/*
* Copyright (c) 2016, Gabor Adam Fodor <fogggab@yahoo.com>
* All rights reserved.
*
* License:
*
* This program is provided for scientific and educational purposed only.
* Feel free to use and/or modify it for such purposes, but you are kindly
* asked not to redistribute this or derivative works in source or executable
* form. A license must be obtained from the author of the code for any other use.
*
*/
#include "Shearlet.h"
#include <opencv/cv.h>
#include <vector>
#include <opencv2/opencv.hpp>

/**
* Discrete shearlet transform implementation finite-length filters following the Matlab
* implementation of Glenn R. Easley, Demetrio Labate. and Wang-Q Lim. The matlab code and detailed
* publications on their method can be found at:
* http://www.math.uh.edu/~dlabate/software.html
*
* NOTE: This implementation ONLY works for images of size 512x512, skipping over
* generating the filter-banks on the run to speed up computation speed. Instead,
* all frequency domain filters are stored in a big lookup table that is initalized
* by calling function getFilterBank().
**/

namespace Shearlet{

	//Use it to store precomputed filters
	std::vector<std::vector<cv::Mat>> shear_filter;

	void getFilterBank(int L, std::vector<int> &decomp, std::vector<int> &dsize){
		int level = (decomp).size();

		shear_filter = std::vector<std::vector<cv::Mat>>(level);
		cv::Mat tmp(L, L, CV_32F, cv::Scalar(0));
		
		//Calculate the filter frequency response
		for (int i = 1; i <= level; i++){
			//Based on the filter..
			if (((dsize)[i - 1] == 32) && ((decomp)[i - 1] == 3)){
				shear_filter[i - 1] = std::vector<cv::Mat>(8);

				//Extend fft2
				cv::Mat tmp2;
				for (int k = 0; k < 8; k++){
					//Fill out upscaled matrix
					for (int j = 0; j < 32; j++){
						for (int l = 0; l < 32; l++){
							tmp.at<float>(j, l) = f32_3[k][j][l];
						}
					}

					cv::dft(tmp, tmp2, cv::DFT_COMPLEX_OUTPUT);     //Forward Fourier transform
					shear_filter[i - 1][k] = tmp2.clone() / L;

					//Erase filter
					for (int j = 0; j < 32; j++){
						for (int l = 0; l < 32; l++){
							tmp.at<float>(j, l) = 0;
						}
					}
				}
			}
			else if (((dsize)[i - 1] == 16) && ((decomp)[i - 1] == 4)){
				shear_filter[i - 1] = std::vector<cv::Mat>(16);

				//Extend fft2
				cv::Mat tmp2;
				for (int k = 0; k < 16; k++){
					//Fill out upscaled matrix
					for (int j = 0; j < 16; j++){
						for (int l = 0; l < 16; l++){
							tmp.at<float>(j, l) = f16_4[k][j][l];
						}
					}

					cv::dft(tmp, tmp2, cv::DFT_COMPLEX_OUTPUT);     //Forward Fourier transform
					shear_filter[i - 1][k] = tmp2.clone() / L;

					//Erase filter
					for (int j = 0; j < 16; j++){
						for (int l = 0; l < 16; l++){
							tmp.at<float>(j, l) = 0;
						}
					}
				}
			}
			else if (((dsize)[i - 1] == 32) && ((decomp)[i - 1] == 6)){
				shear_filter[i - 1] = std::vector<cv::Mat>(64);

				//Extend fft2
				cv::Mat tmp2;
				for (int k = 0; k < 64; k++){
					//Fill out upscaled matrix
					for (int j = 0; j < 32; j++){
						for (int l = 0; l < 32; l++){
							tmp.at<float>(j, l) = f32_6[k][j][l];
						}
					}

					cv::dft(tmp, tmp2, cv::DFT_COMPLEX_OUTPUT);     //Forward Fourier transform
					shear_filter[i - 1][k] = tmp2.clone() / L;

					//Erase filter
					for (int j = 0; j < 32; j++){
						for (int l = 0; l < 32; l++){
							tmp.at<float>(j, l) = 0;
						}
					}
				}
			}
			else if ((dsize[i - 1] == 128) && (decomp[i - 1] == 4)){
				shear_filter[i - 1] = std::vector<cv::Mat>(16);

				//Extend fft2
				cv::Mat tmp2;
				for (int k = 0; k < 16; k++){
					//Fill out upscaled matrix
					for (int j = 0; j < 128; j++){
						for (int l = 0; l < 128; l++){
							tmp.at<float>(j, l) = f128_4[k][j][l];
						}
					}

					cv::dft(tmp, tmp2, cv::DFT_COMPLEX_OUTPUT);     //Forward Fourier transform
					shear_filter[i - 1][k] = tmp2.clone() / L;

					//Erase filter
					for (int j = 0; j < 128; j++){
						for (int l = 0; l < 128; l++){
							tmp.at<float>(j, l) = 0;
						}
					}
				}
			}
			else if ((dsize[i - 1] == 64) && (decomp[i - 1] == 4)){
				shear_filter[i - 1] = std::vector<cv::Mat>(16);

				//Extend fft2
				cv::Mat tmp2;
				for (int k = 0; k < 16; k++){
					//Fill out upscaled matrix
					for (int j = 0; j < 64; j++){
						for (int l = 0; l < 64; l++){
							tmp.at<float>(j, l) = f64_4[k][j][l];
						}
					}

					cv::dft(tmp, tmp2, cv::DFT_COMPLEX_OUTPUT);     //Forward Fourier transform
					shear_filter[i - 1][k] = tmp2.clone() / L;

					//Erase filter
					for (int j = 0; j < 64; j++){
						for (int l = 0; l < 64; l++){
							tmp.at<float>(j, l) = 0;
						}
					}
				}
			}
			else if ((dsize[i - 1] == 32) && (decomp[i - 1] == 4)){
				shear_filter[i - 1] = std::vector<cv::Mat>(16);

				//Extend fft2
				cv::Mat tmp2;
				for (int k = 0; k < 16; k++){
					//Fill out upscaled matrix
					for (int j = 0; j < 32; j++){
						for (int l = 0; l < 32; l++){
							tmp.at<float>(j, l) = f32_4[k][j][l];
						}
					}

					cv::dft(tmp, tmp2, cv::DFT_COMPLEX_OUTPUT);     //Forward Fourier transform
					shear_filter[i - 1][k] = tmp2.clone() / L;

					//Erase filter
					for (int j = 0; j < 32; j++){
						for (int l = 0; l < 32; l++){
							tmp.at<float>(j, l) = 0;
						}
					}
				}
			}
			else if ((dsize[i - 1] == 128) && (decomp[i - 1] == 5)){
				shear_filter[i - 1] = std::vector<cv::Mat>(32);

				//Extend fft2
				cv::Mat tmp2;
				for (int k = 0; k < 32; k++){
					//Fill out upscaled matrix
					for (int j = 0; j < 128; j++){
						for (int l = 0; l < 128; l++){
							tmp.at<float>(j, l) = f128_5[k][j][l];
						}
					}

					cv::dft(tmp, tmp2, cv::DFT_COMPLEX_OUTPUT);     //Forward Fourier transform
					shear_filter[i - 1][k] = tmp2.clone() / L;

					//Erase filter
					for (int j = 0; j < 128; j++){
						for (int l = 0; l < 128; l++){
							tmp.at<float>(j, l) = 0;
						}
					}
				}
			}
			else if ((dsize[i - 1] == 128) && (decomp[i - 1] == 6)){
				shear_filter[i - 1] = std::vector<cv::Mat>(64);

				//Extend fft2
				cv::Mat tmp2;
				for (int k = 0; k < 64; k++){
					//Fill out upscaled matrix
					for (int j = 0; j < 128; j++){
						for (int l = 0; l < 128; l++){
							tmp.at<float>(j, l) = f128_6[k][j][l];
						}
					}

					cv::dft(tmp, tmp2, cv::DFT_COMPLEX_OUTPUT);     //Forward Fourier transform
					shear_filter[i - 1][k] = tmp2.clone() / L;

					//Erase filter
					for (int j = 0; j < 128; j++){
						for (int l = 0; l < 128; l++){
							tmp.at<float>(j, l) = 0;
						}
					}
				}
			}
			else{
				std::cout << " Unimplemented combination of parameters for the shearlet decomposition!!" << std::endl;
			}
		}

		//Normalize
		for (int i = 1; i <= level; i++){
			cv::Mat d(L, L, CV_32F, cv::Scalar(0));
			//Get normalization matrix
			for (int j = 0; j < shear_filter[i - 1].size(); j++){
				for (int k = 0; k < L; k++){
					for (int l = 0; l < L; l++){
						cv::Point2f tmp = shear_filter[i - 1][j].at<cv::Point2f>(k, l);
						d.at<float>(k, l) += tmp.x*tmp.x + tmp.y*tmp.y;		//(a+bi)*(a-bi)
					}
				}
			}
			//Normalize
			for (int j = 0; j < shear_filter[i - 1].size(); j++){
				for (int k = 0; k < L; k++){
					for (int l = 0; l < L; l++){
						float val = std::sqrt(d.at<float>(k, l));
						shear_filter[i - 1][j].at<cv::Point2f>(k, l).x /= val;
						shear_filter[i - 1][j].at<cv::Point2f>(k, l).y /= val;
					}
				}
			}
		}
	}

	void nsst_rec2(std::vector<std::vector<cv::Mat>> &dst, std::vector<std::vector<cv::Mat>> &shear_f, cv::Mat &out){
		int level = dst.size() - 1;
		int L = dst[0][0].rows;
		std::vector<cv::Mat> y(level + 1);

		y[0] = dst[0][0];
		cv::Mat tmp2(L, L, CV_32FC2);

		for (int i = 1; i <= level; i++){
			int l = dst[i].size();
			y[i] = cv::Mat(L, L, CV_32F, cv::Scalar(0));
			for (int j = 0; j < l; j++){
				cv::Mat tmp, tmpi;
				
				//Forward fft2
				cv::dft(dst[i][j], tmp, cv::DFT_COMPLEX_OUTPUT);

				for (int k = 0; k < L; k++){
					for (int l = 0; l < L; l++){
						float a, b, c, d;

						a = shear_f[i - 1][j].at<cv::Point2f>(k, l).x;
						b = -shear_f[i - 1][j].at<cv::Point2f>(k, l).y; //Conjugate
						c = tmp.at<cv::Point2f>(k, l).x;
						d = tmp.at<cv::Point2f>(k, l).y;

						tmp2.at<cv::Point2f>(k, l).x = a*c - b*d;
						tmp2.at<cv::Point2f>(k, l).y = a*d + b*c;
					}
				}

				//Inverse fft2
				cv::dft(tmp2, tmpi, cv::DFT_INVERSE | cv::DFT_REAL_OUTPUT);

				//Add it to the sum
				y[i] += tmpi / L / L;	//Stupid normalization..
			}
		}

		//Reconstruct image
		out = atrousrec(y);
	}

	void nsst_dec2(cv::Mat &in, std::vector<int> &decomp, std::vector<int> &dsize, std::vector<std::vector<cv::Mat>> &dst, std::vector<std::vector<cv::Mat>> &shear_f){
		//Output:
		// dst - the array of arrays containing the discrete shearlet
		//                          tranform coefficients
		// shear_f - the array of arrays containing the shearing filters
		
		int L = in.rows;
		int level = decomp.size();

		//LP decomposition
		std::vector<cv::Mat> y = atrousdec(in, level);

		dst = std::vector<std::vector<cv::Mat>>(level + 1);
		//(*shear_f) = std::vector<std::vector<cv::Mat>>(level);

		dst[0] = std::vector<cv::Mat>(1);
		dst[0][0] = y[0].clone(); // assign low-pass coefficients to first decomposition index

		for (int i = 1; i <= level; i++){
			//Precalculate FFT2 of decomposed image
			cv::Mat ffty;

			cv::dft(y[i], ffty, cv::DFT_COMPLEX_OUTPUT);     //Forward Fourier transform
			
			//Get decomposition
			//Initialize (*dst)
			dst[i] = std::vector<cv::Mat>((shear_f)[i - 1].size());
			for (int j = 0; j < (shear_f)[i - 1].size(); j++){

				//Convolution in frequency domain
				cv::Mat tmp(L, L, CV_32FC2);
				for (int k = 0; k < L; k++){
					for (int l = 0; l < L; l++){
						float a, b, c, d;
						a = (shear_f)[i - 1][j].at<cv::Point2f>(k, l).x;
						b = (shear_f)[i - 1][j].at<cv::Point2f>(k, l).y;
						c = ffty.at<cv::Point2f>(k, l).x;
						d = ffty.at<cv::Point2f>(k, l).y;

						tmp.at<cv::Point2f>(k, l).x = a*c - b*d;
						tmp.at<cv::Point2f>(k, l).y = a*d + b*c;
					}
				}

				//Get real of inverse fft
				cv::dft(tmp, dst[i][j], cv::DFT_REAL_OUTPUT | cv::DFT_INVERSE);
				dst[i][j] /= L*L;
			}
		}

		return;
	}

	void symextfilter(cv::Mat &in, cv::Mat &filter, cv::Mat &out){
		int hf = std::floor((filter).rows *1.0 / 2);

		cv::Mat tmp, tmp2;
		tmp = symext(in, filter, 1, 1);
		cv::filter2D(tmp, tmp2, CV_32F, filter, cv::Point(-1, -1), 0, cv::BORDER_REFLECT);
		
		out = cv::Mat(tmp2.rows - 2 * hf, tmp2.cols - 2 * hf, CV_32F);	//The sloppy way to initialize it..
		for (int i = hf; i < tmp2.rows - hf; i++){
			for (int j = hf; j < tmp2.cols - hf; j++){
				(out).at<float>(i - hf, j - hf) = tmp2.at<float>(i, j);
			}
		}
	}

	cv::Mat atrousrec(std::vector<cv::Mat> &y){
		cv::Mat out = y[0].clone(), y1;
		int level = y.size() - 1;

		for (int i = level - 1; i >= 1; i--){
			y1 = y[level - i];
			int shift = -std::exp2(i - 1) + 2;
			int L = std::exp2(i);

			cv::Mat tmp1, tmp2, uf1, uf2;
			//Upsample filters
			uf1 = upsample2df(g0, i);
			uf2 = upsample2df(g1, i);

			tmp1 = symext(out, uf1, shift, shift);
			tmp2 = symext(y1, uf2, shift, shift);

			out = atrousc(tmp1, g0, L) + atrousc(tmp2, g1, L);
		}

		//Reconstruct first level
		cv::Mat tmp1, tmp2;
		symextfilter(out, g0, tmp1);
		symextfilter(y[level], g1, tmp2);

		out = tmp1 + tmp2;

		return out;
	}

	std::vector<cv::Mat> atrousdec(cv::Mat &in, int level){
		std::vector<cv::Mat> y(level + 1);
		cv::Mat y0, y1, x, tmp, ttmp;

		//First level
		symextfilter(in, h0, y0);
		symextfilter(in, h1, y1);

		y[level] = y1.clone();
		x = y0;

		//Remaining levels
		for (int i = 1; i <= level - 1; i++){

			int s1, s2, L;
			s1 = s2 = (int)(-std::exp2(i-1)+2);
			
			L = 1;
			for (int j = 1; j <= i; j++){
				L *= 2;
			} // L = 2^i; (avoiding std::exp2())

            ttmp = upsample2df(h0, i);
			tmp = symext(x, ttmp, s1, s2);
			y0 = atrousc(tmp, h0, L);
            
            ttmp = upsample2df(h1, i);
			tmp = symext(x, ttmp, s1, s2);
			y1 = atrousc(tmp, h1, L);
			y[level - i] = y1.clone();

			x = y0;
		}

		y[0] = x.clone();
		return y;
	}

	cv::Mat upsample2df(cv::Mat &filter, int power){
		int p = 1;
		for (int i = 1; i <= power; i++){
			p *= 2;
		}

		cv::Mat out(p*(filter).rows, p*(filter).cols, CV_32F, cv::Scalar(0));
		for (int i = 0; i < filter.rows; i++){
			for (int j = 0; j < filter.rows; j++){
				out.at<float>(i*p, j*p) = filter.at<float>(i, j);
			}
		}
		return out;
	}

	cv::Mat symext(cv::Mat &in, cv::Mat &filter, int s1, int s2){
		int m = (in).rows;
		int n = (in).cols;
		int p = (filter).rows;
		int q = (filter).cols;
		int p2 = std::floor(p / 2 + 0.1);
		int q2 = std::floor(q / 2 + 0.1);

		int ss = p2 - s1 + 1;
		int rr = q2 - s2 + 1;

		cv::Mat out(m + rr + q + s2, n + ss + p + s1, CV_32F);

		//Horizontal fill
		for (int i = 0; i < n; i++){
			for (int j = 0; j < m; j++){
				//Fill out middle
				out.at<float>(j + rr, i + ss) = (in).at <float>(j, i);
			}
		}
		for (int i = 0; i < ss; i++){
			for (int j = 0; j < m; j++){
				//Left side
				out.at<float>(j + rr, i) = (in).at < float>(j, ss - i - 1);
			}
		}
		for (int i = 0; i < p + s1; i++){
			for (int j = 0; j < m; j++){
				//Left side
				out.at<float>(j + rr, ss + n + i) = (in).at <float>(j, n - i - 1);
			}
		}

		//Vertical fill
		for (int i = 0; i < out.cols; i++){
			for (int j = 0; j < rr; j++){
				//Fill out middle
				out.at<float>(j, i) = (out).at <float>(rr + (rr - j - 1), i);
			}
		}
		for (int i = 0; i < out.cols; i++){
			for (int j = 0; j < q+s2; j++){
				//Fill out middle	
				out.at<float>(j + rr + m, i) = (out).at <float>(rr + m - j - 1, i);
			}
		}

		//Resize, for whatever kinky reason
		cv::Mat truesol(m+p-1, n+q-1, CV_32F);
		for (int i = 0; i < m + p - 1; i++){
			for (int j = 0; j < n + q - 1; j++){
				truesol.at<float>(i, j) = out.at<float>(i, j);
			}
		}
		
		return truesol;
	}

	cv::Mat atrousc(cv::Mat &S, cv::Mat &F, int L){

		int SColLength, SRowLength, FColLength, FRowLength, O_SColLength, O_SRowLength;
		int SFColLength, SFRowLength;
		int n1, n2, k1, k2, f1, f2, kk2, kk1, ind;
		double sum;
		int M0, M3, sM0, sM3;

		SColLength = S.cols;
		SRowLength = S.rows;
		FColLength = F.cols;
		FRowLength = F.rows;

		//Speed up the bastard
		double* f = new double[(F).cols*(F).rows];
		double* s = new double[(S).cols*(S).rows];

		ind = 0;
		for (n1 = 0; n1 < SRowLength; n1++){
			for (n2 = 0; n2 < SColLength; n2++){
				s[ind] = (S).at<float>(n1, n2);
				ind++;
			}
		}
		ind = 0;
		for (n1 = 0; n1 < FRowLength; n1++){
			for (n2 = 0; n2 < FColLength; n2++){
				f[ind] = (F).at<float>(n1, n2);
				ind++;
			}
		}

		SFColLength = FColLength - 1;
		SFRowLength = FRowLength - 1;

		M0 = M3 = L;
		sM0 = M0 - 1;
		sM3 = M3 - 1;

		O_SColLength = SColLength - M0*FColLength + 1;
		O_SRowLength = SRowLength - M3*FRowLength + 1;

		cv::Mat out(O_SColLength, O_SRowLength, CV_32F);

		/* Convolution loop */
		for (n1 = 0; n1 < O_SRowLength; n1++){
			for (n2 = 0; n2 < O_SColLength; n2++){
				sum = 0;
				kk1 = n1 + sM0;
				for (k1 = 0; k1 < FRowLength; k1++){
					kk2 = n2 + sM3;
					for (k2 = 0; k2 < FColLength; k2++){
						f1 = SFRowLength - k1; /* flipped index */
						f2 = SFColLength - k2;
						//sum += (*F).at<float>(f1, f2) * (*S).at<float>(kk1, kk2);
						sum += f[f1*FColLength + f2] * s[kk1*SColLength + kk2];
						kk2 += M3;
					}
					kk1 += M0;
				}
				(out).at<float>(n1, n2) = sum;
			}
		}

		//Deallocate vectors
		delete[] f;
		delete[] s;

		return out;
	}

	//Constant defitions
	//Filters for LP decomposition
	cv::Mat g0 = (cv::Mat_<float>(19, 19) <<
		-6.39158767662235e-10, 0, 1.72572867268803e-08, 3.06796208477873e-08, -1.38058293815043e-07, -5.52233175260171e-07, -3.37475829325660e-07, 1.93281611341060e-06, 5.69490461987051e-06, 7.64945213138162e-06, 5.69490461987051e-06, 1.93281611341060e-06, -3.37475829325660e-07, -5.52233175260171e-07, -1.38058293815043e-07, 3.06796208477873e-08, 1.72572867268803e-08, 0, -6.39158767662235e-10,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		1.72572867268803e-08, 0, -4.65946741625769e-07, -8.28349762890256e-07, 3.72757393300615e-06, 1.49102957320246e-05, 9.11184739179282e-06, -5.21860350620861e-05, -0.000153762424736504, -0.000206535207547304, -0.000153762424736504, -5.21860350620861e-05, 9.11184739179282e-06, 1.49102957320246e-05, 3.72757393300615e-06, -8.28349762890256e-07, -4.65946741625769e-07, 0, 1.72572867268803e-08,
		3.06796208477873e-08, 0, -8.28349762890256e-07, -1.28092360544931e-06, 6.62679810312205e-06, 2.30566248980877e-05, 1.00644975598085e-05, -8.06981871433068e-05, -0.000218146341523376, -0.000286660460303639, -0.000218146341523376, -8.06981871433068e-05, 1.00644975598085e-05, 2.30566248980877e-05, 6.62679810312205e-06, -1.28092360544931e-06, -8.28349762890256e-07, 0, 3.06796208477873e-08,
		-1.38058293815043e-07, 0, 3.72757393300615e-06, 6.62679810312205e-06, -2.98205914640492e-05, -0.000119282365856197, -7.28947791343425e-05, 0.000417488280496689, 0.00123009939789203, 0.00165228166037843, 0.00123009939789203, 0.000417488280496689, -7.28947791343425e-05, -0.000119282365856197, -2.98205914640492e-05, 6.62679810312205e-06, 3.72757393300615e-06, 0, -1.38058293815043e-07,
		-5.52233175260171e-07, 0, 1.49102957320246e-05, 2.30566248980877e-05, -0.000119282365856197, -0.000415019248165578, -0.000181160956076553, 0.00145256736857952, 0.00392663414742077, 0.00515988828546550, 0.00392663414742077, 0.00145256736857952, -0.000181160956076553, -0.000415019248165578, -0.000119282365856197, 2.30566248980877e-05, 1.49102957320246e-05, 0, -5.52233175260171e-07,
		-3.37475829325660e-07, 0, 9.11184739179282e-06, 1.00644975598085e-05, -7.28947791343425e-05, -0.000181160956076553, 0.00146858180607625, 0.000634063346267936, -0.0118140117563501, -0.0217450344911939, -0.0118140117563501, 0.000634063346267936, 0.00146858180607625, -0.000181160956076553, -7.28947791343425e-05, 1.00644975598085e-05, 9.11184739179282e-06, 0, -3.37475829325660e-07,
		1.93281611341060e-06, 0, -5.21860350620861e-05, -8.06981871433068e-05, 0.000417488280496689, 0.00145256736857952, 0.000634063346267936, -0.00508398579002833, -0.0137432195159727, -0.0180596089991292, -0.0137432195159727, -0.00508398579002833, 0.000634063346267936, 0.00145256736857952, 0.000417488280496689, -8.06981871433068e-05, -5.21860350620861e-05, 0, 1.93281611341060e-06,
		5.69490461987051e-06, 0, -0.000153762424736504, -0.000218146341523376, 0.00123009939789203, 0.00392663414742077, -0.0118140117563501, -0.0137432195159727, 0.0826466923977296, 0.163898888458460, 0.0826466923977296, -0.0137432195159727, -0.0118140117563501, 0.00392663414742077, 0.00123009939789203, -0.000218146341523376, -0.000153762424736504, 0, 5.69490461987051e-06,
		7.64945213138162e-06, 0, -0.000206535207547304, -0.000286660460303639, 0.00165228166037843, 0.00515988828546550, -0.0217450344911939, -0.0180596089991292, 0.163898888458460, 0.313587262092392, 0.163898888458460, -0.0180596089991292, -0.0217450344911939, 0.00515988828546550, 0.00165228166037843, -0.000286660460303639, -0.000206535207547304, 0, 7.64945213138162e-06,
		5.69490461987051e-06, 0, -0.000153762424736504, -0.000218146341523376, 0.00123009939789203, 0.00392663414742077, -0.0118140117563501, -0.0137432195159727, 0.0826466923977296, 0.163898888458460, 0.0826466923977296, -0.0137432195159727, -0.0118140117563501, 0.00392663414742077, 0.00123009939789203, -0.000218146341523376, -0.000153762424736504, 0, 5.69490461987051e-06,
		1.93281611341060e-06, 0, -5.21860350620861e-05, -8.06981871433068e-05, 0.000417488280496689, 0.00145256736857952, 0.000634063346267936, -0.00508398579002833, -0.0137432195159727, -0.0180596089991292, -0.0137432195159727, -0.00508398579002833, 0.000634063346267936, 0.00145256736857952, 0.000417488280496689, -8.06981871433068e-05, -5.21860350620861e-05, 0, 1.93281611341060e-06,
		-3.37475829325660e-07, 0, 9.11184739179282e-06, 1.00644975598085e-05, -7.28947791343425e-05, -0.000181160956076553, 0.00146858180607625, 0.000634063346267936, -0.0118140117563501, -0.0217450344911939, -0.0118140117563501, 0.000634063346267936, 0.00146858180607625, -0.000181160956076553, -7.28947791343425e-05, 1.00644975598085e-05, 9.11184739179282e-06, 0, -3.37475829325660e-07,
		-5.52233175260171e-07, 0, 1.49102957320246e-05, 2.30566248980877e-05, -0.000119282365856197, -0.000415019248165578, -0.000181160956076553, 0.00145256736857952, 0.00392663414742077, 0.00515988828546550, 0.00392663414742077, 0.00145256736857952, -0.000181160956076553, -0.000415019248165578, -0.000119282365856197, 2.30566248980877e-05, 1.49102957320246e-05, 0, -5.52233175260171e-07,
		-1.38058293815043e-07, 0, 3.72757393300615e-06, 6.62679810312205e-06, -2.98205914640492e-05, -0.000119282365856197, -7.28947791343425e-05, 0.000417488280496689, 0.00123009939789203, 0.00165228166037843, 0.00123009939789203, 0.000417488280496689, -7.28947791343425e-05, -0.000119282365856197, -2.98205914640492e-05, 6.62679810312205e-06, 3.72757393300615e-06, 0, -1.38058293815043e-07,
		3.06796208477873e-08, 0, -8.28349762890256e-07, -1.28092360544931e-06, 6.62679810312205e-06, 2.30566248980877e-05, 1.00644975598085e-05, -8.06981871433068e-05, -0.000218146341523376, -0.000286660460303639, -0.000218146341523376, -8.06981871433068e-05, 1.00644975598085e-05, 2.30566248980877e-05, 6.62679810312205e-06, -1.28092360544931e-06, -8.28349762890256e-07, 0, 3.06796208477873e-08,
		1.72572867268803e-08, 0, -4.65946741625769e-07, -8.28349762890256e-07, 3.72757393300615e-06, 1.49102957320246e-05, 9.11184739179282e-06, -5.21860350620861e-05, -0.000153762424736504, -0.000206535207547304, -0.000153762424736504, -5.21860350620861e-05, 9.11184739179282e-06, 1.49102957320246e-05, 3.72757393300615e-06, -8.28349762890256e-07, -4.65946741625769e-07, 0, 1.72572867268803e-08,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		-6.39158767662235e-10, 0, 1.72572867268803e-08, 3.06796208477873e-08, -1.38058293815043e-07, -5.52233175260171e-07, -3.37475829325660e-07, 1.93281611341060e-06, 5.69490461987051e-06, 7.64945213138162e-06, 5.69490461987051e-06, 1.93281611341060e-06, -3.37475829325660e-07, -5.52233175260171e-07, -1.38058293815043e-07, 3.06796208477873e-08, 1.72572867268803e-08, 0, -6.39158767662235e-10);

	cv::Mat g1 = (cv::Mat_<float>(13, 13) <<
		-7.90049671884718e-07, 0, 1.42208940939249e-05, 2.52815895003110e-05, -4.97731293287372e-05, -0.000227534305502799, -0.000331820862191582, -0.000227534305502799, -4.97731293287372e-05, 2.52815895003110e-05, 1.42208940939249e-05, 0, -7.90049671884718e-07,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		1.42208940939249e-05, 0, -0.000255976093690649, -0.000455068611005598, 0.000895916327917271, 0.00409561749905038, 0.00597277551944847, 0.00409561749905038, 0.000895916327917271, -0.000455068611005598, -0.000255976093690649, 0, 1.42208940939249e-05,
		2.52815895003110e-05, 0, -0.000455068611005598, -0.000976562500000000, 0.00159274013851959, 0.00878906250000000, 0.0132990937659714, 0.00878906250000000, 0.00159274013851959, -0.000976562500000000, -0.000455068611005598, 0, 2.52815895003110e-05,
		-4.97731293287372e-05, 0, 0.000895916327917271, 0.00159274013851959, -0.00313570714771045, -0.0143346612466763, -0.0209047143180696, -0.0143346612466763, -0.00313570714771045, 0.00159274013851959, 0.000895916327917271, 0, -4.97731293287372e-05,
		-0.000227534305502799, 0, 0.00409561749905038, 0.00878906250000000, -0.0143346612466763, -0.0791015625000000, -0.119691843893743, -0.0791015625000000, -0.0143346612466763, 0.00878906250000000, 0.00409561749905038, 0, -0.000227534305502799,
		-0.000331820862191582, 0, 0.00597277551944847, 0.0132990937659714, -0.0209047143180696, -0.119691843893743, 0.817742019066083, -0.119691843893743, -0.0209047143180696, 0.0132990937659714, 0.00597277551944847, 0, -0.000331820862191582,
		-0.000227534305502799, 0, 0.00409561749905038, 0.00878906250000000, -0.0143346612466763, -0.0791015625000000, -0.119691843893743, -0.0791015625000000, -0.0143346612466763, 0.00878906250000000, 0.00409561749905038, 0, -0.000227534305502799,
		-4.97731293287372e-05, 0, 0.000895916327917271, 0.00159274013851959, -0.00313570714771045, -0.0143346612466763, -0.0209047143180696, -0.0143346612466763, -0.00313570714771045, 0.00159274013851959, 0.000895916327917271, 0, -4.97731293287372e-05,
		2.52815895003110e-05, 0, -0.000455068611005598, -0.000976562500000000, 0.00159274013851959, 0.00878906250000000, 0.0132990937659714, 0.00878906250000000, 0.00159274013851959, -0.000976562500000000, -0.000455068611005598, 0, 2.52815895003110e-05,
		1.42208940939249e-05, 0, -0.000255976093690649, -0.000455068611005598, 0.000895916327917271, 0.00409561749905038, 0.00597277551944847, 0.00409561749905038, 0.000895916327917271, -0.000455068611005598, -0.000255976093690649, 0, 1.42208940939249e-05,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		-7.90049671884718e-07, 0, 1.42208940939249e-05, 2.52815895003110e-05, -4.97731293287372e-05, -0.000227534305502799, -0.000331820862191582, -0.000227534305502799, -4.97731293287372e-05, 2.52815895003110e-05, 1.42208940939249e-05, 0, -7.90049671884718e-07);

	cv::Mat h1 = (cv::Mat_<float>(19, 19) <<
		6.39158767662235e-10, 0, -1.72572867268803e-08, -3.06796208477873e-08, 1.38058293815043e-07, 5.52233175260171e-07, 3.37475829325660e-07, -1.93281611341060e-06, -5.69490461987051e-06, -7.64945213138162e-06, -5.69490461987051e-06, -1.93281611341060e-06, 3.37475829325660e-07, 5.52233175260171e-07, 1.38058293815043e-07, -3.06796208477873e-08, -1.72572867268803e-08, 0, 6.39158767662235e-10,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		-1.72572867268803e-08, 0, 4.65946741625769e-07, 8.28349762890256e-07, -3.72757393300615e-06, -1.49102957320246e-05, -9.11184739179282e-06, 5.21860350620861e-05, 0.000153762424736504, 0.000206535207547304, 0.000153762424736504, 5.21860350620861e-05, -9.11184739179282e-06, -1.49102957320246e-05, -3.72757393300615e-06, 8.28349762890256e-07, 4.65946741625769e-07, 0, -1.72572867268803e-08,
		-3.06796208477873e-08, 0, 8.28349762890256e-07, -2.99175738320122e-07, -6.62679810312205e-06, 5.38516328976220e-06, 4.04986814408135e-05, -1.88480715141677e-05, -0.000236922269482222, -0.000376981264079525, -0.000236922269482222, -1.88480715141677e-05, 4.04986814408135e-05, 5.38516328976220e-06, -6.62679810312205e-06, -2.99175738320122e-07, 8.28349762890256e-07, 0, -3.06796208477873e-08,
		1.38058293815043e-07, 0, -3.72757393300615e-06, -6.62679810312205e-06, 2.98205914640492e-05, 0.000119282365856197, 7.28947791343425e-05, -0.000417488280496689, -0.00123009939789203, -0.00165228166037843, -0.00123009939789203, -0.000417488280496689, 7.28947791343425e-05, 0.000119282365856197, 2.98205914640492e-05, -6.62679810312205e-06, -3.72757393300615e-06, 0, 1.38058293815043e-07,
		5.52233175260171e-07, 0, -1.49102957320246e-05, 5.38516328976220e-06, 0.000119282365856197, -9.69329392157196e-05, -0.000728976265934642, 0.000339265287255018, 0.00426460085067999, 0.00678566275343144, 0.00426460085067999, 0.000339265287255018, -0.000728976265934642, -9.69329392157196e-05, 0.000119282365856197, 5.38516328976220e-06, -1.49102957320246e-05, 0, 5.52233175260171e-07,
		3.37475829325660e-07, 0, -9.11184739179282e-06, 4.04986814408135e-05, 7.28947791343425e-05, -0.000728976265934642, -0.00146858180607625, 0.00255141693077125, 0.0118140117563501, 0.0170932220231367, 0.0118140117563501, 0.00255141693077125, -0.00146858180607625, -0.000728976265934642, 7.28947791343425e-05, 4.04986814408135e-05, -9.11184739179282e-06, 0, 3.37475829325660e-07,
		-1.93281611341060e-06, 0, 5.21860350620861e-05, -1.88480715141677e-05, -0.000417488280496689, 0.000339265287255018, 0.00255141693077125, -0.00118742850539256, -0.0149261029773800, -0.0237498196370100, -0.0149261029773800, -0.00118742850539256, 0.00255141693077125, 0.000339265287255018, -0.000417488280496689, -1.88480715141677e-05, 5.21860350620861e-05, 0, -1.93281611341060e-06,
		-5.69490461987051e-06, 0, 0.000153762424736504, -0.000236922269482222, -0.00123009939789203, 0.00426460085067999, 0.0118140117563501, -0.0149261029773800, -0.0826466923977296, -0.122032576245945, -0.0826466923977296, -0.0149261029773800, 0.0118140117563501, 0.00426460085067999, -0.00123009939789203, -0.000236922269482222, 0.000153762424736504, 0, -5.69490461987051e-06,
		-7.64945213138162e-06, 0, 0.000206535207547304, -0.000376981264079525, -0.00165228166037843, 0.00678566275343144, 0.0170932220231367, -0.0237498196370100, -0.122032576245945, 0.821896776039774, -0.122032576245945, -0.0237498196370100, 0.0170932220231367, 0.00678566275343144, -0.00165228166037843, -0.000376981264079525, 0.000206535207547304, 0, -7.64945213138162e-06,
		-5.69490461987051e-06, 0, 0.000153762424736504, -0.000236922269482222, -0.00123009939789203, 0.00426460085067999, 0.0118140117563501, -0.0149261029773800, -0.0826466923977296, -0.122032576245945, -0.0826466923977296, -0.0149261029773800, 0.0118140117563501, 0.00426460085067999, -0.00123009939789203, -0.000236922269482222, 0.000153762424736504, 0, -5.69490461987051e-06,
		-1.93281611341060e-06, 0, 5.21860350620861e-05, -1.88480715141677e-05, -0.000417488280496689, 0.000339265287255018, 0.00255141693077125, -0.00118742850539256, -0.0149261029773800, -0.0237498196370100, -0.0149261029773800, -0.00118742850539256, 0.00255141693077125, 0.000339265287255018, -0.000417488280496689, -1.88480715141677e-05, 5.21860350620861e-05, 0, -1.93281611341060e-06,
		3.37475829325660e-07, 0, -9.11184739179282e-06, 4.04986814408135e-05, 7.28947791343425e-05, -0.000728976265934642, -0.00146858180607625, 0.00255141693077125, 0.0118140117563501, 0.0170932220231367, 0.0118140117563501, 0.00255141693077125, -0.00146858180607625, -0.000728976265934642, 7.28947791343425e-05, 4.04986814408135e-05, -9.11184739179282e-06, 0, 3.37475829325660e-07,
		5.52233175260171e-07, 0, -1.49102957320246e-05, 5.38516328976220e-06, 0.000119282365856197, -9.69329392157196e-05, -0.000728976265934642, 0.000339265287255018, 0.00426460085067999, 0.00678566275343144, 0.00426460085067999, 0.000339265287255018, -0.000728976265934642, -9.69329392157196e-05, 0.000119282365856197, 5.38516328976220e-06, -1.49102957320246e-05, 0, 5.52233175260171e-07,
		1.38058293815043e-07, 0, -3.72757393300615e-06, -6.62679810312205e-06, 2.98205914640492e-05, 0.000119282365856197, 7.28947791343425e-05, -0.000417488280496689, -0.00123009939789203, -0.00165228166037843, -0.00123009939789203, -0.000417488280496689, 7.28947791343425e-05, 0.000119282365856197, 2.98205914640492e-05, -6.62679810312205e-06, -3.72757393300615e-06, 0, 1.38058293815043e-07,
		-3.06796208477873e-08, 0, 8.28349762890256e-07, -2.99175738320122e-07, -6.62679810312205e-06, 5.38516328976220e-06, 4.04986814408135e-05, -1.88480715141677e-05, -0.000236922269482222, -0.000376981264079525, -0.000236922269482222, -1.88480715141677e-05, 4.04986814408135e-05, 5.38516328976220e-06, -6.62679810312205e-06, -2.99175738320122e-07, 8.28349762890256e-07, 0, -3.06796208477873e-08,
		-1.72572867268803e-08, 0, 4.65946741625769e-07, 8.28349762890256e-07, -3.72757393300615e-06, -1.49102957320246e-05, -9.11184739179282e-06, 5.21860350620861e-05, 0.000153762424736504, 0.000206535207547304, 0.000153762424736504, 5.21860350620861e-05, -9.11184739179282e-06, -1.49102957320246e-05, -3.72757393300615e-06, 8.28349762890256e-07, 4.65946741625769e-07, 0, -1.72572867268803e-08,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		6.39158767662235e-10, 0, -1.72572867268803e-08, -3.06796208477873e-08, 1.38058293815043e-07, 5.52233175260171e-07, 3.37475829325660e-07, -1.93281611341060e-06, -5.69490461987051e-06, -7.64945213138162e-06, -5.69490461987051e-06, -1.93281611341060e-06, 3.37475829325660e-07, 5.52233175260171e-07, 1.38058293815043e-07, -3.06796208477873e-08, -1.72572867268803e-08, 0, 6.39158767662235e-10);

	cv::Mat h0 = (cv::Mat_<float>(13, 13) <<
		-7.90049671884718e-07, 0, 1.42208940939249e-05, 2.52815895003110e-05, -4.97731293287372e-05, -0.000227534305502799, -0.000331820862191582, -0.000227534305502799, -4.97731293287372e-05, 2.52815895003110e-05, 1.42208940939249e-05, 0, -7.90049671884718e-07,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		1.42208940939249e-05, 0, -0.000255976093690649, -0.000455068611005598, 0.000895916327917271, 0.00409561749905038, 0.00597277551944847, 0.00409561749905038, 0.000895916327917271, -0.000455068611005598, -0.000255976093690649, 0, 1.42208940939249e-05,
		2.52815895003110e-05, 0, -0.000455068611005598, 0.000976562500000000, 0.00159274013851959, -0.00878906250000000, -0.0179509062340286, -0.00878906250000000, 0.00159274013851959, 0.000976562500000000, -0.000455068611005598, 0, 2.52815895003110e-05,
		-4.97731293287372e-05, 0, 0.000895916327917271, 0.00159274013851959, -0.00313570714771045, -0.0143346612466763, -0.0209047143180696, -0.0143346612466763, -0.00313570714771045, 0.00159274013851959, 0.000895916327917271, 0, -4.97731293287372e-05,
		-0.000227534305502799, 0, 0.00409561749905038, -0.00878906250000000, -0.0143346612466763, 0.0791015625000000, 0.161558156106257, 0.0791015625000000, -0.0143346612466763, -0.00878906250000000, 0.00409561749905038, 0, -0.000227534305502799,
		-0.000331820862191582, 0, 0.00597277551944847, -0.0179509062340286, -0.0209047143180696, 0.161558156106257, 0.317742019066083, 0.161558156106257, -0.0209047143180696, -0.0179509062340286, 0.00597277551944847, 0, -0.000331820862191582,
		-0.000227534305502799, 0, 0.00409561749905038, -0.00878906250000000, -0.0143346612466763, 0.0791015625000000, 0.161558156106257, 0.0791015625000000, -0.0143346612466763, -0.00878906250000000, 0.00409561749905038, 0, -0.000227534305502799,
		-4.97731293287372e-05, 0, 0.000895916327917271, 0.00159274013851959, -0.00313570714771045, -0.0143346612466763, -0.0209047143180696, -0.0143346612466763, -0.00313570714771045, 0.00159274013851959, 0.000895916327917271, 0, -4.97731293287372e-05,
		2.52815895003110e-05, 0, -0.000455068611005598, 0.000976562500000000, 0.00159274013851959, -0.00878906250000000, -0.0179509062340286, -0.00878906250000000, 0.00159274013851959, 0.000976562500000000, -0.000455068611005598, 0, 2.52815895003110e-05,
		1.42208940939249e-05, 0, -0.000255976093690649, -0.000455068611005598, 0.000895916327917271, 0.00409561749905038, 0.00597277551944847, 0.00409561749905038, 0.000895916327917271, -0.000455068611005598, -0.000255976093690649, 0, 1.42208940939249e-05,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		-7.90049671884718e-07, 0, 1.42208940939249e-05, 2.52815895003110e-05, -4.97731293287372e-05, -0.000227534305502799, -0.000331820862191582, -0.000227534305502799, -4.97731293287372e-05, 2.52815895003110e-05, 1.42208940939249e-05, 0, -7.90049671884718e-07);


	//DFT2 of the Myer Shearing Filters for fixed size/comp pairs, normalized
	//Myer filters for fixed decomposition levels/sizes (f[size]_[dcomp])
	float f16_4[16][16][16] = {
		{ { -0.00181071, -0.00008226, 0.00056328, 0.00119464, -0.00018311, -0.00038204, 0.00102364, -0.00048620, -0.00050863, -0.00048620, 0.00102364, -0.00038204, -0.00018311, 0.00119464, 0.00056328, -0.00008226, },
		{ 0.00022431, -0.00200128, -0.00000200, -0.00036052, 0.00084460, 0.00071876, -0.00137376, 0.00082036, 0.00025349, -0.00028641, -0.00095340, 0.00031256, 0.00041804, -0.00029220, 0.00107700, 0.00157700, },
		{ 0.00270639, 0.00108673, -0.00270858, 0.00023672, -0.00058592, -0.00050689, 0.00190077, -0.00143114, -0.00022669, 0.00061867, 0.00015229, -0.00080519, -0.00055460, 0.00037763, -0.00019537, 0.00091175, },
		{ 0.00030483, 0.00381072, 0.00314951, -0.00369312, -0.00034137, 0.00051986, -0.00229011, 0.00204943, -0.00024320, -0.00119588, 0.00033333, 0.00037784, -0.00031566, -0.00087885, -0.00010904, -0.00050173, },
		{ -0.00091553, -0.00116115, 0.00381370, 0.00665688, -0.00351969, -0.00272282, 0.00228961, -0.00292645, 0.00054932, 0.00119945, -0.00125022, -0.00023622, 0.00022380, -0.00000103, -0.00070270, -0.00032038, },
		{ -0.00018276, -0.00068931, -0.00279251, 0.00149948, 0.01015428, -0.00034297, -0.00624835, 0.00272163, -0.00146938, -0.00159081, 0.00157863, -0.00049185, -0.00051312, -0.00007327, -0.00003851, -0.00054463, },
		{ -0.00063119, -0.00024991, 0.00017443, -0.00305984, -0.00274129, 0.01085723, 0.00592310, -0.00825274, 0.00034876, 0.00106252, -0.00261224, 0.00062203, 0.00021971, -0.00032808, -0.00019298, -0.00016293, },
		{ -0.00010224, -0.00064244, -0.00077722, 0.00064136, -0.00127253, -0.00644390, 0.00702878, 0.01246640, -0.00610927, -0.00386841, 0.00241315, -0.00180034, -0.00067346, 0.00023516, -0.00001895, -0.00009955, },
		{ -0.00002035, 0.00004592, -0.00026349, -0.00117842, -0.00018311, 0.00130049, -0.00693866, 0.00007615, 0.01527913, 0.00007615, -0.00693866, 0.00130049, -0.00018311, -0.00117842, -0.00026349, 0.00004592, },
		{ -0.00010224, -0.00009955, -0.00001895, 0.00023516, -0.00067346, -0.00180034, 0.00241315, -0.00386841, -0.00610927, 0.01246640, 0.00702878, -0.00644390, -0.00127253, 0.00064136, -0.00077722, -0.00064244, },
		{ -0.00063119, -0.00016293, -0.00019298, -0.00032808, 0.00021971, 0.00062203, -0.00261224, 0.00106252, 0.00034876, -0.00825274, 0.00592310, 0.01085723, -0.00274129, -0.00305984, 0.00017443, -0.00024991, },
		{ -0.00018276, -0.00054463, -0.00003851, -0.00007327, -0.00051312, -0.00049185, 0.00157863, -0.00159081, -0.00146938, 0.00272163, -0.00624835, -0.00034297, 0.01015428, 0.00149948, -0.00279251, -0.00068931, },
		{ -0.00091553, -0.00032038, -0.00070270, -0.00000103, 0.00022380, -0.00023622, -0.00125022, 0.00119945, 0.00054932, -0.00292645, 0.00228961, -0.00272282, -0.00351969, 0.00665688, 0.00381370, -0.00116115, },
		{ 0.00030483, -0.00050173, -0.00010904, -0.00087885, -0.00031566, 0.00037784, 0.00033333, -0.00119588, -0.00024320, 0.00204943, -0.00229011, 0.00051986, -0.00034137, -0.00369312, 0.00314951, 0.00381072, },
		{ 0.00270639, 0.00091175, -0.00019537, 0.00037763, -0.00055460, -0.00080519, 0.00015229, 0.00061867, -0.00022669, -0.00143114, 0.00190077, -0.00050689, -0.00058592, 0.00023672, -0.00270858, 0.00108673, },
		{ 0.00022431, 0.00157700, 0.00107700, -0.00029220, 0.00041804, 0.00031256, -0.00095340, -0.00028641, 0.00025349, 0.00082036, -0.00137376, 0.00071876, 0.00084460, -0.00036052, -0.00000200, -0.00200128, },
		},
		{ { -0.00042725, -0.00014260, -0.00032612, -0.00067472, 0.00006104, -0.00003675, -0.00101665, 0.00012166, 0.00152588, 0.00012166, -0.00101665, -0.00003675, 0.00006104, -0.00067472, -0.00032612, -0.00014260, },
		{ -0.00028423, -0.00056521, -0.00042725, 0.00013132, -0.00033445, -0.00027311, 0.00055599, -0.00041392, -0.00144026, 0.00012166, 0.00118815, 0.00015844, -0.00006587, 0.00026161, -0.00050826, -0.00073695, },
		{ -0.00091553, -0.00021378, -0.00020764, -0.00107217, -0.00042725, 0.00026536, -0.00032612, 0.00007225, 0.00075157, -0.00041392, -0.00101665, 0.00015844, 0.00059839, 0.00043893, 0.00065044, -0.00012789, },
		{ 0.00040630, -0.00154450, -0.00103677, 0.00081512, -0.00011553, -0.00153578, -0.00042725, 0.00031056, -0.00032055, 0.00007225, 0.00055599, -0.00003675, -0.00006587, 0.00043893, 0.00074453, 0.00122103, },
		{ 0.00250244, 0.00259818, -0.00101665, -0.00377256, -0.00091553, 0.00195660, -0.00004009, -0.00181403, -0.00042725, 0.00031056, -0.00032612, -0.00027311, 0.00006104, 0.00026161, 0.00065044, 0.00122103, },
		{ 0.00040630, 0.00259818, 0.00597653, 0.00436075, -0.00314626, -0.00607431, -0.00071995, 0.00270237, -0.00001096, -0.00181403, -0.00042725, 0.00026536, -0.00033445, -0.00067472, -0.00050826, -0.00012789, },
		{ -0.00091553, -0.00154450, -0.00101665, 0.00436075, 0.01026587, 0.00602784, -0.00504138, -0.00753570, -0.00062950, 0.00270237, -0.00004009, -0.00153578, -0.00042725, 0.00013132, -0.00032612, -0.00073695, },
		{ -0.00028423, -0.00021378, -0.00103677, -0.00377256, -0.00314626, 0.00602784, 0.01381931, 0.00704510, -0.00579659, -0.00753570, -0.00071995, 0.00195660, -0.00011553, -0.00107217, -0.00042725, -0.00014260, },
		{ -0.00042725, -0.00056521, -0.00020764, 0.00081512, -0.00091553, -0.00607431, -0.00504138, 0.00704510, 0.01519775, 0.00704510, -0.00504138, -0.00607431, -0.00091553, 0.00081512, -0.00020764, -0.00056521, },
		{ -0.00028423, -0.00014260, -0.00042725, -0.00107217, -0.00011553, 0.00195660, -0.00071995, -0.00753570, -0.00579659, 0.00704510, 0.01381931, 0.00602784, -0.00314626, -0.00377256, -0.00103677, -0.00021378, },
		{ -0.00091553, -0.00073695, -0.00032612, 0.00013132, -0.00042725, -0.00153578, -0.00004009, 0.00270237, -0.00062950, -0.00753570, -0.00504138, 0.00602784, 0.01026587, 0.00436075, -0.00101665, -0.00154450, },
		{ 0.00040630, -0.00012789, -0.00050826, -0.00067472, -0.00033445, 0.00026536, -0.00042725, -0.00181403, -0.00001096, 0.00270237, -0.00071995, -0.00607431, -0.00314626, 0.00436075, 0.00597653, 0.00259818, },
		{ 0.00250244, 0.00122103, 0.00065044, 0.00026161, 0.00006104, -0.00027311, -0.00032612, 0.00031056, -0.00042725, -0.00181403, -0.00004009, 0.00195660, -0.00091553, -0.00377256, -0.00101665, 0.00259818, },
		{ 0.00040630, 0.00122103, 0.00074453, 0.00043893, -0.00006587, -0.00003675, 0.00055599, 0.00007225, -0.00032055, 0.00031056, -0.00042725, -0.00153578, -0.00011553, 0.00081512, -0.00103677, -0.00154450, },
		{ -0.00091553, -0.00012789, 0.00065044, 0.00043893, 0.00059839, 0.00015844, -0.00101665, -0.00041392, 0.00075157, 0.00007225, -0.00032612, 0.00026536, -0.00042725, -0.00107217, -0.00020764, -0.00021378, },
		{ -0.00028423, -0.00073695, -0.00050826, 0.00026161, -0.00006587, 0.00015844, 0.00118815, 0.00012166, -0.00144026, -0.00041392, 0.00055599, -0.00027311, -0.00033445, 0.00013132, -0.00042725, -0.00056521, },
		},
		{ { -0.00002035, -0.00009955, -0.00019298, -0.00007327, 0.00022380, 0.00037784, 0.00015229, -0.00028641, -0.00050863, -0.00028641, 0.00015229, 0.00037784, 0.00022380, -0.00007327, -0.00019298, -0.00009955, },
		{ -0.00010224, 0.00004592, -0.00001895, -0.00032808, -0.00051312, -0.00023622, 0.00033333, 0.00061867, 0.00025349, -0.00048620, -0.00095340, -0.00080519, -0.00031566, -0.00000103, -0.00003851, -0.00016293, },
		{ -0.00063119, -0.00064244, -0.00026349, 0.00023516, 0.00021971, -0.00049185, -0.00125022, -0.00119588, -0.00022669, 0.00082036, 0.00102364, 0.00031256, -0.00055460, -0.00087885, -0.00070270, -0.00054463, },
		{ -0.00018276, -0.00024991, -0.00077722, -0.00117842, -0.00067346, 0.00062203, 0.00157863, 0.00119945, -0.00024320, -0.00143114, -0.00137376, -0.00038204, 0.00041804, 0.00037763, -0.00010904, -0.00032038, },
		{ -0.00091553, -0.00068931, 0.00017443, 0.00064136, -0.00018311, -0.00180034, -0.00261224, -0.00159081, 0.00054932, 0.00204943, 0.00190077, 0.00071876, -0.00018311, -0.00029220, -0.00019537, -0.00050173, },
		{ 0.00030483, -0.00116115, -0.00279251, -0.00305984, -0.00127253, 0.00130049, 0.00241315, 0.00106252, -0.00146938, -0.00292645, -0.00229011, -0.00050689, 0.00084460, 0.00119464, 0.00107700, 0.00091175, },
		{ 0.00270639, 0.00381072, 0.00381370, 0.00149948, -0.00274129, -0.00644390, -0.00693866, -0.00386841, 0.00034876, 0.00272163, 0.00228961, 0.00051986, -0.00058592, -0.00036052, 0.00056328, 0.00157700, },
		{ 0.00022431, 0.00108673, 0.00314951, 0.00665688, 0.01015428, 0.01085723, 0.00702878, 0.00007615, -0.00610927, -0.00825274, -0.00624835, -0.00272282, -0.00034137, 0.00023672, -0.00000200, -0.00008226, },
		{ -0.00181071, -0.00200128, -0.00270858, -0.00369312, -0.00351969, -0.00034297, 0.00592310, 0.01246640, 0.01527913, 0.01246640, 0.00592310, -0.00034297, -0.00351969, -0.00369312, -0.00270858, -0.00200128, },
		{ 0.00022431, -0.00008226, -0.00000200, 0.00023672, -0.00034137, -0.00272282, -0.00624835, -0.00825274, -0.00610927, 0.00007615, 0.00702878, 0.01085723, 0.01015428, 0.00665688, 0.00314951, 0.00108673, },
		{ 0.00270639, 0.00157700, 0.00056328, -0.00036052, -0.00058592, 0.00051986, 0.00228961, 0.00272163, 0.00034876, -0.00386841, -0.00693866, -0.00644390, -0.00274129, 0.00149948, 0.00381370, 0.00381072, },
		{ 0.00030483, 0.00091175, 0.00107700, 0.00119464, 0.00084460, -0.00050689, -0.00229011, -0.00292645, -0.00146938, 0.00106252, 0.00241315, 0.00130049, -0.00127253, -0.00305984, -0.00279251, -0.00116115, },
		{ -0.00091553, -0.00050173, -0.00019537, -0.00029220, -0.00018311, 0.00071876, 0.00190077, 0.00204943, 0.00054932, -0.00159081, -0.00261224, -0.00180034, -0.00018311, 0.00064136, 0.00017443, -0.00068931, },
		{ -0.00018276, -0.00032038, -0.00010904, 0.00037763, 0.00041804, -0.00038204, -0.00137376, -0.00143114, -0.00024320, 0.00119945, 0.00157863, 0.00062203, -0.00067346, -0.00117842, -0.00077722, -0.00024991, },
		{ -0.00063119, -0.00054463, -0.00070270, -0.00087885, -0.00055460, 0.00031256, 0.00102364, 0.00082036, -0.00022669, -0.00119588, -0.00125022, -0.00049185, 0.00021971, 0.00023516, -0.00026349, -0.00064244, },
		{ -0.00010224, -0.00016293, -0.00003851, -0.00000103, -0.00031566, -0.00080519, -0.00095340, -0.00048620, 0.00025349, 0.00061867, 0.00033333, -0.00023622, -0.00051312, -0.00032808, -0.00001895, 0.00004592, },
		},
		{ { -0.00026449, -0.00026449, -0.00026449, -0.00026449, -0.00026449, -0.00026449, -0.00026449, -0.00026449, -0.00026449, -0.00026449, -0.00026449, -0.00026449, -0.00026449, -0.00026449, -0.00026449, -0.00026449, },
		{ -0.00016938, -0.00015676, -0.00012082, -0.00006703, -0.00000358, 0.00005987, 0.00011366, 0.00014960, 0.00016222, 0.00014960, 0.00011366, 0.00005987, -0.00000358, -0.00006703, -0.00012082, -0.00015676, },
		{ 0.00035045, 0.00033506, 0.00029121, 0.00022560, 0.00014820, 0.00007080, 0.00000518, -0.00003866, -0.00005405, -0.00003866, 0.00000518, 0.00007080, 0.00014820, 0.00022560, 0.00029121, 0.00033506, },
		{ -0.00017035, -0.00017762, -0.00019834, -0.00022934, -0.00026591, -0.00030247, -0.00033348, -0.00035419, -0.00036147, -0.00035419, -0.00033348, -0.00030247, -0.00026591, -0.00022934, -0.00019834, -0.00017762, },
		{ -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, },
		{ -0.00039812, -0.00038182, -0.00033543, -0.00026599, -0.00018408, -0.00010217, -0.00003273, 0.00001366, 0.00002996, 0.00001366, -0.00003273, -0.00010217, -0.00018408, -0.00026599, -0.00033543, -0.00038182, },
		{ -0.00218151, -0.00209177, -0.00183624, -0.00145380, -0.00100269, -0.00055158, -0.00016914, 0.00008639, 0.00017612, 0.00008639, -0.00016914, -0.00055158, -0.00100269, -0.00145380, -0.00183624, -0.00209177, },
		{ 0.00098199, 0.00066300, -0.00024539, -0.00160490, -0.00320854, -0.00481219, -0.00617170, -0.00708009, -0.00739907, -0.00708009, -0.00617170, -0.00481219, -0.00320854, -0.00160490, -0.00024539, 0.00066300, },
		{ 0.00526937, 0.00571539, 0.00698554, 0.00888646, 0.01112874, 0.01337103, 0.01527195, 0.01654210, 0.01698812, 0.01654210, 0.01527195, 0.01337103, 0.01112874, 0.00888646, 0.00698554, 0.00571539, },
		{ 0.00098199, 0.00066300, -0.00024539, -0.00160490, -0.00320854, -0.00481219, -0.00617170, -0.00708009, -0.00739907, -0.00708009, -0.00617170, -0.00481219, -0.00320854, -0.00160490, -0.00024539, 0.00066300, },
		{ -0.00218151, -0.00209177, -0.00183624, -0.00145380, -0.00100269, -0.00055158, -0.00016914, 0.00008639, 0.00017612, 0.00008639, -0.00016914, -0.00055158, -0.00100269, -0.00145380, -0.00183624, -0.00209177, },
		{ -0.00039812, -0.00038182, -0.00033543, -0.00026599, -0.00018408, -0.00010217, -0.00003273, 0.00001366, 0.00002996, 0.00001366, -0.00003273, -0.00010217, -0.00018408, -0.00026599, -0.00033543, -0.00038182, },
		{ -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, -0.00042725, },
		{ -0.00017035, -0.00017762, -0.00019834, -0.00022934, -0.00026591, -0.00030247, -0.00033348, -0.00035419, -0.00036147, -0.00035419, -0.00033348, -0.00030247, -0.00026591, -0.00022934, -0.00019834, -0.00017762, },
		{ 0.00035045, 0.00033506, 0.00029121, 0.00022560, 0.00014820, 0.00007080, 0.00000518, -0.00003866, -0.00005405, -0.00003866, 0.00000518, 0.00007080, 0.00014820, 0.00022560, 0.00029121, 0.00033506, },
		{ -0.00016938, -0.00015676, -0.00012082, -0.00006703, -0.00000358, 0.00005987, 0.00011366, 0.00014960, 0.00016222, 0.00014960, 0.00011366, 0.00005987, -0.00000358, -0.00006703, -0.00012082, -0.00015676, },
		},
		{ { -0.00002035, -0.00009955, -0.00019298, -0.00007327, 0.00022380, 0.00037784, 0.00015229, -0.00028641, -0.00050863, -0.00028641, 0.00015229, 0.00037784, 0.00022380, -0.00007327, -0.00019298, -0.00009955, },
		{ -0.00010224, -0.00016293, -0.00003851, -0.00000103, -0.00031566, -0.00080519, -0.00095340, -0.00048620, 0.00025349, 0.00061867, 0.00033333, -0.00023622, -0.00051312, -0.00032808, -0.00001895, 0.00004592, },
		{ -0.00063119, -0.00054463, -0.00070270, -0.00087885, -0.00055460, 0.00031256, 0.00102364, 0.00082036, -0.00022669, -0.00119588, -0.00125022, -0.00049185, 0.00021971, 0.00023516, -0.00026349, -0.00064244, },
		{ -0.00018276, -0.00032038, -0.00010904, 0.00037763, 0.00041804, -0.00038204, -0.00137376, -0.00143114, -0.00024320, 0.00119945, 0.00157863, 0.00062203, -0.00067346, -0.00117842, -0.00077722, -0.00024991, },
		{ -0.00091553, -0.00050173, -0.00019537, -0.00029220, -0.00018311, 0.00071876, 0.00190077, 0.00204943, 0.00054932, -0.00159081, -0.00261224, -0.00180034, -0.00018311, 0.00064136, 0.00017443, -0.00068931, },
		{ 0.00030483, 0.00091175, 0.00107700, 0.00119464, 0.00084460, -0.00050689, -0.00229011, -0.00292645, -0.00146938, 0.00106252, 0.00241315, 0.00130049, -0.00127253, -0.00305984, -0.00279251, -0.00116115, },
		{ 0.00270639, 0.00157700, 0.00056328, -0.00036052, -0.00058592, 0.00051986, 0.00228961, 0.00272163, 0.00034876, -0.00386841, -0.00693866, -0.00644390, -0.00274129, 0.00149948, 0.00381370, 0.00381072, },
		{ 0.00022431, -0.00008226, -0.00000200, 0.00023672, -0.00034137, -0.00272282, -0.00624835, -0.00825274, -0.00610927, 0.00007615, 0.00702878, 0.01085723, 0.01015428, 0.00665688, 0.00314951, 0.00108673, },
		{ -0.00181071, -0.00200128, -0.00270858, -0.00369312, -0.00351969, -0.00034297, 0.00592310, 0.01246640, 0.01527913, 0.01246640, 0.00592310, -0.00034297, -0.00351969, -0.00369312, -0.00270858, -0.00200128, },
		{ 0.00022431, 0.00108673, 0.00314951, 0.00665688, 0.01015428, 0.01085723, 0.00702878, 0.00007615, -0.00610927, -0.00825274, -0.00624835, -0.00272282, -0.00034137, 0.00023672, -0.00000200, -0.00008226, },
		{ 0.00270639, 0.00381072, 0.00381370, 0.00149948, -0.00274129, -0.00644390, -0.00693866, -0.00386841, 0.00034876, 0.00272163, 0.00228961, 0.00051986, -0.00058592, -0.00036052, 0.00056328, 0.00157700, },
		{ 0.00030483, -0.00116115, -0.00279251, -0.00305984, -0.00127253, 0.00130049, 0.00241315, 0.00106252, -0.00146938, -0.00292645, -0.00229011, -0.00050689, 0.00084460, 0.00119464, 0.00107700, 0.00091175, },
		{ -0.00091553, -0.00068931, 0.00017443, 0.00064136, -0.00018311, -0.00180034, -0.00261224, -0.00159081, 0.00054932, 0.00204943, 0.00190077, 0.00071876, -0.00018311, -0.00029220, -0.00019537, -0.00050173, },
		{ -0.00018276, -0.00024991, -0.00077722, -0.00117842, -0.00067346, 0.00062203, 0.00157863, 0.00119945, -0.00024320, -0.00143114, -0.00137376, -0.00038204, 0.00041804, 0.00037763, -0.00010904, -0.00032038, },
		{ -0.00063119, -0.00064244, -0.00026349, 0.00023516, 0.00021971, -0.00049185, -0.00125022, -0.00119588, -0.00022669, 0.00082036, 0.00102364, 0.00031256, -0.00055460, -0.00087885, -0.00070270, -0.00054463, },
		{ -0.00010224, 0.00004592, -0.00001895, -0.00032808, -0.00051312, -0.00023622, 0.00033333, 0.00061867, 0.00025349, -0.00048620, -0.00095340, -0.00080519, -0.00031566, -0.00000103, -0.00003851, -0.00016293, },
		},
		{ { -0.00043887, -0.00015334, -0.00033434, -0.00067917, 0.00006104, -0.00003230, -0.00100843, 0.00013240, 0.00153750, 0.00013240, -0.00100843, -0.00003230, 0.00006104, -0.00067917, -0.00033434, -0.00015334, },
		{ -0.00029497, -0.00074517, -0.00051271, 0.00026161, -0.00006142, 0.00016666, 0.00119889, 0.00013328, -0.00142952, -0.00040570, 0.00056044, -0.00027311, -0.00033890, 0.00012310, -0.00043799, -0.00057683, },
		{ -0.00092375, -0.00013234, 0.00065044, 0.00044338, 0.00060661, 0.00016918, -0.00100503, -0.00040318, 0.00075979, 0.00007669, -0.00032612, 0.00026091, -0.00043547, -0.00108291, -0.00021927, -0.00022452, },
		{ 0.00040185, 0.00122103, 0.00074898, 0.00044715, -0.00005512, -0.00002513, 0.00056673, 0.00008047, -0.00031610, 0.00031056, -0.00043170, -0.00154400, -0.00012628, 0.00080350, -0.00104751, -0.00155272, },
		{ 0.00250244, 0.00122548, 0.00065866, 0.00027235, 0.00007266, -0.00026237, -0.00031790, 0.00031501, -0.00042725, -0.00181848, -0.00004831, 0.00194586, -0.00092715, -0.00378330, -0.00102487, 0.00259374, },
		{ 0.00041075, -0.00011967, -0.00049752, -0.00066310, -0.00032371, 0.00027358, -0.00042280, -0.00181403, -0.00001541, 0.00269415, -0.00073069, -0.00608594, -0.00315700, 0.00435253, 0.00597208, 0.00259818, },
		{ -0.00090731, -0.00072621, -0.00031449, 0.00014206, -0.00041903, -0.00153133, -0.00004009, 0.00269792, -0.00063772, -0.00754644, -0.00505301, 0.00601710, 0.01025765, 0.00435630, -0.00101665, -0.00154005, },
		{ -0.00027349, -0.00013097, -0.00041651, -0.00106395, -0.00011109, 0.00195660, -0.00072440, -0.00754392, -0.00580733, 0.00703347, 0.01380857, 0.00601962, -0.00315071, -0.00377256, -0.00103232, -0.00020556, },
		{ -0.00041562, -0.00055447, -0.00019942, 0.00081957, -0.00091553, -0.00607876, -0.00504960, 0.00703436, 0.01518613, 0.00703436, -0.00504960, -0.00607876, -0.00091553, 0.00081957, -0.00019942, -0.00055447, },
		{ -0.00027349, -0.00020556, -0.00103232, -0.00377256, -0.00315071, 0.00601962, 0.01380857, 0.00703347, -0.00580733, -0.00754392, -0.00072440, 0.00195660, -0.00011109, -0.00106395, -0.00041651, -0.00013097, },
		{ -0.00090731, -0.00154005, -0.00101665, 0.00435630, 0.01025765, 0.00601710, -0.00505301, -0.00754644, -0.00063772, 0.00269792, -0.00004009, -0.00153133, -0.00041903, 0.00014206, -0.00031449, -0.00072621, },
		{ 0.00041075, 0.00259818, 0.00597208, 0.00435253, -0.00315700, -0.00608594, -0.00073069, 0.00269415, -0.00001541, -0.00181403, -0.00042280, 0.00027358, -0.00032371, -0.00066310, -0.00049752, -0.00011967, },
		{ 0.00250244, 0.00259374, -0.00102487, -0.00378330, -0.00092715, 0.00194586, -0.00004831, -0.00181848, -0.00042725, 0.00031501, -0.00031790, -0.00026237, 0.00007266, 0.00027235, 0.00065866, 0.00122548, },
		{ 0.00040185, -0.00155272, -0.00104751, 0.00080350, -0.00012628, -0.00154400, -0.00043170, 0.00031056, -0.00031610, 0.00008047, 0.00056673, -0.00002513, -0.00005512, 0.00044715, 0.00074898, 0.00122103, },
		{ -0.00092375, -0.00022452, -0.00021927, -0.00108291, -0.00043547, 0.00026091, -0.00032612, 0.00007669, 0.00075979, -0.00040318, -0.00100503, 0.00016918, 0.00060661, 0.00044338, 0.00065044, -0.00013234, },
		{ -0.00029497, -0.00057683, -0.00043799, 0.00012310, -0.00033890, -0.00027311, 0.00056044, -0.00040570, -0.00142952, 0.00013328, 0.00119889, 0.00016666, -0.00006142, 0.00026161, -0.00051271, -0.00074517, },
		},
		{ { -0.00191689, -0.00009812, 0.00065014, 0.00122063, -0.00023193, -0.00036199, 0.00109953, -0.00051638, -0.00063031, -0.00051638, 0.00109953, -0.00036199, -0.00023193, 0.00122063, 0.00065014, -0.00009812, },
		{ 0.00020845, 0.00166386, 0.00110299, -0.00034103, 0.00043809, 0.00038846, -0.00098358, -0.00040809, 0.00022331, 0.00089626, -0.00135371, 0.00066993, 0.00087059, -0.00027366, -0.00001786, -0.00210746, },
		{ 0.00279325, 0.00093773, -0.00024420, 0.00039768, -0.00047870, -0.00083537, 0.00003061, 0.00058849, -0.00015079, -0.00141109, 0.00185194, -0.00048090, -0.00049906, 0.00022087, -0.00281476, 0.00107087, },
		{ 0.00033082, -0.00055056, -0.00008899, -0.00080295, -0.00034584, 0.00025616, 0.00030315, -0.00111998, -0.00022315, 0.00200060, -0.00226412, 0.00060672, -0.00035723, -0.00379930, 0.00313365, 0.00389758, },
		{ -0.00096436, -0.00030032, -0.00062680, -0.00003120, 0.00010211, -0.00026640, -0.00117432, 0.00121950, 0.00050049, -0.00290046, 0.00237647, -0.00273867, -0.00362588, 0.00664102, 0.00390056, -0.00113517, },
		{ -0.00016271, -0.00046873, -0.00006869, -0.00019495, -0.00054330, -0.00041595, 0.00159868, -0.00163964, -0.00144339, 0.00280850, -0.00626420, -0.00044915, 0.01013843, 0.00158635, -0.00276653, -0.00073814, },
		{ -0.00055529, -0.00019311, -0.00031466, -0.00035826, 0.00029561, 0.00064208, -0.00266107, 0.00108850, 0.00043562, -0.00826859, 0.00581691, 0.01084138, -0.00265443, -0.00303386, 0.00012560, -0.00022986, },
		{ -0.00013241, -0.00022123, -0.00004912, 0.00031106, -0.00065341, -0.00184917, 0.00243914, -0.00378155, -0.00612513, 0.01236022, 0.00701293, -0.00635704, -0.00124655, 0.00059253, -0.00075716, -0.00056654, },
		{ -0.00014203, 0.00001574, -0.00018759, -0.00115837, -0.00023193, 0.00132647, -0.00685179, 0.00006029, 0.01517295, 0.00006029, -0.00685179, 0.00132647, -0.00023193, -0.00115837, -0.00018759, 0.00001574, },
		{ -0.00013241, -0.00056654, -0.00075716, 0.00059253, -0.00124655, -0.00635704, 0.00701293, 0.01236022, -0.00612513, -0.00378155, 0.00243914, -0.00184917, -0.00065341, 0.00031106, -0.00004912, -0.00022123, },
		{ -0.00055529, -0.00022986, 0.00012560, -0.00303386, -0.00265443, 0.01084138, 0.00581691, -0.00826859, 0.00043562, 0.00108850, -0.00266107, 0.00064208, 0.00029561, -0.00035826, -0.00031466, -0.00019311, },
		{ -0.00016271, -0.00073814, -0.00276653, 0.00158635, 0.01013843, -0.00044915, -0.00626420, 0.00280850, -0.00144339, -0.00163964, 0.00159868, -0.00041595, -0.00054330, -0.00019495, -0.00006869, -0.00046873, },
		{ -0.00096436, -0.00113517, 0.00390056, 0.00664102, -0.00362588, -0.00273867, 0.00237647, -0.00290046, 0.00050049, 0.00121950, -0.00117432, -0.00026640, 0.00010211, -0.00003120, -0.00062680, -0.00030032, },
		{ 0.00033082, 0.00389758, 0.00313365, -0.00379930, -0.00035723, 0.00060672, -0.00226412, 0.00200060, -0.00022315, -0.00111998, 0.00030315, 0.00025616, -0.00034584, -0.00080295, -0.00008899, -0.00055056, },
		{ 0.00279325, 0.00107087, -0.00281476, 0.00022087, -0.00049906, -0.00048090, 0.00185194, -0.00141109, -0.00015079, 0.00058849, 0.00003061, -0.00083537, -0.00047870, 0.00039768, -0.00024420, 0.00093773, },
		{ 0.00020845, -0.00210746, -0.00001786, -0.00027366, 0.00087059, 0.00066993, -0.00135371, 0.00089626, 0.00022331, -0.00040809, -0.00098358, 0.00038846, 0.00043809, -0.00034103, 0.00110299, 0.00166386, },
		},
		{ { 0.00550498, 0.00026842, -0.00091583, -0.00128896, 0.00015869, -0.00069341, 0.00022410, 0.00000496, -0.00004437, 0.00000496, 0.00022410, -0.00069341, 0.00015869, -0.00128896, -0.00091583, 0.00026842, },
		{ 0.00026842, -0.00089572, -0.00147964, 0.00030838, -0.00088137, 0.00043107, -0.00005110, -0.00004437, 0.00006563, -0.00002004, -0.00044782, -0.00006533, -0.00093663, -0.00097311, -0.00025282, 0.00595100, },
		{ -0.00091583, -0.00147964, 0.00036094, -0.00098310, 0.00056937, -0.00009400, -0.00004437, 0.00012169, -0.00026418, -0.00018200, -0.00032959, -0.00047629, -0.00105884, -0.00121594, 0.00722115, -0.00025282, },
		{ -0.00128896, 0.00030838, -0.00098310, 0.00061793, -0.00011722, -0.00004437, 0.00016459, -0.00047116, 0.00006359, -0.00059385, 0.00002198, -0.00115997, -0.00247432, 0.00912207, -0.00121594, -0.00097311, },
		{ 0.00015869, -0.00088137, 0.00056937, -0.00011722, -0.00004437, 0.00018781, -0.00060945, 0.00025156, -0.00081787, 0.00048233, -0.00126110, -0.00383638, 0.01136436, -0.00247432, -0.00105884, -0.00093663, },
		{ -0.00069341, 0.00043107, -0.00009400, -0.00004437, 0.00018781, -0.00065801, 0.00035328, -0.00096756, 0.00083466, -0.00134683, -0.00509476, 0.01360664, -0.00383638, -0.00115997, -0.00047629, -0.00006533, },
		{ 0.00022410, -0.00005110, -0.00004437, 0.00016459, -0.00060945, 0.00035328, -0.00102012, 0.00102534, -0.00140411, -0.00605788, 0.01550756, -0.00509476, -0.00126110, 0.00002198, -0.00032959, -0.00044782, },
		{ 0.00000496, -0.00004437, 0.00012169, -0.00047116, 0.00025156, -0.00096756, 0.00102534, -0.00142423, -0.00657912, 0.01677772, -0.00605788, -0.00134683, 0.00048233, -0.00059385, -0.00018200, -0.00002004, },
		{ -0.00004437, 0.00006563, -0.00026418, 0.00006359, -0.00081787, 0.00083466, -0.00140411, -0.00657912, 0.01722373, -0.00657912, -0.00140411, 0.00083466, -0.00081787, 0.00006359, -0.00026418, 0.00006563, },
		{ 0.00000496, -0.00002004, -0.00018200, -0.00059385, 0.00048233, -0.00134683, -0.00605788, 0.01677772, -0.00657912, -0.00142423, 0.00102534, -0.00096756, 0.00025156, -0.00047116, 0.00012169, -0.00004437, },
		{ 0.00022410, -0.00044782, -0.00032959, 0.00002198, -0.00126110, -0.00509476, 0.01550756, -0.00605788, -0.00140411, 0.00102534, -0.00102012, 0.00035328, -0.00060945, 0.00016459, -0.00004437, -0.00005110, },
		{ -0.00069341, -0.00006533, -0.00047629, -0.00115997, -0.00383638, 0.01360664, -0.00509476, -0.00134683, 0.00083466, -0.00096756, 0.00035328, -0.00065801, 0.00018781, -0.00004437, -0.00009400, 0.00043107, },
		{ 0.00015869, -0.00093663, -0.00105884, -0.00247432, 0.01136436, -0.00383638, -0.00126110, 0.00048233, -0.00081787, 0.00025156, -0.00060945, 0.00018781, -0.00004437, -0.00011722, 0.00056937, -0.00088137, },
		{ -0.00128896, -0.00097311, -0.00121594, 0.00912207, -0.00247432, -0.00115997, 0.00002198, -0.00059385, 0.00006359, -0.00047116, 0.00016459, -0.00004437, -0.00011722, 0.00061793, -0.00098310, 0.00030838, },
		{ -0.00091583, -0.00025282, 0.00722115, -0.00121594, -0.00105884, -0.00047629, -0.00032959, -0.00018200, -0.00026418, 0.00012169, -0.00004437, -0.00009400, 0.00056937, -0.00098310, 0.00036094, -0.00147964, },
		{ 0.00026842, 0.00595100, -0.00025282, -0.00097311, -0.00093663, -0.00006533, -0.00044782, -0.00002004, 0.00006563, -0.00004437, -0.00005110, 0.00043107, -0.00088137, 0.00030838, -0.00147964, -0.00089572, },
		},
		{ { -0.00191689, 0.00020845, 0.00279325, 0.00033082, -0.00096436, -0.00016271, -0.00055529, -0.00013241, -0.00014203, -0.00013241, -0.00055529, -0.00016271, -0.00096436, 0.00033082, 0.00279325, 0.00020845, },
		{ -0.00009812, 0.00166386, 0.00093773, -0.00055056, -0.00030032, -0.00046873, -0.00019311, -0.00022123, 0.00001574, -0.00056654, -0.00022986, -0.00073814, -0.00113517, 0.00389758, 0.00107087, -0.00210746, },
		{ 0.00065014, 0.00110299, -0.00024420, -0.00008899, -0.00062680, -0.00006869, -0.00031466, -0.00004912, -0.00018759, -0.00075716, 0.00012560, -0.00276653, 0.00390056, 0.00313365, -0.00281476, -0.00001786, },
		{ 0.00122063, -0.00034103, 0.00039768, -0.00080295, -0.00003120, -0.00019495, -0.00035826, 0.00031106, -0.00115837, 0.00059253, -0.00303386, 0.00158635, 0.00664102, -0.00379930, 0.00022087, -0.00027366, },
		{ -0.00023193, 0.00043809, -0.00047870, -0.00034584, 0.00010211, -0.00054330, 0.00029561, -0.00065341, -0.00023193, -0.00124655, -0.00265443, 0.01013843, -0.00362588, -0.00035723, -0.00049906, 0.00087059, },
		{ -0.00036199, 0.00038846, -0.00083537, 0.00025616, -0.00026640, -0.00041595, 0.00064208, -0.00184917, 0.00132647, -0.00635704, 0.01084138, -0.00044915, -0.00273867, 0.00060672, -0.00048090, 0.00066993, },
		{ 0.00109953, -0.00098358, 0.00003061, 0.00030315, -0.00117432, 0.00159868, -0.00266107, 0.00243914, -0.00685179, 0.00701293, 0.00581691, -0.00626420, 0.00237647, -0.00226412, 0.00185194, -0.00135371, },
		{ -0.00051638, -0.00040809, 0.00058849, -0.00111998, 0.00121950, -0.00163964, 0.00108850, -0.00378155, 0.00006029, 0.01236022, -0.00826859, 0.00280850, -0.00290046, 0.00200060, -0.00141109, 0.00089626, },
		{ -0.00063031, 0.00022331, -0.00015079, -0.00022315, 0.00050049, -0.00144339, 0.00043562, -0.00612513, 0.01517295, -0.00612513, 0.00043562, -0.00144339, 0.00050049, -0.00022315, -0.00015079, 0.00022331, },
		{ -0.00051638, 0.00089626, -0.00141109, 0.00200060, -0.00290046, 0.00280850, -0.00826859, 0.01236022, 0.00006029, -0.00378155, 0.00108850, -0.00163964, 0.00121950, -0.00111998, 0.00058849, -0.00040809, },
		{ 0.00109953, -0.00135371, 0.00185194, -0.00226412, 0.00237647, -0.00626420, 0.00581691, 0.00701293, -0.00685179, 0.00243914, -0.00266107, 0.00159868, -0.00117432, 0.00030315, 0.00003061, -0.00098358, },
		{ -0.00036199, 0.00066993, -0.00048090, 0.00060672, -0.00273867, -0.00044915, 0.01084138, -0.00635704, 0.00132647, -0.00184917, 0.00064208, -0.00041595, -0.00026640, 0.00025616, -0.00083537, 0.00038846, },
		{ -0.00023193, 0.00087059, -0.00049906, -0.00035723, -0.00362588, 0.01013843, -0.00265443, -0.00124655, -0.00023193, -0.00065341, 0.00029561, -0.00054330, 0.00010211, -0.00034584, -0.00047870, 0.00043809, },
		{ 0.00122063, -0.00027366, 0.00022087, -0.00379930, 0.00664102, 0.00158635, -0.00303386, 0.00059253, -0.00115837, 0.00031106, -0.00035826, -0.00019495, -0.00003120, -0.00080295, 0.00039768, -0.00034103, },
		{ 0.00065014, -0.00001786, -0.00281476, 0.00313365, 0.00390056, -0.00276653, 0.00012560, -0.00075716, -0.00018759, -0.00004912, -0.00031466, -0.00006869, -0.00062680, -0.00008899, -0.00024420, 0.00110299, },
		{ -0.00009812, -0.00210746, 0.00107087, 0.00389758, -0.00113517, -0.00073814, -0.00022986, -0.00056654, 0.00001574, -0.00022123, -0.00019311, -0.00046873, -0.00030032, -0.00055056, 0.00093773, 0.00166386, },
		},
		{ { -0.00043887, -0.00029497, -0.00092375, 0.00040185, 0.00250244, 0.00041075, -0.00090731, -0.00027349, -0.00041562, -0.00027349, -0.00090731, 0.00041075, 0.00250244, 0.00040185, -0.00092375, -0.00029497, },
		{ -0.00015334, -0.00074517, -0.00013234, 0.00122103, 0.00122548, -0.00011967, -0.00072621, -0.00013097, -0.00055447, -0.00020556, -0.00154005, 0.00259818, 0.00259374, -0.00155272, -0.00022452, -0.00057683, },
		{ -0.00033434, -0.00051271, 0.00065044, 0.00074898, 0.00065866, -0.00049752, -0.00031449, -0.00041651, -0.00019942, -0.00103232, -0.00101665, 0.00597208, -0.00102487, -0.00104751, -0.00021927, -0.00043799, },
		{ -0.00067917, 0.00026161, 0.00044338, 0.00044715, 0.00027235, -0.00066310, 0.00014206, -0.00106395, 0.00081957, -0.00377256, 0.00435630, 0.00435253, -0.00378330, 0.00080350, -0.00108291, 0.00012310, },
		{ 0.00006104, -0.00006142, 0.00060661, -0.00005512, 0.00007266, -0.00032371, -0.00041903, -0.00011109, -0.00091553, -0.00315071, 0.01025765, -0.00315700, -0.00092715, -0.00012628, -0.00043547, -0.00033890, },
		{ -0.00003230, 0.00016666, 0.00016918, -0.00002513, -0.00026237, 0.00027358, -0.00153133, 0.00195660, -0.00607876, 0.00601962, 0.00601710, -0.00608594, 0.00194586, -0.00154400, 0.00026091, -0.00027311, },
		{ -0.00100843, 0.00119889, -0.00100503, 0.00056673, -0.00031790, -0.00042280, -0.00004009, -0.00072440, -0.00504960, 0.01380857, -0.00505301, -0.00073069, -0.00004831, -0.00043170, -0.00032612, 0.00056044, },
		{ 0.00013240, 0.00013328, -0.00040318, 0.00008047, 0.00031501, -0.00181403, 0.00269792, -0.00754392, 0.00703436, 0.00703347, -0.00754644, 0.00269415, -0.00181848, 0.00031056, 0.00007669, -0.00040570, },
		{ 0.00153750, -0.00142952, 0.00075979, -0.00031610, -0.00042725, -0.00001541, -0.00063772, -0.00580733, 0.01518613, -0.00580733, -0.00063772, -0.00001541, -0.00042725, -0.00031610, 0.00075979, -0.00142952, },
		{ 0.00013240, -0.00040570, 0.00007669, 0.00031056, -0.00181848, 0.00269415, -0.00754644, 0.00703347, 0.00703436, -0.00754392, 0.00269792, -0.00181403, 0.00031501, 0.00008047, -0.00040318, 0.00013328, },
		{ -0.00100843, 0.00056044, -0.00032612, -0.00043170, -0.00004831, -0.00073069, -0.00505301, 0.01380857, -0.00504960, -0.00072440, -0.00004009, -0.00042280, -0.00031790, 0.00056673, -0.00100503, 0.00119889, },
		{ -0.00003230, -0.00027311, 0.00026091, -0.00154400, 0.00194586, -0.00608594, 0.00601710, 0.00601962, -0.00607876, 0.00195660, -0.00153133, 0.00027358, -0.00026237, -0.00002513, 0.00016918, 0.00016666, },
		{ 0.00006104, -0.00033890, -0.00043547, -0.00012628, -0.00092715, -0.00315700, 0.01025765, -0.00315071, -0.00091553, -0.00011109, -0.00041903, -0.00032371, 0.00007266, -0.00005512, 0.00060661, -0.00006142, },
		{ -0.00067917, 0.00012310, -0.00108291, 0.00080350, -0.00378330, 0.00435253, 0.00435630, -0.00377256, 0.00081957, -0.00106395, 0.00014206, -0.00066310, 0.00027235, 0.00044715, 0.00044338, 0.00026161, },
		{ -0.00033434, -0.00043799, -0.00021927, -0.00104751, -0.00102487, 0.00597208, -0.00101665, -0.00103232, -0.00019942, -0.00041651, -0.00031449, -0.00049752, 0.00065866, 0.00074898, 0.00065044, -0.00051271, },
		{ -0.00015334, -0.00057683, -0.00022452, -0.00155272, 0.00259374, 0.00259818, -0.00154005, -0.00020556, -0.00055447, -0.00013097, -0.00072621, -0.00011967, 0.00122548, 0.00122103, -0.00013234, -0.00074517, },
		},
		{ { -0.00002035, -0.00010224, -0.00063119, -0.00018276, -0.00091553, 0.00030483, 0.00270639, 0.00022431, -0.00181071, 0.00022431, 0.00270639, 0.00030483, -0.00091553, -0.00018276, -0.00063119, -0.00010224, },
		{ -0.00009955, -0.00016293, -0.00054463, -0.00032038, -0.00050173, 0.00091175, 0.00157700, -0.00008226, -0.00200128, 0.00108673, 0.00381072, -0.00116115, -0.00068931, -0.00024991, -0.00064244, 0.00004592, },
		{ -0.00019298, -0.00003851, -0.00070270, -0.00010904, -0.00019537, 0.00107700, 0.00056328, -0.00000200, -0.00270858, 0.00314951, 0.00381370, -0.00279251, 0.00017443, -0.00077722, -0.00026349, -0.00001895, },
		{ -0.00007327, -0.00000103, -0.00087885, 0.00037763, -0.00029220, 0.00119464, -0.00036052, 0.00023672, -0.00369312, 0.00665688, 0.00149948, -0.00305984, 0.00064136, -0.00117842, 0.00023516, -0.00032808, },
		{ 0.00022380, -0.00031566, -0.00055460, 0.00041804, -0.00018311, 0.00084460, -0.00058592, -0.00034137, -0.00351969, 0.01015428, -0.00274129, -0.00127253, -0.00018311, -0.00067346, 0.00021971, -0.00051312, },
		{ 0.00037784, -0.00080519, 0.00031256, -0.00038204, 0.00071876, -0.00050689, 0.00051986, -0.00272282, -0.00034297, 0.01085723, -0.00644390, 0.00130049, -0.00180034, 0.00062203, -0.00049185, -0.00023622, },
		{ 0.00015229, -0.00095340, 0.00102364, -0.00137376, 0.00190077, -0.00229011, 0.00228961, -0.00624835, 0.00592310, 0.00702878, -0.00693866, 0.00241315, -0.00261224, 0.00157863, -0.00125022, 0.00033333, },
		{ -0.00028641, -0.00048620, 0.00082036, -0.00143114, 0.00204943, -0.00292645, 0.00272163, -0.00825274, 0.01246640, 0.00007615, -0.00386841, 0.00106252, -0.00159081, 0.00119945, -0.00119588, 0.00061867, },
		{ -0.00050863, 0.00025349, -0.00022669, -0.00024320, 0.00054932, -0.00146938, 0.00034876, -0.00610927, 0.01527913, -0.00610927, 0.00034876, -0.00146938, 0.00054932, -0.00024320, -0.00022669, 0.00025349, },
		{ -0.00028641, 0.00061867, -0.00119588, 0.00119945, -0.00159081, 0.00106252, -0.00386841, 0.00007615, 0.01246640, -0.00825274, 0.00272163, -0.00292645, 0.00204943, -0.00143114, 0.00082036, -0.00048620, },
		{ 0.00015229, 0.00033333, -0.00125022, 0.00157863, -0.00261224, 0.00241315, -0.00693866, 0.00702878, 0.00592310, -0.00624835, 0.00228961, -0.00229011, 0.00190077, -0.00137376, 0.00102364, -0.00095340, },
		{ 0.00037784, -0.00023622, -0.00049185, 0.00062203, -0.00180034, 0.00130049, -0.00644390, 0.01085723, -0.00034297, -0.00272282, 0.00051986, -0.00050689, 0.00071876, -0.00038204, 0.00031256, -0.00080519, },
		{ 0.00022380, -0.00051312, 0.00021971, -0.00067346, -0.00018311, -0.00127253, -0.00274129, 0.01015428, -0.00351969, -0.00034137, -0.00058592, 0.00084460, -0.00018311, 0.00041804, -0.00055460, -0.00031566, },
		{ -0.00007327, -0.00032808, 0.00023516, -0.00117842, 0.00064136, -0.00305984, 0.00149948, 0.00665688, -0.00369312, 0.00023672, -0.00036052, 0.00119464, -0.00029220, 0.00037763, -0.00087885, -0.00000103, },
		{ -0.00019298, -0.00001895, -0.00026349, -0.00077722, 0.00017443, -0.00279251, 0.00381370, 0.00314951, -0.00270858, -0.00000200, 0.00056328, 0.00107700, -0.00019537, -0.00010904, -0.00070270, -0.00003851, },
		{ -0.00009955, 0.00004592, -0.00064244, -0.00024991, -0.00068931, -0.00116115, 0.00381072, 0.00108673, -0.00200128, -0.00008226, 0.00157700, 0.00091175, -0.00050173, -0.00032038, -0.00054463, -0.00016293, },
		},
		{ { -0.00026449, -0.00016938, 0.00035045, -0.00017035, -0.00042725, -0.00039812, -0.00218151, 0.00098199, 0.00526937, 0.00098199, -0.00218151, -0.00039812, -0.00042725, -0.00017035, 0.00035045, -0.00016938, },
		{ -0.00026449, -0.00015676, 0.00033506, -0.00017762, -0.00042725, -0.00038182, -0.00209177, 0.00066300, 0.00571539, 0.00066300, -0.00209177, -0.00038182, -0.00042725, -0.00017762, 0.00033506, -0.00015676, },
		{ -0.00026449, -0.00012082, 0.00029121, -0.00019834, -0.00042725, -0.00033543, -0.00183624, -0.00024539, 0.00698554, -0.00024539, -0.00183624, -0.00033543, -0.00042725, -0.00019834, 0.00029121, -0.00012082, },
		{ -0.00026449, -0.00006703, 0.00022560, -0.00022934, -0.00042725, -0.00026599, -0.00145380, -0.00160490, 0.00888646, -0.00160490, -0.00145380, -0.00026599, -0.00042725, -0.00022934, 0.00022560, -0.00006703, },
		{ -0.00026449, -0.00000358, 0.00014820, -0.00026591, -0.00042725, -0.00018408, -0.00100269, -0.00320854, 0.01112874, -0.00320854, -0.00100269, -0.00018408, -0.00042725, -0.00026591, 0.00014820, -0.00000358, },
		{ -0.00026449, 0.00005987, 0.00007080, -0.00030247, -0.00042725, -0.00010217, -0.00055158, -0.00481219, 0.01337103, -0.00481219, -0.00055158, -0.00010217, -0.00042725, -0.00030247, 0.00007080, 0.00005987, },
		{ -0.00026449, 0.00011366, 0.00000518, -0.00033348, -0.00042725, -0.00003273, -0.00016914, -0.00617170, 0.01527195, -0.00617170, -0.00016914, -0.00003273, -0.00042725, -0.00033348, 0.00000518, 0.00011366, },
		{ -0.00026449, 0.00014960, -0.00003866, -0.00035419, -0.00042725, 0.00001366, 0.00008639, -0.00708009, 0.01654210, -0.00708009, 0.00008639, 0.00001366, -0.00042725, -0.00035419, -0.00003866, 0.00014960, },
		{ -0.00026449, 0.00016222, -0.00005405, -0.00036147, -0.00042725, 0.00002996, 0.00017612, -0.00739907, 0.01698812, -0.00739907, 0.00017612, 0.00002996, -0.00042725, -0.00036147, -0.00005405, 0.00016222, },
		{ -0.00026449, 0.00014960, -0.00003866, -0.00035419, -0.00042725, 0.00001366, 0.00008639, -0.00708009, 0.01654210, -0.00708009, 0.00008639, 0.00001366, -0.00042725, -0.00035419, -0.00003866, 0.00014960, },
		{ -0.00026449, 0.00011366, 0.00000518, -0.00033348, -0.00042725, -0.00003273, -0.00016914, -0.00617170, 0.01527195, -0.00617170, -0.00016914, -0.00003273, -0.00042725, -0.00033348, 0.00000518, 0.00011366, },
		{ -0.00026449, 0.00005987, 0.00007080, -0.00030247, -0.00042725, -0.00010217, -0.00055158, -0.00481219, 0.01337103, -0.00481219, -0.00055158, -0.00010217, -0.00042725, -0.00030247, 0.00007080, 0.00005987, },
		{ -0.00026449, -0.00000358, 0.00014820, -0.00026591, -0.00042725, -0.00018408, -0.00100269, -0.00320854, 0.01112874, -0.00320854, -0.00100269, -0.00018408, -0.00042725, -0.00026591, 0.00014820, -0.00000358, },
		{ -0.00026449, -0.00006703, 0.00022560, -0.00022934, -0.00042725, -0.00026599, -0.00145380, -0.00160490, 0.00888646, -0.00160490, -0.00145380, -0.00026599, -0.00042725, -0.00022934, 0.00022560, -0.00006703, },
		{ -0.00026449, -0.00012082, 0.00029121, -0.00019834, -0.00042725, -0.00033543, -0.00183624, -0.00024539, 0.00698554, -0.00024539, -0.00183624, -0.00033543, -0.00042725, -0.00019834, 0.00029121, -0.00012082, },
		{ -0.00026449, -0.00015676, 0.00033506, -0.00017762, -0.00042725, -0.00038182, -0.00209177, 0.00066300, 0.00571539, 0.00066300, -0.00209177, -0.00038182, -0.00042725, -0.00017762, 0.00033506, -0.00015676, },
		},
		{ { -0.00002035, -0.00010224, -0.00063119, -0.00018276, -0.00091553, 0.00030483, 0.00270639, 0.00022431, -0.00181071, 0.00022431, 0.00270639, 0.00030483, -0.00091553, -0.00018276, -0.00063119, -0.00010224, },
		{ -0.00009955, 0.00004592, -0.00064244, -0.00024991, -0.00068931, -0.00116115, 0.00381072, 0.00108673, -0.00200128, -0.00008226, 0.00157700, 0.00091175, -0.00050173, -0.00032038, -0.00054463, -0.00016293, },
		{ -0.00019298, -0.00001895, -0.00026349, -0.00077722, 0.00017443, -0.00279251, 0.00381370, 0.00314951, -0.00270858, -0.00000200, 0.00056328, 0.00107700, -0.00019537, -0.00010904, -0.00070270, -0.00003851, },
		{ -0.00007327, -0.00032808, 0.00023516, -0.00117842, 0.00064136, -0.00305984, 0.00149948, 0.00665688, -0.00369312, 0.00023672, -0.00036052, 0.00119464, -0.00029220, 0.00037763, -0.00087885, -0.00000103, },
		{ 0.00022380, -0.00051312, 0.00021971, -0.00067346, -0.00018311, -0.00127253, -0.00274129, 0.01015428, -0.00351969, -0.00034137, -0.00058592, 0.00084460, -0.00018311, 0.00041804, -0.00055460, -0.00031566, },
		{ 0.00037784, -0.00023622, -0.00049185, 0.00062203, -0.00180034, 0.00130049, -0.00644390, 0.01085723, -0.00034297, -0.00272282, 0.00051986, -0.00050689, 0.00071876, -0.00038204, 0.00031256, -0.00080519, },
		{ 0.00015229, 0.00033333, -0.00125022, 0.00157863, -0.00261224, 0.00241315, -0.00693866, 0.00702878, 0.00592310, -0.00624835, 0.00228961, -0.00229011, 0.00190077, -0.00137376, 0.00102364, -0.00095340, },
		{ -0.00028641, 0.00061867, -0.00119588, 0.00119945, -0.00159081, 0.00106252, -0.00386841, 0.00007615, 0.01246640, -0.00825274, 0.00272163, -0.00292645, 0.00204943, -0.00143114, 0.00082036, -0.00048620, },
		{ -0.00050863, 0.00025349, -0.00022669, -0.00024320, 0.00054932, -0.00146938, 0.00034876, -0.00610927, 0.01527913, -0.00610927, 0.00034876, -0.00146938, 0.00054932, -0.00024320, -0.00022669, 0.00025349, },
		{ -0.00028641, -0.00048620, 0.00082036, -0.00143114, 0.00204943, -0.00292645, 0.00272163, -0.00825274, 0.01246640, 0.00007615, -0.00386841, 0.00106252, -0.00159081, 0.00119945, -0.00119588, 0.00061867, },
		{ 0.00015229, -0.00095340, 0.00102364, -0.00137376, 0.00190077, -0.00229011, 0.00228961, -0.00624835, 0.00592310, 0.00702878, -0.00693866, 0.00241315, -0.00261224, 0.00157863, -0.00125022, 0.00033333, },
		{ 0.00037784, -0.00080519, 0.00031256, -0.00038204, 0.00071876, -0.00050689, 0.00051986, -0.00272282, -0.00034297, 0.01085723, -0.00644390, 0.00130049, -0.00180034, 0.00062203, -0.00049185, -0.00023622, },
		{ 0.00022380, -0.00031566, -0.00055460, 0.00041804, -0.00018311, 0.00084460, -0.00058592, -0.00034137, -0.00351969, 0.01015428, -0.00274129, -0.00127253, -0.00018311, -0.00067346, 0.00021971, -0.00051312, },
		{ -0.00007327, -0.00000103, -0.00087885, 0.00037763, -0.00029220, 0.00119464, -0.00036052, 0.00023672, -0.00369312, 0.00665688, 0.00149948, -0.00305984, 0.00064136, -0.00117842, 0.00023516, -0.00032808, },
		{ -0.00019298, -0.00003851, -0.00070270, -0.00010904, -0.00019537, 0.00107700, 0.00056328, -0.00000200, -0.00270858, 0.00314951, 0.00381370, -0.00279251, 0.00017443, -0.00077722, -0.00026349, -0.00001895, },
		{ -0.00009955, -0.00016293, -0.00054463, -0.00032038, -0.00050173, 0.00091175, 0.00157700, -0.00008226, -0.00200128, 0.00108673, 0.00381072, -0.00116115, -0.00068931, -0.00024991, -0.00064244, 0.00004592, },
		},
		{ { -0.00042725, -0.00028423, -0.00091553, 0.00040630, 0.00250244, 0.00040630, -0.00091553, -0.00028423, -0.00042725, -0.00028423, -0.00091553, 0.00040630, 0.00250244, 0.00040630, -0.00091553, -0.00028423, },
		{ -0.00014260, -0.00056521, -0.00021378, -0.00154450, 0.00259818, 0.00259818, -0.00154450, -0.00021378, -0.00056521, -0.00014260, -0.00073695, -0.00012789, 0.00122103, 0.00122103, -0.00012789, -0.00073695, },
		{ -0.00032612, -0.00042725, -0.00020764, -0.00103677, -0.00101665, 0.00597653, -0.00101665, -0.00103677, -0.00020764, -0.00042725, -0.00032612, -0.00050826, 0.00065044, 0.00074453, 0.00065044, -0.00050826, },
		{ -0.00067472, 0.00013132, -0.00107217, 0.00081512, -0.00377256, 0.00436075, 0.00436075, -0.00377256, 0.00081512, -0.00107217, 0.00013132, -0.00067472, 0.00026161, 0.00043893, 0.00043893, 0.00026161, },
		{ 0.00006104, -0.00033445, -0.00042725, -0.00011553, -0.00091553, -0.00314626, 0.01026587, -0.00314626, -0.00091553, -0.00011553, -0.00042725, -0.00033445, 0.00006104, -0.00006587, 0.00059839, -0.00006587, },
		{ -0.00003675, -0.00027311, 0.00026536, -0.00153578, 0.00195660, -0.00607431, 0.00602784, 0.00602784, -0.00607431, 0.00195660, -0.00153578, 0.00026536, -0.00027311, -0.00003675, 0.00015844, 0.00015844, },
		{ -0.00101665, 0.00055599, -0.00032612, -0.00042725, -0.00004009, -0.00071995, -0.00504138, 0.01381931, -0.00504138, -0.00071995, -0.00004009, -0.00042725, -0.00032612, 0.00055599, -0.00101665, 0.00118815, },
		{ 0.00012166, -0.00041392, 0.00007225, 0.00031056, -0.00181403, 0.00270237, -0.00753570, 0.00704510, 0.00704510, -0.00753570, 0.00270237, -0.00181403, 0.00031056, 0.00007225, -0.00041392, 0.00012166, },
		{ 0.00152588, -0.00144026, 0.00075157, -0.00032055, -0.00042725, -0.00001096, -0.00062950, -0.00579659, 0.01519775, -0.00579659, -0.00062950, -0.00001096, -0.00042725, -0.00032055, 0.00075157, -0.00144026, },
		{ 0.00012166, 0.00012166, -0.00041392, 0.00007225, 0.00031056, -0.00181403, 0.00270237, -0.00753570, 0.00704510, 0.00704510, -0.00753570, 0.00270237, -0.00181403, 0.00031056, 0.00007225, -0.00041392, },
		{ -0.00101665, 0.00118815, -0.00101665, 0.00055599, -0.00032612, -0.00042725, -0.00004009, -0.00071995, -0.00504138, 0.01381931, -0.00504138, -0.00071995, -0.00004009, -0.00042725, -0.00032612, 0.00055599, },
		{ -0.00003675, 0.00015844, 0.00015844, -0.00003675, -0.00027311, 0.00026536, -0.00153578, 0.00195660, -0.00607431, 0.00602784, 0.00602784, -0.00607431, 0.00195660, -0.00153578, 0.00026536, -0.00027311, },
		{ 0.00006104, -0.00006587, 0.00059839, -0.00006587, 0.00006104, -0.00033445, -0.00042725, -0.00011553, -0.00091553, -0.00314626, 0.01026587, -0.00314626, -0.00091553, -0.00011553, -0.00042725, -0.00033445, },
		{ -0.00067472, 0.00026161, 0.00043893, 0.00043893, 0.00026161, -0.00067472, 0.00013132, -0.00107217, 0.00081512, -0.00377256, 0.00436075, 0.00436075, -0.00377256, 0.00081512, -0.00107217, 0.00013132, },
		{ -0.00032612, -0.00050826, 0.00065044, 0.00074453, 0.00065044, -0.00050826, -0.00032612, -0.00042725, -0.00020764, -0.00103677, -0.00101665, 0.00597653, -0.00101665, -0.00103677, -0.00020764, -0.00042725, },
		{ -0.00014260, -0.00073695, -0.00012789, 0.00122103, 0.00122103, -0.00012789, -0.00073695, -0.00014260, -0.00056521, -0.00021378, -0.00154450, 0.00259818, 0.00259818, -0.00154450, -0.00021378, -0.00056521, },
		},
		{ { -0.00156657, -0.00001983, 0.00295053, 0.00006069, -0.00067139, -0.00042691, -0.00038705, -0.00034638, 0.00022380, -0.00034638, -0.00038705, -0.00042691, -0.00067139, 0.00006069, 0.00295053, -0.00001983, },
		{ -0.00030782, -0.00177572, 0.00086117, 0.00403627, -0.00138671, -0.00046375, -0.00047547, -0.00041688, -0.00017963, 0.00012601, -0.00038848, -0.00031907, -0.00054593, -0.00027617, 0.00068619, 0.00180256, },
		{ 0.00073591, -0.00017463, -0.00253594, 0.00297687, 0.00398633, -0.00296515, 0.00034706, -0.00094985, -0.00009086, -0.00019158, -0.00002035, -0.00021114, -0.00053007, -0.00028168, -0.00002274, 0.00090437, },
		{ 0.00110121, -0.00026709, 0.00014329, -0.00359969, 0.00656345, 0.00159291, -0.00315327, 0.00073479, -0.00127185, 0.00032859, -0.00042151, 0.00002016, -0.00009445, -0.00078542, 0.00028420, -0.00019878, },
		{ -0.00018311, 0.00084460, -0.00058592, -0.00034137, -0.00351969, 0.01015428, -0.00274129, -0.00127253, -0.00018311, -0.00067346, 0.00021971, -0.00051312, 0.00022380, -0.00031566, -0.00055460, 0.00041804, },
		{ -0.00028861, 0.00062533, -0.00041346, 0.00042643, -0.00262939, -0.00043640, 0.01095066, -0.00653733, 0.00139392, -0.00189377, 0.00071546, -0.00058528, -0.00014280, 0.00028442, -0.00071176, 0.00021913, },
		{ 0.00085100, -0.00120112, 0.00172813, -0.00211747, 0.00211698, -0.00607571, 0.00575046, 0.00720142, -0.00711129, 0.00258578, -0.00278488, 0.00175126, -0.00142285, 0.00050596, -0.00002035, -0.00078077, },
		{ -0.00026065, 0.00059480, -0.00120558, 0.00182387, -0.00270089, 0.00249608, -0.00802718, 0.01224084, 0.00030170, -0.00409397, 0.00128807, -0.00181637, 0.00142500, -0.00142143, 0.00084422, -0.00051196, },
		{ -0.00075277, 0.00049763, -0.00047083, 0.00000094, 0.00030518, -0.00122524, 0.00010462, -0.00586513, 0.01503499, -0.00586513, 0.00010462, -0.00122524, 0.00030518, 0.00000094, -0.00047083, 0.00049763, },
		{ -0.00026065, -0.00051196, 0.00084422, -0.00142143, 0.00142500, -0.00181637, 0.00128807, -0.00409397, 0.00030170, 0.01224084, -0.00802718, 0.00249608, -0.00270089, 0.00182387, -0.00120558, 0.00059480, },
		{ 0.00085100, -0.00078077, -0.00002035, 0.00050596, -0.00142285, 0.00175126, -0.00278488, 0.00258578, -0.00711129, 0.00720142, 0.00575046, -0.00607571, 0.00211698, -0.00211747, 0.00172813, -0.00120112, },
		{ -0.00028861, 0.00021913, -0.00071176, 0.00028442, -0.00014280, -0.00058528, 0.00071546, -0.00189377, 0.00139392, -0.00653733, 0.01095066, -0.00043640, -0.00262939, 0.00042643, -0.00041346, 0.00062533, },
		{ -0.00018311, 0.00041804, -0.00055460, -0.00031566, 0.00022380, -0.00051312, 0.00021971, -0.00067346, -0.00018311, -0.00127253, -0.00274129, 0.01015428, -0.00351969, -0.00034137, -0.00058592, 0.00084460, },
		{ 0.00110121, -0.00019878, 0.00028420, -0.00078542, -0.00009445, 0.00002016, -0.00042151, 0.00032859, -0.00127185, 0.00073479, -0.00315327, 0.00159291, 0.00656345, -0.00359969, 0.00014329, -0.00026709, },
		{ 0.00073591, 0.00090437, -0.00002274, -0.00028168, -0.00053007, -0.00021114, -0.00002035, -0.00019158, -0.00009086, -0.00094985, 0.00034706, -0.00296515, 0.00398633, 0.00297687, -0.00253594, -0.00017463, },
		{ -0.00030782, 0.00180256, 0.00068619, -0.00027617, -0.00054593, -0.00031907, -0.00038848, 0.00012601, -0.00017963, -0.00041688, -0.00047547, -0.00046375, -0.00138671, 0.00403627, 0.00086117, -0.00177572, },
		},
		{ { 0.00404867, 0.00143593, -0.00197925, -0.00002519, -0.00115967, 0.00057630, -0.00082836, 0.00118679, -0.00148519, 0.00118679, -0.00082836, 0.00057630, -0.00115967, -0.00002519, -0.00197925, 0.00143593, },
		{ 0.00141734, 0.00451327, 0.00089611, -0.00201795, 0.00030856, -0.00136511, 0.00080330, -0.00105392, 0.00122888, -0.00146660, 0.00111215, -0.00060281, 0.00036975, -0.00099139, -0.00023445, -0.00194055, },
		{ -0.00190775, 0.00084318, 0.00583634, -0.00011994, -0.00205076, 0.00071598, -0.00157644, 0.00101620, -0.00124514, 0.00123201, -0.00141368, 0.00101632, -0.00041159, 0.00021510, -0.00088591, -0.00028737, },
		{ -0.00017590, -0.00188582, -0.00019914, 0.00781647, -0.00145752, -0.00207268, 0.00113505, -0.00176149, 0.00118258, -0.00137291, 0.00119571, -0.00133448, 0.00091389, -0.00028382, 0.00013590, -0.00085927, },
		{ -0.00091553, 0.00008300, -0.00187813, -0.00155095, 0.01015218, -0.00291301, -0.00208038, 0.00150196, -0.00189209, 0.00127712, -0.00141777, 0.00112550, -0.00124105, 0.00082047, -0.00023896, 0.00014419, },
		{ 0.00023873, -0.00104612, 0.00044992, -0.00188582, -0.00300644, 0.01248790, -0.00426482, -0.00207268, 0.00176086, -0.00194835, 0.00128542, -0.00137291, 0.00103207, -0.00114762, 0.00075026, -0.00028382, },
		{ -0.00041159, 0.00040511, -0.00123118, 0.00086898, -0.00190775, -0.00434402, 0.01446802, -0.00530715, -0.00205076, 0.00187234, -0.00192171, 0.00120621, -0.00124514, 0.00092965, -0.00106841, 0.00071396, },
		{ 0.00071709, -0.00060281, 0.00061801, -0.00144251, 0.00127640, -0.00194055, -0.00536007, 0.01579109, -0.00588131, -0.00201795, 0.00181941, -0.00181622, 0.00105156, -0.00105392, 0.00083382, -0.00101549, },
		{ -0.00099691, 0.00075918, -0.00082836, 0.00084501, -0.00164795, 0.00161015, -0.00197925, -0.00589989, 0.01625570, -0.00589989, -0.00197925, 0.00161015, -0.00164795, 0.00084501, -0.00082836, 0.00075918, },
		{ 0.00071709, -0.00101549, 0.00083382, -0.00105392, 0.00105156, -0.00181622, 0.00181941, -0.00201795, -0.00588131, 0.01579109, -0.00536007, -0.00194055, 0.00127640, -0.00144251, 0.00061801, -0.00060281, },
		{ -0.00041159, 0.00071396, -0.00106841, 0.00092965, -0.00124514, 0.00120621, -0.00192171, 0.00187234, -0.00205076, -0.00530715, 0.01446802, -0.00434402, -0.00190775, 0.00086898, -0.00123118, 0.00040511, },
		{ 0.00023873, -0.00028382, 0.00075026, -0.00114762, 0.00103207, -0.00137291, 0.00128542, -0.00194835, 0.00176086, -0.00207268, -0.00426482, 0.01248790, -0.00300644, -0.00188582, 0.00044992, -0.00104612, },
		{ -0.00091553, 0.00014419, -0.00023896, 0.00082047, -0.00124105, 0.00112550, -0.00141777, 0.00127712, -0.00189209, 0.00150196, -0.00208038, -0.00291301, 0.01015218, -0.00155095, -0.00187813, 0.00008300, },
		{ -0.00017590, -0.00085927, 0.00013590, -0.00028382, 0.00091389, -0.00133448, 0.00119571, -0.00137291, 0.00118258, -0.00176149, 0.00113505, -0.00207268, -0.00145752, 0.00781647, -0.00019914, -0.00188582, },
		{ -0.00190775, -0.00028737, -0.00088591, 0.00021510, -0.00041159, 0.00101632, -0.00141368, 0.00123201, -0.00124514, 0.00101620, -0.00157644, 0.00071598, -0.00205076, -0.00011994, 0.00583634, 0.00084318, },
		{ 0.00141734, -0.00194055, -0.00023445, -0.00099139, 0.00036975, -0.00060281, 0.00111215, -0.00146660, 0.00122888, -0.00105392, 0.00080330, -0.00136511, 0.00030856, -0.00201795, 0.00089611, 0.00451327, },
		},
	}
	;

	float f32_3[8][32][32] = {
		{ { -0.00088590, -0.00016349, 0.00021680, 0.00020719, 0.00007881, 0.00025997, -0.00006529, -0.00008603, -0.00000723, -0.00002000, 0.00012104, 0.00008664, 0.00000753, -0.00030965, -0.00010005, 0.00002543, 0.00038255, 0.00002543, -0.00010005, -0.00030965, 0.00000753, 0.00008664, 0.00012104, -0.00002000, -0.00000723, -0.00008603, -0.00006529, 0.00025997, 0.00007881, 0.00020719, 0.00021680, -0.00016349, },
		{ -0.00011199, -0.00092438, -0.00019036, 0.00010299, 0.00018288, -0.00002520, 0.00016705, -0.00002183, 0.00001311, 0.00008754, -0.00014843, -0.00005429, 0.00002800, 0.00026028, -0.00015608, -0.00005516, -0.00031617, 0.00004783, 0.00025399, 0.00022630, -0.00013860, -0.00019202, -0.00012124, 0.00008562, 0.00015191, 0.00010761, -0.00009158, -0.00019343, 0.00019557, 0.00014729, 0.00028869, 0.00040733, },
		{ 0.00061483, 0.00005763, -0.00094736, -0.00023568, -0.00004694, 0.00018779, -0.00001344, 0.00007666, -0.00011686, -0.00004182, 0.00025627, -0.00005533, -0.00012514, -0.00024365, 0.00030368, -0.00010654, 0.00014267, -0.00007173, -0.00030460, -0.00008723, 0.00026458, 0.00022928, 0.00001474, -0.00019661, -0.00013577, 0.00009157, 0.00021725, 0.00006379, -0.00018680, 0.00008920, 0.00008904, 0.00026904, },
		{ 0.00014948, 0.00068338, 0.00022128, -0.00091991, -0.00017300, -0.00019158, 0.00008513, 0.00001030, 0.00010913, -0.00012523, -0.00025046, 0.00023582, 0.00007527, 0.00008776, -0.00049188, 0.00023255, -0.00010364, -0.00006436, 0.00027084, -0.00004785, -0.00029731, -0.00011543, 0.00018719, 0.00028317, 0.00007780, -0.00014025, -0.00006922, 0.00012888, 0.00015626, -0.00006276, 0.00011141, 0.00006359, },
		{ 0.00015481, 0.00011775, 0.00070565, 0.00027609, -0.00103984, -0.00005694, -0.00015770, -0.00000790, -0.00008592, 0.00006919, 0.00004243, -0.00032353, 0.00007586, -0.00009092, 0.00046128, -0.00050252, -0.00000995, 0.00012218, -0.00033731, 0.00007221, 0.00021982, -0.00000506, -0.00021435, -0.00011942, 0.00014755, 0.00027254, 0.00011119, -0.00000311, 0.00004080, 0.00011019, -0.00001330, 0.00017045, },
		{ 0.00016235, 0.00021678, 0.00015193, 0.00084502, 0.00043012, -0.00133158, -0.00016084, -0.00008547, 0.00008876, -0.00007440, -0.00016763, 0.00008715, -0.00021321, 0.00020456, -0.00056390, 0.00054645, -0.00009485, -0.00037846, 0.00025199, -0.00018937, -0.00022154, -0.00001094, 0.00013580, 0.00004260, -0.00009510, -0.00007504, 0.00013873, 0.00021640, 0.00018425, 0.00013566, 0.00012666, 0.00000400, },
		{ 0.00010820, 0.00022388, 0.00027465, 0.00010962, 0.00094233, 0.00078815, -0.00148679, -0.00040659, -0.00026940, 0.00008954, 0.00022986, -0.00023838, -0.00008439, -0.00042284, 0.00074019, -0.00072002, -0.00000769, 0.00043175, -0.00042869, 0.00003075, 0.00001749, -0.00011318, -0.00022662, -0.00013473, 0.00002648, 0.00008445, -0.00000645, 0.00004310, 0.00014478, 0.00023230, 0.00023913, 0.00021539, },
		{ 0.00025120, 0.00014935, 0.00031493, 0.00049086, 0.00018933, 0.00091404, 0.00106759, -0.00152942, -0.00043125, -0.00043412, -0.00032511, 0.00036199, 0.00001094, 0.00015765, -0.00113794, 0.00090064, -0.00004249, -0.00071034, 0.00043522, -0.00010092, -0.00012308, -0.00007669, 0.00003360, -0.00007983, -0.00025738, -0.00025547, -0.00006268, 0.00001665, 0.00008580, 0.00016210, 0.00026924, 0.00031139, },
		{ 0.00051054, 0.00044804, 0.00018325, 0.00020039, 0.00059202, 0.00049641, 0.00113389, 0.00134550, -0.00185540, -0.00049969, -0.00014286, -0.00064305, 0.00003249, -0.00023052, 0.00108084, -0.00136844, 0.00002156, 0.00080859, -0.00070872, -0.00001367, 0.00004079, 0.00001095, -0.00014655, -0.00006479, 0.00002873, -0.00008130, -0.00028097, -0.00029055, -0.00023391, -0.00008926, 0.00008955, 0.00031666, },
		{ -0.00000444, 0.00049553, 0.00081685, 0.00063749, 0.00025666, 0.00037430, 0.00049568, 0.00152024, 0.00217668, -0.00212888, -0.00113525, 0.00005950, -0.00044052, 0.00020765, -0.00139524, 0.00141950, -0.00027325, -0.00103436, 0.00077116, -0.00017202, -0.00018514, -0.00014131, 0.00004774, -0.00002043, -0.00010567, -0.00006435, -0.00001562, -0.00010486, -0.00021243, -0.00035542, -0.00038461, -0.00027991, },
		{ -0.00051245, -0.00057172, -0.00016252, 0.00059491, 0.00115720, 0.00101493, 0.00064247, 0.00020238, 0.00126700, 0.00316773, -0.00166142, -0.00167043, -0.00033993, -0.00084582, 0.00151863, -0.00193779, 0.00005081, 0.00102606, -0.00096061, 0.00012505, 0.00000399, 0.00000886, -0.00017792, -0.00011199, -0.00000188, -0.00000420, -0.00007118, -0.00009405, -0.00008619, -0.00006147, -0.00018854, -0.00036508, },
		{ -0.00015951, -0.00030015, -0.00066977, -0.00085479, -0.00047239, 0.00056793, 0.00152544, 0.00182217, 0.00094556, 0.00084135, 0.00331722, -0.00137898, -0.00144374, 0.00010948, -0.00268447, 0.00211628, -0.00036138, -0.00164429, 0.00086948, -0.00018211, -0.00004378, -0.00021509, 0.00001657, -0.00001496, -0.00013429, -0.00011531, -0.00001495, 0.00000747, -0.00007132, -0.00013968, -0.00008165, -0.00008214, },
		{ -0.00011168, -0.00013487, -0.00008038, -0.00028678, -0.00071747, -0.00116956, -0.00093374, 0.00030637, 0.00211063, 0.00275702, 0.00216153, 0.00396212, -0.00246401, -0.00253902, 0.00259214, -0.00362157, 0.00005308, 0.00170607, -0.00153287, -0.00018876, -0.00006096, 0.00021971, -0.00018263, -0.00016560, -0.00001477, -0.00003435, -0.00010618, -0.00009770, -0.00002427, -0.00002201, -0.00009890, -0.00010176, },
		{ -0.00006593, -0.00007718, -0.00016454, -0.00012087, -0.00013065, -0.00020270, -0.00068489, -0.00146627, -0.00165459, -0.00011038, 0.00268763, 0.00442520, 0.00792790, -0.00208212, -0.00723312, 0.00397729, -0.00097178, -0.00269835, 0.00155609, -0.00024030, -0.00040753, -0.00048547, 0.00011423, 0.00002842, -0.00014064, -0.00011487, -0.00006392, -0.00003860, -0.00007141, -0.00007536, -0.00006461, -0.00008595, },
		{ -0.00012433, -0.00010221, -0.00005398, -0.00009225, -0.00009487, -0.00023672, -0.00019668, -0.00007846, -0.00054254, -0.00178818, -0.00266632, -0.00078325, 0.00295013, 0.01326069, 0.00468360, -0.01181426, -0.00009954, 0.00279760, -0.00258048, -0.00010707, 0.00004558, 0.00009903, -0.00042087, -0.00020837, -0.00005453, -0.00004440, -0.00006459, -0.00012590, -0.00007414, -0.00003187, -0.00004206, -0.00008023, },
		{ -0.00004840, -0.00009950, -0.00012878, -0.00012445, -0.00012281, 0.00002233, -0.00011372, -0.00037166, -0.00026117, 0.00003868, -0.00025632, -0.00229248, -0.00369681, -0.00378957, 0.01247644, 0.01595992, -0.00871224, -0.00725303, 0.00243679, -0.00041888, -0.00048532, -0.00057469, 0.00002826, -0.00005235, -0.00013732, -0.00020750, -0.00015246, -0.00001859, -0.00008134, -0.00010384, -0.00006104, -0.00003035, },
		{ -0.00004893, -0.00006073, -0.00007611, -0.00007436, -0.00010651, -0.00023054, -0.00008100, 0.00004144, -0.00015108, -0.00040855, -0.00045960, 0.00027389, -0.00023883, -0.00174957, -0.00939589, 0.00255363, 0.02175766, 0.00255363, -0.00939589, -0.00174957, -0.00023883, 0.00027389, -0.00045960, -0.00040855, -0.00015108, 0.00004144, -0.00008100, -0.00023054, -0.00010651, -0.00007436, -0.00007611, -0.00006073, },
		{ -0.00004840, -0.00003035, -0.00006104, -0.00010384, -0.00008134, -0.00001859, -0.00015246, -0.00020750, -0.00013732, -0.00005235, 0.00002826, -0.00057469, -0.00048532, -0.00041888, 0.00243679, -0.00725303, -0.00871224, 0.01595992, 0.01247644, -0.00378957, -0.00369681, -0.00229248, -0.00025632, 0.00003868, -0.00026117, -0.00037166, -0.00011372, 0.00002233, -0.00012281, -0.00012445, -0.00012878, -0.00009950, },
		{ -0.00012433, -0.00008023, -0.00004206, -0.00003187, -0.00007414, -0.00012590, -0.00006459, -0.00004440, -0.00005453, -0.00020837, -0.00042087, 0.00009903, 0.00004558, -0.00010707, -0.00258048, 0.00279760, -0.00009954, -0.01181426, 0.00468360, 0.01326069, 0.00295013, -0.00078325, -0.00266632, -0.00178818, -0.00054254, -0.00007846, -0.00019668, -0.00023672, -0.00009487, -0.00009225, -0.00005398, -0.00010221, },
		{ -0.00006593, -0.00008595, -0.00006461, -0.00007536, -0.00007141, -0.00003860, -0.00006392, -0.00011487, -0.00014064, 0.00002842, 0.00011423, -0.00048547, -0.00040753, -0.00024030, 0.00155609, -0.00269835, -0.00097178, 0.00397729, -0.00723312, -0.00208212, 0.00792790, 0.00442520, 0.00268763, -0.00011038, -0.00165459, -0.00146627, -0.00068489, -0.00020270, -0.00013065, -0.00012087, -0.00016454, -0.00007718, },
		{ -0.00011168, -0.00010176, -0.00009890, -0.00002201, -0.00002427, -0.00009770, -0.00010618, -0.00003435, -0.00001477, -0.00016560, -0.00018263, 0.00021971, -0.00006096, -0.00018876, -0.00153287, 0.00170607, 0.00005308, -0.00362157, 0.00259214, -0.00253902, -0.00246401, 0.00396212, 0.00216153, 0.00275702, 0.00211063, 0.00030637, -0.00093374, -0.00116956, -0.00071747, -0.00028678, -0.00008038, -0.00013487, },
		{ -0.00015951, -0.00008214, -0.00008165, -0.00013968, -0.00007132, 0.00000747, -0.00001495, -0.00011531, -0.00013429, -0.00001496, 0.00001657, -0.00021509, -0.00004378, -0.00018211, 0.00086948, -0.00164429, -0.00036138, 0.00211628, -0.00268447, 0.00010948, -0.00144374, -0.00137898, 0.00331722, 0.00084135, 0.00094556, 0.00182217, 0.00152544, 0.00056793, -0.00047239, -0.00085479, -0.00066977, -0.00030015, },
		{ -0.00051245, -0.00036508, -0.00018854, -0.00006147, -0.00008619, -0.00009405, -0.00007118, -0.00000420, -0.00000188, -0.00011199, -0.00017792, 0.00000886, 0.00000399, 0.00012505, -0.00096061, 0.00102606, 0.00005081, -0.00193779, 0.00151863, -0.00084582, -0.00033993, -0.00167043, -0.00166142, 0.00316773, 0.00126700, 0.00020238, 0.00064247, 0.00101493, 0.00115720, 0.00059491, -0.00016252, -0.00057172, },
		{ -0.00000444, -0.00027991, -0.00038461, -0.00035542, -0.00021243, -0.00010486, -0.00001562, -0.00006435, -0.00010567, -0.00002043, 0.00004774, -0.00014131, -0.00018514, -0.00017202, 0.00077116, -0.00103436, -0.00027325, 0.00141950, -0.00139524, 0.00020765, -0.00044052, 0.00005950, -0.00113525, -0.00212888, 0.00217668, 0.00152024, 0.00049568, 0.00037430, 0.00025666, 0.00063749, 0.00081685, 0.00049553, },
		{ 0.00051054, 0.00031666, 0.00008955, -0.00008926, -0.00023391, -0.00029055, -0.00028097, -0.00008130, 0.00002873, -0.00006479, -0.00014655, 0.00001095, 0.00004079, -0.00001367, -0.00070872, 0.00080859, 0.00002156, -0.00136844, 0.00108084, -0.00023052, 0.00003249, -0.00064305, -0.00014286, -0.00049969, -0.00185540, 0.00134550, 0.00113389, 0.00049641, 0.00059202, 0.00020039, 0.00018325, 0.00044804, },
		{ 0.00025120, 0.00031139, 0.00026924, 0.00016210, 0.00008580, 0.00001665, -0.00006268, -0.00025547, -0.00025738, -0.00007983, 0.00003360, -0.00007669, -0.00012308, -0.00010092, 0.00043522, -0.00071034, -0.00004249, 0.00090064, -0.00113794, 0.00015765, 0.00001094, 0.00036199, -0.00032511, -0.00043412, -0.00043125, -0.00152942, 0.00106759, 0.00091404, 0.00018933, 0.00049086, 0.00031493, 0.00014935, },
		{ 0.00010820, 0.00021539, 0.00023913, 0.00023230, 0.00014478, 0.00004310, -0.00000645, 0.00008445, 0.00002648, -0.00013473, -0.00022662, -0.00011318, 0.00001749, 0.00003075, -0.00042869, 0.00043175, -0.00000769, -0.00072002, 0.00074019, -0.00042284, -0.00008439, -0.00023838, 0.00022986, 0.00008954, -0.00026940, -0.00040659, -0.00148679, 0.00078815, 0.00094233, 0.00010962, 0.00027465, 0.00022388, },
		{ 0.00016235, 0.00000400, 0.00012666, 0.00013566, 0.00018425, 0.00021640, 0.00013873, -0.00007504, -0.00009510, 0.00004260, 0.00013580, -0.00001094, -0.00022154, -0.00018937, 0.00025199, -0.00037846, -0.00009485, 0.00054645, -0.00056390, 0.00020456, -0.00021321, 0.00008715, -0.00016763, -0.00007440, 0.00008876, -0.00008547, -0.00016084, -0.00133158, 0.00043012, 0.00084502, 0.00015193, 0.00021678, },
		{ 0.00015481, 0.00017045, -0.00001330, 0.00011019, 0.00004080, -0.00000311, 0.00011119, 0.00027254, 0.00014755, -0.00011942, -0.00021435, -0.00000506, 0.00021982, 0.00007221, -0.00033731, 0.00012218, -0.00000995, -0.00050252, 0.00046128, -0.00009092, 0.00007586, -0.00032353, 0.00004243, 0.00006919, -0.00008592, -0.00000790, -0.00015770, -0.00005694, -0.00103984, 0.00027609, 0.00070565, 0.00011775, },
		{ 0.00014948, 0.00006359, 0.00011141, -0.00006276, 0.00015626, 0.00012888, -0.00006922, -0.00014025, 0.00007780, 0.00028317, 0.00018719, -0.00011543, -0.00029731, -0.00004785, 0.00027084, -0.00006436, -0.00010364, 0.00023255, -0.00049188, 0.00008776, 0.00007527, 0.00023582, -0.00025046, -0.00012523, 0.00010913, 0.00001030, 0.00008513, -0.00019158, -0.00017300, -0.00091991, 0.00022128, 0.00068338, },
		{ 0.00061483, 0.00026904, 0.00008904, 0.00008920, -0.00018680, 0.00006379, 0.00021725, 0.00009157, -0.00013577, -0.00019661, 0.00001474, 0.00022928, 0.00026458, -0.00008723, -0.00030460, -0.00007173, 0.00014267, -0.00010654, 0.00030368, -0.00024365, -0.00012514, -0.00005533, 0.00025627, -0.00004182, -0.00011686, 0.00007666, -0.00001344, 0.00018779, -0.00004694, -0.00023568, -0.00094736, 0.00005763, },
		{ -0.00011199, 0.00040733, 0.00028869, 0.00014729, 0.00019557, -0.00019343, -0.00009158, 0.00010761, 0.00015191, 0.00008562, -0.00012124, -0.00019202, -0.00013860, 0.00022630, 0.00025399, 0.00004783, -0.00031617, -0.00005516, -0.00015608, 0.00026028, 0.00002800, -0.00005429, -0.00014843, 0.00008754, 0.00001311, -0.00002183, 0.00016705, -0.00002520, 0.00018288, 0.00010299, -0.00019036, -0.00092438, },
		},
		{ { -0.00005752, -0.00005577, -0.00005746, -0.00008248, -0.00013930, -0.00018367, -0.00014281, -0.00002324, 0.00004603, -0.00004345, -0.00020292, -0.00021350, -0.00001318, 0.00017468, 0.00009820, -0.00018247, -0.00033941, -0.00018247, 0.00009820, 0.00017468, -0.00001318, -0.00021350, -0.00020292, -0.00004345, 0.00004603, -0.00002324, -0.00014281, -0.00018367, -0.00013930, -0.00008248, -0.00005746, -0.00005577, },
		{ -0.00004525, -0.00002391, -0.00003997, -0.00005858, -0.00004854, -0.00003916, -0.00008455, -0.00015648, -0.00014322, -0.00001044, 0.00009904, 0.00002827, -0.00014993, -0.00017992, 0.00003633, 0.00025885, 0.00018504, -0.00014760, -0.00037756, -0.00025881, 0.00004425, 0.00018028, 0.00003983, -0.00015935, -0.00018525, -0.00005721, 0.00003987, 0.00000511, -0.00009791, -0.00016280, -0.00015367, -0.00009974, },
		{ -0.00013831, -0.00009620, -0.00001404, 0.00003774, 0.00001549, -0.00003068, -0.00002210, 0.00001840, -0.00001249, -0.00011949, -0.00015390, -0.00002587, 0.00010787, 0.00001637, -0.00025273, -0.00036003, -0.00009953, 0.00028743, 0.00037301, 0.00007144, -0.00025786, -0.00026399, -0.00001215, 0.00015783, 0.00007565, -0.00010278, -0.00015588, -0.00006426, 0.00003070, 0.00003082, -0.00004083, -0.00011383, },
		{ -0.00002674, -0.00010530, -0.00017529, -0.00017026, -0.00007718, 0.00000805, -0.00001284, -0.00009049, -0.00008782, 0.00000008, 0.00001587, -0.00010225, -0.00016263, 0.00002745, 0.00031780, 0.00032954, -0.00004170, -0.00042228, -0.00039739, -0.00003675, 0.00020095, 0.00005662, -0.00023528, -0.00030001, -0.00009856, 0.00009105, 0.00007381, -0.00005805, -0.00011822, -0.00006753, 0.00000158, 0.00001647, },
		{ -0.00004455, -0.00001664, -0.00001103, -0.00006135, -0.00015812, -0.00019482, -0.00009599, 0.00003396, 0.00002359, -0.00010478, -0.00014462, -0.00003093, 0.00001172, -0.00020457, -0.00047475, -0.00040074, 0.00005309, 0.00043767, 0.00036497, 0.00000001, -0.00017846, -0.00001082, 0.00018439, 0.00009433, -0.00017592, -0.00029967, -0.00018051, -0.00002190, -0.00000375, -0.00007882, -0.00011505, -0.00008593, },
		{ -0.00006678, -0.00005351, -0.00006494, -0.00006179, -0.00002240, -0.00003286, -0.00014980, -0.00024373, -0.00014862, 0.00005488, 0.00010425, -0.00004677, -0.00010187, 0.00015494, 0.00046932, 0.00039440, -0.00008393, -0.00046740, -0.00037212, -0.00000740, 0.00014695, -0.00001729, -0.00016956, -0.00007040, 0.00010871, 0.00008938, -0.00010349, -0.00021878, -0.00015412, -0.00005016, -0.00003553, -0.00006745, },
		{ -0.00017194, -0.00015486, -0.00006768, -0.00000932, -0.00004587, -0.00007167, 0.00000698, 0.00006751, -0.00005284, -0.00024014, -0.00021063, 0.00003144, 0.00010422, -0.00022221, -0.00060759, -0.00051856, 0.00005083, 0.00052934, 0.00048957, 0.00015010, 0.00000010, 0.00013612, 0.00023490, 0.00009441, -0.00010462, -0.00013506, -0.00006385, -0.00009306, -0.00022397, -0.00028565, -0.00022600, -0.00016678, },
		{ -0.00013111, -0.00012423, -0.00021852, -0.00022984, -0.00008783, 0.00001769, -0.00006363, -0.00017376, -0.00009198, 0.00006813, -0.00000069, -0.00026444, -0.00028135, 0.00016601, 0.00063225, 0.00049913, -0.00020324, -0.00079005, -0.00076374, -0.00036072, -0.00011974, -0.00016690, -0.00019654, -0.00002582, 0.00016232, 0.00015977, 0.00003523, -0.00004009, -0.00007035, -0.00015364, -0.00024739, -0.00022756, },
		{ -0.00046740, -0.00036909, -0.00016885, -0.00015878, -0.00035365, -0.00043615, -0.00023712, -0.00002808, -0.00011940, -0.00035372, -0.00032910, -0.00007145, -0.00008880, -0.00060833, -0.00109541, -0.00085863, 0.00002158, 0.00073700, 0.00072860, 0.00027943, 0.00000843, 0.00007876, 0.00017801, 0.00010462, 0.00001874, 0.00010008, 0.00025830, 0.00030901, 0.00023260, 0.00010349, -0.00008258, -0.00032421, },
		{ -0.00002930, -0.00047325, -0.00074201, -0.00057249, -0.00017410, -0.00004424, -0.00029908, -0.00047895, -0.00020788, 0.00021878, 0.00024852, -0.00006231, -0.00005450, 0.00055760, 0.00113590, 0.00087622, -0.00011253, -0.00089721, -0.00085264, -0.00031175, 0.00002433, -0.00003240, -0.00013947, -0.00006064, 0.00006044, 0.00004806, -0.00001436, 0.00003698, 0.00018934, 0.00031854, 0.00035923, 0.00026895, },
		{ 0.00066248, 0.00071853, 0.00038264, -0.00030625, -0.00084011, -0.00071227, -0.00011197, 0.00019399, -0.00016662, -0.00064049, -0.00051231, 0.00002715, 0.00007321, -0.00070058, -0.00144962, -0.00117319, -0.00000767, 0.00093389, 0.00091314, 0.00030541, -0.00008929, -0.00004381, 0.00007436, 0.00000143, -0.00011852, -0.00009412, 0.00000757, 0.00003750, 0.00001820, 0.00007045, 0.00022755, 0.00044544, },
		{ 0.00015030, 0.00032959, 0.00064231, 0.00085266, 0.00051288, -0.00048023, -0.00145444, -0.00151290, -0.00064566, 0.00010026, -0.00011980, -0.00080363, -0.00071058, 0.00044289, 0.00146210, 0.00110595, -0.00037233, -0.00155588, -0.00154087, -0.00077913, -0.00023043, -0.00019268, -0.00026881, -0.00015130, 0.00002438, 0.00005568, 0.00000354, 0.00002065, 0.00009514, 0.00012418, 0.00010003, 0.00009161, },
		{ 0.00008771, 0.00013298, 0.00018529, 0.00037857, 0.00085992, 0.00138405, 0.00123245, -0.00001921, -0.00165072, -0.00223998, -0.00128607, -0.00003245, -0.00014395, -0.00165365, -0.00279748, -0.00207642, -0.00000993, 0.00154976, 0.00153893, 0.00058944, -0.00009124, -0.00012367, 0.00002441, -0.00003500, -0.00018187, -0.00019142, -0.00010935, -0.00008900, -0.00012351, -0.00011179, -0.00004213, 0.00003026, },
		{ 0.00002954, 0.00003824, 0.00008457, 0.00012049, 0.00009509, 0.00019684, 0.00077915, 0.00171501, 0.00201782, 0.00071989, -0.00165273, -0.00303430, -0.00186626, 0.00093260, 0.00265226, 0.00162920, -0.00103375, -0.00284193, -0.00259457, -0.00123567, -0.00029472, -0.00020253, -0.00035439, -0.00025320, -0.00002147, 0.00006654, 0.00003345, 0.00004256, 0.00010275, 0.00011950, 0.00008164, 0.00004513, },
		{ -0.00026591, -0.00013915, -0.00002888, 0.00010510, 0.00032745, 0.00054896, 0.00058885, 0.00061216, 0.00121785, 0.00256808, 0.00346513, 0.00205009, -0.00198631, -0.00622185, -0.00733532, -0.00438637, 0.00014269, 0.00283415, 0.00253341, 0.00080408, -0.00033913, -0.00041851, -0.00022369, -0.00036079, -0.00066578, -0.00078791, -0.00073205, -0.00068956, -0.00069213, -0.00064520, -0.00052911, -0.00039619, },
		{ 0.00029199, 0.00035865, 0.00042650, 0.00046960, 0.00044034, 0.00042288, 0.00059618, 0.00090601, 0.00103567, 0.00106051, 0.00190739, 0.00441185, 0.00758631, 0.00848701, 0.00475660, -0.00253658, -0.00921348, -0.01157736, -0.00953224, -0.00599705, -0.00368551, -0.00298947, -0.00275059, -0.00216788, -0.00143361, -0.00096284, -0.00073797, -0.00051539, -0.00023963, -0.00001711, 0.00011661, 0.00021174, },
		{ 0.00107864, 0.00109652, 0.00114351, 0.00124322, 0.00144132, 0.00167610, 0.00180694, 0.00189076, 0.00221218, 0.00278418, 0.00307771, 0.00283994, 0.00322792, 0.00630948, 0.01263220, 0.01948072, 0.02247966, 0.01948072, 0.01263220, 0.00630948, 0.00322792, 0.00283994, 0.00307771, 0.00278418, 0.00221218, 0.00189076, 0.00180694, 0.00167610, 0.00144132, 0.00124322, 0.00114351, 0.00109652, },
		{ 0.00029199, 0.00021174, 0.00011661, -0.00001711, -0.00023963, -0.00051539, -0.00073797, -0.00096284, -0.00143361, -0.00216788, -0.00275059, -0.00298947, -0.00368551, -0.00599705, -0.00953224, -0.01157736, -0.00921348, -0.00253658, 0.00475660, 0.00848701, 0.00758631, 0.00441185, 0.00190739, 0.00106051, 0.00103567, 0.00090601, 0.00059618, 0.00042288, 0.00044034, 0.00046960, 0.00042650, 0.00035865, },
		{ -0.00026591, -0.00039619, -0.00052911, -0.00064520, -0.00069213, -0.00068956, -0.00073205, -0.00078791, -0.00066578, -0.00036079, -0.00022369, -0.00041851, -0.00033913, 0.00080408, 0.00253341, 0.00283415, 0.00014269, -0.00438637, -0.00733532, -0.00622185, -0.00198631, 0.00205009, 0.00346513, 0.00256808, 0.00121785, 0.00061216, 0.00058885, 0.00054896, 0.00032745, 0.00010510, -0.00002888, -0.00013915, },
		{ 0.00002954, 0.00004513, 0.00008164, 0.00011950, 0.00010275, 0.00004256, 0.00003345, 0.00006654, -0.00002147, -0.00025320, -0.00035439, -0.00020253, -0.00029472, -0.00123567, -0.00259457, -0.00284193, -0.00103375, 0.00162920, 0.00265226, 0.00093260, -0.00186626, -0.00303430, -0.00165273, 0.00071989, 0.00201782, 0.00171501, 0.00077915, 0.00019684, 0.00009509, 0.00012049, 0.00008457, 0.00003824, },
		{ 0.00008771, 0.00003026, -0.00004213, -0.00011179, -0.00012351, -0.00008900, -0.00010935, -0.00019142, -0.00018187, -0.00003500, 0.00002441, -0.00012367, -0.00009124, 0.00058944, 0.00153893, 0.00154976, -0.00000993, -0.00207642, -0.00279748, -0.00165365, -0.00014395, -0.00003245, -0.00128607, -0.00223998, -0.00165072, -0.00001921, 0.00123245, 0.00138405, 0.00085992, 0.00037857, 0.00018529, 0.00013298, },
		{ 0.00015030, 0.00009161, 0.00010003, 0.00012418, 0.00009514, 0.00002065, 0.00000354, 0.00005568, 0.00002438, -0.00015130, -0.00026881, -0.00019268, -0.00023043, -0.00077913, -0.00154087, -0.00155588, -0.00037233, 0.00110595, 0.00146210, 0.00044289, -0.00071058, -0.00080363, -0.00011980, 0.00010026, -0.00064566, -0.00151290, -0.00145444, -0.00048023, 0.00051288, 0.00085266, 0.00064231, 0.00032959, },
		{ 0.00066248, 0.00044544, 0.00022755, 0.00007045, 0.00001820, 0.00003750, 0.00000757, -0.00009412, -0.00011852, 0.00000143, 0.00007436, -0.00004381, -0.00008929, 0.00030541, 0.00091314, 0.00093389, -0.00000767, -0.00117319, -0.00144962, -0.00070058, 0.00007321, 0.00002715, -0.00051231, -0.00064049, -0.00016662, 0.00019399, -0.00011197, -0.00071227, -0.00084011, -0.00030625, 0.00038264, 0.00071853, },
		{ -0.00002930, 0.00026895, 0.00035923, 0.00031854, 0.00018934, 0.00003698, -0.00001436, 0.00004806, 0.00006044, -0.00006064, -0.00013947, -0.00003240, 0.00002433, -0.00031175, -0.00085264, -0.00089721, -0.00011253, 0.00087622, 0.00113590, 0.00055760, -0.00005450, -0.00006231, 0.00024852, 0.00021878, -0.00020788, -0.00047895, -0.00029908, -0.00004424, -0.00017410, -0.00057249, -0.00074201, -0.00047325, },
		{ -0.00046740, -0.00032421, -0.00008258, 0.00010349, 0.00023260, 0.00030901, 0.00025830, 0.00010008, 0.00001874, 0.00010462, 0.00017801, 0.00007876, 0.00000843, 0.00027943, 0.00072860, 0.00073700, 0.00002158, -0.00085863, -0.00109541, -0.00060833, -0.00008880, -0.00007145, -0.00032910, -0.00035372, -0.00011940, -0.00002808, -0.00023712, -0.00043615, -0.00035365, -0.00015878, -0.00016885, -0.00036909, },
		{ -0.00013111, -0.00022756, -0.00024739, -0.00015364, -0.00007035, -0.00004009, 0.00003523, 0.00015977, 0.00016232, -0.00002582, -0.00019654, -0.00016690, -0.00011974, -0.00036072, -0.00076374, -0.00079005, -0.00020324, 0.00049913, 0.00063225, 0.00016601, -0.00028135, -0.00026444, -0.00000069, 0.00006813, -0.00009198, -0.00017376, -0.00006363, 0.00001769, -0.00008783, -0.00022984, -0.00021852, -0.00012423, },
		{ -0.00017194, -0.00016678, -0.00022600, -0.00028565, -0.00022397, -0.00009306, -0.00006385, -0.00013506, -0.00010462, 0.00009441, 0.00023490, 0.00013612, 0.00000010, 0.00015010, 0.00048957, 0.00052934, 0.00005083, -0.00051856, -0.00060759, -0.00022221, 0.00010422, 0.00003144, -0.00021063, -0.00024014, -0.00005284, 0.00006751, 0.00000698, -0.00007167, -0.00004587, -0.00000932, -0.00006768, -0.00015486, },
		{ -0.00006678, -0.00006745, -0.00003553, -0.00005016, -0.00015412, -0.00021878, -0.00010349, 0.00008938, 0.00010871, -0.00007040, -0.00016956, -0.00001729, 0.00014695, -0.00000740, -0.00037212, -0.00046740, -0.00008393, 0.00039440, 0.00046932, 0.00015494, -0.00010187, -0.00004677, 0.00010425, 0.00005488, -0.00014862, -0.00024373, -0.00014980, -0.00003286, -0.00002240, -0.00006179, -0.00006494, -0.00005351, },
		{ -0.00004455, -0.00008593, -0.00011505, -0.00007882, -0.00000375, -0.00002190, -0.00018051, -0.00029967, -0.00017592, 0.00009433, 0.00018439, -0.00001082, -0.00017846, 0.00000001, 0.00036497, 0.00043767, 0.00005309, -0.00040074, -0.00047475, -0.00020457, 0.00001172, -0.00003093, -0.00014462, -0.00010478, 0.00002359, 0.00003396, -0.00009599, -0.00019482, -0.00015812, -0.00006135, -0.00001103, -0.00001664, },
		{ -0.00002674, 0.00001647, 0.00000158, -0.00006753, -0.00011822, -0.00005805, 0.00007381, 0.00009105, -0.00009856, -0.00030001, -0.00023528, 0.00005662, 0.00020095, -0.00003675, -0.00039739, -0.00042228, -0.00004170, 0.00032954, 0.00031780, 0.00002745, -0.00016263, -0.00010225, 0.00001587, 0.00000008, -0.00008782, -0.00009049, -0.00001284, 0.00000805, -0.00007718, -0.00017026, -0.00017529, -0.00010530, },
		{ -0.00013831, -0.00011383, -0.00004083, 0.00003082, 0.00003070, -0.00006426, -0.00015588, -0.00010278, 0.00007565, 0.00015783, -0.00001215, -0.00026399, -0.00025786, 0.00007144, 0.00037301, 0.00028743, -0.00009953, -0.00036003, -0.00025273, 0.00001637, 0.00010787, -0.00002587, -0.00015390, -0.00011949, -0.00001249, 0.00001840, -0.00002210, -0.00003068, 0.00001549, 0.00003774, -0.00001404, -0.00009620, },
		{ -0.00004525, -0.00009974, -0.00015367, -0.00016280, -0.00009791, 0.00000511, 0.00003987, -0.00005721, -0.00018525, -0.00015935, 0.00003983, 0.00018028, 0.00004425, -0.00025881, -0.00037756, -0.00014760, 0.00018504, 0.00025885, 0.00003633, -0.00017992, -0.00014993, 0.00002827, 0.00009904, -0.00001044, -0.00014322, -0.00015648, -0.00008455, -0.00003916, -0.00004854, -0.00005858, -0.00003997, -0.00002391, },
		},
		{ { -0.00004951, -0.00003092, -0.00004260, -0.00007584, -0.00002468, 0.00000716, -0.00007139, -0.00006446, 0.00002874, -0.00007969, -0.00022637, -0.00001059, 0.00022026, -0.00004734, -0.00030402, 0.00004844, 0.00038316, 0.00004844, -0.00030402, -0.00004734, 0.00022026, -0.00001059, -0.00022637, -0.00007969, 0.00002874, -0.00006446, -0.00007139, 0.00000716, -0.00002468, -0.00007584, -0.00004260, -0.00003092, },
		{ -0.00004899, -0.00008079, -0.00006511, -0.00002244, -0.00007166, -0.00009429, -0.00001575, -0.00008131, -0.00025729, -0.00013452, 0.00013610, -0.00000467, -0.00029684, -0.00008670, 0.00025454, 0.00002599, -0.00031561, -0.00007120, 0.00027131, 0.00007261, -0.00022124, -0.00011297, 0.00003370, -0.00006481, -0.00010580, -0.00000444, -0.00001529, -0.00009813, -0.00007191, -0.00003243, -0.00006163, -0.00006133, },
		{ -0.00012487, -0.00008644, -0.00009931, -0.00014000, -0.00008641, -0.00010496, -0.00028095, -0.00025534, 0.00002672, 0.00004295, -0.00021391, -0.00011491, 0.00026515, 0.00022691, -0.00009943, -0.00005455, 0.00014324, -0.00006384, -0.00033686, -0.00018902, 0.00001773, -0.00007655, -0.00014653, -0.00002053, -0.00000210, -0.00011563, -0.00010659, -0.00003908, -0.00007468, -0.00010441, -0.00007669, -0.00010007, },
		{ -0.00006643, -0.00010219, -0.00008199, -0.00006171, -0.00021256, -0.00029057, -0.00006258, 0.00008465, -0.00009480, -0.00011902, 0.00018766, 0.00022981, -0.00013804, -0.00030908, -0.00015552, -0.00010601, -0.00010317, 0.00012258, 0.00025230, 0.00003096, -0.00012298, 0.00001094, 0.00004761, -0.00011223, -0.00013463, -0.00003478, -0.00006442, -0.00012645, -0.00008193, -0.00007496, -0.00012937, -0.00010277, },
		{ -0.00011209, -0.00008245, -0.00018875, -0.00035552, -0.00023389, 0.00001678, -0.00000620, -0.00007469, 0.00014799, 0.00028369, 0.00001531, -0.00019141, 0.00000815, 0.00026089, 0.00030425, 0.00023306, -0.00000951, -0.00037811, -0.00042844, -0.00010078, 0.00004081, -0.00014141, -0.00017813, -0.00001528, -0.00001518, -0.00011535, -0.00006513, -0.00001916, -0.00010710, -0.00012502, -0.00005452, -0.00007766, },
		{ -0.00015985, -0.00036532, -0.00038474, -0.00008928, 0.00008590, 0.00004331, 0.00013904, 0.00027293, 0.00007827, -0.00019608, -0.00012069, 0.00008721, 0.00002855, -0.00024312, -0.00049141, -0.00050212, -0.00009454, 0.00043195, 0.00043532, -0.00001369, -0.00018527, 0.00000862, 0.00001623, -0.00016603, -0.00014114, -0.00004496, -0.00015305, -0.00023114, -0.00012339, -0.00009280, -0.00016504, -0.00013529, },
		{ -0.00051267, -0.00028001, 0.00008956, 0.00016224, 0.00014502, 0.00021675, 0.00011163, -0.00013974, -0.00013520, 0.00008623, 0.00012166, -0.00005369, -0.00012457, 0.00008827, 0.00046172, 0.00054680, -0.00000744, -0.00071021, -0.00070870, -0.00017212, 0.00000378, -0.00021541, -0.00018303, 0.00002794, -0.00005507, -0.00020807, -0.00008159, 0.00002176, -0.00009541, -0.00012135, -0.00008079, -0.00030047, },
		{ -0.00000457, 0.00031664, 0.00026933, 0.00023251, 0.00018456, -0.00000272, -0.00006875, 0.00009210, 0.00015247, -0.00001943, -0.00014788, -0.00005481, 0.00007574, -0.00009052, -0.00056359, -0.00071981, -0.00004239, 0.00080857, 0.00077103, 0.00012481, -0.00004412, 0.00021928, 0.00011372, -0.00020893, -0.00013791, 0.00004084, -0.00011431, -0.00023727, -0.00013116, -0.00028721, -0.00067011, -0.00057196, },
		{ 0.00051056, 0.00031152, 0.00023938, 0.00013601, 0.00004124, 0.00012939, 0.00021782, 0.00010822, -0.00000662, 0.00008815, 0.00025684, 0.00023633, 0.00007631, 0.00020491, 0.00074043, 0.00090077, 0.00002158, -0.00103446, -0.00096083, -0.00018243, -0.00006137, -0.00048595, -0.00042141, -0.00005292, -0.00015166, -0.00037224, -0.00019721, -0.00020319, -0.00071788, -0.00085511, -0.00016273, 0.00049543, },
		{ 0.00025130, 0.00021559, 0.00012697, 0.00011059, 0.00015673, 0.00006431, -0.00009103, -0.00008547, 0.00001367, -0.00004130, -0.00024999, -0.00032313, -0.00021290, -0.00042263, -0.00113784, -0.00136846, -0.00027338, 0.00102582, 0.00086914, -0.00018919, -0.00040803, 0.00009847, 0.00002767, -0.00040915, -0.00026176, -0.00007902, -0.00068540, -0.00116999, -0.00047273, 0.00059467, 0.00081672, 0.00044802, },
		{ 0.00010845, 0.00000435, -0.00001285, -0.00006225, -0.00018623, -0.00019283, -0.00006467, -0.00002122, -0.00011629, -0.00012472, 0.00004287, 0.00008750, -0.00008415, 0.00015778, 0.00108086, 0.00141940, 0.00005060, -0.00164461, -0.00153328, -0.00024079, 0.00004504, -0.00057527, -0.00046018, 0.00003811, -0.00054308, -0.00146676, -0.00093415, 0.00056762, 0.00115699, 0.00063739, 0.00018326, 0.00014949, },
		{ 0.00016266, 0.00017085, 0.00011188, 0.00008972, 0.00019613, 0.00026054, 0.00016760, 0.00007719, 0.00010960, 0.00006959, -0.00016732, -0.00023817, 0.00001104, -0.00023053, -0.00139537, -0.00193803, -0.00036172, 0.00170564, 0.00155559, -0.00010762, -0.00048591, 0.00027329, -0.00025690, -0.00178874, -0.00165509, 0.00030594, 0.00152510, 0.00101469, 0.00025653, 0.00020038, 0.00031503, 0.00022408, },
		{ 0.00015525, 0.00006411, 0.00008961, 0.00014789, 0.00007942, -0.00002460, -0.00001287, 0.00001081, -0.00008548, -0.00007405, 0.00023011, 0.00036212, 0.00003251, 0.00020755, 0.00151841, 0.00211596, 0.00005267, -0.00269884, -0.00258102, -0.00041945, -0.00023941, -0.00229305, -0.00266686, -0.00011087, 0.00211022, 0.00182185, 0.00064226, 0.00037420, 0.00059204, 0.00049100, 0.00027490, 0.00021713, },
		{ 0.00014995, 0.00026956, 0.00028924, 0.00020776, 0.00018344, 0.00018831, 0.00008560, -0.00000750, 0.00008907, 0.00008974, -0.00032501, -0.00064307, -0.00044065, -0.00084606, -0.00268481, -0.00362200, -0.00097228, 0.00279705, 0.00243620, -0.00175017, -0.00369740, -0.00078380, 0.00268713, 0.00275659, 0.00094522, 0.00020214, 0.00049555, 0.00049640, 0.00018943, 0.00010983, 0.00015224, 0.00011815, },
		{ 0.00061540, 0.00040793, 0.00021741, 0.00010360, -0.00004636, -0.00019106, -0.00015726, -0.00008512, -0.00026915, -0.00043399, -0.00014284, 0.00005940, -0.00034014, 0.00010916, 0.00259173, 0.00397681, -0.00010008, -0.00725360, -0.00939648, -0.00379014, 0.00294959, 0.00442472, 0.00216113, 0.00084103, 0.00126678, 0.00152014, 0.00113391, 0.00091417, 0.00094258, 0.00084537, 0.00070609, 0.00068390, },
		{ -0.00011143, -0.00016293, -0.00018980, -0.00023516, -0.00017253, -0.00005654, -0.00016054, -0.00040638, -0.00043116, -0.00049970, -0.00113538, -0.00167067, -0.00144408, -0.00253945, -0.00723362, -0.01181482, -0.00871283, 0.00255303, 0.01247585, 0.01326013, 0.00792739, 0.00396169, 0.00331688, 0.00316749, 0.00217655, 0.00134548, 0.00106769, 0.00078835, 0.00043042, 0.00027649, 0.00022175, 0.00005816, },
		{ -0.00088529, -0.00092377, -0.00094679, -0.00091939, -0.00103940, -0.00133123, -0.00148654, -0.00152929, -0.00185538, -0.00212898, -0.00166164, -0.00137930, -0.00246442, -0.00208260, 0.00468306, 0.01595935, 0.02175708, 0.01595935, 0.00468306, -0.00208260, -0.00246442, -0.00137930, -0.00166164, -0.00212898, -0.00185538, -0.00152929, -0.00148654, -0.00133123, -0.00103940, -0.00091939, -0.00094679, -0.00092377, },
		{ -0.00011143, 0.00005816, 0.00022175, 0.00027649, 0.00043042, 0.00078835, 0.00106769, 0.00134548, 0.00217655, 0.00316749, 0.00331688, 0.00396169, 0.00792739, 0.01326013, 0.01247585, 0.00255303, -0.00871283, -0.01181482, -0.00723362, -0.00253945, -0.00144408, -0.00167067, -0.00113538, -0.00049970, -0.00043116, -0.00040638, -0.00016054, -0.00005654, -0.00017253, -0.00023516, -0.00018980, -0.00016293, },
		{ 0.00061540, 0.00068390, 0.00070609, 0.00084537, 0.00094258, 0.00091417, 0.00113391, 0.00152014, 0.00126678, 0.00084103, 0.00216113, 0.00442472, 0.00294959, -0.00379014, -0.00939648, -0.00725360, -0.00010008, 0.00397681, 0.00259173, 0.00010916, -0.00034014, 0.00005940, -0.00014284, -0.00043399, -0.00026915, -0.00008512, -0.00015726, -0.00019106, -0.00004636, 0.00010360, 0.00021741, 0.00040793, },
		{ 0.00014995, 0.00011815, 0.00015224, 0.00010983, 0.00018943, 0.00049640, 0.00049555, 0.00020214, 0.00094522, 0.00275659, 0.00268713, -0.00078380, -0.00369740, -0.00175017, 0.00243620, 0.00279705, -0.00097228, -0.00362200, -0.00268481, -0.00084606, -0.00044065, -0.00064307, -0.00032501, 0.00008974, 0.00008907, -0.00000750, 0.00008560, 0.00018831, 0.00018344, 0.00020776, 0.00028924, 0.00026956, },
		{ 0.00015525, 0.00021713, 0.00027490, 0.00049100, 0.00059204, 0.00037420, 0.00064226, 0.00182185, 0.00211022, -0.00011087, -0.00266686, -0.00229305, -0.00023941, -0.00041945, -0.00258102, -0.00269884, 0.00005267, 0.00211596, 0.00151841, 0.00020755, 0.00003251, 0.00036212, 0.00023011, -0.00007405, -0.00008548, 0.00001081, -0.00001287, -0.00002460, 0.00007942, 0.00014789, 0.00008961, 0.00006411, },
		{ 0.00016266, 0.00022408, 0.00031503, 0.00020038, 0.00025653, 0.00101469, 0.00152510, 0.00030594, -0.00165509, -0.00178874, -0.00025690, 0.00027329, -0.00048591, -0.00010762, 0.00155559, 0.00170564, -0.00036172, -0.00193803, -0.00139537, -0.00023053, 0.00001104, -0.00023817, -0.00016732, 0.00006959, 0.00010960, 0.00007719, 0.00016760, 0.00026054, 0.00019613, 0.00008972, 0.00011188, 0.00017085, },
		{ 0.00010845, 0.00014949, 0.00018326, 0.00063739, 0.00115699, 0.00056762, -0.00093415, -0.00146676, -0.00054308, 0.00003811, -0.00046018, -0.00057527, 0.00004504, -0.00024079, -0.00153328, -0.00164461, 0.00005060, 0.00141940, 0.00108086, 0.00015778, -0.00008415, 0.00008750, 0.00004287, -0.00012472, -0.00011629, -0.00002122, -0.00006467, -0.00019283, -0.00018623, -0.00006225, -0.00001285, 0.00000435, },
		{ 0.00025130, 0.00044802, 0.00081672, 0.00059467, -0.00047273, -0.00116999, -0.00068540, -0.00007902, -0.00026176, -0.00040915, 0.00002767, 0.00009847, -0.00040803, -0.00018919, 0.00086914, 0.00102582, -0.00027338, -0.00136846, -0.00113784, -0.00042263, -0.00021290, -0.00032313, -0.00024999, -0.00004130, 0.00001367, -0.00008547, -0.00009103, 0.00006431, 0.00015673, 0.00011059, 0.00012697, 0.00021559, },
		{ 0.00051056, 0.00049543, -0.00016273, -0.00085511, -0.00071788, -0.00020319, -0.00019721, -0.00037224, -0.00015166, -0.00005292, -0.00042141, -0.00048595, -0.00006137, -0.00018243, -0.00096083, -0.00103446, 0.00002158, 0.00090077, 0.00074043, 0.00020491, 0.00007631, 0.00023633, 0.00025684, 0.00008815, -0.00000662, 0.00010822, 0.00021782, 0.00012939, 0.00004124, 0.00013601, 0.00023938, 0.00031152, },
		{ -0.00000457, -0.00057196, -0.00067011, -0.00028721, -0.00013116, -0.00023727, -0.00011431, 0.00004084, -0.00013791, -0.00020893, 0.00011372, 0.00021928, -0.00004412, 0.00012481, 0.00077103, 0.00080857, -0.00004239, -0.00071981, -0.00056359, -0.00009052, 0.00007574, -0.00005481, -0.00014788, -0.00001943, 0.00015247, 0.00009210, -0.00006875, -0.00000272, 0.00018456, 0.00023251, 0.00026933, 0.00031664, },
		{ -0.00051267, -0.00030047, -0.00008079, -0.00012135, -0.00009541, 0.00002176, -0.00008159, -0.00020807, -0.00005507, 0.00002794, -0.00018303, -0.00021541, 0.00000378, -0.00017212, -0.00070870, -0.00071021, -0.00000744, 0.00054680, 0.00046172, 0.00008827, -0.00012457, -0.00005369, 0.00012166, 0.00008623, -0.00013520, -0.00013974, 0.00011163, 0.00021675, 0.00014502, 0.00016224, 0.00008956, -0.00028001, },
		{ -0.00015985, -0.00013529, -0.00016504, -0.00009280, -0.00012339, -0.00023114, -0.00015305, -0.00004496, -0.00014114, -0.00016603, 0.00001623, 0.00000862, -0.00018527, -0.00001369, 0.00043532, 0.00043195, -0.00009454, -0.00050212, -0.00049141, -0.00024312, 0.00002855, 0.00008721, -0.00012069, -0.00019608, 0.00007827, 0.00027293, 0.00013904, 0.00004331, 0.00008590, -0.00008928, -0.00038474, -0.00036532, },
		{ -0.00011209, -0.00007766, -0.00005452, -0.00012502, -0.00010710, -0.00001916, -0.00006513, -0.00011535, -0.00001518, -0.00001528, -0.00017813, -0.00014141, 0.00004081, -0.00010078, -0.00042844, -0.00037811, -0.00000951, 0.00023306, 0.00030425, 0.00026089, 0.00000815, -0.00019141, 0.00001531, 0.00028369, 0.00014799, -0.00007469, -0.00000620, 0.00001678, -0.00023389, -0.00035552, -0.00018875, -0.00008245, },
		{ -0.00006643, -0.00010277, -0.00012937, -0.00007496, -0.00008193, -0.00012645, -0.00006442, -0.00003478, -0.00013463, -0.00011223, 0.00004761, 0.00001094, -0.00012298, 0.00003096, 0.00025230, 0.00012258, -0.00010317, -0.00010601, -0.00015552, -0.00030908, -0.00013804, 0.00022981, 0.00018766, -0.00011902, -0.00009480, 0.00008465, -0.00006258, -0.00029057, -0.00021256, -0.00006171, -0.00008199, -0.00010219, },
		{ -0.00012487, -0.00010007, -0.00007669, -0.00010441, -0.00007468, -0.00003908, -0.00010659, -0.00011563, -0.00000210, -0.00002053, -0.00014653, -0.00007655, 0.00001773, -0.00018902, -0.00033686, -0.00006384, 0.00014324, -0.00005455, -0.00009943, 0.00022691, 0.00026515, -0.00011491, -0.00021391, 0.00004295, 0.00002672, -0.00025534, -0.00028095, -0.00010496, -0.00008641, -0.00014000, -0.00009931, -0.00008644, },
		{ -0.00004899, -0.00006133, -0.00006163, -0.00003243, -0.00007191, -0.00009813, -0.00001529, -0.00000444, -0.00010580, -0.00006481, 0.00003370, -0.00011297, -0.00022124, 0.00007261, 0.00027131, -0.00007120, -0.00031561, 0.00002599, 0.00025454, -0.00008670, -0.00029684, -0.00000467, 0.00013610, -0.00013452, -0.00025729, -0.00008131, -0.00001575, -0.00009429, -0.00007166, -0.00002244, -0.00006511, -0.00008079, },
		},
		{ { 0.00110013, 0.00012094, -0.00030981, -0.00010469, 0.00003289, -0.00015447, 0.00021180, -0.00008730, 0.00003818, 0.00000482, 0.00025984, -0.00008443, -0.00014504, -0.00005378, 0.00038485, -0.00015900, -0.00031927, -0.00015900, 0.00038485, -0.00005378, -0.00014504, -0.00008443, 0.00025984, 0.00000482, 0.00003818, -0.00008730, 0.00021180, -0.00015447, 0.00003289, -0.00010469, -0.00030981, 0.00012094, },
		{ 0.00017245, -0.00015552, -0.00014642, 0.00005366, -0.00009946, 0.00024920, -0.00014053, 0.00011951, 0.00017020, 0.00014711, -0.00025335, 0.00003402, 0.00017108, 0.00009181, -0.00039052, -0.00016232, 0.00017578, 0.00028974, -0.00040110, 0.00002105, 0.00009886, 0.00013176, -0.00014418, 0.00012408, -0.00007876, 0.00009423, -0.00014555, 0.00005270, -0.00010908, -0.00045620, 0.00005824, 0.00111802, },
		{ -0.00001364, -0.00017816, 0.00012636, -0.00008257, 0.00022868, -0.00007503, 0.00027772, 0.00014471, -0.00002678, -0.00016780, 0.00023924, 0.00001489, -0.00023031, -0.00027267, 0.00011834, 0.00025225, -0.00010740, -0.00041270, 0.00037312, -0.00003478, -0.00003396, -0.00009472, 0.00019747, -0.00019821, 0.00004616, -0.00006178, 0.00001255, -0.00014890, -0.00054120, -0.00004061, 0.00116501, 0.00021518, },
		{ -0.00018053, 0.00019569, -0.00011109, 0.00027108, 0.00009593, 0.00032843, -0.00000207, -0.00003564, 0.00000125, 0.00015740, -0.00028742, -0.00023089, 0.00003018, 0.00019482, 0.00003279, -0.00037834, -0.00001936, 0.00043293, -0.00037791, 0.00008705, -0.00003037, 0.00009824, -0.00026997, 0.00013555, -0.00005710, -0.00009365, -0.00013041, -0.00058303, -0.00022714, 0.00126472, 0.00026481, 0.00011454, },
		{ 0.00024407, -0.00011593, 0.00041006, 0.00024745, 0.00025203, -0.00009806, 0.00005275, -0.00002417, -0.00010675, -0.00036073, 0.00002469, 0.00016669, 0.00000696, -0.00018010, -0.00028133, 0.00036362, 0.00003596, -0.00045155, 0.00039938, -0.00025748, 0.00002792, -0.00015068, 0.00017219, -0.00019382, -0.00011165, -0.00006353, -0.00067451, -0.00046718, 0.00146281, 0.00029610, 0.00021597, -0.00015069, },
		{ -0.00004585, 0.00060225, 0.00031333, 0.00012293, -0.00014665, 0.00003566, -0.00021896, -0.00022673, -0.00016570, 0.00019643, 0.00002740, -0.00019336, -0.00014659, -0.00002197, 0.00036213, -0.00042929, -0.00004724, 0.00041489, -0.00065046, 0.00029891, -0.00007705, 0.00001340, -0.00027087, 0.00000531, -0.00011287, -0.00078210, -0.00065568, 0.00169759, 0.00026191, 0.00033120, -0.00007609, 0.00027463, },
		{ 0.00078702, 0.00025012, -0.00006314, -0.00024521, -0.00008865, -0.00033193, -0.00010630, 0.00001990, 0.00011397, -0.00016998, -0.00018278, 0.00003515, 0.00006069, 0.00008017, -0.00051331, 0.00045035, -0.00008407, -0.00067095, 0.00074808, -0.00039218, -0.00007545, -0.00015434, 0.00003361, -0.00030253, -0.00071244, -0.00084943, 0.00182844, 0.00024658, 0.00052561, -0.00000585, 0.00030715, 0.00015222, },
		{ -0.00002019, -0.00030475, -0.00035059, -0.00014952, -0.00026076, 0.00005105, 0.00000122, -0.00006678, -0.00019352, -0.00002331, 0.00010936, -0.00008062, -0.00010373, -0.00025135, 0.00054219, -0.00066930, -0.00008278, 0.00075647, -0.00090890, 0.00027476, -0.00015330, -0.00014561, -0.00035987, -0.00045864, -0.00129321, 0.00191226, 0.00042899, 0.00071105, 0.00000779, 0.00047630, 0.00049038, 0.00080545, },
		{ -0.00044793, -0.00033829, -0.00009487, -0.00014639, 0.00006543, -0.00012945, -0.00012414, -0.00006265, 0.00006617, 0.00000307, -0.00021469, -0.00003958, -0.00004117, 0.00024175, -0.00076896, 0.00061646, 0.00004104, -0.00092698, 0.00083862, -0.00066633, -0.00014312, -0.00016406, -0.00036949, -0.00200570, 0.00223368, 0.00075454, 0.00070815, 0.00015182, 0.00093012, 0.00073185, 0.00042804, -0.00043640, },
		{ -0.00024500, -0.00004626, -0.00011782, -0.00001573, -0.00018587, -0.00003855, 0.00003763, -0.00000309, -0.00012688, -0.00018452, 0.00007971, -0.00008758, -0.00000463, -0.00038860, 0.00074206, -0.00083919, -0.00011453, 0.00081779, -0.00139688, 0.00050995, -0.00021391, -0.00060717, -0.00257264, 0.00280568, 0.00090597, 0.00068356, 0.00077801, 0.00142435, 0.00042770, -0.00030469, -0.00067871, -0.00034961, },
		{ -0.00006726, -0.00013282, -0.00006018, -0.00012899, 0.00004882, 0.00000631, -0.00012267, -0.00013778, -0.00007984, 0.00006249, -0.00020254, 0.00005698, -0.00006137, 0.00026422, -0.00107598, 0.00090220, -0.00016144, -0.00138638, 0.00143523, -0.00111575, -0.00056391, -0.00280239, 0.00309921, 0.00095776, 0.00123810, 0.00175768, 0.00124166, -0.00052666, -0.00088300, -0.00048503, -0.00014936, -0.00023676, },
		{ -0.00011291, -0.00004107, -0.00005149, 0.00004011, -0.00009317, -0.00016353, -0.00006406, -0.00004923, -0.00002941, -0.00016143, 0.00021033, -0.00012756, -0.00019840, -0.00058891, 0.00118900, -0.00135930, -0.00018399, 0.00142619, -0.00244027, 0.00055132, -0.00349628, 0.00286144, 0.00183575, 0.00253588, 0.00210254, -0.00004111, -0.00146048, -0.00079855, -0.00006569, -0.00013929, -0.00032524, -0.00007062, },
		{ -0.00001111, -0.00002635, -0.00004129, -0.00015453, -0.00011916, -0.00001751, -0.00008797, -0.00003850, -0.00002930, 0.00015904, -0.00035752, -0.00019981, -0.00006939, 0.00063591, -0.00166146, 0.00130574, -0.00014826, -0.00265931, 0.00226188, -0.00581275, 0.00324941, 0.00437429, 0.00338119, 0.00084330, -0.00170255, -0.00147845, -0.00023889, 0.00008109, -0.00033417, -0.00032653, -0.00000770, -0.00007880, },
		{ -0.00005782, -0.00012457, -0.00014205, -0.00006234, -0.00002640, -0.00009283, 0.00003057, -0.00001283, -0.00005056, -0.00036984, 0.00004326, -0.00005205, 0.00004615, -0.00093057, 0.00166549, -0.00222383, -0.00082996, 0.00255377, -0.00935974, 0.00633098, 0.00758447, 0.00191712, -0.00149550, -0.00231941, -0.00057219, 0.00003074, -0.00016150, -0.00041668, -0.00017063, 0.00002352, -0.00006859, 0.00000440, },
		{ -0.00015949, -0.00008506, -0.00003732, -0.00003665, -0.00004109, 0.00004104, -0.00013455, -0.00015573, -0.00016096, 0.00008983, -0.00030970, 0.00005694, -0.00016663, 0.00064191, -0.00294795, 0.00184618, -0.00013625, -0.01142310, 0.01265369, 0.00852118, -0.00216370, -0.00284939, -0.00138971, 0.00020978, -0.00036049, -0.00033427, -0.00021766, -0.00004792, -0.00004203, -0.00004385, -0.00000287, -0.00012362, },
		{ -0.00002792, -0.00003563, -0.00001894, -0.00001166, -0.00005604, -0.00022338, -0.00007541, -0.00001545, -0.00009323, -0.00033431, 0.00038191, -0.00021389, -0.00052374, -0.00180104, 0.00287396, -0.00465367, -0.00908317, 0.01950222, 0.00482568, -0.00643737, -0.00166092, -0.00015596, 0.00002140, -0.00085810, -0.00006152, -0.00000864, -0.00010940, -0.00009755, 0.00001626, -0.00006609, -0.00018031, -0.00012756, },
		{ -0.00003738, -0.00000443, -0.00005493, -0.00016196, -0.00017526, 0.00002486, -0.00004821, -0.00019779, -0.00009998, 0.00036133, -0.00074584, -0.00063631, -0.00028223, 0.00115037, -0.00758120, -0.00243503, 0.02250115, -0.00243503, -0.00758120, 0.00115037, -0.00028223, -0.00063631, -0.00074584, 0.00036133, -0.00009998, -0.00019779, -0.00004821, 0.00002486, -0.00017526, -0.00016196, -0.00005493, -0.00000443, },
		{ -0.00002792, -0.00012756, -0.00018031, -0.00006609, 0.00001626, -0.00009755, -0.00010940, -0.00000864, -0.00006152, -0.00085810, 0.00002140, -0.00015596, -0.00166092, -0.00643737, 0.00482568, 0.01950222, -0.00908317, -0.00465367, 0.00287396, -0.00180104, -0.00052374, -0.00021389, 0.00038191, -0.00033431, -0.00009323, -0.00001545, -0.00007541, -0.00022338, -0.00005604, -0.00001166, -0.00001894, -0.00003563, },
		{ -0.00015949, -0.00012362, -0.00000287, -0.00004385, -0.00004203, -0.00004792, -0.00021766, -0.00033427, -0.00036049, 0.00020978, -0.00138971, -0.00284939, -0.00216370, 0.00852118, 0.01265369, -0.01142310, -0.00013625, 0.00184618, -0.00294795, 0.00064191, -0.00016663, 0.00005694, -0.00030970, 0.00008983, -0.00016096, -0.00015573, -0.00013455, 0.00004104, -0.00004109, -0.00003665, -0.00003732, -0.00008506, },
		{ -0.00005782, 0.00000440, -0.00006859, 0.00002352, -0.00017063, -0.00041668, -0.00016150, 0.00003074, -0.00057219, -0.00231941, -0.00149550, 0.00191712, 0.00758447, 0.00633098, -0.00935974, 0.00255377, -0.00082996, -0.00222383, 0.00166549, -0.00093057, 0.00004615, -0.00005205, 0.00004326, -0.00036984, -0.00005056, -0.00001283, 0.00003057, -0.00009283, -0.00002640, -0.00006234, -0.00014205, -0.00012457, },
		{ -0.00001111, -0.00007880, -0.00000770, -0.00032653, -0.00033417, 0.00008109, -0.00023889, -0.00147845, -0.00170255, 0.00084330, 0.00338119, 0.00437429, 0.00324941, -0.00581275, 0.00226188, -0.00265931, -0.00014826, 0.00130574, -0.00166146, 0.00063591, -0.00006939, -0.00019981, -0.00035752, 0.00015904, -0.00002930, -0.00003850, -0.00008797, -0.00001751, -0.00011916, -0.00015453, -0.00004129, -0.00002635, },
		{ -0.00011291, -0.00007062, -0.00032524, -0.00013929, -0.00006569, -0.00079855, -0.00146048, -0.00004111, 0.00210254, 0.00253588, 0.00183575, 0.00286144, -0.00349628, 0.00055132, -0.00244027, 0.00142619, -0.00018399, -0.00135930, 0.00118900, -0.00058891, -0.00019840, -0.00012756, 0.00021033, -0.00016143, -0.00002941, -0.00004923, -0.00006406, -0.00016353, -0.00009317, 0.00004011, -0.00005149, -0.00004107, },
		{ -0.00006726, -0.00023676, -0.00014936, -0.00048503, -0.00088300, -0.00052666, 0.00124166, 0.00175768, 0.00123810, 0.00095776, 0.00309921, -0.00280239, -0.00056391, -0.00111575, 0.00143523, -0.00138638, -0.00016144, 0.00090220, -0.00107598, 0.00026422, -0.00006137, 0.00005698, -0.00020254, 0.00006249, -0.00007984, -0.00013778, -0.00012267, 0.00000631, 0.00004882, -0.00012899, -0.00006018, -0.00013282, },
		{ -0.00024500, -0.00034961, -0.00067871, -0.00030469, 0.00042770, 0.00142435, 0.00077801, 0.00068356, 0.00090597, 0.00280568, -0.00257264, -0.00060717, -0.00021391, 0.00050995, -0.00139688, 0.00081779, -0.00011453, -0.00083919, 0.00074206, -0.00038860, -0.00000463, -0.00008758, 0.00007971, -0.00018452, -0.00012688, -0.00000309, 0.00003763, -0.00003855, -0.00018587, -0.00001573, -0.00011782, -0.00004626, },
		{ -0.00044793, -0.00043640, 0.00042804, 0.00073185, 0.00093012, 0.00015182, 0.00070815, 0.00075454, 0.00223368, -0.00200570, -0.00036949, -0.00016406, -0.00014312, -0.00066633, 0.00083862, -0.00092698, 0.00004104, 0.00061646, -0.00076896, 0.00024175, -0.00004117, -0.00003958, -0.00021469, 0.00000307, 0.00006617, -0.00006265, -0.00012414, -0.00012945, 0.00006543, -0.00014639, -0.00009487, -0.00033829, },
		{ -0.00002019, 0.00080545, 0.00049038, 0.00047630, 0.00000779, 0.00071105, 0.00042899, 0.00191226, -0.00129321, -0.00045864, -0.00035987, -0.00014561, -0.00015330, 0.00027476, -0.00090890, 0.00075647, -0.00008278, -0.00066930, 0.00054219, -0.00025135, -0.00010373, -0.00008062, 0.00010936, -0.00002331, -0.00019352, -0.00006678, 0.00000122, 0.00005105, -0.00026076, -0.00014952, -0.00035059, -0.00030475, },
		{ 0.00078702, 0.00015222, 0.00030715, -0.00000585, 0.00052561, 0.00024658, 0.00182844, -0.00084943, -0.00071244, -0.00030253, 0.00003361, -0.00015434, -0.00007545, -0.00039218, 0.00074808, -0.00067095, -0.00008407, 0.00045035, -0.00051331, 0.00008017, 0.00006069, 0.00003515, -0.00018278, -0.00016998, 0.00011397, 0.00001990, -0.00010630, -0.00033193, -0.00008865, -0.00024521, -0.00006314, 0.00025012, },
		{ -0.00004585, 0.00027463, -0.00007609, 0.00033120, 0.00026191, 0.00169759, -0.00065568, -0.00078210, -0.00011287, 0.00000531, -0.00027087, 0.00001340, -0.00007705, 0.00029891, -0.00065046, 0.00041489, -0.00004724, -0.00042929, 0.00036213, -0.00002197, -0.00014659, -0.00019336, 0.00002740, 0.00019643, -0.00016570, -0.00022673, -0.00021896, 0.00003566, -0.00014665, 0.00012293, 0.00031333, 0.00060225, },
		{ 0.00024407, -0.00015069, 0.00021597, 0.00029610, 0.00146281, -0.00046718, -0.00067451, -0.00006353, -0.00011165, -0.00019382, 0.00017219, -0.00015068, 0.00002792, -0.00025748, 0.00039938, -0.00045155, 0.00003596, 0.00036362, -0.00028133, -0.00018010, 0.00000696, 0.00016669, 0.00002469, -0.00036073, -0.00010675, -0.00002417, 0.00005275, -0.00009806, 0.00025203, 0.00024745, 0.00041006, -0.00011593, },
		{ -0.00018053, 0.00011454, 0.00026481, 0.00126472, -0.00022714, -0.00058303, -0.00013041, -0.00009365, -0.00005710, 0.00013555, -0.00026997, 0.00009824, -0.00003037, 0.00008705, -0.00037791, 0.00043293, -0.00001936, -0.00037834, 0.00003279, 0.00019482, 0.00003018, -0.00023089, -0.00028742, 0.00015740, 0.00000125, -0.00003564, -0.00000207, 0.00032843, 0.00009593, 0.00027108, -0.00011109, 0.00019569, },
		{ -0.00001364, 0.00021518, 0.00116501, -0.00004061, -0.00054120, -0.00014890, 0.00001255, -0.00006178, 0.00004616, -0.00019821, 0.00019747, -0.00009472, -0.00003396, -0.00003478, 0.00037312, -0.00041270, -0.00010740, 0.00025225, 0.00011834, -0.00027267, -0.00023031, 0.00001489, 0.00023924, -0.00016780, -0.00002678, 0.00014471, 0.00027772, -0.00007503, 0.00022868, -0.00008257, 0.00012636, -0.00017816, },
		{ 0.00017245, 0.00111802, 0.00005824, -0.00045620, -0.00010908, 0.00005270, -0.00014555, 0.00009423, -0.00007876, 0.00012408, -0.00014418, 0.00013176, 0.00009886, 0.00002105, -0.00040110, 0.00028974, 0.00017578, -0.00016232, -0.00039052, 0.00009181, 0.00017108, 0.00003402, -0.00025335, 0.00014711, 0.00017020, 0.00011951, -0.00014053, 0.00024920, -0.00009946, 0.00005366, -0.00014642, -0.00015552, },
		},
		{ { -0.00090678, -0.00011679, 0.00062868, 0.00015319, 0.00014614, 0.00016661, 0.00012305, 0.00024773, 0.00049111, -0.00000787, -0.00049754, -0.00015515, -0.00012024, -0.00006206, -0.00011034, -0.00005302, -0.00006966, -0.00005302, -0.00011034, -0.00006206, -0.00012024, -0.00015515, -0.00049754, -0.00000787, 0.00049111, 0.00024773, 0.00012305, 0.00016661, 0.00014614, 0.00015319, 0.00062868, -0.00011679, },
		{ -0.00016830, 0.00042117, 0.00027275, 0.00005493, 0.00017471, 0.00001886, 0.00021192, 0.00029196, 0.00031322, -0.00026500, -0.00036071, -0.00009069, -0.00009789, -0.00007197, -0.00008486, -0.00005107, -0.00006535, -0.00008552, -0.00009834, -0.00008574, -0.00013050, -0.00028524, -0.00057516, 0.00047609, 0.00044457, 0.00016420, 0.00022814, 0.00020811, 0.00012146, 0.00069722, 0.00005283, -0.00094525, },
		{ 0.00023064, 0.00029240, 0.00008038, 0.00011567, 0.00000156, 0.00012319, 0.00021971, 0.00026580, 0.00010446, -0.00038025, -0.00019709, -0.00007779, -0.00008492, -0.00006923, -0.00006279, -0.00006567, -0.00006213, -0.00012491, -0.00006254, -0.00016017, -0.00006548, -0.00067320, -0.00018196, 0.00081339, 0.00019810, 0.00031920, 0.00026598, 0.00015565, 0.00071949, 0.00021647, -0.00096824, -0.00019516, },
		{ 0.00021090, 0.00013862, 0.00009346, -0.00004791, 0.00010672, 0.00011623, 0.00022887, 0.00017702, -0.00008490, -0.00036397, -0.00005761, -0.00012570, -0.00002664, -0.00009608, -0.00003650, -0.00008985, -0.00007049, -0.00013301, -0.00008788, -0.00010596, -0.00029021, -0.00087423, 0.00059145, 0.00065233, 0.00020466, 0.00048220, 0.00011334, 0.00085886, 0.00027128, -0.00094079, -0.00024049, 0.00011683, },
		{ 0.00007015, 0.00019983, -0.00017194, 0.00015279, 0.00002138, 0.00018081, 0.00015970, 0.00009016, -0.00024246, -0.00020857, -0.00007221, -0.00007595, -0.00004499, -0.00007604, -0.00006016, -0.00007747, -0.00011507, -0.00011844, -0.00007996, -0.00013408, -0.00073691, -0.00047585, 0.00117205, 0.00026093, 0.00058335, 0.00019305, 0.00095617, 0.00042531, -0.00106072, -0.00017781, -0.00003309, 0.00018659, },
		{ 0.00026423, -0.00017858, 0.00006031, 0.00010945, -0.00000655, 0.00023131, 0.00004746, 0.00000810, -0.00028670, -0.00009087, -0.00009869, -0.00001325, -0.00010233, -0.00002462, -0.00012203, -0.00002715, -0.00022617, 0.00003724, -0.00024015, -0.00022214, -0.00117302, 0.00058278, 0.00101920, 0.00036562, 0.00050013, 0.00092788, 0.00078334, -0.00135246, -0.00006174, -0.00017774, 0.00019150, -0.00003386, },
		{ -0.00005043, -0.00009506, 0.00019782, -0.00007266, 0.00012611, 0.00014308, -0.00001500, -0.00005882, -0.00026698, -0.00002025, -0.00009190, -0.00001958, -0.00009219, -0.00006005, -0.00007315, -0.00014809, -0.00006610, -0.00011715, -0.00021611, -0.00068836, -0.00091890, 0.00152971, 0.00063380, 0.00049940, 0.00114773, 0.00106279, -0.00150767, -0.00016565, -0.00014386, 0.00008884, -0.00002210, 0.00017130, },
		{ -0.00008951, 0.00008819, 0.00008813, -0.00012533, 0.00027689, -0.00008358, 0.00008830, -0.00024148, -0.00008593, -0.00008507, -0.00000883, -0.00010133, -0.00003049, -0.00012343, -0.00004004, -0.00019259, 0.00003801, -0.00039110, -0.00008192, -0.00145143, 0.00031065, 0.00181349, 0.00020610, 0.00153408, 0.00134070, -0.00155031, -0.00041139, -0.00007163, -0.00000418, 0.00000163, 0.00008092, -0.00000697, },
		{ -0.00002666, 0.00014847, -0.00012085, 0.00008215, 0.00013901, -0.00009125, 0.00004047, -0.00026202, 0.00000801, -0.00011030, 0.00001210, -0.00013043, -0.00002333, -0.00013627, -0.00003962, -0.00014075, -0.00017052, -0.00026463, -0.00052770, -0.00165032, 0.00210195, 0.00094928, 0.00128083, 0.00217188, -0.00187629, -0.00043606, -0.00025556, 0.00009248, -0.00009459, 0.00011339, -0.00010200, 0.00000963, },
		{ -0.00002344, 0.00010054, -0.00019225, 0.00027463, -0.00011556, 0.00005659, -0.00013937, -0.00010054, -0.00006943, -0.00000644, -0.00010813, -0.00002352, -0.00016123, 0.00004333, -0.00021180, -0.00007179, -0.00041201, 0.00005353, -0.00178390, -0.00011906, 0.00276074, 0.00085518, 0.00316293, -0.00214977, -0.00050449, -0.00042029, 0.00009325, -0.00008306, 0.00007345, -0.00011038, -0.00004530, 0.00006812, },
		{ 0.00013596, -0.00011689, 0.00000619, 0.00019104, -0.00020035, 0.00013116, -0.00024734, 0.00002897, -0.00013255, 0.00005160, -0.00018647, 0.00002093, -0.00016772, 0.00011080, -0.00044031, 0.00002480, -0.00044476, -0.00025204, -0.00267500, 0.00269136, 0.00217536, 0.00331242, -0.00168231, -0.00114005, -0.00012902, -0.00032139, 0.00022120, -0.00016337, 0.00005728, -0.00025394, 0.00023684, -0.00015188, },
		{ 0.00009099, -0.00020056, 0.00023313, -0.00010143, -0.00000970, -0.00003165, -0.00011782, -0.00006269, 0.00001481, -0.00014987, 0.00001322, -0.00020018, 0.00021628, -0.00050491, 0.00009557, -0.00055985, 0.00027817, -0.00230116, -0.00077952, 0.00443903, 0.00395732, -0.00139987, -0.00167522, 0.00007333, -0.00063933, 0.00035332, -0.00023411, 0.00010201, -0.00032700, 0.00021639, -0.00005878, -0.00003937, },
		{ -0.00000101, -0.00013475, 0.00027858, -0.00030195, 0.00019910, -0.00022618, 0.00003148, -0.00011922, 0.00003224, -0.00018078, 0.00001891, -0.00004721, -0.00008040, -0.00041099, 0.00006042, -0.00048104, -0.00024751, -0.00369308, 0.00296396, 0.00792310, -0.00248491, -0.00144853, -0.00032610, -0.00043680, 0.00002382, 0.00001520, -0.00006954, -0.00021668, 0.00005644, 0.00007183, -0.00011022, 0.00003235, },
		{ -0.00030580, 0.00024030, -0.00009187, -0.00006857, 0.00006757, -0.00017538, 0.00003461, -0.00010947, -0.00000931, -0.00015711, 0.00012161, -0.00020155, -0.00019222, -0.00022546, -0.00010279, -0.00042755, -0.00174584, -0.00377574, 0.01325590, -0.00210301, -0.00254381, 0.00012331, -0.00084210, 0.00019897, -0.00022625, 0.00017250, -0.00042631, 0.00018513, -0.00009436, 0.00010268, -0.00023929, 0.00025174, },
		{ -0.00008605, 0.00024934, -0.00032531, 0.00026620, -0.00032331, 0.00025585, -0.00043724, 0.00043958, -0.00069381, 0.00076773, -0.00098005, 0.00086602, -0.00151802, 0.00156036, -0.00258916, 0.00244052, -0.00938207, 0.01247165, 0.00466271, -0.00723791, 0.00260597, -0.00268074, 0.00150995, -0.00139097, 0.00109569, -0.00114140, 0.00072076, -0.00056734, 0.00047620, -0.00048752, 0.00029513, -0.00015223, },
		{ 0.00002078, 0.00002712, -0.00007637, -0.00005036, 0.00012603, -0.00038700, 0.00043610, -0.00069543, 0.00080515, -0.00105379, 0.00102259, -0.00162945, 0.00171034, -0.00270703, 0.00280133, -0.00723921, 0.00254883, 0.01593902, -0.01181906, 0.00399112, -0.00361785, 0.00210760, -0.00193352, 0.00143435, -0.00137191, 0.00088121, -0.00072346, 0.00056137, -0.00049816, 0.00022400, -0.00010269, -0.00004116, },
		{ 0.00036184, -0.00032081, 0.00015667, -0.00009978, -0.00001850, -0.00009049, 0.00000723, -0.00004593, 0.00000213, -0.00027671, 0.00006566, -0.00035711, 0.00004440, -0.00096805, -0.00008572, -0.00871703, 0.02173677, -0.00871703, -0.00008572, -0.00096805, 0.00004440, -0.00035711, 0.00006566, -0.00027671, 0.00000213, -0.00004593, 0.00000723, -0.00009049, -0.00001850, -0.00009978, 0.00015667, -0.00032081, },
		{ 0.00002078, -0.00004116, -0.00010269, 0.00022400, -0.00049816, 0.00056137, -0.00072346, 0.00088121, -0.00137191, 0.00143435, -0.00193352, 0.00210760, -0.00361785, 0.00399112, -0.01181906, 0.01593902, 0.00254883, -0.00723921, 0.00280133, -0.00270703, 0.00171034, -0.00162945, 0.00102259, -0.00105379, 0.00080515, -0.00069543, 0.00043610, -0.00038700, 0.00012603, -0.00005036, -0.00007637, 0.00002712, },
		{ -0.00008605, -0.00015223, 0.00029513, -0.00048752, 0.00047620, -0.00056734, 0.00072076, -0.00114140, 0.00109569, -0.00139097, 0.00150995, -0.00268074, 0.00260597, -0.00723791, 0.00466271, 0.01247165, -0.00938207, 0.00244052, -0.00258916, 0.00156036, -0.00151802, 0.00086602, -0.00098005, 0.00076773, -0.00069381, 0.00043958, -0.00043724, 0.00025585, -0.00032331, 0.00026620, -0.00032531, 0.00024934, },
		{ -0.00030580, 0.00025174, -0.00023929, 0.00010268, -0.00009436, 0.00018513, -0.00042631, 0.00017250, -0.00022625, 0.00019897, -0.00084210, 0.00012331, -0.00254381, -0.00210301, 0.01325590, -0.00377574, -0.00174584, -0.00042755, -0.00010279, -0.00022546, -0.00019222, -0.00020155, 0.00012161, -0.00015711, -0.00000931, -0.00010947, 0.00003461, -0.00017538, 0.00006757, -0.00006857, -0.00009187, 0.00024030, },
		{ -0.00000101, 0.00003235, -0.00011022, 0.00007183, 0.00005644, -0.00021668, -0.00006954, 0.00001520, 0.00002382, -0.00043680, -0.00032610, -0.00144853, -0.00248491, 0.00792310, 0.00296396, -0.00369308, -0.00024751, -0.00048104, 0.00006042, -0.00041099, -0.00008040, -0.00004721, 0.00001891, -0.00018078, 0.00003224, -0.00011922, 0.00003148, -0.00022618, 0.00019910, -0.00030195, 0.00027858, -0.00013475, },
		{ 0.00009099, -0.00003937, -0.00005878, 0.00021639, -0.00032700, 0.00010201, -0.00023411, 0.00035332, -0.00063933, 0.00007333, -0.00167522, -0.00139987, 0.00395732, 0.00443903, -0.00077952, -0.00230116, 0.00027817, -0.00055985, 0.00009557, -0.00050491, 0.00021628, -0.00020018, 0.00001322, -0.00014987, 0.00001481, -0.00006269, -0.00011782, -0.00003165, -0.00000970, -0.00010143, 0.00023313, -0.00020056, },
		{ 0.00013596, -0.00015188, 0.00023684, -0.00025394, 0.00005728, -0.00016337, 0.00022120, -0.00032139, -0.00012902, -0.00114005, -0.00168231, 0.00331242, 0.00217536, 0.00269136, -0.00267500, -0.00025204, -0.00044476, 0.00002480, -0.00044031, 0.00011080, -0.00016772, 0.00002093, -0.00018647, 0.00005160, -0.00013255, 0.00002897, -0.00024734, 0.00013116, -0.00020035, 0.00019104, 0.00000619, -0.00011689, },
		{ -0.00002344, 0.00006812, -0.00004530, -0.00011038, 0.00007345, -0.00008306, 0.00009325, -0.00042029, -0.00050449, -0.00214977, 0.00316293, 0.00085518, 0.00276074, -0.00011906, -0.00178390, 0.00005353, -0.00041201, -0.00007179, -0.00021180, 0.00004333, -0.00016123, -0.00002352, -0.00010813, -0.00000644, -0.00006943, -0.00010054, -0.00013937, 0.00005659, -0.00011556, 0.00027463, -0.00019225, 0.00010054, },
		{ -0.00002666, 0.00000963, -0.00010200, 0.00011339, -0.00009459, 0.00009248, -0.00025556, -0.00043606, -0.00187629, 0.00217188, 0.00128083, 0.00094928, 0.00210195, -0.00165032, -0.00052770, -0.00026463, -0.00017052, -0.00014075, -0.00003962, -0.00013627, -0.00002333, -0.00013043, 0.00001210, -0.00011030, 0.00000801, -0.00026202, 0.00004047, -0.00009125, 0.00013901, 0.00008215, -0.00012085, 0.00014847, },
		{ -0.00008951, -0.00000697, 0.00008092, 0.00000163, -0.00000418, -0.00007163, -0.00041139, -0.00155031, 0.00134070, 0.00153408, 0.00020610, 0.00181349, 0.00031065, -0.00145143, -0.00008192, -0.00039110, 0.00003801, -0.00019259, -0.00004004, -0.00012343, -0.00003049, -0.00010133, -0.00000883, -0.00008507, -0.00008593, -0.00024148, 0.00008830, -0.00008358, 0.00027689, -0.00012533, 0.00008813, 0.00008819, },
		{ -0.00005043, 0.00017130, -0.00002210, 0.00008884, -0.00014386, -0.00016565, -0.00150767, 0.00106279, 0.00114773, 0.00049940, 0.00063380, 0.00152971, -0.00091890, -0.00068836, -0.00021611, -0.00011715, -0.00006610, -0.00014809, -0.00007315, -0.00006005, -0.00009219, -0.00001958, -0.00009190, -0.00002025, -0.00026698, -0.00005882, -0.00001500, 0.00014308, 0.00012611, -0.00007266, 0.00019782, -0.00009506, },
		{ 0.00026423, -0.00003386, 0.00019150, -0.00017774, -0.00006174, -0.00135246, 0.00078334, 0.00092788, 0.00050013, 0.00036562, 0.00101920, 0.00058278, -0.00117302, -0.00022214, -0.00024015, 0.00003724, -0.00022617, -0.00002715, -0.00012203, -0.00002462, -0.00010233, -0.00001325, -0.00009869, -0.00009087, -0.00028670, 0.00000810, 0.00004746, 0.00023131, -0.00000655, 0.00010945, 0.00006031, -0.00017858, },
		{ 0.00007015, 0.00018659, -0.00003309, -0.00017781, -0.00106072, 0.00042531, 0.00095617, 0.00019305, 0.00058335, 0.00026093, 0.00117205, -0.00047585, -0.00073691, -0.00013408, -0.00007996, -0.00011844, -0.00011507, -0.00007747, -0.00006016, -0.00007604, -0.00004499, -0.00007595, -0.00007221, -0.00020857, -0.00024246, 0.00009016, 0.00015970, 0.00018081, 0.00002138, 0.00015279, -0.00017194, 0.00019983, },
		{ 0.00021090, 0.00011683, -0.00024049, -0.00094079, 0.00027128, 0.00085886, 0.00011334, 0.00048220, 0.00020466, 0.00065233, 0.00059145, -0.00087423, -0.00029021, -0.00010596, -0.00008788, -0.00013301, -0.00007049, -0.00008985, -0.00003650, -0.00009608, -0.00002664, -0.00012570, -0.00005761, -0.00036397, -0.00008490, 0.00017702, 0.00022887, 0.00011623, 0.00010672, -0.00004791, 0.00009346, 0.00013862, },
		{ 0.00023064, -0.00019516, -0.00096824, 0.00021647, 0.00071949, 0.00015565, 0.00026598, 0.00031920, 0.00019810, 0.00081339, -0.00018196, -0.00067320, -0.00006548, -0.00016017, -0.00006254, -0.00012491, -0.00006213, -0.00006567, -0.00006279, -0.00006923, -0.00008492, -0.00007779, -0.00019709, -0.00038025, 0.00010446, 0.00026580, 0.00021971, 0.00012319, 0.00000156, 0.00011567, 0.00008038, 0.00029240, },
		{ -0.00016830, -0.00094525, 0.00005283, 0.00069722, 0.00012146, 0.00020811, 0.00022814, 0.00016420, 0.00044457, 0.00047609, -0.00057516, -0.00028524, -0.00013050, -0.00008574, -0.00009834, -0.00008552, -0.00006535, -0.00005107, -0.00008486, -0.00007197, -0.00009789, -0.00009069, -0.00036071, -0.00026500, 0.00031322, 0.00029196, 0.00021192, 0.00001886, 0.00017471, 0.00005493, 0.00027275, 0.00042117, },
		},
		{ { -0.00005752, -0.00004525, -0.00013831, -0.00002674, -0.00004455, -0.00006678, -0.00017194, -0.00013111, -0.00046740, -0.00002930, 0.00066248, 0.00015030, 0.00008771, 0.00002954, -0.00026591, 0.00029199, 0.00107864, 0.00029199, -0.00026591, 0.00002954, 0.00008771, 0.00015030, 0.00066248, -0.00002930, -0.00046740, -0.00013111, -0.00017194, -0.00006678, -0.00004455, -0.00002674, -0.00013831, -0.00004525, },
		{ -0.00005577, -0.00009974, -0.00011383, 0.00001647, -0.00008593, -0.00006745, -0.00016678, -0.00022756, -0.00032421, 0.00026895, 0.00044544, 0.00009161, 0.00003026, 0.00004513, -0.00039619, 0.00021174, 0.00109652, 0.00035865, -0.00013915, 0.00003824, 0.00013298, 0.00032959, 0.00071853, -0.00047325, -0.00036909, -0.00012423, -0.00015486, -0.00005351, -0.00001664, -0.00010530, -0.00009620, -0.00002391, },
		{ -0.00005746, -0.00015367, -0.00004083, 0.00000158, -0.00011505, -0.00003553, -0.00022600, -0.00024739, -0.00008258, 0.00035923, 0.00022755, 0.00010003, -0.00004213, 0.00008164, -0.00052911, 0.00011661, 0.00114351, 0.00042650, -0.00002888, 0.00008457, 0.00018529, 0.00064231, 0.00038264, -0.00074201, -0.00016885, -0.00021852, -0.00006768, -0.00006494, -0.00001103, -0.00017529, -0.00001404, -0.00003997, },
		{ -0.00008248, -0.00016280, 0.00003082, -0.00006753, -0.00007882, -0.00005016, -0.00028565, -0.00015364, 0.00010349, 0.00031854, 0.00007045, 0.00012418, -0.00011179, 0.00011950, -0.00064520, -0.00001711, 0.00124322, 0.00046960, 0.00010510, 0.00012049, 0.00037857, 0.00085266, -0.00030625, -0.00057249, -0.00015878, -0.00022984, -0.00000932, -0.00006179, -0.00006135, -0.00017026, 0.00003774, -0.00005858, },
		{ -0.00013930, -0.00009791, 0.00003070, -0.00011822, -0.00000375, -0.00015412, -0.00022397, -0.00007035, 0.00023260, 0.00018934, 0.00001820, 0.00009514, -0.00012351, 0.00010275, -0.00069213, -0.00023963, 0.00144132, 0.00044034, 0.00032745, 0.00009509, 0.00085992, 0.00051288, -0.00084011, -0.00017410, -0.00035365, -0.00008783, -0.00004587, -0.00002240, -0.00015812, -0.00007718, 0.00001549, -0.00004854, },
		{ -0.00018367, 0.00000511, -0.00006426, -0.00005805, -0.00002190, -0.00021878, -0.00009306, -0.00004009, 0.00030901, 0.00003698, 0.00003750, 0.00002065, -0.00008900, 0.00004256, -0.00068956, -0.00051539, 0.00167610, 0.00042288, 0.00054896, 0.00019684, 0.00138405, -0.00048023, -0.00071227, -0.00004424, -0.00043615, 0.00001769, -0.00007167, -0.00003286, -0.00019482, 0.00000805, -0.00003068, -0.00003916, },
		{ -0.00014281, 0.00003987, -0.00015588, 0.00007381, -0.00018051, -0.00010349, -0.00006385, 0.00003523, 0.00025830, -0.00001436, 0.00000757, 0.00000354, -0.00010935, 0.00003345, -0.00073205, -0.00073797, 0.00180694, 0.00059618, 0.00058885, 0.00077915, 0.00123245, -0.00145444, -0.00011197, -0.00029908, -0.00023712, -0.00006363, 0.00000698, -0.00014980, -0.00009599, -0.00001284, -0.00002210, -0.00008455, },
		{ -0.00002324, -0.00005721, -0.00010278, 0.00009105, -0.00029967, 0.00008938, -0.00013506, 0.00015977, 0.00010008, 0.00004806, -0.00009412, 0.00005568, -0.00019142, 0.00006654, -0.00078791, -0.00096284, 0.00189076, 0.00090601, 0.00061216, 0.00171501, -0.00001921, -0.00151290, 0.00019399, -0.00047895, -0.00002808, -0.00017376, 0.00006751, -0.00024373, 0.00003396, -0.00009049, 0.00001840, -0.00015648, },
		{ 0.00004603, -0.00018525, 0.00007565, -0.00009856, -0.00017592, 0.00010871, -0.00010462, 0.00016232, 0.00001874, 0.00006044, -0.00011852, 0.00002438, -0.00018187, -0.00002147, -0.00066578, -0.00143361, 0.00221218, 0.00103567, 0.00121785, 0.00201782, -0.00165072, -0.00064566, -0.00016662, -0.00020788, -0.00011940, -0.00009198, -0.00005284, -0.00014862, 0.00002359, -0.00008782, -0.00001249, -0.00014322, },
		{ -0.00004345, -0.00015935, 0.00015783, -0.00030001, 0.00009433, -0.00007040, 0.00009441, -0.00002582, 0.00010462, -0.00006064, 0.00000143, -0.00015130, -0.00003500, -0.00025320, -0.00036079, -0.00216788, 0.00278418, 0.00106051, 0.00256808, 0.00071989, -0.00223998, 0.00010026, -0.00064049, 0.00021878, -0.00035372, 0.00006813, -0.00024014, 0.00005488, -0.00010478, 0.00000008, -0.00011949, -0.00001044, },
		{ -0.00020292, 0.00003983, -0.00001215, -0.00023528, 0.00018439, -0.00016956, 0.00023490, -0.00019654, 0.00017801, -0.00013947, 0.00007436, -0.00026881, 0.00002441, -0.00035439, -0.00022369, -0.00275059, 0.00307771, 0.00190739, 0.00346513, -0.00165273, -0.00128607, -0.00011980, -0.00051231, 0.00024852, -0.00032910, -0.00000069, -0.00021063, 0.00010425, -0.00014462, 0.00001587, -0.00015390, 0.00009904, },
		{ -0.00021350, 0.00018028, -0.00026399, 0.00005662, -0.00001082, -0.00001729, 0.00013612, -0.00016690, 0.00007876, -0.00003240, -0.00004381, -0.00019268, -0.00012367, -0.00020253, -0.00041851, -0.00298947, 0.00283994, 0.00441185, 0.00205008, -0.00303430, -0.00003245, -0.00080363, 0.00002715, -0.00006231, -0.00007145, -0.00026444, 0.00003144, -0.00004677, -0.00003093, -0.00010225, -0.00002587, 0.00002827, },
		{ -0.00001318, 0.00004425, -0.00025786, 0.00020095, -0.00017846, 0.00014695, 0.00000010, -0.00011974, 0.00000843, 0.00002433, -0.00008929, -0.00023043, -0.00009124, -0.00029472, -0.00033913, -0.00368551, 0.00322792, 0.00758631, -0.00198631, -0.00186626, -0.00014395, -0.00071058, 0.00007321, -0.00005450, -0.00008880, -0.00028135, 0.00010422, -0.00010187, 0.00001172, -0.00016263, 0.00010787, -0.00014993, },
		{ 0.00017468, -0.00025881, 0.00007144, -0.00003675, 0.00000001, -0.00000740, 0.00015010, -0.00036072, 0.00027943, -0.00031175, 0.00030541, -0.00077913, 0.00058944, -0.00123567, 0.00080408, -0.00599705, 0.00630948, 0.00848701, -0.00622185, 0.00093260, -0.00165365, 0.00044289, -0.00070058, 0.00055760, -0.00060833, 0.00016601, -0.00022221, 0.00015494, -0.00020457, 0.00002745, 0.00001637, -0.00017992, },
		{ 0.00009820, -0.00037756, 0.00037301, -0.00039739, 0.00036497, -0.00037212, 0.00048957, -0.00076374, 0.00072860, -0.00085264, 0.00091314, -0.00154087, 0.00153893, -0.00259457, 0.00253341, -0.00953224, 0.01263220, 0.00475660, -0.00733532, 0.00265226, -0.00279748, 0.00146210, -0.00144962, 0.00113590, -0.00109541, 0.00063225, -0.00060759, 0.00046932, -0.00047475, 0.00031780, -0.00025273, 0.00003633, },
		{ -0.00018247, -0.00014760, 0.00028743, -0.00042228, 0.00043767, -0.00046740, 0.00052934, -0.00079005, 0.00073700, -0.00089721, 0.00093389, -0.00155588, 0.00154976, -0.00284193, 0.00283415, -0.01157736, 0.01948072, -0.00253658, -0.00438637, 0.00162920, -0.00207642, 0.00110595, -0.00117319, 0.00087622, -0.00085863, 0.00049913, -0.00051856, 0.00039440, -0.00040074, 0.00032954, -0.00036003, 0.00025885, },
		{ -0.00033941, 0.00018504, -0.00009953, -0.00004170, 0.00005309, -0.00008393, 0.00005083, -0.00020324, 0.00002158, -0.00011253, -0.00000767, -0.00037233, -0.00000993, -0.00103375, 0.00014269, -0.00921348, 0.02247966, -0.00921348, 0.00014269, -0.00103375, -0.00000993, -0.00037233, -0.00000767, -0.00011253, 0.00002158, -0.00020324, 0.00005083, -0.00008393, 0.00005309, -0.00004170, -0.00009953, 0.00018504, },
		{ -0.00018247, 0.00025885, -0.00036003, 0.00032954, -0.00040074, 0.00039440, -0.00051856, 0.00049913, -0.00085863, 0.00087622, -0.00117319, 0.00110595, -0.00207642, 0.00162920, -0.00438637, -0.00253658, 0.01948072, -0.01157736, 0.00283415, -0.00284193, 0.00154976, -0.00155588, 0.00093389, -0.00089721, 0.00073700, -0.00079005, 0.00052934, -0.00046740, 0.00043767, -0.00042228, 0.00028743, -0.00014760, },
		{ 0.00009820, 0.00003633, -0.00025273, 0.00031780, -0.00047475, 0.00046932, -0.00060759, 0.00063225, -0.00109541, 0.00113590, -0.00144962, 0.00146210, -0.00279748, 0.00265226, -0.00733532, 0.00475660, 0.01263220, -0.00953224, 0.00253341, -0.00259457, 0.00153893, -0.00154087, 0.00091314, -0.00085264, 0.00072860, -0.00076374, 0.00048957, -0.00037212, 0.00036497, -0.00039739, 0.00037301, -0.00037756, },
		{ 0.00017468, -0.00017992, 0.00001637, 0.00002745, -0.00020457, 0.00015494, -0.00022221, 0.00016601, -0.00060833, 0.00055760, -0.00070058, 0.00044289, -0.00165365, 0.00093260, -0.00622185, 0.00848701, 0.00630948, -0.00599705, 0.00080408, -0.00123567, 0.00058944, -0.00077913, 0.00030541, -0.00031175, 0.00027943, -0.00036072, 0.00015010, -0.00000740, 0.00000001, -0.00003675, 0.00007144, -0.00025881, },
		{ -0.00001318, -0.00014993, 0.00010787, -0.00016263, 0.00001172, -0.00010187, 0.00010422, -0.00028135, -0.00008880, -0.00005450, 0.00007321, -0.00071058, -0.00014395, -0.00186626, -0.00198631, 0.00758631, 0.00322792, -0.00368551, -0.00033913, -0.00029472, -0.00009124, -0.00023043, -0.00008929, 0.00002433, 0.00000843, -0.00011974, 0.00000010, 0.00014695, -0.00017846, 0.00020095, -0.00025786, 0.00004425, },
		{ -0.00021350, 0.00002827, -0.00002587, -0.00010225, -0.00003093, -0.00004677, 0.00003144, -0.00026444, -0.00007145, -0.00006231, 0.00002715, -0.00080363, -0.00003245, -0.00303430, 0.00205008, 0.00441185, 0.00283994, -0.00298947, -0.00041851, -0.00020253, -0.00012367, -0.00019268, -0.00004381, -0.00003240, 0.00007876, -0.00016690, 0.00013612, -0.00001729, -0.00001082, 0.00005662, -0.00026399, 0.00018028, },
		{ -0.00020292, 0.00009904, -0.00015390, 0.00001587, -0.00014462, 0.00010425, -0.00021063, -0.00000069, -0.00032910, 0.00024852, -0.00051231, -0.00011980, -0.00128607, -0.00165273, 0.00346513, 0.00190739, 0.00307771, -0.00275059, -0.00022369, -0.00035439, 0.00002441, -0.00026881, 0.00007436, -0.00013947, 0.00017801, -0.00019654, 0.00023490, -0.00016956, 0.00018439, -0.00023528, -0.00001215, 0.00003983, },
		{ -0.00004345, -0.00001044, -0.00011949, 0.00000008, -0.00010478, 0.00005488, -0.00024014, 0.00006813, -0.00035372, 0.00021878, -0.00064049, 0.00010026, -0.00223998, 0.00071989, 0.00256808, 0.00106051, 0.00278418, -0.00216788, -0.00036079, -0.00025320, -0.00003500, -0.00015130, 0.00000143, -0.00006064, 0.00010462, -0.00002582, 0.00009441, -0.00007040, 0.00009433, -0.00030001, 0.00015783, -0.00015935, },
		{ 0.00004603, -0.00014322, -0.00001249, -0.00008782, 0.00002359, -0.00014862, -0.00005284, -0.00009198, -0.00011940, -0.00020788, -0.00016662, -0.00064566, -0.00165072, 0.00201782, 0.00121785, 0.00103567, 0.00221218, -0.00143361, -0.00066578, -0.00002147, -0.00018187, 0.00002438, -0.00011852, 0.00006044, 0.00001874, 0.00016232, -0.00010462, 0.00010871, -0.00017592, -0.00009856, 0.00007565, -0.00018525, },
		{ -0.00002324, -0.00015648, 0.00001840, -0.00009049, 0.00003396, -0.00024373, 0.00006751, -0.00017376, -0.00002808, -0.00047895, 0.00019399, -0.00151290, -0.00001921, 0.00171501, 0.00061216, 0.00090601, 0.00189076, -0.00096284, -0.00078791, 0.00006654, -0.00019142, 0.00005568, -0.00009412, 0.00004806, 0.00010008, 0.00015977, -0.00013506, 0.00008938, -0.00029967, 0.00009105, -0.00010278, -0.00005721, },
		{ -0.00014281, -0.00008455, -0.00002210, -0.00001284, -0.00009599, -0.00014980, 0.00000698, -0.00006363, -0.00023712, -0.00029908, -0.00011197, -0.00145444, 0.00123245, 0.00077915, 0.00058885, 0.00059618, 0.00180694, -0.00073797, -0.00073205, 0.00003345, -0.00010935, 0.00000354, 0.00000757, -0.00001436, 0.00025830, 0.00003523, -0.00006385, -0.00010349, -0.00018051, 0.00007381, -0.00015588, 0.00003987, },
		{ -0.00018367, -0.00003916, -0.00003068, 0.00000805, -0.00019482, -0.00003286, -0.00007167, 0.00001769, -0.00043615, -0.00004424, -0.00071227, -0.00048023, 0.00138405, 0.00019684, 0.00054896, 0.00042288, 0.00167610, -0.00051539, -0.00068956, 0.00004256, -0.00008900, 0.00002065, 0.00003750, 0.00003698, 0.00030901, -0.00004009, -0.00009306, -0.00021878, -0.00002190, -0.00005805, -0.00006426, 0.00000511, },
		{ -0.00013930, -0.00004854, 0.00001549, -0.00007718, -0.00015812, -0.00002240, -0.00004587, -0.00008783, -0.00035365, -0.00017410, -0.00084011, 0.00051288, 0.00085992, 0.00009509, 0.00032745, 0.00044034, 0.00144132, -0.00023963, -0.00069213, 0.00010275, -0.00012351, 0.00009514, 0.00001820, 0.00018934, 0.00023260, -0.00007035, -0.00022397, -0.00015412, -0.00000375, -0.00011822, 0.00003070, -0.00009791, },
		{ -0.00008248, -0.00005858, 0.00003774, -0.00017026, -0.00006135, -0.00006179, -0.00000932, -0.00022984, -0.00015878, -0.00057249, -0.00030625, 0.00085266, 0.00037857, 0.00012049, 0.00010510, 0.00046960, 0.00124322, -0.00001711, -0.00064520, 0.00011950, -0.00011179, 0.00012418, 0.00007045, 0.00031854, 0.00010349, -0.00015364, -0.00028565, -0.00005016, -0.00007882, -0.00006753, 0.00003082, -0.00016280, },
		{ -0.00005746, -0.00003997, -0.00001404, -0.00017529, -0.00001103, -0.00006494, -0.00006768, -0.00021852, -0.00016885, -0.00074201, 0.00038264, 0.00064231, 0.00018529, 0.00008457, -0.00002888, 0.00042650, 0.00114351, 0.00011661, -0.00052911, 0.00008164, -0.00004213, 0.00010003, 0.00022755, 0.00035923, -0.00008258, -0.00024739, -0.00022600, -0.00003553, -0.00011505, 0.00000158, -0.00004083, -0.00015367, },
		{ -0.00005577, -0.00002391, -0.00009620, -0.00010530, -0.00001664, -0.00005351, -0.00015486, -0.00012423, -0.00036909, -0.00047325, 0.00071853, 0.00032959, 0.00013298, 0.00003824, -0.00013915, 0.00035865, 0.00109652, 0.00021174, -0.00039619, 0.00004513, 0.00003026, 0.00009161, 0.00044544, 0.00026895, -0.00032421, -0.00022756, -0.00016678, -0.00006745, -0.00008593, 0.00001647, -0.00011383, -0.00009974, },
		},
		{ { -0.00004892, -0.00004840, -0.00012432, -0.00006594, -0.00011167, -0.00015953, -0.00051244, -0.00000445, 0.00051056, 0.00025118, 0.00010822, 0.00016233, 0.00015483, 0.00014945, 0.00061486, -0.00011201, -0.00088588, -0.00011201, 0.00061486, 0.00014945, 0.00015483, 0.00016233, 0.00010822, 0.00025118, 0.00051056, -0.00000445, -0.00051244, -0.00015953, -0.00011167, -0.00006594, -0.00012432, -0.00004840, },
		{ -0.00003034, -0.00006074, -0.00009949, -0.00010222, -0.00007717, -0.00013488, -0.00030014, -0.00057174, 0.00049554, 0.00044802, 0.00014937, 0.00022386, 0.00021680, 0.00011773, 0.00068341, 0.00005761, -0.00092435, -0.00016352, 0.00040735, 0.00026901, 0.00006361, 0.00017043, 0.00000402, 0.00021537, 0.00031141, 0.00031664, -0.00027990, -0.00036509, -0.00008212, -0.00010177, -0.00008594, -0.00008024, },
		{ -0.00004205, -0.00006105, -0.00007610, -0.00012879, -0.00005397, -0.00016455, -0.00008037, -0.00066978, -0.00016251, 0.00081684, 0.00018326, 0.00031491, 0.00027467, 0.00015191, 0.00070567, 0.00022126, -0.00094734, -0.00019038, 0.00021682, 0.00028866, 0.00008906, 0.00011139, -0.00001327, 0.00012664, 0.00023915, 0.00026922, 0.00008956, -0.00038462, -0.00018852, -0.00008166, -0.00009889, -0.00006462, },
		{ -0.00007535, -0.00003188, -0.00010383, -0.00007437, -0.00012444, -0.00009226, -0.00012086, -0.00028679, -0.00085478, 0.00059490, 0.00063750, 0.00020038, 0.00049088, 0.00010960, 0.00084504, 0.00027607, -0.00091989, -0.00023571, 0.00010302, 0.00020717, 0.00014731, 0.00008918, -0.00006274, 0.00011017, 0.00013568, 0.00023228, 0.00016212, -0.00008928, -0.00035540, -0.00006149, -0.00013967, -0.00002202, },
		{ -0.00002426, -0.00007142, -0.00007414, -0.00008135, -0.00010650, -0.00012281, -0.00009486, -0.00013066, -0.00071746, -0.00047240, 0.00115722, 0.00025664, 0.00059204, 0.00018931, 0.00094235, 0.00043010, -0.00103982, -0.00017303, -0.00004691, 0.00018286, 0.00007883, 0.00019555, -0.00018678, 0.00015624, 0.00004082, 0.00018423, 0.00014480, 0.00008578, -0.00023389, -0.00021244, -0.00008618, -0.00007133, },
		{ 0.00000749, -0.00009771, -0.00003859, -0.00012591, -0.00001858, -0.00023055, 0.00002234, -0.00023673, -0.00020269, -0.00116957, 0.00056795, 0.00101492, 0.00037431, 0.00049640, 0.00091406, 0.00078813, -0.00133156, -0.00005696, -0.00019155, 0.00018777, -0.00002518, 0.00025995, -0.00019341, 0.00006377, 0.00012890, -0.00000313, 0.00021642, 0.00004308, 0.00001667, -0.00029057, -0.00010484, -0.00009407, },
		{ -0.00007116, -0.00001496, -0.00010617, -0.00006392, -0.00006458, -0.00015247, -0.00008100, -0.00011373, -0.00019667, -0.00068490, -0.00093373, 0.00152543, 0.00064248, 0.00049566, 0.00113391, 0.00106757, -0.00148677, -0.00016086, -0.00015768, 0.00008511, -0.00001342, 0.00016702, -0.00006527, -0.00009161, 0.00021727, -0.00006924, 0.00011121, 0.00013871, -0.00000643, -0.00006270, -0.00028095, -0.00001563, },
		{ -0.00006434, -0.00000421, -0.00011530, -0.00003436, -0.00011486, -0.00004441, -0.00020749, 0.00004143, -0.00037166, -0.00007847, -0.00146626, 0.00030636, 0.00182218, 0.00020237, 0.00152026, 0.00134548, -0.00152940, -0.00040661, -0.00008545, -0.00000792, 0.00001032, 0.00007664, -0.00002180, -0.00008606, 0.00010764, 0.00009155, -0.00014023, 0.00027252, -0.00007501, 0.00008443, -0.00025546, -0.00008131, },
		{ 0.00002874, -0.00010568, -0.00000187, -0.00013430, -0.00001476, -0.00014065, -0.00005452, -0.00013733, -0.00015107, -0.00026118, -0.00054253, -0.00165460, 0.00211064, 0.00094555, 0.00126701, 0.00217666, -0.00185538, -0.00043127, -0.00026938, 0.00008874, -0.00008590, 0.00010910, -0.00011683, 0.00001309, -0.00000721, 0.00015189, -0.00013574, 0.00007777, 0.00014758, -0.00009512, 0.00002650, -0.00025740, },
		{ -0.00007981, -0.00006481, -0.00002042, -0.00011200, -0.00001495, -0.00016561, 0.00002843, -0.00020838, -0.00005234, -0.00040856, 0.00003869, -0.00178819, -0.00011037, 0.00275701, 0.00084136, 0.00316771, -0.00212886, -0.00049970, -0.00043410, 0.00008952, -0.00007438, 0.00006917, -0.00012521, -0.00004185, 0.00008757, -0.00002002, 0.00008565, -0.00019663, 0.00028319, -0.00011944, 0.00004262, -0.00013475, },
		{ -0.00022660, 0.00003358, -0.00014653, 0.00004772, -0.00017790, 0.00001656, -0.00018261, 0.00011422, -0.00042086, 0.00002825, -0.00045959, -0.00025632, -0.00266631, 0.00268763, 0.00216154, 0.00331720, -0.00166141, -0.00113527, -0.00014284, -0.00032512, 0.00022988, -0.00016765, 0.00004245, -0.00025048, 0.00025629, -0.00014846, 0.00012106, -0.00012127, 0.00001476, 0.00018717, -0.00021433, 0.00013578, },
		{ -0.00001092, -0.00011320, -0.00007667, 0.00001094, -0.00014130, 0.00000884, -0.00021508, 0.00021970, -0.00048546, 0.00009902, -0.00057469, 0.00027389, -0.00229247, -0.00078326, 0.00442521, 0.00396211, -0.00137897, -0.00167044, 0.00005951, -0.00064307, 0.00036200, -0.00023840, 0.00008717, -0.00032355, 0.00023584, -0.00005536, -0.00005427, 0.00008661, -0.00019199, 0.00022926, -0.00011540, -0.00000509, },
		{ 0.00021984, -0.00022157, 0.00001751, -0.00012310, 0.00004081, -0.00018516, 0.00000401, -0.00004379, -0.00006095, -0.00040754, 0.00004559, -0.00048532, -0.00023882, -0.00369682, 0.00295014, 0.00792789, -0.00246400, -0.00144375, -0.00033992, -0.00044054, 0.00003251, 0.00001092, -0.00008437, -0.00021323, 0.00007589, 0.00007525, -0.00012511, 0.00002797, 0.00000756, -0.00013862, 0.00026460, -0.00029733, },
		{ -0.00004783, 0.00007219, -0.00018935, 0.00003073, -0.00010090, -0.00001369, -0.00017200, 0.00012503, -0.00018210, -0.00018877, -0.00024029, -0.00010708, -0.00041887, -0.00174958, -0.00378956, 0.01326068, -0.00208211, -0.00253903, 0.00010949, -0.00084583, 0.00020766, -0.00023053, 0.00015767, -0.00042286, 0.00020458, -0.00009094, 0.00008778, -0.00024367, 0.00026031, -0.00030967, 0.00022633, -0.00008725, },
		{ -0.00030457, 0.00027081, -0.00033728, 0.00025197, -0.00042867, 0.00043521, -0.00070870, 0.00077115, -0.00096060, 0.00086947, -0.00153286, 0.00155608, -0.00258047, 0.00243678, -0.00939589, 0.01247643, 0.00468361, -0.00723313, 0.00259215, -0.00268448, 0.00151864, -0.00139525, 0.00108086, -0.00113795, 0.00074020, -0.00056392, 0.00046130, -0.00049190, 0.00030370, -0.00015610, -0.00010002, 0.00025396, },
		{ 0.00004786, -0.00007175, -0.00006433, 0.00012216, -0.00037844, 0.00043173, -0.00071033, 0.00080857, -0.00103434, 0.00102604, -0.00164428, 0.00170606, -0.00269834, 0.00279759, -0.00725302, 0.00255362, 0.01595993, -0.01181427, 0.00397730, -0.00362158, 0.00211629, -0.00193780, 0.00141951, -0.00136846, 0.00090066, -0.00072004, 0.00054647, -0.00050254, 0.00023257, -0.00010656, -0.00005513, 0.00002540, },
		{ 0.00038257, -0.00031619, 0.00014269, -0.00010366, -0.00000993, -0.00009487, -0.00000767, -0.00004251, 0.00002158, -0.00027326, 0.00005083, -0.00036139, 0.00005309, -0.00097179, -0.00009953, -0.00871225, 0.02175767, -0.00871225, -0.00009953, -0.00097179, 0.00005309, -0.00036139, 0.00005083, -0.00027326, 0.00002158, -0.00004251, -0.00000767, -0.00009487, -0.00000993, -0.00010366, 0.00014269, -0.00031619, },
		{ 0.00004786, 0.00002540, -0.00005513, -0.00010656, 0.00023257, -0.00050254, 0.00054647, -0.00072004, 0.00090066, -0.00136846, 0.00141951, -0.00193780, 0.00211629, -0.00362158, 0.00397730, -0.01181427, 0.01595993, 0.00255362, -0.00725302, 0.00279759, -0.00269834, 0.00170606, -0.00164428, 0.00102604, -0.00103434, 0.00080857, -0.00071033, 0.00043173, -0.00037844, 0.00012216, -0.00006433, -0.00007175, },
		{ -0.00030457, 0.00025396, -0.00010002, -0.00015610, 0.00030370, -0.00049190, 0.00046130, -0.00056392, 0.00074020, -0.00113795, 0.00108086, -0.00139525, 0.00151864, -0.00268448, 0.00259215, -0.00723313, 0.00468361, 0.01247643, -0.00939589, 0.00243678, -0.00258047, 0.00155608, -0.00153286, 0.00086947, -0.00096060, 0.00077115, -0.00070870, 0.00043521, -0.00042867, 0.00025197, -0.00033728, 0.00027081, },
		{ -0.00004783, -0.00008725, 0.00022633, -0.00030967, 0.00026031, -0.00024367, 0.00008778, -0.00009094, 0.00020458, -0.00042286, 0.00015767, -0.00023053, 0.00020766, -0.00084583, 0.00010949, -0.00253903, -0.00208211, 0.01326068, -0.00378956, -0.00174958, -0.00041887, -0.00010708, -0.00024029, -0.00018877, -0.00018210, 0.00012503, -0.00017200, -0.00001369, -0.00010090, 0.00003073, -0.00018935, 0.00007219, },
		{ 0.00021984, -0.00029733, 0.00026460, -0.00013862, 0.00000756, 0.00002797, -0.00012511, 0.00007525, 0.00007589, -0.00021323, -0.00008437, 0.00001092, 0.00003251, -0.00044054, -0.00033992, -0.00144375, -0.00246400, 0.00792789, 0.00295014, -0.00369682, -0.00023882, -0.00048532, 0.00004559, -0.00040754, -0.00006095, -0.00004379, 0.00000401, -0.00018516, 0.00004081, -0.00012310, 0.00001751, -0.00022157, },
		{ -0.00001092, -0.00000509, -0.00011540, 0.00022926, -0.00019199, 0.00008661, -0.00005427, -0.00005536, 0.00023584, -0.00032355, 0.00008717, -0.00023840, 0.00036200, -0.00064307, 0.00005951, -0.00167044, -0.00137897, 0.00396211, 0.00442521, -0.00078326, -0.00229247, 0.00027389, -0.00057469, 0.00009902, -0.00048546, 0.00021970, -0.00021508, 0.00000884, -0.00014130, 0.00001094, -0.00007667, -0.00011320, },
		{ -0.00022660, 0.00013578, -0.00021433, 0.00018717, 0.00001476, -0.00012127, 0.00012106, -0.00014846, 0.00025629, -0.00025048, 0.00004245, -0.00016765, 0.00022988, -0.00032512, -0.00014284, -0.00113527, -0.00166141, 0.00331720, 0.00216154, 0.00268763, -0.00266631, -0.00025632, -0.00045959, 0.00002825, -0.00042086, 0.00011422, -0.00018261, 0.00001656, -0.00017790, 0.00004772, -0.00014653, 0.00003358, },
		{ -0.00007981, -0.00013475, 0.00004262, -0.00011944, 0.00028319, -0.00019663, 0.00008565, -0.00002002, 0.00008757, -0.00004185, -0.00012521, 0.00006917, -0.00007438, 0.00008952, -0.00043410, -0.00049970, -0.00212886, 0.00316771, 0.00084136, 0.00275701, -0.00011037, -0.00178819, 0.00003869, -0.00040856, -0.00005234, -0.00020838, 0.00002843, -0.00016561, -0.00001495, -0.00011200, -0.00002042, -0.00006481, },
		{ 0.00002874, -0.00025740, 0.00002650, -0.00009512, 0.00014758, 0.00007777, -0.00013574, 0.00015189, -0.00000721, 0.00001309, -0.00011683, 0.00010910, -0.00008590, 0.00008874, -0.00026938, -0.00043127, -0.00185538, 0.00217666, 0.00126701, 0.00094555, 0.00211064, -0.00165460, -0.00054253, -0.00026118, -0.00015107, -0.00013733, -0.00005452, -0.00014065, -0.00001476, -0.00013430, -0.00000187, -0.00010568, },
		{ -0.00006434, -0.00008131, -0.00025546, 0.00008443, -0.00007501, 0.00027252, -0.00014023, 0.00009155, 0.00010764, -0.00008606, -0.00002180, 0.00007664, 0.00001032, -0.00000792, -0.00008545, -0.00040661, -0.00152940, 0.00134548, 0.00152026, 0.00020237, 0.00182218, 0.00030636, -0.00146626, -0.00007847, -0.00037166, 0.00004143, -0.00020749, -0.00004441, -0.00011486, -0.00003436, -0.00011530, -0.00000421, },
		{ -0.00007116, -0.00001563, -0.00028095, -0.00006270, -0.00000643, 0.00013871, 0.00011121, -0.00006924, 0.00021727, -0.00009161, -0.00006527, 0.00016702, -0.00001342, 0.00008511, -0.00015768, -0.00016086, -0.00148677, 0.00106757, 0.00113391, 0.00049566, 0.00064248, 0.00152543, -0.00093373, -0.00068490, -0.00019667, -0.00011373, -0.00008100, -0.00015247, -0.00006458, -0.00006392, -0.00010617, -0.00001496, },
		{ 0.00000749, -0.00009407, -0.00010484, -0.00029057, 0.00001667, 0.00004308, 0.00021642, -0.00000313, 0.00012890, 0.00006377, -0.00019341, 0.00025995, -0.00002518, 0.00018777, -0.00019155, -0.00005696, -0.00133156, 0.00078813, 0.00091406, 0.00049640, 0.00037431, 0.00101492, 0.00056795, -0.00116957, -0.00020269, -0.00023673, 0.00002234, -0.00023055, -0.00001858, -0.00012591, -0.00003859, -0.00009771, },
		{ -0.00002426, -0.00007133, -0.00008618, -0.00021244, -0.00023389, 0.00008578, 0.00014480, 0.00018423, 0.00004082, 0.00015624, -0.00018678, 0.00019555, 0.00007883, 0.00018286, -0.00004691, -0.00017303, -0.00103982, 0.00043010, 0.00094235, 0.00018931, 0.00059204, 0.00025664, 0.00115722, -0.00047240, -0.00071746, -0.00013066, -0.00009486, -0.00012281, -0.00010650, -0.00008135, -0.00007414, -0.00007142, },
		{ -0.00007535, -0.00002202, -0.00013967, -0.00006149, -0.00035540, -0.00008928, 0.00016212, 0.00023228, 0.00013568, 0.00011017, -0.00006274, 0.00008918, 0.00014731, 0.00020717, 0.00010302, -0.00023571, -0.00091989, 0.00027607, 0.00084504, 0.00010960, 0.00049088, 0.00020038, 0.00063750, 0.00059490, -0.00085478, -0.00028679, -0.00012086, -0.00009226, -0.00012444, -0.00007437, -0.00010383, -0.00003188, },
		{ -0.00004205, -0.00006462, -0.00009889, -0.00008166, -0.00018852, -0.00038462, 0.00008956, 0.00026922, 0.00023915, 0.00012664, -0.00001327, 0.00011139, 0.00008906, 0.00028866, 0.00021682, -0.00019038, -0.00094734, 0.00022126, 0.00070567, 0.00015191, 0.00027467, 0.00031491, 0.00018326, 0.00081684, -0.00016251, -0.00066978, -0.00008037, -0.00016455, -0.00005397, -0.00012879, -0.00007610, -0.00006105, },
		{ -0.00003034, -0.00008024, -0.00008594, -0.00010177, -0.00008212, -0.00036509, -0.00027990, 0.00031664, 0.00031141, 0.00021537, 0.00000402, 0.00017043, 0.00006361, 0.00026901, 0.00040735, -0.00016352, -0.00092435, 0.00005761, 0.00068341, 0.00011773, 0.00021680, 0.00022386, 0.00014937, 0.00044802, 0.00049554, -0.00057174, -0.00030014, -0.00013488, -0.00007717, -0.00010222, -0.00009949, -0.00006074, },
		},
		{ { 0.00090603, 0.00033969, -0.00017298, -0.00000468, 0.00006236, 0.00013071, 0.00062902, 0.00014885, -0.00063999, -0.00007569, -0.00022475, 0.00006441, -0.00019187, 0.00011916, -0.00031758, 0.00014065, -0.00023013, 0.00014065, -0.00031758, 0.00011916, -0.00019187, 0.00006441, -0.00022475, -0.00007569, -0.00063999, 0.00014885, 0.00062902, 0.00013071, 0.00006236, -0.00000468, -0.00017298, 0.00033969, },
		{ 0.00028818, 0.00092391, 0.00038242, -0.00004480, 0.00002516, 0.00009292, 0.00032878, 0.00064745, -0.00026736, -0.00054167, -0.00006745, -0.00022811, 0.00009852, -0.00017636, 0.00005336, -0.00028565, 0.00016414, -0.00022838, 0.00008351, -0.00028266, 0.00015063, -0.00022183, 0.00004450, -0.00020375, -0.00016899, -0.00049681, 0.00041917, 0.00044425, 0.00006063, 0.00001398, -0.00000230, -0.00031486, },
		{ -0.00046915, 0.00022548, 0.00097091, 0.00043205, 0.00005663, 0.00009976, 0.00012543, 0.00066694, 0.00027003, -0.00050967, -0.00034142, -0.00015594, -0.00016518, 0.00010873, -0.00018363, -0.00000333, -0.00021302, 0.00014964, -0.00023007, 0.00002652, -0.00019937, 0.00012549, -0.00024094, 0.00005950, -0.00025235, -0.00018128, -0.00025520, 0.00048237, 0.00025205, 0.00006547, -0.00005535, 0.00002944, },
		{ 0.00007116, -0.00061554, 0.00012664, 0.00107061, 0.00046335, 0.00017186, 0.00017000, 0.00029458, 0.00090842, -0.00046269, -0.00031599, -0.00033135, -0.00015722, -0.00013397, 0.00013347, -0.00024685, 0.00001502, -0.00016975, 0.00013193, -0.00025510, 0.00001404, -0.00011797, 0.00004799, -0.00019649, 0.00003093, -0.00030700, -0.00007591, -0.00006913, 0.00041649, 0.00011308, 0.00009400, -0.00012806, },
		{ -0.00014883, 0.00006677, -0.00070055, -0.00005989, 0.00126871, 0.00042916, 0.00036627, 0.00018365, 0.00074840, 0.00060426, -0.00104101, 0.00010335, -0.00052623, -0.00000132, -0.00019951, 0.00019358, -0.00035602, 0.00012094, -0.00019918, 0.00014217, -0.00031192, 0.00007540, -0.00010926, -0.00000889, -0.00011533, -0.00008344, -0.00024614, 0.00002266, 0.00005997, 0.00026497, 0.00007068, 0.00007711, },
		{ 0.00002210, -0.00012902, 0.00002695, -0.00074237, -0.00029993, 0.00150349, 0.00041383, 0.00055171, 0.00032767, 0.00124263, -0.00035009, -0.00095656, 0.00025014, -0.00060875, 0.00012139, -0.00025503, 0.00020218, -0.00040414, 0.00021802, -0.00025092, 0.00015107, -0.00035629, 0.00017488, -0.00019664, 0.00004753, -0.00012972, -0.00015461, -0.00012183, 0.00007125, 0.00013637, 0.00009402, 0.00009120, },
		{ 0.00005380, 0.00003102, -0.00016917, 0.00004545, -0.00083385, -0.00048843, 0.00163433, 0.00059623, 0.00054881, 0.00095387, 0.00105994, -0.00128391, -0.00039690, 0.00000754, -0.00040972, 0.00005992, -0.00020570, 0.00010191, -0.00031532, 0.00020755, -0.00024606, 0.00010452, -0.00031542, 0.00020620, -0.00028223, 0.00017820, -0.00028707, -0.00004164, -0.00010474, 0.00016725, 0.00008566, 0.00002851, },
		{ 0.00008175, -0.00006378, 0.00011479, -0.00027537, 0.00011233, -0.00094144, -0.00068218, 0.00171815, 0.00092179, 0.00052422, 0.00193354, -0.00022284, -0.00130188, -0.00012727, -0.00016522, -0.00020070, -0.00002848, -0.00017294, 0.00002159, -0.00019359, 0.00013849, -0.00020733, 0.00003080, -0.00019585, 0.00010593, -0.00022487, 0.00019689, -0.00040750, 0.00015315, -0.00019313, 0.00031402, -0.00007255, },
		{ -0.00015388, 0.00009029, -0.00011185, 0.00011947, -0.00029337, 0.00006299, -0.00087179, -0.00112596, 0.00203957, 0.00107322, 0.00107875, 0.00227840, -0.00188427, -0.00039562, -0.00051850, 0.00010753, -0.00029204, 0.00007608, -0.00031845, 0.00012676, -0.00021006, 0.00014757, -0.00023793, 0.00004170, -0.00012659, -0.00002494, -0.00004413, 0.00001129, -0.00028751, 0.00017858, -0.00018427, 0.00033952, },
		{ 0.00017414, -0.00006799, -0.00002915, -0.00002246, -0.00001724, -0.00017641, -0.00012667, -0.00061799, -0.00183844, 0.00261156, 0.00112501, 0.00237653, 0.00101916, -0.00250114, 0.00038635, -0.00101611, 0.00053038, -0.00052637, 0.00025915, -0.00052734, 0.00033637, -0.00034220, 0.00023948, -0.00034262, 0.00017165, -0.00021607, -0.00000140, 0.00003833, -0.00018374, -0.00002337, 0.00000953, -0.00001039, },
		{ 0.00010235, 0.00002514, 0.00000540, -0.00010092, 0.00001418, -0.00009429, -0.00014812, -0.00018400, -0.00052884, -0.00240538, 0.00290509, 0.00200300, 0.00322184, -0.00131963, -0.00157143, 0.00019798, -0.00090385, 0.00055096, -0.00050177, 0.00021257, -0.00051502, 0.00038766, -0.00038331, 0.00025670, -0.00037278, 0.00027794, -0.00037554, 0.00019599, -0.00013341, -0.00011043, 0.00005847, -0.00007603, },
		{ 0.00009290, -0.00002573, 0.00007460, -0.00009383, 0.00001837, -0.00014461, 0.00002224, -0.00032733, 0.00001180, -0.00076652, -0.00263514, 0.00266732, 0.00454154, 0.00175777, -0.00267353, -0.00033769, -0.00045973, -0.00037191, 0.00022600, -0.00024412, -0.00003049, -0.00028506, 0.00023431, -0.00026835, 0.00013741, -0.00023872, 0.00020373, -0.00038612, 0.00033528, -0.00038899, 0.00019188, -0.00014675, },
		{ -0.00032581, 0.00027619, -0.00019146, 0.00013895, -0.00016416, 0.00009201, -0.00023347, 0.00002327, -0.00032485, -0.00003804, -0.00072326, -0.00332903, 0.00305529, 0.00775173, -0.00232306, -0.00148505, -0.00046396, -0.00034716, -0.00032465, 0.00021521, -0.00026146, -0.00002908, -0.00021887, 0.00017270, -0.00022194, 0.00007326, -0.00009741, 0.00002199, -0.00018580, 0.00019877, -0.00038842, 0.00034807, },
		{ 0.00012321, -0.00015973, 0.00014256, -0.00007045, -0.00008816, 0.00010684, -0.00022312, 0.00011674, -0.00048976, 0.00032822, -0.00093988, 0.00039197, -0.00564549, 0.00613686, 0.00868844, -0.00659672, 0.00132624, -0.00198277, 0.00081849, -0.00108858, 0.00080497, -0.00078098, 0.00043354, -0.00054610, 0.00041908, -0.00043212, 0.00025716, -0.00018008, -0.00001152, 0.00000205, -0.00010408, -0.00006630, },
		{ 0.00022675, -0.00022411, 0.00019235, -0.00020058, 0.00024188, -0.00048114, 0.00055601, -0.00073984, 0.00068060, -0.00122031, 0.00125350, -0.00226440, 0.00210252, -0.00919248, 0.01245957, 0.00499294, -0.00774056, 0.00304983, -0.00312968, 0.00184207, -0.00181948, 0.00135806, -0.00126805, 0.00091138, -0.00092646, 0.00071952, -0.00069408, 0.00053913, -0.00043943, 0.00020138, -0.00007442, -0.00022193, },
		{ 0.00000959, 0.00013163, -0.00023570, 0.00025216, -0.00027422, 0.00025738, -0.00050163, 0.00056440, -0.00075792, 0.00065977, -0.00120980, 0.00124446, -0.00248344, 0.00239442, -0.01125584, 0.01930810, -0.00226777, -0.00481302, 0.00202205, -0.00240556, 0.00148232, -0.00151732, 0.00107125, -0.00103126, 0.00078579, -0.00082680, 0.00062768, -0.00061006, 0.00054061, -0.00053644, 0.00042083, -0.00035509, },
		{ -0.00051204, 0.00034437, -0.00026551, 0.00015764, -0.00014481, 0.00013009, -0.00024157, 0.00008655, -0.00015103, 0.00005453, -0.00031946, -0.00000741, -0.00032999, -0.00065410, -0.00029561, -0.00891591, 0.02230703, -0.00891591, -0.00029561, -0.00065410, -0.00032999, -0.00000741, -0.00031946, 0.00005453, -0.00015103, 0.00008655, -0.00024157, 0.00013009, -0.00014481, 0.00015764, -0.00026551, 0.00034437, },
		{ 0.00000959, -0.00035509, 0.00042083, -0.00053644, 0.00054061, -0.00061006, 0.00062768, -0.00082680, 0.00078579, -0.00103126, 0.00107125, -0.00151732, 0.00148232, -0.00240556, 0.00202205, -0.00481302, -0.00226777, 0.01930810, -0.01125584, 0.00239442, -0.00248344, 0.00124446, -0.00120980, 0.00065977, -0.00075792, 0.00056440, -0.00050163, 0.00025738, -0.00027422, 0.00025216, -0.00023570, 0.00013163, },
		{ 0.00022675, -0.00022193, -0.00007442, 0.00020138, -0.00043943, 0.00053913, -0.00069408, 0.00071952, -0.00092646, 0.00091138, -0.00126805, 0.00135806, -0.00181948, 0.00184207, -0.00312968, 0.00304983, -0.00774056, 0.00499294, 0.01245957, -0.00919248, 0.00210252, -0.00226440, 0.00125350, -0.00122031, 0.00068060, -0.00073984, 0.00055601, -0.00048114, 0.00024188, -0.00020058, 0.00019235, -0.00022411, },
		{ 0.00012321, -0.00006630, -0.00010408, 0.00000205, -0.00001152, -0.00018008, 0.00025716, -0.00043212, 0.00041908, -0.00054610, 0.00043354, -0.00078098, 0.00080497, -0.00108858, 0.00081849, -0.00198277, 0.00132624, -0.00659672, 0.00868844, 0.00613686, -0.00564549, 0.00039197, -0.00093988, 0.00032822, -0.00048976, 0.00011674, -0.00022312, 0.00010684, -0.00008816, -0.00007045, 0.00014256, -0.00015973, },
		{ -0.00032581, 0.00034807, -0.00038842, 0.00019877, -0.00018580, 0.00002199, -0.00009741, 0.00007326, -0.00022194, 0.00017270, -0.00021887, -0.00002908, -0.00026146, 0.00021521, -0.00032465, -0.00034716, -0.00046396, -0.00148505, -0.00232306, 0.00775173, 0.00305529, -0.00332903, -0.00072326, -0.00003804, -0.00032485, 0.00002327, -0.00023347, 0.00009201, -0.00016416, 0.00013895, -0.00019146, 0.00027619, },
		{ 0.00009290, -0.00014675, 0.00019188, -0.00038899, 0.00033528, -0.00038612, 0.00020373, -0.00023872, 0.00013741, -0.00026835, 0.00023431, -0.00028506, -0.00003049, -0.00024412, 0.00022600, -0.00037191, -0.00045973, -0.00033769, -0.00267353, 0.00175777, 0.00454154, 0.00266732, -0.00263514, -0.00076652, 0.00001180, -0.00032733, 0.00002224, -0.00014461, 0.00001837, -0.00009383, 0.00007460, -0.00002573, },
		{ 0.00010235, -0.00007603, 0.00005847, -0.00011043, -0.00013341, 0.00019599, -0.00037554, 0.00027794, -0.00037278, 0.00025670, -0.00038331, 0.00038766, -0.00051502, 0.00021257, -0.00050177, 0.00055096, -0.00090385, 0.00019798, -0.00157143, -0.00131963, 0.00322184, 0.00200300, 0.00290509, -0.00240538, -0.00052884, -0.00018400, -0.00014812, -0.00009429, 0.00001418, -0.00010092, 0.00000540, 0.00002514, },
		{ 0.00017414, -0.00001039, 0.00000953, -0.00002337, -0.00018374, 0.00003833, -0.00000140, -0.00021607, 0.00017165, -0.00034262, 0.00023948, -0.00034220, 0.00033637, -0.00052734, 0.00025915, -0.00052637, 0.00053038, -0.00101611, 0.00038635, -0.00250114, 0.00101916, 0.00237653, 0.00112501, 0.00261156, -0.00183844, -0.00061799, -0.00012667, -0.00017641, -0.00001724, -0.00002246, -0.00002915, -0.00006799, },
		{ -0.00015388, 0.00033952, -0.00018427, 0.00017858, -0.00028751, 0.00001129, -0.00004413, -0.00002494, -0.00012659, 0.00004170, -0.00023793, 0.00014757, -0.00021006, 0.00012676, -0.00031845, 0.00007608, -0.00029204, 0.00010753, -0.00051850, -0.00039562, -0.00188427, 0.00227840, 0.00107875, 0.00107322, 0.00203957, -0.00112596, -0.00087179, 0.00006299, -0.00029337, 0.00011947, -0.00011185, 0.00009029, },
		{ 0.00008175, -0.00007255, 0.00031402, -0.00019313, 0.00015315, -0.00040750, 0.00019689, -0.00022487, 0.00010593, -0.00019585, 0.00003080, -0.00020733, 0.00013849, -0.00019359, 0.00002159, -0.00017294, -0.00002848, -0.00020070, -0.00016522, -0.00012727, -0.00130188, -0.00022284, 0.00193354, 0.00052422, 0.00092179, 0.00171815, -0.00068218, -0.00094144, 0.00011233, -0.00027537, 0.00011479, -0.00006378, },
		{ 0.00005380, 0.00002851, 0.00008566, 0.00016725, -0.00010474, -0.00004164, -0.00028707, 0.00017820, -0.00028223, 0.00020620, -0.00031542, 0.00010452, -0.00024606, 0.00020755, -0.00031532, 0.00010191, -0.00020570, 0.00005992, -0.00040972, 0.00000754, -0.00039690, -0.00128391, 0.00105994, 0.00095387, 0.00054881, 0.00059623, 0.00163433, -0.00048843, -0.00083385, 0.00004545, -0.00016917, 0.00003102, },
		{ 0.00002210, 0.00009120, 0.00009402, 0.00013637, 0.00007125, -0.00012183, -0.00015461, -0.00012972, 0.00004753, -0.00019664, 0.00017488, -0.00035629, 0.00015107, -0.00025092, 0.00021802, -0.00040414, 0.00020218, -0.00025503, 0.00012139, -0.00060875, 0.00025014, -0.00095656, -0.00035009, 0.00124263, 0.00032767, 0.00055171, 0.00041383, 0.00150349, -0.00029993, -0.00074237, 0.00002695, -0.00012902, },
		{ -0.00014883, 0.00007711, 0.00007068, 0.00026497, 0.00005997, 0.00002266, -0.00024614, -0.00008344, -0.00011533, -0.00000889, -0.00010926, 0.00007540, -0.00031192, 0.00014217, -0.00019918, 0.00012094, -0.00035602, 0.00019358, -0.00019951, -0.00000132, -0.00052623, 0.00010335, -0.00104101, 0.00060426, 0.00074840, 0.00018365, 0.00036627, 0.00042916, 0.00126871, -0.00005989, -0.00070055, 0.00006677, },
		{ 0.00007116, -0.00012806, 0.00009400, 0.00011308, 0.00041649, -0.00006913, -0.00007591, -0.00030700, 0.00003093, -0.00019649, 0.00004799, -0.00011797, 0.00001404, -0.00025510, 0.00013193, -0.00016975, 0.00001502, -0.00024685, 0.00013347, -0.00013397, -0.00015722, -0.00033135, -0.00031599, -0.00046269, 0.00090842, 0.00029458, 0.00017000, 0.00017186, 0.00046335, 0.00107061, 0.00012664, -0.00061554, },
		{ -0.00046915, 0.00002944, -0.00005535, 0.00006547, 0.00025205, 0.00048237, -0.00025520, -0.00018128, -0.00025235, 0.00005950, -0.00024094, 0.00012549, -0.00019937, 0.00002652, -0.00023007, 0.00014964, -0.00021302, -0.00000333, -0.00018363, 0.00010873, -0.00016518, -0.00015594, -0.00034142, -0.00050967, 0.00027003, 0.00066694, 0.00012543, 0.00009976, 0.00005663, 0.00043205, 0.00097091, 0.00022548, },
		{ 0.00028818, -0.00031486, -0.00000230, 0.00001398, 0.00006063, 0.00044425, 0.00041917, -0.00049681, -0.00016899, -0.00020375, 0.00004450, -0.00022183, 0.00015063, -0.00028266, 0.00008351, -0.00022838, 0.00016414, -0.00028565, 0.00005336, -0.00017636, 0.00009852, -0.00022811, -0.00006745, -0.00054167, -0.00026736, 0.00064745, 0.00032878, 0.00009292, 0.00002516, -0.00004480, 0.00038242, 0.00092391, },
		},
	}
	;
	float f32_6[64][32][32] = {
		{ { -0.00141721, 0.00155891, -0.00014121, -0.00081775, 0.00026383, -0.00024333, 0.00038147, -0.00023992, 0.00001708, -0.00008226, 0.00014242, -0.00012728, 0.00008683, -0.00002156, -0.00002662, -0.00000524, 0.00001277, -0.00000524, -0.00002662, -0.00002156, 0.00008683, -0.00012728, 0.00014242, -0.00008226, 0.00001708, -0.00023992, 0.00038147, -0.00024333, 0.00026383, -0.00081775, -0.00014121, 0.00155891, },
		{ 0.00180871, -0.00137741, 0.00126122, 0.00012396, -0.00092411, 0.00031566, -0.00029190, 0.00037929, -0.00020750, 0.00001708, -0.00009974, 0.00013640, -0.00010969, 0.00007599, -0.00001001, -0.00003726, -0.00000281, 0.00001277, -0.00000812, -0.00001613, -0.00003135, 0.00009572, -0.00014115, 0.00014145, -0.00006043, 0.00001708, -0.00026586, 0.00036799, -0.00018179, 0.00020068, -0.00069228, -0.00041012, },
		{ -0.00067243, 0.00200102, -0.00125952, 0.00092707, 0.00037520, -0.00100728, 0.00035418, -0.00032565, 0.00036154, -0.00016986, 0.00001708, -0.00011218, 0.00012363, -0.00008909, 0.00006363, 0.00000285, -0.00004764, -0.00000095, 0.00001277, -0.00001136, -0.00000619, -0.00003900, 0.00010233, -0.00015080, 0.00013352, -0.00003508, 0.00001708, -0.00028434, 0.00033937, -0.00010967, 0.00012862, -0.00055252, },
		{ -0.00040384, -0.00091807, 0.00212844, -0.00106809, 0.00056931, 0.00060286, -0.00106406, 0.00037789, -0.00034328, 0.00032889, -0.00012844, 0.00001708, -0.00011912, 0.00010458, -0.00006625, 0.00005022, 0.00001653, -0.00005736, 0.00000029, 0.00001277, -0.00001483, 0.00000282, -0.00004423, 0.00010640, -0.00015583, 0.00011895, -0.00000718, 0.00001708, -0.00029464, 0.00029671, -0.00002971, 0.00005044, },
		{ -0.00003087, -0.00025196, -0.00113759, 0.00218608, -0.00081045, 0.00020168, 0.00079818, -0.00109226, 0.00038590, -0.00034411, 0.00028260, -0.00008484, 0.00001708, -0.00012028, 0.00008000, -0.00004205, 0.00003627, 0.00003050, -0.00006605, 0.00000085, 0.00001277, -0.00001840, 0.00001054, -0.00004682, 0.00010777, -0.00015607, 0.00009828, 0.00002219, 0.00001708, -0.00029636, 0.00024165, 0.00005499, },
		{ 0.00014118, -0.00011218, -0.00010271, -0.00132256, 0.00217172, -0.00049653, -0.00016168, 0.00095367, -0.00109080, 0.00037789, -0.00032810, 0.00022445, -0.00004072, 0.00001708, -0.00011562, 0.00005082, -0.00001742, 0.00002232, 0.00004424, -0.00007337, 0.00000071, 0.00001277, -0.00002193, 0.00001670, -0.00004669, 0.00010640, -0.00015150, 0.00007232, 0.00005190, 0.00001708, -0.00028944, 0.00017630, },
		{ 0.00010317, 0.00022556, -0.00019036, 0.00003816, -0.00146587, 0.00208592, -0.00013838, -0.00050681, 0.00106334, -0.00105974, 0.00035418, -0.00029588, 0.00015668, 0.00000222, 0.00001708, -0.00010532, 0.00001817, 0.00000669, 0.00000890, 0.00005720, -0.00007904, -0.00000012, 0.00001277, -0.00002527, 0.00002104, -0.00004383, 0.00010233, -0.00014229, 0.00004206, 0.00008082, 0.00001708, -0.00027415, },
		{ -0.00025107, 0.00002508, 0.00030488, -0.00026242, 0.00016526, -0.00156201, 0.00193198, 0.00025024, -0.00082046, 0.00112298, -0.00100027, 0.00031566, -0.00024868, 0.00008189, 0.00004232, 0.00001708, -0.00008977, -0.00001669, 0.00002935, -0.00000346, 0.00006890, -0.00008284, -0.00000162, 0.00001277, -0.00002832, 0.00002340, -0.00003836, 0.00009572, -0.00012881, 0.00000867, 0.00010783, 0.00001708, },
		{ 0.00001708, -0.00022109, -0.00005497, 0.00037608, -0.00032558, 0.00027368, -0.00160729, 0.00171580, 0.00065439, -0.00109056, 0.00113031, -0.00091468, 0.00026383, -0.00018832, 0.00000295, 0.00007805, 0.00001708, -0.00006958, -0.00005243, 0.00004969, -0.00001430, 0.00007887, -0.00008463, -0.00000371, 0.00001277, -0.00003094, 0.00002368, -0.00003048, 0.00008683, -0.00011156, -0.00002657, 0.00013189, },
		{ 0.00015208, 0.00001708, -0.00018536, -0.00013391, 0.00043645, -0.00037741, 0.00035928, -0.00159997, 0.00144570, 0.00105854, -0.00130673, 0.00108503, -0.00080625, 0.00020068, -0.00011711, -0.00007710, 0.00010803, 0.00001708, -0.00004552, -0.00008768, 0.00006694, -0.00002319, 0.00008675, -0.00008434, -0.00000633, 0.00001277, -0.00003303, 0.00002189, -0.00002050, 0.00007599, -0.00009121, -0.00006231, },
		{ -0.00009718, 0.00016763, 0.00001708, -0.00014526, -0.00020870, 0.00048365, -0.00041592, 0.00041874, -0.00154032, 0.00113205, 0.00144716, -0.00146068, 0.00098889, -0.00067916, 0.00012862, -0.00003779, -0.00015519, 0.00013111, 0.00001708, -0.00001851, -0.00012107, 0.00008042, -0.00002980, 0.00009222, -0.00008198, -0.00000938, 0.00001277, -0.00003453, 0.00001809, -0.00000881, 0.00006363, -0.00006855, },
		{ -0.00004445, -0.00012983, 0.00017793, 0.00001708, -0.00010233, -0.00027647, 0.00051587, -0.00043964, 0.00044981, -0.00143065, 0.00078692, 0.00180531, -0.00154648, 0.00084558, -0.00053828, 0.00005044, 0.00004658, -0.00022832, 0.00014640, 0.00001708, 0.00001041, -0.00015133, 0.00008963, -0.00003387, 0.00009508, -0.00007764, -0.00001272, 0.00001277, -0.00003536, 0.00001242, 0.00000416, 0.00005022, },
		{ 0.00003627, -0.00001982, -0.00015900, 0.00018259, 0.00001708, -0.00005821, -0.00033462, 0.00053187, -0.00044765, 0.00045126, -0.00127516, 0.00042356, 0.00211923, -0.00156083, 0.00066061, -0.00038903, -0.00003087, 0.00013278, -0.00029367, 0.00015332, 0.00001708, 0.00004012, -0.00017729, 0.00009420, -0.00003524, 0.00009521, -0.00007149, -0.00001625, 0.00001277, -0.00003550, 0.00000510, 0.00001789, },
		{ 0.00003187, 0.00002232, 0.00000438, -0.00018359, 0.00018143, 0.00001708, -0.00001460, -0.00038091, 0.00053105, -0.00043964, 0.00042306, -0.00107984, 0.00005593, 0.00237686, -0.00150319, 0.00044109, -0.00023715, -0.00011218, 0.00021748, -0.00034873, 0.00015159, 0.00001708, 0.00006949, -0.00019795, 0.00009397, -0.00003387, 0.00009262, -0.00006376, -0.00001981, 0.00001277, -0.00003494, -0.00000358, },
		{ -0.00001330, 0.00004555, 0.00000890, 0.00002722, -0.00020263, 0.00017449, 0.00001708, 0.00002682, -0.00041356, 0.00051342, -0.00041592, 0.00036629, -0.00085218, -0.00030183, 0.00256830, -0.00137577, 0.00019545, -0.00008848, -0.00019036, 0.00029743, -0.00039139, 0.00014130, 0.00001708, 0.00009739, -0.00021253, 0.00008893, -0.00002980, 0.00008739, -0.00005475, -0.00002329, 0.00001277, -0.00003370, },
		{ -0.00003184, -0.00002368, 0.00005841, -0.00000346, 0.00004783, -0.00021541, 0.00016205, 0.00001708, 0.00006446, -0.00043131, 0.00047967, -0.00037741, 0.00028312, -0.00060094, -0.00063597, 0.00268619, -0.00118347, -0.00006686, 0.00005128, -0.00026242, 0.00036956, -0.00042001, 0.00012282, 0.00001708, 0.00012274, -0.00022046, 0.00007929, -0.00002319, 0.00007974, -0.00004481, -0.00002653, 0.00001277, },
		{ 0.00001277, -0.00002941, -0.00003432, 0.00006995, -0.00001430, 0.00006541, -0.00022143, 0.00014457, 0.00001708, 0.00009688, -0.00043349, 0.00043109, -0.00032558, 0.00017676, -0.00033577, -0.00093367, 0.00272599, -0.00093367, -0.00033577, 0.00017676, -0.00032558, 0.00043109, -0.00043349, 0.00009688, 0.00001708, 0.00014457, -0.00022143, 0.00006541, -0.00001430, 0.00006995, -0.00003432, -0.00002941, },
		{ -0.00003184, 0.00001277, -0.00002653, -0.00004481, 0.00007974, -0.00002319, 0.00007929, -0.00022046, 0.00012274, 0.00001708, 0.00012282, -0.00042001, 0.00036956, -0.00026242, 0.00005128, -0.00006686, -0.00118347, 0.00268619, -0.00063597, -0.00060094, 0.00028312, -0.00037741, 0.00047967, -0.00043131, 0.00006446, 0.00001708, 0.00016205, -0.00021541, 0.00004783, -0.00000346, 0.00005841, -0.00002368, },
		{ -0.00001330, -0.00003370, 0.00001277, -0.00002329, -0.00005475, 0.00008739, -0.00002980, 0.00008893, -0.00021253, 0.00009739, 0.00001708, 0.00014130, -0.00039139, 0.00029743, -0.00019036, -0.00008848, 0.00019545, -0.00137577, 0.00256830, -0.00030183, -0.00085218, 0.00036629, -0.00041592, 0.00051342, -0.00041356, 0.00002682, 0.00001708, 0.00017449, -0.00020263, 0.00002722, 0.00000890, 0.00004555, },
		{ 0.00003187, -0.00000358, -0.00003494, 0.00001277, -0.00001981, -0.00006376, 0.00009262, -0.00003387, 0.00009397, -0.00019795, 0.00006949, 0.00001708, 0.00015159, -0.00034873, 0.00021748, -0.00011218, -0.00023715, 0.00044109, -0.00150319, 0.00237686, 0.00005593, -0.00107984, 0.00042306, -0.00043964, 0.00053105, -0.00038091, -0.00001460, 0.00001708, 0.00018143, -0.00018359, 0.00000438, 0.00002232, },
		{ 0.00003627, 0.00001789, 0.00000510, -0.00003550, 0.00001277, -0.00001625, -0.00007149, 0.00009521, -0.00003524, 0.00009420, -0.00017729, 0.00004012, 0.00001708, 0.00015332, -0.00029367, 0.00013278, -0.00003087, -0.00038903, 0.00066061, -0.00156083, 0.00211923, 0.00042356, -0.00127516, 0.00045126, -0.00044765, 0.00053187, -0.00033462, -0.00005821, 0.00001708, 0.00018259, -0.00015900, -0.00001982, },
		{ -0.00004445, 0.00005022, 0.00000416, 0.00001242, -0.00003536, 0.00001277, -0.00001272, -0.00007764, 0.00009508, -0.00003387, 0.00008963, -0.00015133, 0.00001041, 0.00001708, 0.00014640, -0.00022832, 0.00004658, 0.00005044, -0.00053828, 0.00084558, -0.00154648, 0.00180531, 0.00078692, -0.00143065, 0.00044981, -0.00043964, 0.00051587, -0.00027647, -0.00010233, 0.00001708, 0.00017793, -0.00012983, },
		{ -0.00009718, -0.00006855, 0.00006363, -0.00000881, 0.00001809, -0.00003453, 0.00001277, -0.00000938, -0.00008198, 0.00009222, -0.00002980, 0.00008042, -0.00012107, -0.00001851, 0.00001708, 0.00013111, -0.00015519, -0.00003779, 0.00012862, -0.00067916, 0.00098889, -0.00146068, 0.00144716, 0.00113205, -0.00154032, 0.00041874, -0.00041592, 0.00048365, -0.00020870, -0.00014526, 0.00001708, 0.00016763, },
		{ 0.00015208, -0.00006231, -0.00009121, 0.00007599, -0.00002050, 0.00002189, -0.00003303, 0.00001277, -0.00000633, -0.00008434, 0.00008675, -0.00002319, 0.00006694, -0.00008768, -0.00004552, 0.00001708, 0.00010803, -0.00007710, -0.00011711, 0.00020068, -0.00080625, 0.00108503, -0.00130673, 0.00105854, 0.00144570, -0.00159997, 0.00035928, -0.00037741, 0.00043645, -0.00013391, -0.00018536, 0.00001708, },
		{ 0.00001708, 0.00013189, -0.00002657, -0.00011156, 0.00008683, -0.00003048, 0.00002368, -0.00003094, 0.00001277, -0.00000371, -0.00008463, 0.00007887, -0.00001430, 0.00004969, -0.00005243, -0.00006958, 0.00001708, 0.00007805, 0.00000295, -0.00018832, 0.00026383, -0.00091468, 0.00113031, -0.00109056, 0.00065439, 0.00171580, -0.00160729, 0.00027368, -0.00032558, 0.00037608, -0.00005497, -0.00022109, },
		{ -0.00025107, 0.00001708, 0.00010783, 0.00000867, -0.00012881, 0.00009572, -0.00003836, 0.00002340, -0.00002832, 0.00001277, -0.00000162, -0.00008284, 0.00006890, -0.00000346, 0.00002935, -0.00001669, -0.00008977, 0.00001708, 0.00004232, 0.00008189, -0.00024868, 0.00031566, -0.00100027, 0.00112298, -0.00082046, 0.00025024, 0.00193198, -0.00156201, 0.00016526, -0.00026242, 0.00030488, 0.00002508, },
		{ 0.00010317, -0.00027415, 0.00001708, 0.00008082, 0.00004206, -0.00014229, 0.00010233, -0.00004383, 0.00002104, -0.00002527, 0.00001277, -0.00000012, -0.00007904, 0.00005720, 0.00000890, 0.00000669, 0.00001817, -0.00010532, 0.00001708, 0.00000222, 0.00015668, -0.00029588, 0.00035418, -0.00105974, 0.00106334, -0.00050681, -0.00013838, 0.00208592, -0.00146587, 0.00003816, -0.00019036, 0.00022556, },
		{ 0.00014118, 0.00017630, -0.00028944, 0.00001708, 0.00005190, 0.00007232, -0.00015150, 0.00010640, -0.00004669, 0.00001670, -0.00002193, 0.00001277, 0.00000071, -0.00007337, 0.00004424, 0.00002232, -0.00001742, 0.00005082, -0.00011562, 0.00001708, -0.00004072, 0.00022445, -0.00032810, 0.00037789, -0.00109080, 0.00095367, -0.00016168, -0.00049653, 0.00217172, -0.00132256, -0.00010271, -0.00011218, },
		{ -0.00003087, 0.00005499, 0.00024165, -0.00029636, 0.00001708, 0.00002219, 0.00009828, -0.00015607, 0.00010777, -0.00004682, 0.00001054, -0.00001840, 0.00001277, 0.00000085, -0.00006605, 0.00003050, 0.00003627, -0.00004205, 0.00008000, -0.00012028, 0.00001708, -0.00008484, 0.00028260, -0.00034411, 0.00038590, -0.00109226, 0.00079818, 0.00020168, -0.00081045, 0.00218608, -0.00113759, -0.00025196, },
		{ -0.00040384, 0.00005044, -0.00002971, 0.00029671, -0.00029464, 0.00001708, -0.00000718, 0.00011895, -0.00015583, 0.00010640, -0.00004423, 0.00000282, -0.00001483, 0.00001277, 0.00000029, -0.00005736, 0.00001653, 0.00005022, -0.00006625, 0.00010458, -0.00011912, 0.00001708, -0.00012844, 0.00032889, -0.00034328, 0.00037789, -0.00106406, 0.00060286, 0.00056931, -0.00106809, 0.00212844, -0.00091807, },
		{ -0.00067243, -0.00055252, 0.00012862, -0.00010967, 0.00033937, -0.00028434, 0.00001708, -0.00003508, 0.00013352, -0.00015080, 0.00010233, -0.00003900, -0.00000619, -0.00001136, 0.00001277, -0.00000095, -0.00004764, 0.00000285, 0.00006363, -0.00008909, 0.00012363, -0.00011218, 0.00001708, -0.00016986, 0.00036154, -0.00032565, 0.00035418, -0.00100728, 0.00037520, 0.00092707, -0.00125952, 0.00200102, },
		{ 0.00180871, -0.00041012, -0.00069228, 0.00020068, -0.00018179, 0.00036799, -0.00026586, 0.00001708, -0.00006043, 0.00014145, -0.00014115, 0.00009572, -0.00003135, -0.00001613, -0.00000812, 0.00001277, -0.00000281, -0.00003726, -0.00001001, 0.00007599, -0.00010969, 0.00013640, -0.00009974, 0.00001708, -0.00020750, 0.00037929, -0.00029190, 0.00031566, -0.00092411, 0.00012396, 0.00126122, -0.00137741, },
		},
		{ { 0.00005017, -0.00081722, 0.00101400, 0.00063390, -0.00089392, 0.00054782, -0.00041605, -0.00014940, 0.00014656, -0.00012991, 0.00009532, -0.00001282, 0.00000737, 0.00002189, 0.00000804, -0.00007269, 0.00007031, -0.00007269, 0.00000804, 0.00002189, 0.00000737, -0.00001282, 0.00009532, -0.00012991, 0.00014656, -0.00014940, -0.00041605, 0.00054782, -0.00089392, 0.00063390, 0.00101400, -0.00081722, },
		{ -0.00080966, -0.00001120, -0.00069426, 0.00056291, 0.00096032, -0.00091544, 0.00043380, -0.00022924, -0.00027539, 0.00020214, -0.00014886, 0.00008728, -0.00000509, -0.00000579, 0.00001603, 0.00003132, -0.00008553, 0.00006402, -0.00005116, -0.00001547, 0.00002390, 0.00001760, -0.00001113, 0.00008957, -0.00009354, 0.00007563, -0.00001885, -0.00056498, 0.00057600, -0.00075991, 0.00022658, 0.00141473, },
		{ 0.00170692, -0.00064738, -0.00018322, -0.00048420, 0.00012373, 0.00116522, -0.00082074, 0.00025254, -0.00002932, -0.00037941, 0.00023346, -0.00014649, 0.00006502, 0.00001042, -0.00001964, 0.00000665, 0.00005124, -0.00008727, 0.00004613, -0.00002461, -0.00003620, 0.00002224, 0.00002313, -0.00000006, 0.00007265, -0.00004606, 0.00000033, 0.00009859, -0.00065652, 0.00051234, -0.00053382, -0.00020934, },
		{ -0.00061750, 0.00184532, -0.00033006, -0.00043160, -0.00024211, -0.00024635, 0.00122549, -0.00062331, 0.00003231, 0.00015726, -0.00044685, 0.00023508, -0.00012191, 0.00003050, 0.00003069, -0.00003181, -0.00000544, 0.00006508, -0.00007716, 0.00001945, 0.00000260, -0.00005162, 0.00001755, 0.00002295, 0.00001871, 0.00004899, 0.00000481, -0.00006792, 0.00018765, -0.00067928, 0.00036438, -0.00024957, },
		{ 0.00005051, -0.00094561, 0.00180446, 0.00011774, -0.00070536, -0.00002545, -0.00050357, 0.00113875, -0.00035188, -0.00019323, 0.00030631, -0.00046810, 0.00020585, -0.00007749, -0.00001220, 0.00005181, -0.00004018, -0.00001902, 0.00007088, -0.00005621, -0.00001179, 0.00002610, -0.00006009, 0.00001084, 0.00001694, 0.00004209, 0.00002400, 0.00005115, -0.00011907, 0.00023763, -0.00063165, 0.00015216, },
		{ -0.00009476, 0.00032208, -0.00115308, 0.00158273, 0.00064939, -0.00094512, 0.00011546, -0.00062379, 0.00092344, -0.00004605, -0.00039019, 0.00039938, -0.00043998, 0.00014917, -0.00001846, -0.00005746, 0.00006960, -0.00004319, -0.00003268, 0.00006778, -0.00002710, -0.00004264, 0.00004217, -0.00006109, 0.00000330, 0.00000596, 0.00006609, 0.00000326, 0.00008601, -0.00014590, 0.00024396, -0.00052194, },
		{ -0.00036709, -0.00034165, 0.00052553, -0.00121686, 0.00120289, 0.00120248, -0.00109277, 0.00014581, -0.00060558, 0.00061580, 0.00024962, -0.00052959, 0.00042640, -0.00036623, 0.00007246, 0.00004774, -0.00009895, 0.00008025, -0.00004014, -0.00004494, 0.00005610, 0.00000624, -0.00006816, 0.00004833, -0.00005521, -0.00000387, -0.00000837, 0.00008639, -0.00000846, 0.00010451, -0.00014510, 0.00020876, },
		{ 0.00014041, -0.00019008, -0.00055385, 0.00063191, -0.00113455, 0.00070909, 0.00170759, -0.00110123, 0.00005204, -0.00046978, 0.00026423, 0.00049237, -0.00059172, 0.00038708, -0.00025694, -0.00001387, 0.00011260, -0.00013056, 0.00008099, -0.00003125, -0.00005452, 0.00003734, 0.00003921, -0.00008417, 0.00004371, -0.00004411, -0.00000966, -0.00002384, 0.00009908, -0.00000797, 0.00010454, -0.00011778, },
		{ -0.00006923, 0.00005213, -0.00001646, -0.00070188, 0.00062728, -0.00092456, 0.00016060, 0.00209832, -0.00094247, -0.00015567, -0.00025565, -0.00007811, 0.00064762, -0.00056910, 0.00029079, -0.00012709, -0.00009801, 0.00016773, -0.00014729, 0.00007054, -0.00001766, -0.00006046, 0.00001393, 0.00006719, -0.00008794, 0.00002918, -0.00003017, -0.00001339, -0.00003801, 0.00010126, 0.00000581, 0.00008706, },
		{ 0.00005592, -0.00000810, -0.00004010, 0.00012949, -0.00076584, 0.00051472, -0.00062301, -0.00037653, 0.00232095, -0.00061288, -0.00044486, -0.00001417, -0.00036143, 0.00069416, -0.00046751, 0.00015504, 0.00000570, -0.00016845, 0.00020610, -0.00014609, 0.00004934, -0.00000128, -0.00006228, -0.00001108, 0.00008627, -0.00007856, 0.00000721, -0.00001615, -0.00001480, -0.00004860, 0.00009150, 0.00003164, },
		{ 0.00006611, 0.00001723, 0.00005500, -0.00012010, 0.00022838, -0.00073836, 0.00031380, -0.00027815, -0.00083891, 0.00234244, -0.00013485, -0.00076572, 0.00020042, -0.00054677, 0.00062730, -0.00030507, 0.00000262, 0.00012374, -0.00021572, 0.00022305, -0.00012630, 0.00001953, 0.00001554, -0.00006001, -0.00003453, 0.00009380, -0.00005715, -0.00001852, -0.00000472, -0.00001410, -0.00005382, 0.00007014, },
		{ 0.00003927, 0.00010412, -0.00002163, 0.00010915, -0.00017402, 0.00026853, -0.00062558, 0.00005749, 0.00005715, -0.00117475, 0.00215541, 0.00044558, -0.00105891, 0.00033957, -0.00061178, 0.00045945, -0.00010934, -0.00014184, 0.00021189, -0.00023376, 0.00021698, -0.00008976, -0.00001532, 0.00003030, -0.00005420, -0.00005355, 0.00008880, -0.00002666, -0.00004367, 0.00000197, -0.00001190, -0.00005268, },
		{ -0.00004511, 0.00000243, 0.00013969, -0.00005320, 0.00014481, -0.00019253, 0.00024768, -0.00044619, -0.00021296, 0.00033341, -0.00135097, 0.00177941, 0.00106466, -0.00126502, 0.00036839, -0.00055389, 0.00021821, 0.00008697, -0.00025577, 0.00025975, -0.00022080, 0.00018952, -0.00004063, -0.00005073, 0.00004078, -0.00004581, -0.00006599, 0.00007210, 0.00000855, -0.00006396, 0.00000261, -0.00000910, },
		{ -0.00000676, -0.00003200, -0.00003585, 0.00016687, -0.00007109, 0.00015537, -0.00017226, 0.00017308, -0.00022851, -0.00045434, 0.00051204, -0.00135763, 0.00125831, 0.00164972, -0.00133414, 0.00027153, -0.00039042, -0.00005799, 0.00025110, -0.00032206, 0.00026310, -0.00017960, 0.00014527, 0.00001523, -0.00008197, 0.00004536, -0.00003608, -0.00007072, 0.00004621, 0.00004338, -0.00007578, -0.00000303, },
		{ -0.00001408, -0.00000587, -0.00001508, -0.00007073, 0.00018061, -0.00007109, 0.00013814, -0.00011635, 0.00006035, -0.00000612, -0.00062821, 0.00057108, -0.00120873, 0.00065411, 0.00212942, -0.00123419, 0.00005588, -0.00015564, -0.00032572, 0.00035531, -0.00033169, 0.00022448, -0.00011701, 0.00009107, 0.00007107, -0.00010472, 0.00004327, -0.00002630, -0.00006777, 0.00001492, 0.00007271, -0.00007683, },
		{ -0.00006642, -0.00002872, -0.00000719, 0.00000334, -0.00009772, 0.00017766, -0.00005179, 0.00009488, -0.00003397, -0.00006916, 0.00018726, -0.00070676, 0.00050833, -0.00093952, 0.00003820, 0.00244389, -0.00095666, -0.00025011, 0.00010486, -0.00054330, 0.00038126, -0.00028502, 0.00015264, -0.00004287, 0.00003501, 0.00012022, -0.00011575, 0.00003468, -0.00001769, -0.00005830, -0.00001722, 0.00009224, },
		{ 0.00009908, -0.00004564, -0.00004440, -0.00001115, 0.00002073, -0.00011337, 0.00015703, -0.00001489, 0.00003147, 0.00006110, -0.00019172, 0.00032286, -0.00067710, 0.00034143, -0.00060065, -0.00051875, 0.00255336, -0.00051875, -0.00060065, 0.00034143, -0.00067710, 0.00032286, -0.00019172, 0.00006110, 0.00003147, -0.00001489, 0.00015703, -0.00011337, 0.00002073, -0.00001115, -0.00004440, -0.00004564, },
		{ -0.00006642, 0.00009224, -0.00001722, -0.00005830, -0.00001769, 0.00003468, -0.00011575, 0.00012022, 0.00003501, -0.00004287, 0.00015264, -0.00028502, 0.00038126, -0.00054330, 0.00010486, -0.00025011, -0.00095666, 0.00244389, 0.00003820, -0.00093952, 0.00050833, -0.00070676, 0.00018726, -0.00006916, -0.00003397, 0.00009488, -0.00005179, 0.00017766, -0.00009772, 0.00000334, -0.00000719, -0.00002872, },
		{ -0.00001408, -0.00007683, 0.00007271, 0.00001492, -0.00006777, -0.00002630, 0.00004327, -0.00010472, 0.00007107, 0.00009107, -0.00011701, 0.00022448, -0.00033169, 0.00035531, -0.00032572, -0.00015564, 0.00005588, -0.00123419, 0.00212942, 0.00065411, -0.00120873, 0.00057108, -0.00062821, -0.00000612, 0.00006035, -0.00011635, 0.00013814, -0.00007109, 0.00018061, -0.00007073, -0.00001508, -0.00000587, },
		{ -0.00000676, -0.00000303, -0.00007578, 0.00004338, 0.00004621, -0.00007072, -0.00003608, 0.00004536, -0.00008197, 0.00001523, 0.00014527, -0.00017960, 0.00026310, -0.00032206, 0.00025110, -0.00005799, -0.00039042, 0.00027153, -0.00133414, 0.00164972, 0.00125831, -0.00135763, 0.00051204, -0.00045434, -0.00022851, 0.00017308, -0.00017226, 0.00015537, -0.00007109, 0.00016687, -0.00003585, -0.00003200, },
		{ -0.00004511, -0.00000910, 0.00000261, -0.00006396, 0.00000855, 0.00007210, -0.00006599, -0.00004581, 0.00004078, -0.00005073, -0.00004063, 0.00018952, -0.00022080, 0.00025975, -0.00025577, 0.00008697, 0.00021821, -0.00055389, 0.00036839, -0.00126502, 0.00106466, 0.00177941, -0.00135097, 0.00033341, -0.00021296, -0.00044619, 0.00024768, -0.00019253, 0.00014481, -0.00005320, 0.00013969, 0.00000243, },
		{ 0.00003927, -0.00005268, -0.00001190, 0.00000197, -0.00004367, -0.00002666, 0.00008880, -0.00005355, -0.00005420, 0.00003030, -0.00001532, -0.00008976, 0.00021698, -0.00023376, 0.00021189, -0.00014184, -0.00010934, 0.00045945, -0.00061178, 0.00033957, -0.00105891, 0.00044558, 0.00215541, -0.00117475, 0.00005715, 0.00005749, -0.00062558, 0.00026853, -0.00017402, 0.00010915, -0.00002163, 0.00010412, },
		{ 0.00006611, 0.00007014, -0.00005382, -0.00001410, -0.00000472, -0.00001852, -0.00005715, 0.00009380, -0.00003453, -0.00006001, 0.00001554, 0.00001953, -0.00012630, 0.00022305, -0.00021572, 0.00012374, 0.00000262, -0.00030507, 0.00062730, -0.00054677, 0.00020042, -0.00076572, -0.00013485, 0.00234244, -0.00083891, -0.00027815, 0.00031380, -0.00073836, 0.00022838, -0.00012010, 0.00005500, 0.00001723, },
		{ 0.00005592, 0.00003164, 0.00009150, -0.00004860, -0.00001480, -0.00001615, 0.00000721, -0.00007856, 0.00008627, -0.00001108, -0.00006228, -0.00000128, 0.00004934, -0.00014609, 0.00020610, -0.00016845, 0.00000570, 0.00015504, -0.00046751, 0.00069416, -0.00036143, -0.00001417, -0.00044486, -0.00061288, 0.00232095, -0.00037653, -0.00062301, 0.00051472, -0.00076584, 0.00012949, -0.00004010, -0.00000810, },
		{ -0.00006923, 0.00008706, 0.00000581, 0.00010126, -0.00003801, -0.00001339, -0.00003017, 0.00002918, -0.00008794, 0.00006719, 0.00001393, -0.00006046, -0.00001766, 0.00007054, -0.00014729, 0.00016773, -0.00009801, -0.00012709, 0.00029079, -0.00056910, 0.00064762, -0.00007811, -0.00025565, -0.00015567, -0.00094247, 0.00209832, 0.00016060, -0.00092456, 0.00062728, -0.00070188, -0.00001646, 0.00005213, },
		{ 0.00014041, -0.00011778, 0.00010454, -0.00000797, 0.00009908, -0.00002384, -0.00000966, -0.00004411, 0.00004371, -0.00008417, 0.00003921, 0.00003734, -0.00005452, -0.00003125, 0.00008099, -0.00013056, 0.00011260, -0.00001387, -0.00025694, 0.00038708, -0.00059172, 0.00049237, 0.00026423, -0.00046978, 0.00005204, -0.00110123, 0.00170759, 0.00070909, -0.00113455, 0.00063191, -0.00055385, -0.00019008, },
		{ -0.00036709, 0.00020876, -0.00014510, 0.00010451, -0.00000846, 0.00008639, -0.00000837, -0.00000387, -0.00005521, 0.00004833, -0.00006816, 0.00000624, 0.00005610, -0.00004494, -0.00004014, 0.00008025, -0.00009895, 0.00004774, 0.00007246, -0.00036623, 0.00042640, -0.00052959, 0.00024962, 0.00061580, -0.00060558, 0.00014581, -0.00109277, 0.00120248, 0.00120289, -0.00121686, 0.00052553, -0.00034165, },
		{ -0.00009476, -0.00052194, 0.00024396, -0.00014590, 0.00008601, 0.00000326, 0.00006609, 0.00000596, 0.00000330, -0.00006109, 0.00004217, -0.00004264, -0.00002710, 0.00006778, -0.00003268, -0.00004319, 0.00006960, -0.00005746, -0.00001846, 0.00014917, -0.00043998, 0.00039938, -0.00039019, -0.00004605, 0.00092344, -0.00062379, 0.00011546, -0.00094512, 0.00064939, 0.00158273, -0.00115308, 0.00032208, },
		{ 0.00005051, 0.00015216, -0.00063165, 0.00023763, -0.00011907, 0.00005115, 0.00002400, 0.00004209, 0.00001694, 0.00001084, -0.00006009, 0.00002610, -0.00001179, -0.00005621, 0.00007088, -0.00001902, -0.00004018, 0.00005181, -0.00001220, -0.00007749, 0.00020585, -0.00046810, 0.00030631, -0.00019323, -0.00035188, 0.00113875, -0.00050357, -0.00002545, -0.00070536, 0.00011774, 0.00180446, -0.00094561, },
		{ -0.00061750, -0.00024957, 0.00036438, -0.00067928, 0.00018765, -0.00006792, 0.00000481, 0.00004899, 0.00001871, 0.00002295, 0.00001755, -0.00005162, 0.00000260, 0.00001945, -0.00007716, 0.00006508, -0.00000544, -0.00003181, 0.00003069, 0.00003050, -0.00012191, 0.00023508, -0.00044685, 0.00015726, 0.00003231, -0.00062331, 0.00122549, -0.00024635, -0.00024211, -0.00043160, -0.00033006, 0.00184532, },
		{ 0.00170692, -0.00020934, -0.00053382, 0.00051234, -0.00065652, 0.00009859, 0.00000033, -0.00004606, 0.00007265, -0.00000006, 0.00002313, 0.00002224, -0.00003620, -0.00002461, 0.00004613, -0.00008727, 0.00005124, 0.00000665, -0.00001964, 0.00001042, 0.00006502, -0.00014649, 0.00023346, -0.00037941, -0.00002932, 0.00025254, -0.00082074, 0.00116522, 0.00012373, -0.00048420, -0.00018322, -0.00064738, },
		{ -0.00080966, 0.00141473, 0.00022658, -0.00075991, 0.00057600, -0.00056498, -0.00001885, 0.00007563, -0.00009354, 0.00008957, -0.00001113, 0.00001760, 0.00002390, -0.00001547, -0.00005116, 0.00006402, -0.00008553, 0.00003132, 0.00001603, -0.00000579, -0.00000509, 0.00008728, -0.00014886, 0.00020214, -0.00027539, -0.00022924, 0.00043380, -0.00091544, 0.00096032, 0.00056291, -0.00069426, -0.00001120, },
		},
		{ { -0.00008218, 0.00002517, -0.00078237, 0.00035262, 0.00138830, -0.00077055, -0.00025026, 0.00047858, -0.00035120, 0.00015381, 0.00001038, -0.00017214, 0.00005570, -0.00005019, -0.00000276, 0.00000428, -0.00001025, 0.00000428, -0.00000276, -0.00005019, 0.00005570, -0.00017214, 0.00001038, 0.00015381, -0.00035120, 0.00047858, -0.00025026, -0.00077055, 0.00138830, 0.00035262, -0.00078237, 0.00002517, },
		{ -0.00010334, -0.00003699, 0.00001869, -0.00053425, -0.00017992, 0.00140298, -0.00029831, -0.00060539, 0.00054744, -0.00023520, -0.00000995, 0.00011896, -0.00017055, 0.00003953, -0.00003202, -0.00000271, 0.00000084, -0.00001756, 0.00001406, 0.00000295, -0.00006735, 0.00005559, -0.00013268, -0.00012652, 0.00026824, -0.00034233, 0.00027104, 0.00014364, -0.00106654, 0.00106946, 0.00098352, -0.00088901, },
		{ -0.00074920, -0.00037543, 0.00006897, -0.00005888, -0.00026990, -0.00050544, 0.00114200, 0.00022312, -0.00081851, 0.00045122, -0.00003122, -0.00016674, 0.00016815, -0.00012983, 0.00001686, -0.00001602, -0.00000256, 0.00001019, -0.00003563, 0.00002026, 0.00001727, -0.00007778, 0.00003434, -0.00006418, -0.00025114, 0.00029216, -0.00021274, -0.00000526, 0.00046262, -0.00110836, 0.00050384, 0.00155413, },
		{ 0.00190161, -0.00032072, -0.00072993, 0.00015945, -0.00012245, -0.00008989, -0.00059208, 0.00070631, 0.00065659, -0.00082933, 0.00021578, 0.00019499, -0.00026351, 0.00014645, -0.00006495, -0.00000163, -0.00000207, -0.00000779, 0.00003238, -0.00005430, 0.00001375, 0.00003829, -0.00007541, -0.00000558, 0.00001143, -0.00032652, 0.00021307, -0.00000630, -0.00025987, 0.00061832, -0.00089120, -0.00017859, },
		{ -0.00081152, 0.00190921, 0.00035104, -0.00104783, 0.00014475, -0.00010599, -0.00003819, -0.00048527, 0.00023419, 0.00089647, -0.00063919, -0.00008923, 0.00037051, -0.00026778, 0.00006558, 0.00000123, -0.00000849, 0.00001266, -0.00002072, 0.00006078, -0.00006117, -0.00000966, 0.00005941, -0.00005644, -0.00005443, 0.00006967, -0.00033138, 0.00005126, 0.00020775, -0.00041140, 0.00057504, -0.00048453, },
		{ -0.00000997, -0.00124069, 0.00154491, 0.00113586, -0.00118812, -0.00003697, 0.00000659, -0.00009355, -0.00028301, -0.00014343, 0.00089921, -0.00030977, -0.00037304, 0.00043910, -0.00017797, -0.00004365, 0.00004589, -0.00000230, 0.00003167, -0.00003883, 0.00008432, -0.00004653, -0.00004681, 0.00007128, -0.00002160, -0.00009795, 0.00009178, -0.00026675, -0.00014680, 0.00035758, -0.00041389, 0.00035975, },
		{ 0.00005181, 0.00039678, -0.00136880, 0.00087552, 0.00184992, -0.00103621, -0.00038582, 0.00017161, -0.00018602, -0.00009583, -0.00034269, 0.00069350, 0.00005751, -0.00055328, 0.00037935, -0.00002353, -0.00014169, 0.00005413, 0.00001193, 0.00005685, -0.00005504, 0.00009172, -0.00000765, -0.00008742, 0.00006529, 0.00002306, -0.00012185, 0.00007116, -0.00015536, -0.00032362, 0.00039225, -0.00027128, },
		{ -0.00003527, -0.00024420, 0.00062764, -0.00118315, 0.00005098, 0.00230674, -0.00054873, -0.00082739, 0.00030099, -0.00023149, -0.00000568, -0.00034866, 0.00036719, 0.00035454, -0.00058240, 0.00021113, 0.00014519, -0.00019348, 0.00002409, 0.00002474, 0.00008664, -0.00005996, 0.00007600, 0.00004933, -0.00011708, 0.00003709, 0.00006712, -0.00011656, 0.00001571, -0.00003374, -0.00042924, 0.00029823, },
		{ 0.00010340, 0.00021266, -0.00043308, 0.00063405, -0.00075723, -0.00073679, 0.00237239, 0.00022135, -0.00122732, 0.00029742, -0.00017082, -0.00003834, -0.00021304, 0.00003571, 0.00050097, -0.00046227, -0.00001169, 0.00027381, -0.00018000, -0.00003235, 0.00002605, 0.00011554, -0.00004541, 0.00003783, 0.00011059, -0.00012206, -0.00001094, 0.00009847, -0.00008080, -0.00005472, 0.00006006, -0.00043716, },
		{ -0.00035228, -0.00013256, 0.00038879, -0.00045973, 0.00043901, -0.00022603, -0.00130791, 0.00200750, 0.00112896, -0.00143013, 0.00009676, 0.00000140, -0.00015832, -0.00002951, -0.00019683, 0.00046888, -0.00024239, -0.00021966, 0.00032190, -0.00010448, -0.00009364, 0.00000934, 0.00013528, -0.00000792, -0.00001378, 0.00015871, -0.00009427, -0.00006928, 0.00010679, -0.00002250, -0.00011490, 0.00009910, },
		{ 0.00007626, -0.00020837, -0.00033821, 0.00043512, -0.00032573, 0.00012583, 0.00025653, -0.00154493, 0.00128228, 0.00197122, -0.00131036, -0.00029829, 0.00022957, -0.00028743, 0.00010382, -0.00027107, 0.00028950, -0.00000218, -0.00035111, 0.00027603, 0.00000892, -0.00013596, -0.00002534, 0.00013731, 0.00004900, -0.00006363, 0.00017827, -0.00003472, -0.00012338, 0.00008693, 0.00004313, -0.00014270, },
		{ -0.00012661, 0.00000613, -0.00005604, -0.00045396, 0.00033827, -0.00008659, -0.00019202, 0.00056439, -0.00142381, 0.00035935, 0.00254230, -0.00081912, -0.00080635, 0.00041459, -0.00033960, 0.00012145, -0.00018912, 0.00004027, 0.00017621, -0.00037244, 0.00015360, 0.00012339, -0.00014143, -0.00007063, 0.00011590, 0.00011477, -0.00009491, 0.00016115, 0.00004578, -0.00015756, 0.00004097, 0.00009797, },
		{ 0.00012638, -0.00006966, -0.00008045, 0.00005469, -0.00045143, 0.00013311, 0.00016907, -0.00040966, 0.00063527, -0.00101646, -0.00055076, 0.00269230, -0.00000988, -0.00128355, 0.00045052, -0.00025887, 0.00001389, -0.00000960, -0.00018317, 0.00023566, -0.00028867, -0.00000371, 0.00020194, -0.00010447, -0.00011371, 0.00007070, 0.00017412, -0.00009463, 0.00010973, 0.00012914, -0.00015969, -0.00002184, },
		{ -0.00008715, 0.00011997, 0.00001136, -0.00014740, 0.00009117, -0.00034197, -0.00010997, 0.00035106, -0.00046482, 0.00048475, -0.00046461, -0.00124796, 0.00237185, 0.00096638, -0.00156403, 0.00026753, -0.00004617, -0.00017055, 0.00017670, -0.00030021, 0.00016278, -0.00014069, -0.00014460, 0.00021973, -0.00003394, -0.00014034, 0.00000785, 0.00021121, -0.00005823, 0.00003679, 0.00019496, -0.00012514, },
		{ -0.00005861, -0.00013925, 0.00008009, 0.00009139, -0.00016601, 0.00004879, -0.00017177, -0.00030962, 0.00039895, -0.00035614, 0.00019480, 0.00006597, -0.00159699, 0.00164832, 0.00189695, -0.00151246, -0.00013446, 0.00023525, -0.00034585, 0.00028170, -0.00027266, -0.00000844, 0.00000974, -0.00022580, 0.00017226, 0.00004953, -0.00013964, -0.00006092, 0.00021426, 0.00000837, -0.00003789, 0.00022648, },
		{ 0.00021564, 0.00002665, -0.00016530, 0.00001744, 0.00014526, -0.00012506, -0.00004825, -0.00000522, -0.00040164, 0.00030246, -0.00013982, -0.00011849, 0.00043589, -0.00156128, 0.00068822, 0.00256425, -0.00107167, -0.00067439, 0.00047870, -0.00042295, 0.00025302, -0.00011670, -0.00020930, 0.00010192, -0.00022669, 0.00007659, 0.00012026, -0.00010808, -0.00012099, 0.00017940, 0.00008984, -0.00009338, },
		{ -0.00011383, 0.00016559, 0.00011222, -0.00015873, -0.00005121, 0.00015597, -0.00003476, -0.00015601, 0.00009765, -0.00036083, 0.00010412, 0.00009346, -0.00034621, 0.00057113, -0.00120659, -0.00028900, 0.00280655, -0.00028900, -0.00120659, 0.00057113, -0.00034621, 0.00009346, 0.00010412, -0.00036083, 0.00009765, -0.00015601, -0.00003476, 0.00015597, -0.00005121, -0.00015873, 0.00011222, 0.00016559, },
		{ 0.00021564, -0.00009338, 0.00008984, 0.00017940, -0.00012099, -0.00010808, 0.00012026, 0.00007659, -0.00022669, 0.00010192, -0.00020930, -0.00011670, 0.00025302, -0.00042295, 0.00047870, -0.00067439, -0.00107167, 0.00256425, 0.00068822, -0.00156128, 0.00043589, -0.00011849, -0.00013982, 0.00030246, -0.00040164, -0.00000522, -0.00004825, -0.00012506, 0.00014526, 0.00001744, -0.00016530, 0.00002665, },
		{ -0.00005861, 0.00022648, -0.00003789, 0.00000837, 0.00021426, -0.00006092, -0.00013964, 0.00004953, 0.00017226, -0.00022580, 0.00000974, -0.00000844, -0.00027266, 0.00028170, -0.00034585, 0.00023525, -0.00013446, -0.00151246, 0.00189695, 0.00164832, -0.00159699, 0.00006597, 0.00019480, -0.00035614, 0.00039895, -0.00030962, -0.00017177, 0.00004879, -0.00016601, 0.00009139, 0.00008009, -0.00013925, },
		{ -0.00008715, -0.00012514, 0.00019496, 0.00003679, -0.00005823, 0.00021121, 0.00000785, -0.00014034, -0.00003394, 0.00021973, -0.00014460, -0.00014069, 0.00016278, -0.00030021, 0.00017670, -0.00017055, -0.00004617, 0.00026753, -0.00156403, 0.00096638, 0.00237185, -0.00124796, -0.00046461, 0.00048475, -0.00046482, 0.00035106, -0.00010997, -0.00034197, 0.00009117, -0.00014740, 0.00001136, 0.00011997, },
		{ 0.00012638, -0.00002184, -0.00015969, 0.00012914, 0.00010973, -0.00009463, 0.00017412, 0.00007070, -0.00011371, -0.00010447, 0.00020194, -0.00000371, -0.00028867, 0.00023566, -0.00018317, -0.00000960, 0.00001389, -0.00025887, 0.00045052, -0.00128355, -0.00000988, 0.00269230, -0.00055076, -0.00101646, 0.00063527, -0.00040966, 0.00016907, 0.00013311, -0.00045143, 0.00005469, -0.00008045, -0.00006966, },
		{ -0.00012661, 0.00009797, 0.00004097, -0.00015756, 0.00004578, 0.00016115, -0.00009491, 0.00011477, 0.00011590, -0.00007063, -0.00014143, 0.00012339, 0.00015360, -0.00037244, 0.00017621, 0.00004027, -0.00018912, 0.00012145, -0.00033960, 0.00041459, -0.00080635, -0.00081912, 0.00254230, 0.00035935, -0.00142381, 0.00056439, -0.00019202, -0.00008659, 0.00033827, -0.00045396, -0.00005604, 0.00000613, },
		{ 0.00007626, -0.00014270, 0.00004313, 0.00008693, -0.00012338, -0.00003472, 0.00017827, -0.00006363, 0.00004900, 0.00013731, -0.00002534, -0.00013596, 0.00000892, 0.00027603, -0.00035111, -0.00000218, 0.00028950, -0.00027107, 0.00010382, -0.00028743, 0.00022957, -0.00029829, -0.00131036, 0.00197122, 0.00128228, -0.00154493, 0.00025653, 0.00012583, -0.00032573, 0.00043512, -0.00033821, -0.00020837, },
		{ -0.00035228, 0.00009910, -0.00011490, -0.00002250, 0.00010679, -0.00006928, -0.00009427, 0.00015871, -0.00001378, -0.00000792, 0.00013528, 0.00000934, -0.00009364, -0.00010448, 0.00032190, -0.00021966, -0.00024239, 0.00046888, -0.00019683, -0.00002951, -0.00015832, 0.00000140, 0.00009676, -0.00143013, 0.00112896, 0.00200750, -0.00130791, -0.00022603, 0.00043901, -0.00045973, 0.00038879, -0.00013256, },
		{ 0.00010340, -0.00043716, 0.00006006, -0.00005472, -0.00008080, 0.00009847, -0.00001094, -0.00012206, 0.00011059, 0.00003783, -0.00004541, 0.00011554, 0.00002605, -0.00003235, -0.00018000, 0.00027381, -0.00001169, -0.00046227, 0.00050097, 0.00003571, -0.00021304, -0.00003834, -0.00017082, 0.00029742, -0.00122732, 0.00022135, 0.00237239, -0.00073679, -0.00075723, 0.00063405, -0.00043308, 0.00021266, },
		{ -0.00003527, 0.00029823, -0.00042924, -0.00003374, 0.00001571, -0.00011656, 0.00006712, 0.00003709, -0.00011708, 0.00004933, 0.00007600, -0.00005996, 0.00008664, 0.00002474, 0.00002409, -0.00019348, 0.00014519, 0.00021113, -0.00058240, 0.00035454, 0.00036719, -0.00034866, -0.00000568, -0.00023149, 0.00030099, -0.00082739, -0.00054873, 0.00230674, 0.00005098, -0.00118315, 0.00062764, -0.00024420, },
		{ 0.00005181, -0.00027128, 0.00039225, -0.00032362, -0.00015536, 0.00007116, -0.00012185, 0.00002306, 0.00006529, -0.00008742, -0.00000765, 0.00009172, -0.00005504, 0.00005685, 0.00001193, 0.00005413, -0.00014169, -0.00002353, 0.00037935, -0.00055328, 0.00005751, 0.00069350, -0.00034269, -0.00009583, -0.00018602, 0.00017161, -0.00038582, -0.00103621, 0.00184992, 0.00087552, -0.00136880, 0.00039678, },
		{ -0.00000997, 0.00035975, -0.00041389, 0.00035758, -0.00014680, -0.00026675, 0.00009178, -0.00009795, -0.00002160, 0.00007128, -0.00004681, -0.00004653, 0.00008432, -0.00003883, 0.00003167, -0.00000230, 0.00004589, -0.00004365, -0.00017797, 0.00043910, -0.00037304, -0.00030977, 0.00089921, -0.00014343, -0.00028301, -0.00009355, 0.00000659, -0.00003697, -0.00118812, 0.00113586, 0.00154491, -0.00124069, },
		{ -0.00081152, -0.00048453, 0.00057504, -0.00041140, 0.00020775, 0.00005126, -0.00033138, 0.00006967, -0.00005443, -0.00005644, 0.00005941, -0.00000966, -0.00006117, 0.00006078, -0.00002072, 0.00001266, -0.00000849, 0.00000123, 0.00006558, -0.00026778, 0.00037051, -0.00008923, -0.00063919, 0.00089647, 0.00023419, -0.00048527, -0.00003819, -0.00010599, 0.00014475, -0.00104783, 0.00035104, 0.00190921, },
		{ 0.00190161, -0.00017859, -0.00089120, 0.00061832, -0.00025987, -0.00000630, 0.00021307, -0.00032652, 0.00001143, -0.00000558, -0.00007541, 0.00003829, 0.00001375, -0.00005430, 0.00003238, -0.00000779, -0.00000207, -0.00000163, -0.00006495, 0.00014645, -0.00026351, 0.00019499, 0.00021578, -0.00082933, 0.00065659, 0.00070631, -0.00059208, -0.00008989, -0.00012245, 0.00015945, -0.00072993, -0.00032072, },
		{ -0.00074920, 0.00155413, 0.00050384, -0.00110836, 0.00046262, -0.00000526, -0.00021274, 0.00029216, -0.00025114, -0.00006418, 0.00003434, -0.00007778, 0.00001727, 0.00002026, -0.00003563, 0.00001019, -0.00000256, -0.00001602, 0.00001686, -0.00012983, 0.00016815, -0.00016674, -0.00003122, 0.00045122, -0.00081851, 0.00022312, 0.00114200, -0.00050544, -0.00026990, -0.00005888, 0.00006897, -0.00037543, },
		{ -0.00010334, -0.00088901, 0.00098352, 0.00106946, -0.00106654, 0.00014364, 0.00027104, -0.00034233, 0.00026824, -0.00012652, -0.00013268, 0.00005559, -0.00006735, 0.00000295, 0.00001406, -0.00001756, 0.00000084, -0.00000271, -0.00003202, 0.00003953, -0.00017055, 0.00011896, -0.00000995, -0.00023520, 0.00054744, -0.00060539, -0.00029831, 0.00140298, -0.00017992, -0.00053425, 0.00001869, -0.00003699, },
		},
		{ { -0.00006779, -0.00010094, 0.00018362, -0.00033108, -0.00058526, 0.00099560, 0.00058189, -0.00078163, 0.00015951, 0.00020827, -0.00046693, 0.00016471, 0.00018785, -0.00016740, 0.00005748, 0.00003405, -0.00012534, 0.00003405, 0.00005748, -0.00016740, 0.00018785, 0.00016471, -0.00046693, 0.00020827, 0.00015951, -0.00078163, 0.00058189, 0.00099560, -0.00058526, -0.00033108, 0.00018362, -0.00010094, },
		{ -0.00003788, -0.00011397, -0.00006432, 0.00007221, -0.00002703, -0.00067159, 0.00039928, 0.00096075, -0.00051851, -0.00023033, 0.00041304, -0.00036342, -0.00009241, 0.00029793, -0.00010476, -0.00003724, 0.00008920, -0.00009762, -0.00004991, 0.00011352, -0.00012937, 0.00000629, 0.00033552, -0.00035868, -0.00012872, 0.00044696, -0.00070965, -0.00004636, 0.00136933, -0.00012644, -0.00070064, 0.00019114, },
		{ -0.00002361, 0.00013211, -0.00020025, -0.00002300, -0.00000474, 0.00009668, -0.00048161, -0.00014279, 0.00096248, -0.00005069, -0.00052302, 0.00038976, -0.00010478, -0.00031217, 0.00027288, 0.00003234, -0.00011435, 0.00008724, -0.00002961, -0.00011848, 0.00009767, -0.00002020, -0.00014970, 0.00034252, -0.00008968, -0.00042850, 0.00048100, -0.00033818, -0.00066119, 0.00130476, 0.00061166, -0.00090826, },
		{ -0.00073363, -0.00045124, 0.00029694, -0.00021195, -0.00006810, 0.00001109, 0.00007160, -0.00021184, -0.00043415, 0.00064470, 0.00039956, -0.00057713, 0.00016400, 0.00017655, -0.00038950, 0.00011784, 0.00017447, -0.00012579, 0.00003214, 0.00004196, -0.00013660, 0.00001731, 0.00009315, -0.00020237, 0.00019413, 0.00020765, -0.00053746, 0.00024179, 0.00014489, -0.00099239, 0.00077172, 0.00135698, },
		{ 0.00177544, -0.00010547, -0.00091390, 0.00028417, -0.00006906, -0.00020328, 0.00006893, 0.00001575, -0.00003153, -0.00044079, 0.00020141, 0.00063671, -0.00038087, -0.00013657, 0.00034396, -0.00029230, -0.00009481, 0.00024587, -0.00005926, -0.00004174, 0.00007986, -0.00009750, -0.00008522, 0.00014430, -0.00013398, -0.00002252, 0.00038960, -0.00039981, -0.00014691, 0.00049581, -0.00090051, -0.00003946, },
		{ -0.00079820, 0.00163833, 0.00081941, -0.00113110, -0.00001918, 0.00019024, -0.00033002, 0.00006360, 0.00002925, 0.00000645, -0.00027558, -0.00015192, 0.00058913, -0.00005055, -0.00035569, 0.00032775, -0.00007594, -0.00026327, 0.00020549, 0.00005377, -0.00008903, 0.00006755, -0.00002511, -0.00015736, 0.00010372, 0.00000654, -0.00019083, 0.00037674, -0.00008671, -0.00048344, 0.00054246, -0.00045017, },
		{ 0.00012260, -0.00119155, 0.00095077, 0.00169631, -0.00086736, -0.00056023, 0.00040330, -0.00031516, -0.00006900, 0.00011743, -0.00003119, -0.00010474, -0.00029131, 0.00033285, 0.00024013, -0.00039377, 0.00015695, 0.00014922, -0.00030952, 0.00006957, 0.00015449, -0.00007934, 0.00001809, 0.00004094, -0.00016561, -0.00000698, 0.00013348, -0.00022836, 0.00019316, 0.00024338, -0.00059582, 0.00027357, },
		{ -0.00015879, 0.00053257, -0.00107793, -0.00003237, 0.00214328, -0.00008288, -0.00110203, 0.00037010, -0.00009724, -0.00028507, 0.00019104, -0.00003255, -0.00003880, -0.00023328, 0.00003406, 0.00035746, -0.00025706, -0.00006405, 0.00027793, -0.00021861, -0.00009678, 0.00018858, -0.00001494, -0.00003428, 0.00006877, -0.00011246, -0.00012964, 0.00017316, -0.00012923, -0.00005114, 0.00043051, -0.00043201, },
		{ -0.00008362, -0.00052390, 0.00059144, -0.00055481, -0.00091809, 0.00192926, 0.00100267, -0.00132053, -0.00001169, 0.00024160, -0.00044533, 0.00014025, 0.00006291, -0.00007445, -0.00009837, -0.00015210, 0.00027603, -0.00004493, -0.00021837, 0.00026392, -0.00004326, -0.00021411, 0.00013592, 0.00007006, -0.00005485, 0.00005199, -0.00003204, -0.00020359, 0.00010360, 0.00003718, -0.00022381, 0.00039730, },
		{ 0.00018512, 0.00026830, -0.00063644, 0.00029962, 0.00009570, -0.00136043, 0.00110028, 0.00198118, -0.00097516, -0.00064667, 0.00049497, -0.00040436, -0.00007046, 0.00021203, -0.00011801, -0.00001865, -0.00016574, 0.00008120, 0.00011428, -0.00024181, 0.00013818, 0.00012761, -0.00023006, 0.00002150, 0.00012870, -0.00002923, 0.00001159, 0.00003209, -0.00019881, -0.00003648, 0.00016903, -0.00024273, },
		{ -0.00011721, -0.00007592, 0.00045333, -0.00045161, -0.00016406, 0.00055372, -0.00122391, -0.00002756, 0.00243678, -0.00006166, -0.00124746, 0.00044376, -0.00012147, -0.00035422, 0.00029477, -0.00006899, -0.00004983, -0.00006225, -0.00009178, 0.00013937, -0.00015305, -0.00001628, 0.00021748, -0.00014493, -0.00009824, 0.00012754, 0.00002800, -0.00001868, 0.00005287, -0.00012958, -0.00017307, 0.00019045, },
		{ 0.00009211, 0.00006392, -0.00024376, 0.00040102, -0.00008053, -0.00054581, 0.00062213, -0.00064059, -0.00101011, 0.00214772, 0.00114342, -0.00145951, 0.00000045, 0.00028196, -0.00053917, 0.00020240, 0.00009505, -0.00014192, 0.00003895, -0.00014266, 0.00003960, -0.00003381, -0.00011635, 0.00020051, -0.00000803, -0.00016583, 0.00006548, 0.00008194, -0.00001436, 0.00003413, -0.00004171, -0.00024032, },
		{ -0.00021634, -0.00006810, 0.00018810, -0.00024292, 0.00017048, 0.00027890, -0.00065415, 0.00031603, 0.00006744, -0.00148196, 0.00120352, 0.00218343, -0.00104766, -0.00070144, 0.00056567, -0.00047453, -0.00007243, 0.00029008, -0.00018405, 0.00004093, -0.00006349, -0.00009866, 0.00002685, -0.00012601, 0.00010885, 0.00011170, -0.00015281, -0.00002571, 0.00009817, 0.00002396, 0.00000500, 0.00001745, },
		{ 0.00002754, -0.00013100, -0.00020285, 0.00018490, -0.00010057, -0.00009301, 0.00045358, -0.00045533, -0.00016268, 0.00055642, -0.00132252, -0.00002717, 0.00262607, -0.00004434, -0.00133670, 0.00050098, -0.00014090, -0.00040687, 0.00037475, -0.00009132, -0.00006462, 0.00006463, -0.00017129, -0.00001107, -0.00007163, 0.00000707, 0.00016343, -0.00007290, -0.00009914, 0.00006349, 0.00007068, -0.00000134, },
		{ 0.00002901, 0.00000490, -0.00004279, -0.00024751, 0.00006352, 0.00007826, -0.00024597, 0.00038523, -0.00007756, -0.00054591, 0.00062972, -0.00069760, -0.00106547, 0.00227024, 0.00122783, -0.00153568, 0.00001802, 0.00030839, -0.00060656, 0.00024692, 0.00012444, -0.00019316, 0.00013108, -0.00012418, -0.00011546, -0.00001826, -0.00004926, 0.00013898, 0.00002831, -0.00011868, -0.00000569, 0.00009120, },
		{ 0.00006422, 0.00008005, -0.00001177, -0.00000030, -0.00019560, -0.00009843, 0.00017878, -0.00022681, 0.00014994, 0.00027234, -0.00064492, 0.00031962, 0.00004079, -0.00154289, 0.00124813, 0.00228295, -0.00107857, -0.00071888, 0.00061117, -0.00052224, -0.00007483, 0.00034831, -0.00022674, 0.00007246, 0.00001274, -0.00020548, -0.00002335, -0.00004559, 0.00007136, 0.00010028, -0.00007817, -0.00007222, },
		{ -0.00009944, -0.00000335, 0.00011508, 0.00000988, -0.00001241, -0.00009728, -0.00020596, 0.00014572, -0.00008218, -0.00009896, 0.00042791, -0.00044077, -0.00015483, 0.00053895, -0.00136205, -0.00003262, 0.00269146, -0.00003262, -0.00136205, 0.00053895, -0.00015483, -0.00044077, 0.00042791, -0.00009896, -0.00008218, 0.00014572, -0.00020596, -0.00009728, -0.00001241, 0.00000988, 0.00011508, -0.00000335, },
		{ 0.00006422, -0.00007222, -0.00007817, 0.00010028, 0.00007136, -0.00004559, -0.00002335, -0.00020548, 0.00001274, 0.00007246, -0.00022674, 0.00034831, -0.00007483, -0.00052224, 0.00061117, -0.00071888, -0.00107857, 0.00228295, 0.00124813, -0.00154289, 0.00004079, 0.00031962, -0.00064492, 0.00027234, 0.00014994, -0.00022681, 0.00017878, -0.00009843, -0.00019560, -0.00000030, -0.00001177, 0.00008005, },
		{ 0.00002901, 0.00009120, -0.00000569, -0.00011868, 0.00002831, 0.00013898, -0.00004926, -0.00001826, -0.00011546, -0.00012418, 0.00013108, -0.00019316, 0.00012444, 0.00024692, -0.00060656, 0.00030839, 0.00001802, -0.00153568, 0.00122783, 0.00227024, -0.00106547, -0.00069760, 0.00062972, -0.00054591, -0.00007756, 0.00038523, -0.00024597, 0.00007826, 0.00006352, -0.00024751, -0.00004279, 0.00000490, },
		{ 0.00002754, -0.00000134, 0.00007068, 0.00006349, -0.00009914, -0.00007290, 0.00016343, 0.00000707, -0.00007163, -0.00001107, -0.00017129, 0.00006463, -0.00006462, -0.00009132, 0.00037475, -0.00040687, -0.00014090, 0.00050098, -0.00133670, -0.00004434, 0.00262607, -0.00002717, -0.00132252, 0.00055642, -0.00016268, -0.00045533, 0.00045358, -0.00009301, -0.00010057, 0.00018490, -0.00020285, -0.00013100, },
		{ -0.00021634, 0.00001745, 0.00000500, 0.00002396, 0.00009817, -0.00002571, -0.00015281, 0.00011170, 0.00010885, -0.00012601, 0.00002685, -0.00009866, -0.00006349, 0.00004093, -0.00018405, 0.00029008, -0.00007243, -0.00047453, 0.00056567, -0.00070144, -0.00104766, 0.00218343, 0.00120352, -0.00148196, 0.00006744, 0.00031603, -0.00065415, 0.00027890, 0.00017048, -0.00024292, 0.00018810, -0.00006810, },
		{ 0.00009211, -0.00024032, -0.00004171, 0.00003413, -0.00001436, 0.00008194, 0.00006548, -0.00016583, -0.00000803, 0.00020051, -0.00011635, -0.00003381, 0.00003960, -0.00014266, 0.00003895, -0.00014192, 0.00009505, 0.00020240, -0.00053917, 0.00028196, 0.00000045, -0.00145951, 0.00114342, 0.00214772, -0.00101011, -0.00064059, 0.00062213, -0.00054581, -0.00008053, 0.00040102, -0.00024376, 0.00006392, },
		{ -0.00011721, 0.00019045, -0.00017307, -0.00012958, 0.00005287, -0.00001868, 0.00002800, 0.00012754, -0.00009824, -0.00014493, 0.00021748, -0.00001628, -0.00015305, 0.00013937, -0.00009178, -0.00006225, -0.00004983, -0.00006899, 0.00029477, -0.00035422, -0.00012147, 0.00044376, -0.00124746, -0.00006166, 0.00243678, -0.00002756, -0.00122391, 0.00055372, -0.00016406, -0.00045161, 0.00045333, -0.00007592, },
		{ 0.00018512, -0.00024273, 0.00016903, -0.00003648, -0.00019881, 0.00003209, 0.00001159, -0.00002923, 0.00012870, 0.00002150, -0.00023006, 0.00012761, 0.00013818, -0.00024181, 0.00011428, 0.00008120, -0.00016574, -0.00001865, -0.00011801, 0.00021203, -0.00007046, -0.00040436, 0.00049497, -0.00064667, -0.00097516, 0.00198118, 0.00110028, -0.00136043, 0.00009570, 0.00029962, -0.00063644, 0.00026830, },
		{ -0.00008362, 0.00039730, -0.00022381, 0.00003718, 0.00010360, -0.00020359, -0.00003204, 0.00005199, -0.00005485, 0.00007006, 0.00013592, -0.00021411, -0.00004326, 0.00026392, -0.00021837, -0.00004493, 0.00027603, -0.00015210, -0.00009837, -0.00007445, 0.00006291, 0.00014025, -0.00044533, 0.00024160, -0.00001169, -0.00132053, 0.00100267, 0.00192926, -0.00091809, -0.00055481, 0.00059144, -0.00052390, },
		{ -0.00015879, -0.00043201, 0.00043051, -0.00005114, -0.00012923, 0.00017316, -0.00012964, -0.00011246, 0.00006877, -0.00003428, -0.00001494, 0.00018858, -0.00009678, -0.00021861, 0.00027793, -0.00006405, -0.00025706, 0.00035746, 0.00003406, -0.00023328, -0.00003880, -0.00003255, 0.00019104, -0.00028507, -0.00009724, 0.00037010, -0.00110203, -0.00008288, 0.00214328, -0.00003237, -0.00107793, 0.00053257, },
		{ 0.00012260, 0.00027357, -0.00059582, 0.00024338, 0.00019316, -0.00022836, 0.00013348, -0.00000698, -0.00016561, 0.00004094, 0.00001809, -0.00007934, 0.00015449, 0.00006957, -0.00030952, 0.00014922, 0.00015695, -0.00039377, 0.00024013, 0.00033285, -0.00029131, -0.00010474, -0.00003119, 0.00011743, -0.00006900, -0.00031516, 0.00040330, -0.00056023, -0.00086736, 0.00169631, 0.00095077, -0.00119155, },
		{ -0.00079820, -0.00045017, 0.00054246, -0.00048344, -0.00008671, 0.00037674, -0.00019083, 0.00000654, 0.00010372, -0.00015736, -0.00002511, 0.00006755, -0.00008903, 0.00005377, 0.00020549, -0.00026327, -0.00007594, 0.00032775, -0.00035569, -0.00005055, 0.00058913, -0.00015192, -0.00027558, 0.00000645, 0.00002925, 0.00006360, -0.00033002, 0.00019024, -0.00001918, -0.00113110, 0.00081941, 0.00163833, },
		{ 0.00177544, -0.00003946, -0.00090051, 0.00049581, -0.00014691, -0.00039981, 0.00038960, -0.00002252, -0.00013398, 0.00014430, -0.00008522, -0.00009750, 0.00007986, -0.00004174, -0.00005926, 0.00024587, -0.00009481, -0.00029230, 0.00034396, -0.00013657, -0.00038087, 0.00063671, 0.00020141, -0.00044079, -0.00003153, 0.00001575, 0.00006893, -0.00020328, -0.00006906, 0.00028417, -0.00091390, -0.00010547, },
		{ -0.00073363, 0.00135698, 0.00077172, -0.00099239, 0.00014489, 0.00024179, -0.00053746, 0.00020765, 0.00019413, -0.00020237, 0.00009315, 0.00001731, -0.00013660, 0.00004196, 0.00003214, -0.00012579, 0.00017447, 0.00011784, -0.00038950, 0.00017655, 0.00016400, -0.00057713, 0.00039956, 0.00064470, -0.00043415, -0.00021184, 0.00007160, 0.00001109, -0.00006810, -0.00021195, 0.00029694, -0.00045124, },
		{ -0.00002361, -0.00090826, 0.00061166, 0.00130476, -0.00066119, -0.00033818, 0.00048100, -0.00042850, -0.00008968, 0.00034252, -0.00014970, -0.00002020, 0.00009767, -0.00011848, -0.00002961, 0.00008724, -0.00011435, 0.00003234, 0.00027288, -0.00031217, -0.00010478, 0.00038976, -0.00052302, -0.00005069, 0.00096248, -0.00014279, -0.00048161, 0.00009668, -0.00000474, -0.00002300, -0.00020025, 0.00013211, },
		{ -0.00003788, 0.00019114, -0.00070064, -0.00012644, 0.00136933, -0.00004636, -0.00070965, 0.00044696, -0.00012872, -0.00035868, 0.00033552, 0.00000629, -0.00012937, 0.00011352, -0.00004991, -0.00009762, 0.00008920, -0.00003724, -0.00010476, 0.00029793, -0.00009241, -0.00036342, 0.00041304, -0.00023033, -0.00051851, 0.00096075, 0.00039928, -0.00067159, -0.00002703, 0.00007221, -0.00006432, -0.00011397, },
		},
		{ { -0.00001025, 0.00006273, -0.00006745, -0.00006229, 0.00004179, -0.00055277, -0.00029526, 0.00095862, 0.00046881, -0.00063281, -0.00002660, 0.00040316, -0.00015149, -0.00018421, 0.00005484, 0.00002916, 0.00000414, 0.00002916, 0.00005484, -0.00018421, -0.00015149, 0.00040316, -0.00002660, -0.00063281, 0.00046881, 0.00095862, -0.00029526, -0.00055277, 0.00004179, -0.00006229, -0.00006745, 0.00006273, },
		{ 0.00004553, 0.00004886, -0.00001104, 0.00002410, -0.00009496, 0.00003093, -0.00020768, -0.00054966, 0.00036917, 0.00082930, -0.00025004, -0.00041165, 0.00032278, 0.00012667, -0.00021584, -0.00004773, 0.00005646, 0.00003231, -0.00001798, 0.00006767, -0.00002500, -0.00031912, 0.00017227, 0.00039173, -0.00058894, -0.00020977, 0.00117027, 0.00039262, -0.00076102, -0.00020107, 0.00007918, -0.00016203, },
		{ -0.00009959, -0.00010375, 0.00013615, -0.00002325, -0.00000128, -0.00001793, -0.00005162, -0.00001993, -0.00041104, -0.00016219, 0.00070137, 0.00025704, -0.00048321, 0.00000787, 0.00029327, -0.00007737, -0.00014020, 0.00000859, 0.00007780, -0.00000914, -0.00000879, 0.00010962, -0.00025017, -0.00019561, 0.00051897, -0.00012969, -0.00076189, 0.00075942, 0.00117523, -0.00048661, -0.00065185, 0.00012609, },
		{ -0.00014347, 0.00010922, -0.00023661, 0.00008899, 0.00009027, -0.00008084, 0.00002611, -0.00004309, -0.00002503, -0.00016526, -0.00035420, 0.00027629, 0.00053024, -0.00022403, -0.00028520, 0.00022036, 0.00011946, -0.00011942, -0.00008236, 0.00006687, 0.00006475, -0.00007791, 0.00010063, -0.00002017, -0.00041666, 0.00025231, 0.00040859, -0.00079885, -0.00008767, 0.00152292, 0.00033431, -0.00096422, },
		{ -0.00071997, -0.00069258, 0.00022069, -0.00016245, -0.00012071, 0.00020922, -0.00006027, -0.00004762, 0.00003700, -0.00006483, -0.00005070, -0.00024677, -0.00009424, 0.00043536, 0.00012777, -0.00033968, -0.00001866, 0.00020541, 0.00001488, -0.00012019, -0.00002265, 0.00012834, -0.00004697, -0.00002877, 0.00016491, -0.00033238, -0.00020201, 0.00062158, -0.00029913, -0.00084350, 0.00111322, 0.00132204, },
		{ 0.00185216, 0.00019261, -0.00113405, -0.00003994, 0.00011272, -0.00030963, 0.00014975, 0.00010740, -0.00015049, 0.00004948, -0.00001165, -0.00007782, -0.00008240, -0.00019424, 0.00013477, 0.00029465, -0.00015430, -0.00021834, 0.00010578, 0.00014807, -0.00004076, -0.00012135, 0.00009542, 0.00007390, -0.00013780, 0.00015210, -0.00003664, -0.00049600, 0.00035674, 0.00036064, -0.00101608, 0.00012164, },
		{ -0.00085473, 0.00150385, 0.00137688, -0.00097158, -0.00066608, 0.00031135, -0.00024604, -0.00011495, 0.00027209, -0.00011561, -0.00007073, 0.00008894, -0.00009600, -0.00004720, -0.00008144, -0.00008822, 0.00019058, 0.00007319, -0.00022425, -0.00009040, 0.00015026, 0.00011108, -0.00012253, -0.00003494, 0.00017178, -0.00010365, -0.00002624, 0.00021229, -0.00042226, -0.00016834, 0.00069188, -0.00052265, },
		{ 0.00024335, -0.00121186, 0.00041161, 0.00212264, -0.00002278, -0.00124779, 0.00009800, 0.00008485, -0.00037191, 0.00022038, 0.00009811, -0.00020993, 0.00008907, -0.00000269, -0.00011791, 0.00002705, -0.00008229, -0.00002953, 0.00013307, -0.00006623, -0.00020938, -0.00000037, 0.00020387, 0.00001243, -0.00015683, 0.00013594, 0.00005791, -0.00018798, 0.00021584, -0.00007464, -0.00054654, 0.00047088, },
		{ -0.00009801, 0.00071373, -0.00077926, -0.00078058, 0.00189715, 0.00132797, -0.00121337, -0.00057174, 0.00038393, -0.00034224, -0.00008632, 0.00031630, -0.00018078, -0.00006855, 0.00013025, -0.00013703, -0.00001169, 0.00006517, -0.00013201, -0.00000898, 0.00007627, -0.00015105, -0.00018618, 0.00013159, 0.00019691, -0.00014223, -0.00002742, 0.00020042, -0.00017060, -0.00000014, 0.00024429, -0.00050947, },
		{ -0.00013104, -0.00056121, 0.00057762, 0.00006152, -0.00135817, 0.00076381, 0.00230355, -0.00029184, -0.00128858, 0.00025403, 0.00002526, -0.00041548, 0.00029185, 0.00006418, -0.00025065, 0.00013746, -0.00001490, -0.00013858, 0.00014616, -0.00002237, -0.00018980, 0.00004441, 0.00001639, -0.00024692, -0.00008149, 0.00027369, 0.00003668, -0.00018188, 0.00018054, 0.00001882, -0.00022027, 0.00028351, },
		{ 0.00004693, 0.00025552, -0.00058335, 0.00000027, 0.00067673, -0.00104202, -0.00061675, 0.00225547, 0.00117579, -0.00141684, -0.00041748, 0.00042664, -0.00044077, -0.00003797, 0.00033600, -0.00024575, -0.00004277, 0.00015516, -0.00017905, 0.00004844, 0.00017784, -0.00020681, -0.00014852, 0.00011433, -0.00012428, -0.00028311, 0.00014580, 0.00026005, -0.00017194, -0.00000246, 0.00020900, -0.00023776, },
		{ -0.00003770, -0.00022929, 0.00034592, -0.00019978, -0.00053742, 0.00066002, -0.00017091, -0.00143196, 0.00114968, 0.00237294, -0.00058678, -0.00124789, 0.00040930, -0.00006175, -0.00043477, 0.00035481, 0.00001107, -0.00026693, 0.00018621, -0.00004392, -0.00013704, 0.00025750, -0.00000983, -0.00032244, 0.00001708, 0.00007539, -0.00031206, -0.00012752, 0.00034200, 0.00003307, -0.00019162, 0.00022091, },
		{ 0.00003483, 0.00019559, -0.00029501, 0.00010901, 0.00024349, -0.00063448, 0.00011370, 0.00057799, -0.00128163, -0.00037094, 0.00254247, 0.00093333, -0.00155707, -0.00021883, 0.00043180, -0.00053054, 0.00002406, 0.00032883, -0.00030051, 0.00000146, 0.00016067, -0.00021385, 0.00012187, 0.00024794, -0.00029056, -0.00022407, 0.00016337, -0.00013783, -0.00036038, 0.00018277, 0.00029205, -0.00020323, },
		{ 0.00000689, -0.00018403, 0.00024961, -0.00010323, -0.00021324, 0.00039434, -0.00027266, -0.00047752, 0.00070388, -0.00043250, -0.00141864, 0.00153401, 0.00232085, -0.00087577, -0.00112699, 0.00054513, -0.00016770, -0.00042738, 0.00040096, -0.00005293, -0.00025669, 0.00022710, -0.00008305, -0.00011478, 0.00034577, -0.00003264, -0.00041061, 0.00003173, 0.00010120, -0.00038181, -0.00013641, 0.00039345, },
		{ 0.00022765, 0.00028957, -0.00022799, 0.00007735, 0.00016187, -0.00033378, 0.00017766, 0.00020903, -0.00065599, 0.00022728, 0.00042292, -0.00147046, -0.00006226, 0.00272800, 0.00062448, -0.00161630, 0.00000318, 0.00039712, -0.00060120, 0.00009174, 0.00029623, -0.00033659, 0.00005641, 0.00014693, -0.00023520, 0.00019515, 0.00027441, -0.00036162, -0.00024240, 0.00020237, -0.00017660, -0.00040263, },
		{ -0.00043308, -0.00011417, 0.00041541, -0.00003334, -0.00016026, 0.00026128, -0.00016790, -0.00017424, 0.00042183, -0.00034049, -0.00038850, 0.00069999, -0.00069702, -0.00131446, 0.00187931, 0.00215077, -0.00112729, -0.00093691, 0.00064564, -0.00028098, -0.00039446, 0.00042426, -0.00011800, -0.00022184, 0.00025348, -0.00012444, -0.00007734, 0.00040006, -0.00007374, -0.00044666, 0.00006508, 0.00009432, },
		{ 0.00021705, -0.00021933, -0.00040247, 0.00026354, 0.00025482, -0.00023985, 0.00011721, 0.00011291, -0.00034832, 0.00024323, 0.00015620, -0.00064463, 0.00032602, 0.00022482, -0.00158648, 0.00028089, 0.00279217, 0.00028089, -0.00158648, 0.00022482, 0.00032602, -0.00064463, 0.00015620, 0.00024323, -0.00034832, 0.00011291, 0.00011721, -0.00023985, 0.00025482, 0.00026354, -0.00040247, -0.00021933, },
		{ -0.00043308, 0.00009432, 0.00006508, -0.00044666, -0.00007374, 0.00040006, -0.00007734, -0.00012444, 0.00025348, -0.00022184, -0.00011800, 0.00042426, -0.00039446, -0.00028098, 0.00064564, -0.00093691, -0.00112729, 0.00215077, 0.00187931, -0.00131446, -0.00069702, 0.00069999, -0.00038850, -0.00034049, 0.00042183, -0.00017424, -0.00016790, 0.00026128, -0.00016026, -0.00003334, 0.00041541, -0.00011417, },
		{ 0.00022765, -0.00040263, -0.00017660, 0.00020237, -0.00024240, -0.00036162, 0.00027441, 0.00019515, -0.00023520, 0.00014693, 0.00005641, -0.00033659, 0.00029623, 0.00009174, -0.00060120, 0.00039712, 0.00000318, -0.00161630, 0.00062448, 0.00272800, -0.00006226, -0.00147046, 0.00042292, 0.00022728, -0.00065599, 0.00020903, 0.00017766, -0.00033378, 0.00016187, 0.00007735, -0.00022799, 0.00028957, },
		{ 0.00000689, 0.00039345, -0.00013641, -0.00038181, 0.00010120, 0.00003173, -0.00041061, -0.00003264, 0.00034577, -0.00011478, -0.00008305, 0.00022710, -0.00025669, -0.00005293, 0.00040096, -0.00042738, -0.00016770, 0.00054513, -0.00112699, -0.00087577, 0.00232085, 0.00153401, -0.00141864, -0.00043250, 0.00070388, -0.00047752, -0.00027266, 0.00039434, -0.00021324, -0.00010323, 0.00024961, -0.00018403, },
		{ 0.00003483, -0.00020323, 0.00029205, 0.00018277, -0.00036038, -0.00013783, 0.00016337, -0.00022407, -0.00029056, 0.00024794, 0.00012187, -0.00021385, 0.00016067, 0.00000146, -0.00030051, 0.00032883, 0.00002406, -0.00053054, 0.00043180, -0.00021883, -0.00155707, 0.00093333, 0.00254247, -0.00037094, -0.00128163, 0.00057799, 0.00011370, -0.00063448, 0.00024349, 0.00010901, -0.00029501, 0.00019559, },
		{ -0.00003770, 0.00022091, -0.00019162, 0.00003307, 0.00034200, -0.00012752, -0.00031206, 0.00007539, 0.00001708, -0.00032244, -0.00000983, 0.00025750, -0.00013704, -0.00004392, 0.00018621, -0.00026693, 0.00001107, 0.00035481, -0.00043477, -0.00006175, 0.00040930, -0.00124789, -0.00058678, 0.00237294, 0.00114968, -0.00143196, -0.00017091, 0.00066002, -0.00053742, -0.00019978, 0.00034592, -0.00022929, },
		{ 0.00004693, -0.00023776, 0.00020900, -0.00000246, -0.00017194, 0.00026005, 0.00014580, -0.00028311, -0.00012428, 0.00011433, -0.00014852, -0.00020681, 0.00017784, 0.00004844, -0.00017905, 0.00015516, -0.00004277, -0.00024575, 0.00033600, -0.00003797, -0.00044077, 0.00042664, -0.00041748, -0.00141684, 0.00117579, 0.00225547, -0.00061675, -0.00104202, 0.00067673, 0.00000027, -0.00058335, 0.00025552, },
		{ -0.00013104, 0.00028351, -0.00022027, 0.00001882, 0.00018054, -0.00018188, 0.00003668, 0.00027369, -0.00008149, -0.00024692, 0.00001639, 0.00004441, -0.00018980, -0.00002237, 0.00014616, -0.00013858, -0.00001490, 0.00013746, -0.00025065, 0.00006418, 0.00029185, -0.00041548, 0.00002526, 0.00025403, -0.00128858, -0.00029184, 0.00230355, 0.00076381, -0.00135817, 0.00006152, 0.00057762, -0.00056121, },
		{ -0.00009801, -0.00050947, 0.00024429, -0.00000014, -0.00017060, 0.00020042, -0.00002742, -0.00014223, 0.00019691, 0.00013159, -0.00018618, -0.00015105, 0.00007627, -0.00000898, -0.00013201, 0.00006517, -0.00001169, -0.00013703, 0.00013025, -0.00006855, -0.00018078, 0.00031630, -0.00008632, -0.00034224, 0.00038393, -0.00057174, -0.00121337, 0.00132797, 0.00189715, -0.00078058, -0.00077926, 0.00071373, },
		{ 0.00024335, 0.00047088, -0.00054654, -0.00007464, 0.00021584, -0.00018798, 0.00005791, 0.00013594, -0.00015683, 0.00001243, 0.00020387, -0.00000037, -0.00020938, -0.00006623, 0.00013307, -0.00002953, -0.00008229, 0.00002705, -0.00011791, -0.00000269, 0.00008907, -0.00020993, 0.00009811, 0.00022038, -0.00037191, 0.00008485, 0.00009800, -0.00124779, -0.00002278, 0.00212264, 0.00041161, -0.00121186, },
		{ -0.00085473, -0.00052265, 0.00069188, -0.00016834, -0.00042226, 0.00021229, -0.00002624, -0.00010365, 0.00017178, -0.00003494, -0.00012253, 0.00011108, 0.00015026, -0.00009040, -0.00022425, 0.00007319, 0.00019058, -0.00008822, -0.00008144, -0.00004720, -0.00009600, 0.00008894, -0.00007073, -0.00011561, 0.00027209, -0.00011495, -0.00024604, 0.00031135, -0.00066608, -0.00097158, 0.00137688, 0.00150385, },
		{ 0.00185216, 0.00012164, -0.00101608, 0.00036064, 0.00035674, -0.00049600, -0.00003664, 0.00015210, -0.00013780, 0.00007390, 0.00009542, -0.00012135, -0.00004076, 0.00014807, 0.00010578, -0.00021834, -0.00015430, 0.00029465, 0.00013477, -0.00019424, -0.00008240, -0.00007782, -0.00001165, 0.00004948, -0.00015049, 0.00010740, 0.00014975, -0.00030963, 0.00011272, -0.00003994, -0.00113405, 0.00019261, },
		{ -0.00071997, 0.00132204, 0.00111322, -0.00084350, -0.00029913, 0.00062158, -0.00020201, -0.00033238, 0.00016491, -0.00002877, -0.00004697, 0.00012834, -0.00002265, -0.00012019, 0.00001488, 0.00020541, -0.00001866, -0.00033968, 0.00012777, 0.00043536, -0.00009424, -0.00024677, -0.00005070, -0.00006483, 0.00003700, -0.00004762, -0.00006027, 0.00020922, -0.00012071, -0.00016245, 0.00022069, -0.00069258, },
		{ -0.00014347, -0.00096422, 0.00033431, 0.00152292, -0.00008767, -0.00079885, 0.00040859, 0.00025231, -0.00041666, -0.00002017, 0.00010063, -0.00007791, 0.00006475, 0.00006687, -0.00008236, -0.00011942, 0.00011946, 0.00022036, -0.00028520, -0.00022403, 0.00053024, 0.00027629, -0.00035420, -0.00016526, -0.00002503, -0.00004309, 0.00002611, -0.00008084, 0.00009027, 0.00008899, -0.00023661, 0.00010922, },
		{ -0.00009959, 0.00012609, -0.00065185, -0.00048661, 0.00117523, 0.00075942, -0.00076189, -0.00012969, 0.00051897, -0.00019561, -0.00025017, 0.00010962, -0.00000879, -0.00000914, 0.00007780, 0.00000859, -0.00014020, -0.00007737, 0.00029327, 0.00000787, -0.00048321, 0.00025704, 0.00070137, -0.00016219, -0.00041104, -0.00001993, -0.00005162, -0.00001793, -0.00000128, -0.00002325, 0.00013615, -0.00010375, },
		{ 0.00004553, -0.00016203, 0.00007918, -0.00020107, -0.00076102, 0.00039262, 0.00117027, -0.00020977, -0.00058894, 0.00039173, 0.00017227, -0.00031912, -0.00002500, 0.00006767, -0.00001798, 0.00003231, 0.00005646, -0.00004773, -0.00021584, 0.00012667, 0.00032278, -0.00041165, -0.00025004, 0.00082930, 0.00036917, -0.00054966, -0.00020768, 0.00003093, -0.00009496, 0.00002410, -0.00001104, 0.00004886, },
		},
		{ { 0.00001852, -0.00010068, -0.00004575, 0.00015077, -0.00000703, -0.00003589, 0.00007795, -0.00034633, -0.00040875, 0.00052106, 0.00071797, -0.00026229, -0.00050548, 0.00015820, 0.00029643, -0.00006326, -0.00022604, -0.00006326, 0.00029643, 0.00015820, -0.00050548, -0.00026229, 0.00071797, 0.00052106, -0.00040875, -0.00034633, 0.00007795, -0.00003589, -0.00000703, 0.00015077, -0.00004575, -0.00010068, },
		{ -0.00011511, -0.00006950, 0.00001525, -0.00011115, 0.00001471, 0.00012744, -0.00002481, 0.00002179, -0.00004611, -0.00041209, -0.00007454, 0.00063399, 0.00027952, -0.00044253, -0.00021077, 0.00026314, 0.00017231, -0.00011532, -0.00024011, 0.00005123, 0.00039915, -0.00013520, -0.00068465, 0.00020672, 0.00101183, 0.00013773, -0.00062672, -0.00013963, 0.00011364, -0.00011523, 0.00005959, 0.00014176, },
		{ 0.00015382, 0.00007683, -0.00016660, -0.00001483, 0.00000052, -0.00011036, 0.00005825, 0.00010171, -0.00001154, 0.00001650, -0.00017347, -0.00031126, 0.00020476, 0.00048117, -0.00008375, -0.00036552, 0.00002136, 0.00024055, 0.00009740, -0.00017736, -0.00024038, 0.00025424, 0.00034851, -0.00055052, -0.00055662, 0.00082780, 0.00096866, -0.00037263, -0.00060736, 0.00009620, 0.00002961, -0.00014936, },
		{ -0.00012801, -0.00009103, 0.00022206, -0.00003839, -0.00016755, 0.00004891, -0.00003501, -0.00010180, 0.00009274, 0.00008739, -0.00000415, -0.00004173, -0.00022146, -0.00011565, 0.00030423, 0.00019579, -0.00023965, -0.00017618, 0.00011856, 0.00019722, 0.00004819, -0.00028233, -0.00013967, 0.00046793, 0.00004848, -0.00088704, -0.00004214, 0.00133655, 0.00053826, -0.00078328, -0.00035171, 0.00022678, },
		{ 0.00001185, 0.00017434, -0.00025983, 0.00005125, 0.00020945, -0.00016220, -0.00010143, 0.00009105, -0.00009307, -0.00007795, 0.00012763, 0.00007515, -0.00002737, -0.00010107, -0.00016313, 0.00005185, 0.00021007, -0.00004754, -0.00019573, -0.00002550, 0.00012440, 0.00019216, -0.00004710, -0.00037644, 0.00010634, 0.00053160, -0.00046112, -0.00089813, 0.00074250, 0.00146536, -0.00014030, -0.00090078, },
		{ -0.00079230, -0.00067014, 0.00029462, -0.00003424, -0.00027222, 0.00021045, 0.00009736, -0.00023363, 0.00001479, 0.00007900, -0.00015824, -0.00002688, 0.00015990, 0.00004456, -0.00007261, -0.00010600, -0.00004495, 0.00009281, 0.00001766, -0.00013386, -0.00006595, 0.00002035, 0.00013216, 0.00019516, -0.00022396, -0.00034863, 0.00042723, 0.00029996, -0.00097450, -0.00043398, 0.00150419, 0.00108822, },
		{ 0.00189741, 0.00030513, -0.00113872, -0.00020876, 0.00035020, -0.00027948, -0.00013604, 0.00029583, -0.00007665, -0.00020575, 0.00013082, -0.00000101, -0.00019945, 0.00005415, 0.00017144, -0.00001404, -0.00010199, -0.00003917, 0.00004242, -0.00000755, -0.00013310, -0.00006411, 0.00001948, 0.00000174, 0.00019507, 0.00012256, -0.00042441, -0.00012028, 0.00064197, -0.00023824, -0.00119528, 0.00044215, },
		{ -0.00088941, 0.00143935, 0.00169604, -0.00058648, -0.00103716, 0.00025175, 0.00015530, -0.00041586, 0.00009055, 0.00024639, -0.00023132, -0.00007686, 0.00018392, -0.00012949, -0.00018410, 0.00014824, 0.00014077, -0.00008420, -0.00007547, 0.00006085, 0.00002264, -0.00017267, -0.00014474, 0.00006259, 0.00000162, 0.00001210, 0.00027687, -0.00007894, -0.00051421, 0.00026669, 0.00055184, -0.00090029, },
		{ 0.00007463, -0.00136087, -0.00003507, 0.00214977, 0.00092149, -0.00122580, -0.00055611, 0.00048580, -0.00017713, -0.00035504, 0.00029063, 0.00007154, -0.00028280, 0.00009991, 0.00013146, -0.00025592, -0.00009801, 0.00022068, 0.00006197, -0.00012685, 0.00001613, 0.00011875, -0.00011451, -0.00028280, -0.00001888, 0.00012495, -0.00007656, 0.00011120, 0.00027669, -0.00036313, -0.00037076, 0.00063097, },
		{ 0.00001333, 0.00073032, -0.00066235, -0.00130385, 0.00111992, 0.00223769, -0.00014643, -0.00136148, 0.00005883, 0.00039088, -0.00046034, -0.00011781, 0.00035354, -0.00014677, -0.00019250, 0.00023722, -0.00002247, -0.00031951, 0.00004084, 0.00023460, -0.00004273, -0.00010299, 0.00013492, 0.00006900, -0.00030036, -0.00024978, 0.00015012, 0.00006364, -0.00011219, 0.00026666, 0.00011104, -0.00058466, },
		{ -0.00020457, -0.00057870, 0.00048380, 0.00040339, -0.00133878, -0.00060025, 0.00213921, 0.00161158, -0.00108920, -0.00097354, 0.00051011, 0.00003794, -0.00053663, 0.00017573, 0.00023936, -0.00029442, 0.00000566, 0.00025590, -0.00022228, -0.00027919, 0.00018450, 0.00017284, -0.00012704, -0.00000231, 0.00020965, -0.00009828, -0.00042345, -0.00007323, 0.00023907, -0.00007967, -0.00002626, 0.00036538, },
		{ 0.00016804, 0.00028977, -0.00053458, -0.00027521, 0.00077860, -0.00030575, -0.00157715, 0.00058703, 0.00258862, 0.00048608, -0.00154323, -0.00027965, 0.00059818, -0.00037168, -0.00036303, 0.00037269, 0.00000521, -0.00028721, 0.00021612, 0.00012743, -0.00038289, -0.00013711, 0.00027470, 0.00005297, -0.00014231, 0.00014077, 0.00017432, -0.00032110, -0.00039454, 0.00015613, 0.00017708, -0.00019201, },
		{ -0.00000646, -0.00016803, 0.00035969, 0.00000956, -0.00068655, 0.00022590, 0.00066538, -0.00112175, -0.00113674, 0.00183987, 0.00224467, -0.00070479, -0.00136933, 0.00037750, 0.00031662, -0.00061746, -0.00003204, 0.00037293, -0.00022339, -0.00012000, 0.00032746, -0.00010669, -0.00042921, 0.00005776, 0.00027110, -0.00007507, -0.00006601, 0.00025636, 0.00001047, -0.00049171, -0.00020252, 0.00030877, },
		{ 0.00007252, 0.00029376, -0.00019719, 0.00001268, 0.00040437, -0.00037336, -0.00052742, 0.00067551, 0.00008983, -0.00164241, -0.00006334, 0.00265815, 0.00121272, -0.00150429, -0.00071117, 0.00070176, -0.00015621, -0.00057686, 0.00028292, 0.00018482, -0.00032093, 0.00012558, 0.00026984, -0.00035076, -0.00033312, 0.00022968, 0.00017349, -0.00015016, 0.00007954, 0.00027213, -0.00023271, -0.00051307, },
		{ -0.00045480, -0.00035450, 0.00029892, 0.00011537, -0.00026582, 0.00026051, 0.00022075, -0.00066365, -0.00008349, 0.00079462, -0.00075416, -0.00153189, 0.00129325, 0.00268958, -0.00011036, -0.00164144, 0.00008576, 0.00058393, -0.00056071, -0.00028448, 0.00042519, -0.00008055, -0.00023204, 0.00032104, 0.00005328, -0.00049913, -0.00012889, 0.00031079, 0.00002522, -0.00013046, 0.00023104, 0.00015342, },
		{ 0.00031144, -0.00007193, -0.00055154, -0.00007443, 0.00036465, -0.00012573, -0.00015074, 0.00042198, -0.00014259, -0.00067964, 0.00044293, 0.00043349, -0.00148517, -0.00070490, 0.00241566, 0.00190057, -0.00121450, -0.00114675, 0.00065004, 0.00013972, -0.00069547, 0.00009671, 0.00033970, -0.00027308, -0.00000355, 0.00035777, -0.00023422, -0.00048669, 0.00010185, 0.00027321, -0.00010534, -0.00001715, },
		{ -0.00015699, 0.00014206, 0.00026808, -0.00031982, -0.00046888, 0.00020476, 0.00024011, -0.00028849, 0.00009765, 0.00037195, -0.00051601, -0.00036876, 0.00076200, -0.00031920, -0.00170772, 0.00059907, 0.00284971, 0.00059907, -0.00170772, -0.00031920, 0.00076200, -0.00036876, -0.00051601, 0.00037195, 0.00009765, -0.00028849, 0.00024011, 0.00020476, -0.00046888, -0.00031982, 0.00026808, 0.00014206, },
		{ 0.00031144, -0.00001715, -0.00010534, 0.00027321, 0.00010185, -0.00048669, -0.00023422, 0.00035777, -0.00000355, -0.00027308, 0.00033970, 0.00009671, -0.00069547, 0.00013972, 0.00065004, -0.00114675, -0.00121450, 0.00190057, 0.00241566, -0.00070490, -0.00148517, 0.00043349, 0.00044293, -0.00067964, -0.00014259, 0.00042198, -0.00015074, -0.00012573, 0.00036465, -0.00007443, -0.00055154, -0.00007193, },
		{ -0.00045480, 0.00015342, 0.00023104, -0.00013046, 0.00002522, 0.00031079, -0.00012889, -0.00049913, 0.00005328, 0.00032104, -0.00023204, -0.00008055, 0.00042519, -0.00028448, -0.00056071, 0.00058393, 0.00008576, -0.00164144, -0.00011036, 0.00268958, 0.00129325, -0.00153189, -0.00075416, 0.00079462, -0.00008349, -0.00066365, 0.00022075, 0.00026051, -0.00026582, 0.00011537, 0.00029892, -0.00035450, },
		{ 0.00007252, -0.00051307, -0.00023271, 0.00027213, 0.00007954, -0.00015016, 0.00017349, 0.00022968, -0.00033312, -0.00035076, 0.00026984, 0.00012558, -0.00032093, 0.00018482, 0.00028292, -0.00057686, -0.00015621, 0.00070176, -0.00071117, -0.00150429, 0.00121272, 0.00265815, -0.00006334, -0.00164241, 0.00008983, 0.00067551, -0.00052742, -0.00037336, 0.00040437, 0.00001268, -0.00019719, 0.00029376, },
		{ -0.00000646, 0.00030877, -0.00020252, -0.00049171, 0.00001047, 0.00025636, -0.00006601, -0.00007507, 0.00027110, 0.00005776, -0.00042921, -0.00010669, 0.00032746, -0.00012000, -0.00022339, 0.00037293, -0.00003204, -0.00061746, 0.00031662, 0.00037750, -0.00136933, -0.00070479, 0.00224467, 0.00183987, -0.00113674, -0.00112175, 0.00066538, 0.00022590, -0.00068655, 0.00000956, 0.00035969, -0.00016803, },
		{ 0.00016804, -0.00019201, 0.00017708, 0.00015613, -0.00039454, -0.00032110, 0.00017432, 0.00014077, -0.00014231, 0.00005297, 0.00027470, -0.00013711, -0.00038289, 0.00012743, 0.00021612, -0.00028721, 0.00000521, 0.00037269, -0.00036303, -0.00037168, 0.00059818, -0.00027965, -0.00154323, 0.00048608, 0.00258862, 0.00058703, -0.00157715, -0.00030575, 0.00077860, -0.00027521, -0.00053458, 0.00028977, },
		{ -0.00020457, 0.00036538, -0.00002626, -0.00007967, 0.00023907, -0.00007323, -0.00042345, -0.00009828, 0.00020965, -0.00000231, -0.00012704, 0.00017284, 0.00018450, -0.00027919, -0.00022228, 0.00025590, 0.00000566, -0.00029442, 0.00023936, 0.00017573, -0.00053663, 0.00003794, 0.00051011, -0.00097354, -0.00108920, 0.00161158, 0.00213921, -0.00060025, -0.00133878, 0.00040339, 0.00048380, -0.00057870, },
		{ 0.00001333, -0.00058466, 0.00011104, 0.00026666, -0.00011219, 0.00006364, 0.00015012, -0.00024978, -0.00030036, 0.00006900, 0.00013492, -0.00010299, -0.00004273, 0.00023460, 0.00004084, -0.00031951, -0.00002247, 0.00023722, -0.00019250, -0.00014677, 0.00035354, -0.00011781, -0.00046034, 0.00039088, 0.00005883, -0.00136148, -0.00014643, 0.00223769, 0.00111992, -0.00130385, -0.00066235, 0.00073032, },
		{ 0.00007463, 0.00063097, -0.00037076, -0.00036313, 0.00027669, 0.00011120, -0.00007656, 0.00012495, -0.00001888, -0.00028280, -0.00011451, 0.00011875, 0.00001613, -0.00012685, 0.00006197, 0.00022068, -0.00009801, -0.00025592, 0.00013146, 0.00009991, -0.00028280, 0.00007154, 0.00029063, -0.00035504, -0.00017713, 0.00048580, -0.00055611, -0.00122580, 0.00092149, 0.00214977, -0.00003507, -0.00136087, },
		{ -0.00088941, -0.00090029, 0.00055184, 0.00026669, -0.00051421, -0.00007894, 0.00027687, 0.00001210, 0.00000162, 0.00006259, -0.00014474, -0.00017267, 0.00002264, 0.00006085, -0.00007547, -0.00008420, 0.00014077, 0.00014824, -0.00018410, -0.00012949, 0.00018392, -0.00007686, -0.00023132, 0.00024639, 0.00009055, -0.00041586, 0.00015530, 0.00025175, -0.00103716, -0.00058648, 0.00169604, 0.00143935, },
		{ 0.00189741, 0.00044215, -0.00119528, -0.00023824, 0.00064197, -0.00012028, -0.00042441, 0.00012256, 0.00019507, 0.00000174, 0.00001948, -0.00006411, -0.00013310, -0.00000755, 0.00004242, -0.00003917, -0.00010199, -0.00001404, 0.00017144, 0.00005415, -0.00019945, -0.00000101, 0.00013082, -0.00020575, -0.00007665, 0.00029583, -0.00013604, -0.00027948, 0.00035020, -0.00020876, -0.00113872, 0.00030513, },
		{ -0.00079230, 0.00108822, 0.00150419, -0.00043398, -0.00097450, 0.00029996, 0.00042723, -0.00034863, -0.00022396, 0.00019516, 0.00013216, 0.00002035, -0.00006595, -0.00013386, 0.00001766, 0.00009281, -0.00004495, -0.00010600, -0.00007261, 0.00004456, 0.00015990, -0.00002688, -0.00015824, 0.00007900, 0.00001479, -0.00023363, 0.00009736, 0.00021045, -0.00027222, -0.00003424, 0.00029462, -0.00067014, },
		{ 0.00001185, -0.00090078, -0.00014030, 0.00146536, 0.00074250, -0.00089813, -0.00046112, 0.00053160, 0.00010634, -0.00037644, -0.00004710, 0.00019216, 0.00012440, -0.00002550, -0.00019573, -0.00004754, 0.00021007, 0.00005185, -0.00016313, -0.00010107, -0.00002737, 0.00007515, 0.00012763, -0.00007795, -0.00009307, 0.00009105, -0.00010143, -0.00016220, 0.00020945, 0.00005125, -0.00025983, 0.00017434, },
		{ -0.00012801, 0.00022678, -0.00035171, -0.00078328, 0.00053826, 0.00133655, -0.00004214, -0.00088704, 0.00004848, 0.00046793, -0.00013967, -0.00028233, 0.00004819, 0.00019722, 0.00011856, -0.00017618, -0.00023965, 0.00019579, 0.00030423, -0.00011565, -0.00022146, -0.00004173, -0.00000415, 0.00008739, 0.00009274, -0.00010180, -0.00003501, 0.00004891, -0.00016755, -0.00003839, 0.00022206, -0.00009103, },
		{ 0.00015382, -0.00014936, 0.00002961, 0.00009620, -0.00060736, -0.00037263, 0.00096866, 0.00082780, -0.00055662, -0.00055052, 0.00034851, 0.00025424, -0.00024038, -0.00017736, 0.00009740, 0.00024055, 0.00002136, -0.00036552, -0.00008375, 0.00048117, 0.00020476, -0.00031126, -0.00017347, 0.00001650, -0.00001154, 0.00010171, 0.00005825, -0.00011036, 0.00000052, -0.00001483, -0.00016660, 0.00007683, },
		{ -0.00011511, 0.00014176, 0.00005959, -0.00011523, 0.00011364, -0.00013963, -0.00062672, 0.00013773, 0.00101183, 0.00020672, -0.00068465, -0.00013520, 0.00039915, 0.00005123, -0.00024011, -0.00011532, 0.00017231, 0.00026314, -0.00021077, -0.00044253, 0.00027952, 0.00063399, -0.00007454, -0.00041209, -0.00004611, 0.00002179, -0.00002481, 0.00012744, 0.00001471, -0.00011115, 0.00001525, -0.00006950, },
		},
		{ { -0.00006779, 0.00004670, 0.00006168, -0.00014842, -0.00010267, 0.00013417, 0.00006966, -0.00004822, 0.00009477, -0.00000423, -0.00039152, -0.00022451, 0.00039579, 0.00041230, -0.00007429, -0.00014622, 0.00000414, -0.00014622, -0.00007429, 0.00041230, 0.00039579, -0.00022451, -0.00039152, -0.00000423, 0.00009477, -0.00004822, 0.00006966, 0.00013417, -0.00010267, -0.00014842, 0.00006168, 0.00004670, },
		{ 0.00007893, 0.00003108, -0.00006616, 0.00006060, 0.00004533, -0.00014631, -0.00007132, 0.00011692, 0.00004061, -0.00001126, 0.00009861, -0.00007303, -0.00034479, -0.00008017, 0.00033202, 0.00018162, -0.00008728, 0.00001211, -0.00004848, -0.00030301, 0.00004955, 0.00065993, 0.00037209, -0.00039538, -0.00039377, 0.00007097, 0.00006774, -0.00007576, 0.00011293, 0.00013810, -0.00014284, -0.00014326, },
		{ -0.00018720, -0.00012823, 0.00009571, 0.00001391, -0.00006133, 0.00007267, 0.00003060, -0.00014026, -0.00005133, 0.00009304, 0.00002897, 0.00002674, 0.00008062, -0.00012381, -0.00026541, 0.00001532, 0.00020119, 0.00000703, -0.00000693, 0.00012349, -0.00019532, -0.00040755, 0.00028193, 0.00087964, 0.00025411, -0.00056516, -0.00034679, 0.00013933, 0.00002006, -0.00008677, 0.00016427, 0.00012377, },
		{ 0.00021553, 0.00008890, -0.00022970, -0.00010202, 0.00011016, -0.00000447, -0.00005349, 0.00008283, 0.00001754, -0.00013372, -0.00004284, 0.00006977, 0.00003458, 0.00005747, 0.00004574, -0.00014835, -0.00017026, 0.00004837, 0.00003375, -0.00008415, 0.00013282, 0.00014995, -0.00040520, -0.00041568, 0.00060041, 0.00102797, 0.00005006, -0.00070516, -0.00025375, 0.00018807, -0.00004218, -0.00007661, },
		{ -0.00011022, -0.00004395, 0.00025780, 0.00003438, -0.00026372, -0.00006492, 0.00012016, -0.00002354, -0.00004293, 0.00009099, 0.00000555, -0.00012965, -0.00004367, 0.00005368, 0.00005400, 0.00007410, 0.00000168, -0.00014316, -0.00007858, 0.00001616, -0.00013485, -0.00008062, 0.00028760, 0.00007064, -0.00063179, -0.00029759, 0.00096534, 0.00106985, -0.00021775, -0.00079013, -0.00012546, 0.00020692, },
		{ 0.00002116, 0.00018981, -0.00017391, 0.00000893, 0.00028280, -0.00003568, -0.00028311, -0.00001893, 0.00012372, -0.00004260, -0.00003007, 0.00009696, -0.00000637, -0.00012998, -0.00004984, 0.00004955, 0.00008113, 0.00007239, -0.00004234, -0.00010995, -0.00000877, -0.00007303, -0.00026989, 0.00001129, 0.00041028, -0.00011436, -0.00082129, -0.00004346, 0.00132557, 0.00098443, -0.00051590, -0.00080223, },
		{ -0.00080471, -0.00073396, 0.00016559, 0.00013588, -0.00022330, 0.00007640, 0.00028421, -0.00011442, -0.00028316, 0.00003242, 0.00011922, -0.00006077, -0.00001540, 0.00010049, -0.00001939, -0.00013526, -0.00005627, 0.00005956, 0.00010828, 0.00005137, -0.00007710, -0.00005535, 0.00002449, -0.00020120, -0.00034307, 0.00016880, 0.00045901, -0.00038358, -0.00092130, 0.00033400, 0.00162608, 0.00076878, },
		{ 0.00181631, 0.00043899, -0.00104404, -0.00058948, 0.00028670, 0.00004975, -0.00025012, 0.00015020, 0.00025875, -0.00019314, -0.00026138, 0.00008446, 0.00010575, -0.00007708, 0.00000052, 0.00010125, -0.00003456, -0.00014457, -0.00005773, 0.00008288, 0.00012744, 0.00001358, -0.00009491, 0.00001014, 0.00001269, -0.00034355, -0.00033708, 0.00035648, 0.00040452, -0.00069717, -0.00088972, 0.00080043, },
		{ -0.00070222, 0.00130422, 0.00185810, 0.00002836, -0.00119929, -0.00038428, 0.00036601, -0.00005902, -0.00024906, 0.00022067, 0.00020680, -0.00026248, -0.00021798, 0.00013195, 0.00008329, -0.00009051, 0.00001708, 0.00009888, -0.00005251, -0.00015566, -0.00004979, 0.00011578, 0.00013172, -0.00003531, -0.00009081, 0.00007385, -0.00004509, -0.00047246, -0.00024849, 0.00053211, 0.00023539, -0.00100294, },
		{ -0.00003944, -0.00124445, -0.00035703, 0.00178361, 0.00173184, -0.00041707, -0.00124653, -0.00014301, 0.00039043, -0.00017724, -0.00021868, 0.00027813, 0.00013242, -0.00031377, -0.00015595, 0.00016982, 0.00005282, -0.00010010, 0.00003364, 0.00009307, -0.00007327, -0.00016532, -0.00002969, 0.00015220, 0.00011646, -0.00008738, -0.00006347, 0.00012300, -0.00014195, -0.00056190, -0.00008845, 0.00065358, },
		{ 0.00011897, 0.00068596, -0.00039139, -0.00137002, 0.00012361, 0.00217548, 0.00143993, -0.00084483, -0.00117622, 0.00010401, 0.00035435, -0.00029004, -0.00016164, 0.00031431, 0.00004296, -0.00034023, -0.00008075, 0.00019383, 0.00001620, -0.00010509, 0.00004956, 0.00008361, -0.00009613, -0.00016992, 0.00000313, 0.00018472, 0.00008017, -0.00013366, -0.00001544, 0.00014669, -0.00026414, -0.00059167, },
		{ -0.00039314, -0.00055072, 0.00034037, 0.00060739, -0.00077608, -0.00134112, 0.00069568, 0.00242463, 0.00100710, -0.00120346, -0.00099473, 0.00032512, 0.00026055, -0.00038293, -0.00008440, 0.00032362, -0.00005197, -0.00033803, 0.00000036, 0.00020117, -0.00002394, -0.00010494, 0.00006424, 0.00007046, -0.00011969, -0.00016606, 0.00004677, 0.00020577, 0.00002485, -0.00016552, 0.00004718, 0.00013778, },
		{ 0.00011561, 0.00009404, -0.00050841, -0.00043905, 0.00053820, 0.00041308, -0.00113974, -0.00113905, 0.00129895, 0.00249200, 0.00047739, -0.00144981, -0.00072352, 0.00049194, 0.00011985, -0.00044394, 0.00000371, 0.00030397, -0.00014177, -0.00030679, 0.00007929, 0.00019082, -0.00006450, -0.00009947, 0.00007710, 0.00005385, -0.00014195, -0.00015120, 0.00009709, 0.00020885, -0.00004414, -0.00017610, },
		{ -0.00011880, -0.00016137, 0.00017971, 0.00001867, -0.00059063, -0.00026801, 0.00067663, 0.00011670, -0.00142720, -0.00076859, 0.00186512, 0.00236063, -0.00009172, -0.00155516, -0.00039614, 0.00058350, -0.00005036, -0.00046538, 0.00009179, 0.00025714, -0.00021633, -0.00024974, 0.00014815, 0.00016376, -0.00010230, -0.00008887, 0.00008766, 0.00003427, -0.00016058, -0.00012412, 0.00014816, 0.00018970, },
		{ 0.00019320, 0.00014702, -0.00018961, -0.00012089, 0.00022948, -0.00008007, -0.00062465, -0.00005873, 0.00072724, -0.00025110, -0.00159028, -0.00025834, 0.00232732, 0.00203853, -0.00063717, -0.00150926, -0.00005338, 0.00058923, -0.00022864, -0.00044490, 0.00016892, 0.00018854, -0.00026744, -0.00017322, 0.00020034, 0.00012281, -0.00013442, -0.00007368, 0.00009550, 0.00001252, -0.00017325, -0.00008529, },
		{ -0.00017792, -0.00003694, 0.00022548, 0.00008291, -0.00024689, -0.00005793, 0.00025664, -0.00019000, -0.00060195, 0.00016086, 0.00067359, -0.00064684, -0.00159532, 0.00034258, 0.00262954, 0.00155791, -0.00109900, -0.00132169, 0.00026254, 0.00051048, -0.00039256, -0.00038584, 0.00022577, 0.00010646, -0.00028992, -0.00008586, 0.00023137, 0.00007222, -0.00015854, -0.00005477, 0.00010033, -0.00001039, },
		{ 0.00010196, -0.00003325, -0.00017319, 0.00001719, 0.00023941, 0.00000265, -0.00028228, 0.00002088, 0.00025589, -0.00029674, -0.00052200, 0.00036017, 0.00051407, -0.00102026, -0.00142861, 0.00097094, 0.00273462, 0.00097094, -0.00142861, -0.00102026, 0.00051407, 0.00036017, -0.00052200, -0.00029674, 0.00025589, 0.00002088, -0.00028228, 0.00000265, 0.00023941, 0.00001719, -0.00017319, -0.00003325, },
		{ -0.00017792, -0.00001039, 0.00010033, -0.00005477, -0.00015854, 0.00007222, 0.00023137, -0.00008586, -0.00028992, 0.00010646, 0.00022577, -0.00038584, -0.00039256, 0.00051048, 0.00026254, -0.00132169, -0.00109900, 0.00155791, 0.00262954, 0.00034258, -0.00159532, -0.00064684, 0.00067359, 0.00016086, -0.00060195, -0.00019000, 0.00025664, -0.00005793, -0.00024689, 0.00008291, 0.00022548, -0.00003694, },
		{ 0.00019320, -0.00008529, -0.00017325, 0.00001252, 0.00009550, -0.00007368, -0.00013442, 0.00012281, 0.00020034, -0.00017322, -0.00026744, 0.00018854, 0.00016892, -0.00044490, -0.00022864, 0.00058923, -0.00005338, -0.00150926, -0.00063717, 0.00203853, 0.00232732, -0.00025834, -0.00159028, -0.00025110, 0.00072724, -0.00005873, -0.00062465, -0.00008007, 0.00022948, -0.00012089, -0.00018961, 0.00014702, },
		{ -0.00011880, 0.00018970, 0.00014816, -0.00012412, -0.00016058, 0.00003427, 0.00008766, -0.00008887, -0.00010230, 0.00016376, 0.00014815, -0.00024974, -0.00021633, 0.00025714, 0.00009179, -0.00046538, -0.00005036, 0.00058350, -0.00039614, -0.00155516, -0.00009172, 0.00236063, 0.00186512, -0.00076859, -0.00142720, 0.00011670, 0.00067663, -0.00026801, -0.00059063, 0.00001867, 0.00017971, -0.00016137, },
		{ 0.00011561, -0.00017610, -0.00004414, 0.00020885, 0.00009709, -0.00015120, -0.00014195, 0.00005385, 0.00007710, -0.00009947, -0.00006450, 0.00019082, 0.00007929, -0.00030679, -0.00014177, 0.00030397, 0.00000371, -0.00044394, 0.00011985, 0.00049194, -0.00072352, -0.00144981, 0.00047739, 0.00249200, 0.00129895, -0.00113905, -0.00113974, 0.00041308, 0.00053820, -0.00043905, -0.00050841, 0.00009404, },
		{ -0.00039314, 0.00013778, 0.00004718, -0.00016552, 0.00002485, 0.00020577, 0.00004677, -0.00016606, -0.00011969, 0.00007046, 0.00006424, -0.00010494, -0.00002394, 0.00020117, 0.00000036, -0.00033803, -0.00005197, 0.00032362, -0.00008440, -0.00038293, 0.00026055, 0.00032512, -0.00099473, -0.00120346, 0.00100710, 0.00242463, 0.00069568, -0.00134112, -0.00077608, 0.00060739, 0.00034037, -0.00055072, },
		{ 0.00011897, -0.00059167, -0.00026414, 0.00014669, -0.00001544, -0.00013366, 0.00008017, 0.00018472, 0.00000313, -0.00016992, -0.00009613, 0.00008361, 0.00004956, -0.00010509, 0.00001620, 0.00019383, -0.00008075, -0.00034023, 0.00004296, 0.00031431, -0.00016164, -0.00029004, 0.00035435, 0.00010401, -0.00117622, -0.00084483, 0.00143993, 0.00217548, 0.00012361, -0.00137002, -0.00039139, 0.00068596, },
		{ -0.00003944, 0.00065358, -0.00008845, -0.00056190, -0.00014195, 0.00012300, -0.00006347, -0.00008738, 0.00011646, 0.00015220, -0.00002969, -0.00016532, -0.00007327, 0.00009307, 0.00003364, -0.00010010, 0.00005282, 0.00016982, -0.00015595, -0.00031377, 0.00013242, 0.00027813, -0.00021868, -0.00017724, 0.00039043, -0.00014301, -0.00124653, -0.00041707, 0.00173184, 0.00178361, -0.00035703, -0.00124445, },
		{ -0.00070222, -0.00100294, 0.00023539, 0.00053211, -0.00024849, -0.00047246, -0.00004509, 0.00007385, -0.00009081, -0.00003531, 0.00013172, 0.00011578, -0.00004979, -0.00015566, -0.00005251, 0.00009888, 0.00001708, -0.00009051, 0.00008329, 0.00013195, -0.00021798, -0.00026248, 0.00020680, 0.00022067, -0.00024906, -0.00005902, 0.00036601, -0.00038428, -0.00119929, 0.00002836, 0.00185810, 0.00130422, },
		{ 0.00181631, 0.00080043, -0.00088972, -0.00069717, 0.00040452, 0.00035648, -0.00033708, -0.00034355, 0.00001269, 0.00001014, -0.00009491, 0.00001358, 0.00012744, 0.00008288, -0.00005773, -0.00014457, -0.00003456, 0.00010125, 0.00000052, -0.00007708, 0.00010575, 0.00008446, -0.00026138, -0.00019314, 0.00025875, 0.00015020, -0.00025012, 0.00004975, 0.00028670, -0.00058948, -0.00104404, 0.00043899, },
		{ -0.00080471, 0.00076878, 0.00162608, 0.00033400, -0.00092130, -0.00038358, 0.00045901, 0.00016880, -0.00034307, -0.00020120, 0.00002449, -0.00005535, -0.00007710, 0.00005137, 0.00010828, 0.00005956, -0.00005627, -0.00013526, -0.00001939, 0.00010049, -0.00001540, -0.00006077, 0.00011922, 0.00003242, -0.00028316, -0.00011442, 0.00028421, 0.00007640, -0.00022330, 0.00013588, 0.00016559, -0.00073396, },
		{ 0.00002116, -0.00080223, -0.00051590, 0.00098443, 0.00132557, -0.00004346, -0.00082129, -0.00011436, 0.00041028, 0.00001129, -0.00026989, -0.00007303, -0.00000877, -0.00010995, -0.00004234, 0.00007239, 0.00008113, 0.00004955, -0.00004984, -0.00012998, -0.00000637, 0.00009696, -0.00003007, -0.00004260, 0.00012372, -0.00001893, -0.00028311, -0.00003568, 0.00028280, 0.00000893, -0.00017391, 0.00018981, },
		{ -0.00011022, 0.00020692, -0.00012546, -0.00079013, -0.00021775, 0.00106985, 0.00096534, -0.00029759, -0.00063179, 0.00007064, 0.00028760, -0.00008062, -0.00013485, 0.00001616, -0.00007858, -0.00014316, 0.00000168, 0.00007410, 0.00005400, 0.00005368, -0.00004367, -0.00012965, 0.00000555, 0.00009099, -0.00004293, -0.00002354, 0.00012016, -0.00006492, -0.00026372, 0.00003438, 0.00025780, -0.00004395, },
		{ 0.00021553, -0.00007661, -0.00004218, 0.00018807, -0.00025375, -0.00070516, 0.00005006, 0.00102797, 0.00060041, -0.00041568, -0.00040520, 0.00014995, 0.00013282, -0.00008415, 0.00003375, 0.00004837, -0.00017026, -0.00014835, 0.00004574, 0.00005747, 0.00003458, 0.00006977, -0.00004284, -0.00013372, 0.00001754, 0.00008283, -0.00005349, -0.00000447, 0.00011016, -0.00010202, -0.00022970, 0.00008890, },
		{ -0.00018720, 0.00012377, 0.00016427, -0.00008677, 0.00002006, 0.00013933, -0.00034679, -0.00056516, 0.00025411, 0.00087964, 0.00028193, -0.00040755, -0.00019532, 0.00012349, -0.00000693, 0.00000703, 0.00020119, 0.00001532, -0.00026541, -0.00012381, 0.00008062, 0.00002674, 0.00002897, 0.00009304, -0.00005133, -0.00014026, 0.00003060, 0.00007267, -0.00006133, 0.00001391, 0.00009571, -0.00012823, },
		{ 0.00007893, -0.00014326, -0.00014284, 0.00013810, 0.00011293, -0.00007576, 0.00006774, 0.00007097, -0.00039377, -0.00039538, 0.00037209, 0.00065993, 0.00004955, -0.00030301, -0.00004848, 0.00001211, -0.00008728, 0.00018162, 0.00033202, -0.00008017, -0.00034479, -0.00007303, 0.00009861, -0.00001126, 0.00004061, 0.00011692, -0.00007132, -0.00014631, 0.00004533, 0.00006060, -0.00006616, 0.00003108, },
		},
		{ { -0.00006923, -0.00005347, 0.00001042, 0.00008262, 0.00005708, -0.00003956, -0.00006403, -0.00003403, -0.00008002, -0.00010017, 0.00009900, 0.00029066, 0.00004182, -0.00044747, -0.00042304, 0.00023669, 0.00065727, 0.00023669, -0.00042304, -0.00044747, 0.00004182, 0.00029066, 0.00009900, -0.00010017, -0.00008002, -0.00003403, -0.00006403, -0.00003956, 0.00005708, 0.00008262, 0.00001042, -0.00005347, },
		{ 0.00001416, -0.00006765, -0.00006923, -0.00003957, 0.00000683, 0.00006448, 0.00005867, -0.00001902, -0.00005861, -0.00004583, -0.00007939, -0.00008692, 0.00009776, 0.00027862, 0.00004737, -0.00042849, -0.00041721, 0.00023669, 0.00067592, 0.00025579, -0.00044028, -0.00048470, 0.00003282, 0.00030813, 0.00010068, -0.00011703, -0.00008079, -0.00002016, -0.00007020, -0.00006238, 0.00005536, 0.00010188, },
		{ 0.00005356, 0.00012151, 0.00001789, -0.00008155, -0.00006923, -0.00002647, 0.00000352, 0.00004815, 0.00006007, -0.00000155, -0.00005417, -0.00005510, -0.00007892, -0.00007780, 0.00009699, 0.00027249, 0.00004924, -0.00042849, -0.00042304, 0.00025579, 0.00073118, 0.00029327, -0.00046828, -0.00053874, 0.00002071, 0.00033036, 0.00010272, -0.00013685, -0.00008167, -0.00000474, -0.00007690, -0.00008661, },
		{ -0.00008388, -0.00011131, 0.00005176, 0.00014077, 0.00002148, -0.00009465, -0.00006923, -0.00001468, 0.00000062, 0.00003427, 0.00006121, 0.00001217, -0.00005087, -0.00006149, -0.00007863, -0.00007315, 0.00009673, 0.00027249, 0.00004737, -0.00044747, -0.00044028, 0.00029327, 0.00082092, 0.00034768, -0.00050597, -0.00060753, 0.00000595, 0.00035651, 0.00010505, -0.00015888, -0.00008263, 0.00001162, },
		{ -0.00008362, 0.00002831, -0.00009085, -0.00013554, 0.00005004, 0.00015891, 0.00002479, -0.00010643, -0.00006923, -0.00000466, -0.00000176, 0.00002336, 0.00006206, 0.00002163, -0.00004883, -0.00006475, -0.00007853, -0.00007315, 0.00009699, 0.00027862, 0.00004182, -0.00048470, -0.00046828, 0.00034768, 0.00094168, 0.00041692, -0.00055189, -0.00068842, -0.00001089, 0.00038556, 0.00010758, -0.00018226, },
		{ 0.00011021, -0.00020610, -0.00008461, 0.00004468, -0.00009755, -0.00015836, 0.00004845, 0.00017524, 0.00002769, -0.00011646, -0.00006923, 0.00000322, -0.00000353, 0.00001584, 0.00006259, 0.00002644, -0.00004815, -0.00006475, -0.00007863, -0.00007780, 0.00009776, 0.00029066, 0.00003282, -0.00053874, -0.00050597, 0.00041692, 0.00108884, 0.00049835, -0.00060428, -0.00077829, -0.00002916, 0.00041640, },
		{ -0.00004817, 0.00044784, 0.00011284, -0.00022948, -0.00008557, 0.00006009, -0.00010373, -0.00017890, 0.00004705, 0.00018912, 0.00003007, -0.00012433, -0.00006923, 0.00000864, -0.00000462, 0.00001201, 0.00006276, 0.00002644, -0.00004883, -0.00006149, -0.00007892, -0.00008692, 0.00009900, 0.00030813, 0.00002071, -0.00060753, -0.00055189, 0.00049835, 0.00125672, 0.00058883, -0.00066112, -0.00087371, },
		{ -0.00072024, -0.00097099, -0.00006717, 0.00047868, 0.00011537, -0.00025151, -0.00008645, 0.00007397, -0.00010914, -0.00019636, 0.00004591, 0.00020003, 0.00003184, -0.00012976, -0.00006923, 0.00001141, -0.00000498, 0.00001201, 0.00006259, 0.00002163, -0.00005087, -0.00005510, -0.00007939, -0.00010017, 0.00010068, 0.00033036, 0.00000595, -0.00068842, -0.00060428, 0.00058883, 0.00143888, 0.00068488, },
		{ 0.00162833, 0.00078281, -0.00077936, -0.00106640, -0.00008544, 0.00050773, 0.00011771, -0.00027133, -0.00008722, 0.00008577, -0.00011358, -0.00021009, 0.00004506, 0.00020755, 0.00003293, -0.00013253, -0.00006923, 0.00001141, -0.00000462, 0.00001584, 0.00006206, 0.00001217, -0.00005417, -0.00004583, -0.00008002, -0.00011703, 0.00010272, 0.00035651, -0.00001089, -0.00077829, -0.00066112, 0.00068488, },
		{ -0.00072024, 0.00078281, 0.00181777, 0.00087886, -0.00083620, -0.00115628, -0.00010228, 0.00053387, 0.00011975, -0.00028819, -0.00008785, 0.00009504, -0.00011689, -0.00021954, 0.00004453, 0.00021138, 0.00003330, -0.00013253, -0.00006923, 0.00000864, -0.00000353, 0.00002336, 0.00006121, -0.00000155, -0.00005861, -0.00003403, -0.00008079, -0.00013685, 0.00010505, 0.00038556, -0.00002916, -0.00087371, },
		{ -0.00004817, -0.00097099, -0.00077936, 0.00087886, 0.00199994, 0.00096933, -0.00088859, -0.00123716, -0.00011704, 0.00055611, 0.00012143, -0.00030144, -0.00008832, 0.00010143, -0.00011892, -0.00022436, 0.00004436, 0.00021138, 0.00003293, -0.00012976, -0.00006923, 0.00000322, -0.00000176, 0.00003427, 0.00006007, -0.00001902, -0.00006403, -0.00002016, -0.00008167, -0.00015888, 0.00010758, 0.00041640, },
		{ 0.00011021, 0.00044784, -0.00006717, -0.00106640, -0.00083620, 0.00096933, 0.00216782, 0.00105076, -0.00093451, -0.00130595, -0.00012915, 0.00057358, 0.00012267, -0.00031056, -0.00008861, 0.00010468, -0.00011961, -0.00022436, 0.00004453, 0.00020755, 0.00003184, -0.00012433, -0.00006923, -0.00000466, 0.00000062, 0.00004815, 0.00005867, -0.00003956, -0.00007020, -0.00000474, -0.00008263, -0.00018226, },
		{ -0.00008362, -0.00020610, 0.00011284, 0.00047868, -0.00008544, -0.00115628, -0.00088859, 0.00105076, 0.00231497, 0.00112001, -0.00097220, -0.00136000, -0.00013815, 0.00058561, 0.00012344, -0.00031521, -0.00008871, 0.00010468, -0.00011892, -0.00021954, 0.00004506, 0.00020003, 0.00003007, -0.00011646, -0.00006923, -0.00001468, 0.00000352, 0.00006448, 0.00005708, -0.00006238, -0.00007690, 0.00001162, },
		{ -0.00008388, 0.00002831, -0.00008461, -0.00022948, 0.00011537, 0.00050773, -0.00010228, -0.00123716, -0.00093451, 0.00112001, 0.00243574, 0.00117442, -0.00100020, -0.00139723, -0.00014370, 0.00059175, 0.00012370, -0.00031521, -0.00008861, 0.00010143, -0.00011689, -0.00021009, 0.00004591, 0.00018912, 0.00002769, -0.00010643, -0.00006923, -0.00002647, 0.00000683, 0.00008262, 0.00005536, -0.00008661, },
		{ 0.00005356, -0.00011131, -0.00009085, 0.00004468, -0.00008557, -0.00025151, 0.00011771, 0.00053387, -0.00011704, -0.00130595, -0.00097220, 0.00117442, 0.00252548, 0.00121189, -0.00101744, -0.00141621, -0.00014557, 0.00059175, 0.00012344, -0.00031056, -0.00008832, 0.00009504, -0.00011358, -0.00019636, 0.00004705, 0.00017524, 0.00002479, -0.00009465, -0.00006923, -0.00003957, 0.00001042, 0.00010188, },
		{ 0.00001416, 0.00012151, 0.00005176, -0.00013554, -0.00009755, 0.00006009, -0.00008645, -0.00027133, 0.00011975, 0.00055611, -0.00012915, -0.00136000, -0.00100020, 0.00121189, 0.00258073, 0.00123100, -0.00102327, -0.00141621, -0.00014370, 0.00058561, 0.00012267, -0.00030144, -0.00008785, 0.00008577, -0.00010914, -0.00017890, 0.00004845, 0.00015891, 0.00002148, -0.00008155, -0.00006923, -0.00005347, },
		{ -0.00006923, -0.00006765, 0.00001789, 0.00014077, 0.00005004, -0.00015836, -0.00010373, 0.00007397, -0.00008722, -0.00028819, 0.00012143, 0.00057358, -0.00013815, -0.00139723, -0.00101744, 0.00123100, 0.00259939, 0.00123100, -0.00101744, -0.00139723, -0.00013815, 0.00057358, 0.00012143, -0.00028819, -0.00008722, 0.00007397, -0.00010373, -0.00015836, 0.00005004, 0.00014077, 0.00001789, -0.00006765, },
		{ 0.00001416, -0.00005347, -0.00006923, -0.00008155, 0.00002148, 0.00015891, 0.00004845, -0.00017890, -0.00010914, 0.00008577, -0.00008785, -0.00030144, 0.00012267, 0.00058561, -0.00014370, -0.00141621, -0.00102327, 0.00123100, 0.00258073, 0.00121189, -0.00100020, -0.00136000, -0.00012915, 0.00055611, 0.00011975, -0.00027133, -0.00008645, 0.00006009, -0.00009755, -0.00013554, 0.00005176, 0.00012151, },
		{ 0.00005356, 0.00010188, 0.00001042, -0.00003957, -0.00006923, -0.00009465, 0.00002479, 0.00017524, 0.00004705, -0.00019636, -0.00011358, 0.00009504, -0.00008832, -0.00031056, 0.00012344, 0.00059175, -0.00014557, -0.00141621, -0.00101744, 0.00121189, 0.00252548, 0.00117442, -0.00097220, -0.00130595, -0.00011704, 0.00053387, 0.00011771, -0.00025151, -0.00008557, 0.00004468, -0.00009085, -0.00011131, },
		{ -0.00008388, -0.00008661, 0.00005536, 0.00008262, 0.00000683, -0.00002647, -0.00006923, -0.00010643, 0.00002769, 0.00018912, 0.00004591, -0.00021009, -0.00011689, 0.00010143, -0.00008861, -0.00031521, 0.00012370, 0.00059175, -0.00014370, -0.00139723, -0.00100020, 0.00117442, 0.00243574, 0.00112001, -0.00093451, -0.00123716, -0.00010228, 0.00050773, 0.00011537, -0.00022948, -0.00008461, 0.00002831, },
		{ -0.00008362, 0.00001162, -0.00007690, -0.00006238, 0.00005708, 0.00006448, 0.00000352, -0.00001468, -0.00006923, -0.00011646, 0.00003007, 0.00020003, 0.00004506, -0.00021954, -0.00011892, 0.00010468, -0.00008871, -0.00031521, 0.00012344, 0.00058561, -0.00013815, -0.00136000, -0.00097220, 0.00112001, 0.00231497, 0.00105076, -0.00088859, -0.00115628, -0.00008544, 0.00047868, 0.00011284, -0.00020610, },
		{ 0.00011021, -0.00018226, -0.00008263, -0.00000474, -0.00007020, -0.00003956, 0.00005867, 0.00004815, 0.00000062, -0.00000466, -0.00006923, -0.00012433, 0.00003184, 0.00020755, 0.00004453, -0.00022436, -0.00011961, 0.00010468, -0.00008861, -0.00031056, 0.00012267, 0.00057358, -0.00012915, -0.00130595, -0.00093451, 0.00105076, 0.00216782, 0.00096933, -0.00083620, -0.00106640, -0.00006717, 0.00044784, },
		{ -0.00004817, 0.00041640, 0.00010758, -0.00015888, -0.00008167, -0.00002016, -0.00006403, -0.00001902, 0.00006007, 0.00003427, -0.00000176, 0.00000322, -0.00006923, -0.00012976, 0.00003293, 0.00021138, 0.00004436, -0.00022436, -0.00011892, 0.00010143, -0.00008832, -0.00030144, 0.00012143, 0.00055611, -0.00011704, -0.00123716, -0.00088859, 0.00096933, 0.00199994, 0.00087886, -0.00077936, -0.00097099, },
		{ -0.00072024, -0.00087371, -0.00002916, 0.00038556, 0.00010505, -0.00013685, -0.00008079, -0.00003403, -0.00005861, -0.00000155, 0.00006121, 0.00002336, -0.00000353, 0.00000864, -0.00006923, -0.00013253, 0.00003330, 0.00021138, 0.00004453, -0.00021954, -0.00011689, 0.00009504, -0.00008785, -0.00028819, 0.00011975, 0.00053387, -0.00010228, -0.00115628, -0.00083620, 0.00087886, 0.00181777, 0.00078281, },
		{ 0.00162833, 0.00068488, -0.00066112, -0.00077829, -0.00001089, 0.00035651, 0.00010272, -0.00011703, -0.00008002, -0.00004583, -0.00005417, 0.00001217, 0.00006206, 0.00001584, -0.00000462, 0.00001141, -0.00006923, -0.00013253, 0.00003293, 0.00020755, 0.00004506, -0.00021009, -0.00011358, 0.00008577, -0.00008722, -0.00027133, 0.00011771, 0.00050773, -0.00008544, -0.00106640, -0.00077936, 0.00078281, },
		{ -0.00072024, 0.00068488, 0.00143888, 0.00058883, -0.00060428, -0.00068842, 0.00000595, 0.00033036, 0.00010068, -0.00010017, -0.00007939, -0.00005510, -0.00005087, 0.00002163, 0.00006259, 0.00001201, -0.00000498, 0.00001141, -0.00006923, -0.00012976, 0.00003184, 0.00020003, 0.00004591, -0.00019636, -0.00010914, 0.00007397, -0.00008645, -0.00025151, 0.00011537, 0.00047868, -0.00006717, -0.00097099, },
		{ -0.00004817, -0.00087371, -0.00066112, 0.00058883, 0.00125672, 0.00049835, -0.00055189, -0.00060753, 0.00002071, 0.00030813, 0.00009900, -0.00008692, -0.00007892, -0.00006149, -0.00004883, 0.00002644, 0.00006276, 0.00001201, -0.00000462, 0.00000864, -0.00006923, -0.00012433, 0.00003007, 0.00018912, 0.00004705, -0.00017890, -0.00010373, 0.00006009, -0.00008557, -0.00022948, 0.00011284, 0.00044784, },
		{ 0.00011021, 0.00041640, -0.00002916, -0.00077829, -0.00060428, 0.00049835, 0.00108884, 0.00041692, -0.00050597, -0.00053874, 0.00003282, 0.00029066, 0.00009776, -0.00007780, -0.00007863, -0.00006475, -0.00004815, 0.00002644, 0.00006259, 0.00001584, -0.00000353, 0.00000322, -0.00006923, -0.00011646, 0.00002769, 0.00017524, 0.00004845, -0.00015836, -0.00009755, 0.00004468, -0.00008461, -0.00020610, },
		{ -0.00008362, -0.00018226, 0.00010758, 0.00038556, -0.00001089, -0.00068842, -0.00055189, 0.00041692, 0.00094168, 0.00034768, -0.00046828, -0.00048470, 0.00004182, 0.00027862, 0.00009699, -0.00007315, -0.00007853, -0.00006475, -0.00004883, 0.00002163, 0.00006206, 0.00002336, -0.00000176, -0.00000466, -0.00006923, -0.00010643, 0.00002479, 0.00015891, 0.00005004, -0.00013554, -0.00009085, 0.00002831, },
		{ -0.00008388, 0.00001162, -0.00008263, -0.00015888, 0.00010505, 0.00035651, 0.00000595, -0.00060753, -0.00050597, 0.00034768, 0.00082092, 0.00029327, -0.00044028, -0.00044747, 0.00004737, 0.00027249, 0.00009673, -0.00007315, -0.00007863, -0.00006149, -0.00005087, 0.00001217, 0.00006121, 0.00003427, 0.00000062, -0.00001468, -0.00006923, -0.00009465, 0.00002148, 0.00014077, 0.00005176, -0.00011131, },
		{ 0.00005356, -0.00008661, -0.00007690, -0.00000474, -0.00008167, -0.00013685, 0.00010272, 0.00033036, 0.00002071, -0.00053874, -0.00046828, 0.00029327, 0.00073118, 0.00025579, -0.00042304, -0.00042849, 0.00004924, 0.00027249, 0.00009699, -0.00007780, -0.00007892, -0.00005510, -0.00005417, -0.00000155, 0.00006007, 0.00004815, 0.00000352, -0.00002647, -0.00006923, -0.00008155, 0.00001789, 0.00012151, },
		{ 0.00001416, 0.00010188, 0.00005536, -0.00006238, -0.00007020, -0.00002016, -0.00008079, -0.00011703, 0.00010068, 0.00030813, 0.00003282, -0.00048470, -0.00044028, 0.00025579, 0.00067592, 0.00023669, -0.00041721, -0.00042849, 0.00004737, 0.00027862, 0.00009776, -0.00008692, -0.00007939, -0.00004583, -0.00005861, -0.00001902, 0.00005867, 0.00006448, 0.00000683, -0.00003957, -0.00006923, -0.00006765, },
		},
		{ { 0.00010196, -0.00001039, -0.00017325, -0.00012412, 0.00009709, 0.00020577, 0.00008017, -0.00008738, -0.00009081, 0.00001014, 0.00002449, -0.00007303, -0.00013485, -0.00008415, -0.00000693, 0.00001211, 0.00000414, 0.00001211, -0.00000693, -0.00008415, -0.00013485, -0.00007303, 0.00002449, 0.00001014, -0.00009081, -0.00008738, 0.00008017, 0.00020577, 0.00009709, -0.00012412, -0.00017325, -0.00001039, },
		{ -0.00017792, -0.00003325, 0.00010033, 0.00001252, -0.00016058, -0.00015120, 0.00004677, 0.00018472, 0.00011646, -0.00003531, -0.00009491, -0.00005535, -0.00000877, 0.00001616, 0.00003375, 0.00000703, -0.00008728, -0.00014622, -0.00004848, 0.00012349, 0.00013282, -0.00008062, -0.00026989, -0.00020120, 0.00001269, 0.00007385, -0.00006347, -0.00013366, 0.00002485, 0.00020885, 0.00014816, -0.00008529, },
		{ 0.00019320, -0.00003694, -0.00017319, -0.00005477, 0.00009550, 0.00003427, -0.00014195, -0.00016606, 0.00000313, 0.00015220, 0.00013172, 0.00001358, -0.00007710, -0.00010995, -0.00007858, 0.00004837, 0.00020119, 0.00018162, -0.00007429, -0.00030301, -0.00019532, 0.00014995, 0.00028760, 0.00001129, -0.00034307, -0.00034355, -0.00004509, 0.00012300, -0.00001544, -0.00016552, -0.00004414, 0.00018970, },
		{ -0.00011880, 0.00014702, 0.00022548, 0.00001719, -0.00015854, -0.00007368, 0.00008766, 0.00005385, -0.00011969, -0.00016992, -0.00002969, 0.00011578, 0.00012744, 0.00005137, -0.00004234, -0.00014316, -0.00017026, 0.00001532, 0.00033202, 0.00041230, 0.00004955, -0.00040755, -0.00040520, 0.00007064, 0.00041028, 0.00016880, -0.00033708, -0.00047246, -0.00014195, 0.00014669, 0.00004718, -0.00017610, },
		{ 0.00011561, -0.00016137, -0.00018961, 0.00008291, 0.00023941, 0.00007222, -0.00013442, -0.00008887, 0.00007710, 0.00007046, -0.00009613, -0.00016532, -0.00004979, 0.00008288, 0.00010828, 0.00007239, 0.00000168, -0.00014835, -0.00026541, -0.00008017, 0.00039579, 0.00065993, 0.00028193, -0.00041568, -0.00063179, -0.00011436, 0.00045901, 0.00035648, -0.00024849, -0.00056190, -0.00026414, 0.00013778, },
		{ -0.00039314, 0.00009404, 0.00017971, -0.00012089, -0.00024689, 0.00000265, 0.00023137, 0.00012281, -0.00010230, -0.00009947, 0.00006424, 0.00008361, -0.00007327, -0.00015566, -0.00005773, 0.00005956, 0.00008113, 0.00007410, 0.00004574, -0.00012381, -0.00034479, -0.00022451, 0.00037209, 0.00087964, 0.00060041, -0.00029759, -0.00082129, -0.00038358, 0.00040452, 0.00053211, -0.00008845, -0.00059167, },
		{ 0.00011897, -0.00055072, -0.00050841, 0.00001867, 0.00022948, -0.00005793, -0.00028228, -0.00008586, 0.00020034, 0.00016376, -0.00006450, -0.00010494, 0.00004956, 0.00009307, -0.00005251, -0.00014457, -0.00005627, 0.00004955, 0.00005400, 0.00005747, 0.00008062, -0.00007303, -0.00039152, -0.00039538, 0.00025411, 0.00102797, 0.00096534, -0.00004346, -0.00092130, -0.00069717, 0.00023539, 0.00065358, },
		{ -0.00003944, 0.00068596, 0.00034037, -0.00043905, -0.00059063, -0.00008007, 0.00025664, 0.00002088, -0.00028992, -0.00017322, 0.00014815, 0.00019082, -0.00002394, -0.00010509, 0.00003364, 0.00009888, -0.00003456, -0.00013526, -0.00004984, 0.00005368, 0.00003458, 0.00002674, 0.00009861, -0.00000423, -0.00039377, -0.00056516, 0.00005006, 0.00106985, 0.00132557, 0.00033400, -0.00088972, -0.00100294, },
		{ -0.00070222, -0.00124445, -0.00039139, 0.00060739, 0.00053820, -0.00026801, -0.00062465, -0.00019000, 0.00025589, 0.00010646, -0.00026744, -0.00024974, 0.00007929, 0.00020117, 0.00001620, -0.00010010, 0.00001708, 0.00010125, -0.00001939, -0.00012998, -0.00004367, 0.00006977, 0.00002897, -0.00001126, 0.00009477, 0.00007097, -0.00034679, -0.00070516, -0.00021775, 0.00098443, 0.00162608, 0.00080043, },
		{ 0.00181631, 0.00130422, -0.00035703, -0.00137002, -0.00077608, 0.00041308, 0.00067663, -0.00005873, -0.00060195, -0.00029674, 0.00022577, 0.00018854, -0.00021633, -0.00030679, 0.00000036, 0.00019383, 0.00005282, -0.00009051, 0.00000052, 0.00010049, -0.00000637, -0.00012965, -0.00004284, 0.00009304, 0.00004061, -0.00004822, 0.00006774, 0.00013933, -0.00025375, -0.00079013, -0.00051590, 0.00076878, },
		{ -0.00080471, 0.00043899, 0.00185810, 0.00178361, 0.00012361, -0.00134112, -0.00113974, 0.00011670, 0.00072724, 0.00016086, -0.00052200, -0.00038584, 0.00016892, 0.00025714, -0.00014177, -0.00033803, -0.00008075, 0.00016982, 0.00008329, -0.00007708, -0.00001540, 0.00009696, 0.00000555, -0.00013372, -0.00005133, 0.00011692, 0.00006966, -0.00007576, 0.00002006, 0.00018807, -0.00012546, -0.00080223, },
		{ 0.00002116, -0.00073396, -0.00104404, 0.00002836, 0.00173184, 0.00217548, 0.00069568, -0.00113905, -0.00142720, -0.00025110, 0.00067359, 0.00036017, -0.00039256, -0.00044490, 0.00009179, 0.00030397, -0.00005197, -0.00034023, -0.00015595, 0.00013195, 0.00010575, -0.00006077, -0.00003007, 0.00009099, 0.00001754, -0.00014026, -0.00007132, 0.00013417, 0.00011293, -0.00008677, -0.00004218, 0.00020692, },
		{ -0.00011022, 0.00018981, 0.00016559, -0.00058948, -0.00119929, -0.00041707, 0.00143993, 0.00242463, 0.00129895, -0.00076859, -0.00159028, -0.00064684, 0.00051407, 0.00051048, -0.00022864, -0.00046538, 0.00000371, 0.00032362, 0.00004296, -0.00031377, -0.00021798, 0.00008446, 0.00011922, -0.00004260, -0.00004293, 0.00008283, 0.00003060, -0.00014631, -0.00010267, 0.00013810, 0.00016427, -0.00007661, },
		{ 0.00021553, -0.00004395, -0.00017391, 0.00013588, 0.00028670, -0.00038428, -0.00124653, -0.00084483, 0.00100710, 0.00249200, 0.00186512, -0.00025834, -0.00159532, -0.00102026, 0.00026254, 0.00058923, -0.00005036, -0.00044394, -0.00008440, 0.00031431, 0.00013242, -0.00026248, -0.00026138, 0.00003242, 0.00012372, -0.00002354, -0.00005349, 0.00007267, 0.00004533, -0.00014842, -0.00014284, 0.00012377, },
		{ -0.00018720, 0.00008890, 0.00025780, 0.00000893, -0.00022330, 0.00004975, 0.00036601, -0.00014301, -0.00117622, -0.00120346, 0.00047739, 0.00236063, 0.00232732, 0.00034258, -0.00142861, -0.00132169, -0.00005338, 0.00058350, 0.00011985, -0.00038293, -0.00016164, 0.00027813, 0.00020680, -0.00019314, -0.00028316, -0.00001893, 0.00012016, -0.00000447, -0.00006133, 0.00006060, 0.00006168, -0.00014326, },
		{ 0.00007893, -0.00012823, -0.00022970, 0.00003438, 0.00028280, 0.00007640, -0.00025012, -0.00005902, 0.00039043, 0.00010401, -0.00099473, -0.00144981, -0.00009172, 0.00203853, 0.00262954, 0.00097094, -0.00109900, -0.00150926, -0.00039614, 0.00049194, 0.00026055, -0.00029004, -0.00021868, 0.00022067, 0.00025875, -0.00011442, -0.00028311, -0.00006492, 0.00011016, 0.00001391, -0.00006616, 0.00004670, },
		{ -0.00006779, 0.00003108, 0.00009571, -0.00010202, -0.00026372, -0.00003568, 0.00028421, 0.00015020, -0.00024906, -0.00017724, 0.00035435, 0.00032512, -0.00072352, -0.00155516, -0.00063717, 0.00155791, 0.00273462, 0.00155791, -0.00063717, -0.00155516, -0.00072352, 0.00032512, 0.00035435, -0.00017724, -0.00024906, 0.00015020, 0.00028421, -0.00003568, -0.00026372, -0.00010202, 0.00009571, 0.00003108, },
		{ 0.00007893, 0.00004670, -0.00006616, 0.00001391, 0.00011016, -0.00006492, -0.00028311, -0.00011442, 0.00025875, 0.00022067, -0.00021868, -0.00029004, 0.00026055, 0.00049194, -0.00039614, -0.00150926, -0.00109900, 0.00097094, 0.00262954, 0.00203853, -0.00009172, -0.00144981, -0.00099473, 0.00010401, 0.00039043, -0.00005902, -0.00025012, 0.00007640, 0.00028280, 0.00003438, -0.00022970, -0.00012823, },
		{ -0.00018720, -0.00014326, 0.00006168, 0.00006060, -0.00006133, -0.00000447, 0.00012016, -0.00001893, -0.00028316, -0.00019314, 0.00020680, 0.00027813, -0.00016164, -0.00038293, 0.00011985, 0.00058350, -0.00005338, -0.00132169, -0.00142861, 0.00034258, 0.00232732, 0.00236063, 0.00047739, -0.00120346, -0.00117622, -0.00014301, 0.00036601, 0.00004975, -0.00022330, 0.00000893, 0.00025780, 0.00008890, },
		{ 0.00021553, 0.00012377, -0.00014284, -0.00014842, 0.00004533, 0.00007267, -0.00005349, -0.00002354, 0.00012372, 0.00003242, -0.00026138, -0.00026248, 0.00013242, 0.00031431, -0.00008440, -0.00044394, -0.00005036, 0.00058923, 0.00026254, -0.00102026, -0.00159532, -0.00025834, 0.00186512, 0.00249200, 0.00100710, -0.00084483, -0.00124653, -0.00038428, 0.00028670, 0.00013588, -0.00017391, -0.00004395, },
		{ -0.00011022, -0.00007661, 0.00016427, 0.00013810, -0.00010267, -0.00014631, 0.00003060, 0.00008283, -0.00004293, -0.00004260, 0.00011922, 0.00008446, -0.00021798, -0.00031377, 0.00004296, 0.00032362, 0.00000371, -0.00046538, -0.00022864, 0.00051048, 0.00051407, -0.00064684, -0.00159028, -0.00076859, 0.00129895, 0.00242463, 0.00143993, -0.00041707, -0.00119929, -0.00058948, 0.00016559, 0.00018981, },
		{ 0.00002116, 0.00020692, -0.00004218, -0.00008677, 0.00011293, 0.00013417, -0.00007132, -0.00014026, 0.00001754, 0.00009099, -0.00003007, -0.00006077, 0.00010575, 0.00013195, -0.00015595, -0.00034023, -0.00005197, 0.00030397, 0.00009179, -0.00044490, -0.00039256, 0.00036017, 0.00067359, -0.00025110, -0.00142720, -0.00113905, 0.00069568, 0.00217548, 0.00173184, 0.00002836, -0.00104404, -0.00073396, },
		{ -0.00080471, -0.00080223, -0.00012546, 0.00018807, 0.00002006, -0.00007576, 0.00006966, 0.00011692, -0.00005133, -0.00013372, 0.00000555, 0.00009696, -0.00001540, -0.00007708, 0.00008329, 0.00016982, -0.00008075, -0.00033803, -0.00014177, 0.00025714, 0.00016892, -0.00038584, -0.00052200, 0.00016086, 0.00072724, 0.00011670, -0.00113974, -0.00134112, 0.00012361, 0.00178361, 0.00185810, 0.00043899, },
		{ 0.00181631, 0.00076878, -0.00051590, -0.00079013, -0.00025375, 0.00013933, 0.00006774, -0.00004822, 0.00004061, 0.00009304, -0.00004284, -0.00012965, -0.00000637, 0.00010049, 0.00000052, -0.00009051, 0.00005282, 0.00019383, 0.00000036, -0.00030679, -0.00021633, 0.00018854, 0.00022577, -0.00029674, -0.00060195, -0.00005873, 0.00067663, 0.00041308, -0.00077608, -0.00137002, -0.00035703, 0.00130422, },
		{ -0.00070222, 0.00080043, 0.00162608, 0.00098443, -0.00021775, -0.00070516, -0.00034679, 0.00007097, 0.00009477, -0.00001126, 0.00002897, 0.00006977, -0.00004367, -0.00012998, -0.00001939, 0.00010125, 0.00001708, -0.00010010, 0.00001620, 0.00020117, 0.00007929, -0.00024974, -0.00026744, 0.00010646, 0.00025589, -0.00019000, -0.00062465, -0.00026801, 0.00053820, 0.00060739, -0.00039139, -0.00124445, },
		{ -0.00003944, -0.00100294, -0.00088972, 0.00033400, 0.00132557, 0.00106985, 0.00005006, -0.00056516, -0.00039377, -0.00000423, 0.00009861, 0.00002674, 0.00003458, 0.00005368, -0.00004984, -0.00013526, -0.00003456, 0.00009888, 0.00003364, -0.00010509, -0.00002394, 0.00019082, 0.00014815, -0.00017322, -0.00028992, 0.00002088, 0.00025664, -0.00008007, -0.00059063, -0.00043905, 0.00034037, 0.00068596, },
		{ 0.00011897, 0.00065358, 0.00023539, -0.00069717, -0.00092130, -0.00004346, 0.00096534, 0.00102797, 0.00025411, -0.00039538, -0.00039152, -0.00007303, 0.00008062, 0.00005747, 0.00005400, 0.00004955, -0.00005627, -0.00014457, -0.00005251, 0.00009307, 0.00004956, -0.00010494, -0.00006450, 0.00016376, 0.00020034, -0.00008586, -0.00028228, -0.00005793, 0.00022948, 0.00001867, -0.00050841, -0.00055072, },
		{ -0.00039314, -0.00059167, -0.00008845, 0.00053211, 0.00040452, -0.00038358, -0.00082129, -0.00029759, 0.00060041, 0.00087964, 0.00037209, -0.00022451, -0.00034479, -0.00012381, 0.00004574, 0.00007410, 0.00008113, 0.00005956, -0.00005773, -0.00015566, -0.00007327, 0.00008361, 0.00006424, -0.00009947, -0.00010230, 0.00012281, 0.00023137, 0.00000265, -0.00024689, -0.00012089, 0.00017971, 0.00009404, },
		{ 0.00011561, 0.00013778, -0.00026414, -0.00056190, -0.00024849, 0.00035648, 0.00045901, -0.00011436, -0.00063179, -0.00041568, 0.00028193, 0.00065993, 0.00039579, -0.00008017, -0.00026541, -0.00014835, 0.00000168, 0.00007239, 0.00010828, 0.00008288, -0.00004979, -0.00016532, -0.00009613, 0.00007046, 0.00007710, -0.00008887, -0.00013442, 0.00007222, 0.00023941, 0.00008291, -0.00018961, -0.00016137, },
		{ -0.00011880, -0.00017610, 0.00004718, 0.00014669, -0.00014195, -0.00047246, -0.00033708, 0.00016880, 0.00041028, 0.00007064, -0.00040520, -0.00040755, 0.00004955, 0.00041230, 0.00033202, 0.00001532, -0.00017026, -0.00014316, -0.00004234, 0.00005137, 0.00012744, 0.00011578, -0.00002969, -0.00016992, -0.00011969, 0.00005385, 0.00008766, -0.00007368, -0.00015854, 0.00001719, 0.00022548, 0.00014702, },
		{ 0.00019320, 0.00018970, -0.00004414, -0.00016552, -0.00001544, 0.00012300, -0.00004509, -0.00034355, -0.00034307, 0.00001129, 0.00028760, 0.00014995, -0.00019532, -0.00030301, -0.00007429, 0.00018162, 0.00020119, 0.00004837, -0.00007858, -0.00010995, -0.00007710, 0.00001358, 0.00013172, 0.00015220, 0.00000313, -0.00016606, -0.00014195, 0.00003427, 0.00009550, -0.00005477, -0.00017319, -0.00003694, },
		{ -0.00017792, -0.00008529, 0.00014816, 0.00020885, 0.00002485, -0.00013366, -0.00006347, 0.00007385, 0.00001269, -0.00020120, -0.00026989, -0.00008062, 0.00013282, 0.00012349, -0.00004848, -0.00014622, -0.00008728, 0.00000703, 0.00003375, 0.00001616, -0.00000877, -0.00005535, -0.00009491, -0.00003531, 0.00011646, 0.00018472, 0.00004677, -0.00015120, -0.00016058, 0.00001252, 0.00010033, -0.00003325, },
		},
		{ { -0.00015699, -0.00001715, 0.00023104, 0.00027213, 0.00001047, -0.00032110, -0.00042345, -0.00024978, -0.00001888, 0.00006259, 0.00001948, 0.00002035, 0.00012440, 0.00019722, 0.00009740, -0.00011532, -0.00022604, -0.00011532, 0.00009740, 0.00019722, 0.00012440, 0.00002035, 0.00001948, 0.00006259, -0.00001888, -0.00024978, -0.00042345, -0.00032110, 0.00001047, 0.00027213, 0.00023104, -0.00001715, },
		{ 0.00031144, 0.00014206, -0.00010534, -0.00013046, 0.00007954, 0.00025636, 0.00017432, -0.00009828, -0.00030036, -0.00028280, -0.00014474, -0.00006411, -0.00006595, -0.00002550, 0.00011856, 0.00024055, 0.00017231, -0.00006326, -0.00024011, -0.00017736, 0.00004819, 0.00019216, 0.00013216, 0.00000174, 0.00000162, 0.00012495, 0.00015012, -0.00007323, -0.00039454, -0.00049171, -0.00023271, 0.00015342, },
		{ -0.00045480, -0.00007193, 0.00026808, 0.00027321, 0.00002522, -0.00015016, -0.00006601, 0.00014077, 0.00020965, 0.00006900, -0.00011451, -0.00017267, -0.00013310, -0.00013386, -0.00019573, -0.00017618, 0.00002136, 0.00026314, 0.00029643, 0.00005123, -0.00024038, -0.00028233, -0.00004710, 0.00019516, 0.00019507, 0.00001210, -0.00007656, 0.00006364, 0.00023907, 0.00015613, -0.00020252, -0.00051307, },
		{ 0.00007252, -0.00035450, -0.00055154, -0.00031982, 0.00010185, 0.00031079, 0.00017349, -0.00007507, -0.00014231, -0.00000231, 0.00013492, 0.00011875, 0.00002264, -0.00000755, 0.00001766, -0.00004754, -0.00023965, -0.00036552, -0.00021077, 0.00015820, 0.00039915, 0.00025424, -0.00013967, -0.00037644, -0.00022396, 0.00012256, 0.00027687, 0.00011120, -0.00011219, -0.00007967, 0.00017708, 0.00030877, },
		{ -0.00000646, 0.00029376, 0.00029892, -0.00007443, -0.00046888, -0.00048669, -0.00012889, 0.00022968, 0.00027110, 0.00005297, -0.00012704, -0.00010299, 0.00001613, 0.00006085, 0.00004242, 0.00009281, 0.00021007, 0.00019579, -0.00008375, -0.00044253, -0.00050548, -0.00013520, 0.00034851, 0.00046793, 0.00010634, -0.00034863, -0.00042441, -0.00007894, 0.00027669, 0.00026666, -0.00002626, -0.00019201, },
		{ 0.00016804, -0.00016803, -0.00019719, 0.00011537, 0.00036465, 0.00020476, -0.00023422, -0.00049913, -0.00033312, 0.00005776, 0.00027470, 0.00017284, -0.00004273, -0.00012685, -0.00007547, -0.00003917, -0.00004495, 0.00005185, 0.00030423, 0.00048117, 0.00027952, -0.00026229, -0.00068465, -0.00055052, 0.00004848, 0.00053160, 0.00042723, -0.00012028, -0.00051421, -0.00036313, 0.00011104, 0.00036538, },
		{ -0.00020457, 0.00028977, 0.00035969, 0.00001268, -0.00026582, -0.00012573, 0.00024011, 0.00035777, 0.00005328, -0.00035076, -0.00042921, -0.00013711, 0.00018450, 0.00023460, 0.00006197, -0.00008420, -0.00010199, -0.00010600, -0.00016313, -0.00011565, 0.00020476, 0.00063399, 0.00071797, 0.00020672, -0.00055662, -0.00088704, -0.00046112, 0.00029996, 0.00064197, 0.00026669, -0.00037076, -0.00058466, },
		{ 0.00001333, -0.00057870, -0.00053458, 0.00000956, 0.00040437, 0.00026051, -0.00015074, -0.00028849, -0.00000355, 0.00032104, 0.00026984, -0.00010669, -0.00038289, -0.00027919, 0.00004084, 0.00022068, 0.00014077, -0.00001404, -0.00007261, -0.00010107, -0.00022146, -0.00031126, -0.00007454, 0.00052106, 0.00101183, 0.00082780, -0.00004214, -0.00089813, -0.00097450, -0.00023824, 0.00055184, 0.00063097, },
		{ 0.00007463, 0.00073032, 0.00048380, -0.00027521, -0.00068655, -0.00037336, 0.00022075, 0.00042198, 0.00009765, -0.00027308, -0.00023204, 0.00012558, 0.00032746, 0.00012743, -0.00022228, -0.00031951, -0.00009801, 0.00014824, 0.00017144, 0.00004456, -0.00002737, -0.00004173, -0.00017347, -0.00041209, -0.00040875, 0.00013773, 0.00096866, 0.00133655, 0.00074250, -0.00043398, -0.00119528, -0.00090029, },
		{ -0.00088941, -0.00136087, -0.00066235, 0.00040339, 0.00077860, 0.00022590, -0.00052742, -0.00066365, -0.00014259, 0.00037195, 0.00033970, -0.00008055, -0.00032093, -0.00012000, 0.00021612, 0.00025590, -0.00002247, -0.00025592, -0.00018410, 0.00005415, 0.00015990, 0.00007515, -0.00000415, 0.00001650, -0.00004611, -0.00034633, -0.00062672, -0.00037263, 0.00053826, 0.00146536, 0.00150419, 0.00044215, },
		{ 0.00189741, 0.00143935, -0.00003507, -0.00130385, -0.00133878, -0.00030575, 0.00066538, 0.00067551, -0.00008349, -0.00067964, -0.00051601, 0.00009671, 0.00042519, 0.00018482, -0.00022339, -0.00028721, 0.00000566, 0.00023722, 0.00013146, -0.00012949, -0.00019945, -0.00002688, 0.00012763, 0.00008739, -0.00001154, 0.00002179, 0.00007795, -0.00013963, -0.00060736, -0.00078328, -0.00014030, 0.00108822, },
		{ -0.00079230, 0.00030513, 0.00169604, 0.00214977, 0.00111992, -0.00060025, -0.00157715, -0.00112175, 0.00008983, 0.00079462, 0.00044293, -0.00036876, -0.00069547, -0.00028448, 0.00028292, 0.00037293, 0.00000521, -0.00029442, -0.00019250, 0.00009991, 0.00018392, -0.00000101, -0.00015824, -0.00007795, 0.00009274, 0.00010171, -0.00002481, -0.00003589, 0.00011364, 0.00009620, -0.00035171, -0.00090078, },
		{ 0.00001185, -0.00067014, -0.00113872, -0.00058648, 0.00092149, 0.00223769, 0.00213921, 0.00058703, -0.00113674, -0.00164241, -0.00075416, 0.00043349, 0.00076200, 0.00013972, -0.00056071, -0.00057686, -0.00003204, 0.00037269, 0.00023936, -0.00014677, -0.00028280, -0.00007686, 0.00013082, 0.00007900, -0.00009307, -0.00010180, 0.00005825, 0.00012744, -0.00000703, -0.00011523, 0.00002961, 0.00022678, },
		{ -0.00012801, 0.00017434, 0.00029462, -0.00020876, -0.00103716, -0.00122580, -0.00014643, 0.00161158, 0.00258862, 0.00183987, -0.00006334, -0.00153189, -0.00148517, -0.00031920, 0.00065004, 0.00058393, -0.00015621, -0.00061746, -0.00036303, 0.00017573, 0.00035354, 0.00007154, -0.00023132, -0.00020575, 0.00001479, 0.00009105, -0.00003501, -0.00011036, 0.00001471, 0.00015077, 0.00005959, -0.00014936, },
		{ 0.00015382, -0.00009103, -0.00025983, -0.00003424, 0.00035020, 0.00025175, -0.00055611, -0.00136148, -0.00108920, 0.00048608, 0.00224467, 0.00265815, 0.00129325, -0.00070490, -0.00170772, -0.00114675, 0.00008576, 0.00070176, 0.00031662, -0.00037168, -0.00053663, -0.00011781, 0.00029063, 0.00024639, -0.00007665, -0.00023363, -0.00010143, 0.00004891, 0.00000052, -0.00011115, -0.00004575, 0.00014176, },
		{ -0.00011511, 0.00007683, 0.00022206, 0.00005125, -0.00027222, -0.00027948, 0.00015530, 0.00048580, 0.00005883, -0.00097354, -0.00154323, -0.00070479, 0.00121272, 0.00268958, 0.00241566, 0.00059907, -0.00121450, -0.00164144, -0.00071117, 0.00037750, 0.00059818, 0.00003794, -0.00046034, -0.00035504, 0.00009055, 0.00029583, 0.00009736, -0.00016220, -0.00016755, -0.00001483, 0.00001525, -0.00010068, },
		{ 0.00001852, -0.00006950, -0.00016660, -0.00003839, 0.00020945, 0.00021045, -0.00013604, -0.00041586, -0.00017713, 0.00039088, 0.00051011, -0.00027965, -0.00136933, -0.00150429, -0.00011036, 0.00190057, 0.00284971, 0.00190057, -0.00011036, -0.00150429, -0.00136933, -0.00027965, 0.00051011, 0.00039088, -0.00017713, -0.00041586, -0.00013604, 0.00021045, 0.00020945, -0.00003839, -0.00016660, -0.00006950, },
		{ -0.00011511, -0.00010068, 0.00001525, -0.00001483, -0.00016755, -0.00016220, 0.00009736, 0.00029583, 0.00009055, -0.00035504, -0.00046034, 0.00003794, 0.00059818, 0.00037750, -0.00071117, -0.00164144, -0.00121450, 0.00059907, 0.00241566, 0.00268958, 0.00121272, -0.00070479, -0.00154323, -0.00097354, 0.00005883, 0.00048580, 0.00015530, -0.00027948, -0.00027222, 0.00005125, 0.00022206, 0.00007683, },
		{ 0.00015382, 0.00014176, -0.00004575, -0.00011115, 0.00000052, 0.00004891, -0.00010143, -0.00023363, -0.00007665, 0.00024639, 0.00029063, -0.00011781, -0.00053663, -0.00037168, 0.00031662, 0.00070176, 0.00008576, -0.00114675, -0.00170772, -0.00070490, 0.00129325, 0.00265815, 0.00224467, 0.00048608, -0.00108920, -0.00136148, -0.00055611, 0.00025175, 0.00035020, -0.00003424, -0.00025983, -0.00009103, },
		{ -0.00012801, -0.00014936, 0.00005959, 0.00015077, 0.00001471, -0.00011036, -0.00003501, 0.00009105, 0.00001479, -0.00020575, -0.00023132, 0.00007154, 0.00035354, 0.00017573, -0.00036303, -0.00061746, -0.00015621, 0.00058393, 0.00065004, -0.00031920, -0.00148517, -0.00153189, -0.00006334, 0.00183987, 0.00258862, 0.00161158, -0.00014643, -0.00122580, -0.00103716, -0.00020876, 0.00029462, 0.00017434, },
		{ 0.00001185, 0.00022678, 0.00002961, -0.00011523, -0.00000703, 0.00012744, 0.00005825, -0.00010180, -0.00009307, 0.00007900, 0.00013082, -0.00007686, -0.00028280, -0.00014677, 0.00023936, 0.00037269, -0.00003204, -0.00057686, -0.00056071, 0.00013972, 0.00076200, 0.00043349, -0.00075416, -0.00164241, -0.00113674, 0.00058703, 0.00213921, 0.00223769, 0.00092149, -0.00058648, -0.00113872, -0.00067014, },
		{ -0.00079230, -0.00090078, -0.00035171, 0.00009620, 0.00011364, -0.00003589, -0.00002481, 0.00010171, 0.00009274, -0.00007795, -0.00015824, -0.00000101, 0.00018392, 0.00009991, -0.00019250, -0.00029442, 0.00000521, 0.00037293, 0.00028292, -0.00028448, -0.00069547, -0.00036876, 0.00044293, 0.00079462, 0.00008983, -0.00112175, -0.00157715, -0.00060025, 0.00111992, 0.00214977, 0.00169604, 0.00030513, },
		{ 0.00189741, 0.00108822, -0.00014030, -0.00078328, -0.00060736, -0.00013963, 0.00007795, 0.00002179, -0.00001154, 0.00008739, 0.00012763, -0.00002688, -0.00019945, -0.00012949, 0.00013146, 0.00023722, 0.00000566, -0.00028721, -0.00022339, 0.00018482, 0.00042519, 0.00009671, -0.00051601, -0.00067964, -0.00008349, 0.00067551, 0.00066538, -0.00030575, -0.00133878, -0.00130385, -0.00003507, 0.00143935, },
		{ -0.00088941, 0.00044215, 0.00150419, 0.00146536, 0.00053826, -0.00037263, -0.00062672, -0.00034633, -0.00004611, 0.00001650, -0.00000415, 0.00007515, 0.00015990, 0.00005415, -0.00018410, -0.00025592, -0.00002247, 0.00025590, 0.00021612, -0.00012000, -0.00032093, -0.00008055, 0.00033970, 0.00037195, -0.00014259, -0.00066365, -0.00052742, 0.00022590, 0.00077860, 0.00040339, -0.00066235, -0.00136087, },
		{ 0.00007463, -0.00090029, -0.00119528, -0.00043398, 0.00074250, 0.00133655, 0.00096866, 0.00013773, -0.00040875, -0.00041209, -0.00017347, -0.00004173, -0.00002737, 0.00004456, 0.00017144, 0.00014824, -0.00009801, -0.00031951, -0.00022228, 0.00012743, 0.00032746, 0.00012558, -0.00023204, -0.00027308, 0.00009765, 0.00042198, 0.00022075, -0.00037336, -0.00068655, -0.00027521, 0.00048380, 0.00073032, },
		{ 0.00001333, 0.00063097, 0.00055184, -0.00023824, -0.00097450, -0.00089813, -0.00004214, 0.00082780, 0.00101183, 0.00052106, -0.00007454, -0.00031126, -0.00022146, -0.00010107, -0.00007261, -0.00001404, 0.00014077, 0.00022068, 0.00004084, -0.00027919, -0.00038289, -0.00010669, 0.00026984, 0.00032104, -0.00000355, -0.00028849, -0.00015074, 0.00026051, 0.00040437, 0.00000956, -0.00053458, -0.00057870, },
		{ -0.00020457, -0.00058466, -0.00037076, 0.00026669, 0.00064197, 0.00029996, -0.00046112, -0.00088704, -0.00055662, 0.00020672, 0.00071797, 0.00063399, 0.00020476, -0.00011565, -0.00016313, -0.00010600, -0.00010199, -0.00008420, 0.00006197, 0.00023460, 0.00018450, -0.00013711, -0.00042921, -0.00035076, 0.00005328, 0.00035777, 0.00024011, -0.00012573, -0.00026582, 0.00001268, 0.00035969, 0.00028977, },
		{ 0.00016804, 0.00036538, 0.00011104, -0.00036313, -0.00051421, -0.00012028, 0.00042723, 0.00053160, 0.00004848, -0.00055052, -0.00068465, -0.00026229, 0.00027952, 0.00048117, 0.00030423, 0.00005185, -0.00004495, -0.00003917, -0.00007547, -0.00012685, -0.00004273, 0.00017284, 0.00027470, 0.00005776, -0.00033312, -0.00049913, -0.00023422, 0.00020476, 0.00036465, 0.00011537, -0.00019719, -0.00016803, },
		{ -0.00000646, -0.00019201, -0.00002626, 0.00026666, 0.00027669, -0.00007894, -0.00042441, -0.00034863, 0.00010634, 0.00046793, 0.00034851, -0.00013520, -0.00050548, -0.00044253, -0.00008375, 0.00019579, 0.00021007, 0.00009281, 0.00004242, 0.00006085, 0.00001613, -0.00010299, -0.00012704, 0.00005297, 0.00027110, 0.00022968, -0.00012889, -0.00048669, -0.00046888, -0.00007443, 0.00029892, 0.00029376, },
		{ 0.00007252, 0.00030877, 0.00017708, -0.00007967, -0.00011219, 0.00011120, 0.00027687, 0.00012256, -0.00022396, -0.00037644, -0.00013967, 0.00025424, 0.00039915, 0.00015820, -0.00021077, -0.00036552, -0.00023965, -0.00004754, 0.00001766, -0.00000755, 0.00002264, 0.00011875, 0.00013492, -0.00000231, -0.00014231, -0.00007507, 0.00017349, 0.00031079, 0.00010185, -0.00031982, -0.00055154, -0.00035450, },
		{ -0.00045480, -0.00051307, -0.00020252, 0.00015613, 0.00023907, 0.00006364, -0.00007656, 0.00001210, 0.00019507, 0.00019516, -0.00004710, -0.00028233, -0.00024038, 0.00005123, 0.00029643, 0.00026314, 0.00002136, -0.00017618, -0.00019573, -0.00013386, -0.00013310, -0.00017267, -0.00011451, 0.00006900, 0.00020965, 0.00014077, -0.00006601, -0.00015016, 0.00002522, 0.00027321, 0.00026808, -0.00007193, },
		{ 0.00031144, 0.00015342, -0.00023271, -0.00049171, -0.00039454, -0.00007323, 0.00015012, 0.00012495, 0.00000162, 0.00000174, 0.00013216, 0.00019216, 0.00004819, -0.00017736, -0.00024011, -0.00006326, 0.00017231, 0.00024055, 0.00011856, -0.00002550, -0.00006595, -0.00006411, -0.00014474, -0.00028280, -0.00030036, -0.00009828, 0.00017432, 0.00025636, 0.00007954, -0.00013046, -0.00010534, 0.00014206, },
		},
		{ { 0.00021705, 0.00009432, -0.00017660, -0.00038181, -0.00036038, -0.00012752, 0.00014580, 0.00027369, 0.00019691, 0.00001243, -0.00012253, -0.00012135, -0.00002265, 0.00006687, 0.00007780, 0.00003231, 0.00000414, 0.00003231, 0.00007780, 0.00006687, -0.00002265, -0.00012135, -0.00012253, 0.00001243, 0.00019691, 0.00027369, 0.00014580, -0.00012752, -0.00036038, -0.00038181, -0.00017660, 0.00009432, },
		{ -0.00043308, -0.00021933, 0.00006508, 0.00020237, 0.00010120, -0.00013783, -0.00031206, -0.00028311, -0.00008149, 0.00013159, 0.00020387, 0.00011108, -0.00004076, -0.00012019, -0.00008236, 0.00000859, 0.00005646, 0.00002916, -0.00001798, -0.00000914, 0.00006475, 0.00012834, 0.00009542, -0.00003494, -0.00015683, -0.00014223, 0.00003668, 0.00026005, 0.00034200, 0.00018277, -0.00013641, -0.00040263, },
		{ 0.00022765, -0.00011417, -0.00040247, -0.00044666, -0.00024240, 0.00003173, 0.00016337, 0.00007539, -0.00012428, -0.00024692, -0.00018618, -0.00000037, 0.00015026, 0.00014807, 0.00001488, -0.00011942, -0.00014020, -0.00004773, 0.00005484, 0.00006767, -0.00000879, -0.00007791, -0.00004697, 0.00007390, 0.00017178, 0.00013594, -0.00002742, -0.00018188, -0.00017194, 0.00003307, 0.00029205, 0.00039345, },
		{ 0.00000689, 0.00028957, 0.00041541, 0.00026354, -0.00007374, -0.00036162, -0.00041061, -0.00022407, 0.00001708, 0.00011433, 0.00001639, -0.00015105, -0.00020938, -0.00009040, 0.00010578, 0.00020541, 0.00011946, -0.00007737, -0.00021584, -0.00018421, -0.00002500, 0.00010962, 0.00010063, -0.00002877, -0.00013780, -0.00010365, 0.00005791, 0.00020042, 0.00018054, -0.00000246, -0.00019162, -0.00020323, },
		{ 0.00003483, -0.00018403, -0.00022799, -0.00003334, 0.00025482, 0.00040006, 0.00027441, -0.00003264, -0.00029056, -0.00032244, -0.00014852, 0.00004441, 0.00007627, -0.00006623, -0.00022425, -0.00021834, -0.00001866, 0.00022036, 0.00029327, 0.00012667, -0.00015149, -0.00031912, -0.00025017, -0.00002017, 0.00016491, 0.00015210, -0.00002624, -0.00018798, -0.00017060, 0.00001882, 0.00020900, 0.00022091, },
		{ -0.00003770, 0.00019559, 0.00024961, 0.00007735, -0.00016026, -0.00023985, -0.00007734, 0.00019515, 0.00034577, 0.00024794, -0.00000983, -0.00020681, -0.00018980, -0.00000898, 0.00013307, 0.00007319, -0.00015430, -0.00033968, -0.00028520, 0.00000787, 0.00032278, 0.00040316, 0.00017227, -0.00019561, -0.00041666, -0.00033238, -0.00003664, 0.00021229, 0.00021584, -0.00000014, -0.00022027, -0.00023776, },
		{ 0.00004693, -0.00022929, -0.00029501, -0.00010323, 0.00016187, 0.00026128, 0.00011721, -0.00012444, -0.00023520, -0.00011478, 0.00012187, 0.00025750, 0.00017784, -0.00002237, -0.00013201, -0.00002953, 0.00019058, 0.00029465, 0.00012777, -0.00022403, -0.00048321, -0.00041165, -0.00002660, 0.00039173, 0.00051897, 0.00025231, -0.00020201, -0.00049600, -0.00042226, -0.00007464, 0.00024429, 0.00028351, },
		{ -0.00013104, 0.00025552, 0.00034592, 0.00010901, -0.00021324, -0.00033378, -0.00016790, 0.00011291, 0.00025348, 0.00014693, -0.00008305, -0.00021385, -0.00013704, 0.00004844, 0.00014616, 0.00006517, -0.00008229, -0.00008822, 0.00013477, 0.00043536, 0.00053024, 0.00025704, -0.00025004, -0.00063281, -0.00058894, -0.00012969, 0.00040859, 0.00062158, 0.00035674, -0.00016834, -0.00054654, -0.00050947, },
		{ -0.00009801, -0.00056121, -0.00058335, -0.00019978, 0.00024349, 0.00039434, 0.00017766, -0.00017424, -0.00034832, -0.00022184, 0.00005641, 0.00022710, 0.00016067, -0.00004392, -0.00017905, -0.00013858, -0.00001169, 0.00002705, -0.00008144, -0.00019424, -0.00009424, 0.00027629, 0.00070137, 0.00082930, 0.00046881, -0.00020977, -0.00076189, -0.00079885, -0.00029913, 0.00036064, 0.00069188, 0.00047088, },
		{ 0.00024335, 0.00071373, 0.00057762, 0.00000027, -0.00053742, -0.00063448, -0.00027266, 0.00020903, 0.00042183, 0.00024323, -0.00011800, -0.00033659, -0.00025669, 0.00000146, 0.00018621, 0.00015516, -0.00001490, -0.00013703, -0.00011791, -0.00004720, -0.00008240, -0.00024677, -0.00035420, -0.00016219, 0.00036917, 0.00095862, 0.00117027, 0.00075942, -0.00008767, -0.00084350, -0.00101608, -0.00052265, },
		{ -0.00085473, -0.00121186, -0.00077926, 0.00006152, 0.00067673, 0.00066002, 0.00011370, -0.00047752, -0.00065599, -0.00034049, 0.00015620, 0.00042426, 0.00029623, -0.00005293, -0.00030051, -0.00026693, -0.00004277, 0.00013746, 0.00013025, -0.00000269, -0.00009600, -0.00007782, -0.00005070, -0.00016526, -0.00041104, -0.00054966, -0.00029526, 0.00039262, 0.00117523, 0.00152292, 0.00111322, 0.00012164, },
		{ 0.00185216, 0.00150385, 0.00041161, -0.00078058, -0.00135817, -0.00104202, -0.00017091, 0.00057799, 0.00070388, 0.00022728, -0.00038850, -0.00064463, -0.00039446, 0.00009174, 0.00040096, 0.00032883, 0.00001107, -0.00024575, -0.00025065, -0.00006855, 0.00008907, 0.00008894, -0.00001165, -0.00006483, -0.00002503, -0.00001993, -0.00020768, -0.00055277, -0.00076102, -0.00048661, 0.00033431, 0.00132204, },
		{ -0.00071997, 0.00019261, 0.00137688, 0.00212264, 0.00189715, 0.00076381, -0.00061675, -0.00143196, -0.00128163, -0.00043250, 0.00042292, 0.00069999, 0.00032602, -0.00028098, -0.00060120, -0.00042738, 0.00002406, 0.00035481, 0.00033600, 0.00006418, -0.00018078, -0.00020993, -0.00007073, 0.00004948, 0.00003700, -0.00004309, -0.00005162, 0.00003093, 0.00004179, -0.00020107, -0.00065185, -0.00096422, },
		{ -0.00014347, -0.00069258, -0.00113405, -0.00097158, -0.00002278, 0.00132797, 0.00230355, 0.00225547, 0.00114968, -0.00037094, -0.00141864, -0.00147046, -0.00069702, 0.00022482, 0.00064564, 0.00039712, -0.00016770, -0.00053054, -0.00043477, -0.00003797, 0.00029185, 0.00031630, 0.00009811, -0.00011561, -0.00015049, -0.00004762, 0.00002611, -0.00001793, -0.00009496, -0.00006229, 0.00007918, 0.00012609, },
		{ -0.00009959, 0.00010922, 0.00022069, -0.00003994, -0.00066608, -0.00124779, -0.00121337, -0.00029184, 0.00117579, 0.00237294, 0.00254247, 0.00153401, -0.00006226, -0.00131446, -0.00158648, -0.00093691, 0.00000318, 0.00054513, 0.00043180, -0.00006175, -0.00044077, -0.00041548, -0.00008632, 0.00022038, 0.00027209, 0.00010740, -0.00006027, -0.00008084, -0.00000128, 0.00002410, -0.00006745, -0.00016203, },
		{ 0.00004553, -0.00010375, -0.00023661, -0.00016245, 0.00011272, 0.00031135, 0.00009800, -0.00057174, -0.00128858, -0.00141684, -0.00058678, 0.00093333, 0.00232085, 0.00272800, 0.00187931, 0.00028089, -0.00112729, -0.00161630, -0.00112699, -0.00021883, 0.00040930, 0.00042664, 0.00002526, -0.00034224, -0.00037191, -0.00011495, 0.00014975, 0.00020922, 0.00009027, -0.00002325, -0.00001104, 0.00006273, },
		{ -0.00001025, 0.00004886, 0.00013615, 0.00008899, -0.00012071, -0.00030963, -0.00024604, 0.00008485, 0.00038393, 0.00025403, -0.00041748, -0.00124789, -0.00155707, -0.00087577, 0.00062448, 0.00215077, 0.00279217, 0.00215077, 0.00062448, -0.00087577, -0.00155707, -0.00124789, -0.00041748, 0.00025403, 0.00038393, 0.00008485, -0.00024604, -0.00030963, -0.00012071, 0.00008899, 0.00013615, 0.00004886, },
		{ 0.00004553, 0.00006273, -0.00001104, -0.00002325, 0.00009027, 0.00020922, 0.00014975, -0.00011495, -0.00037191, -0.00034224, 0.00002526, 0.00042664, 0.00040930, -0.00021883, -0.00112699, -0.00161630, -0.00112729, 0.00028089, 0.00187931, 0.00272800, 0.00232085, 0.00093333, -0.00058678, -0.00141684, -0.00128858, -0.00057174, 0.00009800, 0.00031135, 0.00011272, -0.00016245, -0.00023661, -0.00010375, },
		{ -0.00009959, -0.00016203, -0.00006745, 0.00002410, -0.00000128, -0.00008084, -0.00006027, 0.00010740, 0.00027209, 0.00022038, -0.00008632, -0.00041548, -0.00044077, -0.00006175, 0.00043180, 0.00054513, 0.00000318, -0.00093691, -0.00158648, -0.00131446, -0.00006226, 0.00153401, 0.00254247, 0.00237294, 0.00117579, -0.00029184, -0.00121337, -0.00124779, -0.00066608, -0.00003994, 0.00022069, 0.00010922, },
		{ -0.00014347, 0.00012609, 0.00007918, -0.00006229, -0.00009496, -0.00001793, 0.00002611, -0.00004762, -0.00015049, -0.00011561, 0.00009811, 0.00031630, 0.00029185, -0.00003797, -0.00043477, -0.00053054, -0.00016770, 0.00039712, 0.00064564, 0.00022482, -0.00069702, -0.00147046, -0.00141864, -0.00037094, 0.00114968, 0.00225547, 0.00230355, 0.00132797, -0.00002278, -0.00097158, -0.00113405, -0.00069258, },
		{ -0.00071997, -0.00096422, -0.00065185, -0.00020107, 0.00004179, 0.00003093, -0.00005162, -0.00004309, 0.00003700, 0.00004948, -0.00007073, -0.00020993, -0.00018078, 0.00006418, 0.00033600, 0.00035481, 0.00002406, -0.00042738, -0.00060120, -0.00028098, 0.00032602, 0.00069999, 0.00042292, -0.00043250, -0.00128163, -0.00143196, -0.00061675, 0.00076381, 0.00189715, 0.00212264, 0.00137688, 0.00019261, },
		{ 0.00185216, 0.00132204, 0.00033431, -0.00048661, -0.00076102, -0.00055277, -0.00020768, -0.00001993, -0.00002503, -0.00006483, -0.00001165, 0.00008894, 0.00008907, -0.00006855, -0.00025065, -0.00024575, 0.00001107, 0.00032883, 0.00040096, 0.00009174, -0.00039446, -0.00064463, -0.00038850, 0.00022728, 0.00070388, 0.00057799, -0.00017091, -0.00104202, -0.00135817, -0.00078058, 0.00041161, 0.00150385, },
		{ -0.00085473, 0.00012164, 0.00111322, 0.00152292, 0.00117523, 0.00039262, -0.00029526, -0.00054966, -0.00041104, -0.00016526, -0.00005070, -0.00007782, -0.00009600, -0.00000269, 0.00013025, 0.00013746, -0.00004277, -0.00026693, -0.00030051, -0.00005293, 0.00029623, 0.00042426, 0.00015620, -0.00034049, -0.00065599, -0.00047752, 0.00011370, 0.00066002, 0.00067673, 0.00006152, -0.00077926, -0.00121186, },
		{ 0.00024335, -0.00052265, -0.00101608, -0.00084350, -0.00008767, 0.00075942, 0.00117027, 0.00095862, 0.00036917, -0.00016219, -0.00035420, -0.00024677, -0.00008240, -0.00004720, -0.00011791, -0.00013703, -0.00001490, 0.00015516, 0.00018621, 0.00000146, -0.00025669, -0.00033659, -0.00011800, 0.00024323, 0.00042183, 0.00020903, -0.00027266, -0.00063448, -0.00053742, 0.00000027, 0.00057762, 0.00071373, },
		{ -0.00009801, 0.00047088, 0.00069188, 0.00036064, -0.00029913, -0.00079885, -0.00076189, -0.00020977, 0.00046881, 0.00082930, 0.00070137, 0.00027629, -0.00009424, -0.00019424, -0.00008144, 0.00002705, -0.00001169, -0.00013858, -0.00017905, -0.00004392, 0.00016067, 0.00022710, 0.00005641, -0.00022184, -0.00034832, -0.00017424, 0.00017766, 0.00039434, 0.00024349, -0.00019978, -0.00058335, -0.00056121, },
		{ -0.00013104, -0.00050947, -0.00054654, -0.00016834, 0.00035674, 0.00062158, 0.00040859, -0.00012969, -0.00058894, -0.00063281, -0.00025004, 0.00025704, 0.00053024, 0.00043536, 0.00013477, -0.00008822, -0.00008229, 0.00006517, 0.00014616, 0.00004844, -0.00013704, -0.00021385, -0.00008305, 0.00014693, 0.00025348, 0.00011291, -0.00016790, -0.00033378, -0.00021324, 0.00010901, 0.00034592, 0.00025552, },
		{ 0.00004693, 0.00028351, 0.00024429, -0.00007464, -0.00042226, -0.00049600, -0.00020201, 0.00025231, 0.00051897, 0.00039173, -0.00002660, -0.00041165, -0.00048321, -0.00022403, 0.00012777, 0.00029465, 0.00019058, -0.00002953, -0.00013201, -0.00002237, 0.00017784, 0.00025750, 0.00012187, -0.00011478, -0.00023520, -0.00012444, 0.00011721, 0.00026128, 0.00016187, -0.00010323, -0.00029501, -0.00022929, },
		{ -0.00003770, -0.00023776, -0.00022027, -0.00000014, 0.00021584, 0.00021229, -0.00003664, -0.00033238, -0.00041666, -0.00019561, 0.00017227, 0.00040316, 0.00032278, 0.00000787, -0.00028520, -0.00033968, -0.00015430, 0.00007319, 0.00013307, -0.00000898, -0.00018980, -0.00020681, -0.00000983, 0.00024794, 0.00034577, 0.00019515, -0.00007734, -0.00023985, -0.00016026, 0.00007735, 0.00024961, 0.00019559, },
		{ 0.00003483, 0.00022091, 0.00020900, 0.00001882, -0.00017060, -0.00018798, -0.00002624, 0.00015210, 0.00016491, -0.00002017, -0.00025017, -0.00031912, -0.00015149, 0.00012667, 0.00029327, 0.00022036, -0.00001866, -0.00021834, -0.00022425, -0.00006623, 0.00007627, 0.00004441, -0.00014852, -0.00032244, -0.00029056, -0.00003264, 0.00027441, 0.00040006, 0.00025482, -0.00003334, -0.00022799, -0.00018403, },
		{ 0.00000689, -0.00020323, -0.00019162, -0.00000246, 0.00018054, 0.00020042, 0.00005791, -0.00010365, -0.00013780, -0.00002877, 0.00010063, 0.00010962, -0.00002500, -0.00018421, -0.00021584, -0.00007737, 0.00011946, 0.00020541, 0.00010578, -0.00009040, -0.00020938, -0.00015105, 0.00001639, 0.00011433, 0.00001708, -0.00022407, -0.00041061, -0.00036162, -0.00007374, 0.00026354, 0.00041541, 0.00028957, },
		{ 0.00022765, 0.00039345, 0.00029205, 0.00003307, -0.00017194, -0.00018188, -0.00002742, 0.00013594, 0.00017178, 0.00007390, -0.00004697, -0.00007791, -0.00000879, 0.00006767, 0.00005484, -0.00004773, -0.00014020, -0.00011942, 0.00001488, 0.00014807, 0.00015026, -0.00000037, -0.00018618, -0.00024692, -0.00012428, 0.00007539, 0.00016337, 0.00003173, -0.00024240, -0.00044666, -0.00040247, -0.00011417, },
		{ -0.00043308, -0.00040263, -0.00013641, 0.00018277, 0.00034200, 0.00026005, 0.00003668, -0.00014223, -0.00015683, -0.00003494, 0.00009542, 0.00012834, 0.00006475, -0.00000914, -0.00001798, 0.00002916, 0.00005646, 0.00000859, -0.00008236, -0.00012019, -0.00004076, 0.00011108, 0.00020387, 0.00013159, -0.00008149, -0.00028311, -0.00031206, -0.00013783, 0.00010120, 0.00020237, 0.00006508, -0.00021933, },
		},
		{ { -0.00009944, -0.00007222, -0.00000569, 0.00006349, 0.00009817, 0.00008194, 0.00002800, -0.00002923, -0.00005485, -0.00003428, 0.00001809, 0.00006755, 0.00007986, 0.00004196, -0.00002961, -0.00009762, -0.00012534, -0.00009762, -0.00002961, 0.00004196, 0.00007986, 0.00006755, 0.00001809, -0.00003428, -0.00005485, -0.00002923, 0.00002800, 0.00008194, 0.00009817, 0.00006349, -0.00000569, -0.00007222, },
		{ 0.00006422, -0.00000335, -0.00007817, -0.00011868, -0.00009914, -0.00002571, 0.00006548, 0.00012754, 0.00012870, 0.00007006, -0.00001494, -0.00007934, -0.00008903, -0.00004174, 0.00003214, 0.00008724, 0.00008920, 0.00003405, -0.00004991, -0.00011848, -0.00013660, -0.00009750, -0.00002511, 0.00004094, 0.00006877, 0.00005199, 0.00001159, -0.00001868, -0.00001436, 0.00002396, 0.00007068, 0.00009120, },
		{ 0.00002901, 0.00008005, 0.00011508, 0.00010028, 0.00002831, -0.00007290, -0.00015281, -0.00016583, -0.00009824, 0.00002150, 0.00013592, 0.00018858, 0.00015449, 0.00005377, -0.00005926, -0.00012579, -0.00011435, -0.00003724, 0.00005748, 0.00011352, 0.00009767, 0.00001731, -0.00008522, -0.00015736, -0.00016561, -0.00011246, -0.00003204, 0.00003209, 0.00005287, 0.00003413, 0.00000500, -0.00000134, },
		{ 0.00002754, 0.00000490, -0.00001177, 0.00000988, 0.00007136, 0.00013898, 0.00016343, 0.00011170, -0.00000803, -0.00014493, -0.00023006, -0.00021411, -0.00009678, 0.00006957, 0.00020549, 0.00024587, 0.00017447, 0.00003234, -0.00010476, -0.00016740, -0.00012937, -0.00002020, 0.00009315, 0.00014430, 0.00010372, -0.00000698, -0.00012964, -0.00020359, -0.00019881, -0.00012958, -0.00004171, 0.00001745, },
		{ -0.00021634, -0.00013100, -0.00004279, -0.00000030, -0.00001241, -0.00004559, -0.00004926, 0.00000707, 0.00010885, 0.00020051, 0.00021748, 0.00012761, -0.00004326, -0.00021861, -0.00030952, -0.00026327, -0.00009481, 0.00011784, 0.00027288, 0.00029793, 0.00018785, 0.00000629, -0.00014970, -0.00020237, -0.00013398, 0.00000654, 0.00013348, 0.00017316, 0.00010360, -0.00003648, -0.00017307, -0.00024032, },
		{ 0.00009211, -0.00006810, -0.00020285, -0.00024751, -0.00019560, -0.00009728, -0.00002335, -0.00001826, -0.00007163, -0.00012601, -0.00011635, -0.00001628, 0.00013818, 0.00026392, 0.00027793, 0.00014922, -0.00007594, -0.00029230, -0.00038950, -0.00031217, -0.00009241, 0.00016471, 0.00033552, 0.00034252, 0.00019413, -0.00002252, -0.00019083, -0.00022836, -0.00012923, 0.00003718, 0.00016903, 0.00019045, },
		{ -0.00011721, 0.00006392, 0.00018810, 0.00018490, 0.00006352, -0.00009843, -0.00020596, -0.00020548, -0.00011546, -0.00001107, 0.00002685, -0.00003381, -0.00015305, -0.00024181, -0.00021837, -0.00006405, 0.00015695, 0.00032775, 0.00034396, 0.00017655, -0.00010478, -0.00036342, -0.00046693, -0.00035868, -0.00008968, 0.00020765, 0.00038960, 0.00037674, 0.00019316, -0.00005114, -0.00022381, -0.00024273, },
		{ 0.00018512, -0.00007592, -0.00024376, -0.00024292, -0.00010057, 0.00007826, 0.00017878, 0.00014572, 0.00001274, -0.00012418, -0.00017129, -0.00009866, 0.00003960, 0.00013937, 0.00011428, -0.00004493, -0.00025706, -0.00039377, -0.00035569, -0.00013657, 0.00016400, 0.00038976, 0.00041304, 0.00020827, -0.00012872, -0.00042850, -0.00053746, -0.00039981, -0.00008671, 0.00024338, 0.00043051, 0.00039730, },
		{ -0.00008362, 0.00026830, 0.00045333, 0.00040102, 0.00017048, -0.00009301, -0.00024597, -0.00022681, -0.00008218, 0.00007246, 0.00013108, 0.00006463, -0.00006349, -0.00014266, -0.00009178, 0.00008120, 0.00027603, 0.00035746, 0.00024013, -0.00005055, -0.00038087, -0.00057713, -0.00052302, -0.00023033, 0.00015951, 0.00044696, 0.00048100, 0.00024179, -0.00014691, -0.00048344, -0.00059582, -0.00043201, },
		{ -0.00015879, -0.00052390, -0.00063644, -0.00045161, -0.00008053, 0.00027890, 0.00045358, 0.00038523, 0.00014994, -0.00009896, -0.00022674, -0.00019316, -0.00006462, 0.00004093, 0.00003895, -0.00006225, -0.00016574, -0.00015210, 0.00003406, 0.00033285, 0.00058913, 0.00063671, 0.00039956, -0.00005069, -0.00051851, -0.00078163, -0.00070965, -0.00033818, 0.00014489, 0.00049581, 0.00054246, 0.00027357, },
		{ 0.00012260, 0.00053257, 0.00059144, 0.00029962, -0.00016406, -0.00054581, -0.00065415, -0.00045533, -0.00007756, 0.00027234, 0.00042791, 0.00034831, 0.00012444, -0.00009132, -0.00018405, -0.00014192, -0.00004983, -0.00001865, -0.00009837, -0.00023328, -0.00029131, -0.00015192, 0.00020141, 0.00064470, 0.00096248, 0.00096075, 0.00058189, -0.00004636, -0.00066119, -0.00099239, -0.00090051, -0.00045017, },
		{ -0.00079820, -0.00119155, -0.00107793, -0.00055481, 0.00009570, 0.00055372, 0.00062213, 0.00031603, -0.00016268, -0.00054591, -0.00064492, -0.00044077, -0.00007483, 0.00024692, 0.00037475, 0.00029008, 0.00009505, -0.00006899, -0.00011801, -0.00007445, -0.00003880, -0.00010474, -0.00027558, -0.00044079, -0.00043415, -0.00014279, 0.00039928, 0.00099560, 0.00136933, 0.00130476, 0.00077172, -0.00003946, },
		{ 0.00177544, 0.00163833, 0.00095077, -0.00003237, -0.00091809, -0.00136043, -0.00122391, -0.00064059, 0.00006744, 0.00055642, 0.00062972, 0.00031962, -0.00015483, -0.00052224, -0.00060656, -0.00040687, -0.00007243, 0.00020240, 0.00029477, 0.00021203, 0.00006291, -0.00003255, -0.00003119, 0.00000645, -0.00003153, -0.00021184, -0.00048161, -0.00067159, -0.00058526, -0.00012644, 0.00061166, 0.00135698, },
		{ -0.00073363, -0.00010547, 0.00081941, 0.00169631, 0.00214328, 0.00192926, 0.00110028, -0.00002756, -0.00101011, -0.00148196, -0.00132252, -0.00069760, 0.00004079, 0.00053895, 0.00061117, 0.00030839, -0.00014090, -0.00047453, -0.00053917, -0.00035422, -0.00007046, 0.00014025, 0.00019104, 0.00011743, 0.00002925, 0.00001575, 0.00007160, 0.00009668, -0.00002703, -0.00033108, -0.00070064, -0.00090826, },
		{ -0.00002361, -0.00045124, -0.00091390, -0.00113110, -0.00086736, -0.00008288, 0.00100267, 0.00198118, 0.00243678, 0.00214772, 0.00120352, -0.00002717, -0.00106547, -0.00154289, -0.00136205, -0.00071888, 0.00001802, 0.00050098, 0.00056567, 0.00028196, -0.00012147, -0.00040436, -0.00044533, -0.00028507, -0.00006900, 0.00006360, 0.00006893, 0.00001109, -0.00000474, 0.00007221, 0.00018362, 0.00019114, },
		{ -0.00003788, 0.00013211, 0.00029694, 0.00028417, -0.00001918, -0.00056023, -0.00110203, -0.00132053, -0.00097516, -0.00006166, 0.00114342, 0.00218343, 0.00262607, 0.00227024, 0.00124813, -0.00003262, -0.00107857, -0.00153568, -0.00133670, -0.00070144, 0.00000045, 0.00044376, 0.00049497, 0.00024160, -0.00009724, -0.00031516, -0.00033002, -0.00020328, -0.00006810, -0.00002300, -0.00006432, -0.00010094, },
		{ -0.00006779, -0.00011397, -0.00020025, -0.00021195, -0.00006906, 0.00019024, 0.00040330, 0.00037010, -0.00001169, -0.00064667, -0.00124746, -0.00145951, -0.00104766, -0.00004434, 0.00122783, 0.00228295, 0.00269146, 0.00228295, 0.00122783, -0.00004434, -0.00104766, -0.00145951, -0.00124746, -0.00064667, -0.00001169, 0.00037010, 0.00040330, 0.00019024, -0.00006906, -0.00021195, -0.00020025, -0.00011397, },
		{ -0.00003788, -0.00010094, -0.00006432, -0.00002300, -0.00006810, -0.00020328, -0.00033002, -0.00031516, -0.00009724, 0.00024160, 0.00049497, 0.00044376, 0.00000045, -0.00070144, -0.00133670, -0.00153568, -0.00107857, -0.00003262, 0.00124813, 0.00227024, 0.00262607, 0.00218343, 0.00114342, -0.00006166, -0.00097516, -0.00132053, -0.00110203, -0.00056023, -0.00001918, 0.00028417, 0.00029694, 0.00013211, },
		{ -0.00002361, 0.00019114, 0.00018362, 0.00007221, -0.00000474, 0.00001109, 0.00006893, 0.00006360, -0.00006900, -0.00028507, -0.00044533, -0.00040436, -0.00012147, 0.00028196, 0.00056567, 0.00050098, 0.00001802, -0.00071888, -0.00136205, -0.00154289, -0.00106547, -0.00002717, 0.00120352, 0.00214772, 0.00243678, 0.00198118, 0.00100267, -0.00008288, -0.00086736, -0.00113110, -0.00091390, -0.00045124, },
		{ -0.00073363, -0.00090826, -0.00070064, -0.00033108, -0.00002703, 0.00009668, 0.00007160, 0.00001575, 0.00002925, 0.00011743, 0.00019104, 0.00014025, -0.00007046, -0.00035422, -0.00053917, -0.00047453, -0.00014090, 0.00030839, 0.00061117, 0.00053895, 0.00004079, -0.00069760, -0.00132252, -0.00148196, -0.00101011, -0.00002756, 0.00110028, 0.00192926, 0.00214328, 0.00169631, 0.00081941, -0.00010547, },
		{ 0.00177544, 0.00135698, 0.00061166, -0.00012644, -0.00058526, -0.00067159, -0.00048161, -0.00021184, -0.00003153, 0.00000645, -0.00003119, -0.00003255, 0.00006291, 0.00021203, 0.00029477, 0.00020240, -0.00007243, -0.00040687, -0.00060656, -0.00052224, -0.00015483, 0.00031962, 0.00062972, 0.00055642, 0.00006744, -0.00064059, -0.00122391, -0.00136043, -0.00091809, -0.00003237, 0.00095077, 0.00163833, },
		{ -0.00079820, -0.00003946, 0.00077172, 0.00130476, 0.00136933, 0.00099560, 0.00039928, -0.00014279, -0.00043415, -0.00044079, -0.00027558, -0.00010474, -0.00003880, -0.00007445, -0.00011801, -0.00006899, 0.00009505, 0.00029008, 0.00037475, 0.00024692, -0.00007483, -0.00044077, -0.00064492, -0.00054591, -0.00016268, 0.00031603, 0.00062213, 0.00055372, 0.00009570, -0.00055481, -0.00107793, -0.00119155, },
		{ 0.00012260, -0.00045017, -0.00090051, -0.00099239, -0.00066119, -0.00004636, 0.00058189, 0.00096075, 0.00096248, 0.00064470, 0.00020141, -0.00015192, -0.00029131, -0.00023328, -0.00009837, -0.00001865, -0.00004983, -0.00014192, -0.00018405, -0.00009132, 0.00012444, 0.00034831, 0.00042791, 0.00027234, -0.00007756, -0.00045533, -0.00065415, -0.00054581, -0.00016406, 0.00029962, 0.00059144, 0.00053257, },
		{ -0.00015879, 0.00027357, 0.00054246, 0.00049581, 0.00014489, -0.00033818, -0.00070965, -0.00078163, -0.00051851, -0.00005069, 0.00039956, 0.00063671, 0.00058913, 0.00033285, 0.00003406, -0.00015210, -0.00016574, -0.00006225, 0.00003895, 0.00004093, -0.00006462, -0.00019316, -0.00022674, -0.00009896, 0.00014994, 0.00038523, 0.00045358, 0.00027890, -0.00008053, -0.00045161, -0.00063644, -0.00052390, },
		{ -0.00008362, -0.00043201, -0.00059582, -0.00048344, -0.00014691, 0.00024179, 0.00048100, 0.00044696, 0.00015951, -0.00023033, -0.00052302, -0.00057713, -0.00038087, -0.00005055, 0.00024013, 0.00035746, 0.00027603, 0.00008120, -0.00009178, -0.00014266, -0.00006349, 0.00006463, 0.00013108, 0.00007246, -0.00008218, -0.00022681, -0.00024597, -0.00009301, 0.00017048, 0.00040102, 0.00045333, 0.00026830, },
		{ 0.00018512, 0.00039730, 0.00043051, 0.00024338, -0.00008671, -0.00039981, -0.00053746, -0.00042850, -0.00012872, 0.00020827, 0.00041304, 0.00038976, 0.00016400, -0.00013657, -0.00035569, -0.00039377, -0.00025706, -0.00004493, 0.00011428, 0.00013937, 0.00003960, -0.00009866, -0.00017129, -0.00012418, 0.00001274, 0.00014572, 0.00017878, 0.00007826, -0.00010057, -0.00024292, -0.00024376, -0.00007592, },
		{ -0.00011721, -0.00024273, -0.00022381, -0.00005114, 0.00019316, 0.00037674, 0.00038960, 0.00020765, -0.00008968, -0.00035868, -0.00046693, -0.00036342, -0.00010478, 0.00017655, 0.00034396, 0.00032775, 0.00015695, -0.00006405, -0.00021837, -0.00024181, -0.00015305, -0.00003381, 0.00002685, -0.00001107, -0.00011546, -0.00020548, -0.00020596, -0.00009843, 0.00006352, 0.00018490, 0.00018810, 0.00006392, },
		{ 0.00009211, 0.00019045, 0.00016903, 0.00003718, -0.00012923, -0.00022836, -0.00019083, -0.00002252, 0.00019413, 0.00034252, 0.00033552, 0.00016471, -0.00009241, -0.00031217, -0.00038950, -0.00029230, -0.00007594, 0.00014922, 0.00027793, 0.00026392, 0.00013818, -0.00001628, -0.00011635, -0.00012601, -0.00007163, -0.00001826, -0.00002335, -0.00009728, -0.00019560, -0.00024751, -0.00020285, -0.00006810, },
		{ -0.00021634, -0.00024032, -0.00017307, -0.00003648, 0.00010360, 0.00017316, 0.00013348, 0.00000654, -0.00013398, -0.00020237, -0.00014970, 0.00000629, 0.00018785, 0.00029793, 0.00027288, 0.00011784, -0.00009481, -0.00026327, -0.00030952, -0.00021861, -0.00004326, 0.00012761, 0.00021748, 0.00020051, 0.00010885, 0.00000707, -0.00004926, -0.00004559, -0.00001241, -0.00000030, -0.00004279, -0.00013100, },
		{ 0.00002754, 0.00001745, -0.00004171, -0.00012958, -0.00019881, -0.00020359, -0.00012964, -0.00000698, 0.00010372, 0.00014430, 0.00009315, -0.00002020, -0.00012937, -0.00016740, -0.00010476, 0.00003234, 0.00017447, 0.00024587, 0.00020549, 0.00006957, -0.00009678, -0.00021411, -0.00023006, -0.00014493, -0.00000803, 0.00011170, 0.00016343, 0.00013898, 0.00007136, 0.00000988, -0.00001177, 0.00000490, },
		{ 0.00002901, -0.00000134, 0.00000500, 0.00003413, 0.00005287, 0.00003209, -0.00003204, -0.00011246, -0.00016561, -0.00015736, -0.00008522, 0.00001731, 0.00009767, 0.00011352, 0.00005748, -0.00003724, -0.00011435, -0.00012579, -0.00005926, 0.00005377, 0.00015449, 0.00018858, 0.00013592, 0.00002150, -0.00009824, -0.00016583, -0.00015281, -0.00007290, 0.00002831, 0.00010028, 0.00011508, 0.00008005, },
		{ 0.00006422, 0.00009120, 0.00007068, 0.00002396, -0.00001436, -0.00001868, 0.00001159, 0.00005199, 0.00006877, 0.00004094, -0.00002511, -0.00009750, -0.00013660, -0.00011848, -0.00004991, 0.00003405, 0.00008920, 0.00008724, 0.00003214, -0.00004174, -0.00008903, -0.00007934, -0.00001494, 0.00007006, 0.00012870, 0.00012754, 0.00006548, -0.00002571, -0.00009914, -0.00011868, -0.00007817, -0.00000335, },
		},
		{ { -0.00011383, -0.00009338, -0.00003789, 0.00003679, 0.00010973, 0.00016115, 0.00017827, 0.00015871, 0.00011059, 0.00004933, -0.00000765, -0.00004653, -0.00006117, -0.00005430, -0.00003563, -0.00001756, -0.00001025, -0.00001756, -0.00003563, -0.00005430, -0.00006117, -0.00004653, -0.00000765, 0.00004933, 0.00011059, 0.00015871, 0.00017827, 0.00016115, 0.00010973, 0.00003679, -0.00003789, -0.00009338, },
		{ 0.00021564, 0.00016559, 0.00008984, 0.00000837, -0.00005823, -0.00009463, -0.00009491, -0.00006363, -0.00001378, 0.00003783, 0.00007600, 0.00009172, 0.00008432, 0.00006078, 0.00003238, 0.00001019, 0.00000084, 0.00000428, 0.00001406, 0.00002026, 0.00001375, -0.00000966, -0.00004681, -0.00008742, -0.00011708, -0.00012206, -0.00009427, -0.00003472, 0.00004578, 0.00012914, 0.00019496, 0.00022648, },
		{ -0.00005861, 0.00002665, 0.00011222, 0.00017940, 0.00021426, 0.00021121, 0.00017412, 0.00011477, 0.00004900, -0.00000792, -0.00004541, -0.00005996, -0.00005504, -0.00003883, -0.00002072, -0.00000779, -0.00000256, -0.00000271, -0.00000276, 0.00000295, 0.00001727, 0.00003829, 0.00005941, 0.00007128, 0.00006529, 0.00003709, -0.00001094, -0.00006928, -0.00012338, -0.00015756, -0.00015969, -0.00012514, },
		{ -0.00008715, -0.00013925, -0.00016530, -0.00015873, -0.00012099, -0.00006092, 0.00000785, 0.00007070, 0.00011590, 0.00013731, 0.00013528, 0.00011554, 0.00008664, 0.00005685, 0.00003167, 0.00001266, -0.00000207, -0.00001602, -0.00003202, -0.00005019, -0.00006735, -0.00007778, -0.00007541, -0.00005644, -0.00002160, 0.00002306, 0.00006712, 0.00009847, 0.00010679, 0.00008693, 0.00004097, -0.00002184, },
		{ 0.00012638, 0.00011997, 0.00008009, 0.00001744, -0.00005121, -0.00010808, -0.00013964, -0.00014034, -0.00011371, -0.00007063, -0.00002534, 0.00000934, 0.00002605, 0.00002474, 0.00001193, -0.00000230, -0.00000849, -0.00000163, 0.00001686, 0.00003953, 0.00005570, 0.00005559, 0.00003434, -0.00000558, -0.00005443, -0.00009795, -0.00012185, -0.00011656, -0.00008080, -0.00002250, 0.00004313, 0.00009797, },
		{ -0.00012661, -0.00006966, 0.00001136, 0.00009139, 0.00014526, 0.00015597, 0.00012026, 0.00004953, -0.00003394, -0.00010447, -0.00014143, -0.00013596, -0.00009364, -0.00003235, 0.00002409, 0.00005413, 0.00004589, 0.00000123, -0.00006495, -0.00012983, -0.00017055, -0.00017214, -0.00013268, -0.00006418, 0.00001143, 0.00006967, 0.00009178, 0.00007116, 0.00001571, -0.00005472, -0.00011490, -0.00014270, },
		{ 0.00007626, 0.00000613, -0.00008045, -0.00014740, -0.00016601, -0.00012506, -0.00003476, 0.00007659, 0.00017226, 0.00021973, 0.00020194, 0.00012339, 0.00000892, -0.00010448, -0.00018000, -0.00019348, -0.00014169, -0.00004365, 0.00006558, 0.00014645, 0.00016815, 0.00011896, 0.00001038, -0.00012652, -0.00025114, -0.00032652, -0.00033138, -0.00026675, -0.00015536, -0.00003374, 0.00006006, 0.00009910, },
		{ -0.00035228, -0.00020837, -0.00005604, 0.00005469, 0.00009117, 0.00004879, -0.00004825, -0.00015601, -0.00022669, -0.00022580, -0.00014460, -0.00000371, 0.00015360, 0.00027603, 0.00032190, 0.00027381, 0.00014519, -0.00002353, -0.00017797, -0.00026778, -0.00026351, -0.00016674, -0.00000995, 0.00015381, 0.00026824, 0.00029216, 0.00021307, 0.00005126, -0.00014680, -0.00032362, -0.00042924, -0.00043716, },
		{ 0.00010340, -0.00013256, -0.00033821, -0.00045396, -0.00045143, -0.00034197, -0.00017177, -0.00000522, 0.00009765, 0.00010192, 0.00000974, -0.00014069, -0.00028867, -0.00037244, -0.00035111, -0.00021966, -0.00001169, 0.00021113, 0.00037935, 0.00043910, 0.00037051, 0.00019499, -0.00003122, -0.00023520, -0.00035120, -0.00034233, -0.00021274, -0.00000630, 0.00020775, 0.00035758, 0.00039225, 0.00029823, },
		{ -0.00003527, 0.00021266, 0.00038879, 0.00043512, 0.00033827, 0.00013311, -0.00010997, -0.00030962, -0.00040164, -0.00036083, -0.00020930, -0.00000844, 0.00016278, 0.00023566, 0.00017621, -0.00000218, -0.00024239, -0.00046227, -0.00058240, -0.00055328, -0.00037304, -0.00008923, 0.00021578, 0.00045122, 0.00054744, 0.00047858, 0.00027104, -0.00000526, -0.00025987, -0.00041140, -0.00041389, -0.00027128, },
		{ 0.00005181, -0.00024420, -0.00043308, -0.00045973, -0.00032573, -0.00008659, 0.00016907, 0.00035106, 0.00039895, 0.00030246, 0.00010412, -0.00011670, -0.00027266, -0.00030021, -0.00018317, 0.00004027, 0.00028950, 0.00046888, 0.00050097, 0.00035454, 0.00005751, -0.00030977, -0.00063919, -0.00082933, -0.00081851, -0.00060539, -0.00025026, 0.00014364, 0.00046262, 0.00061832, 0.00057504, 0.00035975, },
		{ -0.00000997, 0.00039678, 0.00062764, 0.00063405, 0.00043901, 0.00012583, -0.00019202, -0.00040966, -0.00046482, -0.00035614, -0.00013982, 0.00009346, 0.00025302, 0.00028170, 0.00017670, -0.00000960, -0.00018912, -0.00027107, -0.00019683, 0.00003571, 0.00036719, 0.00069350, 0.00089921, 0.00089647, 0.00065659, 0.00022312, -0.00029831, -0.00077055, -0.00106654, -0.00110836, -0.00089120, -0.00048453, },
		{ -0.00081152, -0.00124069, -0.00136880, -0.00118315, -0.00075723, -0.00022603, 0.00025653, 0.00056439, 0.00063527, 0.00048475, 0.00019480, -0.00011849, -0.00034621, -0.00042295, -0.00034585, -0.00017055, 0.00001389, 0.00012145, 0.00010382, -0.00002951, -0.00021304, -0.00034866, -0.00034269, -0.00014343, 0.00023419, 0.00070631, 0.00114200, 0.00140298, 0.00138830, 0.00106946, 0.00050384, -0.00017859, },
		{ 0.00190161, 0.00190921, 0.00154491, 0.00087552, 0.00005098, -0.00073679, -0.00130791, -0.00154493, -0.00142381, -0.00101646, -0.00046461, 0.00006597, 0.00043589, 0.00057113, 0.00047870, 0.00023525, -0.00004617, -0.00025887, -0.00033960, -0.00028743, -0.00015832, -0.00003834, -0.00000568, -0.00009583, -0.00028301, -0.00048527, -0.00059208, -0.00050544, -0.00017992, 0.00035262, 0.00098352, 0.00155413, },
		{ -0.00074920, -0.00032072, 0.00035104, 0.00113586, 0.00184992, 0.00230674, 0.00237239, 0.00200750, 0.00128228, 0.00035935, -0.00055076, -0.00124796, -0.00159699, -0.00156128, -0.00120659, -0.00067439, -0.00013446, 0.00026753, 0.00045052, 0.00041459, 0.00022957, 0.00000140, -0.00017082, -0.00023149, -0.00018602, -0.00009355, -0.00003819, -0.00008989, -0.00026990, -0.00053425, -0.00078237, -0.00088901, },
		{ -0.00010334, -0.00037543, -0.00072993, -0.00104783, -0.00118812, -0.00103621, -0.00054873, 0.00022135, 0.00112896, 0.00197122, 0.00254230, 0.00269230, 0.00237185, 0.00164832, 0.00068822, -0.00028900, -0.00107167, -0.00151246, -0.00156403, -0.00128355, -0.00080635, -0.00029829, 0.00009676, 0.00029742, 0.00030099, 0.00017161, 0.00000659, -0.00010599, -0.00012245, -0.00005888, 0.00001869, 0.00002517, },
		{ -0.00008218, -0.00003699, 0.00006897, 0.00015945, 0.00014475, -0.00003697, -0.00038582, -0.00082739, -0.00122732, -0.00143013, -0.00131036, -0.00081912, -0.00000988, 0.00096638, 0.00189695, 0.00256425, 0.00280655, 0.00256425, 0.00189695, 0.00096638, -0.00000988, -0.00081912, -0.00131036, -0.00143013, -0.00122732, -0.00082739, -0.00038582, -0.00003697, 0.00014475, 0.00015945, 0.00006897, -0.00003699, },
		{ -0.00010334, 0.00002517, 0.00001869, -0.00005888, -0.00012245, -0.00010599, 0.00000659, 0.00017161, 0.00030099, 0.00029742, 0.00009676, -0.00029829, -0.00080635, -0.00128355, -0.00156403, -0.00151246, -0.00107167, -0.00028900, 0.00068822, 0.00164832, 0.00237185, 0.00269230, 0.00254230, 0.00197122, 0.00112896, 0.00022135, -0.00054873, -0.00103621, -0.00118812, -0.00104783, -0.00072993, -0.00037543, },
		{ -0.00074920, -0.00088901, -0.00078237, -0.00053425, -0.00026990, -0.00008989, -0.00003819, -0.00009355, -0.00018602, -0.00023149, -0.00017082, 0.00000140, 0.00022957, 0.00041459, 0.00045052, 0.00026753, -0.00013446, -0.00067439, -0.00120659, -0.00156128, -0.00159699, -0.00124796, -0.00055076, 0.00035935, 0.00128228, 0.00200750, 0.00237239, 0.00230674, 0.00184992, 0.00113586, 0.00035104, -0.00032072, },
		{ 0.00190161, 0.00155413, 0.00098352, 0.00035262, -0.00017992, -0.00050544, -0.00059208, -0.00048527, -0.00028301, -0.00009583, -0.00000568, -0.00003834, -0.00015832, -0.00028743, -0.00033960, -0.00025887, -0.00004617, 0.00023525, 0.00047870, 0.00057113, 0.00043589, 0.00006597, -0.00046461, -0.00101646, -0.00142381, -0.00154493, -0.00130791, -0.00073679, 0.00005098, 0.00087552, 0.00154491, 0.00190921, },
		{ -0.00081152, -0.00017859, 0.00050384, 0.00106946, 0.00138830, 0.00140298, 0.00114200, 0.00070631, 0.00023419, -0.00014343, -0.00034269, -0.00034866, -0.00021304, -0.00002951, 0.00010382, 0.00012145, 0.00001389, -0.00017055, -0.00034585, -0.00042295, -0.00034621, -0.00011849, 0.00019480, 0.00048475, 0.00063527, 0.00056439, 0.00025653, -0.00022603, -0.00075723, -0.00118315, -0.00136880, -0.00124069, },
		{ -0.00000997, -0.00048453, -0.00089120, -0.00110836, -0.00106654, -0.00077055, -0.00029831, 0.00022312, 0.00065659, 0.00089647, 0.00089921, 0.00069350, 0.00036719, 0.00003571, -0.00019683, -0.00027107, -0.00018912, -0.00000960, 0.00017670, 0.00028170, 0.00025302, 0.00009346, -0.00013982, -0.00035614, -0.00046482, -0.00040966, -0.00019202, 0.00012583, 0.00043901, 0.00063405, 0.00062764, 0.00039678, },
		{ 0.00005181, 0.00035975, 0.00057504, 0.00061832, 0.00046262, 0.00014364, -0.00025026, -0.00060539, -0.00081851, -0.00082933, -0.00063919, -0.00030977, 0.00005751, 0.00035454, 0.00050097, 0.00046888, 0.00028950, 0.00004027, -0.00018317, -0.00030021, -0.00027266, -0.00011670, 0.00010412, 0.00030246, 0.00039895, 0.00035106, 0.00016907, -0.00008659, -0.00032573, -0.00045973, -0.00043308, -0.00024420, },
		{ -0.00003527, -0.00027128, -0.00041389, -0.00041140, -0.00025987, -0.00000526, 0.00027104, 0.00047858, 0.00054744, 0.00045122, 0.00021578, -0.00008923, -0.00037304, -0.00055328, -0.00058240, -0.00046227, -0.00024239, -0.00000218, 0.00017621, 0.00023566, 0.00016278, -0.00000844, -0.00020930, -0.00036083, -0.00040164, -0.00030962, -0.00010997, 0.00013311, 0.00033827, 0.00043512, 0.00038879, 0.00021266, },
		{ 0.00010340, 0.00029823, 0.00039225, 0.00035758, 0.00020775, -0.00000630, -0.00021274, -0.00034233, -0.00035120, -0.00023520, -0.00003122, 0.00019499, 0.00037051, 0.00043910, 0.00037935, 0.00021113, -0.00001169, -0.00021966, -0.00035111, -0.00037244, -0.00028867, -0.00014069, 0.00000974, 0.00010192, 0.00009765, -0.00000522, -0.00017177, -0.00034197, -0.00045143, -0.00045396, -0.00033821, -0.00013256, },
		{ -0.00035228, -0.00043716, -0.00042924, -0.00032362, -0.00014680, 0.00005126, 0.00021307, 0.00029216, 0.00026824, 0.00015381, -0.00000995, -0.00016674, -0.00026351, -0.00026778, -0.00017797, -0.00002353, 0.00014519, 0.00027381, 0.00032190, 0.00027603, 0.00015360, -0.00000371, -0.00014460, -0.00022580, -0.00022669, -0.00015601, -0.00004825, 0.00004879, 0.00009117, 0.00005469, -0.00005604, -0.00020837, },
		{ 0.00007626, 0.00009910, 0.00006006, -0.00003374, -0.00015536, -0.00026675, -0.00033138, -0.00032652, -0.00025114, -0.00012652, 0.00001038, 0.00011896, 0.00016815, 0.00014645, 0.00006558, -0.00004365, -0.00014169, -0.00019348, -0.00018000, -0.00010448, 0.00000892, 0.00012339, 0.00020194, 0.00021973, 0.00017226, 0.00007659, -0.00003476, -0.00012506, -0.00016601, -0.00014740, -0.00008045, 0.00000613, },
		{ -0.00012661, -0.00014270, -0.00011490, -0.00005472, 0.00001571, 0.00007116, 0.00009178, 0.00006967, 0.00001143, -0.00006418, -0.00013268, -0.00017214, -0.00017055, -0.00012983, -0.00006495, 0.00000123, 0.00004589, 0.00005413, 0.00002409, -0.00003235, -0.00009364, -0.00013596, -0.00014143, -0.00010447, -0.00003394, 0.00004953, 0.00012026, 0.00015597, 0.00014526, 0.00009139, 0.00001136, -0.00006966, },
		{ 0.00012638, 0.00009797, 0.00004313, -0.00002250, -0.00008080, -0.00011656, -0.00012185, -0.00009795, -0.00005443, -0.00000558, 0.00003434, 0.00005559, 0.00005570, 0.00003953, 0.00001686, -0.00000163, -0.00000849, -0.00000230, 0.00001193, 0.00002474, 0.00002605, 0.00000934, -0.00002534, -0.00007063, -0.00011371, -0.00014034, -0.00013964, -0.00010808, -0.00005121, 0.00001744, 0.00008009, 0.00011997, },
		{ -0.00008715, -0.00002184, 0.00004097, 0.00008693, 0.00010679, 0.00009847, 0.00006712, 0.00002306, -0.00002160, -0.00005644, -0.00007541, -0.00007778, -0.00006735, -0.00005019, -0.00003202, -0.00001602, -0.00000207, 0.00001266, 0.00003167, 0.00005685, 0.00008664, 0.00011554, 0.00013528, 0.00013731, 0.00011590, 0.00007070, 0.00000785, -0.00006092, -0.00012099, -0.00015873, -0.00016530, -0.00013925, },
		{ -0.00005861, -0.00012514, -0.00015969, -0.00015756, -0.00012338, -0.00006928, -0.00001094, 0.00003709, 0.00006529, 0.00007128, 0.00005941, 0.00003829, 0.00001727, 0.00000295, -0.00000276, -0.00000271, -0.00000256, -0.00000779, -0.00002072, -0.00003883, -0.00005504, -0.00005996, -0.00004541, -0.00000792, 0.00004900, 0.00011477, 0.00017412, 0.00021121, 0.00021426, 0.00017940, 0.00011222, 0.00002665, },
		{ 0.00021564, 0.00022648, 0.00019496, 0.00012914, 0.00004578, -0.00003472, -0.00009427, -0.00012206, -0.00011708, -0.00008742, -0.00004681, -0.00000966, 0.00001375, 0.00002026, 0.00001406, 0.00000428, 0.00000084, 0.00001019, 0.00003238, 0.00006078, 0.00008432, 0.00009172, 0.00007600, 0.00003783, -0.00001378, -0.00006363, -0.00009491, -0.00009463, -0.00005823, 0.00000837, 0.00008984, 0.00016559, },
		},
		{ { 0.00009908, 0.00009224, 0.00007271, 0.00004338, 0.00000855, -0.00002666, -0.00005715, -0.00007856, -0.00008794, -0.00008417, -0.00006816, -0.00004264, -0.00001179, 0.00001945, 0.00004613, 0.00006402, 0.00007031, 0.00006402, 0.00004613, 0.00001945, -0.00001179, -0.00004264, -0.00006816, -0.00008417, -0.00008794, -0.00007856, -0.00005715, -0.00002666, 0.00000855, 0.00004338, 0.00007271, 0.00009224, },
		{ -0.00006642, -0.00004564, -0.00001722, 0.00001492, 0.00004621, 0.00007210, 0.00008880, 0.00009380, 0.00008627, 0.00006719, 0.00003921, 0.00000624, -0.00002710, -0.00005621, -0.00007716, -0.00008727, -0.00008553, -0.00007269, -0.00005116, -0.00002461, 0.00000260, 0.00002610, 0.00004217, 0.00004833, 0.00004371, 0.00002918, 0.00000721, -0.00001852, -0.00004367, -0.00006396, -0.00007578, -0.00007683, },
		{ -0.00001408, -0.00002872, -0.00004440, -0.00005830, -0.00006777, -0.00007072, -0.00006599, -0.00005355, -0.00003453, -0.00001108, 0.00001393, 0.00003734, 0.00005610, 0.00006778, 0.00007088, 0.00006508, 0.00005124, 0.00003132, 0.00000804, -0.00001547, -0.00003620, -0.00005162, -0.00006009, -0.00006109, -0.00005521, -0.00004411, -0.00003017, -0.00001615, -0.00000472, 0.00000197, 0.00000261, -0.00000303, },
		{ -0.00000676, -0.00000587, -0.00000719, -0.00001115, -0.00001769, -0.00002630, -0.00003608, -0.00004581, -0.00005420, -0.00006001, -0.00006228, -0.00006046, -0.00005452, -0.00004494, -0.00003268, -0.00001902, -0.00000544, 0.00000665, 0.00001603, 0.00002189, 0.00002390, 0.00002224, 0.00001755, 0.00001084, 0.00000330, -0.00000387, -0.00000966, -0.00001339, -0.00001480, -0.00001410, -0.00001190, -0.00000910, },
		{ -0.00004511, -0.00003200, -0.00001508, 0.00000334, 0.00002073, 0.00003468, 0.00004327, 0.00004536, 0.00004078, 0.00003030, 0.00001554, -0.00000128, -0.00001766, -0.00003125, -0.00004014, -0.00004319, -0.00004018, -0.00003181, -0.00001964, -0.00000579, 0.00000737, 0.00001760, 0.00002313, 0.00002295, 0.00001694, 0.00000596, -0.00000837, -0.00002384, -0.00003801, -0.00004860, -0.00005382, -0.00005268, },
		{ 0.00003927, 0.00000243, -0.00003585, -0.00007073, -0.00009772, -0.00011337, -0.00011575, -0.00010472, -0.00008197, -0.00005073, -0.00001532, 0.00001953, 0.00004934, 0.00007054, 0.00008099, 0.00008025, 0.00006960, 0.00005181, 0.00003069, 0.00001042, -0.00000509, -0.00001282, -0.00001113, -0.00000006, 0.00001871, 0.00004209, 0.00006609, 0.00008639, 0.00009908, 0.00010126, 0.00009150, 0.00007014, },
		{ 0.00006611, 0.00010412, 0.00013969, 0.00016687, 0.00018061, 0.00017766, 0.00015703, 0.00012022, 0.00007107, 0.00001523, -0.00004063, -0.00008976, -0.00012630, -0.00014609, -0.00014729, -0.00013056, -0.00009895, -0.00005746, -0.00001220, 0.00003050, 0.00006502, 0.00008728, 0.00009532, 0.00008957, 0.00007265, 0.00004899, 0.00002400, 0.00000326, -0.00000846, -0.00000797, 0.00000581, 0.00003164, },
		{ 0.00005592, 0.00001723, -0.00002163, -0.00005320, -0.00007109, -0.00007109, -0.00005179, -0.00001489, 0.00003501, 0.00009107, 0.00014527, 0.00018952, 0.00021698, 0.00022305, 0.00020610, 0.00016773, 0.00011260, 0.00004774, -0.00001846, -0.00007749, -0.00012191, -0.00014649, -0.00014886, -0.00012991, -0.00009354, -0.00004606, 0.00000481, 0.00005115, 0.00008601, 0.00010451, 0.00010454, 0.00008706, },
		{ -0.00006923, -0.00000810, 0.00005500, 0.00010915, 0.00014481, 0.00015537, 0.00013814, 0.00009488, 0.00003147, -0.00004287, -0.00011701, -0.00017960, -0.00022080, -0.00023376, -0.00021572, -0.00016845, -0.00009801, -0.00001387, 0.00007246, 0.00014917, 0.00020585, 0.00023508, 0.00023346, 0.00020214, 0.00014656, 0.00007563, 0.00000033, -0.00006792, -0.00011907, -0.00014590, -0.00014510, -0.00011778, },
		{ 0.00014041, 0.00005213, -0.00004010, -0.00012010, -0.00017402, -0.00019253, -0.00017226, -0.00011635, -0.00003397, 0.00006110, 0.00015264, 0.00022448, 0.00026310, 0.00025975, 0.00021189, 0.00012374, 0.00000570, -0.00012709, -0.00025694, -0.00036623, -0.00043998, -0.00046810, -0.00044685, -0.00037941, -0.00027539, -0.00014940, -0.00001885, 0.00009859, 0.00018765, 0.00023763, 0.00024396, 0.00020876, },
		{ -0.00036709, -0.00019008, -0.00001646, 0.00012949, 0.00022838, 0.00026853, 0.00024768, 0.00017308, 0.00006035, -0.00006916, -0.00019172, -0.00028502, -0.00033169, -0.00032206, -0.00025577, -0.00014184, 0.00000262, 0.00015504, 0.00029079, 0.00038708, 0.00042640, 0.00039938, 0.00030631, 0.00015726, -0.00002932, -0.00022924, -0.00041605, -0.00056498, -0.00065652, -0.00067928, -0.00063165, -0.00052194, },
		{ -0.00009476, -0.00034165, -0.00055385, -0.00070188, -0.00076584, -0.00073836, -0.00062558, -0.00044619, -0.00022851, -0.00000612, 0.00018726, 0.00032286, 0.00038126, 0.00035531, 0.00025110, 0.00008697, -0.00010934, -0.00030507, -0.00046751, -0.00056910, -0.00059172, -0.00052959, -0.00039019, -0.00019323, 0.00003231, 0.00025254, 0.00043380, 0.00054782, 0.00057600, 0.00051234, 0.00036438, 0.00015216, },
		{ 0.00005051, 0.00032208, 0.00052553, 0.00063191, 0.00062728, 0.00051472, 0.00031380, 0.00005749, -0.00021296, -0.00045434, -0.00062821, -0.00070676, -0.00067710, -0.00054330, -0.00032572, -0.00005799, 0.00021821, 0.00045945, 0.00062730, 0.00069416, 0.00064762, 0.00049237, 0.00024962, -0.00004605, -0.00035188, -0.00062331, -0.00082074, -0.00091544, -0.00089392, -0.00075991, -0.00053382, -0.00024957, },
		{ -0.00061750, -0.00094561, -0.00115308, -0.00121686, -0.00113455, -0.00092456, -0.00062301, -0.00027815, 0.00005715, 0.00033341, 0.00051204, 0.00057108, 0.00050833, 0.00034143, 0.00010486, -0.00015564, -0.00039042, -0.00055389, -0.00061178, -0.00054677, -0.00036143, -0.00007811, 0.00026423, 0.00061580, 0.00092344, 0.00113875, 0.00122549, 0.00116522, 0.00096032, 0.00063390, 0.00022658, -0.00020934, },
		{ 0.00170692, 0.00184532, 0.00180446, 0.00158273, 0.00120289, 0.00070909, 0.00016060, -0.00037653, -0.00083891, -0.00117475, -0.00135097, -0.00135763, -0.00120873, -0.00093952, -0.00060065, -0.00025011, 0.00005588, 0.00027153, 0.00036839, 0.00033957, 0.00020042, -0.00001417, -0.00025565, -0.00046978, -0.00060558, -0.00062379, -0.00050357, -0.00024635, 0.00012373, 0.00056291, 0.00101400, 0.00141473, },
		{ -0.00080966, -0.00064738, -0.00033006, 0.00011774, 0.00064939, 0.00120248, 0.00170759, 0.00209832, 0.00232095, 0.00234244, 0.00215541, 0.00177941, 0.00125831, 0.00065411, 0.00003820, -0.00051875, -0.00095666, -0.00123419, -0.00133414, -0.00126502, -0.00105891, -0.00076572, -0.00044486, -0.00015567, 0.00005204, 0.00014581, 0.00011546, -0.00002545, -0.00024211, -0.00048420, -0.00069426, -0.00081722, },
		{ 0.00005017, -0.00001120, -0.00018322, -0.00043160, -0.00070536, -0.00094512, -0.00109277, -0.00110123, -0.00094247, -0.00061288, -0.00013485, 0.00044558, 0.00106466, 0.00164972, 0.00212942, 0.00244389, 0.00255336, 0.00244389, 0.00212942, 0.00164972, 0.00106466, 0.00044558, -0.00013485, -0.00061288, -0.00094247, -0.00110123, -0.00109277, -0.00094512, -0.00070536, -0.00043160, -0.00018322, -0.00001120, },
		{ -0.00080966, -0.00081722, -0.00069426, -0.00048420, -0.00024211, -0.00002545, 0.00011546, 0.00014581, 0.00005204, -0.00015567, -0.00044486, -0.00076572, -0.00105891, -0.00126502, -0.00133414, -0.00123419, -0.00095666, -0.00051875, 0.00003820, 0.00065411, 0.00125831, 0.00177941, 0.00215541, 0.00234244, 0.00232095, 0.00209832, 0.00170759, 0.00120248, 0.00064939, 0.00011774, -0.00033006, -0.00064738, },
		{ 0.00170692, 0.00141473, 0.00101400, 0.00056291, 0.00012373, -0.00024635, -0.00050357, -0.00062379, -0.00060558, -0.00046978, -0.00025565, -0.00001417, 0.00020042, 0.00033957, 0.00036839, 0.00027153, 0.00005588, -0.00025011, -0.00060065, -0.00093952, -0.00120873, -0.00135763, -0.00135097, -0.00117475, -0.00083891, -0.00037653, 0.00016060, 0.00070909, 0.00120289, 0.00158273, 0.00180446, 0.00184532, },
		{ -0.00061750, -0.00020934, 0.00022658, 0.00063390, 0.00096032, 0.00116522, 0.00122549, 0.00113875, 0.00092344, 0.00061580, 0.00026423, -0.00007811, -0.00036143, -0.00054677, -0.00061178, -0.00055389, -0.00039042, -0.00015564, 0.00010486, 0.00034143, 0.00050833, 0.00057108, 0.00051204, 0.00033341, 0.00005715, -0.00027815, -0.00062301, -0.00092456, -0.00113455, -0.00121686, -0.00115308, -0.00094561, },
		{ 0.00005051, -0.00024957, -0.00053382, -0.00075991, -0.00089392, -0.00091544, -0.00082074, -0.00062331, -0.00035188, -0.00004605, 0.00024962, 0.00049237, 0.00064762, 0.00069416, 0.00062730, 0.00045945, 0.00021821, -0.00005799, -0.00032572, -0.00054330, -0.00067710, -0.00070676, -0.00062821, -0.00045434, -0.00021296, 0.00005749, 0.00031380, 0.00051472, 0.00062728, 0.00063191, 0.00052553, 0.00032208, },
		{ -0.00009476, 0.00015216, 0.00036438, 0.00051234, 0.00057600, 0.00054782, 0.00043380, 0.00025254, 0.00003231, -0.00019323, -0.00039019, -0.00052959, -0.00059172, -0.00056910, -0.00046751, -0.00030507, -0.00010934, 0.00008697, 0.00025110, 0.00035531, 0.00038126, 0.00032286, 0.00018726, -0.00000612, -0.00022851, -0.00044619, -0.00062558, -0.00073836, -0.00076584, -0.00070188, -0.00055385, -0.00034165, },
		{ -0.00036709, -0.00052194, -0.00063165, -0.00067928, -0.00065652, -0.00056498, -0.00041605, -0.00022924, -0.00002932, 0.00015726, 0.00030631, 0.00039938, 0.00042640, 0.00038708, 0.00029079, 0.00015504, 0.00000262, -0.00014184, -0.00025577, -0.00032206, -0.00033169, -0.00028502, -0.00019172, -0.00006916, 0.00006035, 0.00017308, 0.00024768, 0.00026853, 0.00022838, 0.00012949, -0.00001646, -0.00019008, },
		{ 0.00014041, 0.00020876, 0.00024396, 0.00023763, 0.00018765, 0.00009859, -0.00001885, -0.00014940, -0.00027539, -0.00037941, -0.00044685, -0.00046810, -0.00043998, -0.00036623, -0.00025694, -0.00012709, 0.00000570, 0.00012374, 0.00021189, 0.00025975, 0.00026310, 0.00022448, 0.00015264, 0.00006110, -0.00003397, -0.00011635, -0.00017226, -0.00019253, -0.00017402, -0.00012010, -0.00004010, 0.00005213, },
		{ -0.00006923, -0.00011778, -0.00014510, -0.00014590, -0.00011907, -0.00006792, 0.00000033, 0.00007563, 0.00014656, 0.00020214, 0.00023346, 0.00023508, 0.00020585, 0.00014917, 0.00007246, -0.00001387, -0.00009801, -0.00016845, -0.00021572, -0.00023376, -0.00022080, -0.00017960, -0.00011701, -0.00004287, 0.00003147, 0.00009488, 0.00013814, 0.00015537, 0.00014481, 0.00010915, 0.00005500, -0.00000810, },
		{ 0.00005592, 0.00008706, 0.00010454, 0.00010451, 0.00008601, 0.00005115, 0.00000481, -0.00004606, -0.00009354, -0.00012991, -0.00014886, -0.00014649, -0.00012191, -0.00007749, -0.00001846, 0.00004774, 0.00011260, 0.00016773, 0.00020610, 0.00022305, 0.00021698, 0.00018952, 0.00014527, 0.00009107, 0.00003501, -0.00001489, -0.00005179, -0.00007109, -0.00007109, -0.00005320, -0.00002163, 0.00001723, },
		{ 0.00006611, 0.00003164, 0.00000581, -0.00000797, -0.00000846, 0.00000326, 0.00002400, 0.00004899, 0.00007265, 0.00008957, 0.00009532, 0.00008728, 0.00006502, 0.00003050, -0.00001220, -0.00005746, -0.00009895, -0.00013056, -0.00014729, -0.00014609, -0.00012630, -0.00008976, -0.00004063, 0.00001523, 0.00007107, 0.00012022, 0.00015703, 0.00017766, 0.00018061, 0.00016687, 0.00013969, 0.00010412, },
		{ 0.00003927, 0.00007014, 0.00009150, 0.00010126, 0.00009908, 0.00008639, 0.00006609, 0.00004209, 0.00001871, -0.00000006, -0.00001113, -0.00001282, -0.00000509, 0.00001042, 0.00003069, 0.00005181, 0.00006960, 0.00008025, 0.00008099, 0.00007054, 0.00004934, 0.00001953, -0.00001532, -0.00005073, -0.00008197, -0.00010472, -0.00011575, -0.00011337, -0.00009772, -0.00007073, -0.00003585, 0.00000243, },
		{ -0.00004511, -0.00005268, -0.00005382, -0.00004860, -0.00003801, -0.00002384, -0.00000837, 0.00000596, 0.00001694, 0.00002295, 0.00002313, 0.00001760, 0.00000737, -0.00000579, -0.00001964, -0.00003181, -0.00004018, -0.00004319, -0.00004014, -0.00003125, -0.00001766, -0.00000128, 0.00001554, 0.00003030, 0.00004078, 0.00004536, 0.00004327, 0.00003468, 0.00002073, 0.00000334, -0.00001508, -0.00003200, },
		{ -0.00000676, -0.00000910, -0.00001190, -0.00001410, -0.00001480, -0.00001339, -0.00000966, -0.00000387, 0.00000330, 0.00001084, 0.00001755, 0.00002224, 0.00002390, 0.00002189, 0.00001603, 0.00000665, -0.00000544, -0.00001902, -0.00003268, -0.00004494, -0.00005452, -0.00006046, -0.00006228, -0.00006001, -0.00005420, -0.00004581, -0.00003608, -0.00002630, -0.00001769, -0.00001115, -0.00000719, -0.00000587, },
		{ -0.00001408, -0.00000303, 0.00000261, 0.00000197, -0.00000472, -0.00001615, -0.00003017, -0.00004411, -0.00005521, -0.00006109, -0.00006009, -0.00005162, -0.00003620, -0.00001547, 0.00000804, 0.00003132, 0.00005124, 0.00006508, 0.00007088, 0.00006778, 0.00005610, 0.00003734, 0.00001393, -0.00001108, -0.00003453, -0.00005355, -0.00006599, -0.00007072, -0.00006777, -0.00005830, -0.00004440, -0.00002872, },
		{ -0.00006642, -0.00007683, -0.00007578, -0.00006396, -0.00004367, -0.00001852, 0.00000721, 0.00002918, 0.00004371, 0.00004833, 0.00004217, 0.00002610, 0.00000260, -0.00002461, -0.00005116, -0.00007269, -0.00008553, -0.00008727, -0.00007716, -0.00005621, -0.00002710, 0.00000624, 0.00003921, 0.00006719, 0.00008627, 0.00009380, 0.00008880, 0.00007210, 0.00004621, 0.00001492, -0.00001722, -0.00004564, },
		},
		{ { 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, },
		{ -0.00003184, -0.00002941, -0.00002653, -0.00002329, -0.00001981, -0.00001625, -0.00001272, -0.00000938, -0.00000633, -0.00000371, -0.00000162, -0.00000012, 0.00000071, 0.00000085, 0.00000029, -0.00000095, -0.00000281, -0.00000524, -0.00000812, -0.00001136, -0.00001483, -0.00001840, -0.00002193, -0.00002527, -0.00002832, -0.00003094, -0.00003303, -0.00003453, -0.00003536, -0.00003550, -0.00003494, -0.00003370, },
		{ -0.00001330, -0.00002368, -0.00003432, -0.00004481, -0.00005475, -0.00006376, -0.00007149, -0.00007764, -0.00008198, -0.00008434, -0.00008463, -0.00008284, -0.00007904, -0.00007337, -0.00006605, -0.00005736, -0.00004764, -0.00003726, -0.00002662, -0.00001613, -0.00000619, 0.00000282, 0.00001054, 0.00001670, 0.00002104, 0.00002340, 0.00002368, 0.00002189, 0.00001809, 0.00001242, 0.00000510, -0.00000358, },
		{ 0.00003187, 0.00004555, 0.00005841, 0.00006995, 0.00007974, 0.00008739, 0.00009262, 0.00009521, 0.00009508, 0.00009222, 0.00008675, 0.00007887, 0.00006890, 0.00005720, 0.00004424, 0.00003050, 0.00001653, 0.00000285, -0.00001001, -0.00002156, -0.00003135, -0.00003900, -0.00004423, -0.00004682, -0.00004669, -0.00004383, -0.00003836, -0.00003048, -0.00002050, -0.00000881, 0.00000416, 0.00001789, },
		{ 0.00003627, 0.00002232, 0.00000890, -0.00000346, -0.00001430, -0.00002319, -0.00002980, -0.00003387, -0.00003524, -0.00003387, -0.00002980, -0.00002319, -0.00001430, -0.00000346, 0.00000890, 0.00002232, 0.00003627, 0.00005022, 0.00006363, 0.00007599, 0.00008683, 0.00009572, 0.00010233, 0.00010640, 0.00010777, 0.00010640, 0.00010233, 0.00009572, 0.00008683, 0.00007599, 0.00006363, 0.00005022, },
		{ -0.00004445, -0.00001982, 0.00000438, 0.00002722, 0.00004783, 0.00006541, 0.00007929, 0.00008893, 0.00009397, 0.00009420, 0.00008963, 0.00008042, 0.00006694, 0.00004969, 0.00002935, 0.00000669, -0.00001742, -0.00004205, -0.00006625, -0.00008909, -0.00010969, -0.00012728, -0.00014115, -0.00015080, -0.00015583, -0.00015607, -0.00015150, -0.00014229, -0.00012881, -0.00011156, -0.00009121, -0.00006855, },
		{ -0.00009718, -0.00012983, -0.00015900, -0.00018359, -0.00020263, -0.00021541, -0.00022143, -0.00022046, -0.00021253, -0.00019795, -0.00017729, -0.00015133, -0.00012107, -0.00008768, -0.00005243, -0.00001669, 0.00001817, 0.00005082, 0.00008000, 0.00010458, 0.00012363, 0.00013640, 0.00014242, 0.00014145, 0.00013352, 0.00011895, 0.00009828, 0.00007232, 0.00004206, 0.00000867, -0.00002657, -0.00006231, },
		{ 0.00015208, 0.00016763, 0.00017793, 0.00018259, 0.00018143, 0.00017449, 0.00016205, 0.00014457, 0.00012274, 0.00009739, 0.00006949, 0.00004012, 0.00001041, -0.00001851, -0.00004552, -0.00006958, -0.00008977, -0.00010532, -0.00011562, -0.00012028, -0.00011912, -0.00011218, -0.00009974, -0.00008226, -0.00006043, -0.00003508, -0.00000718, 0.00002219, 0.00005190, 0.00008082, 0.00010783, 0.00013189, },
		{ 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, },
		{ -0.00025107, -0.00022109, -0.00018536, -0.00014526, -0.00010233, -0.00005821, -0.00001460, 0.00002682, 0.00006446, 0.00009688, 0.00012282, 0.00014130, 0.00015159, 0.00015332, 0.00014640, 0.00013111, 0.00010803, 0.00007805, 0.00004232, 0.00000222, -0.00004072, -0.00008484, -0.00012844, -0.00016986, -0.00020750, -0.00023992, -0.00026586, -0.00028434, -0.00029464, -0.00029636, -0.00028944, -0.00027415, },
		{ 0.00010317, 0.00002508, -0.00005497, -0.00013391, -0.00020870, -0.00027647, -0.00033462, -0.00038091, -0.00041356, -0.00043131, -0.00043349, -0.00042001, -0.00039139, -0.00034873, -0.00029367, -0.00022832, -0.00015519, -0.00007710, 0.00000295, 0.00008189, 0.00015668, 0.00022445, 0.00028260, 0.00032889, 0.00036154, 0.00037929, 0.00038147, 0.00036799, 0.00033937, 0.00029671, 0.00024165, 0.00017630, },
		{ 0.00014118, 0.00022556, 0.00030488, 0.00037608, 0.00043645, 0.00048365, 0.00051587, 0.00053187, 0.00053105, 0.00051342, 0.00047967, 0.00043109, 0.00036956, 0.00029743, 0.00021748, 0.00013278, 0.00004658, -0.00003779, -0.00011711, -0.00018832, -0.00024868, -0.00029588, -0.00032810, -0.00034411, -0.00034328, -0.00032565, -0.00029190, -0.00024333, -0.00018179, -0.00010967, -0.00002971, 0.00005499, },
		{ -0.00003087, -0.00011218, -0.00019036, -0.00026242, -0.00032558, -0.00037741, -0.00041592, -0.00043964, -0.00044765, -0.00043964, -0.00041592, -0.00037741, -0.00032558, -0.00026242, -0.00019036, -0.00011218, -0.00003087, 0.00005044, 0.00012862, 0.00020068, 0.00026383, 0.00031566, 0.00035418, 0.00037789, 0.00038590, 0.00037789, 0.00035418, 0.00031566, 0.00026383, 0.00020068, 0.00012862, 0.00005044, },
		{ -0.00040384, -0.00025196, -0.00010271, 0.00003816, 0.00016526, 0.00027368, 0.00035928, 0.00041874, 0.00044981, 0.00045126, 0.00042306, 0.00036629, 0.00028312, 0.00017676, 0.00005128, -0.00008848, -0.00023715, -0.00038903, -0.00053828, -0.00067916, -0.00080625, -0.00091468, -0.00100027, -0.00105974, -0.00109080, -0.00109226, -0.00106406, -0.00100728, -0.00092411, -0.00081775, -0.00069228, -0.00055252, },
		{ -0.00067243, -0.00091807, -0.00113759, -0.00132256, -0.00146587, -0.00156201, -0.00160729, -0.00159997, -0.00154032, -0.00143065, -0.00127516, -0.00107984, -0.00085218, -0.00060094, -0.00033577, -0.00006686, 0.00019545, 0.00044109, 0.00066061, 0.00084558, 0.00098889, 0.00108503, 0.00113031, 0.00112298, 0.00106334, 0.00095367, 0.00079818, 0.00060286, 0.00037520, 0.00012396, -0.00014121, -0.00041012, },
		{ 0.00180871, 0.00200102, 0.00212844, 0.00218608, 0.00217172, 0.00208592, 0.00193198, 0.00171580, 0.00144570, 0.00113205, 0.00078692, 0.00042356, 0.00005593, -0.00030183, -0.00063597, -0.00093367, -0.00118347, -0.00137577, -0.00150319, -0.00156083, -0.00154648, -0.00146068, -0.00130673, -0.00109056, -0.00082046, -0.00050681, -0.00016168, 0.00020168, 0.00056931, 0.00092707, 0.00126122, 0.00155891, },
		{ -0.00141721, -0.00137741, -0.00125952, -0.00106809, -0.00081045, -0.00049653, -0.00013838, 0.00025024, 0.00065439, 0.00105854, 0.00144716, 0.00180531, 0.00211923, 0.00237686, 0.00256830, 0.00268619, 0.00272599, 0.00268619, 0.00256830, 0.00237686, 0.00211923, 0.00180531, 0.00144716, 0.00105854, 0.00065439, 0.00025024, -0.00013838, -0.00049653, -0.00081045, -0.00106809, -0.00125952, -0.00137741, },
		{ 0.00180871, 0.00155891, 0.00126122, 0.00092707, 0.00056931, 0.00020168, -0.00016168, -0.00050681, -0.00082046, -0.00109056, -0.00130673, -0.00146068, -0.00154648, -0.00156083, -0.00150319, -0.00137577, -0.00118347, -0.00093367, -0.00063597, -0.00030183, 0.00005593, 0.00042356, 0.00078692, 0.00113205, 0.00144570, 0.00171580, 0.00193198, 0.00208592, 0.00217172, 0.00218608, 0.00212844, 0.00200102, },
		{ -0.00067243, -0.00041012, -0.00014121, 0.00012396, 0.00037520, 0.00060286, 0.00079818, 0.00095367, 0.00106334, 0.00112298, 0.00113031, 0.00108503, 0.00098889, 0.00084558, 0.00066061, 0.00044109, 0.00019545, -0.00006686, -0.00033577, -0.00060094, -0.00085218, -0.00107984, -0.00127516, -0.00143065, -0.00154032, -0.00159997, -0.00160729, -0.00156201, -0.00146587, -0.00132256, -0.00113759, -0.00091807, },
		{ -0.00040384, -0.00055252, -0.00069228, -0.00081775, -0.00092411, -0.00100728, -0.00106406, -0.00109226, -0.00109080, -0.00105974, -0.00100027, -0.00091468, -0.00080625, -0.00067916, -0.00053828, -0.00038903, -0.00023715, -0.00008848, 0.00005128, 0.00017676, 0.00028312, 0.00036629, 0.00042306, 0.00045126, 0.00044981, 0.00041874, 0.00035928, 0.00027368, 0.00016526, 0.00003816, -0.00010271, -0.00025196, },
		{ -0.00003087, 0.00005044, 0.00012862, 0.00020068, 0.00026383, 0.00031566, 0.00035418, 0.00037789, 0.00038590, 0.00037789, 0.00035418, 0.00031566, 0.00026383, 0.00020068, 0.00012862, 0.00005044, -0.00003087, -0.00011218, -0.00019036, -0.00026242, -0.00032558, -0.00037741, -0.00041592, -0.00043964, -0.00044765, -0.00043964, -0.00041592, -0.00037741, -0.00032558, -0.00026242, -0.00019036, -0.00011218, },
		{ 0.00014118, 0.00005499, -0.00002971, -0.00010967, -0.00018179, -0.00024333, -0.00029190, -0.00032565, -0.00034328, -0.00034411, -0.00032810, -0.00029588, -0.00024868, -0.00018832, -0.00011711, -0.00003779, 0.00004658, 0.00013278, 0.00021748, 0.00029743, 0.00036956, 0.00043109, 0.00047967, 0.00051342, 0.00053105, 0.00053187, 0.00051587, 0.00048365, 0.00043645, 0.00037608, 0.00030488, 0.00022556, },
		{ 0.00010317, 0.00017630, 0.00024165, 0.00029671, 0.00033937, 0.00036799, 0.00038147, 0.00037929, 0.00036154, 0.00032889, 0.00028260, 0.00022445, 0.00015668, 0.00008189, 0.00000295, -0.00007710, -0.00015519, -0.00022832, -0.00029367, -0.00034873, -0.00039139, -0.00042001, -0.00043349, -0.00043131, -0.00041356, -0.00038091, -0.00033462, -0.00027647, -0.00020870, -0.00013391, -0.00005497, 0.00002508, },
		{ -0.00025107, -0.00027415, -0.00028944, -0.00029636, -0.00029464, -0.00028434, -0.00026586, -0.00023992, -0.00020750, -0.00016986, -0.00012844, -0.00008484, -0.00004072, 0.00000222, 0.00004232, 0.00007805, 0.00010803, 0.00013111, 0.00014640, 0.00015332, 0.00015159, 0.00014130, 0.00012282, 0.00009688, 0.00006446, 0.00002682, -0.00001460, -0.00005821, -0.00010233, -0.00014526, -0.00018536, -0.00022109, },
		{ 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, },
		{ 0.00015208, 0.00013189, 0.00010783, 0.00008082, 0.00005190, 0.00002219, -0.00000718, -0.00003508, -0.00006043, -0.00008226, -0.00009974, -0.00011218, -0.00011912, -0.00012028, -0.00011562, -0.00010532, -0.00008977, -0.00006958, -0.00004552, -0.00001851, 0.00001041, 0.00004012, 0.00006949, 0.00009739, 0.00012274, 0.00014457, 0.00016205, 0.00017449, 0.00018143, 0.00018259, 0.00017793, 0.00016763, },
		{ -0.00009718, -0.00006231, -0.00002657, 0.00000867, 0.00004206, 0.00007232, 0.00009828, 0.00011895, 0.00013352, 0.00014145, 0.00014242, 0.00013640, 0.00012363, 0.00010458, 0.00008000, 0.00005082, 0.00001817, -0.00001669, -0.00005243, -0.00008768, -0.00012107, -0.00015133, -0.00017729, -0.00019795, -0.00021253, -0.00022046, -0.00022143, -0.00021541, -0.00020263, -0.00018359, -0.00015900, -0.00012983, },
		{ -0.00004445, -0.00006855, -0.00009121, -0.00011156, -0.00012881, -0.00014229, -0.00015150, -0.00015607, -0.00015583, -0.00015080, -0.00014115, -0.00012728, -0.00010969, -0.00008909, -0.00006625, -0.00004205, -0.00001742, 0.00000669, 0.00002935, 0.00004969, 0.00006694, 0.00008042, 0.00008963, 0.00009420, 0.00009397, 0.00008893, 0.00007929, 0.00006541, 0.00004783, 0.00002722, 0.00000438, -0.00001982, },
		{ 0.00003627, 0.00005022, 0.00006363, 0.00007599, 0.00008683, 0.00009572, 0.00010233, 0.00010640, 0.00010777, 0.00010640, 0.00010233, 0.00009572, 0.00008683, 0.00007599, 0.00006363, 0.00005022, 0.00003627, 0.00002232, 0.00000890, -0.00000346, -0.00001430, -0.00002319, -0.00002980, -0.00003387, -0.00003524, -0.00003387, -0.00002980, -0.00002319, -0.00001430, -0.00000346, 0.00000890, 0.00002232, },
		{ 0.00003187, 0.00001789, 0.00000416, -0.00000881, -0.00002050, -0.00003048, -0.00003836, -0.00004383, -0.00004669, -0.00004682, -0.00004423, -0.00003900, -0.00003135, -0.00002156, -0.00001001, 0.00000285, 0.00001653, 0.00003050, 0.00004424, 0.00005720, 0.00006890, 0.00007887, 0.00008675, 0.00009222, 0.00009508, 0.00009521, 0.00009262, 0.00008739, 0.00007974, 0.00006995, 0.00005841, 0.00004555, },
		{ -0.00001330, -0.00000358, 0.00000510, 0.00001242, 0.00001809, 0.00002189, 0.00002368, 0.00002340, 0.00002104, 0.00001670, 0.00001054, 0.00000282, -0.00000619, -0.00001613, -0.00002662, -0.00003726, -0.00004764, -0.00005736, -0.00006605, -0.00007337, -0.00007904, -0.00008284, -0.00008463, -0.00008434, -0.00008198, -0.00007764, -0.00007149, -0.00006376, -0.00005475, -0.00004481, -0.00003432, -0.00002368, },
		{ -0.00003184, -0.00003370, -0.00003494, -0.00003550, -0.00003536, -0.00003453, -0.00003303, -0.00003094, -0.00002832, -0.00002527, -0.00002193, -0.00001840, -0.00001483, -0.00001136, -0.00000812, -0.00000524, -0.00000281, -0.00000095, 0.00000029, 0.00000085, 0.00000071, -0.00000012, -0.00000162, -0.00000371, -0.00000633, -0.00000938, -0.00001272, -0.00001625, -0.00001981, -0.00002329, -0.00002653, -0.00002941, },
		},
		{ { -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, -0.00007355, },
		{ -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, },
		{ 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, },
		{ -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, },
		{ -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, },
		{ 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, },
		{ 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, },
		{ -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, },
		{ -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, },
		{ 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, },
		{ 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, },
		{ -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, },
		{ -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, },
		{ 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, },
		{ -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, },
		{ -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, },
		{ 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, 0.00333021, },
		{ -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, -0.00144016, },
		{ -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, -0.00015217, },
		{ 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, 0.00024860, },
		{ -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, -0.00011719, },
		{ -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, -0.00022909, },
		{ 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, 0.00006031, },
		{ 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, 0.00013883, },
		{ -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, -0.00006923, },
		{ -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, -0.00011052, },
		{ 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, 0.00004681, },
		{ 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, 0.00006134, },
		{ -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, -0.00005005, },
		{ -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, -0.00002817, },
		{ 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, 0.00005584, },
		{ -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, -0.00000032, },
		},
		{ { 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, 0.00001277, },
		{ -0.00003184, -0.00003370, -0.00003494, -0.00003550, -0.00003536, -0.00003453, -0.00003303, -0.00003094, -0.00002832, -0.00002527, -0.00002193, -0.00001840, -0.00001483, -0.00001136, -0.00000812, -0.00000524, -0.00000281, -0.00000095, 0.00000029, 0.00000085, 0.00000071, -0.00000012, -0.00000162, -0.00000371, -0.00000633, -0.00000938, -0.00001272, -0.00001625, -0.00001981, -0.00002329, -0.00002653, -0.00002941, },
		{ -0.00001330, -0.00000358, 0.00000510, 0.00001242, 0.00001809, 0.00002189, 0.00002368, 0.00002340, 0.00002104, 0.00001670, 0.00001054, 0.00000282, -0.00000619, -0.00001613, -0.00002662, -0.00003726, -0.00004764, -0.00005736, -0.00006605, -0.00007337, -0.00007904, -0.00008284, -0.00008463, -0.00008434, -0.00008198, -0.00007764, -0.00007149, -0.00006376, -0.00005475, -0.00004481, -0.00003432, -0.00002368, },
		{ 0.00003187, 0.00001789, 0.00000416, -0.00000881, -0.00002050, -0.00003048, -0.00003836, -0.00004383, -0.00004669, -0.00004682, -0.00004423, -0.00003900, -0.00003135, -0.00002156, -0.00001001, 0.00000285, 0.00001653, 0.00003050, 0.00004424, 0.00005720, 0.00006890, 0.00007887, 0.00008675, 0.00009222, 0.00009508, 0.00009521, 0.00009262, 0.00008739, 0.00007974, 0.00006995, 0.00005841, 0.00004555, },
		{ 0.00003627, 0.00005022, 0.00006363, 0.00007599, 0.00008683, 0.00009572, 0.00010233, 0.00010640, 0.00010777, 0.00010640, 0.00010233, 0.00009572, 0.00008683, 0.00007599, 0.00006363, 0.00005022, 0.00003627, 0.00002232, 0.00000890, -0.00000346, -0.00001430, -0.00002319, -0.00002980, -0.00003387, -0.00003524, -0.00003387, -0.00002980, -0.00002319, -0.00001430, -0.00000346, 0.00000890, 0.00002232, },
		{ -0.00004445, -0.00006855, -0.00009121, -0.00011156, -0.00012881, -0.00014229, -0.00015150, -0.00015607, -0.00015583, -0.00015080, -0.00014115, -0.00012728, -0.00010969, -0.00008909, -0.00006625, -0.00004205, -0.00001742, 0.00000669, 0.00002935, 0.00004969, 0.00006694, 0.00008042, 0.00008963, 0.00009420, 0.00009397, 0.00008893, 0.00007929, 0.00006541, 0.00004783, 0.00002722, 0.00000438, -0.00001982, },
		{ -0.00009718, -0.00006231, -0.00002657, 0.00000867, 0.00004206, 0.00007232, 0.00009828, 0.00011895, 0.00013352, 0.00014145, 0.00014242, 0.00013640, 0.00012363, 0.00010458, 0.00008000, 0.00005082, 0.00001817, -0.00001669, -0.00005243, -0.00008768, -0.00012107, -0.00015133, -0.00017729, -0.00019795, -0.00021253, -0.00022046, -0.00022143, -0.00021541, -0.00020263, -0.00018359, -0.00015900, -0.00012983, },
		{ 0.00015208, 0.00013189, 0.00010783, 0.00008082, 0.00005190, 0.00002219, -0.00000718, -0.00003508, -0.00006043, -0.00008226, -0.00009974, -0.00011218, -0.00011912, -0.00012028, -0.00011562, -0.00010532, -0.00008977, -0.00006958, -0.00004552, -0.00001851, 0.00001041, 0.00004012, 0.00006949, 0.00009739, 0.00012274, 0.00014457, 0.00016205, 0.00017449, 0.00018143, 0.00018259, 0.00017793, 0.00016763, },
		{ 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, },
		{ -0.00025107, -0.00027415, -0.00028944, -0.00029636, -0.00029464, -0.00028434, -0.00026586, -0.00023992, -0.00020750, -0.00016986, -0.00012844, -0.00008484, -0.00004072, 0.00000222, 0.00004232, 0.00007805, 0.00010803, 0.00013111, 0.00014640, 0.00015332, 0.00015159, 0.00014130, 0.00012282, 0.00009688, 0.00006446, 0.00002682, -0.00001460, -0.00005821, -0.00010233, -0.00014526, -0.00018536, -0.00022109, },
		{ 0.00010317, 0.00017630, 0.00024165, 0.00029671, 0.00033937, 0.00036799, 0.00038147, 0.00037929, 0.00036154, 0.00032889, 0.00028260, 0.00022445, 0.00015668, 0.00008189, 0.00000295, -0.00007710, -0.00015519, -0.00022832, -0.00029367, -0.00034873, -0.00039139, -0.00042001, -0.00043349, -0.00043131, -0.00041356, -0.00038091, -0.00033462, -0.00027647, -0.00020870, -0.00013391, -0.00005497, 0.00002508, },
		{ 0.00014118, 0.00005499, -0.00002971, -0.00010967, -0.00018179, -0.00024333, -0.00029190, -0.00032565, -0.00034328, -0.00034411, -0.00032810, -0.00029588, -0.00024868, -0.00018832, -0.00011711, -0.00003779, 0.00004658, 0.00013278, 0.00021748, 0.00029743, 0.00036956, 0.00043109, 0.00047967, 0.00051342, 0.00053105, 0.00053187, 0.00051587, 0.00048365, 0.00043645, 0.00037608, 0.00030488, 0.00022556, },
		{ -0.00003087, 0.00005044, 0.00012862, 0.00020068, 0.00026383, 0.00031566, 0.00035418, 0.00037789, 0.00038590, 0.00037789, 0.00035418, 0.00031566, 0.00026383, 0.00020068, 0.00012862, 0.00005044, -0.00003087, -0.00011218, -0.00019036, -0.00026242, -0.00032558, -0.00037741, -0.00041592, -0.00043964, -0.00044765, -0.00043964, -0.00041592, -0.00037741, -0.00032558, -0.00026242, -0.00019036, -0.00011218, },
		{ -0.00040384, -0.00055252, -0.00069228, -0.00081775, -0.00092411, -0.00100728, -0.00106406, -0.00109226, -0.00109080, -0.00105974, -0.00100027, -0.00091468, -0.00080625, -0.00067916, -0.00053828, -0.00038903, -0.00023715, -0.00008848, 0.00005128, 0.00017676, 0.00028312, 0.00036629, 0.00042306, 0.00045126, 0.00044981, 0.00041874, 0.00035928, 0.00027368, 0.00016526, 0.00003816, -0.00010271, -0.00025196, },
		{ -0.00067243, -0.00041012, -0.00014121, 0.00012396, 0.00037520, 0.00060286, 0.00079818, 0.00095367, 0.00106334, 0.00112298, 0.00113031, 0.00108503, 0.00098889, 0.00084558, 0.00066061, 0.00044109, 0.00019545, -0.00006686, -0.00033577, -0.00060094, -0.00085218, -0.00107984, -0.00127516, -0.00143065, -0.00154032, -0.00159997, -0.00160729, -0.00156201, -0.00146587, -0.00132256, -0.00113759, -0.00091807, },
		{ 0.00180871, 0.00155891, 0.00126122, 0.00092707, 0.00056931, 0.00020168, -0.00016168, -0.00050681, -0.00082046, -0.00109056, -0.00130673, -0.00146068, -0.00154648, -0.00156083, -0.00150319, -0.00137577, -0.00118347, -0.00093367, -0.00063597, -0.00030183, 0.00005593, 0.00042356, 0.00078692, 0.00113205, 0.00144570, 0.00171580, 0.00193198, 0.00208592, 0.00217172, 0.00218608, 0.00212844, 0.00200102, },
		{ -0.00141721, -0.00137741, -0.00125952, -0.00106809, -0.00081045, -0.00049653, -0.00013838, 0.00025024, 0.00065439, 0.00105854, 0.00144716, 0.00180531, 0.00211923, 0.00237686, 0.00256830, 0.00268619, 0.00272599, 0.00268619, 0.00256830, 0.00237686, 0.00211923, 0.00180531, 0.00144716, 0.00105854, 0.00065439, 0.00025024, -0.00013838, -0.00049653, -0.00081045, -0.00106809, -0.00125952, -0.00137741, },
		{ 0.00180871, 0.00200102, 0.00212844, 0.00218608, 0.00217172, 0.00208592, 0.00193198, 0.00171580, 0.00144570, 0.00113205, 0.00078692, 0.00042356, 0.00005593, -0.00030183, -0.00063597, -0.00093367, -0.00118347, -0.00137577, -0.00150319, -0.00156083, -0.00154648, -0.00146068, -0.00130673, -0.00109056, -0.00082046, -0.00050681, -0.00016168, 0.00020168, 0.00056931, 0.00092707, 0.00126122, 0.00155891, },
		{ -0.00067243, -0.00091807, -0.00113759, -0.00132256, -0.00146587, -0.00156201, -0.00160729, -0.00159997, -0.00154032, -0.00143065, -0.00127516, -0.00107984, -0.00085218, -0.00060094, -0.00033577, -0.00006686, 0.00019545, 0.00044109, 0.00066061, 0.00084558, 0.00098889, 0.00108503, 0.00113031, 0.00112298, 0.00106334, 0.00095367, 0.00079818, 0.00060286, 0.00037520, 0.00012396, -0.00014121, -0.00041012, },
		{ -0.00040384, -0.00025196, -0.00010271, 0.00003816, 0.00016526, 0.00027368, 0.00035928, 0.00041874, 0.00044981, 0.00045126, 0.00042306, 0.00036629, 0.00028312, 0.00017676, 0.00005128, -0.00008848, -0.00023715, -0.00038903, -0.00053828, -0.00067916, -0.00080625, -0.00091468, -0.00100027, -0.00105974, -0.00109080, -0.00109226, -0.00106406, -0.00100728, -0.00092411, -0.00081775, -0.00069228, -0.00055252, },
		{ -0.00003087, -0.00011218, -0.00019036, -0.00026242, -0.00032558, -0.00037741, -0.00041592, -0.00043964, -0.00044765, -0.00043964, -0.00041592, -0.00037741, -0.00032558, -0.00026242, -0.00019036, -0.00011218, -0.00003087, 0.00005044, 0.00012862, 0.00020068, 0.00026383, 0.00031566, 0.00035418, 0.00037789, 0.00038590, 0.00037789, 0.00035418, 0.00031566, 0.00026383, 0.00020068, 0.00012862, 0.00005044, },
		{ 0.00014118, 0.00022556, 0.00030488, 0.00037608, 0.00043645, 0.00048365, 0.00051587, 0.00053187, 0.00053105, 0.00051342, 0.00047967, 0.00043109, 0.00036956, 0.00029743, 0.00021748, 0.00013278, 0.00004658, -0.00003779, -0.00011711, -0.00018832, -0.00024868, -0.00029588, -0.00032810, -0.00034411, -0.00034328, -0.00032565, -0.00029190, -0.00024333, -0.00018179, -0.00010967, -0.00002971, 0.00005499, },
		{ 0.00010317, 0.00002508, -0.00005497, -0.00013391, -0.00020870, -0.00027647, -0.00033462, -0.00038091, -0.00041356, -0.00043131, -0.00043349, -0.00042001, -0.00039139, -0.00034873, -0.00029367, -0.00022832, -0.00015519, -0.00007710, 0.00000295, 0.00008189, 0.00015668, 0.00022445, 0.00028260, 0.00032889, 0.00036154, 0.00037929, 0.00038147, 0.00036799, 0.00033937, 0.00029671, 0.00024165, 0.00017630, },
		{ -0.00025107, -0.00022109, -0.00018536, -0.00014526, -0.00010233, -0.00005821, -0.00001460, 0.00002682, 0.00006446, 0.00009688, 0.00012282, 0.00014130, 0.00015159, 0.00015332, 0.00014640, 0.00013111, 0.00010803, 0.00007805, 0.00004232, 0.00000222, -0.00004072, -0.00008484, -0.00012844, -0.00016986, -0.00020750, -0.00023992, -0.00026586, -0.00028434, -0.00029464, -0.00029636, -0.00028944, -0.00027415, },
		{ 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, 0.00001708, },
		{ 0.00015208, 0.00016763, 0.00017793, 0.00018259, 0.00018143, 0.00017449, 0.00016205, 0.00014457, 0.00012274, 0.00009739, 0.00006949, 0.00004012, 0.00001041, -0.00001851, -0.00004552, -0.00006958, -0.00008977, -0.00010532, -0.00011562, -0.00012028, -0.00011912, -0.00011218, -0.00009974, -0.00008226, -0.00006043, -0.00003508, -0.00000718, 0.00002219, 0.00005190, 0.00008082, 0.00010783, 0.00013189, },
		{ -0.00009718, -0.00012983, -0.00015900, -0.00018359, -0.00020263, -0.00021541, -0.00022143, -0.00022046, -0.00021253, -0.00019795, -0.00017729, -0.00015133, -0.00012107, -0.00008768, -0.00005243, -0.00001669, 0.00001817, 0.00005082, 0.00008000, 0.00010458, 0.00012363, 0.00013640, 0.00014242, 0.00014145, 0.00013352, 0.00011895, 0.00009828, 0.00007232, 0.00004206, 0.00000867, -0.00002657, -0.00006231, },
		{ -0.00004445, -0.00001982, 0.00000438, 0.00002722, 0.00004783, 0.00006541, 0.00007929, 0.00008893, 0.00009397, 0.00009420, 0.00008963, 0.00008042, 0.00006694, 0.00004969, 0.00002935, 0.00000669, -0.00001742, -0.00004205, -0.00006625, -0.00008909, -0.00010969, -0.00012728, -0.00014115, -0.00015080, -0.00015583, -0.00015607, -0.00015150, -0.00014229, -0.00012881, -0.00011156, -0.00009121, -0.00006855, },
		{ 0.00003627, 0.00002232, 0.00000890, -0.00000346, -0.00001430, -0.00002319, -0.00002980, -0.00003387, -0.00003524, -0.00003387, -0.00002980, -0.00002319, -0.00001430, -0.00000346, 0.00000890, 0.00002232, 0.00003627, 0.00005022, 0.00006363, 0.00007599, 0.00008683, 0.00009572, 0.00010233, 0.00010640, 0.00010777, 0.00010640, 0.00010233, 0.00009572, 0.00008683, 0.00007599, 0.00006363, 0.00005022, },
		{ 0.00003187, 0.00004555, 0.00005841, 0.00006995, 0.00007974, 0.00008739, 0.00009262, 0.00009521, 0.00009508, 0.00009222, 0.00008675, 0.00007887, 0.00006890, 0.00005720, 0.00004424, 0.00003050, 0.00001653, 0.00000285, -0.00001001, -0.00002156, -0.00003135, -0.00003900, -0.00004423, -0.00004682, -0.00004669, -0.00004383, -0.00003836, -0.00003048, -0.00002050, -0.00000881, 0.00000416, 0.00001789, },
		{ -0.00001330, -0.00002368, -0.00003432, -0.00004481, -0.00005475, -0.00006376, -0.00007149, -0.00007764, -0.00008198, -0.00008434, -0.00008463, -0.00008284, -0.00007904, -0.00007337, -0.00006605, -0.00005736, -0.00004764, -0.00003726, -0.00002662, -0.00001613, -0.00000619, 0.00000282, 0.00001054, 0.00001670, 0.00002104, 0.00002340, 0.00002368, 0.00002189, 0.00001809, 0.00001242, 0.00000510, -0.00000358, },
		{ -0.00003184, -0.00002941, -0.00002653, -0.00002329, -0.00001981, -0.00001625, -0.00001272, -0.00000938, -0.00000633, -0.00000371, -0.00000162, -0.00000012, 0.00000071, 0.00000085, 0.00000029, -0.00000095, -0.00000281, -0.00000524, -0.00000812, -0.00001136, -0.00001483, -0.00001840, -0.00002193, -0.00002527, -0.00002832, -0.00003094, -0.00003303, -0.00003453, -0.00003536, -0.00003550, -0.00003494, -0.00003370, },
		},
		{ { 0.00009908, 0.00009224, 0.00007271, 0.00004338, 0.00000855, -0.00002666, -0.00005715, -0.00007856, -0.00008794, -0.00008417, -0.00006816, -0.00004264, -0.00001179, 0.00001945, 0.00004613, 0.00006402, 0.00007031, 0.00006402, 0.00004613, 0.00001945, -0.00001179, -0.00004264, -0.00006816, -0.00008417, -0.00008794, -0.00007856, -0.00005715, -0.00002666, 0.00000855, 0.00004338, 0.00007271, 0.00009224, },
		{ -0.00006642, -0.00007683, -0.00007578, -0.00006396, -0.00004367, -0.00001852, 0.00000721, 0.00002918, 0.00004371, 0.00004833, 0.00004217, 0.00002610, 0.00000260, -0.00002461, -0.00005116, -0.00007269, -0.00008553, -0.00008727, -0.00007716, -0.00005621, -0.00002710, 0.00000624, 0.00003921, 0.00006719, 0.00008627, 0.00009380, 0.00008880, 0.00007210, 0.00004621, 0.00001492, -0.00001722, -0.00004564, },
		{ -0.00001408, -0.00000303, 0.00000261, 0.00000197, -0.00000472, -0.00001615, -0.00003017, -0.00004411, -0.00005521, -0.00006109, -0.00006009, -0.00005162, -0.00003620, -0.00001547, 0.00000804, 0.00003132, 0.00005124, 0.00006508, 0.00007088, 0.00006778, 0.00005610, 0.00003734, 0.00001393, -0.00001108, -0.00003453, -0.00005355, -0.00006599, -0.00007072, -0.00006777, -0.00005830, -0.00004440, -0.00002872, },
		{ -0.00000676, -0.00000910, -0.00001190, -0.00001410, -0.00001480, -0.00001339, -0.00000966, -0.00000387, 0.00000330, 0.00001084, 0.00001755, 0.00002224, 0.00002390, 0.00002189, 0.00001603, 0.00000665, -0.00000544, -0.00001902, -0.00003268, -0.00004494, -0.00005452, -0.00006046, -0.00006228, -0.00006001, -0.00005420, -0.00004581, -0.00003608, -0.00002630, -0.00001769, -0.00001115, -0.00000719, -0.00000587, },
		{ -0.00004511, -0.00005268, -0.00005382, -0.00004860, -0.00003801, -0.00002384, -0.00000837, 0.00000596, 0.00001694, 0.00002295, 0.00002313, 0.00001760, 0.00000737, -0.00000579, -0.00001964, -0.00003181, -0.00004018, -0.00004319, -0.00004014, -0.00003125, -0.00001766, -0.00000128, 0.00001554, 0.00003030, 0.00004078, 0.00004536, 0.00004327, 0.00003468, 0.00002073, 0.00000334, -0.00001508, -0.00003200, },
		{ 0.00003927, 0.00007014, 0.00009150, 0.00010126, 0.00009908, 0.00008639, 0.00006609, 0.00004209, 0.00001871, -0.00000006, -0.00001113, -0.00001282, -0.00000509, 0.00001042, 0.00003069, 0.00005181, 0.00006960, 0.00008025, 0.00008099, 0.00007054, 0.00004934, 0.00001953, -0.00001532, -0.00005073, -0.00008197, -0.00010472, -0.00011575, -0.00011337, -0.00009772, -0.00007073, -0.00003585, 0.00000243, },
		{ 0.00006611, 0.00003164, 0.00000581, -0.00000797, -0.00000846, 0.00000326, 0.00002400, 0.00004899, 0.00007265, 0.00008957, 0.00009532, 0.00008728, 0.00006502, 0.00003050, -0.00001220, -0.00005746, -0.00009895, -0.00013056, -0.00014729, -0.00014609, -0.00012630, -0.00008976, -0.00004063, 0.00001523, 0.00007107, 0.00012022, 0.00015703, 0.00017766, 0.00018061, 0.00016687, 0.00013969, 0.00010412, },
		{ 0.00005592, 0.00008706, 0.00010454, 0.00010451, 0.00008601, 0.00005115, 0.00000481, -0.00004606, -0.00009354, -0.00012991, -0.00014886, -0.00014649, -0.00012191, -0.00007749, -0.00001846, 0.00004774, 0.00011260, 0.00016773, 0.00020610, 0.00022305, 0.00021698, 0.00018952, 0.00014527, 0.00009107, 0.00003501, -0.00001489, -0.00005179, -0.00007109, -0.00007109, -0.00005320, -0.00002163, 0.00001723, },
		{ -0.00006923, -0.00011778, -0.00014510, -0.00014590, -0.00011907, -0.00006792, 0.00000033, 0.00007563, 0.00014656, 0.00020214, 0.00023346, 0.00023508, 0.00020585, 0.00014917, 0.00007246, -0.00001387, -0.00009801, -0.00016845, -0.00021572, -0.00023376, -0.00022080, -0.00017960, -0.00011701, -0.00004287, 0.00003147, 0.00009488, 0.00013814, 0.00015537, 0.00014481, 0.00010915, 0.00005500, -0.00000810, },
		{ 0.00014041, 0.00020876, 0.00024396, 0.00023763, 0.00018765, 0.00009859, -0.00001885, -0.00014940, -0.00027539, -0.00037941, -0.00044685, -0.00046810, -0.00043998, -0.00036623, -0.00025694, -0.00012709, 0.00000570, 0.00012374, 0.00021189, 0.00025975, 0.00026310, 0.00022448, 0.00015264, 0.00006110, -0.00003397, -0.00011635, -0.00017226, -0.00019253, -0.00017402, -0.00012010, -0.00004010, 0.00005213, },
		{ -0.00036709, -0.00052194, -0.00063165, -0.00067928, -0.00065652, -0.00056498, -0.00041605, -0.00022924, -0.00002932, 0.00015726, 0.00030631, 0.00039938, 0.00042640, 0.00038708, 0.00029079, 0.00015504, 0.00000262, -0.00014184, -0.00025577, -0.00032206, -0.00033169, -0.00028502, -0.00019172, -0.00006916, 0.00006035, 0.00017308, 0.00024768, 0.00026853, 0.00022838, 0.00012949, -0.00001646, -0.00019008, },
		{ -0.00009476, 0.00015216, 0.00036438, 0.00051234, 0.00057600, 0.00054782, 0.00043380, 0.00025254, 0.00003231, -0.00019323, -0.00039019, -0.00052959, -0.00059172, -0.00056910, -0.00046751, -0.00030507, -0.00010934, 0.00008697, 0.00025110, 0.00035531, 0.00038126, 0.00032286, 0.00018726, -0.00000612, -0.00022851, -0.00044619, -0.00062558, -0.00073836, -0.00076584, -0.00070188, -0.00055385, -0.00034165, },
		{ 0.00005051, -0.00024957, -0.00053382, -0.00075991, -0.00089392, -0.00091544, -0.00082074, -0.00062331, -0.00035188, -0.00004605, 0.00024962, 0.00049237, 0.00064762, 0.00069416, 0.00062730, 0.00045945, 0.00021821, -0.00005799, -0.00032572, -0.00054330, -0.00067710, -0.00070676, -0.00062821, -0.00045434, -0.00021296, 0.00005749, 0.00031380, 0.00051472, 0.00062728, 0.00063191, 0.00052553, 0.00032208, },
		{ -0.00061750, -0.00020934, 0.00022658, 0.00063390, 0.00096032, 0.00116522, 0.00122549, 0.00113875, 0.00092344, 0.00061580, 0.00026423, -0.00007811, -0.00036143, -0.00054677, -0.00061178, -0.00055389, -0.00039042, -0.00015564, 0.00010486, 0.00034143, 0.00050833, 0.00057108, 0.00051204, 0.00033341, 0.00005715, -0.00027815, -0.00062301, -0.00092456, -0.00113455, -0.00121686, -0.00115308, -0.00094561, },
		{ 0.00170692, 0.00141473, 0.00101400, 0.00056291, 0.00012373, -0.00024635, -0.00050357, -0.00062379, -0.00060558, -0.00046978, -0.00025565, -0.00001417, 0.00020042, 0.00033957, 0.00036839, 0.00027153, 0.00005588, -0.00025011, -0.00060065, -0.00093952, -0.00120873, -0.00135763, -0.00135097, -0.00117475, -0.00083891, -0.00037653, 0.00016060, 0.00070909, 0.00120289, 0.00158273, 0.00180446, 0.00184532, },
		{ -0.00080966, -0.00081722, -0.00069426, -0.00048420, -0.00024211, -0.00002545, 0.00011546, 0.00014581, 0.00005204, -0.00015567, -0.00044486, -0.00076572, -0.00105891, -0.00126502, -0.00133414, -0.00123419, -0.00095666, -0.00051875, 0.00003820, 0.00065411, 0.00125831, 0.00177941, 0.00215541, 0.00234244, 0.00232095, 0.00209832, 0.00170759, 0.00120248, 0.00064939, 0.00011774, -0.00033006, -0.00064738, },
		{ 0.00005017, -0.00001120, -0.00018322, -0.00043160, -0.00070536, -0.00094512, -0.00109277, -0.00110123, -0.00094247, -0.00061288, -0.00013485, 0.00044558, 0.00106466, 0.00164972, 0.00212942, 0.00244389, 0.00255336, 0.00244389, 0.00212942, 0.00164972, 0.00106466, 0.00044558, -0.00013485, -0.00061288, -0.00094247, -0.00110123, -0.00109277, -0.00094512, -0.00070536, -0.00043160, -0.00018322, -0.00001120, },
		{ -0.00080966, -0.00064738, -0.00033006, 0.00011774, 0.00064939, 0.00120248, 0.00170759, 0.00209832, 0.00232095, 0.00234244, 0.00215541, 0.00177941, 0.00125831, 0.00065411, 0.00003820, -0.00051875, -0.00095666, -0.00123419, -0.00133414, -0.00126502, -0.00105891, -0.00076572, -0.00044486, -0.00015567, 0.00005204, 0.00014581, 0.00011546, -0.00002545, -0.00024211, -0.00048420, -0.00069426, -0.00081722, },
		{ 0.00170692, 0.00184532, 0.00180446, 0.00158273, 0.00120289, 0.00070909, 0.00016060, -0.00037653, -0.00083891, -0.00117475, -0.00135097, -0.00135763, -0.00120873, -0.00093952, -0.00060065, -0.00025011, 0.00005588, 0.00027153, 0.00036839, 0.00033957, 0.00020042, -0.00001417, -0.00025565, -0.00046978, -0.00060558, -0.00062379, -0.00050357, -0.00024635, 0.00012373, 0.00056291, 0.00101400, 0.00141473, },
		{ -0.00061750, -0.00094561, -0.00115308, -0.00121686, -0.00113455, -0.00092456, -0.00062301, -0.00027815, 0.00005715, 0.00033341, 0.00051204, 0.00057108, 0.00050833, 0.00034143, 0.00010486, -0.00015564, -0.00039042, -0.00055389, -0.00061178, -0.00054677, -0.00036143, -0.00007811, 0.00026423, 0.00061580, 0.00092344, 0.00113875, 0.00122549, 0.00116522, 0.00096032, 0.00063390, 0.00022658, -0.00020934, },
		{ 0.00005051, 0.00032208, 0.00052553, 0.00063191, 0.00062728, 0.00051472, 0.00031380, 0.00005749, -0.00021296, -0.00045434, -0.00062821, -0.00070676, -0.00067710, -0.00054330, -0.00032572, -0.00005799, 0.00021821, 0.00045945, 0.00062730, 0.00069416, 0.00064762, 0.00049237, 0.00024962, -0.00004605, -0.00035188, -0.00062331, -0.00082074, -0.00091544, -0.00089392, -0.00075991, -0.00053382, -0.00024957, },
		{ -0.00009476, -0.00034165, -0.00055385, -0.00070188, -0.00076584, -0.00073836, -0.00062558, -0.00044619, -0.00022851, -0.00000612, 0.00018726, 0.00032286, 0.00038126, 0.00035531, 0.00025110, 0.00008697, -0.00010934, -0.00030507, -0.00046751, -0.00056910, -0.00059172, -0.00052959, -0.00039019, -0.00019323, 0.00003231, 0.00025254, 0.00043380, 0.00054782, 0.00057600, 0.00051234, 0.00036438, 0.00015216, },
		{ -0.00036709, -0.00019008, -0.00001646, 0.00012949, 0.00022838, 0.00026853, 0.00024768, 0.00017308, 0.00006035, -0.00006916, -0.00019172, -0.00028502, -0.00033169, -0.00032206, -0.00025577, -0.00014184, 0.00000262, 0.00015504, 0.00029079, 0.00038708, 0.00042640, 0.00039938, 0.00030631, 0.00015726, -0.00002932, -0.00022924, -0.00041605, -0.00056498, -0.00065652, -0.00067928, -0.00063165, -0.00052194, },
		{ 0.00014041, 0.00005213, -0.00004010, -0.00012010, -0.00017402, -0.00019253, -0.00017226, -0.00011635, -0.00003397, 0.00006110, 0.00015264, 0.00022448, 0.00026310, 0.00025975, 0.00021189, 0.00012374, 0.00000570, -0.00012709, -0.00025694, -0.00036623, -0.00043998, -0.00046810, -0.00044685, -0.00037941, -0.00027539, -0.00014940, -0.00001885, 0.00009859, 0.00018765, 0.00023763, 0.00024396, 0.00020876, },
		{ -0.00006923, -0.00000810, 0.00005500, 0.00010915, 0.00014481, 0.00015537, 0.00013814, 0.00009488, 0.00003147, -0.00004287, -0.00011701, -0.00017960, -0.00022080, -0.00023376, -0.00021572, -0.00016845, -0.00009801, -0.00001387, 0.00007246, 0.00014917, 0.00020585, 0.00023508, 0.00023346, 0.00020214, 0.00014656, 0.00007563, 0.00000033, -0.00006792, -0.00011907, -0.00014590, -0.00014510, -0.00011778, },
		{ 0.00005592, 0.00001723, -0.00002163, -0.00005320, -0.00007109, -0.00007109, -0.00005179, -0.00001489, 0.00003501, 0.00009107, 0.00014527, 0.00018952, 0.00021698, 0.00022305, 0.00020610, 0.00016773, 0.00011260, 0.00004774, -0.00001846, -0.00007749, -0.00012191, -0.00014649, -0.00014886, -0.00012991, -0.00009354, -0.00004606, 0.00000481, 0.00005115, 0.00008601, 0.00010451, 0.00010454, 0.00008706, },
		{ 0.00006611, 0.00010412, 0.00013969, 0.00016687, 0.00018061, 0.00017766, 0.00015703, 0.00012022, 0.00007107, 0.00001523, -0.00004063, -0.00008976, -0.00012630, -0.00014609, -0.00014729, -0.00013056, -0.00009895, -0.00005746, -0.00001220, 0.00003050, 0.00006502, 0.00008728, 0.00009532, 0.00008957, 0.00007265, 0.00004899, 0.00002400, 0.00000326, -0.00000846, -0.00000797, 0.00000581, 0.00003164, },
		{ 0.00003927, 0.00000243, -0.00003585, -0.00007073, -0.00009772, -0.00011337, -0.00011575, -0.00010472, -0.00008197, -0.00005073, -0.00001532, 0.00001953, 0.00004934, 0.00007054, 0.00008099, 0.00008025, 0.00006960, 0.00005181, 0.00003069, 0.00001042, -0.00000509, -0.00001282, -0.00001113, -0.00000006, 0.00001871, 0.00004209, 0.00006609, 0.00008639, 0.00009908, 0.00010126, 0.00009150, 0.00007014, },
		{ -0.00004511, -0.00003200, -0.00001508, 0.00000334, 0.00002073, 0.00003468, 0.00004327, 0.00004536, 0.00004078, 0.00003030, 0.00001554, -0.00000128, -0.00001766, -0.00003125, -0.00004014, -0.00004319, -0.00004018, -0.00003181, -0.00001964, -0.00000579, 0.00000737, 0.00001760, 0.00002313, 0.00002295, 0.00001694, 0.00000596, -0.00000837, -0.00002384, -0.00003801, -0.00004860, -0.00005382, -0.00005268, },
		{ -0.00000676, -0.00000587, -0.00000719, -0.00001115, -0.00001769, -0.00002630, -0.00003608, -0.00004581, -0.00005420, -0.00006001, -0.00006228, -0.00006046, -0.00005452, -0.00004494, -0.00003268, -0.00001902, -0.00000544, 0.00000665, 0.00001603, 0.00002189, 0.00002390, 0.00002224, 0.00001755, 0.00001084, 0.00000330, -0.00000387, -0.00000966, -0.00001339, -0.00001480, -0.00001410, -0.00001190, -0.00000910, },
		{ -0.00001408, -0.00002872, -0.00004440, -0.00005830, -0.00006777, -0.00007072, -0.00006599, -0.00005355, -0.00003453, -0.00001108, 0.00001393, 0.00003734, 0.00005610, 0.00006778, 0.00007088, 0.00006508, 0.00005124, 0.00003132, 0.00000804, -0.00001547, -0.00003620, -0.00005162, -0.00006009, -0.00006109, -0.00005521, -0.00004411, -0.00003017, -0.00001615, -0.00000472, 0.00000197, 0.00000261, -0.00000303, },
		{ -0.00006642, -0.00004564, -0.00001722, 0.00001492, 0.00004621, 0.00007210, 0.00008880, 0.00009380, 0.00008627, 0.00006719, 0.00003921, 0.00000624, -0.00002710, -0.00005621, -0.00007716, -0.00008727, -0.00008553, -0.00007269, -0.00005116, -0.00002461, 0.00000260, 0.00002610, 0.00004217, 0.00004833, 0.00004371, 0.00002918, 0.00000721, -0.00001852, -0.00004367, -0.00006396, -0.00007578, -0.00007683, },
		},
		{ { -0.00011383, -0.00009338, -0.00003789, 0.00003679, 0.00010973, 0.00016115, 0.00017827, 0.00015871, 0.00011059, 0.00004933, -0.00000765, -0.00004653, -0.00006117, -0.00005430, -0.00003563, -0.00001756, -0.00001025, -0.00001756, -0.00003563, -0.00005430, -0.00006117, -0.00004653, -0.00000765, 0.00004933, 0.00011059, 0.00015871, 0.00017827, 0.00016115, 0.00010973, 0.00003679, -0.00003789, -0.00009338, },
		{ 0.00021564, 0.00022648, 0.00019496, 0.00012914, 0.00004578, -0.00003472, -0.00009427, -0.00012206, -0.00011708, -0.00008742, -0.00004681, -0.00000966, 0.00001375, 0.00002026, 0.00001406, 0.00000428, 0.00000084, 0.00001019, 0.00003238, 0.00006078, 0.00008432, 0.00009172, 0.00007600, 0.00003783, -0.00001378, -0.00006363, -0.00009491, -0.00009463, -0.00005823, 0.00000837, 0.00008984, 0.00016559, },
		{ -0.00005861, -0.00012514, -0.00015969, -0.00015756, -0.00012338, -0.00006928, -0.00001094, 0.00003709, 0.00006529, 0.00007128, 0.00005941, 0.00003829, 0.00001727, 0.00000295, -0.00000276, -0.00000271, -0.00000256, -0.00000779, -0.00002072, -0.00003883, -0.00005504, -0.00005996, -0.00004541, -0.00000792, 0.00004900, 0.00011477, 0.00017412, 0.00021121, 0.00021426, 0.00017940, 0.00011222, 0.00002665, },
		{ -0.00008715, -0.00002184, 0.00004097, 0.00008693, 0.00010679, 0.00009847, 0.00006712, 0.00002306, -0.00002160, -0.00005644, -0.00007541, -0.00007778, -0.00006735, -0.00005019, -0.00003202, -0.00001602, -0.00000207, 0.00001266, 0.00003167, 0.00005685, 0.00008664, 0.00011554, 0.00013528, 0.00013731, 0.00011590, 0.00007070, 0.00000785, -0.00006092, -0.00012099, -0.00015873, -0.00016530, -0.00013925, },
		{ 0.00012638, 0.00009797, 0.00004313, -0.00002250, -0.00008080, -0.00011656, -0.00012185, -0.00009795, -0.00005443, -0.00000558, 0.00003434, 0.00005559, 0.00005570, 0.00003953, 0.00001686, -0.00000163, -0.00000849, -0.00000230, 0.00001193, 0.00002474, 0.00002605, 0.00000934, -0.00002534, -0.00007063, -0.00011371, -0.00014034, -0.00013964, -0.00010808, -0.00005121, 0.00001744, 0.00008009, 0.00011997, },
		{ -0.00012661, -0.00014270, -0.00011490, -0.00005472, 0.00001571, 0.00007116, 0.00009178, 0.00006967, 0.00001143, -0.00006418, -0.00013268, -0.00017214, -0.00017055, -0.00012983, -0.00006495, 0.00000123, 0.00004589, 0.00005413, 0.00002409, -0.00003235, -0.00009364, -0.00013596, -0.00014143, -0.00010447, -0.00003394, 0.00004953, 0.00012026, 0.00015597, 0.00014526, 0.00009139, 0.00001136, -0.00006966, },
		{ 0.00007626, 0.00009910, 0.00006006, -0.00003374, -0.00015536, -0.00026675, -0.00033138, -0.00032652, -0.00025114, -0.00012652, 0.00001038, 0.00011896, 0.00016815, 0.00014645, 0.00006558, -0.00004365, -0.00014169, -0.00019348, -0.00018000, -0.00010448, 0.00000892, 0.00012339, 0.00020194, 0.00021973, 0.00017226, 0.00007659, -0.00003476, -0.00012506, -0.00016601, -0.00014740, -0.00008045, 0.00000613, },
		{ -0.00035228, -0.00043716, -0.00042924, -0.00032362, -0.00014680, 0.00005126, 0.00021307, 0.00029216, 0.00026824, 0.00015381, -0.00000995, -0.00016674, -0.00026351, -0.00026778, -0.00017797, -0.00002353, 0.00014519, 0.00027381, 0.00032190, 0.00027603, 0.00015360, -0.00000371, -0.00014460, -0.00022580, -0.00022669, -0.00015601, -0.00004825, 0.00004879, 0.00009117, 0.00005469, -0.00005604, -0.00020837, },
		{ 0.00010340, 0.00029823, 0.00039225, 0.00035758, 0.00020775, -0.00000630, -0.00021274, -0.00034233, -0.00035120, -0.00023520, -0.00003122, 0.00019499, 0.00037051, 0.00043910, 0.00037935, 0.00021113, -0.00001169, -0.00021966, -0.00035111, -0.00037244, -0.00028867, -0.00014069, 0.00000974, 0.00010192, 0.00009765, -0.00000522, -0.00017177, -0.00034197, -0.00045143, -0.00045396, -0.00033821, -0.00013256, },
		{ -0.00003527, -0.00027128, -0.00041389, -0.00041140, -0.00025987, -0.00000526, 0.00027104, 0.00047858, 0.00054744, 0.00045122, 0.00021578, -0.00008923, -0.00037304, -0.00055328, -0.00058240, -0.00046227, -0.00024239, -0.00000218, 0.00017621, 0.00023566, 0.00016278, -0.00000844, -0.00020930, -0.00036083, -0.00040164, -0.00030962, -0.00010997, 0.00013311, 0.00033827, 0.00043512, 0.00038879, 0.00021266, },
		{ 0.00005181, 0.00035975, 0.00057504, 0.00061832, 0.00046262, 0.00014364, -0.00025026, -0.00060539, -0.00081851, -0.00082933, -0.00063919, -0.00030977, 0.00005751, 0.00035454, 0.00050097, 0.00046888, 0.00028950, 0.00004027, -0.00018317, -0.00030021, -0.00027266, -0.00011670, 0.00010412, 0.00030246, 0.00039895, 0.00035106, 0.00016907, -0.00008659, -0.00032573, -0.00045973, -0.00043308, -0.00024420, },
		{ -0.00000997, -0.00048453, -0.00089120, -0.00110836, -0.00106654, -0.00077055, -0.00029831, 0.00022312, 0.00065659, 0.00089647, 0.00089921, 0.00069350, 0.00036719, 0.00003571, -0.00019683, -0.00027107, -0.00018912, -0.00000960, 0.00017670, 0.00028170, 0.00025302, 0.00009346, -0.00013982, -0.00035614, -0.00046482, -0.00040966, -0.00019202, 0.00012583, 0.00043901, 0.00063405, 0.00062764, 0.00039678, },
		{ -0.00081152, -0.00017859, 0.00050384, 0.00106946, 0.00138830, 0.00140298, 0.00114200, 0.00070631, 0.00023419, -0.00014343, -0.00034269, -0.00034866, -0.00021304, -0.00002951, 0.00010382, 0.00012145, 0.00001389, -0.00017055, -0.00034585, -0.00042295, -0.00034621, -0.00011849, 0.00019480, 0.00048475, 0.00063527, 0.00056439, 0.00025653, -0.00022603, -0.00075723, -0.00118315, -0.00136880, -0.00124069, },
		{ 0.00190161, 0.00155413, 0.00098352, 0.00035262, -0.00017992, -0.00050544, -0.00059208, -0.00048527, -0.00028301, -0.00009583, -0.00000568, -0.00003834, -0.00015832, -0.00028743, -0.00033960, -0.00025887, -0.00004617, 0.00023525, 0.00047870, 0.00057113, 0.00043589, 0.00006597, -0.00046461, -0.00101646, -0.00142381, -0.00154493, -0.00130791, -0.00073679, 0.00005098, 0.00087552, 0.00154491, 0.00190921, },
		{ -0.00074920, -0.00088901, -0.00078237, -0.00053425, -0.00026990, -0.00008989, -0.00003819, -0.00009355, -0.00018602, -0.00023149, -0.00017082, 0.00000140, 0.00022957, 0.00041459, 0.00045052, 0.00026753, -0.00013446, -0.00067439, -0.00120659, -0.00156128, -0.00159699, -0.00124796, -0.00055076, 0.00035935, 0.00128228, 0.00200750, 0.00237239, 0.00230674, 0.00184992, 0.00113586, 0.00035104, -0.00032072, },
		{ -0.00010334, 0.00002517, 0.00001869, -0.00005888, -0.00012245, -0.00010599, 0.00000659, 0.00017161, 0.00030099, 0.00029742, 0.00009676, -0.00029829, -0.00080635, -0.00128355, -0.00156403, -0.00151246, -0.00107167, -0.00028900, 0.00068822, 0.00164832, 0.00237185, 0.00269230, 0.00254230, 0.00197122, 0.00112896, 0.00022135, -0.00054873, -0.00103621, -0.00118812, -0.00104783, -0.00072993, -0.00037543, },
		{ -0.00008218, -0.00003699, 0.00006897, 0.00015945, 0.00014475, -0.00003697, -0.00038582, -0.00082739, -0.00122732, -0.00143013, -0.00131036, -0.00081912, -0.00000988, 0.00096638, 0.00189695, 0.00256425, 0.00280655, 0.00256425, 0.00189695, 0.00096638, -0.00000988, -0.00081912, -0.00131036, -0.00143013, -0.00122732, -0.00082739, -0.00038582, -0.00003697, 0.00014475, 0.00015945, 0.00006897, -0.00003699, },
		{ -0.00010334, -0.00037543, -0.00072993, -0.00104783, -0.00118812, -0.00103621, -0.00054873, 0.00022135, 0.00112896, 0.00197122, 0.00254230, 0.00269230, 0.00237185, 0.00164832, 0.00068822, -0.00028900, -0.00107167, -0.00151246, -0.00156403, -0.00128355, -0.00080635, -0.00029829, 0.00009676, 0.00029742, 0.00030099, 0.00017161, 0.00000659, -0.00010599, -0.00012245, -0.00005888, 0.00001869, 0.00002517, },
		{ -0.00074920, -0.00032072, 0.00035104, 0.00113586, 0.00184992, 0.00230674, 0.00237239, 0.00200750, 0.00128228, 0.00035935, -0.00055076, -0.00124796, -0.00159699, -0.00156128, -0.00120659, -0.00067439, -0.00013446, 0.00026753, 0.00045052, 0.00041459, 0.00022957, 0.00000140, -0.00017082, -0.00023149, -0.00018602, -0.00009355, -0.00003819, -0.00008989, -0.00026990, -0.00053425, -0.00078237, -0.00088901, },
		{ 0.00190161, 0.00190921, 0.00154491, 0.00087552, 0.00005098, -0.00073679, -0.00130791, -0.00154493, -0.00142381, -0.00101646, -0.00046461, 0.00006597, 0.00043589, 0.00057113, 0.00047870, 0.00023525, -0.00004617, -0.00025887, -0.00033960, -0.00028743, -0.00015832, -0.00003834, -0.00000568, -0.00009583, -0.00028301, -0.00048527, -0.00059208, -0.00050544, -0.00017992, 0.00035262, 0.00098352, 0.00155413, },
		{ -0.00081152, -0.00124069, -0.00136880, -0.00118315, -0.00075723, -0.00022603, 0.00025653, 0.00056439, 0.00063527, 0.00048475, 0.00019480, -0.00011849, -0.00034621, -0.00042295, -0.00034585, -0.00017055, 0.00001389, 0.00012145, 0.00010382, -0.00002951, -0.00021304, -0.00034866, -0.00034269, -0.00014343, 0.00023419, 0.00070631, 0.00114200, 0.00140298, 0.00138830, 0.00106946, 0.00050384, -0.00017859, },
		{ -0.00000997, 0.00039678, 0.00062764, 0.00063405, 0.00043901, 0.00012583, -0.00019202, -0.00040966, -0.00046482, -0.00035614, -0.00013982, 0.00009346, 0.00025302, 0.00028170, 0.00017670, -0.00000960, -0.00018912, -0.00027107, -0.00019683, 0.00003571, 0.00036719, 0.00069350, 0.00089921, 0.00089647, 0.00065659, 0.00022312, -0.00029831, -0.00077055, -0.00106654, -0.00110836, -0.00089120, -0.00048453, },
		{ 0.00005181, -0.00024420, -0.00043308, -0.00045973, -0.00032573, -0.00008659, 0.00016907, 0.00035106, 0.00039895, 0.00030246, 0.00010412, -0.00011670, -0.00027266, -0.00030021, -0.00018317, 0.00004027, 0.00028950, 0.00046888, 0.00050097, 0.00035454, 0.00005751, -0.00030977, -0.00063919, -0.00082933, -0.00081851, -0.00060539, -0.00025026, 0.00014364, 0.00046262, 0.00061832, 0.00057504, 0.00035975, },
		{ -0.00003527, 0.00021266, 0.00038879, 0.00043512, 0.00033827, 0.00013311, -0.00010997, -0.00030962, -0.00040164, -0.00036083, -0.00020930, -0.00000844, 0.00016278, 0.00023566, 0.00017621, -0.00000218, -0.00024239, -0.00046227, -0.00058240, -0.00055328, -0.00037304, -0.00008923, 0.00021578, 0.00045122, 0.00054744, 0.00047858, 0.00027104, -0.00000526, -0.00025987, -0.00041140, -0.00041389, -0.00027128, },
		{ 0.00010340, -0.00013256, -0.00033821, -0.00045396, -0.00045143, -0.00034197, -0.00017177, -0.00000522, 0.00009765, 0.00010192, 0.00000974, -0.00014069, -0.00028867, -0.00037244, -0.00035111, -0.00021966, -0.00001169, 0.00021113, 0.00037935, 0.00043910, 0.00037051, 0.00019499, -0.00003122, -0.00023520, -0.00035120, -0.00034233, -0.00021274, -0.00000630, 0.00020775, 0.00035758, 0.00039225, 0.00029823, },
		{ -0.00035228, -0.00020837, -0.00005604, 0.00005469, 0.00009117, 0.00004879, -0.00004825, -0.00015601, -0.00022669, -0.00022580, -0.00014460, -0.00000371, 0.00015360, 0.00027603, 0.00032190, 0.00027381, 0.00014519, -0.00002353, -0.00017797, -0.00026778, -0.00026351, -0.00016674, -0.00000995, 0.00015381, 0.00026824, 0.00029216, 0.00021307, 0.00005126, -0.00014680, -0.00032362, -0.00042924, -0.00043716, },
		{ 0.00007626, 0.00000613, -0.00008045, -0.00014740, -0.00016601, -0.00012506, -0.00003476, 0.00007659, 0.00017226, 0.00021973, 0.00020194, 0.00012339, 0.00000892, -0.00010448, -0.00018000, -0.00019348, -0.00014169, -0.00004365, 0.00006558, 0.00014645, 0.00016815, 0.00011896, 0.00001038, -0.00012652, -0.00025114, -0.00032652, -0.00033138, -0.00026675, -0.00015536, -0.00003374, 0.00006006, 0.00009910, },
		{ -0.00012661, -0.00006966, 0.00001136, 0.00009139, 0.00014526, 0.00015597, 0.00012026, 0.00004953, -0.00003394, -0.00010447, -0.00014143, -0.00013596, -0.00009364, -0.00003235, 0.00002409, 0.00005413, 0.00004589, 0.00000123, -0.00006495, -0.00012983, -0.00017055, -0.00017214, -0.00013268, -0.00006418, 0.00001143, 0.00006967, 0.00009178, 0.00007116, 0.00001571, -0.00005472, -0.00011490, -0.00014270, },
		{ 0.00012638, 0.00011997, 0.00008009, 0.00001744, -0.00005121, -0.00010808, -0.00013964, -0.00014034, -0.00011371, -0.00007063, -0.00002534, 0.00000934, 0.00002605, 0.00002474, 0.00001193, -0.00000230, -0.00000849, -0.00000163, 0.00001686, 0.00003953, 0.00005570, 0.00005559, 0.00003434, -0.00000558, -0.00005443, -0.00009795, -0.00012185, -0.00011656, -0.00008080, -0.00002250, 0.00004313, 0.00009797, },
		{ -0.00008715, -0.00013925, -0.00016530, -0.00015873, -0.00012099, -0.00006092, 0.00000785, 0.00007070, 0.00011590, 0.00013731, 0.00013528, 0.00011554, 0.00008664, 0.00005685, 0.00003167, 0.00001266, -0.00000207, -0.00001602, -0.00003202, -0.00005019, -0.00006735, -0.00007778, -0.00007541, -0.00005644, -0.00002160, 0.00002306, 0.00006712, 0.00009847, 0.00010679, 0.00008693, 0.00004097, -0.00002184, },
		{ -0.00005861, 0.00002665, 0.00011222, 0.00017940, 0.00021426, 0.00021121, 0.00017412, 0.00011477, 0.00004900, -0.00000792, -0.00004541, -0.00005996, -0.00005504, -0.00003883, -0.00002072, -0.00000779, -0.00000256, -0.00000271, -0.00000276, 0.00000295, 0.00001727, 0.00003829, 0.00005941, 0.00007128, 0.00006529, 0.00003709, -0.00001094, -0.00006928, -0.00012338, -0.00015756, -0.00015969, -0.00012514, },
		{ 0.00021564, 0.00016559, 0.00008984, 0.00000837, -0.00005823, -0.00009463, -0.00009491, -0.00006363, -0.00001378, 0.00003783, 0.00007600, 0.00009172, 0.00008432, 0.00006078, 0.00003238, 0.00001019, 0.00000084, 0.00000428, 0.00001406, 0.00002026, 0.00001375, -0.00000966, -0.00004681, -0.00008742, -0.00011708, -0.00012206, -0.00009427, -0.00003472, 0.00004578, 0.00012914, 0.00019496, 0.00022648, },
		},
		{ { -0.00009944, -0.00007222, -0.00000569, 0.00006349, 0.00009817, 0.00008194, 0.00002800, -0.00002923, -0.00005485, -0.00003428, 0.00001809, 0.00006755, 0.00007986, 0.00004196, -0.00002961, -0.00009762, -0.00012534, -0.00009762, -0.00002961, 0.00004196, 0.00007986, 0.00006755, 0.00001809, -0.00003428, -0.00005485, -0.00002923, 0.00002800, 0.00008194, 0.00009817, 0.00006349, -0.00000569, -0.00007222, },
		{ 0.00006422, 0.00009120, 0.00007068, 0.00002396, -0.00001436, -0.00001868, 0.00001159, 0.00005199, 0.00006877, 0.00004094, -0.00002511, -0.00009750, -0.00013660, -0.00011848, -0.00004991, 0.00003405, 0.00008920, 0.00008724, 0.00003214, -0.00004174, -0.00008903, -0.00007934, -0.00001494, 0.00007006, 0.00012870, 0.00012754, 0.00006548, -0.00002571, -0.00009914, -0.00011868, -0.00007817, -0.00000335, },
		{ 0.00002901, -0.00000134, 0.00000500, 0.00003413, 0.00005287, 0.00003209, -0.00003204, -0.00011246, -0.00016561, -0.00015736, -0.00008522, 0.00001731, 0.00009767, 0.00011352, 0.00005748, -0.00003724, -0.00011435, -0.00012579, -0.00005926, 0.00005377, 0.00015449, 0.00018858, 0.00013592, 0.00002150, -0.00009824, -0.00016583, -0.00015281, -0.00007290, 0.00002831, 0.00010028, 0.00011508, 0.00008005, },
		{ 0.00002754, 0.00001745, -0.00004171, -0.00012958, -0.00019881, -0.00020359, -0.00012964, -0.00000698, 0.00010372, 0.00014430, 0.00009315, -0.00002020, -0.00012937, -0.00016740, -0.00010476, 0.00003234, 0.00017447, 0.00024587, 0.00020549, 0.00006957, -0.00009678, -0.00021411, -0.00023006, -0.00014493, -0.00000803, 0.00011170, 0.00016343, 0.00013898, 0.00007136, 0.00000988, -0.00001177, 0.00000490, },
		{ -0.00021634, -0.00024032, -0.00017307, -0.00003648, 0.00010360, 0.00017316, 0.00013348, 0.00000654, -0.00013398, -0.00020237, -0.00014970, 0.00000629, 0.00018785, 0.00029793, 0.00027288, 0.00011784, -0.00009481, -0.00026327, -0.00030952, -0.00021861, -0.00004326, 0.00012761, 0.00021748, 0.00020051, 0.00010885, 0.00000707, -0.00004926, -0.00004559, -0.00001241, -0.00000030, -0.00004279, -0.00013100, },
		{ 0.00009211, 0.00019045, 0.00016903, 0.00003718, -0.00012923, -0.00022836, -0.00019083, -0.00002252, 0.00019413, 0.00034252, 0.00033552, 0.00016471, -0.00009241, -0.00031217, -0.00038950, -0.00029230, -0.00007594, 0.00014922, 0.00027793, 0.00026392, 0.00013818, -0.00001628, -0.00011635, -0.00012601, -0.00007163, -0.00001826, -0.00002335, -0.00009728, -0.00019560, -0.00024751, -0.00020285, -0.00006810, },
		{ -0.00011721, -0.00024273, -0.00022381, -0.00005114, 0.00019316, 0.00037674, 0.00038960, 0.00020765, -0.00008968, -0.00035868, -0.00046693, -0.00036342, -0.00010478, 0.00017655, 0.00034396, 0.00032775, 0.00015695, -0.00006405, -0.00021837, -0.00024181, -0.00015305, -0.00003381, 0.00002685, -0.00001107, -0.00011546, -0.00020548, -0.00020596, -0.00009843, 0.00006352, 0.00018490, 0.00018810, 0.00006392, },
		{ 0.00018512, 0.00039730, 0.00043051, 0.00024338, -0.00008671, -0.00039981, -0.00053746, -0.00042850, -0.00012872, 0.00020827, 0.00041304, 0.00038976, 0.00016400, -0.00013657, -0.00035569, -0.00039377, -0.00025706, -0.00004493, 0.00011428, 0.00013937, 0.00003960, -0.00009866, -0.00017129, -0.00012418, 0.00001274, 0.00014572, 0.00017878, 0.00007826, -0.00010057, -0.00024292, -0.00024376, -0.00007592, },
		{ -0.00008362, -0.00043201, -0.00059582, -0.00048344, -0.00014691, 0.00024179, 0.00048100, 0.00044696, 0.00015951, -0.00023033, -0.00052302, -0.00057713, -0.00038087, -0.00005055, 0.00024013, 0.00035746, 0.00027603, 0.00008120, -0.00009178, -0.00014266, -0.00006349, 0.00006463, 0.00013108, 0.00007246, -0.00008218, -0.00022681, -0.00024597, -0.00009301, 0.00017048, 0.00040102, 0.00045333, 0.00026830, },
		{ -0.00015879, 0.00027357, 0.00054246, 0.00049581, 0.00014489, -0.00033818, -0.00070965, -0.00078163, -0.00051851, -0.00005069, 0.00039956, 0.00063671, 0.00058913, 0.00033285, 0.00003406, -0.00015210, -0.00016574, -0.00006225, 0.00003895, 0.00004093, -0.00006462, -0.00019316, -0.00022674, -0.00009896, 0.00014994, 0.00038523, 0.00045358, 0.00027890, -0.00008053, -0.00045161, -0.00063644, -0.00052390, },
		{ 0.00012260, -0.00045017, -0.00090051, -0.00099239, -0.00066119, -0.00004636, 0.00058189, 0.00096075, 0.00096248, 0.00064470, 0.00020141, -0.00015192, -0.00029131, -0.00023328, -0.00009837, -0.00001865, -0.00004983, -0.00014192, -0.00018405, -0.00009132, 0.00012444, 0.00034831, 0.00042791, 0.00027234, -0.00007756, -0.00045533, -0.00065415, -0.00054581, -0.00016406, 0.00029962, 0.00059144, 0.00053257, },
		{ -0.00079820, -0.00003946, 0.00077172, 0.00130476, 0.00136933, 0.00099560, 0.00039928, -0.00014279, -0.00043415, -0.00044079, -0.00027558, -0.00010474, -0.00003880, -0.00007445, -0.00011801, -0.00006899, 0.00009505, 0.00029008, 0.00037475, 0.00024692, -0.00007483, -0.00044077, -0.00064492, -0.00054591, -0.00016268, 0.00031603, 0.00062213, 0.00055372, 0.00009570, -0.00055481, -0.00107793, -0.00119155, },
		{ 0.00177544, 0.00135698, 0.00061166, -0.00012644, -0.00058526, -0.00067159, -0.00048161, -0.00021184, -0.00003153, 0.00000645, -0.00003119, -0.00003255, 0.00006291, 0.00021203, 0.00029477, 0.00020240, -0.00007243, -0.00040687, -0.00060656, -0.00052224, -0.00015483, 0.00031962, 0.00062972, 0.00055642, 0.00006744, -0.00064059, -0.00122391, -0.00136043, -0.00091809, -0.00003237, 0.00095077, 0.00163833, },
		{ -0.00073363, -0.00090826, -0.00070064, -0.00033108, -0.00002703, 0.00009668, 0.00007160, 0.00001575, 0.00002925, 0.00011743, 0.00019104, 0.00014025, -0.00007046, -0.00035422, -0.00053917, -0.00047453, -0.00014090, 0.00030839, 0.00061117, 0.00053895, 0.00004079, -0.00069760, -0.00132252, -0.00148196, -0.00101011, -0.00002756, 0.00110028, 0.00192926, 0.00214328, 0.00169631, 0.00081941, -0.00010547, },
		{ -0.00002361, 0.00019114, 0.00018362, 0.00007221, -0.00000474, 0.00001109, 0.00006893, 0.00006360, -0.00006900, -0.00028507, -0.00044533, -0.00040436, -0.00012147, 0.00028196, 0.00056567, 0.00050098, 0.00001802, -0.00071888, -0.00136205, -0.00154289, -0.00106547, -0.00002717, 0.00120352, 0.00214772, 0.00243678, 0.00198118, 0.00100267, -0.00008288, -0.00086736, -0.00113110, -0.00091390, -0.00045124, },
		{ -0.00003788, -0.00010094, -0.00006432, -0.00002300, -0.00006810, -0.00020328, -0.00033002, -0.00031516, -0.00009724, 0.00024160, 0.00049497, 0.00044376, 0.00000045, -0.00070144, -0.00133670, -0.00153568, -0.00107857, -0.00003262, 0.00124813, 0.00227024, 0.00262607, 0.00218343, 0.00114342, -0.00006166, -0.00097516, -0.00132053, -0.00110203, -0.00056023, -0.00001918, 0.00028417, 0.00029694, 0.00013211, },
		{ -0.00006779, -0.00011397, -0.00020025, -0.00021195, -0.00006906, 0.00019024, 0.00040330, 0.00037010, -0.00001169, -0.00064667, -0.00124746, -0.00145951, -0.00104766, -0.00004434, 0.00122783, 0.00228295, 0.00269146, 0.00228295, 0.00122783, -0.00004434, -0.00104766, -0.00145951, -0.00124746, -0.00064667, -0.00001169, 0.00037010, 0.00040330, 0.00019024, -0.00006906, -0.00021195, -0.00020025, -0.00011397, },
		{ -0.00003788, 0.00013211, 0.00029694, 0.00028417, -0.00001918, -0.00056023, -0.00110203, -0.00132053, -0.00097516, -0.00006166, 0.00114342, 0.00218343, 0.00262607, 0.00227024, 0.00124813, -0.00003262, -0.00107857, -0.00153568, -0.00133670, -0.00070144, 0.00000045, 0.00044376, 0.00049497, 0.00024160, -0.00009724, -0.00031516, -0.00033002, -0.00020328, -0.00006810, -0.00002300, -0.00006432, -0.00010094, },
		{ -0.00002361, -0.00045124, -0.00091390, -0.00113110, -0.00086736, -0.00008288, 0.00100267, 0.00198118, 0.00243678, 0.00214772, 0.00120352, -0.00002717, -0.00106547, -0.00154289, -0.00136205, -0.00071888, 0.00001802, 0.00050098, 0.00056567, 0.00028196, -0.00012147, -0.00040436, -0.00044533, -0.00028507, -0.00006900, 0.00006360, 0.00006893, 0.00001109, -0.00000474, 0.00007221, 0.00018362, 0.00019114, },
		{ -0.00073363, -0.00010547, 0.00081941, 0.00169631, 0.00214328, 0.00192926, 0.00110028, -0.00002756, -0.00101011, -0.00148196, -0.00132252, -0.00069760, 0.00004079, 0.00053895, 0.00061117, 0.00030839, -0.00014090, -0.00047453, -0.00053917, -0.00035422, -0.00007046, 0.00014025, 0.00019104, 0.00011743, 0.00002925, 0.00001575, 0.00007160, 0.00009668, -0.00002703, -0.00033108, -0.00070064, -0.00090826, },
		{ 0.00177544, 0.00163833, 0.00095077, -0.00003237, -0.00091809, -0.00136043, -0.00122391, -0.00064059, 0.00006744, 0.00055642, 0.00062972, 0.00031962, -0.00015483, -0.00052224, -0.00060656, -0.00040687, -0.00007243, 0.00020240, 0.00029477, 0.00021203, 0.00006291, -0.00003255, -0.00003119, 0.00000645, -0.00003153, -0.00021184, -0.00048161, -0.00067159, -0.00058526, -0.00012644, 0.00061166, 0.00135698, },
		{ -0.00079820, -0.00119155, -0.00107793, -0.00055481, 0.00009570, 0.00055372, 0.00062213, 0.00031603, -0.00016268, -0.00054591, -0.00064492, -0.00044077, -0.00007483, 0.00024692, 0.00037475, 0.00029008, 0.00009505, -0.00006899, -0.00011801, -0.00007445, -0.00003880, -0.00010474, -0.00027558, -0.00044079, -0.00043415, -0.00014279, 0.00039928, 0.00099560, 0.00136933, 0.00130476, 0.00077172, -0.00003946, },
		{ 0.00012260, 0.00053257, 0.00059144, 0.00029962, -0.00016406, -0.00054581, -0.00065415, -0.00045533, -0.00007756, 0.00027234, 0.00042791, 0.00034831, 0.00012444, -0.00009132, -0.00018405, -0.00014192, -0.00004983, -0.00001865, -0.00009837, -0.00023328, -0.00029131, -0.00015192, 0.00020141, 0.00064470, 0.00096248, 0.00096075, 0.00058189, -0.00004636, -0.00066119, -0.00099239, -0.00090051, -0.00045017, },
		{ -0.00015879, -0.00052390, -0.00063644, -0.00045161, -0.00008053, 0.00027890, 0.00045358, 0.00038523, 0.00014994, -0.00009896, -0.00022674, -0.00019316, -0.00006462, 0.00004093, 0.00003895, -0.00006225, -0.00016574, -0.00015210, 0.00003406, 0.00033285, 0.00058913, 0.00063671, 0.00039956, -0.00005069, -0.00051851, -0.00078163, -0.00070965, -0.00033818, 0.00014489, 0.00049581, 0.00054246, 0.00027357, },
		{ -0.00008362, 0.00026830, 0.00045333, 0.00040102, 0.00017048, -0.00009301, -0.00024597, -0.00022681, -0.00008218, 0.00007246, 0.00013108, 0.00006463, -0.00006349, -0.00014266, -0.00009178, 0.00008120, 0.00027603, 0.00035746, 0.00024013, -0.00005055, -0.00038087, -0.00057713, -0.00052302, -0.00023033, 0.00015951, 0.00044696, 0.00048100, 0.00024179, -0.00014691, -0.00048344, -0.00059582, -0.00043201, },
		{ 0.00018512, -0.00007592, -0.00024376, -0.00024292, -0.00010057, 0.00007826, 0.00017878, 0.00014572, 0.00001274, -0.00012418, -0.00017129, -0.00009866, 0.00003960, 0.00013937, 0.00011428, -0.00004493, -0.00025706, -0.00039377, -0.00035569, -0.00013657, 0.00016400, 0.00038976, 0.00041304, 0.00020827, -0.00012872, -0.00042850, -0.00053746, -0.00039981, -0.00008671, 0.00024338, 0.00043051, 0.00039730, },
		{ -0.00011721, 0.00006392, 0.00018810, 0.00018490, 0.00006352, -0.00009843, -0.00020596, -0.00020548, -0.00011546, -0.00001107, 0.00002685, -0.00003381, -0.00015305, -0.00024181, -0.00021837, -0.00006405, 0.00015695, 0.00032775, 0.00034396, 0.00017655, -0.00010478, -0.00036342, -0.00046693, -0.00035868, -0.00008968, 0.00020765, 0.00038960, 0.00037674, 0.00019316, -0.00005114, -0.00022381, -0.00024273, },
		{ 0.00009211, -0.00006810, -0.00020285, -0.00024751, -0.00019560, -0.00009728, -0.00002335, -0.00001826, -0.00007163, -0.00012601, -0.00011635, -0.00001628, 0.00013818, 0.00026392, 0.00027793, 0.00014922, -0.00007594, -0.00029230, -0.00038950, -0.00031217, -0.00009241, 0.00016471, 0.00033552, 0.00034252, 0.00019413, -0.00002252, -0.00019083, -0.00022836, -0.00012923, 0.00003718, 0.00016903, 0.00019045, },
		{ -0.00021634, -0.00013100, -0.00004279, -0.00000030, -0.00001241, -0.00004559, -0.00004926, 0.00000707, 0.00010885, 0.00020051, 0.00021748, 0.00012761, -0.00004326, -0.00021861, -0.00030952, -0.00026327, -0.00009481, 0.00011784, 0.00027288, 0.00029793, 0.00018785, 0.00000629, -0.00014970, -0.00020237, -0.00013398, 0.00000654, 0.00013348, 0.00017316, 0.00010360, -0.00003648, -0.00017307, -0.00024032, },
		{ 0.00002754, 0.00000490, -0.00001177, 0.00000988, 0.00007136, 0.00013898, 0.00016343, 0.00011170, -0.00000803, -0.00014493, -0.00023006, -0.00021411, -0.00009678, 0.00006957, 0.00020549, 0.00024587, 0.00017447, 0.00003234, -0.00010476, -0.00016740, -0.00012937, -0.00002020, 0.00009315, 0.00014430, 0.00010372, -0.00000698, -0.00012964, -0.00020359, -0.00019881, -0.00012958, -0.00004171, 0.00001745, },
		{ 0.00002901, 0.00008005, 0.00011508, 0.00010028, 0.00002831, -0.00007290, -0.00015281, -0.00016583, -0.00009824, 0.00002150, 0.00013592, 0.00018858, 0.00015449, 0.00005377, -0.00005926, -0.00012579, -0.00011435, -0.00003724, 0.00005748, 0.00011352, 0.00009767, 0.00001731, -0.00008522, -0.00015736, -0.00016561, -0.00011246, -0.00003204, 0.00003209, 0.00005287, 0.00003413, 0.00000500, -0.00000134, },
		{ 0.00006422, -0.00000335, -0.00007817, -0.00011868, -0.00009914, -0.00002571, 0.00006548, 0.00012754, 0.00012870, 0.00007006, -0.00001494, -0.00007934, -0.00008903, -0.00004174, 0.00003214, 0.00008724, 0.00008920, 0.00003405, -0.00004991, -0.00011848, -0.00013660, -0.00009750, -0.00002511, 0.00004094, 0.00006877, 0.00005199, 0.00001159, -0.00001868, -0.00001436, 0.00002396, 0.00007068, 0.00009120, },
		},
		{ { 0.00021705, 0.00009432, -0.00017660, -0.00038181, -0.00036038, -0.00012752, 0.00014580, 0.00027369, 0.00019691, 0.00001243, -0.00012253, -0.00012135, -0.00002265, 0.00006687, 0.00007780, 0.00003231, 0.00000414, 0.00003231, 0.00007780, 0.00006687, -0.00002265, -0.00012135, -0.00012253, 0.00001243, 0.00019691, 0.00027369, 0.00014580, -0.00012752, -0.00036038, -0.00038181, -0.00017660, 0.00009432, },
		{ -0.00043308, -0.00040263, -0.00013641, 0.00018277, 0.00034200, 0.00026005, 0.00003668, -0.00014223, -0.00015683, -0.00003494, 0.00009542, 0.00012834, 0.00006475, -0.00000914, -0.00001798, 0.00002916, 0.00005646, 0.00000859, -0.00008236, -0.00012019, -0.00004076, 0.00011108, 0.00020387, 0.00013159, -0.00008149, -0.00028311, -0.00031206, -0.00013783, 0.00010120, 0.00020237, 0.00006508, -0.00021933, },
		{ 0.00022765, 0.00039345, 0.00029205, 0.00003307, -0.00017194, -0.00018188, -0.00002742, 0.00013594, 0.00017178, 0.00007390, -0.00004697, -0.00007791, -0.00000879, 0.00006767, 0.00005484, -0.00004773, -0.00014020, -0.00011942, 0.00001488, 0.00014807, 0.00015026, -0.00000037, -0.00018618, -0.00024692, -0.00012428, 0.00007539, 0.00016337, 0.00003173, -0.00024240, -0.00044666, -0.00040247, -0.00011417, },
		{ 0.00000689, -0.00020323, -0.00019162, -0.00000246, 0.00018054, 0.00020042, 0.00005791, -0.00010365, -0.00013780, -0.00002877, 0.00010063, 0.00010962, -0.00002500, -0.00018421, -0.00021584, -0.00007737, 0.00011946, 0.00020541, 0.00010578, -0.00009040, -0.00020938, -0.00015105, 0.00001639, 0.00011433, 0.00001708, -0.00022407, -0.00041061, -0.00036162, -0.00007374, 0.00026354, 0.00041541, 0.00028957, },
		{ 0.00003483, 0.00022091, 0.00020900, 0.00001882, -0.00017060, -0.00018798, -0.00002624, 0.00015210, 0.00016491, -0.00002017, -0.00025017, -0.00031912, -0.00015149, 0.00012667, 0.00029327, 0.00022036, -0.00001866, -0.00021834, -0.00022425, -0.00006623, 0.00007627, 0.00004441, -0.00014852, -0.00032244, -0.00029056, -0.00003264, 0.00027441, 0.00040006, 0.00025482, -0.00003334, -0.00022799, -0.00018403, },
		{ -0.00003770, -0.00023776, -0.00022027, -0.00000014, 0.00021584, 0.00021229, -0.00003664, -0.00033238, -0.00041666, -0.00019561, 0.00017227, 0.00040316, 0.00032278, 0.00000787, -0.00028520, -0.00033968, -0.00015430, 0.00007319, 0.00013307, -0.00000898, -0.00018980, -0.00020681, -0.00000983, 0.00024794, 0.00034577, 0.00019515, -0.00007734, -0.00023985, -0.00016026, 0.00007735, 0.00024961, 0.00019559, },
		{ 0.00004693, 0.00028351, 0.00024429, -0.00007464, -0.00042226, -0.00049600, -0.00020201, 0.00025231, 0.00051897, 0.00039173, -0.00002660, -0.00041165, -0.00048321, -0.00022403, 0.00012777, 0.00029465, 0.00019058, -0.00002953, -0.00013201, -0.00002237, 0.00017784, 0.00025750, 0.00012187, -0.00011478, -0.00023520, -0.00012444, 0.00011721, 0.00026128, 0.00016187, -0.00010323, -0.00029501, -0.00022929, },
		{ -0.00013104, -0.00050947, -0.00054654, -0.00016834, 0.00035674, 0.00062158, 0.00040859, -0.00012969, -0.00058894, -0.00063281, -0.00025004, 0.00025704, 0.00053024, 0.00043536, 0.00013477, -0.00008822, -0.00008229, 0.00006517, 0.00014616, 0.00004844, -0.00013704, -0.00021385, -0.00008305, 0.00014693, 0.00025348, 0.00011291, -0.00016790, -0.00033378, -0.00021324, 0.00010901, 0.00034592, 0.00025552, },
		{ -0.00009801, 0.00047088, 0.00069188, 0.00036064, -0.00029913, -0.00079885, -0.00076189, -0.00020977, 0.00046881, 0.00082930, 0.00070137, 0.00027629, -0.00009424, -0.00019424, -0.00008144, 0.00002705, -0.00001169, -0.00013858, -0.00017905, -0.00004392, 0.00016067, 0.00022710, 0.00005641, -0.00022184, -0.00034832, -0.00017424, 0.00017766, 0.00039434, 0.00024349, -0.00019978, -0.00058335, -0.00056121, },
		{ 0.00024335, -0.00052265, -0.00101608, -0.00084350, -0.00008767, 0.00075942, 0.00117027, 0.00095862, 0.00036917, -0.00016219, -0.00035420, -0.00024677, -0.00008240, -0.00004720, -0.00011791, -0.00013703, -0.00001490, 0.00015516, 0.00018621, 0.00000146, -0.00025669, -0.00033659, -0.00011800, 0.00024323, 0.00042183, 0.00020903, -0.00027266, -0.00063448, -0.00053742, 0.00000027, 0.00057762, 0.00071373, },
		{ -0.00085473, 0.00012164, 0.00111322, 0.00152292, 0.00117523, 0.00039262, -0.00029526, -0.00054966, -0.00041104, -0.00016526, -0.00005070, -0.00007782, -0.00009600, -0.00000269, 0.00013025, 0.00013746, -0.00004277, -0.00026693, -0.00030051, -0.00005293, 0.00029623, 0.00042426, 0.00015620, -0.00034049, -0.00065599, -0.00047752, 0.00011370, 0.00066002, 0.00067673, 0.00006152, -0.00077926, -0.00121186, },
		{ 0.00185216, 0.00132204, 0.00033431, -0.00048661, -0.00076102, -0.00055277, -0.00020768, -0.00001993, -0.00002503, -0.00006483, -0.00001165, 0.00008894, 0.00008907, -0.00006855, -0.00025065, -0.00024575, 0.00001107, 0.00032883, 0.00040096, 0.00009174, -0.00039446, -0.00064463, -0.00038850, 0.00022728, 0.00070388, 0.00057799, -0.00017091, -0.00104202, -0.00135817, -0.00078058, 0.00041161, 0.00150385, },
		{ -0.00071997, -0.00096422, -0.00065185, -0.00020107, 0.00004179, 0.00003093, -0.00005162, -0.00004309, 0.00003700, 0.00004948, -0.00007073, -0.00020993, -0.00018078, 0.00006418, 0.00033600, 0.00035481, 0.00002406, -0.00042738, -0.00060120, -0.00028098, 0.00032602, 0.00069999, 0.00042292, -0.00043250, -0.00128163, -0.00143196, -0.00061675, 0.00076381, 0.00189715, 0.00212264, 0.00137688, 0.00019261, },
		{ -0.00014347, 0.00012609, 0.00007918, -0.00006229, -0.00009496, -0.00001793, 0.00002611, -0.00004762, -0.00015049, -0.00011561, 0.00009811, 0.00031630, 0.00029185, -0.00003797, -0.00043477, -0.00053054, -0.00016770, 0.00039712, 0.00064564, 0.00022482, -0.00069702, -0.00147046, -0.00141864, -0.00037094, 0.00114968, 0.00225547, 0.00230355, 0.00132797, -0.00002278, -0.00097158, -0.00113405, -0.00069258, },
		{ -0.00009959, -0.00016203, -0.00006745, 0.00002410, -0.00000128, -0.00008084, -0.00006027, 0.00010740, 0.00027209, 0.00022038, -0.00008632, -0.00041548, -0.00044077, -0.00006175, 0.00043180, 0.00054513, 0.00000318, -0.00093691, -0.00158648, -0.00131446, -0.00006226, 0.00153401, 0.00254247, 0.00237294, 0.00117579, -0.00029184, -0.00121337, -0.00124779, -0.00066608, -0.00003994, 0.00022069, 0.00010922, },
		{ 0.00004553, 0.00006273, -0.00001104, -0.00002325, 0.00009027, 0.00020922, 0.00014975, -0.00011495, -0.00037191, -0.00034224, 0.00002526, 0.00042664, 0.00040930, -0.00021883, -0.00112699, -0.00161630, -0.00112729, 0.00028089, 0.00187931, 0.00272800, 0.00232085, 0.00093333, -0.00058678, -0.00141684, -0.00128858, -0.00057174, 0.00009800, 0.00031135, 0.00011272, -0.00016245, -0.00023661, -0.00010375, },
		{ -0.00001025, 0.00004886, 0.00013615, 0.00008899, -0.00012071, -0.00030963, -0.00024604, 0.00008485, 0.00038393, 0.00025403, -0.00041748, -0.00124789, -0.00155707, -0.00087577, 0.00062448, 0.00215077, 0.00279217, 0.00215077, 0.00062448, -0.00087577, -0.00155707, -0.00124789, -0.00041748, 0.00025403, 0.00038393, 0.00008485, -0.00024604, -0.00030963, -0.00012071, 0.00008899, 0.00013615, 0.00004886, },
		{ 0.00004553, -0.00010375, -0.00023661, -0.00016245, 0.00011272, 0.00031135, 0.00009800, -0.00057174, -0.00128858, -0.00141684, -0.00058678, 0.00093333, 0.00232085, 0.00272800, 0.00187931, 0.00028089, -0.00112729, -0.00161630, -0.00112699, -0.00021883, 0.00040930, 0.00042664, 0.00002526, -0.00034224, -0.00037191, -0.00011495, 0.00014975, 0.00020922, 0.00009027, -0.00002325, -0.00001104, 0.00006273, },
		{ -0.00009959, 0.00010922, 0.00022069, -0.00003994, -0.00066608, -0.00124779, -0.00121337, -0.00029184, 0.00117579, 0.00237294, 0.00254247, 0.00153401, -0.00006226, -0.00131446, -0.00158648, -0.00093691, 0.00000318, 0.00054513, 0.00043180, -0.00006175, -0.00044077, -0.00041548, -0.00008632, 0.00022038, 0.00027209, 0.00010740, -0.00006027, -0.00008084, -0.00000128, 0.00002410, -0.00006745, -0.00016203, },
		{ -0.00014347, -0.00069258, -0.00113405, -0.00097158, -0.00002278, 0.00132797, 0.00230355, 0.00225547, 0.00114968, -0.00037094, -0.00141864, -0.00147046, -0.00069702, 0.00022482, 0.00064564, 0.00039712, -0.00016770, -0.00053054, -0.00043477, -0.00003797, 0.00029185, 0.00031630, 0.00009811, -0.00011561, -0.00015049, -0.00004762, 0.00002611, -0.00001793, -0.00009496, -0.00006229, 0.00007918, 0.00012609, },
		{ -0.00071997, 0.00019261, 0.00137688, 0.00212264, 0.00189715, 0.00076381, -0.00061675, -0.00143196, -0.00128163, -0.00043250, 0.00042292, 0.00069999, 0.00032602, -0.00028098, -0.00060120, -0.00042738, 0.00002406, 0.00035481, 0.00033600, 0.00006418, -0.00018078, -0.00020993, -0.00007073, 0.00004948, 0.00003700, -0.00004309, -0.00005162, 0.00003093, 0.00004179, -0.00020107, -0.00065185, -0.00096422, },
		{ 0.00185216, 0.00150385, 0.00041161, -0.00078058, -0.00135817, -0.00104202, -0.00017091, 0.00057799, 0.00070388, 0.00022728, -0.00038850, -0.00064463, -0.00039446, 0.00009174, 0.00040096, 0.00032883, 0.00001107, -0.00024575, -0.00025065, -0.00006855, 0.00008907, 0.00008894, -0.00001165, -0.00006483, -0.00002503, -0.00001993, -0.00020768, -0.00055277, -0.00076102, -0.00048661, 0.00033431, 0.00132204, },
		{ -0.00085473, -0.00121186, -0.00077926, 0.00006152, 0.00067673, 0.00066002, 0.00011370, -0.00047752, -0.00065599, -0.00034049, 0.00015620, 0.00042426, 0.00029623, -0.00005293, -0.00030051, -0.00026693, -0.00004277, 0.00013746, 0.00013025, -0.00000269, -0.00009600, -0.00007782, -0.00005070, -0.00016526, -0.00041104, -0.00054966, -0.00029526, 0.00039262, 0.00117523, 0.00152292, 0.00111322, 0.00012164, },
		{ 0.00024335, 0.00071373, 0.00057762, 0.00000027, -0.00053742, -0.00063448, -0.00027266, 0.00020903, 0.00042183, 0.00024323, -0.00011800, -0.00033659, -0.00025669, 0.00000146, 0.00018621, 0.00015516, -0.00001490, -0.00013703, -0.00011791, -0.00004720, -0.00008240, -0.00024677, -0.00035420, -0.00016219, 0.00036917, 0.00095862, 0.00117027, 0.00075942, -0.00008767, -0.00084350, -0.00101608, -0.00052265, },
		{ -0.00009801, -0.00056121, -0.00058335, -0.00019978, 0.00024349, 0.00039434, 0.00017766, -0.00017424, -0.00034832, -0.00022184, 0.00005641, 0.00022710, 0.00016067, -0.00004392, -0.00017905, -0.00013858, -0.00001169, 0.00002705, -0.00008144, -0.00019424, -0.00009424, 0.00027629, 0.00070137, 0.00082930, 0.00046881, -0.00020977, -0.00076189, -0.00079885, -0.00029913, 0.00036064, 0.00069188, 0.00047088, },
		{ -0.00013104, 0.00025552, 0.00034592, 0.00010901, -0.00021324, -0.00033378, -0.00016790, 0.00011291, 0.00025348, 0.00014693, -0.00008305, -0.00021385, -0.00013704, 0.00004844, 0.00014616, 0.00006517, -0.00008229, -0.00008822, 0.00013477, 0.00043536, 0.00053024, 0.00025704, -0.00025004, -0.00063281, -0.00058894, -0.00012969, 0.00040859, 0.00062158, 0.00035674, -0.00016834, -0.00054654, -0.00050947, },
		{ 0.00004693, -0.00022929, -0.00029501, -0.00010323, 0.00016187, 0.00026128, 0.00011721, -0.00012444, -0.00023520, -0.00011478, 0.00012187, 0.00025750, 0.00017784, -0.00002237, -0.00013201, -0.00002953, 0.00019058, 0.00029465, 0.00012777, -0.00022403, -0.00048321, -0.00041165, -0.00002660, 0.00039173, 0.00051897, 0.00025231, -0.00020201, -0.00049600, -0.00042226, -0.00007464, 0.00024429, 0.00028351, },
		{ -0.00003770, 0.00019559, 0.00024961, 0.00007735, -0.00016026, -0.00023985, -0.00007734, 0.00019515, 0.00034577, 0.00024794, -0.00000983, -0.00020681, -0.00018980, -0.00000898, 0.00013307, 0.00007319, -0.00015430, -0.00033968, -0.00028520, 0.00000787, 0.00032278, 0.00040316, 0.00017227, -0.00019561, -0.00041666, -0.00033238, -0.00003664, 0.00021229, 0.00021584, -0.00000014, -0.00022027, -0.00023776, },
		{ 0.00003483, -0.00018403, -0.00022799, -0.00003334, 0.00025482, 0.00040006, 0.00027441, -0.00003264, -0.00029056, -0.00032244, -0.00014852, 0.00004441, 0.00007627, -0.00006623, -0.00022425, -0.00021834, -0.00001866, 0.00022036, 0.00029327, 0.00012667, -0.00015149, -0.00031912, -0.00025017, -0.00002017, 0.00016491, 0.00015210, -0.00002624, -0.00018798, -0.00017060, 0.00001882, 0.00020900, 0.00022091, },
		{ 0.00000689, 0.00028957, 0.00041541, 0.00026354, -0.00007374, -0.00036162, -0.00041061, -0.00022407, 0.00001708, 0.00011433, 0.00001639, -0.00015105, -0.00020938, -0.00009040, 0.00010578, 0.00020541, 0.00011946, -0.00007737, -0.00021584, -0.00018421, -0.00002500, 0.00010962, 0.00010063, -0.00002877, -0.00013780, -0.00010365, 0.00005791, 0.00020042, 0.00018054, -0.00000246, -0.00019162, -0.00020323, },
		{ 0.00022765, -0.00011417, -0.00040247, -0.00044666, -0.00024240, 0.00003173, 0.00016337, 0.00007539, -0.00012428, -0.00024692, -0.00018618, -0.00000037, 0.00015026, 0.00014807, 0.00001488, -0.00011942, -0.00014020, -0.00004773, 0.00005484, 0.00006767, -0.00000879, -0.00007791, -0.00004697, 0.00007390, 0.00017178, 0.00013594, -0.00002742, -0.00018188, -0.00017194, 0.00003307, 0.00029205, 0.00039345, },
		{ -0.00043308, -0.00021933, 0.00006508, 0.00020237, 0.00010120, -0.00013783, -0.00031206, -0.00028311, -0.00008149, 0.00013159, 0.00020387, 0.00011108, -0.00004076, -0.00012019, -0.00008236, 0.00000859, 0.00005646, 0.00002916, -0.00001798, -0.00000914, 0.00006475, 0.00012834, 0.00009542, -0.00003494, -0.00015683, -0.00014223, 0.00003668, 0.00026005, 0.00034200, 0.00018277, -0.00013641, -0.00040263, },
		},
		{ { -0.00015699, -0.00001715, 0.00023104, 0.00027213, 0.00001047, -0.00032110, -0.00042345, -0.00024978, -0.00001888, 0.00006259, 0.00001948, 0.00002035, 0.00012440, 0.00019722, 0.00009740, -0.00011532, -0.00022604, -0.00011532, 0.00009740, 0.00019722, 0.00012440, 0.00002035, 0.00001948, 0.00006259, -0.00001888, -0.00024978, -0.00042345, -0.00032110, 0.00001047, 0.00027213, 0.00023104, -0.00001715, },
		{ 0.00031144, 0.00015342, -0.00023271, -0.00049171, -0.00039454, -0.00007323, 0.00015012, 0.00012495, 0.00000162, 0.00000174, 0.00013216, 0.00019216, 0.00004819, -0.00017736, -0.00024011, -0.00006326, 0.00017231, 0.00024055, 0.00011856, -0.00002550, -0.00006595, -0.00006411, -0.00014474, -0.00028280, -0.00030036, -0.00009828, 0.00017432, 0.00025636, 0.00007954, -0.00013046, -0.00010534, 0.00014206, },
		{ -0.00045480, -0.00051307, -0.00020252, 0.00015613, 0.00023907, 0.00006364, -0.00007656, 0.00001210, 0.00019507, 0.00019516, -0.00004710, -0.00028233, -0.00024038, 0.00005123, 0.00029643, 0.00026314, 0.00002136, -0.00017618, -0.00019573, -0.00013386, -0.00013310, -0.00017267, -0.00011451, 0.00006900, 0.00020965, 0.00014077, -0.00006601, -0.00015016, 0.00002522, 0.00027321, 0.00026808, -0.00007193, },
		{ 0.00007252, 0.00030877, 0.00017708, -0.00007967, -0.00011219, 0.00011120, 0.00027687, 0.00012256, -0.00022396, -0.00037644, -0.00013967, 0.00025424, 0.00039915, 0.00015820, -0.00021077, -0.00036552, -0.00023965, -0.00004754, 0.00001766, -0.00000755, 0.00002264, 0.00011875, 0.00013492, -0.00000231, -0.00014231, -0.00007507, 0.00017349, 0.00031079, 0.00010185, -0.00031982, -0.00055154, -0.00035450, },
		{ -0.00000646, -0.00019201, -0.00002626, 0.00026666, 0.00027669, -0.00007894, -0.00042441, -0.00034863, 0.00010634, 0.00046793, 0.00034851, -0.00013520, -0.00050548, -0.00044253, -0.00008375, 0.00019579, 0.00021007, 0.00009281, 0.00004242, 0.00006085, 0.00001613, -0.00010299, -0.00012704, 0.00005297, 0.00027110, 0.00022968, -0.00012889, -0.00048669, -0.00046888, -0.00007443, 0.00029892, 0.00029376, },
		{ 0.00016804, 0.00036538, 0.00011104, -0.00036313, -0.00051421, -0.00012028, 0.00042723, 0.00053160, 0.00004848, -0.00055052, -0.00068465, -0.00026229, 0.00027952, 0.00048117, 0.00030423, 0.00005185, -0.00004495, -0.00003917, -0.00007547, -0.00012685, -0.00004273, 0.00017284, 0.00027470, 0.00005776, -0.00033312, -0.00049913, -0.00023422, 0.00020476, 0.00036465, 0.00011537, -0.00019719, -0.00016803, },
		{ -0.00020457, -0.00058466, -0.00037076, 0.00026669, 0.00064197, 0.00029996, -0.00046112, -0.00088704, -0.00055662, 0.00020672, 0.00071797, 0.00063399, 0.00020476, -0.00011565, -0.00016313, -0.00010600, -0.00010199, -0.00008420, 0.00006197, 0.00023460, 0.00018450, -0.00013711, -0.00042921, -0.00035076, 0.00005328, 0.00035777, 0.00024011, -0.00012573, -0.00026582, 0.00001268, 0.00035969, 0.00028977, },
		{ 0.00001333, 0.00063097, 0.00055184, -0.00023824, -0.00097450, -0.00089813, -0.00004214, 0.00082780, 0.00101183, 0.00052106, -0.00007454, -0.00031126, -0.00022146, -0.00010107, -0.00007261, -0.00001404, 0.00014077, 0.00022068, 0.00004084, -0.00027919, -0.00038289, -0.00010669, 0.00026984, 0.00032104, -0.00000355, -0.00028849, -0.00015074, 0.00026051, 0.00040437, 0.00000956, -0.00053458, -0.00057870, },
		{ 0.00007463, -0.00090029, -0.00119528, -0.00043398, 0.00074250, 0.00133655, 0.00096866, 0.00013773, -0.00040875, -0.00041209, -0.00017347, -0.00004173, -0.00002737, 0.00004456, 0.00017144, 0.00014824, -0.00009801, -0.00031951, -0.00022228, 0.00012743, 0.00032746, 0.00012558, -0.00023204, -0.00027308, 0.00009765, 0.00042198, 0.00022075, -0.00037336, -0.00068655, -0.00027521, 0.00048380, 0.00073032, },
		{ -0.00088941, 0.00044215, 0.00150419, 0.00146536, 0.00053826, -0.00037263, -0.00062672, -0.00034633, -0.00004611, 0.00001650, -0.00000415, 0.00007515, 0.00015990, 0.00005415, -0.00018410, -0.00025592, -0.00002247, 0.00025590, 0.00021612, -0.00012000, -0.00032093, -0.00008055, 0.00033970, 0.00037195, -0.00014259, -0.00066365, -0.00052742, 0.00022590, 0.00077860, 0.00040339, -0.00066235, -0.00136087, },
		{ 0.00189741, 0.00108822, -0.00014030, -0.00078328, -0.00060736, -0.00013963, 0.00007795, 0.00002179, -0.00001154, 0.00008739, 0.00012763, -0.00002688, -0.00019945, -0.00012949, 0.00013146, 0.00023722, 0.00000566, -0.00028721, -0.00022339, 0.00018482, 0.00042519, 0.00009671, -0.00051601, -0.00067964, -0.00008349, 0.00067551, 0.00066538, -0.00030575, -0.00133878, -0.00130385, -0.00003507, 0.00143935, },
		{ -0.00079230, -0.00090078, -0.00035171, 0.00009620, 0.00011364, -0.00003589, -0.00002481, 0.00010171, 0.00009274, -0.00007795, -0.00015824, -0.00000101, 0.00018392, 0.00009991, -0.00019250, -0.00029442, 0.00000521, 0.00037293, 0.00028292, -0.00028448, -0.00069547, -0.00036876, 0.00044293, 0.00079462, 0.00008983, -0.00112175, -0.00157715, -0.00060025, 0.00111992, 0.00214977, 0.00169604, 0.00030513, },
		{ 0.00001185, 0.00022678, 0.00002961, -0.00011523, -0.00000703, 0.00012744, 0.00005825, -0.00010180, -0.00009307, 0.00007900, 0.00013082, -0.00007686, -0.00028280, -0.00014677, 0.00023936, 0.00037269, -0.00003204, -0.00057686, -0.00056071, 0.00013972, 0.00076200, 0.00043349, -0.00075416, -0.00164241, -0.00113674, 0.00058703, 0.00213921, 0.00223769, 0.00092149, -0.00058648, -0.00113872, -0.00067014, },
		{ -0.00012801, -0.00014936, 0.00005959, 0.00015077, 0.00001471, -0.00011036, -0.00003501, 0.00009105, 0.00001479, -0.00020575, -0.00023132, 0.00007154, 0.00035354, 0.00017573, -0.00036303, -0.00061746, -0.00015621, 0.00058393, 0.00065004, -0.00031920, -0.00148517, -0.00153189, -0.00006334, 0.00183987, 0.00258862, 0.00161158, -0.00014643, -0.00122580, -0.00103716, -0.00020876, 0.00029462, 0.00017434, },
		{ 0.00015382, 0.00014176, -0.00004575, -0.00011115, 0.00000052, 0.00004891, -0.00010143, -0.00023363, -0.00007665, 0.00024639, 0.00029063, -0.00011781, -0.00053663, -0.00037168, 0.00031662, 0.00070176, 0.00008576, -0.00114675, -0.00170772, -0.00070490, 0.00129325, 0.00265815, 0.00224467, 0.00048608, -0.00108920, -0.00136148, -0.00055611, 0.00025175, 0.00035020, -0.00003424, -0.00025983, -0.00009103, },
		{ -0.00011511, -0.00010068, 0.00001525, -0.00001483, -0.00016755, -0.00016220, 0.00009736, 0.00029583, 0.00009055, -0.00035504, -0.00046034, 0.00003794, 0.00059818, 0.00037750, -0.00071117, -0.00164144, -0.00121450, 0.00059907, 0.00241566, 0.00268958, 0.00121272, -0.00070479, -0.00154323, -0.00097354, 0.00005883, 0.00048580, 0.00015530, -0.00027948, -0.00027222, 0.00005125, 0.00022206, 0.00007683, },
		{ 0.00001852, -0.00006950, -0.00016660, -0.00003839, 0.00020945, 0.00021045, -0.00013604, -0.00041586, -0.00017713, 0.00039088, 0.00051011, -0.00027965, -0.00136933, -0.00150429, -0.00011036, 0.00190057, 0.00284971, 0.00190057, -0.00011036, -0.00150429, -0.00136933, -0.00027965, 0.00051011, 0.00039088, -0.00017713, -0.00041586, -0.00013604, 0.00021045, 0.00020945, -0.00003839, -0.00016660, -0.00006950, },
		{ -0.00011511, 0.00007683, 0.00022206, 0.00005125, -0.00027222, -0.00027948, 0.00015530, 0.00048580, 0.00005883, -0.00097354, -0.00154323, -0.00070479, 0.00121272, 0.00268958, 0.00241566, 0.00059907, -0.00121450, -0.00164144, -0.00071117, 0.00037750, 0.00059818, 0.00003794, -0.00046034, -0.00035504, 0.00009055, 0.00029583, 0.00009736, -0.00016220, -0.00016755, -0.00001483, 0.00001525, -0.00010068, },
		{ 0.00015382, -0.00009103, -0.00025983, -0.00003424, 0.00035020, 0.00025175, -0.00055611, -0.00136148, -0.00108920, 0.00048608, 0.00224467, 0.00265815, 0.00129325, -0.00070490, -0.00170772, -0.00114675, 0.00008576, 0.00070176, 0.00031662, -0.00037168, -0.00053663, -0.00011781, 0.00029063, 0.00024639, -0.00007665, -0.00023363, -0.00010143, 0.00004891, 0.00000052, -0.00011115, -0.00004575, 0.00014176, },
		{ -0.00012801, 0.00017434, 0.00029462, -0.00020876, -0.00103716, -0.00122580, -0.00014643, 0.00161158, 0.00258862, 0.00183987, -0.00006334, -0.00153189, -0.00148517, -0.00031920, 0.00065004, 0.00058393, -0.00015621, -0.00061746, -0.00036303, 0.00017573, 0.00035354, 0.00007154, -0.00023132, -0.00020575, 0.00001479, 0.00009105, -0.00003501, -0.00011036, 0.00001471, 0.00015077, 0.00005959, -0.00014936, },
		{ 0.00001185, -0.00067014, -0.00113872, -0.00058648, 0.00092149, 0.00223769, 0.00213921, 0.00058703, -0.00113674, -0.00164241, -0.00075416, 0.00043349, 0.00076200, 0.00013972, -0.00056071, -0.00057686, -0.00003204, 0.00037269, 0.00023936, -0.00014677, -0.00028280, -0.00007686, 0.00013082, 0.00007900, -0.00009307, -0.00010180, 0.00005825, 0.00012744, -0.00000703, -0.00011523, 0.00002961, 0.00022678, },
		{ -0.00079230, 0.00030513, 0.00169604, 0.00214977, 0.00111992, -0.00060025, -0.00157715, -0.00112175, 0.00008983, 0.00079462, 0.00044293, -0.00036876, -0.00069547, -0.00028448, 0.00028292, 0.00037293, 0.00000521, -0.00029442, -0.00019250, 0.00009991, 0.00018392, -0.00000101, -0.00015824, -0.00007795, 0.00009274, 0.00010171, -0.00002481, -0.00003589, 0.00011364, 0.00009620, -0.00035171, -0.00090078, },
		{ 0.00189741, 0.00143935, -0.00003507, -0.00130385, -0.00133878, -0.00030575, 0.00066538, 0.00067551, -0.00008349, -0.00067964, -0.00051601, 0.00009671, 0.00042519, 0.00018482, -0.00022339, -0.00028721, 0.00000566, 0.00023722, 0.00013146, -0.00012949, -0.00019945, -0.00002688, 0.00012763, 0.00008739, -0.00001154, 0.00002179, 0.00007795, -0.00013963, -0.00060736, -0.00078328, -0.00014030, 0.00108822, },
		{ -0.00088941, -0.00136087, -0.00066235, 0.00040339, 0.00077860, 0.00022590, -0.00052742, -0.00066365, -0.00014259, 0.00037195, 0.00033970, -0.00008055, -0.00032093, -0.00012000, 0.00021612, 0.00025590, -0.00002247, -0.00025592, -0.00018410, 0.00005415, 0.00015990, 0.00007515, -0.00000415, 0.00001650, -0.00004611, -0.00034633, -0.00062672, -0.00037263, 0.00053826, 0.00146536, 0.00150419, 0.00044215, },
		{ 0.00007463, 0.00073032, 0.00048380, -0.00027521, -0.00068655, -0.00037336, 0.00022075, 0.00042198, 0.00009765, -0.00027308, -0.00023204, 0.00012558, 0.00032746, 0.00012743, -0.00022228, -0.00031951, -0.00009801, 0.00014824, 0.00017144, 0.00004456, -0.00002737, -0.00004173, -0.00017347, -0.00041209, -0.00040875, 0.00013773, 0.00096866, 0.00133655, 0.00074250, -0.00043398, -0.00119528, -0.00090029, },
		{ 0.00001333, -0.00057870, -0.00053458, 0.00000956, 0.00040437, 0.00026051, -0.00015074, -0.00028849, -0.00000355, 0.00032104, 0.00026984, -0.00010669, -0.00038289, -0.00027919, 0.00004084, 0.00022068, 0.00014077, -0.00001404, -0.00007261, -0.00010107, -0.00022146, -0.00031126, -0.00007454, 0.00052106, 0.00101183, 0.00082780, -0.00004214, -0.00089813, -0.00097450, -0.00023824, 0.00055184, 0.00063097, },
		{ -0.00020457, 0.00028977, 0.00035969, 0.00001268, -0.00026582, -0.00012573, 0.00024011, 0.00035777, 0.00005328, -0.00035076, -0.00042921, -0.00013711, 0.00018450, 0.00023460, 0.00006197, -0.00008420, -0.00010199, -0.00010600, -0.00016313, -0.00011565, 0.00020476, 0.00063399, 0.00071797, 0.00020672, -0.00055662, -0.00088704, -0.00046112, 0.00029996, 0.00064197, 0.00026669, -0.00037076, -0.00058466, },
		{ 0.00016804, -0.00016803, -0.00019719, 0.00011537, 0.00036465, 0.00020476, -0.00023422, -0.00049913, -0.00033312, 0.00005776, 0.00027470, 0.00017284, -0.00004273, -0.00012685, -0.00007547, -0.00003917, -0.00004495, 0.00005185, 0.00030423, 0.00048117, 0.00027952, -0.00026229, -0.00068465, -0.00055052, 0.00004848, 0.00053160, 0.00042723, -0.00012028, -0.00051421, -0.00036313, 0.00011104, 0.00036538, },
		{ -0.00000646, 0.00029376, 0.00029892, -0.00007443, -0.00046888, -0.00048669, -0.00012889, 0.00022968, 0.00027110, 0.00005297, -0.00012704, -0.00010299, 0.00001613, 0.00006085, 0.00004242, 0.00009281, 0.00021007, 0.00019579, -0.00008375, -0.00044253, -0.00050548, -0.00013520, 0.00034851, 0.00046793, 0.00010634, -0.00034863, -0.00042441, -0.00007894, 0.00027669, 0.00026666, -0.00002626, -0.00019201, },
		{ 0.00007252, -0.00035450, -0.00055154, -0.00031982, 0.00010185, 0.00031079, 0.00017349, -0.00007507, -0.00014231, -0.00000231, 0.00013492, 0.00011875, 0.00002264, -0.00000755, 0.00001766, -0.00004754, -0.00023965, -0.00036552, -0.00021077, 0.00015820, 0.00039915, 0.00025424, -0.00013967, -0.00037644, -0.00022396, 0.00012256, 0.00027687, 0.00011120, -0.00011219, -0.00007967, 0.00017708, 0.00030877, },
		{ -0.00045480, -0.00007193, 0.00026808, 0.00027321, 0.00002522, -0.00015016, -0.00006601, 0.00014077, 0.00020965, 0.00006900, -0.00011451, -0.00017267, -0.00013310, -0.00013386, -0.00019573, -0.00017618, 0.00002136, 0.00026314, 0.00029643, 0.00005123, -0.00024038, -0.00028233, -0.00004710, 0.00019516, 0.00019507, 0.00001210, -0.00007656, 0.00006364, 0.00023907, 0.00015613, -0.00020252, -0.00051307, },
		{ 0.00031144, 0.00014206, -0.00010534, -0.00013046, 0.00007954, 0.00025636, 0.00017432, -0.00009828, -0.00030036, -0.00028280, -0.00014474, -0.00006411, -0.00006595, -0.00002550, 0.00011856, 0.00024055, 0.00017231, -0.00006326, -0.00024011, -0.00017736, 0.00004819, 0.00019216, 0.00013216, 0.00000174, 0.00000162, 0.00012495, 0.00015012, -0.00007323, -0.00039454, -0.00049171, -0.00023271, 0.00015342, },
		},
		{ { 0.00010196, -0.00001039, -0.00017325, -0.00012412, 0.00009709, 0.00020577, 0.00008017, -0.00008738, -0.00009081, 0.00001014, 0.00002449, -0.00007303, -0.00013485, -0.00008415, -0.00000693, 0.00001211, 0.00000414, 0.00001211, -0.00000693, -0.00008415, -0.00013485, -0.00007303, 0.00002449, 0.00001014, -0.00009081, -0.00008738, 0.00008017, 0.00020577, 0.00009709, -0.00012412, -0.00017325, -0.00001039, },
		{ -0.00017792, -0.00008529, 0.00014816, 0.00020885, 0.00002485, -0.00013366, -0.00006347, 0.00007385, 0.00001269, -0.00020120, -0.00026989, -0.00008062, 0.00013282, 0.00012349, -0.00004848, -0.00014622, -0.00008728, 0.00000703, 0.00003375, 0.00001616, -0.00000877, -0.00005535, -0.00009491, -0.00003531, 0.00011646, 0.00018472, 0.00004677, -0.00015120, -0.00016058, 0.00001252, 0.00010033, -0.00003325, },
		{ 0.00019320, 0.00018970, -0.00004414, -0.00016552, -0.00001544, 0.00012300, -0.00004509, -0.00034355, -0.00034307, 0.00001129, 0.00028760, 0.00014995, -0.00019532, -0.00030301, -0.00007429, 0.00018162, 0.00020119, 0.00004837, -0.00007858, -0.00010995, -0.00007710, 0.00001358, 0.00013172, 0.00015220, 0.00000313, -0.00016606, -0.00014195, 0.00003427, 0.00009550, -0.00005477, -0.00017319, -0.00003694, },
		{ -0.00011880, -0.00017610, 0.00004718, 0.00014669, -0.00014195, -0.00047246, -0.00033708, 0.00016880, 0.00041028, 0.00007064, -0.00040520, -0.00040755, 0.00004955, 0.00041230, 0.00033202, 0.00001532, -0.00017026, -0.00014316, -0.00004234, 0.00005137, 0.00012744, 0.00011578, -0.00002969, -0.00016992, -0.00011969, 0.00005385, 0.00008766, -0.00007368, -0.00015854, 0.00001719, 0.00022548, 0.00014702, },
		{ 0.00011561, 0.00013778, -0.00026414, -0.00056190, -0.00024849, 0.00035648, 0.00045901, -0.00011436, -0.00063179, -0.00041568, 0.00028193, 0.00065993, 0.00039579, -0.00008017, -0.00026541, -0.00014835, 0.00000168, 0.00007239, 0.00010828, 0.00008288, -0.00004979, -0.00016532, -0.00009613, 0.00007046, 0.00007710, -0.00008887, -0.00013442, 0.00007222, 0.00023941, 0.00008291, -0.00018961, -0.00016137, },
		{ -0.00039314, -0.00059167, -0.00008845, 0.00053211, 0.00040452, -0.00038358, -0.00082129, -0.00029759, 0.00060041, 0.00087964, 0.00037209, -0.00022451, -0.00034479, -0.00012381, 0.00004574, 0.00007410, 0.00008113, 0.00005956, -0.00005773, -0.00015566, -0.00007327, 0.00008361, 0.00006424, -0.00009947, -0.00010230, 0.00012281, 0.00023137, 0.00000265, -0.00024689, -0.00012089, 0.00017971, 0.00009404, },
		{ 0.00011897, 0.00065358, 0.00023539, -0.00069717, -0.00092130, -0.00004346, 0.00096534, 0.00102797, 0.00025411, -0.00039538, -0.00039152, -0.00007303, 0.00008062, 0.00005747, 0.00005400, 0.00004955, -0.00005627, -0.00014457, -0.00005251, 0.00009307, 0.00004956, -0.00010494, -0.00006450, 0.00016376, 0.00020034, -0.00008586, -0.00028228, -0.00005793, 0.00022948, 0.00001867, -0.00050841, -0.00055072, },
		{ -0.00003944, -0.00100294, -0.00088972, 0.00033400, 0.00132557, 0.00106985, 0.00005006, -0.00056516, -0.00039377, -0.00000423, 0.00009861, 0.00002674, 0.00003458, 0.00005368, -0.00004984, -0.00013526, -0.00003456, 0.00009888, 0.00003364, -0.00010509, -0.00002394, 0.00019082, 0.00014815, -0.00017322, -0.00028992, 0.00002088, 0.00025664, -0.00008007, -0.00059063, -0.00043905, 0.00034037, 0.00068596, },
		{ -0.00070222, 0.00080043, 0.00162608, 0.00098443, -0.00021775, -0.00070516, -0.00034679, 0.00007097, 0.00009477, -0.00001126, 0.00002897, 0.00006977, -0.00004367, -0.00012998, -0.00001939, 0.00010125, 0.00001708, -0.00010010, 0.00001620, 0.00020117, 0.00007929, -0.00024974, -0.00026744, 0.00010646, 0.00025589, -0.00019000, -0.00062465, -0.00026801, 0.00053820, 0.00060739, -0.00039139, -0.00124445, },
		{ 0.00181631, 0.00076878, -0.00051590, -0.00079013, -0.00025375, 0.00013933, 0.00006774, -0.00004822, 0.00004061, 0.00009304, -0.00004284, -0.00012965, -0.00000637, 0.00010049, 0.00000052, -0.00009051, 0.00005282, 0.00019383, 0.00000036, -0.00030679, -0.00021633, 0.00018854, 0.00022577, -0.00029674, -0.00060195, -0.00005873, 0.00067663, 0.00041308, -0.00077608, -0.00137002, -0.00035703, 0.00130422, },
		{ -0.00080471, -0.00080223, -0.00012546, 0.00018807, 0.00002006, -0.00007576, 0.00006966, 0.00011692, -0.00005133, -0.00013372, 0.00000555, 0.00009696, -0.00001540, -0.00007708, 0.00008329, 0.00016982, -0.00008075, -0.00033803, -0.00014177, 0.00025714, 0.00016892, -0.00038584, -0.00052200, 0.00016086, 0.00072724, 0.00011670, -0.00113974, -0.00134112, 0.00012361, 0.00178361, 0.00185810, 0.00043899, },
		{ 0.00002116, 0.00020692, -0.00004218, -0.00008677, 0.00011293, 0.00013417, -0.00007132, -0.00014026, 0.00001754, 0.00009099, -0.00003007, -0.00006077, 0.00010575, 0.00013195, -0.00015595, -0.00034023, -0.00005197, 0.00030397, 0.00009179, -0.00044490, -0.00039256, 0.00036017, 0.00067359, -0.00025110, -0.00142720, -0.00113905, 0.00069568, 0.00217548, 0.00173184, 0.00002836, -0.00104404, -0.00073396, },
		{ -0.00011022, -0.00007661, 0.00016427, 0.00013810, -0.00010267, -0.00014631, 0.00003060, 0.00008283, -0.00004293, -0.00004260, 0.00011922, 0.00008446, -0.00021798, -0.00031377, 0.00004296, 0.00032362, 0.00000371, -0.00046538, -0.00022864, 0.00051048, 0.00051407, -0.00064684, -0.00159028, -0.00076859, 0.00129895, 0.00242463, 0.00143993, -0.00041707, -0.00119929, -0.00058948, 0.00016559, 0.00018981, },
		{ 0.00021553, 0.00012377, -0.00014284, -0.00014842, 0.00004533, 0.00007267, -0.00005349, -0.00002354, 0.00012372, 0.00003242, -0.00026138, -0.00026248, 0.00013242, 0.00031431, -0.00008440, -0.00044394, -0.00005036, 0.00058923, 0.00026254, -0.00102026, -0.00159532, -0.00025834, 0.00186512, 0.00249200, 0.00100710, -0.00084483, -0.00124653, -0.00038428, 0.00028670, 0.00013588, -0.00017391, -0.00004395, },
		{ -0.00018720, -0.00014326, 0.00006168, 0.00006060, -0.00006133, -0.00000447, 0.00012016, -0.00001893, -0.00028316, -0.00019314, 0.00020680, 0.00027813, -0.00016164, -0.00038293, 0.00011985, 0.00058350, -0.00005338, -0.00132169, -0.00142861, 0.00034258, 0.00232732, 0.00236063, 0.00047739, -0.00120346, -0.00117622, -0.00014301, 0.00036601, 0.00004975, -0.00022330, 0.00000893, 0.00025780, 0.00008890, },
		{ 0.00007893, 0.00004670, -0.00006616, 0.00001391, 0.00011016, -0.00006492, -0.00028311, -0.00011442, 0.00025875, 0.00022067, -0.00021868, -0.00029004, 0.00026055, 0.00049194, -0.00039614, -0.00150926, -0.00109900, 0.00097094, 0.00262954, 0.00203853, -0.00009172, -0.00144981, -0.00099473, 0.00010401, 0.00039043, -0.00005902, -0.00025012, 0.00007640, 0.00028280, 0.00003438, -0.00022970, -0.00012823, },
		{ -0.00006779, 0.00003108, 0.00009571, -0.00010202, -0.00026372, -0.00003568, 0.00028421, 0.00015020, -0.00024906, -0.00017724, 0.00035435, 0.00032512, -0.00072352, -0.00155516, -0.00063717, 0.00155791, 0.00273462, 0.00155791, -0.00063717, -0.00155516, -0.00072352, 0.00032512, 0.00035435, -0.00017724, -0.00024906, 0.00015020, 0.00028421, -0.00003568, -0.00026372, -0.00010202, 0.00009571, 0.00003108, },
		{ 0.00007893, -0.00012823, -0.00022970, 0.00003438, 0.00028280, 0.00007640, -0.00025012, -0.00005902, 0.00039043, 0.00010401, -0.00099473, -0.00144981, -0.00009172, 0.00203853, 0.00262954, 0.00097094, -0.00109900, -0.00150926, -0.00039614, 0.00049194, 0.00026055, -0.00029004, -0.00021868, 0.00022067, 0.00025875, -0.00011442, -0.00028311, -0.00006492, 0.00011016, 0.00001391, -0.00006616, 0.00004670, },
		{ -0.00018720, 0.00008890, 0.00025780, 0.00000893, -0.00022330, 0.00004975, 0.00036601, -0.00014301, -0.00117622, -0.00120346, 0.00047739, 0.00236063, 0.00232732, 0.00034258, -0.00142861, -0.00132169, -0.00005338, 0.00058350, 0.00011985, -0.00038293, -0.00016164, 0.00027813, 0.00020680, -0.00019314, -0.00028316, -0.00001893, 0.00012016, -0.00000447, -0.00006133, 0.00006060, 0.00006168, -0.00014326, },
		{ 0.00021553, -0.00004395, -0.00017391, 0.00013588, 0.00028670, -0.00038428, -0.00124653, -0.00084483, 0.00100710, 0.00249200, 0.00186512, -0.00025834, -0.00159532, -0.00102026, 0.00026254, 0.00058923, -0.00005036, -0.00044394, -0.00008440, 0.00031431, 0.00013242, -0.00026248, -0.00026138, 0.00003242, 0.00012372, -0.00002354, -0.00005349, 0.00007267, 0.00004533, -0.00014842, -0.00014284, 0.00012377, },
		{ -0.00011022, 0.00018981, 0.00016559, -0.00058948, -0.00119929, -0.00041707, 0.00143993, 0.00242463, 0.00129895, -0.00076859, -0.00159028, -0.00064684, 0.00051407, 0.00051048, -0.00022864, -0.00046538, 0.00000371, 0.00032362, 0.00004296, -0.00031377, -0.00021798, 0.00008446, 0.00011922, -0.00004260, -0.00004293, 0.00008283, 0.00003060, -0.00014631, -0.00010267, 0.00013810, 0.00016427, -0.00007661, },
		{ 0.00002116, -0.00073396, -0.00104404, 0.00002836, 0.00173184, 0.00217548, 0.00069568, -0.00113905, -0.00142720, -0.00025110, 0.00067359, 0.00036017, -0.00039256, -0.00044490, 0.00009179, 0.00030397, -0.00005197, -0.00034023, -0.00015595, 0.00013195, 0.00010575, -0.00006077, -0.00003007, 0.00009099, 0.00001754, -0.00014026, -0.00007132, 0.00013417, 0.00011293, -0.00008677, -0.00004218, 0.00020692, },
		{ -0.00080471, 0.00043899, 0.00185810, 0.00178361, 0.00012361, -0.00134112, -0.00113974, 0.00011670, 0.00072724, 0.00016086, -0.00052200, -0.00038584, 0.00016892, 0.00025714, -0.00014177, -0.00033803, -0.00008075, 0.00016982, 0.00008329, -0.00007708, -0.00001540, 0.00009696, 0.00000555, -0.00013372, -0.00005133, 0.00011692, 0.00006966, -0.00007576, 0.00002006, 0.00018807, -0.00012546, -0.00080223, },
		{ 0.00181631, 0.00130422, -0.00035703, -0.00137002, -0.00077608, 0.00041308, 0.00067663, -0.00005873, -0.00060195, -0.00029674, 0.00022577, 0.00018854, -0.00021633, -0.00030679, 0.00000036, 0.00019383, 0.00005282, -0.00009051, 0.00000052, 0.00010049, -0.00000637, -0.00012965, -0.00004284, 0.00009304, 0.00004061, -0.00004822, 0.00006774, 0.00013933, -0.00025375, -0.00079013, -0.00051590, 0.00076878, },
		{ -0.00070222, -0.00124445, -0.00039139, 0.00060739, 0.00053820, -0.00026801, -0.00062465, -0.00019000, 0.00025589, 0.00010646, -0.00026744, -0.00024974, 0.00007929, 0.00020117, 0.00001620, -0.00010010, 0.00001708, 0.00010125, -0.00001939, -0.00012998, -0.00004367, 0.00006977, 0.00002897, -0.00001126, 0.00009477, 0.00007097, -0.00034679, -0.00070516, -0.00021775, 0.00098443, 0.00162608, 0.00080043, },
		{ -0.00003944, 0.00068596, 0.00034037, -0.00043905, -0.00059063, -0.00008007, 0.00025664, 0.00002088, -0.00028992, -0.00017322, 0.00014815, 0.00019082, -0.00002394, -0.00010509, 0.00003364, 0.00009888, -0.00003456, -0.00013526, -0.00004984, 0.00005368, 0.00003458, 0.00002674, 0.00009861, -0.00000423, -0.00039377, -0.00056516, 0.00005006, 0.00106985, 0.00132557, 0.00033400, -0.00088972, -0.00100294, },
		{ 0.00011897, -0.00055072, -0.00050841, 0.00001867, 0.00022948, -0.00005793, -0.00028228, -0.00008586, 0.00020034, 0.00016376, -0.00006450, -0.00010494, 0.00004956, 0.00009307, -0.00005251, -0.00014457, -0.00005627, 0.00004955, 0.00005400, 0.00005747, 0.00008062, -0.00007303, -0.00039152, -0.00039538, 0.00025411, 0.00102797, 0.00096534, -0.00004346, -0.00092130, -0.00069717, 0.00023539, 0.00065358, },
		{ -0.00039314, 0.00009404, 0.00017971, -0.00012089, -0.00024689, 0.00000265, 0.00023137, 0.00012281, -0.00010230, -0.00009947, 0.00006424, 0.00008361, -0.00007327, -0.00015566, -0.00005773, 0.00005956, 0.00008113, 0.00007410, 0.00004574, -0.00012381, -0.00034479, -0.00022451, 0.00037209, 0.00087964, 0.00060041, -0.00029759, -0.00082129, -0.00038358, 0.00040452, 0.00053211, -0.00008845, -0.00059167, },
		{ 0.00011561, -0.00016137, -0.00018961, 0.00008291, 0.00023941, 0.00007222, -0.00013442, -0.00008887, 0.00007710, 0.00007046, -0.00009613, -0.00016532, -0.00004979, 0.00008288, 0.00010828, 0.00007239, 0.00000168, -0.00014835, -0.00026541, -0.00008017, 0.00039579, 0.00065993, 0.00028193, -0.00041568, -0.00063179, -0.00011436, 0.00045901, 0.00035648, -0.00024849, -0.00056190, -0.00026414, 0.00013778, },
		{ -0.00011880, 0.00014702, 0.00022548, 0.00001719, -0.00015854, -0.00007368, 0.00008766, 0.00005385, -0.00011969, -0.00016992, -0.00002969, 0.00011578, 0.00012744, 0.00005137, -0.00004234, -0.00014316, -0.00017026, 0.00001532, 0.00033202, 0.00041230, 0.00004955, -0.00040755, -0.00040520, 0.00007064, 0.00041028, 0.00016880, -0.00033708, -0.00047246, -0.00014195, 0.00014669, 0.00004718, -0.00017610, },
		{ 0.00019320, -0.00003694, -0.00017319, -0.00005477, 0.00009550, 0.00003427, -0.00014195, -0.00016606, 0.00000313, 0.00015220, 0.00013172, 0.00001358, -0.00007710, -0.00010995, -0.00007858, 0.00004837, 0.00020119, 0.00018162, -0.00007429, -0.00030301, -0.00019532, 0.00014995, 0.00028760, 0.00001129, -0.00034307, -0.00034355, -0.00004509, 0.00012300, -0.00001544, -0.00016552, -0.00004414, 0.00018970, },
		{ -0.00017792, -0.00003325, 0.00010033, 0.00001252, -0.00016058, -0.00015120, 0.00004677, 0.00018472, 0.00011646, -0.00003531, -0.00009491, -0.00005535, -0.00000877, 0.00001616, 0.00003375, 0.00000703, -0.00008728, -0.00014622, -0.00004848, 0.00012349, 0.00013282, -0.00008062, -0.00026989, -0.00020120, 0.00001269, 0.00007385, -0.00006347, -0.00013366, 0.00002485, 0.00020885, 0.00014816, -0.00008529, },
		},
		{ { -0.00006987, -0.00005409, 0.00000984, 0.00008209, 0.00005663, -0.00003991, -0.00006427, -0.00003416, -0.00008002, -0.00010005, 0.00009925, 0.00029101, 0.00004227, -0.00044694, -0.00042245, 0.00023731, 0.00065790, 0.00023731, -0.00042245, -0.00044694, 0.00004227, 0.00029101, 0.00009925, -0.00010005, -0.00008002, -0.00003416, -0.00006427, -0.00003991, 0.00005663, 0.00008209, 0.00000984, -0.00005409, },
		{ 0.00001353, 0.00010129, 0.00005483, -0.00006283, -0.00007055, -0.00002040, -0.00008092, -0.00011703, 0.00010080, 0.00030837, 0.00003317, -0.00048425, -0.00043975, 0.00025638, 0.00067655, 0.00023732, -0.00041659, -0.00042790, 0.00004789, 0.00027907, 0.00009811, -0.00008668, -0.00007927, -0.00004583, -0.00005873, -0.00001926, 0.00005832, 0.00006403, 0.00000630, -0.00004015, -0.00006986, -0.00006828, },
		{ 0.00005297, -0.00008713, -0.00007735, -0.00000510, -0.00008192, -0.00013697, 0.00010272, 0.00033048, 0.00002095, -0.00053839, -0.00046783, 0.00029380, 0.00073177, 0.00025641, -0.00042240, -0.00042787, 0.00004982, 0.00027302, 0.00009744, -0.00007745, -0.00007868, -0.00005498, -0.00005417, -0.00000168, 0.00005982, 0.00004780, 0.00000307, -0.00002700, -0.00006982, -0.00008217, 0.00001726, 0.00012089, },
		{ -0.00008440, 0.00001118, -0.00008298, -0.00015912, 0.00010493, 0.00035651, 0.00000607, -0.00060729, -0.00050561, 0.00034812, 0.00082145, 0.00029385, -0.00043966, -0.00044683, 0.00004799, 0.00027308, 0.00009726, -0.00007270, -0.00007828, -0.00006125, -0.00005074, 0.00001217, 0.00006109, 0.00003402, 0.00000027, -0.00001513, -0.00006976, -0.00009523, 0.00002086, 0.00014014, 0.00005114, -0.00011189, },
		{ -0.00008407, -0.00018261, 0.00010734, 0.00038543, -0.00001089, -0.00068829, -0.00055164, 0.00041728, 0.00094213, 0.00034820, -0.00046770, -0.00048407, 0.00004246, 0.00027925, 0.00009758, -0.00007262, -0.00007808, -0.00006439, -0.00004859, 0.00002175, 0.00006206, 0.00002323, -0.00000200, -0.00000501, -0.00006968, -0.00010696, 0.00002420, 0.00015829, 0.00004940, -0.00013616, -0.00009143, 0.00002778, },
		{ 0.00010986, 0.00041615, -0.00002929, -0.00077829, -0.00060415, 0.00049859, 0.00108919, 0.00041737, -0.00050544, -0.00053816, 0.00003344, 0.00029129, 0.00009838, -0.00007721, -0.00007810, -0.00006430, -0.00004779, 0.00002669, 0.00006271, 0.00001584, -0.00000365, 0.00000298, -0.00006959, -0.00011691, 0.00002716, 0.00017465, 0.00004782, -0.00015899, -0.00009817, 0.00004409, -0.00008514, -0.00020655, },
		{ -0.00004841, -0.00087383, -0.00066112, 0.00058895, 0.00125696, 0.00049870, -0.00055144, -0.00060700, 0.00002130, 0.00030875, 0.00009964, -0.00008630, -0.00007833, -0.00006096, -0.00004838, 0.00002680, 0.00006301, 0.00001214, -0.00000462, 0.00000852, -0.00006948, -0.00012469, 0.00002962, 0.00018860, 0.00004647, -0.00017952, -0.00010436, 0.00005947, -0.00008615, -0.00023001, 0.00011240, 0.00044749, },
		{ -0.00072036, 0.00068488, 0.00143901, 0.00058907, -0.00060392, -0.00068797, 0.00000648, 0.00033095, 0.00010130, -0.00009954, -0.00007877, -0.00005452, -0.00005034, 0.00002207, 0.00006294, 0.00001225, -0.00000486, 0.00001141, -0.00006936, -0.00013000, 0.00003149, 0.00019958, 0.00004538, -0.00019695, -0.00010976, 0.00007333, -0.00008707, -0.00025210, 0.00011485, 0.00047823, -0.00006752, -0.00097123, },
		{ 0.00162833, 0.00068500, -0.00066088, -0.00077794, -0.00001044, 0.00035703, 0.00010331, -0.00011641, -0.00007939, -0.00004521, -0.00005358, 0.00001270, 0.00006251, 0.00001619, -0.00000437, 0.00001153, -0.00006923, -0.00013265, 0.00003268, 0.00020719, 0.00004461, -0.00021061, -0.00011417, 0.00008515, -0.00008785, -0.00027196, 0.00011712, 0.00050720, -0.00008589, -0.00106675, -0.00077960, 0.00078268, },
		{ -0.00072012, -0.00087346, -0.00002881, 0.00038600, 0.00010558, -0.00013626, -0.00008017, -0.00003340, -0.00005799, -0.00000096, 0.00006174, 0.00002381, -0.00000317, 0.00000889, -0.00006911, -0.00013253, 0.00003317, 0.00021114, 0.00004418, -0.00021999, -0.00011741, 0.00009445, -0.00008847, -0.00028883, 0.00011913, 0.00053329, -0.00010281, -0.00115673, -0.00083655, 0.00087861, 0.00181765, 0.00078281, },
		{ -0.00004792, 0.00041675, 0.00010803, -0.00015835, -0.00008109, -0.00001954, -0.00006339, -0.00001840, 0.00006065, 0.00003479, -0.00000131, 0.00000357, -0.00006899, -0.00012964, 0.00003293, 0.00021125, 0.00004411, -0.00022471, -0.00011937, 0.00010090, -0.00008890, -0.00030206, 0.00012079, 0.00055549, -0.00011763, -0.00123769, -0.00088904, 0.00096898, 0.00199970, 0.00087873, -0.00077936, -0.00097086, },
		{ 0.00011057, -0.00018181, -0.00008210, -0.00000416, -0.00006958, -0.00003892, 0.00005930, 0.00004874, 0.00000115, -0.00000421, -0.00006888, -0.00012409, 0.00003196, 0.00020755, 0.00004441, -0.00022460, -0.00011996, 0.00010423, -0.00008914, -0.00031115, 0.00012205, 0.00057294, -0.00012977, -0.00130654, -0.00093504, 0.00105031, 0.00216747, 0.00096909, -0.00083633, -0.00106640, -0.00006704, 0.00044808, },
		{ -0.00008317, 0.00001215, -0.00007632, -0.00006176, 0.00005772, 0.00006510, 0.00000411, -0.00001415, -0.00006878, -0.00011611, 0.00003031, 0.00020016, 0.00004506, -0.00021966, -0.00011916, 0.00010433, -0.00008915, -0.00031574, 0.00012285, 0.00058499, -0.00013879, -0.00136062, -0.00097278, 0.00111948, 0.00231453, 0.00105041, -0.00088883, -0.00115640, -0.00008544, 0.00047880, 0.00011309, -0.00020575, },
		{ -0.00008335, -0.00008602, 0.00005598, 0.00008325, 0.00000745, -0.00002588, -0.00006871, -0.00010599, 0.00002804, 0.00018937, 0.00004603, -0.00021009, -0.00011701, 0.00010118, -0.00008896, -0.00031566, 0.00012317, 0.00059116, -0.00014432, -0.00139786, -0.00100082, 0.00117383, 0.00243521, 0.00111956, -0.00093486, -0.00123741, -0.00010240, 0.00050773, 0.00011550, -0.00022924, -0.00008426, 0.00002876, },
		{ 0.00005415, 0.00010250, 0.00001106, -0.00003894, -0.00006865, -0.00009412, 0.00002524, 0.00017559, 0.00004730, -0.00019624, -0.00011358, 0.00009492, -0.00008856, -0.00031091, 0.00012299, 0.00059122, -0.00014615, -0.00141683, -0.00101808, 0.00121127, 0.00252489, 0.00117389, -0.00097264, -0.00130630, -0.00011728, 0.00053375, 0.00011771, -0.00025139, -0.00008532, 0.00004503, -0.00009040, -0.00011078, },
		{ 0.00001478, -0.00005284, -0.00006861, -0.00008096, 0.00002201, 0.00015936, 0.00004880, -0.00017865, -0.00010902, 0.00008577, -0.00008797, -0.00030168, 0.00012232, 0.00058516, -0.00014422, -0.00141679, -0.00102389, 0.00123036, 0.00258011, 0.00121131, -0.00100073, -0.00136045, -0.00012951, 0.00055587, 0.00011962, -0.00027133, -0.00008632, 0.00006034, -0.00009720, -0.00013509, 0.00005229, 0.00012210, },
		{ -0.00006860, -0.00006702, 0.00001848, 0.00014130, 0.00005049, -0.00015800, -0.00010348, 0.00007409, -0.00008722, -0.00028831, 0.00012118, 0.00057323, -0.00013860, -0.00139775, -0.00101803, 0.00123038, 0.00259876, 0.00123038, -0.00101803, -0.00139775, -0.00013860, 0.00057323, 0.00012118, -0.00028831, -0.00008722, 0.00007409, -0.00010348, -0.00015800, 0.00005049, 0.00014130, 0.00001848, -0.00006702, },
		{ 0.00001478, 0.00012210, 0.00005229, -0.00013509, -0.00009720, 0.00006034, -0.00008632, -0.00027133, 0.00011962, 0.00055587, -0.00012951, -0.00136045, -0.00100073, 0.00121131, 0.00258011, 0.00123036, -0.00102389, -0.00141679, -0.00014422, 0.00058516, 0.00012232, -0.00030168, -0.00008797, 0.00008577, -0.00010902, -0.00017865, 0.00004880, 0.00015936, 0.00002201, -0.00008096, -0.00006861, -0.00005284, },
		{ 0.00005415, -0.00011078, -0.00009040, 0.00004503, -0.00008532, -0.00025139, 0.00011771, 0.00053375, -0.00011728, -0.00130630, -0.00097264, 0.00117389, 0.00252489, 0.00121127, -0.00101808, -0.00141683, -0.00014615, 0.00059122, 0.00012299, -0.00031091, -0.00008856, 0.00009492, -0.00011358, -0.00019624, 0.00004730, 0.00017559, 0.00002524, -0.00009412, -0.00006865, -0.00003894, 0.00001106, 0.00010250, },
		{ -0.00008335, 0.00002876, -0.00008426, -0.00022924, 0.00011550, 0.00050773, -0.00010240, -0.00123741, -0.00093486, 0.00111956, 0.00243521, 0.00117383, -0.00100082, -0.00139786, -0.00014432, 0.00059116, 0.00012317, -0.00031566, -0.00008896, 0.00010118, -0.00011701, -0.00021009, 0.00004603, 0.00018937, 0.00002804, -0.00010599, -0.00006871, -0.00002588, 0.00000745, 0.00008325, 0.00005598, -0.00008602, },
		{ -0.00008317, -0.00020575, 0.00011309, 0.00047880, -0.00008544, -0.00115640, -0.00088883, 0.00105041, 0.00231453, 0.00111948, -0.00097278, -0.00136062, -0.00013879, 0.00058499, 0.00012285, -0.00031574, -0.00008915, 0.00010433, -0.00011916, -0.00021966, 0.00004506, 0.00020016, 0.00003031, -0.00011611, -0.00006878, -0.00001415, 0.00000411, 0.00006510, 0.00005772, -0.00006176, -0.00007632, 0.00001215, },
		{ 0.00011057, 0.00044808, -0.00006704, -0.00106640, -0.00083633, 0.00096909, 0.00216747, 0.00105031, -0.00093504, -0.00130654, -0.00012977, 0.00057294, 0.00012205, -0.00031115, -0.00008914, 0.00010423, -0.00011996, -0.00022460, 0.00004441, 0.00020755, 0.00003196, -0.00012409, -0.00006888, -0.00000421, 0.00000115, 0.00004874, 0.00005930, -0.00003892, -0.00006958, -0.00000416, -0.00008210, -0.00018181, },
		{ -0.00004792, -0.00097086, -0.00077936, 0.00087873, 0.00199970, 0.00096898, -0.00088904, -0.00123769, -0.00011763, 0.00055549, 0.00012079, -0.00030206, -0.00008890, 0.00010090, -0.00011937, -0.00022471, 0.00004411, 0.00021125, 0.00003293, -0.00012964, -0.00006899, 0.00000357, -0.00000131, 0.00003479, 0.00006065, -0.00001840, -0.00006339, -0.00001954, -0.00008109, -0.00015835, 0.00010803, 0.00041675, },
		{ -0.00072012, 0.00078281, 0.00181765, 0.00087861, -0.00083655, -0.00115673, -0.00010281, 0.00053329, 0.00011913, -0.00028883, -0.00008847, 0.00009445, -0.00011741, -0.00021999, 0.00004418, 0.00021114, 0.00003317, -0.00013253, -0.00006911, 0.00000889, -0.00000317, 0.00002381, 0.00006174, -0.00000096, -0.00005799, -0.00003340, -0.00008017, -0.00013626, 0.00010558, 0.00038600, -0.00002881, -0.00087346, },
		{ 0.00162833, 0.00078268, -0.00077960, -0.00106675, -0.00008589, 0.00050720, 0.00011712, -0.00027196, -0.00008785, 0.00008515, -0.00011417, -0.00021061, 0.00004461, 0.00020719, 0.00003268, -0.00013265, -0.00006923, 0.00001153, -0.00000437, 0.00001619, 0.00006251, 0.00001270, -0.00005358, -0.00004521, -0.00007939, -0.00011641, 0.00010331, 0.00035703, -0.00001044, -0.00077794, -0.00066088, 0.00068500, },
		{ -0.00072036, -0.00097123, -0.00006752, 0.00047823, 0.00011485, -0.00025210, -0.00008707, 0.00007333, -0.00010976, -0.00019695, 0.00004538, 0.00019958, 0.00003149, -0.00013000, -0.00006936, 0.00001141, -0.00000486, 0.00001225, 0.00006294, 0.00002207, -0.00005034, -0.00005452, -0.00007877, -0.00009954, 0.00010130, 0.00033095, 0.00000648, -0.00068797, -0.00060392, 0.00058907, 0.00143901, 0.00068488, },
		{ -0.00004841, 0.00044749, 0.00011240, -0.00023001, -0.00008615, 0.00005947, -0.00010436, -0.00017952, 0.00004647, 0.00018860, 0.00002962, -0.00012469, -0.00006948, 0.00000852, -0.00000462, 0.00001214, 0.00006301, 0.00002680, -0.00004838, -0.00006096, -0.00007833, -0.00008630, 0.00009964, 0.00030875, 0.00002130, -0.00060700, -0.00055144, 0.00049870, 0.00125696, 0.00058895, -0.00066112, -0.00087383, },
		{ 0.00010986, -0.00020655, -0.00008514, 0.00004409, -0.00009817, -0.00015899, 0.00004782, 0.00017465, 0.00002716, -0.00011691, -0.00006959, 0.00000298, -0.00000365, 0.00001584, 0.00006271, 0.00002669, -0.00004779, -0.00006430, -0.00007810, -0.00007721, 0.00009838, 0.00029129, 0.00003344, -0.00053816, -0.00050544, 0.00041737, 0.00108919, 0.00049859, -0.00060415, -0.00077829, -0.00002929, 0.00041615, },
		{ -0.00008407, 0.00002778, -0.00009143, -0.00013616, 0.00004940, 0.00015829, 0.00002420, -0.00010696, -0.00006968, -0.00000501, -0.00000200, 0.00002323, 0.00006206, 0.00002175, -0.00004859, -0.00006439, -0.00007808, -0.00007262, 0.00009758, 0.00027925, 0.00004246, -0.00048407, -0.00046770, 0.00034820, 0.00094213, 0.00041728, -0.00055164, -0.00068829, -0.00001089, 0.00038543, 0.00010734, -0.00018261, },
		{ -0.00008440, -0.00011189, 0.00005114, 0.00014014, 0.00002086, -0.00009523, -0.00006976, -0.00001513, 0.00000027, 0.00003402, 0.00006109, 0.00001217, -0.00005074, -0.00006125, -0.00007828, -0.00007270, 0.00009726, 0.00027308, 0.00004799, -0.00044683, -0.00043966, 0.00029385, 0.00082145, 0.00034812, -0.00050561, -0.00060729, 0.00000607, 0.00035651, 0.00010493, -0.00015912, -0.00008298, 0.00001118, },
		{ 0.00005297, 0.00012089, 0.00001726, -0.00008217, -0.00006982, -0.00002700, 0.00000307, 0.00004780, 0.00005982, -0.00000168, -0.00005417, -0.00005498, -0.00007868, -0.00007745, 0.00009744, 0.00027302, 0.00004982, -0.00042787, -0.00042240, 0.00025641, 0.00073177, 0.00029380, -0.00046783, -0.00053839, 0.00002095, 0.00033048, 0.00010272, -0.00013697, -0.00008192, -0.00000510, -0.00007735, -0.00008713, },
		{ 0.00001353, -0.00006828, -0.00006986, -0.00004015, 0.00000630, 0.00006403, 0.00005832, -0.00001926, -0.00005873, -0.00004583, -0.00007927, -0.00008668, 0.00009811, 0.00027907, 0.00004789, -0.00042790, -0.00041659, 0.00023732, 0.00067655, 0.00025638, -0.00043975, -0.00048425, 0.00003317, 0.00030837, 0.00010080, -0.00011703, -0.00008092, -0.00002040, -0.00007055, -0.00006283, 0.00005483, 0.00010129, },
		},
		{ { -0.00006771, 0.00004679, 0.00006176, -0.00014835, -0.00010261, 0.00013422, 0.00006969, -0.00004820, 0.00009477, -0.00000425, -0.00039155, -0.00022455, 0.00039573, 0.00041223, -0.00007437, -0.00014630, 0.00000405, -0.00014630, -0.00007437, 0.00041223, 0.00039573, -0.00022455, -0.00039155, -0.00000425, 0.00009477, -0.00004820, 0.00006969, 0.00013422, -0.00010261, -0.00014835, 0.00006176, 0.00004679, },
		{ 0.00007902, -0.00014318, -0.00014276, 0.00013816, 0.00011297, -0.00007573, 0.00006776, 0.00007097, -0.00039379, -0.00039541, 0.00037204, 0.00065987, 0.00004948, -0.00030309, -0.00004856, 0.00001203, -0.00008736, 0.00018155, 0.00033195, -0.00008023, -0.00034484, -0.00007307, 0.00009860, -0.00001126, 0.00004062, 0.00011696, -0.00007128, -0.00014625, 0.00004540, 0.00006068, -0.00006608, 0.00003116, },
		{ -0.00018712, 0.00012384, 0.00016433, -0.00008672, 0.00002009, 0.00013934, -0.00034679, -0.00056518, 0.00025408, 0.00087959, 0.00028187, -0.00040762, -0.00019540, 0.00012341, -0.00000701, 0.00000695, 0.00020111, 0.00001525, -0.00026547, -0.00012385, 0.00008058, 0.00002672, 0.00002897, 0.00009306, -0.00005130, -0.00014021, 0.00003066, 0.00007274, -0.00006126, 0.00001399, 0.00009580, -0.00012815, },
		{ 0.00021560, -0.00007655, -0.00004214, 0.00018810, -0.00025373, -0.00070516, 0.00005004, 0.00102794, 0.00060036, -0.00041574, -0.00040527, 0.00014988, 0.00013274, -0.00008424, 0.00003366, 0.00004829, -0.00017033, -0.00014841, 0.00004569, 0.00005744, 0.00003457, 0.00006977, -0.00004282, -0.00013369, 0.00001758, 0.00008289, -0.00005342, -0.00000439, 0.00011024, -0.00010194, -0.00022962, 0.00008898, },
		{ -0.00011016, 0.00020697, -0.00012543, -0.00079011, -0.00021775, 0.00106984, 0.00096531, -0.00029764, -0.00063185, 0.00007057, 0.00028753, -0.00008071, -0.00013493, 0.00001608, -0.00007865, -0.00014323, 0.00000162, 0.00007405, 0.00005397, 0.00005366, -0.00004367, -0.00012964, 0.00000558, 0.00009103, -0.00004287, -0.00002347, 0.00012024, -0.00006484, -0.00026364, 0.00003446, 0.00025788, -0.00004388, },
		{ 0.00002121, -0.00080220, -0.00051588, 0.00098443, 0.00132556, -0.00004349, -0.00082134, -0.00011442, 0.00041021, 0.00001121, -0.00026997, -0.00007311, -0.00000886, -0.00011002, -0.00004241, 0.00007233, 0.00008108, 0.00004952, -0.00004986, -0.00012998, -0.00000635, 0.00009699, -0.00003003, -0.00004254, 0.00012379, -0.00001885, -0.00028303, -0.00003559, 0.00028288, 0.00000901, -0.00017384, 0.00018987, },
		{ -0.00080468, 0.00076879, 0.00162608, 0.00033398, -0.00092133, -0.00038363, 0.00045895, 0.00016873, -0.00034314, -0.00020128, 0.00002440, -0.00005543, -0.00007718, 0.00005130, 0.00010822, 0.00005952, -0.00005630, -0.00013528, -0.00001939, 0.00010050, -0.00001537, -0.00006073, 0.00011928, 0.00003249, -0.00028308, -0.00011434, 0.00028429, 0.00007648, -0.00022323, 0.00013595, 0.00016565, -0.00073391, },
		{ 0.00181633, 0.00080043, -0.00088974, -0.00069720, 0.00040448, 0.00035642, -0.00033715, -0.00034363, 0.00001261, 0.00001005, -0.00009499, 0.00001351, 0.00012737, 0.00008282, -0.00005777, -0.00014461, -0.00003458, 0.00010125, 0.00000054, -0.00007705, 0.00010580, 0.00008452, -0.00026131, -0.00019306, 0.00025884, 0.00015029, -0.00025003, 0.00004982, 0.00028677, -0.00058942, -0.00104400, 0.00043903, },
		{ -0.00070222, -0.00100295, 0.00023536, 0.00053206, -0.00024855, -0.00047253, -0.00004516, 0.00007377, -0.00009090, -0.00003539, 0.00013164, 0.00011571, -0.00004985, -0.00015571, -0.00005255, 0.00009887, 0.00001708, -0.00009049, 0.00008333, 0.00013199, -0.00021792, -0.00026241, 0.00020687, 0.00022076, -0.00024898, -0.00005893, 0.00036608, -0.00038421, -0.00119923, 0.00002840, 0.00185814, 0.00130424, },
		{ -0.00003945, 0.00065355, -0.00008849, -0.00056196, -0.00014202, 0.00012292, -0.00006356, -0.00008747, 0.00011638, 0.00015212, -0.00002976, -0.00016538, -0.00007332, 0.00009304, 0.00003363, -0.00010010, 0.00005283, 0.00016985, -0.00015590, -0.00031371, 0.00013249, 0.00027821, -0.00021860, -0.00017716, 0.00039051, -0.00014293, -0.00124646, -0.00041701, 0.00173189, 0.00178364, -0.00035701, -0.00124445, },
		{ 0.00011894, -0.00059172, -0.00026420, 0.00014662, -0.00001552, -0.00013374, 0.00008008, 0.00018464, 0.00000305, -0.00016999, -0.00009619, 0.00008356, 0.00004953, -0.00010510, 0.00001620, 0.00019385, -0.00008072, -0.00034019, 0.00004302, 0.00031438, -0.00016157, -0.00028996, 0.00035444, 0.00010410, -0.00117614, -0.00084476, 0.00143999, 0.00217552, 0.00012364, -0.00137000, -0.00039139, 0.00068594, },
		{ -0.00039318, 0.00013772, 0.00004711, -0.00016560, 0.00002476, 0.00020568, 0.00004669, -0.00016614, -0.00011976, 0.00007040, 0.00006419, -0.00010497, -0.00002395, 0.00020117, 0.00000038, -0.00033799, -0.00005192, 0.00032368, -0.00008433, -0.00038285, 0.00026063, 0.00032520, -0.00099465, -0.00120338, 0.00100717, 0.00242469, 0.00069573, -0.00134109, -0.00077606, 0.00060739, 0.00034035, -0.00055075, },
		{ 0.00011555, -0.00017617, -0.00004422, 0.00020877, 0.00009700, -0.00015128, -0.00014202, 0.00005378, 0.00007704, -0.00009952, -0.00006453, 0.00019080, 0.00007929, -0.00030677, -0.00014173, 0.00030402, 0.00000377, -0.00044387, 0.00011992, 0.00049202, -0.00072344, -0.00144972, 0.00047747, 0.00249207, 0.00129901, -0.00113900, -0.00113971, 0.00041310, 0.00053820, -0.00043907, -0.00050844, 0.00009399, },
		{ -0.00011887, 0.00018962, 0.00014808, -0.00012420, -0.00016066, 0.00003420, 0.00008759, -0.00008893, -0.00010235, 0.00016373, 0.00014813, -0.00024974, -0.00021631, 0.00025718, 0.00009183, -0.00046532, -0.00005029, 0.00058358, -0.00039606, -0.00155507, -0.00009164, 0.00236071, 0.00186519, -0.00076853, -0.00142715, 0.00011674, 0.00067665, -0.00026801, -0.00059064, 0.00001864, 0.00017966, -0.00016143, },
		{ 0.00019312, -0.00008538, -0.00017333, 0.00001243, 0.00009542, -0.00007375, -0.00013448, 0.00012276, 0.00020031, -0.00017324, -0.00026744, 0.00018856, 0.00016895, -0.00044485, -0.00022858, 0.00058930, -0.00005330, -0.00150918, -0.00063708, 0.00203861, 0.00232740, -0.00025827, -0.00159022, -0.00025106, 0.00072728, -0.00005872, -0.00062465, -0.00008009, 0.00022944, -0.00012094, -0.00018967, 0.00014695, },
		{ -0.00017800, -0.00001048, 0.00010025, -0.00005484, -0.00015861, 0.00007216, 0.00023132, -0.00008589, -0.00028994, 0.00010646, 0.00022578, -0.00038581, -0.00039252, 0.00051054, 0.00026261, -0.00132162, -0.00109892, 0.00155800, 0.00262963, 0.00034266, -0.00159525, -0.00064678, 0.00067363, 0.00016089, -0.00060193, -0.00019000, 0.00025662, -0.00005796, -0.00024694, 0.00008285, 0.00022541, -0.00003702, },
		{ 0.00010188, -0.00003333, -0.00017327, 0.00001712, 0.00023935, 0.00000261, -0.00028231, 0.00002087, 0.00025589, -0.00029672, -0.00052197, 0.00036022, 0.00051413, -0.00102019, -0.00142853, 0.00097102, 0.00273471, 0.00097102, -0.00142853, -0.00102019, 0.00051413, 0.00036022, -0.00052197, -0.00029672, 0.00025589, 0.00002087, -0.00028231, 0.00000261, 0.00023935, 0.00001712, -0.00017327, -0.00003333, },
		{ -0.00017800, -0.00003702, 0.00022541, 0.00008285, -0.00024694, -0.00005796, 0.00025662, -0.00019000, -0.00060193, 0.00016089, 0.00067363, -0.00064678, -0.00159525, 0.00034266, 0.00262963, 0.00155800, -0.00109892, -0.00132162, 0.00026261, 0.00051054, -0.00039252, -0.00038581, 0.00022578, 0.00010646, -0.00028994, -0.00008589, 0.00023132, 0.00007216, -0.00015861, -0.00005484, 0.00010025, -0.00001048, },
		{ 0.00019312, 0.00014695, -0.00018967, -0.00012094, 0.00022944, -0.00008009, -0.00062465, -0.00005872, 0.00072728, -0.00025106, -0.00159022, -0.00025827, 0.00232740, 0.00203861, -0.00063708, -0.00150918, -0.00005330, 0.00058930, -0.00022858, -0.00044485, 0.00016895, 0.00018856, -0.00026744, -0.00017324, 0.00020031, 0.00012276, -0.00013448, -0.00007375, 0.00009542, 0.00001243, -0.00017333, -0.00008538, },
		{ -0.00011887, -0.00016143, 0.00017966, 0.00001864, -0.00059064, -0.00026801, 0.00067665, 0.00011674, -0.00142715, -0.00076853, 0.00186519, 0.00236071, -0.00009164, -0.00155507, -0.00039606, 0.00058358, -0.00005029, -0.00046532, 0.00009183, 0.00025718, -0.00021631, -0.00024974, 0.00014813, 0.00016373, -0.00010235, -0.00008893, 0.00008759, 0.00003420, -0.00016066, -0.00012420, 0.00014808, 0.00018962, },
		{ 0.00011555, 0.00009399, -0.00050844, -0.00043907, 0.00053820, 0.00041310, -0.00113971, -0.00113900, 0.00129901, 0.00249207, 0.00047747, -0.00144972, -0.00072344, 0.00049202, 0.00011992, -0.00044387, 0.00000377, 0.00030402, -0.00014173, -0.00030677, 0.00007929, 0.00019080, -0.00006453, -0.00009952, 0.00007704, 0.00005378, -0.00014202, -0.00015128, 0.00009700, 0.00020877, -0.00004422, -0.00017617, },
		{ -0.00039318, -0.00055075, 0.00034035, 0.00060739, -0.00077606, -0.00134109, 0.00069573, 0.00242469, 0.00100717, -0.00120338, -0.00099465, 0.00032520, 0.00026063, -0.00038285, -0.00008433, 0.00032368, -0.00005192, -0.00033799, 0.00000038, 0.00020117, -0.00002395, -0.00010497, 0.00006419, 0.00007040, -0.00011976, -0.00016614, 0.00004669, 0.00020568, 0.00002476, -0.00016560, 0.00004711, 0.00013772, },
		{ 0.00011894, 0.00068594, -0.00039139, -0.00137000, 0.00012364, 0.00217552, 0.00143999, -0.00084476, -0.00117614, 0.00010410, 0.00035444, -0.00028996, -0.00016157, 0.00031438, 0.00004302, -0.00034019, -0.00008072, 0.00019385, 0.00001620, -0.00010510, 0.00004953, 0.00008356, -0.00009619, -0.00016999, 0.00000305, 0.00018464, 0.00008008, -0.00013374, -0.00001552, 0.00014662, -0.00026420, -0.00059172, },
		{ -0.00003945, -0.00124445, -0.00035701, 0.00178364, 0.00173189, -0.00041701, -0.00124646, -0.00014293, 0.00039051, -0.00017716, -0.00021860, 0.00027821, 0.00013249, -0.00031371, -0.00015590, 0.00016985, 0.00005283, -0.00010010, 0.00003363, 0.00009304, -0.00007332, -0.00016538, -0.00002976, 0.00015212, 0.00011638, -0.00008747, -0.00006356, 0.00012292, -0.00014202, -0.00056196, -0.00008849, 0.00065355, },
		{ -0.00070222, 0.00130424, 0.00185814, 0.00002840, -0.00119923, -0.00038421, 0.00036608, -0.00005893, -0.00024898, 0.00022076, 0.00020687, -0.00026241, -0.00021792, 0.00013199, 0.00008333, -0.00009049, 0.00001708, 0.00009887, -0.00005255, -0.00015571, -0.00004985, 0.00011571, 0.00013164, -0.00003539, -0.00009090, 0.00007377, -0.00004516, -0.00047253, -0.00024855, 0.00053206, 0.00023536, -0.00100295, },
		{ 0.00181633, 0.00043903, -0.00104400, -0.00058942, 0.00028677, 0.00004982, -0.00025003, 0.00015029, 0.00025884, -0.00019306, -0.00026131, 0.00008452, 0.00010580, -0.00007705, 0.00000054, 0.00010125, -0.00003458, -0.00014461, -0.00005777, 0.00008282, 0.00012737, 0.00001351, -0.00009499, 0.00001005, 0.00001261, -0.00034363, -0.00033715, 0.00035642, 0.00040448, -0.00069720, -0.00088974, 0.00080043, },
		{ -0.00080468, -0.00073391, 0.00016565, 0.00013595, -0.00022323, 0.00007648, 0.00028429, -0.00011434, -0.00028308, 0.00003249, 0.00011928, -0.00006073, -0.00001537, 0.00010050, -0.00001939, -0.00013528, -0.00005630, 0.00005952, 0.00010822, 0.00005130, -0.00007718, -0.00005543, 0.00002440, -0.00020128, -0.00034314, 0.00016873, 0.00045895, -0.00038363, -0.00092133, 0.00033398, 0.00162608, 0.00076879, },
		{ 0.00002121, 0.00018987, -0.00017384, 0.00000901, 0.00028288, -0.00003559, -0.00028303, -0.00001885, 0.00012379, -0.00004254, -0.00003003, 0.00009699, -0.00000635, -0.00012998, -0.00004986, 0.00004952, 0.00008108, 0.00007233, -0.00004241, -0.00011002, -0.00000886, -0.00007311, -0.00026997, 0.00001121, 0.00041021, -0.00011442, -0.00082134, -0.00004349, 0.00132556, 0.00098443, -0.00051588, -0.00080220, },
		{ -0.00011016, -0.00004388, 0.00025788, 0.00003446, -0.00026364, -0.00006484, 0.00012024, -0.00002347, -0.00004287, 0.00009103, 0.00000558, -0.00012964, -0.00004367, 0.00005366, 0.00005397, 0.00007405, 0.00000162, -0.00014323, -0.00007865, 0.00001608, -0.00013493, -0.00008071, 0.00028753, 0.00007057, -0.00063185, -0.00029764, 0.00096531, 0.00106984, -0.00021775, -0.00079011, -0.00012543, 0.00020697, },
		{ 0.00021560, 0.00008898, -0.00022962, -0.00010194, 0.00011024, -0.00000439, -0.00005342, 0.00008289, 0.00001758, -0.00013369, -0.00004282, 0.00006977, 0.00003457, 0.00005744, 0.00004569, -0.00014841, -0.00017033, 0.00004829, 0.00003366, -0.00008424, 0.00013274, 0.00014988, -0.00040527, -0.00041574, 0.00060036, 0.00102794, 0.00005004, -0.00070516, -0.00025373, 0.00018810, -0.00004214, -0.00007655, },
		{ -0.00018712, -0.00012815, 0.00009580, 0.00001399, -0.00006126, 0.00007274, 0.00003066, -0.00014021, -0.00005130, 0.00009306, 0.00002897, 0.00002672, 0.00008058, -0.00012385, -0.00026547, 0.00001525, 0.00020111, 0.00000695, -0.00000701, 0.00012341, -0.00019540, -0.00040762, 0.00028187, 0.00087959, 0.00025408, -0.00056518, -0.00034679, 0.00013934, 0.00002009, -0.00008672, 0.00016433, 0.00012384, },
		{ 0.00007902, 0.00003116, -0.00006608, 0.00006068, 0.00004540, -0.00014625, -0.00007128, 0.00011696, 0.00004062, -0.00001126, 0.00009860, -0.00007307, -0.00034484, -0.00008023, 0.00033195, 0.00018155, -0.00008736, 0.00001203, -0.00004856, -0.00030309, 0.00004948, 0.00065987, 0.00037204, -0.00039541, -0.00039379, 0.00007097, 0.00006776, -0.00007573, 0.00011297, 0.00013816, -0.00014276, -0.00014318, },
		},
		{ { 0.00001861, -0.00010060, -0.00004568, 0.00015084, -0.00000697, -0.00003584, 0.00007798, -0.00034631, -0.00040875, 0.00052105, 0.00071794, -0.00026234, -0.00050554, 0.00015813, 0.00029635, -0.00006335, -0.00022613, -0.00006335, 0.00029635, 0.00015813, -0.00050554, -0.00026234, 0.00071794, 0.00052105, -0.00040875, -0.00034631, 0.00007798, -0.00003584, -0.00000697, 0.00015084, -0.00004568, -0.00010060, },
		{ -0.00011503, 0.00014184, 0.00005966, -0.00011517, 0.00011368, -0.00013959, -0.00062671, 0.00013773, 0.00101181, 0.00020669, -0.00068470, -0.00013526, 0.00039908, 0.00005115, -0.00024020, -0.00011540, 0.00017222, 0.00026306, -0.00021084, -0.00044259, 0.00027948, 0.00063396, -0.00007456, -0.00041209, -0.00004609, 0.00002183, -0.00002476, 0.00012750, 0.00001478, -0.00011107, 0.00001533, -0.00006941, },
		{ 0.00015390, -0.00014929, 0.00002967, 0.00009625, -0.00060733, -0.00037261, 0.00096866, 0.00082778, -0.00055665, -0.00055057, 0.00034845, 0.00025417, -0.00024046, -0.00017744, 0.00009732, 0.00024046, 0.00002128, -0.00036559, -0.00008381, 0.00048112, 0.00020473, -0.00031128, -0.00017347, 0.00001652, -0.00001150, 0.00010176, 0.00005831, -0.00011029, 0.00000059, -0.00001474, -0.00016652, 0.00007692, },
		{ -0.00012794, 0.00022684, -0.00035167, -0.00078324, 0.00053828, 0.00133655, -0.00004216, -0.00088707, 0.00004843, 0.00046787, -0.00013974, -0.00028241, 0.00004811, 0.00019714, 0.00011848, -0.00017626, -0.00023972, 0.00019573, 0.00030418, -0.00011568, -0.00022147, -0.00004173, -0.00000413, 0.00008743, 0.00009279, -0.00010174, -0.00003494, 0.00004899, -0.00016747, -0.00003831, 0.00022214, -0.00009095, },
		{ 0.00001191, -0.00090073, -0.00014027, 0.00146538, 0.00074250, -0.00089815, -0.00046115, 0.00053156, 0.00010628, -0.00037651, -0.00004717, 0.00019208, 0.00012432, -0.00002559, -0.00019580, -0.00004761, 0.00021001, 0.00005180, -0.00016316, -0.00010108, -0.00002737, 0.00007517, 0.00012766, -0.00007790, -0.00009301, 0.00009112, -0.00010136, -0.00016212, 0.00020953, 0.00005133, -0.00025976, 0.00017441, },
		{ -0.00079225, 0.00108825, 0.00150421, -0.00043398, -0.00097452, 0.00029993, 0.00042718, -0.00034869, -0.00022403, 0.00019508, 0.00013208, 0.00002027, -0.00006603, -0.00013394, 0.00001759, 0.00009275, -0.00004500, -0.00010604, -0.00007263, 0.00004456, 0.00015992, -0.00002685, -0.00015820, 0.00007906, 0.00001486, -0.00023355, 0.00009744, 0.00021054, -0.00027214, -0.00003416, 0.00029469, -0.00067008, },
		{ 0.00189744, 0.00044216, -0.00119528, -0.00023826, 0.00064194, -0.00012032, -0.00042447, 0.00012249, 0.00019499, 0.00000166, 0.00001939, -0.00006419, -0.00013318, -0.00000762, 0.00004236, -0.00003921, -0.00010202, -0.00001406, 0.00017144, 0.00005416, -0.00019942, -0.00000097, 0.00013088, -0.00020568, -0.00007657, 0.00029591, -0.00013595, -0.00027940, 0.00035027, -0.00020869, -0.00113866, 0.00030518, },
		{ -0.00088939, -0.00090029, 0.00055183, 0.00026666, -0.00051425, -0.00007900, 0.00027680, 0.00001202, 0.00000154, 0.00006250, -0.00014483, -0.00017274, 0.00002257, 0.00006079, -0.00007552, -0.00008423, 0.00014076, 0.00014824, -0.00018408, -0.00012946, 0.00018397, -0.00007680, -0.00023125, 0.00024647, 0.00009064, -0.00041578, 0.00015538, 0.00025183, -0.00103709, -0.00058642, 0.00169609, 0.00143938, },
		{ 0.00007463, 0.00063096, -0.00037080, -0.00036318, 0.00027663, 0.00011113, -0.00007664, 0.00012487, -0.00001897, -0.00028288, -0.00011459, 0.00011868, 0.00001607, -0.00012690, 0.00006194, 0.00022066, -0.00009801, -0.00025590, 0.00013149, 0.00009995, -0.00028274, 0.00007161, 0.00029071, -0.00035496, -0.00017704, 0.00048588, -0.00055603, -0.00122573, 0.00092155, 0.00214982, -0.00003504, -0.00136085, },
		{ 0.00001331, -0.00058469, 0.00011099, 0.00026660, -0.00011226, 0.00006356, 0.00015004, -0.00024987, -0.00030044, 0.00006892, 0.00013485, -0.00010305, -0.00004278, 0.00023457, 0.00004082, -0.00031951, -0.00002246, 0.00023725, -0.00019245, -0.00014671, 0.00035361, -0.00011773, -0.00046025, 0.00039097, 0.00005891, -0.00136140, -0.00014636, 0.00223775, 0.00111997, -0.00130382, -0.00066234, 0.00073032, },
		{ -0.00020461, 0.00036534, -0.00002632, -0.00007975, 0.00023899, -0.00007331, -0.00042354, -0.00009837, 0.00020957, -0.00000238, -0.00012710, 0.00017280, 0.00018447, -0.00027920, -0.00022228, 0.00025592, 0.00000569, -0.00029438, 0.00023942, 0.00017580, -0.00053655, 0.00003802, 0.00051020, -0.00097346, -0.00108912, 0.00161165, 0.00213927, -0.00060020, -0.00133875, 0.00040341, 0.00048380, -0.00057872, },
		{ 0.00016799, -0.00019207, 0.00017701, 0.00015606, -0.00039462, -0.00032118, 0.00017424, 0.00014069, -0.00014238, 0.00005291, 0.00027465, -0.00013714, -0.00038291, 0.00012743, 0.00021614, -0.00028718, 0.00000526, 0.00037275, -0.00036296, -0.00037160, 0.00059827, -0.00027957, -0.00154314, 0.00048616, 0.00258869, 0.00058709, -0.00157710, -0.00030572, 0.00077862, -0.00027521, -0.00053460, 0.00028973, },
		{ -0.00000652, 0.00030870, -0.00020259, -0.00049179, 0.00001038, 0.00025628, -0.00006609, -0.00007514, 0.00027104, 0.00005771, -0.00042924, -0.00010670, 0.00032746, -0.00011999, -0.00022336, 0.00037298, -0.00003198, -0.00061739, 0.00031670, 0.00037758, -0.00136924, -0.00070471, 0.00224475, 0.00183994, -0.00113668, -0.00112170, 0.00066541, 0.00022592, -0.00068655, 0.00000954, 0.00035966, -0.00016808, },
		{ 0.00007245, -0.00051315, -0.00023279, 0.00027205, 0.00007946, -0.00015024, 0.00017342, 0.00022962, -0.00033317, -0.00035080, 0.00026982, 0.00012558, -0.00032091, 0.00018486, 0.00028297, -0.00057680, -0.00015614, 0.00070184, -0.00071109, -0.00150420, 0.00121280, 0.00265823, -0.00006326, -0.00164236, 0.00008988, 0.00067554, -0.00052741, -0.00037336, 0.00040436, 0.00001265, -0.00019723, 0.00029370, },
		{ -0.00045488, 0.00015333, 0.00023096, -0.00013054, 0.00002514, 0.00031072, -0.00012895, -0.00049918, 0.00005325, 0.00032102, -0.00023204, -0.00008053, 0.00042523, -0.00028443, -0.00056065, 0.00058400, 0.00008584, -0.00164135, -0.00011028, 0.00268967, 0.00129332, -0.00153182, -0.00075410, 0.00079466, -0.00008346, -0.00066363, 0.00022075, 0.00026049, -0.00026585, 0.00011532, 0.00029886, -0.00035458, },
		{ 0.00031136, -0.00001723, -0.00010542, 0.00027313, 0.00010178, -0.00048675, -0.00023426, 0.00035774, -0.00000357, -0.00027308, 0.00033971, 0.00009674, -0.00069542, 0.00013978, 0.00065011, -0.00114667, -0.00121441, 0.00190065, 0.00241575, -0.00070483, -0.00148510, 0.00043355, 0.00044298, -0.00067960, -0.00014257, 0.00042198, -0.00015076, -0.00012576, 0.00036460, -0.00007449, -0.00055161, -0.00007201, },
		{ -0.00015707, 0.00014198, 0.00026800, -0.00031989, -0.00046894, 0.00020471, 0.00024008, -0.00028851, 0.00009765, 0.00037197, -0.00051598, -0.00036871, 0.00076206, -0.00031913, -0.00170764, 0.00059915, 0.00284980, 0.00059915, -0.00170764, -0.00031913, 0.00076206, -0.00036871, -0.00051598, 0.00037197, 0.00009765, -0.00028851, 0.00024008, 0.00020471, -0.00046894, -0.00031989, 0.00026800, 0.00014198, },
		{ 0.00031136, -0.00007201, -0.00055161, -0.00007449, 0.00036460, -0.00012576, -0.00015076, 0.00042198, -0.00014257, -0.00067960, 0.00044298, 0.00043355, -0.00148510, -0.00070483, 0.00241575, 0.00190065, -0.00121441, -0.00114667, 0.00065011, 0.00013978, -0.00069542, 0.00009674, 0.00033971, -0.00027308, -0.00000357, 0.00035774, -0.00023426, -0.00048675, 0.00010178, 0.00027313, -0.00010542, -0.00001723, },
		{ -0.00045488, -0.00035458, 0.00029886, 0.00011532, -0.00026585, 0.00026049, 0.00022075, -0.00066363, -0.00008346, 0.00079466, -0.00075410, -0.00153182, 0.00129332, 0.00268967, -0.00011028, -0.00164135, 0.00008584, 0.00058400, -0.00056065, -0.00028443, 0.00042523, -0.00008053, -0.00023204, 0.00032102, 0.00005325, -0.00049918, -0.00012895, 0.00031072, 0.00002514, -0.00013054, 0.00023096, 0.00015333, },
		{ 0.00007245, 0.00029370, -0.00019723, 0.00001265, 0.00040436, -0.00037336, -0.00052741, 0.00067554, 0.00008988, -0.00164236, -0.00006326, 0.00265823, 0.00121280, -0.00150420, -0.00071109, 0.00070184, -0.00015614, -0.00057680, 0.00028297, 0.00018486, -0.00032091, 0.00012558, 0.00026982, -0.00035080, -0.00033317, 0.00022962, 0.00017342, -0.00015024, 0.00007946, 0.00027205, -0.00023279, -0.00051315, },
		{ -0.00000652, -0.00016808, 0.00035966, 0.00000954, -0.00068655, 0.00022592, 0.00066541, -0.00112170, -0.00113668, 0.00183994, 0.00224475, -0.00070471, -0.00136924, 0.00037758, 0.00031670, -0.00061739, -0.00003198, 0.00037298, -0.00022336, -0.00011999, 0.00032746, -0.00010670, -0.00042924, 0.00005771, 0.00027104, -0.00007514, -0.00006609, 0.00025628, 0.00001038, -0.00049179, -0.00020259, 0.00030870, },
		{ 0.00016799, 0.00028973, -0.00053460, -0.00027521, 0.00077862, -0.00030572, -0.00157710, 0.00058709, 0.00258869, 0.00048616, -0.00154314, -0.00027957, 0.00059827, -0.00037160, -0.00036296, 0.00037275, 0.00000526, -0.00028718, 0.00021614, 0.00012743, -0.00038291, -0.00013714, 0.00027465, 0.00005291, -0.00014238, 0.00014069, 0.00017424, -0.00032118, -0.00039462, 0.00015606, 0.00017701, -0.00019207, },
		{ -0.00020461, -0.00057872, 0.00048380, 0.00040341, -0.00133875, -0.00060020, 0.00213927, 0.00161165, -0.00108912, -0.00097346, 0.00051020, 0.00003802, -0.00053655, 0.00017580, 0.00023942, -0.00029438, 0.00000569, 0.00025592, -0.00022228, -0.00027920, 0.00018447, 0.00017280, -0.00012710, -0.00000238, 0.00020957, -0.00009837, -0.00042354, -0.00007331, 0.00023899, -0.00007975, -0.00002632, 0.00036534, },
		{ 0.00001331, 0.00073032, -0.00066234, -0.00130382, 0.00111997, 0.00223775, -0.00014636, -0.00136140, 0.00005891, 0.00039097, -0.00046025, -0.00011773, 0.00035361, -0.00014671, -0.00019245, 0.00023725, -0.00002246, -0.00031951, 0.00004082, 0.00023457, -0.00004278, -0.00010305, 0.00013485, 0.00006892, -0.00030044, -0.00024987, 0.00015004, 0.00006356, -0.00011226, 0.00026660, 0.00011099, -0.00058469, },
		{ 0.00007463, -0.00136085, -0.00003504, 0.00214982, 0.00092155, -0.00122573, -0.00055603, 0.00048588, -0.00017704, -0.00035496, 0.00029071, 0.00007161, -0.00028274, 0.00009995, 0.00013149, -0.00025590, -0.00009801, 0.00022066, 0.00006194, -0.00012690, 0.00001607, 0.00011868, -0.00011459, -0.00028288, -0.00001897, 0.00012487, -0.00007664, 0.00011113, 0.00027663, -0.00036318, -0.00037080, 0.00063096, },
		{ -0.00088939, 0.00143938, 0.00169609, -0.00058642, -0.00103709, 0.00025183, 0.00015538, -0.00041578, 0.00009064, 0.00024647, -0.00023125, -0.00007680, 0.00018397, -0.00012946, -0.00018408, 0.00014824, 0.00014076, -0.00008423, -0.00007552, 0.00006079, 0.00002257, -0.00017274, -0.00014483, 0.00006250, 0.00000154, 0.00001202, 0.00027680, -0.00007900, -0.00051425, 0.00026666, 0.00055183, -0.00090029, },
		{ 0.00189744, 0.00030518, -0.00113866, -0.00020869, 0.00035027, -0.00027940, -0.00013595, 0.00029591, -0.00007657, -0.00020568, 0.00013088, -0.00000097, -0.00019942, 0.00005416, 0.00017144, -0.00001406, -0.00010202, -0.00003921, 0.00004236, -0.00000762, -0.00013318, -0.00006419, 0.00001939, 0.00000166, 0.00019499, 0.00012249, -0.00042447, -0.00012032, 0.00064194, -0.00023826, -0.00119528, 0.00044216, },
		{ -0.00079225, -0.00067008, 0.00029469, -0.00003416, -0.00027214, 0.00021054, 0.00009744, -0.00023355, 0.00001486, 0.00007906, -0.00015820, -0.00002685, 0.00015992, 0.00004456, -0.00007263, -0.00010604, -0.00004500, 0.00009275, 0.00001759, -0.00013394, -0.00006603, 0.00002027, 0.00013208, 0.00019508, -0.00022403, -0.00034869, 0.00042718, 0.00029993, -0.00097452, -0.00043398, 0.00150421, 0.00108825, },
		{ 0.00001191, 0.00017441, -0.00025976, 0.00005133, 0.00020953, -0.00016212, -0.00010136, 0.00009112, -0.00009301, -0.00007790, 0.00012766, 0.00007517, -0.00002737, -0.00010108, -0.00016316, 0.00005180, 0.00021001, -0.00004761, -0.00019580, -0.00002559, 0.00012432, 0.00019208, -0.00004717, -0.00037651, 0.00010628, 0.00053156, -0.00046115, -0.00089815, 0.00074250, 0.00146538, -0.00014027, -0.00090073, },
		{ -0.00012794, -0.00009095, 0.00022214, -0.00003831, -0.00016747, 0.00004899, -0.00003494, -0.00010174, 0.00009279, 0.00008743, -0.00000413, -0.00004173, -0.00022147, -0.00011568, 0.00030418, 0.00019573, -0.00023972, -0.00017626, 0.00011848, 0.00019714, 0.00004811, -0.00028241, -0.00013974, 0.00046787, 0.00004843, -0.00088707, -0.00004216, 0.00133655, 0.00053828, -0.00078324, -0.00035167, 0.00022684, },
		{ 0.00015390, 0.00007692, -0.00016652, -0.00001474, 0.00000059, -0.00011029, 0.00005831, 0.00010176, -0.00001150, 0.00001652, -0.00017347, -0.00031128, 0.00020473, 0.00048112, -0.00008381, -0.00036559, 0.00002128, 0.00024046, 0.00009732, -0.00017744, -0.00024046, 0.00025417, 0.00034845, -0.00055057, -0.00055665, 0.00082778, 0.00096866, -0.00037261, -0.00060733, 0.00009625, 0.00002967, -0.00014929, },
		{ -0.00011503, -0.00006941, 0.00001533, -0.00011107, 0.00001478, 0.00012750, -0.00002476, 0.00002183, -0.00004609, -0.00041209, -0.00007456, 0.00063396, 0.00027948, -0.00044259, -0.00021084, 0.00026306, 0.00017222, -0.00011540, -0.00024020, 0.00005115, 0.00039908, -0.00013526, -0.00068470, 0.00020669, 0.00101181, 0.00013773, -0.00062671, -0.00013959, 0.00011368, -0.00011517, 0.00005966, 0.00014184, },
		},
		{ { -0.00001017, 0.00006281, -0.00006738, -0.00006222, 0.00004185, -0.00055273, -0.00029523, 0.00095863, 0.00046881, -0.00063283, -0.00002664, 0.00040312, -0.00015155, -0.00018428, 0.00005476, 0.00002907, 0.00000405, 0.00002907, 0.00005476, -0.00018428, -0.00015155, 0.00040312, -0.00002664, -0.00063283, 0.00046881, 0.00095863, -0.00029523, -0.00055273, 0.00004185, -0.00006222, -0.00006738, 0.00006281, },
		{ 0.00004561, -0.00016195, 0.00007925, -0.00020101, -0.00076098, 0.00039265, 0.00117028, -0.00020977, -0.00058896, 0.00039169, 0.00017223, -0.00031918, -0.00002507, 0.00006759, -0.00001806, 0.00003223, 0.00005638, -0.00004780, -0.00021591, 0.00012661, 0.00032273, -0.00041168, -0.00025005, 0.00082930, 0.00036919, -0.00054963, -0.00020763, 0.00003099, -0.00009489, 0.00002418, -0.00001095, 0.00004894, },
		{ -0.00009951, 0.00012616, -0.00065179, -0.00048656, 0.00117526, 0.00075944, -0.00076189, -0.00012971, 0.00051894, -0.00019565, -0.00025023, 0.00010955, -0.00000887, -0.00000923, 0.00007772, 0.00000850, -0.00014028, -0.00007744, 0.00029321, 0.00000782, -0.00048324, 0.00025702, 0.00070137, -0.00016218, -0.00041101, -0.00001989, -0.00005156, -0.00001785, -0.00000121, -0.00002317, 0.00013624, -0.00010366, },
		{ -0.00014340, -0.00096416, 0.00033436, 0.00152295, -0.00008766, -0.00079885, 0.00040857, 0.00025228, -0.00041671, -0.00002023, 0.00010056, -0.00007799, 0.00006467, 0.00006679, -0.00008244, -0.00011950, 0.00011939, 0.00022030, -0.00028525, -0.00022406, 0.00053022, 0.00027629, -0.00035418, -0.00016522, -0.00002498, -0.00004303, 0.00002618, -0.00008076, 0.00009035, 0.00008908, -0.00023653, 0.00010930, },
		{ -0.00071991, 0.00132208, 0.00111326, -0.00084349, -0.00029913, 0.00062157, -0.00020204, -0.00033243, 0.00016485, -0.00002884, -0.00004705, 0.00012825, -0.00002274, -0.00012028, 0.00001480, 0.00020533, -0.00001872, -0.00033972, 0.00012774, 0.00043534, -0.00009424, -0.00024675, -0.00005066, -0.00006479, 0.00003706, -0.00004755, -0.00006019, 0.00020930, -0.00012063, -0.00016237, 0.00022077, -0.00069251, },
		{ 0.00185221, 0.00012167, -0.00101606, 0.00036064, 0.00035672, -0.00049603, -0.00003669, 0.00015204, -0.00013787, 0.00007382, 0.00009534, -0.00012143, -0.00004085, 0.00014799, 0.00010571, -0.00021840, -0.00015435, 0.00029461, 0.00013475, -0.00019424, -0.00008238, -0.00007779, -0.00001160, 0.00004954, -0.00015042, 0.00010748, 0.00014983, -0.00030955, 0.00011280, -0.00003987, -0.00113398, 0.00019267, },
		{ -0.00085470, -0.00052263, 0.00069188, -0.00016836, -0.00042229, 0.00021224, -0.00002630, -0.00010372, 0.00017170, -0.00003502, -0.00012262, 0.00011100, 0.00015018, -0.00009047, -0.00022431, 0.00007314, 0.00019055, -0.00008824, -0.00008144, -0.00004718, -0.00009597, 0.00008899, -0.00007067, -0.00011554, 0.00027217, -0.00011487, -0.00024596, 0.00031143, -0.00066600, -0.00097151, 0.00137694, 0.00150389, },
		{ 0.00024336, 0.00047088, -0.00054656, -0.00007467, 0.00021579, -0.00018804, 0.00005783, 0.00013586, -0.00015692, 0.00001235, 0.00020378, -0.00000045, -0.00020945, -0.00006629, 0.00013302, -0.00002956, -0.00008231, 0.00002705, -0.00011789, -0.00000266, 0.00008911, -0.00020987, 0.00009818, 0.00022046, -0.00037182, 0.00008494, 0.00009808, -0.00124772, -0.00002271, 0.00212270, 0.00041166, -0.00121182, },
		{ -0.00009801, -0.00050948, 0.00024426, -0.00000019, -0.00017066, 0.00020035, -0.00002750, -0.00014232, 0.00019683, 0.00013151, -0.00018626, -0.00015112, 0.00007621, -0.00000903, -0.00013204, 0.00006515, -0.00001169, -0.00013702, 0.00013028, -0.00006850, -0.00018072, 0.00031637, -0.00008624, -0.00034216, 0.00038401, -0.00057166, -0.00121329, 0.00132804, 0.00189721, -0.00078053, -0.00077923, 0.00071375, },
		{ -0.00013106, 0.00028348, -0.00022032, 0.00001876, 0.00018047, -0.00018196, 0.00003660, 0.00027361, -0.00008157, -0.00024700, 0.00001632, 0.00004435, -0.00018985, -0.00002240, 0.00014614, -0.00013858, -0.00001488, 0.00013750, -0.00025061, 0.00006424, 0.00029192, -0.00041540, 0.00002535, 0.00025411, -0.00128849, -0.00029176, 0.00230362, 0.00076387, -0.00135813, 0.00006155, 0.00057764, -0.00056121, },
		{ 0.00004689, -0.00023781, 0.00020894, -0.00000253, -0.00017202, 0.00025997, 0.00014572, -0.00028320, -0.00012436, 0.00011426, -0.00014858, -0.00020686, 0.00017781, 0.00004843, -0.00017905, 0.00015518, -0.00004274, -0.00024570, 0.00033606, -0.00003790, -0.00044069, 0.00042672, -0.00041740, -0.00141675, 0.00117587, 0.00225554, -0.00061669, -0.00104198, 0.00067676, 0.00000029, -0.00058335, 0.00025551, },
		{ -0.00003774, 0.00022085, -0.00019169, 0.00003299, 0.00034192, -0.00012761, -0.00031214, 0.00007532, 0.00001701, -0.00032250, -0.00000988, 0.00025747, -0.00013705, -0.00004392, 0.00018622, -0.00026690, 0.00001112, 0.00035487, -0.00043470, -0.00006167, 0.00040939, -0.00124781, -0.00058669, 0.00237302, 0.00114975, -0.00143190, -0.00017086, 0.00066005, -0.00053740, -0.00019978, 0.00034591, -0.00022932, },
		{ 0.00003477, -0.00020330, 0.00029197, 0.00018268, -0.00036046, -0.00013791, 0.00016329, -0.00022414, -0.00029062, 0.00024790, 0.00012184, -0.00021387, 0.00016067, 0.00000148, -0.00030048, 0.00032887, 0.00002412, -0.00053047, 0.00043188, -0.00021875, -0.00155699, 0.00093341, 0.00254255, -0.00037087, -0.00128157, 0.00057803, 0.00011373, -0.00063446, 0.00024349, 0.00010899, -0.00029504, 0.00019554, },
		{ 0.00000682, 0.00039338, -0.00013649, -0.00038189, 0.00010112, 0.00003165, -0.00041068, -0.00003270, 0.00034572, -0.00011481, -0.00008307, 0.00022710, -0.00025668, -0.00005290, 0.00040100, -0.00042732, -0.00016763, 0.00054521, -0.00112691, -0.00087568, 0.00232094, 0.00153408, -0.00141857, -0.00043244, 0.00070392, -0.00047749, -0.00027264, 0.00039434, -0.00021326, -0.00010327, 0.00024956, -0.00018409, },
		{ 0.00022757, -0.00040272, -0.00017669, 0.00020228, -0.00024248, -0.00036169, 0.00027435, 0.00019510, -0.00023523, 0.00014691, 0.00005641, -0.00033657, 0.00029626, 0.00009179, -0.00060114, 0.00039719, 0.00000325, -0.00161622, 0.00062456, 0.00272809, -0.00006218, -0.00147039, 0.00042298, 0.00022733, -0.00065596, 0.00020905, 0.00017766, -0.00033380, 0.00016184, 0.00007731, -0.00022805, 0.00028950, },
		{ -0.00043316, 0.00009424, 0.00006500, -0.00044674, -0.00007381, 0.00040000, -0.00007738, -0.00012447, 0.00025346, -0.00022184, -0.00011798, 0.00042429, -0.00039441, -0.00028092, 0.00064571, -0.00093683, -0.00112721, 0.00215086, 0.00187939, -0.00131438, -0.00069695, 0.00070005, -0.00038846, -0.00034046, 0.00042185, -0.00017424, -0.00016792, 0.00026125, -0.00016031, -0.00003340, 0.00041534, -0.00011424, },
		{ 0.00021697, -0.00021942, -0.00040255, 0.00026347, 0.00025476, -0.00023989, 0.00011718, 0.00011290, -0.00034832, 0.00024324, 0.00015624, -0.00064458, 0.00032608, 0.00022489, -0.00158641, 0.00028097, 0.00279225, 0.00028097, -0.00158641, 0.00022489, 0.00032608, -0.00064458, 0.00015624, 0.00024324, -0.00034832, 0.00011290, 0.00011718, -0.00023989, 0.00025476, 0.00026347, -0.00040255, -0.00021942, },
		{ -0.00043316, -0.00011424, 0.00041534, -0.00003340, -0.00016031, 0.00026125, -0.00016792, -0.00017424, 0.00042185, -0.00034046, -0.00038846, 0.00070005, -0.00069695, -0.00131438, 0.00187939, 0.00215086, -0.00112721, -0.00093683, 0.00064571, -0.00028092, -0.00039441, 0.00042429, -0.00011798, -0.00022184, 0.00025346, -0.00012447, -0.00007738, 0.00040000, -0.00007381, -0.00044674, 0.00006500, 0.00009424, },
		{ 0.00022757, 0.00028950, -0.00022805, 0.00007731, 0.00016184, -0.00033380, 0.00017766, 0.00020905, -0.00065596, 0.00022733, 0.00042298, -0.00147039, -0.00006218, 0.00272809, 0.00062456, -0.00161622, 0.00000325, 0.00039719, -0.00060114, 0.00009179, 0.00029626, -0.00033657, 0.00005641, 0.00014691, -0.00023523, 0.00019510, 0.00027435, -0.00036169, -0.00024248, 0.00020228, -0.00017669, -0.00040272, },
		{ 0.00000682, -0.00018409, 0.00024956, -0.00010327, -0.00021326, 0.00039434, -0.00027264, -0.00047749, 0.00070392, -0.00043244, -0.00141857, 0.00153408, 0.00232094, -0.00087568, -0.00112691, 0.00054521, -0.00016763, -0.00042732, 0.00040100, -0.00005290, -0.00025668, 0.00022710, -0.00008307, -0.00011481, 0.00034572, -0.00003270, -0.00041068, 0.00003165, 0.00010112, -0.00038189, -0.00013649, 0.00039338, },
		{ 0.00003477, 0.00019554, -0.00029504, 0.00010899, 0.00024349, -0.00063446, 0.00011373, 0.00057803, -0.00128157, -0.00037087, 0.00254255, 0.00093341, -0.00155699, -0.00021875, 0.00043188, -0.00053047, 0.00002412, 0.00032887, -0.00030048, 0.00000148, 0.00016067, -0.00021387, 0.00012184, 0.00024790, -0.00029062, -0.00022414, 0.00016329, -0.00013791, -0.00036046, 0.00018268, 0.00029197, -0.00020330, },
		{ -0.00003774, -0.00022932, 0.00034591, -0.00019978, -0.00053740, 0.00066005, -0.00017086, -0.00143190, 0.00114975, 0.00237302, -0.00058669, -0.00124781, 0.00040939, -0.00006167, -0.00043470, 0.00035487, 0.00001112, -0.00026690, 0.00018622, -0.00004392, -0.00013705, 0.00025747, -0.00000988, -0.00032250, 0.00001701, 0.00007532, -0.00031214, -0.00012761, 0.00034192, 0.00003299, -0.00019169, 0.00022085, },
		{ 0.00004689, 0.00025551, -0.00058335, 0.00000029, 0.00067676, -0.00104198, -0.00061669, 0.00225554, 0.00117587, -0.00141675, -0.00041740, 0.00042672, -0.00044069, -0.00003790, 0.00033606, -0.00024570, -0.00004274, 0.00015518, -0.00017905, 0.00004843, 0.00017781, -0.00020686, -0.00014858, 0.00011426, -0.00012436, -0.00028320, 0.00014572, 0.00025997, -0.00017202, -0.00000253, 0.00020894, -0.00023781, },
		{ -0.00013106, -0.00056121, 0.00057764, 0.00006155, -0.00135813, 0.00076387, 0.00230362, -0.00029176, -0.00128849, 0.00025411, 0.00002535, -0.00041540, 0.00029192, 0.00006424, -0.00025061, 0.00013750, -0.00001488, -0.00013858, 0.00014614, -0.00002240, -0.00018985, 0.00004435, 0.00001632, -0.00024700, -0.00008157, 0.00027361, 0.00003660, -0.00018196, 0.00018047, 0.00001876, -0.00022032, 0.00028348, },
		{ -0.00009801, 0.00071375, -0.00077923, -0.00078053, 0.00189721, 0.00132804, -0.00121329, -0.00057166, 0.00038401, -0.00034216, -0.00008624, 0.00031637, -0.00018072, -0.00006850, 0.00013028, -0.00013702, -0.00001169, 0.00006515, -0.00013204, -0.00000903, 0.00007621, -0.00015112, -0.00018626, 0.00013151, 0.00019683, -0.00014232, -0.00002750, 0.00020035, -0.00017066, -0.00000019, 0.00024426, -0.00050948, },
		{ 0.00024336, -0.00121182, 0.00041166, 0.00212270, -0.00002271, -0.00124772, 0.00009808, 0.00008494, -0.00037182, 0.00022046, 0.00009818, -0.00020987, 0.00008911, -0.00000266, -0.00011789, 0.00002705, -0.00008231, -0.00002956, 0.00013302, -0.00006629, -0.00020945, -0.00000045, 0.00020378, 0.00001235, -0.00015692, 0.00013586, 0.00005783, -0.00018804, 0.00021579, -0.00007467, -0.00054656, 0.00047088, },
		{ -0.00085470, 0.00150389, 0.00137694, -0.00097151, -0.00066600, 0.00031143, -0.00024596, -0.00011487, 0.00027217, -0.00011554, -0.00007067, 0.00008899, -0.00009597, -0.00004718, -0.00008144, -0.00008824, 0.00019055, 0.00007314, -0.00022431, -0.00009047, 0.00015018, 0.00011100, -0.00012262, -0.00003502, 0.00017170, -0.00010372, -0.00002630, 0.00021224, -0.00042229, -0.00016836, 0.00069188, -0.00052263, },
		{ 0.00185221, 0.00019267, -0.00113398, -0.00003987, 0.00011280, -0.00030955, 0.00014983, 0.00010748, -0.00015042, 0.00004954, -0.00001160, -0.00007779, -0.00008238, -0.00019424, 0.00013475, 0.00029461, -0.00015435, -0.00021840, 0.00010571, 0.00014799, -0.00004085, -0.00012143, 0.00009534, 0.00007382, -0.00013787, 0.00015204, -0.00003669, -0.00049603, 0.00035672, 0.00036064, -0.00101606, 0.00012167, },
		{ -0.00071991, -0.00069251, 0.00022077, -0.00016237, -0.00012063, 0.00020930, -0.00006019, -0.00004755, 0.00003706, -0.00006479, -0.00005066, -0.00024675, -0.00009424, 0.00043534, 0.00012774, -0.00033972, -0.00001872, 0.00020533, 0.00001480, -0.00012028, -0.00002274, 0.00012825, -0.00004705, -0.00002884, 0.00016485, -0.00033243, -0.00020204, 0.00062157, -0.00029913, -0.00084349, 0.00111326, 0.00132208, },
		{ -0.00014340, 0.00010930, -0.00023653, 0.00008908, 0.00009035, -0.00008076, 0.00002618, -0.00004303, -0.00002498, -0.00016522, -0.00035418, 0.00027629, 0.00053022, -0.00022406, -0.00028525, 0.00022030, 0.00011939, -0.00011950, -0.00008244, 0.00006679, 0.00006467, -0.00007799, 0.00010056, -0.00002023, -0.00041671, 0.00025228, 0.00040857, -0.00079885, -0.00008766, 0.00152295, 0.00033436, -0.00096416, },
		{ -0.00009951, -0.00010366, 0.00013624, -0.00002317, -0.00000121, -0.00001785, -0.00005156, -0.00001989, -0.00041101, -0.00016218, 0.00070137, 0.00025702, -0.00048324, 0.00000782, 0.00029321, -0.00007744, -0.00014028, 0.00000850, 0.00007772, -0.00000923, -0.00000887, 0.00010955, -0.00025023, -0.00019565, 0.00051894, -0.00012971, -0.00076189, 0.00075944, 0.00117526, -0.00048656, -0.00065179, 0.00012616, },
		{ 0.00004561, 0.00004894, -0.00001095, 0.00002418, -0.00009489, 0.00003099, -0.00020763, -0.00054963, 0.00036919, 0.00082930, -0.00025005, -0.00041168, 0.00032273, 0.00012661, -0.00021591, -0.00004780, 0.00005638, 0.00003223, -0.00001806, 0.00006759, -0.00002507, -0.00031918, 0.00017223, 0.00039169, -0.00058896, -0.00020977, 0.00117028, 0.00039265, -0.00076098, -0.00020101, 0.00007925, -0.00016195, },
		},
		{ { -0.00007011, -0.00010308, 0.00018200, -0.00033192, -0.00058520, 0.00099656, 0.00058362, -0.00077940, 0.00016190, 0.00021047, -0.00046527, 0.00016558, 0.00018779, -0.00016839, 0.00005571, 0.00003175, -0.00012782, 0.00003175, 0.00005571, -0.00016839, 0.00018779, 0.00016558, -0.00046527, 0.00021047, 0.00016190, -0.00077940, 0.00058362, 0.00099656, -0.00058520, -0.00033192, 0.00018200, -0.00010308, },
		{ -0.00004001, 0.00018952, -0.00070148, -0.00012638, 0.00137029, -0.00004464, -0.00070742, 0.00044936, -0.00012652, -0.00035702, 0.00033639, 0.00000623, -0.00013036, 0.00011175, -0.00005220, -0.00010010, 0.00008691, -0.00003901, -0.00010575, 0.00029787, -0.00009154, -0.00036175, 0.00041523, -0.00022793, -0.00051628, 0.00096248, 0.00040024, -0.00067153, -0.00002788, 0.00007060, -0.00006645, -0.00011628, },
		{ -0.00002523, -0.00090911, 0.00061172, 0.00130573, -0.00065946, -0.00033595, 0.00048339, -0.00042630, -0.00008801, 0.00034339, -0.00014976, -0.00002119, 0.00009590, -0.00012077, -0.00003209, 0.00008494, -0.00011612, 0.00003135, 0.00027282, -0.00031130, -0.00010312, 0.00039196, -0.00052062, -0.00004845, 0.00096421, -0.00014182, -0.00048156, 0.00009583, -0.00000635, -0.00002513, -0.00020256, 0.00012998, },
		{ -0.00073447, 0.00135704, 0.00077268, -0.00099066, 0.00014712, 0.00024419, -0.00053526, 0.00020931, 0.00019500, -0.00020243, 0.00009216, 0.00001554, -0.00013890, 0.00003947, 0.00002984, -0.00012756, 0.00017349, 0.00011778, -0.00038863, 0.00017822, 0.00016620, -0.00057473, 0.00040179, 0.00064643, -0.00043318, -0.00021178, 0.00007075, 0.00000947, -0.00007023, -0.00021427, 0.00029481, -0.00045285, },
		{ 0.00177550, -0.00003850, -0.00089878, 0.00049804, -0.00014451, -0.00039761, 0.00039127, -0.00002165, -0.00013404, 0.00014331, -0.00008699, -0.00009980, 0.00007738, -0.00004404, -0.00006104, 0.00024488, -0.00009487, -0.00029143, 0.00034563, -0.00013437, -0.00037848, 0.00063894, 0.00020314, -0.00043983, -0.00003147, 0.00001490, 0.00006731, -0.00020541, -0.00007138, 0.00028204, -0.00091551, -0.00010632, },
		{ -0.00079723, -0.00044844, 0.00054469, -0.00048104, -0.00008451, 0.00037840, -0.00018996, 0.00000648, 0.00010273, -0.00015913, -0.00002741, 0.00006507, -0.00009133, 0.00005200, 0.00020450, -0.00026333, -0.00007507, 0.00032941, -0.00035349, -0.00004816, 0.00059136, -0.00015019, -0.00027462, 0.00000651, 0.00002841, 0.00006198, -0.00033216, 0.00018793, -0.00002131, -0.00113272, 0.00081857, 0.00163839, },
		{ 0.00012433, 0.00027581, -0.00059343, 0.00024558, 0.00019483, -0.00022749, 0.00013342, -0.00000797, -0.00016738, 0.00003864, 0.00001561, -0.00008164, 0.00015272, 0.00006858, -0.00030958, 0.00015009, 0.00015861, -0.00039157, 0.00024253, 0.00033508, -0.00028958, -0.00010378, -0.00003113, 0.00011658, -0.00007062, -0.00031729, 0.00040098, -0.00056236, -0.00086898, 0.00169547, 0.00095083, -0.00119058, },
		{ -0.00015655, -0.00042962, 0.00043271, -0.00004948, -0.00012836, 0.00017310, -0.00013063, -0.00011424, 0.00006647, -0.00003676, -0.00001724, 0.00018681, -0.00009777, -0.00021867, 0.00027880, -0.00006238, -0.00025486, 0.00035985, 0.00003630, -0.00023155, -0.00003784, -0.00003249, 0.00019020, -0.00028668, -0.00009937, 0.00036778, -0.00110416, -0.00008450, 0.00214243, -0.00003231, -0.00107697, 0.00053430, },
		{ -0.00008122, 0.00039950, -0.00022215, 0.00003805, 0.00010354, -0.00020457, -0.00003381, 0.00004969, -0.00005733, 0.00006776, 0.00013415, -0.00021510, -0.00004332, 0.00026479, -0.00021671, -0.00004273, 0.00027843, -0.00014986, -0.00009664, -0.00007348, 0.00006297, 0.00013940, -0.00044694, 0.00023946, -0.00001400, -0.00132266, 0.00100106, 0.00192841, -0.00091804, -0.00055384, 0.00059316, -0.00052166, },
		{ 0.00018732, -0.00024107, 0.00016990, -0.00003654, -0.00019980, 0.00003032, 0.00000930, -0.00003171, 0.00012640, 0.00001973, -0.00023105, 0.00012755, 0.00013905, -0.00024015, 0.00011648, 0.00008360, -0.00016351, -0.00001693, -0.00011705, 0.00021209, -0.00007131, -0.00040598, 0.00049284, -0.00064898, -0.00097730, 0.00197956, 0.00109943, -0.00136037, 0.00009666, 0.00030135, -0.00063421, 0.00027069, },
		{ -0.00011554, 0.00019132, -0.00017313, -0.00013056, 0.00005110, -0.00002098, 0.00002551, 0.00012524, -0.00010001, -0.00014592, 0.00021742, -0.00001541, -0.00015139, 0.00014157, -0.00008938, -0.00006002, -0.00004810, -0.00006802, 0.00029483, -0.00035506, -0.00012309, 0.00044163, -0.00124977, -0.00006379, 0.00243516, -0.00002841, -0.00122385, 0.00055468, -0.00016233, -0.00044938, 0.00045573, -0.00007372, },
		{ 0.00009298, -0.00024038, -0.00004270, 0.00003236, -0.00001665, 0.00007945, 0.00006318, -0.00016760, -0.00000902, 0.00020045, -0.00011548, -0.00003214, 0.00004179, -0.00014026, 0.00004118, -0.00014019, 0.00009601, 0.00020246, -0.00054002, 0.00028034, -0.00000169, -0.00146183, 0.00114128, 0.00214610, -0.00101096, -0.00064053, 0.00062309, -0.00054408, -0.00007830, 0.00040342, -0.00024156, 0.00006559, },
		{ -0.00021640, 0.00001646, 0.00000323, 0.00002166, 0.00009569, -0.00002801, -0.00015459, 0.00011071, 0.00010879, -0.00012514, 0.00002852, -0.00009646, -0.00006109, 0.00004316, -0.00018232, 0.00029104, -0.00007237, -0.00047538, 0.00056405, -0.00070357, -0.00104997, 0.00218130, 0.00120191, -0.00148280, 0.00006750, 0.00031699, -0.00065242, 0.00028113, 0.00017287, -0.00024072, 0.00018977, -0.00006723, },
		{ 0.00002656, -0.00000311, 0.00006838, 0.00006101, -0.00010144, -0.00007468, 0.00016244, 0.00000701, -0.00007076, -0.00000941, -0.00016909, 0.00006703, -0.00006239, -0.00008959, 0.00037572, -0.00040681, -0.00014175, 0.00049936, -0.00133883, -0.00004665, 0.00262394, -0.00002879, -0.00132336, 0.00055648, -0.00016171, -0.00045360, 0.00045581, -0.00009061, -0.00009837, 0.00018656, -0.00020198, -0.00013106, },
		{ 0.00002723, 0.00008890, -0.00000817, -0.00012098, 0.00002653, 0.00013799, -0.00004932, -0.00001739, -0.00011380, -0.00012199, 0.00013348, -0.00019092, 0.00012617, 0.00024788, -0.00060650, 0.00030755, 0.00001640, -0.00153781, 0.00122552, 0.00226811, -0.00106709, -0.00069845, 0.00062978, -0.00054494, -0.00007584, 0.00038746, -0.00024357, 0.00008046, 0.00006518, -0.00024664, -0.00004285, 0.00000391, },
		{ 0.00006192, -0.00007470, -0.00008047, 0.00009850, 0.00007037, -0.00004565, -0.00002248, -0.00020381, 0.00001494, 0.00007486, -0.00022451, 0.00035004, -0.00007386, -0.00052218, 0.00061032, -0.00072050, -0.00108070, 0.00228064, 0.00124600, -0.00154451, 0.00003994, 0.00031968, -0.00064396, 0.00027407, 0.00015217, -0.00022441, 0.00018098, -0.00009677, -0.00019473, -0.00000036, -0.00001276, 0.00007828, },
		{ -0.00010193, -0.00000564, 0.00011331, 0.00000889, -0.00001247, -0.00009641, -0.00020430, 0.00014791, -0.00007978, -0.00009673, 0.00042964, -0.00043980, -0.00015477, 0.00053811, -0.00136366, -0.00003475, 0.00268915, -0.00003475, -0.00136366, 0.00053811, -0.00015477, -0.00043980, 0.00042964, -0.00009673, -0.00007978, 0.00014791, -0.00020430, -0.00009641, -0.00001247, 0.00000889, 0.00011331, -0.00000564, },
		{ 0.00006192, 0.00007828, -0.00001276, -0.00000036, -0.00019473, -0.00009677, 0.00018098, -0.00022441, 0.00015217, 0.00027407, -0.00064396, 0.00031968, 0.00003994, -0.00154451, 0.00124600, 0.00228064, -0.00108070, -0.00072050, 0.00061032, -0.00052218, -0.00007386, 0.00035004, -0.00022451, 0.00007486, 0.00001494, -0.00020381, -0.00002248, -0.00004565, 0.00007037, 0.00009850, -0.00008047, -0.00007470, },
		{ 0.00002723, 0.00000391, -0.00004285, -0.00024664, 0.00006518, 0.00008046, -0.00024357, 0.00038746, -0.00007584, -0.00054494, 0.00062978, -0.00069845, -0.00106709, 0.00226811, 0.00122552, -0.00153781, 0.00001640, 0.00030755, -0.00060650, 0.00024788, 0.00012617, -0.00019092, 0.00013348, -0.00012199, -0.00011380, -0.00001739, -0.00004932, 0.00013799, 0.00002653, -0.00012098, -0.00000817, 0.00008890, },
		{ 0.00002656, -0.00013106, -0.00020198, 0.00018656, -0.00009837, -0.00009061, 0.00045581, -0.00045360, -0.00016171, 0.00055648, -0.00132336, -0.00002879, 0.00262394, -0.00004665, -0.00133883, 0.00049936, -0.00014175, -0.00040681, 0.00037572, -0.00008959, -0.00006239, 0.00006703, -0.00016909, -0.00000941, -0.00007076, 0.00000701, 0.00016244, -0.00007468, -0.00010144, 0.00006101, 0.00006838, -0.00000311, },
		{ -0.00021640, -0.00006723, 0.00018977, -0.00024072, 0.00017287, 0.00028113, -0.00065242, 0.00031699, 0.00006750, -0.00148280, 0.00120191, 0.00218130, -0.00104997, -0.00070357, 0.00056405, -0.00047538, -0.00007237, 0.00029104, -0.00018232, 0.00004316, -0.00006109, -0.00009646, 0.00002852, -0.00012514, 0.00010879, 0.00011071, -0.00015459, -0.00002801, 0.00009569, 0.00002166, 0.00000323, 0.00001646, },
		{ 0.00009298, 0.00006559, -0.00024156, 0.00040342, -0.00007830, -0.00054408, 0.00062309, -0.00064053, -0.00101096, 0.00214610, 0.00114128, -0.00146183, -0.00000169, 0.00028034, -0.00054002, 0.00020246, 0.00009601, -0.00014019, 0.00004118, -0.00014026, 0.00004179, -0.00003214, -0.00011548, 0.00020045, -0.00000902, -0.00016760, 0.00006318, 0.00007945, -0.00001665, 0.00003236, -0.00004270, -0.00024038, },
		{ -0.00011554, -0.00007372, 0.00045573, -0.00044938, -0.00016233, 0.00055468, -0.00122385, -0.00002841, 0.00243516, -0.00006379, -0.00124977, 0.00044163, -0.00012309, -0.00035506, 0.00029483, -0.00006802, -0.00004810, -0.00006002, -0.00008938, 0.00014157, -0.00015139, -0.00001541, 0.00021742, -0.00014592, -0.00010001, 0.00012524, 0.00002551, -0.00002098, 0.00005110, -0.00013056, -0.00017313, 0.00019132, },
		{ 0.00018732, 0.00027069, -0.00063421, 0.00030135, 0.00009666, -0.00136037, 0.00109943, 0.00197956, -0.00097730, -0.00064898, 0.00049284, -0.00040598, -0.00007131, 0.00021209, -0.00011705, -0.00001693, -0.00016351, 0.00008360, 0.00011648, -0.00024015, 0.00013905, 0.00012755, -0.00023105, 0.00001973, 0.00012640, -0.00003171, 0.00000930, 0.00003032, -0.00019980, -0.00003654, 0.00016990, -0.00024107, },
		{ -0.00008122, -0.00052166, 0.00059316, -0.00055384, -0.00091804, 0.00192841, 0.00100106, -0.00132266, -0.00001400, 0.00023946, -0.00044694, 0.00013940, 0.00006297, -0.00007348, -0.00009664, -0.00014986, 0.00027843, -0.00004273, -0.00021671, 0.00026479, -0.00004332, -0.00021510, 0.00013415, 0.00006776, -0.00005733, 0.00004969, -0.00003381, -0.00020457, 0.00010354, 0.00003805, -0.00022215, 0.00039950, },
		{ -0.00015655, 0.00053430, -0.00107697, -0.00003231, 0.00214243, -0.00008450, -0.00110416, 0.00036778, -0.00009937, -0.00028668, 0.00019020, -0.00003249, -0.00003784, -0.00023155, 0.00003630, 0.00035985, -0.00025486, -0.00006238, 0.00027880, -0.00021867, -0.00009777, 0.00018681, -0.00001724, -0.00003676, 0.00006647, -0.00011424, -0.00013063, 0.00017310, -0.00012836, -0.00004948, 0.00043271, -0.00042962, },
		{ 0.00012433, -0.00119058, 0.00095083, 0.00169547, -0.00086898, -0.00056236, 0.00040098, -0.00031729, -0.00007062, 0.00011658, -0.00003113, -0.00010378, -0.00028958, 0.00033508, 0.00024253, -0.00039157, 0.00015861, 0.00015009, -0.00030958, 0.00006858, 0.00015272, -0.00008164, 0.00001561, 0.00003864, -0.00016738, -0.00000797, 0.00013342, -0.00022749, 0.00019483, 0.00024558, -0.00059343, 0.00027581, },
		{ -0.00079723, 0.00163839, 0.00081857, -0.00113272, -0.00002131, 0.00018793, -0.00033216, 0.00006198, 0.00002841, 0.00000651, -0.00027462, -0.00015019, 0.00059136, -0.00004816, -0.00035349, 0.00032941, -0.00007507, -0.00026333, 0.00020450, 0.00005200, -0.00009133, 0.00006507, -0.00002741, -0.00015913, 0.00010273, 0.00000648, -0.00018996, 0.00037840, -0.00008451, -0.00048104, 0.00054469, -0.00044844, },
		{ 0.00177550, -0.00010632, -0.00091551, 0.00028204, -0.00007138, -0.00020541, 0.00006731, 0.00001490, -0.00003147, -0.00043983, 0.00020314, 0.00063894, -0.00037848, -0.00013437, 0.00034563, -0.00029143, -0.00009487, 0.00024488, -0.00006104, -0.00004404, 0.00007738, -0.00009980, -0.00008699, 0.00014331, -0.00013404, -0.00002165, 0.00039127, -0.00039761, -0.00014451, 0.00049804, -0.00089878, -0.00003850, },
		{ -0.00073447, -0.00045285, 0.00029481, -0.00021427, -0.00007023, 0.00000947, 0.00007075, -0.00021178, -0.00043318, 0.00064643, 0.00040179, -0.00057473, 0.00016620, 0.00017822, -0.00038863, 0.00011778, 0.00017349, -0.00012756, 0.00002984, 0.00003947, -0.00013890, 0.00001554, 0.00009216, -0.00020243, 0.00019500, 0.00020931, -0.00053526, 0.00024419, 0.00014712, -0.00099066, 0.00077268, 0.00135704, },
		{ -0.00002523, 0.00012998, -0.00020256, -0.00002513, -0.00000635, 0.00009583, -0.00048156, -0.00014182, 0.00096421, -0.00004845, -0.00052062, 0.00039196, -0.00010312, -0.00031130, 0.00027282, 0.00003135, -0.00011612, 0.00008494, -0.00003209, -0.00012077, 0.00009590, -0.00002119, -0.00014976, 0.00034339, -0.00008801, -0.00042630, 0.00048339, -0.00033595, -0.00065946, 0.00130573, 0.00061172, -0.00090911, },
		{ -0.00004001, -0.00011628, -0.00006645, 0.00007060, -0.00002788, -0.00067153, 0.00040024, 0.00096248, -0.00051628, -0.00022793, 0.00041523, -0.00036175, -0.00009154, 0.00029787, -0.00010575, -0.00003901, 0.00008691, -0.00010010, -0.00005220, 0.00011175, -0.00013036, 0.00000623, 0.00033639, -0.00035702, -0.00012652, 0.00044936, -0.00070742, -0.00004464, 0.00137029, -0.00012638, -0.00070148, 0.00018952, },
		},
		{ { -0.00008141, 0.00002588, -0.00078181, 0.00035295, 0.00138836, -0.00077076, -0.00025071, 0.00047796, -0.00035189, 0.00015316, 0.00000986, -0.00017244, 0.00005564, -0.00005000, -0.00000235, 0.00000483, -0.00000965, 0.00000483, -0.00000235, -0.00005000, 0.00005564, -0.00017244, 0.00000986, 0.00015316, -0.00035189, 0.00047796, -0.00025071, -0.00077076, 0.00138836, 0.00035295, -0.00078181, 0.00002588, },
		{ -0.00010262, -0.00088845, 0.00098385, 0.00106952, -0.00106675, 0.00014319, 0.00027042, -0.00034302, 0.00026759, -0.00012704, -0.00013299, 0.00005553, -0.00006716, 0.00000336, 0.00001461, -0.00001696, 0.00000139, -0.00000230, -0.00003182, 0.00003947, -0.00017086, 0.00011845, -0.00001060, -0.00023588, 0.00054682, -0.00060584, -0.00029852, 0.00140304, -0.00017959, -0.00053369, 0.00001940, -0.00003622, },
		{ -0.00074863, 0.00155446, 0.00050390, -0.00110857, 0.00046217, -0.00000588, -0.00021342, 0.00029151, -0.00025165, -0.00006449, 0.00003428, -0.00007759, 0.00001768, 0.00002081, -0.00003502, 0.00001074, -0.00000216, -0.00001583, 0.00001680, -0.00013013, 0.00016763, -0.00016739, -0.00003191, 0.00045061, -0.00081896, 0.00022290, 0.00114206, -0.00050511, -0.00026934, -0.00005816, 0.00006974, -0.00037471, },
		{ 0.00190195, -0.00017853, -0.00089142, 0.00061787, -0.00026049, -0.00000699, 0.00021242, -0.00032704, 0.00001112, -0.00000564, -0.00007521, 0.00003870, 0.00001430, -0.00005370, 0.00003293, -0.00000739, -0.00000187, -0.00000169, -0.00006526, 0.00014593, -0.00026416, 0.00019430, 0.00021517, -0.00082978, 0.00065638, 0.00070637, -0.00059175, -0.00008933, -0.00012173, 0.00016022, -0.00072921, -0.00032016, },
		{ -0.00081146, -0.00048475, 0.00057459, -0.00041202, 0.00020706, 0.00005061, -0.00033190, 0.00006936, -0.00005449, -0.00005625, 0.00005981, -0.00000911, -0.00006057, 0.00006133, -0.00002031, 0.00001285, -0.00000855, 0.00000092, 0.00006506, -0.00026843, 0.00036982, -0.00008984, -0.00063964, 0.00089626, 0.00023425, -0.00048493, -0.00003763, -0.00010527, 0.00014552, -0.00104711, 0.00035160, 0.00190954, },
		{ -0.00001019, 0.00035930, -0.00041451, 0.00035689, -0.00014745, -0.00026727, 0.00009148, -0.00009801, -0.00002141, 0.00007169, -0.00004626, -0.00004593, 0.00008487, -0.00003842, 0.00003186, -0.00000236, 0.00004558, -0.00004417, -0.00017862, 0.00043842, -0.00037366, -0.00031022, 0.00089899, -0.00014337, -0.00028268, -0.00009299, 0.00000731, -0.00003620, -0.00118740, 0.00113643, 0.00154525, -0.00124063, },
		{ 0.00005135, -0.00027190, 0.00039157, -0.00032427, -0.00015588, 0.00007085, -0.00012191, 0.00002325, 0.00006570, -0.00008687, -0.00000705, 0.00009227, -0.00005463, 0.00005704, 0.00001187, 0.00005382, -0.00014221, -0.00002418, 0.00037867, -0.00055390, 0.00005706, 0.00069329, -0.00034263, -0.00009549, -0.00018546, 0.00017233, -0.00038505, -0.00103550, 0.00185049, 0.00087585, -0.00136874, 0.00039656, },
		{ -0.00003588, 0.00029755, -0.00042989, -0.00003426, 0.00001540, -0.00011662, 0.00006731, 0.00003749, -0.00011653, 0.00004993, 0.00007655, -0.00005955, 0.00008683, 0.00002468, 0.00002378, -0.00019399, 0.00014454, 0.00021045, -0.00058301, 0.00035409, 0.00036697, -0.00034860, -0.00000534, -0.00023093, 0.00030171, -0.00082662, -0.00054801, 0.00230730, 0.00005131, -0.00118309, 0.00062743, -0.00024465, },
		{ 0.00010272, -0.00043781, 0.00005954, -0.00005503, -0.00008086, 0.00009866, -0.00001053, -0.00012151, 0.00011119, 0.00003838, -0.00004500, 0.00011573, 0.00002600, -0.00003266, -0.00018051, 0.00027316, -0.00001237, -0.00046289, 0.00050052, 0.00003550, -0.00021298, -0.00003801, -0.00017026, 0.00029814, -0.00122655, 0.00022207, 0.00237295, -0.00073646, -0.00075717, 0.00063384, -0.00043353, 0.00021204, },
		{ -0.00035293, 0.00009858, -0.00011520, -0.00002256, 0.00010698, -0.00006887, -0.00009372, 0.00015931, -0.00001323, -0.00000751, 0.00013547, 0.00000928, -0.00009395, -0.00010500, 0.00032125, -0.00022035, -0.00024301, 0.00046843, -0.00019705, -0.00002945, -0.00015799, 0.00000197, 0.00009747, -0.00142936, 0.00112968, 0.00200807, -0.00130758, -0.00022597, 0.00043879, -0.00046018, 0.00038817, -0.00013324, },
		{ 0.00007574, -0.00014301, 0.00004307, 0.00008713, -0.00012297, -0.00003417, 0.00017887, -0.00006308, 0.00004940, 0.00013750, -0.00002540, -0.00013626, 0.00000840, 0.00027538, -0.00035179, -0.00000280, 0.00028905, -0.00027129, 0.00010388, -0.00028709, 0.00023014, -0.00029757, -0.00130959, 0.00197194, 0.00128284, -0.00154460, 0.00025659, 0.00012562, -0.00032618, 0.00043450, -0.00033889, -0.00020902, },
		{ -0.00012692, 0.00009791, 0.00004116, -0.00015715, 0.00004633, 0.00016175, -0.00009436, 0.00011518, 0.00011609, -0.00007069, -0.00014174, 0.00012287, 0.00015295, -0.00037312, 0.00017559, 0.00003981, -0.00018933, 0.00012151, -0.00033927, 0.00041516, -0.00080563, -0.00081835, 0.00254302, 0.00035991, -0.00142348, 0.00056445, -0.00019224, -0.00008704, 0.00033766, -0.00045465, -0.00005669, 0.00000561, },
		{ 0.00012632, -0.00002165, -0.00015929, 0.00012969, 0.00011033, -0.00009408, 0.00017453, 0.00007090, -0.00011377, -0.00010478, 0.00020142, -0.00000436, -0.00028936, 0.00023504, -0.00018362, -0.00000981, 0.00001395, -0.00025853, 0.00045108, -0.00128283, -0.00000911, 0.00269302, -0.00055020, -0.00101612, 0.00063533, -0.00040987, 0.00016862, 0.00013249, -0.00045212, 0.00005404, -0.00008097, -0.00006997, },
		{ -0.00008696, -0.00012474, 0.00019551, 0.00003739, -0.00005768, 0.00021161, 0.00000804, -0.00014040, -0.00003425, 0.00021921, -0.00014525, -0.00014137, 0.00016216, -0.00030066, 0.00017648, -0.00017049, -0.00004584, 0.00026809, -0.00156331, 0.00096715, 0.00237257, -0.00124740, -0.00046428, 0.00048481, -0.00046504, 0.00035061, -0.00011058, -0.00034266, 0.00009052, -0.00014792, 0.00001105, 0.00011991, },
		{ -0.00005820, 0.00022703, -0.00003729, 0.00000892, 0.00021467, -0.00006072, -0.00013970, 0.00004922, 0.00017174, -0.00022645, 0.00000906, -0.00000906, -0.00027311, 0.00028148, -0.00034579, 0.00023558, -0.00013389, -0.00151175, 0.00189772, 0.00164904, -0.00159643, 0.00006630, 0.00019485, -0.00035636, 0.00039849, -0.00031024, -0.00017245, 0.00004814, -0.00016653, 0.00009108, 0.00008003, -0.00013906, },
		{ 0.00021619, -0.00009278, 0.00009039, 0.00017981, -0.00012080, -0.00010814, 0.00011995, 0.00007607, -0.00022734, 0.00010123, -0.00020991, -0.00011715, 0.00025281, -0.00042289, 0.00047903, -0.00067383, -0.00107096, 0.00256502, 0.00068894, -0.00156072, 0.00043622, -0.00011843, -0.00014003, 0.00030200, -0.00040226, -0.00000590, -0.00004890, -0.00012557, 0.00014495, 0.00001738, -0.00016511, 0.00002706, },
		{ -0.00011323, 0.00016614, 0.00011262, -0.00015854, -0.00005127, 0.00015566, -0.00003527, -0.00015666, 0.00009696, -0.00036144, 0.00010367, 0.00009325, -0.00034615, 0.00057146, -0.00120603, -0.00028829, 0.00280732, -0.00028829, -0.00120603, 0.00057146, -0.00034615, 0.00009325, 0.00010367, -0.00036144, 0.00009696, -0.00015666, -0.00003527, 0.00015566, -0.00005127, -0.00015854, 0.00011262, 0.00016614, },
		{ 0.00021619, 0.00002706, -0.00016511, 0.00001738, 0.00014495, -0.00012557, -0.00004890, -0.00000590, -0.00040226, 0.00030200, -0.00014003, -0.00011843, 0.00043622, -0.00156072, 0.00068894, 0.00256502, -0.00107096, -0.00067383, 0.00047903, -0.00042289, 0.00025281, -0.00011715, -0.00020991, 0.00010123, -0.00022734, 0.00007607, 0.00011995, -0.00010814, -0.00012080, 0.00017981, 0.00009039, -0.00009278, },
		{ -0.00005820, -0.00013906, 0.00008003, 0.00009108, -0.00016653, 0.00004814, -0.00017245, -0.00031024, 0.00039849, -0.00035636, 0.00019485, 0.00006630, -0.00159643, 0.00164904, 0.00189772, -0.00151175, -0.00013389, 0.00023558, -0.00034579, 0.00028148, -0.00027311, -0.00000906, 0.00000906, -0.00022645, 0.00017174, 0.00004922, -0.00013970, -0.00006072, 0.00021467, 0.00000892, -0.00003729, 0.00022703, },
		{ -0.00008696, 0.00011991, 0.00001105, -0.00014792, 0.00009052, -0.00034266, -0.00011058, 0.00035061, -0.00046504, 0.00048481, -0.00046428, -0.00124740, 0.00237257, 0.00096715, -0.00156331, 0.00026809, -0.00004584, -0.00017049, 0.00017648, -0.00030066, 0.00016216, -0.00014137, -0.00014525, 0.00021921, -0.00003425, -0.00014040, 0.00000804, 0.00021161, -0.00005768, 0.00003739, 0.00019551, -0.00012474, },
		{ 0.00012632, -0.00006997, -0.00008097, 0.00005404, -0.00045212, 0.00013249, 0.00016862, -0.00040987, 0.00063533, -0.00101612, -0.00055020, 0.00269302, -0.00000911, -0.00128283, 0.00045108, -0.00025853, 0.00001395, -0.00000981, -0.00018362, 0.00023504, -0.00028936, -0.00000436, 0.00020142, -0.00010478, -0.00011377, 0.00007090, 0.00017453, -0.00009408, 0.00011033, 0.00012969, -0.00015929, -0.00002165, },
		{ -0.00012692, 0.00000561, -0.00005669, -0.00045465, 0.00033766, -0.00008704, -0.00019224, 0.00056445, -0.00142348, 0.00035991, 0.00254302, -0.00081835, -0.00080563, 0.00041516, -0.00033927, 0.00012151, -0.00018933, 0.00003981, 0.00017559, -0.00037312, 0.00015295, 0.00012287, -0.00014174, -0.00007069, 0.00011609, 0.00011518, -0.00009436, 0.00016175, 0.00004633, -0.00015715, 0.00004116, 0.00009791, },
		{ 0.00007574, -0.00020902, -0.00033889, 0.00043450, -0.00032618, 0.00012562, 0.00025659, -0.00154460, 0.00128284, 0.00197194, -0.00130959, -0.00029757, 0.00023014, -0.00028709, 0.00010388, -0.00027129, 0.00028905, -0.00000280, -0.00035179, 0.00027538, 0.00000840, -0.00013626, -0.00002540, 0.00013750, 0.00004940, -0.00006308, 0.00017887, -0.00003417, -0.00012297, 0.00008713, 0.00004307, -0.00014301, },
		{ -0.00035293, -0.00013324, 0.00038817, -0.00046018, 0.00043879, -0.00022597, -0.00130758, 0.00200807, 0.00112968, -0.00142936, 0.00009747, 0.00000197, -0.00015799, -0.00002945, -0.00019705, 0.00046843, -0.00024301, -0.00022035, 0.00032125, -0.00010500, -0.00009395, 0.00000928, 0.00013547, -0.00000751, -0.00001323, 0.00015931, -0.00009372, -0.00006887, 0.00010698, -0.00002256, -0.00011520, 0.00009858, },
		{ 0.00010272, 0.00021204, -0.00043353, 0.00063384, -0.00075717, -0.00073646, 0.00237295, 0.00022207, -0.00122655, 0.00029814, -0.00017026, -0.00003801, -0.00021298, 0.00003550, 0.00050052, -0.00046289, -0.00001237, 0.00027316, -0.00018051, -0.00003266, 0.00002600, 0.00011573, -0.00004500, 0.00003838, 0.00011119, -0.00012151, -0.00001053, 0.00009866, -0.00008086, -0.00005503, 0.00005954, -0.00043781, },
		{ -0.00003588, -0.00024465, 0.00062743, -0.00118309, 0.00005131, 0.00230730, -0.00054801, -0.00082662, 0.00030171, -0.00023093, -0.00000534, -0.00034860, 0.00036697, 0.00035409, -0.00058301, 0.00021045, 0.00014454, -0.00019399, 0.00002378, 0.00002468, 0.00008683, -0.00005955, 0.00007655, 0.00004993, -0.00011653, 0.00003749, 0.00006731, -0.00011662, 0.00001540, -0.00003426, -0.00042989, 0.00029755, },
		{ 0.00005135, 0.00039656, -0.00136874, 0.00087585, 0.00185049, -0.00103550, -0.00038505, 0.00017233, -0.00018546, -0.00009549, -0.00034263, 0.00069329, 0.00005706, -0.00055390, 0.00037867, -0.00002418, -0.00014221, 0.00005382, 0.00001187, 0.00005704, -0.00005463, 0.00009227, -0.00000705, -0.00008687, 0.00006570, 0.00002325, -0.00012191, 0.00007085, -0.00015588, -0.00032427, 0.00039157, -0.00027190, },
		{ -0.00001019, -0.00124063, 0.00154525, 0.00113643, -0.00118740, -0.00003620, 0.00000731, -0.00009299, -0.00028268, -0.00014337, 0.00089899, -0.00031022, -0.00037366, 0.00043842, -0.00017862, -0.00004417, 0.00004558, -0.00000236, 0.00003186, -0.00003842, 0.00008487, -0.00004593, -0.00004626, 0.00007169, -0.00002141, -0.00009801, 0.00009148, -0.00026727, -0.00014745, 0.00035689, -0.00041451, 0.00035930, },
		{ -0.00081146, 0.00190954, 0.00035160, -0.00104711, 0.00014552, -0.00010527, -0.00003763, -0.00048493, 0.00023425, 0.00089626, -0.00063964, -0.00008984, 0.00036982, -0.00026843, 0.00006506, 0.00000092, -0.00000855, 0.00001285, -0.00002031, 0.00006133, -0.00006057, -0.00000911, 0.00005981, -0.00005625, -0.00005449, 0.00006936, -0.00033190, 0.00005061, 0.00020706, -0.00041202, 0.00057459, -0.00048475, },
		{ 0.00190195, -0.00032016, -0.00072921, 0.00016022, -0.00012173, -0.00008933, -0.00059175, 0.00070637, 0.00065638, -0.00082978, 0.00021517, 0.00019430, -0.00026416, 0.00014593, -0.00006526, -0.00000169, -0.00000187, -0.00000739, 0.00003293, -0.00005370, 0.00001430, 0.00003870, -0.00007521, -0.00000564, 0.00001112, -0.00032704, 0.00021242, -0.00000699, -0.00026049, 0.00061787, -0.00089142, -0.00017853, },
		{ -0.00074863, -0.00037471, 0.00006974, -0.00005816, -0.00026934, -0.00050511, 0.00114206, 0.00022290, -0.00081896, 0.00045061, -0.00003191, -0.00016739, 0.00016763, -0.00013013, 0.00001680, -0.00001583, -0.00000216, 0.00001074, -0.00003502, 0.00002081, 0.00001768, -0.00007759, 0.00003428, -0.00006449, -0.00025165, 0.00029151, -0.00021342, -0.00000588, 0.00046217, -0.00110857, 0.00050390, 0.00155446, },
		{ -0.00010262, -0.00003622, 0.00001940, -0.00053369, -0.00017959, 0.00140304, -0.00029852, -0.00060584, 0.00054682, -0.00023588, -0.00001060, 0.00011845, -0.00017086, 0.00003947, -0.00003182, -0.00000230, 0.00000139, -0.00001696, 0.00001461, 0.00000336, -0.00006716, 0.00005553, -0.00013299, -0.00012704, 0.00026759, -0.00034302, 0.00027042, 0.00014319, -0.00106675, 0.00106952, 0.00098385, -0.00088845, },
		},
		{ { 0.00004231, -0.00082261, 0.00101457, 0.00064034, -0.00088522, 0.00055371, -0.00041650, -0.00015612, 0.00013724, -0.00013666, 0.00009481, -0.00000702, 0.00001594, 0.00002819, 0.00000845, -0.00007825, 0.00006228, -0.00007825, 0.00000845, 0.00002819, 0.00001594, -0.00000702, 0.00009481, -0.00013666, 0.00013724, -0.00015612, -0.00041650, 0.00055371, -0.00088522, 0.00064034, 0.00101457, -0.00082261, },
		{ -0.00081505, 0.00141529, 0.00023302, -0.00075122, 0.00058189, -0.00056543, -0.00002557, 0.00006631, -0.00010030, 0.00008905, -0.00000534, 0.00002617, 0.00003020, -0.00001507, -0.00005672, 0.00005599, -0.00009109, 0.00003172, 0.00002233, 0.00000278, 0.00000070, 0.00008676, -0.00015561, 0.00019282, -0.00028211, -0.00022969, 0.00043969, -0.00090674, 0.00096676, 0.00056347, -0.00069965, -0.00001906, },
		{ 0.00170748, -0.00020291, -0.00052513, 0.00051823, -0.00065697, 0.00009187, -0.00000898, -0.00005282, 0.00007214, 0.00000573, 0.00003171, 0.00002854, -0.00003579, -0.00003017, 0.00003809, -0.00009283, 0.00005165, 0.00001294, -0.00001107, 0.00001621, 0.00006450, -0.00015324, 0.00022414, -0.00038613, -0.00002977, 0.00025843, -0.00081204, 0.00117165, 0.00012429, -0.00048959, -0.00019108, -0.00065277, },
		{ -0.00061106, -0.00024087, 0.00037027, -0.00067974, 0.00018093, -0.00007724, -0.00000194, 0.00004847, 0.00002450, 0.00003152, 0.00002385, -0.00005121, -0.00000295, 0.00001142, -0.00008271, 0.00006548, 0.00000085, -0.00002324, 0.00003648, 0.00002999, -0.00012867, 0.00022576, -0.00045357, 0.00015681, 0.00003820, -0.00061462, 0.00123193, -0.00024579, -0.00024750, -0.00043946, -0.00033545, 0.00184588, },
		{ 0.00005920, 0.00015805, -0.00063211, 0.00023091, -0.00012839, 0.00004439, 0.00002349, 0.00004789, 0.00002552, 0.00001714, -0.00005969, 0.00002054, -0.00001982, -0.00006176, 0.00007129, -0.00001273, -0.00003161, 0.00005761, -0.00001271, -0.00008424, 0.00019653, -0.00047482, 0.00030586, -0.00018734, -0.00034319, 0.00114519, -0.00050301, -0.00003084, -0.00071323, 0.00011235, 0.00180502, -0.00093917, },
		{ -0.00008887, -0.00052239, 0.00023724, -0.00015521, 0.00007926, 0.00000275, 0.00007188, 0.00001453, 0.00000960, -0.00006068, 0.00003662, -0.00005067, -0.00003265, 0.00006818, -0.00002638, -0.00003462, 0.00007539, -0.00005798, -0.00002521, 0.00013985, -0.00044670, 0.00039893, -0.00038430, -0.00003736, 0.00092987, -0.00062323, 0.00011008, -0.00095298, 0.00064400, 0.00158329, -0.00114665, 0.00033077, },
		{ -0.00036754, 0.00020204, -0.00015441, 0.00009775, -0.00000898, 0.00009219, 0.00000020, 0.00000242, -0.00005481, 0.00004278, -0.00007619, 0.00000069, 0.00005650, -0.00003865, -0.00003157, 0.00008604, -0.00009947, 0.00004099, 0.00006315, -0.00037294, 0.00042594, -0.00052370, 0.00025831, 0.00062223, -0.00060502, 0.00014042, -0.00110064, 0.00119709, 0.00120345, -0.00121043, 0.00053422, -0.00033576, },
		{ 0.00013369, -0.00012709, 0.00009778, -0.00000849, 0.00010488, -0.00001527, -0.00000337, -0.00004370, 0.00003816, -0.00009220, 0.00003365, 0.00003774, -0.00004822, -0.00002268, 0.00008678, -0.00013107, 0.00010585, -0.00002318, -0.00026366, 0.00038662, -0.00058583, 0.00050106, 0.00027067, -0.00046921, 0.00004665, -0.00110909, 0.00170220, 0.00070965, -0.00112812, 0.00064060, -0.00054796, -0.00019053, },
		{ -0.00007855, 0.00008031, 0.00000530, 0.00010705, -0.00002944, -0.00000709, -0.00002976, 0.00002363, -0.00009597, 0.00006163, 0.00001434, -0.00005416, -0.00000909, 0.00007633, -0.00014781, 0.00016097, -0.00010732, -0.00013380, 0.00029034, -0.00056321, 0.00065631, -0.00007167, -0.00025509, -0.00016106, -0.00095033, 0.00209293, 0.00016117, -0.00091812, 0.00063597, -0.00069599, -0.00001691, 0.00004541, },
		{ 0.00004917, 0.00003113, 0.00009729, -0.00004003, -0.00000850, -0.00001574, 0.00000165, -0.00008659, 0.00008071, -0.00001068, -0.00005598, 0.00000729, 0.00005513, -0.00014661, 0.00019934, -0.00017777, -0.00000102, 0.00015458, -0.00046162, 0.00070285, -0.00035499, -0.00001361, -0.00045025, -0.00062074, 0.00231556, -0.00037596, -0.00061658, 0.00052341, -0.00075995, 0.00012904, -0.00004682, -0.00001742, },
		{ 0.00006559, 0.00007594, -0.00004525, -0.00000780, -0.00000431, -0.00002407, -0.00006518, 0.00008824, -0.00003413, -0.00005371, 0.00002411, 0.00002532, -0.00012681, 0.00021630, -0.00022504, 0.00011702, 0.00000217, -0.00029918, 0.00063599, -0.00054033, 0.00020098, -0.00077111, -0.00014271, 0.00233705, -0.00083835, -0.00027171, 0.00032249, -0.00073247, 0.00022792, -0.00012682, 0.00004569, 0.00001048, },
		{ 0.00004507, -0.00004411, -0.00000560, 0.00000238, -0.00004923, -0.00003469, 0.00008325, -0.00005314, -0.00004790, 0.00003887, -0.00000952, -0.00009027, 0.00021023, -0.00024308, 0.00020517, -0.00014229, -0.00010345, 0.00046814, -0.00060534, 0.00034013, -0.00106430, 0.00043772, 0.00215002, -0.00117418, 0.00006358, 0.00006618, -0.00061969, 0.00026808, -0.00018074, 0.00009983, -0.00002839, 0.00010360, },
		{ -0.00003654, -0.00000281, 0.00000302, -0.00006951, 0.00000052, 0.00006655, -0.00006558, -0.00003952, 0.00004935, -0.00004494, -0.00004114, 0.00018277, -0.00023012, 0.00025303, -0.00025622, 0.00009286, 0.00022690, -0.00054745, 0.00036895, -0.00127041, 0.00105679, 0.00177402, -0.00135041, 0.00033984, -0.00020426, -0.00044031, 0.00024722, -0.00019925, 0.00013550, -0.00005995, 0.00013918, 0.00000823, },
		{ -0.00000047, -0.00000262, -0.00008133, 0.00003534, 0.00004065, -0.00007031, -0.00002978, 0.00005393, -0.00007617, 0.00001471, 0.00013851, -0.00018892, 0.00025638, -0.00032252, 0.00025699, -0.00004930, -0.00038399, 0.00027209, -0.00133953, 0.00164186, 0.00125292, -0.00135707, 0.00051848, -0.00044565, -0.00022262, 0.00017263, -0.00017898, 0.00014605, -0.00007785, 0.00016635, -0.00003006, -0.00002343, },
		{ -0.00001368, -0.00008238, 0.00006468, 0.00000937, -0.00006736, -0.00002001, 0.00005184, -0.00009893, 0.00007055, 0.00008432, -0.00012632, 0.00021776, -0.00033214, 0.00036120, -0.00031703, -0.00014920, 0.00005644, -0.00123958, 0.00212156, 0.00064873, -0.00120817, 0.00057752, -0.00061952, -0.00000023, 0.00005990, -0.00012307, 0.00012883, -0.00007784, 0.00018010, -0.00006494, -0.00000651, 0.00000043, },
		{ -0.00007197, 0.00008421, -0.00002277, -0.00005789, -0.00001139, 0.00004325, -0.00010995, 0.00011970, 0.00002825, -0.00005218, 0.00014592, -0.00028547, 0.00038715, -0.00053460, 0.00011130, -0.00024955, -0.00096204, 0.00243603, 0.00003282, -0.00093895, 0.00051476, -0.00069807, 0.00019315, -0.00006961, -0.00004069, 0.00008556, -0.00005854, 0.00017715, -0.00009193, 0.00001192, -0.00000090, -0.00002831, },
		{ 0.00009105, -0.00005120, -0.00004399, -0.00000485, 0.00002931, -0.00010758, 0.00015651, -0.00002164, 0.00002215, 0.00005438, -0.00019218, 0.00032874, -0.00066841, 0.00034786, -0.00060009, -0.00052414, 0.00254550, -0.00052414, -0.00060009, 0.00034786, -0.00066841, 0.00032874, -0.00019218, 0.00005438, 0.00002215, -0.00002164, 0.00015651, -0.00010758, 0.00002931, -0.00000485, -0.00004399, -0.00005120, },
		{ -0.00007197, -0.00002831, -0.00000090, 0.00001192, -0.00009193, 0.00017715, -0.00005854, 0.00008556, -0.00004069, -0.00006961, 0.00019315, -0.00069807, 0.00051476, -0.00093895, 0.00003282, 0.00243603, -0.00096204, -0.00024955, 0.00011130, -0.00053460, 0.00038715, -0.00028547, 0.00014592, -0.00005218, 0.00002825, 0.00011970, -0.00010995, 0.00004325, -0.00001139, -0.00005789, -0.00002277, 0.00008421, },
		{ -0.00001368, 0.00000043, -0.00000651, -0.00006494, 0.00018010, -0.00007784, 0.00012883, -0.00012307, 0.00005990, -0.00000023, -0.00061952, 0.00057752, -0.00120817, 0.00064873, 0.00212156, -0.00123958, 0.00005644, -0.00014920, -0.00031703, 0.00036120, -0.00033214, 0.00021776, -0.00012632, 0.00008432, 0.00007055, -0.00009893, 0.00005184, -0.00002001, -0.00006736, 0.00000937, 0.00006468, -0.00008238, },
		{ -0.00000047, -0.00002343, -0.00003006, 0.00016635, -0.00007785, 0.00014605, -0.00017898, 0.00017263, -0.00022262, -0.00044565, 0.00051848, -0.00135707, 0.00125292, 0.00164186, -0.00133953, 0.00027209, -0.00038399, -0.00004930, 0.00025699, -0.00032252, 0.00025638, -0.00018892, 0.00013851, 0.00001471, -0.00007617, 0.00005393, -0.00002978, -0.00007031, 0.00004065, 0.00003534, -0.00008133, -0.00000262, },
		{ -0.00003654, 0.00000823, 0.00013918, -0.00005995, 0.00013550, -0.00019925, 0.00024722, -0.00044031, -0.00020426, 0.00033984, -0.00135041, 0.00177402, 0.00105679, -0.00127041, 0.00036895, -0.00054745, 0.00022690, 0.00009286, -0.00025622, 0.00025303, -0.00023012, 0.00018277, -0.00004114, -0.00004494, 0.00004935, -0.00003952, -0.00006558, 0.00006655, 0.00000052, -0.00006951, 0.00000302, -0.00000281, },
		{ 0.00004507, 0.00010360, -0.00002839, 0.00009983, -0.00018074, 0.00026808, -0.00061969, 0.00006618, 0.00006358, -0.00117418, 0.00215002, 0.00043772, -0.00106430, 0.00034013, -0.00060534, 0.00046814, -0.00010345, -0.00014229, 0.00020517, -0.00024308, 0.00021023, -0.00009027, -0.00000952, 0.00003887, -0.00004790, -0.00005314, 0.00008325, -0.00003469, -0.00004923, 0.00000238, -0.00000560, -0.00004411, },
		{ 0.00006559, 0.00001048, 0.00004569, -0.00012682, 0.00022792, -0.00073247, 0.00032249, -0.00027171, -0.00083835, 0.00233705, -0.00014271, -0.00077111, 0.00020098, -0.00054033, 0.00063599, -0.00029918, 0.00000217, 0.00011702, -0.00022504, 0.00021630, -0.00012681, 0.00002532, 0.00002411, -0.00005371, -0.00003413, 0.00008824, -0.00006518, -0.00002407, -0.00000431, -0.00000780, -0.00004525, 0.00007594, },
		{ 0.00004917, -0.00001742, -0.00004682, 0.00012904, -0.00075995, 0.00052341, -0.00061658, -0.00037596, 0.00231556, -0.00062074, -0.00045025, -0.00001361, -0.00035499, 0.00070285, -0.00046162, 0.00015458, -0.00000102, -0.00017777, 0.00019934, -0.00014661, 0.00005513, 0.00000729, -0.00005598, -0.00001068, 0.00008071, -0.00008659, 0.00000165, -0.00001574, -0.00000850, -0.00004003, 0.00009729, 0.00003113, },
		{ -0.00007855, 0.00004541, -0.00001691, -0.00069599, 0.00063597, -0.00091812, 0.00016117, 0.00209293, -0.00095033, -0.00016106, -0.00025509, -0.00007167, 0.00065631, -0.00056321, 0.00029034, -0.00013380, -0.00010732, 0.00016097, -0.00014781, 0.00007633, -0.00000909, -0.00005416, 0.00001434, 0.00006163, -0.00009597, 0.00002363, -0.00002976, -0.00000709, -0.00002944, 0.00010705, 0.00000530, 0.00008031, },
		{ 0.00013369, -0.00019053, -0.00054796, 0.00064060, -0.00112812, 0.00070965, 0.00170220, -0.00110909, 0.00004665, -0.00046921, 0.00027067, 0.00050106, -0.00058583, 0.00038662, -0.00026366, -0.00002318, 0.00010585, -0.00013107, 0.00008678, -0.00002268, -0.00004822, 0.00003774, 0.00003365, -0.00009220, 0.00003816, -0.00004370, -0.00000337, -0.00001527, 0.00010488, -0.00000849, 0.00009778, -0.00012709, },
		{ -0.00036754, -0.00033576, 0.00053422, -0.00121043, 0.00120345, 0.00119709, -0.00110064, 0.00014042, -0.00060502, 0.00062223, 0.00025831, -0.00052370, 0.00042594, -0.00037294, 0.00006315, 0.00004099, -0.00009947, 0.00008604, -0.00003157, -0.00003865, 0.00005650, 0.00000069, -0.00007619, 0.00004278, -0.00005481, 0.00000242, 0.00000020, 0.00009219, -0.00000898, 0.00009775, -0.00015441, 0.00020204, },
		{ -0.00008887, 0.00033077, -0.00114665, 0.00158329, 0.00064400, -0.00095298, 0.00011008, -0.00062323, 0.00092987, -0.00003736, -0.00038430, 0.00039893, -0.00044670, 0.00013985, -0.00002521, -0.00005798, 0.00007539, -0.00003462, -0.00002638, 0.00006818, -0.00003265, -0.00005067, 0.00003662, -0.00006068, 0.00000960, 0.00001453, 0.00007188, 0.00000275, 0.00007926, -0.00015521, 0.00023724, -0.00052239, },
		{ 0.00005920, -0.00093917, 0.00180502, 0.00011235, -0.00071323, -0.00003084, -0.00050301, 0.00114519, -0.00034319, -0.00018734, 0.00030586, -0.00047482, 0.00019653, -0.00008424, -0.00001271, 0.00005761, -0.00003161, -0.00001273, 0.00007129, -0.00006176, -0.00001982, 0.00002054, -0.00005969, 0.00001714, 0.00002552, 0.00004789, 0.00002349, 0.00004439, -0.00012839, 0.00023091, -0.00063211, 0.00015805, },
		{ -0.00061106, 0.00184588, -0.00033545, -0.00043946, -0.00024750, -0.00024579, 0.00123193, -0.00061462, 0.00003820, 0.00015681, -0.00045357, 0.00022576, -0.00012867, 0.00002999, 0.00003648, -0.00002324, 0.00000085, 0.00006548, -0.00008271, 0.00001142, -0.00000295, -0.00005121, 0.00002385, 0.00003152, 0.00002450, 0.00004847, -0.00000194, -0.00007724, 0.00018093, -0.00067974, 0.00037027, -0.00024087, },
		{ 0.00170748, -0.00065277, -0.00019108, -0.00048959, 0.00012429, 0.00117165, -0.00081204, 0.00025843, -0.00002977, -0.00038613, 0.00022414, -0.00015324, 0.00006450, 0.00001621, -0.00001107, 0.00001294, 0.00005165, -0.00009283, 0.00003809, -0.00003017, -0.00003579, 0.00002854, 0.00003171, 0.00000573, 0.00007214, -0.00005282, -0.00000898, 0.00009187, -0.00065697, 0.00051823, -0.00052513, -0.00020291, },
		{ -0.00081505, -0.00001906, -0.00069965, 0.00056347, 0.00096676, -0.00090674, 0.00043969, -0.00022969, -0.00028211, 0.00019282, -0.00015561, 0.00008676, 0.00000070, 0.00000278, 0.00002233, 0.00003172, -0.00009109, 0.00005599, -0.00005672, -0.00001507, 0.00003020, 0.00002617, -0.00000534, 0.00008905, -0.00010030, 0.00006631, -0.00002557, -0.00056543, 0.00058189, -0.00075122, 0.00023302, 0.00141529, },
		},
		{ { -0.00143946, 0.00156370, -0.00011188, -0.00082149, 0.00022937, -0.00024761, 0.00040979, -0.00023647, -0.00000662, -0.00007884, 0.00017068, -0.00013165, 0.00005224, -0.00002544, 0.00000255, -0.00000062, -0.00000965, -0.00000062, 0.00000255, -0.00002544, 0.00005224, -0.00013165, 0.00017068, -0.00007884, -0.00000662, -0.00023647, 0.00040979, -0.00024761, 0.00022937, -0.00082149, -0.00011188, 0.00156370, },
		{ 0.00181350, -0.00038079, -0.00069601, 0.00016621, -0.00018608, 0.00039631, -0.00026241, -0.00000662, -0.00005701, 0.00016970, -0.00014553, 0.00006114, -0.00003522, 0.00001305, -0.00000350, -0.00000965, 0.00000181, -0.00000809, -0.00001389, 0.00004141, -0.00011407, 0.00016466, -0.00009632, -0.00000662, -0.00020405, 0.00040761, -0.00029619, 0.00028120, -0.00092785, 0.00015329, 0.00126600, -0.00139966, },
		{ -0.00064310, -0.00055625, 0.00009415, -0.00011395, 0.00036769, -0.00028089, -0.00000662, -0.00003166, 0.00016178, -0.00015517, 0.00006774, -0.00004288, 0.00002299, -0.00000675, -0.00000965, 0.00000367, -0.00001846, -0.00000103, 0.00002904, -0.00009347, 0.00015188, -0.00010876, -0.00000662, -0.00016641, 0.00038986, -0.00032994, 0.00031971, -0.00101102, 0.00040453, 0.00093186, -0.00128177, 0.00200580, },
		{ -0.00040758, 0.00001597, -0.00003400, 0.00032503, -0.00029118, -0.00000662, -0.00000376, 0.00014720, -0.00016021, 0.00007181, -0.00004810, 0.00003199, -0.00001022, -0.00000965, 0.00000491, -0.00002818, 0.00001265, 0.00001563, -0.00007062, 0.00013283, -0.00011570, -0.00000662, -0.00012499, 0.00035721, -0.00034756, 0.00034343, -0.00106779, 0.00063219, 0.00057409, -0.00109033, 0.00213322, -0.00088874, },
		{ -0.00006534, 0.00005070, 0.00026997, -0.00029291, -0.00000662, 0.00002561, 0.00012654, -0.00016045, 0.00007319, -0.00005070, 0.00003972, -0.00001378, -0.00000965, 0.00000547, -0.00003687, 0.00002663, 0.00000168, -0.00004642, 0.00010825, -0.00011686, -0.00000662, -0.00008138, 0.00031092, -0.00034839, 0.00035144, -0.00109599, 0.00082752, 0.00020647, -0.00083270, 0.00219086, -0.00110826, -0.00025570, },
		{ 0.00013690, 0.00020462, -0.00028599, -0.00000662, 0.00005532, 0.00010057, -0.00015587, 0.00007181, -0.00005056, 0.00004587, -0.00001731, -0.00000965, 0.00000533, -0.00004419, 0.00004036, -0.00001227, -0.00002180, 0.00007907, -0.00011220, -0.00000662, -0.00003726, 0.00025277, -0.00033239, 0.00034343, -0.00109454, 0.00098300, -0.00015689, -0.00051878, 0.00217651, -0.00129323, -0.00010645, -0.00014665, },
		{ 0.00013149, -0.00027070, -0.00000662, 0.00008424, 0.00007032, -0.00014667, 0.00006774, -0.00004770, 0.00005021, -0.00002065, -0.00000965, 0.00000450, -0.00004986, 0.00005332, -0.00002568, 0.00000231, 0.00004643, -0.00010190, -0.00000662, 0.00000567, 0.00018500, -0.00030017, 0.00031971, -0.00106348, 0.00109267, -0.00050203, -0.00016063, 0.00209071, -0.00143653, 0.00003443, -0.00022483, 0.00022127, },
		{ -0.00024762, -0.00000662, 0.00011125, 0.00003692, -0.00013318, 0.00006114, -0.00004223, 0.00005257, -0.00002370, -0.00000965, 0.00000300, -0.00005366, 0.00006502, -0.00003805, 0.00002497, 0.00001156, -0.00008635, -0.00000662, 0.00004577, 0.00011021, -0.00025297, 0.00028120, -0.00100401, 0.00115232, -0.00081567, 0.00022799, 0.00193676, -0.00153267, 0.00016152, -0.00029689, 0.00030059, 0.00005340, },
		{ -0.00000662, 0.00013531, 0.00000168, -0.00011594, 0.00005224, -0.00003436, 0.00005286, -0.00002632, -0.00000965, 0.00000091, -0.00005545, 0.00007500, -0.00004888, 0.00004531, -0.00002418, -0.00006616, -0.00000662, 0.00008150, 0.00003127, -0.00019260, 0.00022937, -0.00091841, 0.00115964, -0.00108577, 0.00063214, 0.00172058, -0.00157795, 0.00026995, -0.00036004, 0.00037180, -0.00002665, -0.00021764, },
		{ 0.00015550, -0.00003406, -0.00009559, 0.00004141, -0.00002438, 0.00005107, -0.00002841, -0.00000965, -0.00000171, -0.00005516, 0.00008287, -0.00005778, 0.00006256, -0.00005942, -0.00004210, -0.00000662, 0.00011148, -0.00004878, -0.00012139, 0.00016621, -0.00080999, 0.00111436, -0.00130195, 0.00103629, 0.00145048, -0.00157063, 0.00035554, -0.00041187, 0.00043216, -0.00010559, -0.00018191, -0.00000662, },
		{ -0.00006892, -0.00007293, 0.00002904, -0.00001268, 0.00004727, -0.00002991, -0.00000965, -0.00000476, -0.00005280, 0.00008834, -0.00006438, 0.00007605, -0.00009281, -0.00001509, -0.00000662, 0.00013456, -0.00012687, -0.00004208, 0.00009415, -0.00068289, 0.00101822, -0.00145590, 0.00142491, 0.00113684, -0.00151099, 0.00041501, -0.00045039, 0.00047936, -0.00018038, -0.00014181, -0.00000662, 0.00017105, },
		{ -0.00004882, 0.00001563, 0.00000028, 0.00004160, -0.00003074, -0.00000965, -0.00000810, -0.00004846, 0.00009120, -0.00006845, 0.00008525, -0.00012307, 0.00001383, -0.00000662, 0.00014985, -0.00020000, 0.00004230, 0.00001597, -0.00054202, 0.00087491, -0.00154169, 0.00178306, 0.00079170, -0.00140131, 0.00044607, -0.00047410, 0.00051158, -0.00024815, -0.00009888, -0.00000662, 0.00018135, -0.00010157, },
		{ 0.00000168, 0.00001401, 0.00003428, -0.00003088, -0.00000965, -0.00001163, -0.00004231, 0.00009134, -0.00006983, 0.00008982, -0.00014903, 0.00004354, -0.00000662, 0.00015677, -0.00026535, 0.00012850, -0.00006534, -0.00039277, 0.00068994, -0.00155605, 0.00209698, 0.00042834, -0.00124583, 0.00044753, -0.00048211, 0.00052759, -0.00030630, -0.00005476, -0.00000662, 0.00018601, -0.00013075, -0.00002420, },
		{ 0.00002799, 0.00002559, -0.00003032, -0.00000965, -0.00001520, -0.00003458, 0.00008874, -0.00006845, 0.00008959, -0.00016970, 0.00007291, -0.00000662, 0.00015505, -0.00032041, 0.00021320, -0.00014665, -0.00024089, 0.00047042, -0.00149841, 0.00235461, 0.00006072, -0.00105050, 0.00041933, -0.00047410, 0.00052676, -0.00035259, -0.00001115, -0.00000662, 0.00018485, -0.00015533, 0.00000000, -0.00001227, },
		{ 0.00001587, -0.00002908, -0.00000965, -0.00001867, -0.00002558, 0.00008352, -0.00006438, 0.00008455, -0.00018427, 0.00010081, -0.00000662, 0.00014475, -0.00036307, 0.00029315, -0.00022483, -0.00009221, 0.00022479, -0.00137099, 0.00254605, -0.00029704, -0.00082284, 0.00036255, -0.00045039, 0.00050913, -0.00038524, 0.00003027, -0.00000662, 0.00017791, -0.00017438, 0.00002284, -0.00002568, 0.00004167, },
		{ -0.00002722, -0.00000965, -0.00002191, -0.00001564, 0.00007586, -0.00005778, 0.00007491, -0.00019220, 0.00012616, -0.00000662, 0.00012627, -0.00039169, 0.00036527, -0.00029689, 0.00004755, -0.00003753, -0.00117868, 0.00266394, -0.00063119, -0.00057160, 0.00027938, -0.00041187, 0.00047538, -0.00040299, 0.00006791, -0.00000662, 0.00016547, -0.00018716, 0.00004345, -0.00003805, 0.00005453, 0.00000550, },
		{ -0.00000965, -0.00002480, -0.00000514, 0.00006608, -0.00004888, 0.00006103, -0.00019317, 0.00014799, -0.00000662, 0.00010033, -0.00040517, 0.00042681, -0.00036004, 0.00017302, -0.00030643, -0.00092888, 0.00270374, -0.00092888, -0.00030643, 0.00017302, -0.00036004, 0.00042681, -0.00040517, 0.00010033, -0.00000662, 0.00014799, -0.00019317, 0.00006103, -0.00004888, 0.00006608, -0.00000514, -0.00002480, },
		{ -0.00002722, 0.00000550, 0.00005453, -0.00003805, 0.00004345, -0.00018716, 0.00016547, -0.00000662, 0.00006791, -0.00040299, 0.00047538, -0.00041187, 0.00027938, -0.00057160, -0.00063119, 0.00266394, -0.00117868, -0.00003753, 0.00004755, -0.00029689, 0.00036527, -0.00039169, 0.00012627, -0.00000662, 0.00012616, -0.00019220, 0.00007491, -0.00005778, 0.00007586, -0.00001564, -0.00002191, -0.00000965, },
		{ 0.00001587, 0.00004167, -0.00002568, 0.00002284, -0.00017438, 0.00017791, -0.00000662, 0.00003027, -0.00038524, 0.00050913, -0.00045039, 0.00036255, -0.00082284, -0.00029704, 0.00254605, -0.00137099, 0.00022479, -0.00009221, -0.00022483, 0.00029315, -0.00036307, 0.00014475, -0.00000662, 0.00010081, -0.00018427, 0.00008455, -0.00006438, 0.00008352, -0.00002558, -0.00001867, -0.00000965, -0.00002908, },
		{ 0.00002799, -0.00001227, 0.00000000, -0.00015533, 0.00018485, -0.00000662, -0.00001115, -0.00035259, 0.00052676, -0.00047410, 0.00041933, -0.00105050, 0.00006072, 0.00235461, -0.00149841, 0.00047042, -0.00024089, -0.00014665, 0.00021320, -0.00032041, 0.00015505, -0.00000662, 0.00007291, -0.00016970, 0.00008959, -0.00006845, 0.00008874, -0.00003458, -0.00001520, -0.00000965, -0.00003032, 0.00002559, },
		{ 0.00000168, -0.00002420, -0.00013075, 0.00018601, -0.00000662, -0.00005476, -0.00030630, 0.00052759, -0.00048211, 0.00044753, -0.00124583, 0.00042834, 0.00209698, -0.00155605, 0.00068994, -0.00039277, -0.00006534, 0.00012850, -0.00026535, 0.00015677, -0.00000662, 0.00004354, -0.00014903, 0.00008982, -0.00006983, 0.00009134, -0.00004231, -0.00001163, -0.00000965, -0.00003088, 0.00003428, 0.00001401, },
		{ -0.00004882, -0.00010157, 0.00018135, -0.00000662, -0.00009888, -0.00024815, 0.00051158, -0.00047410, 0.00044607, -0.00140131, 0.00079170, 0.00178306, -0.00154169, 0.00087491, -0.00054202, 0.00001597, 0.00004230, -0.00020000, 0.00014985, -0.00000662, 0.00001383, -0.00012307, 0.00008525, -0.00006845, 0.00009120, -0.00004846, -0.00000810, -0.00000965, -0.00003074, 0.00004160, 0.00000028, 0.00001563, },
		{ -0.00006892, 0.00017105, -0.00000662, -0.00014181, -0.00018038, 0.00047936, -0.00045039, 0.00041501, -0.00151099, 0.00113684, 0.00142491, -0.00145590, 0.00101822, -0.00068289, 0.00009415, -0.00004208, -0.00012687, 0.00013456, -0.00000662, -0.00001509, -0.00009281, 0.00007605, -0.00006438, 0.00008834, -0.00005280, -0.00000476, -0.00000965, -0.00002991, 0.00004727, -0.00001268, 0.00002904, -0.00007293, },
		{ 0.00015550, -0.00000662, -0.00018191, -0.00010559, 0.00043216, -0.00041187, 0.00035554, -0.00157063, 0.00145048, 0.00103629, -0.00130195, 0.00111436, -0.00080999, 0.00016621, -0.00012139, -0.00004878, 0.00011148, -0.00000662, -0.00004210, -0.00005942, 0.00006256, -0.00005778, 0.00008287, -0.00005516, -0.00000171, -0.00000965, -0.00002841, 0.00005107, -0.00002438, 0.00004141, -0.00009559, -0.00003406, },
		{ -0.00000662, -0.00021764, -0.00002665, 0.00037180, -0.00036004, 0.00026995, -0.00157795, 0.00172058, 0.00063214, -0.00108577, 0.00115964, -0.00091841, 0.00022937, -0.00019260, 0.00003127, 0.00008150, -0.00000662, -0.00006616, -0.00002418, 0.00004531, -0.00004888, 0.00007500, -0.00005545, 0.00000091, -0.00000965, -0.00002632, 0.00005286, -0.00003436, 0.00005224, -0.00011594, 0.00000168, 0.00013531, },
		{ -0.00024762, 0.00005340, 0.00030059, -0.00029689, 0.00016152, -0.00153267, 0.00193676, 0.00022799, -0.00081567, 0.00115232, -0.00100401, 0.00028120, -0.00025297, 0.00011021, 0.00004577, -0.00000662, -0.00008635, 0.00001156, 0.00002497, -0.00003805, 0.00006502, -0.00005366, 0.00000300, -0.00000965, -0.00002370, 0.00005257, -0.00004223, 0.00006114, -0.00013318, 0.00003692, 0.00011125, -0.00000662, },
		{ 0.00013149, 0.00022127, -0.00022483, 0.00003443, -0.00143653, 0.00209071, -0.00016063, -0.00050203, 0.00109267, -0.00106348, 0.00031971, -0.00030017, 0.00018500, 0.00000567, -0.00000662, -0.00010190, 0.00004643, 0.00000231, -0.00002568, 0.00005332, -0.00004986, 0.00000450, -0.00000965, -0.00002065, 0.00005021, -0.00004770, 0.00006774, -0.00014667, 0.00007032, 0.00008424, -0.00000662, -0.00027070, },
		{ 0.00013690, -0.00014665, -0.00010645, -0.00129323, 0.00217651, -0.00051878, -0.00015689, 0.00098300, -0.00109454, 0.00034343, -0.00033239, 0.00025277, -0.00003726, -0.00000662, -0.00011220, 0.00007907, -0.00002180, -0.00001227, 0.00004036, -0.00004419, 0.00000533, -0.00000965, -0.00001731, 0.00004587, -0.00005056, 0.00007181, -0.00015587, 0.00010057, 0.00005532, -0.00000662, -0.00028599, 0.00020462, },
		{ -0.00006534, -0.00025570, -0.00110826, 0.00219086, -0.00083270, 0.00020647, 0.00082752, -0.00109599, 0.00035144, -0.00034839, 0.00031092, -0.00008138, -0.00000662, -0.00011686, 0.00010825, -0.00004642, 0.00000168, 0.00002663, -0.00003687, 0.00000547, -0.00000965, -0.00001378, 0.00003972, -0.00005070, 0.00007319, -0.00016045, 0.00012654, 0.00002561, -0.00000662, -0.00029291, 0.00026997, 0.00005070, },
		{ -0.00040758, -0.00088874, 0.00213322, -0.00109033, 0.00057409, 0.00063219, -0.00106779, 0.00034343, -0.00034756, 0.00035721, -0.00012499, -0.00000662, -0.00011570, 0.00013283, -0.00007062, 0.00001563, 0.00001265, -0.00002818, 0.00000491, -0.00000965, -0.00001022, 0.00003199, -0.00004810, 0.00007181, -0.00016021, 0.00014720, -0.00000376, -0.00000662, -0.00029118, 0.00032503, -0.00003400, 0.00001597, },
		{ -0.00064310, 0.00200580, -0.00128177, 0.00093186, 0.00040453, -0.00101102, 0.00031971, -0.00032994, 0.00038986, -0.00016641, -0.00000662, -0.00010876, 0.00015188, -0.00009347, 0.00002904, -0.00000103, -0.00001846, 0.00000367, -0.00000965, -0.00000675, 0.00002299, -0.00004288, 0.00006774, -0.00015517, 0.00016178, -0.00003166, -0.00000662, -0.00028089, 0.00036769, -0.00011395, 0.00009415, -0.00055625, },
		{ 0.00181350, -0.00139966, 0.00126600, 0.00015329, -0.00092785, 0.00028120, -0.00029619, 0.00040761, -0.00020405, -0.00000662, -0.00009632, 0.00016466, -0.00011407, 0.00004141, -0.00001389, -0.00000809, 0.00000181, -0.00000965, -0.00000350, 0.00001305, -0.00003522, 0.00006114, -0.00014553, 0.00016970, -0.00005701, -0.00000662, -0.00026241, 0.00039631, -0.00018608, 0.00016621, -0.00069601, -0.00038079, },
		},
		{ { 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, },
		{ -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, },
		{ -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, },
		{ 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, },
		{ -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, },
		{ -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, },
		{ 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, },
		{ 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, },
		{ -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, },
		{ -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, },
		{ -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, },
		{ 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, },
		{ 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, },
		{ -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, },
		{ -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, },
		{ 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, },
		{ -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, },
		{ 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, },
		{ -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, },
		{ -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, },
		{ 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, },
		{ 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, },
		{ -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, },
		{ -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, },
		{ -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, },
		{ 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, },
		{ 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, },
		{ -0.00023338, -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, },
		{ -0.00006534, 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, },
		{ 0.00024486, -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, },
		{ -0.00020916, -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, },
		{ -0.00143537, 0.00339428, -0.00143537, -0.00020916, 0.00024486, -0.00006534, -0.00023338, 0.00000231, 0.00014229, -0.00000662, -0.00010710, -0.00001125, 0.00005696, 0.00000168, -0.00003205, -0.00000129, 0.00000430, -0.00000965, 0.00000430, -0.00000129, -0.00003205, 0.00000168, 0.00005696, -0.00001125, -0.00010710, -0.00000662, 0.00014229, 0.00000231, -0.00023338, -0.00006534, 0.00024486, -0.00020916, },
		},
		{ { -0.00143946, 0.00181350, -0.00064310, -0.00040758, -0.00006534, 0.00013690, 0.00013149, -0.00024762, -0.00000662, 0.00015550, -0.00006892, -0.00004882, 0.00000168, 0.00002799, 0.00001587, -0.00002722, -0.00000965, -0.00002722, 0.00001587, 0.00002799, 0.00000168, -0.00004882, -0.00006892, 0.00015550, -0.00000662, -0.00024762, 0.00013149, 0.00013690, -0.00006534, -0.00040758, -0.00064310, 0.00181350, },
		{ 0.00156370, -0.00038079, -0.00055625, 0.00001597, 0.00005070, 0.00020462, -0.00027070, -0.00000662, 0.00013531, -0.00003406, -0.00007293, 0.00001563, 0.00001401, 0.00002559, -0.00002908, -0.00000965, -0.00002480, 0.00000550, 0.00004167, -0.00001227, -0.00002420, -0.00010157, 0.00017105, -0.00000662, -0.00021764, 0.00005340, 0.00022127, -0.00014665, -0.00025570, -0.00088874, 0.00200580, -0.00139966, },
		{ -0.00011188, -0.00069601, 0.00009415, -0.00003400, 0.00026997, -0.00028599, -0.00000662, 0.00011125, 0.00000168, -0.00009559, 0.00002904, 0.00000028, 0.00003428, -0.00003032, -0.00000965, -0.00002191, -0.00000514, 0.00005453, -0.00002568, 0.00000000, -0.00013075, 0.00018135, -0.00000662, -0.00018191, -0.00002665, 0.00030059, -0.00022483, -0.00010645, -0.00110826, 0.00213322, -0.00128177, 0.00126600, },
		{ -0.00082149, 0.00016621, -0.00011395, 0.00032503, -0.00029291, -0.00000662, 0.00008424, 0.00003692, -0.00011594, 0.00004141, -0.00001268, 0.00004160, -0.00003088, -0.00000965, -0.00001867, -0.00001564, 0.00006608, -0.00003805, 0.00002284, -0.00015533, 0.00018601, -0.00000662, -0.00014181, -0.00010559, 0.00037180, -0.00029689, 0.00003443, -0.00129323, 0.00219086, -0.00109033, 0.00093186, 0.00015329, },
		{ 0.00022937, -0.00018608, 0.00036769, -0.00029118, -0.00000662, 0.00005532, 0.00007032, -0.00013318, 0.00005224, -0.00002438, 0.00004727, -0.00003074, -0.00000965, -0.00001520, -0.00002558, 0.00007586, -0.00004888, 0.00004345, -0.00017438, 0.00018485, -0.00000662, -0.00009888, -0.00018038, 0.00043216, -0.00036004, 0.00016152, -0.00143653, 0.00217651, -0.00083270, 0.00057409, 0.00040453, -0.00092785, },
		{ -0.00024761, 0.00039631, -0.00028089, -0.00000662, 0.00002561, 0.00010057, -0.00014667, 0.00006114, -0.00003436, 0.00005107, -0.00002991, -0.00000965, -0.00001163, -0.00003458, 0.00008352, -0.00005778, 0.00006103, -0.00018716, 0.00017791, -0.00000662, -0.00005476, -0.00024815, 0.00047936, -0.00041187, 0.00026995, -0.00153267, 0.00209071, -0.00051878, 0.00020647, 0.00063219, -0.00101102, 0.00028120, },
		{ 0.00040979, -0.00026241, -0.00000662, -0.00000376, 0.00012654, -0.00015587, 0.00006774, -0.00004223, 0.00005286, -0.00002841, -0.00000965, -0.00000810, -0.00004231, 0.00008874, -0.00006438, 0.00007491, -0.00019317, 0.00016547, -0.00000662, -0.00001115, -0.00030630, 0.00051158, -0.00045039, 0.00035554, -0.00157795, 0.00193676, -0.00016063, -0.00015689, 0.00082752, -0.00106779, 0.00031971, -0.00029619, },
		{ -0.00023647, -0.00000662, -0.00003166, 0.00014720, -0.00016045, 0.00007181, -0.00004770, 0.00005257, -0.00002632, -0.00000965, -0.00000476, -0.00004846, 0.00009134, -0.00006845, 0.00008455, -0.00019220, 0.00014799, -0.00000662, 0.00003027, -0.00035259, 0.00052759, -0.00047410, 0.00041501, -0.00157063, 0.00172058, 0.00022799, -0.00050203, 0.00098300, -0.00109599, 0.00034343, -0.00032994, 0.00040761, },
		{ -0.00000662, -0.00005701, 0.00016178, -0.00016021, 0.00007319, -0.00005056, 0.00005021, -0.00002370, -0.00000965, -0.00000171, -0.00005280, 0.00009120, -0.00006983, 0.00008959, -0.00018427, 0.00012616, -0.00000662, 0.00006791, -0.00038524, 0.00052676, -0.00048211, 0.00044607, -0.00151099, 0.00145048, 0.00063214, -0.00081567, 0.00109267, -0.00109454, 0.00035144, -0.00034756, 0.00038986, -0.00020405, },
		{ -0.00007884, 0.00016970, -0.00015517, 0.00007181, -0.00005070, 0.00004587, -0.00002065, -0.00000965, 0.00000091, -0.00005516, 0.00008834, -0.00006845, 0.00008982, -0.00016970, 0.00010081, -0.00000662, 0.00010033, -0.00040299, 0.00050913, -0.00047410, 0.00044753, -0.00140131, 0.00113684, 0.00103629, -0.00108577, 0.00115232, -0.00106348, 0.00034343, -0.00034839, 0.00035721, -0.00016641, -0.00000662, },
		{ 0.00017068, -0.00014553, 0.00006774, -0.00004810, 0.00003972, -0.00001731, -0.00000965, 0.00000300, -0.00005545, 0.00008287, -0.00006438, 0.00008525, -0.00014903, 0.00007291, -0.00000662, 0.00012627, -0.00040517, 0.00047538, -0.00045039, 0.00041933, -0.00124583, 0.00079170, 0.00142491, -0.00130195, 0.00115964, -0.00100401, 0.00031971, -0.00033239, 0.00031092, -0.00012499, -0.00000662, -0.00009632, },
		{ -0.00013165, 0.00006114, -0.00004288, 0.00003199, -0.00001378, -0.00000965, 0.00000450, -0.00005366, 0.00007500, -0.00005778, 0.00007605, -0.00012307, 0.00004354, -0.00000662, 0.00014475, -0.00039169, 0.00042681, -0.00041187, 0.00036255, -0.00105050, 0.00042834, 0.00178306, -0.00145590, 0.00111436, -0.00091841, 0.00028120, -0.00030017, 0.00025277, -0.00008138, -0.00000662, -0.00010876, 0.00016466, },
		{ 0.00005224, -0.00003522, 0.00002299, -0.00001022, -0.00000965, 0.00000533, -0.00004986, 0.00006502, -0.00004888, 0.00006256, -0.00009281, 0.00001383, -0.00000662, 0.00015505, -0.00036307, 0.00036527, -0.00036004, 0.00027938, -0.00082284, 0.00006072, 0.00209698, -0.00154169, 0.00101822, -0.00080999, 0.00022937, -0.00025297, 0.00018500, -0.00003726, -0.00000662, -0.00011570, 0.00015188, -0.00011407, },
		{ -0.00002544, 0.00001305, -0.00000675, -0.00000965, 0.00000547, -0.00004419, 0.00005332, -0.00003805, 0.00004531, -0.00005942, -0.00001509, -0.00000662, 0.00015677, -0.00032041, 0.00029315, -0.00029689, 0.00017302, -0.00057160, -0.00029704, 0.00235461, -0.00155605, 0.00087491, -0.00068289, 0.00016621, -0.00019260, 0.00011021, 0.00000567, -0.00000662, -0.00011686, 0.00013283, -0.00009347, 0.00004141, },
		{ 0.00000255, -0.00000350, -0.00000965, 0.00000491, -0.00003687, 0.00004036, -0.00002568, 0.00002497, -0.00002418, -0.00004210, -0.00000662, 0.00014985, -0.00026535, 0.00021320, -0.00022483, 0.00004755, -0.00030643, -0.00063119, 0.00254605, -0.00149841, 0.00068994, -0.00054202, 0.00009415, -0.00012139, 0.00003127, 0.00004577, -0.00000662, -0.00011220, 0.00010825, -0.00007062, 0.00002904, -0.00001389, },
		{ -0.00000062, -0.00000965, 0.00000367, -0.00002818, 0.00002663, -0.00001227, 0.00000231, 0.00001156, -0.00006616, -0.00000662, 0.00013456, -0.00020000, 0.00012850, -0.00014665, -0.00009221, -0.00003753, -0.00092888, 0.00266394, -0.00137099, 0.00047042, -0.00039277, 0.00001597, -0.00004208, -0.00004878, 0.00008150, -0.00000662, -0.00010190, 0.00007907, -0.00004642, 0.00001563, -0.00000103, -0.00000809, },
		{ -0.00000965, 0.00000181, -0.00001846, 0.00001265, 0.00000168, -0.00002180, 0.00004643, -0.00008635, -0.00000662, 0.00011148, -0.00012687, 0.00004230, -0.00006534, -0.00024089, 0.00022479, -0.00117868, 0.00270374, -0.00117868, 0.00022479, -0.00024089, -0.00006534, 0.00004230, -0.00012687, 0.00011148, -0.00000662, -0.00008635, 0.00004643, -0.00002180, 0.00000168, 0.00001265, -0.00001846, 0.00000181, },
		{ -0.00000062, -0.00000809, -0.00000103, 0.00001563, -0.00004642, 0.00007907, -0.00010190, -0.00000662, 0.00008150, -0.00004878, -0.00004208, 0.00001597, -0.00039277, 0.00047042, -0.00137099, 0.00266394, -0.00092888, -0.00003753, -0.00009221, -0.00014665, 0.00012850, -0.00020000, 0.00013456, -0.00000662, -0.00006616, 0.00001156, 0.00000231, -0.00001227, 0.00002663, -0.00002818, 0.00000367, -0.00000965, },
		{ 0.00000255, -0.00001389, 0.00002904, -0.00007062, 0.00010825, -0.00011220, -0.00000662, 0.00004577, 0.00003127, -0.00012139, 0.00009415, -0.00054202, 0.00068994, -0.00149841, 0.00254605, -0.00063119, -0.00030643, 0.00004755, -0.00022483, 0.00021320, -0.00026535, 0.00014985, -0.00000662, -0.00004210, -0.00002418, 0.00002497, -0.00002568, 0.00004036, -0.00003687, 0.00000491, -0.00000965, -0.00000350, },
		{ -0.00002544, 0.00004141, -0.00009347, 0.00013283, -0.00011686, -0.00000662, 0.00000567, 0.00011021, -0.00019260, 0.00016621, -0.00068289, 0.00087491, -0.00155605, 0.00235461, -0.00029704, -0.00057160, 0.00017302, -0.00029689, 0.00029315, -0.00032041, 0.00015677, -0.00000662, -0.00001509, -0.00005942, 0.00004531, -0.00003805, 0.00005332, -0.00004419, 0.00000547, -0.00000965, -0.00000675, 0.00001305, },
		{ 0.00005224, -0.00011407, 0.00015188, -0.00011570, -0.00000662, -0.00003726, 0.00018500, -0.00025297, 0.00022937, -0.00080999, 0.00101822, -0.00154169, 0.00209698, 0.00006072, -0.00082284, 0.00027938, -0.00036004, 0.00036527, -0.00036307, 0.00015505, -0.00000662, 0.00001383, -0.00009281, 0.00006256, -0.00004888, 0.00006502, -0.00004986, 0.00000533, -0.00000965, -0.00001022, 0.00002299, -0.00003522, },
		{ -0.00013165, 0.00016466, -0.00010876, -0.00000662, -0.00008138, 0.00025277, -0.00030017, 0.00028120, -0.00091841, 0.00111436, -0.00145590, 0.00178306, 0.00042834, -0.00105050, 0.00036255, -0.00041187, 0.00042681, -0.00039169, 0.00014475, -0.00000662, 0.00004354, -0.00012307, 0.00007605, -0.00005778, 0.00007500, -0.00005366, 0.00000450, -0.00000965, -0.00001378, 0.00003199, -0.00004288, 0.00006114, },
		{ 0.00017068, -0.00009632, -0.00000662, -0.00012499, 0.00031092, -0.00033239, 0.00031971, -0.00100401, 0.00115964, -0.00130195, 0.00142491, 0.00079170, -0.00124583, 0.00041933, -0.00045039, 0.00047538, -0.00040517, 0.00012627, -0.00000662, 0.00007291, -0.00014903, 0.00008525, -0.00006438, 0.00008287, -0.00005545, 0.00000300, -0.00000965, -0.00001731, 0.00003972, -0.00004810, 0.00006774, -0.00014553, },
		{ -0.00007884, -0.00000662, -0.00016641, 0.00035721, -0.00034839, 0.00034343, -0.00106348, 0.00115232, -0.00108577, 0.00103629, 0.00113684, -0.00140131, 0.00044753, -0.00047410, 0.00050913, -0.00040299, 0.00010033, -0.00000662, 0.00010081, -0.00016970, 0.00008982, -0.00006845, 0.00008834, -0.00005516, 0.00000091, -0.00000965, -0.00002065, 0.00004587, -0.00005070, 0.00007181, -0.00015517, 0.00016970, },
		{ -0.00000662, -0.00020405, 0.00038986, -0.00034756, 0.00035144, -0.00109454, 0.00109267, -0.00081567, 0.00063214, 0.00145048, -0.00151099, 0.00044607, -0.00048211, 0.00052676, -0.00038524, 0.00006791, -0.00000662, 0.00012616, -0.00018427, 0.00008959, -0.00006983, 0.00009120, -0.00005280, -0.00000171, -0.00000965, -0.00002370, 0.00005021, -0.00005056, 0.00007319, -0.00016021, 0.00016178, -0.00005701, },
		{ -0.00023647, 0.00040761, -0.00032994, 0.00034343, -0.00109599, 0.00098300, -0.00050203, 0.00022799, 0.00172058, -0.00157063, 0.00041501, -0.00047410, 0.00052759, -0.00035259, 0.00003027, -0.00000662, 0.00014799, -0.00019220, 0.00008455, -0.00006845, 0.00009134, -0.00004846, -0.00000476, -0.00000965, -0.00002632, 0.00005257, -0.00004770, 0.00007181, -0.00016045, 0.00014720, -0.00003166, -0.00000662, },
		{ 0.00040979, -0.00029619, 0.00031971, -0.00106779, 0.00082752, -0.00015689, -0.00016063, 0.00193676, -0.00157795, 0.00035554, -0.00045039, 0.00051158, -0.00030630, -0.00001115, -0.00000662, 0.00016547, -0.00019317, 0.00007491, -0.00006438, 0.00008874, -0.00004231, -0.00000810, -0.00000965, -0.00002841, 0.00005286, -0.00004223, 0.00006774, -0.00015587, 0.00012654, -0.00000376, -0.00000662, -0.00026241, },
		{ -0.00024761, 0.00028120, -0.00101102, 0.00063219, 0.00020647, -0.00051878, 0.00209071, -0.00153267, 0.00026995, -0.00041187, 0.00047936, -0.00024815, -0.00005476, -0.00000662, 0.00017791, -0.00018716, 0.00006103, -0.00005778, 0.00008352, -0.00003458, -0.00001163, -0.00000965, -0.00002991, 0.00005107, -0.00003436, 0.00006114, -0.00014667, 0.00010057, 0.00002561, -0.00000662, -0.00028089, 0.00039631, },
		{ 0.00022937, -0.00092785, 0.00040453, 0.00057409, -0.00083270, 0.00217651, -0.00143653, 0.00016152, -0.00036004, 0.00043216, -0.00018038, -0.00009888, -0.00000662, 0.00018485, -0.00017438, 0.00004345, -0.00004888, 0.00007586, -0.00002558, -0.00001520, -0.00000965, -0.00003074, 0.00004727, -0.00002438, 0.00005224, -0.00013318, 0.00007032, 0.00005532, -0.00000662, -0.00029118, 0.00036769, -0.00018608, },
		{ -0.00082149, 0.00015329, 0.00093186, -0.00109033, 0.00219086, -0.00129323, 0.00003443, -0.00029689, 0.00037180, -0.00010559, -0.00014181, -0.00000662, 0.00018601, -0.00015533, 0.00002284, -0.00003805, 0.00006608, -0.00001564, -0.00001867, -0.00000965, -0.00003088, 0.00004160, -0.00001268, 0.00004141, -0.00011594, 0.00003692, 0.00008424, -0.00000662, -0.00029291, 0.00032503, -0.00011395, 0.00016621, },
		{ -0.00011188, 0.00126600, -0.00128177, 0.00213322, -0.00110826, -0.00010645, -0.00022483, 0.00030059, -0.00002665, -0.00018191, -0.00000662, 0.00018135, -0.00013075, 0.00000000, -0.00002568, 0.00005453, -0.00000514, -0.00002191, -0.00000965, -0.00003032, 0.00003428, 0.00000028, 0.00002904, -0.00009559, 0.00000168, 0.00011125, -0.00000662, -0.00028599, 0.00026997, -0.00003400, 0.00009415, -0.00069601, },
		{ 0.00156370, -0.00139966, 0.00200580, -0.00088874, -0.00025570, -0.00014665, 0.00022127, 0.00005340, -0.00021764, -0.00000662, 0.00017105, -0.00010157, -0.00002420, -0.00001227, 0.00004167, 0.00000550, -0.00002480, -0.00000965, -0.00002908, 0.00002559, 0.00001401, 0.00001563, -0.00007293, -0.00003406, 0.00013531, -0.00000662, -0.00027070, 0.00020462, 0.00005070, 0.00001597, -0.00055625, -0.00038079, },
		},
		{ { 0.00004231, -0.00081505, 0.00170748, -0.00061106, 0.00005920, -0.00008887, -0.00036754, 0.00013369, -0.00007855, 0.00004917, 0.00006559, 0.00004507, -0.00003654, -0.00000047, -0.00001368, -0.00007197, 0.00009105, -0.00007197, -0.00001368, -0.00000047, -0.00003654, 0.00004507, 0.00006559, 0.00004917, -0.00007855, 0.00013369, -0.00036754, -0.00008887, 0.00005920, -0.00061106, 0.00170748, -0.00081505, },
		{ -0.00082261, 0.00141529, -0.00020291, -0.00024087, 0.00015805, -0.00052239, 0.00020204, -0.00012709, 0.00008031, 0.00003113, 0.00007594, -0.00004411, -0.00000281, -0.00000262, -0.00008238, 0.00008421, -0.00005120, -0.00002831, 0.00000043, -0.00002343, 0.00000823, 0.00010360, 0.00001048, -0.00001742, 0.00004541, -0.00019053, -0.00033576, 0.00033077, -0.00093917, 0.00184588, -0.00065277, -0.00001906, },
		{ 0.00101457, 0.00023302, -0.00052513, 0.00037027, -0.00063211, 0.00023724, -0.00015441, 0.00009778, 0.00000530, 0.00009729, -0.00004525, -0.00000560, 0.00000302, -0.00008133, 0.00006468, -0.00002277, -0.00004399, -0.00000090, -0.00000651, -0.00003006, 0.00013918, -0.00002839, 0.00004569, -0.00004682, -0.00001691, -0.00054796, 0.00053422, -0.00114665, 0.00180502, -0.00033545, -0.00019108, -0.00069965, },
		{ 0.00064034, -0.00075122, 0.00051823, -0.00067974, 0.00023091, -0.00015521, 0.00009775, -0.00000849, 0.00010705, -0.00004003, -0.00000780, 0.00000238, -0.00006951, 0.00003534, 0.00000937, -0.00005789, -0.00000485, 0.00001192, -0.00006494, 0.00016635, -0.00005995, 0.00009983, -0.00012682, 0.00012904, -0.00069599, 0.00064060, -0.00121043, 0.00158329, 0.00011235, -0.00043946, -0.00048959, 0.00056347, },
		{ -0.00088522, 0.00058189, -0.00065697, 0.00018093, -0.00012839, 0.00007926, -0.00000898, 0.00010488, -0.00002944, -0.00000850, -0.00000431, -0.00004923, 0.00000052, 0.00004065, -0.00006736, -0.00001139, 0.00002931, -0.00009193, 0.00018010, -0.00007785, 0.00013550, -0.00018074, 0.00022792, -0.00075995, 0.00063597, -0.00112812, 0.00120345, 0.00064400, -0.00071323, -0.00024750, 0.00012429, 0.00096676, },
		{ 0.00055371, -0.00056543, 0.00009187, -0.00007724, 0.00004439, 0.00000275, 0.00009219, -0.00001527, -0.00000709, -0.00001574, -0.00002407, -0.00003469, 0.00006655, -0.00007031, -0.00002001, 0.00004325, -0.00010758, 0.00017715, -0.00007784, 0.00014605, -0.00019925, 0.00026808, -0.00073247, 0.00052341, -0.00091812, 0.00070965, 0.00119709, -0.00095298, -0.00003084, -0.00024579, 0.00117165, -0.00090674, },
		{ -0.00041650, -0.00002557, -0.00000898, -0.00000194, 0.00002349, 0.00007188, 0.00000020, -0.00000337, -0.00002976, 0.00000165, -0.00006518, 0.00008325, -0.00006558, -0.00002978, 0.00005184, -0.00010995, 0.00015651, -0.00005854, 0.00012883, -0.00017898, 0.00024722, -0.00061969, 0.00032249, -0.00061658, 0.00016117, 0.00170220, -0.00110064, 0.00011008, -0.00050301, 0.00123193, -0.00081204, 0.00043969, },
		{ -0.00015612, 0.00006631, -0.00005282, 0.00004847, 0.00004789, 0.00001453, 0.00000242, -0.00004370, 0.00002363, -0.00008659, 0.00008824, -0.00005314, -0.00003952, 0.00005393, -0.00009893, 0.00011970, -0.00002164, 0.00008556, -0.00012307, 0.00017263, -0.00044031, 0.00006618, -0.00027171, -0.00037596, 0.00209293, -0.00110909, 0.00014042, -0.00062323, 0.00114519, -0.00061462, 0.00025843, -0.00022969, },
		{ 0.00013724, -0.00010030, 0.00007214, 0.00002450, 0.00002552, 0.00000960, -0.00005481, 0.00003816, -0.00009597, 0.00008071, -0.00003413, -0.00004790, 0.00004935, -0.00007617, 0.00007055, 0.00002825, 0.00002215, -0.00004069, 0.00005990, -0.00022262, -0.00020426, 0.00006358, -0.00083835, 0.00231556, -0.00095033, 0.00004665, -0.00060502, 0.00092987, -0.00034319, 0.00003820, -0.00002977, -0.00028211, },
		{ -0.00013666, 0.00008905, 0.00000573, 0.00003152, 0.00001714, -0.00006068, 0.00004278, -0.00009220, 0.00006163, -0.00001068, -0.00005371, 0.00003887, -0.00004494, 0.00001471, 0.00008432, -0.00005218, 0.00005438, -0.00006961, -0.00000023, -0.00044565, 0.00033984, -0.00117418, 0.00233705, -0.00062074, -0.00016106, -0.00046921, 0.00062223, -0.00003736, -0.00018734, 0.00015681, -0.00038613, 0.00019282, },
		{ 0.00009481, -0.00000534, 0.00003171, 0.00002385, -0.00005969, 0.00003662, -0.00007619, 0.00003365, 0.00001434, -0.00005598, 0.00002411, -0.00000952, -0.00004114, 0.00013851, -0.00012632, 0.00014592, -0.00019218, 0.00019315, -0.00061952, 0.00051848, -0.00135041, 0.00215002, -0.00014271, -0.00045025, -0.00025509, 0.00027067, 0.00025831, -0.00038430, 0.00030586, -0.00045357, 0.00022414, -0.00015561, },
		{ -0.00000702, 0.00002617, 0.00002854, -0.00005121, 0.00002054, -0.00005067, 0.00000069, 0.00003774, -0.00005416, 0.00000729, 0.00002532, -0.00009027, 0.00018277, -0.00018892, 0.00021776, -0.00028547, 0.00032874, -0.00069807, 0.00057752, -0.00135707, 0.00177402, 0.00043772, -0.00077111, -0.00001361, -0.00007167, 0.00050106, -0.00052370, 0.00039893, -0.00047482, 0.00022576, -0.00015324, 0.00008676, },
		{ 0.00001594, 0.00003020, -0.00003579, -0.00000295, -0.00001982, -0.00003265, 0.00005650, -0.00004822, -0.00000909, 0.00005513, -0.00012681, 0.00021023, -0.00023012, 0.00025638, -0.00033214, 0.00038715, -0.00066841, 0.00051476, -0.00120817, 0.00125292, 0.00105679, -0.00106430, 0.00020098, -0.00035499, 0.00065631, -0.00058583, 0.00042594, -0.00044670, 0.00019653, -0.00012867, 0.00006450, 0.00000070, },
		{ 0.00002819, -0.00001507, -0.00003017, 0.00001142, -0.00006176, 0.00006818, -0.00003865, -0.00002268, 0.00007633, -0.00014661, 0.00021630, -0.00024308, 0.00025303, -0.00032252, 0.00036120, -0.00053460, 0.00034786, -0.00093895, 0.00064873, 0.00164186, -0.00127041, 0.00034013, -0.00054033, 0.00070285, -0.00056321, 0.00038662, -0.00037294, 0.00013985, -0.00008424, 0.00002999, 0.00001621, 0.00000278, },
		{ 0.00000845, -0.00005672, 0.00003809, -0.00008271, 0.00007129, -0.00002638, -0.00003157, 0.00008678, -0.00014781, 0.00019934, -0.00022504, 0.00020517, -0.00025622, 0.00025699, -0.00031703, 0.00011130, -0.00060009, 0.00003282, 0.00212156, -0.00133953, 0.00036895, -0.00060534, 0.00063599, -0.00046162, 0.00029034, -0.00026366, 0.00006315, -0.00002521, -0.00001271, 0.00003648, -0.00001107, 0.00002233, },
		{ -0.00007825, 0.00005599, -0.00009283, 0.00006548, -0.00001273, -0.00003462, 0.00008604, -0.00013107, 0.00016097, -0.00017777, 0.00011702, -0.00014229, 0.00009286, -0.00004930, -0.00014920, -0.00024955, -0.00052414, 0.00243603, -0.00123958, 0.00027209, -0.00054745, 0.00046814, -0.00029918, 0.00015458, -0.00013380, -0.00002318, 0.00004099, -0.00005798, 0.00005761, -0.00002324, 0.00001294, 0.00003172, },
		{ 0.00006228, -0.00009109, 0.00005165, 0.00000085, -0.00003161, 0.00007539, -0.00009947, 0.00010585, -0.00010732, -0.00000102, 0.00000217, -0.00010345, 0.00022690, -0.00038399, 0.00005644, -0.00096204, 0.00254550, -0.00096204, 0.00005644, -0.00038399, 0.00022690, -0.00010345, 0.00000217, -0.00000102, -0.00010732, 0.00010585, -0.00009947, 0.00007539, -0.00003161, 0.00000085, 0.00005165, -0.00009109, },
		{ -0.00007825, 0.00003172, 0.00001294, -0.00002324, 0.00005761, -0.00005798, 0.00004099, -0.00002318, -0.00013380, 0.00015458, -0.00029918, 0.00046814, -0.00054745, 0.00027209, -0.00123958, 0.00243603, -0.00052414, -0.00024955, -0.00014920, -0.00004930, 0.00009286, -0.00014229, 0.00011702, -0.00017777, 0.00016097, -0.00013107, 0.00008604, -0.00003462, -0.00001273, 0.00006548, -0.00009283, 0.00005599, },
		{ 0.00000845, 0.00002233, -0.00001107, 0.00003648, -0.00001271, -0.00002521, 0.00006315, -0.00026366, 0.00029034, -0.00046162, 0.00063599, -0.00060534, 0.00036895, -0.00133953, 0.00212156, 0.00003282, -0.00060009, 0.00011130, -0.00031703, 0.00025699, -0.00025622, 0.00020517, -0.00022504, 0.00019934, -0.00014781, 0.00008678, -0.00003157, -0.00002638, 0.00007129, -0.00008271, 0.00003809, -0.00005672, },
		{ 0.00002819, 0.00000278, 0.00001621, 0.00002999, -0.00008424, 0.00013985, -0.00037294, 0.00038662, -0.00056321, 0.00070285, -0.00054033, 0.00034013, -0.00127041, 0.00164186, 0.00064873, -0.00093895, 0.00034786, -0.00053460, 0.00036120, -0.00032252, 0.00025303, -0.00024308, 0.00021630, -0.00014661, 0.00007633, -0.00002268, -0.00003865, 0.00006818, -0.00006176, 0.00001142, -0.00003017, -0.00001507, },
		{ 0.00001594, 0.00000070, 0.00006450, -0.00012867, 0.00019653, -0.00044670, 0.00042594, -0.00058583, 0.00065631, -0.00035499, 0.00020098, -0.00106430, 0.00105679, 0.00125292, -0.00120817, 0.00051476, -0.00066841, 0.00038715, -0.00033214, 0.00025638, -0.00023012, 0.00021023, -0.00012681, 0.00005513, -0.00000909, -0.00004822, 0.00005650, -0.00003265, -0.00001982, -0.00000295, -0.00003579, 0.00003020, },
		{ -0.00000702, 0.00008676, -0.00015324, 0.00022576, -0.00047482, 0.00039893, -0.00052370, 0.00050106, -0.00007167, -0.00001361, -0.00077111, 0.00043772, 0.00177402, -0.00135707, 0.00057752, -0.00069807, 0.00032874, -0.00028547, 0.00021776, -0.00018892, 0.00018277, -0.00009027, 0.00002532, 0.00000729, -0.00005416, 0.00003774, 0.00000069, -0.00005067, 0.00002054, -0.00005121, 0.00002854, 0.00002617, },
		{ 0.00009481, -0.00015561, 0.00022414, -0.00045357, 0.00030586, -0.00038430, 0.00025831, 0.00027067, -0.00025509, -0.00045025, -0.00014271, 0.00215002, -0.00135041, 0.00051848, -0.00061952, 0.00019315, -0.00019218, 0.00014592, -0.00012632, 0.00013851, -0.00004114, -0.00000952, 0.00002411, -0.00005598, 0.00001434, 0.00003365, -0.00007619, 0.00003662, -0.00005969, 0.00002385, 0.00003171, -0.00000534, },
		{ -0.00013666, 0.00019282, -0.00038613, 0.00015681, -0.00018734, -0.00003736, 0.00062223, -0.00046921, -0.00016106, -0.00062074, 0.00233705, -0.00117418, 0.00033984, -0.00044565, -0.00000023, -0.00006961, 0.00005438, -0.00005218, 0.00008432, 0.00001471, -0.00004494, 0.00003887, -0.00005371, -0.00001068, 0.00006163, -0.00009220, 0.00004278, -0.00006068, 0.00001714, 0.00003152, 0.00000573, 0.00008905, },
		{ 0.00013724, -0.00028211, -0.00002977, 0.00003820, -0.00034319, 0.00092987, -0.00060502, 0.00004665, -0.00095033, 0.00231556, -0.00083835, 0.00006358, -0.00020426, -0.00022262, 0.00005990, -0.00004069, 0.00002215, 0.00002825, 0.00007055, -0.00007617, 0.00004935, -0.00004790, -0.00003413, 0.00008071, -0.00009597, 0.00003816, -0.00005481, 0.00000960, 0.00002552, 0.00002450, 0.00007214, -0.00010030, },
		{ -0.00015612, -0.00022969, 0.00025843, -0.00061462, 0.00114519, -0.00062323, 0.00014042, -0.00110909, 0.00209293, -0.00037596, -0.00027171, 0.00006618, -0.00044031, 0.00017263, -0.00012307, 0.00008556, -0.00002164, 0.00011970, -0.00009893, 0.00005393, -0.00003952, -0.00005314, 0.00008824, -0.00008659, 0.00002363, -0.00004370, 0.00000242, 0.00001453, 0.00004789, 0.00004847, -0.00005282, 0.00006631, },
		{ -0.00041650, 0.00043969, -0.00081204, 0.00123193, -0.00050301, 0.00011008, -0.00110064, 0.00170220, 0.00016117, -0.00061658, 0.00032249, -0.00061969, 0.00024722, -0.00017898, 0.00012883, -0.00005854, 0.00015651, -0.00010995, 0.00005184, -0.00002978, -0.00006558, 0.00008325, -0.00006518, 0.00000165, -0.00002976, -0.00000337, 0.00000020, 0.00007188, 0.00002349, -0.00000194, -0.00000898, -0.00002557, },
		{ 0.00055371, -0.00090674, 0.00117165, -0.00024579, -0.00003084, -0.00095298, 0.00119709, 0.00070965, -0.00091812, 0.00052341, -0.00073247, 0.00026808, -0.00019925, 0.00014605, -0.00007784, 0.00017715, -0.00010758, 0.00004325, -0.00002001, -0.00007031, 0.00006655, -0.00003469, -0.00002407, -0.00001574, -0.00000709, -0.00001527, 0.00009219, 0.00000275, 0.00004439, -0.00007724, 0.00009187, -0.00056543, },
		{ -0.00088522, 0.00096676, 0.00012429, -0.00024750, -0.00071323, 0.00064400, 0.00120345, -0.00112812, 0.00063597, -0.00075995, 0.00022792, -0.00018074, 0.00013550, -0.00007785, 0.00018010, -0.00009193, 0.00002931, -0.00001139, -0.00006736, 0.00004065, 0.00000052, -0.00004923, -0.00000431, -0.00000850, -0.00002944, 0.00010488, -0.00000898, 0.00007926, -0.00012839, 0.00018093, -0.00065697, 0.00058189, },
		{ 0.00064034, 0.00056347, -0.00048959, -0.00043946, 0.00011235, 0.00158329, -0.00121043, 0.00064060, -0.00069599, 0.00012904, -0.00012682, 0.00009983, -0.00005995, 0.00016635, -0.00006494, 0.00001192, -0.00000485, -0.00005789, 0.00000937, 0.00003534, -0.00006951, 0.00000238, -0.00000780, -0.00004003, 0.00010705, -0.00000849, 0.00009775, -0.00015521, 0.00023091, -0.00067974, 0.00051823, -0.00075122, },
		{ 0.00101457, -0.00069965, -0.00019108, -0.00033545, 0.00180502, -0.00114665, 0.00053422, -0.00054796, -0.00001691, -0.00004682, 0.00004569, -0.00002839, 0.00013918, -0.00003006, -0.00000651, -0.00000090, -0.00004399, -0.00002277, 0.00006468, -0.00008133, 0.00000302, -0.00000560, -0.00004525, 0.00009729, 0.00000530, 0.00009778, -0.00015441, 0.00023724, -0.00063211, 0.00037027, -0.00052513, 0.00023302, },
		{ -0.00082261, -0.00001906, -0.00065277, 0.00184588, -0.00093917, 0.00033077, -0.00033576, -0.00019053, 0.00004541, -0.00001742, 0.00001048, 0.00010360, 0.00000823, -0.00002343, 0.00000043, -0.00002831, -0.00005120, 0.00008421, -0.00008238, -0.00000262, -0.00000281, -0.00004411, 0.00007594, 0.00003113, 0.00008031, -0.00012709, 0.00020204, -0.00052239, 0.00015805, -0.00024087, -0.00020291, 0.00141529, },
		},
		{ { -0.00008141, -0.00010262, -0.00074863, 0.00190195, -0.00081146, -0.00001019, 0.00005135, -0.00003588, 0.00010272, -0.00035293, 0.00007574, -0.00012692, 0.00012632, -0.00008696, -0.00005820, 0.00021619, -0.00011323, 0.00021619, -0.00005820, -0.00008696, 0.00012632, -0.00012692, 0.00007574, -0.00035293, 0.00010272, -0.00003588, 0.00005135, -0.00001019, -0.00081146, 0.00190195, -0.00074863, -0.00010262, },
		{ 0.00002588, -0.00088845, 0.00155446, -0.00017853, -0.00048475, 0.00035930, -0.00027190, 0.00029755, -0.00043781, 0.00009858, -0.00014301, 0.00009791, -0.00002165, -0.00012474, 0.00022703, -0.00009278, 0.00016614, 0.00002706, -0.00013906, 0.00011991, -0.00006997, 0.00000561, -0.00020902, -0.00013324, 0.00021204, -0.00024465, 0.00039656, -0.00124063, 0.00190954, -0.00032016, -0.00037471, -0.00003622, },
		{ -0.00078181, 0.00098385, 0.00050390, -0.00089142, 0.00057459, -0.00041451, 0.00039157, -0.00042989, 0.00005954, -0.00011520, 0.00004307, 0.00004116, -0.00015929, 0.00019551, -0.00003729, 0.00009039, 0.00011262, -0.00016511, 0.00008003, 0.00001105, -0.00008097, -0.00005669, -0.00033889, 0.00038817, -0.00043353, 0.00062743, -0.00136874, 0.00154525, 0.00035160, -0.00072921, 0.00006974, 0.00001940, },
		{ 0.00035295, 0.00106952, -0.00110857, 0.00061787, -0.00041202, 0.00035689, -0.00032427, -0.00003426, -0.00005503, -0.00002256, 0.00008713, -0.00015715, 0.00012969, 0.00003739, 0.00000892, 0.00017981, -0.00015854, 0.00001738, 0.00009108, -0.00014792, 0.00005404, -0.00045465, 0.00043450, -0.00046018, 0.00063384, -0.00118309, 0.00087585, 0.00113643, -0.00104711, 0.00016022, -0.00005816, -0.00053369, },
		{ 0.00138836, -0.00106675, 0.00046217, -0.00026049, 0.00020706, -0.00014745, -0.00015588, 0.00001540, -0.00008086, 0.00010698, -0.00012297, 0.00004633, 0.00011033, -0.00005768, 0.00021467, -0.00012080, -0.00005127, 0.00014495, -0.00016653, 0.00009052, -0.00045212, 0.00033766, -0.00032618, 0.00043879, -0.00075717, 0.00005131, 0.00185049, -0.00118740, 0.00014552, -0.00012173, -0.00026934, -0.00017959, },
		{ -0.00077076, 0.00014319, -0.00000588, -0.00000699, 0.00005061, -0.00026727, 0.00007085, -0.00011662, 0.00009866, -0.00006887, -0.00003417, 0.00016175, -0.00009408, 0.00021161, -0.00006072, -0.00010814, 0.00015566, -0.00012557, 0.00004814, -0.00034266, 0.00013249, -0.00008704, 0.00012562, -0.00022597, -0.00073646, 0.00230730, -0.00103550, -0.00003620, -0.00010527, -0.00008933, -0.00050511, 0.00140304, },
		{ -0.00025071, 0.00027042, -0.00021342, 0.00021242, -0.00033190, 0.00009148, -0.00012191, 0.00006731, -0.00001053, -0.00009372, 0.00017887, -0.00009436, 0.00017453, 0.00000804, -0.00013970, 0.00011995, -0.00003527, -0.00004890, -0.00017245, -0.00011058, 0.00016862, -0.00019224, 0.00025659, -0.00130758, 0.00237295, -0.00054801, -0.00038505, 0.00000731, -0.00003763, -0.00059175, 0.00114206, -0.00029852, },
		{ 0.00047796, -0.00034302, 0.00029151, -0.00032704, 0.00006936, -0.00009801, 0.00002325, 0.00003749, -0.00012151, 0.00015931, -0.00006308, 0.00011518, 0.00007090, -0.00014040, 0.00004922, 0.00007607, -0.00015666, -0.00000590, -0.00031024, 0.00035061, -0.00040987, 0.00056445, -0.00154460, 0.00200807, 0.00022207, -0.00082662, 0.00017233, -0.00009299, -0.00048493, 0.00070637, 0.00022290, -0.00060584, },
		{ -0.00035189, 0.00026759, -0.00025165, 0.00001112, -0.00005449, -0.00002141, 0.00006570, -0.00011653, 0.00011119, -0.00001323, 0.00004940, 0.00011609, -0.00011377, -0.00003425, 0.00017174, -0.00022734, 0.00009696, -0.00040226, 0.00039849, -0.00046504, 0.00063533, -0.00142348, 0.00128284, 0.00112968, -0.00122655, 0.00030171, -0.00018546, -0.00028268, 0.00023425, 0.00065638, -0.00081896, 0.00054682, },
		{ 0.00015316, -0.00012704, -0.00006449, -0.00000564, -0.00005625, 0.00007169, -0.00008687, 0.00004993, 0.00003838, -0.00000751, 0.00013750, -0.00007069, -0.00010478, 0.00021921, -0.00022645, 0.00010123, -0.00036144, 0.00030200, -0.00035636, 0.00048481, -0.00101612, 0.00035991, 0.00197194, -0.00142936, 0.00029814, -0.00023093, -0.00009549, -0.00014337, 0.00089626, -0.00082978, 0.00045061, -0.00023588, },
		{ 0.00000986, -0.00013299, 0.00003428, -0.00007521, 0.00005981, -0.00004626, -0.00000705, 0.00007655, -0.00004500, 0.00013547, -0.00002540, -0.00014174, 0.00020142, -0.00014525, 0.00000906, -0.00020991, 0.00010367, -0.00014003, 0.00019485, -0.00046428, -0.00055020, 0.00254302, -0.00130959, 0.00009747, -0.00017026, -0.00000534, -0.00034263, 0.00089899, -0.00063964, 0.00021517, -0.00003191, -0.00001060, },
		{ -0.00017244, 0.00005553, -0.00007759, 0.00003870, -0.00000911, -0.00004593, 0.00009227, -0.00005955, 0.00011573, 0.00000928, -0.00013626, 0.00012287, -0.00000436, -0.00014137, -0.00000906, -0.00011715, 0.00009325, -0.00011843, 0.00006630, -0.00124740, 0.00269302, -0.00081835, -0.00029757, 0.00000197, -0.00003801, -0.00034860, 0.00069329, -0.00031022, -0.00008984, 0.00019430, -0.00016739, 0.00011845, },
		{ 0.00005564, -0.00006716, 0.00001768, 0.00001430, -0.00006057, 0.00008487, -0.00005463, 0.00008683, 0.00002600, -0.00009395, 0.00000840, 0.00015295, -0.00028936, 0.00016216, -0.00027311, 0.00025281, -0.00034615, 0.00043622, -0.00159643, 0.00237257, -0.00000911, -0.00080563, 0.00023014, -0.00015799, -0.00021298, 0.00036697, 0.00005706, -0.00037366, 0.00036982, -0.00026416, 0.00016763, -0.00017086, },
		{ -0.00005000, 0.00000336, 0.00002081, -0.00005370, 0.00006133, -0.00003842, 0.00005704, 0.00002468, -0.00003266, -0.00010500, 0.00027538, -0.00037312, 0.00023504, -0.00030066, 0.00028148, -0.00042289, 0.00057146, -0.00156072, 0.00164904, 0.00096715, -0.00128283, 0.00041516, -0.00028709, -0.00002945, 0.00003550, 0.00035409, -0.00055390, 0.00043842, -0.00026843, 0.00014593, -0.00013013, 0.00003947, },
		{ -0.00000235, 0.00001461, -0.00003502, 0.00003293, -0.00002031, 0.00003186, 0.00001187, 0.00002378, -0.00018051, 0.00032125, -0.00035179, 0.00017559, -0.00018362, 0.00017648, -0.00034579, 0.00047903, -0.00120603, 0.00068894, 0.00189772, -0.00156331, 0.00045108, -0.00033927, 0.00010388, -0.00019705, 0.00050052, -0.00058301, 0.00037867, -0.00017862, 0.00006506, -0.00006526, 0.00001680, -0.00003182, },
		{ 0.00000483, -0.00001696, 0.00001074, -0.00000739, 0.00001285, -0.00000236, 0.00005382, -0.00019399, 0.00027316, -0.00022035, -0.00000280, 0.00003981, -0.00000981, -0.00017049, 0.00023558, -0.00067383, -0.00028829, 0.00256502, -0.00151175, 0.00026809, -0.00025853, 0.00012151, -0.00027129, 0.00046843, -0.00046289, 0.00021045, -0.00002418, -0.00004417, 0.00000092, -0.00000169, -0.00001583, -0.00000230, },
		{ -0.00000965, 0.00000139, -0.00000216, -0.00000187, -0.00000855, 0.00004558, -0.00014221, 0.00014454, -0.00001237, -0.00024301, 0.00028905, -0.00018933, 0.00001395, -0.00004584, -0.00013389, -0.00107096, 0.00280732, -0.00107096, -0.00013389, -0.00004584, 0.00001395, -0.00018933, 0.00028905, -0.00024301, -0.00001237, 0.00014454, -0.00014221, 0.00004558, -0.00000855, -0.00000187, -0.00000216, 0.00000139, },
		{ 0.00000483, -0.00000230, -0.00001583, -0.00000169, 0.00000092, -0.00004417, -0.00002418, 0.00021045, -0.00046289, 0.00046843, -0.00027129, 0.00012151, -0.00025853, 0.00026809, -0.00151175, 0.00256502, -0.00028829, -0.00067383, 0.00023558, -0.00017049, -0.00000981, 0.00003981, -0.00000280, -0.00022035, 0.00027316, -0.00019399, 0.00005382, -0.00000236, 0.00001285, -0.00000739, 0.00001074, -0.00001696, },
		{ -0.00000235, -0.00003182, 0.00001680, -0.00006526, 0.00006506, -0.00017862, 0.00037867, -0.00058301, 0.00050052, -0.00019705, 0.00010388, -0.00033927, 0.00045108, -0.00156331, 0.00189772, 0.00068894, -0.00120603, 0.00047903, -0.00034579, 0.00017648, -0.00018362, 0.00017559, -0.00035179, 0.00032125, -0.00018051, 0.00002378, 0.00001187, 0.00003186, -0.00002031, 0.00003293, -0.00003502, 0.00001461, },
		{ -0.00005000, 0.00003947, -0.00013013, 0.00014593, -0.00026843, 0.00043842, -0.00055390, 0.00035409, 0.00003550, -0.00002945, -0.00028709, 0.00041516, -0.00128283, 0.00096715, 0.00164904, -0.00156072, 0.00057146, -0.00042289, 0.00028148, -0.00030066, 0.00023504, -0.00037312, 0.00027538, -0.00010500, -0.00003266, 0.00002468, 0.00005704, -0.00003842, 0.00006133, -0.00005370, 0.00002081, 0.00000336, },
		{ 0.00005564, -0.00017086, 0.00016763, -0.00026416, 0.00036982, -0.00037366, 0.00005706, 0.00036697, -0.00021298, -0.00015799, 0.00023014, -0.00080563, -0.00000911, 0.00237257, -0.00159643, 0.00043622, -0.00034615, 0.00025281, -0.00027311, 0.00016216, -0.00028936, 0.00015295, 0.00000840, -0.00009395, 0.00002600, 0.00008683, -0.00005463, 0.00008487, -0.00006057, 0.00001430, 0.00001768, -0.00006716, },
		{ -0.00017244, 0.00011845, -0.00016739, 0.00019430, -0.00008984, -0.00031022, 0.00069329, -0.00034860, -0.00003801, 0.00000197, -0.00029757, -0.00081835, 0.00269302, -0.00124740, 0.00006630, -0.00011843, 0.00009325, -0.00011715, -0.00000906, -0.00014137, -0.00000436, 0.00012287, -0.00013626, 0.00000928, 0.00011573, -0.00005955, 0.00009227, -0.00004593, -0.00000911, 0.00003870, -0.00007759, 0.00005553, },
		{ 0.00000986, -0.00001060, -0.00003191, 0.00021517, -0.00063964, 0.00089899, -0.00034263, -0.00000534, -0.00017026, 0.00009747, -0.00130959, 0.00254302, -0.00055020, -0.00046428, 0.00019485, -0.00014003, 0.00010367, -0.00020991, 0.00000906, -0.00014525, 0.00020142, -0.00014174, -0.00002540, 0.00013547, -0.00004500, 0.00007655, -0.00000705, -0.00004626, 0.00005981, -0.00007521, 0.00003428, -0.00013299, },
		{ 0.00015316, -0.00023588, 0.00045061, -0.00082978, 0.00089626, -0.00014337, -0.00009549, -0.00023093, 0.00029814, -0.00142936, 0.00197194, 0.00035991, -0.00101612, 0.00048481, -0.00035636, 0.00030200, -0.00036144, 0.00010123, -0.00022645, 0.00021921, -0.00010478, -0.00007069, 0.00013750, -0.00000751, 0.00003838, 0.00004993, -0.00008687, 0.00007169, -0.00005625, -0.00000564, -0.00006449, -0.00012704, },
		{ -0.00035189, 0.00054682, -0.00081896, 0.00065638, 0.00023425, -0.00028268, -0.00018546, 0.00030171, -0.00122655, 0.00112968, 0.00128284, -0.00142348, 0.00063533, -0.00046504, 0.00039849, -0.00040226, 0.00009696, -0.00022734, 0.00017174, -0.00003425, -0.00011377, 0.00011609, 0.00004940, -0.00001323, 0.00011119, -0.00011653, 0.00006570, -0.00002141, -0.00005449, 0.00001112, -0.00025165, 0.00026759, },
		{ 0.00047796, -0.00060584, 0.00022290, 0.00070637, -0.00048493, -0.00009299, 0.00017233, -0.00082662, 0.00022207, 0.00200807, -0.00154460, 0.00056445, -0.00040987, 0.00035061, -0.00031024, -0.00000590, -0.00015666, 0.00007607, 0.00004922, -0.00014040, 0.00007090, 0.00011518, -0.00006308, 0.00015931, -0.00012151, 0.00003749, 0.00002325, -0.00009801, 0.00006936, -0.00032704, 0.00029151, -0.00034302, },
		{ -0.00025071, -0.00029852, 0.00114206, -0.00059175, -0.00003763, 0.00000731, -0.00038505, -0.00054801, 0.00237295, -0.00130758, 0.00025659, -0.00019224, 0.00016862, -0.00011058, -0.00017245, -0.00004890, -0.00003527, 0.00011995, -0.00013970, 0.00000804, 0.00017453, -0.00009436, 0.00017887, -0.00009372, -0.00001053, 0.00006731, -0.00012191, 0.00009148, -0.00033190, 0.00021242, -0.00021342, 0.00027042, },
		{ -0.00077076, 0.00140304, -0.00050511, -0.00008933, -0.00010527, -0.00003620, -0.00103550, 0.00230730, -0.00073646, -0.00022597, 0.00012562, -0.00008704, 0.00013249, -0.00034266, 0.00004814, -0.00012557, 0.00015566, -0.00010814, -0.00006072, 0.00021161, -0.00009408, 0.00016175, -0.00003417, -0.00006887, 0.00009866, -0.00011662, 0.00007085, -0.00026727, 0.00005061, -0.00000699, -0.00000588, 0.00014319, },
		{ 0.00138836, -0.00017959, -0.00026934, -0.00012173, 0.00014552, -0.00118740, 0.00185049, 0.00005131, -0.00075717, 0.00043879, -0.00032618, 0.00033766, -0.00045212, 0.00009052, -0.00016653, 0.00014495, -0.00005127, -0.00012080, 0.00021467, -0.00005768, 0.00011033, 0.00004633, -0.00012297, 0.00010698, -0.00008086, 0.00001540, -0.00015588, -0.00014745, 0.00020706, -0.00026049, 0.00046217, -0.00106675, },
		{ 0.00035295, -0.00053369, -0.00005816, 0.00016022, -0.00104711, 0.00113643, 0.00087585, -0.00118309, 0.00063384, -0.00046018, 0.00043450, -0.00045465, 0.00005404, -0.00014792, 0.00009108, 0.00001738, -0.00015854, 0.00017981, 0.00000892, 0.00003739, 0.00012969, -0.00015715, 0.00008713, -0.00002256, -0.00005503, -0.00003426, -0.00032427, 0.00035689, -0.00041202, 0.00061787, -0.00110857, 0.00106952, },
		{ -0.00078181, 0.00001940, 0.00006974, -0.00072921, 0.00035160, 0.00154525, -0.00136874, 0.00062743, -0.00043353, 0.00038817, -0.00033889, -0.00005669, -0.00008097, 0.00001105, 0.00008003, -0.00016511, 0.00011262, 0.00009039, -0.00003729, 0.00019551, -0.00015929, 0.00004116, 0.00004307, -0.00011520, 0.00005954, -0.00042989, 0.00039157, -0.00041451, 0.00057459, -0.00089142, 0.00050390, 0.00098385, },
		{ 0.00002588, -0.00003622, -0.00037471, -0.00032016, 0.00190954, -0.00124063, 0.00039656, -0.00024465, 0.00021204, -0.00013324, -0.00020902, 0.00000561, -0.00006997, 0.00011991, -0.00013906, 0.00002706, 0.00016614, -0.00009278, 0.00022703, -0.00012474, -0.00002165, 0.00009791, -0.00014301, 0.00009858, -0.00043781, 0.00029755, -0.00027190, 0.00035930, -0.00048475, -0.00017853, 0.00155446, -0.00088845, },
		},
		{ { -0.00007011, -0.00004001, -0.00002523, -0.00073447, 0.00177550, -0.00079723, 0.00012433, -0.00015655, -0.00008122, 0.00018732, -0.00011554, 0.00009298, -0.00021640, 0.00002656, 0.00002723, 0.00006192, -0.00010193, 0.00006192, 0.00002723, 0.00002656, -0.00021640, 0.00009298, -0.00011554, 0.00018732, -0.00008122, -0.00015655, 0.00012433, -0.00079723, 0.00177550, -0.00073447, -0.00002523, -0.00004001, },
		{ -0.00010308, 0.00018952, -0.00090911, 0.00135704, -0.00003850, -0.00044844, 0.00027581, -0.00042962, 0.00039950, -0.00024107, 0.00019132, -0.00024038, 0.00001646, -0.00000311, 0.00008890, -0.00007470, -0.00000564, 0.00007828, 0.00000391, -0.00013106, -0.00006723, 0.00006559, -0.00007372, 0.00027069, -0.00052166, 0.00053430, -0.00119058, 0.00163839, -0.00010632, -0.00045285, 0.00012998, -0.00011628, },
		{ 0.00018200, -0.00070148, 0.00061172, 0.00077268, -0.00089878, 0.00054469, -0.00059343, 0.00043271, -0.00022215, 0.00016990, -0.00017313, -0.00004270, 0.00000323, 0.00006838, -0.00000817, -0.00008047, 0.00011331, -0.00001276, -0.00004285, -0.00020198, 0.00018977, -0.00024156, 0.00045573, -0.00063421, 0.00059316, -0.00107697, 0.00095083, 0.00081857, -0.00091551, 0.00029481, -0.00020256, -0.00006645, },
		{ -0.00033192, -0.00012638, 0.00130573, -0.00099066, 0.00049804, -0.00048104, 0.00024558, -0.00004948, 0.00003805, -0.00003654, -0.00013056, 0.00003236, 0.00002166, 0.00006101, -0.00012098, 0.00009850, 0.00000889, -0.00000036, -0.00024664, 0.00018656, -0.00024072, 0.00040342, -0.00044938, 0.00030135, -0.00055384, -0.00003231, 0.00169547, -0.00113272, 0.00028204, -0.00021427, -0.00002513, 0.00007060, },
		{ -0.00058520, 0.00137029, -0.00065946, 0.00014712, -0.00014451, -0.00008451, 0.00019483, -0.00012836, 0.00010354, -0.00019980, 0.00005110, -0.00001665, 0.00009569, -0.00010144, 0.00002653, 0.00007037, -0.00001247, -0.00019473, 0.00006518, -0.00009837, 0.00017287, -0.00007830, -0.00016233, 0.00009666, -0.00091804, 0.00214243, -0.00086898, -0.00002131, -0.00007138, -0.00007023, -0.00000635, -0.00002788, },
		{ 0.00099656, -0.00004464, -0.00033595, 0.00024419, -0.00039761, 0.00037840, -0.00022749, 0.00017310, -0.00020457, 0.00003032, -0.00002098, 0.00007945, -0.00002801, -0.00007468, 0.00013799, -0.00004565, -0.00009641, -0.00009677, 0.00008046, -0.00009061, 0.00028113, -0.00054408, 0.00055468, -0.00136037, 0.00192841, -0.00008450, -0.00056236, 0.00018793, -0.00020541, 0.00000947, 0.00009583, -0.00067153, },
		{ 0.00058362, -0.00070742, 0.00048339, -0.00053526, 0.00039127, -0.00018996, 0.00013342, -0.00013063, -0.00003381, 0.00000930, 0.00002551, 0.00006318, -0.00015459, 0.00016244, -0.00004932, -0.00002248, -0.00020430, 0.00018098, -0.00024357, 0.00045581, -0.00065242, 0.00062309, -0.00122385, 0.00109943, 0.00100106, -0.00110416, 0.00040098, -0.00033216, 0.00006731, 0.00007075, -0.00048156, 0.00040024, },
		{ -0.00077940, 0.00044936, -0.00042630, 0.00020931, -0.00002165, 0.00000648, -0.00000797, -0.00011424, 0.00004969, -0.00003171, 0.00012524, -0.00016760, 0.00011071, 0.00000701, -0.00001739, -0.00020381, 0.00014791, -0.00022441, 0.00038746, -0.00045360, 0.00031699, -0.00064053, -0.00002841, 0.00197956, -0.00132266, 0.00036778, -0.00031729, 0.00006198, 0.00001490, -0.00021178, -0.00014182, 0.00096248, },
		{ 0.00016190, -0.00012652, -0.00008801, 0.00019500, -0.00013404, 0.00010273, -0.00016738, 0.00006647, -0.00005733, 0.00012640, -0.00010001, -0.00000902, 0.00010879, -0.00007076, -0.00011380, 0.00001494, -0.00007978, 0.00015217, -0.00007584, -0.00016171, 0.00006750, -0.00101096, 0.00243516, -0.00097730, -0.00001400, -0.00009937, -0.00007062, 0.00002841, -0.00003147, -0.00043318, 0.00096421, -0.00051628, },
		{ 0.00021047, -0.00035702, 0.00034339, -0.00020243, 0.00014331, -0.00015913, 0.00003864, -0.00003676, 0.00006776, 0.00001973, -0.00014592, 0.00020045, -0.00012514, -0.00000941, -0.00012199, 0.00007486, -0.00009673, 0.00027407, -0.00054494, 0.00055648, -0.00148280, 0.00214610, -0.00006379, -0.00064898, 0.00023946, -0.00028668, 0.00011658, 0.00000651, -0.00043983, 0.00064643, -0.00004845, -0.00022793, },
		{ -0.00046527, 0.00033639, -0.00014976, 0.00009216, -0.00008699, -0.00002741, 0.00001561, -0.00001724, 0.00013415, -0.00023105, 0.00021742, -0.00011548, 0.00002852, -0.00016909, 0.00013348, -0.00022451, 0.00042964, -0.00064396, 0.00062978, -0.00132336, 0.00120191, 0.00114128, -0.00124977, 0.00049284, -0.00044694, 0.00019020, -0.00003113, -0.00027462, 0.00020314, 0.00040179, -0.00052062, 0.00041523, },
		{ 0.00016558, 0.00000623, -0.00002119, 0.00001554, -0.00009980, 0.00006507, -0.00008164, 0.00018681, -0.00021510, 0.00012755, -0.00001541, -0.00003214, -0.00009646, 0.00006703, -0.00019092, 0.00035004, -0.00043980, 0.00031968, -0.00069845, -0.00002879, 0.00218130, -0.00146183, 0.00044163, -0.00040598, 0.00013940, -0.00003249, -0.00010378, -0.00015019, 0.00063894, -0.00057473, 0.00039196, -0.00036175, },
		{ 0.00018779, -0.00013036, 0.00009590, -0.00013890, 0.00007738, -0.00009133, 0.00015272, -0.00009777, -0.00004332, 0.00013905, -0.00015139, 0.00004179, -0.00006109, -0.00006239, 0.00012617, -0.00007386, -0.00015477, 0.00003994, -0.00106709, 0.00262394, -0.00104997, -0.00000169, -0.00012309, -0.00007131, 0.00006297, -0.00003784, -0.00028958, 0.00059136, -0.00037848, 0.00016620, -0.00010312, -0.00009154, },
		{ -0.00016839, 0.00011175, -0.00012077, 0.00003947, -0.00004404, 0.00005200, 0.00006858, -0.00021867, 0.00026479, -0.00024015, 0.00014157, -0.00014026, 0.00004316, -0.00008959, 0.00024788, -0.00052218, 0.00053811, -0.00154451, 0.00226811, -0.00004665, -0.00070357, 0.00028034, -0.00035506, 0.00021209, -0.00007348, -0.00023155, 0.00033508, -0.00004816, -0.00013437, 0.00017822, -0.00031130, 0.00029787, },
		{ 0.00005571, -0.00005220, -0.00003209, 0.00002984, -0.00006104, 0.00020450, -0.00030958, 0.00027880, -0.00021671, 0.00011648, -0.00008938, 0.00004118, -0.00018232, 0.00037572, -0.00060650, 0.00061032, -0.00136366, 0.00124600, 0.00122552, -0.00133883, 0.00056405, -0.00054002, 0.00029483, -0.00011705, -0.00009664, 0.00003630, 0.00024253, -0.00035349, 0.00034563, -0.00038863, 0.00027282, -0.00010575, },
		{ 0.00003175, -0.00010010, 0.00008494, -0.00012756, 0.00024488, -0.00026333, 0.00015009, -0.00006238, -0.00004273, 0.00008360, -0.00006002, -0.00014019, 0.00029104, -0.00040681, 0.00030755, -0.00072050, -0.00003475, 0.00228064, -0.00153781, 0.00049936, -0.00047538, 0.00020246, -0.00006802, -0.00001693, -0.00014986, 0.00035985, -0.00039157, 0.00032941, -0.00029143, 0.00011778, 0.00003135, -0.00003901, },
		{ -0.00012782, 0.00008691, -0.00011612, 0.00017349, -0.00009487, -0.00007507, 0.00015861, -0.00025486, 0.00027843, -0.00016351, -0.00004810, 0.00009601, -0.00007237, -0.00014175, 0.00001640, -0.00108070, 0.00268915, -0.00108070, 0.00001640, -0.00014175, -0.00007237, 0.00009601, -0.00004810, -0.00016351, 0.00027843, -0.00025486, 0.00015861, -0.00007507, -0.00009487, 0.00017349, -0.00011612, 0.00008691, },
		{ 0.00003175, -0.00003901, 0.00003135, 0.00011778, -0.00029143, 0.00032941, -0.00039157, 0.00035985, -0.00014986, -0.00001693, -0.00006802, 0.00020246, -0.00047538, 0.00049936, -0.00153781, 0.00228064, -0.00003475, -0.00072050, 0.00030755, -0.00040681, 0.00029104, -0.00014019, -0.00006002, 0.00008360, -0.00004273, -0.00006238, 0.00015009, -0.00026333, 0.00024488, -0.00012756, 0.00008494, -0.00010010, },
		{ 0.00005571, -0.00010575, 0.00027282, -0.00038863, 0.00034563, -0.00035349, 0.00024253, 0.00003630, -0.00009664, -0.00011705, 0.00029483, -0.00054002, 0.00056405, -0.00133883, 0.00122552, 0.00124600, -0.00136366, 0.00061032, -0.00060650, 0.00037572, -0.00018232, 0.00004118, -0.00008938, 0.00011648, -0.00021671, 0.00027880, -0.00030958, 0.00020450, -0.00006104, 0.00002984, -0.00003209, -0.00005220, },
		{ -0.00016839, 0.00029787, -0.00031130, 0.00017822, -0.00013437, -0.00004816, 0.00033508, -0.00023155, -0.00007348, 0.00021209, -0.00035506, 0.00028034, -0.00070357, -0.00004665, 0.00226811, -0.00154451, 0.00053811, -0.00052218, 0.00024788, -0.00008959, 0.00004316, -0.00014026, 0.00014157, -0.00024015, 0.00026479, -0.00021867, 0.00006858, 0.00005200, -0.00004404, 0.00003947, -0.00012077, 0.00011175, },
		{ 0.00018779, -0.00009154, -0.00010312, 0.00016620, -0.00037848, 0.00059136, -0.00028958, -0.00003784, 0.00006297, -0.00007131, -0.00012309, -0.00000169, -0.00104997, 0.00262394, -0.00106709, 0.00003994, -0.00015477, -0.00007386, 0.00012617, -0.00006239, -0.00006109, 0.00004179, -0.00015139, 0.00013905, -0.00004332, -0.00009777, 0.00015272, -0.00009133, 0.00007738, -0.00013890, 0.00009590, -0.00013036, },
		{ 0.00016558, -0.00036175, 0.00039196, -0.00057473, 0.00063894, -0.00015019, -0.00010378, -0.00003249, 0.00013940, -0.00040598, 0.00044163, -0.00146183, 0.00218130, -0.00002879, -0.00069845, 0.00031968, -0.00043980, 0.00035004, -0.00019092, 0.00006703, -0.00009646, -0.00003214, -0.00001541, 0.00012755, -0.00021510, 0.00018681, -0.00008164, 0.00006507, -0.00009980, 0.00001554, -0.00002119, 0.00000623, },
		{ -0.00046527, 0.00041523, -0.00052062, 0.00040179, 0.00020314, -0.00027462, -0.00003113, 0.00019020, -0.00044694, 0.00049284, -0.00124977, 0.00114128, 0.00120191, -0.00132336, 0.00062978, -0.00064396, 0.00042964, -0.00022451, 0.00013348, -0.00016909, 0.00002852, -0.00011548, 0.00021742, -0.00023105, 0.00013415, -0.00001724, 0.00001561, -0.00002741, -0.00008699, 0.00009216, -0.00014976, 0.00033639, },
		{ 0.00021047, -0.00022793, -0.00004845, 0.00064643, -0.00043983, 0.00000651, 0.00011658, -0.00028668, 0.00023946, -0.00064898, -0.00006379, 0.00214610, -0.00148280, 0.00055648, -0.00054494, 0.00027407, -0.00009673, 0.00007486, -0.00012199, -0.00000941, -0.00012514, 0.00020045, -0.00014592, 0.00001973, 0.00006776, -0.00003676, 0.00003864, -0.00015913, 0.00014331, -0.00020243, 0.00034339, -0.00035702, },
		{ 0.00016190, -0.00051628, 0.00096421, -0.00043318, -0.00003147, 0.00002841, -0.00007062, -0.00009937, -0.00001400, -0.00097730, 0.00243516, -0.00101096, 0.00006750, -0.00016171, -0.00007584, 0.00015217, -0.00007978, 0.00001494, -0.00011380, -0.00007076, 0.00010879, -0.00000902, -0.00010001, 0.00012640, -0.00005733, 0.00006647, -0.00016738, 0.00010273, -0.00013404, 0.00019500, -0.00008801, -0.00012652, },
		{ -0.00077940, 0.00096248, -0.00014182, -0.00021178, 0.00001490, 0.00006198, -0.00031729, 0.00036778, -0.00132266, 0.00197956, -0.00002841, -0.00064053, 0.00031699, -0.00045360, 0.00038746, -0.00022441, 0.00014791, -0.00020381, -0.00001739, 0.00000701, 0.00011071, -0.00016760, 0.00012524, -0.00003171, 0.00004969, -0.00011424, -0.00000797, 0.00000648, -0.00002165, 0.00020931, -0.00042630, 0.00044936, },
		{ 0.00058362, 0.00040024, -0.00048156, 0.00007075, 0.00006731, -0.00033216, 0.00040098, -0.00110416, 0.00100106, 0.00109943, -0.00122385, 0.00062309, -0.00065242, 0.00045581, -0.00024357, 0.00018098, -0.00020430, -0.00002248, -0.00004932, 0.00016244, -0.00015459, 0.00006318, 0.00002551, 0.00000930, -0.00003381, -0.00013063, 0.00013342, -0.00018996, 0.00039127, -0.00053526, 0.00048339, -0.00070742, },
		{ 0.00099656, -0.00067153, 0.00009583, 0.00000947, -0.00020541, 0.00018793, -0.00056236, -0.00008450, 0.00192841, -0.00136037, 0.00055468, -0.00054408, 0.00028113, -0.00009061, 0.00008046, -0.00009677, -0.00009641, -0.00004565, 0.00013799, -0.00007468, -0.00002801, 0.00007945, -0.00002098, 0.00003032, -0.00020457, 0.00017310, -0.00022749, 0.00037840, -0.00039761, 0.00024419, -0.00033595, -0.00004464, },
		{ -0.00058520, -0.00002788, -0.00000635, -0.00007023, -0.00007138, -0.00002131, -0.00086898, 0.00214243, -0.00091804, 0.00009666, -0.00016233, -0.00007830, 0.00017287, -0.00009837, 0.00006518, -0.00019473, -0.00001247, 0.00007037, 0.00002653, -0.00010144, 0.00009569, -0.00001665, 0.00005110, -0.00019980, 0.00010354, -0.00012836, 0.00019483, -0.00008451, -0.00014451, 0.00014712, -0.00065946, 0.00137029, },
		{ -0.00033192, 0.00007060, -0.00002513, -0.00021427, 0.00028204, -0.00113272, 0.00169547, -0.00003231, -0.00055384, 0.00030135, -0.00044938, 0.00040342, -0.00024072, 0.00018656, -0.00024664, -0.00000036, 0.00000889, 0.00009850, -0.00012098, 0.00006101, 0.00002166, 0.00003236, -0.00013056, -0.00003654, 0.00003805, -0.00004948, 0.00024558, -0.00048104, 0.00049804, -0.00099066, 0.00130573, -0.00012638, },
		{ 0.00018200, -0.00006645, -0.00020256, 0.00029481, -0.00091551, 0.00081857, 0.00095083, -0.00107697, 0.00059316, -0.00063421, 0.00045573, -0.00024156, 0.00018977, -0.00020198, -0.00004285, -0.00001276, 0.00011331, -0.00008047, -0.00000817, 0.00006838, 0.00000323, -0.00004270, -0.00017313, 0.00016990, -0.00022215, 0.00043271, -0.00059343, 0.00054469, -0.00089878, 0.00077268, 0.00061172, -0.00070148, },
		{ -0.00010308, -0.00011628, 0.00012998, -0.00045285, -0.00010632, 0.00163839, -0.00119058, 0.00053430, -0.00052166, 0.00027069, -0.00007372, 0.00006559, -0.00006723, -0.00013106, 0.00000391, 0.00007828, -0.00000564, -0.00007470, 0.00008890, -0.00000311, 0.00001646, -0.00024038, 0.00019132, -0.00024107, 0.00039950, -0.00042962, 0.00027581, -0.00044844, -0.00003850, 0.00135704, -0.00090911, 0.00018952, },
		},
		{ { -0.00001017, 0.00004561, -0.00009951, -0.00014340, -0.00071991, 0.00185221, -0.00085470, 0.00024336, -0.00009801, -0.00013106, 0.00004689, -0.00003774, 0.00003477, 0.00000682, 0.00022757, -0.00043316, 0.00021697, -0.00043316, 0.00022757, 0.00000682, 0.00003477, -0.00003774, 0.00004689, -0.00013106, -0.00009801, 0.00024336, -0.00085470, 0.00185221, -0.00071991, -0.00014340, -0.00009951, 0.00004561, },
		{ 0.00006281, -0.00016195, 0.00012616, -0.00096416, 0.00132208, 0.00012167, -0.00052263, 0.00047088, -0.00050948, 0.00028348, -0.00023781, 0.00022085, -0.00020330, 0.00039338, -0.00040272, 0.00009424, -0.00021942, -0.00011424, 0.00028950, -0.00018409, 0.00019554, -0.00022932, 0.00025551, -0.00056121, 0.00071375, -0.00121182, 0.00150389, 0.00019267, -0.00069251, 0.00010930, -0.00010366, 0.00004894, },
		{ -0.00006738, 0.00007925, -0.00065179, 0.00033436, 0.00111326, -0.00101606, 0.00069188, -0.00054656, 0.00024426, -0.00022032, 0.00020894, -0.00019169, 0.00029197, -0.00013649, -0.00017669, 0.00006500, -0.00040255, 0.00041534, -0.00022805, 0.00024956, -0.00029504, 0.00034591, -0.00058335, 0.00057764, -0.00077923, 0.00041166, 0.00137694, -0.00113398, 0.00022077, -0.00023653, 0.00013624, -0.00001095, },
		{ -0.00006222, -0.00020101, -0.00048656, 0.00152295, -0.00084349, 0.00036064, -0.00016836, -0.00007467, -0.00000019, 0.00001876, -0.00000253, 0.00003299, 0.00018268, -0.00038189, 0.00020228, -0.00044674, 0.00026347, -0.00003340, 0.00007731, -0.00010327, 0.00010899, -0.00019978, 0.00000029, 0.00006155, -0.00078053, 0.00212270, -0.00097151, -0.00003987, -0.00016237, 0.00008908, -0.00002317, 0.00002418, },
		{ 0.00004185, -0.00076098, 0.00117526, -0.00008766, -0.00029913, 0.00035672, -0.00042229, 0.00021579, -0.00017066, 0.00018047, -0.00017202, 0.00034192, -0.00036046, 0.00010112, -0.00024248, -0.00007381, 0.00025476, -0.00016031, 0.00016184, -0.00021326, 0.00024349, -0.00053740, 0.00067676, -0.00135813, 0.00189721, -0.00002271, -0.00066600, 0.00011280, -0.00012063, 0.00009035, -0.00000121, -0.00009489, },
		{ -0.00055273, 0.00039265, 0.00075944, -0.00079885, 0.00062157, -0.00049603, 0.00021224, -0.00018804, 0.00020035, -0.00018196, 0.00025997, -0.00012761, -0.00013791, 0.00003165, -0.00036169, 0.00040000, -0.00023989, 0.00026125, -0.00033380, 0.00039434, -0.00063446, 0.00066005, -0.00104198, 0.00076387, 0.00132804, -0.00124772, 0.00031143, -0.00030955, 0.00020930, -0.00008076, -0.00001785, 0.00003099, },
		{ -0.00029523, 0.00117028, -0.00076189, 0.00040857, -0.00020204, -0.00003669, -0.00002630, 0.00005783, -0.00002750, 0.00003660, 0.00014572, -0.00031214, 0.00016329, -0.00041068, 0.00027435, -0.00007738, 0.00011718, -0.00016792, 0.00017766, -0.00027264, 0.00011373, -0.00017086, -0.00061669, 0.00230362, -0.00121329, 0.00009808, -0.00024596, 0.00014983, -0.00006019, 0.00002618, -0.00005156, -0.00020763, },
		{ 0.00095863, -0.00020977, -0.00012971, 0.00025228, -0.00033243, 0.00015204, -0.00010372, 0.00013586, -0.00014232, 0.00027361, -0.00028320, 0.00007532, -0.00022414, -0.00003270, 0.00019510, -0.00012447, 0.00011290, -0.00017424, 0.00020905, -0.00047749, 0.00057803, -0.00143190, 0.00225554, -0.00029176, -0.00057166, 0.00008494, -0.00011487, 0.00010748, -0.00004755, -0.00004303, -0.00001989, -0.00054963, },
		{ 0.00046881, -0.00058896, 0.00051894, -0.00041671, 0.00016485, -0.00013787, 0.00017170, -0.00015692, 0.00019683, -0.00008157, -0.00012436, 0.00001701, -0.00029062, 0.00034572, -0.00023523, 0.00025346, -0.00034832, 0.00042185, -0.00065596, 0.00070392, -0.00128157, 0.00114975, 0.00117587, -0.00128849, 0.00038401, -0.00037182, 0.00027217, -0.00015042, 0.00003706, -0.00002498, -0.00041101, 0.00036919, },
		{ -0.00063283, 0.00039169, -0.00019565, -0.00002023, -0.00002884, 0.00007382, -0.00003502, 0.00001235, 0.00013151, -0.00024700, 0.00011426, -0.00032250, 0.00024790, -0.00011481, 0.00014691, -0.00022184, 0.00024324, -0.00034046, 0.00022733, -0.00043244, -0.00037087, 0.00237302, -0.00141675, 0.00025411, -0.00034216, 0.00022046, -0.00011554, 0.00004954, -0.00006479, -0.00016522, -0.00016218, 0.00082930, },
		{ -0.00002664, 0.00017223, -0.00025023, 0.00010056, -0.00004705, 0.00009534, -0.00012262, 0.00020378, -0.00018626, 0.00001632, -0.00014858, -0.00000988, 0.00012184, -0.00008307, 0.00005641, -0.00011798, 0.00015624, -0.00038846, 0.00042298, -0.00141857, 0.00254255, -0.00058669, -0.00041740, 0.00002535, -0.00008624, 0.00009818, -0.00007067, -0.00001160, -0.00005066, -0.00035418, 0.00070137, -0.00025005, },
		{ 0.00040312, -0.00031918, 0.00010955, -0.00007799, 0.00012825, -0.00012143, 0.00011100, -0.00000045, -0.00015112, 0.00004435, -0.00020686, 0.00025747, -0.00021387, 0.00022710, -0.00033657, 0.00042429, -0.00064458, 0.00070005, -0.00147039, 0.00153408, 0.00093341, -0.00124781, 0.00042672, -0.00041540, 0.00031637, -0.00020987, 0.00008899, -0.00007779, -0.00024675, 0.00027629, 0.00025702, -0.00041168, },
		{ -0.00015155, -0.00002507, -0.00000887, 0.00006467, -0.00002274, -0.00004085, 0.00015018, -0.00020945, 0.00007621, -0.00018985, 0.00017781, -0.00013705, 0.00016067, -0.00025668, 0.00029626, -0.00039441, 0.00032608, -0.00069695, -0.00006218, 0.00232094, -0.00155699, 0.00040939, -0.00044069, 0.00029192, -0.00018072, 0.00008911, -0.00009597, -0.00008238, -0.00009424, 0.00053022, -0.00048324, 0.00032273, },
		{ -0.00018428, 0.00006759, -0.00000923, 0.00006679, -0.00012028, 0.00014799, -0.00009047, -0.00006629, -0.00000903, -0.00002240, 0.00004843, -0.00004392, 0.00000148, -0.00005290, 0.00009179, -0.00028092, 0.00022489, -0.00131438, 0.00272809, -0.00087568, -0.00021875, -0.00006167, -0.00003790, 0.00006424, -0.00006850, -0.00000266, -0.00004718, -0.00019424, 0.00043534, -0.00022406, 0.00000782, 0.00012661, },
		{ 0.00005476, -0.00001806, 0.00007772, -0.00008244, 0.00001480, 0.00010571, -0.00022431, 0.00013302, -0.00013204, 0.00014614, -0.00017905, 0.00018622, -0.00030048, 0.00040100, -0.00060114, 0.00064571, -0.00158641, 0.00187939, 0.00062456, -0.00112691, 0.00043188, -0.00043470, 0.00033606, -0.00025061, 0.00013028, -0.00011789, -0.00008144, 0.00013475, 0.00012774, -0.00028525, 0.00029321, -0.00021591, },
		{ 0.00002907, 0.00003223, 0.00000850, -0.00011950, 0.00020533, -0.00021840, 0.00007314, -0.00002956, 0.00006515, -0.00013858, 0.00015518, -0.00026690, 0.00032887, -0.00042732, 0.00039719, -0.00093683, 0.00028097, 0.00215086, -0.00161622, 0.00054521, -0.00053047, 0.00035487, -0.00024570, 0.00013750, -0.00013702, 0.00002705, -0.00008824, 0.00029461, -0.00033972, 0.00022030, -0.00007744, -0.00004780, },
		{ 0.00000405, 0.00005638, -0.00014028, 0.00011939, -0.00001872, -0.00015435, 0.00019055, -0.00008231, -0.00001169, -0.00001488, -0.00004274, 0.00001112, 0.00002412, -0.00016763, 0.00000325, -0.00112721, 0.00279225, -0.00112721, 0.00000325, -0.00016763, 0.00002412, 0.00001112, -0.00004274, -0.00001488, -0.00001169, -0.00008231, 0.00019055, -0.00015435, -0.00001872, 0.00011939, -0.00014028, 0.00005638, },
		{ 0.00002907, -0.00004780, -0.00007744, 0.00022030, -0.00033972, 0.00029461, -0.00008824, 0.00002705, -0.00013702, 0.00013750, -0.00024570, 0.00035487, -0.00053047, 0.00054521, -0.00161622, 0.00215086, 0.00028097, -0.00093683, 0.00039719, -0.00042732, 0.00032887, -0.00026690, 0.00015518, -0.00013858, 0.00006515, -0.00002956, 0.00007314, -0.00021840, 0.00020533, -0.00011950, 0.00000850, 0.00003223, },
		{ 0.00005476, -0.00021591, 0.00029321, -0.00028525, 0.00012774, 0.00013475, -0.00008144, -0.00011789, 0.00013028, -0.00025061, 0.00033606, -0.00043470, 0.00043188, -0.00112691, 0.00062456, 0.00187939, -0.00158641, 0.00064571, -0.00060114, 0.00040100, -0.00030048, 0.00018622, -0.00017905, 0.00014614, -0.00013204, 0.00013302, -0.00022431, 0.00010571, 0.00001480, -0.00008244, 0.00007772, -0.00001806, },
		{ -0.00018428, 0.00012661, 0.00000782, -0.00022406, 0.00043534, -0.00019424, -0.00004718, -0.00000266, -0.00006850, 0.00006424, -0.00003790, -0.00006167, -0.00021875, -0.00087568, 0.00272809, -0.00131438, 0.00022489, -0.00028092, 0.00009179, -0.00005290, 0.00000148, -0.00004392, 0.00004843, -0.00002240, -0.00000903, -0.00006629, -0.00009047, 0.00014799, -0.00012028, 0.00006679, -0.00000923, 0.00006759, },
		{ -0.00015155, 0.00032273, -0.00048324, 0.00053022, -0.00009424, -0.00008238, -0.00009597, 0.00008911, -0.00018072, 0.00029192, -0.00044069, 0.00040939, -0.00155699, 0.00232094, -0.00006218, -0.00069695, 0.00032608, -0.00039441, 0.00029626, -0.00025668, 0.00016067, -0.00013705, 0.00017781, -0.00018985, 0.00007621, -0.00020945, 0.00015018, -0.00004085, -0.00002274, 0.00006467, -0.00000887, -0.00002507, },
		{ 0.00040312, -0.00041168, 0.00025702, 0.00027629, -0.00024675, -0.00007779, 0.00008899, -0.00020987, 0.00031637, -0.00041540, 0.00042672, -0.00124781, 0.00093341, 0.00153408, -0.00147039, 0.00070005, -0.00064458, 0.00042429, -0.00033657, 0.00022710, -0.00021387, 0.00025747, -0.00020686, 0.00004435, -0.00015112, -0.00000045, 0.00011100, -0.00012143, 0.00012825, -0.00007799, 0.00010955, -0.00031918, },
		{ -0.00002664, -0.00025005, 0.00070137, -0.00035418, -0.00005066, -0.00001160, -0.00007067, 0.00009818, -0.00008624, 0.00002535, -0.00041740, -0.00058669, 0.00254255, -0.00141857, 0.00042298, -0.00038846, 0.00015624, -0.00011798, 0.00005641, -0.00008307, 0.00012184, -0.00000988, -0.00014858, 0.00001632, -0.00018626, 0.00020378, -0.00012262, 0.00009534, -0.00004705, 0.00010056, -0.00025023, 0.00017223, },
		{ -0.00063283, 0.00082930, -0.00016218, -0.00016522, -0.00006479, 0.00004954, -0.00011554, 0.00022046, -0.00034216, 0.00025411, -0.00141675, 0.00237302, -0.00037087, -0.00043244, 0.00022733, -0.00034046, 0.00024324, -0.00022184, 0.00014691, -0.00011481, 0.00024790, -0.00032250, 0.00011426, -0.00024700, 0.00013151, 0.00001235, -0.00003502, 0.00007382, -0.00002884, -0.00002023, -0.00019565, 0.00039169, },
		{ 0.00046881, 0.00036919, -0.00041101, -0.00002498, 0.00003706, -0.00015042, 0.00027217, -0.00037182, 0.00038401, -0.00128849, 0.00117587, 0.00114975, -0.00128157, 0.00070392, -0.00065596, 0.00042185, -0.00034832, 0.00025346, -0.00023523, 0.00034572, -0.00029062, 0.00001701, -0.00012436, -0.00008157, 0.00019683, -0.00015692, 0.00017170, -0.00013787, 0.00016485, -0.00041671, 0.00051894, -0.00058896, },
		{ 0.00095863, -0.00054963, -0.00001989, -0.00004303, -0.00004755, 0.00010748, -0.00011487, 0.00008494, -0.00057166, -0.00029176, 0.00225554, -0.00143190, 0.00057803, -0.00047749, 0.00020905, -0.00017424, 0.00011290, -0.00012447, 0.00019510, -0.00003270, -0.00022414, 0.00007532, -0.00028320, 0.00027361, -0.00014232, 0.00013586, -0.00010372, 0.00015204, -0.00033243, 0.00025228, -0.00012971, -0.00020977, },
		{ -0.00029523, -0.00020763, -0.00005156, 0.00002618, -0.00006019, 0.00014983, -0.00024596, 0.00009808, -0.00121329, 0.00230362, -0.00061669, -0.00017086, 0.00011373, -0.00027264, 0.00017766, -0.00016792, 0.00011718, -0.00007738, 0.00027435, -0.00041068, 0.00016329, -0.00031214, 0.00014572, 0.00003660, -0.00002750, 0.00005783, -0.00002630, -0.00003669, -0.00020204, 0.00040857, -0.00076189, 0.00117028, },
		{ -0.00055273, 0.00003099, -0.00001785, -0.00008076, 0.00020930, -0.00030955, 0.00031143, -0.00124772, 0.00132804, 0.00076387, -0.00104198, 0.00066005, -0.00063446, 0.00039434, -0.00033380, 0.00026125, -0.00023989, 0.00040000, -0.00036169, 0.00003165, -0.00013791, -0.00012761, 0.00025997, -0.00018196, 0.00020035, -0.00018804, 0.00021224, -0.00049603, 0.00062157, -0.00079885, 0.00075944, 0.00039265, },
		{ 0.00004185, -0.00009489, -0.00000121, 0.00009035, -0.00012063, 0.00011280, -0.00066600, -0.00002271, 0.00189721, -0.00135813, 0.00067676, -0.00053740, 0.00024349, -0.00021326, 0.00016184, -0.00016031, 0.00025476, -0.00007381, -0.00024248, 0.00010112, -0.00036046, 0.00034192, -0.00017202, 0.00018047, -0.00017066, 0.00021579, -0.00042229, 0.00035672, -0.00029913, -0.00008766, 0.00117526, -0.00076098, },
		{ -0.00006222, 0.00002418, -0.00002317, 0.00008908, -0.00016237, -0.00003987, -0.00097151, 0.00212270, -0.00078053, 0.00006155, 0.00000029, -0.00019978, 0.00010899, -0.00010327, 0.00007731, -0.00003340, 0.00026347, -0.00044674, 0.00020228, -0.00038189, 0.00018268, 0.00003299, -0.00000253, 0.00001876, -0.00000019, -0.00007467, -0.00016836, 0.00036064, -0.00084349, 0.00152295, -0.00048656, -0.00020101, },
		{ -0.00006738, -0.00001095, 0.00013624, -0.00023653, 0.00022077, -0.00113398, 0.00137694, 0.00041166, -0.00077923, 0.00057764, -0.00058335, 0.00034591, -0.00029504, 0.00024956, -0.00022805, 0.00041534, -0.00040255, 0.00006500, -0.00017669, -0.00013649, 0.00029197, -0.00019169, 0.00020894, -0.00022032, 0.00024426, -0.00054656, 0.00069188, -0.00101606, 0.00111326, 0.00033436, -0.00065179, 0.00007925, },
		{ 0.00006281, 0.00004894, -0.00010366, 0.00010930, -0.00069251, 0.00019267, 0.00150389, -0.00121182, 0.00071375, -0.00056121, 0.00025551, -0.00022932, 0.00019554, -0.00018409, 0.00028950, -0.00011424, -0.00021942, 0.00009424, -0.00040272, 0.00039338, -0.00020330, 0.00022085, -0.00023781, 0.00028348, -0.00050948, 0.00047088, -0.00052263, 0.00012167, 0.00132208, -0.00096416, 0.00012616, -0.00016195, },
		},
		{ { 0.00001861, -0.00011503, 0.00015390, -0.00012794, 0.00001191, -0.00079225, 0.00189744, -0.00088939, 0.00007463, 0.00001331, -0.00020461, 0.00016799, -0.00000652, 0.00007245, -0.00045488, 0.00031136, -0.00015707, 0.00031136, -0.00045488, 0.00007245, -0.00000652, 0.00016799, -0.00020461, 0.00001331, 0.00007463, -0.00088939, 0.00189744, -0.00079225, 0.00001191, -0.00012794, 0.00015390, -0.00011503, },
		{ -0.00010060, 0.00014184, -0.00014929, 0.00022684, -0.00090073, 0.00108825, 0.00044216, -0.00090029, 0.00063096, -0.00058469, 0.00036534, -0.00019207, 0.00030870, -0.00051315, 0.00015333, -0.00001723, 0.00014198, -0.00007201, -0.00035458, 0.00029370, -0.00016808, 0.00028973, -0.00057872, 0.00073032, -0.00136085, 0.00143938, 0.00030518, -0.00067008, 0.00017441, -0.00009095, 0.00007692, -0.00006941, },
		{ -0.00004568, 0.00005966, 0.00002967, -0.00035167, -0.00014027, 0.00150421, -0.00119528, 0.00055183, -0.00037080, 0.00011099, -0.00002632, 0.00017701, -0.00020259, -0.00023279, 0.00023096, -0.00010542, 0.00026800, -0.00055161, 0.00029886, -0.00019723, 0.00035966, -0.00053460, 0.00048380, -0.00066234, -0.00003504, 0.00169609, -0.00113866, 0.00029469, -0.00025976, 0.00022214, -0.00016652, 0.00001533, },
		{ 0.00015084, -0.00011517, 0.00009625, -0.00078324, 0.00146538, -0.00043398, -0.00023826, 0.00026666, -0.00036318, 0.00026660, -0.00007975, 0.00015606, -0.00049179, 0.00027205, -0.00013054, 0.00027313, -0.00031989, -0.00007449, 0.00011532, 0.00001265, 0.00000954, -0.00027521, 0.00040341, -0.00130382, 0.00214982, -0.00058642, -0.00020869, -0.00003416, 0.00005133, -0.00003831, -0.00001474, -0.00011107, },
		{ -0.00000697, 0.00011368, -0.00060733, 0.00053828, 0.00074250, -0.00097452, 0.00064194, -0.00051425, 0.00027663, -0.00011226, 0.00023899, -0.00039462, 0.00001038, 0.00007946, 0.00002514, 0.00010178, -0.00046894, 0.00036460, -0.00026585, 0.00040436, -0.00068655, 0.00077862, -0.00133875, 0.00111997, 0.00092155, -0.00103709, 0.00035027, -0.00027214, 0.00020953, -0.00016747, 0.00000059, 0.00001478, },
		{ -0.00003584, -0.00013959, -0.00037261, 0.00133655, -0.00089815, 0.00029993, -0.00012032, -0.00007900, 0.00011113, 0.00006356, -0.00007331, -0.00032118, 0.00025628, -0.00015024, 0.00031072, -0.00048675, 0.00020471, -0.00012576, 0.00026049, -0.00037336, 0.00022592, -0.00030572, -0.00060020, 0.00223775, -0.00122573, 0.00025183, -0.00027940, 0.00021054, -0.00016212, 0.00004899, -0.00011029, 0.00012750, },
		{ 0.00007798, -0.00062671, 0.00096866, -0.00004216, -0.00046115, 0.00042718, -0.00042447, 0.00027680, -0.00007664, 0.00015004, -0.00042354, 0.00017424, -0.00006609, 0.00017342, -0.00012895, -0.00023426, 0.00024008, -0.00015076, 0.00022075, -0.00052741, 0.00066541, -0.00157710, 0.00213927, -0.00014636, -0.00055603, 0.00015538, -0.00013595, 0.00009744, -0.00010136, -0.00003494, 0.00005831, -0.00002476, },
		{ -0.00034631, 0.00013773, 0.00082778, -0.00088707, 0.00053156, -0.00034869, 0.00012249, 0.00001202, 0.00012487, -0.00024987, -0.00009837, 0.00014069, -0.00007514, 0.00022962, -0.00049918, 0.00035774, -0.00028851, 0.00042198, -0.00066363, 0.00067554, -0.00112170, 0.00058709, 0.00161165, -0.00136140, 0.00048588, -0.00041578, 0.00029591, -0.00023355, 0.00009112, -0.00010174, 0.00010176, 0.00002183, },
		{ -0.00040875, 0.00101181, -0.00055665, 0.00004843, 0.00010628, -0.00022403, 0.00019499, 0.00000154, -0.00001897, -0.00030044, 0.00020957, -0.00014238, 0.00027104, -0.00033317, 0.00005325, -0.00000357, 0.00009765, -0.00014257, -0.00008346, 0.00008988, -0.00113668, 0.00258869, -0.00108912, 0.00005891, -0.00017704, 0.00009064, -0.00007657, 0.00001486, -0.00009301, 0.00009279, -0.00001150, -0.00004609, },
		{ 0.00052105, 0.00020669, -0.00055057, 0.00046787, -0.00037651, 0.00019508, 0.00000166, 0.00006250, -0.00028288, 0.00006892, -0.00000238, 0.00005291, 0.00005771, -0.00035080, 0.00032102, -0.00027308, 0.00037197, -0.00067960, 0.00079466, -0.00164236, 0.00183994, 0.00048616, -0.00097346, 0.00039097, -0.00035496, 0.00024647, -0.00020568, 0.00007906, -0.00007790, 0.00008743, 0.00001652, -0.00041209, },
		{ 0.00071794, -0.00068470, 0.00034845, -0.00013974, -0.00004717, 0.00013208, 0.00001939, -0.00014483, -0.00011459, 0.00013485, -0.00012710, 0.00027465, -0.00042924, 0.00026982, -0.00023204, 0.00033971, -0.00051598, 0.00044298, -0.00075410, -0.00006326, 0.00224475, -0.00154314, 0.00051020, -0.00046025, 0.00029071, -0.00023125, 0.00013088, -0.00015820, 0.00012766, -0.00000413, -0.00017347, -0.00007456, },
		{ -0.00026234, -0.00013526, 0.00025417, -0.00028241, 0.00019208, 0.00002027, -0.00006419, -0.00017274, 0.00011868, -0.00010305, 0.00017280, -0.00013714, -0.00010670, 0.00012558, -0.00008053, 0.00009674, -0.00036871, 0.00043355, -0.00153182, 0.00265823, -0.00070471, -0.00027957, 0.00003802, -0.00011773, 0.00007161, -0.00007680, -0.00000097, -0.00002685, 0.00007517, -0.00004173, -0.00031128, 0.00063396, },
		{ -0.00050554, 0.00039908, -0.00024046, 0.00004811, 0.00012432, -0.00006603, -0.00013318, 0.00002257, 0.00001607, -0.00004278, 0.00018447, -0.00038291, 0.00032746, -0.00032091, 0.00042523, -0.00069542, 0.00076206, -0.00148510, 0.00129332, 0.00121280, -0.00136924, 0.00059827, -0.00053655, 0.00035361, -0.00028274, 0.00018397, -0.00019942, 0.00015992, -0.00002737, -0.00022147, 0.00020473, 0.00027948, },
		{ 0.00015813, 0.00005115, -0.00017744, 0.00019714, -0.00002559, -0.00013394, -0.00000762, 0.00006079, -0.00012690, 0.00023457, -0.00027920, 0.00012743, -0.00011999, 0.00018486, -0.00028443, 0.00013978, -0.00031913, -0.00070483, 0.00268967, -0.00150420, 0.00037758, -0.00037160, 0.00017580, -0.00014671, 0.00009995, -0.00012946, 0.00005416, 0.00004456, -0.00010108, -0.00011568, 0.00048112, -0.00044259, },
		{ 0.00029635, -0.00024020, 0.00009732, 0.00011848, -0.00019580, 0.00001759, 0.00004236, -0.00007552, 0.00006194, 0.00004082, -0.00022228, 0.00021614, -0.00022336, 0.00028297, -0.00056065, 0.00065011, -0.00170764, 0.00241575, -0.00011028, -0.00071109, 0.00031670, -0.00036296, 0.00023942, -0.00019245, 0.00013149, -0.00018408, 0.00017144, -0.00007263, -0.00016316, 0.00030418, -0.00008381, -0.00021084, },
		{ -0.00006335, -0.00011540, 0.00024046, -0.00017626, -0.00004761, 0.00009275, -0.00003921, -0.00008423, 0.00022066, -0.00031951, 0.00025592, -0.00028718, 0.00037298, -0.00057680, 0.00058400, -0.00114667, 0.00059915, 0.00190065, -0.00164135, 0.00070184, -0.00061739, 0.00037275, -0.00029438, 0.00023725, -0.00025590, 0.00014824, -0.00001406, -0.00010604, 0.00005180, 0.00019573, -0.00036559, 0.00026306, },
		{ -0.00022613, 0.00017222, 0.00002128, -0.00023972, 0.00021001, -0.00004500, -0.00010202, 0.00014076, -0.00009801, -0.00002246, 0.00000569, 0.00000526, -0.00003198, -0.00015614, 0.00008584, -0.00121441, 0.00284980, -0.00121441, 0.00008584, -0.00015614, -0.00003198, 0.00000526, 0.00000569, -0.00002246, -0.00009801, 0.00014076, -0.00010202, -0.00004500, 0.00021001, -0.00023972, 0.00002128, 0.00017222, },
		{ -0.00006335, 0.00026306, -0.00036559, 0.00019573, 0.00005180, -0.00010604, -0.00001406, 0.00014824, -0.00025590, 0.00023725, -0.00029438, 0.00037275, -0.00061739, 0.00070184, -0.00164135, 0.00190065, 0.00059915, -0.00114667, 0.00058400, -0.00057680, 0.00037298, -0.00028718, 0.00025592, -0.00031951, 0.00022066, -0.00008423, -0.00003921, 0.00009275, -0.00004761, -0.00017626, 0.00024046, -0.00011540, },
		{ 0.00029635, -0.00021084, -0.00008381, 0.00030418, -0.00016316, -0.00007263, 0.00017144, -0.00018408, 0.00013149, -0.00019245, 0.00023942, -0.00036296, 0.00031670, -0.00071109, -0.00011028, 0.00241575, -0.00170764, 0.00065011, -0.00056065, 0.00028297, -0.00022336, 0.00021614, -0.00022228, 0.00004082, 0.00006194, -0.00007552, 0.00004236, 0.00001759, -0.00019580, 0.00011848, 0.00009732, -0.00024020, },
		{ 0.00015813, -0.00044259, 0.00048112, -0.00011568, -0.00010108, 0.00004456, 0.00005416, -0.00012946, 0.00009995, -0.00014671, 0.00017580, -0.00037160, 0.00037758, -0.00150420, 0.00268967, -0.00070483, -0.00031913, 0.00013978, -0.00028443, 0.00018486, -0.00011999, 0.00012743, -0.00027920, 0.00023457, -0.00012690, 0.00006079, -0.00000762, -0.00013394, -0.00002559, 0.00019714, -0.00017744, 0.00005115, },
		{ -0.00050554, 0.00027948, 0.00020473, -0.00022147, -0.00002737, 0.00015992, -0.00019942, 0.00018397, -0.00028274, 0.00035361, -0.00053655, 0.00059827, -0.00136924, 0.00121280, 0.00129332, -0.00148510, 0.00076206, -0.00069542, 0.00042523, -0.00032091, 0.00032746, -0.00038291, 0.00018447, -0.00004278, 0.00001607, 0.00002257, -0.00013318, -0.00006603, 0.00012432, 0.00004811, -0.00024046, 0.00039908, },
		{ -0.00026234, 0.00063396, -0.00031128, -0.00004173, 0.00007517, -0.00002685, -0.00000097, -0.00007680, 0.00007161, -0.00011773, 0.00003802, -0.00027957, -0.00070471, 0.00265823, -0.00153182, 0.00043355, -0.00036871, 0.00009674, -0.00008053, 0.00012558, -0.00010670, -0.00013714, 0.00017280, -0.00010305, 0.00011868, -0.00017274, -0.00006419, 0.00002027, 0.00019208, -0.00028241, 0.00025417, -0.00013526, },
		{ 0.00071794, -0.00007456, -0.00017347, -0.00000413, 0.00012766, -0.00015820, 0.00013088, -0.00023125, 0.00029071, -0.00046025, 0.00051020, -0.00154314, 0.00224475, -0.00006326, -0.00075410, 0.00044298, -0.00051598, 0.00033971, -0.00023204, 0.00026982, -0.00042924, 0.00027465, -0.00012710, 0.00013485, -0.00011459, -0.00014483, 0.00001939, 0.00013208, -0.00004717, -0.00013974, 0.00034845, -0.00068470, },
		{ 0.00052105, -0.00041209, 0.00001652, 0.00008743, -0.00007790, 0.00007906, -0.00020568, 0.00024647, -0.00035496, 0.00039097, -0.00097346, 0.00048616, 0.00183994, -0.00164236, 0.00079466, -0.00067960, 0.00037197, -0.00027308, 0.00032102, -0.00035080, 0.00005771, 0.00005291, -0.00000238, 0.00006892, -0.00028288, 0.00006250, 0.00000166, 0.00019508, -0.00037651, 0.00046787, -0.00055057, 0.00020669, },
		{ -0.00040875, -0.00004609, -0.00001150, 0.00009279, -0.00009301, 0.00001486, -0.00007657, 0.00009064, -0.00017704, 0.00005891, -0.00108912, 0.00258869, -0.00113668, 0.00008988, -0.00008346, -0.00014257, 0.00009765, -0.00000357, 0.00005325, -0.00033317, 0.00027104, -0.00014238, 0.00020957, -0.00030044, -0.00001897, 0.00000154, 0.00019499, -0.00022403, 0.00010628, 0.00004843, -0.00055665, 0.00101181, },
		{ -0.00034631, 0.00002183, 0.00010176, -0.00010174, 0.00009112, -0.00023355, 0.00029591, -0.00041578, 0.00048588, -0.00136140, 0.00161165, 0.00058709, -0.00112170, 0.00067554, -0.00066363, 0.00042198, -0.00028851, 0.00035774, -0.00049918, 0.00022962, -0.00007514, 0.00014069, -0.00009837, -0.00024987, 0.00012487, 0.00001202, 0.00012249, -0.00034869, 0.00053156, -0.00088707, 0.00082778, 0.00013773, },
		{ 0.00007798, -0.00002476, 0.00005831, -0.00003494, -0.00010136, 0.00009744, -0.00013595, 0.00015538, -0.00055603, -0.00014636, 0.00213927, -0.00157710, 0.00066541, -0.00052741, 0.00022075, -0.00015076, 0.00024008, -0.00023426, -0.00012895, 0.00017342, -0.00006609, 0.00017424, -0.00042354, 0.00015004, -0.00007664, 0.00027680, -0.00042447, 0.00042718, -0.00046115, -0.00004216, 0.00096866, -0.00062671, },
		{ -0.00003584, 0.00012750, -0.00011029, 0.00004899, -0.00016212, 0.00021054, -0.00027940, 0.00025183, -0.00122573, 0.00223775, -0.00060020, -0.00030572, 0.00022592, -0.00037336, 0.00026049, -0.00012576, 0.00020471, -0.00048675, 0.00031072, -0.00015024, 0.00025628, -0.00032118, -0.00007331, 0.00006356, 0.00011113, -0.00007900, -0.00012032, 0.00029993, -0.00089815, 0.00133655, -0.00037261, -0.00013959, },
		{ -0.00000697, 0.00001478, 0.00000059, -0.00016747, 0.00020953, -0.00027214, 0.00035027, -0.00103709, 0.00092155, 0.00111997, -0.00133875, 0.00077862, -0.00068655, 0.00040436, -0.00026585, 0.00036460, -0.00046894, 0.00010178, 0.00002514, 0.00007946, 0.00001038, -0.00039462, 0.00023899, -0.00011226, 0.00027663, -0.00051425, 0.00064194, -0.00097452, 0.00074250, 0.00053828, -0.00060733, 0.00011368, },
		{ 0.00015084, -0.00011107, -0.00001474, -0.00003831, 0.00005133, -0.00003416, -0.00020869, -0.00058642, 0.00214982, -0.00130382, 0.00040341, -0.00027521, 0.00000954, 0.00001265, 0.00011532, -0.00007449, -0.00031989, 0.00027313, -0.00013054, 0.00027205, -0.00049179, 0.00015606, -0.00007975, 0.00026660, -0.00036318, 0.00026666, -0.00023826, -0.00043398, 0.00146538, -0.00078324, 0.00009625, -0.00011517, },
		{ -0.00004568, 0.00001533, -0.00016652, 0.00022214, -0.00025976, 0.00029469, -0.00113866, 0.00169609, -0.00003504, -0.00066234, 0.00048380, -0.00053460, 0.00035966, -0.00019723, 0.00029886, -0.00055161, 0.00026800, -0.00010542, 0.00023096, -0.00023279, -0.00020259, 0.00017701, -0.00002632, 0.00011099, -0.00037080, 0.00055183, -0.00119528, 0.00150421, -0.00014027, -0.00035167, 0.00002967, 0.00005966, },
		{ -0.00010060, -0.00006941, 0.00007692, -0.00009095, 0.00017441, -0.00067008, 0.00030518, 0.00143938, -0.00136085, 0.00073032, -0.00057872, 0.00028973, -0.00016808, 0.00029370, -0.00035458, -0.00007201, 0.00014198, -0.00001723, 0.00015333, -0.00051315, 0.00030870, -0.00019207, 0.00036534, -0.00058469, 0.00063096, -0.00090029, 0.00044216, 0.00108825, -0.00090073, 0.00022684, -0.00014929, 0.00014184, },
		},
		{ { -0.00006771, 0.00007902, -0.00018712, 0.00021560, -0.00011016, 0.00002121, -0.00080468, 0.00181633, -0.00070222, -0.00003945, 0.00011894, -0.00039318, 0.00011555, -0.00011887, 0.00019312, -0.00017800, 0.00010188, -0.00017800, 0.00019312, -0.00011887, 0.00011555, -0.00039318, 0.00011894, -0.00003945, -0.00070222, 0.00181633, -0.00080468, 0.00002121, -0.00011016, 0.00021560, -0.00018712, 0.00007902, },
		{ 0.00004679, -0.00014318, 0.00012384, -0.00007655, 0.00020697, -0.00080220, 0.00076879, 0.00080043, -0.00100295, 0.00065355, -0.00059172, 0.00013772, -0.00017617, 0.00018962, -0.00008538, -0.00001048, -0.00003333, -0.00003702, 0.00014695, -0.00016143, 0.00009399, -0.00055075, 0.00068594, -0.00124445, 0.00130424, 0.00043903, -0.00073391, 0.00018987, -0.00004388, 0.00008898, -0.00012815, 0.00003116, },
		{ 0.00006176, -0.00014276, 0.00016433, -0.00004214, -0.00012543, -0.00051588, 0.00162608, -0.00088974, 0.00023536, -0.00008849, -0.00026420, 0.00004711, -0.00004422, 0.00014808, -0.00017333, 0.00010025, -0.00017327, 0.00022541, -0.00018967, 0.00017966, -0.00050844, 0.00034035, -0.00039139, -0.00035701, 0.00185814, -0.00104400, 0.00016565, -0.00017384, 0.00025788, -0.00022962, 0.00009580, -0.00006608, },
		{ -0.00014835, 0.00013816, -0.00008672, 0.00018810, -0.00079011, 0.00098443, 0.00033398, -0.00069720, 0.00053206, -0.00056196, 0.00014662, -0.00016560, 0.00020877, -0.00012420, 0.00001243, -0.00005484, 0.00001712, 0.00008285, -0.00012094, 0.00001864, -0.00043907, 0.00060739, -0.00137000, 0.00178364, 0.00002840, -0.00058942, 0.00013595, 0.00000901, 0.00003446, -0.00010194, 0.00001399, 0.00006068, },
		{ -0.00010261, 0.00011297, 0.00002009, -0.00025373, -0.00021775, 0.00132556, -0.00092133, 0.00040448, -0.00024855, -0.00014202, -0.00001552, 0.00002476, 0.00009700, -0.00016066, 0.00009542, -0.00015861, 0.00023935, -0.00024694, 0.00022944, -0.00059064, 0.00053820, -0.00077606, 0.00012364, 0.00173189, -0.00119923, 0.00028677, -0.00022323, 0.00028288, -0.00026364, 0.00011024, -0.00006126, 0.00004540, },
		{ 0.00013422, -0.00007573, 0.00013934, -0.00070516, 0.00106984, -0.00004349, -0.00038363, 0.00035642, -0.00047253, 0.00012292, -0.00013374, 0.00020568, -0.00015128, 0.00003420, -0.00007375, 0.00007216, 0.00000261, -0.00005796, -0.00008009, -0.00026801, 0.00041310, -0.00134109, 0.00217552, -0.00041701, -0.00038421, 0.00004982, 0.00007648, -0.00003559, -0.00006484, -0.00000439, 0.00007274, -0.00014625, },
		{ 0.00006969, 0.00006776, -0.00034679, 0.00005004, 0.00096531, -0.00082134, 0.00045895, -0.00033715, -0.00004516, -0.00006356, 0.00008008, 0.00004669, -0.00014202, 0.00008759, -0.00013448, 0.00023132, -0.00028231, 0.00025662, -0.00062465, 0.00067665, -0.00113971, 0.00069573, 0.00143999, -0.00124646, 0.00036608, -0.00025003, 0.00028429, -0.00028303, 0.00012024, -0.00005342, 0.00003066, -0.00007128, },
		{ -0.00004820, 0.00007097, -0.00056518, 0.00102794, -0.00029764, -0.00011442, 0.00016873, -0.00034363, 0.00007377, -0.00008747, 0.00018464, -0.00016614, 0.00005378, -0.00008893, 0.00012276, -0.00008589, 0.00002087, -0.00019000, -0.00005872, 0.00011674, -0.00113900, 0.00242469, -0.00084476, -0.00014293, -0.00005893, 0.00015029, -0.00011434, -0.00001885, -0.00002347, 0.00008289, -0.00014021, 0.00011696, },
		{ 0.00009477, -0.00039379, 0.00025408, 0.00060036, -0.00063185, 0.00041021, -0.00034314, 0.00001261, -0.00009090, 0.00011638, 0.00000305, -0.00011976, 0.00007704, -0.00010235, 0.00020031, -0.00028994, 0.00025589, -0.00060193, 0.00072728, -0.00142715, 0.00129901, 0.00100717, -0.00117614, 0.00039051, -0.00024898, 0.00025884, -0.00028308, 0.00012379, -0.00004287, 0.00001758, -0.00005130, 0.00004062, },
		{ -0.00000425, -0.00039541, 0.00087959, -0.00041574, 0.00007057, 0.00001121, -0.00020128, 0.00001005, -0.00003539, 0.00015212, -0.00016999, 0.00007040, -0.00009952, 0.00016373, -0.00017324, 0.00010646, -0.00029672, 0.00016089, -0.00025106, -0.00076853, 0.00249207, -0.00120338, 0.00010410, -0.00017716, 0.00022076, -0.00019306, 0.00003249, -0.00004254, 0.00009103, -0.00013369, 0.00009306, -0.00001126, },
		{ -0.00039155, 0.00037204, 0.00028187, -0.00040527, 0.00028753, -0.00026997, 0.00002440, -0.00009499, 0.00013164, -0.00002976, -0.00009619, 0.00006419, -0.00006453, 0.00014813, -0.00026744, 0.00022578, -0.00052197, 0.00067363, -0.00159022, 0.00186519, 0.00047747, -0.00099465, 0.00035444, -0.00021860, 0.00020687, -0.00026131, 0.00011928, -0.00003003, 0.00000558, -0.00004282, 0.00002897, 0.00009860, },
		{ -0.00022455, 0.00065987, -0.00040762, 0.00014988, -0.00008071, -0.00007311, -0.00005543, 0.00001351, 0.00011571, -0.00016538, 0.00008356, -0.00010497, 0.00019080, -0.00024974, 0.00018856, -0.00038581, 0.00036022, -0.00064678, -0.00025827, 0.00236071, -0.00144972, 0.00032520, -0.00028996, 0.00027821, -0.00026241, 0.00008452, -0.00006073, 0.00009699, -0.00012964, 0.00006977, 0.00002672, -0.00007307, },
		{ 0.00039573, 0.00004948, -0.00019540, 0.00013274, -0.00013493, -0.00000886, -0.00007718, 0.00012737, -0.00004985, -0.00007332, 0.00004953, -0.00002395, 0.00007929, -0.00021631, 0.00016895, -0.00039252, 0.00051413, -0.00159525, 0.00232740, -0.00009164, -0.00072344, 0.00026063, -0.00016157, 0.00013249, -0.00021792, 0.00010580, -0.00001537, -0.00000635, -0.00004367, 0.00003457, 0.00008058, -0.00034484, },
		{ 0.00041223, -0.00030309, 0.00012341, -0.00008424, 0.00001608, -0.00011002, 0.00005130, 0.00008282, -0.00015571, 0.00009304, -0.00010510, 0.00020117, -0.00030677, 0.00025718, -0.00044485, 0.00051054, -0.00102019, 0.00034266, 0.00203861, -0.00155507, 0.00049202, -0.00038285, 0.00031438, -0.00031371, 0.00013199, -0.00007705, 0.00010050, -0.00012998, 0.00005366, 0.00005744, -0.00012385, -0.00008023, },
		{ -0.00007437, -0.00004856, -0.00000701, 0.00003366, -0.00007865, -0.00004241, 0.00010822, -0.00005777, -0.00005255, 0.00003363, 0.00001620, 0.00000038, -0.00014173, 0.00009183, -0.00022858, 0.00026261, -0.00142853, 0.00262963, -0.00063708, -0.00039606, 0.00011992, -0.00008433, 0.00004302, -0.00015590, 0.00008333, 0.00000054, -0.00001939, -0.00004986, 0.00005397, 0.00004569, -0.00026547, 0.00033195, },
		{ -0.00014630, 0.00001203, 0.00000695, 0.00004829, -0.00014323, 0.00007233, 0.00005952, -0.00014461, 0.00009887, -0.00010010, 0.00019385, -0.00033799, 0.00030402, -0.00046532, 0.00058930, -0.00132162, 0.00097102, 0.00155800, -0.00150918, 0.00058358, -0.00044387, 0.00032368, -0.00034019, 0.00016985, -0.00009049, 0.00010125, -0.00013528, 0.00004952, 0.00007405, -0.00014841, 0.00001525, 0.00018155, },
		{ 0.00000405, -0.00008736, 0.00020111, -0.00017033, 0.00000162, 0.00008108, -0.00005630, -0.00003458, 0.00001708, 0.00005283, -0.00008072, -0.00005192, 0.00000377, -0.00005029, -0.00005330, -0.00109892, 0.00273471, -0.00109892, -0.00005330, -0.00005029, 0.00000377, -0.00005192, -0.00008072, 0.00005283, 0.00001708, -0.00003458, -0.00005630, 0.00008108, 0.00000162, -0.00017033, 0.00020111, -0.00008736, },
		{ -0.00014630, 0.00018155, 0.00001525, -0.00014841, 0.00007405, 0.00004952, -0.00013528, 0.00010125, -0.00009049, 0.00016985, -0.00034019, 0.00032368, -0.00044387, 0.00058358, -0.00150918, 0.00155800, 0.00097102, -0.00132162, 0.00058930, -0.00046532, 0.00030402, -0.00033799, 0.00019385, -0.00010010, 0.00009887, -0.00014461, 0.00005952, 0.00007233, -0.00014323, 0.00004829, 0.00000695, 0.00001203, },
		{ -0.00007437, 0.00033195, -0.00026547, 0.00004569, 0.00005397, -0.00004986, -0.00001939, 0.00000054, 0.00008333, -0.00015590, 0.00004302, -0.00008433, 0.00011992, -0.00039606, -0.00063708, 0.00262963, -0.00142853, 0.00026261, -0.00022858, 0.00009183, -0.00014173, 0.00000038, 0.00001620, 0.00003363, -0.00005255, -0.00005777, 0.00010822, -0.00004241, -0.00007865, 0.00003366, -0.00000701, -0.00004856, },
		{ 0.00041223, -0.00008023, -0.00012385, 0.00005744, 0.00005366, -0.00012998, 0.00010050, -0.00007705, 0.00013199, -0.00031371, 0.00031438, -0.00038285, 0.00049202, -0.00155507, 0.00203861, 0.00034266, -0.00102019, 0.00051054, -0.00044485, 0.00025718, -0.00030677, 0.00020117, -0.00010510, 0.00009304, -0.00015571, 0.00008282, 0.00005130, -0.00011002, 0.00001608, -0.00008424, 0.00012341, -0.00030309, },
		{ 0.00039573, -0.00034484, 0.00008058, 0.00003457, -0.00004367, -0.00000635, -0.00001537, 0.00010580, -0.00021792, 0.00013249, -0.00016157, 0.00026063, -0.00072344, -0.00009164, 0.00232740, -0.00159525, 0.00051413, -0.00039252, 0.00016895, -0.00021631, 0.00007929, -0.00002395, 0.00004953, -0.00007332, -0.00004985, 0.00012737, -0.00007718, -0.00000886, -0.00013493, 0.00013274, -0.00019540, 0.00004948, },
		{ -0.00022455, -0.00007307, 0.00002672, 0.00006977, -0.00012964, 0.00009699, -0.00006073, 0.00008452, -0.00026241, 0.00027821, -0.00028996, 0.00032520, -0.00144972, 0.00236071, -0.00025827, -0.00064678, 0.00036022, -0.00038581, 0.00018856, -0.00024974, 0.00019080, -0.00010497, 0.00008356, -0.00016538, 0.00011571, 0.00001351, -0.00005543, -0.00007311, -0.00008071, 0.00014988, -0.00040762, 0.00065987, },
		{ -0.00039155, 0.00009860, 0.00002897, -0.00004282, 0.00000558, -0.00003003, 0.00011928, -0.00026131, 0.00020687, -0.00021860, 0.00035444, -0.00099465, 0.00047747, 0.00186519, -0.00159022, 0.00067363, -0.00052197, 0.00022578, -0.00026744, 0.00014813, -0.00006453, 0.00006419, -0.00009619, -0.00002976, 0.00013164, -0.00009499, 0.00002440, -0.00026997, 0.00028753, -0.00040527, 0.00028187, 0.00037204, },
		{ -0.00000425, -0.00001126, 0.00009306, -0.00013369, 0.00009103, -0.00004254, 0.00003249, -0.00019306, 0.00022076, -0.00017716, 0.00010410, -0.00120338, 0.00249207, -0.00076853, -0.00025106, 0.00016089, -0.00029672, 0.00010646, -0.00017324, 0.00016373, -0.00009952, 0.00007040, -0.00016999, 0.00015212, -0.00003539, 0.00001005, -0.00020128, 0.00001121, 0.00007057, -0.00041574, 0.00087959, -0.00039541, },
		{ 0.00009477, 0.00004062, -0.00005130, 0.00001758, -0.00004287, 0.00012379, -0.00028308, 0.00025884, -0.00024898, 0.00039051, -0.00117614, 0.00100717, 0.00129901, -0.00142715, 0.00072728, -0.00060193, 0.00025589, -0.00028994, 0.00020031, -0.00010235, 0.00007704, -0.00011976, 0.00000305, 0.00011638, -0.00009090, 0.00001261, -0.00034314, 0.00041021, -0.00063185, 0.00060036, 0.00025408, -0.00039379, },
		{ -0.00004820, 0.00011696, -0.00014021, 0.00008289, -0.00002347, -0.00001885, -0.00011434, 0.00015029, -0.00005893, -0.00014293, -0.00084476, 0.00242469, -0.00113900, 0.00011674, -0.00005872, -0.00019000, 0.00002087, -0.00008589, 0.00012276, -0.00008893, 0.00005378, -0.00016614, 0.00018464, -0.00008747, 0.00007377, -0.00034363, 0.00016873, -0.00011442, -0.00029764, 0.00102794, -0.00056518, 0.00007097, },
		{ 0.00006969, -0.00007128, 0.00003066, -0.00005342, 0.00012024, -0.00028303, 0.00028429, -0.00025003, 0.00036608, -0.00124646, 0.00143999, 0.00069573, -0.00113971, 0.00067665, -0.00062465, 0.00025662, -0.00028231, 0.00023132, -0.00013448, 0.00008759, -0.00014202, 0.00004669, 0.00008008, -0.00006356, -0.00004516, -0.00033715, 0.00045895, -0.00082134, 0.00096531, 0.00005004, -0.00034679, 0.00006776, },
		{ 0.00013422, -0.00014625, 0.00007274, -0.00000439, -0.00006484, -0.00003559, 0.00007648, 0.00004982, -0.00038421, -0.00041701, 0.00217552, -0.00134109, 0.00041310, -0.00026801, -0.00008009, -0.00005796, 0.00000261, 0.00007216, -0.00007375, 0.00003420, -0.00015128, 0.00020568, -0.00013374, 0.00012292, -0.00047253, 0.00035642, -0.00038363, -0.00004349, 0.00106984, -0.00070516, 0.00013934, -0.00007573, },
		{ -0.00010261, 0.00004540, -0.00006126, 0.00011024, -0.00026364, 0.00028288, -0.00022323, 0.00028677, -0.00119923, 0.00173189, 0.00012364, -0.00077606, 0.00053820, -0.00059064, 0.00022944, -0.00024694, 0.00023935, -0.00015861, 0.00009542, -0.00016066, 0.00009700, 0.00002476, -0.00001552, -0.00014202, -0.00024855, 0.00040448, -0.00092133, 0.00132556, -0.00021775, -0.00025373, 0.00002009, 0.00011297, },
		{ -0.00014835, 0.00006068, 0.00001399, -0.00010194, 0.00003446, 0.00000901, 0.00013595, -0.00058942, 0.00002840, 0.00178364, -0.00137000, 0.00060739, -0.00043907, 0.00001864, -0.00012094, 0.00008285, 0.00001712, -0.00005484, 0.00001243, -0.00012420, 0.00020877, -0.00016560, 0.00014662, -0.00056196, 0.00053206, -0.00069720, 0.00033398, 0.00098443, -0.00079011, 0.00018810, -0.00008672, 0.00013816, },
		{ 0.00006176, -0.00006608, 0.00009580, -0.00022962, 0.00025788, -0.00017384, 0.00016565, -0.00104400, 0.00185814, -0.00035701, -0.00039139, 0.00034035, -0.00050844, 0.00017966, -0.00018967, 0.00022541, -0.00017327, 0.00010025, -0.00017333, 0.00014808, -0.00004422, 0.00004711, -0.00026420, -0.00008849, 0.00023536, -0.00088974, 0.00162608, -0.00051588, -0.00012543, -0.00004214, 0.00016433, -0.00014276, },
		{ 0.00004679, 0.00003116, -0.00012815, 0.00008898, -0.00004388, 0.00018987, -0.00073391, 0.00043903, 0.00130424, -0.00124445, 0.00068594, -0.00055075, 0.00009399, -0.00016143, 0.00014695, -0.00003702, -0.00003333, -0.00001048, -0.00008538, 0.00018962, -0.00017617, 0.00013772, -0.00059172, 0.00065355, -0.00100295, 0.00080043, 0.00076879, -0.00080220, 0.00020697, -0.00007655, 0.00012384, -0.00014318, },
		},
		{ { -0.00006987, 0.00001353, 0.00005297, -0.00008440, -0.00008407, 0.00010986, -0.00004841, -0.00072036, 0.00162833, -0.00072012, -0.00004792, 0.00011057, -0.00008317, -0.00008335, 0.00005415, 0.00001478, -0.00006860, 0.00001478, 0.00005415, -0.00008335, -0.00008317, 0.00011057, -0.00004792, -0.00072012, 0.00162833, -0.00072036, -0.00004841, 0.00010986, -0.00008407, -0.00008440, 0.00005297, 0.00001353, },
		{ -0.00005409, 0.00010129, -0.00008713, 0.00001118, -0.00018261, 0.00041615, -0.00087383, 0.00068488, 0.00068500, -0.00087346, 0.00041675, -0.00018181, 0.00001215, -0.00008602, 0.00010250, -0.00005284, -0.00006702, 0.00012210, -0.00011078, 0.00002876, -0.00020575, 0.00044808, -0.00097086, 0.00078281, 0.00078268, -0.00097123, 0.00044749, -0.00020655, 0.00002778, -0.00011189, 0.00012089, -0.00006828, },
		{ 0.00000984, 0.00005483, -0.00007735, -0.00008298, 0.00010734, -0.00002929, -0.00066112, 0.00143901, -0.00066088, -0.00002881, 0.00010803, -0.00008210, -0.00007632, 0.00005598, 0.00001106, -0.00006861, 0.00001848, 0.00005229, -0.00009040, -0.00008426, 0.00011309, -0.00006704, -0.00077936, 0.00181765, -0.00077960, -0.00006752, 0.00011240, -0.00008514, -0.00009143, 0.00005114, 0.00001726, -0.00006986, },
		{ 0.00008209, -0.00006283, -0.00000510, -0.00015912, 0.00038543, -0.00077829, 0.00058895, 0.00058907, -0.00077794, 0.00038600, -0.00015835, -0.00000416, -0.00006176, 0.00008325, -0.00003894, -0.00008096, 0.00014130, -0.00013509, 0.00004503, -0.00022924, 0.00047880, -0.00106640, 0.00087873, 0.00087861, -0.00106675, 0.00047823, -0.00023001, 0.00004409, -0.00013616, 0.00014014, -0.00008217, -0.00004015, },
		{ 0.00005663, -0.00007055, -0.00008192, 0.00010493, -0.00001089, -0.00060415, 0.00125696, -0.00060392, -0.00001044, 0.00010558, -0.00008109, -0.00006958, 0.00005772, 0.00000745, -0.00006865, 0.00002201, 0.00005049, -0.00009720, -0.00008532, 0.00011550, -0.00008544, -0.00083633, 0.00199970, -0.00083655, -0.00008589, 0.00011485, -0.00008615, -0.00009817, 0.00004940, 0.00002086, -0.00006982, 0.00000630, },
		{ -0.00003991, -0.00002040, -0.00013697, 0.00035651, -0.00068829, 0.00049859, 0.00049870, -0.00068797, 0.00035703, -0.00013626, -0.00001954, -0.00003892, 0.00006510, -0.00002588, -0.00009412, 0.00015936, -0.00015800, 0.00006034, -0.00025139, 0.00050773, -0.00115640, 0.00096909, 0.00096898, -0.00115673, 0.00050720, -0.00025210, 0.00005947, -0.00015899, 0.00015829, -0.00009523, -0.00002700, 0.00006403, },
		{ -0.00006427, -0.00008092, 0.00010272, 0.00000607, -0.00055164, 0.00108919, -0.00055144, 0.00000648, 0.00010331, -0.00008017, -0.00006339, 0.00005930, 0.00000411, -0.00006871, 0.00002524, 0.00004880, -0.00010348, -0.00008632, 0.00011771, -0.00010240, -0.00088883, 0.00216747, -0.00088904, -0.00010281, 0.00011712, -0.00008707, -0.00010436, 0.00004782, 0.00002420, -0.00006976, 0.00000307, 0.00005832, },
		{ -0.00003416, -0.00011703, 0.00033048, -0.00060729, 0.00041728, 0.00041737, -0.00060700, 0.00033095, -0.00011641, -0.00003340, -0.00001840, 0.00004874, -0.00001415, -0.00010599, 0.00017559, -0.00017865, 0.00007409, -0.00027133, 0.00053375, -0.00123741, 0.00105041, 0.00105031, -0.00123769, 0.00053329, -0.00027196, 0.00007333, -0.00017952, 0.00017465, -0.00010696, -0.00001513, 0.00004780, -0.00001926, },
		{ -0.00008002, 0.00010080, 0.00002095, -0.00050561, 0.00094213, -0.00050544, 0.00002130, 0.00010130, -0.00007939, -0.00005799, 0.00006065, 0.00000115, -0.00006878, 0.00002804, 0.00004730, -0.00010902, -0.00008722, 0.00011962, -0.00011728, -0.00093486, 0.00231453, -0.00093504, -0.00011763, 0.00011913, -0.00008785, -0.00010976, 0.00004647, 0.00002716, -0.00006968, 0.00000027, 0.00005982, -0.00005873, },
		{ -0.00010005, 0.00030837, -0.00053839, 0.00034812, 0.00034820, -0.00053816, 0.00030875, -0.00009954, -0.00004521, -0.00000096, 0.00003479, -0.00000421, -0.00011611, 0.00018937, -0.00019624, 0.00008577, -0.00028831, 0.00055587, -0.00130630, 0.00111956, 0.00111948, -0.00130654, 0.00055549, -0.00028883, 0.00008515, -0.00019695, 0.00018860, -0.00011691, -0.00000501, 0.00003402, -0.00000168, -0.00004583, },
		{ 0.00009925, 0.00003317, -0.00046783, 0.00082145, -0.00046770, 0.00003344, 0.00009964, -0.00007877, -0.00005358, 0.00006174, -0.00000131, -0.00006888, 0.00003031, 0.00004603, -0.00011358, -0.00008797, 0.00012118, -0.00012951, -0.00097264, 0.00243521, -0.00097278, -0.00012977, 0.00012079, -0.00008847, -0.00011417, 0.00004538, 0.00002962, -0.00006959, -0.00000200, 0.00006109, -0.00005417, -0.00007927, },
		{ 0.00029101, -0.00048425, 0.00029380, 0.00029385, -0.00048407, 0.00029129, -0.00008630, -0.00005452, 0.00001270, 0.00002381, 0.00000357, -0.00012409, 0.00020016, -0.00021009, 0.00009492, -0.00030168, 0.00057323, -0.00136045, 0.00117389, 0.00117383, -0.00136062, 0.00057294, -0.00030206, 0.00009445, -0.00021061, 0.00019958, -0.00012469, 0.00000298, 0.00002323, 0.00001217, -0.00005498, -0.00008668, },
		{ 0.00004227, -0.00043975, 0.00073177, -0.00043966, 0.00004246, 0.00009838, -0.00007833, -0.00005034, 0.00006251, -0.00000317, -0.00006899, 0.00003196, 0.00004506, -0.00011701, -0.00008856, 0.00012232, -0.00013860, -0.00100073, 0.00252489, -0.00100082, -0.00013879, 0.00012205, -0.00008890, -0.00011741, 0.00004461, 0.00003149, -0.00006948, -0.00000365, 0.00006206, -0.00005074, -0.00007868, 0.00009811, },
		{ -0.00044694, 0.00025638, 0.00025641, -0.00044683, 0.00027925, -0.00007721, -0.00006096, 0.00002207, 0.00001619, 0.00000889, -0.00012964, 0.00020755, -0.00021966, 0.00010118, -0.00031091, 0.00058516, -0.00139775, 0.00121131, 0.00121127, -0.00139786, 0.00058499, -0.00031115, 0.00010090, -0.00021999, 0.00020719, -0.00013000, 0.00000852, 0.00001584, 0.00002175, -0.00006125, -0.00007745, 0.00027907, },
		{ -0.00042245, 0.00067655, -0.00042240, 0.00004799, 0.00009758, -0.00007810, -0.00004838, 0.00006294, -0.00000437, -0.00006911, 0.00003293, 0.00004441, -0.00011916, -0.00008896, 0.00012299, -0.00014422, -0.00101803, 0.00258011, -0.00101808, -0.00014432, 0.00012285, -0.00008914, -0.00011937, 0.00004418, 0.00003268, -0.00006936, -0.00000462, 0.00006271, -0.00004859, -0.00007828, 0.00009744, 0.00004789, },
		{ 0.00023731, 0.00023732, -0.00042787, 0.00027308, -0.00007262, -0.00006430, 0.00002680, 0.00001225, 0.00001153, -0.00013253, 0.00021125, -0.00022460, 0.00010433, -0.00031566, 0.00059122, -0.00141679, 0.00123038, 0.00123036, -0.00141683, 0.00059116, -0.00031574, 0.00010423, -0.00022471, 0.00021114, -0.00013265, 0.00001141, 0.00001214, 0.00002669, -0.00006439, -0.00007270, 0.00027302, -0.00042790, },
		{ 0.00065790, -0.00041659, 0.00004982, 0.00009726, -0.00007808, -0.00004779, 0.00006301, -0.00000486, -0.00006923, 0.00003317, 0.00004411, -0.00011996, -0.00008915, 0.00012317, -0.00014615, -0.00102389, 0.00259876, -0.00102389, -0.00014615, 0.00012317, -0.00008915, -0.00011996, 0.00004411, 0.00003317, -0.00006923, -0.00000486, 0.00006301, -0.00004779, -0.00007808, 0.00009726, 0.00004982, -0.00041659, },
		{ 0.00023731, -0.00042790, 0.00027302, -0.00007270, -0.00006439, 0.00002669, 0.00001214, 0.00001141, -0.00013265, 0.00021114, -0.00022471, 0.00010423, -0.00031574, 0.00059116, -0.00141683, 0.00123036, 0.00123038, -0.00141679, 0.00059122, -0.00031566, 0.00010433, -0.00022460, 0.00021125, -0.00013253, 0.00001153, 0.00001225, 0.00002680, -0.00006430, -0.00007262, 0.00027308, -0.00042787, 0.00023732, },
		{ -0.00042245, 0.00004789, 0.00009744, -0.00007828, -0.00004859, 0.00006271, -0.00000462, -0.00006936, 0.00003268, 0.00004418, -0.00011937, -0.00008914, 0.00012285, -0.00014432, -0.00101808, 0.00258011, -0.00101803, -0.00014422, 0.00012299, -0.00008896, -0.00011916, 0.00004441, 0.00003293, -0.00006911, -0.00000437, 0.00006294, -0.00004838, -0.00007810, 0.00009758, 0.00004799, -0.00042240, 0.00067655, },
		{ -0.00044694, 0.00027907, -0.00007745, -0.00006125, 0.00002175, 0.00001584, 0.00000852, -0.00013000, 0.00020719, -0.00021999, 0.00010090, -0.00031115, 0.00058499, -0.00139786, 0.00121127, 0.00121131, -0.00139775, 0.00058516, -0.00031091, 0.00010118, -0.00021966, 0.00020755, -0.00012964, 0.00000889, 0.00001619, 0.00002207, -0.00006096, -0.00007721, 0.00027925, -0.00044683, 0.00025641, 0.00025638, },
		{ 0.00004227, 0.00009811, -0.00007868, -0.00005074, 0.00006206, -0.00000365, -0.00006948, 0.00003149, 0.00004461, -0.00011741, -0.00008890, 0.00012205, -0.00013879, -0.00100082, 0.00252489, -0.00100073, -0.00013860, 0.00012232, -0.00008856, -0.00011701, 0.00004506, 0.00003196, -0.00006899, -0.00000317, 0.00006251, -0.00005034, -0.00007833, 0.00009838, 0.00004246, -0.00043966, 0.00073177, -0.00043975, },
		{ 0.00029101, -0.00008668, -0.00005498, 0.00001217, 0.00002323, 0.00000298, -0.00012469, 0.00019958, -0.00021061, 0.00009445, -0.00030206, 0.00057294, -0.00136062, 0.00117383, 0.00117389, -0.00136045, 0.00057323, -0.00030168, 0.00009492, -0.00021009, 0.00020016, -0.00012409, 0.00000357, 0.00002381, 0.00001270, -0.00005452, -0.00008630, 0.00029129, -0.00048407, 0.00029385, 0.00029380, -0.00048425, },
		{ 0.00009925, -0.00007927, -0.00005417, 0.00006109, -0.00000200, -0.00006959, 0.00002962, 0.00004538, -0.00011417, -0.00008847, 0.00012079, -0.00012977, -0.00097278, 0.00243521, -0.00097264, -0.00012951, 0.00012118, -0.00008797, -0.00011358, 0.00004603, 0.00003031, -0.00006888, -0.00000131, 0.00006174, -0.00005358, -0.00007877, 0.00009964, 0.00003344, -0.00046770, 0.00082145, -0.00046783, 0.00003317, },
		{ -0.00010005, -0.00004583, -0.00000168, 0.00003402, -0.00000501, -0.00011691, 0.00018860, -0.00019695, 0.00008515, -0.00028883, 0.00055549, -0.00130654, 0.00111948, 0.00111956, -0.00130630, 0.00055587, -0.00028831, 0.00008577, -0.00019624, 0.00018937, -0.00011611, -0.00000421, 0.00003479, -0.00000096, -0.00004521, -0.00009954, 0.00030875, -0.00053816, 0.00034820, 0.00034812, -0.00053839, 0.00030837, },
		{ -0.00008002, -0.00005873, 0.00005982, 0.00000027, -0.00006968, 0.00002716, 0.00004647, -0.00010976, -0.00008785, 0.00011913, -0.00011763, -0.00093504, 0.00231453, -0.00093486, -0.00011728, 0.00011962, -0.00008722, -0.00010902, 0.00004730, 0.00002804, -0.00006878, 0.00000115, 0.00006065, -0.00005799, -0.00007939, 0.00010130, 0.00002130, -0.00050544, 0.00094213, -0.00050561, 0.00002095, 0.00010080, },
		{ -0.00003416, -0.00001926, 0.00004780, -0.00001513, -0.00010696, 0.00017465, -0.00017952, 0.00007333, -0.00027196, 0.00053329, -0.00123769, 0.00105031, 0.00105041, -0.00123741, 0.00053375, -0.00027133, 0.00007409, -0.00017865, 0.00017559, -0.00010599, -0.00001415, 0.00004874, -0.00001840, -0.00003340, -0.00011641, 0.00033095, -0.00060700, 0.00041737, 0.00041728, -0.00060729, 0.00033048, -0.00011703, },
		{ -0.00006427, 0.00005832, 0.00000307, -0.00006976, 0.00002420, 0.00004782, -0.00010436, -0.00008707, 0.00011712, -0.00010281, -0.00088904, 0.00216747, -0.00088883, -0.00010240, 0.00011771, -0.00008632, -0.00010348, 0.00004880, 0.00002524, -0.00006871, 0.00000411, 0.00005930, -0.00006339, -0.00008017, 0.00010331, 0.00000648, -0.00055144, 0.00108919, -0.00055164, 0.00000607, 0.00010272, -0.00008092, },
		{ -0.00003991, 0.00006403, -0.00002700, -0.00009523, 0.00015829, -0.00015899, 0.00005947, -0.00025210, 0.00050720, -0.00115673, 0.00096898, 0.00096909, -0.00115640, 0.00050773, -0.00025139, 0.00006034, -0.00015800, 0.00015936, -0.00009412, -0.00002588, 0.00006510, -0.00003892, -0.00001954, -0.00013626, 0.00035703, -0.00068797, 0.00049870, 0.00049859, -0.00068829, 0.00035651, -0.00013697, -0.00002040, },
		{ 0.00005663, 0.00000630, -0.00006982, 0.00002086, 0.00004940, -0.00009817, -0.00008615, 0.00011485, -0.00008589, -0.00083655, 0.00199970, -0.00083633, -0.00008544, 0.00011550, -0.00008532, -0.00009720, 0.00005049, 0.00002201, -0.00006865, 0.00000745, 0.00005772, -0.00006958, -0.00008109, 0.00010558, -0.00001044, -0.00060392, 0.00125696, -0.00060415, -0.00001089, 0.00010493, -0.00008192, -0.00007055, },
		{ 0.00008209, -0.00004015, -0.00008217, 0.00014014, -0.00013616, 0.00004409, -0.00023001, 0.00047823, -0.00106675, 0.00087861, 0.00087873, -0.00106640, 0.00047880, -0.00022924, 0.00004503, -0.00013509, 0.00014130, -0.00008096, -0.00003894, 0.00008325, -0.00006176, -0.00000416, -0.00015835, 0.00038600, -0.00077794, 0.00058907, 0.00058895, -0.00077829, 0.00038543, -0.00015912, -0.00000510, -0.00006283, },
		{ 0.00000984, -0.00006986, 0.00001726, 0.00005114, -0.00009143, -0.00008514, 0.00011240, -0.00006752, -0.00077960, 0.00181765, -0.00077936, -0.00006704, 0.00011309, -0.00008426, -0.00009040, 0.00005229, 0.00001848, -0.00006861, 0.00001106, 0.00005598, -0.00007632, -0.00008210, 0.00010803, -0.00002881, -0.00066088, 0.00143901, -0.00066112, -0.00002929, 0.00010734, -0.00008298, -0.00007735, 0.00005483, },
		{ -0.00005409, -0.00006828, 0.00012089, -0.00011189, 0.00002778, -0.00020655, 0.00044749, -0.00097123, 0.00078268, 0.00078281, -0.00097086, 0.00044808, -0.00020575, 0.00002876, -0.00011078, 0.00012210, -0.00006702, -0.00005284, 0.00010250, -0.00008602, 0.00001215, -0.00018181, 0.00041675, -0.00087346, 0.00068500, 0.00068488, -0.00087383, 0.00041615, -0.00018261, 0.00001118, -0.00008713, 0.00010129, },
		},
		{ { 0.00010196, -0.00017792, 0.00019320, -0.00011880, 0.00011561, -0.00039314, 0.00011897, -0.00003944, -0.00070222, 0.00181631, -0.00080471, 0.00002116, -0.00011022, 0.00021553, -0.00018720, 0.00007893, -0.00006779, 0.00007893, -0.00018720, 0.00021553, -0.00011022, 0.00002116, -0.00080471, 0.00181631, -0.00070222, -0.00003944, 0.00011897, -0.00039314, 0.00011561, -0.00011880, 0.00019320, -0.00017792, },
		{ -0.00001039, -0.00008529, 0.00018970, -0.00017610, 0.00013778, -0.00059167, 0.00065358, -0.00100294, 0.00080043, 0.00076878, -0.00080223, 0.00020692, -0.00007661, 0.00012377, -0.00014326, 0.00004670, 0.00003108, -0.00012823, 0.00008890, -0.00004395, 0.00018981, -0.00073396, 0.00043899, 0.00130422, -0.00124445, 0.00068596, -0.00055072, 0.00009404, -0.00016137, 0.00014702, -0.00003694, -0.00003325, },
		{ -0.00017325, 0.00014816, -0.00004414, 0.00004718, -0.00026414, -0.00008845, 0.00023539, -0.00088972, 0.00162608, -0.00051590, -0.00012546, -0.00004218, 0.00016427, -0.00014284, 0.00006168, -0.00006616, 0.00009571, -0.00022970, 0.00025780, -0.00017391, 0.00016559, -0.00104404, 0.00185810, -0.00035703, -0.00039139, 0.00034037, -0.00050841, 0.00017971, -0.00018961, 0.00022548, -0.00017319, 0.00010033, },
		{ -0.00012412, 0.00020885, -0.00016552, 0.00014669, -0.00056190, 0.00053211, -0.00069717, 0.00033400, 0.00098443, -0.00079013, 0.00018807, -0.00008677, 0.00013810, -0.00014842, 0.00006060, 0.00001391, -0.00010202, 0.00003438, 0.00000893, 0.00013588, -0.00058948, 0.00002836, 0.00178361, -0.00137002, 0.00060739, -0.00043905, 0.00001867, -0.00012089, 0.00008291, 0.00001719, -0.00005477, 0.00001252, },
		{ 0.00009709, 0.00002485, -0.00001544, -0.00014195, -0.00024849, 0.00040452, -0.00092130, 0.00132557, -0.00021775, -0.00025375, 0.00002006, 0.00011293, -0.00010267, 0.00004533, -0.00006133, 0.00011016, -0.00026372, 0.00028280, -0.00022330, 0.00028670, -0.00119929, 0.00173184, 0.00012361, -0.00077608, 0.00053820, -0.00059063, 0.00022948, -0.00024689, 0.00023941, -0.00015854, 0.00009550, -0.00016058, },
		{ 0.00020577, -0.00013366, 0.00012300, -0.00047246, 0.00035648, -0.00038358, -0.00004346, 0.00106985, -0.00070516, 0.00013933, -0.00007576, 0.00013417, -0.00014631, 0.00007267, -0.00000447, -0.00006492, -0.00003568, 0.00007640, 0.00004975, -0.00038428, -0.00041707, 0.00217548, -0.00134112, 0.00041308, -0.00026801, -0.00008007, -0.00005793, 0.00000265, 0.00007222, -0.00007368, 0.00003427, -0.00015120, },
		{ 0.00008017, -0.00006347, -0.00004509, -0.00033708, 0.00045901, -0.00082129, 0.00096534, 0.00005006, -0.00034679, 0.00006774, 0.00006966, -0.00007132, 0.00003060, -0.00005349, 0.00012016, -0.00028311, 0.00028421, -0.00025012, 0.00036601, -0.00124653, 0.00143993, 0.00069568, -0.00113974, 0.00067663, -0.00062465, 0.00025664, -0.00028228, 0.00023137, -0.00013442, 0.00008766, -0.00014195, 0.00004677, },
		{ -0.00008738, 0.00007385, -0.00034355, 0.00016880, -0.00011436, -0.00029759, 0.00102797, -0.00056516, 0.00007097, -0.00004822, 0.00011692, -0.00014026, 0.00008283, -0.00002354, -0.00001893, -0.00011442, 0.00015020, -0.00005902, -0.00014301, -0.00084483, 0.00242463, -0.00113905, 0.00011670, -0.00005873, -0.00019000, 0.00002088, -0.00008586, 0.00012281, -0.00008887, 0.00005385, -0.00016606, 0.00018472, },
		{ -0.00009081, 0.00001269, -0.00034307, 0.00041028, -0.00063179, 0.00060041, 0.00025411, -0.00039377, 0.00009477, 0.00004061, -0.00005133, 0.00001754, -0.00004293, 0.00012372, -0.00028316, 0.00025875, -0.00024906, 0.00039043, -0.00117622, 0.00100710, 0.00129895, -0.00142720, 0.00072724, -0.00060195, 0.00025589, -0.00028992, 0.00020034, -0.00010230, 0.00007710, -0.00011969, 0.00000313, 0.00011646, },
		{ 0.00001014, -0.00020120, 0.00001129, 0.00007064, -0.00041568, 0.00087964, -0.00039538, -0.00000423, -0.00001126, 0.00009304, -0.00013372, 0.00009099, -0.00004260, 0.00003242, -0.00019314, 0.00022067, -0.00017724, 0.00010401, -0.00120346, 0.00249200, -0.00076859, -0.00025110, 0.00016086, -0.00029674, 0.00010646, -0.00017322, 0.00016376, -0.00009947, 0.00007046, -0.00016992, 0.00015220, -0.00003531, },
		{ 0.00002449, -0.00026989, 0.00028760, -0.00040520, 0.00028193, 0.00037209, -0.00039152, 0.00009861, 0.00002897, -0.00004284, 0.00000555, -0.00003007, 0.00011922, -0.00026138, 0.00020680, -0.00021868, 0.00035435, -0.00099473, 0.00047739, 0.00186512, -0.00159028, 0.00067359, -0.00052200, 0.00022577, -0.00026744, 0.00014815, -0.00006450, 0.00006424, -0.00009613, -0.00002969, 0.00013172, -0.00009491, },
		{ -0.00007303, -0.00008062, 0.00014995, -0.00040755, 0.00065993, -0.00022451, -0.00007303, 0.00002674, 0.00006977, -0.00012965, 0.00009696, -0.00006077, 0.00008446, -0.00026248, 0.00027813, -0.00029004, 0.00032512, -0.00144981, 0.00236063, -0.00025834, -0.00064684, 0.00036017, -0.00038584, 0.00018854, -0.00024974, 0.00019082, -0.00010494, 0.00008361, -0.00016532, 0.00011578, 0.00001358, -0.00005535, },
		{ -0.00013485, 0.00013282, -0.00019532, 0.00004955, 0.00039579, -0.00034479, 0.00008062, 0.00003458, -0.00004367, -0.00000637, -0.00001540, 0.00010575, -0.00021798, 0.00013242, -0.00016164, 0.00026055, -0.00072352, -0.00009172, 0.00232732, -0.00159532, 0.00051407, -0.00039256, 0.00016892, -0.00021633, 0.00007929, -0.00002394, 0.00004956, -0.00007327, -0.00004979, 0.00012744, -0.00007710, -0.00000877, },
		{ -0.00008415, 0.00012349, -0.00030301, 0.00041230, -0.00008017, -0.00012381, 0.00005747, 0.00005368, -0.00012998, 0.00010049, -0.00007708, 0.00013195, -0.00031377, 0.00031431, -0.00038293, 0.00049194, -0.00155516, 0.00203853, 0.00034258, -0.00102026, 0.00051048, -0.00044490, 0.00025714, -0.00030679, 0.00020117, -0.00010509, 0.00009307, -0.00015566, 0.00008288, 0.00005137, -0.00010995, 0.00001616, },
		{ -0.00000693, -0.00004848, -0.00007429, 0.00033202, -0.00026541, 0.00004574, 0.00005400, -0.00004984, -0.00001939, 0.00000052, 0.00008329, -0.00015595, 0.00004296, -0.00008440, 0.00011985, -0.00039614, -0.00063717, 0.00262954, -0.00142861, 0.00026254, -0.00022864, 0.00009179, -0.00014177, 0.00000036, 0.00001620, 0.00003364, -0.00005251, -0.00005773, 0.00010828, -0.00004234, -0.00007858, 0.00003375, },
		{ 0.00001211, -0.00014622, 0.00018162, 0.00001532, -0.00014835, 0.00007410, 0.00004955, -0.00013526, 0.00010125, -0.00009051, 0.00016982, -0.00034023, 0.00032362, -0.00044394, 0.00058350, -0.00150926, 0.00155791, 0.00097094, -0.00132169, 0.00058923, -0.00046538, 0.00030397, -0.00033803, 0.00019383, -0.00010010, 0.00009888, -0.00014457, 0.00005956, 0.00007239, -0.00014316, 0.00004837, 0.00000703, },
		{ 0.00000414, -0.00008728, 0.00020119, -0.00017026, 0.00000168, 0.00008113, -0.00005627, -0.00003456, 0.00001708, 0.00005282, -0.00008075, -0.00005197, 0.00000371, -0.00005036, -0.00005338, -0.00109900, 0.00273462, -0.00109900, -0.00005338, -0.00005036, 0.00000371, -0.00005197, -0.00008075, 0.00005282, 0.00001708, -0.00003456, -0.00005627, 0.00008113, 0.00000168, -0.00017026, 0.00020119, -0.00008728, },
		{ 0.00001211, 0.00000703, 0.00004837, -0.00014316, 0.00007239, 0.00005956, -0.00014457, 0.00009888, -0.00010010, 0.00019383, -0.00033803, 0.00030397, -0.00046538, 0.00058923, -0.00132169, 0.00097094, 0.00155791, -0.00150926, 0.00058350, -0.00044394, 0.00032362, -0.00034023, 0.00016982, -0.00009051, 0.00010125, -0.00013526, 0.00004955, 0.00007410, -0.00014835, 0.00001532, 0.00018162, -0.00014622, },
		{ -0.00000693, 0.00003375, -0.00007858, -0.00004234, 0.00010828, -0.00005773, -0.00005251, 0.00003364, 0.00001620, 0.00000036, -0.00014177, 0.00009179, -0.00022864, 0.00026254, -0.00142861, 0.00262954, -0.00063717, -0.00039614, 0.00011985, -0.00008440, 0.00004296, -0.00015595, 0.00008329, 0.00000052, -0.00001939, -0.00004984, 0.00005400, 0.00004574, -0.00026541, 0.00033202, -0.00007429, -0.00004848, },
		{ -0.00008415, 0.00001616, -0.00010995, 0.00005137, 0.00008288, -0.00015566, 0.00009307, -0.00010509, 0.00020117, -0.00030679, 0.00025714, -0.00044490, 0.00051048, -0.00102026, 0.00034258, 0.00203853, -0.00155516, 0.00049194, -0.00038293, 0.00031431, -0.00031377, 0.00013195, -0.00007708, 0.00010049, -0.00012998, 0.00005368, 0.00005747, -0.00012381, -0.00008017, 0.00041230, -0.00030301, 0.00012349, },
		{ -0.00013485, -0.00000877, -0.00007710, 0.00012744, -0.00004979, -0.00007327, 0.00004956, -0.00002394, 0.00007929, -0.00021633, 0.00016892, -0.00039256, 0.00051407, -0.00159532, 0.00232732, -0.00009172, -0.00072352, 0.00026055, -0.00016164, 0.00013242, -0.00021798, 0.00010575, -0.00001540, -0.00000637, -0.00004367, 0.00003458, 0.00008062, -0.00034479, 0.00039579, 0.00004955, -0.00019532, 0.00013282, },
		{ -0.00007303, -0.00005535, 0.00001358, 0.00011578, -0.00016532, 0.00008361, -0.00010494, 0.00019082, -0.00024974, 0.00018854, -0.00038584, 0.00036017, -0.00064684, -0.00025834, 0.00236063, -0.00144981, 0.00032512, -0.00029004, 0.00027813, -0.00026248, 0.00008446, -0.00006077, 0.00009696, -0.00012965, 0.00006977, 0.00002674, -0.00007303, -0.00022451, 0.00065993, -0.00040755, 0.00014995, -0.00008062, },
		{ 0.00002449, -0.00009491, 0.00013172, -0.00002969, -0.00009613, 0.00006424, -0.00006450, 0.00014815, -0.00026744, 0.00022577, -0.00052200, 0.00067359, -0.00159028, 0.00186512, 0.00047739, -0.00099473, 0.00035435, -0.00021868, 0.00020680, -0.00026138, 0.00011922, -0.00003007, 0.00000555, -0.00004284, 0.00002897, 0.00009861, -0.00039152, 0.00037209, 0.00028193, -0.00040520, 0.00028760, -0.00026989, },
		{ 0.00001014, -0.00003531, 0.00015220, -0.00016992, 0.00007046, -0.00009947, 0.00016376, -0.00017322, 0.00010646, -0.00029674, 0.00016086, -0.00025110, -0.00076859, 0.00249200, -0.00120346, 0.00010401, -0.00017724, 0.00022067, -0.00019314, 0.00003242, -0.00004260, 0.00009099, -0.00013372, 0.00009304, -0.00001126, -0.00000423, -0.00039538, 0.00087964, -0.00041568, 0.00007064, 0.00001129, -0.00020120, },
		{ -0.00009081, 0.00011646, 0.00000313, -0.00011969, 0.00007710, -0.00010230, 0.00020034, -0.00028992, 0.00025589, -0.00060195, 0.00072724, -0.00142720, 0.00129895, 0.00100710, -0.00117622, 0.00039043, -0.00024906, 0.00025875, -0.00028316, 0.00012372, -0.00004293, 0.00001754, -0.00005133, 0.00004061, 0.00009477, -0.00039377, 0.00025411, 0.00060041, -0.00063179, 0.00041028, -0.00034307, 0.00001269, },
		{ -0.00008738, 0.00018472, -0.00016606, 0.00005385, -0.00008887, 0.00012281, -0.00008586, 0.00002088, -0.00019000, -0.00005873, 0.00011670, -0.00113905, 0.00242463, -0.00084483, -0.00014301, -0.00005902, 0.00015020, -0.00011442, -0.00001893, -0.00002354, 0.00008283, -0.00014026, 0.00011692, -0.00004822, 0.00007097, -0.00056516, 0.00102797, -0.00029759, -0.00011436, 0.00016880, -0.00034355, 0.00007385, },
		{ 0.00008017, 0.00004677, -0.00014195, 0.00008766, -0.00013442, 0.00023137, -0.00028228, 0.00025664, -0.00062465, 0.00067663, -0.00113974, 0.00069568, 0.00143993, -0.00124653, 0.00036601, -0.00025012, 0.00028421, -0.00028311, 0.00012016, -0.00005349, 0.00003060, -0.00007132, 0.00006966, 0.00006774, -0.00034679, 0.00005006, 0.00096534, -0.00082129, 0.00045901, -0.00033708, -0.00004509, -0.00006347, },
		{ 0.00020577, -0.00015120, 0.00003427, -0.00007368, 0.00007222, 0.00000265, -0.00005793, -0.00008007, -0.00026801, 0.00041308, -0.00134112, 0.00217548, -0.00041707, -0.00038428, 0.00004975, 0.00007640, -0.00003568, -0.00006492, -0.00000447, 0.00007267, -0.00014631, 0.00013417, -0.00007576, 0.00013933, -0.00070516, 0.00106985, -0.00004346, -0.00038358, 0.00035648, -0.00047246, 0.00012300, -0.00013366, },
		{ 0.00009709, -0.00016058, 0.00009550, -0.00015854, 0.00023941, -0.00024689, 0.00022948, -0.00059063, 0.00053820, -0.00077608, 0.00012361, 0.00173184, -0.00119929, 0.00028670, -0.00022330, 0.00028280, -0.00026372, 0.00011016, -0.00006133, 0.00004533, -0.00010267, 0.00011293, 0.00002006, -0.00025375, -0.00021775, 0.00132557, -0.00092130, 0.00040452, -0.00024849, -0.00014195, -0.00001544, 0.00002485, },
		{ -0.00012412, 0.00001252, -0.00005477, 0.00001719, 0.00008291, -0.00012089, 0.00001867, -0.00043905, 0.00060739, -0.00137002, 0.00178361, 0.00002836, -0.00058948, 0.00013588, 0.00000893, 0.00003438, -0.00010202, 0.00001391, 0.00006060, -0.00014842, 0.00013810, -0.00008677, 0.00018807, -0.00079013, 0.00098443, 0.00033400, -0.00069717, 0.00053211, -0.00056190, 0.00014669, -0.00016552, 0.00020885, },
		{ -0.00017325, 0.00010033, -0.00017319, 0.00022548, -0.00018961, 0.00017971, -0.00050841, 0.00034037, -0.00039139, -0.00035703, 0.00185810, -0.00104404, 0.00016559, -0.00017391, 0.00025780, -0.00022970, 0.00009571, -0.00006616, 0.00006168, -0.00014284, 0.00016427, -0.00004218, -0.00012546, -0.00051590, 0.00162608, -0.00088972, 0.00023539, -0.00008845, -0.00026414, 0.00004718, -0.00004414, 0.00014816, },
		{ -0.00001039, -0.00003325, -0.00003694, 0.00014702, -0.00016137, 0.00009404, -0.00055072, 0.00068596, -0.00124445, 0.00130422, 0.00043899, -0.00073396, 0.00018981, -0.00004395, 0.00008890, -0.00012823, 0.00003108, 0.00004670, -0.00014326, 0.00012377, -0.00007661, 0.00020692, -0.00080223, 0.00076878, 0.00080043, -0.00100294, 0.00065358, -0.00059167, 0.00013778, -0.00017610, 0.00018970, -0.00008529, },
		},
		{ { -0.00015699, 0.00031144, -0.00045480, 0.00007252, -0.00000646, 0.00016804, -0.00020457, 0.00001333, 0.00007463, -0.00088941, 0.00189741, -0.00079230, 0.00001185, -0.00012801, 0.00015382, -0.00011511, 0.00001852, -0.00011511, 0.00015382, -0.00012801, 0.00001185, -0.00079230, 0.00189741, -0.00088941, 0.00007463, 0.00001333, -0.00020457, 0.00016804, -0.00000646, 0.00007252, -0.00045480, 0.00031144, },
		{ -0.00001715, 0.00015342, -0.00051307, 0.00030877, -0.00019201, 0.00036538, -0.00058466, 0.00063097, -0.00090029, 0.00044215, 0.00108822, -0.00090078, 0.00022678, -0.00014936, 0.00014176, -0.00010068, -0.00006950, 0.00007683, -0.00009103, 0.00017434, -0.00067014, 0.00030513, 0.00143935, -0.00136087, 0.00073032, -0.00057870, 0.00028977, -0.00016803, 0.00029376, -0.00035450, -0.00007193, 0.00014206, },
		{ 0.00023104, -0.00023271, -0.00020252, 0.00017708, -0.00002626, 0.00011104, -0.00037076, 0.00055184, -0.00119528, 0.00150419, -0.00014030, -0.00035171, 0.00002961, 0.00005959, -0.00004575, 0.00001525, -0.00016660, 0.00022206, -0.00025983, 0.00029462, -0.00113872, 0.00169604, -0.00003507, -0.00066235, 0.00048380, -0.00053458, 0.00035969, -0.00019719, 0.00029892, -0.00055154, 0.00026808, -0.00010534, },
		{ 0.00027213, -0.00049171, 0.00015613, -0.00007967, 0.00026666, -0.00036313, 0.00026669, -0.00023824, -0.00043398, 0.00146536, -0.00078328, 0.00009620, -0.00011523, 0.00015077, -0.00011115, -0.00001483, -0.00003839, 0.00005125, -0.00003424, -0.00020876, -0.00058648, 0.00214977, -0.00130385, 0.00040339, -0.00027521, 0.00000956, 0.00001268, 0.00011537, -0.00007443, -0.00031982, 0.00027321, -0.00013046, },
		{ 0.00001047, -0.00039454, 0.00023907, -0.00011219, 0.00027669, -0.00051421, 0.00064197, -0.00097450, 0.00074250, 0.00053826, -0.00060736, 0.00011364, -0.00000703, 0.00001471, 0.00000052, -0.00016755, 0.00020945, -0.00027222, 0.00035020, -0.00103716, 0.00092149, 0.00111992, -0.00133878, 0.00077860, -0.00068655, 0.00040437, -0.00026582, 0.00036465, -0.00046888, 0.00010185, 0.00002522, 0.00007954, },
		{ -0.00032110, -0.00007323, 0.00006364, 0.00011120, -0.00007894, -0.00012028, 0.00029996, -0.00089813, 0.00133655, -0.00037263, -0.00013963, -0.00003589, 0.00012744, -0.00011036, 0.00004891, -0.00016220, 0.00021045, -0.00027948, 0.00025175, -0.00122580, 0.00223769, -0.00060025, -0.00030575, 0.00022590, -0.00037336, 0.00026051, -0.00012573, 0.00020476, -0.00048669, 0.00031079, -0.00015016, 0.00025636, },
		{ -0.00042345, 0.00015012, -0.00007656, 0.00027687, -0.00042441, 0.00042723, -0.00046112, -0.00004214, 0.00096866, -0.00062672, 0.00007795, -0.00002481, 0.00005825, -0.00003501, -0.00010143, 0.00009736, -0.00013604, 0.00015530, -0.00055611, -0.00014643, 0.00213921, -0.00157715, 0.00066538, -0.00052742, 0.00022075, -0.00015074, 0.00024011, -0.00023422, -0.00012889, 0.00017349, -0.00006601, 0.00017432, },
		{ -0.00024978, 0.00012495, 0.00001210, 0.00012256, -0.00034863, 0.00053160, -0.00088704, 0.00082780, 0.00013773, -0.00034633, 0.00002179, 0.00010171, -0.00010180, 0.00009105, -0.00023363, 0.00029583, -0.00041586, 0.00048580, -0.00136148, 0.00161158, 0.00058703, -0.00112175, 0.00067551, -0.00066365, 0.00042198, -0.00028849, 0.00035777, -0.00049913, 0.00022968, -0.00007507, 0.00014077, -0.00009828, },
		{ -0.00001888, 0.00000162, 0.00019507, -0.00022396, 0.00010634, 0.00004848, -0.00055662, 0.00101183, -0.00040875, -0.00004611, -0.00001154, 0.00009274, -0.00009307, 0.00001479, -0.00007665, 0.00009055, -0.00017713, 0.00005883, -0.00108920, 0.00258862, -0.00113674, 0.00008983, -0.00008349, -0.00014259, 0.00009765, -0.00000355, 0.00005328, -0.00033312, 0.00027110, -0.00014231, 0.00020965, -0.00030036, },
		{ 0.00006259, 0.00000174, 0.00019516, -0.00037644, 0.00046793, -0.00055052, 0.00020672, 0.00052106, -0.00041209, 0.00001650, 0.00008739, -0.00007795, 0.00007900, -0.00020575, 0.00024639, -0.00035504, 0.00039088, -0.00097354, 0.00048608, 0.00183987, -0.00164241, 0.00079462, -0.00067964, 0.00037195, -0.00027308, 0.00032104, -0.00035076, 0.00005776, 0.00005297, -0.00000231, 0.00006900, -0.00028280, },
		{ 0.00001948, 0.00013216, -0.00004710, -0.00013967, 0.00034851, -0.00068465, 0.00071797, -0.00007454, -0.00017347, -0.00000415, 0.00012763, -0.00015824, 0.00013082, -0.00023132, 0.00029063, -0.00046034, 0.00051011, -0.00154323, 0.00224467, -0.00006334, -0.00075416, 0.00044293, -0.00051601, 0.00033970, -0.00023204, 0.00026984, -0.00042921, 0.00027470, -0.00012704, 0.00013492, -0.00011451, -0.00014474, },
		{ 0.00002035, 0.00019216, -0.00028233, 0.00025424, -0.00013520, -0.00026229, 0.00063399, -0.00031126, -0.00004173, 0.00007515, -0.00002688, -0.00000101, -0.00007686, 0.00007154, -0.00011781, 0.00003794, -0.00027965, -0.00070479, 0.00265815, -0.00153189, 0.00043349, -0.00036876, 0.00009671, -0.00008055, 0.00012558, -0.00010669, -0.00013711, 0.00017284, -0.00010299, 0.00011875, -0.00017267, -0.00006411, },
		{ 0.00012440, 0.00004819, -0.00024038, 0.00039915, -0.00050548, 0.00027952, 0.00020476, -0.00022146, -0.00002737, 0.00015990, -0.00019945, 0.00018392, -0.00028280, 0.00035354, -0.00053663, 0.00059818, -0.00136933, 0.00121272, 0.00129325, -0.00148517, 0.00076200, -0.00069547, 0.00042519, -0.00032093, 0.00032746, -0.00038289, 0.00018450, -0.00004273, 0.00001613, 0.00002264, -0.00013310, -0.00006595, },
		{ 0.00019722, -0.00017736, 0.00005123, 0.00015820, -0.00044253, 0.00048117, -0.00011565, -0.00010107, 0.00004456, 0.00005415, -0.00012949, 0.00009991, -0.00014677, 0.00017573, -0.00037168, 0.00037750, -0.00150429, 0.00268958, -0.00070490, -0.00031920, 0.00013972, -0.00028448, 0.00018482, -0.00012000, 0.00012743, -0.00027919, 0.00023460, -0.00012685, 0.00006085, -0.00000755, -0.00013386, -0.00002550, },
		{ 0.00009740, -0.00024011, 0.00029643, -0.00021077, -0.00008375, 0.00030423, -0.00016313, -0.00007261, 0.00017144, -0.00018410, 0.00013146, -0.00019250, 0.00023936, -0.00036303, 0.00031662, -0.00071117, -0.00011036, 0.00241566, -0.00170772, 0.00065004, -0.00056071, 0.00028292, -0.00022339, 0.00021612, -0.00022228, 0.00004084, 0.00006197, -0.00007547, 0.00004242, 0.00001766, -0.00019573, 0.00011856, },
		{ -0.00011532, -0.00006326, 0.00026314, -0.00036552, 0.00019579, 0.00005185, -0.00010600, -0.00001404, 0.00014824, -0.00025592, 0.00023722, -0.00029442, 0.00037269, -0.00061746, 0.00070176, -0.00164144, 0.00190057, 0.00059907, -0.00114675, 0.00058393, -0.00057686, 0.00037293, -0.00028721, 0.00025590, -0.00031951, 0.00022068, -0.00008420, -0.00003917, 0.00009281, -0.00004754, -0.00017618, 0.00024055, },
		{ -0.00022604, 0.00017231, 0.00002136, -0.00023965, 0.00021007, -0.00004495, -0.00010199, 0.00014077, -0.00009801, -0.00002247, 0.00000566, 0.00000521, -0.00003204, -0.00015621, 0.00008576, -0.00121450, 0.00284971, -0.00121450, 0.00008576, -0.00015621, -0.00003204, 0.00000521, 0.00000566, -0.00002247, -0.00009801, 0.00014077, -0.00010199, -0.00004495, 0.00021007, -0.00023965, 0.00002136, 0.00017231, },
		{ -0.00011532, 0.00024055, -0.00017618, -0.00004754, 0.00009281, -0.00003917, -0.00008420, 0.00022068, -0.00031951, 0.00025590, -0.00028721, 0.00037293, -0.00057686, 0.00058393, -0.00114675, 0.00059907, 0.00190057, -0.00164144, 0.00070176, -0.00061746, 0.00037269, -0.00029442, 0.00023722, -0.00025592, 0.00014824, -0.00001404, -0.00010600, 0.00005185, 0.00019579, -0.00036552, 0.00026314, -0.00006326, },
		{ 0.00009740, 0.00011856, -0.00019573, 0.00001766, 0.00004242, -0.00007547, 0.00006197, 0.00004084, -0.00022228, 0.00021612, -0.00022339, 0.00028292, -0.00056071, 0.00065004, -0.00170772, 0.00241566, -0.00011036, -0.00071117, 0.00031662, -0.00036303, 0.00023936, -0.00019250, 0.00013146, -0.00018410, 0.00017144, -0.00007261, -0.00016313, 0.00030423, -0.00008375, -0.00021077, 0.00029643, -0.00024011, },
		{ 0.00019722, -0.00002550, -0.00013386, -0.00000755, 0.00006085, -0.00012685, 0.00023460, -0.00027919, 0.00012743, -0.00012000, 0.00018482, -0.00028448, 0.00013972, -0.00031920, -0.00070490, 0.00268958, -0.00150429, 0.00037750, -0.00037168, 0.00017573, -0.00014677, 0.00009991, -0.00012949, 0.00005415, 0.00004456, -0.00010107, -0.00011565, 0.00048117, -0.00044253, 0.00015820, 0.00005123, -0.00017736, },
		{ 0.00012440, -0.00006595, -0.00013310, 0.00002264, 0.00001613, -0.00004273, 0.00018450, -0.00038289, 0.00032746, -0.00032093, 0.00042519, -0.00069547, 0.00076200, -0.00148517, 0.00129325, 0.00121272, -0.00136933, 0.00059818, -0.00053663, 0.00035354, -0.00028280, 0.00018392, -0.00019945, 0.00015990, -0.00002737, -0.00022146, 0.00020476, 0.00027952, -0.00050548, 0.00039915, -0.00024038, 0.00004819, },
		{ 0.00002035, -0.00006411, -0.00017267, 0.00011875, -0.00010299, 0.00017284, -0.00013711, -0.00010669, 0.00012558, -0.00008055, 0.00009671, -0.00036876, 0.00043349, -0.00153189, 0.00265815, -0.00070479, -0.00027965, 0.00003794, -0.00011781, 0.00007154, -0.00007686, -0.00000101, -0.00002688, 0.00007515, -0.00004173, -0.00031126, 0.00063399, -0.00026229, -0.00013520, 0.00025424, -0.00028233, 0.00019216, },
		{ 0.00001948, -0.00014474, -0.00011451, 0.00013492, -0.00012704, 0.00027470, -0.00042921, 0.00026984, -0.00023204, 0.00033970, -0.00051601, 0.00044293, -0.00075416, -0.00006334, 0.00224467, -0.00154323, 0.00051011, -0.00046034, 0.00029063, -0.00023132, 0.00013082, -0.00015824, 0.00012763, -0.00000415, -0.00017347, -0.00007454, 0.00071797, -0.00068465, 0.00034851, -0.00013967, -0.00004710, 0.00013216, },
		{ 0.00006259, -0.00028280, 0.00006900, -0.00000231, 0.00005297, 0.00005776, -0.00035076, 0.00032104, -0.00027308, 0.00037195, -0.00067964, 0.00079462, -0.00164241, 0.00183987, 0.00048608, -0.00097354, 0.00039088, -0.00035504, 0.00024639, -0.00020575, 0.00007900, -0.00007795, 0.00008739, 0.00001650, -0.00041209, 0.00052106, 0.00020672, -0.00055052, 0.00046793, -0.00037644, 0.00019516, 0.00000174, },
		{ -0.00001888, -0.00030036, 0.00020965, -0.00014231, 0.00027110, -0.00033312, 0.00005328, -0.00000355, 0.00009765, -0.00014259, -0.00008349, 0.00008983, -0.00113674, 0.00258862, -0.00108920, 0.00005883, -0.00017713, 0.00009055, -0.00007665, 0.00001479, -0.00009307, 0.00009274, -0.00001154, -0.00004611, -0.00040875, 0.00101183, -0.00055662, 0.00004848, 0.00010634, -0.00022396, 0.00019507, 0.00000162, },
		{ -0.00024978, -0.00009828, 0.00014077, -0.00007507, 0.00022968, -0.00049913, 0.00035777, -0.00028849, 0.00042198, -0.00066365, 0.00067551, -0.00112175, 0.00058703, 0.00161158, -0.00136148, 0.00048580, -0.00041586, 0.00029583, -0.00023363, 0.00009105, -0.00010180, 0.00010171, 0.00002179, -0.00034633, 0.00013773, 0.00082780, -0.00088704, 0.00053160, -0.00034863, 0.00012256, 0.00001210, 0.00012495, },
		{ -0.00042345, 0.00017432, -0.00006601, 0.00017349, -0.00012889, -0.00023422, 0.00024011, -0.00015074, 0.00022075, -0.00052742, 0.00066538, -0.00157715, 0.00213921, -0.00014643, -0.00055611, 0.00015530, -0.00013604, 0.00009736, -0.00010143, -0.00003501, 0.00005825, -0.00002481, 0.00007795, -0.00062672, 0.00096866, -0.00004214, -0.00046112, 0.00042723, -0.00042441, 0.00027687, -0.00007656, 0.00015012, },
		{ -0.00032110, 0.00025636, -0.00015016, 0.00031079, -0.00048669, 0.00020476, -0.00012573, 0.00026051, -0.00037336, 0.00022590, -0.00030575, -0.00060025, 0.00223769, -0.00122580, 0.00025175, -0.00027948, 0.00021045, -0.00016220, 0.00004891, -0.00011036, 0.00012744, -0.00003589, -0.00013963, -0.00037263, 0.00133655, -0.00089813, 0.00029996, -0.00012028, -0.00007894, 0.00011120, 0.00006364, -0.00007323, },
		{ 0.00001047, 0.00007954, 0.00002522, 0.00010185, -0.00046888, 0.00036465, -0.00026582, 0.00040437, -0.00068655, 0.00077860, -0.00133878, 0.00111992, 0.00092149, -0.00103716, 0.00035020, -0.00027222, 0.00020945, -0.00016755, 0.00000052, 0.00001471, -0.00000703, 0.00011364, -0.00060736, 0.00053826, 0.00074250, -0.00097450, 0.00064197, -0.00051421, 0.00027669, -0.00011219, 0.00023907, -0.00039454, },
		{ 0.00027213, -0.00013046, 0.00027321, -0.00031982, -0.00007443, 0.00011537, 0.00001268, 0.00000956, -0.00027521, 0.00040339, -0.00130385, 0.00214977, -0.00058648, -0.00020876, -0.00003424, 0.00005125, -0.00003839, -0.00001483, -0.00011115, 0.00015077, -0.00011523, 0.00009620, -0.00078328, 0.00146536, -0.00043398, -0.00023824, 0.00026669, -0.00036313, 0.00026666, -0.00007967, 0.00015613, -0.00049171, },
		{ 0.00023104, -0.00010534, 0.00026808, -0.00055154, 0.00029892, -0.00019719, 0.00035969, -0.00053458, 0.00048380, -0.00066235, -0.00003507, 0.00169604, -0.00113872, 0.00029462, -0.00025983, 0.00022206, -0.00016660, 0.00001525, -0.00004575, 0.00005959, 0.00002961, -0.00035171, -0.00014030, 0.00150419, -0.00119528, 0.00055184, -0.00037076, 0.00011104, -0.00002626, 0.00017708, -0.00020252, -0.00023271, },
		{ -0.00001715, 0.00014206, -0.00007193, -0.00035450, 0.00029376, -0.00016803, 0.00028977, -0.00057870, 0.00073032, -0.00136087, 0.00143935, 0.00030513, -0.00067014, 0.00017434, -0.00009103, 0.00007683, -0.00006950, -0.00010068, 0.00014176, -0.00014936, 0.00022678, -0.00090078, 0.00108822, 0.00044215, -0.00090029, 0.00063097, -0.00058466, 0.00036538, -0.00019201, 0.00030877, -0.00051307, 0.00015342, },
		},
		{ { 0.00021705, -0.00043308, 0.00022765, 0.00000689, 0.00003483, -0.00003770, 0.00004693, -0.00013104, -0.00009801, 0.00024335, -0.00085473, 0.00185216, -0.00071997, -0.00014347, -0.00009959, 0.00004553, -0.00001025, 0.00004553, -0.00009959, -0.00014347, -0.00071997, 0.00185216, -0.00085473, 0.00024335, -0.00009801, -0.00013104, 0.00004693, -0.00003770, 0.00003483, 0.00000689, 0.00022765, -0.00043308, },
		{ 0.00009432, -0.00040263, 0.00039345, -0.00020323, 0.00022091, -0.00023776, 0.00028351, -0.00050947, 0.00047088, -0.00052265, 0.00012164, 0.00132204, -0.00096422, 0.00012609, -0.00016203, 0.00006273, 0.00004886, -0.00010375, 0.00010922, -0.00069258, 0.00019261, 0.00150385, -0.00121186, 0.00071373, -0.00056121, 0.00025552, -0.00022929, 0.00019559, -0.00018403, 0.00028957, -0.00011417, -0.00021933, },
		{ -0.00017660, -0.00013641, 0.00029205, -0.00019162, 0.00020900, -0.00022027, 0.00024429, -0.00054654, 0.00069188, -0.00101608, 0.00111322, 0.00033431, -0.00065185, 0.00007918, -0.00006745, -0.00001104, 0.00013615, -0.00023661, 0.00022069, -0.00113405, 0.00137688, 0.00041161, -0.00077926, 0.00057762, -0.00058335, 0.00034592, -0.00029501, 0.00024961, -0.00022799, 0.00041541, -0.00040247, 0.00006508, },
		{ -0.00038181, 0.00018277, 0.00003307, -0.00000246, 0.00001882, -0.00000014, -0.00007464, -0.00016834, 0.00036064, -0.00084350, 0.00152292, -0.00048661, -0.00020107, -0.00006229, 0.00002410, -0.00002325, 0.00008899, -0.00016245, -0.00003994, -0.00097158, 0.00212264, -0.00078058, 0.00006152, 0.00000027, -0.00019978, 0.00010901, -0.00010323, 0.00007735, -0.00003334, 0.00026354, -0.00044666, 0.00020237, },
		{ -0.00036038, 0.00034200, -0.00017194, 0.00018054, -0.00017060, 0.00021584, -0.00042226, 0.00035674, -0.00029913, -0.00008767, 0.00117523, -0.00076102, 0.00004179, -0.00009496, -0.00000128, 0.00009027, -0.00012071, 0.00011272, -0.00066608, -0.00002278, 0.00189715, -0.00135817, 0.00067673, -0.00053742, 0.00024349, -0.00021324, 0.00016187, -0.00016026, 0.00025482, -0.00007374, -0.00024240, 0.00010120, },
		{ -0.00012752, 0.00026005, -0.00018188, 0.00020042, -0.00018798, 0.00021229, -0.00049600, 0.00062158, -0.00079885, 0.00075942, 0.00039262, -0.00055277, 0.00003093, -0.00001793, -0.00008084, 0.00020922, -0.00030963, 0.00031135, -0.00124779, 0.00132797, 0.00076381, -0.00104202, 0.00066002, -0.00063448, 0.00039434, -0.00033378, 0.00026128, -0.00023985, 0.00040006, -0.00036162, 0.00003173, -0.00013783, },
		{ 0.00014580, 0.00003668, -0.00002742, 0.00005791, -0.00002624, -0.00003664, -0.00020201, 0.00040859, -0.00076189, 0.00117027, -0.00029526, -0.00020768, -0.00005162, 0.00002611, -0.00006027, 0.00014975, -0.00024604, 0.00009800, -0.00121337, 0.00230355, -0.00061675, -0.00017091, 0.00011370, -0.00027266, 0.00017766, -0.00016790, 0.00011721, -0.00007734, 0.00027441, -0.00041061, 0.00016337, -0.00031206, },
		{ 0.00027369, -0.00014223, 0.00013594, -0.00010365, 0.00015210, -0.00033238, 0.00025231, -0.00012969, -0.00020977, 0.00095862, -0.00054966, -0.00001993, -0.00004309, -0.00004762, 0.00010740, -0.00011495, 0.00008485, -0.00057174, -0.00029184, 0.00225547, -0.00143196, 0.00057799, -0.00047752, 0.00020903, -0.00017424, 0.00011291, -0.00012444, 0.00019515, -0.00003264, -0.00022407, 0.00007539, -0.00028311, },
		{ 0.00019691, -0.00015683, 0.00017178, -0.00013780, 0.00016491, -0.00041666, 0.00051897, -0.00058894, 0.00046881, 0.00036917, -0.00041104, -0.00002503, 0.00003700, -0.00015049, 0.00027209, -0.00037191, 0.00038393, -0.00128858, 0.00117579, 0.00114968, -0.00128163, 0.00070388, -0.00065599, 0.00042183, -0.00034832, 0.00025348, -0.00023520, 0.00034577, -0.00029056, 0.00001708, -0.00012428, -0.00008149, },
		{ 0.00001243, -0.00003494, 0.00007390, -0.00002877, -0.00002017, -0.00019561, 0.00039173, -0.00063281, 0.00082930, -0.00016219, -0.00016526, -0.00006483, 0.00004948, -0.00011561, 0.00022038, -0.00034224, 0.00025403, -0.00141684, 0.00237294, -0.00037094, -0.00043250, 0.00022728, -0.00034049, 0.00024323, -0.00022184, 0.00014693, -0.00011478, 0.00024794, -0.00032244, 0.00011433, -0.00024692, 0.00013159, },
		{ -0.00012253, 0.00009542, -0.00004697, 0.00010063, -0.00025017, 0.00017227, -0.00002660, -0.00025004, 0.00070137, -0.00035420, -0.00005070, -0.00001165, -0.00007073, 0.00009811, -0.00008632, 0.00002526, -0.00041748, -0.00058678, 0.00254247, -0.00141864, 0.00042292, -0.00038850, 0.00015620, -0.00011800, 0.00005641, -0.00008305, 0.00012187, -0.00000983, -0.00014852, 0.00001639, -0.00018618, 0.00020387, },
		{ -0.00012135, 0.00012834, -0.00007791, 0.00010962, -0.00031912, 0.00040316, -0.00041165, 0.00025704, 0.00027629, -0.00024677, -0.00007782, 0.00008894, -0.00020993, 0.00031630, -0.00041548, 0.00042664, -0.00124789, 0.00093333, 0.00153401, -0.00147046, 0.00069999, -0.00064463, 0.00042426, -0.00033659, 0.00022710, -0.00021385, 0.00025750, -0.00020681, 0.00004441, -0.00015105, -0.00000037, 0.00011108, },
		{ -0.00002265, 0.00006475, -0.00000879, -0.00002500, -0.00015149, 0.00032278, -0.00048321, 0.00053024, -0.00009424, -0.00008240, -0.00009600, 0.00008907, -0.00018078, 0.00029185, -0.00044077, 0.00040930, -0.00155707, 0.00232085, -0.00006226, -0.00069702, 0.00032602, -0.00039446, 0.00029623, -0.00025669, 0.00016067, -0.00013704, 0.00017784, -0.00018980, 0.00007627, -0.00020938, 0.00015026, -0.00004076, },
		{ 0.00006687, -0.00000914, 0.00006767, -0.00018421, 0.00012667, 0.00000787, -0.00022403, 0.00043536, -0.00019424, -0.00004720, -0.00000269, -0.00006855, 0.00006418, -0.00003797, -0.00006175, -0.00021883, -0.00087577, 0.00272800, -0.00131446, 0.00022482, -0.00028098, 0.00009174, -0.00005293, 0.00000146, -0.00004392, 0.00004844, -0.00002237, -0.00000898, -0.00006623, -0.00009040, 0.00014807, -0.00012019, },
		{ 0.00007780, -0.00001798, 0.00005484, -0.00021584, 0.00029327, -0.00028520, 0.00012777, 0.00013477, -0.00008144, -0.00011791, 0.00013025, -0.00025065, 0.00033600, -0.00043477, 0.00043180, -0.00112699, 0.00062448, 0.00187931, -0.00158648, 0.00064564, -0.00060120, 0.00040096, -0.00030051, 0.00018621, -0.00017905, 0.00014616, -0.00013201, 0.00013307, -0.00022425, 0.00010578, 0.00001488, -0.00008236, },
		{ 0.00003231, 0.00002916, -0.00004773, -0.00007737, 0.00022036, -0.00033968, 0.00029465, -0.00008822, 0.00002705, -0.00013703, 0.00013746, -0.00024575, 0.00035481, -0.00053054, 0.00054513, -0.00161630, 0.00215077, 0.00028089, -0.00093691, 0.00039712, -0.00042738, 0.00032883, -0.00026693, 0.00015516, -0.00013858, 0.00006517, -0.00002953, 0.00007319, -0.00021834, 0.00020541, -0.00011942, 0.00000859, },
		{ 0.00000414, 0.00005646, -0.00014020, 0.00011946, -0.00001866, -0.00015430, 0.00019058, -0.00008229, -0.00001169, -0.00001490, -0.00004277, 0.00001107, 0.00002406, -0.00016770, 0.00000318, -0.00112729, 0.00279217, -0.00112729, 0.00000318, -0.00016770, 0.00002406, 0.00001107, -0.00004277, -0.00001490, -0.00001169, -0.00008229, 0.00019058, -0.00015430, -0.00001866, 0.00011946, -0.00014020, 0.00005646, },
		{ 0.00003231, 0.00000859, -0.00011942, 0.00020541, -0.00021834, 0.00007319, -0.00002953, 0.00006517, -0.00013858, 0.00015516, -0.00026693, 0.00032883, -0.00042738, 0.00039712, -0.00093691, 0.00028089, 0.00215077, -0.00161630, 0.00054513, -0.00053054, 0.00035481, -0.00024575, 0.00013746, -0.00013703, 0.00002705, -0.00008822, 0.00029465, -0.00033968, 0.00022036, -0.00007737, -0.00004773, 0.00002916, },
		{ 0.00007780, -0.00008236, 0.00001488, 0.00010578, -0.00022425, 0.00013307, -0.00013201, 0.00014616, -0.00017905, 0.00018621, -0.00030051, 0.00040096, -0.00060120, 0.00064564, -0.00158648, 0.00187931, 0.00062448, -0.00112699, 0.00043180, -0.00043477, 0.00033600, -0.00025065, 0.00013025, -0.00011791, -0.00008144, 0.00013477, 0.00012777, -0.00028520, 0.00029327, -0.00021584, 0.00005484, -0.00001798, },
		{ 0.00006687, -0.00012019, 0.00014807, -0.00009040, -0.00006623, -0.00000898, -0.00002237, 0.00004844, -0.00004392, 0.00000146, -0.00005293, 0.00009174, -0.00028098, 0.00022482, -0.00131446, 0.00272800, -0.00087577, -0.00021883, -0.00006175, -0.00003797, 0.00006418, -0.00006855, -0.00000269, -0.00004720, -0.00019424, 0.00043536, -0.00022403, 0.00000787, 0.00012667, -0.00018421, 0.00006767, -0.00000914, },
		{ -0.00002265, -0.00004076, 0.00015026, -0.00020938, 0.00007627, -0.00018980, 0.00017784, -0.00013704, 0.00016067, -0.00025669, 0.00029623, -0.00039446, 0.00032602, -0.00069702, -0.00006226, 0.00232085, -0.00155707, 0.00040930, -0.00044077, 0.00029185, -0.00018078, 0.00008907, -0.00009600, -0.00008240, -0.00009424, 0.00053024, -0.00048321, 0.00032278, -0.00015149, -0.00002500, -0.00000879, 0.00006475, },
		{ -0.00012135, 0.00011108, -0.00000037, -0.00015105, 0.00004441, -0.00020681, 0.00025750, -0.00021385, 0.00022710, -0.00033659, 0.00042426, -0.00064463, 0.00069999, -0.00147046, 0.00153401, 0.00093333, -0.00124789, 0.00042664, -0.00041548, 0.00031630, -0.00020993, 0.00008894, -0.00007782, -0.00024677, 0.00027629, 0.00025704, -0.00041165, 0.00040316, -0.00031912, 0.00010962, -0.00007791, 0.00012834, },
		{ -0.00012253, 0.00020387, -0.00018618, 0.00001639, -0.00014852, -0.00000983, 0.00012187, -0.00008305, 0.00005641, -0.00011800, 0.00015620, -0.00038850, 0.00042292, -0.00141864, 0.00254247, -0.00058678, -0.00041748, 0.00002526, -0.00008632, 0.00009811, -0.00007073, -0.00001165, -0.00005070, -0.00035420, 0.00070137, -0.00025004, -0.00002660, 0.00017227, -0.00025017, 0.00010063, -0.00004697, 0.00009542, },
		{ 0.00001243, 0.00013159, -0.00024692, 0.00011433, -0.00032244, 0.00024794, -0.00011478, 0.00014693, -0.00022184, 0.00024323, -0.00034049, 0.00022728, -0.00043250, -0.00037094, 0.00237294, -0.00141684, 0.00025403, -0.00034224, 0.00022038, -0.00011561, 0.00004948, -0.00006483, -0.00016526, -0.00016219, 0.00082930, -0.00063281, 0.00039173, -0.00019561, -0.00002017, -0.00002877, 0.00007390, -0.00003494, },
		{ 0.00019691, -0.00008149, -0.00012428, 0.00001708, -0.00029056, 0.00034577, -0.00023520, 0.00025348, -0.00034832, 0.00042183, -0.00065599, 0.00070388, -0.00128163, 0.00114968, 0.00117579, -0.00128858, 0.00038393, -0.00037191, 0.00027209, -0.00015049, 0.00003700, -0.00002503, -0.00041104, 0.00036917, 0.00046881, -0.00058894, 0.00051897, -0.00041666, 0.00016491, -0.00013780, 0.00017178, -0.00015683, },
		{ 0.00027369, -0.00028311, 0.00007539, -0.00022407, -0.00003264, 0.00019515, -0.00012444, 0.00011291, -0.00017424, 0.00020903, -0.00047752, 0.00057799, -0.00143196, 0.00225547, -0.00029184, -0.00057174, 0.00008485, -0.00011495, 0.00010740, -0.00004762, -0.00004309, -0.00001993, -0.00054966, 0.00095862, -0.00020977, -0.00012969, 0.00025231, -0.00033238, 0.00015210, -0.00010365, 0.00013594, -0.00014223, },
		{ 0.00014580, -0.00031206, 0.00016337, -0.00041061, 0.00027441, -0.00007734, 0.00011721, -0.00016790, 0.00017766, -0.00027266, 0.00011370, -0.00017091, -0.00061675, 0.00230355, -0.00121337, 0.00009800, -0.00024604, 0.00014975, -0.00006027, 0.00002611, -0.00005162, -0.00020768, -0.00029526, 0.00117027, -0.00076189, 0.00040859, -0.00020201, -0.00003664, -0.00002624, 0.00005791, -0.00002742, 0.00003668, },
		{ -0.00012752, -0.00013783, 0.00003173, -0.00036162, 0.00040006, -0.00023985, 0.00026128, -0.00033378, 0.00039434, -0.00063448, 0.00066002, -0.00104202, 0.00076381, 0.00132797, -0.00124779, 0.00031135, -0.00030963, 0.00020922, -0.00008084, -0.00001793, 0.00003093, -0.00055277, 0.00039262, 0.00075942, -0.00079885, 0.00062158, -0.00049600, 0.00021229, -0.00018798, 0.00020042, -0.00018188, 0.00026005, },
		{ -0.00036038, 0.00010120, -0.00024240, -0.00007374, 0.00025482, -0.00016026, 0.00016187, -0.00021324, 0.00024349, -0.00053742, 0.00067673, -0.00135817, 0.00189715, -0.00002278, -0.00066608, 0.00011272, -0.00012071, 0.00009027, -0.00000128, -0.00009496, 0.00004179, -0.00076102, 0.00117523, -0.00008767, -0.00029913, 0.00035674, -0.00042226, 0.00021584, -0.00017060, 0.00018054, -0.00017194, 0.00034200, },
		{ -0.00038181, 0.00020237, -0.00044666, 0.00026354, -0.00003334, 0.00007735, -0.00010323, 0.00010901, -0.00019978, 0.00000027, 0.00006152, -0.00078058, 0.00212264, -0.00097158, -0.00003994, -0.00016245, 0.00008899, -0.00002325, 0.00002410, -0.00006229, -0.00020107, -0.00048661, 0.00152292, -0.00084350, 0.00036064, -0.00016834, -0.00007464, -0.00000014, 0.00001882, -0.00000246, 0.00003307, 0.00018277, },
		{ -0.00017660, 0.00006508, -0.00040247, 0.00041541, -0.00022799, 0.00024961, -0.00029501, 0.00034592, -0.00058335, 0.00057762, -0.00077926, 0.00041161, 0.00137688, -0.00113405, 0.00022069, -0.00023661, 0.00013615, -0.00001104, -0.00006745, 0.00007918, -0.00065185, 0.00033431, 0.00111322, -0.00101608, 0.00069188, -0.00054654, 0.00024429, -0.00022027, 0.00020900, -0.00019162, 0.00029205, -0.00013641, },
		{ 0.00009432, -0.00021933, -0.00011417, 0.00028957, -0.00018403, 0.00019559, -0.00022929, 0.00025552, -0.00056121, 0.00071373, -0.00121186, 0.00150385, 0.00019261, -0.00069258, 0.00010922, -0.00010375, 0.00004886, 0.00006273, -0.00016203, 0.00012609, -0.00096422, 0.00132204, 0.00012164, -0.00052265, 0.00047088, -0.00050947, 0.00028351, -0.00023776, 0.00022091, -0.00020323, 0.00039345, -0.00040263, },
		},
		{ { -0.00009944, 0.00006422, 0.00002901, 0.00002754, -0.00021634, 0.00009211, -0.00011721, 0.00018512, -0.00008362, -0.00015879, 0.00012260, -0.00079820, 0.00177544, -0.00073363, -0.00002361, -0.00003788, -0.00006779, -0.00003788, -0.00002361, -0.00073363, 0.00177544, -0.00079820, 0.00012260, -0.00015879, -0.00008362, 0.00018512, -0.00011721, 0.00009211, -0.00021634, 0.00002754, 0.00002901, 0.00006422, },
		{ -0.00007222, 0.00009120, -0.00000134, 0.00001745, -0.00024032, 0.00019045, -0.00024273, 0.00039730, -0.00043201, 0.00027357, -0.00045017, -0.00003946, 0.00135698, -0.00090826, 0.00019114, -0.00010094, -0.00011397, 0.00013211, -0.00045124, -0.00010547, 0.00163833, -0.00119155, 0.00053257, -0.00052390, 0.00026830, -0.00007592, 0.00006392, -0.00006810, -0.00013100, 0.00000490, 0.00008005, -0.00000335, },
		{ -0.00000569, 0.00007068, 0.00000500, -0.00004171, -0.00017307, 0.00016903, -0.00022381, 0.00043051, -0.00059582, 0.00054246, -0.00090051, 0.00077172, 0.00061166, -0.00070064, 0.00018362, -0.00006432, -0.00020025, 0.00029694, -0.00091390, 0.00081941, 0.00095077, -0.00107793, 0.00059144, -0.00063644, 0.00045333, -0.00024376, 0.00018810, -0.00020285, -0.00004279, -0.00001177, 0.00011508, -0.00007817, },
		{ 0.00006349, 0.00002396, 0.00003413, -0.00012958, -0.00003648, 0.00003718, -0.00005114, 0.00024338, -0.00048344, 0.00049581, -0.00099239, 0.00130476, -0.00012644, -0.00033108, 0.00007221, -0.00002300, -0.00021195, 0.00028417, -0.00113110, 0.00169631, -0.00003237, -0.00055481, 0.00029962, -0.00045161, 0.00040102, -0.00024292, 0.00018490, -0.00024751, -0.00000030, 0.00000988, 0.00010028, -0.00011868, },
		{ 0.00009817, -0.00001436, 0.00005287, -0.00019881, 0.00010360, -0.00012923, 0.00019316, -0.00008671, -0.00014691, 0.00014489, -0.00066119, 0.00136933, -0.00058526, -0.00002703, -0.00000474, -0.00006810, -0.00006906, -0.00001918, -0.00086736, 0.00214328, -0.00091809, 0.00009570, -0.00016406, -0.00008053, 0.00017048, -0.00010057, 0.00006352, -0.00019560, -0.00001241, 0.00007136, 0.00002831, -0.00009914, },
		{ 0.00008194, -0.00001868, 0.00003209, -0.00020359, 0.00017316, -0.00022836, 0.00037674, -0.00039981, 0.00024179, -0.00033818, -0.00004636, 0.00099560, -0.00067159, 0.00009668, 0.00001109, -0.00020328, 0.00019024, -0.00056023, -0.00008288, 0.00192926, -0.00136043, 0.00055372, -0.00054581, 0.00027890, -0.00009301, 0.00007826, -0.00009843, -0.00009728, -0.00004559, 0.00013898, -0.00007290, -0.00002571, },
		{ 0.00002800, 0.00001159, -0.00003204, -0.00012964, 0.00013348, -0.00019083, 0.00038960, -0.00053746, 0.00048100, -0.00070965, 0.00058189, 0.00039928, -0.00048161, 0.00007160, 0.00006893, -0.00033002, 0.00040330, -0.00110203, 0.00100267, 0.00110028, -0.00122391, 0.00062213, -0.00065415, 0.00045358, -0.00024597, 0.00017878, -0.00020596, -0.00002335, -0.00004926, 0.00016343, -0.00015281, 0.00006548, },
		{ -0.00002923, 0.00005199, -0.00011246, -0.00000698, 0.00000654, -0.00002252, 0.00020765, -0.00042850, 0.00044696, -0.00078163, 0.00096075, -0.00014279, -0.00021184, 0.00001575, 0.00006360, -0.00031516, 0.00037010, -0.00132053, 0.00198118, -0.00002756, -0.00064059, 0.00031603, -0.00045533, 0.00038523, -0.00022681, 0.00014572, -0.00020548, -0.00001826, 0.00000707, 0.00011170, -0.00016583, 0.00012754, },
		{ -0.00005485, 0.00006877, -0.00016561, 0.00010372, -0.00013398, 0.00019413, -0.00008968, -0.00012872, 0.00015951, -0.00051851, 0.00096248, -0.00043415, -0.00003153, 0.00002925, -0.00006900, -0.00009724, -0.00001169, -0.00097516, 0.00243678, -0.00101011, 0.00006744, -0.00016268, -0.00007756, 0.00014994, -0.00008218, 0.00001274, -0.00011546, -0.00007163, 0.00010885, -0.00000803, -0.00009824, 0.00012870, },
		{ -0.00003428, 0.00004094, -0.00015736, 0.00014430, -0.00020237, 0.00034252, -0.00035868, 0.00020827, -0.00023033, -0.00005069, 0.00064470, -0.00044079, 0.00000645, 0.00011743, -0.00028507, 0.00024160, -0.00064667, -0.00006166, 0.00214772, -0.00148196, 0.00055642, -0.00054591, 0.00027234, -0.00009896, 0.00007246, -0.00012418, -0.00001107, -0.00012601, 0.00020051, -0.00014493, 0.00002150, 0.00007006, },
		{ 0.00001809, -0.00002511, -0.00008522, 0.00009315, -0.00014970, 0.00033552, -0.00046693, 0.00041304, -0.00052302, 0.00039956, 0.00020141, -0.00027558, -0.00003119, 0.00019104, -0.00044533, 0.00049497, -0.00124746, 0.00114342, 0.00120352, -0.00132252, 0.00062972, -0.00064492, 0.00042791, -0.00022674, 0.00013108, -0.00017129, 0.00002685, -0.00011635, 0.00021748, -0.00023006, 0.00013592, -0.00001494, },
		{ 0.00006755, -0.00009750, 0.00001731, -0.00002020, 0.00000629, 0.00016471, -0.00036342, 0.00038976, -0.00057713, 0.00063671, -0.00015192, -0.00010474, -0.00003255, 0.00014025, -0.00040436, 0.00044376, -0.00145951, 0.00218343, -0.00002717, -0.00069760, 0.00031962, -0.00044077, 0.00034831, -0.00019316, 0.00006463, -0.00009866, -0.00003381, -0.00001628, 0.00012761, -0.00021411, 0.00018858, -0.00007934, },
		{ 0.00007986, -0.00013660, 0.00009767, -0.00012937, 0.00018785, -0.00009241, -0.00010478, 0.00016400, -0.00038087, 0.00058913, -0.00029131, -0.00003880, 0.00006291, -0.00007046, -0.00012147, 0.00000045, -0.00104766, 0.00262607, -0.00106547, 0.00004079, -0.00015483, -0.00007483, 0.00012444, -0.00006462, -0.00006349, 0.00003960, -0.00015305, 0.00013818, -0.00004326, -0.00009678, 0.00015449, -0.00008903, },
		{ 0.00004196, -0.00011848, 0.00011352, -0.00016740, 0.00029793, -0.00031217, 0.00017655, -0.00013657, -0.00005055, 0.00033285, -0.00023328, -0.00007445, 0.00021203, -0.00035422, 0.00028196, -0.00070144, -0.00004434, 0.00227024, -0.00154289, 0.00053895, -0.00052224, 0.00024692, -0.00009132, 0.00004093, -0.00014266, 0.00013937, -0.00024181, 0.00026392, -0.00021861, 0.00006957, 0.00005377, -0.00004174, },
		{ -0.00002961, -0.00004991, 0.00005748, -0.00010476, 0.00027288, -0.00038950, 0.00034396, -0.00035569, 0.00024013, 0.00003406, -0.00009837, -0.00011801, 0.00029477, -0.00053917, 0.00056567, -0.00133670, 0.00122783, 0.00124813, -0.00136205, 0.00061117, -0.00060656, 0.00037475, -0.00018405, 0.00003895, -0.00009178, 0.00011428, -0.00021837, 0.00027793, -0.00030952, 0.00020549, -0.00005926, 0.00003214, },
		{ -0.00009762, 0.00003405, -0.00003724, 0.00003234, 0.00011784, -0.00029230, 0.00032775, -0.00039377, 0.00035746, -0.00015210, -0.00001865, -0.00006899, 0.00020240, -0.00047453, 0.00050098, -0.00153568, 0.00228295, -0.00003262, -0.00071888, 0.00030839, -0.00040687, 0.00029008, -0.00014192, -0.00006225, 0.00008120, -0.00004493, -0.00006405, 0.00014922, -0.00026327, 0.00024587, -0.00012579, 0.00008724, },
		{ -0.00012534, 0.00008920, -0.00011435, 0.00017447, -0.00009481, -0.00007594, 0.00015695, -0.00025706, 0.00027603, -0.00016574, -0.00004983, 0.00009505, -0.00007243, -0.00014090, 0.00001802, -0.00107857, 0.00269146, -0.00107857, 0.00001802, -0.00014090, -0.00007243, 0.00009505, -0.00004983, -0.00016574, 0.00027603, -0.00025706, 0.00015695, -0.00007594, -0.00009481, 0.00017447, -0.00011435, 0.00008920, },
		{ -0.00009762, 0.00008724, -0.00012579, 0.00024587, -0.00026327, 0.00014922, -0.00006405, -0.00004493, 0.00008120, -0.00006225, -0.00014192, 0.00029008, -0.00040687, 0.00030839, -0.00071888, -0.00003262, 0.00228295, -0.00153568, 0.00050098, -0.00047453, 0.00020240, -0.00006899, -0.00001865, -0.00015210, 0.00035746, -0.00039377, 0.00032775, -0.00029230, 0.00011784, 0.00003234, -0.00003724, 0.00003405, },
		{ -0.00002961, 0.00003214, -0.00005926, 0.00020549, -0.00030952, 0.00027793, -0.00021837, 0.00011428, -0.00009178, 0.00003895, -0.00018405, 0.00037475, -0.00060656, 0.00061117, -0.00136205, 0.00124813, 0.00122783, -0.00133670, 0.00056567, -0.00053917, 0.00029477, -0.00011801, -0.00009837, 0.00003406, 0.00024013, -0.00035569, 0.00034396, -0.00038950, 0.00027288, -0.00010476, 0.00005748, -0.00004991, },
		{ 0.00004196, -0.00004174, 0.00005377, 0.00006957, -0.00021861, 0.00026392, -0.00024181, 0.00013937, -0.00014266, 0.00004093, -0.00009132, 0.00024692, -0.00052224, 0.00053895, -0.00154289, 0.00227024, -0.00004434, -0.00070144, 0.00028196, -0.00035422, 0.00021203, -0.00007445, -0.00023328, 0.00033285, -0.00005055, -0.00013657, 0.00017655, -0.00031217, 0.00029793, -0.00016740, 0.00011352, -0.00011848, },
		{ 0.00007986, -0.00008903, 0.00015449, -0.00009678, -0.00004326, 0.00013818, -0.00015305, 0.00003960, -0.00006349, -0.00006462, 0.00012444, -0.00007483, -0.00015483, 0.00004079, -0.00106547, 0.00262607, -0.00104766, 0.00000045, -0.00012147, -0.00007046, 0.00006291, -0.00003880, -0.00029131, 0.00058913, -0.00038087, 0.00016400, -0.00010478, -0.00009241, 0.00018785, -0.00012937, 0.00009767, -0.00013660, },
		{ 0.00006755, -0.00007934, 0.00018858, -0.00021411, 0.00012761, -0.00001628, -0.00003381, -0.00009866, 0.00006463, -0.00019316, 0.00034831, -0.00044077, 0.00031962, -0.00069760, -0.00002717, 0.00218343, -0.00145951, 0.00044376, -0.00040436, 0.00014025, -0.00003255, -0.00010474, -0.00015192, 0.00063671, -0.00057713, 0.00038976, -0.00036342, 0.00016471, 0.00000629, -0.00002020, 0.00001731, -0.00009750, },
		{ 0.00001809, -0.00001494, 0.00013592, -0.00023006, 0.00021748, -0.00011635, 0.00002685, -0.00017129, 0.00013108, -0.00022674, 0.00042791, -0.00064492, 0.00062972, -0.00132252, 0.00120352, 0.00114342, -0.00124746, 0.00049497, -0.00044533, 0.00019104, -0.00003119, -0.00027558, 0.00020141, 0.00039956, -0.00052302, 0.00041304, -0.00046693, 0.00033552, -0.00014970, 0.00009315, -0.00008522, -0.00002511, },
		{ -0.00003428, 0.00007006, 0.00002150, -0.00014493, 0.00020051, -0.00012601, -0.00001107, -0.00012418, 0.00007246, -0.00009896, 0.00027234, -0.00054591, 0.00055642, -0.00148196, 0.00214772, -0.00006166, -0.00064667, 0.00024160, -0.00028507, 0.00011743, 0.00000645, -0.00044079, 0.00064470, -0.00005069, -0.00023033, 0.00020827, -0.00035868, 0.00034252, -0.00020237, 0.00014430, -0.00015736, 0.00004094, },
		{ -0.00005485, 0.00012870, -0.00009824, -0.00000803, 0.00010885, -0.00007163, -0.00011546, 0.00001274, -0.00008218, 0.00014994, -0.00007756, -0.00016268, 0.00006744, -0.00101011, 0.00243678, -0.00097516, -0.00001169, -0.00009724, -0.00006900, 0.00002925, -0.00003153, -0.00043415, 0.00096248, -0.00051851, 0.00015951, -0.00012872, -0.00008968, 0.00019413, -0.00013398, 0.00010372, -0.00016561, 0.00006877, },
		{ -0.00002923, 0.00012754, -0.00016583, 0.00011170, 0.00000707, -0.00001826, -0.00020548, 0.00014572, -0.00022681, 0.00038523, -0.00045533, 0.00031603, -0.00064059, -0.00002756, 0.00198118, -0.00132053, 0.00037010, -0.00031516, 0.00006360, 0.00001575, -0.00021184, -0.00014279, 0.00096075, -0.00078163, 0.00044696, -0.00042850, 0.00020765, -0.00002252, 0.00000654, -0.00000698, -0.00011246, 0.00005199, },
		{ 0.00002800, 0.00006548, -0.00015281, 0.00016343, -0.00004926, -0.00002335, -0.00020596, 0.00017878, -0.00024597, 0.00045358, -0.00065415, 0.00062213, -0.00122391, 0.00110028, 0.00100267, -0.00110203, 0.00040330, -0.00033002, 0.00006893, 0.00007160, -0.00048161, 0.00039928, 0.00058189, -0.00070965, 0.00048100, -0.00053746, 0.00038960, -0.00019083, 0.00013348, -0.00012964, -0.00003204, 0.00001159, },
		{ 0.00008194, -0.00002571, -0.00007290, 0.00013898, -0.00004559, -0.00009728, -0.00009843, 0.00007826, -0.00009301, 0.00027890, -0.00054581, 0.00055372, -0.00136043, 0.00192926, -0.00008288, -0.00056023, 0.00019024, -0.00020328, 0.00001109, 0.00009668, -0.00067159, 0.00099560, -0.00004636, -0.00033818, 0.00024179, -0.00039981, 0.00037674, -0.00022836, 0.00017316, -0.00020359, 0.00003209, -0.00001868, },
		{ 0.00009817, -0.00009914, 0.00002831, 0.00007136, -0.00001241, -0.00019560, 0.00006352, -0.00010057, 0.00017048, -0.00008053, -0.00016406, 0.00009570, -0.00091809, 0.00214328, -0.00086736, -0.00001918, -0.00006906, -0.00006810, -0.00000474, -0.00002703, -0.00058526, 0.00136933, -0.00066119, 0.00014489, -0.00014691, -0.00008671, 0.00019316, -0.00012923, 0.00010360, -0.00019881, 0.00005287, -0.00001436, },
		{ 0.00006349, -0.00011868, 0.00010028, 0.00000988, -0.00000030, -0.00024751, 0.00018490, -0.00024292, 0.00040102, -0.00045161, 0.00029962, -0.00055481, -0.00003237, 0.00169631, -0.00113110, 0.00028417, -0.00021195, -0.00002300, 0.00007221, -0.00033108, -0.00012644, 0.00130476, -0.00099239, 0.00049581, -0.00048344, 0.00024338, -0.00005114, 0.00003718, -0.00003648, -0.00012958, 0.00003413, 0.00002396, },
		{ -0.00000569, -0.00007817, 0.00011508, -0.00001177, -0.00004279, -0.00020285, 0.00018810, -0.00024376, 0.00045333, -0.00063644, 0.00059144, -0.00107793, 0.00095077, 0.00081941, -0.00091390, 0.00029694, -0.00020025, -0.00006432, 0.00018362, -0.00070064, 0.00061166, 0.00077172, -0.00090051, 0.00054246, -0.00059582, 0.00043051, -0.00022381, 0.00016903, -0.00017307, -0.00004171, 0.00000500, 0.00007068, },
		{ -0.00007222, -0.00000335, 0.00008005, 0.00000490, -0.00013100, -0.00006810, 0.00006392, -0.00007592, 0.00026830, -0.00052390, 0.00053257, -0.00119155, 0.00163833, -0.00010547, -0.00045124, 0.00013211, -0.00011397, -0.00010094, 0.00019114, -0.00090826, 0.00135698, -0.00003946, -0.00045017, 0.00027357, -0.00043201, 0.00039730, -0.00024273, 0.00019045, -0.00024032, 0.00001745, -0.00000134, 0.00009120, },
		},
		{ { -0.00011383, 0.00021564, -0.00005861, -0.00008715, 0.00012638, -0.00012661, 0.00007626, -0.00035228, 0.00010340, -0.00003527, 0.00005181, -0.00000997, -0.00081152, 0.00190161, -0.00074920, -0.00010334, -0.00008218, -0.00010334, -0.00074920, 0.00190161, -0.00081152, -0.00000997, 0.00005181, -0.00003527, 0.00010340, -0.00035228, 0.00007626, -0.00012661, 0.00012638, -0.00008715, -0.00005861, 0.00021564, },
		{ -0.00009338, 0.00022648, -0.00012514, -0.00002184, 0.00009797, -0.00014270, 0.00009910, -0.00043716, 0.00029823, -0.00027128, 0.00035975, -0.00048453, -0.00017859, 0.00155413, -0.00088901, 0.00002517, -0.00003699, -0.00037543, -0.00032072, 0.00190921, -0.00124069, 0.00039678, -0.00024420, 0.00021266, -0.00013256, -0.00020837, 0.00000613, -0.00006966, 0.00011997, -0.00013925, 0.00002665, 0.00016559, },
		{ -0.00003789, 0.00019496, -0.00015969, 0.00004097, 0.00004313, -0.00011490, 0.00006006, -0.00042924, 0.00039225, -0.00041389, 0.00057504, -0.00089120, 0.00050384, 0.00098352, -0.00078237, 0.00001869, 0.00006897, -0.00072993, 0.00035104, 0.00154491, -0.00136880, 0.00062764, -0.00043308, 0.00038879, -0.00033821, -0.00005604, -0.00008045, 0.00001136, 0.00008009, -0.00016530, 0.00011222, 0.00008984, },
		{ 0.00003679, 0.00012914, -0.00015756, 0.00008693, -0.00002250, -0.00005472, -0.00003374, -0.00032362, 0.00035758, -0.00041140, 0.00061832, -0.00110836, 0.00106946, 0.00035262, -0.00053425, -0.00005888, 0.00015945, -0.00104783, 0.00113586, 0.00087552, -0.00118315, 0.00063405, -0.00045973, 0.00043512, -0.00045396, 0.00005469, -0.00014740, 0.00009139, 0.00001744, -0.00015873, 0.00017940, 0.00000837, },
		{ 0.00010973, 0.00004578, -0.00012338, 0.00010679, -0.00008080, 0.00001571, -0.00015536, -0.00014680, 0.00020775, -0.00025987, 0.00046262, -0.00106654, 0.00138830, -0.00017992, -0.00026990, -0.00012245, 0.00014475, -0.00118812, 0.00184992, 0.00005098, -0.00075723, 0.00043901, -0.00032573, 0.00033827, -0.00045143, 0.00009117, -0.00016601, 0.00014526, -0.00005121, -0.00012099, 0.00021426, -0.00005823, },
		{ 0.00016115, -0.00003472, -0.00006928, 0.00009847, -0.00011656, 0.00007116, -0.00026675, 0.00005126, -0.00000630, -0.00000526, 0.00014364, -0.00077055, 0.00140298, -0.00050544, -0.00008989, -0.00010599, -0.00003697, -0.00103621, 0.00230674, -0.00073679, -0.00022603, 0.00012583, -0.00008659, 0.00013311, -0.00034197, 0.00004879, -0.00012506, 0.00015597, -0.00010808, -0.00006092, 0.00021121, -0.00009463, },
		{ 0.00017827, -0.00009427, -0.00001094, 0.00006712, -0.00012185, 0.00009178, -0.00033138, 0.00021307, -0.00021274, 0.00027104, -0.00025026, -0.00029831, 0.00114200, -0.00059208, -0.00003819, 0.00000659, -0.00038582, -0.00054873, 0.00237239, -0.00130791, 0.00025653, -0.00019202, 0.00016907, -0.00010997, -0.00017177, -0.00004825, -0.00003476, 0.00012026, -0.00013964, 0.00000785, 0.00017412, -0.00009491, },
		{ 0.00015871, -0.00012206, 0.00003709, 0.00002306, -0.00009795, 0.00006967, -0.00032652, 0.00029216, -0.00034233, 0.00047858, -0.00060539, 0.00022312, 0.00070631, -0.00048527, -0.00009355, 0.00017161, -0.00082739, 0.00022135, 0.00200750, -0.00154493, 0.00056439, -0.00040966, 0.00035106, -0.00030962, -0.00000522, -0.00015601, 0.00007659, 0.00004953, -0.00014034, 0.00007070, 0.00011477, -0.00006363, },
		{ 0.00011059, -0.00011708, 0.00006529, -0.00002160, -0.00005443, 0.00001143, -0.00025114, 0.00026824, -0.00035120, 0.00054744, -0.00081851, 0.00065659, 0.00023419, -0.00028301, -0.00018602, 0.00030099, -0.00122732, 0.00112896, 0.00128228, -0.00142381, 0.00063527, -0.00046482, 0.00039895, -0.00040164, 0.00009765, -0.00022669, 0.00017226, -0.00003394, -0.00011371, 0.00011590, 0.00004900, -0.00001378, },
		{ 0.00004933, -0.00008742, 0.00007128, -0.00005644, -0.00000558, -0.00006418, -0.00012652, 0.00015381, -0.00023520, 0.00045122, -0.00082933, 0.00089647, -0.00014343, -0.00009583, -0.00023149, 0.00029742, -0.00143013, 0.00197122, 0.00035935, -0.00101646, 0.00048475, -0.00035614, 0.00030246, -0.00036083, 0.00010192, -0.00022580, 0.00021973, -0.00010447, -0.00007063, 0.00013731, -0.00000792, 0.00003783, },
		{ -0.00000765, -0.00004681, 0.00005941, -0.00007541, 0.00003434, -0.00013268, 0.00001038, -0.00000995, -0.00003122, 0.00021578, -0.00063919, 0.00089921, -0.00034269, -0.00000568, -0.00017082, 0.00009676, -0.00131036, 0.00254230, -0.00055076, -0.00046461, 0.00019480, -0.00013982, 0.00010412, -0.00020930, 0.00000974, -0.00014460, 0.00020194, -0.00014143, -0.00002534, 0.00013528, -0.00004541, 0.00007600, },
		{ -0.00004653, -0.00000966, 0.00003829, -0.00007778, 0.00005559, -0.00017214, 0.00011896, -0.00016674, 0.00019499, -0.00008923, -0.00030977, 0.00069350, -0.00034866, -0.00003834, 0.00000140, -0.00029829, -0.00081912, 0.00269230, -0.00124796, 0.00006597, -0.00011849, 0.00009346, -0.00011670, -0.00000844, -0.00014069, -0.00000371, 0.00012339, -0.00013596, 0.00000934, 0.00011554, -0.00005996, 0.00009172, },
		{ -0.00006117, 0.00001375, 0.00001727, -0.00006735, 0.00005570, -0.00017055, 0.00016815, -0.00026351, 0.00037051, -0.00037304, 0.00005751, 0.00036719, -0.00021304, -0.00015832, 0.00022957, -0.00080635, -0.00000988, 0.00237185, -0.00159699, 0.00043589, -0.00034621, 0.00025302, -0.00027266, 0.00016278, -0.00028867, 0.00015360, 0.00000892, -0.00009364, 0.00002605, 0.00008664, -0.00005504, 0.00008432, },
		{ -0.00005430, 0.00002026, 0.00000295, -0.00005019, 0.00003953, -0.00012983, 0.00014645, -0.00026778, 0.00043910, -0.00055328, 0.00035454, 0.00003571, -0.00002951, -0.00028743, 0.00041459, -0.00128355, 0.00096638, 0.00164832, -0.00156128, 0.00057113, -0.00042295, 0.00028170, -0.00030021, 0.00023566, -0.00037244, 0.00027603, -0.00010448, -0.00003235, 0.00002474, 0.00005685, -0.00003883, 0.00006078, },
		{ -0.00003563, 0.00001406, -0.00000276, -0.00003202, 0.00001686, -0.00006495, 0.00006558, -0.00017797, 0.00037935, -0.00058240, 0.00050097, -0.00019683, 0.00010382, -0.00033960, 0.00045052, -0.00156403, 0.00189695, 0.00068822, -0.00120659, 0.00047870, -0.00034585, 0.00017670, -0.00018317, 0.00017621, -0.00035111, 0.00032190, -0.00018000, 0.00002409, 0.00001193, 0.00003167, -0.00002072, 0.00003238, },
		{ -0.00001756, 0.00000428, -0.00000271, -0.00001602, -0.00000163, 0.00000123, -0.00004365, -0.00002353, 0.00021113, -0.00046227, 0.00046888, -0.00027107, 0.00012145, -0.00025887, 0.00026753, -0.00151246, 0.00256425, -0.00028900, -0.00067439, 0.00023525, -0.00017055, -0.00000960, 0.00004027, -0.00000218, -0.00021966, 0.00027381, -0.00019348, 0.00005413, -0.00000230, 0.00001266, -0.00000779, 0.00001019, },
		{ -0.00001025, 0.00000084, -0.00000256, -0.00000207, -0.00000849, 0.00004589, -0.00014169, 0.00014519, -0.00001169, -0.00024239, 0.00028950, -0.00018912, 0.00001389, -0.00004617, -0.00013446, -0.00107167, 0.00280655, -0.00107167, -0.00013446, -0.00004617, 0.00001389, -0.00018912, 0.00028950, -0.00024239, -0.00001169, 0.00014519, -0.00014169, 0.00004589, -0.00000849, -0.00000207, -0.00000256, 0.00000084, },
		{ -0.00001756, 0.00001019, -0.00000779, 0.00001266, -0.00000230, 0.00005413, -0.00019348, 0.00027381, -0.00021966, -0.00000218, 0.00004027, -0.00000960, -0.00017055, 0.00023525, -0.00067439, -0.00028900, 0.00256425, -0.00151246, 0.00026753, -0.00025887, 0.00012145, -0.00027107, 0.00046888, -0.00046227, 0.00021113, -0.00002353, -0.00004365, 0.00000123, -0.00000163, -0.00001602, -0.00000271, 0.00000428, },
		{ -0.00003563, 0.00003238, -0.00002072, 0.00003167, 0.00001193, 0.00002409, -0.00018000, 0.00032190, -0.00035111, 0.00017621, -0.00018317, 0.00017670, -0.00034585, 0.00047870, -0.00120659, 0.00068822, 0.00189695, -0.00156403, 0.00045052, -0.00033960, 0.00010382, -0.00019683, 0.00050097, -0.00058240, 0.00037935, -0.00017797, 0.00006558, -0.00006495, 0.00001686, -0.00003202, -0.00000276, 0.00001406, },
		{ -0.00005430, 0.00006078, -0.00003883, 0.00005685, 0.00002474, -0.00003235, -0.00010448, 0.00027603, -0.00037244, 0.00023566, -0.00030021, 0.00028170, -0.00042295, 0.00057113, -0.00156128, 0.00164832, 0.00096638, -0.00128355, 0.00041459, -0.00028743, -0.00002951, 0.00003571, 0.00035454, -0.00055328, 0.00043910, -0.00026778, 0.00014645, -0.00012983, 0.00003953, -0.00005019, 0.00000295, 0.00002026, },
		{ -0.00006117, 0.00008432, -0.00005504, 0.00008664, 0.00002605, -0.00009364, 0.00000892, 0.00015360, -0.00028867, 0.00016278, -0.00027266, 0.00025302, -0.00034621, 0.00043589, -0.00159699, 0.00237185, -0.00000988, -0.00080635, 0.00022957, -0.00015832, -0.00021304, 0.00036719, 0.00005751, -0.00037304, 0.00037051, -0.00026351, 0.00016815, -0.00017055, 0.00005570, -0.00006735, 0.00001727, 0.00001375, },
		{ -0.00004653, 0.00009172, -0.00005996, 0.00011554, 0.00000934, -0.00013596, 0.00012339, -0.00000371, -0.00014069, -0.00000844, -0.00011670, 0.00009346, -0.00011849, 0.00006597, -0.00124796, 0.00269230, -0.00081912, -0.00029829, 0.00000140, -0.00003834, -0.00034866, 0.00069350, -0.00030977, -0.00008923, 0.00019499, -0.00016674, 0.00011896, -0.00017214, 0.00005559, -0.00007778, 0.00003829, -0.00000966, },
		{ -0.00000765, 0.00007600, -0.00004541, 0.00013528, -0.00002534, -0.00014143, 0.00020194, -0.00014460, 0.00000974, -0.00020930, 0.00010412, -0.00013982, 0.00019480, -0.00046461, -0.00055076, 0.00254230, -0.00131036, 0.00009676, -0.00017082, -0.00000568, -0.00034269, 0.00089921, -0.00063919, 0.00021578, -0.00003122, -0.00000995, 0.00001038, -0.00013268, 0.00003434, -0.00007541, 0.00005941, -0.00004681, },
		{ 0.00004933, 0.00003783, -0.00000792, 0.00013731, -0.00007063, -0.00010447, 0.00021973, -0.00022580, 0.00010192, -0.00036083, 0.00030246, -0.00035614, 0.00048475, -0.00101646, 0.00035935, 0.00197122, -0.00143013, 0.00029742, -0.00023149, -0.00009583, -0.00014343, 0.00089647, -0.00082933, 0.00045122, -0.00023520, 0.00015381, -0.00012652, -0.00006418, -0.00000558, -0.00005644, 0.00007128, -0.00008742, },
		{ 0.00011059, -0.00001378, 0.00004900, 0.00011590, -0.00011371, -0.00003394, 0.00017226, -0.00022669, 0.00009765, -0.00040164, 0.00039895, -0.00046482, 0.00063527, -0.00142381, 0.00128228, 0.00112896, -0.00122732, 0.00030099, -0.00018602, -0.00028301, 0.00023419, 0.00065659, -0.00081851, 0.00054744, -0.00035120, 0.00026824, -0.00025114, 0.00001143, -0.00005443, -0.00002160, 0.00006529, -0.00011708, },
		{ 0.00015871, -0.00006363, 0.00011477, 0.00007070, -0.00014034, 0.00004953, 0.00007659, -0.00015601, -0.00000522, -0.00030962, 0.00035106, -0.00040966, 0.00056439, -0.00154493, 0.00200750, 0.00022135, -0.00082739, 0.00017161, -0.00009355, -0.00048527, 0.00070631, 0.00022312, -0.00060539, 0.00047858, -0.00034233, 0.00029216, -0.00032652, 0.00006967, -0.00009795, 0.00002306, 0.00003709, -0.00012206, },
		{ 0.00017827, -0.00009491, 0.00017412, 0.00000785, -0.00013964, 0.00012026, -0.00003476, -0.00004825, -0.00017177, -0.00010997, 0.00016907, -0.00019202, 0.00025653, -0.00130791, 0.00237239, -0.00054873, -0.00038582, 0.00000659, -0.00003819, -0.00059208, 0.00114200, -0.00029831, -0.00025026, 0.00027104, -0.00021274, 0.00021307, -0.00033138, 0.00009178, -0.00012185, 0.00006712, -0.00001094, -0.00009427, },
		{ 0.00016115, -0.00009463, 0.00021121, -0.00006092, -0.00010808, 0.00015597, -0.00012506, 0.00004879, -0.00034197, 0.00013311, -0.00008659, 0.00012583, -0.00022603, -0.00073679, 0.00230674, -0.00103621, -0.00003697, -0.00010599, -0.00008989, -0.00050544, 0.00140298, -0.00077055, 0.00014364, -0.00000526, -0.00000630, 0.00005126, -0.00026675, 0.00007116, -0.00011656, 0.00009847, -0.00006928, -0.00003472, },
		{ 0.00010973, -0.00005823, 0.00021426, -0.00012099, -0.00005121, 0.00014526, -0.00016601, 0.00009117, -0.00045143, 0.00033827, -0.00032573, 0.00043901, -0.00075723, 0.00005098, 0.00184992, -0.00118812, 0.00014475, -0.00012245, -0.00026990, -0.00017992, 0.00138830, -0.00106654, 0.00046262, -0.00025987, 0.00020775, -0.00014680, -0.00015536, 0.00001571, -0.00008080, 0.00010679, -0.00012338, 0.00004578, },
		{ 0.00003679, 0.00000837, 0.00017940, -0.00015873, 0.00001744, 0.00009139, -0.00014740, 0.00005469, -0.00045396, 0.00043512, -0.00045973, 0.00063405, -0.00118315, 0.00087552, 0.00113586, -0.00104783, 0.00015945, -0.00005888, -0.00053425, 0.00035262, 0.00106946, -0.00110836, 0.00061832, -0.00041140, 0.00035758, -0.00032362, -0.00003374, -0.00005472, -0.00002250, 0.00008693, -0.00015756, 0.00012914, },
		{ -0.00003789, 0.00008984, 0.00011222, -0.00016530, 0.00008009, 0.00001136, -0.00008045, -0.00005604, -0.00033821, 0.00038879, -0.00043308, 0.00062764, -0.00136880, 0.00154491, 0.00035104, -0.00072993, 0.00006897, 0.00001869, -0.00078237, 0.00098352, 0.00050384, -0.00089120, 0.00057504, -0.00041389, 0.00039225, -0.00042924, 0.00006006, -0.00011490, 0.00004313, 0.00004097, -0.00015969, 0.00019496, },
		{ -0.00009338, 0.00016559, 0.00002665, -0.00013925, 0.00011997, -0.00006966, 0.00000613, -0.00020837, -0.00013256, 0.00021266, -0.00024420, 0.00039678, -0.00124069, 0.00190921, -0.00032072, -0.00037543, -0.00003699, 0.00002517, -0.00088901, 0.00155413, -0.00017859, -0.00048453, 0.00035975, -0.00027128, 0.00029823, -0.00043716, 0.00009910, -0.00014270, 0.00009797, -0.00002184, -0.00012514, 0.00022648, },
		},
		{ { 0.00009908, -0.00006642, -0.00001408, -0.00000676, -0.00004511, 0.00003927, 0.00006611, 0.00005592, -0.00006923, 0.00014041, -0.00036709, -0.00009476, 0.00005051, -0.00061750, 0.00170692, -0.00080966, 0.00005017, -0.00080966, 0.00170692, -0.00061750, 0.00005051, -0.00009476, -0.00036709, 0.00014041, -0.00006923, 0.00005592, 0.00006611, 0.00003927, -0.00004511, -0.00000676, -0.00001408, -0.00006642, },
		{ 0.00009224, -0.00007683, -0.00000303, -0.00000910, -0.00005268, 0.00007014, 0.00003164, 0.00008706, -0.00011778, 0.00020876, -0.00052194, 0.00015216, -0.00024957, -0.00020934, 0.00141473, -0.00081722, -0.00001120, -0.00064738, 0.00184532, -0.00094561, 0.00032208, -0.00034165, -0.00019008, 0.00005213, -0.00000810, 0.00001723, 0.00010412, 0.00000243, -0.00003200, -0.00000587, -0.00002872, -0.00004564, },
		{ 0.00007271, -0.00007578, 0.00000261, -0.00001190, -0.00005382, 0.00009150, 0.00000581, 0.00010454, -0.00014510, 0.00024396, -0.00063165, 0.00036438, -0.00053382, 0.00022658, 0.00101400, -0.00069426, -0.00018322, -0.00033006, 0.00180446, -0.00115308, 0.00052553, -0.00055385, -0.00001646, -0.00004010, 0.00005500, -0.00002163, 0.00013969, -0.00003585, -0.00001508, -0.00000719, -0.00004440, -0.00001722, },
		{ 0.00004338, -0.00006396, 0.00000197, -0.00001410, -0.00004860, 0.00010126, -0.00000797, 0.00010451, -0.00014590, 0.00023763, -0.00067928, 0.00051234, -0.00075991, 0.00063390, 0.00056291, -0.00048420, -0.00043160, 0.00011774, 0.00158273, -0.00121686, 0.00063191, -0.00070188, 0.00012949, -0.00012010, 0.00010915, -0.00005320, 0.00016687, -0.00007073, 0.00000334, -0.00001115, -0.00005830, 0.00001492, },
		{ 0.00000855, -0.00004367, -0.00000472, -0.00001480, -0.00003801, 0.00009908, -0.00000846, 0.00008601, -0.00011907, 0.00018765, -0.00065652, 0.00057600, -0.00089392, 0.00096032, 0.00012373, -0.00024211, -0.00070536, 0.00064939, 0.00120289, -0.00113455, 0.00062728, -0.00076584, 0.00022838, -0.00017402, 0.00014481, -0.00007109, 0.00018061, -0.00009772, 0.00002073, -0.00001769, -0.00006777, 0.00004621, },
		{ -0.00002666, -0.00001852, -0.00001615, -0.00001339, -0.00002384, 0.00008639, 0.00000326, 0.00005115, -0.00006792, 0.00009859, -0.00056498, 0.00054782, -0.00091544, 0.00116522, -0.00024635, -0.00002545, -0.00094512, 0.00120248, 0.00070909, -0.00092456, 0.00051472, -0.00073836, 0.00026853, -0.00019253, 0.00015537, -0.00007109, 0.00017766, -0.00011337, 0.00003468, -0.00002630, -0.00007072, 0.00007210, },
		{ -0.00005715, 0.00000721, -0.00003017, -0.00000966, -0.00000837, 0.00006609, 0.00002400, 0.00000481, 0.00000033, -0.00001885, -0.00041605, 0.00043380, -0.00082074, 0.00122549, -0.00050357, 0.00011546, -0.00109277, 0.00170759, 0.00016060, -0.00062301, 0.00031380, -0.00062558, 0.00024768, -0.00017226, 0.00013814, -0.00005179, 0.00015703, -0.00011575, 0.00004327, -0.00003608, -0.00006599, 0.00008880, },
		{ -0.00007856, 0.00002918, -0.00004411, -0.00000387, 0.00000596, 0.00004209, 0.00004899, -0.00004606, 0.00007563, -0.00014940, -0.00022924, 0.00025254, -0.00062331, 0.00113875, -0.00062379, 0.00014581, -0.00110123, 0.00209832, -0.00037653, -0.00027815, 0.00005749, -0.00044619, 0.00017308, -0.00011635, 0.00009488, -0.00001489, 0.00012022, -0.00010472, 0.00004536, -0.00004581, -0.00005355, 0.00009380, },
		{ -0.00008794, 0.00004371, -0.00005521, 0.00000330, 0.00001694, 0.00001871, 0.00007265, -0.00009354, 0.00014656, -0.00027539, -0.00002932, 0.00003231, -0.00035188, 0.00092344, -0.00060558, 0.00005204, -0.00094247, 0.00232095, -0.00083891, 0.00005715, -0.00021296, -0.00022851, 0.00006035, -0.00003397, 0.00003147, 0.00003501, 0.00007107, -0.00008197, 0.00004078, -0.00005420, -0.00003453, 0.00008627, },
		{ -0.00008417, 0.00004833, -0.00006109, 0.00001084, 0.00002295, -0.00000006, 0.00008957, -0.00012991, 0.00020214, -0.00037941, 0.00015726, -0.00019323, -0.00004605, 0.00061580, -0.00046978, -0.00015567, -0.00061288, 0.00234244, -0.00117475, 0.00033341, -0.00045434, -0.00000612, -0.00006916, 0.00006110, -0.00004287, 0.00009107, 0.00001523, -0.00005073, 0.00003030, -0.00006001, -0.00001108, 0.00006719, },
		{ -0.00006816, 0.00004217, -0.00006009, 0.00001755, 0.00002313, -0.00001113, 0.00009532, -0.00014886, 0.00023346, -0.00044685, 0.00030631, -0.00039019, 0.00024962, 0.00026423, -0.00025565, -0.00044486, -0.00013485, 0.00215541, -0.00135097, 0.00051204, -0.00062821, 0.00018726, -0.00019172, 0.00015264, -0.00011701, 0.00014527, -0.00004063, -0.00001532, 0.00001554, -0.00006228, 0.00001393, 0.00003921, },
		{ -0.00004264, 0.00002610, -0.00005162, 0.00002224, 0.00001760, -0.00001282, 0.00008728, -0.00014649, 0.00023508, -0.00046810, 0.00039938, -0.00052959, 0.00049237, -0.00007811, -0.00001417, -0.00076572, 0.00044558, 0.00177941, -0.00135763, 0.00057108, -0.00070676, 0.00032286, -0.00028502, 0.00022448, -0.00017960, 0.00018952, -0.00008976, 0.00001953, -0.00000128, -0.00006046, 0.00003734, 0.00000624, },
		{ -0.00001179, 0.00000260, -0.00003620, 0.00002390, 0.00000737, -0.00000509, 0.00006502, -0.00012191, 0.00020585, -0.00043998, 0.00042640, -0.00059172, 0.00064762, -0.00036143, 0.00020042, -0.00105891, 0.00106466, 0.00125831, -0.00120873, 0.00050833, -0.00067710, 0.00038126, -0.00033169, 0.00026310, -0.00022080, 0.00021698, -0.00012630, 0.00004934, -0.00001766, -0.00005452, 0.00005610, -0.00002710, },
		{ 0.00001945, -0.00002461, -0.00001547, 0.00002189, -0.00000579, 0.00001042, 0.00003050, -0.00007749, 0.00014917, -0.00036623, 0.00038708, -0.00056910, 0.00069416, -0.00054677, 0.00033957, -0.00126502, 0.00164972, 0.00065411, -0.00093952, 0.00034143, -0.00054330, 0.00035531, -0.00032206, 0.00025975, -0.00023376, 0.00022305, -0.00014609, 0.00007054, -0.00003125, -0.00004494, 0.00006778, -0.00005621, },
		{ 0.00004613, -0.00005116, 0.00000804, 0.00001603, -0.00001964, 0.00003069, -0.00001220, -0.00001846, 0.00007246, -0.00025694, 0.00029079, -0.00046751, 0.00062730, -0.00061178, 0.00036839, -0.00133414, 0.00212942, 0.00003820, -0.00060065, 0.00010486, -0.00032572, 0.00025110, -0.00025577, 0.00021189, -0.00021572, 0.00020610, -0.00014729, 0.00008099, -0.00004014, -0.00003268, 0.00007088, -0.00007716, },
		{ 0.00006402, -0.00007269, 0.00003132, 0.00000665, -0.00003181, 0.00005181, -0.00005746, 0.00004774, -0.00001387, -0.00012709, 0.00015504, -0.00030507, 0.00045945, -0.00055389, 0.00027153, -0.00123419, 0.00244389, -0.00051875, -0.00025011, -0.00015564, -0.00005799, 0.00008697, -0.00014184, 0.00012374, -0.00016845, 0.00016773, -0.00013056, 0.00008025, -0.00004319, -0.00001902, 0.00006508, -0.00008727, },
		{ 0.00007031, -0.00008553, 0.00005124, -0.00000544, -0.00004018, 0.00006960, -0.00009895, 0.00011260, -0.00009801, 0.00000570, 0.00000262, -0.00010934, 0.00021821, -0.00039042, 0.00005588, -0.00095666, 0.00255336, -0.00095666, 0.00005588, -0.00039042, 0.00021821, -0.00010934, 0.00000262, 0.00000570, -0.00009801, 0.00011260, -0.00009895, 0.00006960, -0.00004018, -0.00000544, 0.00005124, -0.00008553, },
		{ 0.00006402, -0.00008727, 0.00006508, -0.00001902, -0.00004319, 0.00008025, -0.00013056, 0.00016773, -0.00016845, 0.00012374, -0.00014184, 0.00008697, -0.00005799, -0.00015564, -0.00025011, -0.00051875, 0.00244389, -0.00123419, 0.00027153, -0.00055389, 0.00045945, -0.00030507, 0.00015504, -0.00012709, -0.00001387, 0.00004774, -0.00005746, 0.00005181, -0.00003181, 0.00000665, 0.00003132, -0.00007269, },
		{ 0.00004613, -0.00007716, 0.00007088, -0.00003268, -0.00004014, 0.00008099, -0.00014729, 0.00020610, -0.00021572, 0.00021189, -0.00025577, 0.00025110, -0.00032572, 0.00010486, -0.00060065, 0.00003820, 0.00212942, -0.00133414, 0.00036839, -0.00061178, 0.00062730, -0.00046751, 0.00029079, -0.00025694, 0.00007246, -0.00001846, -0.00001220, 0.00003069, -0.00001964, 0.00001603, 0.00000804, -0.00005116, },
		{ 0.00001945, -0.00005621, 0.00006778, -0.00004494, -0.00003125, 0.00007054, -0.00014609, 0.00022305, -0.00023376, 0.00025975, -0.00032206, 0.00035531, -0.00054330, 0.00034143, -0.00093952, 0.00065411, 0.00164972, -0.00126502, 0.00033957, -0.00054677, 0.00069416, -0.00056910, 0.00038708, -0.00036623, 0.00014917, -0.00007749, 0.00003050, 0.00001042, -0.00000579, 0.00002189, -0.00001547, -0.00002461, },
		{ -0.00001179, -0.00002710, 0.00005610, -0.00005452, -0.00001766, 0.00004934, -0.00012630, 0.00021698, -0.00022080, 0.00026310, -0.00033169, 0.00038126, -0.00067710, 0.00050833, -0.00120873, 0.00125831, 0.00106466, -0.00105891, 0.00020042, -0.00036143, 0.00064762, -0.00059172, 0.00042640, -0.00043998, 0.00020585, -0.00012191, 0.00006502, -0.00000509, 0.00000737, 0.00002390, -0.00003620, 0.00000260, },
		{ -0.00004264, 0.00000624, 0.00003734, -0.00006046, -0.00000128, 0.00001953, -0.00008976, 0.00018952, -0.00017960, 0.00022448, -0.00028502, 0.00032286, -0.00070676, 0.00057108, -0.00135763, 0.00177941, 0.00044558, -0.00076572, -0.00001417, -0.00007811, 0.00049237, -0.00052959, 0.00039938, -0.00046810, 0.00023508, -0.00014649, 0.00008728, -0.00001282, 0.00001760, 0.00002224, -0.00005162, 0.00002610, },
		{ -0.00006816, 0.00003921, 0.00001393, -0.00006228, 0.00001554, -0.00001532, -0.00004063, 0.00014527, -0.00011701, 0.00015264, -0.00019172, 0.00018726, -0.00062821, 0.00051204, -0.00135097, 0.00215541, -0.00013485, -0.00044486, -0.00025565, 0.00026423, 0.00024962, -0.00039019, 0.00030631, -0.00044685, 0.00023346, -0.00014886, 0.00009532, -0.00001113, 0.00002313, 0.00001755, -0.00006009, 0.00004217, },
		{ -0.00008417, 0.00006719, -0.00001108, -0.00006001, 0.00003030, -0.00005073, 0.00001523, 0.00009107, -0.00004287, 0.00006110, -0.00006916, -0.00000612, -0.00045434, 0.00033341, -0.00117475, 0.00234244, -0.00061288, -0.00015567, -0.00046978, 0.00061580, -0.00004605, -0.00019323, 0.00015726, -0.00037941, 0.00020214, -0.00012991, 0.00008957, -0.00000006, 0.00002295, 0.00001084, -0.00006109, 0.00004833, },
		{ -0.00008794, 0.00008627, -0.00003453, -0.00005420, 0.00004078, -0.00008197, 0.00007107, 0.00003501, 0.00003147, -0.00003397, 0.00006035, -0.00022851, -0.00021296, 0.00005715, -0.00083891, 0.00232095, -0.00094247, 0.00005204, -0.00060558, 0.00092344, -0.00035188, 0.00003231, -0.00002932, -0.00027539, 0.00014656, -0.00009354, 0.00007265, 0.00001871, 0.00001694, 0.00000330, -0.00005521, 0.00004371, },
		{ -0.00007856, 0.00009380, -0.00005355, -0.00004581, 0.00004536, -0.00010472, 0.00012022, -0.00001489, 0.00009488, -0.00011635, 0.00017308, -0.00044619, 0.00005749, -0.00027815, -0.00037653, 0.00209832, -0.00110123, 0.00014581, -0.00062379, 0.00113875, -0.00062331, 0.00025254, -0.00022924, -0.00014940, 0.00007563, -0.00004606, 0.00004899, 0.00004209, 0.00000596, -0.00000387, -0.00004411, 0.00002918, },
		{ -0.00005715, 0.00008880, -0.00006599, -0.00003608, 0.00004327, -0.00011575, 0.00015703, -0.00005179, 0.00013814, -0.00017226, 0.00024768, -0.00062558, 0.00031380, -0.00062301, 0.00016060, 0.00170759, -0.00109277, 0.00011546, -0.00050357, 0.00122549, -0.00082074, 0.00043380, -0.00041605, -0.00001885, 0.00000033, 0.00000481, 0.00002400, 0.00006609, -0.00000837, -0.00000966, -0.00003017, 0.00000721, },
		{ -0.00002666, 0.00007210, -0.00007072, -0.00002630, 0.00003468, -0.00011337, 0.00017766, -0.00007109, 0.00015537, -0.00019253, 0.00026853, -0.00073836, 0.00051472, -0.00092456, 0.00070909, 0.00120248, -0.00094512, -0.00002545, -0.00024635, 0.00116522, -0.00091544, 0.00054782, -0.00056498, 0.00009859, -0.00006792, 0.00005115, 0.00000326, 0.00008639, -0.00002384, -0.00001339, -0.00001615, -0.00001852, },
		{ 0.00000855, 0.00004621, -0.00006777, -0.00001769, 0.00002073, -0.00009772, 0.00018061, -0.00007109, 0.00014481, -0.00017402, 0.00022838, -0.00076584, 0.00062728, -0.00113455, 0.00120289, 0.00064939, -0.00070536, -0.00024211, 0.00012373, 0.00096032, -0.00089392, 0.00057600, -0.00065652, 0.00018765, -0.00011907, 0.00008601, -0.00000846, 0.00009908, -0.00003801, -0.00001480, -0.00000472, -0.00004367, },
		{ 0.00004338, 0.00001492, -0.00005830, -0.00001115, 0.00000334, -0.00007073, 0.00016687, -0.00005320, 0.00010915, -0.00012010, 0.00012949, -0.00070188, 0.00063191, -0.00121686, 0.00158273, 0.00011774, -0.00043160, -0.00048420, 0.00056291, 0.00063390, -0.00075991, 0.00051234, -0.00067928, 0.00023763, -0.00014590, 0.00010451, -0.00000797, 0.00010126, -0.00004860, -0.00001410, 0.00000197, -0.00006396, },
		{ 0.00007271, -0.00001722, -0.00004440, -0.00000719, -0.00001508, -0.00003585, 0.00013969, -0.00002163, 0.00005500, -0.00004010, -0.00001646, -0.00055385, 0.00052553, -0.00115308, 0.00180446, -0.00033006, -0.00018322, -0.00069426, 0.00101400, 0.00022658, -0.00053382, 0.00036438, -0.00063165, 0.00024396, -0.00014510, 0.00010454, 0.00000581, 0.00009150, -0.00005382, -0.00001190, 0.00000261, -0.00007578, },
		{ 0.00009224, -0.00004564, -0.00002872, -0.00000587, -0.00003200, 0.00000243, 0.00010412, 0.00001723, -0.00000810, 0.00005213, -0.00019008, -0.00034165, 0.00032208, -0.00094561, 0.00184532, -0.00064738, -0.00001120, -0.00081722, 0.00141473, -0.00020934, -0.00024957, 0.00015216, -0.00052194, 0.00020876, -0.00011778, 0.00008706, 0.00003164, 0.00007014, -0.00005268, -0.00000910, -0.00000303, -0.00007683, },
		},
		{ { 0.00001277, -0.00003184, -0.00001330, 0.00003187, 0.00003627, -0.00004445, -0.00009718, 0.00015208, 0.00001708, -0.00025107, 0.00010317, 0.00014118, -0.00003087, -0.00040384, -0.00067243, 0.00180871, -0.00141721, 0.00180871, -0.00067243, -0.00040384, -0.00003087, 0.00014118, 0.00010317, -0.00025107, 0.00001708, 0.00015208, -0.00009718, -0.00004445, 0.00003627, 0.00003187, -0.00001330, -0.00003184, },
		{ 0.00001277, -0.00003370, -0.00000358, 0.00001789, 0.00005022, -0.00006855, -0.00006231, 0.00013189, 0.00001708, -0.00027415, 0.00017630, 0.00005499, 0.00005044, -0.00055252, -0.00041012, 0.00155891, -0.00137741, 0.00200102, -0.00091807, -0.00025196, -0.00011218, 0.00022556, 0.00002508, -0.00022109, 0.00001708, 0.00016763, -0.00012983, -0.00001982, 0.00002232, 0.00004555, -0.00002368, -0.00002941, },
		{ 0.00001277, -0.00003494, 0.00000510, 0.00000416, 0.00006363, -0.00009121, -0.00002657, 0.00010783, 0.00001708, -0.00028944, 0.00024165, -0.00002971, 0.00012862, -0.00069228, -0.00014121, 0.00126122, -0.00125952, 0.00212844, -0.00113759, -0.00010271, -0.00019036, 0.00030488, -0.00005497, -0.00018536, 0.00001708, 0.00017793, -0.00015900, 0.00000438, 0.00000890, 0.00005841, -0.00003432, -0.00002653, },
		{ 0.00001277, -0.00003550, 0.00001242, -0.00000881, 0.00007599, -0.00011156, 0.00000867, 0.00008082, 0.00001708, -0.00029636, 0.00029671, -0.00010967, 0.00020068, -0.00081775, 0.00012396, 0.00092707, -0.00106809, 0.00218608, -0.00132256, 0.00003816, -0.00026242, 0.00037608, -0.00013391, -0.00014526, 0.00001708, 0.00018259, -0.00018359, 0.00002722, -0.00000346, 0.00006995, -0.00004481, -0.00002329, },
		{ 0.00001277, -0.00003536, 0.00001809, -0.00002050, 0.00008683, -0.00012881, 0.00004206, 0.00005190, 0.00001708, -0.00029464, 0.00033937, -0.00018179, 0.00026383, -0.00092411, 0.00037520, 0.00056931, -0.00081045, 0.00217172, -0.00146587, 0.00016526, -0.00032558, 0.00043645, -0.00020870, -0.00010233, 0.00001708, 0.00018143, -0.00020263, 0.00004783, -0.00001430, 0.00007974, -0.00005475, -0.00001981, },
		{ 0.00001277, -0.00003453, 0.00002189, -0.00003048, 0.00009572, -0.00014229, 0.00007232, 0.00002219, 0.00001708, -0.00028434, 0.00036799, -0.00024333, 0.00031566, -0.00100728, 0.00060286, 0.00020168, -0.00049653, 0.00208592, -0.00156201, 0.00027368, -0.00037741, 0.00048365, -0.00027647, -0.00005821, 0.00001708, 0.00017449, -0.00021541, 0.00006541, -0.00002319, 0.00008739, -0.00006376, -0.00001625, },
		{ 0.00001277, -0.00003303, 0.00002368, -0.00003836, 0.00010233, -0.00015150, 0.00009828, -0.00000718, 0.00001708, -0.00026586, 0.00038147, -0.00029190, 0.00035418, -0.00106406, 0.00079818, -0.00016168, -0.00013838, 0.00193198, -0.00160729, 0.00035928, -0.00041592, 0.00051587, -0.00033462, -0.00001460, 0.00001708, 0.00016205, -0.00022143, 0.00007929, -0.00002980, 0.00009262, -0.00007149, -0.00001272, },
		{ 0.00001277, -0.00003094, 0.00002340, -0.00004383, 0.00010640, -0.00015607, 0.00011895, -0.00003508, 0.00001708, -0.00023992, 0.00037929, -0.00032565, 0.00037789, -0.00109226, 0.00095367, -0.00050681, 0.00025024, 0.00171580, -0.00159997, 0.00041874, -0.00043964, 0.00053187, -0.00038091, 0.00002682, 0.00001708, 0.00014457, -0.00022046, 0.00008893, -0.00003387, 0.00009521, -0.00007764, -0.00000938, },
		{ 0.00001277, -0.00002832, 0.00002104, -0.00004669, 0.00010777, -0.00015583, 0.00013352, -0.00006043, 0.00001708, -0.00020750, 0.00036154, -0.00034328, 0.00038590, -0.00109080, 0.00106334, -0.00082046, 0.00065439, 0.00144570, -0.00154032, 0.00044981, -0.00044765, 0.00053105, -0.00041356, 0.00006446, 0.00001708, 0.00012274, -0.00021253, 0.00009397, -0.00003524, 0.00009508, -0.00008198, -0.00000633, },
		{ 0.00001277, -0.00002527, 0.00001670, -0.00004682, 0.00010640, -0.00015080, 0.00014145, -0.00008226, 0.00001708, -0.00016986, 0.00032889, -0.00034411, 0.00037789, -0.00105974, 0.00112298, -0.00109056, 0.00105854, 0.00113205, -0.00143065, 0.00045126, -0.00043964, 0.00051342, -0.00043131, 0.00009688, 0.00001708, 0.00009739, -0.00019795, 0.00009420, -0.00003387, 0.00009222, -0.00008434, -0.00000371, },
		{ 0.00001277, -0.00002193, 0.00001054, -0.00004423, 0.00010233, -0.00014115, 0.00014242, -0.00009974, 0.00001708, -0.00012844, 0.00028260, -0.00032810, 0.00035418, -0.00100027, 0.00113031, -0.00130673, 0.00144716, 0.00078692, -0.00127516, 0.00042306, -0.00041592, 0.00047967, -0.00043349, 0.00012282, 0.00001708, 0.00006949, -0.00017729, 0.00008963, -0.00002980, 0.00008675, -0.00008463, -0.00000162, },
		{ 0.00001277, -0.00001840, 0.00000282, -0.00003900, 0.00009572, -0.00012728, 0.00013640, -0.00011218, 0.00001708, -0.00008484, 0.00022445, -0.00029588, 0.00031566, -0.00091468, 0.00108503, -0.00146068, 0.00180531, 0.00042356, -0.00107984, 0.00036629, -0.00037741, 0.00043109, -0.00042001, 0.00014130, 0.00001708, 0.00004012, -0.00015133, 0.00008042, -0.00002319, 0.00007887, -0.00008284, -0.00000012, },
		{ 0.00001277, -0.00001483, -0.00000619, -0.00003135, 0.00008683, -0.00010969, 0.00012363, -0.00011912, 0.00001708, -0.00004072, 0.00015668, -0.00024868, 0.00026383, -0.00080625, 0.00098889, -0.00154648, 0.00211923, 0.00005593, -0.00085218, 0.00028312, -0.00032558, 0.00036956, -0.00039139, 0.00015159, 0.00001708, 0.00001041, -0.00012107, 0.00006694, -0.00001430, 0.00006890, -0.00007904, 0.00000071, },
		{ 0.00001277, -0.00001136, -0.00001613, -0.00002156, 0.00007599, -0.00008909, 0.00010458, -0.00012028, 0.00001708, 0.00000222, 0.00008189, -0.00018832, 0.00020068, -0.00067916, 0.00084558, -0.00156083, 0.00237686, -0.00030183, -0.00060094, 0.00017676, -0.00026242, 0.00029743, -0.00034873, 0.00015332, 0.00001708, -0.00001851, -0.00008768, 0.00004969, -0.00000346, 0.00005720, -0.00007337, 0.00000085, },
		{ 0.00001277, -0.00000812, -0.00002662, -0.00001001, 0.00006363, -0.00006625, 0.00008000, -0.00011562, 0.00001708, 0.00004232, 0.00000295, -0.00011711, 0.00012862, -0.00053828, 0.00066061, -0.00150319, 0.00256830, -0.00063597, -0.00033577, 0.00005128, -0.00019036, 0.00021748, -0.00029367, 0.00014640, 0.00001708, -0.00004552, -0.00005243, 0.00002935, 0.00000890, 0.00004424, -0.00006605, 0.00000029, },
		{ 0.00001277, -0.00000524, -0.00003726, 0.00000285, 0.00005022, -0.00004205, 0.00005082, -0.00010532, 0.00001708, 0.00007805, -0.00007710, -0.00003779, 0.00005044, -0.00038903, 0.00044109, -0.00137577, 0.00268619, -0.00093367, -0.00006686, -0.00008848, -0.00011218, 0.00013278, -0.00022832, 0.00013111, 0.00001708, -0.00006958, -0.00001669, 0.00000669, 0.00002232, 0.00003050, -0.00005736, -0.00000095, },
		{ 0.00001277, -0.00000281, -0.00004764, 0.00001653, 0.00003627, -0.00001742, 0.00001817, -0.00008977, 0.00001708, 0.00010803, -0.00015519, 0.00004658, -0.00003087, -0.00023715, 0.00019545, -0.00118347, 0.00272599, -0.00118347, 0.00019545, -0.00023715, -0.00003087, 0.00004658, -0.00015519, 0.00010803, 0.00001708, -0.00008977, 0.00001817, -0.00001742, 0.00003627, 0.00001653, -0.00004764, -0.00000281, },
		{ 0.00001277, -0.00000095, -0.00005736, 0.00003050, 0.00002232, 0.00000669, -0.00001669, -0.00006958, 0.00001708, 0.00013111, -0.00022832, 0.00013278, -0.00011218, -0.00008848, -0.00006686, -0.00093367, 0.00268619, -0.00137577, 0.00044109, -0.00038903, 0.00005044, -0.00003779, -0.00007710, 0.00007805, 0.00001708, -0.00010532, 0.00005082, -0.00004205, 0.00005022, 0.00000285, -0.00003726, -0.00000524, },
		{ 0.00001277, 0.00000029, -0.00006605, 0.00004424, 0.00000890, 0.00002935, -0.00005243, -0.00004552, 0.00001708, 0.00014640, -0.00029367, 0.00021748, -0.00019036, 0.00005128, -0.00033577, -0.00063597, 0.00256830, -0.00150319, 0.00066061, -0.00053828, 0.00012862, -0.00011711, 0.00000295, 0.00004232, 0.00001708, -0.00011562, 0.00008000, -0.00006625, 0.00006363, -0.00001001, -0.00002662, -0.00000812, },
		{ 0.00001277, 0.00000085, -0.00007337, 0.00005720, -0.00000346, 0.00004969, -0.00008768, -0.00001851, 0.00001708, 0.00015332, -0.00034873, 0.00029743, -0.00026242, 0.00017676, -0.00060094, -0.00030183, 0.00237686, -0.00156083, 0.00084558, -0.00067916, 0.00020068, -0.00018832, 0.00008189, 0.00000222, 0.00001708, -0.00012028, 0.00010458, -0.00008909, 0.00007599, -0.00002156, -0.00001613, -0.00001136, },
		{ 0.00001277, 0.00000071, -0.00007904, 0.00006890, -0.00001430, 0.00006694, -0.00012107, 0.00001041, 0.00001708, 0.00015159, -0.00039139, 0.00036956, -0.00032558, 0.00028312, -0.00085218, 0.00005593, 0.00211923, -0.00154648, 0.00098889, -0.00080625, 0.00026383, -0.00024868, 0.00015668, -0.00004072, 0.00001708, -0.00011912, 0.00012363, -0.00010969, 0.00008683, -0.00003135, -0.00000619, -0.00001483, },
		{ 0.00001277, -0.00000012, -0.00008284, 0.00007887, -0.00002319, 0.00008042, -0.00015133, 0.00004012, 0.00001708, 0.00014130, -0.00042001, 0.00043109, -0.00037741, 0.00036629, -0.00107984, 0.00042356, 0.00180531, -0.00146068, 0.00108503, -0.00091468, 0.00031566, -0.00029588, 0.00022445, -0.00008484, 0.00001708, -0.00011218, 0.00013640, -0.00012728, 0.00009572, -0.00003900, 0.00000282, -0.00001840, },
		{ 0.00001277, -0.00000162, -0.00008463, 0.00008675, -0.00002980, 0.00008963, -0.00017729, 0.00006949, 0.00001708, 0.00012282, -0.00043349, 0.00047967, -0.00041592, 0.00042306, -0.00127516, 0.00078692, 0.00144716, -0.00130673, 0.00113031, -0.00100027, 0.00035418, -0.00032810, 0.00028260, -0.00012844, 0.00001708, -0.00009974, 0.00014242, -0.00014115, 0.00010233, -0.00004423, 0.00001054, -0.00002193, },
		{ 0.00001277, -0.00000371, -0.00008434, 0.00009222, -0.00003387, 0.00009420, -0.00019795, 0.00009739, 0.00001708, 0.00009688, -0.00043131, 0.00051342, -0.00043964, 0.00045126, -0.00143065, 0.00113205, 0.00105854, -0.00109056, 0.00112298, -0.00105974, 0.00037789, -0.00034411, 0.00032889, -0.00016986, 0.00001708, -0.00008226, 0.00014145, -0.00015080, 0.00010640, -0.00004682, 0.00001670, -0.00002527, },
		{ 0.00001277, -0.00000633, -0.00008198, 0.00009508, -0.00003524, 0.00009397, -0.00021253, 0.00012274, 0.00001708, 0.00006446, -0.00041356, 0.00053105, -0.00044765, 0.00044981, -0.00154032, 0.00144570, 0.00065439, -0.00082046, 0.00106334, -0.00109080, 0.00038590, -0.00034328, 0.00036154, -0.00020750, 0.00001708, -0.00006043, 0.00013352, -0.00015583, 0.00010777, -0.00004669, 0.00002104, -0.00002832, },
		{ 0.00001277, -0.00000938, -0.00007764, 0.00009521, -0.00003387, 0.00008893, -0.00022046, 0.00014457, 0.00001708, 0.00002682, -0.00038091, 0.00053187, -0.00043964, 0.00041874, -0.00159997, 0.00171580, 0.00025024, -0.00050681, 0.00095367, -0.00109226, 0.00037789, -0.00032565, 0.00037929, -0.00023992, 0.00001708, -0.00003508, 0.00011895, -0.00015607, 0.00010640, -0.00004383, 0.00002340, -0.00003094, },
		{ 0.00001277, -0.00001272, -0.00007149, 0.00009262, -0.00002980, 0.00007929, -0.00022143, 0.00016205, 0.00001708, -0.00001460, -0.00033462, 0.00051587, -0.00041592, 0.00035928, -0.00160729, 0.00193198, -0.00013838, -0.00016168, 0.00079818, -0.00106406, 0.00035418, -0.00029190, 0.00038147, -0.00026586, 0.00001708, -0.00000718, 0.00009828, -0.00015150, 0.00010233, -0.00003836, 0.00002368, -0.00003303, },
		{ 0.00001277, -0.00001625, -0.00006376, 0.00008739, -0.00002319, 0.00006541, -0.00021541, 0.00017449, 0.00001708, -0.00005821, -0.00027647, 0.00048365, -0.00037741, 0.00027368, -0.00156201, 0.00208592, -0.00049653, 0.00020168, 0.00060286, -0.00100728, 0.00031566, -0.00024333, 0.00036799, -0.00028434, 0.00001708, 0.00002219, 0.00007232, -0.00014229, 0.00009572, -0.00003048, 0.00002189, -0.00003453, },
		{ 0.00001277, -0.00001981, -0.00005475, 0.00007974, -0.00001430, 0.00004783, -0.00020263, 0.00018143, 0.00001708, -0.00010233, -0.00020870, 0.00043645, -0.00032558, 0.00016526, -0.00146587, 0.00217172, -0.00081045, 0.00056931, 0.00037520, -0.00092411, 0.00026383, -0.00018179, 0.00033937, -0.00029464, 0.00001708, 0.00005190, 0.00004206, -0.00012881, 0.00008683, -0.00002050, 0.00001809, -0.00003536, },
		{ 0.00001277, -0.00002329, -0.00004481, 0.00006995, -0.00000346, 0.00002722, -0.00018359, 0.00018259, 0.00001708, -0.00014526, -0.00013391, 0.00037608, -0.00026242, 0.00003816, -0.00132256, 0.00218608, -0.00106809, 0.00092707, 0.00012396, -0.00081775, 0.00020068, -0.00010967, 0.00029671, -0.00029636, 0.00001708, 0.00008082, 0.00000867, -0.00011156, 0.00007599, -0.00000881, 0.00001242, -0.00003550, },
		{ 0.00001277, -0.00002653, -0.00003432, 0.00005841, 0.00000890, 0.00000438, -0.00015900, 0.00017793, 0.00001708, -0.00018536, -0.00005497, 0.00030488, -0.00019036, -0.00010271, -0.00113759, 0.00212844, -0.00125952, 0.00126122, -0.00014121, -0.00069228, 0.00012862, -0.00002971, 0.00024165, -0.00028944, 0.00001708, 0.00010783, -0.00002657, -0.00009121, 0.00006363, 0.00000416, 0.00000510, -0.00003494, },
		{ 0.00001277, -0.00002941, -0.00002368, 0.00004555, 0.00002232, -0.00001982, -0.00012983, 0.00016763, 0.00001708, -0.00022109, 0.00002508, 0.00022556, -0.00011218, -0.00025196, -0.00091807, 0.00200102, -0.00137741, 0.00155891, -0.00041012, -0.00055252, 0.00005044, 0.00005499, 0.00017630, -0.00027415, 0.00001708, 0.00013189, -0.00006231, -0.00006855, 0.00005022, 0.00001789, -0.00000358, -0.00003370, },
		},
		{ { -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		{ -0.00007355, -0.00000032, 0.00005584, -0.00002817, -0.00005005, 0.00006134, 0.00004681, -0.00011052, -0.00006923, 0.00013883, 0.00006031, -0.00022909, -0.00011719, 0.00024860, -0.00015217, -0.00144016, 0.00333021, -0.00144016, -0.00015217, 0.00024860, -0.00011719, -0.00022909, 0.00006031, 0.00013883, -0.00006923, -0.00011052, 0.00004681, 0.00006134, -0.00005005, -0.00002817, 0.00005584, -0.00000032, },
		},
		{ { 0.00001277, -0.00003184, -0.00001330, 0.00003187, 0.00003627, -0.00004445, -0.00009718, 0.00015208, 0.00001708, -0.00025107, 0.00010317, 0.00014118, -0.00003087, -0.00040384, -0.00067243, 0.00180871, -0.00141721, 0.00180871, -0.00067243, -0.00040384, -0.00003087, 0.00014118, 0.00010317, -0.00025107, 0.00001708, 0.00015208, -0.00009718, -0.00004445, 0.00003627, 0.00003187, -0.00001330, -0.00003184, },
		{ 0.00001277, -0.00002941, -0.00002368, 0.00004555, 0.00002232, -0.00001982, -0.00012983, 0.00016763, 0.00001708, -0.00022109, 0.00002508, 0.00022556, -0.00011218, -0.00025196, -0.00091807, 0.00200102, -0.00137741, 0.00155891, -0.00041012, -0.00055252, 0.00005044, 0.00005499, 0.00017630, -0.00027415, 0.00001708, 0.00013189, -0.00006231, -0.00006855, 0.00005022, 0.00001789, -0.00000358, -0.00003370, },
		{ 0.00001277, -0.00002653, -0.00003432, 0.00005841, 0.00000890, 0.00000438, -0.00015900, 0.00017793, 0.00001708, -0.00018536, -0.00005497, 0.00030488, -0.00019036, -0.00010271, -0.00113759, 0.00212844, -0.00125952, 0.00126122, -0.00014121, -0.00069228, 0.00012862, -0.00002971, 0.00024165, -0.00028944, 0.00001708, 0.00010783, -0.00002657, -0.00009121, 0.00006363, 0.00000416, 0.00000510, -0.00003494, },
		{ 0.00001277, -0.00002329, -0.00004481, 0.00006995, -0.00000346, 0.00002722, -0.00018359, 0.00018259, 0.00001708, -0.00014526, -0.00013391, 0.00037608, -0.00026242, 0.00003816, -0.00132256, 0.00218608, -0.00106809, 0.00092707, 0.00012396, -0.00081775, 0.00020068, -0.00010967, 0.00029671, -0.00029636, 0.00001708, 0.00008082, 0.00000867, -0.00011156, 0.00007599, -0.00000881, 0.00001242, -0.00003550, },
		{ 0.00001277, -0.00001981, -0.00005475, 0.00007974, -0.00001430, 0.00004783, -0.00020263, 0.00018143, 0.00001708, -0.00010233, -0.00020870, 0.00043645, -0.00032558, 0.00016526, -0.00146587, 0.00217172, -0.00081045, 0.00056931, 0.00037520, -0.00092411, 0.00026383, -0.00018179, 0.00033937, -0.00029464, 0.00001708, 0.00005190, 0.00004206, -0.00012881, 0.00008683, -0.00002050, 0.00001809, -0.00003536, },
		{ 0.00001277, -0.00001625, -0.00006376, 0.00008739, -0.00002319, 0.00006541, -0.00021541, 0.00017449, 0.00001708, -0.00005821, -0.00027647, 0.00048365, -0.00037741, 0.00027368, -0.00156201, 0.00208592, -0.00049653, 0.00020168, 0.00060286, -0.00100728, 0.00031566, -0.00024333, 0.00036799, -0.00028434, 0.00001708, 0.00002219, 0.00007232, -0.00014229, 0.00009572, -0.00003048, 0.00002189, -0.00003453, },
		{ 0.00001277, -0.00001272, -0.00007149, 0.00009262, -0.00002980, 0.00007929, -0.00022143, 0.00016205, 0.00001708, -0.00001460, -0.00033462, 0.00051587, -0.00041592, 0.00035928, -0.00160729, 0.00193198, -0.00013838, -0.00016168, 0.00079818, -0.00106406, 0.00035418, -0.00029190, 0.00038147, -0.00026586, 0.00001708, -0.00000718, 0.00009828, -0.00015150, 0.00010233, -0.00003836, 0.00002368, -0.00003303, },
		{ 0.00001277, -0.00000938, -0.00007764, 0.00009521, -0.00003387, 0.00008893, -0.00022046, 0.00014457, 0.00001708, 0.00002682, -0.00038091, 0.00053187, -0.00043964, 0.00041874, -0.00159997, 0.00171580, 0.00025024, -0.00050681, 0.00095367, -0.00109226, 0.00037789, -0.00032565, 0.00037929, -0.00023992, 0.00001708, -0.00003508, 0.00011895, -0.00015607, 0.00010640, -0.00004383, 0.00002340, -0.00003094, },
		{ 0.00001277, -0.00000633, -0.00008198, 0.00009508, -0.00003524, 0.00009397, -0.00021253, 0.00012274, 0.00001708, 0.00006446, -0.00041356, 0.00053105, -0.00044765, 0.00044981, -0.00154032, 0.00144570, 0.00065439, -0.00082046, 0.00106334, -0.00109080, 0.00038590, -0.00034328, 0.00036154, -0.00020750, 0.00001708, -0.00006043, 0.00013352, -0.00015583, 0.00010777, -0.00004669, 0.00002104, -0.00002832, },
		{ 0.00001277, -0.00000371, -0.00008434, 0.00009222, -0.00003387, 0.00009420, -0.00019795, 0.00009739, 0.00001708, 0.00009688, -0.00043131, 0.00051342, -0.00043964, 0.00045126, -0.00143065, 0.00113205, 0.00105854, -0.00109056, 0.00112298, -0.00105974, 0.00037789, -0.00034411, 0.00032889, -0.00016986, 0.00001708, -0.00008226, 0.00014145, -0.00015080, 0.00010640, -0.00004682, 0.00001670, -0.00002527, },
		{ 0.00001277, -0.00000162, -0.00008463, 0.00008675, -0.00002980, 0.00008963, -0.00017729, 0.00006949, 0.00001708, 0.00012282, -0.00043349, 0.00047967, -0.00041592, 0.00042306, -0.00127516, 0.00078692, 0.00144716, -0.00130673, 0.00113031, -0.00100027, 0.00035418, -0.00032810, 0.00028260, -0.00012844, 0.00001708, -0.00009974, 0.00014242, -0.00014115, 0.00010233, -0.00004423, 0.00001054, -0.00002193, },
		{ 0.00001277, -0.00000012, -0.00008284, 0.00007887, -0.00002319, 0.00008042, -0.00015133, 0.00004012, 0.00001708, 0.00014130, -0.00042001, 0.00043109, -0.00037741, 0.00036629, -0.00107984, 0.00042356, 0.00180531, -0.00146068, 0.00108503, -0.00091468, 0.00031566, -0.00029588, 0.00022445, -0.00008484, 0.00001708, -0.00011218, 0.00013640, -0.00012728, 0.00009572, -0.00003900, 0.00000282, -0.00001840, },
		{ 0.00001277, 0.00000071, -0.00007904, 0.00006890, -0.00001430, 0.00006694, -0.00012107, 0.00001041, 0.00001708, 0.00015159, -0.00039139, 0.00036956, -0.00032558, 0.00028312, -0.00085218, 0.00005593, 0.00211923, -0.00154648, 0.00098889, -0.00080625, 0.00026383, -0.00024868, 0.00015668, -0.00004072, 0.00001708, -0.00011912, 0.00012363, -0.00010969, 0.00008683, -0.00003135, -0.00000619, -0.00001483, },
		{ 0.00001277, 0.00000085, -0.00007337, 0.00005720, -0.00000346, 0.00004969, -0.00008768, -0.00001851, 0.00001708, 0.00015332, -0.00034873, 0.00029743, -0.00026242, 0.00017676, -0.00060094, -0.00030183, 0.00237686, -0.00156083, 0.00084558, -0.00067916, 0.00020068, -0.00018832, 0.00008189, 0.00000222, 0.00001708, -0.00012028, 0.00010458, -0.00008909, 0.00007599, -0.00002156, -0.00001613, -0.00001136, },
		{ 0.00001277, 0.00000029, -0.00006605, 0.00004424, 0.00000890, 0.00002935, -0.00005243, -0.00004552, 0.00001708, 0.00014640, -0.00029367, 0.00021748, -0.00019036, 0.00005128, -0.00033577, -0.00063597, 0.00256830, -0.00150319, 0.00066061, -0.00053828, 0.00012862, -0.00011711, 0.00000295, 0.00004232, 0.00001708, -0.00011562, 0.00008000, -0.00006625, 0.00006363, -0.00001001, -0.00002662, -0.00000812, },
		{ 0.00001277, -0.00000095, -0.00005736, 0.00003050, 0.00002232, 0.00000669, -0.00001669, -0.00006958, 0.00001708, 0.00013111, -0.00022832, 0.00013278, -0.00011218, -0.00008848, -0.00006686, -0.00093367, 0.00268619, -0.00137577, 0.00044109, -0.00038903, 0.00005044, -0.00003779, -0.00007710, 0.00007805, 0.00001708, -0.00010532, 0.00005082, -0.00004205, 0.00005022, 0.00000285, -0.00003726, -0.00000524, },
		{ 0.00001277, -0.00000281, -0.00004764, 0.00001653, 0.00003627, -0.00001742, 0.00001817, -0.00008977, 0.00001708, 0.00010803, -0.00015519, 0.00004658, -0.00003087, -0.00023715, 0.00019545, -0.00118347, 0.00272599, -0.00118347, 0.00019545, -0.00023715, -0.00003087, 0.00004658, -0.00015519, 0.00010803, 0.00001708, -0.00008977, 0.00001817, -0.00001742, 0.00003627, 0.00001653, -0.00004764, -0.00000281, },
		{ 0.00001277, -0.00000524, -0.00003726, 0.00000285, 0.00005022, -0.00004205, 0.00005082, -0.00010532, 0.00001708, 0.00007805, -0.00007710, -0.00003779, 0.00005044, -0.00038903, 0.00044109, -0.00137577, 0.00268619, -0.00093367, -0.00006686, -0.00008848, -0.00011218, 0.00013278, -0.00022832, 0.00013111, 0.00001708, -0.00006958, -0.00001669, 0.00000669, 0.00002232, 0.00003050, -0.00005736, -0.00000095, },
		{ 0.00001277, -0.00000812, -0.00002662, -0.00001001, 0.00006363, -0.00006625, 0.00008000, -0.00011562, 0.00001708, 0.00004232, 0.00000295, -0.00011711, 0.00012862, -0.00053828, 0.00066061, -0.00150319, 0.00256830, -0.00063597, -0.00033577, 0.00005128, -0.00019036, 0.00021748, -0.00029367, 0.00014640, 0.00001708, -0.00004552, -0.00005243, 0.00002935, 0.00000890, 0.00004424, -0.00006605, 0.00000029, },
		{ 0.00001277, -0.00001136, -0.00001613, -0.00002156, 0.00007599, -0.00008909, 0.00010458, -0.00012028, 0.00001708, 0.00000222, 0.00008189, -0.00018832, 0.00020068, -0.00067916, 0.00084558, -0.00156083, 0.00237686, -0.00030183, -0.00060094, 0.00017676, -0.00026242, 0.00029743, -0.00034873, 0.00015332, 0.00001708, -0.00001851, -0.00008768, 0.00004969, -0.00000346, 0.00005720, -0.00007337, 0.00000085, },
		{ 0.00001277, -0.00001483, -0.00000619, -0.00003135, 0.00008683, -0.00010969, 0.00012363, -0.00011912, 0.00001708, -0.00004072, 0.00015668, -0.00024868, 0.00026383, -0.00080625, 0.00098889, -0.00154648, 0.00211923, 0.00005593, -0.00085218, 0.00028312, -0.00032558, 0.00036956, -0.00039139, 0.00015159, 0.00001708, 0.00001041, -0.00012107, 0.00006694, -0.00001430, 0.00006890, -0.00007904, 0.00000071, },
		{ 0.00001277, -0.00001840, 0.00000282, -0.00003900, 0.00009572, -0.00012728, 0.00013640, -0.00011218, 0.00001708, -0.00008484, 0.00022445, -0.00029588, 0.00031566, -0.00091468, 0.00108503, -0.00146068, 0.00180531, 0.00042356, -0.00107984, 0.00036629, -0.00037741, 0.00043109, -0.00042001, 0.00014130, 0.00001708, 0.00004012, -0.00015133, 0.00008042, -0.00002319, 0.00007887, -0.00008284, -0.00000012, },
		{ 0.00001277, -0.00002193, 0.00001054, -0.00004423, 0.00010233, -0.00014115, 0.00014242, -0.00009974, 0.00001708, -0.00012844, 0.00028260, -0.00032810, 0.00035418, -0.00100027, 0.00113031, -0.00130673, 0.00144716, 0.00078692, -0.00127516, 0.00042306, -0.00041592, 0.00047967, -0.00043349, 0.00012282, 0.00001708, 0.00006949, -0.00017729, 0.00008963, -0.00002980, 0.00008675, -0.00008463, -0.00000162, },
		{ 0.00001277, -0.00002527, 0.00001670, -0.00004682, 0.00010640, -0.00015080, 0.00014145, -0.00008226, 0.00001708, -0.00016986, 0.00032889, -0.00034411, 0.00037789, -0.00105974, 0.00112298, -0.00109056, 0.00105854, 0.00113205, -0.00143065, 0.00045126, -0.00043964, 0.00051342, -0.00043131, 0.00009688, 0.00001708, 0.00009739, -0.00019795, 0.00009420, -0.00003387, 0.00009222, -0.00008434, -0.00000371, },
		{ 0.00001277, -0.00002832, 0.00002104, -0.00004669, 0.00010777, -0.00015583, 0.00013352, -0.00006043, 0.00001708, -0.00020750, 0.00036154, -0.00034328, 0.00038590, -0.00109080, 0.00106334, -0.00082046, 0.00065439, 0.00144570, -0.00154032, 0.00044981, -0.00044765, 0.00053105, -0.00041356, 0.00006446, 0.00001708, 0.00012274, -0.00021253, 0.00009397, -0.00003524, 0.00009508, -0.00008198, -0.00000633, },
		{ 0.00001277, -0.00003094, 0.00002340, -0.00004383, 0.00010640, -0.00015607, 0.00011895, -0.00003508, 0.00001708, -0.00023992, 0.00037929, -0.00032565, 0.00037789, -0.00109226, 0.00095367, -0.00050681, 0.00025024, 0.00171580, -0.00159997, 0.00041874, -0.00043964, 0.00053187, -0.00038091, 0.00002682, 0.00001708, 0.00014457, -0.00022046, 0.00008893, -0.00003387, 0.00009521, -0.00007764, -0.00000938, },
		{ 0.00001277, -0.00003303, 0.00002368, -0.00003836, 0.00010233, -0.00015150, 0.00009828, -0.00000718, 0.00001708, -0.00026586, 0.00038147, -0.00029190, 0.00035418, -0.00106406, 0.00079818, -0.00016168, -0.00013838, 0.00193198, -0.00160729, 0.00035928, -0.00041592, 0.00051587, -0.00033462, -0.00001460, 0.00001708, 0.00016205, -0.00022143, 0.00007929, -0.00002980, 0.00009262, -0.00007149, -0.00001272, },
		{ 0.00001277, -0.00003453, 0.00002189, -0.00003048, 0.00009572, -0.00014229, 0.00007232, 0.00002219, 0.00001708, -0.00028434, 0.00036799, -0.00024333, 0.00031566, -0.00100728, 0.00060286, 0.00020168, -0.00049653, 0.00208592, -0.00156201, 0.00027368, -0.00037741, 0.00048365, -0.00027647, -0.00005821, 0.00001708, 0.00017449, -0.00021541, 0.00006541, -0.00002319, 0.00008739, -0.00006376, -0.00001625, },
		{ 0.00001277, -0.00003536, 0.00001809, -0.00002050, 0.00008683, -0.00012881, 0.00004206, 0.00005190, 0.00001708, -0.00029464, 0.00033937, -0.00018179, 0.00026383, -0.00092411, 0.00037520, 0.00056931, -0.00081045, 0.00217172, -0.00146587, 0.00016526, -0.00032558, 0.00043645, -0.00020870, -0.00010233, 0.00001708, 0.00018143, -0.00020263, 0.00004783, -0.00001430, 0.00007974, -0.00005475, -0.00001981, },
		{ 0.00001277, -0.00003550, 0.00001242, -0.00000881, 0.00007599, -0.00011156, 0.00000867, 0.00008082, 0.00001708, -0.00029636, 0.00029671, -0.00010967, 0.00020068, -0.00081775, 0.00012396, 0.00092707, -0.00106809, 0.00218608, -0.00132256, 0.00003816, -0.00026242, 0.00037608, -0.00013391, -0.00014526, 0.00001708, 0.00018259, -0.00018359, 0.00002722, -0.00000346, 0.00006995, -0.00004481, -0.00002329, },
		{ 0.00001277, -0.00003494, 0.00000510, 0.00000416, 0.00006363, -0.00009121, -0.00002657, 0.00010783, 0.00001708, -0.00028944, 0.00024165, -0.00002971, 0.00012862, -0.00069228, -0.00014121, 0.00126122, -0.00125952, 0.00212844, -0.00113759, -0.00010271, -0.00019036, 0.00030488, -0.00005497, -0.00018536, 0.00001708, 0.00017793, -0.00015900, 0.00000438, 0.00000890, 0.00005841, -0.00003432, -0.00002653, },
		{ 0.00001277, -0.00003370, -0.00000358, 0.00001789, 0.00005022, -0.00006855, -0.00006231, 0.00013189, 0.00001708, -0.00027415, 0.00017630, 0.00005499, 0.00005044, -0.00055252, -0.00041012, 0.00155891, -0.00137741, 0.00200102, -0.00091807, -0.00025196, -0.00011218, 0.00022556, 0.00002508, -0.00022109, 0.00001708, 0.00016763, -0.00012983, -0.00001982, 0.00002232, 0.00004555, -0.00002368, -0.00002941, },
		},
		{ { 0.00009908, -0.00006642, -0.00001408, -0.00000676, -0.00004511, 0.00003927, 0.00006611, 0.00005592, -0.00006923, 0.00014041, -0.00036709, -0.00009476, 0.00005051, -0.00061750, 0.00170692, -0.00080966, 0.00005017, -0.00080966, 0.00170692, -0.00061750, 0.00005051, -0.00009476, -0.00036709, 0.00014041, -0.00006923, 0.00005592, 0.00006611, 0.00003927, -0.00004511, -0.00000676, -0.00001408, -0.00006642, },
		{ 0.00009224, -0.00004564, -0.00002872, -0.00000587, -0.00003200, 0.00000243, 0.00010412, 0.00001723, -0.00000810, 0.00005213, -0.00019008, -0.00034165, 0.00032208, -0.00094561, 0.00184532, -0.00064738, -0.00001120, -0.00081722, 0.00141473, -0.00020934, -0.00024957, 0.00015216, -0.00052194, 0.00020876, -0.00011778, 0.00008706, 0.00003164, 0.00007014, -0.00005268, -0.00000910, -0.00000303, -0.00007683, },
		{ 0.00007271, -0.00001722, -0.00004440, -0.00000719, -0.00001508, -0.00003585, 0.00013969, -0.00002163, 0.00005500, -0.00004010, -0.00001646, -0.00055385, 0.00052553, -0.00115308, 0.00180446, -0.00033006, -0.00018322, -0.00069426, 0.00101400, 0.00022658, -0.00053382, 0.00036438, -0.00063165, 0.00024396, -0.00014510, 0.00010454, 0.00000581, 0.00009150, -0.00005382, -0.00001190, 0.00000261, -0.00007578, },
		{ 0.00004338, 0.00001492, -0.00005830, -0.00001115, 0.00000334, -0.00007073, 0.00016687, -0.00005320, 0.00010915, -0.00012010, 0.00012949, -0.00070188, 0.00063191, -0.00121686, 0.00158273, 0.00011774, -0.00043160, -0.00048420, 0.00056291, 0.00063390, -0.00075991, 0.00051234, -0.00067928, 0.00023763, -0.00014590, 0.00010451, -0.00000797, 0.00010126, -0.00004860, -0.00001410, 0.00000197, -0.00006396, },
		{ 0.00000855, 0.00004621, -0.00006777, -0.00001769, 0.00002073, -0.00009772, 0.00018061, -0.00007109, 0.00014481, -0.00017402, 0.00022838, -0.00076584, 0.00062728, -0.00113455, 0.00120289, 0.00064939, -0.00070536, -0.00024211, 0.00012373, 0.00096032, -0.00089392, 0.00057600, -0.00065652, 0.00018765, -0.00011907, 0.00008601, -0.00000846, 0.00009908, -0.00003801, -0.00001480, -0.00000472, -0.00004367, },
		{ -0.00002666, 0.00007210, -0.00007072, -0.00002630, 0.00003468, -0.00011337, 0.00017766, -0.00007109, 0.00015537, -0.00019253, 0.00026853, -0.00073836, 0.00051472, -0.00092456, 0.00070909, 0.00120248, -0.00094512, -0.00002545, -0.00024635, 0.00116522, -0.00091544, 0.00054782, -0.00056498, 0.00009859, -0.00006792, 0.00005115, 0.00000326, 0.00008639, -0.00002384, -0.00001339, -0.00001615, -0.00001852, },
		{ -0.00005715, 0.00008880, -0.00006599, -0.00003608, 0.00004327, -0.00011575, 0.00015703, -0.00005179, 0.00013814, -0.00017226, 0.00024768, -0.00062558, 0.00031380, -0.00062301, 0.00016060, 0.00170759, -0.00109277, 0.00011546, -0.00050357, 0.00122549, -0.00082074, 0.00043380, -0.00041605, -0.00001885, 0.00000033, 0.00000481, 0.00002400, 0.00006609, -0.00000837, -0.00000966, -0.00003017, 0.00000721, },
		{ -0.00007856, 0.00009380, -0.00005355, -0.00004581, 0.00004536, -0.00010472, 0.00012022, -0.00001489, 0.00009488, -0.00011635, 0.00017308, -0.00044619, 0.00005749, -0.00027815, -0.00037653, 0.00209832, -0.00110123, 0.00014581, -0.00062379, 0.00113875, -0.00062331, 0.00025254, -0.00022924, -0.00014940, 0.00007563, -0.00004606, 0.00004899, 0.00004209, 0.00000596, -0.00000387, -0.00004411, 0.00002918, },
		{ -0.00008794, 0.00008627, -0.00003453, -0.00005420, 0.00004078, -0.00008197, 0.00007107, 0.00003501, 0.00003147, -0.00003397, 0.00006035, -0.00022851, -0.00021296, 0.00005715, -0.00083891, 0.00232095, -0.00094247, 0.00005204, -0.00060558, 0.00092344, -0.00035188, 0.00003231, -0.00002932, -0.00027539, 0.00014656, -0.00009354, 0.00007265, 0.00001871, 0.00001694, 0.00000330, -0.00005521, 0.00004371, },
		{ -0.00008417, 0.00006719, -0.00001108, -0.00006001, 0.00003030, -0.00005073, 0.00001523, 0.00009107, -0.00004287, 0.00006110, -0.00006916, -0.00000612, -0.00045434, 0.00033341, -0.00117475, 0.00234244, -0.00061288, -0.00015567, -0.00046978, 0.00061580, -0.00004605, -0.00019323, 0.00015726, -0.00037941, 0.00020214, -0.00012991, 0.00008957, -0.00000006, 0.00002295, 0.00001084, -0.00006109, 0.00004833, },
		{ -0.00006816, 0.00003921, 0.00001393, -0.00006228, 0.00001554, -0.00001532, -0.00004063, 0.00014527, -0.00011701, 0.00015264, -0.00019172, 0.00018726, -0.00062821, 0.00051204, -0.00135097, 0.00215541, -0.00013485, -0.00044486, -0.00025565, 0.00026423, 0.00024962, -0.00039019, 0.00030631, -0.00044685, 0.00023346, -0.00014886, 0.00009532, -0.00001113, 0.00002313, 0.00001755, -0.00006009, 0.00004217, },
		{ -0.00004264, 0.00000624, 0.00003734, -0.00006046, -0.00000128, 0.00001953, -0.00008976, 0.00018952, -0.00017960, 0.00022448, -0.00028502, 0.00032286, -0.00070676, 0.00057108, -0.00135763, 0.00177941, 0.00044558, -0.00076572, -0.00001417, -0.00007811, 0.00049237, -0.00052959, 0.00039938, -0.00046810, 0.00023508, -0.00014649, 0.00008728, -0.00001282, 0.00001760, 0.00002224, -0.00005162, 0.00002610, },
		{ -0.00001179, -0.00002710, 0.00005610, -0.00005452, -0.00001766, 0.00004934, -0.00012630, 0.00021698, -0.00022080, 0.00026310, -0.00033169, 0.00038126, -0.00067710, 0.00050833, -0.00120873, 0.00125831, 0.00106466, -0.00105891, 0.00020042, -0.00036143, 0.00064762, -0.00059172, 0.00042640, -0.00043998, 0.00020585, -0.00012191, 0.00006502, -0.00000509, 0.00000737, 0.00002390, -0.00003620, 0.00000260, },
		{ 0.00001945, -0.00005621, 0.00006778, -0.00004494, -0.00003125, 0.00007054, -0.00014609, 0.00022305, -0.00023376, 0.00025975, -0.00032206, 0.00035531, -0.00054330, 0.00034143, -0.00093952, 0.00065411, 0.00164972, -0.00126502, 0.00033957, -0.00054677, 0.00069416, -0.00056910, 0.00038708, -0.00036623, 0.00014917, -0.00007749, 0.00003050, 0.00001042, -0.00000579, 0.00002189, -0.00001547, -0.00002461, },
		{ 0.00004613, -0.00007716, 0.00007088, -0.00003268, -0.00004014, 0.00008099, -0.00014729, 0.00020610, -0.00021572, 0.00021189, -0.00025577, 0.00025110, -0.00032572, 0.00010486, -0.00060065, 0.00003820, 0.00212942, -0.00133414, 0.00036839, -0.00061178, 0.00062730, -0.00046751, 0.00029079, -0.00025694, 0.00007246, -0.00001846, -0.00001220, 0.00003069, -0.00001964, 0.00001603, 0.00000804, -0.00005116, },
		{ 0.00006402, -0.00008727, 0.00006508, -0.00001902, -0.00004319, 0.00008025, -0.00013056, 0.00016773, -0.00016845, 0.00012374, -0.00014184, 0.00008697, -0.00005799, -0.00015564, -0.00025011, -0.00051875, 0.00244389, -0.00123419, 0.00027153, -0.00055389, 0.00045945, -0.00030507, 0.00015504, -0.00012709, -0.00001387, 0.00004774, -0.00005746, 0.00005181, -0.00003181, 0.00000665, 0.00003132, -0.00007269, },
		{ 0.00007031, -0.00008553, 0.00005124, -0.00000544, -0.00004018, 0.00006960, -0.00009895, 0.00011260, -0.00009801, 0.00000570, 0.00000262, -0.00010934, 0.00021821, -0.00039042, 0.00005588, -0.00095666, 0.00255336, -0.00095666, 0.00005588, -0.00039042, 0.00021821, -0.00010934, 0.00000262, 0.00000570, -0.00009801, 0.00011260, -0.00009895, 0.00006960, -0.00004018, -0.00000544, 0.00005124, -0.00008553, },
		{ 0.00006402, -0.00007269, 0.00003132, 0.00000665, -0.00003181, 0.00005181, -0.00005746, 0.00004774, -0.00001387, -0.00012709, 0.00015504, -0.00030507, 0.00045945, -0.00055389, 0.00027153, -0.00123419, 0.00244389, -0.00051875, -0.00025011, -0.00015564, -0.00005799, 0.00008697, -0.00014184, 0.00012374, -0.00016845, 0.00016773, -0.00013056, 0.00008025, -0.00004319, -0.00001902, 0.00006508, -0.00008727, },
		{ 0.00004613, -0.00005116, 0.00000804, 0.00001603, -0.00001964, 0.00003069, -0.00001220, -0.00001846, 0.00007246, -0.00025694, 0.00029079, -0.00046751, 0.00062730, -0.00061178, 0.00036839, -0.00133414, 0.00212942, 0.00003820, -0.00060065, 0.00010486, -0.00032572, 0.00025110, -0.00025577, 0.00021189, -0.00021572, 0.00020610, -0.00014729, 0.00008099, -0.00004014, -0.00003268, 0.00007088, -0.00007716, },
		{ 0.00001945, -0.00002461, -0.00001547, 0.00002189, -0.00000579, 0.00001042, 0.00003050, -0.00007749, 0.00014917, -0.00036623, 0.00038708, -0.00056910, 0.00069416, -0.00054677, 0.00033957, -0.00126502, 0.00164972, 0.00065411, -0.00093952, 0.00034143, -0.00054330, 0.00035531, -0.00032206, 0.00025975, -0.00023376, 0.00022305, -0.00014609, 0.00007054, -0.00003125, -0.00004494, 0.00006778, -0.00005621, },
		{ -0.00001179, 0.00000260, -0.00003620, 0.00002390, 0.00000737, -0.00000509, 0.00006502, -0.00012191, 0.00020585, -0.00043998, 0.00042640, -0.00059172, 0.00064762, -0.00036143, 0.00020042, -0.00105891, 0.00106466, 0.00125831, -0.00120873, 0.00050833, -0.00067710, 0.00038126, -0.00033169, 0.00026310, -0.00022080, 0.00021698, -0.00012630, 0.00004934, -0.00001766, -0.00005452, 0.00005610, -0.00002710, },
		{ -0.00004264, 0.00002610, -0.00005162, 0.00002224, 0.00001760, -0.00001282, 0.00008728, -0.00014649, 0.00023508, -0.00046810, 0.00039938, -0.00052959, 0.00049237, -0.00007811, -0.00001417, -0.00076572, 0.00044558, 0.00177941, -0.00135763, 0.00057108, -0.00070676, 0.00032286, -0.00028502, 0.00022448, -0.00017960, 0.00018952, -0.00008976, 0.00001953, -0.00000128, -0.00006046, 0.00003734, 0.00000624, },
		{ -0.00006816, 0.00004217, -0.00006009, 0.00001755, 0.00002313, -0.00001113, 0.00009532, -0.00014886, 0.00023346, -0.00044685, 0.00030631, -0.00039019, 0.00024962, 0.00026423, -0.00025565, -0.00044486, -0.00013485, 0.00215541, -0.00135097, 0.00051204, -0.00062821, 0.00018726, -0.00019172, 0.00015264, -0.00011701, 0.00014527, -0.00004063, -0.00001532, 0.00001554, -0.00006228, 0.00001393, 0.00003921, },
		{ -0.00008417, 0.00004833, -0.00006109, 0.00001084, 0.00002295, -0.00000006, 0.00008957, -0.00012991, 0.00020214, -0.00037941, 0.00015726, -0.00019323, -0.00004605, 0.00061580, -0.00046978, -0.00015567, -0.00061288, 0.00234244, -0.00117475, 0.00033341, -0.00045434, -0.00000612, -0.00006916, 0.00006110, -0.00004287, 0.00009107, 0.00001523, -0.00005073, 0.00003030, -0.00006001, -0.00001108, 0.00006719, },
		{ -0.00008794, 0.00004371, -0.00005521, 0.00000330, 0.00001694, 0.00001871, 0.00007265, -0.00009354, 0.00014656, -0.00027539, -0.00002932, 0.00003231, -0.00035188, 0.00092344, -0.00060558, 0.00005204, -0.00094247, 0.00232095, -0.00083891, 0.00005715, -0.00021296, -0.00022851, 0.00006035, -0.00003397, 0.00003147, 0.00003501, 0.00007107, -0.00008197, 0.00004078, -0.00005420, -0.00003453, 0.00008627, },
		{ -0.00007856, 0.00002918, -0.00004411, -0.00000387, 0.00000596, 0.00004209, 0.00004899, -0.00004606, 0.00007563, -0.00014940, -0.00022924, 0.00025254, -0.00062331, 0.00113875, -0.00062379, 0.00014581, -0.00110123, 0.00209832, -0.00037653, -0.00027815, 0.00005749, -0.00044619, 0.00017308, -0.00011635, 0.00009488, -0.00001489, 0.00012022, -0.00010472, 0.00004536, -0.00004581, -0.00005355, 0.00009380, },
		{ -0.00005715, 0.00000721, -0.00003017, -0.00000966, -0.00000837, 0.00006609, 0.00002400, 0.00000481, 0.00000033, -0.00001885, -0.00041605, 0.00043380, -0.00082074, 0.00122549, -0.00050357, 0.00011546, -0.00109277, 0.00170759, 0.00016060, -0.00062301, 0.00031380, -0.00062558, 0.00024768, -0.00017226, 0.00013814, -0.00005179, 0.00015703, -0.00011575, 0.00004327, -0.00003608, -0.00006599, 0.00008880, },
		{ -0.00002666, -0.00001852, -0.00001615, -0.00001339, -0.00002384, 0.00008639, 0.00000326, 0.00005115, -0.00006792, 0.00009859, -0.00056498, 0.00054782, -0.00091544, 0.00116522, -0.00024635, -0.00002545, -0.00094512, 0.00120248, 0.00070909, -0.00092456, 0.00051472, -0.00073836, 0.00026853, -0.00019253, 0.00015537, -0.00007109, 0.00017766, -0.00011337, 0.00003468, -0.00002630, -0.00007072, 0.00007210, },
		{ 0.00000855, -0.00004367, -0.00000472, -0.00001480, -0.00003801, 0.00009908, -0.00000846, 0.00008601, -0.00011907, 0.00018765, -0.00065652, 0.00057600, -0.00089392, 0.00096032, 0.00012373, -0.00024211, -0.00070536, 0.00064939, 0.00120289, -0.00113455, 0.00062728, -0.00076584, 0.00022838, -0.00017402, 0.00014481, -0.00007109, 0.00018061, -0.00009772, 0.00002073, -0.00001769, -0.00006777, 0.00004621, },
		{ 0.00004338, -0.00006396, 0.00000197, -0.00001410, -0.00004860, 0.00010126, -0.00000797, 0.00010451, -0.00014590, 0.00023763, -0.00067928, 0.00051234, -0.00075991, 0.00063390, 0.00056291, -0.00048420, -0.00043160, 0.00011774, 0.00158273, -0.00121686, 0.00063191, -0.00070188, 0.00012949, -0.00012010, 0.00010915, -0.00005320, 0.00016687, -0.00007073, 0.00000334, -0.00001115, -0.00005830, 0.00001492, },
		{ 0.00007271, -0.00007578, 0.00000261, -0.00001190, -0.00005382, 0.00009150, 0.00000581, 0.00010454, -0.00014510, 0.00024396, -0.00063165, 0.00036438, -0.00053382, 0.00022658, 0.00101400, -0.00069426, -0.00018322, -0.00033006, 0.00180446, -0.00115308, 0.00052553, -0.00055385, -0.00001646, -0.00004010, 0.00005500, -0.00002163, 0.00013969, -0.00003585, -0.00001508, -0.00000719, -0.00004440, -0.00001722, },
		{ 0.00009224, -0.00007683, -0.00000303, -0.00000910, -0.00005268, 0.00007014, 0.00003164, 0.00008706, -0.00011778, 0.00020876, -0.00052194, 0.00015216, -0.00024957, -0.00020934, 0.00141473, -0.00081722, -0.00001120, -0.00064738, 0.00184532, -0.00094561, 0.00032208, -0.00034165, -0.00019008, 0.00005213, -0.00000810, 0.00001723, 0.00010412, 0.00000243, -0.00003200, -0.00000587, -0.00002872, -0.00004564, },
		},
		{ { -0.00011383, 0.00021564, -0.00005861, -0.00008715, 0.00012638, -0.00012661, 0.00007626, -0.00035228, 0.00010340, -0.00003527, 0.00005181, -0.00000997, -0.00081152, 0.00190161, -0.00074920, -0.00010334, -0.00008218, -0.00010334, -0.00074920, 0.00190161, -0.00081152, -0.00000997, 0.00005181, -0.00003527, 0.00010340, -0.00035228, 0.00007626, -0.00012661, 0.00012638, -0.00008715, -0.00005861, 0.00021564, },
		{ -0.00009338, 0.00016559, 0.00002665, -0.00013925, 0.00011997, -0.00006966, 0.00000613, -0.00020837, -0.00013256, 0.00021266, -0.00024420, 0.00039678, -0.00124069, 0.00190921, -0.00032072, -0.00037543, -0.00003699, 0.00002517, -0.00088901, 0.00155413, -0.00017859, -0.00048453, 0.00035975, -0.00027128, 0.00029823, -0.00043716, 0.00009910, -0.00014270, 0.00009797, -0.00002184, -0.00012514, 0.00022648, },
		{ -0.00003789, 0.00008984, 0.00011222, -0.00016530, 0.00008009, 0.00001136, -0.00008045, -0.00005604, -0.00033821, 0.00038879, -0.00043308, 0.00062764, -0.00136880, 0.00154491, 0.00035104, -0.00072993, 0.00006897, 0.00001869, -0.00078237, 0.00098352, 0.00050384, -0.00089120, 0.00057504, -0.00041389, 0.00039225, -0.00042924, 0.00006006, -0.00011490, 0.00004313, 0.00004097, -0.00015969, 0.00019496, },
		{ 0.00003679, 0.00000837, 0.00017940, -0.00015873, 0.00001744, 0.00009139, -0.00014740, 0.00005469, -0.00045396, 0.00043512, -0.00045973, 0.00063405, -0.00118315, 0.00087552, 0.00113586, -0.00104783, 0.00015945, -0.00005888, -0.00053425, 0.00035262, 0.00106946, -0.00110836, 0.00061832, -0.00041140, 0.00035758, -0.00032362, -0.00003374, -0.00005472, -0.00002250, 0.00008693, -0.00015756, 0.00012914, },
		{ 0.00010973, -0.00005823, 0.00021426, -0.00012099, -0.00005121, 0.00014526, -0.00016601, 0.00009117, -0.00045143, 0.00033827, -0.00032573, 0.00043901, -0.00075723, 0.00005098, 0.00184992, -0.00118812, 0.00014475, -0.00012245, -0.00026990, -0.00017992, 0.00138830, -0.00106654, 0.00046262, -0.00025987, 0.00020775, -0.00014680, -0.00015536, 0.00001571, -0.00008080, 0.00010679, -0.00012338, 0.00004578, },
		{ 0.00016115, -0.00009463, 0.00021121, -0.00006092, -0.00010808, 0.00015597, -0.00012506, 0.00004879, -0.00034197, 0.00013311, -0.00008659, 0.00012583, -0.00022603, -0.00073679, 0.00230674, -0.00103621, -0.00003697, -0.00010599, -0.00008989, -0.00050544, 0.00140298, -0.00077055, 0.00014364, -0.00000526, -0.00000630, 0.00005126, -0.00026675, 0.00007116, -0.00011656, 0.00009847, -0.00006928, -0.00003472, },
		{ 0.00017827, -0.00009491, 0.00017412, 0.00000785, -0.00013964, 0.00012026, -0.00003476, -0.00004825, -0.00017177, -0.00010997, 0.00016907, -0.00019202, 0.00025653, -0.00130791, 0.00237239, -0.00054873, -0.00038582, 0.00000659, -0.00003819, -0.00059208, 0.00114200, -0.00029831, -0.00025026, 0.00027104, -0.00021274, 0.00021307, -0.00033138, 0.00009178, -0.00012185, 0.00006712, -0.00001094, -0.00009427, },
		{ 0.00015871, -0.00006363, 0.00011477, 0.00007070, -0.00014034, 0.00004953, 0.00007659, -0.00015601, -0.00000522, -0.00030962, 0.00035106, -0.00040966, 0.00056439, -0.00154493, 0.00200750, 0.00022135, -0.00082739, 0.00017161, -0.00009355, -0.00048527, 0.00070631, 0.00022312, -0.00060539, 0.00047858, -0.00034233, 0.00029216, -0.00032652, 0.00006967, -0.00009795, 0.00002306, 0.00003709, -0.00012206, },
		{ 0.00011059, -0.00001378, 0.00004900, 0.00011590, -0.00011371, -0.00003394, 0.00017226, -0.00022669, 0.00009765, -0.00040164, 0.00039895, -0.00046482, 0.00063527, -0.00142381, 0.00128228, 0.00112896, -0.00122732, 0.00030099, -0.00018602, -0.00028301, 0.00023419, 0.00065659, -0.00081851, 0.00054744, -0.00035120, 0.00026824, -0.00025114, 0.00001143, -0.00005443, -0.00002160, 0.00006529, -0.00011708, },
		{ 0.00004933, 0.00003783, -0.00000792, 0.00013731, -0.00007063, -0.00010447, 0.00021973, -0.00022580, 0.00010192, -0.00036083, 0.00030246, -0.00035614, 0.00048475, -0.00101646, 0.00035935, 0.00197122, -0.00143013, 0.00029742, -0.00023149, -0.00009583, -0.00014343, 0.00089647, -0.00082933, 0.00045122, -0.00023520, 0.00015381, -0.00012652, -0.00006418, -0.00000558, -0.00005644, 0.00007128, -0.00008742, },
		{ -0.00000765, 0.00007600, -0.00004541, 0.00013528, -0.00002534, -0.00014143, 0.00020194, -0.00014460, 0.00000974, -0.00020930, 0.00010412, -0.00013982, 0.00019480, -0.00046461, -0.00055076, 0.00254230, -0.00131036, 0.00009676, -0.00017082, -0.00000568, -0.00034269, 0.00089921, -0.00063919, 0.00021578, -0.00003122, -0.00000995, 0.00001038, -0.00013268, 0.00003434, -0.00007541, 0.00005941, -0.00004681, },
		{ -0.00004653, 0.00009172, -0.00005996, 0.00011554, 0.00000934, -0.00013596, 0.00012339, -0.00000371, -0.00014069, -0.00000844, -0.00011670, 0.00009346, -0.00011849, 0.00006597, -0.00124796, 0.00269230, -0.00081912, -0.00029829, 0.00000140, -0.00003834, -0.00034866, 0.00069350, -0.00030977, -0.00008923, 0.00019499, -0.00016674, 0.00011896, -0.00017214, 0.00005559, -0.00007778, 0.00003829, -0.00000966, },
		{ -0.00006117, 0.00008432, -0.00005504, 0.00008664, 0.00002605, -0.00009364, 0.00000892, 0.00015360, -0.00028867, 0.00016278, -0.00027266, 0.00025302, -0.00034621, 0.00043589, -0.00159699, 0.00237185, -0.00000988, -0.00080635, 0.00022957, -0.00015832, -0.00021304, 0.00036719, 0.00005751, -0.00037304, 0.00037051, -0.00026351, 0.00016815, -0.00017055, 0.00005570, -0.00006735, 0.00001727, 0.00001375, },
		{ -0.00005430, 0.00006078, -0.00003883, 0.00005685, 0.00002474, -0.00003235, -0.00010448, 0.00027603, -0.00037244, 0.00023566, -0.00030021, 0.00028170, -0.00042295, 0.00057113, -0.00156128, 0.00164832, 0.00096638, -0.00128355, 0.00041459, -0.00028743, -0.00002951, 0.00003571, 0.00035454, -0.00055328, 0.00043910, -0.00026778, 0.00014645, -0.00012983, 0.00003953, -0.00005019, 0.00000295, 0.00002026, },
		{ -0.00003563, 0.00003238, -0.00002072, 0.00003167, 0.00001193, 0.00002409, -0.00018000, 0.00032190, -0.00035111, 0.00017621, -0.00018317, 0.00017670, -0.00034585, 0.00047870, -0.00120659, 0.00068822, 0.00189695, -0.00156403, 0.00045052, -0.00033960, 0.00010382, -0.00019683, 0.00050097, -0.00058240, 0.00037935, -0.00017797, 0.00006558, -0.00006495, 0.00001686, -0.00003202, -0.00000276, 0.00001406, },
		{ -0.00001756, 0.00001019, -0.00000779, 0.00001266, -0.00000230, 0.00005413, -0.00019348, 0.00027381, -0.00021966, -0.00000218, 0.00004027, -0.00000960, -0.00017055, 0.00023525, -0.00067439, -0.00028900, 0.00256425, -0.00151246, 0.00026753, -0.00025887, 0.00012145, -0.00027107, 0.00046888, -0.00046227, 0.00021113, -0.00002353, -0.00004365, 0.00000123, -0.00000163, -0.00001602, -0.00000271, 0.00000428, },
		{ -0.00001025, 0.00000084, -0.00000256, -0.00000207, -0.00000849, 0.00004589, -0.00014169, 0.00014519, -0.00001169, -0.00024239, 0.00028950, -0.00018912, 0.00001389, -0.00004617, -0.00013446, -0.00107167, 0.00280655, -0.00107167, -0.00013446, -0.00004617, 0.00001389, -0.00018912, 0.00028950, -0.00024239, -0.00001169, 0.00014519, -0.00014169, 0.00004589, -0.00000849, -0.00000207, -0.00000256, 0.00000084, },
		{ -0.00001756, 0.00000428, -0.00000271, -0.00001602, -0.00000163, 0.00000123, -0.00004365, -0.00002353, 0.00021113, -0.00046227, 0.00046888, -0.00027107, 0.00012145, -0.00025887, 0.00026753, -0.00151246, 0.00256425, -0.00028900, -0.00067439, 0.00023525, -0.00017055, -0.00000960, 0.00004027, -0.00000218, -0.00021966, 0.00027381, -0.00019348, 0.00005413, -0.00000230, 0.00001266, -0.00000779, 0.00001019, },
		{ -0.00003563, 0.00001406, -0.00000276, -0.00003202, 0.00001686, -0.00006495, 0.00006558, -0.00017797, 0.00037935, -0.00058240, 0.00050097, -0.00019683, 0.00010382, -0.00033960, 0.00045052, -0.00156403, 0.00189695, 0.00068822, -0.00120659, 0.00047870, -0.00034585, 0.00017670, -0.00018317, 0.00017621, -0.00035111, 0.00032190, -0.00018000, 0.00002409, 0.00001193, 0.00003167, -0.00002072, 0.00003238, },
		{ -0.00005430, 0.00002026, 0.00000295, -0.00005019, 0.00003953, -0.00012983, 0.00014645, -0.00026778, 0.00043910, -0.00055328, 0.00035454, 0.00003571, -0.00002951, -0.00028743, 0.00041459, -0.00128355, 0.00096638, 0.00164832, -0.00156128, 0.00057113, -0.00042295, 0.00028170, -0.00030021, 0.00023566, -0.00037244, 0.00027603, -0.00010448, -0.00003235, 0.00002474, 0.00005685, -0.00003883, 0.00006078, },
		{ -0.00006117, 0.00001375, 0.00001727, -0.00006735, 0.00005570, -0.00017055, 0.00016815, -0.00026351, 0.00037051, -0.00037304, 0.00005751, 0.00036719, -0.00021304, -0.00015832, 0.00022957, -0.00080635, -0.00000988, 0.00237185, -0.00159699, 0.00043589, -0.00034621, 0.00025302, -0.00027266, 0.00016278, -0.00028867, 0.00015360, 0.00000892, -0.00009364, 0.00002605, 0.00008664, -0.00005504, 0.00008432, },
		{ -0.00004653, -0.00000966, 0.00003829, -0.00007778, 0.00005559, -0.00017214, 0.00011896, -0.00016674, 0.00019499, -0.00008923, -0.00030977, 0.00069350, -0.00034866, -0.00003834, 0.00000140, -0.00029829, -0.00081912, 0.00269230, -0.00124796, 0.00006597, -0.00011849, 0.00009346, -0.00011670, -0.00000844, -0.00014069, -0.00000371, 0.00012339, -0.00013596, 0.00000934, 0.00011554, -0.00005996, 0.00009172, },
		{ -0.00000765, -0.00004681, 0.00005941, -0.00007541, 0.00003434, -0.00013268, 0.00001038, -0.00000995, -0.00003122, 0.00021578, -0.00063919, 0.00089921, -0.00034269, -0.00000568, -0.00017082, 0.00009676, -0.00131036, 0.00254230, -0.00055076, -0.00046461, 0.00019480, -0.00013982, 0.00010412, -0.00020930, 0.00000974, -0.00014460, 0.00020194, -0.00014143, -0.00002534, 0.00013528, -0.00004541, 0.00007600, },
		{ 0.00004933, -0.00008742, 0.00007128, -0.00005644, -0.00000558, -0.00006418, -0.00012652, 0.00015381, -0.00023520, 0.00045122, -0.00082933, 0.00089647, -0.00014343, -0.00009583, -0.00023149, 0.00029742, -0.00143013, 0.00197122, 0.00035935, -0.00101646, 0.00048475, -0.00035614, 0.00030246, -0.00036083, 0.00010192, -0.00022580, 0.00021973, -0.00010447, -0.00007063, 0.00013731, -0.00000792, 0.00003783, },
		{ 0.00011059, -0.00011708, 0.00006529, -0.00002160, -0.00005443, 0.00001143, -0.00025114, 0.00026824, -0.00035120, 0.00054744, -0.00081851, 0.00065659, 0.00023419, -0.00028301, -0.00018602, 0.00030099, -0.00122732, 0.00112896, 0.00128228, -0.00142381, 0.00063527, -0.00046482, 0.00039895, -0.00040164, 0.00009765, -0.00022669, 0.00017226, -0.00003394, -0.00011371, 0.00011590, 0.00004900, -0.00001378, },
		{ 0.00015871, -0.00012206, 0.00003709, 0.00002306, -0.00009795, 0.00006967, -0.00032652, 0.00029216, -0.00034233, 0.00047858, -0.00060539, 0.00022312, 0.00070631, -0.00048527, -0.00009355, 0.00017161, -0.00082739, 0.00022135, 0.00200750, -0.00154493, 0.00056439, -0.00040966, 0.00035106, -0.00030962, -0.00000522, -0.00015601, 0.00007659, 0.00004953, -0.00014034, 0.00007070, 0.00011477, -0.00006363, },
		{ 0.00017827, -0.00009427, -0.00001094, 0.00006712, -0.00012185, 0.00009178, -0.00033138, 0.00021307, -0.00021274, 0.00027104, -0.00025026, -0.00029831, 0.00114200, -0.00059208, -0.00003819, 0.00000659, -0.00038582, -0.00054873, 0.00237239, -0.00130791, 0.00025653, -0.00019202, 0.00016907, -0.00010997, -0.00017177, -0.00004825, -0.00003476, 0.00012026, -0.00013964, 0.00000785, 0.00017412, -0.00009491, },
		{ 0.00016115, -0.00003472, -0.00006928, 0.00009847, -0.00011656, 0.00007116, -0.00026675, 0.00005126, -0.00000630, -0.00000526, 0.00014364, -0.00077055, 0.00140298, -0.00050544, -0.00008989, -0.00010599, -0.00003697, -0.00103621, 0.00230674, -0.00073679, -0.00022603, 0.00012583, -0.00008659, 0.00013311, -0.00034197, 0.00004879, -0.00012506, 0.00015597, -0.00010808, -0.00006092, 0.00021121, -0.00009463, },
		{ 0.00010973, 0.00004578, -0.00012338, 0.00010679, -0.00008080, 0.00001571, -0.00015536, -0.00014680, 0.00020775, -0.00025987, 0.00046262, -0.00106654, 0.00138830, -0.00017992, -0.00026990, -0.00012245, 0.00014475, -0.00118812, 0.00184992, 0.00005098, -0.00075723, 0.00043901, -0.00032573, 0.00033827, -0.00045143, 0.00009117, -0.00016601, 0.00014526, -0.00005121, -0.00012099, 0.00021426, -0.00005823, },
		{ 0.00003679, 0.00012914, -0.00015756, 0.00008693, -0.00002250, -0.00005472, -0.00003374, -0.00032362, 0.00035758, -0.00041140, 0.00061832, -0.00110836, 0.00106946, 0.00035262, -0.00053425, -0.00005888, 0.00015945, -0.00104783, 0.00113586, 0.00087552, -0.00118315, 0.00063405, -0.00045973, 0.00043512, -0.00045396, 0.00005469, -0.00014740, 0.00009139, 0.00001744, -0.00015873, 0.00017940, 0.00000837, },
		{ -0.00003789, 0.00019496, -0.00015969, 0.00004097, 0.00004313, -0.00011490, 0.00006006, -0.00042924, 0.00039225, -0.00041389, 0.00057504, -0.00089120, 0.00050384, 0.00098352, -0.00078237, 0.00001869, 0.00006897, -0.00072993, 0.00035104, 0.00154491, -0.00136880, 0.00062764, -0.00043308, 0.00038879, -0.00033821, -0.00005604, -0.00008045, 0.00001136, 0.00008009, -0.00016530, 0.00011222, 0.00008984, },
		{ -0.00009338, 0.00022648, -0.00012514, -0.00002184, 0.00009797, -0.00014270, 0.00009910, -0.00043716, 0.00029823, -0.00027128, 0.00035975, -0.00048453, -0.00017859, 0.00155413, -0.00088901, 0.00002517, -0.00003699, -0.00037543, -0.00032072, 0.00190921, -0.00124069, 0.00039678, -0.00024420, 0.00021266, -0.00013256, -0.00020837, 0.00000613, -0.00006966, 0.00011997, -0.00013925, 0.00002665, 0.00016559, },
		},
		{ { -0.00009944, 0.00006422, 0.00002901, 0.00002754, -0.00021634, 0.00009211, -0.00011721, 0.00018512, -0.00008362, -0.00015879, 0.00012260, -0.00079820, 0.00177544, -0.00073363, -0.00002361, -0.00003788, -0.00006779, -0.00003788, -0.00002361, -0.00073363, 0.00177544, -0.00079820, 0.00012260, -0.00015879, -0.00008362, 0.00018512, -0.00011721, 0.00009211, -0.00021634, 0.00002754, 0.00002901, 0.00006422, },
		{ -0.00007222, -0.00000335, 0.00008005, 0.00000490, -0.00013100, -0.00006810, 0.00006392, -0.00007592, 0.00026830, -0.00052390, 0.00053257, -0.00119155, 0.00163833, -0.00010547, -0.00045124, 0.00013211, -0.00011397, -0.00010094, 0.00019114, -0.00090826, 0.00135698, -0.00003946, -0.00045017, 0.00027357, -0.00043201, 0.00039730, -0.00024273, 0.00019045, -0.00024032, 0.00001745, -0.00000134, 0.00009120, },
		{ -0.00000569, -0.00007817, 0.00011508, -0.00001177, -0.00004279, -0.00020285, 0.00018810, -0.00024376, 0.00045333, -0.00063644, 0.00059144, -0.00107793, 0.00095077, 0.00081941, -0.00091390, 0.00029694, -0.00020025, -0.00006432, 0.00018362, -0.00070064, 0.00061166, 0.00077172, -0.00090051, 0.00054246, -0.00059582, 0.00043051, -0.00022381, 0.00016903, -0.00017307, -0.00004171, 0.00000500, 0.00007068, },
		{ 0.00006349, -0.00011868, 0.00010028, 0.00000988, -0.00000030, -0.00024751, 0.00018490, -0.00024292, 0.00040102, -0.00045161, 0.00029962, -0.00055481, -0.00003237, 0.00169631, -0.00113110, 0.00028417, -0.00021195, -0.00002300, 0.00007221, -0.00033108, -0.00012644, 0.00130476, -0.00099239, 0.00049581, -0.00048344, 0.00024338, -0.00005114, 0.00003718, -0.00003648, -0.00012958, 0.00003413, 0.00002396, },
		{ 0.00009817, -0.00009914, 0.00002831, 0.00007136, -0.00001241, -0.00019560, 0.00006352, -0.00010057, 0.00017048, -0.00008053, -0.00016406, 0.00009570, -0.00091809, 0.00214328, -0.00086736, -0.00001918, -0.00006906, -0.00006810, -0.00000474, -0.00002703, -0.00058526, 0.00136933, -0.00066119, 0.00014489, -0.00014691, -0.00008671, 0.00019316, -0.00012923, 0.00010360, -0.00019881, 0.00005287, -0.00001436, },
		{ 0.00008194, -0.00002571, -0.00007290, 0.00013898, -0.00004559, -0.00009728, -0.00009843, 0.00007826, -0.00009301, 0.00027890, -0.00054581, 0.00055372, -0.00136043, 0.00192926, -0.00008288, -0.00056023, 0.00019024, -0.00020328, 0.00001109, 0.00009668, -0.00067159, 0.00099560, -0.00004636, -0.00033818, 0.00024179, -0.00039981, 0.00037674, -0.00022836, 0.00017316, -0.00020359, 0.00003209, -0.00001868, },
		{ 0.00002800, 0.00006548, -0.00015281, 0.00016343, -0.00004926, -0.00002335, -0.00020596, 0.00017878, -0.00024597, 0.00045358, -0.00065415, 0.00062213, -0.00122391, 0.00110028, 0.00100267, -0.00110203, 0.00040330, -0.00033002, 0.00006893, 0.00007160, -0.00048161, 0.00039928, 0.00058189, -0.00070965, 0.00048100, -0.00053746, 0.00038960, -0.00019083, 0.00013348, -0.00012964, -0.00003204, 0.00001159, },
		{ -0.00002923, 0.00012754, -0.00016583, 0.00011170, 0.00000707, -0.00001826, -0.00020548, 0.00014572, -0.00022681, 0.00038523, -0.00045533, 0.00031603, -0.00064059, -0.00002756, 0.00198118, -0.00132053, 0.00037010, -0.00031516, 0.00006360, 0.00001575, -0.00021184, -0.00014279, 0.00096075, -0.00078163, 0.00044696, -0.00042850, 0.00020765, -0.00002252, 0.00000654, -0.00000698, -0.00011246, 0.00005199, },
		{ -0.00005485, 0.00012870, -0.00009824, -0.00000803, 0.00010885, -0.00007163, -0.00011546, 0.00001274, -0.00008218, 0.00014994, -0.00007756, -0.00016268, 0.00006744, -0.00101011, 0.00243678, -0.00097516, -0.00001169, -0.00009724, -0.00006900, 0.00002925, -0.00003153, -0.00043415, 0.00096248, -0.00051851, 0.00015951, -0.00012872, -0.00008968, 0.00019413, -0.00013398, 0.00010372, -0.00016561, 0.00006877, },
		{ -0.00003428, 0.00007006, 0.00002150, -0.00014493, 0.00020051, -0.00012601, -0.00001107, -0.00012418, 0.00007246, -0.00009896, 0.00027234, -0.00054591, 0.00055642, -0.00148196, 0.00214772, -0.00006166, -0.00064667, 0.00024160, -0.00028507, 0.00011743, 0.00000645, -0.00044079, 0.00064470, -0.00005069, -0.00023033, 0.00020827, -0.00035868, 0.00034252, -0.00020237, 0.00014430, -0.00015736, 0.00004094, },
		{ 0.00001809, -0.00001494, 0.00013592, -0.00023006, 0.00021748, -0.00011635, 0.00002685, -0.00017129, 0.00013108, -0.00022674, 0.00042791, -0.00064492, 0.00062972, -0.00132252, 0.00120352, 0.00114342, -0.00124746, 0.00049497, -0.00044533, 0.00019104, -0.00003119, -0.00027558, 0.00020141, 0.00039956, -0.00052302, 0.00041304, -0.00046693, 0.00033552, -0.00014970, 0.00009315, -0.00008522, -0.00002511, },
		{ 0.00006755, -0.00007934, 0.00018858, -0.00021411, 0.00012761, -0.00001628, -0.00003381, -0.00009866, 0.00006463, -0.00019316, 0.00034831, -0.00044077, 0.00031962, -0.00069760, -0.00002717, 0.00218343, -0.00145951, 0.00044376, -0.00040436, 0.00014025, -0.00003255, -0.00010474, -0.00015192, 0.00063671, -0.00057713, 0.00038976, -0.00036342, 0.00016471, 0.00000629, -0.00002020, 0.00001731, -0.00009750, },
		{ 0.00007986, -0.00008903, 0.00015449, -0.00009678, -0.00004326, 0.00013818, -0.00015305, 0.00003960, -0.00006349, -0.00006462, 0.00012444, -0.00007483, -0.00015483, 0.00004079, -0.00106547, 0.00262607, -0.00104766, 0.00000045, -0.00012147, -0.00007046, 0.00006291, -0.00003880, -0.00029131, 0.00058913, -0.00038087, 0.00016400, -0.00010478, -0.00009241, 0.00018785, -0.00012937, 0.00009767, -0.00013660, },
		{ 0.00004196, -0.00004174, 0.00005377, 0.00006957, -0.00021861, 0.00026392, -0.00024181, 0.00013937, -0.00014266, 0.00004093, -0.00009132, 0.00024692, -0.00052224, 0.00053895, -0.00154289, 0.00227024, -0.00004434, -0.00070144, 0.00028196, -0.00035422, 0.00021203, -0.00007445, -0.00023328, 0.00033285, -0.00005055, -0.00013657, 0.00017655, -0.00031217, 0.00029793, -0.00016740, 0.00011352, -0.00011848, },
		{ -0.00002961, 0.00003214, -0.00005926, 0.00020549, -0.00030952, 0.00027793, -0.00021837, 0.00011428, -0.00009178, 0.00003895, -0.00018405, 0.00037475, -0.00060656, 0.00061117, -0.00136205, 0.00124813, 0.00122783, -0.00133670, 0.00056567, -0.00053917, 0.00029477, -0.00011801, -0.00009837, 0.00003406, 0.00024013, -0.00035569, 0.00034396, -0.00038950, 0.00027288, -0.00010476, 0.00005748, -0.00004991, },
		{ -0.00009762, 0.00008724, -0.00012579, 0.00024587, -0.00026327, 0.00014922, -0.00006405, -0.00004493, 0.00008120, -0.00006225, -0.00014192, 0.00029008, -0.00040687, 0.00030839, -0.00071888, -0.00003262, 0.00228295, -0.00153568, 0.00050098, -0.00047453, 0.00020240, -0.00006899, -0.00001865, -0.00015210, 0.00035746, -0.00039377, 0.00032775, -0.00029230, 0.00011784, 0.00003234, -0.00003724, 0.00003405, },
		{ -0.00012534, 0.00008920, -0.00011435, 0.00017447, -0.00009481, -0.00007594, 0.00015695, -0.00025706, 0.00027603, -0.00016574, -0.00004983, 0.00009505, -0.00007243, -0.00014090, 0.00001802, -0.00107857, 0.00269146, -0.00107857, 0.00001802, -0.00014090, -0.00007243, 0.00009505, -0.00004983, -0.00016574, 0.00027603, -0.00025706, 0.00015695, -0.00007594, -0.00009481, 0.00017447, -0.00011435, 0.00008920, },
		{ -0.00009762, 0.00003405, -0.00003724, 0.00003234, 0.00011784, -0.00029230, 0.00032775, -0.00039377, 0.00035746, -0.00015210, -0.00001865, -0.00006899, 0.00020240, -0.00047453, 0.00050098, -0.00153568, 0.00228295, -0.00003262, -0.00071888, 0.00030839, -0.00040687, 0.00029008, -0.00014192, -0.00006225, 0.00008120, -0.00004493, -0.00006405, 0.00014922, -0.00026327, 0.00024587, -0.00012579, 0.00008724, },
		{ -0.00002961, -0.00004991, 0.00005748, -0.00010476, 0.00027288, -0.00038950, 0.00034396, -0.00035569, 0.00024013, 0.00003406, -0.00009837, -0.00011801, 0.00029477, -0.00053917, 0.00056567, -0.00133670, 0.00122783, 0.00124813, -0.00136205, 0.00061117, -0.00060656, 0.00037475, -0.00018405, 0.00003895, -0.00009178, 0.00011428, -0.00021837, 0.00027793, -0.00030952, 0.00020549, -0.00005926, 0.00003214, },
		{ 0.00004196, -0.00011848, 0.00011352, -0.00016740, 0.00029793, -0.00031217, 0.00017655, -0.00013657, -0.00005055, 0.00033285, -0.00023328, -0.00007445, 0.00021203, -0.00035422, 0.00028196, -0.00070144, -0.00004434, 0.00227024, -0.00154289, 0.00053895, -0.00052224, 0.00024692, -0.00009132, 0.00004093, -0.00014266, 0.00013937, -0.00024181, 0.00026392, -0.00021861, 0.00006957, 0.00005377, -0.00004174, },
		{ 0.00007986, -0.00013660, 0.00009767, -0.00012937, 0.00018785, -0.00009241, -0.00010478, 0.00016400, -0.00038087, 0.00058913, -0.00029131, -0.00003880, 0.00006291, -0.00007046, -0.00012147, 0.00000045, -0.00104766, 0.00262607, -0.00106547, 0.00004079, -0.00015483, -0.00007483, 0.00012444, -0.00006462, -0.00006349, 0.00003960, -0.00015305, 0.00013818, -0.00004326, -0.00009678, 0.00015449, -0.00008903, },
		{ 0.00006755, -0.00009750, 0.00001731, -0.00002020, 0.00000629, 0.00016471, -0.00036342, 0.00038976, -0.00057713, 0.00063671, -0.00015192, -0.00010474, -0.00003255, 0.00014025, -0.00040436, 0.00044376, -0.00145951, 0.00218343, -0.00002717, -0.00069760, 0.00031962, -0.00044077, 0.00034831, -0.00019316, 0.00006463, -0.00009866, -0.00003381, -0.00001628, 0.00012761, -0.00021411, 0.00018858, -0.00007934, },
		{ 0.00001809, -0.00002511, -0.00008522, 0.00009315, -0.00014970, 0.00033552, -0.00046693, 0.00041304, -0.00052302, 0.00039956, 0.00020141, -0.00027558, -0.00003119, 0.00019104, -0.00044533, 0.00049497, -0.00124746, 0.00114342, 0.00120352, -0.00132252, 0.00062972, -0.00064492, 0.00042791, -0.00022674, 0.00013108, -0.00017129, 0.00002685, -0.00011635, 0.00021748, -0.00023006, 0.00013592, -0.00001494, },
		{ -0.00003428, 0.00004094, -0.00015736, 0.00014430, -0.00020237, 0.00034252, -0.00035868, 0.00020827, -0.00023033, -0.00005069, 0.00064470, -0.00044079, 0.00000645, 0.00011743, -0.00028507, 0.00024160, -0.00064667, -0.00006166, 0.00214772, -0.00148196, 0.00055642, -0.00054591, 0.00027234, -0.00009896, 0.00007246, -0.00012418, -0.00001107, -0.00012601, 0.00020051, -0.00014493, 0.00002150, 0.00007006, },
		{ -0.00005485, 0.00006877, -0.00016561, 0.00010372, -0.00013398, 0.00019413, -0.00008968, -0.00012872, 0.00015951, -0.00051851, 0.00096248, -0.00043415, -0.00003153, 0.00002925, -0.00006900, -0.00009724, -0.00001169, -0.00097516, 0.00243678, -0.00101011, 0.00006744, -0.00016268, -0.00007756, 0.00014994, -0.00008218, 0.00001274, -0.00011546, -0.00007163, 0.00010885, -0.00000803, -0.00009824, 0.00012870, },
		{ -0.00002923, 0.00005199, -0.00011246, -0.00000698, 0.00000654, -0.00002252, 0.00020765, -0.00042850, 0.00044696, -0.00078163, 0.00096075, -0.00014279, -0.00021184, 0.00001575, 0.00006360, -0.00031516, 0.00037010, -0.00132053, 0.00198118, -0.00002756, -0.00064059, 0.00031603, -0.00045533, 0.00038523, -0.00022681, 0.00014572, -0.00020548, -0.00001826, 0.00000707, 0.00011170, -0.00016583, 0.00012754, },
		{ 0.00002800, 0.00001159, -0.00003204, -0.00012964, 0.00013348, -0.00019083, 0.00038960, -0.00053746, 0.00048100, -0.00070965, 0.00058189, 0.00039928, -0.00048161, 0.00007160, 0.00006893, -0.00033002, 0.00040330, -0.00110203, 0.00100267, 0.00110028, -0.00122391, 0.00062213, -0.00065415, 0.00045358, -0.00024597, 0.00017878, -0.00020596, -0.00002335, -0.00004926, 0.00016343, -0.00015281, 0.00006548, },
		{ 0.00008194, -0.00001868, 0.00003209, -0.00020359, 0.00017316, -0.00022836, 0.00037674, -0.00039981, 0.00024179, -0.00033818, -0.00004636, 0.00099560, -0.00067159, 0.00009668, 0.00001109, -0.00020328, 0.00019024, -0.00056023, -0.00008288, 0.00192926, -0.00136043, 0.00055372, -0.00054581, 0.00027890, -0.00009301, 0.00007826, -0.00009843, -0.00009728, -0.00004559, 0.00013898, -0.00007290, -0.00002571, },
		{ 0.00009817, -0.00001436, 0.00005287, -0.00019881, 0.00010360, -0.00012923, 0.00019316, -0.00008671, -0.00014691, 0.00014489, -0.00066119, 0.00136933, -0.00058526, -0.00002703, -0.00000474, -0.00006810, -0.00006906, -0.00001918, -0.00086736, 0.00214328, -0.00091809, 0.00009570, -0.00016406, -0.00008053, 0.00017048, -0.00010057, 0.00006352, -0.00019560, -0.00001241, 0.00007136, 0.00002831, -0.00009914, },
		{ 0.00006349, 0.00002396, 0.00003413, -0.00012958, -0.00003648, 0.00003718, -0.00005114, 0.00024338, -0.00048344, 0.00049581, -0.00099239, 0.00130476, -0.00012644, -0.00033108, 0.00007221, -0.00002300, -0.00021195, 0.00028417, -0.00113110, 0.00169631, -0.00003237, -0.00055481, 0.00029962, -0.00045161, 0.00040102, -0.00024292, 0.00018490, -0.00024751, -0.00000030, 0.00000988, 0.00010028, -0.00011868, },
		{ -0.00000569, 0.00007068, 0.00000500, -0.00004171, -0.00017307, 0.00016903, -0.00022381, 0.00043051, -0.00059582, 0.00054246, -0.00090051, 0.00077172, 0.00061166, -0.00070064, 0.00018362, -0.00006432, -0.00020025, 0.00029694, -0.00091390, 0.00081941, 0.00095077, -0.00107793, 0.00059144, -0.00063644, 0.00045333, -0.00024376, 0.00018810, -0.00020285, -0.00004279, -0.00001177, 0.00011508, -0.00007817, },
		{ -0.00007222, 0.00009120, -0.00000134, 0.00001745, -0.00024032, 0.00019045, -0.00024273, 0.00039730, -0.00043201, 0.00027357, -0.00045017, -0.00003946, 0.00135698, -0.00090826, 0.00019114, -0.00010094, -0.00011397, 0.00013211, -0.00045124, -0.00010547, 0.00163833, -0.00119155, 0.00053257, -0.00052390, 0.00026830, -0.00007592, 0.00006392, -0.00006810, -0.00013100, 0.00000490, 0.00008005, -0.00000335, },
		},
		{ { 0.00021705, -0.00043308, 0.00022765, 0.00000689, 0.00003483, -0.00003770, 0.00004693, -0.00013104, -0.00009801, 0.00024335, -0.00085473, 0.00185216, -0.00071997, -0.00014347, -0.00009959, 0.00004553, -0.00001025, 0.00004553, -0.00009959, -0.00014347, -0.00071997, 0.00185216, -0.00085473, 0.00024335, -0.00009801, -0.00013104, 0.00004693, -0.00003770, 0.00003483, 0.00000689, 0.00022765, -0.00043308, },
		{ 0.00009432, -0.00021933, -0.00011417, 0.00028957, -0.00018403, 0.00019559, -0.00022929, 0.00025552, -0.00056121, 0.00071373, -0.00121186, 0.00150385, 0.00019261, -0.00069258, 0.00010922, -0.00010375, 0.00004886, 0.00006273, -0.00016203, 0.00012609, -0.00096422, 0.00132204, 0.00012164, -0.00052265, 0.00047088, -0.00050947, 0.00028351, -0.00023776, 0.00022091, -0.00020323, 0.00039345, -0.00040263, },
		{ -0.00017660, 0.00006508, -0.00040247, 0.00041541, -0.00022799, 0.00024961, -0.00029501, 0.00034592, -0.00058335, 0.00057762, -0.00077926, 0.00041161, 0.00137688, -0.00113405, 0.00022069, -0.00023661, 0.00013615, -0.00001104, -0.00006745, 0.00007918, -0.00065185, 0.00033431, 0.00111322, -0.00101608, 0.00069188, -0.00054654, 0.00024429, -0.00022027, 0.00020900, -0.00019162, 0.00029205, -0.00013641, },
		{ -0.00038181, 0.00020237, -0.00044666, 0.00026354, -0.00003334, 0.00007735, -0.00010323, 0.00010901, -0.00019978, 0.00000027, 0.00006152, -0.00078058, 0.00212264, -0.00097158, -0.00003994, -0.00016245, 0.00008899, -0.00002325, 0.00002410, -0.00006229, -0.00020107, -0.00048661, 0.00152292, -0.00084350, 0.00036064, -0.00016834, -0.00007464, -0.00000014, 0.00001882, -0.00000246, 0.00003307, 0.00018277, },
		{ -0.00036038, 0.00010120, -0.00024240, -0.00007374, 0.00025482, -0.00016026, 0.00016187, -0.00021324, 0.00024349, -0.00053742, 0.00067673, -0.00135817, 0.00189715, -0.00002278, -0.00066608, 0.00011272, -0.00012071, 0.00009027, -0.00000128, -0.00009496, 0.00004179, -0.00076102, 0.00117523, -0.00008767, -0.00029913, 0.00035674, -0.00042226, 0.00021584, -0.00017060, 0.00018054, -0.00017194, 0.00034200, },
		{ -0.00012752, -0.00013783, 0.00003173, -0.00036162, 0.00040006, -0.00023985, 0.00026128, -0.00033378, 0.00039434, -0.00063448, 0.00066002, -0.00104202, 0.00076381, 0.00132797, -0.00124779, 0.00031135, -0.00030963, 0.00020922, -0.00008084, -0.00001793, 0.00003093, -0.00055277, 0.00039262, 0.00075942, -0.00079885, 0.00062158, -0.00049600, 0.00021229, -0.00018798, 0.00020042, -0.00018188, 0.00026005, },
		{ 0.00014580, -0.00031206, 0.00016337, -0.00041061, 0.00027441, -0.00007734, 0.00011721, -0.00016790, 0.00017766, -0.00027266, 0.00011370, -0.00017091, -0.00061675, 0.00230355, -0.00121337, 0.00009800, -0.00024604, 0.00014975, -0.00006027, 0.00002611, -0.00005162, -0.00020768, -0.00029526, 0.00117027, -0.00076189, 0.00040859, -0.00020201, -0.00003664, -0.00002624, 0.00005791, -0.00002742, 0.00003668, },
		{ 0.00027369, -0.00028311, 0.00007539, -0.00022407, -0.00003264, 0.00019515, -0.00012444, 0.00011291, -0.00017424, 0.00020903, -0.00047752, 0.00057799, -0.00143196, 0.00225547, -0.00029184, -0.00057174, 0.00008485, -0.00011495, 0.00010740, -0.00004762, -0.00004309, -0.00001993, -0.00054966, 0.00095862, -0.00020977, -0.00012969, 0.00025231, -0.00033238, 0.00015210, -0.00010365, 0.00013594, -0.00014223, },
		{ 0.00019691, -0.00008149, -0.00012428, 0.00001708, -0.00029056, 0.00034577, -0.00023520, 0.00025348, -0.00034832, 0.00042183, -0.00065599, 0.00070388, -0.00128163, 0.00114968, 0.00117579, -0.00128858, 0.00038393, -0.00037191, 0.00027209, -0.00015049, 0.00003700, -0.00002503, -0.00041104, 0.00036917, 0.00046881, -0.00058894, 0.00051897, -0.00041666, 0.00016491, -0.00013780, 0.00017178, -0.00015683, },
		{ 0.00001243, 0.00013159, -0.00024692, 0.00011433, -0.00032244, 0.00024794, -0.00011478, 0.00014693, -0.00022184, 0.00024323, -0.00034049, 0.00022728, -0.00043250, -0.00037094, 0.00237294, -0.00141684, 0.00025403, -0.00034224, 0.00022038, -0.00011561, 0.00004948, -0.00006483, -0.00016526, -0.00016219, 0.00082930, -0.00063281, 0.00039173, -0.00019561, -0.00002017, -0.00002877, 0.00007390, -0.00003494, },
		{ -0.00012253, 0.00020387, -0.00018618, 0.00001639, -0.00014852, -0.00000983, 0.00012187, -0.00008305, 0.00005641, -0.00011800, 0.00015620, -0.00038850, 0.00042292, -0.00141864, 0.00254247, -0.00058678, -0.00041748, 0.00002526, -0.00008632, 0.00009811, -0.00007073, -0.00001165, -0.00005070, -0.00035420, 0.00070137, -0.00025004, -0.00002660, 0.00017227, -0.00025017, 0.00010063, -0.00004697, 0.00009542, },
		{ -0.00012135, 0.00011108, -0.00000037, -0.00015105, 0.00004441, -0.00020681, 0.00025750, -0.00021385, 0.00022710, -0.00033659, 0.00042426, -0.00064463, 0.00069999, -0.00147046, 0.00153401, 0.00093333, -0.00124789, 0.00042664, -0.00041548, 0.00031630, -0.00020993, 0.00008894, -0.00007782, -0.00024677, 0.00027629, 0.00025704, -0.00041165, 0.00040316, -0.00031912, 0.00010962, -0.00007791, 0.00012834, },
		{ -0.00002265, -0.00004076, 0.00015026, -0.00020938, 0.00007627, -0.00018980, 0.00017784, -0.00013704, 0.00016067, -0.00025669, 0.00029623, -0.00039446, 0.00032602, -0.00069702, -0.00006226, 0.00232085, -0.00155707, 0.00040930, -0.00044077, 0.00029185, -0.00018078, 0.00008907, -0.00009600, -0.00008240, -0.00009424, 0.00053024, -0.00048321, 0.00032278, -0.00015149, -0.00002500, -0.00000879, 0.00006475, },
		{ 0.00006687, -0.00012019, 0.00014807, -0.00009040, -0.00006623, -0.00000898, -0.00002237, 0.00004844, -0.00004392, 0.00000146, -0.00005293, 0.00009174, -0.00028098, 0.00022482, -0.00131446, 0.00272800, -0.00087577, -0.00021883, -0.00006175, -0.00003797, 0.00006418, -0.00006855, -0.00000269, -0.00004720, -0.00019424, 0.00043536, -0.00022403, 0.00000787, 0.00012667, -0.00018421, 0.00006767, -0.00000914, },
		{ 0.00007780, -0.00008236, 0.00001488, 0.00010578, -0.00022425, 0.00013307, -0.00013201, 0.00014616, -0.00017905, 0.00018621, -0.00030051, 0.00040096, -0.00060120, 0.00064564, -0.00158648, 0.00187931, 0.00062448, -0.00112699, 0.00043180, -0.00043477, 0.00033600, -0.00025065, 0.00013025, -0.00011791, -0.00008144, 0.00013477, 0.00012777, -0.00028520, 0.00029327, -0.00021584, 0.00005484, -0.00001798, },
		{ 0.00003231, 0.00000859, -0.00011942, 0.00020541, -0.00021834, 0.00007319, -0.00002953, 0.00006517, -0.00013858, 0.00015516, -0.00026693, 0.00032883, -0.00042738, 0.00039712, -0.00093691, 0.00028089, 0.00215077, -0.00161630, 0.00054513, -0.00053054, 0.00035481, -0.00024575, 0.00013746, -0.00013703, 0.00002705, -0.00008822, 0.00029465, -0.00033968, 0.00022036, -0.00007737, -0.00004773, 0.00002916, },
		{ 0.00000414, 0.00005646, -0.00014020, 0.00011946, -0.00001866, -0.00015430, 0.00019058, -0.00008229, -0.00001169, -0.00001490, -0.00004277, 0.00001107, 0.00002406, -0.00016770, 0.00000318, -0.00112729, 0.00279217, -0.00112729, 0.00000318, -0.00016770, 0.00002406, 0.00001107, -0.00004277, -0.00001490, -0.00001169, -0.00008229, 0.00019058, -0.00015430, -0.00001866, 0.00011946, -0.00014020, 0.00005646, },
		{ 0.00003231, 0.00002916, -0.00004773, -0.00007737, 0.00022036, -0.00033968, 0.00029465, -0.00008822, 0.00002705, -0.00013703, 0.00013746, -0.00024575, 0.00035481, -0.00053054, 0.00054513, -0.00161630, 0.00215077, 0.00028089, -0.00093691, 0.00039712, -0.00042738, 0.00032883, -0.00026693, 0.00015516, -0.00013858, 0.00006517, -0.00002953, 0.00007319, -0.00021834, 0.00020541, -0.00011942, 0.00000859, },
		{ 0.00007780, -0.00001798, 0.00005484, -0.00021584, 0.00029327, -0.00028520, 0.00012777, 0.00013477, -0.00008144, -0.00011791, 0.00013025, -0.00025065, 0.00033600, -0.00043477, 0.00043180, -0.00112699, 0.00062448, 0.00187931, -0.00158648, 0.00064564, -0.00060120, 0.00040096, -0.00030051, 0.00018621, -0.00017905, 0.00014616, -0.00013201, 0.00013307, -0.00022425, 0.00010578, 0.00001488, -0.00008236, },
		{ 0.00006687, -0.00000914, 0.00006767, -0.00018421, 0.00012667, 0.00000787, -0.00022403, 0.00043536, -0.00019424, -0.00004720, -0.00000269, -0.00006855, 0.00006418, -0.00003797, -0.00006175, -0.00021883, -0.00087577, 0.00272800, -0.00131446, 0.00022482, -0.00028098, 0.00009174, -0.00005293, 0.00000146, -0.00004392, 0.00004844, -0.00002237, -0.00000898, -0.00006623, -0.00009040, 0.00014807, -0.00012019, },
		{ -0.00002265, 0.00006475, -0.00000879, -0.00002500, -0.00015149, 0.00032278, -0.00048321, 0.00053024, -0.00009424, -0.00008240, -0.00009600, 0.00008907, -0.00018078, 0.00029185, -0.00044077, 0.00040930, -0.00155707, 0.00232085, -0.00006226, -0.00069702, 0.00032602, -0.00039446, 0.00029623, -0.00025669, 0.00016067, -0.00013704, 0.00017784, -0.00018980, 0.00007627, -0.00020938, 0.00015026, -0.00004076, },
		{ -0.00012135, 0.00012834, -0.00007791, 0.00010962, -0.00031912, 0.00040316, -0.00041165, 0.00025704, 0.00027629, -0.00024677, -0.00007782, 0.00008894, -0.00020993, 0.00031630, -0.00041548, 0.00042664, -0.00124789, 0.00093333, 0.00153401, -0.00147046, 0.00069999, -0.00064463, 0.00042426, -0.00033659, 0.00022710, -0.00021385, 0.00025750, -0.00020681, 0.00004441, -0.00015105, -0.00000037, 0.00011108, },
		{ -0.00012253, 0.00009542, -0.00004697, 0.00010063, -0.00025017, 0.00017227, -0.00002660, -0.00025004, 0.00070137, -0.00035420, -0.00005070, -0.00001165, -0.00007073, 0.00009811, -0.00008632, 0.00002526, -0.00041748, -0.00058678, 0.00254247, -0.00141864, 0.00042292, -0.00038850, 0.00015620, -0.00011800, 0.00005641, -0.00008305, 0.00012187, -0.00000983, -0.00014852, 0.00001639, -0.00018618, 0.00020387, },
		{ 0.00001243, -0.00003494, 0.00007390, -0.00002877, -0.00002017, -0.00019561, 0.00039173, -0.00063281, 0.00082930, -0.00016219, -0.00016526, -0.00006483, 0.00004948, -0.00011561, 0.00022038, -0.00034224, 0.00025403, -0.00141684, 0.00237294, -0.00037094, -0.00043250, 0.00022728, -0.00034049, 0.00024323, -0.00022184, 0.00014693, -0.00011478, 0.00024794, -0.00032244, 0.00011433, -0.00024692, 0.00013159, },
		{ 0.00019691, -0.00015683, 0.00017178, -0.00013780, 0.00016491, -0.00041666, 0.00051897, -0.00058894, 0.00046881, 0.00036917, -0.00041104, -0.00002503, 0.00003700, -0.00015049, 0.00027209, -0.00037191, 0.00038393, -0.00128858, 0.00117579, 0.00114968, -0.00128163, 0.00070388, -0.00065599, 0.00042183, -0.00034832, 0.00025348, -0.00023520, 0.00034577, -0.00029056, 0.00001708, -0.00012428, -0.00008149, },
		{ 0.00027369, -0.00014223, 0.00013594, -0.00010365, 0.00015210, -0.00033238, 0.00025231, -0.00012969, -0.00020977, 0.00095862, -0.00054966, -0.00001993, -0.00004309, -0.00004762, 0.00010740, -0.00011495, 0.00008485, -0.00057174, -0.00029184, 0.00225547, -0.00143196, 0.00057799, -0.00047752, 0.00020903, -0.00017424, 0.00011291, -0.00012444, 0.00019515, -0.00003264, -0.00022407, 0.00007539, -0.00028311, },
		{ 0.00014580, 0.00003668, -0.00002742, 0.00005791, -0.00002624, -0.00003664, -0.00020201, 0.00040859, -0.00076189, 0.00117027, -0.00029526, -0.00020768, -0.00005162, 0.00002611, -0.00006027, 0.00014975, -0.00024604, 0.00009800, -0.00121337, 0.00230355, -0.00061675, -0.00017091, 0.00011370, -0.00027266, 0.00017766, -0.00016790, 0.00011721, -0.00007734, 0.00027441, -0.00041061, 0.00016337, -0.00031206, },
		{ -0.00012752, 0.00026005, -0.00018188, 0.00020042, -0.00018798, 0.00021229, -0.00049600, 0.00062158, -0.00079885, 0.00075942, 0.00039262, -0.00055277, 0.00003093, -0.00001793, -0.00008084, 0.00020922, -0.00030963, 0.00031135, -0.00124779, 0.00132797, 0.00076381, -0.00104202, 0.00066002, -0.00063448, 0.00039434, -0.00033378, 0.00026128, -0.00023985, 0.00040006, -0.00036162, 0.00003173, -0.00013783, },
		{ -0.00036038, 0.00034200, -0.00017194, 0.00018054, -0.00017060, 0.00021584, -0.00042226, 0.00035674, -0.00029913, -0.00008767, 0.00117523, -0.00076102, 0.00004179, -0.00009496, -0.00000128, 0.00009027, -0.00012071, 0.00011272, -0.00066608, -0.00002278, 0.00189715, -0.00135817, 0.00067673, -0.00053742, 0.00024349, -0.00021324, 0.00016187, -0.00016026, 0.00025482, -0.00007374, -0.00024240, 0.00010120, },
		{ -0.00038181, 0.00018277, 0.00003307, -0.00000246, 0.00001882, -0.00000014, -0.00007464, -0.00016834, 0.00036064, -0.00084350, 0.00152292, -0.00048661, -0.00020107, -0.00006229, 0.00002410, -0.00002325, 0.00008899, -0.00016245, -0.00003994, -0.00097158, 0.00212264, -0.00078058, 0.00006152, 0.00000027, -0.00019978, 0.00010901, -0.00010323, 0.00007735, -0.00003334, 0.00026354, -0.00044666, 0.00020237, },
		{ -0.00017660, -0.00013641, 0.00029205, -0.00019162, 0.00020900, -0.00022027, 0.00024429, -0.00054654, 0.00069188, -0.00101608, 0.00111322, 0.00033431, -0.00065185, 0.00007918, -0.00006745, -0.00001104, 0.00013615, -0.00023661, 0.00022069, -0.00113405, 0.00137688, 0.00041161, -0.00077926, 0.00057762, -0.00058335, 0.00034592, -0.00029501, 0.00024961, -0.00022799, 0.00041541, -0.00040247, 0.00006508, },
		{ 0.00009432, -0.00040263, 0.00039345, -0.00020323, 0.00022091, -0.00023776, 0.00028351, -0.00050947, 0.00047088, -0.00052265, 0.00012164, 0.00132204, -0.00096422, 0.00012609, -0.00016203, 0.00006273, 0.00004886, -0.00010375, 0.00010922, -0.00069258, 0.00019261, 0.00150385, -0.00121186, 0.00071373, -0.00056121, 0.00025552, -0.00022929, 0.00019559, -0.00018403, 0.00028957, -0.00011417, -0.00021933, },
		},
		{ { -0.00015699, 0.00031144, -0.00045480, 0.00007252, -0.00000646, 0.00016804, -0.00020457, 0.00001333, 0.00007463, -0.00088941, 0.00189741, -0.00079230, 0.00001185, -0.00012801, 0.00015382, -0.00011511, 0.00001852, -0.00011511, 0.00015382, -0.00012801, 0.00001185, -0.00079230, 0.00189741, -0.00088941, 0.00007463, 0.00001333, -0.00020457, 0.00016804, -0.00000646, 0.00007252, -0.00045480, 0.00031144, },
		{ -0.00001715, 0.00014206, -0.00007193, -0.00035450, 0.00029376, -0.00016803, 0.00028977, -0.00057870, 0.00073032, -0.00136087, 0.00143935, 0.00030513, -0.00067014, 0.00017434, -0.00009103, 0.00007683, -0.00006950, -0.00010068, 0.00014176, -0.00014936, 0.00022678, -0.00090078, 0.00108822, 0.00044215, -0.00090029, 0.00063097, -0.00058466, 0.00036538, -0.00019201, 0.00030877, -0.00051307, 0.00015342, },
		{ 0.00023104, -0.00010534, 0.00026808, -0.00055154, 0.00029892, -0.00019719, 0.00035969, -0.00053458, 0.00048380, -0.00066235, -0.00003507, 0.00169604, -0.00113872, 0.00029462, -0.00025983, 0.00022206, -0.00016660, 0.00001525, -0.00004575, 0.00005959, 0.00002961, -0.00035171, -0.00014030, 0.00150419, -0.00119528, 0.00055184, -0.00037076, 0.00011104, -0.00002626, 0.00017708, -0.00020252, -0.00023271, },
		{ 0.00027213, -0.00013046, 0.00027321, -0.00031982, -0.00007443, 0.00011537, 0.00001268, 0.00000956, -0.00027521, 0.00040339, -0.00130385, 0.00214977, -0.00058648, -0.00020876, -0.00003424, 0.00005125, -0.00003839, -0.00001483, -0.00011115, 0.00015077, -0.00011523, 0.00009620, -0.00078328, 0.00146536, -0.00043398, -0.00023824, 0.00026669, -0.00036313, 0.00026666, -0.00007967, 0.00015613, -0.00049171, },
		{ 0.00001047, 0.00007954, 0.00002522, 0.00010185, -0.00046888, 0.00036465, -0.00026582, 0.00040437, -0.00068655, 0.00077860, -0.00133878, 0.00111992, 0.00092149, -0.00103716, 0.00035020, -0.00027222, 0.00020945, -0.00016755, 0.00000052, 0.00001471, -0.00000703, 0.00011364, -0.00060736, 0.00053826, 0.00074250, -0.00097450, 0.00064197, -0.00051421, 0.00027669, -0.00011219, 0.00023907, -0.00039454, },
		{ -0.00032110, 0.00025636, -0.00015016, 0.00031079, -0.00048669, 0.00020476, -0.00012573, 0.00026051, -0.00037336, 0.00022590, -0.00030575, -0.00060025, 0.00223769, -0.00122580, 0.00025175, -0.00027948, 0.00021045, -0.00016220, 0.00004891, -0.00011036, 0.00012744, -0.00003589, -0.00013963, -0.00037263, 0.00133655, -0.00089813, 0.00029996, -0.00012028, -0.00007894, 0.00011120, 0.00006364, -0.00007323, },
		{ -0.00042345, 0.00017432, -0.00006601, 0.00017349, -0.00012889, -0.00023422, 0.00024011, -0.00015074, 0.00022075, -0.00052742, 0.00066538, -0.00157715, 0.00213921, -0.00014643, -0.00055611, 0.00015530, -0.00013604, 0.00009736, -0.00010143, -0.00003501, 0.00005825, -0.00002481, 0.00007795, -0.00062672, 0.00096866, -0.00004214, -0.00046112, 0.00042723, -0.00042441, 0.00027687, -0.00007656, 0.00015012, },
		{ -0.00024978, -0.00009828, 0.00014077, -0.00007507, 0.00022968, -0.00049913, 0.00035777, -0.00028849, 0.00042198, -0.00066365, 0.00067551, -0.00112175, 0.00058703, 0.00161158, -0.00136148, 0.00048580, -0.00041586, 0.00029583, -0.00023363, 0.00009105, -0.00010180, 0.00010171, 0.00002179, -0.00034633, 0.00013773, 0.00082780, -0.00088704, 0.00053160, -0.00034863, 0.00012256, 0.00001210, 0.00012495, },
		{ -0.00001888, -0.00030036, 0.00020965, -0.00014231, 0.00027110, -0.00033312, 0.00005328, -0.00000355, 0.00009765, -0.00014259, -0.00008349, 0.00008983, -0.00113674, 0.00258862, -0.00108920, 0.00005883, -0.00017713, 0.00009055, -0.00007665, 0.00001479, -0.00009307, 0.00009274, -0.00001154, -0.00004611, -0.00040875, 0.00101183, -0.00055662, 0.00004848, 0.00010634, -0.00022396, 0.00019507, 0.00000162, },
		{ 0.00006259, -0.00028280, 0.00006900, -0.00000231, 0.00005297, 0.00005776, -0.00035076, 0.00032104, -0.00027308, 0.00037195, -0.00067964, 0.00079462, -0.00164241, 0.00183987, 0.00048608, -0.00097354, 0.00039088, -0.00035504, 0.00024639, -0.00020575, 0.00007900, -0.00007795, 0.00008739, 0.00001650, -0.00041209, 0.00052106, 0.00020672, -0.00055052, 0.00046793, -0.00037644, 0.00019516, 0.00000174, },
		{ 0.00001948, -0.00014474, -0.00011451, 0.00013492, -0.00012704, 0.00027470, -0.00042921, 0.00026984, -0.00023204, 0.00033970, -0.00051601, 0.00044293, -0.00075416, -0.00006334, 0.00224467, -0.00154323, 0.00051011, -0.00046034, 0.00029063, -0.00023132, 0.00013082, -0.00015824, 0.00012763, -0.00000415, -0.00017347, -0.00007454, 0.00071797, -0.00068465, 0.00034851, -0.00013967, -0.00004710, 0.00013216, },
		{ 0.00002035, -0.00006411, -0.00017267, 0.00011875, -0.00010299, 0.00017284, -0.00013711, -0.00010669, 0.00012558, -0.00008055, 0.00009671, -0.00036876, 0.00043349, -0.00153189, 0.00265815, -0.00070479, -0.00027965, 0.00003794, -0.00011781, 0.00007154, -0.00007686, -0.00000101, -0.00002688, 0.00007515, -0.00004173, -0.00031126, 0.00063399, -0.00026229, -0.00013520, 0.00025424, -0.00028233, 0.00019216, },
		{ 0.00012440, -0.00006595, -0.00013310, 0.00002264, 0.00001613, -0.00004273, 0.00018450, -0.00038289, 0.00032746, -0.00032093, 0.00042519, -0.00069547, 0.00076200, -0.00148517, 0.00129325, 0.00121272, -0.00136933, 0.00059818, -0.00053663, 0.00035354, -0.00028280, 0.00018392, -0.00019945, 0.00015990, -0.00002737, -0.00022146, 0.00020476, 0.00027952, -0.00050548, 0.00039915, -0.00024038, 0.00004819, },
		{ 0.00019722, -0.00002550, -0.00013386, -0.00000755, 0.00006085, -0.00012685, 0.00023460, -0.00027919, 0.00012743, -0.00012000, 0.00018482, -0.00028448, 0.00013972, -0.00031920, -0.00070490, 0.00268958, -0.00150429, 0.00037750, -0.00037168, 0.00017573, -0.00014677, 0.00009991, -0.00012949, 0.00005415, 0.00004456, -0.00010107, -0.00011565, 0.00048117, -0.00044253, 0.00015820, 0.00005123, -0.00017736, },
		{ 0.00009740, 0.00011856, -0.00019573, 0.00001766, 0.00004242, -0.00007547, 0.00006197, 0.00004084, -0.00022228, 0.00021612, -0.00022339, 0.00028292, -0.00056071, 0.00065004, -0.00170772, 0.00241566, -0.00011036, -0.00071117, 0.00031662, -0.00036303, 0.00023936, -0.00019250, 0.00013146, -0.00018410, 0.00017144, -0.00007261, -0.00016313, 0.00030423, -0.00008375, -0.00021077, 0.00029643, -0.00024011, },
		{ -0.00011532, 0.00024055, -0.00017618, -0.00004754, 0.00009281, -0.00003917, -0.00008420, 0.00022068, -0.00031951, 0.00025590, -0.00028721, 0.00037293, -0.00057686, 0.00058393, -0.00114675, 0.00059907, 0.00190057, -0.00164144, 0.00070176, -0.00061746, 0.00037269, -0.00029442, 0.00023722, -0.00025592, 0.00014824, -0.00001404, -0.00010600, 0.00005185, 0.00019579, -0.00036552, 0.00026314, -0.00006326, },
		{ -0.00022604, 0.00017231, 0.00002136, -0.00023965, 0.00021007, -0.00004495, -0.00010199, 0.00014077, -0.00009801, -0.00002247, 0.00000566, 0.00000521, -0.00003204, -0.00015621, 0.00008576, -0.00121450, 0.00284971, -0.00121450, 0.00008576, -0.00015621, -0.00003204, 0.00000521, 0.00000566, -0.00002247, -0.00009801, 0.00014077, -0.00010199, -0.00004495, 0.00021007, -0.00023965, 0.00002136, 0.00017231, },
		{ -0.00011532, -0.00006326, 0.00026314, -0.00036552, 0.00019579, 0.00005185, -0.00010600, -0.00001404, 0.00014824, -0.00025592, 0.00023722, -0.00029442, 0.00037269, -0.00061746, 0.00070176, -0.00164144, 0.00190057, 0.00059907, -0.00114675, 0.00058393, -0.00057686, 0.00037293, -0.00028721, 0.00025590, -0.00031951, 0.00022068, -0.00008420, -0.00003917, 0.00009281, -0.00004754, -0.00017618, 0.00024055, },
		{ 0.00009740, -0.00024011, 0.00029643, -0.00021077, -0.00008375, 0.00030423, -0.00016313, -0.00007261, 0.00017144, -0.00018410, 0.00013146, -0.00019250, 0.00023936, -0.00036303, 0.00031662, -0.00071117, -0.00011036, 0.00241566, -0.00170772, 0.00065004, -0.00056071, 0.00028292, -0.00022339, 0.00021612, -0.00022228, 0.00004084, 0.00006197, -0.00007547, 0.00004242, 0.00001766, -0.00019573, 0.00011856, },
		{ 0.00019722, -0.00017736, 0.00005123, 0.00015820, -0.00044253, 0.00048117, -0.00011565, -0.00010107, 0.00004456, 0.00005415, -0.00012949, 0.00009991, -0.00014677, 0.00017573, -0.00037168, 0.00037750, -0.00150429, 0.00268958, -0.00070490, -0.00031920, 0.00013972, -0.00028448, 0.00018482, -0.00012000, 0.00012743, -0.00027919, 0.00023460, -0.00012685, 0.00006085, -0.00000755, -0.00013386, -0.00002550, },
		{ 0.00012440, 0.00004819, -0.00024038, 0.00039915, -0.00050548, 0.00027952, 0.00020476, -0.00022146, -0.00002737, 0.00015990, -0.00019945, 0.00018392, -0.00028280, 0.00035354, -0.00053663, 0.00059818, -0.00136933, 0.00121272, 0.00129325, -0.00148517, 0.00076200, -0.00069547, 0.00042519, -0.00032093, 0.00032746, -0.00038289, 0.00018450, -0.00004273, 0.00001613, 0.00002264, -0.00013310, -0.00006595, },
		{ 0.00002035, 0.00019216, -0.00028233, 0.00025424, -0.00013520, -0.00026229, 0.00063399, -0.00031126, -0.00004173, 0.00007515, -0.00002688, -0.00000101, -0.00007686, 0.00007154, -0.00011781, 0.00003794, -0.00027965, -0.00070479, 0.00265815, -0.00153189, 0.00043349, -0.00036876, 0.00009671, -0.00008055, 0.00012558, -0.00010669, -0.00013711, 0.00017284, -0.00010299, 0.00011875, -0.00017267, -0.00006411, },
		{ 0.00001948, 0.00013216, -0.00004710, -0.00013967, 0.00034851, -0.00068465, 0.00071797, -0.00007454, -0.00017347, -0.00000415, 0.00012763, -0.00015824, 0.00013082, -0.00023132, 0.00029063, -0.00046034, 0.00051011, -0.00154323, 0.00224467, -0.00006334, -0.00075416, 0.00044293, -0.00051601, 0.00033970, -0.00023204, 0.00026984, -0.00042921, 0.00027470, -0.00012704, 0.00013492, -0.00011451, -0.00014474, },
		{ 0.00006259, 0.00000174, 0.00019516, -0.00037644, 0.00046793, -0.00055052, 0.00020672, 0.00052106, -0.00041209, 0.00001650, 0.00008739, -0.00007795, 0.00007900, -0.00020575, 0.00024639, -0.00035504, 0.00039088, -0.00097354, 0.00048608, 0.00183987, -0.00164241, 0.00079462, -0.00067964, 0.00037195, -0.00027308, 0.00032104, -0.00035076, 0.00005776, 0.00005297, -0.00000231, 0.00006900, -0.00028280, },
		{ -0.00001888, 0.00000162, 0.00019507, -0.00022396, 0.00010634, 0.00004848, -0.00055662, 0.00101183, -0.00040875, -0.00004611, -0.00001154, 0.00009274, -0.00009307, 0.00001479, -0.00007665, 0.00009055, -0.00017713, 0.00005883, -0.00108920, 0.00258862, -0.00113674, 0.00008983, -0.00008349, -0.00014259, 0.00009765, -0.00000355, 0.00005328, -0.00033312, 0.00027110, -0.00014231, 0.00020965, -0.00030036, },
		{ -0.00024978, 0.00012495, 0.00001210, 0.00012256, -0.00034863, 0.00053160, -0.00088704, 0.00082780, 0.00013773, -0.00034633, 0.00002179, 0.00010171, -0.00010180, 0.00009105, -0.00023363, 0.00029583, -0.00041586, 0.00048580, -0.00136148, 0.00161158, 0.00058703, -0.00112175, 0.00067551, -0.00066365, 0.00042198, -0.00028849, 0.00035777, -0.00049913, 0.00022968, -0.00007507, 0.00014077, -0.00009828, },
		{ -0.00042345, 0.00015012, -0.00007656, 0.00027687, -0.00042441, 0.00042723, -0.00046112, -0.00004214, 0.00096866, -0.00062672, 0.00007795, -0.00002481, 0.00005825, -0.00003501, -0.00010143, 0.00009736, -0.00013604, 0.00015530, -0.00055611, -0.00014643, 0.00213921, -0.00157715, 0.00066538, -0.00052742, 0.00022075, -0.00015074, 0.00024011, -0.00023422, -0.00012889, 0.00017349, -0.00006601, 0.00017432, },
		{ -0.00032110, -0.00007323, 0.00006364, 0.00011120, -0.00007894, -0.00012028, 0.00029996, -0.00089813, 0.00133655, -0.00037263, -0.00013963, -0.00003589, 0.00012744, -0.00011036, 0.00004891, -0.00016220, 0.00021045, -0.00027948, 0.00025175, -0.00122580, 0.00223769, -0.00060025, -0.00030575, 0.00022590, -0.00037336, 0.00026051, -0.00012573, 0.00020476, -0.00048669, 0.00031079, -0.00015016, 0.00025636, },
		{ 0.00001047, -0.00039454, 0.00023907, -0.00011219, 0.00027669, -0.00051421, 0.00064197, -0.00097450, 0.00074250, 0.00053826, -0.00060736, 0.00011364, -0.00000703, 0.00001471, 0.00000052, -0.00016755, 0.00020945, -0.00027222, 0.00035020, -0.00103716, 0.00092149, 0.00111992, -0.00133878, 0.00077860, -0.00068655, 0.00040437, -0.00026582, 0.00036465, -0.00046888, 0.00010185, 0.00002522, 0.00007954, },
		{ 0.00027213, -0.00049171, 0.00015613, -0.00007967, 0.00026666, -0.00036313, 0.00026669, -0.00023824, -0.00043398, 0.00146536, -0.00078328, 0.00009620, -0.00011523, 0.00015077, -0.00011115, -0.00001483, -0.00003839, 0.00005125, -0.00003424, -0.00020876, -0.00058648, 0.00214977, -0.00130385, 0.00040339, -0.00027521, 0.00000956, 0.00001268, 0.00011537, -0.00007443, -0.00031982, 0.00027321, -0.00013046, },
		{ 0.00023104, -0.00023271, -0.00020252, 0.00017708, -0.00002626, 0.00011104, -0.00037076, 0.00055184, -0.00119528, 0.00150419, -0.00014030, -0.00035171, 0.00002961, 0.00005959, -0.00004575, 0.00001525, -0.00016660, 0.00022206, -0.00025983, 0.00029462, -0.00113872, 0.00169604, -0.00003507, -0.00066235, 0.00048380, -0.00053458, 0.00035969, -0.00019719, 0.00029892, -0.00055154, 0.00026808, -0.00010534, },
		{ -0.00001715, 0.00015342, -0.00051307, 0.00030877, -0.00019201, 0.00036538, -0.00058466, 0.00063097, -0.00090029, 0.00044215, 0.00108822, -0.00090078, 0.00022678, -0.00014936, 0.00014176, -0.00010068, -0.00006950, 0.00007683, -0.00009103, 0.00017434, -0.00067014, 0.00030513, 0.00143935, -0.00136087, 0.00073032, -0.00057870, 0.00028977, -0.00016803, 0.00029376, -0.00035450, -0.00007193, 0.00014206, },
		},
		{ { 0.00010196, -0.00017792, 0.00019320, -0.00011880, 0.00011561, -0.00039314, 0.00011897, -0.00003944, -0.00070222, 0.00181631, -0.00080471, 0.00002116, -0.00011022, 0.00021553, -0.00018720, 0.00007893, -0.00006779, 0.00007893, -0.00018720, 0.00021553, -0.00011022, 0.00002116, -0.00080471, 0.00181631, -0.00070222, -0.00003944, 0.00011897, -0.00039314, 0.00011561, -0.00011880, 0.00019320, -0.00017792, },
		{ -0.00001039, -0.00003325, -0.00003694, 0.00014702, -0.00016137, 0.00009404, -0.00055072, 0.00068596, -0.00124445, 0.00130422, 0.00043899, -0.00073396, 0.00018981, -0.00004395, 0.00008890, -0.00012823, 0.00003108, 0.00004670, -0.00014326, 0.00012377, -0.00007661, 0.00020692, -0.00080223, 0.00076878, 0.00080043, -0.00100294, 0.00065358, -0.00059167, 0.00013778, -0.00017610, 0.00018970, -0.00008529, },
		{ -0.00017325, 0.00010033, -0.00017319, 0.00022548, -0.00018961, 0.00017971, -0.00050841, 0.00034037, -0.00039139, -0.00035703, 0.00185810, -0.00104404, 0.00016559, -0.00017391, 0.00025780, -0.00022970, 0.00009571, -0.00006616, 0.00006168, -0.00014284, 0.00016427, -0.00004218, -0.00012546, -0.00051590, 0.00162608, -0.00088972, 0.00023539, -0.00008845, -0.00026414, 0.00004718, -0.00004414, 0.00014816, },
		{ -0.00012412, 0.00001252, -0.00005477, 0.00001719, 0.00008291, -0.00012089, 0.00001867, -0.00043905, 0.00060739, -0.00137002, 0.00178361, 0.00002836, -0.00058948, 0.00013588, 0.00000893, 0.00003438, -0.00010202, 0.00001391, 0.00006060, -0.00014842, 0.00013810, -0.00008677, 0.00018807, -0.00079013, 0.00098443, 0.00033400, -0.00069717, 0.00053211, -0.00056190, 0.00014669, -0.00016552, 0.00020885, },
		{ 0.00009709, -0.00016058, 0.00009550, -0.00015854, 0.00023941, -0.00024689, 0.00022948, -0.00059063, 0.00053820, -0.00077608, 0.00012361, 0.00173184, -0.00119929, 0.00028670, -0.00022330, 0.00028280, -0.00026372, 0.00011016, -0.00006133, 0.00004533, -0.00010267, 0.00011293, 0.00002006, -0.00025375, -0.00021775, 0.00132557, -0.00092130, 0.00040452, -0.00024849, -0.00014195, -0.00001544, 0.00002485, },
		{ 0.00020577, -0.00015120, 0.00003427, -0.00007368, 0.00007222, 0.00000265, -0.00005793, -0.00008007, -0.00026801, 0.00041308, -0.00134112, 0.00217548, -0.00041707, -0.00038428, 0.00004975, 0.00007640, -0.00003568, -0.00006492, -0.00000447, 0.00007267, -0.00014631, 0.00013417, -0.00007576, 0.00013933, -0.00070516, 0.00106985, -0.00004346, -0.00038358, 0.00035648, -0.00047246, 0.00012300, -0.00013366, },
		{ 0.00008017, 0.00004677, -0.00014195, 0.00008766, -0.00013442, 0.00023137, -0.00028228, 0.00025664, -0.00062465, 0.00067663, -0.00113974, 0.00069568, 0.00143993, -0.00124653, 0.00036601, -0.00025012, 0.00028421, -0.00028311, 0.00012016, -0.00005349, 0.00003060, -0.00007132, 0.00006966, 0.00006774, -0.00034679, 0.00005006, 0.00096534, -0.00082129, 0.00045901, -0.00033708, -0.00004509, -0.00006347, },
		{ -0.00008738, 0.00018472, -0.00016606, 0.00005385, -0.00008887, 0.00012281, -0.00008586, 0.00002088, -0.00019000, -0.00005873, 0.00011670, -0.00113905, 0.00242463, -0.00084483, -0.00014301, -0.00005902, 0.00015020, -0.00011442, -0.00001893, -0.00002354, 0.00008283, -0.00014026, 0.00011692, -0.00004822, 0.00007097, -0.00056516, 0.00102797, -0.00029759, -0.00011436, 0.00016880, -0.00034355, 0.00007385, },
		{ -0.00009081, 0.00011646, 0.00000313, -0.00011969, 0.00007710, -0.00010230, 0.00020034, -0.00028992, 0.00025589, -0.00060195, 0.00072724, -0.00142720, 0.00129895, 0.00100710, -0.00117622, 0.00039043, -0.00024906, 0.00025875, -0.00028316, 0.00012372, -0.00004293, 0.00001754, -0.00005133, 0.00004061, 0.00009477, -0.00039377, 0.00025411, 0.00060041, -0.00063179, 0.00041028, -0.00034307, 0.00001269, },
		{ 0.00001014, -0.00003531, 0.00015220, -0.00016992, 0.00007046, -0.00009947, 0.00016376, -0.00017322, 0.00010646, -0.00029674, 0.00016086, -0.00025110, -0.00076859, 0.00249200, -0.00120346, 0.00010401, -0.00017724, 0.00022067, -0.00019314, 0.00003242, -0.00004260, 0.00009099, -0.00013372, 0.00009304, -0.00001126, -0.00000423, -0.00039538, 0.00087964, -0.00041568, 0.00007064, 0.00001129, -0.00020120, },
		{ 0.00002449, -0.00009491, 0.00013172, -0.00002969, -0.00009613, 0.00006424, -0.00006450, 0.00014815, -0.00026744, 0.00022577, -0.00052200, 0.00067359, -0.00159028, 0.00186512, 0.00047739, -0.00099473, 0.00035435, -0.00021868, 0.00020680, -0.00026138, 0.00011922, -0.00003007, 0.00000555, -0.00004284, 0.00002897, 0.00009861, -0.00039152, 0.00037209, 0.00028193, -0.00040520, 0.00028760, -0.00026989, },
		{ -0.00007303, -0.00005535, 0.00001358, 0.00011578, -0.00016532, 0.00008361, -0.00010494, 0.00019082, -0.00024974, 0.00018854, -0.00038584, 0.00036017, -0.00064684, -0.00025834, 0.00236063, -0.00144981, 0.00032512, -0.00029004, 0.00027813, -0.00026248, 0.00008446, -0.00006077, 0.00009696, -0.00012965, 0.00006977, 0.00002674, -0.00007303, -0.00022451, 0.00065993, -0.00040755, 0.00014995, -0.00008062, },
		{ -0.00013485, -0.00000877, -0.00007710, 0.00012744, -0.00004979, -0.00007327, 0.00004956, -0.00002394, 0.00007929, -0.00021633, 0.00016892, -0.00039256, 0.00051407, -0.00159532, 0.00232732, -0.00009172, -0.00072352, 0.00026055, -0.00016164, 0.00013242, -0.00021798, 0.00010575, -0.00001540, -0.00000637, -0.00004367, 0.00003458, 0.00008062, -0.00034479, 0.00039579, 0.00004955, -0.00019532, 0.00013282, },
		{ -0.00008415, 0.00001616, -0.00010995, 0.00005137, 0.00008288, -0.00015566, 0.00009307, -0.00010509, 0.00020117, -0.00030679, 0.00025714, -0.00044490, 0.00051048, -0.00102026, 0.00034258, 0.00203853, -0.00155516, 0.00049194, -0.00038293, 0.00031431, -0.00031377, 0.00013195, -0.00007708, 0.00010049, -0.00012998, 0.00005368, 0.00005747, -0.00012381, -0.00008017, 0.00041230, -0.00030301, 0.00012349, },
		{ -0.00000693, 0.00003375, -0.00007858, -0.00004234, 0.00010828, -0.00005773, -0.00005251, 0.00003364, 0.00001620, 0.00000036, -0.00014177, 0.00009179, -0.00022864, 0.00026254, -0.00142861, 0.00262954, -0.00063717, -0.00039614, 0.00011985, -0.00008440, 0.00004296, -0.00015595, 0.00008329, 0.00000052, -0.00001939, -0.00004984, 0.00005400, 0.00004574, -0.00026541, 0.00033202, -0.00007429, -0.00004848, },
		{ 0.00001211, 0.00000703, 0.00004837, -0.00014316, 0.00007239, 0.00005956, -0.00014457, 0.00009888, -0.00010010, 0.00019383, -0.00033803, 0.00030397, -0.00046538, 0.00058923, -0.00132169, 0.00097094, 0.00155791, -0.00150926, 0.00058350, -0.00044394, 0.00032362, -0.00034023, 0.00016982, -0.00009051, 0.00010125, -0.00013526, 0.00004955, 0.00007410, -0.00014835, 0.00001532, 0.00018162, -0.00014622, },
		{ 0.00000414, -0.00008728, 0.00020119, -0.00017026, 0.00000168, 0.00008113, -0.00005627, -0.00003456, 0.00001708, 0.00005282, -0.00008075, -0.00005197, 0.00000371, -0.00005036, -0.00005338, -0.00109900, 0.00273462, -0.00109900, -0.00005338, -0.00005036, 0.00000371, -0.00005197, -0.00008075, 0.00005282, 0.00001708, -0.00003456, -0.00005627, 0.00008113, 0.00000168, -0.00017026, 0.00020119, -0.00008728, },
		{ 0.00001211, -0.00014622, 0.00018162, 0.00001532, -0.00014835, 0.00007410, 0.00004955, -0.00013526, 0.00010125, -0.00009051, 0.00016982, -0.00034023, 0.00032362, -0.00044394, 0.00058350, -0.00150926, 0.00155791, 0.00097094, -0.00132169, 0.00058923, -0.00046538, 0.00030397, -0.00033803, 0.00019383, -0.00010010, 0.00009888, -0.00014457, 0.00005956, 0.00007239, -0.00014316, 0.00004837, 0.00000703, },
		{ -0.00000693, -0.00004848, -0.00007429, 0.00033202, -0.00026541, 0.00004574, 0.00005400, -0.00004984, -0.00001939, 0.00000052, 0.00008329, -0.00015595, 0.00004296, -0.00008440, 0.00011985, -0.00039614, -0.00063717, 0.00262954, -0.00142861, 0.00026254, -0.00022864, 0.00009179, -0.00014177, 0.00000036, 0.00001620, 0.00003364, -0.00005251, -0.00005773, 0.00010828, -0.00004234, -0.00007858, 0.00003375, },
		{ -0.00008415, 0.00012349, -0.00030301, 0.00041230, -0.00008017, -0.00012381, 0.00005747, 0.00005368, -0.00012998, 0.00010049, -0.00007708, 0.00013195, -0.00031377, 0.00031431, -0.00038293, 0.00049194, -0.00155516, 0.00203853, 0.00034258, -0.00102026, 0.00051048, -0.00044490, 0.00025714, -0.00030679, 0.00020117, -0.00010509, 0.00009307, -0.00015566, 0.00008288, 0.00005137, -0.00010995, 0.00001616, },
		{ -0.00013485, 0.00013282, -0.00019532, 0.00004955, 0.00039579, -0.00034479, 0.00008062, 0.00003458, -0.00004367, -0.00000637, -0.00001540, 0.00010575, -0.00021798, 0.00013242, -0.00016164, 0.00026055, -0.00072352, -0.00009172, 0.00232732, -0.00159532, 0.00051407, -0.00039256, 0.00016892, -0.00021633, 0.00007929, -0.00002394, 0.00004956, -0.00007327, -0.00004979, 0.00012744, -0.00007710, -0.00000877, },
		{ -0.00007303, -0.00008062, 0.00014995, -0.00040755, 0.00065993, -0.00022451, -0.00007303, 0.00002674, 0.00006977, -0.00012965, 0.00009696, -0.00006077, 0.00008446, -0.00026248, 0.00027813, -0.00029004, 0.00032512, -0.00144981, 0.00236063, -0.00025834, -0.00064684, 0.00036017, -0.00038584, 0.00018854, -0.00024974, 0.00019082, -0.00010494, 0.00008361, -0.00016532, 0.00011578, 0.00001358, -0.00005535, },
		{ 0.00002449, -0.00026989, 0.00028760, -0.00040520, 0.00028193, 0.00037209, -0.00039152, 0.00009861, 0.00002897, -0.00004284, 0.00000555, -0.00003007, 0.00011922, -0.00026138, 0.00020680, -0.00021868, 0.00035435, -0.00099473, 0.00047739, 0.00186512, -0.00159028, 0.00067359, -0.00052200, 0.00022577, -0.00026744, 0.00014815, -0.00006450, 0.00006424, -0.00009613, -0.00002969, 0.00013172, -0.00009491, },
		{ 0.00001014, -0.00020120, 0.00001129, 0.00007064, -0.00041568, 0.00087964, -0.00039538, -0.00000423, -0.00001126, 0.00009304, -0.00013372, 0.00009099, -0.00004260, 0.00003242, -0.00019314, 0.00022067, -0.00017724, 0.00010401, -0.00120346, 0.00249200, -0.00076859, -0.00025110, 0.00016086, -0.00029674, 0.00010646, -0.00017322, 0.00016376, -0.00009947, 0.00007046, -0.00016992, 0.00015220, -0.00003531, },
		{ -0.00009081, 0.00001269, -0.00034307, 0.00041028, -0.00063179, 0.00060041, 0.00025411, -0.00039377, 0.00009477, 0.00004061, -0.00005133, 0.00001754, -0.00004293, 0.00012372, -0.00028316, 0.00025875, -0.00024906, 0.00039043, -0.00117622, 0.00100710, 0.00129895, -0.00142720, 0.00072724, -0.00060195, 0.00025589, -0.00028992, 0.00020034, -0.00010230, 0.00007710, -0.00011969, 0.00000313, 0.00011646, },
		{ -0.00008738, 0.00007385, -0.00034355, 0.00016880, -0.00011436, -0.00029759, 0.00102797, -0.00056516, 0.00007097, -0.00004822, 0.00011692, -0.00014026, 0.00008283, -0.00002354, -0.00001893, -0.00011442, 0.00015020, -0.00005902, -0.00014301, -0.00084483, 0.00242463, -0.00113905, 0.00011670, -0.00005873, -0.00019000, 0.00002088, -0.00008586, 0.00012281, -0.00008887, 0.00005385, -0.00016606, 0.00018472, },
		{ 0.00008017, -0.00006347, -0.00004509, -0.00033708, 0.00045901, -0.00082129, 0.00096534, 0.00005006, -0.00034679, 0.00006774, 0.00006966, -0.00007132, 0.00003060, -0.00005349, 0.00012016, -0.00028311, 0.00028421, -0.00025012, 0.00036601, -0.00124653, 0.00143993, 0.00069568, -0.00113974, 0.00067663, -0.00062465, 0.00025664, -0.00028228, 0.00023137, -0.00013442, 0.00008766, -0.00014195, 0.00004677, },
		{ 0.00020577, -0.00013366, 0.00012300, -0.00047246, 0.00035648, -0.00038358, -0.00004346, 0.00106985, -0.00070516, 0.00013933, -0.00007576, 0.00013417, -0.00014631, 0.00007267, -0.00000447, -0.00006492, -0.00003568, 0.00007640, 0.00004975, -0.00038428, -0.00041707, 0.00217548, -0.00134112, 0.00041308, -0.00026801, -0.00008007, -0.00005793, 0.00000265, 0.00007222, -0.00007368, 0.00003427, -0.00015120, },
		{ 0.00009709, 0.00002485, -0.00001544, -0.00014195, -0.00024849, 0.00040452, -0.00092130, 0.00132557, -0.00021775, -0.00025375, 0.00002006, 0.00011293, -0.00010267, 0.00004533, -0.00006133, 0.00011016, -0.00026372, 0.00028280, -0.00022330, 0.00028670, -0.00119929, 0.00173184, 0.00012361, -0.00077608, 0.00053820, -0.00059063, 0.00022948, -0.00024689, 0.00023941, -0.00015854, 0.00009550, -0.00016058, },
		{ -0.00012412, 0.00020885, -0.00016552, 0.00014669, -0.00056190, 0.00053211, -0.00069717, 0.00033400, 0.00098443, -0.00079013, 0.00018807, -0.00008677, 0.00013810, -0.00014842, 0.00006060, 0.00001391, -0.00010202, 0.00003438, 0.00000893, 0.00013588, -0.00058948, 0.00002836, 0.00178361, -0.00137002, 0.00060739, -0.00043905, 0.00001867, -0.00012089, 0.00008291, 0.00001719, -0.00005477, 0.00001252, },
		{ -0.00017325, 0.00014816, -0.00004414, 0.00004718, -0.00026414, -0.00008845, 0.00023539, -0.00088972, 0.00162608, -0.00051590, -0.00012546, -0.00004218, 0.00016427, -0.00014284, 0.00006168, -0.00006616, 0.00009571, -0.00022970, 0.00025780, -0.00017391, 0.00016559, -0.00104404, 0.00185810, -0.00035703, -0.00039139, 0.00034037, -0.00050841, 0.00017971, -0.00018961, 0.00022548, -0.00017319, 0.00010033, },
		{ -0.00001039, -0.00008529, 0.00018970, -0.00017610, 0.00013778, -0.00059167, 0.00065358, -0.00100294, 0.00080043, 0.00076878, -0.00080223, 0.00020692, -0.00007661, 0.00012377, -0.00014326, 0.00004670, 0.00003108, -0.00012823, 0.00008890, -0.00004395, 0.00018981, -0.00073396, 0.00043899, 0.00130422, -0.00124445, 0.00068596, -0.00055072, 0.00009404, -0.00016137, 0.00014702, -0.00003694, -0.00003325, },
		},
		{ { -0.00006923, 0.00001416, 0.00005356, -0.00008388, -0.00008362, 0.00011021, -0.00004817, -0.00072024, 0.00162833, -0.00072024, -0.00004817, 0.00011021, -0.00008362, -0.00008388, 0.00005356, 0.00001416, -0.00006923, 0.00001416, 0.00005356, -0.00008388, -0.00008362, 0.00011021, -0.00004817, -0.00072024, 0.00162833, -0.00072024, -0.00004817, 0.00011021, -0.00008362, -0.00008388, 0.00005356, 0.00001416, },
		{ -0.00005347, -0.00006765, 0.00012151, -0.00011131, 0.00002831, -0.00020610, 0.00044784, -0.00097099, 0.00078281, 0.00078281, -0.00097099, 0.00044784, -0.00020610, 0.00002831, -0.00011131, 0.00012151, -0.00006765, -0.00005347, 0.00010188, -0.00008661, 0.00001162, -0.00018226, 0.00041640, -0.00087371, 0.00068488, 0.00068488, -0.00087371, 0.00041640, -0.00018226, 0.00001162, -0.00008661, 0.00010188, },
		{ 0.00001042, -0.00006923, 0.00001789, 0.00005176, -0.00009085, -0.00008461, 0.00011284, -0.00006717, -0.00077936, 0.00181777, -0.00077936, -0.00006717, 0.00011284, -0.00008461, -0.00009085, 0.00005176, 0.00001789, -0.00006923, 0.00001042, 0.00005536, -0.00007690, -0.00008263, 0.00010758, -0.00002916, -0.00066112, 0.00143888, -0.00066112, -0.00002916, 0.00010758, -0.00008263, -0.00007690, 0.00005536, },
		{ 0.00008262, -0.00003957, -0.00008155, 0.00014077, -0.00013554, 0.00004468, -0.00022948, 0.00047868, -0.00106640, 0.00087886, 0.00087886, -0.00106640, 0.00047868, -0.00022948, 0.00004468, -0.00013554, 0.00014077, -0.00008155, -0.00003957, 0.00008262, -0.00006238, -0.00000474, -0.00015888, 0.00038556, -0.00077829, 0.00058883, 0.00058883, -0.00077829, 0.00038556, -0.00015888, -0.00000474, -0.00006238, },
		{ 0.00005708, 0.00000683, -0.00006923, 0.00002148, 0.00005004, -0.00009755, -0.00008557, 0.00011537, -0.00008544, -0.00083620, 0.00199994, -0.00083620, -0.00008544, 0.00011537, -0.00008557, -0.00009755, 0.00005004, 0.00002148, -0.00006923, 0.00000683, 0.00005708, -0.00007020, -0.00008167, 0.00010505, -0.00001089, -0.00060428, 0.00125672, -0.00060428, -0.00001089, 0.00010505, -0.00008167, -0.00007020, },
		{ -0.00003956, 0.00006448, -0.00002647, -0.00009465, 0.00015891, -0.00015836, 0.00006009, -0.00025151, 0.00050773, -0.00115628, 0.00096933, 0.00096933, -0.00115628, 0.00050773, -0.00025151, 0.00006009, -0.00015836, 0.00015891, -0.00009465, -0.00002647, 0.00006448, -0.00003956, -0.00002016, -0.00013685, 0.00035651, -0.00068842, 0.00049835, 0.00049835, -0.00068842, 0.00035651, -0.00013685, -0.00002016, },
		{ -0.00006403, 0.00005867, 0.00000352, -0.00006923, 0.00002479, 0.00004845, -0.00010373, -0.00008645, 0.00011771, -0.00010228, -0.00088859, 0.00216782, -0.00088859, -0.00010228, 0.00011771, -0.00008645, -0.00010373, 0.00004845, 0.00002479, -0.00006923, 0.00000352, 0.00005867, -0.00006403, -0.00008079, 0.00010272, 0.00000595, -0.00055189, 0.00108884, -0.00055189, 0.00000595, 0.00010272, -0.00008079, },
		{ -0.00003403, -0.00001902, 0.00004815, -0.00001468, -0.00010643, 0.00017524, -0.00017890, 0.00007397, -0.00027133, 0.00053387, -0.00123716, 0.00105076, 0.00105076, -0.00123716, 0.00053387, -0.00027133, 0.00007397, -0.00017890, 0.00017524, -0.00010643, -0.00001468, 0.00004815, -0.00001902, -0.00003403, -0.00011703, 0.00033036, -0.00060753, 0.00041692, 0.00041692, -0.00060753, 0.00033036, -0.00011703, },
		{ -0.00008002, -0.00005861, 0.00006007, 0.00000062, -0.00006923, 0.00002769, 0.00004705, -0.00010914, -0.00008722, 0.00011975, -0.00011704, -0.00093451, 0.00231497, -0.00093451, -0.00011704, 0.00011975, -0.00008722, -0.00010914, 0.00004705, 0.00002769, -0.00006923, 0.00000062, 0.00006007, -0.00005861, -0.00008002, 0.00010068, 0.00002071, -0.00050597, 0.00094168, -0.00050597, 0.00002071, 0.00010068, },
		{ -0.00010017, -0.00004583, -0.00000155, 0.00003427, -0.00000466, -0.00011646, 0.00018912, -0.00019636, 0.00008577, -0.00028819, 0.00055611, -0.00130595, 0.00112001, 0.00112001, -0.00130595, 0.00055611, -0.00028819, 0.00008577, -0.00019636, 0.00018912, -0.00011646, -0.00000466, 0.00003427, -0.00000155, -0.00004583, -0.00010017, 0.00030813, -0.00053874, 0.00034768, 0.00034768, -0.00053874, 0.00030813, },
		{ 0.00009900, -0.00007939, -0.00005417, 0.00006121, -0.00000176, -0.00006923, 0.00003007, 0.00004591, -0.00011358, -0.00008785, 0.00012143, -0.00012915, -0.00097220, 0.00243574, -0.00097220, -0.00012915, 0.00012143, -0.00008785, -0.00011358, 0.00004591, 0.00003007, -0.00006923, -0.00000176, 0.00006121, -0.00005417, -0.00007939, 0.00009900, 0.00003282, -0.00046828, 0.00082092, -0.00046828, 0.00003282, },
		{ 0.00029066, -0.00008692, -0.00005510, 0.00001217, 0.00002336, 0.00000322, -0.00012433, 0.00020003, -0.00021009, 0.00009504, -0.00030144, 0.00057358, -0.00136000, 0.00117442, 0.00117442, -0.00136000, 0.00057358, -0.00030144, 0.00009504, -0.00021009, 0.00020003, -0.00012433, 0.00000322, 0.00002336, 0.00001217, -0.00005510, -0.00008692, 0.00029066, -0.00048470, 0.00029327, 0.00029327, -0.00048470, },
		{ 0.00004182, 0.00009776, -0.00007892, -0.00005087, 0.00006206, -0.00000353, -0.00006923, 0.00003184, 0.00004506, -0.00011689, -0.00008832, 0.00012267, -0.00013815, -0.00100020, 0.00252548, -0.00100020, -0.00013815, 0.00012267, -0.00008832, -0.00011689, 0.00004506, 0.00003184, -0.00006923, -0.00000353, 0.00006206, -0.00005087, -0.00007892, 0.00009776, 0.00004182, -0.00044028, 0.00073118, -0.00044028, },
		{ -0.00044747, 0.00027862, -0.00007780, -0.00006149, 0.00002163, 0.00001584, 0.00000864, -0.00012976, 0.00020755, -0.00021954, 0.00010143, -0.00031056, 0.00058561, -0.00139723, 0.00121189, 0.00121189, -0.00139723, 0.00058561, -0.00031056, 0.00010143, -0.00021954, 0.00020755, -0.00012976, 0.00000864, 0.00001584, 0.00002163, -0.00006149, -0.00007780, 0.00027862, -0.00044747, 0.00025579, 0.00025579, },
		{ -0.00042304, 0.00004737, 0.00009699, -0.00007863, -0.00004883, 0.00006259, -0.00000462, -0.00006923, 0.00003293, 0.00004453, -0.00011892, -0.00008861, 0.00012344, -0.00014370, -0.00101744, 0.00258073, -0.00101744, -0.00014370, 0.00012344, -0.00008861, -0.00011892, 0.00004453, 0.00003293, -0.00006923, -0.00000462, 0.00006259, -0.00004883, -0.00007863, 0.00009699, 0.00004737, -0.00042304, 0.00067592, },
		{ 0.00023669, -0.00042849, 0.00027249, -0.00007315, -0.00006475, 0.00002644, 0.00001201, 0.00001141, -0.00013253, 0.00021138, -0.00022436, 0.00010468, -0.00031521, 0.00059175, -0.00141621, 0.00123100, 0.00123100, -0.00141621, 0.00059175, -0.00031521, 0.00010468, -0.00022436, 0.00021138, -0.00013253, 0.00001141, 0.00001201, 0.00002644, -0.00006475, -0.00007315, 0.00027249, -0.00042849, 0.00023669, },
		{ 0.00065727, -0.00041721, 0.00004924, 0.00009673, -0.00007853, -0.00004815, 0.00006276, -0.00000498, -0.00006923, 0.00003330, 0.00004436, -0.00011961, -0.00008871, 0.00012370, -0.00014557, -0.00102327, 0.00259939, -0.00102327, -0.00014557, 0.00012370, -0.00008871, -0.00011961, 0.00004436, 0.00003330, -0.00006923, -0.00000498, 0.00006276, -0.00004815, -0.00007853, 0.00009673, 0.00004924, -0.00041721, },
		{ 0.00023669, 0.00023669, -0.00042849, 0.00027249, -0.00007315, -0.00006475, 0.00002644, 0.00001201, 0.00001141, -0.00013253, 0.00021138, -0.00022436, 0.00010468, -0.00031521, 0.00059175, -0.00141621, 0.00123100, 0.00123100, -0.00141621, 0.00059175, -0.00031521, 0.00010468, -0.00022436, 0.00021138, -0.00013253, 0.00001141, 0.00001201, 0.00002644, -0.00006475, -0.00007315, 0.00027249, -0.00042849, },
		{ -0.00042304, 0.00067592, -0.00042304, 0.00004737, 0.00009699, -0.00007863, -0.00004883, 0.00006259, -0.00000462, -0.00006923, 0.00003293, 0.00004453, -0.00011892, -0.00008861, 0.00012344, -0.00014370, -0.00101744, 0.00258073, -0.00101744, -0.00014370, 0.00012344, -0.00008861, -0.00011892, 0.00004453, 0.00003293, -0.00006923, -0.00000462, 0.00006259, -0.00004883, -0.00007863, 0.00009699, 0.00004737, },
		{ -0.00044747, 0.00025579, 0.00025579, -0.00044747, 0.00027862, -0.00007780, -0.00006149, 0.00002163, 0.00001584, 0.00000864, -0.00012976, 0.00020755, -0.00021954, 0.00010143, -0.00031056, 0.00058561, -0.00139723, 0.00121189, 0.00121189, -0.00139723, 0.00058561, -0.00031056, 0.00010143, -0.00021954, 0.00020755, -0.00012976, 0.00000864, 0.00001584, 0.00002163, -0.00006149, -0.00007780, 0.00027862, },
		{ 0.00004182, -0.00044028, 0.00073118, -0.00044028, 0.00004182, 0.00009776, -0.00007892, -0.00005087, 0.00006206, -0.00000353, -0.00006923, 0.00003184, 0.00004506, -0.00011689, -0.00008832, 0.00012267, -0.00013815, -0.00100020, 0.00252548, -0.00100020, -0.00013815, 0.00012267, -0.00008832, -0.00011689, 0.00004506, 0.00003184, -0.00006923, -0.00000353, 0.00006206, -0.00005087, -0.00007892, 0.00009776, },
		{ 0.00029066, -0.00048470, 0.00029327, 0.00029327, -0.00048470, 0.00029066, -0.00008692, -0.00005510, 0.00001217, 0.00002336, 0.00000322, -0.00012433, 0.00020003, -0.00021009, 0.00009504, -0.00030144, 0.00057358, -0.00136000, 0.00117442, 0.00117442, -0.00136000, 0.00057358, -0.00030144, 0.00009504, -0.00021009, 0.00020003, -0.00012433, 0.00000322, 0.00002336, 0.00001217, -0.00005510, -0.00008692, },
		{ 0.00009900, 0.00003282, -0.00046828, 0.00082092, -0.00046828, 0.00003282, 0.00009900, -0.00007939, -0.00005417, 0.00006121, -0.00000176, -0.00006923, 0.00003007, 0.00004591, -0.00011358, -0.00008785, 0.00012143, -0.00012915, -0.00097220, 0.00243574, -0.00097220, -0.00012915, 0.00012143, -0.00008785, -0.00011358, 0.00004591, 0.00003007, -0.00006923, -0.00000176, 0.00006121, -0.00005417, -0.00007939, },
		{ -0.00010017, 0.00030813, -0.00053874, 0.00034768, 0.00034768, -0.00053874, 0.00030813, -0.00010017, -0.00004583, -0.00000155, 0.00003427, -0.00000466, -0.00011646, 0.00018912, -0.00019636, 0.00008577, -0.00028819, 0.00055611, -0.00130595, 0.00112001, 0.00112001, -0.00130595, 0.00055611, -0.00028819, 0.00008577, -0.00019636, 0.00018912, -0.00011646, -0.00000466, 0.00003427, -0.00000155, -0.00004583, },
		{ -0.00008002, 0.00010068, 0.00002071, -0.00050597, 0.00094168, -0.00050597, 0.00002071, 0.00010068, -0.00008002, -0.00005861, 0.00006007, 0.00000062, -0.00006923, 0.00002769, 0.00004705, -0.00010914, -0.00008722, 0.00011975, -0.00011704, -0.00093451, 0.00231497, -0.00093451, -0.00011704, 0.00011975, -0.00008722, -0.00010914, 0.00004705, 0.00002769, -0.00006923, 0.00000062, 0.00006007, -0.00005861, },
		{ -0.00003403, -0.00011703, 0.00033036, -0.00060753, 0.00041692, 0.00041692, -0.00060753, 0.00033036, -0.00011703, -0.00003403, -0.00001902, 0.00004815, -0.00001468, -0.00010643, 0.00017524, -0.00017890, 0.00007397, -0.00027133, 0.00053387, -0.00123716, 0.00105076, 0.00105076, -0.00123716, 0.00053387, -0.00027133, 0.00007397, -0.00017890, 0.00017524, -0.00010643, -0.00001468, 0.00004815, -0.00001902, },
		{ -0.00006403, -0.00008079, 0.00010272, 0.00000595, -0.00055189, 0.00108884, -0.00055189, 0.00000595, 0.00010272, -0.00008079, -0.00006403, 0.00005867, 0.00000352, -0.00006923, 0.00002479, 0.00004845, -0.00010373, -0.00008645, 0.00011771, -0.00010228, -0.00088859, 0.00216782, -0.00088859, -0.00010228, 0.00011771, -0.00008645, -0.00010373, 0.00004845, 0.00002479, -0.00006923, 0.00000352, 0.00005867, },
		{ -0.00003956, -0.00002016, -0.00013685, 0.00035651, -0.00068842, 0.00049835, 0.00049835, -0.00068842, 0.00035651, -0.00013685, -0.00002016, -0.00003956, 0.00006448, -0.00002647, -0.00009465, 0.00015891, -0.00015836, 0.00006009, -0.00025151, 0.00050773, -0.00115628, 0.00096933, 0.00096933, -0.00115628, 0.00050773, -0.00025151, 0.00006009, -0.00015836, 0.00015891, -0.00009465, -0.00002647, 0.00006448, },
		{ 0.00005708, -0.00007020, -0.00008167, 0.00010505, -0.00001089, -0.00060428, 0.00125672, -0.00060428, -0.00001089, 0.00010505, -0.00008167, -0.00007020, 0.00005708, 0.00000683, -0.00006923, 0.00002148, 0.00005004, -0.00009755, -0.00008557, 0.00011537, -0.00008544, -0.00083620, 0.00199994, -0.00083620, -0.00008544, 0.00011537, -0.00008557, -0.00009755, 0.00005004, 0.00002148, -0.00006923, 0.00000683, },
		{ 0.00008262, -0.00006238, -0.00000474, -0.00015888, 0.00038556, -0.00077829, 0.00058883, 0.00058883, -0.00077829, 0.00038556, -0.00015888, -0.00000474, -0.00006238, 0.00008262, -0.00003957, -0.00008155, 0.00014077, -0.00013554, 0.00004468, -0.00022948, 0.00047868, -0.00106640, 0.00087886, 0.00087886, -0.00106640, 0.00047868, -0.00022948, 0.00004468, -0.00013554, 0.00014077, -0.00008155, -0.00003957, },
		{ 0.00001042, 0.00005536, -0.00007690, -0.00008263, 0.00010758, -0.00002916, -0.00066112, 0.00143888, -0.00066112, -0.00002916, 0.00010758, -0.00008263, -0.00007690, 0.00005536, 0.00001042, -0.00006923, 0.00001789, 0.00005176, -0.00009085, -0.00008461, 0.00011284, -0.00006717, -0.00077936, 0.00181777, -0.00077936, -0.00006717, 0.00011284, -0.00008461, -0.00009085, 0.00005176, 0.00001789, -0.00006923, },
		{ -0.00005347, 0.00010188, -0.00008661, 0.00001162, -0.00018226, 0.00041640, -0.00087371, 0.00068488, 0.00068488, -0.00087371, 0.00041640, -0.00018226, 0.00001162, -0.00008661, 0.00010188, -0.00005347, -0.00006765, 0.00012151, -0.00011131, 0.00002831, -0.00020610, 0.00044784, -0.00097099, 0.00078281, 0.00078281, -0.00097099, 0.00044784, -0.00020610, 0.00002831, -0.00011131, 0.00012151, -0.00006765, },
		},
		{ { -0.00006779, 0.00007893, -0.00018720, 0.00021553, -0.00011022, 0.00002116, -0.00080471, 0.00181631, -0.00070222, -0.00003944, 0.00011897, -0.00039314, 0.00011561, -0.00011880, 0.00019320, -0.00017792, 0.00010196, -0.00017792, 0.00019320, -0.00011880, 0.00011561, -0.00039314, 0.00011897, -0.00003944, -0.00070222, 0.00181631, -0.00080471, 0.00002116, -0.00011022, 0.00021553, -0.00018720, 0.00007893, },
		{ 0.00004670, 0.00003108, -0.00012823, 0.00008890, -0.00004395, 0.00018981, -0.00073396, 0.00043899, 0.00130422, -0.00124445, 0.00068596, -0.00055072, 0.00009404, -0.00016137, 0.00014702, -0.00003694, -0.00003325, -0.00001039, -0.00008529, 0.00018970, -0.00017610, 0.00013778, -0.00059167, 0.00065358, -0.00100294, 0.00080043, 0.00076878, -0.00080223, 0.00020692, -0.00007661, 0.00012377, -0.00014326, },
		{ 0.00006168, -0.00006616, 0.00009571, -0.00022970, 0.00025780, -0.00017391, 0.00016559, -0.00104404, 0.00185810, -0.00035703, -0.00039139, 0.00034037, -0.00050841, 0.00017971, -0.00018961, 0.00022548, -0.00017319, 0.00010033, -0.00017325, 0.00014816, -0.00004414, 0.00004718, -0.00026414, -0.00008845, 0.00023539, -0.00088972, 0.00162608, -0.00051590, -0.00012546, -0.00004218, 0.00016427, -0.00014284, },
		{ -0.00014842, 0.00006060, 0.00001391, -0.00010202, 0.00003438, 0.00000893, 0.00013588, -0.00058948, 0.00002836, 0.00178361, -0.00137002, 0.00060739, -0.00043905, 0.00001867, -0.00012089, 0.00008291, 0.00001719, -0.00005477, 0.00001252, -0.00012412, 0.00020885, -0.00016552, 0.00014669, -0.00056190, 0.00053211, -0.00069717, 0.00033400, 0.00098443, -0.00079013, 0.00018807, -0.00008677, 0.00013810, },
		{ -0.00010267, 0.00004533, -0.00006133, 0.00011016, -0.00026372, 0.00028280, -0.00022330, 0.00028670, -0.00119929, 0.00173184, 0.00012361, -0.00077608, 0.00053820, -0.00059063, 0.00022948, -0.00024689, 0.00023941, -0.00015854, 0.00009550, -0.00016058, 0.00009709, 0.00002485, -0.00001544, -0.00014195, -0.00024849, 0.00040452, -0.00092130, 0.00132557, -0.00021775, -0.00025375, 0.00002006, 0.00011293, },
		{ 0.00013417, -0.00014631, 0.00007267, -0.00000447, -0.00006492, -0.00003568, 0.00007640, 0.00004975, -0.00038428, -0.00041707, 0.00217548, -0.00134112, 0.00041308, -0.00026801, -0.00008007, -0.00005793, 0.00000265, 0.00007222, -0.00007368, 0.00003427, -0.00015120, 0.00020577, -0.00013366, 0.00012300, -0.00047246, 0.00035648, -0.00038358, -0.00004346, 0.00106985, -0.00070516, 0.00013933, -0.00007576, },
		{ 0.00006966, -0.00007132, 0.00003060, -0.00005349, 0.00012016, -0.00028311, 0.00028421, -0.00025012, 0.00036601, -0.00124653, 0.00143993, 0.00069568, -0.00113974, 0.00067663, -0.00062465, 0.00025664, -0.00028228, 0.00023137, -0.00013442, 0.00008766, -0.00014195, 0.00004677, 0.00008017, -0.00006347, -0.00004509, -0.00033708, 0.00045901, -0.00082129, 0.00096534, 0.00005006, -0.00034679, 0.00006774, },
		{ -0.00004822, 0.00011692, -0.00014026, 0.00008283, -0.00002354, -0.00001893, -0.00011442, 0.00015020, -0.00005902, -0.00014301, -0.00084483, 0.00242463, -0.00113905, 0.00011670, -0.00005873, -0.00019000, 0.00002088, -0.00008586, 0.00012281, -0.00008887, 0.00005385, -0.00016606, 0.00018472, -0.00008738, 0.00007385, -0.00034355, 0.00016880, -0.00011436, -0.00029759, 0.00102797, -0.00056516, 0.00007097, },
		{ 0.00009477, 0.00004061, -0.00005133, 0.00001754, -0.00004293, 0.00012372, -0.00028316, 0.00025875, -0.00024906, 0.00039043, -0.00117622, 0.00100710, 0.00129895, -0.00142720, 0.00072724, -0.00060195, 0.00025589, -0.00028992, 0.00020034, -0.00010230, 0.00007710, -0.00011969, 0.00000313, 0.00011646, -0.00009081, 0.00001269, -0.00034307, 0.00041028, -0.00063179, 0.00060041, 0.00025411, -0.00039377, },
		{ -0.00000423, -0.00001126, 0.00009304, -0.00013372, 0.00009099, -0.00004260, 0.00003242, -0.00019314, 0.00022067, -0.00017724, 0.00010401, -0.00120346, 0.00249200, -0.00076859, -0.00025110, 0.00016086, -0.00029674, 0.00010646, -0.00017322, 0.00016376, -0.00009947, 0.00007046, -0.00016992, 0.00015220, -0.00003531, 0.00001014, -0.00020120, 0.00001129, 0.00007064, -0.00041568, 0.00087964, -0.00039538, },
		{ -0.00039152, 0.00009861, 0.00002897, -0.00004284, 0.00000555, -0.00003007, 0.00011922, -0.00026138, 0.00020680, -0.00021868, 0.00035435, -0.00099473, 0.00047739, 0.00186512, -0.00159028, 0.00067359, -0.00052200, 0.00022577, -0.00026744, 0.00014815, -0.00006450, 0.00006424, -0.00009613, -0.00002969, 0.00013172, -0.00009491, 0.00002449, -0.00026989, 0.00028760, -0.00040520, 0.00028193, 0.00037209, },
		{ -0.00022451, -0.00007303, 0.00002674, 0.00006977, -0.00012965, 0.00009696, -0.00006077, 0.00008446, -0.00026248, 0.00027813, -0.00029004, 0.00032512, -0.00144981, 0.00236063, -0.00025834, -0.00064684, 0.00036017, -0.00038584, 0.00018854, -0.00024974, 0.00019082, -0.00010494, 0.00008361, -0.00016532, 0.00011578, 0.00001358, -0.00005535, -0.00007303, -0.00008062, 0.00014995, -0.00040755, 0.00065993, },
		{ 0.00039579, -0.00034479, 0.00008062, 0.00003458, -0.00004367, -0.00000637, -0.00001540, 0.00010575, -0.00021798, 0.00013242, -0.00016164, 0.00026055, -0.00072352, -0.00009172, 0.00232732, -0.00159532, 0.00051407, -0.00039256, 0.00016892, -0.00021633, 0.00007929, -0.00002394, 0.00004956, -0.00007327, -0.00004979, 0.00012744, -0.00007710, -0.00000877, -0.00013485, 0.00013282, -0.00019532, 0.00004955, },
		{ 0.00041230, -0.00008017, -0.00012381, 0.00005747, 0.00005368, -0.00012998, 0.00010049, -0.00007708, 0.00013195, -0.00031377, 0.00031431, -0.00038293, 0.00049194, -0.00155516, 0.00203853, 0.00034258, -0.00102026, 0.00051048, -0.00044490, 0.00025714, -0.00030679, 0.00020117, -0.00010509, 0.00009307, -0.00015566, 0.00008288, 0.00005137, -0.00010995, 0.00001616, -0.00008415, 0.00012349, -0.00030301, },
		{ -0.00007429, 0.00033202, -0.00026541, 0.00004574, 0.00005400, -0.00004984, -0.00001939, 0.00000052, 0.00008329, -0.00015595, 0.00004296, -0.00008440, 0.00011985, -0.00039614, -0.00063717, 0.00262954, -0.00142861, 0.00026254, -0.00022864, 0.00009179, -0.00014177, 0.00000036, 0.00001620, 0.00003364, -0.00005251, -0.00005773, 0.00010828, -0.00004234, -0.00007858, 0.00003375, -0.00000693, -0.00004848, },
		{ -0.00014622, 0.00018162, 0.00001532, -0.00014835, 0.00007410, 0.00004955, -0.00013526, 0.00010125, -0.00009051, 0.00016982, -0.00034023, 0.00032362, -0.00044394, 0.00058350, -0.00150926, 0.00155791, 0.00097094, -0.00132169, 0.00058923, -0.00046538, 0.00030397, -0.00033803, 0.00019383, -0.00010010, 0.00009888, -0.00014457, 0.00005956, 0.00007239, -0.00014316, 0.00004837, 0.00000703, 0.00001211, },
		{ 0.00000414, -0.00008728, 0.00020119, -0.00017026, 0.00000168, 0.00008113, -0.00005627, -0.00003456, 0.00001708, 0.00005282, -0.00008075, -0.00005197, 0.00000371, -0.00005036, -0.00005338, -0.00109900, 0.00273462, -0.00109900, -0.00005338, -0.00005036, 0.00000371, -0.00005197, -0.00008075, 0.00005282, 0.00001708, -0.00003456, -0.00005627, 0.00008113, 0.00000168, -0.00017026, 0.00020119, -0.00008728, },
		{ -0.00014622, 0.00001211, 0.00000703, 0.00004837, -0.00014316, 0.00007239, 0.00005956, -0.00014457, 0.00009888, -0.00010010, 0.00019383, -0.00033803, 0.00030397, -0.00046538, 0.00058923, -0.00132169, 0.00097094, 0.00155791, -0.00150926, 0.00058350, -0.00044394, 0.00032362, -0.00034023, 0.00016982, -0.00009051, 0.00010125, -0.00013526, 0.00004955, 0.00007410, -0.00014835, 0.00001532, 0.00018162, },
		{ -0.00007429, -0.00004848, -0.00000693, 0.00003375, -0.00007858, -0.00004234, 0.00010828, -0.00005773, -0.00005251, 0.00003364, 0.00001620, 0.00000036, -0.00014177, 0.00009179, -0.00022864, 0.00026254, -0.00142861, 0.00262954, -0.00063717, -0.00039614, 0.00011985, -0.00008440, 0.00004296, -0.00015595, 0.00008329, 0.00000052, -0.00001939, -0.00004984, 0.00005400, 0.00004574, -0.00026541, 0.00033202, },
		{ 0.00041230, -0.00030301, 0.00012349, -0.00008415, 0.00001616, -0.00010995, 0.00005137, 0.00008288, -0.00015566, 0.00009307, -0.00010509, 0.00020117, -0.00030679, 0.00025714, -0.00044490, 0.00051048, -0.00102026, 0.00034258, 0.00203853, -0.00155516, 0.00049194, -0.00038293, 0.00031431, -0.00031377, 0.00013195, -0.00007708, 0.00010049, -0.00012998, 0.00005368, 0.00005747, -0.00012381, -0.00008017, },
		{ 0.00039579, 0.00004955, -0.00019532, 0.00013282, -0.00013485, -0.00000877, -0.00007710, 0.00012744, -0.00004979, -0.00007327, 0.00004956, -0.00002394, 0.00007929, -0.00021633, 0.00016892, -0.00039256, 0.00051407, -0.00159532, 0.00232732, -0.00009172, -0.00072352, 0.00026055, -0.00016164, 0.00013242, -0.00021798, 0.00010575, -0.00001540, -0.00000637, -0.00004367, 0.00003458, 0.00008062, -0.00034479, },
		{ -0.00022451, 0.00065993, -0.00040755, 0.00014995, -0.00008062, -0.00007303, -0.00005535, 0.00001358, 0.00011578, -0.00016532, 0.00008361, -0.00010494, 0.00019082, -0.00024974, 0.00018854, -0.00038584, 0.00036017, -0.00064684, -0.00025834, 0.00236063, -0.00144981, 0.00032512, -0.00029004, 0.00027813, -0.00026248, 0.00008446, -0.00006077, 0.00009696, -0.00012965, 0.00006977, 0.00002674, -0.00007303, },
		{ -0.00039152, 0.00037209, 0.00028193, -0.00040520, 0.00028760, -0.00026989, 0.00002449, -0.00009491, 0.00013172, -0.00002969, -0.00009613, 0.00006424, -0.00006450, 0.00014815, -0.00026744, 0.00022577, -0.00052200, 0.00067359, -0.00159028, 0.00186512, 0.00047739, -0.00099473, 0.00035435, -0.00021868, 0.00020680, -0.00026138, 0.00011922, -0.00003007, 0.00000555, -0.00004284, 0.00002897, 0.00009861, },
		{ -0.00000423, -0.00039538, 0.00087964, -0.00041568, 0.00007064, 0.00001129, -0.00020120, 0.00001014, -0.00003531, 0.00015220, -0.00016992, 0.00007046, -0.00009947, 0.00016376, -0.00017322, 0.00010646, -0.00029674, 0.00016086, -0.00025110, -0.00076859, 0.00249200, -0.00120346, 0.00010401, -0.00017724, 0.00022067, -0.00019314, 0.00003242, -0.00004260, 0.00009099, -0.00013372, 0.00009304, -0.00001126, },
		{ 0.00009477, -0.00039377, 0.00025411, 0.00060041, -0.00063179, 0.00041028, -0.00034307, 0.00001269, -0.00009081, 0.00011646, 0.00000313, -0.00011969, 0.00007710, -0.00010230, 0.00020034, -0.00028992, 0.00025589, -0.00060195, 0.00072724, -0.00142720, 0.00129895, 0.00100710, -0.00117622, 0.00039043, -0.00024906, 0.00025875, -0.00028316, 0.00012372, -0.00004293, 0.00001754, -0.00005133, 0.00004061, },
		{ -0.00004822, 0.00007097, -0.00056516, 0.00102797, -0.00029759, -0.00011436, 0.00016880, -0.00034355, 0.00007385, -0.00008738, 0.00018472, -0.00016606, 0.00005385, -0.00008887, 0.00012281, -0.00008586, 0.00002088, -0.00019000, -0.00005873, 0.00011670, -0.00113905, 0.00242463, -0.00084483, -0.00014301, -0.00005902, 0.00015020, -0.00011442, -0.00001893, -0.00002354, 0.00008283, -0.00014026, 0.00011692, },
		{ 0.00006966, 0.00006774, -0.00034679, 0.00005006, 0.00096534, -0.00082129, 0.00045901, -0.00033708, -0.00004509, -0.00006347, 0.00008017, 0.00004677, -0.00014195, 0.00008766, -0.00013442, 0.00023137, -0.00028228, 0.00025664, -0.00062465, 0.00067663, -0.00113974, 0.00069568, 0.00143993, -0.00124653, 0.00036601, -0.00025012, 0.00028421, -0.00028311, 0.00012016, -0.00005349, 0.00003060, -0.00007132, },
		{ 0.00013417, -0.00007576, 0.00013933, -0.00070516, 0.00106985, -0.00004346, -0.00038358, 0.00035648, -0.00047246, 0.00012300, -0.00013366, 0.00020577, -0.00015120, 0.00003427, -0.00007368, 0.00007222, 0.00000265, -0.00005793, -0.00008007, -0.00026801, 0.00041308, -0.00134112, 0.00217548, -0.00041707, -0.00038428, 0.00004975, 0.00007640, -0.00003568, -0.00006492, -0.00000447, 0.00007267, -0.00014631, },
		{ -0.00010267, 0.00011293, 0.00002006, -0.00025375, -0.00021775, 0.00132557, -0.00092130, 0.00040452, -0.00024849, -0.00014195, -0.00001544, 0.00002485, 0.00009709, -0.00016058, 0.00009550, -0.00015854, 0.00023941, -0.00024689, 0.00022948, -0.00059063, 0.00053820, -0.00077608, 0.00012361, 0.00173184, -0.00119929, 0.00028670, -0.00022330, 0.00028280, -0.00026372, 0.00011016, -0.00006133, 0.00004533, },
		{ -0.00014842, 0.00013810, -0.00008677, 0.00018807, -0.00079013, 0.00098443, 0.00033400, -0.00069717, 0.00053211, -0.00056190, 0.00014669, -0.00016552, 0.00020885, -0.00012412, 0.00001252, -0.00005477, 0.00001719, 0.00008291, -0.00012089, 0.00001867, -0.00043905, 0.00060739, -0.00137002, 0.00178361, 0.00002836, -0.00058948, 0.00013588, 0.00000893, 0.00003438, -0.00010202, 0.00001391, 0.00006060, },
		{ 0.00006168, -0.00014284, 0.00016427, -0.00004218, -0.00012546, -0.00051590, 0.00162608, -0.00088972, 0.00023539, -0.00008845, -0.00026414, 0.00004718, -0.00004414, 0.00014816, -0.00017325, 0.00010033, -0.00017319, 0.00022548, -0.00018961, 0.00017971, -0.00050841, 0.00034037, -0.00039139, -0.00035703, 0.00185810, -0.00104404, 0.00016559, -0.00017391, 0.00025780, -0.00022970, 0.00009571, -0.00006616, },
		{ 0.00004670, -0.00014326, 0.00012377, -0.00007661, 0.00020692, -0.00080223, 0.00076878, 0.00080043, -0.00100294, 0.00065358, -0.00059167, 0.00013778, -0.00017610, 0.00018970, -0.00008529, -0.00001039, -0.00003325, -0.00003694, 0.00014702, -0.00016137, 0.00009404, -0.00055072, 0.00068596, -0.00124445, 0.00130422, 0.00043899, -0.00073396, 0.00018981, -0.00004395, 0.00008890, -0.00012823, 0.00003108, },
		},
		{ { 0.00001852, -0.00011511, 0.00015382, -0.00012801, 0.00001185, -0.00079230, 0.00189741, -0.00088941, 0.00007463, 0.00001333, -0.00020457, 0.00016804, -0.00000646, 0.00007252, -0.00045480, 0.00031144, -0.00015699, 0.00031144, -0.00045480, 0.00007252, -0.00000646, 0.00016804, -0.00020457, 0.00001333, 0.00007463, -0.00088941, 0.00189741, -0.00079230, 0.00001185, -0.00012801, 0.00015382, -0.00011511, },
		{ -0.00010068, -0.00006950, 0.00007683, -0.00009103, 0.00017434, -0.00067014, 0.00030513, 0.00143935, -0.00136087, 0.00073032, -0.00057870, 0.00028977, -0.00016803, 0.00029376, -0.00035450, -0.00007193, 0.00014206, -0.00001715, 0.00015342, -0.00051307, 0.00030877, -0.00019201, 0.00036538, -0.00058466, 0.00063097, -0.00090029, 0.00044215, 0.00108822, -0.00090078, 0.00022678, -0.00014936, 0.00014176, },
		{ -0.00004575, 0.00001525, -0.00016660, 0.00022206, -0.00025983, 0.00029462, -0.00113872, 0.00169604, -0.00003507, -0.00066235, 0.00048380, -0.00053458, 0.00035969, -0.00019719, 0.00029892, -0.00055154, 0.00026808, -0.00010534, 0.00023104, -0.00023271, -0.00020252, 0.00017708, -0.00002626, 0.00011104, -0.00037076, 0.00055184, -0.00119528, 0.00150419, -0.00014030, -0.00035171, 0.00002961, 0.00005959, },
		{ 0.00015077, -0.00011115, -0.00001483, -0.00003839, 0.00005125, -0.00003424, -0.00020876, -0.00058648, 0.00214977, -0.00130385, 0.00040339, -0.00027521, 0.00000956, 0.00001268, 0.00011537, -0.00007443, -0.00031982, 0.00027321, -0.00013046, 0.00027213, -0.00049171, 0.00015613, -0.00007967, 0.00026666, -0.00036313, 0.00026669, -0.00023824, -0.00043398, 0.00146536, -0.00078328, 0.00009620, -0.00011523, },
		{ -0.00000703, 0.00001471, 0.00000052, -0.00016755, 0.00020945, -0.00027222, 0.00035020, -0.00103716, 0.00092149, 0.00111992, -0.00133878, 0.00077860, -0.00068655, 0.00040437, -0.00026582, 0.00036465, -0.00046888, 0.00010185, 0.00002522, 0.00007954, 0.00001047, -0.00039454, 0.00023907, -0.00011219, 0.00027669, -0.00051421, 0.00064197, -0.00097450, 0.00074250, 0.00053826, -0.00060736, 0.00011364, },
		{ -0.00003589, 0.00012744, -0.00011036, 0.00004891, -0.00016220, 0.00021045, -0.00027948, 0.00025175, -0.00122580, 0.00223769, -0.00060025, -0.00030575, 0.00022590, -0.00037336, 0.00026051, -0.00012573, 0.00020476, -0.00048669, 0.00031079, -0.00015016, 0.00025636, -0.00032110, -0.00007323, 0.00006364, 0.00011120, -0.00007894, -0.00012028, 0.00029996, -0.00089813, 0.00133655, -0.00037263, -0.00013963, },
		{ 0.00007795, -0.00002481, 0.00005825, -0.00003501, -0.00010143, 0.00009736, -0.00013604, 0.00015530, -0.00055611, -0.00014643, 0.00213921, -0.00157715, 0.00066538, -0.00052742, 0.00022075, -0.00015074, 0.00024011, -0.00023422, -0.00012889, 0.00017349, -0.00006601, 0.00017432, -0.00042345, 0.00015012, -0.00007656, 0.00027687, -0.00042441, 0.00042723, -0.00046112, -0.00004214, 0.00096866, -0.00062672, },
		{ -0.00034633, 0.00002179, 0.00010171, -0.00010180, 0.00009105, -0.00023363, 0.00029583, -0.00041586, 0.00048580, -0.00136148, 0.00161158, 0.00058703, -0.00112175, 0.00067551, -0.00066365, 0.00042198, -0.00028849, 0.00035777, -0.00049913, 0.00022968, -0.00007507, 0.00014077, -0.00009828, -0.00024978, 0.00012495, 0.00001210, 0.00012256, -0.00034863, 0.00053160, -0.00088704, 0.00082780, 0.00013773, },
		{ -0.00040875, -0.00004611, -0.00001154, 0.00009274, -0.00009307, 0.00001479, -0.00007665, 0.00009055, -0.00017713, 0.00005883, -0.00108920, 0.00258862, -0.00113674, 0.00008983, -0.00008349, -0.00014259, 0.00009765, -0.00000355, 0.00005328, -0.00033312, 0.00027110, -0.00014231, 0.00020965, -0.00030036, -0.00001888, 0.00000162, 0.00019507, -0.00022396, 0.00010634, 0.00004848, -0.00055662, 0.00101183, },
		{ 0.00052106, -0.00041209, 0.00001650, 0.00008739, -0.00007795, 0.00007900, -0.00020575, 0.00024639, -0.00035504, 0.00039088, -0.00097354, 0.00048608, 0.00183987, -0.00164241, 0.00079462, -0.00067964, 0.00037195, -0.00027308, 0.00032104, -0.00035076, 0.00005776, 0.00005297, -0.00000231, 0.00006900, -0.00028280, 0.00006259, 0.00000174, 0.00019516, -0.00037644, 0.00046793, -0.00055052, 0.00020672, },
		{ 0.00071797, -0.00007454, -0.00017347, -0.00000415, 0.00012763, -0.00015824, 0.00013082, -0.00023132, 0.00029063, -0.00046034, 0.00051011, -0.00154323, 0.00224467, -0.00006334, -0.00075416, 0.00044293, -0.00051601, 0.00033970, -0.00023204, 0.00026984, -0.00042921, 0.00027470, -0.00012704, 0.00013492, -0.00011451, -0.00014474, 0.00001948, 0.00013216, -0.00004710, -0.00013967, 0.00034851, -0.00068465, },
		{ -0.00026229, 0.00063399, -0.00031126, -0.00004173, 0.00007515, -0.00002688, -0.00000101, -0.00007686, 0.00007154, -0.00011781, 0.00003794, -0.00027965, -0.00070479, 0.00265815, -0.00153189, 0.00043349, -0.00036876, 0.00009671, -0.00008055, 0.00012558, -0.00010669, -0.00013711, 0.00017284, -0.00010299, 0.00011875, -0.00017267, -0.00006411, 0.00002035, 0.00019216, -0.00028233, 0.00025424, -0.00013520, },
		{ -0.00050548, 0.00027952, 0.00020476, -0.00022146, -0.00002737, 0.00015990, -0.00019945, 0.00018392, -0.00028280, 0.00035354, -0.00053663, 0.00059818, -0.00136933, 0.00121272, 0.00129325, -0.00148517, 0.00076200, -0.00069547, 0.00042519, -0.00032093, 0.00032746, -0.00038289, 0.00018450, -0.00004273, 0.00001613, 0.00002264, -0.00013310, -0.00006595, 0.00012440, 0.00004819, -0.00024038, 0.00039915, },
		{ 0.00015820, -0.00044253, 0.00048117, -0.00011565, -0.00010107, 0.00004456, 0.00005415, -0.00012949, 0.00009991, -0.00014677, 0.00017573, -0.00037168, 0.00037750, -0.00150429, 0.00268958, -0.00070490, -0.00031920, 0.00013972, -0.00028448, 0.00018482, -0.00012000, 0.00012743, -0.00027919, 0.00023460, -0.00012685, 0.00006085, -0.00000755, -0.00013386, -0.00002550, 0.00019722, -0.00017736, 0.00005123, },
		{ 0.00029643, -0.00021077, -0.00008375, 0.00030423, -0.00016313, -0.00007261, 0.00017144, -0.00018410, 0.00013146, -0.00019250, 0.00023936, -0.00036303, 0.00031662, -0.00071117, -0.00011036, 0.00241566, -0.00170772, 0.00065004, -0.00056071, 0.00028292, -0.00022339, 0.00021612, -0.00022228, 0.00004084, 0.00006197, -0.00007547, 0.00004242, 0.00001766, -0.00019573, 0.00011856, 0.00009740, -0.00024011, },
		{ -0.00006326, 0.00026314, -0.00036552, 0.00019579, 0.00005185, -0.00010600, -0.00001404, 0.00014824, -0.00025592, 0.00023722, -0.00029442, 0.00037269, -0.00061746, 0.00070176, -0.00164144, 0.00190057, 0.00059907, -0.00114675, 0.00058393, -0.00057686, 0.00037293, -0.00028721, 0.00025590, -0.00031951, 0.00022068, -0.00008420, -0.00003917, 0.00009281, -0.00004754, -0.00017618, 0.00024055, -0.00011532, },
		{ -0.00022604, 0.00017231, 0.00002136, -0.00023965, 0.00021007, -0.00004495, -0.00010199, 0.00014077, -0.00009801, -0.00002247, 0.00000566, 0.00000521, -0.00003204, -0.00015621, 0.00008576, -0.00121450, 0.00284971, -0.00121450, 0.00008576, -0.00015621, -0.00003204, 0.00000521, 0.00000566, -0.00002247, -0.00009801, 0.00014077, -0.00010199, -0.00004495, 0.00021007, -0.00023965, 0.00002136, 0.00017231, },
		{ -0.00006326, -0.00011532, 0.00024055, -0.00017618, -0.00004754, 0.00009281, -0.00003917, -0.00008420, 0.00022068, -0.00031951, 0.00025590, -0.00028721, 0.00037293, -0.00057686, 0.00058393, -0.00114675, 0.00059907, 0.00190057, -0.00164144, 0.00070176, -0.00061746, 0.00037269, -0.00029442, 0.00023722, -0.00025592, 0.00014824, -0.00001404, -0.00010600, 0.00005185, 0.00019579, -0.00036552, 0.00026314, },
		{ 0.00029643, -0.00024011, 0.00009740, 0.00011856, -0.00019573, 0.00001766, 0.00004242, -0.00007547, 0.00006197, 0.00004084, -0.00022228, 0.00021612, -0.00022339, 0.00028292, -0.00056071, 0.00065004, -0.00170772, 0.00241566, -0.00011036, -0.00071117, 0.00031662, -0.00036303, 0.00023936, -0.00019250, 0.00013146, -0.00018410, 0.00017144, -0.00007261, -0.00016313, 0.00030423, -0.00008375, -0.00021077, },
		{ 0.00015820, 0.00005123, -0.00017736, 0.00019722, -0.00002550, -0.00013386, -0.00000755, 0.00006085, -0.00012685, 0.00023460, -0.00027919, 0.00012743, -0.00012000, 0.00018482, -0.00028448, 0.00013972, -0.00031920, -0.00070490, 0.00268958, -0.00150429, 0.00037750, -0.00037168, 0.00017573, -0.00014677, 0.00009991, -0.00012949, 0.00005415, 0.00004456, -0.00010107, -0.00011565, 0.00048117, -0.00044253, },
		{ -0.00050548, 0.00039915, -0.00024038, 0.00004819, 0.00012440, -0.00006595, -0.00013310, 0.00002264, 0.00001613, -0.00004273, 0.00018450, -0.00038289, 0.00032746, -0.00032093, 0.00042519, -0.00069547, 0.00076200, -0.00148517, 0.00129325, 0.00121272, -0.00136933, 0.00059818, -0.00053663, 0.00035354, -0.00028280, 0.00018392, -0.00019945, 0.00015990, -0.00002737, -0.00022146, 0.00020476, 0.00027952, },
		{ -0.00026229, -0.00013520, 0.00025424, -0.00028233, 0.00019216, 0.00002035, -0.00006411, -0.00017267, 0.00011875, -0.00010299, 0.00017284, -0.00013711, -0.00010669, 0.00012558, -0.00008055, 0.00009671, -0.00036876, 0.00043349, -0.00153189, 0.00265815, -0.00070479, -0.00027965, 0.00003794, -0.00011781, 0.00007154, -0.00007686, -0.00000101, -0.00002688, 0.00007515, -0.00004173, -0.00031126, 0.00063399, },
		{ 0.00071797, -0.00068465, 0.00034851, -0.00013967, -0.00004710, 0.00013216, 0.00001948, -0.00014474, -0.00011451, 0.00013492, -0.00012704, 0.00027470, -0.00042921, 0.00026984, -0.00023204, 0.00033970, -0.00051601, 0.00044293, -0.00075416, -0.00006334, 0.00224467, -0.00154323, 0.00051011, -0.00046034, 0.00029063, -0.00023132, 0.00013082, -0.00015824, 0.00012763, -0.00000415, -0.00017347, -0.00007454, },
		{ 0.00052106, 0.00020672, -0.00055052, 0.00046793, -0.00037644, 0.00019516, 0.00000174, 0.00006259, -0.00028280, 0.00006900, -0.00000231, 0.00005297, 0.00005776, -0.00035076, 0.00032104, -0.00027308, 0.00037195, -0.00067964, 0.00079462, -0.00164241, 0.00183987, 0.00048608, -0.00097354, 0.00039088, -0.00035504, 0.00024639, -0.00020575, 0.00007900, -0.00007795, 0.00008739, 0.00001650, -0.00041209, },
		{ -0.00040875, 0.00101183, -0.00055662, 0.00004848, 0.00010634, -0.00022396, 0.00019507, 0.00000162, -0.00001888, -0.00030036, 0.00020965, -0.00014231, 0.00027110, -0.00033312, 0.00005328, -0.00000355, 0.00009765, -0.00014259, -0.00008349, 0.00008983, -0.00113674, 0.00258862, -0.00108920, 0.00005883, -0.00017713, 0.00009055, -0.00007665, 0.00001479, -0.00009307, 0.00009274, -0.00001154, -0.00004611, },
		{ -0.00034633, 0.00013773, 0.00082780, -0.00088704, 0.00053160, -0.00034863, 0.00012256, 0.00001210, 0.00012495, -0.00024978, -0.00009828, 0.00014077, -0.00007507, 0.00022968, -0.00049913, 0.00035777, -0.00028849, 0.00042198, -0.00066365, 0.00067551, -0.00112175, 0.00058703, 0.00161158, -0.00136148, 0.00048580, -0.00041586, 0.00029583, -0.00023363, 0.00009105, -0.00010180, 0.00010171, 0.00002179, },
		{ 0.00007795, -0.00062672, 0.00096866, -0.00004214, -0.00046112, 0.00042723, -0.00042441, 0.00027687, -0.00007656, 0.00015012, -0.00042345, 0.00017432, -0.00006601, 0.00017349, -0.00012889, -0.00023422, 0.00024011, -0.00015074, 0.00022075, -0.00052742, 0.00066538, -0.00157715, 0.00213921, -0.00014643, -0.00055611, 0.00015530, -0.00013604, 0.00009736, -0.00010143, -0.00003501, 0.00005825, -0.00002481, },
		{ -0.00003589, -0.00013963, -0.00037263, 0.00133655, -0.00089813, 0.00029996, -0.00012028, -0.00007894, 0.00011120, 0.00006364, -0.00007323, -0.00032110, 0.00025636, -0.00015016, 0.00031079, -0.00048669, 0.00020476, -0.00012573, 0.00026051, -0.00037336, 0.00022590, -0.00030575, -0.00060025, 0.00223769, -0.00122580, 0.00025175, -0.00027948, 0.00021045, -0.00016220, 0.00004891, -0.00011036, 0.00012744, },
		{ -0.00000703, 0.00011364, -0.00060736, 0.00053826, 0.00074250, -0.00097450, 0.00064197, -0.00051421, 0.00027669, -0.00011219, 0.00023907, -0.00039454, 0.00001047, 0.00007954, 0.00002522, 0.00010185, -0.00046888, 0.00036465, -0.00026582, 0.00040437, -0.00068655, 0.00077860, -0.00133878, 0.00111992, 0.00092149, -0.00103716, 0.00035020, -0.00027222, 0.00020945, -0.00016755, 0.00000052, 0.00001471, },
		{ 0.00015077, -0.00011523, 0.00009620, -0.00078328, 0.00146536, -0.00043398, -0.00023824, 0.00026669, -0.00036313, 0.00026666, -0.00007967, 0.00015613, -0.00049171, 0.00027213, -0.00013046, 0.00027321, -0.00031982, -0.00007443, 0.00011537, 0.00001268, 0.00000956, -0.00027521, 0.00040339, -0.00130385, 0.00214977, -0.00058648, -0.00020876, -0.00003424, 0.00005125, -0.00003839, -0.00001483, -0.00011115, },
		{ -0.00004575, 0.00005959, 0.00002961, -0.00035171, -0.00014030, 0.00150419, -0.00119528, 0.00055184, -0.00037076, 0.00011104, -0.00002626, 0.00017708, -0.00020252, -0.00023271, 0.00023104, -0.00010534, 0.00026808, -0.00055154, 0.00029892, -0.00019719, 0.00035969, -0.00053458, 0.00048380, -0.00066235, -0.00003507, 0.00169604, -0.00113872, 0.00029462, -0.00025983, 0.00022206, -0.00016660, 0.00001525, },
		{ -0.00010068, 0.00014176, -0.00014936, 0.00022678, -0.00090078, 0.00108822, 0.00044215, -0.00090029, 0.00063097, -0.00058466, 0.00036538, -0.00019201, 0.00030877, -0.00051307, 0.00015342, -0.00001715, 0.00014206, -0.00007193, -0.00035450, 0.00029376, -0.00016803, 0.00028977, -0.00057870, 0.00073032, -0.00136087, 0.00143935, 0.00030513, -0.00067014, 0.00017434, -0.00009103, 0.00007683, -0.00006950, },
		},
		{ { -0.00001025, 0.00004553, -0.00009959, -0.00014347, -0.00071997, 0.00185216, -0.00085473, 0.00024335, -0.00009801, -0.00013104, 0.00004693, -0.00003770, 0.00003483, 0.00000689, 0.00022765, -0.00043308, 0.00021705, -0.00043308, 0.00022765, 0.00000689, 0.00003483, -0.00003770, 0.00004693, -0.00013104, -0.00009801, 0.00024335, -0.00085473, 0.00185216, -0.00071997, -0.00014347, -0.00009959, 0.00004553, },
		{ 0.00006273, 0.00004886, -0.00010375, 0.00010922, -0.00069258, 0.00019261, 0.00150385, -0.00121186, 0.00071373, -0.00056121, 0.00025552, -0.00022929, 0.00019559, -0.00018403, 0.00028957, -0.00011417, -0.00021933, 0.00009432, -0.00040263, 0.00039345, -0.00020323, 0.00022091, -0.00023776, 0.00028351, -0.00050947, 0.00047088, -0.00052265, 0.00012164, 0.00132204, -0.00096422, 0.00012609, -0.00016203, },
		{ -0.00006745, -0.00001104, 0.00013615, -0.00023661, 0.00022069, -0.00113405, 0.00137688, 0.00041161, -0.00077926, 0.00057762, -0.00058335, 0.00034592, -0.00029501, 0.00024961, -0.00022799, 0.00041541, -0.00040247, 0.00006508, -0.00017660, -0.00013641, 0.00029205, -0.00019162, 0.00020900, -0.00022027, 0.00024429, -0.00054654, 0.00069188, -0.00101608, 0.00111322, 0.00033431, -0.00065185, 0.00007918, },
		{ -0.00006229, 0.00002410, -0.00002325, 0.00008899, -0.00016245, -0.00003994, -0.00097158, 0.00212264, -0.00078058, 0.00006152, 0.00000027, -0.00019978, 0.00010901, -0.00010323, 0.00007735, -0.00003334, 0.00026354, -0.00044666, 0.00020237, -0.00038181, 0.00018277, 0.00003307, -0.00000246, 0.00001882, -0.00000014, -0.00007464, -0.00016834, 0.00036064, -0.00084350, 0.00152292, -0.00048661, -0.00020107, },
		{ 0.00004179, -0.00009496, -0.00000128, 0.00009027, -0.00012071, 0.00011272, -0.00066608, -0.00002278, 0.00189715, -0.00135817, 0.00067673, -0.00053742, 0.00024349, -0.00021324, 0.00016187, -0.00016026, 0.00025482, -0.00007374, -0.00024240, 0.00010120, -0.00036038, 0.00034200, -0.00017194, 0.00018054, -0.00017060, 0.00021584, -0.00042226, 0.00035674, -0.00029913, -0.00008767, 0.00117523, -0.00076102, },
		{ -0.00055277, 0.00003093, -0.00001793, -0.00008084, 0.00020922, -0.00030963, 0.00031135, -0.00124779, 0.00132797, 0.00076381, -0.00104202, 0.00066002, -0.00063448, 0.00039434, -0.00033378, 0.00026128, -0.00023985, 0.00040006, -0.00036162, 0.00003173, -0.00013783, -0.00012752, 0.00026005, -0.00018188, 0.00020042, -0.00018798, 0.00021229, -0.00049600, 0.00062158, -0.00079885, 0.00075942, 0.00039262, },
		{ -0.00029526, -0.00020768, -0.00005162, 0.00002611, -0.00006027, 0.00014975, -0.00024604, 0.00009800, -0.00121337, 0.00230355, -0.00061675, -0.00017091, 0.00011370, -0.00027266, 0.00017766, -0.00016790, 0.00011721, -0.00007734, 0.00027441, -0.00041061, 0.00016337, -0.00031206, 0.00014580, 0.00003668, -0.00002742, 0.00005791, -0.00002624, -0.00003664, -0.00020201, 0.00040859, -0.00076189, 0.00117027, },
		{ 0.00095862, -0.00054966, -0.00001993, -0.00004309, -0.00004762, 0.00010740, -0.00011495, 0.00008485, -0.00057174, -0.00029184, 0.00225547, -0.00143196, 0.00057799, -0.00047752, 0.00020903, -0.00017424, 0.00011291, -0.00012444, 0.00019515, -0.00003264, -0.00022407, 0.00007539, -0.00028311, 0.00027369, -0.00014223, 0.00013594, -0.00010365, 0.00015210, -0.00033238, 0.00025231, -0.00012969, -0.00020977, },
		{ 0.00046881, 0.00036917, -0.00041104, -0.00002503, 0.00003700, -0.00015049, 0.00027209, -0.00037191, 0.00038393, -0.00128858, 0.00117579, 0.00114968, -0.00128163, 0.00070388, -0.00065599, 0.00042183, -0.00034832, 0.00025348, -0.00023520, 0.00034577, -0.00029056, 0.00001708, -0.00012428, -0.00008149, 0.00019691, -0.00015683, 0.00017178, -0.00013780, 0.00016491, -0.00041666, 0.00051897, -0.00058894, },
		{ -0.00063281, 0.00082930, -0.00016219, -0.00016526, -0.00006483, 0.00004948, -0.00011561, 0.00022038, -0.00034224, 0.00025403, -0.00141684, 0.00237294, -0.00037094, -0.00043250, 0.00022728, -0.00034049, 0.00024323, -0.00022184, 0.00014693, -0.00011478, 0.00024794, -0.00032244, 0.00011433, -0.00024692, 0.00013159, 0.00001243, -0.00003494, 0.00007390, -0.00002877, -0.00002017, -0.00019561, 0.00039173, },
		{ -0.00002660, -0.00025004, 0.00070137, -0.00035420, -0.00005070, -0.00001165, -0.00007073, 0.00009811, -0.00008632, 0.00002526, -0.00041748, -0.00058678, 0.00254247, -0.00141864, 0.00042292, -0.00038850, 0.00015620, -0.00011800, 0.00005641, -0.00008305, 0.00012187, -0.00000983, -0.00014852, 0.00001639, -0.00018618, 0.00020387, -0.00012253, 0.00009542, -0.00004697, 0.00010063, -0.00025017, 0.00017227, },
		{ 0.00040316, -0.00041165, 0.00025704, 0.00027629, -0.00024677, -0.00007782, 0.00008894, -0.00020993, 0.00031630, -0.00041548, 0.00042664, -0.00124789, 0.00093333, 0.00153401, -0.00147046, 0.00069999, -0.00064463, 0.00042426, -0.00033659, 0.00022710, -0.00021385, 0.00025750, -0.00020681, 0.00004441, -0.00015105, -0.00000037, 0.00011108, -0.00012135, 0.00012834, -0.00007791, 0.00010962, -0.00031912, },
		{ -0.00015149, 0.00032278, -0.00048321, 0.00053024, -0.00009424, -0.00008240, -0.00009600, 0.00008907, -0.00018078, 0.00029185, -0.00044077, 0.00040930, -0.00155707, 0.00232085, -0.00006226, -0.00069702, 0.00032602, -0.00039446, 0.00029623, -0.00025669, 0.00016067, -0.00013704, 0.00017784, -0.00018980, 0.00007627, -0.00020938, 0.00015026, -0.00004076, -0.00002265, 0.00006475, -0.00000879, -0.00002500, },
		{ -0.00018421, 0.00012667, 0.00000787, -0.00022403, 0.00043536, -0.00019424, -0.00004720, -0.00000269, -0.00006855, 0.00006418, -0.00003797, -0.00006175, -0.00021883, -0.00087577, 0.00272800, -0.00131446, 0.00022482, -0.00028098, 0.00009174, -0.00005293, 0.00000146, -0.00004392, 0.00004844, -0.00002237, -0.00000898, -0.00006623, -0.00009040, 0.00014807, -0.00012019, 0.00006687, -0.00000914, 0.00006767, },
		{ 0.00005484, -0.00021584, 0.00029327, -0.00028520, 0.00012777, 0.00013477, -0.00008144, -0.00011791, 0.00013025, -0.00025065, 0.00033600, -0.00043477, 0.00043180, -0.00112699, 0.00062448, 0.00187931, -0.00158648, 0.00064564, -0.00060120, 0.00040096, -0.00030051, 0.00018621, -0.00017905, 0.00014616, -0.00013201, 0.00013307, -0.00022425, 0.00010578, 0.00001488, -0.00008236, 0.00007780, -0.00001798, },
		{ 0.00002916, -0.00004773, -0.00007737, 0.00022036, -0.00033968, 0.00029465, -0.00008822, 0.00002705, -0.00013703, 0.00013746, -0.00024575, 0.00035481, -0.00053054, 0.00054513, -0.00161630, 0.00215077, 0.00028089, -0.00093691, 0.00039712, -0.00042738, 0.00032883, -0.00026693, 0.00015516, -0.00013858, 0.00006517, -0.00002953, 0.00007319, -0.00021834, 0.00020541, -0.00011942, 0.00000859, 0.00003231, },
		{ 0.00000414, 0.00005646, -0.00014020, 0.00011946, -0.00001866, -0.00015430, 0.00019058, -0.00008229, -0.00001169, -0.00001490, -0.00004277, 0.00001107, 0.00002406, -0.00016770, 0.00000318, -0.00112729, 0.00279217, -0.00112729, 0.00000318, -0.00016770, 0.00002406, 0.00001107, -0.00004277, -0.00001490, -0.00001169, -0.00008229, 0.00019058, -0.00015430, -0.00001866, 0.00011946, -0.00014020, 0.00005646, },
		{ 0.00002916, 0.00003231, 0.00000859, -0.00011942, 0.00020541, -0.00021834, 0.00007319, -0.00002953, 0.00006517, -0.00013858, 0.00015516, -0.00026693, 0.00032883, -0.00042738, 0.00039712, -0.00093691, 0.00028089, 0.00215077, -0.00161630, 0.00054513, -0.00053054, 0.00035481, -0.00024575, 0.00013746, -0.00013703, 0.00002705, -0.00008822, 0.00029465, -0.00033968, 0.00022036, -0.00007737, -0.00004773, },
		{ 0.00005484, -0.00001798, 0.00007780, -0.00008236, 0.00001488, 0.00010578, -0.00022425, 0.00013307, -0.00013201, 0.00014616, -0.00017905, 0.00018621, -0.00030051, 0.00040096, -0.00060120, 0.00064564, -0.00158648, 0.00187931, 0.00062448, -0.00112699, 0.00043180, -0.00043477, 0.00033600, -0.00025065, 0.00013025, -0.00011791, -0.00008144, 0.00013477, 0.00012777, -0.00028520, 0.00029327, -0.00021584, },
		{ -0.00018421, 0.00006767, -0.00000914, 0.00006687, -0.00012019, 0.00014807, -0.00009040, -0.00006623, -0.00000898, -0.00002237, 0.00004844, -0.00004392, 0.00000146, -0.00005293, 0.00009174, -0.00028098, 0.00022482, -0.00131446, 0.00272800, -0.00087577, -0.00021883, -0.00006175, -0.00003797, 0.00006418, -0.00006855, -0.00000269, -0.00004720, -0.00019424, 0.00043536, -0.00022403, 0.00000787, 0.00012667, },
		{ -0.00015149, -0.00002500, -0.00000879, 0.00006475, -0.00002265, -0.00004076, 0.00015026, -0.00020938, 0.00007627, -0.00018980, 0.00017784, -0.00013704, 0.00016067, -0.00025669, 0.00029623, -0.00039446, 0.00032602, -0.00069702, -0.00006226, 0.00232085, -0.00155707, 0.00040930, -0.00044077, 0.00029185, -0.00018078, 0.00008907, -0.00009600, -0.00008240, -0.00009424, 0.00053024, -0.00048321, 0.00032278, },
		{ 0.00040316, -0.00031912, 0.00010962, -0.00007791, 0.00012834, -0.00012135, 0.00011108, -0.00000037, -0.00015105, 0.00004441, -0.00020681, 0.00025750, -0.00021385, 0.00022710, -0.00033659, 0.00042426, -0.00064463, 0.00069999, -0.00147046, 0.00153401, 0.00093333, -0.00124789, 0.00042664, -0.00041548, 0.00031630, -0.00020993, 0.00008894, -0.00007782, -0.00024677, 0.00027629, 0.00025704, -0.00041165, },
		{ -0.00002660, 0.00017227, -0.00025017, 0.00010063, -0.00004697, 0.00009542, -0.00012253, 0.00020387, -0.00018618, 0.00001639, -0.00014852, -0.00000983, 0.00012187, -0.00008305, 0.00005641, -0.00011800, 0.00015620, -0.00038850, 0.00042292, -0.00141864, 0.00254247, -0.00058678, -0.00041748, 0.00002526, -0.00008632, 0.00009811, -0.00007073, -0.00001165, -0.00005070, -0.00035420, 0.00070137, -0.00025004, },
		{ -0.00063281, 0.00039173, -0.00019561, -0.00002017, -0.00002877, 0.00007390, -0.00003494, 0.00001243, 0.00013159, -0.00024692, 0.00011433, -0.00032244, 0.00024794, -0.00011478, 0.00014693, -0.00022184, 0.00024323, -0.00034049, 0.00022728, -0.00043250, -0.00037094, 0.00237294, -0.00141684, 0.00025403, -0.00034224, 0.00022038, -0.00011561, 0.00004948, -0.00006483, -0.00016526, -0.00016219, 0.00082930, },
		{ 0.00046881, -0.00058894, 0.00051897, -0.00041666, 0.00016491, -0.00013780, 0.00017178, -0.00015683, 0.00019691, -0.00008149, -0.00012428, 0.00001708, -0.00029056, 0.00034577, -0.00023520, 0.00025348, -0.00034832, 0.00042183, -0.00065599, 0.00070388, -0.00128163, 0.00114968, 0.00117579, -0.00128858, 0.00038393, -0.00037191, 0.00027209, -0.00015049, 0.00003700, -0.00002503, -0.00041104, 0.00036917, },
		{ 0.00095862, -0.00020977, -0.00012969, 0.00025231, -0.00033238, 0.00015210, -0.00010365, 0.00013594, -0.00014223, 0.00027369, -0.00028311, 0.00007539, -0.00022407, -0.00003264, 0.00019515, -0.00012444, 0.00011291, -0.00017424, 0.00020903, -0.00047752, 0.00057799, -0.00143196, 0.00225547, -0.00029184, -0.00057174, 0.00008485, -0.00011495, 0.00010740, -0.00004762, -0.00004309, -0.00001993, -0.00054966, },
		{ -0.00029526, 0.00117027, -0.00076189, 0.00040859, -0.00020201, -0.00003664, -0.00002624, 0.00005791, -0.00002742, 0.00003668, 0.00014580, -0.00031206, 0.00016337, -0.00041061, 0.00027441, -0.00007734, 0.00011721, -0.00016790, 0.00017766, -0.00027266, 0.00011370, -0.00017091, -0.00061675, 0.00230355, -0.00121337, 0.00009800, -0.00024604, 0.00014975, -0.00006027, 0.00002611, -0.00005162, -0.00020768, },
		{ -0.00055277, 0.00039262, 0.00075942, -0.00079885, 0.00062158, -0.00049600, 0.00021229, -0.00018798, 0.00020042, -0.00018188, 0.00026005, -0.00012752, -0.00013783, 0.00003173, -0.00036162, 0.00040006, -0.00023985, 0.00026128, -0.00033378, 0.00039434, -0.00063448, 0.00066002, -0.00104202, 0.00076381, 0.00132797, -0.00124779, 0.00031135, -0.00030963, 0.00020922, -0.00008084, -0.00001793, 0.00003093, },
		{ 0.00004179, -0.00076102, 0.00117523, -0.00008767, -0.00029913, 0.00035674, -0.00042226, 0.00021584, -0.00017060, 0.00018054, -0.00017194, 0.00034200, -0.00036038, 0.00010120, -0.00024240, -0.00007374, 0.00025482, -0.00016026, 0.00016187, -0.00021324, 0.00024349, -0.00053742, 0.00067673, -0.00135817, 0.00189715, -0.00002278, -0.00066608, 0.00011272, -0.00012071, 0.00009027, -0.00000128, -0.00009496, },
		{ -0.00006229, -0.00020107, -0.00048661, 0.00152292, -0.00084350, 0.00036064, -0.00016834, -0.00007464, -0.00000014, 0.00001882, -0.00000246, 0.00003307, 0.00018277, -0.00038181, 0.00020237, -0.00044666, 0.00026354, -0.00003334, 0.00007735, -0.00010323, 0.00010901, -0.00019978, 0.00000027, 0.00006152, -0.00078058, 0.00212264, -0.00097158, -0.00003994, -0.00016245, 0.00008899, -0.00002325, 0.00002410, },
		{ -0.00006745, 0.00007918, -0.00065185, 0.00033431, 0.00111322, -0.00101608, 0.00069188, -0.00054654, 0.00024429, -0.00022027, 0.00020900, -0.00019162, 0.00029205, -0.00013641, -0.00017660, 0.00006508, -0.00040247, 0.00041541, -0.00022799, 0.00024961, -0.00029501, 0.00034592, -0.00058335, 0.00057762, -0.00077926, 0.00041161, 0.00137688, -0.00113405, 0.00022069, -0.00023661, 0.00013615, -0.00001104, },
		{ 0.00006273, -0.00016203, 0.00012609, -0.00096422, 0.00132204, 0.00012164, -0.00052265, 0.00047088, -0.00050947, 0.00028351, -0.00023776, 0.00022091, -0.00020323, 0.00039345, -0.00040263, 0.00009432, -0.00021933, -0.00011417, 0.00028957, -0.00018403, 0.00019559, -0.00022929, 0.00025552, -0.00056121, 0.00071373, -0.00121186, 0.00150385, 0.00019261, -0.00069258, 0.00010922, -0.00010375, 0.00004886, },
		},
		{ { -0.00006779, -0.00003788, -0.00002361, -0.00073363, 0.00177544, -0.00079820, 0.00012260, -0.00015879, -0.00008362, 0.00018512, -0.00011721, 0.00009211, -0.00021634, 0.00002754, 0.00002901, 0.00006422, -0.00009944, 0.00006422, 0.00002901, 0.00002754, -0.00021634, 0.00009211, -0.00011721, 0.00018512, -0.00008362, -0.00015879, 0.00012260, -0.00079820, 0.00177544, -0.00073363, -0.00002361, -0.00003788, },
		{ -0.00010094, -0.00011397, 0.00013211, -0.00045124, -0.00010547, 0.00163833, -0.00119155, 0.00053257, -0.00052390, 0.00026830, -0.00007592, 0.00006392, -0.00006810, -0.00013100, 0.00000490, 0.00008005, -0.00000335, -0.00007222, 0.00009120, -0.00000134, 0.00001745, -0.00024032, 0.00019045, -0.00024273, 0.00039730, -0.00043201, 0.00027357, -0.00045017, -0.00003946, 0.00135698, -0.00090826, 0.00019114, },
		{ 0.00018362, -0.00006432, -0.00020025, 0.00029694, -0.00091390, 0.00081941, 0.00095077, -0.00107793, 0.00059144, -0.00063644, 0.00045333, -0.00024376, 0.00018810, -0.00020285, -0.00004279, -0.00001177, 0.00011508, -0.00007817, -0.00000569, 0.00007068, 0.00000500, -0.00004171, -0.00017307, 0.00016903, -0.00022381, 0.00043051, -0.00059582, 0.00054246, -0.00090051, 0.00077172, 0.00061166, -0.00070064, },
		{ -0.00033108, 0.00007221, -0.00002300, -0.00021195, 0.00028417, -0.00113110, 0.00169631, -0.00003237, -0.00055481, 0.00029962, -0.00045161, 0.00040102, -0.00024292, 0.00018490, -0.00024751, -0.00000030, 0.00000988, 0.00010028, -0.00011868, 0.00006349, 0.00002396, 0.00003413, -0.00012958, -0.00003648, 0.00003718, -0.00005114, 0.00024338, -0.00048344, 0.00049581, -0.00099239, 0.00130476, -0.00012644, },
		{ -0.00058526, -0.00002703, -0.00000474, -0.00006810, -0.00006906, -0.00001918, -0.00086736, 0.00214328, -0.00091809, 0.00009570, -0.00016406, -0.00008053, 0.00017048, -0.00010057, 0.00006352, -0.00019560, -0.00001241, 0.00007136, 0.00002831, -0.00009914, 0.00009817, -0.00001436, 0.00005287, -0.00019881, 0.00010360, -0.00012923, 0.00019316, -0.00008671, -0.00014691, 0.00014489, -0.00066119, 0.00136933, },
		{ 0.00099560, -0.00067159, 0.00009668, 0.00001109, -0.00020328, 0.00019024, -0.00056023, -0.00008288, 0.00192926, -0.00136043, 0.00055372, -0.00054581, 0.00027890, -0.00009301, 0.00007826, -0.00009843, -0.00009728, -0.00004559, 0.00013898, -0.00007290, -0.00002571, 0.00008194, -0.00001868, 0.00003209, -0.00020359, 0.00017316, -0.00022836, 0.00037674, -0.00039981, 0.00024179, -0.00033818, -0.00004636, },
		{ 0.00058189, 0.00039928, -0.00048161, 0.00007160, 0.00006893, -0.00033002, 0.00040330, -0.00110203, 0.00100267, 0.00110028, -0.00122391, 0.00062213, -0.00065415, 0.00045358, -0.00024597, 0.00017878, -0.00020596, -0.00002335, -0.00004926, 0.00016343, -0.00015281, 0.00006548, 0.00002800, 0.00001159, -0.00003204, -0.00012964, 0.00013348, -0.00019083, 0.00038960, -0.00053746, 0.00048100, -0.00070965, },
		{ -0.00078163, 0.00096075, -0.00014279, -0.00021184, 0.00001575, 0.00006360, -0.00031516, 0.00037010, -0.00132053, 0.00198118, -0.00002756, -0.00064059, 0.00031603, -0.00045533, 0.00038523, -0.00022681, 0.00014572, -0.00020548, -0.00001826, 0.00000707, 0.00011170, -0.00016583, 0.00012754, -0.00002923, 0.00005199, -0.00011246, -0.00000698, 0.00000654, -0.00002252, 0.00020765, -0.00042850, 0.00044696, },
		{ 0.00015951, -0.00051851, 0.00096248, -0.00043415, -0.00003153, 0.00002925, -0.00006900, -0.00009724, -0.00001169, -0.00097516, 0.00243678, -0.00101011, 0.00006744, -0.00016268, -0.00007756, 0.00014994, -0.00008218, 0.00001274, -0.00011546, -0.00007163, 0.00010885, -0.00000803, -0.00009824, 0.00012870, -0.00005485, 0.00006877, -0.00016561, 0.00010372, -0.00013398, 0.00019413, -0.00008968, -0.00012872, },
		{ 0.00020827, -0.00023033, -0.00005069, 0.00064470, -0.00044079, 0.00000645, 0.00011743, -0.00028507, 0.00024160, -0.00064667, -0.00006166, 0.00214772, -0.00148196, 0.00055642, -0.00054591, 0.00027234, -0.00009896, 0.00007246, -0.00012418, -0.00001107, -0.00012601, 0.00020051, -0.00014493, 0.00002150, 0.00007006, -0.00003428, 0.00004094, -0.00015736, 0.00014430, -0.00020237, 0.00034252, -0.00035868, },
		{ -0.00046693, 0.00041304, -0.00052302, 0.00039956, 0.00020141, -0.00027558, -0.00003119, 0.00019104, -0.00044533, 0.00049497, -0.00124746, 0.00114342, 0.00120352, -0.00132252, 0.00062972, -0.00064492, 0.00042791, -0.00022674, 0.00013108, -0.00017129, 0.00002685, -0.00011635, 0.00021748, -0.00023006, 0.00013592, -0.00001494, 0.00001809, -0.00002511, -0.00008522, 0.00009315, -0.00014970, 0.00033552, },
		{ 0.00016471, -0.00036342, 0.00038976, -0.00057713, 0.00063671, -0.00015192, -0.00010474, -0.00003255, 0.00014025, -0.00040436, 0.00044376, -0.00145951, 0.00218343, -0.00002717, -0.00069760, 0.00031962, -0.00044077, 0.00034831, -0.00019316, 0.00006463, -0.00009866, -0.00003381, -0.00001628, 0.00012761, -0.00021411, 0.00018858, -0.00007934, 0.00006755, -0.00009750, 0.00001731, -0.00002020, 0.00000629, },
		{ 0.00018785, -0.00009241, -0.00010478, 0.00016400, -0.00038087, 0.00058913, -0.00029131, -0.00003880, 0.00006291, -0.00007046, -0.00012147, 0.00000045, -0.00104766, 0.00262607, -0.00106547, 0.00004079, -0.00015483, -0.00007483, 0.00012444, -0.00006462, -0.00006349, 0.00003960, -0.00015305, 0.00013818, -0.00004326, -0.00009678, 0.00015449, -0.00008903, 0.00007986, -0.00013660, 0.00009767, -0.00012937, },
		{ -0.00016740, 0.00029793, -0.00031217, 0.00017655, -0.00013657, -0.00005055, 0.00033285, -0.00023328, -0.00007445, 0.00021203, -0.00035422, 0.00028196, -0.00070144, -0.00004434, 0.00227024, -0.00154289, 0.00053895, -0.00052224, 0.00024692, -0.00009132, 0.00004093, -0.00014266, 0.00013937, -0.00024181, 0.00026392, -0.00021861, 0.00006957, 0.00005377, -0.00004174, 0.00004196, -0.00011848, 0.00011352, },
		{ 0.00005748, -0.00010476, 0.00027288, -0.00038950, 0.00034396, -0.00035569, 0.00024013, 0.00003406, -0.00009837, -0.00011801, 0.00029477, -0.00053917, 0.00056567, -0.00133670, 0.00122783, 0.00124813, -0.00136205, 0.00061117, -0.00060656, 0.00037475, -0.00018405, 0.00003895, -0.00009178, 0.00011428, -0.00021837, 0.00027793, -0.00030952, 0.00020549, -0.00005926, 0.00003214, -0.00002961, -0.00004991, },
		{ 0.00003405, -0.00003724, 0.00003234, 0.00011784, -0.00029230, 0.00032775, -0.00039377, 0.00035746, -0.00015210, -0.00001865, -0.00006899, 0.00020240, -0.00047453, 0.00050098, -0.00153568, 0.00228295, -0.00003262, -0.00071888, 0.00030839, -0.00040687, 0.00029008, -0.00014192, -0.00006225, 0.00008120, -0.00004493, -0.00006405, 0.00014922, -0.00026327, 0.00024587, -0.00012579, 0.00008724, -0.00009762, },
		{ -0.00012534, 0.00008920, -0.00011435, 0.00017447, -0.00009481, -0.00007594, 0.00015695, -0.00025706, 0.00027603, -0.00016574, -0.00004983, 0.00009505, -0.00007243, -0.00014090, 0.00001802, -0.00107857, 0.00269146, -0.00107857, 0.00001802, -0.00014090, -0.00007243, 0.00009505, -0.00004983, -0.00016574, 0.00027603, -0.00025706, 0.00015695, -0.00007594, -0.00009481, 0.00017447, -0.00011435, 0.00008920, },
		{ 0.00003405, -0.00009762, 0.00008724, -0.00012579, 0.00024587, -0.00026327, 0.00014922, -0.00006405, -0.00004493, 0.00008120, -0.00006225, -0.00014192, 0.00029008, -0.00040687, 0.00030839, -0.00071888, -0.00003262, 0.00228295, -0.00153568, 0.00050098, -0.00047453, 0.00020240, -0.00006899, -0.00001865, -0.00015210, 0.00035746, -0.00039377, 0.00032775, -0.00029230, 0.00011784, 0.00003234, -0.00003724, },
		{ 0.00005748, -0.00004991, -0.00002961, 0.00003214, -0.00005926, 0.00020549, -0.00030952, 0.00027793, -0.00021837, 0.00011428, -0.00009178, 0.00003895, -0.00018405, 0.00037475, -0.00060656, 0.00061117, -0.00136205, 0.00124813, 0.00122783, -0.00133670, 0.00056567, -0.00053917, 0.00029477, -0.00011801, -0.00009837, 0.00003406, 0.00024013, -0.00035569, 0.00034396, -0.00038950, 0.00027288, -0.00010476, },
		{ -0.00016740, 0.00011352, -0.00011848, 0.00004196, -0.00004174, 0.00005377, 0.00006957, -0.00021861, 0.00026392, -0.00024181, 0.00013937, -0.00014266, 0.00004093, -0.00009132, 0.00024692, -0.00052224, 0.00053895, -0.00154289, 0.00227024, -0.00004434, -0.00070144, 0.00028196, -0.00035422, 0.00021203, -0.00007445, -0.00023328, 0.00033285, -0.00005055, -0.00013657, 0.00017655, -0.00031217, 0.00029793, },
		{ 0.00018785, -0.00012937, 0.00009767, -0.00013660, 0.00007986, -0.00008903, 0.00015449, -0.00009678, -0.00004326, 0.00013818, -0.00015305, 0.00003960, -0.00006349, -0.00006462, 0.00012444, -0.00007483, -0.00015483, 0.00004079, -0.00106547, 0.00262607, -0.00104766, 0.00000045, -0.00012147, -0.00007046, 0.00006291, -0.00003880, -0.00029131, 0.00058913, -0.00038087, 0.00016400, -0.00010478, -0.00009241, },
		{ 0.00016471, 0.00000629, -0.00002020, 0.00001731, -0.00009750, 0.00006755, -0.00007934, 0.00018858, -0.00021411, 0.00012761, -0.00001628, -0.00003381, -0.00009866, 0.00006463, -0.00019316, 0.00034831, -0.00044077, 0.00031962, -0.00069760, -0.00002717, 0.00218343, -0.00145951, 0.00044376, -0.00040436, 0.00014025, -0.00003255, -0.00010474, -0.00015192, 0.00063671, -0.00057713, 0.00038976, -0.00036342, },
		{ -0.00046693, 0.00033552, -0.00014970, 0.00009315, -0.00008522, -0.00002511, 0.00001809, -0.00001494, 0.00013592, -0.00023006, 0.00021748, -0.00011635, 0.00002685, -0.00017129, 0.00013108, -0.00022674, 0.00042791, -0.00064492, 0.00062972, -0.00132252, 0.00120352, 0.00114342, -0.00124746, 0.00049497, -0.00044533, 0.00019104, -0.00003119, -0.00027558, 0.00020141, 0.00039956, -0.00052302, 0.00041304, },
		{ 0.00020827, -0.00035868, 0.00034252, -0.00020237, 0.00014430, -0.00015736, 0.00004094, -0.00003428, 0.00007006, 0.00002150, -0.00014493, 0.00020051, -0.00012601, -0.00001107, -0.00012418, 0.00007246, -0.00009896, 0.00027234, -0.00054591, 0.00055642, -0.00148196, 0.00214772, -0.00006166, -0.00064667, 0.00024160, -0.00028507, 0.00011743, 0.00000645, -0.00044079, 0.00064470, -0.00005069, -0.00023033, },
		{ 0.00015951, -0.00012872, -0.00008968, 0.00019413, -0.00013398, 0.00010372, -0.00016561, 0.00006877, -0.00005485, 0.00012870, -0.00009824, -0.00000803, 0.00010885, -0.00007163, -0.00011546, 0.00001274, -0.00008218, 0.00014994, -0.00007756, -0.00016268, 0.00006744, -0.00101011, 0.00243678, -0.00097516, -0.00001169, -0.00009724, -0.00006900, 0.00002925, -0.00003153, -0.00043415, 0.00096248, -0.00051851, },
		{ -0.00078163, 0.00044696, -0.00042850, 0.00020765, -0.00002252, 0.00000654, -0.00000698, -0.00011246, 0.00005199, -0.00002923, 0.00012754, -0.00016583, 0.00011170, 0.00000707, -0.00001826, -0.00020548, 0.00014572, -0.00022681, 0.00038523, -0.00045533, 0.00031603, -0.00064059, -0.00002756, 0.00198118, -0.00132053, 0.00037010, -0.00031516, 0.00006360, 0.00001575, -0.00021184, -0.00014279, 0.00096075, },
		{ 0.00058189, -0.00070965, 0.00048100, -0.00053746, 0.00038960, -0.00019083, 0.00013348, -0.00012964, -0.00003204, 0.00001159, 0.00002800, 0.00006548, -0.00015281, 0.00016343, -0.00004926, -0.00002335, -0.00020596, 0.00017878, -0.00024597, 0.00045358, -0.00065415, 0.00062213, -0.00122391, 0.00110028, 0.00100267, -0.00110203, 0.00040330, -0.00033002, 0.00006893, 0.00007160, -0.00048161, 0.00039928, },
		{ 0.00099560, -0.00004636, -0.00033818, 0.00024179, -0.00039981, 0.00037674, -0.00022836, 0.00017316, -0.00020359, 0.00003209, -0.00001868, 0.00008194, -0.00002571, -0.00007290, 0.00013898, -0.00004559, -0.00009728, -0.00009843, 0.00007826, -0.00009301, 0.00027890, -0.00054581, 0.00055372, -0.00136043, 0.00192926, -0.00008288, -0.00056023, 0.00019024, -0.00020328, 0.00001109, 0.00009668, -0.00067159, },
		{ -0.00058526, 0.00136933, -0.00066119, 0.00014489, -0.00014691, -0.00008671, 0.00019316, -0.00012923, 0.00010360, -0.00019881, 0.00005287, -0.00001436, 0.00009817, -0.00009914, 0.00002831, 0.00007136, -0.00001241, -0.00019560, 0.00006352, -0.00010057, 0.00017048, -0.00008053, -0.00016406, 0.00009570, -0.00091809, 0.00214328, -0.00086736, -0.00001918, -0.00006906, -0.00006810, -0.00000474, -0.00002703, },
		{ -0.00033108, -0.00012644, 0.00130476, -0.00099239, 0.00049581, -0.00048344, 0.00024338, -0.00005114, 0.00003718, -0.00003648, -0.00012958, 0.00003413, 0.00002396, 0.00006349, -0.00011868, 0.00010028, 0.00000988, -0.00000030, -0.00024751, 0.00018490, -0.00024292, 0.00040102, -0.00045161, 0.00029962, -0.00055481, -0.00003237, 0.00169631, -0.00113110, 0.00028417, -0.00021195, -0.00002300, 0.00007221, },
		{ 0.00018362, -0.00070064, 0.00061166, 0.00077172, -0.00090051, 0.00054246, -0.00059582, 0.00043051, -0.00022381, 0.00016903, -0.00017307, -0.00004171, 0.00000500, 0.00007068, -0.00000569, -0.00007817, 0.00011508, -0.00001177, -0.00004279, -0.00020285, 0.00018810, -0.00024376, 0.00045333, -0.00063644, 0.00059144, -0.00107793, 0.00095077, 0.00081941, -0.00091390, 0.00029694, -0.00020025, -0.00006432, },
		{ -0.00010094, 0.00019114, -0.00090826, 0.00135698, -0.00003946, -0.00045017, 0.00027357, -0.00043201, 0.00039730, -0.00024273, 0.00019045, -0.00024032, 0.00001745, -0.00000134, 0.00009120, -0.00007222, -0.00000335, 0.00008005, 0.00000490, -0.00013100, -0.00006810, 0.00006392, -0.00007592, 0.00026830, -0.00052390, 0.00053257, -0.00119155, 0.00163833, -0.00010547, -0.00045124, 0.00013211, -0.00011397, },
		},
		{ { -0.00008218, -0.00010334, -0.00074920, 0.00190161, -0.00081152, -0.00000997, 0.00005181, -0.00003527, 0.00010340, -0.00035228, 0.00007626, -0.00012661, 0.00012638, -0.00008715, -0.00005861, 0.00021564, -0.00011383, 0.00021564, -0.00005861, -0.00008715, 0.00012638, -0.00012661, 0.00007626, -0.00035228, 0.00010340, -0.00003527, 0.00005181, -0.00000997, -0.00081152, 0.00190161, -0.00074920, -0.00010334, },
		{ 0.00002517, -0.00003699, -0.00037543, -0.00032072, 0.00190921, -0.00124069, 0.00039678, -0.00024420, 0.00021266, -0.00013256, -0.00020837, 0.00000613, -0.00006966, 0.00011997, -0.00013925, 0.00002665, 0.00016559, -0.00009338, 0.00022648, -0.00012514, -0.00002184, 0.00009797, -0.00014270, 0.00009910, -0.00043716, 0.00029823, -0.00027128, 0.00035975, -0.00048453, -0.00017859, 0.00155413, -0.00088901, },
		{ -0.00078237, 0.00001869, 0.00006897, -0.00072993, 0.00035104, 0.00154491, -0.00136880, 0.00062764, -0.00043308, 0.00038879, -0.00033821, -0.00005604, -0.00008045, 0.00001136, 0.00008009, -0.00016530, 0.00011222, 0.00008984, -0.00003789, 0.00019496, -0.00015969, 0.00004097, 0.00004313, -0.00011490, 0.00006006, -0.00042924, 0.00039225, -0.00041389, 0.00057504, -0.00089120, 0.00050384, 0.00098352, },
		{ 0.00035262, -0.00053425, -0.00005888, 0.00015945, -0.00104783, 0.00113586, 0.00087552, -0.00118315, 0.00063405, -0.00045973, 0.00043512, -0.00045396, 0.00005469, -0.00014740, 0.00009139, 0.00001744, -0.00015873, 0.00017940, 0.00000837, 0.00003679, 0.00012914, -0.00015756, 0.00008693, -0.00002250, -0.00005472, -0.00003374, -0.00032362, 0.00035758, -0.00041140, 0.00061832, -0.00110836, 0.00106946, },
		{ 0.00138830, -0.00017992, -0.00026990, -0.00012245, 0.00014475, -0.00118812, 0.00184992, 0.00005098, -0.00075723, 0.00043901, -0.00032573, 0.00033827, -0.00045143, 0.00009117, -0.00016601, 0.00014526, -0.00005121, -0.00012099, 0.00021426, -0.00005823, 0.00010973, 0.00004578, -0.00012338, 0.00010679, -0.00008080, 0.00001571, -0.00015536, -0.00014680, 0.00020775, -0.00025987, 0.00046262, -0.00106654, },
		{ -0.00077055, 0.00140298, -0.00050544, -0.00008989, -0.00010599, -0.00003697, -0.00103621, 0.00230674, -0.00073679, -0.00022603, 0.00012583, -0.00008659, 0.00013311, -0.00034197, 0.00004879, -0.00012506, 0.00015597, -0.00010808, -0.00006092, 0.00021121, -0.00009463, 0.00016115, -0.00003472, -0.00006928, 0.00009847, -0.00011656, 0.00007116, -0.00026675, 0.00005126, -0.00000630, -0.00000526, 0.00014364, },
		{ -0.00025026, -0.00029831, 0.00114200, -0.00059208, -0.00003819, 0.00000659, -0.00038582, -0.00054873, 0.00237239, -0.00130791, 0.00025653, -0.00019202, 0.00016907, -0.00010997, -0.00017177, -0.00004825, -0.00003476, 0.00012026, -0.00013964, 0.00000785, 0.00017412, -0.00009491, 0.00017827, -0.00009427, -0.00001094, 0.00006712, -0.00012185, 0.00009178, -0.00033138, 0.00021307, -0.00021274, 0.00027104, },
		{ 0.00047858, -0.00060539, 0.00022312, 0.00070631, -0.00048527, -0.00009355, 0.00017161, -0.00082739, 0.00022135, 0.00200750, -0.00154493, 0.00056439, -0.00040966, 0.00035106, -0.00030962, -0.00000522, -0.00015601, 0.00007659, 0.00004953, -0.00014034, 0.00007070, 0.00011477, -0.00006363, 0.00015871, -0.00012206, 0.00003709, 0.00002306, -0.00009795, 0.00006967, -0.00032652, 0.00029216, -0.00034233, },
		{ -0.00035120, 0.00054744, -0.00081851, 0.00065659, 0.00023419, -0.00028301, -0.00018602, 0.00030099, -0.00122732, 0.00112896, 0.00128228, -0.00142381, 0.00063527, -0.00046482, 0.00039895, -0.00040164, 0.00009765, -0.00022669, 0.00017226, -0.00003394, -0.00011371, 0.00011590, 0.00004900, -0.00001378, 0.00011059, -0.00011708, 0.00006529, -0.00002160, -0.00005443, 0.00001143, -0.00025114, 0.00026824, },
		{ 0.00015381, -0.00023520, 0.00045122, -0.00082933, 0.00089647, -0.00014343, -0.00009583, -0.00023149, 0.00029742, -0.00143013, 0.00197122, 0.00035935, -0.00101646, 0.00048475, -0.00035614, 0.00030246, -0.00036083, 0.00010192, -0.00022580, 0.00021973, -0.00010447, -0.00007063, 0.00013731, -0.00000792, 0.00003783, 0.00004933, -0.00008742, 0.00007128, -0.00005644, -0.00000558, -0.00006418, -0.00012652, },
		{ 0.00001038, -0.00000995, -0.00003122, 0.00021578, -0.00063919, 0.00089921, -0.00034269, -0.00000568, -0.00017082, 0.00009676, -0.00131036, 0.00254230, -0.00055076, -0.00046461, 0.00019480, -0.00013982, 0.00010412, -0.00020930, 0.00000974, -0.00014460, 0.00020194, -0.00014143, -0.00002534, 0.00013528, -0.00004541, 0.00007600, -0.00000765, -0.00004681, 0.00005941, -0.00007541, 0.00003434, -0.00013268, },
		{ -0.00017214, 0.00011896, -0.00016674, 0.00019499, -0.00008923, -0.00030977, 0.00069350, -0.00034866, -0.00003834, 0.00000140, -0.00029829, -0.00081912, 0.00269230, -0.00124796, 0.00006597, -0.00011849, 0.00009346, -0.00011670, -0.00000844, -0.00014069, -0.00000371, 0.00012339, -0.00013596, 0.00000934, 0.00011554, -0.00005996, 0.00009172, -0.00004653, -0.00000966, 0.00003829, -0.00007778, 0.00005559, },
		{ 0.00005570, -0.00017055, 0.00016815, -0.00026351, 0.00037051, -0.00037304, 0.00005751, 0.00036719, -0.00021304, -0.00015832, 0.00022957, -0.00080635, -0.00000988, 0.00237185, -0.00159699, 0.00043589, -0.00034621, 0.00025302, -0.00027266, 0.00016278, -0.00028867, 0.00015360, 0.00000892, -0.00009364, 0.00002605, 0.00008664, -0.00005504, 0.00008432, -0.00006117, 0.00001375, 0.00001727, -0.00006735, },
		{ -0.00005019, 0.00003953, -0.00012983, 0.00014645, -0.00026778, 0.00043910, -0.00055328, 0.00035454, 0.00003571, -0.00002951, -0.00028743, 0.00041459, -0.00128355, 0.00096638, 0.00164832, -0.00156128, 0.00057113, -0.00042295, 0.00028170, -0.00030021, 0.00023566, -0.00037244, 0.00027603, -0.00010448, -0.00003235, 0.00002474, 0.00005685, -0.00003883, 0.00006078, -0.00005430, 0.00002026, 0.00000295, },
		{ -0.00000276, -0.00003202, 0.00001686, -0.00006495, 0.00006558, -0.00017797, 0.00037935, -0.00058240, 0.00050097, -0.00019683, 0.00010382, -0.00033960, 0.00045052, -0.00156403, 0.00189695, 0.00068822, -0.00120659, 0.00047870, -0.00034585, 0.00017670, -0.00018317, 0.00017621, -0.00035111, 0.00032190, -0.00018000, 0.00002409, 0.00001193, 0.00003167, -0.00002072, 0.00003238, -0.00003563, 0.00001406, },
		{ 0.00000428, -0.00000271, -0.00001602, -0.00000163, 0.00000123, -0.00004365, -0.00002353, 0.00021113, -0.00046227, 0.00046888, -0.00027107, 0.00012145, -0.00025887, 0.00026753, -0.00151246, 0.00256425, -0.00028900, -0.00067439, 0.00023525, -0.00017055, -0.00000960, 0.00004027, -0.00000218, -0.00021966, 0.00027381, -0.00019348, 0.00005413, -0.00000230, 0.00001266, -0.00000779, 0.00001019, -0.00001756, },
		{ -0.00001025, 0.00000084, -0.00000256, -0.00000207, -0.00000849, 0.00004589, -0.00014169, 0.00014519, -0.00001169, -0.00024239, 0.00028950, -0.00018912, 0.00001389, -0.00004617, -0.00013446, -0.00107167, 0.00280655, -0.00107167, -0.00013446, -0.00004617, 0.00001389, -0.00018912, 0.00028950, -0.00024239, -0.00001169, 0.00014519, -0.00014169, 0.00004589, -0.00000849, -0.00000207, -0.00000256, 0.00000084, },
		{ 0.00000428, -0.00001756, 0.00001019, -0.00000779, 0.00001266, -0.00000230, 0.00005413, -0.00019348, 0.00027381, -0.00021966, -0.00000218, 0.00004027, -0.00000960, -0.00017055, 0.00023525, -0.00067439, -0.00028900, 0.00256425, -0.00151246, 0.00026753, -0.00025887, 0.00012145, -0.00027107, 0.00046888, -0.00046227, 0.00021113, -0.00002353, -0.00004365, 0.00000123, -0.00000163, -0.00001602, -0.00000271, },
		{ -0.00000276, 0.00001406, -0.00003563, 0.00003238, -0.00002072, 0.00003167, 0.00001193, 0.00002409, -0.00018000, 0.00032190, -0.00035111, 0.00017621, -0.00018317, 0.00017670, -0.00034585, 0.00047870, -0.00120659, 0.00068822, 0.00189695, -0.00156403, 0.00045052, -0.00033960, 0.00010382, -0.00019683, 0.00050097, -0.00058240, 0.00037935, -0.00017797, 0.00006558, -0.00006495, 0.00001686, -0.00003202, },
		{ -0.00005019, 0.00000295, 0.00002026, -0.00005430, 0.00006078, -0.00003883, 0.00005685, 0.00002474, -0.00003235, -0.00010448, 0.00027603, -0.00037244, 0.00023566, -0.00030021, 0.00028170, -0.00042295, 0.00057113, -0.00156128, 0.00164832, 0.00096638, -0.00128355, 0.00041459, -0.00028743, -0.00002951, 0.00003571, 0.00035454, -0.00055328, 0.00043910, -0.00026778, 0.00014645, -0.00012983, 0.00003953, },
		{ 0.00005570, -0.00006735, 0.00001727, 0.00001375, -0.00006117, 0.00008432, -0.00005504, 0.00008664, 0.00002605, -0.00009364, 0.00000892, 0.00015360, -0.00028867, 0.00016278, -0.00027266, 0.00025302, -0.00034621, 0.00043589, -0.00159699, 0.00237185, -0.00000988, -0.00080635, 0.00022957, -0.00015832, -0.00021304, 0.00036719, 0.00005751, -0.00037304, 0.00037051, -0.00026351, 0.00016815, -0.00017055, },
		{ -0.00017214, 0.00005559, -0.00007778, 0.00003829, -0.00000966, -0.00004653, 0.00009172, -0.00005996, 0.00011554, 0.00000934, -0.00013596, 0.00012339, -0.00000371, -0.00014069, -0.00000844, -0.00011670, 0.00009346, -0.00011849, 0.00006597, -0.00124796, 0.00269230, -0.00081912, -0.00029829, 0.00000140, -0.00003834, -0.00034866, 0.00069350, -0.00030977, -0.00008923, 0.00019499, -0.00016674, 0.00011896, },
		{ 0.00001038, -0.00013268, 0.00003434, -0.00007541, 0.00005941, -0.00004681, -0.00000765, 0.00007600, -0.00004541, 0.00013528, -0.00002534, -0.00014143, 0.00020194, -0.00014460, 0.00000974, -0.00020930, 0.00010412, -0.00013982, 0.00019480, -0.00046461, -0.00055076, 0.00254230, -0.00131036, 0.00009676, -0.00017082, -0.00000568, -0.00034269, 0.00089921, -0.00063919, 0.00021578, -0.00003122, -0.00000995, },
		{ 0.00015381, -0.00012652, -0.00006418, -0.00000558, -0.00005644, 0.00007128, -0.00008742, 0.00004933, 0.00003783, -0.00000792, 0.00013731, -0.00007063, -0.00010447, 0.00021973, -0.00022580, 0.00010192, -0.00036083, 0.00030246, -0.00035614, 0.00048475, -0.00101646, 0.00035935, 0.00197122, -0.00143013, 0.00029742, -0.00023149, -0.00009583, -0.00014343, 0.00089647, -0.00082933, 0.00045122, -0.00023520, },
		{ -0.00035120, 0.00026824, -0.00025114, 0.00001143, -0.00005443, -0.00002160, 0.00006529, -0.00011708, 0.00011059, -0.00001378, 0.00004900, 0.00011590, -0.00011371, -0.00003394, 0.00017226, -0.00022669, 0.00009765, -0.00040164, 0.00039895, -0.00046482, 0.00063527, -0.00142381, 0.00128228, 0.00112896, -0.00122732, 0.00030099, -0.00018602, -0.00028301, 0.00023419, 0.00065659, -0.00081851, 0.00054744, },
		{ 0.00047858, -0.00034233, 0.00029216, -0.00032652, 0.00006967, -0.00009795, 0.00002306, 0.00003709, -0.00012206, 0.00015871, -0.00006363, 0.00011477, 0.00007070, -0.00014034, 0.00004953, 0.00007659, -0.00015601, -0.00000522, -0.00030962, 0.00035106, -0.00040966, 0.00056439, -0.00154493, 0.00200750, 0.00022135, -0.00082739, 0.00017161, -0.00009355, -0.00048527, 0.00070631, 0.00022312, -0.00060539, },
		{ -0.00025026, 0.00027104, -0.00021274, 0.00021307, -0.00033138, 0.00009178, -0.00012185, 0.00006712, -0.00001094, -0.00009427, 0.00017827, -0.00009491, 0.00017412, 0.00000785, -0.00013964, 0.00012026, -0.00003476, -0.00004825, -0.00017177, -0.00010997, 0.00016907, -0.00019202, 0.00025653, -0.00130791, 0.00237239, -0.00054873, -0.00038582, 0.00000659, -0.00003819, -0.00059208, 0.00114200, -0.00029831, },
		{ -0.00077055, 0.00014364, -0.00000526, -0.00000630, 0.00005126, -0.00026675, 0.00007116, -0.00011656, 0.00009847, -0.00006928, -0.00003472, 0.00016115, -0.00009463, 0.00021121, -0.00006092, -0.00010808, 0.00015597, -0.00012506, 0.00004879, -0.00034197, 0.00013311, -0.00008659, 0.00012583, -0.00022603, -0.00073679, 0.00230674, -0.00103621, -0.00003697, -0.00010599, -0.00008989, -0.00050544, 0.00140298, },
		{ 0.00138830, -0.00106654, 0.00046262, -0.00025987, 0.00020775, -0.00014680, -0.00015536, 0.00001571, -0.00008080, 0.00010679, -0.00012338, 0.00004578, 0.00010973, -0.00005823, 0.00021426, -0.00012099, -0.00005121, 0.00014526, -0.00016601, 0.00009117, -0.00045143, 0.00033827, -0.00032573, 0.00043901, -0.00075723, 0.00005098, 0.00184992, -0.00118812, 0.00014475, -0.00012245, -0.00026990, -0.00017992, },
		{ 0.00035262, 0.00106946, -0.00110836, 0.00061832, -0.00041140, 0.00035758, -0.00032362, -0.00003374, -0.00005472, -0.00002250, 0.00008693, -0.00015756, 0.00012914, 0.00003679, 0.00000837, 0.00017940, -0.00015873, 0.00001744, 0.00009139, -0.00014740, 0.00005469, -0.00045396, 0.00043512, -0.00045973, 0.00063405, -0.00118315, 0.00087552, 0.00113586, -0.00104783, 0.00015945, -0.00005888, -0.00053425, },
		{ -0.00078237, 0.00098352, 0.00050384, -0.00089120, 0.00057504, -0.00041389, 0.00039225, -0.00042924, 0.00006006, -0.00011490, 0.00004313, 0.00004097, -0.00015969, 0.00019496, -0.00003789, 0.00008984, 0.00011222, -0.00016530, 0.00008009, 0.00001136, -0.00008045, -0.00005604, -0.00033821, 0.00038879, -0.00043308, 0.00062764, -0.00136880, 0.00154491, 0.00035104, -0.00072993, 0.00006897, 0.00001869, },
		{ 0.00002517, -0.00088901, 0.00155413, -0.00017859, -0.00048453, 0.00035975, -0.00027128, 0.00029823, -0.00043716, 0.00009910, -0.00014270, 0.00009797, -0.00002184, -0.00012514, 0.00022648, -0.00009338, 0.00016559, 0.00002665, -0.00013925, 0.00011997, -0.00006966, 0.00000613, -0.00020837, -0.00013256, 0.00021266, -0.00024420, 0.00039678, -0.00124069, 0.00190921, -0.00032072, -0.00037543, -0.00003699, },
		},
		{ { 0.00005017, -0.00080966, 0.00170692, -0.00061750, 0.00005051, -0.00009476, -0.00036709, 0.00014041, -0.00006923, 0.00005592, 0.00006611, 0.00003927, -0.00004511, -0.00000676, -0.00001408, -0.00006642, 0.00009908, -0.00006642, -0.00001408, -0.00000676, -0.00004511, 0.00003927, 0.00006611, 0.00005592, -0.00006923, 0.00014041, -0.00036709, -0.00009476, 0.00005051, -0.00061750, 0.00170692, -0.00080966, },
		{ -0.00081722, -0.00001120, -0.00064738, 0.00184532, -0.00094561, 0.00032208, -0.00034165, -0.00019008, 0.00005213, -0.00000810, 0.00001723, 0.00010412, 0.00000243, -0.00003200, -0.00000587, -0.00002872, -0.00004564, 0.00009224, -0.00007683, -0.00000303, -0.00000910, -0.00005268, 0.00007014, 0.00003164, 0.00008706, -0.00011778, 0.00020876, -0.00052194, 0.00015216, -0.00024957, -0.00020934, 0.00141473, },
		{ 0.00101400, -0.00069426, -0.00018322, -0.00033006, 0.00180446, -0.00115308, 0.00052553, -0.00055385, -0.00001646, -0.00004010, 0.00005500, -0.00002163, 0.00013969, -0.00003585, -0.00001508, -0.00000719, -0.00004440, -0.00001722, 0.00007271, -0.00007578, 0.00000261, -0.00001190, -0.00005382, 0.00009150, 0.00000581, 0.00010454, -0.00014510, 0.00024396, -0.00063165, 0.00036438, -0.00053382, 0.00022658, },
		{ 0.00063390, 0.00056291, -0.00048420, -0.00043160, 0.00011774, 0.00158273, -0.00121686, 0.00063191, -0.00070188, 0.00012949, -0.00012010, 0.00010915, -0.00005320, 0.00016687, -0.00007073, 0.00000334, -0.00001115, -0.00005830, 0.00001492, 0.00004338, -0.00006396, 0.00000197, -0.00001410, -0.00004860, 0.00010126, -0.00000797, 0.00010451, -0.00014590, 0.00023763, -0.00067928, 0.00051234, -0.00075991, },
		{ -0.00089392, 0.00096032, 0.00012373, -0.00024211, -0.00070536, 0.00064939, 0.00120289, -0.00113455, 0.00062728, -0.00076584, 0.00022838, -0.00017402, 0.00014481, -0.00007109, 0.00018061, -0.00009772, 0.00002073, -0.00001769, -0.00006777, 0.00004621, 0.00000855, -0.00004367, -0.00000472, -0.00001480, -0.00003801, 0.00009908, -0.00000846, 0.00008601, -0.00011907, 0.00018765, -0.00065652, 0.00057600, },
		{ 0.00054782, -0.00091544, 0.00116522, -0.00024635, -0.00002545, -0.00094512, 0.00120248, 0.00070909, -0.00092456, 0.00051472, -0.00073836, 0.00026853, -0.00019253, 0.00015537, -0.00007109, 0.00017766, -0.00011337, 0.00003468, -0.00002630, -0.00007072, 0.00007210, -0.00002666, -0.00001852, -0.00001615, -0.00001339, -0.00002384, 0.00008639, 0.00000326, 0.00005115, -0.00006792, 0.00009859, -0.00056498, },
		{ -0.00041605, 0.00043380, -0.00082074, 0.00122549, -0.00050357, 0.00011546, -0.00109277, 0.00170759, 0.00016060, -0.00062301, 0.00031380, -0.00062558, 0.00024768, -0.00017226, 0.00013814, -0.00005179, 0.00015703, -0.00011575, 0.00004327, -0.00003608, -0.00006599, 0.00008880, -0.00005715, 0.00000721, -0.00003017, -0.00000966, -0.00000837, 0.00006609, 0.00002400, 0.00000481, 0.00000033, -0.00001885, },
		{ -0.00014940, -0.00022924, 0.00025254, -0.00062331, 0.00113875, -0.00062379, 0.00014581, -0.00110123, 0.00209832, -0.00037653, -0.00027815, 0.00005749, -0.00044619, 0.00017308, -0.00011635, 0.00009488, -0.00001489, 0.00012022, -0.00010472, 0.00004536, -0.00004581, -0.00005355, 0.00009380, -0.00007856, 0.00002918, -0.00004411, -0.00000387, 0.00000596, 0.00004209, 0.00004899, -0.00004606, 0.00007563, },
		{ 0.00014656, -0.00027539, -0.00002932, 0.00003231, -0.00035188, 0.00092344, -0.00060558, 0.00005204, -0.00094247, 0.00232095, -0.00083891, 0.00005715, -0.00021296, -0.00022851, 0.00006035, -0.00003397, 0.00003147, 0.00003501, 0.00007107, -0.00008197, 0.00004078, -0.00005420, -0.00003453, 0.00008627, -0.00008794, 0.00004371, -0.00005521, 0.00000330, 0.00001694, 0.00001871, 0.00007265, -0.00009354, },
		{ -0.00012991, 0.00020214, -0.00037941, 0.00015726, -0.00019323, -0.00004605, 0.00061580, -0.00046978, -0.00015567, -0.00061288, 0.00234244, -0.00117475, 0.00033341, -0.00045434, -0.00000612, -0.00006916, 0.00006110, -0.00004287, 0.00009107, 0.00001523, -0.00005073, 0.00003030, -0.00006001, -0.00001108, 0.00006719, -0.00008417, 0.00004833, -0.00006109, 0.00001084, 0.00002295, -0.00000006, 0.00008957, },
		{ 0.00009532, -0.00014886, 0.00023346, -0.00044685, 0.00030631, -0.00039019, 0.00024962, 0.00026423, -0.00025565, -0.00044486, -0.00013485, 0.00215541, -0.00135097, 0.00051204, -0.00062821, 0.00018726, -0.00019172, 0.00015264, -0.00011701, 0.00014527, -0.00004063, -0.00001532, 0.00001554, -0.00006228, 0.00001393, 0.00003921, -0.00006816, 0.00004217, -0.00006009, 0.00001755, 0.00002313, -0.00001113, },
		{ -0.00001282, 0.00008728, -0.00014649, 0.00023508, -0.00046810, 0.00039938, -0.00052959, 0.00049237, -0.00007811, -0.00001417, -0.00076572, 0.00044558, 0.00177941, -0.00135763, 0.00057108, -0.00070676, 0.00032286, -0.00028502, 0.00022448, -0.00017960, 0.00018952, -0.00008976, 0.00001953, -0.00000128, -0.00006046, 0.00003734, 0.00000624, -0.00004264, 0.00002610, -0.00005162, 0.00002224, 0.00001760, },
		{ 0.00000737, -0.00000509, 0.00006502, -0.00012191, 0.00020585, -0.00043998, 0.00042640, -0.00059172, 0.00064762, -0.00036143, 0.00020042, -0.00105891, 0.00106466, 0.00125831, -0.00120873, 0.00050833, -0.00067710, 0.00038126, -0.00033169, 0.00026310, -0.00022080, 0.00021698, -0.00012630, 0.00004934, -0.00001766, -0.00005452, 0.00005610, -0.00002710, -0.00001179, 0.00000260, -0.00003620, 0.00002390, },
		{ 0.00002189, -0.00000579, 0.00001042, 0.00003050, -0.00007749, 0.00014917, -0.00036623, 0.00038708, -0.00056910, 0.00069416, -0.00054677, 0.00033957, -0.00126502, 0.00164972, 0.00065411, -0.00093952, 0.00034143, -0.00054330, 0.00035531, -0.00032206, 0.00025975, -0.00023376, 0.00022305, -0.00014609, 0.00007054, -0.00003125, -0.00004494, 0.00006778, -0.00005621, 0.00001945, -0.00002461, -0.00001547, },
		{ 0.00000804, 0.00001603, -0.00001964, 0.00003069, -0.00001220, -0.00001846, 0.00007246, -0.00025694, 0.00029079, -0.00046751, 0.00062730, -0.00061178, 0.00036839, -0.00133414, 0.00212942, 0.00003820, -0.00060065, 0.00010486, -0.00032572, 0.00025110, -0.00025577, 0.00021189, -0.00021572, 0.00020610, -0.00014729, 0.00008099, -0.00004014, -0.00003268, 0.00007088, -0.00007716, 0.00004613, -0.00005116, },
		{ -0.00007269, 0.00003132, 0.00000665, -0.00003181, 0.00005181, -0.00005746, 0.00004774, -0.00001387, -0.00012709, 0.00015504, -0.00030507, 0.00045945, -0.00055389, 0.00027153, -0.00123419, 0.00244389, -0.00051875, -0.00025011, -0.00015564, -0.00005799, 0.00008697, -0.00014184, 0.00012374, -0.00016845, 0.00016773, -0.00013056, 0.00008025, -0.00004319, -0.00001902, 0.00006508, -0.00008727, 0.00006402, },
		{ 0.00007031, -0.00008553, 0.00005124, -0.00000544, -0.00004018, 0.00006960, -0.00009895, 0.00011260, -0.00009801, 0.00000570, 0.00000262, -0.00010934, 0.00021821, -0.00039042, 0.00005588, -0.00095666, 0.00255336, -0.00095666, 0.00005588, -0.00039042, 0.00021821, -0.00010934, 0.00000262, 0.00000570, -0.00009801, 0.00011260, -0.00009895, 0.00006960, -0.00004018, -0.00000544, 0.00005124, -0.00008553, },
		{ -0.00007269, 0.00006402, -0.00008727, 0.00006508, -0.00001902, -0.00004319, 0.00008025, -0.00013056, 0.00016773, -0.00016845, 0.00012374, -0.00014184, 0.00008697, -0.00005799, -0.00015564, -0.00025011, -0.00051875, 0.00244389, -0.00123419, 0.00027153, -0.00055389, 0.00045945, -0.00030507, 0.00015504, -0.00012709, -0.00001387, 0.00004774, -0.00005746, 0.00005181, -0.00003181, 0.00000665, 0.00003132, },
		{ 0.00000804, -0.00005116, 0.00004613, -0.00007716, 0.00007088, -0.00003268, -0.00004014, 0.00008099, -0.00014729, 0.00020610, -0.00021572, 0.00021189, -0.00025577, 0.00025110, -0.00032572, 0.00010486, -0.00060065, 0.00003820, 0.00212942, -0.00133414, 0.00036839, -0.00061178, 0.00062730, -0.00046751, 0.00029079, -0.00025694, 0.00007246, -0.00001846, -0.00001220, 0.00003069, -0.00001964, 0.00001603, },
		{ 0.00002189, -0.00001547, -0.00002461, 0.00001945, -0.00005621, 0.00006778, -0.00004494, -0.00003125, 0.00007054, -0.00014609, 0.00022305, -0.00023376, 0.00025975, -0.00032206, 0.00035531, -0.00054330, 0.00034143, -0.00093952, 0.00065411, 0.00164972, -0.00126502, 0.00033957, -0.00054677, 0.00069416, -0.00056910, 0.00038708, -0.00036623, 0.00014917, -0.00007749, 0.00003050, 0.00001042, -0.00000579, },
		{ 0.00000737, 0.00002390, -0.00003620, 0.00000260, -0.00001179, -0.00002710, 0.00005610, -0.00005452, -0.00001766, 0.00004934, -0.00012630, 0.00021698, -0.00022080, 0.00026310, -0.00033169, 0.00038126, -0.00067710, 0.00050833, -0.00120873, 0.00125831, 0.00106466, -0.00105891, 0.00020042, -0.00036143, 0.00064762, -0.00059172, 0.00042640, -0.00043998, 0.00020585, -0.00012191, 0.00006502, -0.00000509, },
		{ -0.00001282, 0.00001760, 0.00002224, -0.00005162, 0.00002610, -0.00004264, 0.00000624, 0.00003734, -0.00006046, -0.00000128, 0.00001953, -0.00008976, 0.00018952, -0.00017960, 0.00022448, -0.00028502, 0.00032286, -0.00070676, 0.00057108, -0.00135763, 0.00177941, 0.00044558, -0.00076572, -0.00001417, -0.00007811, 0.00049237, -0.00052959, 0.00039938, -0.00046810, 0.00023508, -0.00014649, 0.00008728, },
		{ 0.00009532, -0.00001113, 0.00002313, 0.00001755, -0.00006009, 0.00004217, -0.00006816, 0.00003921, 0.00001393, -0.00006228, 0.00001554, -0.00001532, -0.00004063, 0.00014527, -0.00011701, 0.00015264, -0.00019172, 0.00018726, -0.00062821, 0.00051204, -0.00135097, 0.00215541, -0.00013485, -0.00044486, -0.00025565, 0.00026423, 0.00024962, -0.00039019, 0.00030631, -0.00044685, 0.00023346, -0.00014886, },
		{ -0.00012991, 0.00008957, -0.00000006, 0.00002295, 0.00001084, -0.00006109, 0.00004833, -0.00008417, 0.00006719, -0.00001108, -0.00006001, 0.00003030, -0.00005073, 0.00001523, 0.00009107, -0.00004287, 0.00006110, -0.00006916, -0.00000612, -0.00045434, 0.00033341, -0.00117475, 0.00234244, -0.00061288, -0.00015567, -0.00046978, 0.00061580, -0.00004605, -0.00019323, 0.00015726, -0.00037941, 0.00020214, },
		{ 0.00014656, -0.00009354, 0.00007265, 0.00001871, 0.00001694, 0.00000330, -0.00005521, 0.00004371, -0.00008794, 0.00008627, -0.00003453, -0.00005420, 0.00004078, -0.00008197, 0.00007107, 0.00003501, 0.00003147, -0.00003397, 0.00006035, -0.00022851, -0.00021296, 0.00005715, -0.00083891, 0.00232095, -0.00094247, 0.00005204, -0.00060558, 0.00092344, -0.00035188, 0.00003231, -0.00002932, -0.00027539, },
		{ -0.00014940, 0.00007563, -0.00004606, 0.00004899, 0.00004209, 0.00000596, -0.00000387, -0.00004411, 0.00002918, -0.00007856, 0.00009380, -0.00005355, -0.00004581, 0.00004536, -0.00010472, 0.00012022, -0.00001489, 0.00009488, -0.00011635, 0.00017308, -0.00044619, 0.00005749, -0.00027815, -0.00037653, 0.00209832, -0.00110123, 0.00014581, -0.00062379, 0.00113875, -0.00062331, 0.00025254, -0.00022924, },
		{ -0.00041605, -0.00001885, 0.00000033, 0.00000481, 0.00002400, 0.00006609, -0.00000837, -0.00000966, -0.00003017, 0.00000721, -0.00005715, 0.00008880, -0.00006599, -0.00003608, 0.00004327, -0.00011575, 0.00015703, -0.00005179, 0.00013814, -0.00017226, 0.00024768, -0.00062558, 0.00031380, -0.00062301, 0.00016060, 0.00170759, -0.00109277, 0.00011546, -0.00050357, 0.00122549, -0.00082074, 0.00043380, },
		{ 0.00054782, -0.00056498, 0.00009859, -0.00006792, 0.00005115, 0.00000326, 0.00008639, -0.00002384, -0.00001339, -0.00001615, -0.00001852, -0.00002666, 0.00007210, -0.00007072, -0.00002630, 0.00003468, -0.00011337, 0.00017766, -0.00007109, 0.00015537, -0.00019253, 0.00026853, -0.00073836, 0.00051472, -0.00092456, 0.00070909, 0.00120248, -0.00094512, -0.00002545, -0.00024635, 0.00116522, -0.00091544, },
		{ -0.00089392, 0.00057600, -0.00065652, 0.00018765, -0.00011907, 0.00008601, -0.00000846, 0.00009908, -0.00003801, -0.00001480, -0.00000472, -0.00004367, 0.00000855, 0.00004621, -0.00006777, -0.00001769, 0.00002073, -0.00009772, 0.00018061, -0.00007109, 0.00014481, -0.00017402, 0.00022838, -0.00076584, 0.00062728, -0.00113455, 0.00120289, 0.00064939, -0.00070536, -0.00024211, 0.00012373, 0.00096032, },
		{ 0.00063390, -0.00075991, 0.00051234, -0.00067928, 0.00023763, -0.00014590, 0.00010451, -0.00000797, 0.00010126, -0.00004860, -0.00001410, 0.00000197, -0.00006396, 0.00004338, 0.00001492, -0.00005830, -0.00001115, 0.00000334, -0.00007073, 0.00016687, -0.00005320, 0.00010915, -0.00012010, 0.00012949, -0.00070188, 0.00063191, -0.00121686, 0.00158273, 0.00011774, -0.00043160, -0.00048420, 0.00056291, },
		{ 0.00101400, 0.00022658, -0.00053382, 0.00036438, -0.00063165, 0.00024396, -0.00014510, 0.00010454, 0.00000581, 0.00009150, -0.00005382, -0.00001190, 0.00000261, -0.00007578, 0.00007271, -0.00001722, -0.00004440, -0.00000719, -0.00001508, -0.00003585, 0.00013969, -0.00002163, 0.00005500, -0.00004010, -0.00001646, -0.00055385, 0.00052553, -0.00115308, 0.00180446, -0.00033006, -0.00018322, -0.00069426, },
		{ -0.00081722, 0.00141473, -0.00020934, -0.00024957, 0.00015216, -0.00052194, 0.00020876, -0.00011778, 0.00008706, 0.00003164, 0.00007014, -0.00005268, -0.00000910, -0.00000303, -0.00007683, 0.00009224, -0.00004564, -0.00002872, -0.00000587, -0.00003200, 0.00000243, 0.00010412, 0.00001723, -0.00000810, 0.00005213, -0.00019008, -0.00034165, 0.00032208, -0.00094561, 0.00184532, -0.00064738, -0.00001120, },
		},
		{ { -0.00133090, 0.00172239, -0.00058612, -0.00049016, 0.00005544, 0.00005486, 0.00018949, -0.00033739, 0.00010340, 0.00006577, -0.00001086, -0.00013076, 0.00012258, -0.00005445, 0.00007301, -0.00011815, 0.00009908, -0.00011815, 0.00007301, -0.00005445, 0.00012258, -0.00013076, -0.00001086, 0.00006577, 0.00010340, -0.00033739, 0.00018949, 0.00005486, 0.00005544, -0.00049016, -0.00058612, 0.00172239, },
		{ 0.00147425, -0.00129275, 0.00191636, -0.00083341, -0.00033662, -0.00002752, 0.00014090, 0.00010974, -0.00030575, 0.00010174, 0.00008297, -0.00004517, -0.00010448, 0.00010697, -0.00003911, 0.00006098, -0.00011407, 0.00009743, -0.00011836, 0.00008107, -0.00006677, 0.00013488, -0.00015321, 0.00002234, 0.00004723, 0.00010174, -0.00035881, 0.00026095, -0.00002967, 0.00013509, -0.00063717, -0.00032546, },
		{ -0.00006147, 0.00118147, -0.00117978, 0.00204869, -0.00105785, -0.00018246, -0.00011062, 0.00022513, 0.00002478, -0.00026511, 0.00009683, 0.00009818, -0.00007926, -0.00007537, 0.00008865, -0.00002134, 0.00004542, -0.00010627, 0.00009251, -0.00011469, 0.00008485, -0.00007559, 0.00014338, -0.00017096, 0.00005317, 0.00002808, 0.00009683, -0.00036919, 0.00032139, -0.00010946, 0.00020837, -0.00077202, },
		{ -0.00088952, 0.00019573, 0.00085530, -0.00099632, 0.00211431, -0.00125079, -0.00003361, -0.00019065, 0.00030432, -0.00006214, -0.00021703, 0.00008885, 0.00011082, -0.00011182, -0.00004455, 0.00006831, -0.00000182, 0.00002696, -0.00009506, 0.00008454, -0.00010727, 0.00008419, -0.00008058, 0.00014776, -0.00018333, 0.00008044, 0.00000905, 0.00008885, -0.00036813, 0.00036848, -0.00018143, 0.00027245, },
		{ 0.00032487, -0.00098515, 0.00043623, 0.00050827, -0.00074942, 0.00211069, -0.00140483, 0.00010422, -0.00026454, 0.00037541, -0.00014767, -0.00016336, 0.00007812, 0.00012039, -0.00014160, -0.00001321, 0.00004674, 0.00001871, 0.00000628, -0.00008085, 0.00007380, -0.00009639, 0.00007912, -0.00008154, 0.00014787, -0.00018984, 0.00010310, -0.00000913, 0.00007812, -0.00035567, 0.00040041, -0.00024283, },
		{ -0.00029128, 0.00036362, -0.00105524, 0.00065081, 0.00015373, -0.00044858, 0.00203797, -0.00151405, 0.00022573, -0.00032945, 0.00043569, -0.00022852, -0.00010616, 0.00006504, 0.00012654, -0.00016746, 0.00001745, 0.00002477, 0.00003944, -0.00001581, -0.00006420, 0.00006072, -0.00008248, 0.00006985, -0.00007843, 0.00014368, -0.00019025, 0.00012027, -0.00002577, 0.00006504, -0.00033229, 0.00041595, },
		{ 0.00041450, -0.00032494, 0.00038721, -0.00109709, 0.00083121, -0.00019471, -0.00010535, 0.00189894, -0.00157426, 0.00032624, -0.00038289, 0.00048284, -0.00030159, -0.00004763, 0.00005012, 0.00012902, -0.00018840, 0.00004625, 0.00000323, 0.00005959, -0.00003846, -0.00004576, 0.00004580, -0.00006607, 0.00005672, -0.00007139, 0.00013536, -0.00018453, 0.00013131, -0.00004021, 0.00005012, -0.00029890, },
		{ -0.00025676, 0.00039613, -0.00034249, 0.00039473, -0.00110910, 0.00097051, -0.00052365, 0.00026708, 0.00169896, -0.00158313, 0.00040191, -0.00042280, 0.00051503, -0.00036407, 0.00000998, 0.00003392, 0.00012773, -0.00020362, 0.00007209, -0.00001703, 0.00007837, -0.00006080, -0.00002622, 0.00002961, -0.00004778, 0.00004023, -0.00006067, 0.00012324, -0.00017291, 0.00013578, -0.00005192, 0.00003392, },
		{ 0.00001708, -0.00020750, 0.00036154, -0.00034328, 0.00038590, -0.00109080, 0.00106334, -0.00082046, 0.00065439, 0.00144570, -0.00154032, 0.00044981, -0.00044765, 0.00053105, -0.00041356, 0.00006446, 0.00001708, 0.00012274, -0.00021253, 0.00009397, -0.00003524, 0.00009508, -0.00008198, -0.00000633, 0.00001277, -0.00002832, 0.00002104, -0.00004669, 0.00010777, -0.00015583, 0.00013352, -0.00006043, },
		{ -0.00006542, 0.00000024, -0.00015302, 0.00031205, -0.00032727, 0.00036105, -0.00104290, 0.00110614, -0.00107372, 0.00104170, 0.00114889, -0.00144749, 0.00046810, -0.00045648, 0.00053026, -0.00044815, 0.00011372, 0.00000024, 0.00011423, -0.00021479, 0.00011104, -0.00005071, 0.00010906, -0.00010118, 0.00001313, -0.00000407, -0.00000843, -0.00000014, -0.00002998, 0.00008956, -0.00013396, 0.00012461, },
		{ 0.00010939, -0.00006670, -0.00001595, -0.00009541, 0.00024957, -0.00029507, 0.00032115, -0.00096724, 0.00109727, -0.00127370, 0.00141412, 0.00081995, -0.00130819, 0.00045610, -0.00044895, 0.00051270, -0.00046652, 0.00015585, -0.00001595, 0.00010252, -0.00021032, 0.00012266, -0.00006283, 0.00011978, -0.00011766, 0.00003142, -0.00002026, 0.00001111, -0.00002249, -0.00001119, 0.00006930, -0.00010812, },
		{ -0.00007932, 0.00008845, -0.00006423, -0.00003087, -0.00003688, 0.00017650, -0.00024793, 0.00026771, -0.00086672, 0.00103707, -0.00141273, 0.00175735, 0.00047151, -0.00112779, 0.00041424, -0.00042536, 0.00047905, -0.00046797, 0.00018925, -0.00003087, 0.00008808, -0.00019928, 0.00012838, -0.00007114, 0.00012683, -0.00013079, 0.00004783, -0.00003519, 0.00002955, -0.00004514, 0.00000895, 0.00004777, },
		{ 0.00002579, -0.00004866, 0.00006259, -0.00005808, -0.00004395, 0.00002032, 0.00009565, -0.00018765, 0.00020280, -0.00074522, 0.00092785, -0.00148544, 0.00205820, 0.00011697, -0.00091321, 0.00034416, -0.00038661, 0.00043059, -0.00045243, 0.00021263, -0.00004395, 0.00007144, -0.00018210, 0.00012797, -0.00007533, 0.00012993, -0.00014007, 0.00006174, -0.00004827, 0.00004620, -0.00006723, 0.00002969, },
		{ 0.00005021, 0.00000422, -0.00001732, 0.00003281, -0.00004851, -0.00005469, 0.00007399, 0.00001012, -0.00011655, 0.00012891, -0.00060739, 0.00077381, -0.00148907, 0.00230509, -0.00023006, -0.00067271, 0.00024853, -0.00033419, 0.00036920, -0.00042050, 0.00022509, -0.00005469, 0.00005326, -0.00015945, 0.00012146, -0.00007523, 0.00012897, -0.00014514, 0.00007262, -0.00005900, 0.00006041, -0.00008790, },
		{ -0.00010637, 0.00006973, -0.00001611, 0.00001350, 0.00000025, -0.00003587, -0.00006266, 0.00012207, -0.00007680, -0.00003736, 0.00004887, -0.00045853, 0.00058086, -0.00142345, 0.00248855, -0.00055623, -0.00041551, 0.00013103, -0.00027011, 0.00029723, -0.00037341, 0.00022614, -0.00006266, 0.00003423, -0.00013218, 0.00010909, -0.00007084, 0.00012398, -0.00014579, 0.00008004, -0.00006698, 0.00007162, },
		{ 0.00007942, -0.00012192, 0.00008750, -0.00003444, 0.00004261, -0.00003384, -0.00002066, -0.00006757, 0.00016271, -0.00016176, 0.00004686, -0.00003422, -0.00030438, 0.00035643, -0.00129111, 0.00260153, -0.00084901, -0.00015152, -0.00000382, -0.00019684, 0.00021744, -0.00031297, 0.00021576, -0.00006757, 0.00001508, -0.00010135, 0.00009134, -0.00006234, 0.00011516, -0.00014202, 0.00008371, -0.00007189, },
		{ -0.00007355, 0.00008350, -0.00013396, 0.00010284, -0.00005005, 0.00006890, -0.00006815, -0.00000346, -0.00006923, 0.00019434, -0.00024151, 0.00013290, -0.00011719, -0.00015084, 0.00010914, -0.00109715, 0.00263967, -0.00109715, 0.00010914, -0.00015084, -0.00011719, 0.00013290, -0.00024151, 0.00019434, -0.00006923, -0.00000346, -0.00006815, 0.00006890, -0.00005005, 0.00010284, -0.00013396, 0.00008350, },
		{ 0.00007942, -0.00007189, 0.00008371, -0.00014202, 0.00011516, -0.00006234, 0.00009134, -0.00010135, 0.00001508, -0.00006757, 0.00021576, -0.00031297, 0.00021744, -0.00019684, -0.00000382, -0.00015152, -0.00084901, 0.00260153, -0.00129111, 0.00035643, -0.00030438, -0.00003422, 0.00004686, -0.00016176, 0.00016271, -0.00006757, -0.00002066, -0.00003384, 0.00004261, -0.00003444, 0.00008750, -0.00012192, },
		{ -0.00010637, 0.00007162, -0.00006698, 0.00008004, -0.00014579, 0.00012398, -0.00007084, 0.00010909, -0.00013218, 0.00003423, -0.00006266, 0.00022614, -0.00037341, 0.00029723, -0.00027011, 0.00013103, -0.00041551, -0.00055623, 0.00248855, -0.00142345, 0.00058086, -0.00045853, 0.00004887, -0.00003736, -0.00007680, 0.00012207, -0.00006266, -0.00003587, 0.00000025, 0.00001350, -0.00001611, 0.00006973, },
		{ 0.00005021, -0.00008790, 0.00006041, -0.00005900, 0.00007262, -0.00014514, 0.00012897, -0.00007523, 0.00012146, -0.00015945, 0.00005326, -0.00005469, 0.00022509, -0.00042050, 0.00036920, -0.00033419, 0.00024853, -0.00067271, -0.00023006, 0.00230509, -0.00148907, 0.00077381, -0.00060739, 0.00012891, -0.00011655, 0.00001012, 0.00007399, -0.00005469, -0.00004851, 0.00003281, -0.00001732, 0.00000422, },
		{ 0.00002579, 0.00002969, -0.00006723, 0.00004620, -0.00004827, 0.00006174, -0.00014007, 0.00012993, -0.00007533, 0.00012797, -0.00018210, 0.00007144, -0.00004395, 0.00021263, -0.00045243, 0.00043059, -0.00038661, 0.00034416, -0.00091321, 0.00011697, 0.00205820, -0.00148544, 0.00092785, -0.00074522, 0.00020280, -0.00018765, 0.00009565, 0.00002032, -0.00004395, -0.00005808, 0.00006259, -0.00004866, },
		{ -0.00007932, 0.00004777, 0.00000895, -0.00004514, 0.00002955, -0.00003519, 0.00004783, -0.00013079, 0.00012683, -0.00007114, 0.00012838, -0.00019928, 0.00008808, -0.00003087, 0.00018925, -0.00046797, 0.00047905, -0.00042536, 0.00041424, -0.00112779, 0.00047151, 0.00175735, -0.00141273, 0.00103707, -0.00086672, 0.00026771, -0.00024793, 0.00017650, -0.00003688, -0.00003087, -0.00006423, 0.00008845, },
		{ 0.00010939, -0.00010812, 0.00006930, -0.00001119, -0.00002249, 0.00001111, -0.00002026, 0.00003142, -0.00011766, 0.00011978, -0.00006283, 0.00012266, -0.00021032, 0.00010252, -0.00001595, 0.00015585, -0.00046652, 0.00051270, -0.00044895, 0.00045610, -0.00130819, 0.00081995, 0.00141412, -0.00127370, 0.00109727, -0.00096724, 0.00032115, -0.00029507, 0.00024957, -0.00009541, -0.00001595, -0.00006670, },
		{ -0.00006542, 0.00012461, -0.00013396, 0.00008956, -0.00002998, -0.00000014, -0.00000843, -0.00000407, 0.00001313, -0.00010118, 0.00010906, -0.00005071, 0.00011104, -0.00021479, 0.00011423, 0.00000024, 0.00011372, -0.00044815, 0.00053026, -0.00045648, 0.00046810, -0.00144749, 0.00114889, 0.00104170, -0.00107372, 0.00110614, -0.00104290, 0.00036105, -0.00032727, 0.00031205, -0.00015302, 0.00000024, },
		{ 0.00001708, -0.00006043, 0.00013352, -0.00015583, 0.00010777, -0.00004669, 0.00002104, -0.00002832, 0.00001277, -0.00000633, -0.00008198, 0.00009508, -0.00003524, 0.00009397, -0.00021253, 0.00012274, 0.00001708, 0.00006446, -0.00041356, 0.00053105, -0.00044765, 0.00044981, -0.00154032, 0.00144570, 0.00065439, -0.00082046, 0.00106334, -0.00109080, 0.00038590, -0.00034328, 0.00036154, -0.00020750, },
		{ -0.00025676, 0.00003392, -0.00005192, 0.00013578, -0.00017291, 0.00012324, -0.00006067, 0.00004023, -0.00004778, 0.00002961, -0.00002622, -0.00006080, 0.00007837, -0.00001703, 0.00007209, -0.00020362, 0.00012773, 0.00003392, 0.00000998, -0.00036407, 0.00051503, -0.00042280, 0.00040191, -0.00158313, 0.00169896, 0.00026708, -0.00052365, 0.00097051, -0.00110910, 0.00039473, -0.00034249, 0.00039613, },
		{ 0.00041450, -0.00029890, 0.00005012, -0.00004021, 0.00013131, -0.00018453, 0.00013536, -0.00007139, 0.00005672, -0.00006607, 0.00004580, -0.00004576, -0.00003846, 0.00005959, 0.00000323, 0.00004625, -0.00018840, 0.00012902, 0.00005012, -0.00004763, -0.00030159, 0.00048284, -0.00038289, 0.00032624, -0.00157426, 0.00189894, -0.00010535, -0.00019471, 0.00083121, -0.00109709, 0.00038721, -0.00032494, },
		{ -0.00029128, 0.00041595, -0.00033229, 0.00006504, -0.00002577, 0.00012027, -0.00019025, 0.00014368, -0.00007843, 0.00006985, -0.00008248, 0.00006072, -0.00006420, -0.00001581, 0.00003944, 0.00002477, 0.00001745, -0.00016746, 0.00012654, 0.00006504, -0.00010616, -0.00022852, 0.00043569, -0.00032945, 0.00022573, -0.00151405, 0.00203797, -0.00044858, 0.00015373, 0.00065081, -0.00105524, 0.00036362, },
		{ 0.00032487, -0.00024283, 0.00040041, -0.00035567, 0.00007812, -0.00000913, 0.00010310, -0.00018984, 0.00014787, -0.00008154, 0.00007912, -0.00009639, 0.00007380, -0.00008085, 0.00000628, 0.00001871, 0.00004674, -0.00001321, -0.00014160, 0.00012039, 0.00007812, -0.00016336, -0.00014767, 0.00037541, -0.00026454, 0.00010422, -0.00140483, 0.00211069, -0.00074942, 0.00050827, 0.00043623, -0.00098515, },
		{ -0.00088952, 0.00027245, -0.00018143, 0.00036848, -0.00036813, 0.00008885, 0.00000905, 0.00008044, -0.00018333, 0.00014776, -0.00008058, 0.00008419, -0.00010727, 0.00008454, -0.00009506, 0.00002696, -0.00000182, 0.00006831, -0.00004455, -0.00011182, 0.00011082, 0.00008885, -0.00021703, -0.00006214, 0.00030432, -0.00019065, -0.00003361, -0.00125079, 0.00211431, -0.00099632, 0.00085530, 0.00019573, },
		{ -0.00006147, -0.00077202, 0.00020837, -0.00010946, 0.00032139, -0.00036919, 0.00009683, 0.00002808, 0.00005317, -0.00017096, 0.00014338, -0.00007559, 0.00008485, -0.00011469, 0.00009251, -0.00010627, 0.00004542, -0.00002134, 0.00008865, -0.00007537, -0.00007926, 0.00009818, 0.00009683, -0.00026511, 0.00002478, 0.00022513, -0.00011062, -0.00018246, -0.00105785, 0.00204869, -0.00117978, 0.00118147, },
		{ 0.00147425, -0.00032546, -0.00063717, 0.00013509, -0.00002967, 0.00026095, -0.00035881, 0.00010174, 0.00004723, 0.00002234, -0.00015321, 0.00013488, -0.00006677, 0.00008107, -0.00011836, 0.00009743, -0.00011407, 0.00006098, -0.00003911, 0.00010697, -0.00010448, -0.00004517, 0.00008297, 0.00010174, -0.00030575, 0.00010974, 0.00014090, -0.00002752, -0.00033662, -0.00083341, 0.00191636, -0.00129275, },
		},
		{ { 0.00307126, -0.00118121, -0.00041112, 0.00050755, -0.00037614, 0.00002986, -0.00019864, 0.00039778, -0.00032818, 0.00014843, -0.00021214, 0.00032029, -0.00030900, 0.00023078, -0.00020311, 0.00025863, -0.00033250, 0.00025863, -0.00020311, 0.00023078, -0.00030900, 0.00032029, -0.00021214, 0.00014843, -0.00032818, 0.00039778, -0.00019864, 0.00002986, -0.00037614, 0.00050755, -0.00041112, -0.00118121, },
		{ -0.00118286, 0.00307292, -0.00118286, -0.00040947, 0.00050589, -0.00037448, 0.00002820, -0.00019698, 0.00039612, -0.00032652, 0.00014677, -0.00021048, 0.00031863, -0.00030734, 0.00022912, -0.00020145, 0.00025697, -0.00033084, 0.00025697, -0.00020145, 0.00022912, -0.00030734, 0.00031863, -0.00021048, 0.00014677, -0.00032652, 0.00039612, -0.00019698, 0.00002820, -0.00037448, 0.00050589, -0.00040947, },
		{ -0.00040455, -0.00118778, 0.00307783, -0.00118778, -0.00040455, 0.00050098, -0.00036957, 0.00002329, -0.00019207, 0.00039121, -0.00032161, 0.00014186, -0.00020557, 0.00031372, -0.00030243, 0.00022421, -0.00019654, 0.00025206, -0.00032593, 0.00025206, -0.00019654, 0.00022421, -0.00030243, 0.00031372, -0.00020557, 0.00014186, -0.00032161, 0.00039121, -0.00019207, 0.00002329, -0.00036957, 0.00050098, },
		{ 0.00049300, -0.00039658, -0.00119575, 0.00308580, -0.00119575, -0.00039658, 0.00049300, -0.00036159, 0.00001531, -0.00018410, 0.00038324, -0.00031364, 0.00013388, -0.00019759, 0.00030574, -0.00029445, 0.00021623, -0.00018856, 0.00024408, -0.00031795, 0.00024408, -0.00018856, 0.00021623, -0.00029445, 0.00030574, -0.00019759, 0.00013388, -0.00031364, 0.00038324, -0.00018410, 0.00001531, -0.00036159, },
		{ -0.00035086, 0.00048227, -0.00038584, -0.00120649, 0.00309654, -0.00120649, -0.00038584, 0.00048227, -0.00035086, 0.00000458, -0.00017336, 0.00037250, -0.00030290, 0.00012315, -0.00018686, 0.00029501, -0.00028372, 0.00020550, -0.00017783, 0.00023335, -0.00030722, 0.00023335, -0.00017783, 0.00020550, -0.00028372, 0.00029501, -0.00018686, 0.00012315, -0.00030290, 0.00037250, -0.00017336, 0.00000458, },
		{ -0.00000850, -0.00033778, 0.00046919, -0.00037276, -0.00121957, 0.00310962, -0.00121957, -0.00037276, 0.00046919, -0.00033778, -0.00000850, -0.00016028, 0.00035942, -0.00028982, 0.00011007, -0.00017378, 0.00028193, -0.00027064, 0.00019242, -0.00016474, 0.00022027, -0.00029414, 0.00022027, -0.00016474, 0.00019242, -0.00027064, 0.00028193, -0.00017378, 0.00011007, -0.00028982, 0.00035942, -0.00016028, },
		{ -0.00014536, -0.00002343, -0.00032285, 0.00045426, -0.00035784, -0.00123449, 0.00312454, -0.00123449, -0.00035784, 0.00045426, -0.00032285, -0.00002343, -0.00014536, 0.00034450, -0.00027490, 0.00009514, -0.00015885, 0.00026701, -0.00025572, 0.00017749, -0.00014982, 0.00020534, -0.00027921, 0.00020534, -0.00014982, 0.00017749, -0.00025572, 0.00026701, -0.00015885, 0.00009514, -0.00027490, 0.00034450, },
		{ 0.00032831, -0.00012917, -0.00003962, -0.00030666, 0.00043807, -0.00034165, -0.00125068, 0.00314073, -0.00125068, -0.00034165, 0.00043807, -0.00030666, -0.00003962, -0.00012917, 0.00032831, -0.00025871, 0.00007895, -0.00014266, 0.00025081, -0.00023952, 0.00016130, -0.00013363, 0.00018915, -0.00026302, 0.00018915, -0.00013363, 0.00016130, -0.00023952, 0.00025081, -0.00014266, 0.00007895, -0.00025871, },
		{ -0.00024187, 0.00031147, -0.00011233, -0.00005646, -0.00028982, 0.00042123, -0.00032481, -0.00126752, 0.00315757, -0.00126752, -0.00032481, 0.00042123, -0.00028982, -0.00005646, -0.00011233, 0.00031147, -0.00024187, 0.00006211, -0.00012582, 0.00023397, -0.00022268, 0.00014446, -0.00011679, 0.00017231, -0.00024618, 0.00017231, -0.00011679, 0.00014446, -0.00022268, 0.00023397, -0.00012582, 0.00006211, },
		{ 0.00004527, -0.00022503, 0.00029463, -0.00009549, -0.00007330, -0.00027298, 0.00040439, -0.00030797, -0.00128436, 0.00317441, -0.00128436, -0.00030797, 0.00040439, -0.00027298, -0.00007330, -0.00009549, 0.00029463, -0.00022503, 0.00004527, -0.00010898, 0.00021714, -0.00020584, 0.00012762, -0.00009995, 0.00015547, -0.00022934, 0.00015547, -0.00009995, 0.00012762, -0.00020584, 0.00021714, -0.00010898, },
		{ -0.00009279, 0.00002908, -0.00020883, 0.00027843, -0.00007929, -0.00008949, -0.00025679, 0.00038820, -0.00029178, -0.00130055, 0.00319061, -0.00130055, -0.00029178, 0.00038820, -0.00025679, -0.00008949, -0.00007929, 0.00027843, -0.00020883, 0.00002908, -0.00009279, 0.00020094, -0.00018965, 0.00011143, -0.00008376, 0.00013928, -0.00021315, 0.00013928, -0.00008376, 0.00011143, -0.00018965, 0.00020094, },
		{ 0.00018602, -0.00007787, 0.00001416, -0.00019391, 0.00026351, -0.00006437, -0.00010441, -0.00024187, 0.00037328, -0.00027685, -0.00131548, 0.00320553, -0.00131548, -0.00027685, 0.00037328, -0.00024187, -0.00010441, -0.00006437, 0.00026351, -0.00019391, 0.00001416, -0.00007787, 0.00018602, -0.00017473, 0.00009651, -0.00006883, 0.00012436, -0.00019823, 0.00012436, -0.00006883, 0.00009651, -0.00017473, },
		{ -0.00016165, 0.00017294, -0.00006478, 0.00000108, -0.00018083, 0.00025043, -0.00005129, -0.00011749, -0.00022879, 0.00036020, -0.00026377, -0.00132856, 0.00321861, -0.00132856, -0.00026377, 0.00036020, -0.00022879, -0.00011749, -0.00005129, 0.00025043, -0.00018083, 0.00000108, -0.00006478, 0.00017294, -0.00016165, 0.00008343, -0.00005575, 0.00011128, -0.00018515, 0.00011128, -0.00005575, 0.00008343, },
		{ 0.00007269, -0.00015091, 0.00016220, -0.00005405, -0.00000966, -0.00017010, 0.00023970, -0.00004056, -0.00012823, -0.00021805, 0.00034946, -0.00025304, -0.00133929, 0.00322934, -0.00133929, -0.00025304, 0.00034946, -0.00021805, -0.00012823, -0.00004056, 0.00023970, -0.00017010, -0.00000966, -0.00005405, 0.00016220, -0.00015091, 0.00007269, -0.00004502, 0.00010054, -0.00017441, 0.00010054, -0.00004502, },
		{ -0.00003704, 0.00006472, -0.00014294, 0.00015423, -0.00004607, -0.00001763, -0.00016212, 0.00023172, -0.00003258, -0.00013620, -0.00021008, 0.00034149, -0.00024506, -0.00134727, 0.00323732, -0.00134727, -0.00024506, 0.00034149, -0.00021008, -0.00013620, -0.00003258, 0.00023172, -0.00016212, -0.00001763, -0.00004607, 0.00015423, -0.00014294, 0.00006472, -0.00003704, 0.00009257, -0.00016644, 0.00009257, },
		{ 0.00008765, -0.00003213, 0.00005980, -0.00013803, 0.00014932, -0.00004116, -0.00002255, -0.00015721, 0.00022681, -0.00002767, -0.00014112, -0.00020516, 0.00033657, -0.00024015, -0.00135218, 0.00324223, -0.00135218, -0.00024015, 0.00033657, -0.00020516, -0.00014112, -0.00002767, 0.00022681, -0.00015721, -0.00002255, -0.00004116, 0.00014932, -0.00013803, 0.00005980, -0.00003213, 0.00008765, -0.00016152, },
		{ -0.00015987, 0.00008599, -0.00003047, 0.00005815, -0.00013637, 0.00014766, -0.00003950, -0.00002420, -0.00015555, 0.00022515, -0.00002601, -0.00014278, -0.00020351, 0.00033492, -0.00023849, -0.00135384, 0.00324389, -0.00135384, -0.00023849, 0.00033492, -0.00020351, -0.00014278, -0.00002601, 0.00022515, -0.00015555, -0.00002420, -0.00003950, 0.00014766, -0.00013637, 0.00005815, -0.00003047, 0.00008599, },
		{ 0.00008765, -0.00016152, 0.00008765, -0.00003213, 0.00005980, -0.00013803, 0.00014932, -0.00004116, -0.00002255, -0.00015721, 0.00022681, -0.00002767, -0.00014112, -0.00020516, 0.00033657, -0.00024015, -0.00135218, 0.00324223, -0.00135218, -0.00024015, 0.00033657, -0.00020516, -0.00014112, -0.00002767, 0.00022681, -0.00015721, -0.00002255, -0.00004116, 0.00014932, -0.00013803, 0.00005980, -0.00003213, },
		{ -0.00003704, 0.00009257, -0.00016644, 0.00009257, -0.00003704, 0.00006472, -0.00014294, 0.00015423, -0.00004607, -0.00001763, -0.00016212, 0.00023172, -0.00003258, -0.00013620, -0.00021008, 0.00034149, -0.00024506, -0.00134727, 0.00323732, -0.00134727, -0.00024506, 0.00034149, -0.00021008, -0.00013620, -0.00003258, 0.00023172, -0.00016212, -0.00001763, -0.00004607, 0.00015423, -0.00014294, 0.00006472, },
		{ 0.00007269, -0.00004502, 0.00010054, -0.00017441, 0.00010054, -0.00004502, 0.00007269, -0.00015091, 0.00016220, -0.00005405, -0.00000966, -0.00017010, 0.00023970, -0.00004056, -0.00012823, -0.00021805, 0.00034946, -0.00025304, -0.00133929, 0.00322934, -0.00133929, -0.00025304, 0.00034946, -0.00021805, -0.00012823, -0.00004056, 0.00023970, -0.00017010, -0.00000966, -0.00005405, 0.00016220, -0.00015091, },
		{ -0.00016165, 0.00008343, -0.00005575, 0.00011128, -0.00018515, 0.00011128, -0.00005575, 0.00008343, -0.00016165, 0.00017294, -0.00006478, 0.00000108, -0.00018083, 0.00025043, -0.00005129, -0.00011749, -0.00022879, 0.00036020, -0.00026377, -0.00132856, 0.00321861, -0.00132856, -0.00026377, 0.00036020, -0.00022879, -0.00011749, -0.00005129, 0.00025043, -0.00018083, 0.00000108, -0.00006478, 0.00017294, },
		{ 0.00018602, -0.00017473, 0.00009651, -0.00006883, 0.00012436, -0.00019823, 0.00012436, -0.00006883, 0.00009651, -0.00017473, 0.00018602, -0.00007787, 0.00001416, -0.00019391, 0.00026351, -0.00006437, -0.00010441, -0.00024187, 0.00037328, -0.00027685, -0.00131548, 0.00320553, -0.00131548, -0.00027685, 0.00037328, -0.00024187, -0.00010441, -0.00006437, 0.00026351, -0.00019391, 0.00001416, -0.00007787, },
		{ -0.00009279, 0.00020094, -0.00018965, 0.00011143, -0.00008376, 0.00013928, -0.00021315, 0.00013928, -0.00008376, 0.00011143, -0.00018965, 0.00020094, -0.00009279, 0.00002908, -0.00020883, 0.00027843, -0.00007929, -0.00008949, -0.00025679, 0.00038820, -0.00029178, -0.00130055, 0.00319061, -0.00130055, -0.00029178, 0.00038820, -0.00025679, -0.00008949, -0.00007929, 0.00027843, -0.00020883, 0.00002908, },
		{ 0.00004527, -0.00010898, 0.00021714, -0.00020584, 0.00012762, -0.00009995, 0.00015547, -0.00022934, 0.00015547, -0.00009995, 0.00012762, -0.00020584, 0.00021714, -0.00010898, 0.00004527, -0.00022503, 0.00029463, -0.00009549, -0.00007330, -0.00027298, 0.00040439, -0.00030797, -0.00128436, 0.00317441, -0.00128436, -0.00030797, 0.00040439, -0.00027298, -0.00007330, -0.00009549, 0.00029463, -0.00022503, },
		{ -0.00024187, 0.00006211, -0.00012582, 0.00023397, -0.00022268, 0.00014446, -0.00011679, 0.00017231, -0.00024618, 0.00017231, -0.00011679, 0.00014446, -0.00022268, 0.00023397, -0.00012582, 0.00006211, -0.00024187, 0.00031147, -0.00011233, -0.00005646, -0.00028982, 0.00042123, -0.00032481, -0.00126752, 0.00315757, -0.00126752, -0.00032481, 0.00042123, -0.00028982, -0.00005646, -0.00011233, 0.00031147, },
		{ 0.00032831, -0.00025871, 0.00007895, -0.00014266, 0.00025081, -0.00023952, 0.00016130, -0.00013363, 0.00018915, -0.00026302, 0.00018915, -0.00013363, 0.00016130, -0.00023952, 0.00025081, -0.00014266, 0.00007895, -0.00025871, 0.00032831, -0.00012917, -0.00003962, -0.00030666, 0.00043807, -0.00034165, -0.00125068, 0.00314073, -0.00125068, -0.00034165, 0.00043807, -0.00030666, -0.00003962, -0.00012917, },
		{ -0.00014536, 0.00034450, -0.00027490, 0.00009514, -0.00015885, 0.00026701, -0.00025572, 0.00017749, -0.00014982, 0.00020534, -0.00027921, 0.00020534, -0.00014982, 0.00017749, -0.00025572, 0.00026701, -0.00015885, 0.00009514, -0.00027490, 0.00034450, -0.00014536, -0.00002343, -0.00032285, 0.00045426, -0.00035784, -0.00123449, 0.00312454, -0.00123449, -0.00035784, 0.00045426, -0.00032285, -0.00002343, },
		{ -0.00000850, -0.00016028, 0.00035942, -0.00028982, 0.00011007, -0.00017378, 0.00028193, -0.00027064, 0.00019242, -0.00016474, 0.00022027, -0.00029414, 0.00022027, -0.00016474, 0.00019242, -0.00027064, 0.00028193, -0.00017378, 0.00011007, -0.00028982, 0.00035942, -0.00016028, -0.00000850, -0.00033778, 0.00046919, -0.00037276, -0.00121957, 0.00310962, -0.00121957, -0.00037276, 0.00046919, -0.00033778, },
		{ -0.00035086, 0.00000458, -0.00017336, 0.00037250, -0.00030290, 0.00012315, -0.00018686, 0.00029501, -0.00028372, 0.00020550, -0.00017783, 0.00023335, -0.00030722, 0.00023335, -0.00017783, 0.00020550, -0.00028372, 0.00029501, -0.00018686, 0.00012315, -0.00030290, 0.00037250, -0.00017336, 0.00000458, -0.00035086, 0.00048227, -0.00038584, -0.00120649, 0.00309654, -0.00120649, -0.00038584, 0.00048227, },
		{ 0.00049300, -0.00036159, 0.00001531, -0.00018410, 0.00038324, -0.00031364, 0.00013388, -0.00019759, 0.00030574, -0.00029445, 0.00021623, -0.00018856, 0.00024408, -0.00031795, 0.00024408, -0.00018856, 0.00021623, -0.00029445, 0.00030574, -0.00019759, 0.00013388, -0.00031364, 0.00038324, -0.00018410, 0.00001531, -0.00036159, 0.00049300, -0.00039658, -0.00119575, 0.00308580, -0.00119575, -0.00039658, },
		{ -0.00040455, 0.00050098, -0.00036957, 0.00002329, -0.00019207, 0.00039121, -0.00032161, 0.00014186, -0.00020557, 0.00031372, -0.00030243, 0.00022421, -0.00019654, 0.00025206, -0.00032593, 0.00025206, -0.00019654, 0.00022421, -0.00030243, 0.00031372, -0.00020557, 0.00014186, -0.00032161, 0.00039121, -0.00019207, 0.00002329, -0.00036957, 0.00050098, -0.00040455, -0.00118778, 0.00307783, -0.00118778, },
		{ -0.00118286, -0.00040947, 0.00050589, -0.00037448, 0.00002820, -0.00019698, 0.00039612, -0.00032652, 0.00014677, -0.00021048, 0.00031863, -0.00030734, 0.00022912, -0.00020145, 0.00025697, -0.00033084, 0.00025697, -0.00020145, 0.00022912, -0.00030734, 0.00031863, -0.00021048, 0.00014677, -0.00032652, 0.00039612, -0.00019698, 0.00002820, -0.00037448, 0.00050589, -0.00040947, -0.00118286, 0.00307292, },
		},
	}
	;

	float f128_4[16][128][128] = {
		{ { -0.00009627, -0.00002363, 0.00004177, 0.00002579, 0.00000853, 0.00000714, -0.00000866, 0.00000996, 0.00000675, 0.00000081, 0.00000440, 0.00000672, 0.00000269, 0.00000293, -0.00000775, 0.00000991, 0.00000752, 0.00000403, 0.00001015, 0.00000902, 0.00001128, 0.00002444, 0.00000324, 0.00000114, -0.00000259, -0.00002199, -0.00001401, 0.00000444, -0.00000793, -0.00000457, -0.00000516, -0.00000873, 0.00000124, -0.00000301, -0.00000688, 0.00000663, -0.00000097, -0.00000422, 0.00000320, -0.00000537, -0.00000019, 0.00000326, -0.00000369, -0.00000371, -0.00000181, -0.00000366, -0.00000147, -0.00000989, 0.00000537, 0.00000973, 0.00000021, -0.00000424, -0.00000411, -0.00000660, 0.00001015, -0.00000191, 0.00000052, -0.00000009, -0.00001108, 0.00000083, 0.00001675, -0.00001342, -0.00000024, 0.00000333, -0.00000518, 0.00000333, -0.00000024, -0.00001342, 0.00001675, 0.00000083, -0.00001108, -0.00000009, 0.00000052, -0.00000191, 0.00001015, -0.00000660, -0.00000411, -0.00000424, 0.00000021, 0.00000973, 0.00000537, -0.00000989, -0.00000147, -0.00000366, -0.00000181, -0.00000371, -0.00000369, 0.00000326, -0.00000019, -0.00000537, 0.00000320, -0.00000422, -0.00000097, 0.00000663, -0.00000688, -0.00000301, 0.00000124, -0.00000873, -0.00000516, -0.00000457, -0.00000793, 0.00000444, -0.00001401, -0.00002199, -0.00000259, 0.00000114, 0.00000324, 0.00002444, 0.00001128, 0.00000902, 0.00001015, 0.00000403, 0.00000752, 0.00000991, -0.00000775, 0.00000293, 0.00000269, 0.00000672, 0.00000440, 0.00000081, 0.00000675, 0.00000996, -0.00000866, 0.00000714, 0.00000853, 0.00002579, 0.00004177, -0.00002363, },
		{ -0.00002138, -0.00009661, -0.00002509, 0.00003647, 0.00002562, 0.00001006, 0.00000690, -0.00001096, 0.00000986, 0.00000657, 0.00000030, 0.00000416, 0.00000607, 0.00000279, 0.00000465, -0.00001028, 0.00000784, 0.00000961, 0.00000186, 0.00000827, 0.00001053, 0.00000637, 0.00002275, 0.00000786, 0.00000062, 0.00000306, -0.00001538, -0.00002147, 0.00000278, -0.00000272, -0.00000784, -0.00000426, -0.00000764, -0.00000155, 0.00000003, -0.00000879, 0.00000322, 0.00000438, -0.00000609, 0.00000152, -0.00000154, -0.00000258, 0.00000266, -0.00000108, -0.00000516, -0.00000079, -0.00000470, -0.00000053, -0.00000858, -0.00000023, 0.00000990, 0.00000482, -0.00000530, -0.00000097, -0.00001012, 0.00000654, 0.00000332, -0.00000136, 0.00000074, -0.00000589, -0.00001057, 0.00002064, -0.00000632, -0.00000714, 0.00000537, -0.00000337, -0.00000127, 0.00000539, -0.00001412, 0.00000709, 0.00001042, -0.00001003, -0.00000507, 0.00000379, -0.00000550, 0.00000926, -0.00000049, -0.00000638, -0.00000428, -0.00000201, 0.00000630, 0.00001037, -0.00000828, -0.00000390, -0.00000229, -0.00000278, -0.00000266, -0.00000475, 0.00000161, 0.00000304, -0.00000738, 0.00000164, -0.00000021, -0.00000518, 0.00000627, -0.00000245, -0.00000576, 0.00000204, -0.00000725, -0.00000792, -0.00000132, -0.00001005, 0.00000149, -0.00000602, -0.00002300, -0.00001120, 0.00000027, -0.00000000, 0.00002311, 0.00001692, 0.00000933, 0.00001015, 0.00000716, 0.00000636, 0.00001015, -0.00000454, 0.00000238, 0.00000183, 0.00000738, 0.00000485, 0.00000094, 0.00000731, 0.00000995, -0.00000704, 0.00000839, 0.00000755, 0.00002456, 0.00004714, },
		{ 0.00005176, -0.00001807, -0.00009737, -0.00002631, 0.00003187, 0.00002428, 0.00001122, 0.00000796, -0.00001343, 0.00000918, 0.00000686, -0.00000042, 0.00000384, 0.00000580, 0.00000203, 0.00000670, -0.00001116, 0.00000429, 0.00001124, 0.00000166, 0.00000483, 0.00001224, 0.00000389, 0.00001813, 0.00001232, 0.00000139, 0.00000423, -0.00000607, -0.00002438, -0.00000391, 0.00000294, -0.00000842, -0.00000632, -0.00000471, -0.00000424, 0.00000102, -0.00000726, -0.00000226, 0.00000788, -0.00000424, -0.00000250, 0.00000172, -0.00000229, -0.00000030, 0.00000198, -0.00000583, -0.00000067, -0.00000460, -0.00000128, -0.00000596, -0.00000369, 0.00000594, 0.00001035, -0.00000546, 0.00000074, -0.00000979, 0.00000031, 0.00000648, 0.00000098, -0.00000304, 0.00000229, -0.00001842, 0.00001656, 0.00000430, -0.00001126, 0.00000324, 0.00000078, -0.00000522, 0.00000676, -0.00000899, -0.00000348, 0.00001374, -0.00000292, -0.00001084, 0.00000488, -0.00000540, 0.00000480, 0.00000518, -0.00000569, -0.00000601, -0.00000234, 0.00000188, 0.00001248, -0.00000383, -0.00000636, -0.00000178, -0.00000294, -0.00000255, -0.00000437, -0.00000094, 0.00000506, -0.00000658, -0.00000202, 0.00000331, -0.00000624, 0.00000267, 0.00000201, -0.00000609, 0.00000034, -0.00000432, -0.00001037, -0.00000019, -0.00000877, -0.00000337, -0.00000115, -0.00001841, -0.00001930, -0.00000393, -0.00000204, 0.00001995, 0.00002142, 0.00001212, 0.00000917, 0.00000988, 0.00000691, 0.00000915, -0.00000180, 0.00000329, 0.00000074, 0.00000764, 0.00000568, 0.00000074, 0.00000797, 0.00001024, -0.00000617, 0.00000995, 0.00000777, 0.00002235, },
		{ 0.00002003, 0.00005502, -0.00001403, -0.00009796, -0.00002778, 0.00002805, 0.00002246, 0.00001130, 0.00001003, -0.00001534, 0.00000769, 0.00000739, -0.00000102, 0.00000315, 0.00000603, 0.00000080, 0.00000816, -0.00001018, 0.00000046, 0.00001111, 0.00000339, 0.00000117, 0.00001239, 0.00000435, 0.00001220, 0.00001436, 0.00000498, 0.00000218, 0.00000166, -0.00002072, -0.00001301, 0.00000539, -0.00000489, -0.00001028, -0.00000228, -0.00000465, -0.00000092, -0.00000358, -0.00000682, 0.00000741, 0.00000070, -0.00000627, 0.00000208, 0.00000110, -0.00000418, 0.00000368, -0.00000472, -0.00000168, -0.00000363, -0.00000255, -0.00000425, -0.00000351, -0.00000075, 0.00001404, -0.00000293, 0.00000027, -0.00000690, -0.00000486, 0.00000452, 0.00000768, -0.00000876, 0.00000875, -0.00001923, 0.00000617, 0.00001360, -0.00000992, -0.00000195, 0.00000425, -0.00000586, 0.00000360, -0.00000153, -0.00001003, 0.00000960, 0.00000664, -0.00001321, 0.00000197, -0.00000235, -0.00000026, 0.00000759, -0.00000173, -0.00000826, -0.00000241, -0.00000138, 0.00001105, 0.00000170, -0.00000705, -0.00000274, -0.00000233, -0.00000305, -0.00000345, -0.00000288, 0.00000490, -0.00000374, -0.00000554, 0.00000434, -0.00000416, -0.00000189, 0.00000402, -0.00000342, -0.00000258, -0.00000198, -0.00001082, -0.00000151, -0.00000584, -0.00000706, -0.00000088, -0.00001092, -0.00002343, -0.00001133, -0.00000450, 0.00001633, 0.00002370, 0.00001682, 0.00000863, 0.00001125, 0.00000897, 0.00000803, -0.00000034, 0.00000527, 0.00000014, 0.00000731, 0.00000675, 0.00000046, 0.00000841, 0.00001099, -0.00000574, 0.00001111, 0.00000930, },
		{ 0.00001163, 0.00001853, 0.00005690, -0.00001003, -0.00009789, -0.00002952, 0.00002461, 0.00002092, 0.00001008, 0.00001237, -0.00001609, 0.00000559, 0.00000772, -0.00000121, 0.00000203, 0.00000655, -0.00000032, 0.00000846, -0.00000794, -0.00000221, 0.00000880, 0.00000585, -0.00000104, 0.00001023, 0.00000664, 0.00000742, 0.00001254, 0.00001067, 0.00000013, 0.00000452, -0.00001181, -0.00002000, 0.00000219, 0.00000155, -0.00001327, -0.00000265, -0.00000209, -0.00000453, -0.00000044, -0.00000802, 0.00000310, 0.00000599, -0.00000695, -0.00000123, 0.00000597, -0.00000659, 0.00000270, -0.00000198, -0.00000296, -0.00000297, -0.00000282, -0.00000481, -0.00000032, -0.00000712, 0.00001347, 0.00000228, -0.00000087, -0.00000447, -0.00000570, -0.00000288, 0.00001578, -0.00001223, 0.00000999, -0.00001315, -0.00000595, 0.00001733, -0.00000343, -0.00000691, 0.00000431, -0.00000277, -0.00000163, 0.00000396, -0.00001013, 0.00000065, 0.00001361, -0.00000980, -0.00000385, 0.00000077, -0.00000293, 0.00000588, 0.00000372, -0.00000895, -0.00000361, -0.00000271, 0.00000728, 0.00000593, -0.00000500, -0.00000460, -0.00000175, -0.00000339, -0.00000293, -0.00000344, 0.00000295, -0.00000066, -0.00000704, 0.00000258, -0.00000073, -0.00000483, 0.00000260, 0.00000090, -0.00000455, -0.00000171, -0.00000917, -0.00000389, -0.00000357, -0.00000783, -0.00000407, -0.00000421, -0.00002206, -0.00001962, -0.00000918, 0.00001277, 0.00002377, 0.00002212, 0.00000972, 0.00001125, 0.00001165, 0.00000783, -0.00000020, 0.00000750, 0.00000053, 0.00000658, 0.00000781, 0.00000032, 0.00000849, 0.00001209, -0.00000524, 0.00001149, },
		{ 0.00001117, 0.00001401, 0.00001834, 0.00005806, -0.00000694, -0.00009711, -0.00003109, 0.00002098, 0.00002008, 0.00000799, 0.00001407, -0.00001553, 0.00000347, 0.00000744, -0.00000095, 0.00000066, 0.00000696, -0.00000087, 0.00000767, -0.00000557, -0.00000279, 0.00000506, 0.00000735, -0.00000086, 0.00000643, 0.00000866, 0.00000567, 0.00000737, 0.00001589, 0.00000122, 0.00000211, -0.00000184, -0.00002088, -0.00000608, 0.00000732, -0.00001238, -0.00000642, 0.00000191, -0.00000720, -0.00000015, -0.00000561, -0.00000254, 0.00000836, -0.00000334, -0.00000661, 0.00000951, -0.00000568, -0.00000071, 0.00000086, -0.00000297, -0.00000378, -0.00000137, -0.00000722, 0.00000348, -0.00001018, 0.00000834, 0.00000778, 0.00000015, -0.00000514, -0.00000140, -0.00001284, 0.00002077, -0.00001031, 0.00000563, -0.00000356, -0.00001480, 0.00001410, 0.00000497, -0.00000814, 0.00000035, 0.00000191, -0.00000542, 0.00000481, -0.00000495, -0.00000821, 0.00001429, -0.00000155, -0.00000912, 0.00000105, -0.00000211, 0.00000164, 0.00000780, -0.00000668, -0.00000588, -0.00000287, 0.00000322, 0.00000723, -0.00000083, -0.00000593, -0.00000208, -0.00000310, -0.00000313, -0.00000287, 0.00000052, 0.00000101, -0.00000605, -0.00000057, 0.00000171, -0.00000481, -0.00000124, 0.00000447, -0.00000402, -0.00000343, -0.00000677, -0.00000545, -0.00000336, -0.00000605, -0.00000807, -0.00000101, -0.00001631, -0.00002553, -0.00001665, 0.00000852, 0.00002222, 0.00002674, 0.00001283, 0.00001063, 0.00001393, 0.00000897, -0.00000079, 0.00000916, 0.00000191, 0.00000591, 0.00000863, 0.00000039, 0.00000827, 0.00001323, -0.00000432, },
		{ -0.00000290, 0.00001058, 0.00001576, 0.00001934, 0.00005950, -0.00000522, -0.00009619, -0.00003180, 0.00001685, 0.00001979, 0.00000577, 0.00001455, -0.00001412, 0.00000204, 0.00000639, -0.00000051, -0.00000059, 0.00000694, -0.00000074, 0.00000634, -0.00000413, -0.00000130, 0.00000146, 0.00000657, 0.00000130, 0.00000278, 0.00000857, 0.00000710, 0.00000117, 0.00001772, 0.00000646, -0.00000282, 0.00000439, -0.00001463, -0.00001559, 0.00000838, -0.00000676, -0.00001160, 0.00000444, -0.00000678, -0.00000296, -0.00000169, -0.00000612, 0.00000602, 0.00000312, -0.00001086, 0.00000933, -0.00000155, -0.00000455, 0.00000168, -0.00000064, -0.00000612, 0.00000102, -0.00000965, 0.00000531, -0.00000890, 0.00000111, 0.00000990, 0.00000548, -0.00000932, 0.00000572, -0.00002065, 0.00001917, -0.00000285, -0.00000159, 0.00000498, -0.00001741, 0.00000594, 0.00001093, -0.00000410, -0.00000552, 0.00000486, -0.00000543, 0.00000141, 0.00000170, -0.00001258, 0.00000838, 0.00000765, -0.00001026, -0.00000246, 0.00000078, -0.00000221, 0.00000840, -0.00000189, -0.00000783, -0.00000328, 0.00000044, 0.00000555, 0.00000365, -0.00000533, -0.00000350, -0.00000242, -0.00000368, -0.00000200, -0.00000115, 0.00000067, -0.00000358, -0.00000310, 0.00000175, -0.00000233, -0.00000524, 0.00000541, -0.00000104, -0.00000567, -0.00000536, -0.00000511, -0.00000493, -0.00000352, -0.00001044, -0.00000179, -0.00000918, -0.00002669, -0.00002553, 0.00000232, 0.00001939, 0.00002998, 0.00001756, 0.00001046, 0.00001523, 0.00001120, -0.00000125, 0.00000986, 0.00000386, 0.00000572, 0.00000924, 0.00000058, 0.00000793, 0.00001415, },
		{ 0.00001476, -0.00000118, 0.00001020, 0.00001668, 0.00002084, 0.00006211, -0.00000459, -0.00009614, -0.00003123, 0.00001242, 0.00001950, 0.00000414, 0.00001380, -0.00001267, 0.00000181, 0.00000488, -0.00000031, -0.00000145, 0.00000639, -0.00000021, 0.00000520, -0.00000412, 0.00000131, -0.00000043, 0.00000341, 0.00000384, 0.00000108, 0.00000586, 0.00001007, -0.00000314, 0.00001484, 0.00001406, -0.00000604, 0.00000421, -0.00000404, -0.00002112, 0.00000282, 0.00000128, -0.00001462, 0.00000326, -0.00000326, -0.00000678, 0.00000058, -0.00000551, -0.00000003, 0.00000900, -0.00001108, 0.00000524, 0.00000351, -0.00000602, -0.00000082, 0.00000357, -0.00000863, 0.00000263, -0.00001023, 0.00000417, -0.00000498, -0.00000411, 0.00000586, 0.00001474, -0.00001456, 0.00001152, -0.00002235, 0.00001054, 0.00000712, -0.00000757, 0.00000922, -0.00001412, -0.00000296, 0.00001136, 0.00000370, -0.00000968, 0.00000379, -0.00000194, -0.00000336, 0.00000576, -0.00001084, -0.00000093, 0.00001318, -0.00000582, -0.00000794, 0.00000283, -0.00000329, 0.00000551, 0.00000333, -0.00000769, -0.00000473, -0.00000080, 0.00000225, 0.00000644, -0.00000241, -0.00000525, -0.00000204, -0.00000401, -0.00000156, -0.00000156, -0.00000106, -0.00000130, -0.00000361, -0.00000042, 0.00000075, -0.00000732, 0.00000338, 0.00000280, -0.00000656, -0.00000587, -0.00000324, -0.00000686, -0.00000206, -0.00001026, -0.00000496, -0.00000385, -0.00002289, -0.00003307, -0.00000645, 0.00001486, 0.00003157, 0.00002315, 0.00001155, 0.00001561, 0.00001385, -0.00000094, 0.00000971, 0.00000578, 0.00000617, 0.00000985, 0.00000080, 0.00000757, },
		{ 0.00000722, 0.00001513, 0.00000055, 0.00001032, 0.00001704, 0.00002200, 0.00006617, -0.00000412, -0.00009786, -0.00002960, 0.00000841, 0.00001862, 0.00000330, 0.00001239, -0.00001204, 0.00000278, 0.00000351, -0.00000074, -0.00000189, 0.00000553, 0.00000026, 0.00000476, -0.00000531, 0.00000369, 0.00000020, -0.00000098, 0.00000484, 0.00000204, 0.00000174, 0.00001209, -0.00000377, 0.00000843, 0.00002043, -0.00000438, -0.00000137, 0.00000552, -0.00001905, -0.00000750, 0.00000724, -0.00001260, -0.00000144, 0.00000092, -0.00000854, -0.00000085, -0.00000128, -0.00000648, 0.00001087, -0.00000665, -0.00000034, 0.00000617, -0.00000337, -0.00000596, 0.00000760, -0.00000931, 0.00000208, -0.00000856, 0.00000142, -0.00000199, -0.00000382, -0.00000419, 0.00002453, -0.00001686, 0.00001241, -0.00001683, -0.00000213, 0.00001511, -0.00000929, 0.00000875, -0.00000792, -0.00000859, 0.00000612, 0.00001153, -0.00000911, -0.00000106, 0.00000239, -0.00000608, 0.00000522, -0.00000484, -0.00000888, 0.00001229, 0.00000243, -0.00001190, 0.00000170, -0.00000133, 0.00000113, 0.00000642, -0.00000477, -0.00000670, -0.00000142, -0.00000091, 0.00000649, 0.00000190, -0.00000602, -0.00000252, -0.00000387, -0.00000170, -0.00000107, -0.00000296, -0.00000050, -0.00000206, -0.00000333, 0.00000251, -0.00000674, -0.00000033, 0.00000541, -0.00000514, -0.00000778, -0.00000125, -0.00000764, -0.00000242, -0.00000834, -0.00000809, -0.00000206, -0.00001613, -0.00003660, -0.00001689, 0.00000773, 0.00003121, 0.00002886, 0.00001424, 0.00001556, 0.00001631, 0.00000033, 0.00000921, 0.00000726, 0.00000705, 0.00001074, 0.00000105, },
		{ 0.00000151, 0.00000685, 0.00001532, 0.00000213, 0.00001087, 0.00001736, 0.00002229, 0.00007124, -0.00000265, -0.00010162, -0.00002777, 0.00000558, 0.00001700, 0.00000292, 0.00001113, -0.00001271, 0.00000439, 0.00000301, -0.00000187, -0.00000215, 0.00000469, 0.00000030, 0.00000509, -0.00000697, 0.00000475, 0.00000301, -0.00000473, 0.00000318, 0.00000477, -0.00000164, 0.00001136, -0.00000110, 0.00000154, 0.00002236, 0.00000227, -0.00000813, 0.00000915, -0.00000966, -0.00001751, 0.00000697, -0.00000560, -0.00000662, 0.00000239, -0.00000636, -0.00000538, 0.00000360, -0.00000969, 0.00000752, 0.00000006, -0.00000361, 0.00000414, 0.00000272, -0.00001105, 0.00000889, -0.00000701, -0.00000036, -0.00000624, 0.00000002, -0.00000318, 0.00000265, -0.00001665, 0.00002995, -0.00001309, 0.00000748, -0.00000644, -0.00001396, 0.00001760, -0.00000644, 0.00000577, -0.00000242, -0.00000917, -0.00000206, 0.00001541, -0.00000331, -0.00000689, 0.00000428, -0.00000499, 0.00000113, 0.00000149, -0.00001179, 0.00000563, 0.00001039, -0.00001132, -0.00000273, 0.00000176, -0.00000199, 0.00000602, -0.00000005, -0.00000771, -0.00000252, -0.00000287, 0.00000418, 0.00000586, -0.00000485, -0.00000374, -0.00000357, -0.00000208, -0.00000042, -0.00000404, -0.00000143, 0.00000043, -0.00000524, 0.00000206, -0.00000434, -0.00000363, 0.00000545, -0.00000191, -0.00000960, -0.00000049, -0.00000676, -0.00000395, -0.00000634, -0.00000945, -0.00000346, -0.00000944, -0.00003499, -0.00002661, -0.00000251, 0.00002812, 0.00003405, 0.00001855, 0.00001571, 0.00001823, 0.00000238, 0.00000884, 0.00000824, 0.00000802, 0.00001199, },
		{ 0.00001340, 0.00000238, 0.00000654, 0.00001534, 0.00000352, 0.00001159, 0.00001808, 0.00002173, 0.00007642, 0.00000065, -0.00010685, -0.00002697, 0.00000428, 0.00001508, 0.00000228, 0.00001065, -0.00001448, 0.00000574, 0.00000381, -0.00000339, -0.00000265, 0.00000416, -0.00000019, 0.00000585, -0.00000827, 0.00000408, 0.00000677, -0.00000628, -0.00000086, 0.00000731, -0.00000252, 0.00000788, 0.00000262, -0.00000265, 0.00001901, 0.00001087, -0.00001117, 0.00000519, 0.00000265, -0.00002159, -0.00000052, 0.00000282, -0.00000810, -0.00000079, -0.00000114, -0.00001004, 0.00000565, -0.00000807, 0.00000097, 0.00000488, -0.00000170, -0.00000211, 0.00000917, -0.00001285, 0.00000615, -0.00000265, -0.00000249, -0.00000606, 0.00000256, -0.00000937, 0.00001241, -0.00002595, 0.00002714, -0.00000303, -0.00000095, 0.00000440, -0.00002072, 0.00001390, -0.00000139, 0.00000333, 0.00000019, -0.00000572, -0.00000921, 0.00001335, 0.00000523, -0.00000997, 0.00000203, -0.00000102, -0.00000340, 0.00000457, -0.00000900, -0.00000314, 0.00001394, -0.00000566, -0.00000815, 0.00000323, -0.00000216, 0.00000279, 0.00000422, -0.00000643, -0.00000438, -0.00000376, 0.00000088, 0.00000790, -0.00000181, -0.00000484, -0.00000366, -0.00000228, -0.00000016, -0.00000403, -0.00000338, 0.00000236, -0.00000521, -0.00000006, -0.00000189, -0.00000496, 0.00000309, 0.00000151, -0.00000987, -0.00000138, -0.00000489, -0.00000530, -0.00000556, -0.00000880, -0.00000622, -0.00000531, -0.00002929, -0.00003280, -0.00001502, 0.00002121, 0.00003783, 0.00002435, 0.00001662, 0.00001957, 0.00000494, 0.00000888, 0.00000890, 0.00000884, },
		{ 0.00000953, 0.00001467, 0.00000368, 0.00000654, 0.00001515, 0.00000478, 0.00001228, 0.00001930, 0.00002081, 0.00008086, 0.00000578, -0.00011234, -0.00002809, 0.00000403, 0.00001381, 0.00000082, 0.00001098, -0.00001664, 0.00000598, 0.00000582, -0.00000464, -0.00000374, 0.00000398, -0.00000104, 0.00000655, -0.00000868, 0.00000207, 0.00001001, -0.00000506, -0.00000570, 0.00000771, -0.00000060, 0.00000336, 0.00000469, -0.00000274, 0.00001244, 0.00001704, -0.00000778, -0.00000369, 0.00001163, -0.00001690, -0.00001187, 0.00000730, -0.00000337, -0.00000735, 0.00000372, -0.00001144, 0.00000335, -0.00000324, -0.00000462, 0.00000434, 0.00000525, -0.00000921, 0.00001208, -0.00000971, 0.00000071, 0.00000084, -0.00000136, -0.00001006, 0.00000936, -0.00001807, 0.00002024, -0.00002726, 0.00001562, 0.00001011, -0.00000864, 0.00001163, -0.00002083, 0.00000631, 0.00000247, 0.00000339, 0.00000007, -0.00000107, -0.00001232, 0.00000640, 0.00001242, -0.00000787, -0.00000338, 0.00000288, -0.00000532, 0.00000322, -0.00000297, -0.00000951, 0.00001126, 0.00000265, -0.00001119, 0.00000136, 0.00000017, -0.00000102, 0.00000593, -0.00000274, -0.00000618, -0.00000446, -0.00000199, 0.00000748, 0.00000201, -0.00000476, -0.00000441, -0.00000221, -0.00000036, -0.00000332, -0.00000523, 0.00000280, -0.00000353, -0.00000239, -0.00000087, -0.00000408, -0.00000023, 0.00000338, -0.00000813, -0.00000313, -0.00000333, -0.00000541, -0.00000621, -0.00000719, -0.00000833, -0.00000448, -0.00002213, -0.00003371, -0.00002745, 0.00000979, 0.00003874, 0.00003132, 0.00001885, 0.00002040, 0.00000773, 0.00000943, 0.00000945, },
		{ 0.00000993, 0.00001026, 0.00001565, 0.00000509, 0.00000718, 0.00001482, 0.00000587, 0.00001287, 0.00002082, 0.00002012, 0.00008420, 0.00001201, -0.00011682, -0.00003116, 0.00000365, 0.00001402, -0.00000144, 0.00001154, -0.00001830, 0.00000475, 0.00000836, -0.00000493, -0.00000548, 0.00000394, -0.00000190, 0.00000673, -0.00000814, -0.00000044, 0.00001164, -0.00000173, -0.00000935, 0.00000514, 0.00000276, 0.00000017, 0.00000362, 0.00000018, 0.00000634, 0.00001767, 0.00000094, -0.00001197, 0.00001271, -0.00000537, -0.00002083, 0.00000393, 0.00000632, -0.00001315, 0.00000452, -0.00000827, -0.00000163, 0.00000091, -0.00000550, -0.00000195, 0.00001380, -0.00001271, 0.00000931, -0.00000306, -0.00000377, -0.00000004, 0.00000476, -0.00001789, 0.00001779, -0.00002465, 0.00002130, -0.00001914, -0.00000109, 0.00002115, -0.00001189, 0.00001357, -0.00001580, -0.00000135, 0.00000270, 0.00000570, -0.00000083, 0.00000211, -0.00001082, -0.00000206, 0.00001473, -0.00000102, -0.00000873, 0.00000377, -0.00000356, -0.00000093, 0.00000240, -0.00001064, 0.00000387, 0.00000952, -0.00000951, -0.00000323, 0.00000281, -0.00000298, 0.00000439, 0.00000194, -0.00000651, -0.00000572, -0.00000384, 0.00000526, 0.00000504, -0.00000294, -0.00000550, -0.00000217, -0.00000072, -0.00000256, -0.00000623, 0.00000176, -0.00000136, -0.00000354, -0.00000172, -0.00000202, -0.00000275, 0.00000293, -0.00000519, -0.00000438, -0.00000303, -0.00000420, -0.00000742, -0.00000601, -0.00000874, -0.00000597, -0.00001633, -0.00002960, -0.00003668, -0.00000543, 0.00003488, 0.00003855, 0.00002307, 0.00002093, 0.00001052, 0.00001058, },
		{ 0.00001249, 0.00001035, 0.00001114, 0.00001654, 0.00000613, 0.00000855, 0.00001466, 0.00000663, 0.00001343, 0.00002235, 0.00001998, 0.00008675, 0.00001826, -0.00011968, -0.00003524, 0.00000176, 0.00001587, -0.00000375, 0.00001146, -0.00001889, 0.00000240, 0.00001046, -0.00000385, -0.00000751, 0.00000364, -0.00000245, 0.00000624, -0.00000699, -0.00000256, 0.00001142, 0.00000229, -0.00001035, 0.00000036, 0.00000533, -0.00000017, 0.00000001, 0.00000332, 0.00000383, 0.00001273, 0.00001071, -0.00001451, 0.00000573, 0.00000701, -0.00002171, -0.00000697, 0.00001602, -0.00001360, -0.00000001, -0.00000260, -0.00000536, 0.00000072, -0.00000083, -0.00001081, 0.00001901, -0.00000987, 0.00000222, 0.00000285, -0.00000309, -0.00000704, 0.00001483, -0.00002637, 0.00002341, -0.00002474, 0.00001372, -0.00000439, -0.00001702, 0.00002564, -0.00000959, 0.00001148, -0.00000900, -0.00000586, -0.00000079, 0.00000821, -0.00000046, 0.00000280, -0.00000651, -0.00000826, 0.00001111, 0.00000747, -0.00001052, 0.00000065, 0.00000034, -0.00000471, 0.00000417, -0.00000674, -0.00000435, 0.00001149, -0.00000341, -0.00000793, 0.00000328, -0.00000210, 0.00000080, 0.00000544, -0.00000451, -0.00000742, -0.00000500, 0.00000249, 0.00000615, 0.00000024, -0.00000611, -0.00000252, -0.00000097, -0.00000217, -0.00000627, 0.00000001, 0.00000005, -0.00000299, -0.00000366, -0.00000032, -0.00000341, 0.00000068, -0.00000258, -0.00000411, -0.00000400, -0.00000253, -0.00000805, -0.00000605, -0.00000775, -0.00000792, -0.00001350, -0.00002266, -0.00004026, -0.00002194, 0.00002476, 0.00004412, 0.00002982, 0.00002169, 0.00001292, },
		{ 0.00001451, 0.00001520, 0.00001089, 0.00001195, 0.00001779, 0.00000648, 0.00001031, 0.00001519, 0.00000697, 0.00001398, 0.00002377, 0.00002027, 0.00008920, 0.00002377, -0.00012128, -0.00003883, -0.00000243, 0.00001856, -0.00000493, 0.00001021, -0.00001859, -0.00000018, 0.00001128, -0.00000159, -0.00000916, 0.00000272, -0.00000256, 0.00000521, -0.00000580, -0.00000372, 0.00000981, 0.00000546, -0.00000845, -0.00000477, 0.00000533, 0.00000208, -0.00000392, 0.00000398, 0.00000569, 0.00000525, 0.00001666, -0.00000956, -0.00000487, 0.00001344, -0.00001294, -0.00002036, 0.00001973, -0.00000689, -0.00000727, 0.00000116, -0.00000429, -0.00000455, 0.00000649, -0.00001716, 0.00001750, -0.00000180, -0.00000452, 0.00000310, 0.00000484, -0.00001861, 0.00002490, -0.00003094, 0.00002234, -0.00001672, -0.00000029, 0.00001123, -0.00002653, 0.00002212, -0.00000371, 0.00000837, -0.00000357, -0.00000615, -0.00000597, 0.00000859, 0.00000187, 0.00000194, -0.00000217, -0.00000999, 0.00000348, 0.00001347, -0.00000706, -0.00000481, 0.00000337, -0.00000555, 0.00000190, -0.00000068, -0.00000936, 0.00000767, 0.00000423, -0.00000963, 0.00000056, 0.00000062, -0.00000254, 0.00000614, -0.00000060, -0.00000853, -0.00000625, 0.00000022, 0.00000521, 0.00000363, -0.00000544, -0.00000332, -0.00000110, -0.00000217, -0.00000576, -0.00000150, 0.00000007, -0.00000123, -0.00000539, -0.00000004, -0.00000235, -0.00000197, -0.00000157, -0.00000228, -0.00000533, -0.00000146, -0.00000748, -0.00000705, -0.00000649, -0.00000884, -0.00001348, -0.00001589, -0.00003767, -0.00003602, 0.00000843, 0.00004519, 0.00003895, 0.00002373, },
		{ 0.00002854, 0.00001515, 0.00001836, 0.00001204, 0.00001229, 0.00001971, 0.00000634, 0.00001172, 0.00001683, 0.00000711, 0.00001431, 0.00002521, 0.00002062, 0.00009213, 0.00002852, -0.00012275, -0.00004067, -0.00000858, 0.00002052, -0.00000409, 0.00000802, -0.00001821, -0.00000203, 0.00001056, 0.00000107, -0.00000971, 0.00000110, -0.00000239, 0.00000399, -0.00000509, -0.00000382, 0.00000774, 0.00000681, -0.00000456, -0.00000825, 0.00000242, 0.00000511, -0.00000581, 0.00000124, 0.00001001, -0.00000066, 0.00001622, 0.00000026, -0.00001254, 0.00001023, 0.00000163, -0.00002898, 0.00001419, 0.00000416, -0.00001191, -0.00000083, 0.00000206, -0.00001194, 0.00001157, -0.00001726, 0.00000992, 0.00000658, -0.00000528, -0.00000487, 0.00001818, -0.00002996, 0.00002985, -0.00002798, 0.00001348, -0.00000275, -0.00001550, 0.00002177, -0.00002700, 0.00001285, 0.00000193, 0.00000696, -0.00000082, -0.00000342, -0.00001008, 0.00000579, 0.00000506, 0.00000147, 0.00000016, -0.00000757, -0.00000449, 0.00001410, 0.00000040, -0.00000923, 0.00000307, -0.00000312, -0.00000231, 0.00000389, -0.00000915, 0.00000024, 0.00000952, -0.00000675, -0.00000419, 0.00000289, -0.00000367, 0.00000399, 0.00000357, -0.00000787, -0.00000797, -0.00000130, 0.00000300, 0.00000593, -0.00000331, -0.00000412, -0.00000134, -0.00000232, -0.00000525, -0.00000219, -0.00000109, 0.00000063, -0.00000587, -0.00000123, -0.00000065, -0.00000362, -0.00000249, 0.00000016, -0.00000591, -0.00000154, -0.00000598, -0.00000809, -0.00000599, -0.00000835, -0.00001479, -0.00001160, -0.00003065, -0.00004416, -0.00001180, 0.00003887, 0.00004887, },
		{ 0.00005627, 0.00003731, 0.00001555, 0.00002109, 0.00001446, 0.00001199, 0.00002212, 0.00000643, 0.00001205, 0.00001948, 0.00000764, 0.00001409, 0.00002685, 0.00002076, 0.00009560, 0.00003320, -0.00012532, -0.00004044, -0.00001528, 0.00002022, -0.00000129, 0.00000596, -0.00001875, -0.00000269, 0.00000882, 0.00000307, -0.00000876, -0.00000090, -0.00000231, 0.00000299, -0.00000514, -0.00000323, 0.00000615, 0.00000615, -0.00000020, -0.00000893, -0.00000223, 0.00000663, -0.00000461, -0.00000362, 0.00001362, -0.00000227, 0.00001052, 0.00000957, -0.00001245, -0.00000092, 0.00001455, -0.00002758, 0.00000146, 0.00001310, -0.00000916, -0.00000890, 0.00001023, -0.00001649, 0.00001075, -0.00001138, 0.00000109, 0.00000901, 0.00000308, -0.00001920, 0.00003144, -0.00003523, 0.00002608, -0.00001704, -0.00000065, 0.00001220, -0.00002623, 0.00002388, -0.00001990, 0.00000248, 0.00000394, 0.00000806, 0.00000002, 0.00000002, -0.00001125, 0.00000070, 0.00000701, 0.00000276, 0.00000033, -0.00000326, -0.00000939, 0.00000922, 0.00000834, -0.00000953, -0.00000081, 0.00000063, -0.00000545, 0.00000470, -0.00000470, -0.00000691, 0.00000974, -0.00000029, -0.00000820, 0.00000265, -0.00000221, 0.00000054, 0.00000603, -0.00000504, -0.00000970, -0.00000252, 0.00000058, 0.00000643, -0.00000030, -0.00000421, -0.00000190, -0.00000246, -0.00000505, -0.00000208, -0.00000267, 0.00000160, -0.00000491, -0.00000303, 0.00000045, -0.00000368, -0.00000461, 0.00000192, -0.00000513, -0.00000249, -0.00000442, -0.00000832, -0.00000652, -0.00000715, -0.00001575, -0.00001042, -0.00002238, -0.00004468, -0.00003158, 0.00002366, },
		{ 0.00000082, 0.00005676, 0.00004976, 0.00001759, 0.00002235, 0.00001839, 0.00001150, 0.00002427, 0.00000761, 0.00001110, 0.00002236, 0.00000936, 0.00001320, 0.00002859, 0.00002080, 0.00009919, 0.00003872, -0.00012959, -0.00003899, -0.00002072, 0.00001703, 0.00000224, 0.00000545, -0.00002072, -0.00000254, 0.00000715, 0.00000349, -0.00000652, -0.00000259, -0.00000276, 0.00000247, -0.00000587, -0.00000258, 0.00000569, 0.00000399, 0.00000318, -0.00000687, -0.00000658, 0.00000526, -0.00000119, -0.00000806, 0.00001417, 0.00000025, 0.00000350, 0.00001346, -0.00000428, -0.00001379, 0.00001894, -0.00001632, -0.00001158, 0.00001353, 0.00000172, -0.00001882, 0.00001469, -0.00001460, 0.00000436, -0.00000412, -0.00000244, 0.00000166, 0.00001885, -0.00003381, 0.00003793, -0.00003047, 0.00001359, -0.00000129, -0.00001505, 0.00002284, -0.00002909, 0.00001821, -0.00000971, -0.00000441, 0.00000127, 0.00001027, 0.00000076, 0.00000235, -0.00000948, -0.00000444, 0.00000610, 0.00000564, -0.00000025, 0.00000033, -0.00000981, 0.00000134, 0.00001291, -0.00000488, -0.00000611, 0.00000284, -0.00000557, 0.00000202, 0.00000091, -0.00001031, 0.00000488, 0.00000662, -0.00000884, -0.00000050, 0.00000042, -0.00000201, 0.00000571, -0.00000089, -0.00001038, -0.00000400, -0.00000136, 0.00000528, 0.00000248, -0.00000309, -0.00000262, -0.00000261, -0.00000518, -0.00000157, -0.00000389, 0.00000131, -0.00000314, -0.00000434, 0.00000030, -0.00000254, -0.00000668, 0.00000217, -0.00000329, -0.00000342, -0.00000360, -0.00000753, -0.00000749, -0.00000630, -0.00001549, -0.00001135, -0.00001595, -0.00003860, -0.00004596, },
		{ -0.00005146, -0.00002528, 0.00004662, 0.00006317, 0.00002375, 0.00002179, 0.00002313, 0.00001192, 0.00002532, 0.00001017, 0.00000957, 0.00002428, 0.00001274, 0.00001204, 0.00003000, 0.00002113, 0.00010233, 0.00004557, -0.00013515, -0.00003788, -0.00002364, 0.00001173, 0.00000457, 0.00000739, -0.00002359, -0.00000269, 0.00000663, 0.00000218, -0.00000389, -0.00000315, -0.00000390, 0.00000233, -0.00000687, -0.00000255, 0.00000654, 0.00000127, 0.00000469, -0.00000312, -0.00000889, 0.00000117, 0.00000231, -0.00001000, 0.00001152, 0.00000421, -0.00000061, 0.00001037, 0.00000771, -0.00002122, 0.00001263, -0.00000127, -0.00001686, 0.00000327, 0.00001600, -0.00002401, 0.00001164, -0.00000702, -0.00000279, -0.00000199, 0.00000361, -0.00001415, 0.00003554, -0.00004115, 0.00003316, -0.00001597, -0.00000366, 0.00001384, -0.00002468, 0.00002605, -0.00002436, 0.00000862, -0.00000148, -0.00000555, -0.00000438, 0.00001113, 0.00000268, 0.00000327, -0.00000623, -0.00000746, 0.00000225, 0.00000845, 0.00000019, 0.00000183, -0.00000675, -0.00000584, 0.00001197, 0.00000282, -0.00000947, 0.00000164, -0.00000294, -0.00000180, 0.00000442, -0.00000883, -0.00000244, 0.00001050, -0.00000524, -0.00000493, 0.00000200, -0.00000223, 0.00000306, 0.00000285, -0.00000912, -0.00000577, -0.00000282, 0.00000327, 0.00000408, -0.00000087, -0.00000305, -0.00000290, -0.00000548, -0.00000111, -0.00000435, 0.00000011, -0.00000154, -0.00000448, -0.00000087, -0.00000118, -0.00000773, 0.00000095, -0.00000131, -0.00000352, -0.00000370, -0.00000624, -0.00000807, -0.00000636, -0.00001427, -0.00001259, -0.00001292, -0.00002920, },
		{ -0.00002047, -0.00004763, -0.00004831, 0.00002482, 0.00007240, 0.00003573, 0.00002068, 0.00002707, 0.00001440, 0.00002506, 0.00001347, 0.00000877, 0.00002431, 0.00001739, 0.00001161, 0.00003051, 0.00002200, 0.00010486, 0.00005349, -0.00014092, -0.00003843, -0.00002416, 0.00000625, 0.00000406, 0.00001139, -0.00002588, -0.00000433, 0.00000765, -0.00000014, -0.00000210, -0.00000214, -0.00000544, 0.00000218, -0.00000755, -0.00000360, 0.00000839, -0.00000101, 0.00000417, 0.00000086, -0.00000833, -0.00000411, 0.00000385, -0.00000910, 0.00000765, 0.00000635, 0.00000010, 0.00000279, 0.00001729, -0.00001933, -0.00000041, 0.00000884, -0.00001006, -0.00001345, 0.00002568, -0.00001968, 0.00000213, 0.00000084, -0.00000409, -0.00000927, 0.00001826, -0.00003186, 0.00004481, -0.00003607, 0.00001745, 0.00000360, -0.00001951, 0.00002330, -0.00002687, 0.00002218, -0.00001544, -0.00000005, 0.00000166, -0.00000189, -0.00000972, 0.00000888, 0.00000555, 0.00000376, -0.00000332, -0.00000750, -0.00000294, 0.00000920, 0.00000243, 0.00000167, -0.00000263, -0.00000948, 0.00000621, 0.00000990, -0.00000848, -0.00000252, 0.00000022, -0.00000408, 0.00000429, -0.00000400, -0.00000852, 0.00000942, 0.00000104, -0.00000798, 0.00000096, -0.00000028, -0.00000015, 0.00000453, -0.00000598, -0.00000715, -0.00000425, 0.00000124, 0.00000410, 0.00000175, -0.00000265, -0.00000335, -0.00000584, -0.00000095, -0.00000413, -0.00000132, -0.00000084, -0.00000352, -0.00000226, -0.00000051, -0.00000753, -0.00000095, -0.00000018, -0.00000258, -0.00000424, -0.00000522, -0.00000774, -0.00000713, -0.00001305, -0.00001284, -0.00001285, },
		{ -0.00001392, -0.00001526, -0.00003733, -0.00006237, -0.00000555, 0.00007144, 0.00005272, 0.00002201, 0.00002853, 0.00001920, 0.00002446, 0.00001611, 0.00000988, 0.00002256, 0.00002197, 0.00001300, 0.00002998, 0.00002319, 0.00010718, 0.00006172, -0.00014586, -0.00004093, -0.00002366, 0.00000265, 0.00000038, 0.00001571, -0.00002591, -0.00000789, 0.00000954, -0.00000214, -0.00000212, 0.00000018, -0.00000661, 0.00000141, -0.00000746, -0.00000574, 0.00001045, -0.00000203, 0.00000203, 0.00000376, -0.00000522, -0.00000869, 0.00000265, -0.00000686, 0.00000516, 0.00000500, 0.00000432, -0.00000460, 0.00002013, -0.00000976, -0.00001224, 0.00000761, 0.00000634, -0.00002774, 0.00002417, -0.00000668, -0.00000744, 0.00000173, 0.00000468, -0.00002488, 0.00003506, -0.00004264, 0.00004070, -0.00001872, -0.00000390, 0.00002126, -0.00002862, 0.00002480, -0.00002237, 0.00001441, -0.00000678, -0.00000440, -0.00000033, 0.00000342, -0.00001182, 0.00000368, 0.00000775, 0.00000508, -0.00000165, -0.00000528, -0.00000718, 0.00000687, 0.00000564, 0.00000145, 0.00000031, -0.00000891, -0.00000148, 0.00001298, -0.00000306, -0.00000716, 0.00000137, -0.00000348, 0.00000135, 0.00000109, -0.00001061, 0.00000410, 0.00000684, -0.00000752, -0.00000252, 0.00000217, -0.00000200, 0.00000358, -0.00000209, -0.00000715, -0.00000596, -0.00000044, 0.00000291, 0.00000384, -0.00000121, -0.00000371, -0.00000628, -0.00000110, -0.00000359, -0.00000234, -0.00000119, -0.00000212, -0.00000301, -0.00000081, -0.00000660, -0.00000254, -0.00000031, -0.00000107, -0.00000448, -0.00000496, -0.00000667, -0.00000783, -0.00001265, -0.00001190, },
		{ -0.00001064, -0.00001427, -0.00001409, -0.00002540, -0.00006447, -0.00003766, 0.00005598, 0.00007031, 0.00002937, 0.00002719, 0.00002500, 0.00002533, 0.00001686, 0.00001301, 0.00002036, 0.00002482, 0.00001667, 0.00002911, 0.00002391, 0.00010994, 0.00006961, -0.00014976, -0.00004445, -0.00002407, 0.00000196, -0.00000509, 0.00001798, -0.00002282, -0.00001249, 0.00001085, -0.00000261, -0.00000402, 0.00000273, -0.00000644, -0.00000031, -0.00000657, -0.00000841, 0.00001177, -0.00000142, -0.00000079, 0.00000471, -0.00000053, -0.00001123, -0.00000053, -0.00000547, 0.00000547, 0.00000104, 0.00000871, -0.00000798, 0.00001623, 0.00000144, -0.00001562, -0.00000498, 0.00002388, -0.00003106, 0.00001100, 0.00000772, -0.00000855, -0.00000876, 0.00002197, -0.00004209, 0.00004503, -0.00003990, 0.00002292, 0.00000516, -0.00002305, 0.00003110, -0.00002887, 0.00001964, -0.00001453, 0.00000685, -0.00000168, -0.00000381, -0.00000493, 0.00000691, -0.00000977, -0.00000241, 0.00000748, 0.00000757, -0.00000083, -0.00000246, -0.00000884, 0.00000214, 0.00000794, 0.00000262, 0.00000131, -0.00000560, -0.00000763, 0.00001075, 0.00000435, -0.00000921, -0.00000069, -0.00000081, -0.00000201, 0.00000371, -0.00000832, -0.00000262, 0.00000922, -0.00000340, -0.00000658, 0.00000313, -0.00000155, 0.00000090, 0.00000085, -0.00000524, -0.00000769, -0.00000195, 0.00000121, 0.00000474, 0.00000095, -0.00000356, -0.00000681, -0.00000149, -0.00000310, -0.00000271, -0.00000219, -0.00000106, -0.00000277, -0.00000172, -0.00000573, -0.00000324, -0.00000139, 0.00000019, -0.00000395, -0.00000532, -0.00000552, -0.00000778, -0.00001315, },
		{ -0.00001387, -0.00001012, -0.00001314, -0.00001534, -0.00001637, -0.00005592, -0.00006315, 0.00002612, 0.00008117, 0.00004480, 0.00002521, 0.00002925, 0.00002931, 0.00001571, 0.00001684, 0.00001968, 0.00002491, 0.00002179, 0.00002935, 0.00002336, 0.00011347, 0.00007727, -0.00015337, -0.00004750, -0.00002662, 0.00000350, -0.00000985, 0.00001652, -0.00001740, -0.00001622, 0.00001021, -0.00000127, -0.00000680, 0.00000403, -0.00000434, -0.00000275, -0.00000543, -0.00001074, 0.00001165, 0.00000049, -0.00000311, 0.00000335, 0.00000454, -0.00001136, -0.00000408, -0.00000642, 0.00000803, -0.00000304, 0.00001041, -0.00000678, 0.00000963, 0.00000768, -0.00000836, -0.00002230, 0.00003274, -0.00002064, -0.00000669, 0.00001324, 0.00000419, -0.00002858, 0.00004030, -0.00005168, 0.00004143, -0.00002294, -0.00000267, 0.00002686, -0.00003335, 0.00003097, -0.00002214, 0.00001167, -0.00000735, 0.00000244, -0.00000090, -0.00000018, -0.00000886, 0.00000673, -0.00000499, -0.00000670, 0.00000407, 0.00001022, 0.00000030, -0.00000046, -0.00000774, -0.00000303, 0.00000764, 0.00000536, 0.00000123, -0.00000194, -0.00001005, 0.00000456, 0.00001019, -0.00000695, -0.00000495, 0.00000170, -0.00000355, 0.00000299, -0.00000357, -0.00000740, 0.00000703, 0.00000249, -0.00000875, 0.00000155, 0.00000056, -0.00000174, 0.00000158, -0.00000184, -0.00000861, -0.00000363, -0.00000045, 0.00000438, 0.00000311, -0.00000257, -0.00000728, -0.00000206, -0.00000287, -0.00000252, -0.00000327, -0.00000084, -0.00000178, -0.00000253, -0.00000546, -0.00000312, -0.00000272, 0.00000060, -0.00000276, -0.00000574, -0.00000489, -0.00000686, },
		{ -0.00000558, -0.00001398, -0.00001079, -0.00001113, -0.00001654, -0.00001248, -0.00004171, -0.00007551, -0.00001257, 0.00007758, 0.00006642, 0.00002704, 0.00002968, 0.00003641, 0.00001436, 0.00001922, 0.00002192, 0.00002283, 0.00002639, 0.00003213, 0.00002145, 0.00011719, 0.00008558, -0.00015787, -0.00004909, -0.00003102, 0.00000516, -0.00001161, 0.00001143, -0.00001200, -0.00001711, 0.00000732, 0.00000084, -0.00000877, 0.00000300, -0.00000062, -0.00000496, -0.00000489, -0.00001197, 0.00001003, 0.00000267, -0.00000376, -0.00000010, 0.00000883, -0.00000940, -0.00000653, -0.00000964, 0.00001086, -0.00000502, 0.00000892, -0.00000365, 0.00000544, 0.00000585, 0.00000541, -0.00003485, 0.00002780, -0.00000219, -0.00001727, 0.00000255, 0.00002867, -0.00004901, 0.00004970, -0.00004690, 0.00002355, 0.00000246, -0.00002677, 0.00003886, -0.00003242, 0.00002325, -0.00001293, 0.00000503, -0.00000336, 0.00000176, -0.00000299, 0.00000362, -0.00000989, 0.00000336, -0.00000021, -0.00000739, -0.00000138, 0.00001130, 0.00000257, 0.00000044, -0.00000509, -0.00000649, 0.00000436, 0.00000839, 0.00000165, 0.00000032, -0.00000875, -0.00000253, 0.00001171, -0.00000103, -0.00000880, 0.00000193, -0.00000254, 0.00000013, 0.00000073, -0.00000821, 0.00000152, 0.00000715, -0.00000753, -0.00000195, 0.00000262, -0.00000274, 0.00000001, 0.00000170, -0.00000793, -0.00000552, -0.00000194, 0.00000318, 0.00000456, -0.00000080, -0.00000740, -0.00000274, -0.00000297, -0.00000207, -0.00000394, -0.00000143, -0.00000063, -0.00000271, -0.00000582, -0.00000268, -0.00000363, 0.00000012, -0.00000146, -0.00000569, -0.00000492, },
		{ -0.00000521, -0.00000463, -0.00001305, -0.00001215, -0.00000961, -0.00001597, -0.00001292, -0.00002817, -0.00007287, -0.00005040, 0.00005513, 0.00008753, 0.00003752, 0.00002628, 0.00004439, 0.00001554, 0.00001861, 0.00002676, 0.00002071, 0.00002829, 0.00003776, 0.00001934, 0.00011983, 0.00009559, -0.00016392, -0.00004947, -0.00003564, 0.00000468, -0.00000971, 0.00000481, -0.00000940, -0.00001445, 0.00000351, 0.00000184, -0.00000853, -0.00000035, 0.00000338, -0.00000571, -0.00000557, -0.00001203, 0.00000776, 0.00000371, -0.00000210, -0.00000471, 0.00001134, -0.00000601, -0.00000720, -0.00001395, 0.00001213, -0.00000458, 0.00000632, -0.00000231, 0.00000644, -0.00000187, 0.00001790, -0.00003606, 0.00001222, 0.00001253, -0.00001106, -0.00002343, 0.00005494, -0.00005861, 0.00004319, -0.00002743, -0.00000262, 0.00002682, -0.00004113, 0.00003825, -0.00002297, 0.00001315, -0.00000578, 0.00000205, -0.00000278, 0.00000341, -0.00000553, 0.00000543, -0.00000785, -0.00000105, 0.00000216, -0.00000463, -0.00000655, 0.00000948, 0.00000567, 0.00000105, -0.00000251, -0.00000717, -0.00000068, 0.00000981, 0.00000358, 0.00000098, -0.00000548, -0.00000747, 0.00000842, 0.00000582, -0.00000964, -0.00000063, -0.00000013, -0.00000259, 0.00000244, -0.00000531, -0.00000442, 0.00000825, -0.00000330, -0.00000541, 0.00000294, -0.00000167, -0.00000268, 0.00000388, -0.00000546, -0.00000711, -0.00000352, 0.00000170, 0.00000493, 0.00000129, -0.00000686, -0.00000343, -0.00000336, -0.00000167, -0.00000406, -0.00000249, 0.00000009, -0.00000214, -0.00000639, -0.00000246, -0.00000392, -0.00000083, -0.00000063, -0.00000507, },
		{ -0.00000422, -0.00000524, -0.00000437, -0.00001142, -0.00001314, -0.00000960, -0.00001362, -0.00001483, -0.00002005, -0.00005883, -0.00007702, 0.00001575, 0.00009818, 0.00005861, 0.00002260, 0.00004955, 0.00002124, 0.00001549, 0.00003196, 0.00002123, 0.00002656, 0.00004482, 0.00001909, 0.00012026, 0.00010757, -0.00017099, -0.00005009, -0.00003855, 0.00000107, -0.00000561, -0.00000024, -0.00001120, -0.00000960, 0.00000127, 0.00000013, -0.00000598, -0.00000463, 0.00000584, -0.00000419, -0.00000721, -0.00001175, 0.00000619, 0.00000254, 0.00000151, -0.00000896, 0.00001130, -0.00000196, -0.00000600, -0.00001802, 0.00001143, -0.00000341, 0.00000542, -0.00000467, 0.00001127, -0.00000981, 0.00002280, -0.00002663, -0.00000362, 0.00001202, 0.00001330, -0.00005441, 0.00006941, -0.00004939, 0.00002103, 0.00000001, -0.00002731, 0.00004157, -0.00004228, 0.00002775, -0.00001092, 0.00000578, -0.00000298, 0.00000247, -0.00000396, 0.00000532, -0.00000671, 0.00000477, -0.00000423, -0.00000413, 0.00000129, -0.00000029, -0.00000916, 0.00000488, 0.00000820, 0.00000248, -0.00000099, -0.00000548, -0.00000534, 0.00000835, 0.00000664, 0.00000115, -0.00000238, -0.00000878, 0.00000213, 0.00001035, -0.00000647, -0.00000458, 0.00000157, -0.00000344, 0.00000125, -0.00000080, -0.00000795, 0.00000535, 0.00000183, -0.00000672, 0.00000095, 0.00000056, -0.00000483, 0.00000384, -0.00000196, -0.00000757, -0.00000537, 0.00000023, 0.00000438, 0.00000307, -0.00000555, -0.00000385, -0.00000401, -0.00000152, -0.00000370, -0.00000354, 0.00000011, -0.00000116, -0.00000666, -0.00000266, -0.00000382, -0.00000171, -0.00000049, },
		{ -0.00000079, -0.00000369, -0.00000476, -0.00000457, -0.00000989, -0.00001294, -0.00001101, -0.00001099, -0.00001523, -0.00001851, -0.00004078, -0.00008582, -0.00003178, 0.00008905, 0.00008673, 0.00002494, 0.00004909, 0.00003095, 0.00001259, 0.00003458, 0.00002583, 0.00002248, 0.00005049, 0.00002247, 0.00011855, 0.00012052, -0.00017760, -0.00005258, -0.00003887, -0.00000466, -0.00000214, -0.00000138, -0.00001667, -0.00000564, 0.00000285, -0.00000440, -0.00000260, -0.00000772, 0.00000547, -0.00000084, -0.00000856, -0.00001246, 0.00000649, -0.00000086, 0.00000559, -0.00001113, 0.00000859, 0.00000168, -0.00000297, -0.00002132, 0.00000984, -0.00000375, 0.00000748, -0.00000943, 0.00001580, -0.00001300, 0.00001948, -0.00001444, -0.00000794, -0.00000777, 0.00004657, -0.00007486, 0.00006193, -0.00002178, -0.00000874, 0.00002493, -0.00004191, 0.00004310, -0.00003271, 0.00001376, -0.00000216, 0.00000366, -0.00000389, 0.00000423, -0.00000489, 0.00000609, -0.00000612, 0.00000261, -0.00000105, -0.00000462, -0.00000179, 0.00000313, -0.00000824, -0.00000084, 0.00000849, 0.00000510, -0.00000031, -0.00000286, -0.00000781, 0.00000424, 0.00000929, 0.00000221, -0.00000068, -0.00000698, -0.00000416, 0.00001055, -0.00000055, -0.00000740, 0.00000100, -0.00000222, -0.00000141, 0.00000275, -0.00000777, 0.00000008, 0.00000524, -0.00000495, -0.00000239, 0.00000227, -0.00000521, 0.00000191, 0.00000120, -0.00000632, -0.00000726, -0.00000132, 0.00000332, 0.00000405, -0.00000374, -0.00000372, -0.00000476, -0.00000173, -0.00000309, -0.00000424, -0.00000048, -0.00000023, -0.00000640, -0.00000309, -0.00000372, -0.00000222, },
		{ -0.00000248, -0.00000109, -0.00000372, -0.00000400, -0.00000463, -0.00000909, -0.00001154, -0.00001271, -0.00000986, -0.00001285, -0.00002111, -0.00002642, -0.00007673, -0.00007448, 0.00005623, 0.00011215, 0.00003932, 0.00004377, 0.00004107, 0.00001367, 0.00003294, 0.00003345, 0.00001922, 0.00005213, 0.00002955, 0.00011647, 0.00013257, -0.00018234, -0.00005749, -0.00003754, -0.00001009, -0.00000178, 0.00000145, -0.00002279, -0.00000578, 0.00000873, -0.00000982, -0.00000082, -0.00000800, 0.00000249, 0.00000260, -0.00000790, -0.00001512, 0.00000870, -0.00000518, 0.00000806, -0.00001018, 0.00000425, 0.00000340, 0.00000204, -0.00002404, 0.00000885, -0.00000649, 0.00001108, -0.00001317, 0.00001648, -0.00001093, 0.00001368, -0.00000994, 0.00000532, -0.00003993, 0.00007241, -0.00007191, 0.00003186, 0.00001469, -0.00003412, 0.00003850, -0.00004306, 0.00003389, -0.00001890, 0.00000282, 0.00000033, 0.00000586, -0.00000592, 0.00000511, -0.00000450, 0.00000561, -0.00000451, 0.00000047, 0.00000042, -0.00000284, -0.00000505, 0.00000394, -0.00000464, -0.00000533, 0.00000577, 0.00000814, 0.00000054, -0.00000077, -0.00000753, -0.00000091, 0.00000980, 0.00000469, -0.00000015, -0.00000392, -0.00000794, 0.00000662, 0.00000534, -0.00000708, -0.00000178, -0.00000024, -0.00000355, 0.00000377, -0.00000459, -0.00000486, 0.00000521, -0.00000091, -0.00000518, 0.00000214, -0.00000381, -0.00000063, 0.00000272, -0.00000352, -0.00000852, -0.00000316, 0.00000211, 0.00000416, -0.00000197, -0.00000284, -0.00000537, -0.00000235, -0.00000250, -0.00000448, -0.00000138, 0.00000030, -0.00000570, -0.00000340, -0.00000384, },
		{ -0.00000408, -0.00000280, -0.00000110, -0.00000414, -0.00000352, -0.00000414, -0.00000903, -0.00000976, -0.00001331, -0.00001098, -0.00000872, -0.00002372, -0.00002023, -0.00005631, -0.00010024, 0.00000441, 0.00012192, 0.00006719, 0.00003907, 0.00004660, 0.00002090, 0.00002834, 0.00004070, 0.00002017, 0.00004906, 0.00003806, 0.00011678, 0.00014232, -0.00018517, -0.00006366, -0.00003674, -0.00001297, -0.00000499, 0.00000577, -0.00002580, -0.00001145, 0.00001681, -0.00001300, -0.00000248, -0.00000561, -0.00000116, 0.00000375, -0.00000416, -0.00001942, 0.00001151, -0.00000830, 0.00000739, -0.00000667, 0.00000053, 0.00000161, 0.00000894, -0.00002627, 0.00000888, -0.00001038, 0.00001326, -0.00001331, 0.00001357, -0.00000834, 0.00001382, -0.00001944, 0.00003192, -0.00006897, 0.00007549, -0.00004256, -0.00001057, 0.00004563, -0.00004532, 0.00003794, -0.00003379, 0.00002050, -0.00000762, -0.00000154, -0.00000256, 0.00000921, -0.00000645, 0.00000420, -0.00000311, 0.00000474, -0.00000298, -0.00000066, 0.00000011, -0.00000023, -0.00000671, 0.00000206, -0.00000048, -0.00000680, 0.00000076, 0.00000998, 0.00000247, 0.00000025, -0.00000536, -0.00000499, 0.00000739, 0.00000777, 0.00000043, -0.00000139, -0.00000838, 0.00000068, 0.00000844, -0.00000332, -0.00000513, 0.00000067, -0.00000396, 0.00000243, -0.00000053, -0.00000721, 0.00000190, 0.00000330, -0.00000578, 0.00000005, -0.00000174, -0.00000226, 0.00000215, -0.00000015, -0.00000837, -0.00000530, 0.00000080, 0.00000368, -0.00000074, -0.00000138, -0.00000548, -0.00000330, -0.00000215, -0.00000430, -0.00000228, 0.00000031, -0.00000486, -0.00000337, },
		{ -0.00000301, -0.00000415, -0.00000333, -0.00000095, -0.00000448, -0.00000370, -0.00000320, -0.00000912, -0.00000862, -0.00001215, -0.00001355, -0.00000524, -0.00002310, -0.00002177, -0.00003454, -0.00010300, -0.00005359, 0.00010537, 0.00010238, 0.00004353, 0.00004443, 0.00003276, 0.00002492, 0.00004375, 0.00002658, 0.00004355, 0.00004428, 0.00012152, 0.00015023, -0.00018783, -0.00006888, -0.00003841, -0.00001290, -0.00000973, 0.00000792, -0.00002345, -0.00002103, 0.00002315, -0.00001139, -0.00000736, -0.00000270, -0.00000273, 0.00000106, 0.00000209, -0.00002358, 0.00001286, -0.00000860, 0.00000377, -0.00000293, 0.00000009, -0.00000435, 0.00001658, -0.00002717, 0.00000894, -0.00001305, 0.00001203, -0.00001078, 0.00001167, -0.00001208, 0.00002509, -0.00004035, 0.00005860, -0.00007842, 0.00004978, 0.00000367, -0.00004865, 0.00005903, -0.00003982, 0.00002749, -0.00002110, 0.00000963, -0.00000258, 0.00000009, -0.00000733, 0.00001056, -0.00000454, 0.00000224, -0.00000188, 0.00000435, -0.00000210, -0.00000069, -0.00000110, 0.00000171, -0.00000614, -0.00000119, 0.00000211, -0.00000500, -0.00000453, 0.00000921, 0.00000543, 0.00000075, -0.00000280, -0.00000660, 0.00000280, 0.00000977, 0.00000212, -0.00000009, -0.00000640, -0.00000454, 0.00000748, 0.00000222, -0.00000684, -0.00000046, -0.00000283, 0.00000015, 0.00000228, -0.00000625, -0.00000290, 0.00000550, -0.00000374, -0.00000288, -0.00000050, -0.00000217, 0.00000013, 0.00000243, -0.00000650, -0.00000725, -0.00000084, 0.00000297, -0.00000026, 0.00000023, -0.00000483, -0.00000431, -0.00000223, -0.00000384, -0.00000293, -0.00000009, -0.00000417, },
		{ -0.00000377, -0.00000255, -0.00000387, -0.00000396, -0.00000099, -0.00000438, -0.00000443, -0.00000240, -0.00000862, -0.00000860, -0.00000979, -0.00001573, -0.00000447, -0.00001881, -0.00002669, -0.00002017, -0.00008542, -0.00010122, 0.00005991, 0.00013135, 0.00006422, 0.00003668, 0.00004389, 0.00002751, 0.00004108, 0.00003632, 0.00004000, 0.00004515, 0.00013026, 0.00015898, -0.00019297, -0.00007151, -0.00004262, -0.00001180, -0.00001266, 0.00000534, -0.00001657, -0.00003044, 0.00002412, -0.00000492, -0.00001271, -0.00000244, -0.00000032, -0.00000498, 0.00000843, -0.00002515, 0.00001133, -0.00000622, -0.00000058, -0.00000213, 0.00000455, -0.00001317, 0.00002247, -0.00002522, 0.00000771, -0.00001331, 0.00000872, -0.00001027, 0.00001686, -0.00002602, 0.00004503, -0.00006156, 0.00006961, -0.00005963, 0.00000257, 0.00004870, -0.00006732, 0.00005141, -0.00002346, 0.00001547, -0.00001165, 0.00000469, -0.00000333, 0.00000427, -0.00001034, 0.00000864, -0.00000122, 0.00000080, -0.00000172, 0.00000464, -0.00000172, -0.00000018, -0.00000216, 0.00000221, -0.00000400, -0.00000395, 0.00000206, -0.00000128, -0.00000792, 0.00000559, 0.00000830, 0.00000179, -0.00000105, -0.00000567, -0.00000212, 0.00000928, 0.00000497, 0.00000054, -0.00000377, -0.00000714, 0.00000324, 0.00000693, -0.00000553, -0.00000313, -0.00000149, -0.00000143, 0.00000281, -0.00000304, -0.00000680, 0.00000461, -0.00000015, -0.00000494, -0.00000101, -0.00000067, -0.00000200, 0.00000323, -0.00000338, -0.00000823, -0.00000295, 0.00000214, -0.00000036, 0.00000146, -0.00000345, -0.00000499, -0.00000283, -0.00000334, -0.00000321, -0.00000072, },
		{ -0.00000140, -0.00000365, -0.00000219, -0.00000330, -0.00000432, -0.00000148, -0.00000394, -0.00000512, -0.00000240, -0.00000731, -0.00000927, -0.00000757, -0.00001590, -0.00000664, -0.00001314, -0.00002961, -0.00001681, -0.00005760, -0.00012493, -0.00000592, 0.00013785, 0.00010133, 0.00003147, 0.00004784, 0.00003818, 0.00003525, 0.00004453, 0.00004240, 0.00004055, 0.00013958, 0.00017226, -0.00020225, -0.00007206, -0.00004718, -0.00001283, -0.00001134, -0.00000160, -0.00000885, -0.00003532, 0.00001875, 0.00000329, -0.00001456, -0.00000696, 0.00000558, -0.00001163, 0.00001172, -0.00002268, 0.00000750, -0.00000342, -0.00000236, -0.00000644, 0.00001297, -0.00002139, 0.00002359, -0.00001958, 0.00000541, -0.00001286, 0.00000790, -0.00001749, 0.00003206, -0.00004702, 0.00006360, -0.00006903, 0.00005533, -0.00001695, -0.00004832, 0.00007445, -0.00006074, 0.00002950, -0.00000686, 0.00000916, -0.00000815, 0.00000486, -0.00000657, 0.00000728, -0.00000987, 0.00000448, 0.00000146, 0.00000105, -0.00000259, 0.00000507, -0.00000130, 0.00000027, -0.00000252, 0.00000148, -0.00000161, -0.00000494, -0.00000017, 0.00000224, -0.00000815, 0.00000032, 0.00000948, 0.00000399, -0.00000025, -0.00000333, -0.00000544, 0.00000604, 0.00000782, 0.00000167, -0.00000185, -0.00000689, -0.00000206, 0.00000864, -0.00000151, -0.00000566, -0.00000133, -0.00000158, 0.00000151, 0.00000050, -0.00000813, 0.00000118, 0.00000308, -0.00000494, -0.00000309, 0.00000109, -0.00000300, 0.00000217, -0.00000016, -0.00000760, -0.00000528, 0.00000100, -0.00000070, 0.00000200, -0.00000170, -0.00000493, -0.00000381, -0.00000310, -0.00000308, },
		{ -0.00000262, -0.00000197, -0.00000376, -0.00000199, -0.00000272, -0.00000414, -0.00000228, -0.00000367, -0.00000518, -0.00000333, -0.00000574, -0.00000962, -0.00000668, -0.00001380, -0.00001000, -0.00000947, -0.00002743, -0.00002185, -0.00003225, -0.00012014, -0.00007437, 0.00011028, 0.00014504, 0.00003991, 0.00004159, 0.00005384, 0.00003217, 0.00004660, 0.00005149, 0.00003420, 0.00014459, 0.00019235, -0.00021458, -0.00007338, -0.00004891, -0.00001812, -0.00000613, -0.00000958, -0.00000476, -0.00003375, 0.00000981, 0.00000826, -0.00001003, -0.00001555, 0.00001188, -0.00001528, 0.00001008, -0.00001715, 0.00000412, -0.00000342, 0.00000071, -0.00001525, 0.00002163, -0.00002480, 0.00001851, -0.00001193, 0.00000503, -0.00001599, 0.00001471, -0.00003474, 0.00005346, -0.00006518, 0.00006813, -0.00005375, 0.00001804, 0.00003371, -0.00008293, 0.00007175, -0.00003514, 0.00000640, 0.00000071, 0.00001072, -0.00000862, 0.00000668, -0.00000886, 0.00000730, -0.00000672, 0.00000045, 0.00000192, 0.00000286, -0.00000353, 0.00000487, -0.00000053, 0.00000048, -0.00000224, 0.00000029, -0.00000008, -0.00000403, -0.00000309, 0.00000381, -0.00000556, -0.00000449, 0.00000790, 0.00000693, 0.00000035, -0.00000105, -0.00000619, 0.00000120, 0.00000904, 0.00000397, -0.00000085, -0.00000494, -0.00000612, 0.00000680, 0.00000335, -0.00000627, -0.00000280, -0.00000074, -0.00000031, 0.00000268, -0.00000670, -0.00000297, 0.00000423, -0.00000285, -0.00000554, 0.00000182, -0.00000241, 0.00000011, 0.00000193, -0.00000533, -0.00000715, -0.00000072, -0.00000107, 0.00000191, -0.00000017, -0.00000398, -0.00000477, -0.00000340, },
		{ -0.00000434, -0.00000208, -0.00000223, -0.00000410, -0.00000187, -0.00000237, -0.00000347, -0.00000290, -0.00000408, -0.00000447, -0.00000463, -0.00000484, -0.00000886, -0.00000727, -0.00001067, -0.00001199, -0.00000984, -0.00002097, -0.00002870, -0.00001869, -0.00009323, -0.00012512, 0.00004850, 0.00017711, 0.00006981, 0.00002935, 0.00006689, 0.00003771, 0.00004156, 0.00006353, 0.00003193, 0.00014193, 0.00021804, -0.00022606, -0.00007915, -0.00004594, -0.00002702, -0.00000025, -0.00001433, -0.00000664, -0.00002784, 0.00000270, 0.00000595, 0.00000034, -0.00002442, 0.00001448, -0.00001371, 0.00000451, -0.00001206, 0.00000482, -0.00000831, 0.00000809, -0.00002486, 0.00002567, -0.00002102, 0.00000932, -0.00000708, 0.00001131, -0.00002678, 0.00003079, -0.00005782, 0.00007094, -0.00006842, 0.00005058, -0.00001740, -0.00002857, 0.00007246, -0.00008882, 0.00004471, -0.00000535, -0.00000561, -0.00000354, 0.00001646, -0.00000861, 0.00000712, -0.00000880, 0.00000497, -0.00000312, -0.00000152, 0.00000006, 0.00000497, -0.00000344, 0.00000380, 0.00000039, 0.00000069, -0.00000175, -0.00000063, 0.00000022, -0.00000207, -0.00000510, 0.00000291, -0.00000178, -0.00000694, 0.00000380, 0.00000930, 0.00000173, 0.00000031, -0.00000472, -0.00000332, 0.00000759, 0.00000703, 0.00000001, -0.00000281, -0.00000768, 0.00000261, 0.00000678, -0.00000424, -0.00000510, -0.00000009, -0.00000144, 0.00000294, -0.00000366, -0.00000589, 0.00000279, 0.00000014, -0.00000697, 0.00000098, -0.00000079, -0.00000167, 0.00000220, -0.00000217, -0.00000774, -0.00000302, -0.00000161, 0.00000157, 0.00000069, -0.00000236, -0.00000517, },
		{ -0.00000459, -0.00000568, -0.00000186, -0.00000203, -0.00000462, -0.00000178, -0.00000228, -0.00000272, -0.00000280, -0.00000527, -0.00000352, -0.00000534, -0.00000529, -0.00000711, -0.00000836, -0.00000844, -0.00001094, -0.00001365, -0.00001408, -0.00003096, -0.00001892, -0.00005849, -0.00014381, -0.00003376, 0.00017774, 0.00011914, 0.00002270, 0.00006937, 0.00005336, 0.00003385, 0.00007194, 0.00003826, 0.00013245, 0.00024444, -0.00023201, -0.00009117, -0.00003956, -0.00003601, 0.00000217, -0.00001353, -0.00001283, -0.00002282, 0.00000240, -0.00000397, 0.00001214, -0.00002850, 0.00001094, -0.00000805, -0.00000099, -0.00001178, 0.00001161, -0.00001704, 0.00001601, -0.00003021, 0.00002211, -0.00001190, 0.00000203, -0.00001127, 0.00002772, -0.00004547, 0.00005158, -0.00007658, 0.00007262, -0.00004942, 0.00001299, 0.00002749, -0.00006590, 0.00008516, -0.00006692, 0.00000828, 0.00001313, -0.00000165, -0.00001473, 0.00001976, -0.00000486, 0.00000568, -0.00000731, 0.00000224, -0.00000102, -0.00000099, -0.00000276, 0.00000586, -0.00000192, 0.00000237, 0.00000087, 0.00000117, -0.00000130, -0.00000102, -0.00000036, -0.00000022, -0.00000536, 0.00000039, 0.00000118, -0.00000631, -0.00000128, 0.00000962, 0.00000419, 0.00000092, -0.00000233, -0.00000587, 0.00000370, 0.00000949, 0.00000169, -0.00000140, -0.00000699, -0.00000184, 0.00000730, -0.00000036, -0.00000669, -0.00000054, -0.00000153, 0.00000186, -0.00000068, -0.00000646, -0.00000040, 0.00000238, -0.00000660, -0.00000103, 0.00000074, -0.00000222, 0.00000091, 0.00000065, -0.00000657, -0.00000541, -0.00000269, 0.00000130, 0.00000081, -0.00000068, },
		{ 0.00000042, -0.00000301, -0.00000687, -0.00000233, -0.00000140, -0.00000514, -0.00000183, -0.00000222, -0.00000235, -0.00000180, -0.00000672, -0.00000322, -0.00000482, -0.00000685, -0.00000539, -0.00000853, -0.00000832, -0.00000721, -0.00001799, -0.00001100, -0.00002616, -0.00002747, -0.00003125, -0.00012852, -0.00011292, 0.00013515, 0.00017309, 0.00003561, 0.00005893, 0.00007374, 0.00003153, 0.00007140, 0.00005294, 0.00012167, 0.00026547, -0.00023012, -0.00010746, -0.00003399, -0.00004091, -0.00000093, -0.00000879, -0.00001835, -0.00002396, 0.00001017, -0.00001717, 0.00001912, -0.00002464, 0.00000273, -0.00000276, -0.00000149, -0.00001871, 0.00002293, -0.00002528, 0.00001941, -0.00002809, 0.00001252, -0.00000395, 0.00000431, -0.00002853, 0.00005291, -0.00006627, 0.00006715, -0.00007940, 0.00005148, -0.00001067, -0.00003237, 0.00006350, -0.00007976, 0.00007008, -0.00003041, -0.00001970, 0.00001213, 0.00001352, -0.00002369, 0.00001580, 0.00000215, 0.00000447, -0.00000627, 0.00000064, -0.00000091, 0.00000092, -0.00000477, 0.00000480, 0.00000039, 0.00000152, 0.00000051, 0.00000188, -0.00000087, -0.00000103, -0.00000114, 0.00000071, -0.00000409, -0.00000225, 0.00000201, -0.00000339, -0.00000528, 0.00000725, 0.00000703, 0.00000163, -0.00000030, -0.00000590, -0.00000119, 0.00000989, 0.00000446, -0.00000053, -0.00000526, -0.00000475, 0.00000499, 0.00000353, -0.00000634, -0.00000207, -0.00000115, 0.00000050, 0.00000103, -0.00000487, -0.00000372, 0.00000271, -0.00000472, -0.00000308, 0.00000120, -0.00000139, -0.00000094, 0.00000207, -0.00000394, -0.00000702, -0.00000457, 0.00000111, 0.00000053, },
		{ 0.00000035, 0.00000061, -0.00000092, -0.00000722, -0.00000360, -0.00000065, -0.00000536, -0.00000219, -0.00000197, -0.00000253, -0.00000032, -0.00000755, -0.00000417, -0.00000329, -0.00000842, -0.00000500, -0.00000702, -0.00000996, -0.00000295, -0.00001957, -0.00001348, -0.00001690, -0.00003537, -0.00002076, -0.00009050, -0.00016484, 0.00005307, 0.00020776, 0.00007608, 0.00004298, 0.00008843, 0.00004144, 0.00006209, 0.00007010, 0.00011704, 0.00027775, -0.00022292, -0.00012259, -0.00003388, -0.00003999, -0.00000756, -0.00000509, -0.00001793, -0.00003308, 0.00002221, -0.00002670, 0.00001686, -0.00001430, -0.00000477, -0.00000340, 0.00000574, -0.00003110, 0.00003357, -0.00002758, 0.00001533, -0.00002008, 0.00000357, -0.00000571, 0.00002118, -0.00005693, 0.00007915, -0.00007876, 0.00006669, -0.00005959, 0.00001020, 0.00003540, -0.00006853, 0.00007694, -0.00006725, 0.00003802, 0.00000249, -0.00002748, -0.00000543, 0.00002875, -0.00002284, 0.00000493, 0.00000841, 0.00000613, -0.00000659, 0.00000023, -0.00000187, 0.00000255, -0.00000494, 0.00000233, 0.00000223, 0.00000194, -0.00000052, 0.00000240, -0.00000022, -0.00000093, -0.00000164, 0.00000065, -0.00000223, -0.00000371, 0.00000067, 0.00000004, -0.00000669, 0.00000284, 0.00000886, 0.00000323, 0.00000081, -0.00000409, -0.00000520, 0.00000759, 0.00000752, 0.00000069, -0.00000370, -0.00000550, 0.00000126, 0.00000563, -0.00000393, -0.00000375, -0.00000115, -0.00000034, 0.00000123, -0.00000223, -0.00000568, 0.00000105, -0.00000247, -0.00000410, 0.00000033, 0.00000012, -0.00000220, 0.00000175, -0.00000082, -0.00000705, -0.00000708, 0.00000061, },
		{ -0.00000073, 0.00000053, 0.00000010, 0.00000089, -0.00000633, -0.00000529, -0.00000032, -0.00000501, -0.00000291, -0.00000153, -0.00000302, 0.00000092, -0.00000713, -0.00000618, -0.00000177, -0.00000874, -0.00000650, -0.00000443, -0.00001168, -0.00000063, -0.00001692, -0.00001962, -0.00000880, -0.00003553, -0.00002641, -0.00004880, -0.00017530, -0.00004802, 0.00019979, 0.00013845, 0.00003742, 0.00008806, 0.00006381, 0.00005096, 0.00008123, 0.00012327, 0.00028371, -0.00021726, -0.00013083, -0.00004078, -0.00003595, -0.00001242, -0.00000765, -0.00000970, -0.00004675, 0.00003148, -0.00002707, 0.00000614, -0.00000375, -0.00000494, -0.00001287, 0.00001871, -0.00004310, 0.00003725, -0.00002110, 0.00000590, -0.00001301, 0.00000424, -0.00002298, 0.00005085, -0.00008727, 0.00009438, -0.00007264, 0.00004460, -0.00001991, -0.00003873, 0.00007179, -0.00008213, 0.00006471, -0.00003764, 0.00000615, 0.00001781, -0.00001480, -0.00002785, 0.00003320, -0.00001094, -0.00000726, 0.00000924, 0.00001135, -0.00000719, -0.00000014, -0.00000261, 0.00000292, -0.00000352, -0.00000022, 0.00000256, 0.00000351, -0.00000148, 0.00000223, 0.00000072, -0.00000081, -0.00000178, 0.00000003, -0.00000075, -0.00000354, -0.00000177, 0.00000222, -0.00000530, -0.00000191, 0.00000846, 0.00000572, 0.00000139, -0.00000176, -0.00000701, 0.00000327, 0.00000942, 0.00000294, -0.00000271, -0.00000469, -0.00000207, 0.00000517, -0.00000057, -0.00000435, -0.00000197, -0.00000060, 0.00000050, 0.00000013, -0.00000567, -0.00000166, -0.00000113, -0.00000372, -0.00000131, 0.00000129, -0.00000222, 0.00000021, 0.00000155, -0.00000527, -0.00000945, },
		{ -0.00001069, -0.00000311, 0.00000081, -0.00000051, 0.00000175, -0.00000438, -0.00000664, -0.00000089, -0.00000408, -0.00000376, -0.00000119, -0.00000332, 0.00000131, -0.00000552, -0.00000825, -0.00000152, -0.00000722, -0.00000916, -0.00000239, -0.00001166, -0.00000133, -0.00001134, -0.00002516, -0.00000678, -0.00002678, -0.00003923, -0.00002112, -0.00014644, -0.00013781, 0.00013827, 0.00020138, 0.00005933, 0.00007185, 0.00009009, 0.00004862, 0.00008064, 0.00013843, 0.00029134, -0.00022065, -0.00013033, -0.00005101, -0.00003496, -0.00001016, -0.00001814, 0.00000294, -0.00005796, 0.00003194, -0.00001812, -0.00000657, -0.00000103, 0.00000584, -0.00002870, 0.00003089, -0.00004771, 0.00003081, -0.00000866, -0.00000136, -0.00001600, 0.00002069, -0.00005424, 0.00008339, -0.00010578, 0.00008761, -0.00004367, 0.00000456, 0.00002747, -0.00007833, 0.00008506, -0.00007009, 0.00003547, -0.00000695, -0.00001108, 0.00001231, 0.00000816, -0.00004167, 0.00002266, 0.00000603, -0.00001344, 0.00000275, 0.00001775, -0.00000582, -0.00000148, -0.00000263, 0.00000218, -0.00000165, -0.00000159, 0.00000127, 0.00000533, -0.00000154, 0.00000124, 0.00000170, -0.00000049, -0.00000173, -0.00000061, -0.00000017, -0.00000220, -0.00000389, 0.00000230, -0.00000221, -0.00000512, 0.00000560, 0.00000821, 0.00000229, 0.00000001, -0.00000647, -0.00000145, 0.00000895, 0.00000606, -0.00000177, -0.00000344, -0.00000380, 0.00000273, 0.00000212, -0.00000324, -0.00000325, -0.00000079, -0.00000042, 0.00000145, -0.00000407, -0.00000408, -0.00000136, -0.00000240, -0.00000280, 0.00000144, -0.00000113, -0.00000148, 0.00000234, -0.00000228, },
		{ 0.00000077, -0.00001005, -0.00000607, 0.00000055, -0.00000063, 0.00000150, -0.00000212, -0.00000685, -0.00000242, -0.00000301, -0.00000425, -0.00000133, -0.00000309, 0.00000087, -0.00000353, -0.00000916, -0.00000310, -0.00000456, -0.00001129, -0.00000251, -0.00000933, -0.00000395, -0.00000607, -0.00002621, -0.00001179, -0.00001406, -0.00004777, -0.00001565, -0.00009566, -0.00018858, 0.00003268, 0.00023419, 0.00011598, 0.00005163, 0.00010651, 0.00006254, 0.00007046, 0.00015382, 0.00030998, -0.00023618, -0.00012566, -0.00005691, -0.00004288, 0.00000054, -0.00003270, 0.00001284, -0.00006050, 0.00002286, -0.00000608, -0.00001219, -0.00001106, 0.00002485, -0.00004322, 0.00003459, -0.00004148, 0.00001742, 0.00000114, 0.00000320, -0.00003513, 0.00005149, -0.00008920, 0.00010377, -0.00010016, 0.00005503, 0.00000269, -0.00004095, 0.00006615, -0.00009500, 0.00007192, -0.00004091, 0.00000467, 0.00001060, -0.00000883, -0.00000624, 0.00002718, -0.00003910, 0.00000182, 0.00001842, -0.00000926, -0.00000826, 0.00002104, -0.00000097, -0.00000345, -0.00000250, 0.00000124, -0.00000045, -0.00000139, -0.00000082, 0.00000631, -0.00000038, -0.00000012, 0.00000222, 0.00000016, -0.00000164, -0.00000096, -0.00000038, -0.00000063, -0.00000471, 0.00000059, 0.00000082, -0.00000571, 0.00000124, 0.00000935, 0.00000407, 0.00000094, -0.00000447, -0.00000484, 0.00000592, 0.00000895, -0.00000007, -0.00000255, -0.00000388, -0.00000024, 0.00000294, -0.00000082, -0.00000408, -0.00000139, -0.00000111, 0.00000168, -0.00000187, -0.00000522, -0.00000292, -0.00000112, -0.00000340, 0.00000056, 0.00000031, -0.00000237, 0.00000154, },
		{ -0.00000011, 0.00000269, -0.00000761, -0.00000862, -0.00000089, -0.00000016, 0.00000069, -0.00000054, -0.00000558, -0.00000435, -0.00000251, -0.00000399, -0.00000207, -0.00000246, 0.00000017, -0.00000226, -0.00000827, -0.00000589, -0.00000238, -0.00001131, -0.00000509, -0.00000594, -0.00000597, -0.00000405, -0.00002185, -0.00002022, -0.00000496, -0.00004485, -0.00002812, -0.00004724, -0.00018751, -0.00008796, 0.00021016, 0.00019630, 0.00004874, 0.00010265, 0.00009070, 0.00006114, 0.00015867, 0.00034422, -0.00025918, -0.00012637, -0.00005140, -0.00006073, 0.00001495, -0.00004373, 0.00001373, -0.00005352, 0.00001027, -0.00000044, -0.00000458, -0.00003160, 0.00004325, -0.00004751, 0.00002617, -0.00002798, 0.00000661, -0.00000236, 0.00002560, -0.00006843, 0.00008603, -0.00011203, 0.00009854, -0.00006628, 0.00000349, 0.00005243, -0.00007593, 0.00008314, -0.00008502, 0.00004064, -0.00000999, -0.00001362, 0.00000957, 0.00000694, -0.00002451, 0.00003200, -0.00002191, -0.00001863, 0.00001869, 0.00000358, -0.00001759, 0.00001784, 0.00000646, -0.00000425, -0.00000321, 0.00000077, -0.00000026, -0.00000020, -0.00000251, 0.00000580, 0.00000158, -0.00000107, 0.00000192, 0.00000108, -0.00000146, -0.00000108, -0.00000092, 0.00000039, -0.00000412, -0.00000173, 0.00000235, -0.00000390, -0.00000290, 0.00000827, 0.00000648, 0.00000156, -0.00000229, -0.00000594, 0.00000134, 0.00001019, 0.00000277, -0.00000195, -0.00000311, -0.00000234, 0.00000178, 0.00000169, -0.00000368, -0.00000237, -0.00000169, 0.00000126, 0.00000004, -0.00000486, -0.00000491, -0.00000071, -0.00000305, -0.00000077, 0.00000127, -0.00000208, },
		{ -0.00000097, -0.00000154, 0.00000292, -0.00000426, -0.00000962, -0.00000355, 0.00000034, 0.00000016, -0.00000028, -0.00000325, -0.00000569, -0.00000316, -0.00000298, -0.00000303, -0.00000199, 0.00000002, -0.00000230, -0.00000601, -0.00000835, -0.00000225, -0.00000887, -0.00000876, -0.00000384, -0.00000523, -0.00000594, -0.00001471, -0.00002624, -0.00000472, -0.00003137, -0.00004541, -0.00002068, -0.00014259, -0.00018571, 0.00012078, 0.00027045, 0.00008334, 0.00008053, 0.00012020, 0.00006621, 0.00014739, 0.00038970, -0.00027866, -0.00014168, -0.00003325, -0.00008256, 0.00002447, -0.00004476, 0.00000500, -0.00004319, 0.00000391, -0.00000815, 0.00001486, -0.00005323, 0.00005046, -0.00003730, 0.00000991, -0.00001768, 0.00000999, -0.00002562, 0.00006314, -0.00010448, 0.00010801, -0.00010844, 0.00006294, -0.00001163, -0.00005123, 0.00008877, -0.00008843, 0.00007504, -0.00005597, 0.00000658, 0.00000840, -0.00001369, -0.00000502, 0.00002405, -0.00003146, 0.00002136, 0.00000023, -0.00002823, 0.00000596, 0.00001794, -0.00001920, 0.00000828, 0.00001310, -0.00000195, -0.00000496, 0.00000061, -0.00000062, 0.00000096, -0.00000298, 0.00000408, 0.00000337, -0.00000096, 0.00000090, 0.00000188, -0.00000100, -0.00000110, -0.00000137, 0.00000060, -0.00000271, -0.00000343, 0.00000193, -0.00000097, -0.00000521, 0.00000507, 0.00000850, 0.00000264, -0.00000077, -0.00000493, -0.00000308, 0.00000886, 0.00000617, -0.00000091, -0.00000241, -0.00000302, -0.00000043, 0.00000303, -0.00000203, -0.00000314, -0.00000246, 0.00000058, 0.00000122, -0.00000346, -0.00000639, -0.00000139, -0.00000230, -0.00000184, 0.00000133, },
		{ 0.00000068, 0.00000021, -0.00000202, 0.00000177, -0.00000132, -0.00000852, -0.00000670, -0.00000002, 0.00000052, -0.00000121, -0.00000093, -0.00000553, -0.00000494, -0.00000185, -0.00000352, -0.00000229, 0.00000078, -0.00000337, -0.00000371, -0.00000894, -0.00000454, -0.00000530, -0.00001115, -0.00000487, -0.00000156, -0.00000961, -0.00000926, -0.00002561, -0.00001287, -0.00001525, -0.00005374, -0.00002189, -0.00007875, -0.00022975, -0.00001643, 0.00029990, 0.00016061, 0.00005823, 0.00013335, 0.00009332, 0.00012508, 0.00043418, -0.00028337, -0.00017400, -0.00000962, -0.00009803, 0.00002208, -0.00003563, -0.00000661, -0.00003961, 0.00001114, -0.00002834, 0.00003659, -0.00006397, 0.00004088, -0.00001737, -0.00000227, -0.00002319, 0.00003447, -0.00006549, 0.00010311, -0.00012616, 0.00010280, -0.00007317, 0.00000467, 0.00004665, -0.00009086, 0.00009992, -0.00007631, 0.00004927, -0.00002256, -0.00001564, 0.00000808, -0.00000013, -0.00002172, 0.00003158, -0.00002405, 0.00000274, 0.00001571, -0.00002266, -0.00001332, 0.00002522, -0.00001108, -0.00000342, 0.00001491, 0.00000388, -0.00000656, -0.00000005, -0.00000083, 0.00000140, -0.00000225, 0.00000207, 0.00000410, 0.00000023, -0.00000026, 0.00000207, -0.00000018, -0.00000107, -0.00000159, 0.00000024, -0.00000135, -0.00000383, 0.00000015, 0.00000147, -0.00000506, 0.00000094, 0.00000900, 0.00000449, 0.00000003, -0.00000283, -0.00000582, 0.00000520, 0.00000881, 0.00000126, -0.00000211, -0.00000268, -0.00000237, 0.00000269, 0.00000007, -0.00000302, -0.00000346, -0.00000028, 0.00000171, -0.00000175, -0.00000682, -0.00000275, -0.00000183, -0.00000230, },
		{ -0.00000231, -0.00000017, 0.00000083, -0.00000146, 0.00000017, 0.00000017, -0.00000570, -0.00000904, -0.00000192, 0.00000162, -0.00000245, 0.00000024, -0.00000365, -0.00000703, -0.00000152, -0.00000299, -0.00000345, 0.00000207, -0.00000445, -0.00000275, -0.00000714, -0.00000807, -0.00000288, -0.00001043, -0.00000890, 0.00000287, -0.00001155, -0.00000877, -0.00001843, -0.00002358, -0.00000602, -0.00004659, -0.00004165, -0.00002580, -0.00021038, -0.00016183, 0.00025500, 0.00026166, 0.00006376, 0.00011921, 0.00013697, 0.00010671, 0.00046461, -0.00026965, -0.00021569, 0.00000684, -0.00009880, 0.00000807, -0.00002418, -0.00001027, -0.00005037, 0.00003120, -0.00005138, 0.00004779, -0.00005685, 0.00001911, -0.00000127, 0.00000379, -0.00005173, 0.00007645, -0.00010809, 0.00012705, -0.00011848, 0.00006522, -0.00001399, -0.00005820, 0.00008880, -0.00010293, 0.00008476, -0.00004804, 0.00001987, 0.00000099, -0.00001905, -0.00000684, 0.00001586, -0.00002976, 0.00002572, -0.00000792, -0.00001283, 0.00001751, -0.00000586, -0.00002899, 0.00002033, 0.00000312, -0.00001109, 0.00000989, 0.00001118, -0.00000602, -0.00000184, -0.00000068, 0.00000117, -0.00000105, 0.00000075, 0.00000351, 0.00000191, -0.00000081, 0.00000146, 0.00000080, -0.00000088, -0.00000165, -0.00000028, -0.00000060, -0.00000307, -0.00000185, 0.00000233, -0.00000298, -0.00000252, 0.00000739, 0.00000667, 0.00000071, -0.00000081, -0.00000628, 0.00000049, 0.00000937, 0.00000455, -0.00000171, -0.00000222, -0.00000320, 0.00000113, 0.00000148, -0.00000176, -0.00000426, -0.00000145, 0.00000170, -0.00000021, -0.00000629, -0.00000406, -0.00000203, },
		{ -0.00000275, -0.00000234, -0.00000078, 0.00000076, -0.00000041, -0.00000088, -0.00000006, -0.00000236, -0.00000943, -0.00000517, 0.00000258, -0.00000294, -0.00000027, -0.00000080, -0.00000820, -0.00000265, -0.00000154, -0.00000487, 0.00000298, -0.00000446, -0.00000365, -0.00000392, -0.00001066, -0.00000342, -0.00000671, -0.00001367, 0.00000491, -0.00000938, -0.00001312, -0.00000908, -0.00002942, -0.00000870, -0.00002798, -0.00006209, -0.00000423, -0.00014351, -0.00026897, 0.00013172, 0.00034564, 0.00012022, 0.00008437, 0.00017855, 0.00010908, 0.00047646, -0.00024571, -0.00025208, 0.00000552, -0.00008524, -0.00000807, -0.00002258, 0.00000196, -0.00007446, 0.00005409, -0.00006382, 0.00004046, -0.00003588, -0.00000017, -0.00000526, 0.00003608, -0.00009893, 0.00012082, -0.00013354, 0.00011906, -0.00007657, 0.00000354, 0.00005052, -0.00010470, 0.00010130, -0.00008659, 0.00005314, -0.00001814, 0.00000012, 0.00000763, -0.00000681, -0.00002532, 0.00002335, -0.00002499, 0.00001101, 0.00000694, -0.00001730, 0.00000670, 0.00001249, -0.00003292, 0.00000483, 0.00001585, -0.00001013, -0.00000026, 0.00001612, -0.00000191, -0.00000427, -0.00000072, 0.00000083, -0.00000017, 0.00000056, 0.00000207, 0.00000316, -0.00000032, 0.00000032, 0.00000149, -0.00000035, -0.00000168, -0.00000065, -0.00000055, -0.00000181, -0.00000303, 0.00000150, -0.00000031, -0.00000404, 0.00000409, 0.00000819, 0.00000189, 0.00000052, -0.00000489, -0.00000357, 0.00000730, 0.00000801, -0.00000036, -0.00000221, -0.00000299, -0.00000049, 0.00000148, 0.00000009, -0.00000423, -0.00000292, 0.00000121, 0.00000100, -0.00000518, -0.00000480, },
		{ -0.00000490, -0.00000352, -0.00000275, -0.00000114, 0.00000030, 0.00000047, -0.00000090, -0.00000137, 0.00000009, -0.00000751, -0.00000869, 0.00000223, -0.00000213, -0.00000199, 0.00000165, -0.00000744, -0.00000507, -0.00000005, -0.00000559, 0.00000273, -0.00000313, -0.00000570, -0.00000117, -0.00001053, -0.00000693, -0.00000213, -0.00001608, 0.00000262, -0.00000364, -0.00001877, -0.00000334, -0.00002616, -0.00002065, -0.00000936, -0.00006744, -0.00001611, -0.00006307, -0.00030480, -0.00004144, 0.00036487, 0.00022885, 0.00005534, 0.00019589, 0.00013913, 0.00047957, -0.00022865, -0.00027023, -0.00001405, -0.00006866, -0.00001203, -0.00003941, 0.00002852, -0.00010102, 0.00006562, -0.00005699, 0.00001824, -0.00001643, 0.00000129, -0.00003885, 0.00008960, -0.00014790, 0.00014624, -0.00012498, 0.00007430, -0.00000973, -0.00006286, 0.00009870, -0.00012026, 0.00008305, -0.00005273, 0.00002082, 0.00000034, -0.00000385, -0.00000013, 0.00001091, -0.00003608, 0.00001769, -0.00001143, -0.00000330, 0.00001243, -0.00000983, -0.00000895, 0.00002268, -0.00002337, -0.00001384, 0.00001982, -0.00000050, -0.00001052, 0.00001532, 0.00000518, -0.00000569, -0.00000174, 0.00000078, 0.00000006, 0.00000120, 0.00000069, 0.00000335, 0.00000103, -0.00000074, 0.00000152, 0.00000051, -0.00000165, -0.00000083, -0.00000090, -0.00000079, -0.00000303, -0.00000028, 0.00000157, -0.00000338, 0.00000037, 0.00000810, 0.00000367, 0.00000126, -0.00000274, -0.00000570, 0.00000320, 0.00001019, 0.00000241, -0.00000247, -0.00000253, -0.00000121, 0.00000022, 0.00000149, -0.00000302, -0.00000426, 0.00000008, 0.00000184, -0.00000388, },
		{ -0.00000249, -0.00000457, -0.00000393, -0.00000345, -0.00000156, -0.00000017, 0.00000080, -0.00000012, -0.00000268, 0.00000077, -0.00000407, -0.00001093, -0.00000002, -0.00000048, -0.00000366, 0.00000242, -0.00000475, -0.00000765, 0.00000031, -0.00000493, 0.00000132, -0.00000117, -0.00000735, -0.00000052, -0.00000748, -0.00001145, 0.00000031, -0.00001421, -0.00000318, 0.00000233, -0.00002116, -0.00000454, -0.00001554, -0.00003328, -0.00000207, -0.00005294, -0.00004589, -0.00000457, -0.00026489, -0.00021165, 0.00028744, 0.00035974, 0.00006835, 0.00017859, 0.00018597, 0.00049520, -0.00023435, -0.00026869, -0.00003921, -0.00006634, 0.00000629, -0.00007248, 0.00005707, -0.00011474, 0.00005667, -0.00003453, -0.00000287, -0.00001830, 0.00003477, -0.00009710, 0.00014540, -0.00017523, 0.00013494, -0.00007751, 0.00000297, 0.00006133, -0.00011165, 0.00011543, -0.00010324, 0.00004552, -0.00001845, 0.00000151, 0.00000205, 0.00000489, -0.00001308, 0.00002316, -0.00003360, 0.00000240, 0.00000199, -0.00000923, 0.00000680, 0.00000352, -0.00001988, 0.00002020, -0.00000581, -0.00002642, 0.00001216, 0.00001284, -0.00001518, 0.00000806, 0.00001242, -0.00000420, -0.00000386, 0.00000085, -0.00000008, 0.00000199, 0.00000009, 0.00000247, 0.00000257, -0.00000107, 0.00000081, 0.00000141, -0.00000140, -0.00000093, -0.00000128, -0.00000038, -0.00000215, -0.00000195, 0.00000192, -0.00000127, -0.00000236, 0.00000616, 0.00000550, 0.00000199, -0.00000086, -0.00000563, -0.00000143, 0.00000993, 0.00000618, -0.00000215, -0.00000261, -0.00000081, -0.00000129, 0.00000155, -0.00000100, -0.00000476, -0.00000171, 0.00000219, },
		{ 0.00000170, -0.00000102, -0.00000403, -0.00000391, -0.00000405, -0.00000233, -0.00000055, 0.00000071, 0.00000083, -0.00000314, -0.00000018, -0.00000060, -0.00001076, -0.00000371, 0.00000072, -0.00000405, 0.00000118, -0.00000125, -0.00000886, -0.00000119, -0.00000317, -0.00000035, 0.00000010, -0.00000734, -0.00000225, -0.00000315, -0.00001414, -0.00000131, -0.00000881, -0.00000915, 0.00000453, -0.00001793, -0.00001125, -0.00000415, -0.00003756, -0.00001021, -0.00002681, -0.00007044, 0.00001175, -0.00017564, -0.00032302, 0.00011647, 0.00045785, 0.00014908, 0.00014046, 0.00022325, 0.00054427, -0.00026552, -0.00026174, -0.00004957, -0.00009101, 0.00004434, -0.00010732, 0.00007059, -0.00010597, 0.00003127, -0.00001336, -0.00000166, -0.00005434, 0.00009582, -0.00015934, 0.00017730, -0.00016176, 0.00008227, -0.00000228, -0.00007276, 0.00011306, -0.00012746, 0.00009890, -0.00006519, 0.00000690, 0.00000145, 0.00000047, -0.00000892, 0.00001706, -0.00002135, 0.00002417, -0.00002034, -0.00001374, 0.00000715, -0.00000466, -0.00000498, 0.00001406, -0.00002033, 0.00000780, 0.00001046, -0.00002693, -0.00000376, 0.00002248, -0.00001115, -0.00000277, 0.00001602, 0.00000091, -0.00000608, 0.00000034, -0.00000012, 0.00000235, 0.00000042, 0.00000112, 0.00000351, -0.00000046, -0.00000026, 0.00000195, -0.00000081, -0.00000101, -0.00000149, -0.00000052, -0.00000104, -0.00000274, 0.00000090, 0.00000101, -0.00000321, 0.00000299, 0.00000646, 0.00000319, 0.00000033, -0.00000399, -0.00000497, 0.00000706, 0.00000966, -0.00000036, -0.00000339, 0.00000004, -0.00000195, 0.00000020, 0.00000080, -0.00000386, -0.00000373, },
		{ -0.00000512, 0.00000009, 0.00000034, -0.00000325, -0.00000374, -0.00000414, -0.00000340, -0.00000111, 0.00000052, 0.00000140, -0.00000259, -0.00000182, 0.00000145, -0.00000816, -0.00000733, 0.00000012, -0.00000277, -0.00000117, 0.00000133, -0.00000767, -0.00000412, -0.00000151, -0.00000097, -0.00000027, -0.00000556, -0.00000503, -0.00000007, -0.00001307, -0.00000642, -0.00000307, -0.00001140, 0.00000108, -0.00001072, -0.00001816, 0.00000088, -0.00002988, -0.00002802, -0.00000459, -0.00007266, -0.00001042, -0.00008165, -0.00034134, -0.00010480, 0.00046463, 0.00029200, 0.00011981, 0.00022347, 0.00063235, -0.00030590, -0.00027366, -0.00002926, -0.00014072, 0.00008510, -0.00012418, 0.00005857, -0.00007955, 0.00000786, -0.00001625, 0.00003602, -0.00012051, 0.00016249, -0.00019655, 0.00016389, -0.00010293, 0.00000079, 0.00007709, -0.00012801, 0.00012953, -0.00010878, 0.00006089, -0.00002484, -0.00001681, 0.00000124, 0.00001287, -0.00002244, 0.00002326, -0.00001994, 0.00001555, -0.00000424, -0.00002223, 0.00000146, 0.00000600, -0.00001494, 0.00001576, -0.00001112, -0.00000669, 0.00001755, -0.00001581, -0.00002007, 0.00002261, -0.00000001, -0.00001187, 0.00001343, 0.00000819, -0.00000654, -0.00000131, 0.00000007, 0.00000222, 0.00000124, 0.00000008, 0.00000347, 0.00000079, -0.00000108, 0.00000182, 0.00000002, -0.00000100, -0.00000155, -0.00000089, -0.00000027, -0.00000252, -0.00000070, 0.00000233, -0.00000223, -0.00000018, 0.00000588, 0.00000481, 0.00000106, -0.00000190, -0.00000642, 0.00000254, 0.00001133, 0.00000306, -0.00000416, 0.00000031, -0.00000124, -0.00000171, 0.00000125, -0.00000171, },
		{ 0.00000069, -0.00000496, -0.00000241, 0.00000104, -0.00000200, -0.00000369, -0.00000369, -0.00000427, -0.00000221, 0.00000042, 0.00000151, -0.00000152, -0.00000300, 0.00000150, -0.00000433, -0.00000911, -0.00000275, -0.00000070, -0.00000299, 0.00000166, -0.00000436, -0.00000697, -0.00000142, 0.00000017, -0.00000206, -0.00000326, -0.00000681, -0.00000002, -0.00000855, -0.00001210, -0.00000074, -0.00000811, -0.00000614, -0.00000436, -0.00001964, -0.00000338, -0.00001434, -0.00004365, 0.00000088, -0.00005087, -0.00004761, -0.00002422, -0.00027017, -0.00030718, 0.00034686, 0.00045160, 0.00016351, 0.00017806, 0.00074113, -0.00032616, -0.00032443, 0.00002072, -0.00019615, 0.00010504, -0.00011086, 0.00002685, -0.00005610, 0.00001119, -0.00005851, 0.00010629, -0.00019366, 0.00020317, -0.00018462, 0.00009996, -0.00001306, -0.00008380, 0.00013441, -0.00014611, 0.00010942, -0.00006814, 0.00002053, 0.00000112, -0.00001904, -0.00001395, 0.00002755, -0.00002873, 0.00001932, -0.00001079, 0.00000410, 0.00000647, -0.00001959, -0.00001112, 0.00001506, -0.00001708, 0.00000843, 0.00000150, -0.00001518, 0.00001311, 0.00000055, -0.00002873, 0.00001231, 0.00001257, -0.00001437, 0.00000515, 0.00001441, -0.00000371, -0.00000378, 0.00000005, 0.00000196, 0.00000196, -0.00000019, 0.00000263, 0.00000201, -0.00000118, 0.00000110, 0.00000080, -0.00000076, -0.00000158, -0.00000122, -0.00000007, -0.00000168, -0.00000197, 0.00000229, -0.00000028, -0.00000219, 0.00000377, 0.00000622, 0.00000186, -0.00000022, -0.00000578, -0.00000200, 0.00001028, 0.00000725, -0.00000376, -0.00000056, 0.00000035, -0.00000284, -0.00000008, },
		{ -0.00000254, 0.00000192, -0.00000296, -0.00000490, 0.00000044, -0.00000025, -0.00000367, -0.00000314, -0.00000437, -0.00000389, 0.00000017, 0.00000149, -0.00000063, -0.00000308, 0.00000008, -0.00000105, -0.00000804, -0.00000699, 0.00000047, -0.00000292, -0.00000028, -0.00000051, -0.00000790, -0.00000361, 0.00000255, -0.00000391, -0.00000219, -0.00000614, -0.00000285, -0.00000311, -0.00001479, -0.00000366, -0.00000102, -0.00001242, -0.00000373, -0.00001366, -0.00001361, -0.00000020, -0.00004669, -0.00001249, -0.00001887, -0.00007094, -0.00002178, -0.00014950, -0.00042405, 0.00011845, 0.00055383, 0.00029967, 0.00011173, 0.00083434, -0.00030096, -0.00041454, 0.00007855, -0.00022948, 0.00008895, -0.00007409, -0.00000078, -0.00006319, 0.00005790, -0.00013605, 0.00018448, -0.00023957, 0.00019098, -0.00011709, 0.00000126, 0.00007937, -0.00014353, 0.00015253, -0.00012589, 0.00006606, -0.00002554, -0.00000512, 0.00000552, -0.00000459, -0.00003218, 0.00003386, -0.00002381, 0.00000803, -0.00000049, -0.00000301, 0.00000791, -0.00000875, -0.00002299, 0.00001639, -0.00001080, -0.00000275, 0.00001007, -0.00001387, 0.00000113, 0.00001355, -0.00002577, -0.00000378, 0.00001976, -0.00000865, -0.00000515, 0.00001611, 0.00000239, -0.00000568, -0.00000085, 0.00000188, 0.00000224, 0.00000024, 0.00000161, 0.00000259, -0.00000049, 0.00000020, 0.00000116, -0.00000021, -0.00000159, -0.00000138, -0.00000027, -0.00000081, -0.00000241, 0.00000123, 0.00000151, -0.00000248, 0.00000087, 0.00000666, 0.00000304, 0.00000080, -0.00000389, -0.00000506, 0.00000672, 0.00001063, -0.00000140, -0.00000221, 0.00000160, -0.00000226, },
		{ -0.00000010, -0.00000464, 0.00000099, 0.00000012, -0.00000610, -0.00000173, 0.00000150, -0.00000321, -0.00000303, -0.00000348, -0.00000562, -0.00000077, 0.00000168, -0.00000026, -0.00000225, -0.00000156, 0.00000029, -0.00000452, -0.00001066, -0.00000077, -0.00000087, -0.00000311, 0.00000179, -0.00000593, -0.00000732, 0.00000450, -0.00000419, -0.00000338, -0.00000330, -0.00000649, 0.00000008, -0.00001259, -0.00001037, 0.00000541, -0.00001309, -0.00001031, -0.00000346, -0.00002232, 0.00000419, -0.00003490, -0.00003475, 0.00000314, -0.00006282, -0.00006258, -0.00003668, -0.00042142, -0.00015750, 0.00052625, 0.00051245, 0.00007956, 0.00087867, -0.00022841, -0.00051911, 0.00011122, -0.00022174, 0.00004364, -0.00004157, 0.00000730, -0.00011926, 0.00014334, -0.00022209, 0.00023359, -0.00022867, 0.00011803, -0.00001063, -0.00010075, 0.00014371, -0.00016062, 0.00013078, -0.00008140, 0.00002045, 0.00000134, -0.00000881, -0.00000711, 0.00001435, -0.00004174, 0.00002748, -0.00001116, -0.00000341, 0.00000446, -0.00000261, 0.00000205, 0.00000342, -0.00002755, 0.00000894, -0.00000073, -0.00001094, 0.00001031, -0.00000485, -0.00001085, 0.00001709, -0.00001331, -0.00001794, 0.00001734, 0.00000249, -0.00001230, 0.00001170, 0.00000967, -0.00000521, -0.00000285, 0.00000187, 0.00000219, 0.00000091, 0.00000102, 0.00000236, 0.00000059, -0.00000036, 0.00000092, 0.00000050, -0.00000150, -0.00000143, -0.00000059, -0.00000033, -0.00000204, -0.00000010, 0.00000233, -0.00000135, -0.00000171, 0.00000571, 0.00000444, 0.00000145, -0.00000180, -0.00000593, 0.00000194, 0.00001170, 0.00000269, -0.00000340, 0.00000143, },
		{ -0.00000038, 0.00000244, -0.00000487, -0.00000194, 0.00000269, -0.00000505, -0.00000489, 0.00000232, -0.00000181, -0.00000353, -0.00000209, -0.00000652, -0.00000270, 0.00000196, -0.00000018, -0.00000127, -0.00000226, -0.00000052, -0.00000022, -0.00001180, -0.00000461, 0.00000180, -0.00000479, 0.00000125, -0.00000175, -0.00001052, 0.00000428, -0.00000218, -0.00000626, -0.00000017, -0.00000825, -0.00000080, -0.00000662, -0.00001677, 0.00000668, -0.00000671, -0.00002077, 0.00000428, -0.00002287, -0.00000308, -0.00001571, -0.00005025, 0.00000137, -0.00002637, -0.00011331, 0.00002117, -0.00031439, -0.00039054, 0.00033471, 0.00073410, 0.00014306, 0.00087070, -0.00013941, -0.00059835, 0.00009497, -0.00017948, 0.00000098, -0.00005064, 0.00007241, -0.00021887, 0.00023782, -0.00027623, 0.00022120, -0.00015161, 0.00000173, 0.00010046, -0.00017212, 0.00016075, -0.00013521, 0.00008449, -0.00003457, -0.00000892, 0.00000504, 0.00000475, -0.00002460, 0.00002553, -0.00003748, 0.00001211, 0.00000116, -0.00000834, 0.00000171, 0.00000321, -0.00000542, 0.00001048, -0.00002287, -0.00000304, 0.00000654, -0.00001182, 0.00000325, 0.00000572, -0.00001620, 0.00001061, 0.00000186, -0.00002353, 0.00000605, 0.00001313, -0.00001246, 0.00000269, 0.00001478, -0.00000134, -0.00000522, 0.00000137, 0.00000213, 0.00000135, 0.00000110, 0.00000163, 0.00000146, -0.00000024, 0.00000025, 0.00000106, -0.00000117, -0.00000147, -0.00000081, -0.00000032, -0.00000131, -0.00000098, 0.00000201, 0.00000035, -0.00000307, 0.00000358, 0.00000547, 0.00000219, -0.00000023, -0.00000489, -0.00000237, 0.00000984, 0.00000721, -0.00000277, },
		{ 0.00000025, -0.00000283, 0.00000372, -0.00000268, -0.00000545, 0.00000306, -0.00000188, -0.00000762, 0.00000128, 0.00000047, -0.00000414, -0.00000108, -0.00000597, -0.00000530, 0.00000173, 0.00000013, -0.00000081, -0.00000169, -0.00000248, 0.00000281, -0.00000966, -0.00000963, 0.00000295, -0.00000393, -0.00000175, 0.00000252, -0.00001111, 0.00000127, 0.00000120, -0.00000892, 0.00000101, -0.00000671, -0.00000490, -0.00000047, -0.00001859, 0.00000127, 0.00000360, -0.00002874, 0.00000387, -0.00001409, -0.00001601, -0.00000153, -0.00004751, -0.00002300, 0.00001744, -0.00013884, 0.00000938, -0.00016081, -0.00050077, 0.00000825, 0.00086379, 0.00033323, 0.00085410, -0.00008689, -0.00062223, 0.00003609, -0.00013953, 0.00000531, -0.00012731, 0.00018855, -0.00032834, 0.00029646, -0.00026277, 0.00013663, -0.00002660, -0.00012085, 0.00017874, -0.00019149, 0.00013110, -0.00008452, 0.00003681, -0.00000441, -0.00001429, -0.00000950, 0.00002301, -0.00003455, 0.00002317, -0.00002273, -0.00000345, 0.00000581, -0.00000473, -0.00000585, 0.00000911, -0.00000905, 0.00000997, -0.00001225, -0.00001294, 0.00000656, -0.00000598, -0.00000596, 0.00001155, -0.00001275, -0.00000115, 0.00001213, -0.00001843, -0.00000873, 0.00001737, -0.00000536, -0.00000682, 0.00001485, 0.00000521, -0.00000640, -0.00000014, 0.00000221, 0.00000142, 0.00000163, 0.00000097, 0.00000174, 0.00000046, -0.00000045, 0.00000120, -0.00000060, -0.00000148, -0.00000089, -0.00000057, -0.00000069, -0.00000113, 0.00000101, 0.00000168, -0.00000298, 0.00000105, 0.00000552, 0.00000322, 0.00000070, -0.00000289, -0.00000485, 0.00000567, 0.00001034, },
		{ 0.00001068, 0.00000501, -0.00000422, 0.00000271, 0.00000106, -0.00000747, 0.00000057, 0.00000198, -0.00000829, -0.00000174, 0.00000274, -0.00000393, -0.00000116, -0.00000411, -0.00000755, 0.00000027, 0.00000091, -0.00000093, -0.00000044, -0.00000407, 0.00000338, -0.00000517, -0.00001343, 0.00000104, -0.00000085, -0.00000522, 0.00000447, -0.00000824, -0.00000330, 0.00000376, -0.00000935, -0.00000077, -0.00000271, -0.00000918, 0.00000200, -0.00001419, -0.00000811, 0.00001187, -0.00002885, -0.00000532, -0.00000151, -0.00002449, -0.00000174, -0.00002669, -0.00005451, 0.00004318, -0.00012224, -0.00004915, -0.00003403, -0.00045530, -0.00036282, 0.00080914, 0.00061841, 0.00091298, -0.00011540, -0.00059761, -0.00002158, -0.00015531, 0.00008900, -0.00026602, 0.00031701, -0.00039671, 0.00027950, -0.00016971, -0.00000077, 0.00010635, -0.00020862, 0.00020023, -0.00015933, 0.00007406, -0.00003381, 0.00000738, 0.00000135, -0.00000070, -0.00002927, 0.00003326, -0.00003108, 0.00001014, -0.00000640, -0.00001094, 0.00000056, 0.00000383, -0.00001245, 0.00001041, -0.00000696, 0.00000416, -0.00000167, -0.00001581, -0.00000050, 0.00000196, -0.00001156, 0.00000986, -0.00000348, -0.00001122, 0.00001313, -0.00000599, -0.00001984, 0.00001254, 0.00000538, -0.00001215, 0.00000925, 0.00001193, -0.00000481, -0.00000253, 0.00000212, 0.00000132, 0.00000217, 0.00000078, 0.00000143, 0.00000133, -0.00000075, 0.00000086, 0.00000002, -0.00000136, -0.00000091, -0.00000084, -0.00000045, -0.00000071, -0.00000003, 0.00000210, -0.00000186, -0.00000094, 0.00000439, 0.00000429, 0.00000134, -0.00000093, -0.00000512, 0.00000076, },
		{ -0.00000317, 0.00000797, 0.00000972, -0.00000310, -0.00000025, 0.00000442, -0.00000654, -0.00000388, 0.00000435, -0.00000606, -0.00000568, 0.00000361, -0.00000225, -0.00000225, -0.00000194, -0.00000824, -0.00000260, 0.00000171, -0.00000107, 0.00000049, -0.00000422, 0.00000175, -0.00000051, -0.00001397, -0.00000372, 0.00000248, -0.00000675, 0.00000292, -0.00000312, -0.00000686, 0.00000345, -0.00000694, -0.00000439, 0.00000105, -0.00001038, -0.00000083, -0.00000579, -0.00001595, 0.00001292, -0.00002043, -0.00001793, 0.00000602, -0.00002101, -0.00001639, 0.00000036, -0.00007309, 0.00003684, -0.00007371, -0.00010847, 0.00001220, -0.00028504, -0.00065891, 0.00053381, 0.00089649, 0.00113749, -0.00022335, -0.00058108, -0.00001130, -0.00026805, 0.00024796, -0.00042257, 0.00039848, -0.00037849, 0.00017339, -0.00001809, -0.00014755, 0.00020231, -0.00023475, 0.00016486, -0.00009680, 0.00001756, -0.00000401, 0.00000358, -0.00001175, 0.00001853, -0.00004106, 0.00002941, -0.00001703, -0.00000484, 0.00000275, -0.00000736, -0.00001071, 0.00001080, -0.00001340, 0.00000605, -0.00000140, -0.00000211, 0.00000391, -0.00001096, -0.00001030, 0.00000646, -0.00001066, 0.00000253, 0.00000578, -0.00001440, 0.00000573, 0.00000703, -0.00002216, 0.00000066, 0.00001402, -0.00001056, 0.00000025, 0.00001562, -0.00000001, -0.00000476, 0.00000132, 0.00000128, 0.00000240, 0.00000112, 0.00000089, 0.00000192, -0.00000052, 0.00000025, 0.00000045, -0.00000104, -0.00000091, -0.00000100, -0.00000055, -0.00000012, -0.00000060, 0.00000167, -0.00000051, -0.00000180, 0.00000245, 0.00000488, 0.00000208, 0.00000043, -0.00000369, },
		{ -0.00000162, -0.00000499, 0.00000325, 0.00001242, 0.00000080, -0.00000351, 0.00000546, -0.00000267, -0.00000806, 0.00000350, -0.00000162, -0.00000857, 0.00000202, 0.00000062, -0.00000335, -0.00000079, -0.00000684, -0.00000607, 0.00000152, -0.00000054, 0.00000054, -0.00000303, -0.00000057, 0.00000213, -0.00001088, -0.00000929, 0.00000350, -0.00000504, -0.00000119, 0.00000152, -0.00000701, -0.00000014, -0.00000319, -0.00000724, 0.00000182, -0.00000720, -0.00000714, 0.00000170, -0.00001732, 0.00000591, -0.00000847, -0.00002562, 0.00000210, -0.00000594, -0.00003586, 0.00001748, -0.00006675, 0.00000462, -0.00002364, -0.00012606, -0.00003104, -0.00007183, -0.00078125, 0.00008755, 0.00102233, 0.00157273, -0.00033862, -0.00066467, 0.00012210, -0.00047629, 0.00043162, -0.00052717, 0.00037910, -0.00025782, 0.00000153, 0.00014430, -0.00025439, 0.00023141, -0.00019836, 0.00009546, -0.00003474, -0.00001542, -0.00000246, 0.00001880, -0.00002990, 0.00003004, -0.00003850, 0.00001436, -0.00000124, -0.00001301, 0.00000111, 0.00000372, -0.00002071, 0.00001093, -0.00000807, -0.00000104, 0.00000329, -0.00000480, 0.00000316, -0.00000188, -0.00001720, 0.00000447, -0.00000464, -0.00000554, 0.00000995, -0.00000991, -0.00000496, 0.00001424, -0.00001517, -0.00001260, 0.00001580, -0.00000291, -0.00000797, 0.00001413, 0.00000670, -0.00000524, -0.00000049, 0.00000128, 0.00000234, 0.00000168, 0.00000059, 0.00000202, 0.00000008, -0.00000026, 0.00000050, -0.00000056, -0.00000087, -0.00000104, -0.00000082, 0.00000029, -0.00000059, 0.00000084, 0.00000038, -0.00000151, 0.00000043, 0.00000456, 0.00000299, 0.00000120, },
		{ 0.00000173, 0.00000019, -0.00000463, -0.00000168, 0.00001187, 0.00000626, -0.00000496, 0.00000345, 0.00000240, -0.00000950, -0.00000074, 0.00000284, -0.00000854, -0.00000190, 0.00000331, -0.00000311, -0.00000139, -0.00000396, -0.00000860, -0.00000051, 0.00000069, 0.00000001, -0.00000152, -0.00000184, 0.00000184, -0.00000574, -0.00001280, 0.00000076, -0.00000097, -0.00000520, 0.00000295, -0.00000318, -0.00000514, -0.00000093, -0.00000696, -0.00000120, -0.00000135, -0.00001273, 0.00000379, -0.00001117, -0.00000474, -0.00000078, -0.00002242, -0.00001251, 0.00001215, -0.00004692, 0.00001469, -0.00003946, -0.00003059, -0.00000383, -0.00008673, -0.00012560, 0.00008934, -0.00069754, -0.00039939, 0.00086435, 0.00217430, -0.00032586, -0.00093453, 0.00038275, -0.00072126, 0.00055655, -0.00051443, 0.00023864, -0.00006102, -0.00018245, 0.00026284, -0.00028810, 0.00019195, -0.00012381, 0.00002585, 0.00000132, -0.00001698, -0.00002127, 0.00003798, -0.00003958, 0.00002749, -0.00002438, -0.00000284, 0.00000736, -0.00001065, -0.00000825, 0.00001487, -0.00002333, 0.00000351, -0.00000009, -0.00000620, 0.00000384, -0.00000300, -0.00000145, 0.00000627, -0.00001763, -0.00000293, 0.00000207, -0.00000969, 0.00000750, -0.00000125, -0.00001277, 0.00001284, -0.00000299, -0.00002097, 0.00000945, 0.00000672, -0.00001139, 0.00000770, 0.00001263, -0.00000277, -0.00000285, 0.00000090, 0.00000220, 0.00000208, 0.00000074, 0.00000178, 0.00000068, -0.00000042, 0.00000021, -0.00000011, -0.00000070, -0.00000103, -0.00000105, 0.00000037, -0.00000023, 0.00000013, 0.00000052, -0.00000057, -0.00000096, 0.00000332, 0.00000379, },
		{ 0.00000406, 0.00000234, 0.00000132, -0.00000284, -0.00000513, 0.00000826, 0.00001109, -0.00000318, -0.00000062, 0.00000608, -0.00000698, -0.00000658, 0.00000470, -0.00000505, -0.00000648, 0.00000400, -0.00000083, -0.00000328, -0.00000120, -0.00000878, -0.00000426, 0.00000174, -0.00000016, -0.00000084, -0.00000126, -0.00000049, -0.00000124, -0.00001226, -0.00000497, 0.00000279, -0.00000635, 0.00000027, 0.00000273, -0.00000833, -0.00000246, -0.00000320, -0.00000616, 0.00000348, -0.00001374, -0.00000063, -0.00000129, -0.00001182, -0.00000309, -0.00000897, -0.00002971, 0.00002171, -0.00004146, -0.00000505, -0.00000912, -0.00004328, -0.00002787, -0.00001063, -0.00020855, 0.00013548, -0.00046039, -0.00076038, 0.00036843, 0.00279558, -0.00003094, -0.00141778, 0.00070071, -0.00090029, 0.00054131, -0.00035707, 0.00000605, 0.00015045, -0.00031694, 0.00030045, -0.00024625, 0.00011001, -0.00004802, -0.00001558, 0.00000246, 0.00000392, -0.00004363, 0.00004693, -0.00003488, 0.00001389, -0.00000771, -0.00001301, 0.00000495, -0.00000074, -0.00001816, 0.00001950, -0.00001717, -0.00000744, 0.00000518, -0.00000628, 0.00000016, 0.00000096, -0.00000586, 0.00000976, -0.00001184, -0.00001156, 0.00000501, -0.00000819, 0.00000083, 0.00000628, -0.00001389, 0.00000478, 0.00000805, -0.00002085, -0.00000208, 0.00001312, -0.00000841, -0.00000090, 0.00001493, 0.00000252, -0.00000452, -0.00000030, 0.00000211, 0.00000215, 0.00000122, 0.00000157, 0.00000098, -0.00000020, -0.00000019, 0.00000013, -0.00000040, -0.00000100, -0.00000117, 0.00000023, 0.00000015, -0.00000016, 0.00000009, 0.00000036, -0.00000137, 0.00000160, },
		{ 0.00000004, 0.00000356, 0.00000304, 0.00000190, -0.00000073, -0.00000627, 0.00000303, 0.00001323, 0.00000163, -0.00000441, 0.00000635, -0.00000130, -0.00001101, 0.00000233, 0.00000050, -0.00000921, 0.00000157, 0.00000285, -0.00000487, -0.00000023, -0.00000628, -0.00000839, 0.00000124, 0.00000079, -0.00000137, 0.00000060, -0.00000284, 0.00000037, -0.00000784, -0.00001093, 0.00000333, -0.00000364, -0.00000485, 0.00000729, -0.00000710, -0.00000782, 0.00000180, -0.00000960, 0.00000409, -0.00000933, -0.00000837, 0.00000588, -0.00000985, -0.00001510, 0.00000728, -0.00003882, 0.00001593, -0.00002237, -0.00002713, 0.00000348, -0.00002135, -0.00008231, 0.00005902, -0.00022746, 0.00006567, -0.00018749, -0.00085896, -0.00040086, 0.00322057, 0.00065079, -0.00204335, 0.00094228, -0.00090729, 0.00034751, -0.00008591, -0.00024760, 0.00030505, -0.00035974, 0.00025410, -0.00015708, 0.00002584, -0.00000240, -0.00001848, -0.00002162, 0.00002852, -0.00005402, 0.00003997, -0.00001957, -0.00000166, 0.00000235, -0.00001206, -0.00000552, 0.00000955, -0.00002203, 0.00001557, -0.00000599, -0.00001582, 0.00000417, -0.00000171, -0.00000485, 0.00000361, -0.00000713, 0.00000815, -0.00000331, -0.00001677, 0.00000231, -0.00000297, -0.00000560, 0.00000877, -0.00000862, -0.00000485, 0.00001283, -0.00001299, -0.00001327, 0.00001272, -0.00000071, -0.00000768, 0.00001221, 0.00000891, -0.00000405, -0.00000236, 0.00000191, 0.00000201, 0.00000170, 0.00000171, 0.00000096, 0.00000016, -0.00000043, 0.00000009, -0.00000004, -0.00000090, -0.00000121, 0.00000002, 0.00000032, -0.00000007, -0.00000050, 0.00000083, -0.00000096, },
		{ -0.00000023, -0.00000086, 0.00000241, 0.00000354, 0.00000231, 0.00000091, -0.00000531, -0.00000183, 0.00001175, 0.00000766, -0.00000550, 0.00000294, 0.00000491, -0.00001138, -0.00000370, 0.00000517, -0.00000814, -0.00000338, 0.00000597, -0.00000446, -0.00000167, -0.00000232, -0.00001085, -0.00000170, 0.00000260, -0.00000231, 0.00000223, -0.00000336, -0.00000133, -0.00000187, -0.00001383, -0.00000055, 0.00000136, -0.00000902, 0.00000754, -0.00000129, -0.00001438, 0.00000447, -0.00000881, 0.00000006, -0.00000237, -0.00001385, 0.00000528, 0.00000122, -0.00003010, 0.00001634, -0.00003347, -0.00000222, -0.00000248, -0.00003431, -0.00001206, 0.00002564, -0.00013408, 0.00008186, -0.00016982, -0.00005940, -0.00001051, -0.00064926, -0.00126435, 0.00323602, 0.00171056, -0.00264035, 0.00095620, -0.00068445, 0.00000660, 0.00021626, -0.00043750, 0.00035291, -0.00030718, 0.00015507, -0.00006477, -0.00002586, 0.00000109, 0.00000698, -0.00005016, 0.00003952, -0.00004650, 0.00002190, -0.00000350, -0.00001017, 0.00000183, -0.00000277, -0.00001678, 0.00001355, -0.00001770, 0.00000633, 0.00000391, -0.00001722, -0.00000268, 0.00000394, -0.00000751, 0.00000278, -0.00000508, 0.00000388, 0.00000365, -0.00001616, -0.00000437, 0.00000199, -0.00000816, 0.00000565, -0.00000064, -0.00001081, 0.00000997, -0.00000189, -0.00001892, 0.00000547, 0.00000753, -0.00000946, 0.00000543, 0.00001370, -0.00000067, -0.00000455, 0.00000116, 0.00000186, 0.00000189, 0.00000223, 0.00000088, 0.00000039, -0.00000040, -0.00000012, 0.00000023, -0.00000070, -0.00000120, -0.00000013, 0.00000026, 0.00000019, -0.00000089, 0.00000075, },
		{ 0.00000037, 0.00000034, -0.00000102, 0.00000107, 0.00000352, 0.00000277, 0.00000184, -0.00000326, -0.00000473, 0.00000731, 0.00001234, -0.00000268, -0.00000235, 0.00000852, -0.00000692, -0.00001064, 0.00000599, -0.00000318, -0.00000851, 0.00000634, -0.00000140, -0.00000452, 0.00000089, -0.00001010, -0.00000658, 0.00000388, -0.00000237, 0.00000238, -0.00000162, -0.00000475, 0.00000245, -0.00001189, -0.00000734, 0.00000521, -0.00000922, 0.00000303, 0.00000635, -0.00001815, 0.00000227, -0.00000392, -0.00000568, 0.00000224, -0.00001267, -0.00000333, 0.00001552, -0.00003888, 0.00001208, -0.00001634, -0.00002087, 0.00000326, -0.00001807, -0.00004829, 0.00007145, -0.00015032, 0.00004521, -0.00007081, -0.00014898, -0.00001566, -0.00020178, -0.00198198, 0.00271627, 0.00299813, -0.00298343, 0.00063974, -0.00026026, -0.00038358, 0.00044678, -0.00050236, 0.00028840, -0.00019481, 0.00005268, -0.00000775, -0.00003144, -0.00002696, 0.00003816, -0.00006382, 0.00003029, -0.00002676, 0.00000380, 0.00000413, -0.00000817, -0.00000646, 0.00000776, -0.00002197, 0.00000898, -0.00000828, -0.00000217, 0.00000761, -0.00001148, -0.00001131, 0.00000635, -0.00000591, -0.00000093, -0.00000192, 0.00000029, 0.00000636, -0.00001064, -0.00001118, 0.00000322, -0.00000606, -0.00000040, 0.00000542, -0.00001054, 0.00000211, 0.00000681, -0.00001683, -0.00000517, 0.00001183, -0.00000559, -0.00000240, 0.00001454, 0.00000498, -0.00000555, -0.00000045, 0.00000173, 0.00000175, 0.00000286, 0.00000106, 0.00000037, -0.00000021, -0.00000030, 0.00000031, -0.00000044, -0.00000114, -0.00000018, 0.00000008, 0.00000036, -0.00000097, },
		{ -0.00000085, 0.00000001, 0.00000053, -0.00000068, 0.00000004, 0.00000289, 0.00000324, 0.00000229, -0.00000120, -0.00000518, 0.00000177, 0.00001363, 0.00000324, -0.00000654, 0.00000760, 0.00000061, -0.00001481, 0.00000171, 0.00000340, -0.00001085, 0.00000296, 0.00000312, -0.00000657, 0.00000142, -0.00000620, -0.00001139, 0.00000289, -0.00000078, 0.00000106, 0.00000115, -0.00000732, 0.00000300, -0.00000603, -0.00001348, 0.00000471, -0.00000487, -0.00000370, 0.00001161, -0.00001638, -0.00000427, 0.00000182, -0.00000865, 0.00000084, -0.00000472, -0.00001493, 0.00002477, -0.00003556, -0.00000332, 0.00000168, -0.00002630, -0.00001234, 0.00001499, -0.00008360, 0.00008891, -0.00011907, -0.00002636, 0.00000524, -0.00012940, -0.00020362, 0.00031925, -0.00233747, 0.00168492, 0.00424856, -0.00287230, -0.00000402, 0.00024726, -0.00069490, 0.00052769, -0.00043235, 0.00015208, -0.00007926, -0.00000999, -0.00000135, -0.00000204, -0.00006200, 0.00005356, -0.00005487, 0.00000714, -0.00000731, -0.00000437, 0.00000029, 0.00000088, -0.00001550, 0.00001281, -0.00001857, -0.00000110, 0.00000023, -0.00000502, 0.00000430, -0.00000247, -0.00001666, 0.00000329, -0.00000117, -0.00000485, -0.00000035, -0.00000060, 0.00000500, -0.00000351, -0.00001442, -0.00000020, -0.00000145, -0.00000552, 0.00000669, -0.00000513, -0.00000593, 0.00000931, -0.00000873, -0.00001413, 0.00000981, 0.00000165, -0.00000769, 0.00001078, 0.00001091, -0.00000415, -0.00000267, 0.00000133, 0.00000147, 0.00000327, 0.00000164, 0.00000026, -0.00000008, -0.00000030, 0.00000023, -0.00000022, -0.00000103, -0.00000016, -0.00000009, 0.00000035, },
		{ 0.00000021, -0.00000073, -0.00000015, 0.00000040, -0.00000028, -0.00000042, 0.00000188, 0.00000343, 0.00000261, 0.00000020, -0.00000374, -0.00000276, 0.00001102, 0.00000984, -0.00000704, 0.00000260, 0.00000766, -0.00001357, -0.00000620, 0.00000794, -0.00000847, -0.00000298, 0.00000652, -0.00000581, -0.00000105, -0.00000109, -0.00001360, -0.00000118, 0.00000176, -0.00000040, 0.00000301, -0.00000726, 0.00000003, 0.00000057, -0.00001534, -0.00000089, 0.00000154, -0.00000846, 0.00001139, -0.00000954, -0.00001149, 0.00000394, -0.00000586, -0.00000628, 0.00000549, -0.00002223, 0.00002347, -0.00002173, -0.00002007, 0.00000784, -0.00001184, -0.00004303, 0.00004608, -0.00009527, 0.00006738, -0.00005857, -0.00008291, 0.00000021, 0.00001670, -0.00048835, 0.00071930, -0.00221958, 0.00032408, 0.00515981, -0.00221471, -0.00084648, 0.00066887, -0.00082314, 0.00044131, -0.00027327, 0.00001039, -0.00001056, -0.00001577, -0.00003341, 0.00003563, -0.00008032, 0.00004421, -0.00002992, -0.00001596, 0.00000068, 0.00000037, -0.00001067, 0.00000959, -0.00001892, 0.00000992, -0.00000939, -0.00001066, 0.00000287, -0.00000127, -0.00000263, 0.00000447, -0.00001570, -0.00000398, 0.00000334, -0.00000619, -0.00000161, 0.00000088, 0.00000190, 0.00000172, -0.00001268, -0.00000630, 0.00000221, -0.00000703, 0.00000337, 0.00000177, -0.00000993, 0.00000537, 0.00000091, -0.00001739, 0.00000236, 0.00000829, -0.00000811, 0.00000390, 0.00001459, -0.00000005, -0.00000459, 0.00000027, 0.00000121, 0.00000328, 0.00000245, 0.00000038, -0.00000017, -0.00000017, 0.00000015, -0.00000012, -0.00000087, -0.00000009, -0.00000020, },
		{ -0.00000023, 0.00000003, -0.00000074, -0.00000014, 0.00000021, -0.00000010, -0.00000038, 0.00000093, 0.00000315, 0.00000298, 0.00000090, -0.00000157, -0.00000492, 0.00000572, 0.00001424, -0.00000305, -0.00000383, 0.00001074, -0.00000690, -0.00001396, 0.00000732, -0.00000200, -0.00000841, 0.00000630, -0.00000187, -0.00000481, 0.00000239, -0.00001169, -0.00000728, 0.00000329, -0.00000032, 0.00000274, -0.00000470, -0.00000404, 0.00000430, -0.00001156, -0.00000917, 0.00000577, -0.00000808, 0.00000589, -0.00000143, -0.00001456, -0.00000023, 0.00000185, -0.00001478, 0.00001158, -0.00002044, 0.00001281, -0.00000608, -0.00002636, -0.00000430, 0.00001688, -0.00007167, 0.00005548, -0.00007437, 0.00002135, -0.00000903, -0.00007656, -0.00010428, 0.00023889, -0.00073558, 0.00085049, -0.00166537, -0.00107681, 0.00549366, -0.00107681, -0.00166537, 0.00085049, -0.00073558, 0.00023889, -0.00010428, -0.00007656, -0.00000903, 0.00002135, -0.00007437, 0.00005548, -0.00007167, 0.00001688, -0.00000430, -0.00002636, -0.00000608, 0.00001281, -0.00002044, 0.00001158, -0.00001478, 0.00000185, -0.00000023, -0.00001456, -0.00000143, 0.00000589, -0.00000808, 0.00000577, -0.00000917, -0.00001156, 0.00000430, -0.00000404, -0.00000470, 0.00000274, -0.00000032, 0.00000329, -0.00000728, -0.00001169, 0.00000239, -0.00000481, -0.00000187, 0.00000630, -0.00000841, -0.00000200, 0.00000732, -0.00001396, -0.00000690, 0.00001074, -0.00000383, -0.00000305, 0.00001424, 0.00000572, -0.00000492, -0.00000157, 0.00000090, 0.00000298, 0.00000315, 0.00000093, -0.00000038, -0.00000010, 0.00000021, -0.00000014, -0.00000074, 0.00000003, },
		{ 0.00000021, -0.00000020, -0.00000009, -0.00000087, -0.00000012, 0.00000015, -0.00000017, -0.00000017, 0.00000038, 0.00000245, 0.00000328, 0.00000121, 0.00000027, -0.00000459, -0.00000005, 0.00001459, 0.00000390, -0.00000811, 0.00000829, 0.00000236, -0.00001739, 0.00000091, 0.00000537, -0.00000993, 0.00000177, 0.00000337, -0.00000703, 0.00000221, -0.00000630, -0.00001268, 0.00000172, 0.00000190, 0.00000088, -0.00000161, -0.00000619, 0.00000334, -0.00000398, -0.00001570, 0.00000447, -0.00000263, -0.00000127, 0.00000287, -0.00001066, -0.00000939, 0.00000992, -0.00001892, 0.00000959, -0.00001067, 0.00000037, 0.00000068, -0.00001596, -0.00002992, 0.00004421, -0.00008032, 0.00003563, -0.00003341, -0.00001577, -0.00001056, 0.00001039, -0.00027327, 0.00044131, -0.00082314, 0.00066887, -0.00084648, -0.00221471, 0.00515981, 0.00032408, -0.00221958, 0.00071930, -0.00048835, 0.00001670, 0.00000021, -0.00008291, -0.00005857, 0.00006738, -0.00009527, 0.00004608, -0.00004303, -0.00001184, 0.00000784, -0.00002007, -0.00002173, 0.00002347, -0.00002223, 0.00000549, -0.00000628, -0.00000586, 0.00000394, -0.00001149, -0.00000954, 0.00001139, -0.00000846, 0.00000154, -0.00000089, -0.00001534, 0.00000057, 0.00000003, -0.00000726, 0.00000301, -0.00000040, 0.00000176, -0.00000118, -0.00001360, -0.00000109, -0.00000105, -0.00000581, 0.00000652, -0.00000298, -0.00000847, 0.00000794, -0.00000620, -0.00001357, 0.00000766, 0.00000260, -0.00000704, 0.00000984, 0.00001102, -0.00000276, -0.00000374, 0.00000020, 0.00000261, 0.00000343, 0.00000188, -0.00000042, -0.00000028, 0.00000040, -0.00000015, -0.00000073, },
		{ -0.00000085, 0.00000035, -0.00000009, -0.00000016, -0.00000103, -0.00000022, 0.00000023, -0.00000030, -0.00000008, 0.00000026, 0.00000164, 0.00000327, 0.00000147, 0.00000133, -0.00000267, -0.00000415, 0.00001091, 0.00001078, -0.00000769, 0.00000165, 0.00000981, -0.00001413, -0.00000873, 0.00000931, -0.00000593, -0.00000513, 0.00000669, -0.00000552, -0.00000145, -0.00000020, -0.00001442, -0.00000351, 0.00000500, -0.00000060, -0.00000035, -0.00000485, -0.00000117, 0.00000329, -0.00001666, -0.00000247, 0.00000430, -0.00000502, 0.00000023, -0.00000110, -0.00001857, 0.00001281, -0.00001550, 0.00000088, 0.00000029, -0.00000437, -0.00000731, 0.00000714, -0.00005487, 0.00005356, -0.00006200, -0.00000204, -0.00000135, -0.00000999, -0.00007926, 0.00015208, -0.00043235, 0.00052769, -0.00069490, 0.00024726, -0.00000402, -0.00287230, 0.00424856, 0.00168492, -0.00233747, 0.00031925, -0.00020362, -0.00012940, 0.00000524, -0.00002636, -0.00011907, 0.00008891, -0.00008360, 0.00001499, -0.00001234, -0.00002630, 0.00000168, -0.00000332, -0.00003556, 0.00002477, -0.00001493, -0.00000472, 0.00000084, -0.00000865, 0.00000182, -0.00000427, -0.00001638, 0.00001161, -0.00000370, -0.00000487, 0.00000471, -0.00001348, -0.00000603, 0.00000300, -0.00000732, 0.00000115, 0.00000106, -0.00000078, 0.00000289, -0.00001139, -0.00000620, 0.00000142, -0.00000657, 0.00000312, 0.00000296, -0.00001085, 0.00000340, 0.00000171, -0.00001481, 0.00000061, 0.00000760, -0.00000654, 0.00000324, 0.00001363, 0.00000177, -0.00000518, -0.00000120, 0.00000229, 0.00000324, 0.00000289, 0.00000004, -0.00000068, 0.00000053, 0.00000001, },
		{ 0.00000037, -0.00000097, 0.00000036, 0.00000008, -0.00000018, -0.00000114, -0.00000044, 0.00000031, -0.00000030, -0.00000021, 0.00000037, 0.00000106, 0.00000286, 0.00000175, 0.00000173, -0.00000045, -0.00000555, 0.00000498, 0.00001454, -0.00000240, -0.00000559, 0.00001183, -0.00000517, -0.00001683, 0.00000681, 0.00000211, -0.00001054, 0.00000542, -0.00000040, -0.00000606, 0.00000322, -0.00001118, -0.00001064, 0.00000636, 0.00000029, -0.00000192, -0.00000093, -0.00000591, 0.00000635, -0.00001131, -0.00001148, 0.00000761, -0.00000217, -0.00000828, 0.00000898, -0.00002197, 0.00000776, -0.00000646, -0.00000817, 0.00000413, 0.00000380, -0.00002676, 0.00003029, -0.00006382, 0.00003816, -0.00002696, -0.00003144, -0.00000775, 0.00005268, -0.00019481, 0.00028840, -0.00050236, 0.00044678, -0.00038358, -0.00026026, 0.00063974, -0.00298343, 0.00299813, 0.00271627, -0.00198198, -0.00020178, -0.00001566, -0.00014898, -0.00007081, 0.00004521, -0.00015032, 0.00007145, -0.00004829, -0.00001807, 0.00000326, -0.00002087, -0.00001634, 0.00001208, -0.00003888, 0.00001552, -0.00000333, -0.00001267, 0.00000224, -0.00000568, -0.00000392, 0.00000227, -0.00001815, 0.00000635, 0.00000303, -0.00000922, 0.00000521, -0.00000734, -0.00001189, 0.00000245, -0.00000475, -0.00000162, 0.00000238, -0.00000237, 0.00000388, -0.00000658, -0.00001010, 0.00000089, -0.00000452, -0.00000140, 0.00000634, -0.00000851, -0.00000318, 0.00000599, -0.00001064, -0.00000692, 0.00000852, -0.00000235, -0.00000268, 0.00001234, 0.00000731, -0.00000473, -0.00000326, 0.00000184, 0.00000277, 0.00000352, 0.00000107, -0.00000102, 0.00000034, },
		{ -0.00000023, 0.00000075, -0.00000089, 0.00000019, 0.00000026, -0.00000013, -0.00000120, -0.00000070, 0.00000023, -0.00000012, -0.00000040, 0.00000039, 0.00000088, 0.00000223, 0.00000189, 0.00000186, 0.00000116, -0.00000455, -0.00000067, 0.00001370, 0.00000543, -0.00000946, 0.00000753, 0.00000547, -0.00001892, -0.00000189, 0.00000997, -0.00001081, -0.00000064, 0.00000565, -0.00000816, 0.00000199, -0.00000437, -0.00001616, 0.00000365, 0.00000388, -0.00000508, 0.00000278, -0.00000751, 0.00000394, -0.00000268, -0.00001722, 0.00000391, 0.00000633, -0.00001770, 0.00001355, -0.00001678, -0.00000277, 0.00000183, -0.00001017, -0.00000350, 0.00002190, -0.00004650, 0.00003952, -0.00005016, 0.00000698, 0.00000109, -0.00002586, -0.00006477, 0.00015507, -0.00030718, 0.00035291, -0.00043750, 0.00021626, 0.00000660, -0.00068445, 0.00095620, -0.00264035, 0.00171056, 0.00323602, -0.00126435, -0.00064926, -0.00001051, -0.00005940, -0.00016982, 0.00008186, -0.00013408, 0.00002564, -0.00001206, -0.00003431, -0.00000248, -0.00000222, -0.00003347, 0.00001634, -0.00003010, 0.00000122, 0.00000528, -0.00001385, -0.00000237, 0.00000006, -0.00000881, 0.00000447, -0.00001438, -0.00000129, 0.00000754, -0.00000902, 0.00000136, -0.00000055, -0.00001383, -0.00000187, -0.00000133, -0.00000336, 0.00000223, -0.00000231, 0.00000260, -0.00000170, -0.00001085, -0.00000232, -0.00000167, -0.00000446, 0.00000597, -0.00000338, -0.00000814, 0.00000517, -0.00000370, -0.00001138, 0.00000491, 0.00000294, -0.00000550, 0.00000766, 0.00001175, -0.00000183, -0.00000531, 0.00000091, 0.00000231, 0.00000354, 0.00000241, -0.00000086, },
		{ 0.00000004, -0.00000096, 0.00000083, -0.00000050, -0.00000007, 0.00000032, 0.00000002, -0.00000121, -0.00000090, -0.00000004, 0.00000009, -0.00000043, 0.00000016, 0.00000096, 0.00000171, 0.00000170, 0.00000201, 0.00000191, -0.00000236, -0.00000405, 0.00000891, 0.00001221, -0.00000768, -0.00000071, 0.00001272, -0.00001327, -0.00001299, 0.00001283, -0.00000485, -0.00000862, 0.00000877, -0.00000560, -0.00000297, 0.00000231, -0.00001677, -0.00000331, 0.00000815, -0.00000713, 0.00000361, -0.00000485, -0.00000171, 0.00000417, -0.00001582, -0.00000599, 0.00001557, -0.00002203, 0.00000955, -0.00000552, -0.00001206, 0.00000235, -0.00000166, -0.00001957, 0.00003997, -0.00005402, 0.00002852, -0.00002162, -0.00001848, -0.00000240, 0.00002584, -0.00015708, 0.00025410, -0.00035974, 0.00030505, -0.00024760, -0.00008591, 0.00034751, -0.00090729, 0.00094228, -0.00204335, 0.00065079, 0.00322057, -0.00040086, -0.00085896, -0.00018749, 0.00006567, -0.00022746, 0.00005902, -0.00008231, -0.00002135, 0.00000348, -0.00002713, -0.00002237, 0.00001593, -0.00003882, 0.00000728, -0.00001510, -0.00000985, 0.00000588, -0.00000837, -0.00000933, 0.00000409, -0.00000960, 0.00000180, -0.00000782, -0.00000710, 0.00000729, -0.00000485, -0.00000364, 0.00000333, -0.00001093, -0.00000784, 0.00000037, -0.00000284, 0.00000060, -0.00000137, 0.00000079, 0.00000124, -0.00000839, -0.00000628, -0.00000023, -0.00000487, 0.00000285, 0.00000157, -0.00000921, 0.00000050, 0.00000233, -0.00001101, -0.00000130, 0.00000635, -0.00000441, 0.00000163, 0.00001323, 0.00000303, -0.00000627, -0.00000073, 0.00000190, 0.00000304, 0.00000356, },
		{ 0.00000406, 0.00000160, -0.00000137, 0.00000036, 0.00000009, -0.00000016, 0.00000015, 0.00000023, -0.00000117, -0.00000100, -0.00000040, 0.00000013, -0.00000019, -0.00000020, 0.00000098, 0.00000157, 0.00000122, 0.00000215, 0.00000211, -0.00000030, -0.00000452, 0.00000252, 0.00001493, -0.00000090, -0.00000841, 0.00001312, -0.00000208, -0.00002085, 0.00000805, 0.00000478, -0.00001389, 0.00000628, 0.00000083, -0.00000819, 0.00000501, -0.00001156, -0.00001184, 0.00000976, -0.00000586, 0.00000096, 0.00000016, -0.00000628, 0.00000518, -0.00000744, -0.00001717, 0.00001950, -0.00001816, -0.00000074, 0.00000495, -0.00001301, -0.00000771, 0.00001389, -0.00003488, 0.00004693, -0.00004363, 0.00000392, 0.00000246, -0.00001558, -0.00004802, 0.00011001, -0.00024625, 0.00030045, -0.00031694, 0.00015045, 0.00000605, -0.00035707, 0.00054131, -0.00090029, 0.00070071, -0.00141778, -0.00003094, 0.00279558, 0.00036843, -0.00076038, -0.00046039, 0.00013548, -0.00020855, -0.00001063, -0.00002787, -0.00004328, -0.00000912, -0.00000505, -0.00004146, 0.00002171, -0.00002971, -0.00000897, -0.00000309, -0.00001182, -0.00000129, -0.00000063, -0.00001374, 0.00000348, -0.00000616, -0.00000320, -0.00000246, -0.00000833, 0.00000273, 0.00000027, -0.00000635, 0.00000279, -0.00000497, -0.00001226, -0.00000124, -0.00000049, -0.00000126, -0.00000084, -0.00000016, 0.00000174, -0.00000426, -0.00000878, -0.00000120, -0.00000328, -0.00000083, 0.00000400, -0.00000648, -0.00000505, 0.00000470, -0.00000658, -0.00000698, 0.00000608, -0.00000062, -0.00000318, 0.00001109, 0.00000826, -0.00000513, -0.00000284, 0.00000132, 0.00000234, },
		{ 0.00000173, 0.00000379, 0.00000332, -0.00000096, -0.00000057, 0.00000052, 0.00000013, -0.00000023, 0.00000037, -0.00000105, -0.00000103, -0.00000070, -0.00000011, 0.00000021, -0.00000042, 0.00000068, 0.00000178, 0.00000074, 0.00000208, 0.00000220, 0.00000090, -0.00000285, -0.00000277, 0.00001263, 0.00000770, -0.00001139, 0.00000672, 0.00000945, -0.00002097, -0.00000299, 0.00001284, -0.00001277, -0.00000125, 0.00000750, -0.00000969, 0.00000207, -0.00000293, -0.00001763, 0.00000627, -0.00000145, -0.00000300, 0.00000384, -0.00000620, -0.00000009, 0.00000351, -0.00002333, 0.00001487, -0.00000825, -0.00001065, 0.00000736, -0.00000284, -0.00002438, 0.00002749, -0.00003958, 0.00003798, -0.00002127, -0.00001698, 0.00000132, 0.00002585, -0.00012381, 0.00019195, -0.00028810, 0.00026284, -0.00018245, -0.00006102, 0.00023864, -0.00051443, 0.00055655, -0.00072126, 0.00038275, -0.00093453, -0.00032586, 0.00217430, 0.00086435, -0.00039939, -0.00069754, 0.00008934, -0.00012560, -0.00008673, -0.00000383, -0.00003059, -0.00003946, 0.00001469, -0.00004692, 0.00001215, -0.00001251, -0.00002242, -0.00000078, -0.00000474, -0.00001117, 0.00000379, -0.00001273, -0.00000135, -0.00000120, -0.00000696, -0.00000093, -0.00000514, -0.00000318, 0.00000295, -0.00000520, -0.00000097, 0.00000076, -0.00001280, -0.00000574, 0.00000184, -0.00000184, -0.00000152, 0.00000001, 0.00000069, -0.00000051, -0.00000860, -0.00000396, -0.00000139, -0.00000311, 0.00000331, -0.00000190, -0.00000854, 0.00000284, -0.00000074, -0.00000950, 0.00000240, 0.00000345, -0.00000496, 0.00000626, 0.00001187, -0.00000168, -0.00000463, 0.00000019, },
		{ -0.00000162, 0.00000120, 0.00000299, 0.00000456, 0.00000043, -0.00000151, 0.00000038, 0.00000084, -0.00000059, 0.00000029, -0.00000082, -0.00000104, -0.00000087, -0.00000056, 0.00000050, -0.00000026, 0.00000008, 0.00000202, 0.00000059, 0.00000168, 0.00000234, 0.00000128, -0.00000049, -0.00000524, 0.00000670, 0.00001413, -0.00000797, -0.00000291, 0.00001580, -0.00001260, -0.00001517, 0.00001424, -0.00000496, -0.00000991, 0.00000995, -0.00000554, -0.00000464, 0.00000447, -0.00001720, -0.00000188, 0.00000316, -0.00000480, 0.00000329, -0.00000104, -0.00000807, 0.00001093, -0.00002071, 0.00000372, 0.00000111, -0.00001301, -0.00000124, 0.00001436, -0.00003850, 0.00003004, -0.00002990, 0.00001880, -0.00000246, -0.00001542, -0.00003474, 0.00009546, -0.00019836, 0.00023141, -0.00025439, 0.00014430, 0.00000153, -0.00025782, 0.00037910, -0.00052717, 0.00043162, -0.00047629, 0.00012210, -0.00066467, -0.00033862, 0.00157273, 0.00102233, 0.00008755, -0.00078125, -0.00007183, -0.00003104, -0.00012606, -0.00002364, 0.00000462, -0.00006675, 0.00001748, -0.00003586, -0.00000594, 0.00000210, -0.00002562, -0.00000847, 0.00000591, -0.00001732, 0.00000170, -0.00000714, -0.00000720, 0.00000182, -0.00000724, -0.00000319, -0.00000014, -0.00000701, 0.00000152, -0.00000119, -0.00000504, 0.00000350, -0.00000929, -0.00001088, 0.00000213, -0.00000057, -0.00000303, 0.00000054, -0.00000054, 0.00000152, -0.00000607, -0.00000684, -0.00000079, -0.00000335, 0.00000062, 0.00000202, -0.00000857, -0.00000162, 0.00000350, -0.00000806, -0.00000267, 0.00000546, -0.00000351, 0.00000080, 0.00001242, 0.00000325, -0.00000499, },
		{ -0.00000317, -0.00000369, 0.00000043, 0.00000208, 0.00000488, 0.00000245, -0.00000180, -0.00000051, 0.00000167, -0.00000060, -0.00000012, -0.00000055, -0.00000100, -0.00000091, -0.00000104, 0.00000045, 0.00000025, -0.00000052, 0.00000192, 0.00000089, 0.00000112, 0.00000240, 0.00000128, 0.00000132, -0.00000476, -0.00000001, 0.00001562, 0.00000025, -0.00001056, 0.00001402, 0.00000066, -0.00002216, 0.00000703, 0.00000573, -0.00001440, 0.00000578, 0.00000253, -0.00001066, 0.00000646, -0.00001030, -0.00001096, 0.00000391, -0.00000211, -0.00000140, 0.00000605, -0.00001340, 0.00001080, -0.00001071, -0.00000736, 0.00000275, -0.00000484, -0.00001703, 0.00002941, -0.00004106, 0.00001853, -0.00001175, 0.00000358, -0.00000401, 0.00001756, -0.00009680, 0.00016486, -0.00023475, 0.00020231, -0.00014755, -0.00001809, 0.00017339, -0.00037849, 0.00039848, -0.00042257, 0.00024796, -0.00026805, -0.00001130, -0.00058108, -0.00022335, 0.00113749, 0.00089649, 0.00053381, -0.00065891, -0.00028504, 0.00001220, -0.00010847, -0.00007371, 0.00003684, -0.00007309, 0.00000036, -0.00001639, -0.00002101, 0.00000602, -0.00001793, -0.00002043, 0.00001292, -0.00001595, -0.00000579, -0.00000083, -0.00001038, 0.00000105, -0.00000439, -0.00000694, 0.00000345, -0.00000686, -0.00000312, 0.00000292, -0.00000675, 0.00000248, -0.00000372, -0.00001397, -0.00000051, 0.00000175, -0.00000422, 0.00000049, -0.00000107, 0.00000171, -0.00000260, -0.00000824, -0.00000194, -0.00000225, -0.00000225, 0.00000361, -0.00000568, -0.00000606, 0.00000435, -0.00000388, -0.00000654, 0.00000442, -0.00000025, -0.00000310, 0.00000972, 0.00000797, },
		{ 0.00001068, 0.00000076, -0.00000512, -0.00000093, 0.00000134, 0.00000429, 0.00000439, -0.00000094, -0.00000186, 0.00000210, -0.00000003, -0.00000071, -0.00000045, -0.00000084, -0.00000091, -0.00000136, 0.00000002, 0.00000086, -0.00000075, 0.00000133, 0.00000143, 0.00000078, 0.00000217, 0.00000132, 0.00000212, -0.00000253, -0.00000481, 0.00001193, 0.00000925, -0.00001215, 0.00000538, 0.00001254, -0.00001984, -0.00000599, 0.00001313, -0.00001122, -0.00000348, 0.00000986, -0.00001156, 0.00000196, -0.00000050, -0.00001581, -0.00000167, 0.00000416, -0.00000696, 0.00001041, -0.00001245, 0.00000383, 0.00000056, -0.00001094, -0.00000640, 0.00001014, -0.00003108, 0.00003326, -0.00002927, -0.00000070, 0.00000135, 0.00000738, -0.00003381, 0.00007406, -0.00015933, 0.00020023, -0.00020862, 0.00010635, -0.00000077, -0.00016971, 0.00027950, -0.00039671, 0.00031701, -0.00026602, 0.00008900, -0.00015531, -0.00002158, -0.00059761, -0.00011540, 0.00091298, 0.00061841, 0.00080914, -0.00036282, -0.00045530, -0.00003403, -0.00004915, -0.00012224, 0.00004318, -0.00005451, -0.00002669, -0.00000174, -0.00002449, -0.00000151, -0.00000532, -0.00002885, 0.00001187, -0.00000811, -0.00001419, 0.00000200, -0.00000918, -0.00000271, -0.00000077, -0.00000935, 0.00000376, -0.00000330, -0.00000824, 0.00000447, -0.00000522, -0.00000085, 0.00000104, -0.00001343, -0.00000517, 0.00000338, -0.00000407, -0.00000044, -0.00000093, 0.00000091, 0.00000027, -0.00000755, -0.00000411, -0.00000116, -0.00000393, 0.00000274, -0.00000174, -0.00000829, 0.00000198, 0.00000057, -0.00000747, 0.00000106, 0.00000271, -0.00000422, 0.00000501, },
		{ 0.00000025, 0.00001034, 0.00000567, -0.00000485, -0.00000289, 0.00000070, 0.00000322, 0.00000552, 0.00000105, -0.00000298, 0.00000168, 0.00000101, -0.00000113, -0.00000069, -0.00000057, -0.00000089, -0.00000148, -0.00000060, 0.00000120, -0.00000045, 0.00000046, 0.00000174, 0.00000097, 0.00000163, 0.00000142, 0.00000221, -0.00000014, -0.00000640, 0.00000521, 0.00001485, -0.00000682, -0.00000536, 0.00001737, -0.00000873, -0.00001843, 0.00001213, -0.00000115, -0.00001275, 0.00001155, -0.00000596, -0.00000598, 0.00000656, -0.00001294, -0.00001225, 0.00000997, -0.00000905, 0.00000911, -0.00000585, -0.00000473, 0.00000581, -0.00000345, -0.00002273, 0.00002317, -0.00003455, 0.00002301, -0.00000950, -0.00001429, -0.00000441, 0.00003681, -0.00008452, 0.00013110, -0.00019149, 0.00017874, -0.00012085, -0.00002660, 0.00013663, -0.00026277, 0.00029646, -0.00032834, 0.00018855, -0.00012731, 0.00000531, -0.00013953, 0.00003609, -0.00062223, -0.00008689, 0.00085410, 0.00033323, 0.00086379, 0.00000825, -0.00050077, -0.00016081, 0.00000938, -0.00013884, 0.00001744, -0.00002300, -0.00004751, -0.00000153, -0.00001601, -0.00001409, 0.00000387, -0.00002874, 0.00000360, 0.00000127, -0.00001859, -0.00000047, -0.00000490, -0.00000671, 0.00000101, -0.00000892, 0.00000120, 0.00000127, -0.00001111, 0.00000252, -0.00000175, -0.00000393, 0.00000295, -0.00000963, -0.00000966, 0.00000281, -0.00000248, -0.00000169, -0.00000081, 0.00000013, 0.00000173, -0.00000530, -0.00000597, -0.00000108, -0.00000414, 0.00000047, 0.00000128, -0.00000762, -0.00000188, 0.00000306, -0.00000545, -0.00000268, 0.00000372, -0.00000283, },
		{ -0.00000038, -0.00000277, 0.00000721, 0.00000984, -0.00000237, -0.00000489, -0.00000023, 0.00000219, 0.00000547, 0.00000358, -0.00000307, 0.00000035, 0.00000201, -0.00000098, -0.00000131, -0.00000032, -0.00000081, -0.00000147, -0.00000117, 0.00000106, 0.00000025, -0.00000024, 0.00000146, 0.00000163, 0.00000110, 0.00000135, 0.00000213, 0.00000137, -0.00000522, -0.00000134, 0.00001478, 0.00000269, -0.00001246, 0.00001313, 0.00000605, -0.00002353, 0.00000186, 0.00001061, -0.00001620, 0.00000572, 0.00000325, -0.00001182, 0.00000654, -0.00000304, -0.00002287, 0.00001048, -0.00000542, 0.00000321, 0.00000171, -0.00000834, 0.00000116, 0.00001211, -0.00003748, 0.00002553, -0.00002460, 0.00000475, 0.00000504, -0.00000892, -0.00003457, 0.00008449, -0.00013521, 0.00016075, -0.00017212, 0.00010046, 0.00000173, -0.00015161, 0.00022120, -0.00027623, 0.00023782, -0.00021887, 0.00007241, -0.00005064, 0.00000098, -0.00017948, 0.00009497, -0.00059835, -0.00013941, 0.00087070, 0.00014306, 0.00073410, 0.00033471, -0.00039054, -0.00031439, 0.00002117, -0.00011331, -0.00002637, 0.00000137, -0.00005025, -0.00001571, -0.00000308, -0.00002287, 0.00000428, -0.00002077, -0.00000671, 0.00000668, -0.00001677, -0.00000662, -0.00000080, -0.00000825, -0.00000017, -0.00000626, -0.00000218, 0.00000428, -0.00001052, -0.00000175, 0.00000125, -0.00000479, 0.00000180, -0.00000461, -0.00001180, -0.00000022, -0.00000052, -0.00000226, -0.00000127, -0.00000018, 0.00000196, -0.00000270, -0.00000652, -0.00000209, -0.00000353, -0.00000181, 0.00000232, -0.00000489, -0.00000505, 0.00000269, -0.00000194, -0.00000487, 0.00000244, },
		{ -0.00000010, 0.00000143, -0.00000340, 0.00000269, 0.00001170, 0.00000194, -0.00000593, -0.00000180, 0.00000145, 0.00000444, 0.00000571, -0.00000171, -0.00000135, 0.00000233, -0.00000010, -0.00000204, -0.00000033, -0.00000059, -0.00000143, -0.00000150, 0.00000050, 0.00000092, -0.00000036, 0.00000059, 0.00000236, 0.00000102, 0.00000091, 0.00000219, 0.00000187, -0.00000285, -0.00000521, 0.00000967, 0.00001170, -0.00001230, 0.00000249, 0.00001734, -0.00001794, -0.00001331, 0.00001709, -0.00001085, -0.00000485, 0.00001031, -0.00001094, -0.00000073, 0.00000894, -0.00002755, 0.00000342, 0.00000205, -0.00000261, 0.00000446, -0.00000341, -0.00001116, 0.00002748, -0.00004174, 0.00001435, -0.00000711, -0.00000881, 0.00000134, 0.00002045, -0.00008140, 0.00013078, -0.00016062, 0.00014371, -0.00010075, -0.00001063, 0.00011803, -0.00022867, 0.00023359, -0.00022209, 0.00014334, -0.00011926, 0.00000730, -0.00004157, 0.00004364, -0.00022174, 0.00011122, -0.00051911, -0.00022841, 0.00087867, 0.00007956, 0.00051245, 0.00052625, -0.00015750, -0.00042142, -0.00003668, -0.00006258, -0.00006282, 0.00000314, -0.00003475, -0.00003490, 0.00000419, -0.00002232, -0.00000346, -0.00001031, -0.00001309, 0.00000541, -0.00001037, -0.00001259, 0.00000008, -0.00000649, -0.00000330, -0.00000338, -0.00000419, 0.00000450, -0.00000732, -0.00000593, 0.00000179, -0.00000311, -0.00000087, -0.00000077, -0.00001066, -0.00000452, 0.00000029, -0.00000156, -0.00000225, -0.00000026, 0.00000168, -0.00000077, -0.00000562, -0.00000348, -0.00000303, -0.00000321, 0.00000150, -0.00000173, -0.00000610, 0.00000012, 0.00000099, -0.00000464, },
		{ -0.00000254, -0.00000226, 0.00000160, -0.00000221, -0.00000140, 0.00001063, 0.00000672, -0.00000506, -0.00000389, 0.00000080, 0.00000304, 0.00000666, 0.00000087, -0.00000248, 0.00000151, 0.00000123, -0.00000241, -0.00000081, -0.00000027, -0.00000138, -0.00000159, -0.00000021, 0.00000116, 0.00000020, -0.00000049, 0.00000259, 0.00000161, 0.00000024, 0.00000224, 0.00000188, -0.00000085, -0.00000568, 0.00000239, 0.00001611, -0.00000515, -0.00000865, 0.00001976, -0.00000378, -0.00002577, 0.00001355, 0.00000113, -0.00001387, 0.00001007, -0.00000275, -0.00001080, 0.00001639, -0.00002299, -0.00000875, 0.00000791, -0.00000301, -0.00000049, 0.00000803, -0.00002381, 0.00003386, -0.00003218, -0.00000459, 0.00000552, -0.00000512, -0.00002554, 0.00006606, -0.00012589, 0.00015253, -0.00014353, 0.00007937, 0.00000126, -0.00011709, 0.00019098, -0.00023957, 0.00018448, -0.00013605, 0.00005790, -0.00006319, -0.00000078, -0.00007409, 0.00008895, -0.00022948, 0.00007855, -0.00041454, -0.00030096, 0.00083434, 0.00011173, 0.00029967, 0.00055383, 0.00011845, -0.00042405, -0.00014950, -0.00002178, -0.00007094, -0.00001887, -0.00001249, -0.00004669, -0.00000020, -0.00001361, -0.00001366, -0.00000373, -0.00001242, -0.00000102, -0.00000366, -0.00001479, -0.00000311, -0.00000285, -0.00000614, -0.00000219, -0.00000391, 0.00000255, -0.00000361, -0.00000790, -0.00000051, -0.00000028, -0.00000292, 0.00000047, -0.00000699, -0.00000804, -0.00000105, 0.00000008, -0.00000308, -0.00000063, 0.00000149, 0.00000017, -0.00000389, -0.00000437, -0.00000314, -0.00000367, -0.00000025, 0.00000044, -0.00000490, -0.00000296, 0.00000192, },
		{ 0.00000069, -0.00000008, -0.00000284, 0.00000035, -0.00000056, -0.00000376, 0.00000725, 0.00001028, -0.00000200, -0.00000578, -0.00000022, 0.00000186, 0.00000622, 0.00000377, -0.00000219, -0.00000028, 0.00000229, -0.00000197, -0.00000168, -0.00000007, -0.00000122, -0.00000158, -0.00000076, 0.00000080, 0.00000110, -0.00000118, 0.00000201, 0.00000263, -0.00000019, 0.00000196, 0.00000196, 0.00000005, -0.00000378, -0.00000371, 0.00001441, 0.00000515, -0.00001437, 0.00001257, 0.00001231, -0.00002873, 0.00000055, 0.00001311, -0.00001518, 0.00000150, 0.00000843, -0.00001708, 0.00001506, -0.00001112, -0.00001959, 0.00000647, 0.00000410, -0.00001079, 0.00001932, -0.00002873, 0.00002755, -0.00001395, -0.00001904, 0.00000112, 0.00002053, -0.00006814, 0.00010942, -0.00014611, 0.00013441, -0.00008380, -0.00001306, 0.00009996, -0.00018462, 0.00020317, -0.00019366, 0.00010629, -0.00005851, 0.00001119, -0.00005610, 0.00002685, -0.00011086, 0.00010504, -0.00019615, 0.00002072, -0.00032443, -0.00032616, 0.00074113, 0.00017806, 0.00016351, 0.00045160, 0.00034686, -0.00030718, -0.00027017, -0.00002422, -0.00004761, -0.00005087, 0.00000088, -0.00004365, -0.00001434, -0.00000338, -0.00001964, -0.00000436, -0.00000614, -0.00000811, -0.00000074, -0.00001210, -0.00000855, -0.00000002, -0.00000681, -0.00000326, -0.00000206, 0.00000017, -0.00000142, -0.00000697, -0.00000436, 0.00000166, -0.00000299, -0.00000070, -0.00000275, -0.00000911, -0.00000433, 0.00000150, -0.00000300, -0.00000152, 0.00000151, 0.00000042, -0.00000221, -0.00000427, -0.00000369, -0.00000369, -0.00000200, 0.00000104, -0.00000241, -0.00000496, },
		{ -0.00000512, -0.00000171, 0.00000125, -0.00000171, -0.00000124, 0.00000031, -0.00000416, 0.00000306, 0.00001133, 0.00000254, -0.00000642, -0.00000190, 0.00000106, 0.00000481, 0.00000588, -0.00000018, -0.00000223, 0.00000233, -0.00000070, -0.00000252, -0.00000027, -0.00000089, -0.00000155, -0.00000100, 0.00000002, 0.00000182, -0.00000108, 0.00000079, 0.00000347, 0.00000008, 0.00000124, 0.00000222, 0.00000007, -0.00000131, -0.00000654, 0.00000819, 0.00001343, -0.00001187, -0.00000001, 0.00002261, -0.00002007, -0.00001581, 0.00001755, -0.00000669, -0.00001112, 0.00001576, -0.00001494, 0.00000600, 0.00000146, -0.00002223, -0.00000424, 0.00001555, -0.00001994, 0.00002326, -0.00002244, 0.00001287, 0.00000124, -0.00001681, -0.00002484, 0.00006089, -0.00010878, 0.00012953, -0.00012801, 0.00007709, 0.00000079, -0.00010293, 0.00016389, -0.00019655, 0.00016249, -0.00012051, 0.00003602, -0.00001625, 0.00000786, -0.00007955, 0.00005857, -0.00012418, 0.00008510, -0.00014072, -0.00002926, -0.00027366, -0.00030590, 0.00063235, 0.00022347, 0.00011981, 0.00029200, 0.00046463, -0.00010480, -0.00034134, -0.00008165, -0.00001042, -0.00007266, -0.00000459, -0.00002802, -0.00002988, 0.00000088, -0.00001816, -0.00001072, 0.00000108, -0.00001140, -0.00000307, -0.00000642, -0.00001307, -0.00000007, -0.00000503, -0.00000556, -0.00000027, -0.00000097, -0.00000151, -0.00000412, -0.00000767, 0.00000133, -0.00000117, -0.00000277, 0.00000012, -0.00000733, -0.00000816, 0.00000145, -0.00000182, -0.00000259, 0.00000140, 0.00000052, -0.00000111, -0.00000340, -0.00000414, -0.00000374, -0.00000325, 0.00000034, 0.00000009, },
		{ 0.00000170, -0.00000373, -0.00000386, 0.00000080, 0.00000020, -0.00000195, 0.00000004, -0.00000339, -0.00000036, 0.00000966, 0.00000706, -0.00000497, -0.00000399, 0.00000033, 0.00000319, 0.00000646, 0.00000299, -0.00000321, 0.00000101, 0.00000090, -0.00000274, -0.00000104, -0.00000052, -0.00000149, -0.00000101, -0.00000081, 0.00000195, -0.00000026, -0.00000046, 0.00000351, 0.00000112, 0.00000042, 0.00000235, -0.00000012, 0.00000034, -0.00000608, 0.00000091, 0.00001602, -0.00000277, -0.00001115, 0.00002248, -0.00000376, -0.00002693, 0.00001046, 0.00000780, -0.00002033, 0.00001406, -0.00000498, -0.00000466, 0.00000715, -0.00001374, -0.00002034, 0.00002417, -0.00002135, 0.00001706, -0.00000892, 0.00000047, 0.00000145, 0.00000690, -0.00006519, 0.00009890, -0.00012746, 0.00011306, -0.00007276, -0.00000228, 0.00008227, -0.00016176, 0.00017730, -0.00015934, 0.00009582, -0.00005434, -0.00000166, -0.00001336, 0.00003127, -0.00010597, 0.00007059, -0.00010732, 0.00004434, -0.00009101, -0.00004957, -0.00026174, -0.00026552, 0.00054427, 0.00022325, 0.00014046, 0.00014908, 0.00045785, 0.00011647, -0.00032302, -0.00017564, 0.00001175, -0.00007044, -0.00002681, -0.00001021, -0.00003756, -0.00000415, -0.00001125, -0.00001793, 0.00000453, -0.00000915, -0.00000881, -0.00000131, -0.00001414, -0.00000315, -0.00000225, -0.00000734, 0.00000010, -0.00000035, -0.00000317, -0.00000119, -0.00000886, -0.00000125, 0.00000118, -0.00000405, 0.00000072, -0.00000371, -0.00001076, -0.00000060, -0.00000018, -0.00000314, 0.00000083, 0.00000071, -0.00000055, -0.00000233, -0.00000405, -0.00000391, -0.00000403, -0.00000102, },
		{ -0.00000249, 0.00000219, -0.00000171, -0.00000476, -0.00000100, 0.00000155, -0.00000129, -0.00000081, -0.00000261, -0.00000215, 0.00000618, 0.00000993, -0.00000143, -0.00000563, -0.00000086, 0.00000199, 0.00000550, 0.00000616, -0.00000236, -0.00000127, 0.00000192, -0.00000195, -0.00000215, -0.00000038, -0.00000128, -0.00000093, -0.00000140, 0.00000141, 0.00000081, -0.00000107, 0.00000257, 0.00000247, 0.00000009, 0.00000199, -0.00000008, 0.00000085, -0.00000386, -0.00000420, 0.00001242, 0.00000806, -0.00001518, 0.00001284, 0.00001216, -0.00002642, -0.00000581, 0.00002020, -0.00001988, 0.00000352, 0.00000680, -0.00000923, 0.00000199, 0.00000240, -0.00003360, 0.00002316, -0.00001308, 0.00000489, 0.00000205, 0.00000151, -0.00001845, 0.00004552, -0.00010324, 0.00011543, -0.00011165, 0.00006133, 0.00000297, -0.00007751, 0.00013494, -0.00017523, 0.00014540, -0.00009710, 0.00003477, -0.00001830, -0.00000287, -0.00003453, 0.00005667, -0.00011474, 0.00005707, -0.00007248, 0.00000629, -0.00006634, -0.00003921, -0.00026869, -0.00023435, 0.00049520, 0.00018597, 0.00017859, 0.00006835, 0.00035974, 0.00028744, -0.00021165, -0.00026489, -0.00000457, -0.00004589, -0.00005294, -0.00000207, -0.00003328, -0.00001554, -0.00000454, -0.00002116, 0.00000233, -0.00000318, -0.00001421, 0.00000031, -0.00001145, -0.00000748, -0.00000052, -0.00000735, -0.00000117, 0.00000132, -0.00000493, 0.00000031, -0.00000765, -0.00000475, 0.00000242, -0.00000366, -0.00000048, -0.00000002, -0.00001093, -0.00000407, 0.00000077, -0.00000268, -0.00000012, 0.00000080, -0.00000017, -0.00000156, -0.00000345, -0.00000393, -0.00000457, },
		{ -0.00000490, -0.00000388, 0.00000184, 0.00000008, -0.00000426, -0.00000302, 0.00000149, 0.00000022, -0.00000121, -0.00000253, -0.00000247, 0.00000241, 0.00001019, 0.00000320, -0.00000570, -0.00000274, 0.00000126, 0.00000367, 0.00000810, 0.00000037, -0.00000338, 0.00000157, -0.00000028, -0.00000303, -0.00000079, -0.00000090, -0.00000083, -0.00000165, 0.00000051, 0.00000152, -0.00000074, 0.00000103, 0.00000335, 0.00000069, 0.00000120, 0.00000006, 0.00000078, -0.00000174, -0.00000569, 0.00000518, 0.00001532, -0.00001052, -0.00000050, 0.00001982, -0.00001384, -0.00002337, 0.00002268, -0.00000895, -0.00000983, 0.00001243, -0.00000330, -0.00001143, 0.00001769, -0.00003608, 0.00001091, -0.00000013, -0.00000385, 0.00000034, 0.00002082, -0.00005273, 0.00008305, -0.00012026, 0.00009870, -0.00006286, -0.00000973, 0.00007430, -0.00012498, 0.00014624, -0.00014790, 0.00008960, -0.00003885, 0.00000129, -0.00001643, 0.00001824, -0.00005699, 0.00006562, -0.00010102, 0.00002852, -0.00003941, -0.00001203, -0.00006866, -0.00001405, -0.00027023, -0.00022865, 0.00047957, 0.00013913, 0.00019589, 0.00005534, 0.00022885, 0.00036487, -0.00004144, -0.00030480, -0.00006307, -0.00001611, -0.00006744, -0.00000936, -0.00002065, -0.00002616, -0.00000334, -0.00001877, -0.00000364, 0.00000262, -0.00001608, -0.00000213, -0.00000693, -0.00001053, -0.00000117, -0.00000570, -0.00000313, 0.00000273, -0.00000559, -0.00000005, -0.00000507, -0.00000744, 0.00000165, -0.00000199, -0.00000213, 0.00000223, -0.00000869, -0.00000751, 0.00000009, -0.00000137, -0.00000090, 0.00000047, 0.00000030, -0.00000114, -0.00000275, -0.00000352, },
		{ -0.00000275, -0.00000480, -0.00000518, 0.00000100, 0.00000121, -0.00000292, -0.00000423, 0.00000009, 0.00000148, -0.00000049, -0.00000299, -0.00000221, -0.00000036, 0.00000801, 0.00000730, -0.00000357, -0.00000489, 0.00000052, 0.00000189, 0.00000819, 0.00000409, -0.00000404, -0.00000031, 0.00000150, -0.00000303, -0.00000181, -0.00000055, -0.00000065, -0.00000168, -0.00000035, 0.00000149, 0.00000032, -0.00000032, 0.00000316, 0.00000207, 0.00000056, -0.00000017, 0.00000083, -0.00000072, -0.00000427, -0.00000191, 0.00001612, -0.00000026, -0.00001013, 0.00001585, 0.00000483, -0.00003292, 0.00001249, 0.00000670, -0.00001730, 0.00000694, 0.00001101, -0.00002499, 0.00002335, -0.00002532, -0.00000681, 0.00000763, 0.00000012, -0.00001814, 0.00005314, -0.00008659, 0.00010130, -0.00010470, 0.00005052, 0.00000354, -0.00007657, 0.00011906, -0.00013354, 0.00012082, -0.00009893, 0.00003608, -0.00000526, -0.00000017, -0.00003588, 0.00004046, -0.00006382, 0.00005409, -0.00007446, 0.00000196, -0.00002258, -0.00000807, -0.00008524, 0.00000552, -0.00025208, -0.00024571, 0.00047646, 0.00010908, 0.00017855, 0.00008437, 0.00012022, 0.00034564, 0.00013172, -0.00026897, -0.00014351, -0.00000423, -0.00006209, -0.00002798, -0.00000870, -0.00002942, -0.00000908, -0.00001312, -0.00000938, 0.00000491, -0.00001367, -0.00000671, -0.00000342, -0.00001066, -0.00000392, -0.00000365, -0.00000446, 0.00000298, -0.00000487, -0.00000154, -0.00000265, -0.00000820, -0.00000080, -0.00000027, -0.00000294, 0.00000258, -0.00000517, -0.00000943, -0.00000236, -0.00000006, -0.00000088, -0.00000041, 0.00000076, -0.00000078, -0.00000234, },
		{ -0.00000231, -0.00000203, -0.00000406, -0.00000629, -0.00000021, 0.00000170, -0.00000145, -0.00000426, -0.00000176, 0.00000148, 0.00000113, -0.00000320, -0.00000222, -0.00000171, 0.00000455, 0.00000937, 0.00000049, -0.00000628, -0.00000081, 0.00000071, 0.00000667, 0.00000739, -0.00000252, -0.00000298, 0.00000233, -0.00000185, -0.00000307, -0.00000060, -0.00000028, -0.00000165, -0.00000088, 0.00000080, 0.00000146, -0.00000081, 0.00000191, 0.00000351, 0.00000075, -0.00000105, 0.00000117, -0.00000068, -0.00000184, -0.00000602, 0.00001118, 0.00000989, -0.00001109, 0.00000312, 0.00002033, -0.00002899, -0.00000586, 0.00001751, -0.00001283, -0.00000792, 0.00002572, -0.00002976, 0.00001586, -0.00000684, -0.00001905, 0.00000099, 0.00001987, -0.00004804, 0.00008476, -0.00010293, 0.00008880, -0.00005820, -0.00001399, 0.00006522, -0.00011848, 0.00012705, -0.00010809, 0.00007645, -0.00005173, 0.00000379, -0.00000127, 0.00001911, -0.00005685, 0.00004779, -0.00005138, 0.00003120, -0.00005037, -0.00001027, -0.00002418, 0.00000807, -0.00009880, 0.00000684, -0.00021569, -0.00026965, 0.00046461, 0.00010671, 0.00013697, 0.00011921, 0.00006376, 0.00026166, 0.00025500, -0.00016183, -0.00021038, -0.00002580, -0.00004165, -0.00004659, -0.00000602, -0.00002358, -0.00001843, -0.00000877, -0.00001155, 0.00000287, -0.00000890, -0.00001043, -0.00000288, -0.00000807, -0.00000714, -0.00000275, -0.00000445, 0.00000207, -0.00000345, -0.00000299, -0.00000152, -0.00000703, -0.00000365, 0.00000024, -0.00000245, 0.00000162, -0.00000192, -0.00000904, -0.00000570, 0.00000017, 0.00000017, -0.00000146, 0.00000083, -0.00000017, },
		{ 0.00000068, -0.00000230, -0.00000183, -0.00000275, -0.00000682, -0.00000175, 0.00000171, -0.00000028, -0.00000346, -0.00000302, 0.00000007, 0.00000269, -0.00000237, -0.00000268, -0.00000211, 0.00000126, 0.00000881, 0.00000520, -0.00000582, -0.00000283, 0.00000003, 0.00000449, 0.00000900, 0.00000094, -0.00000506, 0.00000147, 0.00000015, -0.00000383, -0.00000135, 0.00000024, -0.00000159, -0.00000107, -0.00000018, 0.00000207, -0.00000026, 0.00000023, 0.00000410, 0.00000207, -0.00000225, 0.00000140, -0.00000083, -0.00000005, -0.00000656, 0.00000388, 0.00001491, -0.00000342, -0.00001108, 0.00002522, -0.00001332, -0.00002266, 0.00001571, 0.00000274, -0.00002405, 0.00003158, -0.00002172, -0.00000013, 0.00000808, -0.00001564, -0.00002256, 0.00004927, -0.00007631, 0.00009992, -0.00009086, 0.00004665, 0.00000467, -0.00007317, 0.00010280, -0.00012616, 0.00010311, -0.00006549, 0.00003447, -0.00002319, -0.00000227, -0.00001737, 0.00004088, -0.00006397, 0.00003659, -0.00002834, 0.00001114, -0.00003961, -0.00000661, -0.00003563, 0.00002208, -0.00009803, -0.00000962, -0.00017400, -0.00028337, 0.00043418, 0.00012508, 0.00009332, 0.00013335, 0.00005823, 0.00016061, 0.00029990, -0.00001643, -0.00022975, -0.00007875, -0.00002189, -0.00005374, -0.00001525, -0.00001287, -0.00002561, -0.00000926, -0.00000961, -0.00000156, -0.00000487, -0.00001115, -0.00000530, -0.00000454, -0.00000894, -0.00000371, -0.00000337, 0.00000078, -0.00000229, -0.00000352, -0.00000185, -0.00000494, -0.00000553, -0.00000093, -0.00000121, 0.00000052, -0.00000002, -0.00000670, -0.00000852, -0.00000132, 0.00000177, -0.00000202, 0.00000021, },
		{ -0.00000097, 0.00000133, -0.00000184, -0.00000230, -0.00000139, -0.00000639, -0.00000346, 0.00000122, 0.00000058, -0.00000246, -0.00000314, -0.00000203, 0.00000303, -0.00000043, -0.00000302, -0.00000241, -0.00000091, 0.00000617, 0.00000886, -0.00000308, -0.00000493, -0.00000077, 0.00000264, 0.00000850, 0.00000507, -0.00000521, -0.00000097, 0.00000193, -0.00000343, -0.00000271, 0.00000060, -0.00000137, -0.00000110, -0.00000100, 0.00000188, 0.00000090, -0.00000096, 0.00000337, 0.00000408, -0.00000298, 0.00000096, -0.00000062, 0.00000061, -0.00000496, -0.00000195, 0.00001310, 0.00000828, -0.00001920, 0.00001794, 0.00000596, -0.00002823, 0.00000023, 0.00002136, -0.00003146, 0.00002405, -0.00000502, -0.00001369, 0.00000840, 0.00000658, -0.00005597, 0.00007504, -0.00008843, 0.00008877, -0.00005123, -0.00001163, 0.00006294, -0.00010844, 0.00010801, -0.00010448, 0.00006314, -0.00002562, 0.00000999, -0.00001768, 0.00000991, -0.00003730, 0.00005046, -0.00005323, 0.00001486, -0.00000815, 0.00000391, -0.00004319, 0.00000500, -0.00004476, 0.00002447, -0.00008256, -0.00003325, -0.00014168, -0.00027866, 0.00038970, 0.00014739, 0.00006621, 0.00012020, 0.00008053, 0.00008334, 0.00027045, 0.00012078, -0.00018571, -0.00014259, -0.00002068, -0.00004541, -0.00003137, -0.00000472, -0.00002624, -0.00001471, -0.00000594, -0.00000523, -0.00000384, -0.00000876, -0.00000887, -0.00000225, -0.00000835, -0.00000601, -0.00000230, 0.00000002, -0.00000199, -0.00000303, -0.00000298, -0.00000316, -0.00000569, -0.00000325, -0.00000028, 0.00000016, 0.00000034, -0.00000355, -0.00000962, -0.00000426, 0.00000292, -0.00000154, },
		{ -0.00000011, -0.00000208, 0.00000127, -0.00000077, -0.00000305, -0.00000071, -0.00000491, -0.00000486, 0.00000004, 0.00000126, -0.00000169, -0.00000237, -0.00000368, 0.00000169, 0.00000178, -0.00000234, -0.00000311, -0.00000195, 0.00000277, 0.00001019, 0.00000134, -0.00000594, -0.00000229, 0.00000156, 0.00000648, 0.00000827, -0.00000290, -0.00000390, 0.00000235, -0.00000173, -0.00000412, 0.00000039, -0.00000092, -0.00000108, -0.00000146, 0.00000108, 0.00000192, -0.00000107, 0.00000158, 0.00000580, -0.00000251, -0.00000020, -0.00000026, 0.00000077, -0.00000321, -0.00000425, 0.00000646, 0.00001784, -0.00001759, 0.00000358, 0.00001869, -0.00001863, -0.00002191, 0.00003200, -0.00002451, 0.00000694, 0.00000957, -0.00001362, -0.00000999, 0.00004064, -0.00008502, 0.00008314, -0.00007593, 0.00005243, 0.00000349, -0.00006628, 0.00009854, -0.00011203, 0.00008603, -0.00006843, 0.00002560, -0.00000236, 0.00000661, -0.00002798, 0.00002617, -0.00004751, 0.00004325, -0.00003160, -0.00000458, -0.00000044, 0.00001027, -0.00005352, 0.00001373, -0.00004373, 0.00001495, -0.00006073, -0.00005140, -0.00012637, -0.00025918, 0.00034422, 0.00015867, 0.00006114, 0.00009070, 0.00010265, 0.00004874, 0.00019630, 0.00021016, -0.00008796, -0.00018751, -0.00004724, -0.00002812, -0.00004485, -0.00000496, -0.00002022, -0.00002185, -0.00000405, -0.00000597, -0.00000594, -0.00000509, -0.00001131, -0.00000238, -0.00000589, -0.00000827, -0.00000226, 0.00000017, -0.00000246, -0.00000207, -0.00000399, -0.00000251, -0.00000435, -0.00000558, -0.00000054, 0.00000069, -0.00000016, -0.00000089, -0.00000862, -0.00000761, 0.00000269, },
		{ 0.00000077, 0.00000154, -0.00000237, 0.00000031, 0.00000056, -0.00000340, -0.00000112, -0.00000292, -0.00000522, -0.00000187, 0.00000168, -0.00000111, -0.00000139, -0.00000408, -0.00000082, 0.00000294, -0.00000024, -0.00000388, -0.00000255, -0.00000007, 0.00000895, 0.00000592, -0.00000484, -0.00000447, 0.00000094, 0.00000407, 0.00000935, 0.00000124, -0.00000571, 0.00000082, 0.00000059, -0.00000471, -0.00000063, -0.00000038, -0.00000096, -0.00000164, 0.00000016, 0.00000222, -0.00000012, -0.00000038, 0.00000631, -0.00000082, -0.00000139, -0.00000045, 0.00000124, -0.00000250, -0.00000345, -0.00000097, 0.00002104, -0.00000826, -0.00000926, 0.00001842, 0.00000182, -0.00003910, 0.00002718, -0.00000624, -0.00000883, 0.00001060, 0.00000467, -0.00004091, 0.00007192, -0.00009500, 0.00006615, -0.00004095, 0.00000269, 0.00005503, -0.00010016, 0.00010377, -0.00008920, 0.00005149, -0.00003513, 0.00000320, 0.00000114, 0.00001742, -0.00004148, 0.00003459, -0.00004322, 0.00002485, -0.00001106, -0.00001219, -0.00000608, 0.00002286, -0.00006050, 0.00001284, -0.00003270, 0.00000054, -0.00004288, -0.00005691, -0.00012566, -0.00023618, 0.00030998, 0.00015382, 0.00007046, 0.00006254, 0.00010651, 0.00005163, 0.00011598, 0.00023419, 0.00003268, -0.00018858, -0.00009566, -0.00001565, -0.00004777, -0.00001406, -0.00001179, -0.00002621, -0.00000607, -0.00000395, -0.00000933, -0.00000251, -0.00001129, -0.00000456, -0.00000310, -0.00000916, -0.00000353, 0.00000087, -0.00000309, -0.00000133, -0.00000425, -0.00000301, -0.00000242, -0.00000685, -0.00000212, 0.00000150, -0.00000063, 0.00000055, -0.00000607, -0.00001005, },
		{ -0.00001069, -0.00000228, 0.00000234, -0.00000148, -0.00000113, 0.00000144, -0.00000280, -0.00000240, -0.00000136, -0.00000408, -0.00000407, 0.00000145, -0.00000042, -0.00000079, -0.00000325, -0.00000324, 0.00000212, 0.00000273, -0.00000380, -0.00000344, -0.00000177, 0.00000606, 0.00000895, -0.00000145, -0.00000647, 0.00000001, 0.00000229, 0.00000821, 0.00000560, -0.00000512, -0.00000221, 0.00000230, -0.00000389, -0.00000220, -0.00000017, -0.00000061, -0.00000173, -0.00000049, 0.00000170, 0.00000124, -0.00000154, 0.00000533, 0.00000127, -0.00000159, -0.00000165, 0.00000218, -0.00000263, -0.00000148, -0.00000582, 0.00001775, 0.00000275, -0.00001344, 0.00000603, 0.00002266, -0.00004167, 0.00000816, 0.00001231, -0.00001108, -0.00000695, 0.00003547, -0.00007009, 0.00008506, -0.00007833, 0.00002747, 0.00000456, -0.00004367, 0.00008761, -0.00010578, 0.00008339, -0.00005424, 0.00002069, -0.00001600, -0.00000136, -0.00000866, 0.00003081, -0.00004771, 0.00003089, -0.00002870, 0.00000584, -0.00000103, -0.00000657, -0.00001812, 0.00003194, -0.00005796, 0.00000294, -0.00001814, -0.00001016, -0.00003496, -0.00005101, -0.00013033, -0.00022065, 0.00029134, 0.00013843, 0.00008064, 0.00004862, 0.00009009, 0.00007185, 0.00005933, 0.00020138, 0.00013827, -0.00013781, -0.00014644, -0.00002112, -0.00003923, -0.00002678, -0.00000678, -0.00002516, -0.00001134, -0.00000133, -0.00001166, -0.00000239, -0.00000916, -0.00000722, -0.00000152, -0.00000825, -0.00000552, 0.00000131, -0.00000332, -0.00000119, -0.00000376, -0.00000408, -0.00000089, -0.00000664, -0.00000438, 0.00000175, -0.00000051, 0.00000081, -0.00000311, },
		{ -0.00000073, -0.00000945, -0.00000527, 0.00000155, 0.00000021, -0.00000222, 0.00000129, -0.00000131, -0.00000372, -0.00000113, -0.00000166, -0.00000567, 0.00000013, 0.00000050, -0.00000060, -0.00000197, -0.00000435, -0.00000057, 0.00000517, -0.00000207, -0.00000469, -0.00000271, 0.00000294, 0.00000942, 0.00000327, -0.00000701, -0.00000176, 0.00000139, 0.00000572, 0.00000846, -0.00000191, -0.00000530, 0.00000222, -0.00000177, -0.00000354, -0.00000075, 0.00000003, -0.00000178, -0.00000081, 0.00000072, 0.00000223, -0.00000148, 0.00000351, 0.00000256, -0.00000022, -0.00000352, 0.00000292, -0.00000261, -0.00000014, -0.00000719, 0.00001135, 0.00000924, -0.00000726, -0.00001094, 0.00003320, -0.00002785, -0.00001480, 0.00001781, 0.00000615, -0.00003764, 0.00006471, -0.00008213, 0.00007179, -0.00003873, -0.00001991, 0.00004460, -0.00007264, 0.00009438, -0.00008727, 0.00005085, -0.00002298, 0.00000424, -0.00001301, 0.00000590, -0.00002110, 0.00003725, -0.00004310, 0.00001871, -0.00001287, -0.00000494, -0.00000375, 0.00000614, -0.00002707, 0.00003148, -0.00004675, -0.00000970, -0.00000765, -0.00001242, -0.00003595, -0.00004078, -0.00013083, -0.00021726, 0.00028371, 0.00012327, 0.00008123, 0.00005096, 0.00006381, 0.00008806, 0.00003742, 0.00013845, 0.00019979, -0.00004802, -0.00017530, -0.00004880, -0.00002641, -0.00003553, -0.00000880, -0.00001962, -0.00001692, -0.00000063, -0.00001168, -0.00000443, -0.00000650, -0.00000874, -0.00000177, -0.00000618, -0.00000713, 0.00000092, -0.00000302, -0.00000153, -0.00000291, -0.00000501, -0.00000032, -0.00000529, -0.00000633, 0.00000089, 0.00000010, 0.00000053, },
		{ 0.00000035, 0.00000061, -0.00000708, -0.00000705, -0.00000082, 0.00000175, -0.00000220, 0.00000012, 0.00000033, -0.00000410, -0.00000247, 0.00000105, -0.00000568, -0.00000223, 0.00000123, -0.00000034, -0.00000115, -0.00000375, -0.00000393, 0.00000563, 0.00000126, -0.00000550, -0.00000370, 0.00000069, 0.00000752, 0.00000759, -0.00000520, -0.00000409, 0.00000081, 0.00000323, 0.00000886, 0.00000284, -0.00000669, 0.00000004, 0.00000067, -0.00000371, -0.00000223, 0.00000065, -0.00000164, -0.00000093, -0.00000022, 0.00000240, -0.00000052, 0.00000194, 0.00000223, 0.00000233, -0.00000494, 0.00000255, -0.00000187, 0.00000023, -0.00000659, 0.00000613, 0.00000841, 0.00000493, -0.00002284, 0.00002875, -0.00000543, -0.00002748, 0.00000249, 0.00003802, -0.00006725, 0.00007694, -0.00006853, 0.00003540, 0.00001020, -0.00005959, 0.00006669, -0.00007876, 0.00007915, -0.00005693, 0.00002118, -0.00000571, 0.00000357, -0.00002008, 0.00001533, -0.00002758, 0.00003357, -0.00003110, 0.00000574, -0.00000340, -0.00000477, -0.00001430, 0.00001686, -0.00002670, 0.00002221, -0.00003308, -0.00001793, -0.00000509, -0.00000756, -0.00003999, -0.00003388, -0.00012259, -0.00022292, 0.00027775, 0.00011704, 0.00007010, 0.00006209, 0.00004144, 0.00008843, 0.00004298, 0.00007608, 0.00020776, 0.00005307, -0.00016484, -0.00009050, -0.00002076, -0.00003537, -0.00001690, -0.00001348, -0.00001957, -0.00000295, -0.00000996, -0.00000702, -0.00000500, -0.00000842, -0.00000329, -0.00000417, -0.00000755, -0.00000032, -0.00000253, -0.00000197, -0.00000219, -0.00000536, -0.00000065, -0.00000360, -0.00000722, -0.00000092, 0.00000061, },
		{ 0.00000042, 0.00000053, 0.00000111, -0.00000457, -0.00000702, -0.00000394, 0.00000207, -0.00000094, -0.00000139, 0.00000120, -0.00000308, -0.00000472, 0.00000271, -0.00000372, -0.00000487, 0.00000103, 0.00000050, -0.00000115, -0.00000207, -0.00000634, 0.00000353, 0.00000499, -0.00000475, -0.00000526, -0.00000053, 0.00000446, 0.00000989, -0.00000119, -0.00000590, -0.00000030, 0.00000163, 0.00000703, 0.00000725, -0.00000528, -0.00000339, 0.00000201, -0.00000225, -0.00000409, 0.00000071, -0.00000114, -0.00000103, -0.00000087, 0.00000188, 0.00000051, 0.00000152, 0.00000039, 0.00000480, -0.00000477, 0.00000092, -0.00000091, 0.00000064, -0.00000627, 0.00000447, 0.00000215, 0.00001580, -0.00002369, 0.00001352, 0.00001213, -0.00001970, -0.00003041, 0.00007008, -0.00007976, 0.00006350, -0.00003237, -0.00001067, 0.00005148, -0.00007940, 0.00006715, -0.00006627, 0.00005291, -0.00002853, 0.00000431, -0.00000395, 0.00001252, -0.00002809, 0.00001941, -0.00002528, 0.00002293, -0.00001871, -0.00000149, -0.00000276, 0.00000273, -0.00002464, 0.00001912, -0.00001717, 0.00001017, -0.00002396, -0.00001835, -0.00000879, -0.00000093, -0.00004091, -0.00003399, -0.00010746, -0.00023012, 0.00026547, 0.00012167, 0.00005294, 0.00007140, 0.00003153, 0.00007374, 0.00005893, 0.00003561, 0.00017309, 0.00013515, -0.00011292, -0.00012852, -0.00003125, -0.00002747, -0.00002616, -0.00001100, -0.00001799, -0.00000721, -0.00000832, -0.00000853, -0.00000539, -0.00000685, -0.00000482, -0.00000322, -0.00000672, -0.00000180, -0.00000235, -0.00000222, -0.00000183, -0.00000514, -0.00000140, -0.00000233, -0.00000687, -0.00000301, },
		{ -0.00000459, -0.00000068, 0.00000081, 0.00000130, -0.00000269, -0.00000541, -0.00000657, 0.00000065, 0.00000091, -0.00000222, 0.00000074, -0.00000103, -0.00000660, 0.00000238, -0.00000040, -0.00000646, -0.00000068, 0.00000186, -0.00000153, -0.00000054, -0.00000669, -0.00000036, 0.00000730, -0.00000184, -0.00000699, -0.00000140, 0.00000169, 0.00000949, 0.00000370, -0.00000587, -0.00000233, 0.00000092, 0.00000419, 0.00000962, -0.00000128, -0.00000631, 0.00000118, 0.00000039, -0.00000536, -0.00000022, -0.00000036, -0.00000102, -0.00000130, 0.00000117, 0.00000087, 0.00000237, -0.00000192, 0.00000586, -0.00000276, -0.00000099, -0.00000102, 0.00000224, -0.00000731, 0.00000568, -0.00000486, 0.00001976, -0.00001473, -0.00000165, 0.00001313, 0.00000828, -0.00006692, 0.00008516, -0.00006590, 0.00002749, 0.00001299, -0.00004942, 0.00007262, -0.00007658, 0.00005158, -0.00004547, 0.00002772, -0.00001127, 0.00000203, -0.00001190, 0.00002211, -0.00003021, 0.00001601, -0.00001704, 0.00001161, -0.00001178, -0.00000099, -0.00000805, 0.00001094, -0.00002850, 0.00001214, -0.00000397, 0.00000240, -0.00002282, -0.00001283, -0.00001353, 0.00000217, -0.00003601, -0.00003956, -0.00009117, -0.00023201, 0.00024444, 0.00013245, 0.00003826, 0.00007194, 0.00003385, 0.00005336, 0.00006937, 0.00002270, 0.00011914, 0.00017774, -0.00003376, -0.00014381, -0.00005849, -0.00001892, -0.00003096, -0.00001408, -0.00001365, -0.00001094, -0.00000844, -0.00000836, -0.00000711, -0.00000529, -0.00000534, -0.00000352, -0.00000527, -0.00000280, -0.00000272, -0.00000228, -0.00000178, -0.00000462, -0.00000203, -0.00000186, -0.00000568, },
		{ -0.00000434, -0.00000517, -0.00000236, 0.00000069, 0.00000157, -0.00000161, -0.00000302, -0.00000774, -0.00000217, 0.00000220, -0.00000167, -0.00000079, 0.00000098, -0.00000697, 0.00000014, 0.00000279, -0.00000589, -0.00000366, 0.00000294, -0.00000144, -0.00000009, -0.00000510, -0.00000424, 0.00000678, 0.00000261, -0.00000768, -0.00000281, 0.00000001, 0.00000703, 0.00000759, -0.00000332, -0.00000472, 0.00000031, 0.00000173, 0.00000930, 0.00000380, -0.00000694, -0.00000178, 0.00000291, -0.00000510, -0.00000207, 0.00000022, -0.00000063, -0.00000175, 0.00000069, 0.00000039, 0.00000380, -0.00000344, 0.00000497, 0.00000006, -0.00000152, -0.00000312, 0.00000497, -0.00000880, 0.00000712, -0.00000861, 0.00001646, -0.00000354, -0.00000561, -0.00000535, 0.00004471, -0.00008882, 0.00007246, -0.00002857, -0.00001740, 0.00005058, -0.00006842, 0.00007094, -0.00005782, 0.00003079, -0.00002678, 0.00001131, -0.00000708, 0.00000932, -0.00002102, 0.00002567, -0.00002486, 0.00000809, -0.00000831, 0.00000482, -0.00001206, 0.00000451, -0.00001371, 0.00001448, -0.00002442, 0.00000034, 0.00000595, 0.00000270, -0.00002784, -0.00000664, -0.00001433, -0.00000025, -0.00002702, -0.00004594, -0.00007915, -0.00022606, 0.00021804, 0.00014193, 0.00003193, 0.00006353, 0.00004156, 0.00003771, 0.00006689, 0.00002935, 0.00006981, 0.00017711, 0.00004850, -0.00012512, -0.00009323, -0.00001869, -0.00002870, -0.00002097, -0.00000984, -0.00001199, -0.00001067, -0.00000727, -0.00000886, -0.00000484, -0.00000463, -0.00000447, -0.00000408, -0.00000290, -0.00000347, -0.00000237, -0.00000187, -0.00000410, -0.00000223, -0.00000208, },
		{ -0.00000262, -0.00000340, -0.00000477, -0.00000398, -0.00000017, 0.00000191, -0.00000107, -0.00000072, -0.00000715, -0.00000533, 0.00000193, 0.00000011, -0.00000241, 0.00000182, -0.00000554, -0.00000285, 0.00000423, -0.00000297, -0.00000670, 0.00000268, -0.00000031, -0.00000074, -0.00000280, -0.00000627, 0.00000335, 0.00000680, -0.00000612, -0.00000494, -0.00000085, 0.00000397, 0.00000904, 0.00000120, -0.00000619, -0.00000105, 0.00000035, 0.00000693, 0.00000790, -0.00000449, -0.00000556, 0.00000381, -0.00000309, -0.00000403, -0.00000008, 0.00000029, -0.00000224, 0.00000048, -0.00000053, 0.00000487, -0.00000353, 0.00000286, 0.00000192, 0.00000045, -0.00000672, 0.00000730, -0.00000886, 0.00000668, -0.00000862, 0.00001072, 0.00000071, 0.00000640, -0.00003514, 0.00007175, -0.00008293, 0.00003371, 0.00001804, -0.00005375, 0.00006813, -0.00006518, 0.00005346, -0.00003474, 0.00001471, -0.00001599, 0.00000503, -0.00001193, 0.00001851, -0.00002480, 0.00002163, -0.00001525, 0.00000071, -0.00000342, 0.00000412, -0.00001715, 0.00001008, -0.00001528, 0.00001188, -0.00001555, -0.00001003, 0.00000826, 0.00000981, -0.00003375, -0.00000476, -0.00000958, -0.00000613, -0.00001812, -0.00004891, -0.00007338, -0.00021458, 0.00019235, 0.00014459, 0.00003420, 0.00005149, 0.00004660, 0.00003217, 0.00005384, 0.00004159, 0.00003991, 0.00014504, 0.00011028, -0.00007437, -0.00012014, -0.00003225, -0.00002185, -0.00002743, -0.00000947, -0.00001000, -0.00001380, -0.00000668, -0.00000962, -0.00000574, -0.00000333, -0.00000518, -0.00000367, -0.00000228, -0.00000414, -0.00000272, -0.00000199, -0.00000376, -0.00000197, },
		{ -0.00000140, -0.00000308, -0.00000310, -0.00000381, -0.00000493, -0.00000170, 0.00000200, -0.00000070, 0.00000100, -0.00000528, -0.00000760, -0.00000016, 0.00000217, -0.00000300, 0.00000109, -0.00000309, -0.00000494, 0.00000308, 0.00000118, -0.00000813, 0.00000050, 0.00000151, -0.00000158, -0.00000133, -0.00000566, -0.00000151, 0.00000864, -0.00000206, -0.00000689, -0.00000185, 0.00000167, 0.00000782, 0.00000604, -0.00000544, -0.00000333, -0.00000025, 0.00000399, 0.00000948, 0.00000032, -0.00000815, 0.00000224, -0.00000017, -0.00000494, -0.00000161, 0.00000148, -0.00000252, 0.00000027, -0.00000130, 0.00000507, -0.00000259, 0.00000105, 0.00000146, 0.00000448, -0.00000987, 0.00000728, -0.00000657, 0.00000486, -0.00000815, 0.00000916, -0.00000686, 0.00002950, -0.00006074, 0.00007445, -0.00004832, -0.00001695, 0.00005533, -0.00006903, 0.00006360, -0.00004702, 0.00003206, -0.00001749, 0.00000790, -0.00001286, 0.00000541, -0.00001958, 0.00002359, -0.00002139, 0.00001297, -0.00000644, -0.00000236, -0.00000342, 0.00000750, -0.00002268, 0.00001172, -0.00001163, 0.00000558, -0.00000696, -0.00001456, 0.00000329, 0.00001875, -0.00003532, -0.00000885, -0.00000160, -0.00001134, -0.00001283, -0.00004718, -0.00007206, -0.00020225, 0.00017226, 0.00013958, 0.00004055, 0.00004240, 0.00004453, 0.00003525, 0.00003818, 0.00004784, 0.00003147, 0.00010133, 0.00013785, -0.00000592, -0.00012493, -0.00005760, -0.00001681, -0.00002961, -0.00001314, -0.00000664, -0.00001590, -0.00000757, -0.00000927, -0.00000731, -0.00000240, -0.00000512, -0.00000394, -0.00000148, -0.00000432, -0.00000330, -0.00000219, -0.00000365, },
		{ -0.00000377, -0.00000072, -0.00000321, -0.00000334, -0.00000283, -0.00000499, -0.00000345, 0.00000146, -0.00000036, 0.00000214, -0.00000295, -0.00000823, -0.00000338, 0.00000323, -0.00000200, -0.00000067, -0.00000101, -0.00000494, -0.00000015, 0.00000461, -0.00000680, -0.00000304, 0.00000281, -0.00000143, -0.00000149, -0.00000313, -0.00000553, 0.00000693, 0.00000324, -0.00000714, -0.00000377, 0.00000054, 0.00000497, 0.00000928, -0.00000212, -0.00000567, -0.00000105, 0.00000179, 0.00000830, 0.00000559, -0.00000792, -0.00000128, 0.00000206, -0.00000395, -0.00000400, 0.00000221, -0.00000216, -0.00000018, -0.00000172, 0.00000464, -0.00000172, 0.00000080, -0.00000122, 0.00000864, -0.00001034, 0.00000427, -0.00000333, 0.00000469, -0.00001165, 0.00001547, -0.00002346, 0.00005141, -0.00006732, 0.00004870, 0.00000257, -0.00005963, 0.00006961, -0.00006156, 0.00004503, -0.00002602, 0.00001686, -0.00001027, 0.00000872, -0.00001331, 0.00000771, -0.00002522, 0.00002247, -0.00001317, 0.00000455, -0.00000213, -0.00000058, -0.00000622, 0.00001133, -0.00002515, 0.00000843, -0.00000498, -0.00000032, -0.00000244, -0.00001271, -0.00000492, 0.00002412, -0.00003044, -0.00001657, 0.00000534, -0.00001266, -0.00001180, -0.00004262, -0.00007151, -0.00019297, 0.00015898, 0.00013026, 0.00004515, 0.00004000, 0.00003632, 0.00004108, 0.00002751, 0.00004389, 0.00003668, 0.00006422, 0.00013135, 0.00005991, -0.00010122, -0.00008542, -0.00002017, -0.00002669, -0.00001881, -0.00000447, -0.00001573, -0.00000979, -0.00000860, -0.00000862, -0.00000240, -0.00000443, -0.00000438, -0.00000099, -0.00000396, -0.00000387, -0.00000255, },
		{ -0.00000301, -0.00000417, -0.00000009, -0.00000293, -0.00000384, -0.00000223, -0.00000431, -0.00000483, 0.00000023, -0.00000026, 0.00000297, -0.00000084, -0.00000725, -0.00000650, 0.00000243, 0.00000013, -0.00000217, -0.00000050, -0.00000288, -0.00000374, 0.00000550, -0.00000290, -0.00000625, 0.00000228, 0.00000015, -0.00000283, -0.00000046, -0.00000684, 0.00000222, 0.00000748, -0.00000454, -0.00000640, -0.00000009, 0.00000212, 0.00000977, 0.00000280, -0.00000660, -0.00000280, 0.00000075, 0.00000543, 0.00000921, -0.00000453, -0.00000500, 0.00000211, -0.00000119, -0.00000614, 0.00000171, -0.00000110, -0.00000069, -0.00000210, 0.00000435, -0.00000188, 0.00000224, -0.00000454, 0.00001056, -0.00000733, 0.00000009, -0.00000258, 0.00000963, -0.00002110, 0.00002749, -0.00003982, 0.00005903, -0.00004865, 0.00000367, 0.00004978, -0.00007842, 0.00005860, -0.00004035, 0.00002509, -0.00001208, 0.00001167, -0.00001078, 0.00001203, -0.00001305, 0.00000894, -0.00002717, 0.00001658, -0.00000435, 0.00000009, -0.00000293, 0.00000377, -0.00000860, 0.00001286, -0.00002358, 0.00000209, 0.00000106, -0.00000273, -0.00000270, -0.00000736, -0.00001139, 0.00002315, -0.00002103, -0.00002345, 0.00000792, -0.00000973, -0.00001290, -0.00003841, -0.00006888, -0.00018783, 0.00015023, 0.00012152, 0.00004428, 0.00004355, 0.00002658, 0.00004375, 0.00002492, 0.00003276, 0.00004443, 0.00004353, 0.00010238, 0.00010537, -0.00005359, -0.00010300, -0.00003454, -0.00002177, -0.00002310, -0.00000524, -0.00001355, -0.00001215, -0.00000862, -0.00000912, -0.00000320, -0.00000370, -0.00000448, -0.00000095, -0.00000333, -0.00000415, },
		{ -0.00000408, -0.00000337, -0.00000486, 0.00000031, -0.00000228, -0.00000430, -0.00000215, -0.00000330, -0.00000548, -0.00000138, -0.00000074, 0.00000368, 0.00000080, -0.00000530, -0.00000837, -0.00000015, 0.00000215, -0.00000226, -0.00000174, 0.00000005, -0.00000578, 0.00000330, 0.00000190, -0.00000721, -0.00000053, 0.00000243, -0.00000396, 0.00000067, -0.00000513, -0.00000332, 0.00000844, 0.00000068, -0.00000838, -0.00000139, 0.00000043, 0.00000777, 0.00000739, -0.00000499, -0.00000536, 0.00000025, 0.00000247, 0.00000998, 0.00000076, -0.00000680, -0.00000048, 0.00000206, -0.00000671, -0.00000023, 0.00000011, -0.00000066, -0.00000298, 0.00000474, -0.00000311, 0.00000420, -0.00000645, 0.00000921, -0.00000256, -0.00000154, -0.00000762, 0.00002050, -0.00003379, 0.00003794, -0.00004532, 0.00004563, -0.00001057, -0.00004256, 0.00007549, -0.00006897, 0.00003192, -0.00001944, 0.00001382, -0.00000834, 0.00001357, -0.00001331, 0.00001326, -0.00001038, 0.00000888, -0.00002627, 0.00000894, 0.00000161, 0.00000053, -0.00000667, 0.00000739, -0.00000830, 0.00001151, -0.00001942, -0.00000416, 0.00000375, -0.00000116, -0.00000561, -0.00000248, -0.00001300, 0.00001681, -0.00001145, -0.00002580, 0.00000577, -0.00000499, -0.00001297, -0.00003674, -0.00006366, -0.00018517, 0.00014232, 0.00011678, 0.00003806, 0.00004906, 0.00002017, 0.00004070, 0.00002834, 0.00002090, 0.00004660, 0.00003907, 0.00006719, 0.00012192, 0.00000441, -0.00010024, -0.00005631, -0.00002023, -0.00002372, -0.00000872, -0.00001098, -0.00001331, -0.00000976, -0.00000903, -0.00000414, -0.00000352, -0.00000414, -0.00000110, -0.00000280, },
		{ -0.00000248, -0.00000384, -0.00000340, -0.00000570, 0.00000030, -0.00000138, -0.00000448, -0.00000250, -0.00000235, -0.00000537, -0.00000284, -0.00000197, 0.00000416, 0.00000211, -0.00000316, -0.00000852, -0.00000352, 0.00000272, -0.00000063, -0.00000381, 0.00000214, -0.00000518, -0.00000091, 0.00000521, -0.00000486, -0.00000459, 0.00000377, -0.00000355, -0.00000024, -0.00000178, -0.00000708, 0.00000534, 0.00000662, -0.00000794, -0.00000392, -0.00000015, 0.00000469, 0.00000980, -0.00000091, -0.00000753, -0.00000077, 0.00000054, 0.00000814, 0.00000577, -0.00000533, -0.00000464, 0.00000394, -0.00000505, -0.00000284, 0.00000042, 0.00000047, -0.00000451, 0.00000561, -0.00000450, 0.00000511, -0.00000592, 0.00000586, 0.00000033, 0.00000282, -0.00001890, 0.00003389, -0.00004306, 0.00003850, -0.00003412, 0.00001469, 0.00003186, -0.00007191, 0.00007241, -0.00003993, 0.00000532, -0.00000994, 0.00001368, -0.00001093, 0.00001648, -0.00001317, 0.00001108, -0.00000649, 0.00000885, -0.00002404, 0.00000204, 0.00000340, 0.00000425, -0.00001018, 0.00000806, -0.00000518, 0.00000870, -0.00001512, -0.00000790, 0.00000260, 0.00000249, -0.00000800, -0.00000082, -0.00000982, 0.00000873, -0.00000578, -0.00002279, 0.00000145, -0.00000178, -0.00001009, -0.00003754, -0.00005749, -0.00018234, 0.00013257, 0.00011647, 0.00002955, 0.00005213, 0.00001922, 0.00003345, 0.00003294, 0.00001367, 0.00004107, 0.00004377, 0.00003932, 0.00011215, 0.00005623, -0.00007448, -0.00007673, -0.00002642, -0.00002111, -0.00001285, -0.00000986, -0.00001271, -0.00001154, -0.00000909, -0.00000463, -0.00000400, -0.00000372, -0.00000109, },
		{ -0.00000079, -0.00000222, -0.00000372, -0.00000309, -0.00000640, -0.00000023, -0.00000048, -0.00000424, -0.00000309, -0.00000173, -0.00000476, -0.00000372, -0.00000374, 0.00000405, 0.00000332, -0.00000132, -0.00000726, -0.00000632, 0.00000120, 0.00000191, -0.00000521, 0.00000227, -0.00000239, -0.00000495, 0.00000524, 0.00000008, -0.00000777, 0.00000275, -0.00000141, -0.00000222, 0.00000100, -0.00000740, -0.00000055, 0.00001055, -0.00000416, -0.00000698, -0.00000068, 0.00000221, 0.00000929, 0.00000424, -0.00000781, -0.00000286, -0.00000031, 0.00000510, 0.00000849, -0.00000084, -0.00000824, 0.00000313, -0.00000179, -0.00000462, -0.00000105, 0.00000261, -0.00000612, 0.00000609, -0.00000489, 0.00000423, -0.00000389, 0.00000366, -0.00000216, 0.00001376, -0.00003271, 0.00004310, -0.00004191, 0.00002493, -0.00000874, -0.00002178, 0.00006193, -0.00007486, 0.00004657, -0.00000777, -0.00000794, -0.00001444, 0.00001948, -0.00001300, 0.00001580, -0.00000943, 0.00000748, -0.00000375, 0.00000984, -0.00002132, -0.00000297, 0.00000168, 0.00000859, -0.00001113, 0.00000559, -0.00000086, 0.00000649, -0.00001246, -0.00000856, -0.00000084, 0.00000547, -0.00000772, -0.00000260, -0.00000440, 0.00000285, -0.00000564, -0.00001667, -0.00000138, -0.00000214, -0.00000466, -0.00003887, -0.00005258, -0.00017760, 0.00012052, 0.00011855, 0.00002247, 0.00005049, 0.00002248, 0.00002583, 0.00003458, 0.00001259, 0.00003095, 0.00004909, 0.00002494, 0.00008673, 0.00008905, -0.00003178, -0.00008582, -0.00004078, -0.00001851, -0.00001523, -0.00001099, -0.00001101, -0.00001294, -0.00000989, -0.00000457, -0.00000476, -0.00000369, },
		{ -0.00000422, -0.00000049, -0.00000171, -0.00000382, -0.00000266, -0.00000666, -0.00000116, 0.00000011, -0.00000354, -0.00000370, -0.00000152, -0.00000401, -0.00000385, -0.00000555, 0.00000307, 0.00000438, 0.00000023, -0.00000537, -0.00000757, -0.00000196, 0.00000384, -0.00000483, 0.00000056, 0.00000095, -0.00000672, 0.00000183, 0.00000535, -0.00000795, -0.00000080, 0.00000125, -0.00000344, 0.00000157, -0.00000458, -0.00000647, 0.00001035, 0.00000213, -0.00000878, -0.00000238, 0.00000115, 0.00000664, 0.00000835, -0.00000534, -0.00000548, -0.00000099, 0.00000248, 0.00000820, 0.00000488, -0.00000916, -0.00000029, 0.00000129, -0.00000413, -0.00000423, 0.00000477, -0.00000671, 0.00000532, -0.00000396, 0.00000247, -0.00000298, 0.00000578, -0.00001092, 0.00002775, -0.00004228, 0.00004157, -0.00002731, 0.00000001, 0.00002103, -0.00004939, 0.00006941, -0.00005441, 0.00001330, 0.00001202, -0.00000362, -0.00002663, 0.00002280, -0.00000981, 0.00001127, -0.00000467, 0.00000542, -0.00000341, 0.00001143, -0.00001802, -0.00000600, -0.00000196, 0.00001130, -0.00000896, 0.00000151, 0.00000254, 0.00000619, -0.00001175, -0.00000721, -0.00000419, 0.00000584, -0.00000463, -0.00000598, 0.00000013, 0.00000127, -0.00000960, -0.00001120, -0.00000024, -0.00000561, 0.00000107, -0.00003855, -0.00005009, -0.00017099, 0.00010757, 0.00012026, 0.00001909, 0.00004482, 0.00002656, 0.00002123, 0.00003196, 0.00001549, 0.00002124, 0.00004955, 0.00002260, 0.00005861, 0.00009818, 0.00001575, -0.00007702, -0.00005883, -0.00002005, -0.00001483, -0.00001362, -0.00000960, -0.00001314, -0.00001142, -0.00000437, -0.00000524, },
		{ -0.00000521, -0.00000507, -0.00000063, -0.00000083, -0.00000392, -0.00000246, -0.00000639, -0.00000214, 0.00000009, -0.00000249, -0.00000406, -0.00000167, -0.00000336, -0.00000343, -0.00000686, 0.00000129, 0.00000493, 0.00000170, -0.00000352, -0.00000711, -0.00000546, 0.00000388, -0.00000268, -0.00000167, 0.00000294, -0.00000541, -0.00000330, 0.00000825, -0.00000442, -0.00000531, 0.00000244, -0.00000259, -0.00000013, -0.00000063, -0.00000964, 0.00000582, 0.00000842, -0.00000747, -0.00000548, 0.00000098, 0.00000358, 0.00000981, -0.00000068, -0.00000717, -0.00000251, 0.00000105, 0.00000567, 0.00000948, -0.00000655, -0.00000463, 0.00000216, -0.00000105, -0.00000785, 0.00000543, -0.00000553, 0.00000341, -0.00000278, 0.00000205, -0.00000578, 0.00001315, -0.00002297, 0.00003825, -0.00004113, 0.00002682, -0.00000262, -0.00002743, 0.00004319, -0.00005861, 0.00005494, -0.00002343, -0.00001106, 0.00001253, 0.00001222, -0.00003606, 0.00001790, -0.00000187, 0.00000644, -0.00000231, 0.00000632, -0.00000458, 0.00001213, -0.00001395, -0.00000720, -0.00000601, 0.00001134, -0.00000471, -0.00000210, 0.00000371, 0.00000776, -0.00001203, -0.00000557, -0.00000571, 0.00000338, -0.00000035, -0.00000853, 0.00000184, 0.00000351, -0.00001445, -0.00000940, 0.00000481, -0.00000971, 0.00000468, -0.00003564, -0.00004947, -0.00016392, 0.00009559, 0.00011983, 0.00001934, 0.00003776, 0.00002829, 0.00002071, 0.00002676, 0.00001861, 0.00001554, 0.00004439, 0.00002628, 0.00003752, 0.00008753, 0.00005513, -0.00005040, -0.00007287, -0.00002817, -0.00001292, -0.00001597, -0.00000961, -0.00001215, -0.00001305, -0.00000463, },
		{ -0.00000558, -0.00000492, -0.00000569, -0.00000146, 0.00000012, -0.00000363, -0.00000268, -0.00000582, -0.00000271, -0.00000063, -0.00000143, -0.00000394, -0.00000207, -0.00000297, -0.00000274, -0.00000740, -0.00000080, 0.00000456, 0.00000318, -0.00000194, -0.00000552, -0.00000793, 0.00000170, 0.00000001, -0.00000274, 0.00000262, -0.00000195, -0.00000753, 0.00000715, 0.00000152, -0.00000821, 0.00000073, 0.00000013, -0.00000254, 0.00000193, -0.00000880, -0.00000103, 0.00001171, -0.00000253, -0.00000875, 0.00000032, 0.00000165, 0.00000839, 0.00000436, -0.00000649, -0.00000509, 0.00000044, 0.00000257, 0.00001130, -0.00000138, -0.00000739, -0.00000021, 0.00000336, -0.00000989, 0.00000362, -0.00000299, 0.00000176, -0.00000336, 0.00000503, -0.00001293, 0.00002325, -0.00003242, 0.00003886, -0.00002677, 0.00000246, 0.00002355, -0.00004690, 0.00004970, -0.00004901, 0.00002867, 0.00000255, -0.00001727, -0.00000219, 0.00002780, -0.00003485, 0.00000541, 0.00000585, 0.00000544, -0.00000365, 0.00000892, -0.00000502, 0.00001086, -0.00000964, -0.00000653, -0.00000940, 0.00000883, -0.00000010, -0.00000376, 0.00000267, 0.00001003, -0.00001197, -0.00000489, -0.00000496, -0.00000062, 0.00000300, -0.00000877, 0.00000084, 0.00000732, -0.00001711, -0.00001200, 0.00001143, -0.00001161, 0.00000516, -0.00003102, -0.00004909, -0.00015787, 0.00008558, 0.00011719, 0.00002145, 0.00003213, 0.00002639, 0.00002283, 0.00002192, 0.00001922, 0.00001436, 0.00003641, 0.00002968, 0.00002704, 0.00006642, 0.00007758, -0.00001257, -0.00007551, -0.00004171, -0.00001248, -0.00001654, -0.00001113, -0.00001079, -0.00001398, },
		{ -0.00001387, -0.00000686, -0.00000489, -0.00000574, -0.00000276, 0.00000060, -0.00000272, -0.00000312, -0.00000546, -0.00000253, -0.00000178, -0.00000084, -0.00000327, -0.00000252, -0.00000287, -0.00000206, -0.00000728, -0.00000257, 0.00000311, 0.00000438, -0.00000045, -0.00000363, -0.00000861, -0.00000184, 0.00000158, -0.00000174, 0.00000056, 0.00000155, -0.00000875, 0.00000249, 0.00000703, -0.00000740, -0.00000357, 0.00000299, -0.00000355, 0.00000170, -0.00000495, -0.00000695, 0.00001019, 0.00000456, -0.00001005, -0.00000194, 0.00000123, 0.00000536, 0.00000764, -0.00000303, -0.00000774, -0.00000046, 0.00000030, 0.00001022, 0.00000407, -0.00000670, -0.00000499, 0.00000673, -0.00000886, -0.00000018, -0.00000090, 0.00000244, -0.00000735, 0.00001167, -0.00002214, 0.00003097, -0.00003335, 0.00002686, -0.00000267, -0.00002294, 0.00004143, -0.00005168, 0.00004030, -0.00002858, 0.00000419, 0.00001324, -0.00000669, -0.00002064, 0.00003274, -0.00002230, -0.00000836, 0.00000768, 0.00000963, -0.00000678, 0.00001041, -0.00000304, 0.00000803, -0.00000642, -0.00000408, -0.00001136, 0.00000454, 0.00000335, -0.00000311, 0.00000049, 0.00001165, -0.00001074, -0.00000543, -0.00000275, -0.00000434, 0.00000403, -0.00000680, -0.00000127, 0.00001021, -0.00001622, -0.00001740, 0.00001652, -0.00000985, 0.00000350, -0.00002662, -0.00004750, -0.00015337, 0.00007727, 0.00011347, 0.00002336, 0.00002935, 0.00002179, 0.00002491, 0.00001968, 0.00001684, 0.00001571, 0.00002931, 0.00002925, 0.00002521, 0.00004480, 0.00008117, 0.00002612, -0.00006315, -0.00005592, -0.00001637, -0.00001534, -0.00001314, -0.00001012, },
		{ -0.00001064, -0.00001315, -0.00000778, -0.00000552, -0.00000532, -0.00000395, 0.00000019, -0.00000139, -0.00000324, -0.00000573, -0.00000172, -0.00000277, -0.00000106, -0.00000219, -0.00000271, -0.00000310, -0.00000149, -0.00000681, -0.00000356, 0.00000095, 0.00000474, 0.00000121, -0.00000195, -0.00000769, -0.00000524, 0.00000085, 0.00000090, -0.00000155, 0.00000313, -0.00000658, -0.00000340, 0.00000922, -0.00000262, -0.00000832, 0.00000371, -0.00000201, -0.00000081, -0.00000069, -0.00000921, 0.00000435, 0.00001075, -0.00000763, -0.00000560, 0.00000131, 0.00000262, 0.00000794, 0.00000214, -0.00000884, -0.00000246, -0.00000083, 0.00000757, 0.00000748, -0.00000241, -0.00000977, 0.00000691, -0.00000493, -0.00000381, -0.00000168, 0.00000685, -0.00001453, 0.00001964, -0.00002887, 0.00003110, -0.00002305, 0.00000516, 0.00002292, -0.00003990, 0.00004503, -0.00004209, 0.00002197, -0.00000876, -0.00000855, 0.00000772, 0.00001100, -0.00003106, 0.00002388, -0.00000498, -0.00001562, 0.00000144, 0.00001623, -0.00000798, 0.00000871, 0.00000104, 0.00000547, -0.00000547, -0.00000053, -0.00001123, -0.00000053, 0.00000471, -0.00000079, -0.00000142, 0.00001177, -0.00000841, -0.00000657, -0.00000031, -0.00000644, 0.00000273, -0.00000402, -0.00000261, 0.00001085, -0.00001249, -0.00002282, 0.00001798, -0.00000509, 0.00000196, -0.00002407, -0.00004445, -0.00014976, 0.00006961, 0.00010994, 0.00002391, 0.00002911, 0.00001667, 0.00002482, 0.00002036, 0.00001301, 0.00001686, 0.00002533, 0.00002500, 0.00002719, 0.00002937, 0.00007031, 0.00005598, -0.00003766, -0.00006447, -0.00002540, -0.00001409, -0.00001427, },
		{ -0.00001392, -0.00001190, -0.00001265, -0.00000783, -0.00000667, -0.00000496, -0.00000448, -0.00000107, -0.00000031, -0.00000254, -0.00000660, -0.00000081, -0.00000301, -0.00000212, -0.00000119, -0.00000234, -0.00000359, -0.00000110, -0.00000628, -0.00000371, -0.00000121, 0.00000384, 0.00000291, -0.00000044, -0.00000596, -0.00000715, -0.00000209, 0.00000358, -0.00000200, 0.00000217, -0.00000252, -0.00000752, 0.00000684, 0.00000410, -0.00001061, 0.00000109, 0.00000135, -0.00000348, 0.00000137, -0.00000716, -0.00000306, 0.00001298, -0.00000148, -0.00000891, 0.00000031, 0.00000145, 0.00000564, 0.00000687, -0.00000718, -0.00000528, -0.00000165, 0.00000508, 0.00000775, 0.00000368, -0.00001182, 0.00000342, -0.00000033, -0.00000440, -0.00000678, 0.00001441, -0.00002237, 0.00002480, -0.00002862, 0.00002126, -0.00000390, -0.00001872, 0.00004070, -0.00004264, 0.00003506, -0.00002488, 0.00000468, 0.00000173, -0.00000744, -0.00000668, 0.00002417, -0.00002774, 0.00000634, 0.00000761, -0.00001224, -0.00000976, 0.00002013, -0.00000460, 0.00000432, 0.00000500, 0.00000516, -0.00000686, 0.00000265, -0.00000869, -0.00000522, 0.00000376, 0.00000203, -0.00000203, 0.00001045, -0.00000574, -0.00000746, 0.00000141, -0.00000661, 0.00000018, -0.00000212, -0.00000214, 0.00000954, -0.00000789, -0.00002591, 0.00001571, 0.00000038, 0.00000265, -0.00002366, -0.00004093, -0.00014586, 0.00006172, 0.00010718, 0.00002319, 0.00002998, 0.00001300, 0.00002197, 0.00002256, 0.00000988, 0.00001611, 0.00002446, 0.00001920, 0.00002853, 0.00002201, 0.00005272, 0.00007144, -0.00000555, -0.00006237, -0.00003733, -0.00001526, },
		{ -0.00002047, -0.00001285, -0.00001284, -0.00001305, -0.00000713, -0.00000774, -0.00000522, -0.00000424, -0.00000258, -0.00000018, -0.00000095, -0.00000753, -0.00000051, -0.00000226, -0.00000352, -0.00000084, -0.00000132, -0.00000413, -0.00000095, -0.00000584, -0.00000335, -0.00000265, 0.00000175, 0.00000410, 0.00000124, -0.00000425, -0.00000715, -0.00000598, 0.00000453, -0.00000015, -0.00000028, 0.00000096, -0.00000798, 0.00000104, 0.00000942, -0.00000852, -0.00000400, 0.00000429, -0.00000408, 0.00000022, -0.00000252, -0.00000848, 0.00000990, 0.00000621, -0.00000948, -0.00000263, 0.00000167, 0.00000243, 0.00000920, -0.00000294, -0.00000750, -0.00000332, 0.00000376, 0.00000555, 0.00000888, -0.00000972, -0.00000189, 0.00000166, -0.00000005, -0.00001544, 0.00002218, -0.00002687, 0.00002330, -0.00001951, 0.00000360, 0.00001745, -0.00003607, 0.00004481, -0.00003186, 0.00001826, -0.00000927, -0.00000409, 0.00000084, 0.00000213, -0.00001968, 0.00002568, -0.00001345, -0.00001006, 0.00000884, -0.00000041, -0.00001933, 0.00001729, 0.00000279, 0.00000010, 0.00000635, 0.00000765, -0.00000910, 0.00000385, -0.00000411, -0.00000833, 0.00000086, 0.00000417, -0.00000101, 0.00000839, -0.00000360, -0.00000755, 0.00000218, -0.00000544, -0.00000214, -0.00000210, -0.00000014, 0.00000765, -0.00000433, -0.00002588, 0.00001139, 0.00000406, 0.00000625, -0.00002416, -0.00003843, -0.00014092, 0.00005349, 0.00010486, 0.00002200, 0.00003051, 0.00001161, 0.00001739, 0.00002431, 0.00000877, 0.00001347, 0.00002506, 0.00001440, 0.00002707, 0.00002068, 0.00003573, 0.00007240, 0.00002482, -0.00004831, -0.00004763, },
		{ -0.00005146, -0.00002920, -0.00001292, -0.00001259, -0.00001427, -0.00000636, -0.00000807, -0.00000624, -0.00000370, -0.00000352, -0.00000131, 0.00000095, -0.00000773, -0.00000118, -0.00000087, -0.00000448, -0.00000154, 0.00000011, -0.00000435, -0.00000111, -0.00000548, -0.00000290, -0.00000305, -0.00000087, 0.00000408, 0.00000327, -0.00000282, -0.00000577, -0.00000912, 0.00000285, 0.00000306, -0.00000223, 0.00000200, -0.00000493, -0.00000524, 0.00001050, -0.00000244, -0.00000883, 0.00000442, -0.00000180, -0.00000294, 0.00000164, -0.00000947, 0.00000282, 0.00001197, -0.00000584, -0.00000675, 0.00000183, 0.00000019, 0.00000845, 0.00000225, -0.00000746, -0.00000623, 0.00000327, 0.00000268, 0.00001113, -0.00000438, -0.00000555, -0.00000148, 0.00000862, -0.00002436, 0.00002605, -0.00002468, 0.00001384, -0.00000366, -0.00001597, 0.00003316, -0.00004115, 0.00003554, -0.00001415, 0.00000361, -0.00000199, -0.00000279, -0.00000702, 0.00001164, -0.00002401, 0.00001600, 0.00000327, -0.00001686, -0.00000127, 0.00001263, -0.00002122, 0.00000771, 0.00001037, -0.00000061, 0.00000421, 0.00001152, -0.00001000, 0.00000231, 0.00000117, -0.00000889, -0.00000312, 0.00000469, 0.00000127, 0.00000654, -0.00000255, -0.00000687, 0.00000233, -0.00000390, -0.00000315, -0.00000389, 0.00000218, 0.00000663, -0.00000269, -0.00002359, 0.00000739, 0.00000457, 0.00001173, -0.00002364, -0.00003788, -0.00013515, 0.00004557, 0.00010233, 0.00002113, 0.00003000, 0.00001204, 0.00001274, 0.00002428, 0.00000957, 0.00001017, 0.00002532, 0.00001192, 0.00002313, 0.00002179, 0.00002375, 0.00006317, 0.00004662, -0.00002528, },
		{ 0.00000082, -0.00004596, -0.00003860, -0.00001595, -0.00001135, -0.00001549, -0.00000630, -0.00000749, -0.00000753, -0.00000360, -0.00000342, -0.00000329, 0.00000217, -0.00000668, -0.00000254, 0.00000030, -0.00000434, -0.00000314, 0.00000131, -0.00000389, -0.00000157, -0.00000518, -0.00000261, -0.00000262, -0.00000309, 0.00000248, 0.00000528, -0.00000136, -0.00000400, -0.00001038, -0.00000089, 0.00000571, -0.00000201, 0.00000042, -0.00000050, -0.00000884, 0.00000662, 0.00000488, -0.00001031, 0.00000091, 0.00000202, -0.00000557, 0.00000284, -0.00000611, -0.00000488, 0.00001291, 0.00000134, -0.00000981, 0.00000033, -0.00000025, 0.00000564, 0.00000610, -0.00000444, -0.00000948, 0.00000235, 0.00000076, 0.00001027, 0.00000127, -0.00000441, -0.00000971, 0.00001821, -0.00002909, 0.00002284, -0.00001505, -0.00000129, 0.00001359, -0.00003047, 0.00003793, -0.00003381, 0.00001885, 0.00000166, -0.00000244, -0.00000412, 0.00000436, -0.00001460, 0.00001469, -0.00001882, 0.00000172, 0.00001353, -0.00001158, -0.00001632, 0.00001894, -0.00001379, -0.00000428, 0.00001346, 0.00000350, 0.00000025, 0.00001417, -0.00000806, -0.00000119, 0.00000526, -0.00000658, -0.00000687, 0.00000318, 0.00000399, 0.00000569, -0.00000258, -0.00000587, 0.00000247, -0.00000276, -0.00000259, -0.00000652, 0.00000349, 0.00000715, -0.00000254, -0.00002072, 0.00000545, 0.00000224, 0.00001703, -0.00002072, -0.00003899, -0.00012959, 0.00003872, 0.00009919, 0.00002080, 0.00002859, 0.00001320, 0.00000936, 0.00002236, 0.00001110, 0.00000761, 0.00002427, 0.00001150, 0.00001839, 0.00002235, 0.00001759, 0.00004976, 0.00005676, },
		{ 0.00005627, 0.00002366, -0.00003158, -0.00004468, -0.00002238, -0.00001042, -0.00001575, -0.00000715, -0.00000652, -0.00000832, -0.00000442, -0.00000249, -0.00000513, 0.00000192, -0.00000461, -0.00000368, 0.00000045, -0.00000303, -0.00000491, 0.00000160, -0.00000267, -0.00000208, -0.00000505, -0.00000246, -0.00000190, -0.00000421, -0.00000030, 0.00000643, 0.00000058, -0.00000252, -0.00000970, -0.00000504, 0.00000603, 0.00000054, -0.00000221, 0.00000265, -0.00000820, -0.00000029, 0.00000974, -0.00000691, -0.00000470, 0.00000470, -0.00000545, 0.00000063, -0.00000081, -0.00000953, 0.00000834, 0.00000922, -0.00000939, -0.00000326, 0.00000033, 0.00000276, 0.00000701, 0.00000070, -0.00001125, 0.00000002, 0.00000002, 0.00000806, 0.00000394, 0.00000248, -0.00001990, 0.00002388, -0.00002623, 0.00001220, -0.00000065, -0.00001704, 0.00002608, -0.00003523, 0.00003144, -0.00001920, 0.00000308, 0.00000901, 0.00000109, -0.00001138, 0.00001075, -0.00001649, 0.00001023, -0.00000890, -0.00000916, 0.00001310, 0.00000146, -0.00002758, 0.00001455, -0.00000092, -0.00001245, 0.00000957, 0.00001052, -0.00000227, 0.00001362, -0.00000362, -0.00000461, 0.00000663, -0.00000223, -0.00000893, -0.00000020, 0.00000615, 0.00000615, -0.00000323, -0.00000514, 0.00000299, -0.00000231, -0.00000090, -0.00000876, 0.00000307, 0.00000882, -0.00000269, -0.00001875, 0.00000596, -0.00000129, 0.00002022, -0.00001528, -0.00004044, -0.00012532, 0.00003320, 0.00009560, 0.00002076, 0.00002685, 0.00001409, 0.00000764, 0.00001948, 0.00001205, 0.00000643, 0.00002212, 0.00001199, 0.00001446, 0.00002109, 0.00001555, 0.00003731, },
		{ 0.00002854, 0.00004887, 0.00003887, -0.00001180, -0.00004416, -0.00003065, -0.00001160, -0.00001479, -0.00000835, -0.00000599, -0.00000809, -0.00000598, -0.00000154, -0.00000591, 0.00000016, -0.00000249, -0.00000362, -0.00000065, -0.00000123, -0.00000587, 0.00000063, -0.00000109, -0.00000219, -0.00000525, -0.00000232, -0.00000134, -0.00000412, -0.00000331, 0.00000593, 0.00000300, -0.00000130, -0.00000797, -0.00000787, 0.00000357, 0.00000399, -0.00000367, 0.00000289, -0.00000419, -0.00000675, 0.00000952, 0.00000024, -0.00000915, 0.00000389, -0.00000231, -0.00000312, 0.00000307, -0.00000923, 0.00000040, 0.00001410, -0.00000449, -0.00000757, 0.00000016, 0.00000147, 0.00000506, 0.00000579, -0.00001008, -0.00000342, -0.00000082, 0.00000696, 0.00000193, 0.00001285, -0.00002700, 0.00002177, -0.00001550, -0.00000275, 0.00001348, -0.00002798, 0.00002985, -0.00002996, 0.00001818, -0.00000487, -0.00000528, 0.00000658, 0.00000992, -0.00001726, 0.00001157, -0.00001194, 0.00000206, -0.00000083, -0.00001191, 0.00000416, 0.00001419, -0.00002898, 0.00000163, 0.00001023, -0.00001254, 0.00000026, 0.00001622, -0.00000066, 0.00001001, 0.00000124, -0.00000581, 0.00000511, 0.00000242, -0.00000825, -0.00000456, 0.00000681, 0.00000774, -0.00000382, -0.00000509, 0.00000399, -0.00000239, 0.00000110, -0.00000971, 0.00000107, 0.00001056, -0.00000203, -0.00001821, 0.00000802, -0.00000409, 0.00002052, -0.00000858, -0.00004067, -0.00012275, 0.00002852, 0.00009213, 0.00002062, 0.00002521, 0.00001431, 0.00000711, 0.00001683, 0.00001172, 0.00000634, 0.00001971, 0.00001229, 0.00001204, 0.00001836, 0.00001515, },
		{ 0.00001451, 0.00002373, 0.00003895, 0.00004519, 0.00000843, -0.00003602, -0.00003767, -0.00001589, -0.00001348, -0.00000884, -0.00000649, -0.00000705, -0.00000748, -0.00000146, -0.00000533, -0.00000228, -0.00000157, -0.00000197, -0.00000235, -0.00000004, -0.00000539, -0.00000123, 0.00000007, -0.00000150, -0.00000576, -0.00000217, -0.00000110, -0.00000332, -0.00000544, 0.00000363, 0.00000521, 0.00000022, -0.00000625, -0.00000853, -0.00000060, 0.00000614, -0.00000254, 0.00000062, 0.00000056, -0.00000963, 0.00000423, 0.00000767, -0.00000936, -0.00000068, 0.00000190, -0.00000555, 0.00000337, -0.00000481, -0.00000706, 0.00001347, 0.00000348, -0.00000999, -0.00000217, 0.00000194, 0.00000187, 0.00000859, -0.00000597, -0.00000615, -0.00000357, 0.00000837, -0.00000371, 0.00002212, -0.00002653, 0.00001123, -0.00000029, -0.00001672, 0.00002234, -0.00003094, 0.00002490, -0.00001861, 0.00000484, 0.00000310, -0.00000452, -0.00000180, 0.00001750, -0.00001716, 0.00000649, -0.00000455, -0.00000429, 0.00000116, -0.00000727, -0.00000689, 0.00001973, -0.00002036, -0.00001294, 0.00001344, -0.00000487, -0.00000956, 0.00001666, 0.00000525, 0.00000569, 0.00000398, -0.00000392, 0.00000208, 0.00000533, -0.00000477, -0.00000845, 0.00000546, 0.00000981, -0.00000372, -0.00000580, 0.00000521, -0.00000256, 0.00000272, -0.00000916, -0.00000159, 0.00001128, -0.00000018, -0.00001859, 0.00001021, -0.00000493, 0.00001856, -0.00000243, -0.00003883, -0.00012128, 0.00002377, 0.00008920, 0.00002027, 0.00002377, 0.00001398, 0.00000697, 0.00001519, 0.00001031, 0.00000648, 0.00001779, 0.00001195, 0.00001089, 0.00001520, },
		{ 0.00001249, 0.00001292, 0.00002169, 0.00002982, 0.00004412, 0.00002476, -0.00002194, -0.00004026, -0.00002266, -0.00001350, -0.00000792, -0.00000775, -0.00000605, -0.00000805, -0.00000253, -0.00000400, -0.00000411, -0.00000258, 0.00000068, -0.00000341, -0.00000032, -0.00000366, -0.00000299, 0.00000005, 0.00000001, -0.00000627, -0.00000217, -0.00000097, -0.00000252, -0.00000611, 0.00000024, 0.00000615, 0.00000249, -0.00000500, -0.00000742, -0.00000451, 0.00000544, 0.00000080, -0.00000210, 0.00000328, -0.00000793, -0.00000341, 0.00001149, -0.00000435, -0.00000674, 0.00000417, -0.00000471, 0.00000034, 0.00000065, -0.00001052, 0.00000747, 0.00001111, -0.00000826, -0.00000651, 0.00000280, -0.00000046, 0.00000821, -0.00000079, -0.00000586, -0.00000900, 0.00001148, -0.00000959, 0.00002564, -0.00001702, -0.00000439, 0.00001372, -0.00002474, 0.00002341, -0.00002637, 0.00001483, -0.00000704, -0.00000309, 0.00000285, 0.00000222, -0.00000987, 0.00001901, -0.00001081, -0.00000083, 0.00000072, -0.00000536, -0.00000260, -0.00000001, -0.00001360, 0.00001602, -0.00000697, -0.00002171, 0.00000701, 0.00000573, -0.00001451, 0.00001071, 0.00001273, 0.00000383, 0.00000332, 0.00000001, -0.00000017, 0.00000533, 0.00000036, -0.00001035, 0.00000229, 0.00001142, -0.00000256, -0.00000699, 0.00000624, -0.00000245, 0.00000364, -0.00000751, -0.00000385, 0.00001046, 0.00000240, -0.00001889, 0.00001146, -0.00000375, 0.00001587, 0.00000176, -0.00003524, -0.00011968, 0.00001826, 0.00008675, 0.00001998, 0.00002235, 0.00001343, 0.00000663, 0.00001466, 0.00000855, 0.00000613, 0.00001654, 0.00001114, 0.00001035, },
		{ 0.00000993, 0.00001058, 0.00001052, 0.00002093, 0.00002307, 0.00003855, 0.00003488, -0.00000543, -0.00003668, -0.00002960, -0.00001633, -0.00000597, -0.00000874, -0.00000601, -0.00000742, -0.00000420, -0.00000303, -0.00000438, -0.00000519, 0.00000293, -0.00000275, -0.00000202, -0.00000172, -0.00000354, -0.00000136, 0.00000176, -0.00000623, -0.00000256, -0.00000072, -0.00000217, -0.00000550, -0.00000294, 0.00000504, 0.00000526, -0.00000384, -0.00000572, -0.00000651, 0.00000194, 0.00000439, -0.00000298, 0.00000281, -0.00000323, -0.00000951, 0.00000952, 0.00000387, -0.00001064, 0.00000240, -0.00000093, -0.00000356, 0.00000377, -0.00000873, -0.00000102, 0.00001473, -0.00000206, -0.00001082, 0.00000211, -0.00000083, 0.00000570, 0.00000270, -0.00000135, -0.00001580, 0.00001357, -0.00001189, 0.00002115, -0.00000109, -0.00001914, 0.00002130, -0.00002465, 0.00001779, -0.00001789, 0.00000476, -0.00000004, -0.00000377, -0.00000306, 0.00000931, -0.00001271, 0.00001380, -0.00000195, -0.00000550, 0.00000091, -0.00000163, -0.00000827, 0.00000452, -0.00001315, 0.00000632, 0.00000393, -0.00002083, -0.00000537, 0.00001271, -0.00001197, 0.00000094, 0.00001767, 0.00000634, 0.00000018, 0.00000362, 0.00000017, 0.00000276, 0.00000514, -0.00000935, -0.00000173, 0.00001164, -0.00000044, -0.00000814, 0.00000673, -0.00000190, 0.00000394, -0.00000548, -0.00000493, 0.00000836, 0.00000475, -0.00001830, 0.00001154, -0.00000144, 0.00001402, 0.00000365, -0.00003116, -0.00011682, 0.00001201, 0.00008420, 0.00002012, 0.00002082, 0.00001287, 0.00000587, 0.00001482, 0.00000718, 0.00000509, 0.00001565, 0.00001026, },
		{ 0.00000953, 0.00000945, 0.00000943, 0.00000773, 0.00002040, 0.00001885, 0.00003132, 0.00003874, 0.00000979, -0.00002745, -0.00003371, -0.00002213, -0.00000448, -0.00000833, -0.00000719, -0.00000621, -0.00000541, -0.00000333, -0.00000313, -0.00000813, 0.00000338, -0.00000023, -0.00000408, -0.00000087, -0.00000239, -0.00000353, 0.00000280, -0.00000523, -0.00000332, -0.00000036, -0.00000221, -0.00000441, -0.00000476, 0.00000201, 0.00000748, -0.00000199, -0.00000446, -0.00000618, -0.00000274, 0.00000593, -0.00000102, 0.00000017, 0.00000136, -0.00001119, 0.00000265, 0.00001126, -0.00000951, -0.00000297, 0.00000322, -0.00000532, 0.00000288, -0.00000338, -0.00000787, 0.00001242, 0.00000640, -0.00001232, -0.00000107, 0.00000007, 0.00000339, 0.00000247, 0.00000631, -0.00002083, 0.00001163, -0.00000864, 0.00001011, 0.00001562, -0.00002726, 0.00002024, -0.00001807, 0.00000936, -0.00001006, -0.00000136, 0.00000084, 0.00000071, -0.00000971, 0.00001208, -0.00000921, 0.00000525, 0.00000434, -0.00000462, -0.00000324, 0.00000335, -0.00001144, 0.00000372, -0.00000735, -0.00000337, 0.00000730, -0.00001187, -0.00001690, 0.00001163, -0.00000369, -0.00000778, 0.00001704, 0.00001244, -0.00000274, 0.00000469, 0.00000336, -0.00000060, 0.00000771, -0.00000570, -0.00000506, 0.00001001, 0.00000207, -0.00000868, 0.00000655, -0.00000104, 0.00000398, -0.00000374, -0.00000464, 0.00000582, 0.00000598, -0.00001664, 0.00001098, 0.00000082, 0.00001381, 0.00000403, -0.00002809, -0.00011234, 0.00000578, 0.00008086, 0.00002081, 0.00001930, 0.00001228, 0.00000478, 0.00001515, 0.00000654, 0.00000368, 0.00001467, },
		{ 0.00001340, 0.00000884, 0.00000890, 0.00000888, 0.00000494, 0.00001957, 0.00001662, 0.00002435, 0.00003783, 0.00002121, -0.00001502, -0.00003280, -0.00002929, -0.00000531, -0.00000622, -0.00000880, -0.00000556, -0.00000530, -0.00000489, -0.00000138, -0.00000987, 0.00000151, 0.00000309, -0.00000496, -0.00000189, -0.00000006, -0.00000521, 0.00000236, -0.00000338, -0.00000403, -0.00000016, -0.00000228, -0.00000366, -0.00000484, -0.00000181, 0.00000790, 0.00000088, -0.00000376, -0.00000438, -0.00000643, 0.00000422, 0.00000279, -0.00000216, 0.00000323, -0.00000815, -0.00000566, 0.00001394, -0.00000314, -0.00000900, 0.00000457, -0.00000340, -0.00000102, 0.00000203, -0.00000997, 0.00000523, 0.00001335, -0.00000921, -0.00000572, 0.00000019, 0.00000333, -0.00000139, 0.00001390, -0.00002072, 0.00000440, -0.00000095, -0.00000303, 0.00002714, -0.00002595, 0.00001241, -0.00000937, 0.00000256, -0.00000606, -0.00000249, -0.00000265, 0.00000615, -0.00001285, 0.00000917, -0.00000211, -0.00000170, 0.00000488, 0.00000097, -0.00000807, 0.00000565, -0.00001004, -0.00000114, -0.00000079, -0.00000810, 0.00000282, -0.00000052, -0.00002159, 0.00000265, 0.00000519, -0.00001117, 0.00001087, 0.00001901, -0.00000265, 0.00000262, 0.00000788, -0.00000252, 0.00000731, -0.00000086, -0.00000628, 0.00000677, 0.00000408, -0.00000827, 0.00000585, -0.00000019, 0.00000416, -0.00000265, -0.00000339, 0.00000381, 0.00000574, -0.00001448, 0.00001065, 0.00000228, 0.00001508, 0.00000428, -0.00002697, -0.00010685, 0.00000065, 0.00007642, 0.00002173, 0.00001808, 0.00001159, 0.00000352, 0.00001534, 0.00000654, 0.00000238, },
		{ 0.00000151, 0.00001199, 0.00000802, 0.00000824, 0.00000884, 0.00000238, 0.00001823, 0.00001571, 0.00001855, 0.00003405, 0.00002812, -0.00000251, -0.00002661, -0.00003499, -0.00000944, -0.00000346, -0.00000945, -0.00000634, -0.00000395, -0.00000676, -0.00000049, -0.00000960, -0.00000191, 0.00000545, -0.00000363, -0.00000434, 0.00000206, -0.00000524, 0.00000043, -0.00000143, -0.00000404, -0.00000042, -0.00000208, -0.00000357, -0.00000374, -0.00000485, 0.00000586, 0.00000418, -0.00000287, -0.00000252, -0.00000771, -0.00000005, 0.00000602, -0.00000199, 0.00000176, -0.00000273, -0.00001132, 0.00001039, 0.00000563, -0.00001179, 0.00000149, 0.00000113, -0.00000499, 0.00000428, -0.00000689, -0.00000331, 0.00001541, -0.00000206, -0.00000917, -0.00000242, 0.00000577, -0.00000644, 0.00001760, -0.00001396, -0.00000644, 0.00000748, -0.00001309, 0.00002995, -0.00001665, 0.00000265, -0.00000318, 0.00000002, -0.00000624, -0.00000036, -0.00000701, 0.00000889, -0.00001105, 0.00000272, 0.00000414, -0.00000361, 0.00000006, 0.00000752, -0.00000969, 0.00000360, -0.00000538, -0.00000636, 0.00000239, -0.00000662, -0.00000560, 0.00000697, -0.00001751, -0.00000966, 0.00000915, -0.00000813, 0.00000227, 0.00002236, 0.00000154, -0.00000110, 0.00001136, -0.00000164, 0.00000477, 0.00000318, -0.00000473, 0.00000301, 0.00000475, -0.00000697, 0.00000509, 0.00000030, 0.00000469, -0.00000215, -0.00000187, 0.00000301, 0.00000439, -0.00001271, 0.00001113, 0.00000292, 0.00001700, 0.00000558, -0.00002777, -0.00010162, -0.00000265, 0.00007124, 0.00002229, 0.00001736, 0.00001087, 0.00000213, 0.00001532, 0.00000685, },
		{ 0.00000722, 0.00000105, 0.00001074, 0.00000705, 0.00000726, 0.00000921, 0.00000033, 0.00001631, 0.00001556, 0.00001424, 0.00002886, 0.00003121, 0.00000773, -0.00001689, -0.00003660, -0.00001613, -0.00000206, -0.00000809, -0.00000834, -0.00000242, -0.00000764, -0.00000125, -0.00000778, -0.00000514, 0.00000541, -0.00000033, -0.00000674, 0.00000251, -0.00000333, -0.00000206, -0.00000050, -0.00000296, -0.00000107, -0.00000170, -0.00000387, -0.00000252, -0.00000602, 0.00000190, 0.00000649, -0.00000091, -0.00000142, -0.00000670, -0.00000477, 0.00000642, 0.00000113, -0.00000133, 0.00000170, -0.00001190, 0.00000243, 0.00001229, -0.00000888, -0.00000484, 0.00000522, -0.00000608, 0.00000239, -0.00000106, -0.00000911, 0.00001153, 0.00000612, -0.00000859, -0.00000792, 0.00000875, -0.00000929, 0.00001511, -0.00000213, -0.00001683, 0.00001241, -0.00001686, 0.00002453, -0.00000419, -0.00000382, -0.00000199, 0.00000142, -0.00000856, 0.00000208, -0.00000931, 0.00000760, -0.00000596, -0.00000337, 0.00000617, -0.00000034, -0.00000665, 0.00001087, -0.00000648, -0.00000128, -0.00000085, -0.00000854, 0.00000092, -0.00000144, -0.00001260, 0.00000724, -0.00000750, -0.00001905, 0.00000552, -0.00000137, -0.00000438, 0.00002043, 0.00000843, -0.00000377, 0.00001209, 0.00000174, 0.00000204, 0.00000484, -0.00000098, 0.00000020, 0.00000369, -0.00000531, 0.00000476, 0.00000026, 0.00000553, -0.00000189, -0.00000074, 0.00000351, 0.00000278, -0.00001204, 0.00001239, 0.00000330, 0.00001862, 0.00000841, -0.00002960, -0.00009786, -0.00000412, 0.00006617, 0.00002200, 0.00001704, 0.00001032, 0.00000055, 0.00001513, },
		{ 0.00001476, 0.00000757, 0.00000080, 0.00000985, 0.00000617, 0.00000578, 0.00000971, -0.00000094, 0.00001385, 0.00001561, 0.00001155, 0.00002315, 0.00003157, 0.00001486, -0.00000645, -0.00003307, -0.00002289, -0.00000385, -0.00000496, -0.00001026, -0.00000206, -0.00000686, -0.00000324, -0.00000587, -0.00000656, 0.00000280, 0.00000338, -0.00000732, 0.00000075, -0.00000042, -0.00000361, -0.00000130, -0.00000106, -0.00000156, -0.00000156, -0.00000401, -0.00000204, -0.00000525, -0.00000241, 0.00000644, 0.00000225, -0.00000080, -0.00000473, -0.00000769, 0.00000333, 0.00000551, -0.00000329, 0.00000283, -0.00000794, -0.00000582, 0.00001318, -0.00000093, -0.00001084, 0.00000576, -0.00000336, -0.00000194, 0.00000379, -0.00000968, 0.00000370, 0.00001136, -0.00000296, -0.00001412, 0.00000922, -0.00000757, 0.00000712, 0.00001054, -0.00002235, 0.00001152, -0.00001456, 0.00001474, 0.00000586, -0.00000411, -0.00000498, 0.00000417, -0.00001023, 0.00000263, -0.00000863, 0.00000357, -0.00000082, -0.00000602, 0.00000351, 0.00000524, -0.00001108, 0.00000900, -0.00000003, -0.00000551, 0.00000058, -0.00000678, -0.00000326, 0.00000326, -0.00001462, 0.00000128, 0.00000282, -0.00002112, -0.00000404, 0.00000421, -0.00000604, 0.00001406, 0.00001484, -0.00000314, 0.00001007, 0.00000586, 0.00000108, 0.00000384, 0.00000341, -0.00000043, 0.00000131, -0.00000412, 0.00000520, -0.00000021, 0.00000639, -0.00000145, -0.00000031, 0.00000488, 0.00000181, -0.00001267, 0.00001380, 0.00000414, 0.00001950, 0.00001242, -0.00003123, -0.00009614, -0.00000459, 0.00006211, 0.00002084, 0.00001668, 0.00001020, -0.00000118, },
		{ -0.00000290, 0.00001415, 0.00000793, 0.00000058, 0.00000924, 0.00000572, 0.00000386, 0.00000986, -0.00000125, 0.00001120, 0.00001523, 0.00001046, 0.00001756, 0.00002998, 0.00001939, 0.00000232, -0.00002553, -0.00002669, -0.00000918, -0.00000179, -0.00001044, -0.00000352, -0.00000493, -0.00000511, -0.00000536, -0.00000567, -0.00000104, 0.00000541, -0.00000524, -0.00000233, 0.00000175, -0.00000310, -0.00000358, 0.00000067, -0.00000115, -0.00000200, -0.00000368, -0.00000242, -0.00000350, -0.00000533, 0.00000365, 0.00000555, 0.00000044, -0.00000328, -0.00000783, -0.00000189, 0.00000840, -0.00000221, 0.00000078, -0.00000246, -0.00001026, 0.00000765, 0.00000838, -0.00001258, 0.00000170, 0.00000141, -0.00000543, 0.00000486, -0.00000552, -0.00000410, 0.00001093, 0.00000594, -0.00001741, 0.00000498, -0.00000159, -0.00000285, 0.00001917, -0.00002065, 0.00000572, -0.00000932, 0.00000548, 0.00000990, 0.00000111, -0.00000890, 0.00000531, -0.00000965, 0.00000102, -0.00000612, -0.00000064, 0.00000168, -0.00000455, -0.00000155, 0.00000933, -0.00001086, 0.00000312, 0.00000602, -0.00000612, -0.00000169, -0.00000296, -0.00000678, 0.00000444, -0.00001160, -0.00000676, 0.00000838, -0.00001559, -0.00001463, 0.00000439, -0.00000282, 0.00000646, 0.00001772, 0.00000117, 0.00000710, 0.00000857, 0.00000278, 0.00000130, 0.00000657, 0.00000146, -0.00000130, -0.00000413, 0.00000634, -0.00000074, 0.00000694, -0.00000059, -0.00000051, 0.00000639, 0.00000204, -0.00001412, 0.00001455, 0.00000577, 0.00001979, 0.00001685, -0.00003180, -0.00009619, -0.00000522, 0.00005950, 0.00001934, 0.00001576, 0.00001058, },
		{ 0.00001117, -0.00000432, 0.00001323, 0.00000827, 0.00000039, 0.00000863, 0.00000591, 0.00000191, 0.00000916, -0.00000079, 0.00000897, 0.00001393, 0.00001063, 0.00001283, 0.00002674, 0.00002222, 0.00000852, -0.00001665, -0.00002553, -0.00001631, -0.00000101, -0.00000807, -0.00000605, -0.00000336, -0.00000545, -0.00000677, -0.00000343, -0.00000402, 0.00000447, -0.00000124, -0.00000481, 0.00000171, -0.00000057, -0.00000605, 0.00000101, 0.00000052, -0.00000287, -0.00000313, -0.00000310, -0.00000208, -0.00000593, -0.00000083, 0.00000723, 0.00000322, -0.00000287, -0.00000588, -0.00000668, 0.00000780, 0.00000164, -0.00000211, 0.00000105, -0.00000912, -0.00000155, 0.00001429, -0.00000821, -0.00000495, 0.00000481, -0.00000542, 0.00000191, 0.00000035, -0.00000814, 0.00000497, 0.00001410, -0.00001480, -0.00000356, 0.00000563, -0.00001031, 0.00002077, -0.00001284, -0.00000140, -0.00000514, 0.00000015, 0.00000778, 0.00000834, -0.00001018, 0.00000348, -0.00000722, -0.00000137, -0.00000378, -0.00000297, 0.00000086, -0.00000071, -0.00000568, 0.00000951, -0.00000661, -0.00000334, 0.00000836, -0.00000254, -0.00000561, -0.00000015, -0.00000720, 0.00000191, -0.00000642, -0.00001238, 0.00000732, -0.00000608, -0.00002088, -0.00000184, 0.00000211, 0.00000122, 0.00001589, 0.00000737, 0.00000567, 0.00000866, 0.00000643, -0.00000086, 0.00000735, 0.00000506, -0.00000279, -0.00000557, 0.00000767, -0.00000087, 0.00000696, 0.00000066, -0.00000095, 0.00000744, 0.00000347, -0.00001553, 0.00001407, 0.00000799, 0.00002008, 0.00002098, -0.00003109, -0.00009711, -0.00000694, 0.00005806, 0.00001834, 0.00001401, },
		{ 0.00001163, 0.00001149, -0.00000524, 0.00001209, 0.00000849, 0.00000032, 0.00000781, 0.00000658, 0.00000053, 0.00000750, -0.00000020, 0.00000783, 0.00001165, 0.00001125, 0.00000972, 0.00002212, 0.00002377, 0.00001277, -0.00000918, -0.00001962, -0.00002206, -0.00000421, -0.00000407, -0.00000783, -0.00000357, -0.00000389, -0.00000917, -0.00000171, -0.00000455, 0.00000090, 0.00000260, -0.00000483, -0.00000073, 0.00000258, -0.00000704, -0.00000066, 0.00000295, -0.00000344, -0.00000293, -0.00000339, -0.00000175, -0.00000460, -0.00000500, 0.00000593, 0.00000728, -0.00000271, -0.00000361, -0.00000895, 0.00000372, 0.00000588, -0.00000293, 0.00000077, -0.00000385, -0.00000980, 0.00001361, 0.00000065, -0.00001013, 0.00000396, -0.00000163, -0.00000277, 0.00000431, -0.00000691, -0.00000343, 0.00001733, -0.00000595, -0.00001315, 0.00000999, -0.00001223, 0.00001578, -0.00000288, -0.00000570, -0.00000447, -0.00000087, 0.00000228, 0.00001347, -0.00000712, -0.00000032, -0.00000481, -0.00000282, -0.00000297, -0.00000296, -0.00000198, 0.00000270, -0.00000659, 0.00000597, -0.00000123, -0.00000695, 0.00000599, 0.00000310, -0.00000802, -0.00000044, -0.00000453, -0.00000209, -0.00000265, -0.00001327, 0.00000155, 0.00000219, -0.00002000, -0.00001181, 0.00000452, 0.00000013, 0.00001067, 0.00001254, 0.00000742, 0.00000664, 0.00001023, -0.00000104, 0.00000585, 0.00000880, -0.00000221, -0.00000794, 0.00000846, -0.00000032, 0.00000655, 0.00000203, -0.00000121, 0.00000772, 0.00000559, -0.00001609, 0.00001237, 0.00001008, 0.00002092, 0.00002461, -0.00002952, -0.00009789, -0.00001003, 0.00005690, 0.00001853, },
		{ 0.00002003, 0.00000930, 0.00001111, -0.00000574, 0.00001099, 0.00000841, 0.00000046, 0.00000675, 0.00000731, 0.00000014, 0.00000527, -0.00000034, 0.00000803, 0.00000897, 0.00001125, 0.00000863, 0.00001682, 0.00002370, 0.00001633, -0.00000450, -0.00001133, -0.00002343, -0.00001092, -0.00000088, -0.00000706, -0.00000584, -0.00000151, -0.00001082, -0.00000198, -0.00000258, -0.00000342, 0.00000402, -0.00000189, -0.00000416, 0.00000434, -0.00000554, -0.00000374, 0.00000490, -0.00000288, -0.00000345, -0.00000305, -0.00000233, -0.00000274, -0.00000705, 0.00000170, 0.00001105, -0.00000138, -0.00000241, -0.00000826, -0.00000173, 0.00000759, -0.00000026, -0.00000235, 0.00000197, -0.00001321, 0.00000664, 0.00000960, -0.00001003, -0.00000153, 0.00000360, -0.00000586, 0.00000425, -0.00000195, -0.00000992, 0.00001360, 0.00000617, -0.00001923, 0.00000875, -0.00000876, 0.00000768, 0.00000452, -0.00000486, -0.00000690, 0.00000027, -0.00000293, 0.00001404, -0.00000075, -0.00000351, -0.00000425, -0.00000255, -0.00000363, -0.00000168, -0.00000472, 0.00000368, -0.00000418, 0.00000110, 0.00000208, -0.00000627, 0.00000070, 0.00000741, -0.00000682, -0.00000358, -0.00000092, -0.00000465, -0.00000228, -0.00001028, -0.00000489, 0.00000539, -0.00001301, -0.00002072, 0.00000166, 0.00000218, 0.00000498, 0.00001436, 0.00001220, 0.00000435, 0.00001239, 0.00000117, 0.00000339, 0.00001111, 0.00000046, -0.00001018, 0.00000816, 0.00000080, 0.00000603, 0.00000315, -0.00000102, 0.00000739, 0.00000769, -0.00001534, 0.00001003, 0.00001130, 0.00002246, 0.00002805, -0.00002778, -0.00009796, -0.00001403, 0.00005502, },
		{ 0.00005176, 0.00002235, 0.00000777, 0.00000995, -0.00000617, 0.00001024, 0.00000797, 0.00000074, 0.00000568, 0.00000764, 0.00000074, 0.00000329, -0.00000180, 0.00000915, 0.00000691, 0.00000988, 0.00000917, 0.00001212, 0.00002142, 0.00001995, -0.00000204, -0.00000393, -0.00001930, -0.00001841, -0.00000115, -0.00000337, -0.00000877, -0.00000019, -0.00001037, -0.00000432, 0.00000034, -0.00000609, 0.00000201, 0.00000267, -0.00000624, 0.00000331, -0.00000202, -0.00000658, 0.00000506, -0.00000094, -0.00000437, -0.00000255, -0.00000294, -0.00000178, -0.00000636, -0.00000383, 0.00001248, 0.00000188, -0.00000234, -0.00000601, -0.00000569, 0.00000518, 0.00000480, -0.00000540, 0.00000488, -0.00001084, -0.00000292, 0.00001374, -0.00000348, -0.00000899, 0.00000676, -0.00000522, 0.00000078, 0.00000324, -0.00001126, 0.00000430, 0.00001656, -0.00001842, 0.00000229, -0.00000304, 0.00000098, 0.00000648, 0.00000031, -0.00000979, 0.00000074, -0.00000546, 0.00001035, 0.00000594, -0.00000369, -0.00000596, -0.00000128, -0.00000460, -0.00000067, -0.00000583, 0.00000198, -0.00000030, -0.00000229, 0.00000172, -0.00000250, -0.00000424, 0.00000788, -0.00000226, -0.00000726, 0.00000102, -0.00000424, -0.00000471, -0.00000632, -0.00000842, 0.00000294, -0.00000391, -0.00002438, -0.00000607, 0.00000423, 0.00000139, 0.00001232, 0.00001813, 0.00000389, 0.00001224, 0.00000483, 0.00000166, 0.00001124, 0.00000429, -0.00001116, 0.00000670, 0.00000203, 0.00000580, 0.00000384, -0.00000042, 0.00000686, 0.00000918, -0.00001343, 0.00000796, 0.00001122, 0.00002428, 0.00003187, -0.00002631, -0.00009737, -0.00001807, },
		{ -0.00002138, 0.00004714, 0.00002456, 0.00000755, 0.00000839, -0.00000704, 0.00000995, 0.00000731, 0.00000094, 0.00000485, 0.00000738, 0.00000183, 0.00000238, -0.00000454, 0.00001015, 0.00000636, 0.00000716, 0.00001015, 0.00000933, 0.00001692, 0.00002311, -0.00000000, 0.00000027, -0.00001120, -0.00002300, -0.00000602, 0.00000149, -0.00001005, -0.00000132, -0.00000792, -0.00000725, 0.00000204, -0.00000576, -0.00000245, 0.00000627, -0.00000518, -0.00000021, 0.00000164, -0.00000738, 0.00000304, 0.00000161, -0.00000475, -0.00000266, -0.00000278, -0.00000229, -0.00000390, -0.00000828, 0.00001037, 0.00000630, -0.00000201, -0.00000428, -0.00000638, -0.00000049, 0.00000926, -0.00000550, 0.00000379, -0.00000507, -0.00001003, 0.00001042, 0.00000709, -0.00001412, 0.00000539, -0.00000127, -0.00000337, 0.00000537, -0.00000714, -0.00000632, 0.00002064, -0.00001057, -0.00000589, 0.00000074, -0.00000136, 0.00000332, 0.00000654, -0.00001012, -0.00000097, -0.00000530, 0.00000482, 0.00000990, -0.00000023, -0.00000858, -0.00000053, -0.00000470, -0.00000079, -0.00000516, -0.00000108, 0.00000266, -0.00000258, -0.00000154, 0.00000152, -0.00000609, 0.00000438, 0.00000322, -0.00000879, 0.00000003, -0.00000155, -0.00000764, -0.00000426, -0.00000784, -0.00000272, 0.00000278, -0.00002147, -0.00001538, 0.00000306, 0.00000062, 0.00000786, 0.00002275, 0.00000637, 0.00001053, 0.00000827, 0.00000186, 0.00000961, 0.00000784, -0.00001028, 0.00000465, 0.00000279, 0.00000607, 0.00000416, 0.00000030, 0.00000657, 0.00000986, -0.00001096, 0.00000690, 0.00001006, 0.00002562, 0.00003647, -0.00002509, -0.00009661, },
		},
		{ { -0.00000249, -0.00000221, -0.00000472, -0.00000556, 0.00000064, 0.00000224, -0.00000297, -0.00000390, -0.00000270, -0.00000349, -0.00000229, -0.00000216, -0.00000227, 0.00000255, 0.00000016, -0.00000811, -0.00000489, -0.00000476, -0.00001176, -0.00000769, -0.00001105, -0.00001863, -0.00000666, -0.00000075, 0.00000521, 0.00002503, 0.00001550, -0.00000208, 0.00000765, 0.00000725, 0.00000290, 0.00000651, -0.00000025, 0.00000447, 0.00000695, -0.00000684, 0.00000073, 0.00000873, 0.00000167, 0.00000436, -0.00000081, -0.00000451, 0.00000691, 0.00000531, -0.00000062, -0.00000119, 0.00000132, 0.00000808, -0.00000210, -0.00000602, 0.00000971, 0.00000604, -0.00000114, -0.00000059, -0.00000110, 0.00000884, -0.00000530, -0.00001771, 0.00001597, 0.00002053, -0.00000483, -0.00001351, -0.00002468, 0.00000503, 0.00004268, 0.00000503, -0.00002468, -0.00001351, -0.00000483, 0.00002053, 0.00001597, -0.00001771, -0.00000530, 0.00000884, -0.00000110, -0.00000059, -0.00000114, 0.00000604, 0.00000971, -0.00000602, -0.00000210, 0.00000808, 0.00000132, -0.00000119, -0.00000062, 0.00000531, 0.00000691, -0.00000451, -0.00000081, 0.00000436, 0.00000167, 0.00000873, 0.00000073, -0.00000684, 0.00000695, 0.00000447, -0.00000025, 0.00000651, 0.00000290, 0.00000725, 0.00000765, -0.00000208, 0.00001550, 0.00002503, 0.00000521, -0.00000075, -0.00000666, -0.00001863, -0.00001105, -0.00000769, -0.00001176, -0.00000476, -0.00000489, -0.00000811, 0.00000016, 0.00000255, -0.00000227, -0.00000216, -0.00000229, -0.00000349, -0.00000270, -0.00000390, -0.00000297, 0.00000224, 0.00000064, -0.00000556, -0.00000472, -0.00000221, },
		{ -0.00000203, -0.00000221, -0.00000244, -0.00000371, -0.00000685, -0.00000269, 0.00000321, 0.00000030, -0.00000334, -0.00000322, -0.00000403, -0.00000295, -0.00000182, -0.00000372, 0.00000049, 0.00000397, -0.00000454, -0.00000671, -0.00000320, -0.00000985, -0.00000930, -0.00000853, -0.00001953, -0.00001080, 0.00000201, -0.00000015, 0.00001612, 0.00002231, 0.00000155, 0.00000344, 0.00000896, 0.00000172, 0.00000712, 0.00000302, -0.00000110, 0.00000797, -0.00000136, -0.00000442, 0.00000583, 0.00000116, 0.00000519, 0.00000702, -0.00000651, -0.00000119, 0.00000750, 0.00000425, 0.00000053, -0.00000345, 0.00000488, 0.00000403, -0.00001082, 0.00000043, 0.00001260, 0.00000745, 0.00000096, -0.00001175, 0.00000022, 0.00001414, -0.00000987, -0.00000870, 0.00000899, 0.00000600, 0.00001284, -0.00001078, -0.00003556, 0.00001111, 0.00003643, 0.00000878, -0.00001356, -0.00003149, -0.00000491, 0.00003567, 0.00000893, -0.00001664, -0.00000479, 0.00000114, 0.00000799, 0.00000367, -0.00000427, 0.00000557, 0.00000175, -0.00000308, 0.00000658, 0.00000521, 0.00000187, -0.00000046, -0.00000111, 0.00000659, 0.00000260, -0.00000139, 0.00000169, -0.00000200, 0.00000678, 0.00000936, -0.00000596, 0.00000043, 0.00000799, 0.00000118, 0.00000498, 0.00000373, 0.00000428, 0.00001115, 0.00000024, 0.00000564, 0.00002649, 0.00001584, 0.00000026, -0.00000574, -0.00001799, -0.00001417, -0.00000556, -0.00001120, -0.00000816, -0.00000337, -0.00000818, -0.00000450, 0.00000147, -0.00000093, -0.00000176, -0.00000157, -0.00000335, -0.00000248, -0.00000267, -0.00000443, -0.00000082, 0.00000157, -0.00000341, -0.00000501, },
		{ -0.00000527, -0.00000198, -0.00000100, -0.00000261, -0.00000296, -0.00000618, -0.00000588, 0.00000103, 0.00000237, -0.00000122, -0.00000226, -0.00000410, -0.00000454, -0.00000176, -0.00000333, -0.00000263, 0.00000402, 0.00000011, -0.00000622, -0.00000281, -0.00000746, -0.00001076, -0.00000588, -0.00001704, -0.00001809, 0.00000130, 0.00000145, 0.00000710, 0.00002270, 0.00000731, 0.00000008, 0.00001064, 0.00000231, 0.00000340, 0.00000787, -0.00000160, 0.00000384, 0.00000271, -0.00000532, 0.00000429, 0.00000077, -0.00000121, 0.00001158, 0.00000203, -0.00000629, 0.00000205, 0.00000501, 0.00000595, -0.00000127, -0.00000207, 0.00000626, -0.00000655, -0.00000947, 0.00000672, 0.00001151, 0.00001247, -0.00000569, -0.00001819, 0.00000830, 0.00000997, -0.00000609, -0.00000327, -0.00000977, 0.00001019, 0.00002674, -0.00002028, -0.00003243, 0.00000977, 0.00002921, 0.00002342, -0.00001936, -0.00004329, 0.00000934, 0.00003450, 0.00000440, -0.00001070, -0.00001245, 0.00000276, 0.00001682, 0.00000160, -0.00000394, -0.00000012, -0.00000194, 0.00000634, 0.00000745, 0.00000331, 0.00000384, -0.00000135, 0.00000069, 0.00000367, 0.00000157, 0.00000411, -0.00000302, -0.00000176, 0.00001174, 0.00000223, -0.00000331, 0.00000655, 0.00000334, 0.00000438, 0.00000595, 0.00000146, 0.00000994, 0.00000651, 0.00000028, 0.00002016, 0.00002417, 0.00000662, -0.00000367, -0.00001712, -0.00001933, -0.00000610, -0.00000745, -0.00001031, -0.00000423, -0.00000656, -0.00000713, -0.00000109, -0.00000115, -0.00000202, -0.00000065, -0.00000245, -0.00000293, -0.00000180, -0.00000370, -0.00000301, 0.00000050, -0.00000218, },
		{ -0.00000141, -0.00000600, -0.00000345, 0.00000050, -0.00000147, -0.00000282, -0.00000487, -0.00000711, -0.00000213, 0.00000208, -0.00000001, -0.00000067, -0.00000207, -0.00000533, -0.00000345, -0.00000251, -0.00000398, 0.00000172, 0.00000272, -0.00000455, -0.00000235, -0.00000385, -0.00001198, -0.00000610, -0.00001057, -0.00002213, -0.00000563, 0.00000380, 0.00000217, 0.00001947, 0.00001357, -0.00000330, 0.00000941, 0.00000781, -0.00000097, 0.00000766, 0.00000244, 0.00000098, 0.00000417, -0.00000615, 0.00000180, 0.00000572, -0.00000650, 0.00000499, 0.00001032, -0.00000175, -0.00000127, 0.00000019, 0.00000511, 0.00000619, -0.00000280, 0.00000238, -0.00000136, -0.00001197, -0.00000113, 0.00000585, 0.00001427, 0.00001306, -0.00001672, -0.00001256, 0.00001025, 0.00000539, 0.00000402, -0.00001296, -0.00001922, 0.00002476, 0.00002320, -0.00002219, -0.00002588, -0.00000126, 0.00003508, 0.00002944, -0.00003349, -0.00003488, 0.00001582, 0.00002461, 0.00000980, -0.00001186, -0.00001729, 0.00001135, 0.00001757, 0.00000098, -0.00000536, -0.00000733, 0.00000367, 0.00001162, 0.00000456, 0.00000486, 0.00000367, -0.00000028, 0.00000047, -0.00000020, 0.00000624, 0.00000354, -0.00000674, 0.00000433, 0.00000787, -0.00000102, 0.00000503, 0.00000468, 0.00000260, 0.00000825, 0.00000274, 0.00000537, 0.00001023, 0.00000128, 0.00001227, 0.00002668, 0.00001445, 0.00000060, -0.00001355, -0.00002359, -0.00001165, -0.00000463, -0.00000912, -0.00000554, -0.00000527, -0.00000814, -0.00000338, -0.00000171, -0.00000330, -0.00000120, -0.00000112, -0.00000263, -0.00000175, -0.00000263, -0.00000364, -0.00000064, },
		{ -0.00000169, -0.00000024, -0.00000561, -0.00000648, -0.00000012, 0.00000072, -0.00000190, -0.00000373, -0.00000705, -0.00000468, 0.00000098, 0.00000003, -0.00000087, 0.00000030, -0.00000321, -0.00000483, -0.00000324, -0.00000446, -0.00000045, 0.00000386, -0.00000316, -0.00000343, 0.00000046, -0.00000970, -0.00000963, -0.00000550, -0.00001944, -0.00001351, 0.00000271, -0.00000071, 0.00001355, 0.00002100, -0.00000290, 0.00000165, 0.00001350, 0.00000077, 0.00000348, 0.00000507, 0.00000025, 0.00000603, -0.00000473, -0.00000548, 0.00000950, -0.00000209, -0.00000440, 0.00000874, 0.00000383, 0.00000209, 0.00000004, -0.00000214, 0.00000705, 0.00000331, 0.00000083, 0.00000074, -0.00001185, -0.00000556, 0.00000213, 0.00000377, 0.00001976, 0.00000372, -0.00001932, -0.00000333, 0.00000442, 0.00000912, 0.00000521, -0.00002629, -0.00001075, 0.00002786, 0.00001510, -0.00001277, -0.00002994, -0.00000548, 0.00004465, 0.00001845, -0.00003325, -0.00002030, 0.00000677, 0.00002213, 0.00001457, -0.00001771, -0.00001196, 0.00001559, 0.00001487, 0.00000212, -0.00001124, -0.00000716, 0.00001129, 0.00000999, 0.00000436, 0.00000592, 0.00000355, 0.00000146, -0.00000328, 0.00000125, 0.00000931, -0.00000149, -0.00000283, 0.00000565, 0.00000085, 0.00000486, 0.00000811, 0.00000130, 0.00000675, 0.00000683, 0.00000312, 0.00000959, 0.00000438, 0.00000662, 0.00002500, 0.00002159, 0.00000551, -0.00000850, -0.00002347, -0.00001889, -0.00000645, -0.00000733, -0.00000569, -0.00000364, -0.00000800, -0.00000571, -0.00000207, -0.00000391, -0.00000312, -0.00000126, -0.00000189, -0.00000143, -0.00000170, -0.00000366, },
		{ -0.00000280, -0.00000299, 0.00000018, -0.00000326, -0.00000837, -0.00000346, 0.00000140, -0.00000018, -0.00000202, -0.00000590, -0.00000681, -0.00000055, 0.00000067, -0.00000213, -0.00000007, -0.00000029, -0.00000340, -0.00000368, -0.00000548, -0.00000320, 0.00000444, -0.00000040, -0.00000614, 0.00000136, -0.00000352, -0.00001146, -0.00000514, -0.00001392, -0.00001794, 0.00000005, -0.00000181, 0.00000342, 0.00002486, 0.00000625, -0.00000732, 0.00001134, 0.00000690, 0.00000151, 0.00000584, -0.00000096, 0.00000576, 0.00000219, -0.00001136, 0.00000411, 0.00000601, -0.00000591, 0.00000294, 0.00000320, 0.00000315, 0.00000709, -0.00000322, -0.00000045, 0.00000546, 0.00000340, 0.00000462, -0.00001024, -0.00001189, 0.00000355, -0.00000245, 0.00000783, 0.00001827, -0.00000714, -0.00001090, -0.00000105, 0.00000115, 0.00001303, -0.00000512, -0.00002474, -0.00000131, 0.00001732, 0.00001815, -0.00000767, -0.00003701, 0.00000253, 0.00003921, 0.00000844, -0.00001626, -0.00001911, -0.00000331, 0.00002639, 0.00001103, -0.00001652, -0.00000388, 0.00001270, 0.00001442, -0.00000035, -0.00001465, 0.00000010, 0.00001268, 0.00000766, 0.00000601, 0.00000542, 0.00000538, -0.00000017, -0.00000488, 0.00000584, 0.00000580, -0.00000182, 0.00000252, -0.00000060, 0.00000127, 0.00001149, 0.00000508, 0.00000343, 0.00000846, 0.00000438, 0.00000798, 0.00000711, 0.00000311, 0.00002011, 0.00002761, 0.00001194, -0.00000428, -0.00002042, -0.00002338, -0.00001111, -0.00000783, -0.00000652, -0.00000184, -0.00000564, -0.00000741, -0.00000332, -0.00000361, -0.00000439, -0.00000277, -0.00000233, -0.00000134, -0.00000047, },
		{ -0.00000015, -0.00000090, -0.00000324, -0.00000087, -0.00000086, -0.00000742, -0.00000690, -0.00000038, 0.00000042, -0.00000042, -0.00000311, -0.00000748, -0.00000336, 0.00000141, -0.00000173, -0.00000219, -0.00000009, -0.00000123, -0.00000180, -0.00000487, -0.00000679, 0.00000221, 0.00000393, -0.00000640, -0.00000247, 0.00000121, -0.00000874, -0.00000632, -0.00000920, -0.00001991, -0.00000364, 0.00000156, -0.00000647, 0.00001802, 0.00001920, -0.00000753, 0.00000267, 0.00001004, 0.00000191, 0.00000745, 0.00000015, 0.00000028, 0.00000854, -0.00000801, -0.00000553, 0.00000753, -0.00000230, 0.00000092, 0.00000257, -0.00000282, 0.00000924, 0.00000579, -0.00000382, 0.00000075, 0.00000220, 0.00000903, -0.00000157, -0.00001898, -0.00000020, 0.00000338, -0.00000556, 0.00001348, 0.00000788, -0.00000863, -0.00000168, -0.00000275, 0.00000263, 0.00000589, -0.00001008, -0.00001167, -0.00000643, 0.00001079, 0.00002577, -0.00001256, -0.00003197, 0.00000589, 0.00002227, 0.00001654, -0.00000285, -0.00002841, -0.00000354, 0.00002529, 0.00000719, -0.00000695, -0.00000227, 0.00000922, 0.00001339, -0.00000503, -0.00000974, 0.00000545, 0.00000954, 0.00000816, 0.00000619, 0.00000623, 0.00000590, -0.00000325, -0.00000109, 0.00000587, 0.00000174, 0.00000376, 0.00000063, -0.00000496, 0.00000801, 0.00001123, 0.00000421, 0.00000756, 0.00000593, 0.00000675, 0.00000971, 0.00000262, 0.00001253, 0.00002975, 0.00002097, 0.00000087, -0.00001744, -0.00002539, -0.00001542, -0.00000937, -0.00000914, -0.00000249, -0.00000210, -0.00000650, -0.00000479, -0.00000377, -0.00000467, -0.00000385, -0.00000360, -0.00000262, },
		{ -0.00000433, -0.00000178, 0.00000009, -0.00000185, -0.00000175, 0.00000009, -0.00000506, -0.00000851, -0.00000281, -0.00000027, -0.00000077, -0.00000048, -0.00000521, -0.00000587, 0.00000023, -0.00000010, -0.00000283, -0.00000167, -0.00000151, -0.00000025, -0.00000127, -0.00000755, -0.00000313, 0.00000558, -0.00000240, -0.00000618, 0.00000107, -0.00000495, -0.00000650, -0.00000439, -0.00001921, -0.00001094, 0.00000644, -0.00000752, 0.00000330, 0.00002424, 0.00000139, -0.00000392, 0.00000926, 0.00000124, 0.00000671, 0.00000660, -0.00000425, 0.00000670, 0.00000046, -0.00000955, 0.00000367, -0.00000074, -0.00000093, 0.00000722, -0.00000557, 0.00000012, 0.00001196, 0.00000257, -0.00000069, -0.00000311, 0.00000507, 0.00001093, -0.00001592, -0.00001319, 0.00000983, -0.00000445, -0.00000195, 0.00001198, -0.00000162, -0.00000246, 0.00000265, -0.00000121, -0.00000185, -0.00000498, 0.00000029, -0.00000677, -0.00001720, 0.00001492, 0.00002357, -0.00001320, -0.00002020, -0.00000700, 0.00001565, 0.00003335, -0.00000464, -0.00003198, -0.00000157, 0.00001767, 0.00001220, 0.00000094, -0.00000531, 0.00000755, 0.00000843, -0.00000428, -0.00000133, 0.00000464, 0.00000781, 0.00000880, 0.00000605, 0.00000787, 0.00000312, -0.00000202, 0.00000249, 0.00000118, 0.00000403, 0.00000641, -0.00000531, -0.00000107, 0.00001140, 0.00000819, 0.00000817, 0.00000767, 0.00000526, 0.00001066, 0.00000591, 0.00000602, 0.00002564, 0.00002930, 0.00000966, -0.00001354, -0.00002697, -0.00001967, -0.00001028, -0.00001122, -0.00000622, -0.00000095, -0.00000374, -0.00000442, -0.00000395, -0.00000499, -0.00000436, -0.00000424, },
		{ -0.00000455, -0.00000511, -0.00000387, -0.00000111, -0.00000087, -0.00000154, 0.00000052, -0.00000253, -0.00000880, -0.00000508, -0.00000049, -0.00000213, -0.00000079, -0.00000197, -0.00000559, -0.00000181, 0.00000043, -0.00000217, -0.00000210, -0.00000272, -0.00000188, 0.00000137, -0.00000363, -0.00000657, 0.00000259, 0.00000155, -0.00000634, -0.00000098, -0.00000277, -0.00000762, -0.00000054, -0.00001274, -0.00001919, 0.00000477, 0.00000019, -0.00000755, 0.00001815, 0.00001065, -0.00000589, 0.00000838, 0.00000245, 0.00000003, 0.00001164, -0.00000044, 0.00000065, 0.00000489, -0.00000812, 0.00000067, 0.00000107, -0.00000706, 0.00000872, 0.00000252, -0.00000849, 0.00000622, 0.00000799, 0.00000499, -0.00000157, -0.00000616, 0.00001310, 0.00000013, -0.00002341, 0.00000298, 0.00000833, -0.00000987, 0.00000243, 0.00000473, -0.00000364, 0.00000315, 0.00000700, -0.00000113, -0.00001509, -0.00000313, 0.00001245, -0.00001481, -0.00001784, 0.00001632, 0.00001596, -0.00000154, -0.00002032, -0.00002197, 0.00002607, 0.00003872, -0.00001023, -0.00002690, -0.00000692, 0.00001432, 0.00002169, 0.00000196, -0.00000707, 0.00000363, 0.00000435, 0.00000336, 0.00000261, 0.00000214, 0.00000799, 0.00000810, 0.00000750, 0.00000675, 0.00000151, 0.00000279, 0.00000047, -0.00000045, 0.00000872, 0.00000202, -0.00000567, 0.00000504, 0.00000880, 0.00000935, 0.00001122, 0.00000584, 0.00000875, 0.00000979, 0.00000432, 0.00001804, 0.00003245, 0.00002049, -0.00000610, -0.00002715, -0.00002509, -0.00001205, -0.00001111, -0.00000993, -0.00000322, -0.00000252, -0.00000290, -0.00000264, -0.00000468, -0.00000496, },
		{ -0.00000450, -0.00000494, -0.00000563, -0.00000503, -0.00000293, -0.00000159, -0.00000176, 0.00000066, 0.00000051, -0.00000749, -0.00000795, -0.00000077, -0.00000170, -0.00000297, -0.00000129, -0.00000378, -0.00000236, 0.00000055, -0.00000201, -0.00000213, -0.00000208, -0.00000406, -0.00000034, 0.00000022, -0.00000506, -0.00000055, 0.00000261, -0.00000536, -0.00000268, -0.00000024, -0.00000920, -0.00000216, -0.00000299, -0.00002040, -0.00000428, 0.00000700, -0.00000903, 0.00000813, 0.00001602, -0.00000649, 0.00000496, 0.00000960, -0.00000551, 0.00000705, 0.00000694, -0.00000090, 0.00000550, -0.00000633, -0.00000323, 0.00000676, -0.00000979, -0.00000060, 0.00001127, -0.00000519, -0.00000211, 0.00000454, 0.00000663, 0.00000935, -0.00000877, -0.00000040, 0.00001437, -0.00001686, -0.00001480, 0.00001479, -0.00000031, -0.00000996, 0.00000209, -0.00000033, -0.00000219, 0.00000790, 0.00001231, -0.00001104, -0.00002254, 0.00000987, 0.00001066, -0.00002017, -0.00001212, 0.00000731, 0.00001816, 0.00000823, -0.00003302, -0.00002033, 0.00003606, 0.00003154, -0.00000686, -0.00002443, -0.00001447, 0.00001945, 0.00002607, 0.00000079, -0.00000832, -0.00000259, 0.00000705, 0.00001064, 0.00000193, 0.00000186, 0.00000735, 0.00000824, 0.00000810, 0.00000381, 0.00000472, 0.00000446, -0.00000308, 0.00000353, 0.00000774, -0.00000212, 0.00000041, 0.00000570, 0.00000687, 0.00001322, 0.00001040, 0.00000745, 0.00001101, 0.00000622, 0.00001162, 0.00003015, 0.00002952, 0.00000460, -0.00002382, -0.00003032, -0.00001632, -0.00001062, -0.00001149, -0.00000621, -0.00000373, -0.00000301, -0.00000102, -0.00000265, },
		{ -0.00000077, -0.00000227, -0.00000380, -0.00000610, -0.00000614, -0.00000412, -0.00000262, -0.00000292, -0.00000095, 0.00000273, -0.00000342, -0.00000989, -0.00000324, -0.00000004, -0.00000352, -0.00000270, -0.00000349, -0.00000263, 0.00000179, -0.00000090, -0.00000340, -0.00000139, -0.00000346, -0.00000324, -0.00000014, -0.00000278, -0.00000078, 0.00000387, -0.00000438, -0.00000607, 0.00000211, -0.00000663, -0.00000811, 0.00000150, -0.00001337, -0.00001232, 0.00000758, -0.00000554, -0.00000123, 0.00001936, -0.00000339, -0.00000532, 0.00001551, -0.00000044, -0.00000200, 0.00000817, 0.00000077, 0.00000682, -0.00000181, -0.00001074, 0.00000879, -0.00000141, -0.00001148, 0.00000821, 0.00000295, -0.00000235, 0.00000143, -0.00000206, 0.00001369, 0.00000588, -0.00001324, 0.00001001, 0.00000254, -0.00002459, 0.00000219, 0.00001534, -0.00000777, -0.00000938, 0.00000022, -0.00000178, -0.00000074, 0.00001521, 0.00000954, -0.00002392, -0.00001409, 0.00001551, 0.00000137, -0.00001431, -0.00001443, 0.00000013, 0.00002823, 0.00000234, -0.00003843, -0.00000716, 0.00003162, 0.00002709, 0.00000065, -0.00002863, -0.00001424, 0.00002496, 0.00002463, 0.00000103, -0.00001229, -0.00000487, 0.00001381, 0.00001256, 0.00000161, 0.00000226, 0.00000636, 0.00000924, 0.00000567, 0.00000417, 0.00000869, 0.00000148, -0.00000156, 0.00000617, 0.00000227, 0.00000068, 0.00000482, 0.00000296, 0.00000930, 0.00001429, 0.00001000, 0.00001130, 0.00000885, 0.00000793, 0.00002470, 0.00003487, 0.00001661, -0.00001673, -0.00003351, -0.00002222, -0.00001159, -0.00001205, -0.00000818, -0.00000507, -0.00000492, -0.00000165, },
		{ -0.00000403, -0.00000138, -0.00000077, -0.00000077, -0.00000444, -0.00000717, -0.00000564, -0.00000333, -0.00000336, -0.00000365, 0.00000149, 0.00000129, -0.00000809, -0.00000675, -0.00000010, -0.00000232, -0.00000329, -0.00000393, -0.00000475, 0.00000173, 0.00000262, -0.00000332, -0.00000288, -0.00000227, -0.00000350, -0.00000176, -0.00000336, -0.00000183, 0.00000638, 0.00000024, -0.00000987, -0.00000055, -0.00000061, -0.00001059, -0.00000160, -0.00000604, -0.00001485, 0.00000515, 0.00000014, -0.00001038, 0.00001756, 0.00000814, -0.00001524, 0.00000919, 0.00001036, -0.00000405, 0.00000474, -0.00000002, 0.00000519, 0.00000798, -0.00001350, -0.00000037, 0.00000935, -0.00001075, -0.00000129, 0.00000410, -0.00000102, 0.00000732, -0.00000672, -0.00000066, 0.00002081, -0.00000526, -0.00000784, 0.00001306, -0.00001330, -0.00001771, 0.00001452, 0.00000920, -0.00001264, -0.00000933, 0.00000034, -0.00000317, 0.00000403, 0.00001918, -0.00000341, -0.00002351, -0.00000184, 0.00000701, 0.00000067, -0.00000886, -0.00002320, 0.00000495, 0.00002846, -0.00000900, -0.00002662, -0.00000203, 0.00002102, 0.00003120, 0.00000180, -0.00003040, -0.00000768, 0.00002418, 0.00002274, 0.00000009, -0.00001597, -0.00000067, 0.00001747, 0.00001187, 0.00000300, 0.00000184, 0.00000677, 0.00000799, 0.00000361, 0.00000822, 0.00000802, 0.00000014, 0.00000300, 0.00000276, 0.00000064, 0.00000655, 0.00000386, 0.00000320, 0.00001201, 0.00001293, 0.00001300, 0.00001231, 0.00000723, 0.00001784, 0.00003568, 0.00002854, -0.00000597, -0.00003411, -0.00002883, -0.00001385, -0.00001262, -0.00001009, -0.00000558, -0.00000621, },
		{ -0.00000626, -0.00000576, -0.00000333, -0.00000220, 0.00000088, -0.00000040, -0.00000571, -0.00000706, -0.00000498, -0.00000324, -0.00000466, -0.00000227, 0.00000251, -0.00000379, -0.00000795, -0.00000164, -0.00000162, -0.00000333, -0.00000287, -0.00000646, -0.00000211, 0.00000488, 0.00000039, -0.00000362, -0.00000318, -0.00000334, -0.00000135, -0.00000351, -0.00000577, 0.00000494, 0.00000789, -0.00000748, -0.00000725, 0.00000143, -0.00000731, -0.00000554, -0.00000278, -0.00001525, 0.00000067, 0.00000910, -0.00001447, 0.00000513, 0.00002107, -0.00001210, -0.00000559, 0.00001364, 0.00000034, 0.00000402, 0.00000063, -0.00000337, 0.00001337, -0.00000359, -0.00001174, 0.00000918, -0.00000238, -0.00000498, 0.00000272, -0.00000742, 0.00000990, 0.00000709, -0.00001720, 0.00001176, 0.00001734, -0.00001367, 0.00000194, 0.00000539, -0.00002046, -0.00000456, 0.00001790, 0.00000147, -0.00001602, -0.00000710, -0.00000006, -0.00000268, 0.00001118, 0.00001101, -0.00001194, -0.00000966, -0.00000262, -0.00000137, 0.00000751, -0.00001336, -0.00002362, 0.00001062, 0.00001466, -0.00000672, -0.00001195, -0.00000947, 0.00001805, 0.00003423, -0.00000121, -0.00002360, -0.00000387, 0.00001988, 0.00002132, -0.00000301, -0.00001394, 0.00000459, 0.00001657, 0.00001234, 0.00000406, 0.00000220, 0.00000677, 0.00000474, 0.00000572, 0.00001030, 0.00000509, 0.00000376, 0.00000348, -0.00000147, 0.00000502, 0.00000834, 0.00000244, 0.00000651, 0.00001157, 0.00001352, 0.00001613, 0.00001066, 0.00001237, 0.00003135, 0.00003782, 0.00000874, -0.00003060, -0.00003616, -0.00001785, -0.00001274, -0.00001228, -0.00000652, },
		{ -0.00000829, -0.00000621, -0.00000664, -0.00000429, -0.00000455, -0.00000123, 0.00000226, -0.00000166, -0.00000574, -0.00000654, -0.00000461, -0.00000419, -0.00000474, 0.00000019, -0.00000107, -0.00000715, -0.00000247, -0.00000065, -0.00000438, -0.00000219, -0.00000480, -0.00000631, 0.00000222, 0.00000391, -0.00000088, -0.00000322, -0.00000498, -0.00000185, -0.00000077, -0.00000745, -0.00000196, 0.00001061, 0.00000147, -0.00000942, -0.00000215, -0.00000450, -0.00000698, -0.00000046, -0.00001388, -0.00000873, 0.00001570, -0.00000600, -0.00001122, 0.00002151, 0.00000239, -0.00001434, 0.00000801, 0.00000259, 0.00000306, 0.00000824, -0.00000891, 0.00000480, 0.00000913, -0.00001164, 0.00000163, 0.00000175, -0.00000580, 0.00000807, -0.00001050, -0.00000629, 0.00002381, -0.00000801, -0.00001517, 0.00002360, 0.00000479, -0.00001335, 0.00000658, -0.00000445, -0.00001690, 0.00000559, 0.00001482, -0.00000634, -0.00001570, -0.00000423, -0.00000177, 0.00000330, 0.00001032, -0.00000298, -0.00000413, -0.00000050, -0.00001269, 0.00000099, 0.00000786, -0.00001809, -0.00001453, 0.00000225, 0.00000429, 0.00000796, -0.00000999, -0.00001585, 0.00002088, 0.00003002, 0.00000143, -0.00001416, -0.00000560, 0.00001626, 0.00001777, -0.00000353, -0.00000734, 0.00000577, 0.00001496, 0.00001340, 0.00000488, 0.00000358, 0.00000416, 0.00000391, 0.00000924, 0.00000807, 0.00000587, 0.00000715, -0.00000038, -0.00000006, 0.00000905, 0.00000650, 0.00000465, 0.00000831, 0.00001053, 0.00001641, 0.00001624, 0.00001200, 0.00002459, 0.00004120, 0.00002505, -0.00002058, -0.00004232, -0.00002502, -0.00001307, -0.00001375, },
		{ -0.00001461, -0.00001041, -0.00000611, -0.00000753, -0.00000484, -0.00000508, -0.00000473, 0.00000100, 0.00000114, -0.00000216, -0.00000488, -0.00000619, -0.00000493, -0.00000542, -0.00000232, -0.00000047, -0.00000694, -0.00000395, 0.00000195, -0.00000389, -0.00000426, -0.00000249, -0.00000661, -0.00000222, 0.00000321, 0.00000164, -0.00000011, -0.00000467, -0.00000516, 0.00000029, -0.00000407, -0.00000716, 0.00000539, 0.00000785, -0.00000448, -0.00000412, -0.00000417, -0.00000946, 0.00000044, -0.00000681, -0.00001823, 0.00001026, 0.00000967, -0.00001653, 0.00000910, 0.00001407, -0.00001311, 0.00000208, 0.00000408, -0.00000451, 0.00001342, -0.00000023, -0.00000764, 0.00000957, -0.00000422, -0.00000172, 0.00000430, -0.00001221, 0.00000977, 0.00000582, -0.00002604, 0.00001270, 0.00002185, -0.00002355, -0.00000007, 0.00002412, -0.00000715, -0.00000855, 0.00000493, -0.00000848, -0.00000895, 0.00000977, 0.00000738, -0.00001116, -0.00001144, -0.00000449, 0.00000056, 0.00000837, -0.00000247, -0.00000560, 0.00001007, -0.00000550, -0.00001594, 0.00000492, 0.00000053, -0.00001162, -0.00001212, -0.00001364, 0.00000968, 0.00001680, -0.00001582, -0.00001311, 0.00001906, 0.00002567, 0.00001076, -0.00000964, -0.00000853, 0.00001228, 0.00001372, 0.00000151, -0.00000225, 0.00000418, 0.00001447, 0.00001377, 0.00000690, 0.00000334, 0.00000147, 0.00000666, 0.00000914, 0.00000632, 0.00000974, 0.00000556, -0.00000180, 0.00000488, 0.00000860, 0.00000642, 0.00000797, 0.00000773, 0.00001187, 0.00001810, 0.00001577, 0.00002041, 0.00003920, 0.00003860, -0.00000425, -0.00004360, -0.00003487, -0.00001549, },
		{ -0.00002081, -0.00001545, -0.00001344, -0.00000607, -0.00000755, -0.00000626, -0.00000455, -0.00000659, -0.00000200, 0.00000081, -0.00000054, -0.00000110, -0.00000430, -0.00000590, -0.00000652, -0.00000456, -0.00000002, -0.00000616, -0.00000811, 0.00000261, 0.00000057, -0.00000576, -0.00000319, -0.00000511, -0.00000412, 0.00000113, 0.00000062, 0.00000164, -0.00000031, -0.00000617, -0.00000287, -0.00000134, -0.00000652, -0.00000035, 0.00000724, -0.00000008, -0.00000225, -0.00000105, -0.00001153, -0.00000514, 0.00000173, -0.00001753, -0.00000420, 0.00001728, -0.00000851, -0.00000315, 0.00001755, -0.00000880, -0.00000415, 0.00001101, -0.00000996, 0.00000375, 0.00001288, -0.00000715, 0.00000218, -0.00000196, -0.00000465, 0.00001310, -0.00001315, -0.00000828, 0.00002647, -0.00001606, -0.00002222, 0.00003132, 0.00000315, -0.00002442, 0.00001534, 0.00001449, -0.00001319, -0.00000481, 0.00000238, -0.00000657, -0.00000189, 0.00000752, -0.00000121, -0.00000967, -0.00000844, -0.00000493, 0.00000739, 0.00000199, -0.00001361, 0.00000626, 0.00001289, -0.00001270, -0.00001051, -0.00000031, -0.00000079, -0.00000205, -0.00002338, -0.00001842, 0.00001994, 0.00001248, -0.00001566, -0.00000892, 0.00001228, 0.00002865, 0.00001874, -0.00000935, -0.00001055, 0.00000679, 0.00001385, 0.00000826, -0.00000101, 0.00000312, 0.00001381, 0.00001462, 0.00000875, 0.00000122, 0.00000240, 0.00000845, 0.00000645, 0.00000863, 0.00001080, 0.00000248, 0.00000208, 0.00000718, 0.00000697, 0.00000922, 0.00000904, 0.00000789, 0.00001425, 0.00001770, 0.00001958, 0.00003541, 0.00004743, 0.00001559, -0.00003765, -0.00004505, },
		{ -0.00005314, -0.00002957, -0.00001594, -0.00001711, -0.00000804, -0.00000626, -0.00000749, -0.00000450, -0.00000708, -0.00000456, -0.00000011, -0.00000178, -0.00000023, -0.00000018, -0.00000357, -0.00000662, -0.00000779, -0.00000121, -0.00000273, -0.00001142, -0.00000238, 0.00000468, -0.00000286, -0.00000429, -0.00000516, -0.00000537, 0.00000093, -0.00000006, -0.00000106, 0.00000230, -0.00000214, -0.00000457, -0.00000259, -0.00000551, -0.00000214, 0.00000584, 0.00000050, -0.00000309, 0.00000380, -0.00000607, -0.00001307, 0.00000130, -0.00000827, -0.00001371, 0.00001283, 0.00000141, -0.00000978, 0.00001777, -0.00000088, -0.00001570, 0.00001484, -0.00000010, -0.00001134, 0.00001186, 0.00000196, 0.00000116, -0.00000012, -0.00001592, 0.00001503, 0.00000772, -0.00003004, 0.00001632, 0.00002090, -0.00003578, 0.00000051, 0.00003113, -0.00001541, -0.00001140, 0.00002064, 0.00000225, -0.00001439, -0.00000255, 0.00000208, -0.00000174, 0.00000119, -0.00000031, -0.00000494, -0.00000505, -0.00000884, 0.00000006, 0.00000834, -0.00001203, -0.00000927, 0.00001599, 0.00000484, -0.00001000, -0.00000933, -0.00000754, 0.00000766, -0.00000466, -0.00003378, -0.00000984, 0.00001902, 0.00000638, -0.00000877, -0.00001087, 0.00000966, 0.00003538, 0.00002092, -0.00000919, -0.00001335, 0.00000316, 0.00001816, 0.00001189, -0.00000097, 0.00000270, 0.00001301, 0.00001627, 0.00000828, 0.00000094, 0.00000527, 0.00000652, 0.00000624, 0.00001141, 0.00000791, 0.00000360, 0.00000659, 0.00000621, 0.00000809, 0.00001152, 0.00000887, 0.00001002, 0.00001528, 0.00001823, 0.00003076, 0.00005148, 0.00003664, -0.00002320, },
		{ 0.00000102, -0.00005471, -0.00004272, -0.00001746, -0.00001893, -0.00001271, -0.00000581, -0.00000774, -0.00000451, -0.00000666, -0.00000740, -0.00000060, -0.00000206, -0.00000296, 0.00000077, 0.00000035, -0.00000280, -0.00000876, -0.00000588, -0.00000013, -0.00000951, -0.00000922, 0.00000307, 0.00000107, -0.00000250, -0.00000414, -0.00000801, -0.00000083, 0.00000263, -0.00000284, -0.00000057, 0.00000069, -0.00000177, -0.00000227, -0.00000691, -0.00000510, 0.00000640, 0.00000359, -0.00000690, 0.00000154, 0.00000383, -0.00001177, -0.00000648, -0.00000298, -0.00001436, 0.00000565, 0.00000844, -0.00001480, 0.00001227, 0.00001537, -0.00002268, 0.00000186, 0.00001598, -0.00001159, 0.00000035, 0.00000351, 0.00000157, 0.00001209, -0.00002062, -0.00000821, 0.00003213, -0.00001858, -0.00002229, 0.00003682, -0.00000439, -0.00003320, 0.00002100, 0.00001432, -0.00001982, 0.00000347, 0.00001541, -0.00000712, -0.00001251, -0.00000060, 0.00000402, 0.00000327, -0.00000234, -0.00000801, -0.00000203, -0.00000347, -0.00000684, 0.00000592, -0.00000275, -0.00001687, 0.00000369, 0.00001194, 0.00000115, -0.00000393, -0.00001592, -0.00000476, 0.00001178, -0.00001646, -0.00002958, -0.00000403, 0.00000882, 0.00000807, -0.00000507, -0.00001410, 0.00001438, 0.00003825, 0.00002024, -0.00000890, -0.00001621, 0.00000441, 0.00002187, 0.00001253, -0.00000032, 0.00000207, 0.00001326, 0.00001647, 0.00000733, 0.00000410, 0.00000563, 0.00000431, 0.00000885, 0.00001035, 0.00000630, 0.00000811, 0.00000773, 0.00000612, 0.00001079, 0.00001178, 0.00000987, 0.00001250, 0.00001532, 0.00002386, 0.00004877, 0.00005531, },
		{ 0.00006434, 0.00003163, -0.00004301, -0.00005719, -0.00002409, -0.00001831, -0.00001757, -0.00000758, -0.00000827, -0.00000505, -0.00000433, -0.00000973, -0.00000313, -0.00000031, -0.00000457, -0.00000203, 0.00000065, 0.00000169, -0.00000382, -0.00000925, -0.00000246, -0.00000543, -0.00001201, -0.00000191, 0.00000142, -0.00000123, -0.00000013, -0.00000749, -0.00000623, 0.00000348, 0.00000026, -0.00000304, -0.00000154, -0.00000082, 0.00000168, -0.00000452, -0.00001041, 0.00000238, 0.00000974, -0.00000448, -0.00000680, 0.00000543, -0.00000223, -0.00000879, -0.00000390, -0.00001425, -0.00000158, 0.00001628, -0.00001387, -0.00000527, 0.00002838, -0.00000965, -0.00001864, 0.00001660, 0.00000063, -0.00000367, 0.00000164, -0.00000962, 0.00001916, 0.00000262, -0.00003657, 0.00001978, 0.00002284, -0.00003884, 0.00000536, 0.00003260, -0.00002497, -0.00001314, 0.00002333, -0.00000248, -0.00001116, 0.00001022, 0.00000502, -0.00001112, -0.00000885, 0.00000101, 0.00000798, 0.00000431, -0.00001057, -0.00000909, 0.00000146, -0.00000360, 0.00000009, 0.00000169, -0.00001457, -0.00000668, 0.00000758, 0.00000334, 0.00000620, -0.00000574, -0.00001824, 0.00000294, 0.00000256, -0.00002013, -0.00001827, -0.00001033, 0.00000251, 0.00001208, -0.00000711, -0.00001067, 0.00001994, 0.00003619, 0.00001955, -0.00000957, -0.00001551, 0.00000816, 0.00002231, 0.00001271, 0.00000032, 0.00000215, 0.00001347, 0.00001467, 0.00000926, 0.00000718, 0.00000356, 0.00000520, 0.00000972, 0.00000758, 0.00000889, 0.00001108, 0.00000743, 0.00000857, 0.00001228, 0.00001164, 0.00001229, 0.00001411, 0.00001711, 0.00003825, },
		{ 0.00002562, 0.00005919, 0.00005781, -0.00001646, -0.00006437, -0.00003680, -0.00001842, -0.00002098, -0.00001048, -0.00000890, -0.00000804, -0.00000187, -0.00000849, -0.00000740, -0.00000002, -0.00000345, -0.00000399, -0.00000202, 0.00000094, 0.00000238, -0.00000580, -0.00000572, -0.00000414, -0.00001202, -0.00000578, 0.00000070, -0.00000350, 0.00000119, -0.00000146, -0.00000890, -0.00000098, 0.00000269, -0.00000097, -0.00000268, -0.00000459, 0.00000160, 0.00000271, -0.00000972, -0.00000652, 0.00000946, 0.00000449, -0.00000868, -0.00000211, 0.00000183, -0.00000418, -0.00000170, -0.00001320, -0.00001352, 0.00001874, 0.00000139, -0.00002307, 0.00001971, 0.00001453, -0.00002311, 0.00000288, 0.00000638, -0.00000281, 0.00001104, -0.00001731, -0.00000325, 0.00003322, -0.00002548, -0.00002549, 0.00003994, -0.00000593, -0.00003275, 0.00002622, 0.00001119, -0.00002474, 0.00000398, 0.00001146, -0.00000732, 0.00000003, 0.00000681, -0.00000403, -0.00000951, -0.00000577, 0.00000340, 0.00001162, -0.00000143, -0.00001581, -0.00000509, 0.00000126, 0.00000077, 0.00000301, -0.00001145, -0.00001239, 0.00000420, -0.00000008, 0.00000309, 0.00000870, -0.00001094, -0.00001059, 0.00000075, -0.00000854, -0.00001064, -0.00001562, -0.00001980, 0.00000316, 0.00001009, -0.00000711, -0.00000197, 0.00002089, 0.00003286, 0.00001824, -0.00000905, -0.00001060, 0.00001022, 0.00002102, 0.00001296, 0.00000106, 0.00000320, 0.00001169, 0.00001397, 0.00001323, 0.00000736, 0.00000307, 0.00000694, 0.00000738, 0.00000775, 0.00001254, 0.00001132, 0.00000894, 0.00001148, 0.00001234, 0.00001261, 0.00001512, 0.00001515, },
		{ 0.00001721, 0.00001878, 0.00004532, 0.00006985, 0.00001661, -0.00005816, -0.00005086, -0.00002110, -0.00002352, -0.00001504, -0.00000798, -0.00001179, -0.00000344, -0.00000460, -0.00000952, -0.00000202, -0.00000250, -0.00000431, -0.00000268, -0.00000300, 0.00000230, 0.00000033, -0.00000411, -0.00000306, -0.00001206, -0.00001043, 0.00000146, -0.00000413, -0.00000378, 0.00000259, -0.00000478, -0.00000448, 0.00000070, 0.00000006, 0.00000046, -0.00000535, -0.00000486, 0.00000486, -0.00000169, -0.00000970, 0.00000168, 0.00000822, -0.00000202, -0.00000555, -0.00000193, -0.00000397, 0.00000233, -0.00000342, -0.00002237, 0.00000504, 0.00001988, -0.00002052, -0.00000402, 0.00002449, -0.00001094, -0.00000580, 0.00000703, -0.00001209, 0.00001745, 0.00000496, -0.00003481, 0.00002264, 0.00002169, -0.00004474, 0.00000423, 0.00003296, -0.00002633, -0.00001010, 0.00002492, -0.00000476, -0.00001096, 0.00000615, -0.00000003, -0.00000140, 0.00000434, -0.00000221, -0.00000690, -0.00000556, -0.00000393, 0.00000747, 0.00000985, -0.00000940, -0.00001364, -0.00000308, 0.00000196, 0.00000721, -0.00000564, -0.00001797, 0.00000005, 0.00000260, -0.00000558, 0.00000739, 0.00000340, -0.00000772, -0.00000471, -0.00001131, -0.00000720, -0.00000166, -0.00002250, -0.00002205, 0.00000206, 0.00000478, -0.00000045, 0.00000474, 0.00001865, 0.00002972, 0.00001653, -0.00000476, -0.00000505, 0.00000996, 0.00001961, 0.00001292, 0.00000284, 0.00000326, 0.00000924, 0.00001614, 0.00001547, 0.00000690, 0.00000497, 0.00000599, 0.00000561, 0.00001080, 0.00001389, 0.00001165, 0.00001216, 0.00001316, 0.00001234, 0.00001506, },
		{ 0.00001365, 0.00001825, 0.00001845, 0.00003251, 0.00006828, 0.00004598, -0.00004024, -0.00006237, -0.00002560, -0.00002326, -0.00002254, -0.00000791, -0.00001204, -0.00000851, -0.00000283, -0.00000902, -0.00000371, -0.00000170, -0.00000581, -0.00000209, -0.00000355, -0.00000216, 0.00000136, -0.00000106, 0.00000124, -0.00000804, -0.00001671, -0.00000154, 0.00000037, -0.00000793, -0.00000085, -0.00000088, -0.00000294, 0.00000025, -0.00000253, 0.00000065, -0.00000012, -0.00000788, -0.00000126, 0.00000247, -0.00000492, -0.00000204, 0.00000450, 0.00000112, -0.00000208, -0.00000179, -0.00000792, -0.00000287, 0.00000775, -0.00001406, -0.00001440, 0.00002003, -0.00000191, -0.00001758, 0.00001717, -0.00000131, -0.00001007, 0.00001731, -0.00001726, -0.00000662, 0.00003715, -0.00002365, -0.00002665, 0.00004148, -0.00000761, -0.00003699, 0.00002465, 0.00001016, -0.00002407, 0.00000547, 0.00001134, -0.00000456, -0.00000067, -0.00000173, -0.00000184, 0.00000594, -0.00000064, -0.00000910, -0.00000389, -0.00000311, -0.00000092, 0.00000954, 0.00000189, -0.00001152, -0.00000909, -0.00000371, 0.00000741, 0.00000524, -0.00001712, -0.00000979, 0.00000594, -0.00000518, -0.00000255, 0.00000529, 0.00000030, 0.00000145, -0.00001083, -0.00001779, 0.00000234, -0.00000380, -0.00002701, -0.00002033, -0.00000494, 0.00000405, 0.00000804, 0.00000676, 0.00001638, 0.00002628, 0.00001703, 0.00000190, -0.00000154, 0.00000904, 0.00001804, 0.00001346, 0.00000467, 0.00000166, 0.00000912, 0.00001859, 0.00001550, 0.00000865, 0.00000599, 0.00000357, 0.00000693, 0.00001320, 0.00001383, 0.00001374, 0.00001522, 0.00001371, },
		{ 0.00001703, 0.00001428, 0.00001725, 0.00001987, 0.00002459, 0.00005914, 0.00006772, -0.00001358, -0.00007065, -0.00003510, -0.00001759, -0.00002829, -0.00001327, -0.00000975, -0.00001231, -0.00000409, -0.00000884, -0.00000599, 0.00000103, -0.00000636, -0.00000467, -0.00000161, -0.00000379, -0.00000113, -0.00000243, 0.00000308, 0.00000202, -0.00001661, -0.00001077, 0.00000279, -0.00000547, -0.00000541, -0.00000244, -0.00000297, 0.00000369, -0.00000096, -0.00000432, 0.00000182, -0.00000346, -0.00000519, -0.00000087, -0.00000345, 0.00000037, 0.00000496, -0.00000103, -0.00000310, 0.00000259, -0.00000372, -0.00001227, 0.00000358, 0.00000275, -0.00001691, 0.00000527, 0.00000920, -0.00001687, 0.00000994, 0.00000614, -0.00002429, 0.00002184, 0.00001007, -0.00004038, 0.00002464, 0.00002865, -0.00004690, 0.00000133, 0.00003459, -0.00002774, -0.00001401, 0.00002260, -0.00000479, -0.00001064, 0.00000489, 0.00000320, 0.00000371, -0.00000148, -0.00000770, 0.00000408, 0.00000613, -0.00000944, -0.00000895, -0.00000013, -0.00000146, 0.00000323, 0.00000480, -0.00000497, -0.00000719, -0.00000786, -0.00000109, 0.00001053, -0.00000624, -0.00001662, 0.00000145, 0.00000023, -0.00000636, -0.00000019, -0.00000106, 0.00000640, 0.00000283, -0.00002063, -0.00001163, 0.00000536, -0.00001063, -0.00002438, -0.00002305, -0.00001157, 0.00000844, 0.00001227, 0.00000680, 0.00001460, 0.00002389, 0.00002065, 0.00000759, 0.00000044, 0.00000812, 0.00001676, 0.00001472, 0.00000475, 0.00000092, 0.00001104, 0.00001877, 0.00001623, 0.00001125, 0.00000483, 0.00000311, 0.00000938, 0.00001378, 0.00001462, 0.00001680, },
		{ 0.00001679, 0.00001950, 0.00001744, 0.00001717, 0.00002124, 0.00002081, 0.00004580, 0.00007974, 0.00002158, -0.00007030, -0.00005509, -0.00001284, -0.00002485, -0.00002136, -0.00000989, -0.00001413, -0.00000589, -0.00000796, -0.00001128, 0.00000194, -0.00000186, -0.00000793, -0.00000268, -0.00000308, -0.00000128, -0.00000428, -0.00000272, 0.00000840, -0.00000576, -0.00001658, -0.00000209, -0.00000237, -0.00000567, -0.00000343, -0.00000762, 0.00000208, 0.00000653, -0.00000441, -0.00000281, -0.00000126, -0.00000299, -0.00000129, -0.00000747, -0.00000311, 0.00001010, 0.00000369, -0.00000799, -0.00000146, 0.00000508, -0.00000862, -0.00000987, 0.00000466, -0.00000546, -0.00000155, 0.00001041, -0.00001715, 0.00000075, 0.00002514, -0.00003145, -0.00000963, 0.00004758, -0.00002610, -0.00003130, 0.00005024, -0.00000300, -0.00004332, 0.00002246, 0.00001236, -0.00002608, 0.00000127, 0.00001010, -0.00000419, -0.00000341, -0.00000195, 0.00000570, 0.00000743, -0.00000774, -0.00000538, 0.00000875, -0.00000171, -0.00001368, -0.00000440, 0.00000137, 0.00000219, 0.00000322, -0.00000409, -0.00000442, -0.00000333, -0.00000702, 0.00000429, 0.00000313, -0.00001363, -0.00000498, 0.00000198, -0.00000549, -0.00000219, -0.00000483, -0.00000060, 0.00001244, -0.00000601, -0.00002007, -0.00000376, -0.00000117, -0.00001167, -0.00002183, -0.00002950, -0.00001218, 0.00001163, 0.00001208, 0.00000706, 0.00001344, 0.00002454, 0.00002502, 0.00001118, 0.00000192, 0.00000738, 0.00001661, 0.00001494, 0.00000405, 0.00000277, 0.00001199, 0.00001804, 0.00001852, 0.00001219, 0.00000404, 0.00000476, 0.00001065, 0.00001406, },
		{ 0.00001101, 0.00001526, 0.00001980, 0.00002046, 0.00001842, 0.00002308, 0.00002200, 0.00003188, 0.00007762, 0.00005955, -0.00005067, -0.00008047, -0.00002189, -0.00001489, -0.00002438, -0.00001184, -0.00001601, -0.00000985, -0.00000419, -0.00001522, -0.00000417, 0.00000352, -0.00000611, -0.00000505, -0.00000451, -0.00000151, -0.00000093, -0.00000850, 0.00000304, 0.00000504, -0.00001183, -0.00000661, -0.00000312, -0.00000707, -0.00000088, -0.00000800, -0.00000720, 0.00000818, 0.00000345, -0.00000354, -0.00000428, -0.00000442, 0.00000300, -0.00000467, -0.00001271, 0.00000567, 0.00001392, -0.00000245, -0.00001039, 0.00000142, 0.00000343, -0.00000979, -0.00000615, -0.00000335, -0.00000038, 0.00001745, -0.00001237, -0.00002452, 0.00003603, 0.00000102, -0.00005362, 0.00003320, 0.00003516, -0.00005443, 0.00000343, 0.00004570, -0.00002889, -0.00002148, 0.00002266, -0.00000330, -0.00001402, 0.00000188, 0.00000356, 0.00000150, -0.00000562, -0.00000332, 0.00001092, 0.00000320, -0.00000986, 0.00000154, 0.00000519, -0.00001015, -0.00001073, -0.00000127, 0.00000325, 0.00000551, -0.00000377, -0.00000854, 0.00000016, -0.00000250, -0.00000207, 0.00000310, -0.00000870, -0.00000743, 0.00000267, -0.00000483, -0.00000357, -0.00000276, -0.00000860, 0.00000639, 0.00000776, -0.00001193, -0.00001066, -0.00000526, -0.00000698, -0.00000814, -0.00002467, -0.00003235, -0.00000975, 0.00000972, 0.00001041, 0.00000773, 0.00001455, 0.00002764, 0.00002790, 0.00001336, 0.00000316, 0.00000780, 0.00001660, 0.00001346, 0.00000515, 0.00000531, 0.00001131, 0.00001851, 0.00002020, 0.00001229, 0.00000528, 0.00000636, },
		{ 0.00000739, 0.00000725, 0.00001144, 0.00001781, 0.00002215, 0.00001976, 0.00002343, 0.00002731, 0.00002472, 0.00006263, 0.00008653, -0.00001018, -0.00009411, -0.00004751, -0.00001067, -0.00002223, -0.00001169, -0.00001436, -0.00001767, -0.00000276, -0.00001210, -0.00001301, 0.00000238, -0.00000193, -0.00000404, -0.00000492, -0.00000602, 0.00000186, -0.00000563, -0.00000583, 0.00000528, -0.00000487, -0.00000523, -0.00000066, -0.00001104, -0.00000348, -0.00000129, -0.00001202, -0.00000047, 0.00000705, 0.00000249, -0.00000184, -0.00001011, -0.00000047, 0.00000520, -0.00001258, -0.00000793, 0.00001270, 0.00001044, -0.00000603, -0.00001003, 0.00000177, 0.00000154, -0.00000608, -0.00000945, -0.00000979, 0.00002093, 0.00001512, -0.00004277, 0.00000308, 0.00004890, -0.00004380, -0.00003343, 0.00006214, -0.00000419, -0.00004872, 0.00003114, 0.00001979, -0.00003191, -0.00000377, 0.00001145, -0.00000510, -0.00000687, -0.00000228, 0.00000515, 0.00000204, -0.00000876, 0.00000142, 0.00001041, -0.00000253, -0.00000456, 0.00000404, -0.00000436, -0.00001166, -0.00000584, -0.00000016, 0.00000718, 0.00000269, -0.00001159, -0.00000579, 0.00000230, 0.00000020, 0.00000197, -0.00000743, -0.00001090, 0.00000365, -0.00000065, -0.00000696, -0.00000093, -0.00000786, -0.00000397, 0.00000777, -0.00000211, -0.00000663, -0.00000555, -0.00001208, -0.00000610, -0.00000798, -0.00002832, -0.00002944, -0.00001038, 0.00000428, 0.00000872, 0.00000951, 0.00001869, 0.00003093, 0.00002906, 0.00001458, 0.00000501, 0.00000938, 0.00001513, 0.00001244, 0.00000797, 0.00000647, 0.00001094, 0.00001931, 0.00002059, 0.00001363, },
		{ 0.00002133, 0.00001607, 0.00000938, 0.00000781, 0.00001271, 0.00002132, 0.00002180, 0.00002147, 0.00003069, 0.00002717, 0.00004575, 0.00009403, 0.00003796, -0.00008385, -0.00007800, -0.00001835, -0.00002147, -0.00001350, -0.00000569, -0.00002152, -0.00000940, -0.00000644, -0.00001653, -0.00000335, -0.00000129, -0.00000341, -0.00000043, -0.00000848, -0.00000276, -0.00000024, -0.00000747, 0.00000086, -0.00000378, -0.00000631, 0.00000664, -0.00000666, -0.00001248, 0.00000046, -0.00000717, -0.00000719, 0.00000189, 0.00000305, 0.00000644, -0.00000627, -0.00001129, 0.00000561, -0.00000166, -0.00001241, 0.00000077, 0.00001084, 0.00000745, -0.00000659, -0.00001022, 0.00000014, 0.00000286, -0.00000239, -0.00001981, -0.00000220, 0.00004084, -0.00001314, -0.00004873, 0.00004206, 0.00002231, -0.00006708, 0.00001027, 0.00005269, -0.00003358, -0.00002022, 0.00003255, -0.00000228, -0.00002022, 0.00000031, 0.00000351, -0.00000151, -0.00000676, -0.00000237, 0.00000697, -0.00000163, -0.00000654, 0.00000495, 0.00000436, -0.00000218, 0.00000210, -0.00000215, -0.00001090, -0.00000720, -0.00000373, 0.00000263, 0.00000840, -0.00000560, -0.00001263, -0.00000187, 0.00000301, 0.00000573, -0.00000300, -0.00001602, -0.00000183, 0.00000603, -0.00000581, -0.00000357, -0.00000432, -0.00000771, 0.00000219, -0.00000060, -0.00000407, 0.00000164, -0.00000940, -0.00001426, -0.00000410, -0.00001217, -0.00002672, -0.00002590, -0.00001562, -0.00000198, 0.00000738, 0.00001351, 0.00002448, 0.00003323, 0.00002885, 0.00001550, 0.00000821, 0.00001035, 0.00001308, 0.00001337, 0.00001026, 0.00000708, 0.00001134, 0.00001907, },
		{ 0.00001123, 0.00001870, 0.00002292, 0.00001871, 0.00001136, 0.00000827, 0.00001583, 0.00002307, 0.00002065, 0.00002872, 0.00003256, 0.00003584, 0.00008660, 0.00007952, -0.00005118, -0.00010159, -0.00003494, -0.00002161, -0.00002363, 0.00000057, -0.00001319, -0.00001726, -0.00000601, -0.00001617, -0.00000801, -0.00000132, -0.00000754, 0.00000168, -0.00000263, -0.00000790, -0.00000112, -0.00000596, -0.00000023, -0.00000167, -0.00001374, 0.00000534, 0.00000764, -0.00001323, -0.00000678, -0.00000427, -0.00000667, -0.00000022, -0.00000363, 0.00000509, 0.00000657, -0.00001157, -0.00000469, 0.00000200, -0.00000643, -0.00000217, 0.00000105, 0.00000638, 0.00000781, -0.00000710, -0.00001075, -0.00000283, 0.00000846, -0.00000339, -0.00002925, 0.00001854, 0.00004036, -0.00004505, -0.00002323, 0.00005703, -0.00001937, -0.00005147, 0.00003849, 0.00002089, -0.00003370, 0.00000253, 0.00001691, -0.00000821, -0.00000981, -0.00000266, 0.00000216, -0.00000044, -0.00000721, 0.00000028, 0.00000467, -0.00000298, -0.00000141, 0.00000230, -0.00000006, 0.00000414, 0.00000168, -0.00001117, -0.00000953, -0.00000342, -0.00000228, 0.00000625, 0.00000254, -0.00001110, -0.00000825, -0.00000082, 0.00000675, 0.00000642, -0.00001329, -0.00001255, 0.00000556, 0.00000051, -0.00000462, -0.00000334, -0.00000792, -0.00000099, -0.00000055, -0.00000799, 0.00000262, 0.00000203, -0.00001383, -0.00001131, -0.00000682, -0.00001429, -0.00002114, -0.00002641, -0.00002302, -0.00000794, 0.00000711, 0.00001942, 0.00003033, 0.00003439, 0.00002776, 0.00001738, 0.00001160, 0.00001020, 0.00001264, 0.00001484, 0.00001150, 0.00000829, },
		{ 0.00001284, 0.00000929, 0.00001089, 0.00001907, 0.00002451, 0.00002187, 0.00001285, 0.00001009, 0.00001943, 0.00002131, 0.00002511, 0.00003490, 0.00003200, 0.00007131, 0.00010803, -0.00000123, -0.00011169, -0.00006078, -0.00001826, -0.00003462, -0.00000827, -0.00000089, -0.00001614, -0.00000737, -0.00001614, -0.00001440, 0.00000133, -0.00000923, -0.00000530, 0.00000315, -0.00000527, -0.00000312, -0.00000743, -0.00000375, 0.00000577, -0.00001442, -0.00000923, 0.00001404, -0.00000089, -0.00000907, -0.00000722, -0.00000970, 0.00000268, -0.00000235, -0.00000654, 0.00000877, 0.00000008, -0.00000703, -0.00000296, -0.00000814, 0.00000242, 0.00000407, -0.00000573, 0.00000594, 0.00000717, -0.00000505, -0.00001393, -0.00000607, 0.00001916, -0.00001133, -0.00003057, 0.00003975, 0.00001914, -0.00005656, 0.00001036, 0.00003978, -0.00004129, -0.00001833, 0.00003538, -0.00000236, -0.00001662, 0.00000619, 0.00000317, -0.00000404, -0.00000653, -0.00000465, 0.00000261, -0.00000138, -0.00000515, 0.00000071, 0.00000139, 0.00000030, 0.00000023, -0.00000350, 0.00000246, 0.00000834, -0.00000626, -0.00001411, -0.00000475, -0.00000236, 0.00000146, 0.00000468, -0.00000587, -0.00000920, -0.00000501, 0.00000079, 0.00001045, -0.00000159, -0.00001625, -0.00000288, 0.00000317, -0.00000236, -0.00000178, -0.00000744, -0.00000428, 0.00000197, -0.00000909, -0.00000565, 0.00000676, -0.00000347, -0.00001232, -0.00001037, -0.00001135, -0.00001098, -0.00001726, -0.00003068, -0.00003023, -0.00001313, 0.00000818, 0.00002592, 0.00003548, 0.00003450, 0.00002720, 0.00002004, 0.00001361, 0.00001081, 0.00001373, 0.00001544, },
		{ 0.00001470, 0.00001585, 0.00001401, 0.00000984, 0.00001152, 0.00001916, 0.00002662, 0.00002503, 0.00001430, 0.00001441, 0.00001921, 0.00002164, 0.00003474, 0.00003239, 0.00005155, 0.00011712, 0.00005982, -0.00009714, -0.00009697, -0.00001995, -0.00003265, -0.00002688, 0.00000022, -0.00001052, -0.00000478, -0.00001078, -0.00002390, -0.00000153, -0.00000241, -0.00001255, -0.00000039, -0.00000158, -0.00000037, -0.00000490, -0.00001276, 0.00000625, -0.00000203, -0.00001945, 0.00000370, 0.00000776, -0.00000193, -0.00000317, -0.00001608, -0.00000393, 0.00000741, -0.00000862, -0.00000231, 0.00000351, -0.00000041, 0.00000169, -0.00001460, -0.00000732, 0.00001569, -0.00000093, -0.00000843, 0.00000364, 0.00000822, 0.00000083, -0.00002646, -0.00000166, 0.00003203, -0.00002709, -0.00002144, 0.00005022, -0.00000744, -0.00004531, 0.00002705, 0.00001220, -0.00003473, 0.00000245, 0.00001669, -0.00000497, -0.00000362, -0.00000144, -0.00000065, 0.00000026, -0.00000696, -0.00000440, 0.00000207, -0.00000112, -0.00000314, -0.00000167, 0.00000161, 0.00000404, -0.00000368, -0.00000517, 0.00000836, 0.00000405, -0.00001313, -0.00001005, -0.00000249, -0.00000023, 0.00000375, -0.00000324, -0.00000861, -0.00000450, -0.00000377, 0.00000491, 0.00000677, -0.00000952, -0.00000845, 0.00000044, -0.00000242, -0.00000018, -0.00000378, -0.00000763, 0.00000180, -0.00000399, -0.00001181, 0.00000069, 0.00000283, -0.00000583, -0.00000882, -0.00001369, -0.00001166, -0.00000598, -0.00001793, -0.00003589, -0.00003625, -0.00001730, 0.00000993, 0.00003200, 0.00003941, 0.00003472, 0.00002811, 0.00002161, 0.00001513, 0.00001332, },
		{ 0.00001802, 0.00001647, 0.00001557, 0.00001626, 0.00001422, 0.00001149, 0.00001217, 0.00001912, 0.00002975, 0.00002698, 0.00001853, 0.00001737, 0.00001616, 0.00003064, 0.00003758, 0.00003531, 0.00010161, 0.00011409, -0.00004772, -0.00012841, -0.00004197, -0.00002108, -0.00003834, -0.00000930, -0.00001032, -0.00000587, 0.00000266, -0.00002422, -0.00001483, 0.00000283, -0.00001054, -0.00000639, -0.00000401, -0.00000127, 0.00000470, -0.00001389, -0.00000610, 0.00000723, -0.00001408, -0.00000758, 0.00000442, -0.00000084, 0.00000748, -0.00000867, -0.00001817, 0.00000579, 0.00000166, -0.00000617, -0.00000328, -0.00000462, 0.00001155, -0.00000280, -0.00002470, 0.00000708, 0.00001647, -0.00000426, -0.00000968, -0.00000476, 0.00002210, -0.00000203, -0.00004542, 0.00002068, 0.00003504, -0.00004668, -0.00000288, 0.00004710, -0.00002566, -0.00002629, 0.00002389, -0.00000284, -0.00001747, 0.00000278, 0.00000417, 0.00000275, -0.00000263, -0.00000767, 0.00000158, 0.00000106, -0.00000734, -0.00000372, 0.00000159, 0.00000028, -0.00000366, -0.00000318, 0.00000499, 0.00000315, -0.00000799, -0.00000038, 0.00000886, -0.00000487, -0.00001229, -0.00000560, -0.00000167, 0.00000353, -0.00000071, -0.00001012, -0.00000499, -0.00000209, -0.00000068, 0.00000575, -0.00000260, -0.00000742, -0.00000028, -0.00000389, -0.00000323, 0.00000005, -0.00000661, -0.00000160, 0.00000070, -0.00001056, -0.00000639, 0.00000102, -0.00000269, -0.00000329, -0.00000930, -0.00001607, -0.00000745, -0.00000436, -0.00002151, -0.00003992, -0.00004097, -0.00002073, 0.00001157, 0.00003661, 0.00004250, 0.00003683, 0.00002911, 0.00002170, },
		{ 0.00002867, 0.00002221, 0.00002187, 0.00001955, 0.00001698, 0.00001545, 0.00001500, 0.00001402, 0.00001180, 0.00002076, 0.00003177, 0.00002992, 0.00002501, 0.00001460, 0.00001990, 0.00004070, 0.00003236, 0.00007170, 0.00013897, 0.00002552, -0.00013022, -0.00008163, -0.00001714, -0.00003895, -0.00001924, -0.00001100, -0.00001758, 0.00000874, -0.00000800, -0.00002454, -0.00000267, -0.00000652, -0.00000733, -0.00000450, -0.00001009, 0.00000677, -0.00000114, -0.00001470, 0.00000251, -0.00000648, -0.00000959, 0.00000336, -0.00000836, 0.00000428, 0.00001023, -0.00001694, -0.00000817, 0.00000366, -0.00000041, -0.00000003, -0.00001571, 0.00000102, 0.00001896, -0.00001797, -0.00001607, 0.00001342, 0.00001124, 0.00000293, -0.00002547, -0.00000314, 0.00004523, -0.00002923, -0.00004910, 0.00005527, 0.00001519, -0.00005744, 0.00002008, 0.00003280, -0.00003220, -0.00001237, 0.00001430, -0.00000251, -0.00000793, -0.00000506, 0.00000376, 0.00000695, -0.00000763, -0.00000771, 0.00000382, -0.00000027, -0.00000682, -0.00000316, 0.00000208, 0.00000061, -0.00000541, -0.00000124, 0.00000671, -0.00000171, -0.00000642, 0.00000431, 0.00000124, -0.00000907, -0.00000690, -0.00000415, 0.00000139, 0.00000324, -0.00000916, -0.00000990, -0.00000107, -0.00000026, 0.00000289, -0.00000095, -0.00000694, 0.00000139, -0.00000020, -0.00000721, -0.00000196, -0.00000357, -0.00000344, 0.00000221, -0.00000636, -0.00000920, -0.00000197, -0.00000427, -0.00000318, -0.00000187, -0.00001239, -0.00001340, -0.00000368, -0.00000716, -0.00002501, -0.00004218, -0.00004477, -0.00002373, 0.00001206, 0.00003907, 0.00004676, 0.00004058, },
		{ 0.00005336, 0.00004390, 0.00002800, 0.00002384, 0.00002539, 0.00002350, 0.00001760, 0.00001417, 0.00001761, 0.00001511, 0.00001261, 0.00002296, 0.00003225, 0.00003715, 0.00002692, 0.00001256, 0.00003277, 0.00003788, 0.00004751, 0.00013108, 0.00009452, -0.00009474, -0.00011954, -0.00002755, -0.00003562, -0.00003093, -0.00000552, -0.00002626, -0.00000532, 0.00000779, -0.00001852, -0.00000894, -0.00000714, -0.00001093, 0.00000102, -0.00001253, -0.00000577, 0.00000785, -0.00000889, -0.00000294, -0.00000556, -0.00001429, 0.00000808, -0.00000426, -0.00001322, 0.00001340, 0.00000080, -0.00001081, -0.00000354, -0.00000511, 0.00000956, -0.00000691, -0.00002227, 0.00001522, 0.00000702, -0.00001755, -0.00000318, 0.00000100, 0.00002379, 0.00000095, -0.00005234, 0.00002791, 0.00004871, -0.00007026, -0.00001763, 0.00007471, -0.00002235, -0.00004645, 0.00003619, 0.00001246, -0.00002962, -0.00000551, 0.00000894, 0.00000221, -0.00000869, -0.00000835, 0.00000765, 0.00000392, -0.00001046, -0.00000424, 0.00000346, -0.00000159, -0.00000565, -0.00000235, 0.00000258, -0.00000085, -0.00000519, 0.00000203, 0.00000352, -0.00000414, -0.00000029, 0.00000135, -0.00000706, -0.00000585, -0.00000353, -0.00000229, 0.00000416, -0.00000377, -0.00001331, -0.00000545, 0.00000063, 0.00000332, 0.00000084, -0.00000880, -0.00000170, 0.00000600, -0.00000403, -0.00000625, -0.00000430, -0.00000444, 0.00000254, -0.00000170, -0.00000951, -0.00000357, -0.00000433, -0.00000741, -0.00000112, -0.00000467, -0.00001247, -0.00000795, -0.00000418, -0.00001180, -0.00002666, -0.00004329, -0.00004779, -0.00002674, 0.00000985, 0.00004078, },
		{ 0.00000497, 0.00004405, 0.00006056, 0.00004660, 0.00002885, 0.00002518, 0.00002935, 0.00002757, 0.00001675, 0.00001511, 0.00001948, 0.00001560, 0.00001609, 0.00002252, 0.00003596, 0.00004418, 0.00002325, 0.00002243, 0.00003845, 0.00003523, 0.00010551, 0.00014149, -0.00003335, -0.00014311, -0.00004971, -0.00002647, -0.00004746, -0.00000610, -0.00001742, -0.00002463, 0.00000516, -0.00000873, -0.00000758, -0.00000224, -0.00001920, -0.00000173, -0.00000201, -0.00001529, 0.00000153, -0.00000304, -0.00000122, 0.00000110, -0.00002310, -0.00000065, 0.00001298, -0.00001694, -0.00000348, 0.00000744, -0.00000112, 0.00000244, -0.00001469, -0.00000118, 0.00001452, -0.00002113, -0.00000990, 0.00001085, -0.00000340, 0.00000538, -0.00001724, -0.00000298, 0.00005159, -0.00003396, -0.00005569, 0.00007347, 0.00001023, -0.00008630, 0.00003526, 0.00005565, -0.00005073, -0.00001711, 0.00003439, -0.00000529, -0.00002125, -0.00000252, 0.00000814, 0.00000290, -0.00001179, -0.00000539, 0.00000770, -0.00000149, -0.00000848, -0.00000169, 0.00000133, -0.00000167, -0.00000424, -0.00000145, 0.00000180, -0.00000215, -0.00000216, 0.00000197, -0.00000135, -0.00000023, 0.00000212, -0.00000745, -0.00000829, -0.00000120, -0.00000223, 0.00000165, 0.00000090, -0.00001140, -0.00001058, -0.00000235, 0.00000284, 0.00000480, -0.00000658, -0.00000875, 0.00000533, 0.00000368, -0.00000471, -0.00000618, -0.00000758, 0.00000013, 0.00000293, -0.00000719, -0.00000558, -0.00000238, -0.00000876, -0.00000583, -0.00000213, -0.00000780, -0.00000797, -0.00000483, -0.00000843, -0.00001522, -0.00002672, -0.00004383, -0.00004970, -0.00003145, },
		{ -0.00005141, -0.00003941, 0.00000000, 0.00004846, 0.00006663, 0.00005090, 0.00003035, 0.00002623, 0.00003421, 0.00002961, 0.00001743, 0.00001728, 0.00001873, 0.00001929, 0.00001842, 0.00002202, 0.00004399, 0.00004310, 0.00002702, 0.00003690, 0.00002905, 0.00007218, 0.00016210, 0.00004442, -0.00014518, -0.00008913, -0.00001181, -0.00005278, -0.00002560, -0.00000397, -0.00002866, -0.00000686, -0.00000818, -0.00001122, 0.00001041, -0.00001431, -0.00001694, 0.00000331, -0.00001053, -0.00000558, -0.00000570, -0.00000659, 0.00001368, -0.00001499, -0.00002295, 0.00001527, -0.00000046, -0.00001189, -0.00000126, -0.00000495, 0.00001511, -0.00000021, -0.00002272, 0.00001110, 0.00000130, -0.00001570, -0.00000225, -0.00001414, 0.00002083, 0.00001373, -0.00005277, 0.00003166, 0.00005401, -0.00008364, -0.00001187, 0.00008716, -0.00004713, -0.00005249, 0.00006645, 0.00000972, -0.00004682, 0.00000855, 0.00001621, -0.00001180, -0.00001154, -0.00000137, 0.00000773, -0.00000015, -0.00001107, -0.00000192, 0.00000294, -0.00000362, -0.00000444, -0.00000205, -0.00000032, -0.00000063, -0.00000300, -0.00000121, 0.00000008, -0.00000093, 0.00000035, -0.00000233, -0.00000150, 0.00000516, -0.00000360, -0.00001298, -0.00000409, 0.00000047, 0.00000088, 0.00000272, -0.00000737, -0.00001262, -0.00000601, -0.00000071, 0.00000552, 0.00000020, -0.00001091, -0.00000209, 0.00000621, 0.00000095, -0.00000339, -0.00000901, -0.00000599, 0.00000346, -0.00000234, -0.00000655, -0.00000162, -0.00000634, -0.00000909, -0.00000432, -0.00000609, -0.00000673, -0.00000281, -0.00000640, -0.00001305, -0.00001643, -0.00002649, -0.00004318, },
		{ -0.00002598, -0.00004092, -0.00005540, -0.00004900, -0.00000351, 0.00005135, 0.00007310, 0.00005715, 0.00003144, 0.00002861, 0.00003732, 0.00003119, 0.00002153, 0.00001651, 0.00001930, 0.00002494, 0.00001762, 0.00002756, 0.00004690, 0.00004033, 0.00004511, 0.00003436, 0.00003771, 0.00014855, 0.00012624, -0.00010786, -0.00014470, -0.00001544, -0.00003262, -0.00004832, -0.00000589, -0.00002301, -0.00001374, -0.00000622, -0.00002570, 0.00000849, 0.00000816, -0.00002196, -0.00000594, -0.00000653, -0.00000506, -0.00000212, -0.00002002, 0.00000838, 0.00000994, -0.00002844, -0.00000383, 0.00000805, -0.00000517, 0.00000120, -0.00001798, 0.00000151, 0.00002694, -0.00001472, -0.00001183, 0.00000427, -0.00000568, 0.00000884, -0.00003131, -0.00001651, 0.00006655, -0.00002673, -0.00005975, 0.00008366, 0.00000530, -0.00009577, 0.00004871, 0.00004624, -0.00007046, 0.00000538, 0.00004928, -0.00002349, -0.00001822, 0.00001112, -0.00000260, -0.00000646, -0.00000531, -0.00000133, 0.00000548, -0.00000250, -0.00000715, -0.00000265, -0.00000206, -0.00000143, -0.00000235, -0.00000386, -0.00000017, 0.00000049, -0.00000225, -0.00000210, -0.00000049, 0.00000207, -0.00000137, -0.00000563, 0.00000357, 0.00000404, -0.00001151, -0.00001105, -0.00000108, 0.00000181, 0.00000454, -0.00000291, -0.00001282, -0.00000842, -0.00000346, 0.00000166, 0.00000421, -0.00000630, -0.00000730, 0.00000243, 0.00000302, 0.00000120, -0.00000451, -0.00000967, -0.00000156, 0.00000030, -0.00000533, -0.00000181, -0.00000310, -0.00000909, -0.00000656, -0.00000644, -0.00000815, -0.00000226, -0.00000155, -0.00001064, -0.00001529, -0.00001687, },
		{ -0.00001554, -0.00001784, -0.00002372, -0.00003920, -0.00006210, -0.00005707, -0.00000779, 0.00005254, 0.00008175, 0.00006352, 0.00003396, 0.00003081, 0.00003753, 0.00003660, 0.00002482, 0.00001386, 0.00002426, 0.00002685, 0.00002052, 0.00003473, 0.00004026, 0.00005279, 0.00005736, 0.00002405, 0.00010243, 0.00018034, -0.00002240, -0.00018415, -0.00005998, -0.00000667, -0.00005136, -0.00001717, -0.00002017, -0.00002323, 0.00000406, -0.00002854, -0.00001706, 0.00001992, -0.00000623, -0.00001050, -0.00000989, -0.00001194, 0.00000863, -0.00001680, -0.00001636, 0.00001857, -0.00000979, -0.00001488, 0.00000166, -0.00000890, 0.00001211, -0.00000567, -0.00002833, 0.00002109, 0.00001562, -0.00001049, -0.00000635, -0.00002087, 0.00002483, 0.00000704, -0.00007734, 0.00003620, 0.00007417, -0.00008761, -0.00000918, 0.00009433, -0.00005835, -0.00004976, 0.00006783, -0.00001182, -0.00003844, 0.00003425, 0.00000606, -0.00001929, 0.00000414, -0.00000515, -0.00000769, 0.00000211, -0.00000380, -0.00000194, 0.00000233, -0.00000157, -0.00000452, -0.00000678, -0.00000302, 0.00000157, -0.00000330, -0.00000443, 0.00000096, 0.00000106, -0.00000229, -0.00000320, 0.00000143, 0.00000287, -0.00000563, -0.00000310, 0.00000653, -0.00000364, -0.00001338, -0.00000661, -0.00000048, 0.00000529, 0.00000296, -0.00001064, -0.00001122, -0.00000435, -0.00000164, 0.00000277, -0.00000170, -0.00000719, -0.00000102, 0.00000139, 0.00000140, 0.00000104, -0.00000629, -0.00000567, -0.00000131, -0.00000498, -0.00000265, -0.00000014, -0.00000676, -0.00000769, -0.00000618, -0.00000959, -0.00000568, 0.00000084, -0.00000493, -0.00001363, },
		{ -0.00000945, -0.00001363, -0.00001615, -0.00001802, -0.00002027, -0.00004061, -0.00006819, -0.00006390, -0.00001503, 0.00005456, 0.00009097, 0.00007068, 0.00003873, 0.00002929, 0.00003938, 0.00004452, 0.00002397, 0.00001560, 0.00002812, 0.00002737, 0.00003098, 0.00003119, 0.00004025, 0.00007533, 0.00004659, 0.00005770, 0.00018192, 0.00008096, -0.00017050, -0.00012878, -0.00000628, -0.00004052, -0.00002514, -0.00001143, -0.00003932, 0.00000126, -0.00000783, -0.00003708, 0.00000413, 0.00000592, -0.00000266, -0.00000196, -0.00002489, 0.00000148, 0.00000496, -0.00002676, 0.00000137, 0.00000201, -0.00001046, 0.00000635, -0.00001953, -0.00000287, 0.00002316, -0.00002561, -0.00001234, 0.00001858, 0.00000736, 0.00001351, -0.00003848, -0.00001787, 0.00007002, -0.00005030, -0.00007643, 0.00010544, 0.00001500, -0.00010131, 0.00005685, 0.00004403, -0.00007480, 0.00000876, 0.00003101, -0.00002913, 0.00001012, 0.00001428, -0.00001742, 0.00000583, 0.00000107, -0.00001881, -0.00000044, 0.00000564, -0.00000433, -0.00000320, 0.00000066, 0.00000103, -0.00000596, -0.00000965, -0.00000050, 0.00000176, -0.00000473, -0.00000315, 0.00000166, 0.00000107, -0.00000313, -0.00000256, 0.00000370, -0.00000045, -0.00000644, 0.00000208, 0.00000187, -0.00000960, -0.00000929, -0.00000474, 0.00000147, 0.00000690, -0.00000438, -0.00001294, -0.00000627, -0.00000228, 0.00000028, -0.00000051, -0.00000610, -0.00000230, 0.00000120, -0.00000088, 0.00000117, -0.00000112, -0.00000415, -0.00000219, -0.00000593, -0.00000591, 0.00000052, -0.00000276, -0.00000723, -0.00000576, -0.00000884, -0.00000929, -0.00000134, -0.00000097, },
		{ -0.00000011, -0.00000656, -0.00001125, -0.00001275, -0.00001729, -0.00001591, -0.00001916, -0.00004355, -0.00007131, -0.00007277, -0.00002327, 0.00005731, 0.00009939, 0.00008145, 0.00004181, 0.00002605, 0.00004596, 0.00004844, 0.00002458, 0.00002092, 0.00002641, 0.00003616, 0.00003710, 0.00002197, 0.00006035, 0.00007760, 0.00004878, 0.00015059, 0.00016045, -0.00010514, -0.00018588, -0.00003451, -0.00003105, -0.00003960, 0.00000567, -0.00003743, -0.00002522, 0.00000843, -0.00003012, -0.00001529, 0.00000077, -0.00000557, 0.00001591, -0.00001494, -0.00002298, 0.00001122, -0.00001272, -0.00001081, -0.00000176, -0.00001770, 0.00001742, -0.00000143, -0.00003360, 0.00001846, 0.00000633, -0.00001998, -0.00000023, -0.00000819, 0.00003820, 0.00000947, -0.00008460, 0.00004123, 0.00006500, -0.00011778, -0.00000528, 0.00011654, -0.00005934, -0.00004867, 0.00007156, -0.00001527, -0.00003486, 0.00002104, -0.00001222, 0.00000258, 0.00002258, -0.00002042, -0.00000486, 0.00001751, -0.00001590, -0.00001674, 0.00000693, 0.00000409, -0.00000465, -0.00000464, 0.00000156, 0.00000150, -0.00000931, -0.00000831, 0.00000127, -0.00000041, -0.00000421, -0.00000162, 0.00000166, 0.00000043, -0.00000349, -0.00000003, 0.00000257, -0.00000393, -0.00000187, 0.00000210, -0.00000618, -0.00000855, -0.00000549, -0.00000370, 0.00000458, 0.00000206, -0.00001053, -0.00000893, -0.00000274, -0.00000068, -0.00000031, -0.00000577, -0.00000462, 0.00000199, -0.00000014, -0.00000154, -0.00000059, -0.00000143, 0.00000044, -0.00000373, -0.00000931, -0.00000350, -0.00000050, -0.00000475, -0.00000486, -0.00000687, -0.00001067, -0.00000495, },
		{ -0.00000833, -0.00000060, -0.00000410, -0.00001124, -0.00001049, -0.00001327, -0.00001634, -0.00001396, -0.00002175, -0.00004383, -0.00007404, -0.00008378, -0.00003049, 0.00005789, 0.00010994, 0.00009397, 0.00004111, 0.00002752, 0.00005138, 0.00004908, 0.00003223, 0.00002106, 0.00002782, 0.00004993, 0.00002950, 0.00003074, 0.00007758, 0.00006151, 0.00012106, 0.00020878, -0.00000969, -0.00021586, -0.00008318, -0.00001738, -0.00006319, 0.00000171, -0.00000665, -0.00004421, -0.00000130, -0.00001699, -0.00002069, 0.00000054, -0.00002228, 0.00000924, 0.00001658, -0.00002630, -0.00000565, -0.00000498, -0.00000948, 0.00000609, -0.00003163, -0.00000344, 0.00003352, -0.00002714, -0.00001601, 0.00001260, -0.00000483, 0.00001244, -0.00003260, -0.00000724, 0.00008489, -0.00005178, -0.00007869, 0.00010820, -0.00000880, -0.00011945, 0.00007414, 0.00005476, -0.00007583, 0.00001176, 0.00002937, -0.00002446, 0.00000567, -0.00001122, -0.00001140, 0.00003464, -0.00000508, -0.00002717, 0.00001705, 0.00000488, -0.00002345, -0.00000590, 0.00000729, 0.00000202, -0.00000502, -0.00000470, 0.00000295, -0.00000144, -0.00001005, -0.00000504, -0.00000013, -0.00000170, -0.00000235, -0.00000101, 0.00000124, -0.00000056, -0.00000186, 0.00000126, -0.00000153, -0.00000239, 0.00000201, -0.00000461, -0.00000941, -0.00000421, -0.00000418, -0.00000055, 0.00000343, -0.00000620, -0.00000992, -0.00000369, -0.00000103, 0.00000008, -0.00000421, -0.00000751, -0.00000033, 0.00000241, -0.00000126, -0.00000251, -0.00000238, 0.00000191, 0.00000191, -0.00000726, -0.00000848, -0.00000319, -0.00000352, -0.00000328, -0.00000378, -0.00000971, },
		{ -0.00000557, -0.00001016, -0.00000311, -0.00000091, -0.00001007, -0.00001224, -0.00001065, -0.00001334, -0.00001307, -0.00001611, -0.00002411, -0.00004151, -0.00007905, -0.00009352, -0.00003921, 0.00005716, 0.00012418, 0.00010291, 0.00004289, 0.00003258, 0.00004998, 0.00005565, 0.00003999, 0.00001725, 0.00004019, 0.00005107, 0.00002691, 0.00005708, 0.00006250, 0.00009126, 0.00023084, 0.00011059, -0.00020466, -0.00015984, -0.00000480, -0.00006420, -0.00003784, 0.00001662, -0.00003227, -0.00001502, -0.00001400, -0.00003086, 0.00001035, -0.00001747, -0.00002447, 0.00002439, -0.00000153, -0.00001063, -0.00000895, -0.00002200, 0.00001937, -0.00001200, -0.00004410, 0.00002777, 0.00001218, -0.00002224, -0.00000381, -0.00002116, 0.00003331, 0.00001363, -0.00008475, 0.00005435, 0.00007862, -0.00012072, -0.00000349, 0.00010769, -0.00008405, -0.00004841, 0.00008574, -0.00001197, -0.00003233, 0.00001932, -0.00000939, 0.00000798, 0.00000067, -0.00003298, 0.00002240, 0.00002612, -0.00003018, -0.00000523, 0.00001772, -0.00001237, -0.00001519, 0.00000037, 0.00000383, 0.00000116, -0.00000540, -0.00000297, 0.00000166, -0.00000449, -0.00000709, -0.00000404, -0.00000257, -0.00000078, -0.00000107, -0.00000098, 0.00000031, -0.00000043, 0.00000061, -0.00000124, -0.00000333, 0.00000254, -0.00000091, -0.00001096, -0.00000653, -0.00000190, -0.00000202, 0.00000171, -0.00000384, -0.00001031, -0.00000464, -0.00000068, 0.00000002, -0.00000175, -0.00000763, -0.00000486, 0.00000190, 0.00000121, -0.00000200, -0.00000438, -0.00000091, 0.00000484, -0.00000047, -0.00000781, -0.00000706, -0.00000620, -0.00000427, -0.00000083, },
		{ -0.00000255, -0.00000069, -0.00000752, -0.00000634, -0.00000003, -0.00000592, -0.00001288, -0.00001215, -0.00001245, -0.00001007, -0.00001198, -0.00002103, -0.00002302, -0.00004071, -0.00008372, -0.00010213, -0.00005157, 0.00005926, 0.00013658, 0.00011035, 0.00005240, 0.00003195, 0.00004874, 0.00006860, 0.00003993, 0.00002338, 0.00004954, 0.00004132, 0.00005189, 0.00006126, 0.00005073, 0.00020382, 0.00023571, -0.00011621, -0.00024766, -0.00003596, -0.00002820, -0.00007441, 0.00000256, -0.00001433, -0.00001628, -0.00000109, -0.00004747, -0.00000406, 0.00001355, -0.00003710, -0.00000312, 0.00000922, 0.00000066, 0.00000706, -0.00003834, -0.00000381, 0.00002845, -0.00004337, -0.00001577, 0.00002165, -0.00000328, 0.00001393, -0.00004415, -0.00001959, 0.00009233, -0.00005282, -0.00008190, 0.00012567, 0.00000085, -0.00012074, 0.00007198, 0.00003769, -0.00008686, 0.00001926, 0.00003444, -0.00002052, 0.00000578, -0.00001282, -0.00000193, 0.00002495, -0.00002860, -0.00001241, 0.00003775, -0.00000721, -0.00002223, 0.00001179, -0.00000078, -0.00001414, -0.00000373, -0.00000098, 0.00000158, 0.00000037, -0.00000445, -0.00000195, -0.00000196, -0.00000398, -0.00000410, -0.00000582, -0.00000304, 0.00000063, -0.00000077, -0.00000131, -0.00000048, 0.00000153, 0.00000049, -0.00000473, 0.00000002, 0.00000367, -0.00000800, -0.00001102, -0.00000308, -0.00000084, 0.00000167, -0.00000191, -0.00001138, -0.00000759, -0.00000001, 0.00000053, -0.00000021, -0.00000504, -0.00000738, -0.00000167, 0.00000164, 0.00000006, -0.00000369, -0.00000420, 0.00000234, 0.00000394, -0.00000226, -0.00000598, -0.00000849, -0.00000891, },
		{ -0.00001128, -0.00000900, -0.00000147, -0.00000268, -0.00000590, -0.00000129, -0.00000273, -0.00001039, -0.00001213, -0.00001450, -0.00001209, -0.00000586, -0.00001557, -0.00002344, -0.00002159, -0.00004173, -0.00008513, -0.00011361, -0.00006352, 0.00006202, 0.00014321, 0.00012539, 0.00006247, 0.00002523, 0.00005654, 0.00007564, 0.00004234, 0.00003878, 0.00003952, 0.00005184, 0.00008072, 0.00003355, 0.00012206, 0.00029904, 0.00004985, -0.00027618, -0.00012918, -0.00000022, -0.00007387, -0.00002579, -0.00001115, -0.00002831, 0.00002134, -0.00003214, -0.00004478, 0.00002178, -0.00001294, -0.00002086, -0.00000316, -0.00001056, 0.00003097, -0.00001125, -0.00004732, 0.00002597, -0.00000296, -0.00003126, 0.00000235, -0.00001992, 0.00003928, 0.00001248, -0.00010530, 0.00005523, 0.00009017, -0.00013185, -0.00000251, 0.00012627, -0.00007788, -0.00005054, 0.00007573, -0.00002339, -0.00003050, 0.00002174, -0.00000740, 0.00001126, -0.00000268, -0.00002648, 0.00002467, 0.00000305, -0.00003102, 0.00002009, 0.00001404, -0.00002140, 0.00000243, 0.00000372, -0.00001390, -0.00000315, -0.00000028, -0.00000399, 0.00000070, -0.00000018, -0.00000251, -0.00000381, -0.00000395, -0.00000098, -0.00000423, -0.00000717, -0.00000176, 0.00000100, -0.00000084, -0.00000202, 0.00000031, 0.00000299, -0.00000276, -0.00000414, 0.00000381, -0.00000176, -0.00001157, -0.00000720, -0.00000194, 0.00000207, 0.00000204, -0.00000990, -0.00001270, -0.00000221, 0.00000181, 0.00000087, -0.00000208, -0.00000664, -0.00000480, -0.00000031, 0.00000057, -0.00000190, -0.00000455, -0.00000118, 0.00000319, 0.00000147, -0.00000141, -0.00000552, },
		{ -0.00000045, -0.00000708, -0.00001234, -0.00000796, -0.00000298, -0.00000338, -0.00000048, -0.00000057, -0.00000856, -0.00001017, -0.00001244, -0.00001729, -0.00000825, -0.00000611, -0.00001931, -0.00002293, -0.00002353, -0.00004013, -0.00008656, -0.00012694, -0.00007216, 0.00005753, 0.00015160, 0.00014801, 0.00006450, 0.00002501, 0.00006529, 0.00007564, 0.00006073, 0.00003991, 0.00002997, 0.00009018, 0.00006876, 0.00004981, 0.00026006, 0.00021244, -0.00019823, -0.00022868, -0.00002111, -0.00005736, -0.00004687, -0.00000148, -0.00005248, 0.00000834, 0.00001642, -0.00005707, -0.00000848, 0.00000279, -0.00001503, 0.00000276, -0.00003574, 0.00000735, 0.00004087, -0.00004133, -0.00001562, 0.00000978, -0.00001844, 0.00001715, -0.00004550, -0.00001990, 0.00010472, -0.00006654, -0.00009912, 0.00014191, 0.00000365, -0.00013612, 0.00007991, 0.00005354, -0.00008412, 0.00001143, 0.00002456, -0.00001877, 0.00000850, -0.00001555, 0.00000034, 0.00002312, -0.00002695, -0.00000395, 0.00002244, -0.00002035, 0.00000033, 0.00002002, -0.00001610, -0.00000422, 0.00001014, -0.00001468, -0.00000935, 0.00000527, -0.00000296, -0.00000488, -0.00000033, 0.00000066, -0.00000226, -0.00000658, -0.00000230, 0.00000037, -0.00000598, -0.00000622, -0.00000087, 0.00000069, -0.00000126, -0.00000208, 0.00000224, 0.00000120, -0.00000474, 0.00000020, 0.00000186, -0.00000800, -0.00000924, -0.00000470, -0.00000041, 0.00000480, -0.00000382, -0.00001505, -0.00000811, 0.00000098, 0.00000183, 0.00000034, -0.00000394, -0.00000576, -0.00000246, -0.00000026, -0.00000149, -0.00000390, -0.00000249, 0.00000126, 0.00000141, 0.00000075, },
		{ 0.00000009, 0.00000102, -0.00000134, -0.00000822, -0.00001133, -0.00000846, -0.00000563, 0.00000019, 0.00000401, -0.00000572, -0.00001112, -0.00000807, -0.00001548, -0.00001601, -0.00000630, -0.00001002, -0.00001865, -0.00002476, -0.00002577, -0.00003591, -0.00009119, -0.00013585, -0.00008448, 0.00004680, 0.00016970, 0.00016589, 0.00006671, 0.00003415, 0.00006221, 0.00008731, 0.00008018, 0.00002270, 0.00005498, 0.00010732, 0.00004695, 0.00017612, 0.00029453, -0.00005748, -0.00027692, -0.00007141, -0.00004360, -0.00007990, 0.00001306, -0.00004148, -0.00004481, 0.00003537, -0.00002336, -0.00002810, -0.00000393, -0.00002607, 0.00002166, -0.00001539, -0.00004831, 0.00003871, 0.00000854, -0.00002382, -0.00000266, -0.00004073, 0.00003818, 0.00001644, -0.00011802, 0.00006447, 0.00009820, -0.00015600, -0.00000306, 0.00014581, -0.00008742, -0.00006122, 0.00008756, -0.00001254, -0.00003417, 0.00001085, -0.00000632, 0.00001936, -0.00000530, -0.00002907, 0.00002295, 0.00000080, -0.00002173, 0.00001493, -0.00000586, -0.00000799, 0.00002278, -0.00000971, -0.00001669, 0.00001528, -0.00000477, -0.00001984, 0.00000215, 0.00000475, -0.00000502, -0.00000503, -0.00000074, 0.00000165, -0.00000446, -0.00000654, 0.00000011, -0.00000120, -0.00000610, -0.00000450, -0.00000103, 0.00000022, -0.00000183, -0.00000053, 0.00000237, -0.00000213, -0.00000210, 0.00000201, -0.00000466, -0.00000893, -0.00000560, -0.00000361, 0.00000285, 0.00000229, -0.00001139, -0.00001306, -0.00000300, 0.00000099, 0.00000134, -0.00000059, -0.00000443, -0.00000383, -0.00000097, -0.00000155, -0.00000431, -0.00000339, 0.00000050, 0.00000100, },
		{ 0.00000134, 0.00000012, -0.00000001, 0.00000004, -0.00000275, -0.00000744, -0.00001022, -0.00001128, -0.00000570, 0.00000632, 0.00000238, -0.00001122, -0.00000903, -0.00000916, -0.00001636, -0.00001269, -0.00001028, -0.00001095, -0.00001703, -0.00002972, -0.00002436, -0.00003472, -0.00009363, -0.00014176, -0.00010704, 0.00004159, 0.00018909, 0.00017660, 0.00008274, 0.00003644, 0.00005822, 0.00011462, 0.00007544, 0.00002609, 0.00009299, 0.00007642, 0.00012123, 0.00031234, 0.00008771, -0.00028096, -0.00013697, -0.00001020, -0.00011358, -0.00001906, 0.00000877, -0.00007108, 0.00000078, 0.00000141, -0.00002198, 0.00000961, -0.00004650, -0.00000648, 0.00003749, -0.00005053, -0.00001550, 0.00002045, -0.00000392, 0.00002779, -0.00006451, -0.00003453, 0.00011762, -0.00007688, -0.00011169, 0.00015979, -0.00000277, -0.00015555, 0.00009509, 0.00006312, -0.00009853, 0.00000885, 0.00003547, -0.00001579, -0.00000168, -0.00001917, 0.00001213, 0.00002775, -0.00003116, -0.00000772, 0.00001619, -0.00001484, 0.00000455, 0.00000034, -0.00001635, 0.00002282, 0.00000799, -0.00002829, 0.00000493, 0.00001129, -0.00001929, -0.00000962, 0.00000658, 0.00000096, -0.00000498, -0.00000545, 0.00000004, 0.00000043, -0.00000599, -0.00000379, -0.00000014, -0.00000280, -0.00000438, -0.00000379, -0.00000146, -0.00000060, -0.00000129, 0.00000106, -0.00000046, -0.00000233, 0.00000181, -0.00000217, -0.00000890, -0.00000574, -0.00000395, -0.00000094, 0.00000373, -0.00000540, -0.00001356, -0.00000689, -0.00000137, -0.00000036, 0.00000106, -0.00000131, -0.00000376, -0.00000160, -0.00000082, -0.00000451, -0.00000528, -0.00000080, },
		{ -0.00000326, 0.00000053, 0.00000076, 0.00000029, -0.00000085, -0.00000183, -0.00000295, -0.00000547, -0.00001100, -0.00001403, -0.00000133, 0.00000883, -0.00000371, -0.00001128, -0.00000804, -0.00001161, -0.00001218, -0.00001437, -0.00001527, -0.00000767, -0.00001940, -0.00003207, -0.00002353, -0.00003605, -0.00008869, -0.00015512, -0.00013133, 0.00004338, 0.00019672, 0.00019659, 0.00010450, 0.00002682, 0.00007403, 0.00012776, 0.00006553, 0.00006989, 0.00008089, 0.00008278, 0.00030161, 0.00024263, -0.00023740, -0.00025047, 0.00002011, -0.00008117, -0.00009599, 0.00002498, -0.00004315, -0.00003088, -0.00000284, -0.00003741, 0.00003338, -0.00001190, -0.00006553, 0.00003429, 0.00000210, -0.00003320, -0.00000291, -0.00003357, 0.00006155, 0.00001836, -0.00014574, 0.00006994, 0.00010843, -0.00018054, -0.00000386, 0.00015937, -0.00010335, -0.00006421, 0.00010482, -0.00001479, -0.00004431, 0.00001360, -0.00000123, 0.00001259, -0.00001383, -0.00001983, 0.00003561, 0.00000147, -0.00002514, 0.00000592, -0.00000825, 0.00000070, 0.00001018, -0.00002284, 0.00000702, 0.00002865, -0.00002060, -0.00001561, 0.00001532, -0.00000728, -0.00001660, 0.00000072, 0.00000340, -0.00000064, -0.00000493, -0.00000497, 0.00000007, -0.00000209, -0.00000437, -0.00000207, -0.00000220, -0.00000225, -0.00000293, -0.00000382, -0.00000205, -0.00000113, 0.00000044, 0.00000001, -0.00000283, 0.00000120, 0.00000110, -0.00000801, -0.00000773, -0.00000323, -0.00000228, 0.00000220, -0.00000149, -0.00001169, -0.00000885, -0.00000243, -0.00000287, -0.00000100, 0.00000088, -0.00000188, -0.00000190, 0.00000024, -0.00000317, -0.00000673, },
		{ -0.00000654, -0.00000549, -0.00000123, -0.00000011, 0.00000097, -0.00000002, -0.00000270, -0.00000298, -0.00000182, -0.00000409, -0.00001354, -0.00001285, 0.00000343, 0.00000436, -0.00000662, -0.00000857, -0.00001124, -0.00000909, -0.00000899, -0.00002122, -0.00001511, -0.00000670, -0.00002217, -0.00003118, -0.00002843, -0.00003134, -0.00008425, -0.00017633, -0.00014570, 0.00003591, 0.00020088, 0.00023279, 0.00011340, 0.00002924, 0.00009523, 0.00011356, 0.00009901, 0.00009687, 0.00003852, 0.00023034, 0.00039095, -0.00008139, -0.00038324, -0.00004995, 0.00001444, -0.00012665, -0.00001982, -0.00001956, -0.00003762, 0.00001189, -0.00006517, -0.00000524, 0.00005793, -0.00005912, -0.00002506, 0.00001728, -0.00001342, 0.00002333, -0.00007285, -0.00002310, 0.00014568, -0.00009249, -0.00012968, 0.00018082, -0.00000982, -0.00018009, 0.00010293, 0.00006428, -0.00010898, 0.00001847, 0.00004351, -0.00002195, -0.00000439, -0.00001685, 0.00000713, 0.00001625, -0.00002847, 0.00000803, 0.00002404, -0.00001380, -0.00000218, -0.00000959, -0.00001098, 0.00001947, -0.00001154, -0.00001655, 0.00002987, 0.00000285, -0.00002516, 0.00000533, 0.00000231, -0.00001585, -0.00000410, 0.00000187, -0.00000013, -0.00000079, -0.00000492, -0.00000389, -0.00000197, -0.00000261, -0.00000164, -0.00000307, -0.00000297, -0.00000076, -0.00000258, -0.00000393, -0.00000286, -0.00000031, 0.00000101, -0.00000285, -0.00000125, 0.00000330, -0.00000421, -0.00000959, -0.00000468, -0.00000218, 0.00000115, 0.00000089, -0.00000965, -0.00001094, -0.00000231, -0.00000244, -0.00000439, -0.00000047, -0.00000017, -0.00000159, 0.00000113, -0.00000054, },
		{ 0.00000281, -0.00000432, -0.00000695, -0.00000256, -0.00000149, -0.00000026, 0.00000106, -0.00000173, -0.00000394, -0.00000308, -0.00000021, -0.00000484, -0.00001446, -0.00000775, 0.00000184, -0.00000160, -0.00000350, -0.00000997, -0.00001285, -0.00000280, -0.00001272, -0.00002513, -0.00001350, -0.00000970, -0.00001960, -0.00003465, -0.00003285, -0.00002078, -0.00009014, -0.00019010, -0.00015912, 0.00001269, 0.00022057, 0.00026503, 0.00012020, 0.00005510, 0.00008515, 0.00011834, 0.00016134, 0.00005670, 0.00010077, 0.00042846, 0.00018190, -0.00039750, -0.00023821, 0.00005328, -0.00007614, -0.00005610, -0.00001939, -0.00006792, 0.00003381, -0.00002706, -0.00008684, 0.00005399, 0.00001163, -0.00004028, -0.00000510, -0.00004471, 0.00006020, 0.00001161, -0.00016267, 0.00009378, 0.00012686, -0.00020568, 0.00000017, 0.00017730, -0.00012470, -0.00007879, 0.00010933, -0.00001917, -0.00004157, 0.00002293, -0.00000203, 0.00001047, -0.00001223, -0.00002440, 0.00002147, -0.00000392, -0.00001161, 0.00001892, -0.00000217, 0.00000086, -0.00000072, -0.00002512, 0.00000902, 0.00000987, -0.00002365, 0.00001197, 0.00001972, -0.00001891, -0.00000492, 0.00000588, -0.00001455, -0.00000775, 0.00000292, -0.00000189, -0.00000107, -0.00000125, -0.00000395, -0.00000428, -0.00000389, -0.00000057, -0.00000112, -0.00000445, -0.00000184, -0.00000015, -0.00000258, -0.00000430, -0.00000290, 0.00000089, -0.00000108, -0.00000349, 0.00000209, 0.00000019, -0.00000832, -0.00000703, -0.00000319, 0.00000012, 0.00000303, -0.00000611, -0.00001345, -0.00000455, 0.00000029, -0.00000468, -0.00000360, -0.00000066, -0.00000194, 0.00000107, },
		{ -0.00000165, 0.00000479, 0.00000030, -0.00000695, -0.00000387, -0.00000172, -0.00000192, 0.00000025, -0.00000088, -0.00000309, -0.00000413, -0.00000237, 0.00000024, -0.00000664, -0.00001029, -0.00000525, -0.00000589, -0.00000136, -0.00000106, -0.00001429, -0.00000803, -0.00000154, -0.00001821, -0.00002368, -0.00001721, -0.00000934, -0.00001649, -0.00004308, -0.00002877, -0.00001794, -0.00009713, -0.00019152, -0.00019012, -0.00000664, 0.00024820, 0.00028239, 0.00015543, 0.00006869, 0.00005831, 0.00017935, 0.00016848, 0.00004013, 0.00031748, 0.00040004, -0.00022447, -0.00040724, -0.00003068, -0.00002949, -0.00006221, 0.00001455, -0.00010075, -0.00002121, 0.00005552, -0.00009192, -0.00002677, 0.00003390, -0.00001191, 0.00003070, -0.00008424, -0.00003379, 0.00015636, -0.00011388, -0.00014347, 0.00021222, -0.00000939, -0.00019620, 0.00012175, 0.00006704, -0.00013121, 0.00000968, 0.00004063, -0.00001988, 0.00000523, -0.00001535, 0.00000518, 0.00001997, -0.00002925, -0.00000674, 0.00000979, -0.00000718, 0.00001293, -0.00000046, -0.00000450, 0.00001469, -0.00001864, -0.00001402, 0.00001589, -0.00001287, -0.00000436, 0.00002267, -0.00000880, -0.00001146, 0.00000931, -0.00001098, -0.00001556, 0.00000334, 0.00000123, -0.00000400, -0.00000141, -0.00000131, -0.00000316, -0.00000602, -0.00000323, 0.00000096, -0.00000251, -0.00000394, -0.00000073, -0.00000032, -0.00000268, -0.00000463, -0.00000178, 0.00000009, -0.00000331, -0.00000067, 0.00000195, -0.00000512, -0.00000772, -0.00000452, -0.00000193, 0.00000331, -0.00000088, -0.00001314, -0.00000957, 0.00000084, -0.00000183, -0.00000480, -0.00000184, -0.00000327, },
		{ -0.00000286, -0.00000594, 0.00000211, 0.00000495, -0.00000401, -0.00000499, -0.00000166, -0.00000234, -0.00000070, -0.00000138, -0.00000343, -0.00000298, -0.00000376, -0.00000147, -0.00000174, -0.00000505, -0.00000390, -0.00001079, -0.00001103, 0.00000301, -0.00000430, -0.00001327, -0.00000325, -0.00000689, -0.00001722, -0.00002477, -0.00002245, -0.00000383, -0.00002039, -0.00004522, -0.00002482, -0.00002612, -0.00008847, -0.00020017, -0.00023151, -0.00001016, 0.00025697, 0.00031423, 0.00020872, 0.00004384, 0.00008014, 0.00023364, 0.00012099, 0.00020007, 0.00046735, 0.00002361, -0.00045122, -0.00015880, -0.00002527, -0.00010869, 0.00005384, -0.00003580, -0.00012297, 0.00005223, -0.00001428, -0.00006198, 0.00000349, -0.00004696, 0.00007696, 0.00002137, -0.00018997, 0.00010107, 0.00013592, -0.00024161, -0.00000008, 0.00020189, -0.00013377, -0.00007673, 0.00012607, -0.00002724, -0.00005821, 0.00001172, -0.00000457, 0.00002136, -0.00000628, -0.00002784, 0.00002428, 0.00000129, -0.00002113, -0.00000123, -0.00000642, 0.00001404, 0.00001004, -0.00001616, 0.00001200, 0.00000384, -0.00002195, 0.00000486, -0.00000338, -0.00001230, 0.00002027, 0.00000194, -0.00001878, 0.00000879, 0.00000062, -0.00002154, -0.00000435, 0.00000589, -0.00000229, -0.00000410, -0.00000199, -0.00000064, -0.00000401, -0.00000645, -0.00000126, 0.00000009, -0.00000304, -0.00000237, -0.00000060, -0.00000055, -0.00000307, -0.00000393, -0.00000128, -0.00000254, -0.00000239, 0.00000158, -0.00000250, -0.00000725, -0.00000491, -0.00000334, 0.00000107, 0.00000284, -0.00000889, -0.00001343, -0.00000232, 0.00000050, -0.00000406, -0.00000170, },
		{ -0.00000192, 0.00000041, -0.00000691, -0.00000434, 0.00000508, 0.00000040, -0.00000433, -0.00000150, -0.00000232, -0.00000096, -0.00000058, -0.00000518, -0.00000414, -0.00000247, -0.00000257, -0.00000239, -0.00000358, 0.00000186, -0.00000347, -0.00001935, -0.00000877, 0.00000269, -0.00000674, -0.00000675, -0.00000684, -0.00000921, -0.00001331, -0.00003153, -0.00002129, -0.00000293, -0.00002446, -0.00003883, -0.00003542, -0.00002672, -0.00007226, -0.00022756, -0.00025751, -0.00002297, 0.00025432, 0.00038325, 0.00023069, 0.00003303, 0.00014913, 0.00019938, 0.00018598, 0.00047056, 0.00024831, -0.00041373, -0.00028851, 0.00000541, -0.00016818, -0.00002784, 0.00008979, -0.00011876, -0.00004006, 0.00001520, -0.00004406, 0.00003521, -0.00009677, -0.00003678, 0.00019636, -0.00012582, -0.00016910, 0.00023961, -0.00001830, -0.00023266, 0.00013127, 0.00007788, -0.00013459, 0.00002199, 0.00004744, -0.00002895, -0.00000990, -0.00002822, 0.00001033, 0.00003164, -0.00003010, -0.00000803, 0.00001693, -0.00000777, -0.00000554, -0.00001507, 0.00000183, 0.00002562, -0.00001001, -0.00000637, 0.00001408, -0.00001143, -0.00000330, 0.00000043, -0.00001895, 0.00001223, 0.00001654, -0.00002077, -0.00000290, 0.00001252, -0.00001577, -0.00001528, 0.00000371, 0.00000215, -0.00000335, -0.00000426, -0.00000186, -0.00000075, -0.00000514, -0.00000468, -0.00000092, -0.00000141, -0.00000187, -0.00000148, -0.00000080, -0.00000097, -0.00000305, -0.00000269, -0.00000295, -0.00000372, 0.00000046, -0.00000047, -0.00000660, -0.00000551, -0.00000319, -0.00000116, 0.00000330, -0.00000367, -0.00001379, -0.00000666, 0.00000090, -0.00000341, },
		{ -0.00000230, -0.00000425, 0.00000263, -0.00000251, -0.00000855, 0.00000037, 0.00000216, -0.00000298, -0.00000030, -0.00000128, -0.00000210, 0.00000135, -0.00000400, -0.00000761, -0.00000382, -0.00000164, -0.00000146, -0.00000565, -0.00000026, 0.00000770, -0.00001076, -0.00002114, -0.00000655, -0.00000233, -0.00000110, -0.00000536, -0.00001382, -0.00000427, -0.00001548, -0.00003487, -0.00001887, -0.00000933, -0.00001664, -0.00004064, -0.00005195, -0.00001014, -0.00007329, -0.00025006, -0.00027339, -0.00006627, 0.00028147, 0.00044945, 0.00022789, 0.00009799, 0.00015713, 0.00016593, 0.00047379, 0.00050188, -0.00029059, -0.00048865, 0.00002047, -0.00009111, -0.00019740, 0.00008805, -0.00000516, -0.00007138, -0.00000707, -0.00009396, 0.00007654, 0.00002456, -0.00023114, 0.00012832, 0.00017495, -0.00027831, -0.00000177, 0.00023107, -0.00016086, -0.00010230, 0.00013381, -0.00002306, -0.00005113, 0.00002204, -0.00000327, 0.00001360, -0.00002311, -0.00003499, 0.00003437, 0.00000638, -0.00002373, 0.00000312, -0.00000099, 0.00000317, -0.00000888, -0.00001876, 0.00002063, 0.00001085, -0.00001348, 0.00000707, -0.00000252, -0.00000453, 0.00000761, -0.00001986, -0.00000609, 0.00002548, -0.00000863, -0.00001759, 0.00001206, -0.00000262, -0.00001881, -0.00000337, 0.00000285, -0.00000092, -0.00000330, -0.00000437, -0.00000146, -0.00000205, -0.00000442, -0.00000303, -0.00000231, -0.00000138, -0.00000058, -0.00000134, -0.00000102, -0.00000149, -0.00000206, -0.00000286, -0.00000510, -0.00000192, 0.00000084, -0.00000493, -0.00000680, -0.00000333, -0.00000186, 0.00000228, 0.00000024, -0.00001169, -0.00001051, 0.00000023, },
		{ -0.00000224, 0.00000035, -0.00000617, 0.00000050, 0.00000283, -0.00000709, -0.00000416, 0.00000102, -0.00000385, -0.00000041, 0.00000206, -0.00000256, 0.00000104, 0.00000002, -0.00000785, -0.00000764, -0.00000401, 0.00000028, -0.00000299, -0.00000666, 0.00000701, 0.00000470, -0.00001610, -0.00001703, -0.00001175, -0.00000179, 0.00000615, -0.00001271, -0.00001351, -0.00000151, -0.00002001, -0.00003099, -0.00002618, -0.00000894, -0.00000522, -0.00005811, -0.00005178, 0.00000044, -0.00008656, -0.00024714, -0.00031610, -0.00010788, 0.00033456, 0.00046909, 0.00028723, 0.00017531, 0.00007236, 0.00035000, 0.00074830, 0.00005784, -0.00068042, -0.00017307, 0.00009055, -0.00023049, -0.00006021, 0.00003698, -0.00003991, 0.00005649, -0.00014565, -0.00007448, 0.00023117, -0.00016223, -0.00020609, 0.00029872, -0.00000914, -0.00026992, 0.00015418, 0.00009071, -0.00016634, 0.00000569, 0.00004898, -0.00002191, 0.00000048, -0.00002306, 0.00001126, 0.00002217, -0.00004400, -0.00000708, 0.00002352, -0.00000841, -0.00000382, -0.00000813, -0.00000144, 0.00000847, -0.00001923, -0.00000136, 0.00001906, -0.00000511, 0.00000133, 0.00000014, -0.00000949, 0.00001295, -0.00000600, -0.00002263, 0.00001682, 0.00000909, -0.00002107, 0.00000146, 0.00000604, -0.00001513, -0.00000826, 0.00000132, -0.00000124, -0.00000167, -0.00000355, -0.00000379, -0.00000207, -0.00000263, -0.00000250, -0.00000330, -0.00000315, -0.00000031, -0.00000017, -0.00000128, -0.00000147, -0.00000128, -0.00000134, -0.00000469, -0.00000487, -0.00000010, -0.00000257, -0.00000717, -0.00000466, -0.00000229, 0.00000111, 0.00000293, -0.00000786, -0.00001362, },
		{ -0.00001423, -0.00000722, 0.00000223, -0.00000480, -0.00000368, 0.00000463, -0.00000287, -0.00000582, 0.00000073, -0.00000500, -0.00000502, 0.00000395, 0.00000052, -0.00000037, 0.00000242, -0.00000452, -0.00000836, -0.00000794, -0.00000345, 0.00000121, -0.00000586, -0.00000122, 0.00000879, -0.00000322, -0.00001224, -0.00001791, -0.00001772, 0.00000649, 0.00000392, -0.00001760, -0.00000901, -0.00000620, -0.00001595, -0.00003202, -0.00003664, 0.00000414, -0.00000995, -0.00006924, -0.00004103, -0.00001419, -0.00007733, -0.00024691, -0.00038510, -0.00011183, 0.00034724, 0.00049894, 0.00043241, 0.00014185, 0.00009517, 0.00073599, 0.00057749, -0.00057049, -0.00055550, 0.00010124, -0.00008461, -0.00014596, -0.00000687, -0.00011165, 0.00012543, 0.00003263, -0.00031601, 0.00014369, 0.00020109, -0.00035250, 0.00000035, 0.00028955, -0.00018124, -0.00011659, 0.00016186, -0.00002741, -0.00007710, 0.00000806, -0.00000311, 0.00002382, -0.00001614, -0.00003137, 0.00003406, -0.00000339, -0.00002906, 0.00000586, -0.00000178, 0.00000308, -0.00000200, -0.00001657, 0.00000756, -0.00000201, -0.00001159, 0.00001133, 0.00000040, 0.00000017, 0.00000664, -0.00001394, 0.00000440, 0.00001186, -0.00002101, -0.00000128, 0.00001683, -0.00001405, -0.00000844, 0.00000821, -0.00001042, -0.00001167, 0.00000151, -0.00000128, -0.00000328, -0.00000205, -0.00000345, -0.00000343, -0.00000307, -0.00000153, -0.00000173, -0.00000427, -0.00000251, 0.00000026, -0.00000015, -0.00000135, -0.00000171, -0.00000050, -0.00000230, -0.00000567, -0.00000277, -0.00000163, -0.00000626, -0.00000598, -0.00000317, -0.00000053, 0.00000401, -0.00000250, },
		{ 0.00000223, -0.00001086, -0.00001216, 0.00000069, -0.00000153, -0.00000637, 0.00000333, 0.00000112, -0.00000660, 0.00000112, -0.00000171, -0.00001028, -0.00000059, 0.00000374, 0.00000065, 0.00000340, -0.00000162, -0.00000666, -0.00000736, -0.00000885, -0.00000186, -0.00000130, -0.00000364, 0.00000495, 0.00000078, -0.00000401, -0.00000686, -0.00002567, -0.00001425, 0.00001008, -0.00000340, -0.00001335, -0.00001221, -0.00000742, -0.00000801, -0.00004383, -0.00003342, 0.00001001, -0.00002384, -0.00005744, -0.00005264, -0.00002750, -0.00004469, -0.00028298, -0.00042702, -0.00012349, 0.00030834, 0.00063597, 0.00053304, 0.00007495, 0.00042650, 0.00088837, -0.00009609, -0.00078362, -0.00014408, 0.00000303, -0.00013699, 0.00005448, -0.00020181, -0.00007876, 0.00032215, -0.00021517, -0.00027170, 0.00035907, -0.00003230, -0.00033991, 0.00019152, 0.00011956, -0.00019032, 0.00001374, 0.00006557, -0.00003266, -0.00001937, -0.00003622, 0.00001365, 0.00003002, -0.00003999, -0.00000388, 0.00002097, -0.00001219, -0.00000261, -0.00001085, -0.00000578, 0.00001333, -0.00001263, -0.00000919, 0.00000401, -0.00000673, 0.00000690, 0.00000144, -0.00000676, 0.00001194, -0.00000576, -0.00001000, 0.00001523, -0.00000642, -0.00001164, 0.00001451, -0.00000565, -0.00001537, 0.00000737, -0.00000415, -0.00001633, -0.00000031, 0.00000263, -0.00000444, -0.00000369, -0.00000258, -0.00000289, -0.00000389, -0.00000286, -0.00000035, -0.00000243, -0.00000410, -0.00000161, 0.00000016, -0.00000017, -0.00000159, -0.00000118, -0.00000063, -0.00000390, -0.00000429, -0.00000228, -0.00000551, -0.00000693, -0.00000399, -0.00000211, 0.00000274, },
		{ 0.00000046, 0.00000408, -0.00000538, -0.00001401, -0.00000329, 0.00000105, -0.00000717, 0.00000028, 0.00000508, -0.00000643, -0.00000183, 0.00000411, -0.00000891, -0.00000811, 0.00000155, 0.00000157, 0.00000535, 0.00000198, -0.00000647, -0.00000498, -0.00000785, -0.00000794, -0.00000300, -0.00000295, 0.00000472, 0.00000157, -0.00000691, 0.00000331, -0.00001026, -0.00002802, -0.00000734, 0.00000265, -0.00000210, -0.00001088, -0.00002019, 0.00000157, -0.00001131, -0.00004949, -0.00002347, -0.00000502, -0.00001675, -0.00004871, -0.00008703, -0.00000753, -0.00003618, -0.00032020, -0.00043436, -0.00021627, 0.00032563, 0.00082797, 0.00051180, 0.00027912, 0.00083368, 0.00038074, -0.00070003, -0.00039055, -0.00001903, -0.00024750, 0.00012960, 0.00002019, -0.00043096, 0.00021291, 0.00026902, -0.00045154, -0.00000102, 0.00034195, -0.00024467, -0.00014915, 0.00020036, -0.00002507, -0.00008439, 0.00002123, 0.00000120, 0.00001457, -0.00003232, -0.00004128, 0.00003588, -0.00000005, -0.00002586, 0.00000673, -0.00000203, 0.00000788, -0.00000201, -0.00002519, 0.00000593, 0.00000589, -0.00001267, -0.00000350, -0.00000649, 0.00000648, 0.00000985, -0.00001289, 0.00000353, 0.00000735, -0.00001285, 0.00000612, 0.00000427, -0.00001257, 0.00001060, 0.00000361, -0.00002052, 0.00000088, 0.00000479, -0.00001726, -0.00000798, 0.00000562, -0.00000085, -0.00000531, -0.00000417, -0.00000258, -0.00000277, -0.00000406, -0.00000169, -0.00000059, -0.00000288, -0.00000301, -0.00000137, -0.00000004, -0.00000043, -0.00000139, -0.00000067, -0.00000208, -0.00000385, -0.00000257, -0.00000487, -0.00000801, -0.00000533, -0.00000289, },
		{ -0.00000400, -0.00000127, 0.00000361, -0.00000057, -0.00001281, -0.00000767, 0.00000250, -0.00000597, -0.00000478, 0.00000751, -0.00000222, -0.00000714, 0.00000532, -0.00000170, -0.00001060, -0.00000344, -0.00000166, 0.00000441, 0.00000811, -0.00000433, -0.00000652, -0.00000392, -0.00000773, -0.00000697, -0.00000742, 0.00000290, 0.00000893, -0.00000918, -0.00000402, 0.00000578, -0.00001606, -0.00002006, -0.00001134, -0.00000314, 0.00000588, -0.00001852, -0.00001842, 0.00000632, -0.00002041, -0.00003760, -0.00003341, -0.00001614, 0.00000999, -0.00007265, -0.00009974, 0.00001565, -0.00006119, -0.00030056, -0.00049481, -0.00034231, 0.00045869, 0.00090070, 0.00061283, 0.00084262, 0.00068882, -0.00055632, -0.00059201, 0.00008850, -0.00036387, -0.00017699, 0.00042348, -0.00032805, -0.00036236, 0.00048448, -0.00003766, -0.00042112, 0.00023853, 0.00012709, -0.00025159, 0.00001087, 0.00008162, -0.00003390, -0.00001229, -0.00003026, 0.00001322, 0.00002373, -0.00005201, -0.00001132, 0.00001869, -0.00001135, -0.00000136, -0.00001067, 0.00000177, 0.00002139, -0.00001859, -0.00001814, 0.00000713, -0.00000281, -0.00000391, -0.00001040, -0.00000315, 0.00001839, -0.00000402, -0.00001003, 0.00000861, -0.00000588, -0.00000116, 0.00000813, -0.00001163, 0.00000423, 0.00001513, -0.00001747, -0.00001208, 0.00000899, -0.00001018, -0.00001599, 0.00000206, 0.00000376, -0.00000314, -0.00000537, -0.00000454, -0.00000232, -0.00000301, -0.00000317, -0.00000122, -0.00000145, -0.00000223, -0.00000221, -0.00000147, -0.00000035, -0.00000066, -0.00000070, -0.00000131, -0.00000311, -0.00000223, -0.00000326, -0.00000810, -0.00000767, },
		{ -0.00000912, -0.00000633, -0.00000315, 0.00000204, 0.00000287, -0.00000952, -0.00001192, 0.00000196, -0.00000165, -0.00000953, 0.00000476, 0.00000479, -0.00000874, 0.00000071, 0.00000387, -0.00000768, -0.00000494, -0.00000468, -0.00000283, 0.00001013, 0.00000343, -0.00000768, -0.00000440, -0.00000453, -0.00000504, -0.00001044, -0.00000760, 0.00001187, 0.00000158, -0.00001300, 0.00000200, -0.00000116, -0.00001068, -0.00001583, -0.00002322, 0.00000272, 0.00000531, -0.00002356, -0.00001037, -0.00000385, -0.00001237, -0.00003081, -0.00006035, 0.00000274, 0.00001447, -0.00010099, -0.00008058, -0.00000974, -0.00004904, -0.00026661, -0.00065253, -0.00036759, 0.00054777, 0.00090639, 0.00120537, 0.00123917, -0.00031861, -0.00107467, 0.00014764, 0.00001467, -0.00072053, 0.00028324, 0.00034019, -0.00064501, -0.00000197, 0.00044748, -0.00030139, -0.00017247, 0.00024443, -0.00004931, -0.00011810, 0.00001881, 0.00000060, 0.00002332, -0.00002423, -0.00004229, 0.00003589, -0.00000301, -0.00003346, -0.00000202, -0.00000781, 0.00000664, -0.00000345, -0.00001983, 0.00002015, 0.00000995, -0.00002200, -0.00000773, -0.00000315, 0.00000061, -0.00000388, -0.00001447, 0.00001127, 0.00001136, -0.00001096, 0.00000088, -0.00000203, -0.00000494, 0.00001082, -0.00000717, -0.00000727, 0.00002015, -0.00000270, -0.00002121, 0.00000285, -0.00000116, -0.00001770, -0.00000517, 0.00000438, 0.00000004, -0.00000355, -0.00000564, -0.00000433, -0.00000238, -0.00000269, -0.00000218, -0.00000177, -0.00000154, -0.00000131, -0.00000194, -0.00000158, -0.00000085, -0.00000042, -0.00000056, -0.00000251, -0.00000242, -0.00000156, -0.00000599, },
		{ -0.00000322, -0.00000799, -0.00000810, -0.00000564, -0.00000096, 0.00000433, -0.00000424, -0.00001451, -0.00000243, 0.00000369, -0.00000939, -0.00000222, 0.00000876, -0.00000497, -0.00000438, 0.00000528, -0.00000459, -0.00000462, -0.00000172, -0.00000891, 0.00000251, 0.00000941, -0.00000287, -0.00000563, -0.00000506, -0.00000261, -0.00000444, -0.00001503, 0.00000034, 0.00001114, -0.00000724, -0.00000623, -0.00000255, -0.00000341, 0.00000117, -0.00002523, -0.00002584, 0.00000796, -0.00000244, -0.00001493, -0.00001650, -0.00000950, 0.00000938, -0.00005140, -0.00006579, 0.00002393, -0.00001220, -0.00008557, -0.00009126, -0.00004996, 0.00003318, -0.00032713, -0.00078305, -0.00033441, 0.00042565, 0.00129684, 0.00216000, 0.00069357, -0.00159759, -0.00058947, 0.00072494, -0.00058784, -0.00057474, 0.00068124, -0.00009255, -0.00059472, 0.00030238, 0.00016649, -0.00029956, 0.00002656, 0.00009586, -0.00005512, -0.00002893, -0.00004646, 0.00001319, 0.00003791, -0.00005006, -0.00001408, 0.00002096, -0.00001022, -0.00000756, -0.00002216, -0.00000595, 0.00001656, -0.00001854, -0.00000470, 0.00002089, -0.00000433, -0.00001055, -0.00001076, -0.00000690, 0.00000478, -0.00001087, -0.00000356, 0.00001465, -0.00000217, -0.00000239, -0.00000087, -0.00001035, 0.00000876, 0.00000389, -0.00001503, 0.00001165, 0.00001309, -0.00001758, -0.00000701, 0.00000265, -0.00001483, -0.00001117, 0.00000224, 0.00000107, -0.00000168, -0.00000378, -0.00000560, -0.00000393, -0.00000243, -0.00000187, -0.00000206, -0.00000219, -0.00000095, -0.00000088, -0.00000179, -0.00000186, -0.00000112, -0.00000013, -0.00000134, -0.00000266, -0.00000129, },
		{ -0.00000166, -0.00000176, -0.00000563, -0.00000778, -0.00000701, -0.00000465, 0.00000233, 0.00000101, -0.00001282, -0.00000913, 0.00000512, -0.00000427, -0.00000782, 0.00000765, 0.00000037, -0.00000753, 0.00000495, -0.00000122, -0.00000735, 0.00000221, -0.00000554, -0.00000667, 0.00000620, 0.00000177, -0.00000277, -0.00000451, -0.00000588, 0.00000101, -0.00000885, -0.00001154, 0.00000573, 0.00000028, -0.00000372, -0.00000372, -0.00001246, 0.00000755, 0.00000150, -0.00003441, -0.00001916, 0.00000004, 0.00000175, -0.00000915, -0.00003443, 0.00000839, 0.00001050, -0.00007274, -0.00004344, 0.00000319, -0.00001301, -0.00005503, -0.00016173, -0.00001656, 0.00008123, -0.00044829, -0.00077521, -0.00052612, 0.00035909, 0.00244202, 0.00262164, -0.00088836, -0.00208014, 0.00050498, 0.00051874, -0.00106720, -0.00000504, 0.00061751, -0.00044960, -0.00026324, 0.00030107, -0.00005322, -0.00012868, 0.00003322, -0.00000021, 0.00001858, -0.00004460, -0.00006399, 0.00004235, 0.00000708, -0.00003585, -0.00000306, -0.00000214, 0.00001026, -0.00001116, -0.00003265, 0.00000949, 0.00000322, -0.00001506, 0.00000930, 0.00000482, -0.00000172, -0.00000499, -0.00001741, -0.00000229, -0.00000084, -0.00000789, 0.00000836, 0.00000264, -0.00000226, 0.00000391, -0.00001183, -0.00000210, 0.00001236, -0.00001057, -0.00000201, 0.00001801, -0.00000628, -0.00001201, 0.00000279, -0.00001079, -0.00001617, -0.00000081, 0.00000182, -0.00000200, -0.00000219, -0.00000415, -0.00000496, -0.00000364, -0.00000209, -0.00000147, -0.00000234, -0.00000186, -0.00000052, -0.00000073, -0.00000173, -0.00000218, -0.00000109, -0.00000056, -0.00000193, },
		{ -0.00000137, -0.00000148, -0.00000118, -0.00000380, -0.00000659, -0.00000649, -0.00000635, -0.00000193, 0.00000261, -0.00000785, -0.00001351, 0.00000157, 0.00000116, -0.00000955, 0.00000408, 0.00000571, -0.00000939, 0.00000166, 0.00000494, -0.00000945, -0.00000084, 0.00000187, -0.00000727, 0.00000027, 0.00000024, -0.00000256, 0.00000071, -0.00000664, -0.00000277, -0.00000040, -0.00001011, -0.00000212, -0.00000252, -0.00000431, 0.00000575, -0.00001355, -0.00001033, 0.00001631, -0.00000890, -0.00002892, -0.00002258, -0.00000689, 0.00001732, -0.00002362, -0.00003425, 0.00002535, -0.00001289, -0.00005544, -0.00004658, -0.00003246, 0.00004018, -0.00008694, -0.00020577, 0.00005967, 0.00000773, -0.00042620, -0.00083746, -0.00103758, 0.00101438, 0.00391633, 0.00174493, -0.00255052, -0.00132516, 0.00121749, -0.00023338, -0.00093894, 0.00044384, 0.00021819, -0.00044295, -0.00000622, 0.00011061, -0.00005669, -0.00002071, -0.00004634, 0.00002021, 0.00003737, -0.00007842, -0.00003010, 0.00002749, -0.00000705, -0.00001008, -0.00001971, 0.00000274, 0.00001934, -0.00002751, -0.00001877, 0.00000690, -0.00000721, 0.00000287, 0.00000258, -0.00000323, 0.00000695, -0.00001013, -0.00001514, -0.00000287, -0.00000551, 0.00000528, 0.00000478, -0.00000566, 0.00000630, -0.00000352, -0.00001206, 0.00000909, -0.00000058, -0.00000916, 0.00001375, 0.00000354, -0.00001274, 0.00000211, -0.00000412, -0.00001963, -0.00000687, 0.00000308, -0.00000171, -0.00000308, -0.00000286, -0.00000398, -0.00000407, -0.00000330, -0.00000180, -0.00000162, -0.00000216, -0.00000135, -0.00000044, -0.00000066, -0.00000184, -0.00000224, -0.00000132, },
		{ -0.00000226, -0.00000202, -0.00000161, -0.00000071, -0.00000220, -0.00000557, -0.00000588, -0.00000557, -0.00000457, 0.00000064, -0.00000381, -0.00001424, -0.00000381, 0.00000444, -0.00000875, -0.00000137, 0.00001112, -0.00000689, -0.00000631, 0.00000905, -0.00000464, -0.00000724, 0.00000338, -0.00000299, -0.00000017, -0.00000014, -0.00000881, 0.00000049, 0.00000028, -0.00000600, -0.00000090, -0.00000527, -0.00000087, -0.00000206, -0.00001520, 0.00000462, 0.00000369, -0.00002015, 0.00000128, 0.00000833, -0.00000878, -0.00002197, -0.00003778, 0.00000432, 0.00001772, -0.00003720, -0.00001348, 0.00000702, -0.00001177, -0.00002625, -0.00010010, -0.00001202, 0.00007207, -0.00015681, -0.00015343, 0.00001530, -0.00002730, -0.00023183, -0.00136893, -0.00133291, 0.00268883, 0.00456578, -0.00028845, -0.00310802, -0.00001065, 0.00101989, -0.00077921, -0.00040799, 0.00043632, -0.00008548, -0.00021678, 0.00000343, -0.00000725, 0.00003273, -0.00004240, -0.00006332, 0.00005870, -0.00000132, -0.00006257, -0.00001359, -0.00000198, 0.00000955, -0.00001094, -0.00002622, 0.00001835, 0.00000397, -0.00002439, -0.00000747, -0.00000836, 0.00000272, 0.00000830, -0.00001032, 0.00000284, 0.00000588, -0.00001244, -0.00001072, -0.00000949, 0.00000203, 0.00001065, -0.00000635, -0.00000043, 0.00000544, -0.00001103, 0.00000023, 0.00000465, -0.00000997, 0.00000713, 0.00001066, -0.00001156, -0.00000196, 0.00000431, -0.00001670, -0.00001517, 0.00000114, 0.00000061, -0.00000381, -0.00000362, -0.00000344, -0.00000309, -0.00000328, -0.00000302, -0.00000190, -0.00000165, -0.00000162, -0.00000107, -0.00000046, -0.00000075, -0.00000189, },
		{ -0.00000094, -0.00000184, -0.00000256, -0.00000260, -0.00000127, -0.00000087, -0.00000389, -0.00000581, -0.00000486, -0.00000450, -0.00000121, -0.00000182, -0.00001345, -0.00000969, 0.00000567, -0.00000484, -0.00000876, 0.00001218, 0.00000257, -0.00001278, 0.00000486, 0.00000323, -0.00000875, 0.00000046, -0.00000158, -0.00000015, 0.00000639, -0.00000995, -0.00001006, 0.00000276, -0.00000134, -0.00000176, -0.00000621, -0.00000189, 0.00000863, -0.00001531, -0.00001454, 0.00001067, -0.00000673, -0.00001029, 0.00000014, -0.00000183, 0.00000528, -0.00003188, -0.00004038, 0.00001583, 0.00000010, -0.00002004, -0.00001616, -0.00002052, 0.00003445, -0.00005008, -0.00013472, 0.00004396, 0.00002109, -0.00013014, -0.00013662, -0.00016421, 0.00024498, -0.00027835, -0.00233247, -0.00057323, 0.00464001, 0.00357349, -0.00223463, -0.00231699, 0.00079882, 0.00031446, -0.00072781, -0.00001935, 0.00016542, -0.00009505, -0.00007434, -0.00009125, 0.00001821, 0.00005073, -0.00008206, -0.00001782, 0.00004068, -0.00002019, -0.00002916, -0.00003176, -0.00000358, 0.00001868, -0.00002310, -0.00001033, 0.00001347, -0.00000749, -0.00000809, -0.00001531, -0.00000991, 0.00001513, -0.00000283, -0.00001038, 0.00000716, -0.00000010, -0.00000788, -0.00001354, -0.00000876, 0.00001205, 0.00000218, -0.00000746, 0.00000484, -0.00000443, -0.00000453, 0.00000582, -0.00000873, -0.00000123, 0.00001531, -0.00000562, -0.00000941, 0.00000753, -0.00000687, -0.00001931, -0.00000505, 0.00000178, -0.00000273, -0.00000440, -0.00000444, -0.00000315, -0.00000196, -0.00000279, -0.00000299, -0.00000202, -0.00000137, -0.00000118, -0.00000093, -0.00000059, },
		{ -0.00000087, -0.00000086, -0.00000111, -0.00000191, -0.00000292, -0.00000271, -0.00000123, -0.00000201, -0.00000479, -0.00000503, -0.00000421, -0.00000172, 0.00000021, -0.00001090, -0.00001613, 0.00000246, 0.00000249, -0.00001305, 0.00000493, 0.00001279, -0.00001023, -0.00000370, 0.00000645, -0.00000645, -0.00000123, -0.00000014, -0.00000524, 0.00000982, 0.00000129, -0.00001658, -0.00000636, -0.00000004, 0.00000262, -0.00000286, -0.00001063, 0.00001072, 0.00000316, -0.00002370, -0.00000490, 0.00000114, -0.00000382, -0.00000018, -0.00001656, 0.00000610, 0.00000920, -0.00004374, -0.00002607, 0.00000062, 0.00000315, 0.00000356, -0.00005891, 0.00000093, 0.00006110, -0.00010700, -0.00009436, 0.00001852, -0.00000308, 0.00000034, -0.00036112, -0.00016467, 0.00072277, -0.00104209, -0.00292679, 0.00139867, 0.00551820, 0.00139867, -0.00292679, -0.00104209, 0.00072277, -0.00016467, -0.00036112, 0.00000034, -0.00000308, 0.00001852, -0.00009436, -0.00010700, 0.00006110, 0.00000093, -0.00005891, 0.00000356, 0.00000315, 0.00000062, -0.00002607, -0.00004374, 0.00000920, 0.00000610, -0.00001656, -0.00000018, -0.00000382, 0.00000114, -0.00000490, -0.00002370, 0.00000316, 0.00001072, -0.00001063, -0.00000286, 0.00000262, -0.00000004, -0.00000636, -0.00001658, 0.00000129, 0.00000982, -0.00000524, -0.00000014, -0.00000123, -0.00000645, 0.00000645, -0.00000370, -0.00001023, 0.00001279, 0.00000493, -0.00001305, 0.00000249, 0.00000246, -0.00001613, -0.00001090, 0.00000021, -0.00000172, -0.00000421, -0.00000503, -0.00000479, -0.00000201, -0.00000123, -0.00000271, -0.00000292, -0.00000191, -0.00000111, -0.00000086, },
		{ -0.00000094, -0.00000059, -0.00000093, -0.00000118, -0.00000137, -0.00000202, -0.00000299, -0.00000279, -0.00000196, -0.00000315, -0.00000444, -0.00000440, -0.00000273, 0.00000178, -0.00000505, -0.00001931, -0.00000687, 0.00000753, -0.00000941, -0.00000562, 0.00001531, -0.00000123, -0.00000873, 0.00000582, -0.00000453, -0.00000443, 0.00000484, -0.00000746, 0.00000218, 0.00001205, -0.00000876, -0.00001354, -0.00000788, -0.00000010, 0.00000716, -0.00001038, -0.00000283, 0.00001513, -0.00000991, -0.00001531, -0.00000809, -0.00000749, 0.00001347, -0.00001033, -0.00002310, 0.00001868, -0.00000358, -0.00003176, -0.00002916, -0.00002019, 0.00004068, -0.00001782, -0.00008206, 0.00005073, 0.00001821, -0.00009125, -0.00007434, -0.00009505, 0.00016542, -0.00001935, -0.00072781, 0.00031446, 0.00079882, -0.00231699, -0.00223463, 0.00357349, 0.00464001, -0.00057323, -0.00233247, -0.00027835, 0.00024498, -0.00016421, -0.00013662, -0.00013014, 0.00002109, 0.00004396, -0.00013472, -0.00005008, 0.00003445, -0.00002052, -0.00001616, -0.00002004, 0.00000010, 0.00001583, -0.00004038, -0.00003188, 0.00000528, -0.00000183, 0.00000014, -0.00001029, -0.00000673, 0.00001067, -0.00001454, -0.00001531, 0.00000863, -0.00000189, -0.00000621, -0.00000176, -0.00000134, 0.00000276, -0.00001006, -0.00000995, 0.00000639, -0.00000015, -0.00000158, 0.00000046, -0.00000875, 0.00000323, 0.00000486, -0.00001278, 0.00000257, 0.00001218, -0.00000876, -0.00000484, 0.00000567, -0.00000969, -0.00001345, -0.00000182, -0.00000121, -0.00000450, -0.00000486, -0.00000581, -0.00000389, -0.00000087, -0.00000127, -0.00000260, -0.00000256, -0.00000184, },
		{ -0.00000226, -0.00000189, -0.00000075, -0.00000046, -0.00000107, -0.00000162, -0.00000165, -0.00000190, -0.00000302, -0.00000328, -0.00000309, -0.00000344, -0.00000362, -0.00000381, 0.00000061, 0.00000114, -0.00001517, -0.00001670, 0.00000431, -0.00000196, -0.00001156, 0.00001066, 0.00000713, -0.00000997, 0.00000465, 0.00000023, -0.00001103, 0.00000544, -0.00000043, -0.00000635, 0.00001065, 0.00000203, -0.00000949, -0.00001072, -0.00001244, 0.00000588, 0.00000284, -0.00001032, 0.00000830, 0.00000272, -0.00000836, -0.00000747, -0.00002439, 0.00000397, 0.00001835, -0.00002622, -0.00001094, 0.00000955, -0.00000198, -0.00001359, -0.00006257, -0.00000132, 0.00005870, -0.00006332, -0.00004240, 0.00003273, -0.00000725, 0.00000343, -0.00021678, -0.00008548, 0.00043632, -0.00040799, -0.00077921, 0.00101989, -0.00001065, -0.00310802, -0.00028845, 0.00456578, 0.00268883, -0.00133291, -0.00136893, -0.00023183, -0.00002730, 0.00001530, -0.00015343, -0.00015681, 0.00007207, -0.00001202, -0.00010010, -0.00002625, -0.00001177, 0.00000702, -0.00001348, -0.00003720, 0.00001772, 0.00000432, -0.00003778, -0.00002197, -0.00000878, 0.00000833, 0.00000128, -0.00002015, 0.00000369, 0.00000462, -0.00001520, -0.00000206, -0.00000087, -0.00000527, -0.00000090, -0.00000600, 0.00000028, 0.00000049, -0.00000881, -0.00000014, -0.00000017, -0.00000299, 0.00000338, -0.00000724, -0.00000464, 0.00000905, -0.00000631, -0.00000689, 0.00001112, -0.00000137, -0.00000875, 0.00000444, -0.00000381, -0.00001424, -0.00000381, 0.00000064, -0.00000457, -0.00000557, -0.00000588, -0.00000557, -0.00000220, -0.00000071, -0.00000161, -0.00000202, },
		{ -0.00000137, -0.00000132, -0.00000224, -0.00000184, -0.00000066, -0.00000044, -0.00000135, -0.00000216, -0.00000162, -0.00000180, -0.00000330, -0.00000407, -0.00000398, -0.00000286, -0.00000308, -0.00000171, 0.00000308, -0.00000687, -0.00001963, -0.00000412, 0.00000211, -0.00001274, 0.00000354, 0.00001375, -0.00000916, -0.00000058, 0.00000909, -0.00001206, -0.00000352, 0.00000630, -0.00000566, 0.00000478, 0.00000528, -0.00000551, -0.00000287, -0.00001514, -0.00001013, 0.00000695, -0.00000323, 0.00000258, 0.00000287, -0.00000721, 0.00000690, -0.00001877, -0.00002751, 0.00001934, 0.00000274, -0.00001971, -0.00001008, -0.00000705, 0.00002749, -0.00003010, -0.00007842, 0.00003737, 0.00002021, -0.00004634, -0.00002071, -0.00005669, 0.00011061, -0.00000622, -0.00044295, 0.00021819, 0.00044384, -0.00093894, -0.00023338, 0.00121749, -0.00132516, -0.00255052, 0.00174493, 0.00391633, 0.00101438, -0.00103758, -0.00083746, -0.00042620, 0.00000773, 0.00005967, -0.00020577, -0.00008694, 0.00004018, -0.00003246, -0.00004658, -0.00005544, -0.00001289, 0.00002535, -0.00003425, -0.00002362, 0.00001732, -0.00000689, -0.00002258, -0.00002892, -0.00000890, 0.00001631, -0.00001033, -0.00001355, 0.00000575, -0.00000431, -0.00000252, -0.00000212, -0.00001011, -0.00000040, -0.00000277, -0.00000664, 0.00000071, -0.00000256, 0.00000024, 0.00000027, -0.00000727, 0.00000187, -0.00000084, -0.00000945, 0.00000494, 0.00000166, -0.00000939, 0.00000571, 0.00000408, -0.00000955, 0.00000116, 0.00000157, -0.00001351, -0.00000785, 0.00000261, -0.00000193, -0.00000635, -0.00000649, -0.00000659, -0.00000380, -0.00000118, -0.00000148, },
		{ -0.00000166, -0.00000193, -0.00000056, -0.00000109, -0.00000218, -0.00000173, -0.00000073, -0.00000052, -0.00000186, -0.00000234, -0.00000147, -0.00000209, -0.00000364, -0.00000496, -0.00000415, -0.00000219, -0.00000200, 0.00000182, -0.00000081, -0.00001617, -0.00001079, 0.00000279, -0.00001201, -0.00000628, 0.00001801, -0.00000201, -0.00001057, 0.00001236, -0.00000210, -0.00001183, 0.00000391, -0.00000226, 0.00000264, 0.00000836, -0.00000789, -0.00000084, -0.00000229, -0.00001741, -0.00000499, -0.00000172, 0.00000482, 0.00000930, -0.00001506, 0.00000322, 0.00000949, -0.00003265, -0.00001116, 0.00001026, -0.00000214, -0.00000306, -0.00003585, 0.00000708, 0.00004235, -0.00006399, -0.00004460, 0.00001858, -0.00000021, 0.00003322, -0.00012868, -0.00005322, 0.00030107, -0.00026324, -0.00044960, 0.00061751, -0.00000504, -0.00106720, 0.00051874, 0.00050498, -0.00208014, -0.00088836, 0.00262164, 0.00244202, 0.00035909, -0.00052612, -0.00077521, -0.00044829, 0.00008123, -0.00001656, -0.00016173, -0.00005503, -0.00001301, 0.00000319, -0.00004344, -0.00007274, 0.00001050, 0.00000839, -0.00003443, -0.00000915, 0.00000175, 0.00000004, -0.00001916, -0.00003441, 0.00000150, 0.00000755, -0.00001246, -0.00000372, -0.00000372, 0.00000028, 0.00000573, -0.00001154, -0.00000885, 0.00000101, -0.00000588, -0.00000451, -0.00000277, 0.00000177, 0.00000620, -0.00000667, -0.00000554, 0.00000221, -0.00000735, -0.00000122, 0.00000495, -0.00000753, 0.00000037, 0.00000765, -0.00000782, -0.00000427, 0.00000512, -0.00000913, -0.00001282, 0.00000101, 0.00000233, -0.00000465, -0.00000701, -0.00000778, -0.00000563, -0.00000176, },
		{ -0.00000322, -0.00000129, -0.00000266, -0.00000134, -0.00000013, -0.00000112, -0.00000186, -0.00000179, -0.00000088, -0.00000095, -0.00000219, -0.00000206, -0.00000187, -0.00000243, -0.00000393, -0.00000560, -0.00000378, -0.00000168, 0.00000107, 0.00000224, -0.00001117, -0.00001483, 0.00000265, -0.00000701, -0.00001758, 0.00001309, 0.00001165, -0.00001503, 0.00000389, 0.00000876, -0.00001035, -0.00000087, -0.00000239, -0.00000217, 0.00001465, -0.00000356, -0.00001087, 0.00000478, -0.00000690, -0.00001076, -0.00001055, -0.00000433, 0.00002089, -0.00000470, -0.00001854, 0.00001656, -0.00000595, -0.00002216, -0.00000756, -0.00001022, 0.00002096, -0.00001408, -0.00005006, 0.00003791, 0.00001319, -0.00004646, -0.00002893, -0.00005512, 0.00009586, 0.00002656, -0.00029956, 0.00016649, 0.00030238, -0.00059472, -0.00009255, 0.00068124, -0.00057474, -0.00058784, 0.00072494, -0.00058947, -0.00159759, 0.00069357, 0.00216000, 0.00129684, 0.00042565, -0.00033441, -0.00078305, -0.00032713, 0.00003318, -0.00004996, -0.00009126, -0.00008557, -0.00001220, 0.00002393, -0.00006579, -0.00005140, 0.00000938, -0.00000950, -0.00001650, -0.00001493, -0.00000244, 0.00000796, -0.00002584, -0.00002523, 0.00000117, -0.00000341, -0.00000255, -0.00000623, -0.00000724, 0.00001114, 0.00000034, -0.00001503, -0.00000444, -0.00000261, -0.00000506, -0.00000563, -0.00000287, 0.00000941, 0.00000251, -0.00000891, -0.00000172, -0.00000462, -0.00000459, 0.00000528, -0.00000438, -0.00000497, 0.00000876, -0.00000222, -0.00000939, 0.00000369, -0.00000243, -0.00001451, -0.00000424, 0.00000433, -0.00000096, -0.00000564, -0.00000810, -0.00000799, },
		{ -0.00000912, -0.00000599, -0.00000156, -0.00000242, -0.00000251, -0.00000056, -0.00000042, -0.00000085, -0.00000158, -0.00000194, -0.00000131, -0.00000154, -0.00000177, -0.00000218, -0.00000269, -0.00000238, -0.00000433, -0.00000564, -0.00000355, 0.00000004, 0.00000438, -0.00000517, -0.00001770, -0.00000116, 0.00000285, -0.00002121, -0.00000270, 0.00002015, -0.00000727, -0.00000717, 0.00001082, -0.00000494, -0.00000203, 0.00000088, -0.00001096, 0.00001136, 0.00001127, -0.00001447, -0.00000388, 0.00000061, -0.00000315, -0.00000773, -0.00002200, 0.00000995, 0.00002015, -0.00001983, -0.00000345, 0.00000664, -0.00000781, -0.00000202, -0.00003346, -0.00000301, 0.00003589, -0.00004229, -0.00002423, 0.00002332, 0.00000060, 0.00001881, -0.00011810, -0.00004931, 0.00024443, -0.00017247, -0.00030139, 0.00044748, -0.00000197, -0.00064501, 0.00034019, 0.00028324, -0.00072053, 0.00001467, 0.00014764, -0.00107467, -0.00031861, 0.00123917, 0.00120537, 0.00090639, 0.00054777, -0.00036759, -0.00065253, -0.00026661, -0.00004904, -0.00000974, -0.00008058, -0.00010099, 0.00001447, 0.00000274, -0.00006035, -0.00003081, -0.00001237, -0.00000385, -0.00001037, -0.00002356, 0.00000531, 0.00000272, -0.00002322, -0.00001583, -0.00001068, -0.00000116, 0.00000200, -0.00001300, 0.00000158, 0.00001187, -0.00000760, -0.00001044, -0.00000504, -0.00000453, -0.00000440, -0.00000768, 0.00000343, 0.00001013, -0.00000283, -0.00000468, -0.00000494, -0.00000768, 0.00000387, 0.00000071, -0.00000874, 0.00000479, 0.00000476, -0.00000953, -0.00000165, 0.00000196, -0.00001192, -0.00000952, 0.00000287, 0.00000204, -0.00000315, -0.00000633, },
		{ -0.00000400, -0.00000767, -0.00000810, -0.00000326, -0.00000223, -0.00000311, -0.00000131, -0.00000070, -0.00000066, -0.00000035, -0.00000147, -0.00000221, -0.00000223, -0.00000145, -0.00000122, -0.00000317, -0.00000301, -0.00000232, -0.00000454, -0.00000537, -0.00000314, 0.00000376, 0.00000206, -0.00001599, -0.00001018, 0.00000899, -0.00001208, -0.00001747, 0.00001513, 0.00000423, -0.00001163, 0.00000813, -0.00000116, -0.00000588, 0.00000861, -0.00001003, -0.00000402, 0.00001839, -0.00000315, -0.00001040, -0.00000391, -0.00000281, 0.00000713, -0.00001814, -0.00001859, 0.00002139, 0.00000177, -0.00001067, -0.00000136, -0.00001135, 0.00001869, -0.00001132, -0.00005201, 0.00002373, 0.00001322, -0.00003026, -0.00001229, -0.00003390, 0.00008162, 0.00001087, -0.00025159, 0.00012709, 0.00023853, -0.00042112, -0.00003766, 0.00048448, -0.00036236, -0.00032805, 0.00042348, -0.00017699, -0.00036387, 0.00008850, -0.00059201, -0.00055632, 0.00068882, 0.00084262, 0.00061283, 0.00090070, 0.00045869, -0.00034231, -0.00049481, -0.00030056, -0.00006119, 0.00001565, -0.00009974, -0.00007265, 0.00000999, -0.00001614, -0.00003341, -0.00003760, -0.00002041, 0.00000632, -0.00001842, -0.00001852, 0.00000588, -0.00000314, -0.00001134, -0.00002006, -0.00001606, 0.00000578, -0.00000402, -0.00000918, 0.00000893, 0.00000290, -0.00000742, -0.00000697, -0.00000773, -0.00000392, -0.00000652, -0.00000433, 0.00000811, 0.00000441, -0.00000166, -0.00000344, -0.00001060, -0.00000170, 0.00000532, -0.00000714, -0.00000222, 0.00000751, -0.00000478, -0.00000597, 0.00000250, -0.00000767, -0.00001281, -0.00000057, 0.00000361, -0.00000127, },
		{ 0.00000046, -0.00000289, -0.00000533, -0.00000801, -0.00000487, -0.00000257, -0.00000385, -0.00000208, -0.00000067, -0.00000139, -0.00000043, -0.00000004, -0.00000137, -0.00000301, -0.00000288, -0.00000059, -0.00000169, -0.00000406, -0.00000277, -0.00000258, -0.00000417, -0.00000531, -0.00000085, 0.00000562, -0.00000798, -0.00001726, 0.00000479, 0.00000088, -0.00002052, 0.00000361, 0.00001060, -0.00001257, 0.00000427, 0.00000612, -0.00001285, 0.00000735, 0.00000353, -0.00001289, 0.00000985, 0.00000648, -0.00000649, -0.00000350, -0.00001267, 0.00000589, 0.00000593, -0.00002519, -0.00000201, 0.00000788, -0.00000203, 0.00000673, -0.00002586, -0.00000005, 0.00003588, -0.00004128, -0.00003232, 0.00001457, 0.00000120, 0.00002123, -0.00008439, -0.00002507, 0.00020036, -0.00014915, -0.00024467, 0.00034195, -0.00000102, -0.00045154, 0.00026902, 0.00021291, -0.00043096, 0.00002019, 0.00012960, -0.00024750, -0.00001903, -0.00039055, -0.00070003, 0.00038074, 0.00083368, 0.00027912, 0.00051180, 0.00082797, 0.00032563, -0.00021627, -0.00043436, -0.00032020, -0.00003618, -0.00000753, -0.00008703, -0.00004871, -0.00001675, -0.00000502, -0.00002347, -0.00004949, -0.00001131, 0.00000157, -0.00002019, -0.00001088, -0.00000210, 0.00000265, -0.00000734, -0.00002802, -0.00001026, 0.00000331, -0.00000691, 0.00000157, 0.00000472, -0.00000295, -0.00000300, -0.00000794, -0.00000785, -0.00000498, -0.00000647, 0.00000198, 0.00000535, 0.00000157, 0.00000155, -0.00000811, -0.00000891, 0.00000411, -0.00000183, -0.00000643, 0.00000508, 0.00000028, -0.00000717, 0.00000105, -0.00000329, -0.00001401, -0.00000538, 0.00000408, },
		{ 0.00000223, 0.00000274, -0.00000211, -0.00000399, -0.00000693, -0.00000551, -0.00000228, -0.00000429, -0.00000390, -0.00000063, -0.00000118, -0.00000159, -0.00000017, 0.00000016, -0.00000161, -0.00000410, -0.00000243, -0.00000035, -0.00000286, -0.00000389, -0.00000289, -0.00000258, -0.00000369, -0.00000444, 0.00000263, -0.00000031, -0.00001633, -0.00000415, 0.00000737, -0.00001537, -0.00000565, 0.00001451, -0.00001164, -0.00000642, 0.00001523, -0.00001000, -0.00000576, 0.00001194, -0.00000676, 0.00000144, 0.00000690, -0.00000673, 0.00000401, -0.00000919, -0.00001263, 0.00001333, -0.00000578, -0.00001085, -0.00000261, -0.00001219, 0.00002097, -0.00000388, -0.00003999, 0.00003002, 0.00001365, -0.00003622, -0.00001937, -0.00003266, 0.00006557, 0.00001374, -0.00019032, 0.00011956, 0.00019152, -0.00033991, -0.00003230, 0.00035907, -0.00027170, -0.00021517, 0.00032215, -0.00007876, -0.00020181, 0.00005448, -0.00013699, 0.00000303, -0.00014408, -0.00078362, -0.00009609, 0.00088837, 0.00042650, 0.00007495, 0.00053304, 0.00063597, 0.00030834, -0.00012349, -0.00042702, -0.00028298, -0.00004469, -0.00002750, -0.00005264, -0.00005744, -0.00002384, 0.00001001, -0.00003342, -0.00004383, -0.00000801, -0.00000742, -0.00001221, -0.00001335, -0.00000340, 0.00001008, -0.00001425, -0.00002567, -0.00000686, -0.00000401, 0.00000078, 0.00000495, -0.00000364, -0.00000130, -0.00000186, -0.00000885, -0.00000736, -0.00000666, -0.00000162, 0.00000340, 0.00000065, 0.00000374, -0.00000059, -0.00001028, -0.00000171, 0.00000112, -0.00000660, 0.00000112, 0.00000333, -0.00000637, -0.00000153, 0.00000069, -0.00001216, -0.00001086, },
		{ -0.00001423, -0.00000250, 0.00000401, -0.00000053, -0.00000317, -0.00000598, -0.00000626, -0.00000163, -0.00000277, -0.00000567, -0.00000230, -0.00000050, -0.00000171, -0.00000135, -0.00000015, 0.00000026, -0.00000251, -0.00000427, -0.00000173, -0.00000153, -0.00000307, -0.00000343, -0.00000345, -0.00000205, -0.00000328, -0.00000128, 0.00000151, -0.00001167, -0.00001042, 0.00000821, -0.00000844, -0.00001405, 0.00001683, -0.00000128, -0.00002101, 0.00001186, 0.00000440, -0.00001394, 0.00000664, 0.00000017, 0.00000040, 0.00001133, -0.00001159, -0.00000201, 0.00000756, -0.00001657, -0.00000200, 0.00000308, -0.00000178, 0.00000586, -0.00002906, -0.00000339, 0.00003406, -0.00003137, -0.00001614, 0.00002382, -0.00000311, 0.00000806, -0.00007710, -0.00002741, 0.00016186, -0.00011659, -0.00018124, 0.00028955, 0.00000035, -0.00035250, 0.00020109, 0.00014369, -0.00031601, 0.00003263, 0.00012543, -0.00011165, -0.00000687, -0.00014596, -0.00008461, 0.00010124, -0.00055550, -0.00057049, 0.00057749, 0.00073599, 0.00009517, 0.00014185, 0.00043241, 0.00049894, 0.00034724, -0.00011183, -0.00038510, -0.00024691, -0.00007733, -0.00001419, -0.00004103, -0.00006924, -0.00000995, 0.00000414, -0.00003664, -0.00003202, -0.00001595, -0.00000620, -0.00000901, -0.00001760, 0.00000392, 0.00000649, -0.00001772, -0.00001791, -0.00001224, -0.00000322, 0.00000879, -0.00000122, -0.00000586, 0.00000121, -0.00000345, -0.00000794, -0.00000836, -0.00000452, 0.00000242, -0.00000037, 0.00000052, 0.00000395, -0.00000502, -0.00000500, 0.00000073, -0.00000582, -0.00000287, 0.00000463, -0.00000368, -0.00000480, 0.00000223, -0.00000722, },
		{ -0.00000224, -0.00001362, -0.00000786, 0.00000293, 0.00000111, -0.00000229, -0.00000466, -0.00000717, -0.00000257, -0.00000010, -0.00000487, -0.00000469, -0.00000134, -0.00000128, -0.00000147, -0.00000128, -0.00000017, -0.00000031, -0.00000315, -0.00000330, -0.00000250, -0.00000263, -0.00000207, -0.00000379, -0.00000355, -0.00000167, -0.00000124, 0.00000132, -0.00000826, -0.00001513, 0.00000604, 0.00000146, -0.00002107, 0.00000909, 0.00001682, -0.00002263, -0.00000600, 0.00001295, -0.00000949, 0.00000014, 0.00000133, -0.00000511, 0.00001906, -0.00000136, -0.00001923, 0.00000847, -0.00000144, -0.00000813, -0.00000382, -0.00000841, 0.00002352, -0.00000708, -0.00004400, 0.00002217, 0.00001126, -0.00002306, 0.00000048, -0.00002191, 0.00004898, 0.00000569, -0.00016634, 0.00009071, 0.00015418, -0.00026992, -0.00000914, 0.00029872, -0.00020609, -0.00016223, 0.00023117, -0.00007448, -0.00014565, 0.00005649, -0.00003991, 0.00003698, -0.00006021, -0.00023049, 0.00009055, -0.00017307, -0.00068042, 0.00005784, 0.00074830, 0.00035000, 0.00007236, 0.00017531, 0.00028723, 0.00046909, 0.00033456, -0.00010788, -0.00031610, -0.00024714, -0.00008656, 0.00000044, -0.00005178, -0.00005811, -0.00000522, -0.00000894, -0.00002618, -0.00003099, -0.00002001, -0.00000151, -0.00001351, -0.00001271, 0.00000615, -0.00000179, -0.00001175, -0.00001703, -0.00001610, 0.00000470, 0.00000701, -0.00000666, -0.00000299, 0.00000028, -0.00000401, -0.00000764, -0.00000785, 0.00000002, 0.00000104, -0.00000256, 0.00000206, -0.00000041, -0.00000385, 0.00000102, -0.00000416, -0.00000709, 0.00000283, 0.00000050, -0.00000617, 0.00000035, },
		{ -0.00000230, 0.00000023, -0.00001051, -0.00001169, 0.00000024, 0.00000228, -0.00000186, -0.00000333, -0.00000680, -0.00000493, 0.00000084, -0.00000192, -0.00000510, -0.00000286, -0.00000206, -0.00000149, -0.00000102, -0.00000134, -0.00000058, -0.00000138, -0.00000231, -0.00000303, -0.00000442, -0.00000205, -0.00000146, -0.00000437, -0.00000330, -0.00000092, 0.00000285, -0.00000337, -0.00001881, -0.00000262, 0.00001206, -0.00001759, -0.00000863, 0.00002548, -0.00000609, -0.00001986, 0.00000761, -0.00000453, -0.00000252, 0.00000707, -0.00001348, 0.00001085, 0.00002063, -0.00001876, -0.00000888, 0.00000317, -0.00000099, 0.00000312, -0.00002373, 0.00000638, 0.00003437, -0.00003499, -0.00002311, 0.00001360, -0.00000327, 0.00002204, -0.00005113, -0.00002306, 0.00013381, -0.00010230, -0.00016086, 0.00023107, -0.00000177, -0.00027831, 0.00017495, 0.00012832, -0.00023114, 0.00002456, 0.00007654, -0.00009396, -0.00000707, -0.00007138, -0.00000516, 0.00008805, -0.00019740, -0.00009111, 0.00002047, -0.00048865, -0.00029059, 0.00050188, 0.00047379, 0.00016593, 0.00015713, 0.00009799, 0.00022789, 0.00044945, 0.00028147, -0.00006627, -0.00027339, -0.00025006, -0.00007329, -0.00001014, -0.00005195, -0.00004064, -0.00001664, -0.00000933, -0.00001887, -0.00003487, -0.00001548, -0.00000427, -0.00001382, -0.00000536, -0.00000110, -0.00000233, -0.00000655, -0.00002114, -0.00001076, 0.00000770, -0.00000026, -0.00000565, -0.00000146, -0.00000164, -0.00000382, -0.00000761, -0.00000400, 0.00000135, -0.00000210, -0.00000128, -0.00000030, -0.00000298, 0.00000216, 0.00000037, -0.00000855, -0.00000251, 0.00000263, -0.00000425, },
		{ -0.00000192, -0.00000341, 0.00000090, -0.00000666, -0.00001379, -0.00000367, 0.00000330, -0.00000116, -0.00000319, -0.00000551, -0.00000660, -0.00000047, 0.00000046, -0.00000372, -0.00000295, -0.00000269, -0.00000305, -0.00000097, -0.00000080, -0.00000148, -0.00000187, -0.00000141, -0.00000092, -0.00000468, -0.00000514, -0.00000075, -0.00000186, -0.00000426, -0.00000335, 0.00000215, 0.00000371, -0.00001528, -0.00001577, 0.00001252, -0.00000290, -0.00002077, 0.00001654, 0.00001223, -0.00001895, 0.00000043, -0.00000330, -0.00001143, 0.00001408, -0.00000637, -0.00001001, 0.00002562, 0.00000183, -0.00001507, -0.00000554, -0.00000777, 0.00001693, -0.00000803, -0.00003010, 0.00003164, 0.00001033, -0.00002822, -0.00000990, -0.00002895, 0.00004744, 0.00002199, -0.00013459, 0.00007788, 0.00013127, -0.00023266, -0.00001830, 0.00023961, -0.00016910, -0.00012582, 0.00019636, -0.00003678, -0.00009677, 0.00003521, -0.00004406, 0.00001520, -0.00004006, -0.00011876, 0.00008979, -0.00002784, -0.00016818, 0.00000541, -0.00028851, -0.00041373, 0.00024831, 0.00047056, 0.00018598, 0.00019938, 0.00014913, 0.00003303, 0.00023069, 0.00038325, 0.00025432, -0.00002297, -0.00025751, -0.00022756, -0.00007226, -0.00002672, -0.00003542, -0.00003883, -0.00002446, -0.00000293, -0.00002129, -0.00003153, -0.00001331, -0.00000921, -0.00000684, -0.00000675, -0.00000674, 0.00000269, -0.00000877, -0.00001935, -0.00000347, 0.00000186, -0.00000358, -0.00000239, -0.00000257, -0.00000247, -0.00000414, -0.00000518, -0.00000058, -0.00000096, -0.00000232, -0.00000150, -0.00000433, 0.00000040, 0.00000508, -0.00000434, -0.00000691, 0.00000041, },
		{ -0.00000286, -0.00000170, -0.00000406, 0.00000050, -0.00000232, -0.00001343, -0.00000889, 0.00000284, 0.00000107, -0.00000334, -0.00000491, -0.00000725, -0.00000250, 0.00000158, -0.00000239, -0.00000254, -0.00000128, -0.00000393, -0.00000307, -0.00000055, -0.00000060, -0.00000237, -0.00000304, 0.00000009, -0.00000126, -0.00000645, -0.00000401, -0.00000064, -0.00000199, -0.00000410, -0.00000229, 0.00000589, -0.00000435, -0.00002154, 0.00000062, 0.00000879, -0.00001878, 0.00000194, 0.00002027, -0.00001230, -0.00000338, 0.00000486, -0.00002195, 0.00000384, 0.00001200, -0.00001616, 0.00001004, 0.00001404, -0.00000642, -0.00000123, -0.00002113, 0.00000129, 0.00002428, -0.00002784, -0.00000628, 0.00002136, -0.00000457, 0.00001172, -0.00005821, -0.00002724, 0.00012607, -0.00007673, -0.00013377, 0.00020189, -0.00000008, -0.00024161, 0.00013592, 0.00010107, -0.00018997, 0.00002137, 0.00007696, -0.00004696, 0.00000349, -0.00006198, -0.00001428, 0.00005223, -0.00012297, -0.00003580, 0.00005384, -0.00010869, -0.00002527, -0.00015880, -0.00045122, 0.00002361, 0.00046735, 0.00020007, 0.00012099, 0.00023364, 0.00008014, 0.00004384, 0.00020872, 0.00031423, 0.00025697, -0.00001016, -0.00023151, -0.00020017, -0.00008847, -0.00002612, -0.00002482, -0.00004522, -0.00002039, -0.00000383, -0.00002245, -0.00002477, -0.00001722, -0.00000689, -0.00000325, -0.00001327, -0.00000430, 0.00000301, -0.00001103, -0.00001079, -0.00000390, -0.00000505, -0.00000174, -0.00000147, -0.00000376, -0.00000298, -0.00000343, -0.00000138, -0.00000070, -0.00000234, -0.00000166, -0.00000499, -0.00000401, 0.00000495, 0.00000211, -0.00000594, },
		{ -0.00000165, -0.00000327, -0.00000184, -0.00000480, -0.00000183, 0.00000084, -0.00000957, -0.00001314, -0.00000088, 0.00000331, -0.00000193, -0.00000452, -0.00000772, -0.00000512, 0.00000195, -0.00000067, -0.00000331, 0.00000009, -0.00000178, -0.00000463, -0.00000268, -0.00000032, -0.00000073, -0.00000394, -0.00000251, 0.00000096, -0.00000323, -0.00000602, -0.00000316, -0.00000131, -0.00000141, -0.00000400, 0.00000123, 0.00000334, -0.00001556, -0.00001098, 0.00000931, -0.00001146, -0.00000880, 0.00002267, -0.00000436, -0.00001287, 0.00001589, -0.00001402, -0.00001864, 0.00001469, -0.00000450, -0.00000046, 0.00001293, -0.00000718, 0.00000979, -0.00000674, -0.00002925, 0.00001997, 0.00000518, -0.00001535, 0.00000523, -0.00001988, 0.00004063, 0.00000968, -0.00013121, 0.00006704, 0.00012175, -0.00019620, -0.00000939, 0.00021222, -0.00014347, -0.00011388, 0.00015636, -0.00003379, -0.00008424, 0.00003070, -0.00001191, 0.00003390, -0.00002677, -0.00009192, 0.00005552, -0.00002121, -0.00010075, 0.00001455, -0.00006221, -0.00002949, -0.00003068, -0.00040724, -0.00022447, 0.00040004, 0.00031748, 0.00004013, 0.00016848, 0.00017935, 0.00005831, 0.00006869, 0.00015543, 0.00028239, 0.00024820, -0.00000664, -0.00019012, -0.00019152, -0.00009713, -0.00001794, -0.00002877, -0.00004308, -0.00001649, -0.00000934, -0.00001721, -0.00002368, -0.00001821, -0.00000154, -0.00000803, -0.00001429, -0.00000106, -0.00000136, -0.00000589, -0.00000525, -0.00001029, -0.00000664, 0.00000024, -0.00000237, -0.00000413, -0.00000309, -0.00000088, 0.00000025, -0.00000192, -0.00000172, -0.00000387, -0.00000695, 0.00000030, 0.00000479, },
		{ 0.00000281, 0.00000107, -0.00000194, -0.00000066, -0.00000360, -0.00000468, 0.00000029, -0.00000455, -0.00001345, -0.00000611, 0.00000303, 0.00000012, -0.00000319, -0.00000703, -0.00000832, 0.00000019, 0.00000209, -0.00000349, -0.00000108, 0.00000089, -0.00000290, -0.00000430, -0.00000258, -0.00000015, -0.00000184, -0.00000445, -0.00000112, -0.00000057, -0.00000389, -0.00000428, -0.00000395, -0.00000125, -0.00000107, -0.00000189, 0.00000292, -0.00000775, -0.00001455, 0.00000588, -0.00000492, -0.00001891, 0.00001972, 0.00001197, -0.00002365, 0.00000987, 0.00000902, -0.00002512, -0.00000072, 0.00000086, -0.00000217, 0.00001892, -0.00001161, -0.00000392, 0.00002147, -0.00002440, -0.00001223, 0.00001047, -0.00000203, 0.00002293, -0.00004157, -0.00001917, 0.00010933, -0.00007879, -0.00012470, 0.00017730, 0.00000017, -0.00020568, 0.00012686, 0.00009378, -0.00016267, 0.00001161, 0.00006020, -0.00004471, -0.00000510, -0.00004028, 0.00001163, 0.00005399, -0.00008684, -0.00002706, 0.00003381, -0.00006792, -0.00001939, -0.00005610, -0.00007614, 0.00005328, -0.00023821, -0.00039750, 0.00018190, 0.00042846, 0.00010077, 0.00005670, 0.00016134, 0.00011834, 0.00008515, 0.00005510, 0.00012020, 0.00026503, 0.00022057, 0.00001269, -0.00015912, -0.00019010, -0.00009014, -0.00002078, -0.00003285, -0.00003465, -0.00001960, -0.00000970, -0.00001350, -0.00002513, -0.00001272, -0.00000280, -0.00001285, -0.00000997, -0.00000350, -0.00000160, 0.00000184, -0.00000775, -0.00001446, -0.00000484, -0.00000021, -0.00000308, -0.00000394, -0.00000173, 0.00000106, -0.00000026, -0.00000149, -0.00000256, -0.00000695, -0.00000432, },
		{ -0.00000654, -0.00000054, 0.00000113, -0.00000159, -0.00000017, -0.00000047, -0.00000439, -0.00000244, -0.00000231, -0.00001094, -0.00000965, 0.00000089, 0.00000115, -0.00000218, -0.00000468, -0.00000959, -0.00000421, 0.00000330, -0.00000125, -0.00000285, 0.00000101, -0.00000031, -0.00000286, -0.00000393, -0.00000258, -0.00000076, -0.00000297, -0.00000307, -0.00000164, -0.00000261, -0.00000197, -0.00000389, -0.00000492, -0.00000079, -0.00000013, 0.00000187, -0.00000410, -0.00001585, 0.00000231, 0.00000533, -0.00002516, 0.00000285, 0.00002987, -0.00001655, -0.00001154, 0.00001947, -0.00001098, -0.00000959, -0.00000218, -0.00001380, 0.00002404, 0.00000803, -0.00002847, 0.00001625, 0.00000713, -0.00001685, -0.00000439, -0.00002195, 0.00004351, 0.00001847, -0.00010898, 0.00006428, 0.00010293, -0.00018009, -0.00000982, 0.00018082, -0.00012968, -0.00009249, 0.00014568, -0.00002310, -0.00007285, 0.00002333, -0.00001342, 0.00001728, -0.00002506, -0.00005912, 0.00005793, -0.00000524, -0.00006517, 0.00001189, -0.00003762, -0.00001956, -0.00001982, -0.00012665, 0.00001444, -0.00004995, -0.00038324, -0.00008139, 0.00039095, 0.00023034, 0.00003852, 0.00009687, 0.00009901, 0.00011356, 0.00009523, 0.00002924, 0.00011340, 0.00023279, 0.00020088, 0.00003591, -0.00014570, -0.00017633, -0.00008425, -0.00003134, -0.00002843, -0.00003118, -0.00002217, -0.00000670, -0.00001511, -0.00002122, -0.00000899, -0.00000909, -0.00001124, -0.00000857, -0.00000662, 0.00000436, 0.00000343, -0.00001285, -0.00001354, -0.00000409, -0.00000182, -0.00000298, -0.00000270, -0.00000002, 0.00000097, -0.00000011, -0.00000123, -0.00000549, },
		{ -0.00000326, -0.00000673, -0.00000317, 0.00000024, -0.00000190, -0.00000188, 0.00000088, -0.00000100, -0.00000287, -0.00000243, -0.00000885, -0.00001169, -0.00000149, 0.00000220, -0.00000228, -0.00000323, -0.00000773, -0.00000801, 0.00000110, 0.00000120, -0.00000283, 0.00000001, 0.00000044, -0.00000113, -0.00000205, -0.00000382, -0.00000293, -0.00000225, -0.00000220, -0.00000207, -0.00000437, -0.00000209, 0.00000007, -0.00000497, -0.00000493, -0.00000064, 0.00000340, 0.00000072, -0.00001660, -0.00000728, 0.00001532, -0.00001561, -0.00002060, 0.00002865, 0.00000702, -0.00002284, 0.00001018, 0.00000070, -0.00000825, 0.00000592, -0.00002514, 0.00000147, 0.00003561, -0.00001983, -0.00001383, 0.00001259, -0.00000123, 0.00001360, -0.00004431, -0.00001479, 0.00010482, -0.00006421, -0.00010335, 0.00015937, -0.00000386, -0.00018054, 0.00010843, 0.00006994, -0.00014574, 0.00001836, 0.00006155, -0.00003357, -0.00000291, -0.00003320, 0.00000210, 0.00003429, -0.00006553, -0.00001190, 0.00003338, -0.00003741, -0.00000284, -0.00003088, -0.00004315, 0.00002498, -0.00009599, -0.00008117, 0.00002011, -0.00025047, -0.00023740, 0.00024263, 0.00030161, 0.00008278, 0.00008089, 0.00006989, 0.00006553, 0.00012776, 0.00007403, 0.00002682, 0.00010450, 0.00019659, 0.00019672, 0.00004338, -0.00013133, -0.00015512, -0.00008869, -0.00003605, -0.00002353, -0.00003207, -0.00001940, -0.00000767, -0.00001527, -0.00001437, -0.00001218, -0.00001161, -0.00000804, -0.00001128, -0.00000371, 0.00000883, -0.00000133, -0.00001403, -0.00001100, -0.00000547, -0.00000295, -0.00000183, -0.00000085, 0.00000029, 0.00000076, 0.00000053, },
		{ 0.00000134, -0.00000080, -0.00000528, -0.00000451, -0.00000082, -0.00000160, -0.00000376, -0.00000131, 0.00000106, -0.00000036, -0.00000137, -0.00000689, -0.00001356, -0.00000540, 0.00000373, -0.00000094, -0.00000395, -0.00000574, -0.00000890, -0.00000217, 0.00000181, -0.00000233, -0.00000046, 0.00000106, -0.00000129, -0.00000060, -0.00000146, -0.00000379, -0.00000438, -0.00000280, -0.00000014, -0.00000379, -0.00000599, 0.00000043, 0.00000004, -0.00000545, -0.00000498, 0.00000096, 0.00000658, -0.00000962, -0.00001929, 0.00001129, 0.00000493, -0.00002829, 0.00000799, 0.00002282, -0.00001635, 0.00000034, 0.00000455, -0.00001484, 0.00001619, -0.00000772, -0.00003116, 0.00002775, 0.00001213, -0.00001917, -0.00000168, -0.00001579, 0.00003547, 0.00000885, -0.00009853, 0.00006312, 0.00009509, -0.00015555, -0.00000277, 0.00015979, -0.00011169, -0.00007688, 0.00011762, -0.00003453, -0.00006451, 0.00002779, -0.00000392, 0.00002045, -0.00001550, -0.00005053, 0.00003749, -0.00000648, -0.00004650, 0.00000961, -0.00002198, 0.00000141, 0.00000078, -0.00007108, 0.00000877, -0.00001906, -0.00011358, -0.00001020, -0.00013697, -0.00028096, 0.00008771, 0.00031234, 0.00012123, 0.00007642, 0.00009299, 0.00002609, 0.00007544, 0.00011462, 0.00005822, 0.00003644, 0.00008274, 0.00017660, 0.00018909, 0.00004159, -0.00010704, -0.00014176, -0.00009363, -0.00003472, -0.00002436, -0.00002972, -0.00001703, -0.00001095, -0.00001028, -0.00001269, -0.00001636, -0.00000916, -0.00000903, -0.00001122, 0.00000238, 0.00000632, -0.00000570, -0.00001128, -0.00001022, -0.00000744, -0.00000275, 0.00000004, -0.00000001, 0.00000012, },
		{ 0.00000009, 0.00000100, 0.00000050, -0.00000339, -0.00000431, -0.00000155, -0.00000097, -0.00000383, -0.00000443, -0.00000059, 0.00000134, 0.00000099, -0.00000300, -0.00001306, -0.00001139, 0.00000229, 0.00000285, -0.00000361, -0.00000560, -0.00000893, -0.00000466, 0.00000201, -0.00000210, -0.00000213, 0.00000237, -0.00000053, -0.00000183, 0.00000022, -0.00000103, -0.00000450, -0.00000610, -0.00000120, 0.00000011, -0.00000654, -0.00000446, 0.00000165, -0.00000074, -0.00000503, -0.00000502, 0.00000475, 0.00000215, -0.00001984, -0.00000477, 0.00001528, -0.00001669, -0.00000971, 0.00002278, -0.00000799, -0.00000586, 0.00001493, -0.00002173, 0.00000080, 0.00002295, -0.00002907, -0.00000530, 0.00001936, -0.00000632, 0.00001085, -0.00003417, -0.00001254, 0.00008756, -0.00006122, -0.00008742, 0.00014581, -0.00000306, -0.00015600, 0.00009820, 0.00006447, -0.00011802, 0.00001644, 0.00003818, -0.00004073, -0.00000266, -0.00002382, 0.00000854, 0.00003871, -0.00004831, -0.00001539, 0.00002166, -0.00002607, -0.00000393, -0.00002810, -0.00002336, 0.00003537, -0.00004481, -0.00004148, 0.00001306, -0.00007990, -0.00004360, -0.00007141, -0.00027692, -0.00005748, 0.00029453, 0.00017612, 0.00004695, 0.00010732, 0.00005498, 0.00002270, 0.00008018, 0.00008731, 0.00006221, 0.00003415, 0.00006671, 0.00016589, 0.00016970, 0.00004680, -0.00008448, -0.00013585, -0.00009119, -0.00003591, -0.00002577, -0.00002476, -0.00001865, -0.00001002, -0.00000630, -0.00001601, -0.00001548, -0.00000807, -0.00001112, -0.00000572, 0.00000401, 0.00000019, -0.00000563, -0.00000846, -0.00001133, -0.00000822, -0.00000134, 0.00000102, },
		{ -0.00000045, 0.00000075, 0.00000141, 0.00000126, -0.00000249, -0.00000390, -0.00000149, -0.00000026, -0.00000246, -0.00000576, -0.00000394, 0.00000034, 0.00000183, 0.00000098, -0.00000811, -0.00001505, -0.00000382, 0.00000480, -0.00000041, -0.00000470, -0.00000924, -0.00000800, 0.00000186, 0.00000020, -0.00000474, 0.00000120, 0.00000224, -0.00000208, -0.00000126, 0.00000069, -0.00000087, -0.00000622, -0.00000598, 0.00000037, -0.00000230, -0.00000658, -0.00000226, 0.00000066, -0.00000033, -0.00000488, -0.00000296, 0.00000527, -0.00000935, -0.00001468, 0.00001014, -0.00000422, -0.00001610, 0.00002002, 0.00000033, -0.00002035, 0.00002244, -0.00000395, -0.00002695, 0.00002312, 0.00000034, -0.00001555, 0.00000850, -0.00001877, 0.00002456, 0.00001143, -0.00008412, 0.00005354, 0.00007991, -0.00013612, 0.00000365, 0.00014191, -0.00009912, -0.00006654, 0.00010472, -0.00001990, -0.00004550, 0.00001715, -0.00001844, 0.00000978, -0.00001562, -0.00004133, 0.00004087, 0.00000735, -0.00003574, 0.00000276, -0.00001503, 0.00000279, -0.00000848, -0.00005707, 0.00001642, 0.00000834, -0.00005248, -0.00000148, -0.00004687, -0.00005736, -0.00002111, -0.00022868, -0.00019823, 0.00021244, 0.00026006, 0.00004981, 0.00006876, 0.00009018, 0.00002997, 0.00003991, 0.00006073, 0.00007564, 0.00006529, 0.00002501, 0.00006450, 0.00014801, 0.00015160, 0.00005753, -0.00007216, -0.00012694, -0.00008656, -0.00004013, -0.00002353, -0.00002293, -0.00001931, -0.00000611, -0.00000825, -0.00001729, -0.00001244, -0.00001017, -0.00000856, -0.00000057, -0.00000048, -0.00000338, -0.00000298, -0.00000796, -0.00001234, -0.00000708, },
		{ -0.00001128, -0.00000552, -0.00000141, 0.00000147, 0.00000319, -0.00000118, -0.00000455, -0.00000190, 0.00000057, -0.00000031, -0.00000480, -0.00000664, -0.00000208, 0.00000087, 0.00000181, -0.00000221, -0.00001270, -0.00000990, 0.00000204, 0.00000207, -0.00000194, -0.00000720, -0.00001157, -0.00000176, 0.00000381, -0.00000414, -0.00000276, 0.00000299, 0.00000031, -0.00000202, -0.00000084, 0.00000100, -0.00000176, -0.00000717, -0.00000423, -0.00000098, -0.00000395, -0.00000381, -0.00000251, -0.00000018, 0.00000070, -0.00000399, -0.00000028, -0.00000315, -0.00001390, 0.00000372, 0.00000243, -0.00002140, 0.00001404, 0.00002009, -0.00003102, 0.00000305, 0.00002467, -0.00002648, -0.00000268, 0.00001126, -0.00000740, 0.00002174, -0.00003050, -0.00002339, 0.00007573, -0.00005054, -0.00007788, 0.00012627, -0.00000251, -0.00013185, 0.00009017, 0.00005523, -0.00010530, 0.00001248, 0.00003928, -0.00001992, 0.00000235, -0.00003126, -0.00000296, 0.00002597, -0.00004732, -0.00001125, 0.00003097, -0.00001056, -0.00000316, -0.00002086, -0.00001294, 0.00002178, -0.00004478, -0.00003214, 0.00002134, -0.00002831, -0.00001115, -0.00002579, -0.00007387, -0.00000022, -0.00012918, -0.00027618, 0.00004985, 0.00029904, 0.00012206, 0.00003355, 0.00008072, 0.00005184, 0.00003952, 0.00003878, 0.00004234, 0.00007564, 0.00005654, 0.00002523, 0.00006247, 0.00012539, 0.00014321, 0.00006202, -0.00006352, -0.00011361, -0.00008513, -0.00004173, -0.00002159, -0.00002344, -0.00001557, -0.00000586, -0.00001209, -0.00001450, -0.00001213, -0.00001039, -0.00000273, -0.00000129, -0.00000590, -0.00000268, -0.00000147, -0.00000900, },
		{ -0.00000255, -0.00000891, -0.00000849, -0.00000598, -0.00000226, 0.00000394, 0.00000234, -0.00000420, -0.00000369, 0.00000006, 0.00000164, -0.00000167, -0.00000738, -0.00000504, -0.00000021, 0.00000053, -0.00000001, -0.00000759, -0.00001138, -0.00000191, 0.00000167, -0.00000084, -0.00000308, -0.00001102, -0.00000800, 0.00000367, 0.00000002, -0.00000473, 0.00000049, 0.00000153, -0.00000048, -0.00000131, -0.00000077, 0.00000063, -0.00000304, -0.00000582, -0.00000410, -0.00000398, -0.00000196, -0.00000195, -0.00000445, 0.00000037, 0.00000158, -0.00000098, -0.00000373, -0.00001414, -0.00000078, 0.00001179, -0.00002223, -0.00000721, 0.00003775, -0.00001241, -0.00002860, 0.00002495, -0.00000193, -0.00001282, 0.00000578, -0.00002052, 0.00003444, 0.00001926, -0.00008686, 0.00003769, 0.00007198, -0.00012074, 0.00000085, 0.00012567, -0.00008190, -0.00005282, 0.00009233, -0.00001959, -0.00004415, 0.00001393, -0.00000328, 0.00002165, -0.00001577, -0.00004337, 0.00002845, -0.00000381, -0.00003834, 0.00000706, 0.00000066, 0.00000922, -0.00000312, -0.00003710, 0.00001355, -0.00000406, -0.00004747, -0.00000109, -0.00001628, -0.00001433, 0.00000256, -0.00007441, -0.00002820, -0.00003596, -0.00024766, -0.00011621, 0.00023571, 0.00020382, 0.00005073, 0.00006126, 0.00005189, 0.00004132, 0.00004954, 0.00002338, 0.00003993, 0.00006860, 0.00004874, 0.00003195, 0.00005240, 0.00011035, 0.00013658, 0.00005926, -0.00005157, -0.00010213, -0.00008372, -0.00004071, -0.00002302, -0.00002103, -0.00001198, -0.00001007, -0.00001245, -0.00001215, -0.00001288, -0.00000592, -0.00000003, -0.00000634, -0.00000752, -0.00000069, },
		{ -0.00000557, -0.00000083, -0.00000427, -0.00000620, -0.00000706, -0.00000781, -0.00000047, 0.00000484, -0.00000091, -0.00000438, -0.00000200, 0.00000121, 0.00000190, -0.00000486, -0.00000763, -0.00000175, 0.00000002, -0.00000068, -0.00000464, -0.00001031, -0.00000384, 0.00000171, -0.00000202, -0.00000190, -0.00000653, -0.00001096, -0.00000091, 0.00000254, -0.00000333, -0.00000124, 0.00000061, -0.00000043, 0.00000031, -0.00000098, -0.00000107, -0.00000078, -0.00000257, -0.00000404, -0.00000709, -0.00000449, 0.00000166, -0.00000297, -0.00000540, 0.00000116, 0.00000383, 0.00000037, -0.00001519, -0.00001237, 0.00001772, -0.00000523, -0.00003018, 0.00002612, 0.00002240, -0.00003298, 0.00000067, 0.00000798, -0.00000939, 0.00001932, -0.00003233, -0.00001197, 0.00008574, -0.00004841, -0.00008405, 0.00010769, -0.00000349, -0.00012072, 0.00007862, 0.00005435, -0.00008475, 0.00001363, 0.00003331, -0.00002116, -0.00000381, -0.00002224, 0.00001218, 0.00002777, -0.00004410, -0.00001200, 0.00001937, -0.00002200, -0.00000895, -0.00001063, -0.00000153, 0.00002439, -0.00002447, -0.00001747, 0.00001035, -0.00003086, -0.00001400, -0.00001502, -0.00003227, 0.00001662, -0.00003784, -0.00006420, -0.00000480, -0.00015984, -0.00020466, 0.00011059, 0.00023084, 0.00009126, 0.00006250, 0.00005708, 0.00002691, 0.00005107, 0.00004019, 0.00001725, 0.00003999, 0.00005565, 0.00004998, 0.00003258, 0.00004289, 0.00010291, 0.00012418, 0.00005716, -0.00003921, -0.00009352, -0.00007905, -0.00004151, -0.00002411, -0.00001611, -0.00001307, -0.00001334, -0.00001065, -0.00001224, -0.00001007, -0.00000091, -0.00000311, -0.00001016, },
		{ -0.00000833, -0.00000971, -0.00000378, -0.00000328, -0.00000352, -0.00000319, -0.00000848, -0.00000726, 0.00000191, 0.00000191, -0.00000238, -0.00000251, -0.00000126, 0.00000241, -0.00000033, -0.00000751, -0.00000421, 0.00000008, -0.00000103, -0.00000369, -0.00000992, -0.00000620, 0.00000343, -0.00000055, -0.00000418, -0.00000421, -0.00000941, -0.00000461, 0.00000201, -0.00000239, -0.00000153, 0.00000126, -0.00000186, -0.00000056, 0.00000124, -0.00000101, -0.00000235, -0.00000170, -0.00000013, -0.00000504, -0.00001005, -0.00000144, 0.00000295, -0.00000470, -0.00000502, 0.00000202, 0.00000729, -0.00000590, -0.00002345, 0.00000488, 0.00001705, -0.00002717, -0.00000508, 0.00003464, -0.00001140, -0.00001122, 0.00000567, -0.00002446, 0.00002937, 0.00001176, -0.00007583, 0.00005476, 0.00007414, -0.00011945, -0.00000880, 0.00010820, -0.00007869, -0.00005178, 0.00008489, -0.00000724, -0.00003260, 0.00001244, -0.00000483, 0.00001260, -0.00001601, -0.00002714, 0.00003352, -0.00000344, -0.00003163, 0.00000609, -0.00000948, -0.00000498, -0.00000565, -0.00002630, 0.00001658, 0.00000924, -0.00002228, 0.00000054, -0.00002069, -0.00001699, -0.00000130, -0.00004421, -0.00000665, 0.00000171, -0.00006319, -0.00001738, -0.00008318, -0.00021586, -0.00000969, 0.00020878, 0.00012106, 0.00006151, 0.00007758, 0.00003074, 0.00002950, 0.00004993, 0.00002782, 0.00002106, 0.00003223, 0.00004908, 0.00005138, 0.00002752, 0.00004111, 0.00009397, 0.00010994, 0.00005789, -0.00003049, -0.00008378, -0.00007404, -0.00004383, -0.00002175, -0.00001396, -0.00001634, -0.00001327, -0.00001049, -0.00001124, -0.00000410, -0.00000060, },
		{ -0.00000011, -0.00000495, -0.00001067, -0.00000687, -0.00000486, -0.00000475, -0.00000050, -0.00000350, -0.00000931, -0.00000373, 0.00000044, -0.00000143, -0.00000059, -0.00000154, -0.00000014, 0.00000199, -0.00000462, -0.00000577, -0.00000031, -0.00000068, -0.00000274, -0.00000893, -0.00001053, 0.00000206, 0.00000458, -0.00000370, -0.00000549, -0.00000855, -0.00000618, 0.00000210, -0.00000187, -0.00000393, 0.00000257, -0.00000003, -0.00000349, 0.00000043, 0.00000166, -0.00000162, -0.00000421, -0.00000041, 0.00000127, -0.00000831, -0.00000931, 0.00000150, 0.00000156, -0.00000464, -0.00000465, 0.00000409, 0.00000693, -0.00001674, -0.00001590, 0.00001751, -0.00000486, -0.00002042, 0.00002258, 0.00000258, -0.00001222, 0.00002104, -0.00003486, -0.00001527, 0.00007156, -0.00004867, -0.00005934, 0.00011654, -0.00000528, -0.00011778, 0.00006500, 0.00004123, -0.00008460, 0.00000947, 0.00003820, -0.00000819, -0.00000023, -0.00001998, 0.00000633, 0.00001846, -0.00003360, -0.00000143, 0.00001742, -0.00001770, -0.00000176, -0.00001081, -0.00001272, 0.00001122, -0.00002298, -0.00001494, 0.00001591, -0.00000557, 0.00000077, -0.00001529, -0.00003012, 0.00000843, -0.00002522, -0.00003743, 0.00000567, -0.00003960, -0.00003105, -0.00003451, -0.00018588, -0.00010514, 0.00016045, 0.00015059, 0.00004878, 0.00007760, 0.00006035, 0.00002197, 0.00003710, 0.00003616, 0.00002641, 0.00002092, 0.00002458, 0.00004844, 0.00004596, 0.00002605, 0.00004181, 0.00008145, 0.00009939, 0.00005731, -0.00002327, -0.00007277, -0.00007131, -0.00004355, -0.00001916, -0.00001591, -0.00001729, -0.00001275, -0.00001125, -0.00000656, },
		{ -0.00000945, -0.00000097, -0.00000134, -0.00000929, -0.00000884, -0.00000576, -0.00000723, -0.00000276, 0.00000052, -0.00000591, -0.00000593, -0.00000219, -0.00000415, -0.00000112, 0.00000117, -0.00000088, 0.00000120, -0.00000230, -0.00000610, -0.00000051, 0.00000028, -0.00000228, -0.00000627, -0.00001294, -0.00000438, 0.00000690, 0.00000147, -0.00000474, -0.00000929, -0.00000960, 0.00000187, 0.00000208, -0.00000644, -0.00000045, 0.00000370, -0.00000256, -0.00000313, 0.00000107, 0.00000166, -0.00000315, -0.00000473, 0.00000176, -0.00000050, -0.00000965, -0.00000596, 0.00000103, 0.00000066, -0.00000320, -0.00000433, 0.00000564, -0.00000044, -0.00001881, 0.00000107, 0.00000583, -0.00001742, 0.00001428, 0.00001012, -0.00002913, 0.00003101, 0.00000876, -0.00007480, 0.00004403, 0.00005685, -0.00010131, 0.00001500, 0.00010544, -0.00007643, -0.00005030, 0.00007002, -0.00001787, -0.00003848, 0.00001351, 0.00000736, 0.00001858, -0.00001234, -0.00002561, 0.00002316, -0.00000287, -0.00001953, 0.00000635, -0.00001046, 0.00000201, 0.00000137, -0.00002676, 0.00000496, 0.00000148, -0.00002489, -0.00000196, -0.00000266, 0.00000592, 0.00000413, -0.00003708, -0.00000783, 0.00000126, -0.00003932, -0.00001143, -0.00002514, -0.00004052, -0.00000628, -0.00012878, -0.00017050, 0.00008096, 0.00018192, 0.00005770, 0.00004659, 0.00007533, 0.00004025, 0.00003119, 0.00003098, 0.00002737, 0.00002812, 0.00001560, 0.00002397, 0.00004452, 0.00003938, 0.00002929, 0.00003873, 0.00007068, 0.00009097, 0.00005456, -0.00001503, -0.00006390, -0.00006819, -0.00004061, -0.00002027, -0.00001802, -0.00001615, -0.00001363, },
		{ -0.00001554, -0.00001363, -0.00000493, 0.00000084, -0.00000568, -0.00000959, -0.00000618, -0.00000769, -0.00000676, -0.00000014, -0.00000265, -0.00000498, -0.00000131, -0.00000567, -0.00000629, 0.00000104, 0.00000140, 0.00000139, -0.00000102, -0.00000719, -0.00000170, 0.00000277, -0.00000164, -0.00000435, -0.00001122, -0.00001064, 0.00000296, 0.00000529, -0.00000048, -0.00000661, -0.00001338, -0.00000364, 0.00000653, -0.00000310, -0.00000563, 0.00000287, 0.00000143, -0.00000320, -0.00000229, 0.00000106, 0.00000096, -0.00000443, -0.00000330, 0.00000157, -0.00000302, -0.00000678, -0.00000452, -0.00000157, 0.00000233, -0.00000194, -0.00000380, 0.00000211, -0.00000769, -0.00000515, 0.00000414, -0.00001929, 0.00000606, 0.00003425, -0.00003844, -0.00001182, 0.00006783, -0.00004976, -0.00005835, 0.00009433, -0.00000918, -0.00008761, 0.00007417, 0.00003620, -0.00007734, 0.00000704, 0.00002483, -0.00002087, -0.00000635, -0.00001049, 0.00001562, 0.00002109, -0.00002833, -0.00000567, 0.00001211, -0.00000890, 0.00000166, -0.00001488, -0.00000979, 0.00001857, -0.00001636, -0.00001680, 0.00000863, -0.00001194, -0.00000989, -0.00001050, -0.00000623, 0.00001992, -0.00001706, -0.00002854, 0.00000406, -0.00002323, -0.00002017, -0.00001717, -0.00005136, -0.00000667, -0.00005998, -0.00018415, -0.00002240, 0.00018034, 0.00010243, 0.00002405, 0.00005736, 0.00005279, 0.00004026, 0.00003473, 0.00002052, 0.00002685, 0.00002426, 0.00001386, 0.00002482, 0.00003660, 0.00003753, 0.00003081, 0.00003396, 0.00006352, 0.00008175, 0.00005254, -0.00000779, -0.00005707, -0.00006210, -0.00003920, -0.00002372, -0.00001784, },
		{ -0.00002598, -0.00001687, -0.00001529, -0.00001064, -0.00000155, -0.00000226, -0.00000815, -0.00000644, -0.00000656, -0.00000909, -0.00000310, -0.00000181, -0.00000533, 0.00000030, -0.00000156, -0.00000967, -0.00000451, 0.00000120, 0.00000302, 0.00000243, -0.00000730, -0.00000630, 0.00000421, 0.00000166, -0.00000346, -0.00000842, -0.00001282, -0.00000291, 0.00000454, 0.00000181, -0.00000108, -0.00001105, -0.00001151, 0.00000404, 0.00000357, -0.00000563, -0.00000137, 0.00000207, -0.00000049, -0.00000210, -0.00000225, 0.00000049, -0.00000017, -0.00000386, -0.00000235, -0.00000143, -0.00000206, -0.00000265, -0.00000715, -0.00000250, 0.00000548, -0.00000133, -0.00000531, -0.00000646, -0.00000260, 0.00001112, -0.00001822, -0.00002349, 0.00004928, 0.00000538, -0.00007046, 0.00004624, 0.00004871, -0.00009577, 0.00000530, 0.00008366, -0.00005975, -0.00002673, 0.00006655, -0.00001651, -0.00003131, 0.00000884, -0.00000568, 0.00000427, -0.00001183, -0.00001472, 0.00002694, 0.00000151, -0.00001798, 0.00000120, -0.00000517, 0.00000805, -0.00000383, -0.00002844, 0.00000994, 0.00000838, -0.00002002, -0.00000212, -0.00000506, -0.00000653, -0.00000594, -0.00002196, 0.00000816, 0.00000849, -0.00002570, -0.00000622, -0.00001374, -0.00002301, -0.00000589, -0.00004832, -0.00003262, -0.00001544, -0.00014470, -0.00010786, 0.00012624, 0.00014855, 0.00003771, 0.00003436, 0.00004511, 0.00004033, 0.00004690, 0.00002756, 0.00001762, 0.00002494, 0.00001930, 0.00001651, 0.00002153, 0.00003119, 0.00003732, 0.00002861, 0.00003144, 0.00005715, 0.00007310, 0.00005135, -0.00000351, -0.00004900, -0.00005540, -0.00004092, },
		{ -0.00005141, -0.00004318, -0.00002649, -0.00001643, -0.00001305, -0.00000640, -0.00000281, -0.00000673, -0.00000609, -0.00000432, -0.00000909, -0.00000634, -0.00000162, -0.00000655, -0.00000234, 0.00000346, -0.00000599, -0.00000901, -0.00000339, 0.00000095, 0.00000621, -0.00000209, -0.00001091, 0.00000020, 0.00000552, -0.00000071, -0.00000601, -0.00001262, -0.00000737, 0.00000272, 0.00000088, 0.00000047, -0.00000409, -0.00001298, -0.00000360, 0.00000516, -0.00000150, -0.00000233, 0.00000035, -0.00000093, 0.00000008, -0.00000121, -0.00000300, -0.00000063, -0.00000032, -0.00000205, -0.00000444, -0.00000362, 0.00000294, -0.00000192, -0.00001107, -0.00000015, 0.00000773, -0.00000137, -0.00001154, -0.00001180, 0.00001621, 0.00000855, -0.00004682, 0.00000972, 0.00006645, -0.00005249, -0.00004713, 0.00008716, -0.00001187, -0.00008364, 0.00005401, 0.00003166, -0.00005277, 0.00001373, 0.00002083, -0.00001414, -0.00000225, -0.00001570, 0.00000130, 0.00001110, -0.00002272, -0.00000021, 0.00001511, -0.00000495, -0.00000126, -0.00001189, -0.00000046, 0.00001527, -0.00002295, -0.00001499, 0.00001368, -0.00000659, -0.00000570, -0.00000558, -0.00001053, 0.00000331, -0.00001694, -0.00001431, 0.00001041, -0.00001122, -0.00000818, -0.00000686, -0.00002866, -0.00000397, -0.00002560, -0.00005278, -0.00001181, -0.00008913, -0.00014518, 0.00004442, 0.00016210, 0.00007218, 0.00002905, 0.00003690, 0.00002702, 0.00004310, 0.00004399, 0.00002202, 0.00001842, 0.00001929, 0.00001873, 0.00001728, 0.00001743, 0.00002961, 0.00003421, 0.00002623, 0.00003035, 0.00005090, 0.00006663, 0.00004846, 0.00000000, -0.00003941, },
		{ 0.00000497, -0.00003145, -0.00004970, -0.00004383, -0.00002672, -0.00001522, -0.00000843, -0.00000483, -0.00000797, -0.00000780, -0.00000213, -0.00000583, -0.00000876, -0.00000238, -0.00000558, -0.00000719, 0.00000293, 0.00000013, -0.00000758, -0.00000618, -0.00000471, 0.00000368, 0.00000533, -0.00000875, -0.00000658, 0.00000480, 0.00000284, -0.00000235, -0.00001058, -0.00001140, 0.00000090, 0.00000165, -0.00000223, -0.00000120, -0.00000829, -0.00000745, 0.00000212, -0.00000023, -0.00000135, 0.00000197, -0.00000216, -0.00000215, 0.00000180, -0.00000145, -0.00000424, -0.00000167, 0.00000133, -0.00000169, -0.00000848, -0.00000149, 0.00000770, -0.00000539, -0.00001179, 0.00000290, 0.00000814, -0.00000252, -0.00002125, -0.00000529, 0.00003439, -0.00001711, -0.00005073, 0.00005565, 0.00003526, -0.00008630, 0.00001023, 0.00007347, -0.00005569, -0.00003396, 0.00005159, -0.00000298, -0.00001724, 0.00000538, -0.00000340, 0.00001085, -0.00000990, -0.00002113, 0.00001452, -0.00000118, -0.00001469, 0.00000244, -0.00000112, 0.00000744, -0.00000348, -0.00001694, 0.00001298, -0.00000065, -0.00002310, 0.00000110, -0.00000122, -0.00000304, 0.00000153, -0.00001529, -0.00000201, -0.00000173, -0.00001920, -0.00000224, -0.00000758, -0.00000873, 0.00000516, -0.00002463, -0.00001742, -0.00000610, -0.00004746, -0.00002647, -0.00004971, -0.00014311, -0.00003335, 0.00014149, 0.00010551, 0.00003523, 0.00003845, 0.00002243, 0.00002325, 0.00004418, 0.00003596, 0.00002252, 0.00001609, 0.00001560, 0.00001948, 0.00001511, 0.00001675, 0.00002757, 0.00002935, 0.00002518, 0.00002885, 0.00004660, 0.00006056, 0.00004405, },
		{ 0.00005336, 0.00004078, 0.00000985, -0.00002674, -0.00004779, -0.00004329, -0.00002666, -0.00001180, -0.00000418, -0.00000795, -0.00001247, -0.00000467, -0.00000112, -0.00000741, -0.00000433, -0.00000357, -0.00000951, -0.00000170, 0.00000254, -0.00000444, -0.00000430, -0.00000625, -0.00000403, 0.00000600, -0.00000170, -0.00000880, 0.00000084, 0.00000332, 0.00000063, -0.00000545, -0.00001331, -0.00000377, 0.00000416, -0.00000229, -0.00000353, -0.00000585, -0.00000706, 0.00000135, -0.00000029, -0.00000414, 0.00000352, 0.00000203, -0.00000519, -0.00000085, 0.00000258, -0.00000235, -0.00000565, -0.00000159, 0.00000346, -0.00000424, -0.00001046, 0.00000392, 0.00000765, -0.00000835, -0.00000869, 0.00000221, 0.00000894, -0.00000551, -0.00002962, 0.00001246, 0.00003619, -0.00004645, -0.00002235, 0.00007471, -0.00001763, -0.00007026, 0.00004871, 0.00002791, -0.00005234, 0.00000095, 0.00002379, 0.00000100, -0.00000318, -0.00001755, 0.00000702, 0.00001522, -0.00002227, -0.00000691, 0.00000956, -0.00000511, -0.00000354, -0.00001081, 0.00000080, 0.00001340, -0.00001322, -0.00000426, 0.00000808, -0.00001429, -0.00000556, -0.00000294, -0.00000889, 0.00000785, -0.00000577, -0.00001253, 0.00000102, -0.00001093, -0.00000714, -0.00000894, -0.00001852, 0.00000779, -0.00000532, -0.00002626, -0.00000552, -0.00003093, -0.00003562, -0.00002755, -0.00011954, -0.00009474, 0.00009452, 0.00013108, 0.00004751, 0.00003788, 0.00003277, 0.00001256, 0.00002692, 0.00003715, 0.00003225, 0.00002296, 0.00001261, 0.00001511, 0.00001761, 0.00001417, 0.00001760, 0.00002350, 0.00002539, 0.00002384, 0.00002800, 0.00004390, },
		{ 0.00002867, 0.00004058, 0.00004676, 0.00003907, 0.00001206, -0.00002373, -0.00004477, -0.00004218, -0.00002501, -0.00000716, -0.00000368, -0.00001340, -0.00001239, -0.00000187, -0.00000318, -0.00000427, -0.00000197, -0.00000920, -0.00000636, 0.00000221, -0.00000344, -0.00000357, -0.00000196, -0.00000721, -0.00000020, 0.00000139, -0.00000694, -0.00000095, 0.00000289, -0.00000026, -0.00000107, -0.00000990, -0.00000916, 0.00000324, 0.00000139, -0.00000415, -0.00000690, -0.00000907, 0.00000124, 0.00000431, -0.00000642, -0.00000171, 0.00000671, -0.00000124, -0.00000541, 0.00000061, 0.00000208, -0.00000316, -0.00000682, -0.00000027, 0.00000382, -0.00000771, -0.00000763, 0.00000695, 0.00000376, -0.00000506, -0.00000793, -0.00000251, 0.00001430, -0.00001237, -0.00003220, 0.00003280, 0.00002008, -0.00005744, 0.00001519, 0.00005527, -0.00004910, -0.00002923, 0.00004523, -0.00000314, -0.00002547, 0.00000293, 0.00001124, 0.00001342, -0.00001607, -0.00001797, 0.00001896, 0.00000102, -0.00001571, -0.00000003, -0.00000041, 0.00000366, -0.00000817, -0.00001694, 0.00001023, 0.00000428, -0.00000836, 0.00000336, -0.00000959, -0.00000648, 0.00000251, -0.00001470, -0.00000114, 0.00000677, -0.00001009, -0.00000450, -0.00000733, -0.00000652, -0.00000267, -0.00002454, -0.00000800, 0.00000874, -0.00001758, -0.00001100, -0.00001924, -0.00003895, -0.00001714, -0.00008163, -0.00013022, 0.00002552, 0.00013897, 0.00007170, 0.00003236, 0.00004070, 0.00001990, 0.00001460, 0.00002501, 0.00002992, 0.00003177, 0.00002076, 0.00001180, 0.00001402, 0.00001500, 0.00001545, 0.00001698, 0.00001955, 0.00002187, 0.00002221, },
		{ 0.00001802, 0.00002170, 0.00002911, 0.00003683, 0.00004250, 0.00003661, 0.00001157, -0.00002073, -0.00004097, -0.00003992, -0.00002151, -0.00000436, -0.00000745, -0.00001607, -0.00000930, -0.00000329, -0.00000269, 0.00000102, -0.00000639, -0.00001056, 0.00000070, -0.00000160, -0.00000661, 0.00000005, -0.00000323, -0.00000389, -0.00000028, -0.00000742, -0.00000260, 0.00000575, -0.00000068, -0.00000209, -0.00000499, -0.00001012, -0.00000071, 0.00000353, -0.00000167, -0.00000560, -0.00001229, -0.00000487, 0.00000886, -0.00000038, -0.00000799, 0.00000315, 0.00000499, -0.00000318, -0.00000366, 0.00000028, 0.00000159, -0.00000372, -0.00000734, 0.00000106, 0.00000158, -0.00000767, -0.00000263, 0.00000275, 0.00000417, 0.00000278, -0.00001747, -0.00000284, 0.00002389, -0.00002629, -0.00002566, 0.00004710, -0.00000288, -0.00004668, 0.00003504, 0.00002068, -0.00004542, -0.00000203, 0.00002210, -0.00000476, -0.00000968, -0.00000426, 0.00001647, 0.00000708, -0.00002470, -0.00000280, 0.00001155, -0.00000462, -0.00000328, -0.00000617, 0.00000166, 0.00000579, -0.00001817, -0.00000867, 0.00000748, -0.00000084, 0.00000442, -0.00000758, -0.00001408, 0.00000723, -0.00000610, -0.00001389, 0.00000470, -0.00000127, -0.00000401, -0.00000639, -0.00001054, 0.00000283, -0.00001483, -0.00002422, 0.00000266, -0.00000587, -0.00001032, -0.00000930, -0.00003834, -0.00002108, -0.00004197, -0.00012841, -0.00004772, 0.00011409, 0.00010161, 0.00003531, 0.00003758, 0.00003064, 0.00001616, 0.00001737, 0.00001853, 0.00002698, 0.00002975, 0.00001912, 0.00001217, 0.00001149, 0.00001422, 0.00001626, 0.00001557, 0.00001647, },
		{ 0.00001470, 0.00001332, 0.00001513, 0.00002161, 0.00002811, 0.00003472, 0.00003941, 0.00003200, 0.00000993, -0.00001730, -0.00003625, -0.00003589, -0.00001793, -0.00000598, -0.00001166, -0.00001369, -0.00000882, -0.00000583, 0.00000283, 0.00000069, -0.00001181, -0.00000399, 0.00000180, -0.00000763, -0.00000378, -0.00000018, -0.00000242, 0.00000044, -0.00000845, -0.00000952, 0.00000677, 0.00000491, -0.00000377, -0.00000450, -0.00000861, -0.00000324, 0.00000375, -0.00000023, -0.00000249, -0.00001005, -0.00001313, 0.00000405, 0.00000836, -0.00000517, -0.00000368, 0.00000404, 0.00000161, -0.00000167, -0.00000314, -0.00000112, 0.00000207, -0.00000440, -0.00000696, 0.00000026, -0.00000065, -0.00000144, -0.00000362, -0.00000497, 0.00001669, 0.00000245, -0.00003473, 0.00001220, 0.00002705, -0.00004531, -0.00000744, 0.00005022, -0.00002144, -0.00002709, 0.00003203, -0.00000166, -0.00002646, 0.00000083, 0.00000822, 0.00000364, -0.00000843, -0.00000093, 0.00001569, -0.00000732, -0.00001460, 0.00000169, -0.00000041, 0.00000351, -0.00000231, -0.00000862, 0.00000741, -0.00000393, -0.00001608, -0.00000317, -0.00000193, 0.00000776, 0.00000370, -0.00001945, -0.00000203, 0.00000625, -0.00001276, -0.00000490, -0.00000037, -0.00000158, -0.00000039, -0.00001255, -0.00000241, -0.00000153, -0.00002390, -0.00001078, -0.00000478, -0.00001052, 0.00000022, -0.00002688, -0.00003265, -0.00001995, -0.00009697, -0.00009714, 0.00005982, 0.00011712, 0.00005155, 0.00003239, 0.00003474, 0.00002164, 0.00001921, 0.00001441, 0.00001430, 0.00002503, 0.00002662, 0.00001916, 0.00001152, 0.00000984, 0.00001401, 0.00001585, },
		{ 0.00001284, 0.00001544, 0.00001373, 0.00001081, 0.00001361, 0.00002004, 0.00002720, 0.00003450, 0.00003548, 0.00002592, 0.00000818, -0.00001313, -0.00003023, -0.00003068, -0.00001726, -0.00001098, -0.00001135, -0.00001037, -0.00001232, -0.00000347, 0.00000676, -0.00000565, -0.00000909, 0.00000197, -0.00000428, -0.00000744, -0.00000178, -0.00000236, 0.00000317, -0.00000288, -0.00001625, -0.00000159, 0.00001045, 0.00000079, -0.00000501, -0.00000920, -0.00000587, 0.00000468, 0.00000146, -0.00000236, -0.00000475, -0.00001411, -0.00000626, 0.00000834, 0.00000246, -0.00000350, 0.00000023, 0.00000030, 0.00000139, 0.00000071, -0.00000515, -0.00000138, 0.00000261, -0.00000465, -0.00000653, -0.00000404, 0.00000317, 0.00000619, -0.00001662, -0.00000236, 0.00003538, -0.00001833, -0.00004129, 0.00003978, 0.00001036, -0.00005656, 0.00001914, 0.00003975, -0.00003057, -0.00001133, 0.00001916, -0.00000607, -0.00001393, -0.00000505, 0.00000717, 0.00000594, -0.00000573, 0.00000407, 0.00000242, -0.00000814, -0.00000296, -0.00000703, 0.00000008, 0.00000877, -0.00000654, -0.00000235, 0.00000268, -0.00000970, -0.00000722, -0.00000907, -0.00000089, 0.00001404, -0.00000923, -0.00001442, 0.00000577, -0.00000375, -0.00000743, -0.00000312, -0.00000527, 0.00000315, -0.00000530, -0.00000923, 0.00000133, -0.00001440, -0.00001614, -0.00000737, -0.00001614, -0.00000089, -0.00000827, -0.00003462, -0.00001826, -0.00006078, -0.00011169, -0.00000123, 0.00010803, 0.00007131, 0.00003200, 0.00003490, 0.00002511, 0.00002131, 0.00001943, 0.00001009, 0.00001285, 0.00002187, 0.00002451, 0.00001907, 0.00001089, 0.00000929, },
		{ 0.00001123, 0.00000829, 0.00001150, 0.00001484, 0.00001264, 0.00001020, 0.00001160, 0.00001738, 0.00002776, 0.00003439, 0.00003033, 0.00001942, 0.00000711, -0.00000794, -0.00002302, -0.00002641, -0.00002114, -0.00001429, -0.00000682, -0.00001131, -0.00001383, 0.00000203, 0.00000262, -0.00000799, -0.00000055, -0.00000099, -0.00000792, -0.00000334, -0.00000462, 0.00000051, 0.00000556, -0.00001255, -0.00001329, 0.00000642, 0.00000675, -0.00000082, -0.00000825, -0.00001110, 0.00000254, 0.00000625, -0.00000228, -0.00000342, -0.00000953, -0.00001117, 0.00000168, 0.00000414, -0.00000006, 0.00000230, -0.00000141, -0.00000298, 0.00000467, 0.00000028, -0.00000721, -0.00000044, 0.00000216, -0.00000266, -0.00000981, -0.00000821, 0.00001691, 0.00000253, -0.00003370, 0.00002089, 0.00003849, -0.00005147, -0.00001937, 0.00005703, -0.00002323, -0.00004505, 0.00004036, 0.00001854, -0.00002925, -0.00000339, 0.00000846, -0.00000283, -0.00001075, -0.00000710, 0.00000781, 0.00000638, 0.00000105, -0.00000217, -0.00000643, 0.00000200, -0.00000469, -0.00001157, 0.00000657, 0.00000509, -0.00000363, -0.00000022, -0.00000667, -0.00000427, -0.00000678, -0.00001323, 0.00000764, 0.00000534, -0.00001374, -0.00000167, -0.00000023, -0.00000596, -0.00000112, -0.00000790, -0.00000263, 0.00000168, -0.00000754, -0.00000132, -0.00000801, -0.00001617, -0.00000601, -0.00001726, -0.00001319, 0.00000057, -0.00002363, -0.00002161, -0.00003494, -0.00010159, -0.00005118, 0.00007952, 0.00008660, 0.00003584, 0.00003256, 0.00002872, 0.00002065, 0.00002307, 0.00001583, 0.00000827, 0.00001136, 0.00001871, 0.00002292, 0.00001870, },
		{ 0.00002133, 0.00001907, 0.00001134, 0.00000708, 0.00001026, 0.00001337, 0.00001308, 0.00001035, 0.00000821, 0.00001550, 0.00002885, 0.00003323, 0.00002448, 0.00001351, 0.00000738, -0.00000198, -0.00001562, -0.00002590, -0.00002672, -0.00001217, -0.00000410, -0.00001426, -0.00000940, 0.00000164, -0.00000407, -0.00000060, 0.00000219, -0.00000771, -0.00000432, -0.00000357, -0.00000581, 0.00000603, -0.00000183, -0.00001602, -0.00000300, 0.00000573, 0.00000301, -0.00000187, -0.00001263, -0.00000560, 0.00000840, 0.00000263, -0.00000373, -0.00000720, -0.00001090, -0.00000215, 0.00000210, -0.00000218, 0.00000436, 0.00000495, -0.00000654, -0.00000163, 0.00000697, -0.00000237, -0.00000676, -0.00000151, 0.00000351, 0.00000031, -0.00002022, -0.00000228, 0.00003255, -0.00002022, -0.00003358, 0.00005269, 0.00001027, -0.00006708, 0.00002231, 0.00004206, -0.00004873, -0.00001314, 0.00004084, -0.00000220, -0.00001981, -0.00000239, 0.00000286, 0.00000014, -0.00001022, -0.00000659, 0.00000745, 0.00001084, 0.00000077, -0.00001241, -0.00000166, 0.00000561, -0.00001129, -0.00000627, 0.00000644, 0.00000305, 0.00000189, -0.00000719, -0.00000717, 0.00000046, -0.00001248, -0.00000666, 0.00000664, -0.00000631, -0.00000378, 0.00000086, -0.00000747, -0.00000024, -0.00000276, -0.00000848, -0.00000043, -0.00000341, -0.00000129, -0.00000335, -0.00001653, -0.00000644, -0.00000940, -0.00002152, -0.00000569, -0.00001350, -0.00002147, -0.00001835, -0.00007800, -0.00008385, 0.00003796, 0.00009403, 0.00004575, 0.00002717, 0.00003069, 0.00002147, 0.00002180, 0.00002132, 0.00001271, 0.00000781, 0.00000938, 0.00001607, },
		{ 0.00000739, 0.00001363, 0.00002059, 0.00001931, 0.00001094, 0.00000647, 0.00000797, 0.00001244, 0.00001513, 0.00000938, 0.00000501, 0.00001458, 0.00002906, 0.00003093, 0.00001869, 0.00000951, 0.00000872, 0.00000428, -0.00001038, -0.00002944, -0.00002832, -0.00000798, -0.00000610, -0.00001208, -0.00000555, -0.00000663, -0.00000211, 0.00000777, -0.00000397, -0.00000786, -0.00000093, -0.00000696, -0.00000065, 0.00000365, -0.00001090, -0.00000743, 0.00000197, 0.00000020, 0.00000230, -0.00000579, -0.00001159, 0.00000269, 0.00000718, -0.00000016, -0.00000584, -0.00001166, -0.00000436, 0.00000404, -0.00000456, -0.00000253, 0.00001041, 0.00000142, -0.00000876, 0.00000204, 0.00000515, -0.00000228, -0.00000687, -0.00000510, 0.00001145, -0.00000377, -0.00003191, 0.00001979, 0.00003114, -0.00004872, -0.00000419, 0.00006214, -0.00003343, -0.00004380, 0.00004890, 0.00000308, -0.00004277, 0.00001512, 0.00002093, -0.00000979, -0.00000945, -0.00000608, 0.00000154, 0.00000177, -0.00001003, -0.00000603, 0.00001044, 0.00001270, -0.00000793, -0.00001258, 0.00000520, -0.00000047, -0.00001011, -0.00000184, 0.00000249, 0.00000705, -0.00000047, -0.00001202, -0.00000129, -0.00000348, -0.00001104, -0.00000066, -0.00000523, -0.00000487, 0.00000528, -0.00000583, -0.00000563, 0.00000186, -0.00000602, -0.00000492, -0.00000404, -0.00000193, 0.00000238, -0.00001301, -0.00001210, -0.00000276, -0.00001767, -0.00001436, -0.00001169, -0.00002223, -0.00001067, -0.00004751, -0.00009411, -0.00001018, 0.00008653, 0.00006263, 0.00002472, 0.00002731, 0.00002343, 0.00001976, 0.00002215, 0.00001781, 0.00001144, 0.00000725, },
		{ 0.00001101, 0.00000636, 0.00000528, 0.00001229, 0.00002020, 0.00001851, 0.00001131, 0.00000531, 0.00000515, 0.00001346, 0.00001660, 0.00000780, 0.00000316, 0.00001336, 0.00002790, 0.00002764, 0.00001455, 0.00000773, 0.00001041, 0.00000972, -0.00000975, -0.00003235, -0.00002467, -0.00000814, -0.00000698, -0.00000526, -0.00001066, -0.00001193, 0.00000776, 0.00000639, -0.00000860, -0.00000276, -0.00000357, -0.00000483, 0.00000267, -0.00000743, -0.00000870, 0.00000310, -0.00000207, -0.00000250, 0.00000016, -0.00000854, -0.00000377, 0.00000551, 0.00000325, -0.00000127, -0.00001073, -0.00001015, 0.00000519, 0.00000154, -0.00000986, 0.00000320, 0.00001092, -0.00000332, -0.00000562, 0.00000150, 0.00000356, 0.00000188, -0.00001402, -0.00000330, 0.00002266, -0.00002148, -0.00002889, 0.00004570, 0.00000343, -0.00005443, 0.00003516, 0.00003320, -0.00005362, 0.00000102, 0.00003603, -0.00002452, -0.00001237, 0.00001745, -0.00000038, -0.00000335, -0.00000615, -0.00000979, 0.00000343, 0.00000142, -0.00001039, -0.00000245, 0.00001392, 0.00000567, -0.00001271, -0.00000467, 0.00000300, -0.00000442, -0.00000428, -0.00000354, 0.00000345, 0.00000818, -0.00000720, -0.00000800, -0.00000088, -0.00000707, -0.00000312, -0.00000661, -0.00001183, 0.00000504, 0.00000304, -0.00000850, -0.00000093, -0.00000151, -0.00000451, -0.00000505, -0.00000611, 0.00000352, -0.00000417, -0.00001522, -0.00000419, -0.00000985, -0.00001601, -0.00001184, -0.00002438, -0.00001489, -0.00002189, -0.00008047, -0.00005067, 0.00005955, 0.00007762, 0.00003188, 0.00002200, 0.00002308, 0.00001842, 0.00002046, 0.00001980, 0.00001526, },
		{ 0.00001679, 0.00001406, 0.00001065, 0.00000476, 0.00000404, 0.00001219, 0.00001852, 0.00001804, 0.00001199, 0.00000277, 0.00000405, 0.00001494, 0.00001661, 0.00000738, 0.00000192, 0.00001118, 0.00002502, 0.00002454, 0.00001344, 0.00000706, 0.00001208, 0.00001163, -0.00001218, -0.00002950, -0.00002183, -0.00001167, -0.00000117, -0.00000376, -0.00002007, -0.00000601, 0.00001244, -0.00000060, -0.00000483, -0.00000219, -0.00000549, 0.00000198, -0.00000498, -0.00001363, 0.00000313, 0.00000429, -0.00000702, -0.00000333, -0.00000442, -0.00000409, 0.00000322, 0.00000219, 0.00000137, -0.00000440, -0.00001368, -0.00000171, 0.00000875, -0.00000538, -0.00000774, 0.00000743, 0.00000570, -0.00000195, -0.00000341, -0.00000419, 0.00001010, 0.00000127, -0.00002608, 0.00001236, 0.00002246, -0.00004332, -0.00000300, 0.00005024, -0.00003130, -0.00002610, 0.00004758, -0.00000963, -0.00003145, 0.00002514, 0.00000075, -0.00001715, 0.00001041, -0.00000155, -0.00000546, 0.00000466, -0.00000987, -0.00000862, 0.00000508, -0.00000146, -0.00000799, 0.00000369, 0.00001010, -0.00000311, -0.00000747, -0.00000129, -0.00000299, -0.00000126, -0.00000281, -0.00000441, 0.00000653, 0.00000208, -0.00000762, -0.00000343, -0.00000567, -0.00000237, -0.00000209, -0.00001658, -0.00000576, 0.00000840, -0.00000272, -0.00000428, -0.00000128, -0.00000308, -0.00000268, -0.00000793, -0.00000186, 0.00000194, -0.00001128, -0.00000796, -0.00000589, -0.00001413, -0.00000989, -0.00002136, -0.00002485, -0.00001284, -0.00005509, -0.00007030, 0.00002158, 0.00007974, 0.00004580, 0.00002081, 0.00002124, 0.00001717, 0.00001744, 0.00001950, },
		{ 0.00001703, 0.00001680, 0.00001462, 0.00001378, 0.00000938, 0.00000311, 0.00000483, 0.00001125, 0.00001623, 0.00001877, 0.00001104, 0.00000092, 0.00000475, 0.00001472, 0.00001676, 0.00000812, 0.00000044, 0.00000759, 0.00002065, 0.00002389, 0.00001460, 0.00000680, 0.00001227, 0.00000844, -0.00001157, -0.00002305, -0.00002438, -0.00001063, 0.00000536, -0.00001163, -0.00002063, 0.00000283, 0.00000640, -0.00000106, -0.00000019, -0.00000636, 0.00000023, 0.00000145, -0.00001662, -0.00000624, 0.00001053, -0.00000109, -0.00000786, -0.00000719, -0.00000497, 0.00000480, 0.00000323, -0.00000146, -0.00000013, -0.00000895, -0.00000944, 0.00000613, 0.00000408, -0.00000770, -0.00000148, 0.00000371, 0.00000320, 0.00000489, -0.00001064, -0.00000479, 0.00002260, -0.00001401, -0.00002774, 0.00003459, 0.00000133, -0.00004690, 0.00002865, 0.00002464, -0.00004038, 0.00001007, 0.00002184, -0.00002429, 0.00000614, 0.00000994, -0.00001687, 0.00000920, 0.00000527, -0.00001691, 0.00000275, 0.00000358, -0.00001227, -0.00000372, 0.00000259, -0.00000310, -0.00000103, 0.00000496, 0.00000037, -0.00000345, -0.00000087, -0.00000519, -0.00000346, 0.00000182, -0.00000432, -0.00000096, 0.00000369, -0.00000297, -0.00000244, -0.00000541, -0.00000547, 0.00000279, -0.00001077, -0.00001661, 0.00000202, 0.00000308, -0.00000243, -0.00000113, -0.00000379, -0.00000161, -0.00000467, -0.00000636, 0.00000103, -0.00000599, -0.00000884, -0.00000409, -0.00001231, -0.00000975, -0.00001327, -0.00002829, -0.00001759, -0.00003510, -0.00007065, -0.00001358, 0.00006772, 0.00005914, 0.00002459, 0.00001987, 0.00001725, 0.00001428, },
		{ 0.00001365, 0.00001371, 0.00001522, 0.00001374, 0.00001383, 0.00001320, 0.00000693, 0.00000357, 0.00000599, 0.00000865, 0.00001550, 0.00001859, 0.00000912, 0.00000166, 0.00000467, 0.00001346, 0.00001804, 0.00000904, -0.00000154, 0.00000190, 0.00001703, 0.00002628, 0.00001638, 0.00000676, 0.00000804, 0.00000405, -0.00000494, -0.00002033, -0.00002701, -0.00000380, 0.00000234, -0.00001779, -0.00001083, 0.00000145, 0.00000030, 0.00000529, -0.00000255, -0.00000518, 0.00000594, -0.00000979, -0.00001712, 0.00000524, 0.00000741, -0.00000371, -0.00000909, -0.00001152, 0.00000189, 0.00000954, -0.00000092, -0.00000311, -0.00000389, -0.00000910, -0.00000064, 0.00000594, -0.00000184, -0.00000173, -0.00000067, -0.00000456, 0.00001134, 0.00000547, -0.00002407, 0.00001016, 0.00002465, -0.00003699, -0.00000761, 0.00004148, -0.00002665, -0.00002365, 0.00003715, -0.00000662, -0.00001726, 0.00001731, -0.00001007, -0.00000131, 0.00001717, -0.00001758, -0.00000191, 0.00002003, -0.00001440, -0.00001406, 0.00000775, -0.00000287, -0.00000792, -0.00000179, -0.00000208, 0.00000112, 0.00000450, -0.00000204, -0.00000492, 0.00000247, -0.00000126, -0.00000788, -0.00000012, 0.00000065, -0.00000253, 0.00000025, -0.00000294, -0.00000088, -0.00000085, -0.00000793, 0.00000037, -0.00000154, -0.00001671, -0.00000804, 0.00000124, -0.00000106, 0.00000136, -0.00000216, -0.00000355, -0.00000209, -0.00000581, -0.00000170, -0.00000371, -0.00000902, -0.00000283, -0.00000851, -0.00001204, -0.00000791, -0.00002254, -0.00002326, -0.00002560, -0.00006237, -0.00004024, 0.00004598, 0.00006828, 0.00003251, 0.00001845, 0.00001825, },
		{ 0.00001721, 0.00001506, 0.00001234, 0.00001316, 0.00001216, 0.00001165, 0.00001389, 0.00001080, 0.00000561, 0.00000599, 0.00000497, 0.00000690, 0.00001547, 0.00001614, 0.00000924, 0.00000326, 0.00000284, 0.00001292, 0.00001961, 0.00000996, -0.00000505, -0.00000476, 0.00001653, 0.00002972, 0.00001865, 0.00000474, -0.00000045, 0.00000478, 0.00000206, -0.00002205, -0.00002250, -0.00000166, -0.00000720, -0.00001131, -0.00000471, -0.00000772, 0.00000340, 0.00000739, -0.00000558, 0.00000260, 0.00000005, -0.00001797, -0.00000564, 0.00000721, 0.00000196, -0.00000308, -0.00001364, -0.00000940, 0.00000985, 0.00000747, -0.00000393, -0.00000556, -0.00000690, -0.00000221, 0.00000434, -0.00000140, -0.00000003, 0.00000615, -0.00001096, -0.00000476, 0.00002492, -0.00001010, -0.00002633, 0.00003296, 0.00000423, -0.00004474, 0.00002169, 0.00002264, -0.00003481, 0.00000496, 0.00001745, -0.00001209, 0.00000703, -0.00000580, -0.00001094, 0.00002449, -0.00000402, -0.00002052, 0.00001988, 0.00000504, -0.00002237, -0.00000342, 0.00000233, -0.00000397, -0.00000193, -0.00000555, -0.00000202, 0.00000822, 0.00000168, -0.00000970, -0.00000169, 0.00000486, -0.00000486, -0.00000535, 0.00000046, 0.00000006, 0.00000070, -0.00000448, -0.00000478, 0.00000259, -0.00000378, -0.00000413, 0.00000146, -0.00001043, -0.00001206, -0.00000306, -0.00000411, 0.00000033, 0.00000230, -0.00000300, -0.00000268, -0.00000431, -0.00000250, -0.00000202, -0.00000952, -0.00000460, -0.00000344, -0.00001179, -0.00000798, -0.00001504, -0.00002352, -0.00002110, -0.00005086, -0.00005816, 0.00001661, 0.00006985, 0.00004532, 0.00001878, },
		{ 0.00002562, 0.00001515, 0.00001512, 0.00001261, 0.00001234, 0.00001148, 0.00000894, 0.00001132, 0.00001254, 0.00000775, 0.00000738, 0.00000694, 0.00000307, 0.00000736, 0.00001323, 0.00001397, 0.00001169, 0.00000320, 0.00000106, 0.00001296, 0.00002102, 0.00001022, -0.00001060, -0.00000905, 0.00001824, 0.00003286, 0.00002089, -0.00000197, -0.00000711, 0.00001009, 0.00000316, -0.00001980, -0.00001562, -0.00001064, -0.00000854, 0.00000075, -0.00001059, -0.00001094, 0.00000870, 0.00000309, -0.00000008, 0.00000420, -0.00001239, -0.00001145, 0.00000301, 0.00000077, 0.00000126, -0.00000509, -0.00001581, -0.00000143, 0.00001162, 0.00000340, -0.00000577, -0.00000951, -0.00000403, 0.00000681, 0.00000003, -0.00000732, 0.00001146, 0.00000398, -0.00002474, 0.00001119, 0.00002622, -0.00003275, -0.00000593, 0.00003994, -0.00002549, -0.00002548, 0.00003322, -0.00000325, -0.00001731, 0.00001104, -0.00000281, 0.00000638, 0.00000288, -0.00002311, 0.00001453, 0.00001971, -0.00002307, 0.00000139, 0.00001874, -0.00001352, -0.00001320, -0.00000170, -0.00000418, 0.00000183, -0.00000211, -0.00000868, 0.00000449, 0.00000946, -0.00000652, -0.00000972, 0.00000271, 0.00000160, -0.00000459, -0.00000268, -0.00000097, 0.00000269, -0.00000098, -0.00000890, -0.00000146, 0.00000119, -0.00000350, 0.00000070, -0.00000578, -0.00001202, -0.00000414, -0.00000572, -0.00000580, 0.00000238, 0.00000094, -0.00000202, -0.00000399, -0.00000345, -0.00000002, -0.00000740, -0.00000849, -0.00000187, -0.00000804, -0.00000890, -0.00001048, -0.00002098, -0.00001842, -0.00003680, -0.00006437, -0.00001646, 0.00005781, 0.00005919, },
		{ 0.00006434, 0.00003825, 0.00001711, 0.00001411, 0.00001229, 0.00001164, 0.00001228, 0.00000857, 0.00000743, 0.00001108, 0.00000889, 0.00000758, 0.00000972, 0.00000520, 0.00000356, 0.00000718, 0.00000926, 0.00001467, 0.00001347, 0.00000215, 0.00000032, 0.00001271, 0.00002231, 0.00000816, -0.00001551, -0.00000957, 0.00001955, 0.00003619, 0.00001994, -0.00001067, -0.00000711, 0.00001208, 0.00000251, -0.00001033, -0.00001827, -0.00002013, 0.00000256, 0.00000294, -0.00001824, -0.00000574, 0.00000620, 0.00000334, 0.00000758, -0.00000668, -0.00001457, 0.00000169, 0.00000009, -0.00000360, 0.00000146, -0.00000909, -0.00001057, 0.00000431, 0.00000798, 0.00000101, -0.00000885, -0.00001112, 0.00000502, 0.00001022, -0.00001116, -0.00000248, 0.00002333, -0.00001314, -0.00002497, 0.00003260, 0.00000536, -0.00003884, 0.00002284, 0.00001978, -0.00003657, 0.00000262, 0.00001916, -0.00000962, 0.00000164, -0.00000367, 0.00000063, 0.00001660, -0.00001864, -0.00000965, 0.00002838, -0.00000527, -0.00001387, 0.00001628, -0.00000158, -0.00001425, -0.00000390, -0.00000879, -0.00000223, 0.00000543, -0.00000680, -0.00000448, 0.00000974, 0.00000238, -0.00001041, -0.00000452, 0.00000168, -0.00000082, -0.00000154, -0.00000304, 0.00000026, 0.00000348, -0.00000623, -0.00000749, -0.00000013, -0.00000123, 0.00000142, -0.00000191, -0.00001201, -0.00000543, -0.00000246, -0.00000925, -0.00000382, 0.00000169, 0.00000065, -0.00000203, -0.00000457, -0.00000031, -0.00000313, -0.00000973, -0.00000433, -0.00000505, -0.00000827, -0.00000758, -0.00001757, -0.00001831, -0.00002409, -0.00005719, -0.00004301, 0.00003163, },
		{ 0.00000102, 0.00005531, 0.00004877, 0.00002386, 0.00001532, 0.00001250, 0.00000987, 0.00001178, 0.00001079, 0.00000612, 0.00000773, 0.00000811, 0.00000630, 0.00001035, 0.00000885, 0.00000431, 0.00000563, 0.00000410, 0.00000733, 0.00001647, 0.00001326, 0.00000207, -0.00000032, 0.00001253, 0.00002187, 0.00000441, -0.00001621, -0.00000890, 0.00002024, 0.00003825, 0.00001438, -0.00001410, -0.00000507, 0.00000807, 0.00000882, -0.00000403, -0.00002958, -0.00001646, 0.00001178, -0.00000476, -0.00001592, -0.00000393, 0.00000115, 0.00001194, 0.00000369, -0.00001687, -0.00000275, 0.00000592, -0.00000684, -0.00000347, -0.00000203, -0.00000801, -0.00000234, 0.00000327, 0.00000402, -0.00000060, -0.00001251, -0.00000712, 0.00001541, 0.00000347, -0.00001982, 0.00001432, 0.00002100, -0.00003320, -0.00000439, 0.00003682, -0.00002229, -0.00001858, 0.00003213, -0.00000821, -0.00002062, 0.00001209, 0.00000157, 0.00000351, 0.00000035, -0.00001159, 0.00001598, 0.00000186, -0.00002268, 0.00001537, 0.00001227, -0.00001480, 0.00000844, 0.00000565, -0.00001436, -0.00000298, -0.00000648, -0.00001177, 0.00000383, 0.00000154, -0.00000690, 0.00000359, 0.00000640, -0.00000510, -0.00000691, -0.00000227, -0.00000177, 0.00000069, -0.00000057, -0.00000284, 0.00000263, -0.00000083, -0.00000801, -0.00000414, -0.00000250, 0.00000107, 0.00000307, -0.00000922, -0.00000951, -0.00000013, -0.00000588, -0.00000876, -0.00000280, 0.00000035, 0.00000077, -0.00000296, -0.00000206, -0.00000060, -0.00000740, -0.00000666, -0.00000451, -0.00000774, -0.00000581, -0.00001271, -0.00001893, -0.00001746, -0.00004272, -0.00005471, },
		{ -0.00005314, -0.00002320, 0.00003664, 0.00005148, 0.00003076, 0.00001823, 0.00001528, 0.00001002, 0.00000887, 0.00001152, 0.00000809, 0.00000621, 0.00000659, 0.00000360, 0.00000791, 0.00001141, 0.00000624, 0.00000652, 0.00000527, 0.00000094, 0.00000828, 0.00001627, 0.00001301, 0.00000270, -0.00000097, 0.00001189, 0.00001816, 0.00000316, -0.00001335, -0.00000919, 0.00002092, 0.00003538, 0.00000966, -0.00001087, -0.00000877, 0.00000638, 0.00001902, -0.00000984, -0.00003378, -0.00000466, 0.00000766, -0.00000754, -0.00000933, -0.00001000, 0.00000484, 0.00001599, -0.00000927, -0.00001203, 0.00000834, 0.00000006, -0.00000884, -0.00000505, -0.00000494, -0.00000031, 0.00000119, -0.00000174, 0.00000208, -0.00000255, -0.00001439, 0.00000225, 0.00002064, -0.00001140, -0.00001541, 0.00003113, 0.00000051, -0.00003578, 0.00002090, 0.00001632, -0.00003004, 0.00000772, 0.00001503, -0.00001592, -0.00000012, 0.00000116, 0.00000196, 0.00001186, -0.00001134, -0.00000010, 0.00001484, -0.00001570, -0.00000088, 0.00001777, -0.00000978, 0.00000141, 0.00001283, -0.00001371, -0.00000827, 0.00000130, -0.00001307, -0.00000607, 0.00000380, -0.00000309, 0.00000050, 0.00000584, -0.00000214, -0.00000551, -0.00000259, -0.00000457, -0.00000214, 0.00000230, -0.00000106, -0.00000006, 0.00000093, -0.00000537, -0.00000516, -0.00000429, -0.00000286, 0.00000468, -0.00000238, -0.00001142, -0.00000273, -0.00000121, -0.00000779, -0.00000662, -0.00000357, -0.00000018, -0.00000023, -0.00000178, -0.00000011, -0.00000456, -0.00000708, -0.00000450, -0.00000749, -0.00000626, -0.00000804, -0.00001711, -0.00001594, -0.00002957, },
		{ -0.00002081, -0.00004505, -0.00003765, 0.00001559, 0.00004743, 0.00003541, 0.00001958, 0.00001770, 0.00001425, 0.00000789, 0.00000904, 0.00000922, 0.00000697, 0.00000718, 0.00000208, 0.00000248, 0.00001080, 0.00000863, 0.00000645, 0.00000845, 0.00000240, 0.00000122, 0.00000875, 0.00001462, 0.00001381, 0.00000312, -0.00000101, 0.00000826, 0.00001385, 0.00000679, -0.00001055, -0.00000935, 0.00001874, 0.00002865, 0.00001228, -0.00000892, -0.00001566, 0.00001248, 0.00001994, -0.00001842, -0.00002338, -0.00000205, -0.00000079, -0.00000031, -0.00001051, -0.00001270, 0.00001289, 0.00000626, -0.00001361, 0.00000199, 0.00000739, -0.00000493, -0.00000844, -0.00000967, -0.00000121, 0.00000752, -0.00000189, -0.00000657, 0.00000238, -0.00000481, -0.00001319, 0.00001449, 0.00001534, -0.00002442, 0.00000315, 0.00003132, -0.00002222, -0.00001606, 0.00002647, -0.00000828, -0.00001315, 0.00001310, -0.00000465, -0.00000196, 0.00000218, -0.00000715, 0.00001288, 0.00000375, -0.00000996, 0.00001101, -0.00000415, -0.00000880, 0.00001755, -0.00000315, -0.00000851, 0.00001728, -0.00000420, -0.00001753, 0.00000173, -0.00000514, -0.00001153, -0.00000105, -0.00000225, -0.00000008, 0.00000724, -0.00000035, -0.00000652, -0.00000134, -0.00000287, -0.00000617, -0.00000031, 0.00000164, 0.00000062, 0.00000113, -0.00000412, -0.00000511, -0.00000319, -0.00000576, 0.00000057, 0.00000261, -0.00000811, -0.00000616, -0.00000002, -0.00000456, -0.00000652, -0.00000590, -0.00000430, -0.00000110, -0.00000054, 0.00000081, -0.00000200, -0.00000659, -0.00000455, -0.00000626, -0.00000755, -0.00000607, -0.00001344, -0.00001545, },
		{ -0.00001461, -0.00001549, -0.00003487, -0.00004360, -0.00000425, 0.00003860, 0.00003920, 0.00002041, 0.00001577, 0.00001810, 0.00001187, 0.00000773, 0.00000797, 0.00000642, 0.00000860, 0.00000488, -0.00000180, 0.00000556, 0.00000974, 0.00000632, 0.00000914, 0.00000666, 0.00000147, 0.00000334, 0.00000690, 0.00001377, 0.00001447, 0.00000418, -0.00000225, 0.00000151, 0.00001372, 0.00001228, -0.00000853, -0.00000964, 0.00001076, 0.00002567, 0.00001906, -0.00001311, -0.00001582, 0.00001680, 0.00000968, -0.00001364, -0.00001212, -0.00001162, 0.00000053, 0.00000492, -0.00001594, -0.00000550, 0.00001007, -0.00000560, -0.00000247, 0.00000837, 0.00000056, -0.00000449, -0.00001144, -0.00001116, 0.00000738, 0.00000977, -0.00000895, -0.00000848, 0.00000493, -0.00000855, -0.00000715, 0.00002412, -0.00000007, -0.00002355, 0.00002185, 0.00001270, -0.00002604, 0.00000582, 0.00000977, -0.00001221, 0.00000430, -0.00000172, -0.00000422, 0.00000957, -0.00000764, -0.00000023, 0.00001342, -0.00000451, 0.00000408, 0.00000208, -0.00001311, 0.00001407, 0.00000910, -0.00001653, 0.00000967, 0.00001026, -0.00001823, -0.00000681, 0.00000044, -0.00000946, -0.00000417, -0.00000412, -0.00000448, 0.00000785, 0.00000539, -0.00000716, -0.00000407, 0.00000029, -0.00000516, -0.00000467, -0.00000011, 0.00000164, 0.00000321, -0.00000222, -0.00000661, -0.00000249, -0.00000426, -0.00000389, 0.00000195, -0.00000395, -0.00000694, -0.00000047, -0.00000232, -0.00000542, -0.00000493, -0.00000619, -0.00000488, -0.00000216, 0.00000114, 0.00000100, -0.00000473, -0.00000508, -0.00000484, -0.00000753, -0.00000611, -0.00001041, },
		{ -0.00000829, -0.00001375, -0.00001307, -0.00002502, -0.00004232, -0.00002058, 0.00002505, 0.00004120, 0.00002459, 0.00001200, 0.00001624, 0.00001641, 0.00001053, 0.00000831, 0.00000465, 0.00000650, 0.00000905, -0.00000006, -0.00000038, 0.00000715, 0.00000587, 0.00000807, 0.00000924, 0.00000391, 0.00000416, 0.00000358, 0.00000488, 0.00001340, 0.00001496, 0.00000577, -0.00000734, -0.00000353, 0.00001777, 0.00001626, -0.00000560, -0.00001416, 0.00000143, 0.00003002, 0.00002088, -0.00001585, -0.00000999, 0.00000796, 0.00000429, 0.00000225, -0.00001453, -0.00001809, 0.00000786, 0.00000099, -0.00001269, -0.00000050, -0.00000413, -0.00000298, 0.00001032, 0.00000330, -0.00000177, -0.00000423, -0.00001570, -0.00000634, 0.00001482, 0.00000559, -0.00001690, -0.00000445, 0.00000658, -0.00001335, 0.00000479, 0.00002360, -0.00001517, -0.00000801, 0.00002381, -0.00000629, -0.00001050, 0.00000807, -0.00000580, 0.00000175, 0.00000163, -0.00001164, 0.00000913, 0.00000480, -0.00000891, 0.00000824, 0.00000306, 0.00000259, 0.00000801, -0.00001434, 0.00000239, 0.00002151, -0.00001122, -0.00000600, 0.00001570, -0.00000873, -0.00001388, -0.00000046, -0.00000698, -0.00000450, -0.00000215, -0.00000942, 0.00000147, 0.00001061, -0.00000196, -0.00000745, -0.00000077, -0.00000185, -0.00000498, -0.00000322, -0.00000088, 0.00000391, 0.00000222, -0.00000631, -0.00000480, -0.00000219, -0.00000438, -0.00000065, -0.00000247, -0.00000715, -0.00000107, 0.00000019, -0.00000474, -0.00000419, -0.00000461, -0.00000654, -0.00000574, -0.00000166, 0.00000226, -0.00000123, -0.00000455, -0.00000429, -0.00000664, -0.00000621, },
		{ -0.00000626, -0.00000652, -0.00001228, -0.00001274, -0.00001785, -0.00003616, -0.00003060, 0.00000874, 0.00003782, 0.00003135, 0.00001237, 0.00001066, 0.00001613, 0.00001352, 0.00001157, 0.00000651, 0.00000244, 0.00000834, 0.00000502, -0.00000147, 0.00000348, 0.00000376, 0.00000509, 0.00001030, 0.00000572, 0.00000474, 0.00000677, 0.00000220, 0.00000406, 0.00001234, 0.00001657, 0.00000459, -0.00001394, -0.00000301, 0.00002132, 0.00001988, -0.00000387, -0.00002360, -0.00000121, 0.00003423, 0.00001805, -0.00000947, -0.00001195, -0.00000672, 0.00001466, 0.00001062, -0.00002362, -0.00001336, 0.00000751, -0.00000137, -0.00000262, -0.00000966, -0.00001194, 0.00001101, 0.00001118, -0.00000268, -0.00000006, -0.00000710, -0.00001602, 0.00000147, 0.00001790, -0.00000456, -0.00002046, 0.00000539, 0.00000194, -0.00001367, 0.00001734, 0.00001176, -0.00001720, 0.00000709, 0.00000990, -0.00000742, 0.00000272, -0.00000498, -0.00000238, 0.00000918, -0.00001174, -0.00000359, 0.00001337, -0.00000337, 0.00000063, 0.00000402, 0.00000034, 0.00001364, -0.00000559, -0.00001210, 0.00002107, 0.00000513, -0.00001447, 0.00000910, 0.00000067, -0.00001525, -0.00000278, -0.00000554, -0.00000731, 0.00000143, -0.00000725, -0.00000748, 0.00000789, 0.00000494, -0.00000577, -0.00000351, -0.00000135, -0.00000334, -0.00000318, -0.00000362, 0.00000039, 0.00000488, -0.00000211, -0.00000646, -0.00000287, -0.00000333, -0.00000162, -0.00000164, -0.00000795, -0.00000379, 0.00000251, -0.00000227, -0.00000466, -0.00000324, -0.00000498, -0.00000706, -0.00000571, -0.00000040, 0.00000088, -0.00000220, -0.00000333, -0.00000576, },
		{ -0.00000403, -0.00000621, -0.00000558, -0.00001009, -0.00001262, -0.00001385, -0.00002883, -0.00003411, -0.00000597, 0.00002854, 0.00003568, 0.00001784, 0.00000723, 0.00001231, 0.00001300, 0.00001293, 0.00001201, 0.00000320, 0.00000386, 0.00000655, 0.00000064, 0.00000276, 0.00000300, 0.00000014, 0.00000802, 0.00000822, 0.00000361, 0.00000799, 0.00000677, 0.00000184, 0.00000300, 0.00001187, 0.00001747, -0.00000067, -0.00001597, 0.00000009, 0.00002274, 0.00002418, -0.00000768, -0.00003040, 0.00000180, 0.00003120, 0.00002102, -0.00000203, -0.00002662, -0.00000900, 0.00002846, 0.00000495, -0.00002320, -0.00000886, 0.00000067, 0.00000701, -0.00000184, -0.00002351, -0.00000341, 0.00001918, 0.00000403, -0.00000317, 0.00000034, -0.00000933, -0.00001264, 0.00000920, 0.00001452, -0.00001771, -0.00001330, 0.00001306, -0.00000784, -0.00000526, 0.00002081, -0.00000066, -0.00000672, 0.00000732, -0.00000102, 0.00000410, -0.00000129, -0.00001075, 0.00000935, -0.00000037, -0.00001350, 0.00000798, 0.00000519, -0.00000002, 0.00000474, -0.00000405, 0.00001036, 0.00000919, -0.00001524, 0.00000814, 0.00001756, -0.00001038, 0.00000014, 0.00000515, -0.00001485, -0.00000604, -0.00000160, -0.00001059, -0.00000061, -0.00000055, -0.00000987, 0.00000024, 0.00000638, -0.00000183, -0.00000336, -0.00000176, -0.00000350, -0.00000227, -0.00000288, -0.00000332, 0.00000262, 0.00000173, -0.00000475, -0.00000393, -0.00000329, -0.00000232, -0.00000010, -0.00000675, -0.00000809, 0.00000129, 0.00000149, -0.00000365, -0.00000336, -0.00000333, -0.00000564, -0.00000717, -0.00000444, -0.00000077, -0.00000077, -0.00000138, },
		{ -0.00000077, -0.00000165, -0.00000492, -0.00000507, -0.00000818, -0.00001205, -0.00001159, -0.00002222, -0.00003351, -0.00001673, 0.00001661, 0.00003487, 0.00002470, 0.00000793, 0.00000885, 0.00001130, 0.00001000, 0.00001429, 0.00000930, 0.00000296, 0.00000482, 0.00000068, 0.00000227, 0.00000617, -0.00000156, 0.00000148, 0.00000869, 0.00000417, 0.00000567, 0.00000924, 0.00000636, 0.00000226, 0.00000161, 0.00001256, 0.00001381, -0.00000487, -0.00001229, 0.00000103, 0.00002463, 0.00002496, -0.00001424, -0.00002863, 0.00000065, 0.00002709, 0.00003162, -0.00000716, -0.00003843, 0.00000234, 0.00002823, 0.00000013, -0.00001443, -0.00001431, 0.00000137, 0.00001551, -0.00001409, -0.00002392, 0.00000954, 0.00001521, -0.00000074, -0.00000178, 0.00000022, -0.00000938, -0.00000777, 0.00001534, 0.00000219, -0.00002459, 0.00000254, 0.00001001, -0.00001324, 0.00000588, 0.00001369, -0.00000206, 0.00000143, -0.00000235, 0.00000295, 0.00000821, -0.00001148, -0.00000141, 0.00000879, -0.00001074, -0.00000181, 0.00000682, 0.00000077, 0.00000817, -0.00000200, -0.00000044, 0.00001551, -0.00000532, -0.00000339, 0.00001936, -0.00000123, -0.00000554, 0.00000758, -0.00001232, -0.00001337, 0.00000150, -0.00000811, -0.00000663, 0.00000211, -0.00000607, -0.00000438, 0.00000387, -0.00000078, -0.00000278, -0.00000014, -0.00000324, -0.00000346, -0.00000139, -0.00000340, -0.00000090, 0.00000179, -0.00000263, -0.00000349, -0.00000270, -0.00000352, -0.00000004, -0.00000324, -0.00000989, -0.00000342, 0.00000273, -0.00000095, -0.00000292, -0.00000262, -0.00000412, -0.00000614, -0.00000610, -0.00000380, -0.00000227, },
		{ -0.00000450, -0.00000265, -0.00000102, -0.00000301, -0.00000373, -0.00000621, -0.00001149, -0.00001062, -0.00001632, -0.00003032, -0.00002382, 0.00000460, 0.00002952, 0.00003015, 0.00001162, 0.00000622, 0.00001101, 0.00000745, 0.00001040, 0.00001322, 0.00000687, 0.00000570, 0.00000041, -0.00000212, 0.00000774, 0.00000353, -0.00000308, 0.00000446, 0.00000472, 0.00000381, 0.00000810, 0.00000824, 0.00000735, 0.00000186, 0.00000193, 0.00001064, 0.00000705, -0.00000259, -0.00000832, 0.00000079, 0.00002607, 0.00001945, -0.00001447, -0.00002443, -0.00000686, 0.00003154, 0.00003606, -0.00002033, -0.00003302, 0.00000823, 0.00001816, 0.00000731, -0.00001212, -0.00002017, 0.00001066, 0.00000987, -0.00002254, -0.00001104, 0.00001231, 0.00000790, -0.00000219, -0.00000033, 0.00000209, -0.00000996, -0.00000031, 0.00001479, -0.00001480, -0.00001686, 0.00001437, -0.00000040, -0.00000877, 0.00000935, 0.00000663, 0.00000454, -0.00000211, -0.00000519, 0.00001127, -0.00000060, -0.00000979, 0.00000676, -0.00000323, -0.00000633, 0.00000550, -0.00000090, 0.00000694, 0.00000705, -0.00000551, 0.00000960, 0.00000496, -0.00000649, 0.00001602, 0.00000813, -0.00000903, 0.00000700, -0.00000428, -0.00002040, -0.00000299, -0.00000216, -0.00000920, -0.00000024, -0.00000268, -0.00000536, 0.00000261, -0.00000055, -0.00000506, 0.00000022, -0.00000034, -0.00000406, -0.00000208, -0.00000213, -0.00000201, 0.00000055, -0.00000236, -0.00000378, -0.00000129, -0.00000297, -0.00000170, -0.00000077, -0.00000795, -0.00000749, 0.00000051, 0.00000066, -0.00000176, -0.00000159, -0.00000293, -0.00000503, -0.00000563, -0.00000494, },
		{ -0.00000455, -0.00000496, -0.00000468, -0.00000264, -0.00000290, -0.00000252, -0.00000322, -0.00000993, -0.00001111, -0.00001205, -0.00002509, -0.00002715, -0.00000610, 0.00002049, 0.00003245, 0.00001804, 0.00000432, 0.00000979, 0.00000875, 0.00000584, 0.00001122, 0.00000935, 0.00000880, 0.00000504, -0.00000567, 0.00000202, 0.00000872, -0.00000045, 0.00000047, 0.00000279, 0.00000151, 0.00000675, 0.00000750, 0.00000810, 0.00000799, 0.00000214, 0.00000261, 0.00000336, 0.00000435, 0.00000363, -0.00000707, 0.00000196, 0.00002169, 0.00001432, -0.00000692, -0.00002690, -0.00001023, 0.00003872, 0.00002607, -0.00002197, -0.00002032, -0.00000154, 0.00001596, 0.00001632, -0.00001784, -0.00001481, 0.00001245, -0.00000313, -0.00001509, -0.00000113, 0.00000700, 0.00000315, -0.00000364, 0.00000473, 0.00000243, -0.00000987, 0.00000833, 0.00000298, -0.00002341, 0.00000013, 0.00001310, -0.00000616, -0.00000157, 0.00000499, 0.00000799, 0.00000622, -0.00000849, 0.00000252, 0.00000872, -0.00000706, 0.00000107, 0.00000067, -0.00000812, 0.00000489, 0.00000065, -0.00000044, 0.00001164, 0.00000003, 0.00000245, 0.00000838, -0.00000589, 0.00001065, 0.00001815, -0.00000755, 0.00000019, 0.00000477, -0.00001919, -0.00001274, -0.00000054, -0.00000762, -0.00000277, -0.00000098, -0.00000634, 0.00000155, 0.00000259, -0.00000657, -0.00000363, 0.00000137, -0.00000188, -0.00000272, -0.00000210, -0.00000217, 0.00000043, -0.00000181, -0.00000559, -0.00000197, -0.00000079, -0.00000213, -0.00000049, -0.00000508, -0.00000880, -0.00000253, 0.00000052, -0.00000154, -0.00000087, -0.00000111, -0.00000387, -0.00000511, },
		{ -0.00000433, -0.00000424, -0.00000436, -0.00000499, -0.00000395, -0.00000442, -0.00000374, -0.00000095, -0.00000622, -0.00001122, -0.00001028, -0.00001967, -0.00002697, -0.00001354, 0.00000966, 0.00002930, 0.00002564, 0.00000602, 0.00000591, 0.00001066, 0.00000526, 0.00000767, 0.00000817, 0.00000819, 0.00001140, -0.00000107, -0.00000531, 0.00000641, 0.00000403, 0.00000118, 0.00000249, -0.00000202, 0.00000312, 0.00000787, 0.00000605, 0.00000880, 0.00000781, 0.00000464, -0.00000133, -0.00000428, 0.00000843, 0.00000755, -0.00000531, 0.00000094, 0.00001220, 0.00001767, -0.00000157, -0.00003198, -0.00000464, 0.00003335, 0.00001565, -0.00000700, -0.00002020, -0.00001320, 0.00002357, 0.00001492, -0.00001720, -0.00000677, 0.00000029, -0.00000498, -0.00000185, -0.00000121, 0.00000265, -0.00000246, -0.00000162, 0.00001198, -0.00000195, -0.00000445, 0.00000983, -0.00001319, -0.00001592, 0.00001093, 0.00000507, -0.00000311, -0.00000069, 0.00000257, 0.00001196, 0.00000012, -0.00000557, 0.00000722, -0.00000093, -0.00000074, 0.00000367, -0.00000955, 0.00000046, 0.00000670, -0.00000425, 0.00000660, 0.00000671, 0.00000124, 0.00000926, -0.00000392, 0.00000139, 0.00002424, 0.00000330, -0.00000752, 0.00000644, -0.00001094, -0.00001921, -0.00000439, -0.00000650, -0.00000495, 0.00000107, -0.00000618, -0.00000240, 0.00000558, -0.00000313, -0.00000755, -0.00000127, -0.00000025, -0.00000151, -0.00000167, -0.00000283, -0.00000010, 0.00000023, -0.00000587, -0.00000521, -0.00000048, -0.00000077, -0.00000027, -0.00000281, -0.00000851, -0.00000506, 0.00000009, -0.00000175, -0.00000185, 0.00000009, -0.00000178, },
		{ -0.00000015, -0.00000262, -0.00000360, -0.00000385, -0.00000467, -0.00000377, -0.00000479, -0.00000650, -0.00000210, -0.00000249, -0.00000914, -0.00000937, -0.00001542, -0.00002539, -0.00001744, 0.00000087, 0.00002097, 0.00002975, 0.00001253, 0.00000262, 0.00000971, 0.00000675, 0.00000593, 0.00000756, 0.00000421, 0.00001123, 0.00000801, -0.00000496, 0.00000063, 0.00000376, 0.00000174, 0.00000587, -0.00000109, -0.00000325, 0.00000590, 0.00000623, 0.00000619, 0.00000816, 0.00000954, 0.00000545, -0.00000974, -0.00000503, 0.00001339, 0.00000922, -0.00000227, -0.00000695, 0.00000719, 0.00002529, -0.00000354, -0.00002841, -0.00000285, 0.00001654, 0.00002227, 0.00000589, -0.00003197, -0.00001256, 0.00002577, 0.00001079, -0.00000643, -0.00001167, -0.00001008, 0.00000589, 0.00000263, -0.00000275, -0.00000168, -0.00000863, 0.00000788, 0.00001348, -0.00000556, 0.00000338, -0.00000020, -0.00001898, -0.00000157, 0.00000903, 0.00000220, 0.00000075, -0.00000382, 0.00000579, 0.00000924, -0.00000282, 0.00000257, 0.00000092, -0.00000230, 0.00000753, -0.00000553, -0.00000801, 0.00000854, 0.00000028, 0.00000015, 0.00000745, 0.00000191, 0.00001004, 0.00000267, -0.00000753, 0.00001920, 0.00001802, -0.00000647, 0.00000156, -0.00000364, -0.00001991, -0.00000920, -0.00000632, -0.00000874, 0.00000121, -0.00000247, -0.00000640, 0.00000393, 0.00000221, -0.00000679, -0.00000487, -0.00000180, -0.00000123, -0.00000009, -0.00000219, -0.00000173, 0.00000141, -0.00000336, -0.00000748, -0.00000311, -0.00000042, 0.00000042, -0.00000038, -0.00000690, -0.00000742, -0.00000086, -0.00000087, -0.00000324, -0.00000090, },
		{ -0.00000280, -0.00000047, -0.00000134, -0.00000233, -0.00000277, -0.00000439, -0.00000361, -0.00000332, -0.00000741, -0.00000564, -0.00000184, -0.00000652, -0.00000783, -0.00001111, -0.00002338, -0.00002042, -0.00000428, 0.00001194, 0.00002761, 0.00002011, 0.00000311, 0.00000711, 0.00000798, 0.00000438, 0.00000846, 0.00000343, 0.00000508, 0.00001149, 0.00000127, -0.00000060, 0.00000252, -0.00000182, 0.00000580, 0.00000584, -0.00000488, -0.00000017, 0.00000538, 0.00000542, 0.00000601, 0.00000766, 0.00001268, 0.00000010, -0.00001465, -0.00000035, 0.00001442, 0.00001270, -0.00000388, -0.00001652, 0.00001103, 0.00002639, -0.00000331, -0.00001911, -0.00001626, 0.00000844, 0.00003921, 0.00000253, -0.00003701, -0.00000767, 0.00001815, 0.00001732, -0.00000131, -0.00002474, -0.00000512, 0.00001303, 0.00000115, -0.00000105, -0.00001090, -0.00000714, 0.00001827, 0.00000783, -0.00000245, 0.00000355, -0.00001189, -0.00001024, 0.00000462, 0.00000340, 0.00000546, -0.00000045, -0.00000322, 0.00000709, 0.00000315, 0.00000320, 0.00000294, -0.00000591, 0.00000601, 0.00000411, -0.00001136, 0.00000219, 0.00000576, -0.00000096, 0.00000584, 0.00000151, 0.00000690, 0.00001134, -0.00000732, 0.00000625, 0.00002486, 0.00000342, -0.00000181, 0.00000005, -0.00001794, -0.00001392, -0.00000514, -0.00001146, -0.00000352, 0.00000136, -0.00000614, -0.00000040, 0.00000444, -0.00000320, -0.00000548, -0.00000368, -0.00000340, -0.00000029, -0.00000007, -0.00000213, 0.00000067, -0.00000055, -0.00000681, -0.00000590, -0.00000202, -0.00000018, 0.00000140, -0.00000346, -0.00000837, -0.00000326, 0.00000018, -0.00000299, },
		{ -0.00000169, -0.00000366, -0.00000170, -0.00000143, -0.00000189, -0.00000126, -0.00000312, -0.00000391, -0.00000207, -0.00000571, -0.00000800, -0.00000364, -0.00000569, -0.00000733, -0.00000645, -0.00001889, -0.00002347, -0.00000850, 0.00000551, 0.00002159, 0.00002500, 0.00000662, 0.00000438, 0.00000959, 0.00000312, 0.00000683, 0.00000675, 0.00000130, 0.00000811, 0.00000486, 0.00000085, 0.00000565, -0.00000283, -0.00000149, 0.00000931, 0.00000125, -0.00000328, 0.00000146, 0.00000355, 0.00000592, 0.00000436, 0.00000999, 0.00001129, -0.00000716, -0.00001124, 0.00000212, 0.00001487, 0.00001559, -0.00001196, -0.00001771, 0.00001457, 0.00002213, 0.00000677, -0.00002030, -0.00003325, 0.00001845, 0.00004465, -0.00000548, -0.00002994, -0.00001277, 0.00001510, 0.00002786, -0.00001075, -0.00002629, 0.00000521, 0.00000912, 0.00000442, -0.00000333, -0.00001932, 0.00000372, 0.00001976, 0.00000377, 0.00000213, -0.00000556, -0.00001185, 0.00000074, 0.00000083, 0.00000331, 0.00000705, -0.00000214, 0.00000004, 0.00000209, 0.00000383, 0.00000874, -0.00000440, -0.00000209, 0.00000950, -0.00000548, -0.00000473, 0.00000603, 0.00000025, 0.00000507, 0.00000348, 0.00000077, 0.00001350, 0.00000165, -0.00000290, 0.00002100, 0.00001355, -0.00000071, 0.00000271, -0.00001351, -0.00001944, -0.00000550, -0.00000963, -0.00000970, 0.00000046, -0.00000343, -0.00000316, 0.00000386, -0.00000045, -0.00000446, -0.00000324, -0.00000483, -0.00000321, 0.00000030, -0.00000087, 0.00000003, 0.00000098, -0.00000468, -0.00000705, -0.00000373, -0.00000190, 0.00000072, -0.00000012, -0.00000648, -0.00000561, -0.00000024, },
		{ -0.00000141, -0.00000064, -0.00000364, -0.00000263, -0.00000175, -0.00000263, -0.00000112, -0.00000120, -0.00000330, -0.00000171, -0.00000338, -0.00000814, -0.00000527, -0.00000554, -0.00000912, -0.00000463, -0.00001165, -0.00002359, -0.00001355, 0.00000060, 0.00001445, 0.00002668, 0.00001227, 0.00000128, 0.00001023, 0.00000537, 0.00000274, 0.00000825, 0.00000260, 0.00000468, 0.00000503, -0.00000102, 0.00000787, 0.00000433, -0.00000674, 0.00000354, 0.00000624, -0.00000020, 0.00000047, -0.00000028, 0.00000367, 0.00000486, 0.00000456, 0.00001162, 0.00000367, -0.00000733, -0.00000536, 0.00000098, 0.00001757, 0.00001135, -0.00001729, -0.00001186, 0.00000980, 0.00002461, 0.00001582, -0.00003488, -0.00003349, 0.00002944, 0.00003508, -0.00000126, -0.00002588, -0.00002219, 0.00002320, 0.00002476, -0.00001922, -0.00001296, 0.00000402, 0.00000539, 0.00001025, -0.00001256, -0.00001672, 0.00001306, 0.00001427, 0.00000585, -0.00000113, -0.00001197, -0.00000136, 0.00000238, -0.00000280, 0.00000619, 0.00000511, 0.00000019, -0.00000127, -0.00000175, 0.00001032, 0.00000499, -0.00000650, 0.00000572, 0.00000180, -0.00000615, 0.00000417, 0.00000098, 0.00000244, 0.00000766, -0.00000097, 0.00000781, 0.00000941, -0.00000330, 0.00001357, 0.00001947, 0.00000217, 0.00000380, -0.00000563, -0.00002213, -0.00001057, -0.00000610, -0.00001198, -0.00000385, -0.00000235, -0.00000455, 0.00000272, 0.00000172, -0.00000398, -0.00000251, -0.00000345, -0.00000533, -0.00000207, -0.00000067, -0.00000001, 0.00000208, -0.00000213, -0.00000711, -0.00000487, -0.00000282, -0.00000147, 0.00000050, -0.00000345, -0.00000600, },
		{ -0.00000527, -0.00000218, 0.00000050, -0.00000301, -0.00000370, -0.00000180, -0.00000293, -0.00000245, -0.00000065, -0.00000202, -0.00000115, -0.00000109, -0.00000713, -0.00000656, -0.00000423, -0.00001031, -0.00000745, -0.00000610, -0.00001933, -0.00001712, -0.00000367, 0.00000662, 0.00002417, 0.00002016, 0.00000028, 0.00000651, 0.00000994, 0.00000146, 0.00000595, 0.00000438, 0.00000334, 0.00000655, -0.00000331, 0.00000223, 0.00001174, -0.00000176, -0.00000302, 0.00000411, 0.00000157, 0.00000367, 0.00000069, -0.00000135, 0.00000384, 0.00000331, 0.00000745, 0.00000634, -0.00000194, -0.00000012, -0.00000394, 0.00000160, 0.00001682, 0.00000276, -0.00001245, -0.00001070, 0.00000440, 0.00003450, 0.00000934, -0.00004329, -0.00001936, 0.00002342, 0.00002921, 0.00000977, -0.00003243, -0.00002028, 0.00002674, 0.00001019, -0.00000977, -0.00000327, -0.00000609, 0.00000997, 0.00000830, -0.00001819, -0.00000569, 0.00001247, 0.00001151, 0.00000672, -0.00000947, -0.00000655, 0.00000626, -0.00000207, -0.00000127, 0.00000595, 0.00000501, 0.00000205, -0.00000629, 0.00000203, 0.00001158, -0.00000121, 0.00000077, 0.00000429, -0.00000532, 0.00000271, 0.00000384, -0.00000160, 0.00000787, 0.00000340, 0.00000231, 0.00001064, 0.00000008, 0.00000731, 0.00002270, 0.00000710, 0.00000145, 0.00000130, -0.00001809, -0.00001704, -0.00000588, -0.00001076, -0.00000746, -0.00000281, -0.00000622, 0.00000011, 0.00000402, -0.00000263, -0.00000333, -0.00000176, -0.00000454, -0.00000410, -0.00000226, -0.00000122, 0.00000237, 0.00000103, -0.00000588, -0.00000618, -0.00000296, -0.00000261, -0.00000100, -0.00000198, },
		{ -0.00000203, -0.00000501, -0.00000341, 0.00000157, -0.00000082, -0.00000443, -0.00000267, -0.00000248, -0.00000335, -0.00000157, -0.00000176, -0.00000093, 0.00000147, -0.00000450, -0.00000818, -0.00000337, -0.00000816, -0.00001120, -0.00000556, -0.00001417, -0.00001799, -0.00000574, 0.00000026, 0.00001584, 0.00002649, 0.00000564, 0.00000024, 0.00001115, 0.00000428, 0.00000373, 0.00000498, 0.00000118, 0.00000799, 0.00000043, -0.00000596, 0.00000936, 0.00000678, -0.00000200, 0.00000169, -0.00000139, 0.00000260, 0.00000659, -0.00000111, -0.00000046, 0.00000187, 0.00000521, 0.00000658, -0.00000308, 0.00000175, 0.00000557, -0.00000427, 0.00000367, 0.00000799, 0.00000114, -0.00000479, -0.00001664, 0.00000893, 0.00003567, -0.00000491, -0.00003149, -0.00001356, 0.00000878, 0.00003643, 0.00001111, -0.00003556, -0.00001078, 0.00001284, 0.00000600, 0.00000899, -0.00000870, -0.00000987, 0.00001414, 0.00000022, -0.00001175, 0.00000096, 0.00000745, 0.00001260, 0.00000043, -0.00001082, 0.00000403, 0.00000488, -0.00000345, 0.00000053, 0.00000425, 0.00000750, -0.00000119, -0.00000651, 0.00000702, 0.00000519, 0.00000116, 0.00000583, -0.00000442, -0.00000136, 0.00000797, -0.00000110, 0.00000302, 0.00000712, 0.00000172, 0.00000896, 0.00000344, 0.00000155, 0.00002231, 0.00001612, -0.00000015, 0.00000201, -0.00001080, -0.00001953, -0.00000853, -0.00000930, -0.00000985, -0.00000320, -0.00000671, -0.00000454, 0.00000397, 0.00000049, -0.00000372, -0.00000182, -0.00000295, -0.00000403, -0.00000322, -0.00000334, 0.00000030, 0.00000321, -0.00000269, -0.00000685, -0.00000371, -0.00000244, -0.00000221, },
		},
		{ { -0.00000155, -0.00000172, -0.00000247, -0.00000379, -0.00000457, -0.00000364, -0.00000153, -0.00000013, -0.00000024, -0.00000069, -0.00000055, -0.00000057, -0.00000135, -0.00000162, -0.00000034, 0.00000106, 0.00000082, -0.00000017, -0.00000024, -0.00000044, -0.00000263, -0.00000531, -0.00000559, -0.00000460, -0.00000614, -0.00000990, -0.00001129, -0.00000874, -0.00000565, -0.00000368, -0.00000042, 0.00000431, 0.00000612, 0.00000356, 0.00000123, 0.00000162, 0.00000060, -0.00000375, -0.00000538, -0.00000052, 0.00000330, -0.00000207, -0.00001166, -0.00001488, -0.00001182, -0.00001009, -0.00000939, -0.00000332, 0.00000544, 0.00000773, 0.00000536, 0.00000830, 0.00001291, 0.00000395, -0.00001548, -0.00001996, -0.00000141, 0.00001186, -0.00000216, -0.00001909, -0.00000393, 0.00002756, 0.00002559, -0.00001558, -0.00004140, -0.00001558, 0.00002559, 0.00002756, -0.00000393, -0.00001909, -0.00000216, 0.00001186, -0.00000141, -0.00001996, -0.00001548, 0.00000395, 0.00001291, 0.00000830, 0.00000536, 0.00000773, 0.00000544, -0.00000332, -0.00000939, -0.00001009, -0.00001182, -0.00001488, -0.00001166, -0.00000207, 0.00000330, -0.00000052, -0.00000538, -0.00000375, 0.00000060, 0.00000162, 0.00000123, 0.00000356, 0.00000612, 0.00000431, -0.00000042, -0.00000368, -0.00000565, -0.00000874, -0.00001129, -0.00000990, -0.00000614, -0.00000460, -0.00000559, -0.00000531, -0.00000263, -0.00000044, -0.00000024, -0.00000017, 0.00000082, 0.00000106, -0.00000034, -0.00000162, -0.00000135, -0.00000057, -0.00000055, -0.00000069, -0.00000024, -0.00000013, -0.00000153, -0.00000364, -0.00000457, -0.00000379, -0.00000247, -0.00000172, },
		{ -0.00000276, -0.00000210, -0.00000195, -0.00000196, -0.00000234, -0.00000339, -0.00000433, -0.00000375, -0.00000185, -0.00000060, -0.00000108, -0.00000197, -0.00000197, -0.00000182, -0.00000237, -0.00000233, -0.00000036, 0.00000196, 0.00000229, 0.00000130, 0.00000093, 0.00000040, -0.00000215, -0.00000496, -0.00000455, -0.00000202, -0.00000209, -0.00000549, -0.00000797, -0.00000777, -0.00000795, -0.00000941, -0.00000826, -0.00000321, 0.00000113, 0.00000203, 0.00000294, 0.00000553, 0.00000490, -0.00000110, -0.00000507, -0.00000104, 0.00000455, 0.00000262, -0.00000371, -0.00000543, -0.00000388, -0.00000735, -0.00001452, -0.00001609, -0.00001162, -0.00000871, -0.00000554, 0.00000686, 0.00002218, 0.00002067, 0.00000165, -0.00000923, 0.00000030, 0.00000607, -0.00001217, -0.00002955, -0.00001037, 0.00002879, 0.00003395, -0.00000558, -0.00003602, -0.00001688, 0.00001994, 0.00002065, -0.00001032, -0.00002299, -0.00000051, 0.00002043, 0.00001061, -0.00001144, -0.00001571, -0.00000528, -0.00000166, -0.00000679, -0.00000562, 0.00000475, 0.00001211, 0.00001098, 0.00000755, 0.00000409, -0.00000406, -0.00001483, -0.00001810, -0.00001155, -0.00000515, -0.00000550, -0.00000707, -0.00000387, 0.00000040, 0.00000025, -0.00000175, -0.00000035, 0.00000319, 0.00000436, 0.00000327, 0.00000269, 0.00000170, -0.00000225, -0.00000734, -0.00000946, -0.00000875, -0.00000874, -0.00000976, -0.00000868, -0.00000483, -0.00000137, -0.00000029, -0.00000019, 0.00000024, 0.00000001, -0.00000152, -0.00000276, -0.00000232, -0.00000100, -0.00000016, 0.00000034, 0.00000109, 0.00000137, 0.00000010, -0.00000222, -0.00000381, -0.00000372, },
		{ -0.00000168, -0.00000194, -0.00000147, -0.00000143, -0.00000201, -0.00000253, -0.00000293, -0.00000375, -0.00000458, -0.00000396, -0.00000182, -0.00000031, -0.00000085, -0.00000205, -0.00000227, -0.00000231, -0.00000335, -0.00000394, -0.00000223, 0.00000046, 0.00000153, 0.00000143, 0.00000214, 0.00000253, 0.00000010, -0.00000353, -0.00000411, -0.00000190, -0.00000148, -0.00000383, -0.00000486, -0.00000335, -0.00000347, -0.00000691, -0.00000930, -0.00000794, -0.00000628, -0.00000633, -0.00000400, 0.00000226, 0.00000598, 0.00000282, -0.00000077, 0.00000199, 0.00000598, 0.00000302, -0.00000294, -0.00000225, 0.00000305, 0.00000238, -0.00000513, -0.00001086, -0.00001363, -0.00001932, -0.00002311, -0.00001233, 0.00000841, 0.00001711, 0.00000867, 0.00000548, 0.00001820, 0.00002120, -0.00000618, -0.00003538, -0.00002346, 0.00001742, 0.00003017, -0.00000292, -0.00003133, -0.00001218, 0.00002498, 0.00002455, -0.00001149, -0.00003082, -0.00001080, 0.00001457, 0.00001273, -0.00000317, -0.00000562, 0.00000279, 0.00000136, -0.00001118, -0.00001748, -0.00001085, -0.00000117, 0.00000512, 0.00001112, 0.00001585, 0.00001208, 0.00000038, -0.00000860, -0.00000976, -0.00000974, -0.00001342, -0.00001532, -0.00001013, -0.00000260, 0.00000012, -0.00000101, -0.00000091, 0.00000062, 0.00000070, 0.00000013, 0.00000157, 0.00000369, 0.00000296, -0.00000021, -0.00000262, -0.00000408, -0.00000703, -0.00001098, -0.00001222, -0.00000944, -0.00000551, -0.00000298, -0.00000144, -0.00000008, 0.00000018, -0.00000127, -0.00000299, -0.00000331, -0.00000248, -0.00000159, -0.00000075, 0.00000048, 0.00000157, 0.00000134, -0.00000021, },
		{ -0.00000008, -0.00000050, -0.00000088, -0.00000021, 0.00000096, 0.00000114, -0.00000003, -0.00000149, -0.00000265, -0.00000402, -0.00000531, -0.00000491, -0.00000254, -0.00000057, -0.00000087, -0.00000202, -0.00000202, -0.00000174, -0.00000280, -0.00000394, -0.00000295, -0.00000094, -0.00000054, -0.00000100, 0.00000024, 0.00000211, 0.00000130, -0.00000134, -0.00000166, 0.00000024, -0.00000023, -0.00000372, -0.00000516, -0.00000263, -0.00000082, -0.00000265, -0.00000457, -0.00000426, -0.00000541, -0.00000962, -0.00001119, -0.00000660, -0.00000175, -0.00000192, -0.00000183, 0.00000444, 0.00001066, 0.00000764, -0.00000030, -0.00000160, 0.00000350, 0.00000480, 0.00000067, -0.00000160, -0.00000302, -0.00001177, -0.00002302, -0.00002112, -0.00000720, -0.00000168, -0.00000874, -0.00000428, 0.00002208, 0.00003855, 0.00001551, -0.00002115, -0.00002135, 0.00001207, 0.00002379, -0.00000883, -0.00003794, -0.00001750, 0.00002570, 0.00003237, -0.00000212, -0.00002692, -0.00001433, 0.00000691, 0.00000503, -0.00000854, -0.00000688, 0.00000734, 0.00001114, -0.00000049, -0.00001166, -0.00001337, -0.00001147, -0.00000942, -0.00000272, 0.00000774, 0.00001289, 0.00000978, 0.00000577, 0.00000475, 0.00000093, -0.00000898, -0.00001759, -0.00001729, -0.00001107, -0.00000631, -0.00000410, -0.00000144, 0.00000107, 0.00000065, -0.00000145, -0.00000160, 0.00000019, 0.00000108, 0.00000083, 0.00000137, 0.00000197, -0.00000032, -0.00000535, -0.00000927, -0.00000976, -0.00000838, -0.00000712, -0.00000574, -0.00000368, -0.00000201, -0.00000186, -0.00000252, -0.00000266, -0.00000218, -0.00000189, -0.00000184, -0.00000139, -0.00000051, },
		{ -0.00000101, -0.00000143, -0.00000199, -0.00000288, -0.00000316, -0.00000166, 0.00000104, 0.00000284, 0.00000266, 0.00000143, -0.00000003, -0.00000216, -0.00000462, -0.00000531, -0.00000342, -0.00000141, -0.00000165, -0.00000281, -0.00000258, -0.00000175, -0.00000234, -0.00000324, -0.00000209, -0.00000011, -0.00000039, -0.00000217, -0.00000211, -0.00000064, -0.00000109, -0.00000279, -0.00000154, 0.00000223, 0.00000291, -0.00000088, -0.00000360, -0.00000229, -0.00000103, -0.00000256, -0.00000313, -0.00000070, -0.00000028, -0.00000493, -0.00000913, -0.00000826, -0.00000716, -0.00001026, -0.00001114, -0.00000285, 0.00000776, 0.00000927, 0.00000413, 0.00000348, 0.00000751, 0.00000679, 0.00000130, 0.00000005, 0.00000184, -0.00000381, -0.00001383, -0.00001251, -0.00000162, -0.00000404, -0.00002368, -0.00002906, -0.00000056, 0.00003038, 0.00002344, -0.00000607, -0.00000800, 0.00002205, 0.00003193, -0.00000464, -0.00004186, -0.00002858, 0.00001486, 0.00002853, 0.00000117, -0.00002092, -0.00000904, 0.00001042, 0.00000549, -0.00001294, -0.00001541, -0.00000080, 0.00000852, 0.00000409, -0.00000227, -0.00000337, -0.00000498, -0.00000953, -0.00001036, -0.00000518, -0.00000056, 0.00000062, 0.00000385, 0.00001067, 0.00001297, 0.00000530, -0.00000592, -0.00001170, -0.00001200, -0.00001173, -0.00001122, -0.00000784, -0.00000293, -0.00000054, -0.00000095, -0.00000129, -0.00000096, -0.00000147, -0.00000222, -0.00000094, 0.00000170, 0.00000233, -0.00000006, -0.00000285, -0.00000400, -0.00000455, -0.00000589, -0.00000731, -0.00000751, -0.00000679, -0.00000612, -0.00000542, -0.00000399, -0.00000203, -0.00000070, -0.00000055, },
		{ -0.00000270, -0.00000044, 0.00000042, 0.00000033, -0.00000074, -0.00000291, -0.00000490, -0.00000469, -0.00000212, 0.00000066, 0.00000205, 0.00000256, 0.00000267, 0.00000137, -0.00000140, -0.00000321, -0.00000242, -0.00000122, -0.00000214, -0.00000388, -0.00000379, -0.00000271, -0.00000304, -0.00000374, -0.00000206, 0.00000078, 0.00000109, -0.00000081, -0.00000144, -0.00000097, -0.00000277, -0.00000587, -0.00000510, -0.00000026, 0.00000246, 0.00000063, -0.00000079, 0.00000103, 0.00000169, -0.00000146, -0.00000356, -0.00000127, 0.00000049, -0.00000228, -0.00000482, -0.00000310, -0.00000294, -0.00000951, -0.00001534, -0.00001110, -0.00000198, 0.00000067, -0.00000125, 0.00000276, 0.00001121, 0.00001264, 0.00000677, 0.00000423, 0.00000460, -0.00000308, -0.00001431, -0.00000944, 0.00001007, 0.00001348, -0.00001241, -0.00003480, -0.00002133, 0.00000719, 0.00000816, -0.00001261, -0.00000889, 0.00002713, 0.00004465, 0.00001200, -0.00002985, -0.00002647, 0.00000947, 0.00002118, -0.00000536, -0.00002621, -0.00001147, 0.00001317, 0.00001247, -0.00000672, -0.00001464, -0.00000586, 0.00000089, -0.00000245, -0.00000523, -0.00000131, 0.00000202, -0.00000014, -0.00000286, -0.00000320, -0.00000535, -0.00000970, -0.00000885, 0.00000005, 0.00000862, 0.00000896, 0.00000406, 0.00000075, -0.00000095, -0.00000481, -0.00000976, -0.00001142, -0.00000934, -0.00000682, -0.00000511, -0.00000300, -0.00000098, -0.00000101, -0.00000252, -0.00000276, -0.00000126, -0.00000035, -0.00000098, -0.00000130, -0.00000006, 0.00000118, 0.00000037, -0.00000227, -0.00000515, -0.00000748, -0.00000930, -0.00001013, -0.00000905, -0.00000609, },
		{ -0.00001164, -0.00001096, -0.00000823, -0.00000512, -0.00000246, -0.00000029, 0.00000077, -0.00000029, -0.00000275, -0.00000421, -0.00000358, -0.00000231, -0.00000165, -0.00000068, 0.00000117, 0.00000216, 0.00000115, 0.00000017, 0.00000104, 0.00000175, -0.00000031, -0.00000345, -0.00000433, -0.00000360, -0.00000414, -0.00000510, -0.00000339, 0.00000001, 0.00000107, -0.00000015, -0.00000001, 0.00000114, -0.00000087, -0.00000545, -0.00000670, -0.00000336, -0.00000124, -0.00000278, -0.00000276, 0.00000156, 0.00000460, 0.00000239, -0.00000032, 0.00000060, 0.00000086, -0.00000308, -0.00000552, -0.00000179, 0.00000121, -0.00000383, -0.00001046, -0.00000886, -0.00000360, -0.00000556, -0.00001118, -0.00000781, 0.00000329, 0.00000892, 0.00000790, 0.00001107, 0.00001571, 0.00000580, -0.00001476, -0.00001758, 0.00000593, 0.00002330, 0.00000782, -0.00001783, -0.00001723, -0.00000007, -0.00000450, -0.00002823, -0.00002665, 0.00001218, 0.00003967, 0.00001873, -0.00001706, -0.00001419, 0.00001901, 0.00002785, -0.00000357, -0.00003144, -0.00002141, 0.00000484, 0.00001034, -0.00000332, -0.00000940, -0.00000167, 0.00000259, -0.00000426, -0.00001066, -0.00000827, -0.00000276, 0.00000006, 0.00000252, 0.00000526, 0.00000283, -0.00000548, -0.00001117, -0.00000860, -0.00000309, -0.00000137, -0.00000149, 0.00000150, 0.00000583, 0.00000591, 0.00000130, -0.00000365, -0.00000673, -0.00000892, -0.00001009, -0.00000868, -0.00000546, -0.00000317, -0.00000261, -0.00000217, -0.00000139, -0.00000173, -0.00000336, -0.00000398, -0.00000188, 0.00000153, 0.00000356, 0.00000316, 0.00000106, -0.00000200, -0.00000579, -0.00000955, },
		{ 0.00000024, -0.00000345, -0.00000729, -0.00000953, -0.00000991, -0.00000940, -0.00000831, -0.00000604, -0.00000309, -0.00000134, -0.00000135, -0.00000144, -0.00000073, -0.00000073, -0.00000238, -0.00000371, -0.00000299, -0.00000161, -0.00000118, -0.00000016, 0.00000295, 0.00000543, 0.00000387, -0.00000007, -0.00000221, -0.00000261, -0.00000416, -0.00000605, -0.00000483, -0.00000140, -0.00000023, -0.00000138, -0.00000067, 0.00000176, 0.00000104, -0.00000306, -0.00000463, -0.00000217, -0.00000166, -0.00000541, -0.00000709, -0.00000278, 0.00000188, 0.00000196, 0.00000157, 0.00000440, 0.00000521, -0.00000013, -0.00000490, -0.00000257, 0.00000056, -0.00000344, -0.00000838, -0.00000443, 0.00000244, -0.00000069, -0.00001026, -0.00001148, -0.00000469, -0.00000354, -0.00000687, 0.00000161, 0.00001960, 0.00002127, -0.00000071, -0.00001641, -0.00000305, 0.00001727, 0.00001320, -0.00000428, -0.00000314, 0.00001100, 0.00000264, -0.00002816, -0.00003704, -0.00000614, 0.00002134, 0.00000642, -0.00002288, -0.00001459, 0.00002516, 0.00004088, 0.00001158, -0.00002170, -0.00002069, -0.00000086, 0.00000295, -0.00000962, -0.00001342, -0.00000212, 0.00000622, 0.00000133, -0.00000704, -0.00000921, -0.00000786, -0.00000667, -0.00000277, 0.00000404, 0.00000686, 0.00000244, -0.00000293, -0.00000379, -0.00000371, -0.00000733, -0.00001111, -0.00000878, -0.00000138, 0.00000461, 0.00000620, 0.00000524, 0.00000294, -0.00000127, -0.00000628, -0.00000924, -0.00000928, -0.00000807, -0.00000662, -0.00000443, -0.00000202, -0.00000141, -0.00000311, -0.00000483, -0.00000413, -0.00000126, 0.00000162, 0.00000309, 0.00000327, 0.00000246, },
		{ 0.00000117, 0.00000151, 0.00000155, 0.00000035, -0.00000176, -0.00000350, -0.00000473, -0.00000656, -0.00000892, -0.00001003, -0.00000911, -0.00000755, -0.00000616, -0.00000368, 0.00000011, 0.00000246, 0.00000134, -0.00000134, -0.00000292, -0.00000395, -0.00000541, -0.00000500, -0.00000065, 0.00000447, 0.00000557, 0.00000335, 0.00000192, 0.00000138, -0.00000124, -0.00000477, -0.00000490, -0.00000212, -0.00000133, -0.00000296, -0.00000242, 0.00000052, 0.00000060, -0.00000279, -0.00000347, 0.00000016, 0.00000127, -0.00000338, -0.00000707, -0.00000481, -0.00000175, -0.00000262, -0.00000259, 0.00000253, 0.00000642, 0.00000320, -0.00000084, 0.00000139, 0.00000346, -0.00000292, -0.00001014, -0.00000613, 0.00000303, 0.00000220, -0.00000554, -0.00000419, 0.00000384, -0.00000101, -0.00001825, -0.00002110, 0.00000038, 0.00001873, 0.00001124, -0.00000381, 0.00000055, 0.00001428, 0.00001080, -0.00000307, 0.00000130, 0.00001946, 0.00001566, -0.00001465, -0.00002979, -0.00000796, 0.00001312, -0.00000528, -0.00003784, -0.00003142, 0.00001234, 0.00003847, 0.00001988, -0.00000854, -0.00000867, 0.00000759, 0.00000719, -0.00001065, -0.00001968, -0.00001039, 0.00000068, 0.00000094, -0.00000346, -0.00000461, -0.00000487, -0.00000697, -0.00000672, -0.00000200, 0.00000177, 0.00000093, -0.00000011, 0.00000233, 0.00000361, -0.00000156, -0.00000966, -0.00001293, -0.00000980, -0.00000475, -0.00000077, 0.00000269, 0.00000534, 0.00000519, 0.00000200, -0.00000198, -0.00000511, -0.00000747, -0.00000882, -0.00000813, -0.00000574, -0.00000386, -0.00000390, -0.00000466, -0.00000410, -0.00000200, 0.00000006, 0.00000096, },
		{ 0.00000030, 0.00000094, 0.00000011, -0.00000053, -0.00000060, -0.00000092, -0.00000139, -0.00000090, 0.00000021, -0.00000020, -0.00000299, -0.00000632, -0.00000863, -0.00001051, -0.00001199, -0.00001077, -0.00000590, -0.00000063, 0.00000162, 0.00000159, 0.00000130, -0.00000026, -0.00000415, -0.00000706, -0.00000516, -0.00000066, 0.00000146, 0.00000130, 0.00000256, 0.00000459, 0.00000332, -0.00000050, -0.00000181, -0.00000030, -0.00000078, -0.00000367, -0.00000385, -0.00000083, -0.00000054, -0.00000406, -0.00000465, -0.00000001, 0.00000251, -0.00000124, -0.00000464, -0.00000244, -0.00000029, -0.00000337, -0.00000584, -0.00000198, 0.00000209, -0.00000020, -0.00000230, 0.00000330, 0.00000856, 0.00000280, -0.00000619, -0.00000401, 0.00000332, -0.00000122, -0.00001236, -0.00000721, 0.00001330, 0.00001866, -0.00000391, -0.00002599, -0.00002083, -0.00000263, 0.00000184, -0.00000244, 0.00000527, 0.00001841, 0.00001414, -0.00000122, 0.00000113, 0.00002001, 0.00002085, -0.00000496, -0.00001958, 0.00000021, 0.00001982, -0.00000088, -0.00004021, -0.00004363, -0.00000603, 0.00002181, 0.00001004, -0.00001147, -0.00000593, 0.00001582, 0.00001948, 0.00000110, -0.00001357, -0.00001133, -0.00000466, -0.00000517, -0.00000800, -0.00000638, -0.00000360, -0.00000380, -0.00000403, -0.00000179, -0.00000087, -0.00000347, -0.00000453, -0.00000013, 0.00000483, 0.00000353, -0.00000274, -0.00000737, -0.00000792, -0.00000716, -0.00000667, -0.00000497, -0.00000161, 0.00000142, 0.00000261, 0.00000237, 0.00000117, -0.00000123, -0.00000431, -0.00000642, -0.00000682, -0.00000658, -0.00000680, -0.00000689, -0.00000539, -0.00000235, },
		{ -0.00000779, -0.00000457, -0.00000085, 0.00000086, 0.00000037, -0.00000059, -0.00000129, -0.00000239, -0.00000353, -0.00000286, -0.00000022, 0.00000179, 0.00000137, -0.00000035, -0.00000226, -0.00000549, -0.00001018, -0.00001307, -0.00001158, -0.00000788, -0.00000520, -0.00000290, 0.00000052, 0.00000265, 0.00000067, -0.00000279, -0.00000317, -0.00000157, -0.00000225, -0.00000421, -0.00000272, 0.00000159, 0.00000336, 0.00000205, 0.00000234, 0.00000444, 0.00000323, -0.00000148, -0.00000344, -0.00000128, -0.00000133, -0.00000535, -0.00000626, -0.00000123, 0.00000195, -0.00000151, -0.00000451, -0.00000114, 0.00000226, -0.00000074, -0.00000413, -0.00000132, 0.00000118, -0.00000377, -0.00000811, -0.00000203, 0.00000590, 0.00000252, -0.00000402, 0.00000307, 0.00001530, 0.00000755, -0.00001709, -0.00002395, 0.00000076, 0.00002517, 0.00001828, -0.00000612, -0.00001684, -0.00001287, -0.00001290, -0.00001512, -0.00000366, 0.00001414, 0.00001487, 0.00000207, 0.00000364, 0.00002159, 0.00002373, -0.00000057, -0.00001566, 0.00000391, 0.00002669, 0.00001026, -0.00002963, -0.00004001, -0.00001083, 0.00001190, -0.00000205, -0.00002502, -0.00001911, 0.00000764, 0.00002010, 0.00000945, -0.00000229, -0.00000146, 0.00000181, -0.00000280, -0.00000987, -0.00001129, -0.00000884, -0.00000692, -0.00000429, -0.00000004, 0.00000140, -0.00000235, -0.00000587, -0.00000385, 0.00000061, 0.00000124, -0.00000188, -0.00000375, -0.00000280, -0.00000215, -0.00000364, -0.00000549, -0.00000581, -0.00000488, -0.00000351, -0.00000173, 0.00000014, 0.00000098, 0.00000026, -0.00000127, -0.00000278, -0.00000448, -0.00000668, -0.00000838, },
		{ -0.00000327, -0.00000605, -0.00000718, -0.00000537, -0.00000216, -0.00000035, -0.00000053, -0.00000103, -0.00000116, -0.00000214, -0.00000395, -0.00000431, -0.00000227, -0.00000008, 0.00000042, 0.00000052, 0.00000123, 0.00000042, -0.00000338, -0.00000742, -0.00000872, -0.00000879, -0.00001006, -0.00001053, -0.00000731, -0.00000275, -0.00000103, -0.00000092, 0.00000128, 0.00000370, 0.00000144, -0.00000402, -0.00000612, -0.00000387, -0.00000229, -0.00000221, 0.00000084, 0.00000637, 0.00000765, 0.00000348, 0.00000069, 0.00000163, 0.00000007, -0.00000576, -0.00000789, -0.00000274, 0.00000114, -0.00000219, -0.00000535, -0.00000149, 0.00000262, -0.00000025, -0.00000345, 0.00000069, 0.00000447, -0.00000137, -0.00000805, -0.00000338, 0.00000312, -0.00000475, -0.00001607, -0.00000493, 0.00002282, 0.00002917, 0.00000089, -0.00002612, -0.00001828, 0.00000857, 0.00001855, 0.00000834, -0.00000021, -0.00000232, -0.00000999, -0.00001913, -0.00001313, 0.00000276, 0.00000570, -0.00000363, 0.00000052, 0.00002126, 0.00002716, 0.00000483, -0.00001236, 0.00000415, 0.00002792, 0.00001636, -0.00002010, -0.00003145, -0.00000508, 0.00001582, -0.00000065, -0.00002950, -0.00003096, -0.00000791, 0.00000688, 0.00000209, -0.00000358, 0.00000251, 0.00001013, 0.00000761, -0.00000106, -0.00000701, -0.00000949, -0.00001080, -0.00000915, -0.00000382, 0.00000008, -0.00000127, -0.00000403, -0.00000316, -0.00000057, -0.00000127, -0.00000473, -0.00000592, -0.00000314, 0.00000017, 0.00000082, -0.00000076, -0.00000279, -0.00000461, -0.00000597, -0.00000621, -0.00000501, -0.00000312, -0.00000152, -0.00000056, -0.00000027, -0.00000105, },
		{ -0.00000257, -0.00000186, -0.00000246, -0.00000419, -0.00000519, -0.00000373, -0.00000093, 0.00000030, -0.00000071, -0.00000180, -0.00000186, -0.00000239, -0.00000397, -0.00000441, -0.00000257, -0.00000085, -0.00000119, -0.00000166, -0.00000032, 0.00000078, -0.00000059, -0.00000237, -0.00000209, -0.00000215, -0.00000574, -0.00001037, -0.00001145, -0.00000983, -0.00000915, -0.00000792, -0.00000231, 0.00000473, 0.00000612, 0.00000178, -0.00000163, -0.00000222, -0.00000427, -0.00000753, -0.00000583, 0.00000106, 0.00000538, 0.00000434, 0.00000423, 0.00000709, 0.00000573, -0.00000207, -0.00000690, -0.00000342, -0.00000018, -0.00000392, -0.00000691, -0.00000172, 0.00000365, 0.00000027, -0.00000426, -0.00000009, 0.00000458, -0.00000149, -0.00000784, 0.00000132, 0.00001337, 0.00000186, -0.00002540, -0.00002861, 0.00000401, 0.00003312, 0.00002394, -0.00000511, -0.00001504, -0.00000286, 0.00000584, 0.00000312, 0.00000251, 0.00000579, 0.00000083, -0.00000934, -0.00000816, 0.00000192, 0.00000046, -0.00001206, -0.00000999, 0.00001197, 0.00002307, 0.00000607, -0.00000993, 0.00000483, 0.00002857, 0.00001975, -0.00001479, -0.00002695, -0.00000113, 0.00002221, 0.00000861, -0.00002170, -0.00002942, -0.00001330, -0.00000282, -0.00000952, -0.00001561, -0.00000751, 0.00000550, 0.00001002, 0.00000673, 0.00000260, -0.00000122, -0.00000580, -0.00000817, -0.00000620, -0.00000369, -0.00000439, -0.00000565, -0.00000356, -0.00000030, -0.00000091, -0.00000513, -0.00000768, -0.00000572, -0.00000173, 0.00000085, 0.00000124, 0.00000041, -0.00000112, -0.00000322, -0.00000522, -0.00000626, -0.00000605, -0.00000506, -0.00000380, },
		{ -0.00000512, -0.00000489, -0.00000458, -0.00000425, -0.00000454, -0.00000576, -0.00000629, -0.00000412, -0.00000026, 0.00000182, 0.00000090, -0.00000065, -0.00000118, -0.00000215, -0.00000426, -0.00000500, -0.00000286, -0.00000069, -0.00000115, -0.00000236, -0.00000175, -0.00000100, -0.00000226, -0.00000306, -0.00000048, 0.00000244, 0.00000095, -0.00000350, -0.00000627, -0.00000791, -0.00001168, -0.00001477, -0.00001131, -0.00000321, 0.00000157, 0.00000140, 0.00000190, 0.00000390, 0.00000175, -0.00000450, -0.00000700, -0.00000347, -0.00000099, -0.00000218, -0.00000019, 0.00000683, 0.00000962, 0.00000380, -0.00000118, 0.00000113, 0.00000224, -0.00000477, -0.00001023, -0.00000464, 0.00000301, 0.00000111, -0.00000274, 0.00000278, 0.00000725, -0.00000446, -0.00001820, -0.00000667, 0.00002124, 0.00002442, -0.00000824, -0.00003552, -0.00002255, 0.00000993, 0.00002108, 0.00000872, 0.00000063, 0.00000515, 0.00000616, -0.00000103, -0.00000342, 0.00000117, 0.00000065, -0.00000446, -0.00000016, 0.00001081, 0.00000814, -0.00000871, -0.00001345, 0.00000323, 0.00001383, -0.00000074, -0.00001532, -0.00000017, 0.00002607, 0.00002191, -0.00000997, -0.00002367, -0.00000006, 0.00002475, 0.00001534, -0.00001262, -0.00002169, -0.00000846, -0.00000065, -0.00001081, -0.00002204, -0.00001868, -0.00000693, 0.00000046, 0.00000221, 0.00000323, 0.00000364, 0.00000158, -0.00000068, -0.00000077, -0.00000115, -0.00000436, -0.00000743, -0.00000628, -0.00000281, -0.00000229, -0.00000555, -0.00000817, -0.00000692, -0.00000326, -0.00000036, 0.00000068, 0.00000063, 0.00000011, -0.00000101, -0.00000272, -0.00000433, -0.00000511, },
		{ -0.00000305, -0.00000393, -0.00000397, -0.00000395, -0.00000410, -0.00000418, -0.00000478, -0.00000659, -0.00000810, -0.00000666, -0.00000264, 0.00000025, 0.00000026, -0.00000034, 0.00000014, -0.00000024, -0.00000269, -0.00000432, -0.00000282, -0.00000080, -0.00000132, -0.00000257, -0.00000182, -0.00000109, -0.00000309, -0.00000494, -0.00000260, 0.00000147, 0.00000202, -0.00000003, -0.00000036, -0.00000049, -0.00000512, -0.00001175, -0.00001300, -0.00000893, -0.00000672, -0.00000741, -0.00000468, 0.00000192, 0.00000458, 0.00000121, -0.00000077, 0.00000127, 0.00000028, -0.00000564, -0.00000724, -0.00000059, 0.00000450, 0.00000197, 0.00000079, 0.00000719, 0.00001047, 0.00000157, -0.00000834, -0.00000626, -0.00000062, -0.00000429, -0.00000726, 0.00000568, 0.00001963, 0.00000657, -0.00002371, -0.00002836, 0.00000416, 0.00003157, 0.00001851, -0.00001342, -0.00002267, -0.00000779, 0.00000229, -0.00000067, 0.00000056, 0.00001064, 0.00001430, 0.00000661, 0.00000030, -0.00000005, -0.00000353, -0.00000887, -0.00000247, 0.00001290, 0.00001555, 0.00000134, -0.00000555, 0.00000648, 0.00001334, -0.00000418, -0.00002286, -0.00001151, 0.00001531, 0.00001659, -0.00000992, -0.00002234, 0.00000034, 0.00002613, 0.00001977, -0.00000633, -0.00001589, -0.00000320, 0.00000553, -0.00000423, -0.00001809, -0.00001979, -0.00001272, -0.00000798, -0.00000704, -0.00000526, -0.00000227, -0.00000029, 0.00000150, 0.00000424, 0.00000514, 0.00000171, -0.00000306, -0.00000455, -0.00000337, -0.00000395, -0.00000748, -0.00001031, -0.00000923, -0.00000526, -0.00000165, -0.00000005, 0.00000022, 0.00000018, -0.00000023, -0.00000147, },
		{ -0.00000005, -0.00000028, -0.00000158, -0.00000265, -0.00000305, -0.00000351, -0.00000411, -0.00000410, -0.00000400, -0.00000525, -0.00000703, -0.00000649, -0.00000352, -0.00000170, -0.00000275, -0.00000371, -0.00000229, -0.00000091, -0.00000182, -0.00000260, -0.00000075, 0.00000106, -0.00000037, -0.00000266, -0.00000223, -0.00000113, -0.00000295, -0.00000523, -0.00000344, 0.00000022, 0.00000037, -0.00000163, -0.00000043, 0.00000240, 0.00000058, -0.00000481, -0.00000652, -0.00000449, -0.00000610, -0.00001172, -0.00001286, -0.00000725, -0.00000288, -0.00000291, -0.00000045, 0.00000624, 0.00000778, 0.00000067, -0.00000470, -0.00000211, -0.00000069, -0.00000630, -0.00000783, 0.00000278, 0.00001257, 0.00000828, 0.00000028, 0.00000306, 0.00000598, -0.00000748, -0.00002202, -0.00000831, 0.00002365, 0.00002906, -0.00000454, -0.00003370, -0.00002151, 0.00001059, 0.00002031, 0.00000588, -0.00000366, -0.00000012, -0.00000074, -0.00000957, -0.00001045, 0.00000127, 0.00001041, 0.00000927, 0.00000695, 0.00000698, 0.00000117, -0.00000827, -0.00000602, 0.00000806, 0.00001350, 0.00000374, -0.00000009, 0.00001364, 0.00002191, 0.00000383, -0.00002004, -0.00001685, 0.00000457, 0.00000610, -0.00001759, -0.00002855, -0.00000505, 0.00002382, 0.00002227, -0.00000099, -0.00001112, 0.00000040, 0.00000998, 0.00000210, -0.00001147, -0.00001493, -0.00001021, -0.00000766, -0.00000935, -0.00001067, -0.00000995, -0.00000828, -0.00000493, 0.00000059, 0.00000495, 0.00000488, 0.00000221, 0.00000100, 0.00000123, -0.00000063, -0.00000566, -0.00001043, -0.00001125, -0.00000831, -0.00000472, -0.00000271, -0.00000188, -0.00000095, },
		{ -0.00000580, -0.00000527, -0.00000336, -0.00000175, -0.00000119, -0.00000079, -0.00000043, -0.00000131, -0.00000335, -0.00000474, -0.00000508, -0.00000585, -0.00000674, -0.00000511, -0.00000114, 0.00000069, -0.00000195, -0.00000540, -0.00000595, -0.00000530, -0.00000582, -0.00000518, -0.00000103, 0.00000302, 0.00000254, -0.00000024, -0.00000075, -0.00000042, -0.00000297, -0.00000581, -0.00000391, 0.00000024, 0.00000023, -0.00000276, -0.00000224, 0.00000110, 0.00000060, -0.00000311, -0.00000254, 0.00000197, 0.00000128, -0.00000623, -0.00001115, -0.00000965, -0.00000912, -0.00001211, -0.00000997, 0.00000002, 0.00000669, 0.00000391, 0.00000156, 0.00000595, 0.00000620, -0.00000490, -0.00001340, -0.00000664, 0.00000292, -0.00000006, -0.00000337, 0.00001099, 0.00002648, 0.00001145, -0.00002385, -0.00003179, 0.00000195, 0.00003264, 0.00002099, -0.00001206, -0.00002282, -0.00000826, 0.00000252, 0.00000044, 0.00000185, 0.00001047, 0.00001047, -0.00000165, -0.00000933, -0.00000471, 0.00000097, 0.00000089, 0.00000245, 0.00000747, 0.00000601, -0.00000192, -0.00000176, 0.00000888, 0.00001212, 0.00000148, -0.00000257, 0.00001287, 0.00002603, 0.00001320, -0.00000993, -0.00001138, 0.00000438, 0.00000271, -0.00002309, -0.00003738, -0.00001671, 0.00001357, 0.00001747, -0.00000064, -0.00000881, 0.00000299, 0.00001371, 0.00000749, -0.00000570, -0.00001013, -0.00000612, -0.00000349, -0.00000560, -0.00000886, -0.00001103, -0.00001207, -0.00001068, -0.00000611, -0.00000119, 0.00000079, 0.00000075, 0.00000181, 0.00000390, 0.00000369, -0.00000041, -0.00000564, -0.00000814, -0.00000718, -0.00000539, -0.00000515, },
		{ -0.00000021, -0.00000133, -0.00000485, -0.00000757, -0.00000797, -0.00000725, -0.00000602, -0.00000340, 0.00000006, 0.00000168, 0.00000039, -0.00000190, -0.00000398, -0.00000673, -0.00000915, -0.00000755, -0.00000167, 0.00000286, 0.00000193, -0.00000163, -0.00000378, -0.00000575, -0.00000935, -0.00001097, -0.00000691, -0.00000083, 0.00000142, 0.00000098, 0.00000215, 0.00000314, -0.00000024, -0.00000485, -0.00000421, -0.00000018, -0.00000003, -0.00000322, -0.00000283, 0.00000067, -0.00000005, -0.00000442, -0.00000357, 0.00000300, 0.00000499, -0.00000075, -0.00000489, -0.00000348, -0.00000501, -0.00001251, -0.00001526, -0.00000814, -0.00000231, -0.00000471, -0.00000421, 0.00000652, 0.00001350, 0.00000501, -0.00000489, -0.00000070, 0.00000358, -0.00001110, -0.00002640, -0.00000849, 0.00003108, 0.00004069, 0.00000415, -0.00003127, -0.00002251, 0.00001036, 0.00002163, 0.00000651, -0.00000544, -0.00000357, -0.00000351, -0.00000966, -0.00000762, 0.00000510, 0.00001195, 0.00000616, 0.00000049, 0.00000244, 0.00000316, -0.00000208, -0.00000437, -0.00000021, 0.00000098, -0.00000281, 0.00000060, 0.00001310, 0.00001708, 0.00000512, -0.00000266, 0.00000948, 0.00002336, 0.00001470, -0.00000512, -0.00000623, 0.00000910, 0.00000808, -0.00001835, -0.00003741, -0.00002369, 0.00000257, 0.00000733, -0.00000790, -0.00001403, -0.00000038, 0.00001359, 0.00001109, -0.00000030, -0.00000517, -0.00000190, 0.00000090, -0.00000087, -0.00000468, -0.00000817, -0.00001074, -0.00001102, -0.00000826, -0.00000479, -0.00000341, -0.00000350, -0.00000231, 0.00000046, 0.00000198, 0.00000029, -0.00000284, -0.00000394, -0.00000211, },
		{ -0.00000647, -0.00000220, 0.00000309, 0.00000477, 0.00000209, -0.00000191, -0.00000524, -0.00000841, -0.00001098, -0.00001041, -0.00000626, -0.00000178, 0.00000051, 0.00000135, 0.00000100, -0.00000233, -0.00000757, -0.00000937, -0.00000531, -0.00000030, 0.00000084, 0.00000008, 0.00000060, -0.00000019, -0.00000529, -0.00001057, -0.00001036, -0.00000681, -0.00000540, -0.00000484, -0.00000078, 0.00000392, 0.00000307, -0.00000105, -0.00000110, 0.00000194, 0.00000074, -0.00000396, -0.00000411, 0.00000010, -0.00000004, -0.00000489, -0.00000462, 0.00000235, 0.00000515, -0.00000006, -0.00000307, 0.00000070, 0.00000110, -0.00000672, -0.00001126, -0.00000597, -0.00000309, -0.00001072, -0.00001497, -0.00000450, 0.00000582, 0.00000078, -0.00000389, 0.00001156, 0.00002707, 0.00000726, -0.00003442, -0.00004260, -0.00000058, 0.00004020, 0.00003243, -0.00000362, -0.00001864, -0.00000535, 0.00000603, 0.00000306, 0.00000112, 0.00000605, 0.00000474, -0.00000551, -0.00000958, -0.00000208, 0.00000388, 0.00000147, 0.00000079, 0.00000720, 0.00001072, 0.00000534, -0.00000046, -0.00000125, -0.00000383, -0.00000925, -0.00000544, 0.00000983, 0.00001849, 0.00001027, 0.00000250, 0.00001192, 0.00002385, 0.00001503, -0.00000533, -0.00000800, 0.00000767, 0.00001076, -0.00001105, -0.00002986, -0.00002050, 0.00000093, 0.00000304, -0.00001398, -0.00002253, -0.00001033, 0.00000549, 0.00000742, 0.00000006, -0.00000264, 0.00000190, 0.00000602, 0.00000503, 0.00000067, -0.00000425, -0.00000819, -0.00000955, -0.00000776, -0.00000489, -0.00000364, -0.00000396, -0.00000375, -0.00000240, -0.00000181, -0.00000367, -0.00000646, },
		{ -0.00000368, -0.00000864, -0.00001127, -0.00000835, -0.00000200, 0.00000271, 0.00000389, 0.00000363, 0.00000283, -0.00000042, -0.00000613, -0.00001048, -0.00001097, -0.00000936, -0.00000737, -0.00000395, 0.00000040, 0.00000160, -0.00000163, -0.00000437, -0.00000309, -0.00000139, -0.00000302, -0.00000462, -0.00000204, 0.00000118, -0.00000039, -0.00000430, -0.00000504, -0.00000444, -0.00000744, -0.00001116, -0.00000900, -0.00000310, -0.00000099, -0.00000189, 0.00000110, 0.00000643, 0.00000549, -0.00000103, -0.00000330, -0.00000011, -0.00000077, -0.00000640, -0.00000630, 0.00000177, 0.00000566, 0.00000030, -0.00000298, 0.00000181, 0.00000339, -0.00000454, -0.00000900, -0.00000154, 0.00000340, -0.00000578, -0.00001383, -0.00000580, 0.00000112, -0.00001347, -0.00002806, -0.00000600, 0.00003763, 0.00004465, -0.00000112, -0.00004401, -0.00003363, 0.00000776, 0.00002632, 0.00001287, -0.00000042, 0.00000137, 0.00000293, -0.00000310, -0.00000418, 0.00000371, 0.00000716, 0.00000107, -0.00000260, 0.00000161, 0.00000316, -0.00000284, -0.00000542, 0.00000184, 0.00000925, 0.00000880, 0.00000619, 0.00000570, 0.00000102, -0.00000838, -0.00000931, 0.00000310, 0.00001279, 0.00000815, 0.00000341, 0.00001449, 0.00002806, 0.00002051, -0.00000156, -0.00000921, 0.00000250, 0.00000665, -0.00001054, -0.00002591, -0.00001606, 0.00000509, 0.00000752, -0.00001061, -0.00002350, -0.00001687, -0.00000420, -0.00000200, -0.00000748, -0.00000804, -0.00000060, 0.00000747, 0.00001008, 0.00000739, 0.00000213, -0.00000323, -0.00000649, -0.00000668, -0.00000513, -0.00000389, -0.00000338, -0.00000249, -0.00000103, -0.00000078, },
		{ -0.00000149, 0.00000074, -0.00000000, -0.00000441, -0.00000889, -0.00000929, -0.00000628, -0.00000400, -0.00000350, -0.00000172, 0.00000218, 0.00000443, 0.00000261, -0.00000087, -0.00000368, -0.00000709, -0.00001130, -0.00001251, -0.00000902, -0.00000495, -0.00000342, -0.00000138, 0.00000330, 0.00000558, 0.00000105, -0.00000538, -0.00000661, -0.00000434, -0.00000415, -0.00000449, -0.00000087, 0.00000298, -0.00000005, -0.00000732, -0.00001014, -0.00000848, -0.00000911, -0.00001071, -0.00000553, 0.00000427, 0.00000758, 0.00000319, 0.00000144, 0.00000445, 0.00000233, -0.00000620, -0.00000855, -0.00000074, 0.00000398, -0.00000101, -0.00000349, 0.00000365, 0.00000703, -0.00000218, -0.00000906, -0.00000160, 0.00000456, -0.00000537, -0.00001286, 0.00000321, 0.00001988, -0.00000108, -0.00004371, -0.00004773, 0.00000250, 0.00004786, 0.00003612, -0.00000804, -0.00002701, -0.00001113, 0.00000458, 0.00000316, 0.00000086, 0.00000705, 0.00000925, 0.00000177, -0.00000308, 0.00000074, 0.00000308, -0.00000075, -0.00000105, 0.00000585, 0.00000867, 0.00000160, -0.00000464, -0.00000164, 0.00000367, 0.00000415, 0.00000446, 0.00000777, 0.00000657, -0.00000159, -0.00000470, 0.00000371, 0.00001032, 0.00000431, -0.00000102, 0.00001063, 0.00002746, 0.00002479, 0.00000494, -0.00000551, 0.00000143, 0.00000340, -0.00001297, -0.00002751, -0.00001777, 0.00000476, 0.00001115, -0.00000339, -0.00001656, -0.00001380, -0.00000552, -0.00000655, -0.00001474, -0.00001764, -0.00001058, 0.00000020, 0.00000746, 0.00000921, 0.00000708, 0.00000343, 0.00000024, -0.00000164, -0.00000270, -0.00000368, -0.00000437, -0.00000376, },
		{ -0.00000106, -0.00000380, -0.00000438, -0.00000273, -0.00000142, -0.00000256, -0.00000424, -0.00000319, -0.00000068, -0.00000151, -0.00000621, -0.00000940, -0.00000768, -0.00000383, -0.00000096, 0.00000198, 0.00000521, 0.00000472, -0.00000138, -0.00000835, -0.00001172, -0.00001312, -0.00001412, -0.00001059, -0.00000051, 0.00000856, 0.00000861, 0.00000267, -0.00000144, -0.00000377, -0.00000832, -0.00001142, -0.00000668, 0.00000179, 0.00000401, 0.00000004, -0.00000212, -0.00000244, -0.00000775, -0.00001535, -0.00001443, -0.00000524, 0.00000011, -0.00000123, 0.00000094, 0.00000839, 0.00000897, -0.00000024, -0.00000503, 0.00000073, 0.00000330, -0.00000482, -0.00000899, 0.00000057, 0.00000831, 0.00000131, -0.00000480, 0.00000477, 0.00001047, -0.00000917, -0.00002936, -0.00000964, 0.00003355, 0.00003766, -0.00001324, -0.00005766, -0.00004202, 0.00000690, 0.00002853, 0.00001319, -0.00000195, 0.00000095, 0.00000405, -0.00000330, -0.00000746, -0.00000049, 0.00000580, 0.00000366, 0.00000159, 0.00000456, 0.00000436, -0.00000188, -0.00000363, 0.00000374, 0.00000928, 0.00000539, 0.00000030, 0.00000191, 0.00000457, 0.00000231, 0.00000051, 0.00000356, 0.00000464, -0.00000005, -0.00000077, 0.00000833, 0.00001471, 0.00000732, -0.00000129, 0.00000670, 0.00002288, 0.00002331, 0.00000681, -0.00000335, 0.00000200, 0.00000363, -0.00001191, -0.00002735, -0.00002073, -0.00000006, 0.00000854, -0.00000148, -0.00001155, -0.00000820, -0.00000035, -0.00000226, -0.00001308, -0.00002071, -0.00001836, -0.00000983, -0.00000202, 0.00000200, 0.00000320, 0.00000369, 0.00000449, 0.00000506, 0.00000439, 0.00000215, },
		{ 0.00000750, 0.00000907, 0.00000773, 0.00000375, -0.00000074, -0.00000362, -0.00000546, -0.00000764, -0.00000799, -0.00000309, 0.00000467, 0.00000775, 0.00000315, -0.00000397, -0.00000851, -0.00001114, -0.00001269, -0.00000987, -0.00000168, 0.00000584, 0.00000718, 0.00000431, 0.00000058, -0.00000575, -0.00001491, -0.00001926, -0.00001285, -0.00000191, 0.00000372, 0.00000441, 0.00000581, 0.00000576, -0.00000116, -0.00001008, -0.00001103, -0.00000525, -0.00000235, -0.00000301, 0.00000007, 0.00000501, 0.00000184, -0.00000813, -0.00001212, -0.00000785, -0.00000677, -0.00001132, -0.00000933, 0.00000185, 0.00000761, 0.00000215, 0.00000004, 0.00000848, 0.00001150, -0.00000081, -0.00001109, -0.00000483, 0.00000196, -0.00000645, -0.00001039, 0.00001159, 0.00003281, 0.00001087, -0.00003675, -0.00004427, 0.00000561, 0.00004946, 0.00003192, -0.00001896, -0.00004029, -0.00002220, -0.00000365, -0.00000333, -0.00000281, 0.00000839, 0.00001477, 0.00000690, -0.00000227, -0.00000231, -0.00000036, -0.00000237, -0.00000172, 0.00000444, 0.00000679, 0.00000130, -0.00000202, 0.00000294, 0.00000742, 0.00000436, 0.00000087, 0.00000406, 0.00000804, 0.00000595, 0.00000241, 0.00000271, 0.00000191, -0.00000312, -0.00000314, 0.00000772, 0.00001718, 0.00001261, 0.00000400, 0.00000914, 0.00002268, 0.00002266, 0.00000619, -0.00000549, -0.00000174, 0.00000124, -0.00001066, -0.00002432, -0.00001987, -0.00000254, 0.00000517, -0.00000337, -0.00001200, -0.00000781, 0.00000180, 0.00000273, -0.00000649, -0.00001562, -0.00001694, -0.00001221, -0.00000751, -0.00000609, -0.00000636, -0.00000511, -0.00000123, 0.00000368, },
		{ -0.00001117, -0.00000646, -0.00000040, 0.00000593, 0.00001061, 0.00001146, 0.00000886, 0.00000490, -0.00000045, -0.00000790, -0.00001363, -0.00001138, -0.00000169, 0.00000699, 0.00000895, 0.00000687, 0.00000398, -0.00000171, -0.00001071, -0.00001674, -0.00001475, -0.00000837, -0.00000299, 0.00000211, 0.00000780, 0.00000831, -0.00000012, -0.00000998, -0.00001213, -0.00000912, -0.00000840, -0.00000794, -0.00000152, 0.00000648, 0.00000626, -0.00000037, -0.00000325, -0.00000204, -0.00000492, -0.00001032, -0.00000798, 0.00000092, 0.00000351, -0.00000209, -0.00000380, 0.00000074, -0.00000112, -0.00001177, -0.00001577, -0.00000735, -0.00000196, -0.00000721, -0.00000694, 0.00000798, 0.00001832, 0.00000914, -0.00000143, 0.00000455, 0.00000737, -0.00001528, -0.00003588, -0.00001084, 0.00004079, 0.00004962, -0.00000284, -0.00005048, -0.00003500, 0.00001517, 0.00003496, 0.00001404, -0.00000705, -0.00000843, -0.00000857, -0.00001789, -0.00002029, -0.00000683, 0.00000718, 0.00000904, 0.00000601, 0.00000629, 0.00000487, -0.00000150, -0.00000461, -0.00000019, 0.00000325, 0.00000044, -0.00000084, 0.00000466, 0.00000879, 0.00000485, 0.00000010, 0.00000263, 0.00000764, 0.00000772, 0.00000571, 0.00000607, 0.00000426, -0.00000261, -0.00000515, 0.00000395, 0.00001427, 0.00001280, 0.00000686, 0.00001282, 0.00002665, 0.00002732, 0.00001007, -0.00000564, -0.00000678, -0.00000509, -0.00001390, -0.00002401, -0.00001889, -0.00000265, 0.00000486, -0.00000305, -0.00001227, -0.00000989, -0.00000078, 0.00000231, -0.00000381, -0.00001115, -0.00001217, -0.00000771, -0.00000409, -0.00000550, -0.00001003, -0.00001275, },
		{ -0.00000525, -0.00001103, -0.00001346, -0.00001316, -0.00001070, -0.00000535, 0.00000171, 0.00000727, 0.00001032, 0.00001188, 0.00001038, 0.00000294, -0.00000730, -0.00001201, -0.00000841, -0.00000273, 0.00000024, 0.00000338, 0.00000869, 0.00001066, 0.00000487, -0.00000395, -0.00000948, -0.00001284, -0.00001625, -0.00001514, -0.00000616, 0.00000352, 0.00000559, 0.00000329, 0.00000367, 0.00000342, -0.00000424, -0.00001390, -0.00001424, -0.00000675, -0.00000233, -0.00000191, 0.00000235, 0.00000824, 0.00000491, -0.00000592, -0.00001015, -0.00000526, -0.00000357, -0.00000771, -0.00000530, 0.00000528, 0.00000773, -0.00000317, -0.00001020, -0.00000463, -0.00000286, -0.00001424, -0.00001940, -0.00000459, 0.00000930, 0.00000285, -0.00000241, 0.00001834, 0.00003746, 0.00000986, -0.00004448, -0.00005312, 0.00000312, 0.00005470, 0.00003990, -0.00001239, -0.00003417, -0.00001391, 0.00000659, 0.00000620, 0.00000367, 0.00001017, 0.00001007, -0.00000473, -0.00001760, -0.00001558, -0.00000779, -0.00000478, -0.00000209, 0.00000488, 0.00000903, 0.00000542, 0.00000131, 0.00000223, 0.00000226, -0.00000233, -0.00000384, 0.00000272, 0.00000888, 0.00000660, 0.00000171, 0.00000269, 0.00000653, 0.00000674, 0.00000571, 0.00000741, 0.00000714, 0.00000103, -0.00000273, 0.00000375, 0.00001226, 0.00001078, 0.00000549, 0.00001211, 0.00002787, 0.00003211, 0.00001708, -0.00000092, -0.00000751, -0.00000964, -0.00001829, -0.00002670, -0.00002121, -0.00000515, 0.00000389, -0.00000138, -0.00000980, -0.00000935, -0.00000252, 0.00000014, -0.00000509, -0.00001166, -0.00001179, -0.00000529, 0.00000087, 0.00000058, },
		{ 0.00000107, 0.00000347, -0.00000057, -0.00000552, -0.00000825, -0.00001019, -0.00001183, -0.00001096, -0.00000739, -0.00000372, -0.00000029, 0.00000503, 0.00001033, 0.00001003, 0.00000411, -0.00000038, -0.00000035, -0.00000128, -0.00000618, -0.00000866, -0.00000339, 0.00000449, 0.00000790, 0.00000817, 0.00000841, 0.00000455, -0.00000661, -0.00001756, -0.00001936, -0.00001489, -0.00001127, -0.00000597, 0.00000592, 0.00001658, 0.00001383, 0.00000097, -0.00000802, -0.00001033, -0.00001368, -0.00001647, -0.00000897, 0.00000523, 0.00001027, 0.00000356, -0.00000097, 0.00000085, -0.00000268, -0.00001282, -0.00001338, -0.00000050, 0.00000659, -0.00000127, -0.00000616, 0.00000276, 0.00000655, -0.00000790, -0.00001844, -0.00000614, 0.00000434, -0.00001442, -0.00003352, -0.00000501, 0.00005095, 0.00005885, -0.00000132, -0.00005659, -0.00004178, 0.00001335, 0.00003729, 0.00001697, -0.00000439, -0.00000414, -0.00000136, -0.00000825, -0.00000950, 0.00000310, 0.00001310, 0.00000826, -0.00000088, -0.00000304, -0.00000367, -0.00000908, -0.00001244, -0.00000751, -0.00000066, 0.00000155, 0.00000306, 0.00000716, 0.00000770, 0.00000140, -0.00000317, 0.00000079, 0.00000644, 0.00000544, 0.00000187, 0.00000324, 0.00000697, 0.00000692, 0.00000533, 0.00000670, 0.00000721, 0.00000268, -0.00000020, 0.00000590, 0.00001379, 0.00001189, 0.00000546, 0.00001027, 0.00002584, 0.00003297, 0.00002148, 0.00000378, -0.00000591, -0.00001082, -0.00001964, -0.00002760, -0.00002351, -0.00000950, -0.00000003, -0.00000227, -0.00000815, -0.00000765, -0.00000192, 0.00000028, -0.00000519, -0.00001293, -0.00001438, -0.00000746, },
		{ -0.00001546, -0.00000973, -0.00000079, 0.00000260, -0.00000051, -0.00000406, -0.00000481, -0.00000529, -0.00000688, -0.00000724, -0.00000609, -0.00000655, -0.00000832, -0.00000681, -0.00000164, 0.00000169, 0.00000180, 0.00000449, 0.00001107, 0.00001303, 0.00000502, -0.00000559, -0.00000969, -0.00000869, -0.00000716, -0.00000227, 0.00000845, 0.00001678, 0.00001361, 0.00000295, -0.00000586, -0.00001317, -0.00002256, -0.00002643, -0.00001507, 0.00000413, 0.00001420, 0.00001264, 0.00000989, 0.00000707, -0.00000389, -0.00001837, -0.00002039, -0.00000904, -0.00000084, -0.00000143, 0.00000131, 0.00000954, 0.00000783, -0.00000648, -0.00001305, -0.00000320, 0.00000315, -0.00000604, -0.00001121, 0.00000145, 0.00000953, -0.00000535, -0.00001616, 0.00000566, 0.00002873, 0.00000179, -0.00005428, -0.00006018, 0.00000517, 0.00006492, 0.00005007, -0.00000854, -0.00003535, -0.00001546, 0.00000639, 0.00000594, 0.00000268, 0.00000988, 0.00001210, 0.00000025, -0.00000975, -0.00000578, 0.00000159, 0.00000155, 0.00000062, 0.00000568, 0.00000906, 0.00000318, -0.00000550, -0.00000856, -0.00000845, -0.00000912, -0.00000653, 0.00000145, 0.00000707, 0.00000484, 0.00000154, 0.00000426, 0.00000799, 0.00000559, 0.00000104, 0.00000194, 0.00000613, 0.00000709, 0.00000592, 0.00000683, 0.00000698, 0.00000268, 0.00000004, 0.00000627, 0.00001498, 0.00001449, 0.00000830, 0.00001138, 0.00002533, 0.00003319, 0.00002358, 0.00000574, -0.00000645, -0.00001288, -0.00002022, -0.00002615, -0.00002286, -0.00001181, -0.00000385, -0.00000476, -0.00000874, -0.00000778, -0.00000211, 0.00000097, -0.00000356, -0.00001203, },
		{ -0.00000271, -0.00001078, -0.00001531, -0.00001040, -0.00000123, 0.00000232, -0.00000135, -0.00000530, -0.00000551, -0.00000494, -0.00000531, -0.00000388, -0.00000071, -0.00000072, -0.00000500, -0.00000818, -0.00000806, -0.00000907, -0.00001147, -0.00000740, 0.00000564, 0.00001703, 0.00001712, 0.00001007, 0.00000393, -0.00000254, -0.00001202, -0.00001711, -0.00000999, 0.00000314, 0.00001081, 0.00001299, 0.00001483, 0.00001145, -0.00000388, -0.00002169, -0.00002555, -0.00001625, -0.00000802, -0.00000348, 0.00000615, 0.00001734, 0.00001502, -0.00000013, -0.00000986, -0.00000834, -0.00000904, -0.00001523, -0.00001169, 0.00000394, 0.00001057, -0.00000020, -0.00000711, 0.00000237, 0.00000764, -0.00000562, -0.00001395, 0.00000142, 0.00001188, -0.00001215, -0.00003722, -0.00000951, 0.00004875, 0.00005443, -0.00001394, -0.00007503, -0.00005617, 0.00000975, 0.00004204, 0.00002356, 0.00000093, 0.00000049, 0.00000271, -0.00000630, -0.00001043, 0.00000063, 0.00001118, 0.00000836, 0.00000190, 0.00000220, 0.00000258, -0.00000346, -0.00000734, -0.00000113, 0.00000768, 0.00000913, 0.00000555, 0.00000287, -0.00000087, -0.00000769, -0.00001138, -0.00000748, -0.00000215, -0.00000145, -0.00000079, 0.00000527, 0.00001124, 0.00000961, 0.00000401, 0.00000279, 0.00000551, 0.00000640, 0.00000574, 0.00000695, 0.00000732, 0.00000318, 0.00000002, 0.00000531, 0.00001410, 0.00001525, 0.00001062, 0.00001369, 0.00002714, 0.00003581, 0.00002731, 0.00000798, -0.00000832, -0.00001753, -0.00002343, -0.00002592, -0.00002145, -0.00001211, -0.00000564, -0.00000601, -0.00000911, -0.00000865, -0.00000373, 0.00000016, },
		{ -0.00000506, -0.00000116, -0.00000270, -0.00001027, -0.00001556, -0.00001106, -0.00000117, 0.00000307, -0.00000092, -0.00000550, -0.00000601, -0.00000586, -0.00000695, -0.00000532, -0.00000016, 0.00000254, 0.00000013, -0.00000235, -0.00000290, -0.00000712, -0.00001576, -0.00001857, -0.00000920, 0.00000360, 0.00000955, 0.00001138, 0.00001495, 0.00001500, 0.00000479, -0.00000857, -0.00001341, -0.00001112, -0.00000944, -0.00000539, 0.00000727, 0.00001987, 0.00001790, 0.00000472, -0.00000410, -0.00000694, -0.00001444, -0.00002360, -0.00001887, -0.00000130, 0.00000955, 0.00000740, 0.00000651, 0.00001073, 0.00000476, -0.00001305, -0.00002013, -0.00000767, 0.00000173, -0.00000559, -0.00000899, 0.00000557, 0.00001331, -0.00000459, -0.00001704, 0.00000751, 0.00003397, 0.00000585, -0.00005396, -0.00005923, 0.00001159, 0.00007309, 0.00005006, -0.00002137, -0.00005509, -0.00003265, -0.00000395, 0.00000148, 0.00000278, 0.00001442, 0.00001975, 0.00000767, -0.00000544, -0.00000504, 0.00000020, -0.00000016, -0.00000005, 0.00000648, 0.00001051, 0.00000429, -0.00000393, -0.00000387, 0.00000105, 0.00000312, 0.00000377, 0.00000660, 0.00000751, 0.00000254, -0.00000326, -0.00000447, -0.00000470, -0.00000794, -0.00000890, -0.00000196, 0.00000721, 0.00000962, 0.00000669, 0.00000593, 0.00000791, 0.00000801, 0.00000665, 0.00000726, 0.00000761, 0.00000389, 0.00000056, 0.00000473, 0.00001275, 0.00001452, 0.00001099, 0.00001441, 0.00002835, 0.00003914, 0.00003300, 0.00001275, -0.00000836, -0.00002212, -0.00002830, -0.00002791, -0.00002167, -0.00001303, -0.00000719, -0.00000640, -0.00000825, -0.00000848, },
		{ -0.00000740, -0.00000771, -0.00000580, -0.00000229, -0.00000282, -0.00000992, -0.00001594, -0.00001183, -0.00000115, 0.00000380, -0.00000049, -0.00000551, -0.00000578, -0.00000548, -0.00000735, -0.00000665, -0.00000158, 0.00000150, -0.00000045, -0.00000144, 0.00000120, -0.00000002, -0.00000875, -0.00001531, -0.00001176, -0.00000529, -0.00000477, -0.00000463, 0.00000365, 0.00001393, 0.00001487, 0.00000899, 0.00000551, 0.00000181, -0.00000920, -0.00001970, -0.00001582, -0.00000159, 0.00000709, 0.00000909, 0.00001527, 0.00002173, 0.00001231, -0.00001030, -0.00002355, -0.00001988, -0.00001480, -0.00001402, -0.00000358, 0.00001654, 0.00002220, 0.00000511, -0.00000924, -0.00000469, -0.00000157, -0.00001434, -0.00001822, 0.00000438, 0.00001917, -0.00000718, -0.00003700, -0.00000989, 0.00005125, 0.00005689, -0.00001550, -0.00007744, -0.00005179, 0.00002239, 0.00005479, 0.00002722, -0.00000619, -0.00001290, -0.00001266, -0.00002154, -0.00002312, -0.00000617, 0.00001167, 0.00001403, 0.00000900, 0.00000801, 0.00000619, -0.00000177, -0.00000684, -0.00000136, 0.00000633, 0.00000621, 0.00000235, 0.00000266, 0.00000457, 0.00000255, -0.00000016, 0.00000166, 0.00000487, 0.00000455, 0.00000317, 0.00000409, 0.00000315, -0.00000335, -0.00000924, -0.00000697, -0.00000003, 0.00000317, 0.00000261, 0.00000427, 0.00000837, 0.00001015, 0.00000946, 0.00000967, 0.00000947, 0.00000564, 0.00000196, 0.00000503, 0.00001208, 0.00001390, 0.00001069, 0.00001372, 0.00002769, 0.00004080, 0.00003815, 0.00001855, -0.00000642, -0.00002501, -0.00003248, -0.00003015, -0.00002264, -0.00001500, -0.00000991, -0.00000757, },
		{ -0.00001318, -0.00000972, -0.00000718, -0.00000694, -0.00000629, -0.00000339, -0.00000325, -0.00001013, -0.00001692, -0.00001290, -0.00000082, 0.00000538, 0.00000074, -0.00000536, -0.00000605, -0.00000578, -0.00000788, -0.00000704, -0.00000110, 0.00000230, -0.00000107, -0.00000382, -0.00000099, 0.00000011, -0.00000577, -0.00000980, -0.00000440, 0.00000141, -0.00000332, -0.00001178, -0.00001097, -0.00000333, 0.00000075, 0.00000350, 0.00001254, 0.00002039, 0.00001334, -0.00000362, -0.00001317, -0.00001354, -0.00001585, -0.00001682, -0.00000200, 0.00002263, 0.00003192, 0.00001939, 0.00000481, -0.00000231, -0.00001455, -0.00003188, -0.00003084, -0.00000574, 0.00001389, 0.00000974, 0.00000341, 0.00001171, 0.00001107, -0.00001478, -0.00002920, 0.00000163, 0.00003643, 0.00001045, -0.00005285, -0.00005960, 0.00001450, 0.00007799, 0.00005103, -0.00002541, -0.00005763, -0.00002769, 0.00000692, 0.00001202, 0.00000902, 0.00001586, 0.00001601, -0.00000227, -0.00002043, -0.00002072, -0.00001150, -0.00000593, -0.00000059, 0.00000953, 0.00001560, 0.00001007, 0.00000137, -0.00000027, 0.00000156, -0.00000022, -0.00000198, 0.00000221, 0.00000800, 0.00000808, 0.00000440, 0.00000277, 0.00000251, 0.00000080, 0.00000059, 0.00000462, 0.00000763, 0.00000406, -0.00000172, -0.00000226, 0.00000075, 0.00000056, -0.00000233, -0.00000175, 0.00000303, 0.00000713, 0.00000909, 0.00001109, 0.00001192, 0.00000880, 0.00000510, 0.00000712, 0.00001306, 0.00001466, 0.00001138, 0.00001334, 0.00002629, 0.00004078, 0.00004144, 0.00002322, -0.00000461, -0.00002741, -0.00003593, -0.00003148, -0.00002273, -0.00001666, },
		{ -0.00002184, -0.00001707, -0.00001593, -0.00001226, -0.00000759, -0.00000642, -0.00000653, -0.00000395, -0.00000319, -0.00001037, -0.00001888, -0.00001571, -0.00000211, 0.00000642, 0.00000285, -0.00000337, -0.00000449, -0.00000525, -0.00000917, -0.00000934, -0.00000218, 0.00000339, 0.00000081, -0.00000277, -0.00000109, -0.00000125, -0.00000871, -0.00001306, -0.00000453, 0.00000680, 0.00000598, -0.00000298, -0.00000686, -0.00000697, -0.00001211, -0.00001608, -0.00000602, 0.00001171, 0.00001867, 0.00001370, 0.00001025, 0.00000740, -0.00000747, -0.00002727, -0.00002769, -0.00000633, 0.00001206, 0.00001613, 0.00002063, 0.00002864, 0.00001892, -0.00001189, -0.00003191, -0.00002288, -0.00000967, -0.00001235, -0.00000812, 0.00001926, 0.00003196, -0.00000382, -0.00004313, -0.00001732, 0.00004908, 0.00005743, -0.00001826, -0.00008291, -0.00005348, 0.00002709, 0.00006098, 0.00002959, -0.00000689, -0.00001264, -0.00000997, -0.00001764, -0.00001823, 0.00000060, 0.00001909, 0.00001806, 0.00000637, -0.00000095, -0.00000619, -0.00001461, -0.00001803, -0.00000921, 0.00000313, 0.00000797, 0.00000801, 0.00000978, 0.00000983, 0.00000328, -0.00000388, -0.00000324, 0.00000269, 0.00000620, 0.00000661, 0.00000726, 0.00000679, 0.00000297, -0.00000012, 0.00000188, 0.00000521, 0.00000423, 0.00000160, 0.00000320, 0.00000692, 0.00000591, 0.00000042, -0.00000279, -0.00000123, 0.00000168, 0.00000425, 0.00000769, 0.00001047, 0.00000958, 0.00000750, 0.00000982, 0.00001550, 0.00001740, 0.00001439, 0.00001524, 0.00002648, 0.00004117, 0.00004407, 0.00002665, -0.00000431, -0.00003140, -0.00004057, -0.00003294, },
		{ -0.00004798, -0.00003673, -0.00002169, -0.00001685, -0.00001773, -0.00001431, -0.00000811, -0.00000640, -0.00000731, -0.00000456, -0.00000224, -0.00000902, -0.00001958, -0.00001868, -0.00000527, 0.00000456, 0.00000224, -0.00000260, -0.00000182, -0.00000167, -0.00000714, -0.00001000, -0.00000424, 0.00000152, -0.00000039, -0.00000278, 0.00000109, 0.00000238, -0.00000685, -0.00001536, -0.00000971, 0.00000223, 0.00000464, 0.00000035, 0.00000179, 0.00000481, -0.00000349, -0.00001716, -0.00001833, -0.00000781, -0.00000167, -0.00000015, 0.00001032, 0.00002431, 0.00001974, -0.00000294, -0.00001726, -0.00001378, -0.00001146, -0.00001610, -0.00000622, 0.00002207, 0.00003650, 0.00001971, -0.00000031, -0.00000118, -0.00000619, -0.00003247, -0.00004127, 0.00000132, 0.00004671, 0.00002205, -0.00004730, -0.00005786, 0.00001830, 0.00008290, 0.00005020, -0.00003423, -0.00006806, -0.00003236, 0.00000895, 0.00001721, 0.00001487, 0.00002153, 0.00001981, -0.00000209, -0.00002246, -0.00002078, -0.00000698, 0.00000168, 0.00000654, 0.00001356, 0.00001537, 0.00000502, -0.00000837, -0.00001300, -0.00001091, -0.00000896, -0.00000492, 0.00000469, 0.00001296, 0.00001175, 0.00000486, 0.00000136, 0.00000201, 0.00000215, 0.00000211, 0.00000467, 0.00000732, 0.00000620, 0.00000381, 0.00000438, 0.00000553, 0.00000307, 0.00000012, 0.00000267, 0.00000865, 0.00001043, 0.00000645, 0.00000233, 0.00000133, 0.00000152, 0.00000189, 0.00000377, 0.00000613, 0.00000626, 0.00000580, 0.00000925, 0.00001587, 0.00001942, 0.00001815, 0.00001926, 0.00002959, 0.00004449, 0.00004910, 0.00003153, -0.00000402, -0.00003711, },
		{ 0.00000013, -0.00004111, -0.00005635, -0.00004311, -0.00002418, -0.00001831, -0.00002012, -0.00001607, -0.00000784, -0.00000581, -0.00000848, -0.00000642, -0.00000234, -0.00000742, -0.00001851, -0.00001911, -0.00000636, 0.00000296, -0.00000066, -0.00000610, -0.00000349, -0.00000025, -0.00000390, -0.00000665, -0.00000146, 0.00000282, -0.00000201, -0.00000678, -0.00000194, 0.00000308, -0.00000324, -0.00001160, -0.00000755, 0.00000175, 0.00000040, -0.00000674, -0.00000346, 0.00000652, 0.00000572, -0.00000453, -0.00000750, -0.00000387, -0.00000950, -0.00002051, -0.00001501, 0.00000648, 0.00001783, 0.00001148, 0.00000867, 0.00001500, 0.00000676, -0.00002067, -0.00003337, -0.00001335, 0.00000918, 0.00000974, 0.00001219, 0.00003461, 0.00003743, -0.00001302, -0.00006383, -0.00003751, 0.00003921, 0.00005664, -0.00001663, -0.00007993, -0.00004484, 0.00004114, 0.00007232, 0.00003070, -0.00001511, -0.00002353, -0.00001814, -0.00002091, -0.00001566, 0.00000847, 0.00002852, 0.00002392, 0.00000663, -0.00000377, -0.00000824, -0.00001396, -0.00001457, -0.00000327, 0.00001059, 0.00001453, 0.00001040, 0.00000595, 0.00000044, -0.00000866, -0.00001460, -0.00001021, -0.00000047, 0.00000501, 0.00000590, 0.00000751, 0.00000935, 0.00000745, 0.00000351, 0.00000232, 0.00000331, 0.00000319, 0.00000324, 0.00000608, 0.00000852, 0.00000604, 0.00000170, 0.00000230, 0.00000746, 0.00001033, 0.00000841, 0.00000582, 0.00000543, 0.00000557, 0.00000505, 0.00000502, 0.00000521, 0.00000400, 0.00000293, 0.00000584, 0.00001239, 0.00001735, 0.00001851, 0.00002134, 0.00003256, 0.00004950, 0.00005749, 0.00004076, },
		{ 0.00006462, 0.00005179, 0.00000853, -0.00004029, -0.00006116, -0.00004793, -0.00002698, -0.00002162, -0.00002523, -0.00002018, -0.00000835, -0.00000441, -0.00000842, -0.00000797, -0.00000349, -0.00000770, -0.00001902, -0.00001960, -0.00000528, 0.00000524, 0.00000005, -0.00000857, -0.00000777, -0.00000433, -0.00000679, -0.00000721, 0.00000182, 0.00000914, 0.00000341, -0.00000560, -0.00000458, -0.00000088, -0.00000683, -0.00001338, -0.00000544, 0.00000788, 0.00000641, -0.00000594, -0.00000878, -0.00000167, -0.00000142, -0.00000702, -0.00000158, 0.00001093, 0.00000730, -0.00001244, -0.00002154, -0.00001257, -0.00000767, -0.00001222, -0.00000172, 0.00002730, 0.00003825, 0.00001317, -0.00001381, -0.00001487, -0.00001450, -0.00003280, -0.00003056, 0.00002494, 0.00007649, 0.00004301, -0.00004553, -0.00007122, 0.00000159, 0.00006884, 0.00003739, -0.00004527, -0.00007122, -0.00002363, 0.00002488, 0.00003112, 0.00002125, 0.00002039, 0.00001343, -0.00001063, -0.00002883, -0.00002121, -0.00000158, 0.00000907, 0.00001218, 0.00001641, 0.00001602, 0.00000395, -0.00001038, -0.00001380, -0.00000813, -0.00000220, 0.00000352, 0.00001128, 0.00001512, 0.00000897, -0.00000159, -0.00000691, -0.00000692, -0.00000699, -0.00000628, -0.00000059, 0.00000751, 0.00001140, 0.00001027, 0.00000811, 0.00000579, 0.00000224, 0.00000016, 0.00000275, 0.00000682, 0.00000685, 0.00000420, 0.00000480, 0.00000903, 0.00001123, 0.00000905, 0.00000634, 0.00000628, 0.00000758, 0.00000839, 0.00000876, 0.00000835, 0.00000622, 0.00000394, 0.00000493, 0.00000945, 0.00001371, 0.00001557, 0.00001910, 0.00003106, 0.00005099, },
		{ 0.00002727, 0.00004802, 0.00006645, 0.00005860, 0.00001510, -0.00003872, -0.00006303, -0.00004828, -0.00002497, -0.00002163, -0.00003003, -0.00002672, -0.00001222, -0.00000531, -0.00000878, -0.00000843, -0.00000311, -0.00000736, -0.00002089, -0.00002314, -0.00000710, 0.00000694, 0.00000352, -0.00000582, -0.00000673, -0.00000582, -0.00001191, -0.00001467, -0.00000355, 0.00000934, 0.00000844, 0.00000124, 0.00000185, 0.00000308, -0.00000802, -0.00002007, -0.00001307, 0.00000515, 0.00001009, 0.00000127, -0.00000148, 0.00000252, -0.00000418, -0.00001784, -0.00001459, 0.00000518, 0.00001334, 0.00000249, -0.00000347, 0.00000193, -0.00000612, -0.00003077, -0.00003491, -0.00000286, 0.00002650, 0.00002339, 0.00001572, 0.00002806, 0.00002215, -0.00003492, -0.00008365, -0.00004128, 0.00005762, 0.00008681, 0.00000796, -0.00006893, -0.00004375, 0.00003665, 0.00006125, 0.00001281, -0.00003352, -0.00003400, -0.00001777, -0.00001289, -0.00000450, 0.00001933, 0.00003592, 0.00002566, 0.00000367, -0.00000792, -0.00001080, -0.00001445, -0.00001326, 0.00000007, 0.00001552, 0.00001889, 0.00001191, 0.00000460, -0.00000138, -0.00000834, -0.00001118, -0.00000449, 0.00000603, 0.00001085, 0.00000991, 0.00000853, 0.00000602, -0.00000107, -0.00000890, -0.00000996, -0.00000404, 0.00000271, 0.00000726, 0.00001046, 0.00001122, 0.00000786, 0.00000348, 0.00000272, 0.00000436, 0.00000398, 0.00000236, 0.00000438, 0.00001003, 0.00001350, 0.00001180, 0.00000833, 0.00000708, 0.00000805, 0.00000952, 0.00001069, 0.00001086, 0.00000936, 0.00000743, 0.00000760, 0.00001039, 0.00001337, 0.00001464, 0.00001696, },
		{ 0.00001700, 0.00001877, 0.00002689, 0.00004660, 0.00006731, 0.00006212, 0.00001667, -0.00004287, -0.00006962, -0.00005010, -0.00002007, -0.00001564, -0.00002864, -0.00002950, -0.00001564, -0.00000850, -0.00001270, -0.00001175, -0.00000325, -0.00000513, -0.00002021, -0.00002582, -0.00001094, 0.00000446, 0.00000273, -0.00000478, -0.00000301, -0.00000056, -0.00000924, -0.00001758, -0.00001063, 0.00000164, 0.00000250, -0.00000101, 0.00000576, 0.00001297, 0.00000247, -0.00001546, -0.00001593, -0.00000210, 0.00000201, -0.00000434, -0.00000043, 0.00001199, 0.00000841, -0.00001123, -0.00001830, -0.00000615, -0.00000076, -0.00000909, -0.00000444, 0.00001793, 0.00002113, -0.00000937, -0.00003273, -0.00002047, -0.00000583, -0.00001716, -0.00001422, 0.00003941, 0.00008374, 0.00003430, -0.00007097, -0.00009926, -0.00001143, 0.00007514, 0.00005330, -0.00002943, -0.00005696, -0.00001031, 0.00003319, 0.00002906, 0.00000901, 0.00000359, -0.00000234, -0.00002209, -0.00003361, -0.00001792, 0.00000821, 0.00002113, 0.00002260, 0.00002325, 0.00001842, 0.00000175, -0.00001544, -0.00001799, -0.00000834, 0.00000151, 0.00000852, 0.00001512, 0.00001711, 0.00000963, -0.00000153, -0.00000670, -0.00000549, -0.00000320, 0.00000027, 0.00000741, 0.00001376, 0.00001221, 0.00000413, -0.00000262, -0.00000460, -0.00000412, -0.00000212, 0.00000239, 0.00000720, 0.00000868, 0.00000765, 0.00000752, 0.00000767, 0.00000502, 0.00000124, 0.00000178, 0.00000739, 0.00001245, 0.00001279, 0.00001037, 0.00000905, 0.00000981, 0.00001136, 0.00001247, 0.00001247, 0.00001127, 0.00000996, 0.00001012, 0.00001227, 0.00001517, },
		{ 0.00001339, 0.00001583, 0.00001866, 0.00002120, 0.00002885, 0.00004895, 0.00007271, 0.00006973, 0.00001994, -0.00004954, -0.00008292, -0.00006024, -0.00002143, -0.00001146, -0.00002445, -0.00002627, -0.00001213, -0.00000650, -0.00001528, -0.00001771, -0.00000783, -0.00000614, -0.00001979, -0.00002603, -0.00001140, 0.00000371, 0.00000003, -0.00000927, -0.00000568, 0.00000139, -0.00000410, -0.00001257, -0.00000759, 0.00000134, -0.00000328, -0.00001178, -0.00000431, 0.00001010, 0.00000792, -0.00000537, -0.00000510, 0.00000599, 0.00000274, -0.00001319, -0.00001362, 0.00000394, 0.00001032, -0.00000158, -0.00000512, 0.00000589, 0.00000217, -0.00002204, -0.00002814, -0.00000020, 0.00002062, 0.00000677, -0.00000546, 0.00001251, 0.00001568, -0.00003641, -0.00008135, -0.00003007, 0.00007808, 0.00010425, 0.00000818, -0.00008590, -0.00006462, 0.00002271, 0.00005420, 0.00000915, -0.00003240, -0.00002483, -0.00000246, 0.00000183, 0.00000392, 0.00001934, 0.00002731, 0.00000972, -0.00001533, -0.00002353, -0.00001801, -0.00001193, -0.00000261, 0.00001544, 0.00003058, 0.00002843, 0.00001358, 0.00000051, -0.00000680, -0.00001171, -0.00001152, -0.00000213, 0.00001040, 0.00001602, 0.00001408, 0.00001021, 0.00000536, -0.00000202, -0.00000742, -0.00000467, 0.00000366, 0.00000924, 0.00000943, 0.00000801, 0.00000665, 0.00000364, -0.00000020, -0.00000150, -0.00000006, 0.00000159, 0.00000339, 0.00000693, 0.00001014, 0.00000909, 0.00000490, 0.00000328, 0.00000652, 0.00001052, 0.00001113, 0.00000930, 0.00000853, 0.00001024, 0.00001308, 0.00001504, 0.00001521, 0.00001415, 0.00001301, 0.00001255, },
		{ 0.00001715, 0.00001697, 0.00001656, 0.00001766, 0.00002011, 0.00002216, 0.00002870, 0.00004982, 0.00007894, 0.00008154, 0.00002978, -0.00005029, -0.00009405, -0.00007311, -0.00002942, -0.00001570, -0.00002789, -0.00002705, -0.00000702, 0.00000219, -0.00000933, -0.00001730, -0.00001058, -0.00000976, -0.00002437, -0.00003053, -0.00001268, 0.00000629, 0.00000248, -0.00001067, -0.00001016, -0.00000326, -0.00000749, -0.00001346, -0.00000365, 0.00000988, 0.00000429, -0.00001148, -0.00001192, -0.00000120, -0.00000304, -0.00001251, -0.00000370, 0.00001813, 0.00001977, -0.00000179, -0.00001339, -0.00000495, -0.00000342, -0.00001525, -0.00001009, 0.00001726, 0.00002519, -0.00000368, -0.00002637, -0.00001386, -0.00000344, -0.00002356, -0.00002582, 0.00003159, 0.00008126, 0.00002850, -0.00008474, -0.00011140, -0.00000881, 0.00009190, 0.00007050, -0.00002218, -0.00005827, -0.00001512, 0.00002476, 0.00001493, -0.00000761, -0.00000816, -0.00000452, -0.00001536, -0.00002127, -0.00000412, 0.00001805, 0.00002169, 0.00001214, 0.00000501, -0.00000157, -0.00001398, -0.00002220, -0.00001369, 0.00000494, 0.00001801, 0.00002222, 0.00002300, 0.00001940, 0.00000788, -0.00000547, -0.00001050, -0.00000664, -0.00000043, 0.00000588, 0.00001316, 0.00001735, 0.00001334, 0.00000446, -0.00000104, -0.00000119, -0.00000021, 0.00000058, 0.00000380, 0.00000882, 0.00001132, 0.00000974, 0.00000659, 0.00000340, 0.00000009, -0.00000152, 0.00000100, 0.00000571, 0.00000783, 0.00000662, 0.00000631, 0.00000921, 0.00001230, 0.00001189, 0.00000869, 0.00000658, 0.00000801, 0.00001184, 0.00001516, 0.00001653, 0.00001686, },
		{ 0.00001489, 0.00001616, 0.00001845, 0.00002009, 0.00002040, 0.00002193, 0.00002493, 0.00002605, 0.00002948, 0.00004886, 0.00008169, 0.00009071, 0.00003989, -0.00004757, -0.00009960, -0.00007992, -0.00003301, -0.00001976, -0.00003693, -0.00003784, -0.00001208, 0.00000511, -0.00000250, -0.00001013, -0.00000372, -0.00000511, -0.00002572, -0.00003795, -0.00002052, 0.00000322, 0.00000382, -0.00000799, -0.00000776, -0.00000259, -0.00001133, -0.00002202, -0.00001144, 0.00000911, 0.00001087, -0.00000200, -0.00000341, 0.00000320, -0.00000720, -0.00002706, -0.00002193, 0.00000754, 0.00002238, 0.00001118, 0.00000469, 0.00001251, 0.00000405, -0.00002593, -0.00003401, -0.00000272, 0.00002180, 0.00000860, -0.00000304, 0.00001701, 0.00001848, -0.00004142, -0.00009121, -0.00003257, 0.00008801, 0.00011465, 0.00000348, -0.00010523, -0.00008315, 0.00001675, 0.00005901, 0.00001780, -0.00002218, -0.00001307, 0.00000657, 0.00000162, -0.00000716, 0.00000200, 0.00000987, -0.00000335, -0.00002093, -0.00002026, -0.00000814, -0.00000128, 0.00000282, 0.00001242, 0.00001909, 0.00001111, -0.00000453, -0.00001250, -0.00001093, -0.00000712, -0.00000104, 0.00001102, 0.00002337, 0.00002601, 0.00001886, 0.00000972, 0.00000212, -0.00000454, -0.00000714, -0.00000170, 0.00000792, 0.00001361, 0.00001363, 0.00001214, 0.00001020, 0.00000536, -0.00000087, -0.00000301, 0.00000064, 0.00000623, 0.00001027, 0.00001221, 0.00001155, 0.00000766, 0.00000291, 0.00000118, 0.00000253, 0.00000358, 0.00000334, 0.00000475, 0.00000943, 0.00001422, 0.00001485, 0.00001119, 0.00000737, 0.00000720, 0.00001035, 0.00001349, },
		{ 0.00001139, 0.00001404, 0.00001434, 0.00001477, 0.00001719, 0.00001931, 0.00002006, 0.00002318, 0.00002934, 0.00003266, 0.00003496, 0.00005233, 0.00008671, 0.00009936, 0.00004700, -0.00004925, -0.00010874, -0.00008654, -0.00003100, -0.00001477, -0.00003827, -0.00004673, -0.00002244, -0.00000258, -0.00000760, -0.00001140, 0.00000212, 0.00000662, -0.00001549, -0.00003516, -0.00002418, -0.00000286, -0.00000258, -0.00001297, -0.00000790, 0.00000293, -0.00000551, -0.00002230, -0.00001822, 0.00000011, 0.00000299, -0.00000627, 0.00000014, 0.00001630, 0.00000852, -0.00002006, -0.00002844, -0.00000835, 0.00000322, -0.00000538, -0.00000000, 0.00002753, 0.00003293, -0.00000180, -0.00002813, -0.00001414, -0.00000212, -0.00002420, -0.00002728, 0.00003408, 0.00008539, 0.00002446, -0.00009903, -0.00012219, -0.00000172, 0.00011261, 0.00008581, -0.00002437, -0.00007294, -0.00003093, 0.00001299, 0.00000752, -0.00000857, -0.00000004, 0.00001010, -0.00000155, -0.00001420, -0.00000533, 0.00000945, 0.00000744, -0.00000428, -0.00000871, -0.00000916, -0.00001534, -0.00001964, -0.00001034, 0.00000569, 0.00001335, 0.00001158, 0.00000868, 0.00000497, -0.00000362, -0.00001179, -0.00000980, 0.00000157, 0.00001307, 0.00002025, 0.00002425, 0.00002348, 0.00001534, 0.00000416, -0.00000206, -0.00000169, 0.00000099, 0.00000449, 0.00001036, 0.00001617, 0.00001634, 0.00001031, 0.00000371, 0.00000093, 0.00000153, 0.00000399, 0.00000773, 0.00001099, 0.00001136, 0.00000927, 0.00000749, 0.00000670, 0.00000503, 0.00000250, 0.00000237, 0.00000670, 0.00001269, 0.00001533, 0.00001295, 0.00000909, 0.00000846, },
		{ 0.00000898, 0.00000833, 0.00001198, 0.00001542, 0.00001583, 0.00001606, 0.00001834, 0.00001947, 0.00001815, 0.00002017, 0.00002812, 0.00003453, 0.00003857, 0.00005771, 0.00009716, 0.00011530, 0.00006006, -0.00004993, -0.00012309, -0.00010171, -0.00003556, -0.00001026, -0.00003351, -0.00004551, -0.00002289, -0.00000525, -0.00001603, -0.00002410, -0.00000680, 0.00000716, -0.00000783, -0.00002557, -0.00001502, 0.00000386, -0.00000250, -0.00002087, -0.00001797, -0.00000184, -0.00000369, -0.00001747, -0.00001306, 0.00000341, -0.00000025, -0.00001855, -0.00001473, 0.00000987, 0.00001470, -0.00000599, -0.00001240, 0.00000459, 0.00000441, -0.00002344, -0.00003075, 0.00000365, 0.00002970, 0.00001395, 0.00000089, 0.00002384, 0.00002612, -0.00004044, -0.00009694, -0.00003526, 0.00009295, 0.00011730, -0.00000686, -0.00012243, -0.00008968, 0.00002825, 0.00007801, 0.00003014, -0.00002029, -0.00001774, -0.00000212, -0.00001051, -0.00001923, -0.00000431, 0.00001175, 0.00000433, -0.00001090, -0.00001009, -0.00000020, 0.00000135, -0.00000161, 0.00000196, 0.00000524, -0.00000344, -0.00001725, -0.00002131, -0.00001543, -0.00000917, -0.00000346, 0.00000599, 0.00001439, 0.00001273, 0.00000286, -0.00000494, -0.00000631, -0.00000401, 0.00000151, 0.00001178, 0.00002264, 0.00002674, 0.00002317, 0.00001696, 0.00001067, 0.00000372, -0.00000130, 0.00000024, 0.00000737, 0.00001344, 0.00001450, 0.00001253, 0.00001031, 0.00000808, 0.00000609, 0.00000568, 0.00000684, 0.00000791, 0.00000846, 0.00000955, 0.00001071, 0.00000977, 0.00000663, 0.00000465, 0.00000690, 0.00001195, 0.00001492, 0.00001297, },
		{ 0.00001711, 0.00001503, 0.00001005, 0.00000835, 0.00001171, 0.00001524, 0.00001571, 0.00001664, 0.00002056, 0.00002265, 0.00002005, 0.00001980, 0.00002664, 0.00003257, 0.00003547, 0.00005575, 0.00010319, 0.00013291, 0.00008093, -0.00004058, -0.00013057, -0.00011642, -0.00004576, -0.00001481, -0.00003734, -0.00004803, -0.00001902, 0.00000377, -0.00001082, -0.00002847, -0.00001737, -0.00000374, -0.00001688, -0.00003079, -0.00001177, 0.00001676, 0.00001271, -0.00001276, -0.00001925, -0.00000893, -0.00001341, -0.00002665, -0.00001555, 0.00001142, 0.00001218, -0.00001309, -0.00002077, -0.00000330, -0.00000046, -0.00001890, -0.00001454, 0.00002120, 0.00003326, -0.00000242, -0.00003129, -0.00001576, -0.00000191, -0.00002518, -0.00002624, 0.00004485, 0.00010388, 0.00003653, -0.00010203, -0.00013122, -0.00000308, 0.00011785, 0.00008512, -0.00003529, -0.00008380, -0.00003133, 0.00002141, 0.00001669, -0.00000216, 0.00000483, 0.00001302, -0.00000324, -0.00002068, -0.00001335, 0.00000270, 0.00000261, -0.00000643, -0.00000626, -0.00000138, -0.00000439, -0.00000924, -0.00000381, 0.00000595, 0.00000607, -0.00000233, -0.00000858, -0.00001166, -0.00001690, -0.00002081, -0.00001540, -0.00000323, 0.00000531, 0.00000692, 0.00000611, 0.00000453, 0.00000034, -0.00000369, -0.00000157, 0.00000676, 0.00001558, 0.00002183, 0.00002614, 0.00002701, 0.00002149, 0.00001196, 0.00000521, 0.00000427, 0.00000604, 0.00000758, 0.00000949, 0.00001231, 0.00001425, 0.00001394, 0.00001221, 0.00001012, 0.00000797, 0.00000665, 0.00000745, 0.00000969, 0.00001071, 0.00000934, 0.00000785, 0.00000949, 0.00001403, },
		{ 0.00001031, 0.00001505, 0.00001900, 0.00001771, 0.00001280, 0.00001094, 0.00001436, 0.00001742, 0.00001631, 0.00001587, 0.00002011, 0.00002355, 0.00002175, 0.00002206, 0.00003002, 0.00003557, 0.00003427, 0.00005030, 0.00010106, 0.00014155, 0.00009649, -0.00003061, -0.00013193, -0.00012162, -0.00004654, -0.00001465, -0.00004481, -0.00006179, -0.00002806, 0.00000572, -0.00000252, -0.00002048, -0.00001157, -0.00000188, -0.00002382, -0.00004715, -0.00002846, 0.00001052, 0.00001879, 0.00000038, -0.00000388, 0.00000418, -0.00000988, -0.00003698, -0.00003276, 0.00000033, 0.00001295, -0.00000418, -0.00000835, 0.00000861, 0.00000170, -0.00003558, -0.00004382, -0.00000003, 0.00003391, 0.00001666, -0.00000098, 0.00002136, 0.00002231, -0.00005101, -0.00011029, -0.00003599, 0.00011200, 0.00014206, 0.00000387, -0.00012779, -0.00009701, 0.00002805, 0.00008001, 0.00002766, -0.00002478, -0.00001841, 0.00000061, -0.00000888, -0.00001947, -0.00000309, 0.00001583, 0.00000909, -0.00000754, -0.00000839, -0.00000071, -0.00000286, -0.00000933, -0.00000600, 0.00000114, -0.00000148, -0.00000935, -0.00000926, -0.00000266, -0.00000013, -0.00000168, -0.00000038, 0.00000147, -0.00000385, -0.00001406, -0.00001943, -0.00001771, -0.00001438, -0.00001112, -0.00000507, 0.00000216, 0.00000494, 0.00000285, 0.00000112, 0.00000237, 0.00000511, 0.00000949, 0.00001723, 0.00002582, 0.00002950, 0.00002673, 0.00002157, 0.00001706, 0.00001243, 0.00000750, 0.00000499, 0.00000676, 0.00001109, 0.00001500, 0.00001690, 0.00001639, 0.00001375, 0.00001056, 0.00000899, 0.00000948, 0.00001021, 0.00000959, 0.00000869, },
		{ 0.00001158, 0.00001055, 0.00001158, 0.00001577, 0.00001962, 0.00001816, 0.00001282, 0.00001139, 0.00001683, 0.00002208, 0.00002143, 0.00001977, 0.00002243, 0.00002406, 0.00002012, 0.00001999, 0.00003114, 0.00004067, 0.00003913, 0.00005172, 0.00010288, 0.00014901, 0.00010546, -0.00003114, -0.00014394, -0.00013196, -0.00004270, -0.00000136, -0.00003768, -0.00006764, -0.00004000, -0.00000457, -0.00001013, -0.00002426, -0.00000633, 0.00001273, -0.00001016, -0.00004562, -0.00003950, -0.00000585, 0.00000279, -0.00001083, -0.00000320, 0.00002040, 0.00001332, -0.00002172, -0.00003214, -0.00000988, -0.00000308, -0.00002137, -0.00001602, 0.00002108, 0.00002839, -0.00001615, -0.00004594, -0.00001987, 0.00000351, -0.00002075, -0.00002570, 0.00004791, 0.00010851, 0.00003011, -0.00012404, -0.00015203, -0.00000232, 0.00013876, 0.00010580, -0.00002879, -0.00008704, -0.00003427, 0.00002088, 0.00001609, -0.00000169, 0.00000907, 0.00001881, -0.00000123, -0.00002381, -0.00001792, 0.00000026, 0.00000299, -0.00000336, -0.00000043, 0.00000568, 0.00000024, -0.00000990, -0.00000936, -0.00000170, -0.00000058, -0.00000532, -0.00000600, -0.00000335, -0.00000507, -0.00000897, -0.00000671, 0.00000015, 0.00000245, -0.00000160, -0.00000610, -0.00000945, -0.00001496, -0.00002115, -0.00002191, -0.00001633, -0.00000969, -0.00000498, -0.00000078, 0.00000296, 0.00000437, 0.00000487, 0.00000871, 0.00001631, 0.00002342, 0.00002746, 0.00002965, 0.00003039, 0.00002742, 0.00002018, 0.00001233, 0.00000810, 0.00000843, 0.00001154, 0.00001517, 0.00001744, 0.00001732, 0.00001542, 0.00001347, 0.00001262, 0.00001237, },
		{ 0.00001554, 0.00001485, 0.00001405, 0.00001328, 0.00001457, 0.00001929, 0.00002363, 0.00002128, 0.00001320, 0.00000906, 0.00001415, 0.00002153, 0.00002389, 0.00002526, 0.00003041, 0.00003201, 0.00002410, 0.00001883, 0.00002827, 0.00003961, 0.00003948, 0.00005326, 0.00011055, 0.00016657, 0.00012405, -0.00002907, -0.00016423, -0.00015864, -0.00005480, 0.00000470, -0.00002580, -0.00005916, -0.00003425, -0.00000193, -0.00001682, -0.00004104, -0.00002186, 0.00001012, -0.00000048, -0.00003196, -0.00002715, 0.00000176, -0.00000119, -0.00003003, -0.00002814, 0.00000588, 0.00001611, -0.00000636, -0.00001028, 0.00001270, 0.00000792, -0.00003400, -0.00004615, -0.00000336, 0.00002535, -0.00000118, -0.00001939, 0.00001578, 0.00002807, -0.00004843, -0.00011531, -0.00003589, 0.00012461, 0.00015283, -0.00000554, -0.00015316, -0.00011558, 0.00002941, 0.00009262, 0.00003696, -0.00002235, -0.00001827, 0.00000083, -0.00000886, -0.00001755, 0.00000352, 0.00002513, 0.00001539, -0.00000720, -0.00001211, -0.00000552, -0.00000727, -0.00001193, -0.00000483, 0.00000643, 0.00000548, -0.00000394, -0.00000698, -0.00000349, -0.00000348, -0.00000638, -0.00000443, 0.00000009, -0.00000155, -0.00000809, -0.00001046, -0.00000707, -0.00000418, -0.00000351, -0.00000143, 0.00000099, -0.00000202, -0.00001071, -0.00001879, -0.00002254, -0.00002347, -0.00002231, -0.00001748, -0.00001017, -0.00000440, -0.00000105, 0.00000300, 0.00000916, 0.00001542, 0.00002082, 0.00002691, 0.00003346, 0.00003673, 0.00003391, 0.00002675, 0.00001953, 0.00001509, 0.00001383, 0.00001475, 0.00001641, 0.00001750, 0.00001744, 0.00001653, },
		{ 0.00001897, 0.00001915, 0.00001866, 0.00001770, 0.00001615, 0.00001432, 0.00001497, 0.00002082, 0.00002814, 0.00002836, 0.00002001, 0.00001290, 0.00001431, 0.00001865, 0.00001935, 0.00002206, 0.00003236, 0.00004005, 0.00003466, 0.00002767, 0.00003415, 0.00004200, 0.00003625, 0.00004593, 0.00010958, 0.00018240, 0.00015078, -0.00001288, -0.00017277, -0.00018286, -0.00007545, -0.00000585, -0.00003257, -0.00006339, -0.00002791, 0.00001619, -0.00000020, -0.00003899, -0.00003304, -0.00000507, -0.00001477, -0.00004227, -0.00002567, 0.00001957, 0.00002273, -0.00001763, -0.00003439, -0.00001318, -0.00000861, -0.00003010, -0.00001960, 0.00003056, 0.00004552, -0.00000269, -0.00003895, -0.00001743, -0.00000375, -0.00004287, -0.00005144, 0.00003868, 0.00011762, 0.00003652, -0.00013464, -0.00016643, 0.00000043, 0.00015649, 0.00011647, -0.00003717, -0.00010387, -0.00004443, 0.00001902, 0.00001532, -0.00000428, 0.00000719, 0.00001892, 0.00000011, -0.00002006, -0.00000959, 0.00001162, 0.00001240, 0.00000096, -0.00000027, 0.00000364, -0.00000325, -0.00001396, -0.00001200, -0.00000153, 0.00000177, -0.00000223, -0.00000276, -0.00000002, -0.00000217, -0.00000737, -0.00000680, -0.00000141, -0.00000002, -0.00000385, -0.00000631, -0.00000587, -0.00000694, -0.00000927, -0.00000772, -0.00000229, 0.00000077, -0.00000152, -0.00000620, -0.00001116, -0.00001743, -0.00002404, -0.00002716, -0.00002562, -0.00002216, -0.00001825, -0.00001249, -0.00000467, 0.00000297, 0.00000976, 0.00001792, 0.00002817, 0.00003721, 0.00004106, 0.00003917, 0.00003414, 0.00002871, 0.00002420, 0.00002093, 0.00001905, 0.00001859, },
		{ 0.00002891, 0.00002367, 0.00002128, 0.00002104, 0.00002172, 0.00002252, 0.00002192, 0.00001879, 0.00001616, 0.00001926, 0.00002678, 0.00002992, 0.00002512, 0.00002078, 0.00002343, 0.00002601, 0.00002135, 0.00001796, 0.00002592, 0.00003603, 0.00003528, 0.00003374, 0.00004639, 0.00005746, 0.00004632, 0.00004490, 0.00010478, 0.00018771, 0.00016749, -0.00000030, -0.00017642, -0.00019499, -0.00008138, -0.00000683, -0.00004290, -0.00008603, -0.00004693, 0.00001524, 0.00001367, -0.00002158, -0.00001720, 0.00000609, -0.00001724, -0.00006342, -0.00005359, 0.00000482, 0.00002860, 0.00000162, -0.00001006, 0.00000913, -0.00000237, -0.00005292, -0.00006236, -0.00000318, 0.00004017, 0.00001681, -0.00000210, 0.00003309, 0.00003527, -0.00006510, -0.00014788, -0.00005533, 0.00013330, 0.00016988, -0.00000824, -0.00017533, -0.00013248, 0.00003246, 0.00010532, 0.00004353, -0.00002374, -0.00002087, -0.00000160, -0.00001516, -0.00002824, -0.00000728, 0.00001763, 0.00001201, -0.00000525, -0.00000357, 0.00000743, 0.00000480, -0.00000429, -0.00000119, 0.00000784, 0.00000528, -0.00000539, -0.00000853, -0.00000428, -0.00000399, -0.00000736, -0.00000539, 0.00000045, 0.00000085, -0.00000403, -0.00000574, -0.00000308, -0.00000243, -0.00000471, -0.00000471, -0.00000223, -0.00000270, -0.00000677, -0.00000900, -0.00000726, -0.00000524, -0.00000504, -0.00000472, -0.00000404, -0.00000608, -0.00001177, -0.00001817, -0.00002314, -0.00002727, -0.00003027, -0.00002985, -0.00002541, -0.00001912, -0.00001214, -0.00000268, 0.00001048, 0.00002493, 0.00003654, 0.00004317, 0.00004546, 0.00004471, 0.00004137, 0.00003563, },
		{ 0.00005748, 0.00005753, 0.00005076, 0.00004073, 0.00003143, 0.00002485, 0.00002204, 0.00002348, 0.00002664, 0.00002706, 0.00002485, 0.00002516, 0.00002866, 0.00002819, 0.00002148, 0.00001872, 0.00002707, 0.00003633, 0.00003448, 0.00002848, 0.00003070, 0.00003419, 0.00002733, 0.00002399, 0.00004340, 0.00006653, 0.00006245, 0.00005918, 0.00011702, 0.00020435, 0.00018511, 0.00000192, -0.00019714, -0.00022089, -0.00008740, 0.00000813, -0.00003014, -0.00009253, -0.00006671, -0.00000446, -0.00000200, -0.00003155, -0.00001217, 0.00003116, 0.00001372, -0.00004874, -0.00006216, -0.00001760, 0.00000180, -0.00002251, -0.00001705, 0.00003274, 0.00004147, -0.00002021, -0.00006094, -0.00002801, -0.00000181, -0.00003781, -0.00004196, 0.00006129, 0.00014439, 0.00004119, -0.00016045, -0.00019556, -0.00000128, 0.00017780, 0.00012974, -0.00004873, -0.00012693, -0.00005909, 0.00001624, 0.00001699, -0.00000112, 0.00001364, 0.00002616, 0.00000146, -0.00002758, -0.00002284, -0.00000281, 0.00000024, -0.00000538, 0.00000201, 0.00001323, 0.00000851, -0.00000439, -0.00000537, 0.00000316, 0.00000497, -0.00000031, -0.00000128, 0.00000154, -0.00000138, -0.00000849, -0.00000964, -0.00000445, -0.00000167, -0.00000317, -0.00000310, -0.00000086, -0.00000179, -0.00000580, -0.00000683, -0.00000367, -0.00000162, -0.00000298, -0.00000439, -0.00000432, -0.00000543, -0.00000835, -0.00000975, -0.00000814, -0.00000626, -0.00000642, -0.00000803, -0.00001077, -0.00001593, -0.00002318, -0.00002963, -0.00003330, -0.00003484, -0.00003460, -0.00003051, -0.00002054, -0.00000572, 0.00001059, 0.00002582, 0.00003929, 0.00005054, },
		{ 0.00000203, 0.00002422, 0.00004732, 0.00006488, 0.00007233, 0.00006989, 0.00006027, 0.00004637, 0.00003274, 0.00002502, 0.00002458, 0.00002709, 0.00002997, 0.00003548, 0.00004175, 0.00003923, 0.00002540, 0.00001447, 0.00001927, 0.00003167, 0.00003708, 0.00003918, 0.00004774, 0.00005193, 0.00003700, 0.00002079, 0.00003224, 0.00005670, 0.00005817, 0.00006129, 0.00013189, 0.00023898, 0.00022860, 0.00002410, -0.00021599, -0.00026431, -0.00011916, 0.00000460, -0.00001866, -0.00008241, -0.00005916, 0.00000088, -0.00000984, -0.00005923, -0.00004426, 0.00001838, 0.00002596, -0.00002333, -0.00003360, 0.00000771, 0.00000879, -0.00004848, -0.00006474, -0.00000268, 0.00003963, 0.00000530, -0.00001812, 0.00002761, 0.00003749, -0.00007173, -0.00016197, -0.00005354, 0.00015971, 0.00019524, -0.00001180, -0.00019939, -0.00014430, 0.00004635, 0.00012590, 0.00004841, -0.00003526, -0.00003624, -0.00001427, -0.00002580, -0.00003515, -0.00000621, 0.00002565, 0.00001992, -0.00000339, -0.00000869, -0.00000298, -0.00000836, -0.00001608, -0.00000712, 0.00000887, 0.00001034, 0.00000037, -0.00000292, 0.00000154, 0.00000177, -0.00000214, -0.00000050, 0.00000522, 0.00000458, -0.00000263, -0.00000692, -0.00000584, -0.00000542, -0.00000682, -0.00000512, -0.00000069, 0.00000023, -0.00000345, -0.00000609, -0.00000511, -0.00000391, -0.00000420, -0.00000343, -0.00000116, -0.00000099, -0.00000446, -0.00000847, -0.00001029, -0.00001062, -0.00001049, -0.00000930, -0.00000747, -0.00000751, -0.00001087, -0.00001624, -0.00002223, -0.00002909, -0.00003670, -0.00004261, -0.00004364, -0.00003876, -0.00002913, -0.00001573, },
		{ -0.00005037, -0.00005191, -0.00004677, -0.00003072, -0.00000404, 0.00002717, 0.00005565, 0.00007686, 0.00008747, 0.00008462, 0.00007022, 0.00005228, 0.00003774, 0.00002789, 0.00002431, 0.00003158, 0.00004661, 0.00005396, 0.00004420, 0.00002921, 0.00002496, 0.00002777, 0.00002668, 0.00002924, 0.00004735, 0.00006549, 0.00005928, 0.00004175, 0.00004565, 0.00006023, 0.00004896, 0.00004134, 0.00011820, 0.00025457, 0.00027437, 0.00006885, -0.00020508, -0.00028645, -0.00014759, -0.00001455, -0.00003543, -0.00009994, -0.00006223, 0.00002221, 0.00001852, -0.00005047, -0.00006040, -0.00000913, -0.00000330, -0.00004911, -0.00003981, 0.00003750, 0.00006207, -0.00000937, -0.00006295, -0.00003081, -0.00000728, -0.00005565, -0.00006157, 0.00006333, 0.00016579, 0.00004981, -0.00018190, -0.00022293, -0.00000266, 0.00019835, 0.00014215, -0.00005648, -0.00013651, -0.00005248, 0.00003348, 0.00002872, 0.00000002, 0.00000946, 0.00001932, -0.00000962, -0.00004083, -0.00003205, -0.00000499, 0.00000212, -0.00000357, 0.00000194, 0.00001057, 0.00000291, -0.00001147, -0.00001104, 0.00000014, 0.00000318, -0.00000224, -0.00000255, 0.00000252, 0.00000223, -0.00000289, -0.00000259, 0.00000338, 0.00000548, 0.00000163, -0.00000120, -0.00000114, -0.00000315, -0.00000737, -0.00000800, -0.00000421, -0.00000175, -0.00000300, -0.00000420, -0.00000347, -0.00000353, -0.00000525, -0.00000540, -0.00000274, -0.00000057, -0.00000143, -0.00000389, -0.00000613, -0.00000851, -0.00001116, -0.00001241, -0.00001143, -0.00000980, -0.00000921, -0.00000965, -0.00001104, -0.00001463, -0.00002149, -0.00003053, -0.00003919, -0.00004584, },
		{ -0.00002079, -0.00002772, -0.00003766, -0.00005115, -0.00006421, -0.00006933, -0.00006153, -0.00004149, -0.00001216, 0.00002383, 0.00006090, 0.00008922, 0.00010181, 0.00009972, 0.00008682, 0.00006522, 0.00004186, 0.00003058, 0.00003682, 0.00004731, 0.00004812, 0.00004478, 0.00004809, 0.00005001, 0.00003804, 0.00002520, 0.00003351, 0.00005295, 0.00005618, 0.00005153, 0.00006935, 0.00009264, 0.00007400, 0.00004339, 0.00010266, 0.00024642, 0.00028793, 0.00008874, -0.00020085, -0.00029346, -0.00014391, 0.00000045, -0.00003623, -0.00013194, -0.00010342, 0.00000394, 0.00002832, -0.00002841, -0.00003533, 0.00001465, 0.00000101, -0.00008330, -0.00009844, -0.00000356, 0.00006330, 0.00002443, -0.00000979, 0.00003650, 0.00003987, -0.00009566, -0.00020243, -0.00006932, 0.00018754, 0.00023328, -0.00000731, -0.00022843, -0.00017153, 0.00004130, 0.00013069, 0.00004736, -0.00003797, -0.00002970, -0.00000046, -0.00001494, -0.00003029, -0.00000295, 0.00002857, 0.00001873, -0.00001005, -0.00001732, -0.00001055, -0.00001528, -0.00002317, -0.00001341, 0.00000445, 0.00000768, -0.00000053, -0.00000173, 0.00000365, 0.00000182, -0.00000610, -0.00000717, -0.00000110, 0.00000096, -0.00000256, -0.00000288, 0.00000155, 0.00000338, 0.00000098, 0.00000028, 0.00000237, 0.00000174, -0.00000270, -0.00000558, -0.00000477, -0.00000419, -0.00000555, -0.00000571, -0.00000351, -0.00000227, -0.00000365, -0.00000486, -0.00000388, -0.00000258, -0.00000280, -0.00000354, -0.00000372, -0.00000446, -0.00000669, -0.00000926, -0.00001083, -0.00001167, -0.00001233, -0.00001236, -0.00001139, -0.00001058, -0.00001173, -0.00001540, },
		{ -0.00001357, -0.00001372, -0.00001362, -0.00001350, -0.00001640, -0.00002620, -0.00004249, -0.00005977, -0.00007271, -0.00007927, -0.00007672, -0.00005954, -0.00002593, 0.00001721, 0.00006023, 0.00009715, 0.00012161, 0.00012448, 0.00010542, 0.00007970, 0.00006175, 0.00004880, 0.00003488, 0.00003053, 0.00004669, 0.00006752, 0.00006826, 0.00005482, 0.00005164, 0.00005475, 0.00004206, 0.00002973, 0.00005707, 0.00010450, 0.00010620, 0.00007780, 0.00012967, 0.00027218, 0.00031172, 0.00008868, -0.00024028, -0.00034678, -0.00016240, 0.00003192, 0.00000849, -0.00011615, -0.00012027, -0.00002550, -0.00000473, -0.00006184, -0.00004791, 0.00004556, 0.00005954, -0.00004427, -0.00010499, -0.00004298, 0.00000591, -0.00004654, -0.00005793, 0.00008378, 0.00019635, 0.00004964, -0.00022751, -0.00027071, -0.00000283, 0.00023971, 0.00017634, -0.00005833, -0.00015975, -0.00007237, 0.00002168, 0.00001915, -0.00000508, 0.00001477, 0.00003147, -0.00000024, -0.00003697, -0.00002866, 0.00000093, 0.00000796, -0.00000042, 0.00000284, 0.00000941, -0.00000229, -0.00002179, -0.00002440, -0.00001276, -0.00000645, -0.00000646, -0.00000032, 0.00000932, 0.00000880, -0.00000091, -0.00000617, -0.00000378, -0.00000259, -0.00000496, -0.00000428, 0.00000030, 0.00000202, -0.00000052, -0.00000132, 0.00000149, 0.00000289, 0.00000080, -0.00000099, -0.00000085, -0.00000198, -0.00000536, -0.00000729, -0.00000582, -0.00000381, -0.00000356, -0.00000366, -0.00000274, -0.00000232, -0.00000361, -0.00000513, -0.00000525, -0.00000466, -0.00000475, -0.00000552, -0.00000645, -0.00000786, -0.00001010, -0.00001242, -0.00001366, -0.00001372, },
		{ -0.00001021, -0.00001222, -0.00001416, -0.00001620, -0.00001694, -0.00001501, -0.00001229, -0.00001327, -0.00001983, -0.00002966, -0.00004153, -0.00005782, -0.00007796, -0.00009359, -0.00009533, -0.00008037, -0.00004890, -0.00000019, 0.00005948, 0.00011181, 0.00014126, 0.00014933, 0.00014269, 0.00011722, 0.00007277, 0.00003366, 0.00002720, 0.00004602, 0.00006168, 0.00007046, 0.00008727, 0.00009820, 0.00007347, 0.00003408, 0.00003812, 0.00008068, 0.00009211, 0.00007868, 0.00015437, 0.00033259, 0.00039456, 0.00014281, -0.00026150, -0.00042964, -0.00024153, 0.00000447, 0.00002406, -0.00008761, -0.00008716, 0.00001073, 0.00000856, -0.00009945, -0.00011799, -0.00000291, 0.00006099, -0.00000834, -0.00005006, 0.00002860, 0.00005508, -0.00010025, -0.00022970, -0.00007314, 0.00022812, 0.00027385, -0.00001793, -0.00027815, -0.00020376, 0.00005436, 0.00016375, 0.00006540, -0.00004115, -0.00004159, -0.00001445, -0.00003062, -0.00004226, -0.00000422, 0.00003567, 0.00002508, -0.00000841, -0.00001655, -0.00000700, -0.00000963, -0.00001678, -0.00000595, 0.00001211, 0.00001180, -0.00000345, -0.00001186, -0.00001111, -0.00001389, -0.00001844, -0.00001244, 0.00000123, 0.00000744, 0.00000332, -0.00000040, 0.00000017, -0.00000119, -0.00000551, -0.00000616, -0.00000204, 0.00000024, -0.00000168, -0.00000273, -0.00000057, 0.00000103, 0.00000025, 0.00000020, 0.00000191, 0.00000186, -0.00000151, -0.00000487, -0.00000557, -0.00000510, -0.00000514, -0.00000464, -0.00000281, -0.00000154, -0.00000253, -0.00000470, -0.00000607, -0.00000633, -0.00000620, -0.00000586, -0.00000520, -0.00000485, -0.00000572, -0.00000782, },
		{ -0.00000487, -0.00000532, -0.00000649, -0.00000768, -0.00000935, -0.00001232, -0.00001553, -0.00001671, -0.00001620, -0.00001706, -0.00001979, -0.00002082, -0.00001957, -0.00002266, -0.00003630, -0.00005836, -0.00008277, -0.00010579, -0.00012064, -0.00011320, -0.00007476, -0.00001436, 0.00005226, 0.00011823, 0.00017539, 0.00020040, 0.00017639, 0.00012363, 0.00008033, 0.00005665, 0.00003968, 0.00003747, 0.00007002, 0.00011426, 0.00011862, 0.00008727, 0.00007954, 0.00009980, 0.00008226, 0.00004158, 0.00011761, 0.00034266, 0.00046930, 0.00023771, -0.00021523, -0.00045285, -0.00029578, -0.00004424, -0.00002283, -0.00013860, -0.00010964, 0.00005159, 0.00008901, -0.00004448, -0.00012048, -0.00004410, -0.00000256, -0.00009359, -0.00011276, 0.00008059, 0.00023940, 0.00006786, -0.00026995, -0.00032334, -0.00000194, 0.00028517, 0.00020578, -0.00007296, -0.00018859, -0.00007962, 0.00003519, 0.00003217, -0.00000180, 0.00001257, 0.00002592, -0.00001033, -0.00004747, -0.00003268, 0.00000313, 0.00000962, -0.00000271, -0.00000029, 0.00000886, -0.00000038, -0.00001800, -0.00001758, -0.00000280, 0.00000352, -0.00000067, -0.00000095, 0.00000230, -0.00000298, -0.00001413, -0.00001660, -0.00000896, -0.00000331, -0.00000368, -0.00000271, 0.00000152, 0.00000233, -0.00000173, -0.00000403, -0.00000205, -0.00000089, -0.00000287, -0.00000396, -0.00000218, -0.00000086, -0.00000162, -0.00000142, 0.00000107, 0.00000264, 0.00000111, -0.00000134, -0.00000244, -0.00000315, -0.00000451, -0.00000516, -0.00000391, -0.00000228, -0.00000232, -0.00000384, -0.00000537, -0.00000630, -0.00000692, -0.00000717, -0.00000663, -0.00000554, },
		{ -0.00000738, -0.00000673, -0.00000586, -0.00000577, -0.00000631, -0.00000646, -0.00000645, -0.00000766, -0.00000980, -0.00001098, -0.00001183, -0.00001602, -0.00002353, -0.00002800, -0.00002495, -0.00001878, -0.00001657, -0.00002010, -0.00003049, -0.00005364, -0.00009053, -0.00012705, -0.00014529, -0.00013948, -0.00010877, -0.00004505, 0.00005052, 0.00014627, 0.00020304, 0.00021613, 0.00020651, 0.00017710, 0.00011886, 0.00005848, 0.00004430, 0.00007315, 0.00009254, 0.00009386, 0.00012460, 0.00017386, 0.00015311, 0.00007015, 0.00009810, 0.00031565, 0.00047559, 0.00026781, -0.00020035, -0.00045705, -0.00028177, -0.00000311, -0.00000881, -0.00020093, -0.00021607, -0.00001836, 0.00009022, 0.00000194, -0.00004804, 0.00005462, 0.00007038, -0.00015335, -0.00032440, -0.00011286, 0.00028014, 0.00034177, -0.00002116, -0.00034338, -0.00025475, 0.00005656, 0.00018922, 0.00007566, -0.00004532, -0.00004037, -0.00000505, -0.00002493, -0.00004472, -0.00000922, 0.00003131, 0.00002004, -0.00001256, -0.00001593, -0.00000290, -0.00000830, -0.00002108, -0.00001249, 0.00000730, 0.00000927, -0.00000407, -0.00000934, -0.00000462, -0.00000522, -0.00001067, -0.00000762, 0.00000225, 0.00000456, -0.00000229, -0.00000652, -0.00000492, -0.00000562, -0.00001029, -0.00001107, -0.00000579, -0.00000152, -0.00000167, -0.00000173, 0.00000049, 0.00000100, -0.00000182, -0.00000391, -0.00000302, -0.00000210, -0.00000297, -0.00000313, -0.00000101, 0.00000092, 0.00000059, -0.00000046, -0.00000050, -0.00000055, -0.00000186, -0.00000326, -0.00000324, -0.00000248, -0.00000262, -0.00000384, -0.00000508, -0.00000584, -0.00000648, -0.00000717, },
		{ -0.00000596, -0.00000666, -0.00000734, -0.00000716, -0.00000638, -0.00000624, -0.00000688, -0.00000707, -0.00000670, -0.00000723, -0.00000834, -0.00000753, -0.00000542, -0.00000732, -0.00001554, -0.00002442, -0.00002801, -0.00002790, -0.00002728, -0.00002344, -0.00001539, -0.00001330, -0.00002839, -0.00005828, -0.00009410, -0.00013398, -0.00017196, -0.00018227, -0.00014045, -0.00005715, 0.00003543, 0.00012550, 0.00021543, 0.00027894, 0.00027531, 0.00021498, 0.00015552, 0.00011793, 0.00007369, 0.00003670, 0.00007465, 0.00017524, 0.00021393, 0.00015761, 0.00017859, 0.00038524, 0.00053273, 0.00027636, -0.00027692, -0.00058036, -0.00034565, 0.00005299, 0.00010669, -0.00013473, -0.00023514, -0.00008436, -0.00000062, -0.00011747, -0.00013525, 0.00012424, 0.00031561, 0.00006804, -0.00037267, -0.00042463, -0.00000043, 0.00036378, 0.00025821, -0.00009500, -0.00024383, -0.00011480, 0.00002570, 0.00002897, -0.00000227, 0.00002345, 0.00004483, 0.00000239, -0.00004744, -0.00004013, -0.00000576, 0.00000166, -0.00000660, 0.00000319, 0.00001724, 0.00000546, -0.00001901, -0.00002314, -0.00000887, -0.00000171, -0.00000485, -0.00000275, 0.00000393, 0.00000099, -0.00000991, -0.00001327, -0.00000663, -0.00000180, -0.00000240, -0.00000082, 0.00000376, 0.00000292, -0.00000490, -0.00001101, -0.00001077, -0.00000877, -0.00000814, -0.00000588, -0.00000108, 0.00000171, 0.00000027, -0.00000181, -0.00000183, -0.00000174, -0.00000313, -0.00000387, -0.00000246, -0.00000090, -0.00000107, -0.00000167, -0.00000106, -0.00000010, -0.00000028, -0.00000110, -0.00000131, -0.00000115, -0.00000178, -0.00000336, -0.00000486, -0.00000558, },
		{ -0.00000423, -0.00000508, -0.00000515, -0.00000561, -0.00000657, -0.00000695, -0.00000657, -0.00000666, -0.00000743, -0.00000740, -0.00000649, -0.00000692, -0.00000883, -0.00000875, -0.00000573, -0.00000471, -0.00000903, -0.00001441, -0.00001664, -0.00001960, -0.00002740, -0.00003398, -0.00003156, -0.00002416, -0.00002070, -0.00002103, -0.00002464, -0.00004521, -0.00009407, -0.00015347, -0.00019141, -0.00019907, -0.00018334, -0.00012859, -0.00001158, 0.00014082, 0.00026070, 0.00031923, 0.00034132, 0.00032682, 0.00023921, 0.00011057, 0.00005646, 0.00011066, 0.00015574, 0.00013757, 0.00022076, 0.00051437, 0.00072953, 0.00043759, -0.00027406, -0.00074643, -0.00054924, -0.00004908, 0.00012172, -0.00006925, -0.00014378, 0.00004462, 0.00009286, -0.00020942, -0.00043054, -0.00013592, 0.00037340, 0.00042784, -0.00005372, -0.00045228, -0.00031684, 0.00008599, 0.00024759, 0.00009565, -0.00006525, -0.00006841, -0.00002862, -0.00004847, -0.00006317, -0.00001031, 0.00004676, 0.00003750, -0.00000410, -0.00001673, -0.00000940, -0.00001739, -0.00002776, -0.00001155, 0.00001512, 0.00001727, -0.00000107, -0.00001078, -0.00000775, -0.00000907, -0.00001506, -0.00001123, 0.00000070, 0.00000446, -0.00000255, -0.00000762, -0.00000633, -0.00000626, -0.00000875, -0.00000609, 0.00000217, 0.00000651, 0.00000274, -0.00000288, -0.00000562, -0.00000796, -0.00001116, -0.00001139, -0.00000721, -0.00000297, -0.00000186, -0.00000180, -0.00000071, -0.00000021, -0.00000162, -0.00000294, -0.00000245, -0.00000161, -0.00000214, -0.00000303, -0.00000260, -0.00000131, -0.00000057, -0.00000044, -0.00000013, 0.00000022, -0.00000042, -0.00000229, },
		{ 0.00000069, -0.00000123, -0.00000373, -0.00000502, -0.00000488, -0.00000472, -0.00000522, -0.00000546, -0.00000529, -0.00000615, -0.00000811, -0.00000873, -0.00000717, -0.00000628, -0.00000750, -0.00000786, -0.00000597, -0.00000628, -0.00001132, -0.00001503, -0.00001199, -0.00000797, -0.00001155, -0.00002028, -0.00002683, -0.00003282, -0.00004147, -0.00004324, -0.00002892, -0.00001260, -0.00001825, -0.00004696, -0.00008462, -0.00013573, -0.00020972, -0.00027304, -0.00026432, -0.00017013, -0.00003539, 0.00011286, 0.00028244, 0.00043297, 0.00046955, 0.00038007, 0.00028138, 0.00024168, 0.00017515, 0.00005755, 0.00010878, 0.00048853, 0.00086659, 0.00067257, -0.00010709, -0.00074517, -0.00065603, -0.00016509, -0.00000360, -0.00024242, -0.00027325, 0.00014718, 0.00044146, 0.00008951, -0.00051357, -0.00056851, -0.00000094, 0.00046043, 0.00030897, -0.00013771, -0.00030658, -0.00012885, 0.00004877, 0.00004559, -0.00000379, 0.00001700, 0.00003615, -0.00001605, -0.00007164, -0.00005542, -0.00000620, 0.00000841, -0.00000132, 0.00000485, 0.00001569, 0.00000109, -0.00002315, -0.00002240, -0.00000283, 0.00000488, -0.00000170, -0.00000246, 0.00000360, 0.00000040, -0.00001138, -0.00001511, -0.00000796, -0.00000308, -0.00000496, -0.00000520, -0.00000172, -0.00000219, -0.00000752, -0.00000896, -0.00000351, 0.00000152, 0.00000146, 0.00000009, 0.00000018, -0.00000136, -0.00000596, -0.00000949, -0.00000912, -0.00000742, -0.00000692, -0.00000610, -0.00000358, -0.00000131, -0.00000109, -0.00000158, -0.00000113, -0.00000068, -0.00000166, -0.00000323, -0.00000364, -0.00000280, -0.00000181, -0.00000105, -0.00000008, 0.00000088, },
		{ -0.00000104, 0.00000060, 0.00000140, 0.00000007, -0.00000273, -0.00000483, -0.00000531, -0.00000532, -0.00000547, -0.00000473, -0.00000322, -0.00000344, -0.00000625, -0.00000868, -0.00000856, -0.00000800, -0.00000850, -0.00000716, -0.00000318, -0.00000283, -0.00001020, -0.00001812, -0.00001767, -0.00001194, -0.00000928, -0.00000885, -0.00000678, -0.00001062, -0.00002882, -0.00005001, -0.00005480, -0.00004490, -0.00003596, -0.00002658, -0.00000896, -0.00000811, -0.00006474, -0.00016946, -0.00026196, -0.00030760, -0.00031630, -0.00026914, -0.00011060, 0.00013959, 0.00035785, 0.00047206, 0.00054526, 0.00060462, 0.00052403, 0.00027031, 0.00014022, 0.00042320, 0.00083622, 0.00070970, -0.00007978, -0.00075869, -0.00061086, 0.00000728, 0.00013765, -0.00039903, -0.00073291, -0.00023011, 0.00054577, 0.00061064, -0.00007760, -0.00062198, -0.00043705, 0.00008888, 0.00029342, 0.00010521, -0.00008235, -0.00007264, -0.00001976, -0.00004910, -0.00007489, -0.00001780, 0.00004630, 0.00003329, -0.00001553, -0.00002737, -0.00001433, -0.00002064, -0.00003360, -0.00001888, 0.00000869, 0.00001252, -0.00000261, -0.00000683, 0.00000053, -0.00000156, -0.00001198, -0.00001141, 0.00000033, 0.00000497, -0.00000174, -0.00000665, -0.00000509, -0.00000571, -0.00001045, -0.00001077, -0.00000491, -0.00000147, -0.00000385, -0.00000548, -0.00000285, -0.00000086, -0.00000245, -0.00000354, -0.00000148, 0.00000020, -0.00000137, -0.00000374, -0.00000450, -0.00000525, -0.00000744, -0.00000899, -0.00000792, -0.00000565, -0.00000422, -0.00000314, -0.00000140, 0.00000006, -0.00000026, -0.00000188, -0.00000308, -0.00000320, -0.00000280, -0.00000221, },
		{ -0.00000219, -0.00000239, -0.00000206, -0.00000065, 0.00000085, 0.00000064, -0.00000138, -0.00000334, -0.00000429, -0.00000537, -0.00000686, -0.00000667, -0.00000414, -0.00000230, -0.00000343, -0.00000537, -0.00000604, -0.00000750, -0.00001062, -0.00001065, -0.00000505, -0.00000085, -0.00000509, -0.00001287, -0.00001544, -0.00001490, -0.00001750, -0.00001818, -0.00000909, 0.00000092, -0.00000402, -0.00002066, -0.00003358, -0.00004415, -0.00006240, -0.00007252, -0.00004723, -0.00000195, 0.00001127, -0.00002500, -0.00008084, -0.00015068, -0.00026455, -0.00039359, -0.00043013, -0.00032358, -0.00014733, 0.00004768, 0.00032619, 0.00067823, 0.00087517, 0.00076526, 0.00061980, 0.00081186, 0.00111521, 0.00079734, -0.00030533, -0.00123242, -0.00096329, 0.00015813, 0.00072018, 0.00004972, -0.00091716, -0.00091667, 0.00000174, 0.00068319, 0.00045069, -0.00018474, -0.00042791, -0.00020447, 0.00002443, 0.00003238, -0.00000981, 0.00003223, 0.00006005, -0.00000968, -0.00008578, -0.00007214, -0.00001514, 0.00000312, -0.00000598, 0.00000448, 0.00001985, 0.00000272, -0.00002984, -0.00003605, -0.00001855, -0.00000866, -0.00000908, -0.00000075, 0.00001264, 0.00001040, -0.00000567, -0.00001374, -0.00000804, -0.00000277, -0.00000395, -0.00000321, 0.00000132, 0.00000075, -0.00000667, -0.00001149, -0.00000935, -0.00000646, -0.00000653, -0.00000531, -0.00000096, 0.00000135, -0.00000129, -0.00000454, -0.00000449, -0.00000310, -0.00000296, -0.00000277, -0.00000133, -0.00000082, -0.00000305, -0.00000616, -0.00000763, -0.00000770, -0.00000761, -0.00000695, -0.00000485, -0.00000227, -0.00000096, -0.00000121, -0.00000185, -0.00000209, },
		{ -0.00000305, -0.00000205, -0.00000114, -0.00000106, -0.00000140, -0.00000104, -0.00000017, -0.00000034, -0.00000175, -0.00000259, -0.00000226, -0.00000289, -0.00000549, -0.00000738, -0.00000650, -0.00000505, -0.00000519, -0.00000486, -0.00000263, -0.00000243, -0.00000684, -0.00001052, -0.00000831, -0.00000516, -0.00000776, -0.00001162, -0.00000924, -0.00000598, -0.00001173, -0.00002097, -0.00002034, -0.00001275, -0.00001156, -0.00001386, -0.00000771, -0.00000423, -0.00002648, -0.00006315, -0.00007527, -0.00005783, -0.00004519, -0.00004302, -0.00001913, 0.00001094, -0.00003229, -0.00017240, -0.00032259, -0.00041681, -0.00049121, -0.00053137, -0.00037072, 0.00006750, 0.00054262, 0.00079920, 0.00102737, 0.00158358, 0.00215186, 0.00174244, 0.00000201, -0.00184215, -0.00207581, -0.00055906, 0.00097466, 0.00095183, -0.00026271, -0.00107185, -0.00069988, 0.00013038, 0.00043895, 0.00016941, -0.00010762, -0.00012045, -0.00006550, -0.00009750, -0.00011063, -0.00001995, 0.00006790, 0.00004919, -0.00001795, -0.00003880, -0.00002572, -0.00003190, -0.00004317, -0.00002034, 0.00001677, 0.00002221, -0.00000038, -0.00001476, -0.00001498, -0.00002002, -0.00002728, -0.00001795, 0.00000318, 0.00001215, 0.00000420, -0.00000348, -0.00000326, -0.00000410, -0.00000881, -0.00000853, -0.00000149, 0.00000247, -0.00000113, -0.00000551, -0.00000611, -0.00000676, -0.00000933, -0.00000908, -0.00000406, 0.00000038, 0.00000003, -0.00000244, -0.00000345, -0.00000384, -0.00000500, -0.00000519, -0.00000304, -0.00000072, -0.00000072, -0.00000239, -0.00000384, -0.00000499, -0.00000655, -0.00000782, -0.00000752, -0.00000590, -0.00000436, -0.00000362, },
		{ -0.00000615, -0.00000648, -0.00000643, -0.00000508, -0.00000284, -0.00000115, -0.00000042, 0.00000021, 0.00000068, -0.00000045, -0.00000301, -0.00000421, -0.00000276, -0.00000139, -0.00000248, -0.00000420, -0.00000451, -0.00000543, -0.00000842, -0.00000962, -0.00000594, -0.00000186, -0.00000246, -0.00000442, -0.00000315, -0.00000351, -0.00001131, -0.00001897, -0.00001520, -0.00000517, -0.00000316, -0.00000863, -0.00001041, -0.00001155, -0.00002328, -0.00003546, -0.00002595, -0.00000258, 0.00000219, -0.00001516, -0.00002619, -0.00003098, -0.00006298, -0.00011103, -0.00010773, -0.00003803, 0.00001681, 0.00000056, -0.00004385, -0.00010362, -0.00027847, -0.00058360, -0.00078822, -0.00067983, -0.00039248, -0.00013281, 0.00037502, 0.00161700, 0.00319696, 0.00359533, 0.00174674, -0.00128760, -0.00301609, -0.00215107, 0.00000871, 0.00119205, 0.00066820, -0.00039566, -0.00073648, -0.00035307, 0.00001791, 0.00005204, -0.00000556, 0.00003411, 0.00005457, -0.00003770, -0.00012376, -0.00009331, -0.00001571, 0.00000550, -0.00001085, -0.00000286, 0.00001381, -0.00000537, -0.00004017, -0.00004162, -0.00001475, 0.00000029, -0.00000273, -0.00000134, 0.00000460, -0.00000325, -0.00002047, -0.00002412, -0.00001174, -0.00000244, -0.00000327, -0.00000299, 0.00000229, 0.00000294, -0.00000383, -0.00000818, -0.00000529, -0.00000203, -0.00000297, -0.00000384, -0.00000209, -0.00000209, -0.00000586, -0.00000840, -0.00000631, -0.00000311, -0.00000263, -0.00000337, -0.00000300, -0.00000272, -0.00000405, -0.00000530, -0.00000448, -0.00000271, -0.00000205, -0.00000236, -0.00000241, -0.00000241, -0.00000333, -0.00000494, -0.00000602, -0.00000617, },
		{ -0.00000352, -0.00000369, -0.00000419, -0.00000570, -0.00000743, -0.00000775, -0.00000628, -0.00000428, -0.00000249, -0.00000029, 0.00000179, 0.00000138, -0.00000199, -0.00000487, -0.00000464, -0.00000325, -0.00000312, -0.00000266, -0.00000053, -0.00000040, -0.00000516, -0.00001036, -0.00001047, -0.00000796, -0.00000732, -0.00000534, 0.00000165, 0.00000565, -0.00000291, -0.00001684, -0.00002076, -0.00001511, -0.00001204, -0.00001065, -0.00000120, 0.00000735, -0.00000490, -0.00003077, -0.00004045, -0.00003018, -0.00002512, -0.00002558, -0.00000354, 0.00002875, 0.00001201, -0.00006257, -0.00011989, -0.00011489, -0.00009774, -0.00009580, -0.00003773, 0.00008458, 0.00009367, -0.00014455, -0.00046876, -0.00066311, -0.00085418, -0.00122158, -0.00125512, 0.00004599, 0.00268994, 0.00483099, 0.00428589, 0.00107299, -0.00224382, -0.00314577, -0.00165471, 0.00016156, 0.00071681, 0.00024447, -0.00020687, -0.00023022, -0.00014156, -0.00017072, -0.00017786, -0.00005420, 0.00006247, 0.00004633, -0.00002597, -0.00004205, -0.00002454, -0.00003853, -0.00006063, -0.00003838, 0.00000530, 0.00001430, -0.00000792, -0.00001883, -0.00001249, -0.00001383, -0.00002284, -0.00001803, -0.00000150, 0.00000394, -0.00000492, -0.00001006, -0.00000628, -0.00000615, -0.00001261, -0.00001364, -0.00000537, 0.00000129, 0.00000011, -0.00000246, -0.00000167, -0.00000175, -0.00000507, -0.00000668, -0.00000387, -0.00000126, -0.00000235, -0.00000404, -0.00000328, -0.00000235, -0.00000378, -0.00000567, -0.00000551, -0.00000439, -0.00000434, -0.00000469, -0.00000395, -0.00000277, -0.00000265, -0.00000332, -0.00000343, -0.00000285, -0.00000259, -0.00000302, },
		{ -0.00000315, -0.00000333, -0.00000339, -0.00000284, -0.00000235, -0.00000317, -0.00000515, -0.00000660, -0.00000668, -0.00000632, -0.00000585, -0.00000390, -0.00000055, 0.00000112, -0.00000066, -0.00000315, -0.00000362, -0.00000377, -0.00000547, -0.00000570, -0.00000182, 0.00000188, 0.00000023, -0.00000411, -0.00000600, -0.00000763, -0.00001247, -0.00001434, -0.00000612, 0.00000436, 0.00000352, -0.00000585, -0.00001027, -0.00001046, -0.00001741, -0.00002515, -0.00001622, 0.00000323, 0.00000693, -0.00000744, -0.00001566, -0.00001641, -0.00003501, -0.00006252, -0.00005031, 0.00000448, 0.00003437, 0.00000612, -0.00002611, -0.00003293, -0.00007945, -0.00018370, -0.00021254, -0.00007633, 0.00007240, 0.00005490, -0.00001869, 0.00000363, -0.00020169, -0.00112823, -0.00232624, -0.00219772, 0.00029669, 0.00383989, 0.00553042, 0.00383989, 0.00029669, -0.00219772, -0.00232624, -0.00112823, -0.00020169, 0.00000363, -0.00001869, 0.00005490, 0.00007240, -0.00007633, -0.00021254, -0.00018370, -0.00007945, -0.00003293, -0.00002611, 0.00000612, 0.00003437, 0.00000448, -0.00005031, -0.00006252, -0.00003501, -0.00001641, -0.00001566, -0.00000744, 0.00000693, 0.00000323, -0.00001622, -0.00002515, -0.00001741, -0.00001046, -0.00001027, -0.00000585, 0.00000352, 0.00000436, -0.00000612, -0.00001434, -0.00001247, -0.00000763, -0.00000600, -0.00000411, 0.00000023, 0.00000188, -0.00000182, -0.00000570, -0.00000547, -0.00000377, -0.00000362, -0.00000315, -0.00000066, 0.00000112, -0.00000055, -0.00000390, -0.00000585, -0.00000632, -0.00000668, -0.00000660, -0.00000515, -0.00000317, -0.00000235, -0.00000284, -0.00000339, -0.00000333, },
		{ -0.00000352, -0.00000302, -0.00000259, -0.00000285, -0.00000343, -0.00000332, -0.00000265, -0.00000277, -0.00000395, -0.00000469, -0.00000434, -0.00000439, -0.00000551, -0.00000567, -0.00000378, -0.00000235, -0.00000328, -0.00000404, -0.00000235, -0.00000126, -0.00000387, -0.00000668, -0.00000507, -0.00000175, -0.00000167, -0.00000246, 0.00000011, 0.00000129, -0.00000537, -0.00001364, -0.00001261, -0.00000615, -0.00000628, -0.00001006, -0.00000492, 0.00000394, -0.00000150, -0.00001803, -0.00002284, -0.00001383, -0.00001249, -0.00001883, -0.00000792, 0.00001430, 0.00000530, -0.00003838, -0.00006063, -0.00003853, -0.00002454, -0.00004205, -0.00002597, 0.00004633, 0.00006247, -0.00005420, -0.00017786, -0.00017072, -0.00014156, -0.00023022, -0.00020687, 0.00024447, 0.00071681, 0.00016156, -0.00165471, -0.00314577, -0.00224382, 0.00107299, 0.00428589, 0.00483099, 0.00268994, 0.00004599, -0.00125512, -0.00122158, -0.00085418, -0.00066311, -0.00046876, -0.00014455, 0.00009367, 0.00008458, -0.00003773, -0.00009580, -0.00009774, -0.00011489, -0.00011989, -0.00006257, 0.00001201, 0.00002875, -0.00000354, -0.00002558, -0.00002512, -0.00003018, -0.00004045, -0.00003077, -0.00000490, 0.00000735, -0.00000120, -0.00001065, -0.00001204, -0.00001511, -0.00002076, -0.00001684, -0.00000291, 0.00000565, 0.00000165, -0.00000534, -0.00000732, -0.00000796, -0.00001047, -0.00001036, -0.00000516, -0.00000040, -0.00000053, -0.00000266, -0.00000312, -0.00000325, -0.00000464, -0.00000487, -0.00000199, 0.00000138, 0.00000179, -0.00000029, -0.00000249, -0.00000428, -0.00000628, -0.00000775, -0.00000743, -0.00000570, -0.00000419, -0.00000369, },
		{ -0.00000615, -0.00000617, -0.00000602, -0.00000494, -0.00000333, -0.00000241, -0.00000241, -0.00000236, -0.00000205, -0.00000271, -0.00000448, -0.00000530, -0.00000405, -0.00000272, -0.00000300, -0.00000337, -0.00000263, -0.00000311, -0.00000631, -0.00000840, -0.00000586, -0.00000209, -0.00000209, -0.00000384, -0.00000297, -0.00000203, -0.00000529, -0.00000818, -0.00000383, 0.00000294, 0.00000229, -0.00000299, -0.00000327, -0.00000244, -0.00001174, -0.00002412, -0.00002047, -0.00000325, 0.00000460, -0.00000134, -0.00000273, 0.00000029, -0.00001475, -0.00004162, -0.00004017, -0.00000537, 0.00001381, -0.00000286, -0.00001085, 0.00000550, -0.00001571, -0.00009331, -0.00012376, -0.00003770, 0.00005457, 0.00003411, -0.00000556, 0.00005204, 0.00001791, -0.00035307, -0.00073648, -0.00039566, 0.00066820, 0.00119205, 0.00000871, -0.00215107, -0.00301609, -0.00128760, 0.00174674, 0.00359533, 0.00319696, 0.00161700, 0.00037502, -0.00013281, -0.00039248, -0.00067983, -0.00078822, -0.00058360, -0.00027847, -0.00010362, -0.00004385, 0.00000056, 0.00001681, -0.00003803, -0.00010773, -0.00011103, -0.00006298, -0.00003098, -0.00002619, -0.00001516, 0.00000219, -0.00000258, -0.00002595, -0.00003546, -0.00002328, -0.00001155, -0.00001041, -0.00000863, -0.00000316, -0.00000517, -0.00001520, -0.00001897, -0.00001131, -0.00000351, -0.00000315, -0.00000442, -0.00000246, -0.00000186, -0.00000594, -0.00000962, -0.00000842, -0.00000543, -0.00000451, -0.00000420, -0.00000248, -0.00000139, -0.00000276, -0.00000421, -0.00000301, -0.00000045, 0.00000068, 0.00000021, -0.00000042, -0.00000115, -0.00000284, -0.00000508, -0.00000643, -0.00000648, },
		{ -0.00000305, -0.00000362, -0.00000436, -0.00000590, -0.00000752, -0.00000782, -0.00000655, -0.00000499, -0.00000384, -0.00000239, -0.00000072, -0.00000072, -0.00000304, -0.00000519, -0.00000500, -0.00000384, -0.00000345, -0.00000244, 0.00000003, 0.00000038, -0.00000406, -0.00000908, -0.00000933, -0.00000676, -0.00000611, -0.00000551, -0.00000113, 0.00000247, -0.00000149, -0.00000853, -0.00000881, -0.00000410, -0.00000326, -0.00000348, 0.00000420, 0.00001215, 0.00000318, -0.00001795, -0.00002728, -0.00002002, -0.00001498, -0.00001476, -0.00000038, 0.00002221, 0.00001677, -0.00002034, -0.00004317, -0.00003190, -0.00002572, -0.00003880, -0.00001795, 0.00004919, 0.00006790, -0.00001995, -0.00011063, -0.00009750, -0.00006550, -0.00012045, -0.00010762, 0.00016941, 0.00043895, 0.00013038, -0.00069988, -0.00107185, -0.00026271, 0.00095183, 0.00097466, -0.00055906, -0.00207581, -0.00184215, 0.00000201, 0.00174244, 0.00215186, 0.00158358, 0.00102737, 0.00079920, 0.00054262, 0.00006750, -0.00037072, -0.00053137, -0.00049121, -0.00041681, -0.00032259, -0.00017240, -0.00003229, 0.00001094, -0.00001913, -0.00004302, -0.00004519, -0.00005783, -0.00007527, -0.00006315, -0.00002648, -0.00000423, -0.00000771, -0.00001386, -0.00001156, -0.00001275, -0.00002034, -0.00002097, -0.00001173, -0.00000598, -0.00000924, -0.00001162, -0.00000776, -0.00000516, -0.00000831, -0.00001052, -0.00000684, -0.00000243, -0.00000263, -0.00000486, -0.00000519, -0.00000505, -0.00000650, -0.00000738, -0.00000549, -0.00000289, -0.00000226, -0.00000259, -0.00000175, -0.00000034, -0.00000017, -0.00000104, -0.00000140, -0.00000106, -0.00000114, -0.00000205, },
		{ -0.00000219, -0.00000209, -0.00000185, -0.00000121, -0.00000096, -0.00000227, -0.00000485, -0.00000695, -0.00000761, -0.00000770, -0.00000763, -0.00000616, -0.00000305, -0.00000082, -0.00000133, -0.00000277, -0.00000296, -0.00000310, -0.00000449, -0.00000454, -0.00000129, 0.00000135, -0.00000096, -0.00000531, -0.00000653, -0.00000646, -0.00000935, -0.00001149, -0.00000667, 0.00000075, 0.00000132, -0.00000321, -0.00000395, -0.00000277, -0.00000804, -0.00001374, -0.00000567, 0.00001040, 0.00001264, -0.00000075, -0.00000908, -0.00000866, -0.00001855, -0.00003605, -0.00002984, 0.00000272, 0.00001985, 0.00000448, -0.00000598, 0.00000312, -0.00001514, -0.00007214, -0.00008578, -0.00000968, 0.00006005, 0.00003223, -0.00000981, 0.00003238, 0.00002443, -0.00020447, -0.00042791, -0.00018474, 0.00045069, 0.00068319, 0.00000174, -0.00091667, -0.00091716, 0.00004972, 0.00072018, 0.00015813, -0.00096329, -0.00123242, -0.00030533, 0.00079734, 0.00111521, 0.00081186, 0.00061980, 0.00076526, 0.00087517, 0.00067823, 0.00032619, 0.00004768, -0.00014733, -0.00032358, -0.00043013, -0.00039359, -0.00026455, -0.00015068, -0.00008084, -0.00002500, 0.00001127, -0.00000195, -0.00004723, -0.00007252, -0.00006240, -0.00004415, -0.00003358, -0.00002066, -0.00000402, 0.00000092, -0.00000909, -0.00001818, -0.00001750, -0.00001490, -0.00001544, -0.00001287, -0.00000509, -0.00000085, -0.00000505, -0.00001065, -0.00001062, -0.00000750, -0.00000604, -0.00000537, -0.00000343, -0.00000230, -0.00000414, -0.00000667, -0.00000686, -0.00000537, -0.00000429, -0.00000334, -0.00000138, 0.00000064, 0.00000085, -0.00000065, -0.00000206, -0.00000239, },
		{ -0.00000104, -0.00000221, -0.00000280, -0.00000320, -0.00000308, -0.00000188, -0.00000026, 0.00000006, -0.00000140, -0.00000314, -0.00000422, -0.00000565, -0.00000792, -0.00000899, -0.00000744, -0.00000525, -0.00000450, -0.00000374, -0.00000137, 0.00000020, -0.00000148, -0.00000354, -0.00000245, -0.00000086, -0.00000285, -0.00000548, -0.00000385, -0.00000147, -0.00000491, -0.00001077, -0.00001045, -0.00000571, -0.00000509, -0.00000665, -0.00000174, 0.00000497, 0.00000033, -0.00001141, -0.00001198, -0.00000156, 0.00000053, -0.00000683, -0.00000261, 0.00001252, 0.00000869, -0.00001888, -0.00003360, -0.00002064, -0.00001433, -0.00002737, -0.00001553, 0.00003329, 0.00004630, -0.00001780, -0.00007489, -0.00004910, -0.00001976, -0.00007264, -0.00008235, 0.00010521, 0.00029342, 0.00008888, -0.00043705, -0.00062198, -0.00007760, 0.00061064, 0.00054577, -0.00023011, -0.00073291, -0.00039903, 0.00013765, 0.00000728, -0.00061086, -0.00075869, -0.00007978, 0.00070970, 0.00083622, 0.00042320, 0.00014022, 0.00027031, 0.00052403, 0.00060462, 0.00054526, 0.00047206, 0.00035785, 0.00013959, -0.00011060, -0.00026914, -0.00031630, -0.00030760, -0.00026196, -0.00016946, -0.00006474, -0.00000811, -0.00000896, -0.00002658, -0.00003596, -0.00004490, -0.00005480, -0.00005001, -0.00002882, -0.00001062, -0.00000678, -0.00000885, -0.00000928, -0.00001194, -0.00001767, -0.00001812, -0.00001020, -0.00000283, -0.00000318, -0.00000716, -0.00000850, -0.00000800, -0.00000856, -0.00000868, -0.00000625, -0.00000344, -0.00000322, -0.00000473, -0.00000547, -0.00000532, -0.00000531, -0.00000483, -0.00000273, 0.00000007, 0.00000140, 0.00000060, },
		{ 0.00000069, 0.00000088, -0.00000008, -0.00000105, -0.00000181, -0.00000280, -0.00000364, -0.00000323, -0.00000166, -0.00000068, -0.00000113, -0.00000158, -0.00000109, -0.00000131, -0.00000358, -0.00000610, -0.00000692, -0.00000742, -0.00000912, -0.00000949, -0.00000596, -0.00000136, 0.00000018, 0.00000009, 0.00000146, 0.00000152, -0.00000351, -0.00000896, -0.00000752, -0.00000219, -0.00000172, -0.00000520, -0.00000496, -0.00000308, -0.00000796, -0.00001511, -0.00001138, 0.00000040, 0.00000360, -0.00000246, -0.00000170, 0.00000488, -0.00000283, -0.00002240, -0.00002315, 0.00000109, 0.00001569, 0.00000485, -0.00000132, 0.00000841, -0.00000620, -0.00005542, -0.00007164, -0.00001605, 0.00003615, 0.00001700, -0.00000379, 0.00004559, 0.00004877, -0.00012885, -0.00030658, -0.00013771, 0.00030897, 0.00046043, -0.00000094, -0.00056851, -0.00051357, 0.00008951, 0.00044146, 0.00014718, -0.00027325, -0.00024242, -0.00000360, -0.00016509, -0.00065603, -0.00074517, -0.00010709, 0.00067257, 0.00086659, 0.00048853, 0.00010878, 0.00005755, 0.00017515, 0.00024168, 0.00028138, 0.00038007, 0.00046955, 0.00043297, 0.00028244, 0.00011286, -0.00003539, -0.00017013, -0.00026432, -0.00027304, -0.00020972, -0.00013573, -0.00008462, -0.00004696, -0.00001825, -0.00001260, -0.00002892, -0.00004324, -0.00004147, -0.00003282, -0.00002683, -0.00002028, -0.00001155, -0.00000797, -0.00001199, -0.00001503, -0.00001132, -0.00000628, -0.00000597, -0.00000786, -0.00000750, -0.00000628, -0.00000717, -0.00000873, -0.00000811, -0.00000615, -0.00000529, -0.00000546, -0.00000522, -0.00000472, -0.00000488, -0.00000502, -0.00000373, -0.00000123, },
		{ -0.00000423, -0.00000229, -0.00000042, 0.00000022, -0.00000013, -0.00000044, -0.00000057, -0.00000131, -0.00000260, -0.00000303, -0.00000214, -0.00000161, -0.00000245, -0.00000294, -0.00000162, -0.00000021, -0.00000071, -0.00000180, -0.00000186, -0.00000297, -0.00000721, -0.00001139, -0.00001116, -0.00000796, -0.00000562, -0.00000288, 0.00000274, 0.00000651, 0.00000217, -0.00000609, -0.00000875, -0.00000626, -0.00000633, -0.00000762, -0.00000255, 0.00000446, 0.00000070, -0.00001123, -0.00001506, -0.00000907, -0.00000775, -0.00001078, -0.00000107, 0.00001727, 0.00001512, -0.00001155, -0.00002776, -0.00001739, -0.00000940, -0.00001673, -0.00000410, 0.00003750, 0.00004676, -0.00001031, -0.00006317, -0.00004847, -0.00002862, -0.00006841, -0.00006525, 0.00009565, 0.00024759, 0.00008599, -0.00031684, -0.00045228, -0.00005372, 0.00042784, 0.00037340, -0.00013592, -0.00043054, -0.00020942, 0.00009286, 0.00004462, -0.00014378, -0.00006925, 0.00012172, -0.00004908, -0.00054924, -0.00074643, -0.00027406, 0.00043759, 0.00072953, 0.00051437, 0.00022076, 0.00013757, 0.00015574, 0.00011066, 0.00005646, 0.00011057, 0.00023921, 0.00032682, 0.00034132, 0.00031923, 0.00026070, 0.00014082, -0.00001158, -0.00012859, -0.00018334, -0.00019907, -0.00019141, -0.00015347, -0.00009407, -0.00004521, -0.00002464, -0.00002103, -0.00002070, -0.00002416, -0.00003156, -0.00003398, -0.00002740, -0.00001960, -0.00001664, -0.00001441, -0.00000903, -0.00000471, -0.00000573, -0.00000875, -0.00000883, -0.00000692, -0.00000649, -0.00000740, -0.00000743, -0.00000666, -0.00000657, -0.00000695, -0.00000657, -0.00000561, -0.00000515, -0.00000508, },
		{ -0.00000596, -0.00000558, -0.00000486, -0.00000336, -0.00000178, -0.00000115, -0.00000131, -0.00000110, -0.00000028, -0.00000010, -0.00000106, -0.00000167, -0.00000107, -0.00000090, -0.00000246, -0.00000387, -0.00000313, -0.00000174, -0.00000183, -0.00000181, 0.00000027, 0.00000171, -0.00000108, -0.00000588, -0.00000814, -0.00000877, -0.00001077, -0.00001101, -0.00000490, 0.00000292, 0.00000376, -0.00000082, -0.00000240, -0.00000180, -0.00000663, -0.00001327, -0.00000991, 0.00000099, 0.00000393, -0.00000275, -0.00000485, -0.00000171, -0.00000887, -0.00002314, -0.00001901, 0.00000546, 0.00001724, 0.00000319, -0.00000660, 0.00000166, -0.00000576, -0.00004013, -0.00004744, 0.00000239, 0.00004483, 0.00002345, -0.00000227, 0.00002897, 0.00002570, -0.00011480, -0.00024383, -0.00009500, 0.00025821, 0.00036378, -0.00000043, -0.00042463, -0.00037267, 0.00006804, 0.00031561, 0.00012424, -0.00013525, -0.00011747, -0.00000062, -0.00008436, -0.00023514, -0.00013473, 0.00010669, 0.00005299, -0.00034565, -0.00058036, -0.00027692, 0.00027636, 0.00053273, 0.00038524, 0.00017859, 0.00015761, 0.00021393, 0.00017524, 0.00007465, 0.00003670, 0.00007369, 0.00011793, 0.00015552, 0.00021498, 0.00027531, 0.00027894, 0.00021543, 0.00012550, 0.00003543, -0.00005715, -0.00014045, -0.00018227, -0.00017196, -0.00013398, -0.00009410, -0.00005828, -0.00002839, -0.00001330, -0.00001539, -0.00002344, -0.00002728, -0.00002790, -0.00002801, -0.00002442, -0.00001554, -0.00000732, -0.00000542, -0.00000753, -0.00000834, -0.00000723, -0.00000670, -0.00000707, -0.00000688, -0.00000624, -0.00000638, -0.00000716, -0.00000734, -0.00000666, },
		{ -0.00000738, -0.00000717, -0.00000648, -0.00000584, -0.00000508, -0.00000384, -0.00000262, -0.00000248, -0.00000324, -0.00000326, -0.00000186, -0.00000055, -0.00000050, -0.00000046, 0.00000059, 0.00000092, -0.00000101, -0.00000313, -0.00000297, -0.00000210, -0.00000302, -0.00000391, -0.00000182, 0.00000100, 0.00000049, -0.00000173, -0.00000167, -0.00000152, -0.00000579, -0.00001107, -0.00001029, -0.00000562, -0.00000492, -0.00000652, -0.00000229, 0.00000456, 0.00000225, -0.00000762, -0.00001067, -0.00000522, -0.00000462, -0.00000934, -0.00000407, 0.00000927, 0.00000730, -0.00001249, -0.00002108, -0.00000830, -0.00000290, -0.00001593, -0.00001256, 0.00002004, 0.00003131, -0.00000922, -0.00004472, -0.00002493, -0.00000505, -0.00004037, -0.00004532, 0.00007566, 0.00018922, 0.00005656, -0.00025475, -0.00034338, -0.00002116, 0.00034177, 0.00028014, -0.00011286, -0.00032440, -0.00015335, 0.00007038, 0.00005462, -0.00004804, 0.00000194, 0.00009022, -0.00001836, -0.00021607, -0.00020093, -0.00000881, -0.00000311, -0.00028177, -0.00045705, -0.00020035, 0.00026781, 0.00047559, 0.00031565, 0.00009810, 0.00007015, 0.00015311, 0.00017386, 0.00012460, 0.00009386, 0.00009254, 0.00007315, 0.00004430, 0.00005848, 0.00011886, 0.00017710, 0.00020651, 0.00021613, 0.00020304, 0.00014627, 0.00005052, -0.00004505, -0.00010877, -0.00013948, -0.00014529, -0.00012705, -0.00009053, -0.00005364, -0.00003049, -0.00002010, -0.00001657, -0.00001878, -0.00002495, -0.00002800, -0.00002353, -0.00001602, -0.00001183, -0.00001098, -0.00000980, -0.00000766, -0.00000645, -0.00000646, -0.00000631, -0.00000577, -0.00000586, -0.00000673, },
		{ -0.00000487, -0.00000554, -0.00000663, -0.00000717, -0.00000692, -0.00000630, -0.00000537, -0.00000384, -0.00000232, -0.00000228, -0.00000391, -0.00000516, -0.00000451, -0.00000315, -0.00000244, -0.00000134, 0.00000111, 0.00000264, 0.00000107, -0.00000142, -0.00000162, -0.00000086, -0.00000218, -0.00000396, -0.00000287, -0.00000089, -0.00000205, -0.00000403, -0.00000173, 0.00000233, 0.00000152, -0.00000271, -0.00000368, -0.00000331, -0.00000896, -0.00001660, -0.00001413, -0.00000298, 0.00000230, -0.00000095, -0.00000067, 0.00000352, -0.00000280, -0.00001758, -0.00001800, -0.00000038, 0.00000886, -0.00000029, -0.00000271, 0.00000962, 0.00000313, -0.00003268, -0.00004747, -0.00001033, 0.00002592, 0.00001257, -0.00000180, 0.00003217, 0.00003519, -0.00007962, -0.00018859, -0.00007296, 0.00020578, 0.00028517, -0.00000194, -0.00032334, -0.00026995, 0.00006786, 0.00023940, 0.00008059, -0.00011276, -0.00009359, -0.00000256, -0.00004410, -0.00012048, -0.00004448, 0.00008901, 0.00005159, -0.00010964, -0.00013860, -0.00002283, -0.00004424, -0.00029578, -0.00045285, -0.00021523, 0.00023771, 0.00046930, 0.00034266, 0.00011761, 0.00004158, 0.00008226, 0.00009980, 0.00007954, 0.00008727, 0.00011862, 0.00011426, 0.00007002, 0.00003747, 0.00003968, 0.00005665, 0.00008033, 0.00012363, 0.00017639, 0.00020040, 0.00017539, 0.00011823, 0.00005226, -0.00001436, -0.00007476, -0.00011320, -0.00012064, -0.00010579, -0.00008277, -0.00005836, -0.00003630, -0.00002266, -0.00001957, -0.00002082, -0.00001979, -0.00001706, -0.00001620, -0.00001671, -0.00001553, -0.00001232, -0.00000935, -0.00000768, -0.00000649, -0.00000532, },
		{ -0.00001021, -0.00000782, -0.00000572, -0.00000485, -0.00000520, -0.00000586, -0.00000620, -0.00000633, -0.00000607, -0.00000470, -0.00000253, -0.00000154, -0.00000281, -0.00000464, -0.00000514, -0.00000510, -0.00000557, -0.00000487, -0.00000151, 0.00000186, 0.00000191, 0.00000020, 0.00000025, 0.00000103, -0.00000057, -0.00000273, -0.00000168, 0.00000024, -0.00000204, -0.00000616, -0.00000551, -0.00000119, 0.00000017, -0.00000040, 0.00000332, 0.00000744, 0.00000123, -0.00001244, -0.00001844, -0.00001389, -0.00001111, -0.00001186, -0.00000345, 0.00001180, 0.00001211, -0.00000595, -0.00001678, -0.00000963, -0.00000700, -0.00001655, -0.00000841, 0.00002508, 0.00003567, -0.00000422, -0.00004226, -0.00003062, -0.00001445, -0.00004159, -0.00004115, 0.00006540, 0.00016375, 0.00005436, -0.00020376, -0.00027815, -0.00001793, 0.00027385, 0.00022812, -0.00007314, -0.00022970, -0.00010025, 0.00005508, 0.00002860, -0.00005006, -0.00000834, 0.00006099, -0.00000291, -0.00011799, -0.00009945, 0.00000856, 0.00001073, -0.00008716, -0.00008761, 0.00002406, 0.00000447, -0.00024153, -0.00042964, -0.00026150, 0.00014281, 0.00039456, 0.00033259, 0.00015437, 0.00007868, 0.00009211, 0.00008068, 0.00003812, 0.00003408, 0.00007347, 0.00009820, 0.00008727, 0.00007046, 0.00006168, 0.00004602, 0.00002720, 0.00003366, 0.00007277, 0.00011722, 0.00014269, 0.00014933, 0.00014126, 0.00011181, 0.00005948, -0.00000019, -0.00004890, -0.00008037, -0.00009533, -0.00009359, -0.00007796, -0.00005782, -0.00004153, -0.00002966, -0.00001983, -0.00001327, -0.00001229, -0.00001501, -0.00001694, -0.00001620, -0.00001416, -0.00001222, },
		{ -0.00001357, -0.00001372, -0.00001366, -0.00001242, -0.00001010, -0.00000786, -0.00000645, -0.00000552, -0.00000475, -0.00000466, -0.00000525, -0.00000513, -0.00000361, -0.00000232, -0.00000274, -0.00000366, -0.00000356, -0.00000381, -0.00000582, -0.00000729, -0.00000536, -0.00000198, -0.00000085, -0.00000099, 0.00000080, 0.00000289, 0.00000149, -0.00000132, -0.00000052, 0.00000202, 0.00000030, -0.00000428, -0.00000496, -0.00000259, -0.00000378, -0.00000617, -0.00000091, 0.00000880, 0.00000932, -0.00000032, -0.00000646, -0.00000645, -0.00001276, -0.00002440, -0.00002179, -0.00000229, 0.00000941, 0.00000284, -0.00000042, 0.00000796, 0.00000093, -0.00002866, -0.00003697, -0.00000024, 0.00003147, 0.00001477, -0.00000508, 0.00001915, 0.00002168, -0.00007237, -0.00015975, -0.00005833, 0.00017634, 0.00023971, -0.00000283, -0.00027071, -0.00022751, 0.00004964, 0.00019635, 0.00008378, -0.00005793, -0.00004654, 0.00000591, -0.00004298, -0.00010499, -0.00004427, 0.00005954, 0.00004556, -0.00004791, -0.00006184, -0.00000473, -0.00002550, -0.00012027, -0.00011615, 0.00000849, 0.00003192, -0.00016240, -0.00034678, -0.00024028, 0.00008868, 0.00031172, 0.00027218, 0.00012967, 0.00007780, 0.00010620, 0.00010450, 0.00005707, 0.00002973, 0.00004206, 0.00005475, 0.00005164, 0.00005482, 0.00006826, 0.00006752, 0.00004669, 0.00003053, 0.00003488, 0.00004880, 0.00006175, 0.00007970, 0.00010542, 0.00012448, 0.00012161, 0.00009715, 0.00006023, 0.00001721, -0.00002593, -0.00005954, -0.00007672, -0.00007927, -0.00007271, -0.00005977, -0.00004249, -0.00002620, -0.00001640, -0.00001350, -0.00001362, -0.00001372, },
		{ -0.00002079, -0.00001540, -0.00001173, -0.00001058, -0.00001139, -0.00001236, -0.00001233, -0.00001167, -0.00001083, -0.00000926, -0.00000669, -0.00000446, -0.00000372, -0.00000354, -0.00000280, -0.00000258, -0.00000388, -0.00000486, -0.00000365, -0.00000227, -0.00000351, -0.00000571, -0.00000555, -0.00000419, -0.00000477, -0.00000558, -0.00000270, 0.00000174, 0.00000237, 0.00000028, 0.00000098, 0.00000338, 0.00000155, -0.00000288, -0.00000256, 0.00000096, -0.00000110, -0.00000717, -0.00000610, 0.00000182, 0.00000365, -0.00000173, -0.00000053, 0.00000768, 0.00000445, -0.00001341, -0.00002317, -0.00001528, -0.00001055, -0.00001732, -0.00001005, 0.00001873, 0.00002857, -0.00000295, -0.00003029, -0.00001494, -0.00000046, -0.00002970, -0.00003797, 0.00004736, 0.00013069, 0.00004130, -0.00017153, -0.00022843, -0.00000731, 0.00023328, 0.00018754, -0.00006932, -0.00020243, -0.00009566, 0.00003987, 0.00003650, -0.00000979, 0.00002443, 0.00006330, -0.00000356, -0.00009844, -0.00008330, 0.00000101, 0.00001465, -0.00003533, -0.00002841, 0.00002832, 0.00000394, -0.00010342, -0.00013194, -0.00003623, 0.00000045, -0.00014391, -0.00029346, -0.00020085, 0.00008874, 0.00028793, 0.00024642, 0.00010266, 0.00004339, 0.00007400, 0.00009264, 0.00006935, 0.00005153, 0.00005618, 0.00005295, 0.00003351, 0.00002520, 0.00003804, 0.00005001, 0.00004809, 0.00004478, 0.00004812, 0.00004731, 0.00003682, 0.00003058, 0.00004186, 0.00006522, 0.00008682, 0.00009972, 0.00010181, 0.00008922, 0.00006090, 0.00002383, -0.00001216, -0.00004149, -0.00006153, -0.00006933, -0.00006421, -0.00005115, -0.00003766, -0.00002772, },
		{ -0.00005037, -0.00004584, -0.00003919, -0.00003053, -0.00002149, -0.00001463, -0.00001104, -0.00000965, -0.00000921, -0.00000980, -0.00001143, -0.00001241, -0.00001116, -0.00000851, -0.00000613, -0.00000389, -0.00000143, -0.00000057, -0.00000274, -0.00000540, -0.00000525, -0.00000353, -0.00000347, -0.00000420, -0.00000300, -0.00000175, -0.00000421, -0.00000800, -0.00000737, -0.00000315, -0.00000114, -0.00000120, 0.00000163, 0.00000548, 0.00000338, -0.00000259, -0.00000289, 0.00000223, 0.00000252, -0.00000255, -0.00000224, 0.00000318, 0.00000014, -0.00001104, -0.00001147, 0.00000291, 0.00001057, 0.00000194, -0.00000357, 0.00000212, -0.00000499, -0.00003205, -0.00004083, -0.00000962, 0.00001932, 0.00000946, 0.00000002, 0.00002872, 0.00003348, -0.00005248, -0.00013651, -0.00005648, 0.00014215, 0.00019835, -0.00000266, -0.00022293, -0.00018190, 0.00004981, 0.00016579, 0.00006333, -0.00006157, -0.00005565, -0.00000728, -0.00003081, -0.00006295, -0.00000937, 0.00006207, 0.00003750, -0.00003981, -0.00004911, -0.00000330, -0.00000913, -0.00006040, -0.00005047, 0.00001852, 0.00002221, -0.00006223, -0.00009994, -0.00003543, -0.00001455, -0.00014759, -0.00028645, -0.00020508, 0.00006885, 0.00027437, 0.00025457, 0.00011820, 0.00004134, 0.00004896, 0.00006023, 0.00004565, 0.00004175, 0.00005928, 0.00006549, 0.00004735, 0.00002924, 0.00002668, 0.00002777, 0.00002496, 0.00002921, 0.00004420, 0.00005396, 0.00004661, 0.00003158, 0.00002431, 0.00002789, 0.00003774, 0.00005228, 0.00007022, 0.00008462, 0.00008747, 0.00007686, 0.00005565, 0.00002717, -0.00000404, -0.00003072, -0.00004677, -0.00005191, },
		{ 0.00000203, -0.00001573, -0.00002913, -0.00003876, -0.00004364, -0.00004261, -0.00003670, -0.00002909, -0.00002223, -0.00001624, -0.00001087, -0.00000751, -0.00000747, -0.00000930, -0.00001049, -0.00001062, -0.00001029, -0.00000847, -0.00000446, -0.00000099, -0.00000116, -0.00000343, -0.00000420, -0.00000391, -0.00000511, -0.00000609, -0.00000345, 0.00000023, -0.00000069, -0.00000512, -0.00000682, -0.00000542, -0.00000584, -0.00000692, -0.00000263, 0.00000458, 0.00000522, -0.00000050, -0.00000214, 0.00000177, 0.00000154, -0.00000292, 0.00000037, 0.00001034, 0.00000887, -0.00000712, -0.00001608, -0.00000836, -0.00000298, -0.00000869, -0.00000339, 0.00001992, 0.00002565, -0.00000621, -0.00003515, -0.00002580, -0.00001427, -0.00003624, -0.00003526, 0.00004841, 0.00012590, 0.00004635, -0.00014430, -0.00019939, -0.00001180, 0.00019524, 0.00015971, -0.00005354, -0.00016197, -0.00007173, 0.00003749, 0.00002761, -0.00001812, 0.00000530, 0.00003963, -0.00000268, -0.00006474, -0.00004848, 0.00000879, 0.00000771, -0.00003360, -0.00002333, 0.00002596, 0.00001838, -0.00004426, -0.00005923, -0.00000984, 0.00000088, -0.00005916, -0.00008241, -0.00001866, 0.00000460, -0.00011916, -0.00026431, -0.00021599, 0.00002410, 0.00022860, 0.00023898, 0.00013189, 0.00006129, 0.00005817, 0.00005670, 0.00003224, 0.00002079, 0.00003700, 0.00005193, 0.00004774, 0.00003918, 0.00003708, 0.00003167, 0.00001927, 0.00001447, 0.00002540, 0.00003923, 0.00004175, 0.00003548, 0.00002997, 0.00002709, 0.00002458, 0.00002502, 0.00003274, 0.00004637, 0.00006027, 0.00006989, 0.00007233, 0.00006488, 0.00004732, 0.00002422, },
		{ 0.00005748, 0.00005054, 0.00003929, 0.00002582, 0.00001059, -0.00000572, -0.00002054, -0.00003051, -0.00003460, -0.00003484, -0.00003330, -0.00002963, -0.00002318, -0.00001593, -0.00001077, -0.00000803, -0.00000642, -0.00000626, -0.00000814, -0.00000975, -0.00000835, -0.00000543, -0.00000432, -0.00000439, -0.00000298, -0.00000162, -0.00000367, -0.00000683, -0.00000580, -0.00000179, -0.00000086, -0.00000310, -0.00000317, -0.00000167, -0.00000445, -0.00000964, -0.00000849, -0.00000138, 0.00000154, -0.00000128, -0.00000031, 0.00000497, 0.00000316, -0.00000537, -0.00000439, 0.00000851, 0.00001323, 0.00000201, -0.00000538, 0.00000024, -0.00000281, -0.00002284, -0.00002758, 0.00000146, 0.00002616, 0.00001364, -0.00000112, 0.00001699, 0.00001624, -0.00005909, -0.00012693, -0.00004873, 0.00012974, 0.00017780, -0.00000128, -0.00019556, -0.00016045, 0.00004119, 0.00014439, 0.00006129, -0.00004196, -0.00003781, -0.00000181, -0.00002801, -0.00006094, -0.00002021, 0.00004147, 0.00003274, -0.00001705, -0.00002251, 0.00000180, -0.00001760, -0.00006216, -0.00004874, 0.00001372, 0.00003116, -0.00001217, -0.00003155, -0.00000200, -0.00000446, -0.00006671, -0.00009253, -0.00003014, 0.00000813, -0.00008740, -0.00022089, -0.00019714, 0.00000192, 0.00018511, 0.00020435, 0.00011702, 0.00005918, 0.00006245, 0.00006653, 0.00004340, 0.00002399, 0.00002733, 0.00003419, 0.00003070, 0.00002848, 0.00003448, 0.00003633, 0.00002707, 0.00001872, 0.00002148, 0.00002819, 0.00002866, 0.00002516, 0.00002485, 0.00002706, 0.00002664, 0.00002348, 0.00002204, 0.00002485, 0.00003143, 0.00004073, 0.00005076, 0.00005753, },
		{ 0.00002891, 0.00003563, 0.00004137, 0.00004471, 0.00004546, 0.00004317, 0.00003654, 0.00002493, 0.00001048, -0.00000268, -0.00001214, -0.00001912, -0.00002541, -0.00002985, -0.00003027, -0.00002727, -0.00002314, -0.00001817, -0.00001177, -0.00000608, -0.00000404, -0.00000472, -0.00000504, -0.00000524, -0.00000726, -0.00000900, -0.00000677, -0.00000270, -0.00000223, -0.00000471, -0.00000471, -0.00000243, -0.00000308, -0.00000574, -0.00000403, 0.00000085, 0.00000045, -0.00000539, -0.00000736, -0.00000399, -0.00000428, -0.00000853, -0.00000539, 0.00000528, 0.00000784, -0.00000119, -0.00000429, 0.00000480, 0.00000743, -0.00000357, -0.00000525, 0.00001201, 0.00001763, -0.00000728, -0.00002824, -0.00001516, -0.00000160, -0.00002087, -0.00002374, 0.00004353, 0.00010532, 0.00003246, -0.00013248, -0.00017533, -0.00000824, 0.00016988, 0.00013330, -0.00005533, -0.00014788, -0.00006510, 0.00003527, 0.00003309, -0.00000210, 0.00001681, 0.00004017, -0.00000318, -0.00006236, -0.00005292, -0.00000237, 0.00000913, -0.00001006, 0.00000162, 0.00002860, 0.00000482, -0.00005359, -0.00006342, -0.00001724, 0.00000609, -0.00001720, -0.00002158, 0.00001367, 0.00001524, -0.00004693, -0.00008603, -0.00004290, -0.00000683, -0.00008138, -0.00019499, -0.00017642, -0.00000030, 0.00016749, 0.00018771, 0.00010478, 0.00004490, 0.00004632, 0.00005746, 0.00004639, 0.00003374, 0.00003528, 0.00003603, 0.00002592, 0.00001796, 0.00002135, 0.00002601, 0.00002343, 0.00002078, 0.00002512, 0.00002992, 0.00002678, 0.00001926, 0.00001616, 0.00001879, 0.00002192, 0.00002252, 0.00002172, 0.00002104, 0.00002128, 0.00002367, },
		{ 0.00001897, 0.00001859, 0.00001905, 0.00002093, 0.00002420, 0.00002871, 0.00003414, 0.00003917, 0.00004106, 0.00003721, 0.00002817, 0.00001792, 0.00000976, 0.00000297, -0.00000467, -0.00001249, -0.00001825, -0.00002216, -0.00002562, -0.00002716, -0.00002404, -0.00001743, -0.00001116, -0.00000620, -0.00000152, 0.00000077, -0.00000229, -0.00000772, -0.00000927, -0.00000694, -0.00000587, -0.00000631, -0.00000385, -0.00000002, -0.00000141, -0.00000680, -0.00000737, -0.00000217, -0.00000002, -0.00000276, -0.00000223, 0.00000177, -0.00000153, -0.00001200, -0.00001396, -0.00000325, 0.00000364, -0.00000027, 0.00000096, 0.00001240, 0.00001162, -0.00000959, -0.00002006, 0.00000011, 0.00001892, 0.00000719, -0.00000428, 0.00001532, 0.00001902, -0.00004443, -0.00010387, -0.00003717, 0.00011647, 0.00015649, 0.00000043, -0.00016643, -0.00013464, 0.00003652, 0.00011762, 0.00003868, -0.00005144, -0.00004287, -0.00000375, -0.00001743, -0.00003895, -0.00000269, 0.00004552, 0.00003056, -0.00001960, -0.00003010, -0.00000861, -0.00001318, -0.00003439, -0.00001763, 0.00002273, 0.00001957, -0.00002567, -0.00004227, -0.00001477, -0.00000507, -0.00003304, -0.00003899, -0.00000020, 0.00001619, -0.00002791, -0.00006339, -0.00003257, -0.00000585, -0.00007545, -0.00018286, -0.00017277, -0.00001288, 0.00015078, 0.00018240, 0.00010958, 0.00004593, 0.00003625, 0.00004200, 0.00003415, 0.00002767, 0.00003466, 0.00004005, 0.00003236, 0.00002206, 0.00001935, 0.00001865, 0.00001431, 0.00001290, 0.00002001, 0.00002836, 0.00002814, 0.00002082, 0.00001497, 0.00001432, 0.00001615, 0.00001770, 0.00001866, 0.00001915, },
		{ 0.00001554, 0.00001653, 0.00001744, 0.00001750, 0.00001641, 0.00001475, 0.00001383, 0.00001509, 0.00001953, 0.00002675, 0.00003391, 0.00003673, 0.00003346, 0.00002691, 0.00002082, 0.00001542, 0.00000916, 0.00000300, -0.00000105, -0.00000440, -0.00001017, -0.00001748, -0.00002231, -0.00002347, -0.00002254, -0.00001879, -0.00001071, -0.00000202, 0.00000099, -0.00000143, -0.00000351, -0.00000418, -0.00000707, -0.00001046, -0.00000809, -0.00000155, 0.00000009, -0.00000443, -0.00000638, -0.00000348, -0.00000349, -0.00000698, -0.00000394, 0.00000548, 0.00000643, -0.00000483, -0.00001193, -0.00000727, -0.00000552, -0.00001211, -0.00000720, 0.00001539, 0.00002513, 0.00000352, -0.00001755, -0.00000886, 0.00000083, -0.00001827, -0.00002235, 0.00003696, 0.00009262, 0.00002941, -0.00011558, -0.00015316, -0.00000554, 0.00015283, 0.00012461, -0.00003589, -0.00011531, -0.00004843, 0.00002807, 0.00001578, -0.00001939, -0.00000118, 0.00002535, -0.00000336, -0.00004615, -0.00003400, 0.00000792, 0.00001270, -0.00001028, -0.00000636, 0.00001611, 0.00000588, -0.00002814, -0.00003003, -0.00000119, 0.00000176, -0.00002715, -0.00003196, -0.00000048, 0.00001012, -0.00002186, -0.00004104, -0.00001682, -0.00000193, -0.00003425, -0.00005916, -0.00002580, 0.00000470, -0.00005480, -0.00015864, -0.00016423, -0.00002907, 0.00012405, 0.00016657, 0.00011055, 0.00005326, 0.00003948, 0.00003961, 0.00002827, 0.00001883, 0.00002410, 0.00003201, 0.00003041, 0.00002526, 0.00002389, 0.00002153, 0.00001415, 0.00000906, 0.00001320, 0.00002128, 0.00002363, 0.00001929, 0.00001457, 0.00001328, 0.00001405, 0.00001485, },
		{ 0.00001158, 0.00001237, 0.00001262, 0.00001347, 0.00001542, 0.00001732, 0.00001744, 0.00001517, 0.00001154, 0.00000843, 0.00000810, 0.00001233, 0.00002018, 0.00002742, 0.00003039, 0.00002965, 0.00002746, 0.00002342, 0.00001631, 0.00000871, 0.00000487, 0.00000437, 0.00000296, -0.00000078, -0.00000498, -0.00000969, -0.00001633, -0.00002191, -0.00002115, -0.00001496, -0.00000945, -0.00000610, -0.00000160, 0.00000245, 0.00000015, -0.00000671, -0.00000897, -0.00000507, -0.00000335, -0.00000600, -0.00000532, -0.00000058, -0.00000170, -0.00000936, -0.00000990, 0.00000024, 0.00000568, -0.00000043, -0.00000336, 0.00000299, 0.00000026, -0.00001792, -0.00002381, -0.00000123, 0.00001881, 0.00000907, -0.00000169, 0.00001609, 0.00002088, -0.00003427, -0.00008704, -0.00002879, 0.00010580, 0.00013876, -0.00000232, -0.00015203, -0.00012404, 0.00003011, 0.00010851, 0.00004791, -0.00002570, -0.00002075, 0.00000351, -0.00001987, -0.00004594, -0.00001615, 0.00002839, 0.00002108, -0.00001602, -0.00002137, -0.00000308, -0.00000988, -0.00003214, -0.00002172, 0.00001332, 0.00002040, -0.00000320, -0.00001083, 0.00000279, -0.00000585, -0.00003950, -0.00004562, -0.00001016, 0.00001273, -0.00000633, -0.00002426, -0.00001013, -0.00000457, -0.00004000, -0.00006764, -0.00003768, -0.00000136, -0.00004270, -0.00013196, -0.00014394, -0.00003114, 0.00010546, 0.00014901, 0.00010288, 0.00005172, 0.00003913, 0.00004067, 0.00003114, 0.00001999, 0.00002012, 0.00002406, 0.00002243, 0.00001977, 0.00002143, 0.00002208, 0.00001683, 0.00001139, 0.00001282, 0.00001816, 0.00001962, 0.00001577, 0.00001158, 0.00001055, },
		{ 0.00001031, 0.00000869, 0.00000959, 0.00001021, 0.00000948, 0.00000899, 0.00001056, 0.00001375, 0.00001639, 0.00001690, 0.00001500, 0.00001109, 0.00000676, 0.00000499, 0.00000750, 0.00001243, 0.00001706, 0.00002157, 0.00002673, 0.00002950, 0.00002582, 0.00001723, 0.00000949, 0.00000511, 0.00000237, 0.00000112, 0.00000285, 0.00000494, 0.00000216, -0.00000507, -0.00001112, -0.00001438, -0.00001771, -0.00001943, -0.00001406, -0.00000385, 0.00000147, -0.00000038, -0.00000168, -0.00000013, -0.00000266, -0.00000926, -0.00000935, -0.00000148, 0.00000114, -0.00000600, -0.00000933, -0.00000286, -0.00000071, -0.00000839, -0.00000754, 0.00000909, 0.00001583, -0.00000309, -0.00001947, -0.00000888, 0.00000061, -0.00001841, -0.00002478, 0.00002766, 0.00008001, 0.00002805, -0.00009701, -0.00012779, 0.00000387, 0.00014206, 0.00011200, -0.00003599, -0.00011029, -0.00005101, 0.00002231, 0.00002136, -0.00000098, 0.00001666, 0.00003391, -0.00000003, -0.00004382, -0.00003558, 0.00000170, 0.00000861, -0.00000835, -0.00000418, 0.00001295, 0.00000033, -0.00003276, -0.00003698, -0.00000988, 0.00000418, -0.00000388, 0.00000038, 0.00001879, 0.00001052, -0.00002846, -0.00004715, -0.00002382, -0.00000188, -0.00001157, -0.00002048, -0.00000252, 0.00000572, -0.00002806, -0.00006179, -0.00004481, -0.00001465, -0.00004654, -0.00012162, -0.00013193, -0.00003061, 0.00009649, 0.00014155, 0.00010106, 0.00005030, 0.00003427, 0.00003557, 0.00003002, 0.00002206, 0.00002175, 0.00002355, 0.00002011, 0.00001587, 0.00001631, 0.00001742, 0.00001436, 0.00001094, 0.00001280, 0.00001771, 0.00001900, 0.00001505, },
		{ 0.00001711, 0.00001403, 0.00000949, 0.00000785, 0.00000934, 0.00001071, 0.00000969, 0.00000745, 0.00000665, 0.00000797, 0.00001012, 0.00001221, 0.00001394, 0.00001425, 0.00001231, 0.00000949, 0.00000758, 0.00000604, 0.00000427, 0.00000521, 0.00001196, 0.00002149, 0.00002701, 0.00002614, 0.00002183, 0.00001558, 0.00000676, -0.00000157, -0.00000369, 0.00000034, 0.00000453, 0.00000611, 0.00000692, 0.00000531, -0.00000323, -0.00001540, -0.00002081, -0.00001690, -0.00001166, -0.00000858, -0.00000233, 0.00000607, 0.00000595, -0.00000381, -0.00000924, -0.00000439, -0.00000138, -0.00000626, -0.00000643, 0.00000261, 0.00000270, -0.00001335, -0.00002068, -0.00000324, 0.00001302, 0.00000483, -0.00000216, 0.00001669, 0.00002141, -0.00003133, -0.00008380, -0.00003529, 0.00008512, 0.00011785, -0.00000308, -0.00013122, -0.00010203, 0.00003653, 0.00010388, 0.00004485, -0.00002624, -0.00002518, -0.00000191, -0.00001576, -0.00003129, -0.00000242, 0.00003326, 0.00002120, -0.00001454, -0.00001890, -0.00000046, -0.00000330, -0.00002077, -0.00001309, 0.00001218, 0.00001142, -0.00001555, -0.00002665, -0.00001341, -0.00000893, -0.00001925, -0.00001276, 0.00001271, 0.00001676, -0.00001177, -0.00003079, -0.00001688, -0.00000374, -0.00001737, -0.00002847, -0.00001082, 0.00000377, -0.00001902, -0.00004803, -0.00003734, -0.00001481, -0.00004576, -0.00011642, -0.00013057, -0.00004058, 0.00008093, 0.00013291, 0.00010319, 0.00005575, 0.00003547, 0.00003257, 0.00002664, 0.00001980, 0.00002005, 0.00002265, 0.00002056, 0.00001664, 0.00001571, 0.00001524, 0.00001171, 0.00000835, 0.00001005, 0.00001503, },
		{ 0.00000898, 0.00001297, 0.00001492, 0.00001195, 0.00000690, 0.00000465, 0.00000663, 0.00000977, 0.00001071, 0.00000955, 0.00000846, 0.00000791, 0.00000684, 0.00000568, 0.00000609, 0.00000808, 0.00001031, 0.00001253, 0.00001450, 0.00001344, 0.00000737, 0.00000024, -0.00000130, 0.00000372, 0.00001067, 0.00001696, 0.00002317, 0.00002674, 0.00002264, 0.00001178, 0.00000151, -0.00000401, -0.00000631, -0.00000494, 0.00000286, 0.00001273, 0.00001439, 0.00000599, -0.00000346, -0.00000917, -0.00001543, -0.00002131, -0.00001725, -0.00000344, 0.00000524, 0.00000196, -0.00000161, 0.00000135, -0.00000020, -0.00001009, -0.00001090, 0.00000433, 0.00001175, -0.00000431, -0.00001923, -0.00001051, -0.00000212, -0.00001774, -0.00002029, 0.00003014, 0.00007801, 0.00002825, -0.00008968, -0.00012243, -0.00000686, 0.00011730, 0.00009295, -0.00003526, -0.00009694, -0.00004044, 0.00002612, 0.00002384, 0.00000089, 0.00001395, 0.00002970, 0.00000365, -0.00003075, -0.00002344, 0.00000441, 0.00000459, -0.00001240, -0.00000599, 0.00001470, 0.00000987, -0.00001473, -0.00001855, -0.00000025, 0.00000341, -0.00001306, -0.00001747, -0.00000369, -0.00000184, -0.00001797, -0.00002087, -0.00000250, 0.00000386, -0.00001502, -0.00002557, -0.00000783, 0.00000716, -0.00000680, -0.00002410, -0.00001603, -0.00000525, -0.00002289, -0.00004551, -0.00003351, -0.00001026, -0.00003556, -0.00010171, -0.00012309, -0.00004993, 0.00006006, 0.00011530, 0.00009716, 0.00005771, 0.00003857, 0.00003453, 0.00002812, 0.00002017, 0.00001815, 0.00001947, 0.00001834, 0.00001606, 0.00001583, 0.00001542, 0.00001198, 0.00000833, },
		{ 0.00001139, 0.00000846, 0.00000909, 0.00001295, 0.00001533, 0.00001269, 0.00000670, 0.00000237, 0.00000250, 0.00000503, 0.00000670, 0.00000749, 0.00000927, 0.00001136, 0.00001099, 0.00000773, 0.00000399, 0.00000153, 0.00000093, 0.00000371, 0.00001031, 0.00001634, 0.00001617, 0.00001036, 0.00000449, 0.00000099, -0.00000169, -0.00000206, 0.00000416, 0.00001534, 0.00002348, 0.00002425, 0.00002025, 0.00001307, 0.00000157, -0.00000980, -0.00001179, -0.00000362, 0.00000497, 0.00000868, 0.00001158, 0.00001335, 0.00000569, -0.00001034, -0.00001964, -0.00001534, -0.00000916, -0.00000871, -0.00000428, 0.00000744, 0.00000945, -0.00000533, -0.00001420, -0.00000155, 0.00001010, -0.00000004, -0.00000857, 0.00000752, 0.00001299, -0.00003093, -0.00007294, -0.00002437, 0.00008581, 0.00011261, -0.00000172, -0.00012219, -0.00009903, 0.00002446, 0.00008539, 0.00003408, -0.00002728, -0.00002420, -0.00000212, -0.00001414, -0.00002813, -0.00000180, 0.00003293, 0.00002753, -0.00000000, -0.00000538, 0.00000322, -0.00000835, -0.00002844, -0.00002006, 0.00000852, 0.00001630, 0.00000014, -0.00000627, 0.00000299, 0.00000011, -0.00001822, -0.00002230, -0.00000551, 0.00000293, -0.00000790, -0.00001297, -0.00000258, -0.00000286, -0.00002418, -0.00003516, -0.00001549, 0.00000662, 0.00000212, -0.00001140, -0.00000760, -0.00000258, -0.00002244, -0.00004673, -0.00003827, -0.00001477, -0.00003100, -0.00008654, -0.00010874, -0.00004925, 0.00004700, 0.00009936, 0.00008671, 0.00005233, 0.00003496, 0.00003266, 0.00002934, 0.00002318, 0.00002006, 0.00001931, 0.00001719, 0.00001477, 0.00001434, 0.00001404, },
		{ 0.00001489, 0.00001349, 0.00001035, 0.00000720, 0.00000737, 0.00001119, 0.00001485, 0.00001422, 0.00000943, 0.00000475, 0.00000334, 0.00000358, 0.00000253, 0.00000118, 0.00000291, 0.00000766, 0.00001155, 0.00001221, 0.00001027, 0.00000623, 0.00000064, -0.00000301, -0.00000087, 0.00000536, 0.00001020, 0.00001214, 0.00001363, 0.00001361, 0.00000792, -0.00000170, -0.00000714, -0.00000454, 0.00000212, 0.00000972, 0.00001886, 0.00002601, 0.00002337, 0.00001102, -0.00000104, -0.00000712, -0.00001093, -0.00001250, -0.00000453, 0.00001111, 0.00001909, 0.00001242, 0.00000282, -0.00000128, -0.00000814, -0.00002026, -0.00002093, -0.00000335, 0.00000987, 0.00000200, -0.00000716, 0.00000162, 0.00000657, -0.00001307, -0.00002218, 0.00001780, 0.00005901, 0.00001675, -0.00008315, -0.00010523, 0.00000348, 0.00011465, 0.00008801, -0.00003257, -0.00009121, -0.00004142, 0.00001848, 0.00001701, -0.00000304, 0.00000860, 0.00002180, -0.00000272, -0.00003401, -0.00002593, 0.00000405, 0.00001251, 0.00000469, 0.00001118, 0.00002238, 0.00000754, -0.00002193, -0.00002706, -0.00000720, 0.00000320, -0.00000341, -0.00000200, 0.00001087, 0.00000911, -0.00001144, -0.00002202, -0.00001133, -0.00000259, -0.00000776, -0.00000799, 0.00000382, 0.00000322, -0.00002052, -0.00003795, -0.00002572, -0.00000511, -0.00000372, -0.00001013, -0.00000250, 0.00000511, -0.00001208, -0.00003784, -0.00003693, -0.00001976, -0.00003301, -0.00007992, -0.00009960, -0.00004757, 0.00003989, 0.00009071, 0.00008169, 0.00004886, 0.00002948, 0.00002605, 0.00002493, 0.00002193, 0.00002040, 0.00002009, 0.00001845, 0.00001616, },
		{ 0.00001715, 0.00001686, 0.00001653, 0.00001516, 0.00001184, 0.00000801, 0.00000658, 0.00000869, 0.00001189, 0.00001230, 0.00000921, 0.00000631, 0.00000662, 0.00000783, 0.00000571, 0.00000100, -0.00000152, 0.00000009, 0.00000340, 0.00000659, 0.00000974, 0.00001132, 0.00000882, 0.00000380, 0.00000058, -0.00000021, -0.00000119, -0.00000104, 0.00000446, 0.00001334, 0.00001735, 0.00001316, 0.00000588, -0.00000043, -0.00000664, -0.00001050, -0.00000547, 0.00000788, 0.00001940, 0.00002300, 0.00002222, 0.00001801, 0.00000494, -0.00001369, -0.00002220, -0.00001398, -0.00000157, 0.00000501, 0.00001214, 0.00002169, 0.00001805, -0.00000412, -0.00002127, -0.00001536, -0.00000452, -0.00000816, -0.00000761, 0.00001493, 0.00002476, -0.00001512, -0.00005827, -0.00002218, 0.00007050, 0.00009190, -0.00000881, -0.00011140, -0.00008474, 0.00002850, 0.00008126, 0.00003159, -0.00002582, -0.00002356, -0.00000344, -0.00001386, -0.00002637, -0.00000368, 0.00002519, 0.00001726, -0.00001009, -0.00001525, -0.00000342, -0.00000495, -0.00001339, -0.00000179, 0.00001977, 0.00001813, -0.00000370, -0.00001251, -0.00000304, -0.00000120, -0.00001192, -0.00001148, 0.00000429, 0.00000988, -0.00000365, -0.00001346, -0.00000749, -0.00000326, -0.00001016, -0.00001067, 0.00000248, 0.00000629, -0.00001268, -0.00003053, -0.00002437, -0.00000976, -0.00001058, -0.00001730, -0.00000933, 0.00000219, -0.00000702, -0.00002705, -0.00002789, -0.00001570, -0.00002942, -0.00007311, -0.00009405, -0.00005029, 0.00002978, 0.00008154, 0.00007894, 0.00004982, 0.00002870, 0.00002216, 0.00002011, 0.00001766, 0.00001656, 0.00001697, },
		{ 0.00001339, 0.00001255, 0.00001301, 0.00001415, 0.00001521, 0.00001504, 0.00001308, 0.00001024, 0.00000853, 0.00000930, 0.00001113, 0.00001052, 0.00000652, 0.00000328, 0.00000490, 0.00000909, 0.00001014, 0.00000693, 0.00000339, 0.00000159, -0.00000006, -0.00000150, -0.00000020, 0.00000364, 0.00000665, 0.00000801, 0.00000943, 0.00000924, 0.00000366, -0.00000467, -0.00000742, -0.00000202, 0.00000536, 0.00001021, 0.00001408, 0.00001602, 0.00001040, -0.00000213, -0.00001152, -0.00001171, -0.00000680, 0.00000051, 0.00001358, 0.00002843, 0.00003058, 0.00001544, -0.00000261, -0.00001193, -0.00001801, -0.00002353, -0.00001533, 0.00000972, 0.00002731, 0.00001934, 0.00000392, 0.00000183, -0.00000246, -0.00002483, -0.00003240, 0.00000915, 0.00005420, 0.00002271, -0.00006462, -0.00008590, 0.00000818, 0.00010425, 0.00007808, -0.00003007, -0.00008135, -0.00003641, 0.00001568, 0.00001251, -0.00000546, 0.00000677, 0.00002062, -0.00000020, -0.00002814, -0.00002204, 0.00000217, 0.00000589, -0.00000512, -0.00000158, 0.00001032, 0.00000394, -0.00001362, -0.00001319, 0.00000274, 0.00000599, -0.00000510, -0.00000537, 0.00000792, 0.00001010, -0.00000431, -0.00001178, -0.00000328, 0.00000134, -0.00000759, -0.00001257, -0.00000410, 0.00000139, -0.00000568, -0.00000927, 0.00000003, 0.00000371, -0.00001140, -0.00002603, -0.00001979, -0.00000614, -0.00000783, -0.00001771, -0.00001528, -0.00000650, -0.00001213, -0.00002627, -0.00002445, -0.00001146, -0.00002143, -0.00006024, -0.00008292, -0.00004954, 0.00001994, 0.00006973, 0.00007271, 0.00004895, 0.00002885, 0.00002120, 0.00001866, 0.00001583, },
		{ 0.00001700, 0.00001517, 0.00001227, 0.00001012, 0.00000996, 0.00001127, 0.00001247, 0.00001247, 0.00001136, 0.00000981, 0.00000905, 0.00001037, 0.00001279, 0.00001245, 0.00000739, 0.00000178, 0.00000124, 0.00000502, 0.00000767, 0.00000752, 0.00000765, 0.00000868, 0.00000720, 0.00000239, -0.00000212, -0.00000412, -0.00000460, -0.00000262, 0.00000413, 0.00001221, 0.00001376, 0.00000741, 0.00000027, -0.00000320, -0.00000549, -0.00000670, -0.00000153, 0.00000963, 0.00001711, 0.00001512, 0.00000852, 0.00000151, -0.00000834, -0.00001799, -0.00001544, 0.00000175, 0.00001842, 0.00002325, 0.00002260, 0.00002113, 0.00000821, -0.00001792, -0.00003361, -0.00002209, -0.00000234, 0.00000359, 0.00000901, 0.00002906, 0.00003319, -0.00001031, -0.00005696, -0.00002943, 0.00005330, 0.00007514, -0.00001143, -0.00009926, -0.00007097, 0.00003430, 0.00008374, 0.00003941, -0.00001422, -0.00001716, -0.00000583, -0.00002047, -0.00003273, -0.00000937, 0.00002113, 0.00001793, -0.00000444, -0.00000909, -0.00000076, -0.00000615, -0.00001830, -0.00001123, 0.00000841, 0.00001199, -0.00000043, -0.00000434, 0.00000201, -0.00000210, -0.00001593, -0.00001546, 0.00000247, 0.00001297, 0.00000576, -0.00000101, 0.00000250, 0.00000164, -0.00001063, -0.00001758, -0.00000924, -0.00000056, -0.00000301, -0.00000478, 0.00000273, 0.00000446, -0.00001094, -0.00002582, -0.00002021, -0.00000513, -0.00000325, -0.00001175, -0.00001270, -0.00000850, -0.00001564, -0.00002950, -0.00002864, -0.00001564, -0.00002007, -0.00005010, -0.00006962, -0.00004287, 0.00001667, 0.00006212, 0.00006731, 0.00004660, 0.00002689, 0.00001877, },
		{ 0.00002727, 0.00001696, 0.00001464, 0.00001337, 0.00001039, 0.00000760, 0.00000743, 0.00000936, 0.00001086, 0.00001069, 0.00000952, 0.00000805, 0.00000708, 0.00000833, 0.00001180, 0.00001350, 0.00001003, 0.00000438, 0.00000236, 0.00000398, 0.00000436, 0.00000272, 0.00000348, 0.00000786, 0.00001122, 0.00001046, 0.00000726, 0.00000271, -0.00000404, -0.00000996, -0.00000890, -0.00000107, 0.00000602, 0.00000853, 0.00000991, 0.00001085, 0.00000603, -0.00000449, -0.00001118, -0.00000834, -0.00000138, 0.00000460, 0.00001191, 0.00001889, 0.00001552, 0.00000007, -0.00001326, -0.00001445, -0.00001080, -0.00000792, 0.00000367, 0.00002566, 0.00003592, 0.00001933, -0.00000450, -0.00001289, -0.00001777, -0.00003400, -0.00003352, 0.00001281, 0.00006125, 0.00003665, -0.00004375, -0.00006893, 0.00000796, 0.00008681, 0.00005762, -0.00004128, -0.00008365, -0.00003492, 0.00002215, 0.00002806, 0.00001572, 0.00002339, 0.00002650, -0.00000286, -0.00003491, -0.00003077, -0.00000612, 0.00000193, -0.00000347, 0.00000249, 0.00001334, 0.00000518, -0.00001459, -0.00001784, -0.00000418, 0.00000252, -0.00000148, 0.00000127, 0.00001009, 0.00000515, -0.00001307, -0.00002007, -0.00000802, 0.00000308, 0.00000185, 0.00000124, 0.00000844, 0.00000934, -0.00000355, -0.00001467, -0.00001191, -0.00000582, -0.00000673, -0.00000582, 0.00000352, 0.00000694, -0.00000710, -0.00002314, -0.00002089, -0.00000736, -0.00000311, -0.00000843, -0.00000878, -0.00000531, -0.00001222, -0.00002672, -0.00003003, -0.00002163, -0.00002497, -0.00004828, -0.00006303, -0.00003872, 0.00001510, 0.00005860, 0.00006645, 0.00004802, },
		{ 0.00006462, 0.00005099, 0.00003106, 0.00001910, 0.00001557, 0.00001371, 0.00000945, 0.00000493, 0.00000394, 0.00000622, 0.00000835, 0.00000876, 0.00000839, 0.00000758, 0.00000628, 0.00000634, 0.00000905, 0.00001123, 0.00000903, 0.00000480, 0.00000420, 0.00000685, 0.00000682, 0.00000275, 0.00000016, 0.00000224, 0.00000579, 0.00000811, 0.00001027, 0.00001140, 0.00000751, -0.00000059, -0.00000628, -0.00000699, -0.00000692, -0.00000691, -0.00000159, 0.00000897, 0.00001512, 0.00001128, 0.00000352, -0.00000220, -0.00000813, -0.00001380, -0.00001038, 0.00000395, 0.00001602, 0.00001641, 0.00001218, 0.00000907, -0.00000158, -0.00002121, -0.00002883, -0.00001063, 0.00001343, 0.00002039, 0.00002125, 0.00003112, 0.00002488, -0.00002363, -0.00007122, -0.00004527, 0.00003739, 0.00006884, 0.00000159, -0.00007122, -0.00004553, 0.00004301, 0.00007649, 0.00002494, -0.00003056, -0.00003280, -0.00001450, -0.00001487, -0.00001381, 0.00001317, 0.00003825, 0.00002730, -0.00000172, -0.00001222, -0.00000767, -0.00001257, -0.00002154, -0.00001244, 0.00000730, 0.00001093, -0.00000158, -0.00000702, -0.00000142, -0.00000167, -0.00000878, -0.00000594, 0.00000641, 0.00000788, -0.00000544, -0.00001338, -0.00000683, -0.00000088, -0.00000458, -0.00000560, 0.00000341, 0.00000914, 0.00000182, -0.00000721, -0.00000679, -0.00000433, -0.00000777, -0.00000857, 0.00000005, 0.00000524, -0.00000528, -0.00001960, -0.00001902, -0.00000770, -0.00000349, -0.00000797, -0.00000842, -0.00000441, -0.00000835, -0.00002018, -0.00002523, -0.00002162, -0.00002698, -0.00004793, -0.00006116, -0.00004029, 0.00000853, 0.00005179, },
		{ 0.00000013, 0.00004076, 0.00005749, 0.00004950, 0.00003256, 0.00002134, 0.00001851, 0.00001735, 0.00001239, 0.00000584, 0.00000293, 0.00000400, 0.00000521, 0.00000502, 0.00000505, 0.00000557, 0.00000543, 0.00000582, 0.00000841, 0.00001033, 0.00000746, 0.00000230, 0.00000170, 0.00000604, 0.00000852, 0.00000608, 0.00000324, 0.00000319, 0.00000331, 0.00000232, 0.00000351, 0.00000745, 0.00000935, 0.00000751, 0.00000590, 0.00000501, -0.00000047, -0.00001021, -0.00001460, -0.00000866, 0.00000044, 0.00000595, 0.00001040, 0.00001453, 0.00001059, -0.00000327, -0.00001457, -0.00001396, -0.00000824, -0.00000377, 0.00000663, 0.00002392, 0.00002852, 0.00000847, -0.00001566, -0.00002091, -0.00001814, -0.00002353, -0.00001511, 0.00003070, 0.00007232, 0.00004114, -0.00004484, -0.00007993, -0.00001663, 0.00005664, 0.00003921, -0.00003751, -0.00006383, -0.00001302, 0.00003743, 0.00003461, 0.00001219, 0.00000974, 0.00000918, -0.00001335, -0.00003337, -0.00002067, 0.00000676, 0.00001500, 0.00000867, 0.00001148, 0.00001783, 0.00000648, -0.00001501, -0.00002051, -0.00000950, -0.00000387, -0.00000750, -0.00000453, 0.00000572, 0.00000652, -0.00000346, -0.00000674, 0.00000040, 0.00000175, -0.00000755, -0.00001160, -0.00000324, 0.00000308, -0.00000194, -0.00000678, -0.00000201, 0.00000282, -0.00000146, -0.00000665, -0.00000390, -0.00000025, -0.00000349, -0.00000610, -0.00000066, 0.00000296, -0.00000636, -0.00001911, -0.00001851, -0.00000742, -0.00000234, -0.00000642, -0.00000848, -0.00000581, -0.00000784, -0.00001607, -0.00002012, -0.00001831, -0.00002418, -0.00004311, -0.00005635, -0.00004111, },
		{ -0.00004798, -0.00003711, -0.00000402, 0.00003153, 0.00004910, 0.00004449, 0.00002959, 0.00001926, 0.00001815, 0.00001942, 0.00001587, 0.00000925, 0.00000580, 0.00000626, 0.00000613, 0.00000377, 0.00000189, 0.00000152, 0.00000133, 0.00000233, 0.00000645, 0.00001043, 0.00000865, 0.00000267, 0.00000012, 0.00000307, 0.00000553, 0.00000438, 0.00000381, 0.00000620, 0.00000732, 0.00000467, 0.00000211, 0.00000215, 0.00000201, 0.00000136, 0.00000486, 0.00001175, 0.00001296, 0.00000469, -0.00000492, -0.00000896, -0.00001091, -0.00001300, -0.00000837, 0.00000502, 0.00001537, 0.00001356, 0.00000654, 0.00000168, -0.00000698, -0.00002078, -0.00002246, -0.00000209, 0.00001981, 0.00002153, 0.00001487, 0.00001721, 0.00000895, -0.00003236, -0.00006806, -0.00003423, 0.00005020, 0.00008290, 0.00001830, -0.00005786, -0.00004730, 0.00002205, 0.00004671, 0.00000132, -0.00004127, -0.00003247, -0.00000619, -0.00000118, -0.00000031, 0.00001971, 0.00003650, 0.00002207, -0.00000622, -0.00001610, -0.00001146, -0.00001378, -0.00001726, -0.00000294, 0.00001974, 0.00002431, 0.00001032, -0.00000015, -0.00000167, -0.00000781, -0.00001833, -0.00001716, -0.00000349, 0.00000481, 0.00000179, 0.00000035, 0.00000464, 0.00000223, -0.00000971, -0.00001536, -0.00000685, 0.00000238, 0.00000109, -0.00000278, -0.00000039, 0.00000152, -0.00000424, -0.00001000, -0.00000714, -0.00000167, -0.00000182, -0.00000260, 0.00000224, 0.00000456, -0.00000527, -0.00001868, -0.00001958, -0.00000902, -0.00000224, -0.00000456, -0.00000731, -0.00000640, -0.00000811, -0.00001431, -0.00001773, -0.00001685, -0.00002169, -0.00003673, },
		{ -0.00002184, -0.00003294, -0.00004057, -0.00003140, -0.00000431, 0.00002665, 0.00004407, 0.00004117, 0.00002648, 0.00001524, 0.00001439, 0.00001740, 0.00001550, 0.00000982, 0.00000750, 0.00000958, 0.00001047, 0.00000769, 0.00000425, 0.00000168, -0.00000123, -0.00000279, 0.00000042, 0.00000591, 0.00000692, 0.00000320, 0.00000160, 0.00000423, 0.00000521, 0.00000188, -0.00000012, 0.00000297, 0.00000679, 0.00000726, 0.00000661, 0.00000620, 0.00000269, -0.00000324, -0.00000388, 0.00000328, 0.00000983, 0.00000978, 0.00000801, 0.00000797, 0.00000313, -0.00000921, -0.00001803, -0.00001461, -0.00000619, -0.00000095, 0.00000637, 0.00001806, 0.00001909, 0.00000060, -0.00001823, -0.00001764, -0.00000997, -0.00001264, -0.00000689, 0.00002959, 0.00006098, 0.00002709, -0.00005348, -0.00008291, -0.00001826, 0.00005743, 0.00004908, -0.00001732, -0.00004313, -0.00000382, 0.00003196, 0.00001926, -0.00000812, -0.00001235, -0.00000967, -0.00002288, -0.00003191, -0.00001189, 0.00001892, 0.00002864, 0.00002063, 0.00001613, 0.00001206, -0.00000633, -0.00002769, -0.00002727, -0.00000747, 0.00000740, 0.00001025, 0.00001370, 0.00001867, 0.00001171, -0.00000602, -0.00001608, -0.00001211, -0.00000697, -0.00000686, -0.00000298, 0.00000598, 0.00000680, -0.00000453, -0.00001306, -0.00000871, -0.00000125, -0.00000109, -0.00000277, 0.00000081, 0.00000339, -0.00000218, -0.00000934, -0.00000917, -0.00000525, -0.00000449, -0.00000337, 0.00000285, 0.00000642, -0.00000211, -0.00001571, -0.00001888, -0.00001037, -0.00000319, -0.00000395, -0.00000653, -0.00000642, -0.00000759, -0.00001226, -0.00001593, -0.00001707, },
		{ -0.00001318, -0.00001666, -0.00002273, -0.00003148, -0.00003593, -0.00002741, -0.00000461, 0.00002322, 0.00004144, 0.00004078, 0.00002629, 0.00001334, 0.00001138, 0.00001466, 0.00001306, 0.00000712, 0.00000510, 0.00000880, 0.00001192, 0.00001109, 0.00000909, 0.00000713, 0.00000303, -0.00000175, -0.00000233, 0.00000056, 0.00000075, -0.00000226, -0.00000172, 0.00000406, 0.00000763, 0.00000462, 0.00000059, 0.00000080, 0.00000251, 0.00000277, 0.00000440, 0.00000808, 0.00000800, 0.00000221, -0.00000198, -0.00000022, 0.00000156, -0.00000027, 0.00000137, 0.00001007, 0.00001560, 0.00000953, -0.00000059, -0.00000593, -0.00001150, -0.00002072, -0.00002043, -0.00000227, 0.00001601, 0.00001586, 0.00000902, 0.00001202, 0.00000692, -0.00002769, -0.00005763, -0.00002541, 0.00005103, 0.00007799, 0.00001450, -0.00005960, -0.00005285, 0.00001045, 0.00003643, 0.00000163, -0.00002920, -0.00001478, 0.00001107, 0.00001171, 0.00000341, 0.00000974, 0.00001389, -0.00000574, -0.00003084, -0.00003188, -0.00001455, -0.00000231, 0.00000481, 0.00001939, 0.00003192, 0.00002263, -0.00000200, -0.00001682, -0.00001585, -0.00001354, -0.00001317, -0.00000362, 0.00001334, 0.00002039, 0.00001254, 0.00000350, 0.00000075, -0.00000333, -0.00001097, -0.00001178, -0.00000332, 0.00000141, -0.00000440, -0.00000980, -0.00000577, 0.00000011, -0.00000099, -0.00000382, -0.00000107, 0.00000230, -0.00000110, -0.00000704, -0.00000788, -0.00000578, -0.00000605, -0.00000536, 0.00000074, 0.00000538, -0.00000082, -0.00001290, -0.00001692, -0.00001013, -0.00000325, -0.00000339, -0.00000629, -0.00000694, -0.00000718, -0.00000972, },
		{ -0.00000740, -0.00000757, -0.00000991, -0.00001500, -0.00002264, -0.00003015, -0.00003248, -0.00002501, -0.00000642, 0.00001855, 0.00003815, 0.00004080, 0.00002769, 0.00001372, 0.00001069, 0.00001390, 0.00001208, 0.00000503, 0.00000196, 0.00000564, 0.00000947, 0.00000967, 0.00000946, 0.00001015, 0.00000837, 0.00000427, 0.00000261, 0.00000317, -0.00000003, -0.00000697, -0.00000924, -0.00000335, 0.00000315, 0.00000409, 0.00000317, 0.00000455, 0.00000487, 0.00000166, -0.00000016, 0.00000255, 0.00000457, 0.00000266, 0.00000235, 0.00000621, 0.00000633, -0.00000136, -0.00000684, -0.00000177, 0.00000619, 0.00000801, 0.00000900, 0.00001403, 0.00001167, -0.00000617, -0.00002312, -0.00002154, -0.00001266, -0.00001290, -0.00000619, 0.00002722, 0.00005479, 0.00002239, -0.00005179, -0.00007744, -0.00001550, 0.00005689, 0.00005125, -0.00000989, -0.00003700, -0.00000718, 0.00001917, 0.00000438, -0.00001822, -0.00001434, -0.00000157, -0.00000469, -0.00000924, 0.00000511, 0.00002220, 0.00001654, -0.00000358, -0.00001402, -0.00001480, -0.00001988, -0.00002355, -0.00001030, 0.00001231, 0.00002173, 0.00001527, 0.00000909, 0.00000709, -0.00000159, -0.00001582, -0.00001970, -0.00000920, 0.00000181, 0.00000551, 0.00000899, 0.00001487, 0.00001393, 0.00000365, -0.00000463, -0.00000477, -0.00000529, -0.00001176, -0.00001531, -0.00000875, -0.00000002, 0.00000120, -0.00000144, -0.00000045, 0.00000150, -0.00000158, -0.00000665, -0.00000735, -0.00000548, -0.00000578, -0.00000551, -0.00000049, 0.00000380, -0.00000115, -0.00001183, -0.00001594, -0.00000992, -0.00000282, -0.00000229, -0.00000580, -0.00000771, },
		{ -0.00000506, -0.00000848, -0.00000825, -0.00000640, -0.00000719, -0.00001303, -0.00002167, -0.00002791, -0.00002830, -0.00002212, -0.00000836, 0.00001275, 0.00003300, 0.00003914, 0.00002835, 0.00001441, 0.00001099, 0.00001452, 0.00001275, 0.00000473, 0.00000056, 0.00000389, 0.00000761, 0.00000726, 0.00000665, 0.00000801, 0.00000791, 0.00000593, 0.00000669, 0.00000962, 0.00000721, -0.00000196, -0.00000890, -0.00000794, -0.00000470, -0.00000447, -0.00000326, 0.00000254, 0.00000751, 0.00000660, 0.00000377, 0.00000312, 0.00000105, -0.00000387, -0.00000393, 0.00000429, 0.00001051, 0.00000648, -0.00000005, -0.00000016, 0.00000020, -0.00000504, -0.00000544, 0.00000767, 0.00001975, 0.00001442, 0.00000278, 0.00000148, -0.00000395, -0.00003265, -0.00005509, -0.00002137, 0.00005006, 0.00007309, 0.00001159, -0.00005923, -0.00005396, 0.00000585, 0.00003397, 0.00000751, -0.00001704, -0.00000459, 0.00001331, 0.00000557, -0.00000899, -0.00000559, 0.00000173, -0.00000767, -0.00002013, -0.00001305, 0.00000476, 0.00001073, 0.00000651, 0.00000740, 0.00000955, -0.00000130, -0.00001887, -0.00002360, -0.00001444, -0.00000694, -0.00000410, 0.00000472, 0.00001790, 0.00001987, 0.00000727, -0.00000539, -0.00000944, -0.00001112, -0.00001341, -0.00000857, 0.00000479, 0.00001500, 0.00001495, 0.00001138, 0.00000955, 0.00000360, -0.00000920, -0.00001857, -0.00001576, -0.00000712, -0.00000290, -0.00000235, 0.00000013, 0.00000254, -0.00000016, -0.00000532, -0.00000695, -0.00000586, -0.00000601, -0.00000550, -0.00000092, 0.00000307, -0.00000117, -0.00001106, -0.00001556, -0.00001027, -0.00000270, -0.00000116, },
		{ -0.00000271, 0.00000016, -0.00000373, -0.00000865, -0.00000911, -0.00000601, -0.00000564, -0.00001211, -0.00002145, -0.00002592, -0.00002343, -0.00001753, -0.00000832, 0.00000798, 0.00002731, 0.00003581, 0.00002714, 0.00001369, 0.00001062, 0.00001525, 0.00001410, 0.00000531, 0.00000002, 0.00000318, 0.00000732, 0.00000695, 0.00000574, 0.00000640, 0.00000551, 0.00000279, 0.00000401, 0.00000961, 0.00001124, 0.00000527, -0.00000079, -0.00000145, -0.00000215, -0.00000748, -0.00001138, -0.00000769, -0.00000087, 0.00000287, 0.00000555, 0.00000913, 0.00000768, -0.00000113, -0.00000734, -0.00000346, 0.00000258, 0.00000220, 0.00000190, 0.00000836, 0.00001118, 0.00000063, -0.00001043, -0.00000630, 0.00000271, 0.00000049, 0.00000093, 0.00002356, 0.00004204, 0.00000975, -0.00005617, -0.00007503, -0.00001394, 0.00005443, 0.00004875, -0.00000951, -0.00003722, -0.00001215, 0.00001188, 0.00000142, -0.00001395, -0.00000562, 0.00000764, 0.00000237, -0.00000711, -0.00000020, 0.00001057, 0.00000394, -0.00001169, -0.00001523, -0.00000904, -0.00000834, -0.00000986, -0.00000013, 0.00001502, 0.00001734, 0.00000615, -0.00000348, -0.00000802, -0.00001625, -0.00002555, -0.00002169, -0.00000388, 0.00001145, 0.00001483, 0.00001299, 0.00001081, 0.00000314, -0.00000999, -0.00001711, -0.00001202, -0.00000254, 0.00000393, 0.00001007, 0.00001712, 0.00001703, 0.00000564, -0.00000740, -0.00001147, -0.00000907, -0.00000806, -0.00000818, -0.00000500, -0.00000072, -0.00000071, -0.00000388, -0.00000531, -0.00000494, -0.00000551, -0.00000530, -0.00000135, 0.00000232, -0.00000123, -0.00001040, -0.00001531, -0.00001078, },
		{ -0.00001546, -0.00001203, -0.00000356, 0.00000097, -0.00000211, -0.00000778, -0.00000874, -0.00000476, -0.00000385, -0.00001181, -0.00002286, -0.00002615, -0.00002022, -0.00001288, -0.00000645, 0.00000574, 0.00002358, 0.00003319, 0.00002533, 0.00001138, 0.00000830, 0.00001449, 0.00001498, 0.00000627, 0.00000004, 0.00000268, 0.00000698, 0.00000683, 0.00000592, 0.00000709, 0.00000613, 0.00000194, 0.00000104, 0.00000559, 0.00000799, 0.00000426, 0.00000154, 0.00000484, 0.00000707, 0.00000145, -0.00000653, -0.00000912, -0.00000845, -0.00000856, -0.00000550, 0.00000318, 0.00000906, 0.00000568, 0.00000062, 0.00000155, 0.00000159, -0.00000578, -0.00000975, 0.00000025, 0.00001210, 0.00000988, 0.00000268, 0.00000594, 0.00000639, -0.00001546, -0.00003535, -0.00000854, 0.00005007, 0.00006492, 0.00000517, -0.00006018, -0.00005428, 0.00000179, 0.00002873, 0.00000566, -0.00001616, -0.00000535, 0.00000953, 0.00000145, -0.00001121, -0.00000604, 0.00000315, -0.00000320, -0.00001305, -0.00000648, 0.00000783, 0.00000954, 0.00000131, -0.00000143, -0.00000084, -0.00000904, -0.00002039, -0.00001837, -0.00000389, 0.00000707, 0.00000989, 0.00001264, 0.00001420, 0.00000413, -0.00001507, -0.00002643, -0.00002256, -0.00001317, -0.00000586, 0.00000295, 0.00001361, 0.00001678, 0.00000845, -0.00000227, -0.00000716, -0.00000869, -0.00000969, -0.00000559, 0.00000502, 0.00001303, 0.00001107, 0.00000449, 0.00000180, 0.00000169, -0.00000164, -0.00000681, -0.00000832, -0.00000655, -0.00000609, -0.00000724, -0.00000688, -0.00000529, -0.00000481, -0.00000406, -0.00000051, 0.00000260, -0.00000079, -0.00000973, },
		{ 0.00000107, -0.00000746, -0.00001438, -0.00001293, -0.00000519, 0.00000028, -0.00000192, -0.00000765, -0.00000815, -0.00000227, -0.00000003, -0.00000950, -0.00002351, -0.00002760, -0.00001964, -0.00001082, -0.00000591, 0.00000378, 0.00002148, 0.00003297, 0.00002584, 0.00001027, 0.00000546, 0.00001189, 0.00001379, 0.00000590, -0.00000020, 0.00000268, 0.00000721, 0.00000670, 0.00000533, 0.00000692, 0.00000697, 0.00000324, 0.00000187, 0.00000544, 0.00000644, 0.00000079, -0.00000317, 0.00000140, 0.00000770, 0.00000716, 0.00000306, 0.00000155, -0.00000066, -0.00000751, -0.00001244, -0.00000908, -0.00000367, -0.00000304, -0.00000088, 0.00000826, 0.00001310, 0.00000310, -0.00000950, -0.00000825, -0.00000136, -0.00000414, -0.00000439, 0.00001697, 0.00003729, 0.00001335, -0.00004178, -0.00005659, -0.00000132, 0.00005885, 0.00005095, -0.00000501, -0.00003352, -0.00001442, 0.00000434, -0.00000614, -0.00001844, -0.00000790, 0.00000655, 0.00000276, -0.00000616, -0.00000127, 0.00000659, -0.00000050, -0.00001338, -0.00001282, -0.00000268, 0.00000085, -0.00000097, 0.00000356, 0.00001027, 0.00000523, -0.00000897, -0.00001647, -0.00001368, -0.00001033, -0.00000802, 0.00000097, 0.00001383, 0.00001658, 0.00000592, -0.00000597, -0.00001127, -0.00001489, -0.00001936, -0.00001756, -0.00000661, 0.00000455, 0.00000841, 0.00000817, 0.00000790, 0.00000449, -0.00000339, -0.00000866, -0.00000618, -0.00000128, -0.00000035, -0.00000038, 0.00000411, 0.00001003, 0.00001033, 0.00000503, -0.00000029, -0.00000372, -0.00000739, -0.00001096, -0.00001183, -0.00001019, -0.00000825, -0.00000552, -0.00000057, 0.00000347, },
		{ -0.00000525, 0.00000058, 0.00000087, -0.00000529, -0.00001179, -0.00001166, -0.00000509, 0.00000014, -0.00000252, -0.00000935, -0.00000980, -0.00000138, 0.00000389, -0.00000515, -0.00002121, -0.00002670, -0.00001829, -0.00000964, -0.00000751, -0.00000092, 0.00001708, 0.00003211, 0.00002787, 0.00001211, 0.00000549, 0.00001078, 0.00001226, 0.00000375, -0.00000273, 0.00000103, 0.00000714, 0.00000741, 0.00000571, 0.00000674, 0.00000653, 0.00000269, 0.00000171, 0.00000660, 0.00000888, 0.00000272, -0.00000384, -0.00000233, 0.00000226, 0.00000223, 0.00000131, 0.00000542, 0.00000903, 0.00000488, -0.00000209, -0.00000478, -0.00000779, -0.00001558, -0.00001760, -0.00000473, 0.00001007, 0.00001017, 0.00000367, 0.00000620, 0.00000659, -0.00001391, -0.00003417, -0.00001239, 0.00003990, 0.00005470, 0.00000312, -0.00005312, -0.00004448, 0.00000986, 0.00003746, 0.00001834, -0.00000241, 0.00000285, 0.00000930, -0.00000459, -0.00001940, -0.00001424, -0.00000286, -0.00000463, -0.00001020, -0.00000317, 0.00000773, 0.00000528, -0.00000530, -0.00000771, -0.00000357, -0.00000526, -0.00001015, -0.00000592, 0.00000491, 0.00000824, 0.00000235, -0.00000191, -0.00000233, -0.00000675, -0.00001424, -0.00001390, -0.00000424, 0.00000342, 0.00000367, 0.00000329, 0.00000559, 0.00000352, -0.00000616, -0.00001514, -0.00001625, -0.00001284, -0.00000948, -0.00000395, 0.00000487, 0.00001066, 0.00000869, 0.00000338, 0.00000024, -0.00000273, -0.00000841, -0.00001201, -0.00000730, 0.00000294, 0.00001038, 0.00001188, 0.00001032, 0.00000727, 0.00000171, -0.00000535, -0.00001070, -0.00001316, -0.00001346, -0.00001103, },
		{ -0.00001117, -0.00001275, -0.00001003, -0.00000550, -0.00000409, -0.00000771, -0.00001217, -0.00001115, -0.00000381, 0.00000231, -0.00000078, -0.00000989, -0.00001227, -0.00000305, 0.00000486, -0.00000265, -0.00001889, -0.00002401, -0.00001390, -0.00000509, -0.00000678, -0.00000564, 0.00001007, 0.00002732, 0.00002665, 0.00001282, 0.00000686, 0.00001280, 0.00001427, 0.00000395, -0.00000515, -0.00000261, 0.00000426, 0.00000607, 0.00000571, 0.00000772, 0.00000764, 0.00000263, 0.00000010, 0.00000485, 0.00000879, 0.00000466, -0.00000084, 0.00000044, 0.00000325, -0.00000019, -0.00000461, -0.00000150, 0.00000487, 0.00000629, 0.00000601, 0.00000904, 0.00000718, -0.00000683, -0.00002029, -0.00001789, -0.00000857, -0.00000843, -0.00000705, 0.00001404, 0.00003496, 0.00001517, -0.00003500, -0.00005048, -0.00000284, 0.00004962, 0.00004079, -0.00001084, -0.00003588, -0.00001528, 0.00000737, 0.00000455, -0.00000143, 0.00000914, 0.00001832, 0.00000798, -0.00000694, -0.00000721, -0.00000196, -0.00000735, -0.00001577, -0.00001177, -0.00000112, 0.00000074, -0.00000380, -0.00000209, 0.00000351, 0.00000092, -0.00000798, -0.00001032, -0.00000492, -0.00000204, -0.00000325, -0.00000037, 0.00000626, 0.00000648, -0.00000152, -0.00000794, -0.00000840, -0.00000912, -0.00001213, -0.00000998, -0.00000012, 0.00000831, 0.00000780, 0.00000211, -0.00000299, -0.00000837, -0.00001475, -0.00001674, -0.00001071, -0.00000171, 0.00000398, 0.00000687, 0.00000895, 0.00000699, -0.00000169, -0.00001138, -0.00001363, -0.00000790, -0.00000045, 0.00000490, 0.00000886, 0.00001146, 0.00001061, 0.00000593, -0.00000040, -0.00000646, },
		{ 0.00000750, 0.00000368, -0.00000123, -0.00000511, -0.00000636, -0.00000609, -0.00000751, -0.00001221, -0.00001694, -0.00001562, -0.00000649, 0.00000273, 0.00000180, -0.00000781, -0.00001200, -0.00000337, 0.00000517, -0.00000254, -0.00001987, -0.00002432, -0.00001066, 0.00000124, -0.00000174, -0.00000549, 0.00000619, 0.00002266, 0.00002268, 0.00000914, 0.00000400, 0.00001261, 0.00001718, 0.00000772, -0.00000314, -0.00000312, 0.00000191, 0.00000271, 0.00000241, 0.00000595, 0.00000804, 0.00000406, 0.00000087, 0.00000436, 0.00000742, 0.00000294, -0.00000202, 0.00000130, 0.00000679, 0.00000444, -0.00000172, -0.00000237, -0.00000036, -0.00000231, -0.00000227, 0.00000690, 0.00001477, 0.00000839, -0.00000281, -0.00000333, -0.00000365, -0.00002220, -0.00004029, -0.00001896, 0.00003192, 0.00004946, 0.00000561, -0.00004427, -0.00003675, 0.00001087, 0.00003281, 0.00001159, -0.00001039, -0.00000645, 0.00000196, -0.00000483, -0.00001109, -0.00000081, 0.00001150, 0.00000848, 0.00000004, 0.00000215, 0.00000761, 0.00000185, -0.00000933, -0.00001132, -0.00000677, -0.00000785, -0.00001212, -0.00000813, 0.00000184, 0.00000501, 0.00000007, -0.00000301, -0.00000235, -0.00000525, -0.00001103, -0.00001008, -0.00000116, 0.00000576, 0.00000581, 0.00000441, 0.00000372, -0.00000191, -0.00001285, -0.00001926, -0.00001491, -0.00000575, 0.00000058, 0.00000431, 0.00000718, 0.00000584, -0.00000168, -0.00000987, -0.00001269, -0.00001114, -0.00000851, -0.00000397, 0.00000315, 0.00000775, 0.00000467, -0.00000309, -0.00000799, -0.00000764, -0.00000546, -0.00000362, -0.00000074, 0.00000375, 0.00000773, 0.00000907, },
		{ -0.00000106, 0.00000215, 0.00000439, 0.00000506, 0.00000449, 0.00000369, 0.00000320, 0.00000200, -0.00000202, -0.00000983, -0.00001836, -0.00002071, -0.00001308, -0.00000226, -0.00000035, -0.00000820, -0.00001155, -0.00000148, 0.00000854, -0.00000006, -0.00002073, -0.00002735, -0.00001191, 0.00000363, 0.00000200, -0.00000335, 0.00000681, 0.00002331, 0.00002288, 0.00000670, -0.00000129, 0.00000732, 0.00001471, 0.00000833, -0.00000077, -0.00000005, 0.00000464, 0.00000356, 0.00000051, 0.00000231, 0.00000457, 0.00000191, 0.00000030, 0.00000539, 0.00000928, 0.00000374, -0.00000363, -0.00000188, 0.00000436, 0.00000456, 0.00000159, 0.00000366, 0.00000580, -0.00000049, -0.00000746, -0.00000330, 0.00000405, 0.00000095, -0.00000195, 0.00001319, 0.00002853, 0.00000690, -0.00004202, -0.00005766, -0.00001324, 0.00003766, 0.00003355, -0.00000964, -0.00002936, -0.00000917, 0.00001047, 0.00000477, -0.00000480, 0.00000131, 0.00000831, 0.00000057, -0.00000899, -0.00000482, 0.00000330, 0.00000073, -0.00000503, -0.00000024, 0.00000897, 0.00000839, 0.00000094, -0.00000123, 0.00000011, -0.00000524, -0.00001443, -0.00001535, -0.00000775, -0.00000244, -0.00000212, 0.00000004, 0.00000401, 0.00000179, -0.00000668, -0.00001142, -0.00000832, -0.00000377, -0.00000144, 0.00000267, 0.00000861, 0.00000856, -0.00000051, -0.00001059, -0.00001412, -0.00001312, -0.00001172, -0.00000835, -0.00000138, 0.00000472, 0.00000521, 0.00000198, -0.00000096, -0.00000383, -0.00000768, -0.00000940, -0.00000621, -0.00000151, -0.00000068, -0.00000319, -0.00000424, -0.00000256, -0.00000142, -0.00000273, -0.00000438, -0.00000380, },
		{ -0.00000149, -0.00000376, -0.00000437, -0.00000368, -0.00000270, -0.00000164, 0.00000024, 0.00000343, 0.00000708, 0.00000921, 0.00000746, 0.00000020, -0.00001058, -0.00001764, -0.00001474, -0.00000655, -0.00000552, -0.00001380, -0.00001656, -0.00000339, 0.00001115, 0.00000476, -0.00001777, -0.00002751, -0.00001297, 0.00000340, 0.00000143, -0.00000551, 0.00000494, 0.00002479, 0.00002746, 0.00001063, -0.00000102, 0.00000431, 0.00001032, 0.00000371, -0.00000470, -0.00000159, 0.00000657, 0.00000777, 0.00000446, 0.00000415, 0.00000367, -0.00000164, -0.00000464, 0.00000160, 0.00000867, 0.00000585, -0.00000105, -0.00000075, 0.00000308, 0.00000074, -0.00000308, 0.00000177, 0.00000925, 0.00000705, 0.00000086, 0.00000316, 0.00000458, -0.00001113, -0.00002701, -0.00000804, 0.00003612, 0.00004786, 0.00000250, -0.00004773, -0.00004371, -0.00000108, 0.00001988, 0.00000321, -0.00001286, -0.00000537, 0.00000456, -0.00000160, -0.00000906, -0.00000218, 0.00000703, 0.00000365, -0.00000349, -0.00000101, 0.00000398, -0.00000074, -0.00000855, -0.00000620, 0.00000233, 0.00000445, 0.00000144, 0.00000319, 0.00000758, 0.00000427, -0.00000553, -0.00001071, -0.00000911, -0.00000848, -0.00001014, -0.00000732, -0.00000005, 0.00000298, -0.00000087, -0.00000449, -0.00000415, -0.00000434, -0.00000661, -0.00000538, 0.00000105, 0.00000558, 0.00000330, -0.00000138, -0.00000342, -0.00000495, -0.00000902, -0.00001251, -0.00001130, -0.00000709, -0.00000368, -0.00000087, 0.00000261, 0.00000443, 0.00000218, -0.00000172, -0.00000350, -0.00000400, -0.00000628, -0.00000929, -0.00000889, -0.00000441, -0.00000000, 0.00000074, },
		{ -0.00000368, -0.00000078, -0.00000103, -0.00000249, -0.00000338, -0.00000389, -0.00000513, -0.00000668, -0.00000649, -0.00000323, 0.00000213, 0.00000739, 0.00001008, 0.00000747, -0.00000060, -0.00000804, -0.00000748, -0.00000200, -0.00000420, -0.00001687, -0.00002350, -0.00001061, 0.00000752, 0.00000509, -0.00001606, -0.00002591, -0.00001054, 0.00000665, 0.00000250, -0.00000921, -0.00000156, 0.00002051, 0.00002806, 0.00001449, 0.00000341, 0.00000815, 0.00001279, 0.00000310, -0.00000931, -0.00000838, 0.00000102, 0.00000570, 0.00000619, 0.00000880, 0.00000925, 0.00000184, -0.00000542, -0.00000284, 0.00000316, 0.00000161, -0.00000260, 0.00000107, 0.00000716, 0.00000371, -0.00000418, -0.00000310, 0.00000293, 0.00000137, -0.00000042, 0.00001287, 0.00002632, 0.00000776, -0.00003363, -0.00004401, -0.00000112, 0.00004465, 0.00003763, -0.00000600, -0.00002806, -0.00001347, 0.00000112, -0.00000580, -0.00001383, -0.00000578, 0.00000340, -0.00000154, -0.00000900, -0.00000454, 0.00000339, 0.00000181, -0.00000298, 0.00000030, 0.00000566, 0.00000177, -0.00000630, -0.00000640, -0.00000077, -0.00000011, -0.00000330, -0.00000103, 0.00000549, 0.00000643, 0.00000110, -0.00000189, -0.00000099, -0.00000310, -0.00000900, -0.00001116, -0.00000744, -0.00000444, -0.00000504, -0.00000430, -0.00000039, 0.00000118, -0.00000204, -0.00000462, -0.00000302, -0.00000139, -0.00000309, -0.00000437, -0.00000163, 0.00000160, 0.00000040, -0.00000395, -0.00000737, -0.00000936, -0.00001097, -0.00001048, -0.00000613, -0.00000042, 0.00000283, 0.00000363, 0.00000389, 0.00000271, -0.00000200, -0.00000835, -0.00001127, -0.00000864, },
		{ -0.00000647, -0.00000646, -0.00000367, -0.00000181, -0.00000240, -0.00000375, -0.00000396, -0.00000364, -0.00000489, -0.00000776, -0.00000955, -0.00000819, -0.00000425, 0.00000067, 0.00000503, 0.00000602, 0.00000190, -0.00000264, 0.00000006, 0.00000742, 0.00000549, -0.00001033, -0.00002253, -0.00001398, 0.00000304, 0.00000093, -0.00002050, -0.00002986, -0.00001105, 0.00001076, 0.00000767, -0.00000800, -0.00000533, 0.00001503, 0.00002385, 0.00001192, 0.00000250, 0.00001027, 0.00001849, 0.00000983, -0.00000544, -0.00000925, -0.00000383, -0.00000125, -0.00000046, 0.00000534, 0.00001072, 0.00000720, 0.00000079, 0.00000147, 0.00000388, -0.00000208, -0.00000958, -0.00000551, 0.00000474, 0.00000605, 0.00000112, 0.00000306, 0.00000603, -0.00000535, -0.00001864, -0.00000362, 0.00003243, 0.00004020, -0.00000058, -0.00004260, -0.00003442, 0.00000726, 0.00002707, 0.00001156, -0.00000389, 0.00000078, 0.00000582, -0.00000450, -0.00001497, -0.00001072, -0.00000309, -0.00000597, -0.00001126, -0.00000672, 0.00000110, 0.00000070, -0.00000307, -0.00000006, 0.00000515, 0.00000235, -0.00000462, -0.00000489, -0.00000004, 0.00000010, -0.00000411, -0.00000396, 0.00000074, 0.00000194, -0.00000110, -0.00000105, 0.00000307, 0.00000392, -0.00000078, -0.00000484, -0.00000540, -0.00000681, -0.00001036, -0.00001057, -0.00000529, -0.00000019, 0.00000060, 0.00000008, 0.00000084, -0.00000030, -0.00000531, -0.00000937, -0.00000757, -0.00000233, 0.00000100, 0.00000135, 0.00000051, -0.00000178, -0.00000626, -0.00001041, -0.00001098, -0.00000841, -0.00000524, -0.00000191, 0.00000209, 0.00000477, 0.00000309, -0.00000220, },
		{ -0.00000021, -0.00000211, -0.00000394, -0.00000284, 0.00000029, 0.00000198, 0.00000046, -0.00000231, -0.00000350, -0.00000341, -0.00000479, -0.00000826, -0.00001102, -0.00001074, -0.00000817, -0.00000468, -0.00000087, 0.00000090, -0.00000190, -0.00000517, -0.00000030, 0.00001109, 0.00001359, -0.00000038, -0.00001403, -0.00000790, 0.00000733, 0.00000257, -0.00002369, -0.00003741, -0.00001835, 0.00000808, 0.00000910, -0.00000623, -0.00000512, 0.00001470, 0.00002336, 0.00000948, -0.00000266, 0.00000512, 0.00001708, 0.00001310, 0.00000060, -0.00000281, 0.00000098, -0.00000021, -0.00000437, -0.00000208, 0.00000316, 0.00000244, 0.00000049, 0.00000616, 0.00001195, 0.00000510, -0.00000762, -0.00000966, -0.00000351, -0.00000357, -0.00000544, 0.00000651, 0.00002163, 0.00001036, -0.00002251, -0.00003127, 0.00000415, 0.00004069, 0.00003108, -0.00000849, -0.00002640, -0.00001110, 0.00000358, -0.00000070, -0.00000489, 0.00000501, 0.00001350, 0.00000652, -0.00000421, -0.00000471, -0.00000231, -0.00000814, -0.00001526, -0.00001251, -0.00000501, -0.00000348, -0.00000489, -0.00000075, 0.00000499, 0.00000300, -0.00000357, -0.00000442, -0.00000005, 0.00000067, -0.00000283, -0.00000322, -0.00000003, -0.00000018, -0.00000421, -0.00000485, -0.00000024, 0.00000314, 0.00000215, 0.00000098, 0.00000142, -0.00000083, -0.00000691, -0.00001097, -0.00000935, -0.00000575, -0.00000378, -0.00000163, 0.00000193, 0.00000286, -0.00000167, -0.00000755, -0.00000915, -0.00000673, -0.00000398, -0.00000190, 0.00000039, 0.00000168, 0.00000006, -0.00000340, -0.00000602, -0.00000725, -0.00000797, -0.00000757, -0.00000485, -0.00000133, },
		{ -0.00000580, -0.00000515, -0.00000539, -0.00000718, -0.00000814, -0.00000564, -0.00000041, 0.00000369, 0.00000390, 0.00000181, 0.00000075, 0.00000079, -0.00000119, -0.00000611, -0.00001068, -0.00001207, -0.00001103, -0.00000886, -0.00000560, -0.00000349, -0.00000612, -0.00001013, -0.00000570, 0.00000749, 0.00001371, 0.00000299, -0.00000881, -0.00000064, 0.00001747, 0.00001357, -0.00001671, -0.00003738, -0.00002309, 0.00000271, 0.00000438, -0.00001138, -0.00000993, 0.00001320, 0.00002603, 0.00001287, -0.00000257, 0.00000148, 0.00001212, 0.00000888, -0.00000176, -0.00000192, 0.00000601, 0.00000747, 0.00000245, 0.00000089, 0.00000097, -0.00000471, -0.00000933, -0.00000165, 0.00001047, 0.00001047, 0.00000185, 0.00000044, 0.00000252, -0.00000826, -0.00002282, -0.00001206, 0.00002099, 0.00003264, 0.00000195, -0.00003179, -0.00002385, 0.00001145, 0.00002648, 0.00001099, -0.00000337, -0.00000006, 0.00000292, -0.00000664, -0.00001340, -0.00000490, 0.00000620, 0.00000595, 0.00000156, 0.00000391, 0.00000669, 0.00000002, -0.00000997, -0.00001211, -0.00000912, -0.00000965, -0.00001115, -0.00000623, 0.00000128, 0.00000197, -0.00000254, -0.00000311, 0.00000060, 0.00000110, -0.00000224, -0.00000276, 0.00000023, 0.00000024, -0.00000391, -0.00000581, -0.00000297, -0.00000042, -0.00000075, -0.00000024, 0.00000254, 0.00000302, -0.00000103, -0.00000518, -0.00000582, -0.00000530, -0.00000595, -0.00000540, -0.00000195, 0.00000069, -0.00000114, -0.00000511, -0.00000674, -0.00000585, -0.00000508, -0.00000474, -0.00000335, -0.00000131, -0.00000043, -0.00000079, -0.00000119, -0.00000175, -0.00000336, -0.00000527, },
		{ -0.00000005, -0.00000095, -0.00000188, -0.00000271, -0.00000472, -0.00000831, -0.00001125, -0.00001043, -0.00000566, -0.00000063, 0.00000123, 0.00000100, 0.00000221, 0.00000488, 0.00000495, 0.00000059, -0.00000493, -0.00000828, -0.00000995, -0.00001067, -0.00000935, -0.00000766, -0.00001021, -0.00001493, -0.00001147, 0.00000210, 0.00000998, 0.00000040, -0.00001112, -0.00000099, 0.00002227, 0.00002382, -0.00000505, -0.00002855, -0.00001759, 0.00000610, 0.00000457, -0.00001685, -0.00002004, 0.00000383, 0.00002191, 0.00001364, -0.00000009, 0.00000374, 0.00001350, 0.00000806, -0.00000602, -0.00000827, 0.00000117, 0.00000698, 0.00000695, 0.00000927, 0.00001041, 0.00000127, -0.00001045, -0.00000957, -0.00000074, -0.00000012, -0.00000366, 0.00000588, 0.00002031, 0.00001059, -0.00002151, -0.00003370, -0.00000454, 0.00002906, 0.00002365, -0.00000831, -0.00002202, -0.00000748, 0.00000598, 0.00000306, 0.00000028, 0.00000828, 0.00001257, 0.00000278, -0.00000783, -0.00000630, -0.00000069, -0.00000211, -0.00000470, 0.00000067, 0.00000778, 0.00000624, -0.00000045, -0.00000291, -0.00000288, -0.00000725, -0.00001286, -0.00001172, -0.00000610, -0.00000449, -0.00000652, -0.00000481, 0.00000058, 0.00000240, -0.00000043, -0.00000163, 0.00000037, 0.00000022, -0.00000344, -0.00000523, -0.00000295, -0.00000113, -0.00000223, -0.00000266, -0.00000037, 0.00000106, -0.00000075, -0.00000260, -0.00000182, -0.00000091, -0.00000229, -0.00000371, -0.00000275, -0.00000170, -0.00000352, -0.00000649, -0.00000703, -0.00000525, -0.00000400, -0.00000410, -0.00000411, -0.00000351, -0.00000305, -0.00000265, -0.00000158, -0.00000028, },
		{ -0.00000305, -0.00000147, -0.00000023, 0.00000018, 0.00000022, -0.00000005, -0.00000165, -0.00000526, -0.00000923, -0.00001031, -0.00000748, -0.00000395, -0.00000337, -0.00000455, -0.00000306, 0.00000171, 0.00000514, 0.00000424, 0.00000150, -0.00000029, -0.00000227, -0.00000526, -0.00000704, -0.00000798, -0.00001272, -0.00001979, -0.00001809, -0.00000423, 0.00000553, -0.00000320, -0.00001589, -0.00000633, 0.00001977, 0.00002613, 0.00000034, -0.00002234, -0.00000992, 0.00001659, 0.00001531, -0.00001151, -0.00002286, -0.00000418, 0.00001334, 0.00000648, -0.00000555, 0.00000134, 0.00001555, 0.00001290, -0.00000247, -0.00000887, -0.00000353, -0.00000005, 0.00000030, 0.00000661, 0.00001430, 0.00001064, 0.00000056, -0.00000067, 0.00000229, -0.00000779, -0.00002267, -0.00001342, 0.00001851, 0.00003157, 0.00000416, -0.00002836, -0.00002371, 0.00000657, 0.00001963, 0.00000568, -0.00000726, -0.00000429, -0.00000062, -0.00000626, -0.00000834, 0.00000157, 0.00001047, 0.00000719, 0.00000079, 0.00000197, 0.00000450, -0.00000059, -0.00000724, -0.00000564, 0.00000028, 0.00000127, -0.00000077, 0.00000121, 0.00000458, 0.00000192, -0.00000468, -0.00000741, -0.00000672, -0.00000893, -0.00001300, -0.00001175, -0.00000512, -0.00000049, -0.00000036, -0.00000003, 0.00000202, 0.00000147, -0.00000260, -0.00000494, -0.00000309, -0.00000109, -0.00000182, -0.00000257, -0.00000132, -0.00000080, -0.00000282, -0.00000432, -0.00000269, -0.00000024, 0.00000014, -0.00000034, 0.00000026, 0.00000025, -0.00000264, -0.00000666, -0.00000810, -0.00000659, -0.00000478, -0.00000418, -0.00000410, -0.00000395, -0.00000397, -0.00000393, },
		{ -0.00000512, -0.00000511, -0.00000433, -0.00000272, -0.00000101, 0.00000011, 0.00000063, 0.00000068, -0.00000036, -0.00000326, -0.00000692, -0.00000817, -0.00000555, -0.00000229, -0.00000281, -0.00000628, -0.00000743, -0.00000436, -0.00000115, -0.00000077, -0.00000068, 0.00000158, 0.00000364, 0.00000323, 0.00000221, 0.00000046, -0.00000693, -0.00001868, -0.00002204, -0.00001081, -0.00000065, -0.00000846, -0.00002169, -0.00001262, 0.00001534, 0.00002475, -0.00000006, -0.00002367, -0.00000997, 0.00002191, 0.00002607, -0.00000017, -0.00001532, -0.00000074, 0.00001383, 0.00000323, -0.00001345, -0.00000871, 0.00000814, 0.00001081, -0.00000016, -0.00000446, 0.00000065, 0.00000117, -0.00000342, -0.00000103, 0.00000616, 0.00000515, 0.00000063, 0.00000872, 0.00002108, 0.00000993, -0.00002255, -0.00003552, -0.00000824, 0.00002442, 0.00002124, -0.00000667, -0.00001820, -0.00000446, 0.00000725, 0.00000278, -0.00000274, 0.00000111, 0.00000301, -0.00000464, -0.00001023, -0.00000477, 0.00000224, 0.00000113, -0.00000118, 0.00000380, 0.00000962, 0.00000683, -0.00000019, -0.00000218, -0.00000099, -0.00000347, -0.00000700, -0.00000450, 0.00000175, 0.00000390, 0.00000190, 0.00000140, 0.00000157, -0.00000321, -0.00001131, -0.00001477, -0.00001168, -0.00000791, -0.00000627, -0.00000350, 0.00000095, 0.00000244, -0.00000048, -0.00000306, -0.00000226, -0.00000100, -0.00000175, -0.00000236, -0.00000115, -0.00000069, -0.00000286, -0.00000500, -0.00000426, -0.00000215, -0.00000118, -0.00000065, 0.00000090, 0.00000182, -0.00000026, -0.00000412, -0.00000629, -0.00000576, -0.00000454, -0.00000425, -0.00000458, -0.00000489, },
		{ -0.00000257, -0.00000380, -0.00000506, -0.00000605, -0.00000626, -0.00000522, -0.00000322, -0.00000112, 0.00000041, 0.00000124, 0.00000085, -0.00000173, -0.00000572, -0.00000768, -0.00000513, -0.00000091, -0.00000030, -0.00000356, -0.00000565, -0.00000439, -0.00000369, -0.00000620, -0.00000817, -0.00000580, -0.00000122, 0.00000260, 0.00000673, 0.00001002, 0.00000550, -0.00000751, -0.00001561, -0.00000952, -0.00000282, -0.00001330, -0.00002942, -0.00002170, 0.00000861, 0.00002221, -0.00000113, -0.00002695, -0.00001479, 0.00001975, 0.00002857, 0.00000483, -0.00000993, 0.00000607, 0.00002307, 0.00001197, -0.00000999, -0.00001206, 0.00000046, 0.00000192, -0.00000816, -0.00000934, 0.00000083, 0.00000579, 0.00000251, 0.00000312, 0.00000584, -0.00000286, -0.00001504, -0.00000511, 0.00002394, 0.00003312, 0.00000401, -0.00002861, -0.00002540, 0.00000186, 0.00001337, 0.00000132, -0.00000784, -0.00000149, 0.00000458, -0.00000009, -0.00000426, 0.00000027, 0.00000365, -0.00000172, -0.00000691, -0.00000392, -0.00000018, -0.00000342, -0.00000690, -0.00000207, 0.00000573, 0.00000709, 0.00000423, 0.00000434, 0.00000538, 0.00000106, -0.00000583, -0.00000753, -0.00000427, -0.00000222, -0.00000163, 0.00000178, 0.00000612, 0.00000473, -0.00000231, -0.00000792, -0.00000915, -0.00000983, -0.00001145, -0.00001037, -0.00000574, -0.00000215, -0.00000209, -0.00000237, -0.00000059, 0.00000078, -0.00000032, -0.00000166, -0.00000119, -0.00000085, -0.00000257, -0.00000441, -0.00000397, -0.00000239, -0.00000186, -0.00000180, -0.00000071, 0.00000030, -0.00000093, -0.00000373, -0.00000519, -0.00000419, -0.00000246, -0.00000186, },
		{ -0.00000327, -0.00000105, -0.00000027, -0.00000056, -0.00000152, -0.00000312, -0.00000501, -0.00000621, -0.00000597, -0.00000461, -0.00000279, -0.00000076, 0.00000082, 0.00000017, -0.00000314, -0.00000592, -0.00000473, -0.00000127, -0.00000057, -0.00000316, -0.00000403, -0.00000127, 0.00000008, -0.00000382, -0.00000915, -0.00001080, -0.00000949, -0.00000701, -0.00000106, 0.00000761, 0.00001013, 0.00000251, -0.00000358, 0.00000209, 0.00000688, -0.00000791, -0.00003096, -0.00002950, -0.00000065, 0.00001582, -0.00000508, -0.00003145, -0.00002010, 0.00001636, 0.00002792, 0.00000415, -0.00001236, 0.00000483, 0.00002716, 0.00002126, 0.00000052, -0.00000363, 0.00000570, 0.00000276, -0.00001313, -0.00001913, -0.00000999, -0.00000232, -0.00000021, 0.00000834, 0.00001855, 0.00000857, -0.00001828, -0.00002612, 0.00000089, 0.00002917, 0.00002282, -0.00000493, -0.00001607, -0.00000475, 0.00000312, -0.00000338, -0.00000805, -0.00000137, 0.00000447, 0.00000069, -0.00000345, -0.00000025, 0.00000262, -0.00000149, -0.00000535, -0.00000219, 0.00000114, -0.00000274, -0.00000789, -0.00000576, 0.00000007, 0.00000163, 0.00000069, 0.00000348, 0.00000765, 0.00000637, 0.00000084, -0.00000221, -0.00000229, -0.00000387, -0.00000612, -0.00000402, 0.00000144, 0.00000370, 0.00000128, -0.00000092, -0.00000103, -0.00000275, -0.00000731, -0.00001053, -0.00001006, -0.00000879, -0.00000872, -0.00000742, -0.00000338, 0.00000042, 0.00000123, 0.00000052, 0.00000042, -0.00000008, -0.00000227, -0.00000431, -0.00000395, -0.00000214, -0.00000116, -0.00000103, -0.00000053, -0.00000035, -0.00000216, -0.00000537, -0.00000718, -0.00000605, },
		{ -0.00000779, -0.00000838, -0.00000668, -0.00000448, -0.00000278, -0.00000127, 0.00000026, 0.00000098, 0.00000014, -0.00000173, -0.00000351, -0.00000488, -0.00000581, -0.00000549, -0.00000364, -0.00000215, -0.00000280, -0.00000375, -0.00000188, 0.00000124, 0.00000061, -0.00000385, -0.00000587, -0.00000235, 0.00000140, -0.00000004, -0.00000429, -0.00000692, -0.00000884, -0.00001129, -0.00000987, -0.00000280, 0.00000181, -0.00000146, -0.00000229, 0.00000945, 0.00002010, 0.00000764, -0.00001911, -0.00002502, -0.00000205, 0.00001190, -0.00001083, -0.00004001, -0.00002963, 0.00001026, 0.00002669, 0.00000391, -0.00001566, -0.00000057, 0.00002373, 0.00002159, 0.00000364, 0.00000207, 0.00001487, 0.00001414, -0.00000366, -0.00001512, -0.00001290, -0.00001287, -0.00001684, -0.00000612, 0.00001828, 0.00002517, 0.00000076, -0.00002395, -0.00001709, 0.00000755, 0.00001530, 0.00000307, -0.00000402, 0.00000252, 0.00000590, -0.00000203, -0.00000811, -0.00000377, 0.00000118, -0.00000132, -0.00000413, -0.00000074, 0.00000226, -0.00000114, -0.00000451, -0.00000151, 0.00000195, -0.00000123, -0.00000626, -0.00000535, -0.00000133, -0.00000128, -0.00000344, -0.00000148, 0.00000323, 0.00000444, 0.00000234, 0.00000205, 0.00000336, 0.00000159, -0.00000272, -0.00000421, -0.00000225, -0.00000157, -0.00000317, -0.00000279, 0.00000067, 0.00000265, 0.00000052, -0.00000290, -0.00000520, -0.00000788, -0.00001158, -0.00001307, -0.00001018, -0.00000549, -0.00000226, -0.00000035, 0.00000137, 0.00000179, -0.00000022, -0.00000286, -0.00000353, -0.00000239, -0.00000129, -0.00000059, 0.00000037, 0.00000086, -0.00000085, -0.00000457, },
		{ 0.00000030, -0.00000235, -0.00000539, -0.00000689, -0.00000680, -0.00000658, -0.00000682, -0.00000642, -0.00000431, -0.00000123, 0.00000117, 0.00000237, 0.00000261, 0.00000142, -0.00000161, -0.00000497, -0.00000667, -0.00000716, -0.00000792, -0.00000737, -0.00000274, 0.00000353, 0.00000483, -0.00000013, -0.00000453, -0.00000347, -0.00000087, -0.00000179, -0.00000403, -0.00000380, -0.00000360, -0.00000638, -0.00000800, -0.00000517, -0.00000466, -0.00001133, -0.00001357, 0.00000110, 0.00001948, 0.00001582, -0.00000593, -0.00001147, 0.00001004, 0.00002181, -0.00000603, -0.00004363, -0.00004021, -0.00000088, 0.00001982, 0.00000021, -0.00001958, -0.00000496, 0.00002085, 0.00002001, 0.00000113, -0.00000122, 0.00001414, 0.00001841, 0.00000527, -0.00000244, 0.00000184, -0.00000263, -0.00002083, -0.00002599, -0.00000391, 0.00001866, 0.00001330, -0.00000721, -0.00001236, -0.00000122, 0.00000332, -0.00000401, -0.00000619, 0.00000280, 0.00000856, 0.00000330, -0.00000230, -0.00000020, 0.00000209, -0.00000198, -0.00000584, -0.00000337, -0.00000029, -0.00000244, -0.00000464, -0.00000124, 0.00000251, -0.00000001, -0.00000465, -0.00000406, -0.00000054, -0.00000083, -0.00000385, -0.00000367, -0.00000078, -0.00000030, -0.00000181, -0.00000050, 0.00000332, 0.00000459, 0.00000256, 0.00000130, 0.00000146, -0.00000066, -0.00000516, -0.00000706, -0.00000415, -0.00000026, 0.00000130, 0.00000159, 0.00000162, -0.00000063, -0.00000590, -0.00001077, -0.00001199, -0.00001051, -0.00000863, -0.00000632, -0.00000299, -0.00000020, 0.00000021, -0.00000090, -0.00000139, -0.00000092, -0.00000060, -0.00000053, 0.00000011, 0.00000094, },
		{ 0.00000117, 0.00000096, 0.00000006, -0.00000200, -0.00000410, -0.00000466, -0.00000390, -0.00000386, -0.00000574, -0.00000813, -0.00000882, -0.00000747, -0.00000511, -0.00000198, 0.00000200, 0.00000519, 0.00000534, 0.00000269, -0.00000077, -0.00000475, -0.00000980, -0.00001293, -0.00000966, -0.00000156, 0.00000361, 0.00000233, -0.00000011, 0.00000093, 0.00000177, -0.00000200, -0.00000672, -0.00000697, -0.00000487, -0.00000461, -0.00000346, 0.00000094, 0.00000068, -0.00001039, -0.00001968, -0.00001065, 0.00000719, 0.00000759, -0.00000867, -0.00000854, 0.00001988, 0.00003847, 0.00001234, -0.00003142, -0.00003784, -0.00000528, 0.00001312, -0.00000796, -0.00002979, -0.00001465, 0.00001566, 0.00001946, 0.00000130, -0.00000307, 0.00001080, 0.00001428, 0.00000055, -0.00000381, 0.00001124, 0.00001873, 0.00000038, -0.00002110, -0.00001825, -0.00000101, 0.00000384, -0.00000419, -0.00000554, 0.00000220, 0.00000303, -0.00000613, -0.00001014, -0.00000292, 0.00000346, 0.00000139, -0.00000084, 0.00000320, 0.00000642, 0.00000253, -0.00000259, -0.00000262, -0.00000175, -0.00000481, -0.00000707, -0.00000338, 0.00000127, 0.00000016, -0.00000347, -0.00000279, 0.00000060, 0.00000052, -0.00000242, -0.00000296, -0.00000133, -0.00000212, -0.00000490, -0.00000477, -0.00000124, 0.00000138, 0.00000192, 0.00000335, 0.00000557, 0.00000447, -0.00000065, -0.00000500, -0.00000541, -0.00000395, -0.00000292, -0.00000134, 0.00000134, 0.00000246, 0.00000011, -0.00000368, -0.00000616, -0.00000755, -0.00000911, -0.00001003, -0.00000892, -0.00000656, -0.00000473, -0.00000350, -0.00000176, 0.00000035, 0.00000155, 0.00000151, },
		{ 0.00000024, 0.00000246, 0.00000327, 0.00000309, 0.00000162, -0.00000126, -0.00000413, -0.00000483, -0.00000311, -0.00000141, -0.00000202, -0.00000443, -0.00000662, -0.00000807, -0.00000928, -0.00000924, -0.00000628, -0.00000127, 0.00000294, 0.00000524, 0.00000620, 0.00000461, -0.00000138, -0.00000878, -0.00001111, -0.00000733, -0.00000371, -0.00000379, -0.00000293, 0.00000244, 0.00000686, 0.00000404, -0.00000277, -0.00000667, -0.00000786, -0.00000921, -0.00000704, 0.00000133, 0.00000622, -0.00000212, -0.00001342, -0.00000962, 0.00000295, -0.00000086, -0.00002069, -0.00002170, 0.00001158, 0.00004088, 0.00002516, -0.00001459, -0.00002288, 0.00000642, 0.00002134, -0.00000614, -0.00003704, -0.00002816, 0.00000264, 0.00001100, -0.00000314, -0.00000428, 0.00001320, 0.00001727, -0.00000305, -0.00001641, -0.00000071, 0.00002127, 0.00001960, 0.00000161, -0.00000687, -0.00000354, -0.00000469, -0.00001148, -0.00001026, -0.00000069, 0.00000244, -0.00000443, -0.00000838, -0.00000344, 0.00000056, -0.00000257, -0.00000490, -0.00000013, 0.00000521, 0.00000440, 0.00000157, 0.00000196, 0.00000188, -0.00000278, -0.00000709, -0.00000541, -0.00000166, -0.00000217, -0.00000463, -0.00000306, 0.00000104, 0.00000176, -0.00000067, -0.00000138, -0.00000023, -0.00000140, -0.00000483, -0.00000605, -0.00000416, -0.00000261, -0.00000221, -0.00000007, 0.00000387, 0.00000543, 0.00000295, -0.00000016, -0.00000118, -0.00000161, -0.00000299, -0.00000371, -0.00000238, -0.00000073, -0.00000073, -0.00000144, -0.00000135, -0.00000134, -0.00000309, -0.00000604, -0.00000831, -0.00000940, -0.00000991, -0.00000953, -0.00000729, -0.00000345, },
		{ -0.00001164, -0.00000955, -0.00000579, -0.00000200, 0.00000106, 0.00000316, 0.00000356, 0.00000153, -0.00000188, -0.00000398, -0.00000336, -0.00000173, -0.00000139, -0.00000217, -0.00000261, -0.00000317, -0.00000546, -0.00000868, -0.00001009, -0.00000892, -0.00000673, -0.00000365, 0.00000130, 0.00000591, 0.00000583, 0.00000150, -0.00000149, -0.00000137, -0.00000309, -0.00000860, -0.00001117, -0.00000548, 0.00000283, 0.00000526, 0.00000252, 0.00000006, -0.00000276, -0.00000827, -0.00001066, -0.00000426, 0.00000259, -0.00000167, -0.00000940, -0.00000332, 0.00001034, 0.00000484, -0.00002141, -0.00003144, -0.00000357, 0.00002785, 0.00001901, -0.00001419, -0.00001706, 0.00001873, 0.00003967, 0.00001218, -0.00002665, -0.00002823, -0.00000450, -0.00000007, -0.00001723, -0.00001783, 0.00000782, 0.00002330, 0.00000593, -0.00001758, -0.00001476, 0.00000580, 0.00001571, 0.00001107, 0.00000790, 0.00000892, 0.00000329, -0.00000781, -0.00001118, -0.00000556, -0.00000360, -0.00000886, -0.00001046, -0.00000383, 0.00000121, -0.00000179, -0.00000552, -0.00000308, 0.00000086, 0.00000060, -0.00000032, 0.00000239, 0.00000460, 0.00000156, -0.00000276, -0.00000278, -0.00000124, -0.00000336, -0.00000670, -0.00000545, -0.00000087, 0.00000114, -0.00000001, -0.00000015, 0.00000107, 0.00000001, -0.00000339, -0.00000510, -0.00000414, -0.00000360, -0.00000433, -0.00000345, -0.00000031, 0.00000175, 0.00000104, 0.00000017, 0.00000115, 0.00000216, 0.00000117, -0.00000068, -0.00000165, -0.00000231, -0.00000358, -0.00000421, -0.00000275, -0.00000029, 0.00000077, -0.00000029, -0.00000246, -0.00000512, -0.00000823, -0.00001096, },
		{ -0.00000270, -0.00000609, -0.00000905, -0.00001013, -0.00000930, -0.00000748, -0.00000515, -0.00000227, 0.00000037, 0.00000118, -0.00000006, -0.00000130, -0.00000098, -0.00000035, -0.00000126, -0.00000276, -0.00000252, -0.00000101, -0.00000098, -0.00000300, -0.00000511, -0.00000682, -0.00000934, -0.00001142, -0.00000976, -0.00000481, -0.00000095, 0.00000075, 0.00000406, 0.00000896, 0.00000862, 0.00000005, -0.00000885, -0.00000970, -0.00000535, -0.00000320, -0.00000286, -0.00000014, 0.00000202, -0.00000131, -0.00000523, -0.00000245, 0.00000089, -0.00000586, -0.00001464, -0.00000672, 0.00001247, 0.00001317, -0.00001147, -0.00002621, -0.00000536, 0.00002118, 0.00000947, -0.00002647, -0.00002985, 0.00001200, 0.00004465, 0.00002713, -0.00000889, -0.00001261, 0.00000816, 0.00000719, -0.00002133, -0.00003480, -0.00001241, 0.00001348, 0.00001007, -0.00000944, -0.00001431, -0.00000308, 0.00000460, 0.00000423, 0.00000677, 0.00001264, 0.00001121, 0.00000276, -0.00000125, 0.00000067, -0.00000198, -0.00001110, -0.00001534, -0.00000951, -0.00000294, -0.00000310, -0.00000482, -0.00000228, 0.00000049, -0.00000127, -0.00000356, -0.00000146, 0.00000169, 0.00000103, -0.00000079, 0.00000063, 0.00000246, -0.00000026, -0.00000510, -0.00000587, -0.00000277, -0.00000097, -0.00000144, -0.00000081, 0.00000109, 0.00000078, -0.00000206, -0.00000374, -0.00000304, -0.00000271, -0.00000379, -0.00000388, -0.00000214, -0.00000122, -0.00000242, -0.00000321, -0.00000140, 0.00000137, 0.00000267, 0.00000256, 0.00000205, 0.00000066, -0.00000212, -0.00000469, -0.00000490, -0.00000291, -0.00000074, 0.00000033, 0.00000042, -0.00000044, },
		{ -0.00000101, -0.00000055, -0.00000070, -0.00000203, -0.00000399, -0.00000542, -0.00000612, -0.00000679, -0.00000751, -0.00000731, -0.00000589, -0.00000455, -0.00000400, -0.00000285, -0.00000006, 0.00000233, 0.00000170, -0.00000094, -0.00000222, -0.00000147, -0.00000096, -0.00000129, -0.00000095, -0.00000054, -0.00000293, -0.00000784, -0.00001122, -0.00001173, -0.00001200, -0.00001170, -0.00000592, 0.00000530, 0.00001297, 0.00001067, 0.00000385, 0.00000062, -0.00000056, -0.00000518, -0.00001036, -0.00000953, -0.00000498, -0.00000337, -0.00000227, 0.00000409, 0.00000852, -0.00000080, -0.00001541, -0.00001294, 0.00000549, 0.00001042, -0.00000904, -0.00002092, 0.00000117, 0.00002853, 0.00001486, -0.00002858, -0.00004186, -0.00000464, 0.00003193, 0.00002205, -0.00000800, -0.00000607, 0.00002344, 0.00003038, -0.00000056, -0.00002906, -0.00002368, -0.00000404, -0.00000162, -0.00001251, -0.00001383, -0.00000381, 0.00000184, 0.00000005, 0.00000130, 0.00000679, 0.00000751, 0.00000348, 0.00000413, 0.00000927, 0.00000776, -0.00000285, -0.00001114, -0.00001026, -0.00000716, -0.00000826, -0.00000913, -0.00000493, -0.00000028, -0.00000070, -0.00000313, -0.00000256, -0.00000103, -0.00000229, -0.00000360, -0.00000088, 0.00000291, 0.00000223, -0.00000154, -0.00000279, -0.00000109, -0.00000064, -0.00000211, -0.00000217, -0.00000039, -0.00000011, -0.00000209, -0.00000324, -0.00000234, -0.00000175, -0.00000258, -0.00000281, -0.00000165, -0.00000141, -0.00000342, -0.00000531, -0.00000462, -0.00000216, -0.00000003, 0.00000143, 0.00000266, 0.00000284, 0.00000104, -0.00000166, -0.00000316, -0.00000288, -0.00000199, -0.00000143, },
		{ -0.00000008, -0.00000051, -0.00000139, -0.00000184, -0.00000189, -0.00000218, -0.00000266, -0.00000252, -0.00000186, -0.00000201, -0.00000368, -0.00000574, -0.00000712, -0.00000838, -0.00000976, -0.00000927, -0.00000535, -0.00000032, 0.00000197, 0.00000137, 0.00000083, 0.00000108, 0.00000019, -0.00000160, -0.00000145, 0.00000065, 0.00000107, -0.00000144, -0.00000410, -0.00000631, -0.00001107, -0.00001729, -0.00001759, -0.00000898, 0.00000093, 0.00000475, 0.00000577, 0.00000978, 0.00001289, 0.00000774, -0.00000272, -0.00000942, -0.00001147, -0.00001337, -0.00001166, -0.00000049, 0.00001114, 0.00000734, -0.00000688, -0.00000854, 0.00000503, 0.00000691, -0.00001433, -0.00002692, -0.00000212, 0.00003237, 0.00002570, -0.00001750, -0.00003794, -0.00000883, 0.00002379, 0.00001207, -0.00002135, -0.00002115, 0.00001551, 0.00003855, 0.00002208, -0.00000428, -0.00000874, -0.00000168, -0.00000720, -0.00002112, -0.00002302, -0.00001177, -0.00000302, -0.00000160, 0.00000067, 0.00000480, 0.00000350, -0.00000160, -0.00000030, 0.00000764, 0.00001066, 0.00000444, -0.00000183, -0.00000192, -0.00000175, -0.00000660, -0.00001119, -0.00000962, -0.00000541, -0.00000426, -0.00000457, -0.00000265, -0.00000082, -0.00000263, -0.00000516, -0.00000372, -0.00000023, 0.00000024, -0.00000166, -0.00000134, 0.00000130, 0.00000211, 0.00000024, -0.00000100, -0.00000054, -0.00000094, -0.00000295, -0.00000394, -0.00000280, -0.00000174, -0.00000202, -0.00000202, -0.00000087, -0.00000057, -0.00000254, -0.00000491, -0.00000531, -0.00000402, -0.00000265, -0.00000149, -0.00000003, 0.00000114, 0.00000096, -0.00000021, -0.00000088, -0.00000050, },
		{ -0.00000168, -0.00000021, 0.00000134, 0.00000157, 0.00000048, -0.00000075, -0.00000159, -0.00000248, -0.00000331, -0.00000299, -0.00000127, 0.00000018, -0.00000008, -0.00000144, -0.00000298, -0.00000551, -0.00000944, -0.00001222, -0.00001098, -0.00000703, -0.00000408, -0.00000262, -0.00000021, 0.00000296, 0.00000369, 0.00000157, 0.00000013, 0.00000070, 0.00000062, -0.00000091, -0.00000101, 0.00000012, -0.00000260, -0.00001013, -0.00001532, -0.00001342, -0.00000974, -0.00000976, -0.00000860, 0.00000038, 0.00001208, 0.00001585, 0.00001112, 0.00000512, -0.00000117, -0.00001085, -0.00001748, -0.00001118, 0.00000136, 0.00000279, -0.00000562, -0.00000317, 0.00001273, 0.00001457, -0.00001080, -0.00003082, -0.00001149, 0.00002455, 0.00002498, -0.00001218, -0.00003133, -0.00000292, 0.00003017, 0.00001742, -0.00002346, -0.00003538, -0.00000618, 0.00002120, 0.00001820, 0.00000548, 0.00000867, 0.00001711, 0.00000841, -0.00001233, -0.00002311, -0.00001932, -0.00001363, -0.00001086, -0.00000513, 0.00000238, 0.00000305, -0.00000225, -0.00000294, 0.00000302, 0.00000598, 0.00000199, -0.00000077, 0.00000282, 0.00000598, 0.00000226, -0.00000400, -0.00000633, -0.00000628, -0.00000794, -0.00000930, -0.00000691, -0.00000347, -0.00000335, -0.00000486, -0.00000383, -0.00000148, -0.00000190, -0.00000411, -0.00000353, 0.00000010, 0.00000253, 0.00000214, 0.00000143, 0.00000153, 0.00000046, -0.00000223, -0.00000394, -0.00000335, -0.00000231, -0.00000227, -0.00000205, -0.00000085, -0.00000031, -0.00000182, -0.00000396, -0.00000458, -0.00000375, -0.00000293, -0.00000253, -0.00000201, -0.00000143, -0.00000147, -0.00000194, },
		{ -0.00000276, -0.00000372, -0.00000381, -0.00000222, 0.00000010, 0.00000137, 0.00000109, 0.00000034, -0.00000016, -0.00000100, -0.00000232, -0.00000276, -0.00000152, 0.00000001, 0.00000024, -0.00000019, -0.00000029, -0.00000137, -0.00000483, -0.00000868, -0.00000976, -0.00000874, -0.00000875, -0.00000946, -0.00000734, -0.00000225, 0.00000170, 0.00000269, 0.00000327, 0.00000436, 0.00000319, -0.00000035, -0.00000175, 0.00000025, 0.00000040, -0.00000387, -0.00000707, -0.00000550, -0.00000515, -0.00001155, -0.00001810, -0.00001483, -0.00000406, 0.00000409, 0.00000755, 0.00001098, 0.00001211, 0.00000475, -0.00000562, -0.00000679, -0.00000166, -0.00000528, -0.00001571, -0.00001144, 0.00001061, 0.00002043, -0.00000051, -0.00002299, -0.00001032, 0.00002065, 0.00001994, -0.00001688, -0.00003602, -0.00000558, 0.00003395, 0.00002879, -0.00001037, -0.00002955, -0.00001217, 0.00000607, 0.00000030, -0.00000923, 0.00000165, 0.00002067, 0.00002218, 0.00000686, -0.00000554, -0.00000871, -0.00001162, -0.00001609, -0.00001452, -0.00000735, -0.00000388, -0.00000543, -0.00000371, 0.00000262, 0.00000455, -0.00000104, -0.00000507, -0.00000110, 0.00000490, 0.00000553, 0.00000294, 0.00000203, 0.00000113, -0.00000321, -0.00000826, -0.00000941, -0.00000795, -0.00000777, -0.00000797, -0.00000549, -0.00000209, -0.00000202, -0.00000455, -0.00000496, -0.00000215, 0.00000040, 0.00000093, 0.00000130, 0.00000229, 0.00000196, -0.00000036, -0.00000233, -0.00000237, -0.00000182, -0.00000197, -0.00000197, -0.00000108, -0.00000060, -0.00000185, -0.00000375, -0.00000433, -0.00000339, -0.00000234, -0.00000196, -0.00000195, -0.00000210, },
		},
		{ { -0.00000026, -0.00000022, -0.00000014, -0.00000008, -0.00000013, -0.00000034, -0.00000069, -0.00000114, -0.00000160, -0.00000204, -0.00000242, -0.00000270, -0.00000286, -0.00000285, -0.00000265, -0.00000226, -0.00000178, -0.00000135, -0.00000107, -0.00000094, -0.00000083, -0.00000045, 0.00000046, 0.00000195, 0.00000382, 0.00000551, 0.00000635, 0.00000579, 0.00000369, 0.00000046, -0.00000303, -0.00000580, -0.00000711, -0.00000681, -0.00000538, -0.00000369, -0.00000257, -0.00000238, -0.00000274, -0.00000279, -0.00000153, 0.00000150, 0.00000585, 0.00001018, 0.00001269, 0.00001189, 0.00000731, -0.00000008, -0.00000813, -0.00001425, -0.00001637, -0.00001383, -0.00000767, -0.00000023, 0.00000574, 0.00000817, 0.00000652, 0.00000188, -0.00000349, -0.00000718, -0.00000770, -0.00000504, -0.00000065, 0.00000329, 0.00000485, 0.00000329, -0.00000065, -0.00000504, -0.00000770, -0.00000718, -0.00000349, 0.00000188, 0.00000652, 0.00000817, 0.00000574, -0.00000023, -0.00000767, -0.00001383, -0.00001637, -0.00001425, -0.00000813, -0.00000008, 0.00000731, 0.00001189, 0.00001269, 0.00001018, 0.00000585, 0.00000150, -0.00000153, -0.00000279, -0.00000274, -0.00000238, -0.00000257, -0.00000369, -0.00000538, -0.00000681, -0.00000711, -0.00000580, -0.00000303, 0.00000046, 0.00000369, 0.00000579, 0.00000635, 0.00000551, 0.00000382, 0.00000195, 0.00000046, -0.00000045, -0.00000083, -0.00000094, -0.00000107, -0.00000135, -0.00000178, -0.00000226, -0.00000265, -0.00000285, -0.00000286, -0.00000270, -0.00000242, -0.00000204, -0.00000160, -0.00000114, -0.00000069, -0.00000034, -0.00000013, -0.00000008, -0.00000014, -0.00000022, },
		{ 0.00000028, 0.00000040, 0.00000032, 0.00000014, -0.00000004, -0.00000014, -0.00000018, -0.00000019, -0.00000022, -0.00000030, -0.00000042, -0.00000057, -0.00000076, -0.00000103, -0.00000140, -0.00000188, -0.00000237, -0.00000273, -0.00000279, -0.00000248, -0.00000185, -0.00000111, -0.00000049, -0.00000016, -0.00000006, 0.00000007, 0.00000060, 0.00000178, 0.00000357, 0.00000552, 0.00000692, 0.00000705, 0.00000553, 0.00000255, -0.00000113, -0.00000446, -0.00000649, -0.00000680, -0.00000566, -0.00000386, -0.00000236, -0.00000173, -0.00000187, -0.00000204, -0.00000120, 0.00000135, 0.00000549, 0.00001011, 0.00001344, 0.00001377, 0.00001023, 0.00000331, -0.00000510, -0.00001238, -0.00001613, -0.00001509, -0.00000973, -0.00000207, 0.00000508, 0.00000921, 0.00000913, 0.00000532, -0.00000028, -0.00000518, -0.00000741, -0.00000632, -0.00000280, 0.00000120, 0.00000363, 0.00000324, 0.00000023, -0.00000390, -0.00000706, -0.00000752, -0.00000477, 0.00000017, 0.00000519, 0.00000796, 0.00000696, 0.00000221, -0.00000472, -0.00001128, -0.00001498, -0.00001434, -0.00000950, -0.00000209, 0.00000545, 0.00001083, 0.00001273, 0.00001118, 0.00000738, 0.00000306, -0.00000028, -0.00000194, -0.00000212, -0.00000166, -0.00000150, -0.00000221, -0.00000367, -0.00000521, -0.00000597, -0.00000532, -0.00000318, -0.00000004, 0.00000323, 0.00000574, 0.00000691, 0.00000665, 0.00000530, 0.00000344, 0.00000167, 0.00000035, -0.00000040, -0.00000070, -0.00000076, -0.00000079, -0.00000094, -0.00000122, -0.00000162, -0.00000206, -0.00000245, -0.00000271, -0.00000274, -0.00000252, -0.00000204, -0.00000138, -0.00000068, -0.00000009, },
		{ -0.00000190, -0.00000109, -0.00000013, 0.00000075, 0.00000137, 0.00000164, 0.00000161, 0.00000140, 0.00000113, 0.00000089, 0.00000071, 0.00000059, 0.00000050, 0.00000047, 0.00000050, 0.00000059, 0.00000068, 0.00000063, 0.00000033, -0.00000028, -0.00000108, -0.00000186, -0.00000233, -0.00000227, -0.00000167, -0.00000075, 0.00000014, 0.00000071, 0.00000084, 0.00000076, 0.00000088, 0.00000163, 0.00000315, 0.00000515, 0.00000696, 0.00000774, 0.00000687, 0.00000428, 0.00000055, -0.00000327, -0.00000608, -0.00000718, -0.00000655, -0.00000485, -0.00000306, -0.00000195, -0.00000174, -0.00000188, -0.00000141, 0.00000056, 0.00000423, 0.00000884, 0.00001278, 0.00001423, 0.00001193, 0.00000585, -0.00000260, -0.00001096, -0.00001661, -0.00001775, -0.00001413, -0.00000715, 0.00000065, 0.00000660, 0.00000893, 0.00000741, 0.00000336, -0.00000104, -0.00000371, -0.00000356, -0.00000094, 0.00000262, 0.00000517, 0.00000528, 0.00000275, -0.00000130, -0.00000495, -0.00000634, -0.00000461, -0.00000032, 0.00000473, 0.00000825, 0.00000846, 0.00000489, -0.00000137, -0.00000806, -0.00001268, -0.00001345, -0.00001003, -0.00000357, 0.00000373, 0.00000956, 0.00001233, 0.00001171, 0.00000855, 0.00000442, 0.00000086, -0.00000119, -0.00000170, -0.00000131, -0.00000095, -0.00000130, -0.00000247, -0.00000399, -0.00000506, -0.00000497, -0.00000344, -0.00000075, 0.00000243, 0.00000522, 0.00000694, 0.00000730, 0.00000642, 0.00000474, 0.00000283, 0.00000118, 0.00000006, -0.00000045, -0.00000050, -0.00000029, -0.00000008, -0.00000005, -0.00000029, -0.00000079, -0.00000144, -0.00000205, -0.00000241, -0.00000237, },
		{ -0.00000112, -0.00000212, -0.00000291, -0.00000321, -0.00000291, -0.00000208, -0.00000097, 0.00000014, 0.00000097, 0.00000139, 0.00000143, 0.00000120, 0.00000085, 0.00000047, 0.00000014, -0.00000014, -0.00000033, -0.00000039, -0.00000027, 0.00000003, 0.00000040, 0.00000064, 0.00000054, -0.00000004, -0.00000103, -0.00000215, -0.00000301, -0.00000325, -0.00000276, -0.00000170, -0.00000050, 0.00000041, 0.00000077, 0.00000067, 0.00000053, 0.00000089, 0.00000210, 0.00000407, 0.00000622, 0.00000767, 0.00000761, 0.00000569, 0.00000222, -0.00000184, -0.00000531, -0.00000723, -0.00000725, -0.00000579, -0.00000374, -0.00000204, -0.00000120, -0.00000100, -0.00000064, 0.00000083, 0.00000393, 0.00000832, 0.00001277, 0.00001553, 0.00001512, 0.00001097, 0.00000384, -0.00000434, -0.00001112, -0.00001447, -0.00001357, -0.00000912, -0.00000306, 0.00000220, 0.00000478, 0.00000404, 0.00000080, -0.00000314, -0.00000581, -0.00000595, -0.00000359, 0.00000003, 0.00000309, 0.00000403, 0.00000232, -0.00000126, -0.00000499, -0.00000699, -0.00000609, -0.00000244, 0.00000253, 0.00000666, 0.00000802, 0.00000575, 0.00000046, -0.00000594, -0.00001106, -0.00001293, -0.00001080, -0.00000539, 0.00000146, 0.00000752, 0.00001105, 0.00001138, 0.00000903, 0.00000533, 0.00000178, -0.00000053, -0.00000133, -0.00000106, -0.00000058, -0.00000061, -0.00000145, -0.00000282, -0.00000404, -0.00000440, -0.00000345, -0.00000126, 0.00000167, 0.00000458, 0.00000674, 0.00000766, 0.00000726, 0.00000583, 0.00000385, 0.00000186, 0.00000031, -0.00000056, -0.00000076, -0.00000046, 0.00000001, 0.00000035, 0.00000029, -0.00000023, },
		{ -0.00000038, 0.00000034, 0.00000062, 0.00000025, -0.00000071, -0.00000197, -0.00000312, -0.00000374, -0.00000364, -0.00000282, -0.00000155, -0.00000020, 0.00000091, 0.00000158, 0.00000178, 0.00000163, 0.00000127, 0.00000083, 0.00000038, -0.00000003, -0.00000036, -0.00000050, -0.00000036, 0.00000007, 0.00000068, 0.00000122, 0.00000137, 0.00000091, -0.00000017, -0.00000160, -0.00000289, -0.00000357, -0.00000337, -0.00000238, -0.00000101, 0.00000019, 0.00000079, 0.00000072, 0.00000035, 0.00000025, 0.00000096, 0.00000262, 0.00000482, 0.00000673, 0.00000739, 0.00000618, 0.00000312, -0.00000107, -0.00000522, -0.00000817, -0.00000924, -0.00000847, -0.00000655, -0.00000448, -0.00000304, -0.00000243, -0.00000222, -0.00000162, 0.00000006, 0.00000291, 0.00000625, 0.00000881, 0.00000931, 0.00000702, 0.00000226, -0.00000361, -0.00000862, -0.00001100, -0.00000983, -0.00000555, 0.00000027, 0.00000547, 0.00000822, 0.00000771, 0.00000451, 0.00000023, -0.00000319, -0.00000431, -0.00000284, 0.00000027, 0.00000335, 0.00000476, 0.00000367, 0.00000051, -0.00000328, -0.00000585, -0.00000585, -0.00000304, 0.00000150, 0.00000584, 0.00000801, 0.00000684, 0.00000254, -0.00000342, -0.00000881, -0.00001163, -0.00001080, -0.00000661, -0.00000054, 0.00000537, 0.00000933, 0.00001042, 0.00000884, 0.00000563, 0.00000221, -0.00000030, -0.00000142, -0.00000140, -0.00000093, -0.00000077, -0.00000133, -0.00000250, -0.00000378, -0.00000447, -0.00000404, -0.00000238, 0.00000023, 0.00000314, 0.00000562, 0.00000708, 0.00000724, 0.00000617, 0.00000426, 0.00000207, 0.00000014, -0.00000109, -0.00000148, -0.00000113, },
		{ 0.00000112, -0.00000057, -0.00000130, -0.00000105, -0.00000015, 0.00000087, 0.00000148, 0.00000131, 0.00000032, -0.00000119, -0.00000274, -0.00000379, -0.00000401, -0.00000336, -0.00000208, -0.00000057, 0.00000076, 0.00000164, 0.00000199, 0.00000192, 0.00000156, 0.00000106, 0.00000051, -0.00000004, -0.00000051, -0.00000077, -0.00000069, -0.00000020, 0.00000061, 0.00000145, 0.00000194, 0.00000174, 0.00000076, -0.00000079, -0.00000239, -0.00000347, -0.00000361, -0.00000274, -0.00000122, 0.00000034, 0.00000136, 0.00000162, 0.00000129, 0.00000097, 0.00000129, 0.00000263, 0.00000482, 0.00000718, 0.00000871, 0.00000857, 0.00000642, 0.00000267, -0.00000169, -0.00000540, -0.00000745, -0.00000742, -0.00000556, -0.00000266, 0.00000040, 0.00000299, 0.00000498, 0.00000660, 0.00000819, 0.00000982, 0.00001108, 0.00001118, 0.00000928, 0.00000500, -0.00000121, -0.00000804, -0.00001366, -0.00001638, -0.00001531, -0.00001079, -0.00000432, 0.00000195, 0.00000604, 0.00000687, 0.00000468, 0.00000086, -0.00000271, -0.00000441, -0.00000363, -0.00000095, 0.00000215, 0.00000405, 0.00000371, 0.00000121, -0.00000230, -0.00000512, -0.00000576, -0.00000370, 0.00000036, 0.00000476, 0.00000759, 0.00000750, 0.00000431, -0.00000090, -0.00000619, -0.00000957, -0.00000978, -0.00000675, -0.00000156, 0.00000401, 0.00000821, 0.00000996, 0.00000917, 0.00000662, 0.00000352, 0.00000100, -0.00000034, -0.00000053, -0.00000013, 0.00000017, -0.00000012, -0.00000106, -0.00000227, -0.00000317, -0.00000319, -0.00000203, 0.00000017, 0.00000296, 0.00000564, 0.00000756, 0.00000824, 0.00000758, 0.00000580, 0.00000344, },
		{ 0.00000814, 0.00000798, 0.00000650, 0.00000412, 0.00000152, -0.00000057, -0.00000164, -0.00000154, -0.00000050, 0.00000087, 0.00000192, 0.00000211, 0.00000130, -0.00000029, -0.00000210, -0.00000351, -0.00000405, -0.00000357, -0.00000228, -0.00000060, 0.00000099, 0.00000215, 0.00000274, 0.00000281, 0.00000254, 0.00000206, 0.00000146, 0.00000083, 0.00000024, -0.00000016, -0.00000020, 0.00000024, 0.00000112, 0.00000216, 0.00000294, 0.00000303, 0.00000220, 0.00000060, -0.00000131, -0.00000289, -0.00000356, -0.00000311, -0.00000177, -0.00000011, 0.00000120, 0.00000172, 0.00000145, 0.00000083, 0.00000053, 0.00000113, 0.00000275, 0.00000495, 0.00000687, 0.00000752, 0.00000623, 0.00000293, -0.00000180, -0.00000691, -0.00001123, -0.00001388, -0.00001448, -0.00001313, -0.00001026, -0.00000633, -0.00000171, 0.00000339, 0.00000871, 0.00001377, 0.00001778, 0.00001972, 0.00001874, 0.00001451, 0.00000759, -0.00000057, -0.00000795, -0.00001261, -0.00001336, -0.00001022, -0.00000448, 0.00000180, 0.00000650, 0.00000824, 0.00000688, 0.00000345, -0.00000028, -0.00000264, -0.00000274, -0.00000081, 0.00000195, 0.00000400, 0.00000415, 0.00000221, -0.00000100, -0.00000396, -0.00000518, -0.00000393, -0.00000055, 0.00000360, 0.00000675, 0.00000746, 0.00000527, 0.00000088, -0.00000411, -0.00000783, -0.00000894, -0.00000705, -0.00000289, 0.00000206, 0.00000619, 0.00000835, 0.00000823, 0.00000631, 0.00000362, 0.00000119, -0.00000029, -0.00000071, -0.00000045, -0.00000011, -0.00000019, -0.00000088, -0.00000198, -0.00000298, -0.00000333, -0.00000264, -0.00000084, 0.00000175, 0.00000457, 0.00000689, },
		{ -0.00000119, 0.00000112, 0.00000393, 0.00000655, 0.00000828, 0.00000863, 0.00000746, 0.00000511, 0.00000223, -0.00000034, -0.00000189, -0.00000211, -0.00000114, 0.00000044, 0.00000184, 0.00000238, 0.00000175, 0.00000011, -0.00000200, -0.00000385, -0.00000484, -0.00000470, -0.00000357, -0.00000187, -0.00000012, 0.00000123, 0.00000199, 0.00000219, 0.00000195, 0.00000146, 0.00000082, 0.00000011, -0.00000059, -0.00000112, -0.00000130, -0.00000095, -0.00000008, 0.00000110, 0.00000216, 0.00000263, 0.00000218, 0.00000079, -0.00000115, -0.00000300, -0.00000411, -0.00000407, -0.00000292, -0.00000111, 0.00000069, 0.00000188, 0.00000221, 0.00000189, 0.00000147, 0.00000163, 0.00000278, 0.00000490, 0.00000750, 0.00000977, 0.00001089, 0.00001034, 0.00000803, 0.00000433, -0.00000012, -0.00000463, -0.00000860, -0.00001161, -0.00001338, -0.00001366, -0.00001223, -0.00000899, -0.00000412, 0.00000175, 0.00000752, 0.00001179, 0.00001327, 0.00001123, 0.00000591, -0.00000145, -0.00000887, -0.00001424, -0.00001605, -0.00001393, -0.00000878, -0.00000245, 0.00000294, 0.00000579, 0.00000558, 0.00000301, -0.00000042, -0.00000302, -0.00000371, -0.00000240, 0.00000002, 0.00000216, 0.00000282, 0.00000154, -0.00000114, -0.00000393, -0.00000544, -0.00000478, -0.00000201, 0.00000181, 0.00000513, 0.00000653, 0.00000532, 0.00000189, -0.00000250, -0.00000621, -0.00000786, -0.00000691, -0.00000374, 0.00000050, 0.00000438, 0.00000679, 0.00000725, 0.00000602, 0.00000386, 0.00000170, 0.00000021, -0.00000036, -0.00000023, 0.00000012, 0.00000020, -0.00000026, -0.00000118, -0.00000220, -0.00000281, -0.00000254, },
		{ -0.00000150, -0.00000252, -0.00000334, -0.00000347, -0.00000258, -0.00000063, 0.00000206, 0.00000488, 0.00000706, 0.00000793, 0.00000719, 0.00000498, 0.00000194, -0.00000103, -0.00000308, -0.00000368, -0.00000283, -0.00000109, 0.00000070, 0.00000170, 0.00000143, -0.00000009, -0.00000233, -0.00000452, -0.00000592, -0.00000613, -0.00000518, -0.00000346, -0.00000154, 0.00000006, 0.00000106, 0.00000142, 0.00000127, 0.00000079, 0.00000014, -0.00000060, -0.00000135, -0.00000198, -0.00000230, -0.00000215, -0.00000144, -0.00000033, 0.00000083, 0.00000154, 0.00000141, 0.00000027, -0.00000165, -0.00000382, -0.00000557, -0.00000635, -0.00000595, -0.00000460, -0.00000282, -0.00000128, -0.00000050, -0.00000064, -0.00000155, -0.00000277, -0.00000382, -0.00000430, -0.00000409, -0.00000328, -0.00000211, -0.00000085, 0.00000028, 0.00000114, 0.00000163, 0.00000165, 0.00000110, 0.00000000, -0.00000144, -0.00000275, -0.00000331, -0.00000246, 0.00000015, 0.00000431, 0.00000916, 0.00001336, 0.00001544, 0.00001436, 0.00000994, 0.00000308, -0.00000447, -0.00001065, -0.00001373, -0.00001300, -0.00000895, -0.00000315, 0.00000238, 0.00000592, 0.00000663, 0.00000484, 0.00000178, -0.00000099, -0.00000226, -0.00000170, 0.00000013, 0.00000205, 0.00000290, 0.00000209, -0.00000011, -0.00000269, -0.00000439, -0.00000427, -0.00000220, 0.00000108, 0.00000426, 0.00000600, 0.00000553, 0.00000296, -0.00000078, -0.00000431, -0.00000635, -0.00000624, -0.00000410, -0.00000077, 0.00000258, 0.00000492, 0.00000571, 0.00000501, 0.00000335, 0.00000147, 0.00000002, -0.00000071, -0.00000078, -0.00000054, -0.00000042, -0.00000072, },
		{ -0.00000061, -0.00000042, -0.00000023, -0.00000034, -0.00000090, -0.00000179, -0.00000266, -0.00000307, -0.00000256, -0.00000096, 0.00000160, 0.00000457, 0.00000718, 0.00000863, 0.00000842, 0.00000652, 0.00000345, 0.00000014, -0.00000240, -0.00000345, -0.00000285, -0.00000104, 0.00000110, 0.00000260, 0.00000279, 0.00000153, -0.00000073, -0.00000318, -0.00000500, -0.00000561, -0.00000490, -0.00000322, -0.00000114, 0.00000075, 0.00000205, 0.00000264, 0.00000262, 0.00000221, 0.00000161, 0.00000094, 0.00000030, -0.00000022, -0.00000050, -0.00000036, 0.00000027, 0.00000135, 0.00000263, 0.00000371, 0.00000419, 0.00000379, 0.00000253, 0.00000071, -0.00000111, -0.00000234, -0.00000253, -0.00000147, 0.00000071, 0.00000363, 0.00000676, 0.00000956, 0.00001154, 0.00001237, 0.00001183, 0.00000990, 0.00000672, 0.00000267, -0.00000173, -0.00000582, -0.00000898, -0.00001081, -0.00001122, -0.00001050, -0.00000921, -0.00000794, -0.00000710, -0.00000667, -0.00000623, -0.00000512, -0.00000275, 0.00000102, 0.00000570, 0.00001017, 0.00001305, 0.00001316, 0.00001004, 0.00000421, -0.00000287, -0.00000925, -0.00001313, -0.00001351, -0.00001052, -0.00000537, 0.00000013, 0.00000422, 0.00000584, 0.00000498, 0.00000255, -0.00000008, -0.00000170, -0.00000175, -0.00000052, 0.00000107, 0.00000197, 0.00000154, -0.00000016, -0.00000239, -0.00000405, -0.00000427, -0.00000277, -0.00000005, 0.00000286, 0.00000477, 0.00000490, 0.00000316, 0.00000021, -0.00000287, -0.00000498, -0.00000541, -0.00000412, -0.00000167, 0.00000107, 0.00000321, 0.00000423, 0.00000402, 0.00000293, 0.00000150, 0.00000026, -0.00000045, },
		{ 0.00000342, 0.00000244, 0.00000148, 0.00000082, 0.00000057, 0.00000066, 0.00000083, 0.00000082, 0.00000044, -0.00000033, -0.00000125, -0.00000190, -0.00000180, -0.00000061, 0.00000167, 0.00000465, 0.00000757, 0.00000953, 0.00000986, 0.00000832, 0.00000531, 0.00000171, -0.00000136, -0.00000299, -0.00000281, -0.00000113, 0.00000121, 0.00000314, 0.00000381, 0.00000288, 0.00000067, -0.00000205, -0.00000434, -0.00000550, -0.00000524, -0.00000378, -0.00000170, 0.00000035, 0.00000183, 0.00000252, 0.00000246, 0.00000188, 0.00000106, 0.00000022, -0.00000052, -0.00000111, -0.00000149, -0.00000163, -0.00000146, -0.00000098, -0.00000030, 0.00000041, 0.00000085, 0.00000074, -0.00000014, -0.00000186, -0.00000432, -0.00000723, -0.00001015, -0.00001253, -0.00001378, -0.00001340, -0.00001107, -0.00000681, -0.00000103, 0.00000545, 0.00001154, 0.00001611, 0.00001825, 0.00001758, 0.00001434, 0.00000935, 0.00000379, -0.00000116, -0.00000468, -0.00000646, -0.00000679, -0.00000627, -0.00000555, -0.00000499, -0.00000451, -0.00000367, -0.00000190, 0.00000111, 0.00000511, 0.00000927, 0.00001234, 0.00001313, 0.00001097, 0.00000606, -0.00000046, -0.00000687, -0.00001139, -0.00001283, -0.00001099, -0.00000672, -0.00000158, 0.00000276, 0.00000510, 0.00000516, 0.00000349, 0.00000124, -0.00000046, -0.00000095, -0.00000028, 0.00000089, 0.00000171, 0.00000155, 0.00000034, -0.00000141, -0.00000284, -0.00000317, -0.00000208, 0.00000014, 0.00000268, 0.00000459, 0.00000513, 0.00000414, 0.00000200, -0.00000046, -0.00000237, -0.00000310, -0.00000252, -0.00000091, 0.00000109, 0.00000284, 0.00000387, 0.00000401, },
		{ -0.00000038, 0.00000078, 0.00000160, 0.00000189, 0.00000169, 0.00000114, 0.00000049, -0.00000006, -0.00000036, -0.00000039, -0.00000027, -0.00000023, -0.00000048, -0.00000111, -0.00000201, -0.00000284, -0.00000310, -0.00000235, -0.00000040, 0.00000248, 0.00000560, 0.00000805, 0.00000896, 0.00000790, 0.00000509, 0.00000135, -0.00000218, -0.00000442, -0.00000476, -0.00000331, -0.00000085, 0.00000152, 0.00000278, 0.00000240, 0.00000047, -0.00000230, -0.00000494, -0.00000656, -0.00000671, -0.00000543, -0.00000322, -0.00000081, 0.00000114, 0.00000223, 0.00000240, 0.00000189, 0.00000108, 0.00000033, -0.00000010, -0.00000017, 0.00000006, 0.00000044, 0.00000085, 0.00000124, 0.00000171, 0.00000242, 0.00000355, 0.00000515, 0.00000710, 0.00000896, 0.00001012, 0.00000991, 0.00000782, 0.00000377, -0.00000178, -0.00000780, -0.00001299, -0.00001605, -0.00001616, -0.00001326, -0.00000807, -0.00000199, 0.00000337, 0.00000666, 0.00000718, 0.00000506, 0.00000116, -0.00000328, -0.00000707, -0.00000944, -0.00001024, -0.00000981, -0.00000876, -0.00000761, -0.00000653, -0.00000530, -0.00000346, -0.00000066, 0.00000305, 0.00000708, 0.00001041, 0.00001189, 0.00001076, 0.00000697, 0.00000133, -0.00000470, -0.00000946, -0.00001167, -0.00001087, -0.00000758, -0.00000303, 0.00000126, 0.00000409, 0.00000493, 0.00000406, 0.00000233, 0.00000072, -0.00000011, -0.00000002, 0.00000059, 0.00000108, 0.00000092, -0.00000000, -0.00000136, -0.00000253, -0.00000288, -0.00000214, -0.00000047, 0.00000154, 0.00000318, 0.00000384, 0.00000335, 0.00000192, 0.00000010, -0.00000149, -0.00000235, -0.00000233, -0.00000155, },
		{ -0.00000037, -0.00000108, -0.00000136, -0.00000124, -0.00000084, -0.00000033, 0.00000014, 0.00000046, 0.00000057, 0.00000047, 0.00000020, -0.00000013, -0.00000040, -0.00000050, -0.00000045, -0.00000038, -0.00000050, -0.00000097, -0.00000179, -0.00000270, -0.00000325, -0.00000290, -0.00000134, 0.00000134, 0.00000457, 0.00000741, 0.00000891, 0.00000843, 0.00000597, 0.00000220, -0.00000172, -0.00000459, -0.00000559, -0.00000461, -0.00000224, 0.00000041, 0.00000222, 0.00000241, 0.00000084, -0.00000194, -0.00000499, -0.00000730, -0.00000815, -0.00000738, -0.00000536, -0.00000282, -0.00000060, 0.00000067, 0.00000074, -0.00000022, -0.00000170, -0.00000312, -0.00000397, -0.00000407, -0.00000364, -0.00000317, -0.00000330, -0.00000450, -0.00000683, -0.00000984, -0.00001259, -0.00001399, -0.00001310, -0.00000959, -0.00000389, 0.00000282, 0.00000890, 0.00001285, 0.00001372, 0.00001153, 0.00000722, 0.00000239, -0.00000128, -0.00000260, -0.00000125, 0.00000212, 0.00000618, 0.00000940, 0.00001059, 0.00000926, 0.00000579, 0.00000116, -0.00000342, -0.00000694, -0.00000891, -0.00000939, -0.00000885, -0.00000785, -0.00000673, -0.00000548, -0.00000383, -0.00000146, 0.00000170, 0.00000527, 0.00000842, 0.00001017, 0.00000971, 0.00000683, 0.00000205, -0.00000346, -0.00000824, -0.00001103, -0.00001119, -0.00000895, -0.00000521, -0.00000125, 0.00000180, 0.00000330, 0.00000326, 0.00000223, 0.00000097, 0.00000007, -0.00000025, -0.00000016, -0.00000006, -0.00000031, -0.00000100, -0.00000194, -0.00000268, -0.00000282, -0.00000218, -0.00000088, 0.00000066, 0.00000196, 0.00000262, 0.00000250, 0.00000174, 0.00000066, },
		{ 0.00000177, 0.00000177, 0.00000149, 0.00000108, 0.00000066, 0.00000025, -0.00000014, -0.00000054, -0.00000090, -0.00000113, -0.00000116, -0.00000099, -0.00000069, -0.00000041, -0.00000028, -0.00000032, -0.00000047, -0.00000060, -0.00000062, -0.00000055, -0.00000055, -0.00000084, -0.00000150, -0.00000239, -0.00000310, -0.00000308, -0.00000190, 0.00000053, 0.00000378, 0.00000697, 0.00000907, 0.00000930, 0.00000741, 0.00000389, -0.00000018, -0.00000355, -0.00000519, -0.00000473, -0.00000254, 0.00000041, 0.00000291, 0.00000396, 0.00000317, 0.00000080, -0.00000230, -0.00000503, -0.00000644, -0.00000602, -0.00000391, -0.00000077, 0.00000241, 0.00000467, 0.00000542, 0.00000465, 0.00000296, 0.00000134, 0.00000089, 0.00000232, 0.00000567, 0.00001017, 0.00001438, 0.00001666, 0.00001574, 0.00001115, 0.00000357, -0.00000537, -0.00001353, -0.00001895, -0.00002048, -0.00001818, -0.00001322, -0.00000749, -0.00000287, -0.00000061, -0.00000090, -0.00000295, -0.00000531, -0.00000651, -0.00000568, -0.00000279, 0.00000129, 0.00000523, 0.00000769, 0.00000785, 0.00000566, 0.00000182, -0.00000258, -0.00000642, -0.00000893, -0.00000991, -0.00000962, -0.00000855, -0.00000713, -0.00000555, -0.00000371, -0.00000139, 0.00000152, 0.00000477, 0.00000778, 0.00000973, 0.00000987, 0.00000789, 0.00000409, -0.00000065, -0.00000512, -0.00000817, -0.00000910, -0.00000787, -0.00000510, -0.00000175, 0.00000119, 0.00000308, 0.00000374, 0.00000344, 0.00000265, 0.00000182, 0.00000117, 0.00000069, 0.00000022, -0.00000038, -0.00000111, -0.00000179, -0.00000216, -0.00000205, -0.00000142, -0.00000045, 0.00000057, 0.00000137, },
		{ -0.00000086, 0.00000007, 0.00000078, 0.00000115, 0.00000126, 0.00000129, 0.00000143, 0.00000172, 0.00000205, 0.00000221, 0.00000198, 0.00000132, 0.00000037, -0.00000057, -0.00000117, -0.00000127, -0.00000090, -0.00000030, 0.00000026, 0.00000056, 0.00000058, 0.00000046, 0.00000036, 0.00000036, 0.00000038, 0.00000021, -0.00000032, -0.00000117, -0.00000202, -0.00000234, -0.00000162, 0.00000039, 0.00000340, 0.00000667, 0.00000917, 0.00000998, 0.00000864, 0.00000542, 0.00000121, -0.00000275, -0.00000529, -0.00000577, -0.00000426, -0.00000153, 0.00000126, 0.00000294, 0.00000278, 0.00000067, -0.00000277, -0.00000646, -0.00000921, -0.00001010, -0.00000883, -0.00000586, -0.00000230, 0.00000040, 0.00000096, -0.00000131, -0.00000610, -0.00001214, -0.00001746, -0.00002002, -0.00001839, -0.00001228, -0.00000282, 0.00000779, 0.00001690, 0.00002223, 0.00002267, 0.00001857, 0.00001159, 0.00000410, -0.00000167, -0.00000439, -0.00000394, -0.00000137, 0.00000167, 0.00000364, 0.00000369, 0.00000196, -0.00000058, -0.00000262, -0.00000305, -0.00000148, 0.00000164, 0.00000521, 0.00000790, 0.00000868, 0.00000716, 0.00000371, -0.00000076, -0.00000514, -0.00000850, -0.00001037, -0.00001079, -0.00001011, -0.00000879, -0.00000716, -0.00000528, -0.00000310, -0.00000053, 0.00000228, 0.00000493, 0.00000679, 0.00000726, 0.00000599, 0.00000311, -0.00000074, -0.00000463, -0.00000760, -0.00000900, -0.00000863, -0.00000681, -0.00000422, -0.00000156, 0.00000057, 0.00000192, 0.00000250, 0.00000249, 0.00000208, 0.00000143, 0.00000060, -0.00000035, -0.00000130, -0.00000209, -0.00000249, -0.00000237, -0.00000177, },
		{ -0.00000190, -0.00000234, -0.00000195, -0.00000096, 0.00000022, 0.00000113, 0.00000147, 0.00000127, 0.00000083, 0.00000056, 0.00000077, 0.00000151, 0.00000248, 0.00000319, 0.00000322, 0.00000238, 0.00000087, -0.00000083, -0.00000217, -0.00000276, -0.00000255, -0.00000178, -0.00000090, -0.00000026, -0.00000002, -0.00000009, -0.00000027, -0.00000037, -0.00000038, -0.00000047, -0.00000084, -0.00000155, -0.00000238, -0.00000286, -0.00000246, -0.00000083, 0.00000193, 0.00000523, 0.00000812, 0.00000964, 0.00000915, 0.00000666, 0.00000281, -0.00000128, -0.00000440, -0.00000566, -0.00000477, -0.00000215, 0.00000122, 0.00000413, 0.00000547, 0.00000470, 0.00000195, -0.00000188, -0.00000540, -0.00000708, -0.00000581, -0.00000131, 0.00000565, 0.00001328, 0.00001928, 0.00002148, 0.00001861, 0.00001080, -0.00000034, -0.00001211, -0.00002153, -0.00002631, -0.00002557, -0.00002014, -0.00001218, -0.00000444, 0.00000073, 0.00000211, -0.00000001, -0.00000419, -0.00000842, -0.00001097, -0.00001104, -0.00000894, -0.00000581, -0.00000309, -0.00000184, -0.00000234, -0.00000399, -0.00000569, -0.00000629, -0.00000511, -0.00000227, 0.00000142, 0.00000473, 0.00000653, 0.00000617, 0.00000371, -0.00000017, -0.00000443, -0.00000808, -0.00001043, -0.00001127, -0.00001079, -0.00000940, -0.00000749, -0.00000532, -0.00000296, -0.00000044, 0.00000214, 0.00000450, 0.00000624, 0.00000690, 0.00000622, 0.00000425, 0.00000140, -0.00000168, -0.00000428, -0.00000587, -0.00000622, -0.00000541, -0.00000379, -0.00000177, 0.00000025, 0.00000201, 0.00000333, 0.00000408, 0.00000418, 0.00000360, 0.00000242, 0.00000084, -0.00000074, },
		{ 0.00000412, 0.00000279, 0.00000070, -0.00000153, -0.00000318, -0.00000374, -0.00000312, -0.00000165, -0.00000002, 0.00000110, 0.00000130, 0.00000066, -0.00000033, -0.00000099, -0.00000082, 0.00000028, 0.00000190, 0.00000331, 0.00000382, 0.00000307, 0.00000122, -0.00000114, -0.00000321, -0.00000440, -0.00000448, -0.00000368, -0.00000252, -0.00000152, -0.00000101, -0.00000098, -0.00000123, -0.00000150, -0.00000168, -0.00000185, -0.00000218, -0.00000281, -0.00000362, -0.00000425, -0.00000421, -0.00000312, -0.00000091, 0.00000200, 0.00000484, 0.00000668, 0.00000680, 0.00000496, 0.00000146, -0.00000285, -0.00000686, -0.00000955, -0.00001023, -0.00000882, -0.00000589, -0.00000254, -0.00000016, -0.00000000, -0.00000282, -0.00000847, -0.00001587, -0.00002306, -0.00002779, -0.00002812, -0.00002313, -0.00001336, -0.00000082, 0.00001157, 0.00002082, 0.00002485, 0.00002314, 0.00001691, 0.00000867, 0.00000128, -0.00000301, -0.00000332, -0.00000031, 0.00000418, 0.00000797, 0.00000943, 0.00000801, 0.00000440, 0.00000008, -0.00000331, -0.00000469, -0.00000392, -0.00000175, 0.00000058, 0.00000193, 0.00000174, 0.00000027, -0.00000160, -0.00000279, -0.00000246, -0.00000046, 0.00000269, 0.00000592, 0.00000808, 0.00000835, 0.00000650, 0.00000295, -0.00000140, -0.00000554, -0.00000864, -0.00001026, -0.00001039, -0.00000933, -0.00000750, -0.00000527, -0.00000290, -0.00000054, 0.00000166, 0.00000353, 0.00000482, 0.00000530, 0.00000485, 0.00000353, 0.00000158, -0.00000062, -0.00000265, -0.00000420, -0.00000509, -0.00000525, -0.00000473, -0.00000361, -0.00000199, -0.00000005, 0.00000192, 0.00000353, 0.00000436, },
		{ -0.00000154, 0.00000097, 0.00000348, 0.00000521, 0.00000555, 0.00000429, 0.00000180, -0.00000108, -0.00000334, -0.00000421, -0.00000346, -0.00000153, 0.00000067, 0.00000218, 0.00000244, 0.00000149, -0.00000002, -0.00000116, -0.00000117, 0.00000017, 0.00000239, 0.00000460, 0.00000582, 0.00000545, 0.00000356, 0.00000080, -0.00000188, -0.00000362, -0.00000400, -0.00000317, -0.00000165, -0.00000012, 0.00000094, 0.00000139, 0.00000137, 0.00000117, 0.00000101, 0.00000094, 0.00000083, 0.00000058, 0.00000021, -0.00000005, 0.00000017, 0.00000123, 0.00000324, 0.00000599, 0.00000894, 0.00001139, 0.00001265, 0.00001230, 0.00001030, 0.00000703, 0.00000326, -0.00000000, -0.00000173, -0.00000110, 0.00000222, 0.00000788, 0.00001471, 0.00002094, 0.00002456, 0.00002395, 0.00001844, 0.00000871, -0.00000322, -0.00001457, -0.00002255, -0.00002536, -0.00002274, -0.00001610, -0.00000797, -0.00000111, 0.00000249, 0.00000225, -0.00000089, -0.00000499, -0.00000795, -0.00000837, -0.00000607, -0.00000207, 0.00000195, 0.00000437, 0.00000434, 0.00000206, -0.00000140, -0.00000458, -0.00000627, -0.00000604, -0.00000427, -0.00000197, -0.00000026, 0.00000016, -0.00000074, -0.00000232, -0.00000359, -0.00000366, -0.00000214, 0.00000067, 0.00000394, 0.00000655, 0.00000758, 0.00000656, 0.00000368, -0.00000037, -0.00000463, -0.00000818, -0.00001039, -0.00001103, -0.00001023, -0.00000839, -0.00000596, -0.00000335, -0.00000091, 0.00000114, 0.00000264, 0.00000350, 0.00000371, 0.00000335, 0.00000257, 0.00000153, 0.00000040, -0.00000074, -0.00000187, -0.00000295, -0.00000389, -0.00000447, -0.00000439, -0.00000343, },
		{ -0.00000074, -0.00000232, -0.00000338, -0.00000327, -0.00000165, 0.00000125, 0.00000465, 0.00000745, 0.00000863, 0.00000771, 0.00000501, 0.00000152, -0.00000142, -0.00000273, -0.00000201, 0.00000030, 0.00000309, 0.00000514, 0.00000560, 0.00000443, 0.00000239, 0.00000063, 0.00000019, 0.00000146, 0.00000404, 0.00000687, 0.00000874, 0.00000878, 0.00000689, 0.00000368, 0.00000023, -0.00000237, -0.00000348, -0.00000309, -0.00000168, -0.00000001, 0.00000128, 0.00000187, 0.00000177, 0.00000124, 0.00000057, -0.00000009, -0.00000068, -0.00000123, -0.00000173, -0.00000208, -0.00000211, -0.00000165, -0.00000064, 0.00000084, 0.00000253, 0.00000411, 0.00000516, 0.00000527, 0.00000404, 0.00000114, -0.00000352, -0.00000963, -0.00001633, -0.00002223, -0.00002568, -0.00002525, -0.00002029, -0.00001133, -0.00000012, 0.00001078, 0.00001876, 0.00002207, 0.00002037, 0.00001489, 0.00000796, 0.00000210, -0.00000088, -0.00000053, 0.00000225, 0.00000562, 0.00000771, 0.00000742, 0.00000480, 0.00000099, -0.00000237, -0.00000390, -0.00000305, -0.00000035, 0.00000292, 0.00000526, 0.00000567, 0.00000397, 0.00000090, -0.00000232, -0.00000449, -0.00000498, -0.00000389, -0.00000198, -0.00000029, 0.00000040, -0.00000016, -0.00000154, -0.00000292, -0.00000339, -0.00000243, -0.00000008, 0.00000301, 0.00000584, 0.00000744, 0.00000718, 0.00000499, 0.00000134, -0.00000291, -0.00000681, -0.00000959, -0.00001079, -0.00001036, -0.00000859, -0.00000597, -0.00000306, -0.00000039, 0.00000164, 0.00000285, 0.00000325, 0.00000304, 0.00000256, 0.00000213, 0.00000193, 0.00000196, 0.00000196, 0.00000162, 0.00000071, },
		{ 0.00000133, 0.00000249, 0.00000308, 0.00000252, 0.00000068, -0.00000195, -0.00000439, -0.00000554, -0.00000467, -0.00000179, 0.00000223, 0.00000598, 0.00000802, 0.00000752, 0.00000463, 0.00000046, -0.00000337, -0.00000543, -0.00000504, -0.00000258, 0.00000070, 0.00000328, 0.00000404, 0.00000279, 0.00000027, -0.00000216, -0.00000317, -0.00000213, 0.00000066, 0.00000409, 0.00000674, 0.00000748, 0.00000597, 0.00000273, -0.00000111, -0.00000432, -0.00000600, -0.00000591, -0.00000444, -0.00000231, -0.00000030, 0.00000107, 0.00000163, 0.00000149, 0.00000092, 0.00000022, -0.00000034, -0.00000058, -0.00000035, 0.00000038, 0.00000156, 0.00000300, 0.00000448, 0.00000584, 0.00000716, 0.00000873, 0.00001099, 0.00001418, 0.00001808, 0.00002184, 0.00002406, 0.00002325, 0.00001842, 0.00000968, -0.00000161, -0.00001310, -0.00002216, -0.00002680, -0.00002635, -0.00002171, -0.00001499, -0.00000865, -0.00000460, -0.00000353, -0.00000478, -0.00000682, -0.00000800, -0.00000731, -0.00000483, -0.00000156, 0.00000109, 0.00000198, 0.00000084, -0.00000168, -0.00000431, -0.00000575, -0.00000526, -0.00000300, 0.00000010, 0.00000273, 0.00000384, 0.00000303, 0.00000074, -0.00000205, -0.00000422, -0.00000500, -0.00000429, -0.00000263, -0.00000089, 0.00000009, -0.00000005, -0.00000112, -0.00000242, -0.00000312, -0.00000258, -0.00000068, 0.00000217, 0.00000510, 0.00000715, 0.00000757, 0.00000607, 0.00000291, -0.00000122, -0.00000540, -0.00000874, -0.00001060, -0.00001070, -0.00000917, -0.00000648, -0.00000331, -0.00000039, 0.00000167, 0.00000258, 0.00000238, 0.00000149, 0.00000051, 0.00000002, 0.00000033, },
		{ -0.00000053, -0.00000216, -0.00000290, -0.00000222, -0.00000025, 0.00000226, 0.00000415, 0.00000442, 0.00000263, -0.00000073, -0.00000441, -0.00000685, -0.00000685, -0.00000412, 0.00000051, 0.00000536, 0.00000857, 0.00000888, 0.00000620, 0.00000163, -0.00000295, -0.00000574, -0.00000573, -0.00000316, 0.00000066, 0.00000394, 0.00000525, 0.00000417, 0.00000137, -0.00000164, -0.00000331, -0.00000271, 0.00000004, 0.00000384, 0.00000713, 0.00000854, 0.00000747, 0.00000426, -0.00000002, -0.00000403, -0.00000670, -0.00000752, -0.00000666, -0.00000475, -0.00000257, -0.00000085, -0.00000003, -0.00000029, -0.00000150, -0.00000334, -0.00000528, -0.00000675, -0.00000724, -0.00000660, -0.00000513, -0.00000364, -0.00000321, -0.00000472, -0.00000837, -0.00001340, -0.00001811, -0.00002039, -0.00001855, -0.00001202, -0.00000182, 0.00000969, 0.00001955, 0.00002522, 0.00002550, 0.00002094, 0.00001355, 0.00000598, 0.00000049, -0.00000182, -0.00000127, 0.00000076, 0.00000259, 0.00000305, 0.00000193, -0.00000006, -0.00000174, -0.00000217, -0.00000107, 0.00000098, 0.00000292, 0.00000368, 0.00000280, 0.00000058, -0.00000203, -0.00000387, -0.00000415, -0.00000276, -0.00000035, 0.00000199, 0.00000323, 0.00000283, 0.00000097, -0.00000159, -0.00000384, -0.00000499, -0.00000478, -0.00000353, -0.00000198, -0.00000093, -0.00000084, -0.00000169, -0.00000294, -0.00000384, -0.00000371, -0.00000229, 0.00000021, 0.00000307, 0.00000542, 0.00000643, 0.00000563, 0.00000305, -0.00000081, -0.00000513, -0.00000896, -0.00001147, -0.00001216, -0.00001094, -0.00000820, -0.00000470, -0.00000137, 0.00000095, 0.00000176, 0.00000108, },
		{ -0.00000131, 0.00000141, 0.00000227, 0.00000116, -0.00000127, -0.00000380, -0.00000516, -0.00000448, -0.00000178, 0.00000201, 0.00000534, 0.00000668, 0.00000523, 0.00000132, -0.00000361, -0.00000757, -0.00000880, -0.00000662, -0.00000170, 0.00000408, 0.00000846, 0.00000971, 0.00000738, 0.00000250, -0.00000286, -0.00000653, -0.00000714, -0.00000469, -0.00000049, 0.00000346, 0.00000544, 0.00000473, 0.00000189, -0.00000154, -0.00000379, -0.00000366, -0.00000104, 0.00000308, 0.00000709, 0.00000946, 0.00000933, 0.00000680, 0.00000278, -0.00000141, -0.00000458, -0.00000598, -0.00000544, -0.00000333, -0.00000029, 0.00000285, 0.00000534, 0.00000655, 0.00000619, 0.00000444, 0.00000207, 0.00000034, 0.00000067, 0.00000408, 0.00001062, 0.00001904, 0.00002697, 0.00003156, 0.00003052, 0.00002302, 0.00001024, -0.00000489, -0.00001861, -0.00002756, -0.00002997, -0.00002617, -0.00001839, -0.00000975, -0.00000312, -0.00000010, -0.00000063, -0.00000331, -0.00000621, -0.00000777, -0.00000743, -0.00000565, -0.00000355, -0.00000218, -0.00000203, -0.00000281, -0.00000372, -0.00000388, -0.00000289, -0.00000100, 0.00000100, 0.00000220, 0.00000203, 0.00000058, -0.00000148, -0.00000316, -0.00000364, -0.00000267, -0.00000064, 0.00000158, 0.00000303, 0.00000310, 0.00000178, -0.00000037, -0.00000248, -0.00000377, -0.00000387, -0.00000295, -0.00000159, -0.00000052, -0.00000025, -0.00000087, -0.00000202, -0.00000304, -0.00000325, -0.00000229, -0.00000022, 0.00000246, 0.00000496, 0.00000644, 0.00000630, 0.00000435, 0.00000086, -0.00000349, -0.00000775, -0.00001095, -0.00001234, -0.00001158, -0.00000892, -0.00000513, },
		{ -0.00001349, -0.00001324, -0.00001068, -0.00000652, -0.00000203, 0.00000137, 0.00000258, 0.00000130, -0.00000183, -0.00000533, -0.00000750, -0.00000707, -0.00000381, 0.00000126, 0.00000624, 0.00000906, 0.00000841, 0.00000436, -0.00000158, -0.00000701, -0.00000963, -0.00000821, -0.00000321, 0.00000343, 0.00000907, 0.00001147, 0.00000979, 0.00000486, -0.00000114, -0.00000572, -0.00000710, -0.00000500, -0.00000066, 0.00000382, 0.00000643, 0.00000613, 0.00000330, -0.00000056, -0.00000354, -0.00000423, -0.00000228, 0.00000151, 0.00000560, 0.00000837, 0.00000878, 0.00000667, 0.00000267, -0.00000213, -0.00000655, -0.00000968, -0.00001101, -0.00001040, -0.00000813, -0.00000487, -0.00000172, -0.00000004, -0.00000117, -0.00000582, -0.00001363, -0.00002288, -0.00003075, -0.00003410, -0.00003060, -0.00001973, -0.00000325, 0.00001512, 0.00003088, 0.00004018, 0.00004113, 0.00003442, 0.00002294, 0.00001068, 0.00000121, -0.00000345, -0.00000332, -0.00000014, 0.00000364, 0.00000593, 0.00000580, 0.00000361, 0.00000060, -0.00000187, -0.00000296, -0.00000268, -0.00000173, -0.00000098, -0.00000103, -0.00000187, -0.00000295, -0.00000357, -0.00000327, -0.00000210, -0.00000063, 0.00000037, 0.00000032, -0.00000088, -0.00000274, -0.00000447, -0.00000530, -0.00000485, -0.00000335, -0.00000147, -0.00000006, 0.00000026, -0.00000066, -0.00000243, -0.00000433, -0.00000565, -0.00000594, -0.00000526, -0.00000405, -0.00000296, -0.00000249, -0.00000283, -0.00000376, -0.00000473, -0.00000511, -0.00000447, -0.00000273, -0.00000022, 0.00000242, 0.00000438, 0.00000498, 0.00000382, 0.00000094, -0.00000319, -0.00000767, -0.00001143, },
		{ 0.00000312, -0.00000075, -0.00000545, -0.00000988, -0.00001282, -0.00001334, -0.00001117, -0.00000687, -0.00000180, 0.00000232, 0.00000400, 0.00000261, -0.00000126, -0.00000597, -0.00000937, -0.00000969, -0.00000636, -0.00000033, 0.00000619, 0.00001058, 0.00001096, 0.00000698, 0.00000012, -0.00000689, -0.00001118, -0.00001095, -0.00000629, 0.00000088, 0.00000763, 0.00001122, 0.00001032, 0.00000553, -0.00000097, -0.00000642, -0.00000866, -0.00000708, -0.00000273, 0.00000221, 0.00000552, 0.00000591, 0.00000352, -0.00000025, -0.00000348, -0.00000461, -0.00000308, 0.00000053, 0.00000487, 0.00000843, 0.00001014, 0.00000963, 0.00000719, 0.00000349, -0.00000063, -0.00000430, -0.00000661, -0.00000663, -0.00000357, 0.00000281, 0.00001179, 0.00002148, 0.00002904, 0.00003149, 0.00002675, 0.00001469, -0.00000250, -0.00002076, -0.00003534, -0.00004246, -0.00004060, -0.00003104, -0.00001740, -0.00000424, 0.00000456, 0.00000709, 0.00000387, -0.00000267, -0.00000935, -0.00001356, -0.00001419, -0.00001177, -0.00000795, -0.00000460, -0.00000296, -0.00000324, -0.00000470, -0.00000626, -0.00000701, -0.00000670, -0.00000568, -0.00000464, -0.00000413, -0.00000429, -0.00000480, -0.00000509, -0.00000470, -0.00000356, -0.00000206, -0.00000083, -0.00000044, -0.00000106, -0.00000239, -0.00000379, -0.00000457, -0.00000431, -0.00000308, -0.00000139, 0.00000004, 0.00000060, 0.00000009, -0.00000126, -0.00000287, -0.00000408, -0.00000447, -0.00000399, -0.00000296, -0.00000192, -0.00000136, -0.00000153, -0.00000229, -0.00000322, -0.00000379, -0.00000352, -0.00000224, -0.00000008, 0.00000245, 0.00000467, 0.00000584, 0.00000540, },
		{ 0.00000047, 0.00000278, 0.00000442, 0.00000468, 0.00000310, -0.00000036, -0.00000515, -0.00001018, -0.00001407, -0.00001554, -0.00001395, -0.00000965, -0.00000400, 0.00000100, 0.00000342, 0.00000222, -0.00000222, -0.00000813, -0.00001295, -0.00001436, -0.00001130, -0.00000454, 0.00000349, 0.00000966, 0.00001143, 0.00000794, 0.00000050, -0.00000792, -0.00001394, -0.00001513, -0.00001110, -0.00000364, 0.00000408, 0.00000888, 0.00000893, 0.00000449, -0.00000235, -0.00000861, -0.00001182, -0.00001103, -0.00000708, -0.00000213, 0.00000150, 0.00000230, 0.00000021, -0.00000355, -0.00000717, -0.00000909, -0.00000869, -0.00000638, -0.00000322, -0.00000039, 0.00000128, 0.00000140, -0.00000014, -0.00000343, -0.00000866, -0.00001583, -0.00002439, -0.00003282, -0.00003875, -0.00003960, -0.00003355, -0.00002059, -0.00000291, 0.00001549, 0.00002999, 0.00003698, 0.00003518, 0.00002617, 0.00001375, 0.00000252, -0.00000382, -0.00000378, 0.00000156, 0.00000920, 0.00001562, 0.00001820, 0.00001616, 0.00001061, 0.00000388, -0.00000159, -0.00000430, -0.00000407, -0.00000192, 0.00000058, 0.00000213, 0.00000219, 0.00000104, -0.00000052, -0.00000167, -0.00000202, -0.00000170, -0.00000118, -0.00000097, -0.00000127, -0.00000190, -0.00000241, -0.00000239, -0.00000172, -0.00000064, 0.00000033, 0.00000068, 0.00000018, -0.00000100, -0.00000236, -0.00000329, -0.00000338, -0.00000260, -0.00000133, -0.00000016, 0.00000037, -0.00000000, -0.00000113, -0.00000256, -0.00000373, -0.00000423, -0.00000398, -0.00000320, -0.00000234, -0.00000182, -0.00000191, -0.00000255, -0.00000346, -0.00000418, -0.00000428, -0.00000348, -0.00000180, },
		{ -0.00000474, -0.00000502, -0.00000459, -0.00000333, -0.00000132, 0.00000105, 0.00000318, 0.00000428, 0.00000365, 0.00000094, -0.00000359, -0.00000895, -0.00001363, -0.00001602, -0.00001506, -0.00001076, -0.00000442, 0.00000175, 0.00000536, 0.00000488, 0.00000032, -0.00000651, -0.00001269, -0.00001530, -0.00001270, -0.00000534, 0.00000428, 0.00001254, 0.00001622, 0.00001385, 0.00000643, -0.00000295, -0.00001042, -0.00001292, -0.00000952, -0.00000177, 0.00000703, 0.00001331, 0.00001475, 0.00001124, 0.00000474, -0.00000167, -0.00000527, -0.00000481, -0.00000093, 0.00000432, 0.00000861, 0.00001039, 0.00000949, 0.00000700, 0.00000456, 0.00000346, 0.00000414, 0.00000621, 0.00000892, 0.00001182, 0.00001501, 0.00001898, 0.00002396, 0.00002929, 0.00003322, 0.00003334, 0.00002766, 0.00001568, -0.00000091, -0.00001847, -0.00003257, -0.00003955, -0.00003801, -0.00002933, -0.00001715, -0.00000596, 0.00000068, 0.00000141, -0.00000257, -0.00000827, -0.00001231, -0.00001239, -0.00000822, -0.00000140, 0.00000537, 0.00000956, 0.00000987, 0.00000663, 0.00000147, -0.00000347, -0.00000654, -0.00000710, -0.00000564, -0.00000333, -0.00000139, -0.00000055, -0.00000083, -0.00000171, -0.00000253, -0.00000285, -0.00000266, -0.00000225, -0.00000204, -0.00000222, -0.00000270, -0.00000314, -0.00000319, -0.00000268, -0.00000177, -0.00000089, -0.00000046, -0.00000076, -0.00000169, -0.00000287, -0.00000378, -0.00000403, -0.00000354, -0.00000257, -0.00000161, -0.00000113, -0.00000138, -0.00000227, -0.00000344, -0.00000445, -0.00000493, -0.00000478, -0.00000416, -0.00000341, -0.00000290, -0.00000287, -0.00000331, -0.00000405, },
		{ -0.00000046, -0.00000024, -0.00000041, -0.00000084, -0.00000135, -0.00000167, -0.00000153, -0.00000072, 0.00000089, 0.00000315, 0.00000559, 0.00000743, 0.00000777, 0.00000591, 0.00000177, -0.00000389, -0.00000947, -0.00001307, -0.00001315, -0.00000930, -0.00000261, 0.00000455, 0.00000933, 0.00000961, 0.00000497, -0.00000293, -0.00001086, -0.00001530, -0.00001394, -0.00000677, 0.00000376, 0.00001365, 0.00001897, 0.00001757, 0.00001003, -0.00000058, -0.00000996, -0.00001444, -0.00001252, -0.00000544, 0.00000347, 0.00001032, 0.00001235, 0.00000908, 0.00000232, -0.00000485, -0.00000954, -0.00001036, -0.00000783, -0.00000391, -0.00000086, -0.00000006, -0.00000153, -0.00000413, -0.00000651, -0.00000799, -0.00000905, -0.00001096, -0.00001488, -0.00002079, -0.00002703, -0.00003073, -0.00002898, -0.00002030, -0.00000566, 0.00001155, 0.00002663, 0.00003534, 0.00003557, 0.00002817, 0.00001653, 0.00000514, -0.00000219, -0.00000379, -0.00000060, 0.00000461, 0.00000859, 0.00000917, 0.00000608, 0.00000094, -0.00000369, -0.00000558, -0.00000383, 0.00000077, 0.00000621, 0.00001019, 0.00001117, 0.00000891, 0.00000444, -0.00000049, -0.00000426, -0.00000596, -0.00000566, -0.00000418, -0.00000254, -0.00000153, -0.00000138, -0.00000184, -0.00000243, -0.00000276, -0.00000272, -0.00000250, -0.00000239, -0.00000258, -0.00000300, -0.00000338, -0.00000343, -0.00000299, -0.00000215, -0.00000126, -0.00000071, -0.00000075, -0.00000135, -0.00000221, -0.00000290, -0.00000308, -0.00000264, -0.00000179, -0.00000090, -0.00000038, -0.00000042, -0.00000097, -0.00000175, -0.00000240, -0.00000264, -0.00000237, -0.00000175, -0.00000102, },
		{ -0.00000341, -0.00000328, -0.00000291, -0.00000250, -0.00000223, -0.00000218, -0.00000235, -0.00000271, -0.00000317, -0.00000365, -0.00000395, -0.00000380, -0.00000290, -0.00000112, 0.00000130, 0.00000367, 0.00000491, 0.00000399, 0.00000041, -0.00000537, -0.00001180, -0.00001669, -0.00001802, -0.00001485, -0.00000794, 0.00000028, 0.00000655, 0.00000811, 0.00000396, -0.00000451, -0.00001388, -0.00002006, -0.00001997, -0.00001301, -0.00000146, 0.00001039, 0.00001796, 0.00001836, 0.00001164, 0.00000077, -0.00000962, -0.00001529, -0.00001415, -0.00000708, 0.00000259, 0.00001073, 0.00001424, 0.00001236, 0.00000682, 0.00000072, -0.00000299, -0.00000296, 0.00000016, 0.00000434, 0.00000750, 0.00000880, 0.00000912, 0.00001051, 0.00001486, 0.00002246, 0.00003131, 0.00003767, 0.00003755, 0.00002862, 0.00001158, -0.00000976, -0.00002973, -0.00004291, -0.00004622, -0.00004000, -0.00002776, -0.00001458, -0.00000503, -0.00000150, -0.00000351, -0.00000837, -0.00001268, -0.00001393, -0.00001154, -0.00000685, -0.00000223, 0.00000018, -0.00000046, -0.00000338, -0.00000667, -0.00000831, -0.00000710, -0.00000321, 0.00000194, 0.00000640, 0.00000857, 0.00000787, 0.00000481, 0.00000075, -0.00000282, -0.00000489, -0.00000519, -0.00000419, -0.00000270, -0.00000145, -0.00000080, -0.00000069, -0.00000083, -0.00000092, -0.00000084, -0.00000071, -0.00000071, -0.00000099, -0.00000146, -0.00000191, -0.00000207, -0.00000179, -0.00000116, -0.00000044, 0.00000000, -0.00000005, -0.00000060, -0.00000141, -0.00000212, -0.00000244, -0.00000228, -0.00000179, -0.00000128, -0.00000107, -0.00000129, -0.00000189, -0.00000261, -0.00000318, },
		{ -0.00000254, -0.00000299, -0.00000342, -0.00000363, -0.00000351, -0.00000314, -0.00000270, -0.00000235, -0.00000215, -0.00000207, -0.00000203, -0.00000202, -0.00000210, -0.00000235, -0.00000274, -0.00000300, -0.00000269, -0.00000138, 0.00000099, 0.00000389, 0.00000618, 0.00000650, 0.00000386, -0.00000168, -0.00000874, -0.00001493, -0.00001772, -0.00001556, -0.00000876, 0.00000033, 0.00000809, 0.00001108, 0.00000762, -0.00000133, -0.00001229, -0.00002062, -0.00002247, -0.00001655, -0.00000479, 0.00000830, 0.00001753, 0.00001926, 0.00001292, 0.00000132, -0.00001068, -0.00001828, -0.00001886, -0.00001301, -0.00000406, 0.00000359, 0.00000660, 0.00000411, -0.00000208, -0.00000867, -0.00001276, -0.00001340, -0.00001213, -0.00001197, -0.00001563, -0.00002367, -0.00003372, -0.00004113, -0.00004099, -0.00003044, -0.00001032, 0.00001471, 0.00003778, 0.00005233, 0.00005469, 0.00004533, 0.00002856, 0.00001067, -0.00000262, -0.00000820, -0.00000641, -0.00000032, 0.00000593, 0.00000907, 0.00000799, 0.00000380, -0.00000103, -0.00000410, -0.00000421, -0.00000184, 0.00000132, 0.00000334, 0.00000309, 0.00000071, -0.00000247, -0.00000473, -0.00000478, -0.00000238, 0.00000154, 0.00000542, 0.00000773, 0.00000766, 0.00000535, 0.00000174, -0.00000189, -0.00000452, -0.00000567, -0.00000551, -0.00000460, -0.00000357, -0.00000280, -0.00000240, -0.00000222, -0.00000210, -0.00000197, -0.00000190, -0.00000203, -0.00000242, -0.00000298, -0.00000349, -0.00000370, -0.00000348, -0.00000291, -0.00000224, -0.00000177, -0.00000172, -0.00000210, -0.00000269, -0.00000320, -0.00000338, -0.00000320, -0.00000278, -0.00000241, -0.00000231, },
		{ -0.00000243, -0.00000223, -0.00000218, -0.00000243, -0.00000295, -0.00000355, -0.00000396, -0.00000399, -0.00000366, -0.00000319, -0.00000282, -0.00000271, -0.00000285, -0.00000302, -0.00000302, -0.00000282, -0.00000258, -0.00000260, -0.00000306, -0.00000380, -0.00000425, -0.00000370, -0.00000171, 0.00000146, 0.00000471, 0.00000636, 0.00000495, 0.00000001, -0.00000739, -0.00001476, -0.00001907, -0.00001808, -0.00001153, -0.00000153, 0.00000803, 0.00001304, 0.00001101, 0.00000235, -0.00000953, -0.00001953, -0.00002300, -0.00001782, -0.00000549, 0.00000938, 0.00002104, 0.00002504, 0.00002022, 0.00000919, -0.00000290, -0.00001072, -0.00001114, -0.00000456, 0.00000550, 0.00001429, 0.00001823, 0.00001676, 0.00001254, 0.00000997, 0.00001275, 0.00002167, 0.00003375, 0.00004328, 0.00004426, 0.00003331, 0.00001149, -0.00001566, -0.00004015, -0.00005458, -0.00005506, -0.00004265, -0.00002275, -0.00000286, 0.00001045, 0.00001391, 0.00000838, -0.00000202, -0.00001201, -0.00001750, -0.00001704, -0.00001197, -0.00000539, -0.00000045, 0.00000109, -0.00000062, -0.00000385, -0.00000645, -0.00000694, -0.00000516, -0.00000221, 0.00000032, 0.00000119, 0.00000015, -0.00000198, -0.00000380, -0.00000408, -0.00000231, 0.00000101, 0.00000469, 0.00000737, 0.00000814, 0.00000686, 0.00000412, 0.00000094, -0.00000173, -0.00000330, -0.00000372, -0.00000330, -0.00000249, -0.00000166, -0.00000097, -0.00000043, 0.00000000, 0.00000031, 0.00000040, 0.00000018, -0.00000038, -0.00000112, -0.00000180, -0.00000217, -0.00000211, -0.00000170, -0.00000118, -0.00000085, -0.00000089, -0.00000129, -0.00000186, -0.00000233, -0.00000252, },
		{ -0.00000301, -0.00000332, -0.00000347, -0.00000333, -0.00000298, -0.00000264, -0.00000259, -0.00000294, -0.00000356, -0.00000413, -0.00000429, -0.00000389, -0.00000303, -0.00000207, -0.00000143, -0.00000131, -0.00000160, -0.00000192, -0.00000188, -0.00000133, -0.00000054, -0.00000001, -0.00000022, -0.00000119, -0.00000233, -0.00000266, -0.00000130, 0.00000188, 0.00000591, 0.00000892, 0.00000889, 0.00000475, -0.00000289, -0.00001155, -0.00001779, -0.00001863, -0.00001306, -0.00000283, 0.00000802, 0.00001471, 0.00001386, 0.00000517, -0.00000817, -0.00002066, -0.00002681, -0.00002363, -0.00001208, 0.00000317, 0.00001575, 0.00002037, 0.00001524, 0.00000282, -0.00001132, -0.00002116, -0.00002312, -0.00001776, -0.00000950, -0.00000443, -0.00000717, -0.00001835, -0.00003388, -0.00004652, -0.00004898, -0.00003736, -0.00001313, 0.00001711, 0.00004408, 0.00005952, 0.00005941, 0.00004539, 0.00002394, 0.00000352, -0.00000883, -0.00001007, -0.00000194, 0.00001032, 0.00002055, 0.00002429, 0.00002041, 0.00001115, 0.00000066, -0.00000704, -0.00000972, -0.00000765, -0.00000307, 0.00000118, 0.00000301, 0.00000191, -0.00000111, -0.00000424, -0.00000592, -0.00000560, -0.00000384, -0.00000191, -0.00000099, -0.00000159, -0.00000332, -0.00000514, -0.00000591, -0.00000504, -0.00000267, 0.00000033, 0.00000283, 0.00000394, 0.00000330, 0.00000124, -0.00000151, -0.00000414, -0.00000604, -0.00000700, -0.00000711, -0.00000665, -0.00000587, -0.00000498, -0.00000405, -0.00000318, -0.00000250, -0.00000216, -0.00000226, -0.00000279, -0.00000355, -0.00000424, -0.00000458, -0.00000445, -0.00000394, -0.00000332, -0.00000288, -0.00000279, },
		{ -0.00000134, -0.00000086, -0.00000059, -0.00000068, -0.00000106, -0.00000147, -0.00000164, -0.00000146, -0.00000109, -0.00000087, -0.00000112, -0.00000194, -0.00000310, -0.00000412, -0.00000451, -0.00000405, -0.00000294, -0.00000174, -0.00000105, -0.00000119, -0.00000200, -0.00000291, -0.00000326, -0.00000274, -0.00000160, -0.00000059, -0.00000052, -0.00000172, -0.00000366, -0.00000512, -0.00000476, -0.00000192, 0.00000268, 0.00000709, 0.00000878, 0.00000594, -0.00000138, -0.00001091, -0.00001885, -0.00002144, -0.00001679, -0.00000609, 0.00000658, 0.00001582, 0.00001727, 0.00000975, -0.00000387, -0.00001773, -0.00002544, -0.00002295, -0.00001060, 0.00000688, 0.00002237, 0.00002956, 0.00002598, 0.00001426, 0.00000110, -0.00000583, -0.00000154, 0.00001367, 0.00003388, 0.00004990, 0.00005316, 0.00003952, 0.00001135, -0.00002319, -0.00005330, -0.00006979, -0.00006862, -0.00005227, -0.00002839, -0.00000648, 0.00000600, 0.00000631, -0.00000297, -0.00001569, -0.00002510, -0.00002682, -0.00002039, -0.00000904, 0.00000217, 0.00000873, 0.00000861, 0.00000279, -0.00000554, -0.00001260, -0.00001580, -0.00001459, -0.00001041, -0.00000568, -0.00000257, -0.00000204, -0.00000364, -0.00000597, -0.00000754, -0.00000752, -0.00000607, -0.00000407, -0.00000258, -0.00000226, -0.00000300, -0.00000411, -0.00000464, -0.00000396, -0.00000201, 0.00000066, 0.00000317, 0.00000470, 0.00000482, 0.00000362, 0.00000158, -0.00000067, -0.00000257, -0.00000383, -0.00000437, -0.00000428, -0.00000369, -0.00000271, -0.00000151, -0.00000026, 0.00000077, 0.00000135, 0.00000131, 0.00000069, -0.00000026, -0.00000116, -0.00000169, -0.00000172, },
		{ -0.00000168, -0.00000287, -0.00000364, -0.00000383, -0.00000356, -0.00000318, -0.00000303, -0.00000329, -0.00000382, -0.00000430, -0.00000440, -0.00000402, -0.00000342, -0.00000305, -0.00000337, -0.00000450, -0.00000610, -0.00000752, -0.00000807, -0.00000742, -0.00000583, -0.00000406, -0.00000297, -0.00000309, -0.00000425, -0.00000571, -0.00000648, -0.00000597, -0.00000436, -0.00000261, -0.00000192, -0.00000303, -0.00000562, -0.00000830, -0.00000925, -0.00000720, -0.00000241, 0.00000318, 0.00000657, 0.00000517, -0.00000177, -0.00001235, -0.00002251, -0.00002768, -0.00002498, -0.00001483, -0.00000121, 0.00000983, 0.00001279, 0.00000544, -0.00000990, -0.00002697, -0.00003829, -0.00003854, -0.00002731, -0.00000965, 0.00000585, 0.00001083, 0.00000109, -0.00002106, -0.00004728, -0.00006639, -0.00006894, -0.00005135, -0.00001763, 0.00002202, 0.00005510, 0.00007173, 0.00006831, 0.00004858, 0.00002172, -0.00000165, -0.00001373, -0.00001228, -0.00000086, 0.00001339, 0.00002320, 0.00002430, 0.00001691, 0.00000509, -0.00000546, -0.00001019, -0.00000756, 0.00000056, 0.00001005, 0.00001662, 0.00001766, 0.00001322, 0.00000560, -0.00000189, -0.00000651, -0.00000716, -0.00000460, -0.00000079, 0.00000220, 0.00000312, 0.00000196, -0.00000031, -0.00000237, -0.00000325, -0.00000279, -0.00000154, -0.00000037, 0.00000006, -0.00000039, -0.00000130, -0.00000198, -0.00000181, -0.00000064, 0.00000124, 0.00000319, 0.00000454, 0.00000484, 0.00000404, 0.00000239, 0.00000033, -0.00000171, -0.00000342, -0.00000460, -0.00000517, -0.00000509, -0.00000439, -0.00000318, -0.00000175, -0.00000047, 0.00000028, 0.00000027, -0.00000048, },
		{ -0.00000163, -0.00000038, -0.00000006, -0.00000066, -0.00000187, -0.00000313, -0.00000394, -0.00000407, -0.00000367, -0.00000315, -0.00000292, -0.00000318, -0.00000375, -0.00000420, -0.00000410, -0.00000333, -0.00000223, -0.00000142, -0.00000152, -0.00000274, -0.00000470, -0.00000650, -0.00000720, -0.00000629, -0.00000402, -0.00000134, 0.00000053, 0.00000080, -0.00000047, -0.00000235, -0.00000350, -0.00000292, -0.00000061, 0.00000234, 0.00000429, 0.00000398, 0.00000137, -0.00000211, -0.00000413, -0.00000272, 0.00000253, 0.00000987, 0.00001589, 0.00001714, 0.00001191, 0.00000150, -0.00000988, -0.00001681, -0.00001518, -0.00000436, 0.00001194, 0.00002701, 0.00003398, 0.00002918, 0.00001428, -0.00000390, -0.00001599, -0.00001425, 0.00000366, 0.00003298, 0.00006313, 0.00008173, 0.00007966, 0.00005498, 0.00001407, -0.00003061, -0.00006525, -0.00007995, -0.00007231, -0.00004794, -0.00001775, 0.00000664, 0.00001748, 0.00001346, -0.00000054, -0.00001624, -0.00002590, -0.00002553, -0.00001616, -0.00000280, 0.00000828, 0.00001251, 0.00000888, 0.00000006, -0.00000921, -0.00001443, -0.00001337, -0.00000684, 0.00000199, 0.00000922, 0.00001200, 0.00000964, 0.00000368, -0.00000309, -0.00000799, -0.00000956, -0.00000798, -0.00000470, -0.00000160, -0.00000006, -0.00000046, -0.00000218, -0.00000413, -0.00000536, -0.00000546, -0.00000467, -0.00000365, -0.00000299, -0.00000297, -0.00000341, -0.00000381, -0.00000372, -0.00000290, -0.00000149, 0.00000009, 0.00000135, 0.00000191, 0.00000164, 0.00000062, -0.00000090, -0.00000266, -0.00000440, -0.00000589, -0.00000688, -0.00000716, -0.00000659, -0.00000524, -0.00000342, },
		{ -0.00000452, -0.00000419, -0.00000281, -0.00000066, 0.00000164, 0.00000341, 0.00000411, 0.00000366, 0.00000239, 0.00000097, 0.00000003, -0.00000013, 0.00000034, 0.00000094, 0.00000115, 0.00000072, -0.00000013, -0.00000082, -0.00000079, 0.00000015, 0.00000163, 0.00000281, 0.00000282, 0.00000129, -0.00000136, -0.00000403, -0.00000545, -0.00000481, -0.00000228, 0.00000097, 0.00000338, 0.00000372, 0.00000186, -0.00000115, -0.00000357, -0.00000388, -0.00000176, 0.00000168, 0.00000433, 0.00000428, 0.00000094, -0.00000438, -0.00000890, -0.00000980, -0.00000581, 0.00000171, 0.00000908, 0.00001184, 0.00000705, -0.00000488, -0.00001981, -0.00003140, -0.00003398, -0.00002548, -0.00000908, 0.00000756, 0.00001536, 0.00000794, -0.00001502, -0.00004677, -0.00007555, -0.00008910, -0.00007969, -0.00004763, -0.00000162, 0.00004435, 0.00007625, 0.00008530, 0.00007115, 0.00004155, 0.00000893, -0.00001475, -0.00002253, -0.00001468, 0.00000227, 0.00001899, 0.00002757, 0.00002471, 0.00001264, -0.00000254, -0.00001397, -0.00001721, -0.00001193, -0.00000167, 0.00000818, 0.00001301, 0.00001098, 0.00000358, -0.00000534, -0.00001162, -0.00001261, -0.00000829, -0.00000098, 0.00000590, 0.00000947, 0.00000854, 0.00000393, -0.00000212, -0.00000714, -0.00000947, -0.00000884, -0.00000625, -0.00000330, -0.00000137, -0.00000109, -0.00000218, -0.00000381, -0.00000510, -0.00000552, -0.00000507, -0.00000418, -0.00000332, -0.00000282, -0.00000266, -0.00000257, -0.00000224, -0.00000148, -0.00000032, 0.00000099, 0.00000215, 0.00000291, 0.00000313, 0.00000281, 0.00000197, 0.00000070, -0.00000087, -0.00000250, -0.00000385, },
		{ -0.00000034, -0.00000188, -0.00000384, -0.00000580, -0.00000720, -0.00000750, -0.00000647, -0.00000433, -0.00000177, 0.00000033, 0.00000125, 0.00000077, -0.00000075, -0.00000256, -0.00000385, -0.00000420, -0.00000377, -0.00000315, -0.00000302, -0.00000372, -0.00000501, -0.00000615, -0.00000635, -0.00000527, -0.00000332, -0.00000155, -0.00000113, -0.00000269, -0.00000587, -0.00000937, -0.00001148, -0.00001102, -0.00000797, -0.00000360, 0.00000007, 0.00000130, -0.00000041, -0.00000396, -0.00000716, -0.00000787, -0.00000523, -0.00000025, 0.00000454, 0.00000642, 0.00000403, -0.00000162, -0.00000739, -0.00000945, -0.00000536, 0.00000428, 0.00001565, 0.00002321, 0.00002239, 0.00001223, -0.00000344, -0.00001707, -0.00002046, -0.00000871, 0.00001693, 0.00004861, 0.00007445, 0.00008308, 0.00006850, 0.00003290, -0.00001363, -0.00005680, -0.00008348, -0.00008671, -0.00006819, -0.00003725, -0.00000676, 0.00001208, 0.00001405, 0.00000138, -0.00001793, -0.00003401, -0.00003956, -0.00003267, -0.00001713, -0.00000018, 0.00001098, 0.00001243, 0.00000489, -0.00000701, -0.00001722, -0.00002109, -0.00001733, -0.00000825, 0.00000168, 0.00000802, 0.00000841, 0.00000344, -0.00000395, -0.00000999, -0.00001189, -0.00000897, -0.00000277, 0.00000386, 0.00000815, 0.00000861, 0.00000550, 0.00000050, -0.00000422, -0.00000696, -0.00000713, -0.00000531, -0.00000278, -0.00000081, -0.00000017, -0.00000085, -0.00000229, -0.00000372, -0.00000460, -0.00000474, -0.00000433, -0.00000373, -0.00000320, -0.00000283, -0.00000254, -0.00000217, -0.00000164, -0.00000096, -0.00000024, 0.00000040, 0.00000086, 0.00000111, 0.00000106, 0.00000062, },
		{ -0.00000081, -0.00000046, -0.00000006, 0.00000014, -0.00000017, -0.00000128, -0.00000317, -0.00000545, -0.00000737, -0.00000816, -0.00000736, -0.00000508, -0.00000204, 0.00000068, 0.00000213, 0.00000191, 0.00000034, -0.00000167, -0.00000315, -0.00000352, -0.00000289, -0.00000198, -0.00000169, -0.00000251, -0.00000420, -0.00000584, -0.00000638, -0.00000524, -0.00000277, -0.00000025, 0.00000078, -0.00000072, -0.00000454, -0.00000920, -0.00001256, -0.00001289, -0.00000982, -0.00000469, 0.00000001, 0.00000187, -0.00000015, -0.00000506, -0.00001025, -0.00001279, -0.00001105, -0.00000576, 0.00000018, 0.00000306, 0.00000040, -0.00000744, -0.00001711, -0.00002361, -0.00002289, -0.00001430, -0.00000152, 0.00000845, 0.00000837, -0.00000562, -0.00003151, -0.00006128, -0.00008357, -0.00008818, -0.00007048, -0.00003377, 0.00001151, 0.00005157, 0.00007462, 0.00007529, 0.00005655, 0.00002812, 0.00000223, -0.00001149, -0.00000962, 0.00000413, 0.00002123, 0.00003246, 0.00003203, 0.00001991, 0.00000144, -0.00001551, -0.00002417, -0.00002181, -0.00001069, 0.00000337, 0.00001380, 0.00001629, 0.00001047, -0.00000033, -0.00001089, -0.00001661, -0.00001553, -0.00000895, -0.00000048, 0.00000581, 0.00000724, 0.00000366, -0.00000281, -0.00000892, -0.00001191, -0.00001066, -0.00000604, -0.00000032, 0.00000400, 0.00000528, 0.00000334, -0.00000067, -0.00000493, -0.00000783, -0.00000860, -0.00000747, -0.00000539, -0.00000349, -0.00000255, -0.00000276, -0.00000379, -0.00000507, -0.00000607, -0.00000652, -0.00000644, -0.00000598, -0.00000534, -0.00000462, -0.00000386, -0.00000310, -0.00000239, -0.00000181, -0.00000139, -0.00000109, },
		{ -0.00000260, -0.00000218, -0.00000219, -0.00000247, -0.00000269, -0.00000257, -0.00000205, -0.00000138, -0.00000112, -0.00000179, -0.00000361, -0.00000624, -0.00000883, -0.00001034, -0.00001001, -0.00000778, -0.00000437, -0.00000105, 0.00000095, 0.00000100, -0.00000064, -0.00000296, -0.00000472, -0.00000513, -0.00000423, -0.00000288, -0.00000223, -0.00000301, -0.00000504, -0.00000724, -0.00000818, -0.00000694, -0.00000372, 0.00000002, 0.00000228, 0.00000151, -0.00000239, -0.00000784, -0.00001221, -0.00001307, -0.00000949, -0.00000275, 0.00000422, 0.00000818, 0.00000729, 0.00000223, -0.00000398, -0.00000742, -0.00000534, 0.00000220, 0.00001202, 0.00001926, 0.00001984, 0.00001291, 0.00000184, -0.00000683, -0.00000630, 0.00000702, 0.00003112, 0.00005835, 0.00007804, 0.00008080, 0.00006261, 0.00002701, -0.00001594, -0.00005337, -0.00007457, -0.00007510, -0.00005825, -0.00003325, -0.00001116, -0.00000021, -0.00000259, -0.00001419, -0.00002705, -0.00003333, -0.00002902, -0.00001555, 0.00000114, 0.00001361, 0.00001646, 0.00000878, -0.00000564, -0.00002017, -0.00002849, -0.00002734, -0.00001777, -0.00000437, 0.00000699, 0.00001187, 0.00000907, 0.00000085, -0.00000840, -0.00001431, -0.00001449, -0.00000937, -0.00000172, 0.00000483, 0.00000756, 0.00000572, 0.00000073, -0.00000474, -0.00000810, -0.00000796, -0.00000465, 0.00000014, 0.00000425, 0.00000605, 0.00000507, 0.00000202, -0.00000167, -0.00000456, -0.00000582, -0.00000542, -0.00000402, -0.00000252, -0.00000165, -0.00000170, -0.00000255, -0.00000380, -0.00000503, -0.00000596, -0.00000645, -0.00000649, -0.00000611, -0.00000538, -0.00000442, -0.00000341, },
		{ -0.00000881, -0.00000779, -0.00000626, -0.00000455, -0.00000310, -0.00000230, -0.00000228, -0.00000279, -0.00000331, -0.00000329, -0.00000247, -0.00000112, 0.00000006, 0.00000023, -0.00000112, -0.00000379, -0.00000687, -0.00000903, -0.00000919, -0.00000699, -0.00000314, 0.00000093, 0.00000364, 0.00000408, 0.00000237, -0.00000035, -0.00000256, -0.00000316, -0.00000208, -0.00000030, 0.00000071, -0.00000013, -0.00000282, -0.00000610, -0.00000813, -0.00000748, -0.00000411, 0.00000045, 0.00000366, 0.00000329, -0.00000128, -0.00000853, -0.00001530, -0.00001828, -0.00001577, -0.00000872, -0.00000054, 0.00000448, 0.00000335, -0.00000394, -0.00001409, -0.00002198, -0.00002322, -0.00001667, -0.00000554, 0.00000368, 0.00000410, -0.00000808, -0.00003114, -0.00005776, -0.00007754, -0.00008118, -0.00006456, -0.00003080, 0.00001065, 0.00004756, 0.00006969, 0.00007261, 0.00005914, 0.00003758, 0.00001787, 0.00000731, 0.00000779, 0.00001565, 0.00002402, 0.00002655, 0.00002046, 0.00000774, -0.00000602, -0.00001462, -0.00001429, -0.00000557, 0.00000701, 0.00001709, 0.00001953, 0.00001276, -0.00000065, -0.00001520, -0.00002508, -0.00002674, -0.00002031, -0.00000932, 0.00000109, 0.00000653, 0.00000519, -0.00000159, -0.00001025, -0.00001679, -0.00001861, -0.00001546, -0.00000935, -0.00000336, -0.00000015, -0.00000082, -0.00000458, -0.00000935, -0.00001282, -0.00001351, -0.00001135, -0.00000751, -0.00000376, -0.00000161, -0.00000168, -0.00000361, -0.00000633, -0.00000867, -0.00000983, -0.00000966, -0.00000855, -0.00000716, -0.00000611, -0.00000569, -0.00000594, -0.00000665, -0.00000755, -0.00000841, -0.00000903, -0.00000921, },
		{ -0.00000360, -0.00000473, -0.00000589, -0.00000675, -0.00000696, -0.00000627, -0.00000473, -0.00000276, -0.00000100, -0.00000011, -0.00000038, -0.00000155, -0.00000291, -0.00000360, -0.00000310, -0.00000155, 0.00000020, 0.00000100, -0.00000003, -0.00000294, -0.00000680, -0.00001003, -0.00001110, -0.00000934, -0.00000530, -0.00000060, 0.00000282, 0.00000364, 0.00000181, -0.00000145, -0.00000427, -0.00000515, -0.00000378, -0.00000125, 0.00000058, 0.00000015, -0.00000281, -0.00000697, -0.00000999, -0.00000977, -0.00000576, 0.00000049, 0.00000593, 0.00000743, 0.00000358, -0.00000434, -0.00001265, -0.00001693, -0.00001426, -0.00000494, 0.00000732, 0.00001694, 0.00001920, 0.00001285, 0.00000124, -0.00000881, -0.00000998, 0.00000191, 0.00002537, 0.00005305, 0.00007433, 0.00007951, 0.00006398, 0.00003040, -0.00001208, -0.00005124, -0.00007655, -0.00008298, -0.00007249, -0.00005258, -0.00003262, -0.00001982, -0.00001658, -0.00002023, -0.00002516, -0.00002609, -0.00002078, -0.00001098, -0.00000132, 0.00000337, 0.00000060, -0.00000836, -0.00001912, -0.00002630, -0.00002621, -0.00001865, -0.00000701, 0.00000337, 0.00000766, 0.00000384, -0.00000647, -0.00001878, -0.00002788, -0.00003009, -0.00002486, -0.00001477, -0.00000421, 0.00000265, 0.00000362, -0.00000071, -0.00000759, -0.00001348, -0.00001572, -0.00001359, -0.00000840, -0.00000268, 0.00000113, 0.00000174, -0.00000056, -0.00000421, -0.00000727, -0.00000831, -0.00000699, -0.00000407, -0.00000091, 0.00000122, 0.00000166, 0.00000055, -0.00000137, -0.00000322, -0.00000433, -0.00000451, -0.00000396, -0.00000312, -0.00000240, -0.00000205, -0.00000216, -0.00000270, },
		{ -0.00000456, -0.00000433, -0.00000426, -0.00000448, -0.00000512, -0.00000626, -0.00000775, -0.00000917, -0.00000996, -0.00000960, -0.00000800, -0.00000557, -0.00000320, -0.00000182, -0.00000201, -0.00000360, -0.00000570, -0.00000712, -0.00000697, -0.00000519, -0.00000269, -0.00000095, -0.00000131, -0.00000417, -0.00000869, -0.00001301, -0.00001515, -0.00001398, -0.00000982, -0.00000443, -0.00000017, 0.00000112, -0.00000091, -0.00000498, -0.00000882, -0.00001043, -0.00000912, -0.00000603, -0.00000345, -0.00000362, -0.00000730, -0.00001317, -0.00001838, -0.00001995, -0.00001657, -0.00000956, -0.00000255, 0.00000020, -0.00000396, -0.00001439, -0.00002703, -0.00003605, -0.00003676, -0.00002815, -0.00001404, -0.00000184, 0.00000064, -0.00001104, -0.00003532, -0.00006446, -0.00008724, -0.00009330, -0.00007753, -0.00004238, 0.00000285, 0.00004548, 0.00007423, 0.00008333, 0.00007410, 0.00005372, 0.00003168, 0.00001580, 0.00000937, 0.00001075, 0.00001515, 0.00001764, 0.00001569, 0.00001021, 0.00000458, 0.00000248, 0.00000564, 0.00001278, 0.00002021, 0.00002373, 0.00002089, 0.00001229, 0.00000136, -0.00000724, -0.00000991, -0.00000587, 0.00000248, 0.00001067, 0.00001433, 0.00001116, 0.00000202, -0.00000956, -0.00001901, -0.00002275, -0.00001974, -0.00001178, -0.00000253, 0.00000418, 0.00000600, 0.00000290, -0.00000305, -0.00000884, -0.00001194, -0.00001131, -0.00000772, -0.00000312, 0.00000034, 0.00000133, -0.00000018, -0.00000312, -0.00000593, -0.00000735, -0.00000694, -0.00000515, -0.00000294, -0.00000131, -0.00000083, -0.00000145, -0.00000271, -0.00000400, -0.00000490, -0.00000525, -0.00000517, -0.00000488, },
		{ -0.00000539, -0.00000554, -0.00000564, -0.00000575, -0.00000579, -0.00000565, -0.00000534, -0.00000505, -0.00000514, -0.00000596, -0.00000750, -0.00000934, -0.00001066, -0.00001065, -0.00000893, -0.00000588, -0.00000253, -0.00000022, 0.00000015, -0.00000147, -0.00000408, -0.00000613, -0.00000625, -0.00000405, -0.00000042, 0.00000282, 0.00000383, 0.00000171, -0.00000288, -0.00000788, -0.00001078, -0.00000983, -0.00000503, 0.00000181, 0.00000785, 0.00001062, 0.00000932, 0.00000519, 0.00000095, -0.00000069, 0.00000151, 0.00000650, 0.00001149, 0.00001340, 0.00001069, 0.00000445, -0.00000189, -0.00000422, 0.00000001, 0.00001010, 0.00002195, 0.00002987, 0.00002933, 0.00001970, 0.00000522, -0.00000635, -0.00000701, 0.00000756, 0.00003538, 0.00006805, 0.00009358, 0.00010105, 0.00008507, 0.00004822, 0.00000035, -0.00004508, -0.00007599, -0.00008595, -0.00007602, -0.00005352, -0.00002851, -0.00000960, -0.00000091, -0.00000143, -0.00000659, -0.00001115, -0.00001185, -0.00000864, -0.00000396, -0.00000095, -0.00000137, -0.00000457, -0.00000794, -0.00000846, -0.00000453, 0.00000298, 0.00001107, 0.00001603, 0.00001537, 0.00000911, -0.00000011, -0.00000831, -0.00001205, -0.00001010, -0.00000399, 0.00000275, 0.00000626, 0.00000420, -0.00000314, -0.00001307, -0.00002168, -0.00002566, -0.00002372, -0.00001702, -0.00000855, -0.00000171, 0.00000113, -0.00000043, -0.00000494, -0.00000991, -0.00001302, -0.00001312, -0.00001056, -0.00000680, -0.00000361, -0.00000222, -0.00000285, -0.00000479, -0.00000687, -0.00000810, -0.00000804, -0.00000697, -0.00000553, -0.00000441, -0.00000395, -0.00000411, -0.00000460, -0.00000508, },
		{ -0.00000151, -0.00000129, -0.00000097, -0.00000043, 0.00000019, 0.00000059, 0.00000058, 0.00000020, -0.00000021, -0.00000025, 0.00000026, 0.00000110, 0.00000169, 0.00000136, -0.00000016, -0.00000247, -0.00000459, -0.00000531, -0.00000390, -0.00000054, 0.00000358, 0.00000675, 0.00000756, 0.00000566, 0.00000205, -0.00000136, -0.00000265, -0.00000093, 0.00000306, 0.00000721, 0.00000905, 0.00000705, 0.00000154, -0.00000530, -0.00001036, -0.00001116, -0.00000717, -0.00000017, 0.00000649, 0.00000953, 0.00000746, 0.00000138, -0.00000553, -0.00000965, -0.00000892, -0.00000407, 0.00000157, 0.00000376, -0.00000027, -0.00001005, -0.00002170, -0.00002961, -0.00002935, -0.00002033, -0.00000686, 0.00000325, 0.00000207, -0.00001451, -0.00004409, -0.00007780, -0.00010321, -0.00010915, -0.00009032, -0.00004973, 0.00000207, 0.00005084, 0.00008386, 0.00009437, 0.00008350, 0.00005905, 0.00003178, 0.00001106, 0.00000158, 0.00000254, 0.00000920, 0.00001576, 0.00001826, 0.00001592, 0.00001084, 0.00000618, 0.00000421, 0.00000511, 0.00000717, 0.00000803, 0.00000626, 0.00000228, -0.00000177, -0.00000329, -0.00000071, 0.00000548, 0.00001280, 0.00001794, 0.00001838, 0.00001361, 0.00000549, -0.00000272, -0.00000784, -0.00000836, -0.00000507, -0.00000062, 0.00000176, -0.00000011, -0.00000632, -0.00001486, -0.00002250, -0.00002635, -0.00002505, -0.00001935, -0.00001166, -0.00000494, -0.00000140, -0.00000167, -0.00000474, -0.00000861, -0.00001128, -0.00001161, -0.00000970, -0.00000661, -0.00000373, -0.00000208, -0.00000192, -0.00000278, -0.00000383, -0.00000437, -0.00000416, -0.00000339, -0.00000250, -0.00000185, },
		{ -0.00000496, -0.00000479, -0.00000491, -0.00000541, -0.00000601, -0.00000626, -0.00000586, -0.00000486, -0.00000373, -0.00000310, -0.00000337, -0.00000449, -0.00000587, -0.00000675, -0.00000664, -0.00000569, -0.00000468, -0.00000466, -0.00000626, -0.00000928, -0.00001255, -0.00001446, -0.00001377, -0.00001034, -0.00000540, -0.00000108, 0.00000064, -0.00000109, -0.00000540, -0.00001002, -0.00001236, -0.00001090, -0.00000603, -0.00000009, 0.00000378, 0.00000325, -0.00000192, -0.00000950, -0.00001576, -0.00001728, -0.00001275, -0.00000377, 0.00000578, 0.00001162, 0.00001129, 0.00000549, -0.00000209, -0.00000664, -0.00000476, 0.00000349, 0.00001439, 0.00002218, 0.00002216, 0.00001345, 0.00000020, -0.00000981, -0.00000858, 0.00000806, 0.00003768, 0.00007127, 0.00009622, 0.00010123, 0.00008103, 0.00003891, -0.00001420, -0.00006359, -0.00009625, -0.00010534, -0.00009227, -0.00006530, -0.00003577, -0.00001352, -0.00000350, -0.00000481, -0.00001234, -0.00001976, -0.00002256, -0.00001964, -0.00001306, -0.00000630, -0.00000219, -0.00000155, -0.00000310, -0.00000460, -0.00000429, -0.00000195, 0.00000113, 0.00000304, 0.00000253, -0.00000012, -0.00000313, -0.00000415, -0.00000159, 0.00000445, 0.00001203, 0.00001824, 0.00002059, 0.00001814, 0.00001202, 0.00000476, -0.00000086, -0.00000320, -0.00000244, -0.00000044, 0.00000033, -0.00000198, -0.00000759, -0.00001498, -0.00002159, -0.00002498, -0.00002394, -0.00001903, -0.00001223, -0.00000606, -0.00000247, -0.00000212, -0.00000430, -0.00000738, -0.00000972, -0.00001033, -0.00000922, -0.00000718, -0.00000523, -0.00000413, -0.00000399, -0.00000445, -0.00000495, -0.00000512, },
		{ -0.00000803, -0.00000794, -0.00000750, -0.00000666, -0.00000579, -0.00000542, -0.00000591, -0.00000711, -0.00000837, -0.00000888, -0.00000809, -0.00000614, -0.00000380, -0.00000214, -0.00000193, -0.00000317, -0.00000507, -0.00000646, -0.00000643, -0.00000493, -0.00000290, -0.00000181, -0.00000282, -0.00000603, -0.00001020, -0.00001330, -0.00001345, -0.00001002, -0.00000417, 0.00000161, 0.00000459, 0.00000324, -0.00000187, -0.00000820, -0.00001243, -0.00001215, -0.00000725, -0.00000014, 0.00000531, 0.00000570, -0.00000001, -0.00000976, -0.00001919, -0.00002370, -0.00002088, -0.00001189, -0.00000116, 0.00000572, 0.00000479, -0.00000407, -0.00001679, -0.00002698, -0.00002922, -0.00002211, -0.00000959, 0.00000034, -0.00000074, -0.00001744, -0.00004760, -0.00008215, -0.00010820, -0.00011403, -0.00009402, -0.00005133, 0.00000299, 0.00005395, 0.00008814, 0.00009851, 0.00008636, 0.00006010, 0.00003127, 0.00000991, 0.00000106, 0.00000371, 0.00001249, 0.00002079, 0.00002388, 0.00002064, 0.00001330, 0.00000572, 0.00000115, 0.00000073, 0.00000333, 0.00000652, 0.00000814, 0.00000741, 0.00000517, 0.00000311, 0.00000265, 0.00000401, 0.00000612, 0.00000734, 0.00000647, 0.00000361, 0.00000020, -0.00000164, -0.00000030, 0.00000446, 0.00001122, 0.00001750, 0.00002090, 0.00002016, 0.00001570, 0.00000934, 0.00000331, -0.00000079, -0.00000277, -0.00000371, -0.00000536, -0.00000901, -0.00001481, -0.00002151, -0.00002709, -0.00002964, -0.00002828, -0.00002359, -0.00001731, -0.00001160, -0.00000809, -0.00000733, -0.00000867, -0.00001077, -0.00001231, -0.00001256, -0.00001162, -0.00001013, -0.00000885, -0.00000819, },
		{ -0.00001155, -0.00001058, -0.00001019, -0.00001044, -0.00001087, -0.00001078, -0.00000978, -0.00000807, -0.00000642, -0.00000579, -0.00000670, -0.00000890, -0.00001136, -0.00001274, -0.00001216, -0.00000967, -0.00000632, -0.00000368, -0.00000300, -0.00000454, -0.00000734, -0.00000975, -0.00001030, -0.00000859, -0.00000563, -0.00000333, -0.00000350, -0.00000673, -0.00001187, -0.00001642, -0.00001775, -0.00001453, -0.00000760, 0.00000020, 0.00000532, 0.00000533, 0.00000023, -0.00000726, -0.00001301, -0.00001350, -0.00000774, 0.00000195, 0.00001096, 0.00001458, 0.00001057, 0.00000058, -0.00001040, -0.00001627, -0.00001295, -0.00000069, 0.00001571, 0.00002893, 0.00003276, 0.00002550, 0.00001135, -0.00000097, -0.00000209, 0.00001338, 0.00004362, 0.00007935, 0.00010700, 0.00011400, 0.00009403, 0.00004991, -0.00000710, -0.00006144, -0.00009893, -0.00011180, -0.00010092, -0.00007460, -0.00004462, -0.00002142, -0.00001041, -0.00001090, -0.00001765, -0.00002408, -0.00002544, -0.00002061, -0.00001192, -0.00000335, 0.00000172, 0.00000212, -0.00000089, -0.00000464, -0.00000661, -0.00000573, -0.00000266, 0.00000084, 0.00000310, 0.00000343, 0.00000247, 0.00000160, 0.00000203, 0.00000403, 0.00000672, 0.00000868, 0.00000872, 0.00000678, 0.00000399, 0.00000220, 0.00000307, 0.00000711, 0.00001343, 0.00002001, 0.00002464, 0.00002586, 0.00002351, 0.00001866, 0.00001292, 0.00000762, 0.00000319, -0.00000086, -0.00000550, -0.00001135, -0.00001820, -0.00002487, -0.00002967, -0.00003123, -0.00002915, -0.00002424, -0.00001825, -0.00001308, -0.00001006, -0.00000944, -0.00001047, -0.00001192, -0.00001273, -0.00001250, },
		{ -0.00001673, -0.00001732, -0.00001690, -0.00001542, -0.00001359, -0.00001237, -0.00001233, -0.00001324, -0.00001418, -0.00001405, -0.00001227, -0.00000919, -0.00000605, -0.00000434, -0.00000497, -0.00000772, -0.00001118, -0.00001345, -0.00001307, -0.00000986, -0.00000510, -0.00000090, 0.00000082, -0.00000060, -0.00000416, -0.00000771, -0.00000908, -0.00000732, -0.00000335, 0.00000044, 0.00000146, -0.00000163, -0.00000794, -0.00001460, -0.00001809, -0.00001609, -0.00000894, 0.00000033, 0.00000726, 0.00000819, 0.00000226, -0.00000791, -0.00001732, -0.00002104, -0.00001682, -0.00000655, 0.00000436, 0.00000955, 0.00000497, -0.00000872, -0.00002602, -0.00003894, -0.00004090, -0.00003037, -0.00001231, 0.00000344, 0.00000643, -0.00000938, -0.00004213, -0.00008176, -0.00011347, -0.00012337, -0.00010423, -0.00005865, 0.00000162, 0.00005995, 0.00010094, 0.00011582, 0.00010497, 0.00007684, 0.00004382, 0.00001721, 0.00000325, 0.00000183, 0.00000792, 0.00001483, 0.00001747, 0.00001436, 0.00000748, 0.00000062, -0.00000295, -0.00000212, 0.00000183, 0.00000616, 0.00000831, 0.00000719, 0.00000355, -0.00000069, -0.00000356, -0.00000414, -0.00000283, -0.00000102, -0.00000008, -0.00000062, -0.00000216, -0.00000354, -0.00000369, -0.00000229, -0.00000003, 0.00000179, 0.00000206, 0.00000048, -0.00000211, -0.00000411, -0.00000393, -0.00000081, 0.00000481, 0.00001141, 0.00001708, 0.00002025, 0.00002030, 0.00001756, 0.00001293, 0.00000730, 0.00000107, -0.00000582, -0.00001359, -0.00002210, -0.00003054, -0.00003748, -0.00004141, -0.00004141, -0.00003765, -0.00003144, -0.00002475, -0.00001946, -0.00001660, -0.00001604, },
		{ -0.00002092, -0.00001564, -0.00001348, -0.00001387, -0.00001522, -0.00001584, -0.00001482, -0.00001241, -0.00000983, -0.00000851, -0.00000920, -0.00001149, -0.00001392, -0.00001480, -0.00001314, -0.00000931, -0.00000499, -0.00000234, -0.00000288, -0.00000658, -0.00001178, -0.00001588, -0.00001667, -0.00001351, -0.00000776, -0.00000214, 0.00000061, -0.00000078, -0.00000545, -0.00001074, -0.00001360, -0.00001229, -0.00000742, -0.00000178, 0.00000108, -0.00000117, -0.00000818, -0.00001683, -0.00002256, -0.00002180, -0.00001392, -0.00000197, 0.00000864, 0.00001272, 0.00000805, -0.00000314, -0.00001496, -0.00002063, -0.00001599, -0.00000192, 0.00001558, 0.00002803, 0.00002866, 0.00001623, -0.00000355, -0.00001990, -0.00002161, -0.00000243, 0.00003519, 0.00007999, 0.00011563, 0.00012693, 0.00010601, 0.00005575, -0.00001098, -0.00007596, -0.00012223, -0.00013997, -0.00012931, -0.00009920, -0.00006314, -0.00003364, -0.00001793, -0.00001635, -0.00002376, -0.00003289, -0.00003792, -0.00003663, -0.00003057, -0.00002338, -0.00001849, -0.00001736, -0.00001907, -0.00002119, -0.00002136, -0.00001859, -0.00001363, -0.00000845, -0.00000500, -0.00000420, -0.00000556, -0.00000759, -0.00000870, -0.00000806, -0.00000594, -0.00000343, -0.00000170, -0.00000128, -0.00000181, -0.00000231, -0.00000177, 0.00000017, 0.00000301, 0.00000562, 0.00000691, 0.00000644, 0.00000477, 0.00000320, 0.00000323, 0.00000580, 0.00001087, 0.00001746, 0.00002401, 0.00002904, 0.00003156, 0.00003126, 0.00002826, 0.00002280, 0.00001505, 0.00000515, -0.00000647, -0.00001882, -0.00003024, -0.00003878, -0.00004279, -0.00004168, -0.00003624, -0.00002848, },
		{ -0.00005446, -0.00005254, -0.00004513, -0.00003514, -0.00002594, -0.00002014, -0.00001857, -0.00002018, -0.00002270, -0.00002379, -0.00002227, -0.00001864, -0.00001473, -0.00001266, -0.00001363, -0.00001717, -0.00002130, -0.00002353, -0.00002223, -0.00001757, -0.00001160, -0.00000729, -0.00000701, -0.00001120, -0.00001803, -0.00002417, -0.00002646, -0.00002354, -0.00001666, -0.00000914, -0.00000472, -0.00000563, -0.00001132, -0.00001872, -0.00002374, -0.00002354, -0.00001815, -0.00001064, -0.00000564, -0.00000686, -0.00001491, -0.00002661, -0.00003635, -0.00003886, -0.00003212, -0.00001880, -0.00000532, 0.00000110, -0.00000384, -0.00001893, -0.00003760, -0.00005073, -0.00005121, -0.00003792, -0.00001715, -0.00000046, 0.00000032, -0.00002112, -0.00006172, -0.00010907, -0.00014555, -0.00015504, -0.00012950, -0.00007255, 0.00000150, 0.00007269, 0.00012265, 0.00014090, 0.00012782, 0.00009347, 0.00005280, 0.00001970, 0.00000227, 0.00000097, 0.00001013, 0.00002157, 0.00002847, 0.00002796, 0.00002140, 0.00001271, 0.00000590, 0.00000311, 0.00000398, 0.00000634, 0.00000781, 0.00000713, 0.00000469, 0.00000209, 0.00000100, 0.00000220, 0.00000515, 0.00000837, 0.00001030, 0.00001014, 0.00000824, 0.00000577, 0.00000405, 0.00000381, 0.00000487, 0.00000635, 0.00000724, 0.00000703, 0.00000598, 0.00000494, 0.00000483, 0.00000607, 0.00000832, 0.00001062, 0.00001192, 0.00001169, 0.00001019, 0.00000845, 0.00000778, 0.00000921, 0.00001308, 0.00001889, 0.00002555, 0.00003179, 0.00003644, 0.00003858, 0.00003754, 0.00003274, 0.00002384, 0.00001093, -0.00000508, -0.00002231, -0.00003805, -0.00004946, },
		{ -0.00000362, -0.00002612, -0.00004702, -0.00006223, -0.00006879, -0.00006604, -0.00005591, -0.00004240, -0.00003008, -0.00002244, -0.00002062, -0.00002314, -0.00002682, -0.00002841, -0.00002614, -0.00002058, -0.00001422, -0.00001013, -0.00001026, -0.00001431, -0.00001981, -0.00002334, -0.00002228, -0.00001634, -0.00000780, -0.00000054, 0.00000197, -0.00000151, -0.00000918, -0.00001699, -0.00002057, -0.00001750, -0.00000865, 0.00000211, 0.00000987, 0.00001112, 0.00000574, -0.00000289, -0.00000951, -0.00000977, -0.00000267, 0.00000868, 0.00001847, 0.00002118, 0.00001473, 0.00000194, -0.00001042, -0.00001489, -0.00000721, 0.00001104, 0.00003258, 0.00004761, 0.00004865, 0.00003475, 0.00001289, -0.00000443, -0.00000462, 0.00001888, 0.00006257, 0.00011303, 0.00015144, 0.00016082, 0.00013298, 0.00007221, -0.00000589, -0.00007994, -0.00013053, -0.00014690, -0.00013006, -0.00009126, -0.00004687, -0.00001187, 0.00000527, 0.00000442, -0.00000780, -0.00002207, -0.00003067, -0.00003034, -0.00002260, -0.00001204, -0.00000351, 0.00000017, -0.00000097, -0.00000469, -0.00000813, -0.00000942, -0.00000837, -0.00000619, -0.00000449, -0.00000420, -0.00000514, -0.00000619, -0.00000612, -0.00000433, -0.00000122, 0.00000202, 0.00000415, 0.00000449, 0.00000328, 0.00000148, 0.00000021, 0.00000013, 0.00000115, 0.00000253, 0.00000342, 0.00000332, 0.00000243, 0.00000147, 0.00000128, 0.00000234, 0.00000444, 0.00000682, 0.00000852, 0.00000891, 0.00000805, 0.00000665, 0.00000580, 0.00000655, 0.00000950, 0.00001464, 0.00002145, 0.00002898, 0.00003607, 0.00004136, 0.00004334, 0.00004049, 0.00003165, 0.00001656, },
		{ 0.00006194, 0.00006237, 0.00005490, 0.00003853, 0.00001438, -0.00001402, -0.00004138, -0.00006197, -0.00007152, -0.00006873, -0.00005593, -0.00003827, -0.00002192, -0.00001171, -0.00000943, -0.00001336, -0.00001928, -0.00002264, -0.00002073, -0.00001392, -0.00000534, 0.00000076, 0.00000131, -0.00000391, -0.00001205, -0.00001861, -0.00001967, -0.00001408, -0.00000416, 0.00000528, 0.00000937, 0.00000568, -0.00000438, -0.00001604, -0.00002352, -0.00002295, -0.00001444, -0.00000221, 0.00000743, 0.00000927, 0.00000200, -0.00001094, -0.00002299, -0.00002785, -0.00002286, -0.00001079, 0.00000130, 0.00000549, -0.00000271, -0.00002156, -0.00004333, -0.00005768, -0.00005688, -0.00004026, -0.00001559, 0.00000357, 0.00000379, -0.00002176, -0.00006904, -0.00012343, -0.00016470, -0.00017476, -0.00014503, -0.00008029, 0.00000267, 0.00008109, 0.00013448, 0.00015168, 0.00013407, 0.00009383, 0.00004836, 0.00001334, -0.00000264, 0.00000002, 0.00001384, 0.00002863, 0.00003621, 0.00003340, 0.00002233, 0.00000848, -0.00000247, -0.00000721, -0.00000579, -0.00000093, 0.00000383, 0.00000589, 0.00000466, 0.00000129, -0.00000230, -0.00000462, -0.00000528, -0.00000497, -0.00000475, -0.00000530, -0.00000650, -0.00000757, -0.00000758, -0.00000609, -0.00000347, -0.00000066, 0.00000128, 0.00000177, 0.00000096, -0.00000035, -0.00000122, -0.00000106, 0.00000009, 0.00000162, 0.00000279, 0.00000314, 0.00000279, 0.00000233, 0.00000254, 0.00000390, 0.00000632, 0.00000922, 0.00001177, 0.00001331, 0.00001369, 0.00001332, 0.00001302, 0.00001373, 0.00001623, 0.00002098, 0.00002801, 0.00003693, 0.00004676, 0.00005586, },
		{ 0.00002519, 0.00003397, 0.00004525, 0.00005732, 0.00006722, 0.00007111, 0.00006533, 0.00004773, 0.00001900, -0.00001681, -0.00005289, -0.00008153, -0.00009654, -0.00009551, -0.00008078, -0.00005875, -0.00003748, -0.00002373, -0.00002047, -0.00002594, -0.00003485, -0.00004098, -0.00004016, -0.00003217, -0.00002070, -0.00001136, -0.00000877, -0.00001411, -0.00002440, -0.00003393, -0.00003712, -0.00003148, -0.00001902, -0.00000541, 0.00000276, 0.00000137, -0.00000895, -0.00002293, -0.00003312, -0.00003366, -0.00002334, -0.00000646, 0.00000916, 0.00001607, 0.00001111, -0.00000270, -0.00001739, -0.00002399, -0.00001724, 0.00000129, 0.00002340, 0.00003807, 0.00003703, 0.00001960, -0.00000584, -0.00002481, -0.00002305, 0.00000645, 0.00005903, 0.00011871, 0.00016366, 0.00017453, 0.00014231, 0.00007238, -0.00001708, -0.00010160, -0.00015926, -0.00017818, -0.00015990, -0.00011739, -0.00006909, -0.00003152, -0.00001363, -0.00001486, -0.00002728, -0.00004026, -0.00004548, -0.00004003, -0.00002664, -0.00001143, -0.00000050, 0.00000284, -0.00000090, -0.00000825, -0.00001489, -0.00001771, -0.00001598, -0.00001117, -0.00000581, -0.00000203, -0.00000070, -0.00000130, -0.00000258, -0.00000338, -0.00000328, -0.00000262, -0.00000206, -0.00000206, -0.00000250, -0.00000277, -0.00000217, -0.00000038, 0.00000231, 0.00000509, 0.00000708, 0.00000778, 0.00000730, 0.00000630, 0.00000556, 0.00000562, 0.00000647, 0.00000764, 0.00000848, 0.00000859, 0.00000803, 0.00000730, 0.00000708, 0.00000783, 0.00000960, 0.00001197, 0.00001426, 0.00001587, 0.00001649, 0.00001625, 0.00001564, 0.00001540, 0.00001638, 0.00001941, },
		{ 0.00001768, 0.00001766, 0.00001812, 0.00002017, 0.00002517, 0.00003421, 0.00004748, 0.00006353, 0.00007889, 0.00008845, 0.00008683, 0.00007022, 0.00003837, -0.00000444, -0.00004995, -0.00008806, -0.00011004, -0.00011169, -0.00009492, -0.00006724, -0.00003901, -0.00001954, -0.00001361, -0.00001992, -0.00003211, -0.00004194, -0.00004320, -0.00003455, -0.00001996, -0.00000661, -0.00000117, -0.00000635, -0.00001944, -0.00003352, -0.00004099, -0.00003740, -0.00002392, -0.00000698, 0.00000483, 0.00000506, -0.00000711, -0.00002624, -0.00004304, -0.00004901, -0.00004084, -0.00002240, -0.00000317, 0.00000638, 0.00000012, -0.00002023, -0.00004546, -0.00006310, -0.00006373, -0.00004632, -0.00001979, 0.00000021, -0.00000174, -0.00003321, -0.00008908, -0.00015210, -0.00019876, -0.00020835, -0.00017142, -0.00009402, 0.00000406, 0.00009641, 0.00015944, 0.00018047, 0.00016121, 0.00011566, 0.00006363, 0.00002283, 0.00000293, 0.00000347, 0.00001616, 0.00002991, 0.00003604, 0.00003159, 0.00001945, 0.00000593, -0.00000284, -0.00000379, 0.00000207, 0.00001072, 0.00001742, 0.00001902, 0.00001512, 0.00000782, 0.00000036, -0.00000456, -0.00000588, -0.00000430, -0.00000155, 0.00000061, 0.00000125, 0.00000048, -0.00000096, -0.00000224, -0.00000300, -0.00000336, -0.00000375, -0.00000444, -0.00000529, -0.00000583, -0.00000551, -0.00000411, -0.00000189, 0.00000046, 0.00000221, 0.00000292, 0.00000266, 0.00000196, 0.00000148, 0.00000167, 0.00000253, 0.00000369, 0.00000461, 0.00000494, 0.00000471, 0.00000433, 0.00000435, 0.00000522, 0.00000703, 0.00000955, 0.00001228, 0.00001471, 0.00001646, 0.00001741, },
		{ 0.00001357, 0.00001599, 0.00001817, 0.00001976, 0.00002046, 0.00002025, 0.00001962, 0.00001981, 0.00002279, 0.00003067, 0.00004464, 0.00006385, 0.00008459, 0.00010053, 0.00010427, 0.00008981, 0.00005531, 0.00000477, -0.00005218, -0.00010272, -0.00013490, -0.00014182, -0.00012429, -0.00009071, -0.00005397, -0.00002664, -0.00001619, -0.00002240, -0.00003803, -0.00005247, -0.00005674, -0.00004763, -0.00002897, -0.00000956, 0.00000135, -0.00000124, -0.00001544, -0.00003332, -0.00004475, -0.00004257, -0.00002634, -0.00000288, 0.00001694, 0.00002360, 0.00001386, -0.00000718, -0.00002824, -0.00003726, -0.00002754, -0.00000151, 0.00002987, 0.00005213, 0.00005443, 0.00003557, 0.00000564, -0.00001772, -0.00001717, 0.00001590, 0.00007608, 0.00014455, 0.00019562, 0.00020670, 0.00016748, 0.00008453, -0.00002067, -0.00011951, -0.00018652, -0.00020807, -0.00018608, -0.00013572, -0.00007839, -0.00003325, -0.00001077, -0.00001053, -0.00002374, -0.00003849, -0.00004545, -0.00004129, -0.00002885, -0.00001462, -0.00000487, -0.00000269, -0.00000701, -0.00001368, -0.00001799, -0.00001702, -0.00001079, -0.00000192, 0.00000597, 0.00001007, 0.00000957, 0.00000569, 0.00000085, -0.00000266, -0.00000362, -0.00000220, 0.00000046, 0.00000302, 0.00000459, 0.00000503, 0.00000476, 0.00000431, 0.00000400, 0.00000377, 0.00000342, 0.00000281, 0.00000211, 0.00000171, 0.00000204, 0.00000327, 0.00000515, 0.00000712, 0.00000857, 0.00000913, 0.00000887, 0.00000821, 0.00000769, 0.00000767, 0.00000816, 0.00000885, 0.00000933, 0.00000932, 0.00000885, 0.00000823, 0.00000791, 0.00000825, 0.00000941, 0.00001127, },
		{ 0.00000233, 0.00000340, 0.00000507, 0.00000737, 0.00001027, 0.00001371, 0.00001747, 0.00002105, 0.00002374, 0.00002493, 0.00002463, 0.00002401, 0.00002555, 0.00003247, 0.00004731, 0.00007024, 0.00009770, 0.00012225, 0.00013424, 0.00012486, 0.00008989, 0.00003237, -0.00003703, -0.00010251, -0.00014813, -0.00016326, -0.00014667, -0.00010713, -0.00006030, -0.00002278, -0.00000582, -0.00001123, -0.00003130, -0.00005277, -0.00006310, -0.00005623, -0.00003515, -0.00001021, 0.00000615, 0.00000566, -0.00001158, -0.00003694, -0.00005741, -0.00006213, -0.00004796, -0.00002147, 0.00000394, 0.00001452, 0.00000314, -0.00002664, -0.00006165, -0.00008514, -0.00008499, -0.00006036, -0.00002330, 0.00000578, 0.00000708, -0.00002903, -0.00009629, -0.00017340, -0.00023149, -0.00024525, -0.00020311, -0.00011229, 0.00000357, 0.00011294, 0.00018779, 0.00021306, 0.00019073, 0.00013724, 0.00007585, 0.00002740, 0.00000334, 0.00000334, 0.00001794, 0.00003432, 0.00004233, 0.00003824, 0.00002498, 0.00000939, -0.00000179, -0.00000512, -0.00000156, 0.00000477, 0.00000910, 0.00000851, 0.00000311, -0.00000444, -0.00001054, -0.00001263, -0.00001026, -0.00000515, -0.00000008, 0.00000249, 0.00000153, -0.00000235, -0.00000743, -0.00001179, -0.00001424, -0.00001461, -0.00001359, -0.00001218, -0.00001116, -0.00001081, -0.00001098, -0.00001136, -0.00001174, -0.00001211, -0.00001258, -0.00001318, -0.00001371, -0.00001385, -0.00001327, -0.00001187, -0.00000989, -0.00000779, -0.00000607, -0.00000502, -0.00000456, -0.00000436, -0.00000401, -0.00000325, -0.00000211, -0.00000084, 0.00000026, 0.00000098, 0.00000140, 0.00000174, },
		{ 0.00001485, 0.00001515, 0.00001527, 0.00001539, 0.00001570, 0.00001625, 0.00001707, 0.00001827, 0.00002010, 0.00002285, 0.00002658, 0.00003079, 0.00003438, 0.00003603, 0.00003503, 0.00003220, 0.00003040, 0.00003398, 0.00004720, 0.00007178, 0.00010484, 0.00013812, 0.00015954, 0.00015689, 0.00012262, 0.00005787, -0.00002615, -0.00011058, -0.00017460, -0.00020261, -0.00019000, -0.00014513, -0.00008638, -0.00003521, -0.00000798, -0.00000983, -0.00003331, -0.00006239, -0.00007997, -0.00007578, -0.00005086, -0.00001676, 0.00001029, 0.00001745, 0.00000174, -0.00002820, -0.00005610, -0.00006597, -0.00005015, -0.00001364, 0.00002772, 0.00005450, 0.00005331, 0.00002406, -0.00001862, -0.00005096, -0.00005055, -0.00000701, 0.00007174, 0.00016090, 0.00022752, 0.00024311, 0.00019505, 0.00009192, -0.00003933, -0.00016318, -0.00024817, -0.00027748, -0.00025330, -0.00019389, -0.00012506, -0.00006999, -0.00004153, -0.00003956, -0.00005393, -0.00007067, -0.00007867, -0.00007372, -0.00005883, -0.00004131, -0.00002847, -0.00002411, -0.00002738, -0.00003396, -0.00003878, -0.00003864, -0.00003344, -0.00002579, -0.00001921, -0.00001623, -0.00001722, -0.00002042, -0.00002308, -0.00002296, -0.00001937, -0.00001338, -0.00000708, -0.00000256, -0.00000093, -0.00000199, -0.00000453, -0.00000701, -0.00000830, -0.00000802, -0.00000651, -0.00000448, -0.00000255, -0.00000107, -0.00000005, 0.00000064, 0.00000111, 0.00000139, 0.00000145, 0.00000137, 0.00000134, 0.00000165, 0.00000255, 0.00000406, 0.00000597, 0.00000788, 0.00000941, 0.00001039, 0.00001092, 0.00001127, 0.00001175, 0.00001248, 0.00001338, 0.00001424, },
		{ 0.00000694, 0.00000749, 0.00000808, 0.00000850, 0.00000863, 0.00000858, 0.00000858, 0.00000883, 0.00000932, 0.00000990, 0.00001039, 0.00001091, 0.00001196, 0.00001429, 0.00001841, 0.00002399, 0.00002961, 0.00003314, 0.00003288, 0.00002893, 0.00002421, 0.00002424, 0.00003525, 0.00006113, 0.00010035, 0.00014429, 0.00017840, 0.00018638, 0.00015624, 0.00008599, -0.00001353, -0.00012047, -0.00020837, -0.00025507, -0.00025089, -0.00020250, -0.00013073, -0.00006278, -0.00002180, -0.00001815, -0.00004594, -0.00008645, -0.00011702, -0.00012134, -0.00009683, -0.00005552, -0.00001809, -0.00000382, -0.00002100, -0.00006253, -0.00010880, -0.00013684, -0.00013158, -0.00009398, -0.00004199, -0.00000344, -0.00000382, -0.00005405, -0.00014387, -0.00024408, -0.00031688, -0.00033038, -0.00027124, -0.00015030, 0.00000113, 0.00014255, 0.00023869, 0.00027113, 0.00024296, 0.00017522, 0.00009723, 0.00003517, 0.00000346, 0.00000205, 0.00001958, 0.00004043, 0.00005206, 0.00004950, 0.00003582, 0.00001886, 0.00000662, 0.00000337, 0.00000833, 0.00001695, 0.00002376, 0.00002521, 0.00002105, 0.00001393, 0.00000755, 0.00000469, 0.00000588, 0.00000948, 0.00001285, 0.00001382, 0.00001179, 0.00000787, 0.00000413, 0.00000246, 0.00000364, 0.00000706, 0.00001113, 0.00001413, 0.00001494, 0.00001348, 0.00001054, 0.00000732, 0.00000483, 0.00000356, 0.00000341, 0.00000396, 0.00000469, 0.00000528, 0.00000557, 0.00000558, 0.00000532, 0.00000480, 0.00000407, 0.00000325, 0.00000258, 0.00000229, 0.00000256, 0.00000334, 0.00000439, 0.00000538, 0.00000606, 0.00000637, 0.00000647, 0.00000659, },
		{ 0.00000025, 0.00000087, 0.00000152, 0.00000244, 0.00000370, 0.00000515, 0.00000647, 0.00000741, 0.00000796, 0.00000843, 0.00000923, 0.00001059, 0.00001238, 0.00001412, 0.00001534, 0.00001602, 0.00001690, 0.00001930, 0.00002437, 0.00003218, 0.00004117, 0.00004837, 0.00005077, 0.00004735, 0.00004072, 0.00003738, 0.00004583, 0.00007285, 0.00011911, 0.00017633, 0.00022768, 0.00025216, 0.00023202, 0.00016052, 0.00004684, -0.00008458, -0.00020080, -0.00027125, -0.00027879, -0.00022642, -0.00013663, -0.00004309, 0.00002210, 0.00004077, 0.00001519, -0.00003381, -0.00007663, -0.00008831, -0.00005977, -0.00000175, 0.00006024, 0.00009798, 0.00009420, 0.00005155, -0.00000725, -0.00004872, -0.00004340, 0.00002024, 0.00012881, 0.00024739, 0.00033219, 0.00034728, 0.00027876, 0.00014052, -0.00003069, -0.00018851, -0.00029339, -0.00032545, -0.00028931, -0.00020961, -0.00011981, -0.00004930, -0.00001394, -0.00001313, -0.00003367, -0.00005770, -0.00007093, -0.00006772, -0.00005141, -0.00003085, -0.00001510, -0.00000918, -0.00001250, -0.00002025, -0.00002647, -0.00002719, -0.00002199, -0.00001360, -0.00000600, -0.00000223, -0.00000300, -0.00000668, -0.00001053, -0.00001219, -0.00001090, -0.00000761, -0.00000428, -0.00000267, -0.00000346, -0.00000599, -0.00000869, -0.00000998, -0.00000905, -0.00000620, -0.00000256, 0.00000045, 0.00000179, 0.00000122, -0.00000079, -0.00000335, -0.00000558, -0.00000695, -0.00000736, -0.00000704, -0.00000631, -0.00000551, -0.00000483, -0.00000439, -0.00000426, -0.00000442, -0.00000480, -0.00000519, -0.00000536, -0.00000508, -0.00000427, -0.00000307, -0.00000175, -0.00000060, },
		{ 0.00001308, 0.00001510, 0.00001702, 0.00001855, 0.00001968, 0.00002063, 0.00002180, 0.00002348, 0.00002564, 0.00002793, 0.00002983, 0.00003102, 0.00003158, 0.00003204, 0.00003306, 0.00003503, 0.00003768, 0.00004015, 0.00004150, 0.00004144, 0.00004083, 0.00004160, 0.00004576, 0.00005413, 0.00006522, 0.00007522, 0.00007958, 0.00007560, 0.00006499, 0.00005494, 0.00005642, 0.00007978, 0.00012902, 0.00019718, 0.00026554, 0.00030782, 0.00029893, 0.00022523, 0.00009226, -0.00007390, -0.00023316, -0.00034425, -0.00037935, -0.00033478, -0.00023308, -0.00011510, -0.00002467, 0.00000814, -0.00002170, -0.00009292, -0.00016783, -0.00020900, -0.00019598, -0.00013473, -0.00005567, -0.00000072, -0.00000492, -0.00008017, -0.00020824, -0.00034610, -0.00044174, -0.00045399, -0.00036842, -0.00020311, -0.00000193, 0.00018130, 0.00030174, 0.00033785, 0.00029623, 0.00020587, 0.00010511, 0.00002701, -0.00001114, -0.00001070, 0.00001293, 0.00003950, 0.00005308, 0.00004767, 0.00002743, 0.00000266, -0.00001625, -0.00002351, -0.00001969, -0.00001024, -0.00000193, 0.00000056, -0.00000350, -0.00001130, -0.00001855, -0.00002189, -0.00002043, -0.00001579, -0.00001091, -0.00000834, -0.00000902, -0.00001206, -0.00001546, -0.00001732, -0.00001680, -0.00001444, -0.00001166, -0.00000991, -0.00000989, -0.00001125, -0.00001285, -0.00001342, -0.00001220, -0.00000926, -0.00000543, -0.00000183, 0.00000061, 0.00000152, 0.00000113, 0.00000004, -0.00000102, -0.00000155, -0.00000131, -0.00000034, 0.00000114, 0.00000288, 0.00000460, 0.00000612, 0.00000728, 0.00000807, 0.00000861, 0.00000915, 0.00000999, 0.00001130, },
		{ 0.00000066, 0.00000021, 0.00000016, 0.00000070, 0.00000172, 0.00000288, 0.00000372, 0.00000395, 0.00000365, 0.00000325, 0.00000333, 0.00000427, 0.00000597, 0.00000788, 0.00000923, 0.00000953, 0.00000893, 0.00000825, 0.00000854, 0.00001044, 0.00001363, 0.00001685, 0.00001853, 0.00001787, 0.00001573, 0.00001465, 0.00001781, 0.00002713, 0.00004160, 0.00005676, 0.00006628, 0.00006519, 0.00005362, 0.00003894, 0.00003465, 0.00005553, 0.00011028, 0.00019476, 0.00028912, 0.00036141, 0.00037733, 0.00031348, 0.00016885, -0.00003058, -0.00023713, -0.00039588, -0.00046353, -0.00042467, -0.00029849, -0.00013284, 0.00001316, 0.00009128, 0.00008267, 0.00000510, -0.00009438, -0.00015866, -0.00014422, -0.00003969, 0.00012817, 0.00030367, 0.00042284, 0.00043756, 0.00033354, 0.00013544, -0.00010248, -0.00031607, -0.00045337, -0.00049080, -0.00043767, -0.00032880, -0.00020888, -0.00011572, -0.00006856, -0.00006526, -0.00008801, -0.00011383, -0.00012504, -0.00011530, -0.00008970, -0.00005996, -0.00003772, -0.00002920, -0.00003339, -0.00004384, -0.00005266, -0.00005444, -0.00004826, -0.00003725, -0.00002634, -0.00001947, -0.00001790, -0.00002004, -0.00002277, -0.00002327, -0.00002044, -0.00001513, -0.00000947, -0.00000551, -0.00000424, -0.00000519, -0.00000693, -0.00000791, -0.00000729, -0.00000528, -0.00000289, -0.00000125, -0.00000100, -0.00000199, -0.00000341, -0.00000424, -0.00000379, -0.00000198, 0.00000065, 0.00000325, 0.00000502, 0.00000555, 0.00000489, 0.00000342, 0.00000168, 0.00000018, -0.00000078, -0.00000109, -0.00000082, -0.00000016, 0.00000060, 0.00000118, 0.00000139, 0.00000116, },
		{ -0.00000232, -0.00000085, 0.00000020, 0.00000076, 0.00000102, 0.00000140, 0.00000231, 0.00000396, 0.00000618, 0.00000846, 0.00001018, 0.00001099, 0.00001106, 0.00001101, 0.00001169, 0.00001362, 0.00001668, 0.00002006, 0.00002263, 0.00002364, 0.00002324, 0.00002256, 0.00002314, 0.00002604, 0.00003100, 0.00003630, 0.00003954, 0.00003911, 0.00003555, 0.00003197, 0.00003286, 0.00004175, 0.00005858, 0.00007852, 0.00009326, 0.00009488, 0.00008091, 0.00005809, 0.00004229, 0.00005353, 0.00010693, 0.00020313, 0.00032253, 0.00042718, 0.00047116, 0.00041706, 0.00025249, -0.00000026, -0.00028643, -0.00053392, -0.00067691, -0.00067959, -0.00055094, -0.00034383, -0.00013793, -0.00001187, -0.00001480, -0.00014800, -0.00036384, -0.00058257, -0.00072078, -0.00072097, -0.00057123, -0.00030806, -0.00000209, 0.00026729, 0.00043856, 0.00048608, 0.00042398, 0.00029581, 0.00015549, 0.00004749, -0.00000616, -0.00000854, 0.00001934, 0.00005096, 0.00006617, 0.00005793, 0.00003213, 0.00000195, -0.00001974, -0.00002629, -0.00001919, -0.00000588, 0.00000480, 0.00000696, -0.00000014, -0.00001265, -0.00002478, -0.00003190, -0.00003251, -0.00002836, -0.00002302, -0.00001981, -0.00002018, -0.00002338, -0.00002721, -0.00002943, -0.00002892, -0.00002609, -0.00002247, -0.00001974, -0.00001886, -0.00001970, -0.00002126, -0.00002232, -0.00002213, -0.00002071, -0.00001873, -0.00001708, -0.00001635, -0.00001652, -0.00001707, -0.00001724, -0.00001642, -0.00001448, -0.00001174, -0.00000882, -0.00000638, -0.00000484, -0.00000429, -0.00000455, -0.00000525, -0.00000596, -0.00000633, -0.00000612, -0.00000527, -0.00000391, },
		{ -0.00001015, -0.00001105, -0.00001122, -0.00001069, -0.00000975, -0.00000883, -0.00000831, -0.00000830, -0.00000858, -0.00000864, -0.00000793, -0.00000619, -0.00000367, -0.00000106, 0.00000082, 0.00000149, 0.00000116, 0.00000071, 0.00000131, 0.00000376, 0.00000795, 0.00001281, 0.00001677, 0.00001871, 0.00001867, 0.00001813, 0.00001929, 0.00002389, 0.00003194, 0.00004132, 0.00004865, 0.00005116, 0.00004877, 0.00004496, 0.00004580, 0.00005699, 0.00008023, 0.00011086, 0.00013862, 0.00015196, 0.00014463, 0.00012161, 0.00010085, 0.00010861, 0.00016892, 0.00029039, 0.00045626, 0.00062297, 0.00073009, 0.00071997, 0.00056074, 0.00026393, -0.00011129, -0.00047107, -0.00071428, -0.00076573, -0.00060341, -0.00026943, 0.00014042, 0.00050563, 0.00071885, 0.00071941, 0.00051106, 0.00015791, -0.00023903, -0.00057585, -0.00077786, -0.00081892, -0.00072265, -0.00054711, -0.00036027, -0.00021622, -0.00014046, -0.00012803, -0.00015296, -0.00018334, -0.00019532, -0.00018061, -0.00014595, -0.00010639, -0.00007649, -0.00006366, -0.00006614, -0.00007559, -0.00008223, -0.00007974, -0.00006757, -0.00005022, -0.00003419, -0.00002451, -0.00002262, -0.00002625, -0.00003114, -0.00003336, -0.00003116, -0.00002535, -0.00001843, -0.00001304, -0.00001060, -0.00001082, -0.00001216, -0.00001281, -0.00001169, -0.00000892, -0.00000553, -0.00000286, -0.00000176, -0.00000228, -0.00000374, -0.00000517, -0.00000589, -0.00000575, -0.00000515, -0.00000473, -0.00000496, -0.00000592, -0.00000725, -0.00000842, -0.00000894, -0.00000861, -0.00000758, -0.00000626, -0.00000512, -0.00000454, -0.00000472, -0.00000564, -0.00000708, -0.00000872, },
		{ -0.00003149, -0.00002813, -0.00002543, -0.00002317, -0.00002096, -0.00001839, -0.00001513, -0.00001113, -0.00000668, -0.00000229, 0.00000156, 0.00000465, 0.00000723, 0.00000996, 0.00001352, 0.00001827, 0.00002392, 0.00002955, 0.00003406, 0.00003674, 0.00003780, 0.00003836, 0.00004001, 0.00004389, 0.00005002, 0.00005705, 0.00006282, 0.00006562, 0.00006522, 0.00006337, 0.00006308, 0.00006700, 0.00007568, 0.00008670, 0.00009549, 0.00009764, 0.00009183, 0.00008156, 0.00007452, 0.00007917, 0.00009985, 0.00013286, 0.00016610, 0.00018347, 0.00017316, 0.00013646, 0.00009236, 0.00007438, 0.00011868, 0.00024677, 0.00044927, 0.00067836, 0.00085459, 0.00088819, 0.00070964, 0.00029886, -0.00029800, -0.00097234, -0.00157850, -0.00197394, -0.00206133, -0.00181843, -0.00130532, -0.00064633, 0.00000717, 0.00052044, 0.00081102, 0.00086524, 0.00073207, 0.00049846, 0.00025668, 0.00007554, -0.00001536, -0.00002521, 0.00001162, 0.00005488, 0.00007522, 0.00006273, 0.00002564, -0.00001807, -0.00005117, -0.00006485, -0.00006102, -0.00004908, -0.00004005, -0.00004103, -0.00005261, -0.00006980, -0.00008549, -0.00009427, -0.00009480, -0.00008978, -0.00008402, -0.00008178, -0.00008482, -0.00009196, -0.00010015, -0.00010626, -0.00010857, -0.00010734, -0.00010433, -0.00010164, -0.00010060, -0.00010124, -0.00010252, -0.00010305, -0.00010191, -0.00009903, -0.00009516, -0.00009135, -0.00008842, -0.00008656, -0.00008536, -0.00008414, -0.00008230, -0.00007965, -0.00007639, -0.00007296, -0.00006975, -0.00006688, -0.00006422, -0.00006144, -0.00005820, -0.00005433, -0.00004986, -0.00004502, -0.00004014, -0.00003554, },
		{ 0.00003336, 0.00003593, 0.00003854, 0.00004097, 0.00004295, 0.00004422, 0.00004467, 0.00004441, 0.00004382, 0.00004345, 0.00004377, 0.00004496, 0.00004678, 0.00004865, 0.00004995, 0.00005038, 0.00005024, 0.00005034, 0.00005164, 0.00005468, 0.00005919, 0.00006405, 0.00006784, 0.00006954, 0.00006930, 0.00006856, 0.00006943, 0.00007368, 0.00008160, 0.00009164, 0.00010106, 0.00010733, 0.00010983, 0.00011056, 0.00011354, 0.00012283, 0.00014005, 0.00016286, 0.00018551, 0.00020160, 0.00020797, 0.00020777, 0.00021049, 0.00022839, 0.00027012, 0.00033465, 0.00040897, 0.00047203, 0.00050474, 0.00050250, 0.00048452, 0.00049418, 0.00058744, 0.00081114, 0.00117817, 0.00164921, 0.00213006, 0.00248905, 0.00259146, 0.00234113, 0.00171542, 0.00078078, -0.00031808, -0.00139330, -0.00226277, -0.00279731, -0.00295120, -0.00276710, -0.00235516, -0.00185488, -0.00139337, -0.00105357, -0.00086122, -0.00079191, -0.00079233, -0.00080612, -0.00079421, -0.00074390, -0.00066561, -0.00058162, -0.00051289, -0.00046987, -0.00045007, -0.00044209, -0.00043267, -0.00041319, -0.00038258, -0.00034622, -0.00031195, -0.00028570, -0.00026894, -0.00025866, -0.00024959, -0.00023716, -0.00021962, -0.00019839, -0.00017691, -0.00015858, -0.00014512, -0.00013600, -0.00012903, -0.00012169, -0.00011238, -0.00010101, -0.00008874, -0.00007718, -0.00006746, -0.00005973, -0.00005328, -0.00004701, -0.00004008, -0.00003232, -0.00002419, -0.00001649, -0.00000991, -0.00000471, -0.00000067, 0.00000273, 0.00000600, 0.00000941, 0.00001295, 0.00001643, 0.00001959, 0.00002231, 0.00002464, 0.00002673, 0.00002879, 0.00003098, },
		{ 0.00010699, 0.00010710, 0.00010745, 0.00010813, 0.00010925, 0.00011093, 0.00011320, 0.00011593, 0.00011884, 0.00012154, 0.00012372, 0.00012535, 0.00012674, 0.00012851, 0.00013128, 0.00013536, 0.00014051, 0.00014603, 0.00015109, 0.00015520, 0.00015863, 0.00016233, 0.00016753, 0.00017501, 0.00018453, 0.00019472, 0.00020362, 0.00020975, 0.00021298, 0.00021493, 0.00021829, 0.00022552, 0.00023734, 0.00025199, 0.00026586, 0.00027525, 0.00027860, 0.00027783, 0.00027794, 0.00028465, 0.00030105, 0.00032514, 0.00034973, 0.00036551, 0.00036617, 0.00035314, 0.00033710, 0.00033452, 0.00035984, 0.00041636, 0.00049071, 0.00055496, 0.00057771, 0.00054140, 0.00045912, 0.00038290, 0.00039722, 0.00059665, 0.00105295, 0.00178238, 0.00272575, 0.00375047, 0.00467709, 0.00532387, 0.00555603, 0.00532387, 0.00467709, 0.00375047, 0.00272575, 0.00178238, 0.00105295, 0.00059665, 0.00039722, 0.00038290, 0.00045912, 0.00054140, 0.00057771, 0.00055496, 0.00049071, 0.00041636, 0.00035984, 0.00033452, 0.00033710, 0.00035314, 0.00036617, 0.00036551, 0.00034973, 0.00032514, 0.00030105, 0.00028465, 0.00027794, 0.00027783, 0.00027860, 0.00027525, 0.00026586, 0.00025199, 0.00023734, 0.00022552, 0.00021829, 0.00021493, 0.00021298, 0.00020975, 0.00020362, 0.00019472, 0.00018453, 0.00017501, 0.00016753, 0.00016233, 0.00015863, 0.00015520, 0.00015109, 0.00014603, 0.00014051, 0.00013536, 0.00013128, 0.00012851, 0.00012674, 0.00012535, 0.00012372, 0.00012154, 0.00011884, 0.00011593, 0.00011320, 0.00011093, 0.00010925, 0.00010813, 0.00010745, 0.00010710, },
		{ 0.00003336, 0.00003098, 0.00002879, 0.00002673, 0.00002464, 0.00002231, 0.00001959, 0.00001643, 0.00001295, 0.00000941, 0.00000600, 0.00000273, -0.00000067, -0.00000471, -0.00000991, -0.00001649, -0.00002419, -0.00003232, -0.00004008, -0.00004701, -0.00005328, -0.00005973, -0.00006746, -0.00007718, -0.00008874, -0.00010101, -0.00011238, -0.00012169, -0.00012903, -0.00013600, -0.00014512, -0.00015858, -0.00017691, -0.00019839, -0.00021962, -0.00023716, -0.00024959, -0.00025866, -0.00026894, -0.00028570, -0.00031195, -0.00034622, -0.00038258, -0.00041319, -0.00043267, -0.00044209, -0.00045007, -0.00046987, -0.00051289, -0.00058162, -0.00066561, -0.00074390, -0.00079421, -0.00080612, -0.00079233, -0.00079191, -0.00086122, -0.00105357, -0.00139337, -0.00185488, -0.00235516, -0.00276710, -0.00295120, -0.00279731, -0.00226277, -0.00139330, -0.00031808, 0.00078078, 0.00171542, 0.00234113, 0.00259146, 0.00248905, 0.00213006, 0.00164921, 0.00117817, 0.00081114, 0.00058744, 0.00049418, 0.00048452, 0.00050250, 0.00050474, 0.00047203, 0.00040897, 0.00033465, 0.00027012, 0.00022839, 0.00021049, 0.00020777, 0.00020797, 0.00020160, 0.00018551, 0.00016286, 0.00014005, 0.00012283, 0.00011354, 0.00011056, 0.00010983, 0.00010733, 0.00010106, 0.00009164, 0.00008160, 0.00007368, 0.00006943, 0.00006856, 0.00006930, 0.00006954, 0.00006784, 0.00006405, 0.00005919, 0.00005468, 0.00005164, 0.00005034, 0.00005024, 0.00005038, 0.00004995, 0.00004865, 0.00004678, 0.00004496, 0.00004377, 0.00004345, 0.00004382, 0.00004441, 0.00004467, 0.00004422, 0.00004295, 0.00004097, 0.00003854, 0.00003593, },
		{ -0.00003149, -0.00003554, -0.00004014, -0.00004502, -0.00004986, -0.00005433, -0.00005820, -0.00006144, -0.00006422, -0.00006688, -0.00006975, -0.00007296, -0.00007639, -0.00007965, -0.00008230, -0.00008414, -0.00008536, -0.00008656, -0.00008842, -0.00009135, -0.00009516, -0.00009903, -0.00010191, -0.00010305, -0.00010252, -0.00010124, -0.00010060, -0.00010164, -0.00010433, -0.00010734, -0.00010857, -0.00010626, -0.00010015, -0.00009196, -0.00008482, -0.00008178, -0.00008402, -0.00008978, -0.00009480, -0.00009427, -0.00008549, -0.00006980, -0.00005261, -0.00004103, -0.00004005, -0.00004908, -0.00006102, -0.00006485, -0.00005117, -0.00001807, 0.00002564, 0.00006273, 0.00007522, 0.00005488, 0.00001162, -0.00002521, -0.00001536, 0.00007554, 0.00025668, 0.00049846, 0.00073207, 0.00086524, 0.00081102, 0.00052044, 0.00000717, -0.00064633, -0.00130532, -0.00181843, -0.00206133, -0.00197394, -0.00157850, -0.00097234, -0.00029800, 0.00029886, 0.00070964, 0.00088819, 0.00085459, 0.00067836, 0.00044927, 0.00024677, 0.00011868, 0.00007438, 0.00009236, 0.00013646, 0.00017316, 0.00018347, 0.00016610, 0.00013286, 0.00009985, 0.00007917, 0.00007452, 0.00008156, 0.00009183, 0.00009764, 0.00009549, 0.00008670, 0.00007568, 0.00006700, 0.00006308, 0.00006337, 0.00006522, 0.00006562, 0.00006282, 0.00005705, 0.00005002, 0.00004389, 0.00004001, 0.00003836, 0.00003780, 0.00003674, 0.00003406, 0.00002955, 0.00002392, 0.00001827, 0.00001352, 0.00000996, 0.00000723, 0.00000465, 0.00000156, -0.00000229, -0.00000668, -0.00001113, -0.00001513, -0.00001839, -0.00002096, -0.00002317, -0.00002543, -0.00002813, },
		{ -0.00001015, -0.00000872, -0.00000708, -0.00000564, -0.00000472, -0.00000454, -0.00000512, -0.00000626, -0.00000758, -0.00000861, -0.00000894, -0.00000842, -0.00000725, -0.00000592, -0.00000496, -0.00000473, -0.00000515, -0.00000575, -0.00000589, -0.00000517, -0.00000374, -0.00000228, -0.00000176, -0.00000286, -0.00000553, -0.00000892, -0.00001169, -0.00001281, -0.00001216, -0.00001082, -0.00001060, -0.00001304, -0.00001843, -0.00002535, -0.00003116, -0.00003336, -0.00003114, -0.00002625, -0.00002262, -0.00002451, -0.00003419, -0.00005022, -0.00006757, -0.00007974, -0.00008223, -0.00007559, -0.00006614, -0.00006366, -0.00007649, -0.00010639, -0.00014595, -0.00018061, -0.00019532, -0.00018334, -0.00015296, -0.00012803, -0.00014046, -0.00021622, -0.00036027, -0.00054711, -0.00072265, -0.00081892, -0.00077786, -0.00057585, -0.00023903, 0.00015791, 0.00051106, 0.00071941, 0.00071885, 0.00050563, 0.00014042, -0.00026943, -0.00060341, -0.00076573, -0.00071428, -0.00047107, -0.00011129, 0.00026393, 0.00056074, 0.00071997, 0.00073009, 0.00062297, 0.00045626, 0.00029039, 0.00016892, 0.00010861, 0.00010085, 0.00012161, 0.00014463, 0.00015196, 0.00013862, 0.00011086, 0.00008023, 0.00005699, 0.00004580, 0.00004496, 0.00004877, 0.00005116, 0.00004865, 0.00004132, 0.00003194, 0.00002389, 0.00001929, 0.00001813, 0.00001867, 0.00001871, 0.00001677, 0.00001281, 0.00000795, 0.00000376, 0.00000131, 0.00000071, 0.00000116, 0.00000149, 0.00000082, -0.00000106, -0.00000367, -0.00000619, -0.00000793, -0.00000864, -0.00000858, -0.00000830, -0.00000831, -0.00000883, -0.00000975, -0.00001069, -0.00001122, -0.00001105, },
		{ -0.00000232, -0.00000391, -0.00000527, -0.00000612, -0.00000633, -0.00000596, -0.00000525, -0.00000455, -0.00000429, -0.00000484, -0.00000638, -0.00000882, -0.00001174, -0.00001448, -0.00001642, -0.00001724, -0.00001707, -0.00001652, -0.00001635, -0.00001708, -0.00001873, -0.00002071, -0.00002213, -0.00002232, -0.00002126, -0.00001970, -0.00001886, -0.00001974, -0.00002247, -0.00002609, -0.00002892, -0.00002943, -0.00002721, -0.00002338, -0.00002018, -0.00001981, -0.00002302, -0.00002836, -0.00003251, -0.00003190, -0.00002478, -0.00001265, -0.00000014, 0.00000696, 0.00000480, -0.00000588, -0.00001919, -0.00002629, -0.00001974, 0.00000195, 0.00003213, 0.00005793, 0.00006617, 0.00005096, 0.00001934, -0.00000854, -0.00000616, 0.00004749, 0.00015549, 0.00029581, 0.00042398, 0.00048608, 0.00043856, 0.00026729, -0.00000209, -0.00030806, -0.00057123, -0.00072097, -0.00072078, -0.00058257, -0.00036384, -0.00014800, -0.00001480, -0.00001187, -0.00013793, -0.00034383, -0.00055094, -0.00067959, -0.00067691, -0.00053392, -0.00028643, -0.00000026, 0.00025249, 0.00041706, 0.00047116, 0.00042718, 0.00032253, 0.00020313, 0.00010693, 0.00005353, 0.00004229, 0.00005809, 0.00008091, 0.00009488, 0.00009326, 0.00007852, 0.00005858, 0.00004175, 0.00003286, 0.00003197, 0.00003555, 0.00003911, 0.00003954, 0.00003630, 0.00003100, 0.00002604, 0.00002314, 0.00002256, 0.00002324, 0.00002364, 0.00002263, 0.00002006, 0.00001668, 0.00001362, 0.00001169, 0.00001101, 0.00001106, 0.00001099, 0.00001018, 0.00000846, 0.00000618, 0.00000396, 0.00000231, 0.00000140, 0.00000102, 0.00000076, 0.00000020, -0.00000085, },
		{ 0.00000066, 0.00000116, 0.00000139, 0.00000118, 0.00000060, -0.00000016, -0.00000082, -0.00000109, -0.00000078, 0.00000018, 0.00000168, 0.00000342, 0.00000489, 0.00000555, 0.00000502, 0.00000325, 0.00000065, -0.00000198, -0.00000379, -0.00000424, -0.00000341, -0.00000199, -0.00000100, -0.00000125, -0.00000289, -0.00000528, -0.00000729, -0.00000791, -0.00000693, -0.00000519, -0.00000424, -0.00000551, -0.00000947, -0.00001513, -0.00002044, -0.00002327, -0.00002277, -0.00002004, -0.00001790, -0.00001947, -0.00002634, -0.00003725, -0.00004826, -0.00005444, -0.00005266, -0.00004384, -0.00003339, -0.00002920, -0.00003772, -0.00005996, -0.00008970, -0.00011530, -0.00012504, -0.00011383, -0.00008801, -0.00006526, -0.00006856, -0.00011572, -0.00020888, -0.00032880, -0.00043767, -0.00049080, -0.00045337, -0.00031607, -0.00010248, 0.00013544, 0.00033354, 0.00043756, 0.00042284, 0.00030367, 0.00012817, -0.00003969, -0.00014422, -0.00015866, -0.00009438, 0.00000510, 0.00008267, 0.00009128, 0.00001316, -0.00013284, -0.00029849, -0.00042467, -0.00046353, -0.00039588, -0.00023713, -0.00003058, 0.00016885, 0.00031348, 0.00037733, 0.00036141, 0.00028912, 0.00019476, 0.00011028, 0.00005553, 0.00003465, 0.00003894, 0.00005362, 0.00006519, 0.00006628, 0.00005676, 0.00004160, 0.00002713, 0.00001781, 0.00001465, 0.00001573, 0.00001787, 0.00001853, 0.00001685, 0.00001363, 0.00001044, 0.00000854, 0.00000825, 0.00000893, 0.00000953, 0.00000923, 0.00000788, 0.00000597, 0.00000427, 0.00000333, 0.00000325, 0.00000365, 0.00000395, 0.00000372, 0.00000288, 0.00000172, 0.00000070, 0.00000016, 0.00000021, },
		{ 0.00001308, 0.00001130, 0.00000999, 0.00000915, 0.00000861, 0.00000807, 0.00000728, 0.00000612, 0.00000460, 0.00000288, 0.00000114, -0.00000034, -0.00000131, -0.00000155, -0.00000102, 0.00000004, 0.00000113, 0.00000152, 0.00000061, -0.00000183, -0.00000543, -0.00000926, -0.00001220, -0.00001342, -0.00001285, -0.00001125, -0.00000989, -0.00000991, -0.00001166, -0.00001444, -0.00001680, -0.00001732, -0.00001546, -0.00001206, -0.00000902, -0.00000834, -0.00001091, -0.00001579, -0.00002043, -0.00002189, -0.00001855, -0.00001130, -0.00000350, 0.00000056, -0.00000193, -0.00001024, -0.00001969, -0.00002351, -0.00001625, 0.00000266, 0.00002743, 0.00004767, 0.00005308, 0.00003950, 0.00001293, -0.00001070, -0.00001114, 0.00002701, 0.00010511, 0.00020587, 0.00029623, 0.00033785, 0.00030174, 0.00018130, -0.00000193, -0.00020311, -0.00036842, -0.00045399, -0.00044174, -0.00034610, -0.00020824, -0.00008017, -0.00000492, -0.00000072, -0.00005567, -0.00013473, -0.00019598, -0.00020900, -0.00016783, -0.00009292, -0.00002170, 0.00000814, -0.00002467, -0.00011510, -0.00023308, -0.00033478, -0.00037935, -0.00034425, -0.00023316, -0.00007390, 0.00009226, 0.00022523, 0.00029893, 0.00030782, 0.00026554, 0.00019718, 0.00012902, 0.00007978, 0.00005642, 0.00005494, 0.00006499, 0.00007560, 0.00007958, 0.00007522, 0.00006522, 0.00005413, 0.00004576, 0.00004160, 0.00004083, 0.00004144, 0.00004150, 0.00004015, 0.00003768, 0.00003503, 0.00003306, 0.00003204, 0.00003158, 0.00003102, 0.00002983, 0.00002793, 0.00002564, 0.00002348, 0.00002180, 0.00002063, 0.00001968, 0.00001855, 0.00001702, 0.00001510, },
		{ 0.00000025, -0.00000060, -0.00000175, -0.00000307, -0.00000427, -0.00000508, -0.00000536, -0.00000519, -0.00000480, -0.00000442, -0.00000426, -0.00000439, -0.00000483, -0.00000551, -0.00000631, -0.00000704, -0.00000736, -0.00000695, -0.00000558, -0.00000335, -0.00000079, 0.00000122, 0.00000179, 0.00000045, -0.00000256, -0.00000620, -0.00000905, -0.00000998, -0.00000869, -0.00000599, -0.00000346, -0.00000267, -0.00000428, -0.00000761, -0.00001090, -0.00001219, -0.00001053, -0.00000668, -0.00000300, -0.00000223, -0.00000600, -0.00001360, -0.00002199, -0.00002719, -0.00002647, -0.00002025, -0.00001250, -0.00000918, -0.00001510, -0.00003085, -0.00005141, -0.00006772, -0.00007093, -0.00005770, -0.00003367, -0.00001313, -0.00001394, -0.00004930, -0.00011981, -0.00020961, -0.00028931, -0.00032545, -0.00029339, -0.00018851, -0.00003069, 0.00014052, 0.00027876, 0.00034728, 0.00033219, 0.00024739, 0.00012881, 0.00002024, -0.00004340, -0.00004872, -0.00000725, 0.00005155, 0.00009420, 0.00009798, 0.00006024, -0.00000175, -0.00005977, -0.00008831, -0.00007663, -0.00003381, 0.00001519, 0.00004077, 0.00002210, -0.00004309, -0.00013663, -0.00022642, -0.00027879, -0.00027125, -0.00020080, -0.00008458, 0.00004684, 0.00016052, 0.00023202, 0.00025216, 0.00022768, 0.00017633, 0.00011911, 0.00007285, 0.00004583, 0.00003738, 0.00004072, 0.00004735, 0.00005077, 0.00004837, 0.00004117, 0.00003218, 0.00002437, 0.00001930, 0.00001690, 0.00001602, 0.00001534, 0.00001412, 0.00001238, 0.00001059, 0.00000923, 0.00000843, 0.00000796, 0.00000741, 0.00000647, 0.00000515, 0.00000370, 0.00000244, 0.00000152, 0.00000087, },
		{ 0.00000694, 0.00000659, 0.00000647, 0.00000637, 0.00000606, 0.00000538, 0.00000439, 0.00000334, 0.00000256, 0.00000229, 0.00000258, 0.00000325, 0.00000407, 0.00000480, 0.00000532, 0.00000558, 0.00000557, 0.00000528, 0.00000469, 0.00000396, 0.00000341, 0.00000356, 0.00000483, 0.00000732, 0.00001054, 0.00001348, 0.00001494, 0.00001413, 0.00001113, 0.00000706, 0.00000364, 0.00000246, 0.00000413, 0.00000787, 0.00001179, 0.00001382, 0.00001285, 0.00000948, 0.00000588, 0.00000469, 0.00000755, 0.00001393, 0.00002105, 0.00002521, 0.00002376, 0.00001695, 0.00000833, 0.00000337, 0.00000662, 0.00001886, 0.00003582, 0.00004950, 0.00005206, 0.00004043, 0.00001958, 0.00000205, 0.00000346, 0.00003517, 0.00009723, 0.00017522, 0.00024296, 0.00027113, 0.00023869, 0.00014255, 0.00000113, -0.00015030, -0.00027124, -0.00033038, -0.00031688, -0.00024408, -0.00014387, -0.00005405, -0.00000382, -0.00000344, -0.00004199, -0.00009398, -0.00013158, -0.00013684, -0.00010880, -0.00006253, -0.00002100, -0.00000382, -0.00001809, -0.00005552, -0.00009683, -0.00012134, -0.00011702, -0.00008645, -0.00004594, -0.00001815, -0.00002180, -0.00006278, -0.00013073, -0.00020250, -0.00025089, -0.00025507, -0.00020837, -0.00012047, -0.00001353, 0.00008599, 0.00015624, 0.00018638, 0.00017840, 0.00014429, 0.00010035, 0.00006113, 0.00003525, 0.00002424, 0.00002421, 0.00002893, 0.00003288, 0.00003314, 0.00002961, 0.00002399, 0.00001841, 0.00001429, 0.00001196, 0.00001091, 0.00001039, 0.00000990, 0.00000932, 0.00000883, 0.00000858, 0.00000858, 0.00000863, 0.00000850, 0.00000808, 0.00000749, },
		{ 0.00001485, 0.00001424, 0.00001338, 0.00001248, 0.00001175, 0.00001127, 0.00001092, 0.00001039, 0.00000941, 0.00000788, 0.00000597, 0.00000406, 0.00000255, 0.00000165, 0.00000134, 0.00000137, 0.00000145, 0.00000139, 0.00000111, 0.00000064, -0.00000005, -0.00000107, -0.00000255, -0.00000448, -0.00000651, -0.00000802, -0.00000830, -0.00000701, -0.00000453, -0.00000199, -0.00000093, -0.00000256, -0.00000708, -0.00001338, -0.00001937, -0.00002296, -0.00002308, -0.00002042, -0.00001722, -0.00001623, -0.00001921, -0.00002579, -0.00003344, -0.00003864, -0.00003878, -0.00003396, -0.00002738, -0.00002411, -0.00002847, -0.00004131, -0.00005883, -0.00007372, -0.00007867, -0.00007067, -0.00005393, -0.00003956, -0.00004153, -0.00006999, -0.00012506, -0.00019389, -0.00025330, -0.00027748, -0.00024817, -0.00016318, -0.00003933, 0.00009192, 0.00019505, 0.00024311, 0.00022752, 0.00016090, 0.00007174, -0.00000701, -0.00005055, -0.00005096, -0.00001862, 0.00002406, 0.00005331, 0.00005450, 0.00002772, -0.00001364, -0.00005015, -0.00006597, -0.00005610, -0.00002820, 0.00000174, 0.00001745, 0.00001029, -0.00001676, -0.00005086, -0.00007578, -0.00007997, -0.00006239, -0.00003331, -0.00000983, -0.00000798, -0.00003521, -0.00008638, -0.00014513, -0.00019000, -0.00020261, -0.00017460, -0.00011058, -0.00002615, 0.00005787, 0.00012262, 0.00015689, 0.00015954, 0.00013812, 0.00010484, 0.00007178, 0.00004720, 0.00003398, 0.00003040, 0.00003220, 0.00003503, 0.00003603, 0.00003438, 0.00003079, 0.00002658, 0.00002285, 0.00002010, 0.00001827, 0.00001707, 0.00001625, 0.00001570, 0.00001539, 0.00001527, 0.00001515, },
		{ 0.00000233, 0.00000174, 0.00000140, 0.00000098, 0.00000026, -0.00000084, -0.00000211, -0.00000325, -0.00000401, -0.00000436, -0.00000456, -0.00000502, -0.00000607, -0.00000779, -0.00000989, -0.00001187, -0.00001327, -0.00001385, -0.00001371, -0.00001318, -0.00001258, -0.00001211, -0.00001174, -0.00001136, -0.00001098, -0.00001081, -0.00001116, -0.00001218, -0.00001359, -0.00001461, -0.00001424, -0.00001179, -0.00000743, -0.00000235, 0.00000153, 0.00000249, -0.00000008, -0.00000515, -0.00001026, -0.00001263, -0.00001054, -0.00000444, 0.00000311, 0.00000851, 0.00000910, 0.00000477, -0.00000156, -0.00000512, -0.00000179, 0.00000939, 0.00002498, 0.00003824, 0.00004233, 0.00003432, 0.00001794, 0.00000334, 0.00000334, 0.00002740, 0.00007585, 0.00013724, 0.00019073, 0.00021306, 0.00018779, 0.00011294, 0.00000357, -0.00011229, -0.00020311, -0.00024525, -0.00023149, -0.00017340, -0.00009629, -0.00002903, 0.00000708, 0.00000578, -0.00002330, -0.00006036, -0.00008499, -0.00008514, -0.00006165, -0.00002664, 0.00000314, 0.00001452, 0.00000394, -0.00002147, -0.00004796, -0.00006213, -0.00005741, -0.00003694, -0.00001158, 0.00000566, 0.00000615, -0.00001021, -0.00003515, -0.00005623, -0.00006310, -0.00005277, -0.00003130, -0.00001123, -0.00000582, -0.00002278, -0.00006030, -0.00010713, -0.00014667, -0.00016326, -0.00014813, -0.00010251, -0.00003703, 0.00003237, 0.00008989, 0.00012486, 0.00013424, 0.00012225, 0.00009770, 0.00007024, 0.00004731, 0.00003247, 0.00002555, 0.00002401, 0.00002463, 0.00002493, 0.00002374, 0.00002105, 0.00001747, 0.00001371, 0.00001027, 0.00000737, 0.00000507, 0.00000340, },
		{ 0.00001357, 0.00001127, 0.00000941, 0.00000825, 0.00000791, 0.00000823, 0.00000885, 0.00000932, 0.00000933, 0.00000885, 0.00000816, 0.00000767, 0.00000769, 0.00000821, 0.00000887, 0.00000913, 0.00000857, 0.00000712, 0.00000515, 0.00000327, 0.00000204, 0.00000171, 0.00000211, 0.00000281, 0.00000342, 0.00000377, 0.00000400, 0.00000431, 0.00000476, 0.00000503, 0.00000459, 0.00000302, 0.00000046, -0.00000220, -0.00000362, -0.00000266, 0.00000085, 0.00000569, 0.00000957, 0.00001007, 0.00000597, -0.00000192, -0.00001079, -0.00001702, -0.00001799, -0.00001368, -0.00000701, -0.00000269, -0.00000487, -0.00001462, -0.00002885, -0.00004129, -0.00004545, -0.00003849, -0.00002374, -0.00001053, -0.00001077, -0.00003325, -0.00007839, -0.00013572, -0.00018608, -0.00020807, -0.00018652, -0.00011951, -0.00002067, 0.00008453, 0.00016748, 0.00020670, 0.00019562, 0.00014455, 0.00007608, 0.00001590, -0.00001717, -0.00001772, 0.00000564, 0.00003557, 0.00005443, 0.00005213, 0.00002987, -0.00000151, -0.00002754, -0.00003726, -0.00002824, -0.00000718, 0.00001386, 0.00002360, 0.00001694, -0.00000288, -0.00002634, -0.00004257, -0.00004475, -0.00003332, -0.00001544, -0.00000124, 0.00000135, -0.00000956, -0.00002897, -0.00004763, -0.00005674, -0.00005247, -0.00003803, -0.00002240, -0.00001619, -0.00002664, -0.00005397, -0.00009071, -0.00012429, -0.00014182, -0.00013490, -0.00010272, -0.00005218, 0.00000477, 0.00005531, 0.00008981, 0.00010427, 0.00010053, 0.00008459, 0.00006385, 0.00004464, 0.00003067, 0.00002279, 0.00001981, 0.00001962, 0.00002025, 0.00002046, 0.00001976, 0.00001817, 0.00001599, },
		{ 0.00001768, 0.00001741, 0.00001646, 0.00001471, 0.00001228, 0.00000955, 0.00000703, 0.00000522, 0.00000435, 0.00000433, 0.00000471, 0.00000494, 0.00000461, 0.00000369, 0.00000253, 0.00000167, 0.00000148, 0.00000196, 0.00000266, 0.00000292, 0.00000221, 0.00000046, -0.00000189, -0.00000411, -0.00000551, -0.00000583, -0.00000529, -0.00000444, -0.00000375, -0.00000336, -0.00000300, -0.00000224, -0.00000096, 0.00000048, 0.00000125, 0.00000061, -0.00000155, -0.00000430, -0.00000588, -0.00000456, 0.00000036, 0.00000782, 0.00001512, 0.00001902, 0.00001742, 0.00001072, 0.00000207, -0.00000379, -0.00000284, 0.00000593, 0.00001945, 0.00003159, 0.00003604, 0.00002991, 0.00001616, 0.00000347, 0.00000293, 0.00002283, 0.00006363, 0.00011566, 0.00016121, 0.00018047, 0.00015944, 0.00009641, 0.00000406, -0.00009402, -0.00017142, -0.00020835, -0.00019876, -0.00015210, -0.00008908, -0.00003321, -0.00000174, 0.00000021, -0.00001979, -0.00004632, -0.00006373, -0.00006310, -0.00004546, -0.00002023, 0.00000012, 0.00000638, -0.00000317, -0.00002240, -0.00004084, -0.00004901, -0.00004304, -0.00002624, -0.00000711, 0.00000506, 0.00000483, -0.00000698, -0.00002392, -0.00003740, -0.00004099, -0.00003352, -0.00001944, -0.00000635, -0.00000117, -0.00000661, -0.00001996, -0.00003455, -0.00004320, -0.00004194, -0.00003211, -0.00001992, -0.00001361, -0.00001954, -0.00003901, -0.00006724, -0.00009492, -0.00011169, -0.00011004, -0.00008806, -0.00004995, -0.00000444, 0.00003837, 0.00007022, 0.00008683, 0.00008845, 0.00007889, 0.00006353, 0.00004748, 0.00003421, 0.00002517, 0.00002017, 0.00001812, 0.00001766, },
		{ 0.00002519, 0.00001941, 0.00001638, 0.00001540, 0.00001564, 0.00001625, 0.00001649, 0.00001587, 0.00001426, 0.00001197, 0.00000960, 0.00000783, 0.00000708, 0.00000730, 0.00000803, 0.00000859, 0.00000848, 0.00000764, 0.00000647, 0.00000562, 0.00000556, 0.00000630, 0.00000730, 0.00000778, 0.00000708, 0.00000509, 0.00000231, -0.00000038, -0.00000217, -0.00000277, -0.00000250, -0.00000206, -0.00000206, -0.00000262, -0.00000328, -0.00000338, -0.00000258, -0.00000130, -0.00000070, -0.00000203, -0.00000581, -0.00001117, -0.00001598, -0.00001771, -0.00001489, -0.00000825, -0.00000090, 0.00000284, -0.00000050, -0.00001143, -0.00002664, -0.00004003, -0.00004548, -0.00004026, -0.00002728, -0.00001486, -0.00001363, -0.00003152, -0.00006909, -0.00011739, -0.00015990, -0.00017818, -0.00015926, -0.00010160, -0.00001708, 0.00007238, 0.00014231, 0.00017453, 0.00016366, 0.00011871, 0.00005903, 0.00000645, -0.00002305, -0.00002481, -0.00000584, 0.00001960, 0.00003703, 0.00003807, 0.00002340, 0.00000129, -0.00001724, -0.00002399, -0.00001739, -0.00000270, 0.00001111, 0.00001607, 0.00000916, -0.00000646, -0.00002334, -0.00003366, -0.00003312, -0.00002293, -0.00000895, 0.00000137, 0.00000276, -0.00000541, -0.00001902, -0.00003148, -0.00003712, -0.00003393, -0.00002440, -0.00001411, -0.00000877, -0.00001136, -0.00002070, -0.00003217, -0.00004016, -0.00004098, -0.00003485, -0.00002594, -0.00002047, -0.00002373, -0.00003748, -0.00005875, -0.00008078, -0.00009551, -0.00009654, -0.00008153, -0.00005289, -0.00001681, 0.00001900, 0.00004773, 0.00006533, 0.00007111, 0.00006722, 0.00005732, 0.00004525, 0.00003397, },
		{ 0.00006194, 0.00005586, 0.00004676, 0.00003693, 0.00002801, 0.00002098, 0.00001623, 0.00001373, 0.00001302, 0.00001332, 0.00001369, 0.00001331, 0.00001177, 0.00000922, 0.00000632, 0.00000390, 0.00000254, 0.00000233, 0.00000279, 0.00000314, 0.00000279, 0.00000162, 0.00000009, -0.00000106, -0.00000122, -0.00000035, 0.00000096, 0.00000177, 0.00000128, -0.00000066, -0.00000347, -0.00000609, -0.00000758, -0.00000757, -0.00000650, -0.00000530, -0.00000475, -0.00000497, -0.00000528, -0.00000462, -0.00000230, 0.00000129, 0.00000466, 0.00000589, 0.00000383, -0.00000093, -0.00000579, -0.00000721, -0.00000247, 0.00000848, 0.00002233, 0.00003340, 0.00003621, 0.00002863, 0.00001384, 0.00000002, -0.00000264, 0.00001334, 0.00004836, 0.00009383, 0.00013407, 0.00015168, 0.00013448, 0.00008109, 0.00000267, -0.00008029, -0.00014503, -0.00017476, -0.00016470, -0.00012343, -0.00006904, -0.00002176, 0.00000379, 0.00000357, -0.00001559, -0.00004026, -0.00005688, -0.00005768, -0.00004333, -0.00002156, -0.00000271, 0.00000549, 0.00000130, -0.00001079, -0.00002286, -0.00002785, -0.00002299, -0.00001094, 0.00000200, 0.00000927, 0.00000743, -0.00000221, -0.00001444, -0.00002295, -0.00002352, -0.00001604, -0.00000438, 0.00000568, 0.00000937, 0.00000528, -0.00000416, -0.00001408, -0.00001967, -0.00001861, -0.00001205, -0.00000391, 0.00000131, 0.00000076, -0.00000534, -0.00001392, -0.00002073, -0.00002264, -0.00001928, -0.00001336, -0.00000943, -0.00001171, -0.00002192, -0.00003827, -0.00005593, -0.00006873, -0.00007152, -0.00006197, -0.00004138, -0.00001402, 0.00001438, 0.00003853, 0.00005490, 0.00006237, },
		{ -0.00000362, 0.00001656, 0.00003165, 0.00004049, 0.00004334, 0.00004136, 0.00003607, 0.00002898, 0.00002145, 0.00001464, 0.00000950, 0.00000655, 0.00000580, 0.00000665, 0.00000805, 0.00000891, 0.00000852, 0.00000682, 0.00000444, 0.00000234, 0.00000128, 0.00000147, 0.00000243, 0.00000332, 0.00000342, 0.00000253, 0.00000115, 0.00000013, 0.00000021, 0.00000148, 0.00000328, 0.00000449, 0.00000415, 0.00000202, -0.00000122, -0.00000433, -0.00000612, -0.00000619, -0.00000514, -0.00000420, -0.00000449, -0.00000619, -0.00000837, -0.00000942, -0.00000813, -0.00000469, -0.00000097, 0.00000017, -0.00000351, -0.00001204, -0.00002260, -0.00003034, -0.00003067, -0.00002207, -0.00000780, 0.00000442, 0.00000527, -0.00001187, -0.00004687, -0.00009126, -0.00013006, -0.00014690, -0.00013053, -0.00007994, -0.00000589, 0.00007221, 0.00013298, 0.00016082, 0.00015144, 0.00011303, 0.00006257, 0.00001888, -0.00000462, -0.00000443, 0.00001289, 0.00003475, 0.00004865, 0.00004761, 0.00003258, 0.00001104, -0.00000721, -0.00001489, -0.00001042, 0.00000194, 0.00001473, 0.00002118, 0.00001847, 0.00000868, -0.00000267, -0.00000977, -0.00000951, -0.00000289, 0.00000574, 0.00001112, 0.00000987, 0.00000211, -0.00000865, -0.00001750, -0.00002057, -0.00001699, -0.00000918, -0.00000151, 0.00000197, -0.00000054, -0.00000780, -0.00001634, -0.00002228, -0.00002334, -0.00001981, -0.00001431, -0.00001026, -0.00001013, -0.00001422, -0.00002058, -0.00002614, -0.00002841, -0.00002682, -0.00002314, -0.00002062, -0.00002244, -0.00003008, -0.00004240, -0.00005591, -0.00006604, -0.00006879, -0.00006223, -0.00004702, -0.00002612, },
		{ -0.00005446, -0.00004946, -0.00003805, -0.00002231, -0.00000508, 0.00001093, 0.00002384, 0.00003274, 0.00003754, 0.00003858, 0.00003644, 0.00003179, 0.00002555, 0.00001889, 0.00001308, 0.00000921, 0.00000778, 0.00000845, 0.00001019, 0.00001169, 0.00001192, 0.00001062, 0.00000832, 0.00000607, 0.00000483, 0.00000494, 0.00000598, 0.00000703, 0.00000724, 0.00000635, 0.00000487, 0.00000381, 0.00000405, 0.00000577, 0.00000824, 0.00001014, 0.00001030, 0.00000837, 0.00000515, 0.00000220, 0.00000100, 0.00000209, 0.00000469, 0.00000713, 0.00000781, 0.00000634, 0.00000398, 0.00000311, 0.00000590, 0.00001271, 0.00002140, 0.00002796, 0.00002847, 0.00002157, 0.00001013, 0.00000097, 0.00000227, 0.00001970, 0.00005280, 0.00009347, 0.00012782, 0.00014090, 0.00012265, 0.00007269, 0.00000150, -0.00007255, -0.00012950, -0.00015504, -0.00014555, -0.00010907, -0.00006172, -0.00002112, 0.00000032, -0.00000046, -0.00001715, -0.00003792, -0.00005121, -0.00005073, -0.00003760, -0.00001893, -0.00000384, 0.00000110, -0.00000532, -0.00001880, -0.00003212, -0.00003886, -0.00003635, -0.00002661, -0.00001491, -0.00000686, -0.00000564, -0.00001064, -0.00001815, -0.00002354, -0.00002374, -0.00001872, -0.00001132, -0.00000563, -0.00000472, -0.00000914, -0.00001666, -0.00002354, -0.00002646, -0.00002417, -0.00001803, -0.00001120, -0.00000701, -0.00000729, -0.00001160, -0.00001757, -0.00002223, -0.00002353, -0.00002130, -0.00001717, -0.00001363, -0.00001266, -0.00001473, -0.00001864, -0.00002227, -0.00002379, -0.00002270, -0.00002018, -0.00001857, -0.00002014, -0.00002594, -0.00003514, -0.00004513, -0.00005254, },
		{ -0.00002092, -0.00002848, -0.00003624, -0.00004168, -0.00004279, -0.00003878, -0.00003024, -0.00001882, -0.00000647, 0.00000515, 0.00001505, 0.00002280, 0.00002826, 0.00003126, 0.00003156, 0.00002904, 0.00002401, 0.00001746, 0.00001087, 0.00000580, 0.00000323, 0.00000320, 0.00000477, 0.00000644, 0.00000691, 0.00000562, 0.00000301, 0.00000017, -0.00000177, -0.00000231, -0.00000181, -0.00000128, -0.00000170, -0.00000343, -0.00000594, -0.00000806, -0.00000870, -0.00000759, -0.00000556, -0.00000420, -0.00000500, -0.00000845, -0.00001363, -0.00001859, -0.00002136, -0.00002119, -0.00001907, -0.00001736, -0.00001849, -0.00002338, -0.00003057, -0.00003663, -0.00003792, -0.00003289, -0.00002376, -0.00001635, -0.00001793, -0.00003364, -0.00006314, -0.00009920, -0.00012931, -0.00013997, -0.00012223, -0.00007596, -0.00001098, 0.00005575, 0.00010601, 0.00012693, 0.00011563, 0.00007999, 0.00003519, -0.00000243, -0.00002161, -0.00001990, -0.00000355, 0.00001623, 0.00002866, 0.00002803, 0.00001558, -0.00000192, -0.00001599, -0.00002063, -0.00001496, -0.00000314, 0.00000805, 0.00001272, 0.00000864, -0.00000197, -0.00001392, -0.00002180, -0.00002256, -0.00001683, -0.00000818, -0.00000117, 0.00000108, -0.00000178, -0.00000742, -0.00001229, -0.00001360, -0.00001074, -0.00000545, -0.00000078, 0.00000061, -0.00000214, -0.00000776, -0.00001351, -0.00001667, -0.00001588, -0.00001178, -0.00000658, -0.00000288, -0.00000234, -0.00000499, -0.00000931, -0.00001314, -0.00001480, -0.00001392, -0.00001149, -0.00000920, -0.00000851, -0.00000983, -0.00001241, -0.00001482, -0.00001584, -0.00001522, -0.00001387, -0.00001348, -0.00001564, },
		{ -0.00001673, -0.00001604, -0.00001660, -0.00001946, -0.00002475, -0.00003144, -0.00003765, -0.00004141, -0.00004141, -0.00003748, -0.00003054, -0.00002210, -0.00001359, -0.00000582, 0.00000107, 0.00000730, 0.00001293, 0.00001756, 0.00002030, 0.00002025, 0.00001708, 0.00001141, 0.00000481, -0.00000081, -0.00000393, -0.00000411, -0.00000211, 0.00000048, 0.00000206, 0.00000179, -0.00000003, -0.00000229, -0.00000369, -0.00000354, -0.00000216, -0.00000062, -0.00000008, -0.00000102, -0.00000283, -0.00000414, -0.00000356, -0.00000069, 0.00000355, 0.00000719, 0.00000831, 0.00000616, 0.00000183, -0.00000212, -0.00000295, 0.00000062, 0.00000748, 0.00001436, 0.00001747, 0.00001483, 0.00000792, 0.00000183, 0.00000325, 0.00001721, 0.00004382, 0.00007684, 0.00010497, 0.00011582, 0.00010094, 0.00005995, 0.00000162, -0.00005865, -0.00010423, -0.00012337, -0.00011347, -0.00008176, -0.00004213, -0.00000938, 0.00000643, 0.00000344, -0.00001231, -0.00003037, -0.00004090, -0.00003894, -0.00002602, -0.00000872, 0.00000497, 0.00000955, 0.00000436, -0.00000655, -0.00001682, -0.00002104, -0.00001732, -0.00000791, 0.00000226, 0.00000819, 0.00000726, 0.00000033, -0.00000894, -0.00001609, -0.00001809, -0.00001460, -0.00000794, -0.00000163, 0.00000146, 0.00000044, -0.00000335, -0.00000732, -0.00000908, -0.00000771, -0.00000416, -0.00000060, 0.00000082, -0.00000090, -0.00000510, -0.00000986, -0.00001307, -0.00001345, -0.00001118, -0.00000772, -0.00000497, -0.00000434, -0.00000605, -0.00000919, -0.00001227, -0.00001405, -0.00001418, -0.00001324, -0.00001233, -0.00001237, -0.00001359, -0.00001542, -0.00001690, -0.00001732, },
		{ -0.00001155, -0.00001250, -0.00001273, -0.00001192, -0.00001047, -0.00000944, -0.00001006, -0.00001308, -0.00001825, -0.00002424, -0.00002915, -0.00003123, -0.00002967, -0.00002487, -0.00001820, -0.00001135, -0.00000550, -0.00000086, 0.00000319, 0.00000762, 0.00001292, 0.00001866, 0.00002351, 0.00002586, 0.00002464, 0.00002001, 0.00001343, 0.00000711, 0.00000307, 0.00000220, 0.00000399, 0.00000678, 0.00000872, 0.00000868, 0.00000672, 0.00000403, 0.00000203, 0.00000160, 0.00000247, 0.00000343, 0.00000310, 0.00000084, -0.00000266, -0.00000573, -0.00000661, -0.00000464, -0.00000089, 0.00000212, 0.00000172, -0.00000335, -0.00001192, -0.00002061, -0.00002544, -0.00002408, -0.00001765, -0.00001090, -0.00001041, -0.00002142, -0.00004462, -0.00007460, -0.00010092, -0.00011180, -0.00009893, -0.00006144, -0.00000710, 0.00004991, 0.00009403, 0.00011400, 0.00010700, 0.00007935, 0.00004362, 0.00001338, -0.00000209, -0.00000097, 0.00001135, 0.00002550, 0.00003276, 0.00002893, 0.00001571, -0.00000069, -0.00001295, -0.00001627, -0.00001040, 0.00000058, 0.00001057, 0.00001458, 0.00001096, 0.00000195, -0.00000774, -0.00001350, -0.00001301, -0.00000726, 0.00000023, 0.00000533, 0.00000532, 0.00000020, -0.00000760, -0.00001453, -0.00001775, -0.00001642, -0.00001187, -0.00000673, -0.00000350, -0.00000333, -0.00000563, -0.00000859, -0.00001030, -0.00000975, -0.00000734, -0.00000454, -0.00000300, -0.00000368, -0.00000632, -0.00000967, -0.00001216, -0.00001274, -0.00001136, -0.00000890, -0.00000670, -0.00000579, -0.00000642, -0.00000807, -0.00000978, -0.00001078, -0.00001087, -0.00001044, -0.00001019, -0.00001058, },
		{ -0.00000803, -0.00000819, -0.00000885, -0.00001013, -0.00001162, -0.00001256, -0.00001231, -0.00001077, -0.00000867, -0.00000733, -0.00000809, -0.00001160, -0.00001731, -0.00002359, -0.00002828, -0.00002964, -0.00002709, -0.00002151, -0.00001481, -0.00000901, -0.00000536, -0.00000371, -0.00000277, -0.00000079, 0.00000331, 0.00000934, 0.00001570, 0.00002016, 0.00002090, 0.00001750, 0.00001122, 0.00000446, -0.00000030, -0.00000164, 0.00000020, 0.00000361, 0.00000647, 0.00000734, 0.00000612, 0.00000401, 0.00000265, 0.00000311, 0.00000517, 0.00000741, 0.00000814, 0.00000652, 0.00000333, 0.00000073, 0.00000115, 0.00000572, 0.00001330, 0.00002064, 0.00002388, 0.00002079, 0.00001249, 0.00000371, 0.00000106, 0.00000991, 0.00003127, 0.00006010, 0.00008636, 0.00009851, 0.00008814, 0.00005395, 0.00000299, -0.00005133, -0.00009402, -0.00011403, -0.00010820, -0.00008215, -0.00004760, -0.00001744, -0.00000074, 0.00000034, -0.00000959, -0.00002211, -0.00002922, -0.00002698, -0.00001679, -0.00000407, 0.00000479, 0.00000572, -0.00000116, -0.00001189, -0.00002088, -0.00002370, -0.00001919, -0.00000976, -0.00000001, 0.00000570, 0.00000531, -0.00000014, -0.00000725, -0.00001215, -0.00001243, -0.00000820, -0.00000187, 0.00000324, 0.00000459, 0.00000161, -0.00000417, -0.00001002, -0.00001345, -0.00001330, -0.00001020, -0.00000603, -0.00000282, -0.00000181, -0.00000290, -0.00000493, -0.00000643, -0.00000646, -0.00000507, -0.00000317, -0.00000193, -0.00000214, -0.00000380, -0.00000614, -0.00000809, -0.00000888, -0.00000837, -0.00000711, -0.00000591, -0.00000542, -0.00000579, -0.00000666, -0.00000750, -0.00000794, },
		{ -0.00000496, -0.00000512, -0.00000495, -0.00000445, -0.00000399, -0.00000413, -0.00000523, -0.00000718, -0.00000922, -0.00001033, -0.00000972, -0.00000738, -0.00000430, -0.00000212, -0.00000247, -0.00000606, -0.00001223, -0.00001903, -0.00002394, -0.00002498, -0.00002159, -0.00001498, -0.00000759, -0.00000198, 0.00000033, -0.00000044, -0.00000244, -0.00000320, -0.00000086, 0.00000476, 0.00001202, 0.00001814, 0.00002059, 0.00001824, 0.00001203, 0.00000445, -0.00000159, -0.00000415, -0.00000313, -0.00000012, 0.00000253, 0.00000304, 0.00000113, -0.00000195, -0.00000429, -0.00000460, -0.00000310, -0.00000155, -0.00000219, -0.00000630, -0.00001306, -0.00001964, -0.00002256, -0.00001976, -0.00001234, -0.00000481, -0.00000350, -0.00001352, -0.00003577, -0.00006530, -0.00009227, -0.00010534, -0.00009625, -0.00006359, -0.00001420, 0.00003891, 0.00008103, 0.00010123, 0.00009622, 0.00007127, 0.00003768, 0.00000806, -0.00000858, -0.00000981, 0.00000020, 0.00001345, 0.00002216, 0.00002218, 0.00001439, 0.00000349, -0.00000476, -0.00000664, -0.00000209, 0.00000549, 0.00001129, 0.00001162, 0.00000578, -0.00000377, -0.00001275, -0.00001728, -0.00001576, -0.00000950, -0.00000192, 0.00000325, 0.00000378, -0.00000009, -0.00000603, -0.00001090, -0.00001236, -0.00001002, -0.00000540, -0.00000109, 0.00000064, -0.00000108, -0.00000540, -0.00001034, -0.00001377, -0.00001446, -0.00001255, -0.00000928, -0.00000626, -0.00000466, -0.00000468, -0.00000569, -0.00000664, -0.00000675, -0.00000587, -0.00000449, -0.00000337, -0.00000310, -0.00000373, -0.00000486, -0.00000586, -0.00000626, -0.00000601, -0.00000541, -0.00000491, -0.00000479, },
		{ -0.00000151, -0.00000185, -0.00000250, -0.00000339, -0.00000416, -0.00000437, -0.00000383, -0.00000278, -0.00000192, -0.00000208, -0.00000373, -0.00000661, -0.00000970, -0.00001161, -0.00001128, -0.00000861, -0.00000474, -0.00000167, -0.00000140, -0.00000494, -0.00001166, -0.00001935, -0.00002505, -0.00002635, -0.00002250, -0.00001486, -0.00000632, -0.00000011, 0.00000176, -0.00000062, -0.00000507, -0.00000836, -0.00000784, -0.00000272, 0.00000549, 0.00001361, 0.00001838, 0.00001794, 0.00001280, 0.00000548, -0.00000071, -0.00000329, -0.00000177, 0.00000228, 0.00000626, 0.00000803, 0.00000717, 0.00000511, 0.00000421, 0.00000618, 0.00001084, 0.00001592, 0.00001826, 0.00001576, 0.00000920, 0.00000254, 0.00000158, 0.00001106, 0.00003178, 0.00005905, 0.00008350, 0.00009437, 0.00008386, 0.00005084, 0.00000207, -0.00004973, -0.00009032, -0.00010915, -0.00010321, -0.00007780, -0.00004409, -0.00001451, 0.00000207, 0.00000325, -0.00000686, -0.00002033, -0.00002935, -0.00002961, -0.00002170, -0.00001005, -0.00000027, 0.00000376, 0.00000157, -0.00000407, -0.00000892, -0.00000965, -0.00000553, 0.00000138, 0.00000746, 0.00000953, 0.00000649, -0.00000017, -0.00000717, -0.00001116, -0.00001036, -0.00000530, 0.00000154, 0.00000705, 0.00000905, 0.00000721, 0.00000306, -0.00000093, -0.00000265, -0.00000136, 0.00000205, 0.00000566, 0.00000756, 0.00000675, 0.00000358, -0.00000054, -0.00000390, -0.00000531, -0.00000459, -0.00000247, -0.00000016, 0.00000136, 0.00000169, 0.00000110, 0.00000026, -0.00000025, -0.00000021, 0.00000020, 0.00000058, 0.00000059, 0.00000019, -0.00000043, -0.00000097, -0.00000129, },
		{ -0.00000539, -0.00000508, -0.00000460, -0.00000411, -0.00000395, -0.00000441, -0.00000553, -0.00000697, -0.00000804, -0.00000810, -0.00000687, -0.00000479, -0.00000285, -0.00000222, -0.00000361, -0.00000680, -0.00001056, -0.00001312, -0.00001302, -0.00000991, -0.00000494, -0.00000043, 0.00000113, -0.00000171, -0.00000855, -0.00001702, -0.00002372, -0.00002566, -0.00002168, -0.00001307, -0.00000314, 0.00000420, 0.00000626, 0.00000275, -0.00000399, -0.00001010, -0.00001205, -0.00000831, -0.00000011, 0.00000911, 0.00001537, 0.00001603, 0.00001107, 0.00000298, -0.00000453, -0.00000846, -0.00000794, -0.00000457, -0.00000137, -0.00000095, -0.00000396, -0.00000864, -0.00001185, -0.00001115, -0.00000659, -0.00000143, -0.00000091, -0.00000960, -0.00002851, -0.00005352, -0.00007602, -0.00008595, -0.00007599, -0.00004508, 0.00000035, 0.00004822, 0.00008507, 0.00010105, 0.00009358, 0.00006805, 0.00003538, 0.00000756, -0.00000701, -0.00000635, 0.00000522, 0.00001970, 0.00002933, 0.00002987, 0.00002195, 0.00001010, 0.00000001, -0.00000422, -0.00000189, 0.00000445, 0.00001069, 0.00001340, 0.00001149, 0.00000650, 0.00000151, -0.00000069, 0.00000095, 0.00000519, 0.00000932, 0.00001062, 0.00000785, 0.00000181, -0.00000503, -0.00000983, -0.00001078, -0.00000788, -0.00000288, 0.00000171, 0.00000383, 0.00000282, -0.00000042, -0.00000405, -0.00000625, -0.00000613, -0.00000408, -0.00000147, 0.00000015, -0.00000022, -0.00000253, -0.00000588, -0.00000893, -0.00001065, -0.00001066, -0.00000934, -0.00000750, -0.00000596, -0.00000514, -0.00000505, -0.00000534, -0.00000565, -0.00000579, -0.00000575, -0.00000564, -0.00000554, },
		{ -0.00000456, -0.00000488, -0.00000517, -0.00000525, -0.00000490, -0.00000400, -0.00000271, -0.00000145, -0.00000083, -0.00000131, -0.00000294, -0.00000515, -0.00000694, -0.00000735, -0.00000593, -0.00000312, -0.00000018, 0.00000133, 0.00000034, -0.00000312, -0.00000772, -0.00001131, -0.00001194, -0.00000884, -0.00000305, 0.00000290, 0.00000600, 0.00000418, -0.00000253, -0.00001178, -0.00001974, -0.00002275, -0.00001901, -0.00000956, 0.00000202, 0.00001116, 0.00001433, 0.00001067, 0.00000248, -0.00000587, -0.00000991, -0.00000724, 0.00000136, 0.00001229, 0.00002089, 0.00002373, 0.00002021, 0.00001278, 0.00000564, 0.00000248, 0.00000458, 0.00001021, 0.00001569, 0.00001764, 0.00001515, 0.00001075, 0.00000937, 0.00001580, 0.00003168, 0.00005372, 0.00007410, 0.00008333, 0.00007423, 0.00004548, 0.00000285, -0.00004238, -0.00007753, -0.00009330, -0.00008724, -0.00006446, -0.00003532, -0.00001104, 0.00000064, -0.00000184, -0.00001404, -0.00002815, -0.00003676, -0.00003605, -0.00002703, -0.00001439, -0.00000396, 0.00000020, -0.00000255, -0.00000956, -0.00001657, -0.00001995, -0.00001838, -0.00001317, -0.00000730, -0.00000362, -0.00000345, -0.00000603, -0.00000912, -0.00001043, -0.00000882, -0.00000498, -0.00000091, 0.00000112, -0.00000017, -0.00000443, -0.00000982, -0.00001398, -0.00001515, -0.00001301, -0.00000869, -0.00000417, -0.00000131, -0.00000095, -0.00000269, -0.00000519, -0.00000697, -0.00000712, -0.00000570, -0.00000360, -0.00000201, -0.00000182, -0.00000320, -0.00000557, -0.00000800, -0.00000960, -0.00000996, -0.00000917, -0.00000775, -0.00000626, -0.00000512, -0.00000448, -0.00000426, -0.00000433, },
		{ -0.00000360, -0.00000270, -0.00000216, -0.00000205, -0.00000240, -0.00000312, -0.00000396, -0.00000451, -0.00000433, -0.00000322, -0.00000137, 0.00000055, 0.00000166, 0.00000122, -0.00000091, -0.00000407, -0.00000699, -0.00000831, -0.00000727, -0.00000421, -0.00000056, 0.00000174, 0.00000113, -0.00000268, -0.00000840, -0.00001359, -0.00001572, -0.00001348, -0.00000759, -0.00000071, 0.00000362, 0.00000265, -0.00000421, -0.00001477, -0.00002486, -0.00003009, -0.00002788, -0.00001878, -0.00000647, 0.00000384, 0.00000766, 0.00000337, -0.00000701, -0.00001865, -0.00002621, -0.00002630, -0.00001912, -0.00000836, 0.00000060, 0.00000337, -0.00000132, -0.00001098, -0.00002078, -0.00002609, -0.00002516, -0.00002023, -0.00001658, -0.00001982, -0.00003262, -0.00005258, -0.00007249, -0.00008298, -0.00007655, -0.00005124, -0.00001208, 0.00003040, 0.00006398, 0.00007951, 0.00007433, 0.00005305, 0.00002537, 0.00000191, -0.00000998, -0.00000881, 0.00000124, 0.00001285, 0.00001920, 0.00001694, 0.00000732, -0.00000494, -0.00001426, -0.00001693, -0.00001265, -0.00000434, 0.00000358, 0.00000743, 0.00000593, 0.00000049, -0.00000576, -0.00000977, -0.00000999, -0.00000697, -0.00000281, 0.00000015, 0.00000058, -0.00000125, -0.00000378, -0.00000515, -0.00000427, -0.00000145, 0.00000181, 0.00000364, 0.00000282, -0.00000060, -0.00000530, -0.00000934, -0.00001110, -0.00001003, -0.00000680, -0.00000294, -0.00000003, 0.00000100, 0.00000020, -0.00000155, -0.00000310, -0.00000360, -0.00000291, -0.00000155, -0.00000038, -0.00000011, -0.00000100, -0.00000276, -0.00000473, -0.00000627, -0.00000696, -0.00000675, -0.00000589, -0.00000473, },
		{ -0.00000881, -0.00000921, -0.00000903, -0.00000841, -0.00000755, -0.00000665, -0.00000594, -0.00000569, -0.00000611, -0.00000716, -0.00000855, -0.00000966, -0.00000983, -0.00000867, -0.00000633, -0.00000361, -0.00000168, -0.00000161, -0.00000376, -0.00000751, -0.00001135, -0.00001351, -0.00001282, -0.00000935, -0.00000458, -0.00000082, -0.00000015, -0.00000336, -0.00000935, -0.00001546, -0.00001861, -0.00001679, -0.00001025, -0.00000159, 0.00000519, 0.00000653, 0.00000109, -0.00000932, -0.00002031, -0.00002674, -0.00002508, -0.00001520, -0.00000065, 0.00001276, 0.00001953, 0.00001709, 0.00000701, -0.00000557, -0.00001429, -0.00001462, -0.00000602, 0.00000774, 0.00002046, 0.00002655, 0.00002402, 0.00001565, 0.00000779, 0.00000731, 0.00001787, 0.00003758, 0.00005914, 0.00007261, 0.00006969, 0.00004756, 0.00001065, -0.00003080, -0.00006456, -0.00008118, -0.00007754, -0.00005776, -0.00003114, -0.00000808, 0.00000410, 0.00000368, -0.00000554, -0.00001667, -0.00002322, -0.00002198, -0.00001409, -0.00000394, 0.00000335, 0.00000448, -0.00000054, -0.00000872, -0.00001577, -0.00001828, -0.00001530, -0.00000853, -0.00000128, 0.00000329, 0.00000366, 0.00000045, -0.00000411, -0.00000748, -0.00000813, -0.00000610, -0.00000282, -0.00000013, 0.00000071, -0.00000030, -0.00000208, -0.00000316, -0.00000256, -0.00000035, 0.00000237, 0.00000408, 0.00000364, 0.00000093, -0.00000314, -0.00000699, -0.00000919, -0.00000903, -0.00000687, -0.00000379, -0.00000112, 0.00000023, 0.00000006, -0.00000112, -0.00000247, -0.00000329, -0.00000331, -0.00000279, -0.00000228, -0.00000230, -0.00000310, -0.00000455, -0.00000626, -0.00000779, },
		{ -0.00000260, -0.00000341, -0.00000442, -0.00000538, -0.00000611, -0.00000649, -0.00000645, -0.00000596, -0.00000503, -0.00000380, -0.00000255, -0.00000170, -0.00000165, -0.00000252, -0.00000402, -0.00000542, -0.00000582, -0.00000456, -0.00000167, 0.00000202, 0.00000507, 0.00000605, 0.00000425, 0.00000014, -0.00000465, -0.00000796, -0.00000810, -0.00000474, 0.00000073, 0.00000572, 0.00000756, 0.00000483, -0.00000172, -0.00000937, -0.00001449, -0.00001431, -0.00000840, 0.00000085, 0.00000907, 0.00001187, 0.00000699, -0.00000437, -0.00001777, -0.00002734, -0.00002849, -0.00002017, -0.00000564, 0.00000878, 0.00001646, 0.00001361, 0.00000114, -0.00001555, -0.00002902, -0.00003333, -0.00002705, -0.00001419, -0.00000259, -0.00000021, -0.00001116, -0.00003325, -0.00005825, -0.00007510, -0.00007457, -0.00005337, -0.00001594, 0.00002701, 0.00006261, 0.00008080, 0.00007804, 0.00005835, 0.00003112, 0.00000702, -0.00000630, -0.00000683, 0.00000184, 0.00001291, 0.00001984, 0.00001926, 0.00001202, 0.00000220, -0.00000534, -0.00000742, -0.00000398, 0.00000223, 0.00000729, 0.00000818, 0.00000422, -0.00000275, -0.00000949, -0.00001307, -0.00001221, -0.00000784, -0.00000239, 0.00000151, 0.00000228, 0.00000002, -0.00000372, -0.00000694, -0.00000818, -0.00000724, -0.00000504, -0.00000301, -0.00000223, -0.00000288, -0.00000423, -0.00000513, -0.00000472, -0.00000296, -0.00000064, 0.00000100, 0.00000095, -0.00000105, -0.00000437, -0.00000778, -0.00001001, -0.00001034, -0.00000883, -0.00000624, -0.00000361, -0.00000179, -0.00000112, -0.00000138, -0.00000205, -0.00000257, -0.00000269, -0.00000247, -0.00000219, -0.00000218, },
		{ -0.00000081, -0.00000109, -0.00000139, -0.00000181, -0.00000239, -0.00000310, -0.00000386, -0.00000462, -0.00000534, -0.00000598, -0.00000644, -0.00000652, -0.00000607, -0.00000507, -0.00000379, -0.00000276, -0.00000255, -0.00000349, -0.00000539, -0.00000747, -0.00000860, -0.00000783, -0.00000493, -0.00000067, 0.00000334, 0.00000528, 0.00000400, -0.00000032, -0.00000604, -0.00001066, -0.00001191, -0.00000892, -0.00000281, 0.00000366, 0.00000724, 0.00000581, -0.00000048, -0.00000895, -0.00001553, -0.00001661, -0.00001089, -0.00000033, 0.00001047, 0.00001629, 0.00001380, 0.00000337, -0.00001069, -0.00002181, -0.00002417, -0.00001551, 0.00000144, 0.00001991, 0.00003203, 0.00003246, 0.00002123, 0.00000413, -0.00000962, -0.00001149, 0.00000223, 0.00002812, 0.00005655, 0.00007529, 0.00007462, 0.00005157, 0.00001151, -0.00003377, -0.00007048, -0.00008818, -0.00008357, -0.00006128, -0.00003151, -0.00000562, 0.00000837, 0.00000845, -0.00000152, -0.00001430, -0.00002289, -0.00002361, -0.00001711, -0.00000744, 0.00000040, 0.00000306, 0.00000018, -0.00000576, -0.00001105, -0.00001279, -0.00001025, -0.00000506, -0.00000015, 0.00000187, 0.00000001, -0.00000469, -0.00000982, -0.00001289, -0.00001256, -0.00000920, -0.00000454, -0.00000072, 0.00000078, -0.00000025, -0.00000277, -0.00000524, -0.00000638, -0.00000584, -0.00000420, -0.00000251, -0.00000169, -0.00000198, -0.00000289, -0.00000352, -0.00000315, -0.00000167, 0.00000034, 0.00000191, 0.00000213, 0.00000068, -0.00000204, -0.00000508, -0.00000736, -0.00000816, -0.00000737, -0.00000545, -0.00000317, -0.00000128, -0.00000017, 0.00000014, -0.00000006, -0.00000046, },
		{ -0.00000034, 0.00000062, 0.00000106, 0.00000111, 0.00000086, 0.00000040, -0.00000024, -0.00000096, -0.00000164, -0.00000217, -0.00000254, -0.00000283, -0.00000320, -0.00000373, -0.00000433, -0.00000474, -0.00000460, -0.00000372, -0.00000229, -0.00000085, -0.00000017, -0.00000081, -0.00000278, -0.00000531, -0.00000713, -0.00000696, -0.00000422, 0.00000050, 0.00000550, 0.00000861, 0.00000815, 0.00000386, -0.00000277, -0.00000897, -0.00001189, -0.00000999, -0.00000395, 0.00000344, 0.00000841, 0.00000802, 0.00000168, -0.00000825, -0.00001733, -0.00002109, -0.00001722, -0.00000701, 0.00000489, 0.00001243, 0.00001098, -0.00000018, -0.00001713, -0.00003267, -0.00003956, -0.00003401, -0.00001793, 0.00000138, 0.00001405, 0.00001208, -0.00000676, -0.00003725, -0.00006819, -0.00008671, -0.00008348, -0.00005680, -0.00001363, 0.00003290, 0.00006850, 0.00008308, 0.00007445, 0.00004861, 0.00001693, -0.00000871, -0.00002046, -0.00001707, -0.00000344, 0.00001223, 0.00002239, 0.00002321, 0.00001565, 0.00000428, -0.00000536, -0.00000945, -0.00000739, -0.00000162, 0.00000403, 0.00000642, 0.00000454, -0.00000025, -0.00000523, -0.00000787, -0.00000716, -0.00000396, -0.00000041, 0.00000130, 0.00000007, -0.00000360, -0.00000797, -0.00001102, -0.00001148, -0.00000937, -0.00000587, -0.00000269, -0.00000113, -0.00000155, -0.00000332, -0.00000527, -0.00000635, -0.00000615, -0.00000501, -0.00000372, -0.00000302, -0.00000315, -0.00000377, -0.00000420, -0.00000385, -0.00000256, -0.00000075, 0.00000077, 0.00000125, 0.00000033, -0.00000177, -0.00000433, -0.00000647, -0.00000750, -0.00000720, -0.00000580, -0.00000384, -0.00000188, },
		{ -0.00000452, -0.00000385, -0.00000250, -0.00000087, 0.00000070, 0.00000197, 0.00000281, 0.00000313, 0.00000291, 0.00000215, 0.00000099, -0.00000032, -0.00000148, -0.00000224, -0.00000257, -0.00000266, -0.00000282, -0.00000332, -0.00000418, -0.00000507, -0.00000552, -0.00000510, -0.00000381, -0.00000218, -0.00000109, -0.00000137, -0.00000330, -0.00000625, -0.00000884, -0.00000947, -0.00000714, -0.00000212, 0.00000393, 0.00000854, 0.00000947, 0.00000590, -0.00000098, -0.00000829, -0.00001261, -0.00001162, -0.00000534, 0.00000358, 0.00001098, 0.00001301, 0.00000818, -0.00000167, -0.00001193, -0.00001721, -0.00001397, -0.00000254, 0.00001264, 0.00002471, 0.00002757, 0.00001899, 0.00000227, -0.00001468, -0.00002253, -0.00001475, 0.00000893, 0.00004155, 0.00007115, 0.00008530, 0.00007625, 0.00004435, -0.00000162, -0.00004763, -0.00007969, -0.00008910, -0.00007555, -0.00004677, -0.00001502, 0.00000794, 0.00001536, 0.00000756, -0.00000908, -0.00002548, -0.00003398, -0.00003140, -0.00001981, -0.00000488, 0.00000705, 0.00001184, 0.00000908, 0.00000171, -0.00000581, -0.00000980, -0.00000890, -0.00000438, 0.00000094, 0.00000428, 0.00000433, 0.00000168, -0.00000176, -0.00000388, -0.00000357, -0.00000115, 0.00000186, 0.00000372, 0.00000338, 0.00000097, -0.00000228, -0.00000481, -0.00000545, -0.00000403, -0.00000136, 0.00000129, 0.00000282, 0.00000281, 0.00000163, 0.00000015, -0.00000079, -0.00000082, -0.00000013, 0.00000072, 0.00000115, 0.00000094, 0.00000034, -0.00000013, 0.00000003, 0.00000097, 0.00000239, 0.00000366, 0.00000411, 0.00000341, 0.00000164, -0.00000066, -0.00000281, -0.00000419, },
		{ -0.00000163, -0.00000342, -0.00000524, -0.00000659, -0.00000716, -0.00000688, -0.00000589, -0.00000440, -0.00000266, -0.00000090, 0.00000062, 0.00000164, 0.00000191, 0.00000135, 0.00000009, -0.00000149, -0.00000290, -0.00000372, -0.00000381, -0.00000341, -0.00000297, -0.00000299, -0.00000365, -0.00000467, -0.00000546, -0.00000536, -0.00000413, -0.00000218, -0.00000046, -0.00000006, -0.00000160, -0.00000470, -0.00000798, -0.00000956, -0.00000799, -0.00000309, 0.00000368, 0.00000964, 0.00001200, 0.00000922, 0.00000199, -0.00000684, -0.00001337, -0.00001443, -0.00000921, 0.00000006, 0.00000888, 0.00001251, 0.00000828, -0.00000280, -0.00001616, -0.00002553, -0.00002590, -0.00001624, -0.00000054, 0.00001346, 0.00001748, 0.00000664, -0.00001775, -0.00004794, -0.00007231, -0.00007995, -0.00006525, -0.00003061, 0.00001407, 0.00005498, 0.00007966, 0.00008173, 0.00006313, 0.00003298, 0.00000366, -0.00001425, -0.00001599, -0.00000390, 0.00001428, 0.00002918, 0.00003398, 0.00002701, 0.00001194, -0.00000436, -0.00001518, -0.00001681, -0.00000988, 0.00000150, 0.00001191, 0.00001714, 0.00001589, 0.00000987, 0.00000253, -0.00000272, -0.00000413, -0.00000211, 0.00000137, 0.00000398, 0.00000429, 0.00000234, -0.00000061, -0.00000292, -0.00000350, -0.00000235, -0.00000047, 0.00000080, 0.00000053, -0.00000134, -0.00000402, -0.00000629, -0.00000720, -0.00000650, -0.00000470, -0.00000274, -0.00000152, -0.00000142, -0.00000223, -0.00000333, -0.00000410, -0.00000420, -0.00000375, -0.00000318, -0.00000292, -0.00000315, -0.00000367, -0.00000407, -0.00000394, -0.00000313, -0.00000187, -0.00000066, -0.00000006, -0.00000038, },
		{ -0.00000168, -0.00000048, 0.00000027, 0.00000028, -0.00000047, -0.00000175, -0.00000318, -0.00000439, -0.00000509, -0.00000517, -0.00000460, -0.00000342, -0.00000171, 0.00000033, 0.00000239, 0.00000404, 0.00000484, 0.00000454, 0.00000319, 0.00000124, -0.00000064, -0.00000181, -0.00000198, -0.00000130, -0.00000039, 0.00000006, -0.00000037, -0.00000154, -0.00000279, -0.00000325, -0.00000237, -0.00000031, 0.00000196, 0.00000312, 0.00000220, -0.00000079, -0.00000460, -0.00000716, -0.00000651, -0.00000189, 0.00000560, 0.00001322, 0.00001766, 0.00001662, 0.00001005, 0.00000056, -0.00000756, -0.00001019, -0.00000546, 0.00000509, 0.00001691, 0.00002430, 0.00002320, 0.00001339, -0.00000086, -0.00001228, -0.00001373, -0.00000165, 0.00002172, 0.00004858, 0.00006831, 0.00007173, 0.00005510, 0.00002202, -0.00001763, -0.00005135, -0.00006894, -0.00006639, -0.00004728, -0.00002106, 0.00000109, 0.00001083, 0.00000585, -0.00000965, -0.00002731, -0.00003854, -0.00003829, -0.00002697, -0.00000990, 0.00000544, 0.00001279, 0.00000983, -0.00000121, -0.00001483, -0.00002498, -0.00002768, -0.00002251, -0.00001235, -0.00000177, 0.00000517, 0.00000657, 0.00000318, -0.00000241, -0.00000720, -0.00000925, -0.00000830, -0.00000562, -0.00000303, -0.00000192, -0.00000261, -0.00000436, -0.00000597, -0.00000648, -0.00000571, -0.00000425, -0.00000309, -0.00000297, -0.00000406, -0.00000583, -0.00000742, -0.00000807, -0.00000752, -0.00000610, -0.00000450, -0.00000337, -0.00000305, -0.00000342, -0.00000402, -0.00000440, -0.00000430, -0.00000382, -0.00000329, -0.00000303, -0.00000318, -0.00000356, -0.00000383, -0.00000364, -0.00000287, },
		{ -0.00000134, -0.00000172, -0.00000169, -0.00000116, -0.00000026, 0.00000069, 0.00000131, 0.00000135, 0.00000077, -0.00000026, -0.00000151, -0.00000271, -0.00000369, -0.00000428, -0.00000437, -0.00000383, -0.00000257, -0.00000067, 0.00000158, 0.00000362, 0.00000482, 0.00000470, 0.00000317, 0.00000066, -0.00000201, -0.00000396, -0.00000464, -0.00000411, -0.00000300, -0.00000226, -0.00000258, -0.00000407, -0.00000607, -0.00000752, -0.00000754, -0.00000597, -0.00000364, -0.00000204, -0.00000257, -0.00000568, -0.00001041, -0.00001459, -0.00001580, -0.00001260, -0.00000554, 0.00000279, 0.00000861, 0.00000873, 0.00000217, -0.00000904, -0.00002039, -0.00002682, -0.00002510, -0.00001569, -0.00000297, 0.00000631, 0.00000600, -0.00000648, -0.00002839, -0.00005227, -0.00006862, -0.00006979, -0.00005330, -0.00002319, 0.00001135, 0.00003952, 0.00005316, 0.00004990, 0.00003388, 0.00001367, -0.00000154, -0.00000583, 0.00000110, 0.00001426, 0.00002598, 0.00002956, 0.00002237, 0.00000688, -0.00001060, -0.00002295, -0.00002544, -0.00001773, -0.00000387, 0.00000975, 0.00001727, 0.00001582, 0.00000658, -0.00000609, -0.00001679, -0.00002144, -0.00001885, -0.00001091, -0.00000138, 0.00000594, 0.00000878, 0.00000709, 0.00000268, -0.00000192, -0.00000476, -0.00000512, -0.00000366, -0.00000172, -0.00000052, -0.00000059, -0.00000160, -0.00000274, -0.00000326, -0.00000291, -0.00000200, -0.00000119, -0.00000105, -0.00000174, -0.00000294, -0.00000405, -0.00000451, -0.00000412, -0.00000310, -0.00000194, -0.00000112, -0.00000087, -0.00000109, -0.00000146, -0.00000164, -0.00000147, -0.00000106, -0.00000068, -0.00000059, -0.00000086, },
		{ -0.00000301, -0.00000279, -0.00000288, -0.00000332, -0.00000394, -0.00000445, -0.00000458, -0.00000424, -0.00000355, -0.00000279, -0.00000226, -0.00000216, -0.00000250, -0.00000318, -0.00000405, -0.00000498, -0.00000587, -0.00000665, -0.00000711, -0.00000700, -0.00000604, -0.00000414, -0.00000151, 0.00000124, 0.00000330, 0.00000394, 0.00000283, 0.00000033, -0.00000267, -0.00000504, -0.00000591, -0.00000514, -0.00000332, -0.00000159, -0.00000099, -0.00000191, -0.00000384, -0.00000560, -0.00000592, -0.00000424, -0.00000111, 0.00000191, 0.00000301, 0.00000118, -0.00000307, -0.00000765, -0.00000972, -0.00000704, 0.00000066, 0.00001115, 0.00002041, 0.00002429, 0.00002055, 0.00001032, -0.00000194, -0.00001007, -0.00000883, 0.00000352, 0.00002394, 0.00004539, 0.00005941, 0.00005952, 0.00004408, 0.00001711, -0.00001313, -0.00003736, -0.00004898, -0.00004652, -0.00003388, -0.00001835, -0.00000717, -0.00000443, -0.00000950, -0.00001776, -0.00002312, -0.00002116, -0.00001132, 0.00000282, 0.00001524, 0.00002037, 0.00001575, 0.00000317, -0.00001208, -0.00002363, -0.00002681, -0.00002066, -0.00000817, 0.00000517, 0.00001386, 0.00001471, 0.00000802, -0.00000283, -0.00001306, -0.00001863, -0.00001779, -0.00001155, -0.00000289, 0.00000475, 0.00000889, 0.00000892, 0.00000591, 0.00000188, -0.00000130, -0.00000266, -0.00000233, -0.00000119, -0.00000022, -0.00000001, -0.00000054, -0.00000133, -0.00000188, -0.00000192, -0.00000160, -0.00000131, -0.00000143, -0.00000207, -0.00000303, -0.00000389, -0.00000429, -0.00000413, -0.00000356, -0.00000294, -0.00000259, -0.00000264, -0.00000298, -0.00000333, -0.00000347, -0.00000332, },
		{ -0.00000243, -0.00000252, -0.00000233, -0.00000186, -0.00000129, -0.00000089, -0.00000085, -0.00000118, -0.00000170, -0.00000211, -0.00000217, -0.00000180, -0.00000112, -0.00000038, 0.00000018, 0.00000040, 0.00000031, 0.00000000, -0.00000043, -0.00000097, -0.00000166, -0.00000249, -0.00000330, -0.00000372, -0.00000330, -0.00000173, 0.00000094, 0.00000412, 0.00000686, 0.00000814, 0.00000737, 0.00000469, 0.00000101, -0.00000231, -0.00000408, -0.00000380, -0.00000198, 0.00000015, 0.00000119, 0.00000032, -0.00000221, -0.00000516, -0.00000694, -0.00000645, -0.00000385, -0.00000062, 0.00000109, -0.00000045, -0.00000539, -0.00001197, -0.00001704, -0.00001750, -0.00001201, -0.00000202, 0.00000838, 0.00001391, 0.00001045, -0.00000286, -0.00002275, -0.00004265, -0.00005506, -0.00005458, -0.00004015, -0.00001566, 0.00001149, 0.00003331, 0.00004426, 0.00004328, 0.00003375, 0.00002167, 0.00001275, 0.00000997, 0.00001254, 0.00001676, 0.00001823, 0.00001429, 0.00000550, -0.00000456, -0.00001114, -0.00001072, -0.00000290, 0.00000919, 0.00002022, 0.00002504, 0.00002104, 0.00000938, -0.00000549, -0.00001782, -0.00002300, -0.00001953, -0.00000953, 0.00000235, 0.00001101, 0.00001304, 0.00000803, -0.00000153, -0.00001153, -0.00001808, -0.00001907, -0.00001476, -0.00000739, 0.00000001, 0.00000495, 0.00000636, 0.00000471, 0.00000146, -0.00000171, -0.00000370, -0.00000425, -0.00000380, -0.00000306, -0.00000260, -0.00000258, -0.00000282, -0.00000302, -0.00000302, -0.00000285, -0.00000271, -0.00000282, -0.00000319, -0.00000366, -0.00000399, -0.00000396, -0.00000355, -0.00000295, -0.00000243, -0.00000218, -0.00000223, },
		{ -0.00000254, -0.00000231, -0.00000241, -0.00000278, -0.00000320, -0.00000338, -0.00000320, -0.00000269, -0.00000210, -0.00000172, -0.00000177, -0.00000224, -0.00000291, -0.00000348, -0.00000370, -0.00000349, -0.00000298, -0.00000242, -0.00000203, -0.00000190, -0.00000197, -0.00000210, -0.00000222, -0.00000240, -0.00000280, -0.00000357, -0.00000460, -0.00000551, -0.00000567, -0.00000452, -0.00000189, 0.00000174, 0.00000535, 0.00000766, 0.00000773, 0.00000542, 0.00000154, -0.00000238, -0.00000478, -0.00000473, -0.00000247, 0.00000071, 0.00000309, 0.00000334, 0.00000132, -0.00000184, -0.00000421, -0.00000410, -0.00000103, 0.00000380, 0.00000799, 0.00000907, 0.00000593, -0.00000032, -0.00000641, -0.00000820, -0.00000262, 0.00001067, 0.00002856, 0.00004533, 0.00005469, 0.00005233, 0.00003778, 0.00001471, -0.00001032, -0.00003044, -0.00004099, -0.00004113, -0.00003372, -0.00002367, -0.00001563, -0.00001197, -0.00001213, -0.00001340, -0.00001276, -0.00000867, -0.00000208, 0.00000411, 0.00000660, 0.00000359, -0.00000406, -0.00001301, -0.00001886, -0.00001828, -0.00001068, 0.00000132, 0.00001292, 0.00001926, 0.00001753, 0.00000830, -0.00000479, -0.00001655, -0.00002247, -0.00002062, -0.00001229, -0.00000133, 0.00000762, 0.00001108, 0.00000809, 0.00000033, -0.00000876, -0.00001556, -0.00001772, -0.00001493, -0.00000874, -0.00000168, 0.00000386, 0.00000650, 0.00000618, 0.00000389, 0.00000099, -0.00000138, -0.00000269, -0.00000300, -0.00000274, -0.00000235, -0.00000210, -0.00000202, -0.00000203, -0.00000207, -0.00000215, -0.00000235, -0.00000270, -0.00000314, -0.00000351, -0.00000363, -0.00000342, -0.00000299, },
		{ -0.00000341, -0.00000318, -0.00000261, -0.00000189, -0.00000129, -0.00000107, -0.00000128, -0.00000179, -0.00000228, -0.00000244, -0.00000212, -0.00000141, -0.00000060, -0.00000005, 0.00000000, -0.00000044, -0.00000116, -0.00000179, -0.00000207, -0.00000191, -0.00000146, -0.00000099, -0.00000071, -0.00000071, -0.00000084, -0.00000092, -0.00000083, -0.00000069, -0.00000080, -0.00000145, -0.00000270, -0.00000419, -0.00000519, -0.00000489, -0.00000282, 0.00000075, 0.00000481, 0.00000787, 0.00000857, 0.00000640, 0.00000194, -0.00000321, -0.00000710, -0.00000831, -0.00000667, -0.00000338, -0.00000046, 0.00000018, -0.00000223, -0.00000685, -0.00001154, -0.00001393, -0.00001268, -0.00000837, -0.00000351, -0.00000150, -0.00000503, -0.00001458, -0.00002776, -0.00004000, -0.00004622, -0.00004291, -0.00002973, -0.00000976, 0.00001158, 0.00002862, 0.00003755, 0.00003767, 0.00003131, 0.00002246, 0.00001486, 0.00001051, 0.00000912, 0.00000880, 0.00000750, 0.00000434, 0.00000016, -0.00000296, -0.00000299, 0.00000072, 0.00000682, 0.00001236, 0.00001424, 0.00001073, 0.00000259, -0.00000708, -0.00001415, -0.00001529, -0.00000962, 0.00000077, 0.00001164, 0.00001836, 0.00001796, 0.00001039, -0.00000146, -0.00001301, -0.00001997, -0.00002006, -0.00001388, -0.00000451, 0.00000396, 0.00000811, 0.00000655, 0.00000028, -0.00000794, -0.00001485, -0.00001802, -0.00001669, -0.00001180, -0.00000537, 0.00000041, 0.00000399, 0.00000491, 0.00000367, 0.00000130, -0.00000112, -0.00000290, -0.00000380, -0.00000395, -0.00000365, -0.00000317, -0.00000271, -0.00000235, -0.00000218, -0.00000223, -0.00000250, -0.00000291, -0.00000328, },
		{ -0.00000046, -0.00000102, -0.00000175, -0.00000237, -0.00000264, -0.00000240, -0.00000175, -0.00000097, -0.00000042, -0.00000038, -0.00000090, -0.00000179, -0.00000264, -0.00000308, -0.00000290, -0.00000221, -0.00000135, -0.00000075, -0.00000071, -0.00000126, -0.00000215, -0.00000299, -0.00000343, -0.00000338, -0.00000300, -0.00000258, -0.00000239, -0.00000250, -0.00000272, -0.00000276, -0.00000243, -0.00000184, -0.00000138, -0.00000153, -0.00000254, -0.00000418, -0.00000566, -0.00000596, -0.00000426, -0.00000049, 0.00000444, 0.00000891, 0.00001117, 0.00001019, 0.00000621, 0.00000077, -0.00000383, -0.00000558, -0.00000369, 0.00000094, 0.00000608, 0.00000917, 0.00000859, 0.00000461, -0.00000060, -0.00000379, -0.00000219, 0.00000514, 0.00001653, 0.00002817, 0.00003557, 0.00003534, 0.00002663, 0.00001155, -0.00000566, -0.00002030, -0.00002898, -0.00003073, -0.00002703, -0.00002079, -0.00001488, -0.00001096, -0.00000905, -0.00000799, -0.00000651, -0.00000413, -0.00000153, -0.00000006, -0.00000086, -0.00000391, -0.00000783, -0.00001036, -0.00000954, -0.00000485, 0.00000232, 0.00000908, 0.00001235, 0.00001032, 0.00000347, -0.00000544, -0.00001252, -0.00001444, -0.00000996, -0.00000058, 0.00001003, 0.00001757, 0.00001897, 0.00001365, 0.00000376, -0.00000677, -0.00001394, -0.00001530, -0.00001086, -0.00000293, 0.00000497, 0.00000961, 0.00000933, 0.00000455, -0.00000261, -0.00000930, -0.00001315, -0.00001307, -0.00000947, -0.00000389, 0.00000177, 0.00000591, 0.00000777, 0.00000743, 0.00000559, 0.00000315, 0.00000089, -0.00000072, -0.00000153, -0.00000167, -0.00000135, -0.00000084, -0.00000041, -0.00000024, },
		{ -0.00000474, -0.00000405, -0.00000331, -0.00000287, -0.00000290, -0.00000341, -0.00000416, -0.00000478, -0.00000493, -0.00000445, -0.00000344, -0.00000227, -0.00000138, -0.00000113, -0.00000161, -0.00000257, -0.00000354, -0.00000403, -0.00000378, -0.00000287, -0.00000169, -0.00000076, -0.00000046, -0.00000089, -0.00000177, -0.00000268, -0.00000319, -0.00000314, -0.00000270, -0.00000222, -0.00000204, -0.00000225, -0.00000266, -0.00000285, -0.00000253, -0.00000171, -0.00000083, -0.00000055, -0.00000139, -0.00000333, -0.00000564, -0.00000710, -0.00000654, -0.00000347, 0.00000147, 0.00000663, 0.00000987, 0.00000956, 0.00000537, -0.00000140, -0.00000822, -0.00001239, -0.00001231, -0.00000827, -0.00000257, 0.00000141, 0.00000068, -0.00000596, -0.00001715, -0.00002933, -0.00003801, -0.00003955, -0.00003257, -0.00001847, -0.00000091, 0.00001568, 0.00002766, 0.00003334, 0.00003322, 0.00002929, 0.00002396, 0.00001898, 0.00001501, 0.00001182, 0.00000892, 0.00000621, 0.00000414, 0.00000346, 0.00000456, 0.00000700, 0.00000949, 0.00001039, 0.00000861, 0.00000432, -0.00000093, -0.00000481, -0.00000527, -0.00000167, 0.00000474, 0.00001124, 0.00001475, 0.00001331, 0.00000703, -0.00000177, -0.00000952, -0.00001292, -0.00001042, -0.00000295, 0.00000643, 0.00001385, 0.00001622, 0.00001254, 0.00000428, -0.00000534, -0.00001270, -0.00001530, -0.00001269, -0.00000651, 0.00000032, 0.00000488, 0.00000536, 0.00000175, -0.00000442, -0.00001076, -0.00001506, -0.00001602, -0.00001363, -0.00000895, -0.00000359, 0.00000094, 0.00000365, 0.00000428, 0.00000318, 0.00000105, -0.00000132, -0.00000333, -0.00000459, -0.00000502, },
		{ 0.00000047, -0.00000180, -0.00000348, -0.00000428, -0.00000418, -0.00000346, -0.00000255, -0.00000191, -0.00000182, -0.00000234, -0.00000320, -0.00000398, -0.00000423, -0.00000373, -0.00000256, -0.00000113, -0.00000000, 0.00000037, -0.00000016, -0.00000133, -0.00000260, -0.00000338, -0.00000329, -0.00000236, -0.00000100, 0.00000018, 0.00000068, 0.00000033, -0.00000064, -0.00000172, -0.00000239, -0.00000241, -0.00000190, -0.00000127, -0.00000097, -0.00000118, -0.00000170, -0.00000202, -0.00000167, -0.00000052, 0.00000104, 0.00000219, 0.00000213, 0.00000058, -0.00000192, -0.00000407, -0.00000430, -0.00000159, 0.00000388, 0.00001061, 0.00001616, 0.00001820, 0.00001562, 0.00000920, 0.00000156, -0.00000378, -0.00000382, 0.00000252, 0.00001375, 0.00002617, 0.00003518, 0.00003698, 0.00002999, 0.00001549, -0.00000291, -0.00002059, -0.00003355, -0.00003960, -0.00003875, -0.00003282, -0.00002439, -0.00001583, -0.00000866, -0.00000343, -0.00000014, 0.00000140, 0.00000128, -0.00000039, -0.00000322, -0.00000638, -0.00000869, -0.00000909, -0.00000717, -0.00000355, 0.00000021, 0.00000230, 0.00000150, -0.00000213, -0.00000708, -0.00001103, -0.00001182, -0.00000861, -0.00000235, 0.00000449, 0.00000893, 0.00000888, 0.00000408, -0.00000364, -0.00001110, -0.00001513, -0.00001394, -0.00000792, 0.00000050, 0.00000794, 0.00001143, 0.00000966, 0.00000349, -0.00000454, -0.00001130, -0.00001436, -0.00001295, -0.00000813, -0.00000222, 0.00000222, 0.00000342, 0.00000100, -0.00000400, -0.00000965, -0.00001395, -0.00001554, -0.00001407, -0.00001018, -0.00000515, -0.00000036, 0.00000310, 0.00000468, 0.00000442, 0.00000278, },
		{ 0.00000312, 0.00000540, 0.00000584, 0.00000467, 0.00000245, -0.00000008, -0.00000224, -0.00000352, -0.00000379, -0.00000322, -0.00000229, -0.00000153, -0.00000136, -0.00000192, -0.00000296, -0.00000399, -0.00000447, -0.00000408, -0.00000287, -0.00000126, 0.00000009, 0.00000060, 0.00000004, -0.00000139, -0.00000308, -0.00000431, -0.00000457, -0.00000379, -0.00000239, -0.00000106, -0.00000044, -0.00000083, -0.00000206, -0.00000356, -0.00000470, -0.00000509, -0.00000480, -0.00000429, -0.00000413, -0.00000464, -0.00000568, -0.00000670, -0.00000701, -0.00000626, -0.00000470, -0.00000324, -0.00000296, -0.00000460, -0.00000795, -0.00001177, -0.00001419, -0.00001356, -0.00000935, -0.00000267, 0.00000387, 0.00000709, 0.00000456, -0.00000424, -0.00001740, -0.00003104, -0.00004060, -0.00004246, -0.00003534, -0.00002076, -0.00000250, 0.00001469, 0.00002675, 0.00003149, 0.00002904, 0.00002148, 0.00001179, 0.00000281, -0.00000357, -0.00000663, -0.00000661, -0.00000430, -0.00000063, 0.00000349, 0.00000719, 0.00000963, 0.00001014, 0.00000843, 0.00000487, 0.00000053, -0.00000308, -0.00000461, -0.00000348, -0.00000025, 0.00000352, 0.00000591, 0.00000552, 0.00000221, -0.00000273, -0.00000708, -0.00000866, -0.00000642, -0.00000097, 0.00000553, 0.00001032, 0.00001122, 0.00000763, 0.00000088, -0.00000629, -0.00001095, -0.00001118, -0.00000689, 0.00000012, 0.00000698, 0.00001096, 0.00001058, 0.00000619, -0.00000033, -0.00000636, -0.00000969, -0.00000937, -0.00000597, -0.00000126, 0.00000261, 0.00000400, 0.00000232, -0.00000180, -0.00000687, -0.00001117, -0.00001334, -0.00001282, -0.00000988, -0.00000545, -0.00000075, },
		{ -0.00001349, -0.00001143, -0.00000767, -0.00000319, 0.00000094, 0.00000382, 0.00000498, 0.00000438, 0.00000242, -0.00000022, -0.00000273, -0.00000447, -0.00000511, -0.00000473, -0.00000376, -0.00000283, -0.00000249, -0.00000296, -0.00000405, -0.00000526, -0.00000594, -0.00000565, -0.00000433, -0.00000243, -0.00000066, 0.00000026, -0.00000006, -0.00000147, -0.00000335, -0.00000485, -0.00000530, -0.00000447, -0.00000274, -0.00000088, 0.00000032, 0.00000037, -0.00000063, -0.00000210, -0.00000327, -0.00000357, -0.00000295, -0.00000187, -0.00000103, -0.00000098, -0.00000173, -0.00000268, -0.00000296, -0.00000187, 0.00000060, 0.00000361, 0.00000580, 0.00000593, 0.00000364, -0.00000014, -0.00000332, -0.00000345, 0.00000121, 0.00001068, 0.00002294, 0.00003442, 0.00004113, 0.00004018, 0.00003088, 0.00001512, -0.00000325, -0.00001973, -0.00003060, -0.00003410, -0.00003075, -0.00002288, -0.00001363, -0.00000582, -0.00000117, -0.00000004, -0.00000172, -0.00000487, -0.00000813, -0.00001040, -0.00001101, -0.00000968, -0.00000655, -0.00000213, 0.00000267, 0.00000667, 0.00000878, 0.00000837, 0.00000560, 0.00000151, -0.00000228, -0.00000423, -0.00000354, -0.00000056, 0.00000330, 0.00000613, 0.00000643, 0.00000382, -0.00000066, -0.00000500, -0.00000710, -0.00000572, -0.00000114, 0.00000486, 0.00000979, 0.00001147, 0.00000907, 0.00000343, -0.00000321, -0.00000821, -0.00000963, -0.00000701, -0.00000158, 0.00000436, 0.00000841, 0.00000906, 0.00000624, 0.00000126, -0.00000381, -0.00000707, -0.00000750, -0.00000533, -0.00000183, 0.00000130, 0.00000258, 0.00000137, -0.00000203, -0.00000652, -0.00001068, -0.00001324, },
		{ -0.00000131, -0.00000513, -0.00000892, -0.00001158, -0.00001234, -0.00001095, -0.00000775, -0.00000349, 0.00000086, 0.00000435, 0.00000630, 0.00000644, 0.00000496, 0.00000246, -0.00000022, -0.00000229, -0.00000325, -0.00000304, -0.00000202, -0.00000087, -0.00000025, -0.00000052, -0.00000159, -0.00000295, -0.00000387, -0.00000377, -0.00000248, -0.00000037, 0.00000178, 0.00000310, 0.00000303, 0.00000158, -0.00000064, -0.00000267, -0.00000364, -0.00000316, -0.00000148, 0.00000058, 0.00000203, 0.00000220, 0.00000100, -0.00000100, -0.00000289, -0.00000388, -0.00000372, -0.00000281, -0.00000203, -0.00000218, -0.00000355, -0.00000565, -0.00000743, -0.00000777, -0.00000621, -0.00000331, -0.00000063, -0.00000010, -0.00000312, -0.00000975, -0.00001839, -0.00002617, -0.00002997, -0.00002756, -0.00001861, -0.00000489, 0.00001024, 0.00002302, 0.00003052, 0.00003156, 0.00002697, 0.00001904, 0.00001062, 0.00000408, 0.00000067, 0.00000034, 0.00000207, 0.00000444, 0.00000619, 0.00000655, 0.00000534, 0.00000285, -0.00000029, -0.00000333, -0.00000544, -0.00000598, -0.00000458, -0.00000141, 0.00000278, 0.00000680, 0.00000933, 0.00000946, 0.00000709, 0.00000308, -0.00000104, -0.00000366, -0.00000379, -0.00000154, 0.00000189, 0.00000473, 0.00000544, 0.00000346, -0.00000049, -0.00000469, -0.00000714, -0.00000653, -0.00000286, 0.00000250, 0.00000738, 0.00000971, 0.00000846, 0.00000408, -0.00000170, -0.00000662, -0.00000880, -0.00000757, -0.00000361, 0.00000132, 0.00000523, 0.00000668, 0.00000534, 0.00000201, -0.00000178, -0.00000448, -0.00000516, -0.00000380, -0.00000127, 0.00000116, 0.00000227, 0.00000141, },
		{ -0.00000053, 0.00000108, 0.00000176, 0.00000095, -0.00000137, -0.00000470, -0.00000820, -0.00001094, -0.00001216, -0.00001147, -0.00000896, -0.00000513, -0.00000081, 0.00000305, 0.00000563, 0.00000643, 0.00000542, 0.00000307, 0.00000021, -0.00000229, -0.00000371, -0.00000384, -0.00000294, -0.00000169, -0.00000084, -0.00000093, -0.00000198, -0.00000353, -0.00000478, -0.00000499, -0.00000384, -0.00000159, 0.00000097, 0.00000283, 0.00000323, 0.00000199, -0.00000035, -0.00000276, -0.00000415, -0.00000387, -0.00000203, 0.00000058, 0.00000280, 0.00000368, 0.00000292, 0.00000098, -0.00000107, -0.00000217, -0.00000174, -0.00000006, 0.00000193, 0.00000305, 0.00000259, 0.00000076, -0.00000127, -0.00000182, 0.00000049, 0.00000598, 0.00001355, 0.00002094, 0.00002550, 0.00002522, 0.00001955, 0.00000969, -0.00000182, -0.00001202, -0.00001855, -0.00002039, -0.00001811, -0.00001340, -0.00000837, -0.00000472, -0.00000321, -0.00000364, -0.00000513, -0.00000660, -0.00000724, -0.00000675, -0.00000528, -0.00000334, -0.00000150, -0.00000029, -0.00000003, -0.00000085, -0.00000257, -0.00000475, -0.00000666, -0.00000752, -0.00000670, -0.00000403, -0.00000002, 0.00000426, 0.00000747, 0.00000854, 0.00000713, 0.00000384, 0.00000004, -0.00000271, -0.00000331, -0.00000164, 0.00000137, 0.00000417, 0.00000525, 0.00000394, 0.00000066, -0.00000316, -0.00000573, -0.00000574, -0.00000295, 0.00000163, 0.00000620, 0.00000888, 0.00000857, 0.00000536, 0.00000051, -0.00000412, -0.00000685, -0.00000685, -0.00000441, -0.00000073, 0.00000263, 0.00000442, 0.00000415, 0.00000226, -0.00000025, -0.00000222, -0.00000290, -0.00000216, },
		{ 0.00000133, 0.00000033, 0.00000002, 0.00000051, 0.00000149, 0.00000238, 0.00000258, 0.00000167, -0.00000039, -0.00000331, -0.00000648, -0.00000917, -0.00001070, -0.00001060, -0.00000874, -0.00000540, -0.00000122, 0.00000291, 0.00000607, 0.00000757, 0.00000715, 0.00000510, 0.00000217, -0.00000068, -0.00000258, -0.00000312, -0.00000242, -0.00000112, -0.00000005, 0.00000009, -0.00000089, -0.00000263, -0.00000429, -0.00000500, -0.00000422, -0.00000205, 0.00000074, 0.00000303, 0.00000384, 0.00000273, 0.00000010, -0.00000300, -0.00000526, -0.00000575, -0.00000431, -0.00000168, 0.00000084, 0.00000198, 0.00000109, -0.00000156, -0.00000483, -0.00000731, -0.00000800, -0.00000682, -0.00000478, -0.00000353, -0.00000460, -0.00000865, -0.00001499, -0.00002171, -0.00002635, -0.00002680, -0.00002216, -0.00001310, -0.00000161, 0.00000968, 0.00001842, 0.00002325, 0.00002406, 0.00002184, 0.00001808, 0.00001418, 0.00001099, 0.00000873, 0.00000716, 0.00000584, 0.00000448, 0.00000300, 0.00000156, 0.00000038, -0.00000035, -0.00000058, -0.00000034, 0.00000022, 0.00000092, 0.00000149, 0.00000163, 0.00000107, -0.00000030, -0.00000231, -0.00000444, -0.00000591, -0.00000600, -0.00000432, -0.00000111, 0.00000273, 0.00000597, 0.00000748, 0.00000674, 0.00000409, 0.00000066, -0.00000213, -0.00000317, -0.00000216, 0.00000027, 0.00000279, 0.00000404, 0.00000328, 0.00000070, -0.00000258, -0.00000504, -0.00000543, -0.00000337, 0.00000046, 0.00000463, 0.00000752, 0.00000802, 0.00000598, 0.00000223, -0.00000179, -0.00000467, -0.00000554, -0.00000439, -0.00000195, 0.00000068, 0.00000252, 0.00000308, 0.00000249, },
		{ -0.00000074, 0.00000071, 0.00000162, 0.00000196, 0.00000196, 0.00000193, 0.00000213, 0.00000256, 0.00000304, 0.00000325, 0.00000285, 0.00000164, -0.00000039, -0.00000306, -0.00000597, -0.00000859, -0.00001036, -0.00001079, -0.00000959, -0.00000681, -0.00000291, 0.00000134, 0.00000499, 0.00000718, 0.00000744, 0.00000584, 0.00000301, -0.00000008, -0.00000243, -0.00000339, -0.00000292, -0.00000154, -0.00000016, 0.00000040, -0.00000029, -0.00000198, -0.00000389, -0.00000498, -0.00000449, -0.00000232, 0.00000090, 0.00000397, 0.00000567, 0.00000526, 0.00000292, -0.00000035, -0.00000305, -0.00000390, -0.00000237, 0.00000099, 0.00000480, 0.00000742, 0.00000771, 0.00000562, 0.00000225, -0.00000053, -0.00000088, 0.00000210, 0.00000796, 0.00001489, 0.00002037, 0.00002207, 0.00001876, 0.00001078, -0.00000012, -0.00001133, -0.00002029, -0.00002525, -0.00002568, -0.00002223, -0.00001633, -0.00000963, -0.00000352, 0.00000114, 0.00000404, 0.00000527, 0.00000516, 0.00000411, 0.00000253, 0.00000084, -0.00000064, -0.00000165, -0.00000211, -0.00000208, -0.00000173, -0.00000123, -0.00000068, -0.00000009, 0.00000057, 0.00000124, 0.00000177, 0.00000187, 0.00000128, -0.00000001, -0.00000168, -0.00000309, -0.00000348, -0.00000237, 0.00000023, 0.00000368, 0.00000689, 0.00000878, 0.00000874, 0.00000687, 0.00000404, 0.00000146, 0.00000019, 0.00000063, 0.00000239, 0.00000443, 0.00000560, 0.00000514, 0.00000309, 0.00000030, -0.00000201, -0.00000273, -0.00000142, 0.00000152, 0.00000501, 0.00000771, 0.00000863, 0.00000745, 0.00000465, 0.00000125, -0.00000165, -0.00000327, -0.00000338, -0.00000232, },
		{ -0.00000154, -0.00000343, -0.00000439, -0.00000447, -0.00000389, -0.00000295, -0.00000187, -0.00000074, 0.00000040, 0.00000153, 0.00000257, 0.00000335, 0.00000371, 0.00000350, 0.00000264, 0.00000114, -0.00000091, -0.00000335, -0.00000596, -0.00000839, -0.00001023, -0.00001103, -0.00001039, -0.00000818, -0.00000463, -0.00000037, 0.00000368, 0.00000656, 0.00000758, 0.00000655, 0.00000394, 0.00000067, -0.00000214, -0.00000366, -0.00000359, -0.00000232, -0.00000074, 0.00000016, -0.00000026, -0.00000197, -0.00000427, -0.00000604, -0.00000627, -0.00000458, -0.00000140, 0.00000206, 0.00000434, 0.00000437, 0.00000195, -0.00000207, -0.00000607, -0.00000837, -0.00000795, -0.00000499, -0.00000089, 0.00000225, 0.00000249, -0.00000111, -0.00000797, -0.00001610, -0.00002274, -0.00002536, -0.00002255, -0.00001457, -0.00000322, 0.00000871, 0.00001844, 0.00002395, 0.00002456, 0.00002094, 0.00001471, 0.00000788, 0.00000222, -0.00000110, -0.00000173, -0.00000000, 0.00000326, 0.00000703, 0.00001030, 0.00001230, 0.00001265, 0.00001139, 0.00000894, 0.00000599, 0.00000324, 0.00000123, 0.00000017, -0.00000005, 0.00000021, 0.00000058, 0.00000083, 0.00000094, 0.00000101, 0.00000117, 0.00000137, 0.00000139, 0.00000094, -0.00000012, -0.00000165, -0.00000317, -0.00000400, -0.00000362, -0.00000188, 0.00000080, 0.00000356, 0.00000545, 0.00000582, 0.00000460, 0.00000239, 0.00000017, -0.00000117, -0.00000116, -0.00000002, 0.00000149, 0.00000244, 0.00000218, 0.00000067, -0.00000153, -0.00000346, -0.00000421, -0.00000334, -0.00000108, 0.00000180, 0.00000429, 0.00000555, 0.00000521, 0.00000348, 0.00000097, },
		{ 0.00000412, 0.00000436, 0.00000353, 0.00000192, -0.00000005, -0.00000199, -0.00000361, -0.00000473, -0.00000525, -0.00000509, -0.00000420, -0.00000265, -0.00000062, 0.00000158, 0.00000353, 0.00000485, 0.00000530, 0.00000482, 0.00000353, 0.00000166, -0.00000054, -0.00000290, -0.00000527, -0.00000750, -0.00000933, -0.00001039, -0.00001026, -0.00000864, -0.00000554, -0.00000140, 0.00000295, 0.00000650, 0.00000835, 0.00000808, 0.00000592, 0.00000269, -0.00000046, -0.00000246, -0.00000279, -0.00000160, 0.00000027, 0.00000174, 0.00000193, 0.00000058, -0.00000175, -0.00000392, -0.00000469, -0.00000331, 0.00000008, 0.00000440, 0.00000801, 0.00000943, 0.00000797, 0.00000418, -0.00000031, -0.00000332, -0.00000301, 0.00000128, 0.00000867, 0.00001691, 0.00002314, 0.00002485, 0.00002082, 0.00001157, -0.00000082, -0.00001336, -0.00002313, -0.00002812, -0.00002779, -0.00002306, -0.00001587, -0.00000847, -0.00000282, -0.00000000, -0.00000016, -0.00000254, -0.00000589, -0.00000882, -0.00001023, -0.00000955, -0.00000686, -0.00000285, 0.00000146, 0.00000496, 0.00000680, 0.00000668, 0.00000484, 0.00000200, -0.00000091, -0.00000312, -0.00000421, -0.00000425, -0.00000362, -0.00000281, -0.00000218, -0.00000185, -0.00000168, -0.00000150, -0.00000123, -0.00000098, -0.00000101, -0.00000152, -0.00000252, -0.00000368, -0.00000448, -0.00000440, -0.00000321, -0.00000114, 0.00000122, 0.00000307, 0.00000382, 0.00000331, 0.00000190, 0.00000028, -0.00000082, -0.00000099, -0.00000033, 0.00000066, 0.00000130, 0.00000110, -0.00000002, -0.00000165, -0.00000312, -0.00000374, -0.00000318, -0.00000153, 0.00000070, 0.00000279, },
		{ -0.00000190, -0.00000074, 0.00000084, 0.00000242, 0.00000360, 0.00000418, 0.00000408, 0.00000333, 0.00000201, 0.00000025, -0.00000177, -0.00000379, -0.00000541, -0.00000622, -0.00000587, -0.00000428, -0.00000168, 0.00000140, 0.00000425, 0.00000622, 0.00000690, 0.00000624, 0.00000450, 0.00000214, -0.00000044, -0.00000296, -0.00000532, -0.00000749, -0.00000940, -0.00001079, -0.00001127, -0.00001043, -0.00000808, -0.00000443, -0.00000017, 0.00000371, 0.00000617, 0.00000653, 0.00000473, 0.00000142, -0.00000227, -0.00000511, -0.00000629, -0.00000569, -0.00000399, -0.00000234, -0.00000184, -0.00000309, -0.00000581, -0.00000894, -0.00001104, -0.00001097, -0.00000842, -0.00000419, -0.00000001, 0.00000211, 0.00000073, -0.00000444, -0.00001218, -0.00002014, -0.00002557, -0.00002631, -0.00002153, -0.00001211, -0.00000034, 0.00001080, 0.00001861, 0.00002148, 0.00001928, 0.00001328, 0.00000565, -0.00000131, -0.00000581, -0.00000708, -0.00000540, -0.00000188, 0.00000195, 0.00000470, 0.00000547, 0.00000413, 0.00000122, -0.00000215, -0.00000477, -0.00000566, -0.00000440, -0.00000128, 0.00000281, 0.00000666, 0.00000915, 0.00000964, 0.00000812, 0.00000523, 0.00000193, -0.00000083, -0.00000246, -0.00000286, -0.00000238, -0.00000155, -0.00000084, -0.00000047, -0.00000038, -0.00000037, -0.00000027, -0.00000009, -0.00000002, -0.00000026, -0.00000090, -0.00000178, -0.00000255, -0.00000276, -0.00000217, -0.00000083, 0.00000087, 0.00000238, 0.00000322, 0.00000319, 0.00000248, 0.00000151, 0.00000077, 0.00000056, 0.00000083, 0.00000127, 0.00000147, 0.00000113, 0.00000022, -0.00000096, -0.00000195, -0.00000234, },
		{ -0.00000086, -0.00000177, -0.00000237, -0.00000249, -0.00000209, -0.00000130, -0.00000035, 0.00000060, 0.00000143, 0.00000208, 0.00000249, 0.00000250, 0.00000192, 0.00000057, -0.00000156, -0.00000422, -0.00000681, -0.00000863, -0.00000900, -0.00000760, -0.00000463, -0.00000074, 0.00000311, 0.00000599, 0.00000726, 0.00000679, 0.00000493, 0.00000228, -0.00000053, -0.00000310, -0.00000528, -0.00000716, -0.00000879, -0.00001011, -0.00001079, -0.00001037, -0.00000850, -0.00000514, -0.00000076, 0.00000371, 0.00000716, 0.00000868, 0.00000790, 0.00000521, 0.00000164, -0.00000148, -0.00000305, -0.00000262, -0.00000058, 0.00000196, 0.00000369, 0.00000364, 0.00000167, -0.00000137, -0.00000394, -0.00000439, -0.00000167, 0.00000410, 0.00001159, 0.00001857, 0.00002267, 0.00002223, 0.00001690, 0.00000779, -0.00000282, -0.00001228, -0.00001839, -0.00002002, -0.00001746, -0.00001214, -0.00000610, -0.00000131, 0.00000096, 0.00000040, -0.00000230, -0.00000586, -0.00000883, -0.00001010, -0.00000921, -0.00000646, -0.00000277, 0.00000067, 0.00000278, 0.00000294, 0.00000126, -0.00000153, -0.00000426, -0.00000577, -0.00000529, -0.00000275, 0.00000121, 0.00000542, 0.00000864, 0.00000998, 0.00000917, 0.00000667, 0.00000340, 0.00000039, -0.00000162, -0.00000234, -0.00000202, -0.00000117, -0.00000032, 0.00000021, 0.00000038, 0.00000036, 0.00000036, 0.00000046, 0.00000058, 0.00000056, 0.00000026, -0.00000030, -0.00000090, -0.00000127, -0.00000117, -0.00000057, 0.00000037, 0.00000132, 0.00000198, 0.00000221, 0.00000205, 0.00000172, 0.00000143, 0.00000129, 0.00000126, 0.00000115, 0.00000078, 0.00000007, },
		{ 0.00000177, 0.00000137, 0.00000057, -0.00000045, -0.00000142, -0.00000205, -0.00000216, -0.00000179, -0.00000111, -0.00000038, 0.00000022, 0.00000069, 0.00000117, 0.00000182, 0.00000265, 0.00000344, 0.00000374, 0.00000308, 0.00000119, -0.00000175, -0.00000510, -0.00000787, -0.00000910, -0.00000817, -0.00000512, -0.00000065, 0.00000409, 0.00000789, 0.00000987, 0.00000973, 0.00000778, 0.00000477, 0.00000152, -0.00000139, -0.00000371, -0.00000555, -0.00000713, -0.00000855, -0.00000962, -0.00000991, -0.00000893, -0.00000642, -0.00000258, 0.00000182, 0.00000566, 0.00000785, 0.00000769, 0.00000523, 0.00000129, -0.00000279, -0.00000568, -0.00000651, -0.00000531, -0.00000295, -0.00000090, -0.00000061, -0.00000287, -0.00000749, -0.00001322, -0.00001818, -0.00002048, -0.00001895, -0.00001353, -0.00000537, 0.00000357, 0.00001115, 0.00001574, 0.00001666, 0.00001438, 0.00001017, 0.00000567, 0.00000232, 0.00000089, 0.00000134, 0.00000296, 0.00000465, 0.00000542, 0.00000467, 0.00000241, -0.00000077, -0.00000391, -0.00000602, -0.00000644, -0.00000503, -0.00000230, 0.00000080, 0.00000317, 0.00000396, 0.00000291, 0.00000041, -0.00000254, -0.00000473, -0.00000519, -0.00000355, -0.00000018, 0.00000389, 0.00000741, 0.00000930, 0.00000907, 0.00000697, 0.00000378, 0.00000053, -0.00000190, -0.00000308, -0.00000310, -0.00000239, -0.00000150, -0.00000084, -0.00000055, -0.00000055, -0.00000062, -0.00000060, -0.00000047, -0.00000032, -0.00000028, -0.00000041, -0.00000069, -0.00000099, -0.00000116, -0.00000113, -0.00000090, -0.00000054, -0.00000014, 0.00000025, 0.00000066, 0.00000108, 0.00000149, 0.00000177, },
		{ -0.00000037, 0.00000066, 0.00000174, 0.00000250, 0.00000262, 0.00000196, 0.00000066, -0.00000088, -0.00000218, -0.00000282, -0.00000268, -0.00000194, -0.00000100, -0.00000031, -0.00000006, -0.00000016, -0.00000025, 0.00000007, 0.00000097, 0.00000223, 0.00000326, 0.00000330, 0.00000180, -0.00000125, -0.00000521, -0.00000895, -0.00001119, -0.00001103, -0.00000824, -0.00000346, 0.00000205, 0.00000683, 0.00000971, 0.00001017, 0.00000842, 0.00000527, 0.00000170, -0.00000146, -0.00000383, -0.00000548, -0.00000673, -0.00000785, -0.00000885, -0.00000939, -0.00000891, -0.00000694, -0.00000342, 0.00000116, 0.00000579, 0.00000926, 0.00001059, 0.00000940, 0.00000618, 0.00000212, -0.00000125, -0.00000260, -0.00000128, 0.00000239, 0.00000722, 0.00001153, 0.00001372, 0.00001285, 0.00000890, 0.00000282, -0.00000389, -0.00000959, -0.00001310, -0.00001399, -0.00001259, -0.00000984, -0.00000683, -0.00000450, -0.00000330, -0.00000317, -0.00000364, -0.00000407, -0.00000397, -0.00000312, -0.00000170, -0.00000022, 0.00000074, 0.00000067, -0.00000060, -0.00000282, -0.00000536, -0.00000738, -0.00000815, -0.00000730, -0.00000499, -0.00000194, 0.00000084, 0.00000241, 0.00000222, 0.00000041, -0.00000224, -0.00000461, -0.00000559, -0.00000459, -0.00000172, 0.00000220, 0.00000597, 0.00000843, 0.00000891, 0.00000741, 0.00000457, 0.00000134, -0.00000134, -0.00000290, -0.00000325, -0.00000270, -0.00000179, -0.00000097, -0.00000050, -0.00000038, -0.00000045, -0.00000050, -0.00000040, -0.00000013, 0.00000020, 0.00000047, 0.00000057, 0.00000046, 0.00000014, -0.00000033, -0.00000084, -0.00000124, -0.00000136, -0.00000108, },
		{ -0.00000038, -0.00000155, -0.00000233, -0.00000235, -0.00000149, 0.00000010, 0.00000192, 0.00000335, 0.00000384, 0.00000318, 0.00000154, -0.00000047, -0.00000214, -0.00000288, -0.00000253, -0.00000136, -0.00000000, 0.00000092, 0.00000108, 0.00000059, -0.00000002, -0.00000011, 0.00000072, 0.00000233, 0.00000406, 0.00000493, 0.00000409, 0.00000126, -0.00000303, -0.00000758, -0.00001087, -0.00001167, -0.00000946, -0.00000470, 0.00000133, 0.00000697, 0.00001076, 0.00001189, 0.00001041, 0.00000708, 0.00000305, -0.00000066, -0.00000346, -0.00000530, -0.00000653, -0.00000761, -0.00000876, -0.00000981, -0.00001024, -0.00000944, -0.00000707, -0.00000328, 0.00000116, 0.00000506, 0.00000718, 0.00000666, 0.00000337, -0.00000199, -0.00000807, -0.00001326, -0.00001616, -0.00001605, -0.00001299, -0.00000780, -0.00000178, 0.00000377, 0.00000782, 0.00000991, 0.00001012, 0.00000896, 0.00000710, 0.00000515, 0.00000355, 0.00000242, 0.00000171, 0.00000124, 0.00000085, 0.00000044, 0.00000006, -0.00000017, -0.00000010, 0.00000033, 0.00000108, 0.00000189, 0.00000240, 0.00000223, 0.00000114, -0.00000081, -0.00000322, -0.00000543, -0.00000671, -0.00000656, -0.00000494, -0.00000230, 0.00000047, 0.00000240, 0.00000278, 0.00000152, -0.00000085, -0.00000331, -0.00000476, -0.00000442, -0.00000218, 0.00000135, 0.00000509, 0.00000790, 0.00000896, 0.00000805, 0.00000560, 0.00000248, -0.00000040, -0.00000235, -0.00000310, -0.00000284, -0.00000201, -0.00000111, -0.00000048, -0.00000023, -0.00000027, -0.00000039, -0.00000036, -0.00000006, 0.00000049, 0.00000114, 0.00000169, 0.00000189, 0.00000160, 0.00000078, },
		{ 0.00000342, 0.00000401, 0.00000387, 0.00000284, 0.00000109, -0.00000091, -0.00000252, -0.00000310, -0.00000237, -0.00000046, 0.00000200, 0.00000414, 0.00000513, 0.00000459, 0.00000268, 0.00000014, -0.00000208, -0.00000317, -0.00000284, -0.00000141, 0.00000034, 0.00000155, 0.00000171, 0.00000089, -0.00000028, -0.00000095, -0.00000046, 0.00000124, 0.00000349, 0.00000516, 0.00000510, 0.00000276, -0.00000158, -0.00000672, -0.00001099, -0.00001283, -0.00001139, -0.00000687, -0.00000046, 0.00000606, 0.00001097, 0.00001313, 0.00001234, 0.00000927, 0.00000511, 0.00000111, -0.00000190, -0.00000367, -0.00000451, -0.00000499, -0.00000555, -0.00000627, -0.00000679, -0.00000646, -0.00000468, -0.00000116, 0.00000379, 0.00000935, 0.00001434, 0.00001758, 0.00001825, 0.00001611, 0.00001154, 0.00000545, -0.00000103, -0.00000681, -0.00001107, -0.00001340, -0.00001378, -0.00001253, -0.00001015, -0.00000723, -0.00000432, -0.00000186, -0.00000014, 0.00000074, 0.00000085, 0.00000041, -0.00000030, -0.00000098, -0.00000146, -0.00000163, -0.00000149, -0.00000111, -0.00000052, 0.00000022, 0.00000106, 0.00000188, 0.00000246, 0.00000252, 0.00000183, 0.00000035, -0.00000170, -0.00000378, -0.00000524, -0.00000550, -0.00000434, -0.00000205, 0.00000067, 0.00000288, 0.00000381, 0.00000314, 0.00000121, -0.00000113, -0.00000281, -0.00000299, -0.00000136, 0.00000171, 0.00000531, 0.00000832, 0.00000986, 0.00000953, 0.00000757, 0.00000465, 0.00000167, -0.00000061, -0.00000180, -0.00000190, -0.00000125, -0.00000033, 0.00000044, 0.00000082, 0.00000083, 0.00000066, 0.00000057, 0.00000082, 0.00000148, 0.00000244, },
		{ -0.00000061, -0.00000045, 0.00000026, 0.00000150, 0.00000293, 0.00000402, 0.00000423, 0.00000321, 0.00000107, -0.00000167, -0.00000412, -0.00000541, -0.00000498, -0.00000287, 0.00000021, 0.00000316, 0.00000490, 0.00000477, 0.00000286, -0.00000005, -0.00000277, -0.00000427, -0.00000405, -0.00000239, -0.00000016, 0.00000154, 0.00000197, 0.00000107, -0.00000052, -0.00000175, -0.00000170, -0.00000008, 0.00000255, 0.00000498, 0.00000584, 0.00000422, 0.00000013, -0.00000537, -0.00001052, -0.00001351, -0.00001313, -0.00000925, -0.00000287, 0.00000421, 0.00001004, 0.00001316, 0.00001305, 0.00001017, 0.00000570, 0.00000102, -0.00000275, -0.00000512, -0.00000623, -0.00000667, -0.00000710, -0.00000794, -0.00000921, -0.00001050, -0.00001122, -0.00001081, -0.00000898, -0.00000582, -0.00000173, 0.00000267, 0.00000672, 0.00000990, 0.00001183, 0.00001237, 0.00001154, 0.00000956, 0.00000676, 0.00000363, 0.00000071, -0.00000147, -0.00000253, -0.00000234, -0.00000111, 0.00000071, 0.00000253, 0.00000379, 0.00000419, 0.00000371, 0.00000263, 0.00000135, 0.00000027, -0.00000036, -0.00000050, -0.00000022, 0.00000030, 0.00000094, 0.00000161, 0.00000221, 0.00000262, 0.00000264, 0.00000205, 0.00000075, -0.00000114, -0.00000322, -0.00000490, -0.00000561, -0.00000500, -0.00000318, -0.00000073, 0.00000153, 0.00000279, 0.00000260, 0.00000110, -0.00000104, -0.00000285, -0.00000345, -0.00000240, 0.00000014, 0.00000345, 0.00000652, 0.00000842, 0.00000863, 0.00000718, 0.00000457, 0.00000160, -0.00000096, -0.00000256, -0.00000307, -0.00000266, -0.00000179, -0.00000090, -0.00000034, -0.00000023, -0.00000042, },
		{ -0.00000150, -0.00000072, -0.00000042, -0.00000054, -0.00000078, -0.00000071, 0.00000002, 0.00000147, 0.00000335, 0.00000501, 0.00000571, 0.00000492, 0.00000258, -0.00000077, -0.00000410, -0.00000624, -0.00000635, -0.00000431, -0.00000078, 0.00000296, 0.00000553, 0.00000600, 0.00000426, 0.00000108, -0.00000220, -0.00000427, -0.00000439, -0.00000269, -0.00000011, 0.00000209, 0.00000290, 0.00000205, 0.00000013, -0.00000170, -0.00000226, -0.00000099, 0.00000178, 0.00000484, 0.00000663, 0.00000592, 0.00000238, -0.00000315, -0.00000895, -0.00001300, -0.00001373, -0.00001065, -0.00000447, 0.00000308, 0.00000994, 0.00001436, 0.00001544, 0.00001336, 0.00000916, 0.00000431, 0.00000015, -0.00000246, -0.00000331, -0.00000275, -0.00000144, 0.00000000, 0.00000110, 0.00000165, 0.00000163, 0.00000114, 0.00000028, -0.00000085, -0.00000211, -0.00000328, -0.00000409, -0.00000430, -0.00000382, -0.00000277, -0.00000155, -0.00000064, -0.00000050, -0.00000128, -0.00000282, -0.00000460, -0.00000595, -0.00000635, -0.00000557, -0.00000382, -0.00000165, 0.00000027, 0.00000141, 0.00000154, 0.00000083, -0.00000033, -0.00000144, -0.00000215, -0.00000230, -0.00000198, -0.00000135, -0.00000060, 0.00000014, 0.00000079, 0.00000127, 0.00000142, 0.00000106, 0.00000006, -0.00000154, -0.00000346, -0.00000518, -0.00000613, -0.00000592, -0.00000452, -0.00000233, -0.00000009, 0.00000143, 0.00000170, 0.00000070, -0.00000109, -0.00000283, -0.00000368, -0.00000308, -0.00000103, 0.00000194, 0.00000498, 0.00000719, 0.00000793, 0.00000706, 0.00000488, 0.00000206, -0.00000063, -0.00000258, -0.00000347, -0.00000334, -0.00000252, },
		{ -0.00000119, -0.00000254, -0.00000281, -0.00000220, -0.00000118, -0.00000026, 0.00000020, 0.00000012, -0.00000023, -0.00000036, 0.00000021, 0.00000170, 0.00000386, 0.00000602, 0.00000725, 0.00000679, 0.00000438, 0.00000050, -0.00000374, -0.00000691, -0.00000786, -0.00000621, -0.00000250, 0.00000189, 0.00000532, 0.00000653, 0.00000513, 0.00000181, -0.00000201, -0.00000478, -0.00000544, -0.00000393, -0.00000114, 0.00000154, 0.00000282, 0.00000216, 0.00000002, -0.00000240, -0.00000371, -0.00000302, -0.00000042, 0.00000301, 0.00000558, 0.00000579, 0.00000294, -0.00000245, -0.00000878, -0.00001393, -0.00001605, -0.00001424, -0.00000887, -0.00000145, 0.00000591, 0.00001123, 0.00001327, 0.00001179, 0.00000752, 0.00000175, -0.00000412, -0.00000899, -0.00001223, -0.00001366, -0.00001338, -0.00001161, -0.00000860, -0.00000463, -0.00000012, 0.00000433, 0.00000803, 0.00001034, 0.00001089, 0.00000977, 0.00000750, 0.00000490, 0.00000278, 0.00000163, 0.00000147, 0.00000189, 0.00000221, 0.00000188, 0.00000069, -0.00000111, -0.00000292, -0.00000407, -0.00000411, -0.00000300, -0.00000115, 0.00000079, 0.00000218, 0.00000263, 0.00000216, 0.00000110, -0.00000008, -0.00000095, -0.00000130, -0.00000112, -0.00000059, 0.00000011, 0.00000082, 0.00000146, 0.00000195, 0.00000219, 0.00000199, 0.00000123, -0.00000012, -0.00000187, -0.00000357, -0.00000470, -0.00000484, -0.00000385, -0.00000200, 0.00000011, 0.00000175, 0.00000238, 0.00000184, 0.00000044, -0.00000114, -0.00000211, -0.00000189, -0.00000034, 0.00000223, 0.00000511, 0.00000746, 0.00000863, 0.00000828, 0.00000655, 0.00000393, 0.00000112, },
		{ 0.00000814, 0.00000689, 0.00000457, 0.00000175, -0.00000084, -0.00000264, -0.00000333, -0.00000298, -0.00000198, -0.00000088, -0.00000019, -0.00000011, -0.00000045, -0.00000071, -0.00000029, 0.00000119, 0.00000362, 0.00000631, 0.00000823, 0.00000835, 0.00000619, 0.00000206, -0.00000289, -0.00000705, -0.00000894, -0.00000783, -0.00000411, 0.00000088, 0.00000527, 0.00000746, 0.00000675, 0.00000360, -0.00000055, -0.00000393, -0.00000518, -0.00000396, -0.00000100, 0.00000221, 0.00000415, 0.00000400, 0.00000195, -0.00000081, -0.00000274, -0.00000264, -0.00000028, 0.00000345, 0.00000688, 0.00000824, 0.00000650, 0.00000180, -0.00000448, -0.00001022, -0.00001336, -0.00001261, -0.00000795, -0.00000057, 0.00000759, 0.00001451, 0.00001874, 0.00001972, 0.00001778, 0.00001377, 0.00000871, 0.00000339, -0.00000171, -0.00000633, -0.00001026, -0.00001313, -0.00001448, -0.00001388, -0.00001123, -0.00000691, -0.00000180, 0.00000293, 0.00000623, 0.00000752, 0.00000687, 0.00000495, 0.00000275, 0.00000113, 0.00000053, 0.00000083, 0.00000145, 0.00000172, 0.00000120, -0.00000011, -0.00000177, -0.00000311, -0.00000356, -0.00000289, -0.00000131, 0.00000060, 0.00000220, 0.00000303, 0.00000294, 0.00000216, 0.00000112, 0.00000024, -0.00000020, -0.00000016, 0.00000024, 0.00000083, 0.00000146, 0.00000206, 0.00000254, 0.00000281, 0.00000274, 0.00000215, 0.00000099, -0.00000060, -0.00000228, -0.00000357, -0.00000405, -0.00000351, -0.00000210, -0.00000029, 0.00000130, 0.00000211, 0.00000192, 0.00000087, -0.00000050, -0.00000154, -0.00000164, -0.00000057, 0.00000152, 0.00000412, 0.00000650, 0.00000798, },
		{ 0.00000112, 0.00000344, 0.00000580, 0.00000758, 0.00000824, 0.00000756, 0.00000564, 0.00000296, 0.00000017, -0.00000203, -0.00000319, -0.00000317, -0.00000227, -0.00000106, -0.00000012, 0.00000017, -0.00000013, -0.00000053, -0.00000034, 0.00000100, 0.00000352, 0.00000662, 0.00000917, 0.00000996, 0.00000821, 0.00000401, -0.00000156, -0.00000675, -0.00000978, -0.00000957, -0.00000619, -0.00000090, 0.00000431, 0.00000750, 0.00000759, 0.00000476, 0.00000036, -0.00000370, -0.00000576, -0.00000512, -0.00000230, 0.00000121, 0.00000371, 0.00000405, 0.00000215, -0.00000095, -0.00000363, -0.00000441, -0.00000271, 0.00000086, 0.00000468, 0.00000687, 0.00000604, 0.00000195, -0.00000432, -0.00001079, -0.00001531, -0.00001638, -0.00001366, -0.00000804, -0.00000121, 0.00000500, 0.00000928, 0.00001118, 0.00001108, 0.00000982, 0.00000819, 0.00000660, 0.00000498, 0.00000299, 0.00000040, -0.00000266, -0.00000556, -0.00000742, -0.00000745, -0.00000540, -0.00000169, 0.00000267, 0.00000642, 0.00000857, 0.00000871, 0.00000718, 0.00000482, 0.00000263, 0.00000129, 0.00000097, 0.00000129, 0.00000162, 0.00000136, 0.00000034, -0.00000122, -0.00000274, -0.00000361, -0.00000347, -0.00000239, -0.00000079, 0.00000076, 0.00000174, 0.00000194, 0.00000145, 0.00000061, -0.00000020, -0.00000069, -0.00000077, -0.00000051, -0.00000004, 0.00000051, 0.00000106, 0.00000156, 0.00000192, 0.00000199, 0.00000164, 0.00000076, -0.00000057, -0.00000208, -0.00000336, -0.00000401, -0.00000379, -0.00000274, -0.00000119, 0.00000032, 0.00000131, 0.00000148, 0.00000087, -0.00000015, -0.00000105, -0.00000130, -0.00000057, },
		{ -0.00000038, -0.00000113, -0.00000148, -0.00000109, 0.00000014, 0.00000207, 0.00000426, 0.00000617, 0.00000724, 0.00000708, 0.00000562, 0.00000314, 0.00000023, -0.00000238, -0.00000404, -0.00000447, -0.00000378, -0.00000250, -0.00000133, -0.00000077, -0.00000093, -0.00000140, -0.00000142, -0.00000030, 0.00000221, 0.00000563, 0.00000884, 0.00001042, 0.00000933, 0.00000537, -0.00000054, -0.00000661, -0.00001080, -0.00001163, -0.00000881, -0.00000342, 0.00000254, 0.00000684, 0.00000801, 0.00000584, 0.00000150, -0.00000304, -0.00000585, -0.00000585, -0.00000328, 0.00000051, 0.00000367, 0.00000476, 0.00000335, 0.00000027, -0.00000284, -0.00000431, -0.00000319, 0.00000023, 0.00000451, 0.00000771, 0.00000822, 0.00000547, 0.00000027, -0.00000555, -0.00000983, -0.00001100, -0.00000862, -0.00000361, 0.00000226, 0.00000702, 0.00000931, 0.00000881, 0.00000625, 0.00000291, 0.00000006, -0.00000162, -0.00000222, -0.00000243, -0.00000304, -0.00000448, -0.00000655, -0.00000847, -0.00000924, -0.00000817, -0.00000522, -0.00000107, 0.00000312, 0.00000618, 0.00000739, 0.00000673, 0.00000482, 0.00000262, 0.00000096, 0.00000025, 0.00000035, 0.00000072, 0.00000079, 0.00000019, -0.00000101, -0.00000238, -0.00000337, -0.00000357, -0.00000289, -0.00000160, -0.00000017, 0.00000091, 0.00000137, 0.00000122, 0.00000068, 0.00000007, -0.00000036, -0.00000050, -0.00000036, -0.00000003, 0.00000038, 0.00000083, 0.00000127, 0.00000163, 0.00000178, 0.00000158, 0.00000091, -0.00000020, -0.00000155, -0.00000282, -0.00000364, -0.00000374, -0.00000312, -0.00000197, -0.00000071, 0.00000025, 0.00000062, 0.00000034, },
		{ -0.00000112, -0.00000023, 0.00000029, 0.00000035, 0.00000001, -0.00000046, -0.00000076, -0.00000056, 0.00000031, 0.00000186, 0.00000385, 0.00000583, 0.00000726, 0.00000766, 0.00000674, 0.00000458, 0.00000167, -0.00000126, -0.00000345, -0.00000440, -0.00000404, -0.00000282, -0.00000145, -0.00000061, -0.00000058, -0.00000106, -0.00000133, -0.00000053, 0.00000178, 0.00000533, 0.00000903, 0.00001138, 0.00001105, 0.00000752, 0.00000146, -0.00000539, -0.00001080, -0.00001293, -0.00001106, -0.00000594, 0.00000046, 0.00000575, 0.00000802, 0.00000666, 0.00000253, -0.00000244, -0.00000609, -0.00000699, -0.00000499, -0.00000126, 0.00000232, 0.00000403, 0.00000309, 0.00000003, -0.00000359, -0.00000595, -0.00000581, -0.00000314, 0.00000080, 0.00000404, 0.00000478, 0.00000220, -0.00000306, -0.00000912, -0.00001357, -0.00001447, -0.00001112, -0.00000434, 0.00000384, 0.00001097, 0.00001512, 0.00001553, 0.00001277, 0.00000832, 0.00000393, 0.00000083, -0.00000064, -0.00000100, -0.00000120, -0.00000204, -0.00000374, -0.00000579, -0.00000725, -0.00000723, -0.00000531, -0.00000184, 0.00000222, 0.00000569, 0.00000761, 0.00000767, 0.00000622, 0.00000407, 0.00000210, 0.00000089, 0.00000053, 0.00000067, 0.00000077, 0.00000041, -0.00000050, -0.00000170, -0.00000276, -0.00000325, -0.00000301, -0.00000215, -0.00000103, -0.00000004, 0.00000054, 0.00000064, 0.00000040, 0.00000003, -0.00000027, -0.00000039, -0.00000033, -0.00000014, 0.00000014, 0.00000047, 0.00000085, 0.00000120, 0.00000143, 0.00000139, 0.00000097, 0.00000014, -0.00000097, -0.00000208, -0.00000291, -0.00000321, -0.00000291, -0.00000212, },
		{ -0.00000190, -0.00000237, -0.00000241, -0.00000205, -0.00000144, -0.00000079, -0.00000029, -0.00000005, -0.00000008, -0.00000029, -0.00000050, -0.00000045, 0.00000006, 0.00000118, 0.00000283, 0.00000474, 0.00000642, 0.00000730, 0.00000694, 0.00000522, 0.00000243, -0.00000075, -0.00000344, -0.00000497, -0.00000506, -0.00000399, -0.00000247, -0.00000130, -0.00000095, -0.00000131, -0.00000170, -0.00000119, 0.00000086, 0.00000442, 0.00000855, 0.00001171, 0.00001233, 0.00000956, 0.00000373, -0.00000357, -0.00001003, -0.00001345, -0.00001268, -0.00000806, -0.00000137, 0.00000489, 0.00000846, 0.00000825, 0.00000473, -0.00000032, -0.00000461, -0.00000634, -0.00000495, -0.00000130, 0.00000275, 0.00000528, 0.00000517, 0.00000262, -0.00000094, -0.00000356, -0.00000371, -0.00000104, 0.00000336, 0.00000741, 0.00000893, 0.00000660, 0.00000065, -0.00000715, -0.00001413, -0.00001775, -0.00001661, -0.00001096, -0.00000260, 0.00000585, 0.00001193, 0.00001423, 0.00001278, 0.00000884, 0.00000423, 0.00000056, -0.00000141, -0.00000188, -0.00000174, -0.00000195, -0.00000306, -0.00000485, -0.00000655, -0.00000718, -0.00000608, -0.00000327, 0.00000055, 0.00000428, 0.00000687, 0.00000774, 0.00000696, 0.00000515, 0.00000315, 0.00000163, 0.00000088, 0.00000076, 0.00000084, 0.00000071, 0.00000014, -0.00000075, -0.00000167, -0.00000227, -0.00000233, -0.00000186, -0.00000108, -0.00000028, 0.00000033, 0.00000063, 0.00000068, 0.00000059, 0.00000050, 0.00000047, 0.00000050, 0.00000059, 0.00000071, 0.00000089, 0.00000113, 0.00000140, 0.00000161, 0.00000164, 0.00000137, 0.00000075, -0.00000013, -0.00000109, },
		{ 0.00000028, -0.00000009, -0.00000068, -0.00000138, -0.00000204, -0.00000252, -0.00000274, -0.00000271, -0.00000245, -0.00000206, -0.00000162, -0.00000122, -0.00000094, -0.00000079, -0.00000076, -0.00000070, -0.00000040, 0.00000035, 0.00000167, 0.00000344, 0.00000530, 0.00000665, 0.00000691, 0.00000574, 0.00000323, -0.00000004, -0.00000318, -0.00000532, -0.00000597, -0.00000521, -0.00000367, -0.00000221, -0.00000150, -0.00000166, -0.00000212, -0.00000194, -0.00000028, 0.00000306, 0.00000738, 0.00001118, 0.00001273, 0.00001083, 0.00000545, -0.00000209, -0.00000950, -0.00001434, -0.00001498, -0.00001128, -0.00000472, 0.00000221, 0.00000696, 0.00000796, 0.00000519, 0.00000017, -0.00000477, -0.00000752, -0.00000706, -0.00000390, 0.00000023, 0.00000324, 0.00000363, 0.00000120, -0.00000280, -0.00000632, -0.00000741, -0.00000518, -0.00000028, 0.00000532, 0.00000913, 0.00000921, 0.00000508, -0.00000207, -0.00000973, -0.00001509, -0.00001613, -0.00001238, -0.00000510, 0.00000331, 0.00001023, 0.00001377, 0.00001344, 0.00001011, 0.00000549, 0.00000135, -0.00000120, -0.00000204, -0.00000187, -0.00000173, -0.00000236, -0.00000386, -0.00000566, -0.00000680, -0.00000649, -0.00000446, -0.00000113, 0.00000255, 0.00000553, 0.00000705, 0.00000692, 0.00000552, 0.00000357, 0.00000178, 0.00000060, 0.00000007, -0.00000006, -0.00000016, -0.00000049, -0.00000111, -0.00000185, -0.00000248, -0.00000279, -0.00000273, -0.00000237, -0.00000188, -0.00000140, -0.00000103, -0.00000076, -0.00000057, -0.00000042, -0.00000030, -0.00000022, -0.00000019, -0.00000018, -0.00000014, -0.00000004, 0.00000014, 0.00000032, 0.00000040, },
		},
		{ { -0.00000010, -0.00000006, 0.00000004, 0.00000022, 0.00000040, 0.00000046, 0.00000029, -0.00000009, -0.00000045, -0.00000045, 0.00000012, 0.00000116, 0.00000217, 0.00000249, 0.00000168, -0.00000011, -0.00000206, -0.00000304, -0.00000218, 0.00000056, 0.00000429, 0.00000758, 0.00000920, 0.00000871, 0.00000669, 0.00000429, 0.00000252, 0.00000163, 0.00000105, -0.00000005, -0.00000207, -0.00000446, -0.00000592, -0.00000516, -0.00000184, 0.00000309, 0.00000768, 0.00001010, 0.00000960, 0.00000696, 0.00000389, 0.00000197, 0.00000156, 0.00000165, 0.00000065, -0.00000228, -0.00000639, -0.00000945, -0.00000903, -0.00000412, 0.00000385, 0.00001149, 0.00001511, 0.00001280, 0.00000562, -0.00000293, -0.00000872, -0.00000929, -0.00000513, 0.00000075, 0.00000470, 0.00000465, 0.00000118, -0.00000297, -0.00000478, -0.00000297, 0.00000118, 0.00000465, 0.00000470, 0.00000075, -0.00000513, -0.00000929, -0.00000872, -0.00000293, 0.00000562, 0.00001280, 0.00001511, 0.00001149, 0.00000385, -0.00000412, -0.00000903, -0.00000945, -0.00000639, -0.00000228, 0.00000065, 0.00000165, 0.00000156, 0.00000197, 0.00000389, 0.00000696, 0.00000960, 0.00001010, 0.00000768, 0.00000309, -0.00000184, -0.00000516, -0.00000592, -0.00000446, -0.00000207, -0.00000005, 0.00000105, 0.00000163, 0.00000252, 0.00000429, 0.00000669, 0.00000871, 0.00000920, 0.00000758, 0.00000429, 0.00000056, -0.00000218, -0.00000304, -0.00000206, -0.00000011, 0.00000168, 0.00000249, 0.00000217, 0.00000116, 0.00000012, -0.00000045, -0.00000045, -0.00000009, 0.00000029, 0.00000046, 0.00000040, 0.00000022, 0.00000004, -0.00000006, },
		{ 0.00000007, 0.00000022, 0.00000022, 0.00000005, -0.00000020, -0.00000030, -0.00000001, 0.00000069, 0.00000152, 0.00000195, 0.00000152, 0.00000019, -0.00000151, -0.00000265, -0.00000236, -0.00000036, 0.00000281, 0.00000597, 0.00000791, 0.00000799, 0.00000647, 0.00000429, 0.00000243, 0.00000135, 0.00000072, -0.00000022, -0.00000200, -0.00000433, -0.00000615, -0.00000612, -0.00000358, 0.00000093, 0.00000577, 0.00000900, 0.00000949, 0.00000747, 0.00000439, 0.00000190, 0.00000083, 0.00000066, -0.00000000, -0.00000227, -0.00000594, -0.00000925, -0.00000982, -0.00000619, 0.00000098, 0.00000886, 0.00001374, 0.00001311, 0.00000710, -0.00000140, -0.00000826, -0.00001035, -0.00000728, -0.00000145, 0.00000340, 0.00000447, 0.00000151, -0.00000317, -0.00000626, -0.00000562, -0.00000167, 0.00000293, 0.00000497, 0.00000284, -0.00000235, -0.00000731, -0.00000853, -0.00000450, 0.00000331, 0.00001114, 0.00001502, 0.00001297, 0.00000602, -0.00000244, -0.00000863, -0.00001036, -0.00000794, -0.00000362, -0.00000003, 0.00000148, 0.00000132, 0.00000111, 0.00000229, 0.00000503, 0.00000807, 0.00000949, 0.00000804, 0.00000394, -0.00000121, -0.00000529, -0.00000689, -0.00000595, -0.00000361, -0.00000132, 0.00000005, 0.00000066, 0.00000138, 0.00000299, 0.00000549, 0.00000798, 0.00000913, 0.00000805, 0.00000486, 0.00000073, -0.00000272, -0.00000424, -0.00000356, -0.00000145, 0.00000083, 0.00000216, 0.00000212, 0.00000106, -0.00000026, -0.00000114, -0.00000129, -0.00000086, -0.00000027, 0.00000014, 0.00000023, 0.00000009, -0.00000012, -0.00000026, -0.00000031, -0.00000030, -0.00000023, -0.00000010, },
		{ -0.00000072, -0.00000083, -0.00000076, -0.00000036, 0.00000023, 0.00000066, 0.00000052, -0.00000036, -0.00000171, -0.00000283, -0.00000292, -0.00000155, 0.00000103, 0.00000394, 0.00000605, 0.00000664, 0.00000568, 0.00000386, 0.00000208, 0.00000091, 0.00000024, -0.00000056, -0.00000207, -0.00000425, -0.00000624, -0.00000677, -0.00000496, -0.00000094, 0.00000395, 0.00000785, 0.00000931, 0.00000809, 0.00000525, 0.00000241, 0.00000072, 0.00000015, -0.00000038, -0.00000207, -0.00000516, -0.00000843, -0.00000972, -0.00000729, -0.00000114, 0.00000655, 0.00001233, 0.00001327, 0.00000871, 0.00000077, -0.00000669, -0.00001014, -0.00000836, -0.00000300, 0.00000241, 0.00000467, 0.00000278, -0.00000155, -0.00000503, -0.00000502, -0.00000123, 0.00000400, 0.00000716, 0.00000579, 0.00000016, -0.00000674, -0.00001086, -0.00000952, -0.00000303, 0.00000537, 0.00001133, 0.00001176, 0.00000651, -0.00000167, -0.00000882, -0.00001194, -0.00001044, -0.00000613, -0.00000180, 0.00000053, 0.00000069, 0.00000010, 0.00000054, 0.00000277, 0.00000598, 0.00000828, 0.00000797, 0.00000468, -0.00000036, -0.00000503, -0.00000751, -0.00000723, -0.00000506, -0.00000252, -0.00000077, 0.00000002, 0.00000062, 0.00000199, 0.00000445, 0.00000728, 0.00000910, 0.00000870, 0.00000583, 0.00000146, -0.00000268, -0.00000500, -0.00000486, -0.00000278, -0.00000012, 0.00000176, 0.00000210, 0.00000107, -0.00000053, -0.00000180, -0.00000224, -0.00000188, -0.00000115, -0.00000054, -0.00000030, -0.00000041, -0.00000067, -0.00000091, -0.00000103, -0.00000107, -0.00000106, -0.00000101, -0.00000089, -0.00000074, -0.00000061, -0.00000060, },
		{ 0.00000024, 0.00000061, 0.00000069, 0.00000022, -0.00000071, -0.00000165, -0.00000194, -0.00000109, 0.00000090, 0.00000342, 0.00000555, 0.00000650, 0.00000605, 0.00000463, 0.00000302, 0.00000181, 0.00000107, 0.00000032, -0.00000101, -0.00000304, -0.00000514, -0.00000616, -0.00000506, -0.00000168, 0.00000304, 0.00000735, 0.00000964, 0.00000924, 0.00000677, 0.00000370, 0.00000140, 0.00000026, -0.00000038, -0.00000168, -0.00000422, -0.00000729, -0.00000909, -0.00000779, -0.00000285, 0.00000423, 0.00001042, 0.00001264, 0.00000955, 0.00000247, -0.00000523, -0.00000990, -0.00000959, -0.00000518, 0.00000028, 0.00000339, 0.00000250, -0.00000141, -0.00000538, -0.00000647, -0.00000375, 0.00000102, 0.00000442, 0.00000356, -0.00000193, -0.00000939, -0.00001452, -0.00001382, -0.00000672, 0.00000390, 0.00001320, 0.00001694, 0.00001371, 0.00000554, -0.00000335, -0.00000898, -0.00000955, -0.00000611, -0.00000145, 0.00000173, 0.00000242, 0.00000157, 0.00000110, 0.00000240, 0.00000528, 0.00000810, 0.00000885, 0.00000653, 0.00000180, -0.00000335, -0.00000682, -0.00000749, -0.00000579, -0.00000317, -0.00000107, -0.00000004, 0.00000047, 0.00000151, 0.00000371, 0.00000666, 0.00000904, 0.00000937, 0.00000700, 0.00000260, -0.00000214, -0.00000535, -0.00000593, -0.00000411, -0.00000118, 0.00000127, 0.00000213, 0.00000130, -0.00000047, -0.00000213, -0.00000292, -0.00000269, -0.00000185, -0.00000098, -0.00000049, -0.00000047, -0.00000072, -0.00000099, -0.00000115, -0.00000119, -0.00000118, -0.00000114, -0.00000104, -0.00000084, -0.00000058, -0.00000035, -0.00000025, -0.00000027, -0.00000028, -0.00000012, },
		{ -0.00000009, -0.00000082, -0.00000123, -0.00000082, 0.00000057, 0.00000259, 0.00000454, 0.00000567, 0.00000562, 0.00000460, 0.00000321, 0.00000202, 0.00000123, 0.00000051, -0.00000068, -0.00000255, -0.00000470, -0.00000608, -0.00000562, -0.00000287, 0.00000156, 0.00000615, 0.00000918, 0.00000966, 0.00000776, 0.00000470, 0.00000194, 0.00000027, -0.00000058, -0.00000157, -0.00000349, -0.00000612, -0.00000809, -0.00000764, -0.00000385, 0.00000244, 0.00000873, 0.00001204, 0.00001052, 0.00000469, -0.00000272, -0.00000813, -0.00000912, -0.00000575, -0.00000045, 0.00000344, 0.00000374, 0.00000071, -0.00000318, -0.00000493, -0.00000301, 0.00000152, 0.00000561, 0.00000615, 0.00000210, -0.00000455, -0.00000973, -0.00000963, -0.00000308, 0.00000749, 0.00001715, 0.00002104, 0.00001696, 0.00000657, -0.00000556, -0.00001440, -0.00001701, -0.00001373, -0.00000753, -0.00000203, 0.00000057, 0.00000046, -0.00000045, -0.00000009, 0.00000224, 0.00000542, 0.00000737, 0.00000646, 0.00000260, -0.00000259, -0.00000684, -0.00000850, -0.00000741, -0.00000476, -0.00000219, -0.00000067, -0.00000000, 0.00000084, 0.00000276, 0.00000572, 0.00000858, 0.00000975, 0.00000816, 0.00000406, -0.00000102, -0.00000503, -0.00000645, -0.00000506, -0.00000198, 0.00000104, 0.00000257, 0.00000214, 0.00000035, -0.00000165, -0.00000284, -0.00000284, -0.00000197, -0.00000086, -0.00000010, 0.00000010, -0.00000011, -0.00000043, -0.00000065, -0.00000074, -0.00000076, -0.00000076, -0.00000072, -0.00000056, -0.00000026, 0.00000010, 0.00000037, 0.00000044, 0.00000037, 0.00000029, 0.00000035, 0.00000054, 0.00000066, 0.00000048, },
		{ -0.00000009, 0.00000146, 0.00000318, 0.00000441, 0.00000473, 0.00000414, 0.00000307, 0.00000204, 0.00000129, 0.00000063, -0.00000039, -0.00000208, -0.00000417, -0.00000581, -0.00000589, -0.00000376, 0.00000029, 0.00000499, 0.00000865, 0.00000997, 0.00000873, 0.00000583, 0.00000270, 0.00000045, -0.00000077, -0.00000166, -0.00000307, -0.00000520, -0.00000716, -0.00000742, -0.00000482, 0.00000038, 0.00000632, 0.00001025, 0.00001004, 0.00000550, -0.00000135, -0.00000728, -0.00000955, -0.00000750, -0.00000287, 0.00000129, 0.00000248, 0.00000027, -0.00000349, -0.00000592, -0.00000508, -0.00000131, 0.00000283, 0.00000418, 0.00000110, -0.00000523, -0.00001123, -0.00001297, -0.00000861, 0.00000026, 0.00000921, 0.00001334, 0.00001002, 0.00000043, -0.00001097, -0.00001882, -0.00001963, -0.00001355, -0.00000396, 0.00000459, 0.00000883, 0.00000836, 0.00000538, 0.00000294, 0.00000301, 0.00000532, 0.00000784, 0.00000822, 0.00000542, 0.00000034, -0.00000479, -0.00000781, -0.00000784, -0.00000562, -0.00000280, -0.00000079, 0.00000012, 0.00000074, 0.00000217, 0.00000480, 0.00000784, 0.00000969, 0.00000893, 0.00000531, 0.00000004, -0.00000476, -0.00000720, -0.00000656, -0.00000364, -0.00000020, 0.00000199, 0.00000207, 0.00000037, -0.00000195, -0.00000364, -0.00000403, -0.00000327, -0.00000200, -0.00000098, -0.00000056, -0.00000069, -0.00000107, -0.00000139, -0.00000154, -0.00000159, -0.00000163, -0.00000163, -0.00000151, -0.00000118, -0.00000070, -0.00000025, -0.00000000, -0.00000005, -0.00000026, -0.00000043, -0.00000043, -0.00000030, -0.00000024, -0.00000041, -0.00000082, -0.00000115, -0.00000100, },
		{ 0.00000420, 0.00000393, 0.00000313, 0.00000222, 0.00000149, 0.00000085, -0.00000008, -0.00000164, -0.00000373, -0.00000561, -0.00000624, -0.00000479, -0.00000125, 0.00000339, 0.00000749, 0.00000957, 0.00000902, 0.00000638, 0.00000299, 0.00000015, -0.00000154, -0.00000247, -0.00000346, -0.00000502, -0.00000673, -0.00000742, -0.00000585, -0.00000179, 0.00000352, 0.00000777, 0.00000873, 0.00000563, -0.00000025, -0.00000616, -0.00000932, -0.00000847, -0.00000462, -0.00000036, 0.00000171, 0.00000057, -0.00000259, -0.00000521, -0.00000509, -0.00000194, 0.00000227, 0.00000454, 0.00000278, -0.00000259, -0.00000862, -0.00001140, -0.00000849, -0.00000054, 0.00000878, 0.00001468, 0.00001401, 0.00000704, -0.00000253, -0.00000963, -0.00001068, -0.00000549, 0.00000269, 0.00000915, 0.00001039, 0.00000602, -0.00000120, -0.00000718, -0.00000890, -0.00000606, -0.00000100, 0.00000285, 0.00000308, -0.00000042, -0.00000560, -0.00000971, -0.00001089, -0.00000912, -0.00000588, -0.00000297, -0.00000131, -0.00000049, 0.00000063, 0.00000289, 0.00000599, 0.00000851, 0.00000880, 0.00000607, 0.00000106, -0.00000425, -0.00000764, -0.00000788, -0.00000529, -0.00000152, 0.00000143, 0.00000225, 0.00000091, -0.00000153, -0.00000366, -0.00000449, -0.00000390, -0.00000250, -0.00000115, -0.00000041, -0.00000036, -0.00000071, -0.00000109, -0.00000131, -0.00000138, -0.00000141, -0.00000143, -0.00000134, -0.00000101, -0.00000045, 0.00000017, 0.00000060, 0.00000064, 0.00000035, -0.00000005, -0.00000030, -0.00000026, -0.00000007, 0.00000006, -0.00000003, -0.00000025, -0.00000029, 0.00000017, 0.00000120, 0.00000255, 0.00000370, },
		{ 0.00000159, 0.00000105, 0.00000028, -0.00000108, -0.00000304, -0.00000505, -0.00000609, -0.00000523, -0.00000217, 0.00000237, 0.00000688, 0.00000975, 0.00001000, 0.00000782, 0.00000435, 0.00000106, -0.00000112, -0.00000216, -0.00000277, -0.00000367, -0.00000492, -0.00000574, -0.00000500, -0.00000209, 0.00000235, 0.00000654, 0.00000837, 0.00000665, 0.00000194, -0.00000362, -0.00000738, -0.00000771, -0.00000489, -0.00000095, 0.00000161, 0.00000137, -0.00000114, -0.00000382, -0.00000442, -0.00000215, 0.00000167, 0.00000434, 0.00000352, -0.00000105, -0.00000711, -0.00001102, -0.00000990, -0.00000352, 0.00000525, 0.00001190, 0.00001279, 0.00000726, -0.00000195, -0.00001020, -0.00001361, -0.00001123, -0.00000547, -0.00000049, 0.00000034, -0.00000344, -0.00000916, -0.00001263, -0.00001070, -0.00000325, 0.00000672, 0.00001472, 0.00001721, 0.00001352, 0.00000593, -0.00000173, -0.00000623, -0.00000652, -0.00000379, -0.00000034, 0.00000200, 0.00000292, 0.00000342, 0.00000474, 0.00000723, 0.00000989, 0.00001097, 0.00000916, 0.00000457, -0.00000115, -0.00000563, -0.00000710, -0.00000529, -0.00000156, 0.00000197, 0.00000354, 0.00000265, 0.00000013, -0.00000250, -0.00000395, -0.00000377, -0.00000243, -0.00000086, 0.00000018, 0.00000039, 0.00000003, -0.00000048, -0.00000085, -0.00000102, -0.00000111, -0.00000119, -0.00000118, -0.00000093, -0.00000038, 0.00000034, 0.00000090, 0.00000105, 0.00000071, 0.00000010, -0.00000041, -0.00000056, -0.00000034, 0.00000003, 0.00000025, 0.00000024, 0.00000013, 0.00000024, 0.00000080, 0.00000175, 0.00000275, 0.00000338, 0.00000340, 0.00000290, 0.00000220, },
		{ -0.00000148, -0.00000355, -0.00000497, -0.00000470, -0.00000221, 0.00000210, 0.00000688, 0.00001044, 0.00001150, 0.00000983, 0.00000636, 0.00000260, -0.00000017, -0.00000151, -0.00000191, -0.00000220, -0.00000289, -0.00000366, -0.00000357, -0.00000183, 0.00000150, 0.00000524, 0.00000756, 0.00000704, 0.00000362, -0.00000124, -0.00000525, -0.00000655, -0.00000484, -0.00000148, 0.00000131, 0.00000188, 0.00000018, -0.00000221, -0.00000322, -0.00000168, 0.00000170, 0.00000464, 0.00000477, 0.00000126, -0.00000434, -0.00000879, -0.00000899, -0.00000398, 0.00000425, 0.00001165, 0.00001432, 0.00001078, 0.00000282, -0.00000542, -0.00000986, -0.00000872, -0.00000345, 0.00000237, 0.00000528, 0.00000419, 0.00000092, -0.00000096, 0.00000138, 0.00000808, 0.00001617, 0.00002117, 0.00001956, 0.00001093, -0.00000174, -0.00001357, -0.00002033, -0.00002045, -0.00001542, -0.00000862, -0.00000320, -0.00000045, 0.00000038, 0.00000111, 0.00000304, 0.00000590, 0.00000809, 0.00000781, 0.00000440, -0.00000106, -0.00000623, -0.00000884, -0.00000793, -0.00000434, -0.00000018, 0.00000240, 0.00000234, 0.00000010, -0.00000282, -0.00000484, -0.00000510, -0.00000384, -0.00000198, -0.00000052, 0.00000005, -0.00000016, -0.00000071, -0.00000116, -0.00000138, -0.00000145, -0.00000148, -0.00000146, -0.00000122, -0.00000064, 0.00000020, 0.00000097, 0.00000133, 0.00000108, 0.00000041, -0.00000029, -0.00000061, -0.00000037, 0.00000023, 0.00000082, 0.00000111, 0.00000109, 0.00000101, 0.00000119, 0.00000177, 0.00000259, 0.00000329, 0.00000357, 0.00000337, 0.00000291, 0.00000245, 0.00000204, 0.00000146, 0.00000033, },
		{ -0.00000331, 0.00000061, 0.00000552, 0.00000968, 0.00001154, 0.00001046, 0.00000708, 0.00000289, -0.00000054, -0.00000233, -0.00000266, -0.00000241, -0.00000242, -0.00000289, -0.00000323, -0.00000256, -0.00000046, 0.00000250, 0.00000494, 0.00000539, 0.00000328, -0.00000061, -0.00000450, -0.00000653, -0.00000593, -0.00000341, -0.00000074, 0.00000038, -0.00000054, -0.00000252, -0.00000375, -0.00000291, -0.00000019, 0.00000266, 0.00000337, 0.00000075, -0.00000434, -0.00000918, -0.00001067, -0.00000727, -0.00000013, 0.00000730, 0.00001111, 0.00000915, 0.00000227, -0.00000605, -0.00001171, -0.00001229, -0.00000833, -0.00000285, 0.00000065, 0.00000037, -0.00000275, -0.00000579, -0.00000590, -0.00000231, 0.00000292, 0.00000608, 0.00000412, -0.00000328, -0.00001324, -0.00002105, -0.00002274, -0.00001722, -0.00000683, 0.00000402, 0.00001125, 0.00001320, 0.00001111, 0.00000800, 0.00000660, 0.00000770, 0.00000991, 0.00001078, 0.00000853, 0.00000334, -0.00000275, -0.00000703, -0.00000772, -0.00000495, -0.00000060, 0.00000282, 0.00000364, 0.00000176, -0.00000148, -0.00000423, -0.00000522, -0.00000430, -0.00000233, -0.00000048, 0.00000047, 0.00000041, -0.00000020, -0.00000083, -0.00000119, -0.00000129, -0.00000131, -0.00000128, -0.00000108, -0.00000054, 0.00000032, 0.00000123, 0.00000177, 0.00000165, 0.00000093, 0.00000003, -0.00000052, -0.00000040, 0.00000030, 0.00000118, 0.00000177, 0.00000187, 0.00000164, 0.00000143, 0.00000158, 0.00000209, 0.00000272, 0.00000313, 0.00000314, 0.00000285, 0.00000248, 0.00000215, 0.00000170, 0.00000078, -0.00000087, -0.00000298, -0.00000477, -0.00000514, },
		{ 0.00001083, 0.00001049, 0.00000737, 0.00000285, -0.00000124, -0.00000359, -0.00000398, -0.00000321, -0.00000243, -0.00000234, -0.00000280, -0.00000300, -0.00000217, -0.00000022, 0.00000199, 0.00000316, 0.00000235, -0.00000033, -0.00000369, -0.00000609, -0.00000644, -0.00000484, -0.00000254, -0.00000105, -0.00000121, -0.00000254, -0.00000366, -0.00000325, -0.00000110, 0.00000157, 0.00000275, 0.00000105, -0.00000324, -0.00000798, -0.00001027, -0.00000818, -0.00000209, 0.00000524, 0.00001011, 0.00000989, 0.00000460, -0.00000304, -0.00000913, -0.00001081, -0.00000777, -0.00000234, 0.00000209, 0.00000322, 0.00000117, -0.00000171, -0.00000253, 0.00000014, 0.00000520, 0.00000966, 0.00001055, 0.00000694, 0.00000071, -0.00000447, -0.00000530, -0.00000101, 0.00000611, 0.00001191, 0.00001280, 0.00000786, -0.00000056, -0.00000826, -0.00001151, -0.00000921, -0.00000334, 0.00000229, 0.00000430, 0.00000163, -0.00000403, -0.00000937, -0.00001148, -0.00000944, -0.00000466, 0.00000012, 0.00000248, 0.00000160, -0.00000148, -0.00000475, -0.00000642, -0.00000586, -0.00000372, -0.00000134, 0.00000019, 0.00000048, -0.00000010, -0.00000089, -0.00000140, -0.00000154, -0.00000150, -0.00000142, -0.00000124, -0.00000079, 0.00000003, 0.00000099, 0.00000168, 0.00000170, 0.00000099, -0.00000008, -0.00000091, -0.00000101, -0.00000033, 0.00000075, 0.00000162, 0.00000190, 0.00000158, 0.00000104, 0.00000073, 0.00000088, 0.00000136, 0.00000184, 0.00000202, 0.00000188, 0.00000160, 0.00000134, 0.00000103, 0.00000034, -0.00000107, -0.00000312, -0.00000518, -0.00000613, -0.00000499, -0.00000154, 0.00000340, 0.00000812, },
		{ 0.00000010, -0.00000293, -0.00000356, -0.00000237, -0.00000073, 0.00000014, -0.00000014, -0.00000099, -0.00000140, -0.00000069, 0.00000094, 0.00000248, 0.00000282, 0.00000147, -0.00000104, -0.00000346, -0.00000458, -0.00000399, -0.00000232, -0.00000079, -0.00000035, -0.00000101, -0.00000185, -0.00000171, -0.00000014, 0.00000214, 0.00000348, 0.00000244, -0.00000111, -0.00000563, -0.00000853, -0.00000773, -0.00000296, 0.00000376, 0.00000908, 0.00001015, 0.00000627, -0.00000065, -0.00000712, -0.00001000, -0.00000830, -0.00000361, 0.00000098, 0.00000284, 0.00000137, -0.00000172, -0.00000368, -0.00000260, 0.00000124, 0.00000548, 0.00000719, 0.00000478, -0.00000089, -0.00000700, -0.00001052, -0.00001019, -0.00000730, -0.00000486, -0.00000554, -0.00000971, -0.00001495, -0.00001725, -0.00001348, -0.00000341, 0.00000972, 0.00002077, 0.00002526, 0.00002175, 0.00001248, 0.00000212, -0.00000476, -0.00000590, -0.00000223, 0.00000307, 0.00000657, 0.00000640, 0.00000304, -0.00000137, -0.00000444, -0.00000492, -0.00000314, -0.00000052, 0.00000142, 0.00000193, 0.00000120, 0.00000002, -0.00000089, -0.00000127, -0.00000128, -0.00000117, -0.00000104, -0.00000072, -0.00000006, 0.00000087, 0.00000166, 0.00000187, 0.00000127, 0.00000012, -0.00000093, -0.00000126, -0.00000064, 0.00000062, 0.00000184, 0.00000242, 0.00000220, 0.00000148, 0.00000081, 0.00000063, 0.00000097, 0.00000153, 0.00000194, 0.00000201, 0.00000188, 0.00000173, 0.00000159, 0.00000120, 0.00000014, -0.00000171, -0.00000387, -0.00000531, -0.00000487, -0.00000196, 0.00000289, 0.00000812, 0.00001173, 0.00001227, 0.00000958, 0.00000486, },
		{ -0.00000022, 0.00000159, 0.00000177, 0.00000051, -0.00000107, -0.00000177, -0.00000111, 0.00000041, 0.00000167, 0.00000168, 0.00000027, -0.00000183, -0.00000347, -0.00000386, -0.00000303, -0.00000174, -0.00000091, -0.00000090, -0.00000128, -0.00000118, -0.00000004, 0.00000178, 0.00000309, 0.00000257, -0.00000024, -0.00000429, -0.00000746, -0.00000764, -0.00000410, 0.00000185, 0.00000736, 0.00000956, 0.00000718, 0.00000138, -0.00000489, -0.00000848, -0.00000784, -0.00000383, 0.00000093, 0.00000366, 0.00000319, 0.00000055, -0.00000178, -0.00000160, 0.00000152, 0.00000590, 0.00000880, 0.00000819, 0.00000412, -0.00000133, -0.00000530, -0.00000598, -0.00000371, -0.00000057, 0.00000114, 0.00000067, -0.00000047, 0.00000055, 0.00000588, 0.00001513, 0.00002492, 0.00003033, 0.00002752, 0.00001600, -0.00000072, -0.00001664, -0.00002616, -0.00002664, -0.00001954, -0.00000936, -0.00000113, 0.00000201, 0.00000001, -0.00000460, -0.00000853, -0.00000957, -0.00000753, -0.00000392, -0.00000074, 0.00000072, 0.00000040, -0.00000081, -0.00000188, -0.00000224, -0.00000198, -0.00000149, -0.00000105, -0.00000062, 0.00000002, 0.00000096, 0.00000193, 0.00000244, 0.00000213, 0.00000111, -0.00000005, -0.00000061, -0.00000013, 0.00000122, 0.00000275, 0.00000369, 0.00000364, 0.00000280, 0.00000178, 0.00000118, 0.00000126, 0.00000178, 0.00000232, 0.00000255, 0.00000248, 0.00000234, 0.00000225, 0.00000205, 0.00000130, -0.00000031, -0.00000254, -0.00000445, -0.00000479, -0.00000263, 0.00000191, 0.00000745, 0.00001189, 0.00001337, 0.00001123, 0.00000640, 0.00000094, -0.00000296, -0.00000407, -0.00000268, },
		{ -0.00000026, -0.00000243, -0.00000301, -0.00000188, 0.00000002, 0.00000137, 0.00000130, -0.00000010, -0.00000196, -0.00000322, -0.00000335, -0.00000258, -0.00000161, -0.00000102, -0.00000084, -0.00000058, 0.00000025, 0.00000160, 0.00000268, 0.00000242, 0.00000022, -0.00000332, -0.00000654, -0.00000750, -0.00000516, -0.00000023, 0.00000504, 0.00000793, 0.00000683, 0.00000216, -0.00000373, -0.00000787, -0.00000832, -0.00000528, -0.00000085, 0.00000229, 0.00000253, 0.00000025, -0.00000248, -0.00000336, -0.00000140, 0.00000240, 0.00000563, 0.00000603, 0.00000294, -0.00000228, -0.00000708, -0.00000934, -0.00000866, -0.00000641, -0.00000468, -0.00000464, -0.00000570, -0.00000586, -0.00000333, 0.00000188, 0.00000738, 0.00000950, 0.00000547, -0.00000458, -0.00001704, -0.00002633, -0.00002764, -0.00001952, -0.00000476, 0.00001086, 0.00002140, 0.00002352, 0.00001785, 0.00000826, -0.00000040, -0.00000474, -0.00000414, -0.00000051, 0.00000330, 0.00000511, 0.00000445, 0.00000231, 0.00000017, -0.00000095, -0.00000097, -0.00000046, 0.00000000, 0.00000022, 0.00000043, 0.00000091, 0.00000163, 0.00000216, 0.00000199, 0.00000102, -0.00000034, -0.00000128, -0.00000118, 0.00000003, 0.00000171, 0.00000297, 0.00000316, 0.00000229, 0.00000094, -0.00000010, -0.00000036, 0.00000008, 0.00000075, 0.00000120, 0.00000127, 0.00000117, 0.00000114, 0.00000113, 0.00000076, -0.00000044, -0.00000250, -0.00000468, -0.00000568, -0.00000428, -0.00000018, 0.00000558, 0.00001088, 0.00001349, 0.00001220, 0.00000752, 0.00000149, -0.00000332, -0.00000510, -0.00000366, -0.00000039, 0.00000256, 0.00000356, 0.00000228, },
		{ -0.00000238, 0.00000009, 0.00000142, 0.00000097, -0.00000082, -0.00000283, -0.00000404, -0.00000407, -0.00000325, -0.00000225, -0.00000147, -0.00000089, -0.00000019, 0.00000073, 0.00000147, 0.00000127, -0.00000043, -0.00000340, -0.00000643, -0.00000786, -0.00000652, -0.00000262, 0.00000219, 0.00000551, 0.00000554, 0.00000215, -0.00000299, -0.00000721, -0.00000842, -0.00000624, -0.00000218, 0.00000130, 0.00000235, 0.00000077, -0.00000185, -0.00000330, -0.00000217, 0.00000117, 0.00000477, 0.00000631, 0.00000461, 0.00000034, -0.00000438, -0.00000730, -0.00000743, -0.00000549, -0.00000321, -0.00000202, -0.00000194, -0.00000161, 0.00000061, 0.00000522, 0.00001075, 0.00001444, 0.00001387, 0.00000874, 0.00000150, -0.00000372, -0.00000343, 0.00000304, 0.00001280, 0.00002062, 0.00002168, 0.00001409, 0.00000008, -0.00001513, -0.00002587, -0.00002873, -0.00002401, -0.00001509, -0.00000644, -0.00000134, -0.00000057, -0.00000266, -0.00000521, -0.00000640, -0.00000581, -0.00000421, -0.00000269, -0.00000182, -0.00000144, -0.00000103, -0.00000026, 0.00000065, 0.00000112, 0.00000075, -0.00000028, -0.00000124, -0.00000131, -0.00000016, 0.00000179, 0.00000358, 0.00000431, 0.00000371, 0.00000227, 0.00000088, 0.00000028, 0.00000060, 0.00000141, 0.00000213, 0.00000240, 0.00000235, 0.00000229, 0.00000236, 0.00000227, 0.00000146, -0.00000035, -0.00000271, -0.00000438, -0.00000391, -0.00000053, 0.00000512, 0.00001105, 0.00001472, 0.00001438, 0.00000997, 0.00000338, -0.00000255, -0.00000536, -0.00000427, -0.00000048, 0.00000359, 0.00000560, 0.00000456, 0.00000121, -0.00000253, -0.00000471, -0.00000448, },
		{ 0.00000052, -0.00000188, -0.00000404, -0.00000494, -0.00000445, -0.00000311, -0.00000165, -0.00000048, 0.00000037, 0.00000105, 0.00000146, 0.00000122, -0.00000010, -0.00000247, -0.00000511, -0.00000672, -0.00000615, -0.00000324, 0.00000091, 0.00000433, 0.00000522, 0.00000301, -0.00000121, -0.00000526, -0.00000703, -0.00000572, -0.00000228, 0.00000122, 0.00000280, 0.00000183, -0.00000063, -0.00000258, -0.00000237, 0.00000018, 0.00000362, 0.00000577, 0.00000507, 0.00000156, -0.00000312, -0.00000681, -0.00000810, -0.00000709, -0.00000512, -0.00000367, -0.00000327, -0.00000311, -0.00000178, 0.00000148, 0.00000590, 0.00000933, 0.00000950, 0.00000554, -0.00000115, -0.00000751, -0.00001057, -0.00000929, -0.00000537, -0.00000238, -0.00000357, -0.00000975, -0.00001845, -0.00002503, -0.00002511, -0.00001709, -0.00000312, 0.00001184, 0.00002248, 0.00002570, 0.00002183, 0.00001407, 0.00000652, 0.00000211, 0.00000143, 0.00000310, 0.00000501, 0.00000578, 0.00000531, 0.00000444, 0.00000406, 0.00000436, 0.00000479, 0.00000460, 0.00000354, 0.00000212, 0.00000127, 0.00000170, 0.00000327, 0.00000510, 0.00000608, 0.00000559, 0.00000387, 0.00000187, 0.00000055, 0.00000037, 0.00000102, 0.00000180, 0.00000217, 0.00000205, 0.00000180, 0.00000172, 0.00000171, 0.00000121, -0.00000029, -0.00000268, -0.00000490, -0.00000538, -0.00000291, 0.00000238, 0.00000877, 0.00001356, 0.00001441, 0.00001064, 0.00000377, -0.00000319, -0.00000717, -0.00000668, -0.00000249, 0.00000281, 0.00000622, 0.00000594, 0.00000226, -0.00000282, -0.00000668, -0.00000762, -0.00000560, -0.00000210, 0.00000083, 0.00000176, },
		{ -0.00000495, -0.00000356, -0.00000153, 0.00000027, 0.00000140, 0.00000190, 0.00000191, 0.00000143, 0.00000025, -0.00000165, -0.00000384, -0.00000543, -0.00000544, -0.00000348, -0.00000015, 0.00000305, 0.00000450, 0.00000335, 0.00000014, -0.00000345, -0.00000549, -0.00000488, -0.00000208, 0.00000126, 0.00000326, 0.00000295, 0.00000089, -0.00000122, -0.00000166, 0.00000020, 0.00000341, 0.00000606, 0.00000639, 0.00000394, -0.00000023, -0.00000414, -0.00000614, -0.00000580, -0.00000400, -0.00000213, -0.00000103, -0.00000041, 0.00000084, 0.00000361, 0.00000767, 0.00001144, 0.00001288, 0.00001077, 0.00000577, 0.00000017, -0.00000329, -0.00000311, 0.00000008, 0.00000388, 0.00000571, 0.00000456, 0.00000181, 0.00000041, 0.00000294, 0.00000974, 0.00001817, 0.00002379, 0.00002271, 0.00001385, -0.00000029, -0.00001469, -0.00002433, -0.00002653, -0.00002203, -0.00001420, -0.00000702, -0.00000304, -0.00000246, -0.00000365, -0.00000459, -0.00000420, -0.00000282, -0.00000159, -0.00000144, -0.00000232, -0.00000326, -0.00000318, -0.00000166, 0.00000064, 0.00000247, 0.00000280, 0.00000150, -0.00000060, -0.00000230, -0.00000277, -0.00000205, -0.00000085, 0.00000002, 0.00000025, 0.00000011, 0.00000006, 0.00000024, 0.00000022, -0.00000068, -0.00000269, -0.00000507, -0.00000625, -0.00000467, 0.00000013, 0.00000685, 0.00001277, 0.00001507, 0.00001235, 0.00000556, -0.00000227, -0.00000756, -0.00000798, -0.00000370, 0.00000273, 0.00000774, 0.00000866, 0.00000505, -0.00000116, -0.00000684, -0.00000936, -0.00000791, -0.00000374, 0.00000072, 0.00000322, 0.00000284, 0.00000028, -0.00000280, -0.00000479, },
		{ 0.00000152, 0.00000201, 0.00000166, 0.00000080, -0.00000040, -0.00000195, -0.00000367, -0.00000507, -0.00000544, -0.00000431, -0.00000189, 0.00000081, 0.00000246, 0.00000213, -0.00000012, -0.00000312, -0.00000521, -0.00000522, -0.00000316, -0.00000024, 0.00000184, 0.00000196, 0.00000025, -0.00000198, -0.00000307, -0.00000207, 0.00000056, 0.00000324, 0.00000418, 0.00000250, -0.00000124, -0.00000540, -0.00000823, -0.00000890, -0.00000779, -0.00000610, -0.00000484, -0.00000414, -0.00000326, -0.00000130, 0.00000186, 0.00000515, 0.00000683, 0.00000554, 0.00000130, -0.00000423, -0.00000868, -0.00001029, -0.00000898, -0.00000636, -0.00000465, -0.00000521, -0.00000761, -0.00001004, -0.00001062, -0.00000895, -0.00000663, -0.00000644, -0.00001044, -0.00001826, -0.00002675, -0.00003131, -0.00002835, -0.00001740, -0.00000164, 0.00001349, 0.00002295, 0.00002445, 0.00001930, 0.00001134, 0.00000471, 0.00000173, 0.00000216, 0.00000392, 0.00000481, 0.00000391, 0.00000195, 0.00000052, 0.00000078, 0.00000260, 0.00000471, 0.00000560, 0.00000455, 0.00000208, -0.00000046, -0.00000182, -0.00000161, -0.00000041, 0.00000075, 0.00000119, 0.00000098, 0.00000064, 0.00000060, 0.00000067, 0.00000016, -0.00000149, -0.00000395, -0.00000585, -0.00000537, -0.00000150, 0.00000506, 0.00001178, 0.00001547, 0.00001399, 0.00000756, -0.00000104, -0.00000782, -0.00000960, -0.00000573, 0.00000152, 0.00000816, 0.00001061, 0.00000753, 0.00000048, -0.00000705, -0.00001151, -0.00001115, -0.00000672, -0.00000088, 0.00000335, 0.00000420, 0.00000183, -0.00000203, -0.00000521, -0.00000626, -0.00000507, -0.00000258, -0.00000008, },
		{ -0.00000231, -0.00000369, -0.00000499, -0.00000609, -0.00000662, -0.00000618, -0.00000470, -0.00000270, -0.00000113, -0.00000090, -0.00000223, -0.00000447, -0.00000634, -0.00000671, -0.00000528, -0.00000282, -0.00000073, -0.00000021, -0.00000145, -0.00000352, -0.00000491, -0.00000451, -0.00000236, 0.00000034, 0.00000191, 0.00000118, -0.00000175, -0.00000563, -0.00000877, -0.00001002, -0.00000934, -0.00000761, -0.00000588, -0.00000463, -0.00000346, -0.00000161, 0.00000126, 0.00000456, 0.00000688, 0.00000683, 0.00000405, -0.00000043, -0.00000460, -0.00000665, -0.00000596, -0.00000350, -0.00000107, -0.00000013, -0.00000088, -0.00000215, -0.00000242, -0.00000098, 0.00000142, 0.00000303, 0.00000247, 0.00000003, -0.00000211, -0.00000098, 0.00000514, 0.00001521, 0.00002527, 0.00003022, 0.00002644, 0.00001394, -0.00000330, -0.00001917, -0.00002828, -0.00002853, -0.00002181, -0.00001261, -0.00000546, -0.00000267, -0.00000359, -0.00000562, -0.00000616, -0.00000425, -0.00000094, 0.00000174, 0.00000218, 0.00000032, -0.00000250, -0.00000452, -0.00000470, -0.00000329, -0.00000142, -0.00000022, -0.00000005, -0.00000041, -0.00000062, -0.00000048, -0.00000053, -0.00000155, -0.00000368, -0.00000590, -0.00000637, -0.00000354, 0.00000256, 0.00000983, 0.00001493, 0.00001500, 0.00000940, 0.00000039, -0.00000782, -0.00001125, -0.00000827, -0.00000057, 0.00000767, 0.00001197, 0.00000995, 0.00000249, -0.00000676, -0.00001344, -0.00001468, -0.00001051, -0.00000357, 0.00000246, 0.00000484, 0.00000299, -0.00000148, -0.00000591, -0.00000814, -0.00000747, -0.00000473, -0.00000157, 0.00000061, 0.00000121, 0.00000051, -0.00000084, },
		{ -0.00000508, -0.00000500, -0.00000426, -0.00000295, -0.00000160, -0.00000092, -0.00000134, -0.00000263, -0.00000395, -0.00000434, -0.00000332, -0.00000128, 0.00000072, 0.00000156, 0.00000081, -0.00000094, -0.00000246, -0.00000255, -0.00000093, 0.00000160, 0.00000353, 0.00000357, 0.00000138, -0.00000217, -0.00000557, -0.00000749, -0.00000749, -0.00000612, -0.00000435, -0.00000288, -0.00000167, -0.00000015, 0.00000211, 0.00000488, 0.00000710, 0.00000752, 0.00000548, 0.00000151, -0.00000280, -0.00000573, -0.00000633, -0.00000500, -0.00000313, -0.00000218, -0.00000269, -0.00000406, -0.00000510, -0.00000500, -0.00000403, -0.00000337, -0.00000418, -0.00000651, -0.00000902, -0.00000970, -0.00000740, -0.00000303, 0.00000046, -0.00000042, -0.00000742, -0.00001898, -0.00003032, -0.00003559, -0.00003083, -0.00001632, 0.00000325, 0.00002092, 0.00003076, 0.00003064, 0.00002293, 0.00001283, 0.00000547, 0.00000335, 0.00000556, 0.00000892, 0.00001028, 0.00000836, 0.00000421, 0.00000018, -0.00000178, -0.00000124, 0.00000067, 0.00000231, 0.00000268, 0.00000195, 0.00000102, 0.00000055, 0.00000040, -0.00000026, -0.00000204, -0.00000446, -0.00000588, -0.00000438, 0.00000073, 0.00000802, 0.00001424, 0.00001594, 0.00001154, 0.00000253, -0.00000698, -0.00001230, -0.00001071, -0.00000306, 0.00000658, 0.00001298, 0.00001258, 0.00000533, -0.00000531, -0.00001427, -0.00001752, -0.00001410, -0.00000633, 0.00000157, 0.00000585, 0.00000503, 0.00000036, -0.00000519, -0.00000869, -0.00000871, -0.00000583, -0.00000189, 0.00000113, 0.00000216, 0.00000134, -0.00000042, -0.00000217, -0.00000342, -0.00000421, -0.00000475, },
		{ -0.00000256, -0.00000183, -0.00000162, -0.00000209, -0.00000284, -0.00000318, -0.00000257, -0.00000108, 0.00000057, 0.00000145, 0.00000100, -0.00000053, -0.00000215, -0.00000272, -0.00000172, 0.00000042, 0.00000244, 0.00000302, 0.00000150, -0.00000165, -0.00000515, -0.00000761, -0.00000829, -0.00000735, -0.00000562, -0.00000390, -0.00000247, -0.00000100, 0.00000094, 0.00000336, 0.00000558, 0.00000653, 0.00000542, 0.00000239, -0.00000142, -0.00000447, -0.00000563, -0.00000486, -0.00000311, -0.00000168, -0.00000135, -0.00000193, -0.00000254, -0.00000238, -0.00000141, -0.00000039, -0.00000027, -0.00000128, -0.00000256, -0.00000257, -0.00000024, 0.00000397, 0.00000799, 0.00000927, 0.00000654, 0.00000109, -0.00000339, -0.00000271, 0.00000521, 0.00001854, 0.00003180, 0.00003828, 0.00003351, 0.00001780, -0.00000356, -0.00002272, -0.00003302, -0.00003210, -0.00002265, -0.00001061, -0.00000186, 0.00000063, -0.00000227, -0.00000705, -0.00001006, -0.00000960, -0.00000645, -0.00000286, -0.00000076, -0.00000066, -0.00000163, -0.00000238, -0.00000240, -0.00000228, -0.00000300, -0.00000475, -0.00000636, -0.00000582, -0.00000171, 0.00000537, 0.00001260, 0.00001620, 0.00001361, 0.00000532, -0.00000498, -0.00001214, -0.00001230, -0.00000520, 0.00000553, 0.00001417, 0.00001594, 0.00000959, -0.00000194, -0.00001313, -0.00001879, -0.00001670, -0.00000859, 0.00000104, 0.00000746, 0.00000807, 0.00000354, -0.00000305, -0.00000800, -0.00000907, -0.00000634, -0.00000176, 0.00000216, 0.00000376, 0.00000289, 0.00000062, -0.00000165, -0.00000305, -0.00000351, -0.00000352, -0.00000357, -0.00000373, -0.00000374, -0.00000334, },
		{ -0.00000251, -0.00000250, -0.00000205, -0.00000099, 0.00000034, 0.00000122, 0.00000106, -0.00000012, -0.00000164, -0.00000247, -0.00000193, -0.00000016, 0.00000190, 0.00000295, 0.00000214, -0.00000044, -0.00000382, -0.00000663, -0.00000789, -0.00000743, -0.00000584, -0.00000396, -0.00000228, -0.00000078, 0.00000090, 0.00000292, 0.00000492, 0.00000606, 0.00000556, 0.00000326, -0.00000013, -0.00000332, -0.00000514, -0.00000523, -0.00000413, -0.00000291, -0.00000243, -0.00000281, -0.00000350, -0.00000378, -0.00000345, -0.00000295, -0.00000306, -0.00000411, -0.00000556, -0.00000623, -0.00000512, -0.00000238, 0.00000047, 0.00000129, -0.00000121, -0.00000635, -0.00001144, -0.00001324, -0.00001009, -0.00000343, 0.00000248, 0.00000273, -0.00000524, -0.00001956, -0.00003417, -0.00004150, -0.00003642, -0.00001906, 0.00000483, 0.00002653, 0.00003853, 0.00003793, 0.00002744, 0.00001345, 0.00000252, -0.00000184, -0.00000020, 0.00000398, 0.00000701, 0.00000704, 0.00000471, 0.00000198, 0.00000040, -0.00000001, -0.00000038, -0.00000171, -0.00000367, -0.00000447, -0.00000216, 0.00000368, 0.00001098, 0.00001593, 0.00001514, 0.00000788, -0.00000295, -0.00001204, -0.00001454, -0.00000881, 0.00000233, 0.00001297, 0.00001722, 0.00001239, 0.00000051, -0.00001275, -0.00002112, -0.00002099, -0.00001313, -0.00000203, 0.00000667, 0.00000917, 0.00000523, -0.00000213, -0.00000861, -0.00001098, -0.00000865, -0.00000351, 0.00000143, 0.00000386, 0.00000317, 0.00000048, -0.00000239, -0.00000408, -0.00000429, -0.00000369, -0.00000319, -0.00000330, -0.00000386, -0.00000431, -0.00000423, -0.00000364, -0.00000293, -0.00000254, },
		{ 0.00000088, 0.00000154, 0.00000148, 0.00000053, -0.00000091, -0.00000199, -0.00000196, -0.00000069, 0.00000117, 0.00000246, 0.00000219, 0.00000011, -0.00000310, -0.00000621, -0.00000808, -0.00000822, -0.00000696, -0.00000506, -0.00000320, -0.00000162, -0.00000012, 0.00000154, 0.00000325, 0.00000445, 0.00000443, 0.00000284, 0.00000006, -0.00000294, -0.00000508, -0.00000574, -0.00000513, -0.00000403, -0.00000327, -0.00000318, -0.00000348, -0.00000362, -0.00000329, -0.00000274, -0.00000253, -0.00000305, -0.00000397, -0.00000435, -0.00000329, -0.00000068, 0.00000238, 0.00000408, 0.00000310, -0.00000032, -0.00000415, -0.00000563, -0.00000298, 0.00000311, 0.00000955, 0.00001246, 0.00000969, 0.00000259, -0.00000415, -0.00000492, 0.00000344, 0.00001921, 0.00003589, 0.00004516, 0.00004104, 0.00002325, -0.00000234, -0.00002645, -0.00004081, -0.00004197, -0.00003240, -0.00001860, -0.00000733, -0.00000230, -0.00000305, -0.00000629, -0.00000860, -0.00000854, -0.00000705, -0.00000602, -0.00000635, -0.00000696, -0.00000559, -0.00000074, 0.00000668, 0.00001331, 0.00001512, 0.00001009, -0.00000007, -0.00001033, -0.00001500, -0.00001098, 0.00000016, 0.00001273, 0.00001993, 0.00001750, 0.00000621, -0.00000857, -0.00001969, -0.00002201, -0.00001502, -0.00000285, 0.00000819, 0.00001297, 0.00001011, 0.00000225, -0.00000579, -0.00000980, -0.00000828, -0.00000285, 0.00000310, 0.00000648, 0.00000611, 0.00000299, -0.00000069, -0.00000297, -0.00000320, -0.00000210, -0.00000101, -0.00000093, -0.00000194, -0.00000333, -0.00000420, -0.00000410, -0.00000323, -0.00000216, -0.00000137, -0.00000095, -0.00000059, 0.00000003, },
		{ -0.00000118, -0.00000229, -0.00000259, -0.00000171, -0.00000002, 0.00000151, 0.00000182, 0.00000039, -0.00000243, -0.00000559, -0.00000787, -0.00000851, -0.00000754, -0.00000559, -0.00000346, -0.00000162, -0.00000009, 0.00000138, 0.00000285, 0.00000405, 0.00000441, 0.00000350, 0.00000141, -0.00000122, -0.00000345, -0.00000457, -0.00000448, -0.00000369, -0.00000290, -0.00000254, -0.00000259, -0.00000266, -0.00000246, -0.00000209, -0.00000195, -0.00000239, -0.00000326, -0.00000384, -0.00000328, -0.00000134, 0.00000121, 0.00000286, 0.00000226, -0.00000070, -0.00000458, -0.00000699, -0.00000608, -0.00000195, 0.00000312, 0.00000578, 0.00000375, -0.00000250, -0.00000960, -0.00001309, -0.00001024, -0.00000217, 0.00000621, 0.00000865, 0.00000130, -0.00001451, -0.00003221, -0.00004298, -0.00004024, -0.00002328, 0.00000214, 0.00002650, 0.00004126, 0.00004279, 0.00003359, 0.00002021, 0.00000934, 0.00000441, 0.00000458, 0.00000642, 0.00000686, 0.00000543, 0.00000422, 0.00000589, 0.00001112, 0.00001753, 0.00002082, 0.00001762, 0.00000807, -0.00000363, -0.00001129, -0.00001032, -0.00000069, 0.00001266, 0.00002231, 0.00002243, 0.00001214, -0.00000394, -0.00001810, -0.00002368, -0.00001858, -0.00000614, 0.00000689, 0.00001409, 0.00001270, 0.00000464, -0.00000499, -0.00001095, -0.00001067, -0.00000526, 0.00000164, 0.00000617, 0.00000639, 0.00000299, -0.00000153, -0.00000459, -0.00000501, -0.00000342, -0.00000153, -0.00000093, -0.00000209, -0.00000426, -0.00000612, -0.00000668, -0.00000578, -0.00000406, -0.00000238, -0.00000128, -0.00000073, -0.00000035, 0.00000011, 0.00000057, 0.00000063, 0.00000000, },
		{ 0.00000011, 0.00000176, 0.00000257, 0.00000178, -0.00000058, -0.00000369, -0.00000632, -0.00000748, -0.00000689, -0.00000501, -0.00000267, -0.00000055, 0.00000109, 0.00000241, 0.00000363, 0.00000469, 0.00000525, 0.00000488, 0.00000344, 0.00000129, -0.00000087, -0.00000230, -0.00000270, -0.00000226, -0.00000153, -0.00000098, -0.00000073, -0.00000060, -0.00000035, 0.00000002, 0.00000022, -0.00000003, -0.00000068, -0.00000117, -0.00000076, 0.00000091, 0.00000335, 0.00000529, 0.00000546, 0.00000344, 0.00000021, -0.00000218, -0.00000189, 0.00000149, 0.00000637, 0.00000990, 0.00000973, 0.00000564, 0.00000005, -0.00000322, -0.00000131, 0.00000563, 0.00001418, 0.00001938, 0.00001787, 0.00001028, 0.00000145, -0.00000209, 0.00000396, 0.00001861, 0.00003540, 0.00004536, 0.00004176, 0.00002388, -0.00000230, -0.00002700, -0.00004160, -0.00004276, -0.00003349, -0.00002081, -0.00001137, -0.00000788, -0.00000835, -0.00000835, -0.00000460, 0.00000240, 0.00000865, 0.00000947, 0.00000298, -0.00000804, -0.00001737, -0.00001899, -0.00001088, 0.00000332, 0.00001605, 0.00002000, 0.00001234, -0.00000358, -0.00001990, -0.00002856, -0.00002560, -0.00001315, 0.00000202, 0.00001229, 0.00001328, 0.00000588, -0.00000482, -0.00001263, -0.00001375, -0.00000848, -0.00000050, 0.00000556, 0.00000683, 0.00000351, -0.00000174, -0.00000567, -0.00000643, -0.00000438, -0.00000154, -0.00000009, -0.00000106, -0.00000382, -0.00000670, -0.00000810, -0.00000740, -0.00000515, -0.00000255, -0.00000061, 0.00000033, 0.00000059, 0.00000069, 0.00000088, 0.00000099, 0.00000066, -0.00000022, -0.00000127, -0.00000181, -0.00000133, },
		{ 0.00000055, -0.00000250, -0.00000551, -0.00000729, -0.00000727, -0.00000566, -0.00000328, -0.00000098, 0.00000073, 0.00000186, 0.00000268, 0.00000339, 0.00000386, 0.00000374, 0.00000275, 0.00000097, -0.00000113, -0.00000290, -0.00000389, -0.00000405, -0.00000369, -0.00000325, -0.00000296, -0.00000282, -0.00000269, -0.00000254, -0.00000255, -0.00000294, -0.00000370, -0.00000441, -0.00000443, -0.00000335, -0.00000145, 0.00000028, 0.00000061, -0.00000106, -0.00000414, -0.00000698, -0.00000778, -0.00000577, -0.00000194, 0.00000132, 0.00000168, -0.00000163, -0.00000701, -0.00001125, -0.00001147, -0.00000704, -0.00000033, 0.00000449, 0.00000396, -0.00000233, -0.00001118, -0.00001753, -0.00001762, -0.00001174, -0.00000448, -0.00000234, -0.00000961, -0.00002522, -0.00004243, -0.00005201, -0.00004722, -0.00002774, -0.00000030, 0.00002462, 0.00003820, 0.00003765, 0.00002721, 0.00001519, 0.00000888, 0.00001045, 0.00001620, 0.00001951, 0.00001555, 0.00000469, -0.00000749, -0.00001366, -0.00000948, 0.00000336, 0.00001773, 0.00002501, 0.00002013, 0.00000464, -0.00001402, -0.00002660, -0.00002711, -0.00001598, 0.00000038, 0.00001339, 0.00001693, 0.00001045, -0.00000131, -0.00001137, -0.00001454, -0.00001006, -0.00000140, 0.00000614, 0.00000863, 0.00000552, -0.00000054, -0.00000566, -0.00000716, -0.00000494, -0.00000118, 0.00000129, 0.00000076, -0.00000246, -0.00000642, -0.00000892, -0.00000874, -0.00000621, -0.00000274, 0.00000012, 0.00000158, 0.00000181, 0.00000152, 0.00000133, 0.00000134, 0.00000122, 0.00000065, -0.00000030, -0.00000109, -0.00000107, -0.00000003, 0.00000156, 0.00000271, 0.00000247, },
		{ -0.00001046, -0.00000922, -0.00000684, -0.00000433, -0.00000243, -0.00000135, -0.00000081, -0.00000043, -0.00000008, 0.00000002, -0.00000045, -0.00000167, -0.00000344, -0.00000523, -0.00000654, -0.00000708, -0.00000695, -0.00000647, -0.00000594, -0.00000550, -0.00000513, -0.00000483, -0.00000469, -0.00000488, -0.00000539, -0.00000590, -0.00000586, -0.00000485, -0.00000299, -0.00000105, -0.00000011, -0.00000090, -0.00000319, -0.00000566, -0.00000664, -0.00000514, -0.00000162, 0.00000204, 0.00000360, 0.00000192, -0.00000211, -0.00000587, -0.00000655, -0.00000295, 0.00000351, 0.00000939, 0.00001126, 0.00000791, 0.00000135, -0.00000424, -0.00000512, -0.00000046, 0.00000686, 0.00001194, 0.00001111, 0.00000472, -0.00000244, -0.00000374, 0.00000502, 0.00002230, 0.00004063, 0.00005002, 0.00004351, 0.00002134, -0.00000843, -0.00003382, -0.00004503, -0.00003954, -0.00002318, -0.00000658, 0.00000098, -0.00000328, -0.00001447, -0.00002345, -0.00002260, -0.00001070, 0.00000626, 0.00001865, 0.00001891, 0.00000617, -0.00001311, -0.00002880, -0.00003272, -0.00002318, -0.00000563, 0.00001064, 0.00001772, 0.00001328, 0.00000129, -0.00001069, -0.00001604, -0.00001260, -0.00000333, 0.00000593, 0.00001009, 0.00000764, 0.00000097, -0.00000546, -0.00000802, -0.00000592, -0.00000132, 0.00000234, 0.00000252, -0.00000096, -0.00000606, -0.00000996, -0.00001071, -0.00000825, -0.00000412, -0.00000036, 0.00000166, 0.00000186, 0.00000106, 0.00000029, -0.00000002, -0.00000010, -0.00000047, -0.00000135, -0.00000242, -0.00000296, -0.00000250, -0.00000118, 0.00000013, 0.00000036, -0.00000111, -0.00000405, -0.00000740, -0.00000983, },
		{ -0.00000097, 0.00000037, 0.00000086, 0.00000105, 0.00000131, 0.00000165, 0.00000172, 0.00000118, -0.00000006, -0.00000167, -0.00000314, -0.00000402, -0.00000417, -0.00000374, -0.00000307, -0.00000238, -0.00000183, -0.00000144, -0.00000129, -0.00000146, -0.00000192, -0.00000244, -0.00000255, -0.00000186, -0.00000036, 0.00000138, 0.00000242, 0.00000198, 0.00000004, -0.00000245, -0.00000401, -0.00000349, -0.00000097, 0.00000219, 0.00000394, 0.00000291, -0.00000061, -0.00000462, -0.00000649, -0.00000464, 0.00000030, 0.00000559, 0.00000796, 0.00000563, -0.00000042, -0.00000687, -0.00001008, -0.00000846, -0.00000362, 0.00000053, 0.00000034, -0.00000492, -0.00001231, -0.00001685, -0.00001496, -0.00000737, 0.00000049, 0.00000139, -0.00000910, -0.00002891, -0.00004934, -0.00005893, -0.00004970, -0.00002208, 0.00001439, 0.00004533, 0.00005874, 0.00005113, 0.00002922, 0.00000613, -0.00000587, -0.00000191, 0.00001328, 0.00002851, 0.00003296, 0.00002243, 0.00000160, -0.00001884, -0.00002844, -0.00002286, -0.00000606, 0.00001230, 0.00002249, 0.00002001, 0.00000762, -0.00000686, -0.00001535, -0.00001404, -0.00000501, 0.00000556, 0.00001139, 0.00000971, 0.00000239, -0.00000568, -0.00000983, -0.00000833, -0.00000310, 0.00000185, 0.00000308, -0.00000033, -0.00000644, -0.00001182, -0.00001374, -0.00001153, -0.00000672, -0.00000184, 0.00000110, 0.00000160, 0.00000054, -0.00000069, -0.00000121, -0.00000108, -0.00000096, -0.00000142, -0.00000241, -0.00000325, -0.00000317, -0.00000200, -0.00000036, 0.00000059, -0.00000010, -0.00000255, -0.00000590, -0.00000872, -0.00000983, -0.00000883, -0.00000630, -0.00000333, },
		{ 0.00000059, 0.00000078, 0.00000107, 0.00000095, 0.00000010, -0.00000143, -0.00000317, -0.00000454, -0.00000519, -0.00000507, -0.00000446, -0.00000370, -0.00000305, -0.00000264, -0.00000255, -0.00000277, -0.00000325, -0.00000379, -0.00000400, -0.00000354, -0.00000231, -0.00000073, 0.00000044, 0.00000043, -0.00000098, -0.00000315, -0.00000483, -0.00000485, -0.00000294, -0.00000003, 0.00000214, 0.00000211, -0.00000034, -0.00000376, -0.00000584, -0.00000478, -0.00000057, 0.00000478, 0.00000828, 0.00000781, 0.00000350, -0.00000214, -0.00000573, -0.00000512, -0.00000089, 0.00000398, 0.00000603, 0.00000373, -0.00000125, -0.00000499, -0.00000392, 0.00000243, 0.00001059, 0.00001505, 0.00001199, 0.00000247, -0.00000710, -0.00000821, 0.00000452, 0.00002900, 0.00005526, 0.00006972, 0.00006223, 0.00003216, -0.00001028, -0.00004836, -0.00006704, -0.00006047, -0.00003487, -0.00000509, 0.00001339, 0.00001238, -0.00000519, -0.00002762, -0.00004144, -0.00003887, -0.00002169, 0.00000044, 0.00001591, 0.00001771, 0.00000689, -0.00000876, -0.00001970, -0.00002015, -0.00001095, 0.00000167, 0.00001023, 0.00001047, 0.00000340, -0.00000581, -0.00001141, -0.00001051, -0.00000450, 0.00000221, 0.00000515, 0.00000245, -0.00000424, -0.00001104, -0.00001429, -0.00001255, -0.00000714, -0.00000105, 0.00000303, 0.00000402, 0.00000274, 0.00000092, -0.00000002, 0.00000015, 0.00000068, 0.00000064, -0.00000029, -0.00000154, -0.00000210, -0.00000140, 0.00000024, 0.00000167, 0.00000163, -0.00000042, -0.00000389, -0.00000733, -0.00000923, -0.00000884, -0.00000649, -0.00000334, -0.00000067, 0.00000077, 0.00000104, 0.00000076, },
		{ 0.00000017, -0.00000093, -0.00000262, -0.00000424, -0.00000521, -0.00000527, -0.00000460, -0.00000359, -0.00000265, -0.00000204, -0.00000183, -0.00000199, -0.00000239, -0.00000285, -0.00000304, -0.00000268, -0.00000166, -0.00000024, 0.00000096, 0.00000124, 0.00000026, -0.00000160, -0.00000336, -0.00000392, -0.00000277, -0.00000044, 0.00000167, 0.00000212, 0.00000036, -0.00000273, -0.00000520, -0.00000523, -0.00000229, 0.00000227, 0.00000587, 0.00000623, 0.00000282, -0.00000266, -0.00000721, -0.00000842, -0.00000600, -0.00000208, 0.00000019, -0.00000124, -0.00000577, -0.00001043, -0.00001182, -0.00000865, -0.00000297, 0.00000079, -0.00000120, -0.00000926, -0.00001918, -0.00002436, -0.00002002, -0.00000702, 0.00000738, 0.00001291, 0.00000223, -0.00002375, -0.00005472, -0.00007517, -0.00007221, -0.00004286, 0.00000338, 0.00004842, 0.00007417, 0.00007148, 0.00004445, 0.00000802, -0.00001995, -0.00002772, -0.00001504, 0.00000793, 0.00002696, 0.00003178, 0.00002115, 0.00000254, -0.00001303, -0.00001749, -0.00001002, 0.00000334, 0.00001388, 0.00001558, 0.00000826, -0.00000292, -0.00001110, -0.00001198, -0.00000614, 0.00000181, 0.00000631, 0.00000433, -0.00000307, -0.00001170, -0.00001691, -0.00001630, -0.00001078, -0.00000365, 0.00000159, 0.00000315, 0.00000163, -0.00000092, -0.00000251, -0.00000245, -0.00000150, -0.00000098, -0.00000166, -0.00000317, -0.00000437, -0.00000422, -0.00000263, -0.00000068, 0.00000013, -0.00000121, -0.00000449, -0.00000833, -0.00001094, -0.00001111, -0.00000888, -0.00000537, -0.00000212, -0.00000023, 0.00000014, -0.00000038, -0.00000085, -0.00000074, -0.00000014, 0.00000037, },
		{ -0.00000434, -0.00000491, -0.00000450, -0.00000354, -0.00000255, -0.00000195, -0.00000185, -0.00000220, -0.00000279, -0.00000339, -0.00000374, -0.00000362, -0.00000294, -0.00000186, -0.00000085, -0.00000050, -0.00000119, -0.00000279, -0.00000455, -0.00000550, -0.00000502, -0.00000329, -0.00000134, -0.00000049, -0.00000151, -0.00000396, -0.00000635, -0.00000693, -0.00000485, -0.00000084, 0.00000299, 0.00000442, 0.00000241, -0.00000205, -0.00000648, -0.00000839, -0.00000686, -0.00000318, -0.00000004, 0.00000025, -0.00000257, -0.00000639, -0.00000815, -0.00000597, -0.00000076, 0.00000410, 0.00000489, 0.00000037, -0.00000690, -0.00001155, -0.00000880, 0.00000207, 0.00001635, 0.00002603, 0.00002451, 0.00001134, -0.00000636, -0.00001695, -0.00001075, 0.00001361, 0.00004687, 0.00007271, 0.00007579, 0.00005007, 0.00000285, -0.00004805, -0.00008266, -0.00008856, -0.00006667, -0.00003016, 0.00000274, 0.00001805, 0.00001254, -0.00000602, -0.00002429, -0.00003124, -0.00002385, -0.00000772, 0.00000713, 0.00001261, 0.00000699, -0.00000481, -0.00001476, -0.00001691, -0.00001064, -0.00000057, 0.00000670, 0.00000671, -0.00000035, -0.00001015, -0.00001706, -0.00001749, -0.00001170, -0.00000317, 0.00000378, 0.00000650, 0.00000514, 0.00000203, -0.00000019, -0.00000026, 0.00000120, 0.00000252, 0.00000236, 0.00000074, -0.00000112, -0.00000170, -0.00000043, 0.00000190, 0.00000353, 0.00000297, -0.00000007, -0.00000436, -0.00000787, -0.00000894, -0.00000716, -0.00000354, 0.00000016, 0.00000242, 0.00000277, 0.00000182, 0.00000072, 0.00000037, 0.00000091, 0.00000173, 0.00000197, 0.00000113, -0.00000068, -0.00000276, },
		{ -0.00000460, -0.00000367, -0.00000338, -0.00000362, -0.00000412, -0.00000459, -0.00000478, -0.00000455, -0.00000385, -0.00000281, -0.00000176, -0.00000116, -0.00000140, -0.00000246, -0.00000387, -0.00000484, -0.00000470, -0.00000339, -0.00000159, -0.00000041, -0.00000070, -0.00000240, -0.00000446, -0.00000534, -0.00000401, -0.00000072, 0.00000298, 0.00000502, 0.00000408, 0.00000050, -0.00000381, -0.00000649, -0.00000616, -0.00000341, -0.00000037, 0.00000062, -0.00000131, -0.00000493, -0.00000759, -0.00000717, -0.00000365, 0.00000053, 0.00000200, -0.00000118, -0.00000773, -0.00001353, -0.00001397, -0.00000726, 0.00000379, 0.00001274, 0.00001317, 0.00000292, -0.00001379, -0.00002809, -0.00003139, -0.00002097, -0.00000251, 0.00001221, 0.00001171, -0.00000808, -0.00004023, -0.00006927, -0.00007871, -0.00005991, -0.00001723, 0.00003344, 0.00007220, 0.00008491, 0.00006983, 0.00003778, 0.00000591, -0.00001131, -0.00000905, 0.00000672, 0.00002384, 0.00003138, 0.00002552, 0.00001082, -0.00000348, -0.00000941, -0.00000484, 0.00000583, 0.00001489, 0.00001622, 0.00000878, -0.00000320, -0.00001310, -0.00001596, -0.00001111, -0.00000202, 0.00000613, 0.00000962, 0.00000801, 0.00000374, 0.00000016, -0.00000073, 0.00000074, 0.00000261, 0.00000287, 0.00000102, -0.00000175, -0.00000349, -0.00000298, -0.00000063, 0.00000170, 0.00000199, -0.00000065, -0.00000532, -0.00000981, -0.00001196, -0.00001082, -0.00000716, -0.00000292, -0.00000005, 0.00000054, -0.00000066, -0.00000232, -0.00000313, -0.00000265, -0.00000141, -0.00000049, -0.00000076, -0.00000234, -0.00000460, -0.00000656, -0.00000746, -0.00000711, -0.00000594, },
		{ -0.00000235, -0.00000281, -0.00000293, -0.00000264, -0.00000198, -0.00000108, -0.00000017, 0.00000042, 0.00000041, -0.00000031, -0.00000145, -0.00000243, -0.00000267, -0.00000191, -0.00000052, 0.00000066, 0.00000081, -0.00000029, -0.00000198, -0.00000303, -0.00000237, 0.00000015, 0.00000348, 0.00000584, 0.00000580, 0.00000319, -0.00000065, -0.00000362, -0.00000415, -0.00000218, 0.00000074, 0.00000244, 0.00000164, -0.00000115, -0.00000384, -0.00000425, -0.00000169, 0.00000231, 0.00000484, 0.00000365, -0.00000105, -0.00000623, -0.00000772, -0.00000299, 0.00000670, 0.00001639, 0.00002010, 0.00001455, 0.00000182, -0.00001122, -0.00001651, -0.00000972, 0.00000668, 0.00002421, 0.00003307, 0.00002812, 0.00001254, -0.00000331, -0.00000763, 0.00000562, 0.00003222, 0.00005888, 0.00006989, 0.00005561, 0.00001830, -0.00002830, -0.00006564, -0.00007952, -0.00006701, -0.00003740, -0.00000672, 0.00001082, 0.00000978, -0.00000481, -0.00002151, -0.00002956, -0.00002480, -0.00001124, 0.00000217, 0.00000733, 0.00000164, -0.00001096, -0.00002282, -0.00002739, -0.00002271, -0.00001209, -0.00000152, 0.00000405, 0.00000331, -0.00000131, -0.00000572, -0.00000699, -0.00000494, -0.00000178, -0.00000022, -0.00000148, -0.00000457, -0.00000713, -0.00000719, -0.00000461, -0.00000120, 0.00000051, -0.00000110, -0.00000560, -0.00001074, -0.00001383, -0.00001328, -0.00000948, -0.00000445, -0.00000061, 0.00000063, -0.00000053, -0.00000260, -0.00000387, -0.00000348, -0.00000180, -0.00000008, 0.00000044, -0.00000072, -0.00000304, -0.00000536, -0.00000663, -0.00000641, -0.00000505, -0.00000336, -0.00000209, -0.00000161, -0.00000182, },
		{ -0.00000168, -0.00000100, -0.00000037, 0.00000002, -0.00000000, -0.00000053, -0.00000147, -0.00000246, -0.00000303, -0.00000286, -0.00000205, -0.00000115, -0.00000085, -0.00000153, -0.00000290, -0.00000404, -0.00000399, -0.00000234, 0.00000031, 0.00000256, 0.00000300, 0.00000112, -0.00000229, -0.00000553, -0.00000696, -0.00000606, -0.00000383, -0.00000210, -0.00000238, -0.00000472, -0.00000765, -0.00000913, -0.00000803, -0.00000511, -0.00000267, -0.00000305, -0.00000682, -0.00001199, -0.00001501, -0.00001299, -0.00000599, 0.00000244, 0.00000692, 0.00000356, -0.00000734, -0.00002073, -0.00002935, -0.00002794, -0.00001668, -0.00000162, 0.00000849, 0.00000727, -0.00000532, -0.00002263, -0.00003517, -0.00003638, -0.00002694, -0.00001474, -0.00001024, -0.00001973, -0.00004063, -0.00006184, -0.00006930, -0.00005401, -0.00001766, 0.00002724, 0.00006325, 0.00007670, 0.00006448, 0.00003524, 0.00000445, -0.00001383, -0.00001390, -0.00000033, 0.00001566, 0.00002316, 0.00001751, 0.00000222, -0.00001392, -0.00002240, -0.00001976, -0.00000892, 0.00000326, 0.00001032, 0.00000982, 0.00000401, -0.00000226, -0.00000494, -0.00000315, 0.00000082, 0.00000347, 0.00000265, -0.00000106, -0.00000498, -0.00000632, -0.00000417, -0.00000011, 0.00000286, 0.00000231, -0.00000201, -0.00000797, -0.00001240, -0.00001294, -0.00000944, -0.00000387, 0.00000089, 0.00000279, 0.00000164, -0.00000103, -0.00000309, -0.00000316, -0.00000134, 0.00000100, 0.00000224, 0.00000146, -0.00000099, -0.00000385, -0.00000573, -0.00000590, -0.00000457, -0.00000264, -0.00000112, -0.00000056, -0.00000089, -0.00000164, -0.00000228, -0.00000249, -0.00000224, },
		{ -0.00000349, -0.00000389, -0.00000457, -0.00000540, -0.00000609, -0.00000635, -0.00000607, -0.00000550, -0.00000514, -0.00000539, -0.00000626, -0.00000719, -0.00000738, -0.00000633, -0.00000425, -0.00000213, -0.00000120, -0.00000221, -0.00000482, -0.00000775, -0.00000948, -0.00000921, -0.00000739, -0.00000548, -0.00000497, -0.00000637, -0.00000877, -0.00001042, -0.00000997, -0.00000752, -0.00000476, -0.00000386, -0.00000592, -0.00000988, -0.00001287, -0.00001198, -0.00000637, 0.00000167, 0.00000763, 0.00000736, -0.00000013, -0.00001149, -0.00002045, -0.00002134, -0.00001257, 0.00000199, 0.00001491, 0.00001930, 0.00001298, -0.00000006, -0.00001190, -0.00001528, -0.00000803, 0.00000573, 0.00001822, 0.00002311, 0.00001978, 0.00001398, 0.00001412, 0.00002538, 0.00004531, 0.00006388, 0.00006839, 0.00005084, 0.00001325, -0.00003212, -0.00006829, -0.00008190, -0.00006986, -0.00004068, -0.00000977, 0.00000861, 0.00000828, -0.00000664, -0.00002513, -0.00003589, -0.00003341, -0.00002016, -0.00000406, 0.00000664, 0.00000808, 0.00000221, -0.00000526, -0.00000890, -0.00000681, -0.00000125, 0.00000343, 0.00000398, 0.00000030, -0.00000471, -0.00000737, -0.00000577, -0.00000103, 0.00000348, 0.00000443, 0.00000070, -0.00000589, -0.00001176, -0.00001369, -0.00001079, -0.00000483, 0.00000090, 0.00000364, 0.00000262, -0.00000070, -0.00000378, -0.00000463, -0.00000295, -0.00000012, 0.00000183, 0.00000150, -0.00000107, -0.00000457, -0.00000725, -0.00000800, -0.00000683, -0.00000473, -0.00000295, -0.00000229, -0.00000278, -0.00000385, -0.00000477, -0.00000512, -0.00000489, -0.00000437, -0.00000384, -0.00000350, -0.00000337, },
		{ -0.00000330, -0.00000366, -0.00000370, -0.00000339, -0.00000297, -0.00000278, -0.00000301, -0.00000345, -0.00000358, -0.00000288, -0.00000129, 0.00000060, 0.00000181, 0.00000156, -0.00000021, -0.00000265, -0.00000448, -0.00000476, -0.00000351, -0.00000176, -0.00000088, -0.00000164, -0.00000363, -0.00000550, -0.00000584, -0.00000429, -0.00000191, -0.00000059, -0.00000171, -0.00000497, -0.00000824, -0.00000882, -0.00000529, 0.00000115, 0.00000692, 0.00000798, 0.00000243, -0.00000789, -0.00001790, -0.00002197, -0.00001728, -0.00000574, 0.00000664, 0.00001307, 0.00000986, -0.00000128, -0.00001414, -0.00002157, -0.00001968, -0.00001011, 0.00000111, 0.00000735, 0.00000553, -0.00000218, -0.00001012, -0.00001330, -0.00001123, -0.00000863, -0.00001237, -0.00002630, -0.00004730, -0.00006544, -0.00006868, -0.00004967, -0.00001092, 0.00003515, 0.00007155, 0.00008483, 0.00007170, 0.00004043, 0.00000637, -0.00001584, -0.00001925, -0.00000700, 0.00001072, 0.00002297, 0.00002399, 0.00001538, 0.00000393, -0.00000318, -0.00000277, 0.00000305, 0.00000896, 0.00001026, 0.00000590, -0.00000119, -0.00000634, -0.00000642, -0.00000185, 0.00000380, 0.00000621, 0.00000315, -0.00000412, -0.00001170, -0.00001548, -0.00001361, -0.00000748, -0.00000069, 0.00000319, 0.00000265, -0.00000111, -0.00000518, -0.00000689, -0.00000537, -0.00000191, 0.00000109, 0.00000163, -0.00000068, -0.00000453, -0.00000784, -0.00000903, -0.00000783, -0.00000526, -0.00000284, -0.00000171, -0.00000205, -0.00000319, -0.00000420, -0.00000447, -0.00000398, -0.00000314, -0.00000243, -0.00000208, -0.00000204, -0.00000212, -0.00000225, -0.00000247, -0.00000283, },
		{ -0.00000379, -0.00000354, -0.00000341, -0.00000351, -0.00000360, -0.00000331, -0.00000239, -0.00000106, 0.00000003, 0.00000017, -0.00000096, -0.00000293, -0.00000477, -0.00000554, -0.00000496, -0.00000362, -0.00000266, -0.00000295, -0.00000451, -0.00000637, -0.00000725, -0.00000647, -0.00000454, -0.00000296, -0.00000318, -0.00000544, -0.00000838, -0.00000965, -0.00000753, -0.00000232, 0.00000334, 0.00000579, 0.00000259, -0.00000569, -0.00001519, -0.00002073, -0.00001866, -0.00000928, 0.00000303, 0.00001201, 0.00001309, 0.00000611, -0.00000450, -0.00001218, -0.00001209, -0.00000392, 0.00000806, 0.00001764, 0.00002047, 0.00001657, 0.00000996, 0.00000576, 0.00000675, 0.00001159, 0.00001611, 0.00001678, 0.00001386, 0.00001195, 0.00001715, 0.00003233, 0.00005373, 0.00007120, 0.00007287, 0.00005172, 0.00001045, -0.00003849, -0.00007810, -0.00009478, -0.00008480, -0.00005576, -0.00002236, 0.00000119, 0.00000783, 0.00000006, -0.00001290, -0.00002117, -0.00001977, -0.00001062, -0.00000032, 0.00000462, 0.00000187, -0.00000570, -0.00001230, -0.00001325, -0.00000793, 0.00000005, 0.00000532, 0.00000431, -0.00000254, -0.00001118, -0.00001652, -0.00001556, -0.00000904, -0.00000075, 0.00000485, 0.00000534, 0.00000150, -0.00000350, -0.00000621, -0.00000506, -0.00000105, 0.00000306, 0.00000462, 0.00000268, -0.00000159, -0.00000577, -0.00000773, -0.00000680, -0.00000395, -0.00000101, 0.00000049, 0.00000014, -0.00000135, -0.00000279, -0.00000332, -0.00000282, -0.00000186, -0.00000111, -0.00000096, -0.00000129, -0.00000173, -0.00000197, -0.00000202, -0.00000211, -0.00000245, -0.00000303, -0.00000359, -0.00000387, },
		{ -0.00000278, -0.00000253, -0.00000202, -0.00000115, -0.00000020, 0.00000028, -0.00000014, -0.00000141, -0.00000292, -0.00000387, -0.00000378, -0.00000285, -0.00000191, -0.00000182, -0.00000290, -0.00000458, -0.00000579, -0.00000568, -0.00000434, -0.00000281, -0.00000245, -0.00000391, -0.00000645, -0.00000823, -0.00000745, -0.00000377, 0.00000109, 0.00000404, 0.00000238, -0.00000431, -0.00001340, -0.00002039, -0.00002128, -0.00001502, -0.00000443, 0.00000513, 0.00000876, 0.00000481, -0.00000407, -0.00001234, -0.00001489, -0.00000993, -0.00000010, 0.00000921, 0.00001310, 0.00001002, 0.00000244, -0.00000494, -0.00000832, -0.00000711, -0.00000398, -0.00000263, -0.00000498, -0.00000974, -0.00001347, -0.00001360, -0.00001106, -0.00001055, -0.00001777, -0.00003501, -0.00005808, -0.00007684, -0.00007965, -0.00005983, -0.00002015, 0.00002724, 0.00006592, 0.00008292, 0.00007483, 0.00004907, 0.00001967, -0.00000007, -0.00000404, 0.00000470, 0.00001688, 0.00002309, 0.00001915, 0.00000786, -0.00000358, -0.00000849, -0.00000488, 0.00000368, 0.00001065, 0.00001079, 0.00000337, -0.00000758, -0.00001596, -0.00001736, -0.00001161, -0.00000260, 0.00000434, 0.00000564, 0.00000141, -0.00000500, -0.00000935, -0.00000915, -0.00000499, 0.00000013, 0.00000280, 0.00000137, -0.00000326, -0.00000839, -0.00001127, -0.00001068, -0.00000748, -0.00000378, -0.00000159, -0.00000168, -0.00000334, -0.00000511, -0.00000577, -0.00000507, -0.00000365, -0.00000250, -0.00000220, -0.00000267, -0.00000331, -0.00000355, -0.00000325, -0.00000277, -0.00000259, -0.00000297, -0.00000373, -0.00000441, -0.00000460, -0.00000424, -0.00000362, -0.00000309, },
		{ -0.00000049, -0.00000018, -0.00000031, -0.00000113, -0.00000243, -0.00000362, -0.00000414, -0.00000390, -0.00000335, -0.00000324, -0.00000406, -0.00000561, -0.00000711, -0.00000767, -0.00000704, -0.00000583, -0.00000518, -0.00000595, -0.00000795, -0.00000987, -0.00001005, -0.00000768, -0.00000367, -0.00000048, -0.00000075, -0.00000562, -0.00001361, -0.00002107, -0.00002398, -0.00002032, -0.00001142, -0.00000145, 0.00000479, 0.00000460, -0.00000098, -0.00000775, -0.00001079, -0.00000739, 0.00000132, 0.00001108, 0.00001703, 0.00001658, 0.00001088, 0.00000386, -0.00000026, 0.00000045, 0.00000466, 0.00000905, 0.00001074, 0.00000925, 0.00000664, 0.00000574, 0.00000778, 0.00001138, 0.00001355, 0.00001232, 0.00000900, 0.00000818, 0.00001498, 0.00003105, 0.00005183, 0.00006740, 0.00006694, 0.00004484, 0.00000479, -0.00004072, -0.00007571, -0.00008822, -0.00007610, -0.00004784, -0.00001795, 0.00000053, 0.00000224, -0.00000874, -0.00002226, -0.00002856, -0.00002365, -0.00001100, 0.00000132, 0.00000597, 0.00000071, -0.00001056, -0.00002056, -0.00002318, -0.00001703, -0.00000585, 0.00000405, 0.00000775, 0.00000444, -0.00000250, -0.00000795, -0.00000833, -0.00000356, 0.00000316, 0.00000768, 0.00000743, 0.00000287, -0.00000304, -0.00000687, -0.00000671, -0.00000315, 0.00000143, 0.00000444, 0.00000466, 0.00000271, 0.00000034, -0.00000080, -0.00000017, 0.00000150, 0.00000292, 0.00000317, 0.00000227, 0.00000098, 0.00000023, 0.00000039, 0.00000109, 0.00000154, 0.00000115, -0.00000008, -0.00000155, -0.00000254, -0.00000264, -0.00000202, -0.00000120, -0.00000071, -0.00000067, -0.00000081, -0.00000079, },
		{ -0.00000454, -0.00000539, -0.00000602, -0.00000607, -0.00000565, -0.00000530, -0.00000557, -0.00000659, -0.00000789, -0.00000868, -0.00000845, -0.00000742, -0.00000642, -0.00000639, -0.00000757, -0.00000919, -0.00000984, -0.00000848, -0.00000535, -0.00000219, -0.00000140, -0.00000461, -0.00001131, -0.00001881, -0.00002340, -0.00002239, -0.00001578, -0.00000647, 0.00000126, 0.00000412, 0.00000171, -0.00000322, -0.00000643, -0.00000477, 0.00000188, 0.00001049, 0.00001666, 0.00001729, 0.00001236, 0.00000488, -0.00000103, -0.00000263, -0.00000011, 0.00000386, 0.00000600, 0.00000466, 0.00000070, -0.00000337, -0.00000534, -0.00000490, -0.00000372, -0.00000406, -0.00000679, -0.00001067, -0.00001334, -0.00001359, -0.00001313, -0.00001625, -0.00002705, -0.00004573, -0.00006652, -0.00007900, -0.00007287, -0.00004390, 0.00000247, 0.00005221, 0.00008855, 0.00009972, 0.00008456, 0.00005295, 0.00002063, 0.00000133, 0.00000030, 0.00001269, 0.00002738, 0.00003370, 0.00002732, 0.00001195, -0.00000360, -0.00001109, -0.00000751, 0.00000357, 0.00001473, 0.00001936, 0.00001543, 0.00000620, -0.00000231, -0.00000516, -0.00000140, 0.00000575, 0.00001126, 0.00001148, 0.00000622, -0.00000142, -0.00000724, -0.00000847, -0.00000528, -0.00000028, 0.00000330, 0.00000361, 0.00000105, -0.00000233, -0.00000435, -0.00000408, -0.00000223, -0.00000044, -0.00000008, -0.00000135, -0.00000331, -0.00000466, -0.00000464, -0.00000358, -0.00000251, -0.00000247, -0.00000376, -0.00000574, -0.00000736, -0.00000779, -0.00000696, -0.00000552, -0.00000434, -0.00000393, -0.00000422, -0.00000466, -0.00000476, -0.00000443, -0.00000402, -0.00000399, },
		{ -0.00000490, -0.00000459, -0.00000456, -0.00000519, -0.00000633, -0.00000739, -0.00000772, -0.00000714, -0.00000615, -0.00000563, -0.00000612, -0.00000738, -0.00000836, -0.00000794, -0.00000584, -0.00000307, -0.00000167, -0.00000346, -0.00000878, -0.00001591, -0.00002167, -0.00002304, -0.00001890, -0.00001080, -0.00000219, 0.00000346, 0.00000463, 0.00000258, 0.00000053, 0.00000161, 0.00000686, 0.00001446, 0.00002082, 0.00002275, 0.00001940, 0.00001282, 0.00000671, 0.00000421, 0.00000606, 0.00001037, 0.00001398, 0.00001453, 0.00001188, 0.00000800, 0.00000550, 0.00000581, 0.00000836, 0.00001119, 0.00001245, 0.00001174, 0.00001015, 0.00000915, 0.00000917, 0.00000923, 0.00000799, 0.00000569, 0.00000530, 0.00001157, 0.00002797, 0.00005316, 0.00007939, 0.00009462, 0.00008783, 0.00005525, 0.00000370, -0.00005102, -0.00009065, -0.00010273, -0.00008639, -0.00005254, -0.00001827, 0.00000163, 0.00000142, -0.00001379, -0.00003201, -0.00004158, -0.00003733, -0.00002258, -0.00000625, 0.00000289, 0.00000125, -0.00000811, -0.00001802, -0.00002189, -0.00001756, -0.00000815, 0.00000037, 0.00000300, -0.00000130, -0.00000940, -0.00001625, -0.00001802, -0.00001424, -0.00000769, -0.00000232, -0.00000079, -0.00000305, -0.00000679, -0.00000920, -0.00000877, -0.00000611, -0.00000322, -0.00000203, -0.00000309, -0.00000539, -0.00000716, -0.00000717, -0.00000551, -0.00000346, -0.00000255, -0.00000351, -0.00000580, -0.00000805, -0.00000894, -0.00000802, -0.00000594, -0.00000393, -0.00000297, -0.00000327, -0.00000421, -0.00000491, -0.00000485, -0.00000415, -0.00000342, -0.00000323, -0.00000370, -0.00000445, -0.00000493, },
		{ -0.00000590, -0.00000719, -0.00000811, -0.00000817, -0.00000748, -0.00000672, -0.00000665, -0.00000752, -0.00000871, -0.00000921, -0.00000829, -0.00000635, -0.00000486, -0.00000569, -0.00000983, -0.00001652, -0.00002328, -0.00002709, -0.00002593, -0.00002001, -0.00001173, -0.00000432, -0.00000013, 0.00000074, 0.00000022, 0.00000099, 0.00000458, 0.00001028, 0.00001549, 0.00001732, 0.00001445, 0.00000804, 0.00000116, -0.00000303, -0.00000312, -0.00000020, 0.00000294, 0.00000361, 0.00000087, -0.00000398, -0.00000842, -0.00001035, -0.00000944, -0.00000720, -0.00000581, -0.00000675, -0.00000997, -0.00001427, -0.00001833, -0.00002162, -0.00002434, -0.00002656, -0.00002746, -0.00002550, -0.00001985, -0.00001219, -0.00000734, -0.00001152, -0.00002877, -0.00005717, -0.00008751, -0.00010600, -0.00010039, -0.00006650, -0.00001184, 0.00004645, 0.00008860, 0.00010111, 0.00008286, 0.00004534, 0.00000673, -0.00001698, -0.00001921, -0.00000475, 0.00001419, 0.00002538, 0.00002302, 0.00000999, -0.00000499, -0.00001315, -0.00001079, -0.00000088, 0.00000941, 0.00001341, 0.00000884, -0.00000124, -0.00001080, -0.00001470, -0.00001162, -0.00000441, 0.00000214, 0.00000433, 0.00000166, -0.00000334, -0.00000710, -0.00000739, -0.00000454, -0.00000092, 0.00000085, -0.00000033, -0.00000346, -0.00000630, -0.00000698, -0.00000521, -0.00000242, -0.00000070, -0.00000136, -0.00000412, -0.00000734, -0.00000914, -0.00000856, -0.00000611, -0.00000332, -0.00000175, -0.00000201, -0.00000353, -0.00000507, -0.00000559, -0.00000489, -0.00000360, -0.00000268, -0.00000268, -0.00000348, -0.00000443, -0.00000493, -0.00000487, -0.00000468, -0.00000495, },
		{ -0.00001355, -0.00001281, -0.00001231, -0.00001271, -0.00001392, -0.00001514, -0.00001539, -0.00001443, -0.00001311, -0.00001313, -0.00001596, -0.00002175, -0.00002891, -0.00003460, -0.00003621, -0.00003268, -0.00002514, -0.00001626, -0.00000879, -0.00000410, -0.00000165, 0.00000040, 0.00000364, 0.00000826, 0.00001271, 0.00001475, 0.00001294, 0.00000776, 0.00000156, -0.00000275, -0.00000338, -0.00000071, 0.00000300, 0.00000507, 0.00000399, 0.00000038, -0.00000355, -0.00000544, -0.00000425, -0.00000083, 0.00000266, 0.00000419, 0.00000299, -0.00000020, -0.00000375, -0.00000621, -0.00000688, -0.00000569, -0.00000267, 0.00000234, 0.00000923, 0.00001687, 0.00002275, 0.00002387, 0.00001878, 0.00000958, 0.00000229, 0.00000456, 0.00002142, 0.00005109, 0.00008362, 0.00010400, 0.00009876, 0.00006315, 0.00000487, -0.00005800, -0.00010450, -0.00012007, -0.00010305, -0.00006501, -0.00002466, 0.00000129, 0.00000569, -0.00000695, -0.00002444, -0.00003440, -0.00003093, -0.00001688, -0.00000110, 0.00000750, 0.00000509, -0.00000538, -0.00001665, -0.00002183, -0.00001844, -0.00000926, -0.00000009, 0.00000408, 0.00000196, -0.00000378, -0.00000867, -0.00000946, -0.00000602, -0.00000106, 0.00000196, 0.00000119, -0.00000256, -0.00000655, -0.00000809, -0.00000633, -0.00000269, 0.00000015, 0.00000011, -0.00000297, -0.00000730, -0.00001035, -0.00001047, -0.00000787, -0.00000434, -0.00000204, -0.00000216, -0.00000434, -0.00000699, -0.00000850, -0.00000818, -0.00000660, -0.00000503, -0.00000458, -0.00000547, -0.00000704, -0.00000834, -0.00000883, -0.00000874, -0.00000879, -0.00000958, -0.00001111, -0.00001273, -0.00001365, },
		{ -0.00001097, -0.00001224, -0.00001318, -0.00001295, -0.00001170, -0.00001078, -0.00001196, -0.00001626, -0.00002303, -0.00002998, -0.00003409, -0.00003318, -0.00002699, -0.00001728, -0.00000684, 0.00000193, 0.00000812, 0.00001237, 0.00001594, 0.00001955, 0.00002271, 0.00002404, 0.00002236, 0.00001773, 0.00001180, 0.00000702, 0.00000527, 0.00000671, 0.00000968, 0.00001170, 0.00001097, 0.00000749, 0.00000297, -0.00000021, -0.00000055, 0.00000171, 0.00000474, 0.00000627, 0.00000490, 0.00000082, -0.00000445, -0.00000903, -0.00001157, -0.00001169, -0.00000972, -0.00000633, -0.00000227, 0.00000137, 0.00000310, 0.00000132, -0.00000463, -0.00001345, -0.00002167, -0.00002507, -0.00002128, -0.00001215, -0.00000420, -0.00000595, -0.00002321, -0.00005447, -0.00008942, -0.00011225, -0.00010865, -0.00007337, -0.00001419, 0.00005033, 0.00009851, 0.00011524, 0.00009872, 0.00006072, 0.00002037, -0.00000528, -0.00000897, 0.00000459, 0.00002286, 0.00003311, 0.00002926, 0.00001422, -0.00000299, -0.00001315, -0.00001220, -0.00000291, 0.00000763, 0.00001267, 0.00000979, 0.00000171, -0.00000600, -0.00000879, -0.00000582, -0.00000007, 0.00000395, 0.00000334, -0.00000150, -0.00000735, -0.00001053, -0.00000933, -0.00000505, -0.00000095, -0.00000008, -0.00000327, -0.00000872, -0.00001324, -0.00001432, -0.00001164, -0.00000712, -0.00000361, -0.00000308, -0.00000550, -0.00000908, -0.00001152, -0.00001151, -0.00000939, -0.00000673, -0.00000524, -0.00000564, -0.00000736, -0.00000912, -0.00000989, -0.00000954, -0.00000888, -0.00000891, -0.00001005, -0.00001184, -0.00001325, -0.00001346, -0.00001247, -0.00001111, -0.00001045, },
		{ -0.00001358, -0.00001250, -0.00001257, -0.00001560, -0.00002202, -0.00003028, -0.00003730, -0.00003989, -0.00003631, -0.00002707, -0.00001464, -0.00000212, 0.00000812, 0.00001525, 0.00001978, 0.00002261, 0.00002417, 0.00002410, 0.00002182, 0.00001732, 0.00001168, 0.00000677, 0.00000429, 0.00000474, 0.00000703, 0.00000906, 0.00000896, 0.00000625, 0.00000213, -0.00000126, -0.00000214, -0.00000022, 0.00000314, 0.00000570, 0.00000570, 0.00000278, -0.00000186, -0.00000623, -0.00000854, -0.00000799, -0.00000488, -0.00000019, 0.00000487, 0.00000916, 0.00001166, 0.00001161, 0.00000890, 0.00000455, 0.00000082, 0.00000038, 0.00000483, 0.00001324, 0.00002189, 0.00002593, 0.00002236, 0.00001277, 0.00000382, 0.00000447, 0.00002092, 0.00005171, 0.00008624, 0.00010822, 0.00010301, 0.00006542, 0.00000374, -0.00006267, -0.00011146, -0.00012728, -0.00010861, -0.00006797, -0.00002537, 0.00000150, 0.00000519, -0.00000935, -0.00002912, -0.00004091, -0.00003835, -0.00002423, -0.00000757, 0.00000249, 0.00000197, -0.00000632, -0.00001537, -0.00001879, -0.00001464, -0.00000614, 0.00000098, 0.00000237, -0.00000223, -0.00000919, -0.00001360, -0.00001258, -0.00000699, -0.00000069, 0.00000215, -0.00000031, -0.00000646, -0.00001243, -0.00001462, -0.00001193, -0.00000625, -0.00000113, 0.00000049, -0.00000194, -0.00000652, -0.00001024, -0.00001094, -0.00000855, -0.00000488, -0.00000228, -0.00000216, -0.00000419, -0.00000678, -0.00000828, -0.00000811, -0.00000701, -0.00000641, -0.00000730, -0.00000951, -0.00001187, -0.00001302, -0.00001241, -0.00001071, -0.00000929, -0.00000929, -0.00001081, -0.00001281, -0.00001394, },
		{ -0.00002082, -0.00002955, -0.00003896, -0.00004504, -0.00004452, -0.00003639, -0.00002231, -0.00000580, 0.00000941, 0.00002082, 0.00002774, 0.00003082, 0.00003118, 0.00002959, 0.00002635, 0.00002172, 0.00001642, 0.00001172, 0.00000893, 0.00000862, 0.00001014, 0.00001184, 0.00001200, 0.00000983, 0.00000603, 0.00000238, 0.00000069, 0.00000165, 0.00000439, 0.00000693, 0.00000733, 0.00000477, -0.00000001, -0.00000513, -0.00000862, -0.00000929, -0.00000722, -0.00000347, 0.00000051, 0.00000342, 0.00000437, 0.00000302, -0.00000042, -0.00000503, -0.00000928, -0.00001143, -0.00001048, -0.00000696, -0.00000322, -0.00000241, -0.00000672, -0.00001552, -0.00002513, -0.00003046, -0.00002833, -0.00002032, -0.00001324, -0.00001623, -0.00003536, -0.00006865, -0.00010469, -0.00012637, -0.00011868, -0.00007677, -0.00000992, 0.00006115, 0.00011288, 0.00012935, 0.00010925, 0.00006589, 0.00002029, -0.00000907, -0.00001444, -0.00000086, 0.00001838, 0.00002990, 0.00002735, 0.00001375, -0.00000180, -0.00001031, -0.00000831, 0.00000075, 0.00000945, 0.00001145, 0.00000535, -0.00000486, -0.00001280, -0.00001398, -0.00000851, -0.00000077, 0.00000365, 0.00000155, -0.00000594, -0.00001434, -0.00001874, -0.00001693, -0.00001053, -0.00000382, -0.00000090, -0.00000324, -0.00000898, -0.00001439, -0.00001626, -0.00001386, -0.00000910, -0.00000507, -0.00000405, -0.00000615, -0.00000955, -0.00001189, -0.00001192, -0.00001020, -0.00000856, -0.00000869, -0.00001093, -0.00001406, -0.00001617, -0.00001601, -0.00001384, -0.00001127, -0.00001016, -0.00001130, -0.00001393, -0.00001628, -0.00001686, -0.00001567, -0.00001443, -0.00001560, },
		{ -0.00005111, -0.00004561, -0.00003122, -0.00001144, 0.00000880, 0.00002514, 0.00003525, 0.00003912, 0.00003822, 0.00003444, 0.00002925, 0.00002354, 0.00001796, 0.00001322, 0.00001011, 0.00000902, 0.00000957, 0.00001055, 0.00001055, 0.00000867, 0.00000523, 0.00000157, -0.00000064, -0.00000046, 0.00000169, 0.00000423, 0.00000523, 0.00000356, -0.00000047, -0.00000532, -0.00000897, -0.00000998, -0.00000808, -0.00000420, 0.00000016, 0.00000353, 0.00000505, 0.00000451, 0.00000237, -0.00000045, -0.00000278, -0.00000344, -0.00000178, 0.00000183, 0.00000590, 0.00000836, 0.00000768, 0.00000405, -0.00000030, -0.00000209, 0.00000098, 0.00000835, 0.00001626, 0.00001966, 0.00001557, 0.00000613, -0.00000113, 0.00000342, 0.00002582, 0.00006335, 0.00010332, 0.00012725, 0.00011925, 0.00007448, 0.00000315, -0.00007258, -0.00012784, -0.00014594, -0.00012572, -0.00008124, -0.00003442, -0.00000426, 0.00000150, -0.00001161, -0.00002976, -0.00003944, -0.00003474, -0.00001943, -0.00000336, 0.00000427, 0.00000042, -0.00001054, -0.00002030, -0.00002210, -0.00001479, -0.00000320, 0.00000542, 0.00000602, -0.00000115, -0.00001103, -0.00001717, -0.00001580, -0.00000796, 0.00000142, 0.00000684, 0.00000557, -0.00000085, -0.00000794, -0.00001122, -0.00000897, -0.00000300, 0.00000281, 0.00000510, 0.00000306, -0.00000136, -0.00000502, -0.00000576, -0.00000373, -0.00000110, -0.00000040, -0.00000271, -0.00000696, -0.00001069, -0.00001173, -0.00000974, -0.00000636, -0.00000419, -0.00000496, -0.00000843, -0.00001256, -0.00001493, -0.00001452, -0.00001260, -0.00001215, -0.00001604, -0.00002506, -0.00003699, -0.00004729, },
		{ 0.00000046, 0.00002330, 0.00003851, 0.00004482, 0.00004374, 0.00003817, 0.00003094, 0.00002396, 0.00001816, 0.00001385, 0.00001116, 0.00001007, 0.00001024, 0.00001085, 0.00001085, 0.00000948, 0.00000676, 0.00000363, 0.00000146, 0.00000126, 0.00000299, 0.00000546, 0.00000695, 0.00000618, 0.00000302, -0.00000137, -0.00000512, -0.00000667, -0.00000545, -0.00000211, 0.00000191, 0.00000509, 0.00000644, 0.00000579, 0.00000364, 0.00000092, -0.00000136, -0.00000237, -0.00000175, 0.00000016, 0.00000234, 0.00000339, 0.00000223, -0.00000117, -0.00000548, -0.00000859, -0.00000889, -0.00000650, -0.00000369, -0.00000369, -0.00000857, -0.00001729, -0.00002552, -0.00002782, -0.00002134, -0.00000889, 0.00000091, -0.00000258, -0.00002583, -0.00006570, -0.00010829, -0.00013374, -0.00012517, -0.00007769, -0.00000245, 0.00007692, 0.00013424, 0.00015236, 0.00013062, 0.00008433, 0.00003666, 0.00000723, 0.00000344, 0.00001866, 0.00003784, 0.00004679, 0.00003983, 0.00002177, 0.00000366, -0.00000450, 0.00000023, 0.00001267, 0.00002333, 0.00002463, 0.00001539, 0.00000108, -0.00000998, -0.00001196, -0.00000488, 0.00000576, 0.00001277, 0.00001172, 0.00000345, -0.00000687, -0.00001325, -0.00001258, -0.00000631, 0.00000091, 0.00000435, 0.00000215, -0.00000382, -0.00000957, -0.00001175, -0.00000981, -0.00000615, -0.00000426, -0.00000629, -0.00001157, -0.00001715, -0.00001981, -0.00001824, -0.00001391, -0.00001010, -0.00000973, -0.00001342, -0.00001908, -0.00002339, -0.00002401, -0.00002141, -0.00001881, -0.00002036, -0.00002849, -0.00004203, -0.00005625, -0.00006481, -0.00006271, -0.00004866, -0.00002563, },
		{ 0.00005663, 0.00004924, 0.00003932, 0.00003019, 0.00002345, 0.00001924, 0.00001700, 0.00001610, 0.00001595, 0.00001600, 0.00001563, 0.00001433, 0.00001205, 0.00000937, 0.00000731, 0.00000676, 0.00000792, 0.00001001, 0.00001164, 0.00001150, 0.00000920, 0.00000546, 0.00000186, -0.00000004, 0.00000059, 0.00000339, 0.00000716, 0.00001039, 0.00001202, 0.00001179, 0.00001018, 0.00000810, 0.00000651, 0.00000604, 0.00000690, 0.00000873, 0.00001079, 0.00001211, 0.00001198, 0.00001036, 0.00000807, 0.00000651, 0.00000679, 0.00000900, 0.00001180, 0.00001314, 0.00001161, 0.00000777, 0.00000439, 0.00000508, 0.00001187, 0.00002318, 0.00003381, 0.00003749, 0.00003104, 0.00001764, 0.00000683, 0.00001020, 0.00003467, 0.00007665, 0.00012101, 0.00014642, 0.00013512, 0.00008244, 0.00000096, -0.00008363, -0.00014336, -0.00016046, -0.00013505, -0.00008437, -0.00003366, -0.00000386, -0.00000232, -0.00002113, -0.00004349, -0.00005388, -0.00004636, -0.00002662, -0.00000717, 0.00000081, -0.00000595, -0.00002160, -0.00003542, -0.00003875, -0.00002996, -0.00001496, -0.00000296, -0.00000064, -0.00000833, -0.00002029, -0.00002870, -0.00002859, -0.00002061, -0.00001015, -0.00000357, -0.00000419, -0.00001054, -0.00001775, -0.00002098, -0.00001858, -0.00001293, -0.00000858, -0.00000906, -0.00001450, -0.00002163, -0.00002599, -0.00002493, -0.00001939, -0.00001324, -0.00001076, -0.00001383, -0.00002071, -0.00002716, -0.00002933, -0.00002641, -0.00002149, -0.00001993, -0.00002612, -0.00004048, -0.00005851, -0.00007255, -0.00007519, -0.00006282, -0.00003732, -0.00000521, 0.00002516, 0.00004697, 0.00005712, },
		{ 0.00002818, 0.00002336, 0.00002131, 0.00002055, 0.00002001, 0.00001918, 0.00001784, 0.00001592, 0.00001346, 0.00001081, 0.00000858, 0.00000743, 0.00000763, 0.00000875, 0.00000975, 0.00000951, 0.00000744, 0.00000389, 0.00000007, -0.00000261, -0.00000319, -0.00000170, 0.00000091, 0.00000330, 0.00000436, 0.00000370, 0.00000169, -0.00000082, -0.00000293, -0.00000408, -0.00000419, -0.00000359, -0.00000293, -0.00000290, -0.00000395, -0.00000608, -0.00000875, -0.00001106, -0.00001229, -0.00001235, -0.00001206, -0.00001276, -0.00001541, -0.00001971, -0.00002386, -0.00002543, -0.00002295, -0.00001742, -0.00001232, -0.00001203, -0.00001896, -0.00003134, -0.00004333, -0.00004794, -0.00004170, -0.00002819, -0.00001789, -0.00002339, -0.00005193, -0.00009912, -0.00014809, -0.00017542, -0.00016194, -0.00010311, -0.00001323, 0.00007921, 0.00014355, 0.00016059, 0.00013095, 0.00007391, 0.00001741, -0.00001556, -0.00001712, 0.00000378, 0.00002838, 0.00003938, 0.00003016, 0.00000700, -0.00001610, -0.00002649, -0.00002020, -0.00000346, 0.00001195, 0.00001620, 0.00000714, -0.00000911, -0.00002256, -0.00002576, -0.00001807, -0.00000547, 0.00000372, 0.00000412, -0.00000363, -0.00001382, -0.00001981, -0.00001834, -0.00001143, -0.00000472, -0.00000353, -0.00000927, -0.00001863, -0.00002582, -0.00002645, -0.00002043, -0.00001202, -0.00000701, -0.00000904, -0.00001714, -0.00002641, -0.00003122, -0.00002896, -0.00002212, -0.00001708, -0.00002045, -0.00003483, -0.00005660, -0.00007703, -0.00008612, -0.00007734, -0.00005072, -0.00001290, 0.00002579, 0.00005565, 0.00007110, 0.00007206, 0.00006282, 0.00004941, 0.00003695, },
		{ 0.00001488, 0.00001330, 0.00001089, 0.00000811, 0.00000532, 0.00000277, 0.00000071, -0.00000055, -0.00000086, -0.00000040, 0.00000019, 0.00000006, -0.00000140, -0.00000414, -0.00000735, -0.00000988, -0.00001076, -0.00000978, -0.00000755, -0.00000518, -0.00000374, -0.00000373, -0.00000492, -0.00000658, -0.00000788, -0.00000830, -0.00000776, -0.00000662, -0.00000543, -0.00000473, -0.00000478, -0.00000551, -0.00000652, -0.00000721, -0.00000712, -0.00000623, -0.00000510, -0.00000457, -0.00000525, -0.00000690, -0.00000826, -0.00000761, -0.00000386, 0.00000238, 0.00000871, 0.00001199, 0.00001034, 0.00000477, -0.00000074, -0.00000125, 0.00000598, 0.00001895, 0.00003112, 0.00003481, 0.00002638, 0.00001012, -0.00000206, 0.00000398, 0.00003627, 0.00008979, 0.00014558, 0.00017748, 0.00016406, 0.00010013, 0.00000142, -0.00010059, -0.00017197, -0.00019133, -0.00015918, -0.00009665, -0.00003460, 0.00000146, 0.00000260, -0.00002148, -0.00005002, -0.00006368, -0.00005474, -0.00002979, -0.00000416, 0.00000798, 0.00000182, -0.00001620, -0.00003333, -0.00003868, -0.00002957, -0.00001236, 0.00000225, 0.00000617, -0.00000135, -0.00001394, -0.00002271, -0.00002216, -0.00001344, -0.00000321, 0.00000105, -0.00000397, -0.00001536, -0.00002597, -0.00002923, -0.00002339, -0.00001267, -0.00000451, -0.00000475, -0.00001384, -0.00002655, -0.00003529, -0.00003513, -0.00002723, -0.00001866, -0.00001843, -0.00003198, -0.00005733, -0.00008514, -0.00010271, -0.00009995, -0.00007418, -0.00003147, 0.00001607, 0.00005545, 0.00007790, 0.00008150, 0.00007065, 0.00005306, 0.00003608, 0.00002410, 0.00001795, 0.00001591, 0.00001551, },
		{ 0.00001212, 0.00001021, 0.00000905, 0.00000850, 0.00000846, 0.00000883, 0.00000932, 0.00000946, 0.00000877, 0.00000711, 0.00000487, 0.00000289, 0.00000198, 0.00000253, 0.00000425, 0.00000630, 0.00000776, 0.00000807, 0.00000727, 0.00000594, 0.00000481, 0.00000437, 0.00000470, 0.00000547, 0.00000616, 0.00000636, 0.00000587, 0.00000484, 0.00000363, 0.00000263, 0.00000211, 0.00000198, 0.00000183, 0.00000107, -0.00000065, -0.00000308, -0.00000529, -0.00000602, -0.00000451, -0.00000115, 0.00000230, 0.00000345, 0.00000077, -0.00000526, -0.00001190, -0.00001555, -0.00001398, -0.00000830, -0.00000301, -0.00000364, -0.00001310, -0.00002892, -0.00004344, -0.00004773, -0.00003751, -0.00001758, -0.00000147, -0.00000545, -0.00003904, -0.00009728, -0.00015952, -0.00019673, -0.00018479, -0.00011725, -0.00001079, 0.00010031, 0.00017859, 0.00020008, 0.00016471, 0.00009531, 0.00002570, -0.00001572, -0.00001861, 0.00000685, 0.00003802, 0.00005342, 0.00004408, 0.00001674, -0.00001198, -0.00002644, -0.00002101, -0.00000236, 0.00001579, 0.00002157, 0.00001200, -0.00000608, -0.00002115, -0.00002464, -0.00001616, -0.00000323, 0.00000403, -0.00000036, -0.00001451, -0.00003005, -0.00003774, -0.00003346, -0.00002083, -0.00000889, -0.00000621, -0.00001546, -0.00003158, -0.00004506, -0.00004800, -0.00003946, -0.00002664, -0.00002100, -0.00003150, -0.00005875, -0.00009349, -0.00012040, -0.00012520, -0.00010150, -0.00005388, 0.00000412, 0.00005594, 0.00008874, 0.00009787, 0.00008722, 0.00006610, 0.00004447, 0.00002904, 0.00002167, 0.00002034, 0.00002144, 0.00002199, 0.00002073, 0.00001800, 0.00001483, },
		{ 0.00001031, 0.00000983, 0.00000929, 0.00000863, 0.00000765, 0.00000617, 0.00000425, 0.00000231, 0.00000093, 0.00000054, 0.00000114, 0.00000223, 0.00000309, 0.00000319, 0.00000243, 0.00000119, 0.00000007, -0.00000047, -0.00000034, 0.00000018, 0.00000065, 0.00000069, 0.00000019, -0.00000066, -0.00000153, -0.00000208, -0.00000219, -0.00000202, -0.00000195, -0.00000240, -0.00000355, -0.00000510, -0.00000627, -0.00000611, -0.00000401, -0.00000030, 0.00000370, 0.00000615, 0.00000578, 0.00000279, -0.00000087, -0.00000244, -0.00000004, 0.00000585, 0.00001214, 0.00001474, 0.00001120, 0.00000298, -0.00000453, -0.00000471, 0.00000607, 0.00002507, 0.00004347, 0.00005067, 0.00004108, 0.00001930, 0.00000003, 0.00000148, 0.00003495, 0.00009600, 0.00016267, 0.00020334, 0.00019115, 0.00011819, 0.00000189, -0.00012069, -0.00020847, -0.00023460, -0.00019821, -0.00012367, -0.00004788, -0.00000226, 0.00000123, -0.00002682, -0.00006186, -0.00008018, -0.00007153, -0.00004289, -0.00001221, 0.00000340, -0.00000234, -0.00002223, -0.00004131, -0.00004677, -0.00003571, -0.00001624, -0.00000159, -0.00000138, -0.00001555, -0.00003474, -0.00004653, -0.00004348, -0.00002784, -0.00001003, -0.00000189, -0.00000917, -0.00002779, -0.00004639, -0.00005353, -0.00004520, -0.00002790, -0.00001543, -0.00002089, -0.00004856, -0.00009026, -0.00012827, -0.00014356, -0.00012486, -0.00007411, -0.00000565, 0.00006019, 0.00010565, 0.00012221, 0.00011244, 0.00008703, 0.00005893, 0.00003783, 0.00002737, 0.00002554, 0.00002753, 0.00002887, 0.00002742, 0.00002348, 0.00001875, 0.00001482, 0.00001241, 0.00001125, 0.00001073, },
		{ 0.00000915, 0.00000846, 0.00000764, 0.00000666, 0.00000578, 0.00000533, 0.00000552, 0.00000619, 0.00000691, 0.00000724, 0.00000699, 0.00000632, 0.00000567, 0.00000540, 0.00000563, 0.00000615, 0.00000657, 0.00000654, 0.00000600, 0.00000514, 0.00000430, 0.00000377, 0.00000360, 0.00000356, 0.00000326, 0.00000235, 0.00000073, -0.00000126, -0.00000294, -0.00000350, -0.00000249, -0.00000016, 0.00000239, 0.00000371, 0.00000268, -0.00000067, -0.00000494, -0.00000802, -0.00000830, -0.00000590, -0.00000299, -0.00000272, -0.00000723, -0.00001582, -0.00002468, -0.00002876, -0.00002489, -0.00001454, -0.00000398, -0.00000126, -0.00001128, -0.00003182, -0.00005344, -0.00006418, -0.00005705, -0.00003607, -0.00001662, -0.00001849, -0.00005452, -0.00012069, -0.00019382, -0.00023974, -0.00022883, -0.00015177, -0.00002676, 0.00010620, 0.00020217, 0.00023127, 0.00019180, 0.00010987, 0.00002570, -0.00002600, -0.00003157, -0.00000217, 0.00003549, 0.00005534, 0.00004595, 0.00001486, -0.00001820, -0.00003445, -0.00002727, -0.00000519, 0.00001455, 0.00001734, 0.00000021, -0.00002684, -0.00004766, -0.00005040, -0.00003492, -0.00001265, 0.00000071, -0.00000476, -0.00002699, -0.00005312, -0.00006763, -0.00006228, -0.00004184, -0.00002209, -0.00002089, -0.00004745, -0.00009589, -0.00014667, -0.00017554, -0.00016508, -0.00011317, -0.00003402, 0.00004824, 0.00011001, 0.00013758, 0.00013107, 0.00010209, 0.00006690, 0.00003902, 0.00002463, 0.00002229, 0.00002597, 0.00002936, 0.00002892, 0.00002466, 0.00001886, 0.00001406, 0.00001157, 0.00001115, 0.00001163, 0.00001193, 0.00001158, 0.00001077, 0.00000989, },
		{ 0.00001378, 0.00001237, 0.00001113, 0.00001024, 0.00000963, 0.00000903, 0.00000820, 0.00000715, 0.00000608, 0.00000525, 0.00000479, 0.00000456, 0.00000424, 0.00000355, 0.00000242, 0.00000106, -0.00000019, -0.00000106, -0.00000150, -0.00000176, -0.00000227, -0.00000336, -0.00000510, -0.00000714, -0.00000879, -0.00000935, -0.00000840, -0.00000617, -0.00000354, -0.00000176, -0.00000178, -0.00000369, -0.00000647, -0.00000844, -0.00000822, -0.00000567, -0.00000230, -0.00000052, -0.00000219, -0.00000707, -0.00001243, -0.00001429, -0.00000976, 0.00000060, 0.00001240, 0.00001937, 0.00001711, 0.00000638, -0.00000634, -0.00001197, -0.00000436, 0.00001507, 0.00003680, 0.00004799, 0.00004065, 0.00001843, -0.00000269, -0.00000148, 0.00003620, 0.00010646, 0.00018465, 0.00023376, 0.00022136, 0.00013678, -0.00000062, -0.00014740, -0.00025425, -0.00028796, -0.00024610, -0.00015694, -0.00006463, -0.00000762, -0.00000121, -0.00003328, -0.00007423, -0.00009514, -0.00008347, -0.00004815, -0.00001192, 0.00000362, -0.00000897, -0.00003932, -0.00006660, -0.00007294, -0.00005436, -0.00002284, 0.00000153, 0.00000330, -0.00001871, -0.00005070, -0.00007249, -0.00007027, -0.00004570, -0.00001592, -0.00000404, -0.00002560, -0.00007854, -0.00014238, -0.00018730, -0.00018836, -0.00013758, -0.00004779, 0.00005331, 0.00013539, 0.00017799, 0.00017736, 0.00014531, 0.00010160, 0.00006450, 0.00004382, 0.00003936, 0.00004407, 0.00004940, 0.00004986, 0.00004471, 0.00003674, 0.00002963, 0.00002566, 0.00002479, 0.00002541, 0.00002567, 0.00002463, 0.00002248, 0.00002008, 0.00001816, 0.00001692, 0.00001603, 0.00001505, },
		{ 0.00000112, 0.00000066, 0.00000041, 0.00000019, -0.00000001, -0.00000008, 0.00000006, 0.00000034, 0.00000054, 0.00000041, -0.00000010, -0.00000082, -0.00000142, -0.00000161, -0.00000135, -0.00000090, -0.00000070, -0.00000113, -0.00000226, -0.00000378, -0.00000502, -0.00000530, -0.00000422, -0.00000196, 0.00000071, 0.00000270, 0.00000314, 0.00000183, -0.00000049, -0.00000249, -0.00000295, -0.00000160, 0.00000052, 0.00000153, -0.00000018, -0.00000470, -0.00001012, -0.00001330, -0.00001168, -0.00000520, 0.00000302, 0.00000800, 0.00000569, -0.00000422, -0.00001742, -0.00002695, -0.00002731, -0.00001835, -0.00000630, -0.00000066, -0.00000836, -0.00002855, -0.00005158, -0.00006401, -0.00005718, -0.00003471, -0.00001354, -0.00001643, -0.00005868, -0.00013622, -0.00022226, -0.00027638, -0.00026292, -0.00016986, -0.00001824, 0.00014415, 0.00026270, 0.00030031, 0.00025396, 0.00015494, 0.00005254, -0.00001013, -0.00001591, 0.00002122, 0.00006725, 0.00008877, 0.00007132, 0.00002551, -0.00002149, -0.00004342, -0.00003073, 0.00000448, 0.00003669, 0.00004303, 0.00001728, -0.00002653, -0.00006298, -0.00007111, -0.00004777, -0.00001059, 0.00001157, -0.00000475, -0.00006326, -0.00014371, -0.00021058, -0.00022996, -0.00018592, -0.00008818, 0.00003246, 0.00013804, 0.00019955, 0.00020752, 0.00017285, 0.00011888, 0.00006955, 0.00003964, 0.00003120, 0.00003640, 0.00004415, 0.00004630, 0.00004077, 0.00003064, 0.00002093, 0.00001522, 0.00001403, 0.00001538, 0.00001652, 0.00001577, 0.00001312, 0.00000982, 0.00000725, 0.00000607, 0.00000597, 0.00000611, 0.00000575, 0.00000471, 0.00000330, 0.00000200, },
		{ 0.00000761, 0.00000797, 0.00000832, 0.00000867, 0.00000888, 0.00000880, 0.00000836, 0.00000771, 0.00000715, 0.00000697, 0.00000726, 0.00000779, 0.00000812, 0.00000783, 0.00000679, 0.00000527, 0.00000390, 0.00000340, 0.00000420, 0.00000620, 0.00000874, 0.00001083, 0.00001165, 0.00001092, 0.00000918, 0.00000751, 0.00000698, 0.00000800, 0.00000988, 0.00001119, 0.00001054, 0.00000764, 0.00000383, 0.00000158, 0.00000316, 0.00000903, 0.00001710, 0.00002343, 0.00002440, 0.00001902, 0.00001008, 0.00000296, 0.00000265, 0.00001041, 0.00002246, 0.00003172, 0.00003216, 0.00002308, 0.00001048, 0.00000412, 0.00001134, 0.00003140, 0.00005428, 0.00006581, 0.00005685, 0.00003145, 0.00000804, 0.00001136, 0.00005807, 0.00014350, 0.00023798, 0.00029662, 0.00027965, 0.00017354, 0.00000168, -0.00018231, -0.00031680, -0.00035960, -0.00030704, -0.00019467, -0.00007913, -0.00001043, -0.00000889, -0.00005843, -0.00011893, -0.00015019, -0.00013394, -0.00008217, -0.00002773, -0.00000369, -0.00002350, -0.00007357, -0.00012209, -0.00013869, -0.00011315, -0.00006242, -0.00002162, -0.00002405, -0.00008148, -0.00017546, -0.00026471, -0.00030461, -0.00026878, -0.00016163, -0.00001622, 0.00012099, 0.00021038, 0.00023442, 0.00020161, 0.00013869, 0.00007632, 0.00003586, 0.00002295, 0.00002961, 0.00004186, 0.00004812, 0.00004417, 0.00003304, 0.00002119, 0.00001397, 0.00001290, 0.00001580, 0.00001893, 0.00001954, 0.00001717, 0.00001335, 0.00001023, 0.00000912, 0.00000992, 0.00001146, 0.00001241, 0.00001208, 0.00001068, 0.00000892, 0.00000756, 0.00000692, 0.00000692, 0.00000724, },
		{ 0.00001437, 0.00001376, 0.00001284, 0.00001170, 0.00001060, 0.00000984, 0.00000955, 0.00000956, 0.00000946, 0.00000878, 0.00000729, 0.00000520, 0.00000307, 0.00000165, 0.00000142, 0.00000239, 0.00000400, 0.00000535, 0.00000560, 0.00000438, 0.00000203, -0.00000055, -0.00000240, -0.00000305, -0.00000289, -0.00000304, -0.00000468, -0.00000823, -0.00001283, -0.00001655, -0.00001743, -0.00001476, -0.00000991, -0.00000594, -0.00000608, -0.00001174, -0.00002131, -0.00003072, -0.00003559, -0.00003402, -0.00002810, -0.00002307, -0.00002421, -0.00003332, -0.00004698, -0.00005819, -0.00006072, -0.00005374, -0.00004340, -0.00003981, -0.00005057, -0.00007466, -0.00010120, -0.00011492, -0.00010615, -0.00007972, -0.00005627, -0.00006336, -0.00011942, -0.00021902, -0.00032850, -0.00039686, -0.00037864, -0.00025796, -0.00006116, 0.00015031, 0.00030492, 0.00035277, 0.00028827, 0.00015194, 0.00001026, -0.00007670, -0.00008305, -0.00002649, 0.00004507, 0.00008172, 0.00005888, -0.00001077, -0.00008687, -0.00012621, -0.00010861, -0.00004930, 0.00000922, 0.00002101, -0.00003684, -0.00014967, -0.00027034, -0.00034151, -0.00032320, -0.00021183, -0.00004163, 0.00013152, 0.00025480, 0.00029970, 0.00027024, 0.00019609, 0.00011575, 0.00005910, 0.00003719, 0.00004273, 0.00005869, 0.00006911, 0.00006663, 0.00005357, 0.00003790, 0.00002723, 0.00002464, 0.00002805, 0.00003264, 0.00003427, 0.00003165, 0.00002644, 0.00002160, 0.00001935, 0.00002003, 0.00002222, 0.00002397, 0.00002392, 0.00002200, 0.00001912, 0.00001646, 0.00001482, 0.00001429, 0.00001444, 0.00001477, 0.00001496, 0.00001493, 0.00001474, },
		{ -0.00000251, -0.00000393, -0.00000485, -0.00000534, -0.00000577, -0.00000664, -0.00000826, -0.00001056, -0.00001303, -0.00001494, -0.00001569, -0.00001514, -0.00001372, -0.00001228, -0.00001165, -0.00001232, -0.00001410, -0.00001626, -0.00001792, -0.00001852, -0.00001822, -0.00001784, -0.00001839, -0.00002037, -0.00002324, -0.00002554, -0.00002557, -0.00002248, -0.00001708, -0.00001168, -0.00000908, -0.00001089, -0.00001636, -0.00002249, -0.00002551, -0.00002317, -0.00001629, -0.00000858, -0.00000450, -0.00000638, -0.00001257, -0.00001810, -0.00001769, -0.00000946, 0.00000316, 0.00001306, 0.00001395, 0.00000521, -0.00000641, -0.00001007, 0.00000215, 0.00002871, 0.00005726, 0.00007077, 0.00005874, 0.00002702, -0.00000061, 0.00000746, 0.00007257, 0.00018870, 0.00031716, 0.00039863, 0.00037924, 0.00023878, 0.00000619, -0.00024746, -0.00043736, -0.00050222, -0.00043221, -0.00027313, -0.00010414, 0.00000035, 0.00000537, -0.00007067, -0.00016968, -0.00022732, -0.00020841, -0.00012545, -0.00003090, 0.00001235, -0.00003388, -0.00016065, -0.00031423, -0.00042126, -0.00042389, -0.00030701, -0.00010462, 0.00011665, 0.00028762, 0.00036506, 0.00034584, 0.00026186, 0.00016081, 0.00008355, 0.00004896, 0.00005185, 0.00007208, 0.00008830, 0.00008867, 0.00007399, 0.00005361, 0.00003797, 0.00003243, 0.00003552, 0.00004136, 0.00004404, 0.00004099, 0.00003375, 0.00002613, 0.00002154, 0.00002103, 0.00002315, 0.00002524, 0.00002515, 0.00002231, 0.00001773, 0.00001309, 0.00000972, 0.00000800, 0.00000747, 0.00000734, 0.00000700, 0.00000625, 0.00000516, 0.00000391, 0.00000251, 0.00000094, -0.00000079, },
		{ 0.00001077, 0.00001110, 0.00001072, 0.00000955, 0.00000799, 0.00000674, 0.00000650, 0.00000756, 0.00000964, 0.00001197, 0.00001368, 0.00001416, 0.00001341, 0.00001200, 0.00001075, 0.00001027, 0.00001060, 0.00001116, 0.00001109, 0.00000984, 0.00000765, 0.00000560, 0.00000510, 0.00000704, 0.00001104, 0.00001536, 0.00001765, 0.00001619, 0.00001111, 0.00000449, -0.00000055, -0.00000168, 0.00000115, 0.00000548, 0.00000763, 0.00000506, -0.00000188, -0.00000985, -0.00001459, -0.00001378, -0.00000904, -0.00000533, -0.00000789, -0.00001837, -0.00003292, -0.00004394, -0.00004497, -0.00003583, -0.00002438, -0.00002277, -0.00003960, -0.00007267, -0.00010750, -0.00012436, -0.00011103, -0.00007426, -0.00004173, -0.00005080, -0.00012746, -0.00026617, -0.00042272, -0.00052732, -0.00051457, -0.00035721, -0.00008605, 0.00021612, 0.00044665, 0.00052755, 0.00044118, 0.00023875, 0.00001656, -0.00012953, -0.00014912, -0.00005954, 0.00006553, 0.00013534, 0.00008919, -0.00007198, -0.00028519, -0.00045545, -0.00050092, -0.00039070, -0.00015766, 0.00011828, 0.00034669, 0.00046446, 0.00045767, 0.00035956, 0.00022866, 0.00012003, 0.00006356, 0.00005802, 0.00008032, 0.00010243, 0.00010629, 0.00008986, 0.00006358, 0.00004120, 0.00003128, 0.00003359, 0.00004130, 0.00004633, 0.00004426, 0.00003604, 0.00002630, 0.00001987, 0.00001893, 0.00002217, 0.00002625, 0.00002797, 0.00002607, 0.00002146, 0.00001634, 0.00001266, 0.00001126, 0.00001169, 0.00001284, 0.00001373, 0.00001394, 0.00001361, 0.00001306, 0.00001249, 0.00001190, 0.00001121, 0.00001048, 0.00000993, 0.00000981, 0.00001018, },
		{ 0.00001203, 0.00000970, 0.00000817, 0.00000795, 0.00000893, 0.00001046, 0.00001162, 0.00001169, 0.00001048, 0.00000838, 0.00000617, 0.00000453, 0.00000369, 0.00000330, 0.00000267, 0.00000120, -0.00000106, -0.00000337, -0.00000456, -0.00000373, -0.00000093, 0.00000260, 0.00000493, 0.00000441, 0.00000072, -0.00000476, -0.00000952, -0.00001129, -0.00000939, -0.00000529, -0.00000183, -0.00000148, -0.00000469, -0.00000944, -0.00001238, -0.00001107, -0.00000587, 0.00000001, 0.00000242, -0.00000068, -0.00000732, -0.00001219, -0.00000993, 0.00000076, 0.00001517, 0.00002468, 0.00002228, 0.00000826, -0.00000788, -0.00001184, 0.00000647, 0.00004452, 0.00008528, 0.00010521, 0.00008911, 0.00004380, 0.00000114, 0.00000546, 0.00008915, 0.00024811, 0.00043176, 0.00055669, 0.00054145, 0.00034765, 0.00000674, -0.00038345, -0.00069477, -0.00082301, -0.00073544, -0.00048822, -0.00020349, -0.00001552, -0.00001037, -0.00018735, -0.00046025, -0.00069739, -0.00078110, -0.00065876, -0.00036267, 0.00000840, 0.00033486, 0.00052641, 0.00055400, 0.00045177, 0.00029217, 0.00014926, 0.00006853, 0.00005553, 0.00008456, 0.00011941, 0.00013356, 0.00012041, 0.00009091, 0.00006276, 0.00004885, 0.00005120, 0.00006233, 0.00007165, 0.00007219, 0.00006379, 0.00005176, 0.00004267, 0.00004028, 0.00004383, 0.00004935, 0.00005243, 0.00005080, 0.00004513, 0.00003808, 0.00003245, 0.00002968, 0.00002944, 0.00003032, 0.00003085, 0.00003035, 0.00002895, 0.00002721, 0.00002558, 0.00002414, 0.00002272, 0.00002120, 0.00001968, 0.00001846, 0.00001774, 0.00001743, 0.00001707, 0.00001616, 0.00001440, },
		{ 0.00001963, 0.00002195, 0.00002416, 0.00002539, 0.00002522, 0.00002388, 0.00002206, 0.00002052, 0.00001968, 0.00001940, 0.00001911, 0.00001823, 0.00001661, 0.00001470, 0.00001343, 0.00001364, 0.00001550, 0.00001819, 0.00002016, 0.00001986, 0.00001668, 0.00001138, 0.00000590, 0.00000231, 0.00000163, 0.00000317, 0.00000483, 0.00000436, 0.00000076, -0.00000497, -0.00001039, -0.00001326, -0.00001318, -0.00001208, -0.00001310, -0.00001839, -0.00002728, -0.00003626, -0.00004116, -0.00004023, -0.00003618, -0.00003519, -0.00004310, -0.00006094, -0.00008277, -0.00009824, -0.00009900, -0.00008544, -0.00006885, -0.00006652, -0.00009119, -0.00014089, -0.00019632, -0.00022965, -0.00022190, -0.00017964, -0.00013969, -0.00015546, -0.00026820, -0.00047644, -0.00072141, -0.00090043, -0.00090743, -0.00068458, -0.00026040, 0.00024712, 0.00066874, 0.00085035, 0.00071916, 0.00031912, -0.00020192, -0.00064939, -0.00085910, -0.00076052, -0.00039954, 0.00008741, 0.00053366, 0.00080898, 0.00086363, 0.00073395, 0.00051228, 0.00029951, 0.00016334, 0.00011964, 0.00014028, 0.00017840, 0.00019570, 0.00017836, 0.00013677, 0.00009312, 0.00006600, 0.00006092, 0.00007024, 0.00008041, 0.00008099, 0.00006986, 0.00005269, 0.00003801, 0.00003167, 0.00003393, 0.00004028, 0.00004487, 0.00004400, 0.00003777, 0.00002925, 0.00002217, 0.00001878, 0.00001903, 0.00002112, 0.00002303, 0.00002358, 0.00002285, 0.00002165, 0.00002078, 0.00002045, 0.00002040, 0.00002025, 0.00001990, 0.00001961, 0.00001974, 0.00002043, 0.00002135, 0.00002190, 0.00002161, 0.00002045, 0.00001895, 0.00001795, 0.00001813, },
		{ 0.00005216, 0.00004754, 0.00004217, 0.00003687, 0.00003227, 0.00002845, 0.00002501, 0.00002138, 0.00001724, 0.00001281, 0.00000879, 0.00000597, 0.00000466, 0.00000441, 0.00000402, 0.00000213, -0.00000206, -0.00000822, -0.00001492, -0.00002036, -0.00002329, -0.00002381, -0.00002332, -0.00002374, -0.00002629, -0.00003069, -0.00003532, -0.00003824, -0.00003857, -0.00003724, -0.00003645, -0.00003813, -0.00004234, -0.00004691, -0.00004864, -0.00004568, -0.00003930, -0.00003374, -0.00003364, -0.00004054, -0.00005078, -0.00005669, -0.00005118, -0.00003304, -0.00000942, 0.00000704, 0.00000572, -0.00001386, -0.00003903, -0.00004939, -0.00002909, 0.00002089, 0.00007871, 0.00011138, 0.00009513, 0.00003624, -0.00002143, -0.00001115, 0.00012225, 0.00038289, 0.00070085, 0.00094242, 0.00095634, 0.00063988, -0.00000388, -0.00084634, -0.00166524, -0.00221944, -0.00233733, -0.00198184, -0.00126421, -0.00040072, 0.00036857, 0.00086450, 0.00102247, 0.00089664, 0.00061856, 0.00033339, 0.00014322, 0.00007972, 0.00011190, 0.00017823, 0.00022364, 0.00022343, 0.00018615, 0.00013931, 0.00010927, 0.00010691, 0.00012528, 0.00014760, 0.00015888, 0.00015405, 0.00013866, 0.00012350, 0.00011728, 0.00012191, 0.00013271, 0.00014219, 0.00014486, 0.00013985, 0.00013054, 0.00012181, 0.00011707, 0.00011677, 0.00011886, 0.00012057, 0.00012015, 0.00011751, 0.00011379, 0.00011028, 0.00010752, 0.00010521, 0.00010268, 0.00009954, 0.00009596, 0.00009250, 0.00008957, 0.00008712, 0.00008457, 0.00008124, 0.00007680, 0.00007162, 0.00006656, 0.00006250, 0.00005990, 0.00005845, 0.00005730, 0.00005542, },
		{ -0.00002177, -0.00002403, -0.00002549, -0.00002671, -0.00002818, -0.00002992, -0.00003148, -0.00003219, -0.00003162, -0.00002998, -0.00002816, -0.00002736, -0.00002847, -0.00003153, -0.00003561, -0.00003920, -0.00004104, -0.00004080, -0.00003935, -0.00003823, -0.00003878, -0.00004127, -0.00004479, -0.00004783, -0.00004941, -0.00004979, -0.00005040, -0.00005289, -0.00005779, -0.00006395, -0.00006916, -0.00007179, -0.00007233, -0.00007365, -0.00007941, -0.00009143, -0.00010771, -0.00012283, -0.00013106, -0.00013056, -0.00012589, -0.00012659, -0.00014189, -0.00017421, -0.00021589, -0.00025227, -0.00027041, -0.00026887, -0.00026192, -0.00027384, -0.00032460, -0.00041470, -0.00051927, -0.00059851, -0.00062238, -0.00059776, -0.00058123, -0.00066482, -0.00093467, -0.00141792, -0.00204349, -0.00264049, -0.00298357, -0.00287243, -0.00221485, -0.00107694, 0.00032395, 0.00168478, 0.00271613, 0.00323588, 0.00322043, 0.00279544, 0.00217416, 0.00157259, 0.00113734, 0.00091283, 0.00085394, 0.00087054, 0.00087851, 0.00083417, 0.00074096, 0.00063217, 0.00054409, 0.00049502, 0.00047938, 0.00047627, 0.00046441, 0.00043398, 0.00038949, 0.00034401, 0.00030976, 0.00029111, 0.00028348, 0.00027751, 0.00026522, 0.00024419, 0.00021778, 0.00019208, 0.00017198, 0.00015870, 0.00014994, 0.00014203, 0.00013228, 0.00012022, 0.00010726, 0.00009527, 0.00008526, 0.00007695, 0.00006928, 0.00006138, 0.00005315, 0.00004522, 0.00003836, 0.00003284, 0.00002815, 0.00002340, 0.00001789, 0.00001163, 0.00000540, 0.00000026, -0.00000303, -0.00000451, -0.00000498, -0.00000561, -0.00000734, -0.00001042, -0.00001442, -0.00001847, },
		{ -0.00009587, -0.00009621, -0.00009697, -0.00009756, -0.00009749, -0.00009671, -0.00009580, -0.00009575, -0.00009747, -0.00010123, -0.00010647, -0.00011196, -0.00011644, -0.00011931, -0.00012091, -0.00012239, -0.00012496, -0.00012923, -0.00013480, -0.00014057, -0.00014552, -0.00014942, -0.00015304, -0.00015755, -0.00016360, -0.00017068, -0.00017730, -0.00018205, -0.00018488, -0.00018754, -0.00019269, -0.00020197, -0.00021431, -0.00022580, -0.00023175, -0.00022987, -0.00022268, -0.00021702, -0.00022043, -0.00023596, -0.00025897, -0.00027845, -0.00028317, -0.00026945, -0.00024552, -0.00022847, -0.00023417, -0.00026535, -0.00030573, -0.00032600, -0.00030079, -0.00022825, -0.00013925, -0.00008674, -0.00011525, -0.00022320, -0.00033847, -0.00032572, -0.00003080, 0.00065093, 0.00171070, 0.00299827, 0.00424869, 0.00515994, 0.00549379, 0.00515994, 0.00424869, 0.00299827, 0.00171070, 0.00065093, -0.00003080, -0.00032572, -0.00033847, -0.00022320, -0.00011525, -0.00008674, -0.00013925, -0.00022825, -0.00030079, -0.00032600, -0.00030573, -0.00026535, -0.00023417, -0.00022847, -0.00024552, -0.00026945, -0.00028317, -0.00027845, -0.00025897, -0.00023596, -0.00022043, -0.00021702, -0.00022268, -0.00022987, -0.00023175, -0.00022580, -0.00021431, -0.00020197, -0.00019269, -0.00018754, -0.00018488, -0.00018205, -0.00017730, -0.00017068, -0.00016360, -0.00015755, -0.00015304, -0.00014942, -0.00014552, -0.00014057, -0.00013480, -0.00012923, -0.00012496, -0.00012239, -0.00012091, -0.00011931, -0.00011644, -0.00011196, -0.00010647, -0.00010123, -0.00009747, -0.00009575, -0.00009580, -0.00009671, -0.00009749, -0.00009756, -0.00009697, -0.00009621, },
		{ -0.00002177, -0.00001847, -0.00001442, -0.00001042, -0.00000734, -0.00000561, -0.00000498, -0.00000451, -0.00000303, 0.00000026, 0.00000540, 0.00001163, 0.00001789, 0.00002340, 0.00002815, 0.00003284, 0.00003836, 0.00004522, 0.00005315, 0.00006138, 0.00006928, 0.00007695, 0.00008526, 0.00009527, 0.00010726, 0.00012022, 0.00013228, 0.00014203, 0.00014994, 0.00015870, 0.00017198, 0.00019208, 0.00021778, 0.00024419, 0.00026522, 0.00027751, 0.00028348, 0.00029111, 0.00030976, 0.00034401, 0.00038949, 0.00043398, 0.00046441, 0.00047627, 0.00047938, 0.00049502, 0.00054409, 0.00063217, 0.00074096, 0.00083417, 0.00087851, 0.00087054, 0.00085394, 0.00091283, 0.00113734, 0.00157259, 0.00217416, 0.00279544, 0.00322043, 0.00323588, 0.00271613, 0.00168478, 0.00032395, -0.00107694, -0.00221485, -0.00287243, -0.00298357, -0.00264049, -0.00204349, -0.00141792, -0.00093467, -0.00066482, -0.00058123, -0.00059776, -0.00062238, -0.00059851, -0.00051927, -0.00041470, -0.00032460, -0.00027384, -0.00026192, -0.00026887, -0.00027041, -0.00025227, -0.00021589, -0.00017421, -0.00014189, -0.00012659, -0.00012589, -0.00013056, -0.00013106, -0.00012283, -0.00010771, -0.00009143, -0.00007941, -0.00007365, -0.00007233, -0.00007179, -0.00006916, -0.00006395, -0.00005779, -0.00005289, -0.00005040, -0.00004979, -0.00004941, -0.00004783, -0.00004479, -0.00004127, -0.00003878, -0.00003823, -0.00003935, -0.00004080, -0.00004104, -0.00003920, -0.00003561, -0.00003153, -0.00002847, -0.00002736, -0.00002816, -0.00002998, -0.00003162, -0.00003219, -0.00003148, -0.00002992, -0.00002818, -0.00002671, -0.00002549, -0.00002403, },
		{ 0.00005216, 0.00005542, 0.00005730, 0.00005845, 0.00005990, 0.00006250, 0.00006656, 0.00007162, 0.00007680, 0.00008124, 0.00008457, 0.00008712, 0.00008957, 0.00009250, 0.00009596, 0.00009954, 0.00010268, 0.00010521, 0.00010752, 0.00011028, 0.00011379, 0.00011751, 0.00012015, 0.00012057, 0.00011886, 0.00011677, 0.00011707, 0.00012181, 0.00013054, 0.00013985, 0.00014486, 0.00014219, 0.00013271, 0.00012191, 0.00011728, 0.00012350, 0.00013866, 0.00015405, 0.00015888, 0.00014760, 0.00012528, 0.00010691, 0.00010927, 0.00013931, 0.00018615, 0.00022343, 0.00022364, 0.00017823, 0.00011190, 0.00007972, 0.00014322, 0.00033339, 0.00061856, 0.00089664, 0.00102247, 0.00086450, 0.00036857, -0.00040072, -0.00126421, -0.00198184, -0.00233733, -0.00221944, -0.00166524, -0.00084634, -0.00000388, 0.00063988, 0.00095634, 0.00094242, 0.00070085, 0.00038289, 0.00012225, -0.00001115, -0.00002143, 0.00003624, 0.00009513, 0.00011138, 0.00007871, 0.00002089, -0.00002909, -0.00004939, -0.00003903, -0.00001386, 0.00000572, 0.00000704, -0.00000942, -0.00003304, -0.00005118, -0.00005669, -0.00005078, -0.00004054, -0.00003364, -0.00003374, -0.00003930, -0.00004568, -0.00004864, -0.00004691, -0.00004234, -0.00003813, -0.00003645, -0.00003724, -0.00003857, -0.00003824, -0.00003532, -0.00003069, -0.00002629, -0.00002374, -0.00002332, -0.00002381, -0.00002329, -0.00002036, -0.00001492, -0.00000822, -0.00000206, 0.00000213, 0.00000402, 0.00000441, 0.00000466, 0.00000597, 0.00000879, 0.00001281, 0.00001724, 0.00002138, 0.00002501, 0.00002845, 0.00003227, 0.00003687, 0.00004217, 0.00004754, },
		{ 0.00001963, 0.00001813, 0.00001795, 0.00001895, 0.00002045, 0.00002161, 0.00002190, 0.00002135, 0.00002043, 0.00001974, 0.00001961, 0.00001990, 0.00002025, 0.00002040, 0.00002045, 0.00002078, 0.00002165, 0.00002285, 0.00002358, 0.00002303, 0.00002112, 0.00001903, 0.00001878, 0.00002217, 0.00002925, 0.00003777, 0.00004400, 0.00004487, 0.00004028, 0.00003393, 0.00003167, 0.00003801, 0.00005269, 0.00006986, 0.00008099, 0.00008041, 0.00007024, 0.00006092, 0.00006600, 0.00009312, 0.00013677, 0.00017836, 0.00019570, 0.00017840, 0.00014028, 0.00011964, 0.00016334, 0.00029951, 0.00051228, 0.00073395, 0.00086363, 0.00080898, 0.00053366, 0.00008741, -0.00039954, -0.00076052, -0.00085910, -0.00064939, -0.00020192, 0.00031912, 0.00071916, 0.00085035, 0.00066874, 0.00024712, -0.00026040, -0.00068458, -0.00090743, -0.00090043, -0.00072141, -0.00047644, -0.00026820, -0.00015546, -0.00013969, -0.00017964, -0.00022190, -0.00022965, -0.00019632, -0.00014089, -0.00009119, -0.00006652, -0.00006885, -0.00008544, -0.00009900, -0.00009824, -0.00008277, -0.00006094, -0.00004310, -0.00003519, -0.00003618, -0.00004023, -0.00004116, -0.00003626, -0.00002728, -0.00001839, -0.00001310, -0.00001208, -0.00001318, -0.00001326, -0.00001039, -0.00000497, 0.00000076, 0.00000436, 0.00000483, 0.00000317, 0.00000163, 0.00000231, 0.00000590, 0.00001138, 0.00001668, 0.00001986, 0.00002016, 0.00001819, 0.00001550, 0.00001364, 0.00001343, 0.00001470, 0.00001661, 0.00001823, 0.00001911, 0.00001940, 0.00001968, 0.00002052, 0.00002206, 0.00002388, 0.00002522, 0.00002539, 0.00002416, 0.00002195, },
		{ 0.00001203, 0.00001440, 0.00001616, 0.00001707, 0.00001743, 0.00001774, 0.00001846, 0.00001968, 0.00002120, 0.00002272, 0.00002414, 0.00002558, 0.00002721, 0.00002895, 0.00003035, 0.00003085, 0.00003032, 0.00002944, 0.00002968, 0.00003245, 0.00003808, 0.00004513, 0.00005080, 0.00005243, 0.00004935, 0.00004383, 0.00004028, 0.00004267, 0.00005176, 0.00006379, 0.00007219, 0.00007165, 0.00006233, 0.00005120, 0.00004885, 0.00006276, 0.00009091, 0.00012041, 0.00013356, 0.00011941, 0.00008456, 0.00005553, 0.00006853, 0.00014926, 0.00029217, 0.00045177, 0.00055400, 0.00052641, 0.00033486, 0.00000840, -0.00036267, -0.00065876, -0.00078110, -0.00069739, -0.00046025, -0.00018735, -0.00001037, -0.00001552, -0.00020349, -0.00048822, -0.00073544, -0.00082301, -0.00069477, -0.00038345, 0.00000674, 0.00034765, 0.00054145, 0.00055669, 0.00043176, 0.00024811, 0.00008915, 0.00000546, 0.00000114, 0.00004380, 0.00008911, 0.00010521, 0.00008528, 0.00004452, 0.00000647, -0.00001184, -0.00000788, 0.00000826, 0.00002228, 0.00002468, 0.00001517, 0.00000076, -0.00000993, -0.00001219, -0.00000732, -0.00000068, 0.00000242, 0.00000001, -0.00000587, -0.00001107, -0.00001238, -0.00000944, -0.00000469, -0.00000148, -0.00000183, -0.00000529, -0.00000939, -0.00001129, -0.00000952, -0.00000476, 0.00000072, 0.00000441, 0.00000493, 0.00000260, -0.00000093, -0.00000373, -0.00000456, -0.00000337, -0.00000106, 0.00000120, 0.00000267, 0.00000330, 0.00000369, 0.00000453, 0.00000617, 0.00000838, 0.00001048, 0.00001169, 0.00001162, 0.00001046, 0.00000893, 0.00000795, 0.00000817, 0.00000970, },
		{ 0.00001077, 0.00001018, 0.00000981, 0.00000993, 0.00001048, 0.00001121, 0.00001190, 0.00001249, 0.00001306, 0.00001361, 0.00001394, 0.00001373, 0.00001284, 0.00001169, 0.00001126, 0.00001266, 0.00001634, 0.00002146, 0.00002607, 0.00002797, 0.00002625, 0.00002217, 0.00001893, 0.00001987, 0.00002630, 0.00003604, 0.00004426, 0.00004633, 0.00004130, 0.00003359, 0.00003128, 0.00004120, 0.00006358, 0.00008986, 0.00010629, 0.00010243, 0.00008032, 0.00005802, 0.00006356, 0.00012003, 0.00022866, 0.00035956, 0.00045767, 0.00046446, 0.00034669, 0.00011828, -0.00015766, -0.00039070, -0.00050092, -0.00045545, -0.00028519, -0.00007198, 0.00008919, 0.00013534, 0.00006553, -0.00005954, -0.00014912, -0.00012953, 0.00001656, 0.00023875, 0.00044118, 0.00052755, 0.00044665, 0.00021612, -0.00008605, -0.00035721, -0.00051457, -0.00052732, -0.00042272, -0.00026617, -0.00012746, -0.00005080, -0.00004173, -0.00007426, -0.00011103, -0.00012436, -0.00010750, -0.00007267, -0.00003960, -0.00002277, -0.00002438, -0.00003583, -0.00004497, -0.00004394, -0.00003292, -0.00001837, -0.00000789, -0.00000533, -0.00000904, -0.00001378, -0.00001459, -0.00000985, -0.00000188, 0.00000506, 0.00000763, 0.00000548, 0.00000115, -0.00000168, -0.00000055, 0.00000449, 0.00001111, 0.00001619, 0.00001765, 0.00001536, 0.00001104, 0.00000704, 0.00000510, 0.00000560, 0.00000765, 0.00000984, 0.00001109, 0.00001116, 0.00001060, 0.00001027, 0.00001075, 0.00001200, 0.00001341, 0.00001416, 0.00001368, 0.00001197, 0.00000964, 0.00000756, 0.00000650, 0.00000674, 0.00000799, 0.00000955, 0.00001072, 0.00001110, },
		{ -0.00000251, -0.00000079, 0.00000094, 0.00000251, 0.00000391, 0.00000516, 0.00000625, 0.00000700, 0.00000734, 0.00000747, 0.00000800, 0.00000972, 0.00001309, 0.00001773, 0.00002231, 0.00002515, 0.00002524, 0.00002315, 0.00002103, 0.00002154, 0.00002613, 0.00003375, 0.00004099, 0.00004404, 0.00004136, 0.00003552, 0.00003243, 0.00003797, 0.00005361, 0.00007399, 0.00008867, 0.00008830, 0.00007208, 0.00005185, 0.00004896, 0.00008355, 0.00016081, 0.00026186, 0.00034584, 0.00036506, 0.00028762, 0.00011665, -0.00010462, -0.00030701, -0.00042389, -0.00042126, -0.00031423, -0.00016065, -0.00003388, 0.00001235, -0.00003090, -0.00012545, -0.00020841, -0.00022732, -0.00016968, -0.00007067, 0.00000537, 0.00000035, -0.00010414, -0.00027313, -0.00043221, -0.00050222, -0.00043736, -0.00024746, 0.00000619, 0.00023878, 0.00037924, 0.00039863, 0.00031716, 0.00018870, 0.00007257, 0.00000746, -0.00000061, 0.00002702, 0.00005874, 0.00007077, 0.00005726, 0.00002871, 0.00000215, -0.00001007, -0.00000641, 0.00000521, 0.00001395, 0.00001306, 0.00000316, -0.00000946, -0.00001769, -0.00001810, -0.00001257, -0.00000638, -0.00000450, -0.00000858, -0.00001629, -0.00002317, -0.00002551, -0.00002249, -0.00001636, -0.00001089, -0.00000908, -0.00001168, -0.00001708, -0.00002248, -0.00002557, -0.00002554, -0.00002324, -0.00002037, -0.00001839, -0.00001784, -0.00001822, -0.00001852, -0.00001792, -0.00001626, -0.00001410, -0.00001232, -0.00001165, -0.00001228, -0.00001372, -0.00001514, -0.00001569, -0.00001494, -0.00001303, -0.00001056, -0.00000826, -0.00000664, -0.00000577, -0.00000534, -0.00000485, -0.00000393, },
		{ 0.00001437, 0.00001474, 0.00001493, 0.00001496, 0.00001477, 0.00001444, 0.00001429, 0.00001482, 0.00001646, 0.00001912, 0.00002200, 0.00002392, 0.00002397, 0.00002222, 0.00002003, 0.00001935, 0.00002160, 0.00002644, 0.00003165, 0.00003427, 0.00003264, 0.00002805, 0.00002464, 0.00002723, 0.00003790, 0.00005357, 0.00006663, 0.00006911, 0.00005869, 0.00004273, 0.00003719, 0.00005910, 0.00011575, 0.00019609, 0.00027024, 0.00029970, 0.00025480, 0.00013152, -0.00004163, -0.00021183, -0.00032320, -0.00034151, -0.00027034, -0.00014967, -0.00003684, 0.00002101, 0.00000922, -0.00004930, -0.00010861, -0.00012621, -0.00008687, -0.00001077, 0.00005888, 0.00008172, 0.00004507, -0.00002649, -0.00008305, -0.00007670, 0.00001026, 0.00015194, 0.00028827, 0.00035277, 0.00030492, 0.00015031, -0.00006116, -0.00025796, -0.00037864, -0.00039686, -0.00032850, -0.00021902, -0.00011942, -0.00006336, -0.00005627, -0.00007972, -0.00010615, -0.00011492, -0.00010120, -0.00007466, -0.00005057, -0.00003981, -0.00004340, -0.00005374, -0.00006072, -0.00005819, -0.00004698, -0.00003332, -0.00002421, -0.00002307, -0.00002810, -0.00003402, -0.00003559, -0.00003072, -0.00002131, -0.00001174, -0.00000608, -0.00000594, -0.00000991, -0.00001476, -0.00001743, -0.00001655, -0.00001283, -0.00000823, -0.00000468, -0.00000304, -0.00000289, -0.00000305, -0.00000240, -0.00000055, 0.00000203, 0.00000438, 0.00000560, 0.00000535, 0.00000400, 0.00000239, 0.00000142, 0.00000165, 0.00000307, 0.00000520, 0.00000729, 0.00000878, 0.00000946, 0.00000956, 0.00000955, 0.00000984, 0.00001060, 0.00001170, 0.00001284, 0.00001376, },
		{ 0.00000761, 0.00000724, 0.00000692, 0.00000692, 0.00000756, 0.00000892, 0.00001068, 0.00001208, 0.00001241, 0.00001146, 0.00000992, 0.00000912, 0.00001023, 0.00001335, 0.00001717, 0.00001954, 0.00001893, 0.00001580, 0.00001290, 0.00001397, 0.00002119, 0.00003304, 0.00004417, 0.00004812, 0.00004186, 0.00002961, 0.00002295, 0.00003586, 0.00007632, 0.00013869, 0.00020161, 0.00023442, 0.00021038, 0.00012099, -0.00001622, -0.00016163, -0.00026878, -0.00030461, -0.00026471, -0.00017546, -0.00008148, -0.00002405, -0.00002162, -0.00006242, -0.00011315, -0.00013869, -0.00012209, -0.00007357, -0.00002350, -0.00000369, -0.00002773, -0.00008217, -0.00013394, -0.00015019, -0.00011893, -0.00005843, -0.00000889, -0.00001043, -0.00007913, -0.00019467, -0.00030704, -0.00035960, -0.00031680, -0.00018231, 0.00000168, 0.00017354, 0.00027965, 0.00029662, 0.00023798, 0.00014350, 0.00005807, 0.00001136, 0.00000804, 0.00003145, 0.00005685, 0.00006581, 0.00005428, 0.00003140, 0.00001134, 0.00000412, 0.00001048, 0.00002308, 0.00003216, 0.00003172, 0.00002246, 0.00001041, 0.00000265, 0.00000296, 0.00001008, 0.00001902, 0.00002440, 0.00002343, 0.00001710, 0.00000903, 0.00000316, 0.00000158, 0.00000383, 0.00000764, 0.00001054, 0.00001119, 0.00000988, 0.00000800, 0.00000698, 0.00000751, 0.00000918, 0.00001092, 0.00001165, 0.00001083, 0.00000874, 0.00000620, 0.00000420, 0.00000340, 0.00000390, 0.00000527, 0.00000679, 0.00000783, 0.00000812, 0.00000779, 0.00000726, 0.00000697, 0.00000715, 0.00000771, 0.00000836, 0.00000880, 0.00000888, 0.00000867, 0.00000832, 0.00000797, },
		{ 0.00000112, 0.00000200, 0.00000330, 0.00000471, 0.00000575, 0.00000611, 0.00000597, 0.00000607, 0.00000725, 0.00000982, 0.00001312, 0.00001577, 0.00001652, 0.00001538, 0.00001403, 0.00001522, 0.00002093, 0.00003064, 0.00004077, 0.00004630, 0.00004415, 0.00003640, 0.00003120, 0.00003964, 0.00006955, 0.00011888, 0.00017285, 0.00020752, 0.00019955, 0.00013804, 0.00003246, -0.00008818, -0.00018592, -0.00022996, -0.00021058, -0.00014371, -0.00006326, -0.00000475, 0.00001157, -0.00001059, -0.00004777, -0.00007111, -0.00006298, -0.00002653, 0.00001728, 0.00004303, 0.00003669, 0.00000448, -0.00003073, -0.00004342, -0.00002149, 0.00002551, 0.00007132, 0.00008877, 0.00006725, 0.00002122, -0.00001591, -0.00001013, 0.00005254, 0.00015494, 0.00025396, 0.00030031, 0.00026270, 0.00014415, -0.00001824, -0.00016986, -0.00026292, -0.00027638, -0.00022226, -0.00013622, -0.00005868, -0.00001643, -0.00001354, -0.00003471, -0.00005718, -0.00006401, -0.00005158, -0.00002855, -0.00000836, -0.00000066, -0.00000630, -0.00001835, -0.00002731, -0.00002695, -0.00001742, -0.00000422, 0.00000569, 0.00000800, 0.00000302, -0.00000520, -0.00001168, -0.00001330, -0.00001012, -0.00000470, -0.00000018, 0.00000153, 0.00000052, -0.00000160, -0.00000295, -0.00000249, -0.00000049, 0.00000183, 0.00000314, 0.00000270, 0.00000071, -0.00000196, -0.00000422, -0.00000530, -0.00000502, -0.00000378, -0.00000226, -0.00000113, -0.00000070, -0.00000090, -0.00000135, -0.00000161, -0.00000142, -0.00000082, -0.00000010, 0.00000041, 0.00000054, 0.00000034, 0.00000006, -0.00000008, -0.00000001, 0.00000019, 0.00000041, 0.00000066, },
		{ 0.00001378, 0.00001505, 0.00001603, 0.00001692, 0.00001816, 0.00002008, 0.00002248, 0.00002463, 0.00002567, 0.00002541, 0.00002479, 0.00002566, 0.00002963, 0.00003674, 0.00004471, 0.00004986, 0.00004940, 0.00004407, 0.00003936, 0.00004382, 0.00006450, 0.00010160, 0.00014531, 0.00017736, 0.00017799, 0.00013539, 0.00005331, -0.00004779, -0.00013758, -0.00018836, -0.00018730, -0.00014238, -0.00007854, -0.00002560, -0.00000404, -0.00001592, -0.00004570, -0.00007027, -0.00007249, -0.00005070, -0.00001871, 0.00000330, 0.00000153, -0.00002284, -0.00005436, -0.00007294, -0.00006660, -0.00003932, -0.00000897, 0.00000362, -0.00001192, -0.00004815, -0.00008347, -0.00009514, -0.00007423, -0.00003328, -0.00000121, -0.00000762, -0.00006463, -0.00015694, -0.00024610, -0.00028796, -0.00025425, -0.00014740, -0.00000062, 0.00013678, 0.00022136, 0.00023376, 0.00018465, 0.00010646, 0.00003620, -0.00000148, -0.00000269, 0.00001843, 0.00004065, 0.00004799, 0.00003680, 0.00001507, -0.00000436, -0.00001197, -0.00000634, 0.00000638, 0.00001711, 0.00001937, 0.00001240, 0.00000060, -0.00000976, -0.00001429, -0.00001243, -0.00000707, -0.00000219, -0.00000052, -0.00000230, -0.00000567, -0.00000822, -0.00000844, -0.00000647, -0.00000369, -0.00000178, -0.00000176, -0.00000354, -0.00000617, -0.00000840, -0.00000935, -0.00000879, -0.00000714, -0.00000510, -0.00000336, -0.00000227, -0.00000176, -0.00000150, -0.00000106, -0.00000019, 0.00000106, 0.00000242, 0.00000355, 0.00000424, 0.00000456, 0.00000479, 0.00000525, 0.00000608, 0.00000715, 0.00000820, 0.00000903, 0.00000963, 0.00001024, 0.00001113, 0.00001237, },
		{ 0.00000915, 0.00000989, 0.00001077, 0.00001158, 0.00001193, 0.00001163, 0.00001115, 0.00001157, 0.00001406, 0.00001886, 0.00002466, 0.00002892, 0.00002936, 0.00002597, 0.00002229, 0.00002463, 0.00003902, 0.00006690, 0.00010209, 0.00013107, 0.00013758, 0.00011001, 0.00004824, -0.00003402, -0.00011317, -0.00016508, -0.00017554, -0.00014667, -0.00009589, -0.00004745, -0.00002089, -0.00002209, -0.00004184, -0.00006228, -0.00006763, -0.00005312, -0.00002699, -0.00000476, 0.00000071, -0.00001265, -0.00003492, -0.00005040, -0.00004766, -0.00002684, 0.00000021, 0.00001734, 0.00001455, -0.00000519, -0.00002727, -0.00003445, -0.00001820, 0.00001486, 0.00004595, 0.00005534, 0.00003549, -0.00000217, -0.00003157, -0.00002600, 0.00002570, 0.00010987, 0.00019180, 0.00023127, 0.00020217, 0.00010620, -0.00002676, -0.00015177, -0.00022883, -0.00023974, -0.00019382, -0.00012069, -0.00005452, -0.00001849, -0.00001662, -0.00003607, -0.00005705, -0.00006418, -0.00005344, -0.00003182, -0.00001128, -0.00000126, -0.00000398, -0.00001454, -0.00002489, -0.00002876, -0.00002468, -0.00001582, -0.00000723, -0.00000272, -0.00000299, -0.00000590, -0.00000830, -0.00000802, -0.00000494, -0.00000067, 0.00000268, 0.00000371, 0.00000239, -0.00000016, -0.00000249, -0.00000350, -0.00000294, -0.00000126, 0.00000073, 0.00000235, 0.00000326, 0.00000356, 0.00000360, 0.00000377, 0.00000430, 0.00000514, 0.00000600, 0.00000654, 0.00000657, 0.00000615, 0.00000563, 0.00000540, 0.00000567, 0.00000632, 0.00000699, 0.00000724, 0.00000691, 0.00000619, 0.00000552, 0.00000533, 0.00000578, 0.00000666, 0.00000764, 0.00000846, },
		{ 0.00001031, 0.00001073, 0.00001125, 0.00001241, 0.00001482, 0.00001875, 0.00002348, 0.00002742, 0.00002887, 0.00002753, 0.00002554, 0.00002737, 0.00003783, 0.00005893, 0.00008703, 0.00011244, 0.00012221, 0.00010565, 0.00006019, -0.00000565, -0.00007411, -0.00012486, -0.00014356, -0.00012827, -0.00009026, -0.00004856, -0.00002089, -0.00001543, -0.00002790, -0.00004520, -0.00005353, -0.00004639, -0.00002779, -0.00000917, -0.00000189, -0.00001003, -0.00002784, -0.00004348, -0.00004653, -0.00003474, -0.00001555, -0.00000138, -0.00000159, -0.00001624, -0.00003571, -0.00004677, -0.00004131, -0.00002223, -0.00000234, 0.00000340, -0.00001221, -0.00004289, -0.00007153, -0.00008018, -0.00006186, -0.00002682, 0.00000123, -0.00000226, -0.00004788, -0.00012367, -0.00019821, -0.00023460, -0.00020847, -0.00012069, 0.00000189, 0.00011819, 0.00019115, 0.00020334, 0.00016267, 0.00009600, 0.00003495, 0.00000148, 0.00000003, 0.00001930, 0.00004108, 0.00005067, 0.00004347, 0.00002507, 0.00000607, -0.00000471, -0.00000453, 0.00000298, 0.00001120, 0.00001474, 0.00001214, 0.00000585, -0.00000004, -0.00000244, -0.00000087, 0.00000279, 0.00000578, 0.00000615, 0.00000370, -0.00000030, -0.00000401, -0.00000611, -0.00000627, -0.00000510, -0.00000355, -0.00000240, -0.00000195, -0.00000202, -0.00000219, -0.00000208, -0.00000153, -0.00000066, 0.00000019, 0.00000069, 0.00000065, 0.00000018, -0.00000034, -0.00000047, 0.00000007, 0.00000119, 0.00000243, 0.00000319, 0.00000309, 0.00000223, 0.00000114, 0.00000054, 0.00000093, 0.00000231, 0.00000425, 0.00000617, 0.00000765, 0.00000863, 0.00000929, 0.00000983, },
		{ 0.00001212, 0.00001483, 0.00001800, 0.00002073, 0.00002199, 0.00002144, 0.00002034, 0.00002167, 0.00002904, 0.00004447, 0.00006610, 0.00008722, 0.00009787, 0.00008874, 0.00005594, 0.00000412, -0.00005388, -0.00010150, -0.00012520, -0.00012040, -0.00009349, -0.00005875, -0.00003150, -0.00002100, -0.00002664, -0.00003946, -0.00004800, -0.00004506, -0.00003158, -0.00001546, -0.00000621, -0.00000889, -0.00002083, -0.00003346, -0.00003774, -0.00003005, -0.00001451, -0.00000036, 0.00000403, -0.00000323, -0.00001616, -0.00002464, -0.00002115, -0.00000608, 0.00001200, 0.00002157, 0.00001579, -0.00000236, -0.00002101, -0.00002644, -0.00001198, 0.00001674, 0.00004408, 0.00005342, 0.00003802, 0.00000685, -0.00001861, -0.00001572, 0.00002570, 0.00009531, 0.00016471, 0.00020008, 0.00017859, 0.00010031, -0.00001079, -0.00011725, -0.00018479, -0.00019673, -0.00015952, -0.00009728, -0.00003904, -0.00000545, -0.00000147, -0.00001758, -0.00003751, -0.00004773, -0.00004344, -0.00002892, -0.00001310, -0.00000364, -0.00000301, -0.00000830, -0.00001398, -0.00001555, -0.00001190, -0.00000526, 0.00000077, 0.00000345, 0.00000230, -0.00000115, -0.00000451, -0.00000602, -0.00000529, -0.00000308, -0.00000065, 0.00000107, 0.00000183, 0.00000198, 0.00000211, 0.00000263, 0.00000363, 0.00000484, 0.00000587, 0.00000636, 0.00000616, 0.00000547, 0.00000470, 0.00000437, 0.00000481, 0.00000594, 0.00000727, 0.00000807, 0.00000776, 0.00000630, 0.00000425, 0.00000253, 0.00000198, 0.00000289, 0.00000487, 0.00000711, 0.00000877, 0.00000946, 0.00000932, 0.00000883, 0.00000846, 0.00000850, 0.00000905, 0.00001021, },
		{ 0.00001488, 0.00001551, 0.00001591, 0.00001795, 0.00002410, 0.00003608, 0.00005306, 0.00007065, 0.00008150, 0.00007790, 0.00005545, 0.00001607, -0.00003147, -0.00007418, -0.00009995, -0.00010271, -0.00008514, -0.00005733, -0.00003198, -0.00001843, -0.00001866, -0.00002723, -0.00003513, -0.00003529, -0.00002655, -0.00001384, -0.00000475, -0.00000451, -0.00001267, -0.00002339, -0.00002923, -0.00002597, -0.00001536, -0.00000397, 0.00000105, -0.00000321, -0.00001344, -0.00002216, -0.00002271, -0.00001394, -0.00000135, 0.00000617, 0.00000225, -0.00001236, -0.00002957, -0.00003868, -0.00003333, -0.00001620, 0.00000182, 0.00000798, -0.00000416, -0.00002979, -0.00005474, -0.00006368, -0.00005002, -0.00002148, 0.00000260, 0.00000146, -0.00003460, -0.00009665, -0.00015918, -0.00019133, -0.00017197, -0.00010059, 0.00000142, 0.00010013, 0.00016406, 0.00017748, 0.00014558, 0.00008979, 0.00003627, 0.00000398, -0.00000206, 0.00001012, 0.00002638, 0.00003481, 0.00003112, 0.00001895, 0.00000598, -0.00000125, -0.00000074, 0.00000477, 0.00001034, 0.00001199, 0.00000871, 0.00000238, -0.00000386, -0.00000761, -0.00000826, -0.00000690, -0.00000525, -0.00000457, -0.00000510, -0.00000623, -0.00000712, -0.00000721, -0.00000652, -0.00000551, -0.00000478, -0.00000473, -0.00000543, -0.00000662, -0.00000776, -0.00000830, -0.00000788, -0.00000658, -0.00000492, -0.00000373, -0.00000374, -0.00000518, -0.00000755, -0.00000978, -0.00001076, -0.00000988, -0.00000735, -0.00000414, -0.00000140, 0.00000006, 0.00000019, -0.00000040, -0.00000086, -0.00000055, 0.00000071, 0.00000277, 0.00000532, 0.00000811, 0.00001089, 0.00001330, },
		{ 0.00002818, 0.00003695, 0.00004941, 0.00006282, 0.00007206, 0.00007110, 0.00005565, 0.00002579, -0.00001290, -0.00005072, -0.00007734, -0.00008612, -0.00007703, -0.00005660, -0.00003483, -0.00002045, -0.00001708, -0.00002212, -0.00002896, -0.00003122, -0.00002641, -0.00001714, -0.00000904, -0.00000701, -0.00001202, -0.00002043, -0.00002645, -0.00002582, -0.00001863, -0.00000927, -0.00000353, -0.00000472, -0.00001143, -0.00001834, -0.00001981, -0.00001382, -0.00000363, 0.00000412, 0.00000372, -0.00000547, -0.00001807, -0.00002576, -0.00002256, -0.00000911, 0.00000714, 0.00001620, 0.00001195, -0.00000346, -0.00002020, -0.00002649, -0.00001610, 0.00000700, 0.00003016, 0.00003938, 0.00002838, 0.00000378, -0.00001712, -0.00001556, 0.00001741, 0.00007391, 0.00013095, 0.00016059, 0.00014355, 0.00007921, -0.00001323, -0.00010311, -0.00016194, -0.00017542, -0.00014809, -0.00009912, -0.00005193, -0.00002339, -0.00001789, -0.00002819, -0.00004170, -0.00004794, -0.00004333, -0.00003134, -0.00001896, -0.00001203, -0.00001232, -0.00001742, -0.00002295, -0.00002543, -0.00002386, -0.00001971, -0.00001541, -0.00001276, -0.00001206, -0.00001235, -0.00001229, -0.00001106, -0.00000875, -0.00000608, -0.00000395, -0.00000290, -0.00000293, -0.00000359, -0.00000419, -0.00000408, -0.00000293, -0.00000082, 0.00000169, 0.00000370, 0.00000436, 0.00000330, 0.00000091, -0.00000170, -0.00000319, -0.00000261, 0.00000007, 0.00000389, 0.00000744, 0.00000951, 0.00000975, 0.00000875, 0.00000763, 0.00000743, 0.00000858, 0.00001081, 0.00001346, 0.00001592, 0.00001784, 0.00001918, 0.00002001, 0.00002055, 0.00002131, 0.00002336, },
		{ 0.00005663, 0.00005712, 0.00004697, 0.00002516, -0.00000521, -0.00003732, -0.00006282, -0.00007519, -0.00007255, -0.00005851, -0.00004048, -0.00002612, -0.00001993, -0.00002149, -0.00002641, -0.00002933, -0.00002716, -0.00002071, -0.00001383, -0.00001076, -0.00001324, -0.00001939, -0.00002493, -0.00002599, -0.00002163, -0.00001450, -0.00000906, -0.00000858, -0.00001293, -0.00001858, -0.00002098, -0.00001775, -0.00001054, -0.00000419, -0.00000357, -0.00001015, -0.00002061, -0.00002859, -0.00002870, -0.00002029, -0.00000833, -0.00000064, -0.00000296, -0.00001496, -0.00002996, -0.00003875, -0.00003542, -0.00002160, -0.00000595, 0.00000081, -0.00000717, -0.00002662, -0.00004636, -0.00005388, -0.00004349, -0.00002113, -0.00000232, -0.00000386, -0.00003366, -0.00008437, -0.00013505, -0.00016046, -0.00014336, -0.00008363, 0.00000096, 0.00008244, 0.00013512, 0.00014642, 0.00012101, 0.00007665, 0.00003467, 0.00001020, 0.00000683, 0.00001764, 0.00003104, 0.00003749, 0.00003381, 0.00002318, 0.00001187, 0.00000508, 0.00000439, 0.00000777, 0.00001161, 0.00001314, 0.00001180, 0.00000900, 0.00000679, 0.00000651, 0.00000807, 0.00001036, 0.00001198, 0.00001211, 0.00001079, 0.00000873, 0.00000690, 0.00000604, 0.00000651, 0.00000810, 0.00001018, 0.00001179, 0.00001202, 0.00001039, 0.00000716, 0.00000339, 0.00000059, -0.00000004, 0.00000186, 0.00000546, 0.00000920, 0.00001150, 0.00001164, 0.00001001, 0.00000792, 0.00000676, 0.00000731, 0.00000937, 0.00001205, 0.00001433, 0.00001563, 0.00001600, 0.00001595, 0.00001610, 0.00001700, 0.00001924, 0.00002345, 0.00003019, 0.00003932, 0.00004924, },
		{ 0.00000046, -0.00002563, -0.00004866, -0.00006271, -0.00006481, -0.00005625, -0.00004203, -0.00002849, -0.00002036, -0.00001881, -0.00002141, -0.00002401, -0.00002339, -0.00001908, -0.00001342, -0.00000973, -0.00001010, -0.00001391, -0.00001824, -0.00001981, -0.00001715, -0.00001157, -0.00000629, -0.00000426, -0.00000615, -0.00000981, -0.00001175, -0.00000957, -0.00000382, 0.00000215, 0.00000435, 0.00000091, -0.00000631, -0.00001258, -0.00001325, -0.00000687, 0.00000345, 0.00001172, 0.00001277, 0.00000576, -0.00000488, -0.00001196, -0.00000998, 0.00000108, 0.00001539, 0.00002463, 0.00002333, 0.00001267, 0.00000023, -0.00000450, 0.00000366, 0.00002177, 0.00003983, 0.00004679, 0.00003784, 0.00001866, 0.00000344, 0.00000723, 0.00003666, 0.00008433, 0.00013062, 0.00015236, 0.00013424, 0.00007692, -0.00000245, -0.00007769, -0.00012517, -0.00013374, -0.00010829, -0.00006570, -0.00002583, -0.00000258, 0.00000091, -0.00000889, -0.00002134, -0.00002782, -0.00002552, -0.00001729, -0.00000857, -0.00000369, -0.00000369, -0.00000650, -0.00000889, -0.00000859, -0.00000548, -0.00000117, 0.00000223, 0.00000339, 0.00000234, 0.00000016, -0.00000175, -0.00000237, -0.00000136, 0.00000092, 0.00000364, 0.00000579, 0.00000644, 0.00000509, 0.00000191, -0.00000211, -0.00000545, -0.00000667, -0.00000512, -0.00000137, 0.00000302, 0.00000618, 0.00000695, 0.00000546, 0.00000299, 0.00000126, 0.00000146, 0.00000363, 0.00000676, 0.00000948, 0.00001085, 0.00001085, 0.00001024, 0.00001007, 0.00001116, 0.00001385, 0.00001816, 0.00002396, 0.00003094, 0.00003817, 0.00004374, 0.00004482, 0.00003851, 0.00002330, },
		{ -0.00005111, -0.00004729, -0.00003699, -0.00002506, -0.00001604, -0.00001215, -0.00001260, -0.00001452, -0.00001493, -0.00001256, -0.00000843, -0.00000496, -0.00000419, -0.00000636, -0.00000974, -0.00001173, -0.00001069, -0.00000696, -0.00000271, -0.00000040, -0.00000110, -0.00000373, -0.00000576, -0.00000502, -0.00000136, 0.00000306, 0.00000510, 0.00000281, -0.00000300, -0.00000897, -0.00001122, -0.00000794, -0.00000085, 0.00000557, 0.00000684, 0.00000142, -0.00000796, -0.00001580, -0.00001717, -0.00001103, -0.00000115, 0.00000602, 0.00000542, -0.00000320, -0.00001479, -0.00002210, -0.00002030, -0.00001054, 0.00000042, 0.00000427, -0.00000336, -0.00001943, -0.00003474, -0.00003944, -0.00002976, -0.00001161, 0.00000150, -0.00000426, -0.00003442, -0.00008124, -0.00012572, -0.00014594, -0.00012784, -0.00007258, 0.00000315, 0.00007448, 0.00011925, 0.00012725, 0.00010332, 0.00006335, 0.00002582, 0.00000342, -0.00000113, 0.00000613, 0.00001557, 0.00001966, 0.00001626, 0.00000835, 0.00000098, -0.00000209, -0.00000030, 0.00000405, 0.00000768, 0.00000836, 0.00000590, 0.00000183, -0.00000178, -0.00000344, -0.00000278, -0.00000045, 0.00000237, 0.00000451, 0.00000505, 0.00000353, 0.00000016, -0.00000420, -0.00000808, -0.00000998, -0.00000897, -0.00000532, -0.00000047, 0.00000356, 0.00000523, 0.00000423, 0.00000169, -0.00000046, -0.00000064, 0.00000157, 0.00000523, 0.00000867, 0.00001055, 0.00001055, 0.00000957, 0.00000902, 0.00001011, 0.00001322, 0.00001796, 0.00002354, 0.00002925, 0.00003444, 0.00003822, 0.00003912, 0.00003525, 0.00002514, 0.00000880, -0.00001144, -0.00003122, -0.00004561, },
		{ -0.00002082, -0.00001560, -0.00001443, -0.00001567, -0.00001686, -0.00001628, -0.00001393, -0.00001130, -0.00001016, -0.00001127, -0.00001384, -0.00001601, -0.00001617, -0.00001406, -0.00001093, -0.00000869, -0.00000856, -0.00001020, -0.00001192, -0.00001189, -0.00000955, -0.00000615, -0.00000405, -0.00000507, -0.00000910, -0.00001386, -0.00001626, -0.00001439, -0.00000898, -0.00000324, -0.00000090, -0.00000382, -0.00001053, -0.00001693, -0.00001874, -0.00001434, -0.00000594, 0.00000155, 0.00000365, -0.00000077, -0.00000851, -0.00001398, -0.00001280, -0.00000486, 0.00000535, 0.00001145, 0.00000945, 0.00000075, -0.00000831, -0.00001031, -0.00000180, 0.00001375, 0.00002735, 0.00002990, 0.00001838, -0.00000086, -0.00001444, -0.00000907, 0.00002029, 0.00006589, 0.00010925, 0.00012935, 0.00011288, 0.00006115, -0.00000992, -0.00007677, -0.00011868, -0.00012637, -0.00010469, -0.00006865, -0.00003536, -0.00001623, -0.00001324, -0.00002032, -0.00002833, -0.00003046, -0.00002513, -0.00001552, -0.00000672, -0.00000241, -0.00000322, -0.00000696, -0.00001048, -0.00001143, -0.00000928, -0.00000503, -0.00000042, 0.00000302, 0.00000437, 0.00000342, 0.00000051, -0.00000347, -0.00000722, -0.00000929, -0.00000862, -0.00000513, -0.00000001, 0.00000477, 0.00000733, 0.00000693, 0.00000439, 0.00000165, 0.00000069, 0.00000238, 0.00000603, 0.00000983, 0.00001200, 0.00001184, 0.00001014, 0.00000862, 0.00000893, 0.00001172, 0.00001642, 0.00002172, 0.00002635, 0.00002959, 0.00003118, 0.00003082, 0.00002774, 0.00002082, 0.00000941, -0.00000580, -0.00002231, -0.00003639, -0.00004452, -0.00004504, -0.00003896, -0.00002955, },
		{ -0.00001358, -0.00001394, -0.00001281, -0.00001081, -0.00000929, -0.00000929, -0.00001071, -0.00001241, -0.00001302, -0.00001187, -0.00000951, -0.00000730, -0.00000641, -0.00000701, -0.00000811, -0.00000828, -0.00000678, -0.00000419, -0.00000216, -0.00000228, -0.00000488, -0.00000855, -0.00001094, -0.00001024, -0.00000652, -0.00000194, 0.00000049, -0.00000113, -0.00000625, -0.00001193, -0.00001462, -0.00001243, -0.00000646, -0.00000031, 0.00000215, -0.00000069, -0.00000699, -0.00001258, -0.00001360, -0.00000919, -0.00000223, 0.00000237, 0.00000098, -0.00000614, -0.00001464, -0.00001879, -0.00001537, -0.00000632, 0.00000197, 0.00000249, -0.00000757, -0.00002423, -0.00003835, -0.00004091, -0.00002912, -0.00000935, 0.00000519, 0.00000150, -0.00002537, -0.00006797, -0.00010861, -0.00012728, -0.00011146, -0.00006267, 0.00000374, 0.00006542, 0.00010301, 0.00010822, 0.00008624, 0.00005171, 0.00002092, 0.00000447, 0.00000382, 0.00001277, 0.00002236, 0.00002593, 0.00002189, 0.00001324, 0.00000483, 0.00000038, 0.00000082, 0.00000455, 0.00000890, 0.00001161, 0.00001166, 0.00000916, 0.00000487, -0.00000019, -0.00000488, -0.00000799, -0.00000854, -0.00000623, -0.00000186, 0.00000278, 0.00000570, 0.00000570, 0.00000314, -0.00000022, -0.00000214, -0.00000126, 0.00000213, 0.00000625, 0.00000896, 0.00000906, 0.00000703, 0.00000474, 0.00000429, 0.00000677, 0.00001168, 0.00001732, 0.00002182, 0.00002410, 0.00002417, 0.00002261, 0.00001978, 0.00001525, 0.00000812, -0.00000212, -0.00001464, -0.00002707, -0.00003631, -0.00003989, -0.00003730, -0.00003028, -0.00002202, -0.00001560, -0.00001257, -0.00001250, },
		{ -0.00001097, -0.00001045, -0.00001111, -0.00001247, -0.00001346, -0.00001325, -0.00001184, -0.00001005, -0.00000891, -0.00000888, -0.00000954, -0.00000989, -0.00000912, -0.00000736, -0.00000564, -0.00000524, -0.00000673, -0.00000939, -0.00001151, -0.00001152, -0.00000908, -0.00000550, -0.00000308, -0.00000361, -0.00000712, -0.00001164, -0.00001432, -0.00001324, -0.00000872, -0.00000327, -0.00000008, -0.00000095, -0.00000505, -0.00000933, -0.00001053, -0.00000735, -0.00000150, 0.00000334, 0.00000395, -0.00000007, -0.00000582, -0.00000879, -0.00000600, 0.00000171, 0.00000979, 0.00001267, 0.00000763, -0.00000291, -0.00001220, -0.00001315, -0.00000299, 0.00001422, 0.00002926, 0.00003311, 0.00002286, 0.00000459, -0.00000897, -0.00000528, 0.00002037, 0.00006072, 0.00009872, 0.00011524, 0.00009851, 0.00005033, -0.00001419, -0.00007337, -0.00010865, -0.00011225, -0.00008942, -0.00005447, -0.00002321, -0.00000595, -0.00000420, -0.00001215, -0.00002128, -0.00002507, -0.00002167, -0.00001345, -0.00000463, 0.00000132, 0.00000310, 0.00000137, -0.00000227, -0.00000633, -0.00000972, -0.00001169, -0.00001157, -0.00000903, -0.00000445, 0.00000082, 0.00000490, 0.00000627, 0.00000474, 0.00000171, -0.00000055, -0.00000021, 0.00000297, 0.00000749, 0.00001097, 0.00001170, 0.00000968, 0.00000671, 0.00000527, 0.00000702, 0.00001180, 0.00001773, 0.00002236, 0.00002404, 0.00002271, 0.00001955, 0.00001594, 0.00001237, 0.00000812, 0.00000193, -0.00000684, -0.00001728, -0.00002699, -0.00003318, -0.00003409, -0.00002998, -0.00002303, -0.00001626, -0.00001196, -0.00001078, -0.00001170, -0.00001295, -0.00001318, -0.00001224, },
		{ -0.00001355, -0.00001365, -0.00001273, -0.00001111, -0.00000958, -0.00000879, -0.00000874, -0.00000883, -0.00000834, -0.00000704, -0.00000547, -0.00000458, -0.00000503, -0.00000660, -0.00000818, -0.00000850, -0.00000699, -0.00000434, -0.00000216, -0.00000204, -0.00000434, -0.00000787, -0.00001047, -0.00001035, -0.00000730, -0.00000297, 0.00000011, 0.00000015, -0.00000269, -0.00000633, -0.00000809, -0.00000655, -0.00000256, 0.00000119, 0.00000196, -0.00000106, -0.00000602, -0.00000946, -0.00000867, -0.00000378, 0.00000196, 0.00000408, -0.00000009, -0.00000926, -0.00001844, -0.00002183, -0.00001665, -0.00000538, 0.00000509, 0.00000750, -0.00000110, -0.00001688, -0.00003093, -0.00003440, -0.00002444, -0.00000695, 0.00000569, 0.00000129, -0.00002466, -0.00006501, -0.00010305, -0.00012007, -0.00010450, -0.00005800, 0.00000487, 0.00006315, 0.00009876, 0.00010400, 0.00008362, 0.00005109, 0.00002142, 0.00000456, 0.00000229, 0.00000958, 0.00001878, 0.00002387, 0.00002275, 0.00001687, 0.00000923, 0.00000234, -0.00000267, -0.00000569, -0.00000688, -0.00000621, -0.00000375, -0.00000020, 0.00000299, 0.00000419, 0.00000266, -0.00000083, -0.00000425, -0.00000544, -0.00000355, 0.00000038, 0.00000399, 0.00000507, 0.00000300, -0.00000071, -0.00000338, -0.00000275, 0.00000156, 0.00000776, 0.00001294, 0.00001475, 0.00001271, 0.00000826, 0.00000364, 0.00000040, -0.00000165, -0.00000410, -0.00000879, -0.00001626, -0.00002514, -0.00003268, -0.00003621, -0.00003460, -0.00002891, -0.00002175, -0.00001596, -0.00001313, -0.00001311, -0.00001443, -0.00001539, -0.00001514, -0.00001392, -0.00001271, -0.00001231, -0.00001281, },
		{ -0.00000590, -0.00000495, -0.00000468, -0.00000487, -0.00000493, -0.00000443, -0.00000348, -0.00000268, -0.00000268, -0.00000360, -0.00000489, -0.00000559, -0.00000507, -0.00000353, -0.00000201, -0.00000175, -0.00000332, -0.00000611, -0.00000856, -0.00000914, -0.00000734, -0.00000412, -0.00000136, -0.00000070, -0.00000242, -0.00000521, -0.00000698, -0.00000630, -0.00000346, -0.00000033, 0.00000085, -0.00000092, -0.00000454, -0.00000739, -0.00000710, -0.00000334, 0.00000166, 0.00000433, 0.00000214, -0.00000441, -0.00001162, -0.00001470, -0.00001080, -0.00000124, 0.00000884, 0.00001341, 0.00000941, -0.00000088, -0.00001079, -0.00001315, -0.00000499, 0.00000999, 0.00002302, 0.00002538, 0.00001419, -0.00000475, -0.00001921, -0.00001698, 0.00000673, 0.00004534, 0.00008286, 0.00010111, 0.00008860, 0.00004645, -0.00001184, -0.00006650, -0.00010039, -0.00010600, -0.00008751, -0.00005717, -0.00002877, -0.00001152, -0.00000734, -0.00001219, -0.00001985, -0.00002550, -0.00002746, -0.00002656, -0.00002434, -0.00002162, -0.00001833, -0.00001427, -0.00000997, -0.00000675, -0.00000581, -0.00000720, -0.00000944, -0.00001035, -0.00000842, -0.00000398, 0.00000087, 0.00000361, 0.00000294, -0.00000020, -0.00000312, -0.00000303, 0.00000116, 0.00000804, 0.00001445, 0.00001732, 0.00001549, 0.00001028, 0.00000458, 0.00000099, 0.00000022, 0.00000074, -0.00000013, -0.00000432, -0.00001173, -0.00002001, -0.00002593, -0.00002709, -0.00002328, -0.00001652, -0.00000983, -0.00000569, -0.00000486, -0.00000635, -0.00000829, -0.00000921, -0.00000871, -0.00000752, -0.00000665, -0.00000672, -0.00000748, -0.00000817, -0.00000811, -0.00000719, },
		{ -0.00000490, -0.00000493, -0.00000445, -0.00000370, -0.00000323, -0.00000342, -0.00000415, -0.00000485, -0.00000491, -0.00000421, -0.00000327, -0.00000297, -0.00000393, -0.00000594, -0.00000802, -0.00000894, -0.00000805, -0.00000580, -0.00000351, -0.00000255, -0.00000346, -0.00000551, -0.00000717, -0.00000716, -0.00000539, -0.00000309, -0.00000203, -0.00000322, -0.00000611, -0.00000877, -0.00000920, -0.00000679, -0.00000305, -0.00000079, -0.00000232, -0.00000769, -0.00001424, -0.00001802, -0.00001625, -0.00000940, -0.00000130, 0.00000300, 0.00000037, -0.00000815, -0.00001756, -0.00002189, -0.00001802, -0.00000811, 0.00000125, 0.00000289, -0.00000625, -0.00002258, -0.00003733, -0.00004158, -0.00003201, -0.00001379, 0.00000142, 0.00000163, -0.00001827, -0.00005254, -0.00008639, -0.00010273, -0.00009065, -0.00005102, 0.00000370, 0.00005525, 0.00008783, 0.00009462, 0.00007939, 0.00005316, 0.00002797, 0.00001157, 0.00000530, 0.00000569, 0.00000799, 0.00000923, 0.00000917, 0.00000915, 0.00001015, 0.00001174, 0.00001245, 0.00001119, 0.00000836, 0.00000581, 0.00000550, 0.00000800, 0.00001188, 0.00001453, 0.00001398, 0.00001037, 0.00000606, 0.00000421, 0.00000671, 0.00001282, 0.00001940, 0.00002275, 0.00002082, 0.00001446, 0.00000686, 0.00000161, 0.00000053, 0.00000258, 0.00000463, 0.00000346, -0.00000219, -0.00001080, -0.00001890, -0.00002304, -0.00002167, -0.00001591, -0.00000878, -0.00000346, -0.00000167, -0.00000307, -0.00000584, -0.00000794, -0.00000836, -0.00000738, -0.00000612, -0.00000563, -0.00000615, -0.00000714, -0.00000772, -0.00000739, -0.00000633, -0.00000519, -0.00000456, -0.00000459, },
		{ -0.00000454, -0.00000399, -0.00000402, -0.00000443, -0.00000476, -0.00000466, -0.00000422, -0.00000393, -0.00000434, -0.00000552, -0.00000696, -0.00000779, -0.00000736, -0.00000574, -0.00000376, -0.00000247, -0.00000251, -0.00000358, -0.00000464, -0.00000466, -0.00000331, -0.00000135, -0.00000008, -0.00000044, -0.00000223, -0.00000408, -0.00000435, -0.00000233, 0.00000105, 0.00000361, 0.00000330, -0.00000028, -0.00000528, -0.00000847, -0.00000724, -0.00000142, 0.00000622, 0.00001148, 0.00001126, 0.00000575, -0.00000140, -0.00000516, -0.00000231, 0.00000620, 0.00001543, 0.00001936, 0.00001473, 0.00000357, -0.00000751, -0.00001109, -0.00000360, 0.00001195, 0.00002732, 0.00003370, 0.00002738, 0.00001269, 0.00000030, 0.00000133, 0.00002063, 0.00005295, 0.00008456, 0.00009972, 0.00008855, 0.00005221, 0.00000247, -0.00004390, -0.00007287, -0.00007900, -0.00006652, -0.00004573, -0.00002705, -0.00001625, -0.00001313, -0.00001359, -0.00001334, -0.00001067, -0.00000679, -0.00000406, -0.00000372, -0.00000490, -0.00000534, -0.00000337, 0.00000070, 0.00000466, 0.00000600, 0.00000386, -0.00000011, -0.00000263, -0.00000103, 0.00000488, 0.00001236, 0.00001729, 0.00001666, 0.00001049, 0.00000188, -0.00000477, -0.00000643, -0.00000322, 0.00000171, 0.00000412, 0.00000126, -0.00000647, -0.00001578, -0.00002239, -0.00002340, -0.00001881, -0.00001131, -0.00000461, -0.00000140, -0.00000219, -0.00000535, -0.00000848, -0.00000984, -0.00000919, -0.00000757, -0.00000639, -0.00000642, -0.00000742, -0.00000845, -0.00000868, -0.00000789, -0.00000659, -0.00000557, -0.00000530, -0.00000565, -0.00000607, -0.00000602, -0.00000539, },
		{ -0.00000049, -0.00000079, -0.00000081, -0.00000067, -0.00000071, -0.00000120, -0.00000202, -0.00000264, -0.00000254, -0.00000155, -0.00000008, 0.00000115, 0.00000154, 0.00000109, 0.00000039, 0.00000023, 0.00000098, 0.00000227, 0.00000317, 0.00000292, 0.00000150, -0.00000017, -0.00000080, 0.00000034, 0.00000271, 0.00000466, 0.00000444, 0.00000143, -0.00000315, -0.00000671, -0.00000687, -0.00000304, 0.00000287, 0.00000743, 0.00000768, 0.00000316, -0.00000356, -0.00000833, -0.00000795, -0.00000250, 0.00000444, 0.00000775, 0.00000405, -0.00000585, -0.00001703, -0.00002318, -0.00002056, -0.00001056, 0.00000071, 0.00000597, 0.00000132, -0.00001100, -0.00002365, -0.00002856, -0.00002226, -0.00000874, 0.00000224, 0.00000053, -0.00001795, -0.00004784, -0.00007610, -0.00008822, -0.00007571, -0.00004072, 0.00000479, 0.00004484, 0.00006694, 0.00006740, 0.00005183, 0.00003105, 0.00001498, 0.00000818, 0.00000900, 0.00001232, 0.00001355, 0.00001138, 0.00000778, 0.00000574, 0.00000664, 0.00000925, 0.00001074, 0.00000905, 0.00000466, 0.00000045, -0.00000026, 0.00000386, 0.00001088, 0.00001658, 0.00001703, 0.00001108, 0.00000132, -0.00000739, -0.00001079, -0.00000775, -0.00000098, 0.00000460, 0.00000479, -0.00000145, -0.00001142, -0.00002032, -0.00002398, -0.00002107, -0.00001361, -0.00000562, -0.00000075, -0.00000048, -0.00000367, -0.00000768, -0.00001005, -0.00000987, -0.00000795, -0.00000595, -0.00000518, -0.00000583, -0.00000704, -0.00000767, -0.00000711, -0.00000561, -0.00000406, -0.00000324, -0.00000335, -0.00000390, -0.00000414, -0.00000362, -0.00000243, -0.00000113, -0.00000031, -0.00000018, },
		{ -0.00000278, -0.00000309, -0.00000362, -0.00000424, -0.00000460, -0.00000441, -0.00000373, -0.00000297, -0.00000259, -0.00000277, -0.00000325, -0.00000355, -0.00000331, -0.00000267, -0.00000220, -0.00000250, -0.00000365, -0.00000507, -0.00000577, -0.00000511, -0.00000334, -0.00000168, -0.00000159, -0.00000378, -0.00000748, -0.00001068, -0.00001127, -0.00000839, -0.00000326, 0.00000137, 0.00000280, 0.00000013, -0.00000499, -0.00000915, -0.00000935, -0.00000500, 0.00000141, 0.00000564, 0.00000434, -0.00000260, -0.00001161, -0.00001736, -0.00001596, -0.00000758, 0.00000337, 0.00001079, 0.00001065, 0.00000368, -0.00000488, -0.00000849, -0.00000358, 0.00000786, 0.00001915, 0.00002309, 0.00001688, 0.00000470, -0.00000404, -0.00000007, 0.00001967, 0.00004907, 0.00007483, 0.00008292, 0.00006592, 0.00002724, -0.00002015, -0.00005983, -0.00007965, -0.00007684, -0.00005808, -0.00003501, -0.00001777, -0.00001055, -0.00001106, -0.00001360, -0.00001347, -0.00000974, -0.00000498, -0.00000263, -0.00000398, -0.00000711, -0.00000832, -0.00000494, 0.00000244, 0.00001002, 0.00001310, 0.00000921, -0.00000010, -0.00000993, -0.00001489, -0.00001234, -0.00000407, 0.00000481, 0.00000876, 0.00000513, -0.00000443, -0.00001502, -0.00002128, -0.00002039, -0.00001340, -0.00000431, 0.00000238, 0.00000404, 0.00000109, -0.00000377, -0.00000745, -0.00000823, -0.00000645, -0.00000391, -0.00000245, -0.00000281, -0.00000434, -0.00000568, -0.00000579, -0.00000458, -0.00000290, -0.00000182, -0.00000191, -0.00000285, -0.00000378, -0.00000387, -0.00000292, -0.00000141, -0.00000014, 0.00000028, -0.00000020, -0.00000115, -0.00000202, -0.00000253, },
		{ -0.00000379, -0.00000387, -0.00000359, -0.00000303, -0.00000245, -0.00000211, -0.00000202, -0.00000197, -0.00000173, -0.00000129, -0.00000096, -0.00000111, -0.00000186, -0.00000282, -0.00000332, -0.00000279, -0.00000135, 0.00000014, 0.00000049, -0.00000101, -0.00000395, -0.00000680, -0.00000773, -0.00000577, -0.00000159, 0.00000268, 0.00000462, 0.00000306, -0.00000105, -0.00000506, -0.00000621, -0.00000350, 0.00000150, 0.00000534, 0.00000485, -0.00000075, -0.00000904, -0.00001556, -0.00001652, -0.00001118, -0.00000254, 0.00000431, 0.00000532, 0.00000005, -0.00000793, -0.00001325, -0.00001230, -0.00000570, 0.00000187, 0.00000462, -0.00000032, -0.00001062, -0.00001977, -0.00002117, -0.00001290, 0.00000006, 0.00000783, 0.00000119, -0.00002236, -0.00005576, -0.00008480, -0.00009478, -0.00007810, -0.00003849, 0.00001045, 0.00005172, 0.00007287, 0.00007120, 0.00005373, 0.00003233, 0.00001715, 0.00001195, 0.00001386, 0.00001678, 0.00001611, 0.00001159, 0.00000675, 0.00000576, 0.00000996, 0.00001657, 0.00002047, 0.00001764, 0.00000806, -0.00000392, -0.00001209, -0.00001218, -0.00000450, 0.00000611, 0.00001309, 0.00001201, 0.00000303, -0.00000928, -0.00001866, -0.00002073, -0.00001519, -0.00000569, 0.00000259, 0.00000579, 0.00000334, -0.00000232, -0.00000753, -0.00000965, -0.00000838, -0.00000544, -0.00000318, -0.00000296, -0.00000454, -0.00000647, -0.00000725, -0.00000637, -0.00000451, -0.00000295, -0.00000266, -0.00000362, -0.00000496, -0.00000554, -0.00000477, -0.00000293, -0.00000096, 0.00000017, 0.00000003, -0.00000106, -0.00000239, -0.00000331, -0.00000360, -0.00000351, -0.00000341, -0.00000354, },
		{ -0.00000330, -0.00000283, -0.00000247, -0.00000225, -0.00000212, -0.00000204, -0.00000208, -0.00000243, -0.00000314, -0.00000398, -0.00000447, -0.00000420, -0.00000319, -0.00000205, -0.00000171, -0.00000284, -0.00000526, -0.00000783, -0.00000903, -0.00000784, -0.00000453, -0.00000068, 0.00000163, 0.00000109, -0.00000191, -0.00000537, -0.00000689, -0.00000518, -0.00000111, 0.00000265, 0.00000319, -0.00000069, -0.00000748, -0.00001361, -0.00001548, -0.00001170, -0.00000412, 0.00000315, 0.00000621, 0.00000380, -0.00000185, -0.00000642, -0.00000634, -0.00000119, 0.00000590, 0.00001026, 0.00000896, 0.00000305, -0.00000277, -0.00000318, 0.00000393, 0.00001538, 0.00002399, 0.00002297, 0.00001072, -0.00000700, -0.00001925, -0.00001584, 0.00000637, 0.00004043, 0.00007170, 0.00008483, 0.00007155, 0.00003515, -0.00001092, -0.00004967, -0.00006868, -0.00006544, -0.00004730, -0.00002630, -0.00001237, -0.00000863, -0.00001123, -0.00001330, -0.00001012, -0.00000218, 0.00000553, 0.00000735, 0.00000111, -0.00001011, -0.00001968, -0.00002157, -0.00001414, -0.00000128, 0.00000986, 0.00001307, 0.00000664, -0.00000574, -0.00001728, -0.00002197, -0.00001790, -0.00000789, 0.00000243, 0.00000798, 0.00000692, 0.00000115, -0.00000529, -0.00000882, -0.00000824, -0.00000497, -0.00000171, -0.00000059, -0.00000191, -0.00000429, -0.00000584, -0.00000550, -0.00000363, -0.00000164, -0.00000088, -0.00000176, -0.00000351, -0.00000476, -0.00000448, -0.00000265, -0.00000021, 0.00000156, 0.00000181, 0.00000060, -0.00000129, -0.00000288, -0.00000358, -0.00000345, -0.00000301, -0.00000278, -0.00000297, -0.00000339, -0.00000370, -0.00000366, },
		{ -0.00000349, -0.00000337, -0.00000350, -0.00000384, -0.00000437, -0.00000489, -0.00000512, -0.00000477, -0.00000385, -0.00000278, -0.00000229, -0.00000295, -0.00000473, -0.00000683, -0.00000800, -0.00000725, -0.00000457, -0.00000107, 0.00000150, 0.00000183, -0.00000012, -0.00000295, -0.00000463, -0.00000378, -0.00000070, 0.00000262, 0.00000364, 0.00000090, -0.00000483, -0.00001079, -0.00001369, -0.00001176, -0.00000589, 0.00000070, 0.00000443, 0.00000348, -0.00000103, -0.00000577, -0.00000737, -0.00000471, 0.00000030, 0.00000398, 0.00000343, -0.00000125, -0.00000681, -0.00000890, -0.00000526, 0.00000221, 0.00000808, 0.00000664, -0.00000406, -0.00002016, -0.00003341, -0.00003589, -0.00002513, -0.00000664, 0.00000828, 0.00000861, -0.00000977, -0.00004068, -0.00006986, -0.00008190, -0.00006829, -0.00003212, 0.00001325, 0.00005084, 0.00006839, 0.00006388, 0.00004531, 0.00002538, 0.00001412, 0.00001398, 0.00001978, 0.00002311, 0.00001822, 0.00000573, -0.00000803, -0.00001528, -0.00001190, -0.00000006, 0.00001298, 0.00001930, 0.00001491, 0.00000199, -0.00001257, -0.00002134, -0.00002045, -0.00001149, -0.00000013, 0.00000736, 0.00000763, 0.00000167, -0.00000637, -0.00001198, -0.00001287, -0.00000988, -0.00000592, -0.00000386, -0.00000476, -0.00000752, -0.00000997, -0.00001042, -0.00000877, -0.00000637, -0.00000497, -0.00000548, -0.00000739, -0.00000921, -0.00000948, -0.00000775, -0.00000482, -0.00000221, -0.00000120, -0.00000213, -0.00000425, -0.00000633, -0.00000738, -0.00000719, -0.00000626, -0.00000539, -0.00000514, -0.00000550, -0.00000607, -0.00000635, -0.00000609, -0.00000540, -0.00000457, -0.00000389, },
		{ -0.00000168, -0.00000224, -0.00000249, -0.00000228, -0.00000164, -0.00000089, -0.00000056, -0.00000112, -0.00000264, -0.00000457, -0.00000590, -0.00000573, -0.00000385, -0.00000099, 0.00000146, 0.00000224, 0.00000100, -0.00000134, -0.00000316, -0.00000309, -0.00000103, 0.00000164, 0.00000279, 0.00000089, -0.00000387, -0.00000944, -0.00001294, -0.00001240, -0.00000797, -0.00000201, 0.00000231, 0.00000286, -0.00000011, -0.00000417, -0.00000632, -0.00000498, -0.00000106, 0.00000265, 0.00000347, 0.00000082, -0.00000315, -0.00000494, -0.00000226, 0.00000401, 0.00000982, 0.00001032, 0.00000326, -0.00000892, -0.00001976, -0.00002240, -0.00001392, 0.00000222, 0.00001751, 0.00002316, 0.00001566, -0.00000033, -0.00001390, -0.00001383, 0.00000445, 0.00003524, 0.00006448, 0.00007670, 0.00006325, 0.00002724, -0.00001766, -0.00005401, -0.00006930, -0.00006184, -0.00004063, -0.00001973, -0.00001024, -0.00001474, -0.00002694, -0.00003638, -0.00003517, -0.00002263, -0.00000532, 0.00000727, 0.00000849, -0.00000162, -0.00001668, -0.00002794, -0.00002935, -0.00002073, -0.00000734, 0.00000356, 0.00000692, 0.00000244, -0.00000599, -0.00001299, -0.00001501, -0.00001199, -0.00000682, -0.00000305, -0.00000267, -0.00000511, -0.00000803, -0.00000913, -0.00000765, -0.00000472, -0.00000238, -0.00000210, -0.00000383, -0.00000606, -0.00000696, -0.00000553, -0.00000229, 0.00000112, 0.00000300, 0.00000256, 0.00000031, -0.00000234, -0.00000399, -0.00000404, -0.00000290, -0.00000153, -0.00000085, -0.00000115, -0.00000205, -0.00000286, -0.00000303, -0.00000246, -0.00000147, -0.00000053, -0.00000000, 0.00000002, -0.00000037, -0.00000100, },
		{ -0.00000235, -0.00000182, -0.00000161, -0.00000209, -0.00000336, -0.00000505, -0.00000641, -0.00000663, -0.00000536, -0.00000304, -0.00000072, 0.00000044, -0.00000008, -0.00000180, -0.00000348, -0.00000387, -0.00000260, -0.00000053, 0.00000063, -0.00000061, -0.00000445, -0.00000948, -0.00001328, -0.00001383, -0.00001074, -0.00000560, -0.00000110, 0.00000051, -0.00000120, -0.00000461, -0.00000719, -0.00000713, -0.00000457, -0.00000148, -0.00000022, -0.00000178, -0.00000494, -0.00000699, -0.00000572, -0.00000131, 0.00000331, 0.00000405, -0.00000152, -0.00001209, -0.00002271, -0.00002739, -0.00002282, -0.00001096, 0.00000164, 0.00000733, 0.00000217, -0.00001124, -0.00002480, -0.00002956, -0.00002151, -0.00000481, 0.00000978, 0.00001082, -0.00000672, -0.00003740, -0.00006701, -0.00007952, -0.00006564, -0.00002830, 0.00001830, 0.00005561, 0.00006989, 0.00005888, 0.00003222, 0.00000562, -0.00000763, -0.00000331, 0.00001254, 0.00002812, 0.00003307, 0.00002421, 0.00000668, -0.00000972, -0.00001651, -0.00001122, 0.00000182, 0.00001455, 0.00002010, 0.00001639, 0.00000670, -0.00000299, -0.00000772, -0.00000623, -0.00000105, 0.00000365, 0.00000484, 0.00000231, -0.00000169, -0.00000425, -0.00000384, -0.00000115, 0.00000164, 0.00000244, 0.00000074, -0.00000218, -0.00000415, -0.00000362, -0.00000065, 0.00000319, 0.00000580, 0.00000584, 0.00000348, 0.00000015, -0.00000237, -0.00000303, -0.00000198, -0.00000029, 0.00000081, 0.00000066, -0.00000052, -0.00000191, -0.00000267, -0.00000243, -0.00000145, -0.00000031, 0.00000041, 0.00000042, -0.00000017, -0.00000108, -0.00000198, -0.00000264, -0.00000293, -0.00000281, },
		{ -0.00000460, -0.00000594, -0.00000711, -0.00000746, -0.00000656, -0.00000460, -0.00000234, -0.00000076, -0.00000049, -0.00000141, -0.00000265, -0.00000313, -0.00000232, -0.00000066, 0.00000054, -0.00000005, -0.00000292, -0.00000716, -0.00001082, -0.00001196, -0.00000981, -0.00000532, -0.00000065, 0.00000199, 0.00000170, -0.00000063, -0.00000298, -0.00000349, -0.00000175, 0.00000102, 0.00000287, 0.00000261, 0.00000074, -0.00000073, 0.00000016, 0.00000374, 0.00000801, 0.00000962, 0.00000613, -0.00000202, -0.00001111, -0.00001596, -0.00001310, -0.00000320, 0.00000878, 0.00001622, 0.00001489, 0.00000583, -0.00000484, -0.00000941, -0.00000348, 0.00001082, 0.00002552, 0.00003138, 0.00002384, 0.00000672, -0.00000905, -0.00001131, 0.00000591, 0.00003778, 0.00006983, 0.00008491, 0.00007220, 0.00003344, -0.00001723, -0.00005991, -0.00007871, -0.00006927, -0.00004023, -0.00000808, 0.00001171, 0.00001221, -0.00000251, -0.00002097, -0.00003139, -0.00002809, -0.00001379, 0.00000292, 0.00001317, 0.00001274, 0.00000379, -0.00000726, -0.00001397, -0.00001353, -0.00000773, -0.00000118, 0.00000200, 0.00000053, -0.00000365, -0.00000717, -0.00000759, -0.00000493, -0.00000131, 0.00000062, -0.00000037, -0.00000341, -0.00000616, -0.00000649, -0.00000381, 0.00000050, 0.00000408, 0.00000502, 0.00000298, -0.00000072, -0.00000401, -0.00000534, -0.00000446, -0.00000240, -0.00000070, -0.00000041, -0.00000159, -0.00000339, -0.00000470, -0.00000484, -0.00000387, -0.00000246, -0.00000140, -0.00000116, -0.00000176, -0.00000281, -0.00000385, -0.00000455, -0.00000478, -0.00000459, -0.00000412, -0.00000362, -0.00000338, -0.00000367, },
		{ -0.00000434, -0.00000276, -0.00000068, 0.00000113, 0.00000197, 0.00000173, 0.00000091, 0.00000037, 0.00000072, 0.00000182, 0.00000277, 0.00000242, 0.00000016, -0.00000354, -0.00000716, -0.00000894, -0.00000787, -0.00000436, -0.00000007, 0.00000297, 0.00000353, 0.00000190, -0.00000043, -0.00000170, -0.00000112, 0.00000074, 0.00000236, 0.00000252, 0.00000120, -0.00000026, -0.00000019, 0.00000203, 0.00000514, 0.00000650, 0.00000378, -0.00000317, -0.00001170, -0.00001749, -0.00001706, -0.00001015, -0.00000035, 0.00000671, 0.00000670, -0.00000057, -0.00001064, -0.00001691, -0.00001476, -0.00000481, 0.00000699, 0.00001261, 0.00000713, -0.00000772, -0.00002385, -0.00003124, -0.00002429, -0.00000602, 0.00001254, 0.00001805, 0.00000274, -0.00003016, -0.00006667, -0.00008856, -0.00008266, -0.00004805, 0.00000285, 0.00005007, 0.00007579, 0.00007271, 0.00004687, 0.00001361, -0.00001075, -0.00001695, -0.00000636, 0.00001134, 0.00002451, 0.00002603, 0.00001635, 0.00000207, -0.00000880, -0.00001155, -0.00000690, 0.00000037, 0.00000489, 0.00000410, -0.00000076, -0.00000597, -0.00000815, -0.00000639, -0.00000257, 0.00000025, -0.00000004, -0.00000318, -0.00000686, -0.00000839, -0.00000648, -0.00000205, 0.00000241, 0.00000442, 0.00000299, -0.00000084, -0.00000485, -0.00000693, -0.00000635, -0.00000396, -0.00000151, -0.00000049, -0.00000134, -0.00000329, -0.00000502, -0.00000550, -0.00000455, -0.00000279, -0.00000119, -0.00000050, -0.00000085, -0.00000186, -0.00000294, -0.00000362, -0.00000374, -0.00000339, -0.00000279, -0.00000220, -0.00000185, -0.00000195, -0.00000255, -0.00000354, -0.00000450, -0.00000491, },
		{ 0.00000017, 0.00000037, -0.00000014, -0.00000074, -0.00000085, -0.00000038, 0.00000014, -0.00000023, -0.00000212, -0.00000537, -0.00000888, -0.00001111, -0.00001094, -0.00000833, -0.00000449, -0.00000121, 0.00000013, -0.00000068, -0.00000263, -0.00000422, -0.00000437, -0.00000317, -0.00000166, -0.00000098, -0.00000150, -0.00000245, -0.00000251, -0.00000092, 0.00000163, 0.00000315, 0.00000159, -0.00000365, -0.00001078, -0.00001630, -0.00001691, -0.00001170, -0.00000307, 0.00000433, 0.00000631, 0.00000181, -0.00000614, -0.00001198, -0.00001110, -0.00000292, 0.00000826, 0.00001558, 0.00001388, 0.00000334, -0.00001002, -0.00001749, -0.00001303, 0.00000254, 0.00002115, 0.00003178, 0.00002696, 0.00000793, -0.00001504, -0.00002772, -0.00001995, 0.00000802, 0.00004445, 0.00007148, 0.00007417, 0.00004842, 0.00000338, -0.00004286, -0.00007221, -0.00007517, -0.00005472, -0.00002375, 0.00000223, 0.00001291, 0.00000738, -0.00000702, -0.00002002, -0.00002436, -0.00001918, -0.00000926, -0.00000120, 0.00000079, -0.00000297, -0.00000865, -0.00001182, -0.00001043, -0.00000577, -0.00000124, 0.00000019, -0.00000208, -0.00000600, -0.00000842, -0.00000721, -0.00000266, 0.00000282, 0.00000623, 0.00000587, 0.00000227, -0.00000229, -0.00000523, -0.00000520, -0.00000273, 0.00000036, 0.00000212, 0.00000167, -0.00000044, -0.00000277, -0.00000392, -0.00000336, -0.00000160, 0.00000026, 0.00000124, 0.00000096, -0.00000024, -0.00000166, -0.00000268, -0.00000304, -0.00000285, -0.00000239, -0.00000199, -0.00000183, -0.00000204, -0.00000265, -0.00000359, -0.00000460, -0.00000527, -0.00000521, -0.00000424, -0.00000262, -0.00000093, },
		{ 0.00000059, 0.00000076, 0.00000104, 0.00000077, -0.00000067, -0.00000334, -0.00000649, -0.00000884, -0.00000923, -0.00000733, -0.00000389, -0.00000042, 0.00000163, 0.00000167, 0.00000024, -0.00000140, -0.00000210, -0.00000154, -0.00000029, 0.00000064, 0.00000068, 0.00000015, -0.00000002, 0.00000092, 0.00000274, 0.00000402, 0.00000303, -0.00000105, -0.00000714, -0.00001255, -0.00001429, -0.00001104, -0.00000424, 0.00000245, 0.00000515, 0.00000221, -0.00000450, -0.00001051, -0.00001141, -0.00000581, 0.00000340, 0.00001047, 0.00001023, 0.00000167, -0.00001095, -0.00002015, -0.00001970, -0.00000876, 0.00000689, 0.00001771, 0.00001591, 0.00000044, -0.00002169, -0.00003887, -0.00004144, -0.00002762, -0.00000519, 0.00001238, 0.00001339, -0.00000509, -0.00003487, -0.00006047, -0.00006704, -0.00004836, -0.00001028, 0.00003216, 0.00006223, 0.00006972, 0.00005526, 0.00002900, 0.00000452, -0.00000821, -0.00000710, 0.00000247, 0.00001199, 0.00001505, 0.00001059, 0.00000243, -0.00000392, -0.00000499, -0.00000125, 0.00000373, 0.00000603, 0.00000398, -0.00000089, -0.00000512, -0.00000573, -0.00000214, 0.00000350, 0.00000781, 0.00000828, 0.00000478, -0.00000057, -0.00000478, -0.00000584, -0.00000376, -0.00000034, 0.00000211, 0.00000214, -0.00000003, -0.00000294, -0.00000485, -0.00000483, -0.00000315, -0.00000098, 0.00000043, 0.00000044, -0.00000073, -0.00000231, -0.00000354, -0.00000400, -0.00000379, -0.00000325, -0.00000277, -0.00000255, -0.00000264, -0.00000305, -0.00000370, -0.00000446, -0.00000507, -0.00000519, -0.00000454, -0.00000317, -0.00000143, 0.00000010, 0.00000095, 0.00000107, 0.00000078, },
		{ -0.00000097, -0.00000333, -0.00000630, -0.00000883, -0.00000983, -0.00000872, -0.00000590, -0.00000255, -0.00000010, 0.00000059, -0.00000036, -0.00000200, -0.00000317, -0.00000325, -0.00000241, -0.00000142, -0.00000096, -0.00000108, -0.00000121, -0.00000069, 0.00000054, 0.00000160, 0.00000110, -0.00000184, -0.00000672, -0.00001153, -0.00001374, -0.00001182, -0.00000644, -0.00000033, 0.00000308, 0.00000185, -0.00000310, -0.00000833, -0.00000983, -0.00000568, 0.00000239, 0.00000971, 0.00001139, 0.00000556, -0.00000501, -0.00001404, -0.00001535, -0.00000686, 0.00000762, 0.00002001, 0.00002249, 0.00001230, -0.00000606, -0.00002286, -0.00002844, -0.00001884, 0.00000160, 0.00002243, 0.00003296, 0.00002851, 0.00001328, -0.00000191, -0.00000587, 0.00000613, 0.00002922, 0.00005113, 0.00005874, 0.00004533, 0.00001439, -0.00002208, -0.00004970, -0.00005893, -0.00004934, -0.00002891, -0.00000910, 0.00000139, 0.00000049, -0.00000737, -0.00001496, -0.00001685, -0.00001231, -0.00000492, 0.00000034, 0.00000053, -0.00000362, -0.00000846, -0.00001008, -0.00000687, -0.00000042, 0.00000563, 0.00000796, 0.00000559, 0.00000030, -0.00000464, -0.00000649, -0.00000462, -0.00000061, 0.00000291, 0.00000394, 0.00000219, -0.00000097, -0.00000349, -0.00000401, -0.00000245, 0.00000004, 0.00000198, 0.00000242, 0.00000138, -0.00000036, -0.00000186, -0.00000255, -0.00000244, -0.00000192, -0.00000146, -0.00000129, -0.00000144, -0.00000183, -0.00000238, -0.00000307, -0.00000374, -0.00000417, -0.00000402, -0.00000314, -0.00000167, -0.00000006, 0.00000118, 0.00000172, 0.00000165, 0.00000131, 0.00000105, 0.00000086, 0.00000037, },
		{ -0.00001046, -0.00000983, -0.00000740, -0.00000405, -0.00000111, 0.00000036, 0.00000013, -0.00000118, -0.00000250, -0.00000296, -0.00000242, -0.00000135, -0.00000047, -0.00000010, -0.00000002, 0.00000029, 0.00000106, 0.00000186, 0.00000166, -0.00000036, -0.00000412, -0.00000825, -0.00001071, -0.00000996, -0.00000606, -0.00000096, 0.00000252, 0.00000234, -0.00000132, -0.00000592, -0.00000802, -0.00000546, 0.00000097, 0.00000764, 0.00001009, 0.00000593, -0.00000333, -0.00001260, -0.00001604, -0.00001069, 0.00000129, 0.00001328, 0.00001772, 0.00001064, -0.00000563, -0.00002318, -0.00003272, -0.00002880, -0.00001311, 0.00000617, 0.00001891, 0.00001865, 0.00000626, -0.00001070, -0.00002260, -0.00002345, -0.00001447, -0.00000328, 0.00000098, -0.00000658, -0.00002318, -0.00003954, -0.00004503, -0.00003382, -0.00000843, 0.00002134, 0.00004351, 0.00005002, 0.00004063, 0.00002230, 0.00000502, -0.00000374, -0.00000244, 0.00000472, 0.00001111, 0.00001194, 0.00000686, -0.00000046, -0.00000512, -0.00000424, 0.00000135, 0.00000791, 0.00001126, 0.00000939, 0.00000351, -0.00000295, -0.00000655, -0.00000587, -0.00000211, 0.00000192, 0.00000360, 0.00000204, -0.00000162, -0.00000514, -0.00000664, -0.00000566, -0.00000319, -0.00000090, -0.00000011, -0.00000105, -0.00000299, -0.00000485, -0.00000586, -0.00000590, -0.00000539, -0.00000488, -0.00000469, -0.00000483, -0.00000513, -0.00000550, -0.00000594, -0.00000647, -0.00000695, -0.00000708, -0.00000654, -0.00000523, -0.00000344, -0.00000167, -0.00000045, 0.00000002, -0.00000008, -0.00000043, -0.00000081, -0.00000135, -0.00000243, -0.00000433, -0.00000684, -0.00000922, },
		{ 0.00000055, 0.00000247, 0.00000271, 0.00000156, -0.00000003, -0.00000107, -0.00000109, -0.00000030, 0.00000065, 0.00000122, 0.00000134, 0.00000133, 0.00000152, 0.00000181, 0.00000158, 0.00000012, -0.00000274, -0.00000621, -0.00000874, -0.00000892, -0.00000642, -0.00000246, 0.00000076, 0.00000129, -0.00000118, -0.00000494, -0.00000716, -0.00000566, -0.00000054, 0.00000552, 0.00000863, 0.00000614, -0.00000140, -0.00001006, -0.00001454, -0.00001137, -0.00000131, 0.00001045, 0.00001693, 0.00001339, 0.00000038, -0.00001598, -0.00002711, -0.00002660, -0.00001402, 0.00000464, 0.00002013, 0.00002501, 0.00001773, 0.00000336, -0.00000948, -0.00001366, -0.00000749, 0.00000469, 0.00001555, 0.00001951, 0.00001620, 0.00001045, 0.00000888, 0.00001519, 0.00002721, 0.00003765, 0.00003820, 0.00002462, -0.00000030, -0.00002774, -0.00004722, -0.00005201, -0.00004243, -0.00002522, -0.00000961, -0.00000234, -0.00000448, -0.00001174, -0.00001762, -0.00001753, -0.00001118, -0.00000233, 0.00000396, 0.00000449, -0.00000033, -0.00000704, -0.00001147, -0.00001125, -0.00000701, -0.00000163, 0.00000168, 0.00000132, -0.00000194, -0.00000577, -0.00000778, -0.00000698, -0.00000414, -0.00000106, 0.00000061, 0.00000028, -0.00000145, -0.00000335, -0.00000443, -0.00000441, -0.00000370, -0.00000294, -0.00000255, -0.00000254, -0.00000269, -0.00000282, -0.00000296, -0.00000325, -0.00000369, -0.00000405, -0.00000389, -0.00000290, -0.00000113, 0.00000097, 0.00000275, 0.00000374, 0.00000386, 0.00000339, 0.00000268, 0.00000186, 0.00000073, -0.00000098, -0.00000328, -0.00000566, -0.00000727, -0.00000729, -0.00000551, -0.00000250, },
		{ 0.00000011, -0.00000133, -0.00000181, -0.00000127, -0.00000022, 0.00000066, 0.00000099, 0.00000088, 0.00000069, 0.00000059, 0.00000033, -0.00000061, -0.00000255, -0.00000515, -0.00000740, -0.00000810, -0.00000670, -0.00000382, -0.00000106, -0.00000009, -0.00000154, -0.00000438, -0.00000643, -0.00000567, -0.00000174, 0.00000351, 0.00000683, 0.00000556, -0.00000050, -0.00000848, -0.00001375, -0.00001263, -0.00000482, 0.00000588, 0.00001328, 0.00001229, 0.00000202, -0.00001315, -0.00002560, -0.00002856, -0.00001990, -0.00000358, 0.00001234, 0.00002000, 0.00001605, 0.00000332, -0.00001088, -0.00001899, -0.00001737, -0.00000804, 0.00000298, 0.00000947, 0.00000865, 0.00000240, -0.00000460, -0.00000835, -0.00000835, -0.00000788, -0.00001137, -0.00002081, -0.00003349, -0.00004276, -0.00004160, -0.00002700, -0.00000230, 0.00002388, 0.00004176, 0.00004536, 0.00003540, 0.00001861, 0.00000396, -0.00000209, 0.00000145, 0.00001028, 0.00001787, 0.00001938, 0.00001418, 0.00000563, -0.00000131, -0.00000322, 0.00000005, 0.00000564, 0.00000973, 0.00000990, 0.00000637, 0.00000149, -0.00000189, -0.00000218, 0.00000021, 0.00000344, 0.00000546, 0.00000529, 0.00000335, 0.00000091, -0.00000076, -0.00000117, -0.00000068, -0.00000003, 0.00000022, 0.00000002, -0.00000035, -0.00000060, -0.00000073, -0.00000098, -0.00000153, -0.00000226, -0.00000270, -0.00000230, -0.00000087, 0.00000129, 0.00000344, 0.00000488, 0.00000525, 0.00000469, 0.00000363, 0.00000241, 0.00000109, -0.00000055, -0.00000267, -0.00000501, -0.00000689, -0.00000748, -0.00000632, -0.00000369, -0.00000058, 0.00000178, 0.00000257, 0.00000176, },
		{ -0.00000118, 0.00000000, 0.00000063, 0.00000057, 0.00000011, -0.00000035, -0.00000073, -0.00000128, -0.00000238, -0.00000406, -0.00000578, -0.00000668, -0.00000612, -0.00000426, -0.00000209, -0.00000093, -0.00000153, -0.00000342, -0.00000501, -0.00000459, -0.00000153, 0.00000299, 0.00000639, 0.00000617, 0.00000164, -0.00000526, -0.00001067, -0.00001095, -0.00000499, 0.00000464, 0.00001270, 0.00001409, 0.00000689, -0.00000614, -0.00001858, -0.00002368, -0.00001810, -0.00000394, 0.00001214, 0.00002243, 0.00002231, 0.00001266, -0.00000069, -0.00001032, -0.00001129, -0.00000363, 0.00000807, 0.00001762, 0.00002082, 0.00001753, 0.00001112, 0.00000589, 0.00000422, 0.00000543, 0.00000686, 0.00000642, 0.00000458, 0.00000441, 0.00000934, 0.00002021, 0.00003359, 0.00004279, 0.00004126, 0.00002650, 0.00000214, -0.00002328, -0.00004024, -0.00004298, -0.00003221, -0.00001451, 0.00000130, 0.00000865, 0.00000621, -0.00000217, -0.00001024, -0.00001309, -0.00000960, -0.00000250, 0.00000375, 0.00000578, 0.00000312, -0.00000195, -0.00000608, -0.00000699, -0.00000458, -0.00000070, 0.00000226, 0.00000286, 0.00000121, -0.00000134, -0.00000328, -0.00000384, -0.00000326, -0.00000239, -0.00000195, -0.00000209, -0.00000246, -0.00000266, -0.00000259, -0.00000254, -0.00000290, -0.00000369, -0.00000448, -0.00000457, -0.00000345, -0.00000122, 0.00000141, 0.00000350, 0.00000441, 0.00000405, 0.00000285, 0.00000138, -0.00000009, -0.00000162, -0.00000346, -0.00000559, -0.00000754, -0.00000851, -0.00000787, -0.00000559, -0.00000243, 0.00000039, 0.00000182, 0.00000151, -0.00000002, -0.00000171, -0.00000259, -0.00000229, },
		{ 0.00000088, 0.00000003, -0.00000059, -0.00000095, -0.00000137, -0.00000216, -0.00000323, -0.00000410, -0.00000420, -0.00000333, -0.00000194, -0.00000093, -0.00000101, -0.00000210, -0.00000320, -0.00000297, -0.00000069, 0.00000299, 0.00000611, 0.00000648, 0.00000310, -0.00000285, -0.00000828, -0.00000980, -0.00000579, 0.00000225, 0.00001011, 0.00001297, 0.00000819, -0.00000285, -0.00001502, -0.00002201, -0.00001969, -0.00000857, 0.00000621, 0.00001750, 0.00001993, 0.00001273, 0.00000016, -0.00001098, -0.00001500, -0.00001033, -0.00000007, 0.00001009, 0.00001512, 0.00001331, 0.00000668, -0.00000074, -0.00000559, -0.00000696, -0.00000635, -0.00000602, -0.00000705, -0.00000854, -0.00000860, -0.00000629, -0.00000305, -0.00000230, -0.00000733, -0.00001860, -0.00003240, -0.00004197, -0.00004081, -0.00002645, -0.00000234, 0.00002325, 0.00004104, 0.00004516, 0.00003589, 0.00001921, 0.00000344, -0.00000492, -0.00000415, 0.00000259, 0.00000969, 0.00001246, 0.00000955, 0.00000311, -0.00000298, -0.00000563, -0.00000415, -0.00000032, 0.00000310, 0.00000408, 0.00000238, -0.00000068, -0.00000329, -0.00000435, -0.00000397, -0.00000305, -0.00000253, -0.00000274, -0.00000329, -0.00000362, -0.00000348, -0.00000318, -0.00000327, -0.00000403, -0.00000513, -0.00000574, -0.00000508, -0.00000294, 0.00000006, 0.00000284, 0.00000443, 0.00000445, 0.00000325, 0.00000154, -0.00000012, -0.00000162, -0.00000320, -0.00000506, -0.00000696, -0.00000822, -0.00000808, -0.00000621, -0.00000310, 0.00000011, 0.00000219, 0.00000246, 0.00000117, -0.00000069, -0.00000196, -0.00000199, -0.00000091, 0.00000053, 0.00000148, 0.00000154, },
		{ -0.00000251, -0.00000254, -0.00000293, -0.00000364, -0.00000423, -0.00000431, -0.00000386, -0.00000330, -0.00000319, -0.00000369, -0.00000429, -0.00000408, -0.00000239, 0.00000048, 0.00000317, 0.00000386, 0.00000143, -0.00000351, -0.00000865, -0.00001098, -0.00000861, -0.00000213, 0.00000523, 0.00000917, 0.00000667, -0.00000203, -0.00001313, -0.00002099, -0.00002112, -0.00001275, 0.00000051, 0.00001239, 0.00001722, 0.00001297, 0.00000233, -0.00000881, -0.00001454, -0.00001204, -0.00000295, 0.00000788, 0.00001514, 0.00001593, 0.00001098, 0.00000368, -0.00000216, -0.00000447, -0.00000367, -0.00000171, -0.00000038, -0.00000001, 0.00000040, 0.00000198, 0.00000471, 0.00000704, 0.00000701, 0.00000398, -0.00000020, -0.00000184, 0.00000252, 0.00001345, 0.00002744, 0.00003793, 0.00003853, 0.00002653, 0.00000483, -0.00001906, -0.00003642, -0.00004150, -0.00003417, -0.00001956, -0.00000524, 0.00000273, 0.00000248, -0.00000343, -0.00001009, -0.00001324, -0.00001144, -0.00000635, -0.00000121, 0.00000129, 0.00000047, -0.00000238, -0.00000512, -0.00000623, -0.00000556, -0.00000411, -0.00000306, -0.00000295, -0.00000345, -0.00000378, -0.00000350, -0.00000281, -0.00000243, -0.00000291, -0.00000413, -0.00000523, -0.00000514, -0.00000332, -0.00000013, 0.00000326, 0.00000556, 0.00000606, 0.00000492, 0.00000292, 0.00000090, -0.00000078, -0.00000228, -0.00000396, -0.00000584, -0.00000743, -0.00000789, -0.00000663, -0.00000382, -0.00000044, 0.00000214, 0.00000295, 0.00000190, -0.00000016, -0.00000193, -0.00000247, -0.00000164, -0.00000012, 0.00000106, 0.00000122, 0.00000034, -0.00000099, -0.00000205, -0.00000250, },
		{ -0.00000256, -0.00000334, -0.00000374, -0.00000373, -0.00000357, -0.00000352, -0.00000351, -0.00000305, -0.00000165, 0.00000062, 0.00000289, 0.00000376, 0.00000216, -0.00000176, -0.00000634, -0.00000907, -0.00000800, -0.00000305, 0.00000354, 0.00000807, 0.00000746, 0.00000104, -0.00000859, -0.00001670, -0.00001879, -0.00001313, -0.00000194, 0.00000959, 0.00001594, 0.00001417, 0.00000553, -0.00000520, -0.00001230, -0.00001214, -0.00000498, 0.00000532, 0.00001361, 0.00001620, 0.00001260, 0.00000537, -0.00000171, -0.00000582, -0.00000636, -0.00000475, -0.00000300, -0.00000228, -0.00000240, -0.00000238, -0.00000163, -0.00000066, -0.00000076, -0.00000286, -0.00000645, -0.00000960, -0.00001006, -0.00000705, -0.00000227, 0.00000063, -0.00000186, -0.00001061, -0.00002265, -0.00003210, -0.00003302, -0.00002272, -0.00000356, 0.00001780, 0.00003351, 0.00003828, 0.00003180, 0.00001854, 0.00000521, -0.00000271, -0.00000339, 0.00000109, 0.00000654, 0.00000927, 0.00000799, 0.00000397, -0.00000024, -0.00000257, -0.00000256, -0.00000128, -0.00000027, -0.00000039, -0.00000141, -0.00000238, -0.00000254, -0.00000193, -0.00000135, -0.00000168, -0.00000311, -0.00000486, -0.00000563, -0.00000447, -0.00000142, 0.00000239, 0.00000542, 0.00000653, 0.00000558, 0.00000336, 0.00000094, -0.00000100, -0.00000247, -0.00000390, -0.00000562, -0.00000735, -0.00000829, -0.00000761, -0.00000515, -0.00000165, 0.00000150, 0.00000302, 0.00000244, 0.00000042, -0.00000172, -0.00000272, -0.00000215, -0.00000053, 0.00000100, 0.00000145, 0.00000057, -0.00000108, -0.00000257, -0.00000318, -0.00000284, -0.00000209, -0.00000162, -0.00000183, },
		{ -0.00000508, -0.00000475, -0.00000421, -0.00000342, -0.00000217, -0.00000042, 0.00000134, 0.00000216, 0.00000113, -0.00000189, -0.00000583, -0.00000871, -0.00000869, -0.00000519, 0.00000036, 0.00000503, 0.00000585, 0.00000157, -0.00000633, -0.00001410, -0.00001752, -0.00001427, -0.00000531, 0.00000533, 0.00001258, 0.00001298, 0.00000658, -0.00000306, -0.00001071, -0.00001230, -0.00000698, 0.00000253, 0.00001154, 0.00001594, 0.00001424, 0.00000802, 0.00000073, -0.00000438, -0.00000588, -0.00000446, -0.00000204, -0.00000026, 0.00000040, 0.00000055, 0.00000102, 0.00000195, 0.00000268, 0.00000231, 0.00000067, -0.00000124, -0.00000178, 0.00000018, 0.00000421, 0.00000836, 0.00001028, 0.00000892, 0.00000556, 0.00000335, 0.00000547, 0.00001283, 0.00002293, 0.00003064, 0.00003076, 0.00002092, 0.00000325, -0.00001632, -0.00003083, -0.00003559, -0.00003032, -0.00001898, -0.00000742, -0.00000042, 0.00000046, -0.00000303, -0.00000740, -0.00000970, -0.00000902, -0.00000651, -0.00000418, -0.00000337, -0.00000403, -0.00000500, -0.00000510, -0.00000406, -0.00000269, -0.00000218, -0.00000313, -0.00000500, -0.00000633, -0.00000573, -0.00000280, 0.00000151, 0.00000548, 0.00000752, 0.00000710, 0.00000488, 0.00000211, -0.00000015, -0.00000167, -0.00000288, -0.00000435, -0.00000612, -0.00000749, -0.00000749, -0.00000557, -0.00000217, 0.00000138, 0.00000357, 0.00000353, 0.00000160, -0.00000093, -0.00000255, -0.00000246, -0.00000094, 0.00000081, 0.00000156, 0.00000072, -0.00000128, -0.00000332, -0.00000434, -0.00000395, -0.00000263, -0.00000134, -0.00000092, -0.00000160, -0.00000295, -0.00000426, -0.00000500, },
		{ -0.00000231, -0.00000084, 0.00000051, 0.00000121, 0.00000061, -0.00000157, -0.00000473, -0.00000747, -0.00000814, -0.00000591, -0.00000148, 0.00000299, 0.00000484, 0.00000246, -0.00000357, -0.00001051, -0.00001468, -0.00001344, -0.00000676, 0.00000249, 0.00000995, 0.00001197, 0.00000767, -0.00000057, -0.00000827, -0.00001125, -0.00000782, 0.00000039, 0.00000940, 0.00001500, 0.00001493, 0.00000983, 0.00000256, -0.00000354, -0.00000637, -0.00000590, -0.00000368, -0.00000155, -0.00000053, -0.00000048, -0.00000062, -0.00000041, -0.00000005, -0.00000022, -0.00000142, -0.00000329, -0.00000470, -0.00000452, -0.00000250, 0.00000032, 0.00000218, 0.00000174, -0.00000094, -0.00000425, -0.00000616, -0.00000562, -0.00000359, -0.00000267, -0.00000546, -0.00001261, -0.00002181, -0.00002853, -0.00002828, -0.00001917, -0.00000330, 0.00001394, 0.00002644, 0.00003022, 0.00002527, 0.00001521, 0.00000514, -0.00000098, -0.00000211, 0.00000003, 0.00000247, 0.00000303, 0.00000142, -0.00000098, -0.00000242, -0.00000215, -0.00000088, -0.00000013, -0.00000107, -0.00000350, -0.00000596, -0.00000665, -0.00000460, -0.00000043, 0.00000405, 0.00000683, 0.00000688, 0.00000456, 0.00000126, -0.00000161, -0.00000346, -0.00000463, -0.00000588, -0.00000761, -0.00000934, -0.00001002, -0.00000877, -0.00000563, -0.00000175, 0.00000118, 0.00000191, 0.00000034, -0.00000236, -0.00000451, -0.00000491, -0.00000352, -0.00000145, -0.00000021, -0.00000073, -0.00000282, -0.00000528, -0.00000671, -0.00000634, -0.00000447, -0.00000223, -0.00000090, -0.00000113, -0.00000270, -0.00000470, -0.00000618, -0.00000662, -0.00000609, -0.00000499, -0.00000369, },
		{ 0.00000152, -0.00000008, -0.00000258, -0.00000507, -0.00000626, -0.00000521, -0.00000203, 0.00000183, 0.00000420, 0.00000335, -0.00000088, -0.00000672, -0.00001115, -0.00001151, -0.00000705, 0.00000048, 0.00000753, 0.00001061, 0.00000816, 0.00000152, -0.00000573, -0.00000960, -0.00000782, -0.00000104, 0.00000756, 0.00001399, 0.00001547, 0.00001178, 0.00000506, -0.00000150, -0.00000537, -0.00000585, -0.00000395, -0.00000149, 0.00000016, 0.00000067, 0.00000060, 0.00000064, 0.00000098, 0.00000119, 0.00000075, -0.00000041, -0.00000161, -0.00000182, -0.00000046, 0.00000208, 0.00000455, 0.00000560, 0.00000471, 0.00000260, 0.00000078, 0.00000052, 0.00000195, 0.00000391, 0.00000481, 0.00000392, 0.00000216, 0.00000173, 0.00000471, 0.00001134, 0.00001930, 0.00002445, 0.00002295, 0.00001349, -0.00000164, -0.00001740, -0.00002835, -0.00003131, -0.00002675, -0.00001826, -0.00001044, -0.00000644, -0.00000663, -0.00000895, -0.00001062, -0.00001004, -0.00000761, -0.00000521, -0.00000465, -0.00000636, -0.00000898, -0.00001029, -0.00000868, -0.00000423, 0.00000130, 0.00000554, 0.00000683, 0.00000515, 0.00000186, -0.00000130, -0.00000326, -0.00000414, -0.00000484, -0.00000610, -0.00000779, -0.00000890, -0.00000823, -0.00000540, -0.00000124, 0.00000250, 0.00000418, 0.00000324, 0.00000056, -0.00000207, -0.00000307, -0.00000198, 0.00000025, 0.00000196, 0.00000184, -0.00000024, -0.00000316, -0.00000522, -0.00000521, -0.00000312, -0.00000012, 0.00000213, 0.00000246, 0.00000081, -0.00000189, -0.00000431, -0.00000544, -0.00000507, -0.00000367, -0.00000195, -0.00000040, 0.00000080, 0.00000166, 0.00000201, },
		{ -0.00000495, -0.00000479, -0.00000280, 0.00000028, 0.00000284, 0.00000322, 0.00000072, -0.00000374, -0.00000791, -0.00000936, -0.00000684, -0.00000116, 0.00000505, 0.00000866, 0.00000774, 0.00000273, -0.00000370, -0.00000798, -0.00000756, -0.00000227, 0.00000556, 0.00001235, 0.00001507, 0.00001277, 0.00000685, 0.00000013, -0.00000467, -0.00000625, -0.00000507, -0.00000269, -0.00000068, 0.00000022, 0.00000024, 0.00000006, 0.00000011, 0.00000025, 0.00000002, -0.00000085, -0.00000205, -0.00000277, -0.00000230, -0.00000060, 0.00000150, 0.00000280, 0.00000247, 0.00000064, -0.00000166, -0.00000318, -0.00000326, -0.00000232, -0.00000144, -0.00000159, -0.00000282, -0.00000420, -0.00000459, -0.00000365, -0.00000246, -0.00000304, -0.00000702, -0.00001420, -0.00002203, -0.00002653, -0.00002433, -0.00001469, -0.00000029, 0.00001385, 0.00002271, 0.00002379, 0.00001817, 0.00000974, 0.00000294, 0.00000041, 0.00000181, 0.00000456, 0.00000571, 0.00000388, 0.00000008, -0.00000311, -0.00000329, 0.00000017, 0.00000577, 0.00001077, 0.00001288, 0.00001144, 0.00000767, 0.00000361, 0.00000084, -0.00000041, -0.00000103, -0.00000213, -0.00000400, -0.00000580, -0.00000614, -0.00000414, -0.00000023, 0.00000394, 0.00000639, 0.00000606, 0.00000341, 0.00000020, -0.00000166, -0.00000122, 0.00000089, 0.00000295, 0.00000326, 0.00000126, -0.00000208, -0.00000488, -0.00000549, -0.00000345, 0.00000014, 0.00000335, 0.00000450, 0.00000305, -0.00000015, -0.00000348, -0.00000544, -0.00000543, -0.00000384, -0.00000165, 0.00000025, 0.00000143, 0.00000191, 0.00000190, 0.00000140, 0.00000027, -0.00000153, -0.00000356, },
		{ 0.00000052, 0.00000176, 0.00000083, -0.00000210, -0.00000560, -0.00000762, -0.00000668, -0.00000282, 0.00000226, 0.00000594, 0.00000622, 0.00000281, -0.00000249, -0.00000668, -0.00000717, -0.00000319, 0.00000377, 0.00001064, 0.00001441, 0.00001356, 0.00000877, 0.00000238, -0.00000291, -0.00000538, -0.00000490, -0.00000268, -0.00000029, 0.00000121, 0.00000171, 0.00000172, 0.00000180, 0.00000205, 0.00000217, 0.00000180, 0.00000102, 0.00000037, 0.00000055, 0.00000187, 0.00000387, 0.00000559, 0.00000608, 0.00000510, 0.00000327, 0.00000170, 0.00000127, 0.00000212, 0.00000354, 0.00000460, 0.00000479, 0.00000436, 0.00000406, 0.00000444, 0.00000531, 0.00000578, 0.00000501, 0.00000310, 0.00000143, 0.00000211, 0.00000652, 0.00001407, 0.00002183, 0.00002570, 0.00002248, 0.00001184, -0.00000312, -0.00001709, -0.00002511, -0.00002503, -0.00001845, -0.00000975, -0.00000357, -0.00000238, -0.00000537, -0.00000929, -0.00001057, -0.00000751, -0.00000115, 0.00000554, 0.00000950, 0.00000933, 0.00000590, 0.00000148, -0.00000178, -0.00000311, -0.00000327, -0.00000367, -0.00000512, -0.00000709, -0.00000810, -0.00000681, -0.00000312, 0.00000156, 0.00000507, 0.00000577, 0.00000362, 0.00000018, -0.00000237, -0.00000258, -0.00000063, 0.00000183, 0.00000280, 0.00000122, -0.00000228, -0.00000572, -0.00000703, -0.00000526, -0.00000121, 0.00000301, 0.00000522, 0.00000433, 0.00000091, -0.00000324, -0.00000615, -0.00000672, -0.00000511, -0.00000247, -0.00000010, 0.00000122, 0.00000146, 0.00000105, 0.00000037, -0.00000048, -0.00000165, -0.00000311, -0.00000445, -0.00000494, -0.00000404, -0.00000188, },
		{ -0.00000238, -0.00000448, -0.00000471, -0.00000253, 0.00000121, 0.00000456, 0.00000560, 0.00000359, -0.00000048, -0.00000427, -0.00000536, -0.00000255, 0.00000338, 0.00000997, 0.00001438, 0.00001472, 0.00001105, 0.00000512, -0.00000053, -0.00000391, -0.00000438, -0.00000271, -0.00000035, 0.00000146, 0.00000227, 0.00000236, 0.00000229, 0.00000235, 0.00000240, 0.00000213, 0.00000141, 0.00000060, 0.00000028, 0.00000088, 0.00000227, 0.00000371, 0.00000431, 0.00000358, 0.00000179, -0.00000016, -0.00000131, -0.00000124, -0.00000028, 0.00000075, 0.00000112, 0.00000065, -0.00000026, -0.00000103, -0.00000144, -0.00000182, -0.00000269, -0.00000421, -0.00000581, -0.00000640, -0.00000521, -0.00000266, -0.00000057, -0.00000134, -0.00000644, -0.00001509, -0.00002401, -0.00002873, -0.00002587, -0.00001513, 0.00000008, 0.00001409, 0.00002168, 0.00002062, 0.00001280, 0.00000304, -0.00000343, -0.00000372, 0.00000150, 0.00000874, 0.00001387, 0.00001444, 0.00001075, 0.00000522, 0.00000061, -0.00000161, -0.00000194, -0.00000202, -0.00000321, -0.00000549, -0.00000743, -0.00000730, -0.00000438, 0.00000034, 0.00000461, 0.00000631, 0.00000477, 0.00000117, -0.00000217, -0.00000330, -0.00000185, 0.00000077, 0.00000235, 0.00000130, -0.00000218, -0.00000624, -0.00000842, -0.00000721, -0.00000299, 0.00000215, 0.00000554, 0.00000551, 0.00000219, -0.00000262, -0.00000652, -0.00000786, -0.00000643, -0.00000340, -0.00000043, 0.00000127, 0.00000147, 0.00000073, -0.00000019, -0.00000089, -0.00000147, -0.00000225, -0.00000325, -0.00000407, -0.00000404, -0.00000283, -0.00000082, 0.00000097, 0.00000142, 0.00000009, },
		{ -0.00000026, 0.00000228, 0.00000356, 0.00000256, -0.00000039, -0.00000366, -0.00000510, -0.00000332, 0.00000149, 0.00000752, 0.00001220, 0.00001349, 0.00001088, 0.00000558, -0.00000018, -0.00000428, -0.00000568, -0.00000468, -0.00000250, -0.00000044, 0.00000076, 0.00000113, 0.00000114, 0.00000117, 0.00000127, 0.00000120, 0.00000075, 0.00000008, -0.00000036, -0.00000010, 0.00000094, 0.00000229, 0.00000316, 0.00000297, 0.00000171, 0.00000003, -0.00000118, -0.00000128, -0.00000034, 0.00000102, 0.00000199, 0.00000216, 0.00000163, 0.00000091, 0.00000043, 0.00000022, 0.00000000, -0.00000046, -0.00000097, -0.00000095, 0.00000017, 0.00000231, 0.00000445, 0.00000511, 0.00000330, -0.00000051, -0.00000414, -0.00000474, -0.00000040, 0.00000826, 0.00001785, 0.00002352, 0.00002140, 0.00001086, -0.00000476, -0.00001952, -0.00002764, -0.00002633, -0.00001704, -0.00000458, 0.00000547, 0.00000950, 0.00000738, 0.00000188, -0.00000333, -0.00000586, -0.00000570, -0.00000464, -0.00000468, -0.00000641, -0.00000866, -0.00000934, -0.00000708, -0.00000228, 0.00000294, 0.00000603, 0.00000563, 0.00000240, -0.00000140, -0.00000336, -0.00000248, 0.00000025, 0.00000253, 0.00000229, -0.00000085, -0.00000528, -0.00000832, -0.00000787, -0.00000373, 0.00000216, 0.00000683, 0.00000793, 0.00000504, -0.00000023, -0.00000516, -0.00000750, -0.00000654, -0.00000332, 0.00000022, 0.00000242, 0.00000268, 0.00000160, 0.00000025, -0.00000058, -0.00000084, -0.00000102, -0.00000161, -0.00000258, -0.00000335, -0.00000322, -0.00000196, -0.00000010, 0.00000130, 0.00000137, 0.00000002, -0.00000188, -0.00000301, -0.00000243, },
		{ -0.00000022, -0.00000268, -0.00000407, -0.00000296, 0.00000094, 0.00000640, 0.00001123, 0.00001337, 0.00001189, 0.00000745, 0.00000191, -0.00000263, -0.00000479, -0.00000445, -0.00000254, -0.00000031, 0.00000130, 0.00000205, 0.00000225, 0.00000234, 0.00000248, 0.00000255, 0.00000232, 0.00000178, 0.00000126, 0.00000118, 0.00000178, 0.00000280, 0.00000364, 0.00000369, 0.00000275, 0.00000122, -0.00000013, -0.00000061, -0.00000005, 0.00000111, 0.00000213, 0.00000244, 0.00000193, 0.00000096, 0.00000002, -0.00000062, -0.00000105, -0.00000149, -0.00000198, -0.00000224, -0.00000188, -0.00000081, 0.00000040, 0.00000072, -0.00000074, -0.00000392, -0.00000753, -0.00000957, -0.00000853, -0.00000460, 0.00000001, 0.00000201, -0.00000113, -0.00000936, -0.00001954, -0.00002664, -0.00002616, -0.00001664, -0.00000072, 0.00001600, 0.00002752, 0.00003033, 0.00002492, 0.00001513, 0.00000588, 0.00000055, -0.00000047, 0.00000067, 0.00000114, -0.00000057, -0.00000371, -0.00000598, -0.00000530, -0.00000133, 0.00000412, 0.00000819, 0.00000880, 0.00000590, 0.00000152, -0.00000160, -0.00000178, 0.00000055, 0.00000319, 0.00000366, 0.00000093, -0.00000383, -0.00000784, -0.00000848, -0.00000489, 0.00000138, 0.00000718, 0.00000956, 0.00000736, 0.00000185, -0.00000410, -0.00000764, -0.00000746, -0.00000429, -0.00000024, 0.00000257, 0.00000309, 0.00000178, -0.00000004, -0.00000118, -0.00000128, -0.00000090, -0.00000091, -0.00000174, -0.00000303, -0.00000386, -0.00000347, -0.00000183, 0.00000027, 0.00000168, 0.00000167, 0.00000041, -0.00000111, -0.00000177, -0.00000107, 0.00000051, 0.00000177, 0.00000159, },
		{ 0.00000010, 0.00000486, 0.00000958, 0.00001227, 0.00001173, 0.00000812, 0.00000289, -0.00000196, -0.00000487, -0.00000531, -0.00000387, -0.00000171, 0.00000014, 0.00000120, 0.00000159, 0.00000173, 0.00000188, 0.00000201, 0.00000194, 0.00000153, 0.00000097, 0.00000063, 0.00000081, 0.00000148, 0.00000220, 0.00000242, 0.00000184, 0.00000062, -0.00000064, -0.00000126, -0.00000093, 0.00000012, 0.00000127, 0.00000187, 0.00000166, 0.00000087, -0.00000006, -0.00000072, -0.00000104, -0.00000117, -0.00000128, -0.00000127, -0.00000089, 0.00000002, 0.00000120, 0.00000193, 0.00000142, -0.00000052, -0.00000314, -0.00000492, -0.00000444, -0.00000137, 0.00000304, 0.00000640, 0.00000657, 0.00000307, -0.00000223, -0.00000590, -0.00000476, 0.00000212, 0.00001248, 0.00002175, 0.00002526, 0.00002077, 0.00000972, -0.00000341, -0.00001348, -0.00001725, -0.00001495, -0.00000971, -0.00000554, -0.00000486, -0.00000730, -0.00001019, -0.00001052, -0.00000700, -0.00000089, 0.00000478, 0.00000719, 0.00000548, 0.00000124, -0.00000260, -0.00000368, -0.00000172, 0.00000137, 0.00000284, 0.00000098, -0.00000361, -0.00000830, -0.00001000, -0.00000712, -0.00000065, 0.00000627, 0.00001015, 0.00000908, 0.00000376, -0.00000296, -0.00000773, -0.00000853, -0.00000563, -0.00000111, 0.00000244, 0.00000348, 0.00000214, -0.00000014, -0.00000171, -0.00000185, -0.00000101, -0.00000035, -0.00000079, -0.00000232, -0.00000399, -0.00000458, -0.00000346, -0.00000104, 0.00000147, 0.00000282, 0.00000248, 0.00000094, -0.00000069, -0.00000140, -0.00000099, -0.00000014, 0.00000014, -0.00000073, -0.00000237, -0.00000356, -0.00000293, },
		{ 0.00001083, 0.00000812, 0.00000340, -0.00000154, -0.00000499, -0.00000613, -0.00000518, -0.00000312, -0.00000107, 0.00000034, 0.00000103, 0.00000134, 0.00000160, 0.00000188, 0.00000202, 0.00000184, 0.00000136, 0.00000088, 0.00000073, 0.00000104, 0.00000158, 0.00000190, 0.00000162, 0.00000075, -0.00000033, -0.00000101, -0.00000091, -0.00000008, 0.00000099, 0.00000170, 0.00000168, 0.00000099, 0.00000003, -0.00000079, -0.00000124, -0.00000142, -0.00000150, -0.00000154, -0.00000140, -0.00000089, -0.00000010, 0.00000048, 0.00000019, -0.00000134, -0.00000372, -0.00000586, -0.00000642, -0.00000475, -0.00000148, 0.00000160, 0.00000248, 0.00000012, -0.00000466, -0.00000944, -0.00001148, -0.00000937, -0.00000403, 0.00000163, 0.00000430, 0.00000229, -0.00000334, -0.00000921, -0.00001151, -0.00000826, -0.00000056, 0.00000786, 0.00001280, 0.00001191, 0.00000611, -0.00000101, -0.00000530, -0.00000447, 0.00000071, 0.00000694, 0.00001055, 0.00000966, 0.00000520, 0.00000014, -0.00000253, -0.00000171, 0.00000117, 0.00000322, 0.00000209, -0.00000234, -0.00000777, -0.00001081, -0.00000913, -0.00000304, 0.00000460, 0.00000989, 0.00001011, 0.00000524, -0.00000209, -0.00000818, -0.00001027, -0.00000798, -0.00000324, 0.00000105, 0.00000275, 0.00000157, -0.00000110, -0.00000325, -0.00000366, -0.00000254, -0.00000121, -0.00000105, -0.00000254, -0.00000484, -0.00000644, -0.00000609, -0.00000369, -0.00000033, 0.00000235, 0.00000316, 0.00000199, -0.00000022, -0.00000217, -0.00000300, -0.00000280, -0.00000234, -0.00000243, -0.00000321, -0.00000398, -0.00000359, -0.00000124, 0.00000285, 0.00000737, 0.00001049, },
		{ -0.00000331, -0.00000514, -0.00000477, -0.00000298, -0.00000087, 0.00000078, 0.00000170, 0.00000215, 0.00000248, 0.00000285, 0.00000314, 0.00000313, 0.00000272, 0.00000209, 0.00000158, 0.00000143, 0.00000164, 0.00000187, 0.00000177, 0.00000118, 0.00000030, -0.00000040, -0.00000052, 0.00000003, 0.00000093, 0.00000165, 0.00000177, 0.00000123, 0.00000032, -0.00000054, -0.00000108, -0.00000128, -0.00000131, -0.00000129, -0.00000119, -0.00000083, -0.00000020, 0.00000041, 0.00000047, -0.00000048, -0.00000233, -0.00000430, -0.00000522, -0.00000423, -0.00000148, 0.00000176, 0.00000364, 0.00000282, -0.00000060, -0.00000495, -0.00000772, -0.00000703, -0.00000275, 0.00000334, 0.00000853, 0.00001078, 0.00000991, 0.00000770, 0.00000660, 0.00000800, 0.00001111, 0.00001320, 0.00001125, 0.00000402, -0.00000683, -0.00001722, -0.00002274, -0.00002105, -0.00001324, -0.00000328, 0.00000412, 0.00000608, 0.00000292, -0.00000231, -0.00000590, -0.00000579, -0.00000275, 0.00000037, 0.00000065, -0.00000285, -0.00000833, -0.00001229, -0.00001171, -0.00000605, 0.00000227, 0.00000915, 0.00001111, 0.00000730, -0.00000013, -0.00000727, -0.00001067, -0.00000918, -0.00000434, 0.00000075, 0.00000337, 0.00000266, -0.00000019, -0.00000291, -0.00000375, -0.00000252, -0.00000054, 0.00000038, -0.00000074, -0.00000341, -0.00000593, -0.00000653, -0.00000450, -0.00000061, 0.00000328, 0.00000539, 0.00000494, 0.00000250, -0.00000046, -0.00000256, -0.00000323, -0.00000289, -0.00000242, -0.00000241, -0.00000266, -0.00000233, -0.00000054, 0.00000289, 0.00000708, 0.00001046, 0.00001154, 0.00000968, 0.00000552, 0.00000061, },
		{ -0.00000148, 0.00000033, 0.00000146, 0.00000204, 0.00000245, 0.00000291, 0.00000337, 0.00000357, 0.00000329, 0.00000259, 0.00000177, 0.00000119, 0.00000101, 0.00000109, 0.00000111, 0.00000082, 0.00000023, -0.00000037, -0.00000061, -0.00000029, 0.00000041, 0.00000108, 0.00000133, 0.00000097, 0.00000020, -0.00000064, -0.00000122, -0.00000146, -0.00000148, -0.00000145, -0.00000138, -0.00000116, -0.00000071, -0.00000016, 0.00000005, -0.00000052, -0.00000198, -0.00000384, -0.00000510, -0.00000484, -0.00000282, 0.00000010, 0.00000234, 0.00000240, -0.00000018, -0.00000434, -0.00000793, -0.00000884, -0.00000623, -0.00000106, 0.00000440, 0.00000781, 0.00000809, 0.00000590, 0.00000304, 0.00000111, 0.00000038, -0.00000045, -0.00000320, -0.00000862, -0.00001542, -0.00002045, -0.00002033, -0.00001357, -0.00000174, 0.00001093, 0.00001956, 0.00002117, 0.00001617, 0.00000808, 0.00000138, -0.00000096, 0.00000092, 0.00000419, 0.00000528, 0.00000237, -0.00000345, -0.00000872, -0.00000986, -0.00000542, 0.00000282, 0.00001078, 0.00001432, 0.00001165, 0.00000425, -0.00000398, -0.00000899, -0.00000879, -0.00000434, 0.00000126, 0.00000477, 0.00000464, 0.00000170, -0.00000168, -0.00000322, -0.00000221, 0.00000018, 0.00000188, 0.00000131, -0.00000148, -0.00000484, -0.00000655, -0.00000525, -0.00000124, 0.00000362, 0.00000704, 0.00000756, 0.00000524, 0.00000150, -0.00000183, -0.00000357, -0.00000366, -0.00000289, -0.00000220, -0.00000191, -0.00000151, -0.00000017, 0.00000260, 0.00000636, 0.00000983, 0.00001150, 0.00001044, 0.00000688, 0.00000210, -0.00000221, -0.00000470, -0.00000497, -0.00000355, },
		{ 0.00000159, 0.00000220, 0.00000290, 0.00000340, 0.00000338, 0.00000275, 0.00000175, 0.00000080, 0.00000024, 0.00000013, 0.00000024, 0.00000025, 0.00000003, -0.00000034, -0.00000056, -0.00000041, 0.00000010, 0.00000071, 0.00000105, 0.00000090, 0.00000034, -0.00000038, -0.00000093, -0.00000118, -0.00000119, -0.00000111, -0.00000102, -0.00000085, -0.00000048, 0.00000003, 0.00000039, 0.00000018, -0.00000086, -0.00000243, -0.00000377, -0.00000395, -0.00000250, 0.00000013, 0.00000265, 0.00000354, 0.00000197, -0.00000156, -0.00000529, -0.00000710, -0.00000563, -0.00000115, 0.00000457, 0.00000916, 0.00001097, 0.00000989, 0.00000723, 0.00000474, 0.00000342, 0.00000292, 0.00000200, -0.00000034, -0.00000379, -0.00000652, -0.00000623, -0.00000173, 0.00000593, 0.00001352, 0.00001721, 0.00001472, 0.00000672, -0.00000325, -0.00001070, -0.00001263, -0.00000916, -0.00000344, 0.00000034, -0.00000049, -0.00000547, -0.00001123, -0.00001361, -0.00001020, -0.00000195, 0.00000726, 0.00001279, 0.00001190, 0.00000525, -0.00000352, -0.00000990, -0.00001102, -0.00000711, -0.00000105, 0.00000352, 0.00000434, 0.00000167, -0.00000215, -0.00000442, -0.00000382, -0.00000114, 0.00000137, 0.00000161, -0.00000095, -0.00000489, -0.00000771, -0.00000738, -0.00000362, 0.00000194, 0.00000665, 0.00000837, 0.00000654, 0.00000235, -0.00000209, -0.00000500, -0.00000574, -0.00000492, -0.00000367, -0.00000277, -0.00000216, -0.00000112, 0.00000106, 0.00000435, 0.00000782, 0.00001000, 0.00000975, 0.00000688, 0.00000237, -0.00000217, -0.00000523, -0.00000609, -0.00000505, -0.00000304, -0.00000108, 0.00000028, 0.00000105, },
		{ 0.00000420, 0.00000370, 0.00000255, 0.00000120, 0.00000017, -0.00000029, -0.00000025, -0.00000003, 0.00000006, -0.00000007, -0.00000026, -0.00000030, -0.00000005, 0.00000035, 0.00000064, 0.00000060, 0.00000017, -0.00000045, -0.00000101, -0.00000134, -0.00000143, -0.00000141, -0.00000138, -0.00000131, -0.00000109, -0.00000071, -0.00000036, -0.00000041, -0.00000115, -0.00000250, -0.00000390, -0.00000449, -0.00000366, -0.00000153, 0.00000091, 0.00000225, 0.00000143, -0.00000152, -0.00000529, -0.00000788, -0.00000764, -0.00000425, 0.00000106, 0.00000607, 0.00000880, 0.00000851, 0.00000599, 0.00000289, 0.00000063, -0.00000049, -0.00000131, -0.00000297, -0.00000588, -0.00000912, -0.00001089, -0.00000971, -0.00000560, -0.00000042, 0.00000308, 0.00000285, -0.00000100, -0.00000606, -0.00000890, -0.00000718, -0.00000120, 0.00000602, 0.00001039, 0.00000915, 0.00000269, -0.00000549, -0.00001068, -0.00000963, -0.00000253, 0.00000704, 0.00001401, 0.00001468, 0.00000878, -0.00000054, -0.00000849, -0.00001140, -0.00000862, -0.00000259, 0.00000278, 0.00000454, 0.00000227, -0.00000194, -0.00000509, -0.00000521, -0.00000259, 0.00000057, 0.00000171, -0.00000036, -0.00000462, -0.00000847, -0.00000932, -0.00000616, -0.00000025, 0.00000563, 0.00000873, 0.00000777, 0.00000352, -0.00000179, -0.00000585, -0.00000742, -0.00000673, -0.00000502, -0.00000346, -0.00000247, -0.00000154, 0.00000015, 0.00000299, 0.00000638, 0.00000902, 0.00000957, 0.00000749, 0.00000339, -0.00000125, -0.00000479, -0.00000624, -0.00000561, -0.00000373, -0.00000164, -0.00000008, 0.00000085, 0.00000149, 0.00000222, 0.00000313, 0.00000393, },
		{ -0.00000009, -0.00000100, -0.00000115, -0.00000082, -0.00000041, -0.00000024, -0.00000030, -0.00000043, -0.00000043, -0.00000026, -0.00000005, -0.00000000, -0.00000025, -0.00000070, -0.00000118, -0.00000151, -0.00000163, -0.00000163, -0.00000159, -0.00000154, -0.00000139, -0.00000107, -0.00000069, -0.00000056, -0.00000098, -0.00000200, -0.00000327, -0.00000403, -0.00000364, -0.00000195, 0.00000037, 0.00000207, 0.00000199, -0.00000020, -0.00000364, -0.00000656, -0.00000720, -0.00000476, 0.00000004, 0.00000531, 0.00000893, 0.00000969, 0.00000784, 0.00000480, 0.00000217, 0.00000074, 0.00000012, -0.00000079, -0.00000280, -0.00000562, -0.00000784, -0.00000781, -0.00000479, 0.00000034, 0.00000542, 0.00000822, 0.00000784, 0.00000532, 0.00000301, 0.00000294, 0.00000538, 0.00000836, 0.00000883, 0.00000459, -0.00000396, -0.00001355, -0.00001963, -0.00001882, -0.00001097, 0.00000043, 0.00001002, 0.00001334, 0.00000921, 0.00000026, -0.00000861, -0.00001297, -0.00001123, -0.00000523, 0.00000110, 0.00000418, 0.00000283, -0.00000131, -0.00000508, -0.00000592, -0.00000349, 0.00000027, 0.00000248, 0.00000129, -0.00000287, -0.00000750, -0.00000955, -0.00000728, -0.00000135, 0.00000550, 0.00001004, 0.00001025, 0.00000632, 0.00000038, -0.00000482, -0.00000742, -0.00000716, -0.00000520, -0.00000307, -0.00000166, -0.00000077, 0.00000045, 0.00000270, 0.00000583, 0.00000873, 0.00000997, 0.00000865, 0.00000499, 0.00000029, -0.00000376, -0.00000589, -0.00000581, -0.00000417, -0.00000208, -0.00000039, 0.00000063, 0.00000129, 0.00000204, 0.00000307, 0.00000414, 0.00000473, 0.00000441, 0.00000318, 0.00000146, },
		{ -0.00000009, 0.00000048, 0.00000066, 0.00000054, 0.00000035, 0.00000029, 0.00000037, 0.00000044, 0.00000037, 0.00000010, -0.00000026, -0.00000056, -0.00000072, -0.00000076, -0.00000076, -0.00000074, -0.00000065, -0.00000043, -0.00000011, 0.00000010, -0.00000010, -0.00000086, -0.00000197, -0.00000284, -0.00000284, -0.00000165, 0.00000035, 0.00000214, 0.00000257, 0.00000104, -0.00000198, -0.00000506, -0.00000645, -0.00000503, -0.00000102, 0.00000406, 0.00000816, 0.00000975, 0.00000858, 0.00000572, 0.00000276, 0.00000084, -0.00000000, -0.00000067, -0.00000219, -0.00000476, -0.00000741, -0.00000850, -0.00000684, -0.00000259, 0.00000260, 0.00000646, 0.00000737, 0.00000542, 0.00000224, -0.00000009, -0.00000045, 0.00000046, 0.00000057, -0.00000203, -0.00000753, -0.00001373, -0.00001701, -0.00001440, -0.00000556, 0.00000657, 0.00001696, 0.00002104, 0.00001715, 0.00000749, -0.00000308, -0.00000963, -0.00000973, -0.00000455, 0.00000210, 0.00000615, 0.00000561, 0.00000152, -0.00000301, -0.00000493, -0.00000318, 0.00000071, 0.00000374, 0.00000344, -0.00000045, -0.00000575, -0.00000912, -0.00000813, -0.00000272, 0.00000469, 0.00001052, 0.00001204, 0.00000873, 0.00000244, -0.00000385, -0.00000764, -0.00000809, -0.00000612, -0.00000349, -0.00000157, -0.00000058, 0.00000027, 0.00000194, 0.00000470, 0.00000776, 0.00000966, 0.00000918, 0.00000615, 0.00000156, -0.00000287, -0.00000562, -0.00000608, -0.00000470, -0.00000255, -0.00000068, 0.00000051, 0.00000123, 0.00000202, 0.00000321, 0.00000460, 0.00000562, 0.00000567, 0.00000454, 0.00000259, 0.00000057, -0.00000082, -0.00000123, -0.00000082, },
		{ 0.00000024, -0.00000012, -0.00000028, -0.00000027, -0.00000025, -0.00000035, -0.00000058, -0.00000084, -0.00000104, -0.00000114, -0.00000118, -0.00000119, -0.00000115, -0.00000099, -0.00000072, -0.00000047, -0.00000049, -0.00000098, -0.00000185, -0.00000269, -0.00000292, -0.00000213, -0.00000047, 0.00000130, 0.00000213, 0.00000127, -0.00000118, -0.00000411, -0.00000593, -0.00000535, -0.00000214, 0.00000260, 0.00000700, 0.00000937, 0.00000904, 0.00000666, 0.00000371, 0.00000151, 0.00000047, -0.00000004, -0.00000107, -0.00000317, -0.00000579, -0.00000749, -0.00000682, -0.00000335, 0.00000180, 0.00000653, 0.00000885, 0.00000810, 0.00000528, 0.00000240, 0.00000110, 0.00000157, 0.00000242, 0.00000173, -0.00000145, -0.00000611, -0.00000955, -0.00000898, -0.00000335, 0.00000554, 0.00001371, 0.00001694, 0.00001320, 0.00000390, -0.00000672, -0.00001382, -0.00001452, -0.00000939, -0.00000193, 0.00000356, 0.00000442, 0.00000102, -0.00000375, -0.00000647, -0.00000538, -0.00000141, 0.00000250, 0.00000339, 0.00000028, -0.00000518, -0.00000959, -0.00000990, -0.00000523, 0.00000247, 0.00000955, 0.00001264, 0.00001042, 0.00000423, -0.00000285, -0.00000779, -0.00000909, -0.00000729, -0.00000422, -0.00000168, -0.00000038, 0.00000026, 0.00000140, 0.00000370, 0.00000677, 0.00000924, 0.00000964, 0.00000735, 0.00000304, -0.00000168, -0.00000506, -0.00000616, -0.00000514, -0.00000304, -0.00000101, 0.00000032, 0.00000107, 0.00000181, 0.00000302, 0.00000463, 0.00000605, 0.00000650, 0.00000555, 0.00000342, 0.00000090, -0.00000109, -0.00000194, -0.00000165, -0.00000071, 0.00000022, 0.00000069, 0.00000061, },
		{ -0.00000072, -0.00000060, -0.00000061, -0.00000074, -0.00000089, -0.00000101, -0.00000106, -0.00000107, -0.00000103, -0.00000091, -0.00000067, -0.00000041, -0.00000030, -0.00000054, -0.00000115, -0.00000188, -0.00000224, -0.00000180, -0.00000053, 0.00000107, 0.00000210, 0.00000176, -0.00000012, -0.00000278, -0.00000486, -0.00000500, -0.00000268, 0.00000146, 0.00000583, 0.00000870, 0.00000910, 0.00000728, 0.00000445, 0.00000199, 0.00000062, 0.00000002, -0.00000077, -0.00000252, -0.00000506, -0.00000723, -0.00000751, -0.00000503, -0.00000036, 0.00000468, 0.00000797, 0.00000828, 0.00000598, 0.00000277, 0.00000054, 0.00000010, 0.00000069, 0.00000053, -0.00000180, -0.00000613, -0.00001044, -0.00001194, -0.00000882, -0.00000167, 0.00000651, 0.00001176, 0.00001133, 0.00000537, -0.00000303, -0.00000952, -0.00001086, -0.00000674, 0.00000016, 0.00000579, 0.00000716, 0.00000400, -0.00000123, -0.00000502, -0.00000503, -0.00000155, 0.00000278, 0.00000467, 0.00000241, -0.00000300, -0.00000836, -0.00001014, -0.00000669, 0.00000077, 0.00000871, 0.00001327, 0.00001233, 0.00000655, -0.00000114, -0.00000729, -0.00000972, -0.00000843, -0.00000516, -0.00000207, -0.00000038, 0.00000015, 0.00000072, 0.00000241, 0.00000525, 0.00000809, 0.00000931, 0.00000785, 0.00000395, -0.00000094, -0.00000496, -0.00000677, -0.00000624, -0.00000425, -0.00000207, -0.00000056, 0.00000024, 0.00000091, 0.00000208, 0.00000386, 0.00000568, 0.00000664, 0.00000605, 0.00000394, 0.00000103, -0.00000155, -0.00000292, -0.00000283, -0.00000171, -0.00000036, 0.00000052, 0.00000066, 0.00000023, -0.00000036, -0.00000076, -0.00000083, },
		{ 0.00000007, -0.00000010, -0.00000023, -0.00000030, -0.00000031, -0.00000026, -0.00000012, 0.00000009, 0.00000023, 0.00000014, -0.00000027, -0.00000086, -0.00000129, -0.00000114, -0.00000026, 0.00000106, 0.00000212, 0.00000216, 0.00000083, -0.00000145, -0.00000356, -0.00000424, -0.00000272, 0.00000073, 0.00000486, 0.00000805, 0.00000913, 0.00000798, 0.00000549, 0.00000299, 0.00000138, 0.00000066, 0.00000005, -0.00000132, -0.00000361, -0.00000595, -0.00000689, -0.00000529, -0.00000121, 0.00000394, 0.00000804, 0.00000949, 0.00000807, 0.00000503, 0.00000229, 0.00000111, 0.00000132, 0.00000148, -0.00000003, -0.00000362, -0.00000794, -0.00001036, -0.00000863, -0.00000244, 0.00000602, 0.00001297, 0.00001502, 0.00001114, 0.00000331, -0.00000450, -0.00000853, -0.00000731, -0.00000235, 0.00000284, 0.00000497, 0.00000293, -0.00000167, -0.00000562, -0.00000626, -0.00000317, 0.00000151, 0.00000447, 0.00000340, -0.00000145, -0.00000728, -0.00001035, -0.00000826, -0.00000140, 0.00000710, 0.00001311, 0.00001374, 0.00000886, 0.00000098, -0.00000619, -0.00000982, -0.00000925, -0.00000594, -0.00000227, -0.00000000, 0.00000066, 0.00000083, 0.00000190, 0.00000439, 0.00000747, 0.00000949, 0.00000900, 0.00000577, 0.00000093, -0.00000358, -0.00000612, -0.00000615, -0.00000433, -0.00000200, -0.00000022, 0.00000072, 0.00000135, 0.00000243, 0.00000429, 0.00000647, 0.00000799, 0.00000791, 0.00000597, 0.00000281, -0.00000036, -0.00000236, -0.00000265, -0.00000151, 0.00000019, 0.00000152, 0.00000195, 0.00000152, 0.00000069, -0.00000001, -0.00000030, -0.00000020, 0.00000005, 0.00000022, 0.00000022, },
		},
		{ { -0.00000087, -0.00000059, -0.00000075, -0.00000184, -0.00000218, -0.00000112, -0.00000042, -0.00000070, -0.00000067, -0.00000063, -0.00000230, -0.00000469, -0.00000510, -0.00000372, -0.00000239, -0.00000067, 0.00000208, 0.00000330, 0.00000110, -0.00000217, -0.00000466, -0.00000800, -0.00001157, -0.00001103, -0.00000653, -0.00000421, -0.00000549, -0.00000474, -0.00000048, 0.00000181, 0.00000088, 0.00000165, 0.00000416, 0.00000324, -0.00000071, -0.00000324, -0.00000587, -0.00001110, -0.00001263, -0.00000579, 0.00000016, -0.00000333, -0.00000786, -0.00000371, 0.00000196, 0.00000077, 0.00000009, 0.00000592, 0.00000834, 0.00000199, -0.00000247, -0.00000298, -0.00001193, -0.00002351, -0.00001409, 0.00000987, 0.00001245, -0.00000676, -0.00000642, 0.00001732, 0.00001510, -0.00002219, -0.00003243, 0.00001111, 0.00004268, 0.00001111, -0.00003243, -0.00002219, 0.00001510, 0.00001732, -0.00000642, -0.00000676, 0.00001245, 0.00000987, -0.00001409, -0.00002351, -0.00001193, -0.00000298, -0.00000247, 0.00000199, 0.00000834, 0.00000592, 0.00000009, 0.00000077, 0.00000196, -0.00000371, -0.00000786, -0.00000333, 0.00000016, -0.00000579, -0.00001263, -0.00001110, -0.00000587, -0.00000324, -0.00000071, 0.00000324, 0.00000416, 0.00000165, 0.00000088, 0.00000181, -0.00000048, -0.00000474, -0.00000549, -0.00000421, -0.00000653, -0.00001103, -0.00001157, -0.00000800, -0.00000466, -0.00000217, 0.00000110, 0.00000330, 0.00000208, -0.00000067, -0.00000239, -0.00000372, -0.00000510, -0.00000469, -0.00000230, -0.00000063, -0.00000067, -0.00000070, -0.00000042, -0.00000112, -0.00000218, -0.00000184, -0.00000075, -0.00000059, },
		{ -0.00000094, -0.00000189, -0.00000224, -0.00000109, -0.00000013, -0.00000056, -0.00000131, -0.00000209, -0.00000390, -0.00000568, -0.00000487, -0.00000192, 0.00000046, 0.00000158, 0.00000195, 0.00000019, -0.00000421, -0.00000801, -0.00000890, -0.00000893, -0.00000924, -0.00000720, -0.00000308, -0.00000190, -0.00000418, -0.00000370, 0.00000147, 0.00000529, 0.00000454, 0.00000272, 0.00000090, -0.00000377, -0.00000916, -0.00001012, -0.00000861, -0.00000920, -0.00000825, -0.00000187, 0.00000230, -0.00000250, -0.00000701, -0.00000109, 0.00000741, 0.00000722, 0.00000301, 0.00000169, -0.00000275, -0.00001202, -0.00001361, -0.00000560, -0.00000413, -0.00000966, -0.00000184, 0.00001551, 0.00001066, -0.00001481, -0.00001720, 0.00001079, 0.00001815, -0.00001277, -0.00002588, 0.00000977, 0.00003643, 0.00000503, -0.00003556, -0.00002028, 0.00002320, 0.00002786, -0.00000131, -0.00001167, 0.00000030, -0.00000313, -0.00002254, -0.00002392, -0.00000341, 0.00001101, 0.00001032, 0.00000837, 0.00000739, 0.00000006, -0.00000684, -0.00000360, 0.00000126, -0.00000308, -0.00000909, -0.00000719, -0.00000442, -0.00000854, -0.00001159, -0.00000560, 0.00000254, 0.00000468, 0.00000375, 0.00000353, 0.00000139, -0.00000229, -0.00000223, 0.00000047, -0.00000108, -0.00000661, -0.00000929, -0.00000855, -0.00000941, -0.00001096, -0.00000800, -0.00000176, 0.00000186, 0.00000201, 0.00000181, 0.00000120, -0.00000126, -0.00000349, -0.00000331, -0.00000254, -0.00000295, -0.00000286, -0.00000134, -0.00000050, -0.00000118, -0.00000139, -0.00000066, -0.00000085, -0.00000186, -0.00000173, -0.00000066, -0.00000046, -0.00000093, -0.00000086, },
		{ -0.00000226, -0.00000132, -0.00000056, -0.00000134, -0.00000251, -0.00000311, -0.00000385, -0.00000430, -0.00000277, -0.00000010, 0.00000084, -0.00000047, -0.00000250, -0.00000513, -0.00000832, -0.00000959, -0.00000773, -0.00000574, -0.00000560, -0.00000470, -0.00000194, -0.00000084, -0.00000202, -0.00000055, 0.00000458, 0.00000690, 0.00000296, -0.00000291, -0.00000737, -0.00001140, -0.00001331, -0.00000990, -0.00000499, -0.00000450, -0.00000501, -0.00000082, 0.00000301, 0.00000020, -0.00000207, 0.00000429, 0.00001054, 0.00000524, -0.00000563, -0.00001145, -0.00001457, -0.00001687, -0.00000926, 0.00000626, 0.00001007, -0.00000050, -0.00000262, 0.00000702, 0.00000137, -0.00002016, -0.00001784, 0.00001492, 0.00002578, -0.00000767, -0.00002994, -0.00000125, 0.00002922, 0.00000878, -0.00002468, -0.00001078, 0.00002674, 0.00002476, -0.00001075, -0.00002474, -0.00001008, -0.00000497, -0.00001509, -0.00001104, 0.00000954, 0.00001918, 0.00001118, 0.00000330, 0.00000056, -0.00000493, -0.00000884, -0.00000347, 0.00000146, -0.00000509, -0.00001364, -0.00001151, -0.00000497, -0.00000409, -0.00000377, 0.00000269, 0.00000840, 0.00000625, 0.00000146, -0.00000023, -0.00000167, -0.00000415, -0.00000353, -0.00000120, -0.00000409, -0.00001105, -0.00001338, -0.00000960, -0.00000618, -0.00000461, -0.00000091, 0.00000366, 0.00000381, 0.00000020, -0.00000210, -0.00000233, -0.00000283, -0.00000285, -0.00000108, 0.00000009, -0.00000128, -0.00000269, -0.00000206, -0.00000129, -0.00000171, -0.00000159, -0.00000043, -0.00000035, -0.00000158, -0.00000179, -0.00000073, -0.00000044, -0.00000107, -0.00000118, -0.00000111, -0.00000184, },
		{ -0.00000137, -0.00000194, -0.00000266, -0.00000242, -0.00000223, -0.00000257, -0.00000228, -0.00000163, -0.00000257, -0.00000493, -0.00000660, -0.00000725, -0.00000772, -0.00000703, -0.00000468, -0.00000323, -0.00000395, -0.00000361, -0.00000041, 0.00000207, 0.00000166, 0.00000171, 0.00000343, 0.00000206, -0.00000438, -0.00001064, -0.00001282, -0.00001262, -0.00001058, -0.00000545, -0.00000107, -0.00000209, -0.00000377, 0.00000079, 0.00000675, 0.00000573, 0.00000197, 0.00000310, 0.00000313, -0.00000624, -0.00001712, -0.00001797, -0.00001239, -0.00000668, 0.00000369, 0.00001599, 0.00001289, -0.00000550, -0.00001269, -0.00000137, 0.00000067, -0.00001431, -0.00001211, 0.00001632, 0.00002357, -0.00001256, -0.00003701, -0.00000548, 0.00003509, 0.00002342, -0.00001356, -0.00001351, 0.00001285, 0.00001019, -0.00001921, -0.00002629, -0.00000512, 0.00000589, -0.00000185, -0.00000113, 0.00001231, 0.00001521, 0.00000404, -0.00000268, -0.00000177, -0.00000449, -0.00000844, -0.00000505, -0.00000203, -0.00000909, -0.00001580, -0.00000940, 0.00000189, 0.00000480, 0.00000322, 0.00000551, 0.00000718, 0.00000263, -0.00000228, -0.00000236, -0.00000249, -0.00000560, -0.00000690, -0.00000585, -0.00000829, -0.00001298, -0.00001151, -0.00000364, 0.00000187, 0.00000210, 0.00000201, 0.00000254, 0.00000001, -0.00000414, -0.00000474, -0.00000213, -0.00000046, 0.00000001, 0.00000101, 0.00000089, -0.00000178, -0.00000393, -0.00000305, -0.00000149, -0.00000147, -0.00000135, -0.00000017, -0.00000004, -0.00000147, -0.00000194, -0.00000088, -0.00000053, -0.00000135, -0.00000163, -0.00000137, -0.00000191, -0.00000257, -0.00000203, },
		{ -0.00000166, -0.00000130, -0.00000157, -0.00000326, -0.00000487, -0.00000551, -0.00000627, -0.00000717, -0.00000680, -0.00000551, -0.00000491, -0.00000453, -0.00000319, -0.00000218, -0.00000228, -0.00000094, 0.00000285, 0.00000480, 0.00000204, -0.00000191, -0.00000384, -0.00000620, -0.00001053, -0.00001294, -0.00001123, -0.00000842, -0.00000601, -0.00000235, 0.00000063, -0.00000026, -0.00000068, 0.00000491, 0.00001045, 0.00000642, -0.00000300, -0.00000743, -0.00000870, -0.00001363, -0.00001662, -0.00000979, 0.00000005, 0.00000420, 0.00000758, 0.00001194, 0.00000484, -0.00001270, -0.00001594, 0.00000099, 0.00000751, -0.00000886, -0.00001443, 0.00000731, 0.00001596, -0.00001320, -0.00003197, 0.00000253, 0.00004465, 0.00002944, -0.00001936, -0.00003149, -0.00000483, 0.00000600, -0.00000977, -0.00001296, 0.00000521, 0.00001303, 0.00000263, -0.00000121, 0.00000700, 0.00000790, -0.00000074, -0.00000317, -0.00000006, -0.00000423, -0.00001143, -0.00000967, -0.00000494, -0.00000801, -0.00001057, -0.00000143, 0.00000985, 0.00000954, 0.00000323, 0.00000219, 0.00000325, -0.00000016, -0.00000373, -0.00000342, -0.00000475, -0.00001005, -0.00001229, -0.00000907, -0.00000706, -0.00000745, -0.00000360, 0.00000404, 0.00000653, 0.00000208, -0.00000187, -0.00000239, -0.00000333, -0.00000473, -0.00000276, 0.00000120, 0.00000237, 0.00000106, 0.00000044, -0.00000031, -0.00000290, -0.00000464, -0.00000307, -0.00000097, -0.00000102, -0.00000128, -0.00000015, 0.00000016, -0.00000138, -0.00000221, -0.00000132, -0.00000095, -0.00000187, -0.00000216, -0.00000165, -0.00000202, -0.00000292, -0.00000260, -0.00000161, -0.00000148, },
		{ -0.00000322, -0.00000600, -0.00000810, -0.00000801, -0.00000693, -0.00000598, -0.00000466, -0.00000333, -0.00000319, -0.00000334, -0.00000194, 0.00000012, 0.00000115, 0.00000220, 0.00000373, 0.00000229, -0.00000383, -0.00000990, -0.00001138, -0.00001031, -0.00000992, -0.00000893, -0.00000627, -0.00000435, -0.00000347, -0.00000071, 0.00000284, 0.00000332, 0.00000289, 0.00000575, 0.00000677, -0.00000159, -0.00001329, -0.00001602, -0.00001090, -0.00000743, -0.00000498, 0.00000145, 0.00000594, 0.00000261, -0.00000008, 0.00000334, 0.00000115, -0.00001000, -0.00001050, 0.00000492, 0.00000786, -0.00001336, -0.00002320, 0.00000013, 0.00001816, -0.00000153, -0.00002020, 0.00000589, 0.00003921, 0.00001845, -0.00003349, -0.00004328, -0.00000491, 0.00002053, 0.00000899, -0.00000327, 0.00000402, 0.00000912, 0.00000115, -0.00000275, 0.00000265, 0.00000315, -0.00000219, -0.00000178, 0.00000035, -0.00000710, -0.00001570, -0.00001116, -0.00000121, -0.00000031, -0.00000234, 0.00000431, 0.00001162, 0.00000747, -0.00000092, -0.00000146, 0.00000137, -0.00000127, -0.00000584, -0.00000720, -0.00000953, -0.00001411, -0.00001313, -0.00000487, 0.00000124, 0.00000135, 0.00000212, 0.00000516, 0.00000357, -0.00000310, -0.00000644, -0.00000393, -0.00000153, -0.00000124, 0.00000049, 0.00000299, 0.00000224, -0.00000053, -0.00000129, -0.00000113, -0.00000286, -0.00000431, -0.00000268, -0.00000055, -0.00000080, -0.00000134, -0.00000017, 0.00000026, -0.00000161, -0.00000301, -0.00000223, -0.00000154, -0.00000219, -0.00000234, -0.00000162, -0.00000191, -0.00000299, -0.00000271, -0.00000127, -0.00000071, -0.00000118, -0.00000176, },
		{ -0.00000912, -0.00000767, -0.00000533, -0.00000400, -0.00000317, -0.00000229, -0.00000187, -0.00000117, 0.00000107, 0.00000331, 0.00000303, 0.00000089, -0.00000149, -0.00000540, -0.00001139, -0.00001505, -0.00001270, -0.00000759, -0.00000465, -0.00000369, -0.00000274, -0.00000228, -0.00000164, 0.00000166, 0.00000552, 0.00000480, 0.00000084, -0.00000095, -0.00000260, -0.00000952, -0.00001625, -0.00001255, -0.00000183, 0.00000365, 0.00000267, 0.00000198, 0.00000023, -0.00000518, -0.00000558, 0.00000309, 0.00000620, -0.00000393, -0.00000933, -0.00000031, 0.00000053, -0.00001809, -0.00002362, 0.00000495, 0.00002824, 0.00000823, -0.00002032, -0.00000700, 0.00002228, 0.00000844, -0.00003324, -0.00003488, 0.00000934, 0.00003567, 0.00001597, -0.00000870, -0.00000609, 0.00000539, 0.00000442, -0.00000105, -0.00000167, -0.00000246, -0.00000364, -0.00000033, 0.00000022, -0.00000933, -0.00001602, -0.00000634, 0.00000738, 0.00000752, 0.00000119, 0.00000327, 0.00000798, 0.00000340, -0.00000393, -0.00000311, -0.00000013, -0.00000440, -0.00001073, -0.00001166, -0.00001090, -0.00001117, -0.00000626, 0.00000405, 0.00000886, 0.00000431, -0.00000029, -0.00000023, -0.00000150, -0.00000563, -0.00000563, -0.00000045, 0.00000257, 0.00000125, 0.00000061, 0.00000153, 0.00000031, -0.00000209, -0.00000183, -0.00000060, -0.00000205, -0.00000393, -0.00000258, -0.00000032, -0.00000060, -0.00000148, -0.00000058, -0.00000032, -0.00000251, -0.00000410, -0.00000288, -0.00000145, -0.00000177, -0.00000206, -0.00000147, -0.00000181, -0.00000302, -0.00000279, -0.00000123, -0.00000087, -0.00000221, -0.00000380, -0.00000563, -0.00000799, },
		{ -0.00000401, -0.00000289, -0.00000211, -0.00000053, 0.00000111, 0.00000228, 0.00000330, 0.00000284, -0.00000088, -0.00000611, -0.00000965, -0.00001170, -0.00001356, -0.00001306, -0.00000811, -0.00000221, -0.00000001, -0.00000068, -0.00000103, -0.00000068, 0.00000028, 0.00000277, 0.00000421, 0.00000019, -0.00000658, -0.00000880, -0.00000694, -0.00000742, -0.00000845, -0.00000288, 0.00000556, 0.00000603, -0.00000065, -0.00000483, -0.00000549, -0.00000636, -0.00000255, 0.00000739, 0.00000870, -0.00000574, -0.00001592, -0.00000754, -0.00000079, -0.00001161, -0.00001453, 0.00001062, 0.00002846, 0.00000234, -0.00003302, -0.00002197, 0.00001566, 0.00001654, -0.00001626, -0.00002030, 0.00001582, 0.00003450, 0.00000893, -0.00001770, -0.00000987, 0.00000997, 0.00001025, -0.00000333, -0.00001090, -0.00000863, -0.00000162, 0.00000474, 0.00000209, -0.00000937, -0.00001264, 0.00000147, 0.00001482, 0.00000977, -0.00000189, -0.00000174, 0.00000402, 0.00000101, -0.00000577, -0.00000556, -0.00000389, -0.00000895, -0.00001368, -0.00001015, -0.00000436, -0.00000215, 0.00000168, 0.00000834, 0.00000836, -0.00000038, -0.00000642, -0.00000414, -0.00000135, -0.00000233, -0.00000137, 0.00000287, 0.00000370, -0.00000003, -0.00000186, -0.00000043, -0.00000048, -0.00000202, -0.00000126, 0.00000022, -0.00000146, -0.00000382, -0.00000258, -0.00000015, -0.00000073, -0.00000237, -0.00000187, -0.00000138, -0.00000315, -0.00000427, -0.00000243, -0.00000059, -0.00000122, -0.00000218, -0.00000187, -0.00000209, -0.00000330, -0.00000328, -0.00000196, -0.00000201, -0.00000390, -0.00000557, -0.00000659, -0.00000778, -0.00000810, -0.00000633, },
		{ 0.00000046, 0.00000274, 0.00000401, 0.00000293, 0.00000024, -0.00000367, -0.00000889, -0.00001314, -0.00001345, -0.00001094, -0.00000885, -0.00000689, -0.00000300, 0.00000098, 0.00000181, 0.00000053, 0.00000002, 0.00000008, -0.00000031, -0.00000051, -0.00000170, -0.00000630, -0.00001091, -0.00000875, -0.00000170, 0.00000139, -0.00000028, 0.00000044, 0.00000317, 0.00000051, -0.00000581, -0.00000696, -0.00000357, -0.00000219, -0.00000019, 0.00000529, 0.00000340, -0.00001094, -0.00001824, -0.00000476, 0.00000766, -0.00000205, -0.00001212, 0.00000225, 0.00001466, -0.00000900, -0.00003843, -0.00002033, 0.00002608, 0.00003335, -0.00000285, -0.00001911, 0.00000677, 0.00002461, 0.00000440, -0.00001664, -0.00000530, 0.00001415, 0.00000830, -0.00001256, -0.00001931, -0.00000714, 0.00000788, 0.00001198, 0.00000243, -0.00000996, -0.00000777, 0.00000920, 0.00001790, 0.00000559, -0.00000895, -0.00000657, 0.00000208, -0.00000060, -0.00000885, -0.00000951, -0.00000690, -0.00000910, -0.00000944, -0.00000171, 0.00000519, 0.00000404, 0.00000210, 0.00000414, 0.00000246, -0.00000517, -0.00000799, -0.00000171, 0.00000352, 0.00000197, 0.00000035, 0.00000206, 0.00000143, -0.00000256, -0.00000349, -0.00000056, 0.00000031, -0.00000131, -0.00000084, 0.00000069, -0.00000103, -0.00000379, -0.00000293, -0.00000076, -0.00000185, -0.00000394, -0.00000304, -0.00000141, -0.00000231, -0.00000331, -0.00000173, -0.00000036, -0.00000169, -0.00000317, -0.00000269, -0.00000243, -0.00000364, -0.00000408, -0.00000310, -0.00000316, -0.00000479, -0.00000581, -0.00000588, -0.00000649, -0.00000702, -0.00000565, -0.00000315, -0.00000127, },
		{ 0.00000223, -0.00000251, -0.00000786, -0.00001169, -0.00001379, -0.00001343, -0.00000957, -0.00000455, -0.00000231, -0.00000243, -0.00000137, 0.00000099, 0.00000183, 0.00000087, -0.00000021, -0.00000175, -0.00000421, -0.00000577, -0.00000610, -0.00000719, -0.00000730, -0.00000209, 0.00000533, 0.00000600, -0.00000020, -0.00000389, -0.00000242, -0.00000236, -0.00000462, -0.00000357, -0.00000093, -0.00000276, -0.00000483, -0.00000106, 0.00000030, -0.00000772, -0.00001059, 0.00000294, 0.00001178, -0.00000466, -0.00002338, -0.00001364, 0.00000430, -0.00000672, -0.00002661, -0.00000716, 0.00003607, 0.00003873, -0.00000464, -0.00002841, -0.00000331, 0.00002213, 0.00000980, -0.00001070, -0.00000479, 0.00000884, 0.00000022, -0.00001818, -0.00001672, 0.00000373, 0.00001828, 0.00001348, -0.00000195, -0.00000987, -0.00000031, 0.00001534, 0.00001452, -0.00000455, -0.00001690, -0.00000848, 0.00000238, -0.00000254, -0.00001251, -0.00001111, -0.00000403, -0.00000221, -0.00000064, 0.00000613, 0.00000875, 0.00000154, -0.00000456, -0.00000218, -0.00000006, -0.00000350, -0.00000368, 0.00000315, 0.00000671, 0.00000203, -0.00000216, -0.00000093, -0.00000049, -0.00000320, -0.00000313, 0.00000043, 0.00000124, -0.00000098, -0.00000077, 0.00000100, -0.00000087, -0.00000451, -0.00000438, -0.00000226, -0.00000297, -0.00000445, -0.00000251, 0.00000009, -0.00000092, -0.00000303, -0.00000250, -0.00000153, -0.00000286, -0.00000406, -0.00000301, -0.00000238, -0.00000393, -0.00000496, -0.00000398, -0.00000344, -0.00000444, -0.00000503, -0.00000486, -0.00000557, -0.00000635, -0.00000465, -0.00000096, 0.00000204, 0.00000361, 0.00000408, },
		{ -0.00001423, -0.00001362, -0.00001051, -0.00000666, -0.00000232, 0.00000084, 0.00000029, -0.00000245, -0.00000287, -0.00000036, 0.00000134, 0.00000034, -0.00000208, -0.00000504, -0.00000763, -0.00000751, -0.00000462, -0.00000230, -0.00000102, 0.00000243, 0.00000621, 0.00000368, -0.00000403, -0.00000721, -0.00000323, -0.00000018, -0.00000178, -0.00000334, -0.00000432, -0.00000785, -0.00000860, -0.00000060, 0.00000640, 0.00000145, -0.00000471, 0.00000075, 0.00000256, -0.00001646, -0.00003378, -0.00001842, 0.00000968, 0.00000796, -0.00001195, -0.00000203, 0.00003162, 0.00003154, -0.00001022, -0.00003197, -0.00000354, 0.00002639, 0.00001457, -0.00001185, -0.00001245, 0.00000114, -0.00000110, -0.00001175, -0.00000569, 0.00001306, 0.00001976, 0.00000783, -0.00000556, -0.00000445, 0.00000833, 0.00001479, 0.00000219, -0.00001771, -0.00002046, -0.00000445, 0.00000493, -0.00000481, -0.00001439, -0.00000712, 0.00000502, 0.00000681, 0.00000434, 0.00000594, 0.00000408, -0.00000538, -0.00000986, -0.00000253, 0.00000436, 0.00000230, 0.00000023, 0.00000404, 0.00000499, -0.00000124, -0.00000519, -0.00000215, 0.00000008, -0.00000210, -0.00000230, 0.00000107, 0.00000166, -0.00000101, -0.00000107, 0.00000063, -0.00000176, -0.00000622, -0.00000610, -0.00000280, -0.00000220, -0.00000307, -0.00000112, 0.00000096, -0.00000126, -0.00000468, -0.00000442, -0.00000263, -0.00000307, -0.00000389, -0.00000277, -0.00000232, -0.00000433, -0.00000560, -0.00000415, -0.00000286, -0.00000362, -0.00000440, -0.00000421, -0.00000450, -0.00000457, -0.00000193, 0.00000233, 0.00000433, 0.00000287, -0.00000057, -0.00000538, -0.00001086, },
		{ -0.00000224, 0.00000023, 0.00000090, 0.00000050, -0.00000183, -0.00000468, -0.00000439, -0.00000100, 0.00000106, -0.00000059, -0.00000394, -0.00000664, -0.00000738, -0.00000486, -0.00000033, 0.00000199, 0.00000120, 0.00000139, 0.00000302, 0.00000095, -0.00000471, -0.00000625, -0.00000196, 0.00000005, -0.00000378, -0.00000744, -0.00000792, -0.00000771, -0.00000397, 0.00000639, 0.00001244, 0.00000283, -0.00001083, -0.00001131, -0.00000854, -0.00002012, -0.00002958, -0.00000984, 0.00001994, 0.00001680, -0.00000999, -0.00000947, 0.00002102, 0.00002709, -0.00000685, -0.00002690, -0.00000157, 0.00002529, 0.00001103, -0.00001771, -0.00001729, 0.00000276, 0.00000800, -0.00000058, 0.00000096, 0.00001248, 0.00001427, 0.00000377, -0.00000245, 0.00000338, 0.00000984, 0.00000298, -0.00001480, -0.00002459, -0.00001330, 0.00000539, 0.00000658, -0.00000855, -0.00001319, 0.00000225, 0.00001541, 0.00001022, 0.00000003, -0.00000140, -0.00000184, -0.00000770, -0.00000774, 0.00000320, 0.00001042, 0.00000495, -0.00000141, 0.00000030, 0.00000161, -0.00000318, -0.00000541, -0.00000085, 0.00000180, -0.00000121, -0.00000225, 0.00000106, 0.00000166, -0.00000162, -0.00000235, -0.00000078, -0.00000304, -0.00000717, -0.00000598, -0.00000120, -0.00000014, -0.00000207, -0.00000164, -0.00000057, -0.00000323, -0.00000645, -0.00000514, -0.00000205, -0.00000207, -0.00000343, -0.00000289, -0.00000258, -0.00000454, -0.00000565, -0.00000378, -0.00000219, -0.00000308, -0.00000381, -0.00000273, -0.00000172, -0.00000121, 0.00000064, 0.00000261, 0.00000100, -0.00000425, -0.00000953, -0.00001282, -0.00001401, -0.00001216, -0.00000722, },
		{ -0.00000230, -0.00000341, -0.00000406, -0.00000481, -0.00000360, -0.00000047, 0.00000088, -0.00000131, -0.00000443, -0.00000576, -0.00000480, -0.00000168, 0.00000190, 0.00000241, -0.00000014, -0.00000088, 0.00000140, 0.00000120, -0.00000339, -0.00000618, -0.00000430, -0.00000357, -0.00000661, -0.00000763, -0.00000428, -0.00000099, 0.00000219, 0.00000777, 0.00000776, -0.00000601, -0.00002063, -0.00001779, -0.00000720, -0.00001064, -0.00001827, -0.00000403, 0.00001902, 0.00001248, -0.00001582, -0.00001585, 0.00001805, 0.00003120, 0.00000065, -0.00002443, -0.00000692, 0.00001767, 0.00000719, -0.00001652, -0.00001195, 0.00001135, 0.00001682, 0.00000367, -0.00000114, 0.00000745, 0.00001151, 0.00000585, 0.00000213, 0.00000355, -0.00000020, -0.00001319, -0.00002341, -0.00001685, 0.00000254, 0.00001306, 0.00000194, -0.00001335, -0.00000715, 0.00001450, 0.00002064, 0.00000347, -0.00001116, -0.00000732, -0.00000003, -0.00000173, -0.00000148, 0.00000743, 0.00001092, 0.00000142, -0.00000654, -0.00000298, 0.00000139, -0.00000167, -0.00000366, 0.00000061, 0.00000258, -0.00000145, -0.00000300, 0.00000049, 0.00000096, -0.00000315, -0.00000421, -0.00000170, -0.00000257, -0.00000582, -0.00000423, 0.00000037, 0.00000011, -0.00000379, -0.00000437, -0.00000261, -0.00000389, -0.00000602, -0.00000401, -0.00000076, -0.00000146, -0.00000379, -0.00000345, -0.00000258, -0.00000417, -0.00000537, -0.00000355, -0.00000168, -0.00000200, -0.00000171, 0.00000061, 0.00000178, 0.00000021, -0.00000183, -0.00000381, -0.00000785, -0.00001282, -0.00001451, -0.00001192, -0.00000767, -0.00000329, 0.00000069, 0.00000223, 0.00000035, },
		{ -0.00000192, -0.00000170, -0.00000184, -0.00000067, -0.00000017, -0.00000188, -0.00000376, -0.00000384, -0.00000247, -0.00000031, 0.00000164, 0.00000121, -0.00000126, -0.00000154, 0.00000117, 0.00000104, -0.00000451, -0.00000901, -0.00000758, -0.00000444, -0.00000344, -0.00000160, 0.00000180, 0.00000197, -0.00000055, -0.00000060, -0.00000210, -0.00001192, -0.00002007, -0.00001163, 0.00000234, -0.00000166, -0.00001562, -0.00001033, 0.00000882, 0.00000638, -0.00001566, -0.00001311, 0.00002088, 0.00003423, 0.00000180, -0.00002863, -0.00001447, 0.00001432, 0.00001220, -0.00000695, -0.00000388, 0.00001559, 0.00001757, 0.00000160, -0.00000426, 0.00000604, 0.00001260, 0.00000672, -0.00000113, -0.00000556, -0.00001189, -0.00001897, -0.00001592, 0.00000013, 0.00001437, 0.00001001, -0.00000784, -0.00001366, 0.00000479, 0.00002413, 0.00001534, -0.00001140, -0.00001982, -0.00000248, 0.00001146, 0.00000615, -0.00000067, 0.00000371, 0.00000570, -0.00000332, -0.00000876, -0.00000163, 0.00000467, 0.00000071, -0.00000314, 0.00000028, 0.00000208, -0.00000235, -0.00000424, -0.00000063, -0.00000017, -0.00000443, -0.00000473, -0.00000041, -0.00000013, -0.00000404, -0.00000410, -0.00000098, -0.00000230, -0.00000654, -0.00000599, -0.00000209, -0.00000197, -0.00000428, -0.00000316, -0.00000064, -0.00000186, -0.00000437, -0.00000355, -0.00000205, -0.00000369, -0.00000531, -0.00000314, 0.00000004, 0.00000107, 0.00000182, 0.00000308, 0.00000114, -0.00000505, -0.00001090, -0.00001345, -0.00001424, -0.00001351, -0.00000913, -0.00000243, 0.00000196, 0.00000250, 0.00000105, -0.00000153, -0.00000480, -0.00000618, -0.00000425, },
		{ -0.00000286, -0.00000328, -0.00000194, -0.00000159, -0.00000190, -0.00000160, -0.00000097, -0.00000026, 0.00000056, 0.00000006, -0.00000200, -0.00000251, -0.00000059, -0.00000112, -0.00000629, -0.00000967, -0.00000599, 0.00000013, 0.00000254, 0.00000221, 0.00000070, -0.00000399, -0.00000909, -0.00000799, -0.00000407, -0.00000663, -0.00001066, -0.00000376, 0.00000536, -0.00000380, -0.00002250, -0.00001980, 0.00000251, 0.00000807, -0.00000877, -0.00000892, 0.00001906, 0.00003002, -0.00000121, -0.00003040, -0.00001424, 0.00001945, 0.00002169, 0.00000094, -0.00000227, 0.00001271, 0.00001487, 0.00000098, -0.00000394, 0.00000558, 0.00000971, 0.00000044, -0.00000947, -0.00001197, -0.00001185, -0.00001023, -0.00000157, 0.00001093, 0.00001310, -0.00000040, -0.00001324, -0.00000526, 0.00001734, 0.00002361, -0.00000007, -0.00002442, -0.00001541, 0.00001432, 0.00002333, 0.00000399, -0.00001096, -0.00000456, 0.00000320, -0.00000195, -0.00000562, 0.00000204, 0.00000697, 0.00000028, -0.00000515, -0.00000112, 0.00000159, -0.00000316, -0.00000565, -0.00000167, -0.00000032, -0.00000386, -0.00000330, 0.00000176, 0.00000127, -0.00000504, -0.00000709, -0.00000399, -0.00000395, -0.00000658, -0.00000447, 0.00000043, 0.00000007, -0.00000389, -0.00000395, -0.00000131, -0.00000199, -0.00000426, -0.00000330, -0.00000167, -0.00000328, -0.00000445, -0.00000085, 0.00000376, 0.00000438, 0.00000223, -0.00000082, -0.00000687, -0.00001517, -0.00001932, -0.00001613, -0.00000969, -0.00000381, 0.00000156, 0.00000512, 0.00000369, -0.00000165, -0.00000597, -0.00000717, -0.00000637, -0.00000368, 0.00000050, 0.00000262, 0.00000041, },
		{ -0.00000165, 0.00000107, 0.00000113, 0.00000023, -0.00000082, -0.00000155, -0.00000149, -0.00000190, -0.00000369, -0.00000438, -0.00000238, -0.00000143, -0.00000415, -0.00000567, -0.00000156, 0.00000346, 0.00000293, -0.00000170, -0.00000636, -0.00001056, -0.00001181, -0.00000565, 0.00000262, 0.00000164, -0.00000555, -0.00000526, -0.00000117, -0.00001063, -0.00002701, -0.00002205, 0.00000316, 0.00001208, -0.00000507, -0.00001086, 0.00001228, 0.00002567, 0.00000143, -0.00002360, -0.00000768, 0.00002496, 0.00002607, 0.00000196, -0.00000531, 0.00000922, 0.00001442, 0.00000212, -0.00000536, -0.00000012, 0.00000175, -0.00000602, -0.00001082, -0.00000654, -0.00000136, 0.00000074, 0.00000462, 0.00000903, 0.00000507, -0.00000616, -0.00000877, 0.00000589, 0.00002082, 0.00001176, -0.00001517, -0.00002355, 0.00000315, 0.00003113, 0.00002100, -0.00001314, -0.00002473, -0.00000476, 0.00001134, 0.00000489, -0.00000341, 0.00000150, 0.00000515, -0.00000237, -0.00000721, -0.00000138, 0.00000207, -0.00000372, -0.00000682, -0.00000159, 0.00000133, -0.00000205, -0.00000235, 0.00000157, -0.00000050, -0.00000831, -0.00001005, -0.00000449, -0.00000196, -0.00000381, -0.00000226, 0.00000165, 0.00000004, -0.00000497, -0.00000492, -0.00000125, -0.00000141, -0.00000410, -0.00000335, -0.00000092, -0.00000124, -0.00000128, 0.00000263, 0.00000562, 0.00000206, -0.00000517, -0.00001117, -0.00001617, -0.00001963, -0.00001670, -0.00000687, 0.00000246, 0.00000567, 0.00000444, 0.00000116, -0.00000427, -0.00000939, -0.00000954, -0.00000478, 0.00000028, 0.00000333, 0.00000463, 0.00000283, -0.00000251, -0.00000692, -0.00000594, },
		{ 0.00000281, -0.00000054, -0.00000317, -0.00000451, -0.00000431, -0.00000390, -0.00000455, -0.00000420, -0.00000091, 0.00000191, 0.00000044, -0.00000219, -0.00000131, 0.00000030, -0.00000234, -0.00000719, -0.00000951, -0.00000920, -0.00000639, 0.00000069, 0.00000676, 0.00000203, -0.00000940, -0.00001208, -0.00000698, -0.00001167, -0.00002438, -0.00002033, 0.00000206, 0.00001009, -0.00000711, -0.00001410, 0.00000966, 0.00002865, 0.00001076, -0.00001416, -0.00000387, 0.00002418, 0.00002463, 0.00000079, -0.00000707, 0.00000755, 0.00001339, -0.00000035, -0.00001124, -0.00000733, -0.00000194, -0.00000307, -0.00000209, 0.00000403, 0.00000626, 0.00000238, 0.00000083, 0.00000340, 0.00000220, -0.00000311, -0.00000157, 0.00000935, 0.00001369, -0.00000066, -0.00001720, -0.00000801, 0.00002186, 0.00003132, 0.00000051, -0.00003320, -0.00002497, 0.00001119, 0.00002492, 0.00000547, -0.00001064, -0.00000419, 0.00000356, -0.00000227, -0.00000676, -0.00000044, 0.00000261, -0.00000440, -0.00000734, -0.00000027, 0.00000346, -0.00000169, -0.00000444, -0.00000143, -0.00000302, -0.00000965, -0.00000931, -0.00000144, 0.00000166, -0.00000195, -0.00000251, 0.00000066, -0.00000074, -0.00000545, -0.00000493, -0.00000079, -0.00000108, -0.00000401, -0.00000229, 0.00000215, 0.00000285, 0.00000132, 0.00000151, -0.00000031, -0.00000799, -0.00001599, -0.00001770, -0.00001483, -0.00001079, -0.00000412, 0.00000431, 0.00000753, 0.00000249, -0.00000484, -0.00000875, -0.00000955, -0.00000782, -0.00000222, 0.00000476, 0.00000751, 0.00000508, 0.00000112, -0.00000287, -0.00000710, -0.00000855, -0.00000434, 0.00000211, 0.00000479, },
		{ -0.00000654, -0.00000673, -0.00000528, -0.00000339, -0.00000249, -0.00000118, 0.00000234, 0.00000484, 0.00000191, -0.00000373, -0.00000593, -0.00000498, -0.00000533, -0.00000655, -0.00000558, -0.00000357, -0.00000197, 0.00000102, 0.00000283, -0.00000347, -0.00001383, -0.00001426, -0.00000610, -0.00000814, -0.00002183, -0.00002305, -0.00000494, 0.00000478, -0.00000711, -0.00001066, 0.00001438, 0.00003538, 0.00001874, -0.00000964, -0.00000560, 0.00001988, 0.00002274, 0.00000104, -0.00000832, 0.00000363, 0.00000843, -0.00000503, -0.00001465, -0.00000716, 0.00000367, 0.00000634, 0.00000658, 0.00000808, 0.00000488, -0.00000206, -0.00000280, 0.00000331, 0.00000546, 0.00000076, -0.00000069, 0.00000499, 0.00000664, -0.00000206, -0.00000672, 0.00000709, 0.00002381, 0.00001270, -0.00002222, -0.00003578, -0.00000439, 0.00003261, 0.00002622, -0.00001010, -0.00002407, -0.00000479, 0.00001010, 0.00000188, -0.00000687, -0.00000151, 0.00000216, -0.00000465, -0.00000696, 0.00000107, 0.00000382, -0.00000424, -0.00000848, -0.00000362, -0.00000206, -0.00000678, -0.00000596, 0.00000150, 0.00000295, -0.00000297, -0.00000445, -0.00000018, -0.00000033, -0.00000503, -0.00000498, -0.00000064, -0.00000013, -0.00000189, 0.00000123, 0.00000589, 0.00000371, -0.00000337, -0.00000827, -0.00001167, -0.00001633, -0.00001726, -0.00001019, -0.00000116, 0.00000265, 0.00000279, 0.00000211, -0.00000196, -0.00000941, -0.00001305, -0.00000876, -0.00000137, 0.00000408, 0.00000765, 0.00000876, 0.00000479, -0.00000222, -0.00000643, -0.00000660, -0.00000583, -0.00000417, 0.00000037, 0.00000507, 0.00000494, 0.00000029, -0.00000432, },
		{ -0.00000326, -0.00000080, 0.00000050, 0.00000126, 0.00000319, 0.00000394, -0.00000047, -0.00000726, -0.00000931, -0.00000591, -0.00000265, -0.00000181, -0.00000162, -0.00000238, -0.00000433, -0.00000427, -0.00000269, -0.00000583, -0.00001232, -0.00001131, -0.00000410, -0.00000798, -0.00002467, -0.00002950, -0.00001157, 0.00000405, -0.00000045, -0.00000197, 0.00001995, 0.00003825, 0.00002092, -0.00000935, -0.00000853, 0.00001626, 0.00002132, 0.00000009, -0.00001228, -0.00000259, 0.00000435, -0.00000428, -0.00000974, 0.00000010, 0.00001130, 0.00001162, 0.00000745, 0.00000521, 0.00000132, -0.00000345, -0.00000127, 0.00000619, 0.00000705, -0.00000044, -0.00000381, 0.00000257, 0.00000799, 0.00000454, 0.00000143, 0.00000732, 0.00000990, -0.00000629, -0.00002604, -0.00001606, 0.00002091, 0.00003682, 0.00000536, -0.00003275, -0.00002633, 0.00001016, 0.00002260, 0.00000127, -0.00001402, -0.00000510, 0.00000351, -0.00000266, -0.00000653, 0.00000026, 0.00000158, -0.00000771, -0.00001046, -0.00000149, 0.00000294, -0.00000265, -0.00000452, 0.00000103, 0.00000155, -0.00000470, -0.00000540, 0.00000037, 0.00000070, -0.00000488, -0.00000502, 0.00000096, 0.00000340, 0.00000187, 0.00000292, 0.00000334, -0.00000435, -0.00001529, -0.00001881, -0.00001513, -0.00001042, -0.00000415, 0.00000479, 0.00000899, 0.00000285, -0.00000701, -0.00001202, -0.00001274, -0.00001156, -0.00000562, 0.00000492, 0.00001218, 0.00001112, 0.00000571, 0.00000036, -0.00000498, -0.00000874, -0.00000714, -0.00000183, 0.00000112, 0.00000072, 0.00000102, 0.00000215, 0.00000040, -0.00000401, -0.00000695, -0.00000695, -0.00000549, },
		{ 0.00000134, 0.00000100, 0.00000141, 0.00000147, -0.00000226, -0.00000781, -0.00000848, -0.00000350, 0.00000052, -0.00000014, -0.00000310, -0.00000634, -0.00000876, -0.00000741, -0.00000318, -0.00000329, -0.00000882, -0.00001036, -0.00000682, -0.00001217, -0.00002832, -0.00003235, -0.00001218, 0.00000845, 0.00000804, 0.00000474, 0.00002089, 0.00003619, 0.00002024, -0.00000919, -0.00001055, 0.00001229, 0.00001777, -0.00000301, -0.00001597, -0.00000487, 0.00000705, 0.00000336, -0.00000133, 0.00000545, 0.00001268, 0.00000999, 0.00000456, 0.00000331, 0.00000187, -0.00000118, 0.00000053, 0.00000595, 0.00000511, -0.00000214, -0.00000322, 0.00000579, 0.00001196, 0.00000622, -0.00000211, -0.00000235, -0.00000102, -0.00000742, -0.00001050, 0.00000582, 0.00002647, 0.00001632, -0.00002229, -0.00003884, -0.00000593, 0.00003296, 0.00002465, -0.00001401, -0.00002608, -0.00000330, 0.00001145, 0.00000031, -0.00000981, -0.00000404, -0.00000065, -0.00000767, -0.00000763, 0.00000392, 0.00000770, -0.00000192, -0.00000715, -0.00000157, 0.00000065, -0.00000464, -0.00000502, 0.00000116, 0.00000158, -0.00000399, -0.00000296, 0.00000475, 0.00000658, 0.00000072, -0.00000410, -0.00000776, -0.00001556, -0.00002154, -0.00001577, -0.00000262, 0.00000604, 0.00000821, 0.00000736, 0.00000088, -0.00001208, -0.00002121, -0.00001758, -0.00000628, 0.00000353, 0.00001066, 0.00001531, 0.00001279, 0.00000257, -0.00000689, -0.00000939, -0.00000753, -0.00000439, 0.00000071, 0.00000532, 0.00000411, -0.00000171, -0.00000500, -0.00000385, -0.00000298, -0.00000433, -0.00000499, -0.00000387, -0.00000257, -0.00000123, 0.00000053, },
		{ 0.00000009, 0.00000075, -0.00000141, -0.00000598, -0.00000706, -0.00000319, -0.00000050, -0.00000276, -0.00000676, -0.00000909, -0.00000909, -0.00000583, -0.00000112, -0.00000187, -0.00000930, -0.00001368, -0.00001135, -0.00001429, -0.00002672, -0.00002944, -0.00000975, 0.00001163, 0.00001227, 0.00000676, 0.00001865, 0.00003286, 0.00001955, -0.00000890, -0.00001334, 0.00000679, 0.00001372, -0.00000353, -0.00001394, -0.00000067, 0.00001381, 0.00001064, 0.00000261, 0.00000465, 0.00000954, 0.00000767, 0.00000436, 0.00000487, 0.00000384, -0.00000045, -0.00000062, 0.00000425, 0.00000501, 0.00000019, 0.00000004, 0.00000709, 0.00000924, 0.00000012, -0.00000849, -0.00000519, 0.00000295, 0.00000410, 0.00000272, 0.00000807, 0.00000977, -0.00000828, -0.00003004, -0.00001858, 0.00002284, 0.00003994, 0.00000423, -0.00003699, -0.00002774, 0.00001236, 0.00002266, -0.00000377, -0.00002022, -0.00000821, 0.00000317, -0.00000144, -0.00000263, 0.00000695, 0.00000765, -0.00000539, -0.00001107, -0.00000250, 0.00000233, -0.00000320, -0.00000465, 0.00000202, 0.00000383, -0.00000098, -0.00000028, 0.00000527, 0.00000215, -0.00000962, -0.00001660, -0.00001585, -0.00001455, -0.00001098, 0.00000062, 0.00001252, 0.00001206, 0.00000146, -0.00000844, -0.00001538, -0.00002052, -0.00001748, -0.00000270, 0.00001309, 0.00001801, 0.00001375, 0.00000713, -0.00000123, -0.00001023, -0.00001278, -0.00000631, 0.00000166, 0.00000494, 0.00000528, 0.00000387, -0.00000171, -0.00000891, -0.00001029, -0.00000502, -0.00000041, -0.00000030, -0.00000150, -0.00000166, -0.00000172, -0.00000149, -0.00000011, 0.00000076, 0.00000012, },
		{ -0.00000045, -0.00000552, -0.00000850, -0.00000620, -0.00000352, -0.00000475, -0.00000723, -0.00000769, -0.00000656, -0.00000432, -0.00000213, -0.00000467, -0.00001239, -0.00001607, -0.00001166, -0.00001098, -0.00002114, -0.00002590, -0.00001038, 0.00000972, 0.00001208, 0.00000680, 0.00001638, 0.00002972, 0.00001824, -0.00000957, -0.00001621, 0.00000316, 0.00001385, 0.00000151, -0.00000734, 0.00000460, 0.00001747, 0.00001256, 0.00000193, 0.00000215, 0.00000781, 0.00000816, 0.00000601, 0.00000592, 0.00000367, -0.00000135, -0.00000111, 0.00000531, 0.00000751, 0.00000206, -0.00000127, 0.00000209, 0.00000315, -0.00000282, -0.00000556, 0.00000252, 0.00001127, 0.00000821, -0.00000129, -0.00000498, -0.00000580, -0.00001220, -0.00001314, 0.00000772, 0.00003213, 0.00001978, -0.00002549, -0.00004474, -0.00000761, 0.00003459, 0.00002246, -0.00002148, -0.00003191, -0.00000228, 0.00001691, 0.00000619, -0.00000362, 0.00000275, 0.00000376, -0.00000835, -0.00001179, -0.00000015, 0.00000548, -0.00000194, -0.00000433, 0.00000409, 0.00000729, 0.00000037, -0.00000373, -0.00000315, -0.00000935, -0.00001984, -0.00001929, -0.00000728, 0.00000230, 0.00000588, 0.00000931, 0.00000879, -0.00000290, -0.00001759, -0.00002108, -0.00001405, -0.00000565, 0.00000361, 0.00001513, 0.00002015, 0.00001165, -0.00000201, -0.00000916, -0.00000997, -0.00000873, -0.00000370, 0.00000486, 0.00000904, 0.00000494, -0.00000122, -0.00000459, -0.00000768, -0.00001060, -0.00000811, -0.00000060, 0.00000395, 0.00000206, -0.00000128, -0.00000232, -0.00000234, -0.00000193, -0.00000026, 0.00000097, 0.00000029, -0.00000001, 0.00000102, },
		{ -0.00001128, -0.00000891, -0.00000427, -0.00000328, -0.00000486, -0.00000576, -0.00000618, -0.00000644, -0.00000609, -0.00000780, -0.00001247, -0.00001340, -0.00000745, -0.00000598, -0.00001726, -0.00002641, -0.00001562, 0.00000428, 0.00001041, 0.00000706, 0.00001460, 0.00002628, 0.00001653, -0.00000905, -0.00001551, 0.00000442, 0.00001816, 0.00000826, -0.00000225, 0.00000577, 0.00001657, 0.00001187, 0.00000161, 0.00000186, 0.00000799, 0.00000880, 0.00000619, 0.00000543, 0.00000355, -0.00000028, 0.00000070, 0.00000659, 0.00000691, -0.00000119, -0.00000629, -0.00000175, 0.00000383, 0.00000320, 0.00000257, 0.00000722, 0.00000873, -0.00000060, -0.00001148, -0.00001075, -0.00000238, 0.00000175, 0.00000430, 0.00001310, 0.00001503, -0.00000821, -0.00003656, -0.00002548, 0.00002169, 0.00004148, 0.00000133, -0.00004332, -0.00002889, 0.00001979, 0.00003255, 0.00000253, -0.00001662, -0.00000497, 0.00000417, -0.00000506, -0.00000869, 0.00000290, 0.00000773, -0.00000133, -0.00000380, 0.00000564, 0.00000693, -0.00000590, -0.00001519, -0.00001415, -0.00001390, -0.00001468, -0.00000477, 0.00001129, 0.00001532, 0.00000533, -0.00000492, -0.00001146, -0.00001878, -0.00002078, -0.00000863, 0.00000908, 0.00001683, 0.00001451, 0.00001060, 0.00000423, -0.00000727, -0.00001503, -0.00001058, -0.00000058, 0.00000465, 0.00000582, 0.00000645, 0.00000323, -0.00000464, -0.00000945, -0.00000736, -0.00000462, -0.00000494, -0.00000344, 0.00000155, 0.00000374, 0.00000052, -0.00000256, -0.00000210, -0.00000097, -0.00000070, 0.00000025, 0.00000106, -0.00000002, -0.00000085, 0.00000004, -0.00000134, -0.00000708, },
		{ -0.00000255, -0.00000083, -0.00000378, -0.00000687, -0.00000884, -0.00000959, -0.00000815, -0.00000673, -0.00000797, -0.00000795, -0.00000368, -0.00000436, -0.00001793, -0.00003068, -0.00002302, -0.00000198, 0.00000872, 0.00000773, 0.00001344, 0.00002389, 0.00001703, -0.00000476, -0.00001060, 0.00000816, 0.00002187, 0.00001189, -0.00000101, 0.00000418, 0.00001497, 0.00001234, 0.00000301, 0.00000226, 0.00000735, 0.00000810, 0.00000605, 0.00000623, 0.00000538, 0.00000146, 0.00000047, 0.00000367, 0.00000261, -0.00000451, -0.00000650, 0.00000203, 0.00001032, 0.00000874, 0.00000294, 0.00000092, -0.00000092, -0.00000706, -0.00000979, -0.00000141, 0.00000935, 0.00000918, 0.00000163, -0.00000172, -0.00000465, -0.00001591, -0.00002062, 0.00000262, 0.00003322, 0.00002264, -0.00002665, -0.00004690, -0.00000300, 0.00004571, 0.00003114, -0.00002022, -0.00003370, -0.00000236, 0.00001669, 0.00000278, -0.00000793, 0.00000221, 0.00000814, -0.00000137, -0.00000531, 0.00000211, -0.00000045, -0.00001674, -0.00002345, -0.00001237, -0.00000078, 0.00000372, 0.00001014, 0.00001527, 0.00000492, -0.00001561, -0.00002516, -0.00001891, -0.00000880, 0.00000194, 0.00001654, 0.00002548, 0.00001682, -0.00000128, -0.00001164, -0.00001258, -0.00001163, -0.00000718, 0.00000389, 0.00001236, 0.00000909, 0.00000023, -0.00000453, -0.00000645, -0.00000876, -0.00000724, -0.00000084, 0.00000221, -0.00000172, -0.00000469, -0.00000166, 0.00000157, 0.00000065, -0.00000037, 0.00000104, 0.00000135, -0.00000058, -0.00000138, -0.00000088, -0.00000173, -0.00000270, -0.00000183, -0.00000275, -0.00000822, -0.00001234, -0.00000900, },
		{ -0.00000557, -0.00000971, -0.00001067, -0.00000930, -0.00000568, -0.00000226, -0.00000281, -0.00000483, -0.00000418, -0.00000716, -0.00002151, -0.00003589, -0.00003023, -0.00000794, 0.00000738, 0.00000951, 0.00001455, 0.00002454, 0.00002065, 0.00000190, -0.00000505, 0.00001022, 0.00002231, 0.00001253, -0.00000097, 0.00000312, 0.00001447, 0.00001340, 0.00000406, 0.00000184, 0.00000636, 0.00000825, 0.00000750, 0.00000787, 0.00000590, -0.00000017, -0.00000328, -0.00000020, 0.00000158, -0.00000139, -0.00000081, 0.00000702, 0.00001158, 0.00000499, -0.00000440, -0.00000591, -0.00000229, -0.00000074, 0.00000107, 0.00000676, 0.00000879, -0.00000037, -0.00001174, -0.00001164, -0.00000422, -0.00000196, -0.00000012, 0.00001209, 0.00001916, -0.00000325, -0.00003481, -0.00002365, 0.00002865, 0.00005024, 0.00000343, -0.00004872, -0.00003358, 0.00002089, 0.00003538, 0.00000245, -0.00001747, -0.00000251, 0.00000894, -0.00000252, -0.00001154, -0.00000646, -0.00000769, -0.00001881, -0.00001590, 0.00000488, 0.00001772, 0.00001179, 0.00000243, -0.00000422, -0.00001669, -0.00002829, -0.00002060, 0.00000285, 0.00001972, 0.00002267, 0.00002027, 0.00001223, -0.00000609, -0.00002263, -0.00002101, -0.00000643, 0.00000427, 0.00000813, 0.00001082, 0.00000876, -0.00000210, -0.00001206, -0.00001103, -0.00000444, -0.00000124, 0.00000046, 0.00000338, 0.00000186, -0.00000554, -0.00000891, -0.00000283, 0.00000441, 0.00000535, 0.00000340, 0.00000242, 0.00000002, -0.00000400, -0.00000518, -0.00000343, -0.00000309, -0.00000394, -0.00000298, -0.00000295, -0.00000744, -0.00001133, -0.00000796, -0.00000147, -0.00000069, },
		{ -0.00000834, -0.00000495, -0.00000134, 0.00000084, -0.00000155, -0.00000640, -0.00000843, -0.00001180, -0.00002501, -0.00003992, -0.00003625, -0.00001313, 0.00000711, 0.00001351, 0.00001869, 0.00002764, 0.00002502, 0.00000760, -0.00000154, 0.00000996, 0.00002102, 0.00001271, -0.00000032, 0.00000270, 0.00001381, 0.00001377, 0.00000488, 0.00000220, 0.00000677, 0.00000924, 0.00000810, 0.00000675, 0.00000312, -0.00000325, -0.00000488, 0.00000125, 0.00000624, 0.00000411, 0.00000169, 0.00000436, 0.00000519, -0.00000121, -0.00000650, -0.00000209, 0.00000601, 0.00000754, 0.00000367, 0.00000067, -0.00000323, -0.00001074, -0.00001350, -0.00000358, 0.00000913, 0.00000957, 0.00000218, 0.00000116, 0.00000157, -0.00000962, -0.00001731, 0.00000496, 0.00003715, 0.00002464, -0.00003130, -0.00005443, -0.00000419, 0.00005269, 0.00003849, -0.00001833, -0.00003472, -0.00000284, 0.00001430, -0.00000551, -0.00002125, -0.00001180, -0.00000260, -0.00000515, 0.00000107, 0.00001751, 0.00001705, -0.00000523, -0.00002223, -0.00002140, -0.00001610, -0.00000971, 0.00000799, 0.00002865, 0.00002987, 0.00001197, -0.00000436, -0.00001231, -0.00001895, -0.00001986, -0.00000600, 0.00001186, 0.00001523, 0.00000611, -0.00000116, -0.00000494, -0.00001036, -0.00001184, -0.00000352, 0.00000544, 0.00000484, -0.00000014, -0.00000159, -0.00000299, -0.00000728, -0.00000667, 0.00000251, 0.00001013, 0.00000811, 0.00000198, -0.00000162, -0.00000452, -0.00000785, -0.00000761, -0.00000414, -0.00000298, -0.00000413, -0.00000308, -0.00000182, -0.00000547, -0.00001022, -0.00000846, -0.00000298, -0.00000268, -0.00000752, -0.00001016, },
		{ -0.00000011, -0.00000097, -0.00000493, -0.00001064, -0.00001305, -0.00001522, -0.00002666, -0.00004218, -0.00004097, -0.00001730, 0.00000818, 0.00001942, 0.00002448, 0.00003093, 0.00002790, 0.00001119, 0.00000044, 0.00000905, 0.00001961, 0.00001296, 0.00000032, 0.00000207, 0.00001301, 0.00001462, 0.00000690, 0.00000358, 0.00000677, 0.00000799, 0.00000567, 0.00000381, 0.00000151, -0.00000202, -0.00000108, 0.00000584, 0.00000931, 0.00000354, -0.00000302, -0.00000200, 0.00000167, 0.00000116, 0.00000077, 0.00000572, 0.00000950, 0.00000411, -0.00000553, -0.00000955, -0.00000812, -0.00000633, -0.00000181, 0.00000798, 0.00001337, 0.00000480, -0.00000764, -0.00000715, 0.00000196, 0.00000351, 0.00000164, 0.00001104, 0.00001745, -0.00000662, -0.00004038, -0.00002610, 0.00003516, 0.00006214, 0.00001027, -0.00005147, -0.00004129, 0.00001220, 0.00002389, -0.00001237, -0.00002962, -0.00000529, 0.00001621, 0.00001112, 0.00000414, 0.00000583, -0.00000486, -0.00002717, -0.00003018, -0.00000721, 0.00001404, 0.00002002, 0.00002278, 0.00002282, 0.00000702, -0.00001655, -0.00002365, -0.00001287, -0.00000338, 0.00000043, 0.00000761, 0.00001295, 0.00000440, -0.00001000, -0.00001285, -0.00000588, -0.00000203, -0.00000087, 0.00000391, 0.00000630, -0.00000043, -0.00000746, -0.00000524, -0.00000016, -0.00000017, 0.00000026, 0.00000620, 0.00000941, 0.00000342, -0.00000434, -0.00000647, -0.00000666, -0.00000836, -0.00000764, -0.00000382, -0.00000247, -0.00000376, -0.00000237, -0.00000021, -0.00000409, -0.00001100, -0.00001128, -0.00000563, -0.00000338, -0.00000590, -0.00000634, -0.00000311, -0.00000060, },
		{ -0.00000946, -0.00001363, -0.00001529, -0.00001643, -0.00002672, -0.00004329, -0.00004477, -0.00002073, 0.00000993, 0.00002592, 0.00003033, 0.00003323, 0.00002906, 0.00001336, 0.00000192, 0.00000812, 0.00001804, 0.00001292, 0.00000106, 0.00000215, 0.00001326, 0.00001628, 0.00000875, 0.00000334, 0.00000416, 0.00000474, 0.00000361, 0.00000418, 0.00000472, 0.00000279, 0.00000249, 0.00000587, 0.00000580, -0.00000149, -0.00000674, -0.00000176, 0.00000678, 0.00000873, 0.00000583, 0.00000429, 0.00000180, -0.00000548, -0.00001136, -0.00000800, 0.00000046, 0.00000489, 0.00000550, 0.00000683, 0.00000519, -0.00000337, -0.00000891, -0.00000023, 0.00001288, 0.00001186, 0.00000035, -0.00000367, -0.00000281, -0.00001209, -0.00001726, 0.00001007, 0.00004758, 0.00003320, -0.00003343, -0.00006708, -0.00001937, 0.00003978, 0.00002705, -0.00002629, -0.00003220, 0.00001246, 0.00003439, 0.00000855, -0.00001822, -0.00001929, -0.00001742, -0.00002042, -0.00000508, 0.00002612, 0.00003775, 0.00002009, 0.00000033, -0.00000799, -0.00001635, -0.00002284, -0.00001154, 0.00000987, 0.00001589, 0.00000486, -0.00000330, -0.00000453, -0.00000949, -0.00001394, -0.00000576, 0.00000735, 0.00000861, 0.00000088, -0.00000239, -0.00000226, -0.00000566, -0.00000635, 0.00000218, 0.00000982, 0.00000639, -0.00000014, 0.00000023, 0.00000177, -0.00000287, -0.00000768, -0.00000652, -0.00000498, -0.00000736, -0.00000794, -0.00000401, -0.00000164, -0.00000258, -0.00000147, 0.00000024, -0.00000484, -0.00001354, -0.00001403, -0.00000570, 0.00000019, -0.00000048, -0.00000129, -0.00000003, -0.00000091, -0.00000410, -0.00000656, },
		{ -0.00001554, -0.00001687, -0.00002649, -0.00004383, -0.00004779, -0.00002373, 0.00001157, 0.00003200, 0.00003548, 0.00003439, 0.00002885, 0.00001458, 0.00000316, 0.00000738, 0.00001676, 0.00001346, 0.00000284, 0.00000320, 0.00001347, 0.00001647, 0.00000828, 0.00000122, 0.00000148, 0.00000391, 0.00000572, 0.00000822, 0.00000869, 0.00000446, 0.00000048, 0.00000118, 0.00000174, -0.00000182, -0.00000283, 0.00000433, 0.00001174, 0.00000936, 0.00000073, -0.00000442, -0.00000532, -0.00000615, -0.00000473, 0.00000219, 0.00000854, 0.00000670, 0.00000065, -0.00000090, 0.00000077, -0.00000002, 0.00000063, 0.00000824, 0.00001342, 0.00000375, -0.00001134, -0.00001159, 0.00000063, 0.00000638, 0.00000703, 0.00001731, 0.00002184, -0.00000963, -0.00005362, -0.00004380, 0.00002231, 0.00005703, 0.00001036, -0.00004531, -0.00002566, 0.00003280, 0.00003619, -0.00001711, -0.00004682, -0.00002349, 0.00000606, 0.00001428, 0.00002258, 0.00003464, 0.00002240, -0.00001241, -0.00003102, -0.00002035, -0.00000586, 0.00000034, 0.00001018, 0.00001947, 0.00000902, -0.00001402, -0.00002195, -0.00001143, -0.00000252, 0.00000014, 0.00000664, 0.00001193, 0.00000353, -0.00001003, -0.00001096, -0.00000217, 0.00000264, 0.00000477, 0.00001064, 0.00001205, 0.00000129, -0.00000995, -0.00000882, -0.00000256, -0.00000278, -0.00000563, -0.00000440, -0.00000393, -0.00000785, -0.00000885, -0.00000345, 0.00000028, -0.00000146, -0.00000239, -0.00000174, -0.00000664, -0.00001446, -0.00001285, -0.00000133, 0.00000632, 0.00000401, -0.00000057, -0.00000273, -0.00000592, -0.00001007, -0.00001124, -0.00001125, -0.00001363, },
		{ -0.00002598, -0.00004318, -0.00004970, -0.00002674, 0.00001206, 0.00003661, 0.00003941, 0.00003450, 0.00002776, 0.00001550, 0.00000501, 0.00000780, 0.00001661, 0.00001472, 0.00000467, 0.00000326, 0.00001169, 0.00001468, 0.00000733, 0.00000094, 0.00000240, 0.00000666, 0.00000924, 0.00001030, 0.00000802, 0.00000148, -0.00000308, -0.00000045, 0.00000403, 0.00000376, 0.00000252, 0.00000565, 0.00000787, 0.00000223, -0.00000596, -0.00000684, -0.00000136, 0.00000271, 0.00000417, 0.00000603, 0.00000576, 0.00000029, -0.00000425, -0.00000044, 0.00000694, 0.00000817, 0.00000474, 0.00000402, 0.00000306, -0.00000451, -0.00000996, -0.00000010, 0.00001598, 0.00001660, 0.00000288, -0.00000580, -0.00001006, -0.00002429, -0.00003145, 0.00000102, 0.00004890, 0.00004206, -0.00002323, -0.00005656, -0.00000744, 0.00004710, 0.00002008, -0.00004645, -0.00005073, 0.00000972, 0.00004928, 0.00003425, 0.00001012, 0.00000258, -0.00001140, -0.00003298, -0.00002860, 0.00000305, 0.00002243, 0.00001493, 0.00000455, 0.00000070, -0.00001098, -0.00002513, -0.00001864, 0.00000384, 0.00001408, 0.00000707, 0.00000133, 0.00000017, -0.00000676, -0.00001290, -0.00000402, 0.00001136, 0.00001465, 0.00000836, 0.00000528, 0.00000203, -0.00000876, -0.00001658, -0.00001006, 0.00000049, 0.00000071, -0.00000451, -0.00000506, -0.00000453, -0.00000773, -0.00000794, -0.00000186, 0.00000120, -0.00000299, -0.00000566, -0.00000358, -0.00000505, -0.00001029, -0.00000775, 0.00000343, 0.00000883, 0.00000238, -0.00000572, -0.00000856, -0.00001039, -0.00001288, -0.00001224, -0.00001049, -0.00001275, -0.00001615, -0.00001784, },
		{ -0.00005141, -0.00003145, 0.00000985, 0.00003907, 0.00004250, 0.00003472, 0.00002720, 0.00001738, 0.00000821, 0.00000938, 0.00001661, 0.00001494, 0.00000475, 0.00000166, 0.00000924, 0.00001397, 0.00000926, 0.00000410, 0.00000528, 0.00000845, 0.00000914, 0.00000807, 0.00000510, 0.00000014, -0.00000156, 0.00000353, 0.00000873, 0.00000641, 0.00000063, -0.00000060, 0.00000085, -0.00000102, -0.00000331, 0.00000044, 0.00000695, 0.00000797, 0.00000384, 0.00000098, 0.00000026, -0.00000096, 0.00000015, 0.00000660, 0.00001164, 0.00000705, -0.00000200, -0.00000404, 0.00000034, 0.00000259, 0.00000408, 0.00001101, 0.00001484, 0.00000186, -0.00001864, -0.00002311, -0.00001094, -0.00000130, 0.00000614, 0.00002514, 0.00003603, 0.00000308, -0.00004873, -0.00004505, 0.00001914, 0.00005022, -0.00000288, -0.00005744, -0.00002235, 0.00005565, 0.00006645, 0.00000538, -0.00003844, -0.00002913, -0.00001222, -0.00001122, 0.00000067, 0.00002495, 0.00002467, -0.00000395, -0.00002174, -0.00001484, -0.00000825, -0.00000959, -0.00000073, 0.00001469, 0.00001200, -0.00000637, -0.00001348, -0.00000511, 0.00000040, 0.00000144, 0.00000985, 0.00001839, 0.00001127, -0.00000356, -0.00000789, -0.00000551, -0.00000949, -0.00001354, -0.00000636, 0.00000276, 0.00000028, -0.00000665, -0.00000588, -0.00000261, -0.00000504, -0.00000697, -0.00000300, -0.00000130, -0.00000586, -0.00000666, -0.00000026, 0.00000186, -0.00000391, -0.00000525, 0.00000184, 0.00000436, -0.00000371, -0.00001122, -0.00001112, -0.00001017, -0.00001213, -0.00001215, -0.00001065, -0.00001327, -0.00001729, -0.00001802, -0.00002372, -0.00004092, },
		{ 0.00000497, 0.00004078, 0.00004676, 0.00003683, 0.00002811, 0.00002004, 0.00001160, 0.00001035, 0.00001513, 0.00001346, 0.00000405, 0.00000092, 0.00000913, 0.00001614, 0.00001323, 0.00000718, 0.00000564, 0.00000652, 0.00000645, 0.00000632, 0.00000587, 0.00000376, 0.00000300, 0.00000617, 0.00000774, 0.00000202, -0.00000531, -0.00000496, 0.00000127, 0.00000487, 0.00000503, 0.00000655, 0.00000799, 0.00000447, -0.00000110, -0.00000160, 0.00000244, 0.00000507, 0.00000584, 0.00000745, 0.00000671, 0.00000003, -0.00000551, -0.00000044, 0.00001036, 0.00001364, 0.00000801, 0.00000208, -0.00000415, -0.00001570, -0.00002268, -0.00000965, 0.00001453, 0.00002449, 0.00001717, 0.00000994, 0.00000075, -0.00002452, -0.00004277, -0.00001314, 0.00004036, 0.00003975, -0.00002144, -0.00004668, 0.00001519, 0.00007471, 0.00003526, -0.00005249, -0.00007046, -0.00001182, 0.00003101, 0.00002104, 0.00000567, 0.00000798, -0.00000193, -0.00002648, -0.00002695, 0.00000080, 0.00001619, 0.00000592, -0.00000218, 0.00000086, -0.00000450, -0.00001616, -0.00001001, 0.00001085, 0.00001906, 0.00001132, 0.00000690, 0.00000648, -0.00000315, -0.00001447, -0.00001087, -0.00000084, -0.00000287, -0.00001072, -0.00000788, -0.00000005, -0.00000134, -0.00000600, -0.00000277, 0.00000101, -0.00000444, -0.00001044, -0.00000742, -0.00000295, -0.00000364, -0.00000122, 0.00000701, 0.00000770, -0.00000347, -0.00001079, -0.00000590, -0.00000160, -0.00000662, -0.00001128, -0.00000903, -0.00000807, -0.00001244, -0.00001450, -0.00001245, -0.00001334, -0.00001634, -0.00001591, -0.00002027, -0.00003920, -0.00005540, -0.00003941, },
		{ 0.00005336, 0.00004058, 0.00002911, 0.00002161, 0.00001361, 0.00001020, 0.00001308, 0.00001244, 0.00000515, 0.00000277, 0.00001104, 0.00001859, 0.00001547, 0.00000736, 0.00000356, 0.00000431, 0.00000624, 0.00000863, 0.00000974, 0.00000715, 0.00000348, 0.00000276, 0.00000227, -0.00000212, -0.00000567, -0.00000107, 0.00000801, 0.00001150, 0.00000811, 0.00000468, 0.00000334, 0.00000118, -0.00000025, 0.00000302, 0.00000787, 0.00000766, 0.00000349, 0.00000152, 0.00000191, 0.00000125, 0.00000245, 0.00000960, 0.00001551, 0.00000919, -0.00000559, -0.00001434, -0.00001311, -0.00000880, -0.00000088, 0.00001537, 0.00002839, 0.00001971, -0.00000402, -0.00001758, -0.00001687, -0.00001715, -0.00001237, 0.00001512, 0.00004084, 0.00001854, -0.00003056, -0.00002709, 0.00003504, 0.00005527, -0.00001763, -0.00008630, -0.00004713, 0.00004624, 0.00006783, 0.00000875, -0.00003486, -0.00002446, -0.00000939, -0.00001282, -0.00000268, 0.00002312, 0.00002295, -0.00000772, -0.00002514, -0.00001380, -0.00000217, -0.00000046, 0.00001003, 0.00002561, 0.00002063, -0.00000136, -0.00001159, -0.00000673, -0.00000649, -0.00001041, -0.00000388, 0.00000477, -0.00000229, -0.00001514, -0.00001244, -0.00000010, 0.00000262, -0.00000177, -0.00000090, -0.00000040, -0.00000885, -0.00001504, -0.00000760, 0.00000289, 0.00000472, 0.00000495, 0.00000879, 0.00000470, -0.00001076, -0.00001935, -0.00001103, -0.00000136, -0.00000350, -0.00000857, -0.00000804, -0.00000916, -0.00001548, -0.00001729, -0.00001209, -0.00001007, -0.00001307, -0.00001396, -0.00001916, -0.00004061, -0.00006210, -0.00004900, 0.00000000, 0.00004405, },
		{ 0.00002868, 0.00002170, 0.00001513, 0.00001081, 0.00001264, 0.00001337, 0.00000797, 0.00000531, 0.00001199, 0.00001877, 0.00001550, 0.00000690, 0.00000307, 0.00000521, 0.00000885, 0.00001141, 0.00001080, 0.00000557, -0.00000038, -0.00000147, 0.00000064, 0.00000069, 0.00000041, 0.00000504, 0.00001140, 0.00001123, 0.00000508, 0.00000131, 0.00000260, 0.00000438, 0.00000498, 0.00000651, 0.00000713, 0.00000340, -0.00000097, 0.00000077, 0.00000690, 0.00001004, 0.00000926, 0.00000838, 0.00000496, -0.00000532, -0.00001524, -0.00001210, 0.00000239, 0.00001408, 0.00001755, 0.00001778, 0.00001227, -0.00000527, -0.00002307, -0.00002052, -0.00000191, 0.00000920, 0.00001041, 0.00001745, 0.00002093, -0.00000220, -0.00002925, -0.00001133, 0.00003203, 0.00002068, -0.00004910, -0.00007026, 0.00001023, 0.00008716, 0.00004871, -0.00004976, -0.00007480, -0.00001527, 0.00002937, 0.00001932, 0.00000578, 0.00001126, 0.00000034, -0.00002907, -0.00003116, 0.00000147, 0.00002404, 0.00001892, 0.00001292, 0.00001404, 0.00000183, -0.00001876, -0.00001923, -0.00000201, 0.00000401, -0.00000350, -0.00000391, 0.00000061, -0.00000690, -0.00001741, -0.00001013, 0.00000588, 0.00000716, -0.00000286, -0.00000621, -0.00000528, -0.00001011, -0.00001154, 0.00000034, 0.00001187, 0.00000893, 0.00000157, 0.00000077, -0.00000322, -0.00001610, -0.00002114, -0.00000877, 0.00000301, -0.00000107, -0.00000997, -0.00001124, -0.00001161, -0.00001636, -0.00001601, -0.00000825, -0.00000586, -0.00001198, -0.00001611, -0.00002175, -0.00004355, -0.00006819, -0.00005707, -0.00000351, 0.00004846, 0.00006056, 0.00004390, },
		{ 0.00001802, 0.00001332, 0.00001373, 0.00001484, 0.00001026, 0.00000647, 0.00001131, 0.00001805, 0.00001623, 0.00000865, 0.00000497, 0.00000694, 0.00000972, 0.00001035, 0.00000791, 0.00000248, -0.00000180, -0.00000006, 0.00000502, 0.00000656, 0.00000482, 0.00000570, 0.00000880, 0.00000819, 0.00000421, 0.00000343, 0.00000676, 0.00000825, 0.00000595, 0.00000373, 0.00000290, 0.00000172, 0.00000232, 0.00000782, 0.00001350, 0.00001134, 0.00000267, -0.00000392, -0.00000589, -0.00000649, -0.00000339, 0.00000814, 0.00002107, 0.00002151, 0.00000911, -0.00000315, -0.00000978, -0.00001480, -0.00001387, 0.00000140, 0.00001988, 0.00002003, 0.00000527, -0.00000155, -0.00000038, -0.00000979, -0.00001981, -0.00000339, 0.00001916, -0.00000166, -0.00004542, -0.00002923, 0.00004871, 0.00007347, -0.00001187, -0.00009577, -0.00005835, 0.00004403, 0.00007156, 0.00001176, -0.00003233, -0.00002052, -0.00000740, -0.00001555, -0.00000530, 0.00002775, 0.00003561, 0.00000802, -0.00001161, -0.00000718, -0.00000642, -0.00001507, -0.00000889, 0.00000847, 0.00000756, -0.00000919, -0.00001267, -0.00000281, -0.00000315, -0.00001076, -0.00000499, 0.00000695, 0.00000284, -0.00001038, -0.00001063, -0.00000189, -0.00000087, -0.00000213, 0.00000573, 0.00001114, 0.00000158, -0.00000918, -0.00000691, -0.00000401, -0.00001224, -0.00001703, -0.00000655, 0.00000269, -0.00000430, -0.00001429, -0.00001285, -0.00000909, -0.00001218, -0.00001269, -0.00000630, -0.00000611, -0.00001557, -0.00002103, -0.00002411, -0.00004383, -0.00007131, -0.00006390, -0.00000779, 0.00005135, 0.00006663, 0.00004660, 0.00002800, 0.00002221, },
		{ 0.00001470, 0.00001544, 0.00001150, 0.00000708, 0.00001094, 0.00001851, 0.00001852, 0.00001125, 0.00000599, 0.00000599, 0.00000738, 0.00000758, 0.00000630, 0.00000360, 0.00000208, 0.00000488, 0.00000905, 0.00000834, 0.00000386, 0.00000296, 0.00000687, 0.00000935, 0.00000818, 0.00000756, 0.00000846, 0.00000683, 0.00000275, 0.00000146, 0.00000428, 0.00000726, 0.00000896, 0.00001064, 0.00000941, 0.00000165, -0.00000732, -0.00000753, 0.00000139, 0.00001065, 0.00001602, 0.00001936, 0.00001756, 0.00000513, -0.00001122, -0.00001653, -0.00000851, 0.00000141, 0.00000844, 0.00001628, 0.00001874, 0.00000504, -0.00001440, -0.00001691, -0.00000546, -0.00000335, -0.00000945, -0.00000239, 0.00000846, -0.00000607, -0.00002646, -0.00000203, 0.00004523, 0.00002791, -0.00005569, -0.00008364, 0.00000530, 0.00009433, 0.00005685, -0.00004867, -0.00007583, -0.00001197, 0.00003444, 0.00002173, 0.00000850, 0.00001936, 0.00001213, -0.00001983, -0.00002847, -0.00000392, 0.00000979, -0.00000123, -0.00000554, 0.00000317, -0.00000144, -0.00001657, -0.00001263, 0.00000589, 0.00000713, -0.00000773, -0.00001055, -0.00000172, -0.00000323, -0.00001032, -0.00000283, 0.00001072, 0.00000863, -0.00000206, -0.00000252, 0.00000028, -0.00000724, -0.00001300, -0.00000402, 0.00000331, -0.00000687, -0.00001791, -0.00001175, -0.00000234, -0.00000674, -0.00001327, -0.00000803, -0.00000281, -0.00000899, -0.00001437, -0.00001028, -0.00001002, -0.00001931, -0.00002344, -0.00002302, -0.00004151, -0.00007404, -0.00007277, -0.00001503, 0.00005254, 0.00007310, 0.00005090, 0.00002885, 0.00002384, 0.00002187, 0.00001647, },
		{ 0.00001284, 0.00000829, 0.00001134, 0.00001931, 0.00002020, 0.00001220, 0.00000484, 0.00000358, 0.00000561, 0.00000775, 0.00000889, 0.00000811, 0.00000659, 0.00000718, 0.00000860, 0.00000650, 0.00000244, 0.00000321, 0.00000930, 0.00001322, 0.00001122, 0.00000767, 0.00000593, 0.00000438, 0.00000312, 0.00000537, 0.00000994, 0.00001115, 0.00000766, 0.00000344, 0.00000008, -0.00000330, -0.00000290, 0.00000625, 0.00001920, 0.00002424, 0.00001815, 0.00000813, -0.00000123, -0.00001038, -0.00001447, -0.00000600, 0.00000967, 0.00001728, 0.00001283, 0.00000565, -0.00000158, -0.00001351, -0.00002237, -0.00001406, 0.00000275, 0.00000466, -0.00000615, -0.00000608, 0.00000286, -0.00000283, -0.00001393, 0.00000083, 0.00002210, -0.00000314, -0.00005234, -0.00003396, 0.00005401, 0.00008366, -0.00000918, -0.00010131, -0.00005934, 0.00005476, 0.00008574, 0.00001926, -0.00003050, -0.00001877, -0.00000632, -0.00001917, -0.00001383, 0.00001625, 0.00002147, -0.00000674, -0.00002113, -0.00000777, -0.00000099, -0.00000814, -0.00000200, 0.00001333, 0.00000593, -0.00001814, -0.00002200, -0.00000433, 0.00000482, 0.00000258, 0.00000830, 0.00001513, 0.00000316, -0.00001531, -0.00001520, -0.00000431, -0.00000372, -0.00000623, 0.00000200, 0.00000577, -0.00001026, -0.00002567, -0.00001772, -0.00000179, -0.00000110, -0.00000675, -0.00000325, -0.00000154, -0.00001272, -0.00002122, -0.00001527, -0.00001095, -0.00001865, -0.00002293, -0.00002159, -0.00004071, -0.00007905, -0.00008378, -0.00002327, 0.00005456, 0.00008175, 0.00005715, 0.00003035, 0.00002518, 0.00002539, 0.00001955, 0.00001557, 0.00001585, },
		{ 0.00001123, 0.00001907, 0.00002059, 0.00001229, 0.00000404, 0.00000311, 0.00000693, 0.00001080, 0.00001254, 0.00001108, 0.00000773, 0.00000621, 0.00000697, 0.00000642, 0.00000465, 0.00000651, 0.00001201, 0.00001430, 0.00001040, 0.00000584, 0.00000526, 0.00000675, 0.00000798, 0.00000959, 0.00001023, 0.00000652, 0.00000024, -0.00000208, 0.00000155, 0.00000731, 0.00001357, 0.00002100, 0.00002486, 0.00001802, 0.00000330, -0.00000755, -0.00000902, -0.00000554, 0.00000014, 0.00000911, 0.00001570, 0.00001027, -0.00000420, -0.00001371, -0.00001436, -0.00001425, -0.00001320, -0.00000342, 0.00000775, 0.00000358, -0.00000987, -0.00000979, 0.00000154, 0.00000014, -0.00001075, -0.00000505, 0.00000822, -0.00000476, -0.00002547, 0.00000095, 0.00005159, 0.00003166, -0.00005975, -0.00008761, 0.00001500, 0.00011654, 0.00007414, -0.00004841, -0.00008686, -0.00002339, 0.00002456, 0.00001085, -0.00000168, 0.00001259, 0.00000713, -0.00002440, -0.00002925, 0.00000129, 0.00001693, 0.00000312, -0.00000382, 0.00000308, -0.00000578, -0.00002519, -0.00001859, 0.00000995, 0.00002089, 0.00000930, 0.00000286, 0.00000271, -0.00000991, -0.00002370, -0.00001454, 0.00000462, 0.00000575, -0.00000372, -0.00000255, -0.00000116, -0.00001606, -0.00002802, -0.00001425, 0.00000649, 0.00000615, -0.00000536, -0.00000684, -0.00000689, -0.00001821, -0.00002514, -0.00001511, -0.00000767, -0.00001703, -0.00002476, -0.00002353, -0.00004173, -0.00008372, -0.00009352, -0.00003049, 0.00005731, 0.00009097, 0.00006352, 0.00003144, 0.00002623, 0.00002935, 0.00002350, 0.00001698, 0.00001626, 0.00001401, 0.00000929, },
		{ 0.00002133, 0.00001363, 0.00000528, 0.00000476, 0.00000938, 0.00001320, 0.00001389, 0.00001132, 0.00000743, 0.00000612, 0.00000809, 0.00000922, 0.00000797, 0.00000831, 0.00001157, 0.00001293, 0.00001000, 0.00000745, 0.00000875, 0.00001067, 0.00000972, 0.00000711, 0.00000438, 0.00000128, 0.00000028, 0.00000564, 0.00001550, 0.00002231, 0.00002270, 0.00001947, 0.00001355, 0.00000342, -0.00000647, -0.00000752, 0.00000019, 0.00000700, 0.00000758, 0.00000515, 0.00000068, -0.00000873, -0.00001823, -0.00001753, -0.00000827, -0.00000298, -0.00000390, -0.00000170, 0.00000233, -0.00000287, -0.00001227, -0.00000862, 0.00000343, 0.00000177, -0.00001022, -0.00000710, 0.00000717, 0.00000364, -0.00000968, 0.00000293, 0.00002379, -0.00000298, -0.00005277, -0.00002673, 0.00007417, 0.00010544, -0.00000528, -0.00011945, -0.00008405, 0.00003769, 0.00007573, 0.00001143, -0.00003418, -0.00001579, -0.00000123, -0.00001685, -0.00001223, 0.00001997, 0.00002428, -0.00000804, -0.00002373, -0.00000841, -0.00000178, -0.00001086, -0.00000201, 0.00002139, 0.00002015, -0.00000470, -0.00001506, -0.00000721, -0.00000836, -0.00001531, -0.00000491, 0.00001067, 0.00000369, -0.00001355, -0.00001246, -0.00000341, -0.00001069, -0.00002006, -0.00000734, 0.00001008, 0.00000392, -0.00001271, -0.00001383, -0.00000921, -0.00001722, -0.00002368, -0.00001350, -0.00000670, -0.00001940, -0.00002972, -0.00002577, -0.00004014, -0.00008513, -0.00010213, -0.00003921, 0.00005789, 0.00009939, 0.00007068, 0.00003396, 0.00002861, 0.00003421, 0.00002757, 0.00001760, 0.00001545, 0.00001422, 0.00000984, 0.00001089, 0.00001870, },
		{ 0.00000739, 0.00000636, 0.00001065, 0.00001378, 0.00001383, 0.00001165, 0.00000895, 0.00000857, 0.00001079, 0.00001152, 0.00000904, 0.00000773, 0.00001053, 0.00001352, 0.00001300, 0.00001130, 0.00001101, 0.00000979, 0.00000591, 0.00000262, 0.00000311, 0.00000662, 0.00001227, 0.00002016, 0.00002649, 0.00002503, 0.00001612, 0.00000710, 0.00000217, -0.00000071, -0.00000181, 0.00000156, 0.00000644, 0.00000477, -0.00000428, -0.00001232, -0.00001485, -0.00001525, -0.00001388, -0.00000681, 0.00000173, 0.00000130, -0.00000648, -0.00000879, -0.00000418, -0.00000397, -0.00000792, -0.00000372, 0.00000508, 0.00000142, -0.00001003, -0.00000659, 0.00000781, 0.00000594, -0.00000843, -0.00000426, 0.00001124, 0.00000100, -0.00001724, 0.00001373, 0.00006655, 0.00003620, -0.00007643, -0.00011778, -0.00000880, 0.00010769, 0.00007198, -0.00005054, -0.00008412, -0.00001254, 0.00003547, 0.00001360, -0.00000439, 0.00001047, 0.00000518, -0.00002784, -0.00003010, 0.00000638, 0.00002352, 0.00000585, -0.00000261, 0.00000788, 0.00000177, -0.00001984, -0.00001854, 0.00000322, 0.00000689, -0.00000747, -0.00000809, 0.00000114, -0.00000673, -0.00002015, -0.00001033, 0.00000755, 0.00000117, -0.00001583, -0.00001134, 0.00000265, -0.00000340, -0.00001760, -0.00001351, -0.00000427, -0.00001331, -0.00002477, -0.00001721, -0.00000970, -0.00002217, -0.00003207, -0.00002436, -0.00003591, -0.00008656, -0.00011362, -0.00005157, 0.00005716, 0.00010994, 0.00008145, 0.00003873, 0.00003081, 0.00003732, 0.00002961, 0.00001675, 0.00001417, 0.00001500, 0.00001149, 0.00001152, 0.00001907, 0.00002292, 0.00001607, },
		{ 0.00001101, 0.00001406, 0.00001462, 0.00001374, 0.00001216, 0.00001149, 0.00001228, 0.00001178, 0.00000887, 0.00000789, 0.00001187, 0.00001641, 0.00001613, 0.00001231, 0.00000885, 0.00000622, 0.00000432, 0.00000602, 0.00001253, 0.00002011, 0.00002500, 0.00002668, 0.00002417, 0.00001584, 0.00000521, -0.00000015, 0.00000145, 0.00000380, 0.00000271, 0.00000005, -0.00000363, -0.00001094, -0.00001919, -0.00002040, -0.00001337, -0.00000604, -0.00000278, -0.00000046, 0.00000044, -0.00000514, -0.00001307, -0.00001177, -0.00000223, 0.00000183, -0.00000193, -0.00000179, 0.00000259, -0.00000146, -0.00001038, -0.00000603, 0.00000745, 0.00000638, -0.00000573, -0.00000093, 0.00001647, 0.00001342, -0.00000318, 0.00000538, 0.00002083, -0.00001651, -0.00007734, -0.00005030, 0.00006500, 0.00010820, -0.00000349, -0.00012074, -0.00007788, 0.00005354, 0.00008756, 0.00000885, -0.00004432, -0.00002195, -0.00000203, -0.00001535, -0.00000628, 0.00003164, 0.00003436, -0.00000708, -0.00002906, -0.00001219, -0.00000203, -0.00001067, -0.00000345, 0.00001656, 0.00000949, -0.00001877, -0.00002439, -0.00000750, -0.00000382, -0.00001029, 0.00000128, 0.00001631, 0.00000149, -0.00002523, -0.00002322, -0.00000314, -0.00000210, -0.00001335, -0.00000901, -0.00000151, -0.00001548, -0.00003154, -0.00002245, -0.00000934, -0.00001960, -0.00003118, -0.00002353, -0.00003472, -0.00009119, -0.00012694, -0.00006352, 0.00005926, 0.00012418, 0.00009397, 0.00004181, 0.00002928, 0.00003753, 0.00003119, 0.00001743, 0.00001511, 0.00001761, 0.00001402, 0.00001217, 0.00001916, 0.00002451, 0.00001871, 0.00000938, 0.00000725, },
		{ 0.00001679, 0.00001680, 0.00001522, 0.00001316, 0.00001234, 0.00001164, 0.00000987, 0.00001002, 0.00001425, 0.00001811, 0.00001624, 0.00001066, 0.00000724, 0.00000793, 0.00001162, 0.00001804, 0.00002565, 0.00002975, 0.00002761, 0.00002160, 0.00001445, 0.00000662, 0.00000026, -0.00000075, 0.00000201, 0.00000130, -0.00000563, -0.00001351, -0.00001794, -0.00001991, -0.00001921, -0.00001273, -0.00000299, 0.00000150, -0.00000160, -0.00000554, -0.00000698, -0.00000946, -0.00001153, -0.00000607, 0.00000384, 0.00000543, -0.00000211, -0.00000555, -0.00000208, -0.00000310, -0.00000798, -0.00000245, 0.00001044, 0.00001084, 0.00000105, 0.00000407, 0.00001569, 0.00000708, -0.00001607, -0.00001755, -0.00000340, -0.00001414, -0.00003131, 0.00000704, 0.00007002, 0.00004123, -0.00007869, -0.00012072, 0.00000084, 0.00012627, 0.00007991, -0.00006122, -0.00009853, -0.00001479, 0.00004351, 0.00002293, 0.00000523, 0.00002136, 0.00001033, -0.00003499, -0.00004400, -0.00000339, 0.00002097, 0.00000673, -0.00000136, 0.00000664, -0.00000596, -0.00003265, -0.00002751, 0.00000396, 0.00001347, -0.00000018, 0.00000014, 0.00000833, -0.00000890, -0.00003441, -0.00002584, 0.00000271, 0.00000588, -0.00001088, -0.00001221, -0.00000620, -0.00002001, -0.00003487, -0.00002129, -0.00000383, -0.00001649, -0.00003465, -0.00002843, -0.00003605, -0.00009364, -0.00013585, -0.00007216, 0.00006202, 0.00013658, 0.00010291, 0.00004111, 0.00002605, 0.00003938, 0.00003660, 0.00002153, 0.00001728, 0.00001948, 0.00001511, 0.00001180, 0.00001912, 0.00002662, 0.00002187, 0.00001136, 0.00000781, 0.00001144, 0.00001526, },
		{ 0.00001703, 0.00001371, 0.00001234, 0.00001262, 0.00001229, 0.00001250, 0.00001528, 0.00001770, 0.00001577, 0.00001200, 0.00001237, 0.00001784, 0.00002471, 0.00003015, 0.00003245, 0.00002930, 0.00002097, 0.00001194, 0.00000551, 0.00000061, -0.00000367, -0.00000574, -0.00000666, -0.00001080, -0.00001808, -0.00002213, -0.00001944, -0.00001392, -0.00000920, -0.00000438, -0.00000054, -0.00000216, -0.00000811, -0.00001059, -0.00000731, -0.00000450, -0.00000417, -0.00000105, 0.00000380, 0.00000154, -0.00000680, -0.00000868, -0.00000202, 0.00000112, -0.00000103, 0.00000369, 0.00001392, 0.00001270, 0.00000078, -0.00000217, 0.00000242, -0.00000732, -0.00002470, -0.00001797, 0.00000702, 0.00001085, -0.00000225, 0.00000884, 0.00002483, -0.00001787, -0.00008460, -0.00005178, 0.00007862, 0.00012567, -0.00000251, -0.00013612, -0.00008742, 0.00006312, 0.00010482, 0.00001847, -0.00004157, -0.00001988, -0.00000457, -0.00002822, -0.00002311, 0.00002217, 0.00003406, -0.00000388, -0.00002586, -0.00001135, -0.00000781, -0.00002216, -0.00001116, 0.00001934, 0.00001835, -0.00001033, -0.00001656, -0.00000183, -0.00000878, -0.00002893, -0.00001916, 0.00000796, 0.00000531, -0.00001852, -0.00002019, -0.00000743, -0.00001595, -0.00003099, -0.00001887, -0.00000293, -0.00002039, -0.00004308, -0.00003285, -0.00003134, -0.00008869, -0.00014176, -0.00008448, 0.00005753, 0.00014321, 0.00011035, 0.00004289, 0.00002752, 0.00004596, 0.00004452, 0.00002482, 0.00001651, 0.00001873, 0.00001560, 0.00001261, 0.00002076, 0.00002975, 0.00002503, 0.00001285, 0.00000827, 0.00001271, 0.00001781, 0.00001980, 0.00001950, },
		{ 0.00001365, 0.00001506, 0.00001512, 0.00001411, 0.00001532, 0.00001823, 0.00001958, 0.00002041, 0.00002459, 0.00003135, 0.00003568, 0.00003487, 0.00002952, 0.00002049, 0.00000966, 0.00000087, -0.00000428, -0.00000850, -0.00001355, -0.00001712, -0.00001799, -0.00001863, -0.00001953, -0.00001704, -0.00001057, -0.00000550, -0.00000514, -0.00000632, -0.00000650, -0.00000762, -0.00000920, -0.00000663, -0.00000061, 0.00000143, -0.00000215, -0.00000412, -0.00000225, -0.00000309, -0.00000689, -0.00000448, 0.00000450, 0.00000822, 0.00000450, 0.00000496, 0.00001010, 0.00000567, -0.00000793, -0.00001241, -0.00000643, -0.00000814, -0.00001460, -0.00000280, 0.00001896, 0.00001522, -0.00000990, -0.00001570, -0.00000568, -0.00002087, -0.00003848, 0.00000947, 0.00008489, 0.00005435, -0.00008190, -0.00013185, 0.00000365, 0.00014581, 0.00009509, -0.00006421, -0.00010898, -0.00001917, 0.00004063, 0.00001172, -0.00000990, 0.00001359, 0.00001126, -0.00003137, -0.00004000, -0.00000005, 0.00001869, -0.00000202, -0.00000756, 0.00001026, 0.00000274, -0.00002622, -0.00002310, 0.00000610, 0.00000528, -0.00002197, -0.00002258, 0.00000004, -0.00000244, -0.00002356, -0.00001842, 0.00000157, -0.00000801, -0.00003202, -0.00002618, -0.00000933, -0.00002446, -0.00004522, -0.00002877, -0.00002078, -0.00008425, -0.00015512, -0.00010704, 0.00004680, 0.00015160, 0.00012539, 0.00005240, 0.00003258, 0.00005138, 0.00004844, 0.00002397, 0.00001386, 0.00001930, 0.00001929, 0.00001609, 0.00002296, 0.00003177, 0.00002698, 0.00001430, 0.00001009, 0.00001583, 0.00002132, 0.00002215, 0.00002047, 0.00001744, 0.00001428, },
		{ 0.00001721, 0.00001516, 0.00001711, 0.00002386, 0.00003076, 0.00003542, 0.00003920, 0.00004120, 0.00003782, 0.00002855, 0.00001661, 0.00000460, -0.00000609, -0.00001354, -0.00001744, -0.00002042, -0.00002346, -0.00002359, -0.00001933, -0.00001417, -0.00001105, -0.00000853, -0.00000588, -0.00000610, -0.00000963, -0.00001146, -0.00000874, -0.00000495, -0.00000277, -0.00000024, 0.00000211, -0.00000055, -0.00000725, -0.00000942, -0.00000447, -0.00000008, 0.00000050, 0.00000359, 0.00000974, 0.00000946, 0.00000168, -0.00000204, 0.00000037, -0.00000311, -0.00001271, -0.00001258, -0.00000166, 0.00000200, -0.00000296, 0.00000169, 0.00001155, 0.00000102, -0.00002227, -0.00002113, 0.00000130, 0.00000427, -0.00000635, 0.00001351, 0.00003820, -0.00000724, -0.00008475, -0.00005282, 0.00009017, 0.00014191, -0.00000306, -0.00015555, -0.00010335, 0.00006428, 0.00010933, 0.00000968, -0.00005821, -0.00002895, -0.00000327, -0.00002306, -0.00001614, 0.00003002, 0.00003588, -0.00001132, -0.00003347, -0.00001022, -0.00000214, -0.00001971, -0.00001094, 0.00001868, 0.00000920, -0.00003188, -0.00003778, -0.00000690, 0.00000175, -0.00001494, -0.00001037, 0.00000632, -0.00001131, -0.00004384, -0.00003665, -0.00000894, -0.00001664, -0.00003883, -0.00002482, -0.00001794, -0.00009014, -0.00017633, -0.00013133, 0.00004159, 0.00016970, 0.00014801, 0.00006247, 0.00003195, 0.00004998, 0.00004908, 0.00002458, 0.00001560, 0.00002426, 0.00002494, 0.00001842, 0.00002252, 0.00003225, 0.00002992, 0.00001853, 0.00001441, 0.00001943, 0.00002307, 0.00002180, 0.00001976, 0.00001842, 0.00001717, 0.00001725, 0.00001825, },
		{ 0.00002562, 0.00003825, 0.00004877, 0.00005148, 0.00004743, 0.00003860, 0.00002505, 0.00000874, -0.00000597, -0.00001673, -0.00002382, -0.00002715, -0.00002697, -0.00002539, -0.00002338, -0.00001889, -0.00001165, -0.00000609, -0.00000556, -0.00000769, -0.00000930, -0.00001076, -0.00001198, -0.00000970, -0.00000352, 0.00000121, 0.00000107, -0.00000098, -0.00000268, -0.00000607, -0.00000987, -0.00000748, 0.00000147, 0.00000785, 0.00000724, 0.00000584, 0.00000640, 0.00000238, -0.00000652, -0.00000970, -0.00000492, -0.00000345, -0.00000747, -0.00000467, 0.00000520, 0.00000561, -0.00000469, -0.00000703, -0.00000041, -0.00000462, -0.00001571, -0.00000691, 0.00001452, 0.00001110, -0.00001183, -0.00001049, 0.00000736, -0.00000819, -0.00003260, 0.00001363, 0.00009233, 0.00005523, -0.00009912, -0.00015600, -0.00000277, 0.00015936, 0.00010293, -0.00007879, -0.00013121, -0.00002724, 0.00004744, 0.00002204, 0.00000048, 0.00002382, 0.00001365, -0.00004129, -0.00005201, -0.00000301, 0.00002096, -0.00000306, -0.00001008, 0.00000955, -0.00000358, -0.00004374, -0.00004038, 0.00000432, 0.00001732, -0.00000915, -0.00001650, -0.00000385, -0.00002041, -0.00004949, -0.00003343, 0.00000414, -0.00000522, -0.00004064, -0.00003543, -0.00002612, -0.00009713, -0.00019010, -0.00014570, 0.00004338, 0.00018909, 0.00016589, 0.00006450, 0.00002523, 0.00004874, 0.00005565, 0.00003223, 0.00002092, 0.00002812, 0.00002685, 0.00001762, 0.00002202, 0.00003595, 0.00003715, 0.00002501, 0.00001737, 0.00001921, 0.00002131, 0.00002065, 0.00002148, 0.00002343, 0.00002308, 0.00002124, 0.00001987, 0.00001845, 0.00001879, },
		{ 0.00006434, 0.00005531, 0.00003664, 0.00001559, -0.00000425, -0.00002058, -0.00003060, -0.00003410, -0.00003351, -0.00003032, -0.00002508, -0.00001967, -0.00001542, -0.00001111, -0.00000644, -0.00000463, -0.00000745, -0.00001119, -0.00001176, -0.00000985, -0.00000746, -0.00000385, 0.00000046, 0.00000136, -0.00000246, -0.00000618, -0.00000633, -0.00000536, -0.00000438, 0.00000024, 0.00000789, 0.00001061, 0.00000540, -0.00000035, -0.00000214, -0.00000510, -0.00001041, -0.00000971, -0.00000169, 0.00000247, -0.00000087, -0.00000129, 0.00000300, -0.00000047, -0.00001129, -0.00001157, 0.00000008, 0.00000351, -0.00000328, -0.00000003, 0.00000956, -0.00000118, -0.00002272, -0.00001472, 0.00001562, 0.00001858, -0.00000023, 0.00001244, 0.00003331, -0.00001959, -0.00010530, -0.00006654, 0.00009820, 0.00015979, -0.00000386, -0.00018009, -0.00012470, 0.00006704, 0.00012607, 0.00002199, -0.00005113, -0.00002191, -0.00000312, -0.00003622, -0.00003232, 0.00002373, 0.00003589, -0.00001408, -0.00003585, -0.00000705, -0.00000198, -0.00003176, -0.00002607, 0.00001583, 0.00001772, -0.00002362, -0.00003443, -0.00000950, -0.00001237, -0.00003760, -0.00002347, 0.00001001, -0.00000995, -0.00005811, -0.00005195, -0.00002672, -0.00008847, -0.00019152, -0.00015912, 0.00003591, 0.00019672, 0.00017660, 0.00006671, 0.00002501, 0.00005654, 0.00006860, 0.00003999, 0.00002106, 0.00002641, 0.00002737, 0.00002052, 0.00002756, 0.00004399, 0.00004418, 0.00002692, 0.00001460, 0.00001616, 0.00002164, 0.00002511, 0.00002872, 0.00003069, 0.00002731, 0.00002200, 0.00002081, 0.00002459, 0.00003251, 0.00004532, 0.00005919, },
		{ 0.00000102, -0.00002320, -0.00003765, -0.00004360, -0.00004232, -0.00003616, -0.00002883, -0.00002222, -0.00001632, -0.00001205, -0.00001028, -0.00000936, -0.00000782, -0.00000733, -0.00000912, -0.00001031, -0.00000816, -0.00000476, -0.00000320, -0.00000281, -0.00000235, -0.00000343, -0.00000614, -0.00000639, -0.00000240, 0.00000156, 0.00000261, 0.00000387, 0.00000638, 0.00000495, -0.00000196, -0.00000716, -0.00000652, -0.00000551, -0.00000691, -0.00000452, 0.00000271, 0.00000487, -0.00000126, -0.00000519, -0.00000299, -0.00000442, -0.00001011, -0.00000627, 0.00000657, 0.00000877, -0.00000231, -0.00000617, -0.00000041, -0.00000511, -0.00001469, -0.00000021, 0.00002694, 0.00002109, -0.00001234, -0.00001998, -0.00000483, -0.00002116, -0.00004415, 0.00001248, 0.00010472, 0.00006447, -0.00011169, -0.00018054, -0.00000982, 0.00017730, 0.00012174, -0.00007673, -0.00013459, -0.00002306, 0.00004898, 0.00000806, -0.00001937, 0.00001457, 0.00001321, -0.00004229, -0.00005006, 0.00000708, 0.00002749, -0.00001360, -0.00002916, 0.00000062, 0.00000010, -0.00003720, -0.00003425, 0.00000839, 0.00000938, -0.00003081, -0.00003341, -0.00000503, -0.00002384, -0.00006924, -0.00005178, -0.00001014, -0.00007226, -0.00020017, -0.00019012, 0.00001269, 0.00020088, 0.00019659, 0.00008274, 0.00003415, 0.00006529, 0.00007564, 0.00003993, 0.00001725, 0.00002782, 0.00003616, 0.00003098, 0.00003473, 0.00004690, 0.00004310, 0.00002325, 0.00001256, 0.00001990, 0.00003064, 0.00003474, 0.00003490, 0.00003256, 0.00002717, 0.00002472, 0.00003188, 0.00004580, 0.00005914, 0.00006828, 0.00006985, 0.00005781, 0.00003163, },
		{ -0.00005314, -0.00004505, -0.00003487, -0.00002502, -0.00001785, -0.00001385, -0.00001158, -0.00001062, -0.00001111, -0.00001122, -0.00000914, -0.00000652, -0.00000569, -0.00000554, -0.00000423, -0.00000337, -0.00000489, -0.00000671, -0.00000622, -0.00000455, -0.00000316, -0.00000040, 0.00000393, 0.00000558, 0.00000259, -0.00000054, -0.00000078, -0.00000183, -0.00000577, -0.00000745, -0.00000407, -0.00000134, -0.00000259, -0.00000227, 0.00000168, 0.00000161, -0.00000486, -0.00000788, -0.00000346, -0.00000126, -0.00000428, -0.00000184, 0.00000644, 0.00000509, -0.00000654, -0.00000862, 0.00000166, 0.00000366, -0.00000354, 0.00000244, 0.00001511, 0.00000151, -0.00002833, -0.00002561, 0.00000633, 0.00001260, -0.00000381, 0.00001393, 0.00003927, -0.00001990, -0.00011802, -0.00007688, 0.00010843, 0.00018082, 0.00000017, -0.00019620, -0.00013377, 0.00007788, 0.00013381, 0.00000568, -0.00007711, -0.00003266, 0.00000120, -0.00003027, -0.00002423, 0.00003791, 0.00004235, -0.00003010, -0.00006257, -0.00002019, 0.00000314, -0.00002004, -0.00001349, 0.00002535, 0.00001050, -0.00005140, -0.00006035, -0.00001614, -0.00001675, -0.00005744, -0.00004103, 0.00000044, -0.00007329, -0.00022756, -0.00023151, -0.00000664, 0.00022057, 0.00023278, 0.00010450, 0.00003644, 0.00006221, 0.00007564, 0.00004234, 0.00002338, 0.00004019, 0.00004993, 0.00003710, 0.00003119, 0.00004026, 0.00004033, 0.00002702, 0.00002243, 0.00003277, 0.00004070, 0.00003758, 0.00003239, 0.00003200, 0.00003584, 0.00004575, 0.00006263, 0.00007762, 0.00007974, 0.00006772, 0.00004598, 0.00001661, -0.00001646, -0.00004300, -0.00005471, },
		{ -0.00002081, -0.00001549, -0.00001307, -0.00001274, -0.00001262, -0.00001205, -0.00001149, -0.00000993, -0.00000622, -0.00000249, -0.00000183, -0.00000364, -0.00000527, -0.00000656, -0.00000818, -0.00000811, -0.00000454, 0.00000011, 0.00000272, 0.00000387, 0.00000444, 0.00000221, -0.00000313, -0.00000657, -0.00000506, -0.00000278, -0.00000336, -0.00000351, -0.00000077, 0.00000030, -0.00000287, -0.00000457, -0.00000177, -0.00000082, -0.00000459, -0.00000535, -0.00000012, 0.00000182, -0.00000281, -0.00000353, 0.00000249, 0.00000305, -0.00000363, -0.00000235, 0.00000741, 0.00000579, -0.00000817, -0.00001081, -0.00000112, -0.00000495, -0.00001798, -0.00000567, 0.00002316, 0.00001846, -0.00001601, -0.00002224, -0.00000328, -0.00001992, -0.00004550, 0.00001644, 0.00011762, 0.00006994, -0.00012968, -0.00020568, -0.00000939, 0.00020189, 0.00013127, -0.00010230, -0.00016634, -0.00002741, 0.00006557, 0.00002123, -0.00001229, 0.00002331, 0.00001319, -0.00006399, -0.00007842, -0.00000132, 0.00004068, 0.00000356, -0.00001616, 0.00000701, -0.00001289, -0.00007274, -0.00006579, 0.00000274, 0.00000999, -0.00004871, -0.00005264, -0.00001419, -0.00008656, -0.00025007, -0.00025751, -0.00001016, 0.00024820, 0.00026503, 0.00011340, 0.00002682, 0.00005822, 0.00008731, 0.00006073, 0.00003878, 0.00004954, 0.00005107, 0.00002950, 0.00002197, 0.00004025, 0.00005279, 0.00004511, 0.00003690, 0.00003845, 0.00003788, 0.00003236, 0.00003531, 0.00005155, 0.00007131, 0.00008660, 0.00009403, 0.00008653, 0.00005955, 0.00002158, -0.00001358, -0.00004024, -0.00005816, -0.00006437, -0.00005719, -0.00004272, -0.00002956, },
		{ -0.00001461, -0.00001375, -0.00001227, -0.00001009, -0.00000818, -0.00000621, -0.00000322, -0.00000095, -0.00000210, -0.00000563, -0.00000799, -0.00000814, -0.00000713, -0.00000450, 0.00000016, 0.00000397, 0.00000402, 0.00000172, -0.00000045, -0.00000320, -0.00000679, -0.00000755, -0.00000363, 0.00000022, -0.00000014, -0.00000176, -0.00000135, -0.00000185, -0.00000516, -0.00000616, -0.00000213, 0.00000069, -0.00000154, -0.00000268, 0.00000046, 0.00000065, -0.00000432, -0.00000441, 0.00000345, 0.00000706, 0.00000189, -0.00000022, 0.00000268, -0.00000393, -0.00001817, -0.00001694, 0.00000080, 0.00000744, -0.00000126, 0.00000120, 0.00001211, -0.00000287, -0.00003360, -0.00002714, 0.00001218, 0.00002165, 0.00000235, 0.00001715, 0.00003818, -0.00003453, -0.00014574, -0.00009249, 0.00012686, 0.00021222, -0.00000008, -0.00023266, -0.00016086, 0.00009071, 0.00016186, 0.00001374, -0.00008439, -0.00003390, 0.00000059, -0.00004646, -0.00004460, 0.00003737, 0.00005870, -0.00001782, -0.00005891, -0.00002052, -0.00001177, -0.00005544, -0.00004344, 0.00002393, 0.00001447, -0.00007266, -0.00008703, -0.00002750, -0.00007733, -0.00024714, -0.00027339, -0.00002297, 0.00025697, 0.00028239, 0.00012019, 0.00002924, 0.00007403, 0.00011462, 0.00008018, 0.00003991, 0.00003952, 0.00004132, 0.00002691, 0.00003074, 0.00006035, 0.00007533, 0.00005736, 0.00003436, 0.00002905, 0.00003523, 0.00004751, 0.00007170, 0.00010161, 0.00011712, 0.00010803, 0.00007952, 0.00003796, -0.00001018, -0.00005067, -0.00007030, -0.00007065, -0.00006236, -0.00005086, -0.00003680, -0.00002409, -0.00001746, -0.00001593, -0.00001545, },
		{ -0.00000829, -0.00000652, -0.00000558, -0.00000507, -0.00000373, -0.00000252, -0.00000374, -0.00000650, -0.00000741, -0.00000571, -0.00000338, -0.00000109, 0.00000147, 0.00000255, 0.00000049, -0.00000263, -0.00000398, -0.00000446, -0.00000547, -0.00000487, -0.00000127, 0.00000138, -0.00000034, -0.00000324, -0.00000350, -0.00000334, -0.00000498, -0.00000467, -0.00000031, 0.00000230, -0.00000057, -0.00000304, -0.00000097, 0.00000006, -0.00000253, -0.00000096, 0.00000653, 0.00000818, -0.00000047, -0.00000719, -0.00000667, -0.00000970, -0.00001608, -0.00000867, 0.00001023, 0.00001340, -0.00000348, -0.00001189, -0.00000517, -0.00000890, -0.00001953, -0.00000143, 0.00003352, 0.00002777, -0.00001577, -0.00003126, -0.00001845, -0.00004074, -0.00006451, 0.00001836, 0.00014568, 0.00009377, -0.00014347, -0.00024161, -0.00001830, 0.00023106, 0.00015418, -0.00011659, -0.00019032, -0.00002507, 0.00008162, 0.00001881, -0.00002893, 0.00001858, 0.00002021, -0.00006332, -0.00008206, 0.00000092, 0.00003445, -0.00002625, -0.00004658, 0.00000319, -0.00001220, -0.00010099, -0.00009974, -0.00000753, -0.00004469, -0.00024691, -0.00031610, -0.00006627, 0.00025432, 0.00031423, 0.00015543, 0.00005510, 0.00009523, 0.00012776, 0.00007544, 0.00002270, 0.00002997, 0.00005184, 0.00005189, 0.00005708, 0.00007758, 0.00007760, 0.00004659, 0.00002405, 0.00003771, 0.00007218, 0.00010551, 0.00013108, 0.00013897, 0.00011409, 0.00005982, -0.00000123, -0.00005118, -0.00008385, -0.00009411, -0.00008047, -0.00005509, -0.00003510, -0.00002560, -0.00002110, -0.00001842, -0.00001831, -0.00001893, -0.00001711, -0.00001344, -0.00001041, },
		{ -0.00000626, -0.00000621, -0.00000492, -0.00000301, -0.00000289, -0.00000442, -0.00000479, -0.00000332, -0.00000207, -0.00000171, -0.00000115, -0.00000093, -0.00000227, -0.00000372, -0.00000333, -0.00000251, -0.00000324, -0.00000367, -0.00000180, -0.00000025, -0.00000188, -0.00000406, -0.00000345, -0.00000227, -0.00000318, -0.00000322, -0.00000011, 0.00000164, -0.00000106, -0.00000284, 0.00000026, 0.00000269, 0.00000070, 0.00000025, 0.00000369, 0.00000208, -0.00000720, -0.00001202, -0.00000717, -0.00000427, -0.00000722, -0.00000317, 0.00000748, 0.00000428, -0.00001322, -0.00001694, -0.00000046, 0.00000805, 0.00000166, 0.00000635, 0.00001742, -0.00000344, -0.00004410, -0.00004337, -0.00000296, 0.00000978, -0.00000266, 0.00002779, 0.00006155, -0.00002310, -0.00016267, -0.00011388, 0.00013592, 0.00023961, -0.00000177, -0.00026992, -0.00018124, 0.00011956, 0.00020036, 0.00001087, -0.00011810, -0.00005512, -0.00000021, -0.00004634, -0.00004240, 0.00005073, 0.00006110, -0.00005008, -0.00010010, -0.00003246, -0.00001301, -0.00008557, -0.00008058, 0.00001565, -0.00003618, -0.00028298, -0.00038510, -0.00010788, 0.00028147, 0.00038325, 0.00020872, 0.00006869, 0.00008514, 0.00011356, 0.00006553, 0.00002609, 0.00005498, 0.00009018, 0.00008072, 0.00006126, 0.00006250, 0.00006151, 0.00004878, 0.00005770, 0.00010243, 0.00014855, 0.00016210, 0.00014149, 0.00009452, 0.00002552, -0.00004772, -0.00009714, -0.00011169, -0.00010159, -0.00007800, -0.00004751, -0.00002189, -0.00001284, -0.00001759, -0.00002326, -0.00002352, -0.00002098, -0.00001757, -0.00001271, -0.00000804, -0.00000607, -0.00000611, -0.00000621, },
		{ -0.00000403, -0.00000165, -0.00000102, -0.00000263, -0.00000395, -0.00000377, -0.00000361, -0.00000391, -0.00000330, -0.00000201, -0.00000175, -0.00000216, -0.00000182, -0.00000176, -0.00000345, -0.00000483, -0.00000340, -0.00000123, -0.00000151, -0.00000271, -0.00000208, -0.00000139, -0.00000288, -0.00000362, -0.00000088, 0.00000164, 0.00000062, -0.00000006, 0.00000263, 0.00000348, -0.00000098, -0.00000448, -0.00000294, -0.00000297, -0.00000762, -0.00000800, -0.00000129, 0.00000046, -0.00000678, -0.00000907, -0.00000193, -0.00000084, -0.00000836, -0.00000426, 0.00001298, 0.00001527, -0.00000383, -0.00001488, -0.00001046, -0.00001770, -0.00003163, -0.00001200, 0.00002845, 0.00002596, -0.00001562, -0.00002382, -0.00000392, -0.00003357, -0.00007285, 0.00001161, 0.00015636, 0.00010107, -0.00016910, -0.00027831, -0.00000914, 0.00028955, 0.00019152, -0.00014915, -0.00025159, -0.00004931, 0.00009586, 0.00003322, -0.00002071, 0.00003273, 0.00001821, -0.00010700, -0.00013472, -0.00001202, 0.00004018, -0.00005503, -0.00009126, -0.00000975, -0.00006119, -0.00032020, -0.00042702, -0.00011183, 0.00033456, 0.00044945, 0.00023069, 0.00004384, 0.00005831, 0.00011834, 0.00009901, 0.00006989, 0.00009298, 0.00010732, 0.00006876, 0.00003355, 0.00005073, 0.00009126, 0.00012106, 0.00015059, 0.00018192, 0.00018034, 0.00012624, 0.00004441, -0.00003335, -0.00009474, -0.00013022, -0.00012841, -0.00009697, -0.00006078, -0.00003494, -0.00001835, -0.00001067, -0.00001489, -0.00002485, -0.00002829, -0.00002254, -0.00001504, -0.00001048, -0.00000758, -0.00000581, -0.00000625, -0.00000755, -0.00000753, -0.00000664, -0.00000576, },
		{ -0.00000077, -0.00000265, -0.00000467, -0.00000499, -0.00000466, -0.00000439, -0.00000311, -0.00000120, -0.00000065, -0.00000157, -0.00000229, -0.00000295, -0.00000454, -0.00000533, -0.00000321, -0.00000029, -0.00000009, -0.00000167, -0.00000210, -0.00000213, -0.00000340, -0.00000332, 0.00000039, 0.00000391, 0.00000321, 0.00000113, 0.00000093, -0.00000083, -0.00000623, -0.00000889, -0.00000478, -0.00000088, -0.00000244, -0.00000342, -0.00000088, -0.00000348, -0.00001248, -0.00001323, -0.00000089, 0.00000776, 0.00000442, 0.00000336, 0.00000808, -0.00000065, -0.00002295, -0.00002844, -0.00000979, 0.00000201, -0.00000177, 0.00000609, 0.00001937, -0.00000381, -0.00004732, -0.00004133, 0.00000854, 0.00002045, -0.00000291, 0.00002333, 0.00006020, -0.00003379, -0.00018997, -0.00012582, 0.00017495, 0.00029872, 0.00000035, -0.00033991, -0.00024467, 0.00012709, 0.00024443, 0.00002656, -0.00012868, -0.00005669, -0.00000725, -0.00009126, -0.00009436, 0.00004396, 0.00007207, -0.00008694, -0.00016173, -0.00004996, -0.00004904, -0.00030056, -0.00043436, -0.00012349, 0.00034724, 0.00046909, 0.00022789, 0.00003303, 0.00008014, 0.00017934, 0.00016134, 0.00009687, 0.00008089, 0.00007642, 0.00004695, 0.00004981, 0.00012206, 0.00020382, 0.00023084, 0.00020878, 0.00016045, 0.00008096, -0.00002240, -0.00010786, -0.00014519, -0.00014311, -0.00011954, -0.00008163, -0.00004197, -0.00001995, -0.00001826, -0.00002161, -0.00002147, -0.00002223, -0.00002438, -0.00002136, -0.00001327, -0.00000791, -0.00000797, -0.00000890, -0.00000827, -0.00000773, -0.00000749, -0.00000626, -0.00000484, -0.00000429, -0.00000332, -0.00000138, },
		{ -0.00000450, -0.00000496, -0.00000436, -0.00000385, -0.00000276, -0.00000126, -0.00000112, -0.00000245, -0.00000335, -0.00000349, -0.00000403, -0.00000410, -0.00000207, 0.00000030, -0.00000007, -0.00000219, -0.00000283, -0.00000217, -0.00000201, -0.00000090, 0.00000262, 0.00000488, 0.00000222, -0.00000222, -0.00000412, -0.00000537, -0.00000801, -0.00000748, -0.00000146, 0.00000259, -0.00000085, -0.00000541, -0.00000567, -0.00000707, -0.00001104, -0.00000666, 0.00000764, 0.00001404, 0.00000370, -0.00000758, -0.00000959, -0.00001429, -0.00002310, -0.00001499, 0.00000994, 0.00001857, 0.00000137, -0.00001081, -0.00000949, -0.00002200, -0.00003834, -0.00001125, 0.00004087, 0.00003871, -0.00001551, -0.00003320, -0.00001342, -0.00004472, -0.00008424, 0.00002137, 0.00019636, 0.00012832, -0.00020609, -0.00035250, -0.00003230, 0.00034195, 0.00023853, -0.00017247, -0.00029956, -0.00005322, 0.00011060, 0.00000343, -0.00007435, 0.00001852, 0.00002109, -0.00015681, -0.00020577, -0.00001656, 0.00003318, -0.00026661, -0.00049481, -0.00021627, 0.00030834, 0.00049894, 0.00028723, 0.00009799, 0.00014913, 0.00023364, 0.00016848, 0.00005670, 0.00003852, 0.00008277, 0.00012123, 0.00017612, 0.00026006, 0.00029904, 0.00023571, 0.00011059, -0.00000969, -0.00010514, -0.00017050, -0.00018415, -0.00014470, -0.00008913, -0.00004971, -0.00002755, -0.00001714, -0.00002108, -0.00003265, -0.00003462, -0.00002363, -0.00001350, -0.00001169, -0.00001184, -0.00000989, -0.00000974, -0.00001204, -0.00001179, -0.00000804, -0.00000505, -0.00000451, -0.00000450, -0.00000454, -0.00000508, -0.00000455, -0.00000220, -0.00000077, -0.00000227, },
		{ -0.00000455, -0.00000424, -0.00000360, -0.00000233, -0.00000189, -0.00000263, -0.00000292, -0.00000248, -0.00000270, -0.00000322, -0.00000226, -0.00000066, -0.00000087, -0.00000213, -0.00000173, -0.00000010, 0.00000043, 0.00000055, 0.00000179, 0.00000173, -0.00000211, -0.00000631, -0.00000661, -0.00000511, -0.00000515, -0.00000414, -0.00000013, 0.00000119, -0.00000378, -0.00000793, -0.00000547, -0.00000237, -0.00000312, -0.00000066, 0.00000664, 0.00000534, -0.00000923, -0.00001945, -0.00001408, -0.00000648, -0.00000556, 0.00000110, 0.00001368, 0.00000838, -0.00001636, -0.00002676, -0.00001272, -0.00000498, -0.00000895, 0.00000706, 0.00003097, 0.00000735, -0.00004831, -0.00005053, 0.00000210, 0.00001728, -0.00000510, 0.00003070, 0.00007696, -0.00003678, -0.00023114, -0.00016223, 0.00020109, 0.00035907, -0.00000102, -0.00042112, -0.00030139, 0.00016649, 0.00030107, -0.00000622, -0.00021678, -0.00009505, -0.00000308, -0.00013014, -0.00015343, 0.00005967, 0.00008123, -0.00032713, -0.00065253, -0.00034231, 0.00032563, 0.00063597, 0.00043241, 0.00017531, 0.00015713, 0.00019937, 0.00012099, 0.00004013, 0.00010077, 0.00023033, 0.00030161, 0.00031234, 0.00029453, 0.00021244, 0.00004985, -0.00011622, -0.00020466, -0.00021586, -0.00018588, -0.00012878, -0.00005998, -0.00001544, -0.00001181, -0.00002647, -0.00003562, -0.00003895, -0.00003834, -0.00002688, -0.00000827, 0.00000057, -0.00000569, -0.00001436, -0.00001601, -0.00001413, -0.00001231, -0.00000851, -0.00000344, -0.00000187, -0.00000433, -0.00000666, -0.00000708, -0.00000658, -0.00000473, -0.00000123, 0.00000089, -0.00000076, -0.00000380, -0.00000493, },
		{ -0.00000433, -0.00000262, -0.00000133, -0.00000143, -0.00000174, -0.00000180, -0.00000267, -0.00000390, -0.00000334, -0.00000122, -0.00000000, 0.00000003, 0.00000068, 0.00000141, 0.00000024, -0.00000181, -0.00000236, -0.00000262, -0.00000475, -0.00000646, -0.00000480, -0.00000249, -0.00000318, -0.00000429, -0.00000249, -0.00000123, -0.00000350, -0.00000413, 0.00000037, 0.00000279, -0.00000209, -0.00000661, -0.00000523, -0.00000631, -0.00001374, -0.00001442, -0.00000203, 0.00000723, 0.00000251, -0.00000294, -0.00000122, -0.00000659, -0.00002002, -0.00001680, 0.00000496, 0.00001122, -0.00000565, -0.00001063, 0.00000066, -0.00001056, -0.00003574, -0.00001539, 0.00003749, 0.00003429, -0.00002506, -0.00004029, -0.00001191, -0.00004696, -0.00009677, 0.00002455, 0.00023117, 0.00014369, -0.00027170, -0.00045154, -0.00003766, 0.00044748, 0.00030238, -0.00026324, -0.00044295, -0.00008548, 0.00016542, 0.00000034, -0.00013662, 0.00001530, 0.00000773, -0.00044829, -0.00078305, -0.00036759, 0.00045869, 0.00082797, 0.00053304, 0.00014185, 0.00007236, 0.00016593, 0.00018598, 0.00020007, 0.00031748, 0.00042846, 0.00039094, 0.00024263, 0.00008771, -0.00005748, -0.00019823, -0.00027618, -0.00024766, -0.00015984, -0.00008318, -0.00003451, -0.00000628, -0.00000667, -0.00003262, -0.00005278, -0.00004746, -0.00003093, -0.00001924, -0.00000930, 0.00000022, -0.00000089, -0.00001319, -0.00002152, -0.00001767, -0.00000985, -0.00000589, -0.00000409, -0.00000283, -0.00000460, -0.00000849, -0.00000972, -0.00000740, -0.00000456, -0.00000200, 0.00000100, 0.00000226, -0.00000039, -0.00000444, -0.00000610, -0.00000563, -0.00000511, },
		{ -0.00000015, -0.00000047, -0.00000170, -0.00000263, -0.00000370, -0.00000443, -0.00000297, 0.00000030, 0.00000237, 0.00000209, 0.00000098, -0.00000055, -0.00000336, -0.00000587, -0.00000559, -0.00000378, -0.00000349, -0.00000393, -0.00000287, -0.00000218, -0.00000425, -0.00000576, -0.00000286, 0.00000107, 0.00000142, 0.00000070, 0.00000146, -0.00000154, -0.00001077, -0.00001658, -0.00001183, -0.00000487, -0.00000378, -0.00000167, 0.00000578, 0.00000625, -0.00000610, -0.00001470, -0.00000889, -0.00000304, -0.00000570, -0.00000212, 0.00000863, 0.00000148, -0.00002298, -0.00002630, -0.00000153, 0.00000922, -0.00000316, 0.00000276, 0.00002166, -0.00000649, -0.00006553, -0.00005912, 0.00001163, 0.00003390, 0.00000349, 0.00003520, 0.00007654, -0.00007448, -0.00031601, -0.00021517, 0.00026902, 0.00048448, -0.00000197, -0.00059472, -0.00044960, 0.00021819, 0.00043632, -0.00001935, -0.00036112, -0.00016421, -0.00002730, -0.00042620, -0.00077521, -0.00033441, 0.00054777, 0.00090069, 0.00051180, 0.00007495, 0.00009517, 0.00035000, 0.00047379, 0.00047056, 0.00046735, 0.00040004, 0.00018190, -0.00008139, -0.00023740, -0.00028096, -0.00027692, -0.00022868, -0.00012919, -0.00003596, -0.00000480, -0.00001738, -0.00003105, -0.00004052, -0.00005136, -0.00004832, -0.00002560, -0.00000610, -0.00000552, -0.00001100, -0.00001032, -0.00001052, -0.00001614, -0.00001726, -0.00000940, -0.00000276, -0.00000419, -0.00000796, -0.00000884, -0.00000902, -0.00000952, -0.00000740, -0.00000313, -0.00000060, -0.00000010, 0.00000082, 0.00000114, -0.00000166, -0.00000571, -0.00000717, -0.00000614, -0.00000503, -0.00000387, -0.00000178, },
		{ -0.00000280, -0.00000366, -0.00000364, -0.00000301, -0.00000082, 0.00000224, 0.00000321, 0.00000103, -0.00000213, -0.00000468, -0.00000681, -0.00000748, -0.00000521, -0.00000196, -0.00000128, -0.00000270, -0.00000328, -0.00000333, -0.00000438, -0.00000389, 0.00000057, 0.00000468, 0.00000307, -0.00000191, -0.00000578, -0.00001043, -0.00001671, -0.00001661, -0.00000576, 0.00000504, 0.00000528, 0.00000086, -0.00000023, -0.00000375, -0.00001276, -0.00001389, -0.00000114, 0.00000785, 0.00000153, -0.00000558, -0.00000506, -0.00001194, -0.00002489, -0.00001494, 0.00001658, 0.00002439, -0.00000312, -0.00002086, -0.00001503, -0.00002607, -0.00004650, -0.00001190, 0.00005793, 0.00005399, -0.00002677, -0.00006198, -0.00004406, -0.00009396, -0.00014566, 0.00003263, 0.00032215, 0.00021291, -0.00036236, -0.00064501, -0.00009255, 0.00061751, 0.00044384, -0.00040800, -0.00072781, -0.00016467, 0.00024498, -0.00023183, -0.00083746, -0.00052612, 0.00042565, 0.00090639, 0.00061283, 0.00027911, 0.00042650, 0.00073599, 0.00074830, 0.00050188, 0.00024831, 0.00002361, -0.00022447, -0.00039751, -0.00038324, -0.00025047, -0.00013697, -0.00007142, -0.00002111, -0.00000022, -0.00002820, -0.00006420, -0.00006319, -0.00003960, -0.00002514, -0.00001717, -0.00000589, -0.00000397, -0.00001742, -0.00002626, -0.00001758, -0.00000587, -0.00000478, -0.00000737, -0.00000601, -0.00000644, -0.00001210, -0.00001522, -0.00001127, -0.00000599, -0.00000371, -0.00000202, -0.00000002, -0.00000031, -0.00000206, -0.00000178, -0.00000054, -0.00000216, -0.00000574, -0.00000706, -0.00000564, -0.00000412, -0.00000293, -0.00000111, 0.00000009, -0.00000090, },
		{ -0.00000169, -0.00000064, 0.00000050, 0.00000157, 0.00000064, -0.00000269, -0.00000587, -0.00000711, -0.00000705, -0.00000590, -0.00000311, -0.00000048, -0.00000079, -0.00000297, -0.00000352, -0.00000232, -0.00000162, -0.00000065, 0.00000195, 0.00000261, -0.00000238, -0.00000922, -0.00001201, -0.00001202, -0.00001206, -0.00000804, 0.00000202, 0.00000840, 0.00000304, -0.00000583, -0.00000747, -0.00000596, -0.00000743, -0.00000490, 0.00000470, 0.00000677, -0.00000577, -0.00001529, -0.00001053, -0.00000653, -0.00000989, -0.00000196, 0.00001591, 0.00000924, -0.00002447, -0.00003711, -0.00001294, 0.00000278, -0.00000393, 0.00000961, 0.00003337, -0.00000524, -0.00008684, -0.00009192, -0.00001428, 0.00001520, -0.00000707, 0.00005649, 0.00012543, -0.00007876, -0.00043097, -0.00032805, 0.00034018, 0.00068124, -0.00000504, -0.00093895, -0.00077921, 0.00031446, 0.00072277, -0.00027835, -0.00136893, -0.00103758, 0.00035909, 0.00129684, 0.00120537, 0.00084262, 0.00083368, 0.00088837, 0.00057749, 0.00005784, -0.00029059, -0.00041373, -0.00045122, -0.00040724, -0.00023821, -0.00004995, 0.00002010, -0.00001020, -0.00004360, -0.00005736, -0.00007387, -0.00007441, -0.00003784, 0.00000171, 0.00000567, -0.00001143, -0.00002017, -0.00002301, -0.00002866, -0.00002463, -0.00000532, 0.00000874, 0.00000266, -0.00001078, -0.00001614, -0.00001617, -0.00001653, -0.00001301, -0.00000417, 0.00000194, 0.00000103, -0.00000170, -0.00000250, -0.00000345, -0.00000457, -0.00000296, -0.00000023, -0.00000110, -0.00000488, -0.00000654, -0.00000498, -0.00000333, -0.00000262, -0.00000159, -0.00000087, -0.00000185, -0.00000324, -0.00000299, },
		{ -0.00000141, -0.00000217, -0.00000340, -0.00000556, -0.00000684, -0.00000618, -0.00000487, -0.00000373, -0.00000202, -0.00000041, -0.00000077, -0.00000213, -0.00000170, -0.00000004, -0.00000010, -0.00000163, -0.00000247, -0.00000395, -0.00000811, -0.00001142, -0.00000951, -0.00000543, -0.00000414, -0.00000306, 0.00000124, 0.00000308, -0.00000272, -0.00000850, -0.00000563, -0.00000024, -0.00000112, -0.00000312, -0.00000037, -0.00000127, -0.00001009, -0.00001253, -0.00000201, 0.00000331, -0.00000594, -0.00001050, -0.00000266, -0.00000557, -0.00002228, -0.00001747, 0.00001355, 0.00002178, -0.00000848, -0.00002810, -0.00002198, -0.00003741, -0.00006517, -0.00002706, 0.00005552, 0.00005223, -0.00004006, -0.00007138, -0.00003991, -0.00011166, -0.00020182, 0.00002019, 0.00042348, 0.00028324, -0.00057474, -0.00106721, -0.00023338, 0.00101989, 0.00079882, -0.00104210, -0.00233247, -0.00133291, 0.00101438, 0.00244202, 0.00216000, 0.00123917, 0.00068882, 0.00038074, -0.00009609, -0.00057049, -0.00068042, -0.00048865, -0.00028851, -0.00015880, -0.00003068, 0.00005327, 0.00001444, -0.00008117, -0.00011358, -0.00007991, -0.00004687, -0.00002579, 0.00000256, 0.00001662, -0.00000665, -0.00003743, -0.00003932, -0.00002323, -0.00001374, -0.00000686, 0.00000516, 0.00000779, -0.00000800, -0.00002422, -0.00002390, -0.00001440, -0.00000801, -0.00000335, 0.00000238, 0.00000352, -0.00000186, -0.00000636, -0.00000581, -0.00000431, -0.00000398, -0.00000203, 0.00000077, -0.00000018, -0.00000430, -0.00000619, -0.00000460, -0.00000324, -0.00000336, -0.00000292, -0.00000176, -0.00000154, -0.00000175, -0.00000087, 0.00000018, -0.00000024, },
		{ -0.00000527, -0.00000501, -0.00000472, -0.00000371, -0.00000296, -0.00000282, -0.00000190, -0.00000018, 0.00000042, -0.00000027, -0.00000049, -0.00000076, -0.00000324, -0.00000675, -0.00000794, -0.00000715, -0.00000694, -0.00000616, -0.00000273, -0.00000013, -0.00000246, -0.00000572, -0.00000411, -0.00000106, -0.00000243, -0.00000428, -0.00000093, 0.00000186, -0.00000276, -0.00000790, -0.00000527, -0.00000158, -0.00000401, -0.00000450, 0.00000102, -0.00000173, -0.00001694, -0.00002196, -0.00000623, 0.00000592, 0.00000077, 0.00000054, 0.00001035, -0.00000406, -0.00004478, -0.00005707, -0.00002336, 0.00000141, -0.00000284, 0.00001189, 0.00003381, -0.00002122, -0.00012297, -0.00011876, -0.00000516, 0.00003698, -0.00000687, 0.00005448, 0.00012960, -0.00017699, -0.00072053, -0.00058784, 0.00051874, 0.00121749, -0.00001065, -0.00231699, -0.00292679, -0.00057323, 0.00268883, 0.00391633, 0.00262164, 0.00069357, -0.00031861, -0.00055632, -0.00070004, -0.00078362, -0.00055550, -0.00017307, 0.00002047, 0.00000541, -0.00002527, -0.00002949, -0.00007614, -0.00012665, -0.00009599, -0.00001906, 0.00001306, -0.00000148, -0.00001115, -0.00001433, -0.00003227, -0.00004421, -0.00002522, 0.00000126, 0.00000406, -0.00000622, -0.00000818, -0.00000873, -0.00001852, -0.00002454, -0.00001483, -0.00000153, 0.00000133, -0.00000132, -0.00000129, -0.00000193, -0.00000611, -0.00000793, -0.00000467, -0.00000209, -0.00000268, -0.00000202, 0.00000065, 0.00000035, -0.00000357, -0.00000590, -0.00000493, -0.00000419, -0.00000466, -0.00000365, -0.00000095, 0.00000066, 0.00000052, 0.00000009, -0.00000086, -0.00000326, -0.00000561, -0.00000600, },
		{ -0.00000203, -0.00000221, -0.00000244, -0.00000261, -0.00000147, 0.00000072, 0.00000140, -0.00000038, -0.00000281, -0.00000508, -0.00000795, -0.00000989, -0.00000809, -0.00000378, -0.00000107, -0.00000047, -0.00000002, -0.00000121, -0.00000588, -0.00000925, -0.00000580, 0.00000033, 0.00000136, -0.00000113, -0.00000128, -0.00000151, -0.00000602, -0.00000848, -0.00000263, 0.00000315, -0.00000039, -0.00000639, -0.00000733, -0.00001093, -0.00001921, -0.00001431, 0.00000816, 0.00001992, 0.00000412, -0.00001529, -0.00002069, -0.00003086, -0.00004747, -0.00003214, 0.00001642, 0.00003537, 0.00000078, -0.00003088, -0.00003762, -0.00006793, -0.00010075, -0.00003580, 0.00008979, 0.00008805, -0.00006021, -0.00014596, -0.00013699, -0.00024750, -0.00036387, 0.00001467, 0.00072494, 0.00050498, -0.00132516, -0.00310803, -0.00223463, 0.00139867, 0.00464001, 0.00456578, 0.00174493, -0.00088836, -0.00159759, -0.00107467, -0.00059201, -0.00039055, -0.00014408, 0.00010124, 0.00009055, -0.00009111, -0.00016818, -0.00010869, -0.00006221, -0.00005610, -0.00001982, 0.00002498, 0.00000877, -0.00004148, -0.00005248, -0.00002831, -0.00001628, -0.00001502, -0.00000130, 0.00000843, -0.00000783, -0.00002854, -0.00002570, -0.00001122, -0.00000758, -0.00000894, -0.00000267, 0.00000283, -0.00000241, -0.00000923, -0.00000754, -0.00000341, -0.00000404, -0.00000505, -0.00000268, -0.00000161, -0.00000355, -0.00000300, 0.00000094, 0.00000169, -0.00000280, -0.00000662, -0.00000652, -0.00000542, -0.00000474, -0.00000227, 0.00000149, 0.00000273, 0.00000051, -0.00000253, -0.00000506, -0.00000742, -0.00000836, -0.00000648, -0.00000345, -0.00000198, },
		{ -0.00000249, -0.00000221, -0.00000100, 0.00000051, -0.00000012, -0.00000346, -0.00000690, -0.00000851, -0.00000880, -0.00000748, -0.00000342, 0.00000129, 0.00000251, 0.00000019, -0.00000232, -0.00000456, -0.00000779, -0.00000876, -0.00000382, 0.00000238, 0.00000230, -0.00000216, -0.00000379, -0.00000308, -0.00000451, -0.00000492, -0.00000043, 0.00000168, -0.00000530, -0.00001255, -0.00001054, -0.00000652, -0.00000714, -0.00000224, 0.00001041, 0.00000849, -0.00001707, -0.00003708, -0.00003012, -0.00001699, -0.00001400, -0.00000109, 0.00002134, 0.00000834, -0.00004481, -0.00007108, -0.00004315, -0.00001956, -0.00001939, 0.00001455, 0.00005384, -0.00002784, -0.00019740, -0.00023049, -0.00008461, 0.00000303, -0.00001903, 0.00008850, 0.00014764, -0.00058947, -0.00208014, -0.00255052, -0.00028845, 0.00357349, 0.00551820, 0.00357349, -0.00028845, -0.00255052, -0.00208014, -0.00058947, 0.00014764, 0.00008850, -0.00001903, 0.00000303, -0.00008461, -0.00023049, -0.00019740, -0.00002784, 0.00005384, 0.00001455, -0.00001939, -0.00001956, -0.00004315, -0.00007108, -0.00004481, 0.00000834, 0.00002134, -0.00000109, -0.00001400, -0.00001699, -0.00003012, -0.00003708, -0.00001707, 0.00000849, 0.00001041, -0.00000224, -0.00000714, -0.00000652, -0.00001054, -0.00001255, -0.00000530, 0.00000168, -0.00000043, -0.00000492, -0.00000451, -0.00000308, -0.00000379, -0.00000216, 0.00000230, 0.00000238, -0.00000382, -0.00000876, -0.00000779, -0.00000456, -0.00000232, 0.00000019, 0.00000251, 0.00000129, -0.00000342, -0.00000748, -0.00000880, -0.00000851, -0.00000690, -0.00000346, -0.00000012, 0.00000051, -0.00000100, -0.00000221, },
		{ -0.00000203, -0.00000198, -0.00000345, -0.00000648, -0.00000836, -0.00000742, -0.00000506, -0.00000253, 0.00000051, 0.00000273, 0.00000149, -0.00000227, -0.00000474, -0.00000542, -0.00000652, -0.00000662, -0.00000280, 0.00000169, 0.00000094, -0.00000300, -0.00000355, -0.00000161, -0.00000268, -0.00000505, -0.00000404, -0.00000341, -0.00000754, -0.00000923, -0.00000241, 0.00000283, -0.00000267, -0.00000894, -0.00000758, -0.00001122, -0.00002570, -0.00002854, -0.00000783, 0.00000843, -0.00000130, -0.00001502, -0.00001628, -0.00002831, -0.00005248, -0.00004148, 0.00000877, 0.00002498, -0.00001982, -0.00005610, -0.00006221, -0.00010869, -0.00016818, -0.00009111, 0.00009055, 0.00010124, -0.00014408, -0.00039055, -0.00059201, -0.00107467, -0.00159759, -0.00088836, 0.00174493, 0.00456578, 0.00464001, 0.00139867, -0.00223463, -0.00310803, -0.00132516, 0.00050498, 0.00072494, 0.00001467, -0.00036387, -0.00024750, -0.00013699, -0.00014596, -0.00006021, 0.00008805, 0.00008979, -0.00003580, -0.00010075, -0.00006793, -0.00003762, -0.00003088, 0.00000078, 0.00003537, 0.00001642, -0.00003214, -0.00004747, -0.00003086, -0.00002069, -0.00001529, 0.00000412, 0.00001992, 0.00000816, -0.00001431, -0.00001921, -0.00001093, -0.00000733, -0.00000639, -0.00000039, 0.00000315, -0.00000263, -0.00000848, -0.00000602, -0.00000151, -0.00000128, -0.00000113, 0.00000136, 0.00000033, -0.00000580, -0.00000925, -0.00000588, -0.00000121, -0.00000002, -0.00000047, -0.00000107, -0.00000378, -0.00000809, -0.00000989, -0.00000795, -0.00000508, -0.00000281, -0.00000038, 0.00000140, 0.00000072, -0.00000147, -0.00000261, -0.00000244, -0.00000221, },
		{ -0.00000527, -0.00000600, -0.00000561, -0.00000326, -0.00000086, 0.00000009, 0.00000052, 0.00000066, -0.00000095, -0.00000365, -0.00000466, -0.00000419, -0.00000493, -0.00000590, -0.00000357, 0.00000035, 0.00000065, -0.00000202, -0.00000268, -0.00000209, -0.00000467, -0.00000793, -0.00000611, -0.00000193, -0.00000129, -0.00000132, 0.00000133, -0.00000153, -0.00001483, -0.00002454, -0.00001852, -0.00000873, -0.00000818, -0.00000622, 0.00000406, 0.00000126, -0.00002522, -0.00004421, -0.00003227, -0.00001433, -0.00001115, -0.00000148, 0.00001306, -0.00001906, -0.00009599, -0.00012665, -0.00007614, -0.00002949, -0.00002527, 0.00000541, 0.00002047, -0.00017307, -0.00055550, -0.00078362, -0.00070004, -0.00055632, -0.00031861, 0.00069357, 0.00262164, 0.00391633, 0.00268883, -0.00057323, -0.00292679, -0.00231699, -0.00001065, 0.00121749, 0.00051874, -0.00058784, -0.00072053, -0.00017699, 0.00012960, 0.00005448, -0.00000687, 0.00003698, -0.00000516, -0.00011876, -0.00012297, -0.00002122, 0.00003381, 0.00001189, -0.00000284, 0.00000141, -0.00002336, -0.00005707, -0.00004478, -0.00000406, 0.00001035, 0.00000054, 0.00000077, 0.00000592, -0.00000623, -0.00002196, -0.00001694, -0.00000173, 0.00000102, -0.00000450, -0.00000401, -0.00000158, -0.00000527, -0.00000790, -0.00000276, 0.00000186, -0.00000093, -0.00000428, -0.00000243, -0.00000106, -0.00000411, -0.00000572, -0.00000246, -0.00000013, -0.00000273, -0.00000616, -0.00000694, -0.00000715, -0.00000794, -0.00000675, -0.00000324, -0.00000076, -0.00000049, -0.00000027, 0.00000042, -0.00000018, -0.00000190, -0.00000282, -0.00000296, -0.00000371, -0.00000472, -0.00000501, },
		{ -0.00000141, -0.00000024, 0.00000018, -0.00000087, -0.00000175, -0.00000154, -0.00000176, -0.00000292, -0.00000336, -0.00000324, -0.00000460, -0.00000619, -0.00000430, -0.00000018, 0.00000077, -0.00000203, -0.00000398, -0.00000431, -0.00000581, -0.00000636, -0.00000186, 0.00000352, 0.00000238, -0.00000335, -0.00000801, -0.00001440, -0.00002390, -0.00002422, -0.00000800, 0.00000779, 0.00000516, -0.00000686, -0.00001374, -0.00002323, -0.00003932, -0.00003743, -0.00000665, 0.00001662, 0.00000256, -0.00002579, -0.00004687, -0.00007991, -0.00011358, -0.00008117, 0.00001444, 0.00005327, -0.00003068, -0.00015880, -0.00028851, -0.00048865, -0.00068042, -0.00057049, -0.00009609, 0.00038074, 0.00068882, 0.00123917, 0.00216000, 0.00244202, 0.00101438, -0.00133291, -0.00233247, -0.00104210, 0.00079882, 0.00101989, -0.00023338, -0.00106721, -0.00057474, 0.00028324, 0.00042348, 0.00002019, -0.00020182, -0.00011166, -0.00003991, -0.00007138, -0.00004006, 0.00005223, 0.00005552, -0.00002706, -0.00006517, -0.00003741, -0.00002198, -0.00002810, -0.00000848, 0.00002178, 0.00001355, -0.00001747, -0.00002228, -0.00000557, -0.00000266, -0.00001050, -0.00000594, 0.00000331, -0.00000201, -0.00001253, -0.00001009, -0.00000127, -0.00000037, -0.00000312, -0.00000112, -0.00000024, -0.00000563, -0.00000850, -0.00000272, 0.00000308, 0.00000124, -0.00000306, -0.00000414, -0.00000543, -0.00000951, -0.00001142, -0.00000811, -0.00000395, -0.00000247, -0.00000163, -0.00000010, -0.00000004, -0.00000170, -0.00000213, -0.00000077, -0.00000041, -0.00000202, -0.00000373, -0.00000487, -0.00000618, -0.00000684, -0.00000556, -0.00000340, -0.00000217, },
		{ -0.00000169, -0.00000299, -0.00000324, -0.00000185, -0.00000087, -0.00000159, -0.00000262, -0.00000333, -0.00000498, -0.00000654, -0.00000488, -0.00000110, -0.00000023, -0.00000296, -0.00000457, -0.00000345, -0.00000250, -0.00000170, 0.00000103, 0.00000194, -0.00000417, -0.00001301, -0.00001653, -0.00001617, -0.00001614, -0.00001078, 0.00000266, 0.00000874, -0.00000532, -0.00002463, -0.00002866, -0.00002301, -0.00002017, -0.00001143, 0.00000567, 0.00000171, -0.00003784, -0.00007441, -0.00007387, -0.00005736, -0.00004360, -0.00001020, 0.00002010, -0.00004995, -0.00023821, -0.00040724, -0.00045122, -0.00041373, -0.00029059, 0.00005784, 0.00057749, 0.00088837, 0.00083368, 0.00084262, 0.00120537, 0.00129684, 0.00035909, -0.00103758, -0.00136893, -0.00027835, 0.00072277, 0.00031446, -0.00077921, -0.00093895, -0.00000504, 0.00068124, 0.00034018, -0.00032805, -0.00043097, -0.00007876, 0.00012543, 0.00005649, -0.00000707, 0.00001520, -0.00001428, -0.00009192, -0.00008684, -0.00000524, 0.00003337, 0.00000961, -0.00000393, 0.00000278, -0.00001294, -0.00003711, -0.00002447, 0.00000924, 0.00001591, -0.00000196, -0.00000989, -0.00000653, -0.00001053, -0.00001529, -0.00000577, 0.00000677, 0.00000470, -0.00000490, -0.00000743, -0.00000596, -0.00000747, -0.00000583, 0.00000304, 0.00000840, 0.00000202, -0.00000804, -0.00001206, -0.00001202, -0.00001201, -0.00000922, -0.00000238, 0.00000261, 0.00000195, -0.00000065, -0.00000162, -0.00000232, -0.00000352, -0.00000297, -0.00000079, -0.00000048, -0.00000311, -0.00000590, -0.00000705, -0.00000711, -0.00000587, -0.00000269, 0.00000064, 0.00000157, 0.00000050, -0.00000064, },
		{ -0.00000280, -0.00000090, 0.00000009, -0.00000111, -0.00000293, -0.00000412, -0.00000564, -0.00000706, -0.00000574, -0.00000216, -0.00000054, -0.00000178, -0.00000206, -0.00000031, -0.00000002, -0.00000202, -0.00000371, -0.00000599, -0.00001127, -0.00001522, -0.00001210, -0.00000644, -0.00000601, -0.00000737, -0.00000478, -0.00000587, -0.00001758, -0.00002626, -0.00001742, -0.00000397, -0.00000589, -0.00001717, -0.00002514, -0.00003960, -0.00006319, -0.00006420, -0.00002820, -0.00000022, -0.00002111, -0.00007142, -0.00013697, -0.00025047, -0.00038324, -0.00039751, -0.00022447, 0.00002361, 0.00024831, 0.00050188, 0.00074830, 0.00073599, 0.00042650, 0.00027911, 0.00061283, 0.00090639, 0.00042565, -0.00052612, -0.00083746, -0.00023183, 0.00024498, -0.00016467, -0.00072781, -0.00040800, 0.00044384, 0.00061751, -0.00009255, -0.00064501, -0.00036236, 0.00021291, 0.00032215, 0.00003263, -0.00014566, -0.00009396, -0.00004406, -0.00006198, -0.00002677, 0.00005399, 0.00005793, -0.00001190, -0.00004650, -0.00002607, -0.00001503, -0.00002086, -0.00000312, 0.00002439, 0.00001658, -0.00001494, -0.00002489, -0.00001194, -0.00000506, -0.00000558, 0.00000153, 0.00000785, -0.00000114, -0.00001389, -0.00001276, -0.00000375, -0.00000023, 0.00000086, 0.00000528, 0.00000504, -0.00000576, -0.00001661, -0.00001671, -0.00001043, -0.00000578, -0.00000191, 0.00000307, 0.00000468, 0.00000057, -0.00000389, -0.00000438, -0.00000333, -0.00000328, -0.00000270, -0.00000128, -0.00000196, -0.00000521, -0.00000748, -0.00000681, -0.00000468, -0.00000213, 0.00000103, 0.00000321, 0.00000224, -0.00000082, -0.00000301, -0.00000364, -0.00000366, },
		{ -0.00000015, -0.00000178, -0.00000387, -0.00000503, -0.00000614, -0.00000717, -0.00000571, -0.00000166, 0.00000114, 0.00000082, -0.00000010, -0.00000060, -0.00000313, -0.00000740, -0.00000952, -0.00000902, -0.00000884, -0.00000796, -0.00000419, -0.00000276, -0.00000940, -0.00001726, -0.00001614, -0.00001052, -0.00001032, -0.00001100, -0.00000552, -0.00000610, -0.00002560, -0.00004832, -0.00005136, -0.00004052, -0.00003105, -0.00001738, -0.00000480, -0.00003596, -0.00012919, -0.00022868, -0.00027692, -0.00028096, -0.00023740, -0.00008139, 0.00018190, 0.00040004, 0.00046735, 0.00047056, 0.00047379, 0.00035000, 0.00009517, 0.00007495, 0.00051180, 0.00090069, 0.00054777, -0.00033441, -0.00077521, -0.00042620, -0.00002730, -0.00016421, -0.00036112, -0.00001935, 0.00043632, 0.00021819, -0.00044960, -0.00059472, -0.00000197, 0.00048448, 0.00026902, -0.00021517, -0.00031601, -0.00007448, 0.00007654, 0.00003520, 0.00000349, 0.00003390, 0.00001163, -0.00005912, -0.00006553, -0.00000649, 0.00002166, 0.00000276, -0.00000316, 0.00000922, -0.00000153, -0.00002630, -0.00002298, 0.00000148, 0.00000863, -0.00000212, -0.00000570, -0.00000304, -0.00000889, -0.00001470, -0.00000610, 0.00000625, 0.00000578, -0.00000167, -0.00000378, -0.00000487, -0.00001183, -0.00001658, -0.00001077, -0.00000154, 0.00000146, 0.00000070, 0.00000142, 0.00000107, -0.00000286, -0.00000576, -0.00000425, -0.00000218, -0.00000287, -0.00000393, -0.00000349, -0.00000378, -0.00000559, -0.00000587, -0.00000336, -0.00000055, 0.00000098, 0.00000209, 0.00000237, 0.00000030, -0.00000297, -0.00000443, -0.00000370, -0.00000263, -0.00000170, -0.00000047, },
		{ -0.00000433, -0.00000511, -0.00000563, -0.00000610, -0.00000444, -0.00000039, 0.00000226, 0.00000100, -0.00000200, -0.00000456, -0.00000740, -0.00000972, -0.00000849, -0.00000460, -0.00000283, -0.00000409, -0.00000589, -0.00000985, -0.00001767, -0.00002152, -0.00001319, -0.00000089, 0.00000022, -0.00000930, -0.00001924, -0.00003093, -0.00004746, -0.00005278, -0.00003262, -0.00000667, -0.00000628, -0.00003451, -0.00008318, -0.00015984, -0.00024766, -0.00027618, -0.00019823, -0.00005748, 0.00008771, 0.00024263, 0.00039094, 0.00042846, 0.00031748, 0.00020007, 0.00018598, 0.00016593, 0.00007236, 0.00014185, 0.00053304, 0.00082797, 0.00045869, -0.00036759, -0.00078305, -0.00044829, 0.00000773, 0.00001530, -0.00013662, 0.00000034, 0.00016542, -0.00008548, -0.00044295, -0.00026324, 0.00030238, 0.00044748, -0.00003766, -0.00045154, -0.00027170, 0.00014369, 0.00023117, 0.00002455, -0.00009677, -0.00004696, -0.00001191, -0.00004029, -0.00002506, 0.00003429, 0.00003749, -0.00001539, -0.00003574, -0.00001056, 0.00000066, -0.00001063, -0.00000565, 0.00001122, 0.00000496, -0.00001680, -0.00002002, -0.00000659, -0.00000122, -0.00000294, 0.00000251, 0.00000723, -0.00000203, -0.00001442, -0.00001374, -0.00000631, -0.00000523, -0.00000661, -0.00000209, 0.00000279, 0.00000037, -0.00000413, -0.00000350, -0.00000123, -0.00000249, -0.00000429, -0.00000318, -0.00000249, -0.00000480, -0.00000646, -0.00000475, -0.00000262, -0.00000236, -0.00000181, 0.00000024, 0.00000141, 0.00000068, 0.00000003, -0.00000000, -0.00000122, -0.00000334, -0.00000390, -0.00000267, -0.00000180, -0.00000174, -0.00000143, -0.00000133, -0.00000262, },
		{ -0.00000455, -0.00000493, -0.00000380, -0.00000076, 0.00000089, -0.00000123, -0.00000473, -0.00000658, -0.00000708, -0.00000666, -0.00000433, -0.00000187, -0.00000344, -0.00000851, -0.00001231, -0.00001413, -0.00001601, -0.00001436, -0.00000569, 0.00000057, -0.00000827, -0.00002688, -0.00003834, -0.00003895, -0.00003562, -0.00002647, -0.00001181, -0.00001544, -0.00005998, -0.00012878, -0.00018588, -0.00021586, -0.00020466, -0.00011622, 0.00004985, 0.00021244, 0.00029453, 0.00031234, 0.00030161, 0.00023033, 0.00010077, 0.00004013, 0.00012099, 0.00019937, 0.00015713, 0.00017531, 0.00043241, 0.00063597, 0.00032563, -0.00034231, -0.00065253, -0.00032713, 0.00008123, 0.00005967, -0.00015343, -0.00013014, -0.00000308, -0.00009505, -0.00021678, -0.00000622, 0.00030107, 0.00016649, -0.00030139, -0.00042112, -0.00000102, 0.00035907, 0.00020109, -0.00016223, -0.00023114, -0.00003678, 0.00007696, 0.00003070, -0.00000510, 0.00001728, 0.00000210, -0.00005053, -0.00004831, 0.00000735, 0.00003097, 0.00000706, -0.00000895, -0.00000498, -0.00001272, -0.00002676, -0.00001636, 0.00000838, 0.00001368, 0.00000110, -0.00000556, -0.00000648, -0.00001408, -0.00001945, -0.00000923, 0.00000534, 0.00000664, -0.00000066, -0.00000312, -0.00000237, -0.00000547, -0.00000793, -0.00000378, 0.00000119, -0.00000013, -0.00000414, -0.00000515, -0.00000511, -0.00000661, -0.00000631, -0.00000211, 0.00000173, 0.00000179, 0.00000055, 0.00000043, -0.00000010, -0.00000173, -0.00000213, -0.00000087, -0.00000066, -0.00000226, -0.00000322, -0.00000270, -0.00000248, -0.00000292, -0.00000263, -0.00000189, -0.00000233, -0.00000360, -0.00000424, },
		{ -0.00000450, -0.00000227, -0.00000077, -0.00000220, -0.00000455, -0.00000508, -0.00000454, -0.00000450, -0.00000451, -0.00000505, -0.00000804, -0.00001179, -0.00001204, -0.00000974, -0.00000989, -0.00001184, -0.00001169, -0.00001350, -0.00002363, -0.00003462, -0.00003265, -0.00002108, -0.00001714, -0.00002755, -0.00004971, -0.00008913, -0.00014470, -0.00018415, -0.00017050, -0.00010514, -0.00000969, 0.00011059, 0.00023571, 0.00029904, 0.00026006, 0.00017612, 0.00012123, 0.00008277, 0.00003852, 0.00005670, 0.00016848, 0.00023364, 0.00014913, 0.00009799, 0.00028723, 0.00049894, 0.00030834, -0.00021627, -0.00049481, -0.00026661, 0.00003318, -0.00001656, -0.00020577, -0.00015681, 0.00002109, 0.00001852, -0.00007435, 0.00000343, 0.00011060, -0.00005322, -0.00029956, -0.00017247, 0.00023853, 0.00034195, -0.00003230, -0.00035250, -0.00020609, 0.00012832, 0.00019636, 0.00002137, -0.00008424, -0.00004472, -0.00001342, -0.00003320, -0.00001551, 0.00003871, 0.00004087, -0.00001125, -0.00003834, -0.00002200, -0.00000949, -0.00001081, 0.00000137, 0.00001857, 0.00000994, -0.00001499, -0.00002310, -0.00001429, -0.00000959, -0.00000758, 0.00000370, 0.00001404, 0.00000764, -0.00000666, -0.00001104, -0.00000707, -0.00000567, -0.00000541, -0.00000085, 0.00000259, -0.00000146, -0.00000748, -0.00000801, -0.00000537, -0.00000412, -0.00000222, 0.00000222, 0.00000488, 0.00000262, -0.00000090, -0.00000201, -0.00000217, -0.00000283, -0.00000219, -0.00000007, 0.00000030, -0.00000207, -0.00000410, -0.00000403, -0.00000349, -0.00000335, -0.00000245, -0.00000112, -0.00000126, -0.00000276, -0.00000385, -0.00000436, -0.00000496, },
		{ -0.00000077, -0.00000138, -0.00000332, -0.00000429, -0.00000484, -0.00000626, -0.00000749, -0.00000773, -0.00000827, -0.00000890, -0.00000797, -0.00000791, -0.00001327, -0.00002136, -0.00002438, -0.00002223, -0.00002147, -0.00002161, -0.00001826, -0.00001995, -0.00004197, -0.00008163, -0.00011954, -0.00014311, -0.00014519, -0.00010786, -0.00002240, 0.00008096, 0.00016045, 0.00020878, 0.00023084, 0.00020382, 0.00012206, 0.00004981, 0.00004695, 0.00007642, 0.00008089, 0.00009687, 0.00016134, 0.00017934, 0.00008014, 0.00003303, 0.00022789, 0.00046909, 0.00034724, -0.00012349, -0.00043436, -0.00030056, -0.00004904, -0.00004996, -0.00016173, -0.00008694, 0.00007207, 0.00004396, -0.00009436, -0.00009126, -0.00000725, -0.00005669, -0.00012868, 0.00002656, 0.00024443, 0.00012709, -0.00024467, -0.00033991, 0.00000035, 0.00029872, 0.00017495, -0.00012582, -0.00018997, -0.00003379, 0.00006020, 0.00002333, -0.00000291, 0.00002045, 0.00000854, -0.00004133, -0.00004732, -0.00000381, 0.00001937, 0.00000609, -0.00000177, 0.00000201, -0.00000979, -0.00002844, -0.00002295, -0.00000065, 0.00000808, 0.00000336, 0.00000442, 0.00000776, -0.00000089, -0.00001323, -0.00001248, -0.00000348, -0.00000088, -0.00000342, -0.00000244, -0.00000088, -0.00000478, -0.00000889, -0.00000623, -0.00000083, 0.00000093, 0.00000113, 0.00000321, 0.00000391, 0.00000039, -0.00000332, -0.00000340, -0.00000213, -0.00000210, -0.00000167, -0.00000009, -0.00000029, -0.00000321, -0.00000533, -0.00000454, -0.00000295, -0.00000229, -0.00000157, -0.00000065, -0.00000120, -0.00000311, -0.00000439, -0.00000466, -0.00000499, -0.00000467, -0.00000265, },
		{ -0.00000403, -0.00000576, -0.00000664, -0.00000753, -0.00000755, -0.00000625, -0.00000581, -0.00000758, -0.00001048, -0.00001504, -0.00002254, -0.00002829, -0.00002485, -0.00001489, -0.00001067, -0.00001835, -0.00003494, -0.00006078, -0.00009697, -0.00012841, -0.00013022, -0.00009474, -0.00003335, 0.00004441, 0.00012624, 0.00018034, 0.00018192, 0.00015059, 0.00012106, 0.00009126, 0.00005073, 0.00003355, 0.00006876, 0.00010732, 0.00009298, 0.00006989, 0.00009901, 0.00011834, 0.00005831, 0.00004384, 0.00023069, 0.00044945, 0.00033456, -0.00011183, -0.00042702, -0.00032020, -0.00006119, -0.00000975, -0.00009126, -0.00005503, 0.00004018, -0.00001202, -0.00013472, -0.00010700, 0.00001821, 0.00003273, -0.00002071, 0.00003322, 0.00009586, -0.00004931, -0.00025159, -0.00014915, 0.00019152, 0.00028955, -0.00000914, -0.00027831, -0.00016910, 0.00010107, 0.00015636, 0.00001161, -0.00007285, -0.00003357, -0.00000392, -0.00002382, -0.00001562, 0.00002596, 0.00002845, -0.00001200, -0.00003163, -0.00001770, -0.00001046, -0.00001488, -0.00000383, 0.00001527, 0.00001298, -0.00000426, -0.00000836, -0.00000084, -0.00000193, -0.00000907, -0.00000678, 0.00000046, -0.00000129, -0.00000800, -0.00000762, -0.00000297, -0.00000294, -0.00000448, -0.00000098, 0.00000348, 0.00000263, -0.00000006, 0.00000062, 0.00000164, -0.00000088, -0.00000362, -0.00000288, -0.00000139, -0.00000208, -0.00000271, -0.00000151, -0.00000123, -0.00000340, -0.00000483, -0.00000345, -0.00000176, -0.00000182, -0.00000216, -0.00000175, -0.00000201, -0.00000330, -0.00000391, -0.00000361, -0.00000377, -0.00000395, -0.00000263, -0.00000102, -0.00000165, },
		{ -0.00000626, -0.00000621, -0.00000611, -0.00000607, -0.00000804, -0.00001271, -0.00001757, -0.00002098, -0.00002352, -0.00002326, -0.00001759, -0.00001284, -0.00002189, -0.00004751, -0.00007800, -0.00010159, -0.00011169, -0.00009714, -0.00004772, 0.00002552, 0.00009452, 0.00014149, 0.00016210, 0.00014855, 0.00010243, 0.00005770, 0.00004878, 0.00006151, 0.00006250, 0.00006126, 0.00008072, 0.00009018, 0.00005498, 0.00002609, 0.00006553, 0.00011356, 0.00008514, 0.00006869, 0.00020872, 0.00038325, 0.00028147, -0.00010788, -0.00038510, -0.00028298, -0.00003618, 0.00001565, -0.00008058, -0.00008557, -0.00001301, -0.00003246, -0.00010010, -0.00005008, 0.00006110, 0.00005073, -0.00004240, -0.00004634, -0.00000021, -0.00005512, -0.00011810, 0.00001087, 0.00020036, 0.00011956, -0.00018124, -0.00026992, -0.00000177, 0.00023961, 0.00013592, -0.00011388, -0.00016267, -0.00002310, 0.00006155, 0.00002779, -0.00000266, 0.00000978, -0.00000296, -0.00004337, -0.00004410, -0.00000344, 0.00001742, 0.00000635, 0.00000166, 0.00000805, -0.00000046, -0.00001694, -0.00001322, 0.00000428, 0.00000748, -0.00000317, -0.00000722, -0.00000427, -0.00000717, -0.00001202, -0.00000720, 0.00000208, 0.00000369, 0.00000025, 0.00000070, 0.00000269, 0.00000026, -0.00000284, -0.00000106, 0.00000164, -0.00000011, -0.00000322, -0.00000318, -0.00000227, -0.00000345, -0.00000406, -0.00000188, -0.00000025, -0.00000180, -0.00000367, -0.00000324, -0.00000251, -0.00000333, -0.00000372, -0.00000227, -0.00000093, -0.00000115, -0.00000171, -0.00000207, -0.00000332, -0.00000479, -0.00000442, -0.00000289, -0.00000301, -0.00000492, -0.00000621, },
		{ -0.00000829, -0.00001041, -0.00001344, -0.00001711, -0.00001893, -0.00001831, -0.00001842, -0.00002110, -0.00002560, -0.00003510, -0.00005509, -0.00008047, -0.00009411, -0.00008385, -0.00005118, -0.00000123, 0.00005982, 0.00011409, 0.00013897, 0.00013108, 0.00010551, 0.00007218, 0.00003771, 0.00002405, 0.00004659, 0.00007760, 0.00007758, 0.00005708, 0.00005189, 0.00005184, 0.00002997, 0.00002270, 0.00007544, 0.00012776, 0.00009523, 0.00005510, 0.00015543, 0.00031423, 0.00025432, -0.00006627, -0.00031610, -0.00024691, -0.00004469, -0.00000753, -0.00009974, -0.00010099, -0.00001220, 0.00000319, -0.00004658, -0.00002625, 0.00003445, 0.00000092, -0.00008206, -0.00006332, 0.00002021, 0.00001858, -0.00002893, 0.00001881, 0.00008162, -0.00002507, -0.00019032, -0.00011659, 0.00015418, 0.00023106, -0.00001830, -0.00024161, -0.00014347, 0.00009377, 0.00014568, 0.00001836, -0.00006451, -0.00004074, -0.00001845, -0.00003126, -0.00001577, 0.00002777, 0.00003352, -0.00000143, -0.00001953, -0.00000890, -0.00000517, -0.00001189, -0.00000348, 0.00001340, 0.00001023, -0.00000867, -0.00001608, -0.00000970, -0.00000667, -0.00000719, -0.00000047, 0.00000818, 0.00000653, -0.00000096, -0.00000253, 0.00000006, -0.00000097, -0.00000304, -0.00000057, 0.00000230, -0.00000031, -0.00000467, -0.00000498, -0.00000334, -0.00000350, -0.00000324, -0.00000034, 0.00000138, -0.00000127, -0.00000487, -0.00000547, -0.00000446, -0.00000398, -0.00000263, 0.00000049, 0.00000255, 0.00000147, -0.00000109, -0.00000338, -0.00000571, -0.00000741, -0.00000650, -0.00000374, -0.00000252, -0.00000373, -0.00000507, -0.00000558, -0.00000652, },
		{ -0.00001461, -0.00001545, -0.00001593, -0.00001746, -0.00002409, -0.00003680, -0.00005086, -0.00006236, -0.00007065, -0.00007030, -0.00005067, -0.00001018, 0.00003796, 0.00007952, 0.00010803, 0.00011712, 0.00010161, 0.00007170, 0.00004751, 0.00003523, 0.00002905, 0.00003436, 0.00005736, 0.00007533, 0.00006035, 0.00003074, 0.00002691, 0.00004132, 0.00003952, 0.00003991, 0.00008018, 0.00011462, 0.00007403, 0.00002924, 0.00012019, 0.00028239, 0.00025697, -0.00002297, -0.00027339, -0.00024714, -0.00007733, -0.00002750, -0.00008703, -0.00007266, 0.00001447, 0.00002393, -0.00004344, -0.00005544, -0.00001177, -0.00002052, -0.00005891, -0.00001782, 0.00005870, 0.00003737, -0.00004460, -0.00004646, 0.00000059, -0.00003390, -0.00008439, 0.00001374, 0.00016186, 0.00009071, -0.00016086, -0.00023266, -0.00000008, 0.00021222, 0.00012686, -0.00009249, -0.00014574, -0.00003453, 0.00003818, 0.00001715, 0.00000235, 0.00002165, 0.00001218, -0.00002714, -0.00003360, -0.00000287, 0.00001211, 0.00000120, -0.00000126, 0.00000744, 0.00000080, -0.00001694, -0.00001817, -0.00000393, 0.00000268, -0.00000022, 0.00000189, 0.00000706, 0.00000345, -0.00000441, -0.00000432, 0.00000065, 0.00000046, -0.00000268, -0.00000154, 0.00000069, -0.00000213, -0.00000616, -0.00000516, -0.00000185, -0.00000135, -0.00000176, -0.00000014, 0.00000022, -0.00000363, -0.00000755, -0.00000679, -0.00000320, -0.00000045, 0.00000172, 0.00000402, 0.00000397, 0.00000016, -0.00000450, -0.00000713, -0.00000814, -0.00000799, -0.00000563, -0.00000210, -0.00000095, -0.00000322, -0.00000621, -0.00000818, -0.00001009, -0.00001227, -0.00001375, },
		{ -0.00002081, -0.00002956, -0.00004272, -0.00005719, -0.00006437, -0.00005816, -0.00004024, -0.00001358, 0.00002158, 0.00005955, 0.00008653, 0.00009403, 0.00008660, 0.00007131, 0.00005155, 0.00003531, 0.00003236, 0.00003788, 0.00003845, 0.00003690, 0.00004511, 0.00005279, 0.00004025, 0.00002197, 0.00002950, 0.00005107, 0.00004954, 0.00003878, 0.00006073, 0.00008731, 0.00005822, 0.00002682, 0.00011340, 0.00026503, 0.00024820, -0.00001016, -0.00025751, -0.00025007, -0.00008656, -0.00001419, -0.00005264, -0.00004871, 0.00000999, 0.00000274, -0.00006579, -0.00007274, -0.00001289, 0.00000701, -0.00001616, 0.00000356, 0.00004068, -0.00000132, -0.00007842, -0.00006399, 0.00001319, 0.00002331, -0.00001229, 0.00002123, 0.00006557, -0.00002741, -0.00016634, -0.00010230, 0.00013127, 0.00020189, -0.00000939, -0.00020568, -0.00012968, 0.00006994, 0.00011762, 0.00001644, -0.00004550, -0.00001992, -0.00000328, -0.00002224, -0.00001601, 0.00001846, 0.00002316, -0.00000567, -0.00001798, -0.00000495, -0.00000112, -0.00001081, -0.00000817, 0.00000579, 0.00000741, -0.00000235, -0.00000363, 0.00000305, 0.00000249, -0.00000353, -0.00000281, 0.00000182, -0.00000012, -0.00000535, -0.00000459, -0.00000082, -0.00000177, -0.00000457, -0.00000287, 0.00000030, -0.00000077, -0.00000351, -0.00000336, -0.00000278, -0.00000506, -0.00000657, -0.00000313, 0.00000221, 0.00000444, 0.00000387, 0.00000272, 0.00000011, -0.00000454, -0.00000811, -0.00000818, -0.00000656, -0.00000527, -0.00000364, -0.00000183, -0.00000249, -0.00000622, -0.00000993, -0.00001149, -0.00001205, -0.00001262, -0.00001274, -0.00001307, -0.00001549, },
		{ -0.00005314, -0.00005471, -0.00004300, -0.00001646, 0.00001661, 0.00004598, 0.00006772, 0.00007974, 0.00007762, 0.00006263, 0.00004575, 0.00003584, 0.00003200, 0.00003239, 0.00003758, 0.00004070, 0.00003277, 0.00002243, 0.00002702, 0.00004033, 0.00004026, 0.00003119, 0.00003710, 0.00004993, 0.00004019, 0.00002338, 0.00004234, 0.00007564, 0.00006221, 0.00003644, 0.00010450, 0.00023278, 0.00022057, -0.00000664, -0.00023151, -0.00022756, -0.00007329, 0.00000044, -0.00004103, -0.00005744, -0.00001675, -0.00001614, -0.00006035, -0.00005140, 0.00001050, 0.00002535, -0.00001349, -0.00002004, 0.00000314, -0.00002019, -0.00006257, -0.00003010, 0.00004235, 0.00003791, -0.00002423, -0.00003027, 0.00000120, -0.00003266, -0.00007711, 0.00000568, 0.00013381, 0.00007788, -0.00013377, -0.00019620, 0.00000017, 0.00018082, 0.00010843, -0.00007688, -0.00011802, -0.00001990, 0.00003927, 0.00001393, -0.00000381, 0.00001260, 0.00000633, -0.00002561, -0.00002833, 0.00000151, 0.00001511, 0.00000244, -0.00000354, 0.00000366, 0.00000166, -0.00000862, -0.00000654, 0.00000509, 0.00000644, -0.00000184, -0.00000428, -0.00000126, -0.00000346, -0.00000788, -0.00000486, 0.00000161, 0.00000168, -0.00000227, -0.00000259, -0.00000134, -0.00000407, -0.00000745, -0.00000577, -0.00000183, -0.00000078, -0.00000054, 0.00000259, 0.00000558, 0.00000393, -0.00000040, -0.00000316, -0.00000455, -0.00000622, -0.00000671, -0.00000489, -0.00000337, -0.00000423, -0.00000554, -0.00000569, -0.00000652, -0.00000914, -0.00001122, -0.00001111, -0.00001062, -0.00001158, -0.00001385, -0.00001785, -0.00002502, -0.00003487, -0.00004505, },
		{ 0.00000102, 0.00003163, 0.00005781, 0.00006985, 0.00006828, 0.00005914, 0.00004580, 0.00003188, 0.00002472, 0.00002717, 0.00003256, 0.00003490, 0.00003474, 0.00003064, 0.00001990, 0.00001256, 0.00002325, 0.00004310, 0.00004690, 0.00003473, 0.00003098, 0.00003616, 0.00002782, 0.00001725, 0.00003993, 0.00007564, 0.00006529, 0.00003415, 0.00008274, 0.00019659, 0.00020088, 0.00001269, -0.00019012, -0.00020017, -0.00007226, -0.00001014, -0.00005178, -0.00006924, -0.00002384, -0.00000503, -0.00003341, -0.00003081, 0.00000938, 0.00000839, -0.00003425, -0.00003720, 0.00000010, 0.00000062, -0.00002916, -0.00001360, 0.00002749, 0.00000708, -0.00005006, -0.00004229, 0.00001321, 0.00001457, -0.00001937, 0.00000806, 0.00004898, -0.00002306, -0.00013459, -0.00007673, 0.00012174, 0.00017730, -0.00000982, -0.00018054, -0.00011169, 0.00006447, 0.00010472, 0.00001248, -0.00004415, -0.00002116, -0.00000483, -0.00001998, -0.00001234, 0.00002109, 0.00002694, -0.00000021, -0.00001469, -0.00000511, -0.00000041, -0.00000617, -0.00000231, 0.00000877, 0.00000657, -0.00000627, -0.00001011, -0.00000442, -0.00000299, -0.00000519, -0.00000126, 0.00000487, 0.00000271, -0.00000452, -0.00000691, -0.00000551, -0.00000652, -0.00000716, -0.00000196, 0.00000495, 0.00000638, 0.00000387, 0.00000261, 0.00000156, -0.00000240, -0.00000639, -0.00000614, -0.00000343, -0.00000235, -0.00000281, -0.00000320, -0.00000476, -0.00000816, -0.00001031, -0.00000912, -0.00000733, -0.00000782, -0.00000936, -0.00001028, -0.00001205, -0.00001632, -0.00002222, -0.00002883, -0.00003616, -0.00004232, -0.00004360, -0.00003765, -0.00002320, },
		{ 0.00006434, 0.00005919, 0.00004532, 0.00003251, 0.00002459, 0.00002081, 0.00002200, 0.00002731, 0.00003069, 0.00002872, 0.00002511, 0.00002164, 0.00001616, 0.00001460, 0.00002692, 0.00004418, 0.00004399, 0.00002756, 0.00002052, 0.00002737, 0.00002641, 0.00002106, 0.00003999, 0.00006860, 0.00005654, 0.00002501, 0.00006671, 0.00017660, 0.00019672, 0.00003591, -0.00015912, -0.00019152, -0.00008847, -0.00002672, -0.00005195, -0.00005811, -0.00000995, 0.00001001, -0.00002347, -0.00003760, -0.00001237, -0.00000950, -0.00003443, -0.00002362, 0.00001772, 0.00001583, -0.00002607, -0.00003176, -0.00000198, -0.00000705, -0.00003585, -0.00001408, 0.00003589, 0.00002373, -0.00003232, -0.00003622, -0.00000312, -0.00002191, -0.00005113, 0.00002199, 0.00012607, 0.00006704, -0.00012470, -0.00018009, -0.00000386, 0.00015979, 0.00009820, -0.00006654, -0.00010530, -0.00001959, 0.00003331, 0.00001244, -0.00000023, 0.00001858, 0.00001562, -0.00001472, -0.00002272, -0.00000118, 0.00000956, -0.00000003, -0.00000328, 0.00000351, 0.00000008, -0.00001157, -0.00001129, -0.00000047, 0.00000300, -0.00000129, -0.00000087, 0.00000247, -0.00000169, -0.00000971, -0.00001041, -0.00000510, -0.00000214, -0.00000035, 0.00000540, 0.00001061, 0.00000789, 0.00000024, -0.00000438, -0.00000536, -0.00000633, -0.00000618, -0.00000246, 0.00000136, 0.00000046, -0.00000385, -0.00000746, -0.00000985, -0.00001176, -0.00001119, -0.00000745, -0.00000463, -0.00000644, -0.00001111, -0.00001542, -0.00001967, -0.00002508, -0.00003032, -0.00003351, -0.00003410, -0.00003060, -0.00002058, -0.00000425, 0.00001559, 0.00003664, 0.00005531, },
		{ 0.00002562, 0.00001879, 0.00001845, 0.00001987, 0.00002124, 0.00002308, 0.00002343, 0.00002148, 0.00002065, 0.00002131, 0.00001921, 0.00001737, 0.00002501, 0.00003715, 0.00003595, 0.00002202, 0.00001762, 0.00002685, 0.00002812, 0.00002092, 0.00003223, 0.00005565, 0.00004874, 0.00002523, 0.00006450, 0.00016589, 0.00018909, 0.00004338, -0.00014570, -0.00019010, -0.00009713, -0.00002612, -0.00003543, -0.00004064, -0.00000522, 0.00000414, -0.00003343, -0.00004949, -0.00002041, -0.00000385, -0.00001650, -0.00000915, 0.00001732, 0.00000432, -0.00004038, -0.00004374, -0.00000358, 0.00000955, -0.00001008, -0.00000306, 0.00002096, -0.00000301, -0.00005201, -0.00004129, 0.00001365, 0.00002382, 0.00000048, 0.00002204, 0.00004744, -0.00002724, -0.00013121, -0.00007879, 0.00010293, 0.00015936, -0.00000277, -0.00015600, -0.00009912, 0.00005523, 0.00009233, 0.00001363, -0.00003260, -0.00000819, 0.00000736, -0.00001049, -0.00001183, 0.00001110, 0.00001452, -0.00000691, -0.00001571, -0.00000462, -0.00000041, -0.00000703, -0.00000469, 0.00000561, 0.00000520, -0.00000467, -0.00000747, -0.00000345, -0.00000492, -0.00000970, -0.00000652, 0.00000238, 0.00000640, 0.00000584, 0.00000724, 0.00000785, 0.00000147, -0.00000748, -0.00000987, -0.00000607, -0.00000268, -0.00000098, 0.00000107, 0.00000121, -0.00000352, -0.00000970, -0.00001198, -0.00001076, -0.00000930, -0.00000769, -0.00000556, -0.00000609, -0.00001165, -0.00001889, -0.00002338, -0.00002539, -0.00002697, -0.00002715, -0.00002382, -0.00001673, -0.00000597, 0.00000874, 0.00002505, 0.00003860, 0.00004743, 0.00005148, 0.00004877, 0.00003825, },
		{ 0.00001721, 0.00001825, 0.00001725, 0.00001717, 0.00001842, 0.00001976, 0.00002180, 0.00002307, 0.00001943, 0.00001441, 0.00001853, 0.00002992, 0.00003225, 0.00002252, 0.00001842, 0.00002494, 0.00002426, 0.00001560, 0.00002458, 0.00004908, 0.00004998, 0.00003195, 0.00006247, 0.00014801, 0.00016970, 0.00004159, -0.00013133, -0.00017633, -0.00009014, -0.00001794, -0.00002482, -0.00003883, -0.00001664, -0.00000894, -0.00003665, -0.00004384, -0.00001131, 0.00000632, -0.00001037, -0.00001494, 0.00000175, -0.00000690, -0.00003778, -0.00003188, 0.00000920, 0.00001868, -0.00001094, -0.00001971, -0.00000214, -0.00001022, -0.00003347, -0.00001132, 0.00003588, 0.00003002, -0.00001614, -0.00002306, -0.00000327, -0.00002895, -0.00005821, 0.00000968, 0.00010933, 0.00006428, -0.00010335, -0.00015555, -0.00000306, 0.00014191, 0.00009017, -0.00005282, -0.00008475, -0.00000724, 0.00003820, 0.00001351, -0.00000635, 0.00000427, 0.00000130, -0.00002113, -0.00002227, 0.00000102, 0.00001155, 0.00000169, -0.00000296, 0.00000200, -0.00000166, -0.00001258, -0.00001271, -0.00000311, 0.00000037, -0.00000204, 0.00000168, 0.00000946, 0.00000974, 0.00000359, 0.00000050, -0.00000008, -0.00000447, -0.00000942, -0.00000725, -0.00000055, 0.00000211, -0.00000024, -0.00000277, -0.00000495, -0.00000874, -0.00001146, -0.00000963, -0.00000610, -0.00000588, -0.00000853, -0.00001105, -0.00001417, -0.00001933, -0.00002359, -0.00002346, -0.00002042, -0.00001744, -0.00001354, -0.00000609, 0.00000460, 0.00001661, 0.00002855, 0.00003782, 0.00004120, 0.00003920, 0.00003542, 0.00003076, 0.00002386, 0.00001711, 0.00001516, },
		{ 0.00001365, 0.00001428, 0.00001744, 0.00002047, 0.00002215, 0.00002132, 0.00001583, 0.00001009, 0.00001430, 0.00002698, 0.00003177, 0.00002296, 0.00001609, 0.00001929, 0.00001930, 0.00001386, 0.00002397, 0.00004844, 0.00005138, 0.00003258, 0.00005240, 0.00012539, 0.00015160, 0.00004680, -0.00010704, -0.00015512, -0.00008425, -0.00002078, -0.00002877, -0.00004522, -0.00002446, -0.00000933, -0.00002618, -0.00003202, -0.00000801, 0.00000157, -0.00001842, -0.00002356, -0.00000244, 0.00000004, -0.00002258, -0.00002197, 0.00000528, 0.00000610, -0.00002310, -0.00002622, 0.00000274, 0.00001026, -0.00000756, -0.00000202, 0.00001869, -0.00000005, -0.00004000, -0.00003137, 0.00001126, 0.00001359, -0.00000990, 0.00001172, 0.00004063, -0.00001917, -0.00010898, -0.00006421, 0.00009509, 0.00014581, 0.00000365, -0.00013185, -0.00008190, 0.00005435, 0.00008489, 0.00000947, -0.00003848, -0.00002087, -0.00000568, -0.00001570, -0.00000990, 0.00001522, 0.00001896, -0.00000280, -0.00001460, -0.00000814, -0.00000643, -0.00001241, -0.00000793, 0.00000567, 0.00001010, 0.00000496, 0.00000450, 0.00000822, 0.00000450, -0.00000448, -0.00000689, -0.00000309, -0.00000225, -0.00000412, -0.00000215, 0.00000143, -0.00000061, -0.00000663, -0.00000920, -0.00000762, -0.00000650, -0.00000632, -0.00000514, -0.00000550, -0.00001057, -0.00001704, -0.00001953, -0.00001863, -0.00001799, -0.00001712, -0.00001355, -0.00000850, -0.00000428, 0.00000087, 0.00000966, 0.00002049, 0.00002952, 0.00003487, 0.00003568, 0.00003135, 0.00002459, 0.00002041, 0.00001958, 0.00001823, 0.00001532, 0.00001411, 0.00001512, 0.00001506, },
		{ 0.00001703, 0.00001950, 0.00001980, 0.00001781, 0.00001271, 0.00000827, 0.00001285, 0.00002503, 0.00002975, 0.00002076, 0.00001261, 0.00001560, 0.00001873, 0.00001651, 0.00002482, 0.00004452, 0.00004596, 0.00002752, 0.00004289, 0.00011035, 0.00014321, 0.00005753, -0.00008448, -0.00014176, -0.00008869, -0.00003134, -0.00003285, -0.00004308, -0.00002039, -0.00000293, -0.00001887, -0.00003099, -0.00001595, -0.00000743, -0.00002019, -0.00001852, 0.00000531, 0.00000796, -0.00001916, -0.00002893, -0.00000878, -0.00000183, -0.00001656, -0.00001033, 0.00001835, 0.00001934, -0.00001116, -0.00002216, -0.00000781, -0.00001135, -0.00002586, -0.00000388, 0.00003406, 0.00002217, -0.00002311, -0.00002822, -0.00000457, -0.00001988, -0.00004157, 0.00001847, 0.00010482, 0.00006312, -0.00008742, -0.00013612, -0.00000251, 0.00012567, 0.00007862, -0.00005178, -0.00008460, -0.00001787, 0.00002483, 0.00000884, -0.00000225, 0.00001085, 0.00000702, -0.00001797, -0.00002470, -0.00000732, 0.00000242, -0.00000217, 0.00000078, 0.00001270, 0.00001392, 0.00000369, -0.00000103, 0.00000112, -0.00000202, -0.00000868, -0.00000680, 0.00000154, 0.00000380, -0.00000105, -0.00000417, -0.00000450, -0.00000731, -0.00001059, -0.00000811, -0.00000216, -0.00000054, -0.00000438, -0.00000920, -0.00001392, -0.00001944, -0.00002213, -0.00001808, -0.00001080, -0.00000666, -0.00000574, -0.00000367, 0.00000061, 0.00000551, 0.00001194, 0.00002097, 0.00002930, 0.00003245, 0.00003015, 0.00002471, 0.00001784, 0.00001237, 0.00001200, 0.00001577, 0.00001770, 0.00001528, 0.00001250, 0.00001229, 0.00001262, 0.00001234, 0.00001371, },
		{ 0.00001679, 0.00001526, 0.00001144, 0.00000781, 0.00001136, 0.00002187, 0.00002662, 0.00001912, 0.00001180, 0.00001511, 0.00001948, 0.00001728, 0.00002153, 0.00003660, 0.00003938, 0.00002605, 0.00004111, 0.00010291, 0.00013658, 0.00006202, -0.00007216, -0.00013585, -0.00009364, -0.00003605, -0.00002843, -0.00003465, -0.00001649, -0.00000383, -0.00002129, -0.00003487, -0.00002001, -0.00000620, -0.00001221, -0.00001088, 0.00000588, 0.00000271, -0.00002584, -0.00003441, -0.00000890, 0.00000833, 0.00000014, -0.00000018, 0.00001347, 0.00000396, -0.00002751, -0.00003265, -0.00000596, 0.00000664, -0.00000136, 0.00000673, 0.00002097, -0.00000339, -0.00004400, -0.00003499, 0.00001033, 0.00002136, 0.00000523, 0.00002293, 0.00004351, -0.00001479, -0.00009853, -0.00006122, 0.00007991, 0.00012627, 0.00000084, -0.00012072, -0.00007869, 0.00004123, 0.00007002, 0.00000704, -0.00003131, -0.00001414, -0.00000340, -0.00001755, -0.00001607, 0.00000708, 0.00001569, 0.00000407, 0.00000105, 0.00001084, 0.00001044, -0.00000245, -0.00000798, -0.00000310, -0.00000208, -0.00000555, -0.00000211, 0.00000543, 0.00000384, -0.00000607, -0.00001153, -0.00000946, -0.00000698, -0.00000554, -0.00000160, 0.00000150, -0.00000299, -0.00001273, -0.00001921, -0.00001991, -0.00001794, -0.00001351, -0.00000563, 0.00000130, 0.00000201, -0.00000075, 0.00000026, 0.00000662, 0.00001445, 0.00002160, 0.00002761, 0.00002975, 0.00002565, 0.00001804, 0.00001162, 0.00000793, 0.00000724, 0.00001066, 0.00001624, 0.00001811, 0.00001425, 0.00001002, 0.00000987, 0.00001164, 0.00001234, 0.00001316, 0.00001522, 0.00001680, },
		{ 0.00001101, 0.00000725, 0.00000938, 0.00001871, 0.00002451, 0.00001916, 0.00001217, 0.00001402, 0.00001761, 0.00001511, 0.00001743, 0.00003119, 0.00003753, 0.00002928, 0.00004181, 0.00009397, 0.00012418, 0.00005926, -0.00006352, -0.00012694, -0.00009119, -0.00003472, -0.00002353, -0.00003118, -0.00001960, -0.00000934, -0.00002245, -0.00003154, -0.00001548, -0.00000151, -0.00000901, -0.00001335, -0.00000210, -0.00000314, -0.00002322, -0.00002523, 0.00000149, 0.00001631, 0.00000128, -0.00001029, -0.00000382, -0.00000750, -0.00002439, -0.00001877, 0.00000949, 0.00001656, -0.00000345, -0.00001067, -0.00000203, -0.00001219, -0.00002906, -0.00000708, 0.00003436, 0.00003164, -0.00000628, -0.00001535, -0.00000203, -0.00002195, -0.00004432, 0.00000885, 0.00008756, 0.00005354, -0.00007788, -0.00012074, -0.00000349, 0.00010820, 0.00006500, -0.00005030, -0.00007734, -0.00001651, 0.00002083, 0.00000538, -0.00000318, 0.00001342, 0.00001647, -0.00000093, -0.00000573, 0.00000638, 0.00000745, -0.00000603, -0.00001038, -0.00000146, 0.00000259, -0.00000179, -0.00000193, 0.00000183, -0.00000223, -0.00001177, -0.00001307, -0.00000514, 0.00000044, -0.00000046, -0.00000278, -0.00000604, -0.00001337, -0.00002040, -0.00001919, -0.00001094, -0.00000363, 0.00000005, 0.00000271, 0.00000380, 0.00000145, -0.00000015, 0.00000521, 0.00001584, 0.00002417, 0.00002668, 0.00002500, 0.00002011, 0.00001253, 0.00000602, 0.00000432, 0.00000622, 0.00000885, 0.00001231, 0.00001613, 0.00001641, 0.00001187, 0.00000789, 0.00000887, 0.00001178, 0.00001228, 0.00001149, 0.00001216, 0.00001374, 0.00001462, 0.00001406, },
		{ 0.00000739, 0.00001607, 0.00002292, 0.00001907, 0.00001152, 0.00001149, 0.00001500, 0.00001417, 0.00001675, 0.00002961, 0.00003732, 0.00003081, 0.00003873, 0.00008145, 0.00010994, 0.00005716, -0.00005157, -0.00011362, -0.00008656, -0.00003591, -0.00002436, -0.00003207, -0.00002217, -0.00000970, -0.00001721, -0.00002477, -0.00001331, -0.00000427, -0.00001351, -0.00001760, -0.00000340, 0.00000265, -0.00001134, -0.00001583, 0.00000117, 0.00000755, -0.00001033, -0.00002015, -0.00000673, 0.00000114, -0.00000809, -0.00000747, 0.00000689, 0.00000322, -0.00001854, -0.00001984, 0.00000177, 0.00000788, -0.00000261, 0.00000585, 0.00002352, 0.00000638, -0.00003010, -0.00002784, 0.00000518, 0.00001047, -0.00000439, 0.00001360, 0.00003547, -0.00001254, -0.00008412, -0.00005054, 0.00007198, 0.00010769, -0.00000880, -0.00011778, -0.00007643, 0.00003620, 0.00006655, 0.00001373, -0.00001724, 0.00000100, 0.00001124, -0.00000426, -0.00000843, 0.00000594, 0.00000781, -0.00000659, -0.00001003, 0.00000142, 0.00000508, -0.00000372, -0.00000792, -0.00000397, -0.00000418, -0.00000879, -0.00000648, 0.00000130, 0.00000173, -0.00000681, -0.00001388, -0.00001525, -0.00001485, -0.00001232, -0.00000428, 0.00000477, 0.00000644, 0.00000156, -0.00000181, -0.00000071, 0.00000217, 0.00000710, 0.00001612, 0.00002503, 0.00002649, 0.00002016, 0.00001227, 0.00000662, 0.00000311, 0.00000262, 0.00000591, 0.00000979, 0.00001101, 0.00001130, 0.00001300, 0.00001352, 0.00001053, 0.00000773, 0.00000904, 0.00001152, 0.00001079, 0.00000857, 0.00000895, 0.00001165, 0.00001383, 0.00001378, 0.00001065, 0.00000636, },
		{ 0.00002133, 0.00001870, 0.00001089, 0.00000984, 0.00001422, 0.00001545, 0.00001760, 0.00002757, 0.00003421, 0.00002861, 0.00003396, 0.00007068, 0.00009939, 0.00005789, -0.00003921, -0.00010213, -0.00008513, -0.00004014, -0.00002577, -0.00002972, -0.00001940, -0.00000670, -0.00001350, -0.00002368, -0.00001722, -0.00000921, -0.00001383, -0.00001271, 0.00000392, 0.00001008, -0.00000734, -0.00002006, -0.00001069, -0.00000341, -0.00001246, -0.00001355, 0.00000369, 0.00001067, -0.00000491, -0.00001531, -0.00000836, -0.00000721, -0.00001506, -0.00000470, 0.00002015, 0.00002139, -0.00000201, -0.00001086, -0.00000178, -0.00000841, -0.00002373, -0.00000804, 0.00002428, 0.00001997, -0.00001223, -0.00001685, -0.00000123, -0.00001579, -0.00003418, 0.00001143, 0.00007573, 0.00003769, -0.00008405, -0.00011945, -0.00000528, 0.00010544, 0.00007417, -0.00002673, -0.00005277, -0.00000298, 0.00002379, 0.00000293, -0.00000968, 0.00000364, 0.00000717, -0.00000710, -0.00001022, 0.00000177, 0.00000343, -0.00000862, -0.00001227, -0.00000287, 0.00000233, -0.00000170, -0.00000390, -0.00000298, -0.00000827, -0.00001753, -0.00001823, -0.00000873, 0.00000068, 0.00000515, 0.00000758, 0.00000700, 0.00000019, -0.00000752, -0.00000647, 0.00000342, 0.00001355, 0.00001947, 0.00002270, 0.00002231, 0.00001550, 0.00000564, 0.00000028, 0.00000128, 0.00000438, 0.00000711, 0.00000972, 0.00001067, 0.00000875, 0.00000745, 0.00001000, 0.00001293, 0.00001157, 0.00000831, 0.00000797, 0.00000922, 0.00000809, 0.00000612, 0.00000743, 0.00001132, 0.00001389, 0.00001320, 0.00000938, 0.00000476, 0.00000528, 0.00001363, },
		{ 0.00001123, 0.00000929, 0.00001401, 0.00001626, 0.00001698, 0.00002350, 0.00002935, 0.00002623, 0.00003144, 0.00006352, 0.00009097, 0.00005731, -0.00003049, -0.00009352, -0.00008372, -0.00004173, -0.00002353, -0.00002476, -0.00001703, -0.00000767, -0.00001511, -0.00002514, -0.00001821, -0.00000689, -0.00000684, -0.00000536, 0.00000615, 0.00000649, -0.00001425, -0.00002802, -0.00001606, -0.00000116, -0.00000255, -0.00000372, 0.00000575, 0.00000462, -0.00001454, -0.00002370, -0.00000991, 0.00000271, 0.00000286, 0.00000930, 0.00002089, 0.00000995, -0.00001859, -0.00002519, -0.00000578, 0.00000308, -0.00000382, 0.00000312, 0.00001693, 0.00000129, -0.00002925, -0.00002440, 0.00000713, 0.00001259, -0.00000168, 0.00001085, 0.00002456, -0.00002339, -0.00008686, -0.00004841, 0.00007414, 0.00011654, 0.00001500, -0.00008761, -0.00005975, 0.00003166, 0.00005159, 0.00000095, -0.00002547, -0.00000476, 0.00000822, -0.00000505, -0.00001075, 0.00000014, 0.00000154, -0.00000979, -0.00000987, 0.00000358, 0.00000775, -0.00000342, -0.00001320, -0.00001425, -0.00001436, -0.00001371, -0.00000420, 0.00001027, 0.00001570, 0.00000911, 0.00000014, -0.00000554, -0.00000902, -0.00000755, 0.00000330, 0.00001802, 0.00002486, 0.00002100, 0.00001357, 0.00000731, 0.00000155, -0.00000208, 0.00000024, 0.00000652, 0.00001023, 0.00000959, 0.00000798, 0.00000675, 0.00000526, 0.00000584, 0.00001040, 0.00001430, 0.00001201, 0.00000651, 0.00000465, 0.00000642, 0.00000697, 0.00000621, 0.00000773, 0.00001108, 0.00001254, 0.00001080, 0.00000693, 0.00000311, 0.00000404, 0.00001229, 0.00002059, 0.00001907, },
		{ 0.00001284, 0.00001585, 0.00001557, 0.00001955, 0.00002539, 0.00002518, 0.00003035, 0.00005715, 0.00008175, 0.00005456, -0.00002327, -0.00008378, -0.00007905, -0.00004071, -0.00002159, -0.00002293, -0.00001865, -0.00001095, -0.00001527, -0.00002122, -0.00001272, -0.00000154, -0.00000325, -0.00000675, -0.00000110, -0.00000179, -0.00001772, -0.00002567, -0.00001026, 0.00000577, 0.00000200, -0.00000623, -0.00000372, -0.00000431, -0.00001520, -0.00001531, 0.00000316, 0.00001513, 0.00000830, 0.00000258, 0.00000482, -0.00000433, -0.00002200, -0.00001814, 0.00000593, 0.00001333, -0.00000200, -0.00000814, -0.00000099, -0.00000777, -0.00002113, -0.00000674, 0.00002147, 0.00001625, -0.00001383, -0.00001917, -0.00000632, -0.00001877, -0.00003050, 0.00001926, 0.00008574, 0.00005476, -0.00005934, -0.00010131, -0.00000918, 0.00008366, 0.00005401, -0.00003396, -0.00005234, -0.00000314, 0.00002210, 0.00000083, -0.00001393, -0.00000283, 0.00000286, -0.00000608, -0.00000615, 0.00000466, 0.00000275, -0.00001406, -0.00002237, -0.00001351, -0.00000158, 0.00000565, 0.00001283, 0.00001728, 0.00000967, -0.00000600, -0.00001447, -0.00001038, -0.00000123, 0.00000813, 0.00001815, 0.00002424, 0.00001920, 0.00000625, -0.00000290, -0.00000330, 0.00000008, 0.00000344, 0.00000766, 0.00001115, 0.00000994, 0.00000537, 0.00000312, 0.00000438, 0.00000593, 0.00000767, 0.00001122, 0.00001322, 0.00000930, 0.00000321, 0.00000244, 0.00000650, 0.00000860, 0.00000718, 0.00000659, 0.00000811, 0.00000889, 0.00000775, 0.00000561, 0.00000358, 0.00000484, 0.00001220, 0.00002020, 0.00001931, 0.00001134, 0.00000829, },
		{ 0.00001470, 0.00001647, 0.00002187, 0.00002384, 0.00002885, 0.00005090, 0.00007310, 0.00005254, -0.00001503, -0.00007277, -0.00007404, -0.00004151, -0.00002302, -0.00002344, -0.00001931, -0.00001002, -0.00001028, -0.00001437, -0.00000899, -0.00000281, -0.00000803, -0.00001327, -0.00000674, -0.00000234, -0.00001175, -0.00001791, -0.00000687, 0.00000331, -0.00000402, -0.00001300, -0.00000724, 0.00000028, -0.00000252, -0.00000206, 0.00000863, 0.00001072, -0.00000283, -0.00001032, -0.00000323, -0.00000172, -0.00001055, -0.00000773, 0.00000713, 0.00000589, -0.00001263, -0.00001657, -0.00000144, 0.00000317, -0.00000554, -0.00000123, 0.00000979, -0.00000392, -0.00002847, -0.00001983, 0.00001213, 0.00001936, 0.00000850, 0.00002173, 0.00003444, -0.00001197, -0.00007583, -0.00004867, 0.00005685, 0.00009433, 0.00000530, -0.00008364, -0.00005569, 0.00002791, 0.00004523, -0.00000203, -0.00002646, -0.00000607, 0.00000846, -0.00000239, -0.00000945, -0.00000335, -0.00000546, -0.00001691, -0.00001440, 0.00000504, 0.00001874, 0.00001628, 0.00000844, 0.00000141, -0.00000851, -0.00001653, -0.00001122, 0.00000513, 0.00001756, 0.00001936, 0.00001602, 0.00001065, 0.00000139, -0.00000753, -0.00000732, 0.00000165, 0.00000941, 0.00001064, 0.00000896, 0.00000726, 0.00000428, 0.00000146, 0.00000275, 0.00000683, 0.00000846, 0.00000756, 0.00000818, 0.00000935, 0.00000687, 0.00000296, 0.00000386, 0.00000834, 0.00000905, 0.00000488, 0.00000208, 0.00000360, 0.00000630, 0.00000758, 0.00000738, 0.00000599, 0.00000599, 0.00001125, 0.00001852, 0.00001851, 0.00001094, 0.00000708, 0.00001150, 0.00001544, },
		{ 0.00001802, 0.00002221, 0.00002800, 0.00004660, 0.00006663, 0.00005135, -0.00000779, -0.00006390, -0.00007131, -0.00004383, -0.00002411, -0.00002103, -0.00001557, -0.00000611, -0.00000630, -0.00001269, -0.00001218, -0.00000909, -0.00001285, -0.00001429, -0.00000430, 0.00000269, -0.00000655, -0.00001703, -0.00001224, -0.00000401, -0.00000691, -0.00000918, 0.00000158, 0.00001114, 0.00000573, -0.00000213, -0.00000087, -0.00000189, -0.00001063, -0.00001038, 0.00000284, 0.00000695, -0.00000499, -0.00001076, -0.00000315, -0.00000281, -0.00001267, -0.00000919, 0.00000756, 0.00000847, -0.00000889, -0.00001507, -0.00000642, -0.00000718, -0.00001161, 0.00000802, 0.00003561, 0.00002775, -0.00000530, -0.00001555, -0.00000740, -0.00002052, -0.00003233, 0.00001176, 0.00007156, 0.00004403, -0.00005835, -0.00009577, -0.00001187, 0.00007347, 0.00004871, -0.00002923, -0.00004542, -0.00000166, 0.00001916, -0.00000339, -0.00001981, -0.00000979, -0.00000038, -0.00000155, 0.00000527, 0.00002003, 0.00001988, 0.00000140, -0.00001387, -0.00001480, -0.00000978, -0.00000315, 0.00000911, 0.00002151, 0.00002107, 0.00000814, -0.00000339, -0.00000649, -0.00000589, -0.00000392, 0.00000267, 0.00001134, 0.00001350, 0.00000782, 0.00000232, 0.00000172, 0.00000290, 0.00000373, 0.00000595, 0.00000825, 0.00000676, 0.00000343, 0.00000421, 0.00000819, 0.00000880, 0.00000570, 0.00000482, 0.00000656, 0.00000502, -0.00000006, -0.00000180, 0.00000248, 0.00000791, 0.00001035, 0.00000972, 0.00000694, 0.00000497, 0.00000865, 0.00001623, 0.00001805, 0.00001131, 0.00000647, 0.00001026, 0.00001484, 0.00001373, 0.00001332, },
		{ 0.00002868, 0.00004390, 0.00006056, 0.00004846, -0.00000351, -0.00005707, -0.00006819, -0.00004355, -0.00002175, -0.00001611, -0.00001198, -0.00000586, -0.00000825, -0.00001601, -0.00001636, -0.00001161, -0.00001124, -0.00000997, -0.00000107, 0.00000301, -0.00000877, -0.00002114, -0.00001610, -0.00000322, 0.00000077, 0.00000157, 0.00000893, 0.00001187, 0.00000034, -0.00001154, -0.00001011, -0.00000528, -0.00000621, -0.00000286, 0.00000716, 0.00000588, -0.00001013, -0.00001741, -0.00000690, 0.00000061, -0.00000391, -0.00000350, 0.00000401, -0.00000201, -0.00001923, -0.00001876, 0.00000183, 0.00001404, 0.00001292, 0.00001892, 0.00002404, 0.00000147, -0.00003116, -0.00002907, 0.00000034, 0.00001126, 0.00000578, 0.00001932, 0.00002937, -0.00001527, -0.00007480, -0.00004976, 0.00004871, 0.00008716, 0.00001023, -0.00007026, -0.00004910, 0.00002068, 0.00003203, -0.00001133, -0.00002925, -0.00000220, 0.00002093, 0.00001745, 0.00001041, 0.00000920, -0.00000191, -0.00002052, -0.00002307, -0.00000527, 0.00001227, 0.00001778, 0.00001755, 0.00001408, 0.00000239, -0.00001210, -0.00001524, -0.00000532, 0.00000496, 0.00000838, 0.00000926, 0.00001004, 0.00000690, 0.00000077, -0.00000097, 0.00000340, 0.00000713, 0.00000651, 0.00000498, 0.00000438, 0.00000260, 0.00000131, 0.00000508, 0.00001123, 0.00001140, 0.00000504, 0.00000041, 0.00000069, 0.00000064, -0.00000147, -0.00000038, 0.00000557, 0.00001080, 0.00001141, 0.00000885, 0.00000521, 0.00000307, 0.00000690, 0.00001550, 0.00001877, 0.00001199, 0.00000531, 0.00000797, 0.00001337, 0.00001264, 0.00001081, 0.00001513, 0.00002170, },
		{ 0.00005336, 0.00004405, 0.00000000, -0.00004900, -0.00006210, -0.00004061, -0.00001916, -0.00001396, -0.00001307, -0.00001007, -0.00001209, -0.00001729, -0.00001548, -0.00000916, -0.00000804, -0.00000857, -0.00000350, -0.00000136, -0.00001103, -0.00001935, -0.00001076, 0.00000470, 0.00000879, 0.00000495, 0.00000472, 0.00000289, -0.00000760, -0.00001504, -0.00000885, -0.00000040, -0.00000090, -0.00000177, 0.00000262, -0.00000010, -0.00001244, -0.00001514, -0.00000229, 0.00000477, -0.00000388, -0.00001041, -0.00000649, -0.00000673, -0.00001159, -0.00000136, 0.00002063, 0.00002561, 0.00001003, -0.00000046, -0.00000217, -0.00001380, -0.00002514, -0.00000772, 0.00002295, 0.00002312, -0.00000268, -0.00001282, -0.00000939, -0.00002446, -0.00003486, 0.00000875, 0.00006783, 0.00004624, -0.00004713, -0.00008630, -0.00001763, 0.00005527, 0.00003504, -0.00002709, -0.00003056, 0.00001854, 0.00004084, 0.00001512, -0.00001237, -0.00001715, -0.00001687, -0.00001758, -0.00000402, 0.00001971, 0.00002839, 0.00001537, -0.00000088, -0.00000880, -0.00001311, -0.00001434, -0.00000559, 0.00000919, 0.00001551, 0.00000960, 0.00000245, 0.00000125, 0.00000191, 0.00000152, 0.00000349, 0.00000766, 0.00000787, 0.00000302, -0.00000025, 0.00000118, 0.00000334, 0.00000468, 0.00000811, 0.00001150, 0.00000801, -0.00000107, -0.00000567, -0.00000212, 0.00000227, 0.00000276, 0.00000348, 0.00000715, 0.00000974, 0.00000863, 0.00000624, 0.00000431, 0.00000356, 0.00000736, 0.00001547, 0.00001859, 0.00001104, 0.00000277, 0.00000515, 0.00001244, 0.00001308, 0.00001020, 0.00001361, 0.00002161, 0.00002911, 0.00004058, },
		{ 0.00000497, -0.00003941, -0.00005540, -0.00003920, -0.00002027, -0.00001591, -0.00001634, -0.00001334, -0.00001245, -0.00001450, -0.00001244, -0.00000807, -0.00000903, -0.00001128, -0.00000662, -0.00000160, -0.00000590, -0.00001079, -0.00000347, 0.00000770, 0.00000701, -0.00000122, -0.00000364, -0.00000295, -0.00000742, -0.00001044, -0.00000444, 0.00000101, -0.00000277, -0.00000600, -0.00000134, -0.00000005, -0.00000788, -0.00001072, -0.00000287, -0.00000084, -0.00001087, -0.00001447, -0.00000315, 0.00000648, 0.00000690, 0.00001132, 0.00001906, 0.00001085, -0.00001001, -0.00001616, -0.00000450, 0.00000086, -0.00000218, 0.00000592, 0.00001619, 0.00000080, -0.00002695, -0.00002648, -0.00000193, 0.00000798, 0.00000567, 0.00002104, 0.00003101, -0.00001182, -0.00007046, -0.00005249, 0.00003526, 0.00007471, 0.00001519, -0.00004668, -0.00002144, 0.00003975, 0.00004036, -0.00001314, -0.00004277, -0.00002452, 0.00000075, 0.00000994, 0.00001717, 0.00002449, 0.00001453, -0.00000965, -0.00002268, -0.00001570, -0.00000415, 0.00000208, 0.00000801, 0.00001364, 0.00001036, -0.00000044, -0.00000551, 0.00000003, 0.00000671, 0.00000745, 0.00000584, 0.00000507, 0.00000244, -0.00000160, -0.00000110, 0.00000447, 0.00000799, 0.00000655, 0.00000503, 0.00000487, 0.00000127, -0.00000496, -0.00000531, 0.00000202, 0.00000774, 0.00000617, 0.00000300, 0.00000376, 0.00000587, 0.00000632, 0.00000645, 0.00000652, 0.00000564, 0.00000718, 0.00001323, 0.00001614, 0.00000913, 0.00000092, 0.00000405, 0.00001346, 0.00001513, 0.00001035, 0.00001160, 0.00002004, 0.00002811, 0.00003683, 0.00004676, 0.00004078, },
		{ -0.00005141, -0.00004092, -0.00002372, -0.00001802, -0.00001729, -0.00001327, -0.00001065, -0.00001215, -0.00001213, -0.00001017, -0.00001112, -0.00001122, -0.00000371, 0.00000436, 0.00000184, -0.00000525, -0.00000391, 0.00000186, -0.00000026, -0.00000666, -0.00000586, -0.00000130, -0.00000300, -0.00000697, -0.00000504, -0.00000261, -0.00000588, -0.00000665, 0.00000028, 0.00000276, -0.00000636, -0.00001354, -0.00000949, -0.00000551, -0.00000789, -0.00000356, 0.00001127, 0.00001839, 0.00000985, 0.00000144, 0.00000040, -0.00000511, -0.00001348, -0.00000637, 0.00001200, 0.00001469, -0.00000073, -0.00000959, -0.00000825, -0.00001484, -0.00002174, -0.00000395, 0.00002467, 0.00002495, 0.00000067, -0.00001122, -0.00001222, -0.00002913, -0.00003844, 0.00000538, 0.00006645, 0.00005565, -0.00002235, -0.00005744, -0.00000288, 0.00005022, 0.00001914, -0.00004505, -0.00004873, 0.00000308, 0.00003603, 0.00002514, 0.00000614, -0.00000130, -0.00001094, -0.00002311, -0.00001864, 0.00000186, 0.00001484, 0.00001101, 0.00000408, 0.00000259, 0.00000034, -0.00000404, -0.00000200, 0.00000705, 0.00001164, 0.00000660, 0.00000015, -0.00000096, 0.00000026, 0.00000098, 0.00000384, 0.00000797, 0.00000695, 0.00000044, -0.00000331, -0.00000102, 0.00000085, -0.00000060, 0.00000063, 0.00000641, 0.00000873, 0.00000353, -0.00000156, 0.00000014, 0.00000510, 0.00000807, 0.00000914, 0.00000845, 0.00000528, 0.00000410, 0.00000926, 0.00001397, 0.00000924, 0.00000166, 0.00000475, 0.00001494, 0.00001661, 0.00000938, 0.00000821, 0.00001738, 0.00002720, 0.00003472, 0.00004250, 0.00003907, 0.00000985, -0.00003145, },
		{ -0.00002598, -0.00001784, -0.00001615, -0.00001275, -0.00001049, -0.00001224, -0.00001288, -0.00001039, -0.00000856, -0.00000572, 0.00000238, 0.00000883, 0.00000343, -0.00000775, -0.00001029, -0.00000505, -0.00000358, -0.00000566, -0.00000299, 0.00000120, -0.00000186, -0.00000794, -0.00000773, -0.00000453, -0.00000506, -0.00000451, 0.00000071, 0.00000049, -0.00001006, -0.00001658, -0.00000876, 0.00000203, 0.00000528, 0.00000836, 0.00001465, 0.00001136, -0.00000402, -0.00001290, -0.00000676, 0.00000017, 0.00000133, 0.00000707, 0.00001408, 0.00000384, -0.00001864, -0.00002513, -0.00001098, 0.00000070, 0.00000455, 0.00001493, 0.00002243, 0.00000305, -0.00002860, -0.00003298, -0.00001140, 0.00000258, 0.00001012, 0.00003425, 0.00004928, 0.00000972, -0.00005073, -0.00004645, 0.00002008, 0.00004710, -0.00000744, -0.00005656, -0.00002323, 0.00004206, 0.00004890, 0.00000102, -0.00003145, -0.00002429, -0.00001006, -0.00000580, 0.00000288, 0.00001660, 0.00001598, -0.00000010, -0.00000996, -0.00000451, 0.00000306, 0.00000402, 0.00000474, 0.00000817, 0.00000694, -0.00000044, -0.00000425, 0.00000029, 0.00000576, 0.00000603, 0.00000417, 0.00000271, -0.00000136, -0.00000684, -0.00000596, 0.00000223, 0.00000787, 0.00000565, 0.00000252, 0.00000376, 0.00000403, -0.00000045, -0.00000308, 0.00000148, 0.00000802, 0.00001030, 0.00000924, 0.00000666, 0.00000240, 0.00000094, 0.00000733, 0.00001468, 0.00001169, 0.00000326, 0.00000467, 0.00001472, 0.00001661, 0.00000780, 0.00000501, 0.00001550, 0.00002776, 0.00003450, 0.00003941, 0.00003661, 0.00001206, -0.00002674, -0.00004970, -0.00004318, },
		{ -0.00001554, -0.00001363, -0.00001125, -0.00001124, -0.00001007, -0.00000592, -0.00000273, -0.00000057, 0.00000401, 0.00000632, -0.00000133, -0.00001285, -0.00001446, -0.00000664, -0.00000174, -0.00000239, -0.00000146, 0.00000028, -0.00000345, -0.00000885, -0.00000785, -0.00000393, -0.00000440, -0.00000563, -0.00000278, -0.00000256, -0.00000882, -0.00000995, 0.00000129, 0.00001205, 0.00001064, 0.00000477, 0.00000264, -0.00000217, -0.00001096, -0.00001003, 0.00000353, 0.00001193, 0.00000664, 0.00000014, -0.00000252, -0.00001143, -0.00002195, -0.00001402, 0.00000902, 0.00001947, 0.00001018, 0.00000034, -0.00000586, -0.00002035, -0.00003102, -0.00001241, 0.00002240, 0.00003464, 0.00002258, 0.00001428, 0.00000606, -0.00002349, -0.00004682, -0.00001711, 0.00003619, 0.00003280, -0.00002566, -0.00004531, 0.00001036, 0.00005703, 0.00002231, -0.00004380, -0.00005362, -0.00000963, 0.00002184, 0.00001731, 0.00000703, 0.00000638, 0.00000063, -0.00001159, -0.00001134, 0.00000375, 0.00001342, 0.00000824, 0.00000063, -0.00000002, 0.00000077, -0.00000090, 0.00000065, 0.00000670, 0.00000854, 0.00000219, -0.00000473, -0.00000615, -0.00000532, -0.00000442, 0.00000073, 0.00000936, 0.00001174, 0.00000433, -0.00000283, -0.00000182, 0.00000174, 0.00000118, 0.00000048, 0.00000446, 0.00000869, 0.00000822, 0.00000572, 0.00000391, 0.00000148, 0.00000122, 0.00000828, 0.00001647, 0.00001347, 0.00000320, 0.00000284, 0.00001346, 0.00001676, 0.00000738, 0.00000316, 0.00001458, 0.00002885, 0.00003439, 0.00003548, 0.00003200, 0.00001157, -0.00002373, -0.00004779, -0.00004383, -0.00002649, -0.00001687, },
		{ -0.00000946, -0.00000656, -0.00000410, -0.00000091, -0.00000003, -0.00000129, -0.00000048, 0.00000019, -0.00000570, -0.00001403, -0.00001354, -0.00000484, 0.00000024, -0.00000147, -0.00000258, -0.00000164, -0.00000401, -0.00000794, -0.00000736, -0.00000498, -0.00000652, -0.00000768, -0.00000287, 0.00000177, 0.00000023, -0.00000014, 0.00000639, 0.00000982, 0.00000218, -0.00000635, -0.00000566, -0.00000226, -0.00000239, 0.00000088, 0.00000861, 0.00000735, -0.00000576, -0.00001394, -0.00000949, -0.00000453, -0.00000330, 0.00000486, 0.00001589, 0.00000987, -0.00001154, -0.00002284, -0.00001635, -0.00000799, 0.00000033, 0.00002009, 0.00003775, 0.00002612, -0.00000508, -0.00002042, -0.00001742, -0.00001929, -0.00001822, 0.00000855, 0.00003439, 0.00001246, -0.00003220, -0.00002629, 0.00002705, 0.00003978, -0.00001937, -0.00006708, -0.00003343, 0.00003320, 0.00004758, 0.00001007, -0.00001726, -0.00001209, -0.00000281, -0.00000367, 0.00000035, 0.00001186, 0.00001288, -0.00000023, -0.00000891, -0.00000337, 0.00000519, 0.00000683, 0.00000550, 0.00000489, 0.00000046, -0.00000800, -0.00001136, -0.00000548, 0.00000180, 0.00000429, 0.00000583, 0.00000873, 0.00000678, -0.00000176, -0.00000674, -0.00000149, 0.00000580, 0.00000587, 0.00000249, 0.00000279, 0.00000472, 0.00000418, 0.00000361, 0.00000474, 0.00000416, 0.00000334, 0.00000875, 0.00001628, 0.00001326, 0.00000215, 0.00000106, 0.00001292, 0.00001804, 0.00000812, 0.00000192, 0.00001336, 0.00002906, 0.00003323, 0.00003033, 0.00002592, 0.00000993, -0.00002073, -0.00004477, -0.00004329, -0.00002672, -0.00001643, -0.00001529, -0.00001363, },
		{ -0.00000011, -0.00000060, -0.00000311, -0.00000634, -0.00000590, -0.00000338, -0.00000563, -0.00001128, -0.00001100, -0.00000409, -0.00000021, -0.00000237, -0.00000376, -0.00000247, -0.00000382, -0.00000764, -0.00000836, -0.00000666, -0.00000647, -0.00000434, 0.00000342, 0.00000941, 0.00000620, 0.00000026, -0.00000017, -0.00000016, -0.00000524, -0.00000746, -0.00000043, 0.00000630, 0.00000391, -0.00000087, -0.00000203, -0.00000588, -0.00001285, -0.00001000, 0.00000440, 0.00001295, 0.00000761, 0.00000043, -0.00000338, -0.00001287, -0.00002365, -0.00001655, 0.00000702, 0.00002282, 0.00002278, 0.00002002, 0.00001404, -0.00000721, -0.00003018, -0.00002717, -0.00000486, 0.00000583, 0.00000414, 0.00001112, 0.00001621, -0.00000529, -0.00002962, -0.00001237, 0.00002389, 0.00001220, -0.00004129, -0.00005147, 0.00001027, 0.00006214, 0.00003516, -0.00002610, -0.00004038, -0.00000662, 0.00001745, 0.00001104, 0.00000164, 0.00000351, 0.00000196, -0.00000715, -0.00000764, 0.00000480, 0.00001337, 0.00000798, -0.00000181, -0.00000633, -0.00000812, -0.00000955, -0.00000553, 0.00000411, 0.00000950, 0.00000572, 0.00000077, 0.00000116, 0.00000167, -0.00000200, -0.00000302, 0.00000354, 0.00000931, 0.00000584, -0.00000108, -0.00000202, 0.00000151, 0.00000381, 0.00000567, 0.00000799, 0.00000677, 0.00000358, 0.00000690, 0.00001462, 0.00001301, 0.00000207, 0.00000032, 0.00001296, 0.00001961, 0.00000905, 0.00000044, 0.00001119, 0.00002790, 0.00003093, 0.00002448, 0.00001942, 0.00000818, -0.00001730, -0.00004097, -0.00004218, -0.00002666, -0.00001522, -0.00001305, -0.00001064, -0.00000493, -0.00000097, },
		{ -0.00000834, -0.00001016, -0.00000752, -0.00000268, -0.00000298, -0.00000846, -0.00001022, -0.00000547, -0.00000182, -0.00000308, -0.00000413, -0.00000298, -0.00000414, -0.00000761, -0.00000785, -0.00000452, -0.00000162, 0.00000198, 0.00000811, 0.00001013, 0.00000251, -0.00000667, -0.00000728, -0.00000299, -0.00000159, -0.00000014, 0.00000484, 0.00000544, -0.00000352, -0.00001184, -0.00001036, -0.00000494, -0.00000116, 0.00000611, 0.00001523, 0.00001186, -0.00000600, -0.00001986, -0.00001895, -0.00001231, -0.00000436, 0.00001197, 0.00002987, 0.00002865, 0.00000799, -0.00000971, -0.00001610, -0.00002140, -0.00002223, -0.00000523, 0.00001705, 0.00001751, 0.00000107, -0.00000515, -0.00000260, -0.00001180, -0.00002125, -0.00000551, 0.00001430, -0.00000284, -0.00003472, -0.00001833, 0.00003849, 0.00005269, -0.00000419, -0.00005443, -0.00003130, 0.00002464, 0.00003715, 0.00000496, -0.00001731, -0.00000962, 0.00000157, 0.00000116, 0.00000218, 0.00000957, 0.00000913, -0.00000358, -0.00001350, -0.00001074, -0.00000323, 0.00000067, 0.00000367, 0.00000754, 0.00000601, -0.00000209, -0.00000650, -0.00000121, 0.00000519, 0.00000436, 0.00000169, 0.00000411, 0.00000624, 0.00000125, -0.00000488, -0.00000325, 0.00000312, 0.00000675, 0.00000810, 0.00000924, 0.00000677, 0.00000220, 0.00000488, 0.00001377, 0.00001381, 0.00000270, -0.00000032, 0.00001271, 0.00002102, 0.00000996, -0.00000154, 0.00000760, 0.00002502, 0.00002764, 0.00001869, 0.00001351, 0.00000711, -0.00001313, -0.00003625, -0.00003992, -0.00002501, -0.00001180, -0.00000843, -0.00000640, -0.00000155, 0.00000084, -0.00000134, -0.00000495, },
		{ -0.00000557, -0.00000069, -0.00000147, -0.00000796, -0.00001133, -0.00000744, -0.00000295, -0.00000298, -0.00000394, -0.00000309, -0.00000343, -0.00000518, -0.00000400, 0.00000002, 0.00000242, 0.00000340, 0.00000535, 0.00000441, -0.00000283, -0.00000891, -0.00000554, 0.00000186, 0.00000338, 0.00000046, -0.00000124, -0.00000444, -0.00001103, -0.00001206, -0.00000210, 0.00000876, 0.00001082, 0.00000813, 0.00000427, -0.00000643, -0.00002101, -0.00002263, -0.00000609, 0.00001223, 0.00002027, 0.00002267, 0.00001972, 0.00000285, -0.00002060, -0.00002829, -0.00001669, -0.00000422, 0.00000243, 0.00001179, 0.00001772, 0.00000488, -0.00001590, -0.00001881, -0.00000769, -0.00000646, -0.00001154, -0.00000252, 0.00000894, -0.00000251, -0.00001747, 0.00000245, 0.00003538, 0.00002089, -0.00003358, -0.00004872, 0.00000343, 0.00005024, 0.00002865, -0.00002365, -0.00003481, -0.00000325, 0.00001916, 0.00001209, -0.00000012, -0.00000196, -0.00000422, -0.00001164, -0.00001174, -0.00000037, 0.00000879, 0.00000676, 0.00000107, -0.00000074, -0.00000229, -0.00000591, -0.00000440, 0.00000499, 0.00001158, 0.00000702, -0.00000081, -0.00000139, 0.00000158, -0.00000020, -0.00000328, -0.00000017, 0.00000590, 0.00000787, 0.00000750, 0.00000825, 0.00000636, 0.00000184, 0.00000406, 0.00001340, 0.00001447, 0.00000312, -0.00000097, 0.00001253, 0.00002231, 0.00001022, -0.00000505, 0.00000190, 0.00002065, 0.00002454, 0.00001455, 0.00000951, 0.00000738, -0.00000794, -0.00003023, -0.00003589, -0.00002151, -0.00000716, -0.00000418, -0.00000483, -0.00000281, -0.00000226, -0.00000568, -0.00000930, -0.00001067, -0.00000971, },
		{ -0.00000255, -0.00000900, -0.00001234, -0.00000822, -0.00000275, -0.00000183, -0.00000270, -0.00000173, -0.00000088, -0.00000138, -0.00000058, 0.00000135, 0.00000104, -0.00000037, 0.00000065, 0.00000157, -0.00000166, -0.00000469, -0.00000172, 0.00000221, -0.00000084, -0.00000724, -0.00000876, -0.00000645, -0.00000453, 0.00000023, 0.00000909, 0.00001236, 0.00000389, -0.00000718, -0.00001163, -0.00001258, -0.00001164, -0.00000128, 0.00001682, 0.00002548, 0.00001654, 0.00000194, -0.00000880, -0.00001891, -0.00002516, -0.00001561, 0.00000492, 0.00001527, 0.00001014, 0.00000372, -0.00000078, -0.00001237, -0.00002345, -0.00001674, -0.00000045, 0.00000211, -0.00000531, -0.00000137, 0.00000814, 0.00000221, -0.00000793, 0.00000278, 0.00001669, -0.00000236, -0.00003370, -0.00002022, 0.00003114, 0.00004571, -0.00000300, -0.00004690, -0.00002665, 0.00002264, 0.00003322, 0.00000262, -0.00002062, -0.00001591, -0.00000465, -0.00000172, 0.00000163, 0.00000918, 0.00000935, -0.00000141, -0.00000979, -0.00000706, -0.00000092, 0.00000092, 0.00000294, 0.00000874, 0.00001032, 0.00000203, -0.00000650, -0.00000451, 0.00000261, 0.00000367, 0.00000047, 0.00000146, 0.00000538, 0.00000623, 0.00000605, 0.00000810, 0.00000735, 0.00000226, 0.00000301, 0.00001234, 0.00001497, 0.00000418, -0.00000101, 0.00001189, 0.00002187, 0.00000816, -0.00001060, -0.00000476, 0.00001703, 0.00002389, 0.00001344, 0.00000773, 0.00000872, -0.00000198, -0.00002302, -0.00003068, -0.00001793, -0.00000436, -0.00000368, -0.00000795, -0.00000797, -0.00000673, -0.00000815, -0.00000959, -0.00000884, -0.00000687, -0.00000378, -0.00000083, },
		{ -0.00001128, -0.00000708, -0.00000134, 0.00000004, -0.00000085, -0.00000002, 0.00000106, 0.00000025, -0.00000070, -0.00000097, -0.00000210, -0.00000256, 0.00000052, 0.00000374, 0.00000155, -0.00000344, -0.00000494, -0.00000462, -0.00000736, -0.00000945, -0.00000464, 0.00000323, 0.00000645, 0.00000582, 0.00000465, -0.00000058, -0.00001058, -0.00001503, -0.00000727, 0.00000423, 0.00001060, 0.00001451, 0.00001683, 0.00000908, -0.00000863, -0.00002078, -0.00001878, -0.00001146, -0.00000492, 0.00000533, 0.00001532, 0.00001129, -0.00000477, -0.00001468, -0.00001390, -0.00001415, -0.00001519, -0.00000590, 0.00000693, 0.00000564, -0.00000380, -0.00000133, 0.00000773, 0.00000290, -0.00000869, -0.00000506, 0.00000417, -0.00000497, -0.00001662, 0.00000253, 0.00003255, 0.00001979, -0.00002889, -0.00004332, 0.00000133, 0.00004148, 0.00002169, -0.00002548, -0.00003656, -0.00000821, 0.00001503, 0.00001310, 0.00000430, 0.00000175, -0.00000238, -0.00001075, -0.00001148, -0.00000060, 0.00000873, 0.00000722, 0.00000257, 0.00000320, 0.00000383, -0.00000175, -0.00000629, -0.00000119, 0.00000691, 0.00000659, 0.00000070, -0.00000028, 0.00000355, 0.00000543, 0.00000619, 0.00000880, 0.00000799, 0.00000186, 0.00000161, 0.00001187, 0.00001657, 0.00000577, -0.00000225, 0.00000826, 0.00001816, 0.00000442, -0.00001551, -0.00000905, 0.00001653, 0.00002628, 0.00001460, 0.00000706, 0.00001041, 0.00000428, -0.00001562, -0.00002641, -0.00001726, -0.00000598, -0.00000745, -0.00001340, -0.00001247, -0.00000780, -0.00000609, -0.00000644, -0.00000618, -0.00000576, -0.00000486, -0.00000328, -0.00000427, -0.00000891, },
		{ -0.00000045, 0.00000102, -0.00000001, 0.00000029, 0.00000097, -0.00000026, -0.00000193, -0.00000234, -0.00000232, -0.00000128, 0.00000206, 0.00000395, -0.00000060, -0.00000811, -0.00001060, -0.00000768, -0.00000459, -0.00000122, 0.00000494, 0.00000904, 0.00000486, -0.00000370, -0.00000873, -0.00000997, -0.00000916, -0.00000201, 0.00001165, 0.00002015, 0.00001513, 0.00000361, -0.00000565, -0.00001405, -0.00002108, -0.00001759, -0.00000290, 0.00000879, 0.00000931, 0.00000588, 0.00000230, -0.00000728, -0.00001929, -0.00001984, -0.00000935, -0.00000315, -0.00000373, 0.00000037, 0.00000729, 0.00000409, -0.00000433, -0.00000194, 0.00000548, -0.00000015, -0.00001179, -0.00000835, 0.00000376, 0.00000275, -0.00000362, 0.00000619, 0.00001691, -0.00000228, -0.00003191, -0.00002148, 0.00002246, 0.00003459, -0.00000761, -0.00004474, -0.00002549, 0.00001978, 0.00003213, 0.00000772, -0.00001314, -0.00001220, -0.00000580, -0.00000498, -0.00000129, 0.00000821, 0.00001127, 0.00000252, -0.00000556, -0.00000282, 0.00000315, 0.00000209, -0.00000127, 0.00000206, 0.00000751, 0.00000531, -0.00000111, -0.00000135, 0.00000367, 0.00000592, 0.00000601, 0.00000816, 0.00000781, 0.00000215, 0.00000193, 0.00001256, 0.00001747, 0.00000460, -0.00000734, 0.00000151, 0.00001385, 0.00000316, -0.00001621, -0.00000957, 0.00001824, 0.00002972, 0.00001638, 0.00000680, 0.00001208, 0.00000972, -0.00001038, -0.00002590, -0.00002114, -0.00001098, -0.00001166, -0.00001607, -0.00001239, -0.00000467, -0.00000213, -0.00000432, -0.00000656, -0.00000769, -0.00000723, -0.00000475, -0.00000352, -0.00000620, -0.00000850, -0.00000552, },
		{ 0.00000009, 0.00000012, 0.00000076, -0.00000011, -0.00000149, -0.00000172, -0.00000166, -0.00000150, -0.00000030, -0.00000041, -0.00000502, -0.00001029, -0.00000891, -0.00000171, 0.00000387, 0.00000528, 0.00000494, 0.00000166, -0.00000631, -0.00001278, -0.00001023, -0.00000123, 0.00000713, 0.00001375, 0.00001801, 0.00001309, -0.00000270, -0.00001748, -0.00002052, -0.00001538, -0.00000844, 0.00000146, 0.00001206, 0.00001252, 0.00000062, -0.00001098, -0.00001455, -0.00001585, -0.00001660, -0.00000962, 0.00000215, 0.00000527, -0.00000028, -0.00000098, 0.00000383, 0.00000202, -0.00000465, -0.00000320, 0.00000233, -0.00000250, -0.00001107, -0.00000539, 0.00000765, 0.00000695, -0.00000263, -0.00000144, 0.00000317, -0.00000821, -0.00002022, -0.00000377, 0.00002266, 0.00001236, -0.00002774, -0.00003699, 0.00000423, 0.00003994, 0.00002284, -0.00001858, -0.00003004, -0.00000828, 0.00000977, 0.00000807, 0.00000272, 0.00000410, 0.00000295, -0.00000519, -0.00000849, 0.00000012, 0.00000924, 0.00000709, 0.00000004, 0.00000019, 0.00000501, 0.00000425, -0.00000062, -0.00000045, 0.00000384, 0.00000487, 0.00000436, 0.00000767, 0.00000954, 0.00000465, 0.00000261, 0.00001064, 0.00001381, -0.00000067, -0.00001394, -0.00000353, 0.00001372, 0.00000679, -0.00001334, -0.00000890, 0.00001955, 0.00003286, 0.00001865, 0.00000676, 0.00001227, 0.00001163, -0.00000975, -0.00002944, -0.00002672, -0.00001429, -0.00001135, -0.00001368, -0.00000930, -0.00000187, -0.00000112, -0.00000583, -0.00000909, -0.00000909, -0.00000676, -0.00000276, -0.00000050, -0.00000319, -0.00000706, -0.00000598, -0.00000141, 0.00000075, },
		{ 0.00000134, 0.00000053, -0.00000123, -0.00000257, -0.00000387, -0.00000499, -0.00000433, -0.00000298, -0.00000385, -0.00000500, -0.00000171, 0.00000411, 0.00000532, 0.00000071, -0.00000439, -0.00000753, -0.00000939, -0.00000689, 0.00000257, 0.00001279, 0.00001531, 0.00001066, 0.00000353, -0.00000628, -0.00001758, -0.00002121, -0.00001208, 0.00000088, 0.00000736, 0.00000821, 0.00000604, -0.00000262, -0.00001577, -0.00002154, -0.00001556, -0.00000776, -0.00000410, 0.00000072, 0.00000658, 0.00000475, -0.00000296, -0.00000399, 0.00000158, 0.00000116, -0.00000502, -0.00000464, 0.00000065, -0.00000157, -0.00000715, -0.00000192, 0.00000770, 0.00000392, -0.00000763, -0.00000767, -0.00000065, -0.00000404, -0.00000981, 0.00000031, 0.00001145, -0.00000330, -0.00002608, -0.00001401, 0.00002465, 0.00003296, -0.00000593, -0.00003884, -0.00002229, 0.00001632, 0.00002647, 0.00000582, -0.00001050, -0.00000742, -0.00000102, -0.00000235, -0.00000211, 0.00000622, 0.00001196, 0.00000579, -0.00000322, -0.00000214, 0.00000511, 0.00000595, 0.00000053, -0.00000118, 0.00000187, 0.00000331, 0.00000456, 0.00000999, 0.00001268, 0.00000545, -0.00000133, 0.00000336, 0.00000705, -0.00000487, -0.00001597, -0.00000301, 0.00001777, 0.00001229, -0.00001055, -0.00000919, 0.00002024, 0.00003619, 0.00002089, 0.00000474, 0.00000804, 0.00000845, -0.00001218, -0.00003235, -0.00002832, -0.00001217, -0.00000682, -0.00001036, -0.00000882, -0.00000329, -0.00000318, -0.00000741, -0.00000876, -0.00000634, -0.00000310, -0.00000014, 0.00000052, -0.00000350, -0.00000848, -0.00000781, -0.00000226, 0.00000147, 0.00000141, 0.00000100, },
		{ -0.00000326, -0.00000549, -0.00000695, -0.00000695, -0.00000401, 0.00000040, 0.00000215, 0.00000102, 0.00000072, 0.00000112, -0.00000183, -0.00000714, -0.00000874, -0.00000498, 0.00000036, 0.00000571, 0.00001112, 0.00001218, 0.00000492, -0.00000562, -0.00001156, -0.00001274, -0.00001202, -0.00000701, 0.00000285, 0.00000899, 0.00000479, -0.00000415, -0.00001042, -0.00001513, -0.00001881, -0.00001529, -0.00000435, 0.00000334, 0.00000292, 0.00000187, 0.00000340, 0.00000096, -0.00000502, -0.00000488, 0.00000070, 0.00000037, -0.00000540, -0.00000470, 0.00000155, 0.00000103, -0.00000452, -0.00000265, 0.00000294, -0.00000149, -0.00001046, -0.00000771, 0.00000158, 0.00000026, -0.00000653, -0.00000266, 0.00000351, -0.00000510, -0.00001402, 0.00000127, 0.00002260, 0.00001016, -0.00002633, -0.00003275, 0.00000536, 0.00003682, 0.00002091, -0.00001606, -0.00002604, -0.00000629, 0.00000990, 0.00000732, 0.00000143, 0.00000454, 0.00000799, 0.00000257, -0.00000381, -0.00000044, 0.00000705, 0.00000619, -0.00000127, -0.00000345, 0.00000132, 0.00000521, 0.00000745, 0.00001162, 0.00001130, 0.00000010, -0.00000974, -0.00000428, 0.00000435, -0.00000259, -0.00001228, 0.00000009, 0.00002132, 0.00001626, -0.00000853, -0.00000935, 0.00002092, 0.00003825, 0.00001995, -0.00000197, -0.00000045, 0.00000405, -0.00001157, -0.00002950, -0.00002467, -0.00000798, -0.00000410, -0.00001131, -0.00001232, -0.00000583, -0.00000269, -0.00000427, -0.00000433, -0.00000238, -0.00000162, -0.00000181, -0.00000265, -0.00000591, -0.00000931, -0.00000726, -0.00000047, 0.00000394, 0.00000319, 0.00000126, 0.00000050, -0.00000080, },
		{ -0.00000654, -0.00000432, 0.00000029, 0.00000494, 0.00000507, 0.00000037, -0.00000417, -0.00000583, -0.00000660, -0.00000643, -0.00000222, 0.00000479, 0.00000876, 0.00000765, 0.00000408, -0.00000137, -0.00000876, -0.00001305, -0.00000941, -0.00000196, 0.00000211, 0.00000279, 0.00000265, -0.00000116, -0.00001019, -0.00001726, -0.00001633, -0.00001167, -0.00000827, -0.00000337, 0.00000371, 0.00000589, 0.00000123, -0.00000189, -0.00000013, -0.00000064, -0.00000498, -0.00000503, -0.00000033, -0.00000018, -0.00000445, -0.00000297, 0.00000295, 0.00000150, -0.00000596, -0.00000678, -0.00000206, -0.00000362, -0.00000848, -0.00000424, 0.00000382, 0.00000107, -0.00000696, -0.00000465, 0.00000216, -0.00000151, -0.00000687, 0.00000188, 0.00001010, -0.00000479, -0.00002407, -0.00001010, 0.00002622, 0.00003261, -0.00000439, -0.00003578, -0.00002222, 0.00001270, 0.00002381, 0.00000709, -0.00000672, -0.00000206, 0.00000664, 0.00000499, -0.00000069, 0.00000076, 0.00000546, 0.00000331, -0.00000280, -0.00000206, 0.00000488, 0.00000808, 0.00000658, 0.00000634, 0.00000367, -0.00000716, -0.00001465, -0.00000503, 0.00000843, 0.00000363, -0.00000832, 0.00000104, 0.00002274, 0.00001988, -0.00000560, -0.00000964, 0.00001874, 0.00003538, 0.00001438, -0.00001066, -0.00000711, 0.00000478, -0.00000494, -0.00002305, -0.00002183, -0.00000814, -0.00000610, -0.00001426, -0.00001383, -0.00000347, 0.00000283, 0.00000102, -0.00000197, -0.00000357, -0.00000558, -0.00000655, -0.00000533, -0.00000498, -0.00000593, -0.00000373, 0.00000191, 0.00000484, 0.00000234, -0.00000118, -0.00000249, -0.00000339, -0.00000528, -0.00000673, },
		{ 0.00000281, 0.00000479, 0.00000211, -0.00000434, -0.00000855, -0.00000710, -0.00000287, 0.00000112, 0.00000508, 0.00000751, 0.00000476, -0.00000222, -0.00000782, -0.00000955, -0.00000875, -0.00000484, 0.00000249, 0.00000753, 0.00000431, -0.00000412, -0.00001079, -0.00001483, -0.00001770, -0.00001599, -0.00000799, -0.00000031, 0.00000151, 0.00000132, 0.00000285, 0.00000215, -0.00000229, -0.00000401, -0.00000108, -0.00000079, -0.00000493, -0.00000545, -0.00000074, 0.00000066, -0.00000251, -0.00000195, 0.00000166, -0.00000144, -0.00000931, -0.00000965, -0.00000302, -0.00000143, -0.00000444, -0.00000169, 0.00000346, -0.00000027, -0.00000734, -0.00000440, 0.00000261, -0.00000044, -0.00000676, -0.00000227, 0.00000356, -0.00000419, -0.00001064, 0.00000547, 0.00002492, 0.00001119, -0.00002497, -0.00003320, 0.00000051, 0.00003132, 0.00002186, -0.00000801, -0.00001720, -0.00000066, 0.00001369, 0.00000935, -0.00000157, -0.00000311, 0.00000220, 0.00000340, 0.00000083, 0.00000238, 0.00000626, 0.00000403, -0.00000209, -0.00000307, -0.00000194, -0.00000733, -0.00001124, -0.00000035, 0.00001339, 0.00000755, -0.00000707, 0.00000079, 0.00002463, 0.00002418, -0.00000387, -0.00001416, 0.00001076, 0.00002865, 0.00000966, -0.00001410, -0.00000711, 0.00001009, 0.00000206, -0.00002033, -0.00002438, -0.00001167, -0.00000698, -0.00001208, -0.00000940, 0.00000203, 0.00000676, 0.00000069, -0.00000639, -0.00000920, -0.00000951, -0.00000719, -0.00000234, 0.00000030, -0.00000131, -0.00000219, 0.00000044, 0.00000191, -0.00000091, -0.00000420, -0.00000455, -0.00000390, -0.00000431, -0.00000451, -0.00000317, -0.00000054, },
		{ -0.00000165, -0.00000594, -0.00000692, -0.00000251, 0.00000283, 0.00000463, 0.00000333, 0.00000028, -0.00000478, -0.00000954, -0.00000939, -0.00000427, 0.00000116, 0.00000444, 0.00000567, 0.00000246, -0.00000687, -0.00001670, -0.00001963, -0.00001617, -0.00001117, -0.00000517, 0.00000206, 0.00000562, 0.00000263, -0.00000128, -0.00000124, -0.00000092, -0.00000335, -0.00000410, -0.00000141, -0.00000125, -0.00000492, -0.00000497, 0.00000004, 0.00000165, -0.00000226, -0.00000381, -0.00000196, -0.00000449, -0.00001005, -0.00000831, -0.00000050, 0.00000157, -0.00000235, -0.00000205, 0.00000133, -0.00000159, -0.00000682, -0.00000372, 0.00000207, -0.00000138, -0.00000721, -0.00000237, 0.00000515, 0.00000150, -0.00000341, 0.00000489, 0.00001134, -0.00000476, -0.00002473, -0.00001314, 0.00002100, 0.00003113, 0.00000315, -0.00002355, -0.00001517, 0.00001176, 0.00002082, 0.00000589, -0.00000877, -0.00000616, 0.00000507, 0.00000903, 0.00000462, 0.00000074, -0.00000136, -0.00000654, -0.00001082, -0.00000602, 0.00000175, -0.00000012, -0.00000536, 0.00000212, 0.00001442, 0.00000922, -0.00000531, 0.00000196, 0.00002607, 0.00002496, -0.00000768, -0.00002360, 0.00000143, 0.00002567, 0.00001228, -0.00001086, -0.00000507, 0.00001208, 0.00000316, -0.00002205, -0.00002701, -0.00001063, -0.00000117, -0.00000526, -0.00000555, 0.00000164, 0.00000262, -0.00000565, -0.00001181, -0.00001056, -0.00000636, -0.00000170, 0.00000293, 0.00000346, -0.00000156, -0.00000567, -0.00000415, -0.00000143, -0.00000238, -0.00000438, -0.00000369, -0.00000190, -0.00000149, -0.00000155, -0.00000082, 0.00000023, 0.00000113, 0.00000107, },
		{ -0.00000286, 0.00000041, 0.00000262, 0.00000050, -0.00000368, -0.00000637, -0.00000717, -0.00000597, -0.00000165, 0.00000369, 0.00000512, 0.00000156, -0.00000381, -0.00000969, -0.00001613, -0.00001932, -0.00001517, -0.00000687, -0.00000082, 0.00000223, 0.00000438, 0.00000376, -0.00000085, -0.00000445, -0.00000328, -0.00000167, -0.00000330, -0.00000426, -0.00000199, -0.00000131, -0.00000395, -0.00000389, 0.00000007, 0.00000043, -0.00000447, -0.00000658, -0.00000395, -0.00000399, -0.00000709, -0.00000504, 0.00000127, 0.00000176, -0.00000330, -0.00000386, -0.00000032, -0.00000167, -0.00000565, -0.00000316, 0.00000159, -0.00000112, -0.00000515, 0.00000028, 0.00000697, 0.00000204, -0.00000562, -0.00000195, 0.00000320, -0.00000456, -0.00001096, 0.00000399, 0.00002333, 0.00001432, -0.00001541, -0.00002442, -0.00000007, 0.00002361, 0.00001734, -0.00000526, -0.00001324, -0.00000040, 0.00001310, 0.00001093, -0.00000157, -0.00001023, -0.00001185, -0.00001197, -0.00000947, 0.00000044, 0.00000971, 0.00000558, -0.00000394, 0.00000098, 0.00001487, 0.00001271, -0.00000227, 0.00000094, 0.00002169, 0.00001945, -0.00001424, -0.00003040, -0.00000121, 0.00003002, 0.00001906, -0.00000892, -0.00000877, 0.00000807, 0.00000251, -0.00001980, -0.00002250, -0.00000380, 0.00000536, -0.00000376, -0.00001066, -0.00000663, -0.00000407, -0.00000799, -0.00000909, -0.00000399, 0.00000070, 0.00000221, 0.00000254, 0.00000013, -0.00000599, -0.00000967, -0.00000629, -0.00000112, -0.00000059, -0.00000251, -0.00000200, 0.00000006, 0.00000056, -0.00000026, -0.00000097, -0.00000160, -0.00000190, -0.00000159, -0.00000194, -0.00000328, },
		{ -0.00000192, -0.00000425, -0.00000618, -0.00000480, -0.00000153, 0.00000105, 0.00000250, 0.00000196, -0.00000243, -0.00000913, -0.00001351, -0.00001424, -0.00001345, -0.00001090, -0.00000505, 0.00000114, 0.00000308, 0.00000182, 0.00000107, 0.00000004, -0.00000314, -0.00000531, -0.00000369, -0.00000205, -0.00000355, -0.00000437, -0.00000186, -0.00000064, -0.00000316, -0.00000428, -0.00000197, -0.00000209, -0.00000599, -0.00000654, -0.00000230, -0.00000098, -0.00000410, -0.00000404, -0.00000013, -0.00000041, -0.00000473, -0.00000443, -0.00000017, -0.00000063, -0.00000424, -0.00000235, 0.00000208, 0.00000028, -0.00000314, 0.00000071, 0.00000467, -0.00000163, -0.00000876, -0.00000332, 0.00000570, 0.00000371, -0.00000067, 0.00000615, 0.00001146, -0.00000248, -0.00001982, -0.00001140, 0.00001534, 0.00002413, 0.00000479, -0.00001366, -0.00000784, 0.00001001, 0.00001437, 0.00000013, -0.00001592, -0.00001897, -0.00001189, -0.00000556, -0.00000113, 0.00000672, 0.00001260, 0.00000604, -0.00000426, 0.00000160, 0.00001757, 0.00001559, -0.00000388, -0.00000695, 0.00001220, 0.00001432, -0.00001447, -0.00002863, 0.00000180, 0.00003423, 0.00002088, -0.00001311, -0.00001566, 0.00000638, 0.00000882, -0.00001033, -0.00001562, -0.00000166, 0.00000234, -0.00001163, -0.00002007, -0.00001192, -0.00000210, -0.00000060, -0.00000055, 0.00000197, 0.00000180, -0.00000160, -0.00000344, -0.00000444, -0.00000758, -0.00000901, -0.00000451, 0.00000104, 0.00000117, -0.00000154, -0.00000126, 0.00000121, 0.00000164, -0.00000031, -0.00000247, -0.00000384, -0.00000376, -0.00000188, -0.00000017, -0.00000067, -0.00000184, -0.00000170, },
		{ -0.00000230, 0.00000035, 0.00000223, 0.00000069, -0.00000329, -0.00000767, -0.00001192, -0.00001451, -0.00001282, -0.00000785, -0.00000381, -0.00000183, 0.00000021, 0.00000178, 0.00000061, -0.00000171, -0.00000200, -0.00000168, -0.00000355, -0.00000537, -0.00000417, -0.00000258, -0.00000345, -0.00000379, -0.00000146, -0.00000076, -0.00000401, -0.00000602, -0.00000389, -0.00000261, -0.00000437, -0.00000379, 0.00000011, 0.00000037, -0.00000423, -0.00000582, -0.00000257, -0.00000170, -0.00000421, -0.00000315, 0.00000096, 0.00000049, -0.00000300, -0.00000145, 0.00000258, 0.00000061, -0.00000366, -0.00000167, 0.00000139, -0.00000298, -0.00000654, 0.00000142, 0.00001092, 0.00000743, -0.00000148, -0.00000173, -0.00000003, -0.00000732, -0.00001116, 0.00000347, 0.00002064, 0.00001450, -0.00000715, -0.00001335, 0.00000194, 0.00001306, 0.00000254, -0.00001685, -0.00002341, -0.00001319, -0.00000020, 0.00000355, 0.00000213, 0.00000585, 0.00001151, 0.00000745, -0.00000114, 0.00000367, 0.00001682, 0.00001135, -0.00001195, -0.00001652, 0.00000719, 0.00001767, -0.00000692, -0.00002443, 0.00000065, 0.00003120, 0.00001805, -0.00001585, -0.00001582, 0.00001248, 0.00001902, -0.00000403, -0.00001827, -0.00001064, -0.00000720, -0.00001779, -0.00002063, -0.00000601, 0.00000776, 0.00000777, 0.00000219, -0.00000099, -0.00000428, -0.00000763, -0.00000661, -0.00000357, -0.00000430, -0.00000618, -0.00000339, 0.00000120, 0.00000140, -0.00000088, -0.00000014, 0.00000241, 0.00000190, -0.00000168, -0.00000480, -0.00000576, -0.00000443, -0.00000131, 0.00000088, -0.00000047, -0.00000360, -0.00000481, -0.00000406, -0.00000341, },
		{ -0.00000224, -0.00000722, -0.00001216, -0.00001401, -0.00001282, -0.00000953, -0.00000425, 0.00000100, 0.00000261, 0.00000064, -0.00000121, -0.00000172, -0.00000273, -0.00000381, -0.00000308, -0.00000219, -0.00000378, -0.00000565, -0.00000454, -0.00000258, -0.00000289, -0.00000343, -0.00000207, -0.00000205, -0.00000514, -0.00000645, -0.00000323, -0.00000057, -0.00000164, -0.00000207, -0.00000014, -0.00000120, -0.00000598, -0.00000717, -0.00000304, -0.00000078, -0.00000235, -0.00000162, 0.00000166, 0.00000106, -0.00000225, -0.00000121, 0.00000180, -0.00000085, -0.00000541, -0.00000318, 0.00000161, 0.00000030, -0.00000141, 0.00000495, 0.00001042, 0.00000320, -0.00000774, -0.00000770, -0.00000184, -0.00000140, 0.00000003, 0.00001022, 0.00001541, 0.00000225, -0.00001319, -0.00000855, 0.00000658, 0.00000539, -0.00001330, -0.00002459, -0.00001480, 0.00000298, 0.00000984, 0.00000338, -0.00000245, 0.00000377, 0.00001427, 0.00001248, 0.00000096, -0.00000058, 0.00000800, 0.00000276, -0.00001729, -0.00001771, 0.00001103, 0.00002529, -0.00000157, -0.00002690, -0.00000685, 0.00002709, 0.00002102, -0.00000947, -0.00000999, 0.00001680, 0.00001994, -0.00000984, -0.00002958, -0.00002012, -0.00000854, -0.00001131, -0.00001083, 0.00000283, 0.00001244, 0.00000639, -0.00000397, -0.00000771, -0.00000792, -0.00000744, -0.00000378, 0.00000005, -0.00000196, -0.00000625, -0.00000471, 0.00000095, 0.00000302, 0.00000139, 0.00000120, 0.00000199, -0.00000033, -0.00000486, -0.00000738, -0.00000664, -0.00000394, -0.00000059, 0.00000106, -0.00000100, -0.00000439, -0.00000468, -0.00000183, 0.00000050, 0.00000090, 0.00000023, },
		{ -0.00001423, -0.00001086, -0.00000538, -0.00000057, 0.00000287, 0.00000433, 0.00000233, -0.00000193, -0.00000457, -0.00000450, -0.00000421, -0.00000440, -0.00000362, -0.00000286, -0.00000415, -0.00000560, -0.00000433, -0.00000232, -0.00000277, -0.00000389, -0.00000307, -0.00000263, -0.00000442, -0.00000468, -0.00000126, 0.00000096, -0.00000112, -0.00000307, -0.00000220, -0.00000280, -0.00000610, -0.00000622, -0.00000176, 0.00000063, -0.00000107, -0.00000101, 0.00000166, 0.00000107, -0.00000230, -0.00000210, 0.00000008, -0.00000215, -0.00000519, -0.00000124, 0.00000499, 0.00000404, 0.00000023, 0.00000230, 0.00000436, -0.00000253, -0.00000986, -0.00000538, 0.00000408, 0.00000594, 0.00000434, 0.00000681, 0.00000502, -0.00000712, -0.00001439, -0.00000481, 0.00000493, -0.00000445, -0.00002046, -0.00001771, 0.00000219, 0.00001479, 0.00000833, -0.00000445, -0.00000556, 0.00000783, 0.00001976, 0.00001306, -0.00000569, -0.00001175, -0.00000110, 0.00000114, -0.00001245, -0.00001185, 0.00001457, 0.00002639, -0.00000354, -0.00003197, -0.00001022, 0.00003154, 0.00003162, -0.00000203, -0.00001195, 0.00000796, 0.00000968, -0.00001842, -0.00003378, -0.00001646, 0.00000256, 0.00000075, -0.00000471, 0.00000145, 0.00000640, -0.00000060, -0.00000860, -0.00000785, -0.00000432, -0.00000334, -0.00000178, -0.00000018, -0.00000323, -0.00000721, -0.00000403, 0.00000368, 0.00000621, 0.00000243, -0.00000102, -0.00000230, -0.00000462, -0.00000751, -0.00000763, -0.00000504, -0.00000208, 0.00000034, 0.00000134, -0.00000036, -0.00000287, -0.00000245, 0.00000029, 0.00000084, -0.00000232, -0.00000666, -0.00001051, -0.00001362, },
		{ 0.00000223, 0.00000408, 0.00000361, 0.00000204, -0.00000096, -0.00000465, -0.00000635, -0.00000557, -0.00000486, -0.00000503, -0.00000444, -0.00000344, -0.00000398, -0.00000496, -0.00000393, -0.00000238, -0.00000301, -0.00000406, -0.00000286, -0.00000153, -0.00000250, -0.00000303, -0.00000092, 0.00000009, -0.00000251, -0.00000445, -0.00000297, -0.00000226, -0.00000438, -0.00000451, -0.00000087, 0.00000100, -0.00000077, -0.00000098, 0.00000124, 0.00000043, -0.00000313, -0.00000320, -0.00000049, -0.00000093, -0.00000216, 0.00000203, 0.00000671, 0.00000315, -0.00000368, -0.00000350, -0.00000006, -0.00000218, -0.00000456, 0.00000154, 0.00000875, 0.00000613, -0.00000064, -0.00000221, -0.00000403, -0.00001111, -0.00001251, -0.00000254, 0.00000238, -0.00000848, -0.00001690, -0.00000455, 0.00001452, 0.00001534, -0.00000031, -0.00000987, -0.00000195, 0.00001348, 0.00001828, 0.00000373, -0.00001672, -0.00001818, 0.00000022, 0.00000884, -0.00000479, -0.00001070, 0.00000980, 0.00002213, -0.00000331, -0.00002841, -0.00000464, 0.00003873, 0.00003607, -0.00000716, -0.00002661, -0.00000672, 0.00000430, -0.00001364, -0.00002338, -0.00000466, 0.00001178, 0.00000294, -0.00001059, -0.00000772, 0.00000030, -0.00000106, -0.00000483, -0.00000276, -0.00000093, -0.00000357, -0.00000462, -0.00000236, -0.00000242, -0.00000389, -0.00000020, 0.00000600, 0.00000533, -0.00000209, -0.00000730, -0.00000719, -0.00000610, -0.00000577, -0.00000421, -0.00000175, -0.00000021, 0.00000087, 0.00000183, 0.00000099, -0.00000137, -0.00000243, -0.00000231, -0.00000455, -0.00000957, -0.00001343, -0.00001379, -0.00001169, -0.00000786, -0.00000251, },
		{ 0.00000046, -0.00000127, -0.00000315, -0.00000565, -0.00000702, -0.00000649, -0.00000588, -0.00000581, -0.00000479, -0.00000316, -0.00000310, -0.00000408, -0.00000364, -0.00000243, -0.00000269, -0.00000317, -0.00000169, -0.00000036, -0.00000173, -0.00000331, -0.00000231, -0.00000141, -0.00000304, -0.00000394, -0.00000185, -0.00000076, -0.00000293, -0.00000379, -0.00000103, 0.00000069, -0.00000084, -0.00000131, 0.00000031, -0.00000056, -0.00000349, -0.00000256, 0.00000143, 0.00000206, 0.00000035, 0.00000197, 0.00000352, -0.00000171, -0.00000799, -0.00000517, 0.00000246, 0.00000414, 0.00000210, 0.00000404, 0.00000519, -0.00000171, -0.00000944, -0.00000910, -0.00000690, -0.00000951, -0.00000885, -0.00000060, 0.00000208, -0.00000657, -0.00000895, 0.00000559, 0.00001790, 0.00000920, -0.00000777, -0.00000996, 0.00000243, 0.00001198, 0.00000788, -0.00000714, -0.00001931, -0.00001256, 0.00000830, 0.00001415, -0.00000530, -0.00001664, 0.00000440, 0.00002461, 0.00000677, -0.00001911, -0.00000285, 0.00003335, 0.00002608, -0.00002033, -0.00003843, -0.00000900, 0.00001466, 0.00000225, -0.00001212, -0.00000205, 0.00000766, -0.00000476, -0.00001824, -0.00001094, 0.00000340, 0.00000529, -0.00000019, -0.00000219, -0.00000357, -0.00000696, -0.00000581, 0.00000051, 0.00000317, 0.00000044, -0.00000028, 0.00000139, -0.00000170, -0.00000875, -0.00001091, -0.00000630, -0.00000170, -0.00000051, -0.00000031, 0.00000008, 0.00000002, 0.00000053, 0.00000181, 0.00000098, -0.00000300, -0.00000689, -0.00000885, -0.00001094, -0.00001345, -0.00001314, -0.00000889, -0.00000367, 0.00000024, 0.00000293, 0.00000401, 0.00000274, },
		{ -0.00000401, -0.00000633, -0.00000810, -0.00000778, -0.00000659, -0.00000557, -0.00000390, -0.00000201, -0.00000196, -0.00000328, -0.00000330, -0.00000209, -0.00000187, -0.00000218, -0.00000122, -0.00000059, -0.00000243, -0.00000427, -0.00000315, -0.00000138, -0.00000187, -0.00000237, -0.00000073, -0.00000015, -0.00000258, -0.00000382, -0.00000146, 0.00000022, -0.00000126, -0.00000202, -0.00000048, -0.00000043, -0.00000186, -0.00000003, 0.00000370, 0.00000287, -0.00000137, -0.00000233, -0.00000135, -0.00000414, -0.00000642, -0.00000038, 0.00000836, 0.00000834, 0.00000168, -0.00000215, -0.00000436, -0.00001015, -0.00001368, -0.00000895, -0.00000389, -0.00000556, -0.00000577, 0.00000101, 0.00000402, -0.00000174, -0.00000189, 0.00000977, 0.00001482, 0.00000147, -0.00001264, -0.00000937, 0.00000209, 0.00000474, -0.00000162, -0.00000863, -0.00001090, -0.00000333, 0.00001025, 0.00000997, -0.00000987, -0.00001770, 0.00000893, 0.00003450, 0.00001582, -0.00002030, -0.00001626, 0.00001654, 0.00001566, -0.00002197, -0.00003302, 0.00000234, 0.00002846, 0.00001062, -0.00001453, -0.00001161, -0.00000079, -0.00000754, -0.00001592, -0.00000574, 0.00000870, 0.00000739, -0.00000255, -0.00000636, -0.00000549, -0.00000483, -0.00000065, 0.00000603, 0.00000556, -0.00000288, -0.00000845, -0.00000742, -0.00000694, -0.00000880, -0.00000658, 0.00000019, 0.00000421, 0.00000277, 0.00000028, -0.00000068, -0.00000103, -0.00000068, -0.00000001, -0.00000221, -0.00000811, -0.00001306, -0.00001356, -0.00001170, -0.00000965, -0.00000611, -0.00000088, 0.00000284, 0.00000330, 0.00000228, 0.00000111, -0.00000053, -0.00000211, -0.00000289, },
		{ -0.00000912, -0.00000799, -0.00000563, -0.00000380, -0.00000221, -0.00000087, -0.00000123, -0.00000279, -0.00000302, -0.00000181, -0.00000147, -0.00000206, -0.00000177, -0.00000145, -0.00000288, -0.00000410, -0.00000251, -0.00000032, -0.00000058, -0.00000148, -0.00000060, -0.00000032, -0.00000258, -0.00000393, -0.00000205, -0.00000060, -0.00000183, -0.00000209, 0.00000031, 0.00000153, 0.00000061, 0.00000125, 0.00000257, -0.00000045, -0.00000563, -0.00000563, -0.00000150, -0.00000023, -0.00000029, 0.00000431, 0.00000886, 0.00000405, -0.00000626, -0.00001117, -0.00001090, -0.00001166, -0.00001073, -0.00000440, -0.00000013, -0.00000311, -0.00000393, 0.00000340, 0.00000798, 0.00000327, 0.00000119, 0.00000752, 0.00000738, -0.00000634, -0.00001602, -0.00000933, 0.00000022, -0.00000033, -0.00000364, -0.00000246, -0.00000167, -0.00000105, 0.00000442, 0.00000539, -0.00000609, -0.00000870, 0.00001597, 0.00003567, 0.00000934, -0.00003488, -0.00003324, 0.00000844, 0.00002228, -0.00000700, -0.00002032, 0.00000823, 0.00002824, 0.00000495, -0.00002362, -0.00001809, 0.00000053, -0.00000031, -0.00000933, -0.00000393, 0.00000620, 0.00000309, -0.00000558, -0.00000518, 0.00000023, 0.00000198, 0.00000267, 0.00000365, -0.00000183, -0.00001255, -0.00001625, -0.00000952, -0.00000260, -0.00000095, 0.00000084, 0.00000480, 0.00000552, 0.00000166, -0.00000164, -0.00000228, -0.00000274, -0.00000369, -0.00000465, -0.00000759, -0.00001270, -0.00001505, -0.00001139, -0.00000540, -0.00000149, 0.00000089, 0.00000303, 0.00000331, 0.00000107, -0.00000117, -0.00000187, -0.00000229, -0.00000317, -0.00000400, -0.00000533, -0.00000767, },
		{ -0.00000322, -0.00000176, -0.00000118, -0.00000071, -0.00000127, -0.00000271, -0.00000299, -0.00000191, -0.00000162, -0.00000234, -0.00000219, -0.00000154, -0.00000223, -0.00000301, -0.00000161, 0.00000026, -0.00000017, -0.00000134, -0.00000080, -0.00000055, -0.00000268, -0.00000431, -0.00000286, -0.00000113, -0.00000129, -0.00000053, 0.00000224, 0.00000299, 0.00000049, -0.00000124, -0.00000153, -0.00000393, -0.00000644, -0.00000310, 0.00000357, 0.00000516, 0.00000212, 0.00000135, 0.00000124, -0.00000487, -0.00001313, -0.00001411, -0.00000953, -0.00000720, -0.00000584, -0.00000127, 0.00000137, -0.00000146, -0.00000092, 0.00000747, 0.00001162, 0.00000431, -0.00000234, -0.00000031, -0.00000121, -0.00001116, -0.00001570, -0.00000710, 0.00000035, -0.00000178, -0.00000219, 0.00000315, 0.00000265, -0.00000275, 0.00000115, 0.00000912, 0.00000402, -0.00000327, 0.00000899, 0.00002053, -0.00000491, -0.00004328, -0.00003349, 0.00001845, 0.00003921, 0.00000589, -0.00002020, -0.00000153, 0.00001816, 0.00000013, -0.00002320, -0.00001336, 0.00000786, 0.00000492, -0.00001050, -0.00001000, 0.00000115, 0.00000334, -0.00000008, 0.00000261, 0.00000594, 0.00000145, -0.00000498, -0.00000743, -0.00001090, -0.00001602, -0.00001329, -0.00000159, 0.00000677, 0.00000575, 0.00000289, 0.00000332, 0.00000284, -0.00000071, -0.00000347, -0.00000435, -0.00000627, -0.00000893, -0.00000992, -0.00001031, -0.00001138, -0.00000990, -0.00000383, 0.00000229, 0.00000373, 0.00000220, 0.00000115, 0.00000012, -0.00000194, -0.00000334, -0.00000319, -0.00000333, -0.00000466, -0.00000598, -0.00000693, -0.00000801, -0.00000810, -0.00000600, },
		{ -0.00000166, -0.00000148, -0.00000161, -0.00000260, -0.00000292, -0.00000202, -0.00000165, -0.00000216, -0.00000187, -0.00000095, -0.00000132, -0.00000221, -0.00000138, 0.00000016, -0.00000015, -0.00000128, -0.00000102, -0.00000097, -0.00000307, -0.00000464, -0.00000290, -0.00000031, 0.00000044, 0.00000106, 0.00000237, 0.00000120, -0.00000276, -0.00000473, -0.00000333, -0.00000239, -0.00000187, 0.00000208, 0.00000653, 0.00000404, -0.00000360, -0.00000745, -0.00000706, -0.00000907, -0.00001229, -0.00001005, -0.00000475, -0.00000342, -0.00000373, -0.00000016, 0.00000325, 0.00000219, 0.00000323, 0.00000954, 0.00000985, -0.00000143, -0.00001057, -0.00000801, -0.00000494, -0.00000967, -0.00001143, -0.00000423, -0.00000006, -0.00000317, -0.00000074, 0.00000790, 0.00000700, -0.00000121, 0.00000263, 0.00001303, 0.00000521, -0.00001296, -0.00000977, 0.00000600, -0.00000483, -0.00003149, -0.00001936, 0.00002944, 0.00004465, 0.00000253, -0.00003197, -0.00001320, 0.00001596, 0.00000731, -0.00001443, -0.00000886, 0.00000751, 0.00000099, -0.00001594, -0.00001270, 0.00000484, 0.00001194, 0.00000758, 0.00000420, 0.00000005, -0.00000979, -0.00001662, -0.00001363, -0.00000870, -0.00000743, -0.00000300, 0.00000642, 0.00001045, 0.00000491, -0.00000068, -0.00000026, 0.00000063, -0.00000235, -0.00000601, -0.00000842, -0.00001123, -0.00001294, -0.00001053, -0.00000620, -0.00000384, -0.00000191, 0.00000204, 0.00000480, 0.00000285, -0.00000094, -0.00000228, -0.00000218, -0.00000319, -0.00000453, -0.00000491, -0.00000551, -0.00000680, -0.00000717, -0.00000627, -0.00000551, -0.00000487, -0.00000326, -0.00000157, -0.00000130, },
		{ -0.00000137, -0.00000203, -0.00000257, -0.00000191, -0.00000137, -0.00000163, -0.00000135, -0.00000053, -0.00000088, -0.00000194, -0.00000147, -0.00000004, -0.00000017, -0.00000135, -0.00000147, -0.00000149, -0.00000305, -0.00000393, -0.00000178, 0.00000089, 0.00000101, 0.00000001, -0.00000046, -0.00000213, -0.00000474, -0.00000414, 0.00000001, 0.00000254, 0.00000201, 0.00000210, 0.00000187, -0.00000364, -0.00001151, -0.00001298, -0.00000829, -0.00000585, -0.00000690, -0.00000560, -0.00000249, -0.00000236, -0.00000228, 0.00000263, 0.00000718, 0.00000551, 0.00000322, 0.00000480, 0.00000189, -0.00000940, -0.00001580, -0.00000909, -0.00000203, -0.00000505, -0.00000844, -0.00000449, -0.00000177, -0.00000268, 0.00000404, 0.00001521, 0.00001231, -0.00000113, -0.00000185, 0.00000589, -0.00000512, -0.00002629, -0.00001921, 0.00001019, 0.00001285, -0.00001351, -0.00001356, 0.00002342, 0.00003509, -0.00000548, -0.00003701, -0.00001256, 0.00002357, 0.00001632, -0.00001211, -0.00001431, 0.00000067, -0.00000137, -0.00001269, -0.00000550, 0.00001289, 0.00001599, 0.00000369, -0.00000668, -0.00001239, -0.00001797, -0.00001712, -0.00000624, 0.00000313, 0.00000310, 0.00000197, 0.00000573, 0.00000675, 0.00000079, -0.00000377, -0.00000209, -0.00000107, -0.00000545, -0.00001058, -0.00001262, -0.00001282, -0.00001064, -0.00000438, 0.00000206, 0.00000343, 0.00000171, 0.00000166, 0.00000207, -0.00000041, -0.00000361, -0.00000395, -0.00000323, -0.00000468, -0.00000703, -0.00000772, -0.00000725, -0.00000660, -0.00000493, -0.00000257, -0.00000163, -0.00000228, -0.00000257, -0.00000223, -0.00000242, -0.00000266, -0.00000194, },
		{ -0.00000226, -0.00000184, -0.00000111, -0.00000118, -0.00000107, -0.00000044, -0.00000073, -0.00000179, -0.00000158, -0.00000035, -0.00000043, -0.00000159, -0.00000171, -0.00000129, -0.00000206, -0.00000269, -0.00000128, 0.00000009, -0.00000108, -0.00000285, -0.00000283, -0.00000233, -0.00000210, 0.00000020, 0.00000381, 0.00000366, -0.00000091, -0.00000461, -0.00000618, -0.00000960, -0.00001338, -0.00001105, -0.00000409, -0.00000120, -0.00000353, -0.00000415, -0.00000167, -0.00000023, 0.00000146, 0.00000625, 0.00000840, 0.00000269, -0.00000377, -0.00000409, -0.00000497, -0.00001151, -0.00001364, -0.00000509, 0.00000146, -0.00000347, -0.00000884, -0.00000493, 0.00000056, 0.00000330, 0.00001118, 0.00001918, 0.00000954, -0.00001104, -0.00001509, -0.00000497, -0.00001008, -0.00002474, -0.00001075, 0.00002476, 0.00002674, -0.00001078, -0.00002468, 0.00000878, 0.00002922, -0.00000125, -0.00002994, -0.00000767, 0.00002578, 0.00001492, -0.00001784, -0.00002016, 0.00000137, 0.00000702, -0.00000262, -0.00000050, 0.00001007, 0.00000626, -0.00000926, -0.00001687, -0.00001457, -0.00001145, -0.00000563, 0.00000524, 0.00001054, 0.00000429, -0.00000207, 0.00000020, 0.00000301, -0.00000082, -0.00000501, -0.00000450, -0.00000499, -0.00000990, -0.00001331, -0.00001140, -0.00000737, -0.00000291, 0.00000296, 0.00000690, 0.00000458, -0.00000055, -0.00000202, -0.00000084, -0.00000194, -0.00000470, -0.00000560, -0.00000574, -0.00000773, -0.00000959, -0.00000832, -0.00000513, -0.00000250, -0.00000047, 0.00000084, -0.00000010, -0.00000277, -0.00000430, -0.00000385, -0.00000311, -0.00000251, -0.00000134, -0.00000056, -0.00000132, },
		{ -0.00000094, -0.00000086, -0.00000093, -0.00000046, -0.00000066, -0.00000173, -0.00000186, -0.00000085, -0.00000066, -0.00000139, -0.00000118, -0.00000050, -0.00000134, -0.00000286, -0.00000295, -0.00000254, -0.00000331, -0.00000349, -0.00000126, 0.00000120, 0.00000181, 0.00000201, 0.00000186, -0.00000176, -0.00000800, -0.00001096, -0.00000941, -0.00000855, -0.00000929, -0.00000661, -0.00000108, 0.00000047, -0.00000223, -0.00000229, 0.00000139, 0.00000353, 0.00000375, 0.00000468, 0.00000254, -0.00000560, -0.00001159, -0.00000854, -0.00000442, -0.00000719, -0.00000909, -0.00000308, 0.00000126, -0.00000360, -0.00000684, 0.00000006, 0.00000739, 0.00000837, 0.00001032, 0.00001101, -0.00000341, -0.00002392, -0.00002254, -0.00000313, 0.00000030, -0.00001167, -0.00000131, 0.00002786, 0.00002320, -0.00002028, -0.00003556, 0.00000503, 0.00003643, 0.00000977, -0.00002588, -0.00001277, 0.00001815, 0.00001079, -0.00001720, -0.00001481, 0.00001066, 0.00001551, -0.00000184, -0.00000966, -0.00000413, -0.00000560, -0.00001361, -0.00001202, -0.00000275, 0.00000169, 0.00000301, 0.00000722, 0.00000741, -0.00000109, -0.00000701, -0.00000250, 0.00000230, -0.00000187, -0.00000825, -0.00000920, -0.00000861, -0.00001012, -0.00000916, -0.00000377, 0.00000090, 0.00000272, 0.00000454, 0.00000529, 0.00000147, -0.00000370, -0.00000418, -0.00000190, -0.00000308, -0.00000720, -0.00000924, -0.00000893, -0.00000890, -0.00000801, -0.00000421, 0.00000019, 0.00000195, 0.00000158, 0.00000046, -0.00000192, -0.00000487, -0.00000568, -0.00000390, -0.00000209, -0.00000131, -0.00000056, -0.00000013, -0.00000109, -0.00000224, -0.00000189, },
		},
		{ { -0.00000316, -0.00000303, -0.00000602, -0.00000590, -0.00000096, -0.00000188, -0.00000364, -0.00000132, -0.00000028, -0.00000326, -0.00000391, -0.00000154, -0.00000361, -0.00000354, -0.00000613, -0.00001062, -0.00000642, -0.00001817, -0.00002562, -0.00000440, 0.00000487, 0.00001723, 0.00002701, 0.00000372, 0.00000450, 0.00001215, -0.00000118, 0.00000925, 0.00000414, -0.00000995, 0.00000752, 0.00000745, 0.00000211, 0.00000726, 0.00000252, 0.00000455, -0.00000325, -0.00000748, 0.00000707, 0.00000140, -0.00000384, 0.00000466, 0.00000742, 0.00000539, -0.00000464, 0.00000184, 0.00001072, -0.00000208, 0.00000245, 0.00000698, -0.00000353, -0.00000446, -0.00000816, 0.00000276, 0.00001487, -0.00000122, 0.00000130, 0.00001099, -0.00000451, -0.00001262, -0.00000800, 0.00001206, 0.00003017, -0.00000559, -0.00004141, -0.00000559, 0.00003017, 0.00001206, -0.00000800, -0.00001262, -0.00000451, 0.00001099, 0.00000130, -0.00000122, 0.00001487, 0.00000276, -0.00000816, -0.00000446, -0.00000353, 0.00000698, 0.00000245, -0.00000208, 0.00001072, 0.00000184, -0.00000464, 0.00000539, 0.00000742, 0.00000466, -0.00000384, 0.00000140, 0.00000707, -0.00000748, -0.00000325, 0.00000455, 0.00000252, 0.00000726, 0.00000211, 0.00000745, 0.00000752, -0.00000995, 0.00000414, 0.00000925, -0.00000118, 0.00001215, 0.00000450, 0.00000372, 0.00002701, 0.00001723, 0.00000487, -0.00000440, -0.00002562, -0.00001817, -0.00000642, -0.00001062, -0.00000613, -0.00000354, -0.00000361, -0.00000154, -0.00000391, -0.00000326, -0.00000028, -0.00000132, -0.00000364, -0.00000188, -0.00000096, -0.00000590, -0.00000602, -0.00000303, },
		{ -0.00000352, -0.00000617, -0.00000437, -0.00000121, -0.00000308, -0.00000280, -0.00000057, -0.00000110, -0.00000324, -0.00000228, -0.00000253, -0.00000513, -0.00000372, -0.00000851, -0.00001049, -0.00000803, -0.00002314, -0.00002216, -0.00000105, 0.00000871, 0.00002582, 0.00002150, -0.00000130, 0.00001036, 0.00001020, -0.00000021, 0.00000944, -0.00000261, -0.00000404, 0.00001140, 0.00000352, 0.00000467, 0.00000680, 0.00000080, 0.00000318, -0.00000447, -0.00000215, 0.00000484, -0.00000316, 0.00000273, 0.00000879, 0.00000436, 0.00000030, -0.00000164, 0.00000925, 0.00000534, -0.00000437, 0.00000747, 0.00000116, -0.00000887, -0.00000016, 0.00000192, 0.00000570, 0.00000206, 0.00000113, 0.00001945, 0.00000264, -0.00002824, -0.00000890, 0.00002205, 0.00002379, -0.00000293, -0.00003603, -0.00001559, 0.00003394, 0.00001742, -0.00002136, -0.00000608, 0.00000815, -0.00000007, -0.00000314, -0.00000307, 0.00001414, 0.00001413, -0.00001314, -0.00000934, 0.00000065, -0.00000005, 0.00000695, 0.00000089, 0.00000316, 0.00000720, -0.00000542, 0.00000160, 0.00000928, 0.00000294, -0.00000083, -0.00000233, 0.00000771, 0.00000145, -0.00001137, 0.00000255, 0.00000488, 0.00000277, 0.00000661, 0.00000216, 0.00000936, -0.00000058, -0.00000890, 0.00001222, 0.00000367, -0.00000103, 0.00001363, 0.00000099, 0.00001068, 0.00002615, 0.00000949, 0.00000437, -0.00001017, -0.00002715, -0.00001177, -0.00000626, -0.00001029, -0.00000390, -0.00000281, -0.00000232, -0.00000281, -0.00000516, -0.00000186, -0.00000011, -0.00000261, -0.00000323, -0.00000027, -0.00000227, -0.00000753, -0.00000494, -0.00000259, -0.00000333, },
		{ -0.00000616, -0.00000362, -0.00000185, -0.00000320, -0.00000182, -0.00000044, -0.00000131, -0.00000248, -0.00000232, -0.00000471, -0.00000525, -0.00000446, -0.00001116, -0.00000930, -0.00001077, -0.00002727, -0.00001825, 0.00000300, 0.00001631, 0.00002950, 0.00001196, 0.00000024, 0.00001617, 0.00000537, 0.00000059, 0.00000801, -0.00000459, 0.00000271, 0.00001028, 0.00000232, 0.00000732, 0.00000298, 0.00000059, 0.00000409, -0.00000469, -0.00000145, 0.00000154, 0.00000080, 0.00000889, 0.00000486, 0.00000087, 0.00000191, 0.00000367, 0.00000881, -0.00000046, -0.00000021, 0.00000601, -0.00000827, -0.00000248, 0.00001081, 0.00000046, -0.00000363, 0.00000364, 0.00002001, 0.00001566, -0.00002817, -0.00002666, 0.00002713, 0.00003192, -0.00000883, -0.00003134, -0.00001689, 0.00002558, 0.00002879, -0.00002346, -0.00002116, 0.00002343, 0.00000718, -0.00001723, -0.00000428, 0.00001080, 0.00001840, -0.00000366, -0.00001914, 0.00000083, 0.00000117, 0.00000030, 0.00000927, 0.00000097, 0.00000244, 0.00000079, -0.00000284, 0.00000867, 0.00000375, -0.00000201, 0.00000045, 0.00000227, 0.00000716, -0.00000653, -0.00000768, 0.00000752, 0.00000166, 0.00000440, 0.00000621, 0.00000201, 0.00000752, -0.00000627, -0.00000107, 0.00001376, -0.00000466, 0.00000446, 0.00001361, -0.00000169, 0.00001697, 0.00002183, 0.00000512, 0.00000297, -0.00001747, -0.00002404, -0.00000608, -0.00000814, -0.00000847, -0.00000143, -0.00000259, -0.00000275, -0.00000464, -0.00000452, -0.00000056, -0.00000106, -0.00000303, -0.00000167, 0.00000005, -0.00000485, -0.00000782, -0.00000333, -0.00000286, -0.00000340, -0.00000369, },
		{ -0.00000305, -0.00000209, -0.00000280, -0.00000105, -0.00000013, -0.00000115, -0.00000262, -0.00000384, -0.00000608, -0.00000467, -0.00000669, -0.00001241, -0.00000747, -0.00001593, -0.00003026, -0.00001249, 0.00000916, 0.00002342, 0.00002674, 0.00000522, 0.00000737, 0.00001634, -0.00000087, 0.00000380, 0.00000666, -0.00000412, 0.00000726, 0.00000811, 0.00000331, 0.00000620, -0.00000012, 0.00000463, 0.00000315, -0.00000793, -0.00000078, 0.00000426, 0.00000644, 0.00000661, 0.00000011, 0.00000407, 0.00000458, 0.00000415, 0.00000619, -0.00000125, 0.00000098, -0.00000193, -0.00000602, 0.00001289, 0.00000814, -0.00001206, 0.00000051, 0.00002159, 0.00002085, -0.00001466, -0.00003704, 0.00001217, 0.00004464, -0.00000464, -0.00003795, -0.00001218, 0.00001994, 0.00002756, -0.00001037, -0.00003538, 0.00001551, 0.00003037, -0.00002133, -0.00001784, 0.00001319, 0.00001428, 0.00000526, -0.00001513, -0.00001000, 0.00000579, -0.00000342, 0.00000660, 0.00001041, -0.00000471, 0.00000049, 0.00000147, 0.00000316, 0.00000585, -0.00000363, 0.00000130, 0.00000325, 0.00000223, 0.00000306, -0.00000911, -0.00000087, 0.00000660, -0.00000015, 0.00000809, 0.00000270, 0.00000136, 0.00000590, -0.00000699, 0.00000602, 0.00000742, -0.00000742, 0.00001334, 0.00000793, -0.00000206, 0.00002317, 0.00001558, 0.00000238, -0.00000078, -0.00002231, -0.00001743, -0.00000404, -0.00000975, -0.00000446, -0.00000057, -0.00000388, -0.00000366, -0.00000514, -0.00000315, -0.00000050, -0.00000167, -0.00000214, -0.00000069, -0.00000140, -0.00000695, -0.00000656, -0.00000241, -0.00000343, -0.00000285, -0.00000420, -0.00000648, },
		{ -0.00000219, -0.00000221, -0.00000008, 0.00000022, -0.00000178, -0.00000384, -0.00000537, -0.00000633, -0.00000475, -0.00000926, -0.00001143, -0.00000751, -0.00002318, -0.00002985, -0.00000467, 0.00001542, 0.00002746, 0.00002158, 0.00000427, 0.00001344, 0.00001031, -0.00000301, 0.00000882, 0.00000364, -0.00000211, 0.00001046, 0.00000580, 0.00000319, 0.00000382, 0.00000188, 0.00000764, -0.00000335, -0.00000889, 0.00000527, 0.00000800, 0.00000545, 0.00000171, 0.00000263, 0.00000804, 0.00000231, 0.00000446, 0.00000570, -0.00000383, -0.00000281, -0.00000176, 0.00000806, 0.00001555, -0.00000871, -0.00000999, 0.00002126, 0.00002372, -0.00000496, -0.00002979, -0.00000614, 0.00003967, 0.00001199, -0.00004187, -0.00001750, 0.00002498, 0.00002065, -0.00000393, -0.00002956, -0.00000619, 0.00003855, -0.00000057, -0.00003480, 0.00000782, 0.00001726, 0.00000055, -0.00000245, -0.00001290, -0.00000232, 0.00000250, -0.00000103, 0.00001430, 0.00000127, -0.00000933, 0.00000616, 0.00000388, 0.00000162, -0.00000105, -0.00000188, 0.00000679, -0.00000019, 0.00000132, 0.00000155, -0.00000844, 0.00000287, 0.00000377, 0.00000255, 0.00000801, -0.00000324, 0.00000486, 0.00000502, -0.00000692, 0.00000854, 0.00000028, -0.00000201, 0.00001735, -0.00000169, 0.00000416, 0.00002674, 0.00000676, 0.00000112, -0.00000497, -0.00002347, -0.00001116, -0.00000472, -0.00000835, -0.00000099, -0.00000274, -0.00000486, -0.00000357, -0.00000510, -0.00000244, -0.00000046, -0.00000107, -0.00000161, -0.00000113, -0.00000315, -0.00000762, -0.00000500, -0.00000242, -0.00000332, -0.00000235, -0.00000570, -0.00000644, -0.00000206, },
		{ -0.00000105, 0.00000088, -0.00000043, -0.00000336, -0.00000508, -0.00000631, -0.00000620, -0.00000552, -0.00001083, -0.00000980, -0.00001087, -0.00002963, -0.00002540, 0.00000297, 0.00002082, 0.00002965, 0.00001707, 0.00000604, 0.00001450, 0.00000372, 0.00000065, 0.00001132, -0.00000019, 0.00000239, 0.00001122, 0.00000224, 0.00000325, 0.00000438, 0.00000521, 0.00000407, -0.00000923, -0.00000195, 0.00001125, 0.00000559, 0.00000187, 0.00000269, 0.00000764, 0.00000595, 0.00000051, 0.00000777, 0.00000102, -0.00000925, 0.00000060, 0.00000888, 0.00001350, 0.00000134, -0.00001345, 0.00001197, 0.00002716, -0.00000057, -0.00001959, -0.00000797, 0.00002133, 0.00001873, -0.00002986, -0.00002858, 0.00002570, 0.00002454, -0.00001033, -0.00001909, -0.00001218, 0.00002120, 0.00002207, -0.00002906, -0.00001241, 0.00002330, -0.00000306, -0.00000382, 0.00000183, -0.00001287, -0.00000021, 0.00000311, 0.00000615, 0.00001064, -0.00001045, -0.00000165, 0.00001195, -0.00000207, -0.00000260, -0.00000075, 0.00000436, 0.00000444, -0.00000461, 0.00000542, -0.00000065, -0.00000856, 0.00000556, 0.00000312, 0.00000458, 0.00000222, -0.00000387, 0.00001175, -0.00000047, -0.00000691, 0.00000992, -0.00000320, 0.00000537, 0.00001316, -0.00000713, 0.00001535, 0.00002264, -0.00000156, 0.00000285, -0.00000968, -0.00002254, -0.00000620, -0.00000504, -0.00000543, -0.00000116, -0.00000540, -0.00000365, -0.00000381, -0.00000557, -0.00000134, 0.00000059, -0.00000091, -0.00000245, -0.00000158, -0.00000422, -0.00000770, -0.00000384, -0.00000236, -0.00000266, -0.00000317, -0.00000744, -0.00000508, -0.00000114, -0.00000239, },
		{ 0.00000069, -0.00000229, -0.00000486, -0.00000585, -0.00000693, -0.00000586, -0.00000645, -0.00001167, -0.00000921, -0.00001624, -0.00003330, -0.00001912, 0.00000977, 0.00002691, 0.00003039, 0.00001243, 0.00000758, 0.00001254, 0.00000093, 0.00000624, 0.00000975, -0.00000149, 0.00000721, 0.00000786, 0.00000016, 0.00000609, 0.00000554, 0.00000423, -0.00000171, -0.00000697, 0.00000722, 0.00000962, 0.00000105, 0.00000325, 0.00000653, 0.00000772, 0.00000241, 0.00000356, 0.00000657, -0.00000838, -0.00000544, 0.00001310, 0.00001212, 0.00000373, -0.00000555, 0.00000323, 0.00002307, 0.00000483, -0.00001567, 0.00000021, 0.00001311, 0.00000641, -0.00001706, -0.00002647, 0.00001486, 0.00003236, -0.00001149, -0.00002299, -0.00000216, 0.00000607, 0.00001819, -0.00000428, -0.00002368, 0.00001348, 0.00000592, -0.00001641, 0.00001124, -0.00000263, -0.00001684, 0.00000834, 0.00000584, 0.00000515, 0.00000056, -0.00000957, 0.00001047, 0.00000510, -0.00000958, 0.00000107, 0.00000309, 0.00000456, -0.00000172, -0.00000150, 0.00000903, -0.00000751, -0.00000549, 0.00000913, 0.00000105, 0.00000267, -0.00000198, 0.00000329, 0.00001297, -0.00001021, -0.00000159, 0.00001085, -0.00000549, 0.00001022, 0.00000588, -0.00000454, 0.00002349, 0.00001178, -0.00000369, 0.00000494, -0.00001633, -0.00001879, -0.00000152, -0.00000524, -0.00000432, -0.00000343, -0.00000525, -0.00000227, -0.00000583, -0.00000487, 0.00000111, 0.00000092, -0.00000247, -0.00000294, -0.00000109, -0.00000566, -0.00000763, -0.00000239, -0.00000206, -0.00000277, -0.00000515, -0.00000776, -0.00000285, -0.00000106, -0.00000207, 0.00000060, },
		{ -0.00000423, -0.00000558, -0.00000649, -0.00000717, -0.00000520, -0.00000786, -0.00001233, -0.00000965, -0.00002223, -0.00003484, -0.00001214, 0.00001792, 0.00003346, 0.00002742, 0.00000751, 0.00000949, 0.00001031, 0.00000153, 0.00001027, 0.00000659, -0.00000005, 0.00000869, 0.00000349, 0.00000276, 0.00000853, 0.00000307, 0.00000160, -0.00000225, -0.00000002, 0.00000962, 0.00000401, 0.00000195, 0.00000698, 0.00000674, 0.00000571, 0.00000271, 0.00000465, -0.00000159, -0.00000931, 0.00000983, 0.00001708, 0.00000148, -0.00000009, 0.00000648, 0.00001383, 0.00000607, -0.00001236, 0.00000391, 0.00001982, -0.00000528, -0.00002289, -0.00001419, 0.00000946, 0.00002853, -0.00000213, -0.00003082, -0.00000051, 0.00001185, 0.00000030, 0.00000548, -0.00000875, -0.00000404, 0.00001007, -0.00001759, -0.00000072, 0.00001873, -0.00002083, -0.00000613, 0.00001855, -0.00000287, 0.00000063, -0.00000067, -0.00000074, 0.00001047, -0.00000762, -0.00000551, 0.00000716, 0.00000074, 0.00000159, -0.00000237, 0.00000487, 0.00000489, -0.00001243, 0.00000318, 0.00000768, -0.00000387, 0.00000236, -0.00000021, 0.00000984, 0.00000469, -0.00001459, 0.00000897, 0.00000603, -0.00000669, 0.00001408, -0.00000043, 0.00000213, 0.00002425, 0.00000151, 0.00000034, 0.00000216, -0.00002191, -0.00001070, 0.00000077, -0.00000726, -0.00000439, -0.00000420, -0.00000353, -0.00000351, -0.00000729, -0.00000151, 0.00000264, -0.00000101, -0.00000388, -0.00000163, -0.00000131, -0.00000792, -0.00000616, -0.00000072, -0.00000272, -0.00000395, -0.00000660, -0.00000628, -0.00000115, -0.00000140, -0.00000066, 0.00000140, -0.00000123, },
		{ -0.00000596, -0.00000717, -0.00000663, -0.00000485, -0.00001011, -0.00001236, -0.00001104, -0.00002909, -0.00003460, -0.00000268, 0.00002817, 0.00003673, 0.00002019, 0.00000500, 0.00001231, 0.00000809, 0.00000399, 0.00001221, 0.00000341, 0.00000159, 0.00000765, 0.00000272, 0.00000683, 0.00000605, 0.00000013, 0.00000321, 0.00000076, 0.00000317, 0.00000670, 0.00000279, 0.00000614, 0.00000692, 0.00000571, 0.00000607, 0.00000192, -0.00000005, -0.00000470, 0.00000310, 0.00001849, 0.00000512, -0.00000257, 0.00001363, 0.00001334, -0.00000075, -0.00000993, 0.00000414, 0.00002669, -0.00000088, -0.00003784, -0.00001459, 0.00001901, 0.00002117, 0.00000117, -0.00002692, -0.00001081, 0.00002042, -0.00000141, -0.00000923, 0.00000867, -0.00000168, -0.00000162, -0.00000945, -0.00001476, 0.00002127, 0.00000037, -0.00002600, 0.00001828, 0.00000857, -0.00001504, 0.00000872, 0.00000228, -0.00000013, 0.00000185, -0.00000966, 0.00000474, 0.00000372, -0.00000308, 0.00000366, -0.00000036, 0.00000630, -0.00000209, -0.00000907, 0.00000906, -0.00000113, -0.00000393, 0.00000621, 0.00000156, 0.00000978, -0.00000491, -0.00000866, 0.00001513, -0.00000448, -0.00000153, 0.00001602, -0.00000663, 0.00000972, 0.00002025, -0.00000401, 0.00000453, -0.00000507, -0.00002114, -0.00000202, -0.00000229, -0.00000900, -0.00000298, -0.00000391, -0.00000347, -0.00000571, -0.00000536, 0.00000186, 0.00000107, -0.00000314, -0.00000313, -0.00000021, -0.00000359, -0.00000899, -0.00000305, -0.00000072, -0.00000449, -0.00000470, -0.00000668, -0.00000429, -0.00000042, -0.00000104, 0.00000084, 0.00000007, -0.00000373, -0.00000508, },
		{ -0.00000738, -0.00000554, -0.00000572, -0.00001242, -0.00001139, -0.00001463, -0.00003670, -0.00003051, 0.00001049, 0.00003721, 0.00003391, 0.00001233, 0.00000677, 0.00001425, 0.00000609, 0.00000773, 0.00001155, 0.00000009, 0.00000340, 0.00000752, 0.00000437, 0.00000686, 0.00000170, 0.00000267, 0.00000692, 0.00000057, 0.00000261, 0.00000594, 0.00000552, 0.00000710, 0.00000534, 0.00000742, 0.00000426, -0.00000312, -0.00000077, 0.00000371, 0.00001279, 0.00001027, -0.00000266, 0.00001287, 0.00002191, -0.00000419, -0.00001532, 0.00000483, 0.00002792, 0.00001025, -0.00004022, -0.00003143, 0.00002516, 0.00002785, -0.00000536, -0.00002092, -0.00001434, 0.00001456, 0.00001060, -0.00001997, 0.00000164, 0.00001711, -0.00000721, -0.00001251, -0.00001432, 0.00000579, 0.00001960, -0.00002110, -0.00000391, 0.00002517, -0.00001828, -0.00000511, 0.00002108, -0.00000779, -0.00000366, 0.00000044, -0.00000351, 0.00000605, -0.00000418, 0.00000177, 0.00000580, -0.00000231, 0.00000602, -0.00000478, -0.00000366, 0.00000568, -0.00000734, 0.00000429, 0.00000633, -0.00000026, 0.00000801, -0.00000896, 0.00000044, 0.00001128, -0.00001118, 0.00000963, 0.00001040, -0.00001049, 0.00001886, 0.00001307, -0.00000631, 0.00000611, -0.00001112, -0.00001496, 0.00000099, -0.00000772, -0.00000677, -0.00000162, -0.00000511, -0.00000420, -0.00000555, -0.00000198, 0.00000191, -0.00000143, -0.00000297, -0.00000175, -0.00000071, -0.00000611, -0.00000745, -0.00000082, -0.00000305, -0.00000530, -0.00000435, -0.00000632, -0.00000249, 0.00000021, -0.00000017, 0.00000064, -0.00000273, -0.00000502, -0.00000516, -0.00000666, },
		{ -0.00000487, -0.00000783, -0.00001366, -0.00001058, -0.00002149, -0.00004261, -0.00002054, 0.00002493, 0.00004106, 0.00002675, 0.00000810, 0.00001109, 0.00001395, 0.00000568, 0.00001100, 0.00000766, -0.00000152, 0.00000693, 0.00000767, 0.00000398, 0.00000420, 0.00000231, 0.00000865, 0.00000591, -0.00000232, 0.00000428, 0.00000792, 0.00000641, 0.00000593, 0.00000670, 0.00000715, -0.00000261, -0.00000314, 0.00000834, 0.00001033, 0.00000815, 0.00000250, 0.00000948, 0.00002602, 0.00000383, -0.00002286, -0.00000018, 0.00002857, 0.00001635, -0.00002964, -0.00004363, 0.00001234, 0.00004088, -0.00000357, -0.00002621, -0.00000904, 0.00000691, 0.00001273, -0.00001144, -0.00001548, 0.00002067, 0.00000840, -0.00002113, -0.00001384, -0.00000308, 0.00001570, 0.00000161, -0.00001826, 0.00001865, 0.00000076, -0.00002612, 0.00002394, 0.00000993, -0.00002267, 0.00000588, 0.00000252, -0.00000357, 0.00000112, -0.00000310, 0.00000925, -0.00000049, -0.00000227, 0.00000904, -0.00000779, -0.00000303, 0.00000062, -0.00000345, 0.00001051, -0.00000136, 0.00000137, 0.00000797, -0.00001091, 0.00000596, 0.00000353, -0.00000833, 0.00001711, -0.00000212, -0.00000547, 0.00002601, 0.00000157, -0.00000494, 0.00000692, -0.00001437, -0.00000945, -0.00000142, -0.00000927, -0.00000270, -0.00000367, -0.00000609, -0.00000300, -0.00000419, -0.00000086, 0.00000020, -0.00000162, -0.00000210, -0.00000183, -0.00000180, -0.00000692, -0.00000525, -0.00000134, -0.00000519, -0.00000406, -0.00000439, -0.00000585, -0.00000030, 0.00000068, -0.00000034, -0.00000138, -0.00000483, -0.00000488, -0.00000561, -0.00000734, -0.00000673, },
		{ -0.00001022, -0.00001372, -0.00001173, -0.00003053, -0.00004364, -0.00000572, 0.00003654, 0.00003918, 0.00001953, 0.00000843, 0.00001500, 0.00001221, 0.00000684, 0.00001137, 0.00000292, 0.00000100, 0.00001014, 0.00000502, 0.00000237, 0.00000481, 0.00000746, 0.00001043, 0.00000043, -0.00000174, 0.00000837, 0.00000801, 0.00000574, 0.00000683, 0.00000721, 0.00000104, -0.00000514, 0.00000773, 0.00001471, 0.00000432, 0.00000342, 0.00001192, 0.00002336, 0.00001320, -0.00002005, -0.00001152, 0.00002607, 0.00001975, -0.00002011, -0.00004001, -0.00000603, 0.00003847, 0.00001158, -0.00003145, -0.00001148, 0.00001042, 0.00000503, -0.00000318, -0.00001571, 0.00000394, 0.00002218, -0.00001234, -0.00002303, -0.00000381, 0.00000459, 0.00001107, -0.00000687, -0.00000102, 0.00001330, -0.00002396, 0.00000089, 0.00003312, -0.00002255, -0.00001342, 0.00002031, -0.00000826, -0.00000544, 0.00000306, 0.00000293, 0.00000705, -0.00000746, 0.00000690, 0.00000718, -0.00001558, -0.00000088, 0.00000156, 0.00000258, 0.00000649, -0.00000684, 0.00001008, 0.00000313, -0.00001299, 0.00001040, -0.00000219, -0.00000138, 0.00001512, -0.00001152, 0.00000788, 0.00002338, -0.00000980, 0.00000286, 0.00000532, -0.00001770, -0.00000610, -0.00000351, -0.00000694, -0.00000223, -0.00000683, -0.00000345, -0.00000175, -0.00000477, -0.00000099, 0.00000025, -0.00000087, -0.00000303, -0.00000182, -0.00000186, -0.00000743, -0.00000450, -0.00000278, -0.00000500, -0.00000273, -0.00000551, -0.00000390, 0.00000179, -0.00000045, -0.00000176, -0.00000334, -0.00000532, -0.00000473, -0.00000657, -0.00000716, -0.00000586, -0.00000532, },
		{ -0.00001357, -0.00001540, -0.00003919, -0.00003876, 0.00001059, 0.00004317, 0.00003414, 0.00001509, 0.00001155, 0.00001690, 0.00001013, 0.00000791, 0.00000928, 0.00000118, 0.00000571, 0.00000910, 0.00000124, 0.00000439, 0.00000904, 0.00001034, 0.00000645, -0.00000279, 0.00000303, 0.00001015, 0.00000665, 0.00000695, 0.00000698, 0.00000269, -0.00000273, 0.00000396, 0.00001718, 0.00000732, -0.00000101, 0.00001449, 0.00002385, 0.00001470, -0.00000993, -0.00001685, 0.00001531, 0.00002191, -0.00001480, -0.00003146, -0.00001084, 0.00002181, 0.00001988, -0.00002170, -0.00002141, 0.00001317, 0.00000548, -0.00000855, -0.00000563, -0.00000529, 0.00001290, 0.00000686, -0.00002311, -0.00001177, 0.00000184, 0.00000422, 0.00000790, -0.00000355, 0.00000383, -0.00000722, -0.00001710, 0.00002917, 0.00000400, -0.00003552, 0.00001851, 0.00001059, -0.00002282, 0.00000651, 0.00000603, 0.00000137, 0.00000086, -0.00000330, 0.00001477, -0.00000683, -0.00001759, 0.00000827, 0.00000160, 0.00000220, -0.00000004, -0.00000177, 0.00001560, -0.00000920, -0.00000837, 0.00001454, -0.00000812, 0.00000461, 0.00000853, -0.00001171, 0.00001940, 0.00001103, -0.00001179, 0.00001273, -0.00000323, -0.00001943, -0.00000160, -0.00000418, -0.00000587, -0.00000471, -0.00000580, 0.00000023, -0.00000421, -0.00000558, 0.00000080, 0.00000103, -0.00000218, -0.00000391, 0.00000027, -0.00000297, -0.00000912, -0.00000374, -0.00000296, -0.00000385, -0.00000300, -0.00000567, -0.00000055, 0.00000137, -0.00000302, -0.00000259, -0.00000429, -0.00000532, -0.00000522, -0.00000696, -0.00000638, -0.00000577, -0.00000649, -0.00001222, },
		{ -0.00002079, -0.00004584, -0.00002913, 0.00002582, 0.00004546, 0.00002871, 0.00001383, 0.00001518, 0.00001640, 0.00000797, 0.00000846, 0.00000749, 0.00000253, 0.00000783, 0.00000490, 0.00000179, 0.00001003, 0.00001124, 0.00000841, 0.00000233, -0.00000122, 0.00000714, 0.00000947, 0.00000726, 0.00000733, 0.00000269, -0.00000020, 0.00000375, 0.00001427, 0.00001261, -0.00000129, 0.00001063, 0.00002806, 0.00001503, -0.00000512, -0.00001138, 0.00000457, 0.00001659, -0.00000997, -0.00002695, -0.00000508, 0.00001190, 0.00001003, -0.00000854, -0.00002070, 0.00000483, 0.00001247, -0.00001294, -0.00000688, 0.00000278, -0.00000166, 0.00000829, -0.00000554, -0.00001933, -0.00000303, 0.00000005, 0.00000677, 0.00000891, -0.00000469, -0.00000419, -0.00001236, 0.00000755, 0.00002281, -0.00002862, -0.00000824, 0.00003157, -0.00002151, -0.00001206, 0.00002163, -0.00000534, -0.00000042, 0.00000316, 0.00000405, 0.00000839, -0.00002029, -0.00000473, 0.00001310, -0.00000578, 0.00000190, -0.00000016, 0.00000620, 0.00000953, -0.00001803, 0.00000503, 0.00001060, -0.00001379, 0.00001191, 0.00000152, -0.00000680, 0.00002301, -0.00000104, -0.00000362, 0.00001440, -0.00001539, -0.00001405, 0.00000246, -0.00000706, -0.00000631, -0.00000471, -0.00000179, -0.00000069, -0.00000800, -0.00000270, 0.00000289, -0.00000057, -0.00000396, -0.00000182, 0.00000171, -0.00000721, -0.00000949, -0.00000137, -0.00000311, -0.00000346, -0.00000338, -0.00000378, 0.00000112, -0.00000199, -0.00000421, -0.00000226, -0.00000538, -0.00000548, -0.00000546, -0.00000658, -0.00000624, -0.00000631, -0.00000768, -0.00001416, -0.00001372, },
		{ -0.00005037, -0.00001573, 0.00003929, 0.00004471, 0.00002420, 0.00001475, 0.00001744, 0.00001375, 0.00000666, 0.00000955, 0.00000670, 0.00000358, 0.00000663, 0.00000329, 0.00000739, 0.00001350, 0.00000906, 0.00000582, 0.00000133, 0.00000169, 0.00000910, 0.00000968, 0.00000761, 0.00000318, 0.00000004, 0.00000590, 0.00001227, 0.00001280, 0.00000400, 0.00000671, 0.00002746, 0.00002052, -0.00000533, -0.00000623, 0.00000438, 0.00000610, -0.00000992, -0.00002367, -0.00000114, 0.00001582, -0.00000205, -0.00001147, -0.00000868, -0.00000087, 0.00001033, -0.00000672, -0.00001541, 0.00000733, 0.00000136, -0.00000679, 0.00000536, -0.00000872, -0.00001364, -0.00000160, 0.00000129, 0.00001263, 0.00000328, -0.00001149, -0.00000554, -0.00000122, 0.00001530, -0.00000493, -0.00002540, 0.00002442, 0.00000416, -0.00003370, 0.00002099, 0.00001036, -0.00001864, 0.00001287, 0.00000458, 0.00000095, -0.00000281, -0.00001789, 0.00001007, 0.00000311, -0.00000974, 0.00000836, 0.00000021, 0.00000801, -0.00000058, -0.00001461, 0.00001537, -0.00000327, -0.00001037, 0.00001889, -0.00000833, 0.00000051, 0.00002222, -0.00000711, 0.00000498, 0.00000600, -0.00002081, -0.00000385, 0.00000016, -0.00001046, -0.00000385, -0.00000243, -0.00000086, -0.00000512, -0.00000737, 0.00000174, 0.00000149, -0.00000273, -0.00000287, 0.00000100, -0.00000108, -0.00001140, -0.00000596, 0.00000019, -0.00000449, -0.00000244, -0.00000263, -0.00000235, -0.00000067, -0.00000488, -0.00000276, -0.00000289, -0.00000687, -0.00000474, -0.00000529, -0.00000666, -0.00000688, -0.00000647, -0.00000935, -0.00001620, -0.00001362, -0.00002772, },
		{ 0.00000203, 0.00005054, 0.00004137, 0.00002094, 0.00001642, 0.00001733, 0.00001056, 0.00000746, 0.00001071, 0.00000503, 0.00000334, 0.00000631, 0.00000652, 0.00001245, 0.00001180, 0.00000634, 0.00000543, 0.00000153, 0.00000426, 0.00001109, 0.00000947, 0.00000390, 0.00000003, 0.00000627, 0.00001379, 0.00001078, 0.00000686, 0.00000915, 0.00002288, 0.00002479, -0.00000155, -0.00000800, 0.00000910, 0.00000271, -0.00001760, -0.00002234, -0.00000006, 0.00002221, -0.00000065, -0.00002502, -0.00000593, 0.00000759, 0.00000295, -0.00000332, -0.00001465, -0.00000081, 0.00001113, -0.00001119, -0.00000563, 0.00000773, -0.00001163, -0.00001087, 0.00000066, 0.00000679, 0.00001120, -0.00000782, -0.00001026, 0.00000219, 0.00000332, 0.00000307, -0.00001608, 0.00000186, 0.00002124, -0.00002836, -0.00000454, 0.00003264, -0.00002251, -0.00000362, 0.00002632, -0.00001113, -0.00000194, -0.00000333, -0.00000857, 0.00001018, -0.00000950, 0.00000025, 0.00001118, -0.00000503, 0.00000900, -0.00000593, -0.00000618, 0.00001356, -0.00001457, 0.00000395, 0.00001553, -0.00001798, 0.00001358, 0.00001801, -0.00001092, 0.00000869, -0.00000345, -0.00001690, 0.00000147, -0.00000671, -0.00000809, -0.00000002, -0.00000308, -0.00000310, -0.00000682, -0.00000315, 0.00000237, -0.00000132, -0.00000168, -0.00000089, 0.00000048, -0.00000588, -0.00001116, -0.00000136, -0.00000148, -0.00000454, 0.00000003, -0.00000312, -0.00000328, -0.00000316, -0.00000464, -0.00000140, -0.00000549, -0.00000668, -0.00000323, -0.00000616, -0.00000743, -0.00000707, -0.00000646, -0.00001232, -0.00001695, -0.00001350, -0.00003766, -0.00005191, },
		{ 0.00005748, 0.00003564, 0.00001905, 0.00001751, 0.00001542, 0.00000900, 0.00000969, 0.00000978, 0.00000250, 0.00000475, 0.00000921, 0.00001052, 0.00001280, 0.00000834, 0.00000628, 0.00000557, 0.00000190, 0.00000770, 0.00001193, 0.00000564, 0.00000057, 0.00000532, 0.00001499, 0.00001189, 0.00000549, 0.00001282, 0.00002268, 0.00002332, 0.00000494, -0.00000921, 0.00000767, 0.00000808, -0.00002309, -0.00002855, 0.00000034, 0.00002475, 0.00000861, -0.00002950, -0.00001912, 0.00001581, 0.00000719, -0.00000962, -0.00000940, -0.00000586, 0.00000852, -0.00000050, -0.00001748, 0.00000474, 0.00000543, -0.00001609, -0.00000514, 0.00000479, 0.00000751, 0.00000276, -0.00001118, -0.00000070, 0.00000302, -0.00000401, -0.00000402, -0.00000475, 0.00001336, -0.00000667, -0.00002371, 0.00002906, 0.00000195, -0.00003127, 0.00003243, 0.00000776, -0.00002700, 0.00001319, -0.00000365, -0.00000843, 0.00000367, -0.00000824, 0.00001210, 0.00000063, -0.00000543, 0.00001403, -0.00001150, -0.00000094, 0.00000655, -0.00001396, 0.00001602, 0.00000007, -0.00001543, 0.00002843, 0.00000494, -0.00001249, 0.00001159, -0.00000917, -0.00001166, -0.00000037, -0.00000897, -0.00000155, -0.00000141, -0.00000574, -0.00000317, -0.00000542, -0.00000114, 0.00000028, -0.00000052, 0.00000023, -0.00000206, -0.00000173, -0.00000815, -0.00000796, 0.00000018, -0.00000354, -0.00000129, 0.00000037, -0.00000631, -0.00000404, -0.00000362, -0.00000325, -0.00000249, -0.00000738, -0.00000414, -0.00000344, -0.00000812, -0.00000740, -0.00000670, -0.00000766, -0.00001553, -0.00001501, -0.00001640, -0.00005115, -0.00004677, 0.00002422, },
		{ 0.00002891, 0.00001859, 0.00001744, 0.00001347, 0.00000948, 0.00001071, 0.00000663, 0.00000238, 0.00000943, 0.00001231, 0.00001114, 0.00001037, 0.00000709, 0.00000758, 0.00000506, 0.00000378, 0.00001047, 0.00000880, 0.00000196, 0.00000474, 0.00001410, 0.00001449, 0.00000546, 0.00001211, 0.00002665, 0.00002266, 0.00000681, -0.00000551, 0.00000250, 0.00001076, -0.00001835, -0.00003738, -0.00000505, 0.00002613, 0.00001533, -0.00002170, -0.00003097, 0.00000764, 0.00001947, -0.00001066, -0.00001342, -0.00000168, 0.00000089, 0.00000409, -0.00001167, -0.00001086, 0.00001211, -0.00000333, -0.00001452, 0.00000238, 0.00000349, 0.00000348, -0.00000125, -0.00000556, 0.00000244, -0.00000614, -0.00000619, 0.00000252, 0.00000312, 0.00000132, -0.00001820, 0.00000657, 0.00002364, -0.00003179, 0.00000415, 0.00004020, -0.00003363, -0.00000804, 0.00002854, -0.00002220, -0.00000705, 0.00000621, -0.00000135, 0.00000988, -0.00001043, 0.00000768, 0.00001168, -0.00002071, 0.00000638, 0.00000168, -0.00000823, 0.00001641, -0.00001326, 0.00000175, 0.00003059, -0.00001369, -0.00000453, 0.00001336, -0.00001543, -0.00000858, -0.00000168, -0.00000507, 0.00000009, -0.00000680, -0.00000403, -0.00000167, -0.00000584, -0.00000120, 0.00000098, 0.00000202, -0.00000204, -0.00000403, -0.00000167, -0.00000878, -0.00000563, 0.00000008, -0.00000246, 0.00000135, -0.00000406, -0.00000841, -0.00000236, -0.00000378, -0.00000312, -0.00000420, -0.00000651, -0.00000230, -0.00000625, -0.00000873, -0.00000649, -0.00000723, -0.00000980, -0.00001672, -0.00001229, -0.00002620, -0.00006421, -0.00003072, 0.00004732, 0.00005753, },
		{ 0.00001897, 0.00001653, 0.00001262, 0.00001021, 0.00000934, 0.00000465, 0.00000670, 0.00001422, 0.00001189, 0.00000931, 0.00000905, 0.00000805, 0.00000840, 0.00000503, 0.00000613, 0.00000959, 0.00000510, 0.00000503, 0.00001276, 0.00001525, 0.00000831, 0.00001027, 0.00002788, 0.00002733, 0.00000620, -0.00000335, 0.00000143, 0.00000665, -0.00001105, -0.00003741, -0.00001671, 0.00002382, 0.00001976, -0.00001262, -0.00002942, -0.00000791, 0.00002010, 0.00000109, -0.00001969, -0.00000213, 0.00000258, -0.00000246, -0.00000227, -0.00001337, -0.00000118, 0.00001098, -0.00000940, -0.00000736, 0.00000304, -0.00000161, 0.00000412, 0.00000067, -0.00000361, -0.00000443, -0.00001014, 0.00000279, 0.00000590, -0.00000339, -0.00000784, -0.00000446, 0.00001962, -0.00000831, -0.00002385, 0.00004069, -0.00000058, -0.00004401, 0.00003612, 0.00000690, -0.00004028, 0.00001405, 0.00000659, -0.00000414, 0.00000269, -0.00000629, 0.00001975, -0.00000617, -0.00002042, 0.00001807, -0.00000698, -0.00000376, 0.00001218, -0.00001445, 0.00001843, 0.00001544, -0.00002220, 0.00001112, 0.00000570, -0.00002130, -0.00000232, -0.00000013, -0.00000335, -0.00000443, -0.00000737, 0.00000085, -0.00000445, -0.00000692, 0.00000163, 0.00000338, 0.00000030, -0.00000616, -0.00000174, -0.00000152, -0.00001077, -0.00000289, 0.00000145, -0.00000087, -0.00000096, -0.00000909, -0.00000586, -0.00000126, -0.00000547, -0.00000266, -0.00000451, -0.00000506, -0.00000343, -0.00000868, -0.00000717, -0.00000692, -0.00000834, -0.00001098, -0.00001620, -0.00001327, -0.00004249, -0.00006933, -0.00000404, 0.00006488, 0.00005076, 0.00002367, },
		{ 0.00001554, 0.00001237, 0.00000959, 0.00000786, 0.00000690, 0.00001270, 0.00001486, 0.00000869, 0.00000853, 0.00000982, 0.00000952, 0.00000877, 0.00000522, 0.00000627, 0.00000751, 0.00000713, 0.00001208, 0.00001453, 0.00001062, 0.00001139, 0.00002584, 0.00003212, 0.00001007, -0.00000549, 0.00000200, 0.00000340, -0.00001054, -0.00002986, -0.00002369, 0.00001357, 0.00002227, -0.00000633, -0.00002169, -0.00001330, 0.00000687, 0.00000945, -0.00001357, -0.00001039, 0.00000622, -0.00000426, -0.00000524, -0.00000338, -0.00001148, 0.00000512, 0.00000754, -0.00001010, -0.00000389, -0.00000225, -0.00000031, 0.00000927, -0.00000198, -0.00000886, -0.00000839, -0.00000292, 0.00000856, -0.00000204, -0.00000805, -0.00000149, 0.00000725, 0.00000568, -0.00002203, 0.00001145, 0.00003108, -0.00004260, -0.00000112, 0.00004786, -0.00004202, -0.00001896, 0.00003496, -0.00001391, -0.00000439, 0.00000595, 0.00000271, 0.00001442, -0.00002311, -0.00000226, 0.00001909, -0.00002077, 0.00000663, 0.00000907, -0.00001079, 0.00002325, -0.00000260, -0.00001397, 0.00001910, -0.00001034, -0.00001725, 0.00000608, -0.00000266, -0.00000599, -0.00000638, -0.00000217, 0.00000046, -0.00000964, -0.00000263, 0.00000548, 0.00000154, -0.00000428, -0.00000551, 0.00000233, -0.00000579, -0.00001101, 0.00000274, 0.00000152, -0.00000286, -0.00000532, -0.00000933, -0.00000209, -0.00000387, -0.00000570, -0.00000054, -0.00000544, -0.00000520, -0.00000537, -0.00000856, -0.00000628, -0.00000884, -0.00000753, -0.00001184, -0.00001706, -0.00001983, -0.00005977, -0.00006153, 0.00002717, 0.00007233, 0.00004073, 0.00002128, 0.00001915, },
		{ 0.00001158, 0.00000869, 0.00000949, 0.00001195, 0.00001534, 0.00001119, 0.00000658, 0.00001025, 0.00001136, 0.00001069, 0.00000836, 0.00000401, 0.00000581, 0.00000983, 0.00001306, 0.00001390, 0.00001099, 0.00001369, 0.00002533, 0.00003298, 0.00001708, -0.00000564, -0.00000174, 0.00000363, -0.00001297, -0.00002590, -0.00002050, 0.00000257, 0.00001747, -0.00000099, -0.00001590, -0.00000846, -0.00000283, 0.00000209, -0.00000229, -0.00001133, 0.00000067, 0.00000132, -0.00001067, -0.00000131, -0.00000499, -0.00000943, 0.00001111, 0.00000408, -0.00001182, -0.00000544, -0.00000294, 0.00000763, 0.00000775, -0.00001111, -0.00001046, -0.00000345, 0.00000345, 0.00000329, -0.00000811, -0.00000137, 0.00000457, 0.00000278, -0.00000726, -0.00000748, 0.00002648, -0.00000849, -0.00003442, 0.00004465, 0.00000250, -0.00005766, 0.00003192, 0.00001517, -0.00003417, 0.00001697, 0.00000640, 0.00000049, 0.00000278, -0.00002154, 0.00001602, 0.00000060, -0.00002245, 0.00002392, -0.00000157, -0.00000791, 0.00002261, -0.00001193, -0.00000156, 0.00001243, -0.00001964, -0.00000344, 0.00000595, -0.00000925, -0.00000532, -0.00000348, -0.00000002, -0.00000539, -0.00000849, 0.00000458, 0.00000338, -0.00000288, -0.00000496, -0.00000119, 0.00000152, -0.00001107, -0.00000491, 0.00000651, -0.00000351, -0.00000548, -0.00000654, -0.00000677, -0.00000209, -0.00000669, -0.00000183, -0.00000040, -0.00000843, -0.00000486, -0.00000604, -0.00000801, -0.00000751, -0.00000876, -0.00000543, -0.00001602, -0.00001979, -0.00002966, -0.00007271, -0.00004149, 0.00005565, 0.00006989, 0.00003143, 0.00002104, 0.00001866, 0.00001485, },
		{ 0.00001031, 0.00001403, 0.00001493, 0.00001295, 0.00000737, 0.00000802, 0.00001309, 0.00001247, 0.00001086, 0.00000622, 0.00000293, 0.00000926, 0.00001551, 0.00001466, 0.00001069, 0.00001442, 0.00002714, 0.00003319, 0.00002148, -0.00000092, -0.00000678, 0.00000124, -0.00001191, -0.00002751, -0.00001606, 0.00000093, 0.00000733, -0.00000064, -0.00001112, -0.00000320, -0.00000065, -0.00000952, -0.00000358, -0.00000147, -0.00000466, 0.00000094, -0.00000705, -0.00000827, 0.00000201, -0.00000953, -0.00000272, 0.00001584, -0.00000407, -0.00001489, -0.00000372, 0.00000301, 0.00001065, -0.00000286, -0.00001535, -0.00000384, 0.00000055, 0.00000138, -0.00000230, -0.00000377, 0.00000447, -0.00000009, -0.00000274, -0.00000430, 0.00000598, 0.00001099, -0.00002640, 0.00000726, 0.00003763, -0.00004773, -0.00001324, 0.00004946, -0.00003500, -0.00001238, 0.00003729, -0.00001545, 0.00000093, 0.00000148, -0.00001265, 0.00001587, -0.00001823, -0.00000209, 0.00002853, -0.00002121, 0.00000368, 0.00002114, -0.00001801, 0.00000501, 0.00000283, -0.00001534, 0.00000524, -0.00000381, -0.00000935, -0.00000058, -0.00000349, -0.00000276, -0.00000736, -0.00000138, 0.00000522, -0.00000259, -0.00000256, -0.00000259, 0.00000017, -0.00000271, -0.00001029, 0.00000292, 0.00000217, -0.00000897, -0.00000386, -0.00000646, -0.00000612, -0.00000384, -0.00000508, 0.00000188, -0.00000517, -0.00000962, -0.00000263, -0.00000750, -0.00000850, -0.00000786, -0.00000574, -0.00000732, -0.00002353, -0.00002082, -0.00004153, -0.00007927, -0.00001217, 0.00007686, 0.00006027, 0.00002485, 0.00002172, 0.00001770, 0.00001405, 0.00001055, },
		{ 0.00001711, 0.00001298, 0.00000910, 0.00000720, 0.00001185, 0.00001505, 0.00001247, 0.00000937, 0.00000395, 0.00000585, 0.00001588, 0.00001741, 0.00001139, 0.00001373, 0.00002836, 0.00003582, 0.00002359, 0.00000379, -0.00000751, -0.00000509, -0.00001066, -0.00002735, -0.00001777, 0.00000509, 0.00000304, -0.00000790, -0.00000881, 0.00000040, 0.00000553, -0.00001081, -0.00001561, 0.00000251, 0.00000181, -0.00000518, -0.00000347, -0.00000922, -0.00000277, -0.00000015, -0.00001036, 0.00000773, 0.00001208, -0.00001484, -0.00001167, 0.00000262, 0.00000598, 0.00000443, -0.00001114, -0.00000952, 0.00000120, -0.00000258, -0.00000084, -0.00000020, 0.00000118, 0.00000069, -0.00000426, 0.00000111, -0.00000062, 0.00000306, -0.00000337, -0.00001110, 0.00002707, -0.00000600, -0.00004371, 0.00003767, 0.00000562, -0.00005047, 0.00003990, 0.00001335, -0.00003535, 0.00002357, -0.00000394, -0.00001289, 0.00000903, -0.00001764, 0.00001981, 0.00000847, -0.00002883, 0.00002567, 0.00000821, -0.00002353, 0.00001214, -0.00000128, -0.00000915, 0.00000196, -0.00000924, -0.00000148, -0.00000170, -0.00000698, -0.00000223, -0.00000399, 0.00000154, -0.00000050, -0.00000289, 0.00000096, -0.00000378, -0.00000041, -0.00000368, -0.00000562, 0.00000376, -0.00000609, -0.00000753, -0.00000148, -0.00000935, -0.00000551, -0.00000298, -0.00000175, 0.00000022, -0.00001036, -0.00000595, -0.00000243, -0.00001063, -0.00000716, -0.00000597, -0.00000472, -0.00001554, -0.00002801, -0.00001957, -0.00005783, -0.00007672, 0.00002383, 0.00008747, 0.00004637, 0.00002204, 0.00002252, 0.00001616, 0.00001328, 0.00001158, 0.00001505, },
		{ 0.00000898, 0.00000846, 0.00001036, 0.00001517, 0.00001522, 0.00001128, 0.00000744, 0.00000494, 0.00001240, 0.00001943, 0.00001439, 0.00001335, 0.00002769, 0.00003915, 0.00002732, 0.00000574, -0.00000590, -0.00000964, -0.00001389, -0.00002432, -0.00002073, 0.00000476, 0.00000752, -0.00001398, -0.00001403, 0.00000299, 0.00000998, -0.00000423, -0.00002204, -0.00000751, 0.00001013, -0.00000281, -0.00000800, -0.00000461, -0.00000787, 0.00000005, -0.00000286, -0.00000518, 0.00001289, 0.00000037, -0.00001810, -0.00000207, 0.00000455, 0.00000199, -0.00000183, -0.00001027, -0.00000295, -0.00000179, -0.00000490, 0.00000320, 0.00000209, -0.00000133, -0.00000345, 0.00000026, 0.00000300, -0.00000626, 0.00000028, -0.00000006, 0.00000358, 0.00001156, -0.00002806, -0.00000108, 0.00003355, -0.00004426, -0.00000284, 0.00005471, -0.00004178, -0.00000854, 0.00004204, -0.00003264, -0.00000618, 0.00001202, -0.00000997, 0.00002154, -0.00001565, -0.00001063, 0.00003592, -0.00001792, -0.00001533, 0.00002169, -0.00000814, -0.00000870, -0.00000161, -0.00000439, 0.00000114, -0.00000935, -0.00000394, 0.00000177, -0.00000428, -0.00000128, -0.00000214, 0.00000223, -0.00000110, -0.00000617, 0.00000332, -0.00000332, -0.00000492, -0.00000082, -0.00000875, -0.00000219, -0.00000491, -0.00001149, -0.00000114, -0.00000203, -0.00000168, -0.00000412, -0.00001048, -0.00000187, -0.00000685, -0.00001065, -0.00000319, -0.00000628, -0.00000903, -0.00002442, -0.00002495, -0.00002266, -0.00007796, -0.00005955, 0.00006090, 0.00008462, 0.00003274, 0.00002348, 0.00002192, 0.00001432, 0.00001457, 0.00001578, 0.00001900, 0.00001503, },
		{ 0.00001139, 0.00001349, 0.00001653, 0.00001416, 0.00000996, 0.00000761, 0.00000945, 0.00001736, 0.00001816, 0.00001524, 0.00002629, 0.00004081, 0.00003301, 0.00000798, -0.00000645, -0.00001081, -0.00001828, -0.00002401, -0.00001987, -0.00000006, 0.00001115, -0.00001061, -0.00002253, -0.00000038, 0.00001371, 0.00000209, -0.00001809, -0.00001868, 0.00000550, 0.00000761, -0.00000988, -0.00000638, -0.00000487, -0.00000667, 0.00000251, -0.00000320, -0.00000057, 0.00000978, -0.00000860, -0.00001156, 0.00000330, -0.00000105, -0.00000078, -0.00000192, -0.00000716, -0.00000310, -0.00000552, -0.00000014, 0.00000642, -0.00000198, -0.00000413, -0.00000025, 0.00000365, -0.00000464, -0.00000834, 0.00000827, 0.00000292, -0.00000070, -0.00000389, -0.00001347, 0.00001988, -0.00000963, -0.00003675, 0.00004963, 0.00000313, -0.00005659, 0.00005007, 0.00000975, -0.00005508, 0.00002723, 0.00000692, -0.00001263, 0.00001487, -0.00002090, 0.00001344, 0.00001933, -0.00003360, 0.00000972, 0.00001806, -0.00002025, -0.00000428, 0.00000135, -0.00000138, -0.00000600, -0.00000990, 0.00000548, -0.00000153, -0.00000853, -0.00000031, 0.00000177, 0.00000252, -0.00000717, -0.00000092, 0.00000743, -0.00000896, -0.00000652, -0.00000240, -0.00000626, -0.00000173, -0.00001078, -0.00000668, 0.00000247, -0.00000529, -0.00000247, -0.00000600, -0.00000796, -0.00000246, -0.00001052, -0.00000506, -0.00000283, -0.00001132, -0.00001442, -0.00002801, -0.00001878, -0.00003630, -0.00009359, -0.00002594, 0.00008922, 0.00007022, 0.00002502, 0.00002664, 0.00001879, 0.00001497, 0.00001929, 0.00001963, 0.00001772, 0.00001006, 0.00000833, },
		{ 0.00001489, 0.00001686, 0.00001301, 0.00001012, 0.00001040, 0.00001371, 0.00001852, 0.00001927, 0.00002649, 0.00004078, 0.00003815, 0.00001275, -0.00000832, -0.00001288, -0.00001963, -0.00002670, -0.00001888, -0.00000253, 0.00000854, -0.00000339, -0.00002350, -0.00001033, 0.00001359, 0.00000749, -0.00001147, -0.00001979, -0.00000693, 0.00001002, -0.00000106, -0.00001129, -0.00000360, -0.00000697, -0.00000278, 0.00000526, -0.00000536, 0.00000061, 0.00000576, -0.00000976, -0.00000515, -0.00000053, -0.00000508, 0.00000281, -0.00000175, -0.00000827, -0.00000483, -0.00000309, 0.00000521, 0.00000253, -0.00000585, -0.00000074, 0.00000262, -0.00000172, -0.00001024, 0.00000157, 0.00001257, -0.00000664, -0.00000489, 0.00000078, 0.00000112, 0.00000321, -0.00002936, 0.00001087, 0.00004079, -0.00005311, -0.00000132, 0.00006492, -0.00005617, -0.00002136, 0.00005480, -0.00002769, -0.00000689, 0.00001721, -0.00001814, 0.00002039, -0.00000449, -0.00002208, 0.00002731, -0.00000412, -0.00002093, 0.00000744, -0.00000019, -0.00000626, -0.00000933, 0.00000024, 0.00000644, -0.00001200, -0.00000539, 0.00000497, 0.00000154, -0.00000255, -0.00000610, 0.00000880, 0.00000123, -0.00001660, -0.00000229, -0.00000181, -0.00000633, -0.00000520, -0.00001045, 0.00000074, -0.00000150, -0.00000818, 0.00000011, -0.00000763, -0.00000733, -0.00000443, -0.00000831, -0.00000086, -0.00001020, -0.00001504, -0.00001664, -0.00002790, -0.00001657, -0.00005836, -0.00009534, 0.00001721, 0.00010181, 0.00005228, 0.00002458, 0.00002706, 0.00001616, 0.00002082, 0.00002363, 0.00001816, 0.00001281, 0.00000836, 0.00001199, 0.00001404, },
		{ 0.00001716, 0.00001255, 0.00001227, 0.00001338, 0.00001557, 0.00002134, 0.00002960, 0.00004117, 0.00004145, 0.00001855, -0.00000836, -0.00001753, -0.00002021, -0.00002760, -0.00002120, -0.00000265, 0.00000517, -0.00000148, -0.00001656, -0.00001687, 0.00000549, 0.00001109, -0.00000570, -0.00001494, -0.00001272, 0.00000046, 0.00000673, -0.00000701, -0.00000885, -0.00000380, -0.00000673, 0.00000404, 0.00000283, -0.00000971, 0.00000384, 0.00000475, -0.00000974, -0.00000550, -0.00000538, -0.00000110, 0.00000597, -0.00000660, -0.00000913, -0.00000229, 0.00000085, 0.00000439, -0.00000259, -0.00000337, 0.00000225, -0.00000150, -0.00000691, -0.00000477, 0.00001047, 0.00000278, -0.00001340, 0.00000501, 0.00000582, -0.00000580, -0.00001285, -0.00000917, 0.00003281, -0.00001084, -0.00004447, 0.00005886, 0.00000517, -0.00007503, 0.00005007, 0.00002239, -0.00005763, 0.00002959, 0.00000896, -0.00002353, 0.00002125, -0.00001288, -0.00000233, 0.00001934, -0.00002126, -0.00000334, 0.00000945, -0.00001009, -0.00000642, -0.00000286, 0.00000568, -0.00000483, -0.00001396, 0.00000528, 0.00000316, -0.00000292, -0.00000224, 0.00000182, 0.00000931, -0.00001244, -0.00001413, 0.00000456, -0.00000663, -0.00000763, -0.00000496, -0.00000572, 0.00000132, -0.00000854, -0.00000383, 0.00000129, -0.00001248, -0.00000534, -0.00000315, -0.00000516, -0.00000510, -0.00001812, -0.00001199, -0.00001961, -0.00002729, -0.00002011, -0.00008277, -0.00008037, 0.00006023, 0.00009972, 0.00003774, 0.00002709, 0.00002485, 0.00001926, 0.00002814, 0.00002128, 0.00001282, 0.00001094, 0.00001171, 0.00001542, 0.00001435, 0.00001616, },
		{ 0.00001339, 0.00001517, 0.00001464, 0.00001911, 0.00003256, 0.00004450, 0.00004408, 0.00002323, -0.00000642, -0.00002212, -0.00002342, -0.00002614, -0.00002351, -0.00000515, 0.00000486, -0.00000337, -0.00001155, -0.00001380, -0.00000420, 0.00000742, -0.00000030, -0.00001013, -0.00001021, -0.00000799, 0.00000221, 0.00000260, -0.00000949, -0.00000692, -0.00000403, -0.00000200, 0.00000686, -0.00000548, -0.00000885, 0.00001067, 0.00000093, -0.00001343, -0.00000708, -0.00000375, 0.00000489, 0.00000226, -0.00001120, -0.00000494, 0.00000048, 0.00000060, 0.00000156, -0.00000262, -0.00000029, -0.00000114, -0.00000535, -0.00000393, 0.00000223, 0.00000719, -0.00000783, -0.00000490, 0.00001350, -0.00000450, -0.00001383, -0.00000536, 0.00001047, 0.00001159, -0.00003588, 0.00000986, 0.00005095, -0.00006017, -0.00001393, 0.00007310, -0.00005178, -0.00002541, 0.00006099, -0.00003236, -0.00001511, 0.00003113, -0.00001777, 0.00000359, 0.00000392, -0.00001536, 0.00000988, -0.00000533, -0.00001089, 0.00000261, -0.00000071, -0.00000042, -0.00001193, -0.00000325, 0.00000784, -0.00000537, 0.00000037, 0.00000318, 0.00000365, -0.00000032, -0.00001844, -0.00000299, 0.00000225, -0.00001327, -0.00000256, -0.00000308, -0.00000510, -0.00000322, -0.00000881, 0.00000294, -0.00000538, -0.00001434, 0.00000164, -0.00000351, -0.00000777, -0.00001287, -0.00001767, -0.00000797, -0.00002740, -0.00002344, -0.00003049, -0.00010579, -0.00004890, 0.00009715, 0.00008681, 0.00002789, 0.00002997, 0.00002516, 0.00002678, 0.00002836, 0.00001321, 0.00001139, 0.00001436, 0.00001524, 0.00001583, 0.00001478, 0.00001845, 0.00001698, },
		{ 0.00001700, 0.00001697, 0.00003106, 0.00004951, 0.00004911, 0.00002665, -0.00000461, -0.00002500, -0.00002829, -0.00002591, -0.00002286, -0.00000950, 0.00000390, -0.00000305, -0.00001200, -0.00000820, -0.00000552, -0.00000200, 0.00000006, -0.00000517, -0.00000612, -0.00000767, -0.00000704, 0.00000322, -0.00000122, -0.00001081, -0.00000429, -0.00000180, 0.00000176, 0.00000244, -0.00001118, 0.00000005, 0.00001297, -0.00000899, -0.00001532, -0.00000387, 0.00000060, 0.00000553, -0.00000401, -0.00000963, -0.00000028, -0.00000127, -0.00000033, 0.00000196, -0.00000175, -0.00000244, -0.00000451, -0.00000220, -0.00000018, 0.00000113, 0.00000078, -0.00000630, 0.00000620, 0.00000652, -0.00001497, -0.00000578, 0.00000456, 0.00000477, -0.00001039, -0.00001528, 0.00003746, -0.00000500, -0.00005427, 0.00005444, 0.00001160, -0.00007744, 0.00005103, 0.00002710, -0.00006806, 0.00003071, 0.00002488, -0.00003400, 0.00000902, 0.00000183, -0.00000452, 0.00000200, -0.00001419, 0.00000433, 0.00000270, -0.00000839, -0.00000336, -0.00000726, 0.00000364, -0.00000119, -0.00000439, 0.00001034, 0.00000014, -0.00000173, -0.00000646, -0.00001389, 0.00000230, -0.00000763, -0.00000991, 0.00000446, -0.00000796, -0.00000665, -0.00000395, -0.00000411, 0.00000229, -0.00001364, -0.00000612, 0.00000565, -0.00001131, -0.00001163, -0.00001544, -0.00001194, -0.00001155, -0.00003398, -0.00001539, -0.00005365, -0.00012064, -0.00000019, 0.00012160, 0.00006522, 0.00002431, 0.00003548, 0.00002866, 0.00002992, 0.00002001, 0.00000906, 0.00001683, 0.00001742, 0.00001571, 0.00001607, 0.00001719, 0.00002009, 0.00001656, 0.00001584, },
		{ 0.00002727, 0.00005100, 0.00005750, 0.00003153, -0.00000431, -0.00002741, -0.00003247, -0.00002791, -0.00002145, -0.00001181, -0.00000003, -0.00000138, -0.00001227, -0.00000781, -0.00000035, -0.00000655, -0.00000748, -0.00000264, -0.00000190, -0.00000349, -0.00000936, -0.00000526, 0.00000364, -0.00000580, -0.00000915, -0.00000004, -0.00000088, 0.00000092, -0.00000293, -0.00000860, 0.00000862, 0.00000529, -0.00001760, -0.00001013, 0.00000040, 0.00000161, 0.00000293, -0.00000633, -0.00000541, -0.00000070, -0.00000356, 0.00000239, 0.00000188, -0.00000482, -0.00000464, -0.00000152, 0.00000114, -0.00000343, -0.00000119, 0.00000197, -0.00000070, 0.00000595, -0.00000421, -0.00001072, 0.00000340, -0.00000160, -0.00000480, -0.00000645, 0.00000737, 0.00001834, -0.00003352, 0.00000179, 0.00004875, -0.00005923, -0.00001549, 0.00007800, -0.00005347, -0.00003422, 0.00007233, -0.00002363, -0.00003351, 0.00002906, -0.00000245, -0.00000816, -0.00000716, -0.00000154, 0.00001175, -0.00001335, -0.00000754, 0.00000300, -0.00000552, -0.00000027, -0.00000429, 0.00000851, 0.00000887, -0.00001104, -0.00000053, -0.00000645, -0.00001111, -0.00000095, -0.00001067, 0.00000098, 0.00000070, -0.00001512, -0.00000174, -0.00000277, -0.00000327, -0.00000300, -0.00001262, 0.00000436, -0.00000292, -0.00001897, -0.00000924, -0.00001491, -0.00000929, -0.00002028, -0.00003156, -0.00001330, -0.00009053, -0.00011320, 0.00005948, 0.00012448, 0.00004186, 0.00003158, 0.00004175, 0.00002819, 0.00002512, 0.00001290, 0.00001416, 0.00002209, 0.00001631, 0.00001665, 0.00001834, 0.00001931, 0.00002040, 0.00001767, 0.00001866, 0.00001877, },
		{ 0.00006462, 0.00004076, -0.00000402, -0.00003139, -0.00003593, -0.00003014, -0.00002166, -0.00001211, -0.00000385, -0.00000227, -0.00000980, -0.00000989, 0.00000181, -0.00000226, -0.00001474, -0.00000804, 0.00000190, 0.00000090, -0.00000560, -0.00001067, -0.00000227, 0.00000158, -0.00000818, -0.00000382, 0.00000139, -0.00000347, -0.00000011, -0.00000379, -0.00000309, 0.00000896, -0.00000593, -0.00001730, -0.00000261, 0.00000024, 0.00000122, 0.00000202, -0.00000629, -0.00000426, -0.00000314, -0.00000146, 0.00000459, -0.00000278, -0.00000707, -0.00000125, 0.00000194, -0.00000274, -0.00000690, 0.00000380, 0.00000450, -0.00000211, 0.00000156, -0.00000471, -0.00000309, -0.00000154, -0.00000905, 0.00000131, 0.00000196, 0.00000455, -0.00000241, -0.00001442, 0.00002874, -0.00000950, -0.00005395, 0.00005689, 0.00001451, -0.00008290, 0.00005020, 0.00004115, -0.00007122, 0.00001281, 0.00003320, -0.00002482, -0.00000760, 0.00000163, 0.00001010, -0.00000430, -0.00002068, 0.00000909, 0.00000026, -0.00001211, 0.00000096, 0.00000481, 0.00001323, -0.00000712, -0.00001147, 0.00000768, -0.00001276, -0.00001186, -0.00000068, -0.00000522, 0.00000392, -0.00001123, -0.00001138, 0.00000497, -0.00000804, -0.00000348, -0.00000327, -0.00000616, 0.00000352, -0.00001684, -0.00001521, -0.00000598, -0.00001751, -0.00000886, -0.00002683, -0.00002416, -0.00002839, -0.00012705, -0.00007477, 0.00011180, 0.00010542, 0.00003058, 0.00004661, 0.00003923, 0.00002148, 0.00002078, 0.00001432, 0.00002153, 0.00002143, 0.00001587, 0.00002056, 0.00001948, 0.00002007, 0.00002193, 0.00002011, 0.00002120, 0.00002690, 0.00004803, },
		{ 0.00000014, -0.00003711, -0.00004057, -0.00003148, -0.00002263, -0.00001302, -0.00000564, -0.00000476, -0.00000815, -0.00000935, -0.00000077, 0.00000273, -0.00001308, -0.00001764, -0.00000060, 0.00000602, -0.00000087, -0.00000886, -0.00000995, -0.00000029, -0.00000068, -0.00000621, 0.00000008, -0.00000235, -0.00000454, 0.00000232, -0.00000371, -0.00000137, 0.00000405, -0.00001170, -0.00001108, 0.00000011, -0.00000176, 0.00000355, 0.00000112, -0.00000794, -0.00000457, -0.00000256, 0.00000168, 0.00000155, -0.00000709, -0.00000338, 0.00000250, -0.00000124, -0.00000789, -0.00000207, 0.00000962, -0.00000060, -0.00000470, 0.00000391, -0.00000231, -0.00000597, -0.00000900, -0.00000218, 0.00000831, -0.00000483, -0.00000142, 0.00000285, 0.00000434, 0.00000566, -0.00003721, 0.00000586, 0.00005126, -0.00005960, -0.00001825, 0.00008290, -0.00004483, -0.00004526, 0.00006125, -0.00001031, -0.00003239, 0.00001493, 0.00000658, -0.00000004, -0.00001923, -0.00000324, 0.00001583, -0.00001792, -0.00000719, 0.00001240, 0.00000743, 0.00000201, -0.00001608, 0.00000291, 0.00000445, -0.00002440, -0.00000345, 0.00000351, -0.00000462, -0.00000275, -0.00001506, 0.00000040, 0.00000032, -0.00001375, 0.00000420, -0.00000245, -0.00000629, -0.00000585, -0.00002076, -0.00000517, -0.00001174, -0.00001819, -0.00000678, -0.00003282, -0.00002070, -0.00005828, -0.00014530, -0.00001437, 0.00014125, 0.00007970, 0.00003682, 0.00005396, 0.00002540, 0.00001873, 0.00002343, 0.00001865, 0.00002389, 0.00001977, 0.00002011, 0.00002265, 0.00001816, 0.00002318, 0.00002493, 0.00002216, 0.00002885, 0.00004661, 0.00006645, 0.00005179, },
		{ -0.00004798, -0.00003294, -0.00002273, -0.00001500, -0.00000719, -0.00000601, -0.00000874, -0.00000765, -0.00000251, 0.00000231, -0.00000649, -0.00002071, -0.00001058, 0.00000747, 0.00000503, -0.00000468, -0.00001103, -0.00000829, 0.00000150, -0.00000077, -0.00000370, -0.00000127, -0.00000588, -0.00000014, 0.00000360, -0.00000733, -0.00000150, 0.00000074, -0.00001200, -0.00000632, -0.00000101, -0.00000036, 0.00000611, -0.00000322, -0.00000930, -0.00000266, -0.00000103, 0.00000102, -0.00000276, -0.00000541, 0.00000127, -0.00000002, -0.00000626, -0.00000577, 0.00000572, 0.00000683, -0.00000724, 0.00000067, 0.00000668, -0.00000814, -0.00001126, -0.00000454, 0.00000703, 0.00000057, -0.00001109, 0.00000915, 0.00000930, -0.00000613, -0.00001615, -0.00001215, 0.00003398, -0.00000988, -0.00005285, 0.00005743, 0.00001831, -0.00007992, 0.00003740, 0.00003665, -0.00005696, 0.00000916, 0.00002476, -0.00001306, -0.00000856, -0.00001050, 0.00001302, -0.00000309, -0.00002381, 0.00001539, 0.00001162, -0.00000357, -0.00000538, -0.00000836, 0.00001057, -0.00001341, -0.00002179, 0.00001180, -0.00000280, -0.00000934, -0.00000485, -0.00000907, 0.00000360, -0.00001142, -0.00000567, 0.00001215, -0.00001175, -0.00001006, -0.00001027, -0.00001511, -0.00000316, -0.00002098, -0.00000910, -0.00001063, -0.00004147, -0.00002103, -0.00009410, -0.00013948, 0.00005226, 0.00014933, 0.00006175, 0.00004731, 0.00004420, 0.00001447, 0.00002707, 0.00002601, 0.00001935, 0.00002526, 0.00002243, 0.00002355, 0.00002005, 0.00002017, 0.00002934, 0.00002605, 0.00002871, 0.00004896, 0.00006732, 0.00005860, 0.00000853, -0.00004110, },
		{ -0.00002183, -0.00001665, -0.00000990, -0.00000639, -0.00000910, -0.00000778, -0.00000192, 0.00000014, -0.00000381, -0.00001561, -0.00001835, 0.00000020, 0.00001008, 0.00000067, -0.00000817, -0.00001207, -0.00000493, 0.00000424, -0.00000115, -0.00000439, -0.00000403, -0.00000385, 0.00000482, -0.00000156, -0.00001111, 0.00000149, -0.00000096, -0.00001173, -0.00000411, -0.00000091, 0.00000318, 0.00000430, -0.00000826, -0.00000692, -0.00000083, -0.00000229, -0.00000080, -0.00000279, -0.00000166, 0.00000016, -0.00000465, -0.00000536, 0.00000007, 0.00000708, -0.00000019, -0.00000564, 0.00000778, 0.00000002, -0.00001526, -0.00000672, 0.00000340, 0.00000365, -0.00000898, -0.00000081, 0.00001832, -0.00000458, -0.00001844, -0.00000534, 0.00001188, 0.00000751, -0.00003700, 0.00001046, 0.00004908, -0.00005786, -0.00001663, 0.00006885, -0.00004375, -0.00002942, 0.00005420, -0.00001512, -0.00002217, 0.00000752, -0.00000212, 0.00000484, -0.00001947, -0.00000123, 0.00002514, -0.00000958, -0.00000525, 0.00000024, -0.00000298, 0.00000193, -0.00002317, -0.00000229, 0.00001211, -0.00001758, -0.00000407, -0.00000172, -0.00000775, -0.00000246, -0.00001199, 0.00001040, 0.00000318, -0.00002413, -0.00000493, -0.00001046, -0.00001205, -0.00000863, -0.00002034, 0.00000091, -0.00002882, -0.00004324, -0.00002465, -0.00013398, -0.00010877, 0.00011822, 0.00014269, 0.00004880, 0.00004811, 0.00002921, 0.00001927, 0.00003633, 0.00002135, 0.00002206, 0.00003041, 0.00002406, 0.00002176, 0.00001980, 0.00002813, 0.00003266, 0.00002948, 0.00004983, 0.00007272, 0.00006213, 0.00001510, -0.00004029, -0.00005635, -0.00003672, },
		{ -0.00001317, -0.00000756, -0.00000825, -0.00000864, -0.00000211, 0.00000029, -0.00000509, -0.00001114, -0.00001694, -0.00000983, 0.00000746, 0.00000739, -0.00000425, -0.00001074, -0.00001068, 0.00000059, 0.00000514, -0.00000436, -0.00000565, -0.00000316, 0.00000061, 0.00000353, -0.00000966, -0.00000878, 0.00000582, -0.00000481, -0.00001123, -0.00000144, 0.00000062, 0.00000435, -0.00000042, -0.00000941, -0.00000348, -0.00000263, -0.00000361, 0.00000062, -0.00000125, -0.00000218, -0.00000348, -0.00000406, -0.00000134, 0.00000163, 0.00000422, -0.00000218, 0.00000028, 0.00000624, -0.00000997, -0.00001251, 0.00000110, 0.00000181, -0.00000349, -0.00000481, 0.00001150, 0.00000798, -0.00001940, -0.00000790, 0.00000953, 0.00000142, -0.00001703, -0.00000717, 0.00003643, -0.00001732, -0.00004730, 0.00005665, 0.00000160, -0.00006892, 0.00005330, 0.00002271, -0.00005826, 0.00001780, 0.00001299, -0.00001773, -0.00000216, -0.00000888, 0.00001881, 0.00000352, -0.00002006, 0.00001201, -0.00000281, -0.00000869, -0.00000357, -0.00001528, 0.00000941, -0.00000595, -0.00001800, 0.00000926, -0.00000888, -0.00001078, -0.00000170, -0.00000156, 0.00001264, -0.00001796, -0.00002047, 0.00000394, -0.00001742, -0.00001066, -0.00001041, -0.00001275, -0.00000402, -0.00005002, -0.00002892, -0.00004521, -0.00017196, -0.00004505, 0.00017539, 0.00011722, 0.00003487, 0.00004478, 0.00002496, 0.00003167, 0.00003448, 0.00001796, 0.00003237, 0.00003201, 0.00002012, 0.00002206, 0.00002664, 0.00003454, 0.00003496, 0.00004887, 0.00007894, 0.00006974, 0.00001668, -0.00003871, -0.00006115, -0.00004311, -0.00002169, -0.00001706, },
		{ -0.00000739, -0.00000848, -0.00000372, 0.00000097, -0.00000519, -0.00001166, -0.00001217, -0.00001220, -0.00000202, 0.00000921, 0.00000213, -0.00000819, -0.00001102, -0.00000611, 0.00000495, 0.00000171, -0.00000743, -0.00000356, -0.00000057, 0.00000123, -0.00000274, -0.00001294, -0.00000138, 0.00000591, -0.00000977, -0.00000784, 0.00000107, 0.00000070, 0.00000326, -0.00000369, -0.00000796, -0.00000335, -0.00000516, -0.00000088, 0.00000246, -0.00000336, -0.00000463, -0.00000279, -0.00000054, -0.00000129, 0.00000069, 0.00000434, -0.00000099, 0.00000127, -0.00000045, -0.00001211, -0.00000501, 0.00000070, -0.00000298, -0.00000101, 0.00000330, 0.00000848, -0.00000693, -0.00001424, 0.00000655, 0.00000146, -0.00001394, -0.00000459, 0.00001917, 0.00000164, -0.00004312, 0.00002205, 0.00003922, -0.00007121, 0.00000796, 0.00007515, -0.00006461, -0.00002217, 0.00005902, -0.00003093, -0.00002029, 0.00001669, 0.00000062, 0.00000908, -0.00001755, 0.00000011, 0.00001763, -0.00002284, -0.00000339, 0.00000212, -0.00001055, 0.00000284, -0.00001678, -0.00000038, 0.00000730, -0.00002314, -0.00000107, 0.00000488, 0.00000053, -0.00000075, -0.00002729, -0.00000325, -0.00000150, -0.00002515, -0.00000121, -0.00001155, -0.00001157, -0.00002067, -0.00005481, -0.00001261, -0.00009407, -0.00018227, 0.00005052, 0.00020039, 0.00007277, 0.00003053, 0.00004809, 0.00002777, 0.00003708, 0.00002848, 0.00002592, 0.00004005, 0.00002411, 0.00001999, 0.00003002, 0.00003257, 0.00003858, 0.00005233, 0.00008169, 0.00008154, 0.00001995, -0.00004287, -0.00006303, -0.00004792, -0.00002417, -0.00001684, -0.00001593, -0.00000972, },
		{ -0.00000505, 0.00000016, -0.00000356, -0.00001293, -0.00001179, -0.00000770, -0.00000751, 0.00000200, 0.00000708, -0.00000323, -0.00000955, -0.00000826, -0.00000119, 0.00000488, -0.00000306, -0.00000628, -0.00000031, -0.00000127, -0.00000189, -0.00000737, -0.00000980, 0.00000461, 0.00000130, -0.00001142, -0.00000294, 0.00000064, 0.00000012, 0.00000269, -0.00000566, -0.00000777, -0.00000487, -0.00000372, 0.00000290, -0.00000027, -0.00000670, -0.00000307, 0.00000060, -0.00000084, -0.00000345, 0.00000348, 0.00000537, -0.00000347, -0.00000077, -0.00000291, -0.00000912, -0.00000348, -0.00000307, 0.00000030, 0.00000398, 0.00000073, 0.00000004, -0.00000720, -0.00000286, 0.00000276, -0.00001120, -0.00000561, 0.00001331, 0.00000439, -0.00002920, -0.00000381, 0.00004671, -0.00003751, -0.00004553, 0.00008681, -0.00001142, -0.00008589, 0.00007050, 0.00001675, -0.00007293, 0.00003014, 0.00002141, -0.00001841, -0.00000168, -0.00000886, 0.00001892, -0.00000728, -0.00002758, 0.00001992, -0.00000499, -0.00001732, -0.00000042, -0.00000963, 0.00000885, -0.00001249, -0.00001901, 0.00001727, -0.00000283, -0.00000684, -0.00000909, -0.00002002, 0.00000460, -0.00001804, -0.00001622, 0.00000735, -0.00002328, -0.00001387, -0.00003358, -0.00004491, -0.00001825, -0.00015347, -0.00014045, 0.00014627, 0.00017639, 0.00003366, 0.00004669, 0.00005001, 0.00002668, 0.00003918, 0.00003070, 0.00003603, 0.00003466, 0.00001883, 0.00003114, 0.00003557, 0.00003547, 0.00005772, 0.00008672, 0.00009072, 0.00002979, -0.00004954, -0.00006962, -0.00004827, -0.00002698, -0.00001830, -0.00001772, -0.00001226, -0.00000718, -0.00000771, },
		{ -0.00000270, -0.00001202, -0.00001437, -0.00000529, -0.00000409, -0.00000608, 0.00000321, 0.00000343, -0.00000649, -0.00000776, -0.00000479, 0.00000079, 0.00000221, -0.00000455, -0.00000281, -0.00000092, -0.00000473, -0.00000375, -0.00000793, -0.00000476, 0.00000620, -0.00000366, -0.00000934, -0.00000054, -0.00000146, 0.00000157, 0.00000170, -0.00000875, -0.00000797, -0.00000384, -0.00000024, 0.00000222, -0.00000511, -0.00000546, 0.00000103, 0.00000052, -0.00000385, -0.00000149, 0.00000765, 0.00000106, -0.00000700, 0.00000121, -0.00000288, -0.00000965, -0.00000489, -0.00000006, 0.00000566, -0.00000074, -0.00000503, 0.00000215, -0.00000196, -0.00000463, -0.00000616, -0.00000603, 0.00000764, 0.00000558, -0.00001821, -0.00001477, 0.00003197, 0.00000133, -0.00006382, 0.00004301, 0.00005763, -0.00009926, 0.00000819, 0.00009190, -0.00008315, -0.00002437, 0.00007802, -0.00003132, -0.00002478, 0.00001609, 0.00000083, 0.00000720, -0.00002824, 0.00000146, 0.00002565, -0.00003205, -0.00001005, 0.00000796, -0.00000700, -0.00000029, -0.00002108, 0.00000546, 0.00001512, -0.00002240, -0.00000261, -0.00000867, -0.00001498, -0.00000134, -0.00002285, 0.00000322, -0.00000491, -0.00003546, -0.00000772, -0.00004415, -0.00003596, -0.00004696, -0.00019141, -0.00005715, 0.00020304, 0.00012363, 0.00002720, 0.00006752, 0.00003804, 0.00002924, 0.00004774, 0.00003420, 0.00003528, 0.00002767, 0.00002827, 0.00004068, 0.00003428, 0.00005575, 0.00009716, 0.00009936, 0.00003990, -0.00005029, -0.00008292, -0.00005009, -0.00002497, -0.00002162, -0.00002012, -0.00001430, -0.00000759, -0.00000694, -0.00000580, -0.00000115, },
		{ -0.00001545, -0.00000746, 0.00000087, -0.00000550, -0.00000636, 0.00000370, 0.00000024, -0.00000668, -0.00000489, -0.00000341, 0.00000075, 0.00000100, -0.00000337, -0.00000229, -0.00000513, -0.00000592, -0.00000280, -0.00000717, -0.00000077, 0.00000523, -0.00000674, -0.00000683, -0.00000095, -0.00000161, 0.00000368, -0.00000225, -0.00001129, -0.00000550, -0.00000149, 0.00000023, -0.00000154, -0.00000588, -0.00000088, 0.00000176, -0.00000242, -0.00000367, 0.00000323, 0.00000637, -0.00000584, -0.00000450, 0.00000458, -0.00000725, -0.00001115, -0.00000075, 0.00000515, 0.00000177, -0.00000854, -0.00000023, 0.00000761, -0.00000735, -0.00001020, -0.00000126, 0.00000316, 0.00000237, -0.00000898, -0.00001434, 0.00001108, 0.00001926, -0.00004127, -0.00001302, 0.00007649, -0.00004127, -0.00007097, 0.00010425, -0.00000881, -0.00010522, 0.00008581, 0.00002825, -0.00008380, 0.00002767, 0.00002088, -0.00001827, -0.00000428, -0.00001516, 0.00002616, -0.00000621, -0.00004083, 0.00001873, 0.00000093, -0.00001656, -0.00000271, -0.00000830, 0.00001724, -0.00001155, -0.00002315, 0.00001251, -0.00001856, -0.00001477, -0.00000273, -0.00001384, 0.00000692, -0.00003077, -0.00002595, -0.00000423, -0.00006240, -0.00002658, -0.00008462, -0.00019907, 0.00003543, 0.00021612, 0.00008033, 0.00004602, 0.00006826, 0.00002519, 0.00004735, 0.00005193, 0.00002733, 0.00003374, 0.00003415, 0.00003961, 0.00003914, 0.00005030, 0.00010320, 0.00011530, 0.00004701, -0.00004757, -0.00009405, -0.00006024, -0.00002006, -0.00002163, -0.00002523, -0.00001607, -0.00000810, -0.00000642, -0.00000629, -0.00000229, -0.00000270, -0.00001078, },
		{ 0.00000107, 0.00000058, -0.00001003, -0.00000511, 0.00000449, -0.00000164, -0.00000512, -0.00000364, -0.00000350, 0.00000181, 0.00000123, -0.00000395, -0.00000555, -0.00000768, -0.00000314, -0.00000215, -0.00000667, 0.00000269, 0.00000294, -0.00000893, -0.00000512, -0.00000130, 0.00000018, 0.00000296, -0.00000735, -0.00000990, -0.00000209, -0.00000191, -0.00000167, -0.00000279, -0.00000278, 0.00000113, -0.00000068, -0.00000296, -0.00000078, 0.00000444, 0.00000084, -0.00000753, 0.00000174, 0.00000191, -0.00001286, -0.00000623, 0.00000499, 0.00000235, -0.00000630, -0.00000620, 0.00000897, 0.00000186, -0.00001577, -0.00000317, 0.00000659, -0.00000319, -0.00000710, -0.00000559, -0.00000157, 0.00001171, -0.00000812, -0.00003247, 0.00003743, 0.00002495, -0.00008365, 0.00003430, 0.00007808, -0.00011140, 0.00000349, 0.00011261, -0.00008968, -0.00003529, 0.00008002, -0.00003427, -0.00002234, 0.00001532, -0.00000160, 0.00001364, -0.00003515, -0.00000962, 0.00002856, -0.00002866, -0.00000841, 0.00000962, -0.00000290, 0.00000319, -0.00002776, 0.00000109, 0.00000869, -0.00003605, -0.00000039, 0.00000028, -0.00001249, -0.00000744, -0.00004046, -0.00000258, -0.00002649, -0.00007253, -0.00000896, -0.00013573, -0.00018334, 0.00012549, 0.00020651, 0.00005664, 0.00006168, 0.00005482, 0.00003351, 0.00006549, 0.00003700, 0.00002399, 0.00004639, 0.00004201, 0.00003948, 0.00005172, 0.00010106, 0.00013292, 0.00006007, -0.00004925, -0.00009959, -0.00007311, -0.00002143, -0.00001563, -0.00003002, -0.00002017, -0.00000783, -0.00000639, -0.00000652, -0.00000339, -0.00000282, -0.00001027, -0.00001530, -0.00000973, },
		{ -0.00000525, -0.00001275, -0.00000123, 0.00000506, -0.00000270, -0.00000389, -0.00000396, -0.00000231, 0.00000390, -0.00000063, -0.00000748, -0.00000818, -0.00000573, 0.00000016, -0.00000364, -0.00000497, 0.00000534, -0.00000127, -0.00001009, -0.00000300, -0.00000096, 0.00000107, -0.00000022, -0.00000946, -0.00000615, -0.00000203, -0.00000412, -0.00000135, -0.00000109, -0.00000097, -0.00000002, -0.00000138, -0.00000134, -0.00000030, 0.00000233, -0.00000222, -0.00000427, 0.00000390, -0.00000468, -0.00001172, 0.00000127, 0.00000300, -0.00000462, -0.00000640, 0.00000233, 0.00000839, -0.00000933, -0.00001176, 0.00000774, -0.00000050, -0.00001304, -0.00000019, 0.00000173, -0.00000468, 0.00000341, -0.00001235, -0.00000619, 0.00003461, -0.00003055, -0.00003492, 0.00008375, -0.00003007, -0.00008474, 0.00011465, -0.00000171, -0.00012243, 0.00008512, 0.00002805, -0.00008704, 0.00003697, 0.00001902, -0.00002087, -0.00000112, -0.00002580, 0.00001932, -0.00000295, -0.00003697, 0.00002508, 0.00000313, -0.00001593, -0.00000661, -0.00001739, 0.00001568, -0.00001888, -0.00002985, 0.00002220, -0.00001475, -0.00001884, -0.00001567, -0.00003018, 0.00000219, -0.00006315, -0.00004724, -0.00000811, -0.00020973, -0.00012859, 0.00021543, 0.00017710, 0.00003968, 0.00007046, 0.00005164, 0.00005295, 0.00005927, 0.00002079, 0.00004340, 0.00005746, 0.00003625, 0.00005326, 0.00010288, 0.00014155, 0.00008093, -0.00004993, -0.00010873, -0.00007992, -0.00002942, -0.00001145, -0.00002864, -0.00002672, -0.00000835, -0.00000581, -0.00000730, -0.00000394, -0.00000325, -0.00000992, -0.00001556, -0.00001039, -0.00000079, 0.00000347, },
		{ -0.00001117, 0.00000368, 0.00000440, -0.00000368, -0.00000337, -0.00000375, 0.00000046, 0.00000369, -0.00000566, -0.00001031, -0.00000692, -0.00000173, 0.00000081, -0.00000549, -0.00000161, 0.00000519, -0.00000629, -0.00000869, -0.00000098, -0.00000148, 0.00000082, -0.00000263, -0.00000876, -0.00000460, -0.00000455, -0.00000354, 0.00000129, -0.00000065, -0.00000145, -0.00000016, -0.00000023, -0.00000212, -0.00000181, 0.00000205, -0.00000229, -0.00000222, 0.00000190, -0.00000741, -0.00000610, 0.00000197, -0.00000357, -0.00000489, -0.00000077, 0.00000445, 0.00000095, -0.00001131, -0.00000112, 0.00000529, -0.00001337, -0.00000648, 0.00001058, -0.00000767, -0.00000924, 0.00000974, -0.00000967, -0.00000118, 0.00001219, -0.00003279, 0.00002215, 0.00003941, -0.00008135, 0.00002850, 0.00008802, -0.00012219, -0.00000686, 0.00011786, -0.00009701, -0.00002879, 0.00009262, -0.00004443, -0.00002374, 0.00001699, -0.00001427, 0.00000946, -0.00003029, -0.00000024, 0.00003567, -0.00003269, -0.00001256, 0.00000166, -0.00000941, 0.00000485, -0.00003360, 0.00000272, 0.00001676, -0.00004162, -0.00000793, -0.00001642, -0.00002512, -0.00001516, -0.00007528, -0.00000195, -0.00006475, -0.00027304, -0.00001159, 0.00027893, 0.00011885, 0.00003747, 0.00008726, 0.00005475, 0.00005618, 0.00004175, 0.00003224, 0.00006653, 0.00004632, 0.00004593, 0.00011055, 0.00014901, 0.00009649, -0.00004058, -0.00012308, -0.00008653, -0.00003300, -0.00001570, -0.00002444, -0.00002949, -0.00001222, -0.00000440, -0.00000848, -0.00000456, -0.00000319, -0.00001012, -0.00001594, -0.00001106, -0.00000123, 0.00000261, -0.00000057, -0.00001103, },
		{ 0.00000750, 0.00000215, -0.00000437, -0.00000249, -0.00000240, 0.00000198, -0.00000041, -0.00001043, -0.00000923, -0.00000327, 0.00000084, -0.00000076, -0.00000581, 0.00000142, 0.00000199, -0.00000925, -0.00000546, -0.00000102, -0.00000223, 0.00000137, -0.00000409, -0.00000874, -0.00000560, -0.00000496, 0.00000009, 0.00000210, -0.00000212, -0.00000082, 0.00000107, -0.00000141, -0.00000490, -0.00000051, 0.00000336, -0.00000387, -0.00000163, 0.00000140, -0.00000672, -0.00000449, -0.00000254, -0.00000442, -0.00000004, -0.00000011, 0.00000145, -0.00000123, -0.00000677, 0.00000074, -0.00000529, -0.00001282, 0.00000784, 0.00000395, -0.00002012, 0.00000511, 0.00001390, -0.00002287, -0.00000031, 0.00000975, -0.00001449, 0.00002806, -0.00001421, -0.00003640, 0.00008126, -0.00003257, -0.00009902, 0.00011730, -0.00000308, -0.00012779, 0.00010580, 0.00002941, -0.00010387, 0.00004353, 0.00001624, -0.00003624, 0.00000002, -0.00001494, 0.00003147, -0.00000422, -0.00004747, 0.00002004, -0.00000576, -0.00001673, -0.00000133, -0.00002064, 0.00001984, -0.00002035, -0.00004018, 0.00001429, -0.00003502, -0.00002558, -0.00002619, -0.00005783, 0.00001126, -0.00016947, -0.00026432, 0.00014082, 0.00027531, 0.00005848, 0.00007001, 0.00009820, 0.00004206, 0.00005152, 0.00004564, 0.00005670, 0.00006245, 0.00004490, 0.00010958, 0.00016657, 0.00010546, -0.00003061, -0.00013057, -0.00010171, -0.00003100, -0.00001976, -0.00002788, -0.00002626, -0.00001564, -0.00000530, -0.00000841, -0.00000642, -0.00000223, -0.00001037, -0.00001692, -0.00001182, -0.00000117, 0.00000233, -0.00000051, -0.00000551, -0.00001346, -0.00000645, },
		{ -0.00000106, -0.00000375, -0.00000103, -0.00000181, 0.00000029, -0.00000564, -0.00001125, -0.00000526, -0.00000036, 0.00000124, -0.00000279, -0.00000488, 0.00000261, -0.00000198, -0.00000928, -0.00000318, -0.00000252, -0.00000094, 0.00000197, -0.00000703, -0.00000976, -0.00000532, -0.00000215, 0.00000252, 0.00000023, -0.00000217, 0.00000109, 0.00000001, -0.00000483, -0.00000477, 0.00000331, 0.00000158, -0.00000612, 0.00000177, 0.00000157, -0.00000894, -0.00000652, -0.00000312, -0.00000005, 0.00000010, -0.00000330, 0.00000319, 0.00000011, -0.00000784, -0.00000379, -0.00000771, -0.00000268, 0.00000954, -0.00001168, -0.00001305, 0.00002220, -0.00000574, -0.00003191, 0.00001972, 0.00000919, -0.00001487, 0.00001573, -0.00001715, 0.00001568, 0.00003160, -0.00009120, 0.00002446, 0.00009295, -0.00013121, 0.00000388, 0.00013876, -0.00011558, -0.00003717, 0.00010532, -0.00005909, -0.00003526, 0.00002872, -0.00000046, 0.00001477, -0.00004226, -0.00001033, 0.00003131, -0.00004013, -0.00000411, 0.00000841, -0.00001433, 0.00000448, -0.00004318, -0.00000538, 0.00000529, -0.00006252, -0.00000355, -0.00003099, -0.00004519, -0.00002501, -0.00026196, -0.00017013, 0.00026070, 0.00021498, 0.00004430, 0.00011426, 0.00007347, 0.00002973, 0.00006935, 0.00006023, 0.00005817, 0.00005918, 0.00010478, 0.00018240, 0.00012405, -0.00003114, -0.00013193, -0.00011641, -0.00003556, -0.00001477, -0.00003693, -0.00002704, -0.00001213, -0.00000850, -0.00000878, -0.00000797, -0.00000233, -0.00000901, -0.00001888, -0.00001289, -0.00000115, 0.00000307, -0.00000135, -0.00000406, -0.00000824, -0.00001316, -0.00000039, 0.00000907, },
		{ -0.00000149, -0.00000078, -0.00000367, -0.00000284, -0.00000814, -0.00000831, -0.00000165, 0.00000067, 0.00000041, -0.00000461, -0.00000352, 0.00000236, -0.00000511, -0.00000808, -0.00000262, -0.00000276, 0.00000170, -0.00000033, -0.00001098, -0.00000869, -0.00000264, 0.00000040, 0.00000213, -0.00000100, -0.00000040, 0.00000077, -0.00000339, -0.00000606, -0.00000124, 0.00000459, -0.00000273, -0.00000402, 0.00000612, -0.00000321, -0.00001300, -0.00000481, 0.00000060, 0.00000067, -0.00000411, -0.00000103, 0.00000758, -0.00000524, -0.00001212, -0.00000209, -0.00000356, 0.00000085, 0.00000131, -0.00001523, 0.00000476, 0.00001655, -0.00003083, -0.00001188, 0.00003651, -0.00001334, -0.00001380, 0.00002340, -0.00000583, 0.00001251, -0.00002582, -0.00004141, 0.00008539, -0.00003525, -0.00010203, 0.00014206, -0.00000232, -0.00015315, 0.00011647, 0.00003246, -0.00012693, 0.00004841, 0.00003348, -0.00002970, -0.00000508, -0.00003062, 0.00002592, -0.00000922, -0.00004745, 0.00003750, -0.00000620, -0.00002737, -0.00000598, -0.00003191, 0.00001380, -0.00003838, -0.00005031, 0.00002874, -0.00006299, -0.00004302, -0.00008084, -0.00030760, -0.00003539, 0.00031922, 0.00015551, 0.00007315, 0.00011862, 0.00003408, 0.00005706, 0.00009264, 0.00004896, 0.00006129, 0.00011702, 0.00018771, 0.00015078, -0.00002907, -0.00014393, -0.00012162, -0.00004576, -0.00001026, -0.00003826, -0.00003784, -0.00000702, -0.00000649, -0.00001270, -0.00000843, -0.00000349, -0.00000742, -0.00001957, -0.00001571, -0.00000082, 0.00000381, -0.00000092, -0.00000529, -0.00000481, -0.00001018, -0.00001070, 0.00000593, 0.00000773, -0.00000380, },
		{ -0.00000368, -0.00000646, -0.00000394, -0.00000718, -0.00000472, -0.00000005, 0.00000063, -0.00000112, -0.00000597, -0.00000173, 0.00000117, -0.00000747, -0.00000663, -0.00000218, -0.00000127, 0.00000233, -0.00000536, -0.00001222, -0.00000484, -0.00000045, 0.00000093, 0.00000143, -0.00000054, -0.00000012, -0.00000207, -0.00000511, -0.00000416, 0.00000137, 0.00000256, -0.00000421, 0.00000143, 0.00000473, -0.00001131, -0.00001175, 0.00000058, 0.00000110, -0.00000283, -0.00000396, 0.00000549, 0.00000427, -0.00001443, -0.00000812, 0.00000351, -0.00000526, -0.00000097, -0.00000142, -0.00000903, 0.00001073, -0.00000358, -0.00003188, 0.00001893, 0.00002207, -0.00003337, 0.00001318, 0.00002651, -0.00002046, -0.00000546, -0.00002356, 0.00001849, 0.00003409, -0.00009694, 0.00003653, 0.00011201, -0.00015203, -0.00000554, 0.00015650, -0.00013247, -0.00004873, 0.00012590, -0.00005248, -0.00003798, 0.00001915, -0.00001445, 0.00001257, -0.00004472, 0.00000239, 0.00004676, -0.00005542, -0.00001553, 0.00000312, -0.00002573, -0.00000286, -0.00006063, 0.00000447, 0.00001201, -0.00011103, -0.00001914, -0.00015068, -0.00031630, 0.00011285, 0.00034132, 0.00011793, 0.00009254, 0.00008727, 0.00003812, 0.00010449, 0.00007400, 0.00004134, 0.00013189, 0.00020435, 0.00016749, -0.00001288, -0.00016423, -0.00013196, -0.00004654, -0.00001481, -0.00003350, -0.00004673, -0.00001208, 0.00000219, -0.00001528, -0.00001175, -0.00000310, -0.00000769, -0.00001851, -0.00001868, -0.00000211, 0.00000539, -0.00000048, -0.00000550, -0.00000551, -0.00000529, -0.00001182, -0.00000534, 0.00001061, 0.00000376, -0.00000437, 0.00000074, },
		{ -0.00000646, -0.00000211, -0.00000540, -0.00000271, 0.00000022, 0.00000011, -0.00000322, -0.00000621, 0.00000014, -0.00000123, -0.00000883, -0.00000444, -0.00000139, -0.00000035, -0.00000007, -0.00000928, -0.00000945, -0.00000137, -0.00000024, 0.00000129, 0.00000152, -0.00000095, -0.00000209, -0.00000375, -0.00000414, -0.00000261, 0.00000192, 0.00000130, -0.00000225, 0.00000370, -0.00000231, -0.00001477, -0.00000512, 0.00000240, -0.00000224, -0.00000322, 0.00000074, 0.00000643, -0.00000553, -0.00001535, 0.00000184, 0.00000092, -0.00001015, 0.00000357, -0.00000084, -0.00000834, 0.00000651, -0.00001401, -0.00001455, 0.00002864, -0.00000622, -0.00002067, 0.00003826, -0.00000286, -0.00003273, 0.00000678, -0.00000344, 0.00001701, -0.00002728, -0.00004043, 0.00010388, -0.00003599, -0.00012404, 0.00015283, 0.00000043, -0.00017533, 0.00012974, 0.00004635, -0.00013651, 0.00004736, 0.00002168, -0.00004159, -0.00000180, -0.00002493, 0.00004483, -0.00001031, -0.00007164, 0.00003328, -0.00001515, -0.00003880, -0.00001085, -0.00003853, 0.00003437, -0.00006258, -0.00010773, 0.00001094, -0.00026455, -0.00026915, 0.00028244, 0.00032682, 0.00007369, 0.00009386, 0.00007954, 0.00008068, 0.00010620, 0.00004339, 0.00011820, 0.00023898, 0.00018511, -0.00000030, -0.00017277, -0.00015864, -0.00004269, -0.00001465, -0.00003734, -0.00004550, -0.00002244, 0.00000511, -0.00000933, -0.00001771, -0.00000325, -0.00000735, -0.00001902, -0.00001910, -0.00000527, 0.00000643, 0.00000075, -0.00000551, -0.00000600, -0.00000493, -0.00000688, -0.00001096, 0.00000171, 0.00001146, -0.00000074, -0.00000273, -0.00000000, -0.00000864, },
		{ -0.00000021, -0.00000515, -0.00000188, 0.00000018, -0.00000101, -0.00000522, -0.00000502, 0.00000098, -0.00000431, -0.00000814, -0.00000203, -0.00000173, -0.00000098, -0.00000285, -0.00000976, -0.00000551, -0.00000029, -0.00000017, 0.00000229, 0.00000045, -0.00000295, -0.00000325, -0.00000304, -0.00000360, -0.00000222, 0.00000335, 0.00000146, -0.00000157, 0.00000128, -0.00000792, -0.00001168, -0.00000049, -0.00000043, -0.00000276, -0.00000003, 0.00000194, 0.00000110, -0.00001071, -0.00000774, 0.00000501, -0.00000797, -0.00000591, 0.00001027, -0.00000903, -0.00000985, 0.00000740, -0.00001479, -0.00000231, 0.00002063, -0.00001609, 0.00000676, 0.00002731, -0.00003491, -0.00000936, 0.00002062, -0.00001385, -0.00000304, -0.00002420, 0.00002612, 0.00004485, -0.00011029, 0.00003011, 0.00012461, -0.00016642, -0.00000824, 0.00017780, -0.00014430, -0.00005648, 0.00013069, -0.00007237, -0.00004115, 0.00003216, -0.00000505, 0.00002345, -0.00006317, -0.00001606, 0.00004629, -0.00007215, -0.00001795, 0.00000550, -0.00002454, 0.00000612, -0.00011989, -0.00003803, -0.00003230, -0.00039359, -0.00011061, 0.00043297, 0.00023921, 0.00003670, 0.00012460, 0.00009980, 0.00009211, 0.00007780, 0.00010266, 0.00025457, 0.00022860, 0.00000192, -0.00017642, -0.00018285, -0.00005479, -0.00000136, -0.00004481, -0.00004803, -0.00002288, -0.00000258, -0.00000249, -0.00001730, -0.00000783, -0.00000513, -0.00002089, -0.00001960, -0.00000636, 0.00000456, 0.00000286, -0.00000535, -0.00000577, -0.00000585, -0.00000531, -0.00000723, -0.00000739, 0.00000727, 0.00000886, -0.00000362, -0.00000142, -0.00000441, -0.00001126, -0.00000220, },
		{ -0.00000580, -0.00000095, -0.00000023, -0.00000272, -0.00000626, -0.00000313, 0.00000025, -0.00000643, -0.00000575, -0.00000141, -0.00000337, -0.00000130, -0.00000400, -0.00000838, -0.00000299, -0.00000019, 0.00000081, 0.00000195, -0.00000223, -0.00000394, -0.00000234, -0.00000271, -0.00000433, -0.00000007, 0.00000557, -0.00000066, -0.00000317, -0.00000092, -0.00000916, -0.00000792, -0.00000037, -0.00000163, 0.00000023, -0.00000018, -0.00000110, -0.00000189, -0.00000911, -0.00000244, 0.00000008, -0.00001032, 0.00000492, 0.00000524, -0.00002039, -0.00000013, 0.00000955, -0.00001987, 0.00000482, 0.00001614, -0.00001146, 0.00001500, -0.00000171, -0.00003076, 0.00002114, -0.00000020, -0.00002636, 0.00000861, -0.00000212, 0.00002385, -0.00002623, -0.00005101, 0.00010852, -0.00003589, -0.00013464, 0.00016988, -0.00000128, -0.00019939, 0.00014215, 0.00004130, -0.00015975, 0.00006540, 0.00003519, -0.00004038, -0.00000227, -0.00004848, 0.00003615, -0.00001780, -0.00008578, 0.00004919, -0.00001571, -0.00004206, -0.00002612, -0.00011490, 0.00001680, -0.00017240, -0.00043014, 0.00013959, 0.00046954, 0.00011057, 0.00007465, 0.00017385, 0.00008226, 0.00007868, 0.00012967, 0.00024642, 0.00027437, 0.00002410, -0.00019714, -0.00019499, -0.00007545, 0.00000470, -0.00003768, -0.00006179, -0.00001901, -0.00000525, -0.00000760, -0.00001013, -0.00001057, -0.00000614, -0.00002021, -0.00002314, -0.00000527, 0.00000297, 0.00000224, -0.00000337, -0.00000605, -0.00000548, -0.00000695, -0.00000388, -0.00000609, -0.00000372, 0.00001032, 0.00000490, -0.00000546, -0.00000255, -0.00000889, -0.00000835, 0.00000309, -0.00000133, },
		{ -0.00000005, -0.00000147, -0.00000433, -0.00000606, -0.00000153, -0.00000127, -0.00000683, -0.00000386, -0.00000312, -0.00000398, -0.00000006, -0.00000456, -0.00000712, -0.00000144, 0.00000024, 0.00000105, -0.00000036, -0.00000395, -0.00000280, -0.00000175, -0.00000380, -0.00000345, 0.00000386, 0.00000446, -0.00000517, -0.00000280, -0.00000103, -0.00000984, -0.00000627, -0.00000004, 0.00000037, 0.00000024, -0.00000421, -0.00000105, -0.00000099, -0.00000848, -0.00000211, -0.00000301, -0.00000492, 0.00000824, -0.00000896, -0.00001837, 0.00001503, -0.00000130, -0.00002354, 0.00001940, 0.00001206, -0.00001377, 0.00000868, -0.00001222, -0.00000611, 0.00001793, -0.00002814, -0.00000368, 0.00002181, -0.00001414, 0.00000089, -0.00002517, 0.00002231, 0.00004791, -0.00011531, 0.00003652, 0.00013330, -0.00019556, -0.00001180, 0.00019834, -0.00017153, -0.00005833, 0.00016375, -0.00007963, -0.00004533, 0.00002897, -0.00002863, 0.00001700, -0.00007489, -0.00000968, 0.00006789, -0.00009331, -0.00002598, -0.00003293, -0.00009775, 0.00000056, -0.00032259, -0.00032359, 0.00035784, 0.00038007, 0.00005646, 0.00017524, 0.00015311, 0.00004158, 0.00015437, 0.00027217, 0.00028793, 0.00006885, -0.00021599, -0.00022088, -0.00008138, -0.00000585, -0.00002579, -0.00006764, -0.00002805, 0.00000377, -0.00001603, -0.00001139, -0.00000372, -0.00000975, -0.00001979, -0.00002582, -0.00000710, 0.00000525, -0.00000065, -0.00000260, -0.00000449, -0.00000577, -0.00000735, -0.00000532, -0.00000070, -0.00000655, -0.00000028, 0.00001188, -0.00000045, -0.00000764, -0.00000424, -0.00000929, -0.00000200, 0.00000478, -0.00000485, -0.00000528, },
		{ -0.00000305, -0.00000511, -0.00000506, -0.00000056, -0.00000278, -0.00000658, -0.00000390, -0.00000484, -0.00000188, 0.00000117, -0.00000590, -0.00000574, -0.00000008, 0.00000000, -0.00000034, -0.00000233, -0.00000335, -0.00000174, -0.00000258, -0.00000389, -0.00000031, 0.00000543, -0.00000065, -0.00000706, 0.00000066, -0.00000276, -0.00001145, -0.00000350, 0.00000202, 0.00000022, -0.00000391, -0.00000485, 0.00000307, -0.00000310, -0.00001014, 0.00000004, -0.00000235, -0.00000203, 0.00000236, -0.00001646, -0.00000389, 0.00001734, -0.00001887, -0.00001030, 0.00003193, -0.00000632, -0.00001726, 0.00001149, -0.00000766, 0.00000193, -0.00000444, -0.00002204, 0.00002519, -0.00000271, -0.00002812, 0.00001396, -0.00000191, 0.00002137, -0.00002570, -0.00004842, 0.00011763, -0.00005532, -0.00016045, 0.00019524, -0.00000266, -0.00022843, 0.00017634, 0.00005436, -0.00018859, 0.00007566, 0.00002570, -0.00006842, -0.00000380, -0.00004910, 0.00006005, -0.00001996, -0.00012376, 0.00004633, -0.00007945, -0.00009580, -0.00004386, -0.00041681, -0.00014733, 0.00047206, 0.00028138, 0.00011066, 0.00021393, 0.00007015, 0.00011761, 0.00033259, 0.00031172, 0.00008873, -0.00020508, -0.00026431, -0.00008740, -0.00000683, -0.00003257, -0.00005916, -0.00003999, 0.00000572, -0.00001082, -0.00002409, 0.00000212, -0.00000510, -0.00002437, -0.00002603, -0.00001093, 0.00000695, 0.00000005, -0.00000610, -0.00000181, -0.00000525, -0.00000787, -0.00000664, -0.00000016, -0.00000071, -0.00000832, 0.00000504, 0.00001038, -0.00000790, -0.00000799, -0.00000318, -0.00000628, 0.00000271, 0.00000209, -0.00000757, -0.00000336, -0.00000028, },
		{ -0.00000512, -0.00000380, -0.00000027, -0.00000449, -0.00000681, -0.00000466, -0.00000413, 0.00000153, 0.00000036, -0.00000731, -0.00000368, 0.00000018, -0.00000152, -0.00000162, -0.00000238, -0.00000232, -0.00000203, -0.00000282, -0.00000215, 0.00000175, 0.00000294, -0.00000501, -0.00000416, 0.00000265, -0.00000732, -0.00001038, 0.00000095, 0.00000147, -0.00000344, -0.00000581, -0.00000024, 0.00000392, -0.00000900, -0.00000732, 0.00000401, -0.00000525, -0.00000324, -0.00000191, -0.00001368, 0.00000707, 0.00000616, -0.00002360, 0.00001231, 0.00002264, -0.00002769, -0.00000293, 0.00001784, -0.00001257, -0.00000346, -0.00000909, 0.00000218, 0.00001726, -0.00003401, -0.00000180, 0.00002970, -0.00001575, -0.00000098, -0.00002075, 0.00002808, 0.00003868, -0.00014788, 0.00004119, 0.00015971, -0.00022293, -0.00000731, 0.00023971, -0.00020377, -0.00007296, 0.00018921, -0.00011480, -0.00006525, 0.00004559, -0.00001976, 0.00003222, -0.00011063, -0.00003770, 0.00006247, -0.00018370, -0.00003773, -0.00010363, -0.00049122, 0.00004768, 0.00054526, 0.00024168, 0.00015574, 0.00015761, 0.00009809, 0.00034265, 0.00039456, 0.00008868, -0.00020086, -0.00028646, -0.00011916, 0.00000813, -0.00004290, -0.00006339, -0.00003424, -0.00000457, -0.00000252, -0.00002847, -0.00000680, 0.00000662, -0.00002572, -0.00003053, -0.00001140, 0.00000447, 0.00000352, -0.00000856, -0.00000348, -0.00000167, -0.00000917, -0.00000704, -0.00000158, 0.00000254, -0.00000500, -0.00000681, 0.00001034, 0.00000295, -0.00001362, -0.00000309, -0.00000068, -0.00000400, 0.00000390, -0.00000191, -0.00000797, -0.00000175, -0.00000158, -0.00000393, },
		{ -0.00000257, -0.00000105, -0.00000668, -0.00000689, -0.00000411, -0.00000127, 0.00000356, -0.00000228, -0.00000752, -0.00000202, -0.00000128, -0.00000277, -0.00000135, -0.00000182, -0.00000227, -0.00000202, -0.00000165, -0.00000122, 0.00000104, -0.00000017, -0.00000541, -0.00000026, 0.00000052, -0.00001053, -0.00000575, 0.00000244, -0.00000260, -0.00000523, -0.00000297, 0.00000314, -0.00000078, -0.00001116, -0.00000005, 0.00000179, -0.00001103, -0.00000037, -0.00000232, -0.00001033, 0.00000989, -0.00000348, -0.00001444, 0.00002174, -0.00000200, -0.00002726, 0.00001975, 0.00000649, -0.00002154, 0.00000250, -0.00000076, 0.00000590, -0.00001008, -0.00002593, 0.00003293, 0.00000366, -0.00003129, 0.00001667, 0.00000351, 0.00001578, -0.00005144, -0.00006510, 0.00014439, -0.00005354, -0.00018190, 0.00023328, -0.00000283, -0.00027815, 0.00020578, 0.00005656, -0.00024383, 0.00009565, 0.00004876, -0.00007265, -0.00000982, -0.00009750, 0.00005457, -0.00005420, -0.00021255, 0.00008458, -0.00027847, -0.00053137, 0.00032619, 0.00060462, 0.00017515, 0.00013757, 0.00017859, 0.00031565, 0.00046930, 0.00014281, -0.00024028, -0.00029346, -0.00014759, 0.00000460, -0.00003014, -0.00008603, -0.00002791, -0.00000193, -0.00001012, -0.00002048, -0.00001736, 0.00000716, -0.00001549, -0.00003795, -0.00001267, 0.00000372, 0.00000273, -0.00000581, -0.00000776, -0.00000025, -0.00000714, -0.00000934, -0.00000109, 0.00000151, 0.00000014, -0.00000818, -0.00000164, 0.00001004, -0.00000730, -0.00001137, 0.00000467, -0.00000151, -0.00000350, 0.00000364, -0.00000524, -0.00000725, -0.00000119, -0.00000265, -0.00000398, -0.00000489, },
		{ -0.00000327, -0.00000838, -0.00000540, -0.00000201, 0.00000161, 0.00000316, -0.00000516, -0.00000679, -0.00000186, -0.00000299, -0.00000232, -0.00000057, -0.00000197, -0.00000206, -0.00000087, -0.00000142, -0.00000243, 0.00000016, -0.00000119, -0.00000395, 0.00000130, -0.00000291, -0.00001006, -0.00000215, -0.00000048, -0.00000495, -0.00000295, -0.00000042, 0.00000215, -0.00000484, -0.00000744, 0.00000298, -0.00000667, -0.00001008, 0.00000626, -0.00000674, -0.00000802, 0.00001264, -0.00000802, -0.00000693, 0.00001528, -0.00001682, -0.00000747, 0.00002432, -0.00001500, -0.00001244, 0.00001335, -0.00000614, -0.00000512, -0.00001524, 0.00000405, 0.00002754, -0.00003075, -0.00000242, 0.00003392, -0.00001987, -0.00001939, -0.00004286, 0.00003527, 0.00006129, -0.00016197, 0.00004981, 0.00018754, -0.00027071, -0.00001793, 0.00028517, -0.00025475, -0.00009500, 0.00024759, -0.00012886, -0.00008235, 0.00003238, -0.00006550, 0.00003410, -0.00017786, -0.00007633, 0.00009367, -0.00058360, -0.00037072, 0.00067823, 0.00052403, 0.00005755, 0.00022076, 0.00038524, 0.00047558, 0.00023770, -0.00026150, -0.00034678, -0.00014391, -0.00001455, -0.00001866, -0.00009253, -0.00004693, 0.00001619, -0.00001681, -0.00002426, -0.00001156, -0.00000374, -0.00000783, -0.00003516, -0.00002052, 0.00000629, 0.00000003, -0.00000478, -0.00000673, -0.00000432, -0.00000390, -0.00001000, -0.00000217, 0.00000231, -0.00000045, -0.00000235, -0.00000806, 0.00000169, 0.00000411, -0.00001201, -0.00000169, 0.00000776, -0.00000621, -0.00000172, 0.00000283, -0.00000841, -0.00000602, -0.00000079, -0.00000305, -0.00000395, -0.00000459, -0.00000187, },
		{ -0.00000780, -0.00000236, 0.00000006, 0.00000309, 0.00000106, -0.00000749, -0.00000612, -0.00000253, -0.00000331, -0.00000101, -0.00000055, -0.00000198, -0.00000085, -0.00000058, -0.00000343, -0.00000321, 0.00000115, -0.00000161, -0.00000293, 0.00000159, -0.00000520, -0.00000879, -0.00000210, -0.00000306, -0.00000309, -0.00000113, -0.00000075, 0.00000098, -0.00000540, -0.00000444, -0.00000087, -0.00001141, -0.00000115, 0.00000649, -0.00001423, 0.00000097, 0.00001421, -0.00001625, -0.00000410, 0.00000909, -0.00001585, 0.00000741, 0.00001032, -0.00002050, 0.00000730, 0.00000519, -0.00001830, -0.00000158, -0.00000341, 0.00001251, 0.00000000, -0.00002344, 0.00003326, -0.00000003, -0.00004594, -0.00000117, -0.00000375, 0.00003309, -0.00004196, -0.00007173, 0.00016579, -0.00006932, -0.00022751, 0.00027384, -0.00000195, -0.00034338, 0.00025821, 0.00008599, -0.00030658, 0.00010521, 0.00002443, -0.00012045, -0.00000557, -0.00017072, 0.00007239, -0.00014455, -0.00078823, 0.00006750, 0.00087516, 0.00027031, 0.00010877, 0.00051437, 0.00053273, 0.00026780, -0.00021523, -0.00042965, -0.00016240, 0.00000045, -0.00003543, -0.00008241, -0.00006671, 0.00001524, -0.00000020, -0.00004104, -0.00000633, -0.00000188, -0.00001688, -0.00002556, -0.00002418, 0.00000322, 0.00000248, -0.00000926, -0.00000300, -0.00000582, -0.00000679, -0.00000665, -0.00000424, 0.00000339, -0.00000106, -0.00000144, -0.00000289, -0.00000907, 0.00000181, -0.00000037, -0.00000841, 0.00000699, 0.00000316, -0.00000940, 0.00000218, -0.00000042, -0.00001098, -0.00000340, -0.00000043, -0.00000351, -0.00000410, -0.00000425, -0.00000246, -0.00000605, },
		{ 0.00000030, 0.00000095, 0.00000327, -0.00000201, -0.00000930, -0.00000543, -0.00000266, -0.00000248, -0.00000017, -0.00000070, -0.00000108, -0.00000032, -0.00000255, -0.00000532, -0.00000141, 0.00000216, -0.00000299, -0.00000134, 0.00000162, -0.00000789, -0.00000872, -0.00000237, -0.00000227, -0.00000110, -0.00000223, -0.00000024, 0.00000142, -0.00000681, -0.00000504, -0.00000449, -0.00000832, 0.00000576, -0.00000151, -0.00001390, 0.00001383, 0.00000413, -0.00002555, 0.00000473, 0.00000710, -0.00001354, 0.00001025, -0.00000014, -0.00000950, 0.00001093, -0.00001459, -0.00001123, 0.00001033, -0.00000495, 0.00000469, -0.00000538, 0.00000441, 0.00002120, -0.00004382, -0.00001615, 0.00002535, -0.00001743, -0.00000210, -0.00003781, 0.00003749, 0.00006333, -0.00020243, 0.00004963, 0.00022812, -0.00032334, -0.00002116, 0.00036377, -0.00031684, -0.00013771, 0.00029341, -0.00020447, -0.00010763, 0.00005204, -0.00014156, 0.00005490, -0.00046876, -0.00067983, 0.00054261, 0.00076526, 0.00014021, 0.00048853, 0.00072953, 0.00027635, -0.00020035, -0.00045285, -0.00024153, 0.00003192, -0.00003623, -0.00009994, -0.00005916, -0.00000446, 0.00001367, -0.00003899, -0.00002186, 0.00001273, -0.00002382, -0.00003078, -0.00001502, -0.00000285, 0.00000383, -0.00001067, -0.00000568, -0.00000056, -0.00001191, -0.00000721, -0.00000145, 0.00000153, 0.00000081, -0.00000382, 0.00000121, -0.00000711, -0.00001147, 0.00000450, -0.00000034, -0.00000273, 0.00000896, -0.00000396, -0.00000768, 0.00000443, -0.00000613, -0.00001041, 0.00000006, -0.00000131, -0.00000411, -0.00000418, -0.00000455, -0.00000420, -0.00000718, -0.00000458, },
		{ 0.00000116, 0.00000245, -0.00000580, -0.00001014, -0.00000399, -0.00000219, -0.00000160, 0.00000034, -0.00000024, -0.00000060, -0.00000182, -0.00000492, -0.00000463, 0.00000137, 0.00000117, -0.00000372, 0.00000133, -0.00000063, -0.00001158, -0.00000743, -0.00000059, -0.00000100, -0.00000182, -0.00000266, 0.00000254, -0.00000083, -0.00001036, -0.00000430, -0.00000415, -0.00000377, 0.00000581, -0.00000794, -0.00000424, 0.00001659, -0.00001506, -0.00002169, 0.00001791, -0.00000159, -0.00001317, 0.00001371, -0.00000167, -0.00000387, -0.00000157, -0.00001784, 0.00000842, 0.00000394, -0.00001338, 0.00001118, 0.00000322, 0.00000459, -0.00001453, -0.00003558, 0.00002839, -0.00000336, -0.00003895, 0.00001682, -0.00000181, 0.00002761, -0.00006157, -0.00009566, 0.00019635, -0.00007314, -0.00026995, 0.00034177, -0.00000044, -0.00045229, 0.00030897, 0.00008887, -0.00042791, 0.00016941, 0.00001790, -0.00023022, -0.00001870, -0.00066311, -0.00039248, 0.00079919, 0.00061979, 0.00042320, 0.00086659, 0.00043758, -0.00027692, -0.00045705, -0.00029579, 0.00000447, 0.00000849, -0.00013194, -0.00006223, 0.00000088, -0.00000200, -0.00002158, -0.00003303, 0.00001013, -0.00001016, -0.00004715, -0.00001177, 0.00000386, -0.00000258, -0.00000799, -0.00001015, 0.00000140, -0.00000924, -0.00001466, 0.00000183, 0.00000282, -0.00000039, -0.00000277, -0.00000098, -0.00000002, -0.00001576, -0.00000740, 0.00001108, -0.00000127, 0.00000024, 0.00000687, -0.00000851, -0.00000383, 0.00000261, -0.00001048, -0.00000626, 0.00000168, -0.00000335, -0.00000410, -0.00000478, -0.00000577, -0.00000519, -0.00000537, -0.00000085, 0.00000093, },
		{ 0.00000024, -0.00000955, -0.00000906, -0.00000204, -0.00000189, -0.00000076, 0.00000108, -0.00000013, -0.00000185, -0.00000396, -0.00000532, -0.00000216, 0.00000266, -0.00000069, -0.00000238, 0.00000245, -0.00000590, -0.00001307, -0.00000338, 0.00000078, -0.00000175, -0.00000257, -0.00000037, 0.00000302, -0.00000691, -0.00001057, -0.00000039, -0.00000434, -0.00000143, 0.00000441, -0.00000840, 0.00000343, 0.00000593, -0.00002643, -0.00000387, 0.00001988, -0.00001582, -0.00000361, 0.00001867, -0.00000780, -0.00000749, -0.00000701, -0.00000418, 0.00001199, -0.00001361, -0.00000178, 0.00002239, -0.00000834, -0.00001240, -0.00001890, 0.00000170, 0.00002108, -0.00004615, -0.00000269, 0.00004017, -0.00002801, -0.00001812, -0.00005565, 0.00003986, 0.00008378, -0.00022970, 0.00006786, 0.00028014, -0.00042463, -0.00005372, 0.00046043, -0.00043705, -0.00018474, 0.00043894, -0.00035308, -0.00020688, 0.00000362, -0.00085418, -0.00013282, 0.00102737, 0.00081186, 0.00083622, 0.00067257, -0.00027406, -0.00058036, -0.00028177, -0.00004424, 0.00002406, -0.00011615, -0.00010342, 0.00002221, -0.00000984, -0.00003155, -0.00001720, -0.00000506, -0.00000048, -0.00004562, -0.00002846, 0.00001676, -0.00000250, -0.00001297, -0.00000776, -0.00000326, -0.00000410, -0.00001757, -0.00000355, 0.00000914, -0.00000201, -0.00000277, -0.00000108, 0.00000011, -0.00000874, -0.00001857, 0.00000564, 0.00001304, -0.00000618, 0.00000338, 0.00000398, -0.00001114, -0.00000096, -0.00000087, -0.00001097, -0.00000178, 0.00000040, -0.00000475, -0.00000400, -0.00000660, -0.00000629, -0.00000373, -0.00000217, 0.00000086, 0.00000011, 0.00000150, },
		{ -0.00001164, -0.00000609, -0.00000070, -0.00000185, 0.00000048, 0.00000136, -0.00000153, -0.00000376, -0.00000459, -0.00000403, -0.00000003, 0.00000255, -0.00000166, -0.00000073, 0.00000010, -0.00001077, -0.00001019, 0.00000042, -0.00000033, -0.00000236, -0.00000133, 0.00000106, -0.00000103, -0.00001097, -0.00000529, 0.00000118, -0.00000661, 0.00000267, 0.00000372, -0.00000912, 0.00000367, -0.00000597, -0.00002256, 0.00001145, 0.00000728, -0.00001970, 0.00001334, 0.00001172, -0.00001832, -0.00000453, -0.00000141, 0.00000252, -0.00000042, -0.00001318, 0.00001977, 0.00000754, -0.00002844, -0.00000598, -0.00000046, 0.00000861, -0.00001602, -0.00003400, 0.00004552, -0.00000318, -0.00006094, 0.00000530, -0.00000728, 0.00003650, -0.00005793, -0.00010025, 0.00023939, -0.00011286, -0.00037267, 0.00042783, -0.00000094, -0.00062199, 0.00045069, 0.00013038, -0.00073649, 0.00024446, -0.00020169, -0.00122158, 0.00037502, 0.00158358, 0.00111520, 0.00070969, -0.00010709, -0.00074644, -0.00034566, -0.00000311, -0.00002283, -0.00008762, -0.00012027, 0.00000394, 0.00001852, -0.00005923, -0.00001217, 0.00000610, -0.00001477, -0.00003196, -0.00003950, 0.00001052, 0.00001271, -0.00002087, -0.00000790, -0.00000259, -0.00000748, -0.00001257, -0.00001062, 0.00000934, 0.00000342, -0.00000678, 0.00000110, -0.00000124, -0.00000577, -0.00001530, -0.00000920, 0.00001704, 0.00000502, -0.00000866, 0.00000869, -0.00000171, -0.00001269, 0.00000199, -0.00000368, -0.00000936, 0.00000052, -0.00000190, -0.00000508, -0.00000525, -0.00000811, -0.00000412, -0.00000094, -0.00000035, 0.00000037, -0.00000053, 0.00000154, -0.00000346, },
		{ -0.00000271, -0.00000055, -0.00000139, 0.00000156, 0.00000010, -0.00000365, -0.00000433, -0.00000376, -0.00000266, 0.00000143, 0.00000204, -0.00000232, -0.00000073, -0.00000368, -0.00001200, -0.00000549, 0.00000122, -0.00000166, -0.00000116, -0.00000080, -0.00000075, -0.00000518, -0.00000935, -0.00000019, -0.00000204, -0.00000538, 0.00000861, -0.00000191, -0.00001213, 0.00000329, -0.00001127, -0.00001317, 0.00001483, -0.00000539, -0.00000919, 0.00002040, -0.00000602, -0.00001715, 0.00000572, -0.00000166, -0.00000148, -0.00000434, 0.00000274, 0.00001814, -0.00002193, -0.00002006, 0.00001470, -0.00000330, -0.00000835, -0.00002136, 0.00000792, 0.00003056, -0.00006235, -0.00002021, 0.00003963, -0.00003082, -0.00000979, -0.00004654, 0.00005508, 0.00008058, -0.00032441, 0.00006804, 0.00037339, -0.00056852, -0.00007760, 0.00068319, -0.00069989, -0.00039566, 0.00071681, -0.00112824, -0.00125513, 0.00161699, 0.00215185, 0.00079734, -0.00007979, -0.00074518, -0.00054924, 0.00005299, -0.00000881, -0.00013860, -0.00008716, -0.00002550, 0.00002832, -0.00005047, -0.00004426, 0.00003116, -0.00001724, -0.00004227, -0.00002715, -0.00000585, 0.00001879, -0.00001275, -0.00001796, 0.00000293, -0.00001132, -0.00001346, -0.00000758, 0.00000165, 0.00000844, -0.00000560, -0.00000194, 0.00000238, -0.00000871, -0.00000980, -0.00001176, 0.00000360, 0.00001713, -0.00000559, -0.00000338, 0.00001067, -0.00001071, -0.00000987, 0.00000521, -0.00000709, -0.00000737, 0.00000135, -0.00000398, -0.00000585, -0.00000703, -0.00000666, -0.00000027, 0.00000030, -0.00000053, -0.00000060, -0.00000060, 0.00000034, -0.00000729, -0.00001097, },
		{ -0.00000101, -0.00000051, 0.00000134, -0.00000223, -0.00000458, -0.00000340, -0.00000293, -0.00000149, 0.00000265, 0.00000066, -0.00000358, -0.00000145, -0.00000616, -0.00001052, -0.00000226, 0.00000052, -0.00000120, -0.00000069, -0.00000282, -0.00000260, -0.00000582, -0.00000575, 0.00000060, -0.00000462, 0.00000105, 0.00000857, -0.00001285, -0.00000998, 0.00000560, -0.00001488, -0.00000585, 0.00001299, -0.00000944, 0.00000182, 0.00001255, -0.00001607, -0.00000348, 0.00000652, -0.00000878, 0.00000127, 0.00000201, 0.00000599, -0.00000369, -0.00002705, 0.00000852, 0.00000988, -0.00002076, -0.00000418, -0.00000308, 0.00001270, -0.00001960, -0.00005292, 0.00004147, -0.00000268, -0.00006295, 0.00002443, 0.00000591, 0.00002860, -0.00011276, -0.00015335, 0.00031560, -0.00013592, -0.00051357, 0.00061064, 0.00000174, -0.00107186, 0.00066820, 0.00016156, -0.00232624, 0.00004599, 0.00319695, 0.00174244, -0.00030533, -0.00075869, -0.00065604, -0.00004908, 0.00010669, -0.00020093, -0.00010964, 0.00001073, -0.00000474, -0.00002841, -0.00006040, 0.00001838, 0.00001372, -0.00006342, -0.00002567, 0.00000177, 0.00000280, 0.00000038, -0.00001925, -0.00000184, -0.00000550, -0.00002201, -0.00000365, 0.00000134, 0.00000250, 0.00000125, -0.00000457, 0.00000308, -0.00000684, -0.00001305, -0.00000439, -0.00000529, 0.00000955, 0.00001008, -0.00000968, 0.00000449, 0.00000487, -0.00001674, -0.00000167, 0.00000472, -0.00001130, -0.00000395, 0.00000100, -0.00000673, -0.00000674, -0.00000649, -0.00000264, 0.00000182, -0.00000071, -0.00000103, -0.00000129, -0.00000092, -0.00000177, -0.00000953, -0.00000824, -0.00000044, },
		{ -0.00000008, -0.00000021, -0.00000381, -0.00000379, -0.00000234, -0.00000254, -0.00000004, 0.00000283, -0.00000213, -0.00000422, -0.00000135, -0.00000756, -0.00000863, -0.00000036, 0.00000042, -0.00000086, -0.00000287, -0.00000432, -0.00000182, -0.00000530, -0.00000378, 0.00000008, -0.00000302, 0.00000558, -0.00000051, -0.00001925, -0.00000011, 0.00000352, -0.00001936, 0.00000295, 0.00001082, -0.00001112, 0.00000552, 0.00000351, -0.00001211, 0.00000481, -0.00000345, -0.00000593, 0.00001009, -0.00000209, -0.00000510, -0.00001251, -0.00000720, 0.00001630, -0.00001472, -0.00001309, 0.00001295, -0.00000987, -0.00001028, -0.00003010, -0.00000237, 0.00003274, -0.00006474, -0.00000937, 0.00006330, -0.00004298, -0.00005007, -0.00009359, 0.00007038, 0.00012423, -0.00043054, 0.00008951, 0.00054576, -0.00091668, -0.00026272, 0.00119205, -0.00165471, -0.00219773, 0.00268993, 0.00359533, 0.00000201, -0.00123242, -0.00061086, -0.00016510, 0.00012172, -0.00013473, -0.00021607, 0.00005159, 0.00000856, -0.00006185, -0.00003533, -0.00000913, 0.00002596, -0.00004874, -0.00005359, 0.00001957, -0.00000119, -0.00001082, -0.00000388, -0.00000893, -0.00000369, -0.00002230, -0.00001144, 0.00000988, -0.00000328, -0.00000101, 0.00000185, -0.00000088, -0.00000324, -0.00001536, -0.00000453, 0.00000141, -0.00000476, 0.00001139, 0.00000393, -0.00000869, 0.00000790, -0.00000394, -0.00001475, 0.00000585, -0.00000137, -0.00001251, 0.00000040, -0.00000233, -0.00000915, -0.00000511, -0.00000352, 0.00000025, 0.00000090, -0.00000181, -0.00000117, -0.00000239, -0.00000139, -0.00000350, -0.00000991, -0.00000512, 0.00000042, -0.00000143, },
		{ -0.00000168, -0.00000373, -0.00000248, -0.00000197, -0.00000202, 0.00000113, 0.00000104, -0.00000470, -0.00000275, -0.00000135, -0.00000911, -0.00000632, 0.00000137, -0.00000008, -0.00000257, -0.00000500, -0.00000269, -0.00000091, -0.00000595, -0.00000163, 0.00000084, -0.00000138, 0.00000330, -0.00001059, -0.00001491, 0.00000831, -0.00000616, -0.00001756, 0.00001362, 0.00000314, -0.00001340, 0.00000899, 0.00000076, -0.00000696, 0.00000179, -0.00000674, 0.00000642, 0.00000516, -0.00001592, -0.00000537, -0.00000303, 0.00000321, 0.00000014, -0.00001855, 0.00001219, 0.00000033, -0.00003214, -0.00000636, -0.00000860, 0.00000913, -0.00001705, -0.00004848, 0.00006207, -0.00000356, -0.00010499, -0.00000834, -0.00000256, 0.00005461, -0.00013526, -0.00020943, 0.00044146, -0.00023011, -0.00091717, 0.00095183, 0.00000871, -0.00314578, 0.00029669, 0.00483099, 0.00174673, -0.00184215, -0.00096329, 0.00000728, -0.00000360, -0.00006926, -0.00023514, -0.00001836, 0.00008901, -0.00009945, -0.00004791, 0.00001465, -0.00000330, -0.00002333, -0.00006216, 0.00000482, 0.00002273, -0.00003003, -0.00000320, 0.00000418, -0.00001340, -0.00001747, -0.00001822, 0.00000911, 0.00000430, -0.00001178, 0.00000576, 0.00000309, -0.00000682, -0.00001159, -0.00000971, 0.00000681, -0.00000331, -0.00000463, 0.00001496, -0.00000254, -0.00000715, 0.00000818, -0.00000947, -0.00000836, 0.00000718, -0.00000835, -0.00000902, 0.00000160, -0.00000757, -0.00000755, -0.00000114, -0.00000170, 0.00000025, -0.00000065, -0.00000186, -0.00000214, -0.00000353, -0.00000091, -0.00000473, -0.00000941, -0.00000247, 0.00000033, -0.00000200, -0.00000051, },
		{ -0.00000276, -0.00000173, -0.00000195, -0.00000144, 0.00000095, -0.00000167, -0.00000490, -0.00000029, -0.00000309, -0.00001003, -0.00000299, 0.00000179, -0.00000227, -0.00000441, -0.00000427, -0.00000024, -0.00000230, -0.00000540, 0.00000193, -0.00000030, -0.00000309, -0.00000138, -0.00001412, -0.00000575, 0.00000780, -0.00001514, -0.00000661, 0.00001679, -0.00000998, -0.00000857, 0.00001487, -0.00000333, -0.00000686, 0.00000035, 0.00000040, 0.00000788, -0.00001306, -0.00001545, 0.00000792, -0.00000120, -0.00000341, -0.00000626, -0.00000025, 0.00001143, -0.00003276, -0.00002172, 0.00001611, -0.00001318, -0.00001006, -0.00002251, 0.00000879, 0.00003750, -0.00009844, -0.00004427, 0.00006098, -0.00004410, -0.00004804, -0.00011747, 0.00009285, 0.00014718, -0.00073292, 0.00004971, 0.00097465, -0.00215107, -0.00224382, 0.00383989, 0.00428589, -0.00128761, -0.00207581, 0.00015812, 0.00013765, -0.00024242, -0.00014378, -0.00008436, 0.00009022, -0.00004448, -0.00011799, 0.00004556, 0.00000101, -0.00004911, -0.00003360, -0.00001760, 0.00002860, -0.00001763, -0.00002814, 0.00002041, -0.00000988, -0.00002665, -0.00001305, 0.00000012, 0.00001088, -0.00001148, -0.00000431, 0.00001297, -0.00000802, -0.00001337, -0.00000755, 0.00000223, 0.00000599, -0.00001178, 0.00000366, 0.00001501, -0.00001202, -0.00000227, 0.00000841, -0.00001284, -0.00000299, 0.00000431, -0.00001172, -0.00000495, -0.00000163, -0.00000937, -0.00000167, 0.00000069, -0.00000275, -0.00000034, -0.00000118, -0.00000239, -0.00000395, -0.00000286, 0.00000021, -0.00000656, -0.00000832, -0.00000029, -0.00000075, -0.00000289, -0.00000088, -0.00000194, },
		{ -0.00000156, -0.00000211, -0.00000147, -0.00000022, -0.00000317, -0.00000292, 0.00000077, -0.00000604, -0.00000892, -0.00000020, -0.00000023, -0.00000431, -0.00000397, -0.00000215, 0.00000014, -0.00000371, -0.00000195, 0.00000286, -0.00000531, -0.00000437, -0.00000342, -0.00001312, 0.00000058, 0.00000212, -0.00001625, 0.00000455, 0.00000845, -0.00001711, 0.00000480, 0.00001394, -0.00001096, -0.00000298, 0.00000465, 0.00000175, -0.00000544, -0.00002007, 0.00000247, 0.00001010, -0.00001191, -0.00000199, 0.00000300, 0.00000341, -0.00001554, -0.00003697, 0.00001332, 0.00000589, -0.00003439, 0.00000163, 0.00000180, 0.00000771, -0.00003981, -0.00008330, 0.00005954, -0.00000292, -0.00012048, 0.00000194, -0.00000062, 0.00004462, -0.00027325, -0.00039903, 0.00072017, -0.00055907, -0.00301610, 0.00107298, 0.00553041, 0.00107298, -0.00301610, -0.00055907, 0.00072017, -0.00039903, -0.00027325, 0.00004462, -0.00000062, 0.00000194, -0.00012048, -0.00000292, 0.00005954, -0.00008330, -0.00003981, 0.00000771, 0.00000180, 0.00000163, -0.00003439, 0.00000589, 0.00001332, -0.00003697, -0.00001554, 0.00000341, 0.00000300, -0.00000199, -0.00001191, 0.00001010, 0.00000247, -0.00002007, -0.00000544, 0.00000175, 0.00000465, -0.00000298, -0.00001096, 0.00001394, 0.00000480, -0.00001711, 0.00000845, 0.00000455, -0.00001625, 0.00000212, 0.00000058, -0.00001312, -0.00000342, -0.00000437, -0.00000531, 0.00000286, -0.00000195, -0.00000371, 0.00000014, -0.00000215, -0.00000397, -0.00000431, -0.00000023, -0.00000020, -0.00000892, -0.00000604, 0.00000077, -0.00000292, -0.00000317, -0.00000022, -0.00000147, -0.00000211, },
		{ -0.00000276, -0.00000194, -0.00000088, -0.00000289, -0.00000075, -0.00000029, -0.00000832, -0.00000656, 0.00000021, -0.00000286, -0.00000395, -0.00000239, -0.00000118, -0.00000034, -0.00000275, 0.00000069, -0.00000167, -0.00000937, -0.00000163, -0.00000495, -0.00001172, 0.00000431, -0.00000299, -0.00001284, 0.00000841, -0.00000227, -0.00001202, 0.00001501, 0.00000366, -0.00001178, 0.00000599, 0.00000223, -0.00000755, -0.00001337, -0.00000802, 0.00001297, -0.00000431, -0.00001148, 0.00001088, 0.00000012, -0.00001305, -0.00002665, -0.00000988, 0.00002041, -0.00002814, -0.00001763, 0.00002860, -0.00001760, -0.00003360, -0.00004911, 0.00000101, 0.00004556, -0.00011799, -0.00004448, 0.00009022, -0.00008436, -0.00014378, -0.00024242, 0.00013765, 0.00015812, -0.00207581, -0.00128761, 0.00428589, 0.00383989, -0.00224382, -0.00215107, 0.00097465, 0.00004971, -0.00073292, 0.00014718, 0.00009285, -0.00011747, -0.00004804, -0.00004410, 0.00006098, -0.00004427, -0.00009844, 0.00003750, 0.00000879, -0.00002251, -0.00001006, -0.00001318, 0.00001611, -0.00002172, -0.00003276, 0.00001143, -0.00000025, -0.00000626, -0.00000341, -0.00000120, 0.00000792, -0.00001545, -0.00001306, 0.00000788, 0.00000040, 0.00000035, -0.00000686, -0.00000333, 0.00001487, -0.00000857, -0.00000998, 0.00001679, -0.00000661, -0.00001514, 0.00000780, -0.00000575, -0.00001412, -0.00000138, -0.00000309, -0.00000030, 0.00000193, -0.00000540, -0.00000230, -0.00000024, -0.00000427, -0.00000441, -0.00000227, 0.00000179, -0.00000299, -0.00001003, -0.00000309, -0.00000029, -0.00000490, -0.00000167, 0.00000095, -0.00000144, -0.00000195, -0.00000173, },
		{ -0.00000168, -0.00000051, -0.00000200, 0.00000033, -0.00000247, -0.00000941, -0.00000473, -0.00000091, -0.00000353, -0.00000214, -0.00000186, -0.00000065, 0.00000025, -0.00000170, -0.00000114, -0.00000755, -0.00000757, 0.00000160, -0.00000902, -0.00000835, 0.00000718, -0.00000836, -0.00000947, 0.00000818, -0.00000715, -0.00000254, 0.00001496, -0.00000463, -0.00000331, 0.00000681, -0.00000971, -0.00001159, -0.00000682, 0.00000309, 0.00000576, -0.00001178, 0.00000430, 0.00000911, -0.00001822, -0.00001747, -0.00001340, 0.00000418, -0.00000320, -0.00003003, 0.00002273, 0.00000482, -0.00006216, -0.00002333, -0.00000330, 0.00001465, -0.00004791, -0.00009945, 0.00008901, -0.00001836, -0.00023514, -0.00006926, -0.00000360, 0.00000728, -0.00096329, -0.00184215, 0.00174673, 0.00483099, 0.00029669, -0.00314578, 0.00000871, 0.00095183, -0.00091717, -0.00023011, 0.00044146, -0.00020943, -0.00013526, 0.00005461, -0.00000256, -0.00000834, -0.00010499, -0.00000356, 0.00006207, -0.00004848, -0.00001705, 0.00000913, -0.00000860, -0.00000636, -0.00003214, 0.00000033, 0.00001219, -0.00001855, 0.00000014, 0.00000321, -0.00000303, -0.00000537, -0.00001592, 0.00000516, 0.00000642, -0.00000674, 0.00000179, -0.00000696, 0.00000076, 0.00000899, -0.00001340, 0.00000314, 0.00001362, -0.00001756, -0.00000616, 0.00000831, -0.00001491, -0.00001059, 0.00000330, -0.00000138, 0.00000084, -0.00000163, -0.00000595, -0.00000091, -0.00000269, -0.00000500, -0.00000257, -0.00000008, 0.00000137, -0.00000632, -0.00000911, -0.00000135, -0.00000275, -0.00000470, 0.00000104, 0.00000113, -0.00000202, -0.00000197, -0.00000248, -0.00000373, },
		{ -0.00000008, -0.00000143, 0.00000042, -0.00000512, -0.00000991, -0.00000350, -0.00000139, -0.00000239, -0.00000117, -0.00000181, 0.00000090, 0.00000025, -0.00000352, -0.00000511, -0.00000915, -0.00000233, 0.00000040, -0.00001251, -0.00000137, 0.00000585, -0.00001475, -0.00000394, 0.00000790, -0.00000869, 0.00000393, 0.00001139, -0.00000476, 0.00000141, -0.00000453, -0.00001536, -0.00000324, -0.00000088, 0.00000185, -0.00000101, -0.00000328, 0.00000988, -0.00001144, -0.00002230, -0.00000369, -0.00000893, -0.00000388, -0.00001082, -0.00000119, 0.00001957, -0.00005359, -0.00004874, 0.00002596, -0.00000913, -0.00003533, -0.00006185, 0.00000856, 0.00005159, -0.00021607, -0.00013473, 0.00012172, -0.00016510, -0.00061086, -0.00123242, 0.00000201, 0.00359533, 0.00268993, -0.00219773, -0.00165471, 0.00119205, -0.00026272, -0.00091668, 0.00054576, 0.00008951, -0.00043054, 0.00012423, 0.00007038, -0.00009359, -0.00005007, -0.00004298, 0.00006330, -0.00000937, -0.00006474, 0.00003274, -0.00000237, -0.00003010, -0.00001028, -0.00000987, 0.00001295, -0.00001309, -0.00001472, 0.00001630, -0.00000720, -0.00001251, -0.00000510, -0.00000209, 0.00001009, -0.00000593, -0.00000345, 0.00000481, -0.00001211, 0.00000351, 0.00000552, -0.00001112, 0.00001082, 0.00000295, -0.00001936, 0.00000352, -0.00000011, -0.00001925, -0.00000051, 0.00000558, -0.00000302, 0.00000008, -0.00000378, -0.00000530, -0.00000182, -0.00000432, -0.00000287, -0.00000086, 0.00000042, -0.00000036, -0.00000863, -0.00000756, -0.00000135, -0.00000422, -0.00000213, 0.00000283, -0.00000004, -0.00000254, -0.00000234, -0.00000379, -0.00000381, -0.00000021, },
		{ -0.00000101, -0.00000044, -0.00000824, -0.00000953, -0.00000177, -0.00000092, -0.00000129, -0.00000103, -0.00000071, 0.00000182, -0.00000264, -0.00000649, -0.00000674, -0.00000673, 0.00000100, -0.00000395, -0.00001130, 0.00000472, -0.00000167, -0.00001674, 0.00000487, 0.00000449, -0.00000968, 0.00001008, 0.00000955, -0.00000529, -0.00000439, -0.00001305, -0.00000684, 0.00000308, -0.00000457, 0.00000125, 0.00000250, 0.00000134, -0.00000365, -0.00002201, -0.00000550, -0.00000184, -0.00001925, 0.00000038, 0.00000280, 0.00000177, -0.00002567, -0.00006342, 0.00001372, 0.00001838, -0.00006040, -0.00002841, -0.00000474, 0.00001073, -0.00010964, -0.00020093, 0.00010669, -0.00004908, -0.00065604, -0.00075869, -0.00030533, 0.00174244, 0.00319695, 0.00004599, -0.00232624, 0.00016156, 0.00066820, -0.00107186, 0.00000174, 0.00061064, -0.00051357, -0.00013592, 0.00031560, -0.00015335, -0.00011276, 0.00002860, 0.00000591, 0.00002443, -0.00006295, -0.00000268, 0.00004147, -0.00005292, -0.00001960, 0.00001270, -0.00000308, -0.00000418, -0.00002076, 0.00000988, 0.00000852, -0.00002705, -0.00000369, 0.00000599, 0.00000201, 0.00000127, -0.00000878, 0.00000652, -0.00000348, -0.00001607, 0.00001255, 0.00000182, -0.00000944, 0.00001299, -0.00000585, -0.00001488, 0.00000560, -0.00000998, -0.00001285, 0.00000857, 0.00000105, -0.00000462, 0.00000060, -0.00000575, -0.00000582, -0.00000260, -0.00000282, -0.00000069, -0.00000120, 0.00000052, -0.00000226, -0.00001052, -0.00000616, -0.00000145, -0.00000358, 0.00000066, 0.00000265, -0.00000149, -0.00000293, -0.00000340, -0.00000458, -0.00000223, 0.00000134, -0.00000051, },
		{ -0.00000271, -0.00001097, -0.00000729, 0.00000034, -0.00000060, -0.00000060, -0.00000053, 0.00000030, -0.00000027, -0.00000666, -0.00000703, -0.00000585, -0.00000398, 0.00000135, -0.00000737, -0.00000709, 0.00000521, -0.00000987, -0.00001071, 0.00001067, -0.00000338, -0.00000559, 0.00001713, 0.00000360, -0.00001176, -0.00000980, -0.00000871, 0.00000238, -0.00000194, -0.00000560, 0.00000844, 0.00000165, -0.00000758, -0.00001346, -0.00001132, 0.00000293, -0.00001796, -0.00001275, 0.00001879, -0.00000585, -0.00002715, -0.00004227, -0.00001724, 0.00003116, -0.00004426, -0.00005047, 0.00002832, -0.00002550, -0.00008716, -0.00013860, -0.00000881, 0.00005299, -0.00054924, -0.00074518, -0.00007979, 0.00079734, 0.00215185, 0.00161699, -0.00125513, -0.00112824, 0.00071681, -0.00039566, -0.00069989, 0.00068319, -0.00007760, -0.00056852, 0.00037339, 0.00006804, -0.00032441, 0.00008058, 0.00005508, -0.00004654, -0.00000979, -0.00003082, 0.00003963, -0.00002021, -0.00006235, 0.00003056, 0.00000792, -0.00002136, -0.00000835, -0.00000330, 0.00001470, -0.00002006, -0.00002193, 0.00001814, 0.00000274, -0.00000434, -0.00000148, -0.00000166, 0.00000572, -0.00001715, -0.00000602, 0.00002040, -0.00000919, -0.00000539, 0.00001483, -0.00001317, -0.00001127, 0.00000329, -0.00001213, -0.00000191, 0.00000861, -0.00000538, -0.00000204, -0.00000019, -0.00000935, -0.00000518, -0.00000075, -0.00000080, -0.00000116, -0.00000166, 0.00000122, -0.00000549, -0.00001200, -0.00000368, -0.00000073, -0.00000232, 0.00000204, 0.00000143, -0.00000266, -0.00000376, -0.00000433, -0.00000365, 0.00000010, 0.00000156, -0.00000139, -0.00000055, },
		{ -0.00001164, -0.00000346, 0.00000154, -0.00000053, 0.00000037, -0.00000035, -0.00000094, -0.00000412, -0.00000811, -0.00000525, -0.00000508, -0.00000190, 0.00000052, -0.00000936, -0.00000368, 0.00000199, -0.00001269, -0.00000171, 0.00000869, -0.00000866, 0.00000502, 0.00001704, -0.00000920, -0.00001530, -0.00000577, -0.00000124, 0.00000110, -0.00000678, 0.00000342, 0.00000934, -0.00001062, -0.00001257, -0.00000748, -0.00000259, -0.00000790, -0.00002087, 0.00001271, 0.00001052, -0.00003950, -0.00003196, -0.00001477, 0.00000610, -0.00001217, -0.00005923, 0.00001852, 0.00000394, -0.00012027, -0.00008762, -0.00002283, -0.00000311, -0.00034566, -0.00074644, -0.00010709, 0.00070969, 0.00111520, 0.00158358, 0.00037502, -0.00122158, -0.00020169, 0.00024446, -0.00073649, 0.00013038, 0.00045069, -0.00062199, -0.00000094, 0.00042783, -0.00037267, -0.00011286, 0.00023939, -0.00010025, -0.00005793, 0.00003650, -0.00000728, 0.00000530, -0.00006094, -0.00000318, 0.00004552, -0.00003400, -0.00001602, 0.00000861, -0.00000046, -0.00000598, -0.00002844, 0.00000754, 0.00001977, -0.00001318, -0.00000042, 0.00000252, -0.00000141, -0.00000453, -0.00001832, 0.00001172, 0.00001334, -0.00001970, 0.00000728, 0.00001145, -0.00002256, -0.00000597, 0.00000367, -0.00000912, 0.00000372, 0.00000267, -0.00000661, 0.00000118, -0.00000529, -0.00001097, -0.00000103, 0.00000106, -0.00000133, -0.00000236, -0.00000033, 0.00000042, -0.00001019, -0.00001077, 0.00000010, -0.00000073, -0.00000166, 0.00000255, -0.00000003, -0.00000403, -0.00000459, -0.00000376, -0.00000153, 0.00000136, 0.00000048, -0.00000185, -0.00000070, -0.00000609, },
		{ 0.00000024, 0.00000150, 0.00000011, 0.00000086, -0.00000217, -0.00000373, -0.00000629, -0.00000660, -0.00000400, -0.00000475, 0.00000040, -0.00000178, -0.00001097, -0.00000087, -0.00000096, -0.00001114, 0.00000398, 0.00000338, -0.00000618, 0.00001304, 0.00000564, -0.00001857, -0.00000874, 0.00000011, -0.00000108, -0.00000277, -0.00000201, 0.00000914, -0.00000355, -0.00001757, -0.00000410, -0.00000326, -0.00000776, -0.00001297, -0.00000250, 0.00001676, -0.00002846, -0.00004562, -0.00000048, -0.00000506, -0.00001720, -0.00003155, -0.00000984, 0.00002221, -0.00010342, -0.00011615, 0.00002406, -0.00004424, -0.00028177, -0.00058036, -0.00027406, 0.00067257, 0.00083622, 0.00081186, 0.00102737, -0.00013282, -0.00085418, 0.00000362, -0.00020688, -0.00035308, 0.00043894, -0.00018474, -0.00043705, 0.00046043, -0.00005372, -0.00042463, 0.00028014, 0.00006786, -0.00022970, 0.00008378, 0.00003986, -0.00005565, -0.00001812, -0.00002801, 0.00004017, -0.00000269, -0.00004615, 0.00002108, 0.00000170, -0.00001890, -0.00001240, -0.00000834, 0.00002239, -0.00000178, -0.00001361, 0.00001199, -0.00000418, -0.00000701, -0.00000749, -0.00000780, 0.00001867, -0.00000361, -0.00001582, 0.00001988, -0.00000387, -0.00002643, 0.00000593, 0.00000343, -0.00000840, 0.00000441, -0.00000143, -0.00000434, -0.00000039, -0.00001057, -0.00000691, 0.00000302, -0.00000037, -0.00000257, -0.00000175, 0.00000078, -0.00000338, -0.00001307, -0.00000590, 0.00000245, -0.00000238, -0.00000069, 0.00000266, -0.00000216, -0.00000532, -0.00000396, -0.00000185, -0.00000013, 0.00000108, -0.00000076, -0.00000189, -0.00000204, -0.00000906, -0.00000955, },
		{ 0.00000116, 0.00000093, -0.00000085, -0.00000537, -0.00000519, -0.00000577, -0.00000478, -0.00000410, -0.00000335, 0.00000168, -0.00000626, -0.00001048, 0.00000261, -0.00000383, -0.00000851, 0.00000687, 0.00000024, -0.00000127, 0.00001108, -0.00000740, -0.00001576, -0.00000002, -0.00000098, -0.00000277, -0.00000039, 0.00000282, 0.00000183, -0.00001466, -0.00000924, 0.00000140, -0.00001015, -0.00000799, -0.00000258, 0.00000386, -0.00001177, -0.00004715, -0.00001016, 0.00001013, -0.00003303, -0.00002158, -0.00000200, 0.00000088, -0.00006223, -0.00013194, 0.00000849, 0.00000447, -0.00029579, -0.00045705, -0.00027692, 0.00043758, 0.00086659, 0.00042320, 0.00061979, 0.00079919, -0.00039248, -0.00066311, -0.00001870, -0.00023022, 0.00001790, 0.00016941, -0.00042791, 0.00008887, 0.00030897, -0.00045229, -0.00000044, 0.00034177, -0.00026995, -0.00007314, 0.00019635, -0.00009566, -0.00006157, 0.00002761, -0.00000181, 0.00001682, -0.00003895, -0.00000336, 0.00002839, -0.00003558, -0.00001453, 0.00000459, 0.00000322, 0.00001118, -0.00001338, 0.00000394, 0.00000842, -0.00001784, -0.00000157, -0.00000387, -0.00000167, 0.00001371, -0.00001317, -0.00000159, 0.00001791, -0.00002169, -0.00001506, 0.00001659, -0.00000424, -0.00000794, 0.00000581, -0.00000377, -0.00000415, -0.00000430, -0.00001036, -0.00000083, 0.00000254, -0.00000266, -0.00000182, -0.00000100, -0.00000059, -0.00000743, -0.00001158, -0.00000063, 0.00000133, -0.00000372, 0.00000117, 0.00000137, -0.00000463, -0.00000492, -0.00000182, -0.00000060, -0.00000024, 0.00000034, -0.00000160, -0.00000219, -0.00000399, -0.00001014, -0.00000580, 0.00000245, },
		{ 0.00000030, -0.00000458, -0.00000718, -0.00000420, -0.00000455, -0.00000418, -0.00000411, -0.00000131, 0.00000006, -0.00001041, -0.00000613, 0.00000443, -0.00000768, -0.00000396, 0.00000896, -0.00000273, -0.00000034, 0.00000450, -0.00001147, -0.00000711, 0.00000121, -0.00000382, 0.00000081, 0.00000153, -0.00000145, -0.00000721, -0.00001191, -0.00000056, -0.00000568, -0.00001067, 0.00000383, -0.00000285, -0.00001502, -0.00003078, -0.00002382, 0.00001273, -0.00002186, -0.00003899, 0.00001367, -0.00000446, -0.00005916, -0.00009994, -0.00003623, 0.00003192, -0.00024153, -0.00045285, -0.00020035, 0.00027635, 0.00072953, 0.00048853, 0.00014021, 0.00076526, 0.00054261, -0.00067983, -0.00046876, 0.00005490, -0.00014156, 0.00005204, -0.00010763, -0.00020447, 0.00029341, -0.00013771, -0.00031684, 0.00036377, -0.00002116, -0.00032334, 0.00022812, 0.00004963, -0.00020243, 0.00006333, 0.00003749, -0.00003781, -0.00000210, -0.00001743, 0.00002535, -0.00001615, -0.00004382, 0.00002120, 0.00000441, -0.00000538, 0.00000469, -0.00000495, 0.00001033, -0.00001123, -0.00001459, 0.00001093, -0.00000950, -0.00000014, 0.00001025, -0.00001354, 0.00000710, 0.00000473, -0.00002555, 0.00000413, 0.00001383, -0.00001390, -0.00000151, 0.00000576, -0.00000832, -0.00000449, -0.00000504, -0.00000681, 0.00000142, -0.00000024, -0.00000223, -0.00000110, -0.00000227, -0.00000237, -0.00000872, -0.00000789, 0.00000162, -0.00000134, -0.00000299, 0.00000216, -0.00000141, -0.00000532, -0.00000255, -0.00000032, -0.00000108, -0.00000070, -0.00000017, -0.00000248, -0.00000266, -0.00000543, -0.00000930, -0.00000201, 0.00000327, 0.00000095, },
		{ -0.00000780, -0.00000605, -0.00000246, -0.00000425, -0.00000410, -0.00000351, -0.00000043, -0.00000340, -0.00001098, -0.00000042, 0.00000218, -0.00000940, 0.00000316, 0.00000699, -0.00000841, -0.00000037, 0.00000181, -0.00000907, -0.00000289, -0.00000144, -0.00000106, 0.00000339, -0.00000424, -0.00000665, -0.00000679, -0.00000582, -0.00000300, -0.00000926, 0.00000248, 0.00000322, -0.00002418, -0.00002556, -0.00001688, -0.00000188, -0.00000633, -0.00004104, -0.00000020, 0.00001524, -0.00006671, -0.00008241, -0.00003543, 0.00000045, -0.00016240, -0.00042965, -0.00021523, 0.00026780, 0.00053273, 0.00051437, 0.00010877, 0.00027031, 0.00087516, 0.00006750, -0.00078823, -0.00014455, 0.00007239, -0.00017072, -0.00000557, -0.00012045, 0.00002443, 0.00010521, -0.00030658, 0.00008599, 0.00025821, -0.00034338, -0.00000195, 0.00027384, -0.00022751, -0.00006932, 0.00016579, -0.00007173, -0.00004196, 0.00003309, -0.00000375, -0.00000117, -0.00004594, -0.00000003, 0.00003326, -0.00002344, 0.00000000, 0.00001251, -0.00000341, -0.00000158, -0.00001830, 0.00000519, 0.00000730, -0.00002050, 0.00001032, 0.00000741, -0.00001585, 0.00000909, -0.00000410, -0.00001625, 0.00001421, 0.00000097, -0.00001423, 0.00000649, -0.00000115, -0.00001141, -0.00000087, -0.00000444, -0.00000540, 0.00000098, -0.00000075, -0.00000113, -0.00000309, -0.00000306, -0.00000210, -0.00000879, -0.00000520, 0.00000159, -0.00000293, -0.00000161, 0.00000115, -0.00000321, -0.00000343, -0.00000058, -0.00000085, -0.00000198, -0.00000055, -0.00000101, -0.00000331, -0.00000253, -0.00000612, -0.00000749, 0.00000106, 0.00000309, 0.00000006, -0.00000236, },
		{ -0.00000327, -0.00000187, -0.00000459, -0.00000395, -0.00000305, -0.00000079, -0.00000602, -0.00000841, 0.00000283, -0.00000172, -0.00000621, 0.00000776, -0.00000169, -0.00001201, 0.00000411, 0.00000169, -0.00000806, -0.00000235, -0.00000045, 0.00000231, -0.00000217, -0.00001000, -0.00000390, -0.00000432, -0.00000673, -0.00000478, 0.00000003, 0.00000629, -0.00002052, -0.00003516, -0.00000783, -0.00000374, -0.00001156, -0.00002426, -0.00001681, 0.00001619, -0.00004693, -0.00009253, -0.00001866, -0.00001455, -0.00014391, -0.00034678, -0.00026150, 0.00023770, 0.00047558, 0.00038524, 0.00022076, 0.00005755, 0.00052403, 0.00067823, -0.00037072, -0.00058360, 0.00009367, -0.00007633, -0.00017786, 0.00003410, -0.00006550, 0.00003238, -0.00008235, -0.00012886, 0.00024759, -0.00009500, -0.00025475, 0.00028517, -0.00001793, -0.00027071, 0.00018754, 0.00004981, -0.00016197, 0.00006129, 0.00003527, -0.00004286, -0.00001939, -0.00001987, 0.00003392, -0.00000242, -0.00003075, 0.00002754, 0.00000405, -0.00001524, -0.00000512, -0.00000614, 0.00001335, -0.00001244, -0.00001500, 0.00002432, -0.00000747, -0.00001682, 0.00001528, -0.00000693, -0.00000802, 0.00001264, -0.00000802, -0.00000674, 0.00000626, -0.00001008, -0.00000667, 0.00000298, -0.00000744, -0.00000484, 0.00000215, -0.00000042, -0.00000295, -0.00000495, -0.00000048, -0.00000215, -0.00001006, -0.00000291, 0.00000130, -0.00000395, -0.00000119, 0.00000016, -0.00000243, -0.00000142, -0.00000087, -0.00000206, -0.00000197, -0.00000057, -0.00000232, -0.00000299, -0.00000186, -0.00000679, -0.00000516, 0.00000316, 0.00000161, -0.00000201, -0.00000540, -0.00000838, },
		{ -0.00000257, -0.00000489, -0.00000398, -0.00000265, -0.00000119, -0.00000725, -0.00000524, 0.00000364, -0.00000350, -0.00000151, 0.00000467, -0.00001137, -0.00000730, 0.00001004, -0.00000164, -0.00000818, 0.00000014, 0.00000151, -0.00000109, -0.00000934, -0.00000714, -0.00000025, -0.00000776, -0.00000581, 0.00000273, 0.00000372, -0.00001267, -0.00003795, -0.00001549, 0.00000716, -0.00001736, -0.00002048, -0.00001012, -0.00000193, -0.00002791, -0.00008603, -0.00003014, 0.00000460, -0.00014759, -0.00029346, -0.00024028, 0.00014281, 0.00046930, 0.00031565, 0.00017859, 0.00013757, 0.00017515, 0.00060462, 0.00032619, -0.00053137, -0.00027847, 0.00008458, -0.00021255, -0.00005420, 0.00005457, -0.00009750, -0.00000982, -0.00007265, 0.00004876, 0.00009565, -0.00024383, 0.00005656, 0.00020578, -0.00027815, -0.00000283, 0.00023328, -0.00018190, -0.00005354, 0.00014439, -0.00006510, -0.00005144, 0.00001578, 0.00000351, 0.00001667, -0.00003129, 0.00000366, 0.00003293, -0.00002593, -0.00001008, 0.00000590, -0.00000076, 0.00000250, -0.00002154, 0.00000649, 0.00001975, -0.00002726, -0.00000200, 0.00002174, -0.00001444, -0.00000348, 0.00000989, -0.00001033, -0.00000232, -0.00000037, -0.00001103, 0.00000179, -0.00000005, -0.00001116, -0.00000078, 0.00000314, -0.00000297, -0.00000523, -0.00000260, 0.00000244, -0.00000575, -0.00001053, 0.00000052, -0.00000026, -0.00000541, -0.00000017, 0.00000104, -0.00000122, -0.00000165, -0.00000202, -0.00000227, -0.00000182, -0.00000135, -0.00000277, -0.00000128, -0.00000202, -0.00000752, -0.00000228, 0.00000356, -0.00000127, -0.00000411, -0.00000689, -0.00000668, -0.00000105, },
		{ -0.00000512, -0.00000393, -0.00000158, -0.00000175, -0.00000797, -0.00000191, 0.00000390, -0.00000400, -0.00000068, -0.00000309, -0.00001362, 0.00000295, 0.00001034, -0.00000681, -0.00000500, 0.00000254, -0.00000158, -0.00000704, -0.00000917, -0.00000167, -0.00000348, -0.00000856, 0.00000352, 0.00000447, -0.00001140, -0.00003053, -0.00002572, 0.00000662, -0.00000680, -0.00002847, -0.00000252, -0.00000457, -0.00003424, -0.00006339, -0.00004290, 0.00000813, -0.00011916, -0.00028646, -0.00020086, 0.00008868, 0.00039456, 0.00034265, 0.00009809, 0.00015761, 0.00015574, 0.00024168, 0.00054526, 0.00004768, -0.00049122, -0.00010363, -0.00003773, -0.00018370, 0.00006247, -0.00003770, -0.00011063, 0.00003222, -0.00001976, 0.00004559, -0.00006525, -0.00011480, 0.00018921, -0.00007296, -0.00020377, 0.00023971, -0.00000731, -0.00022293, 0.00015971, 0.00004119, -0.00014788, 0.00003868, 0.00002808, -0.00002075, -0.00000098, -0.00001575, 0.00002970, -0.00000180, -0.00003401, 0.00001726, 0.00000218, -0.00000909, -0.00000346, -0.00001257, 0.00001784, -0.00000293, -0.00002769, 0.00002264, 0.00001231, -0.00002360, 0.00000616, 0.00000707, -0.00001368, -0.00000191, -0.00000324, -0.00000525, 0.00000401, -0.00000732, -0.00000900, 0.00000392, -0.00000024, -0.00000581, -0.00000344, 0.00000147, 0.00000095, -0.00001038, -0.00000732, 0.00000265, -0.00000416, -0.00000501, 0.00000294, 0.00000175, -0.00000215, -0.00000282, -0.00000203, -0.00000232, -0.00000238, -0.00000162, -0.00000152, 0.00000018, -0.00000368, -0.00000731, 0.00000036, 0.00000153, -0.00000413, -0.00000466, -0.00000681, -0.00000449, -0.00000027, -0.00000380, },
		{ -0.00000305, -0.00000028, -0.00000336, -0.00000757, 0.00000209, 0.00000271, -0.00000628, -0.00000318, -0.00000799, -0.00000790, 0.00001038, 0.00000504, -0.00000832, -0.00000071, -0.00000016, -0.00000664, -0.00000787, -0.00000525, -0.00000181, -0.00000610, 0.00000005, 0.00000695, -0.00001093, -0.00002603, -0.00002437, -0.00000510, 0.00000212, -0.00002409, -0.00001082, 0.00000572, -0.00003999, -0.00005916, -0.00003257, -0.00000683, -0.00008740, -0.00026431, -0.00020508, 0.00008873, 0.00031172, 0.00033259, 0.00011761, 0.00007015, 0.00021393, 0.00011066, 0.00028138, 0.00047206, -0.00014733, -0.00041681, -0.00004386, -0.00009580, -0.00007945, 0.00004633, -0.00012376, -0.00001996, 0.00006005, -0.00004910, -0.00000380, -0.00006842, 0.00002570, 0.00007566, -0.00018859, 0.00005436, 0.00017634, -0.00022843, -0.00000266, 0.00019524, -0.00016045, -0.00005532, 0.00011763, -0.00004842, -0.00002570, 0.00002137, -0.00000191, 0.00001396, -0.00002812, -0.00000271, 0.00002519, -0.00002204, -0.00000444, 0.00000193, -0.00000766, 0.00001149, -0.00001726, -0.00000632, 0.00003193, -0.00001030, -0.00001887, 0.00001734, -0.00000389, -0.00001646, 0.00000236, -0.00000203, -0.00000235, 0.00000004, -0.00001014, -0.00000310, 0.00000307, -0.00000485, -0.00000391, 0.00000022, 0.00000202, -0.00000350, -0.00001145, -0.00000276, 0.00000066, -0.00000706, -0.00000065, 0.00000543, -0.00000031, -0.00000389, -0.00000258, -0.00000174, -0.00000335, -0.00000233, -0.00000034, 0.00000000, -0.00000008, -0.00000574, -0.00000590, 0.00000117, -0.00000188, -0.00000484, -0.00000390, -0.00000658, -0.00000278, -0.00000056, -0.00000506, -0.00000511, },
		{ -0.00000005, -0.00000528, -0.00000485, 0.00000478, -0.00000200, -0.00000929, -0.00000424, -0.00000764, -0.00000045, 0.00001188, -0.00000028, -0.00000655, -0.00000070, -0.00000532, -0.00000735, -0.00000577, -0.00000449, -0.00000260, -0.00000065, 0.00000525, -0.00000710, -0.00002582, -0.00001979, -0.00000975, -0.00000372, -0.00001139, -0.00001603, 0.00000377, -0.00002805, -0.00006764, -0.00002579, -0.00000585, -0.00008138, -0.00022088, -0.00021599, 0.00006885, 0.00028793, 0.00027217, 0.00015437, 0.00004158, 0.00015311, 0.00017524, 0.00005646, 0.00038007, 0.00035784, -0.00032359, -0.00032259, 0.00000056, -0.00009775, -0.00003293, -0.00002598, -0.00009331, 0.00006789, -0.00000968, -0.00007489, 0.00001700, -0.00002863, 0.00002897, -0.00004533, -0.00007963, 0.00016375, -0.00005833, -0.00017153, 0.00019834, -0.00001180, -0.00019556, 0.00013330, 0.00003652, -0.00011531, 0.00004791, 0.00002231, -0.00002517, 0.00000089, -0.00001414, 0.00002181, -0.00000368, -0.00002814, 0.00001793, -0.00000611, -0.00001222, 0.00000868, -0.00001377, 0.00001206, 0.00001940, -0.00002354, -0.00000130, 0.00001503, -0.00001837, -0.00000896, 0.00000824, -0.00000492, -0.00000301, -0.00000211, -0.00000848, -0.00000099, -0.00000105, -0.00000421, 0.00000024, 0.00000037, -0.00000004, -0.00000627, -0.00000984, -0.00000103, -0.00000280, -0.00000517, 0.00000446, 0.00000386, -0.00000345, -0.00000380, -0.00000175, -0.00000280, -0.00000395, -0.00000036, 0.00000105, 0.00000024, -0.00000144, -0.00000712, -0.00000456, -0.00000006, -0.00000398, -0.00000312, -0.00000386, -0.00000683, -0.00000127, -0.00000153, -0.00000606, -0.00000433, -0.00000147, },
		{ -0.00000580, -0.00000133, 0.00000309, -0.00000835, -0.00000889, -0.00000255, -0.00000546, 0.00000490, 0.00001032, -0.00000372, -0.00000609, -0.00000388, -0.00000695, -0.00000548, -0.00000605, -0.00000337, 0.00000224, 0.00000297, -0.00000527, -0.00002314, -0.00002021, -0.00000614, -0.00001057, -0.00001013, -0.00000760, -0.00000525, -0.00001901, -0.00006179, -0.00003768, 0.00000470, -0.00007545, -0.00019499, -0.00019714, 0.00002410, 0.00027437, 0.00024642, 0.00012967, 0.00007868, 0.00008226, 0.00017385, 0.00007465, 0.00011057, 0.00046954, 0.00013959, -0.00043014, -0.00017240, 0.00001680, -0.00011490, -0.00002612, -0.00004206, -0.00001571, 0.00004919, -0.00008578, -0.00001780, 0.00003615, -0.00004848, -0.00000227, -0.00004038, 0.00003519, 0.00006540, -0.00015975, 0.00004130, 0.00014215, -0.00019939, -0.00000128, 0.00016988, -0.00013464, -0.00003589, 0.00010852, -0.00005101, -0.00002623, 0.00002385, -0.00000212, 0.00000861, -0.00002636, -0.00000020, 0.00002114, -0.00003076, -0.00000171, 0.00001500, -0.00001146, 0.00001614, 0.00000482, -0.00001987, 0.00000955, -0.00000013, -0.00002039, 0.00000524, 0.00000492, -0.00001032, 0.00000008, -0.00000244, -0.00000911, -0.00000189, -0.00000110, -0.00000018, 0.00000023, -0.00000163, -0.00000037, -0.00000792, -0.00000916, -0.00000092, -0.00000317, -0.00000066, 0.00000557, -0.00000007, -0.00000433, -0.00000271, -0.00000234, -0.00000394, -0.00000223, 0.00000195, 0.00000081, -0.00000019, -0.00000299, -0.00000838, -0.00000400, -0.00000130, -0.00000337, -0.00000141, -0.00000575, -0.00000643, 0.00000025, -0.00000313, -0.00000626, -0.00000272, -0.00000023, -0.00000095, },
		{ -0.00000021, -0.00000220, -0.00001126, -0.00000441, -0.00000142, -0.00000362, 0.00000886, 0.00000727, -0.00000739, -0.00000723, -0.00000531, -0.00000585, -0.00000577, -0.00000535, 0.00000286, 0.00000456, -0.00000636, -0.00001960, -0.00002089, -0.00000513, -0.00000783, -0.00001730, -0.00000249, -0.00000258, -0.00002288, -0.00004803, -0.00004481, -0.00000136, -0.00005479, -0.00018285, -0.00017642, 0.00000192, 0.00022860, 0.00025457, 0.00010266, 0.00007780, 0.00009211, 0.00009980, 0.00012460, 0.00003670, 0.00023921, 0.00043297, -0.00011061, -0.00039359, -0.00003230, -0.00003803, -0.00011989, 0.00000612, -0.00002454, 0.00000550, -0.00001795, -0.00007215, 0.00004629, -0.00001606, -0.00006317, 0.00002345, -0.00000505, 0.00003216, -0.00004115, -0.00007237, 0.00013069, -0.00005648, -0.00014430, 0.00017780, -0.00000824, -0.00016642, 0.00012461, 0.00003011, -0.00011029, 0.00004485, 0.00002612, -0.00002420, -0.00000304, -0.00001385, 0.00002062, -0.00000936, -0.00003491, 0.00002731, 0.00000676, -0.00001609, 0.00002063, -0.00000231, -0.00001479, 0.00000740, -0.00000985, -0.00000903, 0.00001027, -0.00000591, -0.00000797, 0.00000501, -0.00000774, -0.00001071, 0.00000110, 0.00000194, -0.00000003, -0.00000276, -0.00000043, -0.00000049, -0.00001168, -0.00000792, 0.00000128, -0.00000157, 0.00000146, 0.00000335, -0.00000222, -0.00000360, -0.00000304, -0.00000325, -0.00000295, 0.00000045, 0.00000229, -0.00000017, -0.00000029, -0.00000551, -0.00000976, -0.00000285, -0.00000098, -0.00000173, -0.00000203, -0.00000814, -0.00000431, 0.00000098, -0.00000502, -0.00000522, -0.00000101, 0.00000018, -0.00000188, -0.00000515, },
		{ -0.00000646, -0.00000864, -0.00000000, -0.00000273, -0.00000074, 0.00001146, 0.00000171, -0.00001096, -0.00000688, -0.00000493, -0.00000600, -0.00000551, 0.00000075, 0.00000643, -0.00000527, -0.00001910, -0.00001902, -0.00000735, -0.00000325, -0.00001771, -0.00000933, 0.00000511, -0.00002244, -0.00004550, -0.00003734, -0.00001465, -0.00004269, -0.00015864, -0.00017277, -0.00000030, 0.00018511, 0.00023898, 0.00011820, 0.00004339, 0.00010620, 0.00008068, 0.00007954, 0.00009386, 0.00007369, 0.00032682, 0.00028244, -0.00026915, -0.00026455, 0.00001094, -0.00010773, -0.00006258, 0.00003437, -0.00003853, -0.00001085, -0.00003880, -0.00001515, 0.00003328, -0.00007164, -0.00001031, 0.00004483, -0.00002493, -0.00000180, -0.00004159, 0.00002168, 0.00004736, -0.00013651, 0.00004635, 0.00012974, -0.00017533, 0.00000043, 0.00015283, -0.00012404, -0.00003599, 0.00010388, -0.00004043, -0.00002728, 0.00001701, -0.00000344, 0.00000678, -0.00003273, -0.00000286, 0.00003826, -0.00002067, -0.00000622, 0.00002864, -0.00001455, -0.00001401, 0.00000651, -0.00000834, -0.00000084, 0.00000357, -0.00001015, 0.00000092, 0.00000184, -0.00001535, -0.00000553, 0.00000643, 0.00000074, -0.00000322, -0.00000224, 0.00000240, -0.00000512, -0.00001477, -0.00000231, 0.00000370, -0.00000225, 0.00000130, 0.00000192, -0.00000261, -0.00000414, -0.00000375, -0.00000209, -0.00000095, 0.00000152, 0.00000129, -0.00000024, -0.00000137, -0.00000945, -0.00000928, -0.00000007, -0.00000035, -0.00000139, -0.00000444, -0.00000883, -0.00000123, 0.00000014, -0.00000621, -0.00000322, 0.00000011, 0.00000022, -0.00000271, -0.00000540, -0.00000211, },
		{ -0.00000368, 0.00000074, -0.00000437, 0.00000376, 0.00001061, -0.00000534, -0.00001182, -0.00000529, -0.00000551, -0.00000550, -0.00000048, 0.00000539, -0.00000211, -0.00001868, -0.00001851, -0.00000769, -0.00000310, -0.00001175, -0.00001528, 0.00000219, -0.00001208, -0.00004673, -0.00003350, -0.00001481, -0.00004654, -0.00013196, -0.00016423, -0.00001288, 0.00016749, 0.00020435, 0.00013189, 0.00004134, 0.00007400, 0.00010449, 0.00003812, 0.00008727, 0.00009254, 0.00011793, 0.00034132, 0.00011285, -0.00031630, -0.00015068, -0.00001914, -0.00011103, 0.00001201, 0.00000447, -0.00006063, -0.00000286, -0.00002573, 0.00000312, -0.00001553, -0.00005542, 0.00004676, 0.00000239, -0.00004472, 0.00001257, -0.00001445, 0.00001915, -0.00003798, -0.00005248, 0.00012590, -0.00004873, -0.00013247, 0.00015650, -0.00000554, -0.00015203, 0.00011201, 0.00003653, -0.00009694, 0.00003409, 0.00001849, -0.00002356, -0.00000546, -0.00002046, 0.00002651, 0.00001318, -0.00003337, 0.00002207, 0.00001893, -0.00003188, -0.00000358, 0.00001073, -0.00000903, -0.00000142, -0.00000097, -0.00000526, 0.00000351, -0.00000812, -0.00001443, 0.00000427, 0.00000549, -0.00000396, -0.00000283, 0.00000110, 0.00000058, -0.00001175, -0.00001131, 0.00000473, 0.00000143, -0.00000421, 0.00000256, 0.00000137, -0.00000416, -0.00000511, -0.00000207, -0.00000012, -0.00000054, 0.00000143, 0.00000093, -0.00000045, -0.00000484, -0.00001222, -0.00000536, 0.00000233, -0.00000127, -0.00000218, -0.00000663, -0.00000747, 0.00000117, -0.00000173, -0.00000597, -0.00000112, 0.00000063, -0.00000005, -0.00000472, -0.00000718, -0.00000394, -0.00000646, },
		{ -0.00000149, -0.00000380, 0.00000773, 0.00000593, -0.00001070, -0.00001018, -0.00000481, -0.00000529, -0.00000092, 0.00000381, -0.00000082, -0.00001571, -0.00001957, -0.00000742, -0.00000349, -0.00000843, -0.00001270, -0.00000649, -0.00000702, -0.00003784, -0.00003826, -0.00001026, -0.00004576, -0.00012162, -0.00014393, -0.00002907, 0.00015078, 0.00018771, 0.00011702, 0.00006129, 0.00004896, 0.00009264, 0.00005706, 0.00003408, 0.00011862, 0.00007315, 0.00015551, 0.00031922, -0.00003539, -0.00030760, -0.00008084, -0.00004302, -0.00006299, 0.00002874, -0.00005031, -0.00003838, 0.00001380, -0.00003191, -0.00000598, -0.00002737, -0.00000620, 0.00003750, -0.00004745, -0.00000922, 0.00002592, -0.00003062, -0.00000508, -0.00002970, 0.00003348, 0.00004841, -0.00012693, 0.00003246, 0.00011647, -0.00015315, -0.00000232, 0.00014206, -0.00010203, -0.00003525, 0.00008539, -0.00004141, -0.00002582, 0.00001251, -0.00000583, 0.00002340, -0.00001380, -0.00001334, 0.00003651, -0.00001188, -0.00003083, 0.00001655, 0.00000476, -0.00001523, 0.00000131, 0.00000085, -0.00000356, -0.00000209, -0.00001212, -0.00000524, 0.00000758, -0.00000103, -0.00000411, 0.00000067, 0.00000060, -0.00000481, -0.00001300, -0.00000321, 0.00000612, -0.00000402, -0.00000273, 0.00000459, -0.00000124, -0.00000606, -0.00000339, 0.00000077, -0.00000040, -0.00000100, 0.00000213, 0.00000040, -0.00000264, -0.00000869, -0.00001098, -0.00000033, 0.00000170, -0.00000276, -0.00000262, -0.00000808, -0.00000511, 0.00000236, -0.00000352, -0.00000461, 0.00000041, 0.00000067, -0.00000165, -0.00000831, -0.00000814, -0.00000284, -0.00000367, -0.00000078, },
		{ -0.00000106, 0.00000907, -0.00000039, -0.00001316, -0.00000824, -0.00000406, -0.00000135, 0.00000307, -0.00000115, -0.00001289, -0.00001888, -0.00000901, -0.00000233, -0.00000797, -0.00000878, -0.00000850, -0.00001213, -0.00002704, -0.00003693, -0.00001477, -0.00003556, -0.00011641, -0.00013193, -0.00003114, 0.00012405, 0.00018240, 0.00010478, 0.00005918, 0.00005817, 0.00006023, 0.00006935, 0.00002973, 0.00007347, 0.00011426, 0.00004430, 0.00021498, 0.00026070, -0.00017013, -0.00026196, -0.00002501, -0.00004519, -0.00003099, -0.00000355, -0.00006252, 0.00000529, -0.00000538, -0.00004318, 0.00000448, -0.00001433, 0.00000841, -0.00000411, -0.00004013, 0.00003131, -0.00001033, -0.00004226, 0.00001477, -0.00000046, 0.00002872, -0.00003526, -0.00005909, 0.00010532, -0.00003717, -0.00011558, 0.00013876, 0.00000388, -0.00013121, 0.00009295, 0.00002446, -0.00009120, 0.00003160, 0.00001568, -0.00001715, 0.00001573, -0.00001487, 0.00000919, 0.00001972, -0.00003191, -0.00000574, 0.00002220, -0.00001305, -0.00001168, 0.00000954, -0.00000268, -0.00000771, -0.00000379, -0.00000784, 0.00000011, 0.00000319, -0.00000330, 0.00000010, -0.00000005, -0.00000312, -0.00000652, -0.00000894, 0.00000157, 0.00000177, -0.00000612, 0.00000158, 0.00000331, -0.00000477, -0.00000483, 0.00000001, 0.00000109, -0.00000217, 0.00000023, 0.00000252, -0.00000215, -0.00000532, -0.00000976, -0.00000703, 0.00000197, -0.00000094, -0.00000252, -0.00000318, -0.00000928, -0.00000198, 0.00000261, -0.00000488, -0.00000279, 0.00000124, -0.00000036, -0.00000526, -0.00001125, -0.00000564, 0.00000029, -0.00000181, -0.00000103, -0.00000375, },
		{ 0.00000750, -0.00000645, -0.00001346, -0.00000551, -0.00000051, 0.00000233, -0.00000117, -0.00001182, -0.00001692, -0.00001037, -0.00000223, -0.00000642, -0.00000841, -0.00000530, -0.00001564, -0.00002626, -0.00002788, -0.00001976, -0.00003100, -0.00010171, -0.00013057, -0.00003061, 0.00010546, 0.00016657, 0.00010958, 0.00004490, 0.00006245, 0.00005670, 0.00004564, 0.00005152, 0.00004206, 0.00009820, 0.00007001, 0.00005848, 0.00027531, 0.00014082, -0.00026432, -0.00016947, 0.00001126, -0.00005783, -0.00002619, -0.00002558, -0.00003502, 0.00001429, -0.00004018, -0.00002035, 0.00001984, -0.00002064, -0.00000133, -0.00001673, -0.00000576, 0.00002004, -0.00004747, -0.00000422, 0.00003147, -0.00001494, 0.00000002, -0.00003624, 0.00001624, 0.00004353, -0.00010387, 0.00002941, 0.00010580, -0.00012779, -0.00000308, 0.00011730, -0.00009902, -0.00003257, 0.00008126, -0.00003640, -0.00001421, 0.00002806, -0.00001449, 0.00000975, -0.00000031, -0.00002287, 0.00001390, 0.00000511, -0.00002012, 0.00000395, 0.00000784, -0.00001282, -0.00000529, 0.00000074, -0.00000677, -0.00000123, 0.00000145, -0.00000011, -0.00000004, -0.00000442, -0.00000254, -0.00000449, -0.00000672, 0.00000140, -0.00000163, -0.00000387, 0.00000336, -0.00000051, -0.00000490, -0.00000141, 0.00000107, -0.00000082, -0.00000212, 0.00000210, 0.00000009, -0.00000496, -0.00000560, -0.00000874, -0.00000409, 0.00000137, -0.00000223, -0.00000102, -0.00000546, -0.00000925, 0.00000199, 0.00000142, -0.00000581, -0.00000076, 0.00000084, -0.00000327, -0.00000923, -0.00001043, -0.00000041, 0.00000198, -0.00000240, -0.00000249, -0.00000437, 0.00000215, },
		{ -0.00001117, -0.00001103, -0.00000057, 0.00000261, -0.00000123, -0.00001106, -0.00001594, -0.00001012, -0.00000319, -0.00000456, -0.00000848, -0.00000440, -0.00001222, -0.00002949, -0.00002444, -0.00001570, -0.00003300, -0.00008653, -0.00012308, -0.00004058, 0.00009649, 0.00014901, 0.00011055, 0.00004593, 0.00004632, 0.00006653, 0.00003224, 0.00004175, 0.00005618, 0.00005475, 0.00008726, 0.00003747, 0.00011885, 0.00027893, -0.00001159, -0.00027304, -0.00006475, -0.00000195, -0.00007528, -0.00001516, -0.00002512, -0.00001642, -0.00000793, -0.00004162, 0.00001676, 0.00000272, -0.00003360, 0.00000485, -0.00000941, 0.00000166, -0.00001256, -0.00003269, 0.00003567, -0.00000024, -0.00003029, 0.00000946, -0.00001427, 0.00001699, -0.00002374, -0.00004443, 0.00009262, -0.00002879, -0.00009701, 0.00011786, -0.00000686, -0.00012219, 0.00008802, 0.00002850, -0.00008135, 0.00003941, 0.00002215, -0.00003279, 0.00001219, -0.00000118, -0.00000967, 0.00000974, -0.00000924, -0.00000767, 0.00001058, -0.00000648, -0.00001337, 0.00000529, -0.00000112, -0.00001131, 0.00000095, 0.00000445, -0.00000077, -0.00000489, -0.00000357, 0.00000197, -0.00000610, -0.00000741, 0.00000190, -0.00000222, -0.00000229, 0.00000205, -0.00000181, -0.00000212, -0.00000023, -0.00000016, -0.00000145, -0.00000065, 0.00000129, -0.00000354, -0.00000455, -0.00000460, -0.00000876, -0.00000263, 0.00000082, -0.00000148, -0.00000098, -0.00000869, -0.00000629, 0.00000519, -0.00000161, -0.00000549, 0.00000081, -0.00000173, -0.00000692, -0.00001031, -0.00000566, 0.00000369, 0.00000046, -0.00000375, -0.00000337, -0.00000368, 0.00000440, 0.00000368, },
		{ -0.00000525, 0.00000347, -0.00000079, -0.00001039, -0.00001556, -0.00000992, -0.00000325, -0.00000394, -0.00000730, -0.00000581, -0.00000835, -0.00002672, -0.00002864, -0.00001145, -0.00002942, -0.00007992, -0.00010873, -0.00004993, 0.00008093, 0.00014155, 0.00010288, 0.00005326, 0.00003625, 0.00005746, 0.00004340, 0.00002079, 0.00005927, 0.00005295, 0.00005164, 0.00007046, 0.00003968, 0.00017710, 0.00021543, -0.00012859, -0.00020973, -0.00000811, -0.00004724, -0.00006315, 0.00000219, -0.00003018, -0.00001567, -0.00001884, -0.00001475, 0.00002220, -0.00002985, -0.00001888, 0.00001568, -0.00001739, -0.00000661, -0.00001593, 0.00000313, 0.00002508, -0.00003697, -0.00000295, 0.00001932, -0.00002580, -0.00000112, -0.00002087, 0.00001902, 0.00003697, -0.00008704, 0.00002805, 0.00008512, -0.00012243, -0.00000171, 0.00011465, -0.00008474, -0.00003007, 0.00008375, -0.00003492, -0.00003055, 0.00003461, -0.00000619, -0.00001235, 0.00000341, -0.00000468, 0.00000173, -0.00000019, -0.00001304, -0.00000050, 0.00000774, -0.00001176, -0.00000933, 0.00000839, 0.00000233, -0.00000640, -0.00000462, 0.00000300, 0.00000127, -0.00001172, -0.00000468, 0.00000390, -0.00000427, -0.00000222, 0.00000233, -0.00000030, -0.00000134, -0.00000138, -0.00000002, -0.00000097, -0.00000109, -0.00000135, -0.00000412, -0.00000203, -0.00000615, -0.00000946, -0.00000022, 0.00000107, -0.00000096, -0.00000300, -0.00001009, -0.00000127, 0.00000534, -0.00000497, -0.00000364, 0.00000016, -0.00000573, -0.00000818, -0.00000748, -0.00000063, 0.00000390, -0.00000231, -0.00000396, -0.00000389, -0.00000270, 0.00000506, -0.00000123, -0.00001275, },
		{ 0.00000107, -0.00000973, -0.00001530, -0.00001027, -0.00000282, -0.00000339, -0.00000652, -0.00000639, -0.00000783, -0.00002017, -0.00003002, -0.00001563, -0.00002143, -0.00007311, -0.00009959, -0.00004925, 0.00006007, 0.00013292, 0.00010106, 0.00005172, 0.00003948, 0.00004201, 0.00004639, 0.00002399, 0.00003700, 0.00006549, 0.00003351, 0.00005482, 0.00006168, 0.00005664, 0.00020651, 0.00012549, -0.00018334, -0.00013573, -0.00000896, -0.00007253, -0.00002649, -0.00000258, -0.00004046, -0.00000744, -0.00001249, 0.00000028, -0.00000039, -0.00003605, 0.00000869, 0.00000109, -0.00002776, 0.00000319, -0.00000290, 0.00000962, -0.00000841, -0.00002866, 0.00002856, -0.00000962, -0.00003515, 0.00001364, -0.00000160, 0.00001532, -0.00002234, -0.00003427, 0.00008002, -0.00003529, -0.00008968, 0.00011261, 0.00000349, -0.00011140, 0.00007808, 0.00003430, -0.00008365, 0.00002495, 0.00003743, -0.00003247, -0.00000812, 0.00001171, -0.00000157, -0.00000559, -0.00000710, -0.00000319, 0.00000659, -0.00000317, -0.00001577, 0.00000186, 0.00000897, -0.00000620, -0.00000630, 0.00000235, 0.00000499, -0.00000623, -0.00001286, 0.00000191, 0.00000174, -0.00000753, 0.00000084, 0.00000444, -0.00000078, -0.00000296, -0.00000068, 0.00000113, -0.00000278, -0.00000279, -0.00000167, -0.00000191, -0.00000209, -0.00000990, -0.00000735, 0.00000296, 0.00000018, -0.00000130, -0.00000512, -0.00000893, 0.00000294, 0.00000269, -0.00000667, -0.00000215, -0.00000314, -0.00000768, -0.00000555, -0.00000395, 0.00000123, 0.00000181, -0.00000350, -0.00000364, -0.00000512, -0.00000164, 0.00000449, -0.00000511, -0.00001003, 0.00000058, },
		{ -0.00001545, -0.00001078, -0.00000270, -0.00000229, -0.00000629, -0.00000642, -0.00000810, -0.00001607, -0.00002523, -0.00002163, -0.00002006, -0.00006024, -0.00009405, -0.00004757, 0.00004701, 0.00011530, 0.00010320, 0.00005030, 0.00003914, 0.00003961, 0.00003415, 0.00003374, 0.00002733, 0.00005193, 0.00004735, 0.00002519, 0.00006826, 0.00004602, 0.00008033, 0.00021612, 0.00003543, -0.00019907, -0.00008462, -0.00002658, -0.00006240, -0.00000423, -0.00002595, -0.00003077, 0.00000692, -0.00001384, -0.00000273, -0.00001477, -0.00001856, 0.00001251, -0.00002315, -0.00001155, 0.00001724, -0.00000830, -0.00000271, -0.00001656, 0.00000093, 0.00001873, -0.00004083, -0.00000621, 0.00002616, -0.00001516, -0.00000428, -0.00001827, 0.00002088, 0.00002767, -0.00008380, 0.00002825, 0.00008581, -0.00010522, -0.00000881, 0.00010425, -0.00007097, -0.00004127, 0.00007649, -0.00001302, -0.00004127, 0.00001926, 0.00001108, -0.00001434, -0.00000898, 0.00000237, 0.00000316, -0.00000126, -0.00001020, -0.00000735, 0.00000761, -0.00000023, -0.00000854, 0.00000177, 0.00000515, -0.00000075, -0.00001115, -0.00000725, 0.00000458, -0.00000450, -0.00000584, 0.00000637, 0.00000323, -0.00000367, -0.00000242, 0.00000176, -0.00000088, -0.00000588, -0.00000154, 0.00000023, -0.00000149, -0.00000550, -0.00001129, -0.00000225, 0.00000368, -0.00000161, -0.00000095, -0.00000683, -0.00000674, 0.00000523, -0.00000077, -0.00000717, -0.00000280, -0.00000592, -0.00000513, -0.00000229, -0.00000337, 0.00000100, 0.00000075, -0.00000341, -0.00000489, -0.00000668, 0.00000024, 0.00000370, -0.00000636, -0.00000550, 0.00000087, -0.00000746, },
		{ -0.00000270, -0.00000115, -0.00000580, -0.00000694, -0.00000759, -0.00001430, -0.00002012, -0.00002162, -0.00002497, -0.00005009, -0.00008292, -0.00005029, 0.00003990, 0.00009936, 0.00009716, 0.00005575, 0.00003428, 0.00004068, 0.00002827, 0.00002767, 0.00003528, 0.00003420, 0.00004774, 0.00002924, 0.00003804, 0.00006752, 0.00002720, 0.00012363, 0.00020304, -0.00005715, -0.00019141, -0.00004696, -0.00003596, -0.00004415, -0.00000772, -0.00003546, -0.00000491, 0.00000322, -0.00002285, -0.00000134, -0.00001498, -0.00000867, -0.00000261, -0.00002240, 0.00001512, 0.00000546, -0.00002108, -0.00000029, -0.00000700, 0.00000796, -0.00001005, -0.00003205, 0.00002565, 0.00000146, -0.00002824, 0.00000720, 0.00000083, 0.00001609, -0.00002478, -0.00003132, 0.00007802, -0.00002437, -0.00008315, 0.00009190, 0.00000819, -0.00009926, 0.00005763, 0.00004301, -0.00006382, 0.00000133, 0.00003197, -0.00001477, -0.00001821, 0.00000558, 0.00000764, -0.00000603, -0.00000616, -0.00000463, -0.00000196, 0.00000215, -0.00000503, -0.00000074, 0.00000566, -0.00000006, -0.00000489, -0.00000965, -0.00000288, 0.00000121, -0.00000700, 0.00000106, 0.00000765, -0.00000149, -0.00000385, 0.00000052, 0.00000103, -0.00000546, -0.00000511, 0.00000222, -0.00000024, -0.00000384, -0.00000797, -0.00000875, 0.00000170, 0.00000157, -0.00000146, -0.00000054, -0.00000934, -0.00000366, 0.00000620, -0.00000476, -0.00000793, -0.00000375, -0.00000473, -0.00000092, -0.00000281, -0.00000455, 0.00000221, 0.00000079, -0.00000479, -0.00000776, -0.00000649, 0.00000343, 0.00000321, -0.00000608, -0.00000409, -0.00000529, -0.00001437, -0.00001202, },
		{ -0.00000505, -0.00000771, -0.00000718, -0.00001226, -0.00001772, -0.00001830, -0.00002698, -0.00004827, -0.00006962, -0.00004954, 0.00002979, 0.00009072, 0.00008672, 0.00005772, 0.00003547, 0.00003557, 0.00003114, 0.00001883, 0.00003466, 0.00003603, 0.00003070, 0.00003918, 0.00002668, 0.00005001, 0.00004669, 0.00003366, 0.00017639, 0.00014627, -0.00014045, -0.00015347, -0.00001825, -0.00004491, -0.00003358, -0.00001387, -0.00002328, 0.00000735, -0.00001622, -0.00001804, 0.00000460, -0.00002002, -0.00000909, -0.00000684, -0.00000283, 0.00001727, -0.00001901, -0.00001249, 0.00000885, -0.00000963, -0.00000042, -0.00001732, -0.00000499, 0.00001992, -0.00002758, -0.00000728, 0.00001892, -0.00000886, -0.00000168, -0.00001841, 0.00002141, 0.00003014, -0.00007293, 0.00001675, 0.00007050, -0.00008589, -0.00001142, 0.00008681, -0.00004553, -0.00003751, 0.00004671, -0.00000381, -0.00002920, 0.00000439, 0.00001331, -0.00000561, -0.00001120, 0.00000276, -0.00000286, -0.00000720, 0.00000004, 0.00000073, 0.00000398, 0.00000030, -0.00000307, -0.00000348, -0.00000912, -0.00000291, -0.00000077, -0.00000347, 0.00000537, 0.00000348, -0.00000345, -0.00000084, 0.00000060, -0.00000307, -0.00000670, -0.00000027, 0.00000290, -0.00000372, -0.00000487, -0.00000777, -0.00000566, 0.00000269, 0.00000012, 0.00000064, -0.00000294, -0.00001142, 0.00000130, 0.00000461, -0.00000980, -0.00000737, -0.00000189, -0.00000127, -0.00000031, -0.00000628, -0.00000306, 0.00000488, -0.00000119, -0.00000826, -0.00000955, -0.00000323, 0.00000708, 0.00000200, -0.00000751, -0.00000770, -0.00001179, -0.00001293, -0.00000356, 0.00000016, },
		{ -0.00000739, -0.00000972, -0.00001593, -0.00001684, -0.00002417, -0.00004792, -0.00006303, -0.00004287, 0.00001995, 0.00008154, 0.00008169, 0.00005233, 0.00003858, 0.00003257, 0.00003002, 0.00001999, 0.00002411, 0.00004005, 0.00002592, 0.00002848, 0.00003708, 0.00002777, 0.00004809, 0.00003053, 0.00007277, 0.00020039, 0.00005052, -0.00018227, -0.00009407, -0.00001261, -0.00005481, -0.00002067, -0.00001157, -0.00001155, -0.00000121, -0.00002515, -0.00000150, -0.00000325, -0.00002729, -0.00000075, 0.00000053, 0.00000488, -0.00000107, -0.00002314, 0.00000730, -0.00000038, -0.00001678, 0.00000284, -0.00001055, 0.00000212, -0.00000339, -0.00002284, 0.00001763, 0.00000011, -0.00001755, 0.00000908, 0.00000062, 0.00001669, -0.00002029, -0.00003093, 0.00005902, -0.00002217, -0.00006461, 0.00007515, 0.00000796, -0.00007121, 0.00003922, 0.00002205, -0.00004312, 0.00000164, 0.00001917, -0.00000459, -0.00001394, 0.00000146, 0.00000655, -0.00001424, -0.00000693, 0.00000848, 0.00000330, -0.00000101, -0.00000298, 0.00000070, -0.00000501, -0.00001211, -0.00000045, 0.00000127, -0.00000099, 0.00000434, 0.00000069, -0.00000129, -0.00000054, -0.00000279, -0.00000463, -0.00000336, 0.00000246, -0.00000088, -0.00000516, -0.00000335, -0.00000796, -0.00000369, 0.00000326, 0.00000070, 0.00000107, -0.00000784, -0.00000977, 0.00000591, -0.00000138, -0.00001294, -0.00000274, 0.00000123, -0.00000057, -0.00000356, -0.00000743, 0.00000171, 0.00000495, -0.00000611, -0.00001102, -0.00000819, 0.00000213, 0.00000921, -0.00000202, -0.00001220, -0.00001217, -0.00001166, -0.00000519, 0.00000097, -0.00000372, -0.00000848, },
		{ -0.00001317, -0.00001706, -0.00002169, -0.00004311, -0.00006115, -0.00003871, 0.00001668, 0.00006974, 0.00007894, 0.00004887, 0.00003496, 0.00003454, 0.00002664, 0.00002206, 0.00002012, 0.00003201, 0.00003237, 0.00001796, 0.00003448, 0.00003167, 0.00002496, 0.00004478, 0.00003487, 0.00011722, 0.00017539, -0.00004505, -0.00017196, -0.00004521, -0.00002892, -0.00005002, -0.00000402, -0.00001275, -0.00001041, -0.00001066, -0.00001742, 0.00000394, -0.00002047, -0.00001796, 0.00001264, -0.00000156, -0.00000170, -0.00001078, -0.00000888, 0.00000926, -0.00001800, -0.00000595, 0.00000941, -0.00001528, -0.00000357, -0.00000869, -0.00000281, 0.00001201, -0.00002006, 0.00000352, 0.00001881, -0.00000888, -0.00000216, -0.00001773, 0.00001299, 0.00001780, -0.00005826, 0.00002271, 0.00005330, -0.00006892, 0.00000160, 0.00005665, -0.00004730, -0.00001732, 0.00003643, -0.00000717, -0.00001703, 0.00000142, 0.00000953, -0.00000790, -0.00001940, 0.00000798, 0.00001150, -0.00000481, -0.00000349, 0.00000181, 0.00000110, -0.00001251, -0.00000997, 0.00000624, 0.00000028, -0.00000218, 0.00000422, 0.00000163, -0.00000134, -0.00000406, -0.00000348, -0.00000218, -0.00000125, 0.00000062, -0.00000361, -0.00000263, -0.00000348, -0.00000941, -0.00000042, 0.00000435, 0.00000062, -0.00000144, -0.00001123, -0.00000481, 0.00000582, -0.00000878, -0.00000966, 0.00000353, 0.00000061, -0.00000316, -0.00000565, -0.00000436, 0.00000514, 0.00000059, -0.00001068, -0.00001074, -0.00000425, 0.00000739, 0.00000746, -0.00000983, -0.00001694, -0.00001114, -0.00000509, 0.00000029, -0.00000211, -0.00000864, -0.00000825, -0.00000756, },
		{ -0.00002183, -0.00003672, -0.00005635, -0.00004029, 0.00001510, 0.00006213, 0.00007272, 0.00004983, 0.00002948, 0.00003266, 0.00002813, 0.00001980, 0.00002176, 0.00002406, 0.00003041, 0.00002206, 0.00002135, 0.00003633, 0.00001927, 0.00002921, 0.00004811, 0.00004880, 0.00014269, 0.00011822, -0.00010877, -0.00013398, -0.00002465, -0.00004324, -0.00002882, 0.00000091, -0.00002034, -0.00000863, -0.00001205, -0.00001046, -0.00000493, -0.00002413, 0.00000318, 0.00001040, -0.00001199, -0.00000246, -0.00000775, -0.00000172, -0.00000407, -0.00001758, 0.00001211, -0.00000229, -0.00002317, 0.00000193, -0.00000298, 0.00000024, -0.00000525, -0.00000958, 0.00002514, -0.00000123, -0.00001947, 0.00000484, -0.00000212, 0.00000752, -0.00002217, -0.00001512, 0.00005420, -0.00002942, -0.00004375, 0.00006885, -0.00001663, -0.00005786, 0.00004908, 0.00001046, -0.00003700, 0.00000751, 0.00001188, -0.00000534, -0.00001844, -0.00000458, 0.00001832, -0.00000081, -0.00000898, 0.00000365, 0.00000340, -0.00000672, -0.00001526, 0.00000002, 0.00000778, -0.00000564, -0.00000019, 0.00000708, 0.00000007, -0.00000536, -0.00000465, 0.00000016, -0.00000166, -0.00000279, -0.00000080, -0.00000229, -0.00000083, -0.00000692, -0.00000826, 0.00000430, 0.00000318, -0.00000091, -0.00000411, -0.00001173, -0.00000096, 0.00000149, -0.00001111, -0.00000156, 0.00000482, -0.00000385, -0.00000403, -0.00000439, -0.00000115, 0.00000424, -0.00000493, -0.00001207, -0.00000817, 0.00000067, 0.00001008, 0.00000020, -0.00001835, -0.00001561, -0.00000381, 0.00000014, -0.00000192, -0.00000778, -0.00000910, -0.00000639, -0.00000990, -0.00001665, },
		{ -0.00004798, -0.00004110, 0.00000853, 0.00005860, 0.00006732, 0.00004896, 0.00002871, 0.00002605, 0.00002934, 0.00002017, 0.00002005, 0.00002355, 0.00002243, 0.00002526, 0.00001935, 0.00002601, 0.00002707, 0.00001447, 0.00004420, 0.00004731, 0.00006175, 0.00014933, 0.00005226, -0.00013948, -0.00009410, -0.00002103, -0.00004147, -0.00001063, -0.00000910, -0.00002098, -0.00000316, -0.00001511, -0.00001027, -0.00001006, -0.00001175, 0.00001215, -0.00000567, -0.00001142, 0.00000360, -0.00000907, -0.00000485, -0.00000934, -0.00000280, 0.00001180, -0.00002179, -0.00001341, 0.00001057, -0.00000836, -0.00000538, -0.00000357, 0.00001162, 0.00001539, -0.00002381, -0.00000309, 0.00001302, -0.00001050, -0.00000856, -0.00001306, 0.00002476, 0.00000916, -0.00005696, 0.00003665, 0.00003740, -0.00007992, 0.00001831, 0.00005743, -0.00005285, -0.00000988, 0.00003398, -0.00001215, -0.00001615, -0.00000613, 0.00000930, 0.00000915, -0.00001109, 0.00000057, 0.00000703, -0.00000454, -0.00001126, -0.00000814, 0.00000668, 0.00000067, -0.00000724, 0.00000683, 0.00000572, -0.00000577, -0.00000626, -0.00000002, 0.00000127, -0.00000541, -0.00000276, 0.00000102, -0.00000103, -0.00000266, -0.00000930, -0.00000322, 0.00000611, -0.00000036, -0.00000101, -0.00000632, -0.00001200, 0.00000074, -0.00000150, -0.00000733, 0.00000360, -0.00000014, -0.00000588, -0.00000127, -0.00000370, -0.00000077, 0.00000150, -0.00000829, -0.00001103, -0.00000468, 0.00000503, 0.00000747, -0.00001058, -0.00002071, -0.00000649, 0.00000231, -0.00000251, -0.00000765, -0.00000874, -0.00000601, -0.00000719, -0.00001500, -0.00002273, -0.00003294, },
		{ 0.00000014, 0.00005179, 0.00006645, 0.00004661, 0.00002885, 0.00002216, 0.00002493, 0.00002318, 0.00001816, 0.00002265, 0.00002011, 0.00001977, 0.00002389, 0.00001865, 0.00002343, 0.00001873, 0.00002540, 0.00005396, 0.00003682, 0.00007970, 0.00014125, -0.00001437, -0.00014530, -0.00005828, -0.00002070, -0.00003282, -0.00000678, -0.00001819, -0.00001174, -0.00000517, -0.00002076, -0.00000585, -0.00000629, -0.00000245, 0.00000420, -0.00001375, 0.00000032, 0.00000040, -0.00001506, -0.00000275, -0.00000462, 0.00000351, -0.00000345, -0.00002440, 0.00000445, 0.00000291, -0.00001608, 0.00000201, 0.00000743, 0.00001240, -0.00000719, -0.00001792, 0.00001583, -0.00000324, -0.00001923, -0.00000004, 0.00000658, 0.00001493, -0.00003239, -0.00001031, 0.00006125, -0.00004526, -0.00004483, 0.00008290, -0.00001825, -0.00005960, 0.00005126, 0.00000586, -0.00003721, 0.00000566, 0.00000434, 0.00000285, -0.00000142, -0.00000483, 0.00000831, -0.00000218, -0.00000900, -0.00000597, -0.00000231, 0.00000391, -0.00000470, -0.00000060, 0.00000962, -0.00000207, -0.00000789, -0.00000124, 0.00000250, -0.00000338, -0.00000709, 0.00000155, 0.00000168, -0.00000256, -0.00000457, -0.00000794, 0.00000112, 0.00000355, -0.00000176, 0.00000011, -0.00001108, -0.00001170, 0.00000405, -0.00000137, -0.00000371, 0.00000232, -0.00000454, -0.00000235, 0.00000008, -0.00000621, -0.00000068, -0.00000029, -0.00000995, -0.00000886, -0.00000087, 0.00000602, -0.00000060, -0.00001764, -0.00001308, 0.00000273, -0.00000077, -0.00000935, -0.00000815, -0.00000476, -0.00000564, -0.00001302, -0.00002263, -0.00003148, -0.00004057, -0.00003711, },
		{ 0.00006462, 0.00004803, 0.00002690, 0.00002120, 0.00002011, 0.00002193, 0.00002007, 0.00001948, 0.00002056, 0.00001587, 0.00002143, 0.00002153, 0.00001432, 0.00002078, 0.00002148, 0.00003923, 0.00004661, 0.00003058, 0.00010542, 0.00011180, -0.00007477, -0.00012705, -0.00002839, -0.00002416, -0.00002683, -0.00000886, -0.00001751, -0.00000598, -0.00001521, -0.00001684, 0.00000352, -0.00000616, -0.00000327, -0.00000348, -0.00000804, 0.00000497, -0.00001138, -0.00001123, 0.00000392, -0.00000522, -0.00000068, -0.00001186, -0.00001276, 0.00000768, -0.00001147, -0.00000712, 0.00001323, 0.00000481, 0.00000096, -0.00001211, 0.00000026, 0.00000909, -0.00002068, -0.00000430, 0.00001010, 0.00000163, -0.00000760, -0.00002482, 0.00003320, 0.00001281, -0.00007122, 0.00004115, 0.00005020, -0.00008290, 0.00001451, 0.00005689, -0.00005395, -0.00000950, 0.00002874, -0.00001442, -0.00000241, 0.00000455, 0.00000196, 0.00000131, -0.00000905, -0.00000154, -0.00000309, -0.00000471, 0.00000156, -0.00000211, 0.00000450, 0.00000380, -0.00000690, -0.00000274, 0.00000194, -0.00000125, -0.00000707, -0.00000278, 0.00000459, -0.00000146, -0.00000314, -0.00000426, -0.00000629, 0.00000202, 0.00000122, 0.00000024, -0.00000261, -0.00001730, -0.00000593, 0.00000896, -0.00000309, -0.00000379, -0.00000011, -0.00000347, 0.00000139, -0.00000382, -0.00000818, 0.00000158, -0.00000227, -0.00001067, -0.00000560, 0.00000090, 0.00000190, -0.00000804, -0.00001474, -0.00000226, 0.00000181, -0.00000989, -0.00000980, -0.00000227, -0.00000385, -0.00001211, -0.00002166, -0.00003014, -0.00003593, -0.00003139, -0.00000402, 0.00004076, },
		{ 0.00002727, 0.00001877, 0.00001866, 0.00001767, 0.00002040, 0.00001931, 0.00001834, 0.00001665, 0.00001631, 0.00002209, 0.00001416, 0.00001290, 0.00002512, 0.00002819, 0.00004175, 0.00003158, 0.00004186, 0.00012448, 0.00005948, -0.00011320, -0.00009053, -0.00001330, -0.00003156, -0.00002028, -0.00000929, -0.00001491, -0.00000924, -0.00001897, -0.00000292, 0.00000436, -0.00001262, -0.00000300, -0.00000327, -0.00000277, -0.00000174, -0.00001512, 0.00000070, 0.00000098, -0.00001067, -0.00000095, -0.00001111, -0.00000645, -0.00000053, -0.00001104, 0.00000887, 0.00000851, -0.00000429, -0.00000027, -0.00000552, 0.00000300, -0.00000754, -0.00001335, 0.00001175, -0.00000154, -0.00000716, -0.00000816, -0.00000245, 0.00002906, -0.00003351, -0.00002363, 0.00007233, -0.00003422, -0.00005347, 0.00007800, -0.00001549, -0.00005923, 0.00004875, 0.00000179, -0.00003352, 0.00001834, 0.00000737, -0.00000645, -0.00000480, -0.00000160, 0.00000340, -0.00001072, -0.00000421, 0.00000595, -0.00000070, 0.00000197, -0.00000119, -0.00000343, 0.00000114, -0.00000152, -0.00000464, -0.00000482, 0.00000188, 0.00000239, -0.00000356, -0.00000070, -0.00000541, -0.00000633, 0.00000293, 0.00000161, 0.00000040, -0.00001013, -0.00001760, 0.00000529, 0.00000862, -0.00000860, -0.00000293, 0.00000092, -0.00000088, -0.00000004, -0.00000915, -0.00000580, 0.00000364, -0.00000526, -0.00000936, -0.00000349, -0.00000190, -0.00000264, -0.00000748, -0.00000655, -0.00000035, -0.00000781, -0.00001227, -0.00000138, -0.00000003, -0.00001181, -0.00002145, -0.00002791, -0.00003247, -0.00002741, -0.00000431, 0.00003153, 0.00005750, 0.00005100, },
		{ 0.00001700, 0.00001584, 0.00001656, 0.00002009, 0.00001719, 0.00001607, 0.00001571, 0.00001742, 0.00001683, 0.00000906, 0.00002001, 0.00002992, 0.00002866, 0.00003548, 0.00002431, 0.00006522, 0.00012160, -0.00000019, -0.00012064, -0.00005365, -0.00001539, -0.00003398, -0.00001155, -0.00001194, -0.00001544, -0.00001163, -0.00001131, 0.00000565, -0.00000612, -0.00001364, 0.00000229, -0.00000411, -0.00000395, -0.00000665, -0.00000796, 0.00000446, -0.00000991, -0.00000763, 0.00000230, -0.00001389, -0.00000646, -0.00000173, 0.00000014, 0.00001034, -0.00000439, -0.00000119, 0.00000364, -0.00000726, -0.00000336, -0.00000839, 0.00000270, 0.00000433, -0.00001419, 0.00000200, -0.00000452, 0.00000183, 0.00000902, -0.00003400, 0.00002488, 0.00003071, -0.00006806, 0.00002710, 0.00005103, -0.00007744, 0.00001160, 0.00005444, -0.00005427, -0.00000500, 0.00003746, -0.00001528, -0.00001039, 0.00000477, 0.00000456, -0.00000578, -0.00001497, 0.00000652, 0.00000620, -0.00000630, 0.00000078, 0.00000113, -0.00000018, -0.00000220, -0.00000451, -0.00000244, -0.00000175, 0.00000196, -0.00000033, -0.00000127, -0.00000028, -0.00000963, -0.00000401, 0.00000553, 0.00000060, -0.00000387, -0.00001532, -0.00000899, 0.00001297, 0.00000005, -0.00001118, 0.00000244, 0.00000176, -0.00000180, -0.00000429, -0.00001081, -0.00000122, 0.00000322, -0.00000704, -0.00000767, -0.00000612, -0.00000517, 0.00000006, -0.00000200, -0.00000552, -0.00000820, -0.00001200, -0.00000305, 0.00000390, -0.00000950, -0.00002286, -0.00002591, -0.00002829, -0.00002500, -0.00000461, 0.00002665, 0.00004911, 0.00004951, 0.00003106, 0.00001697, },
		{ 0.00001339, 0.00001698, 0.00001845, 0.00001478, 0.00001583, 0.00001524, 0.00001436, 0.00001139, 0.00001321, 0.00002836, 0.00002678, 0.00002516, 0.00002997, 0.00002789, 0.00008681, 0.00009715, -0.00004890, -0.00010579, -0.00003049, -0.00002344, -0.00002740, -0.00000797, -0.00001767, -0.00001287, -0.00000777, -0.00000351, 0.00000164, -0.00001434, -0.00000538, 0.00000294, -0.00000881, -0.00000322, -0.00000510, -0.00000308, -0.00000256, -0.00001327, 0.00000225, -0.00000299, -0.00001844, -0.00000032, 0.00000365, 0.00000318, 0.00000037, -0.00000537, 0.00000784, -0.00000325, -0.00001193, -0.00000042, -0.00000071, 0.00000261, -0.00001089, -0.00000533, 0.00000988, -0.00001536, 0.00000392, 0.00000359, -0.00001777, 0.00003113, -0.00001511, -0.00003236, 0.00006099, -0.00002541, -0.00005178, 0.00007310, -0.00001393, -0.00006017, 0.00005095, 0.00000986, -0.00003588, 0.00001159, 0.00001047, -0.00000536, -0.00001383, -0.00000450, 0.00001350, -0.00000490, -0.00000783, 0.00000719, 0.00000223, -0.00000393, -0.00000535, -0.00000114, -0.00000029, -0.00000262, 0.00000156, 0.00000060, 0.00000048, -0.00000494, -0.00001120, 0.00000226, 0.00000489, -0.00000375, -0.00000708, -0.00001343, 0.00000093, 0.00001067, -0.00000885, -0.00000548, 0.00000686, -0.00000200, -0.00000403, -0.00000692, -0.00000949, 0.00000260, 0.00000221, -0.00000799, -0.00001021, -0.00001013, -0.00000030, 0.00000742, -0.00000420, -0.00001380, -0.00001155, -0.00000337, 0.00000486, -0.00000515, -0.00002351, -0.00002614, -0.00002342, -0.00002212, -0.00000642, 0.00002323, 0.00004408, 0.00004450, 0.00003256, 0.00001911, 0.00001464, 0.00001517, },
		{ 0.00001716, 0.00001616, 0.00001435, 0.00001542, 0.00001171, 0.00001094, 0.00001282, 0.00002128, 0.00002814, 0.00001926, 0.00002485, 0.00002709, 0.00003774, 0.00009972, 0.00006023, -0.00008037, -0.00008277, -0.00002011, -0.00002729, -0.00001961, -0.00001199, -0.00001812, -0.00000510, -0.00000516, -0.00000315, -0.00000534, -0.00001248, 0.00000129, -0.00000383, -0.00000854, 0.00000132, -0.00000572, -0.00000496, -0.00000763, -0.00000663, 0.00000456, -0.00001413, -0.00001244, 0.00000931, 0.00000182, -0.00000224, -0.00000292, 0.00000316, 0.00000528, -0.00001396, -0.00000483, 0.00000568, -0.00000286, -0.00000642, -0.00001009, 0.00000945, -0.00000334, -0.00002126, 0.00001934, -0.00000233, -0.00001288, 0.00002125, -0.00002353, 0.00000896, 0.00002959, -0.00005763, 0.00002239, 0.00005007, -0.00007503, 0.00000517, 0.00005886, -0.00004447, -0.00001084, 0.00003281, -0.00000917, -0.00001285, -0.00000580, 0.00000582, 0.00000501, -0.00001340, 0.00000278, 0.00001047, -0.00000477, -0.00000691, -0.00000150, 0.00000225, -0.00000337, -0.00000259, 0.00000439, 0.00000085, -0.00000229, -0.00000913, -0.00000660, 0.00000597, -0.00000110, -0.00000538, -0.00000550, -0.00000974, 0.00000475, 0.00000384, -0.00000971, 0.00000283, 0.00000404, -0.00000673, -0.00000380, -0.00000885, -0.00000701, 0.00000673, 0.00000046, -0.00001272, -0.00001494, -0.00000570, 0.00001109, 0.00000549, -0.00001687, -0.00001656, -0.00000148, 0.00000517, -0.00000265, -0.00002120, -0.00002760, -0.00002021, -0.00001753, -0.00000836, 0.00001855, 0.00004145, 0.00004117, 0.00002960, 0.00002134, 0.00001557, 0.00001338, 0.00001227, 0.00001255, },
		{ 0.00001489, 0.00001404, 0.00001199, 0.00000836, 0.00001281, 0.00001816, 0.00002363, 0.00002082, 0.00001616, 0.00002706, 0.00002458, 0.00005228, 0.00010181, 0.00001721, -0.00009534, -0.00005836, -0.00001657, -0.00002790, -0.00001664, -0.00001504, -0.00001020, -0.00000086, -0.00000831, -0.00000443, -0.00000733, -0.00000763, 0.00000011, -0.00000818, -0.00000150, 0.00000074, -0.00001045, -0.00000520, -0.00000633, -0.00000181, -0.00000229, -0.00001660, 0.00000123, 0.00000880, -0.00000610, -0.00000255, 0.00000154, 0.00000497, -0.00000539, -0.00001200, 0.00000644, 0.00000024, -0.00000933, -0.00000626, -0.00000019, 0.00000744, -0.00002093, -0.00000412, 0.00002731, -0.00002208, -0.00000449, 0.00002039, -0.00001814, 0.00001721, -0.00000689, -0.00002769, 0.00005480, -0.00002136, -0.00005617, 0.00006492, -0.00000132, -0.00005311, 0.00004079, 0.00001087, -0.00002936, 0.00000321, 0.00000112, 0.00000078, -0.00000489, -0.00000664, 0.00001257, 0.00000157, -0.00001024, -0.00000172, 0.00000262, -0.00000074, -0.00000585, 0.00000253, 0.00000521, -0.00000309, -0.00000483, -0.00000827, -0.00000175, 0.00000281, -0.00000508, -0.00000053, -0.00000515, -0.00000976, 0.00000576, 0.00000061, -0.00000536, 0.00000526, -0.00000278, -0.00000697, -0.00000360, -0.00001129, -0.00000106, 0.00001002, -0.00000693, -0.00001979, -0.00001147, 0.00000749, 0.00001359, -0.00001033, -0.00002350, -0.00000339, 0.00000854, -0.00000253, -0.00001888, -0.00002670, -0.00001963, -0.00001288, -0.00000832, 0.00001275, 0.00003815, 0.00004078, 0.00002649, 0.00001927, 0.00001852, 0.00001371, 0.00001040, 0.00001012, 0.00001301, 0.00001686, },
		{ 0.00001139, 0.00000833, 0.00001006, 0.00001772, 0.00001963, 0.00001929, 0.00001497, 0.00001879, 0.00002664, 0.00002502, 0.00007022, 0.00008922, -0.00002594, -0.00009359, -0.00003630, -0.00001878, -0.00002801, -0.00001442, -0.00001132, -0.00000283, -0.00000506, -0.00001052, -0.00000246, -0.00000796, -0.00000600, -0.00000247, -0.00000529, 0.00000247, -0.00000668, -0.00001078, -0.00000173, -0.00000626, -0.00000240, -0.00000652, -0.00000896, 0.00000743, -0.00000092, -0.00000717, 0.00000252, 0.00000177, -0.00000031, -0.00000853, -0.00000153, 0.00000548, -0.00000990, -0.00000600, -0.00000138, 0.00000135, -0.00000428, -0.00002025, 0.00001806, 0.00000972, -0.00003360, 0.00001933, 0.00001344, -0.00002090, 0.00001487, -0.00001263, 0.00000692, 0.00002723, -0.00005508, 0.00000975, 0.00005007, -0.00005659, 0.00000313, 0.00004963, -0.00003675, -0.00000963, 0.00001988, -0.00001347, -0.00000389, -0.00000070, 0.00000292, 0.00000827, -0.00000834, -0.00000464, 0.00000365, -0.00000025, -0.00000413, -0.00000198, 0.00000642, -0.00000014, -0.00000552, -0.00000310, -0.00000716, -0.00000192, -0.00000078, -0.00000105, 0.00000330, -0.00001156, -0.00000860, 0.00000978, -0.00000057, -0.00000320, 0.00000251, -0.00000667, -0.00000487, -0.00000638, -0.00000988, 0.00000761, 0.00000550, -0.00001868, -0.00001809, 0.00000209, 0.00001371, -0.00000038, -0.00002253, -0.00001061, 0.00001115, -0.00000006, -0.00001987, -0.00002401, -0.00001828, -0.00001081, -0.00000645, 0.00000798, 0.00003301, 0.00004081, 0.00002629, 0.00001524, 0.00001816, 0.00001736, 0.00000945, 0.00000761, 0.00000996, 0.00001416, 0.00001653, 0.00001349, },
		{ 0.00000898, 0.00001503, 0.00001900, 0.00001578, 0.00001457, 0.00001432, 0.00002192, 0.00002348, 0.00003274, 0.00008462, 0.00006090, -0.00005955, -0.00007796, -0.00002266, -0.00002495, -0.00002442, -0.00000903, -0.00000628, -0.00000319, -0.00001065, -0.00000685, -0.00000187, -0.00001048, -0.00000412, -0.00000168, -0.00000203, -0.00000114, -0.00001149, -0.00000491, -0.00000219, -0.00000875, -0.00000082, -0.00000492, -0.00000332, 0.00000332, -0.00000617, -0.00000110, 0.00000223, -0.00000214, -0.00000128, -0.00000428, 0.00000177, -0.00000394, -0.00000935, 0.00000114, -0.00000439, -0.00000161, -0.00000870, -0.00000814, 0.00002169, -0.00001533, -0.00001792, 0.00003592, -0.00001063, -0.00001565, 0.00002154, -0.00000997, 0.00001202, -0.00000618, -0.00003264, 0.00004204, -0.00000854, -0.00004178, 0.00005471, -0.00000284, -0.00004426, 0.00003355, -0.00000108, -0.00002806, 0.00001156, 0.00000358, -0.00000006, 0.00000028, -0.00000626, 0.00000300, 0.00000026, -0.00000345, -0.00000133, 0.00000209, 0.00000320, -0.00000490, -0.00000179, -0.00000295, -0.00001027, -0.00000183, 0.00000199, 0.00000455, -0.00000207, -0.00001810, 0.00000037, 0.00001289, -0.00000518, -0.00000286, 0.00000005, -0.00000787, -0.00000461, -0.00000800, -0.00000281, 0.00001013, -0.00000751, -0.00002204, -0.00000423, 0.00000998, 0.00000299, -0.00001403, -0.00001398, 0.00000752, 0.00000476, -0.00002073, -0.00002432, -0.00001389, -0.00000964, -0.00000590, 0.00000574, 0.00002732, 0.00003915, 0.00002769, 0.00001335, 0.00001439, 0.00001943, 0.00001240, 0.00000494, 0.00000744, 0.00001128, 0.00001522, 0.00001517, 0.00001036, 0.00000846, },
		{ 0.00001711, 0.00001505, 0.00001158, 0.00001328, 0.00001616, 0.00002252, 0.00002204, 0.00004637, 0.00008747, 0.00002383, -0.00007672, -0.00005783, -0.00001957, -0.00002801, -0.00001554, -0.00000472, -0.00000597, -0.00000716, -0.00001063, -0.00000243, -0.00000595, -0.00001036, 0.00000022, -0.00000175, -0.00000298, -0.00000551, -0.00000935, -0.00000148, -0.00000753, -0.00000609, 0.00000376, -0.00000562, -0.00000368, -0.00000041, -0.00000378, 0.00000096, -0.00000289, -0.00000050, 0.00000154, -0.00000399, -0.00000223, -0.00000698, -0.00000170, -0.00000148, -0.00000924, 0.00000196, -0.00000915, -0.00000128, 0.00001214, -0.00002353, 0.00000821, 0.00002567, -0.00002883, 0.00000847, 0.00001981, -0.00001764, 0.00000903, -0.00001289, -0.00000394, 0.00002357, -0.00003535, 0.00001335, 0.00003990, -0.00005047, 0.00000562, 0.00003767, -0.00004371, -0.00000600, 0.00002707, -0.00001110, -0.00000337, 0.00000306, -0.00000062, 0.00000111, -0.00000426, 0.00000069, 0.00000118, -0.00000020, -0.00000084, -0.00000258, 0.00000120, -0.00000952, -0.00001114, 0.00000443, 0.00000598, 0.00000262, -0.00001167, -0.00001484, 0.00001208, 0.00000773, -0.00001036, -0.00000015, -0.00000277, -0.00000922, -0.00000347, -0.00000518, 0.00000181, 0.00000251, -0.00001561, -0.00001081, 0.00000553, 0.00000040, -0.00000881, -0.00000790, 0.00000304, 0.00000509, -0.00001777, -0.00002735, -0.00001066, -0.00000509, -0.00000751, 0.00000379, 0.00002359, 0.00003582, 0.00002836, 0.00001373, 0.00001139, 0.00001741, 0.00001588, 0.00000585, 0.00000395, 0.00000937, 0.00001247, 0.00001505, 0.00001185, 0.00000720, 0.00000910, 0.00001298, },
		{ 0.00001031, 0.00001055, 0.00001405, 0.00001770, 0.00002172, 0.00002485, 0.00006027, 0.00007686, -0.00001217, -0.00007927, -0.00004153, -0.00002082, -0.00002353, -0.00000732, -0.00000574, -0.00000786, -0.00000850, -0.00000750, -0.00000263, -0.00000962, -0.00000517, 0.00000188, -0.00000508, -0.00000384, -0.00000612, -0.00000646, -0.00000386, -0.00000897, 0.00000217, 0.00000292, -0.00001029, -0.00000271, 0.00000017, -0.00000259, -0.00000256, -0.00000259, 0.00000522, -0.00000138, -0.00000736, -0.00000276, -0.00000349, -0.00000058, -0.00000935, -0.00000381, 0.00000524, -0.00001534, 0.00000283, 0.00000501, -0.00001801, 0.00002114, 0.00000368, -0.00002121, 0.00002853, -0.00000209, -0.00001823, 0.00001587, -0.00001265, 0.00000148, 0.00000093, -0.00001545, 0.00003729, -0.00001238, -0.00003500, 0.00004946, -0.00001324, -0.00004773, 0.00003763, 0.00000726, -0.00002640, 0.00001099, 0.00000598, -0.00000430, -0.00000274, -0.00000009, 0.00000447, -0.00000377, -0.00000230, 0.00000138, 0.00000055, -0.00000384, -0.00001535, -0.00000286, 0.00001065, 0.00000301, -0.00000372, -0.00001489, -0.00000407, 0.00001584, -0.00000272, -0.00000953, 0.00000201, -0.00000827, -0.00000705, 0.00000094, -0.00000466, -0.00000147, -0.00000358, -0.00000952, -0.00000065, -0.00000320, -0.00001112, -0.00000064, 0.00000733, 0.00000093, -0.00001606, -0.00002751, -0.00001191, 0.00000124, -0.00000678, -0.00000092, 0.00002148, 0.00003319, 0.00002714, 0.00001442, 0.00001069, 0.00001466, 0.00001551, 0.00000926, 0.00000293, 0.00000622, 0.00001086, 0.00001247, 0.00001309, 0.00000802, 0.00000737, 0.00001295, 0.00001493, 0.00001403, },
		{ 0.00001158, 0.00001485, 0.00001866, 0.00002104, 0.00003143, 0.00006989, 0.00005565, -0.00004149, -0.00007271, -0.00002966, -0.00001979, -0.00001602, -0.00000543, -0.00000876, -0.00000751, -0.00000801, -0.00000604, -0.00000486, -0.00000843, -0.00000040, -0.00000183, -0.00000669, -0.00000209, -0.00000677, -0.00000654, -0.00000548, -0.00000351, 0.00000651, -0.00000491, -0.00001107, 0.00000152, -0.00000119, -0.00000496, -0.00000288, 0.00000338, 0.00000458, -0.00000849, -0.00000539, -0.00000002, -0.00000348, -0.00000532, -0.00000925, 0.00000595, -0.00000344, -0.00001964, 0.00001243, -0.00000156, -0.00001193, 0.00002261, -0.00000791, -0.00000157, 0.00002392, -0.00002245, 0.00000060, 0.00001602, -0.00002154, 0.00000278, 0.00000049, 0.00000640, 0.00001697, -0.00003417, 0.00001517, 0.00003192, -0.00005766, 0.00000250, 0.00004465, -0.00003442, -0.00000849, 0.00002648, -0.00000748, -0.00000726, 0.00000278, 0.00000457, -0.00000137, -0.00000811, 0.00000329, 0.00000345, -0.00000345, -0.00001046, -0.00001111, 0.00000775, 0.00000763, -0.00000294, -0.00000544, -0.00001182, 0.00000408, 0.00001111, -0.00000943, -0.00000499, -0.00000131, -0.00001067, 0.00000132, 0.00000067, -0.00001133, -0.00000229, 0.00000209, -0.00000283, -0.00000846, -0.00001590, -0.00000099, 0.00001747, 0.00000257, -0.00002050, -0.00002590, -0.00001297, 0.00000363, -0.00000174, -0.00000564, 0.00001708, 0.00003298, 0.00002533, 0.00001369, 0.00001099, 0.00001390, 0.00001306, 0.00000983, 0.00000581, 0.00000401, 0.00000836, 0.00001069, 0.00001136, 0.00001025, 0.00000658, 0.00001119, 0.00001534, 0.00001195, 0.00000949, 0.00000869, },
		{ 0.00001554, 0.00001915, 0.00002128, 0.00004073, 0.00007233, 0.00002717, -0.00006153, -0.00005977, -0.00001983, -0.00001706, -0.00001184, -0.00000753, -0.00000884, -0.00000628, -0.00000856, -0.00000537, -0.00000520, -0.00000544, -0.00000054, -0.00000570, -0.00000387, -0.00000209, -0.00000933, -0.00000532, -0.00000286, 0.00000152, 0.00000274, -0.00001101, -0.00000579, 0.00000233, -0.00000551, -0.00000428, 0.00000154, 0.00000548, -0.00000263, -0.00000964, 0.00000046, -0.00000217, -0.00000638, -0.00000599, -0.00000266, 0.00000608, -0.00001725, -0.00001034, 0.00001910, -0.00001397, -0.00000260, 0.00002325, -0.00001079, 0.00000907, 0.00000663, -0.00002077, 0.00001909, -0.00000226, -0.00002311, 0.00001442, 0.00000271, 0.00000595, -0.00000439, -0.00001391, 0.00003496, -0.00001896, -0.00004202, 0.00004786, -0.00000112, -0.00004260, 0.00003108, 0.00001145, -0.00002203, 0.00000568, 0.00000725, -0.00000149, -0.00000805, -0.00000204, 0.00000856, -0.00000292, -0.00000839, -0.00000886, -0.00000198, 0.00000927, -0.00000031, -0.00000225, -0.00000389, -0.00001010, 0.00000754, 0.00000512, -0.00001148, -0.00000338, -0.00000524, -0.00000426, 0.00000622, -0.00001039, -0.00001357, 0.00000945, 0.00000687, -0.00001330, -0.00002169, -0.00000633, 0.00002227, 0.00001357, -0.00002369, -0.00002986, -0.00001054, 0.00000340, 0.00000200, -0.00000549, 0.00001007, 0.00003212, 0.00002584, 0.00001139, 0.00001062, 0.00001453, 0.00001208, 0.00000713, 0.00000751, 0.00000627, 0.00000522, 0.00000877, 0.00000952, 0.00000982, 0.00000853, 0.00000869, 0.00001486, 0.00001270, 0.00000690, 0.00000786, 0.00000959, 0.00001237, },
		{ 0.00001897, 0.00002367, 0.00005076, 0.00006488, -0.00000404, -0.00006933, -0.00004249, -0.00001327, -0.00001620, -0.00001098, -0.00000834, -0.00000692, -0.00000717, -0.00000868, -0.00000343, -0.00000506, -0.00000451, -0.00000266, -0.00000547, -0.00000126, -0.00000586, -0.00000909, -0.00000096, -0.00000087, 0.00000145, -0.00000289, -0.00001077, -0.00000152, -0.00000174, -0.00000616, 0.00000030, 0.00000338, 0.00000163, -0.00000692, -0.00000445, 0.00000085, -0.00000737, -0.00000443, -0.00000335, -0.00000013, -0.00000232, -0.00002130, 0.00000570, 0.00001112, -0.00002220, 0.00001544, 0.00001843, -0.00001445, 0.00001218, -0.00000376, -0.00000698, 0.00001807, -0.00002042, -0.00000617, 0.00001975, -0.00000629, 0.00000269, -0.00000414, 0.00000659, 0.00001405, -0.00004028, 0.00000690, 0.00003612, -0.00004401, -0.00000058, 0.00004069, -0.00002385, -0.00000831, 0.00001962, -0.00000446, -0.00000784, -0.00000339, 0.00000590, 0.00000279, -0.00001014, -0.00000443, -0.00000361, 0.00000067, 0.00000412, -0.00000161, 0.00000304, -0.00000736, -0.00000940, 0.00001098, -0.00000118, -0.00001337, -0.00000227, -0.00000246, 0.00000258, -0.00000213, -0.00001969, 0.00000109, 0.00002010, -0.00000791, -0.00002942, -0.00001262, 0.00001976, 0.00002382, -0.00001671, -0.00003741, -0.00001105, 0.00000665, 0.00000143, -0.00000335, 0.00000620, 0.00002733, 0.00002788, 0.00001027, 0.00000831, 0.00001525, 0.00001276, 0.00000503, 0.00000510, 0.00000959, 0.00000613, 0.00000503, 0.00000840, 0.00000805, 0.00000905, 0.00000931, 0.00001189, 0.00001422, 0.00000670, 0.00000465, 0.00000934, 0.00001021, 0.00001262, 0.00001653, },
		{ 0.00002891, 0.00005753, 0.00004732, -0.00003072, -0.00006421, -0.00002620, -0.00001229, -0.00001672, -0.00000980, -0.00000723, -0.00000649, -0.00000873, -0.00000625, -0.00000230, -0.00000651, -0.00000420, -0.00000312, -0.00000378, -0.00000236, -0.00000841, -0.00000406, 0.00000135, -0.00000246, 0.00000008, -0.00000563, -0.00000878, -0.00000167, -0.00000403, -0.00000204, 0.00000202, 0.00000098, -0.00000120, -0.00000584, -0.00000167, -0.00000403, -0.00000680, 0.00000009, -0.00000507, -0.00000168, -0.00000858, -0.00001543, 0.00001336, -0.00000453, -0.00001369, 0.00003059, 0.00000175, -0.00001326, 0.00001641, -0.00000823, 0.00000168, 0.00000638, -0.00002071, 0.00001168, 0.00000768, -0.00001043, 0.00000988, -0.00000135, 0.00000621, -0.00000705, -0.00002220, 0.00002854, -0.00000804, -0.00003363, 0.00004020, 0.00000415, -0.00003179, 0.00002364, 0.00000657, -0.00001820, 0.00000132, 0.00000312, 0.00000252, -0.00000619, -0.00000614, 0.00000244, -0.00000556, -0.00000125, 0.00000348, 0.00000349, 0.00000238, -0.00001452, -0.00000333, 0.00001211, -0.00001086, -0.00001167, 0.00000409, 0.00000089, -0.00000168, -0.00001342, -0.00001066, 0.00001947, 0.00000764, -0.00003097, -0.00002170, 0.00001533, 0.00002613, -0.00000505, -0.00003738, -0.00001835, 0.00001076, 0.00000250, -0.00000551, 0.00000681, 0.00002266, 0.00002665, 0.00001211, 0.00000546, 0.00001449, 0.00001410, 0.00000474, 0.00000196, 0.00000880, 0.00001047, 0.00000378, 0.00000506, 0.00000758, 0.00000709, 0.00001037, 0.00001114, 0.00001231, 0.00000943, 0.00000238, 0.00000663, 0.00001071, 0.00000948, 0.00001347, 0.00001744, 0.00001859, },
		{ 0.00005748, 0.00002422, -0.00004677, -0.00005115, -0.00001640, -0.00001501, -0.00001553, -0.00000766, -0.00000670, -0.00000740, -0.00000812, -0.00000344, -0.00000414, -0.00000738, -0.00000249, -0.00000325, -0.00000362, -0.00000404, -0.00000631, 0.00000037, -0.00000129, -0.00000354, 0.00000018, -0.00000796, -0.00000815, -0.00000173, -0.00000206, 0.00000023, -0.00000052, 0.00000028, -0.00000114, -0.00000542, -0.00000317, -0.00000574, -0.00000141, -0.00000155, -0.00000897, -0.00000037, -0.00001166, -0.00000917, 0.00001159, -0.00001249, 0.00000494, 0.00002843, -0.00001543, 0.00000007, 0.00001602, -0.00001396, 0.00000655, -0.00000094, -0.00001150, 0.00001403, -0.00000543, 0.00000063, 0.00001210, -0.00000824, 0.00000367, -0.00000843, -0.00000365, 0.00001319, -0.00002700, 0.00000776, 0.00003243, -0.00003127, 0.00000195, 0.00002906, -0.00002371, -0.00000667, 0.00001336, -0.00000475, -0.00000402, -0.00000401, 0.00000302, -0.00000070, -0.00001118, 0.00000276, 0.00000751, 0.00000479, -0.00000514, -0.00001609, 0.00000543, 0.00000474, -0.00001748, -0.00000050, 0.00000852, -0.00000586, -0.00000940, -0.00000962, 0.00000719, 0.00001581, -0.00001912, -0.00002950, 0.00000861, 0.00002475, 0.00000034, -0.00002855, -0.00002309, 0.00000808, 0.00000767, -0.00000921, 0.00000494, 0.00002332, 0.00002268, 0.00001282, 0.00000549, 0.00001189, 0.00001499, 0.00000532, 0.00000057, 0.00000564, 0.00001193, 0.00000770, 0.00000190, 0.00000557, 0.00000628, 0.00000834, 0.00001280, 0.00001052, 0.00000921, 0.00000475, 0.00000250, 0.00000978, 0.00000969, 0.00000900, 0.00001542, 0.00001751, 0.00001905, 0.00003564, },
		{ 0.00000203, -0.00005191, -0.00003766, -0.00001350, -0.00001695, -0.00001232, -0.00000646, -0.00000707, -0.00000743, -0.00000616, -0.00000323, -0.00000668, -0.00000549, -0.00000140, -0.00000464, -0.00000316, -0.00000328, -0.00000312, 0.00000003, -0.00000454, -0.00000148, -0.00000136, -0.00001116, -0.00000588, 0.00000048, -0.00000089, -0.00000168, -0.00000132, 0.00000237, -0.00000315, -0.00000682, -0.00000310, -0.00000308, -0.00000002, -0.00000809, -0.00000671, 0.00000147, -0.00001690, -0.00000345, 0.00000869, -0.00001092, 0.00001801, 0.00001358, -0.00001798, 0.00001553, 0.00000395, -0.00001457, 0.00001356, -0.00000618, -0.00000593, 0.00000900, -0.00000503, 0.00001118, 0.00000025, -0.00000950, 0.00001018, -0.00000857, -0.00000333, -0.00000194, -0.00001113, 0.00002632, -0.00000362, -0.00002251, 0.00003264, -0.00000454, -0.00002836, 0.00002124, 0.00000186, -0.00001608, 0.00000307, 0.00000332, 0.00000219, -0.00001026, -0.00000782, 0.00001120, 0.00000679, 0.00000066, -0.00001087, -0.00001163, 0.00000773, -0.00000563, -0.00001119, 0.00001113, -0.00000081, -0.00001465, -0.00000332, 0.00000295, 0.00000759, -0.00000593, -0.00002502, -0.00000065, 0.00002221, -0.00000006, -0.00002234, -0.00001760, 0.00000271, 0.00000910, -0.00000800, -0.00000155, 0.00002479, 0.00002288, 0.00000915, 0.00000686, 0.00001078, 0.00001379, 0.00000627, 0.00000003, 0.00000390, 0.00000947, 0.00001109, 0.00000426, 0.00000153, 0.00000543, 0.00000634, 0.00001180, 0.00001245, 0.00000652, 0.00000631, 0.00000334, 0.00000503, 0.00001071, 0.00000746, 0.00001056, 0.00001733, 0.00001642, 0.00002094, 0.00004137, 0.00005054, },
		{ -0.00005037, -0.00002772, -0.00001362, -0.00001620, -0.00000935, -0.00000647, -0.00000688, -0.00000666, -0.00000529, -0.00000474, -0.00000687, -0.00000289, -0.00000276, -0.00000488, -0.00000067, -0.00000235, -0.00000263, -0.00000244, -0.00000449, 0.00000019, -0.00000596, -0.00001140, -0.00000108, 0.00000100, -0.00000287, -0.00000273, 0.00000149, 0.00000174, -0.00000737, -0.00000512, -0.00000086, -0.00000243, -0.00000385, -0.00001046, 0.00000016, -0.00000385, -0.00002081, 0.00000600, 0.00000498, -0.00000711, 0.00002222, 0.00000051, -0.00000833, 0.00001889, -0.00001037, -0.00000327, 0.00001537, -0.00001461, -0.00000058, 0.00000801, 0.00000021, 0.00000836, -0.00000974, 0.00000311, 0.00001007, -0.00001789, -0.00000281, 0.00000095, 0.00000458, 0.00001287, -0.00001864, 0.00001036, 0.00002099, -0.00003370, 0.00000416, 0.00002442, -0.00002540, -0.00000493, 0.00001530, -0.00000122, -0.00000554, -0.00001149, 0.00000328, 0.00001263, 0.00000129, -0.00000160, -0.00001364, -0.00000872, 0.00000536, -0.00000679, 0.00000136, 0.00000733, -0.00001541, -0.00000672, 0.00001033, -0.00000087, -0.00000868, -0.00001147, -0.00000205, 0.00001582, -0.00000114, -0.00002367, -0.00000992, 0.00000610, 0.00000438, -0.00000623, -0.00000533, 0.00002052, 0.00002746, 0.00000671, 0.00000400, 0.00001280, 0.00001227, 0.00000590, 0.00000004, 0.00000318, 0.00000761, 0.00000968, 0.00000910, 0.00000169, 0.00000133, 0.00000582, 0.00000906, 0.00001350, 0.00000739, 0.00000329, 0.00000663, 0.00000358, 0.00000670, 0.00000955, 0.00000666, 0.00001375, 0.00001744, 0.00001475, 0.00002420, 0.00004471, 0.00003929, -0.00001573, },
		{ -0.00002079, -0.00001372, -0.00001416, -0.00000768, -0.00000631, -0.00000624, -0.00000658, -0.00000546, -0.00000548, -0.00000538, -0.00000226, -0.00000421, -0.00000199, 0.00000112, -0.00000378, -0.00000338, -0.00000346, -0.00000311, -0.00000137, -0.00000949, -0.00000721, 0.00000171, -0.00000182, -0.00000396, -0.00000057, 0.00000289, -0.00000270, -0.00000800, -0.00000069, -0.00000179, -0.00000471, -0.00000631, -0.00000706, 0.00000246, -0.00001405, -0.00001539, 0.00001440, -0.00000362, -0.00000104, 0.00002301, -0.00000680, 0.00000152, 0.00001191, -0.00001379, 0.00001060, 0.00000503, -0.00001803, 0.00000953, 0.00000620, -0.00000016, 0.00000190, -0.00000578, 0.00001310, -0.00000473, -0.00002029, 0.00000839, 0.00000405, 0.00000316, -0.00000042, -0.00000534, 0.00002163, -0.00001206, -0.00002151, 0.00003157, -0.00000824, -0.00002862, 0.00002281, 0.00000755, -0.00001236, -0.00000419, -0.00000469, 0.00000891, 0.00000677, 0.00000005, -0.00000303, -0.00001933, -0.00000554, 0.00000829, -0.00000166, 0.00000278, -0.00000688, -0.00001294, 0.00001247, 0.00000483, -0.00002070, -0.00000854, 0.00001003, 0.00001190, -0.00000508, -0.00002695, -0.00000997, 0.00001659, 0.00000457, -0.00001138, -0.00000512, 0.00001503, 0.00002806, 0.00001063, -0.00000129, 0.00001261, 0.00001427, 0.00000375, -0.00000020, 0.00000269, 0.00000733, 0.00000726, 0.00000947, 0.00000714, -0.00000122, 0.00000233, 0.00000841, 0.00001124, 0.00001003, 0.00000179, 0.00000490, 0.00000783, 0.00000253, 0.00000749, 0.00000846, 0.00000797, 0.00001640, 0.00001518, 0.00001383, 0.00002871, 0.00004546, 0.00002582, -0.00002913, -0.00004584, },
		{ -0.00001357, -0.00001222, -0.00000649, -0.00000577, -0.00000638, -0.00000696, -0.00000522, -0.00000532, -0.00000429, -0.00000259, -0.00000302, 0.00000137, -0.00000055, -0.00000567, -0.00000300, -0.00000385, -0.00000296, -0.00000374, -0.00000912, -0.00000297, 0.00000027, -0.00000391, -0.00000218, 0.00000103, 0.00000080, -0.00000558, -0.00000421, 0.00000023, -0.00000580, -0.00000471, -0.00000587, -0.00000418, -0.00000160, -0.00001943, -0.00000323, 0.00001273, -0.00001179, 0.00001103, 0.00001940, -0.00001171, 0.00000853, 0.00000461, -0.00000812, 0.00001454, -0.00000837, -0.00000920, 0.00001560, -0.00000177, -0.00000004, 0.00000220, 0.00000160, 0.00000827, -0.00001759, -0.00000683, 0.00001477, -0.00000330, 0.00000086, 0.00000137, 0.00000603, 0.00000651, -0.00002282, 0.00001059, 0.00001851, -0.00003552, 0.00000400, 0.00002917, -0.00001710, -0.00000722, 0.00000383, -0.00000355, 0.00000790, 0.00000422, 0.00000184, -0.00001177, -0.00002311, 0.00000686, 0.00001290, -0.00000529, -0.00000563, -0.00000855, 0.00000548, 0.00001317, -0.00002141, -0.00002170, 0.00001988, 0.00002181, -0.00001084, -0.00003146, -0.00001480, 0.00002191, 0.00001531, -0.00001685, -0.00000993, 0.00001470, 0.00002385, 0.00001449, -0.00000101, 0.00000732, 0.00001718, 0.00000396, -0.00000273, 0.00000269, 0.00000698, 0.00000695, 0.00000665, 0.00001015, 0.00000303, -0.00000279, 0.00000645, 0.00001034, 0.00000904, 0.00000439, 0.00000124, 0.00000910, 0.00000571, 0.00000118, 0.00000928, 0.00000791, 0.00001013, 0.00001690, 0.00001155, 0.00001509, 0.00003414, 0.00004317, 0.00001059, -0.00003876, -0.00003919, -0.00001540, },
		{ -0.00001022, -0.00000532, -0.00000586, -0.00000716, -0.00000657, -0.00000473, -0.00000532, -0.00000334, -0.00000176, -0.00000045, 0.00000179, -0.00000390, -0.00000551, -0.00000273, -0.00000500, -0.00000278, -0.00000450, -0.00000743, -0.00000186, -0.00000182, -0.00000303, -0.00000087, 0.00000025, -0.00000099, -0.00000477, -0.00000175, -0.00000345, -0.00000683, -0.00000223, -0.00000694, -0.00000351, -0.00000610, -0.00001770, 0.00000532, 0.00000286, -0.00000980, 0.00002338, 0.00000788, -0.00001152, 0.00001512, -0.00000138, -0.00000219, 0.00001040, -0.00001299, 0.00000313, 0.00001008, -0.00000684, 0.00000649, 0.00000258, 0.00000156, -0.00000088, -0.00001558, 0.00000718, 0.00000690, -0.00000746, 0.00000705, 0.00000293, 0.00000306, -0.00000544, -0.00000826, 0.00002031, -0.00001342, -0.00002255, 0.00003312, 0.00000089, -0.00002396, 0.00001330, -0.00000102, -0.00000687, 0.00001107, 0.00000459, -0.00000381, -0.00002303, -0.00001234, 0.00002218, 0.00000394, -0.00001571, -0.00000318, 0.00000503, 0.00001042, -0.00001148, -0.00003145, 0.00001158, 0.00003847, -0.00000603, -0.00004001, -0.00002011, 0.00001975, 0.00002607, -0.00001152, -0.00002005, 0.00001320, 0.00002336, 0.00001192, 0.00000342, 0.00000432, 0.00001471, 0.00000773, -0.00000514, 0.00000104, 0.00000721, 0.00000683, 0.00000574, 0.00000801, 0.00000837, -0.00000174, 0.00000043, 0.00001043, 0.00000746, 0.00000481, 0.00000237, 0.00000502, 0.00001014, 0.00000100, 0.00000292, 0.00001137, 0.00000684, 0.00001221, 0.00001500, 0.00000843, 0.00001953, 0.00003918, 0.00003654, -0.00000572, -0.00004364, -0.00003053, -0.00001173, -0.00001372, },
		{ -0.00000487, -0.00000673, -0.00000734, -0.00000561, -0.00000488, -0.00000483, -0.00000138, -0.00000034, 0.00000068, -0.00000030, -0.00000585, -0.00000439, -0.00000406, -0.00000519, -0.00000134, -0.00000525, -0.00000692, -0.00000180, -0.00000183, -0.00000210, -0.00000162, 0.00000020, -0.00000086, -0.00000419, -0.00000300, -0.00000609, -0.00000367, -0.00000270, -0.00000927, -0.00000142, -0.00000945, -0.00001437, 0.00000692, -0.00000494, 0.00000157, 0.00002601, -0.00000547, -0.00000212, 0.00001711, -0.00000833, 0.00000353, 0.00000596, -0.00001091, 0.00000797, 0.00000137, -0.00000136, 0.00001051, -0.00000345, 0.00000062, -0.00000303, -0.00000779, 0.00000904, -0.00000227, -0.00000049, 0.00000925, -0.00000310, 0.00000112, -0.00000357, 0.00000252, 0.00000588, -0.00002267, 0.00000993, 0.00002394, -0.00002612, 0.00000076, 0.00001865, -0.00001826, 0.00000161, 0.00001570, -0.00000308, -0.00001384, -0.00002113, 0.00000840, 0.00002067, -0.00001548, -0.00001144, 0.00001273, 0.00000691, -0.00000904, -0.00002621, -0.00000357, 0.00004088, 0.00001234, -0.00004363, -0.00002964, 0.00001635, 0.00002857, -0.00000018, -0.00002286, 0.00000383, 0.00002602, 0.00000948, 0.00000250, 0.00000815, 0.00001033, 0.00000834, -0.00000314, -0.00000261, 0.00000715, 0.00000670, 0.00000593, 0.00000641, 0.00000792, 0.00000428, -0.00000232, 0.00000591, 0.00000865, 0.00000231, 0.00000420, 0.00000398, 0.00000767, 0.00000693, -0.00000152, 0.00000766, 0.00001100, 0.00000568, 0.00001395, 0.00001109, 0.00000810, 0.00002675, 0.00004106, 0.00002493, -0.00002054, -0.00004261, -0.00002149, -0.00001058, -0.00001366, -0.00000783, },
		{ -0.00000738, -0.00000666, -0.00000516, -0.00000502, -0.00000273, 0.00000064, -0.00000017, 0.00000021, -0.00000249, -0.00000632, -0.00000435, -0.00000530, -0.00000305, -0.00000082, -0.00000745, -0.00000611, -0.00000071, -0.00000175, -0.00000297, -0.00000143, 0.00000191, -0.00000198, -0.00000555, -0.00000420, -0.00000511, -0.00000162, -0.00000677, -0.00000772, 0.00000099, -0.00001496, -0.00001112, 0.00000611, -0.00000631, 0.00001307, 0.00001886, -0.00001049, 0.00001040, 0.00000963, -0.00001118, 0.00001128, 0.00000044, -0.00000896, 0.00000801, -0.00000026, 0.00000633, 0.00000429, -0.00000734, 0.00000568, -0.00000366, -0.00000478, 0.00000602, -0.00000231, 0.00000580, 0.00000177, -0.00000418, 0.00000605, -0.00000351, 0.00000044, -0.00000366, -0.00000779, 0.00002108, -0.00000511, -0.00001828, 0.00002517, -0.00000391, -0.00002110, 0.00001960, 0.00000579, -0.00001432, -0.00001251, -0.00000721, 0.00001711, 0.00000164, -0.00001997, 0.00001060, 0.00001456, -0.00001434, -0.00002092, -0.00000536, 0.00002785, 0.00002516, -0.00003143, -0.00004022, 0.00001025, 0.00002792, 0.00000483, -0.00001532, -0.00000419, 0.00002191, 0.00001287, -0.00000266, 0.00001027, 0.00001279, 0.00000371, -0.00000077, -0.00000312, 0.00000426, 0.00000742, 0.00000534, 0.00000710, 0.00000552, 0.00000594, 0.00000261, 0.00000057, 0.00000692, 0.00000267, 0.00000170, 0.00000686, 0.00000437, 0.00000752, 0.00000340, 0.00000009, 0.00001155, 0.00000773, 0.00000609, 0.00001425, 0.00000677, 0.00001233, 0.00003391, 0.00003721, 0.00001049, -0.00003051, -0.00003670, -0.00001463, -0.00001139, -0.00001242, -0.00000572, -0.00000554, },
		{ -0.00000596, -0.00000508, -0.00000373, 0.00000007, 0.00000084, -0.00000104, -0.00000042, -0.00000429, -0.00000668, -0.00000470, -0.00000449, -0.00000072, -0.00000305, -0.00000899, -0.00000359, -0.00000021, -0.00000313, -0.00000314, 0.00000107, 0.00000186, -0.00000536, -0.00000571, -0.00000347, -0.00000391, -0.00000298, -0.00000900, -0.00000229, -0.00000202, -0.00002114, -0.00000507, 0.00000453, -0.00000401, 0.00002025, 0.00000972, -0.00000663, 0.00001602, -0.00000153, -0.00000448, 0.00001513, -0.00000866, -0.00000491, 0.00000978, 0.00000156, 0.00000621, -0.00000393, -0.00000113, 0.00000906, -0.00000907, -0.00000209, 0.00000630, -0.00000036, 0.00000366, -0.00000308, 0.00000372, 0.00000474, -0.00000966, 0.00000185, -0.00000013, 0.00000228, 0.00000872, -0.00001504, 0.00000857, 0.00001828, -0.00002600, 0.00000037, 0.00002127, -0.00001476, -0.00000945, -0.00000162, -0.00000168, 0.00000867, -0.00000923, -0.00000141, 0.00002042, -0.00001081, -0.00002692, 0.00000117, 0.00002117, 0.00001901, -0.00001459, -0.00003784, -0.00000088, 0.00002669, 0.00000414, -0.00000993, -0.00000075, 0.00001334, 0.00001363, -0.00000257, 0.00000512, 0.00001849, 0.00000310, -0.00000470, -0.00000005, 0.00000192, 0.00000607, 0.00000571, 0.00000692, 0.00000614, 0.00000279, 0.00000670, 0.00000317, 0.00000076, 0.00000321, 0.00000013, 0.00000605, 0.00000683, 0.00000272, 0.00000765, 0.00000159, 0.00000341, 0.00001221, 0.00000399, 0.00000809, 0.00001231, 0.00000500, 0.00002019, 0.00003673, 0.00002817, -0.00000268, -0.00003460, -0.00002909, -0.00001104, -0.00001236, -0.00001011, -0.00000485, -0.00000663, -0.00000717, },
		{ -0.00000423, -0.00000123, 0.00000140, -0.00000066, -0.00000140, -0.00000115, -0.00000628, -0.00000660, -0.00000395, -0.00000272, -0.00000072, -0.00000616, -0.00000792, -0.00000131, -0.00000163, -0.00000388, -0.00000101, 0.00000264, -0.00000151, -0.00000729, -0.00000351, -0.00000353, -0.00000420, -0.00000439, -0.00000726, 0.00000077, -0.00001070, -0.00002191, 0.00000216, 0.00000034, 0.00000151, 0.00002425, 0.00000213, -0.00000043, 0.00001408, -0.00000669, 0.00000603, 0.00000897, -0.00001459, 0.00000469, 0.00000984, -0.00000021, 0.00000236, -0.00000387, 0.00000768, 0.00000318, -0.00001243, 0.00000489, 0.00000487, -0.00000237, 0.00000159, 0.00000074, 0.00000716, -0.00000551, -0.00000762, 0.00001047, -0.00000074, -0.00000067, 0.00000063, -0.00000287, 0.00001855, -0.00000613, -0.00002083, 0.00001873, -0.00000072, -0.00001759, 0.00001007, -0.00000404, -0.00000875, 0.00000548, 0.00000030, 0.00001185, -0.00000051, -0.00003082, -0.00000213, 0.00002853, 0.00000946, -0.00001419, -0.00002289, -0.00000528, 0.00001982, 0.00000391, -0.00001236, 0.00000607, 0.00001383, 0.00000648, -0.00000009, 0.00000148, 0.00001708, 0.00000983, -0.00000931, -0.00000159, 0.00000465, 0.00000271, 0.00000571, 0.00000674, 0.00000698, 0.00000195, 0.00000401, 0.00000962, -0.00000002, -0.00000225, 0.00000160, 0.00000307, 0.00000853, 0.00000276, 0.00000349, 0.00000869, -0.00000005, 0.00000659, 0.00001027, 0.00000153, 0.00001031, 0.00000949, 0.00000751, 0.00002742, 0.00003346, 0.00001792, -0.00001214, -0.00003484, -0.00002223, -0.00000965, -0.00001233, -0.00000786, -0.00000520, -0.00000717, -0.00000649, -0.00000558, },
		{ 0.00000069, 0.00000060, -0.00000207, -0.00000106, -0.00000285, -0.00000776, -0.00000515, -0.00000277, -0.00000206, -0.00000239, -0.00000763, -0.00000566, -0.00000109, -0.00000294, -0.00000247, 0.00000092, 0.00000111, -0.00000487, -0.00000583, -0.00000227, -0.00000525, -0.00000343, -0.00000432, -0.00000524, -0.00000152, -0.00001879, -0.00001633, 0.00000494, -0.00000369, 0.00001178, 0.00002349, -0.00000454, 0.00000588, 0.00001022, -0.00000549, 0.00001085, -0.00000159, -0.00001021, 0.00001297, 0.00000329, -0.00000198, 0.00000267, 0.00000105, 0.00000913, -0.00000549, -0.00000751, 0.00000903, -0.00000150, -0.00000172, 0.00000456, 0.00000309, 0.00000107, -0.00000958, 0.00000510, 0.00001047, -0.00000957, 0.00000056, 0.00000515, 0.00000584, 0.00000834, -0.00001684, -0.00000263, 0.00001124, -0.00001641, 0.00000592, 0.00001348, -0.00002368, -0.00000428, 0.00001819, 0.00000607, -0.00000216, -0.00002299, -0.00001149, 0.00003236, 0.00001486, -0.00002647, -0.00001706, 0.00000641, 0.00001311, 0.00000021, -0.00001567, 0.00000483, 0.00002307, 0.00000323, -0.00000555, 0.00000373, 0.00001212, 0.00001310, -0.00000544, -0.00000838, 0.00000657, 0.00000356, 0.00000241, 0.00000772, 0.00000653, 0.00000325, 0.00000105, 0.00000962, 0.00000722, -0.00000697, -0.00000171, 0.00000423, 0.00000554, 0.00000609, 0.00000016, 0.00000786, 0.00000721, -0.00000149, 0.00000975, 0.00000624, 0.00000093, 0.00001254, 0.00000758, 0.00001243, 0.00003039, 0.00002691, 0.00000977, -0.00001912, -0.00003330, -0.00001624, -0.00000921, -0.00001167, -0.00000645, -0.00000586, -0.00000693, -0.00000585, -0.00000486, -0.00000229, },
		{ -0.00000105, -0.00000239, -0.00000114, -0.00000508, -0.00000744, -0.00000317, -0.00000266, -0.00000236, -0.00000384, -0.00000770, -0.00000422, -0.00000158, -0.00000245, -0.00000091, 0.00000059, -0.00000134, -0.00000557, -0.00000381, -0.00000365, -0.00000540, -0.00000116, -0.00000543, -0.00000504, -0.00000620, -0.00002254, -0.00000968, 0.00000285, -0.00000156, 0.00002264, 0.00001535, -0.00000713, 0.00001316, 0.00000537, -0.00000320, 0.00000992, -0.00000691, -0.00000047, 0.00001175, -0.00000387, 0.00000222, 0.00000458, 0.00000312, 0.00000556, -0.00000856, -0.00000065, 0.00000542, -0.00000461, 0.00000444, 0.00000436, -0.00000075, -0.00000260, -0.00000207, 0.00001195, -0.00000165, -0.00001045, 0.00001064, 0.00000615, 0.00000311, -0.00000021, -0.00001287, 0.00000183, -0.00000382, -0.00000306, 0.00002330, -0.00001241, -0.00002906, 0.00002207, 0.00002120, -0.00001218, -0.00001909, -0.00001033, 0.00002454, 0.00002570, -0.00002858, -0.00002986, 0.00001873, 0.00002133, -0.00000797, -0.00001959, -0.00000057, 0.00002716, 0.00001197, -0.00001345, 0.00000134, 0.00001350, 0.00000888, 0.00000060, -0.00000925, 0.00000102, 0.00000777, 0.00000051, 0.00000595, 0.00000764, 0.00000269, 0.00000187, 0.00000559, 0.00001125, -0.00000195, -0.00000923, 0.00000407, 0.00000521, 0.00000438, 0.00000325, 0.00000224, 0.00001122, 0.00000239, -0.00000019, 0.00001132, 0.00000065, 0.00000372, 0.00001450, 0.00000604, 0.00001707, 0.00002965, 0.00002082, 0.00000297, -0.00002540, -0.00002963, -0.00001087, -0.00000980, -0.00001083, -0.00000552, -0.00000620, -0.00000631, -0.00000508, -0.00000336, -0.00000043, 0.00000088, },
		{ -0.00000219, -0.00000206, -0.00000644, -0.00000570, -0.00000235, -0.00000332, -0.00000242, -0.00000500, -0.00000762, -0.00000315, -0.00000113, -0.00000161, -0.00000107, -0.00000046, -0.00000244, -0.00000510, -0.00000357, -0.00000486, -0.00000274, -0.00000099, -0.00000835, -0.00000472, -0.00001116, -0.00002347, -0.00000497, 0.00000112, 0.00000676, 0.00002674, 0.00000416, -0.00000169, 0.00001735, -0.00000201, 0.00000028, 0.00000854, -0.00000692, 0.00000502, 0.00000486, -0.00000324, 0.00000801, 0.00000255, 0.00000377, 0.00000287, -0.00000844, 0.00000155, 0.00000132, -0.00000019, 0.00000679, -0.00000188, -0.00000105, 0.00000162, 0.00000388, 0.00000616, -0.00000933, 0.00000127, 0.00001430, -0.00000103, 0.00000250, -0.00000232, -0.00001290, -0.00000245, 0.00000055, 0.00001726, 0.00000782, -0.00003480, -0.00000057, 0.00003855, -0.00000619, -0.00002956, -0.00000393, 0.00002065, 0.00002498, -0.00001750, -0.00004187, 0.00001199, 0.00003967, -0.00000614, -0.00002979, -0.00000496, 0.00002372, 0.00002126, -0.00000999, -0.00000871, 0.00001555, 0.00000806, -0.00000176, -0.00000281, -0.00000383, 0.00000570, 0.00000446, 0.00000231, 0.00000804, 0.00000263, 0.00000171, 0.00000545, 0.00000800, 0.00000527, -0.00000889, -0.00000335, 0.00000764, 0.00000188, 0.00000382, 0.00000319, 0.00000580, 0.00001046, -0.00000211, 0.00000364, 0.00000882, -0.00000301, 0.00001031, 0.00001344, 0.00000427, 0.00002158, 0.00002746, 0.00001542, -0.00000467, -0.00002985, -0.00002318, -0.00000751, -0.00001143, -0.00000926, -0.00000475, -0.00000633, -0.00000537, -0.00000384, -0.00000178, 0.00000022, -0.00000008, -0.00000221, },
		{ -0.00000305, -0.00000648, -0.00000420, -0.00000285, -0.00000343, -0.00000241, -0.00000656, -0.00000695, -0.00000140, -0.00000069, -0.00000214, -0.00000167, -0.00000050, -0.00000315, -0.00000514, -0.00000366, -0.00000388, -0.00000057, -0.00000446, -0.00000975, -0.00000404, -0.00001743, -0.00002231, -0.00000078, 0.00000238, 0.00001558, 0.00002317, -0.00000206, 0.00000793, 0.00001334, -0.00000742, 0.00000742, 0.00000602, -0.00000699, 0.00000590, 0.00000136, 0.00000270, 0.00000809, -0.00000015, 0.00000660, -0.00000087, -0.00000911, 0.00000306, 0.00000223, 0.00000325, 0.00000130, -0.00000363, 0.00000585, 0.00000316, 0.00000147, 0.00000049, -0.00000471, 0.00001041, 0.00000660, -0.00000342, 0.00000579, -0.00001000, -0.00001513, 0.00000526, 0.00001428, 0.00001319, -0.00001784, -0.00002133, 0.00003037, 0.00001551, -0.00003538, -0.00001037, 0.00002756, 0.00001994, -0.00001218, -0.00003795, -0.00000464, 0.00004464, 0.00001217, -0.00003704, -0.00001466, 0.00002085, 0.00002159, 0.00000051, -0.00001206, 0.00000814, 0.00001289, -0.00000602, -0.00000193, 0.00000098, -0.00000125, 0.00000619, 0.00000415, 0.00000458, 0.00000407, 0.00000011, 0.00000661, 0.00000644, 0.00000426, -0.00000078, -0.00000793, 0.00000315, 0.00000463, -0.00000012, 0.00000620, 0.00000331, 0.00000811, 0.00000726, -0.00000412, 0.00000666, 0.00000380, -0.00000087, 0.00001634, 0.00000737, 0.00000522, 0.00002674, 0.00002342, 0.00000916, -0.00001249, -0.00003026, -0.00001593, -0.00000747, -0.00001241, -0.00000669, -0.00000467, -0.00000608, -0.00000384, -0.00000262, -0.00000115, -0.00000013, -0.00000105, -0.00000280, -0.00000209, },
		{ -0.00000616, -0.00000369, -0.00000340, -0.00000286, -0.00000333, -0.00000782, -0.00000485, 0.00000005, -0.00000167, -0.00000303, -0.00000106, -0.00000056, -0.00000452, -0.00000464, -0.00000275, -0.00000259, -0.00000143, -0.00000847, -0.00000814, -0.00000608, -0.00002404, -0.00001747, 0.00000297, 0.00000512, 0.00002183, 0.00001697, -0.00000169, 0.00001361, 0.00000446, -0.00000466, 0.00001376, -0.00000107, -0.00000627, 0.00000752, 0.00000201, 0.00000621, 0.00000440, 0.00000166, 0.00000752, -0.00000768, -0.00000653, 0.00000716, 0.00000227, 0.00000045, -0.00000201, 0.00000375, 0.00000867, -0.00000284, 0.00000079, 0.00000244, 0.00000097, 0.00000927, 0.00000030, 0.00000117, 0.00000083, -0.00001914, -0.00000366, 0.00001840, 0.00001080, -0.00000428, -0.00001723, 0.00000718, 0.00002343, -0.00002116, -0.00002346, 0.00002879, 0.00002558, -0.00001689, -0.00003134, -0.00000883, 0.00003192, 0.00002713, -0.00002666, -0.00002817, 0.00001566, 0.00002001, 0.00000364, -0.00000363, 0.00000046, 0.00001081, -0.00000248, -0.00000827, 0.00000601, -0.00000021, -0.00000046, 0.00000881, 0.00000367, 0.00000191, 0.00000087, 0.00000486, 0.00000889, 0.00000080, 0.00000154, -0.00000145, -0.00000469, 0.00000409, 0.00000059, 0.00000298, 0.00000732, 0.00000232, 0.00001028, 0.00000271, -0.00000459, 0.00000801, 0.00000059, 0.00000537, 0.00001617, 0.00000024, 0.00001196, 0.00002950, 0.00001631, 0.00000300, -0.00001825, -0.00002727, -0.00001077, -0.00000930, -0.00001116, -0.00000446, -0.00000525, -0.00000471, -0.00000232, -0.00000248, -0.00000131, -0.00000044, -0.00000182, -0.00000320, -0.00000185, -0.00000362, },
		{ -0.00000352, -0.00000333, -0.00000259, -0.00000494, -0.00000753, -0.00000227, -0.00000027, -0.00000323, -0.00000261, -0.00000011, -0.00000186, -0.00000516, -0.00000281, -0.00000232, -0.00000281, -0.00000390, -0.00001029, -0.00000626, -0.00001177, -0.00002715, -0.00001017, 0.00000437, 0.00000949, 0.00002615, 0.00001068, 0.00000099, 0.00001363, -0.00000103, 0.00000367, 0.00001222, -0.00000890, -0.00000058, 0.00000936, 0.00000216, 0.00000661, 0.00000277, 0.00000488, 0.00000255, -0.00001137, 0.00000145, 0.00000771, -0.00000233, -0.00000083, 0.00000294, 0.00000928, 0.00000160, -0.00000542, 0.00000720, 0.00000316, 0.00000089, 0.00000695, -0.00000005, 0.00000065, -0.00000934, -0.00001314, 0.00001413, 0.00001414, -0.00000307, -0.00000314, -0.00000007, 0.00000815, -0.00000608, -0.00002136, 0.00001742, 0.00003394, -0.00001559, -0.00003603, -0.00000293, 0.00002379, 0.00002205, -0.00000890, -0.00002824, 0.00000264, 0.00001945, 0.00000113, 0.00000206, 0.00000570, 0.00000192, -0.00000016, -0.00000887, 0.00000116, 0.00000747, -0.00000437, 0.00000534, 0.00000925, -0.00000164, 0.00000030, 0.00000436, 0.00000879, 0.00000273, -0.00000316, 0.00000484, -0.00000215, -0.00000447, 0.00000318, 0.00000080, 0.00000680, 0.00000467, 0.00000352, 0.00001140, -0.00000404, -0.00000261, 0.00000944, -0.00000021, 0.00001020, 0.00001036, -0.00000130, 0.00002150, 0.00002582, 0.00000871, -0.00000105, -0.00002216, -0.00002314, -0.00000803, -0.00001049, -0.00000851, -0.00000372, -0.00000513, -0.00000253, -0.00000228, -0.00000324, -0.00000110, -0.00000057, -0.00000280, -0.00000308, -0.00000121, -0.00000437, -0.00000617, },
		},
		{ { 0.00010754, 0.00002759, -0.00003312, -0.00001138, -0.00000259, -0.00000535, 0.00001342, -0.00000823, 0.00000237, 0.00000609, 0.00000143, 0.00000150, 0.00000880, 0.00000193, 0.00001335, 0.00000451, 0.00000826, 0.00001632, 0.00002527, 0.00000242, -0.00000229, -0.00001887, -0.00001946, -0.00000568, -0.00000246, -0.00001430, 0.00000400, -0.00001012, -0.00000245, 0.00000513, -0.00000217, -0.00000855, 0.00000220, -0.00000678, 0.00000118, -0.00000724, 0.00000381, 0.00000547, -0.00000063, -0.00000680, 0.00000196, -0.00000610, 0.00000017, -0.00000644, 0.00000446, -0.00000285, -0.00000121, 0.00000238, 0.00000053, -0.00000771, 0.00000363, -0.00000779, 0.00000698, 0.00000446, -0.00000381, -0.00000906, -0.00000272, 0.00000241, 0.00001816, -0.00000835, -0.00000957, 0.00000221, 0.00000315, -0.00000632, 0.00000535, -0.00000632, 0.00000315, 0.00000221, -0.00000957, -0.00000835, 0.00001816, 0.00000241, -0.00000272, -0.00000906, -0.00000381, 0.00000446, 0.00000698, -0.00000779, 0.00000363, -0.00000771, 0.00000053, 0.00000238, -0.00000121, -0.00000285, 0.00000446, -0.00000644, 0.00000017, -0.00000610, 0.00000196, -0.00000680, -0.00000063, 0.00000547, 0.00000381, -0.00000724, 0.00000118, -0.00000678, 0.00000220, -0.00000855, -0.00000217, 0.00000513, -0.00000245, -0.00001012, 0.00000400, -0.00001430, -0.00000246, -0.00000568, -0.00001946, -0.00001887, -0.00000229, 0.00000242, 0.00002527, 0.00001632, 0.00000826, 0.00000451, 0.00001335, 0.00000193, 0.00000880, 0.00000150, 0.00000143, 0.00000609, 0.00000237, -0.00000823, 0.00001342, -0.00000535, -0.00000259, -0.00001138, -0.00003312, 0.00002759, },
		{ 0.00002982, -0.00002851, -0.00001270, -0.00000243, -0.00000463, 0.00001416, -0.00000824, 0.00000316, 0.00000739, 0.00000179, 0.00000200, 0.00000915, 0.00000160, 0.00001624, 0.00000383, 0.00000962, 0.00002273, 0.00002263, -0.00000197, -0.00000594, -0.00002560, -0.00001369, -0.00000277, -0.00000823, -0.00000933, 0.00000353, -0.00001353, 0.00000328, 0.00000194, -0.00000446, -0.00000535, -0.00000012, -0.00000549, 0.00000078, -0.00000763, 0.00000778, 0.00000226, -0.00000220, -0.00000487, 0.00000039, -0.00000521, -0.00000050, -0.00000553, 0.00000528, -0.00000558, 0.00000159, 0.00000234, -0.00000285, -0.00000469, 0.00000206, -0.00000706, 0.00001028, 0.00000047, -0.00000552, -0.00000824, -0.00000188, 0.00000816, 0.00001398, -0.00001401, -0.00000526, 0.00000476, -0.00000123, -0.00000280, 0.00000379, -0.00000740, 0.00000718, -0.00000303, -0.00001075, -0.00000148, 0.00001909, -0.00000343, -0.00000217, -0.00001031, -0.00000038, 0.00000667, 0.00000284, -0.00000647, 0.00000342, -0.00000971, 0.00000485, -0.00000002, -0.00000216, -0.00000023, 0.00000247, -0.00000618, 0.00000005, -0.00000629, 0.00000311, -0.00000910, 0.00000299, 0.00000634, -0.00000021, -0.00000530, 0.00000005, -0.00000664, 0.00000342, -0.00001192, 0.00000279, 0.00000475, -0.00000708, -0.00000459, 0.00000057, -0.00001621, 0.00000348, -0.00001258, -0.00002084, -0.00001135, -0.00000065, 0.00000769, 0.00002510, 0.00000988, 0.00000901, 0.00000393, 0.00001092, 0.00000276, 0.00000786, 0.00000153, 0.00000080, 0.00000441, 0.00000210, -0.00000800, 0.00001231, -0.00000596, -0.00000217, -0.00001059, -0.00003804, 0.00002558, 0.00010765, },
		{ -0.00002445, -0.00001420, -0.00000162, -0.00000409, 0.00001464, -0.00000780, 0.00000423, 0.00000814, 0.00000229, 0.00000271, 0.00000894, 0.00000226, 0.00001875, 0.00000261, 0.00001342, 0.00002761, 0.00001811, -0.00000599, -0.00001173, -0.00002911, -0.00000595, -0.00000426, -0.00001131, -0.00000381, -0.00000005, -0.00001342, 0.00000760, -0.00000314, -0.00000380, -0.00000370, -0.00000221, -0.00000365, -0.00000075, -0.00000596, 0.00001019, -0.00000195, -0.00000179, -0.00000402, -0.00000077, -0.00000429, -0.00000142, -0.00000372, 0.00000445, -0.00000711, 0.00000495, 0.00000004, -0.00000423, -0.00000207, -0.00000033, -0.00000427, 0.00001153, -0.00000406, -0.00000577, -0.00000782, 0.00000073, 0.00001240, 0.00000711, -0.00001677, 0.00000058, 0.00000400, -0.00000387, 0.00000120, -0.00000015, -0.00000517, 0.00000867, -0.00000906, -0.00000840, 0.00000477, 0.00001709, -0.00000828, -0.00000084, -0.00001157, 0.00000449, 0.00000625, -0.00000061, -0.00000400, 0.00000130, -0.00000954, 0.00000846, -0.00000401, -0.00000074, 0.00000102, 0.00000036, -0.00000518, -0.00000088, -0.00000540, 0.00000305, -0.00001053, 0.00000777, 0.00000433, -0.00000271, -0.00000305, -0.00000210, -0.00000491, 0.00000256, -0.00001359, 0.00000878, 0.00000056, -0.00000836, 0.00000048, -0.00000565, -0.00001376, 0.00000656, -0.00002110, -0.00001707, -0.00000562, 0.00000029, 0.00001340, 0.00002180, 0.00000496, 0.00001083, 0.00000207, 0.00000956, 0.00000344, 0.00000667, 0.00000206, -0.00000043, 0.00000274, 0.00000239, -0.00000779, 0.00001085, -0.00000618, -0.00000140, -0.00001053, -0.00004292, 0.00002370, 0.00010800, 0.00003223, },
		{ -0.00001550, -0.00000030, -0.00000394, 0.00001512, -0.00000684, 0.00000526, 0.00000844, 0.00000318, 0.00000321, 0.00000857, 0.00000395, 0.00002025, 0.00000196, 0.00001916, 0.00003000, 0.00001257, -0.00001058, -0.00001843, -0.00002820, 0.00000083, -0.00000870, -0.00001067, -0.00000020, -0.00000466, -0.00001012, 0.00000888, -0.00000793, -0.00000119, -0.00000398, -0.00000314, -0.00000233, -0.00000237, -0.00000273, 0.00001020, -0.00000567, -0.00000019, -0.00000428, -0.00000114, -0.00000390, -0.00000190, -0.00000179, 0.00000227, -0.00000671, 0.00000739, -0.00000344, -0.00000345, -0.00000093, -0.00000222, -0.00000028, 0.00001027, -0.00000794, -0.00000518, -0.00000739, 0.00000488, 0.00001384, -0.00000099, -0.00001574, 0.00000580, 0.00000073, -0.00000370, 0.00000362, -0.00000454, -0.00000026, 0.00000632, -0.00001349, -0.00000341, 0.00000909, 0.00001304, -0.00001149, 0.00000060, -0.00001227, 0.00000996, 0.00000306, -0.00000204, -0.00000184, -0.00000190, -0.00000689, 0.00000988, -0.00000798, 0.00000251, 0.00000023, -0.00000080, -0.00000428, -0.00000202, -0.00000372, 0.00000150, -0.00000994, 0.00001208, 0.00000004, -0.00000277, -0.00000188, -0.00000408, -0.00000243, -0.00000038, -0.00001210, 0.00001331, -0.00000596, -0.00000541, 0.00000241, -0.00001200, -0.00000764, 0.00000473, -0.00002785, -0.00000951, -0.00000319, 0.00000226, 0.00001824, 0.00001600, 0.00000254, 0.00001241, -0.00000046, 0.00000932, 0.00000345, 0.00000578, 0.00000274, -0.00000233, 0.00000145, 0.00000307, -0.00000779, 0.00000916, -0.00000583, -0.00000065, -0.00001122, -0.00004745, 0.00002180, 0.00010868, 0.00003468, -0.00002110, },
		{ 0.00000126, -0.00000422, 0.00001589, -0.00000549, 0.00000597, 0.00000856, 0.00000445, 0.00000323, 0.00000860, 0.00000624, 0.00002058, 0.00000292, 0.00002576, 0.00002955, 0.00000642, -0.00001639, -0.00002399, -0.00002341, 0.00000442, -0.00001358, -0.00000706, 0.00000021, -0.00000797, -0.00000534, 0.00000691, -0.00001073, 0.00000178, -0.00000559, -0.00000272, -0.00000218, -0.00000296, 0.00000087, 0.00000797, -0.00000787, 0.00000154, -0.00000514, -0.00000084, -0.00000418, -0.00000145, -0.00000065, -0.00000030, -0.00000454, 0.00000788, -0.00000662, -0.00000127, -0.00000153, -0.00000250, 0.00000356, 0.00000687, -0.00001039, -0.00000440, -0.00000628, 0.00000972, 0.00001178, -0.00000832, -0.00001125, 0.00000857, -0.00000324, -0.00000106, 0.00000323, -0.00000720, 0.00000535, 0.00000034, -0.00001437, 0.00000244, 0.00001103, 0.00000793, -0.00001290, 0.00000170, -0.00001182, 0.00001486, -0.00000220, -0.00000080, -0.00000143, -0.00000463, -0.00000258, 0.00000843, -0.00001025, 0.00000621, -0.00000231, -0.00000045, -0.00000431, -0.00000249, -0.00000212, -0.00000101, -0.00000683, 0.00001417, -0.00000492, -0.00000063, -0.00000261, -0.00000462, -0.00000068, -0.00000414, -0.00000726, 0.00001415, -0.00001205, 0.00000066, -0.00000021, -0.00001540, -0.00000054, -0.00000200, -0.00002984, -0.00000108, -0.00000383, 0.00000637, 0.00002058, 0.00000925, 0.00000267, 0.00001271, -0.00000271, 0.00000973, 0.00000274, 0.00000556, 0.00000303, -0.00000463, 0.00000079, 0.00000391, -0.00000809, 0.00000747, -0.00000482, -0.00000034, -0.00001246, -0.00005139, 0.00001966, 0.00010980, 0.00003697, -0.00001842, -0.00001626, },
		{ -0.00000476, 0.00001713, -0.00000401, 0.00000633, 0.00000880, 0.00000584, 0.00000280, 0.00000947, 0.00000846, 0.00002014, 0.00000608, 0.00003193, 0.00002641, -0.00000039, -0.00002319, -0.00002652, -0.00001673, 0.00000418, -0.00001661, -0.00000245, -0.00000228, -0.00000852, -0.00000121, 0.00000288, -0.00001095, 0.00000371, -0.00000742, -0.00000151, -0.00000309, -0.00000191, 0.00000347, 0.00000444, -0.00000831, 0.00000264, -0.00000592, -0.00000034, -0.00000481, -0.00000014, -0.00000086, -0.00000211, -0.00000154, 0.00000626, -0.00000832, 0.00000106, -0.00000330, -0.00000079, 0.00000593, 0.00000229, -0.00001127, -0.00000378, -0.00000392, 0.00001391, 0.00000642, -0.00001292, -0.00000482, 0.00000809, -0.00000593, 0.00000252, 0.00000022, -0.00000668, 0.00000916, -0.00000748, -0.00001099, 0.00000717, 0.00001097, 0.00000256, -0.00001260, 0.00000224, -0.00000996, 0.00001801, -0.00000818, 0.00000279, -0.00000357, -0.00000523, 0.00000168, 0.00000463, -0.00000979, 0.00000871, -0.00000547, 0.00000115, -0.00000556, -0.00000160, -0.00000171, -0.00000316, -0.00000183, 0.00001301, -0.00000863, 0.00000244, -0.00000499, -0.00000317, -0.00000106, -0.00000664, -0.00000055, 0.00001048, -0.00001484, 0.00000707, -0.00000663, -0.00001386, 0.00000403, -0.00001127, -0.00002589, 0.00000503, -0.00000571, 0.00001239, 0.00001937, 0.00000346, 0.00000439, 0.00001148, -0.00000390, 0.00001003, 0.00000171, 0.00000600, 0.00000252, -0.00000682, 0.00000070, 0.00000466, -0.00000869, 0.00000602, -0.00000328, -0.00000083, -0.00001388, -0.00005472, 0.00001714, 0.00011148, 0.00003881, -0.00001618, -0.00001629, 0.00000270, },
		{ 0.00001884, -0.00000271, 0.00000647, 0.00000931, 0.00000704, 0.00000223, 0.00001129, 0.00001000, 0.00001976, 0.00001144, 0.00003652, 0.00002081, -0.00000809, -0.00002980, -0.00002514, -0.00001068, 0.00000114, -0.00001661, 0.00000103, -0.00000611, -0.00000623, 0.00000061, -0.00000141, -0.00000919, 0.00000400, -0.00000852, -0.00000036, -0.00000437, 0.00000063, 0.00000417, 0.00000088, -0.00000744, 0.00000287, -0.00000620, -0.00000016, -0.00000521, 0.00000146, -0.00000237, -0.00000235, 0.00000101, 0.00000327, -0.00000808, 0.00000242, -0.00000511, 0.00000247, 0.00000601, -0.00000223, -0.00001091, -0.00000325, -0.00000014, 0.00001598, -0.00000097, -0.00001362, 0.00000142, 0.00000491, -0.00000610, 0.00000509, -0.00000392, -0.00000298, 0.00000925, -0.00001448, -0.00000419, 0.00000944, 0.00000975, -0.00000259, -0.00001081, 0.00000223, -0.00000676, 0.00001858, -0.00001326, 0.00000753, -0.00000806, -0.00000267, 0.00000389, 0.00000014, -0.00000679, 0.00000888, -0.00000784, 0.00000306, -0.00000753, 0.00000066, -0.00000322, -0.00000341, 0.00000339, 0.00000884, -0.00000966, 0.00000469, -0.00000780, -0.00000028, -0.00000403, -0.00000593, 0.00000537, 0.00000348, -0.00001282, 0.00001057, -0.00001413, -0.00000759, 0.00000328, -0.00001924, -0.00001731, 0.00000679, -0.00000634, 0.00001874, 0.00001476, 0.00000016, 0.00000622, 0.00000925, -0.00000383, 0.00000960, 0.00000100, 0.00000665, 0.00000112, -0.00000844, 0.00000096, 0.00000522, -0.00000953, 0.00000508, -0.00000150, -0.00000232, -0.00001499, -0.00005760, 0.00001418, 0.00011374, 0.00003996, -0.00001402, -0.00001562, 0.00000373, -0.00000526, },
		{ -0.00000171, 0.00000666, 0.00000999, 0.00000786, 0.00000196, 0.00001381, 0.00001057, 0.00002038, 0.00001846, 0.00003861, 0.00001291, -0.00001654, -0.00003455, -0.00002043, -0.00000721, -0.00000271, -0.00001395, 0.00000204, -0.00000935, -0.00000228, -0.00000034, -0.00000434, -0.00000670, 0.00000290, -0.00000849, 0.00000005, -0.00000508, 0.00000380, 0.00000283, -0.00000168, -0.00000607, 0.00000247, -0.00000595, -0.00000048, -0.00000492, 0.00000258, -0.00000446, -0.00000092, 0.00000203, 0.00000012, -0.00000631, 0.00000224, -0.00000582, 0.00000618, 0.00000375, -0.00000570, -0.00000994, -0.00000236, 0.00000453, 0.00001489, -0.00000843, -0.00001043, 0.00000547, 0.00000065, -0.00000377, 0.00000522, -0.00000708, 0.00000238, 0.00000511, -0.00001812, 0.00000400, 0.00000893, 0.00000816, -0.00000726, -0.00000777, 0.00000174, -0.00000263, 0.00001633, -0.00001606, 0.00001174, -0.00001367, 0.00000295, 0.00000258, -0.00000285, -0.00000264, 0.00000665, -0.00000841, 0.00000409, -0.00000916, 0.00000347, -0.00000643, -0.00000072, 0.00000670, 0.00000322, -0.00000783, 0.00000482, -0.00000941, 0.00000250, -0.00000862, -0.00000125, 0.00000768, -0.00000396, -0.00000663, 0.00000905, -0.00001909, 0.00000077, -0.00000334, -0.00002226, -0.00000741, 0.00000429, -0.00000387, 0.00002318, 0.00000821, -0.00000020, 0.00000686, 0.00000707, -0.00000297, 0.00000835, 0.00000110, 0.00000686, -0.00000078, -0.00000926, 0.00000126, 0.00000552, -0.00001049, 0.00000486, 0.00000002, -0.00000470, -0.00001539, -0.00006038, 0.00001092, 0.00011648, 0.00004028, -0.00001149, -0.00001455, 0.00000428, -0.00000536, 0.00002078, },
		{ 0.00000706, 0.00001063, 0.00000835, 0.00000237, 0.00001654, 0.00001028, 0.00002278, 0.00002622, 0.00003751, 0.00000287, -0.00002491, -0.00003605, -0.00001412, -0.00000698, -0.00000536, -0.00001009, 0.00000054, -0.00001056, 0.00000157, -0.00000335, -0.00000514, -0.00000478, 0.00000125, -0.00000757, -0.00000043, -0.00000457, 0.00000646, 0.00000004, -0.00000277, -0.00000493, 0.00000193, -0.00000548, -0.00000114, -0.00000389, 0.00000264, -0.00000617, 0.00000151, 0.00000114, -0.00000199, -0.00000401, 0.00000078, -0.00000476, 0.00000900, -0.00000016, -0.00000762, -0.00000886, -0.00000054, 0.00000901, 0.00001047, -0.00001383, -0.00000463, 0.00000626, -0.00000275, -0.00000017, 0.00000271, -0.00000753, 0.00000702, -0.00000203, -0.00001700, 0.00001115, 0.00000634, 0.00000661, -0.00001124, -0.00000378, 0.00000087, 0.00000182, 0.00001166, -0.00001584, 0.00001385, -0.00001850, 0.00001029, -0.00000251, -0.00000254, 0.00000055, 0.00000315, -0.00000712, 0.00000354, -0.00000935, 0.00000544, -0.00001011, 0.00000474, 0.00000647, -0.00000159, -0.00000434, 0.00000266, -0.00000875, 0.00000337, -0.00001266, 0.00000629, 0.00000489, -0.00000842, 0.00000091, 0.00000267, -0.00001873, 0.00000724, -0.00001367, -0.00001854, -0.00000008, -0.00000028, 0.00000188, 0.00002391, 0.00000194, 0.00000160, 0.00000574, 0.00000589, -0.00000217, 0.00000671, 0.00000199, 0.00000609, -0.00000258, -0.00000936, 0.00000142, 0.00000557, -0.00001138, 0.00000539, 0.00000074, -0.00000757, -0.00001494, -0.00006339, 0.00000760, 0.00011939, 0.00003990, -0.00000829, -0.00001349, 0.00000453, -0.00000487, 0.00002260, -0.00000094, },
		{ 0.00001102, 0.00000876, 0.00000360, 0.00001896, 0.00000963, 0.00002741, 0.00003354, 0.00003268, -0.00000890, -0.00003180, -0.00003392, -0.00000838, -0.00000928, -0.00000569, -0.00000678, -0.00000238, -0.00000940, 0.00000381, -0.00000698, -0.00000405, -0.00000418, -0.00000007, -0.00000641, -0.00000135, -0.00000280, 0.00000772, -0.00000319, -0.00000255, -0.00000439, 0.00000165, -0.00000520, -0.00000168, -0.00000250, 0.00000155, -0.00000668, 0.00000388, -0.00000123, -0.00000241, -0.00000228, -0.00000108, -0.00000203, 0.00000991, -0.00000462, -0.00000805, -0.00000785, 0.00000250, 0.00001191, 0.00000361, -0.00001569, 0.00000170, 0.00000404, -0.00000385, 0.00000286, -0.00000132, -0.00000479, 0.00000868, -0.00000968, -0.00001137, 0.00001533, 0.00000299, 0.00000503, -0.00001415, 0.00000074, -0.00000026, 0.00000593, 0.00000546, -0.00001268, 0.00001288, -0.00002068, 0.00001712, -0.00001019, 0.00000176, 0.00000084, 0.00000024, -0.00000495, 0.00000161, -0.00000763, 0.00000535, -0.00001240, 0.00001144, 0.00000237, -0.00000355, -0.00000123, -0.00000063, -0.00000597, 0.00000129, -0.00001373, 0.00001397, -0.00000235, -0.00000758, 0.00000605, -0.00000592, -0.00001262, 0.00000826, -0.00002353, -0.00000910, 0.00000196, -0.00000371, 0.00000927, 0.00002049, -0.00000205, 0.00000397, 0.00000330, 0.00000606, -0.00000215, 0.00000545, 0.00000317, 0.00000427, -0.00000372, -0.00000907, 0.00000136, 0.00000532, -0.00001186, 0.00000647, 0.00000026, -0.00001026, -0.00001382, -0.00006676, 0.00000442, 0.00012209, 0.00003921, -0.00000435, -0.00001283, 0.00000490, -0.00000389, 0.00002405, -0.00000015, 0.00000769, },
		{ 0.00000941, 0.00000554, 0.00002071, 0.00000937, 0.00003432, 0.00003906, 0.00002374, -0.00002139, -0.00003569, -0.00002896, -0.00000483, -0.00001247, -0.00000378, -0.00000527, -0.00000518, -0.00000664, 0.00000381, -0.00000980, -0.00000202, -0.00000490, -0.00000051, -0.00000563, -0.00000204, -0.00000030, 0.00000723, -0.00000590, -0.00000160, -0.00000438, 0.00000173, -0.00000530, -0.00000169, -0.00000138, -0.00000021, -0.00000573, 0.00000515, -0.00000410, -0.00000126, -0.00000184, -0.00000225, 0.00000154, 0.00000852, -0.00000851, -0.00000748, -0.00000667, 0.00000654, 0.00001203, -0.00000396, -0.00001360, 0.00000645, 0.00000018, -0.00000236, 0.00000377, -0.00000495, 0.00000015, 0.00000625, -0.00001503, -0.00000303, 0.00001576, 0.00000012, 0.00000308, -0.00001554, 0.00000521, -0.00000152, 0.00000915, -0.00000113, -0.00000739, 0.00000885, -0.00001902, 0.00002107, -0.00001809, 0.00000915, -0.00000270, -0.00000025, -0.00000357, -0.00000052, -0.00000456, 0.00000283, -0.00001164, 0.00001696, -0.00000430, -0.00000188, -0.00000039, -0.00000314, -0.00000256, -0.00000321, -0.00001040, 0.00001846, -0.00001118, -0.00000148, 0.00000583, -0.00001263, -0.00000315, 0.00000249, -0.00002853, 0.00000247, -0.00000155, -0.00000333, 0.00001554, 0.00001417, -0.00000284, 0.00000532, 0.00000064, 0.00000715, -0.00000310, 0.00000515, 0.00000382, 0.00000184, -0.00000395, -0.00000874, 0.00000110, 0.00000480, -0.00001160, 0.00000755, -0.00000146, -0.00001216, -0.00001251, -0.00007036, 0.00000139, 0.00012427, 0.00003874, 0.00000003, -0.00001269, 0.00000581, -0.00000278, 0.00002507, 0.00000091, 0.00000835, 0.00001112, },
		{ 0.00000793, 0.00002165, 0.00001032, 0.00004310, 0.00004127, 0.00001080, -0.00003296, -0.00003564, -0.00002291, -0.00000403, -0.00001492, -0.00000071, -0.00000578, -0.00000659, -0.00000363, 0.00000194, -0.00001098, -0.00000016, -0.00000637, -0.00000011, -0.00000547, -0.00000188, 0.00000203, 0.00000530, -0.00000751, -0.00000051, -0.00000461, 0.00000197, -0.00000569, -0.00000105, -0.00000108, -0.00000185, -0.00000371, 0.00000477, -0.00000635, 0.00000070, -0.00000273, -0.00000192, 0.00000474, 0.00000519, -0.00001108, -0.00000644, -0.00000489, 0.00001073, 0.00000893, -0.00001014, -0.00000849, 0.00000825, -0.00000342, 0.00000077, 0.00000204, -0.00000633, 0.00000517, 0.00000028, -0.00001607, 0.00000539, 0.00001301, -0.00000158, 0.00000053, -0.00001499, 0.00000893, -0.00000270, 0.00001114, -0.00000703, -0.00000125, 0.00000267, -0.00001347, 0.00002049, -0.00002346, 0.00001740, -0.00000950, 0.00000259, -0.00000452, -0.00000118, -0.00000171, -0.00000117, -0.00000730, 0.00001897, -0.00001101, 0.00000253, -0.00000260, -0.00000317, -0.00000067, -0.00000805, -0.00000321, 0.00001747, -0.00001770, 0.00000726, -0.00000032, -0.00001385, 0.00000530, -0.00000826, -0.00002608, 0.00001160, -0.00000828, 0.00000165, 0.00001813, 0.00000737, -0.00000097, 0.00000481, -0.00000111, 0.00000825, -0.00000453, 0.00000587, 0.00000331, -0.00000046, -0.00000344, -0.00000854, 0.00000062, 0.00000413, -0.00001037, 0.00000793, -0.00000398, -0.00001295, -0.00001157, -0.00007381, -0.00000177, 0.00012589, 0.00003896, 0.00000435, -0.00001284, 0.00000747, -0.00000198, 0.00002592, 0.00000233, 0.00000884, 0.00001103, 0.00001052, },
		{ 0.00002189, 0.00001334, 0.00005277, 0.00003865, -0.00000524, -0.00004164, -0.00003184, -0.00001769, -0.00000532, -0.00001564, 0.00000215, -0.00000760, -0.00000617, -0.00000165, -0.00000074, -0.00001050, 0.00000074, -0.00000785, 0.00000075, -0.00000577, -0.00000070, 0.00000338, 0.00000268, -0.00000795, 0.00000030, -0.00000478, 0.00000208, -0.00000604, -0.00000001, -0.00000180, -0.00000258, -0.00000150, 0.00000293, -0.00000721, 0.00000245, -0.00000428, 0.00000010, 0.00000643, 0.00000083, -0.00001211, -0.00000528, -0.00000215, 0.00001383, 0.00000312, -0.00001323, -0.00000219, 0.00000693, -0.00000515, 0.00000387, -0.00000157, -0.00000458, 0.00000794, -0.00000716, -0.00001232, 0.00001145, 0.00000858, -0.00000220, -0.00000249, -0.00001237, 0.00001125, -0.00000351, 0.00001174, -0.00001140, 0.00000440, -0.00000412, -0.00000518, 0.00001502, -0.00002414, 0.00002364, -0.00001756, 0.00000830, -0.00000844, 0.00000102, -0.00000099, -0.00000458, -0.00000049, 0.00001636, -0.00001507, 0.00000744, -0.00000701, -0.00000026, -0.00000201, -0.00001036, 0.00000524, 0.00001094, -0.00001872, 0.00001446, -0.00001005, -0.00000828, 0.00000835, -0.00001960, -0.00001676, 0.00001476, -0.00001438, 0.00000969, 0.00001593, 0.00000256, 0.00000194, 0.00000272, -0.00000146, 0.00000850, -0.00000558, 0.00000701, 0.00000151, -0.00000196, -0.00000260, -0.00000845, -0.00000019, 0.00000366, -0.00000828, 0.00000700, -0.00000654, -0.00001277, -0.00001133, -0.00007668, -0.00000556, 0.00012729, 0.00004006, 0.00000810, -0.00001276, 0.00000969, -0.00000178, 0.00002698, 0.00000392, 0.00000905, 0.00001096, 0.00001222, 0.00001044, },
		{ 0.00001911, 0.00006171, 0.00003004, -0.00002248, -0.00004580, -0.00002559, -0.00001462, -0.00000738, -0.00001463, 0.00000376, -0.00000955, -0.00000437, -0.00000134, -0.00000305, -0.00000902, 0.00000049, -0.00000875, 0.00000156, -0.00000604, 0.00000119, 0.00000332, 0.00000017, -0.00000755, 0.00000074, -0.00000479, 0.00000190, -0.00000598, 0.00000092, -0.00000323, -0.00000206, -0.00000006, 0.00000039, -0.00000657, 0.00000318, -0.00000553, 0.00000326, 0.00000596, -0.00000344, -0.00001185, -0.00000400, 0.00000151, 0.00001464, -0.00000393, -0.00001251, 0.00000316, 0.00000356, -0.00000440, 0.00000530, -0.00000533, -0.00000028, 0.00000695, -0.00001332, -0.00000503, 0.00001373, 0.00000421, -0.00000243, -0.00000535, -0.00000809, 0.00001181, -0.00000373, 0.00001098, -0.00001384, 0.00000855, -0.00000991, 0.00000389, 0.00000578, -0.00001932, 0.00002536, -0.00002414, 0.00001508, -0.00001464, 0.00000638, -0.00000383, -0.00000493, 0.00000617, 0.00000990, -0.00001486, 0.00001029, -0.00001150, 0.00000442, -0.00000675, -0.00000794, 0.00001119, 0.00000150, -0.00001346, 0.00001628, -0.00001892, 0.00000210, 0.00000373, -0.00002638, -0.00000421, 0.00001109, -0.00001631, 0.00001765, 0.00000987, 0.00000116, 0.00000401, 0.00000019, -0.00000077, 0.00000765, -0.00000552, 0.00000763, -0.00000106, -0.00000237, -0.00000185, -0.00000836, -0.00000146, 0.00000388, -0.00000586, 0.00000452, -0.00000827, -0.00001221, -0.00001173, -0.00007874, -0.00001051, 0.00012906, 0.00004180, 0.00001110, -0.00001192, 0.00001198, -0.00000210, 0.00002853, 0.00000539, 0.00000908, 0.00001107, 0.00001453, 0.00001271, 0.00002186, },
		{ 0.00006762, 0.00001519, -0.00003822, -0.00004484, -0.00001887, -0.00001394, -0.00000887, -0.00001269, 0.00000387, -0.00001063, -0.00000215, -0.00000258, -0.00000418, -0.00000742, -0.00000059, -0.00000892, 0.00000201, -0.00000581, 0.00000316, 0.00000194, -0.00000162, -0.00000675, 0.00000095, -0.00000472, 0.00000147, -0.00000536, 0.00000124, -0.00000473, -0.00000048, 0.00000001, -0.00000184, -0.00000498, 0.00000266, -0.00000557, 0.00000650, 0.00000339, -0.00000671, -0.00001081, -0.00000233, 0.00000550, 0.00001249, -0.00001029, -0.00000847, 0.00000597, -0.00000013, -0.00000175, 0.00000423, -0.00000735, 0.00000479, 0.00000220, -0.00001585, 0.00000339, 0.00001226, 0.00000112, -0.00000305, -0.00000717, -0.00000303, 0.00001070, -0.00000326, 0.00000906, -0.00001433, 0.00001075, -0.00001352, 0.00001159, -0.00000496, -0.00000986, 0.00002134, -0.00002668, 0.00002041, -0.00002123, 0.00001383, -0.00001026, -0.00000051, 0.00000952, 0.00000223, -0.00001069, 0.00000948, -0.00001369, 0.00000846, -0.00001312, -0.00000050, 0.00001143, -0.00000658, -0.00000413, 0.00001122, -0.00002246, 0.00001272, -0.00000736, -0.00002509, 0.00000632, 0.00000289, -0.00001253, 0.00002223, 0.00000246, 0.00000293, 0.00000410, -0.00000150, -0.00000003, 0.00000620, -0.00000425, 0.00000694, -0.00000354, -0.00000194, -0.00000134, -0.00000828, -0.00000319, 0.00000523, -0.00000397, 0.00000088, -0.00000864, -0.00001203, -0.00001227, -0.00008022, -0.00001683, 0.00013183, 0.00004360, 0.00001358, -0.00001005, 0.00001371, -0.00000253, 0.00003056, 0.00000647, 0.00000916, 0.00001145, 0.00001744, 0.00001441, 0.00002229, 0.00002791, },
		{ -0.00000475, -0.00004964, -0.00003955, -0.00001356, -0.00001492, -0.00000908, -0.00001083, 0.00000295, -0.00001043, -0.00000052, -0.00000463, -0.00000400, -0.00000645, -0.00000191, -0.00000855, 0.00000211, -0.00000485, 0.00000452, -0.00000027, -0.00000247, -0.00000592, 0.00000110, -0.00000479, 0.00000106, -0.00000434, 0.00000071, -0.00000560, 0.00000147, -0.00000130, -0.00000297, -0.00000332, 0.00000131, -0.00000403, 0.00000868, -0.00000059, -0.00000849, -0.00000942, 0.00000006, 0.00000878, 0.00000760, -0.00001415, -0.00000267, 0.00000571, -0.00000244, 0.00000133, 0.00000109, -0.00000647, 0.00000832, -0.00000473, -0.00001373, 0.00001031, 0.00000830, -0.00000034, -0.00000451, -0.00000717, 0.00000166, 0.00000844, -0.00000223, 0.00000633, -0.00001319, 0.00001106, -0.00001449, 0.00001630, -0.00001455, 0.00000194, 0.00001210, -0.00002379, 0.00002195, -0.00002573, 0.00002109, -0.00001871, 0.00000873, 0.00000711, -0.00000314, -0.00000485, 0.00000525, -0.00001223, 0.00000939, -0.00001797, 0.00000974, 0.00000499, -0.00000918, 0.00000477, 0.00000112, -0.00001836, 0.00001835, -0.00002046, -0.00001558, 0.00001036, -0.00000546, -0.00000430, 0.00002161, -0.00000341, 0.00000625, 0.00000232, -0.00000166, -0.00000025, 0.00000515, -0.00000243, 0.00000479, -0.00000511, -0.00000121, -0.00000094, -0.00000842, -0.00000502, 0.00000780, -0.00000351, -0.00000302, -0.00000771, -0.00001278, -0.00001235, -0.00008169, -0.00002426, 0.00013591, 0.00004483, 0.00001619, -0.00000741, 0.00001453, -0.00000246, 0.00003270, 0.00000718, 0.00000949, 0.00001210, 0.00002086, 0.00001523, 0.00002432, 0.00003920, 0.00006787, },
		{ -0.00005462, -0.00003194, -0.00001068, -0.00001632, -0.00000808, -0.00000980, 0.00000183, -0.00000920, -0.00000003, -0.00000659, -0.00000289, -0.00000640, -0.00000296, -0.00000797, 0.00000213, -0.00000331, 0.00000485, -0.00000269, -0.00000248, -0.00000523, 0.00000126, -0.00000512, 0.00000098, -0.00000333, -0.00000051, -0.00000543, 0.00000294, -0.00000344, -0.00000274, -0.00000240, -0.00000004, -0.00000118, 0.00000895, -0.00000497, -0.00000881, -0.00000786, 0.00000324, 0.00001023, 0.00000108, -0.00001450, 0.00000285, 0.00000309, -0.00000248, 0.00000320, -0.00000268, -0.00000284, 0.00000855, -0.00001129, -0.00000762, 0.00001386, 0.00000368, -0.00000069, -0.00000660, -0.00000508, 0.00000493, 0.00000585, -0.00000101, 0.00000326, -0.00001093, 0.00000999, -0.00001318, 0.00001735, -0.00002077, 0.00001311, -0.00000028, -0.00001572, 0.00001849, -0.00002604, 0.00002552, -0.00002629, 0.00002077, -0.00000157, -0.00000322, -0.00000073, -0.00000020, -0.00000772, 0.00000608, -0.00001822, 0.00001874, -0.00000611, -0.00000442, 0.00000849, -0.00000940, -0.00000791, 0.00001564, -0.00002973, -0.00000140, 0.00000628, -0.00000950, 0.00000497, 0.00001626, -0.00000580, 0.00000902, -0.00000020, -0.00000060, -0.00000177, 0.00000541, -0.00000118, 0.00000174, -0.00000547, -0.00000065, -0.00000043, -0.00000908, -0.00000633, 0.00001110, -0.00000496, -0.00000603, -0.00000618, -0.00001445, -0.00001154, -0.00008384, -0.00003212, 0.00014106, 0.00004520, 0.00001964, -0.00000467, 0.00001460, -0.00000153, 0.00003445, 0.00000781, 0.00001007, 0.00001314, 0.00002459, 0.00001514, 0.00002928, 0.00005130, 0.00006019, -0.00002702, },
		{ -0.00002453, -0.00001010, -0.00001708, -0.00000654, -0.00000974, 0.00000120, -0.00000761, -0.00000064, -0.00000782, -0.00000151, -0.00000713, -0.00000345, -0.00000747, 0.00000239, -0.00000159, 0.00000407, -0.00000477, -0.00000193, -0.00000468, 0.00000135, -0.00000564, 0.00000138, -0.00000278, -0.00000191, -0.00000430, 0.00000332, -0.00000554, -0.00000147, -0.00000261, -0.00000049, 0.00000213, 0.00000711, -0.00000872, -0.00000805, -0.00000603, 0.00000682, 0.00000906, -0.00000532, -0.00001148, 0.00000639, -0.00000037, -0.00000050, 0.00000283, -0.00000524, 0.00000211, 0.00000499, -0.00001499, 0.00000044, 0.00001353, -0.00000002, -0.00000087, -0.00000853, -0.00000133, 0.00000619, 0.00000373, -0.00000012, 0.00000041, -0.00000812, 0.00000822, -0.00001050, 0.00001518, -0.00002252, 0.00002094, -0.00001281, -0.00000429, 0.00001034, -0.00002126, 0.00002510, -0.00002988, 0.00003201, -0.00001462, 0.00000307, -0.00000131, -0.00000364, -0.00000265, -0.00000033, -0.00001251, 0.00002236, -0.00001735, 0.00000616, 0.00000436, -0.00001495, 0.00000425, 0.00000499, -0.00003056, 0.00001163, -0.00000388, -0.00000684, 0.00001128, 0.00000873, -0.00000459, 0.00000978, -0.00000199, 0.00000060, -0.00000407, 0.00000714, -0.00000146, -0.00000113, -0.00000496, -0.00000041, 0.00000022, -0.00001043, -0.00000642, 0.00001411, -0.00000806, -0.00000734, -0.00000506, -0.00001648, -0.00000999, -0.00008708, -0.00003961, 0.00014658, 0.00004501, 0.00002427, -0.00000266, 0.00001456, 0.00000019, 0.00003547, 0.00000876, 0.00001072, 0.00001482, 0.00002813, 0.00001464, 0.00003828, 0.00006124, 0.00004362, -0.00004769, -0.00005269, },
		{ -0.00001078, -0.00001679, -0.00000521, -0.00001026, 0.00000131, -0.00000631, -0.00000190, -0.00000809, -0.00000047, -0.00000827, -0.00000341, -0.00000712, 0.00000308, -0.00000020, 0.00000246, -0.00000620, -0.00000108, -0.00000423, 0.00000124, -0.00000613, 0.00000216, -0.00000294, -0.00000287, -0.00000270, 0.00000246, -0.00000682, 0.00000009, -0.00000374, 0.00000050, 0.00000479, 0.00000358, -0.00001117, -0.00000666, -0.00000372, 0.00000998, 0.00000524, -0.00000984, -0.00000629, 0.00000710, -0.00000302, 0.00000231, 0.00000030, -0.00000522, 0.00000621, -0.00000126, -0.00001435, 0.00000783, 0.00001021, -0.00000200, -0.00000171, -0.00000932, 0.00000306, 0.00000551, 0.00000259, 0.00000002, -0.00000169, -0.00000532, 0.00000637, -0.00000758, 0.00001109, -0.00002007, 0.00002382, -0.00002257, 0.00000774, -0.00000072, -0.00001206, 0.00001929, -0.00002726, 0.00003850, -0.00002815, 0.00001426, -0.00000779, -0.00000229, -0.00000043, -0.00000659, -0.00000229, 0.00001836, -0.00002367, 0.00001799, -0.00000660, -0.00001200, 0.00001205, -0.00000930, -0.00002191, 0.00001803, -0.00001533, 0.00000176, 0.00001201, 0.00000233, -0.00000138, 0.00000844, -0.00000220, 0.00000079, -0.00000597, 0.00000961, -0.00000351, -0.00000288, -0.00000427, -0.00000019, 0.00000071, -0.00001217, -0.00000494, 0.00001566, -0.00001178, -0.00000688, -0.00000517, -0.00001796, -0.00000840, -0.00009122, -0.00004625, 0.00015163, 0.00004507, 0.00002978, -0.00000186, 0.00001524, 0.00000212, 0.00003585, 0.00001028, 0.00001130, 0.00001744, 0.00003077, 0.00001497, 0.00005151, 0.00006518, 0.00001926, -0.00006267, -0.00004526, -0.00001940, },
		{ -0.00001578, -0.00000450, -0.00001081, 0.00000196, -0.00000568, -0.00000319, -0.00000764, -0.00000009, -0.00000939, -0.00000300, -0.00000685, 0.00000416, 0.00000049, 0.00000044, -0.00000691, -0.00000012, -0.00000389, 0.00000086, -0.00000632, 0.00000299, -0.00000374, -0.00000299, -0.00000136, 0.00000073, -0.00000690, 0.00000116, -0.00000511, 0.00000286, 0.00000586, -0.00000076, -0.00001210, -0.00000496, -0.00000082, 0.00001174, -0.00000043, -0.00001128, -0.00000075, 0.00000530, -0.00000375, 0.00000440, -0.00000323, -0.00000240, 0.00000755, -0.00000794, -0.00000950, 0.00001241, 0.00000560, -0.00000249, -0.00000339, -0.00000827, 0.00000685, 0.00000356, 0.00000244, -0.00000078, -0.00000268, -0.00000298, 0.00000485, -0.00000532, 0.00000671, -0.00001485, 0.00002163, -0.00002755, 0.00001756, -0.00001210, -0.00000043, 0.00000930, -0.00001810, 0.00003733, -0.00003760, 0.00002665, -0.00001882, 0.00000479, -0.00000376, -0.00000855, 0.00000833, 0.00000788, -0.00002185, 0.00002540, -0.00001983, -0.00000087, 0.00001087, -0.00002066, -0.00000703, 0.00001520, -0.00002269, 0.00001268, 0.00000713, -0.00000048, 0.00000147, 0.00000619, -0.00000108, -0.00000051, -0.00000639, 0.00001150, -0.00000667, -0.00000315, -0.00000401, 0.00000040, 0.00000050, -0.00001353, -0.00000227, 0.00001493, -0.00001470, -0.00000540, -0.00000667, -0.00001823, -0.00000772, -0.00009543, -0.00005224, 0.00015575, 0.00004634, 0.00003527, -0.00000206, 0.00001717, 0.00000350, 0.00003611, 0.00001224, 0.00001188, 0.00002105, 0.00003188, 0.00001810, 0.00006750, 0.00005946, -0.00000938, -0.00006901, -0.00003524, -0.00001739, -0.00001136, },
		{ -0.00000430, -0.00001102, 0.00000272, -0.00000567, -0.00000409, -0.00000689, -0.00000038, -0.00001020, -0.00000239, -0.00000650, 0.00000538, 0.00000036, -0.00000156, -0.00000698, 0.00000087, -0.00000377, 0.00000030, -0.00000607, 0.00000347, -0.00000477, -0.00000225, -0.00000088, -0.00000113, -0.00000596, 0.00000127, -0.00000586, 0.00000595, 0.00000490, -0.00000489, -0.00001172, -0.00000303, 0.00000244, 0.00001129, -0.00000648, -0.00000945, 0.00000337, 0.00000224, -0.00000248, 0.00000460, -0.00000606, 0.00000215, 0.00000526, -0.00001256, -0.00000209, 0.00001322, 0.00000147, -0.00000236, -0.00000543, -0.00000533, 0.00000902, 0.00000133, 0.00000286, -0.00000233, -0.00000242, -0.00000136, 0.00000381, -0.00000420, 0.00000347, -0.00000890, 0.00001563, -0.00002721, 0.00002317, -0.00002123, 0.00001100, -0.00000236, -0.00000417, 0.00002781, -0.00003938, 0.00003562, -0.00003083, 0.00001592, -0.00001310, -0.00000313, 0.00001409, -0.00000463, -0.00001234, 0.00002430, -0.00002904, 0.00001400, -0.00000001, -0.00002327, 0.00000773, 0.00000474, -0.00002259, 0.00002128, -0.00000087, 0.00000079, 0.00000219, 0.00000468, 0.00000014, -0.00000270, -0.00000505, 0.00001164, -0.00000967, -0.00000238, -0.00000431, 0.00000151, -0.00000082, -0.00001351, 0.00000046, 0.00001201, -0.00001568, -0.00000418, -0.00000889, -0.00001726, -0.00000864, -0.00009868, -0.00005840, 0.00015918, 0.00004936, 0.00003962, -0.00000237, 0.00002023, 0.00000385, 0.00003693, 0.00001415, 0.00001301, 0.00002515, 0.00003149, 0.00002613, 0.00008280, 0.00004194, -0.00003698, -0.00006603, -0.00002601, -0.00001791, -0.00001095, -0.00001476, },
		{ -0.00001089, 0.00000323, -0.00000597, -0.00000444, -0.00000625, -0.00000113, -0.00001057, -0.00000170, -0.00000598, 0.00000645, -0.00000052, -0.00000321, -0.00000654, 0.00000178, -0.00000395, -0.00000021, -0.00000544, 0.00000334, -0.00000553, -0.00000100, -0.00000148, -0.00000240, -0.00000458, 0.00000044, -0.00000538, 0.00000879, 0.00000207, -0.00000798, -0.00001041, -0.00000079, 0.00000546, 0.00000841, -0.00001126, -0.00000519, 0.00000503, -0.00000053, -0.00000011, 0.00000270, -0.00000672, 0.00000651, -0.00000005, -0.00001332, 0.00000546, 0.00001071, -0.00000108, -0.00000259, -0.00000689, -0.00000120, 0.00000920, -0.00000031, 0.00000318, -0.00000412, -0.00000111, -0.00000059, 0.00000318, -0.00000414, 0.00000214, -0.00000419, 0.00000800, -0.00002247, 0.00002389, -0.00002636, 0.00001989, -0.00001271, 0.00001117, 0.00001184, -0.00003217, 0.00003737, -0.00003935, 0.00002731, -0.00002608, 0.00000999, 0.00001087, -0.00001301, 0.00000047, 0.00001449, -0.00002925, 0.00002575, -0.00001667, -0.00001488, 0.00001577, -0.00000820, -0.00001522, 0.00002412, -0.00000829, 0.00000459, 0.00000050, 0.00000497, 0.00000009, -0.00000439, -0.00000280, 0.00000968, -0.00001134, -0.00000148, -0.00000469, 0.00000276, -0.00000312, -0.00001142, 0.00000170, 0.00000801, -0.00001442, -0.00000436, -0.00001071, -0.00001582, -0.00001110, -0.00010021, -0.00006584, 0.00016287, 0.00005385, 0.00004215, -0.00000163, 0.00002359, 0.00000331, 0.00003868, 0.00001553, 0.00001541, 0.00002864, 0.00003076, 0.00004027, 0.00009229, 0.00001330, -0.00005782, -0.00005567, -0.00002016, -0.00001940, -0.00000948, -0.00001431, -0.00000417, },
		{ 0.00000341, -0.00000626, -0.00000436, -0.00000597, -0.00000209, -0.00001048, -0.00000098, -0.00000531, 0.00000711, -0.00000191, -0.00000435, -0.00000573, 0.00000250, -0.00000446, -0.00000043, -0.00000473, 0.00000265, -0.00000563, 0.00000014, -0.00000292, -0.00000266, -0.00000348, -0.00000080, -0.00000353, 0.00001039, -0.00000192, -0.00000959, -0.00000859, 0.00000183, 0.00000742, 0.00000362, -0.00001349, -0.00000007, 0.00000421, -0.00000180, 0.00000199, -0.00000047, -0.00000468, 0.00000869, -0.00000647, -0.00000989, 0.00001084, 0.00000638, -0.00000191, -0.00000371, -0.00000686, 0.00000298, 0.00000770, -0.00000094, 0.00000285, -0.00000549, 0.00000078, -0.00000059, 0.00000279, -0.00000467, 0.00000269, -0.00000200, 0.00000103, -0.00001527, 0.00002037, -0.00002696, 0.00002485, -0.00001938, 0.00002416, -0.00000674, -0.00001751, 0.00003034, -0.00004072, 0.00003446, -0.00003813, 0.00002772, -0.00000221, -0.00001202, 0.00000984, 0.00000024, -0.00001941, 0.00002828, -0.00003192, 0.00000180, 0.00001328, -0.00001730, -0.00000430, 0.00002060, -0.00001211, 0.00000830, -0.00000235, 0.00000684, -0.00000186, -0.00000422, -0.00000120, 0.00000643, -0.00001127, -0.00000126, -0.00000451, 0.00000335, -0.00000550, -0.00000742, 0.00000025, 0.00000466, -0.00001174, -0.00000629, -0.00001114, -0.00001512, -0.00001425, -0.00010008, -0.00007528, 0.00016807, 0.00005871, 0.00004304, 0.00000096, 0.00002613, 0.00000271, 0.00004109, 0.00001639, 0.00001959, 0.00003021, 0.00003218, 0.00005966, 0.00009059, -0.00002241, -0.00006763, -0.00004194, -0.00001855, -0.00002016, -0.00000767, -0.00001450, -0.00000370, -0.00001073, },
		{ -0.00000637, -0.00000408, -0.00000615, -0.00000302, -0.00000998, -0.00000026, -0.00000460, 0.00000723, -0.00000354, -0.00000487, -0.00000467, 0.00000287, -0.00000515, -0.00000023, -0.00000429, 0.00000168, -0.00000499, 0.00000065, -0.00000461, -0.00000196, -0.00000318, -0.00000168, -0.00000073, 0.00001011, -0.00000612, -0.00000972, -0.00000650, 0.00000468, 0.00000759, -0.00000189, -0.00001271, 0.00000425, 0.00000182, -0.00000124, 0.00000259, -0.00000339, -0.00000064, 0.00000753, -0.00001160, -0.00000343, 0.00001275, 0.00000203, -0.00000174, -0.00000552, -0.00000493, 0.00000603, 0.00000537, -0.00000067, 0.00000166, -0.00000590, 0.00000267, -0.00000114, 0.00000247, -0.00000517, 0.00000440, -0.00000263, -0.00000348, -0.00000786, 0.00001425, -0.00002368, 0.00002581, -0.00002137, 0.00003187, -0.00002342, 0.00000082, 0.00001581, -0.00003356, 0.00003397, -0.00004414, 0.00004431, -0.00002187, 0.00000009, 0.00000961, -0.00001119, -0.00000352, 0.00001936, -0.00003852, 0.00002000, 0.00000117, -0.00001800, 0.00000483, 0.00001317, -0.00001147, 0.00000981, -0.00000445, 0.00000889, -0.00000505, -0.00000175, -0.00000178, 0.00000344, -0.00001006, -0.00000182, -0.00000341, 0.00000249, -0.00000663, -0.00000272, -0.00000409, 0.00000355, -0.00000921, -0.00000928, -0.00000995, -0.00001615, -0.00001677, -0.00009922, -0.00008655, 0.00017556, 0.00006250, 0.00004342, 0.00000530, 0.00002711, 0.00000310, 0.00004330, 0.00001746, 0.00002523, 0.00002915, 0.00003896, 0.00008058, 0.00007390, -0.00005837, -0.00006511, -0.00002940, -0.00002011, -0.00001927, -0.00000647, -0.00001485, -0.00000283, -0.00001085, 0.00000345, },
		{ -0.00000377, -0.00000674, -0.00000370, -0.00000912, 0.00000042, -0.00000397, 0.00000679, -0.00000515, -0.00000473, -0.00000357, 0.00000278, -0.00000577, 0.00000033, -0.00000433, 0.00000088, -0.00000387, 0.00000025, -0.00000591, -0.00000073, -0.00000377, -0.00000154, 0.00000215, 0.00000790, -0.00000960, -0.00000865, -0.00000422, 0.00000736, 0.00000564, -0.00000668, -0.00000943, 0.00000658, -0.00000077, 0.00000059, 0.00000127, -0.00000458, 0.00000375, 0.00000324, -0.00001349, 0.00000387, 0.00001120, -0.00000100, -0.00000159, -0.00000724, -0.00000146, 0.00000723, 0.00000321, -0.00000006, -0.00000014, -0.00000513, 0.00000401, -0.00000194, 0.00000213, -0.00000513, 0.00000618, -0.00000540, -0.00000477, -0.00000220, 0.00000753, -0.00001795, 0.00002373, -0.00001923, 0.00003312, -0.00003448, 0.00001808, -0.00000263, -0.00001928, 0.00002501, -0.00004064, 0.00005351, -0.00004153, 0.00002027, -0.00000278, -0.00001266, 0.00001068, 0.00000229, -0.00003267, 0.00003183, -0.00001515, -0.00000977, 0.00000774, 0.00000603, -0.00000797, 0.00000866, -0.00000451, 0.00000940, -0.00000779, 0.00000207, -0.00000500, 0.00000220, -0.00000895, -0.00000248, -0.00000184, 0.00000012, -0.00000537, 0.00000079, -0.00001021, 0.00000529, -0.00000840, -0.00001186, -0.00000797, -0.00001906, -0.00001760, -0.00009901, -0.00009853, 0.00018508, 0.00006422, 0.00004487, 0.00001031, 0.00002667, 0.00000512, 0.00004438, 0.00002005, 0.00003090, 0.00002628, 0.00005364, 0.00009672, 0.00004196, -0.00008687, -0.00005259, -0.00002155, -0.00002257, -0.00001699, -0.00000647, -0.00001469, -0.00000193, -0.00001133, 0.00000358, -0.00000632, },
		{ -0.00000770, -0.00000397, -0.00000803, 0.00000097, -0.00000350, 0.00000584, -0.00000650, -0.00000397, -0.00000266, 0.00000226, -0.00000607, 0.00000098, -0.00000486, 0.00000058, -0.00000275, -0.00000089, -0.00000638, 0.00000039, -0.00000489, -0.00000010, 0.00000416, 0.00000423, -0.00001174, -0.00000682, -0.00000178, 0.00000925, 0.00000186, -0.00000947, -0.00000489, 0.00000660, -0.00000235, 0.00000252, -0.00000145, -0.00000329, 0.00000657, -0.00000271, -0.00001138, 0.00000970, 0.00000741, -0.00000225, -0.00000221, -0.00000787, 0.00000261, 0.00000656, 0.00000190, 0.00000019, -0.00000196, -0.00000340, 0.00000448, -0.00000272, 0.00000181, -0.00000433, 0.00000704, -0.00000901, -0.00000312, 0.00000062, 0.00000199, -0.00001142, 0.00002010, -0.00001467, 0.00002873, -0.00003816, 0.00003019, -0.00002034, -0.00000154, 0.00000968, -0.00002727, 0.00005099, -0.00005371, 0.00004149, -0.00002462, -0.00000060, 0.00001492, -0.00001502, -0.00001628, 0.00003197, -0.00002825, 0.00000365, 0.00000320, 0.00000287, -0.00000463, 0.00000630, -0.00000275, 0.00000756, -0.00000832, 0.00000530, -0.00000995, 0.00000324, -0.00000904, -0.00000218, -0.00000094, -0.00000285, -0.00000155, 0.00000137, -0.00001604, 0.00000909, -0.00000997, -0.00001262, -0.00000677, -0.00002287, -0.00001665, -0.00010049, -0.00010961, 0.00019526, 0.00006401, 0.00004856, 0.00001444, 0.00002592, 0.00000846, 0.00004404, 0.00002531, 0.00003447, 0.00002444, 0.00007639, 0.00010066, -0.00000097, -0.00010174, -0.00003523, -0.00001951, -0.00002360, -0.00001453, -0.00000742, -0.00001361, -0.00000156, -0.00001189, 0.00000391, -0.00000621, -0.00000354, },
		{ -0.00000366, -0.00000689, 0.00000135, -0.00000325, 0.00000444, -0.00000736, -0.00000271, -0.00000222, 0.00000149, -0.00000593, 0.00000142, -0.00000564, 0.00000088, -0.00000212, -0.00000233, -0.00000598, 0.00000087, -0.00000587, 0.00000240, 0.00000455, -0.00000006, -0.00001230, -0.00000458, 0.00000073, 0.00000973, -0.00000285, -0.00000965, -0.00000060, 0.00000491, -0.00000235, 0.00000337, -0.00000429, 0.00000012, 0.00000647, -0.00000810, -0.00000598, 0.00001248, 0.00000310, -0.00000221, -0.00000371, -0.00000679, 0.00000607, 0.00000464, 0.00000160, -0.00000034, -0.00000312, -0.00000124, 0.00000408, -0.00000329, 0.00000164, -0.00000296, 0.00000636, -0.00001203, 0.00000036, 0.00000048, -0.00000121, -0.00000547, 0.00001633, -0.00000978, 0.00002100, -0.00003499, 0.00003497, -0.00003334, 0.00001522, -0.00000777, -0.00000712, 0.00003627, -0.00005299, 0.00005528, -0.00004843, 0.00002275, 0.00000432, -0.00002323, 0.00000333, 0.00002046, -0.00003219, 0.00001602, -0.00000634, 0.00000497, -0.00000406, 0.00000509, -0.00000093, 0.00000414, -0.00000597, 0.00000598, -0.00001468, 0.00000578, -0.00001054, -0.00000023, -0.00000195, -0.00000468, 0.00000363, -0.00000159, -0.00001945, 0.00001306, -0.00001324, -0.00001103, -0.00000777, -0.00002592, -0.00001498, -0.00010364, -0.00011863, 0.00020417, 0.00006330, 0.00005448, 0.00001654, 0.00002643, 0.00001184, 0.00004313, 0.00003333, 0.00003424, 0.00002802, 0.00010366, 0.00008642, -0.00004659, -0.00010054, -0.00001917, -0.00002182, -0.00002202, -0.00001332, -0.00000842, -0.00001180, -0.00000205, -0.00001209, 0.00000430, -0.00000610, -0.00000349, -0.00000887, },
		{ -0.00000595, 0.00000162, -0.00000328, 0.00000275, -0.00000750, -0.00000122, -0.00000238, 0.00000079, -0.00000543, 0.00000143, -0.00000631, 0.00000160, -0.00000220, -0.00000352, -0.00000502, 0.00000054, -0.00000605, 0.00000528, 0.00000303, -0.00000400, -0.00001143, -0.00000220, 0.00000303, 0.00000847, -0.00000728, -0.00000742, 0.00000222, 0.00000263, -0.00000108, 0.00000257, -0.00000584, 0.00000431, 0.00000327, -0.00001091, 0.00000086, 0.00001185, -0.00000024, -0.00000184, -0.00000550, -0.00000398, 0.00000800, 0.00000242, 0.00000195, -0.00000166, -0.00000317, 0.00000067, 0.00000307, -0.00000354, 0.00000175, -0.00000152, 0.00000418, -0.00001333, 0.00000415, -0.00000190, -0.00000172, -0.00000098, 0.00001327, -0.00000626, 0.00001277, -0.00002728, 0.00003256, -0.00003941, 0.00002738, -0.00002265, 0.00001439, 0.00001304, -0.00003833, 0.00005518, -0.00006472, 0.00004958, -0.00001983, -0.00001581, 0.00001631, 0.00000307, -0.00002554, 0.00002178, -0.00001602, 0.00001060, -0.00000691, 0.00000675, -0.00000130, 0.00000129, -0.00000185, 0.00000340, -0.00001723, 0.00000816, -0.00001250, 0.00000297, -0.00000541, -0.00000360, 0.00000784, -0.00000704, -0.00001941, 0.00001513, -0.00001645, -0.00000792, -0.00001144, -0.00002667, -0.00001441, -0.00010712, -0.00012569, 0.00021030, 0.00006423, 0.00006128, 0.00001668, 0.00002925, 0.00001370, 0.00004357, 0.00004252, 0.00003032, 0.00004124, 0.00012800, 0.00005215, -0.00008482, -0.00008565, -0.00000927, -0.00002529, -0.00001832, -0.00001412, -0.00000836, -0.00001006, -0.00000326, -0.00001162, 0.00000446, -0.00000593, -0.00000372, -0.00001007, -0.00000274, },
		{ 0.00000187, -0.00000370, 0.00000097, -0.00000680, 0.00000013, -0.00000311, 0.00000041, -0.00000482, 0.00000101, -0.00000658, 0.00000234, -0.00000293, -0.00000403, -0.00000406, -0.00000039, -0.00000505, 0.00000764, -0.00000016, -0.00000682, -0.00000951, 0.00000015, 0.00000470, 0.00000561, -0.00001028, -0.00000370, 0.00000306, 0.00000090, 0.00000055, 0.00000037, -0.00000522, 0.00000749, -0.00000194, -0.00001007, 0.00000691, 0.00000869, -0.00000189, -0.00000204, -0.00000664, -0.00000017, 0.00000806, 0.00000076, 0.00000230, -0.00000333, -0.00000207, 0.00000179, 0.00000186, -0.00000349, 0.00000223, -0.00000060, 0.00000106, -0.00001248, 0.00000686, -0.00000524, 0.00000013, 0.00000176, 0.00001114, -0.00000489, 0.00000643, -0.00001818, 0.00002502, -0.00003859, 0.00003324, -0.00003156, 0.00003160, -0.00001215, -0.00001361, 0.00003953, -0.00006584, 0.00006927, -0.00004977, 0.00000749, 0.00001465, -0.00001097, -0.00001232, 0.00001893, -0.00002124, 0.00001609, -0.00001141, 0.00001108, -0.00000512, 0.00000143, 0.00000156, -0.00000136, -0.00001684, 0.00000877, -0.00001332, 0.00000595, -0.00001045, 0.00000107, 0.00000869, -0.00001254, -0.00001658, 0.00001419, -0.00001763, -0.00000518, -0.00001681, -0.00002471, -0.00001648, -0.00010882, -0.00013238, 0.00021353, 0.00006853, 0.00006682, 0.00001632, 0.00003412, 0.00001324, 0.00004740, 0.00004993, 0.00002562, 0.00006585, 0.00013975, 0.00000184, -0.00010695, -0.00006342, -0.00000731, -0.00002667, -0.00001433, -0.00001650, -0.00000666, -0.00000932, -0.00000455, -0.00001053, 0.00000419, -0.00000557, -0.00000437, -0.00001096, -0.00000131, -0.00000539, },
		{ -0.00000464, -0.00000053, -0.00000532, 0.00000094, -0.00000415, 0.00000050, -0.00000438, 0.00000037, -0.00000636, 0.00000271, -0.00000395, -0.00000379, -0.00000358, -0.00000135, -0.00000293, 0.00000865, -0.00000430, -0.00000808, -0.00000699, 0.00000236, 0.00000531, 0.00000175, -0.00001116, 0.00000025, 0.00000220, 0.00000037, 0.00000149, -0.00000221, -0.00000250, 0.00000823, -0.00000723, -0.00000588, 0.00001043, 0.00000456, -0.00000206, -0.00000315, -0.00000632, 0.00000354, 0.00000661, 0.00000004, 0.00000207, -0.00000468, -0.00000023, 0.00000191, 0.00000085, -0.00000319, 0.00000302, -0.00000071, -0.00000201, -0.00000975, 0.00000759, -0.00000817, 0.00000352, 0.00000283, 0.00000968, -0.00000545, 0.00000320, -0.00001050, 0.00001534, -0.00003268, 0.00003330, -0.00003355, 0.00004069, -0.00003242, 0.00001376, 0.00001235, -0.00004930, 0.00007266, -0.00007389, 0.00003974, -0.00000390, -0.00001312, -0.00000035, 0.00001024, -0.00002022, 0.00001804, -0.00001445, 0.00001583, -0.00001159, 0.00000573, 0.00000156, -0.00000574, -0.00001451, 0.00000723, -0.00001178, 0.00000678, -0.00001493, 0.00000826, 0.00000495, -0.00001529, -0.00001323, 0.00001086, -0.00001585, -0.00000482, -0.00002179, -0.00002115, -0.00002148, -0.00010700, -0.00014123, 0.00021547, 0.00007652, 0.00006937, 0.00001767, 0.00003930, 0.00001133, 0.00005542, 0.00005255, 0.00002566, 0.00009895, 0.00013013, -0.00005486, -0.00010874, -0.00004194, -0.00001157, -0.00002420, -0.00001219, -0.00001901, -0.00000368, -0.00001012, -0.00000509, -0.00000928, 0.00000354, -0.00000503, -0.00000552, -0.00001119, 0.00000028, -0.00000527, 0.00000219, },
		{ -0.00000136, -0.00000336, 0.00000098, -0.00000515, 0.00000095, -0.00000432, -0.00000019, -0.00000580, 0.00000251, -0.00000480, -0.00000301, -0.00000382, -0.00000174, -0.00000020, 0.00000786, -0.00000850, -0.00000771, -0.00000429, 0.00000422, 0.00000458, -0.00000215, -0.00000988, 0.00000327, 0.00000047, 0.00000099, 0.00000111, -0.00000390, 0.00000129, 0.00000606, -0.00001061, 0.00000019, 0.00001076, 0.00000099, -0.00000159, -0.00000483, -0.00000428, 0.00000612, 0.00000445, 0.00000012, 0.00000106, -0.00000513, 0.00000169, 0.00000118, 0.00000029, -0.00000267, 0.00000386, -0.00000201, -0.00000407, -0.00000608, 0.00000618, -0.00000957, 0.00000741, 0.00000256, 0.00000848, -0.00000707, 0.00000304, -0.00000598, 0.00000636, -0.00002428, 0.00002955, -0.00003007, 0.00004097, -0.00004312, 0.00003579, -0.00001807, -0.00001915, 0.00005598, -0.00008126, 0.00006896, -0.00003409, 0.00000031, 0.00000238, 0.00000209, -0.00001520, 0.00001555, -0.00001354, 0.00001782, -0.00001788, 0.00001303, -0.00000310, -0.00000701, -0.00001260, 0.00000479, -0.00000808, 0.00000430, -0.00001633, 0.00001529, -0.00000257, -0.00001360, -0.00001209, 0.00000728, -0.00001186, -0.00000774, -0.00002429, -0.00001825, -0.00002799, -0.00010139, -0.00015443, 0.00021884, 0.00008666, 0.00006871, 0.00002248, 0.00004240, 0.00001049, 0.00006614, 0.00004936, 0.00003667, 0.00013230, 0.00009490, -0.00010514, -0.00009213, -0.00002798, -0.00001782, -0.00001844, -0.00001309, -0.00001996, -0.00000065, -0.00001219, -0.00000436, -0.00000852, 0.00000284, -0.00000450, -0.00000702, -0.00001052, 0.00000161, -0.00000540, 0.00000254, -0.00000614, },
		{ -0.00000143, 0.00000029, -0.00000576, 0.00000150, -0.00000461, -0.00000043, -0.00000522, 0.00000178, -0.00000512, -0.00000214, -0.00000465, -0.00000112, 0.00000234, 0.00000528, -0.00001185, -0.00000601, -0.00000176, 0.00000551, 0.00000256, -0.00000514, -0.00000706, 0.00000475, -0.00000114, 0.00000213, -0.00000055, -0.00000377, 0.00000457, 0.00000169, -0.00001082, 0.00000610, 0.00000840, -0.00000107, -0.00000143, -0.00000627, -0.00000095, 0.00000699, 0.00000248, 0.00000049, -0.00000048, -0.00000446, 0.00000306, 0.00000002, 0.00000022, -0.00000200, 0.00000440, -0.00000425, -0.00000444, -0.00000264, 0.00000322, -0.00000889, 0.00001081, 0.00000135, 0.00000723, -0.00000866, 0.00000494, -0.00000497, -0.00000003, -0.00001574, 0.00002439, -0.00002404, 0.00003476, -0.00004341, 0.00004696, -0.00004249, 0.00001543, 0.00002285, -0.00006629, 0.00008275, -0.00006474, 0.00002593, -0.00000823, 0.00000133, -0.00001149, 0.00001115, -0.00000878, 0.00001498, -0.00002050, 0.00002007, -0.00001126, -0.00000385, -0.00001353, 0.00000385, -0.00000410, -0.00000087, -0.00001318, 0.00001892, -0.00001106, -0.00000788, -0.00001491, 0.00000610, -0.00000801, -0.00001299, -0.00002343, -0.00001819, -0.00003337, -0.00009369, -0.00017250, 0.00022607, 0.00009617, 0.00006659, 0.00003074, 0.00004182, 0.00001392, 0.00007576, 0.00004296, 0.00006255, 0.00015380, 0.00003710, -0.00013705, -0.00006462, -0.00002448, -0.00002153, -0.00001189, -0.00001656, -0.00001836, 0.00000093, -0.00001450, -0.00000248, -0.00000864, 0.00000250, -0.00000437, -0.00000846, -0.00000899, 0.00000228, -0.00000549, 0.00000271, -0.00000805, -0.00000120, },
		{ -0.00000085, -0.00000586, 0.00000186, -0.00000505, -0.00000031, -0.00000492, 0.00000082, -0.00000480, -0.00000162, -0.00000566, 0.00000061, 0.00000387, 0.00000141, -0.00001370, -0.00000347, 0.00000039, 0.00000599, -0.00000033, -0.00000655, -0.00000365, 0.00000469, -0.00000196, 0.00000290, -0.00000273, -0.00000169, 0.00000592, -0.00000328, -0.00000778, 0.00001005, 0.00000469, -0.00000158, -0.00000213, -0.00000656, 0.00000274, 0.00000620, 0.00000127, 0.00000055, -0.00000198, -0.00000288, 0.00000350, -0.00000111, 0.00000056, -0.00000122, 0.00000424, -0.00000681, -0.00000300, -0.00000054, -0.00000021, -0.00000621, 0.00001294, -0.00000037, 0.00000586, -0.00000942, 0.00000747, -0.00000674, -0.00000310, -0.00000858, 0.00001959, -0.00001834, 0.00002605, -0.00003613, 0.00004619, -0.00005428, 0.00004362, -0.00001685, -0.00003137, 0.00007332, -0.00008297, 0.00005408, -0.00002998, 0.00001163, -0.00001469, 0.00000977, -0.00000341, 0.00000806, -0.00001740, 0.00002302, -0.00001952, 0.00000258, -0.00001823, 0.00000655, -0.00000268, -0.00000615, -0.00000620, 0.00001704, -0.00001690, -0.00000072, -0.00002135, 0.00000897, -0.00000706, -0.00001802, -0.00002034, -0.00002173, -0.00003513, -0.00008705, -0.00019373, 0.00023789, 0.00010256, 0.00006602, 0.00004035, 0.00003810, 0.00002353, 0.00007973, 0.00003984, 0.00010176, 0.00015125, -0.00003259, -0.00014369, -0.00003658, -0.00002970, -0.00001997, -0.00000761, -0.00002061, -0.00001458, 0.00000016, -0.00001574, -0.00000022, -0.00000958, 0.00000269, -0.00000495, -0.00000926, -0.00000701, 0.00000217, -0.00000523, 0.00000235, -0.00000996, 0.00000001, -0.00000005, },
		{ -0.00000555, 0.00000186, -0.00000539, 0.00000002, -0.00000506, -0.00000001, -0.00000402, -0.00000170, -0.00000629, 0.00000310, 0.00000379, -0.00000291, -0.00001379, -0.00000065, 0.00000199, 0.00000553, -0.00000345, -0.00000621, -0.00000065, 0.00000364, -0.00000186, 0.00000270, -0.00000435, 0.00000160, 0.00000468, -0.00000699, -0.00000259, 0.00001115, 0.00000115, -0.00000120, -0.00000358, -0.00000525, 0.00000572, 0.00000439, 0.00000090, -0.00000006, -0.00000284, -0.00000096, 0.00000300, -0.00000182, 0.00000114, -0.00000046, 0.00000317, -0.00000884, -0.00000023, -0.00000037, -0.00000290, -0.00000219, 0.00001342, -0.00000221, 0.00000451, -0.00000903, 0.00000936, -0.00000997, -0.00000312, -0.00000336, 0.00001579, -0.00001463, 0.00001858, -0.00002622, 0.00003686, -0.00005198, 0.00005727, -0.00005062, 0.00001360, 0.00004106, -0.00007946, 0.00007255, -0.00005466, 0.00003110, -0.00002746, 0.00001621, -0.00000267, 0.00000104, -0.00000967, 0.00001962, -0.00002392, 0.00000904, -0.00002524, 0.00001328, -0.00000606, -0.00000800, 0.00000148, 0.00001005, -0.00001749, 0.00000430, -0.00002887, 0.00001533, -0.00001062, -0.00001989, -0.00001782, -0.00002741, -0.00003245, -0.00008451, -0.00021471, 0.00025254, 0.00010518, 0.00006964, 0.00004791, 0.00003433, 0.00003831, 0.00007534, 0.00004853, 0.00014567, 0.00011710, -0.00009840, -0.00012588, -0.00001729, -0.00003834, -0.00001351, -0.00000754, -0.00002278, -0.00001029, -0.00000274, -0.00001499, 0.00000134, -0.00001077, 0.00000320, -0.00000625, -0.00000893, -0.00000525, 0.00000152, -0.00000455, 0.00000119, -0.00001129, 0.00000197, 0.00000042, -0.00000199, },
		{ 0.00000154, -0.00000544, 0.00000031, -0.00000555, -0.00000043, -0.00000312, -0.00000234, -0.00000609, 0.00000571, 0.00000190, -0.00000675, -0.00001224, 0.00000193, 0.00000295, 0.00000420, -0.00000609, -0.00000443, 0.00000130, 0.00000235, -0.00000123, 0.00000143, -0.00000453, 0.00000479, 0.00000121, -0.00000808, 0.00000299, 0.00000955, -0.00000117, -0.00000085, -0.00000511, -0.00000248, 0.00000725, 0.00000238, 0.00000104, -0.00000126, -0.00000269, 0.00000065, 0.00000191, -0.00000199, 0.00000184, 0.00000007, 0.00000122, -0.00000961, 0.00000290, -0.00000203, -0.00000386, 0.00000211, 0.00001228, -0.00000383, 0.00000342, -0.00000766, 0.00000986, -0.00001330, -0.00000086, -0.00000008, 0.00001276, -0.00001291, 0.00001435, -0.00001830, 0.00002496, -0.00003950, 0.00005402, -0.00006827, 0.00005480, -0.00000528, -0.00005252, 0.00007163, -0.00007120, 0.00005243, -0.00004740, 0.00003209, -0.00001113, -0.00000029, -0.00000183, 0.00001101, -0.00002214, 0.00001175, -0.00003114, 0.00002190, -0.00001430, -0.00000369, 0.00000555, 0.00000111, -0.00001279, 0.00000451, -0.00003393, 0.00002245, -0.00001799, -0.00001697, -0.00001899, -0.00003192, -0.00002711, -0.00008725, -0.00023202, 0.00026641, 0.00010604, 0.00007801, 0.00005067, 0.00003502, 0.00005381, 0.00006431, 0.00007579, 0.00017975, 0.00005249, -0.00014434, -0.00009200, -0.00001157, -0.00004419, -0.00000530, -0.00001141, -0.00002141, -0.00000765, -0.00000638, -0.00001232, 0.00000144, -0.00001143, 0.00000343, -0.00000785, -0.00000734, -0.00000434, 0.00000085, -0.00000374, -0.00000076, -0.00001151, 0.00000407, -0.00000004, -0.00000271, -0.00000511, },
		{ -0.00000520, 0.00000033, -0.00000616, -0.00000037, -0.00000246, -0.00000323, -0.00000486, 0.00000764, -0.00000153, -0.00000928, -0.00000953, 0.00000386, 0.00000327, 0.00000224, -0.00000771, -0.00000187, 0.00000203, 0.00000142, -0.00000072, -0.00000037, -0.00000298, 0.00000654, -0.00000321, -0.00000618, 0.00000721, 0.00000634, -0.00000196, -0.00000122, -0.00000587, 0.00000101, 0.00000712, 0.00000086, 0.00000113, -0.00000263, -0.00000160, 0.00000148, 0.00000071, -0.00000167, 0.00000256, 0.00000009, -0.00000119, -0.00000876, 0.00000529, -0.00000480, -0.00000267, 0.00000560, 0.00000996, -0.00000497, 0.00000277, -0.00000581, 0.00000877, -0.00001563, 0.00000276, 0.00000145, 0.00001003, -0.00001204, 0.00001307, -0.00001476, 0.00001623, -0.00002410, 0.00003800, -0.00006577, 0.00007977, -0.00005164, -0.00000910, 0.00004834, -0.00007038, 0.00006570, -0.00006731, 0.00005412, -0.00002981, 0.00000882, -0.00000020, 0.00000191, -0.00001528, 0.00000872, -0.00003228, 0.00002838, -0.00002456, 0.00000684, 0.00000254, -0.00000504, -0.00000573, -0.00000019, -0.00003395, 0.00002658, -0.00002613, -0.00001016, -0.00002545, -0.00003169, -0.00002300, -0.00009350, -0.00024423, 0.00027580, 0.00010918, 0.00008872, 0.00004854, 0.00004369, 0.00006368, 0.00005383, 0.00012229, 0.00018769, -0.00003137, -0.00015982, -0.00005495, -0.00001809, -0.00004308, 0.00000043, -0.00001674, -0.00001668, -0.00000819, -0.00000884, -0.00000878, 0.00000007, -0.00001108, 0.00000275, -0.00000893, -0.00000493, -0.00000458, 0.00000071, -0.00000332, -0.00000309, -0.00001039, 0.00000563, -0.00000110, -0.00000286, -0.00000488, 0.00000111, },
		{ 0.00000001, -0.00000662, 0.00000006, -0.00000227, -0.00000393, -0.00000276, 0.00000820, -0.00000582, -0.00000998, -0.00000633, 0.00000496, 0.00000302, 0.00000004, -0.00000807, 0.00000074, 0.00000180, 0.00000108, -0.00000085, -0.00000185, -0.00000021, 0.00000608, -0.00000695, -0.00000202, 0.00000898, 0.00000291, -0.00000170, -0.00000240, -0.00000521, 0.00000424, 0.00000573, 0.00000011, 0.00000077, -0.00000363, 0.00000001, 0.00000141, -0.00000019, -0.00000104, 0.00000322, -0.00000064, -0.00000340, -0.00000645, 0.00000610, -0.00000750, 0.00000037, 0.00000745, 0.00000713, -0.00000556, 0.00000262, -0.00000406, 0.00000639, -0.00001628, 0.00000681, 0.00000134, 0.00000748, -0.00001073, 0.00001285, -0.00001504, 0.00001362, -0.00001304, 0.00001777, -0.00004672, 0.00008213, -0.00008404, 0.00003768, 0.00000803, -0.00004890, 0.00006258, -0.00007781, 0.00007450, -0.00005457, 0.00002917, -0.00000992, -0.00000105, -0.00000811, 0.00000143, -0.00002711, 0.00002880, -0.00003198, 0.00002026, -0.00000806, -0.00000438, -0.00000102, -0.00000688, -0.00002901, 0.00002510, -0.00003106, -0.00000282, -0.00003595, -0.00002494, -0.00002440, -0.00009901, -0.00025310, 0.00027915, 0.00011864, 0.00009717, 0.00004500, 0.00006028, 0.00006297, 0.00005487, 0.00017975, 0.00015750, -0.00011563, -0.00014359, -0.00002730, -0.00003051, -0.00003479, 0.00000062, -0.00002009, -0.00001065, -0.00001189, -0.00000867, -0.00000593, -0.00000199, -0.00000988, 0.00000086, -0.00000868, -0.00000257, -0.00000572, 0.00000135, -0.00000383, -0.00000508, -0.00000815, 0.00000612, -0.00000221, -0.00000255, -0.00000499, 0.00000082, -0.00000485, },
		{ -0.00000672, 0.00000058, -0.00000257, -0.00000401, -0.00000028, 0.00000699, -0.00001001, -0.00000879, -0.00000337, 0.00000524, 0.00000235, -0.00000201, -0.00000729, 0.00000283, 0.00000109, 0.00000114, -0.00000176, -0.00000224, 0.00000274, 0.00000352, -0.00000864, 0.00000294, 0.00000820, 0.00000039, -0.00000121, -0.00000393, -0.00000306, 0.00000634, 0.00000378, -0.00000001, -0.00000015, -0.00000383, 0.00000156, 0.00000065, -0.00000057, -0.00000030, 0.00000368, -0.00000215, -0.00000470, -0.00000339, 0.00000505, -0.00000897, 0.00000438, 0.00000731, 0.00000450, -0.00000565, 0.00000292, -0.00000292, 0.00000335, -0.00001497, 0.00001043, -0.00000032, 0.00000557, -0.00000852, 0.00001165, -0.00001641, 0.00001636, -0.00001041, 0.00000255, -0.00002055, 0.00006381, -0.00009379, 0.00007351, -0.00003746, -0.00001120, 0.00004030, -0.00007141, 0.00008373, -0.00007681, 0.00005606, -0.00003186, 0.00000749, -0.00000720, -0.00000502, -0.00001786, 0.00002182, -0.00003200, 0.00003089, -0.00002295, 0.00000409, -0.00000280, -0.00001080, -0.00002232, 0.00001825, -0.00003005, 0.00000077, -0.00004654, -0.00001320, -0.00003359, -0.00009896, -0.00026319, 0.00027838, 0.00013604, 0.00009897, 0.00004613, 0.00007997, 0.00005178, 0.00007778, 0.00023134, 0.00008720, -0.00017946, -0.00010442, -0.00001653, -0.00004058, -0.00002301, -0.00000498, -0.00001886, -0.00000628, -0.00001705, -0.00000570, -0.00000503, -0.00000355, -0.00000860, -0.00000184, -0.00000672, -0.00000122, -0.00000704, 0.00000253, -0.00000549, -0.00000595, -0.00000548, 0.00000543, -0.00000287, -0.00000213, -0.00000538, 0.00000081, -0.00000463, -0.00000054, },
		{ 0.00000093, -0.00000316, -0.00000319, 0.00000189, 0.00000405, -0.00001313, -0.00000608, -0.00000125, 0.00000491, 0.00000143, -0.00000357, -0.00000568, 0.00000412, 0.00000033, 0.00000121, -0.00000309, -0.00000121, 0.00000470, -0.00000021, -0.00000771, 0.00000707, 0.00000567, -0.00000074, -0.00000124, -0.00000501, 0.00000009, 0.00000691, 0.00000200, 0.00000007, -0.00000137, -0.00000315, 0.00000255, -0.00000037, -0.00000042, 0.00000044, 0.00000374, -0.00000422, -0.00000458, -0.00000059, 0.00000256, -0.00000845, 0.00000814, 0.00000547, 0.00000261, -0.00000540, 0.00000354, -0.00000273, 0.00000040, -0.00001184, 0.00001279, -0.00000315, 0.00000503, -0.00000607, 0.00000856, -0.00001586, 0.00002074, -0.00001558, -0.00000178, 0.00000154, 0.00003396, -0.00008056, 0.00008830, -0.00007438, 0.00003200, 0.00000324, -0.00004614, 0.00007464, -0.00008648, 0.00008014, -0.00006108, 0.00002854, -0.00001779, -0.00000395, -0.00001034, 0.00001031, -0.00002308, 0.00003328, -0.00003601, 0.00001732, -0.00001216, -0.00000781, -0.00001895, 0.00000958, -0.00002353, -0.00000221, -0.00005229, -0.00000109, -0.00004914, -0.00009065, -0.00027977, 0.00027849, 0.00015904, 0.00009280, 0.00005761, 0.00009443, 0.00003751, 0.00012656, 0.00025607, -0.00001238, -0.00020726, -0.00005798, -0.00002242, -0.00004208, -0.00001334, -0.00001351, -0.00001275, -0.00000594, -0.00002110, -0.00000118, -0.00000633, -0.00000370, -0.00000828, -0.00000434, -0.00000343, -0.00000143, -0.00000768, 0.00000352, -0.00000795, -0.00000523, -0.00000325, 0.00000393, -0.00000285, -0.00000195, -0.00000579, 0.00000101, -0.00000471, -0.00000109, -0.00000644, },
		{ -0.00000370, -0.00000149, 0.00000302, -0.00000013, -0.00001443, -0.00000262, -0.00000029, 0.00000429, 0.00000042, -0.00000446, -0.00000367, 0.00000464, -0.00000024, 0.00000093, -0.00000422, 0.00000095, 0.00000487, -0.00000373, -0.00000457, 0.00000923, 0.00000259, -0.00000073, -0.00000207, -0.00000494, 0.00000337, 0.00000607, 0.00000082, -0.00000007, -0.00000243, -0.00000185, 0.00000275, -0.00000123, 0.00000008, 0.00000109, 0.00000319, -0.00000632, -0.00000298, 0.00000099, -0.00000039, -0.00000585, 0.00001055, 0.00000263, 0.00000168, -0.00000500, 0.00000428, -0.00000360, -0.00000164, -0.00000741, 0.00001320, -0.00000630, 0.00000628, -0.00000475, 0.00000445, -0.00001186, 0.00002234, -0.00002391, 0.00000511, 0.00001122, 0.00000493, -0.00005209, 0.00007986, -0.00009227, 0.00006783, -0.00003833, -0.00000704, 0.00004588, -0.00007626, 0.00009063, -0.00008776, 0.00005706, -0.00004060, 0.00000968, -0.00001184, 0.00000107, -0.00000849, 0.00002528, -0.00004116, 0.00002917, -0.00002601, 0.00000303, -0.00002321, 0.00000456, -0.00001546, -0.00001101, -0.00005008, 0.00000561, -0.00006600, -0.00007541, -0.00030585, 0.00028520, 0.00018190, 0.00008216, 0.00008115, 0.00009576, 0.00003365, 0.00019427, 0.00023616, -0.00011940, -0.00019442, -0.00002086, -0.00003772, -0.00003368, -0.00001041, -0.00002044, -0.00000405, -0.00001008, -0.00002182, 0.00000275, -0.00000889, -0.00000241, -0.00000957, -0.00000547, 0.00000015, -0.00000304, -0.00000712, 0.00000350, -0.00001037, -0.00000303, -0.00000219, 0.00000226, -0.00000229, -0.00000218, -0.00000594, 0.00000118, -0.00000506, -0.00000140, -0.00000592, 0.00000093, },
		{ 0.00000081, 0.00000254, -0.00000468, -0.00001359, 0.00000068, -0.00000042, 0.00000372, -0.00000060, -0.00000464, -0.00000166, 0.00000456, -0.00000065, 0.00000020, -0.00000455, 0.00000343, 0.00000318, -0.00000576, -0.00000035, 0.00000909, 0.00000009, -0.00000027, -0.00000343, -0.00000343, 0.00000589, 0.00000438, 0.00000030, -0.00000063, -0.00000293, -0.00000043, 0.00000226, -0.00000165, 0.00000074, 0.00000157, 0.00000193, -0.00000779, -0.00000037, 0.00000083, -0.00000259, -0.00000184, 0.00001094, -0.00000030, 0.00000157, -0.00000455, 0.00000488, -0.00000537, -0.00000212, -0.00000259, 0.00001137, -0.00000860, 0.00000886, -0.00000567, 0.00000139, -0.00000535, 0.00001853, -0.00002934, 0.00001822, 0.00000619, -0.00001271, -0.00002063, 0.00005420, -0.00008783, 0.00008623, -0.00007219, 0.00003537, 0.00000325, -0.00004508, 0.00007976, -0.00010054, 0.00008303, -0.00007018, 0.00003599, -0.00002748, 0.00000230, 0.00000406, 0.00001007, -0.00003564, 0.00003334, -0.00003806, 0.00001830, -0.00003597, 0.00000775, -0.00001180, -0.00002096, -0.00004082, 0.00000306, -0.00007767, -0.00005867, -0.00033998, 0.00030160, 0.00019820, 0.00007473, 0.00011209, 0.00008149, 0.00005497, 0.00026217, 0.00016490, -0.00020737, -0.00014931, -0.00000431, -0.00005181, -0.00001936, -0.00001541, -0.00002189, 0.00000337, -0.00001686, -0.00001857, 0.00000419, -0.00001103, -0.00000061, -0.00001225, -0.00000446, 0.00000269, -0.00000516, -0.00000550, 0.00000195, -0.00001170, -0.00000009, -0.00000251, 0.00000106, -0.00000161, -0.00000269, -0.00000573, 0.00000105, -0.00000548, -0.00000141, -0.00000536, 0.00000054, -0.00000384, },
		{ 0.00000023, -0.00000852, -0.00001085, 0.00000295, -0.00000124, 0.00000337, -0.00000161, -0.00000414, 0.00000010, 0.00000409, -0.00000104, -0.00000077, -0.00000377, 0.00000526, 0.00000027, -0.00000564, 0.00000351, 0.00000717, -0.00000122, -0.00000013, -0.00000464, -0.00000075, 0.00000707, 0.00000252, 0.00000018, -0.00000153, -0.00000270, 0.00000065, 0.00000141, -0.00000158, 0.00000141, 0.00000172, 0.00000008, -0.00000806, 0.00000236, -0.00000095, -0.00000306, 0.00000244, 0.00000934, -0.00000251, 0.00000194, -0.00000404, 0.00000504, -0.00000757, -0.00000068, 0.00000146, 0.00000769, -0.00000898, 0.00001148, -0.00000885, 0.00000140, 0.00000078, 0.00000987, -0.00002743, 0.00002988, -0.00000937, -0.00001417, 0.00000205, 0.00002268, -0.00006570, 0.00008366, -0.00008888, 0.00006886, -0.00004198, 0.00000062, 0.00004638, -0.00009116, 0.00009488, -0.00009612, 0.00006872, -0.00005693, 0.00001967, 0.00000523, -0.00000416, -0.00002236, 0.00002692, -0.00004185, 0.00003125, -0.00005341, 0.00001993, -0.00001737, -0.00002528, -0.00002994, -0.00000802, -0.00007968, -0.00004752, -0.00037621, 0.00032569, 0.00020480, 0.00007899, 0.00014018, 0.00005835, 0.00011017, 0.00030398, 0.00005145, -0.00025415, -0.00009030, -0.00001014, -0.00005573, -0.00000629, -0.00002534, -0.00001668, 0.00000603, -0.00002297, -0.00001271, 0.00000247, -0.00001130, 0.00000025, -0.00001517, -0.00000151, 0.00000328, -0.00000660, -0.00000367, -0.00000093, -0.00001119, 0.00000252, -0.00000387, 0.00000063, -0.00000122, -0.00000315, -0.00000532, 0.00000051, -0.00000570, -0.00000117, -0.00000493, -0.00000012, -0.00000328, 0.00000315, },
		{ -0.00001067, -0.00000695, 0.00000370, -0.00000213, 0.00000324, -0.00000257, -0.00000303, 0.00000142, 0.00000336, -0.00000156, -0.00000154, -0.00000206, 0.00000574, -0.00000277, -0.00000349, 0.00000590, 0.00000446, -0.00000138, -0.00000070, -0.00000499, 0.00000240, 0.00000681, 0.00000102, 0.00000008, -0.00000244, -0.00000186, 0.00000112, 0.00000060, -0.00000115, 0.00000201, 0.00000136, -0.00000197, -0.00000690, 0.00000424, -0.00000357, -0.00000144, 0.00000573, 0.00000636, -0.00000362, 0.00000242, -0.00000346, 0.00000449, -0.00000951, 0.00000252, 0.00000363, 0.00000322, -0.00000703, 0.00001251, -0.00001287, 0.00000515, 0.00000322, -0.00000009, -0.00001761, 0.00003338, -0.00002708, -0.00000181, 0.00000921, -0.00000282, -0.00003607, 0.00006414, -0.00008529, 0.00008429, -0.00007713, 0.00004877, -0.00000275, -0.00005824, 0.00008423, -0.00010664, 0.00009660, -0.00009296, 0.00005271, -0.00001158, -0.00000722, -0.00000966, 0.00001291, -0.00003452, 0.00003504, -0.00006823, 0.00003681, -0.00003276, -0.00001854, -0.00002532, -0.00002173, -0.00007256, -0.00004695, -0.00040670, 0.00035027, 0.00020484, 0.00009927, 0.00015385, 0.00004222, 0.00019511, 0.00029462, -0.00007975, -0.00025005, -0.00003887, -0.00003064, -0.00004627, -0.00000113, -0.00003450, -0.00000714, 0.00000270, -0.00002523, -0.00000697, -0.00000141, -0.00000930, -0.00000102, -0.00001672, 0.00000219, 0.00000189, -0.00000647, -0.00000279, -0.00000423, -0.00000877, 0.00000385, -0.00000549, 0.00000083, -0.00000132, -0.00000324, -0.00000504, -0.00000032, -0.00000553, -0.00000087, -0.00000471, -0.00000085, -0.00000188, 0.00000481, -0.00000366, },
		{ -0.00000303, 0.00000296, -0.00000252, 0.00000309, -0.00000339, -0.00000142, 0.00000217, 0.00000245, -0.00000221, -0.00000172, -0.00000001, 0.00000468, -0.00000482, -0.00000015, 0.00000635, 0.00000202, -0.00000095, -0.00000190, -0.00000403, 0.00000517, 0.00000547, 0.00000012, -0.00000025, -0.00000304, -0.00000075, 0.00000097, 0.00000010, -0.00000052, 0.00000246, 0.00000036, -0.00000363, -0.00000460, 0.00000457, -0.00000590, 0.00000190, 0.00000714, 0.00000296, -0.00000367, 0.00000277, -0.00000285, 0.00000312, -0.00001038, 0.00000674, 0.00000329, -0.00000053, -0.00000340, 0.00001086, -0.00001550, 0.00001127, 0.00000016, -0.00000657, -0.00000349, 0.00002611, -0.00003836, 0.00001663, 0.00000116, -0.00001445, -0.00001030, 0.00003705, -0.00006554, 0.00007900, -0.00009291, 0.00008589, -0.00005491, -0.00000858, 0.00004976, -0.00009343, 0.00010720, -0.00012306, 0.00009333, -0.00004627, 0.00000838, -0.00000815, -0.00000005, -0.00001941, 0.00002676, -0.00007313, 0.00005041, -0.00005294, -0.00000014, -0.00003341, -0.00002923, -0.00006264, -0.00005638, -0.00042609, 0.00036606, 0.00020783, 0.00013187, 0.00014686, 0.00005312, 0.00028994, 0.00022046, -0.00019576, -0.00020176, -0.00001135, -0.00005254, -0.00002726, -0.00000659, -0.00003732, 0.00000192, -0.00000483, -0.00002228, -0.00000406, -0.00000530, -0.00000603, -0.00000449, -0.00001565, 0.00000494, -0.00000051, -0.00000471, -0.00000372, -0.00000662, -0.00000517, 0.00000351, -0.00000658, 0.00000115, -0.00000173, -0.00000286, -0.00000522, -0.00000111, -0.00000501, -0.00000065, -0.00000471, -0.00000140, 0.00000026, 0.00000510, -0.00000839, -0.00001050, },
		{ 0.00000128, -0.00000210, 0.00000259, -0.00000384, 0.00000050, 0.00000229, 0.00000142, -0.00000280, -0.00000112, 0.00000167, 0.00000247, -0.00000521, 0.00000321, 0.00000510, 0.00000052, -0.00000062, -0.00000319, -0.00000183, 0.00000688, 0.00000362, -0.00000026, -0.00000086, -0.00000309, 0.00000022, 0.00000044, 0.00000001, 0.00000015, 0.00000265, -0.00000124, -0.00000431, -0.00000186, 0.00000325, -0.00000684, 0.00000595, 0.00000642, 0.00000011, -0.00000305, 0.00000298, -0.00000237, 0.00000110, -0.00000955, 0.00001075, 0.00000060, -0.00000221, 0.00000039, 0.00000668, -0.00001477, 0.00001682, -0.00000753, -0.00000608, 0.00000893, 0.00001053, -0.00003815, 0.00003179, -0.00001570, -0.00001096, 0.00000360, 0.00001308, -0.00003892, 0.00005759, -0.00008685, 0.00010216, -0.00009583, 0.00004477, -0.00000145, -0.00005561, 0.00009212, -0.00013373, 0.00012763, -0.00009034, 0.00004302, -0.00002579, -0.00000085, -0.00000586, 0.00001012, -0.00006509, 0.00005288, -0.00006886, 0.00002412, -0.00005523, -0.00002306, -0.00005971, -0.00006869, -0.00043543, 0.00036672, 0.00022600, 0.00016470, 0.00012418, 0.00010617, 0.00036270, 0.00008704, -0.00026600, -0.00013009, -0.00001290, -0.00006332, -0.00000750, -0.00001996, -0.00003131, 0.00000564, -0.00001251, -0.00001538, -0.00000529, -0.00000702, -0.00000342, -0.00000894, -0.00001184, 0.00000533, -0.00000235, -0.00000222, -0.00000653, -0.00000701, -0.00000165, 0.00000181, -0.00000673, 0.00000106, -0.00000201, -0.00000226, -0.00000596, -0.00000150, -0.00000433, -0.00000060, -0.00000488, -0.00000149, 0.00000276, 0.00000355, -0.00001283, -0.00000804, -0.00000020, },
		{ -0.00000099, 0.00000150, -0.00000364, 0.00000246, 0.00000173, 0.00000038, -0.00000310, 0.00000014, 0.00000238, -0.00000006, -0.00000388, 0.00000556, 0.00000294, 0.00000007, -0.00000087, -0.00000390, 0.00000112, 0.00000722, 0.00000186, -0.00000037, -0.00000160, -0.00000262, 0.00000079, -0.00000017, 0.00000026, 0.00000078, 0.00000241, -0.00000315, -0.00000368, 0.00000037, 0.00000082, -0.00000579, 0.00000943, 0.00000400, -0.00000162, -0.00000225, 0.00000315, -0.00000232, -0.00000105, -0.00000687, 0.00001327, -0.00000338, -0.00000124, 0.00000252, 0.00000154, -0.00001015, 0.00001869, -0.00001650, 0.00000181, 0.00001408, -0.00000706, -0.00002687, 0.00003684, -0.00003229, 0.00000274, 0.00000359, -0.00000014, -0.00001584, 0.00002976, -0.00006408, 0.00009496, -0.00011500, 0.00008712, -0.00005601, -0.00000097, 0.00005095, -0.00011603, 0.00014066, -0.00012883, 0.00008813, -0.00006311, 0.00001881, -0.00000576, -0.00000472, -0.00004831, 0.00004116, -0.00007169, 0.00004384, -0.00008455, -0.00000175, -0.00007209, -0.00007288, -0.00044336, 0.00035369, 0.00026800, 0.00018163, 0.00010376, 0.00020201, 0.00037910, -0.00007957, -0.00027334, -0.00006223, -0.00003608, -0.00005690, 0.00000377, -0.00003434, -0.00001846, 0.00000157, -0.00001595, -0.00000785, -0.00000978, -0.00000564, -0.00000326, -0.00001234, -0.00000650, 0.00000302, -0.00000227, -0.00000050, -0.00001028, -0.00000506, 0.00000056, -0.00000036, -0.00000607, 0.00000033, -0.00000175, -0.00000190, -0.00000702, -0.00000133, -0.00000374, -0.00000067, -0.00000516, -0.00000091, 0.00000500, 0.00000013, -0.00001577, -0.00000399, 0.00000076, -0.00000042, },
		{ -0.00000013, -0.00000255, 0.00000405, 0.00000056, -0.00000049, -0.00000290, 0.00000164, 0.00000194, -0.00000205, -0.00000138, 0.00000636, 0.00000077, 0.00000031, -0.00000177, -0.00000350, 0.00000405, 0.00000631, 0.00000055, -0.00000049, -0.00000222, -0.00000184, 0.00000088, -0.00000059, 0.00000072, 0.00000131, 0.00000160, -0.00000486, -0.00000185, 0.00000132, -0.00000171, -0.00000287, 0.00001127, 0.00000082, -0.00000212, -0.00000164, 0.00000340, -0.00000293, -0.00000256, -0.00000283, 0.00001338, -0.00000705, 0.00000182, 0.00000174, -0.00000224, -0.00000304, 0.00001516, -0.00002237, 0.00001395, 0.00000949, -0.00001953, -0.00000983, 0.00003010, -0.00004097, 0.00001846, -0.00000644, -0.00000059, -0.00000362, 0.00000616, -0.00003494, 0.00006970, -0.00010937, 0.00010753, -0.00009893, 0.00005633, -0.00000753, -0.00006981, 0.00012239, -0.00014536, 0.00012802, -0.00011119, 0.00005937, -0.00002793, -0.00000494, -0.00003393, 0.00002021, -0.00005805, 0.00004875, -0.00010975, 0.00002806, -0.00010157, -0.00005951, -0.00046312, 0.00033765, 0.00033282, 0.00017090, 0.00011188, 0.00032105, 0.00031574, -0.00023927, -0.00022079, -0.00002135, -0.00006484, -0.00003627, 0.00000154, -0.00004211, -0.00000447, -0.00000885, -0.00001276, -0.00000344, -0.00001486, -0.00000206, -0.00000615, -0.00001279, -0.00000177, -0.00000102, 0.00000012, -0.00000088, -0.00001339, -0.00000146, 0.00000082, -0.00000199, -0.00000521, -0.00000082, -0.00000082, -0.00000223, -0.00000793, -0.00000072, -0.00000342, -0.00000079, -0.00000544, 0.00000046, 0.00000622, -0.00000463, -0.00001629, 0.00000038, -0.00000023, -0.00000125, 0.00000029, },
		{ -0.00000053, 0.00000487, -0.00000099, -0.00000099, -0.00000219, 0.00000290, 0.00000056, -0.00000287, 0.00000140, 0.00000565, -0.00000072, 0.00000062, -0.00000295, -0.00000185, 0.00000622, 0.00000461, -0.00000020, -0.00000077, -0.00000253, -0.00000102, 0.00000062, -0.00000072, 0.00000125, 0.00000163, 0.00000020, -0.00000580, 0.00000061, 0.00000068, -0.00000322, 0.00000110, 0.00001101, -0.00000210, -0.00000173, -0.00000132, 0.00000368, -0.00000425, -0.00000269, 0.00000148, 0.00001097, -0.00000887, 0.00000542, -0.00000185, -0.00000260, 0.00000368, 0.00000699, -0.00002178, 0.00002502, -0.00000303, -0.00002214, 0.00000547, 0.00001525, -0.00003856, 0.00002865, -0.00001925, 0.00000813, -0.00000398, -0.00000590, -0.00001067, 0.00003731, -0.00008426, 0.00010272, -0.00011914, 0.00010070, -0.00006792, -0.00000491, 0.00007240, -0.00012869, 0.00014530, -0.00015349, 0.00011118, -0.00007276, 0.00001892, -0.00003589, 0.00000259, -0.00003362, 0.00003467, -0.00011911, 0.00005351, -0.00014072, -0.00002672, -0.00050613, 0.00033507, 0.00040743, 0.00013377, 0.00017254, 0.00042532, 0.00017160, -0.00035078, -0.00013082, -0.00001779, -0.00008211, -0.00001178, -0.00001232, -0.00003881, 0.00000408, -0.00002087, -0.00000381, -0.00000435, -0.00001749, 0.00000133, -0.00001094, -0.00000969, 0.00000037, -0.00000484, 0.00000393, -0.00000376, -0.00001433, 0.00000234, -0.00000060, -0.00000252, -0.00000484, -0.00000183, 0.00000046, -0.00000337, -0.00000819, -0.00000002, -0.00000333, -0.00000094, -0.00000554, 0.00000248, 0.00000571, -0.00000963, -0.00001411, 0.00000365, -0.00000253, -0.00000066, 0.00000101, -0.00000189, },
		{ 0.00000461, -0.00000254, -0.00000097, -0.00000113, 0.00000352, -0.00000121, -0.00000236, 0.00000359, 0.00000400, -0.00000129, 0.00000050, -0.00000381, 0.00000073, 0.00000715, 0.00000268, -0.00000049, -0.00000120, -0.00000246, -0.00000041, 0.00000023, -0.00000057, 0.00000178, 0.00000160, -0.00000159, -0.00000559, 0.00000288, -0.00000121, -0.00000296, 0.00000492, 0.00000885, -0.00000400, -0.00000100, -0.00000117, 0.00000377, -0.00000592, -0.00000107, 0.00000474, 0.00000690, -0.00000804, 0.00000763, -0.00000678, 0.00000108, 0.00000692, -0.00000272, -0.00001417, 0.00002985, -0.00001821, -0.00001458, 0.00001280, -0.00000085, -0.00002710, 0.00002948, -0.00002787, 0.00001924, -0.00001312, -0.00000490, 0.00000101, 0.00000977, -0.00005084, 0.00007795, -0.00011350, 0.00012090, -0.00011359, 0.00006172, 0.00000118, -0.00007781, 0.00012778, -0.00017158, 0.00015653, -0.00012980, 0.00006718, -0.00006437, 0.00000391, -0.00001267, 0.00000761, -0.00010753, 0.00006142, -0.00017499, 0.00001672, -0.00057486, 0.00036039, 0.00047080, 0.00008923, 0.00029436, 0.00046893, -0.00002736, -0.00038629, -0.00003713, -0.00004568, -0.00007758, 0.00000413, -0.00002984, -0.00002563, 0.00000308, -0.00002871, 0.00000707, -0.00001006, -0.00001594, 0.00000193, -0.00001513, -0.00000416, -0.00000094, -0.00000653, 0.00000734, -0.00000829, -0.00001243, 0.00000479, -0.00000270, -0.00000207, -0.00000528, -0.00000213, 0.00000143, -0.00000500, -0.00000759, 0.00000035, -0.00000328, -0.00000118, -0.00000518, 0.00000469, 0.00000311, -0.00001349, -0.00000979, 0.00000473, -0.00000493, 0.00000122, 0.00000059, -0.00000314, 0.00000210, },
		{ -0.00000368, -0.00000046, -0.00000002, 0.00000336, -0.00000273, -0.00000083, 0.00000465, 0.00000212, -0.00000116, -0.00000023, -0.00000378, 0.00000362, 0.00000678, 0.00000101, -0.00000054, -0.00000168, -0.00000210, -0.00000010, -0.00000009, -0.00000024, 0.00000225, 0.00000106, -0.00000334, -0.00000417, 0.00000414, -0.00000351, -0.00000080, 0.00000744, 0.00000561, -0.00000462, -0.00000039, -0.00000100, 0.00000340, -0.00000732, 0.00000206, 0.00000587, 0.00000272, -0.00000490, 0.00000705, -0.00001069, 0.00000758, 0.00000487, -0.00000984, -0.00000216, 0.00002586, -0.00002967, -0.00000087, 0.00000981, -0.00001147, -0.00001228, 0.00002190, -0.00002858, 0.00002618, -0.00002414, 0.00000496, -0.00000148, -0.00000454, -0.00002153, 0.00004452, -0.00008762, 0.00011411, -0.00013278, 0.00011185, -0.00007287, -0.00000347, 0.00007409, -0.00015248, 0.00017559, -0.00017990, 0.00012842, -0.00011945, 0.00003552, -0.00001308, -0.00001650, -0.00008135, 0.00004589, -0.00018933, 0.00005357, -0.00065888, 0.00041691, 0.00050383, 0.00007081, 0.00046058, 0.00041447, -0.00023344, -0.00034212, 0.00002821, -0.00008629, -0.00005233, 0.00000318, -0.00004093, -0.00000898, -0.00000686, -0.00002848, 0.00001492, -0.00001747, -0.00001082, -0.00000158, -0.00001610, 0.00000129, -0.00000482, -0.00000521, 0.00000851, -0.00001263, -0.00000824, 0.00000494, -0.00000423, -0.00000135, -0.00000629, -0.00000151, 0.00000149, -0.00000645, -0.00000634, 0.00000020, -0.00000306, -0.00000161, -0.00000406, 0.00000630, -0.00000131, -0.00001493, -0.00000469, 0.00000330, -0.00000615, 0.00000362, -0.00000111, -0.00000322, 0.00000472, 0.00000326, },
		{ 0.00000034, 0.00000086, 0.00000256, -0.00000354, 0.00000109, 0.00000449, 0.00000067, -0.00000081, -0.00000136, -0.00000258, 0.00000608, 0.00000536, -0.00000013, -0.00000057, -0.00000207, -0.00000161, -0.00000004, -0.00000023, 0.00000019, 0.00000259, -0.00000004, -0.00000452, -0.00000197, 0.00000395, -0.00000515, 0.00000264, 0.00000798, 0.00000231, -0.00000419, -0.00000012, -0.00000075, 0.00000240, -0.00000770, 0.00000582, 0.00000452, 0.00000002, -0.00000087, 0.00000355, -0.00001145, 0.00001415, -0.00000198, -0.00001105, 0.00000948, 0.00001420, -0.00003275, 0.00001260, -0.00000124, -0.00001293, -0.00000044, 0.00001049, -0.00002206, 0.00002559, -0.00003055, 0.00001637, -0.00001368, -0.00000362, -0.00000521, 0.00001503, -0.00005321, 0.00008664, -0.00012290, 0.00013263, -0.00012983, 0.00007490, -0.00000462, -0.00009467, 0.00015439, -0.00020164, 0.00018178, -0.00018832, 0.00009748, -0.00004804, -0.00001740, -0.00005827, 0.00001397, -0.00017700, 0.00006585, -0.00073721, 0.00049126, 0.00050191, 0.00011467, 0.00062764, 0.00024953, -0.00039172, -0.00024081, 0.00004701, -0.00011695, -0.00001789, -0.00001452, -0.00003841, 0.00000253, -0.00002045, -0.00002030, 0.00001627, -0.00002252, -0.00000490, -0.00000833, -0.00001250, 0.00000393, -0.00000901, -0.00000162, 0.00000654, -0.00001478, -0.00000339, 0.00000295, -0.00000440, -0.00000125, -0.00000714, -0.00000036, 0.00000043, -0.00000703, -0.00000498, -0.00000037, -0.00000262, -0.00000217, -0.00000207, 0.00000641, -0.00000658, -0.00001331, -0.00000053, 0.00000003, -0.00000533, 0.00000541, -0.00000359, -0.00000180, 0.00000658, 0.00000116, -0.00000408, },
		{ 0.00000135, 0.00000143, -0.00000345, 0.00000275, 0.00000344, -0.00000005, -0.00000072, -0.00000240, -0.00000033, 0.00000751, 0.00000340, -0.00000067, -0.00000071, -0.00000227, -0.00000113, -0.00000011, -0.00000018, 0.00000065, 0.00000264, -0.00000158, -0.00000471, 0.00000028, 0.00000245, -0.00000527, 0.00000631, 0.00000655, -0.00000018, -0.00000323, -0.00000013, -0.00000051, 0.00000093, -0.00000656, 0.00000897, 0.00000131, -0.00000020, 0.00000214, -0.00000158, -0.00000820, 0.00001773, -0.00001089, -0.00000546, 0.00001586, -0.00000069, -0.00002648, 0.00001994, -0.00001488, -0.00000582, 0.00000429, 0.00000078, -0.00001235, 0.00001832, -0.00002916, 0.00002251, -0.00002779, 0.00000820, -0.00000427, -0.00000150, -0.00002318, 0.00005106, -0.00009119, 0.00012116, -0.00015537, 0.00013594, -0.00008784, -0.00001002, 0.00009120, -0.00017986, 0.00020400, -0.00024779, 0.00017537, -0.00011814, 0.00002030, -0.00006126, -0.00001419, -0.00014630, 0.00004532, -0.00078760, 0.00055551, 0.00048426, 0.00024234, 0.00073510, -0.00000359, -0.00045892, -0.00012319, 0.00002154, -0.00012136, 0.00000950, -0.00004032, -0.00002186, 0.00000247, -0.00003027, -0.00000839, 0.00001097, -0.00002236, -0.00000175, -0.00001543, -0.00000520, 0.00000223, -0.00001093, 0.00000214, 0.00000208, -0.00001362, 0.00000020, -0.00000006, -0.00000325, -0.00000227, -0.00000705, 0.00000049, -0.00000137, -0.00000642, -0.00000412, -0.00000098, -0.00000215, -0.00000261, 0.00000058, 0.00000432, -0.00001116, -0.00000898, 0.00000120, -0.00000364, -0.00000247, 0.00000555, -0.00000587, 0.00000091, 0.00000703, -0.00000108, -0.00000365, 0.00000117, },
		{ 0.00000035, -0.00000261, 0.00000369, 0.00000206, -0.00000012, -0.00000116, -0.00000280, 0.00000250, 0.00000764, 0.00000144, -0.00000074, -0.00000099, -0.00000228, -0.00000075, -0.00000019, -0.00000001, 0.00000111, 0.00000225, -0.00000320, -0.00000378, 0.00000181, 0.00000031, -0.00000358, 0.00000903, 0.00000377, -0.00000143, -0.00000227, -0.00000022, -0.00000055, -0.00000054, -0.00000392, 0.00001038, -0.00000240, 0.00000194, 0.00000263, -0.00000612, -0.00000190, 0.00001635, -0.00001795, 0.00000484, 0.00001419, -0.00001306, -0.00001385, 0.00001832, -0.00002503, 0.00000571, 0.00000149, -0.00000339, -0.00000444, 0.00000844, -0.00002111, 0.00002004, -0.00003641, 0.00002294, -0.00001460, -0.00000265, -0.00000657, 0.00002094, -0.00005152, 0.00008534, -0.00014586, 0.00016403, -0.00015282, 0.00007955, -0.00000129, -0.00011264, 0.00017864, -0.00027204, 0.00024302, -0.00020755, 0.00009797, -0.00010813, -0.00001206, -0.00012108, 0.00000191, -0.00079959, 0.00057826, 0.00049426, 0.00044440, 0.00072532, -0.00029011, -0.00041961, -0.00003309, -0.00002737, -0.00009694, 0.00001701, -0.00006123, 0.00000192, -0.00000999, -0.00003045, 0.00000093, 0.00000241, -0.00001701, -0.00000369, -0.00001937, 0.00000297, -0.00000322, -0.00000904, 0.00000362, -0.00000280, -0.00000948, 0.00000120, -0.00000257, -0.00000166, -0.00000417, -0.00000567, 0.00000017, -0.00000302, -0.00000487, -0.00000408, -0.00000123, -0.00000194, -0.00000247, 0.00000316, -0.00000006, -0.00001346, -0.00000331, -0.00000013, -0.00000597, 0.00000147, 0.00000365, -0.00000685, 0.00000411, 0.00000588, -0.00000275, -0.00000262, 0.00000177, 0.00000144, },
		{ -0.00000136, 0.00000375, 0.00000087, 0.00000012, -0.00000207, -0.00000215, 0.00000523, 0.00000655, -0.00000009, -0.00000056, -0.00000138, -0.00000212, -0.00000049, -0.00000022, 0.00000025, 0.00000148, 0.00000134, -0.00000444, -0.00000198, 0.00000210, -0.00000151, -0.00000045, 0.00000999, 0.00000061, -0.00000151, -0.00000162, -0.00000021, -0.00000108, -0.00000133, -0.00000043, 0.00000952, -0.00000505, 0.00000526, 0.00000023, -0.00000790, 0.00000497, 0.00001013, -0.00001983, 0.00001560, 0.00000502, -0.00001829, -0.00000036, 0.00000889, -0.00002774, 0.00001625, -0.00000589, -0.00000145, -0.00000170, 0.00000083, -0.00001077, 0.00001020, -0.00003571, 0.00003280, -0.00002826, 0.00000742, -0.00000556, 0.00000556, -0.00001870, 0.00004033, -0.00010937, 0.00015490, -0.00018276, 0.00014957, -0.00009949, -0.00001362, 0.00010373, -0.00024281, 0.00027110, -0.00028687, 0.00019832, -0.00020103, 0.00004136, -0.00013324, -0.00003504, -0.00078591, 0.00054195, 0.00058785, 0.00067307, 0.00056641, -0.00053796, -0.00029301, -0.00000060, -0.00007011, -0.00005571, 0.00000164, -0.00006638, 0.00002205, -0.00002939, -0.00001976, 0.00000248, -0.00000415, -0.00000954, -0.00001039, -0.00001784, 0.00000846, -0.00000990, -0.00000375, 0.00000127, -0.00000559, -0.00000410, -0.00000045, -0.00000348, -0.00000076, -0.00000604, -0.00000341, -0.00000169, -0.00000357, -0.00000315, -0.00000470, -0.00000099, -0.00000217, -0.00000135, 0.00000463, -0.00000594, -0.00001241, 0.00000171, -0.00000394, -0.00000572, 0.00000495, 0.00000016, -0.00000580, 0.00000671, 0.00000350, -0.00000333, -0.00000140, 0.00000203, 0.00000124, -0.00000041, },
		{ 0.00000307, 0.00000020, 0.00000024, -0.00000307, -0.00000041, 0.00000719, 0.00000461, -0.00000094, -0.00000037, -0.00000182, -0.00000183, -0.00000029, -0.00000017, 0.00000057, 0.00000165, -0.00000002, -0.00000486, 0.00000007, 0.00000111, -0.00000212, 0.00000321, 0.00000899, -0.00000201, -0.00000087, -0.00000133, -0.00000013, -0.00000209, -0.00000091, 0.00000285, 0.00000675, -0.00000554, 0.00000798, -0.00000405, -0.00000586, 0.00000946, 0.00000137, -0.00001531, 0.00002221, -0.00000800, -0.00001489, 0.00000849, -0.00000398, -0.00002255, 0.00002160, -0.00001341, 0.00000419, -0.00000448, -0.00000134, -0.00000328, -0.00000229, -0.00002669, 0.00003357, -0.00003745, 0.00002071, -0.00001554, 0.00000674, -0.00000278, 0.00000248, -0.00006198, 0.00011672, -0.00017290, 0.00018185, -0.00017699, 0.00009174, -0.00000567, -0.00015792, 0.00023858, -0.00032262, 0.00028830, -0.00032071, 0.00014915, -0.00020876, -0.00002528, -0.00078561, 0.00045967, 0.00081189, 0.00084888, 0.00026991, -0.00068128, -0.00012694, -0.00003072, -0.00008175, -0.00001836, -0.00002795, -0.00005264, 0.00002900, -0.00004646, -0.00000252, -0.00000507, -0.00000436, -0.00000460, -0.00001871, -0.00001114, 0.00000884, -0.00001447, 0.00000260, -0.00000459, -0.00000455, 0.00000015, -0.00000352, -0.00000271, -0.00000128, -0.00000680, -0.00000132, -0.00000467, -0.00000251, -0.00000211, -0.00000540, -0.00000049, -0.00000267, 0.00000084, 0.00000394, -0.00001176, -0.00000803, 0.00000415, -0.00000857, -0.00000265, 0.00000646, -0.00000366, -0.00000274, 0.00000768, 0.00000073, -0.00000271, -0.00000044, 0.00000202, 0.00000092, -0.00000074, -0.00000013, },
		{ 0.00000008, -0.00000007, -0.00000366, 0.00000214, 0.00000792, 0.00000239, -0.00000111, -0.00000035, -0.00000220, -0.00000146, -0.00000014, -0.00000007, 0.00000092, 0.00000147, -0.00000156, -0.00000426, 0.00000167, -0.00000066, -0.00000109, 0.00000629, 0.00000645, -0.00000350, -0.00000005, -0.00000123, -0.00000015, -0.00000326, 0.00000085, 0.00000482, 0.00000318, -0.00000376, 0.00000856, -0.00000818, -0.00000062, 0.00000944, -0.00000633, -0.00000587, 0.00002168, -0.00001966, -0.00000497, 0.00000952, -0.00001488, -0.00001238, 0.00002046, -0.00001736, 0.00000969, -0.00001030, 0.00000198, -0.00000193, -0.00001184, -0.00001407, 0.00002588, -0.00003786, 0.00002942, -0.00002807, 0.00001908, -0.00000564, -0.00001695, -0.00002135, 0.00006648, -0.00013180, 0.00017103, -0.00021400, 0.00017482, -0.00012153, -0.00003420, 0.00014235, -0.00029019, 0.00032956, -0.00042939, 0.00028930, -0.00035261, 0.00006557, -0.00085508, 0.00038344, 0.00117927, 0.00088208, -0.00010500, -0.00068430, 0.00001946, -0.00010220, -0.00005318, -0.00000416, -0.00005551, -0.00002656, 0.00001955, -0.00005270, 0.00001344, -0.00001827, 0.00000279, -0.00000588, -0.00002410, -0.00000233, 0.00000411, -0.00001454, 0.00000689, -0.00001175, 0.00000033, 0.00000144, -0.00000614, -0.00000124, -0.00000306, -0.00000589, -0.00000058, -0.00000767, -0.00000010, -0.00000228, -0.00000555, -0.00000027, -0.00000289, 0.00000356, 0.00000059, -0.00001558, -0.00000161, 0.00000298, -0.00001188, 0.00000223, 0.00000518, -0.00000624, 0.00000146, 0.00000656, -0.00000142, -0.00000126, -0.00000005, 0.00000194, 0.00000059, -0.00000067, 0.00000076, 0.00000200, },
		{ -0.00000086, -0.00000337, 0.00000490, 0.00000731, 0.00000044, -0.00000080, -0.00000062, -0.00000245, -0.00000102, -0.00000002, 0.00000007, 0.00000125, 0.00000082, -0.00000288, -0.00000279, 0.00000224, -0.00000239, 0.00000143, 0.00000791, 0.00000322, -0.00000377, 0.00000054, -0.00000111, -0.00000049, -0.00000407, 0.00000355, 0.00000486, 0.00000025, -0.00000062, 0.00000643, -0.00001005, 0.00000569, 0.00000467, -0.00000970, 0.00000481, 0.00001398, -0.00002549, 0.00000682, 0.00000303, -0.00001974, -0.00000179, 0.00001452, -0.00001640, 0.00001190, -0.00001546, 0.00000806, -0.00000671, -0.00001475, -0.00000364, 0.00001423, -0.00003007, 0.00002943, -0.00003518, 0.00003323, -0.00002122, -0.00001551, 0.00000013, 0.00002299, -0.00007756, 0.00012610, -0.00020447, 0.00021415, -0.00021233, 0.00009693, 0.00000085, -0.00018496, 0.00029311, -0.00048234, 0.00041732, -0.00053928, 0.00024668, -0.00104759, 0.00039777, 0.00165065, 0.00070313, -0.00046450, -0.00055596, 0.00009794, -0.00017734, 0.00000485, -0.00002140, -0.00006504, -0.00000155, -0.00000122, -0.00004478, 0.00002068, -0.00003057, 0.00001416, -0.00001398, -0.00002299, 0.00000413, -0.00000302, -0.00000997, 0.00000675, -0.00001711, 0.00000723, -0.00000060, -0.00000681, -0.00000055, -0.00000500, -0.00000366, -0.00000181, -0.00000941, 0.00000271, -0.00000350, -0.00000481, -0.00000079, -0.00000211, 0.00000566, -0.00000496, -0.00001587, 0.00000463, -0.00000145, -0.00001211, 0.00000701, 0.00000145, -0.00000639, 0.00000540, 0.00000366, -0.00000219, 0.00000037, -0.00000019, 0.00000196, 0.00000027, -0.00000033, 0.00000113, 0.00000094, 0.00000031, },
		{ -0.00000199, 0.00000717, 0.00000559, -0.00000083, -0.00000030, -0.00000114, -0.00000248, -0.00000055, 0.00000006, 0.00000026, 0.00000146, -0.00000027, -0.00000357, -0.00000091, 0.00000165, -0.00000322, 0.00000472, 0.00000766, 0.00000024, -0.00000316, 0.00000075, -0.00000089, -0.00000122, -0.00000401, 0.00000632, 0.00000309, -0.00000095, 0.00000230, 0.00000233, -0.00000844, 0.00001028, -0.00000301, -0.00000717, 0.00001243, 0.00000198, -0.00002353, 0.00001537, -0.00000761, -0.00001737, 0.00000518, 0.00000726, -0.00001180, 0.00000983, -0.00001705, 0.00001298, -0.00001464, -0.00001071, 0.00000051, 0.00000434, -0.00001851, 0.00002159, -0.00003290, 0.00004046, -0.00003908, 0.00000056, -0.00000061, -0.00000032, -0.00003040, 0.00006641, -0.00015785, 0.00020316, -0.00025442, 0.00019991, -0.00015065, -0.00002679, 0.00017236, -0.00044478, 0.00047930, -0.00071467, 0.00048847, -0.00138758, 0.00059720, 0.00213184, 0.00029196, -0.00070744, -0.00034914, 0.00008968, -0.00021872, 0.00006682, -0.00006300, -0.00004889, 0.00000864, -0.00002202, -0.00002658, 0.00001603, -0.00003565, 0.00002396, -0.00002575, -0.00001475, 0.00000460, -0.00000842, -0.00000319, 0.00000197, -0.00001820, 0.00001331, -0.00000468, -0.00000523, -0.00000172, -0.00000568, -0.00000138, -0.00000466, -0.00000909, 0.00000471, -0.00000501, -0.00000345, -0.00000207, 0.00000011, 0.00000579, -0.00001116, -0.00001223, 0.00000831, -0.00000735, -0.00000867, 0.00000962, -0.00000331, -0.00000381, 0.00000762, -0.00000002, -0.00000134, 0.00000148, -0.00000058, 0.00000213, -0.00000004, 0.00000013, 0.00000101, 0.00000019, 0.00000055, -0.00000188, },
		{ 0.00000835, 0.00000327, -0.00000124, 0.00000006, -0.00000181, -0.00000225, -0.00000010, 0.00000009, 0.00000051, 0.00000140, -0.00000161, -0.00000339, 0.00000074, 0.00000020, -0.00000260, 0.00000776, 0.00000577, -0.00000180, -0.00000217, 0.00000067, -0.00000064, -0.00000220, -0.00000283, 0.00000822, 0.00000037, -0.00000009, 0.00000353, -0.00000202, -0.00000363, 0.00001098, -0.00001044, 0.00000031, 0.00001392, -0.00000972, -0.00001509, 0.00001729, -0.00001762, -0.00000966, 0.00000661, 0.00000206, -0.00000638, 0.00000483, -0.00001434, 0.00001370, -0.00002143, -0.00000265, -0.00000224, 0.00000029, -0.00000880, 0.00001062, -0.00002265, 0.00003657, -0.00004939, 0.00002005, -0.00001724, 0.00000005, -0.00000496, 0.00001380, -0.00009485, 0.00015215, -0.00024059, 0.00024780, -0.00027166, 0.00014474, -0.00001115, -0.00030736, 0.00043169, -0.00081083, 0.00072211, -0.00184929, 0.00105349, 0.00249117, -0.00030527, -0.00075864, -0.00014351, 0.00001243, -0.00020586, 0.00010452, -0.00010932, -0.00001205, -0.00000311, -0.00003081, -0.00000757, 0.00000253, -0.00003088, 0.00002668, -0.00003556, -0.00000243, -0.00000179, -0.00000856, 0.00000194, -0.00000524, -0.00001447, 0.00001611, -0.00000853, -0.00000248, -0.00000473, -0.00000417, -0.00000060, -0.00000783, -0.00000688, 0.00000510, -0.00000584, -0.00000221, -0.00000353, 0.00000351, 0.00000297, -0.00001583, -0.00000575, 0.00000788, -0.00001216, -0.00000253, 0.00000876, -0.00000708, 0.00000067, 0.00000732, -0.00000316, 0.00000067, 0.00000167, -0.00000079, 0.00000234, -0.00000034, 0.00000054, 0.00000055, -0.00000015, 0.00000048, -0.00000270, 0.00000035, },
		{ 0.00000100, -0.00000092, 0.00000003, -0.00000244, -0.00000177, 0.00000025, 0.00000010, 0.00000080, 0.00000095, -0.00000286, -0.00000238, 0.00000162, -0.00000143, -0.00000054, 0.00000964, 0.00000293, -0.00000265, -0.00000128, 0.00000052, -0.00000054, -0.00000305, -0.00000076, 0.00000857, -0.00000209, 0.00000217, 0.00000236, -0.00000464, 0.00000261, 0.00000719, -0.00001445, 0.00000971, 0.00000871, -0.00001684, -0.00000400, 0.00001215, -0.00002295, -0.00000036, 0.00000325, 0.00000068, -0.00000291, -0.00000038, -0.00000897, 0.00000957, -0.00002367, 0.00000502, -0.00000925, 0.00000278, -0.00000482, 0.00000226, -0.00000985, 0.00002326, -0.00004717, 0.00003217, -0.00003795, 0.00001754, -0.00000525, -0.00001576, -0.00003897, 0.00008372, -0.00018240, 0.00023202, -0.00033085, 0.00028330, -0.00021257, -0.00009370, 0.00026037, -0.00076969, 0.00085536, -0.00234979, 0.00178293, 0.00259392, -0.00097996, -0.00059609, -0.00001771, -0.00008676, -0.00014475, 0.00010067, -0.00013740, 0.00002963, -0.00003325, -0.00002087, 0.00000176, -0.00001188, -0.00001899, 0.00002022, -0.00003815, 0.00000855, -0.00001242, -0.00000247, 0.00000203, -0.00001114, -0.00000771, 0.00001482, -0.00001014, -0.00000043, -0.00000834, -0.00000070, -0.00000234, -0.00000961, -0.00000392, 0.00000393, -0.00000538, -0.00000191, -0.00000416, 0.00000698, -0.00000275, -0.00001693, 0.00000120, 0.00000339, -0.00001356, 0.00000411, 0.00000448, -0.00000814, 0.00000536, 0.00000463, -0.00000467, 0.00000295, 0.00000103, -0.00000051, 0.00000238, -0.00000061, 0.00000080, -0.00000002, -0.00000016, -0.00000004, -0.00000285, 0.00000319, 0.00000814, },
		{ -0.00000019, -0.00000047, -0.00000283, -0.00000111, 0.00000046, 0.00000014, 0.00000103, 0.00000009, -0.00000365, -0.00000088, 0.00000149, -0.00000245, 0.00000245, 0.00000982, 0.00000004, -0.00000249, -0.00000075, 0.00000045, -0.00000071, -0.00000335, 0.00000158, 0.00000735, -0.00000324, 0.00000452, -0.00000077, -0.00000420, 0.00000777, 0.00000026, -0.00001323, 0.00001713, -0.00000106, -0.00001711, 0.00000508, 0.00000244, -0.00002187, 0.00000673, -0.00000230, 0.00000273, -0.00000270, -0.00000327, -0.00000386, 0.00000241, -0.00002040, 0.00000862, -0.00001627, 0.00000928, -0.00000703, 0.00000026, -0.00000088, 0.00000674, -0.00003398, 0.00003121, -0.00005118, 0.00003954, -0.00002446, -0.00001759, -0.00000749, 0.00002356, -0.00010515, 0.00016521, -0.00031740, 0.00035210, -0.00037810, 0.00014426, -0.00000903, -0.00056763, 0.00080124, -0.00276197, 0.00272629, 0.00234392, -0.00158646, -0.00026190, -0.00002086, -0.00015078, -0.00006579, 0.00005794, -0.00013217, 0.00005673, -0.00006860, 0.00000536, -0.00000469, -0.00001821, -0.00000708, 0.00000742, -0.00003149, 0.00001300, -0.00002227, 0.00000754, -0.00000398, -0.00001232, -0.00000128, 0.00001068, -0.00000884, -0.00000063, -0.00001070, 0.00000330, -0.00000641, -0.00000882, -0.00000181, 0.00000215, -0.00000384, -0.00000287, -0.00000311, 0.00000885, -0.00000999, -0.00001353, 0.00000578, -0.00000340, -0.00001050, 0.00000864, -0.00000168, -0.00000576, 0.00000841, 0.00000066, -0.00000413, 0.00000455, 0.00000003, 0.00000030, 0.00000203, -0.00000075, 0.00000085, -0.00000048, -0.00000007, -0.00000088, -0.00000197, 0.00000585, 0.00000659, -0.00000064, },
		{ -0.00000133, -0.00000282, -0.00000041, 0.00000051, 0.00000025, 0.00000108, -0.00000103, -0.00000373, 0.00000059, 0.00000051, -0.00000229, 0.00000551, 0.00000834, -0.00000216, -0.00000176, -0.00000056, 0.00000050, -0.00000114, -0.00000284, 0.00000343, 0.00000512, -0.00000266, 0.00000559, -0.00000444, -0.00000071, 0.00000970, -0.00000710, -0.00000720, 0.00001950, -0.00001146, -0.00001112, 0.00000868, -0.00000771, -0.00001537, 0.00000933, -0.00000706, 0.00000622, -0.00000519, -0.00000282, -0.00000164, -0.00000454, -0.00001334, 0.00000687, -0.00001965, 0.00001566, -0.00001265, 0.00000465, 0.00000019, -0.00000567, -0.00001637, 0.00001842, -0.00005063, 0.00005358, -0.00004915, 0.00000127, -0.00000548, -0.00000953, -0.00003774, 0.00007584, -0.00024465, 0.00033728, -0.00046384, 0.00034236, -0.00031499, -0.00023079, 0.00051080, -0.00294632, 0.00375102, 0.00171853, -0.00198221, 0.00014813, -0.00015427, -0.00013610, -0.00001091, -0.00000097, -0.00009459, 0.00005584, -0.00009241, 0.00003692, -0.00002522, -0.00001120, -0.00000303, -0.00000499, -0.00001824, 0.00000870, -0.00002642, 0.00001685, -0.00001401, -0.00000747, 0.00000157, 0.00000641, -0.00000574, -0.00000335, -0.00001038, 0.00000582, -0.00001132, -0.00000544, -0.00000175, 0.00000109, -0.00000215, -0.00000457, -0.00000023, 0.00000766, -0.00001636, -0.00000640, 0.00000588, -0.00000953, -0.00000374, 0.00000915, -0.00000734, -0.00000066, 0.00000857, -0.00000298, -0.00000201, 0.00000488, -0.00000064, 0.00000133, 0.00000124, -0.00000070, 0.00000069, -0.00000068, -0.00000017, -0.00000172, -0.00000003, 0.00000762, 0.00000415, -0.00000131, 0.00000050, },
		{ -0.00000237, 0.00000019, 0.00000045, 0.00000042, 0.00000082, -0.00000217, -0.00000308, 0.00000153, -0.00000080, -0.00000078, 0.00000773, 0.00000570, -0.00000326, -0.00000096, -0.00000056, 0.00000057, -0.00000161, -0.00000159, 0.00000418, 0.00000286, -0.00000076, 0.00000464, -0.00000685, 0.00000448, 0.00000754, -0.00001195, 0.00000125, 0.00001582, -0.00001850, -0.00000194, 0.00000584, -0.00001443, -0.00000644, 0.00000740, -0.00000906, 0.00000882, -0.00000843, 0.00000010, -0.00000329, -0.00000833, -0.00000578, 0.00000118, -0.00001809, 0.00001849, -0.00001746, 0.00001205, -0.00000653, -0.00000916, -0.00000235, 0.00000057, -0.00003743, 0.00005266, -0.00006577, 0.00002659, -0.00002541, -0.00001012, -0.00000145, -0.00000314, -0.00014447, 0.00025274, -0.00045174, 0.00044636, -0.00057813, 0.00016614, -0.00000231, -0.00279269, 0.00467764, 0.00078422, -0.00206989, 0.00051350, -0.00037030, -0.00003137, -0.00001514, -0.00004259, -0.00004261, 0.00002688, -0.00009255, 0.00005933, -0.00005062, 0.00000756, -0.00001135, -0.00000951, -0.00000484, -0.00000211, -0.00002262, 0.00002075, -0.00002370, 0.00000181, -0.00000067, 0.00000474, -0.00000317, -0.00000727, -0.00000733, 0.00000531, -0.00001492, -0.00000075, -0.00000389, 0.00000175, -0.00000152, -0.00000579, 0.00000350, 0.00000292, -0.00001932, 0.00000203, 0.00000106, -0.00001211, 0.00000429, 0.00000532, -0.00001019, 0.00000531, 0.00000572, -0.00000483, 0.00000061, 0.00000399, -0.00000046, 0.00000212, 0.00000016, -0.00000044, 0.00000035, -0.00000062, -0.00000065, -0.00000208, 0.00000265, 0.00000800, 0.00000149, -0.00000102, 0.00000075, -0.00000231, },
		{ 0.00000056, 0.00000037, 0.00000060, 0.00000023, -0.00000302, -0.00000191, 0.00000166, -0.00000176, 0.00000174, 0.00000849, 0.00000269, -0.00000327, -0.00000042, -0.00000054, 0.00000053, -0.00000185, 0.00000003, 0.00000366, 0.00000149, 0.00000144, 0.00000188, -0.00000670, 0.00000917, 0.00000215, -0.00001235, 0.00000871, 0.00000754, -0.00001971, 0.00000633, -0.00000177, -0.00001559, 0.00000146, 0.00000275, -0.00000806, 0.00000911, -0.00001035, 0.00000337, -0.00000776, -0.00000770, -0.00000089, -0.00000533, -0.00001289, 0.00001669, -0.00001813, 0.00001764, -0.00001708, -0.00000349, 0.00000258, -0.00001386, -0.00001869, 0.00003773, -0.00006654, 0.00004388, -0.00005186, 0.00001318, -0.00000254, -0.00004692, -0.00005466, 0.00013439, -0.00035621, 0.00043157, -0.00072780, 0.00051924, -0.00065562, -0.00225843, 0.00532442, -0.00031433, -0.00182726, 0.00072685, -0.00058918, 0.00013666, -0.00009039, -0.00003742, -0.00000407, -0.00001552, -0.00006820, 0.00006177, -0.00006841, 0.00002989, -0.00003030, -0.00000188, 0.00000149, -0.00001355, -0.00001276, 0.00001702, -0.00002846, 0.00001164, -0.00000690, 0.00000687, -0.00000334, -0.00001004, -0.00000310, 0.00000166, -0.00001549, 0.00000324, -0.00000716, 0.00000411, -0.00000268, -0.00000518, 0.00000621, -0.00000428, -0.00001733, 0.00000851, -0.00000701, -0.00000948, 0.00001045, -0.00000139, -0.00000901, 0.00000988, 0.00000099, -0.00000425, 0.00000247, 0.00000254, 0.00000063, 0.00000222, -0.00000090, -0.00000005, -0.00000004, -0.00000041, -0.00000151, -0.00000154, 0.00000546, 0.00000688, -0.00000067, -0.00000007, 0.00000036, -0.00000307, -0.00000157, },
		{ 0.00000033, 0.00000068, -0.00000063, -0.00000335, -0.00000062, 0.00000100, -0.00000181, 0.00000457, 0.00000762, 0.00000010, -0.00000253, -0.00000028, -0.00000037, 0.00000031, -0.00000163, 0.00000146, 0.00000225, 0.00000145, 0.00000276, -0.00000160, -0.00000376, 0.00001131, -0.00000425, -0.00000823, 0.00001221, -0.00000209, -0.00001511, 0.00001028, -0.00001070, -0.00001146, 0.00000589, -0.00000210, -0.00000532, 0.00000712, -0.00000981, 0.00000477, -0.00001266, -0.00000356, -0.00000021, -0.00000958, -0.00000701, 0.00001179, -0.00001392, 0.00001774, -0.00002586, 0.00000726, -0.00000224, -0.00001886, -0.00000351, 0.00001629, -0.00005214, 0.00004475, -0.00006917, 0.00004302, -0.00003139, -0.00004759, -0.00000445, 0.00002603, -0.00021842, 0.00031230, -0.00072751, 0.00072752, -0.00131439, -0.00138925, 0.00555658, -0.00138925, -0.00131439, 0.00072752, -0.00072751, 0.00031230, -0.00021842, 0.00002603, -0.00000445, -0.00004759, -0.00003139, 0.00004302, -0.00006917, 0.00004475, -0.00005214, 0.00001629, -0.00000351, -0.00001886, -0.00000224, 0.00000726, -0.00002586, 0.00001774, -0.00001392, 0.00001179, -0.00000701, -0.00000958, -0.00000021, -0.00000356, -0.00001266, 0.00000477, -0.00000981, 0.00000712, -0.00000532, -0.00000210, 0.00000589, -0.00001146, -0.00001070, 0.00001028, -0.00001511, -0.00000209, 0.00001221, -0.00000823, -0.00000425, 0.00001131, -0.00000376, -0.00000160, 0.00000276, 0.00000145, 0.00000225, 0.00000146, -0.00000163, 0.00000031, -0.00000037, -0.00000028, -0.00000253, 0.00000010, 0.00000762, 0.00000457, -0.00000181, 0.00000100, -0.00000062, -0.00000335, -0.00000063, 0.00000068, },
		{ 0.00000056, -0.00000157, -0.00000307, 0.00000036, -0.00000007, -0.00000067, 0.00000688, 0.00000546, -0.00000154, -0.00000151, -0.00000041, -0.00000004, -0.00000005, -0.00000090, 0.00000222, 0.00000063, 0.00000254, 0.00000247, -0.00000425, 0.00000099, 0.00000988, -0.00000901, -0.00000139, 0.00001045, -0.00000948, -0.00000701, 0.00000851, -0.00001733, -0.00000428, 0.00000621, -0.00000518, -0.00000268, 0.00000411, -0.00000716, 0.00000324, -0.00001549, 0.00000166, -0.00000310, -0.00001004, -0.00000334, 0.00000687, -0.00000690, 0.00001164, -0.00002846, 0.00001702, -0.00001276, -0.00001355, 0.00000149, -0.00000188, -0.00003030, 0.00002989, -0.00006841, 0.00006177, -0.00006820, -0.00001552, -0.00000407, -0.00003742, -0.00009039, 0.00013666, -0.00058918, 0.00072685, -0.00182726, -0.00031433, 0.00532442, -0.00225843, -0.00065562, 0.00051924, -0.00072780, 0.00043157, -0.00035621, 0.00013439, -0.00005466, -0.00004692, -0.00000254, 0.00001318, -0.00005186, 0.00004388, -0.00006654, 0.00003773, -0.00001869, -0.00001386, 0.00000258, -0.00000349, -0.00001708, 0.00001764, -0.00001813, 0.00001669, -0.00001289, -0.00000533, -0.00000089, -0.00000770, -0.00000776, 0.00000337, -0.00001035, 0.00000911, -0.00000806, 0.00000275, 0.00000146, -0.00001559, -0.00000177, 0.00000633, -0.00001971, 0.00000754, 0.00000871, -0.00001235, 0.00000215, 0.00000917, -0.00000670, 0.00000188, 0.00000144, 0.00000149, 0.00000366, 0.00000003, -0.00000185, 0.00000053, -0.00000054, -0.00000042, -0.00000327, 0.00000269, 0.00000849, 0.00000174, -0.00000176, 0.00000166, -0.00000191, -0.00000302, 0.00000023, 0.00000060, 0.00000037, },
		{ -0.00000237, -0.00000231, 0.00000075, -0.00000102, 0.00000149, 0.00000800, 0.00000265, -0.00000208, -0.00000065, -0.00000062, 0.00000035, -0.00000044, 0.00000016, 0.00000212, -0.00000046, 0.00000399, 0.00000061, -0.00000483, 0.00000572, 0.00000531, -0.00001019, 0.00000532, 0.00000429, -0.00001211, 0.00000106, 0.00000203, -0.00001932, 0.00000292, 0.00000350, -0.00000579, -0.00000152, 0.00000175, -0.00000389, -0.00000075, -0.00001492, 0.00000531, -0.00000733, -0.00000727, -0.00000317, 0.00000474, -0.00000067, 0.00000181, -0.00002370, 0.00002075, -0.00002262, -0.00000211, -0.00000484, -0.00000951, -0.00001135, 0.00000756, -0.00005062, 0.00005933, -0.00009255, 0.00002688, -0.00004261, -0.00004259, -0.00001514, -0.00003137, -0.00037030, 0.00051350, -0.00206989, 0.00078422, 0.00467764, -0.00279269, -0.00000231, 0.00016614, -0.00057813, 0.00044636, -0.00045174, 0.00025274, -0.00014447, -0.00000314, -0.00000145, -0.00001012, -0.00002541, 0.00002659, -0.00006577, 0.00005266, -0.00003743, 0.00000057, -0.00000235, -0.00000916, -0.00000653, 0.00001205, -0.00001746, 0.00001849, -0.00001809, 0.00000118, -0.00000578, -0.00000833, -0.00000329, 0.00000010, -0.00000843, 0.00000882, -0.00000906, 0.00000740, -0.00000644, -0.00001443, 0.00000584, -0.00000194, -0.00001850, 0.00001582, 0.00000125, -0.00001195, 0.00000754, 0.00000448, -0.00000685, 0.00000464, -0.00000076, 0.00000286, 0.00000418, -0.00000159, -0.00000161, 0.00000057, -0.00000056, -0.00000096, -0.00000326, 0.00000570, 0.00000773, -0.00000078, -0.00000080, 0.00000153, -0.00000308, -0.00000217, 0.00000082, 0.00000042, 0.00000045, 0.00000019, },
		{ -0.00000133, 0.00000050, -0.00000131, 0.00000415, 0.00000762, -0.00000003, -0.00000172, -0.00000017, -0.00000068, 0.00000069, -0.00000070, 0.00000124, 0.00000133, -0.00000064, 0.00000488, -0.00000201, -0.00000298, 0.00000857, -0.00000066, -0.00000734, 0.00000915, -0.00000374, -0.00000953, 0.00000588, -0.00000640, -0.00001636, 0.00000766, -0.00000023, -0.00000457, -0.00000215, 0.00000109, -0.00000175, -0.00000544, -0.00001132, 0.00000582, -0.00001038, -0.00000335, -0.00000574, 0.00000641, 0.00000157, -0.00000747, -0.00001401, 0.00001685, -0.00002642, 0.00000870, -0.00001824, -0.00000499, -0.00000303, -0.00001120, -0.00002522, 0.00003692, -0.00009241, 0.00005584, -0.00009459, -0.00000097, -0.00001091, -0.00013610, -0.00015427, 0.00014813, -0.00198221, 0.00171853, 0.00375102, -0.00294632, 0.00051080, -0.00023079, -0.00031499, 0.00034236, -0.00046384, 0.00033728, -0.00024465, 0.00007584, -0.00003774, -0.00000953, -0.00000548, 0.00000127, -0.00004915, 0.00005358, -0.00005063, 0.00001842, -0.00001637, -0.00000567, 0.00000019, 0.00000465, -0.00001265, 0.00001566, -0.00001965, 0.00000687, -0.00001334, -0.00000454, -0.00000164, -0.00000282, -0.00000519, 0.00000622, -0.00000706, 0.00000933, -0.00001537, -0.00000771, 0.00000868, -0.00001112, -0.00001146, 0.00001950, -0.00000720, -0.00000710, 0.00000970, -0.00000071, -0.00000444, 0.00000559, -0.00000266, 0.00000512, 0.00000343, -0.00000284, -0.00000114, 0.00000050, -0.00000056, -0.00000176, -0.00000216, 0.00000834, 0.00000551, -0.00000229, 0.00000051, 0.00000059, -0.00000373, -0.00000103, 0.00000108, 0.00000025, 0.00000051, -0.00000041, -0.00000282, },
		{ -0.00000019, -0.00000064, 0.00000659, 0.00000585, -0.00000197, -0.00000088, -0.00000007, -0.00000048, 0.00000085, -0.00000075, 0.00000203, 0.00000030, 0.00000003, 0.00000455, -0.00000413, 0.00000066, 0.00000841, -0.00000576, -0.00000168, 0.00000864, -0.00001050, -0.00000340, 0.00000578, -0.00001353, -0.00000999, 0.00000885, -0.00000311, -0.00000287, -0.00000384, 0.00000215, -0.00000181, -0.00000882, -0.00000641, 0.00000330, -0.00001070, -0.00000063, -0.00000884, 0.00001068, -0.00000128, -0.00001232, -0.00000398, 0.00000754, -0.00002227, 0.00001300, -0.00003149, 0.00000742, -0.00000708, -0.00001821, -0.00000469, 0.00000536, -0.00006860, 0.00005673, -0.00013217, 0.00005794, -0.00006579, -0.00015078, -0.00002086, -0.00026190, -0.00158646, 0.00234392, 0.00272629, -0.00276197, 0.00080124, -0.00056763, -0.00000903, 0.00014426, -0.00037810, 0.00035210, -0.00031740, 0.00016521, -0.00010515, 0.00002356, -0.00000749, -0.00001759, -0.00002446, 0.00003954, -0.00005118, 0.00003121, -0.00003398, 0.00000674, -0.00000088, 0.00000026, -0.00000703, 0.00000928, -0.00001627, 0.00000862, -0.00002040, 0.00000241, -0.00000386, -0.00000327, -0.00000270, 0.00000273, -0.00000230, 0.00000673, -0.00002187, 0.00000244, 0.00000508, -0.00001711, -0.00000106, 0.00001713, -0.00001323, 0.00000026, 0.00000777, -0.00000420, -0.00000077, 0.00000452, -0.00000324, 0.00000735, 0.00000158, -0.00000335, -0.00000071, 0.00000045, -0.00000075, -0.00000249, 0.00000004, 0.00000982, 0.00000245, -0.00000245, 0.00000149, -0.00000088, -0.00000365, 0.00000009, 0.00000103, 0.00000014, 0.00000046, -0.00000111, -0.00000283, -0.00000047, },
		{ 0.00000100, 0.00000814, 0.00000319, -0.00000285, -0.00000004, -0.00000016, -0.00000002, 0.00000080, -0.00000061, 0.00000238, -0.00000051, 0.00000103, 0.00000295, -0.00000467, 0.00000463, 0.00000536, -0.00000814, 0.00000448, 0.00000411, -0.00001356, 0.00000339, 0.00000120, -0.00001693, -0.00000275, 0.00000698, -0.00000416, -0.00000191, -0.00000538, 0.00000393, -0.00000392, -0.00000961, -0.00000234, -0.00000070, -0.00000834, -0.00000043, -0.00001014, 0.00001482, -0.00000771, -0.00001114, 0.00000203, -0.00000247, -0.00001242, 0.00000855, -0.00003815, 0.00002022, -0.00001899, -0.00001188, 0.00000176, -0.00002087, -0.00003325, 0.00002963, -0.00013740, 0.00010067, -0.00014475, -0.00008676, -0.00001771, -0.00059609, -0.00097996, 0.00259392, 0.00178293, -0.00234979, 0.00085536, -0.00076969, 0.00026037, -0.00009370, -0.00021257, 0.00028330, -0.00033085, 0.00023202, -0.00018240, 0.00008372, -0.00003897, -0.00001576, -0.00000525, 0.00001754, -0.00003795, 0.00003217, -0.00004717, 0.00002326, -0.00000985, 0.00000226, -0.00000482, 0.00000278, -0.00000925, 0.00000502, -0.00002367, 0.00000957, -0.00000897, -0.00000038, -0.00000291, 0.00000068, 0.00000325, -0.00000036, -0.00002295, 0.00001215, -0.00000400, -0.00001684, 0.00000871, 0.00000971, -0.00001445, 0.00000719, 0.00000261, -0.00000464, 0.00000236, 0.00000217, -0.00000209, 0.00000857, -0.00000076, -0.00000305, -0.00000054, 0.00000052, -0.00000128, -0.00000265, 0.00000293, 0.00000964, -0.00000054, -0.00000143, 0.00000162, -0.00000238, -0.00000286, 0.00000095, 0.00000080, 0.00000010, 0.00000025, -0.00000177, -0.00000244, 0.00000003, -0.00000092, },
		{ 0.00000835, 0.00000035, -0.00000270, 0.00000048, -0.00000015, 0.00000055, 0.00000054, -0.00000034, 0.00000234, -0.00000079, 0.00000167, 0.00000067, -0.00000316, 0.00000732, 0.00000067, -0.00000708, 0.00000876, -0.00000253, -0.00001216, 0.00000788, -0.00000575, -0.00001583, 0.00000297, 0.00000351, -0.00000353, -0.00000221, -0.00000584, 0.00000510, -0.00000688, -0.00000783, -0.00000060, -0.00000417, -0.00000473, -0.00000248, -0.00000853, 0.00001611, -0.00001447, -0.00000524, 0.00000194, -0.00000856, -0.00000179, -0.00000243, -0.00003556, 0.00002668, -0.00003088, 0.00000253, -0.00000757, -0.00003081, -0.00000311, -0.00001205, -0.00010932, 0.00010452, -0.00020586, 0.00001243, -0.00014351, -0.00075864, -0.00030527, 0.00249117, 0.00105349, -0.00184929, 0.00072211, -0.00081083, 0.00043169, -0.00030736, -0.00001115, 0.00014474, -0.00027166, 0.00024780, -0.00024059, 0.00015215, -0.00009485, 0.00001380, -0.00000496, 0.00000005, -0.00001724, 0.00002005, -0.00004939, 0.00003657, -0.00002265, 0.00001062, -0.00000880, 0.00000029, -0.00000224, -0.00000265, -0.00002143, 0.00001370, -0.00001434, 0.00000483, -0.00000638, 0.00000206, 0.00000661, -0.00000966, -0.00001762, 0.00001729, -0.00001509, -0.00000972, 0.00001392, 0.00000031, -0.00001044, 0.00001098, -0.00000363, -0.00000202, 0.00000353, -0.00000009, 0.00000037, 0.00000822, -0.00000283, -0.00000220, -0.00000064, 0.00000067, -0.00000217, -0.00000180, 0.00000577, 0.00000776, -0.00000260, 0.00000020, 0.00000074, -0.00000339, -0.00000161, 0.00000140, 0.00000051, 0.00000009, -0.00000010, -0.00000225, -0.00000181, 0.00000006, -0.00000124, 0.00000327, },
		{ -0.00000199, -0.00000188, 0.00000055, 0.00000019, 0.00000101, 0.00000013, -0.00000004, 0.00000213, -0.00000058, 0.00000148, -0.00000134, -0.00000002, 0.00000762, -0.00000381, -0.00000331, 0.00000962, -0.00000867, -0.00000735, 0.00000831, -0.00001223, -0.00001116, 0.00000579, 0.00000011, -0.00000207, -0.00000345, -0.00000501, 0.00000471, -0.00000909, -0.00000466, -0.00000138, -0.00000568, -0.00000172, -0.00000523, -0.00000468, 0.00001331, -0.00001820, 0.00000197, -0.00000319, -0.00000842, 0.00000460, -0.00001475, -0.00002575, 0.00002396, -0.00003565, 0.00001603, -0.00002658, -0.00002202, 0.00000864, -0.00004889, -0.00006300, 0.00006682, -0.00021872, 0.00008968, -0.00034914, -0.00070744, 0.00029196, 0.00213184, 0.00059720, -0.00138758, 0.00048847, -0.00071467, 0.00047930, -0.00044478, 0.00017236, -0.00002679, -0.00015065, 0.00019991, -0.00025442, 0.00020316, -0.00015785, 0.00006641, -0.00003040, -0.00000032, -0.00000061, 0.00000056, -0.00003908, 0.00004046, -0.00003290, 0.00002159, -0.00001851, 0.00000434, 0.00000051, -0.00001071, -0.00001464, 0.00001298, -0.00001705, 0.00000983, -0.00001180, 0.00000726, 0.00000518, -0.00001737, -0.00000761, 0.00001537, -0.00002353, 0.00000198, 0.00001243, -0.00000717, -0.00000301, 0.00001028, -0.00000844, 0.00000233, 0.00000230, -0.00000095, 0.00000309, 0.00000632, -0.00000401, -0.00000122, -0.00000089, 0.00000075, -0.00000316, 0.00000024, 0.00000766, 0.00000472, -0.00000322, 0.00000165, -0.00000091, -0.00000357, -0.00000027, 0.00000146, 0.00000026, 0.00000006, -0.00000055, -0.00000248, -0.00000114, -0.00000030, -0.00000083, 0.00000559, 0.00000717, },
		{ -0.00000086, 0.00000031, 0.00000094, 0.00000113, -0.00000033, 0.00000027, 0.00000196, -0.00000019, 0.00000037, -0.00000219, 0.00000366, 0.00000540, -0.00000639, 0.00000145, 0.00000701, -0.00001211, -0.00000145, 0.00000463, -0.00001587, -0.00000496, 0.00000566, -0.00000211, -0.00000079, -0.00000481, -0.00000350, 0.00000271, -0.00000941, -0.00000181, -0.00000366, -0.00000500, -0.00000055, -0.00000681, -0.00000060, 0.00000723, -0.00001711, 0.00000675, -0.00000997, -0.00000302, 0.00000413, -0.00002299, -0.00001398, 0.00001416, -0.00003057, 0.00002068, -0.00004478, -0.00000122, -0.00000155, -0.00006504, -0.00002140, 0.00000485, -0.00017734, 0.00009794, -0.00055596, -0.00046450, 0.00070313, 0.00165065, 0.00039777, -0.00104759, 0.00024668, -0.00053928, 0.00041732, -0.00048234, 0.00029311, -0.00018496, 0.00000085, 0.00009693, -0.00021233, 0.00021415, -0.00020447, 0.00012610, -0.00007756, 0.00002299, 0.00000013, -0.00001551, -0.00002122, 0.00003323, -0.00003518, 0.00002943, -0.00003007, 0.00001423, -0.00000364, -0.00001475, -0.00000671, 0.00000806, -0.00001546, 0.00001190, -0.00001640, 0.00001452, -0.00000179, -0.00001974, 0.00000303, 0.00000682, -0.00002549, 0.00001398, 0.00000481, -0.00000970, 0.00000467, 0.00000569, -0.00001005, 0.00000643, -0.00000062, 0.00000025, 0.00000486, 0.00000355, -0.00000407, -0.00000049, -0.00000111, 0.00000054, -0.00000377, 0.00000322, 0.00000791, 0.00000143, -0.00000239, 0.00000224, -0.00000279, -0.00000288, 0.00000082, 0.00000125, 0.00000007, -0.00000002, -0.00000102, -0.00000245, -0.00000062, -0.00000080, 0.00000044, 0.00000731, 0.00000490, -0.00000337, },
		{ 0.00000008, 0.00000200, 0.00000076, -0.00000067, 0.00000059, 0.00000194, -0.00000005, -0.00000126, -0.00000142, 0.00000656, 0.00000146, -0.00000624, 0.00000518, 0.00000223, -0.00001188, 0.00000298, -0.00000161, -0.00001558, 0.00000059, 0.00000356, -0.00000289, -0.00000027, -0.00000555, -0.00000228, -0.00000010, -0.00000767, -0.00000058, -0.00000589, -0.00000306, -0.00000124, -0.00000614, 0.00000144, 0.00000033, -0.00001175, 0.00000689, -0.00001454, 0.00000411, -0.00000233, -0.00002410, -0.00000588, 0.00000279, -0.00001827, 0.00001344, -0.00005270, 0.00001955, -0.00002656, -0.00005551, -0.00000416, -0.00005318, -0.00010220, 0.00001946, -0.00068430, -0.00010500, 0.00088208, 0.00117927, 0.00038344, -0.00085508, 0.00006557, -0.00035261, 0.00028930, -0.00042939, 0.00032956, -0.00029019, 0.00014235, -0.00003420, -0.00012153, 0.00017482, -0.00021400, 0.00017103, -0.00013180, 0.00006648, -0.00002135, -0.00001695, -0.00000564, 0.00001908, -0.00002807, 0.00002942, -0.00003786, 0.00002588, -0.00001407, -0.00001184, -0.00000193, 0.00000198, -0.00001030, 0.00000969, -0.00001736, 0.00002046, -0.00001238, -0.00001488, 0.00000952, -0.00000497, -0.00001966, 0.00002168, -0.00000587, -0.00000633, 0.00000944, -0.00000062, -0.00000818, 0.00000856, -0.00000376, 0.00000318, 0.00000482, 0.00000085, -0.00000326, -0.00000015, -0.00000123, -0.00000005, -0.00000350, 0.00000645, 0.00000629, -0.00000109, -0.00000066, 0.00000167, -0.00000426, -0.00000156, 0.00000147, 0.00000092, -0.00000007, -0.00000014, -0.00000146, -0.00000220, -0.00000035, -0.00000111, 0.00000239, 0.00000792, 0.00000214, -0.00000366, -0.00000007, },
		{ 0.00000307, -0.00000013, -0.00000074, 0.00000092, 0.00000202, -0.00000044, -0.00000271, 0.00000073, 0.00000768, -0.00000274, -0.00000366, 0.00000646, -0.00000265, -0.00000857, 0.00000415, -0.00000803, -0.00001176, 0.00000394, 0.00000084, -0.00000267, -0.00000049, -0.00000540, -0.00000211, -0.00000251, -0.00000467, -0.00000132, -0.00000680, -0.00000128, -0.00000271, -0.00000352, 0.00000015, -0.00000455, -0.00000459, 0.00000260, -0.00001447, 0.00000884, -0.00001114, -0.00001871, -0.00000460, -0.00000436, -0.00000507, -0.00000252, -0.00004646, 0.00002900, -0.00005264, -0.00002795, -0.00001836, -0.00008175, -0.00003072, -0.00012694, -0.00068128, 0.00026991, 0.00084888, 0.00081189, 0.00045967, -0.00078561, -0.00002528, -0.00020876, 0.00014915, -0.00032071, 0.00028830, -0.00032262, 0.00023858, -0.00015792, -0.00000567, 0.00009174, -0.00017699, 0.00018185, -0.00017290, 0.00011672, -0.00006198, 0.00000248, -0.00000278, 0.00000674, -0.00001554, 0.00002071, -0.00003745, 0.00003357, -0.00002669, -0.00000229, -0.00000328, -0.00000134, -0.00000448, 0.00000419, -0.00001341, 0.00002160, -0.00002255, -0.00000398, 0.00000849, -0.00001489, -0.00000800, 0.00002221, -0.00001531, 0.00000137, 0.00000946, -0.00000586, -0.00000405, 0.00000798, -0.00000554, 0.00000675, 0.00000285, -0.00000091, -0.00000209, -0.00000013, -0.00000133, -0.00000087, -0.00000201, 0.00000899, 0.00000321, -0.00000212, 0.00000111, 0.00000007, -0.00000486, -0.00000002, 0.00000165, 0.00000057, -0.00000017, -0.00000029, -0.00000183, -0.00000182, -0.00000037, -0.00000094, 0.00000461, 0.00000719, -0.00000041, -0.00000307, 0.00000024, 0.00000020, },
		{ -0.00000136, -0.00000041, 0.00000124, 0.00000203, -0.00000140, -0.00000333, 0.00000350, 0.00000671, -0.00000580, 0.00000016, 0.00000495, -0.00000572, -0.00000394, 0.00000171, -0.00001241, -0.00000594, 0.00000463, -0.00000135, -0.00000217, -0.00000099, -0.00000470, -0.00000315, -0.00000357, -0.00000169, -0.00000341, -0.00000604, -0.00000076, -0.00000348, -0.00000045, -0.00000410, -0.00000559, 0.00000127, -0.00000375, -0.00000990, 0.00000846, -0.00001784, -0.00001039, -0.00000954, -0.00000415, 0.00000248, -0.00001976, -0.00002939, 0.00002205, -0.00006638, 0.00000164, -0.00005571, -0.00007011, -0.00000060, -0.00029301, -0.00053796, 0.00056641, 0.00067307, 0.00058785, 0.00054195, -0.00078591, -0.00003504, -0.00013324, 0.00004136, -0.00020103, 0.00019832, -0.00028687, 0.00027110, -0.00024281, 0.00010373, -0.00001362, -0.00009949, 0.00014957, -0.00018276, 0.00015490, -0.00010937, 0.00004033, -0.00001870, 0.00000556, -0.00000556, 0.00000742, -0.00002826, 0.00003280, -0.00003571, 0.00001020, -0.00001077, 0.00000083, -0.00000170, -0.00000145, -0.00000589, 0.00001625, -0.00002774, 0.00000889, -0.00000036, -0.00001829, 0.00000502, 0.00001560, -0.00001983, 0.00001013, 0.00000497, -0.00000790, 0.00000023, 0.00000526, -0.00000505, 0.00000952, -0.00000043, -0.00000133, -0.00000108, -0.00000021, -0.00000162, -0.00000151, 0.00000061, 0.00000999, -0.00000045, -0.00000151, 0.00000210, -0.00000198, -0.00000444, 0.00000134, 0.00000148, 0.00000025, -0.00000022, -0.00000049, -0.00000212, -0.00000138, -0.00000056, -0.00000009, 0.00000655, 0.00000523, -0.00000215, -0.00000207, 0.00000012, 0.00000087, 0.00000375, },
		{ 0.00000035, 0.00000144, 0.00000177, -0.00000262, -0.00000275, 0.00000588, 0.00000411, -0.00000685, 0.00000365, 0.00000147, -0.00000597, -0.00000013, -0.00000331, -0.00001346, -0.00000006, 0.00000316, -0.00000247, -0.00000194, -0.00000123, -0.00000408, -0.00000487, -0.00000302, 0.00000017, -0.00000567, -0.00000417, -0.00000166, -0.00000257, 0.00000120, -0.00000948, -0.00000280, 0.00000362, -0.00000904, -0.00000322, 0.00000297, -0.00001937, -0.00000369, -0.00001701, 0.00000241, 0.00000093, -0.00003045, -0.00000999, 0.00000192, -0.00006123, 0.00001701, -0.00009694, -0.00002737, -0.00003309, -0.00041961, -0.00029011, 0.00072532, 0.00044440, 0.00049426, 0.00057826, -0.00079959, 0.00000191, -0.00012108, -0.00001206, -0.00010813, 0.00009797, -0.00020755, 0.00024302, -0.00027204, 0.00017864, -0.00011264, -0.00000129, 0.00007955, -0.00015282, 0.00016403, -0.00014586, 0.00008534, -0.00005152, 0.00002094, -0.00000657, -0.00000265, -0.00001460, 0.00002294, -0.00003641, 0.00002004, -0.00002111, 0.00000844, -0.00000444, -0.00000339, 0.00000149, 0.00000571, -0.00002503, 0.00001832, -0.00001385, -0.00001306, 0.00001419, 0.00000484, -0.00001795, 0.00001635, -0.00000190, -0.00000612, 0.00000263, 0.00000194, -0.00000240, 0.00001038, -0.00000392, -0.00000054, -0.00000055, -0.00000022, -0.00000227, -0.00000143, 0.00000377, 0.00000903, -0.00000358, 0.00000031, 0.00000181, -0.00000378, -0.00000320, 0.00000225, 0.00000111, -0.00000001, -0.00000019, -0.00000075, -0.00000228, -0.00000099, -0.00000074, 0.00000144, 0.00000764, 0.00000250, -0.00000280, -0.00000116, -0.00000012, 0.00000206, 0.00000369, -0.00000261, },
		{ 0.00000135, 0.00000117, -0.00000365, -0.00000108, 0.00000703, 0.00000091, -0.00000587, 0.00000555, -0.00000247, -0.00000364, 0.00000120, -0.00000898, -0.00001116, 0.00000432, 0.00000058, -0.00000261, -0.00000215, -0.00000098, -0.00000412, -0.00000642, -0.00000137, 0.00000049, -0.00000705, -0.00000227, -0.00000325, -0.00000006, 0.00000020, -0.00001362, 0.00000208, 0.00000214, -0.00001093, 0.00000223, -0.00000520, -0.00001543, -0.00000175, -0.00002236, 0.00001097, -0.00000839, -0.00003027, 0.00000247, -0.00002186, -0.00004032, 0.00000950, -0.00012136, 0.00002154, -0.00012319, -0.00045892, -0.00000359, 0.00073510, 0.00024234, 0.00048426, 0.00055551, -0.00078760, 0.00004532, -0.00014630, -0.00001419, -0.00006126, 0.00002030, -0.00011814, 0.00017537, -0.00024779, 0.00020400, -0.00017986, 0.00009120, -0.00001002, -0.00008784, 0.00013594, -0.00015537, 0.00012116, -0.00009119, 0.00005106, -0.00002318, -0.00000150, -0.00000427, 0.00000820, -0.00002779, 0.00002251, -0.00002916, 0.00001832, -0.00001235, 0.00000078, 0.00000429, -0.00000582, -0.00001488, 0.00001994, -0.00002648, -0.00000069, 0.00001586, -0.00000546, -0.00001089, 0.00001773, -0.00000820, -0.00000158, 0.00000214, -0.00000020, 0.00000131, 0.00000897, -0.00000656, 0.00000093, -0.00000051, -0.00000013, -0.00000323, -0.00000018, 0.00000655, 0.00000631, -0.00000527, 0.00000245, 0.00000028, -0.00000471, -0.00000158, 0.00000264, 0.00000065, -0.00000018, -0.00000011, -0.00000113, -0.00000227, -0.00000071, -0.00000067, 0.00000340, 0.00000751, -0.00000033, -0.00000240, -0.00000072, -0.00000005, 0.00000344, 0.00000275, -0.00000345, 0.00000143, },
		{ 0.00000034, -0.00000408, 0.00000116, 0.00000658, -0.00000180, -0.00000359, 0.00000541, -0.00000533, 0.00000003, -0.00000053, -0.00001331, -0.00000658, 0.00000641, -0.00000207, -0.00000217, -0.00000262, -0.00000037, -0.00000498, -0.00000703, 0.00000043, -0.00000036, -0.00000714, -0.00000125, -0.00000440, 0.00000295, -0.00000339, -0.00001478, 0.00000654, -0.00000162, -0.00000901, 0.00000393, -0.00001250, -0.00000833, -0.00000490, -0.00002252, 0.00001627, -0.00002030, -0.00002045, 0.00000253, -0.00003841, -0.00001452, -0.00001789, -0.00011695, 0.00004701, -0.00024081, -0.00039172, 0.00024953, 0.00062764, 0.00011467, 0.00050191, 0.00049126, -0.00073721, 0.00006585, -0.00017700, 0.00001397, -0.00005827, -0.00001740, -0.00004804, 0.00009748, -0.00018832, 0.00018178, -0.00020164, 0.00015439, -0.00009467, -0.00000462, 0.00007490, -0.00012983, 0.00013263, -0.00012290, 0.00008664, -0.00005321, 0.00001503, -0.00000521, -0.00000362, -0.00001368, 0.00001637, -0.00003055, 0.00002559, -0.00002206, 0.00001049, -0.00000044, -0.00001293, -0.00000124, 0.00001260, -0.00003275, 0.00001420, 0.00000948, -0.00001105, -0.00000198, 0.00001415, -0.00001145, 0.00000355, -0.00000087, 0.00000002, 0.00000452, 0.00000582, -0.00000770, 0.00000240, -0.00000075, -0.00000012, -0.00000419, 0.00000231, 0.00000798, 0.00000264, -0.00000515, 0.00000395, -0.00000197, -0.00000452, -0.00000004, 0.00000259, 0.00000019, -0.00000023, -0.00000004, -0.00000161, -0.00000207, -0.00000057, -0.00000013, 0.00000536, 0.00000608, -0.00000258, -0.00000136, -0.00000081, 0.00000067, 0.00000449, 0.00000109, -0.00000354, 0.00000256, 0.00000086, },
		{ -0.00000368, 0.00000326, 0.00000472, -0.00000322, -0.00000111, 0.00000362, -0.00000615, 0.00000330, -0.00000469, -0.00001493, -0.00000131, 0.00000630, -0.00000406, -0.00000161, -0.00000306, 0.00000020, -0.00000634, -0.00000645, 0.00000149, -0.00000151, -0.00000629, -0.00000135, -0.00000423, 0.00000494, -0.00000824, -0.00001263, 0.00000851, -0.00000521, -0.00000482, 0.00000129, -0.00001610, -0.00000158, -0.00001082, -0.00001747, 0.00001492, -0.00002848, -0.00000686, -0.00000898, -0.00004093, 0.00000318, -0.00005233, -0.00008629, 0.00002821, -0.00034212, -0.00023344, 0.00041447, 0.00046058, 0.00007081, 0.00050383, 0.00041691, -0.00065888, 0.00005357, -0.00018933, 0.00004589, -0.00008135, -0.00001650, -0.00001308, 0.00003552, -0.00011945, 0.00012842, -0.00017990, 0.00017559, -0.00015248, 0.00007409, -0.00000347, -0.00007287, 0.00011185, -0.00013278, 0.00011411, -0.00008762, 0.00004452, -0.00002153, -0.00000454, -0.00000148, 0.00000496, -0.00002414, 0.00002618, -0.00002858, 0.00002190, -0.00001228, -0.00001147, 0.00000981, -0.00000087, -0.00002967, 0.00002586, -0.00000216, -0.00000984, 0.00000487, 0.00000758, -0.00001069, 0.00000705, -0.00000490, 0.00000272, 0.00000587, 0.00000206, -0.00000732, 0.00000340, -0.00000100, -0.00000039, -0.00000462, 0.00000561, 0.00000744, -0.00000080, -0.00000351, 0.00000414, -0.00000417, -0.00000334, 0.00000106, 0.00000225, -0.00000024, -0.00000009, -0.00000010, -0.00000210, -0.00000168, -0.00000054, 0.00000101, 0.00000678, 0.00000362, -0.00000378, -0.00000023, -0.00000116, 0.00000212, 0.00000465, -0.00000083, -0.00000273, 0.00000336, -0.00000002, -0.00000046, },
		{ 0.00000461, 0.00000210, -0.00000314, 0.00000059, 0.00000122, -0.00000493, 0.00000473, -0.00000979, -0.00001349, 0.00000311, 0.00000469, -0.00000518, -0.00000118, -0.00000328, 0.00000035, -0.00000759, -0.00000500, 0.00000143, -0.00000213, -0.00000528, -0.00000207, -0.00000270, 0.00000479, -0.00001243, -0.00000829, 0.00000734, -0.00000653, -0.00000094, -0.00000416, -0.00001513, 0.00000193, -0.00001594, -0.00001006, 0.00000707, -0.00002871, 0.00000308, -0.00002563, -0.00002984, 0.00000413, -0.00007758, -0.00004568, -0.00003713, -0.00038629, -0.00002736, 0.00046893, 0.00029436, 0.00008923, 0.00047080, 0.00036039, -0.00057486, 0.00001672, -0.00017499, 0.00006142, -0.00010753, 0.00000761, -0.00001267, 0.00000391, -0.00006437, 0.00006718, -0.00012980, 0.00015653, -0.00017158, 0.00012778, -0.00007781, 0.00000118, 0.00006172, -0.00011359, 0.00012090, -0.00011350, 0.00007795, -0.00005084, 0.00000977, 0.00000101, -0.00000490, -0.00001312, 0.00001924, -0.00002787, 0.00002948, -0.00002710, -0.00000085, 0.00001280, -0.00001458, -0.00001821, 0.00002985, -0.00001417, -0.00000272, 0.00000692, 0.00000108, -0.00000678, 0.00000763, -0.00000804, 0.00000690, 0.00000474, -0.00000107, -0.00000592, 0.00000377, -0.00000117, -0.00000100, -0.00000400, 0.00000885, 0.00000492, -0.00000296, -0.00000121, 0.00000288, -0.00000559, -0.00000159, 0.00000160, 0.00000178, -0.00000057, 0.00000023, -0.00000041, -0.00000246, -0.00000120, -0.00000049, 0.00000268, 0.00000715, 0.00000073, -0.00000381, 0.00000050, -0.00000129, 0.00000400, 0.00000359, -0.00000236, -0.00000121, 0.00000352, -0.00000113, -0.00000097, -0.00000254, },
		{ -0.00000053, -0.00000189, 0.00000101, -0.00000066, -0.00000253, 0.00000365, -0.00001411, -0.00000963, 0.00000571, 0.00000248, -0.00000554, -0.00000094, -0.00000333, -0.00000002, -0.00000819, -0.00000337, 0.00000046, -0.00000183, -0.00000484, -0.00000252, -0.00000060, 0.00000234, -0.00001433, -0.00000376, 0.00000393, -0.00000484, 0.00000037, -0.00000969, -0.00001094, 0.00000133, -0.00001749, -0.00000435, -0.00000381, -0.00002087, 0.00000408, -0.00003881, -0.00001232, -0.00001178, -0.00008211, -0.00001779, -0.00013082, -0.00035078, 0.00017160, 0.00042532, 0.00017254, 0.00013377, 0.00040743, 0.00033507, -0.00050613, -0.00002672, -0.00014072, 0.00005351, -0.00011911, 0.00003467, -0.00003362, 0.00000259, -0.00003589, 0.00001892, -0.00007276, 0.00011118, -0.00015349, 0.00014530, -0.00012869, 0.00007240, -0.00000491, -0.00006792, 0.00010070, -0.00011914, 0.00010272, -0.00008426, 0.00003731, -0.00001067, -0.00000590, -0.00000398, 0.00000813, -0.00001925, 0.00002865, -0.00003856, 0.00001525, 0.00000547, -0.00002214, -0.00000303, 0.00002502, -0.00002178, 0.00000699, 0.00000368, -0.00000260, -0.00000185, 0.00000542, -0.00000887, 0.00001097, 0.00000148, -0.00000269, -0.00000425, 0.00000368, -0.00000132, -0.00000173, -0.00000210, 0.00001101, 0.00000110, -0.00000322, 0.00000068, 0.00000061, -0.00000580, 0.00000020, 0.00000163, 0.00000125, -0.00000072, 0.00000062, -0.00000102, -0.00000253, -0.00000077, -0.00000020, 0.00000461, 0.00000622, -0.00000185, -0.00000295, 0.00000062, -0.00000072, 0.00000565, 0.00000140, -0.00000287, 0.00000056, 0.00000290, -0.00000219, -0.00000099, -0.00000099, 0.00000487, },
		{ -0.00000013, 0.00000029, -0.00000125, -0.00000023, 0.00000038, -0.00001629, -0.00000463, 0.00000622, 0.00000046, -0.00000544, -0.00000079, -0.00000342, -0.00000072, -0.00000793, -0.00000223, -0.00000082, -0.00000082, -0.00000521, -0.00000199, 0.00000082, -0.00000146, -0.00001339, -0.00000088, 0.00000012, -0.00000102, -0.00000177, -0.00001279, -0.00000615, -0.00000206, -0.00001486, -0.00000344, -0.00001276, -0.00000885, -0.00000447, -0.00004211, 0.00000154, -0.00003627, -0.00006484, -0.00002135, -0.00022079, -0.00023927, 0.00031574, 0.00032105, 0.00011188, 0.00017090, 0.00033282, 0.00033765, -0.00046312, -0.00005951, -0.00010157, 0.00002806, -0.00010975, 0.00004875, -0.00005805, 0.00002021, -0.00003393, -0.00000494, -0.00002793, 0.00005937, -0.00011119, 0.00012802, -0.00014536, 0.00012239, -0.00006981, -0.00000753, 0.00005633, -0.00009893, 0.00010753, -0.00010937, 0.00006970, -0.00003494, 0.00000616, -0.00000362, -0.00000059, -0.00000644, 0.00001846, -0.00004097, 0.00003010, -0.00000983, -0.00001953, 0.00000949, 0.00001395, -0.00002237, 0.00001516, -0.00000304, -0.00000224, 0.00000174, 0.00000182, -0.00000705, 0.00001338, -0.00000283, -0.00000256, -0.00000293, 0.00000340, -0.00000164, -0.00000212, 0.00000082, 0.00001127, -0.00000287, -0.00000171, 0.00000132, -0.00000185, -0.00000486, 0.00000160, 0.00000131, 0.00000072, -0.00000059, 0.00000088, -0.00000184, -0.00000222, -0.00000049, 0.00000055, 0.00000631, 0.00000405, -0.00000350, -0.00000177, 0.00000031, 0.00000077, 0.00000636, -0.00000138, -0.00000205, 0.00000194, 0.00000164, -0.00000290, -0.00000049, 0.00000056, 0.00000405, -0.00000255, },
		{ -0.00000099, -0.00000042, 0.00000076, -0.00000399, -0.00001577, 0.00000013, 0.00000500, -0.00000091, -0.00000516, -0.00000067, -0.00000374, -0.00000133, -0.00000702, -0.00000190, -0.00000175, 0.00000033, -0.00000607, -0.00000036, 0.00000056, -0.00000506, -0.00001028, -0.00000050, -0.00000227, 0.00000302, -0.00000650, -0.00001234, -0.00000326, -0.00000564, -0.00000978, -0.00000785, -0.00001595, 0.00000157, -0.00001846, -0.00003434, 0.00000377, -0.00005690, -0.00003608, -0.00006223, -0.00027334, -0.00007957, 0.00037910, 0.00020201, 0.00010376, 0.00018163, 0.00026800, 0.00035369, -0.00044336, -0.00007288, -0.00007209, -0.00000175, -0.00008455, 0.00004384, -0.00007169, 0.00004116, -0.00004831, -0.00000472, -0.00000576, 0.00001881, -0.00006311, 0.00008813, -0.00012883, 0.00014066, -0.00011603, 0.00005095, -0.00000097, -0.00005601, 0.00008712, -0.00011500, 0.00009496, -0.00006408, 0.00002976, -0.00001584, -0.00000014, 0.00000359, 0.00000274, -0.00003229, 0.00003684, -0.00002687, -0.00000706, 0.00001408, 0.00000181, -0.00001650, 0.00001869, -0.00001015, 0.00000154, 0.00000252, -0.00000124, -0.00000338, 0.00001327, -0.00000687, -0.00000105, -0.00000232, 0.00000315, -0.00000225, -0.00000162, 0.00000400, 0.00000943, -0.00000579, 0.00000082, 0.00000037, -0.00000368, -0.00000315, 0.00000241, 0.00000078, 0.00000026, -0.00000017, 0.00000079, -0.00000262, -0.00000160, -0.00000037, 0.00000186, 0.00000722, 0.00000112, -0.00000390, -0.00000087, 0.00000007, 0.00000294, 0.00000556, -0.00000388, -0.00000006, 0.00000238, 0.00000014, -0.00000310, 0.00000038, 0.00000173, 0.00000246, -0.00000364, 0.00000150, },
		{ 0.00000128, -0.00000020, -0.00000804, -0.00001283, 0.00000355, 0.00000276, -0.00000149, -0.00000488, -0.00000060, -0.00000433, -0.00000150, -0.00000596, -0.00000226, -0.00000201, 0.00000106, -0.00000673, 0.00000181, -0.00000165, -0.00000701, -0.00000653, -0.00000222, -0.00000235, 0.00000533, -0.00001184, -0.00000894, -0.00000342, -0.00000702, -0.00000529, -0.00001538, -0.00001251, 0.00000564, -0.00003131, -0.00001996, -0.00000750, -0.00006332, -0.00001290, -0.00013009, -0.00026600, 0.00008704, 0.00036270, 0.00010617, 0.00012418, 0.00016470, 0.00022600, 0.00036672, -0.00043543, -0.00006869, -0.00005971, -0.00002306, -0.00005523, 0.00002412, -0.00006886, 0.00005288, -0.00006509, 0.00001012, -0.00000586, -0.00000085, -0.00002579, 0.00004302, -0.00009034, 0.00012763, -0.00013373, 0.00009212, -0.00005561, -0.00000145, 0.00004477, -0.00009583, 0.00010216, -0.00008685, 0.00005759, -0.00003892, 0.00001308, 0.00000360, -0.00001096, -0.00001570, 0.00003179, -0.00003815, 0.00001053, 0.00000893, -0.00000608, -0.00000753, 0.00001682, -0.00001477, 0.00000668, 0.00000039, -0.00000221, 0.00000060, 0.00001075, -0.00000955, 0.00000110, -0.00000237, 0.00000298, -0.00000305, 0.00000011, 0.00000642, 0.00000595, -0.00000684, 0.00000325, -0.00000186, -0.00000431, -0.00000124, 0.00000265, 0.00000015, 0.00000001, 0.00000044, 0.00000022, -0.00000309, -0.00000086, -0.00000026, 0.00000362, 0.00000688, -0.00000183, -0.00000319, -0.00000062, 0.00000052, 0.00000510, 0.00000321, -0.00000521, 0.00000247, 0.00000167, -0.00000112, -0.00000280, 0.00000142, 0.00000229, 0.00000050, -0.00000384, 0.00000259, -0.00000210, },
		{ -0.00000303, -0.00001050, -0.00000839, 0.00000510, 0.00000026, -0.00000140, -0.00000471, -0.00000065, -0.00000501, -0.00000111, -0.00000522, -0.00000286, -0.00000173, 0.00000115, -0.00000658, 0.00000351, -0.00000517, -0.00000662, -0.00000372, -0.00000471, -0.00000051, 0.00000494, -0.00001565, -0.00000449, -0.00000603, -0.00000530, -0.00000406, -0.00002228, -0.00000483, 0.00000192, -0.00003732, -0.00000659, -0.00002726, -0.00005254, -0.00001135, -0.00020176, -0.00019576, 0.00022046, 0.00028994, 0.00005312, 0.00014686, 0.00013187, 0.00020783, 0.00036606, -0.00042609, -0.00005638, -0.00006264, -0.00002923, -0.00003341, -0.00000014, -0.00005294, 0.00005041, -0.00007313, 0.00002676, -0.00001941, -0.00000005, -0.00000815, 0.00000838, -0.00004627, 0.00009333, -0.00012306, 0.00010720, -0.00009343, 0.00004976, -0.00000858, -0.00005491, 0.00008589, -0.00009291, 0.00007900, -0.00006554, 0.00003705, -0.00001030, -0.00001445, 0.00000116, 0.00001663, -0.00003836, 0.00002611, -0.00000349, -0.00000657, 0.00000016, 0.00001127, -0.00001550, 0.00001086, -0.00000340, -0.00000053, 0.00000329, 0.00000674, -0.00001038, 0.00000312, -0.00000285, 0.00000277, -0.00000367, 0.00000296, 0.00000714, 0.00000190, -0.00000590, 0.00000457, -0.00000460, -0.00000363, 0.00000036, 0.00000246, -0.00000052, 0.00000010, 0.00000097, -0.00000075, -0.00000304, -0.00000025, 0.00000012, 0.00000547, 0.00000517, -0.00000403, -0.00000190, -0.00000095, 0.00000202, 0.00000635, -0.00000015, -0.00000482, 0.00000468, -0.00000001, -0.00000172, -0.00000221, 0.00000245, 0.00000217, -0.00000142, -0.00000339, 0.00000309, -0.00000252, 0.00000296, },
		{ -0.00001067, -0.00000366, 0.00000481, -0.00000188, -0.00000085, -0.00000471, -0.00000087, -0.00000553, -0.00000032, -0.00000504, -0.00000324, -0.00000132, 0.00000083, -0.00000549, 0.00000385, -0.00000877, -0.00000423, -0.00000279, -0.00000647, 0.00000189, 0.00000219, -0.00001672, -0.00000102, -0.00000930, -0.00000141, -0.00000697, -0.00002523, 0.00000270, -0.00000714, -0.00003450, -0.00000113, -0.00004627, -0.00003064, -0.00003887, -0.00025005, -0.00007975, 0.00029462, 0.00019511, 0.00004222, 0.00015385, 0.00009927, 0.00020484, 0.00035027, -0.00040670, -0.00004695, -0.00007256, -0.00002173, -0.00002532, -0.00001854, -0.00003276, 0.00003681, -0.00006823, 0.00003504, -0.00003452, 0.00001291, -0.00000966, -0.00000722, -0.00001158, 0.00005271, -0.00009296, 0.00009660, -0.00010664, 0.00008423, -0.00005824, -0.00000275, 0.00004877, -0.00007713, 0.00008429, -0.00008529, 0.00006414, -0.00003607, -0.00000282, 0.00000921, -0.00000181, -0.00002708, 0.00003338, -0.00001761, -0.00000009, 0.00000322, 0.00000515, -0.00001287, 0.00001251, -0.00000703, 0.00000322, 0.00000363, 0.00000252, -0.00000951, 0.00000449, -0.00000346, 0.00000242, -0.00000362, 0.00000636, 0.00000573, -0.00000144, -0.00000357, 0.00000424, -0.00000690, -0.00000197, 0.00000136, 0.00000201, -0.00000115, 0.00000060, 0.00000112, -0.00000186, -0.00000244, 0.00000008, 0.00000102, 0.00000681, 0.00000240, -0.00000499, -0.00000070, -0.00000138, 0.00000446, 0.00000590, -0.00000349, -0.00000277, 0.00000574, -0.00000206, -0.00000154, -0.00000156, 0.00000336, 0.00000142, -0.00000303, -0.00000257, 0.00000324, -0.00000213, 0.00000370, -0.00000695, },
		{ 0.00000023, 0.00000315, -0.00000328, -0.00000012, -0.00000493, -0.00000117, -0.00000570, 0.00000051, -0.00000532, -0.00000315, -0.00000122, 0.00000063, -0.00000387, 0.00000252, -0.00001119, -0.00000093, -0.00000367, -0.00000660, 0.00000328, -0.00000151, -0.00001517, 0.00000025, -0.00001130, 0.00000247, -0.00001271, -0.00002297, 0.00000603, -0.00001668, -0.00002534, -0.00000629, -0.00005573, -0.00001014, -0.00009030, -0.00025415, 0.00005145, 0.00030398, 0.00011017, 0.00005835, 0.00014018, 0.00007899, 0.00020480, 0.00032569, -0.00037621, -0.00004752, -0.00007968, -0.00000802, -0.00002994, -0.00002528, -0.00001737, 0.00001993, -0.00005341, 0.00003125, -0.00004185, 0.00002692, -0.00002236, -0.00000416, 0.00000523, 0.00001967, -0.00005693, 0.00006872, -0.00009612, 0.00009488, -0.00009116, 0.00004638, 0.00000062, -0.00004198, 0.00006886, -0.00008888, 0.00008366, -0.00006570, 0.00002268, 0.00000205, -0.00001417, -0.00000937, 0.00002988, -0.00002743, 0.00000987, 0.00000078, 0.00000140, -0.00000885, 0.00001148, -0.00000898, 0.00000769, 0.00000146, -0.00000068, -0.00000757, 0.00000504, -0.00000404, 0.00000194, -0.00000251, 0.00000934, 0.00000244, -0.00000306, -0.00000095, 0.00000236, -0.00000806, 0.00000008, 0.00000172, 0.00000141, -0.00000158, 0.00000141, 0.00000065, -0.00000270, -0.00000153, 0.00000018, 0.00000252, 0.00000707, -0.00000075, -0.00000464, -0.00000013, -0.00000122, 0.00000717, 0.00000351, -0.00000564, 0.00000027, 0.00000526, -0.00000377, -0.00000077, -0.00000104, 0.00000409, 0.00000010, -0.00000414, -0.00000161, 0.00000337, -0.00000124, 0.00000295, -0.00001085, -0.00000852, },
		{ 0.00000081, -0.00000384, 0.00000054, -0.00000536, -0.00000141, -0.00000548, 0.00000105, -0.00000573, -0.00000269, -0.00000161, 0.00000106, -0.00000251, -0.00000009, -0.00001170, 0.00000195, -0.00000550, -0.00000516, 0.00000269, -0.00000446, -0.00001225, -0.00000061, -0.00001103, 0.00000419, -0.00001857, -0.00001686, 0.00000337, -0.00002189, -0.00001541, -0.00001936, -0.00005181, -0.00000431, -0.00014931, -0.00020737, 0.00016490, 0.00026217, 0.00005497, 0.00008149, 0.00011209, 0.00007473, 0.00019820, 0.00030160, -0.00033998, -0.00005867, -0.00007767, 0.00000306, -0.00004082, -0.00002096, -0.00001180, 0.00000775, -0.00003597, 0.00001830, -0.00003806, 0.00003334, -0.00003564, 0.00001007, 0.00000406, 0.00000230, -0.00002748, 0.00003599, -0.00007018, 0.00008303, -0.00010054, 0.00007976, -0.00004508, 0.00000325, 0.00003537, -0.00007219, 0.00008623, -0.00008783, 0.00005420, -0.00002063, -0.00001271, 0.00000619, 0.00001822, -0.00002934, 0.00001853, -0.00000535, 0.00000139, -0.00000567, 0.00000886, -0.00000860, 0.00001137, -0.00000259, -0.00000212, -0.00000537, 0.00000488, -0.00000455, 0.00000157, -0.00000030, 0.00001094, -0.00000184, -0.00000259, 0.00000083, -0.00000037, -0.00000779, 0.00000193, 0.00000157, 0.00000074, -0.00000165, 0.00000226, -0.00000043, -0.00000293, -0.00000063, 0.00000030, 0.00000438, 0.00000589, -0.00000343, -0.00000343, -0.00000027, 0.00000009, 0.00000909, -0.00000035, -0.00000576, 0.00000318, 0.00000343, -0.00000455, 0.00000020, -0.00000065, 0.00000456, -0.00000166, -0.00000464, -0.00000060, 0.00000372, -0.00000042, 0.00000068, -0.00001359, -0.00000468, 0.00000254, },
		{ -0.00000370, 0.00000093, -0.00000592, -0.00000140, -0.00000506, 0.00000118, -0.00000594, -0.00000218, -0.00000229, 0.00000226, -0.00000219, -0.00000303, -0.00001037, 0.00000350, -0.00000712, -0.00000304, 0.00000015, -0.00000547, -0.00000957, -0.00000241, -0.00000889, 0.00000275, -0.00002182, -0.00001008, -0.00000405, -0.00002044, -0.00001041, -0.00003368, -0.00003772, -0.00002086, -0.00019442, -0.00011940, 0.00023616, 0.00019427, 0.00003365, 0.00009576, 0.00008115, 0.00008216, 0.00018190, 0.00028520, -0.00030585, -0.00007541, -0.00006600, 0.00000561, -0.00005008, -0.00001101, -0.00001546, 0.00000456, -0.00002321, 0.00000303, -0.00002601, 0.00002917, -0.00004116, 0.00002528, -0.00000849, 0.00000107, -0.00001184, 0.00000968, -0.00004060, 0.00005706, -0.00008776, 0.00009063, -0.00007626, 0.00004588, -0.00000704, -0.00003833, 0.00006783, -0.00009227, 0.00007986, -0.00005209, 0.00000493, 0.00001122, 0.00000511, -0.00002391, 0.00002234, -0.00001186, 0.00000445, -0.00000475, 0.00000628, -0.00000630, 0.00001320, -0.00000741, -0.00000164, -0.00000360, 0.00000428, -0.00000500, 0.00000168, 0.00000263, 0.00001055, -0.00000585, -0.00000039, 0.00000099, -0.00000298, -0.00000632, 0.00000319, 0.00000109, 0.00000008, -0.00000123, 0.00000275, -0.00000185, -0.00000243, -0.00000007, 0.00000082, 0.00000607, 0.00000337, -0.00000494, -0.00000207, -0.00000073, 0.00000259, 0.00000923, -0.00000457, -0.00000373, 0.00000487, 0.00000095, -0.00000422, 0.00000093, -0.00000024, 0.00000464, -0.00000367, -0.00000446, 0.00000042, 0.00000429, -0.00000029, -0.00000262, -0.00001443, -0.00000013, 0.00000302, -0.00000149, },
		{ 0.00000093, -0.00000644, -0.00000109, -0.00000471, 0.00000101, -0.00000579, -0.00000195, -0.00000285, 0.00000393, -0.00000325, -0.00000523, -0.00000795, 0.00000352, -0.00000768, -0.00000143, -0.00000343, -0.00000434, -0.00000828, -0.00000370, -0.00000633, -0.00000118, -0.00002110, -0.00000594, -0.00001275, -0.00001351, -0.00001334, -0.00004208, -0.00002242, -0.00005798, -0.00020726, -0.00001238, 0.00025607, 0.00012656, 0.00003751, 0.00009443, 0.00005761, 0.00009280, 0.00015904, 0.00027849, -0.00027977, -0.00009065, -0.00004914, -0.00000109, -0.00005229, -0.00000221, -0.00002353, 0.00000958, -0.00001895, -0.00000781, -0.00001216, 0.00001732, -0.00003601, 0.00003328, -0.00002308, 0.00001031, -0.00001034, -0.00000395, -0.00001779, 0.00002854, -0.00006108, 0.00008014, -0.00008648, 0.00007464, -0.00004614, 0.00000324, 0.00003200, -0.00007438, 0.00008830, -0.00008056, 0.00003396, 0.00000154, -0.00000178, -0.00001558, 0.00002074, -0.00001586, 0.00000856, -0.00000607, 0.00000503, -0.00000315, 0.00001279, -0.00001184, 0.00000040, -0.00000273, 0.00000354, -0.00000540, 0.00000261, 0.00000547, 0.00000814, -0.00000845, 0.00000256, -0.00000059, -0.00000458, -0.00000422, 0.00000374, 0.00000044, -0.00000042, -0.00000037, 0.00000255, -0.00000315, -0.00000137, 0.00000007, 0.00000200, 0.00000691, 0.00000009, -0.00000501, -0.00000124, -0.00000074, 0.00000567, 0.00000707, -0.00000771, -0.00000021, 0.00000470, -0.00000121, -0.00000309, 0.00000121, 0.00000033, 0.00000412, -0.00000568, -0.00000357, 0.00000143, 0.00000491, -0.00000125, -0.00000608, -0.00001313, 0.00000405, 0.00000189, -0.00000319, -0.00000316, },
		{ -0.00000672, -0.00000054, -0.00000463, 0.00000081, -0.00000538, -0.00000213, -0.00000287, 0.00000543, -0.00000548, -0.00000595, -0.00000549, 0.00000253, -0.00000704, -0.00000122, -0.00000672, -0.00000184, -0.00000860, -0.00000355, -0.00000503, -0.00000570, -0.00001705, -0.00000628, -0.00001886, -0.00000498, -0.00002301, -0.00004058, -0.00001653, -0.00010442, -0.00017946, 0.00008720, 0.00023134, 0.00007778, 0.00005178, 0.00007997, 0.00004613, 0.00009897, 0.00013604, 0.00027838, -0.00026319, -0.00009896, -0.00003359, -0.00001320, -0.00004654, 0.00000077, -0.00003005, 0.00001825, -0.00002232, -0.00001080, -0.00000280, 0.00000409, -0.00002295, 0.00003089, -0.00003200, 0.00002182, -0.00001786, -0.00000502, -0.00000720, 0.00000749, -0.00003186, 0.00005606, -0.00007681, 0.00008373, -0.00007141, 0.00004030, -0.00001120, -0.00003746, 0.00007351, -0.00009379, 0.00006381, -0.00002055, 0.00000255, -0.00001041, 0.00001636, -0.00001641, 0.00001165, -0.00000852, 0.00000557, -0.00000032, 0.00001043, -0.00001497, 0.00000335, -0.00000292, 0.00000292, -0.00000565, 0.00000450, 0.00000731, 0.00000438, -0.00000897, 0.00000505, -0.00000339, -0.00000470, -0.00000215, 0.00000368, -0.00000030, -0.00000057, 0.00000065, 0.00000156, -0.00000383, -0.00000015, -0.00000001, 0.00000378, 0.00000634, -0.00000306, -0.00000393, -0.00000121, 0.00000039, 0.00000820, 0.00000294, -0.00000864, 0.00000352, 0.00000274, -0.00000224, -0.00000176, 0.00000114, 0.00000109, 0.00000283, -0.00000729, -0.00000201, 0.00000235, 0.00000524, -0.00000337, -0.00000879, -0.00001001, 0.00000699, -0.00000028, -0.00000401, -0.00000257, 0.00000058, },
		{ 0.00000001, -0.00000485, 0.00000082, -0.00000499, -0.00000255, -0.00000221, 0.00000612, -0.00000815, -0.00000508, -0.00000383, 0.00000135, -0.00000572, -0.00000257, -0.00000868, 0.00000086, -0.00000988, -0.00000199, -0.00000593, -0.00000867, -0.00001189, -0.00001065, -0.00002009, 0.00000062, -0.00003479, -0.00003051, -0.00002730, -0.00014359, -0.00011563, 0.00015750, 0.00017975, 0.00005487, 0.00006297, 0.00006028, 0.00004500, 0.00009717, 0.00011864, 0.00027915, -0.00025310, -0.00009901, -0.00002440, -0.00002494, -0.00003595, -0.00000282, -0.00003106, 0.00002510, -0.00002901, -0.00000688, -0.00000102, -0.00000438, -0.00000806, 0.00002026, -0.00003198, 0.00002880, -0.00002711, 0.00000143, -0.00000811, -0.00000105, -0.00000992, 0.00002917, -0.00005457, 0.00007450, -0.00007781, 0.00006258, -0.00004890, 0.00000803, 0.00003768, -0.00008404, 0.00008213, -0.00004672, 0.00001777, -0.00001304, 0.00001362, -0.00001504, 0.00001285, -0.00001073, 0.00000748, 0.00000134, 0.00000681, -0.00001628, 0.00000639, -0.00000406, 0.00000262, -0.00000556, 0.00000713, 0.00000745, 0.00000037, -0.00000750, 0.00000610, -0.00000645, -0.00000340, -0.00000064, 0.00000322, -0.00000104, -0.00000019, 0.00000141, 0.00000001, -0.00000363, 0.00000077, 0.00000011, 0.00000573, 0.00000424, -0.00000521, -0.00000240, -0.00000170, 0.00000291, 0.00000898, -0.00000202, -0.00000695, 0.00000608, -0.00000021, -0.00000185, -0.00000085, 0.00000108, 0.00000180, 0.00000074, -0.00000807, 0.00000004, 0.00000302, 0.00000496, -0.00000633, -0.00000998, -0.00000582, 0.00000820, -0.00000276, -0.00000393, -0.00000227, 0.00000006, -0.00000662, },
		{ -0.00000520, 0.00000111, -0.00000488, -0.00000286, -0.00000110, 0.00000563, -0.00001039, -0.00000309, -0.00000332, 0.00000071, -0.00000458, -0.00000493, -0.00000893, 0.00000275, -0.00001108, 0.00000007, -0.00000878, -0.00000884, -0.00000819, -0.00001668, -0.00001674, 0.00000043, -0.00004308, -0.00001809, -0.00005495, -0.00015982, -0.00003137, 0.00018769, 0.00012229, 0.00005383, 0.00006368, 0.00004369, 0.00004854, 0.00008872, 0.00010918, 0.00027580, -0.00024423, -0.00009350, -0.00002300, -0.00003169, -0.00002545, -0.00001016, -0.00002613, 0.00002658, -0.00003395, -0.00000019, -0.00000573, -0.00000504, 0.00000254, 0.00000684, -0.00002456, 0.00002838, -0.00003228, 0.00000872, -0.00001528, 0.00000191, -0.00000020, 0.00000882, -0.00002981, 0.00005412, -0.00006731, 0.00006570, -0.00007038, 0.00004834, -0.00000910, -0.00005164, 0.00007977, -0.00006577, 0.00003800, -0.00002410, 0.00001623, -0.00001476, 0.00001307, -0.00001204, 0.00001003, 0.00000145, 0.00000276, -0.00001563, 0.00000877, -0.00000581, 0.00000277, -0.00000497, 0.00000996, 0.00000560, -0.00000267, -0.00000480, 0.00000529, -0.00000876, -0.00000119, 0.00000009, 0.00000256, -0.00000167, 0.00000071, 0.00000148, -0.00000160, -0.00000263, 0.00000113, 0.00000086, 0.00000712, 0.00000101, -0.00000587, -0.00000122, -0.00000196, 0.00000634, 0.00000721, -0.00000618, -0.00000321, 0.00000654, -0.00000298, -0.00000037, -0.00000072, 0.00000142, 0.00000203, -0.00000187, -0.00000771, 0.00000224, 0.00000327, 0.00000386, -0.00000953, -0.00000928, -0.00000153, 0.00000764, -0.00000486, -0.00000323, -0.00000246, -0.00000037, -0.00000616, 0.00000033, },
		{ 0.00000154, -0.00000511, -0.00000271, -0.00000004, 0.00000407, -0.00001151, -0.00000076, -0.00000374, 0.00000085, -0.00000434, -0.00000734, -0.00000785, 0.00000343, -0.00001143, 0.00000144, -0.00001232, -0.00000638, -0.00000765, -0.00002141, -0.00001141, -0.00000530, -0.00004419, -0.00001157, -0.00009200, -0.00014434, 0.00005249, 0.00017975, 0.00007579, 0.00006431, 0.00005381, 0.00003502, 0.00005067, 0.00007801, 0.00010604, 0.00026641, -0.00023202, -0.00008725, -0.00002711, -0.00003192, -0.00001899, -0.00001697, -0.00001799, 0.00002245, -0.00003393, 0.00000451, -0.00001279, 0.00000111, 0.00000555, -0.00000369, -0.00001430, 0.00002190, -0.00003114, 0.00001175, -0.00002214, 0.00001101, -0.00000183, -0.00000029, -0.00001113, 0.00003209, -0.00004740, 0.00005243, -0.00007120, 0.00007163, -0.00005252, -0.00000528, 0.00005480, -0.00006827, 0.00005402, -0.00003950, 0.00002496, -0.00001830, 0.00001435, -0.00001291, 0.00001276, -0.00000008, -0.00000086, -0.00001330, 0.00000986, -0.00000766, 0.00000342, -0.00000383, 0.00001228, 0.00000211, -0.00000386, -0.00000203, 0.00000290, -0.00000961, 0.00000122, 0.00000007, 0.00000184, -0.00000199, 0.00000191, 0.00000065, -0.00000269, -0.00000126, 0.00000104, 0.00000238, 0.00000725, -0.00000248, -0.00000511, -0.00000085, -0.00000117, 0.00000955, 0.00000299, -0.00000808, 0.00000121, 0.00000479, -0.00000453, 0.00000143, -0.00000123, 0.00000235, 0.00000130, -0.00000443, -0.00000609, 0.00000420, 0.00000295, 0.00000193, -0.00001224, -0.00000675, 0.00000190, 0.00000571, -0.00000609, -0.00000234, -0.00000312, -0.00000043, -0.00000555, 0.00000031, -0.00000544, },
		{ -0.00000555, -0.00000199, 0.00000042, 0.00000197, -0.00001129, 0.00000119, -0.00000455, 0.00000152, -0.00000525, -0.00000893, -0.00000625, 0.00000320, -0.00001077, 0.00000134, -0.00001499, -0.00000274, -0.00001029, -0.00002278, -0.00000754, -0.00001351, -0.00003834, -0.00001729, -0.00012588, -0.00009840, 0.00011710, 0.00014567, 0.00004853, 0.00007534, 0.00003831, 0.00003433, 0.00004791, 0.00006964, 0.00010518, 0.00025254, -0.00021471, -0.00008451, -0.00003245, -0.00002741, -0.00001782, -0.00001989, -0.00001062, 0.00001533, -0.00002887, 0.00000430, -0.00001749, 0.00001005, 0.00000148, -0.00000800, -0.00000606, 0.00001328, -0.00002524, 0.00000904, -0.00002392, 0.00001962, -0.00000967, 0.00000104, -0.00000267, 0.00001621, -0.00002746, 0.00003110, -0.00005466, 0.00007255, -0.00007946, 0.00004106, 0.00001360, -0.00005062, 0.00005727, -0.00005198, 0.00003686, -0.00002622, 0.00001858, -0.00001463, 0.00001579, -0.00000336, -0.00000312, -0.00000997, 0.00000936, -0.00000903, 0.00000451, -0.00000221, 0.00001342, -0.00000219, -0.00000290, -0.00000037, -0.00000023, -0.00000884, 0.00000317, -0.00000046, 0.00000114, -0.00000182, 0.00000300, -0.00000096, -0.00000284, -0.00000006, 0.00000090, 0.00000439, 0.00000572, -0.00000525, -0.00000358, -0.00000120, 0.00000115, 0.00001115, -0.00000259, -0.00000699, 0.00000468, 0.00000160, -0.00000435, 0.00000270, -0.00000186, 0.00000364, -0.00000065, -0.00000621, -0.00000345, 0.00000553, 0.00000199, -0.00000065, -0.00001379, -0.00000291, 0.00000379, 0.00000310, -0.00000629, -0.00000170, -0.00000402, -0.00000001, -0.00000506, 0.00000002, -0.00000539, 0.00000186, },
		{ -0.00000085, -0.00000005, 0.00000001, -0.00000996, 0.00000235, -0.00000523, 0.00000217, -0.00000701, -0.00000926, -0.00000495, 0.00000269, -0.00000958, -0.00000022, -0.00001574, 0.00000016, -0.00001458, -0.00002061, -0.00000761, -0.00001997, -0.00002970, -0.00003658, -0.00014369, -0.00003259, 0.00015125, 0.00010176, 0.00003984, 0.00007973, 0.00002353, 0.00003810, 0.00004035, 0.00006602, 0.00010256, 0.00023789, -0.00019373, -0.00008705, -0.00003513, -0.00002173, -0.00002034, -0.00001802, -0.00000706, 0.00000897, -0.00002135, -0.00000072, -0.00001690, 0.00001704, -0.00000620, -0.00000615, -0.00000268, 0.00000655, -0.00001823, 0.00000258, -0.00001952, 0.00002302, -0.00001740, 0.00000806, -0.00000341, 0.00000977, -0.00001469, 0.00001163, -0.00002998, 0.00005408, -0.00008297, 0.00007332, -0.00003137, -0.00001685, 0.00004362, -0.00005428, 0.00004619, -0.00003613, 0.00002605, -0.00001834, 0.00001959, -0.00000858, -0.00000310, -0.00000674, 0.00000747, -0.00000942, 0.00000586, -0.00000037, 0.00001294, -0.00000621, -0.00000021, -0.00000054, -0.00000300, -0.00000681, 0.00000424, -0.00000122, 0.00000056, -0.00000111, 0.00000350, -0.00000288, -0.00000198, 0.00000055, 0.00000127, 0.00000620, 0.00000274, -0.00000656, -0.00000213, -0.00000158, 0.00000469, 0.00001005, -0.00000778, -0.00000328, 0.00000592, -0.00000169, -0.00000273, 0.00000290, -0.00000196, 0.00000469, -0.00000365, -0.00000655, -0.00000033, 0.00000599, 0.00000039, -0.00000347, -0.00001370, 0.00000141, 0.00000387, 0.00000061, -0.00000566, -0.00000162, -0.00000480, 0.00000082, -0.00000492, -0.00000031, -0.00000505, 0.00000186, -0.00000586, },
		{ -0.00000143, -0.00000120, -0.00000805, 0.00000271, -0.00000549, 0.00000228, -0.00000899, -0.00000846, -0.00000437, 0.00000250, -0.00000864, -0.00000248, -0.00001450, 0.00000093, -0.00001836, -0.00001656, -0.00001189, -0.00002153, -0.00002448, -0.00006462, -0.00013705, 0.00003710, 0.00015380, 0.00006255, 0.00004296, 0.00007576, 0.00001392, 0.00004182, 0.00003074, 0.00006659, 0.00009617, 0.00022607, -0.00017250, -0.00009369, -0.00003337, -0.00001819, -0.00002343, -0.00001299, -0.00000801, 0.00000610, -0.00001491, -0.00000788, -0.00001106, 0.00001892, -0.00001318, -0.00000087, -0.00000410, 0.00000385, -0.00001353, -0.00000385, -0.00001126, 0.00002007, -0.00002050, 0.00001498, -0.00000878, 0.00001115, -0.00001149, 0.00000133, -0.00000823, 0.00002593, -0.00006474, 0.00008275, -0.00006629, 0.00002285, 0.00001543, -0.00004249, 0.00004696, -0.00004341, 0.00003476, -0.00002404, 0.00002439, -0.00001574, -0.00000003, -0.00000497, 0.00000494, -0.00000866, 0.00000723, 0.00000135, 0.00001081, -0.00000889, 0.00000322, -0.00000264, -0.00000444, -0.00000425, 0.00000440, -0.00000200, 0.00000022, 0.00000002, 0.00000306, -0.00000446, -0.00000048, 0.00000049, 0.00000248, 0.00000699, -0.00000095, -0.00000627, -0.00000143, -0.00000107, 0.00000840, 0.00000610, -0.00001082, 0.00000169, 0.00000457, -0.00000377, -0.00000055, 0.00000213, -0.00000114, 0.00000475, -0.00000706, -0.00000514, 0.00000256, 0.00000551, -0.00000176, -0.00000601, -0.00001185, 0.00000528, 0.00000234, -0.00000112, -0.00000465, -0.00000214, -0.00000512, 0.00000178, -0.00000522, -0.00000043, -0.00000461, 0.00000150, -0.00000576, 0.00000029, },
		{ -0.00000136, -0.00000614, 0.00000254, -0.00000540, 0.00000161, -0.00001052, -0.00000702, -0.00000450, 0.00000284, -0.00000852, -0.00000436, -0.00001219, -0.00000065, -0.00001996, -0.00001309, -0.00001844, -0.00001782, -0.00002798, -0.00009213, -0.00010514, 0.00009490, 0.00013230, 0.00003667, 0.00004936, 0.00006614, 0.00001049, 0.00004240, 0.00002248, 0.00006871, 0.00008666, 0.00021884, -0.00015443, -0.00010139, -0.00002799, -0.00001825, -0.00002429, -0.00000774, -0.00001186, 0.00000728, -0.00001209, -0.00001360, -0.00000257, 0.00001529, -0.00001633, 0.00000430, -0.00000808, 0.00000479, -0.00001260, -0.00000701, -0.00000310, 0.00001303, -0.00001788, 0.00001782, -0.00001354, 0.00001555, -0.00001520, 0.00000209, 0.00000238, 0.00000031, -0.00003409, 0.00006896, -0.00008126, 0.00005598, -0.00001915, -0.00001807, 0.00003579, -0.00004312, 0.00004097, -0.00003007, 0.00002955, -0.00002428, 0.00000636, -0.00000598, 0.00000304, -0.00000707, 0.00000848, 0.00000256, 0.00000741, -0.00000957, 0.00000618, -0.00000608, -0.00000407, -0.00000201, 0.00000386, -0.00000267, 0.00000029, 0.00000118, 0.00000169, -0.00000513, 0.00000106, 0.00000012, 0.00000445, 0.00000612, -0.00000428, -0.00000483, -0.00000159, 0.00000099, 0.00001076, 0.00000019, -0.00001061, 0.00000606, 0.00000129, -0.00000390, 0.00000111, 0.00000099, 0.00000047, 0.00000327, -0.00000988, -0.00000215, 0.00000458, 0.00000422, -0.00000429, -0.00000771, -0.00000850, 0.00000786, -0.00000020, -0.00000174, -0.00000382, -0.00000301, -0.00000480, 0.00000251, -0.00000580, -0.00000019, -0.00000432, 0.00000095, -0.00000515, 0.00000098, -0.00000336, },
		{ -0.00000464, 0.00000219, -0.00000527, 0.00000028, -0.00001119, -0.00000552, -0.00000503, 0.00000354, -0.00000928, -0.00000509, -0.00001012, -0.00000368, -0.00001901, -0.00001219, -0.00002420, -0.00001157, -0.00004194, -0.00010874, -0.00005486, 0.00013013, 0.00009895, 0.00002566, 0.00005255, 0.00005542, 0.00001133, 0.00003930, 0.00001767, 0.00006937, 0.00007652, 0.00021547, -0.00014123, -0.00010700, -0.00002148, -0.00002115, -0.00002179, -0.00000482, -0.00001585, 0.00001086, -0.00001323, -0.00001529, 0.00000495, 0.00000826, -0.00001493, 0.00000678, -0.00001178, 0.00000723, -0.00001451, -0.00000574, 0.00000156, 0.00000573, -0.00001159, 0.00001583, -0.00001445, 0.00001804, -0.00002022, 0.00001024, -0.00000035, -0.00001312, -0.00000390, 0.00003974, -0.00007389, 0.00007266, -0.00004930, 0.00001235, 0.00001376, -0.00003242, 0.00004069, -0.00003355, 0.00003330, -0.00003268, 0.00001534, -0.00001050, 0.00000320, -0.00000545, 0.00000968, 0.00000283, 0.00000352, -0.00000817, 0.00000759, -0.00000975, -0.00000201, -0.00000071, 0.00000302, -0.00000319, 0.00000085, 0.00000191, -0.00000023, -0.00000468, 0.00000207, 0.00000004, 0.00000661, 0.00000354, -0.00000632, -0.00000315, -0.00000206, 0.00000456, 0.00001043, -0.00000588, -0.00000723, 0.00000823, -0.00000250, -0.00000221, 0.00000149, 0.00000037, 0.00000220, 0.00000025, -0.00001116, 0.00000175, 0.00000531, 0.00000236, -0.00000699, -0.00000808, -0.00000430, 0.00000865, -0.00000293, -0.00000135, -0.00000358, -0.00000379, -0.00000395, 0.00000271, -0.00000636, 0.00000037, -0.00000438, 0.00000050, -0.00000415, 0.00000094, -0.00000532, -0.00000053, },
		{ 0.00000187, -0.00000539, -0.00000131, -0.00001096, -0.00000437, -0.00000557, 0.00000419, -0.00001053, -0.00000455, -0.00000932, -0.00000666, -0.00001650, -0.00001433, -0.00002667, -0.00000731, -0.00006342, -0.00010695, 0.00000184, 0.00013975, 0.00006585, 0.00002562, 0.00004993, 0.00004740, 0.00001324, 0.00003412, 0.00001632, 0.00006682, 0.00006853, 0.00021353, -0.00013238, -0.00010882, -0.00001648, -0.00002471, -0.00001681, -0.00000518, -0.00001763, 0.00001419, -0.00001658, -0.00001254, 0.00000869, 0.00000107, -0.00001045, 0.00000595, -0.00001332, 0.00000877, -0.00001684, -0.00000136, 0.00000156, 0.00000143, -0.00000512, 0.00001108, -0.00001141, 0.00001609, -0.00002124, 0.00001893, -0.00001232, -0.00001097, 0.00001465, 0.00000749, -0.00004977, 0.00006927, -0.00006584, 0.00003953, -0.00001361, -0.00001215, 0.00003160, -0.00003156, 0.00003324, -0.00003859, 0.00002502, -0.00001818, 0.00000643, -0.00000489, 0.00001114, 0.00000176, 0.00000013, -0.00000524, 0.00000686, -0.00001248, 0.00000106, -0.00000060, 0.00000223, -0.00000349, 0.00000186, 0.00000179, -0.00000207, -0.00000333, 0.00000230, 0.00000076, 0.00000806, -0.00000017, -0.00000664, -0.00000204, -0.00000189, 0.00000869, 0.00000691, -0.00001007, -0.00000194, 0.00000749, -0.00000522, 0.00000037, 0.00000055, 0.00000090, 0.00000306, -0.00000370, -0.00001028, 0.00000561, 0.00000470, 0.00000015, -0.00000951, -0.00000682, -0.00000016, 0.00000764, -0.00000505, -0.00000039, -0.00000406, -0.00000403, -0.00000293, 0.00000234, -0.00000658, 0.00000101, -0.00000482, 0.00000041, -0.00000311, 0.00000013, -0.00000680, 0.00000097, -0.00000370, },
		{ -0.00000595, -0.00000274, -0.00001007, -0.00000372, -0.00000593, 0.00000446, -0.00001162, -0.00000326, -0.00001006, -0.00000836, -0.00001412, -0.00001832, -0.00002529, -0.00000927, -0.00008565, -0.00008482, 0.00005215, 0.00012800, 0.00004124, 0.00003032, 0.00004252, 0.00004357, 0.00001370, 0.00002925, 0.00001668, 0.00006128, 0.00006423, 0.00021030, -0.00012569, -0.00010712, -0.00001441, -0.00002667, -0.00001144, -0.00000792, -0.00001645, 0.00001513, -0.00001941, -0.00000704, 0.00000784, -0.00000360, -0.00000541, 0.00000297, -0.00001250, 0.00000816, -0.00001723, 0.00000340, -0.00000185, 0.00000129, -0.00000130, 0.00000675, -0.00000691, 0.00001060, -0.00001602, 0.00002178, -0.00002554, 0.00000307, 0.00001631, -0.00001581, -0.00001983, 0.00004958, -0.00006472, 0.00005518, -0.00003833, 0.00001304, 0.00001439, -0.00002265, 0.00002738, -0.00003941, 0.00003256, -0.00002728, 0.00001277, -0.00000626, 0.00001327, -0.00000098, -0.00000172, -0.00000190, 0.00000415, -0.00001333, 0.00000418, -0.00000152, 0.00000175, -0.00000354, 0.00000307, 0.00000067, -0.00000317, -0.00000166, 0.00000195, 0.00000242, 0.00000800, -0.00000398, -0.00000550, -0.00000184, -0.00000024, 0.00001185, 0.00000086, -0.00001091, 0.00000327, 0.00000431, -0.00000584, 0.00000257, -0.00000108, 0.00000263, 0.00000222, -0.00000742, -0.00000728, 0.00000847, 0.00000303, -0.00000220, -0.00001143, -0.00000400, 0.00000303, 0.00000528, -0.00000605, 0.00000054, -0.00000502, -0.00000352, -0.00000220, 0.00000160, -0.00000631, 0.00000143, -0.00000543, 0.00000079, -0.00000238, -0.00000122, -0.00000750, 0.00000275, -0.00000328, 0.00000162, },
		{ -0.00000366, -0.00000887, -0.00000349, -0.00000610, 0.00000430, -0.00001209, -0.00000205, -0.00001180, -0.00000842, -0.00001332, -0.00002202, -0.00002182, -0.00001917, -0.00010054, -0.00004659, 0.00008642, 0.00010366, 0.00002802, 0.00003424, 0.00003333, 0.00004313, 0.00001184, 0.00002643, 0.00001654, 0.00005448, 0.00006330, 0.00020417, -0.00011863, -0.00010364, -0.00001498, -0.00002592, -0.00000777, -0.00001103, -0.00001324, 0.00001306, -0.00001945, -0.00000159, 0.00000363, -0.00000468, -0.00000195, -0.00000023, -0.00001054, 0.00000578, -0.00001468, 0.00000598, -0.00000597, 0.00000414, -0.00000093, 0.00000509, -0.00000406, 0.00000497, -0.00000634, 0.00001602, -0.00003219, 0.00002046, 0.00000333, -0.00002323, 0.00000432, 0.00002275, -0.00004843, 0.00005528, -0.00005299, 0.00003627, -0.00000712, -0.00000777, 0.00001522, -0.00003334, 0.00003497, -0.00003499, 0.00002100, -0.00000978, 0.00001633, -0.00000547, -0.00000121, 0.00000048, 0.00000036, -0.00001203, 0.00000636, -0.00000296, 0.00000164, -0.00000329, 0.00000408, -0.00000124, -0.00000312, -0.00000034, 0.00000160, 0.00000464, 0.00000607, -0.00000679, -0.00000371, -0.00000221, 0.00000310, 0.00001248, -0.00000598, -0.00000810, 0.00000647, 0.00000012, -0.00000429, 0.00000337, -0.00000235, 0.00000491, -0.00000060, -0.00000965, -0.00000285, 0.00000973, 0.00000073, -0.00000458, -0.00001230, -0.00000006, 0.00000455, 0.00000240, -0.00000587, 0.00000087, -0.00000598, -0.00000233, -0.00000212, 0.00000088, -0.00000564, 0.00000142, -0.00000593, 0.00000149, -0.00000222, -0.00000271, -0.00000736, 0.00000444, -0.00000325, 0.00000135, -0.00000689, },
		{ -0.00000770, -0.00000354, -0.00000621, 0.00000391, -0.00001189, -0.00000156, -0.00001361, -0.00000742, -0.00001453, -0.00002360, -0.00001951, -0.00003523, -0.00010174, -0.00000097, 0.00010066, 0.00007639, 0.00002444, 0.00003447, 0.00002531, 0.00004404, 0.00000846, 0.00002592, 0.00001444, 0.00004856, 0.00006401, 0.00019526, -0.00010961, -0.00010049, -0.00001665, -0.00002287, -0.00000677, -0.00001262, -0.00000997, 0.00000909, -0.00001604, 0.00000137, -0.00000155, -0.00000285, -0.00000094, -0.00000218, -0.00000904, 0.00000324, -0.00000995, 0.00000530, -0.00000832, 0.00000756, -0.00000275, 0.00000630, -0.00000463, 0.00000287, 0.00000320, 0.00000365, -0.00002825, 0.00003197, -0.00001628, -0.00001502, 0.00001492, -0.00000060, -0.00002462, 0.00004149, -0.00005371, 0.00005099, -0.00002727, 0.00000968, -0.00000154, -0.00002034, 0.00003019, -0.00003816, 0.00002873, -0.00001467, 0.00002010, -0.00001142, 0.00000199, 0.00000062, -0.00000312, -0.00000901, 0.00000704, -0.00000433, 0.00000181, -0.00000272, 0.00000448, -0.00000340, -0.00000196, 0.00000019, 0.00000190, 0.00000656, 0.00000261, -0.00000787, -0.00000221, -0.00000225, 0.00000741, 0.00000970, -0.00001138, -0.00000271, 0.00000657, -0.00000329, -0.00000145, 0.00000252, -0.00000235, 0.00000660, -0.00000489, -0.00000947, 0.00000186, 0.00000925, -0.00000178, -0.00000682, -0.00001174, 0.00000423, 0.00000416, -0.00000010, -0.00000489, 0.00000039, -0.00000638, -0.00000089, -0.00000275, 0.00000058, -0.00000486, 0.00000098, -0.00000607, 0.00000226, -0.00000266, -0.00000397, -0.00000650, 0.00000584, -0.00000350, 0.00000097, -0.00000803, -0.00000397, },
		{ -0.00000377, -0.00000632, 0.00000358, -0.00001133, -0.00000193, -0.00001469, -0.00000647, -0.00001699, -0.00002257, -0.00002155, -0.00005259, -0.00008687, 0.00004196, 0.00009672, 0.00005364, 0.00002628, 0.00003090, 0.00002005, 0.00004438, 0.00000512, 0.00002667, 0.00001031, 0.00004487, 0.00006422, 0.00018508, -0.00009853, -0.00009901, -0.00001760, -0.00001906, -0.00000797, -0.00001186, -0.00000840, 0.00000529, -0.00001021, 0.00000079, -0.00000537, 0.00000012, -0.00000184, -0.00000248, -0.00000895, 0.00000220, -0.00000500, 0.00000207, -0.00000779, 0.00000940, -0.00000451, 0.00000866, -0.00000797, 0.00000603, 0.00000774, -0.00000977, -0.00001515, 0.00003183, -0.00003267, 0.00000229, 0.00001068, -0.00001266, -0.00000278, 0.00002027, -0.00004153, 0.00005351, -0.00004064, 0.00002501, -0.00001928, -0.00000263, 0.00001808, -0.00003448, 0.00003312, -0.00001923, 0.00002373, -0.00001795, 0.00000753, -0.00000220, -0.00000477, -0.00000540, 0.00000618, -0.00000513, 0.00000213, -0.00000194, 0.00000401, -0.00000513, -0.00000014, -0.00000006, 0.00000321, 0.00000723, -0.00000146, -0.00000724, -0.00000159, -0.00000100, 0.00001120, 0.00000387, -0.00001349, 0.00000324, 0.00000375, -0.00000458, 0.00000127, 0.00000059, -0.00000077, 0.00000658, -0.00000943, -0.00000668, 0.00000564, 0.00000736, -0.00000422, -0.00000865, -0.00000960, 0.00000790, 0.00000215, -0.00000154, -0.00000377, -0.00000073, -0.00000591, 0.00000025, -0.00000387, 0.00000088, -0.00000433, 0.00000033, -0.00000577, 0.00000278, -0.00000357, -0.00000473, -0.00000515, 0.00000679, -0.00000397, 0.00000042, -0.00000912, -0.00000370, -0.00000674, },
		{ -0.00000637, 0.00000345, -0.00001085, -0.00000283, -0.00001485, -0.00000647, -0.00001927, -0.00002011, -0.00002940, -0.00006511, -0.00005837, 0.00007390, 0.00008058, 0.00003896, 0.00002915, 0.00002523, 0.00001746, 0.00004330, 0.00000310, 0.00002711, 0.00000530, 0.00004342, 0.00006250, 0.00017556, -0.00008655, -0.00009922, -0.00001677, -0.00001615, -0.00000995, -0.00000928, -0.00000921, 0.00000355, -0.00000409, -0.00000272, -0.00000663, 0.00000249, -0.00000341, -0.00000182, -0.00001006, 0.00000344, -0.00000178, -0.00000175, -0.00000505, 0.00000889, -0.00000445, 0.00000981, -0.00001147, 0.00001317, 0.00000483, -0.00001800, 0.00000117, 0.00002000, -0.00003852, 0.00001936, -0.00000352, -0.00001119, 0.00000961, 0.00000009, -0.00002187, 0.00004431, -0.00004414, 0.00003397, -0.00003356, 0.00001581, 0.00000082, -0.00002342, 0.00003187, -0.00002137, 0.00002581, -0.00002368, 0.00001425, -0.00000786, -0.00000348, -0.00000263, 0.00000440, -0.00000517, 0.00000247, -0.00000114, 0.00000267, -0.00000590, 0.00000166, -0.00000067, 0.00000537, 0.00000603, -0.00000493, -0.00000552, -0.00000174, 0.00000203, 0.00001275, -0.00000343, -0.00001160, 0.00000753, -0.00000064, -0.00000339, 0.00000259, -0.00000124, 0.00000182, 0.00000425, -0.00001271, -0.00000189, 0.00000759, 0.00000468, -0.00000650, -0.00000972, -0.00000612, 0.00001011, -0.00000073, -0.00000168, -0.00000318, -0.00000196, -0.00000461, 0.00000065, -0.00000499, 0.00000168, -0.00000429, -0.00000023, -0.00000515, 0.00000287, -0.00000467, -0.00000487, -0.00000354, 0.00000723, -0.00000460, -0.00000026, -0.00000998, -0.00000302, -0.00000615, -0.00000408, },
		{ 0.00000341, -0.00001073, -0.00000370, -0.00001450, -0.00000767, -0.00002016, -0.00001855, -0.00004194, -0.00006763, -0.00002241, 0.00009059, 0.00005966, 0.00003218, 0.00003021, 0.00001959, 0.00001639, 0.00004109, 0.00000271, 0.00002613, 0.00000096, 0.00004304, 0.00005871, 0.00016807, -0.00007528, -0.00010008, -0.00001425, -0.00001512, -0.00001114, -0.00000629, -0.00001174, 0.00000466, 0.00000025, -0.00000742, -0.00000550, 0.00000335, -0.00000451, -0.00000126, -0.00001127, 0.00000643, -0.00000120, -0.00000422, -0.00000186, 0.00000684, -0.00000235, 0.00000830, -0.00001211, 0.00002060, -0.00000430, -0.00001730, 0.00001328, 0.00000180, -0.00003192, 0.00002828, -0.00001941, 0.00000024, 0.00000984, -0.00001202, -0.00000221, 0.00002772, -0.00003813, 0.00003446, -0.00004072, 0.00003034, -0.00001751, -0.00000674, 0.00002416, -0.00001938, 0.00002485, -0.00002696, 0.00002037, -0.00001527, 0.00000103, -0.00000200, 0.00000269, -0.00000467, 0.00000279, -0.00000059, 0.00000078, -0.00000549, 0.00000285, -0.00000094, 0.00000770, 0.00000298, -0.00000686, -0.00000371, -0.00000191, 0.00000638, 0.00001084, -0.00000989, -0.00000647, 0.00000869, -0.00000468, -0.00000047, 0.00000199, -0.00000180, 0.00000421, -0.00000007, -0.00001349, 0.00000362, 0.00000742, 0.00000183, -0.00000859, -0.00000959, -0.00000192, 0.00001039, -0.00000353, -0.00000080, -0.00000348, -0.00000266, -0.00000292, 0.00000014, -0.00000563, 0.00000265, -0.00000473, -0.00000043, -0.00000446, 0.00000250, -0.00000573, -0.00000435, -0.00000191, 0.00000711, -0.00000531, -0.00000098, -0.00001048, -0.00000209, -0.00000597, -0.00000436, -0.00000626, },
		{ -0.00001089, -0.00000417, -0.00001431, -0.00000948, -0.00001940, -0.00002016, -0.00005567, -0.00005782, 0.00001330, 0.00009229, 0.00004027, 0.00003076, 0.00002864, 0.00001541, 0.00001553, 0.00003868, 0.00000331, 0.00002359, -0.00000163, 0.00004215, 0.00005385, 0.00016287, -0.00006584, -0.00010021, -0.00001110, -0.00001582, -0.00001071, -0.00000436, -0.00001442, 0.00000801, 0.00000170, -0.00001142, -0.00000312, 0.00000276, -0.00000469, -0.00000148, -0.00001134, 0.00000968, -0.00000280, -0.00000439, 0.00000009, 0.00000497, 0.00000050, 0.00000459, -0.00000829, 0.00002412, -0.00001522, -0.00000820, 0.00001577, -0.00001488, -0.00001667, 0.00002575, -0.00002925, 0.00001449, 0.00000047, -0.00001301, 0.00001087, 0.00000999, -0.00002608, 0.00002731, -0.00003935, 0.00003737, -0.00003217, 0.00001184, 0.00001117, -0.00001271, 0.00001989, -0.00002636, 0.00002389, -0.00002247, 0.00000800, -0.00000419, 0.00000214, -0.00000414, 0.00000318, -0.00000059, -0.00000111, -0.00000412, 0.00000318, -0.00000031, 0.00000920, -0.00000120, -0.00000689, -0.00000259, -0.00000108, 0.00001071, 0.00000546, -0.00001332, -0.00000005, 0.00000651, -0.00000672, 0.00000270, -0.00000011, -0.00000053, 0.00000503, -0.00000519, -0.00001126, 0.00000841, 0.00000546, -0.00000079, -0.00001041, -0.00000798, 0.00000207, 0.00000879, -0.00000538, 0.00000044, -0.00000458, -0.00000240, -0.00000148, -0.00000100, -0.00000553, 0.00000334, -0.00000544, -0.00000021, -0.00000395, 0.00000178, -0.00000654, -0.00000321, -0.00000052, 0.00000645, -0.00000598, -0.00000170, -0.00001057, -0.00000113, -0.00000625, -0.00000444, -0.00000597, 0.00000323, },
		{ -0.00000430, -0.00001476, -0.00001095, -0.00001791, -0.00002601, -0.00006603, -0.00003698, 0.00004194, 0.00008280, 0.00002613, 0.00003149, 0.00002515, 0.00001301, 0.00001415, 0.00003693, 0.00000385, 0.00002023, -0.00000237, 0.00003962, 0.00004936, 0.00015918, -0.00005840, -0.00009868, -0.00000864, -0.00001726, -0.00000889, -0.00000418, -0.00001568, 0.00001201, 0.00000046, -0.00001351, -0.00000082, 0.00000151, -0.00000431, -0.00000238, -0.00000967, 0.00001164, -0.00000505, -0.00000270, 0.00000014, 0.00000468, 0.00000219, 0.00000079, -0.00000087, 0.00002128, -0.00002259, 0.00000474, 0.00000773, -0.00002327, -0.00000001, 0.00001400, -0.00002904, 0.00002430, -0.00001234, -0.00000463, 0.00001409, -0.00000313, -0.00001310, 0.00001592, -0.00003083, 0.00003562, -0.00003938, 0.00002781, -0.00000417, -0.00000236, 0.00001100, -0.00002123, 0.00002317, -0.00002721, 0.00001563, -0.00000890, 0.00000347, -0.00000420, 0.00000381, -0.00000136, -0.00000242, -0.00000233, 0.00000286, 0.00000133, 0.00000902, -0.00000533, -0.00000543, -0.00000236, 0.00000147, 0.00001322, -0.00000209, -0.00001256, 0.00000526, 0.00000215, -0.00000606, 0.00000460, -0.00000248, 0.00000224, 0.00000337, -0.00000945, -0.00000648, 0.00001129, 0.00000244, -0.00000303, -0.00001172, -0.00000489, 0.00000490, 0.00000595, -0.00000586, 0.00000127, -0.00000596, -0.00000113, -0.00000088, -0.00000225, -0.00000477, 0.00000347, -0.00000607, 0.00000030, -0.00000377, 0.00000087, -0.00000698, -0.00000156, 0.00000036, 0.00000538, -0.00000650, -0.00000239, -0.00001020, -0.00000038, -0.00000689, -0.00000409, -0.00000567, 0.00000272, -0.00001102, },
		{ -0.00001578, -0.00001136, -0.00001739, -0.00003524, -0.00006901, -0.00000938, 0.00005946, 0.00006750, 0.00001810, 0.00003188, 0.00002105, 0.00001188, 0.00001224, 0.00003611, 0.00000350, 0.00001717, -0.00000206, 0.00003527, 0.00004634, 0.00015575, -0.00005224, -0.00009543, -0.00000772, -0.00001823, -0.00000667, -0.00000540, -0.00001470, 0.00001493, -0.00000227, -0.00001353, 0.00000050, 0.00000040, -0.00000401, -0.00000315, -0.00000667, 0.00001150, -0.00000639, -0.00000051, -0.00000108, 0.00000619, 0.00000147, -0.00000048, 0.00000713, 0.00001268, -0.00002269, 0.00001520, -0.00000703, -0.00002066, 0.00001087, -0.00000087, -0.00001983, 0.00002540, -0.00002185, 0.00000788, 0.00000833, -0.00000855, -0.00000376, 0.00000479, -0.00001882, 0.00002665, -0.00003760, 0.00003733, -0.00001810, 0.00000930, -0.00000043, -0.00001210, 0.00001756, -0.00002755, 0.00002163, -0.00001485, 0.00000671, -0.00000532, 0.00000485, -0.00000298, -0.00000268, -0.00000078, 0.00000244, 0.00000356, 0.00000685, -0.00000827, -0.00000339, -0.00000249, 0.00000560, 0.00001241, -0.00000950, -0.00000794, 0.00000755, -0.00000240, -0.00000323, 0.00000440, -0.00000375, 0.00000530, -0.00000075, -0.00001128, -0.00000043, 0.00001174, -0.00000082, -0.00000496, -0.00001210, -0.00000076, 0.00000586, 0.00000286, -0.00000511, 0.00000116, -0.00000690, 0.00000073, -0.00000136, -0.00000299, -0.00000374, 0.00000299, -0.00000632, 0.00000086, -0.00000389, -0.00000012, -0.00000691, 0.00000044, 0.00000049, 0.00000416, -0.00000685, -0.00000300, -0.00000939, -0.00000009, -0.00000764, -0.00000319, -0.00000568, 0.00000196, -0.00001081, -0.00000450, },
		{ -0.00001078, -0.00001940, -0.00004526, -0.00006267, 0.00001926, 0.00006518, 0.00005151, 0.00001497, 0.00003077, 0.00001744, 0.00001130, 0.00001028, 0.00003585, 0.00000212, 0.00001524, -0.00000186, 0.00002978, 0.00004507, 0.00015163, -0.00004625, -0.00009122, -0.00000840, -0.00001796, -0.00000517, -0.00000688, -0.00001178, 0.00001566, -0.00000494, -0.00001217, 0.00000071, -0.00000019, -0.00000427, -0.00000288, -0.00000351, 0.00000961, -0.00000597, 0.00000079, -0.00000220, 0.00000844, -0.00000138, 0.00000233, 0.00001201, 0.00000176, -0.00001533, 0.00001803, -0.00002191, -0.00000930, 0.00001205, -0.00001200, -0.00000660, 0.00001799, -0.00002367, 0.00001836, -0.00000229, -0.00000659, -0.00000043, -0.00000229, -0.00000779, 0.00001426, -0.00002815, 0.00003850, -0.00002726, 0.00001929, -0.00001206, -0.00000072, 0.00000774, -0.00002257, 0.00002382, -0.00002007, 0.00001109, -0.00000758, 0.00000637, -0.00000532, -0.00000169, 0.00000002, 0.00000259, 0.00000551, 0.00000306, -0.00000932, -0.00000171, -0.00000200, 0.00001021, 0.00000783, -0.00001435, -0.00000126, 0.00000621, -0.00000522, 0.00000030, 0.00000231, -0.00000302, 0.00000710, -0.00000629, -0.00000984, 0.00000524, 0.00000998, -0.00000372, -0.00000666, -0.00001117, 0.00000358, 0.00000479, 0.00000050, -0.00000374, 0.00000009, -0.00000682, 0.00000246, -0.00000270, -0.00000287, -0.00000294, 0.00000216, -0.00000613, 0.00000124, -0.00000423, -0.00000108, -0.00000620, 0.00000246, -0.00000020, 0.00000308, -0.00000712, -0.00000341, -0.00000827, -0.00000047, -0.00000809, -0.00000190, -0.00000631, 0.00000131, -0.00001026, -0.00000521, -0.00001679, },
		{ -0.00002453, -0.00005269, -0.00004769, 0.00004362, 0.00006124, 0.00003828, 0.00001464, 0.00002813, 0.00001482, 0.00001072, 0.00000876, 0.00003547, 0.00000019, 0.00001456, -0.00000266, 0.00002427, 0.00004501, 0.00014658, -0.00003961, -0.00008708, -0.00000999, -0.00001648, -0.00000506, -0.00000734, -0.00000806, 0.00001411, -0.00000642, -0.00001043, 0.00000022, -0.00000041, -0.00000496, -0.00000113, -0.00000146, 0.00000714, -0.00000407, 0.00000060, -0.00000199, 0.00000978, -0.00000459, 0.00000873, 0.00001128, -0.00000684, -0.00000388, 0.00001163, -0.00003056, 0.00000499, 0.00000425, -0.00001495, 0.00000436, 0.00000616, -0.00001735, 0.00002236, -0.00001251, -0.00000033, -0.00000265, -0.00000364, -0.00000131, 0.00000307, -0.00001462, 0.00003201, -0.00002988, 0.00002510, -0.00002126, 0.00001034, -0.00000429, -0.00001281, 0.00002094, -0.00002252, 0.00001518, -0.00001050, 0.00000822, -0.00000812, 0.00000041, -0.00000012, 0.00000373, 0.00000619, -0.00000133, -0.00000853, -0.00000087, -0.00000002, 0.00001353, 0.00000044, -0.00001499, 0.00000499, 0.00000211, -0.00000524, 0.00000283, -0.00000050, -0.00000037, 0.00000639, -0.00001148, -0.00000532, 0.00000906, 0.00000682, -0.00000603, -0.00000805, -0.00000872, 0.00000711, 0.00000213, -0.00000049, -0.00000261, -0.00000147, -0.00000554, 0.00000332, -0.00000430, -0.00000191, -0.00000278, 0.00000138, -0.00000564, 0.00000135, -0.00000468, -0.00000193, -0.00000477, 0.00000407, -0.00000159, 0.00000239, -0.00000747, -0.00000345, -0.00000713, -0.00000151, -0.00000782, -0.00000064, -0.00000761, 0.00000120, -0.00000974, -0.00000654, -0.00001708, -0.00001010, },
		{ -0.00005462, -0.00002702, 0.00006019, 0.00005130, 0.00002928, 0.00001514, 0.00002459, 0.00001314, 0.00001007, 0.00000781, 0.00003445, -0.00000153, 0.00001460, -0.00000467, 0.00001964, 0.00004520, 0.00014106, -0.00003212, -0.00008384, -0.00001154, -0.00001445, -0.00000618, -0.00000603, -0.00000496, 0.00001110, -0.00000633, -0.00000908, -0.00000043, -0.00000065, -0.00000547, 0.00000174, -0.00000118, 0.00000541, -0.00000177, -0.00000060, -0.00000020, 0.00000902, -0.00000580, 0.00001626, 0.00000497, -0.00000950, 0.00000628, -0.00000140, -0.00002973, 0.00001564, -0.00000791, -0.00000940, 0.00000849, -0.00000442, -0.00000611, 0.00001874, -0.00001822, 0.00000608, -0.00000772, -0.00000020, -0.00000073, -0.00000322, -0.00000157, 0.00002077, -0.00002629, 0.00002552, -0.00002604, 0.00001849, -0.00001572, -0.00000028, 0.00001311, -0.00002077, 0.00001735, -0.00001318, 0.00000999, -0.00001093, 0.00000326, -0.00000101, 0.00000585, 0.00000493, -0.00000508, -0.00000660, -0.00000069, 0.00000368, 0.00001386, -0.00000762, -0.00001129, 0.00000855, -0.00000284, -0.00000268, 0.00000320, -0.00000248, 0.00000309, 0.00000285, -0.00001450, 0.00000108, 0.00001023, 0.00000324, -0.00000786, -0.00000881, -0.00000497, 0.00000895, -0.00000118, -0.00000004, -0.00000240, -0.00000274, -0.00000344, 0.00000294, -0.00000543, -0.00000051, -0.00000333, 0.00000098, -0.00000512, 0.00000126, -0.00000523, -0.00000248, -0.00000269, 0.00000485, -0.00000331, 0.00000213, -0.00000797, -0.00000296, -0.00000640, -0.00000289, -0.00000659, -0.00000003, -0.00000920, 0.00000183, -0.00000980, -0.00000808, -0.00001632, -0.00001068, -0.00003194, },
		{ -0.00000475, 0.00006787, 0.00003920, 0.00002432, 0.00001523, 0.00002086, 0.00001210, 0.00000949, 0.00000718, 0.00003270, -0.00000246, 0.00001453, -0.00000741, 0.00001619, 0.00004483, 0.00013591, -0.00002426, -0.00008169, -0.00001235, -0.00001278, -0.00000771, -0.00000302, -0.00000351, 0.00000780, -0.00000502, -0.00000842, -0.00000094, -0.00000121, -0.00000511, 0.00000479, -0.00000243, 0.00000515, -0.00000025, -0.00000166, 0.00000232, 0.00000625, -0.00000341, 0.00002161, -0.00000430, -0.00000546, 0.00001036, -0.00001558, -0.00002046, 0.00001835, -0.00001836, 0.00000112, 0.00000477, -0.00000918, 0.00000499, 0.00000974, -0.00001797, 0.00000939, -0.00001223, 0.00000525, -0.00000485, -0.00000314, 0.00000711, 0.00000873, -0.00001871, 0.00002109, -0.00002573, 0.00002195, -0.00002379, 0.00001210, 0.00000194, -0.00001455, 0.00001630, -0.00001449, 0.00001106, -0.00001319, 0.00000633, -0.00000223, 0.00000844, 0.00000166, -0.00000717, -0.00000451, -0.00000034, 0.00000830, 0.00001031, -0.00001373, -0.00000473, 0.00000832, -0.00000647, 0.00000109, 0.00000133, -0.00000244, 0.00000571, -0.00000267, -0.00001415, 0.00000760, 0.00000878, 0.00000006, -0.00000942, -0.00000849, -0.00000059, 0.00000868, -0.00000403, 0.00000131, -0.00000332, -0.00000297, -0.00000130, 0.00000147, -0.00000560, 0.00000071, -0.00000434, 0.00000106, -0.00000479, 0.00000110, -0.00000592, -0.00000247, -0.00000027, 0.00000452, -0.00000485, 0.00000211, -0.00000855, -0.00000191, -0.00000645, -0.00000400, -0.00000463, -0.00000052, -0.00001043, 0.00000295, -0.00001083, -0.00000908, -0.00001492, -0.00001356, -0.00003955, -0.00004964, },
		{ 0.00006762, 0.00002791, 0.00002229, 0.00001441, 0.00001744, 0.00001145, 0.00000916, 0.00000647, 0.00003056, -0.00000253, 0.00001371, -0.00001005, 0.00001358, 0.00004360, 0.00013183, -0.00001683, -0.00008022, -0.00001227, -0.00001203, -0.00000864, 0.00000088, -0.00000397, 0.00000523, -0.00000319, -0.00000828, -0.00000134, -0.00000194, -0.00000354, 0.00000694, -0.00000425, 0.00000620, -0.00000003, -0.00000150, 0.00000410, 0.00000293, 0.00000246, 0.00002223, -0.00001253, 0.00000289, 0.00000632, -0.00002509, -0.00000736, 0.00001272, -0.00002246, 0.00001122, -0.00000413, -0.00000658, 0.00001143, -0.00000050, -0.00001312, 0.00000846, -0.00001369, 0.00000948, -0.00001069, 0.00000223, 0.00000952, -0.00000051, -0.00001026, 0.00001383, -0.00002123, 0.00002041, -0.00002668, 0.00002134, -0.00000986, -0.00000496, 0.00001159, -0.00001352, 0.00001075, -0.00001433, 0.00000906, -0.00000326, 0.00001070, -0.00000303, -0.00000717, -0.00000305, 0.00000112, 0.00001226, 0.00000339, -0.00001585, 0.00000220, 0.00000479, -0.00000735, 0.00000423, -0.00000175, -0.00000013, 0.00000597, -0.00000847, -0.00001029, 0.00001249, 0.00000550, -0.00000233, -0.00001081, -0.00000671, 0.00000339, 0.00000650, -0.00000557, 0.00000266, -0.00000498, -0.00000184, 0.00000001, -0.00000048, -0.00000473, 0.00000124, -0.00000536, 0.00000147, -0.00000472, 0.00000095, -0.00000675, -0.00000162, 0.00000194, 0.00000316, -0.00000581, 0.00000201, -0.00000892, -0.00000059, -0.00000742, -0.00000418, -0.00000258, -0.00000215, -0.00001063, 0.00000387, -0.00001269, -0.00000887, -0.00001394, -0.00001887, -0.00004484, -0.00003822, 0.00001519, },
		{ 0.00001911, 0.00002186, 0.00001271, 0.00001453, 0.00001107, 0.00000908, 0.00000539, 0.00002853, -0.00000210, 0.00001198, -0.00001192, 0.00001110, 0.00004180, 0.00012906, -0.00001051, -0.00007874, -0.00001173, -0.00001221, -0.00000827, 0.00000452, -0.00000586, 0.00000388, -0.00000146, -0.00000836, -0.00000185, -0.00000237, -0.00000106, 0.00000763, -0.00000552, 0.00000765, -0.00000077, 0.00000019, 0.00000401, 0.00000116, 0.00000987, 0.00001765, -0.00001631, 0.00001109, -0.00000421, -0.00002638, 0.00000373, 0.00000210, -0.00001892, 0.00001628, -0.00001346, 0.00000150, 0.00001119, -0.00000794, -0.00000675, 0.00000442, -0.00001150, 0.00001029, -0.00001486, 0.00000990, 0.00000617, -0.00000493, -0.00000383, 0.00000638, -0.00001464, 0.00001508, -0.00002414, 0.00002536, -0.00001932, 0.00000578, 0.00000389, -0.00000991, 0.00000855, -0.00001384, 0.00001098, -0.00000373, 0.00001181, -0.00000809, -0.00000535, -0.00000243, 0.00000421, 0.00001373, -0.00000503, -0.00001332, 0.00000695, -0.00000028, -0.00000533, 0.00000530, -0.00000440, 0.00000356, 0.00000316, -0.00001251, -0.00000393, 0.00001464, 0.00000151, -0.00000400, -0.00001185, -0.00000344, 0.00000596, 0.00000326, -0.00000553, 0.00000318, -0.00000657, 0.00000039, -0.00000006, -0.00000206, -0.00000323, 0.00000092, -0.00000598, 0.00000190, -0.00000479, 0.00000074, -0.00000755, 0.00000017, 0.00000332, 0.00000119, -0.00000604, 0.00000156, -0.00000875, 0.00000049, -0.00000902, -0.00000305, -0.00000134, -0.00000437, -0.00000955, 0.00000376, -0.00001463, -0.00000738, -0.00001462, -0.00002559, -0.00004580, -0.00002248, 0.00003004, 0.00006171, },
		{ 0.00002189, 0.00001044, 0.00001222, 0.00001096, 0.00000905, 0.00000392, 0.00002698, -0.00000178, 0.00000969, -0.00001276, 0.00000810, 0.00004006, 0.00012729, -0.00000556, -0.00007668, -0.00001133, -0.00001277, -0.00000654, 0.00000700, -0.00000828, 0.00000366, -0.00000019, -0.00000845, -0.00000260, -0.00000196, 0.00000151, 0.00000701, -0.00000558, 0.00000850, -0.00000146, 0.00000272, 0.00000194, 0.00000256, 0.00001593, 0.00000969, -0.00001438, 0.00001476, -0.00001676, -0.00001960, 0.00000835, -0.00000828, -0.00001005, 0.00001446, -0.00001872, 0.00001094, 0.00000524, -0.00001036, -0.00000201, -0.00000026, -0.00000701, 0.00000744, -0.00001507, 0.00001636, -0.00000049, -0.00000458, -0.00000099, 0.00000102, -0.00000844, 0.00000830, -0.00001756, 0.00002364, -0.00002414, 0.00001502, -0.00000518, -0.00000412, 0.00000440, -0.00001140, 0.00001174, -0.00000351, 0.00001125, -0.00001237, -0.00000249, -0.00000220, 0.00000858, 0.00001145, -0.00001232, -0.00000716, 0.00000794, -0.00000458, -0.00000157, 0.00000387, -0.00000515, 0.00000693, -0.00000219, -0.00001323, 0.00000312, 0.00001383, -0.00000215, -0.00000528, -0.00001211, 0.00000083, 0.00000643, 0.00000010, -0.00000428, 0.00000245, -0.00000721, 0.00000293, -0.00000150, -0.00000258, -0.00000180, -0.00000001, -0.00000604, 0.00000208, -0.00000478, 0.00000030, -0.00000795, 0.00000268, 0.00000338, -0.00000070, -0.00000577, 0.00000075, -0.00000785, 0.00000074, -0.00001050, -0.00000074, -0.00000165, -0.00000617, -0.00000760, 0.00000215, -0.00001564, -0.00000532, -0.00001769, -0.00003184, -0.00004164, -0.00000524, 0.00003865, 0.00005277, 0.00001334, },
		{ 0.00000793, 0.00001052, 0.00001103, 0.00000884, 0.00000233, 0.00002592, -0.00000198, 0.00000747, -0.00001284, 0.00000435, 0.00003896, 0.00012589, -0.00000177, -0.00007381, -0.00001157, -0.00001295, -0.00000398, 0.00000793, -0.00001037, 0.00000413, 0.00000062, -0.00000854, -0.00000344, -0.00000046, 0.00000331, 0.00000587, -0.00000453, 0.00000825, -0.00000111, 0.00000481, -0.00000097, 0.00000737, 0.00001813, 0.00000165, -0.00000828, 0.00001160, -0.00002608, -0.00000826, 0.00000530, -0.00001385, -0.00000032, 0.00000726, -0.00001770, 0.00001747, -0.00000321, -0.00000805, -0.00000067, -0.00000317, -0.00000260, 0.00000253, -0.00001101, 0.00001897, -0.00000730, -0.00000117, -0.00000171, -0.00000118, -0.00000452, 0.00000259, -0.00000950, 0.00001740, -0.00002346, 0.00002049, -0.00001347, 0.00000267, -0.00000125, -0.00000703, 0.00001114, -0.00000270, 0.00000893, -0.00001499, 0.00000053, -0.00000158, 0.00001301, 0.00000539, -0.00001607, 0.00000028, 0.00000517, -0.00000633, 0.00000204, 0.00000077, -0.00000342, 0.00000825, -0.00000849, -0.00001014, 0.00000893, 0.00001073, -0.00000489, -0.00000644, -0.00001108, 0.00000519, 0.00000474, -0.00000192, -0.00000273, 0.00000070, -0.00000635, 0.00000477, -0.00000371, -0.00000185, -0.00000108, -0.00000105, -0.00000569, 0.00000197, -0.00000461, -0.00000051, -0.00000751, 0.00000530, 0.00000203, -0.00000188, -0.00000547, -0.00000011, -0.00000637, -0.00000016, -0.00001098, 0.00000194, -0.00000363, -0.00000659, -0.00000578, -0.00000071, -0.00001492, -0.00000403, -0.00002291, -0.00003564, -0.00003296, 0.00001080, 0.00004127, 0.00004310, 0.00001032, 0.00002165, },
		{ 0.00000941, 0.00001112, 0.00000835, 0.00000091, 0.00002507, -0.00000278, 0.00000581, -0.00001269, 0.00000003, 0.00003874, 0.00012427, 0.00000139, -0.00007036, -0.00001251, -0.00001216, -0.00000146, 0.00000755, -0.00001160, 0.00000480, 0.00000110, -0.00000874, -0.00000395, 0.00000184, 0.00000382, 0.00000515, -0.00000310, 0.00000715, 0.00000064, 0.00000532, -0.00000284, 0.00001417, 0.00001554, -0.00000333, -0.00000155, 0.00000247, -0.00002853, 0.00000249, -0.00000315, -0.00001263, 0.00000583, -0.00000148, -0.00001118, 0.00001846, -0.00001040, -0.00000321, -0.00000256, -0.00000314, -0.00000039, -0.00000188, -0.00000430, 0.00001696, -0.00001164, 0.00000283, -0.00000456, -0.00000052, -0.00000357, -0.00000025, -0.00000270, 0.00000915, -0.00001809, 0.00002107, -0.00001902, 0.00000885, -0.00000739, -0.00000113, 0.00000915, -0.00000152, 0.00000521, -0.00001554, 0.00000308, 0.00000012, 0.00001576, -0.00000303, -0.00001503, 0.00000625, 0.00000015, -0.00000495, 0.00000377, -0.00000236, 0.00000018, 0.00000645, -0.00001360, -0.00000396, 0.00001203, 0.00000654, -0.00000667, -0.00000748, -0.00000851, 0.00000852, 0.00000154, -0.00000225, -0.00000184, -0.00000126, -0.00000410, 0.00000515, -0.00000573, -0.00000021, -0.00000138, -0.00000169, -0.00000530, 0.00000173, -0.00000438, -0.00000160, -0.00000590, 0.00000723, -0.00000030, -0.00000204, -0.00000563, -0.00000051, -0.00000490, -0.00000202, -0.00000980, 0.00000381, -0.00000664, -0.00000518, -0.00000527, -0.00000378, -0.00001247, -0.00000483, -0.00002896, -0.00003569, -0.00002139, 0.00002374, 0.00003906, 0.00003432, 0.00000937, 0.00002071, 0.00000554, },
		{ 0.00001102, 0.00000769, -0.00000015, 0.00002405, -0.00000389, 0.00000490, -0.00001283, -0.00000435, 0.00003921, 0.00012209, 0.00000442, -0.00006676, -0.00001382, -0.00001026, 0.00000026, 0.00000647, -0.00001186, 0.00000532, 0.00000136, -0.00000907, -0.00000372, 0.00000427, 0.00000317, 0.00000545, -0.00000215, 0.00000606, 0.00000330, 0.00000397, -0.00000205, 0.00002049, 0.00000927, -0.00000371, 0.00000196, -0.00000910, -0.00002353, 0.00000826, -0.00001262, -0.00000592, 0.00000605, -0.00000758, -0.00000235, 0.00001397, -0.00001373, 0.00000129, -0.00000597, -0.00000063, -0.00000123, -0.00000355, 0.00000237, 0.00001144, -0.00001240, 0.00000535, -0.00000763, 0.00000161, -0.00000495, 0.00000024, 0.00000084, 0.00000176, -0.00001019, 0.00001712, -0.00002068, 0.00001288, -0.00001268, 0.00000546, 0.00000593, -0.00000026, 0.00000074, -0.00001415, 0.00000503, 0.00000299, 0.00001533, -0.00001137, -0.00000968, 0.00000868, -0.00000479, -0.00000132, 0.00000286, -0.00000385, 0.00000404, 0.00000170, -0.00001569, 0.00000361, 0.00001191, 0.00000250, -0.00000785, -0.00000805, -0.00000462, 0.00000991, -0.00000203, -0.00000108, -0.00000228, -0.00000241, -0.00000123, 0.00000388, -0.00000668, 0.00000155, -0.00000250, -0.00000168, -0.00000520, 0.00000165, -0.00000439, -0.00000255, -0.00000319, 0.00000772, -0.00000280, -0.00000135, -0.00000641, -0.00000007, -0.00000418, -0.00000405, -0.00000698, 0.00000381, -0.00000940, -0.00000238, -0.00000678, -0.00000569, -0.00000928, -0.00000838, -0.00003392, -0.00003180, -0.00000890, 0.00003268, 0.00003354, 0.00002741, 0.00000963, 0.00001896, 0.00000360, 0.00000876, },
		{ 0.00000706, -0.00000094, 0.00002260, -0.00000487, 0.00000453, -0.00001349, -0.00000829, 0.00003990, 0.00011939, 0.00000760, -0.00006339, -0.00001494, -0.00000757, 0.00000074, 0.00000539, -0.00001138, 0.00000557, 0.00000142, -0.00000936, -0.00000258, 0.00000609, 0.00000199, 0.00000671, -0.00000217, 0.00000589, 0.00000574, 0.00000160, 0.00000194, 0.00002391, 0.00000188, -0.00000028, -0.00000008, -0.00001854, -0.00001367, 0.00000724, -0.00001873, 0.00000267, 0.00000091, -0.00000842, 0.00000489, 0.00000629, -0.00001266, 0.00000337, -0.00000875, 0.00000266, -0.00000434, -0.00000159, 0.00000647, 0.00000474, -0.00001011, 0.00000544, -0.00000935, 0.00000354, -0.00000712, 0.00000315, 0.00000055, -0.00000254, -0.00000251, 0.00001029, -0.00001850, 0.00001385, -0.00001584, 0.00001166, 0.00000182, 0.00000087, -0.00000378, -0.00001124, 0.00000661, 0.00000634, 0.00001115, -0.00001700, -0.00000203, 0.00000702, -0.00000753, 0.00000271, -0.00000017, -0.00000275, 0.00000626, -0.00000463, -0.00001383, 0.00001047, 0.00000901, -0.00000054, -0.00000886, -0.00000762, -0.00000016, 0.00000900, -0.00000476, 0.00000078, -0.00000401, -0.00000199, 0.00000114, 0.00000151, -0.00000617, 0.00000264, -0.00000389, -0.00000114, -0.00000548, 0.00000193, -0.00000493, -0.00000277, 0.00000004, 0.00000646, -0.00000457, -0.00000043, -0.00000757, 0.00000125, -0.00000478, -0.00000514, -0.00000335, 0.00000157, -0.00001056, 0.00000054, -0.00001009, -0.00000536, -0.00000698, -0.00001412, -0.00003605, -0.00002491, 0.00000287, 0.00003751, 0.00002622, 0.00002278, 0.00001028, 0.00001654, 0.00000237, 0.00000835, 0.00001063, },
		{ -0.00000171, 0.00002078, -0.00000536, 0.00000428, -0.00001455, -0.00001149, 0.00004028, 0.00011648, 0.00001092, -0.00006038, -0.00001539, -0.00000470, 0.00000002, 0.00000486, -0.00001049, 0.00000552, 0.00000126, -0.00000926, -0.00000078, 0.00000686, 0.00000110, 0.00000835, -0.00000297, 0.00000707, 0.00000686, -0.00000020, 0.00000821, 0.00002318, -0.00000387, 0.00000429, -0.00000741, -0.00002226, -0.00000334, 0.00000077, -0.00001909, 0.00000905, -0.00000663, -0.00000396, 0.00000768, -0.00000125, -0.00000862, 0.00000250, -0.00000941, 0.00000482, -0.00000783, 0.00000322, 0.00000670, -0.00000072, -0.00000643, 0.00000347, -0.00000916, 0.00000409, -0.00000841, 0.00000665, -0.00000264, -0.00000285, 0.00000258, 0.00000295, -0.00001367, 0.00001174, -0.00001606, 0.00001633, -0.00000263, 0.00000174, -0.00000777, -0.00000726, 0.00000816, 0.00000893, 0.00000400, -0.00001812, 0.00000511, 0.00000238, -0.00000708, 0.00000522, -0.00000377, 0.00000065, 0.00000547, -0.00001043, -0.00000843, 0.00001489, 0.00000453, -0.00000236, -0.00000994, -0.00000570, 0.00000375, 0.00000618, -0.00000582, 0.00000224, -0.00000631, 0.00000012, 0.00000203, -0.00000092, -0.00000446, 0.00000258, -0.00000492, -0.00000048, -0.00000595, 0.00000247, -0.00000607, -0.00000168, 0.00000283, 0.00000380, -0.00000508, 0.00000005, -0.00000849, 0.00000290, -0.00000670, -0.00000434, -0.00000034, -0.00000228, -0.00000935, 0.00000204, -0.00001395, -0.00000271, -0.00000721, -0.00002043, -0.00003455, -0.00001654, 0.00001291, 0.00003861, 0.00001846, 0.00002038, 0.00001057, 0.00001381, 0.00000196, 0.00000786, 0.00000999, 0.00000666, },
		{ 0.00001884, -0.00000526, 0.00000373, -0.00001562, -0.00001402, 0.00003996, 0.00011374, 0.00001418, -0.00005760, -0.00001499, -0.00000232, -0.00000150, 0.00000508, -0.00000953, 0.00000522, 0.00000096, -0.00000844, 0.00000112, 0.00000665, 0.00000100, 0.00000960, -0.00000383, 0.00000925, 0.00000622, 0.00000016, 0.00001476, 0.00001874, -0.00000634, 0.00000679, -0.00001731, -0.00001924, 0.00000328, -0.00000759, -0.00001413, 0.00001057, -0.00001282, 0.00000348, 0.00000537, -0.00000593, -0.00000403, -0.00000028, -0.00000780, 0.00000469, -0.00000966, 0.00000884, 0.00000339, -0.00000341, -0.00000322, 0.00000066, -0.00000753, 0.00000306, -0.00000784, 0.00000888, -0.00000679, 0.00000014, 0.00000389, -0.00000267, -0.00000806, 0.00000753, -0.00001326, 0.00001858, -0.00000676, 0.00000223, -0.00001081, -0.00000259, 0.00000975, 0.00000944, -0.00000419, -0.00001448, 0.00000925, -0.00000298, -0.00000392, 0.00000509, -0.00000610, 0.00000491, 0.00000142, -0.00001362, -0.00000097, 0.00001598, -0.00000014, -0.00000325, -0.00001091, -0.00000223, 0.00000601, 0.00000247, -0.00000511, 0.00000242, -0.00000808, 0.00000327, 0.00000101, -0.00000235, -0.00000237, 0.00000146, -0.00000521, -0.00000016, -0.00000620, 0.00000287, -0.00000744, 0.00000088, 0.00000417, 0.00000063, -0.00000437, -0.00000036, -0.00000852, 0.00000400, -0.00000919, -0.00000141, 0.00000061, -0.00000623, -0.00000611, 0.00000103, -0.00001661, 0.00000114, -0.00001068, -0.00002514, -0.00002980, -0.00000809, 0.00002081, 0.00003652, 0.00001144, 0.00001976, 0.00001000, 0.00001129, 0.00000223, 0.00000704, 0.00000931, 0.00000647, -0.00000271, },
		{ -0.00000476, 0.00000270, -0.00001629, -0.00001618, 0.00003881, 0.00011148, 0.00001714, -0.00005472, -0.00001388, -0.00000083, -0.00000328, 0.00000602, -0.00000869, 0.00000466, 0.00000070, -0.00000682, 0.00000252, 0.00000600, 0.00000171, 0.00001003, -0.00000390, 0.00001148, 0.00000439, 0.00000346, 0.00001937, 0.00001239, -0.00000571, 0.00000503, -0.00002589, -0.00001127, 0.00000403, -0.00001386, -0.00000663, 0.00000707, -0.00001484, 0.00001048, -0.00000055, -0.00000664, -0.00000106, -0.00000317, -0.00000499, 0.00000244, -0.00000863, 0.00001301, -0.00000183, -0.00000316, -0.00000171, -0.00000160, -0.00000556, 0.00000115, -0.00000547, 0.00000871, -0.00000979, 0.00000463, 0.00000168, -0.00000523, -0.00000357, 0.00000279, -0.00000818, 0.00001801, -0.00000996, 0.00000224, -0.00001260, 0.00000256, 0.00001097, 0.00000717, -0.00001099, -0.00000748, 0.00000916, -0.00000668, 0.00000022, 0.00000252, -0.00000593, 0.00000809, -0.00000482, -0.00001292, 0.00000642, 0.00001391, -0.00000392, -0.00000378, -0.00001127, 0.00000229, 0.00000593, -0.00000079, -0.00000330, 0.00000106, -0.00000832, 0.00000626, -0.00000154, -0.00000211, -0.00000086, -0.00000014, -0.00000481, -0.00000034, -0.00000592, 0.00000264, -0.00000831, 0.00000444, 0.00000347, -0.00000191, -0.00000309, -0.00000151, -0.00000742, 0.00000371, -0.00001095, 0.00000288, -0.00000121, -0.00000852, -0.00000228, -0.00000245, -0.00001661, 0.00000418, -0.00001673, -0.00002652, -0.00002319, -0.00000039, 0.00002641, 0.00003193, 0.00000608, 0.00002014, 0.00000846, 0.00000947, 0.00000280, 0.00000584, 0.00000880, 0.00000633, -0.00000401, 0.00001713, },
		{ 0.00000126, -0.00001626, -0.00001842, 0.00003697, 0.00010980, 0.00001966, -0.00005139, -0.00001246, -0.00000034, -0.00000482, 0.00000747, -0.00000809, 0.00000391, 0.00000079, -0.00000463, 0.00000303, 0.00000556, 0.00000274, 0.00000973, -0.00000271, 0.00001271, 0.00000267, 0.00000925, 0.00002058, 0.00000637, -0.00000383, -0.00000108, -0.00002984, -0.00000200, -0.00000054, -0.00001540, -0.00000021, 0.00000066, -0.00001205, 0.00001415, -0.00000726, -0.00000414, -0.00000068, -0.00000462, -0.00000261, -0.00000063, -0.00000492, 0.00001417, -0.00000683, -0.00000101, -0.00000212, -0.00000249, -0.00000431, -0.00000045, -0.00000231, 0.00000621, -0.00001025, 0.00000843, -0.00000258, -0.00000463, -0.00000143, -0.00000080, -0.00000220, 0.00001486, -0.00001182, 0.00000170, -0.00001290, 0.00000793, 0.00001103, 0.00000244, -0.00001437, 0.00000034, 0.00000535, -0.00000720, 0.00000323, -0.00000106, -0.00000324, 0.00000857, -0.00001125, -0.00000832, 0.00001178, 0.00000972, -0.00000628, -0.00000440, -0.00001039, 0.00000687, 0.00000356, -0.00000250, -0.00000153, -0.00000127, -0.00000662, 0.00000788, -0.00000454, -0.00000030, -0.00000065, -0.00000145, -0.00000418, -0.00000084, -0.00000514, 0.00000154, -0.00000787, 0.00000797, 0.00000087, -0.00000296, -0.00000218, -0.00000272, -0.00000559, 0.00000178, -0.00001073, 0.00000691, -0.00000534, -0.00000797, 0.00000021, -0.00000706, -0.00001358, 0.00000442, -0.00002341, -0.00002399, -0.00001639, 0.00000642, 0.00002955, 0.00002576, 0.00000292, 0.00002058, 0.00000624, 0.00000860, 0.00000323, 0.00000445, 0.00000856, 0.00000597, -0.00000549, 0.00001589, -0.00000422, },
		{ -0.00001550, -0.00002110, 0.00003468, 0.00010868, 0.00002180, -0.00004745, -0.00001122, -0.00000065, -0.00000583, 0.00000916, -0.00000779, 0.00000307, 0.00000145, -0.00000233, 0.00000274, 0.00000578, 0.00000345, 0.00000932, -0.00000046, 0.00001241, 0.00000254, 0.00001600, 0.00001824, 0.00000226, -0.00000319, -0.00000951, -0.00002785, 0.00000473, -0.00000764, -0.00001200, 0.00000241, -0.00000541, -0.00000596, 0.00001331, -0.00001210, -0.00000038, -0.00000243, -0.00000408, -0.00000188, -0.00000277, 0.00000004, 0.00001208, -0.00000994, 0.00000150, -0.00000372, -0.00000202, -0.00000428, -0.00000080, 0.00000023, 0.00000251, -0.00000798, 0.00000988, -0.00000689, -0.00000190, -0.00000184, -0.00000204, 0.00000306, 0.00000996, -0.00001227, 0.00000060, -0.00001149, 0.00001304, 0.00000909, -0.00000341, -0.00001349, 0.00000632, -0.00000026, -0.00000454, 0.00000362, -0.00000370, 0.00000073, 0.00000580, -0.00001574, -0.00000099, 0.00001384, 0.00000488, -0.00000739, -0.00000518, -0.00000794, 0.00001027, -0.00000028, -0.00000222, -0.00000093, -0.00000345, -0.00000344, 0.00000739, -0.00000671, 0.00000227, -0.00000179, -0.00000190, -0.00000390, -0.00000114, -0.00000428, -0.00000019, -0.00000567, 0.00001020, -0.00000273, -0.00000237, -0.00000233, -0.00000314, -0.00000398, -0.00000119, -0.00000793, 0.00000888, -0.00001012, -0.00000466, -0.00000020, -0.00001067, -0.00000870, 0.00000083, -0.00002820, -0.00001843, -0.00001058, 0.00001257, 0.00003000, 0.00001916, 0.00000196, 0.00002025, 0.00000395, 0.00000857, 0.00000321, 0.00000318, 0.00000844, 0.00000526, -0.00000684, 0.00001512, -0.00000394, -0.00000030, },
		{ -0.00002445, 0.00003223, 0.00010800, 0.00002370, -0.00004292, -0.00001053, -0.00000140, -0.00000618, 0.00001085, -0.00000779, 0.00000239, 0.00000274, -0.00000043, 0.00000206, 0.00000667, 0.00000344, 0.00000956, 0.00000207, 0.00001083, 0.00000496, 0.00002180, 0.00001340, 0.00000029, -0.00000562, -0.00001707, -0.00002110, 0.00000656, -0.00001376, -0.00000565, 0.00000048, -0.00000836, 0.00000056, 0.00000878, -0.00001359, 0.00000256, -0.00000491, -0.00000210, -0.00000305, -0.00000271, 0.00000433, 0.00000777, -0.00001053, 0.00000305, -0.00000540, -0.00000088, -0.00000518, 0.00000036, 0.00000102, -0.00000074, -0.00000401, 0.00000846, -0.00000954, 0.00000130, -0.00000400, -0.00000061, 0.00000625, 0.00000449, -0.00001157, -0.00000084, -0.00000828, 0.00001709, 0.00000477, -0.00000840, -0.00000906, 0.00000867, -0.00000517, -0.00000015, 0.00000120, -0.00000387, 0.00000400, 0.00000058, -0.00001677, 0.00000711, 0.00001240, 0.00000073, -0.00000782, -0.00000577, -0.00000406, 0.00001153, -0.00000427, -0.00000033, -0.00000207, -0.00000423, 0.00000004, 0.00000495, -0.00000711, 0.00000445, -0.00000372, -0.00000142, -0.00000429, -0.00000077, -0.00000402, -0.00000179, -0.00000195, 0.00001019, -0.00000596, -0.00000075, -0.00000365, -0.00000221, -0.00000370, -0.00000380, -0.00000314, 0.00000760, -0.00001342, -0.00000005, -0.00000381, -0.00001131, -0.00000426, -0.00000595, -0.00002911, -0.00001173, -0.00000599, 0.00001811, 0.00002761, 0.00001342, 0.00000261, 0.00001875, 0.00000226, 0.00000894, 0.00000271, 0.00000229, 0.00000814, 0.00000423, -0.00000780, 0.00001464, -0.00000409, -0.00000162, -0.00001420, },
		{ 0.00002982, 0.00010765, 0.00002558, -0.00003804, -0.00001059, -0.00000217, -0.00000596, 0.00001231, -0.00000800, 0.00000210, 0.00000441, 0.00000080, 0.00000153, 0.00000786, 0.00000276, 0.00001092, 0.00000393, 0.00000901, 0.00000988, 0.00002510, 0.00000769, -0.00000065, -0.00001135, -0.00002084, -0.00001258, 0.00000348, -0.00001621, 0.00000057, -0.00000459, -0.00000708, 0.00000475, 0.00000279, -0.00001192, 0.00000342, -0.00000664, 0.00000005, -0.00000530, -0.00000021, 0.00000634, 0.00000299, -0.00000910, 0.00000311, -0.00000629, 0.00000005, -0.00000618, 0.00000247, -0.00000023, -0.00000216, -0.00000002, 0.00000485, -0.00000971, 0.00000342, -0.00000647, 0.00000284, 0.00000667, -0.00000038, -0.00001031, -0.00000217, -0.00000343, 0.00001909, -0.00000148, -0.00001075, -0.00000303, 0.00000718, -0.00000740, 0.00000379, -0.00000280, -0.00000123, 0.00000476, -0.00000526, -0.00001401, 0.00001398, 0.00000816, -0.00000188, -0.00000824, -0.00000552, 0.00000047, 0.00001028, -0.00000706, 0.00000206, -0.00000469, -0.00000285, 0.00000234, 0.00000159, -0.00000558, 0.00000528, -0.00000553, -0.00000050, -0.00000521, 0.00000039, -0.00000487, -0.00000220, 0.00000226, 0.00000778, -0.00000763, 0.00000078, -0.00000549, -0.00000012, -0.00000535, -0.00000446, 0.00000194, 0.00000328, -0.00001353, 0.00000353, -0.00000933, -0.00000823, -0.00000277, -0.00001369, -0.00002560, -0.00000594, -0.00000197, 0.00002263, 0.00002273, 0.00000962, 0.00000383, 0.00001624, 0.00000160, 0.00000915, 0.00000200, 0.00000179, 0.00000739, 0.00000316, -0.00000824, 0.00001416, -0.00000463, -0.00000243, -0.00001270, -0.00002851, },
		},
		{ { -0.00009642, -0.00002191, 0.00005250, 0.00001973, 0.00001183, 0.00001089, -0.00000213, 0.00001428, 0.00000712, 0.00000103, 0.00001416, 0.00000927, 0.00001011, 0.00001222, 0.00001523, 0.00002805, 0.00005609, 0.00000033, -0.00005076, -0.00002072, -0.00001378, -0.00001085, -0.00001317, -0.00000599, -0.00000539, -0.00000463, -0.00000011, -0.00000266, -0.00000399, -0.00000320, -0.00000315, -0.00000181, -0.00000289, -0.00000473, -0.00000400, 0.00000027, 0.00000039, -0.00000085, -0.00001008, 0.00000046, -0.00000039, -0.00000127, 0.00000126, -0.00000239, -0.00000276, -0.00000498, -0.00000196, 0.00000139, -0.00000548, 0.00000039, -0.00000203, -0.00000015, -0.00000042, 0.00000022, 0.00001121, -0.00000340, -0.00000197, 0.00000150, 0.00000458, 0.00000002, -0.00000030, 0.00000034, -0.00000037, -0.00000006, -0.00000064, -0.00000006, -0.00000037, 0.00000034, -0.00000030, 0.00000002, 0.00000458, 0.00000150, -0.00000197, -0.00000340, 0.00001121, 0.00000022, -0.00000042, -0.00000015, -0.00000203, 0.00000039, -0.00000548, 0.00000139, -0.00000196, -0.00000498, -0.00000276, -0.00000239, 0.00000126, -0.00000127, -0.00000039, 0.00000046, -0.00001008, -0.00000085, 0.00000039, 0.00000027, -0.00000400, -0.00000473, -0.00000289, -0.00000181, -0.00000315, -0.00000320, -0.00000399, -0.00000266, -0.00000011, -0.00000463, -0.00000539, -0.00000599, -0.00001317, -0.00001085, -0.00001378, -0.00002072, -0.00005076, 0.00000033, 0.00005609, 0.00002805, 0.00001523, 0.00001222, 0.00001011, 0.00000927, 0.00001416, 0.00000103, 0.00000712, 0.00001428, -0.00000213, 0.00001089, 0.00001183, 0.00001973, 0.00005250, -0.00002191, },
		{ -0.00002416, 0.00004788, 0.00002205, 0.00000949, 0.00001121, -0.00000355, 0.00001366, 0.00000747, 0.00000056, 0.00001275, 0.00000858, 0.00000963, 0.00001031, 0.00001365, 0.00002323, 0.00004870, 0.00002317, -0.00004526, -0.00002945, -0.00001270, -0.00001212, -0.00001243, -0.00000728, -0.00000509, -0.00000548, 0.00000019, -0.00000241, -0.00000375, -0.00000356, -0.00000354, -0.00000113, -0.00000335, -0.00000380, -0.00000457, -0.00000083, 0.00000058, 0.00000048, -0.00000885, -0.00000260, 0.00000126, -0.00000238, 0.00000192, -0.00000238, -0.00000203, -0.00000489, -0.00000334, 0.00000188, -0.00000409, -0.00000201, 0.00000044, -0.00000232, 0.00000139, -0.00000281, 0.00001087, 0.00000052, -0.00000405, 0.00000096, 0.00000431, 0.00000158, -0.00000103, 0.00000071, -0.00000049, 0.00000008, -0.00000061, -0.00000023, -0.00000026, -0.00000002, 0.00000027, -0.00000088, 0.00000407, 0.00000211, -0.00000017, -0.00000523, 0.00000850, 0.00000498, -0.00000288, 0.00000239, -0.00000413, 0.00000163, -0.00000533, -0.00000021, -0.00000049, -0.00000464, -0.00000354, -0.00000241, 0.00000041, -0.00000009, -0.00000183, 0.00000238, -0.00000946, -0.00000322, 0.00000056, 0.00000047, -0.00000242, -0.00000607, -0.00000236, -0.00000237, -0.00000303, -0.00000273, -0.00000407, -0.00000297, -0.00000041, -0.00000408, -0.00000543, -0.00000504, -0.00001327, -0.00001033, -0.00001414, -0.00001550, -0.00004694, -0.00002576, 0.00005658, 0.00003682, 0.00001586, 0.00001493, 0.00001053, 0.00001001, 0.00001543, 0.00000191, 0.00000674, 0.00001464, -0.00000042, 0.00001030, 0.00001420, 0.00001823, 0.00005576, -0.00001860, -0.00009675, },
		{ 0.00004251, 0.00002426, 0.00000797, 0.00001083, -0.00000447, 0.00001274, 0.00000783, 0.00000031, 0.00001151, 0.00000776, 0.00000909, 0.00000916, 0.00001124, 0.00002118, 0.00003878, 0.00003838, -0.00003087, -0.00003885, -0.00001277, -0.00001306, -0.00001193, -0.00000820, -0.00000506, -0.00000611, 0.00000006, -0.00000190, -0.00000362, -0.00000360, -0.00000422, -0.00000051, -0.00000347, -0.00000351, -0.00000416, -0.00000252, 0.00000086, 0.00000098, -0.00000646, -0.00000560, 0.00000207, -0.00000268, 0.00000186, -0.00000193, -0.00000182, -0.00000415, -0.00000464, 0.00000153, -0.00000206, -0.00000417, 0.00000177, -0.00000290, 0.00000157, -0.00000344, 0.00000774, 0.00000543, -0.00000546, 0.00000019, 0.00000350, 0.00000330, -0.00000143, 0.00000079, -0.00000041, 0.00000009, -0.00000050, -0.00000036, -0.00000027, -0.00000018, 0.00000046, -0.00000103, 0.00000292, 0.00000281, 0.00000096, -0.00000486, 0.00000377, 0.00000970, -0.00000427, 0.00000367, -0.00000437, 0.00000070, -0.00000333, -0.00000271, 0.00000086, -0.00000410, -0.00000394, -0.00000281, -0.00000021, 0.00000054, -0.00000231, 0.00000262, -0.00000702, -0.00000618, 0.00000084, -0.00000003, -0.00000033, -0.00000725, -0.00000215, -0.00000262, -0.00000315, -0.00000237, -0.00000380, -0.00000350, -0.00000043, -0.00000411, -0.00000495, -0.00000478, -0.00001235, -0.00001099, -0.00001301, -0.00001432, -0.00003664, -0.00004878, 0.00004643, 0.00004928, 0.00001626, 0.00001810, 0.00001105, 0.00001089, 0.00001641, 0.00000321, 0.00000643, 0.00001484, 0.00000132, 0.00000993, 0.00001595, 0.00001805, 0.00005764, -0.00001456, -0.00009751, -0.00002562, },
		{ 0.00002549, 0.00000774, 0.00000967, -0.00000496, 0.00001160, 0.00000817, 0.00000010, 0.00001062, 0.00000678, 0.00000843, 0.00000860, 0.00000846, 0.00002042, 0.00002966, 0.00004469, -0.00001109, -0.00004494, -0.00001580, -0.00001282, -0.00001233, -0.00000826, -0.00000568, -0.00000616, -0.00000076, -0.00000103, -0.00000371, -0.00000329, -0.00000505, -0.00000012, -0.00000319, -0.00000375, -0.00000319, -0.00000414, 0.00000075, 0.00000116, -0.00000395, -0.00000738, 0.00000129, -0.00000180, 0.00000091, -0.00000086, -0.00000229, -0.00000285, -0.00000574, 0.00000067, -0.00000027, -0.00000507, 0.00000132, -0.00000177, 0.00000031, -0.00000225, 0.00000322, 0.00000959, -0.00000520, -0.00000117, 0.00000260, 0.00000453, -0.00000102, 0.00000032, -0.00000002, -0.00000008, -0.00000033, -0.00000043, -0.00000039, -0.00000017, 0.00000033, -0.00000070, 0.00000158, 0.00000331, 0.00000154, -0.00000308, -0.00000116, 0.00001239, -0.00000316, 0.00000267, -0.00000218, -0.00000223, -0.00000025, -0.00000519, 0.00000156, -0.00000332, -0.00000393, -0.00000352, -0.00000058, 0.00000048, -0.00000176, 0.00000147, -0.00000368, -0.00000871, 0.00000056, -0.00000064, 0.00000147, -0.00000759, -0.00000263, -0.00000242, -0.00000350, -0.00000215, -0.00000323, -0.00000412, -0.00000029, -0.00000452, -0.00000420, -0.00000496, -0.00001073, -0.00001235, -0.00001101, -0.00001557, -0.00002471, -0.00006284, 0.00002463, 0.00006269, 0.00001830, 0.00002083, 0.00001221, 0.00001170, 0.00001730, 0.00000462, 0.00000643, 0.00001487, 0.00000289, 0.00001005, 0.00001687, 0.00001905, 0.00005880, -0.00001056, -0.00009811, -0.00002685, 0.00003722, },
		{ 0.00000872, 0.00000810, -0.00000540, 0.00001050, 0.00000839, -0.00000010, 0.00001001, 0.00000590, 0.00000745, 0.00000856, 0.00000567, 0.00001989, 0.00002291, 0.00004361, 0.00000915, -0.00004442, -0.00002222, -0.00001158, -0.00001354, -0.00000757, -0.00000683, -0.00000574, -0.00000205, -0.00000009, -0.00000381, -0.00000287, -0.00000575, -0.00000014, -0.00000252, -0.00000426, -0.00000221, -0.00000511, -0.00000011, 0.00000143, -0.00000206, -0.00000736, -0.00000108, -0.00000011, -0.00000053, 0.00000046, -0.00000304, -0.00000150, -0.00000627, -0.00000053, 0.00000086, -0.00000458, -0.00000047, 0.00000013, -0.00000127, -0.00000061, -0.00000086, 0.00001144, -0.00000271, -0.00000314, 0.00000187, 0.00000485, 0.00000036, -0.00000061, 0.00000057, -0.00000034, -0.00000014, -0.00000045, -0.00000055, -0.00000015, 0.00000014, -0.00000030, 0.00000054, 0.00000329, 0.00000195, -0.00000096, -0.00000462, 0.00001185, 0.00000074, -0.00000029, 0.00000156, -0.00000573, 0.00000231, -0.00000639, 0.00000096, -0.00000206, -0.00000377, -0.00000411, -0.00000100, 0.00000002, -0.00000072, -0.00000012, -0.00000074, -0.00000971, -0.00000088, -0.00000075, 0.00000232, -0.00000670, -0.00000390, -0.00000178, -0.00000402, -0.00000202, -0.00000266, -0.00000448, -0.00000033, -0.00000485, -0.00000373, -0.00000502, -0.00000920, -0.00001334, -0.00000949, -0.00001676, -0.00001569, -0.00006494, -0.00000575, 0.00007193, 0.00002445, 0.00002210, 0.00001462, 0.00001205, 0.00001854, 0.00000566, 0.00000706, 0.00001468, 0.00000428, 0.00001060, 0.00001722, 0.00002055, 0.00006024, -0.00000747, -0.00009803, -0.00002831, 0.00003261, 0.00002532, },
		{ 0.00000686, -0.00000626, 0.00000974, 0.00000831, -0.00000017, 0.00000940, 0.00000545, 0.00000597, 0.00000893, 0.00000312, 0.00001905, 0.00001870, 0.00003804, 0.00002549, -0.00003628, -0.00003048, -0.00001066, -0.00001476, -0.00000681, -0.00000789, -0.00000539, -0.00000324, 0.00000039, -0.00000351, -0.00000268, -0.00000601, -0.00000067, -0.00000162, -0.00000472, -0.00000160, -0.00000517, -0.00000164, 0.00000176, -0.00000097, -0.00000575, -0.00000419, 0.00000142, -0.00000161, 0.00000134, -0.00000338, -0.00000082, -0.00000583, -0.00000208, 0.00000136, -0.00000324, -0.00000248, 0.00000147, -0.00000197, 0.00000026, -0.00000321, 0.00001037, 0.00000160, -0.00000514, 0.00000123, 0.00000426, 0.00000239, -0.00000155, 0.00000100, -0.00000044, -0.00000008, -0.00000040, -0.00000067, -0.00000025, 0.00000008, -0.00000012, 0.00000009, 0.00000266, 0.00000241, 0.00000068, -0.00000575, 0.00000824, 0.00000620, -0.00000355, 0.00000491, -0.00000775, 0.00000268, -0.00000534, -0.00000122, -0.00000031, -0.00000373, -0.00000419, -0.00000178, -0.00000044, 0.00000016, -0.00000116, 0.00000074, -0.00000861, -0.00000355, -0.00000028, 0.00000207, -0.00000474, -0.00000560, -0.00000102, -0.00000455, -0.00000194, -0.00000232, -0.00000429, -0.00000083, -0.00000475, -0.00000391, -0.00000452, -0.00000841, -0.00001313, -0.00000949, -0.00001618, -0.00001180, -0.00005638, -0.00003786, 0.00007097, 0.00003643, 0.00002155, 0.00001855, 0.00001175, 0.00002046, 0.00000602, 0.00000843, 0.00001435, 0.00000554, 0.00001133, 0.00001754, 0.00002171, 0.00006285, -0.00000575, -0.00009725, -0.00003005, 0.00002879, 0.00002398, 0.00001025, },
		{ -0.00000789, 0.00000946, 0.00000787, -0.00000003, 0.00000858, 0.00000564, 0.00000405, 0.00000943, 0.00000106, 0.00001772, 0.00001647, 0.00003081, 0.00003560, -0.00002221, -0.00003750, -0.00001184, -0.00001501, -0.00000674, -0.00000821, -0.00000566, -0.00000377, -0.00000003, -0.00000259, -0.00000291, -0.00000573, -0.00000160, -0.00000072, -0.00000491, -0.00000151, -0.00000450, -0.00000337, 0.00000184, -0.00000043, -0.00000337, -0.00000682, 0.00000173, -0.00000159, 0.00000118, -0.00000277, -0.00000123, -0.00000435, -0.00000380, 0.00000138, -0.00000177, -0.00000369, 0.00000140, -0.00000131, -0.00000002, -0.00000361, 0.00000699, 0.00000639, -0.00000619, 0.00000030, 0.00000319, 0.00000434, -0.00000184, 0.00000086, -0.00000015, -0.00000026, -0.00000025, -0.00000072, -0.00000048, 0.00000016, -0.00000019, 0.00000013, 0.00000166, 0.00000287, 0.00000161, -0.00000480, 0.00000301, 0.00001103, -0.00000499, 0.00000595, -0.00000681, 0.00000018, -0.00000216, -0.00000439, 0.00000145, -0.00000371, -0.00000374, -0.00000285, -0.00000082, 0.00000049, -0.00000118, 0.00000051, -0.00000578, -0.00000670, 0.00000024, 0.00000125, -0.00000247, -0.00000695, -0.00000069, -0.00000478, -0.00000198, -0.00000223, -0.00000362, -0.00000164, -0.00000431, -0.00000464, -0.00000357, -0.00000836, -0.00001172, -0.00001091, -0.00001383, -0.00001225, -0.00004216, -0.00006336, 0.00005552, 0.00005341, 0.00002044, 0.00002328, 0.00001127, 0.00002286, 0.00000588, 0.00001018, 0.00001419, 0.00000662, 0.00001202, 0.00001826, 0.00002200, 0.00006690, -0.00000512, -0.00009634, -0.00003162, 0.00002535, 0.00002216, 0.00001142, 0.00000661, },
		{ 0.00000947, 0.00000721, 0.00000025, 0.00000753, 0.00000631, 0.00000211, 0.00000957, -0.00000020, 0.00001579, 0.00001556, 0.00002384, 0.00003947, -0.00000570, -0.00004009, -0.00001614, -0.00001405, -0.00000761, -0.00000763, -0.00000668, -0.00000352, -0.00000129, -0.00000126, -0.00000335, -0.00000516, -0.00000259, -0.00000012, -0.00000468, -0.00000185, -0.00000349, -0.00000475, 0.00000130, -0.00000005, -0.00000108, -0.00000797, 0.00000030, -0.00000031, -0.00000000, -0.00000128, -0.00000252, -0.00000235, -0.00000520, 0.00000089, -0.00000061, -0.00000372, -0.00000000, 0.00000021, -0.00000087, -0.00000284, 0.00000280, 0.00000995, -0.00000532, -0.00000127, 0.00000215, 0.00000547, -0.00000099, -0.00000002, 0.00000056, -0.00000063, -0.00000005, -0.00000073, -0.00000074, 0.00000024, -0.00000032, 0.00000034, 0.00000071, 0.00000307, 0.00000206, -0.00000275, -0.00000184, 0.00001317, -0.00000322, 0.00000394, -0.00000295, -0.00000427, 0.00000170, -0.00000712, 0.00000227, -0.00000325, -0.00000318, -0.00000373, -0.00000137, 0.00000039, -0.00000039, -0.00000081, -0.00000243, -0.00000905, -0.00000012, 0.00000072, -0.00000089, -0.00000717, -0.00000126, -0.00000443, -0.00000233, -0.00000218, -0.00000286, -0.00000227, -0.00000403, -0.00000534, -0.00000277, -0.00000845, -0.00000993, -0.00001261, -0.00001119, -0.00001417, -0.00002862, -0.00007573, 0.00002567, 0.00007100, 0.00002177, 0.00002722, 0.00001169, 0.00002501, 0.00000598, 0.00001159, 0.00001473, 0.00000738, 0.00001261, 0.00001948, 0.00002145, 0.00007197, -0.00000465, -0.00009629, -0.00003233, 0.00002172, 0.00002062, 0.00001149, 0.00000768, -0.00001019, },
		{ 0.00000665, 0.00000045, 0.00000645, 0.00000703, 0.00000073, 0.00000887, -0.00000051, 0.00001333, 0.00001542, 0.00001803, 0.00003857, 0.00000951, -0.00003651, -0.00002291, -0.00001273, -0.00000881, -0.00000665, -0.00000798, -0.00000297, -0.00000281, -0.00000017, -0.00000348, -0.00000479, -0.00000316, -0.00000014, -0.00000398, -0.00000244, -0.00000255, -0.00000539, 0.00000006, 0.00000029, 0.00000063, -0.00000738, -0.00000252, 0.00000154, -0.00000152, 0.00000037, -0.00000385, -0.00000078, -0.00000557, -0.00000029, 0.00000024, -0.00000291, -0.00000185, 0.00000147, -0.00000128, -0.00000205, -0.00000063, 0.00001100, -0.00000227, -0.00000335, 0.00000140, 0.00000542, 0.00000100, -0.00000137, 0.00000139, -0.00000099, 0.00000010, -0.00000069, -0.00000094, 0.00000016, -0.00000032, 0.00000043, 0.00000015, 0.00000279, 0.00000239, -0.00000069, -0.00000475, 0.00001169, 0.00000159, -0.00000013, 0.00000213, -0.00000845, 0.00000407, -0.00000779, 0.00000124, -0.00000185, -0.00000307, -0.00000382, -0.00000247, 0.00000019, 0.00000056, -0.00000212, 0.00000002, -0.00000944, -0.00000201, 0.00000107, -0.00000063, -0.00000590, -0.00000278, -0.00000351, -0.00000304, -0.00000193, -0.00000247, -0.00000217, -0.00000443, -0.00000541, -0.00000277, -0.00000796, -0.00000879, -0.00001322, -0.00001005, -0.00001458, -0.00002049, -0.00007309, -0.00001303, 0.00008185, 0.00002914, 0.00002867, 0.00001418, 0.00002605, 0.00000716, 0.00001191, 0.00001637, 0.00000772, 0.00001318, 0.00002100, 0.00002053, 0.00007715, -0.00000317, -0.00009801, -0.00003176, 0.00001759, 0.00001978, 0.00001028, 0.00000975, -0.00001266, 0.00000937, },
		{ 0.00000032, 0.00000563, 0.00000736, 0.00000034, 0.00000721, -0.00000005, 0.00001068, 0.00001547, 0.00001372, 0.00003479, 0.00002093, -0.00002727, -0.00002986, -0.00001275, -0.00000930, -0.00000612, -0.00000877, -0.00000286, -0.00000375, -0.00000004, -0.00000278, -0.00000505, -0.00000299, -0.00000085, -0.00000294, -0.00000304, -0.00000194, -0.00000528, -0.00000156, 0.00000040, 0.00000176, -0.00000551, -0.00000568, 0.00000284, -0.00000235, 0.00000125, -0.00000424, -0.00000055, -0.00000444, -0.00000218, 0.00000092, -0.00000190, -0.00000312, 0.00000148, -0.00000057, -0.00000196, -0.00000242, 0.00000934, 0.00000227, -0.00000524, 0.00000075, 0.00000440, 0.00000352, -0.00000249, 0.00000182, -0.00000100, 0.00000001, -0.00000056, -0.00000104, -0.00000011, -0.00000014, 0.00000030, 0.00000004, 0.00000209, 0.00000276, 0.00000071, -0.00000519, 0.00000724, 0.00000762, -0.00000393, 0.00000581, -0.00000990, 0.00000322, -0.00000556, -0.00000178, 0.00000042, -0.00000357, -0.00000295, -0.00000415, 0.00000009, 0.00000113, -0.00000259, 0.00000071, -0.00000753, -0.00000527, 0.00000217, -0.00000155, -0.00000358, -0.00000470, -0.00000244, -0.00000388, -0.00000150, -0.00000265, -0.00000118, -0.00000561, -0.00000471, -0.00000369, -0.00000666, -0.00000876, -0.00001207, -0.00001117, -0.00001221, -0.00001894, -0.00005905, -0.00005084, 0.00007825, 0.00004458, 0.00002733, 0.00001898, 0.00002579, 0.00000973, 0.00001096, 0.00001902, 0.00000785, 0.00001373, 0.00002252, 0.00001984, 0.00008158, 0.00000013, -0.00010178, -0.00003012, 0.00001316, 0.00001950, 0.00000818, 0.00001209, -0.00001456, 0.00000869, 0.00000647, },
		{ 0.00000517, 0.00000711, 0.00000094, 0.00000498, 0.00000054, 0.00000845, 0.00001509, 0.00001103, 0.00002960, 0.00002784, -0.00001484, -0.00003397, -0.00001558, -0.00000839, -0.00000662, -0.00000855, -0.00000368, -0.00000366, -0.00000117, -0.00000119, -0.00000591, -0.00000219, -0.00000199, -0.00000189, -0.00000340, -0.00000173, -0.00000466, -0.00000303, -0.00000008, 0.00000259, -0.00000316, -0.00000796, 0.00000257, -0.00000181, 0.00000080, -0.00000322, -0.00000188, -0.00000202, -0.00000438, 0.00000133, -0.00000113, -0.00000325, 0.00000007, 0.00000106, -0.00000242, -0.00000275, 0.00000587, 0.00000679, -0.00000587, -0.00000027, 0.00000300, 0.00000565, -0.00000257, 0.00000139, -0.00000042, -0.00000040, -0.00000033, -0.00000107, -0.00000047, 0.00000007, 0.00000011, 0.00000014, 0.00000127, 0.00000305, 0.00000141, -0.00000375, 0.00000170, 0.00001231, -0.00000501, 0.00000609, -0.00000738, -0.00000101, -0.00000113, -0.00000572, 0.00000269, -0.00000417, -0.00000156, -0.00000587, -0.00000017, 0.00000125, -0.00000204, -0.00000024, -0.00000409, -0.00000878, 0.00000312, -0.00000279, -0.00000127, -0.00000603, -0.00000195, -0.00000437, -0.00000116, -0.00000314, 0.00000030, -0.00000706, -0.00000377, -0.00000498, -0.00000509, -0.00000942, -0.00000971, -0.00001373, -0.00000808, -0.00002154, -0.00004101, -0.00007746, 0.00005580, 0.00006620, 0.00002534, 0.00002478, 0.00002517, 0.00001303, 0.00000942, 0.00002191, 0.00000837, 0.00001406, 0.00002394, 0.00001971, 0.00008492, 0.00000527, -0.00010701, -0.00002829, 0.00000914, 0.00001921, 0.00000596, 0.00001380, -0.00001531, 0.00000720, 0.00000676, -0.00000019, },
		{ 0.00000644, 0.00000203, 0.00000299, 0.00000041, 0.00000730, 0.00001379, 0.00000994, 0.00002389, 0.00003092, -0.00000232, -0.00003306, -0.00002138, -0.00000644, -0.00000787, -0.00000751, -0.00000524, -0.00000273, -0.00000313, 0.00000070, -0.00000684, -0.00000129, -0.00000298, -0.00000130, -0.00000328, -0.00000189, -0.00000390, -0.00000390, -0.00000130, 0.00000330, -0.00000105, -0.00000860, 0.00000049, -0.00000004, -0.00000073, -0.00000118, -0.00000412, 0.00000068, -0.00000597, 0.00000109, -0.00000054, -0.00000248, -0.00000203, 0.00000260, -0.00000263, -0.00000250, 0.00000210, 0.00000966, -0.00000442, -0.00000196, 0.00000182, 0.00000660, -0.00000120, 0.00000006, 0.00000061, -0.00000099, -0.00000007, -0.00000109, -0.00000077, 0.00000011, 0.00000008, 0.00000016, 0.00000069, 0.00000304, 0.00000172, -0.00000158, -0.00000283, 0.00001360, -0.00000220, 0.00000268, -0.00000170, -0.00000685, 0.00000333, -0.00000861, 0.00000356, -0.00000396, -0.00000055, -0.00000677, -0.00000110, 0.00000124, -0.00000098, -0.00000188, -0.00000063, -0.00001101, 0.00000277, -0.00000327, -0.00000010, -0.00000587, -0.00000261, -0.00000410, -0.00000132, -0.00000343, 0.00000153, -0.00000789, -0.00000347, -0.00000568, -0.00000421, -0.00000977, -0.00000750, -0.00001590, -0.00000461, -0.00002414, -0.00002666, -0.00008625, 0.00001641, 0.00008732, 0.00002717, 0.00002905, 0.00002604, 0.00001568, 0.00000862, 0.00002383, 0.00001009, 0.00001385, 0.00002538, 0.00002000, 0.00008747, 0.00001150, -0.00011250, -0.00002749, 0.00000631, 0.00001833, 0.00000432, 0.00001428, -0.00001476, 0.00000510, 0.00000729, -0.00000091, 0.00000494, },
		{ 0.00000289, 0.00000208, -0.00000105, 0.00000750, 0.00001151, 0.00001011, 0.00001831, 0.00003128, 0.00000792, -0.00002687, -0.00002853, -0.00000495, -0.00000885, -0.00000651, -0.00000673, -0.00000179, -0.00000497, 0.00000191, -0.00000704, -0.00000099, -0.00000321, -0.00000153, -0.00000259, -0.00000229, -0.00000325, -0.00000405, -0.00000306, 0.00000377, 0.00000059, -0.00000763, -0.00000272, 0.00000202, -0.00000235, 0.00000082, -0.00000599, 0.00000234, -0.00000598, -0.00000023, 0.00000016, -0.00000151, -0.00000367, 0.00000294, -0.00000179, -0.00000251, -0.00000066, 0.00000991, -0.00000087, -0.00000405, 0.00000103, 0.00000615, 0.00000138, -0.00000164, 0.00000163, -0.00000142, 0.00000004, -0.00000105, -0.00000093, -0.00000013, 0.00000032, -0.00000007, 0.00000052, 0.00000263, 0.00000198, 0.00000026, -0.00000499, 0.00001099, 0.00000372, -0.00000261, 0.00000451, -0.00001128, 0.00000518, -0.00000858, 0.00000196, -0.00000227, -0.00000063, -0.00000622, -0.00000304, 0.00000143, -0.00000009, -0.00000305, 0.00000142, -0.00001084, 0.00000050, -0.00000245, -0.00000062, -0.00000398, -0.00000439, -0.00000309, -0.00000206, -0.00000319, 0.00000192, -0.00000746, -0.00000442, -0.00000516, -0.00000466, -0.00000900, -0.00000661, -0.00001607, -0.00000384, -0.00002352, -0.00002047, -0.00007716, -0.00003112, 0.00009797, 0.00003763, 0.00002948, 0.00003001, 0.00001643, 0.00000973, 0.00002388, 0.00001347, 0.00001296, 0.00002701, 0.00002035, 0.00008992, 0.00001775, -0.00011698, -0.00002860, 0.00000500, 0.00001671, 0.00000349, 0.00001353, -0.00001335, 0.00000298, 0.00000763, -0.00000151, 0.00000462, 0.00000579, },
		{ 0.00000264, -0.00000379, 0.00000862, 0.00000883, 0.00001072, 0.00001357, 0.00002969, 0.00001505, -0.00001716, -0.00003422, -0.00000578, -0.00000844, -0.00000648, -0.00000730, -0.00000171, -0.00000575, 0.00000166, -0.00000598, -0.00000167, -0.00000245, -0.00000259, -0.00000151, -0.00000275, -0.00000285, -0.00000363, -0.00000486, 0.00000365, 0.00000192, -0.00000568, -0.00000583, 0.00000307, -0.00000293, 0.00000166, -0.00000636, 0.00000200, -0.00000401, -0.00000260, 0.00000108, -0.00000091, -0.00000406, 0.00000159, 0.00000015, -0.00000298, -0.00000201, 0.00000773, 0.00000376, -0.00000569, 0.00000030, 0.00000473, 0.00000428, -0.00000278, 0.00000195, -0.00000127, -0.00000020, -0.00000089, -0.00000097, -0.00000059, 0.00000072, -0.00000043, 0.00000060, 0.00000200, 0.00000226, 0.00000132, -0.00000465, 0.00000569, 0.00001032, -0.00000681, 0.00000812, -0.00001164, 0.00000281, -0.00000508, -0.00000196, 0.00000060, -0.00000173, -0.00000435, -0.00000565, 0.00000171, 0.00000027, -0.00000313, 0.00000146, -0.00000823, -0.00000319, -0.00000080, -0.00000234, -0.00000112, -0.00000649, -0.00000195, -0.00000303, -0.00000255, 0.00000146, -0.00000584, -0.00000644, -0.00000362, -0.00000622, -0.00000724, -0.00000721, -0.00001396, -0.00000602, -0.00001921, -0.00002202, -0.00005673, -0.00007383, 0.00008885, 0.00005873, 0.00002609, 0.00003711, 0.00001529, 0.00001285, 0.00002213, 0.00001811, 0.00001181, 0.00002874, 0.00002050, 0.00009284, 0.00002327, -0.00011985, -0.00003167, 0.00000475, 0.00001480, 0.00000310, 0.00001212, -0.00001190, 0.00000155, 0.00000734, -0.00000170, 0.00000393, 0.00000552, 0.00000299, },
		{ -0.00000700, 0.00000962, 0.00000677, 0.00001072, 0.00001046, 0.00002645, 0.00001958, -0.00000672, -0.00003583, -0.00000992, -0.00000633, -0.00000766, -0.00000667, -0.00000278, -0.00000517, -0.00000011, -0.00000390, -0.00000303, -0.00000106, -0.00000400, -0.00000050, -0.00000294, -0.00000275, -0.00000295, -0.00000616, 0.00000266, 0.00000313, -0.00000355, -0.00000770, 0.00000226, -0.00000192, 0.00000092, -0.00000493, -0.00000025, -0.00000069, -0.00000524, 0.00000182, -0.00000073, -0.00000323, -0.00000092, 0.00000237, -0.00000332, -0.00000240, 0.00000426, 0.00000787, -0.00000577, -0.00000088, 0.00000312, 0.00000640, -0.00000249, 0.00000114, -0.00000039, -0.00000081, -0.00000062, -0.00000097, -0.00000107, 0.00000102, -0.00000065, 0.00000062, 0.00000148, 0.00000240, 0.00000171, -0.00000274, -0.00000008, 0.00001472, -0.00000730, 0.00000720, -0.00000718, -0.00000322, 0.00000047, -0.00000654, 0.00000329, -0.00000283, -0.00000218, -0.00000789, 0.00000149, 0.00000035, -0.00000229, 0.00000004, -0.00000439, -0.00000682, 0.00000041, -0.00000402, 0.00000133, -0.00000766, -0.00000161, -0.00000352, -0.00000209, 0.00000076, -0.00000385, -0.00000852, -0.00000210, -0.00000781, -0.00000551, -0.00000831, -0.00001083, -0.00000939, -0.00001355, -0.00002694, -0.00003495, -0.00009960, 0.00005604, 0.00008683, 0.00002241, 0.00004509, 0.00001394, 0.00001667, 0.00001993, 0.00002269, 0.00001138, 0.00003015, 0.00002055, 0.00009631, 0.00002802, -0.00012145, -0.00003575, 0.00000437, 0.00001353, 0.00000246, 0.00001087, -0.00001128, 0.00000132, 0.00000629, -0.00000144, 0.00000280, 0.00000575, 0.00000223, 0.00000435, },
		{ 0.00000938, 0.00000622, 0.00000935, 0.00000937, 0.00002183, 0.00002241, 0.00000205, -0.00003230, -0.00001661, -0.00000357, -0.00000928, -0.00000545, -0.00000446, -0.00000383, -0.00000255, -0.00000178, -0.00000418, 0.00000012, -0.00000497, -0.00000015, -0.00000258, -0.00000297, -0.00000227, -0.00000671, 0.00000088, 0.00000419, -0.00000171, -0.00000785, -0.00000032, 0.00000022, -0.00000084, -0.00000247, -0.00000325, 0.00000252, -0.00000685, 0.00000162, -0.00000047, -0.00000194, -0.00000334, 0.00000354, -0.00000265, -0.00000270, 0.00000097, 0.00000995, -0.00000365, -0.00000276, 0.00000191, 0.00000698, -0.00000049, -0.00000065, 0.00000093, -0.00000154, -0.00000037, -0.00000095, -0.00000139, 0.00000097, -0.00000050, 0.00000032, 0.00000134, 0.00000221, 0.00000185, -0.00000051, -0.00000418, 0.00001507, -0.00000332, 0.00000219, 0.00000035, -0.00001016, 0.00000514, -0.00000928, 0.00000398, -0.00000260, -0.00000102, -0.00000859, 0.00000003, 0.00000066, -0.00000130, -0.00000160, -0.00000111, -0.00000860, -0.00000019, -0.00000442, 0.00000210, -0.00000691, -0.00000274, -0.00000300, -0.00000238, 0.00000060, -0.00000257, -0.00000943, -0.00000184, -0.00000812, -0.00000512, -0.00000848, -0.00000859, -0.00001138, -0.00000986, -0.00002987, -0.00002058, -0.00010236, 0.00000422, 0.00011224, 0.00002475, 0.00005024, 0.00001513, 0.00001904, 0.00001926, 0.00002553, 0.00001278, 0.00003065, 0.00002088, 0.00009989, 0.00003271, -0.00012293, -0.00003934, 0.00000247, 0.00001374, 0.00000100, 0.00001039, -0.00001194, 0.00000229, 0.00000477, -0.00000099, 0.00000144, 0.00000627, 0.00000099, 0.00000641, -0.00000953, },
		{ 0.00000738, 0.00000663, 0.00000992, 0.00001653, 0.00002397, 0.00000824, -0.00002476, -0.00002338, -0.00000217, -0.00000993, -0.00000480, -0.00000567, -0.00000286, -0.00000438, -0.00000086, -0.00000412, 0.00000027, -0.00000483, -0.00000085, -0.00000156, -0.00000346, -0.00000171, -0.00000658, -0.00000122, 0.00000475, -0.00000018, -0.00000658, -0.00000370, 0.00000224, -0.00000235, -0.00000039, -0.00000534, 0.00000396, -0.00000628, -0.00000008, 0.00000036, -0.00000111, -0.00000446, 0.00000273, -0.00000055, -0.00000339, -0.00000121, 0.00000939, 0.00000041, -0.00000490, 0.00000117, 0.00000602, 0.00000268, -0.00000260, 0.00000199, -0.00000190, -0.00000039, -0.00000086, -0.00000151, 0.00000055, 0.00000001, -0.00000027, 0.00000154, 0.00000174, 0.00000200, 0.00000109, -0.00000558, 0.00001139, 0.00000364, -0.00000423, 0.00000740, -0.00001433, 0.00000596, -0.00000820, 0.00000156, -0.00000031, -0.00000163, -0.00000720, -0.00000284, 0.00000144, -0.00000084, -0.00000231, 0.00000023, -0.00000753, -0.00000305, -0.00000314, 0.00000087, -0.00000423, -0.00000516, -0.00000155, -0.00000353, 0.00000136, -0.00000260, -0.00000855, -0.00000341, -0.00000661, -0.00000661, -0.00000698, -0.00000846, -0.00001034, -0.00001023, -0.00002770, -0.00001721, -0.00008479, -0.00005378, 0.00012201, 0.00003914, 0.00004978, 0.00002084, 0.00001844, 0.00002150, 0.00002562, 0.00001645, 0.00003011, 0.00002176, 0.00010303, 0.00003823, -0.00012550, -0.00004117, -0.00000172, 0.00001560, -0.00000127, 0.00001072, -0.00001372, 0.00000391, 0.00000341, -0.00000079, 0.00000019, 0.00000669, -0.00000013, 0.00000787, -0.00001041, 0.00000731, },
		{ 0.00000350, 0.00001090, 0.00001182, 0.00002390, 0.00001249, -0.00001588, -0.00002718, -0.00000396, -0.00000857, -0.00000557, -0.00000557, -0.00000316, -0.00000466, -0.00000186, -0.00000247, -0.00000083, -0.00000352, -0.00000244, -0.00000014, -0.00000399, -0.00000132, -0.00000610, -0.00000299, 0.00000439, 0.00000129, -0.00000468, -0.00000651, 0.00000282, -0.00000245, 0.00000014, -0.00000535, 0.00000281, -0.00000337, -0.00000305, 0.00000171, -0.00000110, -0.00000387, 0.00000004, 0.00000241, -0.00000415, -0.00000226, 0.00000675, 0.00000511, -0.00000628, 0.00000042, 0.00000420, 0.00000584, -0.00000358, 0.00000202, -0.00000146, -0.00000088, -0.00000063, -0.00000151, -0.00000008, 0.00000062, -0.00000087, 0.00000178, 0.00000126, 0.00000214, 0.00000185, -0.00000458, 0.00000547, 0.00001051, -0.00000851, 0.00001048, -0.00001309, 0.00000168, -0.00000325, -0.00000339, 0.00000336, -0.00000351, -0.00000432, -0.00000630, 0.00000223, -0.00000096, -0.00000174, -0.00000057, -0.00000401, -0.00000729, -0.00000107, -0.00000148, -0.00000073, -0.00000773, -0.00000006, -0.00000495, 0.00000264, -0.00000367, -0.00000630, -0.00000620, -0.00000396, -0.00000927, -0.00000439, -0.00001010, -0.00000661, -0.00001404, -0.00002124, -0.00002225, -0.00005698, -0.00010139, 0.00010545, 0.00006702, 0.00004445, 0.00003055, 0.00001530, 0.00002635, 0.00002353, 0.00002158, 0.00002924, 0.00002295, 0.00010555, 0.00004508, -0.00012977, -0.00004093, -0.00000787, 0.00001829, -0.00000358, 0.00001128, -0.00001589, 0.00000526, 0.00000290, -0.00000122, -0.00000068, 0.00000666, -0.00000067, 0.00000817, -0.00000943, 0.00000376, 0.00000948, },
		{ 0.00001089, 0.00000903, 0.00002162, 0.00001606, -0.00000841, -0.00002602, -0.00000928, -0.00000544, -0.00000757, -0.00000422, -0.00000471, -0.00000341, -0.00000447, 0.00000017, -0.00000252, -0.00000173, -0.00000420, 0.00000106, -0.00000421, -0.00000117, -0.00000557, -0.00000398, 0.00000295, 0.00000276, -0.00000282, -0.00000777, 0.00000130, -0.00000082, -0.00000110, -0.00000329, -0.00000040, 0.00000078, -0.00000608, 0.00000279, -0.00000148, -0.00000220, -0.00000331, 0.00000484, -0.00000407, -0.00000286, 0.00000336, 0.00000877, -0.00000582, -0.00000091, 0.00000242, 0.00000778, -0.00000272, 0.00000069, -0.00000019, -0.00000175, -0.00000031, -0.00000148, -0.00000064, 0.00000096, -0.00000110, 0.00000168, 0.00000111, 0.00000206, 0.00000205, -0.00000240, -0.00000019, 0.00001428, -0.00000809, 0.00000803, -0.00000641, -0.00000623, 0.00000333, -0.00000853, 0.00000648, -0.00000510, -0.00000155, -0.00000884, 0.00000204, -0.00000110, -0.00000050, -0.00000253, 0.00000028, -0.00001095, 0.00000009, -0.00000329, 0.00000185, -0.00000893, 0.00000029, -0.00000565, 0.00000355, -0.00000474, -0.00000400, -0.00000865, -0.00000179, -0.00001139, -0.00000236, -0.00001181, -0.00000236, -0.00001837, -0.00001437, -0.00002909, -0.00003163, -0.00012510, 0.00005999, 0.00010221, 0.00003974, 0.00004068, 0.00001240, 0.00003156, 0.00002140, 0.00002618, 0.00002948, 0.00002368, 0.00010786, 0.00005301, -0.00013534, -0.00003948, -0.00001458, 0.00002026, -0.00000476, 0.00001121, -0.00001754, 0.00000551, 0.00000370, -0.00000235, -0.00000112, 0.00000612, -0.00000054, 0.00000737, -0.00000720, -0.00000006, 0.00001110, 0.00000133, },
		{ 0.00000872, 0.00001712, 0.00001967, -0.00000373, -0.00002011, -0.00001641, -0.00000228, -0.00000950, -0.00000269, -0.00000658, -0.00000166, -0.00000740, 0.00000242, -0.00000358, -0.00000055, -0.00000516, 0.00000134, -0.00000374, -0.00000134, -0.00000512, -0.00000414, 0.00000079, 0.00000396, -0.00000124, -0.00000731, -0.00000185, 0.00000171, -0.00000317, -0.00000037, -0.00000399, 0.00000420, -0.00000751, 0.00000252, -0.00000138, -0.00000067, -0.00000571, 0.00000530, -0.00000233, -0.00000376, 0.00000053, 0.00001009, -0.00000307, -0.00000294, 0.00000125, 0.00000786, 0.00000001, -0.00000158, 0.00000142, -0.00000259, -0.00000010, -0.00000142, -0.00000096, 0.00000081, -0.00000079, 0.00000109, 0.00000141, 0.00000165, 0.00000214, -0.00000034, -0.00000356, 0.00001343, -0.00000281, 0.00000139, 0.00000284, -0.00001399, 0.00000787, -0.00001086, 0.00000685, -0.00000468, -0.00000059, -0.00000901, 0.00000001, -0.00000057, 0.00000043, -0.00000411, 0.00000331, -0.00001209, -0.00000115, -0.00000322, 0.00000218, -0.00000774, -0.00000122, -0.00000499, 0.00000329, -0.00000475, -0.00000305, -0.00000923, -0.00000166, -0.00001140, -0.00000248, -0.00001178, -0.00000005, -0.00001994, -0.00001129, -0.00003134, -0.00001809, -0.00012030, -0.00000585, 0.00013119, 0.00004419, 0.00004621, 0.00001347, 0.00003418, 0.00002191, 0.00002809, 0.00003225, 0.00002313, 0.00011062, 0.00006124, -0.00014111, -0.00003837, -0.00002002, 0.00001996, -0.00000393, 0.00000996, -0.00001814, 0.00000428, 0.00000570, -0.00000387, -0.00000138, 0.00000526, -0.00000002, 0.00000605, -0.00000483, -0.00000273, 0.00001097, 0.00000113, 0.00000902, },
		{ 0.00001148, 0.00002283, -0.00000127, -0.00001182, -0.00002216, -0.00000150, -0.00000967, -0.00000234, -0.00000746, -0.00000078, -0.00000914, 0.00000287, -0.00000291, -0.00000083, -0.00000467, 0.00000036, -0.00000252, -0.00000181, -0.00000475, -0.00000379, -0.00000136, 0.00000432, 0.00000026, -0.00000572, -0.00000535, 0.00000364, -0.00000456, 0.00000171, -0.00000602, 0.00000509, -0.00000617, 0.00000033, -0.00000024, -0.00000023, -0.00000606, 0.00000319, 0.00000100, -0.00000501, -0.00000117, 0.00000885, 0.00000136, -0.00000504, 0.00000058, 0.00000634, 0.00000374, -0.00000370, 0.00000244, -0.00000282, -0.00000031, -0.00000127, -0.00000105, 0.00000024, -0.00000009, 0.00000021, 0.00000195, 0.00000109, 0.00000228, 0.00000086, -0.00000403, 0.00000864, 0.00000503, -0.00000586, 0.00001029, -0.00001742, 0.00000725, -0.00000849, 0.00000347, -0.00000162, -0.00000203, -0.00000651, -0.00000374, 0.00000066, 0.00000048, -0.00000427, 0.00000387, -0.00000995, -0.00000499, -0.00000116, 0.00000022, -0.00000443, -0.00000415, -0.00000323, 0.00000187, -0.00000341, -0.00000396, -0.00000743, -0.00000396, -0.00000896, -0.00000508, -0.00000944, -0.00000075, -0.00001728, -0.00001377, -0.00002654, -0.00001832, -0.00009339, -0.00007431, 0.00013770, 0.00006487, 0.00004406, 0.00002069, 0.00003255, 0.00002651, 0.00002637, 0.00003788, 0.00002123, 0.00011414, 0.00006915, -0.00014606, -0.00003891, -0.00002295, 0.00001678, -0.00000114, 0.00000777, -0.00001785, 0.00000194, 0.00000824, -0.00000511, -0.00000189, 0.00000443, 0.00000045, 0.00000491, -0.00000339, -0.00000331, 0.00000866, 0.00000286, 0.00000558, 0.00001024, },
		{ 0.00002416, 0.00000078, -0.00000442, -0.00002353, -0.00000470, -0.00000730, -0.00000379, -0.00000667, -0.00000154, -0.00000887, 0.00000099, -0.00000039, -0.00000253, -0.00000294, -0.00000150, -0.00000093, -0.00000232, -0.00000445, -0.00000334, -0.00000280, 0.00000341, 0.00000192, -0.00000384, -0.00000781, 0.00000367, -0.00000417, 0.00000183, -0.00000542, 0.00000288, -0.00000227, -0.00000321, 0.00000158, -0.00000089, -0.00000447, -0.00000070, 0.00000474, -0.00000583, -0.00000210, 0.00000595, 0.00000594, -0.00000606, -0.00000022, 0.00000415, 0.00000704, -0.00000437, 0.00000210, -0.00000203, -0.00000107, -0.00000095, -0.00000104, -0.00000047, 0.00000059, -0.00000049, 0.00000227, 0.00000075, 0.00000235, 0.00000124, -0.00000236, 0.00000225, 0.00001181, -0.00000972, 0.00001232, -0.00001416, 0.00000084, -0.00000201, -0.00000247, 0.00000290, -0.00000488, -0.00000255, -0.00000788, 0.00000172, -0.00000006, -0.00000307, 0.00000224, -0.00000545, -0.00001002, 0.00000151, -0.00000261, -0.00000058, -0.00000700, -0.00000156, 0.00000020, -0.00000144, -0.00000600, -0.00000421, -0.00000749, -0.00000538, -0.00000875, -0.00000605, -0.00000338, -0.00001170, -0.00001992, -0.00001726, -0.00002688, -0.00005864, -0.00012506, 0.00011014, 0.00010198, 0.00003631, 0.00003255, 0.00002795, 0.00003413, 0.00002229, 0.00004492, 0.00001913, 0.00011785, 0.00007681, -0.00014997, -0.00004140, -0.00002347, 0.00001148, 0.00000239, 0.00000572, -0.00001747, -0.00000064, 0.00001034, -0.00000539, -0.00000299, 0.00000389, 0.00000049, 0.00000448, -0.00000338, -0.00000183, 0.00000492, 0.00000533, 0.00000192, 0.00001194, 0.00000657, },
		{ 0.00000402, -0.00000022, -0.00001940, -0.00001141, -0.00000330, -0.00000633, -0.00000475, -0.00000353, -0.00000705, -0.00000243, 0.00000293, -0.00000459, -0.00000100, -0.00000326, 0.00000024, -0.00000244, -0.00000432, -0.00000306, -0.00000319, 0.00000131, 0.00000363, -0.00000216, -0.00000849, 0.00000149, -0.00000202, 0.00000012, -0.00000262, -0.00000134, 0.00000254, -0.00000644, 0.00000289, -0.00000173, -0.00000216, -0.00000459, 0.00000706, -0.00000509, -0.00000309, 0.00000282, 0.00000898, -0.00000496, -0.00000173, 0.00000229, 0.00000866, -0.00000285, 0.00000023, -0.00000037, -0.00000217, -0.00000058, -0.00000100, -0.00000102, 0.00000083, -0.00000061, 0.00000200, 0.00000093, 0.00000212, 0.00000124, -0.00000000, -0.00000304, 0.00001453, -0.00000795, 0.00000801, -0.00000520, -0.00000879, 0.00000535, -0.00000790, 0.00000630, -0.00000693, 0.00000066, -0.00001034, 0.00000122, -0.00000023, -0.00000156, -0.00000008, -0.00000079, -0.00001382, 0.00000266, -0.00000429, 0.00000173, -0.00000794, -0.00000147, -0.00000042, -0.00000017, -0.00000766, -0.00000145, -0.00001009, -0.00000296, -0.00001114, -0.00000395, -0.00000541, -0.00000642, -0.00002547, -0.00000916, -0.00003478, -0.00003140, -0.00014376, 0.00004836, 0.00014569, 0.00003111, 0.00004368, 0.00002455, 0.00004137, 0.00001904, 0.00005059, 0.00001888, 0.00012049, 0.00008513, -0.00015358, -0.00004492, -0.00002298, 0.00000600, 0.00000472, 0.00000522, -0.00001802, -0.00000249, 0.00001116, -0.00000432, -0.00000472, 0.00000372, -0.00000001, 0.00000481, -0.00000457, 0.00000079, 0.00000132, 0.00000682, -0.00000029, 0.00001210, 0.00000409, 0.00002248, },
		{ 0.00000064, -0.00001130, -0.00001891, -0.00000010, -0.00000811, -0.00000316, -0.00000540, -0.00000513, -0.00000566, 0.00000529, -0.00000547, -0.00000015, -0.00000382, 0.00000022, -0.00000175, -0.00000451, -0.00000291, -0.00000276, -0.00000132, 0.00000483, -0.00000066, -0.00000755, -0.00000207, 0.00000069, -0.00000211, 0.00000073, -0.00000539, 0.00000586, -0.00000739, 0.00000236, -0.00000159, -0.00000068, -0.00000662, 0.00000655, -0.00000219, -0.00000464, 0.00000057, 0.00000945, -0.00000158, -0.00000391, 0.00000121, 0.00000817, 0.00000060, -0.00000244, 0.00000140, -0.00000305, -0.00000044, -0.00000093, -0.00000127, 0.00000048, -0.00000006, 0.00000112, 0.00000160, 0.00000158, 0.00000127, 0.00000181, -0.00000551, 0.00001223, -0.00000117, -0.00000022, 0.00000544, -0.00001690, 0.00000930, -0.00000942, 0.00000608, -0.00000617, 0.00000120, -0.00000959, -0.00000172, 0.00000072, -0.00000088, -0.00000136, 0.00000185, -0.00001437, 0.00000076, -0.00000344, 0.00000119, -0.00000598, -0.00000366, 0.00000072, -0.00000053, -0.00000766, -0.00000079, -0.00000997, -0.00000349, -0.00001044, -0.00000497, -0.00000467, -0.00000439, -0.00002652, -0.00000714, -0.00003494, -0.00002090, -0.00012847, -0.00003389, 0.00017774, 0.00003955, 0.00004762, 0.00002714, 0.00004441, 0.00001999, 0.00005223, 0.00002227, 0.00012092, 0.00009514, -0.00015809, -0.00004797, -0.00002339, 0.00000241, 0.00000421, 0.00000716, -0.00001999, -0.00000314, 0.00001043, -0.00000205, -0.00000676, 0.00000368, -0.00000086, 0.00000557, -0.00000623, 0.00000317, -0.00000057, 0.00000605, -0.00000011, 0.00000993, 0.00000454, 0.00001785, 0.00000864, },
		{ -0.00000268, -0.00002349, -0.00000037, -0.00000734, -0.00000338, -0.00000574, -0.00000463, -0.00000709, 0.00000525, -0.00000415, -0.00000116, -0.00000267, -0.00000119, -0.00000025, -0.00000501, -0.00000278, -0.00000203, -0.00000354, 0.00000480, 0.00000102, -0.00000582, -0.00000547, 0.00000225, -0.00000319, 0.00000272, -0.00000716, 0.00000589, -0.00000506, -0.00000044, -0.00000002, -0.00000083, -0.00000602, 0.00000312, 0.00000226, -0.00000636, -0.00000066, 0.00000755, 0.00000314, -0.00000590, 0.00000058, 0.00000615, 0.00000473, -0.00000452, 0.00000223, -0.00000304, -0.00000086, -0.00000072, -0.00000128, -0.00000030, 0.00000084, 0.00000005, 0.00000232, 0.00000106, 0.00000138, 0.00000262, -0.00000503, 0.00000631, 0.00000743, -0.00000792, 0.00001269, -0.00001899, 0.00000679, -0.00000542, 0.00000155, -0.00000224, -0.00000127, -0.00000569, -0.00000660, 0.00000253, -0.00000140, -0.00000077, 0.00000157, -0.00001128, -0.00000400, -0.00000035, -0.00000180, -0.00000180, -0.00000736, 0.00000308, -0.00000232, -0.00000589, -0.00000251, -0.00000692, -0.00000700, -0.00000672, -0.00000899, -0.00000101, -0.00000628, -0.00002217, -0.00001214, -0.00002620, -0.00002653, -0.00009046, -0.00011305, 0.00017837, 0.00006946, 0.00004136, 0.00003781, 0.00004174, 0.00002642, 0.00004914, 0.00002935, 0.00011920, 0.00010713, -0.00016414, -0.00004954, -0.00002594, 0.00000173, 0.00000052, 0.00001116, -0.00002286, -0.00000298, 0.00000868, 0.00000062, -0.00000842, 0.00000338, -0.00000173, 0.00000627, -0.00000753, 0.00000423, 0.00000005, 0.00000289, 0.00000204, 0.00000613, 0.00000683, 0.00001192, 0.00001309, 0.00000013, },
		{ -0.00002248, -0.00000524, -0.00000365, -0.00000565, -0.00000419, -0.00000603, -0.00000620, 0.00000265, -0.00000085, -0.00000361, -0.00000034, -0.00000335, 0.00000150, -0.00000552, -0.00000263, -0.00000148, -0.00000466, 0.00000322, 0.00000304, -0.00000411, -0.00000738, 0.00000153, -0.00000220, 0.00000241, -0.00000585, 0.00000249, -0.00000013, -0.00000450, 0.00000226, -0.00000216, -0.00000350, -0.00000174, 0.00000644, -0.00000704, -0.00000154, 0.00000450, 0.00000745, -0.00000643, -0.00000036, 0.00000375, 0.00000792, -0.00000466, 0.00000137, -0.00000185, -0.00000189, -0.00000034, -0.00000120, -0.00000113, 0.00000156, -0.00000063, 0.00000254, 0.00000098, 0.00000130, 0.00000271, -0.00000281, -0.00000041, 0.00001386, -0.00001090, 0.00001309, -0.00001334, -0.00000191, 0.00000262, -0.00000535, 0.00000301, -0.00000503, -0.00000058, -0.00001141, 0.00000381, -0.00000235, 0.00000108, -0.00000077, -0.00000614, -0.00000957, 0.00000297, -0.00000527, 0.00000247, -0.00001056, 0.00000504, -0.00000417, -0.00000359, -0.00000530, -0.00000260, -0.00001152, -0.00000214, -0.00001376, 0.00000341, -0.00000994, -0.00001504, -0.00002056, -0.00001349, -0.00003935, -0.00004876, -0.00016496, 0.00013577, 0.00011879, 0.00002912, 0.00005348, 0.00003590, 0.00003615, 0.00004363, 0.00003787, 0.00011711, 0.00012009, -0.00017122, -0.00004992, -0.00003035, 0.00000327, -0.00000496, 0.00001548, -0.00002516, -0.00000313, 0.00000701, 0.00000261, -0.00000897, 0.00000247, -0.00000228, 0.00000646, -0.00000795, 0.00000357, 0.00000286, -0.00000149, 0.00000458, 0.00000249, 0.00000886, 0.00000714, 0.00001513, 0.00000090, 0.00000297, },
		{ -0.00001324, 0.00000121, -0.00000858, -0.00000181, -0.00000843, -0.00000396, -0.00000119, 0.00000285, -0.00000600, 0.00000177, -0.00000503, 0.00000254, -0.00000548, -0.00000264, -0.00000123, -0.00000458, 0.00000043, 0.00000504, -0.00000267, -0.00000739, -0.00000140, 0.00000043, 0.00000035, -0.00000241, -0.00000264, 0.00000514, -0.00000766, 0.00000360, -0.00000329, -0.00000084, -0.00000574, 0.00000827, -0.00000548, -0.00000296, 0.00000174, 0.00000974, -0.00000461, -0.00000213, 0.00000198, 0.00000900, -0.00000234, -0.00000108, 0.00000014, -0.00000315, 0.00000001, -0.00000111, -0.00000172, 0.00000168, -0.00000053, 0.00000196, 0.00000158, 0.00000086, 0.00000264, -0.00000042, -0.00000520, 0.00001535, -0.00000747, 0.00000668, -0.00000214, -0.00001301, 0.00001048, -0.00001076, 0.00000633, -0.00000725, 0.00000289, -0.00001362, 0.00000282, -0.00000240, 0.00000271, -0.00000311, -0.00000165, -0.00001307, 0.00000398, -0.00000679, 0.00000442, -0.00001115, 0.00000481, -0.00000444, -0.00000253, -0.00000707, 0.00000048, -0.00001420, 0.00000029, -0.00001616, 0.00000545, -0.00001188, -0.00000959, -0.00002658, -0.00000440, -0.00004789, -0.00002109, -0.00017542, 0.00005369, 0.00017276, 0.00002246, 0.00006654, 0.00003281, 0.00004437, 0.00004007, 0.00004410, 0.00011742, 0.00013214, -0.00017784, -0.00005053, -0.00003498, 0.00000493, -0.00000972, 0.00001777, -0.00002519, -0.00000477, 0.00000649, 0.00000305, -0.00000802, 0.00000086, -0.00000239, 0.00000598, -0.00000741, 0.00000156, 0.00000662, -0.00000525, 0.00000558, 0.00000079, 0.00000876, 0.00000539, 0.00001332, 0.00000449, 0.00000413, -0.00001587, },
		{ 0.00000416, -0.00000985, -0.00000049, -0.00001008, -0.00000224, -0.00000417, 0.00000489, -0.00000658, 0.00000222, -0.00000505, 0.00000210, -0.00000448, -0.00000303, -0.00000109, -0.00000378, -0.00000257, 0.00000618, -0.00000120, -0.00000602, -0.00000528, 0.00000311, -0.00000176, 0.00000109, -0.00000686, 0.00000804, -0.00000784, 0.00000256, -0.00000288, 0.00000029, -0.00000705, 0.00000655, -0.00000142, -0.00000509, 0.00000006, 0.00000934, -0.00000060, -0.00000447, 0.00000108, 0.00000784, 0.00000180, -0.00000401, 0.00000193, -0.00000392, -0.00000003, -0.00000094, -0.00000196, 0.00000114, 0.00000029, 0.00000074, 0.00000260, 0.00000018, 0.00000269, 0.00000108, -0.00000678, 0.00001165, 0.00000074, -0.00000295, 0.00000939, -0.00002086, 0.00001334, -0.00001104, 0.00000506, -0.00000575, 0.00000271, -0.00001171, -0.00000125, -0.00000082, 0.00000286, -0.00000363, -0.00000004, -0.00001253, 0.00000124, -0.00000508, 0.00000286, -0.00000827, 0.00000180, -0.00000242, -0.00000372, -0.00000639, 0.00000053, -0.00001312, -0.00000134, -0.00001428, 0.00000316, -0.00000970, -0.00000911, -0.00002594, -0.00000416, -0.00004496, -0.00001563, -0.00014655, -0.00004741, 0.00020743, 0.00003536, 0.00006902, 0.00003835, 0.00004645, 0.00004247, 0.00004497, 0.00012215, 0.00014190, -0.00018259, -0.00005302, -0.00003789, 0.00000446, -0.00001149, 0.00001631, -0.00002211, -0.00000832, 0.00000750, 0.00000174, -0.00000579, -0.00000114, -0.00000222, 0.00000495, -0.00000627, -0.00000095, 0.00000985, -0.00000680, 0.00000391, 0.00000175, 0.00000605, 0.00000683, 0.00000814, 0.00001019, 0.00000208, -0.00000656, -0.00002069, },
		{ -0.00000773, -0.00000161, -0.00000963, -0.00000251, -0.00000469, 0.00000394, -0.00000450, 0.00000046, -0.00000314, 0.00000016, -0.00000262, -0.00000379, -0.00000084, -0.00000299, -0.00000470, 0.00000568, 0.00000074, -0.00000425, -0.00000842, 0.00000405, -0.00000219, 0.00000267, -0.00000807, 0.00000693, -0.00000430, -0.00000099, -0.00000072, -0.00000063, -0.00000533, 0.00000184, 0.00000390, -0.00000705, -0.00000079, 0.00000687, 0.00000429, -0.00000628, 0.00000051, 0.00000536, 0.00000617, -0.00000583, 0.00000236, -0.00000352, -0.00000077, -0.00000057, -0.00000198, 0.00000023, 0.00000137, -0.00000051, 0.00000344, -0.00000026, 0.00000275, 0.00000158, -0.00000560, 0.00000493, 0.00000975, -0.00001060, 0.00001574, -0.00002099, 0.00000856, -0.00000508, -0.00000100, -0.00000063, -0.00000094, -0.00000632, -0.00000735, 0.00000173, 0.00000154, -0.00000189, -0.00000174, -0.00000811, -0.00000449, -0.00000101, -0.00000125, -0.00000314, -0.00000277, 0.00000096, -0.00000660, -0.00000355, -0.00000231, -0.00000860, -0.00000645, -0.00000888, -0.00000265, -0.00000395, -0.00001346, -0.00001876, -0.00001232, -0.00003148, -0.00002811, -0.00009576, -0.00013721, 0.00019946, 0.00007582, 0.00005860, 0.00005399, 0.00004142, 0.00005155, 0.00004038, 0.00013088, 0.00014981, -0.00018542, -0.00005792, -0.00003822, 0.00000086, -0.00000960, 0.00001123, -0.00001670, -0.00001292, 0.00000939, -0.00000058, -0.00000316, -0.00000282, -0.00000215, 0.00000373, -0.00000508, -0.00000306, 0.00001149, -0.00000557, -0.00000013, 0.00000449, 0.00000193, 0.00000980, 0.00000194, 0.00001540, 0.00000003, 0.00000117, -0.00002361, 0.00000249, },
		{ -0.00000487, -0.00000718, -0.00000485, -0.00000272, 0.00000036, -0.00000050, -0.00000263, -0.00000024, -0.00000233, -0.00000067, -0.00000451, -0.00000048, -0.00000263, -0.00000536, 0.00000338, 0.00000316, -0.00000277, -0.00000967, 0.00000237, -0.00000034, 0.00000170, -0.00000589, 0.00000226, 0.00000165, -0.00000551, 0.00000194, -0.00000261, -0.00000198, -0.00000371, 0.00000814, -0.00000730, -0.00000178, 0.00000381, 0.00000819, -0.00000626, -0.00000059, 0.00000285, 0.00000904, -0.00000525, 0.00000084, -0.00000183, -0.00000213, -0.00000006, -0.00000195, -0.00000063, 0.00000208, -0.00000114, 0.00000349, 0.00000001, 0.00000247, 0.00000159, -0.00000323, -0.00000162, 0.00001535, -0.00001219, 0.00001397, -0.00001262, -0.00000247, 0.00000455, -0.00000897, 0.00000543, -0.00000555, -0.00000021, -0.00001275, 0.00000325, 0.00000008, 0.00000089, -0.00000515, -0.00000214, -0.00001045, 0.00000275, -0.00000526, 0.00000149, -0.00000633, 0.00000352, -0.00000926, -0.00000042, -0.00000595, -0.00000315, -0.00001213, -0.00000313, -0.00000863, 0.00000202, -0.00001912, -0.00000940, -0.00002304, -0.00001535, -0.00004540, -0.00004733, -0.00018798, 0.00013795, 0.00013819, 0.00004264, 0.00007436, 0.00003371, 0.00006358, 0.00003403, 0.00014019, 0.00015856, -0.00018808, -0.00006408, -0.00003690, -0.00000487, -0.00000550, 0.00000461, -0.00001130, -0.00001665, 0.00001069, -0.00000258, -0.00000138, -0.00000338, -0.00000260, 0.00000274, -0.00000438, -0.00000421, 0.00001125, -0.00000223, -0.00000497, 0.00000704, -0.00000146, 0.00001182, -0.00000237, 0.00001723, 0.00000112, 0.00000403, -0.00001995, -0.00000419, -0.00000253, },
		{ -0.00000441, -0.00000778, 0.00000020, -0.00000395, 0.00000334, -0.00000510, 0.00000194, -0.00000389, 0.00000027, -0.00000452, -0.00000027, -0.00000268, -0.00000475, -0.00000001, 0.00000538, -0.00000156, -0.00000899, -0.00000137, 0.00000288, -0.00000076, -0.00000184, -0.00000363, 0.00000715, -0.00000841, 0.00000313, -0.00000384, 0.00000081, -0.00000747, 0.00000910, -0.00000471, -0.00000370, 0.00000150, 0.00000965, -0.00000372, -0.00000261, 0.00000125, 0.00000944, -0.00000204, -0.00000219, 0.00000049, -0.00000353, 0.00000030, -0.00000188, -0.00000117, 0.00000206, -0.00000081, 0.00000255, 0.00000105, 0.00000176, 0.00000167, -0.00000122, -0.00000550, 0.00001528, -0.00000687, 0.00000533, 0.00000063, -0.00001465, 0.00001261, -0.00001425, 0.00000854, -0.00000765, 0.00000320, -0.00001449, 0.00000169, 0.00000015, 0.00000274, -0.00000773, 0.00000218, -0.00001335, 0.00000329, -0.00000642, 0.00000292, -0.00000649, 0.00000321, -0.00000969, 0.00000076, -0.00000772, 0.00000004, -0.00001482, -0.00000080, -0.00001087, 0.00000422, -0.00002152, -0.00000366, -0.00002888, -0.00000611, -0.00005374, -0.00002077, -0.00018692, 0.00003237, 0.00020111, 0.00003709, 0.00008904, 0.00003140, 0.00007199, 0.00003177, 0.00014520, 0.00017185, -0.00019323, -0.00006930, -0.00003611, -0.00001029, -0.00000203, -0.00000044, -0.00000870, -0.00001753, 0.00001004, -0.00000304, -0.00000140, -0.00000237, -0.00000375, 0.00000222, -0.00000443, -0.00000431, 0.00000964, 0.00000179, -0.00000862, 0.00000743, -0.00000234, 0.00001110, -0.00000301, 0.00001436, 0.00000636, 0.00000162, -0.00001104, -0.00001329, 0.00000313, -0.00000814, },
		{ -0.00000926, 0.00000190, -0.00000662, 0.00000476, -0.00000513, 0.00000190, -0.00000338, -0.00000053, -0.00000344, -0.00000053, -0.00000275, -0.00000365, -0.00000320, 0.00000633, -0.00000005, -0.00000726, -0.00000553, 0.00000552, -0.00000271, 0.00000165, -0.00000776, 0.00000935, -0.00000761, 0.00000143, -0.00000300, 0.00000138, -0.00000780, 0.00000601, 0.00000050, -0.00000632, 0.00000036, 0.00000844, 0.00000080, -0.00000501, 0.00000053, 0.00000762, 0.00000270, -0.00000527, 0.00000219, -0.00000412, 0.00000008, -0.00000166, -0.00000137, 0.00000137, 0.00000024, 0.00000102, 0.00000240, 0.00000094, 0.00000192, -0.00000032, -0.00000598, 0.00001018, 0.00000264, -0.00000541, 0.00001251, -0.00002164, 0.00001400, -0.00001313, 0.00000605, -0.00000509, 0.00000197, -0.00001125, -0.00000355, 0.00000237, 0.00000247, -0.00000767, 0.00000273, -0.00001141, -0.00000058, -0.00000370, 0.00000025, -0.00000267, -0.00000037, -0.00000729, -0.00000101, -0.00000618, -0.00000083, -0.00001263, -0.00000372, -0.00000759, 0.00000078, -0.00001829, -0.00000485, -0.00002562, -0.00000878, -0.00004659, -0.00002197, -0.00014200, -0.00008827, 0.00023392, 0.00005901, 0.00008867, 0.00004132, 0.00007145, 0.00003811, 0.00014254, 0.00019195, -0.00020252, -0.00007193, -0.00003778, -0.00001316, -0.00000168, -0.00000156, -0.00001051, -0.00001486, 0.00000715, -0.00000169, -0.00000331, -0.00000004, -0.00000530, 0.00000209, -0.00000516, -0.00000372, 0.00000757, 0.00000496, -0.00000963, 0.00000487, -0.00000042, 0.00000761, -0.00000034, 0.00000795, 0.00001396, -0.00000331, -0.00000107, -0.00002028, 0.00000559, -0.00000872, -0.00000351, },
		{ 0.00000109, -0.00000630, 0.00000275, -0.00000219, -0.00000054, -0.00000085, -0.00000281, -0.00000154, -0.00000118, -0.00000255, -0.00000290, -0.00000502, 0.00000522, 0.00000222, -0.00000554, -0.00000836, 0.00000586, -0.00000250, 0.00000270, -0.00000822, 0.00000698, -0.00000284, -0.00000286, -0.00000028, -0.00000031, -0.00000498, 0.00000012, 0.00000644, -0.00000829, -0.00000028, 0.00000560, 0.00000563, -0.00000646, -0.00000008, 0.00000479, 0.00000710, -0.00000665, 0.00000211, -0.00000329, -0.00000095, -0.00000120, -0.00000140, 0.00000040, 0.00000139, -0.00000033, 0.00000327, 0.00000062, 0.00000204, -0.00000030, -0.00000408, 0.00000290, 0.00001165, -0.00001251, 0.00001734, -0.00001931, 0.00000679, -0.00000531, -0.00000149, 0.00000134, -0.00000299, -0.00000444, -0.00001068, 0.00000548, 0.00000061, -0.00000511, -0.00000024, -0.00000555, -0.00000738, 0.00000129, -0.00000487, 0.00000325, -0.00000537, -0.00000355, -0.00000463, -0.00000218, -0.00000493, -0.00000666, -0.00001042, -0.00000050, -0.00000644, -0.00001108, -0.00001156, -0.00001501, -0.00002073, -0.00002799, -0.00004172, -0.00007817, -0.00018601, 0.00020988, 0.00011566, 0.00007245, 0.00006369, 0.00006212, 0.00005279, 0.00013305, 0.00021764, -0.00021485, -0.00007246, -0.00004200, -0.00001308, -0.00000490, 0.00000127, -0.00001599, -0.00001000, 0.00000333, 0.00000043, -0.00000609, 0.00000251, -0.00000647, 0.00000194, -0.00000617, -0.00000306, 0.00000597, 0.00000631, -0.00000773, 0.00000009, 0.00000294, 0.00000310, 0.00000338, 0.00000107, 0.00002032, -0.00000652, 0.00000516, -0.00002116, 0.00000238, -0.00000519, -0.00000557, -0.00000817, },
		{ -0.00000354, -0.00000171, 0.00000237, -0.00000396, 0.00000230, -0.00000528, 0.00000018, -0.00000167, -0.00000218, -0.00000281, -0.00000511, 0.00000219, 0.00000498, -0.00000428, -0.00000902, 0.00000340, 0.00000005, 0.00000112, -0.00000518, 0.00000118, 0.00000388, -0.00000760, 0.00000256, -0.00000271, -0.00000104, -0.00000579, 0.00001036, -0.00000784, -0.00000158, 0.00000275, 0.00000887, -0.00000571, -0.00000145, 0.00000233, 0.00000947, -0.00000524, -0.00000008, -0.00000116, -0.00000253, -0.00000066, -0.00000138, -0.00000041, 0.00000199, -0.00000081, 0.00000308, 0.00000123, 0.00000167, -0.00000048, -0.00000162, -0.00000319, 0.00001605, -0.00001234, 0.00001309, -0.00000820, -0.00000623, 0.00000538, -0.00001015, 0.00000801, -0.00000821, 0.00000225, -0.00001619, 0.00000684, -0.00000087, -0.00000202, -0.00000430, 0.00000105, -0.00001351, 0.00000514, -0.00000903, 0.00000780, -0.00000856, -0.00000128, -0.00000748, 0.00000157, -0.00000921, -0.00000052, -0.00001682, 0.00000592, -0.00001271, -0.00000472, -0.00001847, -0.00000363, -0.00003335, -0.00000937, -0.00006216, -0.00002523, -0.00023005, 0.00012049, 0.00019600, 0.00005222, 0.00008998, 0.00005099, 0.00006996, 0.00012226, 0.00024405, -0.00022635, -0.00007378, -0.00004657, -0.00001198, -0.00000965, 0.00000560, -0.00002211, -0.00000603, 0.00000108, 0.00000144, -0.00000807, 0.00000383, -0.00000631, 0.00000118, -0.00000686, -0.00000303, 0.00000551, 0.00000566, -0.00000385, -0.00000503, 0.00000550, -0.00000009, 0.00000545, -0.00000313, 0.00002225, -0.00000486, 0.00000498, -0.00001490, -0.00000589, 0.00000125, -0.00000953, -0.00000525, -0.00000169, },
		{ -0.00000614, 0.00000597, -0.00000605, 0.00000406, -0.00000627, 0.00000052, -0.00000125, -0.00000204, -0.00000311, -0.00000401, -0.00000162, 0.00000721, -0.00000311, -0.00000792, -0.00000077, 0.00000350, -0.00000151, -0.00000075, -0.00000509, 0.00000920, -0.00000989, 0.00000328, -0.00000372, 0.00000152, -0.00000894, 0.00001016, -0.00000406, -0.00000412, 0.00000106, 0.00000935, -0.00000238, -0.00000374, 0.00000096, 0.00000914, -0.00000123, -0.00000352, 0.00000129, -0.00000387, -0.00000044, -0.00000128, -0.00000087, 0.00000179, -0.00000026, 0.00000182, 0.00000261, 0.00000089, -0.00000043, 0.00000003, -0.00000602, 0.00001434, -0.00000519, 0.00000245, 0.00000658, -0.00001867, 0.00001279, -0.00001464, 0.00001047, -0.00000971, 0.00000494, -0.00001681, 0.00000413, 0.00000002, -0.00000076, -0.00000646, 0.00000477, -0.00001538, 0.00000464, -0.00000923, 0.00000806, -0.00000733, -0.00000281, -0.00000719, 0.00000234, -0.00001041, 0.00000195, -0.00001864, 0.00000719, -0.00001337, -0.00000410, -0.00001994, 0.00000140, -0.00003764, -0.00000209, -0.00006751, -0.00000366, -0.00021067, -0.00001672, 0.00027015, 0.00004933, 0.00010641, 0.00004865, 0.00008109, 0.00011762, 0.00026509, -0.00023229, -0.00007954, -0.00004830, -0.00001300, -0.00001258, 0.00000775, -0.00002513, -0.00000618, 0.00000266, -0.00000027, -0.00000784, 0.00000280, -0.00000421, -0.00000054, -0.00000677, -0.00000408, 0.00000636, 0.00000351, 0.00000051, -0.00000851, 0.00000550, -0.00000043, 0.00000437, -0.00000321, 0.00001890, 0.00000179, -0.00000060, -0.00000432, -0.00001540, 0.00000702, -0.00001253, -0.00000281, -0.00000438, -0.00000050, },
		{ 0.00000633, -0.00000499, 0.00000303, -0.00000476, -0.00000115, 0.00000042, -0.00000248, -0.00000324, -0.00000279, -0.00000467, 0.00000762, -0.00000126, -0.00000623, -0.00000468, 0.00000564, -0.00000296, 0.00000240, -0.00000868, 0.00001027, -0.00000780, 0.00000066, -0.00000218, 0.00000128, -0.00000810, 0.00000562, 0.00000223, -0.00000719, 0.00000049, 0.00000734, 0.00000254, -0.00000609, 0.00000036, 0.00000676, 0.00000386, -0.00000644, 0.00000263, -0.00000404, -0.00000101, -0.00000093, -0.00000104, 0.00000099, 0.00000091, 0.00000013, 0.00000405, 0.00000024, -0.00000029, 0.00000054, -0.00000556, 0.00000813, 0.00000511, -0.00000869, 0.00001788, -0.00002377, 0.00001179, -0.00001146, 0.00000630, -0.00000556, 0.00000201, -0.00001160, -0.00000283, 0.00000361, -0.00000232, -0.00000512, 0.00000382, -0.00001160, -0.00000096, -0.00000488, 0.00000354, -0.00000152, -0.00000818, -0.00000343, -0.00000068, -0.00000722, -0.00000089, -0.00001423, 0.00000177, -0.00000700, -0.00001069, -0.00001395, -0.00000287, -0.00002994, -0.00001023, -0.00005300, -0.00001555, -0.00014380, -0.00016212, 0.00029961, 0.00008392, 0.00010255, 0.00006256, 0.00008052, 0.00012385, 0.00027737, -0.00023042, -0.00009156, -0.00004534, -0.00001829, -0.00001127, 0.00000518, -0.00002279, -0.00001183, 0.00000854, -0.00000479, -0.00000529, -0.00000055, -0.00000050, -0.00000298, -0.00000588, -0.00000621, 0.00000820, 0.00000079, 0.00000388, -0.00000919, 0.00000258, 0.00000184, 0.00000076, -0.00000029, 0.00001232, 0.00001039, -0.00000737, 0.00000525, -0.00002094, 0.00000808, -0.00001164, -0.00000319, -0.00000479, 0.00000049, -0.00000805, },
		{ -0.00000077, -0.00000049, -0.00000124, -0.00000423, 0.00000285, -0.00000335, -0.00000291, -0.00000231, -0.00000583, 0.00000558, 0.00000161, -0.00000497, -0.00000667, 0.00000494, -0.00000182, 0.00000263, -0.00000804, 0.00000639, -0.00000172, -0.00000444, 0.00000120, -0.00000124, -0.00000425, -0.00000124, 0.00000853, -0.00000899, -0.00000004, 0.00000426, 0.00000714, -0.00000702, -0.00000043, 0.00000381, 0.00000796, -0.00000708, 0.00000181, -0.00000259, -0.00000248, -0.00000030, -0.00000113, 0.00000006, 0.00000193, -0.00000107, 0.00000465, 0.00000042, -0.00000052, 0.00000047, -0.00000333, 0.00000084, 0.00001341, -0.00001442, 0.00002030, -0.00001819, 0.00000152, -0.00000140, -0.00000295, 0.00000250, -0.00000471, -0.00000297, -0.00001136, 0.00000787, -0.00000548, -0.00000120, -0.00000069, -0.00000402, -0.00000924, 0.00000153, -0.00000319, 0.00000613, -0.00001474, 0.00000157, -0.00000564, -0.00000137, -0.00000719, -0.00000584, -0.00000761, 0.00000332, -0.00002115, -0.00000375, -0.00001310, -0.00001440, -0.00002805, -0.00002687, -0.00004533, -0.00006335, -0.00026927, 0.00025471, 0.00016119, 0.00008044, 0.00009071, 0.00007034, 0.00013900, 0.00028334, -0.00022322, -0.00010784, -0.00003896, -0.00002718, -0.00000607, -0.00000176, -0.00001591, -0.00002141, 0.00001660, -0.00001021, -0.00000192, -0.00000482, 0.00000350, -0.00000518, -0.00000475, -0.00000887, 0.00001025, -0.00000149, 0.00000539, -0.00000712, -0.00000206, 0.00000487, -0.00000318, 0.00000285, 0.00000622, 0.00001657, -0.00001041, 0.00000888, -0.00001887, 0.00000253, -0.00000603, -0.00000695, -0.00000223, -0.00000145, -0.00000652, 0.00000292, },
		{ -0.00000450, 0.00000241, -0.00000708, 0.00000480, -0.00000393, -0.00000236, -0.00000269, -0.00000506, 0.00000162, 0.00000491, -0.00000427, -0.00000633, 0.00000143, 0.00000152, 0.00000036, -0.00000403, -0.00000053, 0.00000562, -0.00000927, 0.00000414, -0.00000391, 0.00000002, -0.00000716, 0.00001183, -0.00000769, -0.00000172, 0.00000177, 0.00000955, -0.00000542, -0.00000217, 0.00000161, 0.00000955, -0.00000464, -0.00000114, 0.00000004, -0.00000434, 0.00000032, -0.00000117, -0.00000060, 0.00000224, -0.00000118, 0.00000392, 0.00000174, -0.00000139, 0.00000051, -0.00000120, -0.00000427, 0.00001600, -0.00001192, 0.00001311, -0.00000403, -0.00001364, 0.00001036, -0.00001222, 0.00000983, -0.00001071, 0.00000443, -0.00001715, 0.00000949, -0.00000753, 0.00000251, -0.00000543, 0.00000325, -0.00001577, 0.00000576, -0.00000795, 0.00001138, -0.00001851, 0.00000424, -0.00000908, 0.00000346, -0.00001279, 0.00000166, -0.00001545, 0.00001159, -0.00002912, 0.00000399, -0.00002181, -0.00000025, -0.00004368, -0.00000464, -0.00006989, -0.00000484, -0.00030511, 0.00013143, 0.00026223, 0.00005814, 0.00012021, 0.00006103, 0.00015439, 0.00029098, -0.00021756, -0.00012296, -0.00003340, -0.00003616, -0.00000020, -0.00000973, -0.00000820, -0.00003081, 0.00002294, -0.00001339, -0.00000014, -0.00000790, 0.00000595, -0.00000592, -0.00000422, -0.00001119, 0.00001157, -0.00000250, 0.00000486, -0.00000337, -0.00000643, 0.00000639, -0.00000507, 0.00000352, 0.00000371, 0.00001720, -0.00000702, 0.00000492, -0.00000948, -0.00000779, 0.00000202, -0.00001213, 0.00000176, -0.00000506, -0.00000283, -0.00000256, 0.00000457, },
		{ 0.00000397, -0.00000787, 0.00000497, -0.00000337, -0.00000216, -0.00000338, -0.00000331, -0.00000270, 0.00000723, -0.00000339, -0.00000453, -0.00000325, 0.00000512, -0.00000237, 0.00000072, -0.00000699, 0.00001048, -0.00001076, 0.00000427, -0.00000452, 0.00000208, -0.00000943, 0.00001032, -0.00000275, -0.00000482, 0.00000070, 0.00000906, -0.00000134, -0.00000472, 0.00000057, 0.00000838, 0.00000016, -0.00000492, 0.00000256, -0.00000560, 0.00000038, -0.00000102, -0.00000092, 0.00000173, -0.00000023, 0.00000214, 0.00000374, -0.00000259, 0.00000085, -0.00000018, -0.00000577, 0.00001240, -0.00000282, 0.00000054, 0.00001205, -0.00002610, 0.00001684, -0.00001567, 0.00001151, -0.00001162, 0.00000641, -0.00001672, 0.00000599, -0.00000626, 0.00000333, -0.00000703, 0.00000631, -0.00001673, 0.00000446, -0.00000758, 0.00001117, -0.00001675, 0.00000205, -0.00000830, 0.00000407, -0.00001381, 0.00000375, -0.00001683, 0.00001264, -0.00002923, 0.00000359, -0.00002237, 0.00000414, -0.00004673, 0.00000083, -0.00007211, 0.00001148, -0.00026521, -0.00004172, 0.00034621, 0.00006368, 0.00013335, 0.00006610, 0.00015923, 0.00030962, -0.00022097, -0.00013120, -0.00003330, -0.00004105, 0.00000222, -0.00001447, -0.00000412, -0.00003568, 0.00002390, -0.00001177, -0.00000181, -0.00000818, 0.00000558, -0.00000440, -0.00000490, -0.00001242, 0.00001144, -0.00000188, 0.00000272, 0.00000062, -0.00000874, 0.00000502, -0.00000387, 0.00000078, 0.00000556, 0.00001226, 0.00000170, -0.00000395, 0.00000283, -0.00001780, 0.00000798, -0.00001514, 0.00000430, -0.00000773, 0.00000030, -0.00000711, 0.00000808, -0.00000638, },
		{ -0.00000586, 0.00000294, -0.00000143, -0.00000267, -0.00000366, -0.00000189, -0.00000562, 0.00000718, -0.00000143, -0.00000267, -0.00000694, 0.00000666, -0.00000326, 0.00000346, -0.00000988, 0.00001026, -0.00000736, 0.00000076, -0.00000225, 0.00000094, -0.00000738, 0.00000448, 0.00000433, -0.00000809, 0.00000053, 0.00000641, 0.00000380, -0.00000689, 0.00000006, 0.00000551, 0.00000543, -0.00000751, 0.00000345, -0.00000534, -0.00000057, -0.00000051, -0.00000105, 0.00000075, 0.00000112, 0.00000019, 0.00000546, -0.00000330, 0.00000107, -0.00000014, -0.00000436, 0.00000517, 0.00000800, -0.00001060, 0.00002234, -0.00002906, 0.00001330, -0.00001031, 0.00000568, -0.00000601, 0.00000191, -0.00000981, -0.00000216, -0.00000185, 0.00000069, -0.00000437, 0.00000390, -0.00001138, -0.00000248, -0.00000212, 0.00000566, -0.00000990, -0.00000450, -0.00000341, 0.00000005, -0.00000939, -0.00000066, -0.00001068, 0.00000563, -0.00002082, -0.00000559, -0.00001359, -0.00000313, -0.00003494, -0.00001253, -0.00005032, -0.00001068, -0.00017596, -0.00021192, 0.00036544, 0.00012015, 0.00011921, 0.00009322, 0.00014794, 0.00034387, -0.00023650, -0.00013070, -0.00004020, -0.00004013, -0.00000088, -0.00001366, -0.00000600, -0.00003411, 0.00001853, -0.00000529, -0.00000669, -0.00000578, 0.00000259, -0.00000104, -0.00000655, -0.00001248, 0.00000982, 0.00000003, -0.00000011, 0.00000353, -0.00000819, 0.00000094, -0.00000045, -0.00000408, 0.00000988, 0.00000478, 0.00001146, -0.00001223, 0.00001180, -0.00002187, 0.00000770, -0.00001313, 0.00000311, -0.00000730, 0.00000059, -0.00000832, 0.00000760, -0.00000452, 0.00000230, },
		{ -0.00000028, 0.00000112, -0.00000359, -0.00000333, -0.00000155, -0.00000622, 0.00000440, 0.00000173, -0.00000157, -0.00000823, 0.00000496, -0.00000129, 0.00000299, -0.00000818, 0.00000498, -0.00000022, -0.00000484, 0.00000157, -0.00000222, -0.00000275, -0.00000292, 0.00001052, -0.00000938, -0.00000013, 0.00000335, 0.00000791, -0.00000716, -0.00000097, 0.00000255, 0.00000904, -0.00000727, 0.00000188, -0.00000332, -0.00000242, 0.00000027, -0.00000116, -0.00000018, 0.00000210, -0.00000097, 0.00000596, -0.00000283, 0.00000062, -0.00000028, -0.00000193, -0.00000191, 0.00001526, -0.00001462, 0.00002222, -0.00002040, 0.00000029, 0.00000167, -0.00000489, 0.00000320, -0.00000603, -0.00000001, -0.00001124, 0.00000276, -0.00000328, 0.00000065, -0.00000175, -0.00000274, -0.00001149, 0.00000481, -0.00000149, -0.00000179, -0.00001171, 0.00000233, -0.00000569, -0.00000243, -0.00000840, -0.00000080, -0.00000501, -0.00000886, -0.00001820, -0.00000101, -0.00001605, -0.00001575, -0.00003479, -0.00001833, -0.00004787, -0.00008198, -0.00032329, 0.00028799, 0.00022878, 0.00008435, 0.00013687, 0.00012563, 0.00038935, -0.00025951, -0.00012602, -0.00005043, -0.00003608, -0.00000753, -0.00000892, -0.00001220, -0.00002820, 0.00000958, 0.00000293, -0.00001206, -0.00000287, -0.00000107, 0.00000240, -0.00000791, -0.00001219, 0.00000754, 0.00000222, -0.00000243, 0.00000448, -0.00000508, -0.00000433, 0.00000304, -0.00000851, 0.00001349, -0.00000112, 0.00001740, -0.00001477, 0.00001289, -0.00001718, 0.00000021, -0.00000612, -0.00000159, -0.00000378, -0.00000222, -0.00000590, 0.00000330, 0.00000041, -0.00000173, -0.00000204, },
		{ 0.00000277, -0.00000398, -0.00000283, -0.00000213, -0.00000489, -0.00000008, 0.00000502, -0.00000095, -0.00000722, 0.00000068, 0.00000251, 0.00000035, -0.00000349, -0.00000266, 0.00000721, -0.00000928, 0.00000424, -0.00000484, 0.00000141, -0.00000834, 0.00001274, -0.00000695, -0.00000241, 0.00000143, 0.00000936, -0.00000469, -0.00000307, 0.00000064, 0.00000980, -0.00000387, -0.00000165, -0.00000039, -0.00000439, 0.00000085, -0.00000115, -0.00000083, 0.00000226, -0.00000089, 0.00000498, -0.00000114, -0.00000054, -0.00000007, -0.00000015, -0.00000602, 0.00001605, -0.00000995, 0.00001257, -0.00000408, -0.00001608, 0.00001365, -0.00001392, 0.00001026, -0.00001188, 0.00000706, -0.00001609, 0.00000351, -0.00000508, 0.00000432, -0.00000632, 0.00000411, -0.00001724, 0.00000812, -0.00000525, 0.00000251, -0.00001479, 0.00000446, -0.00000867, 0.00000217, -0.00001388, 0.00000637, -0.00001209, -0.00000117, -0.00002589, 0.00000652, -0.00002453, -0.00000158, -0.00005028, 0.00000368, -0.00007120, -0.00002455, -0.00034160, 0.00011703, 0.00035968, 0.00005532, 0.00017846, 0.00010725, 0.00043384, -0.00027899, -0.00012672, -0.00005635, -0.00003508, -0.00001239, -0.00000521, -0.00001773, -0.00002316, 0.00000247, 0.00000791, -0.00001391, -0.00000260, -0.00000264, 0.00000356, -0.00000726, -0.00001289, 0.00000597, 0.00000327, -0.00000309, 0.00000312, -0.00000039, -0.00000891, 0.00000457, -0.00001044, 0.00001403, -0.00000272, 0.00001696, -0.00000981, 0.00000590, -0.00000564, -0.00001114, 0.00000230, -0.00000677, 0.00000040, -0.00000605, -0.00000198, -0.00000235, 0.00000571, -0.00000550, 0.00000123, -0.00000268, },
		{ -0.00000291, -0.00000294, -0.00000274, -0.00000303, -0.00000426, 0.00000670, 0.00000030, -0.00000525, -0.00000403, 0.00000574, -0.00000198, 0.00000110, -0.00000876, 0.00001102, -0.00000949, 0.00000343, -0.00000471, 0.00000260, -0.00000933, 0.00000965, -0.00000080, -0.00000607, 0.00000102, 0.00000793, -0.00000002, -0.00000569, -0.00000021, 0.00000796, 0.00000143, -0.00000538, 0.00000185, -0.00000531, 0.00000056, -0.00000077, -0.00000125, 0.00000174, 0.00000007, 0.00000314, 0.00000096, -0.00000174, 0.00000030, 0.00000051, -0.00000656, 0.00001110, 0.00000031, -0.00000078, 0.00001184, -0.00002720, 0.00001810, -0.00001523, 0.00001003, -0.00001100, 0.00000704, -0.00001323, -0.00000206, -0.00000239, 0.00000377, -0.00000624, 0.00000512, -0.00001584, 0.00000443, -0.00000240, -0.00000013, -0.00001089, 0.00000029, -0.00000588, 0.00000077, -0.00001270, 0.00000577, -0.00001011, -0.00000349, -0.00002269, 0.00000259, -0.00002105, -0.00000179, -0.00004754, 0.00000191, -0.00006307, -0.00002211, -0.00027043, -0.00010425, 0.00045779, 0.00006833, 0.00019580, 0.00010962, 0.00046427, -0.00028371, -0.00014202, -0.00005084, -0.00004300, -0.00001013, -0.00000777, -0.00001732, -0.00002430, 0.00000216, 0.00000560, -0.00000939, -0.00000711, -0.00000024, 0.00000088, -0.00000351, -0.00001554, 0.00000626, 0.00000210, -0.00000143, -0.00000031, 0.00000466, -0.00001145, 0.00000337, -0.00000954, 0.00001138, -0.00000020, 0.00001125, 0.00000001, -0.00000470, 0.00000674, -0.00002010, 0.00000679, -0.00000826, 0.00000187, -0.00000781, 0.00000029, -0.00000593, 0.00000808, -0.00000618, 0.00000158, -0.00000239, 0.00000217, },
		{ -0.00000399, -0.00000258, -0.00000207, -0.00000630, 0.00000541, 0.00000308, -0.00000380, -0.00000695, 0.00000614, -0.00000180, 0.00000296, -0.00001043, 0.00000906, -0.00000448, -0.00000114, -0.00000157, 0.00000039, -0.00000595, 0.00000257, 0.00000690, -0.00000939, 0.00000110, 0.00000490, 0.00000502, -0.00000739, -0.00000088, 0.00000492, 0.00000644, -0.00000719, 0.00000190, -0.00000432, -0.00000096, 0.00000014, -0.00000169, 0.00000102, 0.00000110, 0.00000157, 0.00000226, -0.00000195, 0.00000012, 0.00000066, -0.00000495, 0.00000380, 0.00001046, -0.00001041, 0.00001951, -0.00002669, 0.00001101, -0.00000674, 0.00000147, -0.00000282, -0.00000022, -0.00000333, -0.00001264, 0.00000388, -0.00000091, -0.00000109, -0.00000015, -0.00000746, -0.00000547, 0.00000611, -0.00000865, -0.00000133, -0.00000888, 0.00000184, -0.00000634, -0.00000475, -0.00000285, 0.00000095, -0.00001550, -0.00000923, -0.00001202, -0.00000598, -0.00001644, -0.00002671, -0.00002247, -0.00002662, -0.00006292, -0.00014976, -0.00030664, 0.00046458, 0.00014905, 0.00017851, 0.00013966, 0.00047614, -0.00026999, -0.00017434, -0.00003270, -0.00006084, 0.00000055, -0.00001825, -0.00000909, -0.00003341, 0.00000992, -0.00000431, 0.00000098, -0.00001570, 0.00000565, -0.00000515, 0.00000273, -0.00001984, 0.00000846, -0.00000130, 0.00000217, -0.00000493, 0.00000895, -0.00001156, 0.00000018, -0.00000729, 0.00000750, 0.00000376, 0.00000423, 0.00000933, -0.00001238, 0.00001317, -0.00002100, 0.00000342, -0.00000353, -0.00000131, -0.00000563, -0.00000114, -0.00000533, 0.00000574, -0.00000257, -0.00000172, 0.00000100, -0.00000079, -0.00000030, },
		{ -0.00000161, -0.00000258, -0.00000562, 0.00000117, 0.00000714, -0.00000340, -0.00000708, 0.00000304, 0.00000132, 0.00000149, -0.00000739, 0.00000218, 0.00000375, -0.00000720, 0.00000265, -0.00000337, -0.00000066, -0.00000513, 0.00001267, -0.00000996, 0.00000011, 0.00000215, 0.00000831, -0.00000672, -0.00000240, 0.00000228, 0.00000917, -0.00000572, -0.00000068, -0.00000157, -0.00000335, 0.00000132, -0.00000218, 0.00000053, 0.00000147, 0.00000115, 0.00000193, -0.00000059, -0.00000108, 0.00000112, -0.00000320, -0.00000204, 0.00001549, -0.00001138, 0.00001555, -0.00001412, -0.00000526, 0.00000774, -0.00001116, 0.00000836, -0.00001029, 0.00000865, -0.00002326, 0.00000968, -0.00000647, 0.00000600, -0.00000813, 0.00000349, -0.00001665, 0.00001534, -0.00001806, 0.00000875, -0.00001806, 0.00000991, -0.00001484, 0.00000546, -0.00001445, 0.00001526, -0.00003050, 0.00000701, -0.00002922, 0.00001211, -0.00003591, 0.00000033, -0.00005398, 0.00001719, -0.00011365, -0.00003694, -0.00042351, 0.00034681, 0.00029197, 0.00014038, 0.00018649, 0.00047925, -0.00024606, -0.00021602, -0.00000907, -0.00008267, 0.00001496, -0.00003280, 0.00000354, -0.00004707, 0.00002196, -0.00001751, 0.00001277, -0.00002456, 0.00001194, -0.00001180, 0.00000906, -0.00002399, 0.00001127, -0.00000561, 0.00000625, -0.00000917, 0.00001146, -0.00000960, -0.00000337, -0.00000590, 0.00000500, 0.00000591, 0.00000012, 0.00001322, -0.00001230, 0.00000996, -0.00001223, -0.00000747, 0.00000616, -0.00000787, -0.00000042, -0.00000567, -0.00000109, -0.00000030, 0.00000389, -0.00000710, 0.00000588, -0.00000467, 0.00000275, -0.00000544, },
		{ -0.00000395, -0.00000315, -0.00000436, 0.00001091, -0.00000324, -0.00000514, -0.00000217, 0.00000570, -0.00000160, -0.00000196, -0.00000614, 0.00001115, -0.00001111, 0.00000491, -0.00000581, 0.00000323, -0.00000980, 0.00001361, -0.00000632, -0.00000282, 0.00000097, 0.00000862, -0.00000325, -0.00000497, 0.00000085, 0.00000888, -0.00000124, -0.00000484, 0.00000167, -0.00000548, 0.00000205, -0.00000245, 0.00000032, 0.00000100, 0.00000199, 0.00000010, 0.00000196, -0.00000294, 0.00000206, -0.00000248, -0.00000435, 0.00001368, -0.00000372, 0.00000282, 0.00000455, -0.00002281, 0.00002013, -0.00002036, 0.00001568, -0.00001657, 0.00001609, -0.00002793, 0.00001019, -0.00000856, 0.00001036, -0.00001346, 0.00001091, -0.00002281, 0.00001927, -0.00002239, 0.00001332, -0.00002146, 0.00001279, -0.00001899, 0.00001156, -0.00002175, 0.00002451, -0.00003928, 0.00001607, -0.00003833, 0.00002168, -0.00004698, 0.00001746, -0.00007256, 0.00004294, -0.00013919, 0.00002092, -0.00042088, 0.00011840, 0.00045157, 0.00011974, 0.00022378, 0.00049488, -0.00022901, -0.00025240, 0.00000738, -0.00009814, 0.00002448, -0.00004383, 0.00001344, -0.00005828, 0.00003122, -0.00002704, 0.00001974, -0.00002863, 0.00001454, -0.00001544, 0.00001234, -0.00002556, 0.00001261, -0.00000872, 0.00000871, -0.00001133, 0.00001141, -0.00000621, -0.00000583, -0.00000684, 0.00000531, 0.00000457, 0.00000083, 0.00001014, -0.00000413, -0.00000118, 0.00000234, -0.00002086, 0.00001585, -0.00001366, 0.00000444, -0.00001033, 0.00000378, -0.00000675, 0.00000977, -0.00001134, 0.00000941, -0.00000708, 0.00000445, -0.00000611, -0.00000059, },
		{ -0.00000072, -0.00000881, 0.00001234, -0.00000191, -0.00000286, -0.00000697, 0.00000860, -0.00000356, 0.00000247, -0.00001180, 0.00001382, -0.00000999, 0.00000315, -0.00000496, 0.00000353, -0.00000949, 0.00000905, 0.00000085, -0.00000694, 0.00000119, 0.00000632, 0.00000190, -0.00000761, 0.00000023, 0.00000637, 0.00000448, -0.00000843, 0.00000355, -0.00000604, 0.00000154, -0.00000208, 0.00000011, 0.00000009, 0.00000341, -0.00000220, 0.00000443, -0.00000436, 0.00000279, -0.00000260, -0.00000355, 0.00000706, 0.00000798, -0.00001138, 0.00002004, -0.00003235, 0.00002261, -0.00001991, 0.00001399, -0.00001442, 0.00001476, -0.00002337, 0.00000313, -0.00000492, 0.00000906, -0.00001251, 0.00001077, -0.00002019, 0.00001464, -0.00001852, 0.00000932, -0.00001627, 0.00000775, -0.00001557, 0.00000956, -0.00001996, 0.00002321, -0.00003596, 0.00001182, -0.00003299, 0.00001589, -0.00004152, 0.00001467, -0.00006623, 0.00003660, -0.00012259, 0.00000913, -0.00031386, -0.00015754, 0.00055379, 0.00016344, 0.00022399, 0.00054396, -0.00023471, -0.00027055, 0.00000606, -0.00009890, 0.00002207, -0.00004485, 0.00001432, -0.00006081, 0.00003167, -0.00002740, 0.00001748, -0.00002477, 0.00001100, -0.00001387, 0.00001069, -0.00002308, 0.00001108, -0.00000901, 0.00000803, -0.00001037, 0.00000870, -0.00000215, -0.00000651, -0.00001006, 0.00000787, 0.00000061, 0.00000504, 0.00000256, 0.00000785, -0.00001404, 0.00001525, -0.00002948, 0.00001956, -0.00001410, 0.00000524, -0.00001172, 0.00000583, -0.00000996, 0.00001163, -0.00001156, 0.00000923, -0.00000617, 0.00000348, -0.00000500, -0.00000047, -0.00000499, },
		{ -0.00001042, 0.00001024, 0.00000135, -0.00000167, -0.00000924, 0.00000799, -0.00000248, 0.00000360, -0.00001238, 0.00001028, -0.00000361, -0.00000222, -0.00000119, 0.00000050, -0.00000508, 0.00000111, 0.00000872, -0.00001000, 0.00000134, 0.00000311, 0.00000663, -0.00000871, -0.00000068, 0.00000327, 0.00000907, -0.00000934, 0.00000273, -0.00000438, -0.00000040, -0.00000101, -0.00000035, -0.00000068, 0.00000448, -0.00000372, 0.00000547, -0.00000417, 0.00000242, -0.00000258, -0.00000159, -0.00000037, 0.00001753, -0.00001948, 0.00002492, -0.00002842, 0.00001242, -0.00000897, 0.00000344, -0.00000446, 0.00000570, -0.00001150, -0.00000905, 0.00000255, 0.00000315, -0.00000590, 0.00000380, -0.00001019, 0.00000348, -0.00000860, -0.00000097, -0.00000501, -0.00000279, -0.00000652, 0.00000085, -0.00001019, 0.00001254, -0.00002213, -0.00000359, -0.00001586, -0.00000225, -0.00002243, -0.00000507, -0.00003894, 0.00000439, -0.00007406, -0.00004939, -0.00016028, -0.00039058, 0.00052620, 0.00029961, 0.00017857, 0.00063204, -0.00026589, -0.00026900, -0.00001352, -0.00008533, 0.00000806, -0.00003571, 0.00000558, -0.00005383, 0.00002258, -0.00001844, 0.00000675, -0.00001442, 0.00000278, -0.00000820, 0.00000512, -0.00001755, 0.00000724, -0.00000663, 0.00000440, -0.00000686, 0.00000435, 0.00000149, -0.00000531, -0.00001436, 0.00001069, -0.00000347, 0.00000943, -0.00000482, 0.00001743, -0.00002147, 0.00001964, -0.00002807, 0.00001401, -0.00000739, 0.00000071, -0.00000855, 0.00000353, -0.00000834, 0.00000828, -0.00000713, 0.00000514, -0.00000204, 0.00000006, -0.00000226, -0.00000148, -0.00000489, 0.00000021, },
		{ 0.00000523, 0.00000577, -0.00000159, -0.00000855, 0.00000392, 0.00000136, 0.00000155, -0.00000842, 0.00000233, 0.00000515, -0.00000824, 0.00000295, -0.00000339, 0.00000038, -0.00000635, 0.00001360, -0.00000957, -0.00000016, 0.00000088, 0.00000895, -0.00000705, -0.00000268, 0.00000101, 0.00001088, -0.00000673, -0.00000070, -0.00000111, -0.00000302, 0.00000020, -0.00000087, -0.00000110, 0.00000466, -0.00000380, 0.00000458, -0.00000216, 0.00000078, -0.00000183, -0.00000025, -0.00000522, 0.00002073, -0.00001787, 0.00001764, -0.00001274, -0.00000594, 0.00000668, -0.00000992, 0.00000733, -0.00000497, 0.00000109, -0.00001990, 0.00000842, -0.00000266, 0.00000167, -0.00000476, 0.00000108, -0.00000760, 0.00000076, -0.00001088, 0.00000546, -0.00001208, 0.00000177, -0.00000820, 0.00000077, 0.00000011, -0.00000649, -0.00002033, 0.00000217, -0.00002090, -0.00000254, -0.00002715, -0.00000860, -0.00003082, -0.00002399, -0.00010870, -0.00003350, -0.00050080, 0.00033466, 0.00051239, 0.00011224, 0.00074082, -0.00030627, -0.00026206, -0.00003869, -0.00006874, -0.00000808, -0.00002426, -0.00000603, -0.00004349, 0.00000999, -0.00000639, -0.00000596, -0.00000386, -0.00000473, -0.00000291, -0.00000039, -0.00001245, 0.00000385, -0.00000382, 0.00000005, -0.00000311, 0.00000062, 0.00000322, -0.00000229, -0.00001842, 0.00001195, -0.00000544, 0.00001111, -0.00000820, 0.00002027, -0.00001957, 0.00001333, -0.00001681, 0.00000127, 0.00000366, -0.00000655, -0.00000287, -0.00000146, -0.00000350, 0.00000173, -0.00000042, -0.00000044, 0.00000303, -0.00000378, 0.00000059, -0.00000276, -0.00000392, -0.00000053, -0.00000911, },
		{ 0.00000920, -0.00000126, -0.00000631, -0.00000153, 0.00000560, -0.00000134, -0.00000294, -0.00000592, 0.00001180, -0.00001103, 0.00000430, -0.00000514, 0.00000349, -0.00000980, 0.00001297, -0.00000466, -0.00000376, 0.00000045, 0.00000820, -0.00000280, -0.00000550, -0.00000012, 0.00000980, -0.00000155, -0.00000504, 0.00000198, -0.00000480, 0.00000052, -0.00000085, -0.00000147, 0.00000424, -0.00000286, 0.00000247, 0.00000067, -0.00000114, -0.00000087, 0.00000011, -0.00000658, 0.00001743, -0.00000853, 0.00000327, 0.00000654, -0.00002274, 0.00001751, -0.00001739, 0.00001296, -0.00000955, 0.00000679, -0.00002254, 0.00000698, -0.00000308, 0.00000442, -0.00000837, 0.00000634, -0.00001118, 0.00000240, -0.00001324, 0.00000788, -0.00001303, 0.00000229, -0.00001020, 0.00000461, -0.00000463, 0.00000028, -0.00002662, 0.00000832, -0.00002633, 0.00000320, -0.00003433, 0.00000399, -0.00004351, -0.00000419, -0.00012630, 0.00001272, -0.00045533, 0.00000820, 0.00073405, 0.00008006, 0.00083404, -0.00032654, -0.00027397, -0.00004905, -0.00006642, -0.00001205, -0.00002265, -0.00000969, -0.00003991, 0.00000363, -0.00000075, -0.00001159, -0.00000114, -0.00000491, -0.00000353, -0.00000090, -0.00001216, 0.00000454, -0.00000382, -0.00000174, -0.00000231, 0.00000018, 0.00000144, 0.00000272, -0.00002171, 0.00001124, -0.00000499, 0.00000962, -0.00000699, 0.00001637, -0.00001000, 0.00000028, -0.00000175, -0.00001176, 0.00001261, -0.00001120, 0.00000089, -0.00000519, 0.00000065, -0.00000387, 0.00000440, -0.00000371, 0.00000569, -0.00000525, 0.00000141, -0.00000277, -0.00000327, -0.00000180, -0.00000649, -0.00000037, },
		{ 0.00000096, -0.00000458, -0.00000550, 0.00000731, -0.00000215, 0.00000056, -0.00001036, 0.00001270, -0.00000812, 0.00000122, -0.00000322, 0.00000260, -0.00000801, 0.00000696, 0.00000331, -0.00000807, 0.00000104, 0.00000538, 0.00000240, -0.00000773, -0.00000093, 0.00000714, 0.00000391, -0.00000781, 0.00000285, -0.00000432, -0.00000094, 0.00000027, -0.00000234, 0.00000393, -0.00000198, 0.00000065, 0.00000253, -0.00000167, -0.00000097, 0.00000052, -0.00000597, 0.00001103, 0.00000248, -0.00000957, 0.00001928, -0.00002832, 0.00001571, -0.00001292, 0.00000748, -0.00000362, 0.00000163, -0.00001405, -0.00000372, 0.00000404, -0.00000053, -0.00000346, 0.00000169, -0.00000370, -0.00000675, -0.00000508, -0.00000072, -0.00000286, -0.00000777, -0.00000169, -0.00000302, 0.00000353, -0.00000771, -0.00001623, -0.00000381, -0.00001187, -0.00001241, -0.00001808, -0.00001154, -0.00002158, -0.00002823, -0.00008696, -0.00003052, -0.00028507, -0.00036288, 0.00086373, 0.00014356, 0.00087838, -0.00030134, -0.00032474, -0.00002874, -0.00009109, 0.00000627, -0.00003948, 0.00000253, -0.00005066, 0.00001084, -0.00000845, -0.00000398, -0.00001116, 0.00000587, -0.00001300, 0.00000633, -0.00001909, 0.00001133, -0.00000870, 0.00000133, -0.00000661, 0.00000463, -0.00000451, 0.00000961, -0.00002443, 0.00000965, -0.00000382, 0.00000702, -0.00000386, 0.00000975, 0.00000121, -0.00001155, 0.00000836, -0.00001705, 0.00001304, -0.00000845, -0.00000110, -0.00000412, 0.00000046, -0.00000474, 0.00000387, -0.00000181, 0.00000366, -0.00000261, -0.00000109, -0.00000044, -0.00000407, -0.00000207, -0.00000478, -0.00000383, 0.00000937, },
		{ -0.00000454, -0.00000618, 0.00000490, 0.00000051, 0.00000028, -0.00000922, 0.00000716, -0.00000016, -0.00000511, 0.00000132, -0.00000131, -0.00000266, -0.00000153, 0.00001094, -0.00001049, 0.00000087, 0.00000250, 0.00000625, -0.00000770, -0.00000260, 0.00000465, 0.00000732, -0.00000712, 0.00000049, -0.00000125, -0.00000412, 0.00000241, -0.00000387, 0.00000431, -0.00000213, 0.00000039, 0.00000208, 0.00000028, -0.00000306, 0.00000210, -0.00000565, 0.00000580, 0.00000897, -0.00001376, 0.00001902, -0.00001872, 0.00000023, 0.00000264, -0.00000738, 0.00001069, -0.00001178, 0.00000209, -0.00001982, 0.00001548, -0.00001083, 0.00000798, -0.00001063, 0.00001186, -0.00002307, 0.00000990, -0.00001651, 0.00001434, -0.00002444, 0.00001385, -0.00001908, 0.00002164, -0.00002716, 0.00000687, -0.00002944, 0.00001685, -0.00004310, 0.00001498, -0.00004777, 0.00002542, -0.00008266, -0.00001086, -0.00012508, -0.00007186, -0.00065897, 0.00080909, 0.00033373, 0.00087041, -0.00022879, -0.00041483, 0.00002123, -0.00014079, 0.00004431, -0.00007255, 0.00002908, -0.00007475, 0.00003090, -0.00002864, 0.00001545, -0.00003170, 0.00002487, -0.00002882, 0.00001929, -0.00003147, 0.00002264, -0.00001742, 0.00000870, -0.00001541, 0.00001304, -0.00001333, 0.00001725, -0.00002665, 0.00000865, -0.00000415, 0.00000611, -0.00000251, 0.00000556, 0.00000745, -0.00001494, 0.00000713, -0.00001026, 0.00000279, 0.00000242, -0.00000916, 0.00000223, -0.00000481, -0.00000008, -0.00000242, 0.00000514, -0.00000259, 0.00000349, -0.00000623, 0.00000376, -0.00000641, -0.00000063, -0.00000534, -0.00000365, 0.00000541, 0.00000556, },
		{ -0.00000391, -0.00000077, 0.00000558, -0.00000284, -0.00000394, -0.00000204, 0.00000915, -0.00001111, 0.00000541, -0.00000527, 0.00000275, -0.00000839, 0.00001456, -0.00000877, -0.00000146, 0.00000120, 0.00000717, -0.00000468, -0.00000551, 0.00000332, 0.00000759, -0.00000284, -0.00000428, 0.00000316, -0.00000773, 0.00000456, -0.00000547, 0.00000518, -0.00000335, 0.00000182, -0.00000059, 0.00000432, -0.00000666, 0.00000483, -0.00000667, 0.00000413, 0.00000815, -0.00000758, 0.00000664, 0.00000171, -0.00002190, 0.00002125, -0.00002351, 0.00002539, -0.00002534, 0.00001737, -0.00003307, 0.00002409, -0.00001997, 0.00001927, -0.00002327, 0.00002723, -0.00003782, 0.00002293, -0.00003055, 0.00002938, -0.00003855, 0.00002746, -0.00003439, 0.00003971, -0.00004690, 0.00003003, -0.00005439, 0.00004418, -0.00007174, 0.00004607, -0.00008309, 0.00007123, -0.00013444, 0.00005879, -0.00020804, 0.00008931, -0.00078131, 0.00053376, 0.00061890, 0.00085381, -0.00013979, -0.00051940, 0.00007906, -0.00019622, 0.00008507, -0.00010738, 0.00005763, -0.00010130, 0.00005379, -0.00005167, 0.00003717, -0.00005332, 0.00004327, -0.00004333, 0.00003147, -0.00004346, 0.00003327, -0.00002565, 0.00001661, -0.00002502, 0.00002169, -0.00002155, 0.00002312, -0.00002755, 0.00000868, -0.00000688, 0.00000816, -0.00000487, 0.00000655, 0.00000563, -0.00000768, -0.00000545, 0.00000614, -0.00001393, 0.00001670, -0.00001908, 0.00001038, -0.00001219, 0.00000723, -0.00001127, 0.00001368, -0.00000969, 0.00000993, -0.00001132, 0.00000779, -0.00000891, 0.00000177, -0.00000774, -0.00000046, -0.00000128, 0.00001110, -0.00000559, },
		{ -0.00000688, 0.00001004, -0.00000589, 0.00000187, -0.00001029, 0.00001506, -0.00001285, 0.00000594, -0.00000637, 0.00000501, -0.00001049, 0.00001226, -0.00000258, -0.00000579, 0.00000167, 0.00000522, 0.00000046, -0.00000875, 0.00000282, 0.00000540, 0.00000325, -0.00000906, 0.00000652, -0.00000977, 0.00000522, -0.00000605, 0.00000565, -0.00000474, 0.00000378, -0.00000391, 0.00000846, -0.00000980, 0.00000716, -0.00000816, 0.00000534, 0.00000190, 0.00000460, -0.00001033, 0.00002255, -0.00003908, 0.00003188, -0.00003090, 0.00003124, -0.00003011, 0.00002302, -0.00003555, 0.00002308, -0.00002137, 0.00002321, -0.00002818, 0.00003361, -0.00004208, 0.00002529, -0.00003402, 0.00003323, -0.00004112, 0.00003000, -0.00003909, 0.00004666, -0.00005441, 0.00003926, -0.00006333, 0.00005353, -0.00008038, 0.00005546, -0.00009476, 0.00008868, -0.00015068, 0.00008163, -0.00022695, 0.00013546, -0.00069760, 0.00008751, 0.00089698, 0.00091269, -0.00008728, -0.00059864, 0.00011172, -0.00022955, 0.00010500, -0.00012424, 0.00007115, -0.00011501, 0.00006531, -0.00006410, 0.00004836, -0.00006405, 0.00005047, -0.00004762, 0.00003516, -0.00004807, 0.00003695, -0.00002795, 0.00002000, -0.00003035, 0.00002573, -0.00002495, 0.00002424, -0.00002559, 0.00000873, -0.00001076, 0.00001176, -0.00000962, 0.00001138, -0.00000208, 0.00000608, -0.00002276, 0.00002367, -0.00002822, 0.00002637, -0.00002426, 0.00001484, -0.00001673, 0.00001231, -0.00001762, 0.00001888, -0.00001318, 0.00001284, -0.00001312, 0.00000907, -0.00000959, 0.00000337, -0.00001017, 0.00000334, -0.00000764, 0.00001479, -0.00000575, -0.00000077, },
		{ 0.00001093, -0.00000599, 0.00000479, -0.00001370, 0.00001438, -0.00000849, 0.00000189, -0.00000365, 0.00000312, -0.00000741, 0.00000508, 0.00000589, -0.00001010, 0.00000252, 0.00000203, 0.00000554, -0.00001052, 0.00000190, 0.00000254, 0.00000844, -0.00001110, 0.00000669, -0.00000873, 0.00000340, -0.00000486, 0.00000488, -0.00000512, 0.00000468, -0.00000581, 0.00001038, -0.00001026, 0.00000713, -0.00000822, 0.00000677, -0.00000510, 0.00001546, -0.00002223, 0.00003308, -0.00004164, 0.00002706, -0.00002395, 0.00002371, -0.00002206, 0.00001553, -0.00002479, 0.00001082, -0.00001310, 0.00001700, -0.00002189, 0.00002729, -0.00003251, 0.00001410, -0.00002406, 0.00002298, -0.00002932, 0.00001849, -0.00002941, 0.00003771, -0.00004403, 0.00002825, -0.00004967, 0.00003813, -0.00006206, 0.00003561, -0.00007385, 0.00006716, -0.00011943, 0.00004498, -0.00016931, 0.00006565, -0.00046046, -0.00039943, 0.00102281, 0.00113721, -0.00011580, -0.00062251, 0.00009547, -0.00022180, 0.00008891, -0.00011091, 0.00005912, -0.00010624, 0.00005635, -0.00005727, 0.00004103, -0.00005692, 0.00004088, -0.00003740, 0.00002673, -0.00004183, 0.00003050, -0.00002146, 0.00001592, -0.00002823, 0.00002216, -0.00002115, 0.00001916, -0.00001994, 0.00000749, -0.00001343, 0.00001392, -0.00001335, 0.00001590, -0.00001002, 0.00001857, -0.00003530, 0.00003253, -0.00003153, 0.00002486, -0.00001992, 0.00001178, -0.00001484, 0.00001148, -0.00001772, 0.00001737, -0.00001033, 0.00001007, -0.00000997, 0.00000634, -0.00000730, 0.00000282, -0.00001075, 0.00000517, -0.00001070, 0.00001422, -0.00000322, 0.00000094, -0.00001040, },
		{ -0.00000240, 0.00000369, -0.00001133, 0.00000741, 0.00000037, -0.00000476, 0.00000112, -0.00000120, -0.00000158, -0.00000347, 0.00001283, -0.00001160, 0.00000183, -0.00000030, 0.00000833, -0.00000934, -0.00000044, 0.00000062, 0.00001068, -0.00000899, 0.00000319, -0.00000480, -0.00000040, -0.00000232, 0.00000297, -0.00000419, 0.00000379, -0.00000527, 0.00000902, -0.00000725, 0.00000411, -0.00000592, 0.00000633, -0.00000884, 0.00001942, -0.00002307, 0.00002863, -0.00002782, 0.00000803, -0.00000568, 0.00000659, -0.00000535, -0.00000047, -0.00000630, -0.00000690, -0.00000015, 0.00000482, -0.00000836, 0.00001260, -0.00001428, -0.00000484, -0.00000657, 0.00000471, -0.00000955, -0.00000075, -0.00001126, 0.00001853, -0.00002167, 0.00000365, -0.00002113, 0.00000695, -0.00002703, -0.00000205, -0.00003290, 0.00002113, -0.00005893, -0.00002659, -0.00007030, -0.00005942, -0.00018756, -0.00076042, 0.00086484, 0.00157246, -0.00022375, -0.00059789, 0.00003658, -0.00017953, 0.00004360, -0.00007414, 0.00002740, -0.00007981, 0.00003095, -0.00003480, 0.00001880, -0.00003595, 0.00001910, -0.00001747, 0.00001047, -0.00002832, 0.00001711, -0.00000902, 0.00000648, -0.00002021, 0.00001257, -0.00001203, 0.00000996, -0.00001228, 0.00000519, -0.00001368, 0.00001269, -0.00001348, 0.00001658, -0.00001320, 0.00002345, -0.00003651, 0.00002758, -0.00002111, 0.00001168, -0.00000692, 0.00000227, -0.00000726, 0.00000510, -0.00001183, 0.00000978, -0.00000226, 0.00000297, -0.00000331, 0.00000089, -0.00000293, 0.00000038, -0.00000908, 0.00000403, -0.00000942, 0.00000909, 0.00000199, 0.00000047, -0.00001007, 0.00000732, },
		{ 0.00000043, -0.00000556, -0.00000215, 0.00000932, -0.00000993, 0.00000452, -0.00000469, 0.00000327, -0.00000926, 0.00001488, -0.00000848, -0.00000135, -0.00000066, 0.00000795, -0.00000523, -0.00000388, -0.00000012, 0.00000982, -0.00000365, -0.00000211, -0.00000019, -0.00000404, -0.00000023, 0.00000130, -0.00000300, 0.00000203, -0.00000324, 0.00000566, -0.00000247, -0.00000008, -0.00000267, 0.00000449, -0.00000885, 0.00001610, -0.00001410, 0.00001339, -0.00000539, -0.00001493, 0.00001288, -0.00000919, 0.00000924, -0.00001403, 0.00000863, -0.00001915, 0.00000762, -0.00000392, 0.00000261, 0.00000021, 0.00000092, -0.00001930, 0.00000606, -0.00000885, 0.00000499, -0.00001434, 0.00000185, 0.00000331, -0.00000285, -0.00001725, 0.00000294, -0.00001851, 0.00000103, -0.00003145, -0.00000084, -0.00001600, -0.00000939, -0.00008314, 0.00000575, -0.00014900, -0.00001058, -0.00085900, 0.00036891, 0.00217403, -0.00033902, -0.00058136, -0.00002109, -0.00013958, 0.00000093, -0.00004161, -0.00000024, -0.00005636, 0.00000754, -0.00001363, -0.00000231, -0.00001650, -0.00000018, -0.00000137, -0.00000172, -0.00001802, 0.00000629, 0.00000079, -0.00000078, -0.00001314, 0.00000361, -0.00000407, 0.00000266, -0.00000743, 0.00000480, -0.00001322, 0.00000937, -0.00001094, 0.00001366, -0.00001113, 0.00002013, -0.00002707, 0.00001200, -0.00000264, -0.00000602, 0.00000749, -0.00000730, 0.00000061, -0.00000206, -0.00000457, 0.00000096, 0.00000612, -0.00000377, 0.00000260, -0.00000359, 0.00000056, -0.00000176, -0.00000676, 0.00000128, -0.00000550, 0.00000185, 0.00000749, -0.00000067, -0.00000718, 0.00000108, 0.00000283, },
		{ -0.00000059, -0.00000926, 0.00001346, -0.00000984, 0.00000366, -0.00000468, 0.00000433, -0.00000983, 0.00001101, -0.00000133, -0.00000601, 0.00000025, 0.00000544, -0.00000004, -0.00000661, -0.00000095, 0.00000761, 0.00000200, -0.00000578, 0.00000181, -0.00000463, -0.00000099, 0.00000198, -0.00000357, 0.00000160, -0.00000232, 0.00000346, 0.00000042, -0.00000171, -0.00000192, 0.00000432, -0.00000838, 0.00001036, -0.00000290, -0.00000179, 0.00001217, -0.00002762, 0.00001839, -0.00001145, 0.00001028, -0.00001397, 0.00000895, -0.00001574, 0.00000099, 0.00000005, 0.00000091, 0.00000124, 0.00000113, -0.00001707, 0.00000166, -0.00000516, 0.00000130, -0.00000897, -0.00000391, 0.00000710, -0.00000440, -0.00001569, 0.00000181, -0.00001562, -0.00000246, -0.00002587, -0.00000724, -0.00001022, -0.00001093, -0.00007679, 0.00000072, -0.00012941, -0.00001573, -0.00064929, -0.00040038, 0.00279531, -0.00032626, -0.00066495, -0.00001081, -0.00015536, 0.00000525, -0.00005068, 0.00000783, -0.00006345, 0.00001086, -0.00001652, -0.00000110, -0.00001836, 0.00000127, -0.00000534, 0.00000433, -0.00002352, 0.00000966, -0.00000271, 0.00000377, -0.00001613, 0.00000427, -0.00000583, 0.00000494, -0.00001161, 0.00001107, -0.00001635, 0.00000855, -0.00001044, 0.00001175, -0.00000853, 0.00001433, -0.00001488, -0.00000385, 0.00001207, -0.00001661, 0.00001301, -0.00000842, 0.00000151, -0.00000337, -0.00000243, -0.00000258, 0.00000856, -0.00000454, 0.00000286, -0.00000291, -0.00000031, -0.00000063, -0.00000657, -0.00000013, -0.00000251, -0.00000337, 0.00000961, 0.00000034, -0.00000474, -0.00000409, 0.00000599, -0.00000146, },
		{ -0.00001031, 0.00001013, -0.00000329, -0.00000183, -0.00000089, 0.00000138, -0.00000567, 0.00000318, 0.00000686, -0.00000945, 0.00000036, 0.00000313, 0.00000345, -0.00000633, -0.00000369, 0.00000651, 0.00000468, -0.00000464, -0.00000133, -0.00000030, -0.00000609, 0.00000639, -0.00000755, 0.00000458, -0.00000511, 0.00000558, -0.00000206, 0.00000264, -0.00000695, 0.00000925, -0.00001186, 0.00000879, 0.00000135, -0.00000575, 0.00001318, -0.00001985, 0.00000308, 0.00000578, -0.00000726, 0.00000432, -0.00000943, 0.00000647, -0.00002256, 0.00001979, -0.00001757, 0.00002054, -0.00001876, 0.00000663, -0.00002429, 0.00002049, -0.00002558, 0.00002040, -0.00003407, 0.00003653, -0.00003420, 0.00001728, -0.00003425, 0.00002581, -0.00004808, 0.00002582, -0.00006426, 0.00005245, -0.00007963, 0.00001017, -0.00010377, 0.00001668, -0.00020369, -0.00020182, -0.00126387, 0.00322030, -0.00003134, -0.00093480, 0.00012259, -0.00026809, 0.00008895, -0.00012734, 0.00007295, -0.00011951, 0.00005757, -0.00005877, 0.00003657, -0.00005440, 0.00003475, -0.00003893, 0.00003662, -0.00005206, 0.00003414, -0.00002596, 0.00002616, -0.00003525, 0.00002072, -0.00002309, 0.00002180, -0.00002887, 0.00002748, -0.00002714, 0.00001535, -0.00001765, 0.00001694, -0.00001227, 0.00001446, -0.00001037, -0.00000817, 0.00001157, -0.00001040, 0.00000233, 0.00000431, -0.00000898, 0.00000539, -0.00000971, 0.00000346, 0.00000121, 0.00000382, -0.00000512, 0.00000501, -0.00000731, 0.00000549, -0.00001057, 0.00000240, -0.00000370, -0.00000308, 0.00000557, 0.00000568, -0.00000542, -0.00000492, 0.00000403, 0.00000088, 0.00000025, },
		{ 0.00000054, 0.00000729, -0.00000929, 0.00000434, -0.00000330, 0.00000020, -0.00000463, 0.00001210, -0.00000888, -0.00000224, 0.00000306, 0.00000322, -0.00000182, -0.00000912, 0.00000791, 0.00000267, 0.00000224, -0.00000956, 0.00000837, -0.00001474, 0.00001394, -0.00001473, 0.00001121, -0.00001226, 0.00001294, -0.00001081, 0.00001357, -0.00001823, 0.00002012, -0.00002131, 0.00001510, -0.00000621, 0.00000625, -0.00000529, 0.00000812, -0.00002981, 0.00003765, -0.00003795, 0.00003510, -0.00004034, 0.00004053, -0.00005597, 0.00004919, -0.00004747, 0.00005286, -0.00005304, 0.00004525, -0.00006463, 0.00006084, -0.00006817, 0.00006600, -0.00008090, 0.00008420, -0.00008491, 0.00007378, -0.00009630, 0.00009542, -0.00012387, 0.00010999, -0.00015656, 0.00015485, -0.00019517, 0.00015185, -0.00027276, 0.00023887, -0.00048842, 0.00031922, -0.00198150, 0.00323575, 0.00065039, -0.00141806, 0.00038324, -0.00047633, 0.00024791, -0.00026605, 0.00018908, -0.00021911, 0.00014301, -0.00013631, 0.00010684, -0.00012056, 0.00009580, -0.00009717, 0.00009014, -0.00009925, 0.00007611, -0.00006582, 0.00006370, -0.00006854, 0.00005151, -0.00005435, 0.00005147, -0.00005726, 0.00005266, -0.00004582, 0.00003143, -0.00003489, 0.00003213, -0.00002619, 0.00002572, -0.00001986, 0.00000508, -0.00000821, 0.00001396, -0.00002365, 0.00002879, -0.00002879, 0.00002268, -0.00002531, 0.00001811, -0.00001460, 0.00001958, -0.00001944, 0.00001834, -0.00001887, 0.00001555, -0.00001839, 0.00000920, -0.00000988, 0.00000339, -0.00000448, 0.00001493, -0.00000959, -0.00000063, -0.00000337, 0.00000758, -0.00000354, -0.00000512, },
		{ 0.00001695, -0.00001442, 0.00000751, -0.00000639, 0.00000416, -0.00000867, 0.00001167, -0.00000325, -0.00000773, 0.00000550, -0.00000063, 0.00000583, -0.00001592, 0.00001101, -0.00000296, 0.00001260, -0.00001974, 0.00001796, -0.00002366, 0.00002170, -0.00002256, 0.00001917, -0.00002147, 0.00002304, -0.00002285, 0.00002756, -0.00003203, 0.00003350, -0.00003399, 0.00002712, -0.00002281, 0.00002934, -0.00003508, 0.00004455, -0.00006633, 0.00006970, -0.00006755, 0.00006434, -0.00006952, 0.00007180, -0.00008500, 0.00007494, -0.00007573, 0.00008447, -0.00008689, 0.00008277, -0.00010268, 0.00009885, -0.00010881, 0.00010936, -0.00012537, 0.00013050, -0.00013559, 0.00013082, -0.00015883, 0.00016482, -0.00019841, 0.00019193, -0.00024573, 0.00025387, -0.00030754, 0.00028818, -0.00043184, 0.00044130, -0.00073565, 0.00071926, -0.00233699, 0.00271600, 0.00171016, -0.00204362, 0.00070119, -0.00072130, 0.00043155, -0.00042260, 0.00031753, -0.00032859, 0.00023748, -0.00022235, 0.00018502, -0.00019370, 0.00016247, -0.00015941, 0.00014593, -0.00014822, 0.00012048, -0.00010842, 0.00010367, -0.00010458, 0.00008604, -0.00008930, 0.00008400, -0.00008760, 0.00007890, -0.00006661, 0.00005221, -0.00005797, 0.00005352, -0.00004719, 0.00004566, -0.00004077, 0.00003167, -0.00004036, 0.00004722, -0.00005462, 0.00005506, -0.00004922, 0.00004100, -0.00004252, 0.00003490, -0.00003230, 0.00003627, -0.00003405, 0.00003160, -0.00003022, 0.00002562, -0.00002688, 0.00001763, -0.00001858, 0.00001314, -0.00001694, 0.00002472, -0.00001483, 0.00000649, -0.00001333, 0.00001568, -0.00000925, 0.00000306, -0.00001086, },
		{ -0.00001372, 0.00000614, -0.00000576, 0.00000410, -0.00000744, 0.00000571, 0.00000565, -0.00001393, 0.00000848, -0.00000568, 0.00001342, -0.00002094, 0.00001310, -0.00000884, 0.00002187, -0.00002684, 0.00002363, -0.00002838, 0.00002557, -0.00002706, 0.00002433, -0.00002818, 0.00003075, -0.00003230, 0.00003805, -0.00004159, 0.00004270, -0.00004325, 0.00003756, -0.00003916, 0.00005125, -0.00006067, 0.00007158, -0.00008822, 0.00008478, -0.00008006, 0.00007657, -0.00008155, 0.00008493, -0.00009498, 0.00008304, -0.00008785, 0.00009963, -0.00010323, 0.00010102, -0.00011969, 0.00011537, -0.00012748, 0.00012946, -0.00014560, 0.00015224, -0.00016099, 0.00016047, -0.00019099, 0.00020019, -0.00023480, 0.00023139, -0.00028758, 0.00030022, -0.00036010, 0.00035268, -0.00050185, 0.00052767, -0.00082321, 0.00085046, -0.00221910, 0.00168465, 0.00299773, -0.00264062, 0.00094276, -0.00090033, 0.00055649, -0.00052720, 0.00039900, -0.00039695, 0.00029612, -0.00027647, 0.00023413, -0.00023962, 0.00020314, -0.00019662, 0.00017783, -0.00017554, 0.00014589, -0.00013386, 0.00012760, -0.00012626, 0.00010802, -0.00011212, 0.00010437, -0.00010610, 0.00009412, -0.00007909, 0.00006777, -0.00007672, 0.00007100, -0.00006534, 0.00006422, -0.00006197, 0.00005834, -0.00006940, 0.00007305, -0.00007506, 0.00006951, -0.00005881, 0.00005040, -0.00005211, 0.00004487, -0.00004307, 0.00004553, -0.00004138, 0.00003808, -0.00003548, 0.00003056, -0.00003144, 0.00002325, -0.00002516, 0.00002097, -0.00002623, 0.00003013, -0.00001713, 0.00001229, -0.00002114, 0.00002067, -0.00001272, 0.00000952, -0.00001870, 0.00002083, },
		{ 0.00000050, -0.00000180, 0.00000063, -0.00000249, -0.00000269, 0.00001381, -0.00001722, 0.00000894, -0.00000853, 0.00001712, -0.00002083, 0.00001116, -0.00001114, 0.00002538, -0.00002636, 0.00002151, -0.00002552, 0.00002235, -0.00002487, 0.00002282, -0.00002793, 0.00003087, -0.00003322, 0.00003865, -0.00004043, 0.00004117, -0.00004210, 0.00003811, -0.00004465, 0.00005886, -0.00006724, 0.00007428, -0.00008232, 0.00007207, -0.00006619, 0.00006312, -0.00006795, 0.00007165, -0.00007830, 0.00006604, -0.00007534, 0.00008846, -0.00009115, 0.00008851, -0.00010413, 0.00009863, -0.00011167, 0.00011299, -0.00012749, 0.00013411, -0.00014390, 0.00014342, -0.00017162, 0.00017869, -0.00020867, 0.00020229, -0.00025387, 0.00026261, -0.00031729, 0.00030482, -0.00043699, 0.00044677, -0.00069497, 0.00066884, -0.00166490, 0.00032381, 0.00424815, -0.00298370, 0.00095668, -0.00090733, 0.00054125, -0.00051445, 0.00037962, -0.00037873, 0.00027915, -0.00026301, 0.00022173, -0.00022871, 0.00019095, -0.00018468, 0.00016441, -0.00016207, 0.00013458, -0.00012529, 0.00011960, -0.00011857, 0.00010280, -0.00010852, 0.00009913, -0.00010048, 0.00008734, -0.00007296, 0.00006731, -0.00007953, 0.00007267, -0.00006858, 0.00006874, -0.00006944, 0.00006934, -0.00007884, 0.00007613, -0.00007211, 0.00006203, -0.00004958, 0.00004388, -0.00004731, 0.00004126, -0.00004033, 0.00004141, -0.00003630, 0.00003331, -0.00003073, 0.00002678, -0.00002848, 0.00002217, -0.00002524, 0.00002203, -0.00002754, 0.00002732, -0.00001336, 0.00001317, -0.00002283, 0.00001907, -0.00001080, 0.00001076, -0.00001952, 0.00001675, -0.00000662, },
		{ 0.00000280, -0.00000351, 0.00000270, -0.00000918, 0.00001704, -0.00001461, 0.00000470, -0.00000680, 0.00001462, -0.00001407, 0.00000393, -0.00000788, 0.00002089, -0.00001685, 0.00001097, -0.00001479, 0.00001171, -0.00001523, 0.00001336, -0.00001882, 0.00002102, -0.00002292, 0.00002666, -0.00002607, 0.00002641, -0.00002750, 0.00002453, -0.00003345, 0.00004545, -0.00004857, 0.00004852, -0.00004770, 0.00003330, -0.00002885, 0.00002710, -0.00003178, 0.00003526, -0.00003870, 0.00002736, -0.00004035, 0.00005212, -0.00005152, 0.00004636, -0.00005763, 0.00005045, -0.00006288, 0.00006125, -0.00007224, 0.00007679, -0.00008416, 0.00007908, -0.00010024, 0.00010041, -0.00012090, 0.00010632, -0.00014702, 0.00014407, -0.00018281, 0.00015022, -0.00024709, 0.00021624, -0.00038365, 0.00024722, -0.00084600, -0.00107708, 0.00515940, -0.00287256, 0.00064022, -0.00068448, 0.00034744, -0.00035709, 0.00023916, -0.00025805, 0.00017304, -0.00016995, 0.00013716, -0.00015165, 0.00011799, -0.00011714, 0.00010048, -0.00010323, 0.00008191, -0.00007782, 0.00007483, -0.00007666, 0.00006522, -0.00007325, 0.00006352, -0.00006659, 0.00005476, -0.00004399, 0.00004521, -0.00005971, 0.00005152, -0.00004957, 0.00005118, -0.00005415, 0.00005506, -0.00006004, 0.00005041, -0.00004275, 0.00003195, -0.00002196, 0.00002172, -0.00002784, 0.00002338, -0.00002337, 0.00002363, -0.00001894, 0.00001760, -0.00001622, 0.00001429, -0.00001753, 0.00001331, -0.00001721, 0.00001444, -0.00001942, 0.00001580, -0.00000329, 0.00000824, -0.00001731, 0.00001043, -0.00000334, 0.00000640, -0.00001344, 0.00000637, 0.00000400, -0.00000640, },
		{ -0.00000532, 0.00000484, -0.00001052, 0.00001330, -0.00000576, -0.00000384, -0.00000082, 0.00000663, -0.00000223, -0.00000692, -0.00000019, 0.00000984, -0.00000092, -0.00000466, 0.00000042, -0.00000325, -0.00000083, -0.00000177, -0.00000296, 0.00000336, -0.00000376, 0.00000495, -0.00000196, 0.00000204, -0.00000279, -0.00000040, -0.00000806, 0.00001451, -0.00001049, 0.00000349, 0.00000319, -0.00001736, 0.00001776, -0.00001780, 0.00001359, -0.00001082, 0.00001024, -0.00002003, 0.00000516, 0.00000238, 0.00000321, -0.00001193, 0.00000525, -0.00001407, 0.00000354, -0.00000981, 0.00000350, -0.00000258, 0.00000043, -0.00001335, 0.00000177, -0.00001068, 0.00000169, -0.00002663, -0.00000025, -0.00001833, 0.00000118, -0.00006125, 0.00000656, -0.00008593, 0.00000653, -0.00026030, -0.00000354, -0.00221498, 0.00549325, -0.00221498, -0.00000354, -0.00026030, 0.00000653, -0.00008593, 0.00000656, -0.00006125, 0.00000118, -0.00001833, -0.00000025, -0.00002663, 0.00000169, -0.00001068, 0.00000177, -0.00001335, 0.00000043, -0.00000258, 0.00000350, -0.00000981, 0.00000354, -0.00001407, 0.00000525, -0.00001193, 0.00000321, 0.00000238, 0.00000516, -0.00002003, 0.00001024, -0.00001082, 0.00001359, -0.00001780, 0.00001776, -0.00001736, 0.00000319, 0.00000349, -0.00001049, 0.00001451, -0.00000806, -0.00000040, -0.00000279, 0.00000204, -0.00000196, 0.00000495, -0.00000376, 0.00000336, -0.00000296, -0.00000177, -0.00000083, -0.00000325, 0.00000042, -0.00000466, -0.00000092, 0.00000984, -0.00000019, -0.00000692, -0.00000223, 0.00000663, -0.00000082, -0.00000384, -0.00000576, 0.00001330, -0.00001052, 0.00000484, },
		{ 0.00000280, -0.00000640, 0.00000400, 0.00000637, -0.00001344, 0.00000640, -0.00000334, 0.00001043, -0.00001731, 0.00000824, -0.00000329, 0.00001580, -0.00001942, 0.00001444, -0.00001721, 0.00001331, -0.00001753, 0.00001429, -0.00001622, 0.00001760, -0.00001894, 0.00002363, -0.00002337, 0.00002338, -0.00002784, 0.00002172, -0.00002196, 0.00003195, -0.00004275, 0.00005041, -0.00006004, 0.00005506, -0.00005415, 0.00005118, -0.00004957, 0.00005152, -0.00005971, 0.00004521, -0.00004399, 0.00005476, -0.00006659, 0.00006352, -0.00007325, 0.00006522, -0.00007666, 0.00007483, -0.00007782, 0.00008191, -0.00010323, 0.00010048, -0.00011714, 0.00011799, -0.00015165, 0.00013716, -0.00016995, 0.00017304, -0.00025805, 0.00023916, -0.00035709, 0.00034744, -0.00068448, 0.00064022, -0.00287256, 0.00515940, -0.00107708, -0.00084600, 0.00024722, -0.00038365, 0.00021624, -0.00024709, 0.00015022, -0.00018281, 0.00014407, -0.00014702, 0.00010632, -0.00012090, 0.00010041, -0.00010024, 0.00007908, -0.00008416, 0.00007679, -0.00007224, 0.00006125, -0.00006288, 0.00005045, -0.00005763, 0.00004636, -0.00005152, 0.00005212, -0.00004035, 0.00002736, -0.00003870, 0.00003526, -0.00003178, 0.00002710, -0.00002885, 0.00003330, -0.00004770, 0.00004852, -0.00004857, 0.00004545, -0.00003345, 0.00002453, -0.00002750, 0.00002641, -0.00002607, 0.00002666, -0.00002292, 0.00002102, -0.00001882, 0.00001336, -0.00001523, 0.00001171, -0.00001479, 0.00001097, -0.00001685, 0.00002089, -0.00000788, 0.00000393, -0.00001407, 0.00001462, -0.00000680, 0.00000470, -0.00001461, 0.00001704, -0.00000918, 0.00000270, -0.00000351, },
		{ 0.00000050, -0.00000662, 0.00001675, -0.00001952, 0.00001076, -0.00001080, 0.00001907, -0.00002283, 0.00001317, -0.00001336, 0.00002732, -0.00002754, 0.00002203, -0.00002524, 0.00002217, -0.00002848, 0.00002678, -0.00003073, 0.00003331, -0.00003630, 0.00004141, -0.00004033, 0.00004126, -0.00004731, 0.00004388, -0.00004958, 0.00006203, -0.00007211, 0.00007613, -0.00007884, 0.00006934, -0.00006944, 0.00006874, -0.00006858, 0.00007267, -0.00007953, 0.00006731, -0.00007296, 0.00008734, -0.00010048, 0.00009913, -0.00010852, 0.00010280, -0.00011857, 0.00011960, -0.00012529, 0.00013458, -0.00016207, 0.00016441, -0.00018468, 0.00019095, -0.00022871, 0.00022173, -0.00026301, 0.00027915, -0.00037873, 0.00037962, -0.00051445, 0.00054125, -0.00090733, 0.00095668, -0.00298370, 0.00424815, 0.00032381, -0.00166490, 0.00066884, -0.00069497, 0.00044677, -0.00043699, 0.00030482, -0.00031729, 0.00026261, -0.00025387, 0.00020229, -0.00020867, 0.00017869, -0.00017162, 0.00014342, -0.00014390, 0.00013411, -0.00012749, 0.00011299, -0.00011167, 0.00009863, -0.00010413, 0.00008851, -0.00009115, 0.00008846, -0.00007534, 0.00006604, -0.00007830, 0.00007165, -0.00006795, 0.00006312, -0.00006619, 0.00007207, -0.00008232, 0.00007428, -0.00006724, 0.00005886, -0.00004465, 0.00003811, -0.00004210, 0.00004117, -0.00004043, 0.00003865, -0.00003322, 0.00003087, -0.00002793, 0.00002282, -0.00002487, 0.00002235, -0.00002552, 0.00002151, -0.00002636, 0.00002538, -0.00001114, 0.00001116, -0.00002083, 0.00001712, -0.00000853, 0.00000894, -0.00001722, 0.00001381, -0.00000269, -0.00000249, 0.00000063, -0.00000180, },
		{ -0.00001372, 0.00002083, -0.00001870, 0.00000952, -0.00001272, 0.00002067, -0.00002114, 0.00001229, -0.00001713, 0.00003013, -0.00002623, 0.00002097, -0.00002516, 0.00002325, -0.00003144, 0.00003056, -0.00003548, 0.00003808, -0.00004138, 0.00004553, -0.00004307, 0.00004487, -0.00005211, 0.00005040, -0.00005881, 0.00006951, -0.00007506, 0.00007305, -0.00006940, 0.00005834, -0.00006197, 0.00006422, -0.00006534, 0.00007100, -0.00007672, 0.00006777, -0.00007909, 0.00009412, -0.00010610, 0.00010437, -0.00011212, 0.00010802, -0.00012626, 0.00012760, -0.00013386, 0.00014589, -0.00017554, 0.00017783, -0.00019662, 0.00020314, -0.00023962, 0.00023413, -0.00027647, 0.00029612, -0.00039695, 0.00039900, -0.00052720, 0.00055649, -0.00090033, 0.00094276, -0.00264062, 0.00299773, 0.00168465, -0.00221910, 0.00085046, -0.00082321, 0.00052767, -0.00050185, 0.00035268, -0.00036010, 0.00030022, -0.00028758, 0.00023139, -0.00023480, 0.00020019, -0.00019099, 0.00016047, -0.00016099, 0.00015224, -0.00014560, 0.00012946, -0.00012748, 0.00011537, -0.00011969, 0.00010102, -0.00010323, 0.00009963, -0.00008785, 0.00008304, -0.00009498, 0.00008493, -0.00008155, 0.00007657, -0.00008006, 0.00008478, -0.00008822, 0.00007158, -0.00006067, 0.00005125, -0.00003916, 0.00003756, -0.00004325, 0.00004270, -0.00004159, 0.00003805, -0.00003230, 0.00003075, -0.00002818, 0.00002433, -0.00002706, 0.00002557, -0.00002838, 0.00002363, -0.00002684, 0.00002187, -0.00000884, 0.00001310, -0.00002094, 0.00001342, -0.00000568, 0.00000848, -0.00001393, 0.00000565, 0.00000571, -0.00000744, 0.00000410, -0.00000576, 0.00000614, },
		{ 0.00001695, -0.00001086, 0.00000306, -0.00000925, 0.00001568, -0.00001333, 0.00000649, -0.00001483, 0.00002472, -0.00001694, 0.00001314, -0.00001858, 0.00001763, -0.00002688, 0.00002562, -0.00003022, 0.00003160, -0.00003405, 0.00003627, -0.00003230, 0.00003490, -0.00004252, 0.00004100, -0.00004922, 0.00005506, -0.00005462, 0.00004722, -0.00004036, 0.00003167, -0.00004077, 0.00004566, -0.00004719, 0.00005352, -0.00005797, 0.00005221, -0.00006661, 0.00007890, -0.00008760, 0.00008400, -0.00008930, 0.00008604, -0.00010458, 0.00010367, -0.00010842, 0.00012048, -0.00014822, 0.00014593, -0.00015941, 0.00016247, -0.00019370, 0.00018502, -0.00022235, 0.00023748, -0.00032859, 0.00031753, -0.00042260, 0.00043155, -0.00072130, 0.00070119, -0.00204362, 0.00171016, 0.00271600, -0.00233699, 0.00071926, -0.00073565, 0.00044130, -0.00043184, 0.00028818, -0.00030754, 0.00025387, -0.00024573, 0.00019193, -0.00019841, 0.00016482, -0.00015883, 0.00013082, -0.00013559, 0.00013050, -0.00012537, 0.00010936, -0.00010881, 0.00009885, -0.00010268, 0.00008277, -0.00008689, 0.00008447, -0.00007573, 0.00007494, -0.00008500, 0.00007180, -0.00006952, 0.00006434, -0.00006755, 0.00006970, -0.00006633, 0.00004455, -0.00003508, 0.00002934, -0.00002281, 0.00002712, -0.00003399, 0.00003350, -0.00003203, 0.00002756, -0.00002285, 0.00002304, -0.00002147, 0.00001917, -0.00002256, 0.00002170, -0.00002366, 0.00001796, -0.00001974, 0.00001260, -0.00000296, 0.00001101, -0.00001592, 0.00000583, -0.00000063, 0.00000550, -0.00000773, -0.00000325, 0.00001167, -0.00000867, 0.00000416, -0.00000639, 0.00000751, -0.00001442, },
		{ 0.00000054, -0.00000512, -0.00000354, 0.00000758, -0.00000337, -0.00000063, -0.00000959, 0.00001493, -0.00000448, 0.00000339, -0.00000988, 0.00000920, -0.00001839, 0.00001555, -0.00001887, 0.00001834, -0.00001944, 0.00001958, -0.00001460, 0.00001811, -0.00002531, 0.00002268, -0.00002879, 0.00002879, -0.00002365, 0.00001396, -0.00000821, 0.00000508, -0.00001986, 0.00002572, -0.00002619, 0.00003213, -0.00003489, 0.00003143, -0.00004582, 0.00005266, -0.00005726, 0.00005147, -0.00005435, 0.00005151, -0.00006854, 0.00006370, -0.00006582, 0.00007611, -0.00009925, 0.00009014, -0.00009717, 0.00009580, -0.00012056, 0.00010684, -0.00013631, 0.00014301, -0.00021911, 0.00018908, -0.00026605, 0.00024791, -0.00047633, 0.00038324, -0.00141806, 0.00065039, 0.00323575, -0.00198150, 0.00031922, -0.00048842, 0.00023887, -0.00027276, 0.00015185, -0.00019517, 0.00015485, -0.00015656, 0.00010999, -0.00012387, 0.00009542, -0.00009630, 0.00007378, -0.00008491, 0.00008420, -0.00008090, 0.00006600, -0.00006817, 0.00006084, -0.00006463, 0.00004525, -0.00005304, 0.00005286, -0.00004747, 0.00004919, -0.00005597, 0.00004053, -0.00004034, 0.00003510, -0.00003795, 0.00003765, -0.00002981, 0.00000812, -0.00000529, 0.00000625, -0.00000621, 0.00001510, -0.00002131, 0.00002012, -0.00001823, 0.00001357, -0.00001081, 0.00001294, -0.00001226, 0.00001121, -0.00001473, 0.00001394, -0.00001474, 0.00000837, -0.00000956, 0.00000224, 0.00000267, 0.00000791, -0.00000912, -0.00000182, 0.00000322, 0.00000306, -0.00000224, -0.00000888, 0.00001210, -0.00000463, 0.00000020, -0.00000330, 0.00000434, -0.00000929, 0.00000729, },
		{ -0.00001031, 0.00000025, 0.00000088, 0.00000403, -0.00000492, -0.00000542, 0.00000568, 0.00000557, -0.00000308, -0.00000370, 0.00000240, -0.00001057, 0.00000549, -0.00000731, 0.00000501, -0.00000512, 0.00000382, 0.00000121, 0.00000346, -0.00000971, 0.00000539, -0.00000898, 0.00000431, 0.00000233, -0.00001040, 0.00001157, -0.00000817, -0.00001037, 0.00001446, -0.00001227, 0.00001694, -0.00001765, 0.00001535, -0.00002714, 0.00002748, -0.00002887, 0.00002180, -0.00002309, 0.00002072, -0.00003525, 0.00002616, -0.00002596, 0.00003414, -0.00005206, 0.00003662, -0.00003893, 0.00003475, -0.00005440, 0.00003657, -0.00005877, 0.00005757, -0.00011951, 0.00007295, -0.00012734, 0.00008895, -0.00026809, 0.00012259, -0.00093480, -0.00003134, 0.00322030, -0.00126387, -0.00020182, -0.00020369, 0.00001668, -0.00010377, 0.00001017, -0.00007963, 0.00005245, -0.00006426, 0.00002582, -0.00004808, 0.00002581, -0.00003425, 0.00001728, -0.00003420, 0.00003653, -0.00003407, 0.00002040, -0.00002558, 0.00002049, -0.00002429, 0.00000663, -0.00001876, 0.00002054, -0.00001757, 0.00001979, -0.00002256, 0.00000647, -0.00000943, 0.00000432, -0.00000726, 0.00000578, 0.00000308, -0.00001985, 0.00001318, -0.00000575, 0.00000135, 0.00000879, -0.00001186, 0.00000925, -0.00000695, 0.00000264, -0.00000206, 0.00000558, -0.00000511, 0.00000458, -0.00000755, 0.00000639, -0.00000609, -0.00000030, -0.00000133, -0.00000464, 0.00000468, 0.00000651, -0.00000369, -0.00000633, 0.00000345, 0.00000313, 0.00000036, -0.00000945, 0.00000686, 0.00000318, -0.00000567, 0.00000138, -0.00000089, -0.00000183, -0.00000329, 0.00001013, },
		{ -0.00000059, -0.00000146, 0.00000599, -0.00000409, -0.00000474, 0.00000034, 0.00000961, -0.00000337, -0.00000251, -0.00000013, -0.00000657, -0.00000063, -0.00000031, -0.00000291, 0.00000286, -0.00000454, 0.00000856, -0.00000258, -0.00000243, -0.00000337, 0.00000151, -0.00000842, 0.00001301, -0.00001661, 0.00001207, -0.00000385, -0.00001488, 0.00001433, -0.00000853, 0.00001175, -0.00001044, 0.00000855, -0.00001635, 0.00001107, -0.00001161, 0.00000494, -0.00000583, 0.00000427, -0.00001613, 0.00000377, -0.00000271, 0.00000966, -0.00002352, 0.00000433, -0.00000534, 0.00000127, -0.00001836, -0.00000110, -0.00001652, 0.00001086, -0.00006345, 0.00000783, -0.00005068, 0.00000525, -0.00015536, -0.00001081, -0.00066495, -0.00032626, 0.00279531, -0.00040038, -0.00064929, -0.00001573, -0.00012941, 0.00000072, -0.00007679, -0.00001093, -0.00001022, -0.00000724, -0.00002587, -0.00000246, -0.00001562, 0.00000181, -0.00001569, -0.00000440, 0.00000710, -0.00000391, -0.00000897, 0.00000130, -0.00000516, 0.00000166, -0.00001707, 0.00000113, 0.00000124, 0.00000091, 0.00000005, 0.00000099, -0.00001574, 0.00000895, -0.00001397, 0.00001028, -0.00001145, 0.00001839, -0.00002762, 0.00001217, -0.00000179, -0.00000290, 0.00001036, -0.00000838, 0.00000432, -0.00000192, -0.00000171, 0.00000042, 0.00000346, -0.00000232, 0.00000160, -0.00000357, 0.00000198, -0.00000099, -0.00000463, 0.00000181, -0.00000578, 0.00000200, 0.00000761, -0.00000095, -0.00000661, -0.00000004, 0.00000544, 0.00000025, -0.00000601, -0.00000133, 0.00001101, -0.00000983, 0.00000433, -0.00000468, 0.00000366, -0.00000984, 0.00001346, -0.00000926, },
		{ 0.00000043, 0.00000283, 0.00000108, -0.00000718, -0.00000067, 0.00000749, 0.00000185, -0.00000550, 0.00000128, -0.00000676, -0.00000176, 0.00000056, -0.00000359, 0.00000260, -0.00000377, 0.00000612, 0.00000096, -0.00000457, -0.00000206, 0.00000061, -0.00000730, 0.00000749, -0.00000602, -0.00000264, 0.00001200, -0.00002707, 0.00002013, -0.00001113, 0.00001366, -0.00001094, 0.00000937, -0.00001322, 0.00000480, -0.00000743, 0.00000266, -0.00000407, 0.00000361, -0.00001314, -0.00000078, 0.00000079, 0.00000629, -0.00001802, -0.00000172, -0.00000137, -0.00000018, -0.00001650, -0.00000231, -0.00001363, 0.00000754, -0.00005636, -0.00000024, -0.00004161, 0.00000093, -0.00013958, -0.00002109, -0.00058136, -0.00033902, 0.00217403, 0.00036891, -0.00085900, -0.00001058, -0.00014900, 0.00000575, -0.00008314, -0.00000939, -0.00001600, -0.00000084, -0.00003145, 0.00000103, -0.00001851, 0.00000294, -0.00001725, -0.00000285, 0.00000331, 0.00000185, -0.00001434, 0.00000499, -0.00000885, 0.00000606, -0.00001930, 0.00000092, 0.00000021, 0.00000261, -0.00000392, 0.00000762, -0.00001915, 0.00000863, -0.00001403, 0.00000924, -0.00000919, 0.00001288, -0.00001493, -0.00000539, 0.00001339, -0.00001410, 0.00001610, -0.00000885, 0.00000449, -0.00000267, -0.00000008, -0.00000247, 0.00000566, -0.00000324, 0.00000203, -0.00000300, 0.00000130, -0.00000023, -0.00000404, -0.00000019, -0.00000211, -0.00000365, 0.00000982, -0.00000012, -0.00000388, -0.00000523, 0.00000795, -0.00000066, -0.00000135, -0.00000848, 0.00001488, -0.00000926, 0.00000327, -0.00000469, 0.00000452, -0.00000993, 0.00000932, -0.00000215, -0.00000556, },
		{ -0.00000240, 0.00000732, -0.00001007, 0.00000047, 0.00000199, 0.00000909, -0.00000942, 0.00000403, -0.00000908, 0.00000038, -0.00000293, 0.00000089, -0.00000331, 0.00000297, -0.00000226, 0.00000978, -0.00001183, 0.00000510, -0.00000726, 0.00000227, -0.00000692, 0.00001168, -0.00002111, 0.00002758, -0.00003651, 0.00002345, -0.00001320, 0.00001658, -0.00001348, 0.00001269, -0.00001368, 0.00000519, -0.00001228, 0.00000996, -0.00001203, 0.00001257, -0.00002021, 0.00000648, -0.00000902, 0.00001711, -0.00002832, 0.00001047, -0.00001747, 0.00001910, -0.00003595, 0.00001880, -0.00003480, 0.00003095, -0.00007981, 0.00002740, -0.00007414, 0.00004360, -0.00017953, 0.00003658, -0.00059789, -0.00022375, 0.00157246, 0.00086484, -0.00076042, -0.00018756, -0.00005942, -0.00007030, -0.00002659, -0.00005893, 0.00002113, -0.00003290, -0.00000205, -0.00002703, 0.00000695, -0.00002113, 0.00000365, -0.00002167, 0.00001853, -0.00001126, -0.00000075, -0.00000955, 0.00000471, -0.00000657, -0.00000484, -0.00001428, 0.00001260, -0.00000836, 0.00000482, -0.00000015, -0.00000690, -0.00000630, -0.00000047, -0.00000535, 0.00000659, -0.00000568, 0.00000803, -0.00002782, 0.00002863, -0.00002307, 0.00001942, -0.00000884, 0.00000633, -0.00000592, 0.00000411, -0.00000725, 0.00000902, -0.00000527, 0.00000379, -0.00000419, 0.00000297, -0.00000232, -0.00000040, -0.00000480, 0.00000319, -0.00000899, 0.00001068, 0.00000062, -0.00000044, -0.00000934, 0.00000833, -0.00000030, 0.00000183, -0.00001160, 0.00001283, -0.00000347, -0.00000158, -0.00000120, 0.00000112, -0.00000476, 0.00000037, 0.00000741, -0.00001133, 0.00000369, },
		{ 0.00001093, -0.00001040, 0.00000094, -0.00000322, 0.00001422, -0.00001070, 0.00000517, -0.00001075, 0.00000282, -0.00000730, 0.00000634, -0.00000997, 0.00001007, -0.00001033, 0.00001737, -0.00001772, 0.00001148, -0.00001484, 0.00001178, -0.00001992, 0.00002486, -0.00003153, 0.00003253, -0.00003530, 0.00001857, -0.00001002, 0.00001590, -0.00001335, 0.00001392, -0.00001343, 0.00000749, -0.00001994, 0.00001916, -0.00002115, 0.00002216, -0.00002823, 0.00001592, -0.00002146, 0.00003050, -0.00004183, 0.00002673, -0.00003740, 0.00004088, -0.00005692, 0.00004103, -0.00005727, 0.00005635, -0.00010624, 0.00005912, -0.00011091, 0.00008891, -0.00022180, 0.00009547, -0.00062251, -0.00011580, 0.00113721, 0.00102281, -0.00039943, -0.00046046, 0.00006565, -0.00016931, 0.00004498, -0.00011943, 0.00006716, -0.00007385, 0.00003561, -0.00006206, 0.00003813, -0.00004967, 0.00002825, -0.00004403, 0.00003771, -0.00002941, 0.00001849, -0.00002932, 0.00002298, -0.00002406, 0.00001410, -0.00003251, 0.00002729, -0.00002189, 0.00001700, -0.00001310, 0.00001082, -0.00002479, 0.00001553, -0.00002206, 0.00002371, -0.00002395, 0.00002706, -0.00004164, 0.00003308, -0.00002223, 0.00001546, -0.00000510, 0.00000677, -0.00000822, 0.00000713, -0.00001026, 0.00001038, -0.00000581, 0.00000468, -0.00000512, 0.00000488, -0.00000486, 0.00000340, -0.00000873, 0.00000669, -0.00001110, 0.00000844, 0.00000254, 0.00000190, -0.00001052, 0.00000554, 0.00000203, 0.00000252, -0.00001010, 0.00000589, 0.00000508, -0.00000741, 0.00000312, -0.00000365, 0.00000189, -0.00000849, 0.00001438, -0.00001370, 0.00000479, -0.00000599, },
		{ -0.00000688, -0.00000077, -0.00000575, 0.00001479, -0.00000764, 0.00000334, -0.00001017, 0.00000337, -0.00000959, 0.00000907, -0.00001312, 0.00001284, -0.00001318, 0.00001888, -0.00001762, 0.00001231, -0.00001673, 0.00001484, -0.00002426, 0.00002637, -0.00002822, 0.00002367, -0.00002276, 0.00000608, -0.00000208, 0.00001138, -0.00000962, 0.00001176, -0.00001076, 0.00000873, -0.00002559, 0.00002424, -0.00002495, 0.00002573, -0.00003035, 0.00002000, -0.00002795, 0.00003695, -0.00004807, 0.00003516, -0.00004762, 0.00005047, -0.00006405, 0.00004836, -0.00006410, 0.00006531, -0.00011501, 0.00007115, -0.00012424, 0.00010500, -0.00022955, 0.00011172, -0.00059864, -0.00008728, 0.00091269, 0.00089698, 0.00008751, -0.00069760, 0.00013546, -0.00022695, 0.00008163, -0.00015068, 0.00008868, -0.00009476, 0.00005546, -0.00008038, 0.00005353, -0.00006333, 0.00003926, -0.00005441, 0.00004666, -0.00003909, 0.00003000, -0.00004112, 0.00003323, -0.00003402, 0.00002529, -0.00004208, 0.00003361, -0.00002818, 0.00002321, -0.00002137, 0.00002308, -0.00003555, 0.00002302, -0.00003011, 0.00003124, -0.00003090, 0.00003188, -0.00003908, 0.00002255, -0.00001033, 0.00000460, 0.00000190, 0.00000534, -0.00000816, 0.00000716, -0.00000980, 0.00000846, -0.00000391, 0.00000378, -0.00000474, 0.00000565, -0.00000605, 0.00000522, -0.00000977, 0.00000652, -0.00000906, 0.00000325, 0.00000540, 0.00000282, -0.00000875, 0.00000046, 0.00000522, 0.00000167, -0.00000579, -0.00000258, 0.00001226, -0.00001049, 0.00000501, -0.00000637, 0.00000594, -0.00001285, 0.00001506, -0.00001029, 0.00000187, -0.00000589, 0.00001004, },
		{ -0.00000391, -0.00000559, 0.00001110, -0.00000128, -0.00000046, -0.00000774, 0.00000177, -0.00000891, 0.00000779, -0.00001132, 0.00000993, -0.00000969, 0.00001368, -0.00001127, 0.00000723, -0.00001219, 0.00001038, -0.00001908, 0.00001670, -0.00001393, 0.00000614, -0.00000545, -0.00000768, 0.00000563, 0.00000655, -0.00000487, 0.00000816, -0.00000688, 0.00000868, -0.00002755, 0.00002312, -0.00002155, 0.00002169, -0.00002502, 0.00001661, -0.00002565, 0.00003327, -0.00004346, 0.00003147, -0.00004333, 0.00004327, -0.00005332, 0.00003717, -0.00005167, 0.00005379, -0.00010130, 0.00005763, -0.00010738, 0.00008507, -0.00019622, 0.00007906, -0.00051940, -0.00013979, 0.00085381, 0.00061890, 0.00053376, -0.00078131, 0.00008931, -0.00020804, 0.00005879, -0.00013444, 0.00007123, -0.00008309, 0.00004607, -0.00007174, 0.00004418, -0.00005439, 0.00003003, -0.00004690, 0.00003971, -0.00003439, 0.00002746, -0.00003855, 0.00002938, -0.00003055, 0.00002293, -0.00003782, 0.00002723, -0.00002327, 0.00001927, -0.00001997, 0.00002409, -0.00003307, 0.00001737, -0.00002534, 0.00002539, -0.00002351, 0.00002125, -0.00002190, 0.00000171, 0.00000664, -0.00000758, 0.00000815, 0.00000413, -0.00000667, 0.00000483, -0.00000666, 0.00000432, -0.00000059, 0.00000182, -0.00000335, 0.00000518, -0.00000547, 0.00000456, -0.00000773, 0.00000316, -0.00000428, -0.00000284, 0.00000759, 0.00000332, -0.00000551, -0.00000468, 0.00000717, 0.00000120, -0.00000146, -0.00000877, 0.00001456, -0.00000839, 0.00000275, -0.00000527, 0.00000541, -0.00001111, 0.00000915, -0.00000204, -0.00000394, -0.00000284, 0.00000558, -0.00000077, },
		{ -0.00000454, 0.00000556, 0.00000541, -0.00000365, -0.00000534, -0.00000063, -0.00000641, 0.00000376, -0.00000623, 0.00000349, -0.00000259, 0.00000514, -0.00000242, -0.00000008, -0.00000481, 0.00000223, -0.00000916, 0.00000242, 0.00000279, -0.00001026, 0.00000713, -0.00001494, 0.00000745, 0.00000556, -0.00000251, 0.00000611, -0.00000415, 0.00000865, -0.00002665, 0.00001725, -0.00001333, 0.00001304, -0.00001541, 0.00000870, -0.00001742, 0.00002264, -0.00003147, 0.00001929, -0.00002882, 0.00002487, -0.00003170, 0.00001545, -0.00002864, 0.00003090, -0.00007475, 0.00002908, -0.00007255, 0.00004431, -0.00014079, 0.00002123, -0.00041483, -0.00022879, 0.00087041, 0.00033373, 0.00080909, -0.00065897, -0.00007186, -0.00012508, -0.00001086, -0.00008266, 0.00002542, -0.00004777, 0.00001498, -0.00004310, 0.00001685, -0.00002944, 0.00000687, -0.00002716, 0.00002164, -0.00001908, 0.00001385, -0.00002444, 0.00001434, -0.00001651, 0.00000990, -0.00002307, 0.00001186, -0.00001063, 0.00000798, -0.00001083, 0.00001548, -0.00001982, 0.00000209, -0.00001178, 0.00001069, -0.00000738, 0.00000264, 0.00000023, -0.00001872, 0.00001902, -0.00001376, 0.00000897, 0.00000580, -0.00000565, 0.00000210, -0.00000306, 0.00000028, 0.00000208, 0.00000039, -0.00000213, 0.00000431, -0.00000387, 0.00000241, -0.00000412, -0.00000125, 0.00000049, -0.00000712, 0.00000732, 0.00000465, -0.00000260, -0.00000770, 0.00000625, 0.00000250, 0.00000087, -0.00001049, 0.00001094, -0.00000153, -0.00000266, -0.00000131, 0.00000132, -0.00000511, -0.00000016, 0.00000716, -0.00000922, 0.00000028, 0.00000051, 0.00000490, -0.00000618, },
		{ 0.00000096, 0.00000937, -0.00000383, -0.00000478, -0.00000207, -0.00000407, -0.00000044, -0.00000109, -0.00000261, 0.00000366, -0.00000181, 0.00000387, -0.00000474, 0.00000046, -0.00000412, -0.00000110, -0.00000845, 0.00001304, -0.00001705, 0.00000836, -0.00001155, 0.00000121, 0.00000975, -0.00000386, 0.00000702, -0.00000382, 0.00000965, -0.00002443, 0.00000961, -0.00000451, 0.00000463, -0.00000661, 0.00000133, -0.00000870, 0.00001133, -0.00001909, 0.00000633, -0.00001300, 0.00000587, -0.00001116, -0.00000398, -0.00000845, 0.00001084, -0.00005066, 0.00000253, -0.00003948, 0.00000627, -0.00009109, -0.00002874, -0.00032474, -0.00030134, 0.00087838, 0.00014356, 0.00086373, -0.00036288, -0.00028507, -0.00003052, -0.00008696, -0.00002823, -0.00002158, -0.00001154, -0.00001808, -0.00001241, -0.00001187, -0.00000381, -0.00001623, -0.00000771, 0.00000353, -0.00000302, -0.00000169, -0.00000777, -0.00000286, -0.00000072, -0.00000508, -0.00000675, -0.00000370, 0.00000169, -0.00000346, -0.00000053, 0.00000404, -0.00000372, -0.00001405, 0.00000163, -0.00000362, 0.00000748, -0.00001292, 0.00001571, -0.00002832, 0.00001928, -0.00000957, 0.00000248, 0.00001103, -0.00000597, 0.00000052, -0.00000097, -0.00000167, 0.00000253, 0.00000065, -0.00000198, 0.00000393, -0.00000234, 0.00000027, -0.00000094, -0.00000432, 0.00000285, -0.00000781, 0.00000391, 0.00000714, -0.00000093, -0.00000773, 0.00000240, 0.00000538, 0.00000104, -0.00000807, 0.00000331, 0.00000696, -0.00000801, 0.00000260, -0.00000322, 0.00000122, -0.00000812, 0.00001270, -0.00001036, 0.00000056, -0.00000215, 0.00000731, -0.00000550, -0.00000458, },
		{ 0.00000920, -0.00000037, -0.00000649, -0.00000180, -0.00000327, -0.00000277, 0.00000141, -0.00000525, 0.00000569, -0.00000371, 0.00000440, -0.00000387, 0.00000065, -0.00000519, 0.00000089, -0.00001120, 0.00001261, -0.00001176, -0.00000175, 0.00000028, -0.00001000, 0.00001637, -0.00000699, 0.00000962, -0.00000499, 0.00001124, -0.00002171, 0.00000272, 0.00000144, 0.00000018, -0.00000231, -0.00000174, -0.00000382, 0.00000454, -0.00001216, -0.00000090, -0.00000353, -0.00000491, -0.00000114, -0.00001159, -0.00000075, 0.00000363, -0.00003991, -0.00000969, -0.00002265, -0.00001205, -0.00006642, -0.00004905, -0.00027397, -0.00032654, 0.00083404, 0.00008006, 0.00073405, 0.00000820, -0.00045533, 0.00001272, -0.00012630, -0.00000419, -0.00004351, 0.00000399, -0.00003433, 0.00000320, -0.00002633, 0.00000832, -0.00002662, 0.00000028, -0.00000463, 0.00000461, -0.00001020, 0.00000229, -0.00001303, 0.00000788, -0.00001324, 0.00000240, -0.00001118, 0.00000634, -0.00000837, 0.00000442, -0.00000308, 0.00000698, -0.00002254, 0.00000679, -0.00000955, 0.00001296, -0.00001739, 0.00001751, -0.00002274, 0.00000654, 0.00000327, -0.00000853, 0.00001743, -0.00000658, 0.00000011, -0.00000087, -0.00000114, 0.00000067, 0.00000247, -0.00000286, 0.00000424, -0.00000147, -0.00000085, 0.00000052, -0.00000480, 0.00000198, -0.00000504, -0.00000155, 0.00000980, -0.00000012, -0.00000550, -0.00000280, 0.00000820, 0.00000045, -0.00000376, -0.00000466, 0.00001297, -0.00000980, 0.00000349, -0.00000514, 0.00000430, -0.00001103, 0.00001180, -0.00000592, -0.00000294, -0.00000134, 0.00000560, -0.00000153, -0.00000631, -0.00000126, },
		{ 0.00000523, -0.00000911, -0.00000053, -0.00000392, -0.00000276, 0.00000059, -0.00000378, 0.00000303, -0.00000044, -0.00000042, 0.00000173, -0.00000350, -0.00000146, -0.00000287, -0.00000655, 0.00000366, 0.00000127, -0.00001681, 0.00001333, -0.00001957, 0.00002027, -0.00000820, 0.00001111, -0.00000544, 0.00001195, -0.00001842, -0.00000229, 0.00000322, 0.00000062, -0.00000311, 0.00000005, -0.00000382, 0.00000385, -0.00001245, -0.00000039, -0.00000291, -0.00000473, -0.00000386, -0.00000596, -0.00000639, 0.00000999, -0.00004349, -0.00000603, -0.00002426, -0.00000808, -0.00006874, -0.00003869, -0.00026206, -0.00030627, 0.00074082, 0.00011224, 0.00051239, 0.00033466, -0.00050080, -0.00003350, -0.00010870, -0.00002399, -0.00003082, -0.00000860, -0.00002715, -0.00000254, -0.00002090, 0.00000217, -0.00002033, -0.00000649, 0.00000011, 0.00000077, -0.00000820, 0.00000177, -0.00001208, 0.00000546, -0.00001088, 0.00000076, -0.00000760, 0.00000108, -0.00000476, 0.00000167, -0.00000266, 0.00000842, -0.00001990, 0.00000109, -0.00000497, 0.00000733, -0.00000992, 0.00000668, -0.00000594, -0.00001274, 0.00001764, -0.00001787, 0.00002073, -0.00000522, -0.00000025, -0.00000183, 0.00000078, -0.00000216, 0.00000458, -0.00000380, 0.00000466, -0.00000110, -0.00000087, 0.00000020, -0.00000302, -0.00000111, -0.00000070, -0.00000673, 0.00001088, 0.00000101, -0.00000268, -0.00000705, 0.00000895, 0.00000088, -0.00000016, -0.00000957, 0.00001360, -0.00000635, 0.00000038, -0.00000339, 0.00000295, -0.00000824, 0.00000515, 0.00000233, -0.00000842, 0.00000155, 0.00000136, 0.00000392, -0.00000855, -0.00000159, 0.00000577, },
		{ -0.00001042, 0.00000021, -0.00000489, -0.00000148, -0.00000226, 0.00000006, -0.00000204, 0.00000514, -0.00000713, 0.00000828, -0.00000834, 0.00000353, -0.00000855, 0.00000071, -0.00000739, 0.00001401, -0.00002807, 0.00001964, -0.00002147, 0.00001743, -0.00000482, 0.00000943, -0.00000347, 0.00001069, -0.00001436, -0.00000531, 0.00000149, 0.00000435, -0.00000686, 0.00000440, -0.00000663, 0.00000724, -0.00001755, 0.00000512, -0.00000820, 0.00000278, -0.00001442, 0.00000675, -0.00001844, 0.00002258, -0.00005383, 0.00000558, -0.00003571, 0.00000806, -0.00008533, -0.00001352, -0.00026900, -0.00026589, 0.00063204, 0.00017857, 0.00029961, 0.00052620, -0.00039058, -0.00016028, -0.00004939, -0.00007406, 0.00000439, -0.00003894, -0.00000507, -0.00002243, -0.00000225, -0.00001586, -0.00000359, -0.00002213, 0.00001254, -0.00001019, 0.00000085, -0.00000652, -0.00000279, -0.00000501, -0.00000097, -0.00000860, 0.00000348, -0.00001019, 0.00000380, -0.00000590, 0.00000315, 0.00000255, -0.00000905, -0.00001150, 0.00000570, -0.00000446, 0.00000344, -0.00000897, 0.00001242, -0.00002842, 0.00002492, -0.00001948, 0.00001753, -0.00000037, -0.00000159, -0.00000258, 0.00000242, -0.00000417, 0.00000547, -0.00000372, 0.00000448, -0.00000068, -0.00000035, -0.00000101, -0.00000040, -0.00000438, 0.00000273, -0.00000934, 0.00000907, 0.00000327, -0.00000068, -0.00000871, 0.00000663, 0.00000311, 0.00000134, -0.00001000, 0.00000872, 0.00000111, -0.00000508, 0.00000050, -0.00000119, -0.00000222, -0.00000361, 0.00001028, -0.00001238, 0.00000360, -0.00000248, 0.00000799, -0.00000924, -0.00000167, 0.00000135, 0.00001024, },
		{ -0.00000072, -0.00000499, -0.00000047, -0.00000500, 0.00000348, -0.00000617, 0.00000923, -0.00001156, 0.00001163, -0.00000996, 0.00000583, -0.00001172, 0.00000524, -0.00001410, 0.00001956, -0.00002948, 0.00001525, -0.00001404, 0.00000785, 0.00000256, 0.00000504, 0.00000061, 0.00000787, -0.00001006, -0.00000651, -0.00000215, 0.00000870, -0.00001037, 0.00000803, -0.00000901, 0.00001108, -0.00002308, 0.00001069, -0.00001387, 0.00001100, -0.00002477, 0.00001748, -0.00002740, 0.00003167, -0.00006081, 0.00001432, -0.00004485, 0.00002207, -0.00009890, 0.00000606, -0.00027055, -0.00023471, 0.00054396, 0.00022399, 0.00016344, 0.00055379, -0.00015754, -0.00031386, 0.00000913, -0.00012259, 0.00003660, -0.00006623, 0.00001467, -0.00004152, 0.00001589, -0.00003299, 0.00001182, -0.00003596, 0.00002321, -0.00001996, 0.00000956, -0.00001557, 0.00000775, -0.00001627, 0.00000932, -0.00001852, 0.00001464, -0.00002019, 0.00001077, -0.00001251, 0.00000906, -0.00000492, 0.00000313, -0.00002337, 0.00001476, -0.00001442, 0.00001399, -0.00001991, 0.00002261, -0.00003235, 0.00002004, -0.00001138, 0.00000798, 0.00000706, -0.00000355, -0.00000260, 0.00000279, -0.00000436, 0.00000443, -0.00000220, 0.00000341, 0.00000009, 0.00000011, -0.00000208, 0.00000154, -0.00000604, 0.00000355, -0.00000843, 0.00000448, 0.00000637, 0.00000023, -0.00000761, 0.00000190, 0.00000632, 0.00000119, -0.00000694, 0.00000085, 0.00000905, -0.00000949, 0.00000353, -0.00000496, 0.00000315, -0.00000999, 0.00001382, -0.00001180, 0.00000247, -0.00000356, 0.00000860, -0.00000697, -0.00000286, -0.00000191, 0.00001234, -0.00000881, },
		{ -0.00000395, -0.00000059, -0.00000611, 0.00000445, -0.00000708, 0.00000941, -0.00001134, 0.00000977, -0.00000675, 0.00000378, -0.00001033, 0.00000444, -0.00001366, 0.00001585, -0.00002086, 0.00000234, -0.00000118, -0.00000413, 0.00001014, 0.00000083, 0.00000457, 0.00000531, -0.00000684, -0.00000583, -0.00000621, 0.00001141, -0.00001133, 0.00000871, -0.00000872, 0.00001261, -0.00002556, 0.00001234, -0.00001544, 0.00001454, -0.00002863, 0.00001974, -0.00002704, 0.00003122, -0.00005828, 0.00001344, -0.00004383, 0.00002448, -0.00009814, 0.00000738, -0.00025240, -0.00022901, 0.00049488, 0.00022378, 0.00011974, 0.00045157, 0.00011840, -0.00042088, 0.00002092, -0.00013919, 0.00004294, -0.00007256, 0.00001746, -0.00004698, 0.00002168, -0.00003833, 0.00001607, -0.00003928, 0.00002451, -0.00002175, 0.00001156, -0.00001899, 0.00001279, -0.00002146, 0.00001332, -0.00002239, 0.00001927, -0.00002281, 0.00001091, -0.00001346, 0.00001036, -0.00000856, 0.00001019, -0.00002793, 0.00001609, -0.00001657, 0.00001568, -0.00002036, 0.00002013, -0.00002281, 0.00000455, 0.00000282, -0.00000372, 0.00001368, -0.00000435, -0.00000248, 0.00000206, -0.00000294, 0.00000196, 0.00000010, 0.00000199, 0.00000100, 0.00000032, -0.00000245, 0.00000205, -0.00000548, 0.00000167, -0.00000484, -0.00000124, 0.00000888, 0.00000085, -0.00000497, -0.00000325, 0.00000862, 0.00000097, -0.00000282, -0.00000632, 0.00001361, -0.00000980, 0.00000323, -0.00000581, 0.00000491, -0.00001111, 0.00001115, -0.00000614, -0.00000196, -0.00000160, 0.00000570, -0.00000217, -0.00000514, -0.00000324, 0.00001091, -0.00000436, -0.00000315, },
		{ -0.00000161, -0.00000544, 0.00000275, -0.00000467, 0.00000588, -0.00000710, 0.00000389, -0.00000030, -0.00000109, -0.00000567, -0.00000042, -0.00000787, 0.00000616, -0.00000747, -0.00001223, 0.00000996, -0.00001230, 0.00001322, 0.00000012, 0.00000591, 0.00000500, -0.00000590, -0.00000337, -0.00000960, 0.00001146, -0.00000917, 0.00000625, -0.00000561, 0.00001127, -0.00002399, 0.00000906, -0.00001180, 0.00001194, -0.00002456, 0.00001277, -0.00001751, 0.00002196, -0.00004707, 0.00000354, -0.00003280, 0.00001496, -0.00008267, -0.00000907, -0.00021602, -0.00024606, 0.00047925, 0.00018649, 0.00014038, 0.00029197, 0.00034681, -0.00042351, -0.00003694, -0.00011365, 0.00001719, -0.00005398, 0.00000033, -0.00003591, 0.00001211, -0.00002922, 0.00000701, -0.00003050, 0.00001526, -0.00001445, 0.00000546, -0.00001484, 0.00000991, -0.00001806, 0.00000875, -0.00001806, 0.00001534, -0.00001665, 0.00000349, -0.00000813, 0.00000600, -0.00000647, 0.00000968, -0.00002326, 0.00000865, -0.00001029, 0.00000836, -0.00001116, 0.00000774, -0.00000526, -0.00001412, 0.00001555, -0.00001138, 0.00001549, -0.00000204, -0.00000320, 0.00000112, -0.00000108, -0.00000059, 0.00000193, 0.00000115, 0.00000147, 0.00000053, -0.00000218, 0.00000132, -0.00000335, -0.00000157, -0.00000068, -0.00000572, 0.00000917, 0.00000228, -0.00000240, -0.00000672, 0.00000831, 0.00000215, 0.00000011, -0.00000996, 0.00001267, -0.00000513, -0.00000066, -0.00000337, 0.00000265, -0.00000720, 0.00000375, 0.00000218, -0.00000739, 0.00000149, 0.00000132, 0.00000304, -0.00000708, -0.00000340, 0.00000714, 0.00000117, -0.00000562, -0.00000258, },
		{ -0.00000399, -0.00000030, -0.00000079, 0.00000100, -0.00000172, -0.00000257, 0.00000574, -0.00000533, -0.00000114, -0.00000563, -0.00000131, -0.00000353, 0.00000342, -0.00002100, 0.00001317, -0.00001238, 0.00000933, 0.00000423, 0.00000376, 0.00000750, -0.00000729, 0.00000018, -0.00001156, 0.00000895, -0.00000493, 0.00000217, -0.00000130, 0.00000846, -0.00001984, 0.00000273, -0.00000515, 0.00000565, -0.00001570, 0.00000098, -0.00000431, 0.00000992, -0.00003341, -0.00000909, -0.00001825, 0.00000055, -0.00006084, -0.00003270, -0.00017434, -0.00026999, 0.00047614, 0.00013966, 0.00017851, 0.00014905, 0.00046458, -0.00030664, -0.00014976, -0.00006292, -0.00002662, -0.00002247, -0.00002671, -0.00001644, -0.00000598, -0.00001202, -0.00000923, -0.00001550, 0.00000095, -0.00000285, -0.00000475, -0.00000634, 0.00000184, -0.00000888, -0.00000133, -0.00000865, 0.00000611, -0.00000547, -0.00000746, -0.00000015, -0.00000109, -0.00000091, 0.00000388, -0.00001264, -0.00000333, -0.00000022, -0.00000282, 0.00000147, -0.00000674, 0.00001101, -0.00002669, 0.00001951, -0.00001041, 0.00001046, 0.00000380, -0.00000495, 0.00000066, 0.00000012, -0.00000195, 0.00000226, 0.00000157, 0.00000110, 0.00000102, -0.00000169, 0.00000014, -0.00000096, -0.00000432, 0.00000190, -0.00000719, 0.00000644, 0.00000492, -0.00000088, -0.00000739, 0.00000502, 0.00000490, 0.00000110, -0.00000939, 0.00000690, 0.00000257, -0.00000595, 0.00000039, -0.00000157, -0.00000114, -0.00000448, 0.00000906, -0.00001043, 0.00000296, -0.00000180, 0.00000614, -0.00000695, -0.00000380, 0.00000308, 0.00000541, -0.00000630, -0.00000207, -0.00000258, },
		{ -0.00000291, 0.00000217, -0.00000239, 0.00000158, -0.00000618, 0.00000808, -0.00000593, 0.00000029, -0.00000781, 0.00000187, -0.00000826, 0.00000679, -0.00002010, 0.00000674, -0.00000470, 0.00000001, 0.00001125, -0.00000020, 0.00001138, -0.00000954, 0.00000337, -0.00001145, 0.00000466, -0.00000031, -0.00000143, 0.00000210, 0.00000626, -0.00001554, -0.00000351, 0.00000088, -0.00000024, -0.00000711, -0.00000939, 0.00000560, 0.00000216, -0.00002430, -0.00001732, -0.00000777, -0.00001013, -0.00004300, -0.00005084, -0.00014202, -0.00028371, 0.00046427, 0.00010962, 0.00019580, 0.00006833, 0.00045779, -0.00010425, -0.00027043, -0.00002211, -0.00006307, 0.00000191, -0.00004754, -0.00000179, -0.00002105, 0.00000259, -0.00002269, -0.00000349, -0.00001011, 0.00000577, -0.00001270, 0.00000077, -0.00000588, 0.00000029, -0.00001089, -0.00000013, -0.00000240, 0.00000443, -0.00001584, 0.00000512, -0.00000624, 0.00000377, -0.00000239, -0.00000206, -0.00001323, 0.00000704, -0.00001100, 0.00001003, -0.00001523, 0.00001810, -0.00002720, 0.00001184, -0.00000078, 0.00000031, 0.00001110, -0.00000656, 0.00000051, 0.00000030, -0.00000174, 0.00000096, 0.00000314, 0.00000007, 0.00000174, -0.00000125, -0.00000077, 0.00000056, -0.00000531, 0.00000185, -0.00000538, 0.00000143, 0.00000796, -0.00000021, -0.00000569, -0.00000002, 0.00000793, 0.00000102, -0.00000607, -0.00000080, 0.00000965, -0.00000933, 0.00000260, -0.00000471, 0.00000343, -0.00000949, 0.00001102, -0.00000876, 0.00000110, -0.00000198, 0.00000574, -0.00000403, -0.00000525, 0.00000030, 0.00000670, -0.00000426, -0.00000303, -0.00000274, -0.00000294, },
		{ 0.00000277, -0.00000268, 0.00000123, -0.00000550, 0.00000571, -0.00000235, -0.00000198, -0.00000605, 0.00000040, -0.00000677, 0.00000230, -0.00001114, -0.00000564, 0.00000590, -0.00000981, 0.00001696, -0.00000272, 0.00001403, -0.00001044, 0.00000457, -0.00000891, -0.00000039, 0.00000312, -0.00000309, 0.00000327, 0.00000597, -0.00001289, -0.00000726, 0.00000356, -0.00000264, -0.00000260, -0.00001391, 0.00000791, 0.00000247, -0.00002316, -0.00001773, -0.00000521, -0.00001239, -0.00003508, -0.00005635, -0.00012672, -0.00027899, 0.00043384, 0.00010725, 0.00017846, 0.00005532, 0.00035968, 0.00011703, -0.00034160, -0.00002455, -0.00007120, 0.00000368, -0.00005028, -0.00000158, -0.00002453, 0.00000652, -0.00002589, -0.00000117, -0.00001209, 0.00000637, -0.00001388, 0.00000217, -0.00000867, 0.00000446, -0.00001479, 0.00000251, -0.00000525, 0.00000812, -0.00001724, 0.00000411, -0.00000632, 0.00000432, -0.00000508, 0.00000351, -0.00001609, 0.00000706, -0.00001188, 0.00001026, -0.00001392, 0.00001365, -0.00001608, -0.00000408, 0.00001257, -0.00000995, 0.00001605, -0.00000602, -0.00000015, -0.00000007, -0.00000054, -0.00000114, 0.00000498, -0.00000089, 0.00000226, -0.00000083, -0.00000115, 0.00000085, -0.00000439, -0.00000039, -0.00000165, -0.00000387, 0.00000980, 0.00000064, -0.00000307, -0.00000469, 0.00000936, 0.00000143, -0.00000241, -0.00000695, 0.00001274, -0.00000834, 0.00000141, -0.00000484, 0.00000424, -0.00000928, 0.00000721, -0.00000266, -0.00000349, 0.00000035, 0.00000251, 0.00000068, -0.00000722, -0.00000095, 0.00000502, -0.00000008, -0.00000489, -0.00000213, -0.00000283, -0.00000398, },
		{ -0.00000028, -0.00000204, -0.00000173, 0.00000041, 0.00000330, -0.00000590, -0.00000222, -0.00000378, -0.00000159, -0.00000612, 0.00000021, -0.00001718, 0.00001289, -0.00001477, 0.00001740, -0.00000112, 0.00001349, -0.00000851, 0.00000304, -0.00000433, -0.00000508, 0.00000448, -0.00000243, 0.00000222, 0.00000754, -0.00001219, -0.00000791, 0.00000240, -0.00000107, -0.00000287, -0.00001206, 0.00000293, 0.00000958, -0.00002820, -0.00001220, -0.00000892, -0.00000753, -0.00003608, -0.00005043, -0.00012602, -0.00025951, 0.00038935, 0.00012563, 0.00013687, 0.00008435, 0.00022878, 0.00028799, -0.00032329, -0.00008198, -0.00004787, -0.00001833, -0.00003479, -0.00001575, -0.00001605, -0.00000101, -0.00001820, -0.00000886, -0.00000501, -0.00000080, -0.00000840, -0.00000243, -0.00000569, 0.00000233, -0.00001171, -0.00000179, -0.00000149, 0.00000481, -0.00001149, -0.00000274, -0.00000175, 0.00000065, -0.00000328, 0.00000276, -0.00001124, -0.00000001, -0.00000603, 0.00000320, -0.00000489, 0.00000167, 0.00000029, -0.00002040, 0.00002222, -0.00001462, 0.00001526, -0.00000191, -0.00000193, -0.00000028, 0.00000062, -0.00000283, 0.00000596, -0.00000097, 0.00000210, -0.00000018, -0.00000116, 0.00000027, -0.00000242, -0.00000332, 0.00000188, -0.00000727, 0.00000904, 0.00000255, -0.00000097, -0.00000716, 0.00000791, 0.00000335, -0.00000013, -0.00000938, 0.00001052, -0.00000292, -0.00000275, -0.00000222, 0.00000157, -0.00000484, -0.00000022, 0.00000498, -0.00000818, 0.00000299, -0.00000129, 0.00000496, -0.00000823, -0.00000157, 0.00000173, 0.00000440, -0.00000622, -0.00000155, -0.00000333, -0.00000359, 0.00000112, },
		{ -0.00000586, 0.00000230, -0.00000452, 0.00000760, -0.00000832, 0.00000059, -0.00000730, 0.00000311, -0.00001313, 0.00000770, -0.00002187, 0.00001180, -0.00001223, 0.00001146, 0.00000478, 0.00000988, -0.00000408, -0.00000045, 0.00000094, -0.00000819, 0.00000353, -0.00000011, 0.00000003, 0.00000982, -0.00001248, -0.00000655, -0.00000104, 0.00000259, -0.00000578, -0.00000669, -0.00000529, 0.00001853, -0.00003411, -0.00000600, -0.00001366, -0.00000088, -0.00004013, -0.00004020, -0.00013070, -0.00023650, 0.00034387, 0.00014794, 0.00009322, 0.00011921, 0.00012015, 0.00036544, -0.00021192, -0.00017596, -0.00001068, -0.00005032, -0.00001253, -0.00003494, -0.00000313, -0.00001359, -0.00000559, -0.00002082, 0.00000563, -0.00001068, -0.00000066, -0.00000939, 0.00000005, -0.00000341, -0.00000450, -0.00000990, 0.00000566, -0.00000212, -0.00000248, -0.00001138, 0.00000390, -0.00000437, 0.00000069, -0.00000185, -0.00000216, -0.00000981, 0.00000191, -0.00000601, 0.00000568, -0.00001031, 0.00001330, -0.00002906, 0.00002234, -0.00001060, 0.00000800, 0.00000517, -0.00000436, -0.00000014, 0.00000107, -0.00000330, 0.00000546, 0.00000019, 0.00000112, 0.00000075, -0.00000105, -0.00000051, -0.00000057, -0.00000534, 0.00000345, -0.00000751, 0.00000543, 0.00000551, 0.00000006, -0.00000689, 0.00000380, 0.00000641, 0.00000053, -0.00000809, 0.00000433, 0.00000448, -0.00000738, 0.00000094, -0.00000225, 0.00000076, -0.00000736, 0.00001026, -0.00000988, 0.00000346, -0.00000326, 0.00000666, -0.00000694, -0.00000267, -0.00000143, 0.00000718, -0.00000562, -0.00000189, -0.00000366, -0.00000267, -0.00000143, 0.00000294, },
		{ 0.00000397, -0.00000638, 0.00000808, -0.00000711, 0.00000030, -0.00000773, 0.00000430, -0.00001514, 0.00000798, -0.00001780, 0.00000283, -0.00000395, 0.00000170, 0.00001226, 0.00000556, 0.00000078, -0.00000387, 0.00000502, -0.00000874, 0.00000062, 0.00000272, -0.00000188, 0.00001144, -0.00001242, -0.00000490, -0.00000440, 0.00000558, -0.00000818, -0.00000181, -0.00001177, 0.00002390, -0.00003568, -0.00000412, -0.00001447, 0.00000222, -0.00004105, -0.00003330, -0.00013120, -0.00022097, 0.00030962, 0.00015923, 0.00006610, 0.00013335, 0.00006368, 0.00034621, -0.00004172, -0.00026521, 0.00001148, -0.00007211, 0.00000083, -0.00004673, 0.00000414, -0.00002237, 0.00000359, -0.00002923, 0.00001264, -0.00001683, 0.00000375, -0.00001381, 0.00000407, -0.00000830, 0.00000205, -0.00001675, 0.00001117, -0.00000758, 0.00000446, -0.00001673, 0.00000631, -0.00000703, 0.00000333, -0.00000626, 0.00000599, -0.00001672, 0.00000641, -0.00001162, 0.00001151, -0.00001567, 0.00001684, -0.00002610, 0.00001205, 0.00000054, -0.00000282, 0.00001240, -0.00000577, -0.00000018, 0.00000085, -0.00000259, 0.00000374, 0.00000214, -0.00000023, 0.00000173, -0.00000092, -0.00000102, 0.00000038, -0.00000560, 0.00000256, -0.00000492, 0.00000016, 0.00000838, 0.00000057, -0.00000472, -0.00000134, 0.00000906, 0.00000070, -0.00000482, -0.00000275, 0.00001032, -0.00000943, 0.00000208, -0.00000452, 0.00000427, -0.00001076, 0.00001048, -0.00000699, 0.00000072, -0.00000237, 0.00000512, -0.00000325, -0.00000453, -0.00000339, 0.00000723, -0.00000270, -0.00000331, -0.00000338, -0.00000216, -0.00000337, 0.00000497, -0.00000787, },
		{ -0.00000450, 0.00000457, -0.00000256, -0.00000283, -0.00000506, 0.00000176, -0.00001213, 0.00000202, -0.00000779, -0.00000948, 0.00000492, -0.00000702, 0.00001720, 0.00000371, 0.00000352, -0.00000507, 0.00000639, -0.00000643, -0.00000337, 0.00000486, -0.00000250, 0.00001157, -0.00001119, -0.00000422, -0.00000592, 0.00000595, -0.00000790, -0.00000014, -0.00001339, 0.00002294, -0.00003081, -0.00000820, -0.00000973, -0.00000020, -0.00003616, -0.00003340, -0.00012296, -0.00021756, 0.00029098, 0.00015439, 0.00006103, 0.00012021, 0.00005814, 0.00026223, 0.00013143, -0.00030511, -0.00000484, -0.00006989, -0.00000464, -0.00004368, -0.00000025, -0.00002181, 0.00000399, -0.00002912, 0.00001159, -0.00001545, 0.00000166, -0.00001279, 0.00000346, -0.00000908, 0.00000424, -0.00001851, 0.00001138, -0.00000795, 0.00000576, -0.00001577, 0.00000325, -0.00000543, 0.00000251, -0.00000753, 0.00000949, -0.00001715, 0.00000443, -0.00001071, 0.00000983, -0.00001222, 0.00001036, -0.00001364, -0.00000403, 0.00001311, -0.00001192, 0.00001600, -0.00000427, -0.00000120, 0.00000051, -0.00000139, 0.00000174, 0.00000392, -0.00000118, 0.00000224, -0.00000060, -0.00000117, 0.00000032, -0.00000434, 0.00000004, -0.00000114, -0.00000464, 0.00000955, 0.00000161, -0.00000217, -0.00000542, 0.00000955, 0.00000177, -0.00000172, -0.00000769, 0.00001183, -0.00000716, 0.00000002, -0.00000391, 0.00000414, -0.00000927, 0.00000562, -0.00000053, -0.00000403, 0.00000036, 0.00000152, 0.00000143, -0.00000633, -0.00000427, 0.00000491, 0.00000162, -0.00000506, -0.00000269, -0.00000236, -0.00000393, 0.00000480, -0.00000708, 0.00000241, },
		{ -0.00000077, 0.00000292, -0.00000652, -0.00000145, -0.00000223, -0.00000695, -0.00000603, 0.00000253, -0.00001887, 0.00000888, -0.00001041, 0.00001657, 0.00000622, 0.00000285, -0.00000318, 0.00000487, -0.00000206, -0.00000712, 0.00000539, -0.00000149, 0.00001025, -0.00000887, -0.00000475, -0.00000518, 0.00000350, -0.00000482, -0.00000192, -0.00001021, 0.00001660, -0.00002141, -0.00001591, -0.00000176, -0.00000607, -0.00002718, -0.00003896, -0.00010784, -0.00022322, 0.00028334, 0.00013900, 0.00007034, 0.00009071, 0.00008044, 0.00016119, 0.00025471, -0.00026927, -0.00006335, -0.00004533, -0.00002687, -0.00002805, -0.00001440, -0.00001310, -0.00000375, -0.00002115, 0.00000332, -0.00000761, -0.00000584, -0.00000719, -0.00000137, -0.00000564, 0.00000157, -0.00001474, 0.00000613, -0.00000319, 0.00000153, -0.00000924, -0.00000402, -0.00000069, -0.00000120, -0.00000548, 0.00000787, -0.00001136, -0.00000297, -0.00000471, 0.00000250, -0.00000295, -0.00000140, 0.00000152, -0.00001819, 0.00002030, -0.00001442, 0.00001341, 0.00000084, -0.00000333, 0.00000047, -0.00000052, 0.00000042, 0.00000465, -0.00000107, 0.00000193, 0.00000006, -0.00000113, -0.00000030, -0.00000248, -0.00000259, 0.00000181, -0.00000708, 0.00000796, 0.00000381, -0.00000043, -0.00000702, 0.00000714, 0.00000426, -0.00000004, -0.00000899, 0.00000853, -0.00000124, -0.00000425, -0.00000124, 0.00000120, -0.00000444, -0.00000172, 0.00000639, -0.00000804, 0.00000263, -0.00000182, 0.00000494, -0.00000667, -0.00000497, 0.00000161, 0.00000558, -0.00000583, -0.00000231, -0.00000291, -0.00000335, 0.00000285, -0.00000423, -0.00000124, -0.00000049, },
		{ 0.00000633, -0.00000805, 0.00000049, -0.00000479, -0.00000319, -0.00001164, 0.00000808, -0.00002094, 0.00000525, -0.00000737, 0.00001039, 0.00001232, -0.00000029, 0.00000076, 0.00000184, 0.00000258, -0.00000919, 0.00000388, 0.00000079, 0.00000820, -0.00000621, -0.00000588, -0.00000298, -0.00000050, -0.00000055, -0.00000529, -0.00000479, 0.00000854, -0.00001183, -0.00002279, 0.00000518, -0.00001127, -0.00001829, -0.00004534, -0.00009156, -0.00023042, 0.00027737, 0.00012385, 0.00008052, 0.00006256, 0.00010255, 0.00008392, 0.00029961, -0.00016212, -0.00014380, -0.00001555, -0.00005300, -0.00001023, -0.00002994, -0.00000287, -0.00001395, -0.00001069, -0.00000700, 0.00000177, -0.00001423, -0.00000089, -0.00000722, -0.00000068, -0.00000343, -0.00000818, -0.00000152, 0.00000354, -0.00000488, -0.00000096, -0.00001160, 0.00000382, -0.00000512, -0.00000232, 0.00000361, -0.00000283, -0.00001160, 0.00000201, -0.00000556, 0.00000630, -0.00001146, 0.00001179, -0.00002377, 0.00001788, -0.00000869, 0.00000511, 0.00000813, -0.00000556, 0.00000054, -0.00000029, 0.00000024, 0.00000405, 0.00000013, 0.00000091, 0.00000099, -0.00000104, -0.00000093, -0.00000101, -0.00000404, 0.00000263, -0.00000644, 0.00000386, 0.00000676, 0.00000036, -0.00000609, 0.00000254, 0.00000734, 0.00000049, -0.00000719, 0.00000223, 0.00000562, -0.00000810, 0.00000128, -0.00000218, 0.00000066, -0.00000780, 0.00001027, -0.00000868, 0.00000240, -0.00000296, 0.00000564, -0.00000468, -0.00000623, -0.00000126, 0.00000762, -0.00000467, -0.00000279, -0.00000324, -0.00000248, 0.00000042, -0.00000115, -0.00000476, 0.00000303, -0.00000499, },
		{ -0.00000614, -0.00000050, -0.00000438, -0.00000281, -0.00001253, 0.00000702, -0.00001540, -0.00000432, -0.00000060, 0.00000179, 0.00001890, -0.00000321, 0.00000437, -0.00000043, 0.00000550, -0.00000851, 0.00000051, 0.00000351, 0.00000636, -0.00000408, -0.00000677, -0.00000054, -0.00000421, 0.00000280, -0.00000784, -0.00000027, 0.00000266, -0.00000618, -0.00002513, 0.00000775, -0.00001258, -0.00001300, -0.00004830, -0.00007954, -0.00023229, 0.00026509, 0.00011762, 0.00008109, 0.00004865, 0.00010641, 0.00004933, 0.00027015, -0.00001672, -0.00021067, -0.00000366, -0.00006751, -0.00000209, -0.00003764, 0.00000140, -0.00001994, -0.00000410, -0.00001337, 0.00000719, -0.00001864, 0.00000195, -0.00001041, 0.00000234, -0.00000719, -0.00000281, -0.00000733, 0.00000806, -0.00000923, 0.00000464, -0.00001538, 0.00000477, -0.00000646, -0.00000076, 0.00000002, 0.00000413, -0.00001681, 0.00000494, -0.00000971, 0.00001047, -0.00001464, 0.00001279, -0.00001867, 0.00000658, 0.00000245, -0.00000519, 0.00001434, -0.00000602, 0.00000003, -0.00000043, 0.00000089, 0.00000261, 0.00000182, -0.00000026, 0.00000179, -0.00000087, -0.00000128, -0.00000044, -0.00000387, 0.00000129, -0.00000352, -0.00000123, 0.00000914, 0.00000096, -0.00000374, -0.00000238, 0.00000935, 0.00000106, -0.00000412, -0.00000406, 0.00001016, -0.00000894, 0.00000152, -0.00000372, 0.00000328, -0.00000989, 0.00000920, -0.00000509, -0.00000075, -0.00000151, 0.00000350, -0.00000077, -0.00000792, -0.00000311, 0.00000721, -0.00000162, -0.00000401, -0.00000311, -0.00000204, -0.00000125, 0.00000052, -0.00000627, 0.00000406, -0.00000605, 0.00000597, },
		{ -0.00000354, -0.00000169, -0.00000525, -0.00000953, 0.00000125, -0.00000589, -0.00001490, 0.00000498, -0.00000486, 0.00002225, -0.00000313, 0.00000545, -0.00000009, 0.00000550, -0.00000503, -0.00000385, 0.00000566, 0.00000551, -0.00000303, -0.00000686, 0.00000118, -0.00000631, 0.00000383, -0.00000807, 0.00000144, 0.00000108, -0.00000603, -0.00002211, 0.00000560, -0.00000965, -0.00001198, -0.00004657, -0.00007378, -0.00022635, 0.00024405, 0.00012226, 0.00006996, 0.00005099, 0.00008998, 0.00005222, 0.00019600, 0.00012049, -0.00023005, -0.00002523, -0.00006216, -0.00000937, -0.00003335, -0.00000363, -0.00001847, -0.00000472, -0.00001271, 0.00000592, -0.00001682, -0.00000052, -0.00000921, 0.00000157, -0.00000748, -0.00000128, -0.00000856, 0.00000780, -0.00000903, 0.00000514, -0.00001351, 0.00000105, -0.00000430, -0.00000202, -0.00000087, 0.00000684, -0.00001619, 0.00000225, -0.00000821, 0.00000801, -0.00001015, 0.00000538, -0.00000623, -0.00000820, 0.00001309, -0.00001234, 0.00001605, -0.00000319, -0.00000162, -0.00000048, 0.00000167, 0.00000123, 0.00000308, -0.00000081, 0.00000199, -0.00000041, -0.00000138, -0.00000066, -0.00000253, -0.00000116, -0.00000008, -0.00000524, 0.00000947, 0.00000233, -0.00000145, -0.00000571, 0.00000887, 0.00000275, -0.00000158, -0.00000784, 0.00001036, -0.00000579, -0.00000104, -0.00000271, 0.00000256, -0.00000760, 0.00000388, 0.00000118, -0.00000518, 0.00000112, 0.00000005, 0.00000340, -0.00000902, -0.00000428, 0.00000498, 0.00000219, -0.00000511, -0.00000281, -0.00000218, -0.00000167, 0.00000018, -0.00000528, 0.00000230, -0.00000396, 0.00000237, -0.00000171, },
		{ 0.00000109, -0.00000817, -0.00000557, -0.00000519, 0.00000238, -0.00002116, 0.00000516, -0.00000652, 0.00002032, 0.00000107, 0.00000338, 0.00000310, 0.00000294, 0.00000009, -0.00000773, 0.00000631, 0.00000597, -0.00000306, -0.00000617, 0.00000194, -0.00000647, 0.00000251, -0.00000609, 0.00000043, 0.00000333, -0.00001000, -0.00001599, 0.00000127, -0.00000490, -0.00001308, -0.00004200, -0.00007246, -0.00021485, 0.00021764, 0.00013305, 0.00005279, 0.00006212, 0.00006369, 0.00007245, 0.00011566, 0.00020988, -0.00018601, -0.00007817, -0.00004172, -0.00002799, -0.00002073, -0.00001501, -0.00001156, -0.00001108, -0.00000644, -0.00000050, -0.00001042, -0.00000666, -0.00000493, -0.00000218, -0.00000463, -0.00000355, -0.00000537, 0.00000325, -0.00000487, 0.00000129, -0.00000738, -0.00000555, -0.00000024, -0.00000511, 0.00000061, 0.00000548, -0.00001068, -0.00000444, -0.00000299, 0.00000134, -0.00000149, -0.00000531, 0.00000679, -0.00001931, 0.00001734, -0.00001251, 0.00001165, 0.00000290, -0.00000408, -0.00000030, 0.00000204, 0.00000062, 0.00000327, -0.00000033, 0.00000139, 0.00000040, -0.00000140, -0.00000120, -0.00000095, -0.00000329, 0.00000211, -0.00000665, 0.00000710, 0.00000479, -0.00000008, -0.00000646, 0.00000563, 0.00000560, -0.00000028, -0.00000829, 0.00000644, 0.00000012, -0.00000498, -0.00000031, -0.00000028, -0.00000286, -0.00000284, 0.00000698, -0.00000822, 0.00000270, -0.00000250, 0.00000586, -0.00000836, -0.00000554, 0.00000222, 0.00000522, -0.00000502, -0.00000290, -0.00000255, -0.00000118, -0.00000154, -0.00000281, -0.00000085, -0.00000054, -0.00000219, 0.00000275, -0.00000630, },
		{ -0.00000926, -0.00000351, -0.00000872, 0.00000559, -0.00002028, -0.00000107, -0.00000331, 0.00001396, 0.00000795, -0.00000034, 0.00000761, -0.00000042, 0.00000487, -0.00000963, 0.00000496, 0.00000757, -0.00000372, -0.00000516, 0.00000209, -0.00000530, -0.00000004, -0.00000331, -0.00000169, 0.00000715, -0.00001486, -0.00001051, -0.00000156, -0.00000168, -0.00001316, -0.00003778, -0.00007193, -0.00020252, 0.00019195, 0.00014254, 0.00003811, 0.00007145, 0.00004132, 0.00008867, 0.00005901, 0.00023392, -0.00008827, -0.00014200, -0.00002197, -0.00004659, -0.00000878, -0.00002562, -0.00000485, -0.00001829, 0.00000078, -0.00000759, -0.00000372, -0.00001263, -0.00000083, -0.00000618, -0.00000101, -0.00000729, -0.00000037, -0.00000267, 0.00000025, -0.00000370, -0.00000058, -0.00001141, 0.00000273, -0.00000767, 0.00000247, 0.00000237, -0.00000355, -0.00001125, 0.00000197, -0.00000509, 0.00000605, -0.00001313, 0.00001400, -0.00002164, 0.00001251, -0.00000541, 0.00000264, 0.00001018, -0.00000598, -0.00000032, 0.00000192, 0.00000094, 0.00000240, 0.00000102, 0.00000024, 0.00000137, -0.00000137, -0.00000166, 0.00000008, -0.00000412, 0.00000219, -0.00000527, 0.00000270, 0.00000762, 0.00000053, -0.00000501, 0.00000080, 0.00000844, 0.00000036, -0.00000632, 0.00000050, 0.00000601, -0.00000780, 0.00000138, -0.00000300, 0.00000143, -0.00000761, 0.00000935, -0.00000776, 0.00000165, -0.00000271, 0.00000552, -0.00000553, -0.00000726, -0.00000005, 0.00000633, -0.00000320, -0.00000365, -0.00000275, -0.00000053, -0.00000344, -0.00000053, -0.00000338, 0.00000190, -0.00000513, 0.00000476, -0.00000662, 0.00000190, },
		{ -0.00000441, -0.00000814, 0.00000313, -0.00001329, -0.00001104, 0.00000162, 0.00000636, 0.00001436, -0.00000301, 0.00001110, -0.00000234, 0.00000743, -0.00000862, 0.00000179, 0.00000964, -0.00000431, -0.00000443, 0.00000222, -0.00000375, -0.00000237, -0.00000140, -0.00000304, 0.00001004, -0.00001753, -0.00000870, -0.00000044, -0.00000203, -0.00001029, -0.00003611, -0.00006930, -0.00019323, 0.00017185, 0.00014520, 0.00003177, 0.00007199, 0.00003140, 0.00008904, 0.00003709, 0.00020111, 0.00003237, -0.00018692, -0.00002077, -0.00005374, -0.00000611, -0.00002888, -0.00000366, -0.00002152, 0.00000422, -0.00001087, -0.00000080, -0.00001482, 0.00000004, -0.00000772, 0.00000076, -0.00000969, 0.00000321, -0.00000649, 0.00000292, -0.00000642, 0.00000329, -0.00001335, 0.00000218, -0.00000773, 0.00000274, 0.00000015, 0.00000169, -0.00001449, 0.00000320, -0.00000765, 0.00000854, -0.00001425, 0.00001261, -0.00001465, 0.00000063, 0.00000533, -0.00000687, 0.00001528, -0.00000550, -0.00000122, 0.00000167, 0.00000176, 0.00000105, 0.00000255, -0.00000081, 0.00000206, -0.00000117, -0.00000188, 0.00000030, -0.00000353, 0.00000049, -0.00000219, -0.00000204, 0.00000944, 0.00000125, -0.00000261, -0.00000372, 0.00000965, 0.00000150, -0.00000370, -0.00000471, 0.00000910, -0.00000747, 0.00000081, -0.00000384, 0.00000313, -0.00000841, 0.00000715, -0.00000363, -0.00000184, -0.00000076, 0.00000288, -0.00000137, -0.00000899, -0.00000156, 0.00000538, -0.00000001, -0.00000475, -0.00000268, -0.00000027, -0.00000452, 0.00000027, -0.00000389, 0.00000194, -0.00000510, 0.00000334, -0.00000395, 0.00000020, -0.00000778, },
		{ -0.00000487, -0.00000253, -0.00000419, -0.00001995, 0.00000403, 0.00000112, 0.00001723, -0.00000237, 0.00001182, -0.00000146, 0.00000704, -0.00000497, -0.00000223, 0.00001125, -0.00000421, -0.00000438, 0.00000274, -0.00000260, -0.00000338, -0.00000138, -0.00000258, 0.00001069, -0.00001665, -0.00001130, 0.00000461, -0.00000550, -0.00000487, -0.00003690, -0.00006408, -0.00018808, 0.00015856, 0.00014019, 0.00003403, 0.00006358, 0.00003371, 0.00007436, 0.00004264, 0.00013819, 0.00013795, -0.00018798, -0.00004733, -0.00004540, -0.00001535, -0.00002304, -0.00000940, -0.00001912, 0.00000202, -0.00000863, -0.00000313, -0.00001213, -0.00000315, -0.00000595, -0.00000042, -0.00000926, 0.00000352, -0.00000633, 0.00000149, -0.00000526, 0.00000275, -0.00001045, -0.00000214, -0.00000515, 0.00000089, 0.00000008, 0.00000325, -0.00001275, -0.00000021, -0.00000555, 0.00000543, -0.00000897, 0.00000455, -0.00000247, -0.00001262, 0.00001397, -0.00001219, 0.00001535, -0.00000162, -0.00000323, 0.00000159, 0.00000247, 0.00000001, 0.00000349, -0.00000114, 0.00000208, -0.00000063, -0.00000195, -0.00000006, -0.00000213, -0.00000183, 0.00000084, -0.00000525, 0.00000904, 0.00000285, -0.00000059, -0.00000626, 0.00000819, 0.00000381, -0.00000178, -0.00000730, 0.00000814, -0.00000371, -0.00000198, -0.00000261, 0.00000194, -0.00000551, 0.00000165, 0.00000226, -0.00000589, 0.00000170, -0.00000034, 0.00000237, -0.00000967, -0.00000277, 0.00000316, 0.00000338, -0.00000536, -0.00000263, -0.00000048, -0.00000451, -0.00000067, -0.00000233, -0.00000024, -0.00000263, -0.00000050, 0.00000036, -0.00000272, -0.00000485, -0.00000718, },
		{ -0.00000773, 0.00000249, -0.00002361, 0.00000117, 0.00000003, 0.00001540, 0.00000194, 0.00000980, 0.00000193, 0.00000449, -0.00000013, -0.00000557, 0.00001149, -0.00000306, -0.00000508, 0.00000373, -0.00000215, -0.00000282, -0.00000316, -0.00000058, 0.00000939, -0.00001292, -0.00001670, 0.00001123, -0.00000960, 0.00000086, -0.00003822, -0.00005792, -0.00018542, 0.00014981, 0.00013088, 0.00004038, 0.00005155, 0.00004142, 0.00005399, 0.00005860, 0.00007582, 0.00019946, -0.00013721, -0.00009576, -0.00002811, -0.00003148, -0.00001232, -0.00001876, -0.00001346, -0.00000395, -0.00000265, -0.00000888, -0.00000645, -0.00000860, -0.00000231, -0.00000355, -0.00000660, 0.00000096, -0.00000277, -0.00000314, -0.00000125, -0.00000101, -0.00000449, -0.00000811, -0.00000174, -0.00000189, 0.00000154, 0.00000173, -0.00000735, -0.00000632, -0.00000094, -0.00000063, -0.00000100, -0.00000508, 0.00000856, -0.00002099, 0.00001574, -0.00001060, 0.00000975, 0.00000493, -0.00000560, 0.00000158, 0.00000275, -0.00000026, 0.00000344, -0.00000051, 0.00000137, 0.00000023, -0.00000198, -0.00000057, -0.00000077, -0.00000352, 0.00000236, -0.00000583, 0.00000617, 0.00000536, 0.00000051, -0.00000628, 0.00000429, 0.00000687, -0.00000079, -0.00000705, 0.00000390, 0.00000184, -0.00000533, -0.00000063, -0.00000072, -0.00000099, -0.00000430, 0.00000693, -0.00000807, 0.00000267, -0.00000219, 0.00000405, -0.00000842, -0.00000425, 0.00000074, 0.00000568, -0.00000470, -0.00000299, -0.00000084, -0.00000379, -0.00000262, 0.00000016, -0.00000314, 0.00000046, -0.00000450, 0.00000394, -0.00000469, -0.00000251, -0.00000963, -0.00000161, },
		{ 0.00000416, -0.00002069, -0.00000656, 0.00000208, 0.00001019, 0.00000814, 0.00000683, 0.00000605, 0.00000175, 0.00000391, -0.00000680, 0.00000985, -0.00000095, -0.00000627, 0.00000495, -0.00000222, -0.00000114, -0.00000579, 0.00000174, 0.00000750, -0.00000832, -0.00002211, 0.00001631, -0.00001149, 0.00000446, -0.00003789, -0.00005302, -0.00018259, 0.00014190, 0.00012215, 0.00004497, 0.00004247, 0.00004645, 0.00003835, 0.00006902, 0.00003536, 0.00020743, -0.00004741, -0.00014655, -0.00001563, -0.00004496, -0.00000416, -0.00002594, -0.00000911, -0.00000970, 0.00000316, -0.00001428, -0.00000134, -0.00001312, 0.00000053, -0.00000639, -0.00000372, -0.00000242, 0.00000180, -0.00000827, 0.00000286, -0.00000508, 0.00000124, -0.00001253, -0.00000004, -0.00000363, 0.00000286, -0.00000082, -0.00000125, -0.00001171, 0.00000271, -0.00000575, 0.00000506, -0.00001104, 0.00001334, -0.00002086, 0.00000939, -0.00000295, 0.00000074, 0.00001165, -0.00000678, 0.00000108, 0.00000269, 0.00000018, 0.00000260, 0.00000074, 0.00000029, 0.00000114, -0.00000196, -0.00000094, -0.00000003, -0.00000392, 0.00000193, -0.00000401, 0.00000180, 0.00000784, 0.00000108, -0.00000447, -0.00000060, 0.00000934, 0.00000006, -0.00000509, -0.00000142, 0.00000655, -0.00000705, 0.00000029, -0.00000288, 0.00000256, -0.00000784, 0.00000804, -0.00000686, 0.00000109, -0.00000176, 0.00000311, -0.00000528, -0.00000602, -0.00000120, 0.00000618, -0.00000257, -0.00000378, -0.00000109, -0.00000303, -0.00000448, 0.00000210, -0.00000505, 0.00000222, -0.00000658, 0.00000489, -0.00000417, -0.00000224, -0.00001008, -0.00000049, -0.00000985, },
		{ -0.00001324, -0.00001587, 0.00000413, 0.00000449, 0.00001332, 0.00000539, 0.00000876, 0.00000079, 0.00000558, -0.00000525, 0.00000662, 0.00000156, -0.00000741, 0.00000598, -0.00000239, 0.00000086, -0.00000802, 0.00000305, 0.00000649, -0.00000477, -0.00002519, 0.00001777, -0.00000972, 0.00000493, -0.00003498, -0.00005053, -0.00017784, 0.00013214, 0.00011742, 0.00004410, 0.00004007, 0.00004437, 0.00003281, 0.00006654, 0.00002246, 0.00017276, 0.00005369, -0.00017542, -0.00002109, -0.00004789, -0.00000440, -0.00002658, -0.00000959, -0.00001188, 0.00000545, -0.00001616, 0.00000029, -0.00001420, 0.00000048, -0.00000707, -0.00000253, -0.00000444, 0.00000481, -0.00001115, 0.00000442, -0.00000679, 0.00000398, -0.00001307, -0.00000165, -0.00000311, 0.00000271, -0.00000240, 0.00000282, -0.00001362, 0.00000289, -0.00000725, 0.00000633, -0.00001076, 0.00001048, -0.00001301, -0.00000214, 0.00000668, -0.00000747, 0.00001535, -0.00000520, -0.00000042, 0.00000264, 0.00000086, 0.00000158, 0.00000196, -0.00000053, 0.00000168, -0.00000172, -0.00000111, 0.00000001, -0.00000315, 0.00000014, -0.00000108, -0.00000234, 0.00000900, 0.00000198, -0.00000213, -0.00000461, 0.00000974, 0.00000174, -0.00000296, -0.00000548, 0.00000827, -0.00000574, -0.00000084, -0.00000329, 0.00000360, -0.00000766, 0.00000514, -0.00000264, -0.00000241, 0.00000035, 0.00000043, -0.00000140, -0.00000739, -0.00000267, 0.00000504, 0.00000043, -0.00000458, -0.00000123, -0.00000264, -0.00000548, 0.00000254, -0.00000503, 0.00000177, -0.00000600, 0.00000285, -0.00000119, -0.00000396, -0.00000843, -0.00000181, -0.00000858, 0.00000121, },
		{ -0.00002248, 0.00000297, 0.00000090, 0.00001513, 0.00000714, 0.00000886, 0.00000249, 0.00000458, -0.00000149, 0.00000286, 0.00000357, -0.00000795, 0.00000646, -0.00000228, 0.00000247, -0.00000897, 0.00000261, 0.00000701, -0.00000313, -0.00002516, 0.00001548, -0.00000496, 0.00000327, -0.00003035, -0.00004992, -0.00017122, 0.00012009, 0.00011711, 0.00003787, 0.00004363, 0.00003615, 0.00003590, 0.00005348, 0.00002912, 0.00011879, 0.00013577, -0.00016496, -0.00004876, -0.00003935, -0.00001349, -0.00002056, -0.00001504, -0.00000994, 0.00000341, -0.00001376, -0.00000214, -0.00001152, -0.00000260, -0.00000530, -0.00000359, -0.00000417, 0.00000504, -0.00001056, 0.00000247, -0.00000527, 0.00000297, -0.00000957, -0.00000614, -0.00000077, 0.00000108, -0.00000235, 0.00000381, -0.00001141, -0.00000058, -0.00000503, 0.00000301, -0.00000535, 0.00000262, -0.00000191, -0.00001334, 0.00001309, -0.00001090, 0.00001386, -0.00000041, -0.00000281, 0.00000271, 0.00000130, 0.00000098, 0.00000254, -0.00000063, 0.00000156, -0.00000113, -0.00000120, -0.00000034, -0.00000189, -0.00000185, 0.00000137, -0.00000466, 0.00000792, 0.00000375, -0.00000036, -0.00000643, 0.00000745, 0.00000450, -0.00000154, -0.00000704, 0.00000644, -0.00000174, -0.00000350, -0.00000216, 0.00000226, -0.00000450, -0.00000013, 0.00000249, -0.00000585, 0.00000241, -0.00000220, 0.00000153, -0.00000738, -0.00000411, 0.00000304, 0.00000322, -0.00000466, -0.00000148, -0.00000263, -0.00000552, 0.00000150, -0.00000335, -0.00000034, -0.00000361, -0.00000085, 0.00000265, -0.00000620, -0.00000603, -0.00000419, -0.00000565, -0.00000365, -0.00000524, },
		{ -0.00000268, 0.00000013, 0.00001309, 0.00001192, 0.00000683, 0.00000613, 0.00000204, 0.00000289, 0.00000005, 0.00000423, -0.00000753, 0.00000627, -0.00000173, 0.00000338, -0.00000842, 0.00000062, 0.00000868, -0.00000298, -0.00002286, 0.00001116, 0.00000052, 0.00000173, -0.00002594, -0.00004954, -0.00016414, 0.00010713, 0.00011920, 0.00002935, 0.00004914, 0.00002642, 0.00004174, 0.00003781, 0.00004136, 0.00006946, 0.00017837, -0.00011305, -0.00009046, -0.00002653, -0.00002620, -0.00001214, -0.00002217, -0.00000628, -0.00000101, -0.00000899, -0.00000672, -0.00000700, -0.00000692, -0.00000251, -0.00000589, -0.00000232, 0.00000308, -0.00000736, -0.00000180, -0.00000180, -0.00000035, -0.00000400, -0.00001128, 0.00000157, -0.00000077, -0.00000140, 0.00000253, -0.00000660, -0.00000569, -0.00000127, -0.00000224, 0.00000155, -0.00000542, 0.00000679, -0.00001899, 0.00001269, -0.00000792, 0.00000743, 0.00000631, -0.00000503, 0.00000262, 0.00000138, 0.00000106, 0.00000232, 0.00000005, 0.00000084, -0.00000030, -0.00000128, -0.00000072, -0.00000086, -0.00000304, 0.00000223, -0.00000452, 0.00000473, 0.00000615, 0.00000058, -0.00000590, 0.00000314, 0.00000755, -0.00000066, -0.00000636, 0.00000226, 0.00000312, -0.00000602, -0.00000083, -0.00000002, -0.00000044, -0.00000506, 0.00000589, -0.00000716, 0.00000272, -0.00000319, 0.00000225, -0.00000547, -0.00000582, 0.00000102, 0.00000480, -0.00000354, -0.00000203, -0.00000278, -0.00000501, -0.00000025, -0.00000119, -0.00000267, -0.00000116, -0.00000415, 0.00000525, -0.00000709, -0.00000463, -0.00000574, -0.00000338, -0.00000734, -0.00000037, -0.00002349, },
		{ 0.00000064, 0.00000864, 0.00001785, 0.00000454, 0.00000993, -0.00000011, 0.00000605, -0.00000057, 0.00000317, -0.00000623, 0.00000557, -0.00000086, 0.00000368, -0.00000676, -0.00000205, 0.00001043, -0.00000314, -0.00001999, 0.00000716, 0.00000421, 0.00000241, -0.00002339, -0.00004797, -0.00015809, 0.00009514, 0.00012092, 0.00002227, 0.00005223, 0.00001999, 0.00004441, 0.00002714, 0.00004762, 0.00003955, 0.00017774, -0.00003389, -0.00012847, -0.00002090, -0.00003494, -0.00000714, -0.00002652, -0.00000439, -0.00000467, -0.00000497, -0.00001044, -0.00000349, -0.00000997, -0.00000079, -0.00000766, -0.00000053, 0.00000072, -0.00000366, -0.00000598, 0.00000119, -0.00000344, 0.00000076, -0.00001437, 0.00000185, -0.00000136, -0.00000088, 0.00000072, -0.00000172, -0.00000959, 0.00000120, -0.00000617, 0.00000608, -0.00000942, 0.00000930, -0.00001690, 0.00000544, -0.00000022, -0.00000117, 0.00001223, -0.00000551, 0.00000181, 0.00000127, 0.00000158, 0.00000160, 0.00000112, -0.00000006, 0.00000048, -0.00000127, -0.00000093, -0.00000044, -0.00000305, 0.00000140, -0.00000244, 0.00000060, 0.00000817, 0.00000121, -0.00000391, -0.00000158, 0.00000945, 0.00000057, -0.00000464, -0.00000219, 0.00000655, -0.00000662, -0.00000068, -0.00000159, 0.00000236, -0.00000739, 0.00000586, -0.00000539, 0.00000073, -0.00000211, 0.00000069, -0.00000207, -0.00000755, -0.00000066, 0.00000483, -0.00000132, -0.00000276, -0.00000291, -0.00000451, -0.00000175, 0.00000022, -0.00000382, -0.00000015, -0.00000547, 0.00000529, -0.00000566, -0.00000513, -0.00000540, -0.00000316, -0.00000811, -0.00000010, -0.00001891, -0.00001130, },
		{ 0.00000402, 0.00002248, 0.00000409, 0.00001210, -0.00000029, 0.00000682, 0.00000132, 0.00000079, -0.00000457, 0.00000481, -0.00000001, 0.00000372, -0.00000472, -0.00000432, 0.00001116, -0.00000249, -0.00001802, 0.00000522, 0.00000472, 0.00000600, -0.00002298, -0.00004492, -0.00015358, 0.00008513, 0.00012049, 0.00001888, 0.00005059, 0.00001904, 0.00004137, 0.00002455, 0.00004368, 0.00003111, 0.00014569, 0.00004836, -0.00014376, -0.00003140, -0.00003478, -0.00000916, -0.00002547, -0.00000642, -0.00000541, -0.00000395, -0.00001114, -0.00000296, -0.00001009, -0.00000145, -0.00000766, -0.00000017, -0.00000042, -0.00000147, -0.00000794, 0.00000173, -0.00000429, 0.00000266, -0.00001382, -0.00000079, -0.00000008, -0.00000156, -0.00000023, 0.00000122, -0.00001034, 0.00000066, -0.00000693, 0.00000630, -0.00000790, 0.00000535, -0.00000879, -0.00000520, 0.00000801, -0.00000795, 0.00001453, -0.00000304, -0.00000000, 0.00000124, 0.00000212, 0.00000093, 0.00000200, -0.00000061, 0.00000083, -0.00000102, -0.00000100, -0.00000058, -0.00000217, -0.00000037, 0.00000023, -0.00000285, 0.00000866, 0.00000229, -0.00000173, -0.00000496, 0.00000898, 0.00000282, -0.00000309, -0.00000509, 0.00000706, -0.00000459, -0.00000216, -0.00000173, 0.00000289, -0.00000644, 0.00000254, -0.00000134, -0.00000262, 0.00000012, -0.00000202, 0.00000149, -0.00000849, -0.00000216, 0.00000363, 0.00000131, -0.00000319, -0.00000306, -0.00000432, -0.00000244, 0.00000024, -0.00000326, -0.00000100, -0.00000459, 0.00000293, -0.00000243, -0.00000705, -0.00000353, -0.00000475, -0.00000633, -0.00000330, -0.00001141, -0.00001940, -0.00000022, },
		{ 0.00002416, 0.00000657, 0.00001194, 0.00000192, 0.00000533, 0.00000492, -0.00000183, -0.00000338, 0.00000448, 0.00000049, 0.00000389, -0.00000299, -0.00000539, 0.00001034, -0.00000064, -0.00001747, 0.00000572, 0.00000239, 0.00001148, -0.00002347, -0.00004140, -0.00014997, 0.00007681, 0.00011785, 0.00001913, 0.00004492, 0.00002229, 0.00003413, 0.00002795, 0.00003255, 0.00003631, 0.00010198, 0.00011014, -0.00012506, -0.00005864, -0.00002688, -0.00001726, -0.00001992, -0.00001170, -0.00000338, -0.00000605, -0.00000875, -0.00000538, -0.00000749, -0.00000421, -0.00000600, -0.00000144, 0.00000020, -0.00000156, -0.00000700, -0.00000058, -0.00000261, 0.00000151, -0.00001002, -0.00000545, 0.00000224, -0.00000307, -0.00000006, 0.00000172, -0.00000788, -0.00000255, -0.00000488, 0.00000290, -0.00000247, -0.00000201, 0.00000084, -0.00001416, 0.00001232, -0.00000972, 0.00001181, 0.00000225, -0.00000236, 0.00000124, 0.00000235, 0.00000075, 0.00000227, -0.00000049, 0.00000059, -0.00000047, -0.00000104, -0.00000095, -0.00000107, -0.00000203, 0.00000210, -0.00000437, 0.00000704, 0.00000415, -0.00000022, -0.00000606, 0.00000594, 0.00000595, -0.00000210, -0.00000583, 0.00000474, -0.00000070, -0.00000447, -0.00000089, 0.00000158, -0.00000321, -0.00000227, 0.00000288, -0.00000542, 0.00000183, -0.00000417, 0.00000367, -0.00000781, -0.00000384, 0.00000192, 0.00000341, -0.00000280, -0.00000334, -0.00000445, -0.00000232, -0.00000093, -0.00000150, -0.00000294, -0.00000253, -0.00000039, 0.00000099, -0.00000887, -0.00000154, -0.00000667, -0.00000379, -0.00000730, -0.00000470, -0.00002353, -0.00000442, 0.00000078, },
		{ 0.00001148, 0.00001024, 0.00000558, 0.00000286, 0.00000866, -0.00000331, -0.00000339, 0.00000491, 0.00000045, 0.00000443, -0.00000189, -0.00000511, 0.00000824, 0.00000194, -0.00001785, 0.00000777, -0.00000114, 0.00001678, -0.00002295, -0.00003891, -0.00014606, 0.00006915, 0.00011414, 0.00002123, 0.00003788, 0.00002637, 0.00002651, 0.00003255, 0.00002069, 0.00004406, 0.00006487, 0.00013770, -0.00007431, -0.00009339, -0.00001832, -0.00002654, -0.00001377, -0.00001728, -0.00000075, -0.00000944, -0.00000508, -0.00000896, -0.00000396, -0.00000743, -0.00000396, -0.00000341, 0.00000187, -0.00000323, -0.00000415, -0.00000443, 0.00000022, -0.00000116, -0.00000499, -0.00000995, 0.00000387, -0.00000427, 0.00000048, 0.00000066, -0.00000374, -0.00000651, -0.00000203, -0.00000162, 0.00000347, -0.00000849, 0.00000725, -0.00001742, 0.00001029, -0.00000586, 0.00000503, 0.00000864, -0.00000403, 0.00000086, 0.00000228, 0.00000109, 0.00000195, 0.00000021, -0.00000009, 0.00000024, -0.00000105, -0.00000127, -0.00000031, -0.00000282, 0.00000244, -0.00000370, 0.00000374, 0.00000634, 0.00000058, -0.00000504, 0.00000136, 0.00000885, -0.00000117, -0.00000501, 0.00000100, 0.00000319, -0.00000606, -0.00000023, -0.00000024, 0.00000033, -0.00000617, 0.00000509, -0.00000602, 0.00000171, -0.00000456, 0.00000364, -0.00000535, -0.00000572, 0.00000026, 0.00000432, -0.00000136, -0.00000379, -0.00000475, -0.00000181, -0.00000252, 0.00000036, -0.00000467, -0.00000083, -0.00000291, 0.00000287, -0.00000914, -0.00000078, -0.00000746, -0.00000234, -0.00000967, -0.00000150, -0.00002216, -0.00001182, -0.00000127, 0.00002283, },
		{ 0.00000872, 0.00000902, 0.00000113, 0.00001097, -0.00000273, -0.00000483, 0.00000605, -0.00000002, 0.00000526, -0.00000138, -0.00000387, 0.00000570, 0.00000428, -0.00001814, 0.00000996, -0.00000393, 0.00001996, -0.00002002, -0.00003837, -0.00014111, 0.00006124, 0.00011062, 0.00002313, 0.00003225, 0.00002809, 0.00002191, 0.00003418, 0.00001347, 0.00004621, 0.00004419, 0.00013119, -0.00000585, -0.00012030, -0.00001809, -0.00003134, -0.00001129, -0.00001994, -0.00000005, -0.00001178, -0.00000248, -0.00001140, -0.00000166, -0.00000923, -0.00000305, -0.00000475, 0.00000329, -0.00000499, -0.00000122, -0.00000774, 0.00000218, -0.00000322, -0.00000115, -0.00001209, 0.00000331, -0.00000411, 0.00000043, -0.00000057, 0.00000001, -0.00000901, -0.00000059, -0.00000468, 0.00000685, -0.00001086, 0.00000787, -0.00001399, 0.00000284, 0.00000139, -0.00000281, 0.00001343, -0.00000356, -0.00000034, 0.00000214, 0.00000165, 0.00000141, 0.00000109, -0.00000079, 0.00000081, -0.00000096, -0.00000142, -0.00000010, -0.00000259, 0.00000142, -0.00000158, 0.00000001, 0.00000786, 0.00000125, -0.00000294, -0.00000307, 0.00001009, 0.00000053, -0.00000376, -0.00000233, 0.00000530, -0.00000571, -0.00000067, -0.00000138, 0.00000252, -0.00000751, 0.00000420, -0.00000399, -0.00000037, -0.00000317, 0.00000171, -0.00000185, -0.00000731, -0.00000124, 0.00000396, 0.00000079, -0.00000414, -0.00000512, -0.00000134, -0.00000374, 0.00000134, -0.00000516, -0.00000055, -0.00000358, 0.00000242, -0.00000740, -0.00000166, -0.00000658, -0.00000269, -0.00000950, -0.00000228, -0.00001641, -0.00002011, -0.00000373, 0.00001967, 0.00001712, },
		{ 0.00001089, 0.00000133, 0.00001110, -0.00000006, -0.00000720, 0.00000737, -0.00000054, 0.00000612, -0.00000112, -0.00000235, 0.00000370, 0.00000551, -0.00001754, 0.00001121, -0.00000476, 0.00002026, -0.00001458, -0.00003948, -0.00013534, 0.00005301, 0.00010786, 0.00002368, 0.00002948, 0.00002618, 0.00002140, 0.00003156, 0.00001240, 0.00004068, 0.00003974, 0.00010221, 0.00005999, -0.00012510, -0.00003163, -0.00002909, -0.00001437, -0.00001837, -0.00000236, -0.00001181, -0.00000236, -0.00001139, -0.00000179, -0.00000865, -0.00000400, -0.00000474, 0.00000355, -0.00000565, 0.00000029, -0.00000893, 0.00000185, -0.00000329, 0.00000009, -0.00001095, 0.00000028, -0.00000253, -0.00000050, -0.00000110, 0.00000204, -0.00000884, -0.00000155, -0.00000510, 0.00000648, -0.00000853, 0.00000333, -0.00000623, -0.00000641, 0.00000803, -0.00000809, 0.00001428, -0.00000019, -0.00000240, 0.00000205, 0.00000206, 0.00000111, 0.00000168, -0.00000110, 0.00000096, -0.00000064, -0.00000148, -0.00000031, -0.00000175, -0.00000019, 0.00000069, -0.00000272, 0.00000778, 0.00000242, -0.00000091, -0.00000582, 0.00000877, 0.00000336, -0.00000286, -0.00000407, 0.00000484, -0.00000331, -0.00000220, -0.00000148, 0.00000279, -0.00000608, 0.00000078, -0.00000040, -0.00000329, -0.00000110, -0.00000082, 0.00000130, -0.00000777, -0.00000282, 0.00000276, 0.00000295, -0.00000398, -0.00000557, -0.00000117, -0.00000421, 0.00000106, -0.00000420, -0.00000173, -0.00000252, 0.00000017, -0.00000447, -0.00000341, -0.00000471, -0.00000422, -0.00000757, -0.00000544, -0.00000928, -0.00002602, -0.00000841, 0.00001606, 0.00002162, 0.00000903, },
		{ 0.00000350, 0.00000948, 0.00000376, -0.00000943, 0.00000817, -0.00000067, 0.00000666, -0.00000068, -0.00000122, 0.00000290, 0.00000526, -0.00001589, 0.00001128, -0.00000358, 0.00001829, -0.00000787, -0.00004093, -0.00012977, 0.00004508, 0.00010555, 0.00002295, 0.00002924, 0.00002158, 0.00002353, 0.00002635, 0.00001530, 0.00003055, 0.00004445, 0.00006702, 0.00010545, -0.00010139, -0.00005698, -0.00002225, -0.00002124, -0.00001404, -0.00000661, -0.00001010, -0.00000439, -0.00000927, -0.00000396, -0.00000620, -0.00000630, -0.00000367, 0.00000264, -0.00000495, -0.00000006, -0.00000773, -0.00000073, -0.00000148, -0.00000107, -0.00000729, -0.00000401, -0.00000057, -0.00000174, -0.00000096, 0.00000223, -0.00000630, -0.00000432, -0.00000351, 0.00000336, -0.00000339, -0.00000325, 0.00000168, -0.00001309, 0.00001048, -0.00000851, 0.00001051, 0.00000547, -0.00000458, 0.00000185, 0.00000214, 0.00000126, 0.00000178, -0.00000087, 0.00000062, -0.00000008, -0.00000151, -0.00000063, -0.00000088, -0.00000146, 0.00000202, -0.00000358, 0.00000584, 0.00000420, 0.00000042, -0.00000628, 0.00000511, 0.00000675, -0.00000226, -0.00000415, 0.00000241, 0.00000004, -0.00000387, -0.00000110, 0.00000171, -0.00000305, -0.00000337, 0.00000281, -0.00000535, 0.00000014, -0.00000245, 0.00000282, -0.00000651, -0.00000468, 0.00000129, 0.00000439, -0.00000299, -0.00000610, -0.00000132, -0.00000399, -0.00000014, -0.00000244, -0.00000352, -0.00000083, -0.00000247, -0.00000186, -0.00000466, -0.00000316, -0.00000557, -0.00000557, -0.00000857, -0.00000396, -0.00002718, -0.00001588, 0.00001249, 0.00002390, 0.00001182, 0.00001090, },
		{ 0.00000738, 0.00000731, -0.00001041, 0.00000787, -0.00000013, 0.00000669, 0.00000019, -0.00000079, 0.00000341, 0.00000391, -0.00001372, 0.00001072, -0.00000127, 0.00001560, -0.00000172, -0.00004117, -0.00012550, 0.00003823, 0.00010303, 0.00002176, 0.00003011, 0.00001645, 0.00002562, 0.00002150, 0.00001844, 0.00002084, 0.00004978, 0.00003914, 0.00012201, -0.00005378, -0.00008479, -0.00001721, -0.00002770, -0.00001023, -0.00001034, -0.00000846, -0.00000698, -0.00000661, -0.00000661, -0.00000341, -0.00000855, -0.00000260, 0.00000136, -0.00000353, -0.00000155, -0.00000516, -0.00000423, 0.00000087, -0.00000314, -0.00000305, -0.00000753, 0.00000023, -0.00000231, -0.00000084, 0.00000144, -0.00000284, -0.00000720, -0.00000163, -0.00000031, 0.00000156, -0.00000820, 0.00000596, -0.00001433, 0.00000740, -0.00000423, 0.00000364, 0.00001139, -0.00000558, 0.00000109, 0.00000200, 0.00000174, 0.00000154, -0.00000027, 0.00000001, 0.00000055, -0.00000151, -0.00000086, -0.00000039, -0.00000190, 0.00000199, -0.00000260, 0.00000268, 0.00000602, 0.00000117, -0.00000490, 0.00000041, 0.00000939, -0.00000121, -0.00000339, -0.00000055, 0.00000273, -0.00000446, -0.00000111, 0.00000036, -0.00000008, -0.00000628, 0.00000396, -0.00000534, -0.00000039, -0.00000235, 0.00000224, -0.00000370, -0.00000658, -0.00000018, 0.00000475, -0.00000122, -0.00000658, -0.00000171, -0.00000346, -0.00000156, -0.00000085, -0.00000483, 0.00000027, -0.00000412, -0.00000086, -0.00000438, -0.00000286, -0.00000567, -0.00000480, -0.00000993, -0.00000217, -0.00002338, -0.00002476, 0.00000824, 0.00002397, 0.00001653, 0.00000992, 0.00000663, },
		{ 0.00000938, -0.00000953, 0.00000641, 0.00000099, 0.00000627, 0.00000144, -0.00000099, 0.00000477, 0.00000229, -0.00001194, 0.00001039, 0.00000100, 0.00001374, 0.00000247, -0.00003934, -0.00012293, 0.00003271, 0.00009989, 0.00002088, 0.00003065, 0.00001278, 0.00002553, 0.00001926, 0.00001904, 0.00001513, 0.00005024, 0.00002475, 0.00011224, 0.00000422, -0.00010236, -0.00002058, -0.00002987, -0.00000986, -0.00001138, -0.00000859, -0.00000848, -0.00000512, -0.00000812, -0.00000184, -0.00000943, -0.00000257, 0.00000060, -0.00000238, -0.00000300, -0.00000274, -0.00000691, 0.00000210, -0.00000442, -0.00000019, -0.00000860, -0.00000111, -0.00000160, -0.00000130, 0.00000066, 0.00000003, -0.00000859, -0.00000102, -0.00000260, 0.00000398, -0.00000928, 0.00000514, -0.00001016, 0.00000035, 0.00000219, -0.00000332, 0.00001507, -0.00000418, -0.00000051, 0.00000185, 0.00000221, 0.00000134, 0.00000032, -0.00000050, 0.00000097, -0.00000139, -0.00000095, -0.00000037, -0.00000154, 0.00000093, -0.00000065, -0.00000049, 0.00000698, 0.00000191, -0.00000276, -0.00000365, 0.00000995, 0.00000097, -0.00000270, -0.00000265, 0.00000354, -0.00000334, -0.00000194, -0.00000047, 0.00000162, -0.00000685, 0.00000252, -0.00000325, -0.00000247, -0.00000084, 0.00000022, -0.00000032, -0.00000785, -0.00000171, 0.00000419, 0.00000088, -0.00000671, -0.00000227, -0.00000297, -0.00000258, -0.00000015, -0.00000497, 0.00000012, -0.00000418, -0.00000178, -0.00000255, -0.00000383, -0.00000446, -0.00000545, -0.00000928, -0.00000357, -0.00001661, -0.00003230, 0.00000205, 0.00002241, 0.00002183, 0.00000937, 0.00000935, 0.00000622, },
		{ -0.00000700, 0.00000435, 0.00000223, 0.00000575, 0.00000280, -0.00000144, 0.00000629, 0.00000132, -0.00001128, 0.00001087, 0.00000246, 0.00001353, 0.00000437, -0.00003575, -0.00012145, 0.00002802, 0.00009631, 0.00002055, 0.00003015, 0.00001138, 0.00002269, 0.00001993, 0.00001667, 0.00001394, 0.00004509, 0.00002241, 0.00008683, 0.00005604, -0.00009960, -0.00003495, -0.00002694, -0.00001355, -0.00000939, -0.00001083, -0.00000831, -0.00000551, -0.00000781, -0.00000210, -0.00000852, -0.00000385, 0.00000076, -0.00000209, -0.00000352, -0.00000161, -0.00000766, 0.00000133, -0.00000402, 0.00000041, -0.00000682, -0.00000439, 0.00000004, -0.00000229, 0.00000035, 0.00000149, -0.00000789, -0.00000218, -0.00000283, 0.00000329, -0.00000654, 0.00000047, -0.00000322, -0.00000718, 0.00000720, -0.00000730, 0.00001472, -0.00000008, -0.00000274, 0.00000171, 0.00000240, 0.00000148, 0.00000062, -0.00000065, 0.00000102, -0.00000107, -0.00000097, -0.00000062, -0.00000081, -0.00000039, 0.00000114, -0.00000249, 0.00000640, 0.00000312, -0.00000088, -0.00000577, 0.00000787, 0.00000426, -0.00000240, -0.00000332, 0.00000237, -0.00000092, -0.00000323, -0.00000073, 0.00000182, -0.00000524, -0.00000069, -0.00000025, -0.00000493, 0.00000092, -0.00000192, 0.00000226, -0.00000770, -0.00000355, 0.00000313, 0.00000266, -0.00000616, -0.00000295, -0.00000275, -0.00000294, -0.00000050, -0.00000400, -0.00000106, -0.00000303, -0.00000390, -0.00000011, -0.00000517, -0.00000278, -0.00000667, -0.00000766, -0.00000633, -0.00000992, -0.00003583, -0.00000672, 0.00001958, 0.00002645, 0.00001046, 0.00001072, 0.00000677, 0.00000962, },
		{ 0.00000264, 0.00000299, 0.00000552, 0.00000393, -0.00000170, 0.00000734, 0.00000155, -0.00001190, 0.00001212, 0.00000310, 0.00001480, 0.00000475, -0.00003167, -0.00011985, 0.00002327, 0.00009284, 0.00002050, 0.00002874, 0.00001181, 0.00001811, 0.00002213, 0.00001285, 0.00001529, 0.00003711, 0.00002609, 0.00005873, 0.00008885, -0.00007383, -0.00005673, -0.00002202, -0.00001921, -0.00000602, -0.00001396, -0.00000721, -0.00000724, -0.00000622, -0.00000362, -0.00000644, -0.00000584, 0.00000146, -0.00000255, -0.00000303, -0.00000195, -0.00000649, -0.00000112, -0.00000234, -0.00000080, -0.00000319, -0.00000823, 0.00000146, -0.00000313, 0.00000027, 0.00000171, -0.00000565, -0.00000435, -0.00000173, 0.00000060, -0.00000196, -0.00000508, 0.00000281, -0.00001164, 0.00000812, -0.00000681, 0.00001032, 0.00000569, -0.00000465, 0.00000132, 0.00000226, 0.00000200, 0.00000060, -0.00000043, 0.00000072, -0.00000059, -0.00000097, -0.00000089, -0.00000020, -0.00000127, 0.00000195, -0.00000278, 0.00000428, 0.00000473, 0.00000030, -0.00000569, 0.00000376, 0.00000773, -0.00000201, -0.00000298, 0.00000015, 0.00000159, -0.00000406, -0.00000091, 0.00000108, -0.00000260, -0.00000401, 0.00000200, -0.00000636, 0.00000166, -0.00000293, 0.00000307, -0.00000583, -0.00000568, 0.00000192, 0.00000365, -0.00000486, -0.00000363, -0.00000285, -0.00000275, -0.00000151, -0.00000259, -0.00000245, -0.00000167, -0.00000598, 0.00000166, -0.00000575, -0.00000171, -0.00000730, -0.00000648, -0.00000844, -0.00000578, -0.00003422, -0.00001716, 0.00001505, 0.00002969, 0.00001357, 0.00001072, 0.00000883, 0.00000862, -0.00000379, },
		{ 0.00000289, 0.00000579, 0.00000462, -0.00000151, 0.00000763, 0.00000298, -0.00001335, 0.00001353, 0.00000349, 0.00001671, 0.00000500, -0.00002860, -0.00011698, 0.00001775, 0.00008992, 0.00002035, 0.00002701, 0.00001296, 0.00001347, 0.00002388, 0.00000973, 0.00001643, 0.00003001, 0.00002948, 0.00003763, 0.00009797, -0.00003112, -0.00007716, -0.00002047, -0.00002352, -0.00000384, -0.00001607, -0.00000661, -0.00000900, -0.00000466, -0.00000516, -0.00000442, -0.00000746, 0.00000192, -0.00000319, -0.00000206, -0.00000309, -0.00000439, -0.00000398, -0.00000062, -0.00000245, 0.00000050, -0.00001084, 0.00000142, -0.00000305, -0.00000009, 0.00000143, -0.00000304, -0.00000622, -0.00000063, -0.00000227, 0.00000196, -0.00000858, 0.00000518, -0.00001128, 0.00000451, -0.00000261, 0.00000372, 0.00001099, -0.00000499, 0.00000026, 0.00000198, 0.00000263, 0.00000052, -0.00000007, 0.00000032, -0.00000013, -0.00000093, -0.00000105, 0.00000004, -0.00000142, 0.00000163, -0.00000164, 0.00000138, 0.00000615, 0.00000103, -0.00000405, -0.00000087, 0.00000991, -0.00000066, -0.00000251, -0.00000179, 0.00000294, -0.00000367, -0.00000151, 0.00000016, -0.00000023, -0.00000598, 0.00000234, -0.00000599, 0.00000082, -0.00000235, 0.00000202, -0.00000272, -0.00000763, 0.00000059, 0.00000377, -0.00000306, -0.00000405, -0.00000325, -0.00000229, -0.00000259, -0.00000153, -0.00000321, -0.00000099, -0.00000704, 0.00000191, -0.00000497, -0.00000179, -0.00000673, -0.00000651, -0.00000885, -0.00000495, -0.00002853, -0.00002687, 0.00000792, 0.00003128, 0.00001831, 0.00001011, 0.00001151, 0.00000750, -0.00000105, 0.00000208, },
		{ 0.00000644, 0.00000494, -0.00000091, 0.00000729, 0.00000510, -0.00001476, 0.00001428, 0.00000432, 0.00001833, 0.00000631, -0.00002749, -0.00011250, 0.00001150, 0.00008747, 0.00002000, 0.00002538, 0.00001385, 0.00001009, 0.00002383, 0.00000862, 0.00001568, 0.00002604, 0.00002905, 0.00002717, 0.00008732, 0.00001641, -0.00008625, -0.00002666, -0.00002414, -0.00000461, -0.00001590, -0.00000750, -0.00000977, -0.00000421, -0.00000568, -0.00000347, -0.00000789, 0.00000153, -0.00000343, -0.00000132, -0.00000410, -0.00000261, -0.00000587, -0.00000010, -0.00000327, 0.00000277, -0.00001101, -0.00000063, -0.00000188, -0.00000098, 0.00000124, -0.00000110, -0.00000677, -0.00000055, -0.00000396, 0.00000356, -0.00000861, 0.00000333, -0.00000685, -0.00000170, 0.00000268, -0.00000220, 0.00001360, -0.00000283, -0.00000158, 0.00000172, 0.00000304, 0.00000069, 0.00000016, 0.00000008, 0.00000011, -0.00000077, -0.00000109, -0.00000007, -0.00000099, 0.00000061, 0.00000006, -0.00000120, 0.00000660, 0.00000182, -0.00000196, -0.00000442, 0.00000966, 0.00000210, -0.00000250, -0.00000263, 0.00000260, -0.00000203, -0.00000248, -0.00000054, 0.00000109, -0.00000597, 0.00000068, -0.00000412, -0.00000118, -0.00000073, -0.00000004, 0.00000049, -0.00000860, -0.00000105, 0.00000330, -0.00000130, -0.00000390, -0.00000390, -0.00000189, -0.00000328, -0.00000130, -0.00000298, -0.00000129, -0.00000684, 0.00000070, -0.00000313, -0.00000273, -0.00000524, -0.00000751, -0.00000787, -0.00000644, -0.00002138, -0.00003306, -0.00000232, 0.00003092, 0.00002389, 0.00000994, 0.00001379, 0.00000730, 0.00000041, 0.00000299, 0.00000203, },
		{ 0.00000517, -0.00000019, 0.00000676, 0.00000720, -0.00001531, 0.00001380, 0.00000596, 0.00001921, 0.00000914, -0.00002829, -0.00010701, 0.00000527, 0.00008492, 0.00001971, 0.00002394, 0.00001406, 0.00000837, 0.00002191, 0.00000942, 0.00001303, 0.00002517, 0.00002478, 0.00002534, 0.00006620, 0.00005580, -0.00007746, -0.00004101, -0.00002154, -0.00000808, -0.00001373, -0.00000971, -0.00000942, -0.00000509, -0.00000498, -0.00000377, -0.00000706, 0.00000030, -0.00000314, -0.00000116, -0.00000437, -0.00000195, -0.00000603, -0.00000127, -0.00000279, 0.00000312, -0.00000878, -0.00000409, -0.00000024, -0.00000204, 0.00000125, -0.00000017, -0.00000587, -0.00000156, -0.00000417, 0.00000269, -0.00000572, -0.00000113, -0.00000101, -0.00000738, 0.00000609, -0.00000501, 0.00001231, 0.00000170, -0.00000375, 0.00000141, 0.00000305, 0.00000127, 0.00000014, 0.00000011, 0.00000007, -0.00000047, -0.00000107, -0.00000033, -0.00000040, -0.00000042, 0.00000139, -0.00000257, 0.00000565, 0.00000300, -0.00000027, -0.00000587, 0.00000679, 0.00000587, -0.00000275, -0.00000242, 0.00000106, 0.00000007, -0.00000325, -0.00000113, 0.00000133, -0.00000438, -0.00000202, -0.00000188, -0.00000322, 0.00000080, -0.00000181, 0.00000257, -0.00000796, -0.00000316, 0.00000259, -0.00000008, -0.00000303, -0.00000466, -0.00000173, -0.00000340, -0.00000189, -0.00000199, -0.00000219, -0.00000591, -0.00000119, -0.00000117, -0.00000366, -0.00000368, -0.00000855, -0.00000662, -0.00000839, -0.00001558, -0.00003397, -0.00001484, 0.00002784, 0.00002960, 0.00001103, 0.00001509, 0.00000845, 0.00000054, 0.00000498, 0.00000094, 0.00000711, },
		{ 0.00000032, 0.00000647, 0.00000869, -0.00001456, 0.00001209, 0.00000818, 0.00001950, 0.00001316, -0.00003012, -0.00010178, 0.00000013, 0.00008158, 0.00001984, 0.00002252, 0.00001373, 0.00000785, 0.00001902, 0.00001096, 0.00000973, 0.00002579, 0.00001898, 0.00002733, 0.00004458, 0.00007825, -0.00005084, -0.00005905, -0.00001894, -0.00001221, -0.00001117, -0.00001207, -0.00000876, -0.00000666, -0.00000369, -0.00000471, -0.00000561, -0.00000118, -0.00000265, -0.00000150, -0.00000388, -0.00000244, -0.00000470, -0.00000358, -0.00000155, 0.00000217, -0.00000527, -0.00000753, 0.00000071, -0.00000259, 0.00000113, 0.00000009, -0.00000415, -0.00000295, -0.00000357, 0.00000042, -0.00000178, -0.00000556, 0.00000322, -0.00000990, 0.00000581, -0.00000393, 0.00000762, 0.00000724, -0.00000519, 0.00000071, 0.00000276, 0.00000209, 0.00000004, 0.00000030, -0.00000014, -0.00000011, -0.00000104, -0.00000056, 0.00000001, -0.00000100, 0.00000182, -0.00000249, 0.00000352, 0.00000440, 0.00000075, -0.00000524, 0.00000227, 0.00000934, -0.00000242, -0.00000196, -0.00000057, 0.00000148, -0.00000312, -0.00000190, 0.00000092, -0.00000218, -0.00000444, -0.00000055, -0.00000424, 0.00000125, -0.00000235, 0.00000284, -0.00000568, -0.00000551, 0.00000176, 0.00000040, -0.00000156, -0.00000528, -0.00000194, -0.00000304, -0.00000294, -0.00000085, -0.00000299, -0.00000505, -0.00000278, -0.00000004, -0.00000375, -0.00000286, -0.00000877, -0.00000612, -0.00000930, -0.00001275, -0.00002986, -0.00002727, 0.00002093, 0.00003479, 0.00001372, 0.00001547, 0.00001068, -0.00000005, 0.00000721, 0.00000034, 0.00000736, 0.00000563, },
		{ 0.00000665, 0.00000937, -0.00001266, 0.00000975, 0.00001028, 0.00001978, 0.00001759, -0.00003176, -0.00009801, -0.00000317, 0.00007715, 0.00002053, 0.00002100, 0.00001318, 0.00000772, 0.00001637, 0.00001191, 0.00000716, 0.00002605, 0.00001418, 0.00002867, 0.00002914, 0.00008185, -0.00001303, -0.00007309, -0.00002049, -0.00001458, -0.00001005, -0.00001322, -0.00000879, -0.00000796, -0.00000277, -0.00000541, -0.00000443, -0.00000217, -0.00000247, -0.00000193, -0.00000304, -0.00000351, -0.00000278, -0.00000590, -0.00000063, 0.00000107, -0.00000201, -0.00000944, 0.00000002, -0.00000212, 0.00000056, 0.00000019, -0.00000247, -0.00000382, -0.00000307, -0.00000185, 0.00000124, -0.00000779, 0.00000407, -0.00000845, 0.00000213, -0.00000013, 0.00000159, 0.00001169, -0.00000475, -0.00000069, 0.00000239, 0.00000279, 0.00000015, 0.00000043, -0.00000032, 0.00000016, -0.00000094, -0.00000069, 0.00000010, -0.00000099, 0.00000139, -0.00000137, 0.00000100, 0.00000542, 0.00000140, -0.00000335, -0.00000227, 0.00001100, -0.00000063, -0.00000205, -0.00000128, 0.00000147, -0.00000185, -0.00000291, 0.00000024, -0.00000029, -0.00000557, -0.00000078, -0.00000385, 0.00000037, -0.00000152, 0.00000154, -0.00000252, -0.00000738, 0.00000063, 0.00000029, 0.00000006, -0.00000539, -0.00000255, -0.00000244, -0.00000398, -0.00000014, -0.00000316, -0.00000479, -0.00000348, -0.00000017, -0.00000281, -0.00000297, -0.00000798, -0.00000665, -0.00000881, -0.00001273, -0.00002291, -0.00003651, 0.00000951, 0.00003857, 0.00001803, 0.00001542, 0.00001333, -0.00000051, 0.00000887, 0.00000073, 0.00000703, 0.00000645, 0.00000045, },
		{ 0.00000947, -0.00001019, 0.00000768, 0.00001149, 0.00002062, 0.00002172, -0.00003233, -0.00009629, -0.00000465, 0.00007197, 0.00002145, 0.00001948, 0.00001261, 0.00000738, 0.00001473, 0.00001159, 0.00000598, 0.00002501, 0.00001169, 0.00002722, 0.00002177, 0.00007100, 0.00002567, -0.00007573, -0.00002862, -0.00001417, -0.00001119, -0.00001261, -0.00000993, -0.00000845, -0.00000277, -0.00000534, -0.00000403, -0.00000227, -0.00000286, -0.00000218, -0.00000233, -0.00000443, -0.00000126, -0.00000717, -0.00000089, 0.00000072, -0.00000012, -0.00000905, -0.00000243, -0.00000081, -0.00000039, 0.00000039, -0.00000137, -0.00000373, -0.00000318, -0.00000325, 0.00000227, -0.00000712, 0.00000170, -0.00000427, -0.00000295, 0.00000394, -0.00000322, 0.00001317, -0.00000184, -0.00000275, 0.00000206, 0.00000307, 0.00000071, 0.00000034, -0.00000032, 0.00000024, -0.00000074, -0.00000073, -0.00000005, -0.00000063, 0.00000056, -0.00000002, -0.00000099, 0.00000547, 0.00000215, -0.00000127, -0.00000532, 0.00000995, 0.00000280, -0.00000284, -0.00000087, 0.00000021, -0.00000000, -0.00000372, -0.00000061, 0.00000089, -0.00000520, -0.00000235, -0.00000252, -0.00000128, -0.00000000, -0.00000031, 0.00000030, -0.00000797, -0.00000108, -0.00000005, 0.00000130, -0.00000475, -0.00000349, -0.00000185, -0.00000468, -0.00000012, -0.00000259, -0.00000516, -0.00000335, -0.00000126, -0.00000129, -0.00000352, -0.00000668, -0.00000763, -0.00000761, -0.00001405, -0.00001614, -0.00004009, -0.00000570, 0.00003947, 0.00002384, 0.00001556, 0.00001579, -0.00000020, 0.00000957, 0.00000211, 0.00000631, 0.00000753, 0.00000025, 0.00000721, },
		{ -0.00000789, 0.00000661, 0.00001142, 0.00002216, 0.00002535, -0.00003162, -0.00009634, -0.00000512, 0.00006690, 0.00002200, 0.00001826, 0.00001202, 0.00000662, 0.00001419, 0.00001018, 0.00000588, 0.00002286, 0.00001127, 0.00002328, 0.00002044, 0.00005341, 0.00005552, -0.00006336, -0.00004216, -0.00001225, -0.00001383, -0.00001091, -0.00001172, -0.00000836, -0.00000357, -0.00000464, -0.00000431, -0.00000164, -0.00000362, -0.00000223, -0.00000198, -0.00000478, -0.00000069, -0.00000695, -0.00000247, 0.00000125, 0.00000024, -0.00000670, -0.00000578, 0.00000051, -0.00000118, 0.00000049, -0.00000082, -0.00000285, -0.00000374, -0.00000371, 0.00000145, -0.00000439, -0.00000216, 0.00000018, -0.00000681, 0.00000595, -0.00000499, 0.00001103, 0.00000301, -0.00000480, 0.00000161, 0.00000287, 0.00000166, 0.00000013, -0.00000019, 0.00000016, -0.00000048, -0.00000072, -0.00000025, -0.00000026, -0.00000015, 0.00000086, -0.00000184, 0.00000434, 0.00000319, 0.00000030, -0.00000619, 0.00000639, 0.00000699, -0.00000361, -0.00000002, -0.00000131, 0.00000140, -0.00000369, -0.00000177, 0.00000138, -0.00000380, -0.00000435, -0.00000123, -0.00000277, 0.00000118, -0.00000159, 0.00000173, -0.00000682, -0.00000337, -0.00000043, 0.00000184, -0.00000337, -0.00000450, -0.00000151, -0.00000491, -0.00000072, -0.00000160, -0.00000573, -0.00000291, -0.00000259, -0.00000003, -0.00000377, -0.00000566, -0.00000821, -0.00000674, -0.00001501, -0.00001184, -0.00003750, -0.00002221, 0.00003560, 0.00003081, 0.00001647, 0.00001772, 0.00000106, 0.00000943, 0.00000405, 0.00000564, 0.00000858, -0.00000003, 0.00000787, 0.00000946, },
		{ 0.00000686, 0.00001025, 0.00002398, 0.00002879, -0.00003005, -0.00009725, -0.00000575, 0.00006285, 0.00002171, 0.00001754, 0.00001133, 0.00000554, 0.00001435, 0.00000843, 0.00000602, 0.00002046, 0.00001175, 0.00001855, 0.00002155, 0.00003643, 0.00007097, -0.00003786, -0.00005638, -0.00001180, -0.00001618, -0.00000949, -0.00001313, -0.00000841, -0.00000452, -0.00000391, -0.00000475, -0.00000083, -0.00000429, -0.00000232, -0.00000194, -0.00000455, -0.00000102, -0.00000560, -0.00000474, 0.00000207, -0.00000028, -0.00000355, -0.00000861, 0.00000074, -0.00000116, 0.00000016, -0.00000044, -0.00000178, -0.00000419, -0.00000373, -0.00000031, -0.00000122, -0.00000534, 0.00000268, -0.00000775, 0.00000491, -0.00000355, 0.00000620, 0.00000824, -0.00000575, 0.00000068, 0.00000241, 0.00000266, 0.00000009, -0.00000012, 0.00000008, -0.00000025, -0.00000067, -0.00000040, -0.00000008, -0.00000044, 0.00000100, -0.00000155, 0.00000239, 0.00000426, 0.00000123, -0.00000514, 0.00000160, 0.00001037, -0.00000321, 0.00000026, -0.00000197, 0.00000147, -0.00000248, -0.00000324, 0.00000136, -0.00000208, -0.00000583, -0.00000082, -0.00000338, 0.00000134, -0.00000161, 0.00000142, -0.00000419, -0.00000575, -0.00000097, 0.00000176, -0.00000164, -0.00000517, -0.00000160, -0.00000472, -0.00000162, -0.00000067, -0.00000601, -0.00000268, -0.00000351, 0.00000039, -0.00000324, -0.00000539, -0.00000789, -0.00000681, -0.00001476, -0.00001066, -0.00003048, -0.00003628, 0.00002549, 0.00003804, 0.00001870, 0.00001905, 0.00000312, 0.00000893, 0.00000597, 0.00000545, 0.00000940, -0.00000017, 0.00000831, 0.00000974, -0.00000626, },
		{ 0.00000872, 0.00002532, 0.00003261, -0.00002831, -0.00009803, -0.00000747, 0.00006024, 0.00002055, 0.00001722, 0.00001060, 0.00000428, 0.00001468, 0.00000706, 0.00000566, 0.00001854, 0.00001205, 0.00001462, 0.00002210, 0.00002445, 0.00007193, -0.00000575, -0.00006494, -0.00001569, -0.00001676, -0.00000949, -0.00001334, -0.00000920, -0.00000502, -0.00000373, -0.00000485, -0.00000033, -0.00000448, -0.00000266, -0.00000202, -0.00000402, -0.00000178, -0.00000390, -0.00000670, 0.00000232, -0.00000075, -0.00000088, -0.00000971, -0.00000074, -0.00000012, -0.00000072, 0.00000002, -0.00000100, -0.00000411, -0.00000377, -0.00000206, 0.00000096, -0.00000639, 0.00000231, -0.00000573, 0.00000156, -0.00000029, 0.00000074, 0.00001185, -0.00000462, -0.00000096, 0.00000195, 0.00000329, 0.00000054, -0.00000030, 0.00000014, -0.00000015, -0.00000055, -0.00000045, -0.00000014, -0.00000034, 0.00000057, -0.00000061, 0.00000036, 0.00000485, 0.00000187, -0.00000314, -0.00000271, 0.00001144, -0.00000086, -0.00000061, -0.00000127, 0.00000013, -0.00000047, -0.00000458, 0.00000086, -0.00000053, -0.00000627, -0.00000150, -0.00000304, 0.00000046, -0.00000053, -0.00000011, -0.00000108, -0.00000736, -0.00000206, 0.00000143, -0.00000011, -0.00000511, -0.00000221, -0.00000426, -0.00000252, -0.00000014, -0.00000575, -0.00000287, -0.00000381, -0.00000009, -0.00000205, -0.00000574, -0.00000683, -0.00000757, -0.00001354, -0.00001158, -0.00002222, -0.00004442, 0.00000915, 0.00004361, 0.00002291, 0.00001989, 0.00000567, 0.00000856, 0.00000745, 0.00000590, 0.00001001, -0.00000010, 0.00000839, 0.00001050, -0.00000540, 0.00000810, },
		{ 0.00002549, 0.00003722, -0.00002685, -0.00009811, -0.00001056, 0.00005880, 0.00001905, 0.00001687, 0.00001005, 0.00000289, 0.00001487, 0.00000643, 0.00000462, 0.00001730, 0.00001170, 0.00001221, 0.00002083, 0.00001830, 0.00006269, 0.00002463, -0.00006284, -0.00002471, -0.00001557, -0.00001101, -0.00001235, -0.00001073, -0.00000496, -0.00000420, -0.00000452, -0.00000029, -0.00000412, -0.00000323, -0.00000215, -0.00000350, -0.00000242, -0.00000263, -0.00000759, 0.00000147, -0.00000064, 0.00000056, -0.00000871, -0.00000368, 0.00000147, -0.00000176, 0.00000048, -0.00000058, -0.00000352, -0.00000393, -0.00000332, 0.00000156, -0.00000519, -0.00000025, -0.00000223, -0.00000218, 0.00000267, -0.00000316, 0.00001239, -0.00000116, -0.00000308, 0.00000154, 0.00000331, 0.00000158, -0.00000070, 0.00000033, -0.00000017, -0.00000039, -0.00000043, -0.00000033, -0.00000008, -0.00000002, 0.00000032, -0.00000102, 0.00000453, 0.00000260, -0.00000117, -0.00000520, 0.00000959, 0.00000322, -0.00000225, 0.00000031, -0.00000177, 0.00000132, -0.00000507, -0.00000027, 0.00000067, -0.00000574, -0.00000285, -0.00000229, -0.00000086, 0.00000091, -0.00000180, 0.00000129, -0.00000738, -0.00000395, 0.00000116, 0.00000075, -0.00000414, -0.00000319, -0.00000375, -0.00000319, -0.00000012, -0.00000505, -0.00000329, -0.00000371, -0.00000103, -0.00000076, -0.00000616, -0.00000568, -0.00000826, -0.00001233, -0.00001282, -0.00001580, -0.00004494, -0.00001109, 0.00004469, 0.00002966, 0.00002042, 0.00000846, 0.00000860, 0.00000843, 0.00000678, 0.00001062, 0.00000010, 0.00000817, 0.00001160, -0.00000496, 0.00000967, 0.00000774, },
		{ 0.00004251, -0.00002562, -0.00009751, -0.00001456, 0.00005764, 0.00001805, 0.00001595, 0.00000993, 0.00000132, 0.00001484, 0.00000643, 0.00000321, 0.00001641, 0.00001089, 0.00001105, 0.00001810, 0.00001626, 0.00004928, 0.00004643, -0.00004878, -0.00003664, -0.00001432, -0.00001301, -0.00001099, -0.00001235, -0.00000478, -0.00000495, -0.00000411, -0.00000043, -0.00000350, -0.00000380, -0.00000237, -0.00000315, -0.00000262, -0.00000215, -0.00000725, -0.00000033, -0.00000003, 0.00000084, -0.00000618, -0.00000702, 0.00000262, -0.00000231, 0.00000054, -0.00000021, -0.00000281, -0.00000394, -0.00000410, 0.00000086, -0.00000271, -0.00000333, 0.00000070, -0.00000437, 0.00000367, -0.00000427, 0.00000970, 0.00000377, -0.00000486, 0.00000096, 0.00000281, 0.00000292, -0.00000103, 0.00000046, -0.00000018, -0.00000027, -0.00000036, -0.00000050, 0.00000009, -0.00000041, 0.00000079, -0.00000143, 0.00000330, 0.00000350, 0.00000019, -0.00000546, 0.00000543, 0.00000774, -0.00000344, 0.00000157, -0.00000290, 0.00000177, -0.00000417, -0.00000206, 0.00000153, -0.00000464, -0.00000415, -0.00000182, -0.00000193, 0.00000186, -0.00000268, 0.00000207, -0.00000560, -0.00000646, 0.00000098, 0.00000086, -0.00000252, -0.00000416, -0.00000351, -0.00000347, -0.00000051, -0.00000422, -0.00000360, -0.00000362, -0.00000190, 0.00000006, -0.00000611, -0.00000506, -0.00000820, -0.00001193, -0.00001306, -0.00001277, -0.00003885, -0.00003087, 0.00003838, 0.00003878, 0.00002118, 0.00001124, 0.00000916, 0.00000909, 0.00000776, 0.00001151, 0.00000031, 0.00000783, 0.00001274, -0.00000447, 0.00001083, 0.00000797, 0.00002426, },
		{ -0.00002416, -0.00009675, -0.00001860, 0.00005576, 0.00001823, 0.00001420, 0.00001030, -0.00000042, 0.00001464, 0.00000674, 0.00000191, 0.00001543, 0.00001001, 0.00001053, 0.00001493, 0.00001586, 0.00003682, 0.00005658, -0.00002576, -0.00004694, -0.00001550, -0.00001414, -0.00001033, -0.00001327, -0.00000504, -0.00000543, -0.00000408, -0.00000041, -0.00000297, -0.00000407, -0.00000273, -0.00000303, -0.00000237, -0.00000236, -0.00000607, -0.00000242, 0.00000047, 0.00000056, -0.00000322, -0.00000946, 0.00000238, -0.00000183, -0.00000009, 0.00000041, -0.00000241, -0.00000354, -0.00000464, -0.00000049, -0.00000021, -0.00000533, 0.00000163, -0.00000413, 0.00000239, -0.00000288, 0.00000498, 0.00000850, -0.00000523, -0.00000017, 0.00000211, 0.00000407, -0.00000088, 0.00000027, -0.00000002, -0.00000026, -0.00000023, -0.00000061, 0.00000008, -0.00000049, 0.00000071, -0.00000103, 0.00000158, 0.00000431, 0.00000096, -0.00000405, 0.00000052, 0.00001087, -0.00000281, 0.00000139, -0.00000232, 0.00000044, -0.00000201, -0.00000409, 0.00000188, -0.00000334, -0.00000489, -0.00000203, -0.00000238, 0.00000192, -0.00000238, 0.00000126, -0.00000260, -0.00000885, 0.00000048, 0.00000058, -0.00000083, -0.00000457, -0.00000380, -0.00000335, -0.00000113, -0.00000354, -0.00000356, -0.00000375, -0.00000241, 0.00000019, -0.00000548, -0.00000509, -0.00000728, -0.00001243, -0.00001212, -0.00001270, -0.00002945, -0.00004526, 0.00002317, 0.00004870, 0.00002323, 0.00001365, 0.00001031, 0.00000963, 0.00000858, 0.00001275, 0.00000056, 0.00000747, 0.00001366, -0.00000355, 0.00001121, 0.00000949, 0.00002205, 0.00004788, },
		},
		{ { -0.00000249, -0.00000203, -0.00000527, -0.00000141, -0.00000169, -0.00000280, -0.00000015, -0.00000433, -0.00000455, -0.00000450, -0.00000077, -0.00000403, -0.00000626, -0.00000829, -0.00001461, -0.00002081, -0.00005314, 0.00000102, 0.00006434, 0.00002562, 0.00001721, 0.00001365, 0.00001703, 0.00001679, 0.00001101, 0.00000739, 0.00002133, 0.00001123, 0.00001284, 0.00001470, 0.00001802, 0.00002868, 0.00005336, 0.00000497, -0.00005141, -0.00002598, -0.00001554, -0.00000945, -0.00000011, -0.00000833, -0.00000557, -0.00000255, -0.00001128, -0.00000045, 0.00000009, 0.00000134, -0.00000326, -0.00000654, 0.00000281, -0.00000165, -0.00000285, -0.00000192, -0.00000230, -0.00000224, -0.00001423, 0.00000223, 0.00000046, -0.00000400, -0.00000912, -0.00000322, -0.00000166, -0.00000136, -0.00000226, -0.00000094, -0.00000087, -0.00000094, -0.00000226, -0.00000136, -0.00000166, -0.00000322, -0.00000912, -0.00000400, 0.00000046, 0.00000223, -0.00001423, -0.00000224, -0.00000230, -0.00000192, -0.00000285, -0.00000165, 0.00000281, -0.00000654, -0.00000326, 0.00000134, 0.00000009, -0.00000045, -0.00001128, -0.00000255, -0.00000557, -0.00000833, -0.00000011, -0.00000945, -0.00001554, -0.00002598, -0.00005141, 0.00000497, 0.00005336, 0.00002868, 0.00001802, 0.00001470, 0.00001284, 0.00001123, 0.00002133, 0.00000739, 0.00001101, 0.00001679, 0.00001703, 0.00001365, 0.00001721, 0.00002562, 0.00006434, 0.00000102, -0.00005314, -0.00002081, -0.00001461, -0.00000829, -0.00000626, -0.00000403, -0.00000077, -0.00000450, -0.00000455, -0.00000433, -0.00000015, -0.00000280, -0.00000169, -0.00000141, -0.00000527, -0.00000203, },
		{ -0.00000221, -0.00000501, -0.00000218, -0.00000064, -0.00000366, -0.00000047, -0.00000262, -0.00000424, -0.00000496, -0.00000265, -0.00000165, -0.00000621, -0.00000652, -0.00001375, -0.00001549, -0.00004505, -0.00002320, 0.00005531, 0.00003825, 0.00001515, 0.00001506, 0.00001371, 0.00001680, 0.00001406, 0.00000636, 0.00001363, 0.00001907, 0.00000829, 0.00001544, 0.00001332, 0.00002170, 0.00004058, 0.00004078, -0.00003145, -0.00004318, -0.00001687, -0.00001363, -0.00000097, -0.00000495, -0.00000971, -0.00000083, -0.00000891, -0.00000552, 0.00000076, 0.00000100, -0.00000080, -0.00000673, -0.00000054, 0.00000107, -0.00000327, -0.00000170, -0.00000341, 0.00000023, -0.00001362, -0.00000250, 0.00000274, -0.00000289, -0.00000767, -0.00000599, -0.00000129, -0.00000193, -0.00000132, -0.00000189, -0.00000059, -0.00000086, -0.00000184, -0.00000202, -0.00000148, -0.00000176, -0.00000799, -0.00000633, -0.00000127, 0.00000408, -0.00001086, -0.00000722, 0.00000035, -0.00000425, 0.00000041, -0.00000594, 0.00000479, -0.00000432, -0.00000549, 0.00000053, 0.00000012, 0.00000102, -0.00000708, -0.00000900, -0.00000069, -0.00001016, -0.00000060, -0.00000656, -0.00001363, -0.00001784, -0.00004092, -0.00003941, 0.00004405, 0.00004390, 0.00002221, 0.00001647, 0.00001585, 0.00000929, 0.00001870, 0.00001607, 0.00000725, 0.00001526, 0.00001950, 0.00001428, 0.00001825, 0.00001878, 0.00005919, 0.00003163, -0.00005471, -0.00002957, -0.00001545, -0.00001041, -0.00000621, -0.00000576, -0.00000138, -0.00000227, -0.00000494, -0.00000511, -0.00000178, -0.00000090, -0.00000300, -0.00000024, -0.00000600, -0.00000198, -0.00000221, },
		{ -0.00000473, -0.00000341, 0.00000050, -0.00000364, -0.00000170, -0.00000134, -0.00000360, -0.00000436, -0.00000468, -0.00000102, -0.00000492, -0.00000558, -0.00001228, -0.00001307, -0.00003487, -0.00003765, 0.00003664, 0.00004877, 0.00001711, 0.00001512, 0.00001234, 0.00001522, 0.00001462, 0.00001065, 0.00000528, 0.00002059, 0.00001134, 0.00001150, 0.00001373, 0.00001513, 0.00002911, 0.00004676, 0.00000985, -0.00004970, -0.00002649, -0.00001529, -0.00000492, -0.00000134, -0.00001067, -0.00000378, -0.00000427, -0.00000849, -0.00000141, 0.00000141, 0.00000050, -0.00000528, -0.00000317, 0.00000113, -0.00000194, -0.00000184, -0.00000406, 0.00000090, -0.00001051, -0.00000786, 0.00000401, -0.00000211, -0.00000533, -0.00000810, -0.00000156, -0.00000265, -0.00000056, -0.00000224, -0.00000075, -0.00000093, -0.00000111, -0.00000256, -0.00000161, -0.00000117, -0.00000563, -0.00000810, -0.00000315, 0.00000361, -0.00000538, -0.00001216, 0.00000223, -0.00000617, 0.00000263, -0.00000691, 0.00000211, 0.00000030, -0.00000695, -0.00000123, 0.00000076, -0.00000001, -0.00000134, -0.00001234, -0.00000147, -0.00000752, -0.00000311, -0.00000410, -0.00001125, -0.00001615, -0.00002372, -0.00005540, 0.00000000, 0.00006056, 0.00002800, 0.00002187, 0.00001557, 0.00001401, 0.00001089, 0.00002292, 0.00000938, 0.00001144, 0.00001980, 0.00001744, 0.00001725, 0.00001845, 0.00004532, 0.00005781, -0.00004301, -0.00004273, -0.00001594, -0.00001344, -0.00000611, -0.00000664, -0.00000333, -0.00000077, -0.00000380, -0.00000563, -0.00000387, 0.00000009, -0.00000324, 0.00000018, -0.00000562, -0.00000345, -0.00000100, -0.00000244, },
		{ -0.00000556, 0.00000157, -0.00000301, -0.00000264, -0.00000143, -0.00000233, -0.00000385, -0.00000499, -0.00000264, -0.00000302, -0.00000507, -0.00001009, -0.00001274, -0.00002502, -0.00004360, 0.00001559, 0.00005148, 0.00002386, 0.00001411, 0.00001261, 0.00001316, 0.00001374, 0.00001378, 0.00000476, 0.00001229, 0.00001931, 0.00000708, 0.00001484, 0.00001081, 0.00002161, 0.00003683, 0.00003907, -0.00002674, -0.00004383, -0.00001642, -0.00001064, 0.00000084, -0.00000929, -0.00000687, -0.00000328, -0.00000620, -0.00000598, 0.00000148, 0.00000126, -0.00000339, -0.00000451, 0.00000024, -0.00000159, -0.00000066, -0.00000480, 0.00000050, -0.00000666, -0.00001169, 0.00000293, -0.00000053, -0.00000399, -0.00000801, -0.00000326, -0.00000242, -0.00000134, -0.00000109, -0.00000184, -0.00000046, -0.00000118, -0.00000191, -0.00000260, -0.00000071, -0.00000380, -0.00000778, -0.00000564, 0.00000204, -0.00000057, -0.00001401, 0.00000069, -0.00000480, 0.00000050, -0.00000251, -0.00000434, 0.00000495, -0.00000695, -0.00000256, -0.00000011, 0.00000029, 0.00000004, -0.00000822, -0.00000796, -0.00000267, -0.00000634, -0.00000091, -0.00001124, -0.00001275, -0.00001802, -0.00003920, -0.00004900, 0.00004846, 0.00004660, 0.00002384, 0.00001955, 0.00001626, 0.00000984, 0.00001907, 0.00001871, 0.00000781, 0.00001781, 0.00002046, 0.00001716, 0.00001987, 0.00003251, 0.00006985, -0.00001646, -0.00005719, -0.00001746, -0.00001711, -0.00000607, -0.00000753, -0.00000429, -0.00000220, -0.00000077, -0.00000610, -0.00000503, -0.00000111, -0.00000185, -0.00000087, -0.00000326, -0.00000648, 0.00000050, -0.00000261, -0.00000371, },
		{ 0.00000064, -0.00000082, -0.00000370, -0.00000175, -0.00000189, -0.00000277, -0.00000467, -0.00000396, -0.00000290, -0.00000373, -0.00000818, -0.00001263, -0.00001785, -0.00004232, -0.00000425, 0.00004743, 0.00003076, 0.00001532, 0.00001229, 0.00001234, 0.00001216, 0.00001383, 0.00000938, 0.00000404, 0.00002020, 0.00001094, 0.00001026, 0.00001264, 0.00001361, 0.00002811, 0.00004250, 0.00001206, -0.00004779, -0.00002672, -0.00001305, -0.00000155, -0.00000568, -0.00000884, -0.00000486, -0.00000352, -0.00000706, -0.00000226, 0.00000319, -0.00000249, -0.00000431, -0.00000082, -0.00000190, -0.00000017, -0.00000360, -0.00000183, -0.00000232, -0.00001379, 0.00000024, 0.00000111, -0.00000317, -0.00000693, -0.00000487, -0.00000223, -0.00000251, -0.00000013, -0.00000218, -0.00000066, -0.00000107, -0.00000137, -0.00000292, -0.00000127, -0.00000220, -0.00000659, -0.00000701, -0.00000096, 0.00000287, -0.00001281, -0.00000329, -0.00000153, -0.00000368, 0.00000283, -0.00000855, 0.00000508, -0.00000401, -0.00000387, -0.00000149, 0.00000097, -0.00000085, -0.00000275, -0.00001133, -0.00000298, -0.00000590, -0.00000003, -0.00001007, -0.00001049, -0.00001729, -0.00002027, -0.00006210, -0.00000351, 0.00006663, 0.00002885, 0.00002539, 0.00001698, 0.00001422, 0.00001152, 0.00002451, 0.00001136, 0.00001271, 0.00002215, 0.00001842, 0.00002124, 0.00002459, 0.00006828, 0.00001661, -0.00006437, -0.00002409, -0.00001893, -0.00000804, -0.00000755, -0.00000484, -0.00000455, 0.00000088, -0.00000444, -0.00000614, -0.00000293, -0.00000087, -0.00000175, -0.00000086, -0.00000837, -0.00000012, -0.00000147, -0.00000296, -0.00000685, },
		{ 0.00000224, -0.00000443, -0.00000180, -0.00000263, -0.00000126, -0.00000439, -0.00000377, -0.00000442, -0.00000252, -0.00000621, -0.00001205, -0.00001385, -0.00003616, -0.00002058, 0.00003860, 0.00003541, 0.00001823, 0.00001250, 0.00001164, 0.00001148, 0.00001165, 0.00001320, 0.00000311, 0.00001220, 0.00001851, 0.00000647, 0.00001337, 0.00001020, 0.00002004, 0.00003472, 0.00003661, -0.00002373, -0.00004329, -0.00001522, -0.00000640, -0.00000226, -0.00000959, -0.00000576, -0.00000475, -0.00000319, -0.00000781, 0.00000394, -0.00000118, -0.00000390, -0.00000155, -0.00000159, -0.00000188, -0.00000047, -0.00000468, 0.00000084, -0.00001343, -0.00000367, 0.00000228, -0.00000229, -0.00000598, -0.00000551, -0.00000257, -0.00000311, -0.00000056, -0.00000112, -0.00000173, -0.00000044, -0.00000162, -0.00000202, -0.00000271, -0.00000087, -0.00000557, -0.00000649, -0.00000465, 0.00000433, -0.00000952, -0.00000767, 0.00000105, -0.00000637, 0.00000463, -0.00000709, 0.00000037, 0.00000040, -0.00000499, -0.00000172, -0.00000026, -0.00000002, -0.00000183, -0.00000744, -0.00000846, -0.00000338, -0.00000129, -0.00000592, -0.00001224, -0.00001327, -0.00001591, -0.00004061, -0.00005707, 0.00005135, 0.00005090, 0.00002518, 0.00002350, 0.00001545, 0.00001149, 0.00001916, 0.00002187, 0.00000827, 0.00002132, 0.00001976, 0.00002308, 0.00002081, 0.00005914, 0.00004598, -0.00005816, -0.00003680, -0.00001831, -0.00001271, -0.00000626, -0.00000626, -0.00000508, -0.00000123, -0.00000040, -0.00000717, -0.00000412, -0.00000159, -0.00000154, 0.00000009, -0.00000742, -0.00000346, 0.00000072, -0.00000282, -0.00000618, -0.00000269, },
		{ -0.00000297, -0.00000267, -0.00000293, -0.00000112, -0.00000312, -0.00000361, -0.00000479, -0.00000374, -0.00000322, -0.00001149, -0.00001159, -0.00002883, -0.00003060, 0.00002505, 0.00003920, 0.00001958, 0.00001528, 0.00000987, 0.00001228, 0.00000894, 0.00001389, 0.00000693, 0.00000483, 0.00001852, 0.00001131, 0.00000797, 0.00001308, 0.00001160, 0.00002720, 0.00003941, 0.00001157, -0.00004477, -0.00002666, -0.00000843, -0.00000281, -0.00000815, -0.00000618, -0.00000723, -0.00000050, -0.00000848, -0.00000047, 0.00000234, -0.00000455, -0.00000148, -0.00000097, -0.00000376, 0.00000088, -0.00000439, 0.00000029, -0.00000957, -0.00000889, 0.00000330, -0.00000186, -0.00000466, -0.00000626, -0.00000227, -0.00000385, -0.00000131, -0.00000042, -0.00000186, -0.00000073, -0.00000135, -0.00000165, -0.00000299, -0.00000123, -0.00000389, -0.00000587, -0.00000635, 0.00000233, -0.00000424, -0.00001192, 0.00000250, -0.00000717, 0.00000333, -0.00000287, -0.00000416, 0.00000216, -0.00000433, -0.00000166, -0.00000192, 0.00000106, -0.00000270, -0.00000295, -0.00001022, -0.00000563, -0.00000048, -0.00000273, -0.00001288, -0.00001064, -0.00001633, -0.00001916, -0.00006819, -0.00000779, 0.00007310, 0.00003035, 0.00002935, 0.00001760, 0.00001500, 0.00001217, 0.00002662, 0.00001285, 0.00001583, 0.00002180, 0.00002343, 0.00002200, 0.00004580, 0.00006772, -0.00004024, -0.00005086, -0.00001842, -0.00001757, -0.00000582, -0.00000749, -0.00000455, -0.00000473, 0.00000226, -0.00000571, -0.00000564, -0.00000262, -0.00000176, 0.00000052, -0.00000506, -0.00000690, 0.00000140, -0.00000190, -0.00000487, -0.00000588, 0.00000321, },
		{ -0.00000390, -0.00000248, -0.00000245, -0.00000120, -0.00000391, -0.00000332, -0.00000650, -0.00000095, -0.00000993, -0.00001062, -0.00002222, -0.00003411, 0.00000874, 0.00004120, 0.00002041, 0.00001770, 0.00001002, 0.00001178, 0.00000857, 0.00001132, 0.00001080, 0.00000357, 0.00001125, 0.00001805, 0.00000531, 0.00001244, 0.00001035, 0.00001738, 0.00003450, 0.00003200, -0.00002073, -0.00004218, -0.00001179, -0.00000483, -0.00000673, -0.00000644, -0.00000769, -0.00000276, -0.00000350, -0.00000726, 0.00000484, -0.00000420, -0.00000190, -0.00000025, -0.00000383, -0.00000131, -0.00000100, -0.00000244, -0.00000455, -0.00001314, 0.00000284, -0.00000116, -0.00000333, -0.00000717, -0.00000163, -0.00000429, -0.00000208, -0.00000070, -0.00000085, -0.00000179, -0.00000052, -0.00000216, -0.00000190, -0.00000279, -0.00000201, -0.00000581, -0.00000557, -0.00000193, 0.00000101, -0.00001451, 0.00000196, -0.00000597, 0.00000028, 0.00000112, -0.00000582, 0.00000102, -0.00000298, -0.00000150, -0.00000234, 0.00000025, -0.00000173, -0.00000298, -0.00000547, -0.00001128, 0.00000019, -0.00000057, -0.00001039, -0.00001215, -0.00001334, -0.00001396, -0.00004355, -0.00006390, 0.00005254, 0.00005715, 0.00002623, 0.00002757, 0.00001417, 0.00001402, 0.00001912, 0.00002503, 0.00001009, 0.00002306, 0.00002147, 0.00002731, 0.00003188, 0.00007974, -0.00001358, -0.00006237, -0.00002110, -0.00002098, -0.00000758, -0.00000774, -0.00000450, -0.00000659, 0.00000100, -0.00000166, -0.00000706, -0.00000333, -0.00000292, 0.00000066, -0.00000253, -0.00000851, -0.00000038, -0.00000018, -0.00000373, -0.00000712, 0.00000103, 0.00000030, },
		{ -0.00000271, -0.00000335, -0.00000065, -0.00000330, -0.00000207, -0.00000741, -0.00000210, -0.00000622, -0.00001111, -0.00001632, -0.00003351, -0.00000597, 0.00003782, 0.00002459, 0.00001577, 0.00001425, 0.00000887, 0.00001079, 0.00000743, 0.00001254, 0.00000561, 0.00000599, 0.00001623, 0.00001199, 0.00000515, 0.00001513, 0.00000821, 0.00002776, 0.00003548, 0.00000993, -0.00004097, -0.00002501, -0.00000418, -0.00000797, -0.00000609, -0.00000656, -0.00000676, 0.00000052, -0.00000931, 0.00000191, -0.00000091, -0.00000369, 0.00000057, -0.00000246, -0.00000442, 0.00000106, -0.00000287, -0.00000231, -0.00001345, -0.00000088, 0.00000107, -0.00000319, -0.00000680, -0.00000257, -0.00000277, -0.00000390, -0.00000067, -0.00000066, -0.00000158, -0.00000088, -0.00000186, -0.00000162, -0.00000302, -0.00000196, -0.00000479, -0.00000486, -0.00000457, 0.00000261, -0.00001282, -0.00000243, -0.00000165, -0.00000478, 0.00000508, -0.00000660, 0.00000073, -0.00000385, -0.00000030, -0.00000232, -0.00000070, -0.00000088, -0.00000394, -0.00000182, -0.00001100, -0.00000570, 0.00000401, -0.00000856, -0.00001213, -0.00001244, -0.00001307, -0.00002175, -0.00007131, -0.00001503, 0.00008175, 0.00003144, 0.00003421, 0.00001675, 0.00001761, 0.00001180, 0.00002975, 0.00001430, 0.00001943, 0.00002065, 0.00003069, 0.00002472, 0.00007762, 0.00002158, -0.00007065, -0.00002560, -0.00002352, -0.00001048, -0.00000827, -0.00000451, -0.00000708, -0.00000200, 0.00000114, -0.00000574, -0.00000498, -0.00000337, -0.00000096, 0.00000051, -0.00000880, -0.00000281, 0.00000042, -0.00000202, -0.00000705, -0.00000213, 0.00000237, -0.00000334, },
		{ -0.00000349, -0.00000157, -0.00000202, -0.00000171, -0.00000571, -0.00000564, -0.00000249, -0.00001122, -0.00001206, -0.00003032, -0.00001673, 0.00002854, 0.00003135, 0.00001200, 0.00001810, 0.00000789, 0.00001152, 0.00000612, 0.00001108, 0.00000775, 0.00000599, 0.00000865, 0.00001877, 0.00000277, 0.00001346, 0.00000938, 0.00001550, 0.00003439, 0.00002592, -0.00001730, -0.00003992, -0.00000716, -0.00000795, -0.00000779, -0.00000432, -0.00000909, -0.00000014, -0.00000591, -0.00000373, 0.00000191, -0.00000438, 0.00000006, -0.00000031, -0.00000576, -0.00000059, -0.00000036, -0.00000243, -0.00001094, -0.00000611, 0.00000331, -0.00000334, -0.00000550, -0.00000493, -0.00000010, -0.00000567, -0.00000063, -0.00000139, -0.00000035, -0.00000194, -0.00000095, -0.00000234, -0.00000180, -0.00000328, -0.00000315, -0.00000503, -0.00000450, 0.00000064, -0.00000785, -0.00000913, 0.00000369, -0.00000953, 0.00000751, -0.00000643, 0.00000112, -0.00000500, -0.00000041, -0.00000128, -0.00000096, -0.00000138, -0.00000309, -0.00000308, -0.00000409, -0.00001403, 0.00000632, -0.00000572, -0.00001017, -0.00001450, -0.00001007, -0.00001611, -0.00004383, -0.00007277, 0.00005456, 0.00006352, 0.00002861, 0.00002961, 0.00001511, 0.00001511, 0.00002076, 0.00002698, 0.00001441, 0.00002131, 0.00002872, 0.00002717, 0.00006263, 0.00005955, -0.00007030, -0.00003510, -0.00002326, -0.00001504, -0.00000890, -0.00000505, -0.00000666, -0.00000456, 0.00000081, -0.00000216, -0.00000654, -0.00000324, -0.00000365, 0.00000273, -0.00000749, -0.00000508, -0.00000027, -0.00000042, -0.00000590, -0.00000468, 0.00000208, -0.00000122, -0.00000322, },
		{ -0.00000229, -0.00000176, -0.00000115, -0.00000338, -0.00000800, -0.00000184, -0.00000914, -0.00001028, -0.00002509, -0.00002382, 0.00001661, 0.00003568, 0.00001237, 0.00001624, 0.00001187, 0.00000904, 0.00000809, 0.00000773, 0.00000889, 0.00000738, 0.00000497, 0.00001550, 0.00001104, 0.00000405, 0.00001661, 0.00000501, 0.00002885, 0.00003033, 0.00000818, -0.00003625, -0.00002151, -0.00000368, -0.00001247, -0.00000213, -0.00000909, -0.00000310, -0.00000265, -0.00000593, 0.00000044, -0.00000238, -0.00000200, 0.00000164, -0.00000480, -0.00000394, 0.00000134, -0.00000137, -0.00000885, -0.00000965, 0.00000304, -0.00000193, -0.00000491, -0.00000660, 0.00000084, -0.00000487, -0.00000230, -0.00000118, -0.00000043, -0.00000147, -0.00000131, -0.00000219, -0.00000147, -0.00000330, -0.00000309, -0.00000444, -0.00000421, -0.00000120, -0.00000381, -0.00001351, 0.00000512, -0.00000939, 0.00000476, -0.00000222, -0.00000183, -0.00000171, -0.00000502, 0.00000206, -0.00000210, -0.00000057, -0.00000343, -0.00000413, -0.00000020, -0.00001354, -0.00000133, 0.00000238, -0.00001112, -0.00001244, -0.00001209, -0.00001198, -0.00002411, -0.00007404, -0.00002327, 0.00009097, 0.00003396, 0.00003732, 0.00001743, 0.00001948, 0.00001261, 0.00003177, 0.00001853, 0.00001921, 0.00002511, 0.00003256, 0.00004575, 0.00008653, -0.00005067, -0.00005509, -0.00001759, -0.00002254, -0.00000798, -0.00000804, -0.00000433, -0.00000740, -0.00000011, -0.00000054, -0.00000489, -0.00000461, -0.00000466, 0.00000149, -0.00000343, -0.00000795, -0.00000050, -0.00000077, -0.00000311, -0.00000681, 0.00000098, -0.00000001, -0.00000226, -0.00000403, },
		{ -0.00000216, -0.00000093, -0.00000109, -0.00000814, -0.00000364, -0.00000652, -0.00000937, -0.00001967, -0.00002715, 0.00000460, 0.00003487, 0.00001784, 0.00001066, 0.00001641, 0.00000773, 0.00000922, 0.00000621, 0.00000811, 0.00000758, 0.00000694, 0.00000690, 0.00001859, 0.00000092, 0.00001494, 0.00000780, 0.00001458, 0.00003323, 0.00001942, -0.00001313, -0.00003589, -0.00000436, -0.00001340, -0.00000467, -0.00000583, -0.00000634, -0.00000181, -0.00000498, -0.00000219, -0.00000143, -0.00000251, 0.00000121, -0.00000167, -0.00000664, 0.00000034, 0.00000099, -0.00000689, -0.00001169, 0.00000089, 0.00000012, -0.00000452, -0.00000725, -0.00000047, -0.00000192, -0.00000469, -0.00000050, -0.00000159, -0.00000004, -0.00000221, -0.00000154, -0.00000206, -0.00000209, -0.00000407, -0.00000344, -0.00000440, -0.00000172, -0.00000182, -0.00001424, 0.00000157, -0.00000427, -0.00000222, 0.00000479, -0.00000714, 0.00000411, -0.00001028, 0.00000395, -0.00000256, 0.00000135, -0.00000518, -0.00000298, -0.00000237, -0.00000484, -0.00001285, 0.00000884, -0.00001122, -0.00000806, -0.00001729, -0.00000586, -0.00002103, -0.00004151, -0.00008378, 0.00005731, 0.00007068, 0.00003081, 0.00003119, 0.00001728, 0.00001560, 0.00002296, 0.00002992, 0.00001737, 0.00002164, 0.00003490, 0.00003584, 0.00009403, -0.00001018, -0.00008047, -0.00001284, -0.00002829, -0.00000791, -0.00001179, -0.00000187, -0.00000973, -0.00000060, -0.00000179, -0.00000110, -0.00000619, -0.00000419, -0.00000227, 0.00000129, -0.00000990, -0.00000077, -0.00000213, -0.00000048, -0.00000748, -0.00000055, 0.00000003, -0.00000067, -0.00000410, -0.00000295, },
		{ -0.00000227, 0.00000147, -0.00000714, -0.00000527, -0.00000569, -0.00000783, -0.00001542, -0.00002697, -0.00000610, 0.00002952, 0.00002470, 0.00000723, 0.00001613, 0.00001053, 0.00000797, 0.00000697, 0.00000659, 0.00000630, 0.00000972, 0.00000307, 0.00001547, 0.00000913, 0.00000475, 0.00001661, 0.00000316, 0.00002906, 0.00002448, 0.00000711, -0.00003023, -0.00001793, -0.00000745, -0.00001239, -0.00000112, -0.00000876, -0.00000162, -0.00000533, -0.00000131, -0.00000415, -0.00000059, -0.00000126, 0.00000190, -0.00000738, -0.00000208, 0.00000183, -0.00000300, -0.00001356, -0.00000149, 0.00000115, -0.00000319, -0.00000772, -0.00000250, 0.00000046, -0.00000510, -0.00000134, -0.00000171, -0.00000017, -0.00000137, -0.00000222, -0.00000177, -0.00000187, -0.00000364, -0.00000398, -0.00000362, -0.00000273, 0.00000021, -0.00001345, -0.00000381, 0.00000116, -0.00000782, 0.00000876, -0.00000874, 0.00000532, -0.00000891, -0.00000059, 0.00000052, 0.00000104, -0.00000400, -0.00000414, -0.00000376, 0.00000024, -0.00001446, 0.00000343, -0.00000371, -0.00000903, -0.00001548, -0.00000825, -0.00001557, -0.00002302, -0.00007905, -0.00003049, 0.00009939, 0.00003873, 0.00003753, 0.00002153, 0.00001873, 0.00001609, 0.00003225, 0.00002501, 0.00001616, 0.00003474, 0.00003200, 0.00008660, 0.00003796, -0.00009411, -0.00002189, -0.00002485, -0.00001327, -0.00001204, -0.00000344, -0.00000849, -0.00000313, -0.00000207, -0.00000023, -0.00000430, -0.00000494, -0.00000475, 0.00000251, -0.00000809, -0.00000324, -0.00000170, -0.00000079, -0.00000521, -0.00000336, 0.00000067, -0.00000087, -0.00000207, -0.00000454, -0.00000182, },
		{ 0.00000255, -0.00000450, -0.00000656, -0.00000554, -0.00000733, -0.00001111, -0.00002539, -0.00001354, 0.00002049, 0.00003015, 0.00000793, 0.00001231, 0.00001352, 0.00000831, 0.00000642, 0.00000718, 0.00000360, 0.00001035, 0.00000521, 0.00000736, 0.00001614, 0.00000166, 0.00001472, 0.00000738, 0.00001336, 0.00003093, 0.00001351, -0.00000794, -0.00003068, -0.00000598, -0.00001607, -0.00000187, -0.00000741, -0.00000238, -0.00000655, 0.00000030, -0.00000567, -0.00000112, -0.00000154, 0.00000241, -0.00000486, -0.00000504, 0.00000087, 0.00000098, -0.00001306, -0.00000540, 0.00000220, -0.00000218, -0.00000703, -0.00000512, 0.00000158, -0.00000371, -0.00000286, -0.00000128, -0.00000135, 0.00000016, -0.00000301, -0.00000145, -0.00000218, -0.00000243, -0.00000495, -0.00000286, -0.00000381, 0.00000178, -0.00001090, -0.00000969, 0.00000444, -0.00000955, 0.00000765, -0.00000497, 0.00000071, -0.00000170, -0.00000811, 0.00000374, -0.00000037, 0.00000002, -0.00000761, -0.00000247, -0.00000147, -0.00000664, -0.00000775, 0.00000436, -0.00001128, -0.00000916, -0.00001601, -0.00000611, -0.00002344, -0.00004071, -0.00009352, 0.00005789, 0.00008145, 0.00002929, 0.00003660, 0.00001651, 0.00001929, 0.00002252, 0.00003715, 0.00001460, 0.00003064, 0.00003239, 0.00007131, 0.00007952, -0.00008385, -0.00004751, -0.00001489, -0.00002136, -0.00000975, -0.00000852, -0.00000460, -0.00000740, -0.00000031, -0.00000296, -0.00000018, -0.00000590, -0.00000542, 0.00000019, -0.00000379, -0.00000675, -0.00000004, -0.00000297, -0.00000197, -0.00000587, 0.00000141, -0.00000213, 0.00000030, -0.00000533, -0.00000176, -0.00000372, },
		{ 0.00000016, -0.00000819, -0.00000423, -0.00000912, -0.00000645, -0.00002338, -0.00001744, 0.00000966, 0.00003245, 0.00001161, 0.00000885, 0.00001300, 0.00001157, 0.00000465, 0.00000860, 0.00000208, 0.00000791, 0.00000885, 0.00000356, 0.00001323, 0.00000924, 0.00000467, 0.00001676, 0.00000192, 0.00002790, 0.00001869, 0.00000738, -0.00002302, -0.00001726, -0.00001166, -0.00000930, -0.00000318, -0.00000433, -0.00000558, -0.00000234, -0.00000156, -0.00000629, 0.00000117, -0.00000013, -0.00000033, -0.00000762, -0.00000021, 0.00000182, -0.00000811, -0.00001139, 0.00000373, -0.00000228, -0.00000468, -0.00000832, 0.00000195, -0.00000239, -0.00000295, -0.00000206, -0.00000147, -0.00000015, -0.00000161, -0.00000288, -0.00000122, -0.00000269, -0.00000393, -0.00000415, -0.00000308, 0.00000062, -0.00000505, -0.00001613, 0.00000567, -0.00000875, 0.00000408, 0.00000037, -0.00000438, 0.00000387, -0.00001060, 0.00000155, 0.00000065, 0.00000242, -0.00000785, -0.00000382, -0.00000257, -0.00000174, -0.00001029, 0.00000184, -0.00000662, -0.00000804, -0.00001636, -0.00000630, -0.00001930, -0.00002159, -0.00008372, -0.00003921, 0.00010994, 0.00004181, 0.00003938, 0.00002482, 0.00001930, 0.00001842, 0.00003595, 0.00002692, 0.00001990, 0.00003758, 0.00005155, 0.00010803, -0.00005118, -0.00007800, -0.00001067, -0.00002438, -0.00000989, -0.00001231, -0.00000283, -0.00000952, -0.00000002, -0.00000457, 0.00000076, -0.00000357, -0.00000652, -0.00000232, -0.00000107, -0.00000795, -0.00000010, -0.00000352, -0.00000129, -0.00000559, 0.00000023, -0.00000173, -0.00000007, -0.00000321, -0.00000345, -0.00000333, 0.00000049, },
		{ -0.00000811, -0.00000337, -0.00001031, -0.00000463, -0.00001889, -0.00002042, 0.00000087, 0.00002930, 0.00001804, 0.00000622, 0.00001130, 0.00001293, 0.00000651, 0.00000650, 0.00000488, 0.00000248, 0.00001141, 0.00000431, 0.00000718, 0.00001397, 0.00000326, 0.00001346, 0.00000812, 0.00001118, 0.00002764, 0.00000951, -0.00000198, -0.00002641, -0.00001098, -0.00001368, -0.00000329, -0.00000427, -0.00000357, -0.00000719, 0.00000346, -0.00000967, 0.00000104, -0.00000088, 0.00000199, -0.00000751, -0.00000175, 0.00000053, -0.00000221, -0.00001505, 0.00000229, -0.00000093, -0.00000323, -0.00000959, 0.00000019, -0.00000067, -0.00000254, -0.00000269, -0.00000149, -0.00000128, 0.00000026, -0.00000410, -0.00000059, -0.00000317, -0.00000238, -0.00000560, -0.00000219, -0.00000171, 0.00000114, -0.00001931, 0.00000246, -0.00000484, -0.00000137, 0.00000571, -0.00000753, 0.00000528, -0.00000768, -0.00000344, 0.00000157, 0.00000340, -0.00000452, -0.00000764, -0.00000164, -0.00000239, -0.00000505, -0.00000525, -0.00000160, -0.00000857, -0.00001161, -0.00001269, -0.00001002, -0.00002293, -0.00004173, -0.00010213, 0.00005716, 0.00009397, 0.00002605, 0.00004452, 0.00001386, 0.00002494, 0.00002202, 0.00004418, 0.00001256, 0.00004070, 0.00003531, 0.00011712, -0.00000123, -0.00010159, -0.00001835, -0.00002223, -0.00001184, -0.00001413, -0.00000409, -0.00000902, -0.00000202, -0.00000345, -0.00000203, 0.00000035, -0.00000663, -0.00000456, -0.00000048, -0.00000715, -0.00000164, -0.00000232, -0.00000270, -0.00000379, -0.00000181, -0.00000011, -0.00000219, -0.00000029, -0.00000483, -0.00000251, -0.00000263, 0.00000397, },
		{ -0.00000489, -0.00000816, -0.00000745, -0.00001165, -0.00002347, -0.00000428, 0.00002097, 0.00002564, 0.00000432, 0.00001101, 0.00001000, 0.00001201, 0.00000244, 0.00000905, -0.00000180, 0.00001080, 0.00000624, 0.00000563, 0.00000926, 0.00001169, 0.00000284, 0.00001804, 0.00000044, 0.00002502, 0.00001455, 0.00000872, -0.00001562, -0.00002114, -0.00001135, -0.00000882, -0.00000269, -0.00000197, -0.00000951, 0.00000293, -0.00000599, -0.00000451, 0.00000140, 0.00000120, -0.00000462, -0.00000420, 0.00000002, -0.00000001, -0.00001270, -0.00000382, 0.00000285, -0.00000395, -0.00000772, -0.00000421, 0.00000209, -0.00000331, -0.00000127, -0.00000305, -0.00000102, -0.00000017, -0.00000251, -0.00000243, -0.00000169, -0.00000301, -0.00000433, -0.00000378, -0.00000200, 0.00000308, -0.00001517, -0.00000687, 0.00000249, -0.00000876, 0.00001112, -0.00000939, 0.00000495, -0.00000459, -0.00000494, -0.00000166, 0.00000535, -0.00000162, -0.00000836, -0.00000401, -0.00000146, -0.00000358, -0.00000390, -0.00000589, -0.00000350, -0.00001124, -0.00001218, -0.00001028, -0.00001865, -0.00002353, -0.00008513, -0.00005157, 0.00012418, 0.00004111, 0.00004596, 0.00002397, 0.00002426, 0.00001762, 0.00004399, 0.00002325, 0.00003277, 0.00003236, 0.00010161, 0.00005982, -0.00011169, -0.00003494, -0.00002147, -0.00001169, -0.00001601, -0.00000589, -0.00000884, -0.00000371, -0.00000250, -0.00000399, 0.00000065, -0.00000280, -0.00000779, -0.00000003, -0.00000694, -0.00000247, -0.00000162, -0.00000329, -0.00000349, -0.00000237, 0.00000043, -0.00000283, -0.00000009, -0.00000340, -0.00000324, -0.00000398, 0.00000402, -0.00000454, },
		{ -0.00000476, -0.00001120, -0.00000610, -0.00002359, -0.00000851, 0.00001194, 0.00002975, 0.00000602, 0.00000979, 0.00000745, 0.00001429, 0.00000320, 0.00000834, -0.00000006, 0.00000556, 0.00000863, 0.00000652, 0.00000410, 0.00001467, 0.00000320, 0.00001292, 0.00000904, 0.00000760, 0.00002454, 0.00000773, 0.00000428, -0.00002590, -0.00001429, -0.00001036, -0.00000583, 0.00000102, -0.00000920, -0.00000170, 0.00000013, -0.00000901, 0.00000120, 0.00000139, -0.00000230, -0.00000577, 0.00000009, -0.00000068, -0.00000759, -0.00000990, 0.00000480, -0.00000361, -0.00000574, -0.00000801, 0.00000330, -0.00000349, 0.00000009, -0.00000393, -0.00000097, -0.00000134, -0.00000031, -0.00000427, -0.00000035, -0.00000406, -0.00000232, -0.00000564, -0.00000168, 0.00000182, -0.00000687, -0.00001670, 0.00000753, -0.00001305, 0.00001218, -0.00000689, 0.00000166, -0.00000122, -0.00000462, -0.00000468, 0.00000441, 0.00000198, -0.00000666, -0.00000794, 0.00000028, -0.00000565, 0.00000186, -0.00001079, -0.00000136, -0.00000997, -0.00000909, -0.00001437, -0.00001095, -0.00002476, -0.00004013, -0.00011361, 0.00005926, 0.00010291, 0.00002752, 0.00004844, 0.00001560, 0.00002685, 0.00002756, 0.00004310, 0.00002243, 0.00003788, 0.00007170, 0.00011409, -0.00009714, -0.00006078, -0.00002161, -0.00001350, -0.00001436, -0.00000985, -0.00000796, -0.00000599, -0.00000170, -0.00000432, -0.00000203, 0.00000168, -0.00000877, -0.00000121, -0.00000616, -0.00000395, -0.00000065, -0.00000333, -0.00000393, -0.00000263, 0.00000055, -0.00000218, -0.00000167, -0.00000123, -0.00000368, -0.00000446, 0.00000172, 0.00000011, -0.00000671, },
		{ -0.00001176, -0.00000556, -0.00001933, -0.00001355, 0.00000551, 0.00002761, 0.00001253, 0.00000590, 0.00000875, 0.00001040, 0.00000930, 0.00000386, 0.00000502, -0.00000038, 0.00000974, 0.00000645, 0.00000527, 0.00000733, 0.00001347, 0.00000106, 0.00001961, -0.00000154, 0.00002065, 0.00001344, 0.00001041, -0.00001038, -0.00002672, -0.00000682, -0.00001232, 0.00000283, -0.00000639, -0.00000636, 0.00000254, -0.00000758, -0.00000339, 0.00000302, -0.00000102, -0.00000610, -0.00000031, -0.00000103, -0.00000464, -0.00001138, 0.00000204, -0.00000041, -0.00000560, -0.00000889, 0.00000110, -0.00000125, -0.00000108, -0.00000178, -0.00000307, -0.00000080, -0.00000058, -0.00000315, -0.00000173, -0.00000286, -0.00000277, -0.00000454, -0.00000355, 0.00000107, -0.00000081, -0.00001963, 0.00000431, -0.00000941, 0.00000493, 0.00000257, -0.00000631, 0.00000494, -0.00000735, -0.00000172, -0.00000283, 0.00000811, -0.00000647, -0.00000736, -0.00000345, -0.00000299, -0.00000026, -0.00000347, -0.00001103, -0.00000106, -0.00001285, -0.00000899, -0.00001527, -0.00001703, -0.00002577, -0.00008656, -0.00006352, 0.00013658, 0.00004289, 0.00005138, 0.00002458, 0.00002812, 0.00002052, 0.00004690, 0.00002702, 0.00003845, 0.00004751, 0.00013897, -0.00004772, -0.00009697, -0.00001826, -0.00002363, -0.00000569, -0.00001767, -0.00000419, -0.00001128, 0.00000103, -0.00000581, -0.00000268, 0.00000094, -0.00000382, -0.00000588, -0.00000273, -0.00000812, 0.00000195, -0.00000438, -0.00000287, -0.00000476, 0.00000179, -0.00000201, -0.00000210, -0.00000151, -0.00000180, -0.00000548, -0.00000045, 0.00000272, -0.00000622, -0.00000320, },
		{ -0.00000769, -0.00001417, -0.00001712, 0.00000060, 0.00002159, 0.00002011, 0.00000262, 0.00001066, 0.00000584, 0.00001322, 0.00000296, 0.00000655, -0.00000147, 0.00000715, 0.00000632, 0.00000845, 0.00000094, 0.00001647, 0.00000215, 0.00001296, 0.00000996, 0.00000190, 0.00002389, 0.00000706, 0.00000972, -0.00002944, -0.00001217, -0.00001131, -0.00000347, 0.00000069, -0.00001056, 0.00000221, -0.00000444, -0.00000618, 0.00000095, 0.00000243, -0.00000719, -0.00000051, -0.00000068, -0.00000369, -0.00001031, -0.00000191, 0.00000207, -0.00000470, -0.00000893, -0.00000217, 0.00000120, -0.00000285, 0.00000089, -0.00000463, -0.00000055, -0.00000148, -0.00000138, -0.00000330, -0.00000153, -0.00000389, -0.00000258, -0.00000537, 0.00000004, 0.00000224, -0.00001617, -0.00000412, -0.00000196, -0.00000562, 0.00001279, -0.00001278, 0.00000905, -0.00000945, 0.00000221, -0.00000891, 0.00001013, -0.00000433, -0.00000498, -0.00000885, 0.00000121, -0.00000666, 0.00000770, -0.00001935, 0.00000302, -0.00001429, -0.00000280, -0.00002122, -0.00000767, -0.00002972, -0.00003591, -0.00012694, 0.00006202, 0.00011035, 0.00003258, 0.00004908, 0.00002092, 0.00002737, 0.00003473, 0.00004033, 0.00003690, 0.00003523, 0.00013108, 0.00002552, -0.00012841, -0.00001995, -0.00003462, 0.00000057, -0.00002152, -0.00000276, -0.00001522, 0.00000194, -0.00000636, -0.00000209, -0.00000300, 0.00000237, -0.00000925, -0.00000013, -0.00001142, 0.00000261, -0.00000389, -0.00000219, -0.00000646, 0.00000173, -0.00000090, -0.00000213, -0.00000272, -0.00000025, -0.00000487, -0.00000320, 0.00000386, -0.00000455, -0.00000281, -0.00000985, },
		{ -0.00001105, -0.00001799, -0.00000367, 0.00001445, 0.00002500, 0.00000311, 0.00000971, 0.00000526, 0.00001122, 0.00000687, 0.00000482, 0.00000064, 0.00000348, 0.00000587, 0.00000914, 0.00000240, 0.00000828, 0.00001326, 0.00000032, 0.00002102, -0.00000505, 0.00001703, 0.00001460, 0.00001208, -0.00000975, -0.00002832, -0.00000410, -0.00001383, 0.00000676, -0.00001181, 0.00000070, -0.00000344, -0.00000430, -0.00000471, 0.00000621, -0.00000730, -0.00000170, 0.00000028, -0.00000274, -0.00000992, -0.00000384, 0.00000167, -0.00000194, -0.00000924, -0.00000466, 0.00000181, -0.00000283, 0.00000101, -0.00000290, -0.00000268, -0.00000060, -0.00000187, -0.00000231, -0.00000250, -0.00000307, -0.00000289, -0.00000417, -0.00000314, 0.00000438, -0.00001117, -0.00001079, 0.00000211, -0.00001156, 0.00001531, -0.00001023, 0.00000486, -0.00000464, -0.00000083, -0.00000553, 0.00000251, 0.00000343, -0.00000652, -0.00000785, -0.00000186, -0.00000586, 0.00000701, -0.00001076, -0.00000877, -0.00000430, -0.00000803, -0.00001272, -0.00001511, -0.00001940, -0.00002436, -0.00009119, -0.00007216, 0.00014321, 0.00005240, 0.00004998, 0.00003223, 0.00002641, 0.00003098, 0.00004026, 0.00004511, 0.00002905, 0.00010551, 0.00009452, -0.00013022, -0.00004197, -0.00003265, -0.00000828, -0.00001319, -0.00000940, -0.00001210, -0.00000417, -0.00000186, -0.00000467, -0.00000355, 0.00000230, -0.00000580, -0.00000246, -0.00000951, -0.00000239, 0.00000057, -0.00000426, -0.00000480, -0.00000211, 0.00000262, -0.00000340, -0.00000208, -0.00000188, -0.00000128, -0.00000679, 0.00000444, -0.00000316, -0.00000235, -0.00000746, -0.00000930, },
		{ -0.00001863, -0.00000574, 0.00000662, 0.00002668, 0.00000662, 0.00000711, 0.00000675, 0.00000767, 0.00000935, 0.00000570, 0.00000068, 0.00000276, 0.00000376, 0.00000807, 0.00000666, 0.00000122, 0.00001628, 0.00000207, 0.00001271, 0.00001022, -0.00000476, 0.00002628, 0.00000680, 0.00001163, -0.00003235, -0.00000798, -0.00001426, 0.00000203, -0.00000565, -0.00000399, -0.00000160, -0.00000357, -0.00000625, 0.00000368, -0.00000208, -0.00000630, 0.00000277, -0.00000228, -0.00000893, -0.00000620, 0.00000171, -0.00000084, -0.00000720, -0.00000800, 0.00000201, -0.00000233, 0.00000001, -0.00000031, -0.00000430, -0.00000032, -0.00000236, -0.00000140, -0.00000303, -0.00000263, -0.00000343, -0.00000258, -0.00000531, 0.00000376, -0.00000517, -0.00001483, 0.00000279, -0.00001274, 0.00001066, -0.00000123, -0.00000369, 0.00000323, -0.00000724, 0.00000187, -0.00000667, 0.00000941, -0.00000768, -0.00000392, -0.00000794, -0.00000130, -0.00000122, 0.00000470, -0.00002114, 0.00000269, -0.00001327, -0.00000154, -0.00002513, -0.00000670, -0.00003207, -0.00003472, -0.00013585, 0.00005753, 0.00012539, 0.00003195, 0.00005565, 0.00002106, 0.00003616, 0.00003119, 0.00005279, 0.00003436, 0.00007218, 0.00014149, -0.00009474, -0.00008163, -0.00002108, -0.00002688, -0.00000089, -0.00001726, -0.00000644, -0.00001301, 0.00000352, -0.00000793, -0.00000161, -0.00000216, 0.00000033, -0.00000572, -0.00000543, -0.00000922, 0.00000468, -0.00000576, -0.00000249, -0.00000631, 0.00000488, -0.00000332, -0.00000139, -0.00000406, 0.00000137, -0.00000755, 0.00000221, -0.00000040, -0.00000343, -0.00000385, -0.00001076, -0.00000853, },
		{ -0.00000666, 0.00000026, 0.00002417, 0.00001227, 0.00000438, 0.00000798, 0.00000593, 0.00000817, 0.00000880, 0.00000041, 0.00000227, 0.00000300, 0.00000509, 0.00000924, 0.00000147, 0.00000875, 0.00001301, -0.00000032, 0.00002231, -0.00001060, 0.00001653, 0.00001638, 0.00001227, -0.00001218, -0.00002467, -0.00000610, -0.00000940, 0.00000262, -0.00000909, 0.00000180, -0.00000661, -0.00000196, -0.00000403, 0.00000533, -0.00001091, 0.00000421, -0.00000163, -0.00000627, -0.00001052, 0.00000343, -0.00000202, -0.00000308, -0.00001157, 0.00000186, -0.00000210, -0.00000046, 0.00000044, -0.00000286, -0.00000258, -0.00000073, -0.00000304, -0.00000091, -0.00000442, -0.00000207, -0.00000345, -0.00000369, -0.00000085, 0.00000206, -0.00001770, 0.00000265, -0.00001201, 0.00000354, 0.00000713, -0.00000873, 0.00000645, -0.00000875, 0.00000338, -0.00000727, 0.00000620, -0.00000287, -0.00000440, -0.00000773, -0.00000300, -0.00000364, 0.00000879, -0.00001610, -0.00000654, -0.00000674, -0.00000325, -0.00001821, -0.00001350, -0.00002217, -0.00002353, -0.00009363, -0.00008448, 0.00015160, 0.00006247, 0.00004874, 0.00003999, 0.00002782, 0.00003710, 0.00004025, 0.00005736, 0.00003771, 0.00016210, -0.00003335, -0.00011954, -0.00001714, -0.00003834, 0.00000022, -0.00001614, -0.00000601, -0.00001653, 0.00000238, -0.00000611, -0.00000268, -0.00000379, 0.00000136, -0.00000411, -0.00000414, -0.00001201, 0.00000307, -0.00000286, -0.00000319, -0.00000661, 0.00000222, 0.00000039, -0.00000288, -0.00000346, -0.00000034, -0.00000363, -0.00000314, 0.00000393, -0.00000614, 0.00000046, -0.00001198, -0.00000588, -0.00001953, },
		{ -0.00000075, 0.00001584, 0.00002016, 0.00000128, 0.00000959, 0.00000438, 0.00000756, 0.00000819, 0.00000504, -0.00000212, 0.00000617, 0.00000014, 0.00001030, 0.00000391, 0.00000334, 0.00001462, 0.00000270, 0.00001253, 0.00000816, -0.00000905, 0.00002972, 0.00000676, 0.00000844, -0.00002950, -0.00000814, -0.00001208, 0.00000164, -0.00000799, 0.00000197, -0.00000763, 0.00000005, -0.00000721, 0.00000600, -0.00000875, 0.00000020, 0.00000166, -0.00000435, -0.00001294, 0.00000206, -0.00000055, -0.00000190, -0.00001102, -0.00000176, 0.00000020, -0.00000213, 0.00000106, -0.00000113, -0.00000393, -0.00000015, -0.00000393, 0.00000009, -0.00000468, -0.00000205, -0.00000379, -0.00000205, -0.00000444, 0.00000563, -0.00001599, -0.00000116, -0.00000700, -0.00000628, 0.00001375, -0.00000997, 0.00000582, -0.00000645, 0.00000046, -0.00000299, 0.00000027, 0.00000177, -0.00000563, -0.00000453, -0.00000697, -0.00000295, 0.00000495, -0.00000322, -0.00001703, -0.00000233, -0.00000675, -0.00000689, -0.00002368, -0.00000970, -0.00003118, -0.00003605, -0.00014176, 0.00004680, 0.00014801, 0.00002523, 0.00006860, 0.00001725, 0.00004993, 0.00002197, 0.00007533, 0.00002405, 0.00014855, 0.00004441, -0.00014311, -0.00002755, -0.00003895, -0.00000930, -0.00001052, -0.00000737, -0.00001617, -0.00000335, -0.00000193, -0.00000505, -0.00000308, -0.00000113, -0.00000106, -0.00000306, -0.00001202, -0.00000191, 0.00000106, -0.00000429, -0.00000511, -0.00000222, 0.00000390, -0.00000362, -0.00000227, -0.00000324, 0.00000022, -0.00000657, 0.00000558, -0.00000640, 0.00000136, -0.00000970, -0.00000610, -0.00001704, -0.00001080, },
		{ 0.00000521, 0.00002649, 0.00000028, 0.00001023, 0.00000312, 0.00000846, 0.00000421, 0.00001140, -0.00000567, 0.00000774, -0.00000156, 0.00000802, 0.00000572, 0.00000416, 0.00000690, 0.00001381, -0.00000097, 0.00002187, -0.00001551, 0.00001824, 0.00001865, 0.00000804, -0.00001157, -0.00002183, -0.00000698, -0.00000555, -0.00000407, -0.00000055, -0.00000428, -0.00000378, -0.00000323, -0.00000020, -0.00000170, -0.00000658, 0.00000552, -0.00000346, -0.00001122, -0.00000438, 0.00000458, -0.00000418, -0.00000653, -0.00000800, 0.00000381, -0.00000474, 0.00000237, -0.00000129, -0.00000205, -0.00000258, -0.00000184, -0.00000251, -0.00000126, -0.00000514, -0.00000146, -0.00000355, -0.00000328, 0.00000263, -0.00000798, -0.00001018, 0.00000285, -0.00001758, 0.00001801, -0.00000916, 0.00000465, -0.00000453, -0.00000123, -0.00000158, -0.00000017, 0.00000024, -0.00000277, -0.00000506, -0.00000504, -0.00000742, 0.00000472, 0.00000078, -0.00001224, -0.00001175, -0.00000110, -0.00000684, -0.00001722, -0.00001721, -0.00001960, -0.00002843, -0.00008869, -0.00010704, 0.00016970, 0.00006450, 0.00005654, 0.00003993, 0.00004019, 0.00002950, 0.00006035, 0.00004659, 0.00010243, 0.00012624, -0.00014519, -0.00004971, -0.00003562, -0.00001924, -0.00001032, -0.00000478, -0.00001614, -0.00000801, -0.00000129, -0.00000404, -0.00000451, -0.00000128, -0.00000243, 0.00000124, -0.00001206, -0.00000579, 0.00000142, -0.00000250, -0.00000516, -0.00000412, 0.00000321, -0.00000089, -0.00000318, -0.00000350, -0.00000014, -0.00000506, 0.00000259, -0.00000240, -0.00000247, -0.00000352, -0.00000963, -0.00001057, -0.00001809, 0.00000201, },
		{ 0.00002503, 0.00000564, 0.00000651, 0.00000537, 0.00000683, 0.00000343, 0.00001123, -0.00000107, 0.00000202, 0.00000353, 0.00000148, 0.00000822, 0.00000474, 0.00000358, 0.00001377, 0.00000312, 0.00001189, 0.00000442, -0.00000957, 0.00003286, 0.00000474, 0.00000405, -0.00002305, -0.00001167, -0.00000526, -0.00000663, -0.00000060, -0.00000099, -0.00000744, -0.00000018, -0.00000389, 0.00000139, -0.00000880, 0.00000480, -0.00000071, -0.00000842, -0.00001064, 0.00000690, -0.00000370, -0.00000421, -0.00001096, 0.00000367, -0.00000414, 0.00000120, -0.00000053, -0.00000060, -0.00000382, -0.00000076, -0.00000445, 0.00000096, -0.00000645, -0.00000075, -0.00000437, -0.00000167, -0.00000128, -0.00000031, -0.00001726, 0.00000899, -0.00002121, 0.00001309, -0.00000201, -0.00000058, 0.00000023, -0.00000443, -0.00000014, -0.00000015, -0.00000014, -0.00000256, -0.00000451, -0.00000261, -0.00001044, 0.00000290, 0.00000157, -0.00000401, -0.00001791, -0.00000179, -0.00000536, -0.00000921, -0.00002477, -0.00000934, -0.00003465, -0.00003134, -0.00015512, 0.00004159, 0.00016589, 0.00002501, 0.00007564, 0.00002338, 0.00005107, 0.00003074, 0.00007760, 0.00005770, 0.00018034, -0.00010786, -0.00008913, -0.00002647, -0.00003093, -0.00001100, -0.00000587, -0.00001078, -0.00001440, -0.00000132, -0.00000341, -0.00000493, -0.00000151, -0.00000428, 0.00000308, -0.00000804, -0.00001043, 0.00000070, -0.00000123, -0.00000414, -0.00000537, 0.00000113, 0.00000164, -0.00000322, -0.00000334, -0.00000176, -0.00000278, -0.00000055, 0.00000155, -0.00000618, 0.00000121, -0.00001146, -0.00000550, -0.00002213, 0.00000130, -0.00000015, },
		{ 0.00001550, 0.00000024, 0.00000994, 0.00000274, 0.00000675, 0.00000508, 0.00000801, -0.00000531, 0.00000872, -0.00000308, 0.00000869, 0.00000361, 0.00000677, 0.00000488, 0.00001447, -0.00000101, 0.00001816, -0.00001621, 0.00001955, 0.00002089, -0.00000045, -0.00000494, -0.00002438, -0.00000117, -0.00001066, -0.00000210, 0.00000219, -0.00000792, -0.00000178, -0.00000242, -0.00000028, -0.00000694, 0.00000084, 0.00000284, -0.00000601, -0.00001282, 0.00000296, 0.00000147, -0.00000549, -0.00000941, -0.00000091, 0.00000002, -0.00000276, 0.00000224, -0.00000183, -0.00000146, -0.00000292, -0.00000297, -0.00000112, -0.00000323, -0.00000401, -0.00000186, -0.00000330, -0.00000124, 0.00000151, -0.00001633, 0.00000479, -0.00001208, -0.00000270, 0.00001165, -0.00001057, 0.00000909, -0.00001103, 0.00000484, -0.00000524, 0.00000639, -0.00000881, 0.00000071, -0.00000588, -0.00000444, -0.00000760, 0.00000893, -0.00000691, -0.00000686, -0.00001772, 0.00000615, -0.00001382, -0.00001331, -0.00002245, -0.00001649, -0.00003285, -0.00008425, -0.00013133, 0.00018909, 0.00006671, 0.00006529, 0.00004234, 0.00004954, 0.00002691, 0.00007758, 0.00004878, 0.00018192, -0.00002240, -0.00014470, -0.00001181, -0.00004746, -0.00000552, -0.00001758, 0.00000266, -0.00002390, 0.00000133, -0.00000754, -0.00000043, -0.00000603, -0.00000093, -0.00000272, 0.00000202, -0.00001671, 0.00000146, -0.00000350, -0.00000014, -0.00000801, 0.00000093, 0.00000062, -0.00000011, -0.00000498, -0.00000135, -0.00000336, -0.00000078, 0.00000261, -0.00000634, 0.00000107, -0.00000874, -0.00000514, -0.00001944, -0.00000563, 0.00000145, 0.00001612, },
		{ -0.00000208, 0.00001115, 0.00000146, 0.00000825, 0.00000130, 0.00001149, -0.00000496, 0.00000641, -0.00000045, 0.00000446, 0.00000418, 0.00000799, 0.00000220, 0.00001340, 0.00000418, 0.00000826, 0.00000316, -0.00000890, 0.00003619, -0.00000197, 0.00000478, -0.00002033, -0.00001063, -0.00000376, -0.00001192, 0.00000777, -0.00000771, -0.00000334, -0.00000236, 0.00000044, -0.00000742, -0.00000095, 0.00000332, -0.00000235, -0.00001262, -0.00000291, 0.00000529, -0.00000474, -0.00000855, -0.00000461, 0.00000254, -0.00000473, 0.00000299, -0.00000208, 0.00000022, -0.00000379, -0.00000225, -0.00000307, -0.00000057, -0.00000602, -0.00000064, -0.00000426, -0.00000092, 0.00000132, -0.00001167, -0.00000415, 0.00000088, -0.00001747, 0.00002015, -0.00001503, 0.00001236, -0.00001205, 0.00000544, -0.00000746, 0.00000982, -0.00000995, 0.00000049, -0.00000664, 0.00000101, -0.00001503, 0.00001187, -0.00000918, 0.00000331, -0.00002567, 0.00000649, -0.00001271, -0.00000427, -0.00003153, -0.00000383, -0.00004308, -0.00002078, -0.00017633, 0.00004338, 0.00017660, 0.00003415, 0.00007564, 0.00003878, 0.00004132, 0.00005708, 0.00006151, 0.00015059, 0.00008095, -0.00018415, -0.00001544, -0.00005278, -0.00000610, -0.00002626, 0.00000874, -0.00002422, -0.00000153, -0.00000923, 0.00000168, -0.00000848, 0.00000186, -0.00000850, 0.00000840, -0.00001661, -0.00000154, -0.00000413, 0.00000119, -0.00000749, -0.00000083, -0.00000006, 0.00000164, -0.00000467, -0.00000186, -0.00000351, -0.00000183, 0.00000387, -0.00000536, -0.00000098, -0.00000495, -0.00000632, -0.00001392, -0.00001351, 0.00000380, 0.00000710, 0.00002231, },
		{ 0.00000765, 0.00000428, 0.00000595, 0.00000260, 0.00000811, 0.00000127, 0.00000063, 0.00000403, 0.00000047, 0.00000472, 0.00000567, 0.00000677, 0.00000406, 0.00001497, -0.00000225, 0.00001385, -0.00001334, 0.00002024, 0.00001995, -0.00000711, 0.00000206, -0.00002701, 0.00000536, -0.00002007, 0.00000776, -0.00000397, -0.00000432, -0.00000462, 0.00000317, -0.00000844, -0.00000260, 0.00000289, 0.00000063, -0.00001058, -0.00000737, 0.00000454, -0.00000048, -0.00000929, -0.00000618, 0.00000201, -0.00000333, 0.00000049, 0.00000031, -0.00000126, -0.00000102, -0.00000438, -0.00000220, -0.00000164, -0.00000389, -0.00000316, -0.00000199, -0.00000335, 0.00000285, -0.00000826, -0.00001042, 0.00000737, -0.00002052, 0.00001513, -0.00000727, 0.00000389, -0.00000210, -0.00000352, -0.00000043, 0.00000218, 0.00000129, -0.00001005, 0.00000028, -0.00000277, -0.00000885, 0.00000034, 0.00000158, -0.00000402, -0.00001026, -0.00001425, 0.00000392, -0.00001351, -0.00001548, -0.00002129, -0.00002039, -0.00002877, -0.00009014, -0.00014570, 0.00019672, 0.00008274, 0.00006221, 0.00006073, 0.00003952, 0.00005189, 0.00006250, 0.00012106, 0.00016045, -0.00017050, -0.00005998, -0.00003262, -0.00002560, -0.00001742, -0.00000532, -0.00000800, -0.00001483, -0.00000241, -0.00000530, -0.00000263, -0.00000276, -0.00000564, 0.00000304, -0.00000576, -0.00001077, 0.00000037, -0.00000378, -0.00000146, -0.00000623, 0.00000263, -0.00000107, -0.00000031, -0.00000516, -0.00000077, -0.00000577, 0.00000638, -0.00000438, -0.00000268, -0.00000277, -0.00000650, -0.00000920, -0.00001794, 0.00000271, 0.00000217, 0.00002270, 0.00000155, },
		{ 0.00000725, 0.00000373, 0.00000438, 0.00000468, 0.00000486, -0.00000060, 0.00000376, 0.00000118, 0.00000279, 0.00000381, 0.00000924, 0.00000184, 0.00001234, 0.00000577, 0.00000151, 0.00000679, -0.00000919, 0.00003825, -0.00001066, 0.00001009, -0.00002205, -0.00000380, -0.00001163, -0.00000601, 0.00000639, -0.00000785, -0.00000357, 0.00000051, -0.00000288, -0.00000952, 0.00000575, -0.00000026, -0.00000545, -0.00001140, 0.00000272, 0.00000181, -0.00000661, -0.00000960, 0.00000210, -0.00000239, -0.00000124, 0.00000153, -0.00000202, 0.00000070, -0.00000450, -0.00000280, -0.00000207, -0.00000261, -0.00000428, -0.00000131, -0.00000410, 0.00000215, -0.00000337, -0.00001513, 0.00000821, -0.00001537, 0.00000361, 0.00000423, -0.00000717, 0.00000876, -0.00001183, 0.00000630, -0.00000635, 0.00001205, -0.00001658, 0.00000276, -0.00000600, -0.00000040, -0.00001154, 0.00001114, -0.00001300, 0.00000578, -0.00002802, 0.00001008, -0.00001760, -0.00000151, -0.00003487, -0.00000293, -0.00004522, -0.00001794, -0.00019010, 0.00003591, 0.00019659, 0.00003644, 0.00008731, 0.00003991, 0.00005184, 0.00006126, 0.00009126, 0.00020878, -0.00010514, -0.00012878, -0.00000667, -0.00004832, -0.00000397, -0.00002463, 0.00000779, -0.00002454, 0.00000283, -0.00001255, 0.00000315, -0.00000790, -0.00000025, -0.00000583, 0.00000504, -0.00001658, 0.00000279, -0.00000793, 0.00000259, -0.00000890, 0.00000348, -0.00000284, 0.00000229, -0.00000617, 0.00000029, -0.00000745, 0.00000494, 0.00000024, -0.00000607, -0.00000024, -0.00000762, -0.00000439, -0.00001991, 0.00000005, -0.00000071, 0.00001947, 0.00000731, 0.00000344, },
		{ 0.00000290, 0.00000498, 0.00000334, 0.00000503, 0.00000085, 0.00000252, 0.00000174, 0.00000249, 0.00000151, 0.00000810, 0.00000636, 0.00000301, 0.00001657, -0.00000734, 0.00001372, -0.00001055, 0.00002092, 0.00001438, -0.00000711, 0.00000316, -0.00002250, 0.00000234, -0.00002063, 0.00001244, -0.00000860, -0.00000093, -0.00000581, 0.00000556, -0.00001625, 0.00000677, -0.00000068, -0.00000107, -0.00001331, 0.00000090, 0.00000088, -0.00000108, -0.00001337, 0.00000187, -0.00000187, -0.00000153, 0.00000061, -0.00000047, -0.00000084, -0.00000087, -0.00000610, -0.00000014, -0.00000437, -0.00000197, -0.00000395, -0.00000141, -0.00000229, 0.00000371, -0.00001881, 0.00000604, -0.00000843, -0.00000565, 0.00001060, -0.00001163, 0.00001082, -0.00001035, 0.00000391, -0.00000566, 0.00001065, -0.00000876, -0.00000636, -0.00000134, -0.00000090, -0.00001011, 0.00000573, -0.00000724, 0.00000200, -0.00001606, -0.00000734, -0.00000340, -0.00000901, -0.00002001, -0.00001887, -0.00002446, -0.00002482, -0.00009713, -0.00015912, 0.00020088, 0.00010450, 0.00005822, 0.00008018, 0.00002997, 0.00008072, 0.00005073, 0.00023084, -0.00000969, -0.00018589, -0.00000628, -0.00005136, -0.00000589, -0.00002866, 0.00000516, -0.00001852, -0.00000268, -0.00001054, -0.00000039, -0.00000527, -0.00000112, -0.00000747, 0.00000528, -0.00001183, -0.00000209, -0.00000547, -0.00000085, -0.00000479, -0.00000098, 0.00000026, -0.00000057, -0.00000214, -0.00000287, -0.00000407, -0.00000196, 0.00000789, -0.00000987, 0.00000211, -0.00000920, -0.00000054, -0.00001921, -0.00000364, -0.00000181, 0.00001355, 0.00001357, 0.00000008, 0.00000896, },
		{ 0.00000651, 0.00000118, 0.00000655, -0.00000102, 0.00000565, -0.00000182, 0.00000587, -0.00000202, 0.00000675, 0.00000824, 0.00000226, 0.00001187, 0.00000460, -0.00000353, 0.00001229, -0.00000935, 0.00003538, -0.00001410, 0.00001208, -0.00001980, -0.00000166, -0.00001779, 0.00000283, -0.00000060, -0.00000276, -0.00000696, 0.00000603, -0.00001255, -0.00000159, 0.00000491, -0.00000209, -0.00000990, -0.00000377, 0.00000165, 0.00000047, -0.00001105, -0.00000364, 0.00000208, -0.00000393, 0.00000126, -0.00000043, -0.00000131, 0.00000100, -0.00000622, -0.00000120, -0.00000379, -0.00000209, -0.00000389, -0.00000125, -0.00000400, 0.00000589, -0.00001528, -0.00000261, 0.00000146, -0.00001405, 0.00001451, -0.00001257, 0.00000813, -0.00000494, -0.00000087, -0.00000226, 0.00000478, 0.00000203, -0.00001354, -0.00000004, -0.00000176, -0.00000527, -0.00000212, 0.00000028, -0.00000623, -0.00000116, -0.00002006, 0.00000265, -0.00001335, -0.00000620, -0.00003099, -0.00000933, -0.00003883, -0.00002612, -0.00019152, 0.00001269, 0.00023279, 0.00002682, 0.00011462, 0.00002270, 0.00009018, 0.00003355, 0.00020382, 0.00011059, -0.00021586, -0.00003451, -0.00004052, -0.00001717, -0.00002301, -0.00000686, -0.00000873, -0.00000895, -0.00000652, -0.00000639, -0.00000159, -0.00000312, -0.00000596, 0.00000086, -0.00000487, -0.00000661, -0.00000237, -0.00000541, -0.00000088, -0.00000448, 0.00000269, -0.00000304, 0.00000069, -0.00000457, -0.00000134, -0.00000716, 0.00001061, -0.00000748, -0.00000055, -0.00000663, -0.00000216, -0.00001274, -0.00001094, 0.00000156, 0.00000342, 0.00002100, -0.00000330, 0.00001064, 0.00000172, },
		{ -0.00000025, 0.00000799, -0.00000331, 0.00000787, -0.00000283, 0.00000580, -0.00000109, 0.00000312, 0.00000750, 0.00000735, 0.00000161, 0.00001747, -0.00001394, 0.00001777, -0.00000853, 0.00001874, 0.00000966, -0.00000507, 0.00000251, -0.00001562, -0.00000720, -0.00001083, 0.00000640, -0.00000483, -0.00000357, -0.00000065, -0.00000183, -0.00001329, 0.00001045, -0.00000377, -0.00000499, -0.00000916, 0.00000416, -0.00000223, -0.00000409, -0.00001151, 0.00000653, -0.00000644, 0.00000257, -0.00000186, 0.00000032, -0.00000077, -0.00000176, -0.00000598, 0.00000011, -0.00000599, 0.00000007, -0.00000492, -0.00000107, 0.00000123, -0.00000435, -0.00001577, 0.00001206, -0.00002107, 0.00001683, -0.00001164, 0.00000427, -0.00000116, -0.00000203, -0.00000239, 0.00000264, 0.00000528, -0.00000949, -0.00000788, 0.00000262, -0.00000621, -0.00000087, -0.00000252, -0.00000372, -0.00000255, -0.00001068, -0.00001134, -0.00000210, -0.00001221, -0.00001595, -0.00002618, -0.00001664, -0.00003542, -0.00008847, -0.00019012, 0.00022057, 0.00011340, 0.00007403, 0.00007544, 0.00005498, 0.00006876, 0.00012206, 0.00023571, -0.00020467, -0.00008318, -0.00003105, -0.00002514, -0.00002017, -0.00001374, -0.00000818, -0.00000758, -0.00000714, -0.00000733, -0.00000402, -0.00000037, -0.00000743, -0.00000024, -0.00000378, -0.00000523, -0.00000312, -0.00000567, -0.00000244, -0.00000294, 0.00000070, -0.00000097, -0.00000154, -0.00000177, -0.00000259, -0.00000652, 0.00000539, 0.00000147, -0.00000725, -0.00000061, -0.00000811, -0.00000299, -0.00001919, 0.00000644, -0.00000647, 0.00002486, -0.00000290, 0.00000941, 0.00000231, 0.00000713, },
		{ 0.00000447, 0.00000044, 0.00000223, 0.00000433, -0.00000149, 0.00000584, -0.00000325, 0.00000787, 0.00000810, 0.00000186, 0.00001256, -0.00000067, -0.00000301, 0.00001626, -0.00000964, 0.00002865, -0.00001087, 0.00000807, -0.00001033, -0.00001064, -0.00001131, 0.00000145, -0.00000106, -0.00000219, -0.00000483, 0.00000365, -0.00001602, 0.00000642, 0.00000079, -0.00000450, -0.00001012, 0.00000324, -0.00000229, -0.00000119, -0.00001298, 0.00000404, -0.00000310, -0.00000045, -0.00000003, -0.00000056, -0.00000098, 0.00000063, -0.00000717, 0.00000037, -0.00000654, 0.00000043, -0.00000497, -0.00000079, -0.00000189, 0.00000334, -0.00002154, 0.00001252, -0.00001759, 0.00000909, -0.00000128, -0.00000642, 0.00000612, -0.00000588, 0.00000088, -0.00000217, 0.00000836, -0.00000551, -0.00001072, -0.00000010, -0.00000286, -0.00000189, -0.00000206, -0.00000431, -0.00000372, -0.00000341, -0.00001583, -0.00000314, -0.00001088, -0.00000742, -0.00003202, -0.00000894, -0.00004064, -0.00002672, -0.00020017, -0.00000664, 0.00026503, 0.00002924, 0.00012776, 0.00002609, 0.00010732, 0.00004981, 0.00029904, -0.00011622, -0.00015984, -0.00001738, -0.00003960, -0.00001143, -0.00002323, -0.00000622, -0.00001122, -0.00000224, -0.00001093, -0.00000450, -0.00000127, -0.00000490, -0.00000376, -0.00000167, -0.00000631, -0.00000066, -0.00000708, -0.00000343, -0.00000297, 0.00000025, 0.00000006, -0.00000268, -0.00000082, -0.00000227, -0.00000552, -0.00000035, 0.00000785, -0.00000942, 0.00000143, -0.00001059, 0.00000150, -0.00002040, 0.00000477, -0.00000752, 0.00001802, 0.00000625, 0.00000165, 0.00000782, 0.00000340, 0.00000302, },
		{ 0.00000695, -0.00000596, 0.00001174, -0.00000674, 0.00000931, -0.00000488, 0.00000590, 0.00000605, 0.00000799, 0.00000193, 0.00001381, -0.00001597, 0.00002132, -0.00000560, 0.00001076, 0.00001228, -0.00000877, 0.00000882, -0.00001827, -0.00000854, -0.00000471, 0.00000030, -0.00000019, -0.00000549, 0.00000267, -0.00001090, -0.00000300, 0.00000675, -0.00000501, -0.00000861, -0.00000071, 0.00000139, -0.00000353, -0.00000828, -0.00000360, 0.00000357, -0.00000563, 0.00000370, -0.00000349, 0.00000124, -0.00000107, -0.00000304, -0.00000423, -0.00000230, -0.00000446, 0.00000004, -0.00000493, -0.00000013, 0.00000292, -0.00001556, 0.00000062, -0.00000290, -0.00000863, 0.00001682, -0.00002101, 0.00001523, -0.00001285, 0.00000861, -0.00001096, 0.00001465, -0.00000789, -0.00000287, -0.00001244, 0.00000716, -0.00001063, 0.00000863, -0.00001520, 0.00000575, -0.00001246, 0.00000117, -0.00002322, 0.00000588, -0.00002019, -0.00000801, -0.00003664, -0.00000522, -0.00005195, -0.00007226, -0.00023151, 0.00024820, 0.00012019, 0.00009523, 0.00006553, 0.00009298, 0.00004695, 0.00026006, 0.00004984, -0.00024766, -0.00000480, -0.00006320, 0.00000567, -0.00003932, 0.00000406, -0.00002570, 0.00001041, -0.00001921, 0.00000102, -0.00001009, 0.00000470, -0.00001276, 0.00000577, -0.00001374, 0.00000664, -0.00001104, -0.00000088, -0.00000762, 0.00000369, -0.00000253, 0.00000046, -0.00000459, 0.00000168, -0.00000691, -0.00000214, 0.00000724, -0.00000448, -0.00000215, -0.00000731, -0.00000160, -0.00001337, -0.00000428, 0.00000019, 0.00000330, 0.00001920, -0.00000732, 0.00001350, -0.00000097, 0.00000787, -0.00000110, },
		{ -0.00000684, 0.00000936, -0.00000176, 0.00000354, 0.00000125, -0.00000017, 0.00000623, 0.00000880, 0.00000214, 0.00001064, -0.00000487, 0.00000009, 0.00001988, -0.00001416, 0.00002567, -0.00000892, 0.00000638, -0.00000403, -0.00002013, 0.00000075, -0.00000772, 0.00000529, -0.00000636, 0.00000198, -0.00000743, -0.00000743, 0.00000573, -0.00000082, -0.00000920, -0.00000324, 0.00000353, -0.00000415, -0.00000585, -0.00000745, 0.00000517, -0.00000563, 0.00000288, -0.00000256, 0.00000043, -0.00000101, -0.00000078, -0.00000582, -0.00000098, -0.00000658, 0.00000165, -0.00000545, -0.00000064, 0.00000187, -0.00000775, -0.00001098, 0.00000879, -0.00002077, 0.00002548, -0.00002263, 0.00001186, -0.00001000, 0.00000735, -0.00001003, 0.00001136, -0.00000356, -0.00000084, -0.00001514, 0.00000588, -0.00001038, 0.00001072, -0.00001531, 0.00000462, -0.00001355, 0.00000755, -0.00002523, 0.00000272, -0.00001852, 0.00000157, -0.00004383, 0.00000414, -0.00005811, -0.00001014, -0.00022756, -0.00001016, 0.00028239, 0.00005510, 0.00011356, 0.00006989, 0.00007642, 0.00017612, 0.00021244, -0.00027618, -0.00003596, -0.00006420, 0.00000171, -0.00003743, 0.00000126, -0.00002854, 0.00000849, -0.00001431, -0.00000173, -0.00001253, 0.00000677, -0.00001390, 0.00000625, -0.00001442, 0.00000534, -0.00000666, -0.00000348, -0.00000800, 0.00000208, -0.00000096, 0.00000065, -0.00000535, 0.00000160, -0.00000452, -0.00000510, 0.00000584, -0.00000008, -0.00000412, -0.00000450, -0.00000554, -0.00000604, -0.00001232, 0.00000700, -0.00000755, 0.00002424, -0.00000753, 0.00001134, 0.00000077, 0.00000766, -0.00000160, 0.00000797, },
		{ 0.00000073, 0.00000678, -0.00000302, 0.00000624, -0.00000328, 0.00000538, 0.00000619, 0.00000781, 0.00000261, 0.00000705, -0.00001229, 0.00002274, -0.00000387, 0.00000143, 0.00001906, -0.00001566, 0.00001902, -0.00002958, 0.00000256, -0.00001059, 0.00000340, -0.00000255, 0.00000023, -0.00000498, -0.00000870, 0.00000197, 0.00000301, -0.00000825, -0.00000587, 0.00000375, -0.00000167, -0.00000690, -0.00000706, 0.00000212, -0.00000150, -0.00000137, 0.00000143, -0.00000313, 0.00000166, -0.00000235, -0.00000257, -0.00000410, -0.00000395, -0.00000226, -0.00000074, -0.00000498, 0.00000340, -0.00000409, -0.00001455, 0.00000931, -0.00001878, 0.00001654, -0.00000608, -0.00000600, 0.00000440, -0.00000576, 0.00000353, -0.00000402, 0.00001127, -0.00001087, -0.00000229, -0.00001013, 0.00000284, -0.00000283, 0.00000316, -0.00001454, 0.00000369, -0.00001033, 0.00000150, -0.00002584, 0.00000531, -0.00001842, -0.00001131, -0.00003343, -0.00000995, -0.00005178, -0.00007329, -0.00025751, 0.00025697, 0.00015543, 0.00008514, 0.00009901, 0.00008089, 0.00012123, 0.00029453, -0.00019823, -0.00012919, -0.00002820, -0.00003784, -0.00000665, -0.00002523, -0.00000783, -0.00001707, 0.00000816, -0.00001694, -0.00000201, -0.00000577, -0.00000114, -0.00000610, -0.00000204, -0.00000923, 0.00000764, -0.00001248, -0.00000129, -0.00000720, 0.00000652, -0.00000433, -0.00000012, -0.00000486, 0.00000271, -0.00001041, 0.00000640, 0.00000050, -0.00000225, -0.00000417, -0.00000698, -0.00000278, -0.00001485, 0.00000758, -0.00000903, 0.00001815, 0.00000139, 0.00000267, 0.00000690, 0.00000348, 0.00000244, 0.00000384, -0.00000136, },
		{ 0.00000873, -0.00000200, 0.00000411, -0.00000020, 0.00000146, 0.00000542, 0.00000816, 0.00000465, 0.00000336, -0.00000259, 0.00000104, 0.00002418, -0.00002360, 0.00003002, -0.00001311, 0.00001248, -0.00000984, -0.00001646, 0.00000294, -0.00001094, 0.00000739, -0.00000518, 0.00000145, -0.00001363, 0.00000310, 0.00000020, -0.00000187, -0.00001110, 0.00000468, -0.00000023, -0.00000560, -0.00000907, 0.00000135, -0.00000022, -0.00000233, 0.00000207, -0.00000320, 0.00000107, -0.00000162, -0.00000170, -0.00000404, -0.00000398, -0.00000381, 0.00000066, -0.00000503, 0.00000096, 0.00000072, -0.00001585, 0.00000588, -0.00001146, 0.00000194, 0.00001223, -0.00001986, 0.00001295, -0.00001394, 0.00001194, -0.00001289, 0.00001839, -0.00001447, 0.00000478, -0.00001741, 0.00000695, -0.00001032, 0.00001513, -0.00002370, 0.00001067, -0.00002015, 0.00001631, -0.00003441, 0.00000796, -0.00002356, 0.00000632, -0.00004949, 0.00001001, -0.00006924, 0.00000044, -0.00025007, -0.00002297, 0.00031423, 0.00006869, 0.00011834, 0.00009687, 0.00008277, 0.00031234, -0.00005748, -0.00022868, -0.00000022, -0.00007441, 0.00001661, -0.00004422, 0.00000843, -0.00003708, 0.00001991, -0.00002196, 0.00000331, -0.00001529, 0.00000785, -0.00001470, 0.00000723, -0.00001946, 0.00001404, -0.00001323, 0.00000046, -0.00001202, 0.00000818, -0.00000441, 0.00000182, -0.00000788, 0.00000486, -0.00000972, 0.00000238, 0.00000359, -0.00000310, -0.00000105, -0.00000947, -0.00000046, -0.00001525, 0.00000515, -0.00000554, 0.00000813, 0.00001065, -0.00000392, 0.00001004, 0.00000151, 0.00000507, 0.00000098, 0.00000271, -0.00000442, },
		{ 0.00000167, 0.00000169, 0.00000157, 0.00000047, 0.00000355, 0.00000601, 0.00000954, -0.00000133, 0.00000435, -0.00000832, 0.00002463, -0.00000768, -0.00000121, 0.00002088, -0.00001582, 0.00001994, -0.00003378, 0.00001178, -0.00001824, 0.00000870, -0.00000558, 0.00000594, -0.00001662, 0.00000313, -0.00000207, 0.00000230, -0.00001263, 0.00000254, 0.00000146, -0.00000249, -0.00001229, 0.00000124, -0.00000029, -0.00000135, 0.00000035, -0.00000049, -0.00000229, 0.00000166, -0.00000421, -0.00000013, -0.00000709, -0.00000196, -0.00000251, -0.00000033, -0.00000501, 0.00000658, -0.00001660, 0.00000231, -0.00000492, -0.00000880, 0.00002027, -0.00001895, 0.00000761, -0.00000949, 0.00000664, -0.00000676, 0.00000985, -0.00000315, -0.00000388, -0.00000690, -0.00000499, -0.00000323, 0.00000830, -0.00000991, -0.00000490, -0.00000673, 0.00000128, -0.00000890, -0.00001916, -0.00000244, -0.00001037, -0.00002041, -0.00002347, -0.00002384, -0.00004103, -0.00008656, -0.00027339, 0.00025432, 0.00020872, 0.00005831, 0.00016134, 0.00003852, 0.00030161, 0.00008771, -0.00027692, -0.00002111, -0.00007387, 0.00000256, -0.00003227, -0.00000130, -0.00003012, 0.00000412, -0.00000623, -0.00000594, -0.00001054, 0.00000153, -0.00000889, 0.00000251, -0.00001408, 0.00000370, -0.00000089, -0.00000678, -0.00000717, -0.00000047, 0.00000345, -0.00000281, -0.00000346, -0.00000126, -0.00000169, -0.00000652, 0.00000974, -0.00000690, 0.00000380, -0.00001153, 0.00000044, -0.00001388, 0.00000067, 0.00000014, -0.00000123, 0.00001602, -0.00000589, 0.00000926, 0.00000191, 0.00000584, 0.00000025, 0.00000417, -0.00000532, 0.00000583, },
		{ 0.00000436, -0.00000139, 0.00000367, -0.00000028, 0.00000592, 0.00000767, 0.00000545, -0.00000428, 0.00000363, 0.00000079, 0.00002496, -0.00003040, 0.00003423, -0.00001585, 0.00001680, -0.00001842, -0.00000466, -0.00000476, -0.00000574, 0.00000309, 0.00000260, -0.00000979, -0.00000624, 0.00000429, -0.00000250, -0.00000579, -0.00000560, 0.00000625, -0.00000236, -0.00001005, -0.00000487, 0.00000431, -0.00000414, 0.00000197, -0.00000093, -0.00000210, 0.00000106, -0.00000315, -0.00000041, -0.00000504, -0.00000449, -0.00000195, -0.00000018, -0.00000488, 0.00000475, -0.00000962, -0.00000728, 0.00000534, -0.00001891, 0.00002267, -0.00001230, 0.00000043, -0.00000453, 0.00000014, 0.00000017, 0.00000144, 0.00000648, -0.00001040, 0.00000061, -0.00001076, -0.00000172, 0.00000258, 0.00000272, -0.00001531, 0.00000114, -0.00001029, 0.00000833, -0.00002892, 0.00000004, -0.00001493, -0.00000385, -0.00003760, -0.00000502, -0.00005744, -0.00001419, -0.00024714, -0.00006627, 0.00038325, 0.00004384, 0.00017934, 0.00005670, 0.00023033, 0.00024263, -0.00028096, -0.00007142, -0.00005736, -0.00002580, -0.00001433, -0.00001502, -0.00001699, -0.00001530, 0.00000592, -0.00001051, -0.00000653, -0.00000558, -0.00000304, -0.00000294, -0.00000648, -0.00000758, 0.00000776, -0.00000907, -0.00000427, -0.00000719, 0.00000705, -0.00000354, -0.00000126, -0.00000519, 0.00000247, -0.00000970, 0.00000946, -0.00000448, 0.00000154, -0.00000607, -0.00000514, -0.00000681, -0.00000873, 0.00000911, -0.00001038, 0.00001936, -0.00000649, 0.00000838, 0.00000124, 0.00000745, -0.00000096, 0.00000603, -0.00000615, 0.00000429, 0.00000116, },
		{ -0.00000081, 0.00000260, 0.00000069, 0.00000367, 0.00000436, 0.00001268, -0.00000974, 0.00000843, -0.00000707, 0.00002607, -0.00001424, 0.00000180, 0.00001805, -0.00000999, 0.00000968, -0.00002338, 0.00000766, -0.00001592, 0.00000620, -0.00000008, 0.00000005, -0.00001712, 0.00001054, -0.00000702, 0.00000016, -0.00001159, 0.00000840, -0.00000228, -0.00000475, -0.00001313, 0.00000886, -0.00000642, 0.00000352, -0.00000216, 0.00000009, -0.00000225, 0.00000096, -0.00000473, 0.00000127, -0.00001005, 0.00000166, -0.00000445, 0.00000070, -0.00000296, 0.00000215, -0.00001929, 0.00001532, -0.00002516, 0.00001972, -0.00000436, -0.00000338, -0.00000330, -0.00000252, 0.00000133, 0.00000040, 0.00000690, -0.00000649, -0.00000391, -0.00000315, -0.00001055, 0.00000482, 0.00000287, -0.00000836, -0.00000809, -0.00000382, 0.00000014, -0.00000878, -0.00002258, 0.00000175, -0.00001650, -0.00001237, -0.00003341, -0.00001675, -0.00005264, -0.00007733, -0.00031610, 0.00028147, 0.00023069, 0.00008014, 0.00016848, 0.00010077, 0.00039094, -0.00023740, -0.00013697, -0.00004360, -0.00004687, -0.00001115, -0.00001628, -0.00001400, -0.00002069, 0.00000077, -0.00000266, -0.00000989, -0.00000506, -0.00000570, -0.00000122, -0.00000556, -0.00000959, 0.00000442, -0.00000193, -0.00000722, -0.00000667, 0.00000189, 0.00000249, -0.00000428, -0.00000299, -0.00000087, -0.00000492, 0.00000168, 0.00000449, -0.00000680, 0.00000383, -0.00001307, 0.00000173, -0.00001823, 0.00001570, -0.00001447, 0.00001756, -0.00000339, 0.00000496, 0.00000245, 0.00000671, 0.00000015, 0.00000576, -0.00000473, 0.00000180, 0.00000077, 0.00000519, },
		{ -0.00000451, 0.00000659, -0.00000135, 0.00000486, 0.00000999, 0.00000010, -0.00000503, 0.00000755, 0.00000196, 0.00001945, -0.00002863, 0.00003120, -0.00000947, 0.00000796, -0.00001364, -0.00000205, -0.00000754, -0.00000393, 0.00000334, 0.00000420, -0.00001797, 0.00000524, -0.00000109, -0.00000333, -0.00000854, 0.00000269, 0.00000263, -0.00000342, -0.00001411, 0.00000405, -0.00000038, -0.00000171, 0.00000203, -0.00000215, -0.00000121, 0.00000049, -0.00000443, 0.00000176, -0.00000831, -0.00000144, -0.00000297, 0.00000037, -0.00000399, 0.00000528, -0.00001984, 0.00001129, -0.00001561, 0.00000285, 0.00001197, -0.00001287, 0.00000486, -0.00001143, 0.00000707, -0.00000511, 0.00001133, -0.00000673, -0.00000350, -0.00000281, -0.00000773, -0.00000433, 0.00000930, -0.00000721, -0.00000747, -0.00000749, -0.00000018, -0.00000183, -0.00002197, -0.00000689, -0.00000915, -0.00000950, -0.00003081, -0.00001614, -0.00004871, -0.00002750, -0.00024691, -0.00010788, 0.00044945, 0.00003303, 0.00023364, 0.00004013, 0.00042846, -0.00008139, -0.00025047, -0.00001020, -0.00007991, -0.00000148, -0.00002831, -0.00000109, -0.00003086, 0.00000054, -0.00000557, -0.00000196, -0.00001194, -0.00000212, -0.00000659, 0.00000110, -0.00001429, 0.00000336, -0.00000084, -0.00000317, -0.00000970, -0.00000022, 0.00000305, -0.00000184, -0.00000442, -0.00000129, -0.00000345, -0.00000204, 0.00000822, -0.00000868, 0.00000543, -0.00001177, 0.00000130, -0.00001753, 0.00001026, -0.00000600, 0.00000513, 0.00000814, -0.00000532, 0.00000960, 0.00000003, 0.00000660, 0.00000028, 0.00000219, -0.00000548, 0.00000572, -0.00000121, 0.00000702, },
		{ 0.00000691, -0.00000111, 0.00000384, 0.00000456, 0.00001130, -0.00001465, 0.00001339, -0.00000531, 0.00002169, -0.00001447, 0.00000065, 0.00002102, -0.00001195, 0.00000429, -0.00001212, -0.00000079, -0.00000933, 0.00000115, 0.00000758, -0.00001239, -0.00000564, 0.00000741, -0.00000786, -0.00000442, -0.00000377, 0.00000718, -0.00000373, -0.00000953, -0.00000626, 0.00000836, -0.00000799, 0.00000671, -0.00000519, 0.00000180, -0.00000300, -0.00000017, -0.00000330, -0.00000050, -0.00000931, 0.00000295, -0.00000540, 0.00000158, -0.00000028, -0.00000935, -0.00000476, 0.00000493, -0.00002060, 0.00002987, -0.00002365, 0.00001589, -0.00002195, 0.00001408, -0.00001348, 0.00001906, -0.00001159, 0.00000401, -0.00001267, 0.00000713, -0.00002200, 0.00002089, -0.00001506, 0.00000690, -0.00002439, 0.00001347, -0.00001656, 0.00000528, -0.00003778, 0.00001732, -0.00003443, 0.00000938, -0.00006035, 0.00000999, -0.00008703, -0.00004469, -0.00038510, 0.00033456, 0.00022789, 0.00014913, 0.00012099, 0.00031748, 0.00018190, -0.00038324, 0.00002010, -0.00011358, 0.00001306, -0.00005248, 0.00002134, -0.00004747, 0.00001035, -0.00002228, 0.00001591, -0.00002489, 0.00000863, -0.00002002, 0.00001368, -0.00002310, 0.00000808, -0.00000837, 0.00000748, -0.00001608, 0.00000268, -0.00000363, 0.00000644, -0.00001011, 0.00000300, -0.00000747, 0.00000037, 0.00000450, -0.00000202, -0.00000211, -0.00000223, -0.00000648, -0.00000827, -0.00000420, 0.00000967, -0.00001122, 0.00002107, -0.00001524, 0.00001551, -0.00000551, 0.00001164, -0.00000425, 0.00000854, -0.00001136, 0.00000950, -0.00000650, 0.00001158, -0.00000651, },
		{ 0.00000531, -0.00000046, 0.00000331, 0.00001162, -0.00000716, -0.00000035, 0.00000922, 0.00000094, 0.00001432, -0.00002443, 0.00002709, -0.00000203, -0.00000672, 0.00000225, -0.00001162, -0.00000031, -0.00001000, 0.00001194, -0.00000668, -0.00001145, 0.00000722, -0.00000371, -0.00000719, -0.00000409, 0.00000551, -0.00000016, -0.00000720, -0.00001117, 0.00000834, -0.00000517, 0.00000315, -0.00000124, -0.00000085, -0.00000145, -0.00000063, -0.00000386, 0.00000157, -0.00000965, 0.00000150, -0.00000470, 0.00000116, -0.00000098, -0.00000315, -0.00001468, 0.00001528, -0.00002829, 0.00002865, -0.00001655, 0.00000987, -0.00001402, 0.00000384, -0.00000637, 0.00001085, -0.00000136, -0.00000201, -0.00000919, 0.00000589, -0.00001814, 0.00000995, -0.00000470, 0.00000322, -0.00001877, 0.00000396, -0.00001033, 0.00000610, -0.00003188, 0.00000432, -0.00002362, 0.00000839, -0.00005140, 0.00000274, -0.00007265, -0.00000753, -0.00028298, -0.00011183, 0.00046909, 0.00009799, 0.00019937, 0.00020007, 0.00040004, -0.00039751, -0.00004995, -0.00008117, -0.00001906, -0.00004148, 0.00000834, -0.00003214, -0.00000406, -0.00001747, 0.00000924, -0.00001494, 0.00000148, -0.00001680, 0.00000838, -0.00001499, -0.00000065, -0.00000427, 0.00000428, -0.00000867, -0.00000393, -0.00000235, 0.00000509, -0.00000627, -0.00000047, -0.00000467, -0.00000311, 0.00000496, 0.00000112, -0.00000555, 0.00000183, -0.00000879, -0.00000298, -0.00001371, 0.00001728, -0.00001653, 0.00002151, -0.00001210, 0.00000919, -0.00000044, 0.00000705, -0.00000044, 0.00000670, -0.00000800, 0.00000411, -0.00000209, 0.00000499, 0.00000203, -0.00000119, },
		{ -0.00000062, 0.00000187, 0.00000745, 0.00000367, -0.00001124, 0.00001442, -0.00000227, 0.00001220, -0.00000692, -0.00000686, 0.00003162, -0.00002662, 0.00001466, -0.00001453, 0.00000053, -0.00001051, 0.00000484, 0.00000369, -0.00001457, 0.00000301, 0.00000196, -0.00000909, -0.00000497, 0.00000322, 0.00000325, -0.00000584, -0.00001090, 0.00000168, 0.00000246, -0.00000368, 0.00000499, -0.00000541, 0.00000258, -0.00000424, -0.00000032, -0.00000235, -0.00000302, -0.00000596, 0.00000156, -0.00000501, 0.00000383, -0.00000373, -0.00001390, 0.00001014, -0.00001669, 0.00000800, 0.00000702, -0.00001154, 0.00000902, -0.00001864, 0.00001200, -0.00001001, 0.00002063, -0.00001923, 0.00000756, -0.00001263, 0.00000593, -0.00001859, 0.00002015, -0.00001854, 0.00000949, -0.00002751, 0.00001835, -0.00002310, 0.00000920, -0.00004038, 0.00001772, -0.00003425, 0.00001050, -0.00006579, 0.00001447, -0.00009974, -0.00003618, -0.00042702, 0.00034724, 0.00028723, 0.00015713, 0.00018598, 0.00046735, -0.00022447, -0.00023821, 0.00001444, -0.00009599, 0.00000877, -0.00004481, 0.00001642, -0.00004478, 0.00001355, -0.00002447, 0.00001658, -0.00002298, 0.00000496, -0.00001636, 0.00000994, -0.00002295, 0.00001298, -0.00001322, 0.00001023, -0.00001817, 0.00000740, -0.00000654, 0.00000657, -0.00001129, 0.00000520, -0.00001271, 0.00001010, -0.00000103, -0.00000208, -0.00000193, -0.00000418, -0.00000390, -0.00001436, 0.00001283, -0.00000851, 0.00000910, 0.00000239, -0.00000559, 0.00001036, -0.00000200, 0.00000694, 0.00000065, 0.00000046, -0.00000553, 0.00000601, -0.00000440, 0.00001032, -0.00000629, 0.00000750, },
		{ -0.00000118, 0.00000521, 0.00000634, -0.00000733, 0.00000212, 0.00001270, -0.00000695, 0.00001767, -0.00002690, 0.00003154, -0.00000716, -0.00000900, 0.00001062, -0.00001809, 0.00000492, -0.00001270, 0.00001599, -0.00001687, 0.00000169, 0.00000077, -0.00000308, -0.00001151, 0.00000480, 0.00000219, -0.00000127, -0.00001166, -0.00000215, 0.00000414, -0.00000350, 0.00000404, -0.00000318, 0.00000061, -0.00000235, -0.00000167, -0.00000205, -0.00000143, -0.00000678, 0.00000103, -0.00000464, 0.00000202, 0.00000037, -0.00001414, 0.00000372, -0.00000422, -0.00000971, 0.00002282, -0.00002284, 0.00001947, -0.00002512, 0.00001469, -0.00001616, 0.00002562, -0.00001875, 0.00000847, -0.00001657, 0.00001333, -0.00002519, 0.00002139, -0.00001983, 0.00001656, -0.00003265, 0.00001934, -0.00002622, 0.00001868, -0.00004374, 0.00001583, -0.00003720, 0.00002535, -0.00007274, 0.00002393, -0.00010099, 0.00001565, -0.00032020, -0.00012349, 0.00049894, 0.00017531, 0.00016593, 0.00047056, 0.00002361, -0.00040724, 0.00005327, -0.00012665, 0.00002498, -0.00007108, 0.00003537, -0.00005707, 0.00002178, -0.00003711, 0.00002439, -0.00002630, 0.00001122, -0.00002676, 0.00001857, -0.00002844, 0.00001527, -0.00001694, 0.00001340, -0.00001694, 0.00000578, -0.00000862, 0.00000877, -0.00001157, 0.00000561, -0.00001258, 0.00000567, 0.00000369, -0.00000310, -0.00000179, -0.00000397, -0.00000170, -0.00001425, 0.00000565, 0.00000141, -0.00000315, 0.00001408, -0.00001434, 0.00001364, -0.00000405, 0.00000817, -0.00000090, 0.00000489, -0.00000955, 0.00000753, -0.00000591, 0.00000874, -0.00000175, 0.00000205, 0.00000425, },
		{ 0.00000132, 0.00000658, -0.00000194, -0.00000536, 0.00001487, -0.00000388, 0.00000719, -0.00000157, -0.00001022, 0.00003606, -0.00003843, 0.00002846, -0.00002362, 0.00000786, -0.00001594, 0.00001289, -0.00000926, -0.00000275, 0.00000009, 0.00000126, -0.00001364, 0.00000189, 0.00000323, 0.00000137, -0.00001073, -0.00000436, 0.00000210, -0.00000006, 0.00000023, 0.00000161, -0.00000366, 0.00000208, -0.00000565, 0.00000133, -0.00000443, -0.00000205, -0.00000452, 0.00000066, -0.00000465, 0.00000729, -0.00001519, -0.00000078, 0.00000243, -0.00001610, 0.00002278, -0.00001635, 0.00001018, -0.00001098, -0.00000072, -0.00000450, 0.00001004, 0.00000183, -0.00000888, -0.00000144, -0.00000200, -0.00000578, -0.00000201, 0.00000177, -0.00000345, -0.00000595, -0.00001116, 0.00000274, -0.00001094, -0.00000358, -0.00002607, 0.00000010, -0.00001348, -0.00001289, -0.00004344, -0.00001220, -0.00008058, -0.00006119, -0.00043436, 0.00030834, 0.00043241, 0.00007236, 0.00047379, 0.00024831, -0.00045122, -0.00003068, -0.00007614, -0.00001982, -0.00004315, 0.00000078, -0.00002336, -0.00000848, -0.00001294, -0.00000312, -0.00000153, -0.00000565, -0.00001272, 0.00000136, -0.00000979, -0.00000383, -0.00000046, -0.00000348, 0.00000080, -0.00000817, 0.00000166, -0.00000231, 0.00000008, -0.00000469, -0.00000166, -0.00000793, 0.00001392, -0.00000799, 0.00000259, -0.00000792, 0.00000233, -0.00001320, -0.00000158, 0.00000844, -0.00000978, 0.00001755, -0.00001311, 0.00000801, 0.00000034, 0.00000474, 0.00000077, 0.00000550, -0.00000812, 0.00000367, -0.00000230, 0.00000294, 0.00000383, -0.00000127, 0.00000501, 0.00000053, },
		{ 0.00000808, -0.00000308, -0.00000012, 0.00000098, 0.00001559, -0.00001652, 0.00002529, -0.00003197, 0.00003872, -0.00002033, 0.00000234, 0.00000495, -0.00001336, 0.00000099, -0.00000550, 0.00000626, -0.00001202, 0.00000592, -0.00000360, -0.00000509, -0.00000940, 0.00000954, -0.00000146, -0.00000440, -0.00001015, 0.00000404, -0.00000218, 0.00000230, 0.00000030, -0.00000167, 0.00000028, -0.00000316, -0.00000158, -0.00000169, -0.00000362, -0.00000265, -0.00000157, -0.00000320, 0.00000410, -0.00000589, -0.00001237, 0.00001179, -0.00002140, 0.00002002, -0.00000799, 0.00000034, 0.00000070, -0.00000959, 0.00000086, -0.00000046, 0.00001404, -0.00001506, 0.00000317, -0.00000813, 0.00000308, -0.00001085, 0.00000788, -0.00001067, 0.00000664, -0.00002216, 0.00001026, -0.00001971, 0.00000955, -0.00003176, 0.00000062, -0.00002004, 0.00000701, -0.00005544, 0.00000319, -0.00008557, -0.00000975, -0.00030056, -0.00021627, 0.00063597, 0.00014185, 0.00035000, 0.00050188, -0.00041373, -0.00015880, -0.00002949, -0.00005610, -0.00001956, -0.00003088, 0.00000141, -0.00002810, 0.00000278, -0.00002086, 0.00000922, -0.00001063, -0.00000498, -0.00001081, 0.00000201, -0.00001488, 0.00000805, -0.00001189, 0.00000744, -0.00001081, 0.00000366, -0.00000617, 0.00000351, -0.00000703, 0.00000200, -0.00001241, 0.00001270, -0.00000245, -0.00000146, -0.00000372, -0.00000287, -0.00000342, -0.00001352, 0.00001628, -0.00001480, 0.00001777, -0.00000880, 0.00000208, 0.00000259, 0.00000402, -0.00000002, 0.00000683, -0.00000633, 0.00000067, -0.00000074, 0.00000092, 0.00000320, 0.00000209, 0.00000019, 0.00000595, -0.00000345, },
		{ -0.00000210, 0.00000175, -0.00000394, 0.00001757, -0.00001195, 0.00001103, -0.00000354, -0.00000464, 0.00002608, -0.00003302, 0.00002823, -0.00002320, 0.00000751, -0.00001269, 0.00001007, -0.00001361, 0.00000834, -0.00000684, 0.00000146, -0.00001581, 0.00000985, -0.00000092, -0.00000013, -0.00001368, 0.00000519, -0.00000456, 0.00000436, -0.00000141, 0.00000139, -0.00000314, 0.00000159, -0.00000682, 0.00000346, -0.00000848, 0.00000294, -0.00000715, 0.00000233, -0.00000433, 0.00000693, -0.00002345, 0.00001772, -0.00002223, 0.00001404, 0.00000033, -0.00000586, 0.00000455, -0.00000825, -0.00000218, -0.00000217, 0.00001293, -0.00000642, -0.00000554, -0.00000099, -0.00000382, -0.00000178, -0.00000261, -0.00000203, -0.00000136, -0.00000781, -0.00000756, -0.00000214, -0.00001008, -0.00000198, -0.00002916, 0.00000314, -0.00001616, -0.00001177, -0.00004658, -0.00001301, -0.00009126, -0.00004904, -0.00049481, 0.00032563, 0.00053304, 0.00009517, 0.00074830, -0.00029059, -0.00028851, -0.00002527, -0.00006221, -0.00001939, -0.00003762, -0.00000284, -0.00002198, -0.00000394, -0.00001503, -0.00000316, 0.00000066, -0.00000895, -0.00000949, -0.00000177, -0.00001046, 0.00000166, -0.00000517, -0.00000126, -0.00000112, -0.00000354, -0.00000041, -0.00000328, -0.00000041, -0.00000296, -0.00000643, 0.00000077, 0.00001044, -0.00001039, 0.00000508, -0.00001227, 0.00000775, -0.00002237, 0.00001874, -0.00001387, 0.00001227, -0.00000088, -0.00000415, 0.00000408, 0.00000306, 0.00000063, 0.00000519, -0.00000181, -0.00000323, 0.00000107, -0.00000092, 0.00000257, 0.00000315, 0.00000004, 0.00000511, -0.00000127, 0.00000488, },
		{ -0.00000602, 0.00000557, 0.00000160, 0.00001135, -0.00001771, 0.00002639, -0.00002841, 0.00003335, -0.00002197, 0.00000823, 0.00000013, -0.00000886, -0.00000137, -0.00000050, -0.00000560, 0.00000199, 0.00000006, -0.00000347, -0.00000909, -0.00000143, 0.00000747, -0.00000311, -0.00000895, -0.00000171, 0.00000154, -0.00000253, 0.00000495, -0.00000298, 0.00000071, -0.00000112, -0.00000372, -0.00000027, -0.00000424, -0.00000148, -0.00000192, -0.00000250, -0.00000194, 0.00000564, -0.00001674, 0.00000488, -0.00000523, -0.00000721, 0.00002009, -0.00002035, 0.00001493, -0.00001484, 0.00000592, -0.00001380, 0.00001892, -0.00000718, -0.00000123, -0.00000777, 0.00000312, -0.00000841, 0.00000586, -0.00001219, 0.00000673, -0.00001135, -0.00000202, -0.00001022, -0.00000306, -0.00000705, -0.00001359, -0.00002019, 0.00000356, -0.00002052, -0.00002625, -0.00003246, -0.00005503, -0.00004996, -0.00026661, -0.00034231, 0.00082797, 0.00007495, 0.00073599, 0.00005784, -0.00048865, 0.00000541, -0.00010869, 0.00001455, -0.00006793, 0.00001189, -0.00003741, 0.00000961, -0.00002607, 0.00000276, -0.00001056, 0.00000705, -0.00002200, 0.00000609, -0.00001770, 0.00000635, -0.00000890, 0.00000120, -0.00000495, 0.00000244, -0.00000511, -0.00000003, -0.00000462, 0.00000169, -0.00000814, -0.00000217, 0.00001084, -0.00000603, 0.00000142, -0.00000862, 0.00000358, -0.00001406, 0.00000504, 0.00000139, -0.00000527, 0.00001537, -0.00001570, 0.00001101, -0.00000451, 0.00000824, -0.00000337, 0.00000798, -0.00001074, 0.00000676, -0.00000706, 0.00000722, -0.00000282, 0.00000709, -0.00000214, 0.00000619, -0.00000207, 0.00000403, },
		{ 0.00000971, -0.00000426, 0.00001682, -0.00001729, 0.00001457, -0.00000331, -0.00000285, 0.00001566, -0.00002032, 0.00001816, -0.00001443, 0.00000067, -0.00000262, -0.00000413, -0.00000247, 0.00000739, -0.00000884, -0.00000203, -0.00001057, 0.00001162, -0.00000393, -0.00000389, -0.00000944, 0.00000875, -0.00000986, 0.00001042, -0.00000654, 0.00000467, -0.00000515, 0.00000207, -0.00000734, 0.00000382, -0.00001046, 0.00000770, -0.00001107, 0.00000548, -0.00000380, -0.00000044, -0.00001590, 0.00001705, -0.00003018, 0.00003775, -0.00003102, 0.00002244, -0.00002173, 0.00001619, -0.00002514, 0.00002404, -0.00001161, 0.00000979, -0.00002113, 0.00001693, -0.00002373, 0.00002352, -0.00002906, 0.00002097, -0.00002586, 0.00001869, -0.00003347, 0.00002096, -0.00003585, 0.00002749, -0.00006257, 0.00004068, -0.00005891, 0.00003445, -0.00010010, 0.00004018, -0.00016173, 0.00003318, -0.00065253, 0.00045869, 0.00051180, 0.00042650, 0.00057749, -0.00068042, 0.00002047, -0.00016818, 0.00005384, -0.00010075, 0.00003381, -0.00006517, 0.00003337, -0.00004650, 0.00002166, -0.00003575, 0.00003097, -0.00003834, 0.00001937, -0.00003164, 0.00001742, -0.00001953, 0.00001211, -0.00001798, 0.00001511, -0.00001469, 0.00000956, -0.00001571, 0.00001155, -0.00001460, 0.00000242, 0.00000105, 0.00000745, -0.00001003, 0.00000343, -0.00000987, 0.00000275, -0.00001440, 0.00001988, -0.00002307, 0.00002839, -0.00002268, 0.00001484, -0.00000996, 0.00001342, -0.00000891, 0.00001337, -0.00001350, 0.00000879, -0.00000979, 0.00000872, -0.00000556, 0.00000924, -0.00000322, 0.00000705, -0.00000280, 0.00000626, -0.00001082, },
		{ 0.00000604, 0.00000367, 0.00000276, -0.00001186, 0.00002213, -0.00001911, 0.00001654, -0.00000700, -0.00000154, 0.00000731, -0.00001431, 0.00000702, -0.00000966, -0.00000298, 0.00000837, -0.00000493, -0.00000505, -0.00000801, 0.00000431, 0.00000340, -0.00000556, -0.00000910, 0.00000613, -0.00000538, 0.00000320, 0.00000142, -0.00000163, 0.00000028, -0.00000138, -0.00000440, 0.00000107, -0.00000771, 0.00000392, -0.00000539, -0.00000015, -0.00000133, 0.00000211, -0.00001881, 0.00001751, -0.00002717, 0.00002612, -0.00001241, 0.00000305, -0.00000395, 0.00000080, -0.00000772, 0.00000147, 0.00000803, -0.00000392, -0.00000674, 0.00000129, -0.00000803, 0.00000638, -0.00000708, -0.00000339, -0.00000388, -0.00000005, -0.00001132, -0.00000301, -0.00001408, 0.00000708, -0.00003010, -0.00000132, -0.00001782, 0.00000092, -0.00005008, -0.00001202, -0.00008694, -0.00001656, -0.00032713, -0.00036759, 0.00090070, 0.00027912, 0.00088837, -0.00057049, -0.00017307, -0.00009111, -0.00002784, -0.00003580, -0.00002122, -0.00002706, -0.00000524, -0.00001190, -0.00000649, -0.00001539, 0.00000735, -0.00001125, -0.00000381, -0.00001200, -0.00000344, -0.00000143, -0.00000287, -0.00000567, 0.00000151, -0.00000021, -0.00000118, -0.00000691, 0.00000102, -0.00000280, -0.00000733, 0.00000407, 0.00000638, -0.00000659, 0.00000177, -0.00000979, 0.00000466, -0.00001691, 0.00002003, -0.00002052, 0.00001971, -0.00000965, 0.00000186, -0.00000010, 0.00000375, -0.00000023, 0.00000480, -0.00000358, -0.00000037, -0.00000141, -0.00000060, 0.00000252, 0.00000012, 0.00000579, -0.00000045, 0.00000331, 0.00000238, -0.00000655, 0.00000043, },
		{ -0.00000114, 0.00000799, -0.00001245, 0.00000980, 0.00000677, -0.00001626, 0.00002227, -0.00002020, 0.00001596, -0.00001211, 0.00000137, -0.00000184, -0.00001193, 0.00001032, 0.00000056, -0.00000844, -0.00000494, -0.00000234, 0.00000798, -0.00000577, -0.00000690, -0.00000064, 0.00000408, -0.00000774, 0.00001092, -0.00000876, 0.00000697, -0.00000721, 0.00000261, -0.00000696, 0.00000158, -0.00000763, 0.00000765, -0.00001179, 0.00000773, -0.00000531, -0.00000769, 0.00000107, -0.00000486, -0.00000508, 0.00002240, -0.00002860, 0.00002468, -0.00002695, 0.00002295, -0.00003116, 0.00003561, -0.00002847, 0.00002147, -0.00002925, 0.00002428, -0.00003010, 0.00003437, -0.00004400, 0.00003406, -0.00004000, 0.00003588, -0.00005201, 0.00003589, -0.00005006, 0.00004235, -0.00007842, 0.00005870, -0.00008206, 0.00006110, -0.00013472, 0.00007207, -0.00020577, 0.00008123, -0.00078305, 0.00054777, 0.00061283, 0.00083368, -0.00009609, -0.00055550, 0.00009055, -0.00019740, 0.00008979, -0.00012297, 0.00005552, -0.00008684, 0.00005793, -0.00006553, 0.00003749, -0.00004831, 0.00004087, -0.00004732, 0.00002845, -0.00004410, 0.00003352, -0.00003360, 0.00002316, -0.00002833, 0.00002694, -0.00002272, 0.00001451, -0.00002227, 0.00001896, -0.00002470, 0.00001569, -0.00000573, 0.00000781, -0.00001022, 0.00000154, -0.00000615, -0.00000546, 0.00000527, -0.00000191, -0.00000402, 0.00001453, -0.00001864, 0.00001598, -0.00001134, 0.00001288, -0.00000764, 0.00000913, -0.00001174, 0.00000935, -0.00001148, 0.00001127, -0.00000849, 0.00001196, -0.00000382, 0.00000546, 0.00000083, -0.00000136, -0.00000947, 0.00001260, },
		{ -0.00000059, 0.00000114, -0.00001070, 0.00002461, -0.00002030, 0.00000844, 0.00000589, -0.00001320, 0.00001632, -0.00002016, 0.00001551, -0.00002351, 0.00001101, 0.00000330, -0.00000449, -0.00000967, -0.00000031, 0.00000327, 0.00000101, -0.00000951, -0.00000221, 0.00000594, -0.00000770, 0.00000743, -0.00000332, 0.00000204, -0.00000237, -0.00000044, -0.00000465, 0.00000026, -0.00000767, 0.00000696, -0.00000835, 0.00000290, -0.00000137, -0.00000646, -0.00000514, 0.00000583, -0.00002042, 0.00003464, -0.00003298, 0.00002495, -0.00002648, 0.00002312, -0.00002907, 0.00002775, -0.00001983, 0.00001625, -0.00002440, 0.00001997, -0.00002784, 0.00003164, -0.00003499, 0.00002217, -0.00003137, 0.00003002, -0.00004128, 0.00002373, -0.00004229, 0.00003791, -0.00006399, 0.00003737, -0.00006332, 0.00005073, -0.00010700, 0.00004396, -0.00015681, 0.00005967, -0.00044829, -0.00033441, 0.00090639, 0.00084262, 0.00038074, -0.00078362, 0.00010124, -0.00023049, 0.00008805, -0.00011876, 0.00005223, -0.00009192, 0.00005399, -0.00005912, 0.00003429, -0.00005053, 0.00003871, -0.00004133, 0.00002596, -0.00004337, 0.00002777, -0.00002714, 0.00001846, -0.00002561, 0.00002109, -0.00001472, 0.00001110, -0.00002113, 0.00001521, -0.00001797, 0.00000708, -0.00000093, 0.00000593, -0.00000710, 0.00000014, -0.00000608, -0.00000335, -0.00000155, 0.00000920, -0.00001758, 0.00002449, -0.00002311, 0.00001660, -0.00001159, 0.00001186, -0.00000715, 0.00000957, -0.00001164, 0.00000918, -0.00001075, 0.00000821, -0.00000519, 0.00000622, 0.00000257, 0.00000076, 0.00000340, 0.00000074, -0.00001197, 0.00000672, 0.00000745, },
		{ -0.00000110, -0.00000479, 0.00000440, 0.00001582, -0.00003325, 0.00003921, -0.00003197, 0.00002357, -0.00001784, 0.00001066, -0.00001409, -0.00000341, 0.00001118, -0.00000177, -0.00001144, -0.00000121, 0.00000119, 0.00000402, -0.00000885, -0.00000403, 0.00000434, -0.00000184, -0.00000148, 0.00000570, -0.00000562, 0.00000515, -0.00000676, 0.00000216, -0.00000653, -0.00000065, -0.00000263, 0.00000376, -0.00000869, 0.00000814, -0.00001154, -0.00000260, 0.00000414, -0.00001742, 0.00002258, -0.00001140, 0.00000068, -0.00000193, -0.00000268, 0.00000034, -0.00000530, 0.00001213, -0.00001383, 0.00000713, -0.00001223, 0.00000518, -0.00000628, 0.00001033, -0.00002311, 0.00001126, -0.00001614, 0.00001365, -0.00003232, 0.00001322, -0.00002423, 0.00001319, -0.00004460, 0.00002021, -0.00004240, 0.00001821, -0.00009436, 0.00002109, -0.00015343, 0.00000773, -0.00077521, 0.00042565, 0.00120537, 0.00068882, -0.00070004, -0.00014408, -0.00008461, -0.00006021, -0.00000516, -0.00004006, -0.00001428, -0.00002677, 0.00001163, -0.00002506, 0.00000210, -0.00001551, 0.00000854, -0.00001562, -0.00000296, -0.00001577, 0.00001218, -0.00001601, 0.00000633, -0.00001234, 0.00001562, -0.00001183, 0.00000130, -0.00000990, 0.00000702, -0.00001607, 0.00001647, -0.00000843, 0.00000717, -0.00001075, 0.00000286, -0.00000945, -0.00000038, 0.00001041, -0.00001687, 0.00001717, -0.00001094, 0.00000288, 0.00000063, 0.00000035, 0.00000196, 0.00000218, -0.00000422, 0.00000163, -0.00000238, -0.00000129, 0.00000295, -0.00000211, 0.00000799, -0.00000069, 0.00000220, 0.00000462, -0.00001185, -0.00000113, 0.00001151, 0.00000096, },
		{ 0.00000884, -0.00001664, 0.00003450, -0.00003488, 0.00001845, 0.00000253, -0.00001256, 0.00001492, -0.00001481, 0.00000987, -0.00002392, 0.00001918, -0.00000268, -0.00000423, -0.00001116, 0.00000752, -0.00000174, -0.00000060, -0.00001112, 0.00000681, -0.00000140, -0.00000173, 0.00000371, -0.00000195, 0.00000150, -0.00000227, -0.00000151, -0.00000266, -0.00000404, -0.00000144, 0.00000275, -0.00000506, 0.00000222, -0.00000252, -0.00001180, 0.00001113, -0.00001929, 0.00001428, 0.00000258, -0.00001122, 0.00000798, -0.00001282, 0.00001126, -0.00001555, 0.00001936, -0.00001917, 0.00001259, -0.00001685, 0.00001047, -0.00001535, 0.00002136, -0.00002822, 0.00001360, -0.00002306, 0.00002382, -0.00003622, 0.00001457, -0.00003027, 0.00002331, -0.00004646, 0.00001858, -0.00004634, 0.00003273, -0.00009126, 0.00001852, -0.00013014, 0.00001530, -0.00042620, -0.00052612, 0.00129684, 0.00123917, -0.00055632, -0.00039055, 0.00000303, -0.00014596, 0.00003698, -0.00007138, 0.00001520, -0.00006198, 0.00003390, -0.00004029, 0.00001728, -0.00003320, 0.00002045, -0.00002382, 0.00000978, -0.00003126, 0.00002165, -0.00002224, 0.00001260, -0.00001998, 0.00001858, -0.00001049, 0.00000427, -0.00001570, 0.00001085, -0.00001755, 0.00001342, -0.00000426, 0.00000364, -0.00000505, -0.00000283, -0.00000239, -0.00000979, 0.00001745, -0.00001715, 0.00000994, -0.00000131, -0.00000580, 0.00000638, -0.00000367, 0.00000351, 0.00000116, -0.00000196, -0.00000172, 0.00000175, -0.00000498, 0.00000410, -0.00000235, 0.00000454, 0.00000499, -0.00000311, 0.00000903, -0.00001024, -0.00000556, 0.00000585, 0.00001247, -0.00001175, },
		{ -0.00000530, 0.00000893, 0.00000934, -0.00003349, 0.00004465, -0.00003701, 0.00002578, -0.00001720, 0.00001245, -0.00002254, 0.00000954, 0.00000403, -0.00000006, -0.00001570, 0.00000738, -0.00000189, 0.00000208, -0.00001251, 0.00000502, 0.00000003, -0.00000003, -0.00000067, 0.00000320, -0.00000341, 0.00000356, -0.00000687, 0.00000351, -0.00000981, 0.00000317, -0.00000362, 0.00000417, -0.00000793, 0.00000894, -0.00002125, 0.00001622, -0.00001822, 0.00000606, 0.00001012, -0.00001222, 0.00000567, -0.00000939, 0.00000578, -0.00000740, 0.00000850, -0.00000632, -0.00000168, -0.00000123, -0.00000439, -0.00000203, 0.00000523, -0.00000457, -0.00000990, -0.00000327, 0.00000048, -0.00000311, -0.00001937, 0.00000120, -0.00001229, 0.00000060, -0.00002893, -0.00000021, -0.00002071, -0.00000725, -0.00007435, -0.00000308, -0.00013662, -0.00002730, -0.00083746, 0.00035909, 0.00216000, -0.00031861, -0.00059201, -0.00001903, -0.00013699, -0.00000687, -0.00003991, -0.00000707, -0.00004406, 0.00000349, -0.00001191, -0.00000510, -0.00001342, -0.00000291, -0.00000392, -0.00000266, -0.00001845, 0.00000235, -0.00000328, -0.00000381, -0.00000483, -0.00000023, 0.00000736, -0.00000635, -0.00000568, -0.00000225, -0.00000340, -0.00000318, 0.00001124, -0.00000968, 0.00000822, -0.00001393, 0.00000846, -0.00001981, 0.00002093, -0.00001237, 0.00000075, 0.00000614, -0.00001006, 0.00000703, -0.00000281, 0.00000164, 0.00000157, -0.00000012, -0.00000465, 0.00000430, -0.00000580, 0.00000272, -0.00000102, 0.00000143, 0.00000663, -0.00000157, 0.00000507, -0.00000157, -0.00001189, 0.00000213, 0.00001427, -0.00000569, 0.00000022, },
		{ -0.00001771, 0.00003567, -0.00004329, 0.00002944, -0.00000548, -0.00000767, 0.00001079, -0.00000676, -0.00000313, -0.00001104, 0.00001521, -0.00000317, -0.00000710, -0.00000634, 0.00000977, -0.00000657, -0.00000255, -0.00000712, 0.00001022, -0.00000732, 0.00000615, -0.00000456, 0.00000489, -0.00000419, 0.00000188, -0.00000510, 0.00000031, -0.00000821, 0.00000619, -0.00000497, 0.00000278, -0.00000251, -0.00000551, -0.00000529, 0.00000855, -0.00002349, 0.00003425, -0.00002913, 0.00002104, -0.00002446, 0.00001932, -0.00002052, 0.00002174, -0.00001877, 0.00001085, -0.00001579, 0.00001360, -0.00002195, 0.00002293, -0.00001988, 0.00001172, -0.00002895, 0.00002204, -0.00002191, 0.00000806, -0.00003266, 0.00002123, -0.00003390, 0.00001881, -0.00005512, 0.00003322, -0.00005669, 0.00000343, -0.00009505, 0.00000034, -0.00016421, -0.00023183, -0.00103758, 0.00244202, 0.00069357, -0.00107467, 0.00008850, -0.00024750, 0.00005448, -0.00011166, 0.00005649, -0.00009396, 0.00003520, -0.00004696, 0.00003070, -0.00004472, 0.00002333, -0.00003357, 0.00002779, -0.00004074, 0.00001715, -0.00001992, 0.00001393, -0.00002116, 0.00001244, -0.00000819, 0.00001351, -0.00002087, 0.00000884, -0.00001414, 0.00000538, 0.00000100, 0.00000293, -0.00000476, 0.00000083, -0.00000607, -0.00000339, -0.00000220, 0.00001512, -0.00002452, 0.00002514, -0.00002429, 0.00001731, -0.00001209, 0.00001104, -0.00000962, 0.00001209, -0.00001592, 0.00001310, -0.00001220, 0.00000807, -0.00000742, 0.00000732, -0.00000206, 0.00000935, -0.00000616, 0.00001093, -0.00001898, 0.00000355, 0.00000377, 0.00001306, -0.00001819, 0.00001414, },
		{ 0.00001597, -0.00000491, -0.00001936, 0.00003508, -0.00002994, 0.00001815, -0.00000643, 0.00000029, -0.00001509, 0.00001231, -0.00000074, 0.00000035, -0.00001602, 0.00001482, -0.00000895, 0.00000238, -0.00001439, 0.00001541, -0.00001116, 0.00001146, -0.00001096, 0.00001134, -0.00001064, 0.00001010, -0.00001402, 0.00001145, -0.00002022, 0.00001691, -0.00001662, 0.00001669, -0.00001747, 0.00001430, -0.00002962, 0.00003439, -0.00004682, 0.00004928, -0.00003844, 0.00003101, -0.00003486, 0.00002937, -0.00003233, 0.00003444, -0.00003050, 0.00002456, -0.00003417, 0.00003547, -0.00004431, 0.00004351, -0.00004157, 0.00004063, -0.00005821, 0.00004744, -0.00005113, 0.00004898, -0.00007710, 0.00006557, -0.00008439, 0.00008162, -0.00011810, 0.00009586, -0.00012868, 0.00011061, -0.00021678, 0.00016542, -0.00036112, 0.00024498, -0.00136893, 0.00101438, 0.00262164, -0.00159759, 0.00014764, -0.00036387, 0.00012960, -0.00020182, 0.00012543, -0.00014565, 0.00007654, -0.00009677, 0.00007696, -0.00008424, 0.00006020, -0.00007285, 0.00006155, -0.00006451, 0.00003818, -0.00004550, 0.00003928, -0.00004415, 0.00003331, -0.00003260, 0.00003820, -0.00003848, 0.00002483, -0.00003131, 0.00002083, -0.00001724, 0.00002379, -0.00002547, 0.00002210, -0.00002646, 0.00001916, -0.00002925, 0.00004084, -0.00004277, 0.00003603, -0.00003145, 0.00002184, -0.00001726, 0.00001745, -0.00001731, 0.00001916, -0.00002062, 0.00001503, -0.00001314, 0.00000977, -0.00001050, 0.00000990, -0.00000672, 0.00001369, -0.00000877, 0.00001310, -0.00001592, -0.00000020, -0.00000245, 0.00001976, -0.00001672, 0.00000830, -0.00000987, },
		{ 0.00002053, -0.00003149, 0.00002342, -0.00000125, -0.00001277, 0.00001732, -0.00001167, -0.00000498, -0.00000113, 0.00000790, -0.00000178, -0.00000933, 0.00000147, 0.00000559, -0.00000848, -0.00000481, 0.00000225, 0.00000347, -0.00000248, 0.00000398, -0.00000476, 0.00000547, -0.00000479, 0.00000127, -0.00000330, -0.00000377, -0.00000228, 0.00000253, -0.00000236, 0.00000245, -0.00000284, -0.00001237, 0.00001246, -0.00001711, 0.00000972, 0.00000538, -0.00001182, 0.00000876, -0.00001527, 0.00001176, -0.00001197, 0.00001926, -0.00002339, 0.00001143, -0.00001254, 0.00000885, -0.00001479, 0.00001847, -0.00001917, 0.00000968, -0.00002724, 0.00002199, -0.00002306, 0.00000569, -0.00002741, 0.00001374, -0.00002507, 0.00001087, -0.00004931, 0.00002656, -0.00005322, -0.00000622, -0.00008548, -0.00001935, -0.00016467, -0.00027835, -0.00133291, 0.00391633, -0.00088836, -0.00058947, 0.00001467, -0.00017699, 0.00002019, -0.00007876, 0.00003263, -0.00007448, 0.00002456, -0.00003678, 0.00002137, -0.00003379, 0.00001161, -0.00002310, 0.00001836, -0.00003453, 0.00001644, -0.00001990, 0.00001248, -0.00001959, 0.00001363, -0.00000724, 0.00000947, -0.00001787, 0.00000704, -0.00001651, 0.00001373, -0.00000298, 0.00000095, -0.00000314, -0.00000203, -0.00000166, -0.00001133, 0.00001854, -0.00001314, 0.00000308, 0.00000102, -0.00000963, 0.00001007, -0.00000662, 0.00000496, -0.00000325, 0.00000262, -0.00000821, 0.00000772, -0.00000828, 0.00000582, -0.00000629, 0.00000709, -0.00000066, 0.00000589, -0.00000040, 0.00000013, -0.00001319, 0.00000338, 0.00000783, 0.00000372, -0.00001256, 0.00000997, -0.00000870, },
		{ -0.00000483, -0.00001356, 0.00002922, -0.00002588, 0.00001510, -0.00000131, -0.00001008, -0.00000185, 0.00000700, -0.00000219, 0.00000022, -0.00001264, 0.00001790, -0.00001690, 0.00000493, -0.00001319, 0.00002064, -0.00001982, 0.00002333, -0.00002473, 0.00002492, -0.00002407, 0.00002260, -0.00002608, 0.00002266, -0.00003191, 0.00003255, -0.00003370, 0.00003538, -0.00003472, 0.00002389, -0.00003220, 0.00003619, -0.00005073, 0.00006645, -0.00007046, 0.00006783, -0.00007480, 0.00007156, -0.00007583, 0.00008574, -0.00008686, 0.00007573, -0.00008412, 0.00008756, -0.00009853, 0.00010482, -0.00010898, 0.00010933, -0.00013121, 0.00012607, -0.00013459, 0.00013381, -0.00016634, 0.00016186, -0.00019032, 0.00020036, -0.00025159, 0.00024443, -0.00029956, 0.00030107, -0.00044295, 0.00043632, -0.00072781, 0.00072277, -0.00233247, 0.00268883, 0.00174493, -0.00208014, 0.00072494, -0.00072053, 0.00042348, -0.00043096, 0.00032215, -0.00031601, 0.00023117, -0.00023114, 0.00019636, -0.00018997, 0.00015636, -0.00016267, 0.00014568, -0.00014574, 0.00011762, -0.00011802, 0.00010472, -0.00010530, 0.00009233, -0.00008475, 0.00008489, -0.00008460, 0.00007002, -0.00007734, 0.00006655, -0.00005277, 0.00005159, -0.00005234, 0.00004523, -0.00004542, 0.00003203, -0.00003056, 0.00004036, -0.00004873, 0.00004890, -0.00005362, 0.00004758, -0.00004038, 0.00003715, -0.00003481, 0.00003322, -0.00003656, 0.00003213, -0.00003004, 0.00002647, -0.00002604, 0.00002381, -0.00001720, 0.00002081, -0.00001324, 0.00001437, -0.00002341, 0.00000984, -0.00000556, 0.00001828, -0.00001931, 0.00001025, -0.00000609, 0.00000899, },
		{ -0.00001351, 0.00000878, 0.00000977, -0.00002219, 0.00002786, -0.00002474, 0.00000589, -0.00000121, 0.00000315, -0.00000033, -0.00000937, 0.00000920, -0.00000455, -0.00000445, -0.00000855, 0.00001450, -0.00001140, 0.00001432, -0.00001314, 0.00001119, -0.00001010, 0.00001016, -0.00001401, 0.00001236, -0.00002148, 0.00001979, -0.00002022, 0.00002089, -0.00001833, 0.00001220, -0.00002629, 0.00003280, -0.00004645, 0.00005565, -0.00005249, 0.00004624, -0.00004976, 0.00004403, -0.00004867, 0.00005476, -0.00004841, 0.00003769, -0.00005054, 0.00005354, -0.00006122, 0.00006312, -0.00006421, 0.00006428, -0.00007879, 0.00006704, -0.00007673, 0.00007788, -0.00010230, 0.00009071, -0.00011659, 0.00011956, -0.00014915, 0.00012709, -0.00017247, 0.00016649, -0.00026324, 0.00021819, -0.00040799, 0.00031446, -0.00104209, -0.00057323, 0.00456578, -0.00255052, 0.00050498, -0.00058784, 0.00028324, -0.00032805, 0.00021291, -0.00021517, 0.00014369, -0.00016223, 0.00012832, -0.00012582, 0.00010107, -0.00011388, 0.00009378, -0.00009249, 0.00006994, -0.00007688, 0.00006447, -0.00006654, 0.00005523, -0.00005282, 0.00005435, -0.00005178, 0.00004123, -0.00005030, 0.00003620, -0.00002673, 0.00003166, -0.00003396, 0.00002791, -0.00002923, 0.00002068, -0.00002709, 0.00003975, -0.00004505, 0.00004206, -0.00004380, 0.00003320, -0.00002610, 0.00002464, -0.00002365, 0.00002264, -0.00002548, 0.00001978, -0.00001858, 0.00001632, -0.00001606, 0.00001270, -0.00000801, 0.00001176, -0.00000526, 0.00001001, -0.00001685, 0.00000298, -0.00000445, 0.00001348, -0.00000714, -0.00000333, 0.00000539, -0.00000327, 0.00000600, },
		{ -0.00002468, 0.00003643, -0.00003243, 0.00002320, -0.00001075, -0.00000512, 0.00000263, 0.00000265, -0.00000364, 0.00000209, -0.00000777, 0.00001452, -0.00002046, 0.00000658, -0.00000715, 0.00001534, -0.00001541, 0.00002100, -0.00002497, 0.00002622, -0.00002633, 0.00002465, -0.00002774, 0.00002246, -0.00002889, 0.00003114, -0.00003358, 0.00003849, -0.00004129, 0.00002705, -0.00002566, 0.00002008, -0.00002235, 0.00003526, -0.00004713, 0.00004871, -0.00005835, 0.00005685, -0.00005934, 0.00007414, -0.00008405, 0.00007198, -0.00007788, 0.00007991, -0.00008742, 0.00009509, -0.00010335, 0.00010293, -0.00012470, 0.00012175, -0.00013377, 0.00013127, -0.00016086, 0.00015418, -0.00018124, 0.00019152, -0.00024467, 0.00023853, -0.00030139, 0.00030238, -0.00044960, 0.00044384, -0.00077921, 0.00079882, -0.00292679, 0.00464001, -0.00028845, -0.00132516, 0.00051874, -0.00057474, 0.00034018, -0.00036236, 0.00026902, -0.00027170, 0.00020109, -0.00020609, 0.00017495, -0.00016910, 0.00013592, -0.00014347, 0.00012686, -0.00012968, 0.00010843, -0.00011169, 0.00009820, -0.00009912, 0.00009017, -0.00008190, 0.00007862, -0.00007869, 0.00006500, -0.00007643, 0.00007417, -0.00005975, 0.00005401, -0.00005569, 0.00004871, -0.00004910, 0.00003504, -0.00002144, 0.00001914, -0.00002323, 0.00002231, -0.00003343, 0.00003516, -0.00003130, 0.00002865, -0.00002665, 0.00002169, -0.00002549, 0.00002284, -0.00002229, 0.00002091, -0.00002222, 0.00002186, -0.00001517, 0.00001734, -0.00000784, 0.00000254, -0.00001480, 0.00000833, -0.00000195, 0.00000788, -0.00001090, 0.00000442, 0.00000402, -0.00000977, 0.00001284, },
		{ 0.00000503, 0.00001111, -0.00002028, 0.00002476, -0.00002629, 0.00001303, -0.00000275, -0.00000246, 0.00000473, -0.00000996, 0.00001534, -0.00001771, 0.00000539, -0.00001335, 0.00002413, -0.00002442, 0.00003113, -0.00003320, 0.00003260, -0.00003275, 0.00003296, -0.00003699, 0.00003459, -0.00004332, 0.00004571, -0.00004872, 0.00005269, -0.00005147, 0.00003978, -0.00004531, 0.00004710, -0.00005744, 0.00007471, -0.00008630, 0.00008716, -0.00009577, 0.00009433, -0.00010131, 0.00011654, -0.00011945, 0.00010769, -0.00012074, 0.00012627, -0.00013612, 0.00014581, -0.00015555, 0.00015936, -0.00018009, 0.00017730, -0.00019620, 0.00020189, -0.00023266, 0.00023106, -0.00026992, 0.00028955, -0.00033991, 0.00034195, -0.00042112, 0.00044748, -0.00059472, 0.00061751, -0.00093894, 0.00101989, -0.00231699, 0.00139867, 0.00357349, -0.00310803, 0.00121749, -0.00106721, 0.00068124, -0.00064501, 0.00048448, -0.00045154, 0.00035907, -0.00035250, 0.00029872, -0.00027831, 0.00023961, -0.00024161, 0.00021222, -0.00020568, 0.00018082, -0.00018054, 0.00015979, -0.00015600, 0.00014191, -0.00013185, 0.00012567, -0.00012072, 0.00010820, -0.00011778, 0.00010544, -0.00008761, 0.00008366, -0.00008364, 0.00007347, -0.00007026, 0.00005527, -0.00004667, 0.00005022, -0.00005656, 0.00005703, -0.00006708, 0.00006214, -0.00005443, 0.00005024, -0.00004690, 0.00004148, -0.00004474, 0.00003994, -0.00003884, 0.00003682, -0.00003578, 0.00003132, -0.00002355, 0.00002360, -0.00001366, 0.00001306, -0.00002459, 0.00001479, -0.00000987, 0.00001198, -0.00000863, -0.00000105, 0.00000912, -0.00001296, 0.00001019, -0.00001078, },
		{ 0.00004268, -0.00003556, 0.00002674, -0.00001921, 0.00000521, 0.00000115, -0.00000168, -0.00000162, 0.00000243, -0.00000031, 0.00000219, -0.00001330, 0.00000194, 0.00000479, -0.00000007, 0.00000315, 0.00000051, -0.00000439, 0.00000536, -0.00000593, 0.00000423, -0.00000761, 0.00000133, -0.00000300, 0.00000343, -0.00000419, 0.00001027, -0.00001937, 0.00001036, -0.00000744, -0.00000288, 0.00001519, -0.00001763, 0.00001023, -0.00001187, 0.00000530, -0.00000918, 0.00001500, -0.00000528, -0.00000880, -0.00000349, 0.00000085, -0.00000251, 0.00000365, -0.00000306, -0.00000277, -0.00000386, -0.00000982, 0.00000017, -0.00000939, -0.00000008, -0.00001830, -0.00000177, -0.00000914, 0.00000035, -0.00003230, -0.00000102, -0.00003766, -0.00000197, -0.00009255, -0.00000504, -0.00023338, -0.00001065, -0.00223463, 0.00551820, -0.00223463, -0.00001065, -0.00023338, -0.00000504, -0.00009255, -0.00000197, -0.00003766, -0.00000102, -0.00003230, 0.00000035, -0.00000914, -0.00000177, -0.00001830, -0.00000008, -0.00000939, 0.00000017, -0.00000982, -0.00000386, -0.00000277, -0.00000306, 0.00000365, -0.00000251, 0.00000085, -0.00000349, -0.00000880, -0.00000528, 0.00001500, -0.00000918, 0.00000530, -0.00001187, 0.00001023, -0.00001763, 0.00001519, -0.00000288, -0.00000744, 0.00001036, -0.00001937, 0.00001027, -0.00000419, 0.00000343, -0.00000300, 0.00000133, -0.00000761, 0.00000423, -0.00000593, 0.00000536, -0.00000439, 0.00000051, 0.00000315, -0.00000007, 0.00000479, 0.00000194, -0.00001330, 0.00000219, -0.00000031, 0.00000243, -0.00000162, -0.00000168, 0.00000115, 0.00000521, -0.00001921, 0.00002674, -0.00003556, },
		{ 0.00000503, -0.00001078, 0.00001019, -0.00001296, 0.00000912, -0.00000105, -0.00000863, 0.00001198, -0.00000987, 0.00001479, -0.00002459, 0.00001306, -0.00001366, 0.00002360, -0.00002355, 0.00003132, -0.00003578, 0.00003682, -0.00003884, 0.00003994, -0.00004474, 0.00004148, -0.00004690, 0.00005024, -0.00005443, 0.00006214, -0.00006708, 0.00005703, -0.00005656, 0.00005022, -0.00004667, 0.00005527, -0.00007026, 0.00007347, -0.00008364, 0.00008366, -0.00008761, 0.00010544, -0.00011778, 0.00010820, -0.00012072, 0.00012567, -0.00013185, 0.00014191, -0.00015600, 0.00015979, -0.00018054, 0.00018082, -0.00020568, 0.00021222, -0.00024161, 0.00023961, -0.00027831, 0.00029872, -0.00035250, 0.00035907, -0.00045154, 0.00048448, -0.00064501, 0.00068124, -0.00106721, 0.00121749, -0.00310803, 0.00357349, 0.00139867, -0.00231699, 0.00101989, -0.00093894, 0.00061751, -0.00059472, 0.00044748, -0.00042112, 0.00034195, -0.00033991, 0.00028955, -0.00026992, 0.00023106, -0.00023266, 0.00020189, -0.00019620, 0.00017730, -0.00018009, 0.00015936, -0.00015555, 0.00014581, -0.00013612, 0.00012627, -0.00012074, 0.00010769, -0.00011945, 0.00011654, -0.00010131, 0.00009433, -0.00009577, 0.00008716, -0.00008630, 0.00007471, -0.00005744, 0.00004710, -0.00004531, 0.00003978, -0.00005147, 0.00005269, -0.00004872, 0.00004571, -0.00004332, 0.00003459, -0.00003699, 0.00003296, -0.00003275, 0.00003260, -0.00003320, 0.00003113, -0.00002442, 0.00002413, -0.00001335, 0.00000539, -0.00001771, 0.00001534, -0.00000996, 0.00000473, -0.00000246, -0.00000275, 0.00001303, -0.00002629, 0.00002476, -0.00002028, 0.00001111, },
		{ -0.00002468, 0.00001284, -0.00000977, 0.00000402, 0.00000442, -0.00001090, 0.00000788, -0.00000195, 0.00000833, -0.00001480, 0.00000254, -0.00000784, 0.00001734, -0.00001517, 0.00002186, -0.00002222, 0.00002091, -0.00002229, 0.00002284, -0.00002549, 0.00002169, -0.00002665, 0.00002865, -0.00003130, 0.00003516, -0.00003343, 0.00002231, -0.00002323, 0.00001914, -0.00002144, 0.00003504, -0.00004910, 0.00004871, -0.00005569, 0.00005401, -0.00005975, 0.00007417, -0.00007643, 0.00006500, -0.00007869, 0.00007862, -0.00008190, 0.00009017, -0.00009912, 0.00009820, -0.00011169, 0.00010843, -0.00012968, 0.00012686, -0.00014347, 0.00013592, -0.00016910, 0.00017495, -0.00020609, 0.00020109, -0.00027170, 0.00026902, -0.00036236, 0.00034018, -0.00057474, 0.00051874, -0.00132516, -0.00028845, 0.00464001, -0.00292679, 0.00079882, -0.00077921, 0.00044384, -0.00044960, 0.00030238, -0.00030139, 0.00023853, -0.00024467, 0.00019152, -0.00018124, 0.00015418, -0.00016086, 0.00013127, -0.00013377, 0.00012175, -0.00012470, 0.00010293, -0.00010335, 0.00009509, -0.00008742, 0.00007991, -0.00007788, 0.00007198, -0.00008405, 0.00007414, -0.00005934, 0.00005685, -0.00005835, 0.00004871, -0.00004713, 0.00003526, -0.00002235, 0.00002008, -0.00002566, 0.00002705, -0.00004129, 0.00003849, -0.00003358, 0.00003114, -0.00002889, 0.00002246, -0.00002774, 0.00002465, -0.00002633, 0.00002622, -0.00002497, 0.00002100, -0.00001541, 0.00001534, -0.00000715, 0.00000658, -0.00002046, 0.00001452, -0.00000777, 0.00000209, -0.00000364, 0.00000265, 0.00000263, -0.00000512, -0.00001075, 0.00002320, -0.00003243, 0.00003643, },
		{ -0.00001351, 0.00000600, -0.00000327, 0.00000539, -0.00000333, -0.00000714, 0.00001348, -0.00000445, 0.00000298, -0.00001685, 0.00001001, -0.00000526, 0.00001176, -0.00000801, 0.00001270, -0.00001606, 0.00001632, -0.00001858, 0.00001978, -0.00002548, 0.00002264, -0.00002365, 0.00002464, -0.00002610, 0.00003320, -0.00004380, 0.00004206, -0.00004505, 0.00003975, -0.00002709, 0.00002068, -0.00002923, 0.00002791, -0.00003396, 0.00003166, -0.00002673, 0.00003620, -0.00005030, 0.00004123, -0.00005178, 0.00005435, -0.00005282, 0.00005523, -0.00006654, 0.00006447, -0.00007688, 0.00006994, -0.00009249, 0.00009378, -0.00011388, 0.00010107, -0.00012582, 0.00012832, -0.00016223, 0.00014369, -0.00021517, 0.00021291, -0.00032805, 0.00028324, -0.00058784, 0.00050498, -0.00255052, 0.00456578, -0.00057323, -0.00104209, 0.00031446, -0.00040799, 0.00021819, -0.00026324, 0.00016649, -0.00017247, 0.00012709, -0.00014915, 0.00011956, -0.00011659, 0.00009071, -0.00010230, 0.00007788, -0.00007673, 0.00006704, -0.00007879, 0.00006428, -0.00006421, 0.00006312, -0.00006122, 0.00005354, -0.00005054, 0.00003769, -0.00004841, 0.00005476, -0.00004867, 0.00004403, -0.00004976, 0.00004624, -0.00005249, 0.00005565, -0.00004645, 0.00003280, -0.00002629, 0.00001220, -0.00001833, 0.00002089, -0.00002022, 0.00001979, -0.00002148, 0.00001236, -0.00001401, 0.00001016, -0.00001010, 0.00001119, -0.00001314, 0.00001432, -0.00001140, 0.00001450, -0.00000855, -0.00000445, -0.00000455, 0.00000920, -0.00000937, -0.00000033, 0.00000315, -0.00000121, 0.00000589, -0.00002474, 0.00002786, -0.00002219, 0.00000977, 0.00000878, },
		{ -0.00000483, 0.00000899, -0.00000609, 0.00001025, -0.00001931, 0.00001828, -0.00000556, 0.00000984, -0.00002341, 0.00001437, -0.00001324, 0.00002081, -0.00001720, 0.00002381, -0.00002604, 0.00002647, -0.00003004, 0.00003213, -0.00003656, 0.00003322, -0.00003481, 0.00003715, -0.00004038, 0.00004758, -0.00005362, 0.00004890, -0.00004873, 0.00004036, -0.00003056, 0.00003203, -0.00004542, 0.00004523, -0.00005234, 0.00005159, -0.00005277, 0.00006655, -0.00007734, 0.00007002, -0.00008460, 0.00008489, -0.00008475, 0.00009233, -0.00010530, 0.00010472, -0.00011802, 0.00011762, -0.00014574, 0.00014568, -0.00016267, 0.00015636, -0.00018997, 0.00019636, -0.00023114, 0.00023117, -0.00031601, 0.00032215, -0.00043096, 0.00042348, -0.00072053, 0.00072494, -0.00208014, 0.00174493, 0.00268883, -0.00233247, 0.00072277, -0.00072781, 0.00043632, -0.00044295, 0.00030107, -0.00029956, 0.00024443, -0.00025159, 0.00020036, -0.00019032, 0.00016186, -0.00016634, 0.00013381, -0.00013459, 0.00012607, -0.00013121, 0.00010933, -0.00010898, 0.00010482, -0.00009853, 0.00008756, -0.00008412, 0.00007573, -0.00008686, 0.00008574, -0.00007583, 0.00007156, -0.00007480, 0.00006783, -0.00007046, 0.00006645, -0.00005073, 0.00003619, -0.00003220, 0.00002389, -0.00003472, 0.00003538, -0.00003370, 0.00003255, -0.00003191, 0.00002266, -0.00002608, 0.00002260, -0.00002407, 0.00002492, -0.00002473, 0.00002333, -0.00001982, 0.00002064, -0.00001319, 0.00000493, -0.00001690, 0.00001790, -0.00001264, 0.00000022, -0.00000219, 0.00000700, -0.00000185, -0.00001008, -0.00000131, 0.00001510, -0.00002588, 0.00002922, -0.00001356, },
		{ 0.00002053, -0.00000870, 0.00000997, -0.00001256, 0.00000372, 0.00000783, 0.00000338, -0.00001319, 0.00000013, -0.00000040, 0.00000589, -0.00000066, 0.00000709, -0.00000629, 0.00000582, -0.00000828, 0.00000772, -0.00000821, 0.00000262, -0.00000325, 0.00000496, -0.00000662, 0.00001007, -0.00000963, 0.00000102, 0.00000308, -0.00001314, 0.00001854, -0.00001133, -0.00000166, -0.00000203, -0.00000314, 0.00000095, -0.00000298, 0.00001373, -0.00001651, 0.00000704, -0.00001787, 0.00000947, -0.00000724, 0.00001363, -0.00001959, 0.00001248, -0.00001990, 0.00001644, -0.00003453, 0.00001836, -0.00002310, 0.00001161, -0.00003379, 0.00002137, -0.00003678, 0.00002456, -0.00007448, 0.00003263, -0.00007876, 0.00002019, -0.00017699, 0.00001467, -0.00058947, -0.00088836, 0.00391633, -0.00133291, -0.00027835, -0.00016467, -0.00001935, -0.00008548, -0.00000622, -0.00005322, 0.00002656, -0.00004931, 0.00001087, -0.00002507, 0.00001374, -0.00002741, 0.00000569, -0.00002306, 0.00002199, -0.00002724, 0.00000968, -0.00001917, 0.00001847, -0.00001479, 0.00000885, -0.00001254, 0.00001143, -0.00002339, 0.00001926, -0.00001197, 0.00001176, -0.00001527, 0.00000876, -0.00001182, 0.00000538, 0.00000972, -0.00001711, 0.00001246, -0.00001237, -0.00000284, 0.00000245, -0.00000236, 0.00000253, -0.00000228, -0.00000377, -0.00000330, 0.00000127, -0.00000479, 0.00000547, -0.00000476, 0.00000398, -0.00000248, 0.00000347, 0.00000225, -0.00000481, -0.00000848, 0.00000559, 0.00000147, -0.00000933, -0.00000178, 0.00000790, -0.00000113, -0.00000498, -0.00001167, 0.00001732, -0.00001277, -0.00000125, 0.00002342, -0.00003149, },
		{ 0.00001597, -0.00000987, 0.00000830, -0.00001672, 0.00001976, -0.00000245, -0.00000020, -0.00001592, 0.00001310, -0.00000877, 0.00001369, -0.00000672, 0.00000990, -0.00001050, 0.00000977, -0.00001314, 0.00001503, -0.00002062, 0.00001916, -0.00001731, 0.00001745, -0.00001726, 0.00002184, -0.00003145, 0.00003603, -0.00004277, 0.00004084, -0.00002925, 0.00001916, -0.00002646, 0.00002210, -0.00002547, 0.00002379, -0.00001724, 0.00002083, -0.00003131, 0.00002483, -0.00003848, 0.00003820, -0.00003260, 0.00003331, -0.00004415, 0.00003928, -0.00004550, 0.00003818, -0.00006451, 0.00006155, -0.00007285, 0.00006020, -0.00008424, 0.00007696, -0.00009677, 0.00007654, -0.00014565, 0.00012543, -0.00020182, 0.00012960, -0.00036387, 0.00014764, -0.00159759, 0.00262164, 0.00101438, -0.00136893, 0.00024498, -0.00036112, 0.00016542, -0.00021678, 0.00011061, -0.00012868, 0.00009586, -0.00011810, 0.00008162, -0.00008439, 0.00006557, -0.00007710, 0.00004898, -0.00005113, 0.00004744, -0.00005821, 0.00004063, -0.00004157, 0.00004351, -0.00004431, 0.00003547, -0.00003417, 0.00002456, -0.00003050, 0.00003444, -0.00003233, 0.00002937, -0.00003486, 0.00003101, -0.00003844, 0.00004928, -0.00004682, 0.00003439, -0.00002962, 0.00001430, -0.00001747, 0.00001669, -0.00001662, 0.00001691, -0.00002022, 0.00001145, -0.00001402, 0.00001010, -0.00001064, 0.00001134, -0.00001096, 0.00001146, -0.00001116, 0.00001541, -0.00001439, 0.00000238, -0.00000895, 0.00001482, -0.00001602, 0.00000035, -0.00000074, 0.00001231, -0.00001509, 0.00000029, -0.00000643, 0.00001815, -0.00002994, 0.00003508, -0.00001936, -0.00000491, },
		{ -0.00001771, 0.00001414, -0.00001819, 0.00001306, 0.00000377, 0.00000355, -0.00001898, 0.00001093, -0.00000616, 0.00000935, -0.00000206, 0.00000732, -0.00000742, 0.00000807, -0.00001220, 0.00001310, -0.00001592, 0.00001209, -0.00000962, 0.00001104, -0.00001209, 0.00001731, -0.00002429, 0.00002514, -0.00002452, 0.00001512, -0.00000220, -0.00000339, -0.00000607, 0.00000083, -0.00000476, 0.00000293, 0.00000100, 0.00000538, -0.00001414, 0.00000884, -0.00002087, 0.00001351, -0.00000819, 0.00001244, -0.00002116, 0.00001393, -0.00001992, 0.00001715, -0.00004074, 0.00002779, -0.00003357, 0.00002333, -0.00004472, 0.00003070, -0.00004696, 0.00003520, -0.00009396, 0.00005649, -0.00011166, 0.00005448, -0.00024750, 0.00008850, -0.00107467, 0.00069357, 0.00244202, -0.00103758, -0.00023183, -0.00016421, 0.00000034, -0.00009505, 0.00000343, -0.00005669, 0.00003322, -0.00005512, 0.00001881, -0.00003390, 0.00002123, -0.00003266, 0.00000806, -0.00002191, 0.00002204, -0.00002895, 0.00001172, -0.00001988, 0.00002293, -0.00002195, 0.00001360, -0.00001579, 0.00001085, -0.00001877, 0.00002174, -0.00002052, 0.00001932, -0.00002446, 0.00002104, -0.00002913, 0.00003425, -0.00002349, 0.00000855, -0.00000529, -0.00000551, -0.00000251, 0.00000278, -0.00000497, 0.00000619, -0.00000821, 0.00000031, -0.00000510, 0.00000188, -0.00000419, 0.00000489, -0.00000456, 0.00000615, -0.00000732, 0.00001022, -0.00000712, -0.00000255, -0.00000657, 0.00000977, -0.00000634, -0.00000710, -0.00000317, 0.00001521, -0.00001104, -0.00000313, -0.00000676, 0.00001079, -0.00000767, -0.00000548, 0.00002944, -0.00004329, 0.00003567, },
		{ -0.00000530, 0.00000022, -0.00000569, 0.00001427, 0.00000213, -0.00001189, -0.00000157, 0.00000507, -0.00000157, 0.00000663, 0.00000143, -0.00000102, 0.00000272, -0.00000580, 0.00000430, -0.00000465, -0.00000012, 0.00000157, 0.00000164, -0.00000281, 0.00000703, -0.00001006, 0.00000614, 0.00000075, -0.00001237, 0.00002093, -0.00001981, 0.00000846, -0.00001393, 0.00000822, -0.00000968, 0.00001124, -0.00000318, -0.00000340, -0.00000225, -0.00000568, -0.00000635, 0.00000736, -0.00000023, -0.00000483, -0.00000381, -0.00000328, 0.00000235, -0.00001845, -0.00000266, -0.00000392, -0.00000291, -0.00001342, -0.00000510, -0.00001191, 0.00000349, -0.00004406, -0.00000707, -0.00003991, -0.00000687, -0.00013699, -0.00001903, -0.00059201, -0.00031861, 0.00216000, 0.00035909, -0.00083746, -0.00002730, -0.00013662, -0.00000308, -0.00007435, -0.00000725, -0.00002071, -0.00000021, -0.00002893, 0.00000060, -0.00001229, 0.00000120, -0.00001937, -0.00000311, 0.00000048, -0.00000327, -0.00000990, -0.00000457, 0.00000523, -0.00000203, -0.00000439, -0.00000123, -0.00000168, -0.00000632, 0.00000850, -0.00000740, 0.00000578, -0.00000939, 0.00000567, -0.00001222, 0.00001012, 0.00000606, -0.00001822, 0.00001622, -0.00002125, 0.00000894, -0.00000793, 0.00000417, -0.00000362, 0.00000317, -0.00000981, 0.00000351, -0.00000687, 0.00000356, -0.00000341, 0.00000320, -0.00000067, -0.00000003, 0.00000003, 0.00000502, -0.00001251, 0.00000208, -0.00000189, 0.00000738, -0.00001570, -0.00000006, 0.00000403, 0.00000954, -0.00002254, 0.00001245, -0.00001720, 0.00002578, -0.00003701, 0.00004465, -0.00003349, 0.00000934, 0.00000893, },
		{ 0.00000884, -0.00001175, 0.00001247, 0.00000585, -0.00000556, -0.00001024, 0.00000903, -0.00000311, 0.00000499, 0.00000454, -0.00000235, 0.00000410, -0.00000498, 0.00000175, -0.00000172, -0.00000196, 0.00000116, 0.00000351, -0.00000367, 0.00000638, -0.00000580, -0.00000131, 0.00000994, -0.00001715, 0.00001745, -0.00000979, -0.00000239, -0.00000283, -0.00000505, 0.00000364, -0.00000426, 0.00001342, -0.00001755, 0.00001085, -0.00001570, 0.00000427, -0.00001049, 0.00001858, -0.00001998, 0.00001260, -0.00002224, 0.00002165, -0.00003126, 0.00000978, -0.00002382, 0.00002045, -0.00003320, 0.00001728, -0.00004029, 0.00003390, -0.00006198, 0.00001520, -0.00007138, 0.00003698, -0.00014596, 0.00000303, -0.00039055, -0.00055632, 0.00123917, 0.00129684, -0.00052612, -0.00042620, 0.00001530, -0.00013014, 0.00001852, -0.00009126, 0.00003273, -0.00004634, 0.00001858, -0.00004646, 0.00002331, -0.00003027, 0.00001457, -0.00003622, 0.00002382, -0.00002306, 0.00001360, -0.00002822, 0.00002136, -0.00001535, 0.00001047, -0.00001685, 0.00001259, -0.00001917, 0.00001936, -0.00001555, 0.00001126, -0.00001282, 0.00000798, -0.00001122, 0.00000258, 0.00001428, -0.00001929, 0.00001113, -0.00001180, -0.00000252, 0.00000222, -0.00000506, 0.00000275, -0.00000144, -0.00000404, -0.00000266, -0.00000151, -0.00000227, 0.00000150, -0.00000195, 0.00000371, -0.00000173, -0.00000140, 0.00000681, -0.00001112, -0.00000060, -0.00000174, 0.00000752, -0.00001116, -0.00000423, -0.00000268, 0.00001918, -0.00002392, 0.00000987, -0.00001481, 0.00001492, -0.00001256, 0.00000253, 0.00001845, -0.00003488, 0.00003450, -0.00001664, },
		{ -0.00000110, 0.00000096, 0.00001151, -0.00000113, -0.00001185, 0.00000462, 0.00000220, -0.00000069, 0.00000799, -0.00000211, 0.00000295, -0.00000129, -0.00000238, 0.00000163, -0.00000422, 0.00000218, 0.00000196, 0.00000035, 0.00000063, 0.00000288, -0.00001094, 0.00001717, -0.00001687, 0.00001041, -0.00000038, -0.00000945, 0.00000286, -0.00001075, 0.00000717, -0.00000843, 0.00001647, -0.00001607, 0.00000702, -0.00000990, 0.00000130, -0.00001183, 0.00001562, -0.00001234, 0.00000633, -0.00001601, 0.00001218, -0.00001577, -0.00000296, -0.00001562, 0.00000854, -0.00001551, 0.00000210, -0.00002506, 0.00001163, -0.00002677, -0.00001428, -0.00004006, -0.00000516, -0.00006021, -0.00008461, -0.00014408, -0.00070004, 0.00068882, 0.00120537, 0.00042565, -0.00077521, 0.00000773, -0.00015343, 0.00002109, -0.00009436, 0.00001821, -0.00004240, 0.00002021, -0.00004460, 0.00001319, -0.00002423, 0.00001322, -0.00003232, 0.00001365, -0.00001614, 0.00001126, -0.00002311, 0.00001033, -0.00000628, 0.00000518, -0.00001223, 0.00000713, -0.00001383, 0.00001213, -0.00000530, 0.00000034, -0.00000268, -0.00000193, 0.00000068, -0.00001140, 0.00002258, -0.00001742, 0.00000414, -0.00000260, -0.00001154, 0.00000814, -0.00000869, 0.00000376, -0.00000263, -0.00000065, -0.00000653, 0.00000216, -0.00000676, 0.00000515, -0.00000562, 0.00000570, -0.00000148, -0.00000184, 0.00000434, -0.00000403, -0.00000885, 0.00000402, 0.00000119, -0.00000121, -0.00001144, -0.00000177, 0.00001118, -0.00000341, -0.00001409, 0.00001066, -0.00001784, 0.00002357, -0.00003197, 0.00003921, -0.00003325, 0.00001582, 0.00000440, -0.00000479, },
		{ -0.00000059, 0.00000745, 0.00000672, -0.00001197, 0.00000074, 0.00000340, 0.00000076, 0.00000257, 0.00000622, -0.00000519, 0.00000821, -0.00001075, 0.00000918, -0.00001164, 0.00000957, -0.00000715, 0.00001186, -0.00001159, 0.00001660, -0.00002311, 0.00002449, -0.00001758, 0.00000920, -0.00000155, -0.00000335, -0.00000608, 0.00000014, -0.00000710, 0.00000593, -0.00000093, 0.00000708, -0.00001797, 0.00001521, -0.00002113, 0.00001110, -0.00001472, 0.00002109, -0.00002561, 0.00001846, -0.00002714, 0.00002777, -0.00004337, 0.00002596, -0.00004133, 0.00003871, -0.00005053, 0.00003429, -0.00005912, 0.00005399, -0.00009192, 0.00005223, -0.00011876, 0.00008805, -0.00023049, 0.00010124, -0.00078362, 0.00038074, 0.00084262, 0.00090639, -0.00033441, -0.00044829, 0.00005967, -0.00015681, 0.00004396, -0.00010700, 0.00005073, -0.00006332, 0.00003737, -0.00006399, 0.00003791, -0.00004229, 0.00002373, -0.00004128, 0.00003002, -0.00003137, 0.00002217, -0.00003499, 0.00003164, -0.00002784, 0.00001997, -0.00002440, 0.00001625, -0.00001983, 0.00002775, -0.00002907, 0.00002312, -0.00002648, 0.00002495, -0.00003298, 0.00003464, -0.00002042, 0.00000583, -0.00000514, -0.00000646, -0.00000137, 0.00000290, -0.00000835, 0.00000696, -0.00000767, 0.00000026, -0.00000465, -0.00000044, -0.00000237, 0.00000204, -0.00000332, 0.00000743, -0.00000770, 0.00000594, -0.00000221, -0.00000951, 0.00000101, 0.00000327, -0.00000031, -0.00000967, -0.00000449, 0.00000330, 0.00001101, -0.00002351, 0.00001551, -0.00002016, 0.00001632, -0.00001320, 0.00000589, 0.00000844, -0.00002030, 0.00002461, -0.00001070, 0.00000114, },
		{ -0.00000114, 0.00001260, -0.00000947, -0.00000136, 0.00000083, 0.00000546, -0.00000382, 0.00001196, -0.00000849, 0.00001127, -0.00001148, 0.00000935, -0.00001174, 0.00000913, -0.00000764, 0.00001288, -0.00001134, 0.00001598, -0.00001864, 0.00001453, -0.00000402, -0.00000191, 0.00000527, -0.00000546, -0.00000615, 0.00000154, -0.00001022, 0.00000781, -0.00000573, 0.00001569, -0.00002470, 0.00001896, -0.00002227, 0.00001451, -0.00002272, 0.00002694, -0.00002833, 0.00002316, -0.00003360, 0.00003352, -0.00004410, 0.00002845, -0.00004732, 0.00004087, -0.00004831, 0.00003749, -0.00006553, 0.00005793, -0.00008684, 0.00005552, -0.00012297, 0.00008979, -0.00019740, 0.00009055, -0.00055550, -0.00009609, 0.00083368, 0.00061283, 0.00054777, -0.00078305, 0.00008123, -0.00020577, 0.00007207, -0.00013472, 0.00006110, -0.00008206, 0.00005870, -0.00007842, 0.00004235, -0.00005006, 0.00003589, -0.00005201, 0.00003588, -0.00004000, 0.00003406, -0.00004400, 0.00003437, -0.00003010, 0.00002428, -0.00002925, 0.00002147, -0.00002847, 0.00003561, -0.00003116, 0.00002295, -0.00002695, 0.00002468, -0.00002860, 0.00002240, -0.00000508, -0.00000486, 0.00000107, -0.00000769, -0.00000531, 0.00000773, -0.00001179, 0.00000765, -0.00000763, 0.00000158, -0.00000696, 0.00000261, -0.00000721, 0.00000697, -0.00000876, 0.00001092, -0.00000774, 0.00000408, -0.00000064, -0.00000690, -0.00000577, 0.00000798, -0.00000234, -0.00000494, -0.00000844, 0.00000056, 0.00001032, -0.00001193, -0.00000184, 0.00000137, -0.00001211, 0.00001596, -0.00002020, 0.00002227, -0.00001626, 0.00000677, 0.00000980, -0.00001245, 0.00000799, },
		{ 0.00000604, 0.00000043, -0.00000655, 0.00000238, 0.00000331, -0.00000045, 0.00000579, 0.00000012, 0.00000252, -0.00000060, -0.00000141, -0.00000037, -0.00000358, 0.00000480, -0.00000023, 0.00000375, -0.00000010, 0.00000186, -0.00000965, 0.00001971, -0.00002052, 0.00002003, -0.00001691, 0.00000466, -0.00000979, 0.00000177, -0.00000659, 0.00000638, 0.00000407, -0.00000733, -0.00000280, 0.00000102, -0.00000691, -0.00000118, -0.00000021, 0.00000151, -0.00000567, -0.00000287, -0.00000143, -0.00000344, -0.00001200, -0.00000381, -0.00001125, 0.00000735, -0.00001539, -0.00000649, -0.00001190, -0.00000524, -0.00002706, -0.00002122, -0.00003580, -0.00002784, -0.00009111, -0.00017307, -0.00057049, 0.00088837, 0.00027912, 0.00090070, -0.00036759, -0.00032713, -0.00001656, -0.00008694, -0.00001202, -0.00005008, 0.00000092, -0.00001782, -0.00000132, -0.00003010, 0.00000708, -0.00001408, -0.00000301, -0.00001132, -0.00000005, -0.00000388, -0.00000339, -0.00000708, 0.00000638, -0.00000803, 0.00000129, -0.00000674, -0.00000392, 0.00000803, 0.00000147, -0.00000772, 0.00000080, -0.00000395, 0.00000305, -0.00001241, 0.00002612, -0.00002717, 0.00001751, -0.00001881, 0.00000211, -0.00000133, -0.00000015, -0.00000539, 0.00000392, -0.00000771, 0.00000107, -0.00000440, -0.00000138, 0.00000028, -0.00000163, 0.00000142, 0.00000320, -0.00000538, 0.00000613, -0.00000910, -0.00000556, 0.00000340, 0.00000431, -0.00000801, -0.00000505, -0.00000493, 0.00000837, -0.00000298, -0.00000966, 0.00000702, -0.00001431, 0.00000731, -0.00000154, -0.00000700, 0.00001654, -0.00001911, 0.00002213, -0.00001186, 0.00000276, 0.00000367, },
		{ 0.00000971, -0.00001082, 0.00000626, -0.00000280, 0.00000705, -0.00000322, 0.00000924, -0.00000556, 0.00000872, -0.00000979, 0.00000879, -0.00001350, 0.00001337, -0.00000891, 0.00001342, -0.00000996, 0.00001484, -0.00002268, 0.00002839, -0.00002307, 0.00001988, -0.00001440, 0.00000275, -0.00000987, 0.00000343, -0.00001003, 0.00000745, 0.00000105, 0.00000242, -0.00001460, 0.00001155, -0.00001571, 0.00000956, -0.00001469, 0.00001511, -0.00001798, 0.00001211, -0.00001953, 0.00001742, -0.00003164, 0.00001937, -0.00003834, 0.00003097, -0.00003575, 0.00002166, -0.00004650, 0.00003337, -0.00006517, 0.00003381, -0.00010075, 0.00005384, -0.00016818, 0.00002047, -0.00068042, 0.00057749, 0.00042650, 0.00051180, 0.00045869, -0.00065253, 0.00003318, -0.00016173, 0.00004018, -0.00010010, 0.00003445, -0.00005891, 0.00004068, -0.00006257, 0.00002749, -0.00003585, 0.00002096, -0.00003347, 0.00001869, -0.00002586, 0.00002097, -0.00002906, 0.00002352, -0.00002373, 0.00001693, -0.00002113, 0.00000979, -0.00001161, 0.00002404, -0.00002514, 0.00001619, -0.00002173, 0.00002244, -0.00003102, 0.00003775, -0.00003018, 0.00001705, -0.00001590, -0.00000044, -0.00000380, 0.00000548, -0.00001107, 0.00000770, -0.00001046, 0.00000382, -0.00000734, 0.00000207, -0.00000515, 0.00000467, -0.00000654, 0.00001042, -0.00000986, 0.00000875, -0.00000944, -0.00000389, -0.00000393, 0.00001162, -0.00001057, -0.00000203, -0.00000884, 0.00000739, -0.00000247, -0.00000413, -0.00000262, 0.00000067, -0.00001443, 0.00001816, -0.00002032, 0.00001566, -0.00000285, -0.00000331, 0.00001457, -0.00001729, 0.00001682, -0.00000426, },
		{ -0.00000602, 0.00000403, -0.00000207, 0.00000619, -0.00000214, 0.00000709, -0.00000282, 0.00000722, -0.00000706, 0.00000676, -0.00001074, 0.00000798, -0.00000337, 0.00000824, -0.00000451, 0.00001101, -0.00001570, 0.00001537, -0.00000527, 0.00000139, 0.00000504, -0.00001406, 0.00000358, -0.00000862, 0.00000142, -0.00000603, 0.00001084, -0.00000217, -0.00000814, 0.00000169, -0.00000462, -0.00000003, -0.00000511, 0.00000244, -0.00000495, 0.00000120, -0.00000890, 0.00000635, -0.00001770, 0.00000609, -0.00002200, 0.00000705, -0.00001056, 0.00000276, -0.00002607, 0.00000961, -0.00003741, 0.00001189, -0.00006793, 0.00001455, -0.00010869, 0.00000541, -0.00048865, 0.00005784, 0.00073599, 0.00007495, 0.00082797, -0.00034231, -0.00026661, -0.00004996, -0.00005503, -0.00003246, -0.00002625, -0.00002052, 0.00000356, -0.00002019, -0.00001359, -0.00000705, -0.00000306, -0.00001022, -0.00000202, -0.00001135, 0.00000673, -0.00001219, 0.00000586, -0.00000841, 0.00000312, -0.00000777, -0.00000123, -0.00000718, 0.00001892, -0.00001380, 0.00000592, -0.00001484, 0.00001493, -0.00002035, 0.00002009, -0.00000721, -0.00000523, 0.00000488, -0.00001674, 0.00000564, -0.00000194, -0.00000250, -0.00000192, -0.00000148, -0.00000424, -0.00000027, -0.00000372, -0.00000112, 0.00000071, -0.00000298, 0.00000495, -0.00000253, 0.00000154, -0.00000171, -0.00000895, -0.00000311, 0.00000747, -0.00000143, -0.00000909, -0.00000347, 0.00000006, 0.00000199, -0.00000560, -0.00000050, -0.00000137, -0.00000886, 0.00000013, 0.00000823, -0.00002197, 0.00003335, -0.00002841, 0.00002639, -0.00001771, 0.00001135, 0.00000160, 0.00000557, },
		{ -0.00000210, 0.00000488, -0.00000127, 0.00000511, 0.00000004, 0.00000315, 0.00000257, -0.00000092, 0.00000107, -0.00000323, -0.00000181, 0.00000519, 0.00000063, 0.00000306, 0.00000408, -0.00000415, -0.00000088, 0.00001227, -0.00001387, 0.00001874, -0.00002237, 0.00000775, -0.00001227, 0.00000508, -0.00001039, 0.00001044, 0.00000077, -0.00000643, -0.00000296, -0.00000041, -0.00000328, -0.00000041, -0.00000354, -0.00000112, -0.00000126, -0.00000517, 0.00000166, -0.00001046, -0.00000177, -0.00000949, -0.00000895, 0.00000066, -0.00000316, -0.00001503, -0.00000394, -0.00002198, -0.00000284, -0.00003762, -0.00001939, -0.00006221, -0.00002527, -0.00028851, -0.00029059, 0.00074830, 0.00009517, 0.00053304, 0.00032563, -0.00049481, -0.00004904, -0.00009126, -0.00001301, -0.00004658, -0.00001177, -0.00001616, 0.00000314, -0.00002916, -0.00000198, -0.00001008, -0.00000214, -0.00000756, -0.00000781, -0.00000136, -0.00000203, -0.00000261, -0.00000178, -0.00000382, -0.00000099, -0.00000554, -0.00000642, 0.00001293, -0.00000217, -0.00000218, -0.00000825, 0.00000455, -0.00000586, 0.00000033, 0.00001404, -0.00002223, 0.00001772, -0.00002345, 0.00000693, -0.00000433, 0.00000233, -0.00000715, 0.00000294, -0.00000848, 0.00000346, -0.00000682, 0.00000159, -0.00000314, 0.00000139, -0.00000141, 0.00000436, -0.00000456, 0.00000519, -0.00001368, -0.00000013, -0.00000092, 0.00000985, -0.00001581, 0.00000146, -0.00000684, 0.00000834, -0.00001361, 0.00001007, -0.00001269, 0.00000751, -0.00002320, 0.00002823, -0.00003302, 0.00002608, -0.00000464, -0.00000354, 0.00001103, -0.00001195, 0.00001757, -0.00000394, 0.00000175, },
		{ 0.00000808, -0.00000345, 0.00000595, 0.00000019, 0.00000209, 0.00000320, 0.00000092, -0.00000074, 0.00000067, -0.00000633, 0.00000683, -0.00000002, 0.00000402, 0.00000259, 0.00000208, -0.00000880, 0.00001777, -0.00001480, 0.00001628, -0.00001352, -0.00000342, -0.00000287, -0.00000372, -0.00000146, -0.00000245, 0.00001270, -0.00001241, 0.00000200, -0.00000703, 0.00000351, -0.00000617, 0.00000366, -0.00001081, 0.00000744, -0.00001189, 0.00000805, -0.00001488, 0.00000201, -0.00001081, -0.00000498, -0.00001063, 0.00000922, -0.00002086, 0.00000278, -0.00002810, 0.00000141, -0.00003088, -0.00001956, -0.00005610, -0.00002949, -0.00015880, -0.00041373, 0.00050188, 0.00035000, 0.00014185, 0.00063597, -0.00021627, -0.00030056, -0.00000975, -0.00008557, 0.00000319, -0.00005544, 0.00000701, -0.00002004, 0.00000062, -0.00003176, 0.00000955, -0.00001971, 0.00001026, -0.00002216, 0.00000664, -0.00001067, 0.00000788, -0.00001085, 0.00000308, -0.00000813, 0.00000317, -0.00001506, 0.00001404, -0.00000046, 0.00000086, -0.00000959, 0.00000070, 0.00000034, -0.00000799, 0.00002002, -0.00002140, 0.00001179, -0.00001237, -0.00000589, 0.00000410, -0.00000320, -0.00000157, -0.00000265, -0.00000362, -0.00000169, -0.00000158, -0.00000316, 0.00000028, -0.00000167, 0.00000030, 0.00000230, -0.00000218, 0.00000404, -0.00001015, -0.00000440, -0.00000146, 0.00000954, -0.00000940, -0.00000509, -0.00000360, 0.00000592, -0.00001202, 0.00000626, -0.00000550, 0.00000099, -0.00001336, 0.00000495, 0.00000234, -0.00002033, 0.00003872, -0.00003197, 0.00002529, -0.00001652, 0.00001559, 0.00000098, -0.00000012, -0.00000308, },
		{ 0.00000132, 0.00000053, 0.00000501, -0.00000127, 0.00000383, 0.00000294, -0.00000230, 0.00000367, -0.00000812, 0.00000550, 0.00000077, 0.00000474, 0.00000034, 0.00000801, -0.00001311, 0.00001755, -0.00000978, 0.00000844, -0.00000158, -0.00001320, 0.00000233, -0.00000792, 0.00000259, -0.00000799, 0.00001392, -0.00000793, -0.00000166, -0.00000469, 0.00000008, -0.00000231, 0.00000166, -0.00000817, 0.00000080, -0.00000348, -0.00000046, -0.00000383, -0.00000979, 0.00000136, -0.00001272, -0.00000565, -0.00000153, -0.00000312, -0.00001294, -0.00000848, -0.00002336, 0.00000078, -0.00004315, -0.00001982, -0.00007614, -0.00003068, -0.00045122, 0.00024831, 0.00047379, 0.00007236, 0.00043241, 0.00030834, -0.00043436, -0.00006119, -0.00008058, -0.00001220, -0.00004344, -0.00001289, -0.00001348, 0.00000010, -0.00002607, -0.00000358, -0.00001094, 0.00000274, -0.00001116, -0.00000595, -0.00000345, 0.00000177, -0.00000201, -0.00000578, -0.00000200, -0.00000144, -0.00000888, 0.00000183, 0.00001004, -0.00000450, -0.00000072, -0.00001098, 0.00001018, -0.00001635, 0.00002278, -0.00001610, 0.00000243, -0.00000078, -0.00001519, 0.00000729, -0.00000465, 0.00000066, -0.00000452, -0.00000205, -0.00000443, 0.00000133, -0.00000565, 0.00000208, -0.00000366, 0.00000161, 0.00000023, -0.00000006, 0.00000210, -0.00000436, -0.00001073, 0.00000137, 0.00000323, 0.00000189, -0.00001364, 0.00000126, 0.00000009, -0.00000275, -0.00000926, 0.00001289, -0.00001594, 0.00000786, -0.00002362, 0.00002846, -0.00003843, 0.00003606, -0.00001022, -0.00000157, 0.00000719, -0.00000388, 0.00001487, -0.00000536, -0.00000194, 0.00000658, },
		{ -0.00000118, 0.00000425, 0.00000205, -0.00000175, 0.00000874, -0.00000591, 0.00000753, -0.00000955, 0.00000489, -0.00000090, 0.00000817, -0.00000405, 0.00001364, -0.00001434, 0.00001408, -0.00000315, 0.00000141, 0.00000565, -0.00001425, -0.00000170, -0.00000397, -0.00000179, -0.00000310, 0.00000369, 0.00000567, -0.00001258, 0.00000561, -0.00001157, 0.00000877, -0.00000862, 0.00000578, -0.00001694, 0.00001340, -0.00001694, 0.00001527, -0.00002844, 0.00001857, -0.00002676, 0.00001122, -0.00002630, 0.00002439, -0.00003711, 0.00002178, -0.00005707, 0.00003537, -0.00007108, 0.00002498, -0.00012665, 0.00005327, -0.00040724, 0.00002361, 0.00047056, 0.00016593, 0.00017531, 0.00049894, -0.00012349, -0.00032020, 0.00001565, -0.00010099, 0.00002393, -0.00007274, 0.00002535, -0.00003720, 0.00001583, -0.00004374, 0.00001868, -0.00002622, 0.00001934, -0.00003265, 0.00001656, -0.00001983, 0.00002139, -0.00002519, 0.00001333, -0.00001657, 0.00000847, -0.00001875, 0.00002562, -0.00001616, 0.00001469, -0.00002512, 0.00001947, -0.00002284, 0.00002282, -0.00000971, -0.00000422, 0.00000372, -0.00001414, 0.00000037, 0.00000202, -0.00000464, 0.00000103, -0.00000678, -0.00000143, -0.00000205, -0.00000167, -0.00000235, 0.00000061, -0.00000318, 0.00000404, -0.00000350, 0.00000414, -0.00000215, -0.00001166, -0.00000127, 0.00000219, 0.00000480, -0.00001151, -0.00000308, 0.00000077, 0.00000169, -0.00001687, 0.00001599, -0.00001270, 0.00000492, -0.00001809, 0.00001062, -0.00000900, -0.00000716, 0.00003154, -0.00002690, 0.00001767, -0.00000695, 0.00001270, 0.00000212, -0.00000733, 0.00000634, 0.00000521, },
		{ -0.00000062, 0.00000750, -0.00000629, 0.00001032, -0.00000440, 0.00000601, -0.00000553, 0.00000046, 0.00000065, 0.00000694, -0.00000200, 0.00001036, -0.00000559, 0.00000239, 0.00000910, -0.00000851, 0.00001283, -0.00001436, -0.00000390, -0.00000418, -0.00000193, -0.00000208, -0.00000103, 0.00001010, -0.00001271, 0.00000520, -0.00001129, 0.00000657, -0.00000654, 0.00000740, -0.00001817, 0.00001023, -0.00001322, 0.00001298, -0.00002295, 0.00000994, -0.00001636, 0.00000496, -0.00002298, 0.00001658, -0.00002447, 0.00001355, -0.00004478, 0.00001642, -0.00004481, 0.00000877, -0.00009599, 0.00001444, -0.00023821, -0.00022447, 0.00046735, 0.00018598, 0.00015713, 0.00028723, 0.00034724, -0.00042702, -0.00003618, -0.00009974, 0.00001447, -0.00006579, 0.00001050, -0.00003425, 0.00001772, -0.00004038, 0.00000920, -0.00002310, 0.00001835, -0.00002751, 0.00000949, -0.00001854, 0.00002015, -0.00001859, 0.00000593, -0.00001263, 0.00000756, -0.00001923, 0.00002063, -0.00001001, 0.00001200, -0.00001864, 0.00000902, -0.00001154, 0.00000702, 0.00000800, -0.00001669, 0.00001014, -0.00001390, -0.00000373, 0.00000383, -0.00000501, 0.00000156, -0.00000596, -0.00000302, -0.00000235, -0.00000032, -0.00000424, 0.00000258, -0.00000541, 0.00000499, -0.00000368, 0.00000246, 0.00000168, -0.00001090, -0.00000584, 0.00000325, 0.00000322, -0.00000497, -0.00000909, 0.00000196, 0.00000301, -0.00001457, 0.00000369, 0.00000484, -0.00001051, 0.00000053, -0.00001453, 0.00001466, -0.00002662, 0.00003162, -0.00000686, -0.00000692, 0.00001220, -0.00000227, 0.00001442, -0.00001124, 0.00000367, 0.00000745, 0.00000187, },
		{ 0.00000531, -0.00000119, 0.00000203, 0.00000499, -0.00000209, 0.00000411, -0.00000800, 0.00000670, -0.00000044, 0.00000705, -0.00000044, 0.00000919, -0.00001210, 0.00002151, -0.00001653, 0.00001728, -0.00001371, -0.00000298, -0.00000879, 0.00000183, -0.00000555, 0.00000112, 0.00000496, -0.00000311, -0.00000467, -0.00000047, -0.00000627, 0.00000509, -0.00000235, -0.00000393, -0.00000867, 0.00000428, -0.00000427, -0.00000065, -0.00001499, 0.00000838, -0.00001680, 0.00000148, -0.00001494, 0.00000924, -0.00001747, -0.00000406, -0.00003214, 0.00000834, -0.00004148, -0.00001906, -0.00008117, -0.00004995, -0.00039751, 0.00040004, 0.00020007, 0.00019937, 0.00009799, 0.00046909, -0.00011183, -0.00028298, -0.00000753, -0.00007265, 0.00000274, -0.00005140, 0.00000839, -0.00002362, 0.00000432, -0.00003188, 0.00000610, -0.00001033, 0.00000396, -0.00001877, 0.00000322, -0.00000470, 0.00000995, -0.00001814, 0.00000589, -0.00000919, -0.00000201, -0.00000136, 0.00001085, -0.00000637, 0.00000384, -0.00001402, 0.00000987, -0.00001655, 0.00002865, -0.00002829, 0.00001528, -0.00001468, -0.00000315, -0.00000098, 0.00000116, -0.00000470, 0.00000150, -0.00000965, 0.00000157, -0.00000386, -0.00000063, -0.00000145, -0.00000085, -0.00000124, 0.00000315, -0.00000517, 0.00000834, -0.00001117, -0.00000720, -0.00000016, 0.00000551, -0.00000409, -0.00000719, -0.00000371, 0.00000722, -0.00001145, -0.00000668, 0.00001194, -0.00001000, -0.00000031, -0.00001162, 0.00000225, -0.00000672, -0.00000203, 0.00002709, -0.00002443, 0.00001432, 0.00000094, 0.00000922, -0.00000035, -0.00000716, 0.00001162, 0.00000331, -0.00000046, },
		{ 0.00000691, -0.00000651, 0.00001158, -0.00000650, 0.00000950, -0.00001136, 0.00000854, -0.00000425, 0.00001164, -0.00000551, 0.00001551, -0.00001524, 0.00002107, -0.00001122, 0.00000967, -0.00000420, -0.00000827, -0.00000648, -0.00000223, -0.00000211, -0.00000202, 0.00000450, 0.00000037, -0.00000747, 0.00000300, -0.00001011, 0.00000644, -0.00000363, 0.00000268, -0.00001608, 0.00000748, -0.00000837, 0.00000808, -0.00002310, 0.00001368, -0.00002002, 0.00000863, -0.00002489, 0.00001591, -0.00002228, 0.00001035, -0.00004747, 0.00002134, -0.00005248, 0.00001306, -0.00011358, 0.00002010, -0.00038324, 0.00018190, 0.00031748, 0.00012099, 0.00014913, 0.00022789, 0.00033456, -0.00038510, -0.00004469, -0.00008703, 0.00000999, -0.00006035, 0.00000938, -0.00003443, 0.00001732, -0.00003778, 0.00000528, -0.00001656, 0.00001347, -0.00002439, 0.00000690, -0.00001506, 0.00002089, -0.00002200, 0.00000713, -0.00001267, 0.00000401, -0.00001159, 0.00001906, -0.00001348, 0.00001408, -0.00002195, 0.00001589, -0.00002365, 0.00002987, -0.00002060, 0.00000493, -0.00000476, -0.00000935, -0.00000028, 0.00000158, -0.00000540, 0.00000295, -0.00000931, -0.00000050, -0.00000330, -0.00000017, -0.00000300, 0.00000180, -0.00000519, 0.00000671, -0.00000799, 0.00000836, -0.00000626, -0.00000953, -0.00000373, 0.00000718, -0.00000377, -0.00000442, -0.00000786, 0.00000741, -0.00000564, -0.00001239, 0.00000758, 0.00000115, -0.00000933, -0.00000079, -0.00001212, 0.00000429, -0.00001195, 0.00002102, 0.00000065, -0.00001447, 0.00002169, -0.00000531, 0.00001339, -0.00001465, 0.00001130, 0.00000456, 0.00000384, -0.00000111, },
		{ -0.00000451, 0.00000702, -0.00000121, 0.00000572, -0.00000548, 0.00000219, 0.00000028, 0.00000660, 0.00000003, 0.00000960, -0.00000532, 0.00000814, 0.00000513, -0.00000600, 0.00001026, -0.00001753, 0.00000130, -0.00001177, 0.00000543, -0.00000868, 0.00000822, -0.00000204, -0.00000345, -0.00000129, -0.00000442, -0.00000184, 0.00000305, -0.00000022, -0.00000970, -0.00000317, -0.00000084, 0.00000336, -0.00001429, 0.00000110, -0.00000659, -0.00000212, -0.00001194, -0.00000196, -0.00000557, 0.00000054, -0.00003086, -0.00000109, -0.00002831, -0.00000148, -0.00007991, -0.00001020, -0.00025047, -0.00008139, 0.00042846, 0.00004013, 0.00023364, 0.00003303, 0.00044945, -0.00010788, -0.00024691, -0.00002750, -0.00004871, -0.00001614, -0.00003081, -0.00000950, -0.00000915, -0.00000689, -0.00002197, -0.00000183, -0.00000018, -0.00000749, -0.00000747, -0.00000721, 0.00000930, -0.00000433, -0.00000773, -0.00000281, -0.00000350, -0.00000673, 0.00001133, -0.00000511, 0.00000707, -0.00001143, 0.00000486, -0.00001287, 0.00001197, 0.00000285, -0.00001561, 0.00001129, -0.00001984, 0.00000528, -0.00000399, 0.00000037, -0.00000297, -0.00000144, -0.00000831, 0.00000176, -0.00000443, 0.00000049, -0.00000121, -0.00000215, 0.00000203, -0.00000171, -0.00000038, 0.00000405, -0.00001411, -0.00000342, 0.00000263, 0.00000269, -0.00000854, -0.00000333, -0.00000109, 0.00000524, -0.00001797, 0.00000420, 0.00000334, -0.00000393, -0.00000754, -0.00000205, -0.00001364, 0.00000796, -0.00000947, 0.00003120, -0.00002863, 0.00001945, 0.00000196, 0.00000755, -0.00000503, 0.00000010, 0.00000999, 0.00000486, -0.00000135, 0.00000659, },
		{ -0.00000081, 0.00000519, 0.00000077, 0.00000180, -0.00000473, 0.00000576, 0.00000015, 0.00000671, 0.00000245, 0.00000496, -0.00000339, 0.00001756, -0.00001447, 0.00001570, -0.00001823, 0.00000173, -0.00001307, 0.00000383, -0.00000680, 0.00000449, 0.00000168, -0.00000492, -0.00000087, -0.00000299, -0.00000428, 0.00000249, 0.00000189, -0.00000667, -0.00000722, -0.00000193, 0.00000442, -0.00000959, -0.00000556, -0.00000122, -0.00000570, -0.00000506, -0.00000989, -0.00000266, 0.00000077, -0.00002069, -0.00001400, -0.00001628, -0.00001115, -0.00004687, -0.00004360, -0.00013697, -0.00023740, 0.00039094, 0.00010077, 0.00016848, 0.00008014, 0.00023069, 0.00028147, -0.00031610, -0.00007733, -0.00005264, -0.00001675, -0.00003341, -0.00001237, -0.00001650, 0.00000175, -0.00002258, -0.00000878, 0.00000014, -0.00000382, -0.00000809, -0.00000836, 0.00000287, 0.00000482, -0.00001055, -0.00000315, -0.00000391, -0.00000649, 0.00000690, 0.00000040, 0.00000133, -0.00000252, -0.00000330, -0.00000338, -0.00000436, 0.00001972, -0.00002516, 0.00001532, -0.00001929, 0.00000215, -0.00000296, 0.00000070, -0.00000445, 0.00000166, -0.00001005, 0.00000127, -0.00000473, 0.00000096, -0.00000225, 0.00000009, -0.00000216, 0.00000352, -0.00000642, 0.00000886, -0.00001313, -0.00000475, -0.00000228, 0.00000840, -0.00001159, 0.00000016, -0.00000702, 0.00001054, -0.00001712, 0.00000005, -0.00000008, 0.00000620, -0.00001592, 0.00000766, -0.00002338, 0.00000968, -0.00000999, 0.00001805, 0.00000180, -0.00001424, 0.00002607, -0.00000707, 0.00000843, -0.00000974, 0.00001268, 0.00000436, 0.00000367, 0.00000069, 0.00000260, },
		{ 0.00000436, 0.00000116, 0.00000429, -0.00000615, 0.00000603, -0.00000096, 0.00000745, 0.00000124, 0.00000838, -0.00000649, 0.00001936, -0.00001038, 0.00000911, -0.00000873, -0.00000681, -0.00000514, -0.00000607, 0.00000154, -0.00000448, 0.00000946, -0.00000970, 0.00000247, -0.00000519, -0.00000126, -0.00000354, 0.00000705, -0.00000719, -0.00000427, -0.00000907, 0.00000776, -0.00000758, -0.00000648, -0.00000294, -0.00000304, -0.00000558, -0.00000653, -0.00001051, 0.00000592, -0.00001530, -0.00001699, -0.00001502, -0.00001433, -0.00002580, -0.00005736, -0.00007142, -0.00028096, 0.00024263, 0.00023033, 0.00005670, 0.00017934, 0.00004384, 0.00038325, -0.00006627, -0.00024714, -0.00001419, -0.00005744, -0.00000502, -0.00003760, -0.00000385, -0.00001493, 0.00000004, -0.00002892, 0.00000833, -0.00001029, 0.00000114, -0.00001531, 0.00000272, 0.00000258, -0.00000172, -0.00001076, 0.00000061, -0.00001040, 0.00000648, 0.00000144, 0.00000017, 0.00000014, -0.00000453, 0.00000043, -0.00001230, 0.00002267, -0.00001891, 0.00000534, -0.00000728, -0.00000962, 0.00000475, -0.00000488, -0.00000018, -0.00000195, -0.00000449, -0.00000504, -0.00000041, -0.00000315, 0.00000106, -0.00000210, -0.00000093, 0.00000197, -0.00000414, 0.00000431, -0.00000487, -0.00001005, -0.00000236, 0.00000625, -0.00000560, -0.00000579, -0.00000250, 0.00000429, -0.00000624, -0.00000979, 0.00000260, 0.00000309, -0.00000574, -0.00000476, -0.00000466, -0.00001842, 0.00001680, -0.00001585, 0.00003423, -0.00003040, 0.00002496, 0.00000079, 0.00000363, -0.00000428, 0.00000545, 0.00000767, 0.00000592, -0.00000028, 0.00000367, -0.00000139, },
		{ 0.00000167, 0.00000583, -0.00000532, 0.00000417, 0.00000025, 0.00000584, 0.00000191, 0.00000926, -0.00000589, 0.00001602, -0.00000123, 0.00000014, 0.00000067, -0.00001388, 0.00000044, -0.00001153, 0.00000380, -0.00000690, 0.00000974, -0.00000652, -0.00000169, -0.00000126, -0.00000346, -0.00000281, 0.00000345, -0.00000047, -0.00000717, -0.00000678, -0.00000089, 0.00000370, -0.00001408, 0.00000251, -0.00000889, 0.00000153, -0.00001054, -0.00000594, -0.00000623, 0.00000412, -0.00003012, -0.00000130, -0.00003227, 0.00000256, -0.00007387, -0.00002111, -0.00027692, 0.00008771, 0.00030161, 0.00003852, 0.00016134, 0.00005831, 0.00020872, 0.00025432, -0.00027339, -0.00008656, -0.00004103, -0.00002384, -0.00002347, -0.00002041, -0.00001037, -0.00000244, -0.00001916, -0.00000890, 0.00000128, -0.00000673, -0.00000490, -0.00000991, 0.00000830, -0.00000323, -0.00000499, -0.00000690, -0.00000388, -0.00000315, 0.00000985, -0.00000676, 0.00000664, -0.00000949, 0.00000761, -0.00001895, 0.00002027, -0.00000880, -0.00000492, 0.00000231, -0.00001660, 0.00000658, -0.00000501, -0.00000033, -0.00000251, -0.00000196, -0.00000709, -0.00000013, -0.00000421, 0.00000166, -0.00000229, -0.00000049, 0.00000035, -0.00000135, -0.00000029, 0.00000124, -0.00001229, -0.00000249, 0.00000146, 0.00000254, -0.00001263, 0.00000230, -0.00000207, 0.00000313, -0.00001662, 0.00000594, -0.00000558, 0.00000870, -0.00001824, 0.00001178, -0.00003378, 0.00001994, -0.00001582, 0.00002088, -0.00000121, -0.00000768, 0.00002463, -0.00000832, 0.00000435, -0.00000133, 0.00000954, 0.00000601, 0.00000355, 0.00000047, 0.00000157, 0.00000169, },
		{ 0.00000873, -0.00000442, 0.00000271, 0.00000098, 0.00000507, 0.00000151, 0.00001004, -0.00000392, 0.00001065, 0.00000813, -0.00000554, 0.00000515, -0.00001525, -0.00000046, -0.00000947, -0.00000105, -0.00000310, 0.00000359, 0.00000238, -0.00000972, 0.00000486, -0.00000788, 0.00000182, -0.00000441, 0.00000818, -0.00001202, 0.00000046, -0.00001323, 0.00001404, -0.00001946, 0.00000723, -0.00001470, 0.00000785, -0.00001529, 0.00000331, -0.00002196, 0.00001991, -0.00003708, 0.00000843, -0.00004422, 0.00001661, -0.00007441, -0.00000022, -0.00022868, -0.00005748, 0.00031234, 0.00008277, 0.00009687, 0.00011834, 0.00006869, 0.00031423, -0.00002297, -0.00025007, 0.00000044, -0.00006924, 0.00001001, -0.00004949, 0.00000632, -0.00002356, 0.00000796, -0.00003441, 0.00001631, -0.00002015, 0.00001067, -0.00002370, 0.00001513, -0.00001032, 0.00000695, -0.00001741, 0.00000478, -0.00001447, 0.00001839, -0.00001289, 0.00001194, -0.00001394, 0.00001295, -0.00001986, 0.00001223, 0.00000194, -0.00001146, 0.00000588, -0.00001585, 0.00000072, 0.00000096, -0.00000503, 0.00000066, -0.00000381, -0.00000398, -0.00000404, -0.00000170, -0.00000162, 0.00000107, -0.00000320, 0.00000207, -0.00000233, -0.00000022, 0.00000135, -0.00000907, -0.00000560, -0.00000023, 0.00000468, -0.00001110, -0.00000187, 0.00000020, 0.00000310, -0.00001363, 0.00000145, -0.00000518, 0.00000739, -0.00001094, 0.00000294, -0.00001646, -0.00000984, 0.00001248, -0.00001311, 0.00003002, -0.00002360, 0.00002418, 0.00000104, -0.00000259, 0.00000336, 0.00000465, 0.00000816, 0.00000542, 0.00000146, -0.00000020, 0.00000411, -0.00000200, },
		{ 0.00000073, -0.00000136, 0.00000384, 0.00000244, 0.00000348, 0.00000690, 0.00000267, 0.00000139, 0.00001815, -0.00000903, 0.00000758, -0.00001485, -0.00000278, -0.00000698, -0.00000417, -0.00000225, 0.00000050, 0.00000640, -0.00001041, 0.00000271, -0.00000486, -0.00000012, -0.00000433, 0.00000652, -0.00000720, -0.00000129, -0.00001248, 0.00000764, -0.00000923, -0.00000204, -0.00000610, -0.00000114, -0.00000577, -0.00000201, -0.00001694, 0.00000816, -0.00001707, -0.00000783, -0.00002523, -0.00000665, -0.00003784, -0.00002820, -0.00012919, -0.00019823, 0.00029453, 0.00012123, 0.00008089, 0.00009901, 0.00008514, 0.00015543, 0.00025697, -0.00025751, -0.00007329, -0.00005178, -0.00000995, -0.00003343, -0.00001131, -0.00001842, 0.00000531, -0.00002584, 0.00000150, -0.00001033, 0.00000369, -0.00001454, 0.00000316, -0.00000283, 0.00000284, -0.00001013, -0.00000229, -0.00001087, 0.00001127, -0.00000402, 0.00000353, -0.00000576, 0.00000440, -0.00000600, -0.00000608, 0.00001654, -0.00001878, 0.00000931, -0.00001455, -0.00000409, 0.00000340, -0.00000498, -0.00000074, -0.00000226, -0.00000395, -0.00000410, -0.00000257, -0.00000235, 0.00000166, -0.00000313, 0.00000143, -0.00000137, -0.00000150, 0.00000212, -0.00000706, -0.00000690, -0.00000167, 0.00000375, -0.00000587, -0.00000825, 0.00000301, 0.00000197, -0.00000870, -0.00000498, 0.00000023, -0.00000255, 0.00000340, -0.00001059, 0.00000256, -0.00002958, 0.00001902, -0.00001566, 0.00001906, 0.00000143, -0.00000387, 0.00002274, -0.00001229, 0.00000705, 0.00000261, 0.00000781, 0.00000619, 0.00000538, -0.00000328, 0.00000624, -0.00000302, 0.00000678, },
		{ -0.00000684, 0.00000797, -0.00000160, 0.00000766, 0.00000077, 0.00001134, -0.00000753, 0.00002424, -0.00000755, 0.00000700, -0.00001232, -0.00000604, -0.00000554, -0.00000450, -0.00000412, -0.00000008, 0.00000584, -0.00000510, -0.00000452, 0.00000160, -0.00000535, 0.00000065, -0.00000096, 0.00000208, -0.00000800, -0.00000348, -0.00000666, 0.00000534, -0.00001442, 0.00000625, -0.00001390, 0.00000677, -0.00001253, -0.00000173, -0.00001431, 0.00000849, -0.00002854, 0.00000126, -0.00003743, 0.00000171, -0.00006420, -0.00003596, -0.00027618, 0.00021244, 0.00017612, 0.00007642, 0.00006989, 0.00011356, 0.00005510, 0.00028239, -0.00001016, -0.00022756, -0.00001014, -0.00005811, 0.00000414, -0.00004383, 0.00000157, -0.00001852, 0.00000272, -0.00002523, 0.00000755, -0.00001355, 0.00000462, -0.00001531, 0.00001072, -0.00001038, 0.00000588, -0.00001514, -0.00000084, -0.00000356, 0.00001136, -0.00001003, 0.00000735, -0.00001000, 0.00001186, -0.00002263, 0.00002548, -0.00002077, 0.00000879, -0.00001098, -0.00000775, 0.00000187, -0.00000064, -0.00000545, 0.00000165, -0.00000658, -0.00000098, -0.00000582, -0.00000078, -0.00000101, 0.00000043, -0.00000256, 0.00000288, -0.00000563, 0.00000517, -0.00000745, -0.00000585, -0.00000415, 0.00000353, -0.00000324, -0.00000920, -0.00000082, 0.00000573, -0.00000743, -0.00000743, 0.00000198, -0.00000636, 0.00000529, -0.00000772, 0.00000075, -0.00002013, -0.00000403, 0.00000638, -0.00000892, 0.00002567, -0.00001416, 0.00001988, 0.00000009, -0.00000487, 0.00001064, 0.00000214, 0.00000880, 0.00000623, -0.00000017, 0.00000125, 0.00000354, -0.00000176, 0.00000936, },
		{ 0.00000695, -0.00000110, 0.00000787, -0.00000097, 0.00001350, -0.00000732, 0.00001920, 0.00000330, 0.00000019, -0.00000428, -0.00001337, -0.00000160, -0.00000731, -0.00000215, -0.00000448, 0.00000724, -0.00000214, -0.00000691, 0.00000168, -0.00000459, 0.00000046, -0.00000253, 0.00000369, -0.00000762, -0.00000088, -0.00001104, 0.00000664, -0.00001374, 0.00000577, -0.00001276, 0.00000470, -0.00001009, 0.00000102, -0.00001921, 0.00001041, -0.00002570, 0.00000406, -0.00003932, 0.00000567, -0.00006320, -0.00000480, -0.00024766, 0.00004984, 0.00026006, 0.00004695, 0.00009298, 0.00006553, 0.00009523, 0.00012019, 0.00024820, -0.00023151, -0.00007226, -0.00005195, -0.00000522, -0.00003664, -0.00000801, -0.00002019, 0.00000588, -0.00002322, 0.00000117, -0.00001246, 0.00000575, -0.00001520, 0.00000863, -0.00001063, 0.00000716, -0.00001244, -0.00000287, -0.00000789, 0.00001465, -0.00001096, 0.00000861, -0.00001285, 0.00001523, -0.00002101, 0.00001682, -0.00000863, -0.00000290, 0.00000062, -0.00001556, 0.00000292, -0.00000013, -0.00000493, 0.00000004, -0.00000446, -0.00000230, -0.00000423, -0.00000304, -0.00000107, 0.00000124, -0.00000349, 0.00000370, -0.00000563, 0.00000357, -0.00000360, -0.00000828, -0.00000353, 0.00000139, -0.00000071, -0.00000861, -0.00000501, 0.00000675, -0.00000300, -0.00001090, 0.00000267, -0.00000549, -0.00000019, 0.00000030, -0.00000471, -0.00000854, -0.00001827, 0.00000882, -0.00000877, 0.00001228, 0.00001076, -0.00000560, 0.00002132, -0.00001597, 0.00001381, 0.00000193, 0.00000799, 0.00000605, 0.00000590, -0.00000488, 0.00000931, -0.00000674, 0.00001174, -0.00000596, },
		{ 0.00000447, 0.00000302, 0.00000340, 0.00000782, 0.00000165, 0.00000625, 0.00001802, -0.00000752, 0.00000477, -0.00002040, 0.00000150, -0.00001059, 0.00000143, -0.00000942, 0.00000785, -0.00000035, -0.00000552, -0.00000227, -0.00000082, -0.00000268, 0.00000006, 0.00000025, -0.00000297, -0.00000343, -0.00000708, -0.00000066, -0.00000631, -0.00000167, -0.00000376, -0.00000490, -0.00000127, -0.00000450, -0.00001093, -0.00000224, -0.00001122, -0.00000622, -0.00002323, -0.00001143, -0.00003960, -0.00001738, -0.00015984, -0.00011622, 0.00029904, 0.00004981, 0.00010732, 0.00002609, 0.00012776, 0.00002924, 0.00026503, -0.00000664, -0.00020017, -0.00002672, -0.00004064, -0.00000894, -0.00003202, -0.00000742, -0.00001088, -0.00000314, -0.00001583, -0.00000341, -0.00000372, -0.00000431, -0.00000206, -0.00000189, -0.00000286, -0.00000010, -0.00001072, -0.00000551, 0.00000836, -0.00000217, 0.00000088, -0.00000588, 0.00000612, -0.00000642, -0.00000128, 0.00000909, -0.00001759, 0.00001252, -0.00002154, 0.00000334, -0.00000189, -0.00000079, -0.00000497, 0.00000043, -0.00000654, 0.00000037, -0.00000717, 0.00000063, -0.00000098, -0.00000056, -0.00000003, -0.00000045, -0.00000310, 0.00000404, -0.00001298, -0.00000119, -0.00000229, 0.00000324, -0.00001012, -0.00000450, 0.00000079, 0.00000642, -0.00001602, 0.00000365, -0.00000483, -0.00000219, -0.00000106, 0.00000145, -0.00001131, -0.00001064, -0.00001033, 0.00000807, -0.00001087, 0.00002865, -0.00000964, 0.00001626, -0.00000301, -0.00000067, 0.00001256, 0.00000186, 0.00000810, 0.00000787, -0.00000325, 0.00000584, -0.00000149, 0.00000433, 0.00000223, 0.00000044, },
		{ -0.00000025, 0.00000713, 0.00000231, 0.00000941, -0.00000290, 0.00002486, -0.00000647, 0.00000644, -0.00001919, -0.00000299, -0.00000811, -0.00000061, -0.00000725, 0.00000147, 0.00000539, -0.00000652, -0.00000259, -0.00000177, -0.00000154, -0.00000097, 0.00000070, -0.00000294, -0.00000244, -0.00000567, -0.00000312, -0.00000523, -0.00000378, -0.00000024, -0.00000743, -0.00000037, -0.00000402, -0.00000733, -0.00000714, -0.00000758, -0.00000818, -0.00001374, -0.00002017, -0.00002514, -0.00003105, -0.00008318, -0.00020467, 0.00023571, 0.00012206, 0.00006876, 0.00005498, 0.00007544, 0.00007403, 0.00011340, 0.00022057, -0.00019012, -0.00008847, -0.00003542, -0.00001664, -0.00002618, -0.00001595, -0.00001221, -0.00000210, -0.00001134, -0.00001068, -0.00000255, -0.00000372, -0.00000252, -0.00000087, -0.00000621, 0.00000262, -0.00000788, -0.00000949, 0.00000528, 0.00000264, -0.00000239, -0.00000203, -0.00000116, 0.00000427, -0.00001164, 0.00001683, -0.00002107, 0.00001206, -0.00001577, -0.00000435, 0.00000123, -0.00000107, -0.00000492, 0.00000007, -0.00000599, 0.00000011, -0.00000598, -0.00000176, -0.00000077, 0.00000032, -0.00000186, 0.00000257, -0.00000644, 0.00000653, -0.00001151, -0.00000409, -0.00000223, 0.00000416, -0.00000916, -0.00000499, -0.00000377, 0.00001045, -0.00001329, -0.00000183, -0.00000065, -0.00000357, -0.00000483, 0.00000640, -0.00001083, -0.00000720, -0.00001562, 0.00000251, -0.00000507, 0.00000966, 0.00001874, -0.00000853, 0.00001777, -0.00001394, 0.00001747, 0.00000161, 0.00000735, 0.00000750, 0.00000312, -0.00000109, 0.00000580, -0.00000283, 0.00000787, -0.00000331, 0.00000799, },
		{ 0.00000651, 0.00000172, 0.00001064, -0.00000330, 0.00002100, 0.00000342, 0.00000156, -0.00001094, -0.00001274, -0.00000216, -0.00000663, -0.00000055, -0.00000748, 0.00001061, -0.00000716, -0.00000134, -0.00000457, 0.00000069, -0.00000304, 0.00000269, -0.00000448, -0.00000088, -0.00000541, -0.00000237, -0.00000661, -0.00000487, 0.00000086, -0.00000596, -0.00000312, -0.00000159, -0.00000639, -0.00000652, -0.00000895, -0.00000873, -0.00000686, -0.00002301, -0.00001717, -0.00004052, -0.00003451, -0.00021586, 0.00011059, 0.00020382, 0.00003355, 0.00009018, 0.00002270, 0.00011462, 0.00002682, 0.00023279, 0.00001269, -0.00019152, -0.00002612, -0.00003883, -0.00000933, -0.00003099, -0.00000620, -0.00001335, 0.00000265, -0.00002006, -0.00000116, -0.00000623, 0.00000028, -0.00000212, -0.00000527, -0.00000176, -0.00000004, -0.00001354, 0.00000203, 0.00000478, -0.00000226, -0.00000087, -0.00000494, 0.00000813, -0.00001257, 0.00001451, -0.00001405, 0.00000146, -0.00000261, -0.00001528, 0.00000589, -0.00000400, -0.00000125, -0.00000389, -0.00000209, -0.00000379, -0.00000120, -0.00000622, 0.00000100, -0.00000131, -0.00000043, 0.00000126, -0.00000393, 0.00000208, -0.00000364, -0.00001105, 0.00000047, 0.00000165, -0.00000377, -0.00000990, -0.00000209, 0.00000491, -0.00000159, -0.00001255, 0.00000603, -0.00000696, -0.00000276, -0.00000060, 0.00000283, -0.00001779, -0.00000166, -0.00001980, 0.00001208, -0.00001410, 0.00003538, -0.00000935, 0.00001229, -0.00000353, 0.00000460, 0.00001187, 0.00000226, 0.00000824, 0.00000675, -0.00000202, 0.00000587, -0.00000182, 0.00000565, -0.00000102, 0.00000655, 0.00000118, },
		{ 0.00000290, 0.00000896, 0.00000008, 0.00001357, 0.00001355, -0.00000181, -0.00000364, -0.00001921, -0.00000054, -0.00000920, 0.00000211, -0.00000987, 0.00000789, -0.00000196, -0.00000407, -0.00000287, -0.00000214, -0.00000057, 0.00000026, -0.00000098, -0.00000479, -0.00000085, -0.00000547, -0.00000209, -0.00001183, 0.00000528, -0.00000747, -0.00000112, -0.00000527, -0.00000039, -0.00001054, -0.00000268, -0.00001852, 0.00000516, -0.00002866, -0.00000589, -0.00005136, -0.00000628, -0.00018589, -0.00000969, 0.00023084, 0.00005073, 0.00008072, 0.00002997, 0.00008018, 0.00005822, 0.00010450, 0.00020088, -0.00015912, -0.00009713, -0.00002482, -0.00002446, -0.00001887, -0.00002001, -0.00000901, -0.00000340, -0.00000734, -0.00001606, 0.00000200, -0.00000724, 0.00000573, -0.00001011, -0.00000090, -0.00000134, -0.00000636, -0.00000876, 0.00001065, -0.00000566, 0.00000391, -0.00001035, 0.00001082, -0.00001163, 0.00001060, -0.00000565, -0.00000843, 0.00000604, -0.00001881, 0.00000371, -0.00000229, -0.00000141, -0.00000395, -0.00000197, -0.00000437, -0.00000014, -0.00000610, -0.00000087, -0.00000084, -0.00000047, 0.00000061, -0.00000153, -0.00000187, 0.00000187, -0.00001337, -0.00000108, 0.00000088, 0.00000090, -0.00001331, -0.00000107, -0.00000068, 0.00000677, -0.00001625, 0.00000556, -0.00000581, -0.00000093, -0.00000860, 0.00001244, -0.00002063, 0.00000234, -0.00002250, 0.00000316, -0.00000711, 0.00001438, 0.00002092, -0.00001055, 0.00001372, -0.00000734, 0.00001657, 0.00000301, 0.00000636, 0.00000810, 0.00000151, 0.00000249, 0.00000174, 0.00000252, 0.00000085, 0.00000503, 0.00000334, 0.00000498, },
		{ 0.00000725, 0.00000344, 0.00000731, 0.00001947, -0.00000071, 0.00000005, -0.00001991, -0.00000439, -0.00000762, -0.00000024, -0.00000607, 0.00000024, 0.00000494, -0.00000745, 0.00000029, -0.00000617, 0.00000229, -0.00000284, 0.00000348, -0.00000890, 0.00000259, -0.00000793, 0.00000279, -0.00001658, 0.00000504, -0.00000583, -0.00000025, -0.00000790, 0.00000315, -0.00001255, 0.00000283, -0.00002454, 0.00000779, -0.00002463, -0.00000397, -0.00004832, -0.00000667, -0.00012878, -0.00010514, 0.00020878, 0.00009126, 0.00006126, 0.00005184, 0.00003991, 0.00008731, 0.00003644, 0.00019659, 0.00003591, -0.00019010, -0.00001794, -0.00004522, -0.00000293, -0.00003487, -0.00000151, -0.00001760, 0.00001008, -0.00002802, 0.00000578, -0.00001300, 0.00001114, -0.00001154, -0.00000040, -0.00000600, 0.00000276, -0.00001658, 0.00001205, -0.00000635, 0.00000630, -0.00001183, 0.00000876, -0.00000717, 0.00000423, 0.00000361, -0.00001537, 0.00000821, -0.00001513, -0.00000337, 0.00000215, -0.00000410, -0.00000131, -0.00000428, -0.00000261, -0.00000207, -0.00000280, -0.00000450, 0.00000070, -0.00000202, 0.00000153, -0.00000124, -0.00000239, 0.00000210, -0.00000960, -0.00000661, 0.00000181, 0.00000272, -0.00001140, -0.00000545, -0.00000026, 0.00000575, -0.00000952, -0.00000288, 0.00000051, -0.00000357, -0.00000785, 0.00000639, -0.00000601, -0.00001163, -0.00000380, -0.00002205, 0.00001009, -0.00001066, 0.00003825, -0.00000919, 0.00000679, 0.00000151, 0.00000577, 0.00001234, 0.00000184, 0.00000924, 0.00000381, 0.00000279, 0.00000118, 0.00000376, -0.00000060, 0.00000486, 0.00000468, 0.00000438, 0.00000373, },
		{ 0.00000765, 0.00000155, 0.00002270, 0.00000217, 0.00000271, -0.00001794, -0.00000920, -0.00000650, -0.00000277, -0.00000268, -0.00000438, 0.00000638, -0.00000577, -0.00000077, -0.00000516, -0.00000031, -0.00000107, 0.00000263, -0.00000623, -0.00000146, -0.00000378, 0.00000037, -0.00001077, -0.00000576, 0.00000304, -0.00000564, -0.00000276, -0.00000263, -0.00000530, -0.00000241, -0.00001483, -0.00000800, -0.00000532, -0.00001742, -0.00002560, -0.00003262, -0.00005998, -0.00017050, 0.00016045, 0.00012106, 0.00006250, 0.00005189, 0.00003952, 0.00006073, 0.00006221, 0.00008274, 0.00019672, -0.00014570, -0.00009014, -0.00002877, -0.00002039, -0.00002129, -0.00001548, -0.00001351, 0.00000392, -0.00001425, -0.00001026, -0.00000402, 0.00000158, 0.00000034, -0.00000885, -0.00000277, 0.00000028, -0.00001005, 0.00000129, 0.00000218, -0.00000043, -0.00000352, -0.00000210, 0.00000389, -0.00000727, 0.00001513, -0.00002052, 0.00000737, -0.00001042, -0.00000826, 0.00000285, -0.00000335, -0.00000199, -0.00000316, -0.00000389, -0.00000164, -0.00000220, -0.00000438, -0.00000102, -0.00000126, 0.00000031, 0.00000049, -0.00000333, 0.00000201, -0.00000618, -0.00000929, -0.00000048, 0.00000454, -0.00000737, -0.00001058, 0.00000063, 0.00000289, -0.00000260, -0.00000844, 0.00000317, -0.00000462, -0.00000432, -0.00000397, 0.00000776, -0.00002007, 0.00000536, -0.00002701, 0.00000206, -0.00000711, 0.00001995, 0.00002024, -0.00001334, 0.00001385, -0.00000225, 0.00001497, 0.00000406, 0.00000677, 0.00000567, 0.00000472, 0.00000047, 0.00000403, 0.00000063, 0.00000127, 0.00000811, 0.00000260, 0.00000595, 0.00000428, },
		{ -0.00000208, 0.00002231, 0.00000710, 0.00000380, -0.00001351, -0.00001392, -0.00000632, -0.00000495, -0.00000098, -0.00000536, 0.00000387, -0.00000183, -0.00000351, -0.00000186, -0.00000467, 0.00000164, -0.00000006, -0.00000083, -0.00000749, 0.00000119, -0.00000413, -0.00000154, -0.00001661, 0.00000840, -0.00000850, 0.00000186, -0.00000848, 0.00000168, -0.00000923, -0.00000153, -0.00002422, 0.00000874, -0.00002626, -0.00000610, -0.00005278, -0.00001544, -0.00018415, 0.00008095, 0.00015059, 0.00006151, 0.00005708, 0.00004132, 0.00003878, 0.00007564, 0.00003415, 0.00017660, 0.00004338, -0.00017633, -0.00002078, -0.00004308, -0.00000383, -0.00003153, -0.00000427, -0.00001271, 0.00000649, -0.00002567, 0.00000331, -0.00000918, 0.00001187, -0.00001503, 0.00000101, -0.00000664, 0.00000049, -0.00000995, 0.00000982, -0.00000746, 0.00000544, -0.00001205, 0.00001236, -0.00001503, 0.00002015, -0.00001747, 0.00000088, -0.00000415, -0.00001167, 0.00000132, -0.00000092, -0.00000426, -0.00000064, -0.00000602, -0.00000057, -0.00000307, -0.00000225, -0.00000379, 0.00000022, -0.00000208, 0.00000299, -0.00000473, 0.00000254, -0.00000461, -0.00000855, -0.00000474, 0.00000529, -0.00000291, -0.00001262, -0.00000235, 0.00000332, -0.00000095, -0.00000742, 0.00000044, -0.00000236, -0.00000334, -0.00000771, 0.00000777, -0.00001192, -0.00000376, -0.00001063, -0.00002033, 0.00000478, -0.00000197, 0.00003619, -0.00000890, 0.00000316, 0.00000826, 0.00000418, 0.00001340, 0.00000220, 0.00000799, 0.00000418, 0.00000446, -0.00000045, 0.00000641, -0.00000496, 0.00001149, 0.00000130, 0.00000825, 0.00000146, 0.00001115, },
		{ 0.00001550, 0.00001612, 0.00000145, -0.00000563, -0.00001944, -0.00000514, -0.00000874, 0.00000107, -0.00000634, 0.00000261, -0.00000078, -0.00000336, -0.00000135, -0.00000498, -0.00000011, 0.00000062, 0.00000093, -0.00000801, -0.00000014, -0.00000350, 0.00000146, -0.00001671, 0.00000202, -0.00000272, -0.00000093, -0.00000603, -0.00000043, -0.00000754, 0.00000133, -0.00002390, 0.00000266, -0.00001758, -0.00000552, -0.00004746, -0.00001181, -0.00014470, -0.00002240, 0.00018192, 0.00004878, 0.00007758, 0.00002691, 0.00004954, 0.00004234, 0.00006529, 0.00006671, 0.00018909, -0.00013133, -0.00008425, -0.00003285, -0.00001649, -0.00002245, -0.00001331, -0.00001382, 0.00000615, -0.00001772, -0.00000686, -0.00000691, 0.00000893, -0.00000760, -0.00000444, -0.00000588, 0.00000071, -0.00000881, 0.00000639, -0.00000524, 0.00000484, -0.00001103, 0.00000909, -0.00001057, 0.00001165, -0.00000270, -0.00001208, 0.00000479, -0.00001633, 0.00000151, -0.00000124, -0.00000330, -0.00000186, -0.00000401, -0.00000323, -0.00000112, -0.00000297, -0.00000292, -0.00000146, -0.00000183, 0.00000224, -0.00000276, 0.00000002, -0.00000091, -0.00000941, -0.00000549, 0.00000147, 0.00000296, -0.00001282, -0.00000601, 0.00000284, 0.00000084, -0.00000694, -0.00000028, -0.00000242, -0.00000178, -0.00000792, 0.00000219, -0.00000210, -0.00001066, -0.00000117, -0.00002438, -0.00000494, -0.00000045, 0.00002089, 0.00001955, -0.00001621, 0.00001816, -0.00000101, 0.00001447, 0.00000488, 0.00000677, 0.00000361, 0.00000869, -0.00000308, 0.00000872, -0.00000531, 0.00000801, 0.00000508, 0.00000675, 0.00000274, 0.00000994, 0.00000024, },
		{ 0.00002503, -0.00000015, 0.00000130, -0.00002213, -0.00000550, -0.00001146, 0.00000121, -0.00000618, 0.00000155, -0.00000055, -0.00000278, -0.00000176, -0.00000334, -0.00000322, 0.00000164, 0.00000113, -0.00000537, -0.00000414, -0.00000123, 0.00000070, -0.00001043, -0.00000804, 0.00000308, -0.00000428, -0.00000151, -0.00000493, -0.00000341, -0.00000132, -0.00001440, -0.00001078, -0.00000587, -0.00001100, -0.00003093, -0.00002647, -0.00008913, -0.00010786, 0.00018034, 0.00005770, 0.00007760, 0.00003074, 0.00005107, 0.00002338, 0.00007564, 0.00002501, 0.00016589, 0.00004159, -0.00015512, -0.00003134, -0.00003465, -0.00000934, -0.00002477, -0.00000921, -0.00000536, -0.00000179, -0.00001791, -0.00000401, 0.00000157, 0.00000290, -0.00001044, -0.00000261, -0.00000451, -0.00000256, -0.00000014, -0.00000015, -0.00000014, -0.00000443, 0.00000023, -0.00000058, -0.00000201, 0.00001309, -0.00002121, 0.00000899, -0.00001726, -0.00000031, -0.00000128, -0.00000167, -0.00000437, -0.00000075, -0.00000645, 0.00000096, -0.00000445, -0.00000076, -0.00000382, -0.00000060, -0.00000053, 0.00000120, -0.00000414, 0.00000367, -0.00001096, -0.00000421, -0.00000370, 0.00000690, -0.00001064, -0.00000842, -0.00000071, 0.00000480, -0.00000880, 0.00000139, -0.00000389, -0.00000018, -0.00000744, -0.00000099, -0.00000060, -0.00000663, -0.00000526, -0.00001167, -0.00002305, 0.00000405, 0.00000474, 0.00003286, -0.00000957, 0.00000442, 0.00001189, 0.00000312, 0.00001377, 0.00000358, 0.00000474, 0.00000822, 0.00000148, 0.00000353, 0.00000202, -0.00000107, 0.00001123, 0.00000343, 0.00000683, 0.00000537, 0.00000651, 0.00000564, },
		{ 0.00000521, 0.00000201, -0.00001809, -0.00001057, -0.00000963, -0.00000352, -0.00000247, -0.00000240, 0.00000259, -0.00000506, -0.00000014, -0.00000350, -0.00000318, -0.00000089, 0.00000321, -0.00000412, -0.00000516, -0.00000250, 0.00000142, -0.00000579, -0.00001206, 0.00000124, -0.00000243, -0.00000128, -0.00000451, -0.00000404, -0.00000129, -0.00000801, -0.00001614, -0.00000478, -0.00001032, -0.00001924, -0.00003562, -0.00004971, -0.00014519, 0.00012624, 0.00010243, 0.00004659, 0.00006035, 0.00002950, 0.00004019, 0.00003993, 0.00005654, 0.00006450, 0.00016970, -0.00010704, -0.00008869, -0.00002843, -0.00001960, -0.00001721, -0.00001722, -0.00000684, -0.00000110, -0.00001175, -0.00001224, 0.00000078, 0.00000472, -0.00000742, -0.00000504, -0.00000506, -0.00000277, 0.00000024, -0.00000017, -0.00000158, -0.00000123, -0.00000453, 0.00000465, -0.00000916, 0.00001801, -0.00001758, 0.00000285, -0.00001018, -0.00000798, 0.00000263, -0.00000328, -0.00000355, -0.00000146, -0.00000514, -0.00000126, -0.00000251, -0.00000184, -0.00000258, -0.00000205, -0.00000129, 0.00000237, -0.00000474, 0.00000381, -0.00000800, -0.00000653, -0.00000418, 0.00000458, -0.00000438, -0.00001122, -0.00000346, 0.00000552, -0.00000658, -0.00000170, -0.00000020, -0.00000323, -0.00000378, -0.00000428, -0.00000055, -0.00000407, -0.00000555, -0.00000698, -0.00002183, -0.00001157, 0.00000804, 0.00001865, 0.00001824, -0.00001551, 0.00002187, -0.00000097, 0.00001381, 0.00000690, 0.00000416, 0.00000572, 0.00000802, -0.00000156, 0.00000774, -0.00000567, 0.00001140, 0.00000421, 0.00000846, 0.00000312, 0.00001023, 0.00000028, 0.00002649, },
		{ -0.00000075, -0.00001080, -0.00001704, -0.00000610, -0.00000970, 0.00000136, -0.00000640, 0.00000558, -0.00000657, 0.00000022, -0.00000324, -0.00000227, -0.00000362, 0.00000390, -0.00000222, -0.00000511, -0.00000429, 0.00000106, -0.00000191, -0.00001202, -0.00000306, -0.00000106, -0.00000113, -0.00000308, -0.00000505, -0.00000193, -0.00000335, -0.00001617, -0.00000737, -0.00001052, -0.00000930, -0.00003895, -0.00002755, -0.00014311, 0.00004441, 0.00014855, 0.00002405, 0.00007533, 0.00002197, 0.00004993, 0.00001725, 0.00006860, 0.00002523, 0.00014801, 0.00004680, -0.00014176, -0.00003605, -0.00003118, -0.00000970, -0.00002368, -0.00000689, -0.00000675, -0.00000233, -0.00001703, -0.00000322, 0.00000495, -0.00000295, -0.00000697, -0.00000453, -0.00000563, 0.00000177, 0.00000027, -0.00000299, 0.00000046, -0.00000645, 0.00000582, -0.00000997, 0.00001375, -0.00000628, -0.00000700, -0.00000116, -0.00001599, 0.00000563, -0.00000444, -0.00000205, -0.00000379, -0.00000205, -0.00000468, 0.00000009, -0.00000393, -0.00000015, -0.00000393, -0.00000113, 0.00000106, -0.00000213, 0.00000020, -0.00000176, -0.00001102, -0.00000190, -0.00000055, 0.00000206, -0.00001294, -0.00000435, 0.00000166, 0.00000020, -0.00000875, 0.00000600, -0.00000721, 0.00000005, -0.00000763, 0.00000197, -0.00000799, 0.00000164, -0.00001208, -0.00000814, -0.00002950, 0.00000844, 0.00000676, 0.00002972, -0.00000905, 0.00000816, 0.00001253, 0.00000270, 0.00001462, 0.00000334, 0.00000391, 0.00001030, 0.00000014, 0.00000617, -0.00000212, 0.00000504, 0.00000819, 0.00000756, 0.00000438, 0.00000959, 0.00000128, 0.00002016, 0.00001584, },
		{ -0.00000666, -0.00001953, -0.00000588, -0.00001198, 0.00000046, -0.00000614, 0.00000393, -0.00000314, -0.00000363, -0.00000034, -0.00000346, -0.00000288, 0.00000039, 0.00000222, -0.00000661, -0.00000319, -0.00000286, 0.00000307, -0.00001201, -0.00000414, -0.00000411, 0.00000136, -0.00000379, -0.00000268, -0.00000611, 0.00000238, -0.00001653, -0.00000601, -0.00001614, 0.00000022, -0.00003834, -0.00001714, -0.00011954, -0.00003335, 0.00016210, 0.00003771, 0.00005736, 0.00004025, 0.00003710, 0.00002782, 0.00003999, 0.00004874, 0.00006247, 0.00015160, -0.00008448, -0.00009363, -0.00002353, -0.00002217, -0.00001350, -0.00001821, -0.00000325, -0.00000674, -0.00000654, -0.00001610, 0.00000879, -0.00000364, -0.00000300, -0.00000773, -0.00000440, -0.00000287, 0.00000620, -0.00000727, 0.00000338, -0.00000875, 0.00000645, -0.00000873, 0.00000713, 0.00000354, -0.00001201, 0.00000265, -0.00001770, 0.00000206, -0.00000085, -0.00000369, -0.00000345, -0.00000207, -0.00000442, -0.00000091, -0.00000304, -0.00000073, -0.00000258, -0.00000286, 0.00000044, -0.00000046, -0.00000210, 0.00000186, -0.00001157, -0.00000308, -0.00000202, 0.00000343, -0.00001052, -0.00000627, -0.00000163, 0.00000421, -0.00001091, 0.00000533, -0.00000403, -0.00000196, -0.00000661, 0.00000180, -0.00000909, 0.00000262, -0.00000940, -0.00000610, -0.00002467, -0.00001218, 0.00001227, 0.00001638, 0.00001653, -0.00001060, 0.00002231, -0.00000032, 0.00001301, 0.00000875, 0.00000147, 0.00000924, 0.00000509, 0.00000300, 0.00000227, 0.00000041, 0.00000880, 0.00000817, 0.00000593, 0.00000798, 0.00000438, 0.00001227, 0.00002417, 0.00000026, },
		{ -0.00001863, -0.00000853, -0.00001076, -0.00000385, -0.00000343, -0.00000040, 0.00000221, -0.00000755, 0.00000137, -0.00000406, -0.00000139, -0.00000332, 0.00000488, -0.00000631, -0.00000249, -0.00000576, 0.00000468, -0.00000922, -0.00000543, -0.00000572, 0.00000033, -0.00000216, -0.00000161, -0.00000793, 0.00000352, -0.00001301, -0.00000644, -0.00001726, -0.00000089, -0.00002688, -0.00002108, -0.00008163, -0.00009474, 0.00014149, 0.00007218, 0.00003436, 0.00005279, 0.00003119, 0.00003616, 0.00002106, 0.00005565, 0.00003195, 0.00012539, 0.00005753, -0.00013585, -0.00003472, -0.00003207, -0.00000670, -0.00002513, -0.00000154, -0.00001327, 0.00000269, -0.00002114, 0.00000470, -0.00000122, -0.00000130, -0.00000794, -0.00000392, -0.00000768, 0.00000941, -0.00000667, 0.00000187, -0.00000724, 0.00000323, -0.00000369, -0.00000123, 0.00001066, -0.00001274, 0.00000279, -0.00001483, -0.00000517, 0.00000376, -0.00000531, -0.00000258, -0.00000343, -0.00000263, -0.00000303, -0.00000140, -0.00000236, -0.00000032, -0.00000430, -0.00000031, 0.00000001, -0.00000233, 0.00000201, -0.00000800, -0.00000720, -0.00000084, 0.00000171, -0.00000620, -0.00000893, -0.00000228, 0.00000277, -0.00000630, -0.00000208, 0.00000368, -0.00000625, -0.00000357, -0.00000160, -0.00000399, -0.00000565, 0.00000203, -0.00001426, -0.00000798, -0.00003235, 0.00001163, 0.00000680, 0.00002628, -0.00000476, 0.00001022, 0.00001271, 0.00000207, 0.00001628, 0.00000122, 0.00000666, 0.00000807, 0.00000376, 0.00000276, 0.00000068, 0.00000570, 0.00000935, 0.00000767, 0.00000675, 0.00000711, 0.00000662, 0.00002668, 0.00000662, -0.00000574, },
		{ -0.00001105, -0.00000930, -0.00000746, -0.00000235, -0.00000316, 0.00000444, -0.00000679, -0.00000128, -0.00000188, -0.00000208, -0.00000340, 0.00000262, -0.00000211, -0.00000480, -0.00000426, 0.00000057, -0.00000239, -0.00000951, -0.00000246, -0.00000580, 0.00000230, -0.00000355, -0.00000467, -0.00000186, -0.00000417, -0.00001210, -0.00000940, -0.00001319, -0.00000828, -0.00003265, -0.00004197, -0.00013022, 0.00009452, 0.00010551, 0.00002905, 0.00004511, 0.00004026, 0.00003098, 0.00002641, 0.00003223, 0.00004998, 0.00005240, 0.00014321, -0.00007216, -0.00009119, -0.00002436, -0.00001940, -0.00001511, -0.00001272, -0.00000803, -0.00000430, -0.00000877, -0.00001076, 0.00000701, -0.00000586, -0.00000186, -0.00000785, -0.00000652, 0.00000343, 0.00000251, -0.00000553, -0.00000083, -0.00000464, 0.00000486, -0.00001023, 0.00001531, -0.00001156, 0.00000211, -0.00001079, -0.00001117, 0.00000438, -0.00000314, -0.00000417, -0.00000289, -0.00000307, -0.00000250, -0.00000231, -0.00000187, -0.00000060, -0.00000268, -0.00000290, 0.00000101, -0.00000283, 0.00000181, -0.00000466, -0.00000924, -0.00000194, 0.00000167, -0.00000384, -0.00000992, -0.00000274, 0.00000028, -0.00000170, -0.00000730, 0.00000621, -0.00000471, -0.00000430, -0.00000344, 0.00000070, -0.00001181, 0.00000676, -0.00001383, -0.00000410, -0.00002832, -0.00000975, 0.00001208, 0.00001460, 0.00001703, -0.00000505, 0.00002102, 0.00000032, 0.00001326, 0.00000828, 0.00000240, 0.00000914, 0.00000587, 0.00000348, 0.00000064, 0.00000482, 0.00000687, 0.00001122, 0.00000526, 0.00000971, 0.00000311, 0.00002500, 0.00001445, -0.00000367, -0.00001799, },
		{ -0.00000769, -0.00000985, -0.00000281, -0.00000455, 0.00000386, -0.00000320, -0.00000487, -0.00000025, -0.00000272, -0.00000213, -0.00000090, 0.00000173, -0.00000646, -0.00000219, -0.00000389, 0.00000261, -0.00001142, -0.00000013, -0.00000925, 0.00000237, -0.00000300, -0.00000209, -0.00000636, 0.00000194, -0.00001522, -0.00000276, -0.00002152, 0.00000057, -0.00003462, -0.00001995, -0.00012841, 0.00002552, 0.00013108, 0.00003523, 0.00003690, 0.00004033, 0.00003473, 0.00002737, 0.00002092, 0.00004908, 0.00003258, 0.00011035, 0.00006202, -0.00012694, -0.00003591, -0.00002972, -0.00000767, -0.00002122, -0.00000280, -0.00001429, 0.00000302, -0.00001935, 0.00000770, -0.00000666, 0.00000121, -0.00000885, -0.00000498, -0.00000433, 0.00001013, -0.00000891, 0.00000221, -0.00000945, 0.00000905, -0.00001278, 0.00001279, -0.00000562, -0.00000196, -0.00000412, -0.00001617, 0.00000224, 0.00000004, -0.00000537, -0.00000258, -0.00000389, -0.00000153, -0.00000330, -0.00000138, -0.00000148, -0.00000055, -0.00000463, 0.00000089, -0.00000285, 0.00000120, -0.00000217, -0.00000893, -0.00000470, 0.00000207, -0.00000191, -0.00001031, -0.00000369, -0.00000068, -0.00000051, -0.00000719, 0.00000243, 0.00000095, -0.00000618, -0.00000444, 0.00000221, -0.00001056, 0.00000069, -0.00000347, -0.00001131, -0.00001217, -0.00002944, 0.00000972, 0.00000706, 0.00002389, 0.00000190, 0.00000996, 0.00001296, 0.00000215, 0.00001647, 0.00000094, 0.00000845, 0.00000632, 0.00000715, -0.00000147, 0.00000655, 0.00000296, 0.00001322, 0.00000584, 0.00001066, 0.00000262, 0.00002011, 0.00002159, 0.00000060, -0.00001712, -0.00001417, },
		{ -0.00001176, -0.00000320, -0.00000622, 0.00000272, -0.00000045, -0.00000548, -0.00000180, -0.00000151, -0.00000210, -0.00000201, 0.00000179, -0.00000476, -0.00000287, -0.00000438, 0.00000195, -0.00000812, -0.00000273, -0.00000588, -0.00000382, 0.00000094, -0.00000268, -0.00000581, 0.00000103, -0.00001128, -0.00000419, -0.00001767, -0.00000569, -0.00002363, -0.00001826, -0.00009697, -0.00004772, 0.00013897, 0.00004751, 0.00003845, 0.00002702, 0.00004690, 0.00002052, 0.00002812, 0.00002458, 0.00005138, 0.00004289, 0.00013658, -0.00006352, -0.00008656, -0.00002577, -0.00001703, -0.00001527, -0.00000899, -0.00001285, -0.00000106, -0.00001103, -0.00000347, -0.00000026, -0.00000299, -0.00000345, -0.00000736, -0.00000647, 0.00000811, -0.00000283, -0.00000172, -0.00000735, 0.00000494, -0.00000631, 0.00000257, 0.00000493, -0.00000941, 0.00000431, -0.00001963, -0.00000081, 0.00000107, -0.00000355, -0.00000454, -0.00000277, -0.00000286, -0.00000173, -0.00000315, -0.00000058, -0.00000080, -0.00000307, -0.00000178, -0.00000108, -0.00000125, 0.00000110, -0.00000889, -0.00000560, -0.00000041, 0.00000204, -0.00001138, -0.00000464, -0.00000103, -0.00000031, -0.00000610, -0.00000102, 0.00000302, -0.00000339, -0.00000758, 0.00000254, -0.00000636, -0.00000639, 0.00000283, -0.00001232, -0.00000682, -0.00002672, -0.00001038, 0.00001041, 0.00001344, 0.00002065, -0.00000154, 0.00001961, 0.00000106, 0.00001347, 0.00000733, 0.00000527, 0.00000645, 0.00000974, -0.00000038, 0.00000502, 0.00000386, 0.00000930, 0.00001040, 0.00000875, 0.00000590, 0.00001253, 0.00002761, 0.00000551, -0.00001355, -0.00001933, -0.00000556, },
		{ -0.00000476, -0.00000671, 0.00000011, 0.00000172, -0.00000446, -0.00000368, -0.00000123, -0.00000167, -0.00000218, 0.00000055, -0.00000263, -0.00000393, -0.00000333, -0.00000065, -0.00000395, -0.00000616, -0.00000121, -0.00000877, 0.00000168, -0.00000203, -0.00000432, -0.00000170, -0.00000599, -0.00000796, -0.00000985, -0.00001436, -0.00001350, -0.00002161, -0.00006078, -0.00009714, 0.00011409, 0.00007170, 0.00003788, 0.00002243, 0.00004310, 0.00002756, 0.00002685, 0.00001560, 0.00004844, 0.00002752, 0.00010291, 0.00005926, -0.00011361, -0.00004013, -0.00002476, -0.00001095, -0.00001437, -0.00000909, -0.00000997, -0.00000136, -0.00001079, 0.00000186, -0.00000565, 0.00000028, -0.00000794, -0.00000666, 0.00000198, 0.00000441, -0.00000468, -0.00000462, -0.00000122, 0.00000166, -0.00000689, 0.00001218, -0.00001305, 0.00000753, -0.00001670, -0.00000687, 0.00000182, -0.00000168, -0.00000564, -0.00000232, -0.00000406, -0.00000035, -0.00000427, -0.00000031, -0.00000134, -0.00000097, -0.00000393, 0.00000009, -0.00000349, 0.00000330, -0.00000801, -0.00000574, -0.00000361, 0.00000480, -0.00000990, -0.00000759, -0.00000068, 0.00000009, -0.00000577, -0.00000230, 0.00000139, 0.00000120, -0.00000901, 0.00000013, -0.00000170, -0.00000920, 0.00000102, -0.00000583, -0.00001036, -0.00001429, -0.00002590, 0.00000428, 0.00000773, 0.00002454, 0.00000760, 0.00000904, 0.00001292, 0.00000320, 0.00001467, 0.00000410, 0.00000652, 0.00000863, 0.00000556, -0.00000006, 0.00000834, 0.00000320, 0.00001429, 0.00000745, 0.00000979, 0.00000602, 0.00002975, 0.00001194, -0.00000851, -0.00002359, -0.00000610, -0.00001120, },
		{ -0.00000489, -0.00000454, 0.00000402, -0.00000398, -0.00000324, -0.00000340, -0.00000009, -0.00000283, 0.00000043, -0.00000237, -0.00000349, -0.00000329, -0.00000162, -0.00000247, -0.00000694, -0.00000003, -0.00000779, -0.00000280, 0.00000065, -0.00000399, -0.00000250, -0.00000371, -0.00000884, -0.00000589, -0.00001601, -0.00001169, -0.00002147, -0.00003494, -0.00011169, 0.00005982, 0.00010161, 0.00003236, 0.00003277, 0.00002325, 0.00004399, 0.00001762, 0.00002426, 0.00002397, 0.00004596, 0.00004111, 0.00012418, -0.00005157, -0.00008513, -0.00002353, -0.00001865, -0.00001028, -0.00001218, -0.00001124, -0.00000350, -0.00000589, -0.00000390, -0.00000358, -0.00000146, -0.00000401, -0.00000836, -0.00000162, 0.00000535, -0.00000166, -0.00000494, -0.00000459, 0.00000495, -0.00000939, 0.00001112, -0.00000876, 0.00000249, -0.00000687, -0.00001517, 0.00000308, -0.00000200, -0.00000378, -0.00000433, -0.00000301, -0.00000169, -0.00000243, -0.00000251, -0.00000017, -0.00000102, -0.00000305, -0.00000127, -0.00000331, 0.00000209, -0.00000421, -0.00000772, -0.00000395, 0.00000285, -0.00000382, -0.00001270, -0.00000001, 0.00000002, -0.00000420, -0.00000462, 0.00000120, 0.00000140, -0.00000451, -0.00000599, 0.00000293, -0.00000951, -0.00000197, -0.00000269, -0.00000882, -0.00001135, -0.00002114, -0.00001562, 0.00000872, 0.00001455, 0.00002502, 0.00000044, 0.00001804, 0.00000284, 0.00001169, 0.00000926, 0.00000563, 0.00000624, 0.00001080, -0.00000180, 0.00000905, 0.00000244, 0.00001201, 0.00001000, 0.00001101, 0.00000432, 0.00002564, 0.00002097, -0.00000428, -0.00002347, -0.00001165, -0.00000745, -0.00000816, },
		{ -0.00000811, 0.00000397, -0.00000263, -0.00000251, -0.00000483, -0.00000029, -0.00000219, -0.00000011, -0.00000181, -0.00000379, -0.00000270, -0.00000232, -0.00000164, -0.00000715, -0.00000048, -0.00000456, -0.00000663, 0.00000035, -0.00000203, -0.00000345, -0.00000202, -0.00000902, -0.00000409, -0.00001413, -0.00001184, -0.00002223, -0.00001835, -0.00010159, -0.00000123, 0.00011712, 0.00003531, 0.00004070, 0.00001256, 0.00004418, 0.00002202, 0.00002494, 0.00001386, 0.00004452, 0.00002605, 0.00009397, 0.00005716, -0.00010213, -0.00004173, -0.00002293, -0.00001002, -0.00001269, -0.00001161, -0.00000857, -0.00000160, -0.00000525, -0.00000505, -0.00000239, -0.00000164, -0.00000764, -0.00000452, 0.00000340, 0.00000157, -0.00000344, -0.00000768, 0.00000528, -0.00000753, 0.00000571, -0.00000137, -0.00000484, 0.00000246, -0.00001931, 0.00000114, -0.00000171, -0.00000219, -0.00000560, -0.00000238, -0.00000317, -0.00000059, -0.00000410, 0.00000026, -0.00000128, -0.00000149, -0.00000269, -0.00000254, -0.00000067, 0.00000019, -0.00000959, -0.00000323, -0.00000093, 0.00000229, -0.00001505, -0.00000221, 0.00000053, -0.00000175, -0.00000751, 0.00000199, -0.00000088, 0.00000104, -0.00000967, 0.00000346, -0.00000719, -0.00000357, -0.00000427, -0.00000329, -0.00001368, -0.00001098, -0.00002641, -0.00000198, 0.00000951, 0.00002764, 0.00001118, 0.00000812, 0.00001346, 0.00000326, 0.00001397, 0.00000718, 0.00000431, 0.00001141, 0.00000248, 0.00000488, 0.00000650, 0.00000651, 0.00001293, 0.00001130, 0.00000622, 0.00001804, 0.00002930, 0.00000087, -0.00002042, -0.00001889, -0.00000463, -0.00001031, -0.00000337, },
		{ 0.00000016, 0.00000049, -0.00000333, -0.00000345, -0.00000321, -0.00000007, -0.00000173, 0.00000023, -0.00000559, -0.00000129, -0.00000352, -0.00000010, -0.00000795, -0.00000107, -0.00000232, -0.00000652, -0.00000357, 0.00000076, -0.00000457, -0.00000002, -0.00000952, -0.00000283, -0.00001231, -0.00000989, -0.00002438, -0.00001067, -0.00007800, -0.00005118, 0.00010803, 0.00005155, 0.00003758, 0.00001990, 0.00002692, 0.00003595, 0.00001842, 0.00001930, 0.00002482, 0.00003938, 0.00004181, 0.00010994, -0.00003921, -0.00008372, -0.00002159, -0.00001930, -0.00000630, -0.00001636, -0.00000804, -0.00000662, 0.00000184, -0.00001029, -0.00000174, -0.00000257, -0.00000382, -0.00000785, 0.00000242, 0.00000065, 0.00000155, -0.00001060, 0.00000387, -0.00000438, 0.00000037, 0.00000408, -0.00000875, 0.00000567, -0.00001613, -0.00000505, 0.00000062, -0.00000308, -0.00000415, -0.00000393, -0.00000269, -0.00000122, -0.00000288, -0.00000161, -0.00000015, -0.00000147, -0.00000206, -0.00000295, -0.00000239, 0.00000195, -0.00000832, -0.00000468, -0.00000228, 0.00000373, -0.00001139, -0.00000811, 0.00000182, -0.00000021, -0.00000762, -0.00000033, -0.00000013, 0.00000117, -0.00000629, -0.00000156, -0.00000234, -0.00000558, -0.00000433, -0.00000318, -0.00000930, -0.00001166, -0.00001726, -0.00002302, 0.00000738, 0.00001869, 0.00002790, 0.00000192, 0.00001676, 0.00000467, 0.00000924, 0.00001323, 0.00000356, 0.00000885, 0.00000791, 0.00000208, 0.00000860, 0.00000465, 0.00001157, 0.00001300, 0.00000885, 0.00001161, 0.00003245, 0.00000966, -0.00001744, -0.00002338, -0.00000645, -0.00000912, -0.00000423, -0.00000819, },
		{ 0.00000255, -0.00000372, -0.00000176, -0.00000533, 0.00000030, -0.00000213, 0.00000141, -0.00000587, -0.00000197, -0.00000297, -0.00000004, -0.00000675, -0.00000379, 0.00000019, -0.00000542, -0.00000590, -0.00000018, -0.00000296, -0.00000031, -0.00000740, -0.00000460, -0.00000852, -0.00000975, -0.00002136, -0.00001489, -0.00004751, -0.00008385, 0.00007952, 0.00007131, 0.00003239, 0.00003064, 0.00001460, 0.00003715, 0.00002252, 0.00001929, 0.00001651, 0.00003660, 0.00002929, 0.00008145, 0.00005789, -0.00009352, -0.00004071, -0.00002344, -0.00000611, -0.00001601, -0.00000916, -0.00001128, 0.00000436, -0.00000775, -0.00000664, -0.00000147, -0.00000247, -0.00000761, 0.00000002, -0.00000037, 0.00000374, -0.00000811, -0.00000170, 0.00000071, -0.00000497, 0.00000765, -0.00000955, 0.00000444, -0.00000969, -0.00001090, 0.00000178, -0.00000381, -0.00000286, -0.00000495, -0.00000243, -0.00000218, -0.00000145, -0.00000301, 0.00000016, -0.00000135, -0.00000128, -0.00000286, -0.00000371, 0.00000158, -0.00000512, -0.00000703, -0.00000218, 0.00000220, -0.00000540, -0.00001306, 0.00000098, 0.00000087, -0.00000504, -0.00000486, 0.00000241, -0.00000154, -0.00000112, -0.00000567, 0.00000030, -0.00000655, -0.00000238, -0.00000741, -0.00000187, -0.00001607, -0.00000598, -0.00003068, -0.00000794, 0.00001351, 0.00003093, 0.00001336, 0.00000738, 0.00001472, 0.00000166, 0.00001614, 0.00000736, 0.00000521, 0.00001035, 0.00000360, 0.00000718, 0.00000642, 0.00000831, 0.00001352, 0.00001231, 0.00000793, 0.00003015, 0.00002049, -0.00001354, -0.00002539, -0.00001111, -0.00000733, -0.00000554, -0.00000656, -0.00000450, },
		{ -0.00000227, -0.00000182, -0.00000454, -0.00000207, -0.00000087, 0.00000067, -0.00000336, -0.00000521, -0.00000079, -0.00000170, -0.00000324, -0.00000809, 0.00000251, -0.00000475, -0.00000494, -0.00000430, -0.00000023, -0.00000207, -0.00000313, -0.00000849, -0.00000344, -0.00001204, -0.00001327, -0.00002485, -0.00002189, -0.00009411, 0.00003796, 0.00008660, 0.00003200, 0.00003474, 0.00001616, 0.00002501, 0.00003225, 0.00001609, 0.00001873, 0.00002153, 0.00003753, 0.00003873, 0.00009939, -0.00003049, -0.00007905, -0.00002302, -0.00001557, -0.00000825, -0.00001548, -0.00000903, -0.00000371, 0.00000343, -0.00001446, 0.00000024, -0.00000376, -0.00000414, -0.00000400, 0.00000104, 0.00000052, -0.00000059, -0.00000891, 0.00000532, -0.00000874, 0.00000876, -0.00000782, 0.00000116, -0.00000381, -0.00001345, 0.00000021, -0.00000273, -0.00000362, -0.00000398, -0.00000364, -0.00000187, -0.00000177, -0.00000222, -0.00000137, -0.00000017, -0.00000171, -0.00000134, -0.00000510, 0.00000046, -0.00000250, -0.00000772, -0.00000319, 0.00000115, -0.00000149, -0.00001356, -0.00000300, 0.00000183, -0.00000208, -0.00000738, 0.00000190, -0.00000126, -0.00000059, -0.00000415, -0.00000131, -0.00000533, -0.00000162, -0.00000876, -0.00000112, -0.00001239, -0.00000745, -0.00001793, -0.00003023, 0.00000711, 0.00002448, 0.00002906, 0.00000316, 0.00001661, 0.00000475, 0.00000913, 0.00001547, 0.00000307, 0.00000972, 0.00000630, 0.00000659, 0.00000697, 0.00000797, 0.00001053, 0.00001613, 0.00000723, 0.00002470, 0.00002952, -0.00000610, -0.00002697, -0.00001542, -0.00000783, -0.00000569, -0.00000527, -0.00000714, 0.00000147, },
		{ -0.00000216, -0.00000295, -0.00000410, -0.00000067, 0.00000003, -0.00000055, -0.00000748, -0.00000048, -0.00000213, -0.00000077, -0.00000990, 0.00000129, -0.00000227, -0.00000419, -0.00000619, -0.00000110, -0.00000179, -0.00000060, -0.00000973, -0.00000187, -0.00001179, -0.00000791, -0.00002829, -0.00001284, -0.00008047, -0.00001018, 0.00009403, 0.00003584, 0.00003490, 0.00002164, 0.00001737, 0.00002992, 0.00002296, 0.00001560, 0.00001728, 0.00003119, 0.00003081, 0.00007068, 0.00005731, -0.00008378, -0.00004151, -0.00002103, -0.00000586, -0.00001729, -0.00000806, -0.00001122, 0.00000884, -0.00001285, -0.00000484, -0.00000237, -0.00000298, -0.00000518, 0.00000135, -0.00000256, 0.00000395, -0.00001028, 0.00000411, -0.00000714, 0.00000479, -0.00000222, -0.00000427, 0.00000157, -0.00001424, -0.00000182, -0.00000172, -0.00000440, -0.00000344, -0.00000407, -0.00000209, -0.00000206, -0.00000154, -0.00000221, -0.00000004, -0.00000159, -0.00000050, -0.00000469, -0.00000192, -0.00000047, -0.00000725, -0.00000452, 0.00000012, 0.00000089, -0.00001169, -0.00000689, 0.00000099, 0.00000034, -0.00000664, -0.00000167, 0.00000121, -0.00000251, -0.00000143, -0.00000219, -0.00000498, -0.00000181, -0.00000634, -0.00000583, -0.00000467, -0.00001340, -0.00000436, -0.00003589, -0.00001313, 0.00001942, 0.00003323, 0.00001458, 0.00000780, 0.00001494, 0.00000092, 0.00001859, 0.00000690, 0.00000694, 0.00000758, 0.00000811, 0.00000621, 0.00000922, 0.00000773, 0.00001641, 0.00001066, 0.00001784, 0.00003487, 0.00000460, -0.00002715, -0.00001967, -0.00000937, -0.00000652, -0.00000364, -0.00000814, -0.00000109, -0.00000093, },
		{ -0.00000229, -0.00000403, -0.00000226, -0.00000001, 0.00000098, -0.00000681, -0.00000311, -0.00000077, -0.00000050, -0.00000795, -0.00000343, 0.00000149, -0.00000466, -0.00000461, -0.00000489, -0.00000054, -0.00000011, -0.00000740, -0.00000433, -0.00000804, -0.00000798, -0.00002254, -0.00001759, -0.00005509, -0.00005067, 0.00008653, 0.00004575, 0.00003256, 0.00002511, 0.00001921, 0.00001853, 0.00003177, 0.00001261, 0.00001948, 0.00001743, 0.00003732, 0.00003396, 0.00009097, -0.00002327, -0.00007404, -0.00002411, -0.00001198, -0.00001209, -0.00001244, -0.00001112, 0.00000238, -0.00000133, -0.00001354, -0.00000020, -0.00000413, -0.00000343, -0.00000057, -0.00000210, 0.00000206, -0.00000502, -0.00000171, -0.00000183, -0.00000222, 0.00000476, -0.00000939, 0.00000512, -0.00001351, -0.00000381, -0.00000120, -0.00000421, -0.00000444, -0.00000309, -0.00000330, -0.00000147, -0.00000219, -0.00000131, -0.00000147, -0.00000043, -0.00000118, -0.00000230, -0.00000487, 0.00000084, -0.00000660, -0.00000491, -0.00000193, 0.00000304, -0.00000965, -0.00000885, -0.00000137, 0.00000134, -0.00000394, -0.00000480, 0.00000164, -0.00000200, -0.00000238, 0.00000044, -0.00000593, -0.00000265, -0.00000310, -0.00000909, -0.00000213, -0.00001247, -0.00000368, -0.00002151, -0.00003625, 0.00000818, 0.00003033, 0.00002885, 0.00000501, 0.00001661, 0.00000405, 0.00001104, 0.00001550, 0.00000497, 0.00000738, 0.00000889, 0.00000773, 0.00000809, 0.00000904, 0.00001187, 0.00001624, 0.00001237, 0.00003568, 0.00001661, -0.00002382, -0.00002509, -0.00001028, -0.00000914, -0.00000184, -0.00000800, -0.00000338, -0.00000115, -0.00000176, },
		{ -0.00000349, -0.00000322, -0.00000122, 0.00000208, -0.00000468, -0.00000590, -0.00000042, -0.00000027, -0.00000508, -0.00000749, 0.00000273, -0.00000365, -0.00000324, -0.00000654, -0.00000216, 0.00000081, -0.00000456, -0.00000666, -0.00000505, -0.00000890, -0.00001504, -0.00002326, -0.00003510, -0.00007030, 0.00005955, 0.00006263, 0.00002717, 0.00002872, 0.00002131, 0.00001441, 0.00002698, 0.00002076, 0.00001511, 0.00001511, 0.00002961, 0.00002861, 0.00006352, 0.00005456, -0.00007277, -0.00004383, -0.00001611, -0.00001007, -0.00001450, -0.00001017, -0.00000572, 0.00000632, -0.00001403, -0.00000409, -0.00000308, -0.00000309, -0.00000138, -0.00000096, -0.00000128, -0.00000041, -0.00000500, 0.00000112, -0.00000643, 0.00000751, -0.00000953, 0.00000369, -0.00000913, -0.00000785, 0.00000064, -0.00000450, -0.00000503, -0.00000315, -0.00000328, -0.00000180, -0.00000234, -0.00000095, -0.00000194, -0.00000035, -0.00000139, -0.00000063, -0.00000567, -0.00000010, -0.00000493, -0.00000550, -0.00000334, 0.00000331, -0.00000611, -0.00001094, -0.00000243, -0.00000036, -0.00000059, -0.00000576, -0.00000031, 0.00000006, -0.00000438, 0.00000191, -0.00000373, -0.00000591, -0.00000014, -0.00000909, -0.00000432, -0.00000779, -0.00000795, -0.00000716, -0.00003992, -0.00001730, 0.00002592, 0.00003439, 0.00001550, 0.00000938, 0.00001346, 0.00000277, 0.00001877, 0.00000865, 0.00000599, 0.00000775, 0.00001108, 0.00000612, 0.00001152, 0.00000789, 0.00001810, 0.00001200, 0.00003135, 0.00002854, -0.00001673, -0.00003032, -0.00001206, -0.00001122, -0.00000249, -0.00000564, -0.00000571, -0.00000171, -0.00000202, -0.00000157, },
		{ -0.00000271, -0.00000334, 0.00000237, -0.00000213, -0.00000705, -0.00000202, 0.00000042, -0.00000281, -0.00000880, 0.00000051, -0.00000096, -0.00000337, -0.00000498, -0.00000574, 0.00000114, -0.00000200, -0.00000708, -0.00000451, -0.00000827, -0.00001048, -0.00002352, -0.00002560, -0.00007065, 0.00002158, 0.00007762, 0.00002472, 0.00003069, 0.00002065, 0.00001943, 0.00001430, 0.00002975, 0.00001180, 0.00001761, 0.00001675, 0.00003421, 0.00003144, 0.00008175, -0.00001503, -0.00007131, -0.00002175, -0.00001307, -0.00001244, -0.00001213, -0.00000856, 0.00000401, -0.00000570, -0.00001100, -0.00000182, -0.00000394, -0.00000088, -0.00000070, -0.00000232, -0.00000030, -0.00000385, 0.00000073, -0.00000660, 0.00000508, -0.00000478, -0.00000165, -0.00000243, -0.00001282, 0.00000261, -0.00000457, -0.00000486, -0.00000479, -0.00000196, -0.00000302, -0.00000162, -0.00000186, -0.00000088, -0.00000158, -0.00000066, -0.00000067, -0.00000390, -0.00000277, -0.00000257, -0.00000680, -0.00000319, 0.00000107, -0.00000088, -0.00001345, -0.00000231, -0.00000287, 0.00000106, -0.00000442, -0.00000246, 0.00000057, -0.00000369, -0.00000091, 0.00000191, -0.00000931, 0.00000052, -0.00000676, -0.00000656, -0.00000609, -0.00000797, -0.00000418, -0.00002501, -0.00004097, 0.00000993, 0.00003548, 0.00002776, 0.00000821, 0.00001513, 0.00000515, 0.00001199, 0.00001623, 0.00000599, 0.00000561, 0.00001254, 0.00000743, 0.00001079, 0.00000887, 0.00001425, 0.00001577, 0.00002459, 0.00003782, -0.00000597, -0.00003351, -0.00001632, -0.00001111, -0.00000622, -0.00000210, -0.00000741, -0.00000207, -0.00000330, -0.00000065, -0.00000335, },
		{ -0.00000390, 0.00000030, 0.00000103, -0.00000712, -0.00000373, -0.00000018, -0.00000038, -0.00000851, -0.00000253, 0.00000066, -0.00000292, -0.00000333, -0.00000706, -0.00000166, 0.00000100, -0.00000659, -0.00000450, -0.00000774, -0.00000758, -0.00002098, -0.00002110, -0.00006237, -0.00001358, 0.00007974, 0.00003188, 0.00002731, 0.00002147, 0.00002306, 0.00001009, 0.00002503, 0.00001912, 0.00001402, 0.00001417, 0.00002757, 0.00002623, 0.00005715, 0.00005254, -0.00006390, -0.00004355, -0.00001396, -0.00001334, -0.00001215, -0.00001039, -0.00000057, 0.00000019, -0.00001128, -0.00000547, -0.00000298, -0.00000173, 0.00000025, -0.00000234, -0.00000150, -0.00000298, 0.00000102, -0.00000582, 0.00000112, 0.00000028, -0.00000597, 0.00000196, -0.00001451, 0.00000101, -0.00000193, -0.00000557, -0.00000581, -0.00000201, -0.00000279, -0.00000190, -0.00000216, -0.00000052, -0.00000179, -0.00000085, -0.00000070, -0.00000208, -0.00000429, -0.00000163, -0.00000717, -0.00000333, -0.00000116, 0.00000284, -0.00001314, -0.00000455, -0.00000244, -0.00000100, -0.00000131, -0.00000383, -0.00000025, -0.00000190, -0.00000420, 0.00000484, -0.00000726, -0.00000350, -0.00000276, -0.00000769, -0.00000644, -0.00000673, -0.00000483, -0.00001179, -0.00004218, -0.00002073, 0.00003200, 0.00003450, 0.00001738, 0.00001035, 0.00001244, 0.00000531, 0.00001805, 0.00001125, 0.00000357, 0.00001080, 0.00001132, 0.00000857, 0.00001178, 0.00001002, 0.00001770, 0.00002041, 0.00004120, 0.00000874, -0.00003411, -0.00002222, -0.00001062, -0.00000993, -0.00000095, -0.00000650, -0.00000332, -0.00000391, -0.00000120, -0.00000245, -0.00000248, },
		{ -0.00000297, 0.00000321, -0.00000588, -0.00000487, -0.00000190, 0.00000140, -0.00000690, -0.00000506, 0.00000052, -0.00000176, -0.00000262, -0.00000564, -0.00000571, 0.00000226, -0.00000473, -0.00000455, -0.00000749, -0.00000582, -0.00001757, -0.00001842, -0.00005086, -0.00004024, 0.00006772, 0.00004580, 0.00002200, 0.00002343, 0.00002180, 0.00001583, 0.00001285, 0.00002662, 0.00001217, 0.00001500, 0.00001760, 0.00002935, 0.00003035, 0.00007310, -0.00000779, -0.00006819, -0.00001916, -0.00001633, -0.00001064, -0.00001288, -0.00000273, -0.00000048, -0.00000563, -0.00001022, -0.00000295, -0.00000270, 0.00000106, -0.00000192, -0.00000166, -0.00000433, 0.00000216, -0.00000416, -0.00000287, 0.00000333, -0.00000717, 0.00000250, -0.00001192, -0.00000424, 0.00000233, -0.00000635, -0.00000587, -0.00000389, -0.00000123, -0.00000299, -0.00000165, -0.00000135, -0.00000073, -0.00000186, -0.00000042, -0.00000131, -0.00000385, -0.00000227, -0.00000626, -0.00000466, -0.00000186, 0.00000330, -0.00000889, -0.00000957, 0.00000029, -0.00000439, 0.00000088, -0.00000376, -0.00000097, -0.00000148, -0.00000455, 0.00000234, -0.00000047, -0.00000848, -0.00000050, -0.00000723, -0.00000618, -0.00000815, -0.00000281, -0.00000843, -0.00002666, -0.00004477, 0.00001157, 0.00003941, 0.00002720, 0.00001160, 0.00001308, 0.00000797, 0.00001131, 0.00001852, 0.00000483, 0.00000693, 0.00001389, 0.00000894, 0.00001228, 0.00000987, 0.00001528, 0.00001958, 0.00003920, 0.00002505, -0.00003060, -0.00002883, -0.00001159, -0.00001149, -0.00000322, -0.00000374, -0.00000479, -0.00000361, -0.00000312, -0.00000112, -0.00000293, -0.00000267, },
		{ 0.00000224, -0.00000269, -0.00000618, -0.00000282, 0.00000072, -0.00000346, -0.00000742, 0.00000009, -0.00000154, -0.00000159, -0.00000412, -0.00000717, -0.00000040, -0.00000123, -0.00000508, -0.00000626, -0.00000626, -0.00001271, -0.00001831, -0.00003680, -0.00005816, 0.00004598, 0.00005914, 0.00002081, 0.00002308, 0.00001976, 0.00002132, 0.00000827, 0.00002187, 0.00001916, 0.00001149, 0.00001545, 0.00002350, 0.00002518, 0.00005090, 0.00005135, -0.00005707, -0.00004061, -0.00001591, -0.00001327, -0.00001224, -0.00000592, -0.00000129, -0.00000338, -0.00000846, -0.00000744, -0.00000183, -0.00000002, -0.00000026, -0.00000172, -0.00000499, 0.00000040, 0.00000037, -0.00000709, 0.00000463, -0.00000637, 0.00000105, -0.00000767, -0.00000952, 0.00000433, -0.00000465, -0.00000649, -0.00000557, -0.00000087, -0.00000271, -0.00000202, -0.00000162, -0.00000044, -0.00000173, -0.00000112, -0.00000056, -0.00000311, -0.00000257, -0.00000551, -0.00000598, -0.00000229, 0.00000228, -0.00000367, -0.00001343, 0.00000084, -0.00000468, -0.00000047, -0.00000188, -0.00000159, -0.00000155, -0.00000390, -0.00000118, 0.00000394, -0.00000781, -0.00000319, -0.00000475, -0.00000576, -0.00000959, -0.00000226, -0.00000640, -0.00001522, -0.00004329, -0.00002373, 0.00003661, 0.00003472, 0.00002004, 0.00001020, 0.00001337, 0.00000647, 0.00001851, 0.00001220, 0.00000311, 0.00001320, 0.00001165, 0.00001148, 0.00001164, 0.00001250, 0.00001823, 0.00003541, 0.00003860, -0.00002058, -0.00003616, -0.00001385, -0.00001205, -0.00000621, -0.00000252, -0.00000442, -0.00000377, -0.00000439, -0.00000126, -0.00000263, -0.00000180, -0.00000443, },
		{ 0.00000064, -0.00000685, -0.00000296, -0.00000147, -0.00000012, -0.00000837, -0.00000086, -0.00000175, -0.00000087, -0.00000293, -0.00000614, -0.00000444, 0.00000088, -0.00000455, -0.00000484, -0.00000755, -0.00000804, -0.00001893, -0.00002409, -0.00006437, 0.00001661, 0.00006828, 0.00002459, 0.00002124, 0.00001842, 0.00002215, 0.00001271, 0.00001136, 0.00002451, 0.00001152, 0.00001422, 0.00001698, 0.00002539, 0.00002885, 0.00006663, -0.00000351, -0.00006210, -0.00002027, -0.00001729, -0.00001049, -0.00001007, -0.00000003, -0.00000590, -0.00000298, -0.00001133, -0.00000275, -0.00000085, 0.00000097, -0.00000149, -0.00000387, -0.00000401, 0.00000508, -0.00000855, 0.00000283, -0.00000368, -0.00000153, -0.00000329, -0.00001281, 0.00000287, -0.00000096, -0.00000701, -0.00000659, -0.00000220, -0.00000127, -0.00000292, -0.00000137, -0.00000107, -0.00000066, -0.00000218, -0.00000013, -0.00000251, -0.00000223, -0.00000487, -0.00000693, -0.00000317, 0.00000111, 0.00000024, -0.00001379, -0.00000232, -0.00000183, -0.00000360, -0.00000017, -0.00000190, -0.00000082, -0.00000431, -0.00000249, 0.00000319, -0.00000226, -0.00000706, -0.00000352, -0.00000486, -0.00000884, -0.00000568, -0.00000155, -0.00001305, -0.00002672, -0.00004779, 0.00001206, 0.00004250, 0.00002811, 0.00001361, 0.00001264, 0.00001026, 0.00001094, 0.00002020, 0.00000404, 0.00000938, 0.00001383, 0.00001216, 0.00001234, 0.00001229, 0.00001532, 0.00003076, 0.00004743, -0.00000425, -0.00004232, -0.00001785, -0.00001263, -0.00000818, -0.00000373, -0.00000290, -0.00000396, -0.00000467, -0.00000277, -0.00000189, -0.00000175, -0.00000370, -0.00000082, },
		{ -0.00000556, -0.00000371, -0.00000261, 0.00000050, -0.00000648, -0.00000326, -0.00000087, -0.00000185, -0.00000111, -0.00000503, -0.00000610, -0.00000077, -0.00000220, -0.00000429, -0.00000753, -0.00000607, -0.00001711, -0.00001746, -0.00005719, -0.00001646, 0.00006985, 0.00003251, 0.00001987, 0.00001716, 0.00002046, 0.00001781, 0.00000781, 0.00001871, 0.00001907, 0.00000984, 0.00001626, 0.00001955, 0.00002384, 0.00004660, 0.00004846, -0.00004900, -0.00003920, -0.00001802, -0.00001275, -0.00001124, -0.00000091, -0.00000634, -0.00000267, -0.00000796, -0.00000822, 0.00000004, 0.00000029, -0.00000011, -0.00000256, -0.00000695, 0.00000495, -0.00000434, -0.00000251, 0.00000050, -0.00000480, 0.00000069, -0.00001401, -0.00000057, 0.00000204, -0.00000564, -0.00000778, -0.00000380, -0.00000071, -0.00000260, -0.00000191, -0.00000118, -0.00000046, -0.00000184, -0.00000109, -0.00000134, -0.00000242, -0.00000326, -0.00000801, -0.00000399, -0.00000053, 0.00000293, -0.00001169, -0.00000666, 0.00000050, -0.00000480, -0.00000066, -0.00000159, 0.00000024, -0.00000451, -0.00000339, 0.00000126, 0.00000148, -0.00000598, -0.00000620, -0.00000328, -0.00000687, -0.00000929, 0.00000084, -0.00001064, -0.00001642, -0.00004383, -0.00002674, 0.00003907, 0.00003683, 0.00002161, 0.00001081, 0.00001484, 0.00000708, 0.00001931, 0.00001229, 0.00000476, 0.00001378, 0.00001374, 0.00001316, 0.00001261, 0.00001411, 0.00002386, 0.00005148, 0.00001559, -0.00004360, -0.00002502, -0.00001274, -0.00001009, -0.00000507, -0.00000302, -0.00000264, -0.00000499, -0.00000385, -0.00000233, -0.00000143, -0.00000264, -0.00000301, 0.00000157, },
		{ -0.00000473, -0.00000244, -0.00000100, -0.00000345, -0.00000562, 0.00000018, -0.00000324, 0.00000009, -0.00000387, -0.00000563, -0.00000380, -0.00000077, -0.00000333, -0.00000664, -0.00000611, -0.00001344, -0.00001594, -0.00004273, -0.00004301, 0.00005781, 0.00004532, 0.00001845, 0.00001725, 0.00001744, 0.00001980, 0.00001144, 0.00000938, 0.00002292, 0.00001089, 0.00001401, 0.00001557, 0.00002187, 0.00002800, 0.00006056, 0.00000000, -0.00005540, -0.00002372, -0.00001615, -0.00001125, -0.00000410, -0.00000311, -0.00000752, -0.00000147, -0.00001234, -0.00000134, -0.00000001, 0.00000076, -0.00000123, -0.00000695, 0.00000030, 0.00000211, -0.00000691, 0.00000263, -0.00000617, 0.00000223, -0.00001216, -0.00000538, 0.00000361, -0.00000315, -0.00000810, -0.00000563, -0.00000117, -0.00000161, -0.00000256, -0.00000111, -0.00000093, -0.00000075, -0.00000224, -0.00000056, -0.00000265, -0.00000156, -0.00000810, -0.00000533, -0.00000211, 0.00000401, -0.00000786, -0.00001051, 0.00000090, -0.00000406, -0.00000184, -0.00000194, 0.00000113, -0.00000317, -0.00000528, 0.00000050, 0.00000141, -0.00000141, -0.00000849, -0.00000427, -0.00000378, -0.00001067, -0.00000134, -0.00000492, -0.00001529, -0.00002649, -0.00004970, 0.00000985, 0.00004676, 0.00002911, 0.00001513, 0.00001373, 0.00001150, 0.00001134, 0.00002059, 0.00000528, 0.00001065, 0.00001462, 0.00001522, 0.00001234, 0.00001512, 0.00001711, 0.00004877, 0.00003664, -0.00003765, -0.00003487, -0.00001307, -0.00001228, -0.00000558, -0.00000492, -0.00000102, -0.00000468, -0.00000436, -0.00000360, -0.00000134, -0.00000170, -0.00000364, 0.00000050, -0.00000341, },
		{ -0.00000221, -0.00000221, -0.00000198, -0.00000600, -0.00000024, -0.00000300, -0.00000090, -0.00000178, -0.00000511, -0.00000494, -0.00000227, -0.00000138, -0.00000576, -0.00000621, -0.00001041, -0.00001545, -0.00002957, -0.00005471, 0.00003163, 0.00005919, 0.00001878, 0.00001825, 0.00001428, 0.00001950, 0.00001526, 0.00000725, 0.00001607, 0.00001870, 0.00000929, 0.00001585, 0.00001647, 0.00002221, 0.00004390, 0.00004405, -0.00003941, -0.00004092, -0.00001784, -0.00001363, -0.00000656, -0.00000060, -0.00001016, -0.00000069, -0.00000900, -0.00000708, 0.00000102, 0.00000012, 0.00000053, -0.00000549, -0.00000432, 0.00000479, -0.00000594, 0.00000041, -0.00000425, 0.00000035, -0.00000722, -0.00001086, 0.00000408, -0.00000127, -0.00000633, -0.00000799, -0.00000176, -0.00000148, -0.00000202, -0.00000184, -0.00000086, -0.00000059, -0.00000189, -0.00000132, -0.00000193, -0.00000129, -0.00000599, -0.00000767, -0.00000289, 0.00000274, -0.00000250, -0.00001362, 0.00000023, -0.00000341, -0.00000170, -0.00000327, 0.00000107, -0.00000054, -0.00000673, -0.00000080, 0.00000100, 0.00000076, -0.00000552, -0.00000891, -0.00000083, -0.00000971, -0.00000495, -0.00000097, -0.00001363, -0.00001687, -0.00004318, -0.00003145, 0.00004078, 0.00004058, 0.00002170, 0.00001332, 0.00001544, 0.00000829, 0.00001907, 0.00001363, 0.00000636, 0.00001406, 0.00001680, 0.00001371, 0.00001506, 0.00001515, 0.00003825, 0.00005531, -0.00002320, -0.00004505, -0.00001549, -0.00001375, -0.00000652, -0.00000621, -0.00000165, -0.00000265, -0.00000496, -0.00000424, -0.00000262, -0.00000047, -0.00000366, -0.00000064, -0.00000218, -0.00000501, },
		},
		{ { -0.00000155, -0.00000276, -0.00000168, -0.00000008, -0.00000101, -0.00000270, -0.00001164, 0.00000024, 0.00000117, 0.00000030, -0.00000779, -0.00000327, -0.00000257, -0.00000512, -0.00000305, -0.00000005, -0.00000580, -0.00000021, -0.00000647, -0.00000368, -0.00000149, -0.00000106, 0.00000750, -0.00001117, -0.00000525, 0.00000107, -0.00001546, -0.00000271, -0.00000506, -0.00000740, -0.00001318, -0.00002184, -0.00004798, 0.00000013, 0.00006462, 0.00002727, 0.00001700, 0.00001339, 0.00001715, 0.00001489, 0.00001139, 0.00000898, 0.00001711, 0.00001031, 0.00001158, 0.00001554, 0.00001897, 0.00002891, 0.00005748, 0.00000203, -0.00005037, -0.00002079, -0.00001357, -0.00001021, -0.00000487, -0.00000738, -0.00000596, -0.00000423, 0.00000069, -0.00000104, -0.00000219, -0.00000305, -0.00000615, -0.00000352, -0.00000315, -0.00000352, -0.00000615, -0.00000305, -0.00000219, -0.00000104, 0.00000069, -0.00000423, -0.00000596, -0.00000738, -0.00000487, -0.00001021, -0.00001357, -0.00002079, -0.00005037, 0.00000203, 0.00005748, 0.00002891, 0.00001897, 0.00001554, 0.00001158, 0.00001031, 0.00001711, 0.00000898, 0.00001139, 0.00001489, 0.00001715, 0.00001339, 0.00001700, 0.00002727, 0.00006462, 0.00000013, -0.00004798, -0.00002184, -0.00001318, -0.00000740, -0.00000506, -0.00000271, -0.00001546, 0.00000107, -0.00000525, -0.00001117, 0.00000750, -0.00000106, -0.00000149, -0.00000368, -0.00000647, -0.00000021, -0.00000580, -0.00000005, -0.00000305, -0.00000512, -0.00000257, -0.00000327, -0.00000779, 0.00000030, 0.00000117, 0.00000024, -0.00001164, -0.00000270, -0.00000101, -0.00000008, -0.00000168, -0.00000276, },
		{ -0.00000172, -0.00000372, -0.00000021, -0.00000051, -0.00000055, -0.00000609, -0.00000955, 0.00000246, 0.00000096, -0.00000235, -0.00000838, -0.00000105, -0.00000380, -0.00000511, -0.00000147, -0.00000095, -0.00000515, -0.00000211, -0.00000646, -0.00000078, -0.00000376, 0.00000215, 0.00000368, -0.00001275, 0.00000058, -0.00000746, -0.00001203, 0.00000016, -0.00000848, -0.00000757, -0.00001666, -0.00003294, -0.00003711, 0.00004076, 0.00005099, 0.00001696, 0.00001517, 0.00001255, 0.00001686, 0.00001349, 0.00000846, 0.00001297, 0.00001403, 0.00000869, 0.00001237, 0.00001653, 0.00001859, 0.00003563, 0.00005054, -0.00001573, -0.00004584, -0.00001540, -0.00001372, -0.00000782, -0.00000554, -0.00000717, -0.00000558, -0.00000229, 0.00000088, -0.00000221, -0.00000209, -0.00000362, -0.00000617, -0.00000302, -0.00000333, -0.00000369, -0.00000648, -0.00000205, -0.00000239, 0.00000060, -0.00000123, -0.00000508, -0.00000666, -0.00000673, -0.00000532, -0.00001222, -0.00001372, -0.00002772, -0.00005191, 0.00002422, 0.00005753, 0.00002367, 0.00001915, 0.00001485, 0.00001055, 0.00001505, 0.00001503, 0.00000833, 0.00001404, 0.00001616, 0.00001697, 0.00001583, 0.00001877, 0.00004802, 0.00005179, -0.00004111, -0.00003673, -0.00001707, -0.00000972, -0.00000771, -0.00000116, -0.00001078, -0.00000973, 0.00000347, -0.00001103, -0.00000646, 0.00000907, -0.00000380, 0.00000074, -0.00000864, -0.00000220, -0.00000133, -0.00000527, -0.00000028, -0.00000393, -0.00000489, -0.00000186, -0.00000605, -0.00000457, 0.00000094, 0.00000151, -0.00000345, -0.00001096, -0.00000044, -0.00000143, -0.00000050, -0.00000194, -0.00000210, },
		{ -0.00000247, -0.00000381, 0.00000134, -0.00000139, -0.00000070, -0.00000905, -0.00000579, 0.00000327, 0.00000006, -0.00000539, -0.00000668, -0.00000027, -0.00000506, -0.00000433, -0.00000023, -0.00000188, -0.00000539, -0.00000394, -0.00000367, -0.00000103, -0.00000437, 0.00000439, -0.00000123, -0.00001003, 0.00000087, -0.00001438, -0.00000356, -0.00000373, -0.00000825, -0.00000991, -0.00002273, -0.00004057, -0.00000402, 0.00005749, 0.00003106, 0.00001464, 0.00001227, 0.00001301, 0.00001653, 0.00001035, 0.00000909, 0.00001492, 0.00000949, 0.00000959, 0.00001262, 0.00001744, 0.00001905, 0.00004137, 0.00003929, -0.00002913, -0.00003919, -0.00001173, -0.00001366, -0.00000572, -0.00000663, -0.00000648, -0.00000486, -0.00000042, -0.00000008, -0.00000280, -0.00000185, -0.00000436, -0.00000602, -0.00000259, -0.00000339, -0.00000419, -0.00000643, -0.00000114, -0.00000206, 0.00000140, -0.00000373, -0.00000515, -0.00000734, -0.00000586, -0.00000649, -0.00001416, -0.00001362, -0.00003766, -0.00004677, 0.00004732, 0.00005076, 0.00002128, 0.00001866, 0.00001405, 0.00001158, 0.00001900, 0.00001005, 0.00001198, 0.00001434, 0.00001845, 0.00001656, 0.00001866, 0.00002689, 0.00006645, 0.00000853, -0.00005635, -0.00002169, -0.00001593, -0.00000718, -0.00000580, -0.00000270, -0.00001531, -0.00000079, -0.00000057, -0.00001346, -0.00000040, 0.00000773, -0.00000438, -0.00000000, -0.00001127, 0.00000309, -0.00000485, -0.00000336, -0.00000158, -0.00000397, -0.00000458, -0.00000246, -0.00000718, -0.00000085, 0.00000011, 0.00000155, -0.00000729, -0.00000823, 0.00000042, -0.00000199, -0.00000088, -0.00000147, -0.00000195, },
		{ -0.00000379, -0.00000222, 0.00000157, -0.00000184, -0.00000203, -0.00001013, -0.00000200, 0.00000309, -0.00000200, -0.00000689, -0.00000448, -0.00000056, -0.00000605, -0.00000272, 0.00000018, -0.00000271, -0.00000718, -0.00000284, -0.00000181, -0.00000249, -0.00000368, 0.00000506, -0.00000511, -0.00000550, -0.00000529, -0.00001293, 0.00000097, -0.00000865, -0.00000640, -0.00001500, -0.00003148, -0.00003140, 0.00003153, 0.00004950, 0.00001910, 0.00001337, 0.00001012, 0.00001415, 0.00001516, 0.00000720, 0.00001295, 0.00001195, 0.00000785, 0.00001021, 0.00001347, 0.00001750, 0.00002093, 0.00004471, 0.00002582, -0.00003876, -0.00003053, -0.00001058, -0.00001242, -0.00000485, -0.00000717, -0.00000584, -0.00000336, 0.00000022, -0.00000105, -0.00000320, -0.00000121, -0.00000590, -0.00000494, -0.00000285, -0.00000284, -0.00000570, -0.00000508, -0.00000106, -0.00000065, 0.00000007, -0.00000502, -0.00000561, -0.00000716, -0.00000577, -0.00000768, -0.00001620, -0.00001350, -0.00005115, -0.00003072, 0.00006488, 0.00004073, 0.00002104, 0.00001770, 0.00001328, 0.00001577, 0.00001771, 0.00000835, 0.00001542, 0.00001477, 0.00002009, 0.00001766, 0.00002120, 0.00004660, 0.00005860, -0.00004029, -0.00004311, -0.00001685, -0.00001226, -0.00000694, -0.00000229, -0.00001027, -0.00001040, 0.00000260, -0.00000552, -0.00001316, 0.00000593, 0.00000375, -0.00000273, -0.00000441, -0.00000835, 0.00000477, -0.00000757, -0.00000175, -0.00000265, -0.00000395, -0.00000425, -0.00000419, -0.00000537, 0.00000086, -0.00000053, 0.00000035, -0.00000953, -0.00000512, 0.00000033, -0.00000288, -0.00000021, -0.00000143, -0.00000196, },
		{ -0.00000457, 0.00000010, 0.00000048, -0.00000189, -0.00000399, -0.00000930, 0.00000106, 0.00000162, -0.00000410, -0.00000680, -0.00000278, -0.00000152, -0.00000626, -0.00000101, 0.00000022, -0.00000472, -0.00000814, 0.00000029, -0.00000240, -0.00000338, -0.00000270, 0.00000449, -0.00000636, -0.00000409, -0.00001179, -0.00000519, -0.00000211, -0.00000911, -0.00000719, -0.00002264, -0.00003593, -0.00000431, 0.00004910, 0.00003256, 0.00001557, 0.00001039, 0.00000996, 0.00001521, 0.00001184, 0.00000737, 0.00001533, 0.00000690, 0.00000934, 0.00000948, 0.00001542, 0.00001641, 0.00002420, 0.00004546, 0.00001059, -0.00004364, -0.00002149, -0.00001139, -0.00001010, -0.00000520, -0.00000692, -0.00000508, -0.00000178, -0.00000013, -0.00000181, -0.00000308, -0.00000096, -0.00000752, -0.00000333, -0.00000343, -0.00000235, -0.00000743, -0.00000284, -0.00000140, 0.00000085, -0.00000273, -0.00000488, -0.00000657, -0.00000638, -0.00000631, -0.00000935, -0.00001694, -0.00001640, -0.00006421, -0.00000404, 0.00007233, 0.00003143, 0.00002172, 0.00001615, 0.00001457, 0.00001962, 0.00001280, 0.00001171, 0.00001583, 0.00001719, 0.00002040, 0.00002011, 0.00002885, 0.00006731, 0.00001510, -0.00006116, -0.00002418, -0.00001773, -0.00000759, -0.00000629, -0.00000282, -0.00001556, -0.00000123, -0.00000051, -0.00000825, -0.00001070, 0.00001061, -0.00000074, -0.00000142, -0.00000889, -0.00000200, 0.00000209, -0.00000797, -0.00000119, -0.00000305, -0.00000410, -0.00000454, -0.00000519, -0.00000216, 0.00000037, -0.00000060, -0.00000176, -0.00000991, -0.00000246, -0.00000074, -0.00000316, 0.00000096, -0.00000201, -0.00000234, },
		{ -0.00000364, 0.00000137, -0.00000075, -0.00000218, -0.00000542, -0.00000748, 0.00000316, -0.00000126, -0.00000466, -0.00000658, -0.00000127, -0.00000312, -0.00000522, 0.00000011, -0.00000005, -0.00000831, -0.00000564, 0.00000198, -0.00000375, -0.00000389, -0.00000164, 0.00000369, -0.00000609, -0.00000771, -0.00001166, 0.00000028, -0.00000778, -0.00000601, -0.00001303, -0.00003015, -0.00002741, 0.00002665, 0.00004449, 0.00002134, 0.00001371, 0.00000760, 0.00001127, 0.00001504, 0.00000801, 0.00001119, 0.00001269, 0.00000465, 0.00001071, 0.00000899, 0.00001732, 0.00001475, 0.00002871, 0.00004317, -0.00000572, -0.00004261, -0.00001463, -0.00001236, -0.00000786, -0.00000586, -0.00000630, -0.00000383, -0.00000115, -0.00000044, -0.00000280, -0.00000188, -0.00000227, -0.00000782, -0.00000241, -0.00000332, -0.00000317, -0.00000775, -0.00000115, -0.00000104, 0.00000064, -0.00000483, -0.00000472, -0.00000695, -0.00000624, -0.00000646, -0.00001232, -0.00001501, -0.00002620, -0.00006933, 0.00002717, 0.00006989, 0.00002485, 0.00002252, 0.00001432, 0.00001929, 0.00001816, 0.00001094, 0.00001524, 0.00001606, 0.00001931, 0.00002193, 0.00002216, 0.00004895, 0.00006212, -0.00003872, -0.00004793, -0.00001831, -0.00001431, -0.00000642, -0.00000339, -0.00000992, -0.00001106, 0.00000232, -0.00000406, -0.00001019, -0.00000535, 0.00001146, -0.00000362, -0.00000256, -0.00000929, 0.00000271, -0.00000191, -0.00000725, -0.00000079, -0.00000351, -0.00000418, -0.00000576, -0.00000373, -0.00000035, -0.00000059, -0.00000092, -0.00000350, -0.00000940, -0.00000029, -0.00000291, -0.00000166, 0.00000114, -0.00000253, -0.00000339, },
		{ -0.00000153, 0.00000109, -0.00000159, -0.00000266, -0.00000612, -0.00000515, 0.00000356, -0.00000413, -0.00000390, -0.00000682, 0.00000026, -0.00000501, -0.00000322, 0.00000063, -0.00000165, -0.00001125, -0.00000041, 0.00000046, -0.00000396, -0.00000513, 0.00000024, 0.00000320, -0.00000751, -0.00001217, -0.00000509, -0.00000192, -0.00000874, -0.00000564, -0.00002167, -0.00003248, -0.00000461, 0.00004407, 0.00002959, 0.00001851, 0.00000945, 0.00000743, 0.00001247, 0.00001308, 0.00000658, 0.00001485, 0.00000670, 0.00000663, 0.00000969, 0.00001056, 0.00001744, 0.00001383, 0.00003414, 0.00003654, -0.00002054, -0.00003670, -0.00001104, -0.00001233, -0.00000645, -0.00000620, -0.00000537, -0.00000262, -0.00000131, -0.00000057, -0.00000364, -0.00000026, -0.00000485, -0.00000655, -0.00000241, -0.00000265, -0.00000515, -0.00000628, -0.00000042, -0.00000017, -0.00000138, -0.00000531, -0.00000522, -0.00000657, -0.00000688, -0.00000645, -0.00001553, -0.00001229, -0.00004249, -0.00006153, 0.00005565, 0.00006027, 0.00002204, 0.00002192, 0.00001497, 0.00002363, 0.00001282, 0.00001436, 0.00001571, 0.00001834, 0.00002006, 0.00002493, 0.00002870, 0.00007271, 0.00001667, -0.00006303, -0.00002698, -0.00002012, -0.00000811, -0.00000653, -0.00000325, -0.00001594, -0.00000117, -0.00000135, -0.00000481, -0.00001183, 0.00000171, 0.00000886, -0.00000546, -0.00000424, -0.00000628, 0.00000389, -0.00000524, -0.00000602, -0.00000043, -0.00000411, -0.00000478, -0.00000629, -0.00000093, -0.00000053, -0.00000129, -0.00000139, -0.00000473, -0.00000831, 0.00000077, -0.00000490, 0.00000104, -0.00000003, -0.00000293, -0.00000433, },
		{ -0.00000013, 0.00000034, -0.00000248, -0.00000252, -0.00000679, -0.00000227, 0.00000153, -0.00000483, -0.00000386, -0.00000642, 0.00000098, -0.00000621, -0.00000112, 0.00000068, -0.00000526, -0.00001043, 0.00000369, -0.00000231, -0.00000364, -0.00000668, 0.00000343, 0.00000200, -0.00001221, -0.00001115, 0.00000014, -0.00000765, -0.00000476, -0.00001211, -0.00002791, -0.00002501, 0.00002322, 0.00004117, 0.00001926, 0.00001735, 0.00000493, 0.00000936, 0.00001247, 0.00001024, 0.00000869, 0.00001422, 0.00000237, 0.00000977, 0.00000745, 0.00001375, 0.00001517, 0.00001509, 0.00003917, 0.00002493, -0.00003051, -0.00002909, -0.00000965, -0.00001167, -0.00000552, -0.00000633, -0.00000384, -0.00000248, -0.00000110, -0.00000131, -0.00000323, 0.00000006, -0.00000695, -0.00000499, -0.00000236, -0.00000277, -0.00000660, -0.00000428, 0.00000021, -0.00000034, -0.00000334, -0.00000532, -0.00000546, -0.00000666, -0.00000707, -0.00000766, -0.00001671, -0.00001327, -0.00005977, -0.00004149, 0.00007686, 0.00004637, 0.00002348, 0.00001879, 0.00002082, 0.00002128, 0.00001139, 0.00001742, 0.00001664, 0.00001947, 0.00002318, 0.00002605, 0.00004982, 0.00006973, -0.00004287, -0.00004828, -0.00002162, -0.00001607, -0.00000640, -0.00000395, -0.00001013, -0.00001183, 0.00000307, -0.00000530, -0.00000529, -0.00001096, 0.00000727, 0.00000490, -0.00000764, -0.00000319, -0.00000400, 0.00000363, -0.00000841, -0.00000340, -0.00000131, -0.00000410, -0.00000659, -0.00000412, 0.00000030, -0.00000103, -0.00000239, -0.00000090, -0.00000656, -0.00000604, -0.00000029, -0.00000469, 0.00000284, -0.00000149, -0.00000375, -0.00000375, },
		{ -0.00000024, -0.00000016, -0.00000331, -0.00000186, -0.00000751, 0.00000037, -0.00000188, -0.00000311, -0.00000574, -0.00000431, 0.00000014, -0.00000597, 0.00000041, -0.00000036, -0.00000923, -0.00000566, 0.00000390, -0.00000350, -0.00000489, -0.00000649, 0.00000708, -0.00000202, -0.00001694, -0.00000381, -0.00000252, -0.00000815, -0.00000385, -0.00002145, -0.00002830, -0.00000642, 0.00004144, 0.00002648, 0.00001815, 0.00001239, 0.00000394, 0.00001086, 0.00001136, 0.00000853, 0.00001189, 0.00000943, 0.00000250, 0.00001071, 0.00000665, 0.00001640, 0.00001154, 0.00001953, 0.00004106, 0.00001049, -0.00003460, -0.00002223, -0.00000921, -0.00001083, -0.00000475, -0.00000607, -0.00000232, -0.00000324, -0.00000028, -0.00000260, -0.00000166, -0.00000140, -0.00000761, -0.00000384, -0.00000205, -0.00000395, -0.00000668, -0.00000249, 0.00000068, -0.00000175, -0.00000429, -0.00000547, -0.00000529, -0.00000743, -0.00000670, -0.00000980, -0.00001620, -0.00001983, -0.00007271, -0.00001216, 0.00008747, 0.00003274, 0.00002664, 0.00001616, 0.00002814, 0.00001320, 0.00001683, 0.00001631, 0.00002056, 0.00001815, 0.00002934, 0.00002948, 0.00007894, 0.00001994, -0.00006962, -0.00002497, -0.00002523, -0.00000784, -0.00000731, -0.00000319, -0.00001692, -0.00000115, -0.00000092, -0.00000551, -0.00000688, -0.00000739, 0.00001032, -0.00000045, -0.00000799, -0.00000068, -0.00000350, 0.00000283, -0.00001098, 0.00000006, -0.00000335, -0.00000400, -0.00000810, -0.00000026, -0.00000071, -0.00000116, -0.00000353, 0.00000021, -0.00000892, -0.00000309, -0.00000275, -0.00000212, 0.00000266, -0.00000265, -0.00000458, -0.00000185, },
		{ -0.00000069, -0.00000100, -0.00000299, -0.00000201, -0.00000731, 0.00000118, -0.00000398, -0.00000141, -0.00000813, -0.00000123, -0.00000173, -0.00000461, 0.00000124, -0.00000326, -0.00001031, -0.00000063, 0.00000181, -0.00000341, -0.00000776, -0.00000323, 0.00000921, -0.00000983, -0.00001562, 0.00000231, -0.00000935, -0.00000227, -0.00001181, -0.00002592, -0.00002212, 0.00001855, 0.00004078, 0.00001524, 0.00001942, 0.00000584, 0.00000622, 0.00001069, 0.00000981, 0.00000930, 0.00001230, 0.00000475, 0.00000503, 0.00000955, 0.00000797, 0.00001690, 0.00000843, 0.00002675, 0.00003721, -0.00000268, -0.00003484, -0.00001624, -0.00000980, -0.00000926, -0.00000466, -0.00000470, -0.00000228, -0.00000326, -0.00000010, -0.00000303, -0.00000068, -0.00000314, -0.00000770, -0.00000239, -0.00000271, -0.00000469, -0.00000632, -0.00000029, -0.00000045, -0.00000259, -0.00000537, -0.00000473, -0.00000615, -0.00000740, -0.00000723, -0.00001098, -0.00001706, -0.00002966, -0.00007927, 0.00002383, 0.00008462, 0.00002502, 0.00002706, 0.00001926, 0.00002836, 0.00000906, 0.00002208, 0.00001587, 0.00002265, 0.00002017, 0.00003266, 0.00004886, 0.00008154, -0.00004954, -0.00005010, -0.00002163, -0.00002018, -0.00000581, -0.00000456, -0.00001037, -0.00001290, 0.00000380, -0.00000550, -0.00000494, -0.00000724, -0.00000372, 0.00001188, -0.00000790, -0.00000309, -0.00000151, -0.00000172, -0.00000042, -0.00001041, 0.00000168, -0.00000474, -0.00000525, -0.00000666, 0.00000182, -0.00000180, -0.00000214, -0.00000286, -0.00000020, -0.00001003, -0.00000134, -0.00000421, 0.00000066, 0.00000143, -0.00000402, -0.00000396, -0.00000060, },
		{ -0.00000055, -0.00000232, -0.00000127, -0.00000368, -0.00000589, -0.00000006, -0.00000336, -0.00000202, -0.00000882, 0.00000117, -0.00000351, -0.00000279, 0.00000085, -0.00000692, -0.00000748, 0.00000123, 0.00000075, -0.00000479, -0.00000955, 0.00000213, 0.00000746, -0.00001836, -0.00000649, -0.00000078, -0.00000980, -0.00000003, -0.00002286, -0.00002343, -0.00000836, 0.00003815, 0.00002629, 0.00001439, 0.00001587, 0.00000293, 0.00000835, 0.00000952, 0.00000905, 0.00001113, 0.00000921, 0.00000334, 0.00000670, 0.00000846, 0.00001012, 0.00001500, 0.00000810, 0.00003391, 0.00002817, -0.00001214, -0.00003330, -0.00001087, -0.00001143, -0.00000669, -0.00000525, -0.00000253, -0.00000391, -0.00000186, -0.00000106, -0.00000214, -0.00000113, -0.00000422, -0.00000763, -0.00000072, -0.00000448, -0.00000434, -0.00000585, 0.00000179, -0.00000301, -0.00000226, -0.00000686, -0.00000322, -0.00000811, -0.00000649, -0.00000834, -0.00001183, -0.00001979, -0.00004153, -0.00007672, 0.00006090, 0.00007022, 0.00002458, 0.00002485, 0.00002678, 0.00002001, 0.00001415, 0.00002143, 0.00002011, 0.00002005, 0.00002812, 0.00003496, 0.00008169, 0.00002978, -0.00008292, -0.00002007, -0.00003003, -0.00000835, -0.00000848, -0.00000224, -0.00001888, -0.00000082, -0.00000049, -0.00000601, -0.00000531, -0.00000609, -0.00000029, 0.00001038, -0.00001363, 0.00000467, -0.00000621, 0.00000218, -0.00000613, -0.00000626, 0.00000039, -0.00000508, -0.00000703, -0.00000264, 0.00000090, -0.00000186, -0.00000395, -0.00000022, -0.00000299, -0.00000911, -0.00000135, -0.00000358, 0.00000205, -0.00000003, -0.00000531, -0.00000182, -0.00000108, },
		{ -0.00000057, -0.00000276, 0.00000018, -0.00000574, -0.00000455, -0.00000130, -0.00000173, -0.00000443, -0.00000747, 0.00000237, -0.00000488, -0.00000076, -0.00000173, -0.00000817, -0.00000395, 0.00000100, 0.00000079, -0.00000826, -0.00000819, 0.00000739, 0.00000020, -0.00002071, 0.00000273, -0.00000989, -0.00000138, -0.00000950, -0.00002615, -0.00001753, 0.00001275, 0.00004080, 0.00001334, 0.00001740, 0.00000925, 0.00000400, 0.00000876, 0.00000805, 0.00001037, 0.00001052, 0.00000631, 0.00000358, 0.00000749, 0.00000791, 0.00001221, 0.00001109, 0.00001233, 0.00003673, 0.00001792, -0.00001912, -0.00002963, -0.00000751, -0.00001241, -0.00000446, -0.00000513, -0.00000154, -0.00000516, -0.00000055, -0.00000167, -0.00000161, -0.00000158, -0.00000565, -0.00000616, -0.00000072, -0.00000530, -0.00000439, -0.00000390, 0.00000138, -0.00000421, -0.00000289, -0.00000667, -0.00000344, -0.00000873, -0.00000692, -0.00000753, -0.00001602, -0.00002082, -0.00005782, -0.00005954, 0.00008922, 0.00005228, 0.00002709, 0.00002516, 0.00002992, 0.00001290, 0.00002153, 0.00001977, 0.00002355, 0.00001980, 0.00003453, 0.00005233, 0.00009071, -0.00005029, -0.00006024, -0.00001564, -0.00002672, -0.00000441, -0.00000642, -0.00000902, -0.00001571, 0.00000538, -0.00000551, -0.00000586, -0.00000388, -0.00000655, 0.00000503, 0.00000294, -0.00001138, 0.00000775, -0.00000940, 0.00000443, -0.00001049, -0.00000178, -0.00000190, -0.00000585, -0.00000649, 0.00000025, -0.00000065, -0.00000239, -0.00000431, 0.00000179, -0.00000632, -0.00000755, -0.00000144, -0.00000231, 0.00000256, -0.00000216, -0.00000492, -0.00000031, -0.00000197, },
		{ -0.00000135, -0.00000152, -0.00000008, -0.00000712, -0.00000400, -0.00000098, -0.00000139, -0.00000662, -0.00000511, 0.00000261, -0.00000581, 0.00000082, -0.00000572, -0.00000555, -0.00000337, 0.00000221, -0.00000119, -0.00001102, -0.00000425, 0.00001008, -0.00001058, -0.00001308, 0.00000180, -0.00001227, 0.00000389, -0.00002351, -0.00002022, -0.00000832, 0.00003300, 0.00002769, 0.00001138, 0.00001550, 0.00000580, 0.00000521, 0.00000839, 0.00000708, 0.00001279, 0.00000652, 0.00000662, 0.00000253, 0.00000927, 0.00000684, 0.00001394, 0.00000676, 0.00002018, 0.00003346, 0.00000976, -0.00002541, -0.00002318, -0.00000747, -0.00001116, -0.00000372, -0.00000361, -0.00000281, -0.00000451, -0.00000050, -0.00000107, -0.00000245, -0.00000109, -0.00000792, -0.00000305, -0.00000304, -0.00000405, -0.00000551, -0.00000055, -0.00000199, -0.00000276, -0.00000549, -0.00000414, -0.00000625, -0.00000717, -0.00000883, -0.00000542, -0.00002353, -0.00001957, -0.00007796, -0.00002593, 0.00010181, 0.00003774, 0.00002997, 0.00002866, 0.00002512, 0.00001431, 0.00002389, 0.00002243, 0.00002175, 0.00002664, 0.00003857, 0.00008671, 0.00003989, -0.00009405, -0.00002143, -0.00002864, -0.00001222, -0.00000842, -0.00000234, -0.00001958, -0.00000211, 0.00000074, -0.00000578, -0.00000695, -0.00000071, -0.00000832, 0.00001033, -0.00000730, -0.00000169, 0.00000315, -0.00000768, 0.00000261, -0.00001097, 0.00000051, -0.00000398, -0.00000674, -0.00000352, 0.00000026, -0.00000118, -0.00000397, -0.00000227, 0.00000137, -0.00000863, -0.00000616, -0.00000073, -0.00000165, 0.00000267, -0.00000462, -0.00000254, -0.00000085, -0.00000197, },
		{ -0.00000162, 0.00000001, -0.00000144, -0.00000838, -0.00000285, -0.00000035, -0.00000217, -0.00000807, -0.00000198, 0.00000142, -0.00000549, 0.00000017, -0.00000768, -0.00000229, -0.00000455, 0.00000488, -0.00000611, -0.00001074, 0.00000067, 0.00000747, -0.00001764, -0.00000226, -0.00000781, -0.00000305, -0.00000515, -0.00002760, -0.00001288, 0.00000798, 0.00003914, 0.00001372, 0.00001466, 0.00000982, 0.00000626, 0.00000502, 0.00000758, 0.00000833, 0.00001245, 0.00000328, 0.00000783, 0.00000118, 0.00001136, 0.00000568, 0.00001425, 0.00000499, 0.00002742, 0.00002691, 0.00000297, -0.00002985, -0.00001593, -0.00000930, -0.00000851, -0.00000354, -0.00000232, -0.00000464, -0.00000315, -0.00000046, -0.00000090, -0.00000294, -0.00000131, -0.00000899, -0.00000082, -0.00000519, -0.00000272, -0.00000567, 0.00000112, -0.00000487, -0.00000139, -0.00000738, -0.00000230, -0.00000868, -0.00000628, -0.00000875, -0.00000732, -0.00002800, -0.00002266, -0.00009359, 0.00001721, 0.00009972, 0.00002789, 0.00003548, 0.00002819, 0.00002078, 0.00001865, 0.00002526, 0.00002406, 0.00002206, 0.00003257, 0.00005771, 0.00009936, -0.00004757, -0.00007311, -0.00001146, -0.00002950, -0.00000531, -0.00000797, -0.00000742, -0.00001868, 0.00000642, -0.00000536, -0.00000548, -0.00000532, -0.00000072, -0.00000681, 0.00001003, -0.00001201, 0.00000699, -0.00000397, -0.00000383, -0.00000087, -0.00000936, 0.00000135, -0.00000673, -0.00000511, -0.00000170, -0.00000034, -0.00000215, -0.00000441, -0.00000008, -0.00000035, -0.00001051, -0.00000368, -0.00000073, -0.00000068, 0.00000137, -0.00000531, -0.00000057, -0.00000205, -0.00000182, },
		{ -0.00000034, 0.00000024, -0.00000298, -0.00000976, -0.00000006, -0.00000126, -0.00000261, -0.00000928, 0.00000200, -0.00000161, -0.00000364, -0.00000314, -0.00000513, -0.00000281, -0.00000306, 0.00000495, -0.00001068, -0.00000817, 0.00000503, -0.00000060, -0.00001474, -0.00000035, -0.00001200, 0.00000486, -0.00002121, -0.00001964, -0.00000645, 0.00002731, 0.00002835, 0.00001069, 0.00001306, 0.00000750, 0.00000613, 0.00000505, 0.00000628, 0.00001180, 0.00000739, 0.00000490, 0.00000571, 0.00000292, 0.00001099, 0.00000609, 0.00001231, 0.00000750, 0.00003039, 0.00002082, -0.00000467, -0.00003027, -0.00001077, -0.00001049, -0.00000613, -0.00000280, -0.00000274, -0.00000514, -0.00000244, 0.00000059, -0.00000246, -0.00000162, -0.00000358, -0.00000744, -0.00000133, -0.00000500, -0.00000300, -0.00000378, -0.00000066, -0.00000464, -0.00000248, -0.00000650, -0.00000343, -0.00000856, -0.00000750, -0.00000573, -0.00001554, -0.00002495, -0.00003630, -0.00009533, 0.00006023, 0.00008682, 0.00002431, 0.00004175, 0.00002148, 0.00002343, 0.00001935, 0.00003041, 0.00002012, 0.00003002, 0.00003547, 0.00009716, 0.00004700, -0.00009960, -0.00002942, -0.00002445, -0.00001564, -0.00000878, -0.00000349, -0.00001851, -0.00000527, 0.00000285, -0.00000605, -0.00000735, -0.00000016, -0.00000500, -0.00000164, 0.00000411, -0.00000841, 0.00000895, -0.00000851, -0.00000096, -0.00000368, -0.00000737, 0.00000100, -0.00000915, -0.00000114, -0.00000275, 0.00000014, -0.00000426, -0.00000257, 0.00000042, -0.00000226, -0.00001199, 0.00000011, -0.00000238, 0.00000117, -0.00000140, -0.00000342, -0.00000087, -0.00000227, -0.00000237, },
		{ 0.00000106, -0.00000019, -0.00000551, -0.00000927, 0.00000233, -0.00000276, -0.00000317, -0.00000924, 0.00000519, -0.00000497, -0.00000215, -0.00000592, -0.00000092, -0.00000628, 0.00000171, 0.00000059, -0.00001207, -0.00000468, 0.00000602, -0.00000804, -0.00000655, -0.00000820, -0.00000337, -0.00000265, -0.00002670, -0.00001082, 0.00000574, 0.00003581, 0.00001441, 0.00001390, 0.00000712, 0.00000958, 0.00000377, 0.00000557, 0.00000634, 0.00001350, 0.00000178, 0.00000909, 0.00000100, 0.00000766, 0.00000773, 0.00000808, 0.00000949, 0.00001243, 0.00002965, 0.00001542, -0.00001249, -0.00002727, -0.00000803, -0.00001062, -0.00000389, -0.00000258, -0.00000366, -0.00000510, -0.00000134, 0.00000092, -0.00000387, -0.00000021, -0.00000610, -0.00000525, -0.00000277, -0.00000384, -0.00000337, -0.00000235, -0.00000315, -0.00000325, -0.00000420, -0.00000505, -0.00000537, -0.00000800, -0.00000786, -0.00000471, -0.00002442, -0.00001878, -0.00005836, -0.00008037, 0.00009715, 0.00006522, 0.00003158, 0.00003923, 0.00001872, 0.00002601, 0.00002206, 0.00003201, 0.00001999, 0.00003557, 0.00005575, 0.00011530, -0.00004925, -0.00007992, -0.00001570, -0.00002627, -0.00000850, -0.00000843, -0.00000770, -0.00001911, 0.00000456, -0.00000337, -0.00000578, -0.00000665, 0.00000254, -0.00000818, 0.00000169, -0.00000038, -0.00000273, 0.00000687, -0.00001114, 0.00000198, -0.00000709, -0.00000395, -0.00000233, -0.00000755, 0.00000069, -0.00000371, -0.00000024, -0.00000500, -0.00000085, 0.00000052, -0.00000549, -0.00001077, 0.00000246, -0.00000371, 0.00000216, -0.00000321, -0.00000141, -0.00000202, -0.00000231, -0.00000233, },
		{ 0.00000082, -0.00000029, -0.00000944, -0.00000535, 0.00000170, -0.00000252, -0.00000546, -0.00000628, 0.00000534, -0.00000667, -0.00000280, -0.00000473, -0.00000030, -0.00000743, 0.00000514, -0.00000493, -0.00001103, -0.00000087, 0.00000190, -0.00000748, -0.00000552, -0.00001155, 0.00000517, -0.00001889, -0.00001829, -0.00000591, 0.00002358, 0.00002714, 0.00001099, 0.00001208, 0.00000510, 0.00001047, 0.00000189, 0.00000543, 0.00000905, 0.00001003, 0.00000124, 0.00001014, -0.00000152, 0.00001155, 0.00000399, 0.00001031, 0.00000758, 0.00001706, 0.00002746, 0.00000916, -0.00001825, -0.00002314, -0.00000642, -0.00001029, -0.00000143, -0.00000388, -0.00000356, -0.00000557, 0.00000111, -0.00000101, -0.00000313, -0.00000071, -0.00000692, -0.00000450, -0.00000296, -0.00000345, -0.00000263, -0.00000328, -0.00000362, -0.00000312, -0.00000451, -0.00000519, -0.00000604, -0.00000850, -0.00000597, -0.00000903, -0.00002801, -0.00001657, -0.00008277, -0.00004890, 0.00012161, 0.00004186, 0.00004661, 0.00002540, 0.00002707, 0.00002135, 0.00003236, 0.00002410, 0.00003114, 0.00003427, 0.00010319, 0.00006006, -0.00010874, -0.00003301, -0.00002789, -0.00001213, -0.00001270, -0.00000311, -0.00001902, -0.00000636, 0.00000224, -0.00000449, -0.00000788, -0.00000158, 0.00000013, -0.00000806, 0.00000180, -0.00000035, 0.00000024, 0.00000398, -0.00001269, 0.00000521, -0.00001130, 0.00000040, -0.00000757, -0.00000167, -0.00000195, -0.00000230, -0.00000269, -0.00000286, -0.00000119, 0.00000123, -0.00001018, -0.00000590, 0.00000134, -0.00000299, 0.00000115, -0.00000242, -0.00000165, -0.00000202, -0.00000335, -0.00000036, },
		{ -0.00000017, -0.00000137, -0.00001222, -0.00000032, -0.00000094, -0.00000101, -0.00000868, -0.00000127, 0.00000269, -0.00000716, -0.00000375, -0.00000127, -0.00000356, -0.00000436, 0.00000424, -0.00000828, -0.00000886, 0.00000090, -0.00000264, -0.00000200, -0.00001380, -0.00000148, -0.00000254, -0.00002401, -0.00000964, 0.00000378, 0.00003319, 0.00001369, 0.00001452, 0.00000503, 0.00000880, 0.00000769, 0.00000152, 0.00000582, 0.00001123, 0.00000438, 0.00000502, 0.00000693, 0.00000009, 0.00001221, 0.00000153, 0.00001253, 0.00000604, 0.00002157, 0.00002342, 0.00000300, -0.00002216, -0.00001817, -0.00000626, -0.00000847, -0.00000057, -0.00000486, -0.00000381, -0.00000487, 0.00000264, -0.00000313, -0.00000174, -0.00000180, -0.00000742, -0.00000374, -0.00000310, -0.00000244, -0.00000311, -0.00000404, -0.00000377, -0.00000266, -0.00000543, -0.00000486, -0.00000750, -0.00000716, -0.00000628, -0.00001441, -0.00002790, -0.00002010, -0.00010579, -0.00000019, 0.00012448, 0.00003058, 0.00005396, 0.00001447, 0.00003633, 0.00001796, 0.00004005, 0.00001883, 0.00004067, 0.00005030, 0.00013291, -0.00004993, -0.00008654, -0.00001976, -0.00002705, -0.00000650, -0.00001175, -0.00000736, -0.00001960, 0.00000296, -0.00000260, -0.00000525, -0.00000704, 0.00000150, -0.00000235, -0.00000907, 0.00000449, -0.00000128, 0.00000338, -0.00000171, -0.00000987, 0.00000472, -0.00001251, 0.00000160, -0.00000937, 0.00000286, -0.00000540, -0.00000091, -0.00000432, -0.00000069, -0.00000166, 0.00000042, -0.00001307, -0.00000063, -0.00000134, -0.00000161, 0.00000017, -0.00000122, -0.00000281, -0.00000174, -0.00000394, 0.00000196, },
		{ -0.00000024, -0.00000483, -0.00001098, 0.00000197, -0.00000222, -0.00000098, -0.00001009, 0.00000294, -0.00000077, -0.00000792, -0.00000188, -0.00000057, -0.00000565, -0.00000115, 0.00000150, -0.00000995, -0.00000560, -0.00000190, 0.00000006, -0.00000420, -0.00001656, 0.00000854, -0.00001987, -0.00001390, -0.00000751, 0.00002148, 0.00002533, 0.00001062, 0.00001275, 0.00000196, 0.00001192, 0.00000425, 0.00000133, 0.00000841, 0.00000903, 0.00000236, 0.00000767, 0.00000339, 0.00000340, 0.00001027, 0.00000093, 0.00001450, 0.00000427, 0.00002673, 0.00001631, -0.00000105, -0.00002562, -0.00001177, -0.00000814, -0.00000446, -0.00000274, -0.00000365, -0.00000582, -0.00000151, 0.00000107, -0.00000297, -0.00000183, -0.00000186, -0.00000912, -0.00000137, -0.00000449, 0.00000003, -0.00000631, -0.00000235, -0.00000547, -0.00000053, -0.00000842, -0.00000263, -0.00001062, -0.00000318, -0.00001132, -0.00001664, -0.00002728, -0.00003049, -0.00012064, 0.00005948, 0.00010542, 0.00003682, 0.00004420, 0.00001927, 0.00003448, 0.00002592, 0.00003466, 0.00002827, 0.00003913, 0.00010106, 0.00008093, -0.00012309, -0.00003100, -0.00003693, -0.00000702, -0.00001528, -0.00000325, -0.00002089, -0.00000528, -0.00000066, -0.00000182, -0.00000917, -0.00000110, -0.00000045, -0.00000290, -0.00001147, 0.00001107, -0.00000618, 0.00000869, -0.00001071, -0.00000168, -0.00000138, -0.00000902, -0.00000163, -0.00000531, 0.00000193, -0.00000595, -0.00000182, -0.00000282, -0.00000115, -0.00000032, -0.00000338, -0.00001158, 0.00000162, -0.00000292, -0.00000118, 0.00000104, -0.00000214, -0.00000258, -0.00000280, -0.00000223, 0.00000229, },
		{ -0.00000044, -0.00000868, -0.00000703, 0.00000137, -0.00000147, -0.00000300, -0.00000892, 0.00000524, -0.00000475, -0.00000737, 0.00000124, -0.00000316, -0.00000439, -0.00000077, -0.00000029, -0.00001067, -0.00000349, -0.00000517, 0.00000742, -0.00001687, -0.00000339, -0.00000006, -0.00002432, -0.00000509, -0.00000092, 0.00003297, 0.00001138, 0.00001525, 0.00000473, 0.00000564, 0.00001109, 0.00000168, 0.00000233, 0.00001033, 0.00000480, 0.00000398, 0.00000752, 0.00000159, 0.00000659, 0.00000623, 0.00000371, 0.00001344, 0.00000521, 0.00002950, 0.00000871, -0.00000440, -0.00002716, -0.00000608, -0.00000975, -0.00000099, -0.00000540, -0.00000227, -0.00000729, 0.00000186, -0.00000142, -0.00000210, -0.00000181, -0.00000297, -0.00000949, 0.00000020, -0.00000454, 0.00000038, -0.00000840, -0.00000126, -0.00000569, -0.00000040, -0.00000962, -0.00000243, -0.00001065, -0.00000283, -0.00001503, -0.00001960, -0.00002344, -0.00005364, -0.00011320, 0.00011181, 0.00007970, 0.00004731, 0.00002921, 0.00003167, 0.00002848, 0.00003603, 0.00002767, 0.00003961, 0.00005172, 0.00014155, -0.00004058, -0.00010171, -0.00001477, -0.00003784, 0.00000219, -0.00001771, -0.00000513, -0.00002314, 0.00000524, -0.00000610, -0.00000167, -0.00000934, 0.00000230, -0.00000144, -0.00000712, -0.00000740, 0.00001303, -0.00000866, 0.00001066, -0.00001674, 0.00000584, -0.00000835, -0.00000495, -0.00000437, -0.00000030, -0.00000163, -0.00000530, -0.00000260, -0.00000080, -0.00000236, 0.00000078, -0.00000742, -0.00000788, 0.00000159, -0.00000395, -0.00000016, 0.00000175, -0.00000388, -0.00000175, -0.00000394, 0.00000046, 0.00000130, },
		{ -0.00000263, -0.00000976, -0.00000408, 0.00000083, -0.00000096, -0.00000511, -0.00000673, 0.00000620, -0.00000980, -0.00000274, 0.00000061, -0.00000403, -0.00000369, -0.00000068, -0.00000227, -0.00000935, -0.00000612, -0.00000030, 0.00000549, -0.00002350, 0.00001115, -0.00002073, -0.00001066, -0.00000678, 0.00001708, 0.00002584, 0.00000830, 0.00001410, 0.00000056, 0.00000947, 0.00000909, -0.00000123, 0.00000645, 0.00000746, 0.00000420, 0.00000436, 0.00000765, -0.00000006, 0.00000974, 0.00000064, 0.00001031, 0.00000737, 0.00001196, 0.00002582, 0.00000487, -0.00001017, -0.00002404, -0.00000404, -0.00000835, -0.00000116, -0.00000525, -0.00000351, -0.00000536, 0.00000191, -0.00000162, -0.00000302, 0.00000027, -0.00000721, -0.00000596, -0.00000148, -0.00000129, -0.00000406, -0.00000586, -0.00000387, -0.00000182, -0.00000516, -0.00000594, -0.00000684, -0.00000505, -0.00001020, -0.00001199, -0.00002740, -0.00001539, -0.00009053, -0.00007476, 0.00014126, 0.00006175, 0.00004812, 0.00002496, 0.00003708, 0.00003070, 0.00003528, 0.00003415, 0.00003948, 0.00010288, 0.00009649, -0.00013057, -0.00003556, -0.00003827, -0.00001208, -0.00000933, -0.00000783, -0.00002021, -0.00000710, 0.00000005, -0.00000349, -0.00000714, -0.00000218, -0.00000107, 0.00000120, -0.00001576, 0.00000564, 0.00000502, -0.00000339, 0.00000487, -0.00001475, 0.00000718, -0.00001172, -0.00000342, -0.00000309, 0.00000084, -0.00000378, -0.00000582, -0.00000075, -0.00000132, -0.00000175, -0.00000059, -0.00000872, -0.00000520, 0.00000130, -0.00000541, 0.00000295, -0.00000031, -0.00000379, -0.00000234, -0.00000295, 0.00000153, 0.00000093, },
		{ -0.00000531, -0.00000874, -0.00000262, 0.00000108, -0.00000129, -0.00000682, -0.00000365, 0.00000461, -0.00001293, 0.00000353, -0.00000385, -0.00000127, -0.00000620, 0.00000158, -0.00000526, -0.00000766, -0.00001013, 0.00001109, -0.00001033, -0.00001061, 0.00000476, -0.00002735, 0.00000124, -0.00000564, 0.00003211, 0.00001027, 0.00001449, 0.00000531, 0.00000389, 0.00000967, 0.00000713, -0.00000279, 0.00001043, 0.00000230, 0.00000685, 0.00000272, 0.00000868, -0.00000150, 0.00001132, -0.00000301, 0.00001634, 0.00000024, 0.00002149, 0.00001723, 0.00000437, -0.00001748, -0.00001743, -0.00000472, -0.00000543, -0.00000343, -0.00000353, -0.00000571, -0.00000198, 0.00000020, -0.00000086, -0.00000391, 0.00000171, -0.00001139, -0.00000136, -0.00000354, 0.00000135, -0.00000908, -0.00000209, -0.00000668, 0.00000188, -0.00001036, -0.00000186, -0.00001052, -0.00000085, -0.00001812, -0.00000797, -0.00003398, -0.00001330, -0.00012705, -0.00001436, 0.00014933, 0.00004880, 0.00004478, 0.00002777, 0.00003918, 0.00003419, 0.00003374, 0.00004200, 0.00005326, 0.00014901, -0.00003061, -0.00011642, -0.00001026, -0.00004673, 0.00000511, -0.00001730, -0.00000614, -0.00002582, 0.00000694, -0.00000857, -0.00000025, -0.00001000, 0.00000339, -0.00000382, -0.00000002, -0.00001857, 0.00001703, -0.00000559, 0.00000449, -0.00000395, -0.00000837, 0.00000431, -0.00001312, -0.00000138, -0.00000139, 0.00000008, -0.00000575, -0.00000518, 0.00000106, -0.00000257, -0.00000100, -0.00000237, -0.00000879, -0.00000290, -0.00000026, -0.00000500, 0.00000543, -0.00000345, -0.00000271, -0.00000324, -0.00000094, 0.00000143, 0.00000040, },
		{ -0.00000559, -0.00000875, -0.00000021, 0.00000019, -0.00000095, -0.00000934, 0.00000130, -0.00000138, -0.00000966, 0.00000483, -0.00000587, 0.00000008, -0.00000817, 0.00000364, -0.00000704, -0.00001021, -0.00000570, 0.00001359, -0.00002253, 0.00000752, -0.00001777, -0.00001191, -0.00000174, 0.00001007, 0.00002787, 0.00000546, 0.00001498, 0.00000002, 0.00000761, 0.00000946, 0.00000303, 0.00000042, 0.00000865, 0.00000170, 0.00000682, 0.00000348, 0.00000720, -0.00000020, 0.00000882, -0.00000087, 0.00001617, -0.00000130, 0.00002701, 0.00000949, 0.00000296, -0.00002231, -0.00001116, -0.00000504, -0.00000432, -0.00000420, -0.00000347, -0.00000555, -0.00000085, 0.00000025, -0.00000218, -0.00000182, -0.00000108, -0.00001116, 0.00000018, -0.00000245, -0.00000096, -0.00000933, -0.00000209, -0.00000507, 0.00000023, -0.00001047, -0.00000246, -0.00000831, -0.00000509, -0.00001767, -0.00001155, -0.00003156, -0.00002839, -0.00014529, 0.00005226, 0.00014269, 0.00003488, 0.00004809, 0.00002668, 0.00004774, 0.00002733, 0.00004639, 0.00003625, 0.00011055, 0.00010546, -0.00013193, -0.00004576, -0.00003351, -0.00002244, -0.00000250, -0.00001058, -0.00001979, -0.00001094, 0.00000352, -0.00000777, -0.00000390, -0.00000424, 0.00000081, -0.00000099, -0.00000875, -0.00000920, 0.00001712, -0.00000969, 0.00000790, -0.00000948, -0.00000299, 0.00000058, -0.00001412, 0.00000330, -0.00000302, 0.00000060, -0.00000935, -0.00000103, -0.00000037, -0.00000182, -0.00000226, -0.00000209, -0.00001006, 0.00000052, -0.00000415, -0.00000065, 0.00000387, -0.00000433, -0.00000304, -0.00000209, -0.00000054, 0.00000214, -0.00000215, },
		{ -0.00000460, -0.00000946, 0.00000296, -0.00000160, -0.00000054, -0.00001142, 0.00000591, -0.00000878, -0.00000156, -0.00000013, -0.00000235, -0.00000382, -0.00000580, 0.00000323, -0.00000798, -0.00001493, 0.00000749, -0.00000038, -0.00001398, 0.00000509, -0.00002751, 0.00000363, -0.00000549, 0.00002732, 0.00001211, 0.00001189, 0.00000627, 0.00000318, 0.00000726, 0.00001015, -0.00000175, 0.00000591, 0.00000267, 0.00000604, 0.00000275, 0.00000786, 0.00000239, 0.00000364, 0.00000380, 0.00000536, 0.00001036, 0.00000372, 0.00002615, 0.00000511, -0.00000078, -0.00002347, -0.00000620, -0.00000524, -0.00000439, -0.00000391, -0.00000420, -0.00000419, -0.00000099, 0.00000103, -0.00000396, 0.00000100, -0.00000588, -0.00000796, 0.00000009, -0.00000086, -0.00000531, -0.00000676, -0.00000384, -0.00000175, -0.00000411, -0.00000796, -0.00000442, -0.00000516, -0.00001287, -0.00001194, -0.00002028, -0.00002416, -0.00005828, -0.00013948, 0.00011823, 0.00011722, 0.00003053, 0.00005001, 0.00002924, 0.00005193, 0.00002399, 0.00005746, 0.00004593, 0.00016657, -0.00003114, -0.00012162, -0.00001481, -0.00004551, -0.00000258, -0.00001013, -0.00000976, -0.00002603, 0.00000446, -0.00000582, -0.00000433, -0.00000665, 0.00000152, -0.00000277, 0.00000011, -0.00001531, 0.00000360, 0.00001007, -0.00000869, 0.00000817, -0.00001284, 0.00000211, -0.00000575, -0.00001059, 0.00000558, -0.00000462, -0.00000019, -0.00001097, 0.00000302, -0.00000266, -0.00000109, -0.00000306, -0.00000215, -0.00001053, 0.00000265, -0.00000706, 0.00000447, -0.00000007, -0.00000360, -0.00000374, -0.00000011, -0.00000100, 0.00000253, -0.00000496, },
		{ -0.00000614, -0.00000734, 0.00000369, -0.00000145, -0.00000293, -0.00000976, 0.00000583, -0.00001111, 0.00000361, -0.00000453, 0.00000140, -0.00000915, -0.00000122, 0.00000221, -0.00001272, -0.00001147, 0.00001371, -0.00001403, 0.00000304, -0.00001606, -0.00001297, 0.00000200, 0.00000619, 0.00002665, 0.00000549, 0.00001379, 0.00000004, 0.00000732, 0.00000665, 0.00000837, -0.00000233, 0.00000692, 0.00000012, 0.00000852, 0.00000016, 0.00001122, -0.00000212, 0.00000665, 0.00000058, 0.00001020, 0.00000449, 0.00001067, 0.00002183, 0.00000237, -0.00000498, -0.00002254, -0.00000152, -0.00000726, -0.00000298, -0.00000511, -0.00000300, -0.00000477, 0.00000080, -0.00000057, -0.00000287, 0.00000049, -0.00000814, -0.00000562, 0.00000146, -0.00000285, -0.00000653, -0.00000611, -0.00000297, -0.00000167, -0.00000600, -0.00000732, -0.00000315, -0.00000776, -0.00001544, -0.00000928, -0.00002683, -0.00002070, -0.00009410, -0.00010877, 0.00017539, 0.00007277, 0.00004669, 0.00003804, 0.00004735, 0.00003700, 0.00004340, 0.00004632, 0.00010958, 0.00012405, -0.00014394, -0.00004654, -0.00003734, -0.00002289, -0.00000760, -0.00000372, -0.00002437, -0.00001140, 0.00000273, -0.00000673, -0.00000679, -0.00000146, -0.00000039, -0.00000109, -0.00000577, -0.00001176, 0.00000955, 0.00000393, -0.00000716, 0.00000841, -0.00001625, 0.00000780, -0.00001491, -0.00000051, 0.00000105, -0.00000204, -0.00000529, -0.00000691, 0.00000254, -0.00000223, -0.00000309, -0.00000048, -0.00000574, -0.00000731, 0.00000067, -0.00000516, 0.00000557, -0.00000222, -0.00000414, -0.00000206, -0.00000039, 0.00000024, 0.00000010, -0.00000455, },
		{ -0.00000990, -0.00000225, 0.00000157, 0.00000065, -0.00000784, -0.00000481, 0.00000150, -0.00000733, 0.00000233, -0.00000347, -0.00000004, -0.00001080, 0.00000260, 0.00000046, -0.00001979, 0.00000210, 0.00000299, -0.00000790, 0.00000093, -0.00002591, 0.00000340, -0.00000335, 0.00002266, 0.00001282, 0.00001078, 0.00000590, 0.00000268, 0.00000695, 0.00000801, 0.00000427, 0.00000056, 0.00000320, 0.00000307, 0.00000608, 0.00000224, 0.00001046, -0.00000412, 0.00000801, -0.00000021, 0.00001214, 0.00000099, 0.00001696, 0.00001558, 0.00000112, -0.00000969, -0.00001879, 0.00000077, -0.00000900, -0.00000162, -0.00000609, -0.00000175, -0.00000558, 0.00000289, -0.00000273, -0.00000089, -0.00000173, -0.00000877, -0.00000288, 0.00000152, -0.00000548, -0.00000646, -0.00000551, -0.00000203, -0.00000246, -0.00000763, -0.00000534, -0.00000351, -0.00001162, -0.00001490, -0.00000885, -0.00003282, -0.00002103, -0.00013398, -0.00004505, 0.00020040, 0.00003366, 0.00006752, 0.00002520, 0.00006549, 0.00002079, 0.00006653, 0.00004490, 0.00018240, -0.00002907, -0.00013196, -0.00001465, -0.00004803, -0.00000525, -0.00001140, -0.00000511, -0.00003053, 0.00000371, -0.00000478, -0.00000582, -0.00000721, 0.00000282, -0.00000278, -0.00000125, -0.00000980, -0.00000529, 0.00001138, -0.00000254, -0.00000227, 0.00000455, -0.00001514, 0.00000831, -0.00001926, 0.00000856, -0.00000538, 0.00000118, -0.00001057, -0.00000083, -0.00000024, -0.00000113, -0.00000494, 0.00000244, -0.00001037, -0.00000275, -0.00000279, -0.00000066, 0.00000335, -0.00000261, -0.00000510, 0.00000078, -0.00000217, 0.00000211, -0.00000353, -0.00000202, },
		{ -0.00001129, 0.00000170, 0.00000013, 0.00000107, -0.00001122, -0.00000095, -0.00000149, -0.00000371, -0.00000011, -0.00000087, -0.00000429, -0.00000949, 0.00000673, -0.00000693, -0.00001809, 0.00000998, -0.00000881, 0.00000733, -0.00002050, -0.00001054, 0.00000143, 0.00000681, 0.00002268, 0.00000686, 0.00001226, -0.00000020, 0.00000698, 0.00000574, 0.00000791, 0.00000261, 0.00000075, 0.00000160, 0.00000553, 0.00000324, 0.00000579, 0.00000726, -0.00000460, 0.00000943, -0.00000119, 0.00001363, -0.00000169, 0.00002317, 0.00000676, 0.00000285, -0.00001633, -0.00001071, -0.00000229, -0.00000677, -0.00000367, -0.00000345, -0.00000421, -0.00000270, 0.00000149, -0.00000168, -0.00000205, -0.00000167, -0.00001077, 0.00000274, -0.00000351, -0.00000385, -0.00000935, -0.00000113, -0.00000529, 0.00000011, -0.00001247, 0.00000165, -0.00001131, -0.00000924, -0.00001750, -0.00000678, -0.00004147, -0.00002464, -0.00017196, 0.00005052, 0.00017639, 0.00002720, 0.00006826, 0.00003351, 0.00005928, 0.00003224, 0.00006245, 0.00010478, 0.00015078, -0.00016423, -0.00004270, -0.00004481, -0.00001902, -0.00001603, 0.00000212, -0.00002572, -0.00001268, 0.00000003, -0.00000301, -0.00001191, 0.00000182, -0.00000201, 0.00000109, -0.00000871, -0.00000440, -0.00000477, 0.00001495, -0.00001202, 0.00000845, -0.00000661, -0.00000616, -0.00000012, -0.00001285, 0.00000861, -0.00000661, -0.00000039, -0.00001036, 0.00000142, -0.00000075, -0.00000295, -0.00000260, 0.00000095, -0.00001145, -0.00000103, -0.00000317, 0.00000146, 0.00000192, -0.00000416, -0.00000339, 0.00000109, -0.00000211, 0.00000130, -0.00000411, -0.00000209, },
		{ -0.00000874, 0.00000269, 0.00000070, -0.00000144, -0.00001173, 0.00000075, -0.00000137, -0.00000379, 0.00000093, -0.00000179, -0.00000692, -0.00000701, 0.00001002, -0.00001868, -0.00000423, 0.00000040, -0.00000064, 0.00000257, -0.00002986, 0.00000665, -0.00000551, 0.00002331, 0.00000914, 0.00001280, 0.00000375, 0.00000268, 0.00000683, 0.00000640, 0.00000593, 0.00000317, -0.00000226, 0.00000423, 0.00000438, 0.00000319, 0.00000811, 0.00000271, -0.00000262, 0.00000924, -0.00000104, 0.00001361, -0.00000206, 0.00002674, -0.00000157, 0.00000494, -0.00002191, -0.00000202, -0.00000772, -0.00000270, -0.00000683, 0.00000023, -0.00000800, 0.00000174, -0.00000132, 0.00000024, -0.00000403, -0.00000152, -0.00001101, 0.00000651, -0.00000896, -0.00000147, -0.00001149, 0.00000247, -0.00000818, 0.00000129, -0.00001434, 0.00000565, -0.00001897, -0.00000598, -0.00001818, -0.00001062, -0.00004324, -0.00004521, -0.00018227, 0.00014627, 0.00012363, 0.00004602, 0.00005482, 0.00005295, 0.00004175, 0.00005670, 0.00005918, 0.00018771, -0.00001288, -0.00015864, -0.00000136, -0.00006179, 0.00000377, -0.00002410, 0.00000662, -0.00003795, 0.00000629, -0.00000927, -0.00000056, -0.00001467, 0.00000914, -0.00000678, 0.00000238, -0.00001306, 0.00000141, -0.00000463, 0.00001500, -0.00001711, 0.00001678, -0.00001756, 0.00000352, -0.00000998, -0.00000191, 0.00000267, -0.00000434, -0.00000430, -0.00000681, 0.00000098, -0.00000042, -0.00000523, 0.00000147, -0.00000350, -0.00000983, -0.00000092, -0.00000157, 0.00000130, 0.00000138, -0.00000605, 0.00000001, -0.00000081, -0.00000064, -0.00000134, -0.00000190, -0.00000549, },
		{ -0.00000565, 0.00000327, 0.00000062, -0.00000410, -0.00001200, 0.00000406, -0.00000309, -0.00000293, 0.00000177, -0.00000403, -0.00000884, -0.00000106, 0.00000550, -0.00002204, 0.00000553, -0.00001112, 0.00001747, -0.00002369, -0.00001105, 0.00000250, 0.00000494, 0.00002288, 0.00000400, 0.00001427, -0.00000273, 0.00000721, 0.00000592, 0.00000551, 0.00000669, -0.00000003, -0.00000172, 0.00000521, 0.00000381, 0.00000331, 0.00001027, -0.00000404, 0.00000413, 0.00000366, 0.00000446, 0.00000792, 0.00000416, 0.00002264, -0.00000369, 0.00000216, -0.00002115, 0.00000099, -0.00000927, -0.00000223, -0.00000580, -0.00000069, -0.00000737, 0.00000237, -0.00000052, -0.00000204, -0.00000173, -0.00000579, -0.00000490, 0.00000217, -0.00000752, -0.00000491, -0.00000667, -0.00000149, -0.00000383, -0.00000537, -0.00000612, -0.00000291, -0.00001520, -0.00001173, -0.00000909, -0.00002882, -0.00002892, -0.00009407, -0.00014045, 0.00020304, 0.00008033, 0.00006168, 0.00005164, 0.00005618, 0.00004564, 0.00005817, 0.00011702, 0.00016749, -0.00017277, -0.00005480, -0.00003768, -0.00002806, -0.00001082, -0.00000680, -0.00001549, -0.00002052, 0.00000248, -0.00000568, -0.00000924, -0.00000355, 0.00000341, -0.00000194, -0.00000685, -0.00000453, -0.00000332, 0.00000365, 0.00000479, -0.00000999, 0.00001361, -0.00001936, 0.00000559, -0.00001213, 0.00000372, -0.00000144, -0.00000415, -0.00000504, -0.00000540, 0.00000215, -0.00000297, -0.00000344, 0.00000202, -0.00000627, -0.00000915, 0.00000128, -0.00000225, 0.00000256, -0.00000124, -0.00000483, 0.00000107, -0.00000144, -0.00000109, -0.00000166, -0.00000148, -0.00000797, },
		{ -0.00000368, 0.00000436, -0.00000091, -0.00000631, -0.00001170, 0.00000896, -0.00000860, 0.00000244, -0.00000200, -0.00000380, -0.00001129, 0.00000761, -0.00000751, -0.00001081, -0.00000320, -0.00000099, 0.00001357, -0.00003741, 0.00001076, -0.00000921, 0.00002479, 0.00000670, 0.00001261, 0.00000395, 0.00000103, 0.00000670, 0.00000709, 0.00000279, 0.00000962, -0.00000697, 0.00000406, 0.00000188, 0.00000620, 0.00000232, 0.00001140, -0.00000996, 0.00001221, -0.00000467, 0.00001334, -0.00000170, 0.00001534, 0.00001178, 0.00000034, -0.00000507, -0.00001496, -0.00000143, -0.00000694, -0.00000471, -0.00000179, -0.00000512, -0.00000315, 0.00000028, 0.00000202, -0.00000616, 0.00000233, -0.00001107, 0.00000292, -0.00000609, -0.00000219, -0.00001077, 0.00000075, -0.00000853, 0.00000294, -0.00001364, 0.00000436, -0.00001684, -0.00000517, -0.00002097, 0.00000092, -0.00005001, -0.00001260, -0.00015347, -0.00005715, 0.00021613, 0.00005665, 0.00007046, 0.00005475, 0.00005153, 0.00006023, 0.00006129, 0.00020435, -0.00000030, -0.00018286, 0.00000470, -0.00006764, 0.00000572, -0.00002847, 0.00000716, -0.00003516, 0.00000322, -0.00001067, 0.00000139, -0.00001758, 0.00000934, -0.00000560, 0.00000308, -0.00001536, 0.00000680, -0.00001178, 0.00001393, -0.00000857, 0.00000314, 0.00000295, -0.00001489, 0.00000329, -0.00000912, 0.00000441, -0.00000377, -0.00000449, -0.00000444, -0.00000484, 0.00000314, -0.00000581, 0.00000022, -0.00000003, -0.00000791, -0.00000792, 0.00000370, -0.00000421, 0.00000459, -0.00000477, -0.00000140, -0.00000015, -0.00000097, -0.00000279, 0.00000024, -0.00000383, -0.00000777, },
		{ -0.00000042, 0.00000319, -0.00000101, -0.00001107, -0.00000592, 0.00000862, -0.00001117, 0.00000686, -0.00000672, -0.00000360, -0.00000987, 0.00001013, -0.00001561, -0.00000065, -0.00001589, 0.00002227, -0.00001671, -0.00001835, 0.00000767, -0.00000156, 0.00002746, -0.00000129, 0.00001718, -0.00000515, 0.00000714, 0.00000533, 0.00000613, 0.00000401, 0.00000721, -0.00000924, 0.00000763, -0.00000012, 0.00000732, 0.00000351, 0.00000751, -0.00000890, 0.00001376, -0.00000742, 0.00001735, -0.00000714, 0.00002348, 0.00000151, 0.00000453, -0.00001112, -0.00000945, -0.00000351, -0.00000587, -0.00000471, -0.00000086, -0.00000682, -0.00000114, 0.00000098, 0.00000030, -0.00000551, 0.00000152, -0.00001029, 0.00000376, -0.00000875, -0.00000172, -0.00001045, 0.00000132, -0.00000881, 0.00000229, -0.00001261, 0.00000352, -0.00002076, -0.00000316, -0.00002034, -0.00000402, -0.00005480, -0.00001825, -0.00019141, 0.00003543, 0.00020651, 0.00003968, 0.00008727, 0.00004206, 0.00006935, 0.00004896, 0.00013189, 0.00018511, -0.00017642, -0.00007545, -0.00002580, -0.00004000, -0.00000252, -0.00001737, -0.00000783, -0.00002418, 0.00000382, -0.00001016, -0.00000410, -0.00001063, 0.00000844, -0.00000458, -0.00000324, -0.00000971, 0.00000598, -0.00001097, 0.00001487, -0.00001341, 0.00001081, -0.00000586, -0.00001127, 0.00000367, -0.00000840, 0.00000581, -0.00000832, -0.00000087, -0.00000744, -0.00000078, -0.00000024, -0.00000391, 0.00000037, -0.00000036, -0.00001168, -0.00000231, 0.00000144, -0.00000272, 0.00000332, -0.00000490, -0.00000023, -0.00000001, -0.00000277, -0.00000154, -0.00000023, -0.00000486, -0.00000795, },
		{ 0.00000431, -0.00000035, 0.00000012, -0.00001729, 0.00000530, 0.00000005, -0.00000548, 0.00000404, -0.00000697, -0.00000638, -0.00000280, 0.00000251, -0.00000952, -0.00000846, -0.00000633, 0.00002382, -0.00003738, 0.00000808, -0.00000800, 0.00002051, 0.00001063, 0.00000732, 0.00000772, -0.00000261, 0.00000741, 0.00000692, 0.00000194, 0.00000961, -0.00000196, -0.00000335, 0.00000462, 0.00000297, 0.00000467, 0.00000745, -0.00000059, -0.00000107, 0.00000741, -0.00000202, 0.00001316, -0.00000454, 0.00002425, -0.00000401, 0.00000611, -0.00001438, -0.00000610, -0.00000418, -0.00000631, -0.00000243, -0.00000310, -0.00000542, -0.00000120, 0.00000338, -0.00000428, -0.00000119, -0.00000271, -0.00000562, -0.00000082, -0.00000626, -0.00000520, -0.00000571, -0.00000321, -0.00000410, -0.00000299, -0.00000615, -0.00000585, -0.00001511, -0.00000863, -0.00001275, -0.00002066, -0.00004490, -0.00004696, -0.00019907, 0.00012550, 0.00017710, 0.00003747, 0.00009820, 0.00002973, 0.00009264, 0.00004134, 0.00023898, 0.00000192, -0.00019499, -0.00000585, -0.00005916, -0.00000457, -0.00002048, -0.00000374, -0.00002557, -0.00000286, -0.00000799, -0.00000326, -0.00001257, 0.00000164, 0.00000124, -0.00000088, -0.00001160, 0.00000223, -0.00000298, -0.00000333, 0.00000899, -0.00001112, 0.00001299, -0.00001317, -0.00000597, 0.00000342, -0.00000794, 0.00000576, -0.00001142, 0.00000298, -0.00001116, 0.00000392, -0.00000485, 0.00000024, -0.00000163, -0.00000049, -0.00001477, 0.00000473, -0.00000402, 0.00000159, -0.00000050, -0.00000212, -0.00000138, 0.00000114, -0.00000587, 0.00000223, -0.00000372, -0.00000335, -0.00000941, },
		{ 0.00000612, -0.00000175, -0.00000260, -0.00001759, 0.00001297, -0.00000885, 0.00000283, -0.00000277, -0.00000487, -0.00000800, 0.00000181, -0.00000358, -0.00000282, -0.00002169, 0.00001977, -0.00000505, -0.00002309, 0.00000910, -0.00000533, 0.00002806, -0.00000102, 0.00001471, -0.00000314, 0.00000426, 0.00000571, 0.00000697, 0.00000104, 0.00001124, -0.00000890, 0.00000315, 0.00000059, 0.00000679, 0.00000211, 0.00000935, -0.00000628, 0.00000602, 0.00000027, 0.00000536, 0.00000588, 0.00000212, 0.00002025, -0.00000631, 0.00000692, -0.00001771, -0.00000160, -0.00000707, -0.00000385, -0.00000308, -0.00000317, -0.00000584, 0.00000163, 0.00000155, -0.00000496, 0.00000017, -0.00000368, -0.00000492, -0.00000240, -0.00000633, -0.00000496, -0.00000509, -0.00000395, -0.00000326, -0.00000327, -0.00000628, -0.00001027, -0.00001204, -0.00001041, -0.00001156, -0.00003358, -0.00003596, -0.00008462, -0.00018334, 0.00021543, 0.00011886, 0.00007002, 0.00007347, 0.00005707, 0.00007400, 0.00011820, 0.00022860, -0.00019714, -0.00008138, -0.00003257, -0.00003425, -0.00001013, -0.00001157, -0.00001688, -0.00001502, -0.00000258, -0.00000776, -0.00000749, -0.00000759, 0.00000250, 0.00000185, -0.00000683, -0.00000755, 0.00000464, -0.00000686, 0.00000075, 0.00000551, -0.00000944, 0.00001483, -0.00002256, 0.00000592, -0.00000424, -0.00000152, -0.00000116, -0.00000668, -0.00000005, -0.00000900, 0.00000307, -0.00000421, 0.00000023, -0.00000043, -0.00000512, -0.00001131, 0.00000612, -0.00000612, 0.00000336, -0.00000181, -0.00000133, -0.00000067, -0.00000087, -0.00000510, 0.00000291, -0.00000516, -0.00000347, -0.00000826, },
		{ 0.00000356, 0.00000025, -0.00001013, -0.00000898, 0.00001067, -0.00000970, 0.00000526, -0.00000667, -0.00000461, -0.00000517, -0.00000146, 0.00000209, -0.00001330, -0.00001262, 0.00002613, -0.00002855, 0.00000271, -0.00000623, 0.00001503, 0.00001449, 0.00000431, 0.00000833, -0.00000312, 0.00000607, 0.00000674, 0.00000324, 0.00000559, 0.00000527, -0.00000794, 0.00000409, 0.00000080, 0.00000726, 0.00000215, 0.00000751, -0.00000699, 0.00000853, -0.00000320, 0.00001021, -0.00000043, 0.00000972, 0.00001307, -0.00000494, 0.00000531, -0.00001943, 0.00000245, -0.00001046, -0.00000002, -0.00000574, -0.00000167, -0.00000692, 0.00000548, -0.00000288, -0.00000259, -0.00000040, -0.00000331, -0.00000652, -0.00000180, -0.00000762, -0.00000308, -0.00000665, -0.00000277, -0.00000348, -0.00000244, -0.00001006, -0.00001046, -0.00001065, -0.00001155, -0.00001386, -0.00004415, -0.00002658, -0.00013573, -0.00012859, 0.00027894, 0.00005848, 0.00011426, 0.00003408, 0.00010450, 0.00004339, 0.00025457, 0.00002410, -0.00022089, -0.00000683, -0.00006339, -0.00000193, -0.00002426, -0.00000188, -0.00003079, 0.00000386, -0.00001297, -0.00000259, -0.00001346, 0.00000134, -0.00000101, 0.00000308, -0.00001338, 0.00000175, 0.00000035, -0.00000697, 0.00000350, 0.00000181, -0.00000539, 0.00001145, -0.00002643, 0.00001658, -0.00001390, 0.00000648, -0.00001008, 0.00000179, -0.00000732, -0.00000310, -0.00000105, -0.00000018, -0.00000276, 0.00000240, -0.00001175, -0.00000321, 0.00000178, -0.00000387, 0.00000205, -0.00000030, -0.00000296, 0.00000176, -0.00000545, -0.00000026, -0.00000088, -0.00000263, -0.00000691, -0.00000321, },
		{ 0.00000123, 0.00000040, -0.00001532, 0.00000093, 0.00000385, -0.00000535, 0.00000252, -0.00000786, -0.00000346, -0.00000466, -0.00000229, 0.00000688, -0.00002942, 0.00001534, 0.00000034, -0.00001759, 0.00000438, -0.00000512, 0.00002385, 0.00000341, 0.00001032, -0.00000077, 0.00000191, 0.00000571, 0.00000653, 0.00000187, 0.00000800, -0.00000079, -0.00000470, 0.00000317, 0.00000251, 0.00000661, 0.00000201, 0.00000590, -0.00000692, 0.00000991, -0.00000549, 0.00001408, -0.00000664, 0.00001886, 0.00000157, 0.00000286, -0.00000323, -0.00001406, 0.00000015, -0.00000809, -0.00000141, -0.00000403, -0.00000445, -0.00000263, 0.00000338, -0.00000256, -0.00000378, 0.00000332, -0.00000896, -0.00000229, -0.00000663, -0.00000255, -0.00000796, -0.00000174, -0.00000804, 0.00000420, -0.00001174, -0.00000492, -0.00001741, -0.00000120, -0.00002328, -0.00000771, -0.00006240, -0.00000896, -0.00020972, -0.00001158, 0.00027531, 0.00004430, 0.00011862, 0.00003812, 0.00010620, 0.00010266, 0.00027437, -0.00021599, -0.00008740, -0.00004290, -0.00002791, -0.00001682, -0.00000633, -0.00002382, -0.00001177, -0.00000250, -0.00000790, -0.00001133, -0.00000365, -0.00000328, 0.00000576, -0.00000802, -0.00000544, 0.00000040, 0.00000179, -0.00001211, 0.00001254, -0.00000920, 0.00000727, -0.00000388, -0.00001507, 0.00001383, -0.00001424, 0.00000626, -0.00001103, 0.00000401, -0.00001014, -0.00000099, -0.00000110, -0.00000003, -0.00000224, 0.00000058, -0.00001300, 0.00000157, -0.00000163, -0.00000229, 0.00000234, -0.00000078, -0.00000242, 0.00000104, -0.00000670, 0.00000246, -0.00000360, -0.00000082, -0.00000930, 0.00000113, },
		{ 0.00000162, -0.00000387, -0.00001342, 0.00000475, 0.00000062, -0.00000320, 0.00000006, -0.00000921, 0.00000094, -0.00001133, 0.00000945, -0.00000791, -0.00002170, 0.00002475, -0.00002234, 0.00000610, -0.00001138, 0.00001470, 0.00001192, 0.00000815, 0.00000371, -0.00000005, 0.00000271, 0.00000772, 0.00000269, 0.00000544, 0.00000426, -0.00000145, -0.00000447, 0.00000455, 0.00000277, 0.00000620, 0.00000136, 0.00000501, -0.00000691, 0.00001085, -0.00000670, 0.00001602, -0.00001050, 0.00002601, -0.00000980, 0.00001273, -0.00001540, -0.00000385, -0.00000671, -0.00000155, -0.00000680, 0.00000085, -0.00000964, 0.00000458, -0.00000259, 0.00000096, -0.00000617, 0.00000744, -0.00001660, 0.00000456, -0.00001327, 0.00000446, -0.00001511, 0.00000497, -0.00001374, 0.00001215, -0.00002412, 0.00000394, -0.00002515, 0.00000735, -0.00003546, -0.00000423, -0.00007252, -0.00000811, -0.00027304, 0.00014082, 0.00021498, 0.00007315, 0.00008727, 0.00008068, 0.00007780, 0.00024642, 0.00006885, -0.00026431, 0.00000813, -0.00008603, 0.00001619, -0.00004104, 0.00001273, -0.00004715, 0.00001676, -0.00002087, 0.00000293, -0.00002202, 0.00000988, -0.00001178, 0.00001297, -0.00002007, 0.00000788, -0.00000674, 0.00000480, -0.00001608, 0.00002039, -0.00001970, 0.00001987, -0.00002169, 0.00000413, 0.00000097, -0.00000675, -0.00000037, -0.00000525, 0.00000004, -0.00000848, -0.00000189, 0.00000194, -0.00000322, 0.00000110, -0.00000481, -0.00000893, 0.00000140, -0.00000222, -0.00000221, 0.00000444, -0.00000367, 0.00000052, -0.00000306, -0.00000336, 0.00000063, -0.00000229, -0.00000265, -0.00000794, 0.00000203, },
		{ 0.00000060, -0.00000707, -0.00000974, 0.00000577, -0.00000056, -0.00000286, -0.00000276, -0.00000704, 0.00000068, -0.00001357, 0.00002010, -0.00003096, 0.00000861, -0.00000006, -0.00000992, 0.00000457, -0.00000993, 0.00002336, 0.00000250, 0.00001279, -0.00000470, 0.00000464, 0.00000241, 0.00000764, 0.00000171, 0.00000644, 0.00000154, -0.00000215, -0.00000326, 0.00000487, 0.00000440, 0.00000269, 0.00000486, -0.00000047, -0.00000159, 0.00000603, -0.00000153, 0.00001040, -0.00000547, 0.00002337, -0.00001179, 0.00001439, -0.00002081, 0.00000147, -0.00000897, 0.00000009, -0.00000737, 0.00000045, -0.00000849, 0.00000522, -0.00000289, -0.00000110, -0.00000091, 0.00000123, -0.00001413, 0.00000225, -0.00000991, 0.00000070, -0.00001138, 0.00000033, -0.00000567, 0.00000318, -0.00002047, -0.00000150, -0.00001622, -0.00000490, -0.00002595, -0.00002648, -0.00004723, -0.00006474, -0.00026432, 0.00026070, 0.00015552, 0.00009254, 0.00007954, 0.00009211, 0.00012967, 0.00028793, -0.00020508, -0.00011916, -0.00003014, -0.00004693, -0.00000020, -0.00002186, -0.00001016, -0.00002846, 0.00001270, -0.00001797, -0.00000551, -0.00001144, 0.00000429, -0.00000431, 0.00000247, -0.00001307, 0.00000641, -0.00000346, -0.00000349, -0.00000602, 0.00001334, -0.00001582, 0.00001790, -0.00002555, 0.00001420, -0.00000802, -0.00000233, -0.00000325, -0.00000235, -0.00000212, -0.00000911, 0.00000110, 0.00000074, -0.00000283, 0.00000060, -0.00000652, -0.00000672, 0.00000190, -0.00000427, 0.00000084, 0.00000323, -0.00000385, 0.00000060, -0.00000463, -0.00000124, -0.00000079, -0.00000103, -0.00000457, -0.00000628, 0.00000294, },
		{ -0.00000375, -0.00000550, -0.00000976, 0.00000978, -0.00000518, -0.00000014, -0.00000827, 0.00000133, -0.00001039, 0.00000110, 0.00000764, -0.00002950, 0.00002221, -0.00002367, 0.00001659, -0.00001685, 0.00001320, 0.00000948, 0.00001027, 0.00000310, -0.00000159, 0.00000356, 0.00000595, 0.00000263, 0.00000660, 0.00000079, 0.00000484, -0.00000748, 0.00000254, 0.00000166, 0.00000808, -0.00000324, 0.00001175, -0.00001021, 0.00000897, -0.00000449, 0.00000963, -0.00000213, 0.00000788, 0.00001102, -0.00000362, 0.00000599, -0.00001690, -0.00000038, -0.00000507, -0.00000443, -0.00000217, -0.00000539, -0.00000138, -0.00000050, 0.00000223, -0.00000717, 0.00000880, -0.00001244, -0.00000298, -0.00000762, 0.00000099, -0.00001123, 0.00000040, -0.00001141, 0.00001040, -0.00001795, -0.00000325, -0.00001803, 0.00000323, -0.00003077, -0.00000258, -0.00006315, -0.00000195, -0.00016946, -0.00017013, 0.00031923, 0.00011793, 0.00009386, 0.00009980, 0.00007868, 0.00027218, 0.00008874, -0.00028645, 0.00000460, -0.00009253, 0.00001524, -0.00003899, 0.00001012, -0.00004562, 0.00001052, -0.00001276, -0.00000184, -0.00002230, 0.00000911, -0.00001148, 0.00001010, -0.00001546, 0.00000515, -0.00000594, 0.00000652, -0.00001716, 0.00001171, -0.00000362, -0.00000159, 0.00000472, -0.00001625, 0.00001264, -0.00001033, -0.00000191, -0.00000204, -0.00000301, -0.00000244, -0.00001071, 0.00000643, -0.00000396, 0.00000067, -0.00000311, -0.00000449, -0.00000741, 0.00000390, -0.00000753, 0.00000637, -0.00000148, -0.00000083, -0.00000279, -0.00000217, -0.00000278, 0.00000103, -0.00000256, -0.00000426, -0.00000633, 0.00000553, },
		{ -0.00000538, -0.00000515, -0.00000860, 0.00001289, -0.00001036, 0.00000202, -0.00001066, 0.00000622, -0.00001968, 0.00001948, -0.00001911, -0.00000065, -0.00000113, -0.00000997, 0.00001531, -0.00002004, 0.00002603, -0.00000266, 0.00001849, -0.00000931, 0.00000657, 0.00000051, 0.00000804, 0.00000010, 0.00000888, -0.00000317, 0.00000707, -0.00001138, 0.00000751, -0.00000016, 0.00000800, -0.00000388, 0.00001296, -0.00001460, 0.00001512, -0.00001118, 0.00001711, -0.00001152, 0.00001940, -0.00000104, 0.00000497, -0.00000346, -0.00001166, -0.00000168, -0.00000335, -0.00000638, -0.00000002, -0.00000736, 0.00000154, -0.00000214, 0.00000252, -0.00000610, 0.00000932, -0.00001844, 0.00000230, -0.00001067, 0.00000393, -0.00001506, 0.00000360, -0.00001198, 0.00001264, -0.00002728, 0.00000460, -0.00002284, 0.00000693, -0.00004045, 0.00000219, -0.00007527, 0.00001127, -0.00026196, -0.00003539, 0.00034132, 0.00007369, 0.00012460, 0.00008226, 0.00015437, 0.00031172, -0.00020085, -0.00014759, -0.00001866, -0.00006671, 0.00001367, -0.00003304, -0.00000048, -0.00003950, 0.00001879, -0.00001925, -0.00000369, -0.00001822, 0.00001087, -0.00001192, 0.00000792, -0.00001593, 0.00001009, -0.00000878, 0.00000572, -0.00001833, 0.00001867, -0.00001317, 0.00000709, -0.00000410, -0.00000802, 0.00000989, -0.00001368, 0.00000235, -0.00000492, 0.00000007, -0.00000775, -0.00000553, 0.00000549, -0.00000411, -0.00000005, -0.00000254, -0.00000610, -0.00000468, 0.00000175, -0.00000583, 0.00000765, -0.00000344, -0.00000054, -0.00000347, -0.00000166, -0.00000276, 0.00000169, -0.00000313, -0.00000541, -0.00000400, 0.00000490, },
		{ -0.00000052, -0.00001155, 0.00000038, 0.00000774, -0.00000953, -0.00000131, -0.00000426, -0.00000212, -0.00001065, 0.00001582, -0.00002502, 0.00001582, -0.00002695, 0.00002191, -0.00001151, 0.00000383, 0.00001287, 0.00000512, 0.00000983, -0.00000838, 0.00000777, 0.00000231, 0.00000406, 0.00000485, 0.00000272, 0.00000140, 0.00000145, -0.00000769, 0.00000660, 0.00000255, 0.00000221, 0.00000328, 0.00000469, -0.00000866, 0.00001128, -0.00000834, 0.00001512, -0.00001171, 0.00002300, -0.00000712, 0.00000868, -0.00000917, -0.00000858, -0.00000013, -0.00000600, -0.00000348, -0.00000276, -0.00000399, -0.00000128, 0.00000177, -0.00000255, 0.00000182, -0.00000032, -0.00001389, -0.00000095, -0.00000522, -0.00000275, -0.00000907, -0.00000246, -0.00000156, -0.00000075, -0.00002002, -0.00000134, -0.00001383, -0.00000744, -0.00003018, -0.00001516, -0.00005783, -0.00002500, -0.00030760, 0.00011286, 0.00032682, 0.00003670, 0.00017386, 0.00004158, 0.00033259, 0.00008868, -0.00029346, -0.00001455, -0.00008241, -0.00000446, -0.00002158, -0.00000507, -0.00003196, -0.00000585, 0.00000038, -0.00000893, -0.00001747, 0.00000011, -0.00000200, -0.00000120, -0.00000537, -0.00000210, 0.00000127, -0.00000167, -0.00000453, -0.00000781, 0.00001370, -0.00001354, 0.00000909, -0.00000694, -0.00000348, 0.00000707, -0.00001647, 0.00000824, -0.00001033, 0.00000501, -0.00001535, 0.00000427, -0.00000103, 0.00000010, -0.00000442, 0.00000197, -0.00001172, 0.00000192, -0.00000450, 0.00000106, 0.00000348, -0.00000128, -0.00000406, 0.00000016, -0.00000541, 0.00000156, -0.00000146, -0.00000070, -0.00000962, 0.00000226, -0.00000110, },
		{ 0.00000330, -0.00001810, 0.00001208, -0.00000272, -0.00000498, -0.00000523, 0.00000259, -0.00001342, 0.00000719, -0.00000593, -0.00000205, -0.00000508, -0.00001479, 0.00002607, -0.00002286, 0.00002191, -0.00000257, 0.00001708, -0.00000544, 0.00000102, 0.00000446, 0.00000457, 0.00000087, 0.00000879, -0.00000384, 0.00000770, -0.00000653, -0.00000087, 0.00000377, 0.00000457, -0.00000198, 0.00000983, -0.00000492, 0.00000044, 0.00000352, -0.00000138, 0.00000852, -0.00000680, 0.00002222, -0.00001093, 0.00001158, -0.00001543, -0.00000232, -0.00000266, -0.00000532, -0.00000349, -0.00000223, -0.00000428, -0.00000031, 0.00000154, -0.00000224, 0.00000365, -0.00000646, -0.00001111, -0.00000067, -0.00000462, -0.00000485, -0.00000775, -0.00000170, 0.00000053, -0.00000908, -0.00001498, -0.00000273, -0.00001249, -0.00001566, -0.00002512, -0.00002619, -0.00004519, -0.00008084, -0.00031630, 0.00028244, 0.00023921, 0.00007465, 0.00015311, 0.00011761, 0.00039456, -0.00024028, -0.00014391, -0.00003543, -0.00005916, -0.00000200, -0.00001720, -0.00001477, -0.00002715, 0.00000279, -0.00000388, -0.00001341, -0.00001306, 0.00000299, -0.00000341, -0.00000304, -0.00000510, 0.00000201, -0.00000148, -0.00000142, -0.00000750, -0.00000167, 0.00001025, -0.00001585, 0.00001527, -0.00001444, 0.00000615, -0.00000389, -0.00000897, 0.00000491, -0.00000798, 0.00000184, -0.00001443, 0.00000758, -0.00000330, -0.00000004, -0.00000357, 0.00000128, -0.00001286, 0.00000458, -0.00000700, 0.00000538, 0.00000069, -0.00000133, -0.00000465, 0.00000127, -0.00000709, 0.00000460, -0.00000356, -0.00000028, -0.00001119, 0.00000598, -0.00000507, },
		{ -0.00000207, -0.00001483, 0.00001585, -0.00000942, -0.00000337, -0.00000245, -0.00000167, -0.00000962, 0.00000759, -0.00001146, 0.00001190, -0.00003145, 0.00001975, -0.00000017, -0.00000418, 0.00001364, 0.00000148, 0.00001310, -0.00000925, 0.00000570, 0.00000415, 0.00000191, 0.00000436, 0.00000466, -0.00000233, 0.00000716, -0.00000912, 0.00000287, 0.00000312, 0.00000266, -0.00000022, 0.00000978, -0.00000896, 0.00000595, -0.00000220, 0.00000460, 0.00000151, 0.00000051, 0.00001801, -0.00001250, 0.00001335, -0.00002131, 0.00000607, -0.00000926, -0.00000058, -0.00000698, 0.00000177, -0.00000853, 0.00000497, -0.00000292, 0.00000318, -0.00000173, -0.00000645, -0.00001186, 0.00000352, -0.00000934, -0.00000171, -0.00001078, 0.00000488, -0.00000683, -0.00000866, -0.00001476, 0.00000029, -0.00001883, -0.00001641, -0.00002558, -0.00003098, -0.00004302, -0.00015068, -0.00026914, 0.00043297, 0.00011057, 0.00017524, 0.00007015, 0.00034266, 0.00014281, -0.00034678, 0.00000045, -0.00009994, 0.00000088, -0.00003155, 0.00000609, -0.00004227, 0.00000176, -0.00001083, 0.00000418, -0.00002665, 0.00000341, -0.00000627, 0.00000320, -0.00001251, 0.00000599, -0.00000434, 0.00000252, -0.00000702, -0.00000387, -0.00000015, 0.00000740, -0.00001682, 0.00002173, -0.00002360, 0.00001734, -0.00001837, 0.00000523, -0.00000592, 0.00000092, -0.00000813, -0.00000524, 0.00000319, -0.00000011, -0.00000489, 0.00000300, -0.00000623, -0.00000725, 0.00000121, -0.00000347, 0.00000434, 0.00000163, -0.00000535, -0.00000001, -0.00000338, -0.00000278, 0.00000239, -0.00000127, -0.00000493, -0.00000660, 0.00000282, -0.00000104, },
		{ -0.00001166, -0.00000406, 0.00001112, -0.00001147, -0.00000227, 0.00000089, -0.00000940, 0.00000295, -0.00000867, 0.00001004, -0.00001083, -0.00002010, 0.00002857, -0.00001532, 0.00001334, -0.00000009, 0.00001212, 0.00000060, -0.00000383, 0.00000619, 0.00000367, 0.00000030, 0.00000742, -0.00000084, 0.00000226, 0.00000306, -0.00000845, 0.00000555, 0.00000105, 0.00000235, 0.00000156, 0.00000801, -0.00001091, 0.00001040, -0.00000813, 0.00001191, -0.00000834, 0.00001358, 0.00000494, -0.00000453, 0.00000569, -0.00001725, 0.00000595, -0.00000935, -0.00000170, -0.00000394, -0.00000153, -0.00000539, 0.00000316, 0.00000037, 0.00000014, -0.00000053, -0.00001276, -0.00000345, -0.00000280, -0.00000407, -0.00000887, -0.00000107, -0.00000283, -0.00000261, -0.00001855, -0.00000038, -0.00001475, -0.00000792, -0.00003501, -0.00000354, -0.00006298, -0.00001913, -0.00026455, -0.00011060, 0.00046955, 0.00005646, 0.00021393, 0.00009810, 0.00046930, -0.00026150, -0.00016240, -0.00003623, -0.00006223, -0.00000984, -0.00001217, -0.00001724, -0.00002567, -0.00000119, -0.00000320, -0.00000988, -0.00001555, -0.00000025, 0.00000014, -0.00000720, -0.00000370, 0.00000274, -0.00000043, -0.00000418, -0.00000158, -0.00000950, 0.00001032, -0.00000747, -0.00000200, 0.00001231, -0.00001887, 0.00001502, -0.00002039, 0.00001027, -0.00001015, 0.00000351, -0.00001212, 0.00000011, 0.00000144, -0.00000077, -0.00000462, 0.00000499, -0.00001115, -0.00000288, -0.00000077, -0.00000099, 0.00000423, 0.00000007, -0.00000626, 0.00000251, -0.00000707, 0.00000188, -0.00000032, 0.00000049, -0.00000913, -0.00000175, -0.00000077, 0.00000455, },
		{ -0.00001488, 0.00000409, 0.00000512, -0.00001337, 0.00000409, -0.00000586, -0.00000332, -0.00000086, -0.00000854, 0.00002181, -0.00004001, 0.00001636, 0.00000483, -0.00000074, 0.00000648, 0.00000374, 0.00000888, -0.00000281, -0.00000125, 0.00000880, -0.00000164, 0.00000539, 0.00000294, 0.00000044, 0.00000223, 0.00000155, -0.00000856, 0.00000913, -0.00000387, 0.00000621, -0.00000027, 0.00000797, -0.00001299, 0.00001453, -0.00001380, 0.00001889, -0.00001799, 0.00002843, -0.00001369, 0.00001111, -0.00001034, -0.00000344, -0.00000381, -0.00000148, -0.00000935, 0.00000548, -0.00001200, 0.00000528, -0.00000537, 0.00001034, -0.00001104, 0.00000768, -0.00002440, 0.00001180, -0.00001758, 0.00000927, -0.00002314, 0.00001727, -0.00002240, 0.00001252, -0.00003605, 0.00002221, -0.00004162, 0.00001430, -0.00006252, 0.00002875, -0.00011103, 0.00001094, -0.00039359, 0.00013959, 0.00038007, 0.00011066, 0.00015761, 0.00031565, 0.00023771, -0.00042964, 0.00003192, -0.00013194, 0.00002221, -0.00005923, 0.00003116, -0.00006342, 0.00001957, -0.00003003, 0.00002040, -0.00003698, 0.00001142, -0.00001855, 0.00001630, -0.00002706, 0.00001813, -0.00001319, 0.00001199, -0.00001784, 0.00001093, -0.00002051, 0.00002431, -0.00002727, 0.00002263, -0.00001030, -0.00000130, -0.00000013, -0.00000904, 0.00000356, -0.00000526, -0.00000209, -0.00000785, -0.00000123, 0.00000445, -0.00000640, 0.00000235, -0.00000075, -0.00000965, -0.00000291, 0.00000127, -0.00000218, 0.00000709, -0.00000576, -0.00000123, -0.00000124, -0.00000481, 0.00000196, 0.00000060, -0.00000228, -0.00000826, -0.00000192, 0.00000199, 0.00000262, },
		{ -0.00001182, 0.00000755, -0.00000117, -0.00001166, 0.00000852, -0.00001464, 0.00001034, -0.00002069, 0.00001988, -0.00000603, -0.00002963, 0.00002792, -0.00000993, 0.00001383, -0.00000555, 0.00001350, -0.00000176, 0.00000098, -0.00000046, 0.00000925, -0.00000464, 0.00000928, -0.00000202, 0.00000325, 0.00000131, -0.00000066, -0.00000550, 0.00000768, -0.00000393, 0.00000633, 0.00000137, 0.00000313, -0.00000837, 0.00001059, -0.00001037, 0.00001552, -0.00001544, 0.00003058, -0.00002220, 0.00001909, -0.00001964, 0.00000524, -0.00000924, 0.00000114, -0.00000990, 0.00000643, -0.00001396, 0.00000784, -0.00000439, 0.00000887, -0.00001147, 0.00000445, -0.00002179, 0.00001211, -0.00001800, 0.00000730, -0.00001901, 0.00001512, -0.00002315, 0.00000869, -0.00002984, 0.00001677, -0.00004017, 0.00000530, -0.00005031, 0.00001201, -0.00010773, -0.00003229, -0.00043013, 0.00035785, 0.00028138, 0.00015574, 0.00017859, 0.00047559, -0.00021523, -0.00024153, 0.00000849, -0.00010342, 0.00001852, -0.00004426, 0.00001372, -0.00005359, 0.00002273, -0.00002814, 0.00001332, -0.00003276, 0.00001218, -0.00001473, 0.00000852, -0.00002193, 0.00001977, -0.00001362, 0.00000841, -0.00001459, 0.00000730, -0.00001501, 0.00001974, -0.00002769, 0.00003192, -0.00002355, 0.00000955, -0.00000986, -0.00000084, -0.00000097, -0.00000357, -0.00000380, -0.00000677, 0.00000094, 0.00000233, -0.00000630, 0.00000515, -0.00000489, -0.00000912, -0.00000045, 0.00000028, -0.00000019, 0.00000573, -0.00000789, 0.00000195, -0.00000464, -0.00000175, 0.00000157, 0.00000086, -0.00000482, -0.00000716, -0.00000183, 0.00000598, -0.00000371, },
		{ -0.00001009, 0.00001098, -0.00001085, -0.00000049, -0.00000080, -0.00000672, 0.00000484, -0.00002170, 0.00003847, -0.00004363, 0.00001026, 0.00000415, 0.00000607, 0.00000323, 0.00000134, 0.00000806, -0.00000192, -0.00000021, 0.00000534, 0.00000184, 0.00000160, 0.00000374, 0.00000130, -0.00000019, 0.00000542, -0.00000751, 0.00000318, -0.00000113, 0.00000429, -0.00000136, 0.00001007, -0.00000921, 0.00000502, -0.00000327, 0.00000395, 0.00000007, 0.00000175, 0.00001544, -0.00001398, 0.00001242, -0.00001534, 0.00000196, -0.00000439, -0.00000600, 0.00000024, -0.00000483, -0.00000325, -0.00000119, 0.00000851, -0.00000712, 0.00000291, -0.00001341, -0.00000229, -0.00000595, -0.00000038, -0.00001249, 0.00000546, -0.00001155, 0.00000109, -0.00001888, 0.00000272, -0.00002034, -0.00000537, -0.00003838, 0.00000448, -0.00006257, -0.00003803, -0.00017240, -0.00032358, 0.00047206, 0.00024168, 0.00013757, 0.00038524, 0.00026781, -0.00045285, 0.00000447, -0.00011615, 0.00000394, -0.00005047, 0.00001838, -0.00004874, 0.00000482, -0.00001763, 0.00000588, -0.00002172, 0.00000033, -0.00001309, 0.00000987, -0.00002006, 0.00000754, -0.00000179, 0.00000394, -0.00001123, 0.00000518, -0.00001244, 0.00000648, -0.00000294, -0.00000633, 0.00001939, -0.00001988, 0.00000740, -0.00000834, -0.00000143, 0.00000085, -0.00000771, 0.00000074, -0.00001132, 0.00000839, -0.00000620, 0.00000177, -0.00000006, -0.00000348, -0.00001211, 0.00000624, -0.00000564, 0.00000683, -0.00000207, -0.00000274, -0.00000151, -0.00000244, -0.00000262, 0.00000440, -0.00000308, -0.00000310, -0.00001026, 0.00000444, 0.00000302, -0.00000543, },
		{ -0.00000939, 0.00001211, -0.00001748, 0.00001114, -0.00001541, 0.00001247, -0.00002141, 0.00001158, 0.00001234, -0.00004021, 0.00002669, -0.00001236, 0.00002307, -0.00001345, 0.00001555, -0.00000602, 0.00000601, -0.00000437, 0.00001072, -0.00000542, 0.00000867, -0.00000363, 0.00000679, -0.00000461, 0.00000903, -0.00001244, 0.00000906, -0.00000734, 0.00001051, -0.00000684, 0.00001560, -0.00001803, 0.00001537, -0.00001457, 0.00001602, -0.00001326, 0.00001842, -0.00000261, -0.00000157, 0.00000282, -0.00000916, -0.00000161, -0.00000138, -0.00000933, 0.00000568, -0.00001193, 0.00000364, -0.00000429, 0.00001323, -0.00001608, 0.00001057, -0.00002317, 0.00000941, -0.00001678, 0.00000886, -0.00002108, 0.00001724, -0.00002776, 0.00001569, -0.00003360, 0.00001985, -0.00004317, 0.00001381, -0.00006063, 0.00003437, -0.00011989, 0.00001681, -0.00032259, -0.00014733, 0.00054526, 0.00017515, 0.00022076, 0.00053273, -0.00020035, -0.00029578, 0.00002406, -0.00012027, 0.00002832, -0.00006040, 0.00002596, -0.00006216, 0.00002860, -0.00003439, 0.00001611, -0.00003214, 0.00001295, -0.00002077, 0.00001470, -0.00002844, 0.00002238, -0.00001339, 0.00001032, -0.00001830, 0.00001334, -0.00002154, 0.00001783, -0.00001726, 0.00001206, 0.00000481, -0.00001480, 0.00000651, -0.00000904, 0.00000131, -0.00000268, -0.00000530, -0.00000112, -0.00000933, 0.00000897, -0.00000855, 0.00000566, -0.00000307, -0.00000501, -0.00000997, 0.00000778, -0.00000724, 0.00000962, -0.00000690, 0.00000114, -0.00000451, -0.00000029, -0.00000259, 0.00000521, -0.00000552, -0.00000294, -0.00001114, 0.00001066, -0.00000294, -0.00000388, },
		{ -0.00000332, 0.00000475, -0.00001118, 0.00000734, -0.00001294, 0.00001317, -0.00003144, 0.00004088, -0.00003142, -0.00000088, 0.00000391, 0.00000483, 0.00001197, -0.00000871, 0.00001290, -0.00000827, 0.00000747, -0.00000208, 0.00000720, -0.00000284, 0.00000585, -0.00000188, 0.00000444, -0.00000150, 0.00000488, -0.00000908, 0.00000568, -0.00000346, 0.00000648, -0.00000177, 0.00000953, -0.00001461, 0.00001356, -0.00001396, 0.00001641, -0.00001445, 0.00002325, -0.00001193, 0.00000501, -0.00000128, -0.00000871, 0.00000135, -0.00000626, -0.00000286, -0.00000043, -0.00000727, -0.00000027, 0.00000480, 0.00000201, -0.00000836, 0.00000194, -0.00001528, 0.00000284, -0.00000963, -0.00000029, -0.00000830, 0.00000319, -0.00001739, 0.00000485, -0.00002064, 0.00000448, -0.00003190, -0.00000286, -0.00003853, 0.00000612, -0.00011489, 0.00000056, -0.00041681, 0.00004768, 0.00060462, 0.00005755, 0.00051437, 0.00027636, -0.00045705, -0.00004424, -0.00008761, -0.00002550, -0.00002841, -0.00000913, -0.00002333, -0.00001760, 0.00000162, -0.00001318, -0.00000636, -0.00000988, -0.00000418, -0.00000330, -0.00000599, -0.00000835, 0.00001118, -0.00000495, -0.00000158, -0.00000615, 0.00000249, -0.00001257, 0.00001148, -0.00001378, 0.00001613, -0.00000231, -0.00001402, 0.00001073, -0.00001523, 0.00000954, -0.00001282, 0.00000528, -0.00001177, 0.00000185, -0.00000024, -0.00000074, 0.00000030, 0.00000070, -0.00001251, 0.00000002, 0.00000067, -0.00000059, 0.00000380, -0.00000342, -0.00000219, -0.00000114, -0.00000337, 0.00000253, -0.00000013, -0.00000179, -0.00000951, -0.00000285, 0.00000764, -0.00000225, -0.00000735, },
		{ 0.00000544, -0.00000562, 0.00000136, -0.00000688, 0.00000549, -0.00001147, -0.00000357, 0.00002516, -0.00003784, 0.00001982, -0.00001566, 0.00002716, -0.00000999, 0.00000814, -0.00000247, 0.00000117, 0.00000245, 0.00000316, 0.00000079, 0.00000316, -0.00000105, 0.00000436, -0.00000172, 0.00000487, -0.00000209, -0.00000367, 0.00000062, 0.00000258, -0.00000005, 0.00000619, -0.00000059, -0.00000619, 0.00000654, -0.00000824, 0.00001218, -0.00001080, 0.00002260, -0.00001801, 0.00001214, -0.00000814, -0.00000428, -0.00000020, -0.00000643, -0.00000071, -0.00000336, -0.00000552, 0.00000096, 0.00000743, -0.00000538, -0.00000298, -0.00000357, -0.00001055, -0.00000042, -0.00000700, -0.00000271, -0.00000290, -0.00000660, -0.00000940, -0.00000132, -0.00001433, -0.00000598, -0.00002572, -0.00001085, -0.00002454, -0.00002611, -0.00009774, -0.00004385, -0.00049121, 0.00032619, 0.00052403, 0.00010878, 0.00072953, -0.00027692, -0.00028177, -0.00002283, -0.00008716, -0.00000473, -0.00003533, -0.00000330, -0.00003360, 0.00000180, -0.00001006, -0.00000861, -0.00001028, -0.00000308, -0.00000835, -0.00000046, -0.00001240, 0.00000322, 0.00000469, -0.00000342, -0.00000512, -0.00000076, -0.00000347, -0.00000767, 0.00000867, -0.00001146, 0.00002063, -0.00001455, -0.00000358, 0.00000476, -0.00001169, 0.00000783, -0.00001338, 0.00000773, -0.00001577, 0.00000761, -0.00000503, 0.00000398, -0.00000298, 0.00000110, -0.00001526, 0.00000669, -0.00000470, 0.00000450, -0.00000118, -0.00000018, -0.00000535, 0.00000226, -0.00000584, 0.00000642, -0.00000490, 0.00000121, -0.00001534, 0.00000776, -0.00000030, 0.00000305, -0.00001452, },
		{ 0.00000773, -0.00000679, 0.00000279, -0.00000854, 0.00001042, -0.00002621, 0.00002785, -0.00001459, -0.00000528, 0.00000021, -0.00000057, 0.00002126, -0.00001206, 0.00001081, -0.00000887, 0.00000698, 0.00000089, 0.00000244, 0.00000147, 0.00000161, -0.00000075, 0.00000456, -0.00000237, 0.00000629, -0.00000478, -0.00000304, 0.00000155, 0.00000220, -0.00000016, 0.00000801, -0.00000593, -0.00000095, 0.00000168, -0.00000377, 0.00000907, -0.00000792, 0.00002113, -0.00002353, 0.00002169, -0.00002026, 0.00000744, -0.00001009, 0.00000261, -0.00000839, 0.00000299, -0.00001211, 0.00001240, -0.00000357, 0.00000024, -0.00000869, 0.00000212, -0.00001732, 0.00000796, -0.00001655, 0.00000962, -0.00001593, 0.00000166, -0.00001673, 0.00000841, -0.00002737, 0.00000312, -0.00003880, 0.00000550, -0.00004205, -0.00003293, -0.00009580, -0.00010362, -0.00053137, 0.00067823, 0.00027031, 0.00048853, 0.00043759, -0.00058036, -0.00000311, -0.00013860, 0.00001073, -0.00006184, 0.00001465, -0.00004911, 0.00000771, -0.00002251, 0.00000913, -0.00003010, 0.00001270, -0.00002137, 0.00000861, -0.00001890, 0.00000459, -0.00000538, 0.00001251, -0.00001525, 0.00000589, -0.00000909, 0.00000193, -0.00001222, 0.00001500, -0.00001610, 0.00002864, -0.00003188, 0.00001654, -0.00001305, 0.00000394, -0.00000648, -0.00000050, -0.00000317, -0.00000735, 0.00000215, 0.00000073, -0.00000101, 0.00000181, -0.00000672, -0.00000814, 0.00000391, -0.00000211, 0.00000197, 0.00000113, -0.00000392, -0.00000149, -0.00000074, -0.00000198, 0.00000320, -0.00000257, -0.00000383, -0.00001110, 0.00000927, -0.00000160, 0.00000238, -0.00001609, },
		{ 0.00000536, -0.00000166, -0.00000562, 0.00000503, -0.00000904, -0.00000536, 0.00001901, -0.00002288, 0.00001312, -0.00001958, 0.00002373, 0.00000052, 0.00000046, -0.00000016, -0.00000353, 0.00000695, 0.00000097, 0.00000049, 0.00000388, -0.00000260, 0.00000308, 0.00000159, -0.00000036, 0.00000601, -0.00000779, -0.00000088, 0.00000159, 0.00000190, 0.00000020, 0.00000900, -0.00001150, 0.00000637, -0.00000698, 0.00000663, -0.00000158, 0.00000367, 0.00000821, -0.00001533, 0.00001805, -0.00002093, 0.00000945, -0.00001090, 0.00000270, -0.00000754, 0.00000026, -0.00000720, 0.00001162, -0.00000525, -0.00000281, -0.00000339, -0.00000499, -0.00001005, 0.00000093, -0.00000841, 0.00000313, -0.00001256, -0.00000576, -0.00000410, -0.00000620, -0.00001553, -0.00001514, -0.00001795, -0.00001571, -0.00002597, -0.00007945, -0.00003773, -0.00027847, -0.00037072, 0.00087517, 0.00014022, 0.00086659, -0.00027406, -0.00034565, -0.00000881, -0.00010964, 0.00000856, -0.00004791, 0.00000101, -0.00003981, 0.00000879, -0.00001705, -0.00000237, -0.00001960, 0.00000792, -0.00001602, 0.00000170, -0.00001454, 0.00000441, -0.00000000, 0.00000405, -0.00001009, 0.00000217, -0.00000444, -0.00000612, -0.00000172, 0.00000676, -0.00000622, 0.00001892, -0.00003084, 0.00002220, -0.00002013, 0.00001057, -0.00001305, 0.00000659, -0.00001020, -0.00000196, 0.00000004, 0.00000330, -0.00000349, 0.00000339, -0.00001126, -0.00000231, 0.00000156, -0.00000069, 0.00000079, 0.00000224, -0.00000691, 0.00000262, -0.00000413, 0.00000209, -0.00000084, 0.00000056, -0.00001046, -0.00000198, 0.00000413, 0.00000350, -0.00000513, -0.00001162, },
		{ 0.00000830, -0.00000528, -0.00000317, 0.00000691, -0.00002092, 0.00002118, -0.00001419, 0.00000642, -0.00000796, -0.00000496, 0.00002159, -0.00000363, 0.00000192, -0.00000446, -0.00000005, 0.00000927, -0.00000471, 0.00000616, -0.00000208, 0.00000107, 0.00000074, 0.00000366, -0.00000231, 0.00000904, -0.00001558, 0.00000826, -0.00000578, 0.00000836, -0.00000504, 0.00001403, -0.00002072, 0.00001806, -0.00002078, 0.00002392, -0.00002121, 0.00002566, -0.00001792, 0.00000972, -0.00000412, -0.00000335, -0.00000533, 0.00000433, -0.00001335, 0.00000909, -0.00001792, 0.00001539, -0.00000959, 0.00001201, -0.00002284, 0.00001992, -0.00003205, 0.00001873, -0.00002866, 0.00002508, -0.00003268, 0.00002004, -0.00004013, 0.00003750, -0.00005542, 0.00003329, -0.00007214, 0.00004919, -0.00009331, 0.00004633, -0.00018370, 0.00008458, -0.00058360, 0.00006750, 0.00076526, 0.00042320, 0.00067257, -0.00074643, 0.00005299, -0.00020093, 0.00005159, -0.00009945, 0.00004556, -0.00008330, 0.00003750, -0.00004848, 0.00003274, -0.00005292, 0.00003056, -0.00003400, 0.00002108, -0.00003558, 0.00002120, -0.00002344, 0.00002753, -0.00002594, 0.00001726, -0.00002204, 0.00001793, -0.00003077, 0.00002730, -0.00002067, 0.00002207, -0.00001189, -0.00000574, 0.00000511, -0.00000767, -0.00000020, -0.00000320, -0.00000127, -0.00000463, -0.00000721, 0.00000848, -0.00000482, 0.00000365, -0.00000454, -0.00000597, -0.00000471, 0.00000595, -0.00000630, 0.00000719, -0.00000477, -0.00000172, -0.00000025, -0.00000132, -0.00000020, 0.00000139, -0.00000344, -0.00000886, 0.00000067, 0.00000348, 0.00000480, -0.00001086, -0.00000871, },
		{ 0.00001291, -0.00001571, 0.00001273, -0.00001433, 0.00000117, 0.00000947, -0.00001706, 0.00002134, -0.00002979, 0.00002085, 0.00000364, 0.00000570, -0.00000816, 0.00000065, 0.00000030, 0.00001041, -0.00000933, 0.00001195, -0.00000958, 0.00000716, -0.00000308, 0.00000580, -0.00000227, 0.00000718, -0.00001760, 0.00001310, -0.00000975, 0.00001118, -0.00000544, 0.00001167, -0.00002043, 0.00001909, -0.00002246, 0.00002852, -0.00002883, 0.00003592, -0.00003361, 0.00002731, -0.00002127, 0.00000987, -0.00001420, 0.00001175, -0.00002068, 0.00001583, -0.00002381, 0.00002513, -0.00002006, 0.00001763, -0.00002758, 0.00002565, -0.00004083, 0.00002857, -0.00003697, 0.00003567, -0.00004747, 0.00003131, -0.00004744, 0.00004676, -0.00007164, 0.00004630, -0.00008578, 0.00006790, -0.00012376, 0.00006247, -0.00021254, 0.00009367, -0.00078822, 0.00054262, 0.00061980, 0.00083622, -0.00010709, -0.00054924, 0.00010669, -0.00021607, 0.00008901, -0.00011799, 0.00005954, -0.00009844, 0.00006207, -0.00006474, 0.00004147, -0.00006236, 0.00004552, -0.00004615, 0.00002839, -0.00004382, 0.00003326, -0.00003075, 0.00003293, -0.00003401, 0.00002519, -0.00002814, 0.00002113, -0.00003491, 0.00003825, -0.00003337, 0.00003650, -0.00003191, 0.00001389, -0.00000924, 0.00000173, -0.00000711, 0.00000315, -0.00000616, -0.00000286, -0.00000694, 0.00001150, -0.00000899, 0.00000703, -0.00000900, -0.00000309, -0.00000421, 0.00000620, -0.00000783, 0.00001047, -0.00001023, 0.00000365, -0.00000345, 0.00000118, -0.00000230, 0.00000346, -0.00000838, -0.00000360, -0.00000125, 0.00000751, 0.00000067, -0.00001363, -0.00000554, },
		{ 0.00000395, -0.00001144, 0.00001457, -0.00002692, 0.00002853, -0.00002647, 0.00001873, -0.00000614, -0.00001465, 0.00002001, 0.00000207, 0.00000276, -0.00000934, 0.00000117, 0.00000661, 0.00000127, -0.00000165, 0.00000510, -0.00000551, 0.00000371, 0.00000177, -0.00000049, 0.00000690, -0.00000683, -0.00000473, 0.00000310, 0.00000025, 0.00000063, 0.00000767, -0.00000617, -0.00000227, 0.00000060, -0.00000209, 0.00000847, -0.00001063, 0.00001933, -0.00002209, 0.00001934, -0.00001536, 0.00000200, -0.00000155, -0.00000431, -0.00000324, -0.00000309, -0.00000123, 0.00000352, 0.00000011, -0.00000728, 0.00000146, -0.00000621, -0.00000962, -0.00000295, -0.00000024, -0.00000422, -0.00001033, -0.00000922, 0.00000239, -0.00001031, -0.00001605, -0.00001780, -0.00000968, -0.00001995, -0.00003770, -0.00005420, -0.00007633, -0.00014455, -0.00067983, 0.00079920, 0.00081186, 0.00070970, -0.00074517, -0.00004908, -0.00013473, -0.00001836, -0.00004448, -0.00000291, -0.00004427, -0.00000356, -0.00000937, -0.00000268, -0.00002021, -0.00000318, -0.00000269, -0.00000336, -0.00001615, -0.00000003, -0.00000242, 0.00000365, -0.00000180, -0.00000272, -0.00000368, -0.00000020, -0.00000937, -0.00000286, 0.00001317, -0.00001335, 0.00001971, -0.00002288, 0.00000974, -0.00000469, -0.00000559, 0.00000237, -0.00000604, 0.00000276, -0.00001424, 0.00000798, -0.00000081, 0.00000057, -0.00000218, -0.00000154, -0.00001072, 0.00000652, -0.00000490, 0.00000278, 0.00000157, -0.00000464, 0.00000027, 0.00000069, -0.00000377, 0.00000330, -0.00000292, -0.00000443, -0.00000556, 0.00000276, 0.00000679, -0.00000160, -0.00001932, 0.00000686, },
		{ -0.00001548, 0.00001061, -0.00001080, -0.00000212, 0.00001486, -0.00002985, 0.00003967, -0.00003704, 0.00001566, 0.00000113, 0.00001487, -0.00001313, 0.00000083, -0.00000342, 0.00001430, -0.00001045, 0.00001047, -0.00000762, 0.00000474, -0.00000418, 0.00000925, -0.00000746, 0.00001477, -0.00002029, 0.00001007, -0.00000950, 0.00001210, -0.00001043, 0.00001975, -0.00002312, 0.00001601, -0.00001823, 0.00001981, -0.00001566, 0.00001343, -0.00000450, -0.00000234, 0.00000392, -0.00000452, -0.00000716, 0.00001010, -0.00001923, 0.00001302, -0.00001947, 0.00001881, -0.00001755, 0.00001892, -0.00002824, 0.00002616, -0.00003515, 0.00001932, -0.00003029, 0.00003147, -0.00004226, 0.00002592, -0.00004472, 0.00004483, -0.00006317, 0.00003615, -0.00007489, 0.00006005, -0.00011063, 0.00005457, -0.00017786, 0.00007240, -0.00046876, -0.00039248, 0.00102737, 0.00111521, -0.00007978, -0.00065603, 0.00012172, -0.00023514, 0.00009022, -0.00012048, 0.00006099, -0.00010499, 0.00006330, -0.00006295, 0.00003963, -0.00006094, 0.00004017, -0.00003895, 0.00002535, -0.00004594, 0.00003391, -0.00003129, 0.00002970, -0.00002813, 0.00002180, -0.00002637, 0.00002062, -0.00003273, 0.00002650, -0.00001381, 0.00000918, -0.00000031, -0.00000967, 0.00000341, -0.00000157, -0.00000899, 0.00000764, -0.00001121, 0.00000655, -0.00001940, 0.00001832, -0.00001109, 0.00000831, -0.00000906, 0.00000340, -0.00001497, 0.00001350, -0.00001340, 0.00001257, -0.00000834, 0.00000301, -0.00000426, 0.00000447, -0.00000811, 0.00000856, -0.00001014, 0.00000244, -0.00001118, 0.00001121, 0.00000130, -0.00000302, -0.00002311, 0.00002218, },
		{ -0.00001996, 0.00002043, -0.00003082, 0.00003237, -0.00002858, 0.00001200, 0.00001218, -0.00002816, 0.00001946, -0.00000122, 0.00001414, -0.00001913, 0.00000579, -0.00000103, 0.00001064, -0.00000957, 0.00001047, -0.00000966, 0.00000605, -0.00000310, 0.00000705, -0.00000330, 0.00000839, -0.00001789, 0.00001017, -0.00000825, 0.00000988, -0.00000629, 0.00001442, -0.00002154, 0.00001586, -0.00001764, 0.00002153, -0.00002091, 0.00002039, -0.00001289, 0.00000359, 0.00000183, -0.00000816, 0.00000162, -0.00000004, -0.00001051, 0.00000483, -0.00000888, 0.00000907, -0.00000886, 0.00000719, -0.00001516, 0.00001364, -0.00002580, 0.00000946, -0.00001494, 0.00001477, -0.00003062, 0.00001257, -0.00002493, 0.00002345, -0.00004847, 0.00001700, -0.00004910, 0.00003223, -0.00009750, 0.00003411, -0.00017072, 0.00005490, -0.00066311, -0.00013281, 0.00158358, 0.00079734, -0.00075869, -0.00016509, -0.00006925, -0.00008436, 0.00000194, -0.00004410, -0.00000834, -0.00004298, 0.00002443, -0.00003082, 0.00000530, -0.00002801, 0.00001681, -0.00001743, -0.00000118, -0.00001987, 0.00001666, -0.00001576, 0.00001395, -0.00001414, 0.00000860, -0.00001386, 0.00000677, -0.00002047, 0.00002339, -0.00001487, 0.00000974, -0.00000118, -0.00001235, 0.00001171, -0.00001434, 0.00000557, -0.00000562, 0.00000145, -0.00000790, -0.00000459, 0.00000914, -0.00000483, 0.00000131, -0.00000160, -0.00000578, -0.00000450, 0.00000501, -0.00000664, 0.00000828, -0.00000626, 0.00000111, -0.00000009, -0.00000137, -0.00000203, 0.00000280, -0.00000613, -0.00000069, -0.00000781, 0.00001264, 0.00000005, -0.00001177, -0.00001233, 0.00002067, },
		{ -0.00000141, -0.00000051, -0.00001149, 0.00002570, -0.00004186, 0.00004465, -0.00002665, 0.00000264, 0.00000130, 0.00001414, -0.00000366, -0.00000999, 0.00000251, 0.00000616, 0.00000056, -0.00000074, 0.00000185, -0.00000351, 0.00000112, 0.00000293, 0.00000086, 0.00000405, -0.00000281, -0.00000857, 0.00000367, -0.00000136, 0.00000268, 0.00000271, 0.00000278, -0.00001266, 0.00000902, -0.00000997, 0.00001487, -0.00001814, 0.00002125, -0.00001777, 0.00000901, -0.00000246, -0.00000761, 0.00000657, -0.00000857, -0.00000212, -0.00000216, 0.00000061, -0.00000169, 0.00000083, -0.00000428, -0.00000160, -0.00000112, -0.00001427, 0.00000002, -0.00000046, -0.00000508, -0.00001445, -0.00000180, -0.00000505, -0.00000227, -0.00002862, -0.00000379, -0.00001976, -0.00000981, -0.00006550, -0.00000556, -0.00014156, -0.00001869, -0.00085418, 0.00037502, 0.00215186, -0.00030533, -0.00061086, -0.00000360, -0.00014378, -0.00000062, -0.00004804, -0.00000256, -0.00005006, 0.00000591, -0.00000979, -0.00000728, -0.00001812, -0.00000181, -0.00000210, -0.00000375, -0.00001939, 0.00000351, -0.00000098, -0.00000191, 0.00000089, -0.00000212, -0.00000304, -0.00000344, -0.00000546, -0.00000583, 0.00001572, -0.00001450, 0.00001219, -0.00000619, -0.00000812, 0.00001107, -0.00001822, 0.00001331, -0.00001395, 0.00000953, -0.00001844, 0.00000930, -0.00000143, 0.00000196, -0.00000480, 0.00000456, -0.00001383, 0.00000582, -0.00000489, 0.00000292, 0.00000028, -0.00000062, -0.00000274, 0.00000458, -0.00000805, 0.00000590, -0.00000619, 0.00000303, -0.00001026, 0.00000329, 0.00000677, 0.00000184, -0.00002302, 0.00000841, 0.00000165, },
		{ 0.00001186, -0.00002299, 0.00002455, -0.00001750, -0.00000464, 0.00002713, -0.00002823, 0.00001100, -0.00000307, 0.00001841, -0.00001512, -0.00000232, 0.00000312, 0.00000515, -0.00000067, -0.00000012, 0.00000044, -0.00000357, 0.00000306, 0.00000137, 0.00000316, 0.00000095, -0.00000333, -0.00000843, 0.00000620, -0.00000414, 0.00000594, 0.00000049, 0.00000148, -0.00001290, 0.00001202, -0.00001264, 0.00001721, -0.00002353, 0.00003112, -0.00003400, 0.00002906, -0.00002483, 0.00001493, -0.00001307, 0.00000752, -0.00001774, 0.00001669, -0.00001841, 0.00001609, -0.00001827, 0.00001532, -0.00002087, 0.00001699, -0.00003624, 0.00002872, -0.00002970, 0.00001915, -0.00004159, 0.00003217, -0.00004037, 0.00002897, -0.00006841, 0.00004559, -0.00007264, 0.00003238, -0.00012045, 0.00005204, -0.00023022, 0.00000363, -0.00122158, 0.00161700, 0.00174244, -0.00123242, 0.00000728, -0.00024242, 0.00004462, -0.00011747, 0.00005462, -0.00009359, 0.00002860, -0.00004654, 0.00003650, -0.00005565, 0.00002761, -0.00003781, 0.00003309, -0.00004287, 0.00001578, -0.00002075, 0.00002136, -0.00002518, 0.00002384, -0.00002420, 0.00001701, -0.00002356, 0.00001251, -0.00001716, 0.00002806, -0.00003280, 0.00003461, -0.00003247, 0.00001926, -0.00001478, 0.00000438, -0.00000459, 0.00000142, -0.00000535, -0.00000614, 0.00000285, 0.00000455, -0.00000645, 0.00000477, -0.00000537, -0.00000580, 0.00000078, -0.00000070, -0.00000006, 0.00000306, -0.00000429, 0.00000278, -0.00000149, -0.00000338, 0.00000252, -0.00000401, 0.00000220, -0.00001148, 0.00000892, 0.00000423, -0.00000381, -0.00002112, 0.00001711, -0.00000923, },
		{ -0.00000216, -0.00001032, 0.00002498, -0.00003794, 0.00003193, -0.00000889, -0.00000450, -0.00000314, 0.00001080, 0.00000527, -0.00001290, -0.00000021, 0.00000584, 0.00000063, 0.00000229, -0.00000366, 0.00000252, -0.00000544, 0.00000603, -0.00000042, 0.00000458, -0.00000195, -0.00000365, -0.00000705, 0.00000659, -0.00000439, 0.00000639, 0.00000093, -0.00000395, -0.00000619, 0.00000692, -0.00000689, 0.00000895, -0.00001511, 0.00002488, -0.00003352, 0.00003319, -0.00003240, 0.00002476, -0.00002218, 0.00001299, -0.00002029, 0.00002141, -0.00002478, 0.00002088, -0.00002235, 0.00001902, -0.00002374, 0.00001624, -0.00003526, 0.00003348, -0.00003797, 0.00002168, -0.00004115, 0.00003519, -0.00004532, 0.00002570, -0.00006525, 0.00004877, -0.00008235, 0.00002443, -0.00010762, 0.00001791, -0.00020687, -0.00020169, -0.00125512, 0.00319696, 0.00000201, -0.00096329, 0.00013765, -0.00027325, 0.00009286, -0.00013525, 0.00007038, -0.00011276, 0.00005508, -0.00005793, 0.00003987, -0.00006157, 0.00003749, -0.00004196, 0.00003527, -0.00005144, 0.00002807, -0.00002570, 0.00002231, -0.00002624, 0.00002612, -0.00002728, 0.00001848, -0.00002582, 0.00001568, -0.00001422, 0.00002215, -0.00003056, 0.00003743, -0.00004127, 0.00003196, -0.00002920, 0.00001917, -0.00001704, 0.00001188, -0.00001616, 0.00000434, -0.00000241, 0.00000737, -0.00001039, 0.00001047, -0.00001286, 0.00000112, -0.00000389, 0.00000358, -0.00000337, 0.00000598, -0.00000726, 0.00000725, -0.00000784, 0.00000312, -0.00000402, 0.00000332, -0.00000554, -0.00000469, 0.00000790, 0.00000460, -0.00001383, -0.00000720, 0.00000867, 0.00000030, },
		{ -0.00001909, 0.00002065, -0.00001218, -0.00000882, 0.00002205, -0.00001261, -0.00000007, -0.00000428, 0.00001428, -0.00000244, -0.00001287, 0.00000834, -0.00000286, 0.00000872, -0.00000779, 0.00000588, -0.00000826, 0.00000651, -0.00000535, 0.00001287, -0.00001113, 0.00001319, -0.00002220, 0.00001404, -0.00001391, 0.00001697, -0.00001546, 0.00002356, -0.00003265, 0.00002722, -0.00002769, 0.00002959, -0.00003236, 0.00003070, -0.00002363, 0.00001281, -0.00001031, 0.00000915, -0.00001512, 0.00001780, -0.00003093, 0.00003014, -0.00003133, 0.00002766, -0.00003427, 0.00003696, -0.00004443, 0.00004353, -0.00005909, 0.00004841, -0.00005248, 0.00004736, -0.00007237, 0.00006540, -0.00007962, 0.00007566, -0.00011480, 0.00009565, -0.00012885, 0.00010521, -0.00020447, 0.00016941, -0.00035307, 0.00024447, -0.00112823, 0.00004599, 0.00359533, -0.00184215, 0.00015813, -0.00039903, 0.00014718, -0.00020942, 0.00012424, -0.00015335, 0.00008059, -0.00010025, 0.00008378, -0.00009566, 0.00006333, -0.00007173, 0.00006129, -0.00006510, 0.00003868, -0.00004843, 0.00004791, -0.00005101, 0.00004485, -0.00004044, 0.00003408, -0.00004142, 0.00003159, -0.00003641, 0.00003941, -0.00003492, 0.00002494, -0.00001302, 0.00000132, -0.00000382, 0.00000163, -0.00000718, 0.00000751, -0.00001215, 0.00000566, -0.00001442, 0.00001834, -0.00001528, 0.00001159, -0.00000917, 0.00000321, -0.00001347, 0.00001156, -0.00001110, 0.00001099, -0.00000748, 0.00000568, -0.00000446, 0.00000132, -0.00000475, 0.00000307, -0.00000122, -0.00000419, -0.00000354, 0.00001107, -0.00000308, -0.00001251, -0.00000168, 0.00000548, 0.00000607, },
		{ -0.00000393, 0.00001994, -0.00003133, 0.00002379, -0.00000800, 0.00000816, -0.00001723, 0.00001320, 0.00000055, 0.00000184, -0.00001684, 0.00001855, -0.00001504, 0.00002108, -0.00002267, 0.00002031, -0.00002282, 0.00002163, -0.00001864, 0.00002632, -0.00002701, 0.00002853, -0.00004029, 0.00003496, -0.00003417, 0.00003729, -0.00003535, 0.00004204, -0.00005509, 0.00005479, -0.00005763, 0.00006098, -0.00006806, 0.00007232, -0.00007122, 0.00006125, -0.00005696, 0.00005420, -0.00005827, 0.00005901, -0.00007294, 0.00007801, -0.00008380, 0.00008001, -0.00008704, 0.00009262, -0.00010387, 0.00010532, -0.00012693, 0.00012590, -0.00013651, 0.00013069, -0.00015975, 0.00016375, -0.00018859, 0.00018922, -0.00024383, 0.00024759, -0.00030658, 0.00029342, -0.00042791, 0.00043895, -0.00073648, 0.00071681, -0.00232624, 0.00268994, 0.00174674, -0.00207581, 0.00072018, -0.00073291, 0.00044146, -0.00043054, 0.00031561, -0.00032440, 0.00023940, -0.00022970, 0.00019635, -0.00020243, 0.00016579, -0.00016197, 0.00014439, -0.00014788, 0.00011762, -0.00011531, 0.00010851, -0.00011029, 0.00010388, -0.00009694, 0.00008539, -0.00009121, 0.00008126, -0.00008135, 0.00008374, -0.00008365, 0.00007649, -0.00006383, 0.00004671, -0.00004313, 0.00003643, -0.00003700, 0.00003397, -0.00003722, 0.00002873, -0.00003352, 0.00003746, -0.00003588, 0.00003281, -0.00002936, 0.00001988, -0.00002806, 0.00002707, -0.00002640, 0.00002648, -0.00002202, 0.00001963, -0.00001820, 0.00001337, -0.00001607, 0.00001530, -0.00001236, 0.00000384, -0.00000687, 0.00001571, -0.00001431, -0.00000162, -0.00000874, 0.00001820, -0.00001217, },
		{ 0.00002756, -0.00001688, -0.00000292, 0.00001207, -0.00000607, 0.00000719, -0.00001783, 0.00001727, -0.00000381, -0.00000263, -0.00000612, 0.00000857, -0.00000511, 0.00000993, -0.00001342, 0.00001059, -0.00001206, 0.00001036, -0.00000362, 0.00000776, -0.00000804, 0.00000690, -0.00001896, 0.00001517, -0.00001239, 0.00001335, -0.00000854, 0.00000975, -0.00002137, 0.00002239, -0.00002541, 0.00002709, -0.00003423, 0.00004114, -0.00004527, 0.00003665, -0.00002943, 0.00002271, -0.00002218, 0.00001675, -0.00002437, 0.00002825, -0.00003529, 0.00002805, -0.00002879, 0.00002941, -0.00003717, 0.00003246, -0.00004873, 0.00004635, -0.00005648, 0.00004130, -0.00005833, 0.00005436, -0.00007296, 0.00005656, -0.00009500, 0.00008599, -0.00013771, 0.00008888, -0.00018474, 0.00013038, -0.00039566, 0.00016156, -0.00219772, 0.00483099, -0.00128760, -0.00055906, 0.00004972, -0.00023011, 0.00008951, -0.00013592, 0.00006804, -0.00011286, 0.00006786, -0.00007314, 0.00004964, -0.00006932, 0.00004981, -0.00005354, 0.00004119, -0.00005533, 0.00003652, -0.00003589, 0.00003011, -0.00003599, 0.00003653, -0.00003526, 0.00002446, -0.00003257, 0.00002850, -0.00003007, 0.00003430, -0.00004128, 0.00004301, -0.00003751, 0.00002205, -0.00001732, 0.00001045, -0.00000989, 0.00000585, -0.00000951, 0.00000179, -0.00000501, 0.00000986, -0.00001084, 0.00001087, -0.00000964, -0.00000108, -0.00000600, 0.00000726, -0.00000849, 0.00001145, -0.00000831, 0.00000657, -0.00000667, 0.00000186, -0.00000493, 0.00000755, -0.00000721, -0.00000101, 0.00000161, 0.00000580, -0.00000944, -0.00000404, -0.00000428, 0.00002120, -0.00002955, },
		{ 0.00002559, -0.00003602, 0.00003017, -0.00002135, 0.00002344, -0.00002133, 0.00000782, -0.00000305, 0.00001124, -0.00002083, 0.00001828, -0.00001828, 0.00002394, -0.00002255, 0.00001851, -0.00002151, 0.00002099, -0.00002251, 0.00003243, -0.00003363, 0.00003612, -0.00004202, 0.00003192, -0.00003500, 0.00003990, -0.00004178, 0.00005007, -0.00005617, 0.00005006, -0.00005179, 0.00005103, -0.00005348, 0.00005020, -0.00004484, 0.00003739, -0.00004375, 0.00005330, -0.00006462, 0.00007050, -0.00008315, 0.00008581, -0.00008968, 0.00008512, -0.00009701, 0.00010580, -0.00011558, 0.00011647, -0.00013248, 0.00012974, -0.00014430, 0.00014215, -0.00017153, 0.00017634, -0.00020376, 0.00020578, -0.00025475, 0.00025821, -0.00031684, 0.00030897, -0.00043705, 0.00045069, -0.00069988, 0.00066820, -0.00165471, 0.00029669, 0.00428589, -0.00301609, 0.00097466, -0.00091716, 0.00054577, -0.00051357, 0.00037340, -0.00037267, 0.00028014, -0.00026995, 0.00022812, -0.00022751, 0.00018754, -0.00018190, 0.00015971, -0.00016045, 0.00013330, -0.00013464, 0.00012461, -0.00012404, 0.00011200, -0.00010203, 0.00009295, -0.00009903, 0.00008801, -0.00008474, 0.00007808, -0.00007097, 0.00005762, -0.00004553, 0.00003921, -0.00004730, 0.00004908, -0.00005285, 0.00005125, -0.00005396, 0.00004875, -0.00005428, 0.00005095, -0.00004448, 0.00004079, -0.00003675, 0.00003355, -0.00004371, 0.00003763, -0.00003442, 0.00003108, -0.00002385, 0.00002365, -0.00002371, 0.00002124, -0.00002540, 0.00002282, -0.00001709, 0.00001330, -0.00001825, 0.00001960, -0.00001476, 0.00001007, -0.00002368, 0.00002208, -0.00000618, -0.00001037, },
		{ -0.00001558, -0.00000558, 0.00001742, -0.00002115, 0.00003038, -0.00003480, 0.00002330, -0.00001641, 0.00001873, -0.00002599, 0.00002517, -0.00002612, 0.00003312, -0.00003552, 0.00003157, -0.00003370, 0.00003264, -0.00003127, 0.00004020, -0.00004401, 0.00004786, -0.00005766, 0.00004946, -0.00005048, 0.00005470, -0.00005659, 0.00006492, -0.00007503, 0.00007309, -0.00007744, 0.00007799, -0.00008291, 0.00008290, -0.00007993, 0.00006884, -0.00006893, 0.00007514, -0.00008590, 0.00009190, -0.00010523, 0.00011261, -0.00012243, 0.00011785, -0.00012779, 0.00013876, -0.00015316, 0.00015649, -0.00017533, 0.00017780, -0.00019939, 0.00019835, -0.00022843, 0.00023971, -0.00027815, 0.00028517, -0.00034338, 0.00036378, -0.00045228, 0.00046043, -0.00062198, 0.00068319, -0.00107185, 0.00119205, -0.00314577, 0.00383989, 0.00107299, -0.00215107, 0.00095183, -0.00091667, 0.00061064, -0.00056851, 0.00042784, -0.00042463, 0.00034177, -0.00032334, 0.00027385, -0.00027071, 0.00023328, -0.00022293, 0.00019524, -0.00019556, 0.00016988, -0.00016643, 0.00015283, -0.00015203, 0.00014206, -0.00013122, 0.00011730, -0.00012219, 0.00011465, -0.00011140, 0.00010425, -0.00009926, 0.00008681, -0.00007122, 0.00005664, -0.00005786, 0.00005743, -0.00005960, 0.00005689, -0.00005923, 0.00005443, -0.00006018, 0.00005885, -0.00005312, 0.00004962, -0.00004427, 0.00003766, -0.00004773, 0.00004465, -0.00004260, 0.00004069, -0.00003179, 0.00002906, -0.00002836, 0.00002442, -0.00002861, 0.00002917, -0.00002395, 0.00001866, -0.00002110, 0.00002127, -0.00001758, 0.00001348, -0.00002906, 0.00003855, -0.00003538, 0.00002879, },
		{ -0.00004140, 0.00003395, -0.00002346, 0.00001551, -0.00000056, -0.00001241, 0.00000593, -0.00000071, 0.00000038, -0.00000391, 0.00000076, 0.00000089, 0.00000401, -0.00000824, 0.00000416, -0.00000454, 0.00000195, 0.00000415, -0.00000058, -0.00000112, 0.00000250, -0.00001324, 0.00000561, -0.00000284, 0.00000312, -0.00000132, 0.00000517, -0.00001394, 0.00001159, -0.00001550, 0.00001450, -0.00001826, 0.00001830, -0.00001663, 0.00000159, 0.00000796, -0.00001143, 0.00000818, -0.00000881, 0.00000348, -0.00000172, -0.00000686, -0.00000308, 0.00000387, -0.00000232, -0.00000554, 0.00000043, -0.00000824, -0.00000128, -0.00001180, -0.00000266, -0.00000731, -0.00000283, -0.00001793, -0.00000194, -0.00002116, -0.00000043, -0.00005372, -0.00000094, -0.00007760, 0.00000174, -0.00026271, 0.00000871, -0.00224382, 0.00553042, -0.00224382, 0.00000871, -0.00026271, 0.00000174, -0.00007760, -0.00000094, -0.00005372, -0.00000043, -0.00002116, -0.00000194, -0.00001793, -0.00000283, -0.00000731, -0.00000266, -0.00001180, -0.00000128, -0.00000824, 0.00000043, -0.00000554, -0.00000232, 0.00000387, -0.00000308, -0.00000686, -0.00000172, 0.00000348, -0.00000881, 0.00000818, -0.00001143, 0.00000796, 0.00000159, -0.00001663, 0.00001830, -0.00001826, 0.00001450, -0.00001550, 0.00001159, -0.00001394, 0.00000517, -0.00000132, 0.00000312, -0.00000284, 0.00000561, -0.00001324, 0.00000250, -0.00000112, -0.00000058, 0.00000415, 0.00000195, -0.00000454, 0.00000416, -0.00000824, 0.00000401, 0.00000089, 0.00000076, -0.00000391, 0.00000038, -0.00000071, 0.00000593, -0.00001241, -0.00000056, 0.00001551, -0.00002346, 0.00003395, },
		{ -0.00001558, 0.00002879, -0.00003538, 0.00003855, -0.00002906, 0.00001348, -0.00001758, 0.00002127, -0.00002110, 0.00001866, -0.00002395, 0.00002917, -0.00002861, 0.00002442, -0.00002836, 0.00002906, -0.00003179, 0.00004069, -0.00004260, 0.00004465, -0.00004773, 0.00003766, -0.00004427, 0.00004962, -0.00005312, 0.00005885, -0.00006018, 0.00005443, -0.00005923, 0.00005689, -0.00005960, 0.00005743, -0.00005786, 0.00005664, -0.00007122, 0.00008681, -0.00009926, 0.00010425, -0.00011140, 0.00011465, -0.00012219, 0.00011730, -0.00013122, 0.00014206, -0.00015203, 0.00015283, -0.00016643, 0.00016988, -0.00019556, 0.00019524, -0.00022293, 0.00023328, -0.00027071, 0.00027385, -0.00032334, 0.00034177, -0.00042463, 0.00042784, -0.00056851, 0.00061064, -0.00091667, 0.00095183, -0.00215107, 0.00107299, 0.00383989, -0.00314577, 0.00119205, -0.00107185, 0.00068319, -0.00062198, 0.00046043, -0.00045228, 0.00036378, -0.00034338, 0.00028517, -0.00027815, 0.00023971, -0.00022843, 0.00019835, -0.00019939, 0.00017780, -0.00017533, 0.00015649, -0.00015316, 0.00013876, -0.00012779, 0.00011785, -0.00012243, 0.00011261, -0.00010523, 0.00009190, -0.00008590, 0.00007514, -0.00006893, 0.00006884, -0.00007993, 0.00008290, -0.00008291, 0.00007799, -0.00007744, 0.00007309, -0.00007503, 0.00006492, -0.00005659, 0.00005470, -0.00005048, 0.00004946, -0.00005766, 0.00004786, -0.00004401, 0.00004020, -0.00003127, 0.00003264, -0.00003370, 0.00003157, -0.00003552, 0.00003312, -0.00002612, 0.00002517, -0.00002599, 0.00001873, -0.00001641, 0.00002330, -0.00003480, 0.00003038, -0.00002115, 0.00001742, -0.00000558, },
		{ 0.00002559, -0.00001037, -0.00000618, 0.00002208, -0.00002368, 0.00001007, -0.00001476, 0.00001960, -0.00001825, 0.00001330, -0.00001709, 0.00002282, -0.00002540, 0.00002124, -0.00002371, 0.00002365, -0.00002385, 0.00003108, -0.00003442, 0.00003763, -0.00004371, 0.00003355, -0.00003675, 0.00004079, -0.00004448, 0.00005095, -0.00005428, 0.00004875, -0.00005396, 0.00005125, -0.00005285, 0.00004908, -0.00004730, 0.00003921, -0.00004553, 0.00005762, -0.00007097, 0.00007808, -0.00008474, 0.00008801, -0.00009903, 0.00009295, -0.00010203, 0.00011200, -0.00012404, 0.00012461, -0.00013464, 0.00013330, -0.00016045, 0.00015971, -0.00018190, 0.00018754, -0.00022751, 0.00022812, -0.00026995, 0.00028014, -0.00037267, 0.00037340, -0.00051357, 0.00054577, -0.00091716, 0.00097466, -0.00301609, 0.00428589, 0.00029669, -0.00165471, 0.00066820, -0.00069988, 0.00045069, -0.00043705, 0.00030897, -0.00031684, 0.00025821, -0.00025475, 0.00020578, -0.00020376, 0.00017634, -0.00017153, 0.00014215, -0.00014430, 0.00012974, -0.00013248, 0.00011647, -0.00011558, 0.00010580, -0.00009701, 0.00008512, -0.00008968, 0.00008581, -0.00008315, 0.00007050, -0.00006462, 0.00005330, -0.00004375, 0.00003739, -0.00004484, 0.00005020, -0.00005348, 0.00005103, -0.00005179, 0.00005006, -0.00005617, 0.00005007, -0.00004178, 0.00003990, -0.00003500, 0.00003192, -0.00004202, 0.00003612, -0.00003363, 0.00003243, -0.00002251, 0.00002099, -0.00002151, 0.00001851, -0.00002255, 0.00002394, -0.00001828, 0.00001828, -0.00002083, 0.00001124, -0.00000305, 0.00000782, -0.00002133, 0.00002344, -0.00002135, 0.00003017, -0.00003602, },
		{ 0.00002756, -0.00002955, 0.00002120, -0.00000428, -0.00000404, -0.00000944, 0.00000580, 0.00000161, -0.00000101, -0.00000721, 0.00000755, -0.00000493, 0.00000186, -0.00000667, 0.00000657, -0.00000831, 0.00001145, -0.00000849, 0.00000726, -0.00000600, -0.00000108, -0.00000964, 0.00001087, -0.00001084, 0.00000986, -0.00000501, 0.00000179, -0.00000951, 0.00000585, -0.00000989, 0.00001045, -0.00001732, 0.00002205, -0.00003751, 0.00004301, -0.00004128, 0.00003430, -0.00003007, 0.00002850, -0.00003257, 0.00002446, -0.00003526, 0.00003653, -0.00003599, 0.00003011, -0.00003589, 0.00003652, -0.00005533, 0.00004119, -0.00005354, 0.00004981, -0.00006932, 0.00004964, -0.00007314, 0.00006786, -0.00011286, 0.00006804, -0.00013592, 0.00008951, -0.00023011, 0.00004972, -0.00055906, -0.00128760, 0.00483099, -0.00219772, 0.00016156, -0.00039566, 0.00013038, -0.00018474, 0.00008888, -0.00013771, 0.00008599, -0.00009500, 0.00005656, -0.00007296, 0.00005436, -0.00005833, 0.00004130, -0.00005648, 0.00004635, -0.00004873, 0.00003246, -0.00003717, 0.00002941, -0.00002879, 0.00002805, -0.00003529, 0.00002825, -0.00002437, 0.00001675, -0.00002218, 0.00002271, -0.00002943, 0.00003665, -0.00004527, 0.00004114, -0.00003423, 0.00002709, -0.00002541, 0.00002239, -0.00002137, 0.00000975, -0.00000854, 0.00001335, -0.00001239, 0.00001517, -0.00001896, 0.00000690, -0.00000804, 0.00000776, -0.00000362, 0.00001036, -0.00001206, 0.00001059, -0.00001342, 0.00000993, -0.00000511, 0.00000857, -0.00000612, -0.00000263, -0.00000381, 0.00001727, -0.00001783, 0.00000719, -0.00000607, 0.00001207, -0.00000292, -0.00001688, },
		{ -0.00000393, -0.00001217, 0.00001820, -0.00000874, -0.00000162, -0.00001431, 0.00001571, -0.00000687, 0.00000384, -0.00001236, 0.00001530, -0.00001607, 0.00001337, -0.00001820, 0.00001963, -0.00002202, 0.00002648, -0.00002640, 0.00002707, -0.00002806, 0.00001988, -0.00002936, 0.00003281, -0.00003588, 0.00003746, -0.00003352, 0.00002873, -0.00003722, 0.00003397, -0.00003700, 0.00003643, -0.00004313, 0.00004671, -0.00006383, 0.00007649, -0.00008365, 0.00008374, -0.00008135, 0.00008126, -0.00009121, 0.00008539, -0.00009694, 0.00010388, -0.00011029, 0.00010851, -0.00011531, 0.00011762, -0.00014788, 0.00014439, -0.00016197, 0.00016579, -0.00020243, 0.00019635, -0.00022970, 0.00023940, -0.00032440, 0.00031561, -0.00043054, 0.00044146, -0.00073291, 0.00072018, -0.00207581, 0.00174674, 0.00268994, -0.00232624, 0.00071681, -0.00073648, 0.00043895, -0.00042791, 0.00029342, -0.00030658, 0.00024759, -0.00024383, 0.00018922, -0.00018859, 0.00016375, -0.00015975, 0.00013069, -0.00013651, 0.00012590, -0.00012693, 0.00010532, -0.00010387, 0.00009262, -0.00008704, 0.00008001, -0.00008380, 0.00007801, -0.00007294, 0.00005901, -0.00005827, 0.00005420, -0.00005696, 0.00006125, -0.00007122, 0.00007232, -0.00006806, 0.00006098, -0.00005763, 0.00005479, -0.00005509, 0.00004204, -0.00003535, 0.00003729, -0.00003417, 0.00003496, -0.00004029, 0.00002853, -0.00002701, 0.00002632, -0.00001864, 0.00002163, -0.00002282, 0.00002031, -0.00002267, 0.00002108, -0.00001504, 0.00001855, -0.00001684, 0.00000184, 0.00000055, 0.00001320, -0.00001723, 0.00000816, -0.00000800, 0.00002379, -0.00003133, 0.00001994, },
		{ -0.00001909, 0.00000607, 0.00000548, -0.00000168, -0.00001251, -0.00000308, 0.00001107, -0.00000354, -0.00000419, -0.00000122, 0.00000307, -0.00000475, 0.00000132, -0.00000446, 0.00000568, -0.00000748, 0.00001099, -0.00001110, 0.00001156, -0.00001347, 0.00000321, -0.00000917, 0.00001159, -0.00001528, 0.00001834, -0.00001442, 0.00000566, -0.00001215, 0.00000751, -0.00000718, 0.00000163, -0.00000382, 0.00000132, -0.00001302, 0.00002494, -0.00003492, 0.00003941, -0.00003641, 0.00003159, -0.00004142, 0.00003408, -0.00004044, 0.00004485, -0.00005101, 0.00004791, -0.00004843, 0.00003868, -0.00006510, 0.00006129, -0.00007173, 0.00006333, -0.00009566, 0.00008378, -0.00010025, 0.00008059, -0.00015335, 0.00012424, -0.00020942, 0.00014718, -0.00039903, 0.00015813, -0.00184215, 0.00359533, 0.00004599, -0.00112823, 0.00024447, -0.00035307, 0.00016941, -0.00020447, 0.00010521, -0.00012885, 0.00009565, -0.00011480, 0.00007566, -0.00007962, 0.00006540, -0.00007237, 0.00004736, -0.00005248, 0.00004841, -0.00005909, 0.00004353, -0.00004443, 0.00003696, -0.00003427, 0.00002766, -0.00003133, 0.00003014, -0.00003093, 0.00001780, -0.00001512, 0.00000915, -0.00001031, 0.00001281, -0.00002363, 0.00003070, -0.00003236, 0.00002959, -0.00002769, 0.00002722, -0.00003265, 0.00002356, -0.00001546, 0.00001697, -0.00001391, 0.00001404, -0.00002220, 0.00001319, -0.00001113, 0.00001287, -0.00000535, 0.00000651, -0.00000826, 0.00000588, -0.00000779, 0.00000872, -0.00000286, 0.00000834, -0.00001287, -0.00000244, 0.00001428, -0.00000428, -0.00000007, -0.00001261, 0.00002205, -0.00000882, -0.00001218, 0.00002065, },
		{ -0.00000216, 0.00000030, 0.00000867, -0.00000720, -0.00001383, 0.00000460, 0.00000790, -0.00000469, -0.00000554, 0.00000332, -0.00000402, 0.00000312, -0.00000784, 0.00000725, -0.00000726, 0.00000598, -0.00000337, 0.00000358, -0.00000389, 0.00000112, -0.00001286, 0.00001047, -0.00001039, 0.00000737, -0.00000241, 0.00000434, -0.00001616, 0.00001188, -0.00001704, 0.00001917, -0.00002920, 0.00003196, -0.00004127, 0.00003743, -0.00003056, 0.00002215, -0.00001422, 0.00001568, -0.00002582, 0.00001848, -0.00002728, 0.00002612, -0.00002624, 0.00002231, -0.00002570, 0.00002807, -0.00005144, 0.00003527, -0.00004196, 0.00003749, -0.00006157, 0.00003987, -0.00005793, 0.00005508, -0.00011276, 0.00007038, -0.00013525, 0.00009286, -0.00027325, 0.00013765, -0.00096329, 0.00000201, 0.00319696, -0.00125512, -0.00020169, -0.00020687, 0.00001791, -0.00010762, 0.00002443, -0.00008235, 0.00004877, -0.00006525, 0.00002570, -0.00004532, 0.00003519, -0.00004115, 0.00002168, -0.00003797, 0.00003348, -0.00003526, 0.00001624, -0.00002374, 0.00001902, -0.00002235, 0.00002088, -0.00002478, 0.00002141, -0.00002029, 0.00001299, -0.00002218, 0.00002476, -0.00003240, 0.00003319, -0.00003352, 0.00002488, -0.00001511, 0.00000895, -0.00000689, 0.00000692, -0.00000619, -0.00000395, 0.00000093, 0.00000639, -0.00000439, 0.00000659, -0.00000705, -0.00000365, -0.00000195, 0.00000458, -0.00000042, 0.00000603, -0.00000544, 0.00000252, -0.00000366, 0.00000229, 0.00000063, 0.00000584, -0.00000021, -0.00001290, 0.00000527, 0.00001080, -0.00000314, -0.00000450, -0.00000889, 0.00003193, -0.00003794, 0.00002498, -0.00001032, },
		{ 0.00001186, -0.00000923, 0.00001711, -0.00002112, -0.00000381, 0.00000423, 0.00000892, -0.00001148, 0.00000220, -0.00000401, 0.00000252, -0.00000338, -0.00000149, 0.00000278, -0.00000429, 0.00000306, -0.00000006, -0.00000070, 0.00000078, -0.00000580, -0.00000537, 0.00000477, -0.00000645, 0.00000455, 0.00000285, -0.00000614, -0.00000535, 0.00000142, -0.00000459, 0.00000438, -0.00001478, 0.00001926, -0.00003247, 0.00003461, -0.00003280, 0.00002806, -0.00001716, 0.00001251, -0.00002356, 0.00001701, -0.00002420, 0.00002384, -0.00002518, 0.00002136, -0.00002075, 0.00001578, -0.00004287, 0.00003309, -0.00003781, 0.00002761, -0.00005565, 0.00003650, -0.00004654, 0.00002860, -0.00009359, 0.00005462, -0.00011747, 0.00004462, -0.00024242, 0.00000728, -0.00123242, 0.00174244, 0.00161700, -0.00122158, 0.00000363, -0.00023022, 0.00005204, -0.00012045, 0.00003238, -0.00007264, 0.00004559, -0.00006841, 0.00002897, -0.00004037, 0.00003217, -0.00004159, 0.00001915, -0.00002970, 0.00002872, -0.00003624, 0.00001699, -0.00002087, 0.00001532, -0.00001827, 0.00001609, -0.00001841, 0.00001669, -0.00001774, 0.00000752, -0.00001307, 0.00001493, -0.00002483, 0.00002906, -0.00003400, 0.00003112, -0.00002353, 0.00001721, -0.00001264, 0.00001202, -0.00001290, 0.00000148, 0.00000049, 0.00000594, -0.00000414, 0.00000620, -0.00000843, -0.00000333, 0.00000095, 0.00000316, 0.00000137, 0.00000306, -0.00000357, 0.00000044, -0.00000012, -0.00000067, 0.00000515, 0.00000312, -0.00000232, -0.00001512, 0.00001841, -0.00000307, 0.00001100, -0.00002823, 0.00002713, -0.00000464, -0.00001750, 0.00002455, -0.00002299, },
		{ -0.00000141, 0.00000165, 0.00000841, -0.00002302, 0.00000184, 0.00000677, 0.00000329, -0.00001026, 0.00000303, -0.00000619, 0.00000590, -0.00000805, 0.00000458, -0.00000274, -0.00000062, 0.00000028, 0.00000292, -0.00000489, 0.00000582, -0.00001383, 0.00000456, -0.00000480, 0.00000196, -0.00000143, 0.00000930, -0.00001844, 0.00000953, -0.00001395, 0.00001331, -0.00001822, 0.00001107, -0.00000812, -0.00000619, 0.00001219, -0.00001450, 0.00001572, -0.00000583, -0.00000546, -0.00000344, -0.00000304, -0.00000212, 0.00000089, -0.00000191, -0.00000098, 0.00000351, -0.00001939, -0.00000375, -0.00000210, -0.00000181, -0.00001812, -0.00000728, -0.00000979, 0.00000591, -0.00005006, -0.00000256, -0.00004804, -0.00000062, -0.00014378, -0.00000360, -0.00061086, -0.00030533, 0.00215186, 0.00037502, -0.00085418, -0.00001869, -0.00014156, -0.00000556, -0.00006550, -0.00000981, -0.00001976, -0.00000379, -0.00002862, -0.00000227, -0.00000505, -0.00000180, -0.00001445, -0.00000508, -0.00000046, 0.00000002, -0.00001427, -0.00000112, -0.00000160, -0.00000428, 0.00000083, -0.00000169, 0.00000061, -0.00000216, -0.00000212, -0.00000857, 0.00000657, -0.00000761, -0.00000246, 0.00000901, -0.00001777, 0.00002125, -0.00001814, 0.00001487, -0.00000997, 0.00000902, -0.00001266, 0.00000278, 0.00000271, 0.00000268, -0.00000136, 0.00000367, -0.00000857, -0.00000281, 0.00000405, 0.00000086, 0.00000293, 0.00000112, -0.00000351, 0.00000185, -0.00000074, 0.00000056, 0.00000616, 0.00000251, -0.00000999, -0.00000366, 0.00001414, 0.00000130, 0.00000264, -0.00002665, 0.00004465, -0.00004186, 0.00002570, -0.00001149, -0.00000051, },
		{ -0.00001996, 0.00002067, -0.00001233, -0.00001177, 0.00000005, 0.00001264, -0.00000781, -0.00000069, -0.00000613, 0.00000280, -0.00000203, -0.00000137, -0.00000009, 0.00000111, -0.00000626, 0.00000828, -0.00000664, 0.00000501, -0.00000450, -0.00000578, -0.00000160, 0.00000131, -0.00000483, 0.00000914, -0.00000459, -0.00000790, 0.00000145, -0.00000562, 0.00000557, -0.00001434, 0.00001171, -0.00001235, -0.00000118, 0.00000974, -0.00001487, 0.00002339, -0.00002047, 0.00000677, -0.00001386, 0.00000860, -0.00001414, 0.00001395, -0.00001576, 0.00001666, -0.00001987, -0.00000118, -0.00001743, 0.00001681, -0.00002801, 0.00000530, -0.00003082, 0.00002443, -0.00004298, -0.00000834, -0.00004410, 0.00000194, -0.00008436, -0.00006925, -0.00016509, -0.00075869, 0.00079734, 0.00158358, -0.00013281, -0.00066311, 0.00005490, -0.00017072, 0.00003411, -0.00009750, 0.00003223, -0.00004910, 0.00001700, -0.00004847, 0.00002345, -0.00002493, 0.00001257, -0.00003062, 0.00001477, -0.00001494, 0.00000946, -0.00002580, 0.00001364, -0.00001516, 0.00000719, -0.00000886, 0.00000907, -0.00000888, 0.00000483, -0.00001051, -0.00000004, 0.00000162, -0.00000816, 0.00000183, 0.00000359, -0.00001289, 0.00002039, -0.00002091, 0.00002153, -0.00001764, 0.00001586, -0.00002154, 0.00001442, -0.00000629, 0.00000988, -0.00000825, 0.00001017, -0.00001789, 0.00000839, -0.00000330, 0.00000705, -0.00000310, 0.00000605, -0.00000966, 0.00001047, -0.00000957, 0.00001064, -0.00000103, 0.00000579, -0.00001913, 0.00001414, -0.00000122, 0.00001946, -0.00002816, 0.00001218, 0.00001200, -0.00002858, 0.00003237, -0.00003082, 0.00002043, },
		{ -0.00001548, 0.00002218, -0.00002311, -0.00000302, 0.00000130, 0.00001121, -0.00001118, 0.00000244, -0.00001014, 0.00000856, -0.00000811, 0.00000447, -0.00000426, 0.00000301, -0.00000834, 0.00001257, -0.00001340, 0.00001350, -0.00001497, 0.00000340, -0.00000906, 0.00000831, -0.00001109, 0.00001832, -0.00001940, 0.00000655, -0.00001121, 0.00000764, -0.00000899, -0.00000157, 0.00000341, -0.00000967, -0.00000031, 0.00000918, -0.00001381, 0.00002650, -0.00003273, 0.00002062, -0.00002637, 0.00002180, -0.00002813, 0.00002970, -0.00003129, 0.00003391, -0.00004594, 0.00002535, -0.00003895, 0.00004017, -0.00006094, 0.00003963, -0.00006295, 0.00006330, -0.00010499, 0.00006099, -0.00012048, 0.00009022, -0.00023514, 0.00012172, -0.00065603, -0.00007978, 0.00111521, 0.00102737, -0.00039248, -0.00046876, 0.00007240, -0.00017786, 0.00005457, -0.00011063, 0.00006005, -0.00007489, 0.00003615, -0.00006317, 0.00004483, -0.00004472, 0.00002592, -0.00004226, 0.00003147, -0.00003029, 0.00001932, -0.00003515, 0.00002616, -0.00002824, 0.00001892, -0.00001755, 0.00001881, -0.00001947, 0.00001302, -0.00001923, 0.00001010, -0.00000716, -0.00000452, 0.00000392, -0.00000234, -0.00000450, 0.00001343, -0.00001566, 0.00001981, -0.00001823, 0.00001601, -0.00002312, 0.00001975, -0.00001043, 0.00001210, -0.00000950, 0.00001007, -0.00002029, 0.00001477, -0.00000746, 0.00000925, -0.00000418, 0.00000474, -0.00000762, 0.00001047, -0.00001045, 0.00001430, -0.00000342, 0.00000083, -0.00001313, 0.00001487, 0.00000113, 0.00001566, -0.00003704, 0.00003967, -0.00002985, 0.00001486, -0.00000212, -0.00001080, 0.00001061, },
		{ 0.00000395, 0.00000686, -0.00001932, -0.00000160, 0.00000679, 0.00000276, -0.00000556, -0.00000443, -0.00000292, 0.00000330, -0.00000377, 0.00000069, 0.00000027, -0.00000464, 0.00000157, 0.00000278, -0.00000490, 0.00000652, -0.00001072, -0.00000154, -0.00000218, 0.00000057, -0.00000081, 0.00000798, -0.00001424, 0.00000276, -0.00000604, 0.00000237, -0.00000559, -0.00000469, 0.00000974, -0.00002288, 0.00001971, -0.00001335, 0.00001317, -0.00000286, -0.00000937, -0.00000020, -0.00000368, -0.00000272, -0.00000180, 0.00000365, -0.00000242, -0.00000003, -0.00001615, -0.00000336, -0.00000269, -0.00000318, -0.00002021, -0.00000268, -0.00000937, -0.00000356, -0.00004427, -0.00000291, -0.00004448, -0.00001836, -0.00013473, -0.00004908, -0.00074517, 0.00070970, 0.00081186, 0.00079920, -0.00067983, -0.00014455, -0.00007633, -0.00005420, -0.00003770, -0.00001995, -0.00000968, -0.00001780, -0.00001605, -0.00001031, 0.00000239, -0.00000922, -0.00001033, -0.00000422, -0.00000024, -0.00000295, -0.00000962, -0.00000621, 0.00000146, -0.00000728, 0.00000011, 0.00000352, -0.00000123, -0.00000309, -0.00000324, -0.00000431, -0.00000155, 0.00000200, -0.00001536, 0.00001934, -0.00002209, 0.00001933, -0.00001063, 0.00000847, -0.00000209, 0.00000060, -0.00000227, -0.00000617, 0.00000767, 0.00000063, 0.00000025, 0.00000310, -0.00000473, -0.00000683, 0.00000690, -0.00000049, 0.00000177, 0.00000371, -0.00000551, 0.00000510, -0.00000165, 0.00000127, 0.00000661, 0.00000117, -0.00000934, 0.00000276, 0.00000207, 0.00002001, -0.00001465, -0.00000614, 0.00001873, -0.00002647, 0.00002853, -0.00002692, 0.00001457, -0.00001144, },
		{ 0.00001291, -0.00000554, -0.00001363, 0.00000067, 0.00000751, -0.00000125, -0.00000360, -0.00000838, 0.00000346, -0.00000230, 0.00000118, -0.00000345, 0.00000365, -0.00001023, 0.00001047, -0.00000783, 0.00000620, -0.00000421, -0.00000309, -0.00000900, 0.00000703, -0.00000899, 0.00001150, -0.00000694, -0.00000286, -0.00000616, 0.00000315, -0.00000711, 0.00000173, -0.00000924, 0.00001389, -0.00003191, 0.00003650, -0.00003337, 0.00003825, -0.00003491, 0.00002113, -0.00002814, 0.00002519, -0.00003401, 0.00003293, -0.00003075, 0.00003326, -0.00004382, 0.00002839, -0.00004615, 0.00004552, -0.00006236, 0.00004147, -0.00006474, 0.00006207, -0.00009844, 0.00005954, -0.00011799, 0.00008901, -0.00021607, 0.00010669, -0.00054924, -0.00010709, 0.00083622, 0.00061980, 0.00054262, -0.00078822, 0.00009367, -0.00021254, 0.00006247, -0.00012376, 0.00006790, -0.00008578, 0.00004630, -0.00007164, 0.00004676, -0.00004744, 0.00003131, -0.00004747, 0.00003567, -0.00003697, 0.00002857, -0.00004083, 0.00002565, -0.00002758, 0.00001763, -0.00002006, 0.00002513, -0.00002381, 0.00001583, -0.00002068, 0.00001175, -0.00001420, 0.00000987, -0.00002127, 0.00002731, -0.00003361, 0.00003592, -0.00002883, 0.00002852, -0.00002246, 0.00001909, -0.00002043, 0.00001167, -0.00000544, 0.00001118, -0.00000975, 0.00001310, -0.00001760, 0.00000718, -0.00000227, 0.00000580, -0.00000308, 0.00000716, -0.00000958, 0.00001195, -0.00000933, 0.00001041, 0.00000030, 0.00000065, -0.00000816, 0.00000570, 0.00000364, 0.00002085, -0.00002979, 0.00002134, -0.00001706, 0.00000947, 0.00000117, -0.00001433, 0.00001273, -0.00001571, },
		{ 0.00000830, -0.00000871, -0.00001086, 0.00000480, 0.00000348, 0.00000067, -0.00000886, -0.00000344, 0.00000139, -0.00000020, -0.00000132, -0.00000025, -0.00000172, -0.00000477, 0.00000719, -0.00000630, 0.00000595, -0.00000471, -0.00000597, -0.00000454, 0.00000365, -0.00000482, 0.00000848, -0.00000721, -0.00000463, -0.00000127, -0.00000320, -0.00000020, -0.00000767, 0.00000511, -0.00000574, -0.00001189, 0.00002207, -0.00002067, 0.00002730, -0.00003077, 0.00001793, -0.00002204, 0.00001726, -0.00002594, 0.00002753, -0.00002344, 0.00002120, -0.00003558, 0.00002108, -0.00003400, 0.00003056, -0.00005292, 0.00003274, -0.00004848, 0.00003750, -0.00008330, 0.00004556, -0.00009945, 0.00005159, -0.00020093, 0.00005299, -0.00074643, 0.00067257, 0.00042320, 0.00076526, 0.00006750, -0.00058360, 0.00008458, -0.00018370, 0.00004633, -0.00009331, 0.00004919, -0.00007214, 0.00003329, -0.00005542, 0.00003750, -0.00004013, 0.00002004, -0.00003268, 0.00002508, -0.00002866, 0.00001873, -0.00003205, 0.00001992, -0.00002284, 0.00001201, -0.00000959, 0.00001539, -0.00001792, 0.00000909, -0.00001335, 0.00000433, -0.00000533, -0.00000335, -0.00000412, 0.00000972, -0.00001792, 0.00002566, -0.00002121, 0.00002392, -0.00002078, 0.00001806, -0.00002072, 0.00001403, -0.00000504, 0.00000836, -0.00000578, 0.00000826, -0.00001558, 0.00000904, -0.00000231, 0.00000366, 0.00000074, 0.00000107, -0.00000208, 0.00000616, -0.00000471, 0.00000927, -0.00000005, -0.00000446, 0.00000192, -0.00000363, 0.00002159, -0.00000496, -0.00000796, 0.00000642, -0.00001419, 0.00002118, -0.00002092, 0.00000691, -0.00000317, -0.00000528, },
		{ 0.00000536, -0.00001162, -0.00000513, 0.00000350, 0.00000413, -0.00000198, -0.00001046, 0.00000056, -0.00000084, 0.00000209, -0.00000413, 0.00000262, -0.00000691, 0.00000224, 0.00000079, -0.00000069, 0.00000156, -0.00000231, -0.00001126, 0.00000339, -0.00000349, 0.00000330, 0.00000004, -0.00000196, -0.00001020, 0.00000659, -0.00001305, 0.00001057, -0.00002013, 0.00002220, -0.00003084, 0.00001892, -0.00000622, 0.00000676, -0.00000172, -0.00000612, -0.00000444, 0.00000217, -0.00001009, 0.00000405, -0.00000000, 0.00000441, -0.00001454, 0.00000170, -0.00001602, 0.00000792, -0.00001960, -0.00000237, -0.00001705, 0.00000879, -0.00003981, 0.00000101, -0.00004791, 0.00000856, -0.00010964, -0.00000881, -0.00034565, -0.00027406, 0.00086659, 0.00014022, 0.00087517, -0.00037072, -0.00027847, -0.00003773, -0.00007945, -0.00002597, -0.00001571, -0.00001795, -0.00001514, -0.00001553, -0.00000620, -0.00000410, -0.00000576, -0.00001256, 0.00000313, -0.00000841, 0.00000093, -0.00001005, -0.00000499, -0.00000339, -0.00000281, -0.00000525, 0.00001162, -0.00000720, 0.00000026, -0.00000754, 0.00000270, -0.00001090, 0.00000945, -0.00002093, 0.00001805, -0.00001533, 0.00000821, 0.00000367, -0.00000158, 0.00000663, -0.00000698, 0.00000637, -0.00001150, 0.00000900, 0.00000020, 0.00000190, 0.00000159, -0.00000088, -0.00000779, 0.00000601, -0.00000036, 0.00000159, 0.00000308, -0.00000260, 0.00000388, 0.00000049, 0.00000097, 0.00000695, -0.00000353, -0.00000016, 0.00000046, 0.00000052, 0.00002373, -0.00001958, 0.00001312, -0.00002288, 0.00001901, -0.00000536, -0.00000904, 0.00000503, -0.00000562, -0.00000166, },
		{ 0.00000773, -0.00001609, 0.00000238, -0.00000160, 0.00000927, -0.00001110, -0.00000383, -0.00000257, 0.00000320, -0.00000198, -0.00000074, -0.00000149, -0.00000392, 0.00000113, 0.00000197, -0.00000211, 0.00000391, -0.00000814, -0.00000672, 0.00000181, -0.00000101, 0.00000073, 0.00000215, -0.00000735, -0.00000317, -0.00000050, -0.00000648, 0.00000394, -0.00001305, 0.00001654, -0.00003188, 0.00002864, -0.00001610, 0.00001500, -0.00001222, 0.00000193, -0.00000909, 0.00000589, -0.00001525, 0.00001251, -0.00000538, 0.00000459, -0.00001890, 0.00000861, -0.00002137, 0.00001270, -0.00003010, 0.00000913, -0.00002251, 0.00000771, -0.00004911, 0.00001465, -0.00006184, 0.00001073, -0.00013860, -0.00000311, -0.00058036, 0.00043759, 0.00048853, 0.00027031, 0.00067823, -0.00053137, -0.00010362, -0.00009580, -0.00003293, -0.00004205, 0.00000550, -0.00003880, 0.00000312, -0.00002737, 0.00000841, -0.00001673, 0.00000166, -0.00001593, 0.00000962, -0.00001655, 0.00000796, -0.00001732, 0.00000212, -0.00000869, 0.00000024, -0.00000357, 0.00001240, -0.00001211, 0.00000299, -0.00000839, 0.00000261, -0.00001009, 0.00000744, -0.00002026, 0.00002169, -0.00002353, 0.00002113, -0.00000792, 0.00000907, -0.00000377, 0.00000168, -0.00000095, -0.00000593, 0.00000801, -0.00000016, 0.00000220, 0.00000155, -0.00000304, -0.00000478, 0.00000629, -0.00000237, 0.00000456, -0.00000075, 0.00000161, 0.00000147, 0.00000244, 0.00000089, 0.00000698, -0.00000887, 0.00001081, -0.00001206, 0.00002126, -0.00000057, 0.00000021, -0.00000528, -0.00001459, 0.00002785, -0.00002621, 0.00001042, -0.00000854, 0.00000279, -0.00000679, },
		{ 0.00000544, -0.00001452, 0.00000305, -0.00000030, 0.00000776, -0.00001534, 0.00000121, -0.00000490, 0.00000642, -0.00000584, 0.00000226, -0.00000535, -0.00000018, -0.00000118, 0.00000450, -0.00000470, 0.00000669, -0.00001526, 0.00000110, -0.00000298, 0.00000398, -0.00000503, 0.00000761, -0.00001577, 0.00000773, -0.00001338, 0.00000783, -0.00001169, 0.00000476, -0.00000358, -0.00001455, 0.00002063, -0.00001146, 0.00000867, -0.00000767, -0.00000347, -0.00000076, -0.00000512, -0.00000342, 0.00000469, 0.00000322, -0.00001240, -0.00000046, -0.00000835, -0.00000308, -0.00001028, -0.00000861, -0.00001006, 0.00000180, -0.00003360, -0.00000330, -0.00003533, -0.00000473, -0.00008716, -0.00002283, -0.00028177, -0.00027692, 0.00072953, 0.00010878, 0.00052403, 0.00032619, -0.00049121, -0.00004385, -0.00009774, -0.00002611, -0.00002454, -0.00001085, -0.00002572, -0.00000598, -0.00001433, -0.00000132, -0.00000940, -0.00000660, -0.00000290, -0.00000271, -0.00000700, -0.00000042, -0.00001055, -0.00000357, -0.00000298, -0.00000538, 0.00000743, 0.00000096, -0.00000552, -0.00000336, -0.00000071, -0.00000643, -0.00000020, -0.00000428, -0.00000814, 0.00001214, -0.00001801, 0.00002260, -0.00001080, 0.00001218, -0.00000824, 0.00000654, -0.00000619, -0.00000059, 0.00000619, -0.00000005, 0.00000258, 0.00000062, -0.00000367, -0.00000209, 0.00000487, -0.00000172, 0.00000436, -0.00000105, 0.00000316, 0.00000079, 0.00000316, 0.00000245, 0.00000117, -0.00000247, 0.00000814, -0.00000999, 0.00002716, -0.00001566, 0.00001982, -0.00003784, 0.00002516, -0.00000357, -0.00001147, 0.00000549, -0.00000688, 0.00000136, -0.00000562, },
		{ -0.00000332, -0.00000735, -0.00000225, 0.00000764, -0.00000285, -0.00000951, -0.00000179, -0.00000013, 0.00000253, -0.00000337, -0.00000114, -0.00000219, -0.00000342, 0.00000380, -0.00000059, 0.00000067, 0.00000002, -0.00001251, 0.00000070, 0.00000030, -0.00000074, -0.00000024, 0.00000185, -0.00001177, 0.00000528, -0.00001282, 0.00000954, -0.00001523, 0.00001073, -0.00001402, -0.00000231, 0.00001613, -0.00001378, 0.00001148, -0.00001257, 0.00000249, -0.00000615, -0.00000158, -0.00000495, 0.00001118, -0.00000835, -0.00000599, -0.00000330, -0.00000418, -0.00000988, -0.00000636, -0.00001318, 0.00000162, -0.00001760, -0.00002333, -0.00000913, -0.00002841, -0.00002550, -0.00008761, -0.00004424, -0.00045705, 0.00027636, 0.00051437, 0.00005755, 0.00060462, 0.00004768, -0.00041681, 0.00000056, -0.00011489, 0.00000612, -0.00003853, -0.00000286, -0.00003190, 0.00000448, -0.00002064, 0.00000485, -0.00001739, 0.00000319, -0.00000830, -0.00000029, -0.00000963, 0.00000284, -0.00001528, 0.00000194, -0.00000836, 0.00000201, 0.00000480, -0.00000027, -0.00000727, -0.00000043, -0.00000286, -0.00000626, 0.00000135, -0.00000871, -0.00000128, 0.00000501, -0.00001193, 0.00002325, -0.00001445, 0.00001641, -0.00001396, 0.00001356, -0.00001461, 0.00000953, -0.00000177, 0.00000648, -0.00000346, 0.00000568, -0.00000908, 0.00000488, -0.00000150, 0.00000444, -0.00000188, 0.00000585, -0.00000284, 0.00000720, -0.00000208, 0.00000747, -0.00000827, 0.00001290, -0.00000871, 0.00001197, 0.00000483, 0.00000391, -0.00000088, -0.00003142, 0.00004088, -0.00003144, 0.00001317, -0.00001294, 0.00000734, -0.00001118, 0.00000475, },
		{ -0.00000939, -0.00000388, -0.00000294, 0.00001066, -0.00001114, -0.00000294, -0.00000552, 0.00000521, -0.00000259, -0.00000029, -0.00000451, 0.00000114, -0.00000690, 0.00000962, -0.00000724, 0.00000778, -0.00000997, -0.00000501, -0.00000307, 0.00000566, -0.00000855, 0.00000897, -0.00000933, -0.00000112, -0.00000530, -0.00000268, 0.00000131, -0.00000904, 0.00000651, -0.00001480, 0.00000481, 0.00001206, -0.00001726, 0.00001783, -0.00002154, 0.00001334, -0.00001830, 0.00001032, -0.00001339, 0.00002238, -0.00002844, 0.00001470, -0.00002077, 0.00001295, -0.00003214, 0.00001611, -0.00003439, 0.00002860, -0.00006216, 0.00002596, -0.00006040, 0.00002832, -0.00012027, 0.00002406, -0.00029578, -0.00020035, 0.00053273, 0.00022076, 0.00017515, 0.00054526, -0.00014733, -0.00032259, 0.00001681, -0.00011989, 0.00003437, -0.00006063, 0.00001381, -0.00004317, 0.00001985, -0.00003360, 0.00001569, -0.00002776, 0.00001724, -0.00002108, 0.00000886, -0.00001678, 0.00000941, -0.00002317, 0.00001057, -0.00001608, 0.00001323, -0.00000429, 0.00000364, -0.00001193, 0.00000568, -0.00000933, -0.00000138, -0.00000161, -0.00000916, 0.00000282, -0.00000157, -0.00000261, 0.00001842, -0.00001326, 0.00001602, -0.00001457, 0.00001537, -0.00001803, 0.00001560, -0.00000684, 0.00001051, -0.00000734, 0.00000906, -0.00001244, 0.00000903, -0.00000461, 0.00000679, -0.00000363, 0.00000867, -0.00000542, 0.00001072, -0.00000437, 0.00000601, -0.00000602, 0.00001555, -0.00001345, 0.00002307, -0.00001236, 0.00002669, -0.00004021, 0.00001234, 0.00001158, -0.00002141, 0.00001247, -0.00001541, 0.00001114, -0.00001748, 0.00001211, },
		{ -0.00001009, -0.00000543, 0.00000302, 0.00000444, -0.00001026, -0.00000310, -0.00000308, 0.00000440, -0.00000262, -0.00000244, -0.00000151, -0.00000274, -0.00000207, 0.00000683, -0.00000564, 0.00000624, -0.00001211, -0.00000348, -0.00000006, 0.00000177, -0.00000620, 0.00000839, -0.00001132, 0.00000074, -0.00000771, 0.00000085, -0.00000143, -0.00000834, 0.00000740, -0.00001988, 0.00001939, -0.00000633, -0.00000294, 0.00000648, -0.00001244, 0.00000518, -0.00001123, 0.00000394, -0.00000179, 0.00000754, -0.00002006, 0.00000987, -0.00001309, 0.00000033, -0.00002172, 0.00000588, -0.00001763, 0.00000482, -0.00004874, 0.00001838, -0.00005047, 0.00000394, -0.00011615, 0.00000447, -0.00045285, 0.00026781, 0.00038524, 0.00013757, 0.00024168, 0.00047206, -0.00032358, -0.00017240, -0.00003803, -0.00006257, 0.00000448, -0.00003838, -0.00000537, -0.00002034, 0.00000272, -0.00001888, 0.00000109, -0.00001155, 0.00000546, -0.00001249, -0.00000038, -0.00000595, -0.00000229, -0.00001341, 0.00000291, -0.00000712, 0.00000851, -0.00000119, -0.00000325, -0.00000483, 0.00000024, -0.00000600, -0.00000439, 0.00000196, -0.00001534, 0.00001242, -0.00001398, 0.00001544, 0.00000175, 0.00000007, 0.00000395, -0.00000327, 0.00000502, -0.00000921, 0.00001007, -0.00000136, 0.00000429, -0.00000113, 0.00000318, -0.00000751, 0.00000542, -0.00000019, 0.00000130, 0.00000374, 0.00000160, 0.00000184, 0.00000534, -0.00000021, -0.00000192, 0.00000806, 0.00000134, 0.00000323, 0.00000607, 0.00000415, 0.00001026, -0.00004363, 0.00003847, -0.00002170, 0.00000484, -0.00000672, -0.00000080, -0.00000049, -0.00001085, 0.00001098, },
		{ -0.00001182, -0.00000371, 0.00000598, -0.00000183, -0.00000716, -0.00000482, 0.00000086, 0.00000157, -0.00000175, -0.00000464, 0.00000195, -0.00000789, 0.00000573, -0.00000019, 0.00000028, -0.00000045, -0.00000912, -0.00000489, 0.00000515, -0.00000630, 0.00000233, 0.00000094, -0.00000677, -0.00000380, -0.00000357, -0.00000097, -0.00000084, -0.00000986, 0.00000955, -0.00002355, 0.00003192, -0.00002769, 0.00001974, -0.00001501, 0.00000730, -0.00001459, 0.00000841, -0.00001362, 0.00001977, -0.00002193, 0.00000852, -0.00001473, 0.00001218, -0.00003276, 0.00001332, -0.00002814, 0.00002273, -0.00005359, 0.00001372, -0.00004426, 0.00001852, -0.00010342, 0.00000849, -0.00024153, -0.00021523, 0.00047559, 0.00017859, 0.00015574, 0.00028138, 0.00035785, -0.00043013, -0.00003229, -0.00010773, 0.00001201, -0.00005031, 0.00000530, -0.00004017, 0.00001677, -0.00002984, 0.00000869, -0.00002315, 0.00001512, -0.00001901, 0.00000730, -0.00001800, 0.00001211, -0.00002179, 0.00000445, -0.00001147, 0.00000887, -0.00000439, 0.00000784, -0.00001396, 0.00000643, -0.00000990, 0.00000114, -0.00000924, 0.00000524, -0.00001964, 0.00001909, -0.00002220, 0.00003058, -0.00001544, 0.00001552, -0.00001037, 0.00001059, -0.00000837, 0.00000313, 0.00000137, 0.00000633, -0.00000393, 0.00000768, -0.00000550, -0.00000066, 0.00000131, 0.00000325, -0.00000202, 0.00000928, -0.00000464, 0.00000925, -0.00000046, 0.00000098, -0.00000176, 0.00001350, -0.00000555, 0.00001383, -0.00000993, 0.00002792, -0.00002963, -0.00000603, 0.00001988, -0.00002069, 0.00001034, -0.00001464, 0.00000852, -0.00001166, -0.00000117, 0.00000755, },
		{ -0.00001488, 0.00000262, 0.00000199, -0.00000192, -0.00000826, -0.00000228, 0.00000060, 0.00000196, -0.00000481, -0.00000124, -0.00000123, -0.00000576, 0.00000709, -0.00000218, 0.00000127, -0.00000291, -0.00000965, -0.00000075, 0.00000235, -0.00000640, 0.00000445, -0.00000123, -0.00000785, -0.00000209, -0.00000526, 0.00000356, -0.00000904, -0.00000013, -0.00000130, -0.00001030, 0.00002263, -0.00002727, 0.00002431, -0.00002051, 0.00001093, -0.00001784, 0.00001199, -0.00001319, 0.00001813, -0.00002706, 0.00001630, -0.00001855, 0.00001142, -0.00003698, 0.00002040, -0.00003003, 0.00001957, -0.00006342, 0.00003116, -0.00005923, 0.00002221, -0.00013194, 0.00003192, -0.00042964, 0.00023771, 0.00031565, 0.00015761, 0.00011066, 0.00038007, 0.00013959, -0.00039359, 0.00001094, -0.00011103, 0.00002875, -0.00006252, 0.00001430, -0.00004162, 0.00002221, -0.00003605, 0.00001252, -0.00002240, 0.00001727, -0.00002314, 0.00000927, -0.00001758, 0.00001180, -0.00002440, 0.00000768, -0.00001104, 0.00001034, -0.00000537, 0.00000528, -0.00001200, 0.00000548, -0.00000935, -0.00000148, -0.00000381, -0.00000344, -0.00001034, 0.00001111, -0.00001369, 0.00002843, -0.00001799, 0.00001889, -0.00001380, 0.00001453, -0.00001299, 0.00000797, -0.00000027, 0.00000621, -0.00000387, 0.00000913, -0.00000856, 0.00000155, 0.00000223, 0.00000044, 0.00000294, 0.00000539, -0.00000164, 0.00000880, -0.00000125, -0.00000281, 0.00000888, 0.00000374, 0.00000648, -0.00000074, 0.00000483, 0.00001636, -0.00004001, 0.00002181, -0.00000854, -0.00000086, -0.00000332, -0.00000586, 0.00000409, -0.00001337, 0.00000512, 0.00000409, },
		{ -0.00001166, 0.00000455, -0.00000077, -0.00000175, -0.00000913, 0.00000049, -0.00000032, 0.00000188, -0.00000707, 0.00000251, -0.00000626, 0.00000007, 0.00000423, -0.00000099, -0.00000077, -0.00000288, -0.00001115, 0.00000499, -0.00000462, -0.00000077, 0.00000144, 0.00000011, -0.00001212, 0.00000351, -0.00001015, 0.00001027, -0.00002039, 0.00001502, -0.00001887, 0.00001231, -0.00000200, -0.00000747, 0.00001032, -0.00000950, -0.00000158, -0.00000418, -0.00000043, 0.00000274, -0.00000370, -0.00000720, 0.00000014, -0.00000025, -0.00001555, -0.00000988, -0.00000320, -0.00000119, -0.00002567, -0.00001724, -0.00001217, -0.00000984, -0.00006223, -0.00003623, -0.00016240, -0.00026150, 0.00046930, 0.00009810, 0.00021393, 0.00005646, 0.00046955, -0.00011060, -0.00026455, -0.00001913, -0.00006298, -0.00000354, -0.00003501, -0.00000792, -0.00001475, -0.00000038, -0.00001855, -0.00000261, -0.00000283, -0.00000107, -0.00000887, -0.00000407, -0.00000280, -0.00000345, -0.00001276, -0.00000053, 0.00000014, 0.00000037, 0.00000316, -0.00000539, -0.00000153, -0.00000394, -0.00000170, -0.00000935, 0.00000595, -0.00001725, 0.00000569, -0.00000453, 0.00000494, 0.00001358, -0.00000834, 0.00001191, -0.00000813, 0.00001040, -0.00001091, 0.00000801, 0.00000156, 0.00000235, 0.00000105, 0.00000555, -0.00000845, 0.00000306, 0.00000226, -0.00000084, 0.00000742, 0.00000030, 0.00000367, 0.00000619, -0.00000383, 0.00000060, 0.00001212, -0.00000009, 0.00001334, -0.00001532, 0.00002857, -0.00002010, -0.00001083, 0.00001004, -0.00000867, 0.00000295, -0.00000940, 0.00000089, -0.00000227, -0.00001147, 0.00001112, -0.00000406, },
		{ -0.00000207, -0.00000104, 0.00000282, -0.00000660, -0.00000493, -0.00000127, 0.00000239, -0.00000278, -0.00000338, -0.00000001, -0.00000535, 0.00000163, 0.00000434, -0.00000347, 0.00000121, -0.00000725, -0.00000623, 0.00000300, -0.00000489, -0.00000011, 0.00000319, -0.00000524, -0.00000813, 0.00000092, -0.00000592, 0.00000523, -0.00001837, 0.00001734, -0.00002360, 0.00002173, -0.00001682, 0.00000740, -0.00000015, -0.00000387, -0.00000702, 0.00000252, -0.00000434, 0.00000599, -0.00001251, 0.00000320, -0.00000627, 0.00000341, -0.00002665, 0.00000418, -0.00001083, 0.00000176, -0.00004227, 0.00000609, -0.00003155, 0.00000088, -0.00009994, 0.00000045, -0.00034678, 0.00014281, 0.00034266, 0.00007015, 0.00017524, 0.00011057, 0.00043297, -0.00026914, -0.00015068, -0.00004302, -0.00003098, -0.00002558, -0.00001641, -0.00001883, 0.00000029, -0.00001476, -0.00000866, -0.00000683, 0.00000488, -0.00001078, -0.00000171, -0.00000934, 0.00000352, -0.00001186, -0.00000645, -0.00000173, 0.00000318, -0.00000292, 0.00000497, -0.00000853, 0.00000177, -0.00000698, -0.00000058, -0.00000926, 0.00000607, -0.00002131, 0.00001335, -0.00001250, 0.00001801, 0.00000051, 0.00000151, 0.00000460, -0.00000220, 0.00000595, -0.00000896, 0.00000978, -0.00000022, 0.00000266, 0.00000312, 0.00000287, -0.00000912, 0.00000716, -0.00000233, 0.00000466, 0.00000436, 0.00000191, 0.00000415, 0.00000570, -0.00000925, 0.00001310, 0.00000148, 0.00001364, -0.00000418, -0.00000017, 0.00001975, -0.00003145, 0.00001190, -0.00001146, 0.00000759, -0.00000962, -0.00000167, -0.00000245, -0.00000337, -0.00000942, 0.00001585, -0.00001483, },
		{ 0.00000330, -0.00000507, 0.00000598, -0.00001119, -0.00000028, -0.00000356, 0.00000460, -0.00000709, 0.00000127, -0.00000465, -0.00000133, 0.00000069, 0.00000538, -0.00000700, 0.00000458, -0.00001286, 0.00000128, -0.00000357, -0.00000004, -0.00000330, 0.00000758, -0.00001443, 0.00000184, -0.00000798, 0.00000491, -0.00000897, -0.00000389, 0.00000615, -0.00001444, 0.00001527, -0.00001585, 0.00001025, -0.00000167, -0.00000750, -0.00000142, -0.00000148, 0.00000201, -0.00000510, -0.00000304, -0.00000341, 0.00000299, -0.00001306, -0.00001341, -0.00000388, 0.00000279, -0.00002715, -0.00001477, -0.00001720, -0.00000200, -0.00005916, -0.00003543, -0.00014391, -0.00024028, 0.00039456, 0.00011761, 0.00015311, 0.00007465, 0.00023921, 0.00028244, -0.00031630, -0.00008084, -0.00004519, -0.00002619, -0.00002512, -0.00001566, -0.00001249, -0.00000273, -0.00001498, -0.00000908, 0.00000053, -0.00000170, -0.00000775, -0.00000485, -0.00000462, -0.00000067, -0.00001111, -0.00000646, 0.00000365, -0.00000224, 0.00000154, -0.00000031, -0.00000428, -0.00000223, -0.00000349, -0.00000532, -0.00000266, -0.00000232, -0.00001543, 0.00001158, -0.00001093, 0.00002222, -0.00000680, 0.00000852, -0.00000138, 0.00000352, 0.00000044, -0.00000492, 0.00000983, -0.00000198, 0.00000457, 0.00000377, -0.00000087, -0.00000653, 0.00000770, -0.00000384, 0.00000879, 0.00000087, 0.00000457, 0.00000446, 0.00000102, -0.00000544, 0.00001708, -0.00000257, 0.00002191, -0.00002286, 0.00002607, -0.00001479, -0.00000508, -0.00000205, -0.00000593, 0.00000719, -0.00001342, 0.00000259, -0.00000523, -0.00000498, -0.00000272, 0.00001208, -0.00001810, },
		{ -0.00000052, -0.00000110, 0.00000226, -0.00000962, -0.00000070, -0.00000146, 0.00000156, -0.00000541, 0.00000016, -0.00000406, -0.00000128, 0.00000348, 0.00000106, -0.00000450, 0.00000192, -0.00001172, 0.00000197, -0.00000442, 0.00000010, -0.00000103, 0.00000427, -0.00001535, 0.00000501, -0.00001033, 0.00000824, -0.00001647, 0.00000707, -0.00000348, -0.00000694, 0.00000909, -0.00001354, 0.00001370, -0.00000781, -0.00000453, -0.00000167, 0.00000127, -0.00000210, -0.00000537, -0.00000120, -0.00000200, 0.00000011, -0.00001747, -0.00000893, 0.00000038, -0.00000585, -0.00003196, -0.00000507, -0.00002158, -0.00000446, -0.00008241, -0.00001455, -0.00029346, 0.00008868, 0.00033259, 0.00004158, 0.00017386, 0.00003670, 0.00032682, 0.00011286, -0.00030760, -0.00002500, -0.00005783, -0.00001516, -0.00003018, -0.00000744, -0.00001383, -0.00000134, -0.00002002, -0.00000075, -0.00000156, -0.00000246, -0.00000907, -0.00000275, -0.00000522, -0.00000095, -0.00001389, -0.00000032, 0.00000182, -0.00000255, 0.00000177, -0.00000128, -0.00000399, -0.00000276, -0.00000348, -0.00000600, -0.00000013, -0.00000858, -0.00000917, 0.00000868, -0.00000712, 0.00002300, -0.00001171, 0.00001512, -0.00000834, 0.00001128, -0.00000866, 0.00000469, 0.00000328, 0.00000221, 0.00000255, 0.00000660, -0.00000769, 0.00000145, 0.00000140, 0.00000272, 0.00000485, 0.00000406, 0.00000231, 0.00000777, -0.00000838, 0.00000983, 0.00000512, 0.00001287, 0.00000383, -0.00001151, 0.00002191, -0.00002695, 0.00001582, -0.00002502, 0.00001582, -0.00001065, -0.00000212, -0.00000426, -0.00000131, -0.00000953, 0.00000774, 0.00000038, -0.00001155, },
		{ -0.00000538, 0.00000490, -0.00000400, -0.00000541, -0.00000313, 0.00000169, -0.00000276, -0.00000166, -0.00000347, -0.00000054, -0.00000344, 0.00000765, -0.00000583, 0.00000175, -0.00000468, -0.00000610, -0.00000254, -0.00000005, -0.00000411, 0.00000549, -0.00000553, -0.00000775, 0.00000007, -0.00000492, 0.00000235, -0.00001368, 0.00000989, -0.00000802, -0.00000410, 0.00000709, -0.00001317, 0.00001867, -0.00001833, 0.00000572, -0.00000878, 0.00001009, -0.00001593, 0.00000792, -0.00001192, 0.00001087, -0.00001822, -0.00000369, -0.00001925, 0.00001879, -0.00003950, -0.00000048, -0.00003304, 0.00001367, -0.00006671, -0.00001866, -0.00014759, -0.00020085, 0.00031172, 0.00015437, 0.00008226, 0.00012460, 0.00007369, 0.00034132, -0.00003539, -0.00026196, 0.00001127, -0.00007527, 0.00000219, -0.00004045, 0.00000693, -0.00002284, 0.00000460, -0.00002728, 0.00001264, -0.00001198, 0.00000360, -0.00001506, 0.00000393, -0.00001067, 0.00000230, -0.00001844, 0.00000932, -0.00000610, 0.00000252, -0.00000214, 0.00000154, -0.00000736, -0.00000002, -0.00000638, -0.00000335, -0.00000168, -0.00001166, -0.00000346, 0.00000497, -0.00000104, 0.00001940, -0.00001152, 0.00001711, -0.00001118, 0.00001512, -0.00001460, 0.00001296, -0.00000388, 0.00000800, -0.00000016, 0.00000751, -0.00001138, 0.00000707, -0.00000317, 0.00000888, 0.00000010, 0.00000804, 0.00000051, 0.00000657, -0.00000931, 0.00001849, -0.00000266, 0.00002603, -0.00002004, 0.00001531, -0.00000997, -0.00000113, -0.00000065, -0.00001911, 0.00001948, -0.00001968, 0.00000622, -0.00001066, 0.00000202, -0.00001036, 0.00001289, -0.00000860, -0.00000515, },
		{ -0.00000375, 0.00000553, -0.00000633, -0.00000426, -0.00000256, 0.00000103, -0.00000278, -0.00000217, -0.00000279, -0.00000083, -0.00000148, 0.00000637, -0.00000753, 0.00000390, -0.00000741, -0.00000449, -0.00000311, 0.00000067, -0.00000396, 0.00000643, -0.00001071, -0.00000244, -0.00000301, -0.00000204, -0.00000191, -0.00001033, 0.00001264, -0.00001625, 0.00000472, -0.00000159, -0.00000362, 0.00001171, -0.00001716, 0.00000652, -0.00000594, 0.00000515, -0.00001546, 0.00001010, -0.00001148, 0.00000911, -0.00002230, -0.00000184, -0.00001276, 0.00001052, -0.00004562, 0.00001012, -0.00003899, 0.00001524, -0.00009253, 0.00000460, -0.00028645, 0.00008874, 0.00027218, 0.00007868, 0.00009980, 0.00009386, 0.00011793, 0.00031923, -0.00017013, -0.00016946, -0.00000195, -0.00006315, -0.00000258, -0.00003077, 0.00000323, -0.00001803, -0.00000325, -0.00001795, 0.00001040, -0.00001141, 0.00000040, -0.00001123, 0.00000099, -0.00000762, -0.00000298, -0.00001244, 0.00000880, -0.00000717, 0.00000223, -0.00000050, -0.00000138, -0.00000539, -0.00000217, -0.00000443, -0.00000507, -0.00000038, -0.00001690, 0.00000599, -0.00000362, 0.00001102, 0.00000788, -0.00000213, 0.00000963, -0.00000449, 0.00000897, -0.00001021, 0.00001175, -0.00000324, 0.00000808, 0.00000166, 0.00000254, -0.00000748, 0.00000484, 0.00000079, 0.00000660, 0.00000263, 0.00000595, 0.00000356, -0.00000159, 0.00000310, 0.00001027, 0.00000948, 0.00001320, -0.00001685, 0.00001659, -0.00002367, 0.00002221, -0.00002950, 0.00000764, 0.00000110, -0.00001039, 0.00000133, -0.00000827, -0.00000014, -0.00000518, 0.00000978, -0.00000976, -0.00000550, },
		{ 0.00000060, 0.00000294, -0.00000628, -0.00000457, -0.00000103, -0.00000079, -0.00000124, -0.00000463, 0.00000060, -0.00000385, 0.00000323, 0.00000084, -0.00000427, 0.00000190, -0.00000672, -0.00000652, 0.00000060, -0.00000283, 0.00000074, 0.00000110, -0.00000911, -0.00000212, -0.00000235, -0.00000325, -0.00000233, -0.00000802, 0.00001420, -0.00002555, 0.00001790, -0.00001582, 0.00001334, -0.00000602, -0.00000349, -0.00000346, 0.00000641, -0.00001307, 0.00000247, -0.00000431, 0.00000429, -0.00001144, -0.00000551, -0.00001797, 0.00001270, -0.00002846, -0.00001016, -0.00002186, -0.00000020, -0.00004693, -0.00003014, -0.00011916, -0.00020508, 0.00028793, 0.00012967, 0.00009211, 0.00007954, 0.00009254, 0.00015552, 0.00026070, -0.00026432, -0.00006474, -0.00004723, -0.00002648, -0.00002595, -0.00000490, -0.00001622, -0.00000150, -0.00002047, 0.00000318, -0.00000567, 0.00000033, -0.00001138, 0.00000070, -0.00000991, 0.00000225, -0.00001413, 0.00000123, -0.00000091, -0.00000110, -0.00000289, 0.00000522, -0.00000849, 0.00000045, -0.00000737, 0.00000009, -0.00000897, 0.00000147, -0.00002081, 0.00001439, -0.00001179, 0.00002337, -0.00000547, 0.00001040, -0.00000153, 0.00000603, -0.00000159, -0.00000047, 0.00000486, 0.00000269, 0.00000440, 0.00000487, -0.00000326, -0.00000215, 0.00000154, 0.00000644, 0.00000171, 0.00000764, 0.00000241, 0.00000464, -0.00000470, 0.00001279, 0.00000250, 0.00002336, -0.00000993, 0.00000457, -0.00000992, -0.00000006, 0.00000861, -0.00003096, 0.00002010, -0.00001357, 0.00000068, -0.00000704, -0.00000276, -0.00000286, -0.00000056, 0.00000577, -0.00000974, -0.00000707, },
		{ 0.00000162, 0.00000203, -0.00000794, -0.00000265, -0.00000229, 0.00000063, -0.00000336, -0.00000306, 0.00000052, -0.00000367, 0.00000444, -0.00000221, -0.00000222, 0.00000140, -0.00000893, -0.00000481, 0.00000110, -0.00000322, 0.00000194, -0.00000189, -0.00000848, 0.00000004, -0.00000525, -0.00000037, -0.00000675, 0.00000097, 0.00000413, -0.00002169, 0.00001987, -0.00001970, 0.00002039, -0.00001608, 0.00000480, -0.00000674, 0.00000788, -0.00002007, 0.00001297, -0.00001178, 0.00000988, -0.00002202, 0.00000293, -0.00002087, 0.00001676, -0.00004715, 0.00001273, -0.00004104, 0.00001619, -0.00008603, 0.00000813, -0.00026431, 0.00006885, 0.00024642, 0.00007780, 0.00008068, 0.00008727, 0.00007315, 0.00021498, 0.00014082, -0.00027304, -0.00000811, -0.00007252, -0.00000423, -0.00003546, 0.00000735, -0.00002515, 0.00000394, -0.00002412, 0.00001215, -0.00001374, 0.00000497, -0.00001511, 0.00000446, -0.00001327, 0.00000456, -0.00001660, 0.00000744, -0.00000617, 0.00000096, -0.00000259, 0.00000458, -0.00000964, 0.00000085, -0.00000680, -0.00000155, -0.00000671, -0.00000385, -0.00001540, 0.00001273, -0.00000980, 0.00002601, -0.00001050, 0.00001602, -0.00000670, 0.00001085, -0.00000691, 0.00000501, 0.00000136, 0.00000620, 0.00000277, 0.00000455, -0.00000447, -0.00000145, 0.00000426, 0.00000544, 0.00000269, 0.00000772, 0.00000271, -0.00000005, 0.00000371, 0.00000815, 0.00001192, 0.00001470, -0.00001138, 0.00000610, -0.00002234, 0.00002475, -0.00002170, -0.00000791, 0.00000945, -0.00001133, 0.00000094, -0.00000921, 0.00000006, -0.00000320, 0.00000062, 0.00000475, -0.00001342, -0.00000387, },
		{ 0.00000123, 0.00000113, -0.00000930, -0.00000082, -0.00000360, 0.00000246, -0.00000670, 0.00000104, -0.00000242, -0.00000078, 0.00000234, -0.00000229, -0.00000163, 0.00000157, -0.00001300, 0.00000058, -0.00000224, -0.00000003, -0.00000110, -0.00000099, -0.00001014, 0.00000401, -0.00001103, 0.00000626, -0.00001424, 0.00001383, -0.00001507, -0.00000388, 0.00000727, -0.00000920, 0.00001254, -0.00001211, 0.00000179, 0.00000040, -0.00000544, -0.00000802, 0.00000576, -0.00000328, -0.00000365, -0.00001133, -0.00000790, -0.00000250, -0.00001177, -0.00002382, -0.00000633, -0.00001682, -0.00002791, -0.00004290, -0.00008740, -0.00021599, 0.00027437, 0.00010266, 0.00010620, 0.00003812, 0.00011862, 0.00004430, 0.00027531, -0.00001158, -0.00020972, -0.00000896, -0.00006240, -0.00000771, -0.00002328, -0.00000120, -0.00001741, -0.00000492, -0.00001174, 0.00000420, -0.00000804, -0.00000174, -0.00000796, -0.00000255, -0.00000663, -0.00000229, -0.00000896, 0.00000332, -0.00000378, -0.00000256, 0.00000338, -0.00000263, -0.00000445, -0.00000403, -0.00000141, -0.00000809, 0.00000015, -0.00001406, -0.00000323, 0.00000286, 0.00000157, 0.00001886, -0.00000664, 0.00001408, -0.00000549, 0.00000991, -0.00000692, 0.00000590, 0.00000201, 0.00000661, 0.00000251, 0.00000317, -0.00000470, -0.00000079, 0.00000800, 0.00000187, 0.00000653, 0.00000571, 0.00000191, -0.00000077, 0.00001032, 0.00000341, 0.00002385, -0.00000512, 0.00000438, -0.00001759, 0.00000034, 0.00001534, -0.00002942, 0.00000688, -0.00000229, -0.00000466, -0.00000346, -0.00000786, 0.00000252, -0.00000535, 0.00000385, 0.00000093, -0.00001532, 0.00000040, },
		{ 0.00000356, -0.00000321, -0.00000691, -0.00000263, -0.00000088, -0.00000026, -0.00000545, 0.00000176, -0.00000296, -0.00000030, 0.00000205, -0.00000387, 0.00000178, -0.00000321, -0.00001175, 0.00000240, -0.00000276, -0.00000018, -0.00000105, -0.00000310, -0.00000732, 0.00000179, -0.00001008, 0.00000648, -0.00001390, 0.00001658, -0.00002643, 0.00001145, -0.00000539, 0.00000181, 0.00000350, -0.00000697, 0.00000035, 0.00000175, -0.00001338, 0.00000308, -0.00000101, 0.00000134, -0.00001346, -0.00000259, -0.00001297, 0.00000386, -0.00003079, -0.00000188, -0.00002426, -0.00000193, -0.00006339, -0.00000683, -0.00022089, 0.00002410, 0.00025457, 0.00004339, 0.00010450, 0.00003408, 0.00011426, 0.00005848, 0.00027894, -0.00012859, -0.00013573, -0.00002658, -0.00004415, -0.00001386, -0.00001155, -0.00001065, -0.00001046, -0.00001006, -0.00000244, -0.00000348, -0.00000277, -0.00000665, -0.00000308, -0.00000762, -0.00000180, -0.00000652, -0.00000331, -0.00000040, -0.00000259, -0.00000288, 0.00000548, -0.00000692, -0.00000167, -0.00000574, -0.00000002, -0.00001046, 0.00000245, -0.00001943, 0.00000531, -0.00000494, 0.00001307, 0.00000972, -0.00000043, 0.00001021, -0.00000320, 0.00000853, -0.00000699, 0.00000751, 0.00000215, 0.00000726, 0.00000080, 0.00000409, -0.00000794, 0.00000527, 0.00000559, 0.00000324, 0.00000674, 0.00000607, -0.00000312, 0.00000833, 0.00000431, 0.00001449, 0.00001503, -0.00000623, 0.00000271, -0.00002855, 0.00002613, -0.00001262, -0.00001330, 0.00000209, -0.00000146, -0.00000517, -0.00000461, -0.00000667, 0.00000526, -0.00000970, 0.00001067, -0.00000898, -0.00001013, 0.00000025, },
		{ 0.00000612, -0.00000826, -0.00000347, -0.00000516, 0.00000291, -0.00000510, -0.00000087, -0.00000067, -0.00000133, -0.00000181, 0.00000336, -0.00000612, 0.00000612, -0.00001131, -0.00000512, -0.00000043, 0.00000023, -0.00000421, 0.00000307, -0.00000900, -0.00000005, -0.00000668, -0.00000116, -0.00000152, -0.00000424, 0.00000592, -0.00002256, 0.00001483, -0.00000944, 0.00000551, 0.00000075, -0.00000686, 0.00000464, -0.00000755, -0.00000683, 0.00000185, 0.00000250, -0.00000759, -0.00000749, -0.00000776, -0.00000258, -0.00001502, -0.00001688, -0.00001157, -0.00001013, -0.00003425, -0.00003257, -0.00008138, -0.00019714, 0.00022860, 0.00011820, 0.00007400, 0.00005707, 0.00007347, 0.00007002, 0.00011886, 0.00021543, -0.00018334, -0.00008462, -0.00003596, -0.00003358, -0.00001156, -0.00001041, -0.00001204, -0.00001027, -0.00000628, -0.00000327, -0.00000326, -0.00000395, -0.00000509, -0.00000496, -0.00000633, -0.00000240, -0.00000492, -0.00000368, 0.00000017, -0.00000496, 0.00000155, 0.00000163, -0.00000584, -0.00000317, -0.00000308, -0.00000385, -0.00000707, -0.00000160, -0.00001771, 0.00000692, -0.00000631, 0.00002025, 0.00000212, 0.00000588, 0.00000536, 0.00000027, 0.00000602, -0.00000628, 0.00000935, 0.00000211, 0.00000679, 0.00000059, 0.00000315, -0.00000890, 0.00001124, 0.00000104, 0.00000697, 0.00000571, 0.00000426, -0.00000314, 0.00001471, -0.00000102, 0.00002806, -0.00000533, 0.00000910, -0.00002309, -0.00000505, 0.00001977, -0.00002169, -0.00000282, -0.00000358, 0.00000181, -0.00000800, -0.00000487, -0.00000277, 0.00000283, -0.00000885, 0.00001297, -0.00001759, -0.00000260, -0.00000175, },
		{ 0.00000431, -0.00000941, -0.00000335, -0.00000372, 0.00000223, -0.00000587, 0.00000114, -0.00000138, -0.00000212, -0.00000050, 0.00000159, -0.00000402, 0.00000473, -0.00001477, -0.00000049, -0.00000163, 0.00000024, -0.00000485, 0.00000392, -0.00001116, 0.00000298, -0.00001142, 0.00000576, -0.00000794, 0.00000342, -0.00000597, -0.00001317, 0.00001299, -0.00001112, 0.00000899, -0.00000333, -0.00000298, 0.00000223, -0.00001160, -0.00000088, 0.00000124, 0.00000164, -0.00001257, -0.00000326, -0.00000799, -0.00000286, -0.00002557, -0.00000374, -0.00002048, -0.00000457, -0.00005916, -0.00000585, -0.00019499, 0.00000192, 0.00023898, 0.00004134, 0.00009264, 0.00002973, 0.00009820, 0.00003747, 0.00017710, 0.00012550, -0.00019907, -0.00004696, -0.00004490, -0.00002066, -0.00001275, -0.00000863, -0.00001511, -0.00000585, -0.00000615, -0.00000299, -0.00000410, -0.00000321, -0.00000571, -0.00000520, -0.00000626, -0.00000082, -0.00000562, -0.00000271, -0.00000119, -0.00000428, 0.00000338, -0.00000120, -0.00000542, -0.00000310, -0.00000243, -0.00000631, -0.00000418, -0.00000610, -0.00001438, 0.00000611, -0.00000401, 0.00002425, -0.00000454, 0.00001316, -0.00000202, 0.00000741, -0.00000107, -0.00000059, 0.00000745, 0.00000467, 0.00000297, 0.00000462, -0.00000335, -0.00000196, 0.00000961, 0.00000194, 0.00000692, 0.00000741, -0.00000261, 0.00000772, 0.00000732, 0.00001063, 0.00002051, -0.00000800, 0.00000808, -0.00003738, 0.00002382, -0.00000633, -0.00000846, -0.00000952, 0.00000251, -0.00000280, -0.00000638, -0.00000697, 0.00000404, -0.00000548, 0.00000005, 0.00000530, -0.00001729, 0.00000012, -0.00000035, },
		{ -0.00000042, -0.00000795, -0.00000486, -0.00000023, -0.00000154, -0.00000277, -0.00000001, -0.00000023, -0.00000490, 0.00000332, -0.00000272, 0.00000144, -0.00000231, -0.00001168, -0.00000036, 0.00000037, -0.00000391, -0.00000024, -0.00000078, -0.00000744, -0.00000087, -0.00000832, 0.00000581, -0.00000840, 0.00000367, -0.00001127, -0.00000586, 0.00001081, -0.00001341, 0.00001487, -0.00001097, 0.00000598, -0.00000971, -0.00000324, -0.00000458, 0.00000844, -0.00001063, -0.00000410, -0.00001016, 0.00000382, -0.00002418, -0.00000783, -0.00001737, -0.00000252, -0.00004000, -0.00002580, -0.00007545, -0.00017642, 0.00018511, 0.00013189, 0.00004896, 0.00006935, 0.00004206, 0.00008727, 0.00003968, 0.00020651, 0.00003543, -0.00019141, -0.00001825, -0.00005480, -0.00000402, -0.00002034, -0.00000316, -0.00002076, 0.00000352, -0.00001261, 0.00000229, -0.00000881, 0.00000132, -0.00001045, -0.00000172, -0.00000875, 0.00000376, -0.00001029, 0.00000152, -0.00000551, 0.00000030, 0.00000098, -0.00000114, -0.00000682, -0.00000086, -0.00000471, -0.00000587, -0.00000351, -0.00000945, -0.00001112, 0.00000453, 0.00000151, 0.00002348, -0.00000714, 0.00001735, -0.00000742, 0.00001376, -0.00000890, 0.00000751, 0.00000351, 0.00000732, -0.00000012, 0.00000763, -0.00000924, 0.00000721, 0.00000401, 0.00000613, 0.00000533, 0.00000714, -0.00000515, 0.00001718, -0.00000129, 0.00002746, -0.00000156, 0.00000767, -0.00001835, -0.00001671, 0.00002227, -0.00001589, -0.00000065, -0.00001561, 0.00001013, -0.00000987, -0.00000360, -0.00000672, 0.00000686, -0.00001117, 0.00000862, -0.00000592, -0.00001107, -0.00000101, 0.00000319, },
		{ -0.00000368, -0.00000777, -0.00000383, 0.00000024, -0.00000279, -0.00000097, -0.00000015, -0.00000140, -0.00000477, 0.00000459, -0.00000421, 0.00000370, -0.00000792, -0.00000791, -0.00000003, 0.00000022, -0.00000581, 0.00000314, -0.00000484, -0.00000444, -0.00000449, -0.00000377, 0.00000441, -0.00000912, 0.00000329, -0.00001489, 0.00000295, 0.00000314, -0.00000857, 0.00001393, -0.00001178, 0.00000680, -0.00001536, 0.00000308, -0.00000560, 0.00000934, -0.00001758, 0.00000139, -0.00001067, 0.00000322, -0.00003516, 0.00000716, -0.00002847, 0.00000572, -0.00006764, 0.00000470, -0.00018286, -0.00000030, 0.00020435, 0.00006129, 0.00006023, 0.00005153, 0.00005475, 0.00007046, 0.00005665, 0.00021613, -0.00005715, -0.00015347, -0.00001260, -0.00005001, 0.00000092, -0.00002097, -0.00000517, -0.00001684, 0.00000436, -0.00001364, 0.00000294, -0.00000853, 0.00000075, -0.00001077, -0.00000219, -0.00000609, 0.00000292, -0.00001107, 0.00000233, -0.00000616, 0.00000202, 0.00000028, -0.00000315, -0.00000512, -0.00000179, -0.00000471, -0.00000694, -0.00000143, -0.00001496, -0.00000507, 0.00000034, 0.00001178, 0.00001534, -0.00000170, 0.00001334, -0.00000467, 0.00001221, -0.00000996, 0.00001140, 0.00000232, 0.00000620, 0.00000188, 0.00000406, -0.00000697, 0.00000962, 0.00000279, 0.00000709, 0.00000670, 0.00000103, 0.00000395, 0.00001261, 0.00000670, 0.00002479, -0.00000921, 0.00001076, -0.00003741, 0.00001357, -0.00000099, -0.00000320, -0.00001081, -0.00000751, 0.00000761, -0.00001129, -0.00000380, -0.00000200, 0.00000244, -0.00000860, 0.00000896, -0.00001170, -0.00000631, -0.00000091, 0.00000436, },
		{ -0.00000565, -0.00000797, -0.00000148, -0.00000166, -0.00000109, -0.00000144, 0.00000107, -0.00000483, -0.00000124, 0.00000256, -0.00000225, 0.00000128, -0.00000915, -0.00000627, 0.00000202, -0.00000344, -0.00000297, 0.00000215, -0.00000540, -0.00000504, -0.00000415, -0.00000144, 0.00000372, -0.00001213, 0.00000559, -0.00001936, 0.00001361, -0.00000999, 0.00000479, 0.00000365, -0.00000332, -0.00000453, -0.00000685, -0.00000194, 0.00000341, -0.00000355, -0.00000924, -0.00000568, 0.00000248, -0.00002052, -0.00001549, -0.00000680, -0.00001082, -0.00002806, -0.00003768, -0.00005480, -0.00017277, 0.00016749, 0.00011702, 0.00005817, 0.00004564, 0.00005618, 0.00005164, 0.00006168, 0.00008033, 0.00020304, -0.00014045, -0.00009407, -0.00002892, -0.00002882, -0.00000909, -0.00001173, -0.00001520, -0.00000291, -0.00000612, -0.00000537, -0.00000383, -0.00000149, -0.00000667, -0.00000491, -0.00000752, 0.00000217, -0.00000490, -0.00000579, -0.00000173, -0.00000204, -0.00000052, 0.00000237, -0.00000737, -0.00000069, -0.00000580, -0.00000223, -0.00000927, 0.00000099, -0.00002115, 0.00000216, -0.00000369, 0.00002264, 0.00000416, 0.00000792, 0.00000446, 0.00000366, 0.00000413, -0.00000404, 0.00001027, 0.00000331, 0.00000381, 0.00000521, -0.00000172, -0.00000003, 0.00000669, 0.00000551, 0.00000592, 0.00000721, -0.00000273, 0.00001427, 0.00000400, 0.00002288, 0.00000494, 0.00000250, -0.00001105, -0.00002369, 0.00001747, -0.00001112, 0.00000553, -0.00002204, 0.00000550, -0.00000106, -0.00000884, -0.00000403, 0.00000177, -0.00000293, -0.00000309, 0.00000406, -0.00001200, -0.00000410, 0.00000062, 0.00000327, },
		{ -0.00000874, -0.00000549, -0.00000190, -0.00000134, -0.00000064, -0.00000081, 0.00000001, -0.00000605, 0.00000138, 0.00000130, -0.00000157, -0.00000092, -0.00000983, -0.00000350, 0.00000147, -0.00000523, -0.00000042, 0.00000098, -0.00000681, -0.00000430, -0.00000434, 0.00000267, -0.00000191, -0.00000998, 0.00000352, -0.00001756, 0.00001678, -0.00001711, 0.00001500, -0.00000463, 0.00000141, -0.00001306, 0.00000238, -0.00000678, 0.00000914, -0.00001467, -0.00000056, -0.00000927, 0.00000629, -0.00003795, 0.00000662, -0.00002410, 0.00000377, -0.00006179, -0.00000136, -0.00015864, -0.00001288, 0.00018771, 0.00005918, 0.00005670, 0.00004175, 0.00005295, 0.00005482, 0.00004602, 0.00012363, 0.00014627, -0.00018227, -0.00004521, -0.00004324, -0.00001062, -0.00001818, -0.00000598, -0.00001897, 0.00000565, -0.00001434, 0.00000129, -0.00000818, 0.00000247, -0.00001149, -0.00000147, -0.00000896, 0.00000651, -0.00001101, -0.00000152, -0.00000403, 0.00000024, -0.00000132, 0.00000174, -0.00000800, 0.00000023, -0.00000683, -0.00000270, -0.00000772, -0.00000202, -0.00002191, 0.00000494, -0.00000157, 0.00002674, -0.00000206, 0.00001361, -0.00000104, 0.00000924, -0.00000262, 0.00000271, 0.00000811, 0.00000319, 0.00000438, 0.00000423, -0.00000226, 0.00000317, 0.00000593, 0.00000640, 0.00000683, 0.00000268, 0.00000375, 0.00001280, 0.00000914, 0.00002331, -0.00000551, 0.00000665, -0.00002986, 0.00000257, -0.00000064, 0.00000040, -0.00000423, -0.00001868, 0.00001002, -0.00000701, -0.00000692, -0.00000179, 0.00000093, -0.00000379, -0.00000137, 0.00000075, -0.00001173, -0.00000144, 0.00000070, 0.00000269, },
		{ -0.00001129, -0.00000209, -0.00000411, 0.00000130, -0.00000211, 0.00000109, -0.00000339, -0.00000416, 0.00000192, 0.00000146, -0.00000317, -0.00000103, -0.00001145, 0.00000095, -0.00000260, -0.00000295, -0.00000075, 0.00000142, -0.00001036, -0.00000039, -0.00000661, 0.00000861, -0.00001285, -0.00000012, -0.00000616, -0.00000661, 0.00000845, -0.00001202, 0.00001495, -0.00000477, -0.00000440, -0.00000871, 0.00000109, -0.00000201, 0.00000182, -0.00001191, -0.00000301, 0.00000003, -0.00001268, -0.00002572, 0.00000212, -0.00001603, -0.00001902, -0.00004481, -0.00004270, -0.00016423, 0.00015078, 0.00010478, 0.00006245, 0.00003224, 0.00005928, 0.00003351, 0.00006826, 0.00002720, 0.00017639, 0.00005052, -0.00017196, -0.00002464, -0.00004147, -0.00000678, -0.00001750, -0.00000924, -0.00001131, 0.00000165, -0.00001247, 0.00000011, -0.00000529, -0.00000113, -0.00000935, -0.00000385, -0.00000351, 0.00000274, -0.00001077, -0.00000167, -0.00000205, -0.00000168, 0.00000149, -0.00000270, -0.00000421, -0.00000345, -0.00000367, -0.00000677, -0.00000229, -0.00001071, -0.00001633, 0.00000285, 0.00000676, 0.00002317, -0.00000169, 0.00001363, -0.00000119, 0.00000943, -0.00000460, 0.00000726, 0.00000579, 0.00000324, 0.00000553, 0.00000160, 0.00000075, 0.00000261, 0.00000791, 0.00000574, 0.00000698, -0.00000020, 0.00001226, 0.00000686, 0.00002268, 0.00000681, 0.00000143, -0.00001054, -0.00002050, 0.00000733, -0.00000881, 0.00000998, -0.00001809, -0.00000693, 0.00000673, -0.00000949, -0.00000429, -0.00000087, -0.00000011, -0.00000371, -0.00000149, -0.00000095, -0.00001122, 0.00000107, 0.00000013, 0.00000170, },
		{ -0.00000990, -0.00000202, -0.00000353, 0.00000211, -0.00000217, 0.00000078, -0.00000510, -0.00000261, 0.00000335, -0.00000066, -0.00000279, -0.00000275, -0.00001037, 0.00000244, -0.00000494, -0.00000113, -0.00000024, -0.00000083, -0.00001057, 0.00000118, -0.00000538, 0.00000856, -0.00001926, 0.00000831, -0.00001514, 0.00000455, -0.00000227, -0.00000254, 0.00001138, -0.00000529, -0.00000980, -0.00000125, -0.00000278, 0.00000282, -0.00000721, -0.00000582, -0.00000478, 0.00000371, -0.00003053, -0.00000511, -0.00001140, -0.00000525, -0.00004803, -0.00001465, -0.00013196, -0.00002907, 0.00018240, 0.00004490, 0.00006653, 0.00002079, 0.00006549, 0.00002520, 0.00006752, 0.00003366, 0.00020040, -0.00004505, -0.00013398, -0.00002103, -0.00003282, -0.00000885, -0.00001490, -0.00001162, -0.00000351, -0.00000534, -0.00000763, -0.00000246, -0.00000203, -0.00000551, -0.00000646, -0.00000548, 0.00000152, -0.00000288, -0.00000877, -0.00000173, -0.00000089, -0.00000273, 0.00000289, -0.00000558, -0.00000175, -0.00000609, -0.00000162, -0.00000900, 0.00000077, -0.00001879, -0.00000969, 0.00000112, 0.00001558, 0.00001696, 0.00000099, 0.00001214, -0.00000021, 0.00000801, -0.00000412, 0.00001046, 0.00000224, 0.00000608, 0.00000307, 0.00000320, 0.00000056, 0.00000427, 0.00000801, 0.00000695, 0.00000268, 0.00000590, 0.00001078, 0.00001282, 0.00002266, -0.00000335, 0.00000340, -0.00002591, 0.00000093, -0.00000790, 0.00000299, 0.00000210, -0.00001979, 0.00000046, 0.00000260, -0.00001080, -0.00000004, -0.00000347, 0.00000233, -0.00000733, 0.00000150, -0.00000481, -0.00000784, 0.00000065, 0.00000157, -0.00000225, },
		{ -0.00000614, -0.00000455, 0.00000010, 0.00000024, -0.00000039, -0.00000206, -0.00000414, -0.00000222, 0.00000557, -0.00000516, 0.00000067, -0.00000731, -0.00000574, -0.00000048, -0.00000309, -0.00000223, 0.00000254, -0.00000691, -0.00000529, -0.00000204, 0.00000105, -0.00000051, -0.00001491, 0.00000780, -0.00001625, 0.00000841, -0.00000716, 0.00000393, 0.00000955, -0.00001176, -0.00000577, -0.00000109, -0.00000039, -0.00000146, -0.00000679, -0.00000673, 0.00000273, -0.00001140, -0.00002437, -0.00000372, -0.00000760, -0.00002289, -0.00003734, -0.00004654, -0.00014394, 0.00012405, 0.00010958, 0.00004632, 0.00004340, 0.00003700, 0.00004735, 0.00003804, 0.00004669, 0.00007277, 0.00017539, -0.00010877, -0.00009410, -0.00002070, -0.00002683, -0.00000928, -0.00001544, -0.00000776, -0.00000315, -0.00000732, -0.00000600, -0.00000167, -0.00000297, -0.00000611, -0.00000653, -0.00000285, 0.00000146, -0.00000562, -0.00000814, 0.00000049, -0.00000287, -0.00000057, 0.00000080, -0.00000477, -0.00000300, -0.00000511, -0.00000298, -0.00000726, -0.00000152, -0.00002254, -0.00000498, 0.00000237, 0.00002183, 0.00001067, 0.00000449, 0.00001020, 0.00000058, 0.00000665, -0.00000212, 0.00001122, 0.00000016, 0.00000852, 0.00000012, 0.00000692, -0.00000233, 0.00000837, 0.00000665, 0.00000732, 0.00000004, 0.00001379, 0.00000549, 0.00002665, 0.00000619, 0.00000200, -0.00001297, -0.00001606, 0.00000304, -0.00001403, 0.00001371, -0.00001147, -0.00001272, 0.00000221, -0.00000122, -0.00000915, 0.00000140, -0.00000453, 0.00000361, -0.00001111, 0.00000583, -0.00000976, -0.00000293, -0.00000145, 0.00000369, -0.00000734, },
		{ -0.00000460, -0.00000496, 0.00000253, -0.00000100, -0.00000011, -0.00000374, -0.00000360, -0.00000007, 0.00000447, -0.00000706, 0.00000265, -0.00001053, -0.00000215, -0.00000306, -0.00000109, -0.00000266, 0.00000302, -0.00001097, -0.00000019, -0.00000462, 0.00000558, -0.00001059, -0.00000575, 0.00000211, -0.00001284, 0.00000817, -0.00000869, 0.00001007, 0.00000360, -0.00001531, 0.00000011, -0.00000277, 0.00000152, -0.00000665, -0.00000433, -0.00000582, 0.00000446, -0.00002603, -0.00000976, -0.00001013, -0.00000258, -0.00004551, -0.00001481, -0.00012162, -0.00003114, 0.00016657, 0.00004593, 0.00005746, 0.00002399, 0.00005193, 0.00002924, 0.00005001, 0.00003053, 0.00011722, 0.00011823, -0.00013948, -0.00005828, -0.00002416, -0.00002028, -0.00001194, -0.00001287, -0.00000516, -0.00000442, -0.00000796, -0.00000411, -0.00000175, -0.00000384, -0.00000676, -0.00000531, -0.00000086, 0.00000009, -0.00000796, -0.00000588, 0.00000100, -0.00000396, 0.00000103, -0.00000099, -0.00000419, -0.00000420, -0.00000391, -0.00000439, -0.00000524, -0.00000620, -0.00002347, -0.00000078, 0.00000511, 0.00002615, 0.00000372, 0.00001036, 0.00000536, 0.00000380, 0.00000364, 0.00000239, 0.00000786, 0.00000275, 0.00000604, 0.00000267, 0.00000591, -0.00000175, 0.00001015, 0.00000726, 0.00000318, 0.00000627, 0.00001189, 0.00001211, 0.00002732, -0.00000549, 0.00000363, -0.00002751, 0.00000509, -0.00001398, -0.00000038, 0.00000749, -0.00001493, -0.00000798, 0.00000323, -0.00000580, -0.00000382, -0.00000235, -0.00000013, -0.00000156, -0.00000878, 0.00000591, -0.00001142, -0.00000054, -0.00000160, 0.00000296, -0.00000946, },
		{ -0.00000559, -0.00000215, 0.00000214, -0.00000054, -0.00000209, -0.00000304, -0.00000433, 0.00000387, -0.00000065, -0.00000415, 0.00000052, -0.00001006, -0.00000209, -0.00000226, -0.00000182, -0.00000037, -0.00000103, -0.00000935, 0.00000060, -0.00000302, 0.00000330, -0.00001412, 0.00000058, -0.00000299, -0.00000948, 0.00000790, -0.00000969, 0.00001712, -0.00000920, -0.00000875, -0.00000099, 0.00000081, -0.00000424, -0.00000390, -0.00000777, 0.00000352, -0.00001094, -0.00001979, -0.00001058, -0.00000250, -0.00002244, -0.00003351, -0.00004576, -0.00013193, 0.00010546, 0.00011055, 0.00003625, 0.00004639, 0.00002733, 0.00004774, 0.00002668, 0.00004809, 0.00003488, 0.00014269, 0.00005226, -0.00014529, -0.00002839, -0.00003156, -0.00001155, -0.00001767, -0.00000509, -0.00000831, -0.00000246, -0.00001047, 0.00000023, -0.00000507, -0.00000209, -0.00000933, -0.00000096, -0.00000245, 0.00000018, -0.00001116, -0.00000108, -0.00000182, -0.00000218, 0.00000025, -0.00000085, -0.00000555, -0.00000347, -0.00000420, -0.00000432, -0.00000504, -0.00001116, -0.00002231, 0.00000296, 0.00000949, 0.00002701, -0.00000130, 0.00001617, -0.00000087, 0.00000882, -0.00000020, 0.00000720, 0.00000348, 0.00000682, 0.00000170, 0.00000865, 0.00000042, 0.00000303, 0.00000946, 0.00000761, 0.00000002, 0.00001498, 0.00000546, 0.00002787, 0.00001007, -0.00000174, -0.00001191, -0.00001777, 0.00000752, -0.00002253, 0.00001359, -0.00000570, -0.00001021, -0.00000704, 0.00000364, -0.00000817, 0.00000008, -0.00000587, 0.00000483, -0.00000966, -0.00000138, 0.00000130, -0.00000934, -0.00000095, 0.00000019, -0.00000021, -0.00000875, },
		{ -0.00000531, 0.00000040, 0.00000143, -0.00000094, -0.00000324, -0.00000271, -0.00000345, 0.00000543, -0.00000500, -0.00000026, -0.00000290, -0.00000879, -0.00000237, -0.00000100, -0.00000257, 0.00000106, -0.00000518, -0.00000575, 0.00000008, -0.00000139, -0.00000138, -0.00001312, 0.00000431, -0.00000837, -0.00000395, 0.00000449, -0.00000559, 0.00001703, -0.00001857, -0.00000002, -0.00000382, 0.00000339, -0.00001000, -0.00000025, -0.00000857, 0.00000694, -0.00002582, -0.00000614, -0.00001730, 0.00000511, -0.00004673, -0.00001026, -0.00011642, -0.00003061, 0.00014901, 0.00005326, 0.00004200, 0.00003374, 0.00003419, 0.00003918, 0.00002777, 0.00004478, 0.00004880, 0.00014933, -0.00001436, -0.00012705, -0.00001330, -0.00003398, -0.00000797, -0.00001812, -0.00000085, -0.00001052, -0.00000186, -0.00001036, 0.00000188, -0.00000668, -0.00000209, -0.00000908, 0.00000135, -0.00000354, -0.00000136, -0.00001139, 0.00000171, -0.00000391, -0.00000086, 0.00000020, -0.00000198, -0.00000571, -0.00000353, -0.00000343, -0.00000543, -0.00000472, -0.00001743, -0.00001748, 0.00000437, 0.00001723, 0.00002149, 0.00000024, 0.00001634, -0.00000301, 0.00001132, -0.00000150, 0.00000868, 0.00000272, 0.00000685, 0.00000230, 0.00001043, -0.00000279, 0.00000713, 0.00000967, 0.00000389, 0.00000531, 0.00001449, 0.00001027, 0.00003211, -0.00000564, 0.00000124, -0.00002735, 0.00000476, -0.00001061, -0.00001033, 0.00001109, -0.00001013, -0.00000766, -0.00000526, 0.00000158, -0.00000620, -0.00000127, -0.00000385, 0.00000353, -0.00001293, 0.00000461, -0.00000365, -0.00000682, -0.00000129, 0.00000108, -0.00000262, -0.00000874, },
		{ -0.00000263, 0.00000093, 0.00000153, -0.00000295, -0.00000234, -0.00000379, -0.00000031, 0.00000295, -0.00000541, 0.00000130, -0.00000520, -0.00000872, -0.00000059, -0.00000175, -0.00000132, -0.00000075, -0.00000582, -0.00000378, 0.00000084, -0.00000309, -0.00000342, -0.00001172, 0.00000718, -0.00001475, 0.00000487, -0.00000339, 0.00000502, 0.00000564, -0.00001576, 0.00000120, -0.00000107, -0.00000218, -0.00000714, -0.00000349, 0.00000005, -0.00000710, -0.00002021, -0.00000783, -0.00000933, -0.00001208, -0.00003827, -0.00003556, -0.00013057, 0.00009649, 0.00010288, 0.00003948, 0.00003415, 0.00003528, 0.00003070, 0.00003708, 0.00002496, 0.00004812, 0.00006175, 0.00014126, -0.00007476, -0.00009053, -0.00001539, -0.00002740, -0.00001199, -0.00001020, -0.00000505, -0.00000684, -0.00000594, -0.00000516, -0.00000182, -0.00000387, -0.00000586, -0.00000406, -0.00000129, -0.00000148, -0.00000596, -0.00000721, 0.00000027, -0.00000302, -0.00000162, 0.00000191, -0.00000536, -0.00000351, -0.00000525, -0.00000116, -0.00000835, -0.00000404, -0.00002404, -0.00001017, 0.00000487, 0.00002582, 0.00001196, 0.00000737, 0.00001031, 0.00000064, 0.00000974, -0.00000006, 0.00000765, 0.00000436, 0.00000420, 0.00000746, 0.00000645, -0.00000123, 0.00000909, 0.00000947, 0.00000056, 0.00001410, 0.00000830, 0.00002584, 0.00001708, -0.00000678, -0.00001066, -0.00002073, 0.00001115, -0.00002350, 0.00000549, -0.00000030, -0.00000612, -0.00000935, -0.00000227, -0.00000068, -0.00000369, -0.00000403, 0.00000061, -0.00000274, -0.00000980, 0.00000620, -0.00000673, -0.00000511, -0.00000096, 0.00000083, -0.00000408, -0.00000976, },
		{ -0.00000044, 0.00000130, 0.00000046, -0.00000394, -0.00000175, -0.00000388, 0.00000175, -0.00000016, -0.00000395, 0.00000159, -0.00000788, -0.00000742, 0.00000078, -0.00000236, -0.00000080, -0.00000260, -0.00000530, -0.00000163, -0.00000030, -0.00000437, -0.00000495, -0.00000835, 0.00000584, -0.00001674, 0.00001066, -0.00000866, 0.00001303, -0.00000740, -0.00000712, -0.00000144, 0.00000230, -0.00000934, -0.00000167, -0.00000610, 0.00000524, -0.00002314, -0.00000513, -0.00001771, 0.00000219, -0.00003784, -0.00001477, -0.00010171, -0.00004058, 0.00014155, 0.00005172, 0.00003961, 0.00002767, 0.00003603, 0.00002848, 0.00003167, 0.00002921, 0.00004731, 0.00007970, 0.00011181, -0.00011320, -0.00005364, -0.00002344, -0.00001960, -0.00001503, -0.00000283, -0.00001065, -0.00000243, -0.00000962, -0.00000040, -0.00000569, -0.00000126, -0.00000840, 0.00000038, -0.00000454, 0.00000020, -0.00000949, -0.00000297, -0.00000181, -0.00000210, -0.00000142, 0.00000186, -0.00000729, -0.00000227, -0.00000540, -0.00000099, -0.00000975, -0.00000608, -0.00002716, -0.00000440, 0.00000871, 0.00002950, 0.00000521, 0.00001344, 0.00000371, 0.00000623, 0.00000659, 0.00000159, 0.00000752, 0.00000398, 0.00000480, 0.00001033, 0.00000233, 0.00000168, 0.00001109, 0.00000564, 0.00000473, 0.00001525, 0.00001138, 0.00003297, -0.00000092, -0.00000509, -0.00002432, -0.00000006, -0.00000339, -0.00001687, 0.00000742, -0.00000517, -0.00000349, -0.00001067, -0.00000029, -0.00000077, -0.00000439, -0.00000316, 0.00000124, -0.00000737, -0.00000475, 0.00000524, -0.00000892, -0.00000300, -0.00000147, 0.00000137, -0.00000703, -0.00000868, },
		{ -0.00000024, 0.00000229, -0.00000223, -0.00000280, -0.00000258, -0.00000214, 0.00000104, -0.00000118, -0.00000292, 0.00000162, -0.00001158, -0.00000338, -0.00000032, -0.00000115, -0.00000282, -0.00000182, -0.00000595, 0.00000193, -0.00000531, -0.00000163, -0.00000902, -0.00000138, -0.00000168, -0.00001071, 0.00000869, -0.00000618, 0.00001107, -0.00001147, -0.00000290, -0.00000045, -0.00000110, -0.00000917, -0.00000182, -0.00000066, -0.00000528, -0.00002089, -0.00000325, -0.00001528, -0.00000702, -0.00003693, -0.00003100, -0.00012309, 0.00008093, 0.00010106, 0.00003913, 0.00002827, 0.00003466, 0.00002592, 0.00003448, 0.00001927, 0.00004420, 0.00003682, 0.00010542, 0.00005948, -0.00012064, -0.00003049, -0.00002728, -0.00001664, -0.00001132, -0.00000318, -0.00001062, -0.00000263, -0.00000842, -0.00000053, -0.00000547, -0.00000235, -0.00000631, 0.00000003, -0.00000449, -0.00000137, -0.00000912, -0.00000186, -0.00000183, -0.00000297, 0.00000107, -0.00000151, -0.00000582, -0.00000365, -0.00000274, -0.00000446, -0.00000814, -0.00001177, -0.00002562, -0.00000105, 0.00001631, 0.00002673, 0.00000427, 0.00001450, 0.00000093, 0.00001027, 0.00000340, 0.00000339, 0.00000767, 0.00000236, 0.00000903, 0.00000841, 0.00000133, 0.00000425, 0.00001192, 0.00000196, 0.00001275, 0.00001062, 0.00002533, 0.00002148, -0.00000751, -0.00001390, -0.00001987, 0.00000854, -0.00001656, -0.00000420, 0.00000006, -0.00000190, -0.00000560, -0.00000995, 0.00000150, -0.00000115, -0.00000565, -0.00000057, -0.00000188, -0.00000792, -0.00000077, 0.00000294, -0.00001009, -0.00000098, -0.00000222, 0.00000197, -0.00001098, -0.00000483, },
		{ -0.00000017, 0.00000196, -0.00000394, -0.00000174, -0.00000281, -0.00000122, 0.00000017, -0.00000161, -0.00000134, -0.00000063, -0.00001307, 0.00000042, -0.00000166, -0.00000069, -0.00000432, -0.00000091, -0.00000540, 0.00000286, -0.00000937, 0.00000160, -0.00001251, 0.00000472, -0.00000987, -0.00000171, 0.00000338, -0.00000128, 0.00000449, -0.00000907, -0.00000235, 0.00000150, -0.00000704, -0.00000525, -0.00000260, 0.00000296, -0.00001960, -0.00000736, -0.00001175, -0.00000650, -0.00002705, -0.00001976, -0.00008654, -0.00004993, 0.00013291, 0.00005030, 0.00004067, 0.00001883, 0.00004005, 0.00001796, 0.00003633, 0.00001447, 0.00005396, 0.00003058, 0.00012448, -0.00000019, -0.00010579, -0.00002010, -0.00002790, -0.00001441, -0.00000628, -0.00000716, -0.00000750, -0.00000486, -0.00000543, -0.00000266, -0.00000377, -0.00000404, -0.00000311, -0.00000244, -0.00000310, -0.00000374, -0.00000742, -0.00000180, -0.00000174, -0.00000313, 0.00000264, -0.00000487, -0.00000381, -0.00000486, -0.00000057, -0.00000847, -0.00000626, -0.00001817, -0.00002216, 0.00000300, 0.00002342, 0.00002157, 0.00000604, 0.00001253, 0.00000153, 0.00001221, 0.00000009, 0.00000693, 0.00000502, 0.00000438, 0.00001123, 0.00000582, 0.00000152, 0.00000769, 0.00000880, 0.00000503, 0.00001452, 0.00001369, 0.00003319, 0.00000378, -0.00000964, -0.00002401, -0.00000254, -0.00000148, -0.00001380, -0.00000200, -0.00000264, 0.00000090, -0.00000886, -0.00000828, 0.00000424, -0.00000436, -0.00000356, -0.00000127, -0.00000375, -0.00000716, 0.00000269, -0.00000127, -0.00000868, -0.00000101, -0.00000094, -0.00000032, -0.00001222, -0.00000137, },
		{ 0.00000082, -0.00000036, -0.00000335, -0.00000202, -0.00000165, -0.00000242, 0.00000115, -0.00000299, 0.00000134, -0.00000590, -0.00001018, 0.00000123, -0.00000119, -0.00000286, -0.00000269, -0.00000230, -0.00000195, -0.00000167, -0.00000757, 0.00000040, -0.00001130, 0.00000521, -0.00001269, 0.00000398, 0.00000024, -0.00000035, 0.00000180, -0.00000806, 0.00000013, -0.00000158, -0.00000788, -0.00000449, 0.00000224, -0.00000636, -0.00001902, -0.00000311, -0.00001270, -0.00001213, -0.00002789, -0.00003301, -0.00010874, 0.00006006, 0.00010319, 0.00003427, 0.00003114, 0.00002410, 0.00003236, 0.00002135, 0.00002707, 0.00002540, 0.00004661, 0.00004186, 0.00012161, -0.00004890, -0.00008277, -0.00001657, -0.00002801, -0.00000903, -0.00000597, -0.00000850, -0.00000604, -0.00000519, -0.00000451, -0.00000312, -0.00000362, -0.00000328, -0.00000263, -0.00000345, -0.00000296, -0.00000450, -0.00000692, -0.00000071, -0.00000313, -0.00000101, 0.00000111, -0.00000557, -0.00000356, -0.00000388, -0.00000143, -0.00001029, -0.00000642, -0.00002314, -0.00001825, 0.00000916, 0.00002746, 0.00001706, 0.00000758, 0.00001031, 0.00000399, 0.00001155, -0.00000152, 0.00001014, 0.00000124, 0.00001003, 0.00000905, 0.00000543, 0.00000189, 0.00001047, 0.00000510, 0.00001208, 0.00001099, 0.00002714, 0.00002358, -0.00000591, -0.00001829, -0.00001889, 0.00000517, -0.00001155, -0.00000552, -0.00000748, 0.00000190, -0.00000087, -0.00001103, -0.00000493, 0.00000514, -0.00000743, -0.00000030, -0.00000473, -0.00000280, -0.00000667, 0.00000534, -0.00000628, -0.00000546, -0.00000252, 0.00000170, -0.00000535, -0.00000944, -0.00000029, },
		{ 0.00000106, -0.00000233, -0.00000231, -0.00000202, -0.00000141, -0.00000321, 0.00000216, -0.00000371, 0.00000246, -0.00001077, -0.00000549, 0.00000052, -0.00000085, -0.00000500, -0.00000024, -0.00000371, 0.00000069, -0.00000755, -0.00000233, -0.00000395, -0.00000709, 0.00000198, -0.00001114, 0.00000687, -0.00000273, -0.00000038, 0.00000169, -0.00000818, 0.00000254, -0.00000665, -0.00000578, -0.00000337, 0.00000456, -0.00001911, -0.00000770, -0.00000843, -0.00000850, -0.00002627, -0.00001570, -0.00007992, -0.00004925, 0.00011530, 0.00005575, 0.00003557, 0.00001999, 0.00003201, 0.00002206, 0.00002601, 0.00001872, 0.00003923, 0.00003158, 0.00006522, 0.00009715, -0.00008037, -0.00005836, -0.00001878, -0.00002442, -0.00000471, -0.00000786, -0.00000800, -0.00000537, -0.00000505, -0.00000420, -0.00000325, -0.00000315, -0.00000235, -0.00000337, -0.00000384, -0.00000277, -0.00000525, -0.00000610, -0.00000021, -0.00000387, 0.00000092, -0.00000134, -0.00000510, -0.00000366, -0.00000258, -0.00000389, -0.00001062, -0.00000803, -0.00002727, -0.00001249, 0.00001542, 0.00002965, 0.00001243, 0.00000949, 0.00000808, 0.00000773, 0.00000766, 0.00000100, 0.00000909, 0.00000178, 0.00001350, 0.00000634, 0.00000557, 0.00000377, 0.00000958, 0.00000712, 0.00001390, 0.00001441, 0.00003581, 0.00000574, -0.00001082, -0.00002670, -0.00000265, -0.00000337, -0.00000820, -0.00000655, -0.00000804, 0.00000602, -0.00000468, -0.00001207, 0.00000059, 0.00000171, -0.00000628, -0.00000092, -0.00000592, -0.00000215, -0.00000497, 0.00000519, -0.00000924, -0.00000317, -0.00000276, 0.00000233, -0.00000927, -0.00000551, -0.00000019, },
		{ -0.00000034, -0.00000237, -0.00000227, -0.00000087, -0.00000342, -0.00000140, 0.00000117, -0.00000238, 0.00000011, -0.00001199, -0.00000226, 0.00000042, -0.00000257, -0.00000426, 0.00000014, -0.00000275, -0.00000114, -0.00000915, 0.00000100, -0.00000737, -0.00000368, -0.00000096, -0.00000851, 0.00000895, -0.00000841, 0.00000411, -0.00000164, -0.00000500, -0.00000016, -0.00000735, -0.00000605, 0.00000285, -0.00000527, -0.00001851, -0.00000349, -0.00000878, -0.00001564, -0.00002445, -0.00002942, -0.00009960, 0.00004700, 0.00009716, 0.00003547, 0.00003002, 0.00002012, 0.00003041, 0.00001935, 0.00002343, 0.00002148, 0.00004175, 0.00002431, 0.00008682, 0.00006023, -0.00009533, -0.00003630, -0.00002495, -0.00001554, -0.00000573, -0.00000750, -0.00000856, -0.00000343, -0.00000650, -0.00000248, -0.00000464, -0.00000066, -0.00000378, -0.00000300, -0.00000500, -0.00000133, -0.00000744, -0.00000358, -0.00000162, -0.00000246, 0.00000059, -0.00000244, -0.00000514, -0.00000274, -0.00000280, -0.00000613, -0.00001049, -0.00001077, -0.00003027, -0.00000467, 0.00002082, 0.00003039, 0.00000750, 0.00001231, 0.00000609, 0.00001099, 0.00000292, 0.00000571, 0.00000490, 0.00000739, 0.00001180, 0.00000628, 0.00000505, 0.00000613, 0.00000750, 0.00001306, 0.00001069, 0.00002835, 0.00002731, -0.00000645, -0.00001964, -0.00002121, 0.00000486, -0.00001200, -0.00000035, -0.00001474, -0.00000060, 0.00000503, -0.00000817, -0.00001068, 0.00000495, -0.00000306, -0.00000281, -0.00000513, -0.00000314, -0.00000364, -0.00000161, 0.00000200, -0.00000928, -0.00000261, -0.00000126, -0.00000006, -0.00000976, -0.00000298, 0.00000024, },
		{ -0.00000162, -0.00000182, -0.00000205, -0.00000057, -0.00000531, 0.00000137, -0.00000068, -0.00000073, -0.00000368, -0.00001051, -0.00000035, -0.00000008, -0.00000441, -0.00000215, -0.00000034, -0.00000170, -0.00000511, -0.00000673, 0.00000135, -0.00000936, -0.00000087, -0.00000383, -0.00000397, 0.00000699, -0.00001201, 0.00001003, -0.00000681, -0.00000072, -0.00000532, -0.00000548, -0.00000536, 0.00000642, -0.00001868, -0.00000742, -0.00000797, -0.00000531, -0.00002950, -0.00001146, -0.00007311, -0.00004757, 0.00009936, 0.00005771, 0.00003257, 0.00002206, 0.00002406, 0.00002526, 0.00001865, 0.00002078, 0.00002819, 0.00003548, 0.00002789, 0.00009972, 0.00001721, -0.00009359, -0.00002266, -0.00002800, -0.00000732, -0.00000875, -0.00000628, -0.00000868, -0.00000230, -0.00000738, -0.00000139, -0.00000487, 0.00000112, -0.00000567, -0.00000272, -0.00000519, -0.00000082, -0.00000899, -0.00000131, -0.00000294, -0.00000090, -0.00000046, -0.00000315, -0.00000464, -0.00000232, -0.00000354, -0.00000851, -0.00000930, -0.00001593, -0.00002985, 0.00000297, 0.00002691, 0.00002742, 0.00000499, 0.00001425, 0.00000568, 0.00001136, 0.00000118, 0.00000783, 0.00000328, 0.00001245, 0.00000833, 0.00000758, 0.00000502, 0.00000626, 0.00000982, 0.00001466, 0.00001372, 0.00003914, 0.00000798, -0.00001288, -0.00002760, -0.00000515, -0.00000305, -0.00000781, -0.00000226, -0.00001764, 0.00000747, 0.00000067, -0.00001074, -0.00000611, 0.00000488, -0.00000455, -0.00000229, -0.00000768, 0.00000017, -0.00000549, 0.00000142, -0.00000198, -0.00000807, -0.00000217, -0.00000035, -0.00000285, -0.00000838, -0.00000144, 0.00000001, },
		{ -0.00000135, -0.00000197, -0.00000085, -0.00000254, -0.00000462, 0.00000267, -0.00000165, -0.00000073, -0.00000616, -0.00000863, 0.00000137, -0.00000227, -0.00000397, -0.00000118, 0.00000026, -0.00000352, -0.00000674, -0.00000398, 0.00000051, -0.00001097, 0.00000261, -0.00000768, 0.00000315, -0.00000169, -0.00000730, 0.00001033, -0.00000832, -0.00000071, -0.00000695, -0.00000578, 0.00000074, -0.00000211, -0.00001958, -0.00000234, -0.00000842, -0.00001222, -0.00002864, -0.00002143, -0.00009405, 0.00003989, 0.00008671, 0.00003857, 0.00002664, 0.00002175, 0.00002243, 0.00002389, 0.00001431, 0.00002512, 0.00002866, 0.00002997, 0.00003774, 0.00010181, -0.00002593, -0.00007796, -0.00001957, -0.00002353, -0.00000542, -0.00000883, -0.00000717, -0.00000625, -0.00000414, -0.00000549, -0.00000276, -0.00000199, -0.00000055, -0.00000551, -0.00000405, -0.00000304, -0.00000305, -0.00000792, -0.00000109, -0.00000245, -0.00000107, -0.00000050, -0.00000451, -0.00000281, -0.00000361, -0.00000372, -0.00001116, -0.00000747, -0.00002318, -0.00002541, 0.00000976, 0.00003346, 0.00002018, 0.00000676, 0.00001394, 0.00000684, 0.00000927, 0.00000253, 0.00000662, 0.00000652, 0.00001279, 0.00000708, 0.00000839, 0.00000521, 0.00000580, 0.00001550, 0.00001138, 0.00002769, 0.00003300, -0.00000832, -0.00002022, -0.00002351, 0.00000389, -0.00001227, 0.00000180, -0.00001308, -0.00001058, 0.00001008, -0.00000425, -0.00001102, -0.00000119, 0.00000221, -0.00000337, -0.00000555, -0.00000572, 0.00000082, -0.00000581, 0.00000261, -0.00000511, -0.00000662, -0.00000139, -0.00000098, -0.00000400, -0.00000712, -0.00000008, -0.00000152, },
		{ -0.00000057, -0.00000197, -0.00000031, -0.00000492, -0.00000216, 0.00000256, -0.00000231, -0.00000144, -0.00000755, -0.00000632, 0.00000179, -0.00000431, -0.00000239, -0.00000065, 0.00000025, -0.00000649, -0.00000585, -0.00000190, -0.00000178, -0.00001049, 0.00000443, -0.00000940, 0.00000775, -0.00001138, 0.00000294, 0.00000503, -0.00000655, -0.00000388, -0.00000586, -0.00000551, 0.00000538, -0.00001571, -0.00000902, -0.00000642, -0.00000441, -0.00002672, -0.00001564, -0.00006024, -0.00005029, 0.00009071, 0.00005233, 0.00003453, 0.00001980, 0.00002355, 0.00001977, 0.00002153, 0.00001290, 0.00002992, 0.00002516, 0.00002709, 0.00005228, 0.00008922, -0.00005954, -0.00005782, -0.00002082, -0.00001602, -0.00000753, -0.00000692, -0.00000873, -0.00000344, -0.00000667, -0.00000289, -0.00000421, 0.00000138, -0.00000390, -0.00000439, -0.00000530, -0.00000072, -0.00000616, -0.00000565, -0.00000158, -0.00000161, -0.00000167, -0.00000055, -0.00000516, -0.00000154, -0.00000513, -0.00000446, -0.00001241, -0.00000751, -0.00002963, -0.00001912, 0.00001792, 0.00003673, 0.00001233, 0.00001109, 0.00001221, 0.00000791, 0.00000749, 0.00000358, 0.00000631, 0.00001052, 0.00001037, 0.00000805, 0.00000876, 0.00000400, 0.00000925, 0.00001740, 0.00001334, 0.00004080, 0.00001275, -0.00001753, -0.00002615, -0.00000950, -0.00000138, -0.00000989, 0.00000273, -0.00002071, 0.00000020, 0.00000739, -0.00000819, -0.00000826, 0.00000079, 0.00000100, -0.00000395, -0.00000817, -0.00000173, -0.00000076, -0.00000488, 0.00000237, -0.00000747, -0.00000443, -0.00000173, -0.00000130, -0.00000455, -0.00000574, 0.00000018, -0.00000276, },
		{ -0.00000055, -0.00000108, -0.00000182, -0.00000531, -0.00000003, 0.00000205, -0.00000358, -0.00000135, -0.00000911, -0.00000299, -0.00000022, -0.00000395, -0.00000186, 0.00000090, -0.00000264, -0.00000703, -0.00000508, 0.00000039, -0.00000626, -0.00000613, 0.00000218, -0.00000621, 0.00000467, -0.00001363, 0.00001038, -0.00000029, -0.00000609, -0.00000531, -0.00000601, -0.00000049, -0.00000082, -0.00001888, -0.00000224, -0.00000848, -0.00000835, -0.00003003, -0.00002007, -0.00008292, 0.00002978, 0.00008169, 0.00003496, 0.00002812, 0.00002005, 0.00002011, 0.00002143, 0.00001415, 0.00002001, 0.00002678, 0.00002485, 0.00002458, 0.00007022, 0.00006090, -0.00007672, -0.00004153, -0.00001979, -0.00001183, -0.00000834, -0.00000649, -0.00000811, -0.00000322, -0.00000686, -0.00000226, -0.00000301, 0.00000179, -0.00000585, -0.00000434, -0.00000448, -0.00000072, -0.00000763, -0.00000422, -0.00000113, -0.00000214, -0.00000106, -0.00000186, -0.00000391, -0.00000253, -0.00000525, -0.00000669, -0.00001143, -0.00001087, -0.00003330, -0.00001214, 0.00002817, 0.00003391, 0.00000810, 0.00001500, 0.00001012, 0.00000846, 0.00000670, 0.00000334, 0.00000921, 0.00001113, 0.00000905, 0.00000952, 0.00000835, 0.00000293, 0.00001587, 0.00001439, 0.00002629, 0.00003815, -0.00000836, -0.00002343, -0.00002286, -0.00000003, -0.00000980, -0.00000078, -0.00000649, -0.00001836, 0.00000746, 0.00000213, -0.00000955, -0.00000479, 0.00000075, 0.00000123, -0.00000748, -0.00000692, 0.00000085, -0.00000279, -0.00000351, 0.00000117, -0.00000882, -0.00000202, -0.00000336, -0.00000006, -0.00000589, -0.00000368, -0.00000127, -0.00000232, },
		{ -0.00000069, -0.00000060, -0.00000396, -0.00000402, 0.00000143, 0.00000066, -0.00000421, -0.00000134, -0.00001003, -0.00000020, -0.00000286, -0.00000214, -0.00000180, 0.00000182, -0.00000666, -0.00000525, -0.00000474, 0.00000168, -0.00001041, -0.00000042, -0.00000172, -0.00000151, -0.00000309, -0.00000790, 0.00001188, -0.00000372, -0.00000724, -0.00000494, -0.00000550, 0.00000380, -0.00001290, -0.00001037, -0.00000456, -0.00000581, -0.00002018, -0.00002163, -0.00005010, -0.00004954, 0.00008154, 0.00004886, 0.00003266, 0.00002017, 0.00002265, 0.00001587, 0.00002208, 0.00000906, 0.00002836, 0.00001926, 0.00002706, 0.00002502, 0.00008462, 0.00002383, -0.00007927, -0.00002966, -0.00001706, -0.00001098, -0.00000723, -0.00000740, -0.00000615, -0.00000473, -0.00000537, -0.00000259, -0.00000045, -0.00000029, -0.00000632, -0.00000469, -0.00000271, -0.00000239, -0.00000770, -0.00000314, -0.00000068, -0.00000303, -0.00000010, -0.00000326, -0.00000228, -0.00000470, -0.00000466, -0.00000926, -0.00000980, -0.00001624, -0.00003484, -0.00000268, 0.00003721, 0.00002675, 0.00000843, 0.00001690, 0.00000797, 0.00000955, 0.00000503, 0.00000475, 0.00001230, 0.00000930, 0.00000981, 0.00001069, 0.00000622, 0.00000584, 0.00001942, 0.00001524, 0.00004078, 0.00001855, -0.00002212, -0.00002592, -0.00001181, -0.00000227, -0.00000935, 0.00000231, -0.00001562, -0.00000983, 0.00000921, -0.00000323, -0.00000776, -0.00000341, 0.00000181, -0.00000063, -0.00001031, -0.00000326, 0.00000124, -0.00000461, -0.00000173, -0.00000123, -0.00000813, -0.00000141, -0.00000398, 0.00000118, -0.00000731, -0.00000201, -0.00000299, -0.00000100, },
		{ -0.00000024, -0.00000185, -0.00000458, -0.00000265, 0.00000266, -0.00000212, -0.00000275, -0.00000309, -0.00000892, 0.00000021, -0.00000353, -0.00000116, -0.00000071, -0.00000026, -0.00000810, -0.00000400, -0.00000335, 0.00000006, -0.00001098, 0.00000283, -0.00000350, -0.00000068, -0.00000799, -0.00000045, 0.00001032, -0.00000739, -0.00000688, -0.00000551, -0.00000092, -0.00000115, -0.00001692, -0.00000319, -0.00000731, -0.00000784, -0.00002523, -0.00002497, -0.00006962, 0.00001994, 0.00007894, 0.00002948, 0.00002934, 0.00001815, 0.00002056, 0.00001631, 0.00001683, 0.00001320, 0.00002814, 0.00001616, 0.00002664, 0.00003274, 0.00008747, -0.00001216, -0.00007271, -0.00001983, -0.00001620, -0.00000980, -0.00000670, -0.00000743, -0.00000529, -0.00000547, -0.00000429, -0.00000175, 0.00000068, -0.00000249, -0.00000668, -0.00000395, -0.00000205, -0.00000384, -0.00000761, -0.00000140, -0.00000166, -0.00000260, -0.00000028, -0.00000324, -0.00000232, -0.00000607, -0.00000475, -0.00001083, -0.00000921, -0.00002223, -0.00003460, 0.00001049, 0.00004106, 0.00001953, 0.00001154, 0.00001640, 0.00000665, 0.00001071, 0.00000250, 0.00000943, 0.00001189, 0.00000853, 0.00001136, 0.00001086, 0.00000394, 0.00001239, 0.00001815, 0.00002648, 0.00004144, -0.00000642, -0.00002830, -0.00002145, -0.00000385, -0.00000815, -0.00000252, -0.00000381, -0.00001694, -0.00000202, 0.00000708, -0.00000649, -0.00000489, -0.00000350, 0.00000390, -0.00000566, -0.00000923, -0.00000036, 0.00000041, -0.00000597, 0.00000014, -0.00000431, -0.00000574, -0.00000311, -0.00000188, 0.00000037, -0.00000751, -0.00000186, -0.00000331, -0.00000016, },
		{ -0.00000013, -0.00000375, -0.00000375, -0.00000149, 0.00000284, -0.00000469, -0.00000029, -0.00000604, -0.00000656, -0.00000090, -0.00000239, -0.00000103, 0.00000030, -0.00000412, -0.00000659, -0.00000410, -0.00000131, -0.00000340, -0.00000841, 0.00000363, -0.00000400, -0.00000319, -0.00000764, 0.00000490, 0.00000727, -0.00001096, -0.00000529, -0.00000530, 0.00000307, -0.00001183, -0.00001013, -0.00000395, -0.00000640, -0.00001607, -0.00002162, -0.00004828, -0.00004287, 0.00006973, 0.00004982, 0.00002605, 0.00002318, 0.00001947, 0.00001664, 0.00001742, 0.00001139, 0.00002128, 0.00002082, 0.00001879, 0.00002348, 0.00004637, 0.00007686, -0.00004149, -0.00005977, -0.00001327, -0.00001671, -0.00000766, -0.00000707, -0.00000666, -0.00000546, -0.00000532, -0.00000334, -0.00000034, 0.00000021, -0.00000428, -0.00000660, -0.00000277, -0.00000236, -0.00000499, -0.00000695, 0.00000006, -0.00000323, -0.00000131, -0.00000110, -0.00000248, -0.00000384, -0.00000633, -0.00000552, -0.00001167, -0.00000965, -0.00002909, -0.00003051, 0.00002493, 0.00003917, 0.00001509, 0.00001517, 0.00001375, 0.00000745, 0.00000977, 0.00000237, 0.00001422, 0.00000869, 0.00001024, 0.00001247, 0.00000936, 0.00000493, 0.00001735, 0.00001926, 0.00004117, 0.00002322, -0.00002501, -0.00002791, -0.00001211, -0.00000476, -0.00000765, 0.00000014, -0.00001115, -0.00001221, 0.00000200, 0.00000343, -0.00000668, -0.00000364, -0.00000231, 0.00000369, -0.00001043, -0.00000526, 0.00000068, -0.00000112, -0.00000621, 0.00000098, -0.00000642, -0.00000386, -0.00000483, 0.00000153, -0.00000227, -0.00000679, -0.00000252, -0.00000248, 0.00000034, },
		{ -0.00000153, -0.00000433, -0.00000293, -0.00000003, 0.00000104, -0.00000490, 0.00000077, -0.00000831, -0.00000473, -0.00000139, -0.00000129, -0.00000053, -0.00000093, -0.00000629, -0.00000478, -0.00000411, -0.00000043, -0.00000602, -0.00000524, 0.00000389, -0.00000628, -0.00000424, -0.00000546, 0.00000886, 0.00000171, -0.00001183, -0.00000481, -0.00000135, -0.00000117, -0.00001594, -0.00000325, -0.00000653, -0.00000811, -0.00002012, -0.00002698, -0.00006303, 0.00001667, 0.00007271, 0.00002870, 0.00002493, 0.00002006, 0.00001834, 0.00001571, 0.00001436, 0.00001282, 0.00002363, 0.00001497, 0.00002192, 0.00002204, 0.00006027, 0.00005565, -0.00006153, -0.00004249, -0.00001229, -0.00001553, -0.00000645, -0.00000688, -0.00000657, -0.00000522, -0.00000531, -0.00000138, -0.00000017, -0.00000042, -0.00000628, -0.00000515, -0.00000265, -0.00000241, -0.00000655, -0.00000485, -0.00000026, -0.00000364, -0.00000057, -0.00000131, -0.00000262, -0.00000537, -0.00000620, -0.00000645, -0.00001233, -0.00001104, -0.00003670, -0.00002054, 0.00003654, 0.00003414, 0.00001383, 0.00001744, 0.00001056, 0.00000969, 0.00000663, 0.00000670, 0.00001485, 0.00000658, 0.00001308, 0.00001247, 0.00000743, 0.00000945, 0.00001851, 0.00002959, 0.00004407, -0.00000461, -0.00003248, -0.00002167, -0.00000564, -0.00000874, -0.00000192, -0.00000509, -0.00001217, -0.00000751, 0.00000320, 0.00000024, -0.00000513, -0.00000396, 0.00000046, -0.00000041, -0.00001125, -0.00000165, 0.00000063, -0.00000322, -0.00000501, 0.00000026, -0.00000682, -0.00000390, -0.00000413, 0.00000356, -0.00000515, -0.00000612, -0.00000266, -0.00000159, 0.00000109, },
		{ -0.00000364, -0.00000339, -0.00000253, 0.00000114, -0.00000166, -0.00000291, -0.00000029, -0.00000940, -0.00000350, -0.00000092, -0.00000059, -0.00000035, -0.00000373, -0.00000576, -0.00000418, -0.00000351, -0.00000079, -0.00000725, -0.00000191, 0.00000271, -0.00000929, -0.00000256, -0.00000362, 0.00001146, -0.00000535, -0.00001019, -0.00000406, 0.00000232, -0.00001106, -0.00000992, -0.00000339, -0.00000642, -0.00001431, -0.00001831, -0.00004793, -0.00003872, 0.00006212, 0.00004895, 0.00002216, 0.00002193, 0.00001931, 0.00001606, 0.00001524, 0.00001094, 0.00001816, 0.00001929, 0.00001432, 0.00002252, 0.00002485, 0.00006989, 0.00002717, -0.00006933, -0.00002620, -0.00001501, -0.00001232, -0.00000646, -0.00000624, -0.00000695, -0.00000472, -0.00000483, 0.00000064, -0.00000104, -0.00000115, -0.00000775, -0.00000317, -0.00000332, -0.00000241, -0.00000782, -0.00000227, -0.00000188, -0.00000280, -0.00000044, -0.00000115, -0.00000383, -0.00000630, -0.00000586, -0.00000786, -0.00001236, -0.00001463, -0.00004261, -0.00000572, 0.00004317, 0.00002871, 0.00001475, 0.00001732, 0.00000899, 0.00001071, 0.00000465, 0.00001269, 0.00001119, 0.00000801, 0.00001504, 0.00001127, 0.00000760, 0.00001371, 0.00002134, 0.00004449, 0.00002665, -0.00002741, -0.00003015, -0.00001303, -0.00000601, -0.00000778, 0.00000028, -0.00001166, -0.00000771, -0.00000609, 0.00000369, -0.00000164, -0.00000389, -0.00000375, 0.00000198, -0.00000564, -0.00000831, -0.00000005, 0.00000011, -0.00000522, -0.00000312, -0.00000127, -0.00000658, -0.00000466, -0.00000126, 0.00000316, -0.00000748, -0.00000542, -0.00000218, -0.00000075, 0.00000137, },
		{ -0.00000457, -0.00000234, -0.00000201, 0.00000096, -0.00000316, -0.00000074, -0.00000246, -0.00000991, -0.00000176, -0.00000060, 0.00000037, -0.00000216, -0.00000519, -0.00000454, -0.00000410, -0.00000305, -0.00000119, -0.00000797, 0.00000209, -0.00000200, -0.00000889, -0.00000142, -0.00000074, 0.00001061, -0.00001070, -0.00000825, -0.00000051, -0.00000123, -0.00001556, -0.00000282, -0.00000629, -0.00000759, -0.00001773, -0.00002418, -0.00006116, 0.00001510, 0.00006731, 0.00002885, 0.00002011, 0.00002040, 0.00001719, 0.00001583, 0.00001171, 0.00001280, 0.00001962, 0.00001457, 0.00001615, 0.00002172, 0.00003143, 0.00007233, -0.00000404, -0.00006421, -0.00001640, -0.00001694, -0.00000935, -0.00000631, -0.00000638, -0.00000657, -0.00000488, -0.00000273, 0.00000085, -0.00000140, -0.00000284, -0.00000743, -0.00000235, -0.00000343, -0.00000333, -0.00000752, -0.00000096, -0.00000308, -0.00000181, -0.00000013, -0.00000178, -0.00000508, -0.00000692, -0.00000520, -0.00001010, -0.00001139, -0.00002149, -0.00004364, 0.00001059, 0.00004546, 0.00002420, 0.00001641, 0.00001542, 0.00000948, 0.00000934, 0.00000690, 0.00001533, 0.00000737, 0.00001184, 0.00001521, 0.00000996, 0.00001039, 0.00001557, 0.00003256, 0.00004910, -0.00000431, -0.00003593, -0.00002264, -0.00000719, -0.00000911, -0.00000211, -0.00000519, -0.00001179, -0.00000409, -0.00000636, 0.00000449, -0.00000270, -0.00000338, -0.00000240, 0.00000029, -0.00000814, -0.00000472, 0.00000022, -0.00000101, -0.00000626, -0.00000152, -0.00000278, -0.00000680, -0.00000410, 0.00000162, 0.00000106, -0.00000930, -0.00000399, -0.00000189, 0.00000048, 0.00000010, },
		{ -0.00000379, -0.00000196, -0.00000143, -0.00000021, -0.00000288, 0.00000033, -0.00000512, -0.00000953, 0.00000035, -0.00000053, 0.00000086, -0.00000537, -0.00000419, -0.00000425, -0.00000395, -0.00000265, -0.00000175, -0.00000757, 0.00000477, -0.00000835, -0.00000441, -0.00000273, 0.00000375, 0.00000593, -0.00001316, -0.00000552, 0.00000260, -0.00001040, -0.00001027, -0.00000229, -0.00000694, -0.00001226, -0.00001685, -0.00004311, -0.00004029, 0.00005860, 0.00004660, 0.00002120, 0.00001766, 0.00002009, 0.00001477, 0.00001542, 0.00000835, 0.00001771, 0.00001577, 0.00001328, 0.00001770, 0.00002104, 0.00004073, 0.00006488, -0.00003072, -0.00005115, -0.00001350, -0.00001620, -0.00000768, -0.00000577, -0.00000716, -0.00000561, -0.00000502, 0.00000007, -0.00000065, -0.00000106, -0.00000508, -0.00000570, -0.00000284, -0.00000285, -0.00000494, -0.00000590, -0.00000121, -0.00000320, -0.00000105, 0.00000022, -0.00000336, -0.00000584, -0.00000717, -0.00000485, -0.00001242, -0.00001058, -0.00003053, -0.00003876, 0.00002582, 0.00004471, 0.00002093, 0.00001750, 0.00001347, 0.00001021, 0.00000785, 0.00001195, 0.00001295, 0.00000720, 0.00001516, 0.00001415, 0.00001012, 0.00001337, 0.00001910, 0.00004950, 0.00003153, -0.00003140, -0.00003148, -0.00001500, -0.00000640, -0.00000865, 0.00000097, -0.00001293, -0.00000529, -0.00000550, -0.00000511, 0.00000506, -0.00000368, -0.00000249, -0.00000181, -0.00000284, -0.00000718, -0.00000271, 0.00000018, -0.00000272, -0.00000605, -0.00000056, -0.00000448, -0.00000689, -0.00000200, 0.00000309, -0.00000200, -0.00001013, -0.00000203, -0.00000184, 0.00000157, -0.00000222, },
		{ -0.00000247, -0.00000195, -0.00000147, -0.00000088, -0.00000199, 0.00000042, -0.00000823, -0.00000729, 0.00000155, 0.00000011, -0.00000085, -0.00000718, -0.00000246, -0.00000458, -0.00000397, -0.00000158, -0.00000336, -0.00000485, 0.00000309, -0.00001127, -0.00000000, -0.00000438, 0.00000773, -0.00000040, -0.00001346, -0.00000057, -0.00000079, -0.00001531, -0.00000270, -0.00000580, -0.00000718, -0.00001593, -0.00002169, -0.00005635, 0.00000853, 0.00006645, 0.00002689, 0.00001866, 0.00001656, 0.00001845, 0.00001434, 0.00001198, 0.00001005, 0.00001900, 0.00001158, 0.00001405, 0.00001866, 0.00002128, 0.00005076, 0.00004732, -0.00004677, -0.00003766, -0.00001362, -0.00001416, -0.00000649, -0.00000586, -0.00000734, -0.00000515, -0.00000373, 0.00000140, -0.00000206, -0.00000114, -0.00000643, -0.00000419, -0.00000339, -0.00000259, -0.00000602, -0.00000436, -0.00000185, -0.00000280, -0.00000008, -0.00000042, -0.00000486, -0.00000648, -0.00000663, -0.00000572, -0.00001366, -0.00001173, -0.00003919, -0.00002913, 0.00003929, 0.00004137, 0.00001905, 0.00001744, 0.00001262, 0.00000959, 0.00000949, 0.00001492, 0.00000909, 0.00001035, 0.00001653, 0.00001301, 0.00001227, 0.00001464, 0.00003106, 0.00005749, -0.00000402, -0.00004057, -0.00002273, -0.00000991, -0.00000825, -0.00000373, -0.00000356, -0.00001438, 0.00000087, -0.00001003, -0.00000123, 0.00000439, -0.00000437, -0.00000103, -0.00000367, -0.00000394, -0.00000539, -0.00000188, -0.00000023, -0.00000433, -0.00000506, -0.00000027, -0.00000668, -0.00000539, 0.00000006, 0.00000327, -0.00000579, -0.00000905, -0.00000070, -0.00000139, 0.00000134, -0.00000381, },
		{ -0.00000172, -0.00000210, -0.00000194, -0.00000050, -0.00000143, -0.00000044, -0.00001096, -0.00000345, 0.00000151, 0.00000094, -0.00000457, -0.00000605, -0.00000186, -0.00000489, -0.00000393, -0.00000028, -0.00000527, -0.00000133, -0.00000220, -0.00000864, 0.00000074, -0.00000380, 0.00000907, -0.00000646, -0.00001103, 0.00000347, -0.00000973, -0.00001078, -0.00000116, -0.00000771, -0.00000972, -0.00001707, -0.00003673, -0.00004111, 0.00005179, 0.00004802, 0.00001877, 0.00001583, 0.00001697, 0.00001616, 0.00001404, 0.00000833, 0.00001503, 0.00001505, 0.00001055, 0.00001485, 0.00001915, 0.00002367, 0.00005753, 0.00002422, -0.00005191, -0.00002772, -0.00001372, -0.00001222, -0.00000532, -0.00000673, -0.00000666, -0.00000508, -0.00000123, 0.00000060, -0.00000239, -0.00000205, -0.00000648, -0.00000369, -0.00000333, -0.00000302, -0.00000617, -0.00000362, -0.00000209, -0.00000221, 0.00000088, -0.00000229, -0.00000558, -0.00000717, -0.00000554, -0.00000782, -0.00001372, -0.00001540, -0.00004584, -0.00001573, 0.00005054, 0.00003563, 0.00001859, 0.00001653, 0.00001237, 0.00000869, 0.00001403, 0.00001297, 0.00000846, 0.00001349, 0.00001686, 0.00001255, 0.00001517, 0.00001696, 0.00005099, 0.00004076, -0.00003711, -0.00003294, -0.00001666, -0.00000757, -0.00000848, 0.00000016, -0.00001203, -0.00000746, 0.00000058, -0.00001275, 0.00000368, 0.00000215, -0.00000376, -0.00000078, -0.00000646, -0.00000211, -0.00000515, -0.00000095, -0.00000147, -0.00000511, -0.00000380, -0.00000105, -0.00000838, -0.00000235, 0.00000096, 0.00000246, -0.00000955, -0.00000609, -0.00000055, -0.00000051, -0.00000021, -0.00000372, },
		},
		{ { -0.00000026, 0.00000028, -0.00000190, -0.00000112, -0.00000038, 0.00000112, 0.00000814, -0.00000119, -0.00000150, -0.00000061, 0.00000342, -0.00000038, -0.00000037, 0.00000177, -0.00000086, -0.00000190, 0.00000412, -0.00000154, -0.00000074, 0.00000133, -0.00000053, -0.00000131, -0.00001349, 0.00000312, 0.00000047, -0.00000474, -0.00000046, -0.00000341, -0.00000254, -0.00000243, -0.00000301, -0.00000134, -0.00000168, -0.00000163, -0.00000452, -0.00000034, -0.00000081, -0.00000260, -0.00000881, -0.00000360, -0.00000456, -0.00000539, -0.00000151, -0.00000496, -0.00000803, -0.00001155, -0.00001673, -0.00002092, -0.00005446, -0.00000362, 0.00006194, 0.00002519, 0.00001768, 0.00001357, 0.00000233, 0.00001485, 0.00000694, 0.00000025, 0.00001308, 0.00000066, -0.00000232, -0.00001015, -0.00003149, 0.00003336, 0.00010699, 0.00003336, -0.00003149, -0.00001015, -0.00000232, 0.00000066, 0.00001308, 0.00000025, 0.00000694, 0.00001485, 0.00000233, 0.00001357, 0.00001768, 0.00002519, 0.00006194, -0.00000362, -0.00005446, -0.00002092, -0.00001673, -0.00001155, -0.00000803, -0.00000496, -0.00000151, -0.00000539, -0.00000456, -0.00000360, -0.00000881, -0.00000260, -0.00000081, -0.00000034, -0.00000452, -0.00000163, -0.00000168, -0.00000134, -0.00000301, -0.00000243, -0.00000254, -0.00000341, -0.00000046, -0.00000474, 0.00000047, 0.00000312, -0.00001349, -0.00000131, -0.00000053, 0.00000133, -0.00000074, -0.00000154, 0.00000412, -0.00000190, -0.00000086, 0.00000177, -0.00000037, -0.00000038, 0.00000342, -0.00000061, -0.00000150, -0.00000119, 0.00000814, 0.00000112, -0.00000038, -0.00000112, -0.00000190, 0.00000028, },
		{ -0.00000022, -0.00000009, -0.00000237, -0.00000023, -0.00000113, 0.00000344, 0.00000689, -0.00000254, -0.00000072, -0.00000045, 0.00000401, -0.00000155, 0.00000066, 0.00000137, -0.00000177, -0.00000074, 0.00000436, -0.00000343, 0.00000071, 0.00000033, 0.00000108, -0.00000513, -0.00001143, 0.00000540, -0.00000180, -0.00000405, -0.00000102, -0.00000318, -0.00000231, -0.00000252, -0.00000279, -0.00000172, -0.00000048, -0.00000342, -0.00000385, 0.00000062, -0.00000109, -0.00000341, -0.00000921, -0.00000270, -0.00000488, -0.00000508, -0.00000185, -0.00000512, -0.00000819, -0.00001250, -0.00001604, -0.00002848, -0.00004946, 0.00001656, 0.00005586, 0.00001941, 0.00001741, 0.00001127, 0.00000174, 0.00001424, 0.00000659, -0.00000060, 0.00001130, 0.00000116, -0.00000391, -0.00000872, -0.00003554, 0.00003098, 0.00010710, 0.00003593, -0.00002813, -0.00001105, -0.00000085, 0.00000021, 0.00001510, 0.00000087, 0.00000749, 0.00001515, 0.00000340, 0.00001599, 0.00001766, 0.00003397, 0.00006237, -0.00002612, -0.00005254, -0.00001564, -0.00001732, -0.00001058, -0.00000794, -0.00000479, -0.00000129, -0.00000554, -0.00000433, -0.00000473, -0.00000779, -0.00000218, -0.00000046, -0.00000188, -0.00000419, -0.00000038, -0.00000287, -0.00000086, -0.00000332, -0.00000223, -0.00000299, -0.00000328, -0.00000024, -0.00000502, 0.00000278, -0.00000075, -0.00001324, 0.00000141, -0.00000216, 0.00000249, -0.00000232, 0.00000097, 0.00000279, -0.00000234, 0.00000007, 0.00000177, -0.00000108, 0.00000078, 0.00000244, -0.00000042, -0.00000252, 0.00000112, 0.00000798, -0.00000057, 0.00000034, -0.00000212, -0.00000109, 0.00000040, },
		{ -0.00000014, -0.00000068, -0.00000241, 0.00000029, -0.00000148, 0.00000580, 0.00000457, -0.00000281, -0.00000042, 0.00000026, 0.00000387, -0.00000233, 0.00000174, 0.00000057, -0.00000237, 0.00000084, 0.00000353, -0.00000439, 0.00000162, 0.00000002, 0.00000176, -0.00000892, -0.00000767, 0.00000584, -0.00000348, -0.00000331, -0.00000175, -0.00000261, -0.00000241, -0.00000233, -0.00000288, -0.00000169, 0.00000027, -0.00000524, -0.00000250, 0.00000106, -0.00000139, -0.00000442, -0.00000903, -0.00000216, -0.00000517, -0.00000460, -0.00000250, -0.00000495, -0.00000885, -0.00001273, -0.00001660, -0.00003624, -0.00003805, 0.00003165, 0.00004676, 0.00001638, 0.00001646, 0.00000941, 0.00000140, 0.00001338, 0.00000647, -0.00000175, 0.00000999, 0.00000139, -0.00000527, -0.00000708, -0.00004014, 0.00002879, 0.00010745, 0.00003854, -0.00002543, -0.00001122, 0.00000020, 0.00000016, 0.00001702, 0.00000152, 0.00000808, 0.00001527, 0.00000507, 0.00001817, 0.00001812, 0.00004525, 0.00005490, -0.00004702, -0.00004513, -0.00001348, -0.00001690, -0.00001019, -0.00000750, -0.00000491, -0.00000097, -0.00000564, -0.00000426, -0.00000589, -0.00000626, -0.00000219, -0.00000006, -0.00000384, -0.00000281, -0.00000006, -0.00000364, -0.00000059, -0.00000347, -0.00000218, -0.00000342, -0.00000291, -0.00000041, -0.00000459, 0.00000442, -0.00000545, -0.00001068, 0.00000227, -0.00000290, 0.00000308, -0.00000338, 0.00000348, 0.00000070, -0.00000195, 0.00000078, 0.00000149, -0.00000136, 0.00000160, 0.00000148, -0.00000023, -0.00000334, 0.00000393, 0.00000650, -0.00000130, 0.00000062, -0.00000291, -0.00000013, 0.00000032, },
		{ -0.00000008, -0.00000138, -0.00000205, 0.00000035, -0.00000109, 0.00000758, 0.00000175, -0.00000220, -0.00000054, 0.00000150, 0.00000284, -0.00000235, 0.00000250, -0.00000045, -0.00000249, 0.00000242, 0.00000192, -0.00000447, 0.00000196, 0.00000051, 0.00000095, -0.00001158, -0.00000319, 0.00000467, -0.00000428, -0.00000287, -0.00000237, -0.00000189, -0.00000278, -0.00000186, -0.00000332, -0.00000116, 0.00000028, -0.00000659, -0.00000087, 0.00000111, -0.00000181, -0.00000538, -0.00000841, -0.00000205, -0.00000525, -0.00000411, -0.00000339, -0.00000445, -0.00001013, -0.00001192, -0.00001946, -0.00004168, -0.00002231, 0.00004049, 0.00003693, 0.00001540, 0.00001471, 0.00000825, 0.00000098, 0.00001248, 0.00000637, -0.00000307, 0.00000915, 0.00000118, -0.00000612, -0.00000564, -0.00004502, 0.00002673, 0.00010813, 0.00004097, -0.00002317, -0.00001069, 0.00000076, 0.00000070, 0.00001855, 0.00000244, 0.00000850, 0.00001539, 0.00000737, 0.00001976, 0.00002017, 0.00005732, 0.00003853, -0.00006223, -0.00003514, -0.00001387, -0.00001542, -0.00001044, -0.00000666, -0.00000541, -0.00000043, -0.00000575, -0.00000448, -0.00000675, -0.00000455, -0.00000247, 0.00000014, -0.00000580, -0.00000066, -0.00000066, -0.00000383, -0.00000068, -0.00000333, -0.00000243, -0.00000363, -0.00000250, -0.00000084, -0.00000333, 0.00000468, -0.00000988, -0.00000652, 0.00000116, -0.00000222, 0.00000252, -0.00000327, 0.00000521, -0.00000153, -0.00000096, 0.00000115, 0.00000108, -0.00000124, 0.00000189, 0.00000082, -0.00000034, -0.00000347, 0.00000655, 0.00000412, -0.00000105, 0.00000025, -0.00000321, 0.00000075, 0.00000014, },
		{ -0.00000013, -0.00000204, -0.00000144, 0.00000001, 0.00000014, 0.00000824, -0.00000084, -0.00000118, -0.00000078, 0.00000293, 0.00000109, -0.00000149, 0.00000262, -0.00000142, -0.00000209, 0.00000360, -0.00000005, -0.00000389, 0.00000196, 0.00000149, -0.00000137, -0.00001234, 0.00000094, 0.00000245, -0.00000418, -0.00000290, -0.00000264, -0.00000129, -0.00000320, -0.00000129, -0.00000394, -0.00000026, -0.00000047, -0.00000716, 0.00000070, 0.00000086, -0.00000239, -0.00000611, -0.00000755, -0.00000240, -0.00000490, -0.00000395, -0.00000416, -0.00000399, -0.00001162, -0.00001047, -0.00002475, -0.00004279, -0.00000508, 0.00004334, 0.00002801, 0.00001564, 0.00001228, 0.00000791, 0.00000026, 0.00001175, 0.00000606, -0.00000427, 0.00000861, 0.00000060, -0.00000633, -0.00000472, -0.00004986, 0.00002464, 0.00010925, 0.00004295, -0.00002096, -0.00000975, 0.00000102, 0.00000172, 0.00001968, 0.00000370, 0.00000863, 0.00001570, 0.00001027, 0.00002046, 0.00002517, 0.00006722, 0.00001438, -0.00006879, -0.00002594, -0.00001522, -0.00001359, -0.00001087, -0.00000579, -0.00000601, 0.00000019, -0.00000579, -0.00000512, -0.00000696, -0.00000310, -0.00000269, -0.00000017, -0.00000720, 0.00000164, -0.00000187, -0.00000356, -0.00000106, -0.00000298, -0.00000295, -0.00000351, -0.00000223, -0.00000135, -0.00000132, 0.00000310, -0.00001282, -0.00000203, -0.00000127, -0.00000025, 0.00000068, -0.00000165, 0.00000555, -0.00000318, 0.00000022, 0.00000126, 0.00000066, -0.00000084, 0.00000169, 0.00000057, -0.00000090, -0.00000258, 0.00000828, 0.00000152, -0.00000015, -0.00000071, -0.00000291, 0.00000137, -0.00000004, },
		{ -0.00000034, -0.00000252, -0.00000079, -0.00000046, 0.00000207, 0.00000756, -0.00000264, -0.00000026, -0.00000071, 0.00000402, -0.00000091, 0.00000010, 0.00000196, -0.00000205, -0.00000130, 0.00000418, -0.00000199, -0.00000295, 0.00000193, 0.00000238, -0.00000470, -0.00001095, 0.00000382, -0.00000008, -0.00000346, -0.00000341, -0.00000240, -0.00000107, -0.00000338, -0.00000089, -0.00000445, 0.00000069, -0.00000175, -0.00000688, 0.00000197, 0.00000040, -0.00000310, -0.00000649, -0.00000665, -0.00000312, -0.00000400, -0.00000441, -0.00000437, -0.00000413, -0.00001256, -0.00000944, -0.00003144, -0.00003878, 0.00001093, 0.00004136, 0.00002098, 0.00001625, 0.00000955, 0.00000823, -0.00000084, 0.00001127, 0.00000538, -0.00000508, 0.00000807, -0.00000016, -0.00000596, -0.00000454, -0.00005433, 0.00002231, 0.00011093, 0.00004422, -0.00001839, -0.00000883, 0.00000140, 0.00000288, 0.00002063, 0.00000515, 0.00000858, 0.00001625, 0.00001371, 0.00002025, 0.00003421, 0.00007111, -0.00001402, -0.00006604, -0.00002014, -0.00001584, -0.00001237, -0.00001078, -0.00000542, -0.00000626, 0.00000059, -0.00000565, -0.00000626, -0.00000627, -0.00000230, -0.00000257, -0.00000128, -0.00000750, 0.00000341, -0.00000313, -0.00000318, -0.00000147, -0.00000264, -0.00000355, -0.00000314, -0.00000218, -0.00000167, 0.00000105, -0.00000036, -0.00001334, 0.00000137, -0.00000380, 0.00000226, -0.00000195, 0.00000125, 0.00000429, -0.00000374, 0.00000113, 0.00000129, 0.00000025, -0.00000033, 0.00000114, 0.00000066, -0.00000179, -0.00000063, 0.00000863, -0.00000057, 0.00000087, -0.00000197, -0.00000208, 0.00000164, -0.00000014, },
		{ -0.00000069, -0.00000274, -0.00000029, -0.00000076, 0.00000426, 0.00000564, -0.00000333, 0.00000020, 0.00000002, 0.00000423, -0.00000252, 0.00000192, 0.00000066, -0.00000216, -0.00000035, 0.00000408, -0.00000361, -0.00000187, 0.00000213, 0.00000258, -0.00000820, -0.00000775, 0.00000498, -0.00000224, -0.00000255, -0.00000416, -0.00000175, -0.00000128, -0.00000320, -0.00000085, -0.00000458, 0.00000131, -0.00000318, -0.00000589, 0.00000281, -0.00000024, -0.00000386, -0.00000645, -0.00000594, -0.00000396, -0.00000271, -0.00000553, -0.00000383, -0.00000523, -0.00001231, -0.00001006, -0.00003765, -0.00003024, 0.00002384, 0.00003607, 0.00001623, 0.00001649, 0.00000703, 0.00000885, -0.00000211, 0.00001092, 0.00000439, -0.00000536, 0.00000728, -0.00000082, -0.00000525, -0.00000512, -0.00005820, 0.00001959, 0.00011320, 0.00004467, -0.00001513, -0.00000831, 0.00000231, 0.00000372, 0.00002180, 0.00000647, 0.00000858, 0.00001707, 0.00001747, 0.00001962, 0.00004748, 0.00006533, -0.00004138, -0.00005591, -0.00001857, -0.00001482, -0.00001233, -0.00000978, -0.00000591, -0.00000586, 0.00000058, -0.00000534, -0.00000775, -0.00000473, -0.00000228, -0.00000205, -0.00000317, -0.00000647, 0.00000411, -0.00000394, -0.00000303, -0.00000164, -0.00000259, -0.00000396, -0.00000270, -0.00000235, -0.00000153, 0.00000318, -0.00000515, -0.00001117, 0.00000258, -0.00000516, 0.00000415, -0.00000439, 0.00000465, 0.00000180, -0.00000312, 0.00000147, 0.00000143, -0.00000014, 0.00000014, 0.00000049, 0.00000083, -0.00000266, 0.00000206, 0.00000746, -0.00000164, 0.00000148, -0.00000312, -0.00000097, 0.00000161, -0.00000018, },
		{ -0.00000114, -0.00000271, -0.00000005, -0.00000056, 0.00000617, 0.00000296, -0.00000298, 0.00000012, 0.00000147, 0.00000321, -0.00000310, 0.00000335, -0.00000088, -0.00000179, 0.00000060, 0.00000333, -0.00000473, -0.00000074, 0.00000256, 0.00000167, -0.00001094, -0.00000349, 0.00000438, -0.00000352, -0.00000191, -0.00000478, -0.00000097, -0.00000179, -0.00000269, -0.00000118, -0.00000424, 0.00000135, -0.00000439, -0.00000440, 0.00000313, -0.00000096, -0.00000462, -0.00000596, -0.00000569, -0.00000451, -0.00000145, -0.00000697, -0.00000278, -0.00000718, -0.00001077, -0.00001308, -0.00004141, -0.00001882, 0.00003274, 0.00002898, 0.00001373, 0.00001587, 0.00000522, 0.00000932, -0.00000325, 0.00001039, 0.00000334, -0.00000519, 0.00000612, -0.00000109, -0.00000455, -0.00000626, -0.00006144, 0.00001643, 0.00011593, 0.00004441, -0.00001113, -0.00000830, 0.00000396, 0.00000395, 0.00002348, 0.00000741, 0.00000883, 0.00001827, 0.00002105, 0.00001981, 0.00006353, 0.00004773, -0.00006197, -0.00004240, -0.00002018, -0.00001241, -0.00001324, -0.00000807, -0.00000711, -0.00000486, 0.00000020, -0.00000505, -0.00000917, -0.00000276, -0.00000279, -0.00000138, -0.00000545, -0.00000433, 0.00000366, -0.00000407, -0.00000329, -0.00000146, -0.00000294, -0.00000399, -0.00000235, -0.00000271, -0.00000072, 0.00000428, -0.00001018, -0.00000687, 0.00000130, -0.00000448, 0.00000442, -0.00000554, 0.00000745, -0.00000108, -0.00000165, 0.00000127, 0.00000172, -0.00000054, 0.00000046, -0.00000006, 0.00000082, -0.00000307, 0.00000488, 0.00000511, -0.00000154, 0.00000131, -0.00000374, 0.00000014, 0.00000140, -0.00000019, },
		{ -0.00000160, -0.00000245, -0.00000008, 0.00000031, 0.00000724, 0.00000017, -0.00000198, -0.00000023, 0.00000335, 0.00000107, -0.00000237, 0.00000384, -0.00000218, -0.00000111, 0.00000143, 0.00000201, -0.00000525, 0.00000040, 0.00000304, -0.00000039, -0.00001216, 0.00000086, 0.00000242, -0.00000379, -0.00000182, -0.00000493, -0.00000042, -0.00000228, -0.00000210, -0.00000170, -0.00000355, 0.00000077, -0.00000509, -0.00000266, 0.00000291, -0.00000164, -0.00000534, -0.00000503, -0.00000611, -0.00000433, -0.00000083, -0.00000804, -0.00000192, -0.00000922, -0.00000867, -0.00001825, -0.00004141, -0.00000647, 0.00003754, 0.00002145, 0.00001302, 0.00001426, 0.00000435, 0.00000933, -0.00000401, 0.00000941, 0.00000256, -0.00000480, 0.00000460, -0.00000078, -0.00000429, -0.00000758, -0.00006422, 0.00001295, 0.00011884, 0.00004382, -0.00000668, -0.00000858, 0.00000618, 0.00000365, 0.00002564, 0.00000796, 0.00000932, 0.00002010, 0.00002374, 0.00002279, 0.00007889, 0.00001900, -0.00007152, -0.00003008, -0.00002270, -0.00000983, -0.00001418, -0.00000642, -0.00000837, -0.00000373, -0.00000021, -0.00000514, -0.00000996, -0.00000100, -0.00000331, -0.00000112, -0.00000737, -0.00000177, 0.00000239, -0.00000367, -0.00000382, -0.00000109, -0.00000356, -0.00000366, -0.00000215, -0.00000317, 0.00000089, 0.00000365, -0.00001407, -0.00000180, -0.00000183, -0.00000178, 0.00000263, -0.00000467, 0.00000863, -0.00000334, -0.00000002, 0.00000083, 0.00000205, -0.00000090, 0.00000057, -0.00000036, 0.00000044, -0.00000256, 0.00000706, 0.00000223, -0.00000050, 0.00000032, -0.00000364, 0.00000097, 0.00000113, -0.00000022, },
		{ -0.00000204, -0.00000206, -0.00000029, 0.00000186, 0.00000708, -0.00000203, -0.00000088, -0.00000036, 0.00000501, -0.00000167, -0.00000046, 0.00000318, -0.00000282, -0.00000038, 0.00000208, 0.00000025, -0.00000509, 0.00000153, 0.00000325, -0.00000331, -0.00001147, 0.00000435, -0.00000022, -0.00000322, -0.00000234, -0.00000445, -0.00000038, -0.00000244, -0.00000172, -0.00000211, -0.00000279, -0.00000026, -0.00000517, -0.00000090, 0.00000215, -0.00000217, -0.00000598, -0.00000380, -0.00000716, -0.00000322, -0.00000131, -0.00000810, -0.00000208, -0.00001033, -0.00000733, -0.00002424, -0.00003748, 0.00000515, 0.00003858, 0.00001464, 0.00001332, 0.00001197, 0.00000433, 0.00000885, -0.00000436, 0.00000788, 0.00000229, -0.00000442, 0.00000288, 0.00000018, -0.00000484, -0.00000861, -0.00006688, 0.00000941, 0.00012154, 0.00004345, -0.00000229, -0.00000864, 0.00000846, 0.00000325, 0.00002793, 0.00000843, 0.00000990, 0.00002285, 0.00002493, 0.00003067, 0.00008845, -0.00001681, -0.00006873, -0.00002244, -0.00002379, -0.00000851, -0.00001405, -0.00000579, -0.00000888, -0.00000310, -0.00000025, -0.00000596, -0.00000960, -0.00000011, -0.00000329, -0.00000179, -0.00000816, 0.00000033, 0.00000097, -0.00000315, -0.00000430, -0.00000087, -0.00000413, -0.00000319, -0.00000207, -0.00000365, 0.00000315, 0.00000094, -0.00001554, 0.00000232, -0.00000533, 0.00000201, -0.00000073, -0.00000179, 0.00000771, -0.00000421, 0.00000110, 0.00000056, 0.00000221, -0.00000113, 0.00000047, -0.00000039, -0.00000033, -0.00000096, 0.00000793, -0.00000034, 0.00000087, -0.00000119, -0.00000282, 0.00000139, 0.00000089, -0.00000030, },
		{ -0.00000242, -0.00000162, -0.00000050, 0.00000385, 0.00000562, -0.00000319, -0.00000019, 0.00000021, 0.00000571, -0.00000412, 0.00000200, 0.00000154, -0.00000268, 0.00000022, 0.00000249, -0.00000177, -0.00000420, 0.00000257, 0.00000285, -0.00000648, -0.00000896, 0.00000630, -0.00000273, -0.00000229, -0.00000320, -0.00000344, -0.00000090, -0.00000212, -0.00000177, -0.00000217, -0.00000226, -0.00000151, -0.00000460, 0.00000062, 0.00000099, -0.00000254, -0.00000644, -0.00000255, -0.00000855, -0.00000137, -0.00000294, -0.00000687, -0.00000373, -0.00000972, -0.00000809, -0.00002915, -0.00003054, 0.00001505, 0.00003644, 0.00000950, 0.00001369, 0.00000960, 0.00000471, 0.00000816, -0.00000456, 0.00000597, 0.00000258, -0.00000426, 0.00000114, 0.00000168, -0.00000638, -0.00000894, -0.00006975, 0.00000600, 0.00012372, 0.00004377, 0.00000156, -0.00000793, 0.00001018, 0.00000333, 0.00002983, 0.00000923, 0.00001039, 0.00002658, 0.00002463, 0.00004464, 0.00008683, -0.00005289, -0.00005593, -0.00002062, -0.00002227, -0.00000920, -0.00001227, -0.00000670, -0.00000809, -0.00000337, 0.00000026, -0.00000750, -0.00000800, -0.00000038, -0.00000247, -0.00000361, -0.00000736, 0.00000125, 0.00000003, -0.00000292, -0.00000440, -0.00000112, -0.00000429, -0.00000282, -0.00000203, -0.00000395, 0.00000559, -0.00000359, -0.00001395, 0.00000400, -0.00000750, 0.00000534, -0.00000441, 0.00000223, 0.00000501, -0.00000346, 0.00000130, 0.00000077, 0.00000198, -0.00000116, 0.00000020, -0.00000027, -0.00000125, 0.00000160, 0.00000719, -0.00000189, 0.00000192, -0.00000274, -0.00000155, 0.00000143, 0.00000071, -0.00000042, },
		{ -0.00000270, -0.00000122, -0.00000045, 0.00000583, 0.00000314, -0.00000317, -0.00000011, 0.00000170, 0.00000492, -0.00000541, 0.00000414, -0.00000047, -0.00000194, 0.00000069, 0.00000250, -0.00000379, -0.00000265, 0.00000335, 0.00000164, -0.00000917, -0.00000513, 0.00000644, -0.00000447, -0.00000153, -0.00000398, -0.00000227, -0.00000179, -0.00000141, -0.00000224, -0.00000180, -0.00000216, -0.00000271, -0.00000342, 0.00000164, -0.00000032, -0.00000283, -0.00000652, -0.00000170, -0.00000966, 0.00000055, -0.00000515, -0.00000479, -0.00000661, -0.00000738, -0.00001160, -0.00003123, -0.00002210, 0.00002280, 0.00003179, 0.00000655, 0.00001331, 0.00000783, 0.00000494, 0.00000767, -0.00000502, 0.00000406, 0.00000325, -0.00000439, -0.00000034, 0.00000342, -0.00000882, -0.00000842, -0.00007296, 0.00000273, 0.00012535, 0.00004496, 0.00000465, -0.00000619, 0.00001099, 0.00000427, 0.00003102, 0.00001059, 0.00001091, 0.00003079, 0.00002401, 0.00006385, 0.00007022, -0.00008153, -0.00003827, -0.00002314, -0.00001864, -0.00001149, -0.00000919, -0.00000890, -0.00000614, -0.00000449, 0.00000110, -0.00000934, -0.00000557, -0.00000155, -0.00000112, -0.00000624, -0.00000508, 0.00000077, -0.00000013, -0.00000318, -0.00000402, -0.00000194, -0.00000389, -0.00000271, -0.00000202, -0.00000380, 0.00000743, -0.00000895, -0.00000965, 0.00000261, -0.00000707, 0.00000668, -0.00000685, 0.00000598, 0.00000152, -0.00000153, 0.00000066, 0.00000151, 0.00000132, -0.00000099, -0.00000013, -0.00000023, -0.00000190, 0.00000457, 0.00000498, -0.00000211, 0.00000211, -0.00000379, -0.00000020, 0.00000120, 0.00000059, -0.00000057, },
		{ -0.00000286, -0.00000094, 0.00000006, 0.00000726, 0.00000023, -0.00000227, -0.00000045, 0.00000386, 0.00000258, -0.00000498, 0.00000513, -0.00000214, -0.00000100, 0.00000117, 0.00000192, -0.00000541, -0.00000062, 0.00000371, -0.00000039, -0.00001070, -0.00000081, 0.00000496, -0.00000511, -0.00000136, -0.00000423, -0.00000138, -0.00000264, -0.00000060, -0.00000291, -0.00000112, -0.00000250, -0.00000369, -0.00000171, 0.00000191, -0.00000148, -0.00000320, -0.00000607, -0.00000165, -0.00000983, 0.00000166, -0.00000694, -0.00000285, -0.00000970, -0.00000430, -0.00001731, -0.00002967, -0.00001359, 0.00002826, 0.00002555, 0.00000580, 0.00001177, 0.00000708, 0.00000461, 0.00000769, -0.00000607, 0.00000255, 0.00000407, -0.00000483, -0.00000131, 0.00000489, -0.00001174, -0.00000725, -0.00007639, -0.00000067, 0.00012674, 0.00004678, 0.00000723, -0.00000367, 0.00001106, 0.00000597, 0.00003158, 0.00001238, 0.00001196, 0.00003438, 0.00002555, 0.00008459, 0.00003837, -0.00009654, -0.00002192, -0.00002682, -0.00001473, -0.00001392, -0.00000605, -0.00001136, -0.00000380, -0.00000587, 0.00000169, -0.00001066, -0.00000320, -0.00000291, 0.00000006, -0.00000883, -0.00000204, -0.00000075, 0.00000034, -0.00000375, -0.00000342, -0.00000310, -0.00000303, -0.00000285, -0.00000210, -0.00000290, 0.00000777, -0.00001363, -0.00000400, -0.00000126, -0.00000381, 0.00000523, -0.00000685, 0.00000802, -0.00000142, 0.00000067, -0.00000033, 0.00000248, 0.00000037, -0.00000069, -0.00000040, -0.00000048, -0.00000180, 0.00000718, 0.00000194, -0.00000114, 0.00000130, -0.00000401, 0.00000091, 0.00000085, 0.00000050, -0.00000076, },
		{ -0.00000285, -0.00000079, 0.00000118, 0.00000766, -0.00000238, -0.00000106, -0.00000071, 0.00000602, -0.00000077, -0.00000287, 0.00000459, -0.00000288, -0.00000031, 0.00000182, 0.00000057, -0.00000622, 0.00000158, 0.00000350, -0.00000306, -0.00001060, 0.00000305, 0.00000246, -0.00000473, -0.00000192, -0.00000373, -0.00000113, -0.00000308, -0.00000005, -0.00000348, -0.00000038, -0.00000318, -0.00000428, 0.00000033, 0.00000135, -0.00000224, -0.00000373, -0.00000507, -0.00000252, -0.00000867, 0.00000122, -0.00000735, -0.00000222, -0.00001161, -0.00000212, -0.00002359, -0.00002487, -0.00000582, 0.00003126, 0.00001889, 0.00000665, 0.00000922, 0.00000730, 0.00000369, 0.00000821, -0.00000779, 0.00000165, 0.00000480, -0.00000551, -0.00000155, 0.00000555, -0.00001448, -0.00000592, -0.00007965, -0.00000471, 0.00012851, 0.00004865, 0.00000996, -0.00000106, 0.00001101, 0.00000788, 0.00003204, 0.00001412, 0.00001429, 0.00003603, 0.00003247, 0.00010053, -0.00000444, -0.00009551, -0.00001171, -0.00002841, -0.00001266, -0.00001480, -0.00000434, -0.00001274, -0.00000214, -0.00000675, 0.00000136, -0.00001065, -0.00000182, -0.00000360, 0.00000023, -0.00001034, 0.00000068, -0.00000256, 0.00000094, -0.00000420, -0.00000305, -0.00000412, -0.00000207, -0.00000302, -0.00000235, -0.00000112, 0.00000591, -0.00001602, 0.00000100, -0.00000597, 0.00000126, 0.00000132, -0.00000412, 0.00000752, -0.00000273, 0.00000218, -0.00000099, 0.00000319, -0.00000057, -0.00000041, -0.00000050, -0.00000111, -0.00000061, 0.00000863, -0.00000103, 0.00000044, -0.00000029, -0.00000336, 0.00000158, 0.00000047, 0.00000047, -0.00000103, },
		{ -0.00000265, -0.00000076, 0.00000283, 0.00000674, -0.00000404, -0.00000012, -0.00000029, 0.00000725, -0.00000410, 0.00000021, 0.00000268, -0.00000253, -0.00000006, 0.00000265, -0.00000156, -0.00000587, 0.00000353, 0.00000264, -0.00000597, -0.00000874, 0.00000563, -0.00000022, -0.00000376, -0.00000296, -0.00000256, -0.00000161, -0.00000290, 0.00000000, -0.00000370, 0.00000018, -0.00000405, -0.00000437, 0.00000239, 0.00000009, -0.00000257, -0.00000433, -0.00000379, -0.00000402, -0.00000633, -0.00000091, -0.00000593, -0.00000361, -0.00001128, -0.00000247, -0.00002828, -0.00001820, 0.00000107, 0.00003156, 0.00001308, 0.00000805, 0.00000632, 0.00000803, 0.00000253, 0.00000887, -0.00000989, 0.00000134, 0.00000532, -0.00000631, -0.00000102, 0.00000502, -0.00001642, -0.00000496, -0.00008230, -0.00000991, 0.00013128, 0.00004995, 0.00001352, 0.00000082, 0.00001169, 0.00000923, 0.00003306, 0.00001534, 0.00001841, 0.00003503, 0.00004731, 0.00010427, -0.00004995, -0.00008078, -0.00000943, -0.00002614, -0.00001363, -0.00001314, -0.00000497, -0.00001216, -0.00000193, -0.00000664, -0.00000016, -0.00000893, -0.00000201, -0.00000310, -0.00000112, -0.00001001, 0.00000213, -0.00000385, 0.00000115, -0.00000410, -0.00000337, -0.00000451, -0.00000143, -0.00000302, -0.00000274, 0.00000130, 0.00000177, -0.00001506, 0.00000342, -0.00000937, 0.00000624, -0.00000361, 0.00000051, 0.00000463, -0.00000201, 0.00000244, -0.00000082, 0.00000322, -0.00000117, -0.00000028, -0.00000045, -0.00000201, 0.00000167, 0.00000842, -0.00000308, 0.00000184, -0.00000210, -0.00000208, 0.00000178, 0.00000014, 0.00000050, -0.00000140, },
		{ -0.00000226, -0.00000070, 0.00000474, 0.00000458, -0.00000447, 0.00000017, 0.00000119, 0.00000679, -0.00000624, 0.00000316, 0.00000014, -0.00000136, -0.00000016, 0.00000344, -0.00000422, -0.00000428, 0.00000485, 0.00000114, -0.00000859, -0.00000540, 0.00000643, -0.00000229, -0.00000283, -0.00000399, -0.00000113, -0.00000257, -0.00000221, -0.00000044, -0.00000349, 0.00000040, -0.00000498, -0.00000383, 0.00000404, -0.00000149, -0.00000266, -0.00000474, -0.00000276, -0.00000542, -0.00000361, -0.00000407, -0.00000312, -0.00000680, -0.00000861, -0.00000606, -0.00002964, -0.00001135, 0.00000730, 0.00002904, 0.00000921, 0.00000891, 0.00000390, 0.00000859, 0.00000167, 0.00000913, -0.00001187, 0.00000137, 0.00000558, -0.00000704, 0.00000004, 0.00000325, -0.00001724, -0.00000473, -0.00008414, -0.00001649, 0.00013536, 0.00005038, 0.00001827, 0.00000149, 0.00001362, 0.00000953, 0.00003503, 0.00001602, 0.00002399, 0.00003220, 0.00007024, 0.00008981, -0.00008806, -0.00005875, -0.00001336, -0.00002058, -0.00001717, -0.00000931, -0.00000772, -0.00000967, -0.00000317, -0.00000569, -0.00000247, -0.00000588, -0.00000360, -0.00000155, -0.00000379, -0.00000778, 0.00000191, -0.00000420, 0.00000072, -0.00000333, -0.00000450, -0.00000405, -0.00000131, -0.00000282, -0.00000300, 0.00000367, -0.00000389, -0.00001076, 0.00000222, -0.00000969, 0.00000906, -0.00000757, 0.00000536, 0.00000046, 0.00000030, 0.00000149, 0.00000028, 0.00000238, -0.00000127, -0.00000032, -0.00000038, -0.00000284, 0.00000465, 0.00000652, -0.00000368, 0.00000238, -0.00000351, -0.00000057, 0.00000163, -0.00000014, 0.00000059, -0.00000188, },
		{ -0.00000178, -0.00000040, 0.00000642, 0.00000167, -0.00000378, -0.00000013, 0.00000362, 0.00000438, -0.00000635, 0.00000490, -0.00000208, -0.00000000, -0.00000025, 0.00000374, -0.00000681, -0.00000168, 0.00000530, -0.00000091, -0.00001036, -0.00000122, 0.00000542, -0.00000325, -0.00000249, -0.00000447, -0.00000000, -0.00000354, -0.00000135, -0.00000116, -0.00000298, 0.00000031, -0.00000587, -0.00000257, 0.00000484, -0.00000290, -0.00000282, -0.00000460, -0.00000255, -0.00000582, -0.00000168, -0.00000699, -0.00000018, -0.00001056, -0.00000474, -0.00001223, -0.00002709, -0.00000550, 0.00001293, 0.00002401, 0.00000778, 0.00000852, 0.00000254, 0.00000848, 0.00000148, 0.00000857, -0.00001327, 0.00000145, 0.00000557, -0.00000736, 0.00000113, 0.00000065, -0.00001707, -0.00000515, -0.00008536, -0.00002419, 0.00014051, 0.00005024, 0.00002392, 0.00000116, 0.00001668, 0.00000893, 0.00003768, 0.00001690, 0.00002961, 0.00003040, 0.00009770, 0.00005531, -0.00011004, -0.00003748, -0.00001928, -0.00001422, -0.00002130, -0.00000499, -0.00001118, -0.00000632, -0.00000507, -0.00000468, -0.00000459, -0.00000253, -0.00000570, 0.00000020, -0.00000687, -0.00000437, 0.00000034, -0.00000377, -0.00000013, -0.00000223, -0.00000610, -0.00000294, -0.00000160, -0.00000258, -0.00000269, 0.00000491, -0.00000947, -0.00000442, -0.00000222, -0.00000636, 0.00000841, -0.00000880, 0.00000857, -0.00000337, 0.00000309, -0.00000002, 0.00000190, 0.00000087, -0.00000090, -0.00000047, -0.00000050, -0.00000310, 0.00000757, 0.00000345, -0.00000283, 0.00000175, -0.00000405, 0.00000076, 0.00000127, -0.00000033, 0.00000068, -0.00000237, },
		{ -0.00000135, 0.00000035, 0.00000730, -0.00000126, -0.00000250, -0.00000053, 0.00000631, 0.00000050, -0.00000431, 0.00000477, -0.00000317, 0.00000092, 0.00000007, 0.00000308, -0.00000863, 0.00000140, 0.00000482, -0.00000335, -0.00001079, 0.00000291, 0.00000307, -0.00000304, -0.00000296, -0.00000408, 0.00000037, -0.00000403, -0.00000075, -0.00000179, -0.00000242, 0.00000000, -0.00000665, -0.00000067, 0.00000454, -0.00000372, -0.00000332, -0.00000372, -0.00000349, -0.00000456, -0.00000161, -0.00000831, 0.00000133, -0.00001312, -0.00000167, -0.00001903, -0.00002151, -0.00000086, 0.00001756, 0.00001746, 0.00000845, 0.00000682, 0.00000233, 0.00000764, 0.00000196, 0.00000712, -0.00001385, 0.00000139, 0.00000528, -0.00000695, 0.00000152, -0.00000198, -0.00001652, -0.00000575, -0.00008656, -0.00003232, 0.00014603, 0.00005034, 0.00002955, 0.00000071, 0.00002006, 0.00000825, 0.00004015, 0.00001930, 0.00003314, 0.00003398, 0.00012225, 0.00000477, -0.00011169, -0.00002373, -0.00002264, -0.00001013, -0.00002353, -0.00000234, -0.00001345, -0.00000368, -0.00000646, -0.00000466, -0.00000531, -0.00000022, -0.00000712, 0.00000100, -0.00000903, -0.00000105, -0.00000167, -0.00000315, -0.00000082, -0.00000142, -0.00000752, -0.00000174, -0.00000192, -0.00000260, -0.00000138, 0.00000399, -0.00001307, 0.00000175, -0.00000813, -0.00000033, 0.00000436, -0.00000662, 0.00000888, -0.00000543, 0.00000514, -0.00000116, 0.00000331, -0.00000083, -0.00000030, -0.00000060, -0.00000097, -0.00000235, 0.00000953, 0.00000014, -0.00000109, 0.00000011, -0.00000357, 0.00000164, 0.00000083, -0.00000039, 0.00000063, -0.00000273, },
		{ -0.00000107, 0.00000167, 0.00000694, -0.00000345, -0.00000133, -0.00000034, 0.00000823, -0.00000374, -0.00000078, 0.00000286, -0.00000284, 0.00000108, 0.00000097, 0.00000119, -0.00000900, 0.00000425, 0.00000353, -0.00000596, -0.00000959, 0.00000607, 0.00000021, -0.00000202, -0.00000405, -0.00000287, -0.00000016, -0.00000378, -0.00000071, -0.00000207, -0.00000203, -0.00000043, -0.00000711, 0.00000158, 0.00000319, -0.00000381, -0.00000418, -0.00000229, -0.00000539, -0.00000167, -0.00000376, -0.00000727, 0.00000034, -0.00001302, -0.00000140, -0.00002394, -0.00001481, 0.00000319, 0.00002030, 0.00001087, 0.00001019, 0.00000444, 0.00000279, 0.00000647, 0.00000266, 0.00000515, -0.00001371, 0.00000111, 0.00000469, -0.00000558, 0.00000061, -0.00000379, -0.00001635, -0.00000589, -0.00008842, -0.00004008, 0.00015109, 0.00005164, 0.00003406, 0.00000131, 0.00002263, 0.00000854, 0.00004150, 0.00002437, 0.00003288, 0.00004720, 0.00013424, -0.00005218, -0.00009492, -0.00002047, -0.00002073, -0.00001026, -0.00002223, -0.00000288, -0.00001307, -0.00000300, -0.00000643, -0.00000626, -0.00000390, 0.00000015, -0.00000697, -0.00000003, -0.00000919, 0.00000095, -0.00000315, -0.00000302, -0.00000079, -0.00000152, -0.00000807, -0.00000105, -0.00000188, -0.00000306, 0.00000099, 0.00000041, -0.00001315, 0.00000536, -0.00001295, 0.00000619, -0.00000158, -0.00000170, 0.00000620, -0.00000504, 0.00000560, -0.00000117, 0.00000382, -0.00000217, 0.00000026, -0.00000062, -0.00000179, -0.00000040, 0.00000986, -0.00000240, 0.00000070, -0.00000200, -0.00000228, 0.00000199, 0.00000038, -0.00000027, 0.00000033, -0.00000279, },
		{ -0.00000094, 0.00000344, 0.00000522, -0.00000440, -0.00000077, 0.00000100, 0.00000835, -0.00000691, 0.00000296, -0.00000005, -0.00000141, 0.00000059, 0.00000223, -0.00000175, -0.00000760, 0.00000622, 0.00000166, -0.00000839, -0.00000681, 0.00000757, -0.00000229, -0.00000087, -0.00000526, -0.00000126, -0.00000133, -0.00000287, -0.00000126, -0.00000191, -0.00000190, -0.00000097, -0.00000700, 0.00000362, 0.00000124, -0.00000341, -0.00000507, -0.00000085, -0.00000747, 0.00000202, -0.00000751, -0.00000421, -0.00000312, -0.00000991, -0.00000494, -0.00002498, -0.00000901, 0.00000762, 0.00002025, 0.00000580, 0.00001169, 0.00000234, 0.00000314, 0.00000562, 0.00000292, 0.00000327, -0.00001318, 0.00000064, 0.00000396, -0.00000335, -0.00000183, -0.00000424, -0.00001708, -0.00000517, -0.00009135, -0.00004701, 0.00015520, 0.00005468, 0.00003674, 0.00000376, 0.00002364, 0.00001044, 0.00004144, 0.00003218, 0.00002893, 0.00007178, 0.00012486, -0.00010272, -0.00006724, -0.00002594, -0.00001392, -0.00001431, -0.00001757, -0.00000658, -0.00000986, -0.00000454, -0.00000493, -0.00000928, -0.00000054, -0.00000147, -0.00000519, -0.00000294, -0.00000699, 0.00000100, -0.00000352, -0.00000372, 0.00000015, -0.00000274, -0.00000742, -0.00000119, -0.00000133, -0.00000380, 0.00000389, -0.00000537, -0.00000930, 0.00000488, -0.00001436, 0.00001058, -0.00000701, 0.00000408, 0.00000163, -0.00000258, 0.00000443, 0.00000017, 0.00000307, -0.00000276, 0.00000056, -0.00000055, -0.00000270, 0.00000248, 0.00000832, -0.00000345, 0.00000170, -0.00000385, -0.00000060, 0.00000192, -0.00000003, 0.00000003, -0.00000028, -0.00000248, },
		{ -0.00000083, 0.00000530, 0.00000243, -0.00000404, -0.00000093, 0.00000352, 0.00000619, -0.00000786, 0.00000553, -0.00000277, 0.00000034, -0.00000002, 0.00000326, -0.00000510, -0.00000463, 0.00000690, -0.00000054, -0.00001023, -0.00000291, 0.00000715, -0.00000371, -0.00000025, -0.00000594, 0.00000009, -0.00000260, -0.00000169, -0.00000215, -0.00000146, -0.00000197, -0.00000166, -0.00000604, 0.00000482, -0.00000064, -0.00000297, -0.00000552, -0.00000017, -0.00000860, 0.00000507, -0.00001135, -0.00000056, -0.00000772, -0.00000494, -0.00001166, -0.00002159, -0.00000536, 0.00001292, 0.00001708, 0.00000323, 0.00001192, 0.00000128, 0.00000279, 0.00000556, 0.00000221, 0.00000204, -0.00001258, -0.00000005, 0.00000341, -0.00000079, -0.00000543, -0.00000341, -0.00001873, -0.00000374, -0.00009516, -0.00005328, 0.00015863, 0.00005919, 0.00003780, 0.00000795, 0.00002324, 0.00001363, 0.00004083, 0.00004117, 0.00002421, 0.00010484, 0.00008989, -0.00013490, -0.00003901, -0.00003485, -0.00000534, -0.00001981, -0.00001160, -0.00001178, -0.00000510, -0.00000734, -0.00000290, -0.00001255, 0.00000358, -0.00000408, -0.00000269, -0.00000680, -0.00000314, -0.00000064, -0.00000289, -0.00000501, 0.00000163, -0.00000470, -0.00000583, -0.00000200, -0.00000054, -0.00000425, 0.00000618, -0.00001180, -0.00000261, 0.00000032, -0.00001130, 0.00001096, -0.00000963, 0.00000846, -0.00000295, 0.00000070, 0.00000239, 0.00000239, 0.00000122, -0.00000255, 0.00000058, -0.00000055, -0.00000325, 0.00000560, 0.00000531, -0.00000285, 0.00000143, -0.00000484, 0.00000099, 0.00000156, -0.00000036, 0.00000040, -0.00000108, -0.00000185, },
		{ -0.00000045, 0.00000665, -0.00000075, -0.00000282, -0.00000140, 0.00000662, 0.00000206, -0.00000621, 0.00000600, -0.00000427, 0.00000155, -0.00000011, 0.00000330, -0.00000787, -0.00000074, 0.00000624, -0.00000290, -0.00001103, 0.00000134, 0.00000510, -0.00000384, -0.00000052, -0.00000565, 0.00000060, -0.00000338, -0.00000076, -0.00000299, -0.00000099, -0.00000210, -0.00000249, -0.00000414, 0.00000470, -0.00000181, -0.00000299, -0.00000510, -0.00000081, -0.00000783, 0.00000605, -0.00001351, 0.00000174, -0.00001131, -0.00000043, -0.00001935, -0.00001498, -0.00000371, 0.00001866, 0.00001141, 0.00000320, 0.00001062, 0.00000147, 0.00000162, 0.00000630, 0.00000046, 0.00000171, -0.00001211, -0.00000107, 0.00000356, 0.00000122, -0.00000926, -0.00000199, -0.00002071, -0.00000228, -0.00009903, -0.00005973, 0.00016233, 0.00006405, 0.00003836, 0.00001281, 0.00002256, 0.00001685, 0.00004160, 0.00004837, 0.00002424, 0.00013812, 0.00003237, -0.00014182, -0.00001954, -0.00004098, 0.00000076, -0.00002334, -0.00000729, -0.00001588, -0.00000090, -0.00000975, -0.00000181, -0.00001446, 0.00000675, -0.00000613, -0.00000095, -0.00001003, 0.00000093, -0.00000296, -0.00000198, -0.00000615, 0.00000281, -0.00000650, -0.00000406, -0.00000291, -0.00000001, -0.00000370, 0.00000650, -0.00001669, 0.00000455, -0.00000651, -0.00000454, 0.00000698, -0.00000821, 0.00000971, -0.00000574, 0.00000328, 0.00000063, 0.00000460, -0.00000114, -0.00000178, 0.00000046, -0.00000084, -0.00000290, 0.00000805, 0.00000171, -0.00000104, -0.00000009, -0.00000470, 0.00000215, 0.00000106, -0.00000050, 0.00000064, -0.00000186, -0.00000111, },
		{ 0.00000046, 0.00000691, -0.00000344, -0.00000145, -0.00000142, 0.00000917, -0.00000289, -0.00000250, 0.00000426, -0.00000405, 0.00000171, 0.00000072, 0.00000180, -0.00000910, 0.00000311, 0.00000450, -0.00000527, -0.00001039, 0.00000499, 0.00000217, -0.00000294, -0.00000159, -0.00000433, 0.00000004, -0.00000329, -0.00000046, -0.00000343, -0.00000071, -0.00000222, -0.00000330, -0.00000151, 0.00000317, -0.00000198, -0.00000365, -0.00000381, -0.00000278, -0.00000493, 0.00000425, -0.00001282, 0.00000113, -0.00001194, 0.00000113, -0.00002505, -0.00000759, -0.00000277, 0.00002351, 0.00000481, 0.00000477, 0.00000832, 0.00000243, 0.00000009, 0.00000730, -0.00000189, 0.00000211, -0.00001174, -0.00000255, 0.00000483, 0.00000179, -0.00001220, -0.00000100, -0.00002213, -0.00000176, -0.00010191, -0.00006746, 0.00016753, 0.00006784, 0.00004001, 0.00001677, 0.00002314, 0.00001853, 0.00004576, 0.00005077, 0.00003525, 0.00015954, -0.00003703, -0.00012429, -0.00001361, -0.00004016, 0.00000131, -0.00002228, -0.00000701, -0.00001667, 0.00000082, -0.00001030, -0.00000282, -0.00001377, 0.00000756, -0.00000625, -0.00000131, -0.00001110, 0.00000364, -0.00000472, -0.00000169, -0.00000635, 0.00000282, -0.00000720, -0.00000297, -0.00000326, -0.00000022, -0.00000171, 0.00000386, -0.00001802, 0.00000933, -0.00001269, 0.00000349, 0.00000012, -0.00000321, 0.00000738, -0.00000573, 0.00000404, 0.00000019, 0.00000582, -0.00000321, -0.00000090, 0.00000036, -0.00000150, -0.00000134, 0.00000896, -0.00000136, 0.00000110, -0.00000233, -0.00000357, 0.00000274, 0.00000051, -0.00000036, 0.00000054, -0.00000233, -0.00000049, },
		{ 0.00000195, 0.00000574, -0.00000497, -0.00000061, -0.00000030, 0.00000996, -0.00000705, 0.00000189, 0.00000108, -0.00000239, 0.00000089, 0.00000233, -0.00000125, -0.00000817, 0.00000599, 0.00000214, -0.00000750, -0.00000818, 0.00000718, -0.00000068, -0.00000169, -0.00000295, -0.00000243, -0.00000139, -0.00000236, -0.00000089, -0.00000338, -0.00000071, -0.00000240, -0.00000372, 0.00000124, 0.00000066, -0.00000130, -0.00000467, -0.00000218, -0.00000531, -0.00000067, 0.00000014, -0.00000935, -0.00000268, -0.00000884, -0.00000171, -0.00002635, -0.00000198, -0.00000079, 0.00002586, -0.00000081, 0.00000644, 0.00000607, 0.00000332, -0.00000106, 0.00000778, -0.00000411, 0.00000281, -0.00001136, -0.00000448, 0.00000732, 0.00000045, -0.00001342, -0.00000125, -0.00002232, -0.00000286, -0.00010305, -0.00007718, 0.00017501, 0.00006954, 0.00004389, 0.00001871, 0.00002604, 0.00001787, 0.00005413, 0.00004735, 0.00006113, 0.00015689, -0.00010251, -0.00009071, -0.00001992, -0.00003217, -0.00000391, -0.00001634, -0.00001120, -0.00001351, -0.00000060, -0.00000859, -0.00000603, -0.00001034, 0.00000566, -0.00000405, -0.00000417, -0.00000934, 0.00000408, -0.00000513, -0.00000251, -0.00000527, 0.00000129, -0.00000629, -0.00000309, -0.00000274, -0.00000119, 0.00000146, -0.00000168, -0.00001485, 0.00000961, -0.00001530, 0.00000966, -0.00000689, 0.00000343, 0.00000250, -0.00000316, 0.00000279, 0.00000146, 0.00000545, -0.00000440, -0.00000026, 0.00000036, -0.00000239, 0.00000134, 0.00000790, -0.00000299, 0.00000260, -0.00000452, -0.00000187, 0.00000281, -0.00000004, 0.00000007, -0.00000004, -0.00000227, -0.00000016, },
		{ 0.00000382, 0.00000323, -0.00000506, -0.00000058, 0.00000221, 0.00000821, -0.00000894, 0.00000532, -0.00000220, -0.00000016, -0.00000028, 0.00000406, -0.00000521, -0.00000512, 0.00000726, -0.00000044, -0.00000933, -0.00000463, 0.00000744, -0.00000258, -0.00000084, -0.00000387, -0.00000066, -0.00000308, -0.00000100, -0.00000177, -0.00000300, -0.00000084, -0.00000280, -0.00000330, 0.00000330, -0.00000201, -0.00000039, -0.00000546, -0.00000109, -0.00000713, 0.00000334, -0.00000465, -0.00000458, -0.00000840, -0.00000305, -0.00000855, -0.00002250, 0.00000033, 0.00000331, 0.00002464, -0.00000393, 0.00000691, 0.00000483, 0.00000342, -0.00000122, 0.00000708, -0.00000551, 0.00000342, -0.00001098, -0.00000651, 0.00001054, -0.00000256, -0.00001285, -0.00000289, -0.00002126, -0.00000553, -0.00010252, -0.00008874, 0.00018453, 0.00006930, 0.00005002, 0.00001867, 0.00003100, 0.00001573, 0.00006522, 0.00004072, 0.00010035, 0.00012262, -0.00014813, -0.00005397, -0.00003211, -0.00002070, -0.00001205, -0.00000780, -0.00001803, -0.00000776, -0.00000416, -0.00000563, -0.00001020, -0.00000540, 0.00000205, -0.00000042, -0.00000869, -0.00000530, 0.00000237, -0.00000423, -0.00000420, -0.00000332, -0.00000136, -0.00000402, -0.00000425, -0.00000160, -0.00000233, 0.00000471, -0.00000874, -0.00000794, 0.00000497, -0.00001270, 0.00001143, -0.00001118, 0.00000907, -0.00000286, 0.00000066, 0.00000027, 0.00000404, 0.00000356, -0.00000448, -0.00000002, 0.00000038, -0.00000310, 0.00000457, 0.00000509, -0.00000281, 0.00000279, -0.00000592, -0.00000012, 0.00000254, -0.00000051, 0.00000068, -0.00000103, -0.00000167, -0.00000006, },
		{ 0.00000551, -0.00000004, -0.00000399, -0.00000106, 0.00000563, 0.00000401, -0.00000783, 0.00000653, -0.00000427, 0.00000154, -0.00000095, 0.00000493, -0.00000895, -0.00000065, 0.00000679, -0.00000296, -0.00001039, -0.00000037, 0.00000584, -0.00000312, -0.00000093, -0.00000377, 0.00000026, -0.00000431, 0.00000018, -0.00000268, -0.00000258, -0.00000092, -0.00000357, -0.00000173, 0.00000394, -0.00000396, 0.00000006, -0.00000536, -0.00000137, -0.00000696, 0.00000528, -0.00000796, -0.00000082, -0.00001359, 0.00000290, -0.00001702, -0.00001486, -0.00000044, 0.00000934, 0.00002001, -0.00000411, 0.00000562, 0.00000494, 0.00000253, -0.00000035, 0.00000509, -0.00000583, 0.00000377, -0.00001081, -0.00000802, 0.00001348, -0.00000620, -0.00001125, -0.00000528, -0.00001970, -0.00000892, -0.00010124, -0.00010101, 0.00019472, 0.00006856, 0.00005705, 0.00001813, 0.00003630, 0.00001465, 0.00007522, 0.00003738, 0.00014429, 0.00005787, -0.00016326, -0.00002664, -0.00004194, -0.00001136, -0.00001861, -0.00000054, -0.00002417, -0.00000214, -0.00000771, -0.00000333, -0.00001330, -0.00000108, -0.00000136, 0.00000282, -0.00001301, -0.00000060, -0.00000035, -0.00000288, -0.00000584, -0.00000155, -0.00000403, -0.00000134, -0.00000571, -0.00000059, -0.00000266, 0.00000636, -0.00001493, 0.00000028, -0.00000293, -0.00000534, 0.00000794, -0.00001095, 0.00001147, -0.00000653, 0.00000394, -0.00000216, 0.00000687, 0.00000080, -0.00000368, -0.00000009, 0.00000021, -0.00000308, 0.00000741, 0.00000135, -0.00000113, 0.00000153, -0.00000613, 0.00000123, 0.00000206, -0.00000077, 0.00000122, -0.00000215, -0.00000075, 0.00000007, },
		{ 0.00000635, -0.00000318, -0.00000247, -0.00000133, 0.00000884, -0.00000156, -0.00000411, 0.00000513, -0.00000439, 0.00000197, -0.00000046, 0.00000409, -0.00001119, 0.00000409, 0.00000493, -0.00000532, -0.00001026, 0.00000368, 0.00000301, -0.00000242, -0.00000198, -0.00000248, -0.00000006, -0.00000457, 0.00000068, -0.00000319, -0.00000239, -0.00000083, -0.00000460, 0.00000094, 0.00000283, -0.00000464, -0.00000037, -0.00000413, -0.00000330, -0.00000422, 0.00000400, -0.00000810, -0.00000015, -0.00001572, 0.00000600, -0.00002372, -0.00000632, -0.00000244, 0.00001570, 0.00001343, -0.00000211, 0.00000301, 0.00000598, 0.00000115, 0.00000096, 0.00000231, -0.00000529, 0.00000400, -0.00001116, -0.00000830, 0.00001494, -0.00000905, -0.00000989, -0.00000729, -0.00001886, -0.00001169, -0.00010060, -0.00011238, 0.00020362, 0.00006943, 0.00006282, 0.00001929, 0.00003954, 0.00001781, 0.00007958, 0.00004583, 0.00017840, -0.00002615, -0.00014667, -0.00001619, -0.00004320, -0.00000877, -0.00001967, 0.00000197, -0.00002646, 0.00000061, -0.00000908, -0.00000350, -0.00001345, 0.00000064, -0.00000265, 0.00000383, -0.00001515, 0.00000282, -0.00000256, -0.00000223, -0.00000638, -0.00000113, -0.00000545, 0.00000053, -0.00000648, -0.00000052, -0.00000130, 0.00000495, -0.00001772, 0.00000655, -0.00001086, 0.00000428, 0.00000050, -0.00000629, 0.00000979, -0.00000714, 0.00000525, -0.00000317, 0.00000874, -0.00000188, -0.00000252, -0.00000027, -0.00000032, -0.00000190, 0.00000891, -0.00000218, 0.00000121, -0.00000073, -0.00000518, 0.00000199, 0.00000146, -0.00000069, 0.00000137, -0.00000301, 0.00000014, 0.00000060, },
		{ 0.00000579, -0.00000532, -0.00000130, -0.00000053, 0.00001042, -0.00000675, 0.00000088, 0.00000181, -0.00000269, 0.00000107, 0.00000124, 0.00000126, -0.00001103, 0.00000789, 0.00000228, -0.00000749, -0.00000864, 0.00000656, -0.00000008, -0.00000112, -0.00000353, -0.00000037, -0.00000147, -0.00000379, 0.00000033, -0.00000314, -0.00000250, -0.00000069, -0.00000551, 0.00000412, 0.00000033, -0.00000411, -0.00000154, -0.00000218, -0.00000625, 0.00000050, -0.00000032, -0.00000474, -0.00000336, -0.00001348, 0.00000418, -0.00002566, -0.00000011, -0.00000320, 0.00002016, 0.00000711, 0.00000048, 0.00000017, 0.00000703, 0.00000013, 0.00000177, -0.00000038, -0.00000444, 0.00000431, -0.00001218, -0.00000701, 0.00001413, -0.00000998, -0.00000991, -0.00000791, -0.00001974, -0.00001281, -0.00010164, -0.00012169, 0.00020975, 0.00007368, 0.00006562, 0.00002389, 0.00003911, 0.00002713, 0.00007560, 0.00007285, 0.00018638, -0.00011058, -0.00010713, -0.00002240, -0.00003455, -0.00001411, -0.00001408, -0.00000151, -0.00002354, -0.00000078, -0.00000732, -0.00000673, -0.00001002, -0.00000109, -0.00000093, 0.00000171, -0.00001398, 0.00000364, -0.00000316, -0.00000301, -0.00000524, -0.00000269, -0.00000481, 0.00000080, -0.00000597, -0.00000172, 0.00000188, 0.00000001, -0.00001556, 0.00000811, -0.00001530, 0.00001254, -0.00000792, 0.00000088, 0.00000486, -0.00000469, 0.00000417, -0.00000213, 0.00000878, -0.00000362, -0.00000152, -0.00000037, -0.00000117, 0.00000053, 0.00000843, -0.00000442, 0.00000314, -0.00000318, -0.00000346, 0.00000219, 0.00000083, -0.00000020, 0.00000091, -0.00000325, 0.00000071, 0.00000178, },
		{ 0.00000369, -0.00000597, -0.00000095, 0.00000178, 0.00000933, -0.00000978, 0.00000527, -0.00000201, -0.00000011, -0.00000052, 0.00000349, -0.00000303, -0.00000824, 0.00000987, -0.00000053, -0.00000940, -0.00000554, 0.00000758, -0.00000243, -0.00000005, -0.00000478, 0.00000178, -0.00000335, -0.00000239, -0.00000064, -0.00000270, -0.00000272, -0.00000080, -0.00000567, 0.00000686, -0.00000267, -0.00000300, -0.00000279, -0.00000046, -0.00000884, 0.00000550, -0.00000604, 0.00000073, -0.00000935, -0.00000759, -0.00000253, -0.00002168, 0.00000176, -0.00000086, 0.00002090, 0.00000307, 0.00000206, -0.00000177, 0.00000724, 0.00000021, 0.00000128, -0.00000217, -0.00000375, 0.00000476, -0.00001359, -0.00000453, 0.00001113, -0.00000869, -0.00001166, -0.00000693, -0.00002247, -0.00001216, -0.00010433, -0.00012903, 0.00021298, 0.00008160, 0.00006522, 0.00003194, 0.00003555, 0.00004160, 0.00006499, 0.00011911, 0.00015624, -0.00017460, -0.00006030, -0.00003803, -0.00001996, -0.00002440, -0.00000416, -0.00000918, -0.00001666, -0.00000545, -0.00000335, -0.00001187, -0.00000417, -0.00000540, 0.00000306, -0.00000288, -0.00000982, 0.00000181, -0.00000208, -0.00000504, -0.00000277, -0.00000587, -0.00000228, -0.00000047, -0.00000436, -0.00000366, 0.00000591, -0.00000739, -0.00000876, 0.00000396, -0.00001394, 0.00001622, -0.00001394, 0.00000763, -0.00000114, -0.00000049, 0.00000137, 0.00000066, 0.00000689, -0.00000400, -0.00000101, -0.00000038, -0.00000202, 0.00000378, 0.00000597, -0.00000476, 0.00000381, -0.00000500, -0.00000154, 0.00000195, 0.00000024, 0.00000061, -0.00000017, -0.00000276, 0.00000084, 0.00000357, },
		{ 0.00000046, -0.00000521, -0.00000131, 0.00000533, 0.00000537, -0.00000957, 0.00000746, -0.00000478, 0.00000209, -0.00000175, 0.00000516, -0.00000758, -0.00000346, 0.00000973, -0.00000310, -0.00001079, -0.00000140, 0.00000655, -0.00000339, 0.00000009, -0.00000499, 0.00000310, -0.00000485, -0.00000106, -0.00000172, -0.00000222, -0.00000276, -0.00000145, -0.00000452, 0.00000814, -0.00000504, -0.00000226, -0.00000325, -0.00000006, -0.00000947, 0.00000861, -0.00001066, 0.00000572, -0.00001546, -0.00000071, -0.00001178, -0.00001307, -0.00000062, 0.00000476, 0.00001750, 0.00000220, 0.00000179, -0.00000231, 0.00000635, 0.00000148, -0.00000066, -0.00000277, -0.00000336, 0.00000503, -0.00001461, -0.00000199, 0.00000706, -0.00000599, -0.00001444, -0.00000519, -0.00002609, -0.00001082, -0.00010734, -0.00013600, 0.00021493, 0.00009164, 0.00006337, 0.00004132, 0.00003197, 0.00005676, 0.00005494, 0.00017633, 0.00008599, -0.00020261, -0.00002278, -0.00005247, -0.00000661, -0.00003393, 0.00000528, -0.00001699, -0.00000914, -0.00001074, 0.00000044, -0.00001642, 0.00000161, -0.00001002, 0.00000721, -0.00000788, -0.00000443, -0.00000145, -0.00000030, -0.00000724, -0.00000025, -0.00000937, 0.00000097, -0.00000235, -0.00000261, -0.00000512, 0.00000892, -0.00001476, 0.00000033, -0.00000451, -0.00000677, 0.00001385, -0.00001513, 0.00001122, -0.00000572, 0.00000346, -0.00000164, 0.00000409, 0.00000368, -0.00000317, -0.00000098, -0.00000047, -0.00000234, 0.00000697, 0.00000220, -0.00000331, 0.00000288, -0.00000561, 0.00000006, 0.00000146, -0.00000016, 0.00000145, -0.00000160, -0.00000170, 0.00000076, 0.00000552, },
		{ -0.00000303, -0.00000367, -0.00000170, 0.00000903, -0.00000054, -0.00000619, 0.00000675, -0.00000544, 0.00000290, -0.00000170, 0.00000510, -0.00001087, 0.00000205, 0.00000778, -0.00000528, -0.00001127, 0.00000295, 0.00000394, -0.00000292, -0.00000089, -0.00000384, 0.00000303, -0.00000530, -0.00000044, -0.00000239, -0.00000204, -0.00000243, -0.00000270, -0.00000189, 0.00000737, -0.00000591, -0.00000258, -0.00000237, -0.00000160, -0.00000714, 0.00000815, -0.00001191, 0.00000756, -0.00001861, 0.00000362, -0.00001974, -0.00000314, -0.00000507, 0.00001202, 0.00001122, 0.00000399, -0.00000003, -0.00000181, 0.00000487, 0.00000328, -0.00000347, -0.00000250, -0.00000300, 0.00000459, -0.00001424, -0.00000093, 0.00000364, -0.00000346, -0.00001680, -0.00000424, -0.00002892, -0.00001060, -0.00010857, -0.00014512, 0.00021829, 0.00010106, 0.00006308, 0.00004865, 0.00003286, 0.00006628, 0.00005642, 0.00022768, -0.00001353, -0.00019000, -0.00000582, -0.00005674, -0.00000117, -0.00003712, 0.00000937, -0.00002057, -0.00000472, -0.00001360, 0.00000146, -0.00001775, 0.00000459, -0.00001236, 0.00000905, -0.00001078, -0.00000017, -0.00000427, 0.00000071, -0.00000818, 0.00000078, -0.00001148, 0.00000338, -0.00000350, -0.00000192, -0.00000476, 0.00000889, -0.00001907, 0.00000809, -0.00001388, 0.00000376, 0.00000643, -0.00001110, 0.00001032, -0.00000710, 0.00000544, -0.00000331, 0.00000674, 0.00000023, -0.00000165, -0.00000123, -0.00000084, -0.00000162, 0.00000907, -0.00000172, -0.00000085, 0.00000067, -0.00000490, 0.00000106, 0.00000082, -0.00000020, 0.00000194, -0.00000289, -0.00000050, 0.00000088, 0.00000692, },
		{ -0.00000580, -0.00000221, -0.00000119, 0.00001138, -0.00000661, -0.00000090, 0.00000360, -0.00000393, 0.00000205, -0.00000008, 0.00000276, -0.00001167, 0.00000683, 0.00000477, -0.00000716, -0.00001043, 0.00000650, 0.00000067, -0.00000154, -0.00000263, -0.00000159, 0.00000158, -0.00000447, -0.00000083, -0.00000241, -0.00000225, -0.00000184, -0.00000419, 0.00000174, 0.00000469, -0.00000514, -0.00000407, -0.00000031, -0.00000470, -0.00000212, 0.00000386, -0.00000892, 0.00000483, -0.00001679, 0.00000265, -0.00002275, 0.00000420, -0.00000836, 0.00001814, 0.00000446, 0.00000678, -0.00000229, -0.00000128, 0.00000381, 0.00000449, -0.00000609, -0.00000206, -0.00000224, 0.00000302, -0.00001179, -0.00000256, 0.00000246, -0.00000267, -0.00001732, -0.00000551, -0.00002943, -0.00001304, -0.00010626, -0.00015858, 0.00022552, 0.00010733, 0.00006700, 0.00005116, 0.00004175, 0.00006519, 0.00007978, 0.00025216, -0.00012047, -0.00014513, -0.00001123, -0.00004763, -0.00000635, -0.00003148, 0.00000568, -0.00001750, -0.00000563, -0.00001229, -0.00000163, -0.00001453, 0.00000324, -0.00001090, 0.00000705, -0.00000983, 0.00000112, -0.00000515, -0.00000013, -0.00000694, -0.00000072, -0.00001102, 0.00000372, -0.00000292, -0.00000303, -0.00000192, 0.00000475, -0.00001808, 0.00001108, -0.00002006, 0.00001365, -0.00000295, -0.00000364, 0.00000553, -0.00000500, 0.00000473, -0.00000271, 0.00000748, -0.00000237, -0.00000012, -0.00000150, -0.00000155, 0.00000039, 0.00000930, -0.00000459, 0.00000152, -0.00000205, -0.00000322, 0.00000142, 0.00000011, 0.00000024, 0.00000174, -0.00000357, 0.00000041, 0.00000163, 0.00000705, },
		{ -0.00000711, -0.00000150, 0.00000086, 0.00001105, -0.00001080, 0.00000431, -0.00000055, -0.00000114, 0.00000013, 0.00000255, -0.00000158, -0.00000946, 0.00000971, 0.00000152, -0.00000879, -0.00000808, 0.00000835, -0.00000214, -0.00000016, -0.00000429, 0.00000097, -0.00000064, -0.00000274, -0.00000206, -0.00000190, -0.00000266, -0.00000138, -0.00000519, 0.00000535, 0.00000101, -0.00000332, -0.00000607, 0.00000196, -0.00000798, 0.00000393, -0.00000277, -0.00000281, -0.00000172, -0.00001025, -0.00000421, -0.00001901, 0.00000626, -0.00000784, 0.00002059, -0.00000030, 0.00000872, -0.00000369, -0.00000170, 0.00000405, 0.00000415, -0.00000758, -0.00000206, -0.00000096, 0.00000046, -0.00000743, -0.00000708, 0.00000413, -0.00000428, -0.00001546, -0.00000947, -0.00002721, -0.00001843, -0.00010015, -0.00017691, 0.00023734, 0.00010983, 0.00007568, 0.00004877, 0.00005858, 0.00005362, 0.00012902, 0.00023202, -0.00020837, -0.00008638, -0.00003130, -0.00002897, -0.00001944, -0.00001902, -0.00000438, -0.00000865, -0.00001132, -0.00000742, -0.00000794, -0.00000760, -0.00000187, -0.00000603, 0.00000154, -0.00000503, -0.00000091, -0.00000378, -0.00000282, -0.00000372, -0.00000454, -0.00000797, 0.00000186, -0.00000061, -0.00000562, 0.00000268, -0.00000289, -0.00001153, 0.00000762, -0.00001997, 0.00001897, -0.00001042, 0.00000408, -0.00000097, -0.00000066, 0.00000189, 0.00000004, 0.00000597, -0.00000348, 0.00000094, -0.00000168, -0.00000238, 0.00000340, 0.00000741, -0.00000559, 0.00000278, -0.00000434, -0.00000114, 0.00000127, -0.00000059, 0.00000112, 0.00000076, -0.00000337, 0.00000077, 0.00000315, 0.00000553, },
		{ -0.00000681, -0.00000166, 0.00000442, 0.00000752, -0.00001163, 0.00000750, -0.00000393, 0.00000154, -0.00000170, 0.00000498, -0.00000672, -0.00000470, 0.00001017, -0.00000139, -0.00001011, -0.00000443, 0.00000808, -0.00000366, 0.00000040, -0.00000500, 0.00000283, -0.00000267, -0.00000088, -0.00000356, -0.00000127, -0.00000285, -0.00000153, -0.00000489, 0.00000766, -0.00000231, -0.00000159, -0.00000752, 0.00000312, -0.00000956, 0.00000854, -0.00000897, 0.00000366, -0.00000937, -0.00000159, -0.00001477, -0.00000956, 0.00000275, -0.00000272, 0.00001824, -0.00000164, 0.00000868, -0.00000354, -0.00000343, 0.00000577, 0.00000202, -0.00000757, -0.00000262, 0.00000048, -0.00000220, -0.00000235, -0.00001338, 0.00000787, -0.00000761, -0.00001206, -0.00001513, -0.00002338, -0.00002535, -0.00009196, -0.00019839, 0.00025199, 0.00011056, 0.00008670, 0.00004496, 0.00007852, 0.00003894, 0.00019718, 0.00016052, -0.00025507, -0.00003521, -0.00005277, -0.00000956, -0.00003352, -0.00000541, -0.00001604, 0.00000211, -0.00001872, -0.00000178, -0.00001460, 0.00000020, -0.00000820, -0.00000009, -0.00000530, 0.00000181, -0.00000498, -0.00000125, -0.00000610, 0.00000002, -0.00000920, -0.00000360, -0.00000115, 0.00000234, -0.00000830, 0.00000709, -0.00001155, -0.00000153, -0.00000133, -0.00001301, 0.00001757, -0.00001292, 0.00000888, -0.00000642, 0.00000382, -0.00000154, 0.00000384, 0.00000273, -0.00000309, 0.00000139, -0.00000185, -0.00000286, 0.00000667, 0.00000389, -0.00000461, 0.00000240, -0.00000550, 0.00000075, 0.00000079, -0.00000112, 0.00000216, -0.00000079, -0.00000238, 0.00000067, 0.00000515, 0.00000255, },
		{ -0.00000538, -0.00000212, 0.00000855, 0.00000146, -0.00000881, 0.00000759, -0.00000518, 0.00000282, -0.00000226, 0.00000584, -0.00001099, 0.00000133, 0.00000842, -0.00000371, -0.00001079, -0.00000017, 0.00000592, -0.00000359, -0.00000029, -0.00000422, 0.00000323, -0.00000364, 0.00000032, -0.00000470, -0.00000097, -0.00000253, -0.00000254, -0.00000282, 0.00000773, -0.00000408, -0.00000099, -0.00000754, 0.00000220, -0.00000799, 0.00000947, -0.00001189, 0.00000724, -0.00001449, 0.00000519, -0.00002486, 0.00000202, -0.00000399, 0.00000549, 0.00001203, 0.00000020, 0.00000672, -0.00000216, -0.00000594, 0.00000824, -0.00000122, -0.00000650, -0.00000328, 0.00000125, -0.00000362, 0.00000153, -0.00001937, 0.00001179, -0.00001090, -0.00000902, -0.00002044, -0.00002018, -0.00003116, -0.00008482, -0.00021962, 0.00026586, 0.00011354, 0.00009549, 0.00004580, 0.00009326, 0.00003465, 0.00026554, 0.00004684, -0.00025089, -0.00000798, -0.00006310, 0.00000135, -0.00004099, 0.00000276, -0.00002352, 0.00000987, -0.00002374, 0.00000108, -0.00001809, 0.00000532, -0.00001243, 0.00000378, -0.00001036, 0.00000785, -0.00000882, 0.00000058, -0.00000813, 0.00000228, -0.00001256, 0.00000007, -0.00000357, 0.00000429, -0.00000925, 0.00000878, -0.00001779, 0.00000803, -0.00001229, -0.00000146, 0.00001003, -0.00000952, 0.00000893, -0.00000866, 0.00000643, -0.00000379, 0.00000713, -0.00000111, -0.00000168, 0.00000137, -0.00000218, -0.00000246, 0.00000917, -0.00000018, -0.00000224, 0.00000047, -0.00000524, 0.00000205, 0.00000014, -0.00000130, 0.00000294, -0.00000239, -0.00000101, 0.00000053, 0.00000696, -0.00000113, },
		{ -0.00000369, -0.00000194, 0.00001171, -0.00000539, -0.00000342, 0.00000476, -0.00000396, 0.00000216, -0.00000099, 0.00000422, -0.00001283, 0.00000697, 0.00000527, -0.00000555, -0.00001037, 0.00000371, 0.00000269, -0.00000232, -0.00000198, -0.00000205, 0.00000199, -0.00000316, 0.00000037, -0.00000509, -0.00000118, -0.00000171, -0.00000418, 0.00000075, 0.00000542, -0.00000380, -0.00000191, -0.00000597, -0.00000079, -0.00000309, 0.00000590, -0.00000999, 0.00000581, -0.00001431, 0.00000653, -0.00003009, 0.00001116, -0.00001010, 0.00001361, 0.00000445, 0.00000361, 0.00000403, -0.00000062, -0.00000806, 0.00001014, -0.00000433, -0.00000530, -0.00000338, 0.00000061, -0.00000266, 0.00000249, -0.00002296, 0.00001382, -0.00001219, -0.00000834, -0.00002327, -0.00001981, -0.00003336, -0.00008178, -0.00023716, 0.00027525, 0.00012283, 0.00009764, 0.00005699, 0.00009488, 0.00005553, 0.00030782, -0.00008458, -0.00020250, -0.00000983, -0.00005623, -0.00000124, -0.00003740, 0.00000137, -0.00002295, 0.00001112, -0.00002354, -0.00000117, -0.00001609, 0.00000533, -0.00001215, 0.00000325, -0.00001116, 0.00001062, -0.00001043, 0.00000015, -0.00000748, 0.00000151, -0.00001289, 0.00000130, -0.00000388, 0.00000398, -0.00000720, 0.00000594, -0.00001863, 0.00001304, -0.00002062, 0.00001039, -0.00000058, -0.00000177, 0.00000449, -0.00000708, 0.00000613, -0.00000366, 0.00000854, -0.00000432, -0.00000001, 0.00000117, -0.00000281, -0.00000083, 0.00000998, -0.00000355, 0.00000041, -0.00000230, -0.00000378, 0.00000264, -0.00000060, -0.00000095, 0.00000303, -0.00000347, 0.00000019, 0.00000089, 0.00000774, -0.00000446, },
		{ -0.00000257, -0.00000028, 0.00001233, -0.00001080, 0.00000254, 0.00000036, -0.00000100, 0.00000002, 0.00000178, 0.00000013, -0.00001139, 0.00001076, 0.00000170, -0.00000713, -0.00000850, 0.00000617, -0.00000046, -0.00000074, -0.00000389, 0.00000074, -0.00000035, -0.00000148, -0.00000063, -0.00000480, -0.00000170, -0.00000083, -0.00000566, 0.00000481, 0.00000154, -0.00000198, -0.00000384, -0.00000364, -0.00000460, 0.00000368, -0.00000098, -0.00000395, -0.00000048, -0.00000840, 0.00000109, -0.00002788, 0.00001433, -0.00001205, 0.00001838, -0.00000159, 0.00000647, 0.00000203, -0.00000008, -0.00000870, 0.00001030, -0.00000612, -0.00000475, -0.00000258, -0.00000155, 0.00000085, -0.00000008, -0.00002308, 0.00001285, -0.00001053, -0.00001091, -0.00002277, -0.00002302, -0.00003114, -0.00008402, -0.00024959, 0.00027860, 0.00014005, 0.00009183, 0.00008023, 0.00008091, 0.00011028, 0.00029893, -0.00020080, -0.00013073, -0.00003331, -0.00003515, -0.00001544, -0.00002392, -0.00000895, -0.00001444, 0.00000574, -0.00001815, -0.00000818, -0.00000894, 0.00000023, -0.00000725, -0.00000192, -0.00000717, 0.00000932, -0.00000912, -0.00000281, -0.00000411, -0.00000239, -0.00000982, -0.00000041, -0.00000176, 0.00000137, -0.00000241, -0.00000138, -0.00001306, 0.00001101, -0.00002247, 0.00001796, -0.00000996, 0.00000703, -0.00000235, -0.00000273, 0.00000330, -0.00000104, 0.00000747, -0.00000600, 0.00000128, 0.00000101, -0.00000362, 0.00000193, 0.00000864, -0.00000519, 0.00000222, -0.00000494, -0.00000170, 0.00000262, -0.00000135, -0.00000008, 0.00000220, -0.00000361, 0.00000079, 0.00000210, 0.00000687, -0.00000649, },
		{ -0.00000238, 0.00000306, 0.00000956, -0.00001293, 0.00000684, -0.00000370, 0.00000221, -0.00000240, 0.00000484, -0.00000537, -0.00000687, 0.00001189, -0.00000146, -0.00000855, -0.00000514, 0.00000653, -0.00000246, 0.00000016, -0.00000498, 0.00000303, -0.00000276, 0.00000058, -0.00000210, -0.00000429, -0.00000202, -0.00000055, -0.00000596, 0.00000787, -0.00000238, 0.00000015, -0.00000560, -0.00000204, -0.00000716, 0.00000964, -0.00000829, 0.00000344, -0.00000895, 0.00000085, -0.00000932, -0.00001878, 0.00001067, -0.00000831, 0.00001794, -0.00000415, 0.00000734, 0.00000160, -0.00000102, -0.00000759, 0.00000837, -0.00000619, -0.00000497, -0.00000130, -0.00000430, 0.00000569, -0.00000515, -0.00002042, 0.00000948, -0.00000668, -0.00001579, -0.00002004, -0.00002836, -0.00002625, -0.00008978, -0.00025866, 0.00027783, 0.00016286, 0.00008156, 0.00011086, 0.00005809, 0.00019476, 0.00022523, -0.00027125, -0.00006278, -0.00006239, -0.00001021, -0.00003332, -0.00000698, -0.00002293, -0.00000221, -0.00000289, -0.00001064, -0.00001683, 0.00000033, -0.00000726, -0.00000014, -0.00000950, -0.00000017, 0.00000519, -0.00000603, -0.00000697, 0.00000045, -0.00000784, -0.00000469, -0.00000396, 0.00000168, -0.00000211, 0.00000318, -0.00001091, -0.00000283, 0.00000235, -0.00001655, 0.00001836, -0.00001444, 0.00001331, -0.00000861, 0.00000221, -0.00000056, 0.00000308, 0.00000426, -0.00000591, 0.00000187, 0.00000094, -0.00000425, 0.00000523, 0.00000542, -0.00000473, 0.00000241, -0.00000656, 0.00000035, 0.00000221, -0.00000198, 0.00000110, 0.00000060, -0.00000274, 0.00000072, 0.00000407, 0.00000428, -0.00000680, },
		{ -0.00000274, 0.00000738, 0.00000373, -0.00001106, 0.00000801, -0.00000576, 0.00000415, -0.00000371, 0.00000663, -0.00001052, -0.00000046, 0.00001041, -0.00000383, -0.00000962, -0.00000076, 0.00000473, -0.00000279, -0.00000026, -0.00000449, 0.00000384, -0.00000415, 0.00000203, -0.00000327, -0.00000413, -0.00000167, -0.00000139, -0.00000426, 0.00000857, -0.00000478, 0.00000119, -0.00000592, -0.00000257, -0.00000651, 0.00001200, -0.00001261, 0.00000841, -0.00001553, 0.00000907, -0.00002031, -0.00000647, 0.00000248, -0.00000011, 0.00001280, -0.00000313, 0.00000612, 0.00000247, -0.00000283, -0.00000556, 0.00000515, -0.00000514, -0.00000528, -0.00000070, -0.00000588, 0.00000957, -0.00001026, -0.00001722, 0.00000588, -0.00000300, -0.00002043, -0.00001790, -0.00003251, -0.00002262, -0.00009480, -0.00026894, 0.00027794, 0.00018551, 0.00007452, 0.00013862, 0.00004229, 0.00028912, 0.00009226, -0.00027879, -0.00002180, -0.00007997, 0.00000615, -0.00004475, 0.00000483, -0.00003312, 0.00000743, -0.00000951, -0.00000564, -0.00002256, 0.00000726, -0.00001301, 0.00000531, -0.00001576, 0.00000649, 0.00000095, -0.00000345, -0.00000999, 0.00000366, -0.00001221, 0.00000001, -0.00000716, 0.00000433, -0.00000413, 0.00000657, -0.00001885, 0.00000802, -0.00000953, -0.00000479, 0.00001164, -0.00001252, 0.00001475, -0.00001182, 0.00000552, -0.00000354, 0.00000709, -0.00000002, -0.00000444, 0.00000177, 0.00000083, -0.00000421, 0.00000812, 0.00000121, -0.00000254, 0.00000084, -0.00000671, 0.00000183, 0.00000161, -0.00000230, 0.00000216, -0.00000131, -0.00000122, 0.00000035, 0.00000622, 0.00000055, -0.00000566, },
		{ -0.00000279, 0.00001118, -0.00000357, -0.00000594, 0.00000584, -0.00000512, 0.00000400, -0.00000302, 0.00000592, -0.00001351, 0.00000606, 0.00000708, -0.00000548, -0.00000991, 0.00000371, 0.00000142, -0.00000160, -0.00000197, -0.00000232, 0.00000273, -0.00000387, 0.00000220, -0.00000357, -0.00000464, -0.00000052, -0.00000333, -0.00000049, 0.00000640, -0.00000473, 0.00000032, -0.00000424, -0.00000568, -0.00000189, 0.00000922, -0.00001162, 0.00000802, -0.00001661, 0.00001187, -0.00002674, 0.00000384, -0.00000587, 0.00000911, 0.00000548, -0.00000012, 0.00000401, 0.00000343, -0.00000414, -0.00000420, 0.00000220, -0.00000420, -0.00000462, -0.00000203, -0.00000456, 0.00001007, -0.00001263, -0.00001623, 0.00000469, -0.00000223, -0.00002189, -0.00001947, -0.00003190, -0.00002451, -0.00009427, -0.00028570, 0.00028465, 0.00020160, 0.00007917, 0.00015196, 0.00005353, 0.00036141, -0.00007390, -0.00022642, -0.00001815, -0.00007578, 0.00000566, -0.00004257, 0.00000506, -0.00003366, 0.00000927, -0.00000977, -0.00000686, -0.00002180, 0.00000819, -0.00001350, 0.00000570, -0.00001728, 0.00000953, -0.00000069, -0.00000362, -0.00000977, 0.00000329, -0.00001307, 0.00000187, -0.00000787, 0.00000428, -0.00000272, 0.00000517, -0.00002144, 0.00001471, -0.00001953, 0.00000830, 0.00000077, -0.00000544, 0.00001124, -0.00001103, 0.00000591, -0.00000423, 0.00000946, -0.00000403, -0.00000231, 0.00000124, 0.00000058, -0.00000312, 0.00000964, -0.00000275, 0.00000041, -0.00000194, -0.00000543, 0.00000252, 0.00000094, -0.00000215, 0.00000263, -0.00000289, 0.00000034, 0.00000025, 0.00000767, -0.00000327, -0.00000386, },
		{ -0.00000153, 0.00001273, -0.00001003, 0.00000046, 0.00000150, -0.00000230, 0.00000195, -0.00000042, 0.00000238, -0.00001313, 0.00001097, 0.00000305, -0.00000673, -0.00000893, 0.00000716, -0.00000227, 0.00000027, -0.00000427, 0.00000090, 0.00000010, -0.00000203, 0.00000100, -0.00000295, -0.00000568, 0.00000104, -0.00000564, 0.00000444, 0.00000194, -0.00000247, -0.00000221, -0.00000111, -0.00001041, 0.00000560, 0.00000199, -0.00000534, 0.00000168, -0.00001089, 0.00000699, -0.00002508, 0.00000766, -0.00000991, 0.00001537, -0.00000071, 0.00000253, 0.00000265, 0.00000310, -0.00000356, -0.00000500, 0.00000100, -0.00000449, -0.00000230, -0.00000581, 0.00000036, 0.00000597, -0.00001054, -0.00001921, 0.00000755, -0.00000600, -0.00001855, -0.00002634, -0.00002478, -0.00003419, -0.00008549, -0.00031195, 0.00030105, 0.00020797, 0.00009985, 0.00014463, 0.00010693, 0.00037733, -0.00023316, -0.00013663, -0.00004594, -0.00005086, -0.00001158, -0.00002634, -0.00000711, -0.00002334, 0.00000200, -0.00000267, -0.00001491, -0.00001392, 0.00000226, -0.00000774, -0.00000001, -0.00001275, 0.00000746, 0.00000151, -0.00000730, -0.00000576, -0.00000128, -0.00000949, -0.00000015, -0.00000523, 0.00000094, 0.00000253, -0.00000177, -0.00001679, 0.00001386, -0.00002300, 0.00001753, -0.00000962, 0.00000347, 0.00000474, -0.00000708, 0.00000352, -0.00000228, 0.00000933, -0.00000670, -0.00000030, 0.00000057, 0.00000021, -0.00000091, 0.00000915, -0.00000529, 0.00000291, -0.00000499, -0.00000322, 0.00000246, 0.00000030, -0.00000144, 0.00000218, -0.00000356, 0.00000136, 0.00000096, 0.00000761, -0.00000608, -0.00000236, },
		{ 0.00000150, 0.00001083, -0.00001345, 0.00000575, -0.00000304, 0.00000121, -0.00000081, 0.00000301, -0.00000315, -0.00000925, 0.00001313, -0.00000066, -0.00000785, -0.00000642, 0.00000868, -0.00000511, 0.00000174, -0.00000604, 0.00000397, -0.00000300, 0.00000058, -0.00000100, -0.00000187, -0.00000670, 0.00000219, -0.00000710, 0.00000891, -0.00000321, 0.00000071, -0.00000516, 0.00000191, -0.00001459, 0.00001322, -0.00000684, 0.00000358, -0.00000825, -0.00000033, -0.00000437, -0.00001520, 0.00000337, -0.00000724, 0.00001603, -0.00000329, 0.00000304, 0.00000311, 0.00000084, -0.00000069, -0.00000845, 0.00000209, -0.00000619, 0.00000129, -0.00001117, 0.00000782, -0.00000192, -0.00000444, -0.00002579, 0.00001393, -0.00001360, -0.00001130, -0.00003725, -0.00001265, -0.00005022, -0.00006980, -0.00034622, 0.00032514, 0.00020777, 0.00013286, 0.00012161, 0.00020313, 0.00031348, -0.00034425, -0.00004309, -0.00008645, -0.00001676, -0.00003694, -0.00000288, -0.00002624, -0.00000646, -0.00001094, 0.00000868, -0.00002661, -0.00000197, -0.00000791, 0.00000195, -0.00000976, -0.00000377, 0.00000138, 0.00000650, -0.00001317, 0.00000049, -0.00000853, -0.00000275, -0.00000506, -0.00000025, -0.00000438, 0.00000987, -0.00001235, -0.00000609, 0.00000517, -0.00001782, 0.00001926, -0.00001529, 0.00001032, -0.00000167, -0.00000213, -0.00000025, 0.00000151, 0.00000680, -0.00000752, 0.00000107, -0.00000009, -0.00000005, 0.00000200, 0.00000666, -0.00000577, 0.00000396, -0.00000730, -0.00000081, 0.00000188, -0.00000022, -0.00000033, 0.00000079, -0.00000311, 0.00000162, 0.00000262, 0.00000569, -0.00000718, -0.00000173, },
		{ 0.00000585, 0.00000545, -0.00001268, 0.00000802, -0.00000585, 0.00000371, -0.00000274, 0.00000558, -0.00000895, -0.00000287, 0.00001234, -0.00000346, -0.00000885, -0.00000258, 0.00000790, -0.00000629, 0.00000193, -0.00000627, 0.00000567, -0.00000526, 0.00000280, -0.00000289, -0.00000103, -0.00000701, 0.00000213, -0.00000654, 0.00001117, -0.00000710, 0.00000309, -0.00000694, 0.00000301, -0.00001580, 0.00001766, -0.00001337, 0.00001098, -0.00001733, 0.00001047, -0.00001777, -0.00000065, -0.00000701, 0.00000136, 0.00001107, -0.00000177, 0.00000113, 0.00000517, -0.00000266, 0.00000355, -0.00001363, 0.00000469, -0.00000837, 0.00000466, -0.00001598, 0.00001512, -0.00001079, 0.00000311, -0.00003344, 0.00002105, -0.00002199, -0.00000350, -0.00004826, -0.00000014, -0.00006757, -0.00005261, -0.00038258, 0.00034973, 0.00021049, 0.00016610, 0.00010085, 0.00032253, 0.00016885, -0.00037935, 0.00002210, -0.00011702, 0.00001029, -0.00005741, 0.00001694, -0.00004304, 0.00000916, -0.00002299, 0.00001847, -0.00003635, 0.00000864, -0.00001732, 0.00001096, -0.00001919, 0.00000578, -0.00000553, 0.00001149, -0.00001838, 0.00000593, -0.00001530, 0.00000422, -0.00001025, 0.00000454, -0.00000890, 0.00001589, -0.00002251, 0.00000658, -0.00000817, -0.00000549, 0.00001292, -0.00001415, 0.00001235, -0.00000527, 0.00000150, -0.00000348, 0.00000560, 0.00000278, -0.00000666, 0.00000163, -0.00000068, 0.00000017, 0.00000484, 0.00000281, -0.00000426, 0.00000317, -0.00000815, 0.00000114, 0.00000106, -0.00000050, 0.00000083, -0.00000115, -0.00000177, 0.00000129, 0.00000482, 0.00000222, -0.00000655, -0.00000187, },
		{ 0.00001018, -0.00000209, -0.00000806, 0.00000666, -0.00000585, 0.00000405, -0.00000264, 0.00000579, -0.00001300, 0.00000421, 0.00000927, -0.00000530, -0.00000939, 0.00000182, 0.00000521, -0.00000569, 0.00000058, -0.00000458, 0.00000526, -0.00000575, 0.00000368, -0.00000388, -0.00000098, -0.00000626, 0.00000058, -0.00000347, 0.00001019, -0.00000831, 0.00000334, -0.00000645, 0.00000118, -0.00001260, 0.00001662, -0.00001443, 0.00001301, -0.00002109, 0.00001629, -0.00002734, 0.00001276, -0.00001865, 0.00001229, 0.00000298, 0.00000228, -0.00000195, 0.00000741, -0.00000573, 0.00000719, -0.00001859, 0.00000713, -0.00000942, 0.00000589, -0.00001771, 0.00001902, -0.00001702, 0.00000851, -0.00003864, 0.00002521, -0.00002719, 0.00000056, -0.00005444, 0.00000696, -0.00007974, -0.00004103, -0.00041319, 0.00036551, 0.00022839, 0.00018347, 0.00010861, 0.00042718, -0.00003058, -0.00033478, 0.00004077, -0.00012134, 0.00001745, -0.00006213, 0.00002360, -0.00004901, 0.00001607, -0.00002785, 0.00002118, -0.00003886, 0.00001272, -0.00002104, 0.00001458, -0.00002370, 0.00001162, -0.00000965, 0.00001340, -0.00001995, 0.00000743, -0.00001828, 0.00000818, -0.00001279, 0.00000642, -0.00000980, 0.00001714, -0.00002768, 0.00001582, -0.00002066, 0.00000938, 0.00000132, -0.00000708, 0.00000908, -0.00000481, 0.00000230, -0.00000461, 0.00000837, -0.00000141, -0.00000475, 0.00000149, -0.00000123, 0.00000123, 0.00000668, -0.00000128, -0.00000153, 0.00000080, -0.00000738, 0.00000223, 0.00000022, -0.00000036, 0.00000154, -0.00000300, -0.00000011, 0.00000097, 0.00000673, -0.00000184, -0.00000485, -0.00000204, },
		{ 0.00001269, -0.00000950, -0.00000137, 0.00000253, -0.00000328, 0.00000215, -0.00000028, 0.00000294, -0.00001373, 0.00001004, 0.00000511, -0.00000653, -0.00000891, 0.00000566, 0.00000164, -0.00000399, -0.00000175, -0.00000140, 0.00000292, -0.00000431, 0.00000292, -0.00000372, -0.00000173, -0.00000470, -0.00000192, 0.00000147, 0.00000621, -0.00000667, 0.00000132, -0.00000385, -0.00000307, -0.00000554, 0.00001005, -0.00000921, 0.00000818, -0.00001722, 0.00001380, -0.00002849, 0.00001953, -0.00002621, 0.00002089, -0.00000453, 0.00000626, -0.00000429, 0.00000814, -0.00000661, 0.00000831, -0.00002136, 0.00000781, -0.00000813, 0.00000383, -0.00001489, 0.00001742, -0.00001799, 0.00000910, -0.00003878, 0.00002376, -0.00002647, -0.00000193, -0.00005266, 0.00000480, -0.00008223, -0.00004005, -0.00043267, 0.00036617, 0.00027012, 0.00017316, 0.00016892, 0.00047116, -0.00023713, -0.00023308, 0.00001519, -0.00009683, 0.00000174, -0.00004796, 0.00001386, -0.00004084, 0.00001111, -0.00002286, 0.00001473, -0.00003212, 0.00000805, -0.00001682, 0.00001057, -0.00002088, 0.00001129, -0.00000892, 0.00001069, -0.00001657, 0.00000358, -0.00001577, 0.00000729, -0.00001105, 0.00000403, -0.00000581, 0.00001191, -0.00002498, 0.00001727, -0.00002681, 0.00002104, -0.00001068, 0.00000259, 0.00000232, -0.00000093, 0.00000021, -0.00000308, 0.00000878, -0.00000458, -0.00000257, 0.00000092, -0.00000173, 0.00000324, 0.00000680, -0.00000440, 0.00000126, -0.00000230, -0.00000536, 0.00000240, -0.00000052, 0.00000027, 0.00000141, -0.00000411, 0.00000120, 0.00000129, 0.00000739, -0.00000531, -0.00000306, -0.00000120, },
		{ 0.00001189, -0.00001434, 0.00000489, -0.00000244, 0.00000051, -0.00000095, 0.00000345, -0.00000245, -0.00001065, 0.00001316, 0.00000111, -0.00000761, -0.00000694, 0.00000785, -0.00000148, -0.00000234, -0.00000392, 0.00000206, -0.00000035, -0.00000168, 0.00000098, -0.00000281, -0.00000268, -0.00000324, -0.00000407, 0.00000663, 0.00000077, -0.00000338, -0.00000184, -0.00000062, -0.00000765, 0.00000279, 0.00000056, 0.00000006, -0.00000167, -0.00000701, 0.00000337, -0.00002017, 0.00001709, -0.00002630, 0.00002373, -0.00000846, 0.00000803, -0.00000460, 0.00000652, -0.00000464, 0.00000616, -0.00002119, 0.00000634, -0.00000469, -0.00000093, -0.00000825, 0.00001072, -0.00001368, 0.00000477, -0.00003396, 0.00001695, -0.00002025, -0.00001024, -0.00004384, -0.00000588, -0.00007559, -0.00004908, -0.00044209, 0.00035314, 0.00033465, 0.00013646, 0.00029039, 0.00041706, -0.00039588, -0.00011510, -0.00003381, -0.00005552, -0.00002820, -0.00002147, -0.00000718, -0.00002240, -0.00000270, -0.00001079, 0.00000194, -0.00001880, -0.00000314, -0.00000655, 0.00000058, -0.00001189, 0.00000549, -0.00000407, 0.00000445, -0.00000956, -0.00000434, -0.00000872, 0.00000223, -0.00000576, -0.00000162, 0.00000171, 0.00000150, -0.00001483, 0.00000975, -0.00002363, 0.00002504, -0.00001828, 0.00001073, -0.00000485, 0.00000432, -0.00000355, 0.00000053, 0.00000667, -0.00000598, -0.00000085, 0.00000022, -0.00000208, 0.00000599, 0.00000496, -0.00000566, 0.00000294, -0.00000503, -0.00000282, 0.00000189, -0.00000111, 0.00000135, 0.00000027, -0.00000407, 0.00000172, 0.00000263, 0.00000618, -0.00000723, -0.00000195, 0.00000135, },
		{ 0.00000731, -0.00001498, 0.00000846, -0.00000609, 0.00000367, -0.00000363, 0.00000688, -0.00000878, -0.00000447, 0.00001305, -0.00000190, -0.00000876, -0.00000342, 0.00000769, -0.00000305, -0.00000184, -0.00000469, 0.00000434, -0.00000305, 0.00000084, -0.00000107, -0.00000203, -0.00000296, -0.00000296, -0.00000430, 0.00000987, -0.00000383, -0.00000046, -0.00000421, 0.00000109, -0.00000972, 0.00000861, -0.00000756, 0.00000888, -0.00001193, 0.00000489, -0.00001069, -0.00000564, 0.00000701, -0.00001912, 0.00002021, -0.00000794, 0.00000717, -0.00000310, 0.00000333, -0.00000089, 0.00000183, -0.00001907, 0.00000398, -0.00000097, -0.00000579, -0.00000090, 0.00000207, -0.00000701, -0.00000156, -0.00002738, 0.00000833, -0.00001250, -0.00001969, -0.00003339, -0.00001919, -0.00006614, -0.00006102, -0.00045007, 0.00033710, 0.00040897, 0.00009236, 0.00045626, 0.00025249, -0.00046353, -0.00002467, -0.00007663, -0.00001809, -0.00005610, 0.00000394, -0.00002824, -0.00000317, -0.00001739, 0.00000130, -0.00001042, -0.00000532, -0.00001496, 0.00000436, -0.00001040, -0.00000116, -0.00000209, 0.00000157, -0.00000189, -0.00000255, -0.00001265, -0.00000054, -0.00000398, 0.00000018, -0.00000739, 0.00000908, -0.00000988, -0.00000121, -0.00000387, -0.00001208, 0.00002022, -0.00001886, 0.00001424, -0.00000954, 0.00000861, -0.00000717, 0.00000487, 0.00000267, -0.00000544, -0.00000003, -0.00000034, -0.00000211, 0.00000894, 0.00000146, -0.00000477, 0.00000278, -0.00000644, -0.00000060, 0.00000108, -0.00000149, 0.00000263, -0.00000165, -0.00000292, 0.00000145, 0.00000482, 0.00000312, -0.00000725, -0.00000174, 0.00000549, },
		{ -0.00000008, -0.00001128, 0.00000825, -0.00000699, 0.00000476, -0.00000441, 0.00000824, -0.00001393, 0.00000308, 0.00001017, -0.00000367, -0.00000981, 0.00000116, 0.00000523, -0.00000262, -0.00000309, -0.00000331, 0.00000437, -0.00000390, 0.00000198, -0.00000217, -0.00000218, -0.00000187, -0.00000460, -0.00000159, 0.00000956, -0.00000558, 0.00000018, -0.00000410, -0.00000045, -0.00000704, 0.00000873, -0.00001019, 0.00001251, -0.00001721, 0.00001243, -0.00002181, 0.00000878, -0.00000557, -0.00000836, 0.00001278, -0.00000457, 0.00000511, -0.00000155, 0.00000073, 0.00000212, -0.00000212, -0.00001736, 0.00000311, 0.00000017, -0.00000721, 0.00000284, -0.00000379, -0.00000269, -0.00000512, -0.00002411, 0.00000337, -0.00000918, -0.00002351, -0.00002920, -0.00002629, -0.00006366, -0.00006485, -0.00046987, 0.00033452, 0.00047203, 0.00007438, 0.00062297, -0.00000026, -0.00042467, 0.00000814, -0.00008831, -0.00000382, -0.00006597, 0.00001452, -0.00003726, 0.00000638, -0.00002399, 0.00000549, -0.00001489, 0.00000110, -0.00002063, 0.00000955, -0.00001627, 0.00000572, -0.00000664, 0.00000376, -0.00000422, 0.00000020, -0.00001693, 0.00000448, -0.00000742, 0.00000306, -0.00000945, 0.00001184, -0.00001681, 0.00000983, -0.00001773, 0.00000317, 0.00000919, -0.00001301, 0.00001236, -0.00001036, 0.00001039, -0.00000909, 0.00000843, -0.00000213, -0.00000333, -0.00000029, -0.00000058, -0.00000165, 0.00001139, -0.00000285, -0.00000215, 0.00000067, -0.00000602, 0.00000067, 0.00000033, -0.00000163, 0.00000371, -0.00000382, -0.00000111, 0.00000083, 0.00000718, -0.00000107, -0.00000579, -0.00000188, 0.00001011, },
		{ -0.00000813, -0.00000472, 0.00000473, -0.00000499, 0.00000335, -0.00000271, 0.00000650, -0.00001605, 0.00000994, 0.00000570, -0.00000451, -0.00001024, 0.00000579, 0.00000129, -0.00000058, -0.00000581, 0.00000008, 0.00000195, -0.00000237, 0.00000109, -0.00000174, -0.00000355, 0.00000060, -0.00000795, 0.00000388, 0.00000537, -0.00000369, -0.00000223, -0.00000103, -0.00000539, 0.00000066, 0.00000217, -0.00000546, 0.00000828, -0.00001397, 0.00001098, -0.00002417, 0.00001646, -0.00001429, 0.00000060, 0.00000564, -0.00000137, 0.00000421, -0.00000219, 0.00000115, 0.00000172, -0.00000295, -0.00001849, 0.00000590, -0.00000351, -0.00000247, -0.00000050, -0.00000284, -0.00000487, -0.00000179, -0.00002847, 0.00000662, -0.00001510, -0.00001625, -0.00003772, -0.00001974, -0.00007649, -0.00005117, -0.00051289, 0.00035984, 0.00050474, 0.00011868, 0.00073009, -0.00028643, -0.00029849, -0.00002170, -0.00005977, -0.00002100, -0.00005015, 0.00000314, -0.00002754, 0.00000012, -0.00001724, -0.00000271, -0.00000721, -0.00000384, -0.00001599, 0.00000497, -0.00001295, 0.00000479, -0.00000476, -0.00000027, 0.00000001, -0.00000396, -0.00001426, 0.00000335, -0.00000534, 0.00000040, -0.00000536, 0.00000705, -0.00001518, 0.00001279, -0.00002544, 0.00001575, -0.00000290, -0.00000406, 0.00000682, -0.00000783, 0.00000949, -0.00000869, 0.00001014, -0.00000655, -0.00000029, -0.00000150, -0.00000035, -0.00000064, 0.00001265, -0.00000686, 0.00000122, -0.00000277, -0.00000391, 0.00000074, -0.00000010, -0.00000146, 0.00000419, -0.00000557, 0.00000069, 0.00000053, 0.00000871, -0.00000522, -0.00000374, -0.00000141, 0.00001344, },
		{ -0.00001425, 0.00000221, -0.00000032, -0.00000126, 0.00000027, 0.00000086, 0.00000180, -0.00001424, 0.00001436, 0.00000102, -0.00000499, -0.00000944, 0.00000926, -0.00000279, 0.00000196, -0.00000894, 0.00000440, -0.00000207, 0.00000099, -0.00000156, -0.00000006, -0.00000565, 0.00000361, -0.00001177, 0.00001061, -0.00000140, 0.00000094, -0.00000685, 0.00000380, -0.00001197, 0.00001115, -0.00000904, 0.00000509, -0.00000280, -0.00000254, -0.00000018, -0.00001551, 0.00001361, -0.00001462, 0.00000337, 0.00000248, -0.00000095, 0.00000618, -0.00000630, 0.00000572, -0.00000335, 0.00000062, -0.00002338, 0.00001271, -0.00001204, 0.00000848, -0.00001143, 0.00000593, -0.00001462, 0.00000939, -0.00004131, 0.00001886, -0.00003085, 0.00000266, -0.00005996, 0.00000195, -0.00010639, -0.00001807, -0.00058162, 0.00041636, 0.00050250, 0.00024677, 0.00071997, -0.00053392, -0.00013284, -0.00009292, -0.00000175, -0.00006253, -0.00001364, -0.00002664, -0.00000151, -0.00002023, 0.00000129, -0.00002156, 0.00001104, -0.00001893, -0.00000192, -0.00000872, -0.00000069, -0.00000407, 0.00000349, -0.00001005, 0.00001010, -0.00001439, -0.00000494, -0.00000394, 0.00000220, -0.00000744, 0.00000428, -0.00000488, -0.00000436, 0.00000544, -0.00002295, 0.00002037, -0.00001072, 0.00000359, 0.00000072, -0.00000391, 0.00000700, -0.00000638, 0.00000963, -0.00000968, 0.00000285, -0.00000334, 0.00000038, 0.00000084, 0.00001230, -0.00000955, 0.00000413, -0.00000646, -0.00000077, -0.00000022, -0.00000017, -0.00000098, 0.00000379, -0.00000635, 0.00000188, 0.00000113, 0.00000857, -0.00000817, -0.00000204, 0.00000056, 0.00001377, },
		{ -0.00001637, 0.00000696, -0.00000461, 0.00000232, -0.00000284, 0.00000468, -0.00000448, -0.00000887, 0.00001544, -0.00000275, -0.00000555, -0.00000707, 0.00001059, -0.00000568, 0.00000369, -0.00001104, 0.00000801, -0.00000607, 0.00000480, -0.00000483, 0.00000193, -0.00000743, 0.00000580, -0.00001419, 0.00001616, -0.00000822, 0.00000608, -0.00001154, 0.00000799, -0.00001704, 0.00002041, -0.00002039, 0.00001691, -0.00001616, 0.00001264, -0.00001713, 0.00000144, 0.00000114, -0.00000602, -0.00000132, 0.00000458, -0.00000396, 0.00001084, -0.00001306, 0.00001330, -0.00001192, 0.00000748, -0.00003057, 0.00002140, -0.00002260, 0.00002233, -0.00002664, 0.00001945, -0.00002885, 0.00002498, -0.00005883, 0.00003582, -0.00005141, 0.00002743, -0.00008970, 0.00003213, -0.00014595, 0.00002564, -0.00066561, 0.00049071, 0.00048452, 0.00044927, 0.00056074, -0.00067691, 0.00001316, -0.00016783, 0.00006024, -0.00010880, 0.00002772, -0.00006165, 0.00002987, -0.00004546, 0.00002340, -0.00004333, 0.00003258, -0.00003760, 0.00001558, -0.00002602, 0.00001571, -0.00001679, 0.00001439, -0.00002170, 0.00002195, -0.00002703, 0.00000732, -0.00001409, 0.00001202, -0.00001711, 0.00001565, -0.00001981, 0.00001194, -0.00000990, -0.00001060, 0.00001524, -0.00001114, 0.00000660, -0.00000299, -0.00000086, 0.00000456, -0.00000322, 0.00000719, -0.00001101, 0.00000534, -0.00000528, 0.00000156, 0.00000253, 0.00001030, -0.00001023, 0.00000547, -0.00000921, 0.00000241, -0.00000170, 0.00000006, -0.00000030, 0.00000253, -0.00000595, 0.00000221, 0.00000275, 0.00000642, -0.00000924, -0.00000120, 0.00000423, 0.00001023, },
		{ -0.00001383, 0.00000796, -0.00000634, 0.00000403, -0.00000431, 0.00000687, -0.00001022, -0.00000145, 0.00001336, -0.00000512, -0.00000627, -0.00000328, 0.00000940, -0.00000651, 0.00000364, -0.00001097, 0.00000943, -0.00000837, 0.00000742, -0.00000731, 0.00000305, -0.00000777, 0.00000593, -0.00001356, 0.00001820, -0.00001239, 0.00000917, -0.00001393, 0.00000907, -0.00001750, 0.00002429, -0.00002682, 0.00002430, -0.00002553, 0.00002471, -0.00003267, 0.00001991, -0.00001555, 0.00000774, -0.00001098, 0.00001021, -0.00000864, 0.00001592, -0.00001964, 0.00002064, -0.00002061, 0.00001436, -0.00003663, 0.00002796, -0.00003034, 0.00003340, -0.00004003, 0.00003159, -0.00004129, 0.00003824, -0.00007372, 0.00004950, -0.00006772, 0.00004767, -0.00011530, 0.00005793, -0.00018061, 0.00006273, -0.00074390, 0.00055496, 0.00049418, 0.00067836, 0.00026393, -0.00067959, 0.00009128, -0.00020900, 0.00009798, -0.00013684, 0.00005450, -0.00008514, 0.00005213, -0.00006310, 0.00003807, -0.00005768, 0.00004761, -0.00005073, 0.00002803, -0.00003894, 0.00002893, -0.00002698, 0.00002218, -0.00002961, 0.00002987, -0.00003605, 0.00001694, -0.00002198, 0.00001926, -0.00002361, 0.00002321, -0.00003140, 0.00002701, -0.00002697, 0.00000688, 0.00000282, -0.00000456, 0.00000411, -0.00000296, -0.00000006, 0.00000346, -0.00000039, 0.00000349, -0.00001040, 0.00000655, -0.00000675, 0.00000300, 0.00000411, 0.00000703, -0.00000882, 0.00000470, -0.00001010, 0.00000467, -0.00000312, 0.00000044, 0.00000041, 0.00000071, -0.00000460, 0.00000189, 0.00000495, 0.00000267, -0.00000847, -0.00000100, 0.00000884, 0.00000331, },
		{ -0.00000767, 0.00000519, -0.00000495, 0.00000309, -0.00000319, 0.00000604, -0.00001336, 0.00000591, 0.00000916, -0.00000623, -0.00000679, 0.00000116, 0.00000618, -0.00000531, 0.00000167, -0.00000842, 0.00000797, -0.00000795, 0.00000771, -0.00000800, 0.00000259, -0.00000621, 0.00000364, -0.00000935, 0.00001562, -0.00001231, 0.00000859, -0.00001268, 0.00000593, -0.00001201, 0.00002055, -0.00002510, 0.00002320, -0.00002590, 0.00002757, -0.00003956, 0.00003203, -0.00002902, 0.00002046, -0.00002078, 0.00001569, -0.00001185, 0.00001826, -0.00002256, 0.00002388, -0.00002544, 0.00001747, -0.00003792, 0.00002847, -0.00003067, 0.00003621, -0.00004548, 0.00003604, -0.00004545, 0.00004233, -0.00007867, 0.00005206, -0.00007093, 0.00005308, -0.00012504, 0.00006617, -0.00019532, 0.00007522, -0.00079421, 0.00057771, 0.00058744, 0.00085459, -0.00011129, -0.00055094, 0.00008267, -0.00019598, 0.00009420, -0.00013158, 0.00005331, -0.00008499, 0.00005443, -0.00006373, 0.00003703, -0.00005688, 0.00004865, -0.00005121, 0.00002866, -0.00004090, 0.00003276, -0.00002922, 0.00002216, -0.00002935, 0.00002933, -0.00003676, 0.00001920, -0.00002322, 0.00001984, -0.00002289, 0.00002239, -0.00003398, 0.00003398, -0.00003829, 0.00002237, -0.00001132, 0.00000550, -0.00000208, 0.00000016, -0.00000153, 0.00000414, 0.00000128, -0.00000063, -0.00000813, 0.00000619, -0.00000724, 0.00000448, 0.00000516, 0.00000326, -0.00000589, 0.00000195, -0.00000883, 0.00000542, -0.00000397, 0.00000085, 0.00000085, -0.00000111, -0.00000282, 0.00000147, 0.00000687, -0.00000169, -0.00000655, -0.00000064, 0.00001278, -0.00000510, },
		{ -0.00000023, 0.00000017, -0.00000130, 0.00000003, 0.00000023, 0.00000195, -0.00001261, 0.00001123, 0.00000431, -0.00000667, -0.00000646, 0.00000506, 0.00000212, -0.00000295, -0.00000137, -0.00000419, 0.00000418, -0.00000499, 0.00000562, -0.00000682, 0.00000076, -0.00000331, -0.00000014, -0.00000267, 0.00000920, -0.00000827, 0.00000461, -0.00000837, -0.00000032, -0.00000202, 0.00001032, -0.00001569, 0.00001339, -0.00001624, 0.00001899, -0.00003401, 0.00003246, -0.00003333, 0.00002655, -0.00002609, 0.00001764, -0.00001115, 0.00001576, -0.00001976, 0.00002079, -0.00002408, 0.00001483, -0.00003289, 0.00002157, -0.00002207, 0.00002863, -0.00004026, 0.00002991, -0.00003849, 0.00003432, -0.00007067, 0.00004043, -0.00005770, 0.00003950, -0.00011383, 0.00005096, -0.00018334, 0.00005488, -0.00080612, 0.00054140, 0.00081114, 0.00088819, -0.00047107, -0.00034383, 0.00000510, -0.00013473, 0.00005155, -0.00009398, 0.00002406, -0.00006036, 0.00003557, -0.00004632, 0.00001960, -0.00004026, 0.00003475, -0.00003792, 0.00001623, -0.00003037, 0.00002550, -0.00002211, 0.00001345, -0.00002033, 0.00001970, -0.00002815, 0.00001285, -0.00001667, 0.00001291, -0.00001430, 0.00001223, -0.00002548, 0.00002918, -0.00003854, 0.00002956, -0.00002116, 0.00001429, -0.00000867, 0.00000434, -0.00000413, 0.00000621, 0.00000140, -0.00000430, -0.00000487, 0.00000444, -0.00000660, 0.00000584, 0.00000527, -0.00000000, -0.00000254, -0.00000188, -0.00000586, 0.00000465, -0.00000407, 0.00000124, 0.00000074, -0.00000234, -0.00000128, 0.00000163, 0.00000752, -0.00000540, -0.00000448, 0.00000083, 0.00001423, -0.00001238, },
		{ 0.00000574, -0.00000477, 0.00000275, -0.00000359, 0.00000451, -0.00000432, -0.00000795, 0.00001327, 0.00000015, -0.00000710, -0.00000468, 0.00000718, -0.00000125, -0.00000090, -0.00000394, -0.00000001, -0.00000031, -0.00000089, 0.00000225, -0.00000478, -0.00000127, -0.00000063, -0.00000332, 0.00000387, 0.00000156, -0.00000257, -0.00000060, -0.00000351, -0.00000641, 0.00000838, -0.00000194, -0.00000297, -0.00000086, -0.00000054, 0.00000227, -0.00001793, 0.00002123, -0.00002705, 0.00002402, -0.00002516, 0.00001515, -0.00000659, 0.00000920, -0.00001234, 0.00001249, -0.00001765, 0.00000792, -0.00002376, 0.00001013, -0.00000780, 0.00001384, -0.00002728, 0.00001616, -0.00002374, 0.00001794, -0.00005393, 0.00001958, -0.00003367, 0.00001293, -0.00008801, 0.00001934, -0.00015296, 0.00001162, -0.00079233, 0.00045912, 0.00117817, 0.00070964, -0.00071428, -0.00013793, -0.00009438, -0.00005567, -0.00000725, -0.00004199, -0.00001862, -0.00002330, 0.00000564, -0.00001979, -0.00000584, -0.00001559, 0.00001289, -0.00001715, -0.00000355, -0.00001231, 0.00001135, -0.00000959, 0.00000020, -0.00000686, 0.00000522, -0.00001404, 0.00000124, -0.00000554, 0.00000184, -0.00000152, -0.00000344, -0.00000908, 0.00001428, -0.00002731, 0.00002598, -0.00002312, 0.00001823, -0.00001276, 0.00000750, -0.00000651, 0.00000892, -0.00000014, -0.00000661, -0.00000172, 0.00000207, -0.00000513, 0.00000716, 0.00000404, -0.00000173, -0.00000016, -0.00000540, -0.00000230, 0.00000296, -0.00000364, 0.00000171, -0.00000014, -0.00000253, -0.00000050, 0.00000278, 0.00000623, -0.00000745, -0.00000304, 0.00000393, 0.00001193, -0.00001613, },
		{ 0.00000817, -0.00000752, 0.00000528, -0.00000595, 0.00000771, -0.00001079, -0.00000057, 0.00001179, -0.00000246, -0.00000794, -0.00000116, 0.00000666, -0.00000260, -0.00000061, -0.00000439, 0.00000211, -0.00000332, 0.00000225, -0.00000053, -0.00000353, -0.00000182, -0.00000010, -0.00000345, 0.00000709, -0.00000378, 0.00000141, -0.00000379, -0.00000150, -0.00000820, 0.00001391, -0.00001007, 0.00000631, -0.00001228, 0.00001346, -0.00001468, 0.00000138, 0.00000413, -0.00001419, 0.00001565, -0.00002023, 0.00001075, -0.00000143, 0.00000254, -0.00000481, 0.00000371, -0.00001090, 0.00000183, -0.00001635, 0.00000097, 0.00000442, 0.00000002, -0.00001486, 0.00000347, -0.00001053, 0.00000334, -0.00003956, 0.00000205, -0.00001313, -0.00001070, -0.00006526, -0.00000854, -0.00012803, -0.00002521, -0.00079191, 0.00038290, 0.00164921, 0.00029886, -0.00076573, -0.00001187, -0.00015866, -0.00000072, -0.00004872, -0.00000344, -0.00005096, 0.00000578, -0.00001772, 0.00000021, -0.00002481, 0.00000357, -0.00000443, -0.00000046, -0.00001990, 0.00000344, -0.00000097, 0.00000034, -0.00000981, 0.00000325, -0.00000635, -0.00000184, -0.00000881, 0.00000368, -0.00000683, 0.00000845, -0.00001707, 0.00000756, -0.00000390, -0.00000965, 0.00001426, -0.00001776, 0.00001676, -0.00001340, 0.00000880, -0.00000799, 0.00001182, -0.00000343, -0.00000663, -0.00000004, 0.00000034, -0.00000364, 0.00000873, 0.00000114, -0.00000110, -0.00000000, -0.00000708, 0.00000040, 0.00000134, -0.00000317, 0.00000242, -0.00000186, -0.00000147, -0.00000064, 0.00000490, 0.00000293, -0.00000742, -0.00000243, 0.00000832, 0.00000585, -0.00001509, },
		{ 0.00000652, -0.00000706, 0.00000517, -0.00000581, 0.00000822, -0.00001531, 0.00000759, 0.00000752, -0.00000331, -0.00000921, 0.00000379, 0.00000337, -0.00000128, -0.00000287, -0.00000167, 0.00000073, -0.00000301, 0.00000249, -0.00000088, -0.00000460, 0.00000049, -0.00000312, 0.00000121, 0.00000456, -0.00000382, 0.00000068, -0.00000219, -0.00000503, -0.00000262, 0.00001045, -0.00000883, 0.00000600, -0.00001373, 0.00001748, -0.00002253, 0.00001405, -0.00000962, -0.00000259, 0.00000779, -0.00001658, 0.00000937, -0.00000091, 0.00000158, -0.00000350, 0.00000106, -0.00001041, 0.00000325, -0.00001793, 0.00000227, 0.00000527, -0.00000264, -0.00001363, 0.00000293, -0.00001077, 0.00000334, -0.00004153, 0.00000346, -0.00001394, -0.00001114, -0.00006856, -0.00000616, -0.00014046, -0.00001536, -0.00086122, 0.00039722, 0.00213006, -0.00029800, -0.00060341, -0.00001480, -0.00014422, -0.00000492, -0.00004340, -0.00000382, -0.00005055, 0.00000708, -0.00001717, -0.00000174, -0.00002305, 0.00000379, -0.00000462, 0.00000032, -0.00002161, 0.00000643, -0.00000209, -0.00000074, -0.00000858, 0.00000207, -0.00000701, 0.00000064, -0.00000998, 0.00000410, -0.00000630, 0.00000837, -0.00002046, 0.00001536, -0.00001599, 0.00000585, 0.00000110, -0.00000950, 0.00001254, -0.00001213, 0.00000912, -0.00000905, 0.00001501, -0.00000866, -0.00000357, -0.00000117, 0.00000067, -0.00000321, 0.00001099, -0.00000352, 0.00000222, -0.00000282, -0.00000581, 0.00000096, 0.00000089, -0.00000330, 0.00000355, -0.00000432, 0.00000071, -0.00000155, 0.00000750, -0.00000180, -0.00000556, -0.00000222, 0.00001277, -0.00000260, -0.00000973, },
		{ 0.00000188, -0.00000390, 0.00000262, -0.00000314, 0.00000547, -0.00001638, 0.00001451, 0.00000175, -0.00000275, -0.00001050, 0.00000935, -0.00000199, 0.00000239, -0.00000749, 0.00000410, -0.00000444, 0.00000128, -0.00000111, 0.00000210, -0.00000865, 0.00000598, -0.00000975, 0.00001068, -0.00000424, 0.00000252, -0.00000596, 0.00000514, -0.00001458, 0.00001067, -0.00000286, 0.00000352, -0.00000648, -0.00000165, 0.00000664, -0.00001475, 0.00001208, -0.00001149, -0.00000021, 0.00000731, -0.00001982, 0.00001580, -0.00000960, 0.00001106, -0.00001352, 0.00000991, -0.00002142, 0.00001721, -0.00003364, 0.00001970, -0.00001187, 0.00001334, -0.00003152, 0.00002283, -0.00003325, 0.00002740, -0.00006999, 0.00003517, -0.00004930, 0.00002701, -0.00011572, 0.00004749, -0.00021622, 0.00007554, -0.00105357, 0.00059665, 0.00248905, -0.00097234, -0.00026943, -0.00014800, -0.00003969, -0.00008017, 0.00002024, -0.00005405, -0.00000701, -0.00002903, 0.00001590, -0.00003321, 0.00000645, -0.00002176, 0.00001888, -0.00002112, -0.00000243, -0.00000938, 0.00001338, -0.00001744, 0.00000806, -0.00001451, 0.00000756, -0.00001104, 0.00000191, -0.00000808, 0.00000702, -0.00000562, -0.00000871, 0.00000794, -0.00001425, 0.00001083, -0.00000583, -0.00000443, 0.00000997, -0.00001197, 0.00001051, -0.00001096, 0.00001898, -0.00001583, 0.00000281, -0.00000582, 0.00000408, -0.00000472, 0.00001418, -0.00000963, 0.00000788, -0.00000847, -0.00000131, -0.00000131, 0.00000232, -0.00000450, 0.00000515, -0.00000723, 0.00000363, -0.00000277, 0.00000977, -0.00000691, -0.00000266, -0.00000162, 0.00001553, -0.00001096, -0.00000207, },
		{ -0.00000349, 0.00000023, -0.00000094, 0.00000080, 0.00000027, -0.00001366, 0.00001874, -0.00000412, -0.00000144, -0.00001122, 0.00001434, -0.00000807, 0.00000722, -0.00001322, 0.00001159, -0.00001218, 0.00000867, -0.00000797, 0.00000796, -0.00001499, 0.00001355, -0.00001839, 0.00002294, -0.00001740, 0.00001375, -0.00001715, 0.00001653, -0.00002776, 0.00002856, -0.00002275, 0.00002394, -0.00002839, 0.00002172, -0.00001775, 0.00000893, -0.00000676, 0.00000223, -0.00001116, 0.00001787, -0.00003262, 0.00003168, -0.00002851, 0.00003178, -0.00003577, 0.00003127, -0.00004462, 0.00004382, -0.00006314, 0.00005280, -0.00004687, 0.00004836, -0.00006909, 0.00006363, -0.00007839, 0.00007585, -0.00012506, 0.00009723, -0.00011981, 0.00010511, -0.00020888, 0.00015549, -0.00036027, 0.00025668, -0.00139337, 0.00105295, 0.00259146, -0.00157850, 0.00014042, -0.00036384, 0.00012817, -0.00020824, 0.00012881, -0.00014387, 0.00007174, -0.00009629, 0.00007608, -0.00008908, 0.00005903, -0.00006904, 0.00006257, -0.00006172, 0.00003519, -0.00004213, 0.00004362, -0.00004760, 0.00003768, -0.00004409, 0.00003538, -0.00003532, 0.00002537, -0.00003114, 0.00003112, -0.00003151, 0.00001693, -0.00001502, 0.00000366, 0.00000109, -0.00000154, -0.00000717, 0.00001275, -0.00001563, 0.00001486, -0.00001488, 0.00002396, -0.00002439, 0.00001179, -0.00001363, 0.00001062, -0.00000837, 0.00001808, -0.00001633, 0.00001471, -0.00001587, 0.00000565, -0.00000610, 0.00000567, -0.00000683, 0.00000710, -0.00001015, 0.00000676, -0.00000382, 0.00001089, -0.00001123, 0.00000040, 0.00000006, 0.00001512, -0.00001661, 0.00000508, },
		{ -0.00000718, 0.00000324, -0.00000356, 0.00000404, -0.00000555, -0.00000804, 0.00001972, -0.00000899, 0.00000000, -0.00001081, 0.00001758, -0.00001326, 0.00001153, -0.00001818, 0.00001857, -0.00002014, 0.00001691, -0.00001610, 0.00001489, -0.00002171, 0.00002094, -0.00002617, 0.00003442, -0.00003104, 0.00002617, -0.00002933, 0.00002817, -0.00004000, 0.00004533, -0.00004265, 0.00004539, -0.00005227, 0.00004858, -0.00004794, 0.00004155, -0.00003725, 0.00002812, -0.00003325, 0.00003758, -0.00005258, 0.00005372, -0.00005352, 0.00005905, -0.00006530, 0.00006010, -0.00007460, 0.00007684, -0.00009920, 0.00009347, -0.00009126, 0.00009383, -0.00011739, 0.00011566, -0.00013572, 0.00013724, -0.00019389, 0.00017522, -0.00020961, 0.00020587, -0.00032880, 0.00029581, -0.00054711, 0.00049846, -0.00185488, 0.00178238, 0.00234113, -0.00197394, 0.00050563, -0.00058257, 0.00030367, -0.00034610, 0.00024739, -0.00024408, 0.00016090, -0.00017340, 0.00014455, -0.00015210, 0.00011871, -0.00012343, 0.00011303, -0.00010907, 0.00007999, -0.00008176, 0.00007935, -0.00008215, 0.00007127, -0.00007780, 0.00006805, -0.00006446, 0.00005305, -0.00005776, 0.00005835, -0.00006128, 0.00004861, -0.00004677, 0.00003298, -0.00002106, 0.00001367, -0.00001835, 0.00002167, -0.00002367, 0.00002246, -0.00002079, 0.00002929, -0.00003282, 0.00002148, -0.00002288, 0.00001904, -0.00001340, 0.00002184, -0.00002223, 0.00002094, -0.00002306, 0.00001328, -0.00001214, 0.00001017, -0.00000984, 0.00000896, -0.00001253, 0.00000956, -0.00000430, 0.00001034, -0.00001388, 0.00000299, 0.00000291, 0.00001097, -0.00001775, 0.00000921, },
		{ -0.00000770, 0.00000363, -0.00000371, 0.00000478, -0.00000983, -0.00000121, 0.00001778, -0.00001223, 0.00000110, -0.00000898, 0.00001825, -0.00001616, 0.00001372, -0.00002048, 0.00002267, -0.00002557, 0.00002314, -0.00002274, 0.00002037, -0.00002635, 0.00002550, -0.00002997, 0.00004113, -0.00004060, 0.00003518, -0.00003801, 0.00003557, -0.00004622, 0.00005469, -0.00005506, 0.00005941, -0.00006862, 0.00006831, -0.00007231, 0.00007115, -0.00006819, 0.00005655, -0.00005825, 0.00005914, -0.00007249, 0.00007410, -0.00007602, 0.00008350, -0.00009227, 0.00008636, -0.00010092, 0.00010497, -0.00012931, 0.00012782, -0.00013006, 0.00013407, -0.00015990, 0.00016121, -0.00018608, 0.00019073, -0.00025330, 0.00024296, -0.00028931, 0.00029623, -0.00043767, 0.00042398, -0.00072265, 0.00073207, -0.00235516, 0.00272575, 0.00171542, -0.00206133, 0.00071885, -0.00072078, 0.00042284, -0.00044174, 0.00033219, -0.00031688, 0.00022752, -0.00023149, 0.00019562, -0.00019876, 0.00016366, -0.00016470, 0.00015144, -0.00014555, 0.00011563, -0.00011347, 0.00010700, -0.00010820, 0.00009622, -0.00010321, 0.00009358, -0.00008724, 0.00007433, -0.00007754, 0.00007804, -0.00008357, 0.00007445, -0.00007555, 0.00006313, -0.00004728, 0.00003388, -0.00003388, 0.00003375, -0.00003372, 0.00003131, -0.00002703, 0.00003322, -0.00003875, 0.00002904, -0.00003075, 0.00002697, -0.00001811, 0.00002406, -0.00002568, 0.00002456, -0.00002779, 0.00001928, -0.00001746, 0.00001438, -0.00001259, 0.00001012, -0.00001378, 0.00001154, -0.00000409, 0.00000803, -0.00001448, 0.00000498, 0.00000625, 0.00000384, -0.00001413, 0.00000913, },
		{ -0.00000504, 0.00000120, -0.00000104, 0.00000220, -0.00001100, 0.00000500, 0.00001377, -0.00001366, 0.00000165, -0.00000582, 0.00001611, -0.00001605, 0.00001285, -0.00001895, 0.00002223, -0.00002631, 0.00002485, -0.00002536, 0.00002207, -0.00002680, 0.00002522, -0.00002756, 0.00004018, -0.00004246, 0.00003698, -0.00003955, 0.00003534, -0.00004291, 0.00005233, -0.00005458, 0.00005952, -0.00006979, 0.00007173, -0.00007995, 0.00008530, -0.00008671, 0.00007529, -0.00007510, 0.00007261, -0.00008298, 0.00008333, -0.00008595, 0.00009437, -0.00010534, 0.00009851, -0.00011180, 0.00011582, -0.00013997, 0.00014090, -0.00014690, 0.00015168, -0.00017818, 0.00018047, -0.00020807, 0.00021306, -0.00027748, 0.00027113, -0.00032545, 0.00033785, -0.00049080, 0.00048608, -0.00081892, 0.00086524, -0.00276710, 0.00375047, 0.00078078, -0.00181843, 0.00071941, -0.00072097, 0.00043756, -0.00045399, 0.00034728, -0.00033038, 0.00024311, -0.00024525, 0.00020670, -0.00020835, 0.00017453, -0.00017476, 0.00016082, -0.00015504, 0.00012693, -0.00012337, 0.00011400, -0.00011403, 0.00010123, -0.00010915, 0.00010105, -0.00009330, 0.00007951, -0.00008118, 0.00008080, -0.00008818, 0.00008308, -0.00008910, 0.00008173, -0.00006639, 0.00004990, -0.00004652, 0.00004328, -0.00004113, 0.00003767, -0.00003073, 0.00003334, -0.00003960, 0.00003149, -0.00003410, 0.00003156, -0.00002039, 0.00002325, -0.00002525, 0.00002395, -0.00002812, 0.00002148, -0.00002002, 0.00001666, -0.00001399, 0.00000991, -0.00001340, 0.00001237, -0.00000328, 0.00000433, -0.00001313, 0.00000660, 0.00000881, -0.00000434, -0.00000715, 0.00000532, },
		{ -0.00000065, -0.00000280, 0.00000336, -0.00000306, -0.00000862, 0.00000928, 0.00000871, -0.00001338, 0.00000163, -0.00000173, 0.00001154, -0.00001299, 0.00000890, -0.00001353, 0.00001690, -0.00002153, 0.00002082, -0.00002255, 0.00001876, -0.00002216, 0.00001955, -0.00001861, 0.00003088, -0.00003534, 0.00002999, -0.00003257, 0.00002663, -0.00002973, 0.00003778, -0.00004015, 0.00004408, -0.00005330, 0.00005510, -0.00006525, 0.00007625, -0.00008348, 0.00007462, -0.00007457, 0.00006969, -0.00007655, 0.00007423, -0.00007599, 0.00008386, -0.00009625, 0.00008814, -0.00009893, 0.00010094, -0.00012223, 0.00012265, -0.00013053, 0.00013448, -0.00015926, 0.00015944, -0.00018652, 0.00018779, -0.00024817, 0.00023869, -0.00029339, 0.00030174, -0.00045337, 0.00043856, -0.00077786, 0.00081102, -0.00295120, 0.00467709, -0.00031808, -0.00130532, 0.00051106, -0.00057123, 0.00033354, -0.00036842, 0.00027876, -0.00027124, 0.00019505, -0.00020311, 0.00016748, -0.00017142, 0.00014231, -0.00014503, 0.00013298, -0.00012950, 0.00010601, -0.00010423, 0.00009403, -0.00009402, 0.00008103, -0.00009032, 0.00008507, -0.00007753, 0.00006398, -0.00006456, 0.00006261, -0.00007048, 0.00006850, -0.00007969, 0.00007966, -0.00006894, 0.00005316, -0.00004898, 0.00004426, -0.00004099, 0.00003755, -0.00002898, 0.00002766, -0.00003355, 0.00002675, -0.00003060, 0.00003052, -0.00001855, 0.00001842, -0.00002029, 0.00001844, -0.00002313, 0.00001861, -0.00001839, 0.00001574, -0.00001310, 0.00000782, -0.00001107, 0.00001183, -0.00000211, -0.00000012, -0.00001026, 0.00000819, 0.00000931, -0.00001112, 0.00000065, -0.00000028, },
		{ 0.00000329, -0.00000632, 0.00000741, -0.00000912, -0.00000361, 0.00001118, 0.00000339, -0.00001161, 0.00000114, 0.00000267, 0.00000545, -0.00000780, 0.00000282, -0.00000537, 0.00000779, -0.00001211, 0.00001157, -0.00001457, 0.00001078, -0.00001310, 0.00000969, -0.00000489, 0.00001512, -0.00002076, 0.00001549, -0.00001847, 0.00001155, -0.00000976, 0.00001471, -0.00001566, 0.00001711, -0.00002319, 0.00002202, -0.00003061, 0.00004435, -0.00005680, 0.00005157, -0.00005337, 0.00004756, -0.00005124, 0.00004548, -0.00004508, 0.00005084, -0.00006359, 0.00005395, -0.00006144, 0.00005995, -0.00007596, 0.00007269, -0.00007994, 0.00008109, -0.00010160, 0.00009641, -0.00011951, 0.00011294, -0.00016318, 0.00014255, -0.00018851, 0.00018130, -0.00031607, 0.00026729, -0.00057585, 0.00052044, -0.00279731, 0.00532387, -0.00139330, -0.00064633, 0.00015791, -0.00030806, 0.00013544, -0.00020311, 0.00014052, -0.00015030, 0.00009192, -0.00011229, 0.00008453, -0.00009402, 0.00007238, -0.00008029, 0.00007221, -0.00007255, 0.00005575, -0.00005865, 0.00004991, -0.00005133, 0.00003891, -0.00004973, 0.00004822, -0.00004238, 0.00003040, -0.00003080, 0.00002701, -0.00003377, 0.00003290, -0.00004763, 0.00005498, -0.00005135, 0.00003952, -0.00003736, 0.00003331, -0.00003044, 0.00002862, -0.00002030, 0.00001568, -0.00002059, 0.00001469, -0.00001973, 0.00002302, -0.00001202, 0.00000968, -0.00001133, 0.00000871, -0.00001336, 0.00001080, -0.00001228, 0.00001115, -0.00000959, 0.00000377, -0.00000681, 0.00000990, -0.00000085, -0.00000463, -0.00000633, 0.00000982, 0.00000702, -0.00001447, 0.00000660, -0.00000518, },
		{ 0.00000485, -0.00000741, 0.00000893, -0.00001357, 0.00000226, 0.00001108, -0.00000171, -0.00000860, 0.00000028, 0.00000672, -0.00000103, -0.00000178, -0.00000389, 0.00000357, -0.00000282, -0.00000034, -0.00000082, -0.00000322, -0.00000012, -0.00000161, -0.00000182, 0.00001024, -0.00000325, -0.00000250, -0.00000291, -0.00000091, -0.00000566, 0.00001158, -0.00001032, 0.00001149, -0.00001313, 0.00001135, -0.00001763, 0.00001407, -0.00000162, -0.00001363, 0.00001151, -0.00001594, 0.00001065, -0.00001208, 0.00000285, 0.00000035, 0.00000207, -0.00001420, 0.00000299, -0.00000710, 0.00000162, -0.00001098, 0.00000150, -0.00000589, 0.00000267, -0.00001708, 0.00000406, -0.00002067, 0.00000357, -0.00003933, 0.00000113, -0.00003069, -0.00000193, -0.00010248, -0.00000209, -0.00023903, 0.00000717, -0.00226277, 0.00555603, -0.00226277, 0.00000717, -0.00023903, -0.00000209, -0.00010248, -0.00000193, -0.00003069, 0.00000113, -0.00003933, 0.00000357, -0.00002067, 0.00000406, -0.00001708, 0.00000267, -0.00000589, 0.00000150, -0.00001098, 0.00000162, -0.00000710, 0.00000299, -0.00001420, 0.00000207, 0.00000035, 0.00000285, -0.00001208, 0.00001065, -0.00001594, 0.00001151, -0.00001363, -0.00000162, 0.00001407, -0.00001763, 0.00001135, -0.00001313, 0.00001149, -0.00001032, 0.00001158, -0.00000566, -0.00000091, -0.00000291, -0.00000250, -0.00000325, 0.00001024, -0.00000182, -0.00000161, -0.00000012, -0.00000322, -0.00000082, -0.00000034, -0.00000282, 0.00000357, -0.00000389, -0.00000178, -0.00000103, 0.00000672, 0.00000028, -0.00000860, -0.00000171, 0.00001108, 0.00000226, -0.00001357, 0.00000893, -0.00000741, },
		{ 0.00000329, -0.00000518, 0.00000660, -0.00001447, 0.00000702, 0.00000982, -0.00000633, -0.00000463, -0.00000085, 0.00000990, -0.00000681, 0.00000377, -0.00000959, 0.00001115, -0.00001228, 0.00001080, -0.00001336, 0.00000871, -0.00001133, 0.00000968, -0.00001202, 0.00002302, -0.00001973, 0.00001469, -0.00002059, 0.00001568, -0.00002030, 0.00002862, -0.00003044, 0.00003331, -0.00003736, 0.00003952, -0.00005135, 0.00005498, -0.00004763, 0.00003290, -0.00003377, 0.00002701, -0.00003080, 0.00003040, -0.00004238, 0.00004822, -0.00004973, 0.00003891, -0.00005133, 0.00004991, -0.00005865, 0.00005575, -0.00007255, 0.00007221, -0.00008029, 0.00007238, -0.00009402, 0.00008453, -0.00011229, 0.00009192, -0.00015030, 0.00014052, -0.00020311, 0.00013544, -0.00030806, 0.00015791, -0.00064633, -0.00139330, 0.00532387, -0.00279731, 0.00052044, -0.00057585, 0.00026729, -0.00031607, 0.00018130, -0.00018851, 0.00014255, -0.00016318, 0.00011294, -0.00011951, 0.00009641, -0.00010160, 0.00008109, -0.00007994, 0.00007269, -0.00007596, 0.00005995, -0.00006144, 0.00005395, -0.00006359, 0.00005084, -0.00004508, 0.00004548, -0.00005124, 0.00004756, -0.00005337, 0.00005157, -0.00005680, 0.00004435, -0.00003061, 0.00002202, -0.00002319, 0.00001711, -0.00001566, 0.00001471, -0.00000976, 0.00001155, -0.00001847, 0.00001549, -0.00002076, 0.00001512, -0.00000489, 0.00000969, -0.00001310, 0.00001078, -0.00001457, 0.00001157, -0.00001211, 0.00000779, -0.00000537, 0.00000282, -0.00000780, 0.00000545, 0.00000267, 0.00000114, -0.00001161, 0.00000339, 0.00001118, -0.00000361, -0.00000912, 0.00000741, -0.00000632, },
		{ -0.00000065, -0.00000028, 0.00000065, -0.00001112, 0.00000931, 0.00000819, -0.00001026, -0.00000012, -0.00000211, 0.00001183, -0.00001107, 0.00000782, -0.00001310, 0.00001574, -0.00001839, 0.00001861, -0.00002313, 0.00001844, -0.00002029, 0.00001842, -0.00001855, 0.00003052, -0.00003060, 0.00002675, -0.00003355, 0.00002766, -0.00002898, 0.00003755, -0.00004099, 0.00004426, -0.00004898, 0.00005316, -0.00006894, 0.00007966, -0.00007969, 0.00006850, -0.00007048, 0.00006261, -0.00006456, 0.00006398, -0.00007753, 0.00008507, -0.00009032, 0.00008103, -0.00009402, 0.00009403, -0.00010423, 0.00010601, -0.00012950, 0.00013298, -0.00014503, 0.00014231, -0.00017142, 0.00016748, -0.00020311, 0.00019505, -0.00027124, 0.00027876, -0.00036842, 0.00033354, -0.00057123, 0.00051106, -0.00130532, -0.00031808, 0.00467709, -0.00295120, 0.00081102, -0.00077786, 0.00043856, -0.00045337, 0.00030174, -0.00029339, 0.00023869, -0.00024817, 0.00018779, -0.00018652, 0.00015944, -0.00015926, 0.00013448, -0.00013053, 0.00012265, -0.00012223, 0.00010094, -0.00009893, 0.00008814, -0.00009625, 0.00008386, -0.00007599, 0.00007423, -0.00007655, 0.00006969, -0.00007457, 0.00007462, -0.00008348, 0.00007625, -0.00006525, 0.00005510, -0.00005330, 0.00004408, -0.00004015, 0.00003778, -0.00002973, 0.00002663, -0.00003257, 0.00002999, -0.00003534, 0.00003088, -0.00001861, 0.00001955, -0.00002216, 0.00001876, -0.00002255, 0.00002082, -0.00002153, 0.00001690, -0.00001353, 0.00000890, -0.00001299, 0.00001154, -0.00000173, 0.00000163, -0.00001338, 0.00000871, 0.00000928, -0.00000862, -0.00000306, 0.00000336, -0.00000280, },
		{ -0.00000504, 0.00000532, -0.00000715, -0.00000434, 0.00000881, 0.00000660, -0.00001313, 0.00000433, -0.00000328, 0.00001237, -0.00001340, 0.00000991, -0.00001399, 0.00001666, -0.00002002, 0.00002148, -0.00002812, 0.00002395, -0.00002525, 0.00002325, -0.00002039, 0.00003156, -0.00003410, 0.00003149, -0.00003960, 0.00003334, -0.00003073, 0.00003767, -0.00004113, 0.00004328, -0.00004652, 0.00004990, -0.00006639, 0.00008173, -0.00008910, 0.00008308, -0.00008818, 0.00008080, -0.00008118, 0.00007951, -0.00009330, 0.00010105, -0.00010915, 0.00010123, -0.00011403, 0.00011400, -0.00012337, 0.00012693, -0.00015504, 0.00016082, -0.00017476, 0.00017453, -0.00020835, 0.00020670, -0.00024525, 0.00024311, -0.00033038, 0.00034728, -0.00045399, 0.00043756, -0.00072097, 0.00071941, -0.00181843, 0.00078078, 0.00375047, -0.00276710, 0.00086524, -0.00081892, 0.00048608, -0.00049080, 0.00033785, -0.00032545, 0.00027113, -0.00027748, 0.00021306, -0.00020807, 0.00018047, -0.00017818, 0.00015168, -0.00014690, 0.00014090, -0.00013997, 0.00011582, -0.00011180, 0.00009851, -0.00010534, 0.00009437, -0.00008595, 0.00008333, -0.00008298, 0.00007261, -0.00007510, 0.00007529, -0.00008671, 0.00008530, -0.00007995, 0.00007173, -0.00006979, 0.00005952, -0.00005458, 0.00005233, -0.00004291, 0.00003534, -0.00003955, 0.00003698, -0.00004246, 0.00004018, -0.00002756, 0.00002522, -0.00002680, 0.00002207, -0.00002536, 0.00002485, -0.00002631, 0.00002223, -0.00001895, 0.00001285, -0.00001605, 0.00001611, -0.00000582, 0.00000165, -0.00001366, 0.00001377, 0.00000500, -0.00001100, 0.00000220, -0.00000104, 0.00000120, },
		{ -0.00000770, 0.00000913, -0.00001413, 0.00000384, 0.00000625, 0.00000498, -0.00001448, 0.00000803, -0.00000409, 0.00001154, -0.00001378, 0.00001012, -0.00001259, 0.00001438, -0.00001746, 0.00001928, -0.00002779, 0.00002456, -0.00002568, 0.00002406, -0.00001811, 0.00002697, -0.00003075, 0.00002904, -0.00003875, 0.00003322, -0.00002703, 0.00003131, -0.00003372, 0.00003375, -0.00003388, 0.00003388, -0.00004728, 0.00006313, -0.00007555, 0.00007445, -0.00008357, 0.00007804, -0.00007754, 0.00007433, -0.00008724, 0.00009358, -0.00010321, 0.00009622, -0.00010820, 0.00010700, -0.00011347, 0.00011563, -0.00014555, 0.00015144, -0.00016470, 0.00016366, -0.00019876, 0.00019562, -0.00023149, 0.00022752, -0.00031688, 0.00033219, -0.00044174, 0.00042284, -0.00072078, 0.00071885, -0.00206133, 0.00171542, 0.00272575, -0.00235516, 0.00073207, -0.00072265, 0.00042398, -0.00043767, 0.00029623, -0.00028931, 0.00024296, -0.00025330, 0.00019073, -0.00018608, 0.00016121, -0.00015990, 0.00013407, -0.00013006, 0.00012782, -0.00012931, 0.00010497, -0.00010092, 0.00008636, -0.00009227, 0.00008350, -0.00007602, 0.00007410, -0.00007249, 0.00005914, -0.00005825, 0.00005655, -0.00006819, 0.00007115, -0.00007231, 0.00006831, -0.00006862, 0.00005941, -0.00005506, 0.00005469, -0.00004622, 0.00003557, -0.00003801, 0.00003518, -0.00004060, 0.00004113, -0.00002997, 0.00002550, -0.00002635, 0.00002037, -0.00002274, 0.00002314, -0.00002557, 0.00002267, -0.00002048, 0.00001372, -0.00001616, 0.00001825, -0.00000898, 0.00000110, -0.00001223, 0.00001778, -0.00000121, -0.00000983, 0.00000478, -0.00000371, 0.00000363, },
		{ -0.00000718, 0.00000921, -0.00001775, 0.00001097, 0.00000291, 0.00000299, -0.00001388, 0.00001034, -0.00000430, 0.00000956, -0.00001253, 0.00000896, -0.00000984, 0.00001017, -0.00001214, 0.00001328, -0.00002306, 0.00002094, -0.00002223, 0.00002184, -0.00001340, 0.00001904, -0.00002288, 0.00002148, -0.00003282, 0.00002929, -0.00002079, 0.00002246, -0.00002367, 0.00002167, -0.00001835, 0.00001367, -0.00002106, 0.00003298, -0.00004677, 0.00004861, -0.00006128, 0.00005835, -0.00005776, 0.00005305, -0.00006446, 0.00006805, -0.00007780, 0.00007127, -0.00008215, 0.00007935, -0.00008176, 0.00007999, -0.00010907, 0.00011303, -0.00012343, 0.00011871, -0.00015210, 0.00014455, -0.00017340, 0.00016090, -0.00024408, 0.00024739, -0.00034610, 0.00030367, -0.00058257, 0.00050563, -0.00197394, 0.00234113, 0.00178238, -0.00185488, 0.00049846, -0.00054711, 0.00029581, -0.00032880, 0.00020587, -0.00020961, 0.00017522, -0.00019389, 0.00013724, -0.00013572, 0.00011566, -0.00011739, 0.00009383, -0.00009126, 0.00009347, -0.00009920, 0.00007684, -0.00007460, 0.00006010, -0.00006530, 0.00005905, -0.00005352, 0.00005372, -0.00005258, 0.00003758, -0.00003325, 0.00002812, -0.00003725, 0.00004155, -0.00004794, 0.00004858, -0.00005227, 0.00004539, -0.00004265, 0.00004533, -0.00004000, 0.00002817, -0.00002933, 0.00002617, -0.00003104, 0.00003442, -0.00002617, 0.00002094, -0.00002171, 0.00001489, -0.00001610, 0.00001691, -0.00002014, 0.00001857, -0.00001818, 0.00001153, -0.00001326, 0.00001758, -0.00001081, 0.00000000, -0.00000899, 0.00001972, -0.00000804, -0.00000555, 0.00000404, -0.00000356, 0.00000324, },
		{ -0.00000349, 0.00000508, -0.00001661, 0.00001512, 0.00000006, 0.00000040, -0.00001123, 0.00001089, -0.00000382, 0.00000676, -0.00001015, 0.00000710, -0.00000683, 0.00000567, -0.00000610, 0.00000565, -0.00001587, 0.00001471, -0.00001633, 0.00001808, -0.00000837, 0.00001062, -0.00001363, 0.00001179, -0.00002439, 0.00002396, -0.00001488, 0.00001486, -0.00001563, 0.00001275, -0.00000717, -0.00000154, 0.00000109, 0.00000366, -0.00001502, 0.00001693, -0.00003151, 0.00003112, -0.00003114, 0.00002537, -0.00003532, 0.00003538, -0.00004409, 0.00003768, -0.00004760, 0.00004362, -0.00004213, 0.00003519, -0.00006172, 0.00006257, -0.00006904, 0.00005903, -0.00008908, 0.00007608, -0.00009629, 0.00007174, -0.00014387, 0.00012881, -0.00020824, 0.00012817, -0.00036384, 0.00014042, -0.00157850, 0.00259146, 0.00105295, -0.00139337, 0.00025668, -0.00036027, 0.00015549, -0.00020888, 0.00010511, -0.00011981, 0.00009723, -0.00012506, 0.00007585, -0.00007839, 0.00006363, -0.00006909, 0.00004836, -0.00004687, 0.00005280, -0.00006314, 0.00004382, -0.00004462, 0.00003127, -0.00003577, 0.00003178, -0.00002851, 0.00003168, -0.00003262, 0.00001787, -0.00001116, 0.00000223, -0.00000676, 0.00000893, -0.00001775, 0.00002172, -0.00002839, 0.00002394, -0.00002275, 0.00002856, -0.00002776, 0.00001653, -0.00001715, 0.00001375, -0.00001740, 0.00002294, -0.00001839, 0.00001355, -0.00001499, 0.00000796, -0.00000797, 0.00000867, -0.00001218, 0.00001159, -0.00001322, 0.00000722, -0.00000807, 0.00001434, -0.00001122, -0.00000144, -0.00000412, 0.00001874, -0.00001366, 0.00000027, 0.00000080, -0.00000094, 0.00000023, },
		{ 0.00000188, -0.00000207, -0.00001096, 0.00001553, -0.00000162, -0.00000266, -0.00000691, 0.00000977, -0.00000277, 0.00000363, -0.00000723, 0.00000515, -0.00000450, 0.00000232, -0.00000131, -0.00000131, -0.00000847, 0.00000788, -0.00000963, 0.00001418, -0.00000472, 0.00000408, -0.00000582, 0.00000281, -0.00001583, 0.00001898, -0.00001096, 0.00001051, -0.00001197, 0.00000997, -0.00000443, -0.00000583, 0.00001083, -0.00001425, 0.00000794, -0.00000871, -0.00000562, 0.00000702, -0.00000808, 0.00000191, -0.00001104, 0.00000756, -0.00001451, 0.00000806, -0.00001744, 0.00001338, -0.00000938, -0.00000243, -0.00002112, 0.00001888, -0.00002176, 0.00000645, -0.00003321, 0.00001590, -0.00002903, -0.00000701, -0.00005405, 0.00002024, -0.00008017, -0.00003969, -0.00014800, -0.00026943, -0.00097234, 0.00248905, 0.00059665, -0.00105357, 0.00007554, -0.00021622, 0.00004749, -0.00011572, 0.00002701, -0.00004930, 0.00003517, -0.00006999, 0.00002740, -0.00003325, 0.00002283, -0.00003152, 0.00001334, -0.00001187, 0.00001970, -0.00003364, 0.00001721, -0.00002142, 0.00000991, -0.00001352, 0.00001106, -0.00000960, 0.00001580, -0.00001982, 0.00000731, -0.00000021, -0.00001149, 0.00001208, -0.00001475, 0.00000664, -0.00000165, -0.00000648, 0.00000352, -0.00000286, 0.00001067, -0.00001458, 0.00000514, -0.00000596, 0.00000252, -0.00000424, 0.00001068, -0.00000975, 0.00000598, -0.00000865, 0.00000210, -0.00000111, 0.00000128, -0.00000444, 0.00000410, -0.00000749, 0.00000239, -0.00000199, 0.00000935, -0.00001050, -0.00000275, 0.00000175, 0.00001451, -0.00001638, 0.00000547, -0.00000314, 0.00000262, -0.00000390, },
		{ 0.00000652, -0.00000973, -0.00000260, 0.00001277, -0.00000222, -0.00000556, -0.00000180, 0.00000750, -0.00000155, 0.00000071, -0.00000432, 0.00000355, -0.00000330, 0.00000089, 0.00000096, -0.00000581, -0.00000282, 0.00000222, -0.00000352, 0.00001099, -0.00000321, 0.00000067, -0.00000117, -0.00000357, -0.00000866, 0.00001501, -0.00000905, 0.00000912, -0.00001213, 0.00001254, -0.00000950, 0.00000110, 0.00000585, -0.00001599, 0.00001536, -0.00002046, 0.00000837, -0.00000630, 0.00000410, -0.00000998, 0.00000064, -0.00000701, 0.00000207, -0.00000858, -0.00000074, -0.00000209, 0.00000643, -0.00002161, 0.00000032, -0.00000462, 0.00000379, -0.00002305, -0.00000174, -0.00001717, 0.00000708, -0.00005055, -0.00000382, -0.00004340, -0.00000492, -0.00014422, -0.00001480, -0.00060341, -0.00029800, 0.00213006, 0.00039722, -0.00086122, -0.00001536, -0.00014046, -0.00000616, -0.00006856, -0.00001114, -0.00001394, 0.00000346, -0.00004153, 0.00000334, -0.00001077, 0.00000293, -0.00001363, -0.00000264, 0.00000527, 0.00000227, -0.00001793, 0.00000325, -0.00001041, 0.00000106, -0.00000350, 0.00000158, -0.00000091, 0.00000937, -0.00001658, 0.00000779, -0.00000259, -0.00000962, 0.00001405, -0.00002253, 0.00001748, -0.00001373, 0.00000600, -0.00000883, 0.00001045, -0.00000262, -0.00000503, -0.00000219, 0.00000068, -0.00000382, 0.00000456, 0.00000121, -0.00000312, 0.00000049, -0.00000460, -0.00000088, 0.00000249, -0.00000301, 0.00000073, -0.00000167, -0.00000287, -0.00000128, 0.00000337, 0.00000379, -0.00000921, -0.00000331, 0.00000752, 0.00000759, -0.00001531, 0.00000822, -0.00000581, 0.00000517, -0.00000706, },
		{ 0.00000817, -0.00001509, 0.00000585, 0.00000832, -0.00000243, -0.00000742, 0.00000293, 0.00000490, -0.00000064, -0.00000147, -0.00000186, 0.00000242, -0.00000317, 0.00000134, 0.00000040, -0.00000708, -0.00000000, -0.00000110, 0.00000114, 0.00000873, -0.00000364, 0.00000034, -0.00000004, -0.00000663, -0.00000343, 0.00001182, -0.00000799, 0.00000880, -0.00001340, 0.00001676, -0.00001776, 0.00001426, -0.00000965, -0.00000390, 0.00000756, -0.00001707, 0.00000845, -0.00000683, 0.00000368, -0.00000881, -0.00000184, -0.00000635, 0.00000325, -0.00000981, 0.00000034, -0.00000097, 0.00000344, -0.00001990, -0.00000046, -0.00000443, 0.00000357, -0.00002481, 0.00000021, -0.00001772, 0.00000578, -0.00005096, -0.00000344, -0.00004872, -0.00000072, -0.00015866, -0.00001187, -0.00076573, 0.00029886, 0.00164921, 0.00038290, -0.00079191, -0.00002521, -0.00012803, -0.00000854, -0.00006526, -0.00001070, -0.00001313, 0.00000205, -0.00003956, 0.00000334, -0.00001053, 0.00000347, -0.00001486, 0.00000002, 0.00000442, 0.00000097, -0.00001635, 0.00000183, -0.00001090, 0.00000371, -0.00000481, 0.00000254, -0.00000143, 0.00001075, -0.00002023, 0.00001565, -0.00001419, 0.00000413, 0.00000138, -0.00001468, 0.00001346, -0.00001228, 0.00000631, -0.00001007, 0.00001391, -0.00000820, -0.00000150, -0.00000379, 0.00000141, -0.00000378, 0.00000709, -0.00000345, -0.00000010, -0.00000182, -0.00000353, -0.00000053, 0.00000225, -0.00000332, 0.00000211, -0.00000439, -0.00000061, -0.00000260, 0.00000666, -0.00000116, -0.00000794, -0.00000246, 0.00001179, -0.00000057, -0.00001079, 0.00000771, -0.00000595, 0.00000528, -0.00000752, },
		{ 0.00000574, -0.00001613, 0.00001193, 0.00000393, -0.00000304, -0.00000745, 0.00000623, 0.00000278, -0.00000050, -0.00000253, -0.00000014, 0.00000171, -0.00000364, 0.00000296, -0.00000230, -0.00000540, -0.00000016, -0.00000173, 0.00000404, 0.00000716, -0.00000513, 0.00000207, -0.00000172, -0.00000661, -0.00000014, 0.00000892, -0.00000651, 0.00000750, -0.00001276, 0.00001823, -0.00002312, 0.00002598, -0.00002731, 0.00001428, -0.00000908, -0.00000344, -0.00000152, 0.00000184, -0.00000554, 0.00000124, -0.00001404, 0.00000522, -0.00000686, 0.00000020, -0.00000959, 0.00001135, -0.00001231, -0.00000355, -0.00001715, 0.00001289, -0.00001559, -0.00000584, -0.00001979, 0.00000564, -0.00002330, -0.00001862, -0.00004199, -0.00000725, -0.00005567, -0.00009438, -0.00013793, -0.00071428, 0.00070964, 0.00117817, 0.00045912, -0.00079233, 0.00001162, -0.00015296, 0.00001934, -0.00008801, 0.00001293, -0.00003367, 0.00001958, -0.00005393, 0.00001794, -0.00002374, 0.00001616, -0.00002728, 0.00001384, -0.00000780, 0.00001013, -0.00002376, 0.00000792, -0.00001765, 0.00001249, -0.00001234, 0.00000920, -0.00000659, 0.00001515, -0.00002516, 0.00002402, -0.00002705, 0.00002123, -0.00001793, 0.00000227, -0.00000054, -0.00000086, -0.00000297, -0.00000194, 0.00000838, -0.00000641, -0.00000351, -0.00000060, -0.00000257, 0.00000156, 0.00000387, -0.00000332, -0.00000063, -0.00000127, -0.00000478, 0.00000225, -0.00000089, -0.00000031, -0.00000001, -0.00000394, -0.00000090, -0.00000125, 0.00000718, -0.00000468, -0.00000710, 0.00000015, 0.00001327, -0.00000795, -0.00000432, 0.00000451, -0.00000359, 0.00000275, -0.00000477, },
		{ -0.00000023, -0.00001238, 0.00001423, 0.00000083, -0.00000448, -0.00000540, 0.00000752, 0.00000163, -0.00000128, -0.00000234, 0.00000074, 0.00000124, -0.00000407, 0.00000465, -0.00000586, -0.00000188, -0.00000254, -0.00000000, 0.00000527, 0.00000584, -0.00000660, 0.00000444, -0.00000487, -0.00000430, 0.00000140, 0.00000621, -0.00000413, 0.00000434, -0.00000867, 0.00001429, -0.00002116, 0.00002956, -0.00003854, 0.00002918, -0.00002548, 0.00001223, -0.00001430, 0.00001291, -0.00001667, 0.00001285, -0.00002815, 0.00001970, -0.00002033, 0.00001345, -0.00002211, 0.00002550, -0.00003037, 0.00001623, -0.00003792, 0.00003475, -0.00004026, 0.00001960, -0.00004632, 0.00003557, -0.00006036, 0.00002406, -0.00009398, 0.00005155, -0.00013473, 0.00000510, -0.00034383, -0.00047107, 0.00088819, 0.00081114, 0.00054140, -0.00080612, 0.00005488, -0.00018334, 0.00005096, -0.00011383, 0.00003950, -0.00005770, 0.00004043, -0.00007067, 0.00003432, -0.00003849, 0.00002991, -0.00004026, 0.00002863, -0.00002207, 0.00002157, -0.00003289, 0.00001483, -0.00002408, 0.00002079, -0.00001976, 0.00001576, -0.00001115, 0.00001764, -0.00002609, 0.00002655, -0.00003333, 0.00003246, -0.00003401, 0.00001899, -0.00001624, 0.00001339, -0.00001569, 0.00001032, -0.00000202, -0.00000032, -0.00000837, 0.00000461, -0.00000827, 0.00000920, -0.00000267, -0.00000014, -0.00000331, 0.00000076, -0.00000682, 0.00000562, -0.00000499, 0.00000418, -0.00000419, -0.00000137, -0.00000295, 0.00000212, 0.00000506, -0.00000646, -0.00000667, 0.00000431, 0.00001123, -0.00001261, 0.00000195, 0.00000023, 0.00000003, -0.00000130, 0.00000017, },
		{ -0.00000767, -0.00000510, 0.00001278, -0.00000064, -0.00000655, -0.00000169, 0.00000687, 0.00000147, -0.00000282, -0.00000111, 0.00000085, 0.00000085, -0.00000397, 0.00000542, -0.00000883, 0.00000195, -0.00000589, 0.00000326, 0.00000516, 0.00000448, -0.00000724, 0.00000619, -0.00000813, -0.00000063, 0.00000128, 0.00000414, -0.00000153, 0.00000016, -0.00000208, 0.00000550, -0.00001132, 0.00002237, -0.00003829, 0.00003398, -0.00003398, 0.00002239, -0.00002289, 0.00001984, -0.00002322, 0.00001920, -0.00003676, 0.00002933, -0.00002935, 0.00002216, -0.00002922, 0.00003276, -0.00004090, 0.00002866, -0.00005121, 0.00004865, -0.00005688, 0.00003703, -0.00006373, 0.00005443, -0.00008499, 0.00005331, -0.00013158, 0.00009420, -0.00019598, 0.00008267, -0.00055094, -0.00011129, 0.00085459, 0.00058744, 0.00057771, -0.00079421, 0.00007522, -0.00019532, 0.00006617, -0.00012504, 0.00005308, -0.00007093, 0.00005206, -0.00007867, 0.00004233, -0.00004545, 0.00003604, -0.00004548, 0.00003621, -0.00003067, 0.00002847, -0.00003792, 0.00001747, -0.00002544, 0.00002388, -0.00002256, 0.00001826, -0.00001185, 0.00001569, -0.00002078, 0.00002046, -0.00002902, 0.00003203, -0.00003956, 0.00002757, -0.00002590, 0.00002320, -0.00002510, 0.00002055, -0.00001201, 0.00000593, -0.00001268, 0.00000859, -0.00001231, 0.00001562, -0.00000935, 0.00000364, -0.00000621, 0.00000259, -0.00000800, 0.00000771, -0.00000795, 0.00000797, -0.00000842, 0.00000167, -0.00000531, 0.00000618, 0.00000116, -0.00000679, -0.00000623, 0.00000916, 0.00000591, -0.00001336, 0.00000604, -0.00000319, 0.00000309, -0.00000495, 0.00000519, },
		{ -0.00001383, 0.00000331, 0.00000884, -0.00000100, -0.00000847, 0.00000267, 0.00000495, 0.00000189, -0.00000460, 0.00000071, 0.00000041, 0.00000044, -0.00000312, 0.00000467, -0.00001010, 0.00000470, -0.00000882, 0.00000703, 0.00000411, 0.00000300, -0.00000675, 0.00000655, -0.00001040, 0.00000349, -0.00000039, 0.00000346, -0.00000006, -0.00000296, 0.00000411, -0.00000456, 0.00000282, 0.00000688, -0.00002697, 0.00002701, -0.00003140, 0.00002321, -0.00002361, 0.00001926, -0.00002198, 0.00001694, -0.00003605, 0.00002987, -0.00002961, 0.00002218, -0.00002698, 0.00002893, -0.00003894, 0.00002803, -0.00005073, 0.00004761, -0.00005768, 0.00003807, -0.00006310, 0.00005213, -0.00008514, 0.00005450, -0.00013684, 0.00009798, -0.00020900, 0.00009128, -0.00067959, 0.00026393, 0.00067836, 0.00049418, 0.00055496, -0.00074390, 0.00006273, -0.00018061, 0.00005793, -0.00011530, 0.00004767, -0.00006772, 0.00004950, -0.00007372, 0.00003824, -0.00004129, 0.00003159, -0.00004003, 0.00003340, -0.00003034, 0.00002796, -0.00003663, 0.00001436, -0.00002061, 0.00002064, -0.00001964, 0.00001592, -0.00000864, 0.00001021, -0.00001098, 0.00000774, -0.00001555, 0.00001991, -0.00003267, 0.00002471, -0.00002553, 0.00002430, -0.00002682, 0.00002429, -0.00001750, 0.00000907, -0.00001393, 0.00000917, -0.00001239, 0.00001820, -0.00001356, 0.00000593, -0.00000777, 0.00000305, -0.00000731, 0.00000742, -0.00000837, 0.00000943, -0.00001097, 0.00000364, -0.00000651, 0.00000940, -0.00000328, -0.00000627, -0.00000512, 0.00001336, -0.00000145, -0.00001022, 0.00000687, -0.00000431, 0.00000403, -0.00000634, 0.00000796, },
		{ -0.00001637, 0.00001023, 0.00000423, -0.00000120, -0.00000924, 0.00000642, 0.00000275, 0.00000221, -0.00000595, 0.00000253, -0.00000030, 0.00000006, -0.00000170, 0.00000241, -0.00000921, 0.00000547, -0.00001023, 0.00001030, 0.00000253, 0.00000156, -0.00000528, 0.00000534, -0.00001101, 0.00000719, -0.00000322, 0.00000456, -0.00000086, -0.00000299, 0.00000660, -0.00001114, 0.00001524, -0.00001060, -0.00000990, 0.00001194, -0.00001981, 0.00001565, -0.00001711, 0.00001202, -0.00001409, 0.00000732, -0.00002703, 0.00002195, -0.00002170, 0.00001439, -0.00001679, 0.00001571, -0.00002602, 0.00001558, -0.00003760, 0.00003258, -0.00004333, 0.00002340, -0.00004546, 0.00002987, -0.00006165, 0.00002772, -0.00010880, 0.00006024, -0.00016783, 0.00001316, -0.00067691, 0.00056074, 0.00044927, 0.00048452, 0.00049071, -0.00066561, 0.00002564, -0.00014595, 0.00003213, -0.00008970, 0.00002743, -0.00005141, 0.00003582, -0.00005883, 0.00002498, -0.00002885, 0.00001945, -0.00002664, 0.00002233, -0.00002260, 0.00002140, -0.00003057, 0.00000748, -0.00001192, 0.00001330, -0.00001306, 0.00001084, -0.00000396, 0.00000458, -0.00000132, -0.00000602, 0.00000114, 0.00000144, -0.00001713, 0.00001264, -0.00001616, 0.00001691, -0.00002039, 0.00002041, -0.00001704, 0.00000799, -0.00001154, 0.00000608, -0.00000822, 0.00001616, -0.00001419, 0.00000580, -0.00000743, 0.00000193, -0.00000483, 0.00000480, -0.00000607, 0.00000801, -0.00001104, 0.00000369, -0.00000568, 0.00001059, -0.00000707, -0.00000555, -0.00000275, 0.00001544, -0.00000887, -0.00000448, 0.00000468, -0.00000284, 0.00000232, -0.00000461, 0.00000696, },
		{ -0.00001425, 0.00001377, 0.00000056, -0.00000204, -0.00000817, 0.00000857, 0.00000113, 0.00000188, -0.00000635, 0.00000379, -0.00000098, -0.00000017, -0.00000022, -0.00000077, -0.00000646, 0.00000413, -0.00000955, 0.00001230, 0.00000084, 0.00000038, -0.00000334, 0.00000285, -0.00000968, 0.00000963, -0.00000638, 0.00000700, -0.00000391, 0.00000072, 0.00000359, -0.00001072, 0.00002037, -0.00002295, 0.00000544, -0.00000436, -0.00000488, 0.00000428, -0.00000744, 0.00000220, -0.00000394, -0.00000494, -0.00001439, 0.00001010, -0.00001005, 0.00000349, -0.00000407, -0.00000069, -0.00000872, -0.00000192, -0.00001893, 0.00001104, -0.00002156, 0.00000129, -0.00002023, -0.00000151, -0.00002664, -0.00001364, -0.00006253, -0.00000175, -0.00009292, -0.00013284, -0.00053392, 0.00071997, 0.00024677, 0.00050250, 0.00041636, -0.00058162, -0.00001807, -0.00010639, 0.00000195, -0.00005996, 0.00000266, -0.00003085, 0.00001886, -0.00004131, 0.00000939, -0.00001462, 0.00000593, -0.00001143, 0.00000848, -0.00001204, 0.00001271, -0.00002338, 0.00000062, -0.00000335, 0.00000572, -0.00000630, 0.00000618, -0.00000095, 0.00000248, 0.00000337, -0.00001462, 0.00001361, -0.00001551, -0.00000018, -0.00000254, -0.00000280, 0.00000509, -0.00000904, 0.00001115, -0.00001197, 0.00000380, -0.00000685, 0.00000094, -0.00000140, 0.00001061, -0.00001177, 0.00000361, -0.00000565, -0.00000006, -0.00000156, 0.00000099, -0.00000207, 0.00000440, -0.00000894, 0.00000196, -0.00000279, 0.00000926, -0.00000944, -0.00000499, 0.00000102, 0.00001436, -0.00001424, 0.00000180, 0.00000086, 0.00000027, -0.00000126, -0.00000032, 0.00000221, },
		{ -0.00000813, 0.00001344, -0.00000141, -0.00000374, -0.00000522, 0.00000871, 0.00000053, 0.00000069, -0.00000557, 0.00000419, -0.00000146, -0.00000010, 0.00000074, -0.00000391, -0.00000277, 0.00000122, -0.00000686, 0.00001265, -0.00000064, -0.00000035, -0.00000150, -0.00000029, -0.00000655, 0.00001014, -0.00000869, 0.00000949, -0.00000783, 0.00000682, -0.00000406, -0.00000290, 0.00001575, -0.00002544, 0.00001279, -0.00001518, 0.00000705, -0.00000536, 0.00000040, -0.00000534, 0.00000335, -0.00001426, -0.00000396, 0.00000001, -0.00000027, -0.00000476, 0.00000479, -0.00001295, 0.00000497, -0.00001599, -0.00000384, -0.00000721, -0.00000271, -0.00001724, 0.00000012, -0.00002754, 0.00000314, -0.00005015, -0.00002100, -0.00005977, -0.00002170, -0.00029849, -0.00028643, 0.00073009, 0.00011868, 0.00050474, 0.00035984, -0.00051289, -0.00005117, -0.00007649, -0.00001974, -0.00003772, -0.00001625, -0.00001510, 0.00000662, -0.00002847, -0.00000179, -0.00000487, -0.00000284, -0.00000050, -0.00000247, -0.00000351, 0.00000590, -0.00001849, -0.00000295, 0.00000172, 0.00000115, -0.00000219, 0.00000421, -0.00000137, 0.00000564, 0.00000060, -0.00001429, 0.00001646, -0.00002417, 0.00001098, -0.00001397, 0.00000828, -0.00000546, 0.00000217, 0.00000066, -0.00000539, -0.00000103, -0.00000223, -0.00000369, 0.00000537, 0.00000388, -0.00000795, 0.00000060, -0.00000355, -0.00000174, 0.00000109, -0.00000237, 0.00000195, 0.00000008, -0.00000581, -0.00000058, 0.00000129, 0.00000579, -0.00001024, -0.00000451, 0.00000570, 0.00000994, -0.00001605, 0.00000650, -0.00000271, 0.00000335, -0.00000499, 0.00000473, -0.00000472, },
		{ -0.00000008, 0.00001011, -0.00000188, -0.00000579, -0.00000107, 0.00000718, 0.00000083, -0.00000111, -0.00000382, 0.00000371, -0.00000163, 0.00000033, 0.00000067, -0.00000602, 0.00000067, -0.00000215, -0.00000285, 0.00001139, -0.00000165, -0.00000058, -0.00000029, -0.00000333, -0.00000213, 0.00000843, -0.00000909, 0.00001039, -0.00001036, 0.00001236, -0.00001301, 0.00000919, 0.00000317, -0.00001773, 0.00000983, -0.00001681, 0.00001184, -0.00000945, 0.00000306, -0.00000742, 0.00000448, -0.00001693, 0.00000020, -0.00000422, 0.00000376, -0.00000664, 0.00000572, -0.00001627, 0.00000955, -0.00002063, 0.00000110, -0.00001489, 0.00000549, -0.00002399, 0.00000638, -0.00003726, 0.00001452, -0.00006597, -0.00000382, -0.00008831, 0.00000814, -0.00042467, -0.00000026, 0.00062297, 0.00007438, 0.00047203, 0.00033452, -0.00046987, -0.00006485, -0.00006366, -0.00002629, -0.00002920, -0.00002351, -0.00000918, 0.00000337, -0.00002411, -0.00000512, -0.00000269, -0.00000379, 0.00000284, -0.00000721, 0.00000017, 0.00000311, -0.00001736, -0.00000212, 0.00000212, 0.00000073, -0.00000155, 0.00000511, -0.00000457, 0.00001278, -0.00000836, -0.00000557, 0.00000878, -0.00002181, 0.00001243, -0.00001721, 0.00001251, -0.00001019, 0.00000873, -0.00000704, -0.00000045, -0.00000410, 0.00000018, -0.00000558, 0.00000956, -0.00000159, -0.00000460, -0.00000187, -0.00000218, -0.00000217, 0.00000198, -0.00000390, 0.00000437, -0.00000331, -0.00000309, -0.00000262, 0.00000523, 0.00000116, -0.00000981, -0.00000367, 0.00001017, 0.00000308, -0.00001393, 0.00000824, -0.00000441, 0.00000476, -0.00000699, 0.00000825, -0.00001128, },
		{ 0.00000731, 0.00000549, -0.00000174, -0.00000725, 0.00000312, 0.00000482, 0.00000145, -0.00000292, -0.00000165, 0.00000263, -0.00000149, 0.00000108, -0.00000060, -0.00000644, 0.00000278, -0.00000477, 0.00000146, 0.00000894, -0.00000211, -0.00000034, -0.00000003, -0.00000544, 0.00000267, 0.00000487, -0.00000717, 0.00000861, -0.00000954, 0.00001424, -0.00001886, 0.00002022, -0.00001208, -0.00000387, -0.00000121, -0.00000988, 0.00000908, -0.00000739, 0.00000018, -0.00000398, -0.00000054, -0.00001265, -0.00000255, -0.00000189, 0.00000157, -0.00000209, -0.00000116, -0.00001040, 0.00000436, -0.00001496, -0.00000532, -0.00001042, 0.00000130, -0.00001739, -0.00000317, -0.00002824, 0.00000394, -0.00005610, -0.00001809, -0.00007663, -0.00002467, -0.00046353, 0.00025249, 0.00045626, 0.00009236, 0.00040897, 0.00033710, -0.00045007, -0.00006102, -0.00006614, -0.00001919, -0.00003339, -0.00001969, -0.00001250, 0.00000833, -0.00002738, -0.00000156, -0.00000701, 0.00000207, -0.00000090, -0.00000579, -0.00000097, 0.00000398, -0.00001907, 0.00000183, -0.00000089, 0.00000333, -0.00000310, 0.00000717, -0.00000794, 0.00002021, -0.00001912, 0.00000701, -0.00000564, -0.00001069, 0.00000489, -0.00001193, 0.00000888, -0.00000756, 0.00000861, -0.00000972, 0.00000109, -0.00000421, -0.00000046, -0.00000383, 0.00000987, -0.00000430, -0.00000296, -0.00000296, -0.00000203, -0.00000107, 0.00000084, -0.00000305, 0.00000434, -0.00000469, -0.00000184, -0.00000305, 0.00000769, -0.00000342, -0.00000876, -0.00000190, 0.00001305, -0.00000447, -0.00000878, 0.00000688, -0.00000363, 0.00000367, -0.00000609, 0.00000846, -0.00001498, },
		{ 0.00001189, 0.00000135, -0.00000195, -0.00000723, 0.00000618, 0.00000263, 0.00000172, -0.00000407, 0.00000027, 0.00000135, -0.00000111, 0.00000189, -0.00000282, -0.00000503, 0.00000294, -0.00000566, 0.00000496, 0.00000599, -0.00000208, 0.00000022, -0.00000085, -0.00000598, 0.00000667, 0.00000053, -0.00000355, 0.00000432, -0.00000485, 0.00001073, -0.00001828, 0.00002504, -0.00002363, 0.00000975, -0.00001483, 0.00000150, 0.00000171, -0.00000162, -0.00000576, 0.00000223, -0.00000872, -0.00000434, -0.00000956, 0.00000445, -0.00000407, 0.00000549, -0.00001189, 0.00000058, -0.00000655, -0.00000314, -0.00001880, 0.00000194, -0.00001079, -0.00000270, -0.00002240, -0.00000718, -0.00002147, -0.00002820, -0.00005552, -0.00003381, -0.00011510, -0.00039588, 0.00041706, 0.00029039, 0.00013646, 0.00033465, 0.00035314, -0.00044209, -0.00004908, -0.00007559, -0.00000588, -0.00004384, -0.00001024, -0.00002025, 0.00001695, -0.00003396, 0.00000477, -0.00001368, 0.00001072, -0.00000825, -0.00000093, -0.00000469, 0.00000634, -0.00002119, 0.00000616, -0.00000464, 0.00000652, -0.00000460, 0.00000803, -0.00000846, 0.00002373, -0.00002630, 0.00001709, -0.00002017, 0.00000337, -0.00000701, -0.00000167, 0.00000006, 0.00000056, 0.00000279, -0.00000765, -0.00000062, -0.00000184, -0.00000338, 0.00000077, 0.00000663, -0.00000407, -0.00000324, -0.00000268, -0.00000281, 0.00000098, -0.00000168, -0.00000035, 0.00000206, -0.00000392, -0.00000234, -0.00000148, 0.00000785, -0.00000694, -0.00000761, 0.00000111, 0.00001316, -0.00001065, -0.00000245, 0.00000345, -0.00000095, 0.00000051, -0.00000244, 0.00000489, -0.00001434, },
		{ 0.00001269, -0.00000120, -0.00000306, -0.00000531, 0.00000739, 0.00000129, 0.00000120, -0.00000411, 0.00000141, 0.00000027, -0.00000052, 0.00000240, -0.00000536, -0.00000230, 0.00000126, -0.00000440, 0.00000680, 0.00000324, -0.00000173, 0.00000092, -0.00000257, -0.00000458, 0.00000878, -0.00000308, 0.00000021, -0.00000093, 0.00000232, 0.00000259, -0.00001068, 0.00002104, -0.00002681, 0.00001727, -0.00002498, 0.00001191, -0.00000581, 0.00000403, -0.00001105, 0.00000729, -0.00001577, 0.00000358, -0.00001657, 0.00001069, -0.00000892, 0.00001129, -0.00002088, 0.00001057, -0.00001682, 0.00000805, -0.00003212, 0.00001473, -0.00002286, 0.00001111, -0.00004084, 0.00001386, -0.00004796, 0.00000174, -0.00009683, 0.00001519, -0.00023308, -0.00023713, 0.00047116, 0.00016892, 0.00017316, 0.00027012, 0.00036617, -0.00043267, -0.00004005, -0.00008223, 0.00000480, -0.00005266, -0.00000193, -0.00002647, 0.00002376, -0.00003878, 0.00000910, -0.00001799, 0.00001742, -0.00001489, 0.00000383, -0.00000813, 0.00000781, -0.00002136, 0.00000831, -0.00000661, 0.00000814, -0.00000429, 0.00000626, -0.00000453, 0.00002089, -0.00002621, 0.00001953, -0.00002849, 0.00001380, -0.00001722, 0.00000818, -0.00000921, 0.00001005, -0.00000554, -0.00000307, -0.00000385, 0.00000132, -0.00000667, 0.00000621, 0.00000147, -0.00000192, -0.00000470, -0.00000173, -0.00000372, 0.00000292, -0.00000431, 0.00000292, -0.00000140, -0.00000175, -0.00000399, 0.00000164, 0.00000566, -0.00000891, -0.00000653, 0.00000511, 0.00001004, -0.00001373, 0.00000294, -0.00000028, 0.00000215, -0.00000328, 0.00000253, -0.00000137, -0.00000950, },
		{ 0.00001018, -0.00000204, -0.00000485, -0.00000184, 0.00000673, 0.00000097, -0.00000011, -0.00000300, 0.00000154, -0.00000036, 0.00000022, 0.00000223, -0.00000738, 0.00000080, -0.00000153, -0.00000128, 0.00000668, 0.00000123, -0.00000123, 0.00000149, -0.00000475, -0.00000141, 0.00000837, -0.00000461, 0.00000230, -0.00000481, 0.00000908, -0.00000708, 0.00000132, 0.00000938, -0.00002066, 0.00001582, -0.00002768, 0.00001714, -0.00000980, 0.00000642, -0.00001279, 0.00000818, -0.00001828, 0.00000743, -0.00001995, 0.00001340, -0.00000965, 0.00001162, -0.00002370, 0.00001458, -0.00002104, 0.00001272, -0.00003886, 0.00002118, -0.00002785, 0.00001607, -0.00004901, 0.00002360, -0.00006213, 0.00001745, -0.00012134, 0.00004077, -0.00033478, -0.00003058, 0.00042718, 0.00010861, 0.00018347, 0.00022839, 0.00036551, -0.00041319, -0.00004103, -0.00007974, 0.00000696, -0.00005444, 0.00000056, -0.00002719, 0.00002521, -0.00003864, 0.00000851, -0.00001702, 0.00001902, -0.00001771, 0.00000589, -0.00000942, 0.00000713, -0.00001859, 0.00000719, -0.00000573, 0.00000741, -0.00000195, 0.00000228, 0.00000298, 0.00001229, -0.00001865, 0.00001276, -0.00002734, 0.00001629, -0.00002109, 0.00001301, -0.00001443, 0.00001662, -0.00001260, 0.00000118, -0.00000645, 0.00000334, -0.00000831, 0.00001019, -0.00000347, 0.00000058, -0.00000626, -0.00000098, -0.00000388, 0.00000368, -0.00000575, 0.00000526, -0.00000458, 0.00000058, -0.00000569, 0.00000521, 0.00000182, -0.00000939, -0.00000530, 0.00000927, 0.00000421, -0.00001300, 0.00000579, -0.00000264, 0.00000405, -0.00000585, 0.00000666, -0.00000806, -0.00000209, },
		{ 0.00000585, -0.00000187, -0.00000655, 0.00000222, 0.00000482, 0.00000129, -0.00000177, -0.00000115, 0.00000083, -0.00000050, 0.00000106, 0.00000114, -0.00000815, 0.00000317, -0.00000426, 0.00000281, 0.00000484, 0.00000017, -0.00000068, 0.00000163, -0.00000666, 0.00000278, 0.00000560, -0.00000348, 0.00000150, -0.00000527, 0.00001235, -0.00001415, 0.00001292, -0.00000549, -0.00000817, 0.00000658, -0.00002251, 0.00001589, -0.00000890, 0.00000454, -0.00001025, 0.00000422, -0.00001530, 0.00000593, -0.00001838, 0.00001149, -0.00000553, 0.00000578, -0.00001919, 0.00001096, -0.00001732, 0.00000864, -0.00003635, 0.00001847, -0.00002299, 0.00000916, -0.00004304, 0.00001694, -0.00005741, 0.00001029, -0.00011702, 0.00002210, -0.00037935, 0.00016885, 0.00032253, 0.00010085, 0.00016610, 0.00021049, 0.00034973, -0.00038258, -0.00005261, -0.00006757, -0.00000014, -0.00004826, -0.00000350, -0.00002199, 0.00002105, -0.00003344, 0.00000311, -0.00001079, 0.00001512, -0.00001598, 0.00000466, -0.00000837, 0.00000469, -0.00001363, 0.00000355, -0.00000266, 0.00000517, 0.00000113, -0.00000177, 0.00001107, 0.00000136, -0.00000701, -0.00000065, -0.00001777, 0.00001047, -0.00001733, 0.00001098, -0.00001337, 0.00001766, -0.00001580, 0.00000301, -0.00000694, 0.00000309, -0.00000710, 0.00001117, -0.00000654, 0.00000213, -0.00000701, -0.00000103, -0.00000289, 0.00000280, -0.00000526, 0.00000567, -0.00000627, 0.00000193, -0.00000629, 0.00000790, -0.00000258, -0.00000885, -0.00000346, 0.00001234, -0.00000287, -0.00000895, 0.00000558, -0.00000274, 0.00000371, -0.00000585, 0.00000802, -0.00001268, 0.00000545, },
		{ 0.00000150, -0.00000173, -0.00000718, 0.00000569, 0.00000262, 0.00000162, -0.00000311, 0.00000079, -0.00000033, -0.00000022, 0.00000188, -0.00000081, -0.00000730, 0.00000396, -0.00000577, 0.00000666, 0.00000200, -0.00000005, -0.00000009, 0.00000107, -0.00000752, 0.00000680, 0.00000151, -0.00000025, -0.00000213, -0.00000167, 0.00001032, -0.00001529, 0.00001926, -0.00001782, 0.00000517, -0.00000609, -0.00001235, 0.00000987, -0.00000438, -0.00000025, -0.00000506, -0.00000275, -0.00000853, 0.00000049, -0.00001317, 0.00000650, 0.00000138, -0.00000377, -0.00000976, 0.00000195, -0.00000791, -0.00000197, -0.00002661, 0.00000868, -0.00001094, -0.00000646, -0.00002624, -0.00000288, -0.00003694, -0.00001676, -0.00008645, -0.00004309, -0.00034425, 0.00031348, 0.00020313, 0.00012161, 0.00013286, 0.00020777, 0.00032514, -0.00034622, -0.00006980, -0.00005022, -0.00001265, -0.00003725, -0.00001130, -0.00001360, 0.00001393, -0.00002579, -0.00000444, -0.00000192, 0.00000782, -0.00001117, 0.00000129, -0.00000619, 0.00000209, -0.00000845, -0.00000069, 0.00000084, 0.00000311, 0.00000304, -0.00000329, 0.00001603, -0.00000724, 0.00000337, -0.00001520, -0.00000437, -0.00000033, -0.00000825, 0.00000358, -0.00000684, 0.00001322, -0.00001459, 0.00000191, -0.00000516, 0.00000071, -0.00000321, 0.00000891, -0.00000710, 0.00000219, -0.00000670, -0.00000187, -0.00000100, 0.00000058, -0.00000300, 0.00000397, -0.00000604, 0.00000174, -0.00000511, 0.00000868, -0.00000642, -0.00000785, -0.00000066, 0.00001313, -0.00000925, -0.00000315, 0.00000301, -0.00000081, 0.00000121, -0.00000304, 0.00000575, -0.00001345, 0.00001083, },
		{ -0.00000153, -0.00000236, -0.00000608, 0.00000761, 0.00000096, 0.00000136, -0.00000356, 0.00000218, -0.00000144, 0.00000030, 0.00000246, -0.00000322, -0.00000499, 0.00000291, -0.00000529, 0.00000915, -0.00000091, 0.00000021, 0.00000057, -0.00000030, -0.00000670, 0.00000933, -0.00000228, 0.00000352, -0.00000708, 0.00000474, 0.00000347, -0.00000962, 0.00001753, -0.00002300, 0.00001386, -0.00001679, -0.00000177, 0.00000253, 0.00000094, -0.00000523, -0.00000015, -0.00000949, -0.00000128, -0.00000576, -0.00000730, 0.00000151, 0.00000746, -0.00001275, -0.00000001, -0.00000774, 0.00000226, -0.00001392, -0.00001491, -0.00000267, 0.00000200, -0.00002334, -0.00000711, -0.00002634, -0.00001158, -0.00005086, -0.00004594, -0.00013663, -0.00023316, 0.00037733, 0.00010693, 0.00014463, 0.00009985, 0.00020797, 0.00030105, -0.00031195, -0.00008549, -0.00003419, -0.00002478, -0.00002634, -0.00001855, -0.00000600, 0.00000755, -0.00001921, -0.00001054, 0.00000597, 0.00000036, -0.00000581, -0.00000230, -0.00000449, 0.00000100, -0.00000500, -0.00000356, 0.00000310, 0.00000265, 0.00000253, -0.00000071, 0.00001537, -0.00000991, 0.00000766, -0.00002508, 0.00000699, -0.00001089, 0.00000168, -0.00000534, 0.00000199, 0.00000560, -0.00001041, -0.00000111, -0.00000221, -0.00000247, 0.00000194, 0.00000444, -0.00000564, 0.00000104, -0.00000568, -0.00000295, 0.00000100, -0.00000203, 0.00000010, 0.00000090, -0.00000427, 0.00000027, -0.00000227, 0.00000716, -0.00000893, -0.00000673, 0.00000305, 0.00001097, -0.00001313, 0.00000238, -0.00000042, 0.00000195, -0.00000230, 0.00000150, 0.00000046, -0.00001003, 0.00001273, },
		{ -0.00000279, -0.00000386, -0.00000327, 0.00000767, 0.00000025, 0.00000034, -0.00000289, 0.00000263, -0.00000215, 0.00000094, 0.00000252, -0.00000543, -0.00000194, 0.00000041, -0.00000275, 0.00000964, -0.00000312, 0.00000058, 0.00000124, -0.00000231, -0.00000403, 0.00000946, -0.00000423, 0.00000591, -0.00001103, 0.00001124, -0.00000544, 0.00000077, 0.00000830, -0.00001953, 0.00001471, -0.00002144, 0.00000517, -0.00000272, 0.00000428, -0.00000787, 0.00000187, -0.00001307, 0.00000329, -0.00000977, -0.00000362, -0.00000069, 0.00000953, -0.00001728, 0.00000570, -0.00001350, 0.00000819, -0.00002180, -0.00000686, -0.00000977, 0.00000927, -0.00003366, 0.00000506, -0.00004257, 0.00000566, -0.00007578, -0.00001815, -0.00022642, -0.00007390, 0.00036141, 0.00005353, 0.00015196, 0.00007917, 0.00020160, 0.00028465, -0.00028570, -0.00009427, -0.00002451, -0.00003190, -0.00001947, -0.00002189, -0.00000223, 0.00000469, -0.00001623, -0.00001263, 0.00001007, -0.00000456, -0.00000203, -0.00000462, -0.00000420, 0.00000220, -0.00000420, -0.00000414, 0.00000343, 0.00000401, -0.00000012, 0.00000548, 0.00000911, -0.00000587, 0.00000384, -0.00002674, 0.00001187, -0.00001661, 0.00000802, -0.00001162, 0.00000922, -0.00000189, -0.00000568, -0.00000424, 0.00000032, -0.00000473, 0.00000640, -0.00000049, -0.00000333, -0.00000052, -0.00000464, -0.00000357, 0.00000220, -0.00000387, 0.00000273, -0.00000232, -0.00000197, -0.00000160, 0.00000142, 0.00000371, -0.00000991, -0.00000548, 0.00000708, 0.00000606, -0.00001351, 0.00000592, -0.00000302, 0.00000400, -0.00000512, 0.00000584, -0.00000594, -0.00000357, 0.00001118, },
		{ -0.00000274, -0.00000566, 0.00000055, 0.00000622, 0.00000035, -0.00000122, -0.00000131, 0.00000216, -0.00000230, 0.00000161, 0.00000183, -0.00000671, 0.00000084, -0.00000254, 0.00000121, 0.00000812, -0.00000421, 0.00000083, 0.00000177, -0.00000444, -0.00000002, 0.00000709, -0.00000354, 0.00000552, -0.00001182, 0.00001475, -0.00001252, 0.00001164, -0.00000479, -0.00000953, 0.00000802, -0.00001885, 0.00000657, -0.00000413, 0.00000433, -0.00000716, 0.00000001, -0.00001221, 0.00000366, -0.00000999, -0.00000345, 0.00000095, 0.00000649, -0.00001576, 0.00000531, -0.00001301, 0.00000726, -0.00002256, -0.00000564, -0.00000951, 0.00000743, -0.00003312, 0.00000483, -0.00004475, 0.00000615, -0.00007997, -0.00002180, -0.00027879, 0.00009226, 0.00028912, 0.00004229, 0.00013862, 0.00007452, 0.00018551, 0.00027794, -0.00026894, -0.00009480, -0.00002262, -0.00003251, -0.00001790, -0.00002043, -0.00000300, 0.00000588, -0.00001722, -0.00001026, 0.00000957, -0.00000588, -0.00000070, -0.00000528, -0.00000514, 0.00000515, -0.00000556, -0.00000283, 0.00000247, 0.00000612, -0.00000313, 0.00001280, -0.00000011, 0.00000248, -0.00000647, -0.00002031, 0.00000907, -0.00001553, 0.00000841, -0.00001261, 0.00001200, -0.00000651, -0.00000257, -0.00000592, 0.00000119, -0.00000478, 0.00000857, -0.00000426, -0.00000139, -0.00000167, -0.00000413, -0.00000327, 0.00000203, -0.00000415, 0.00000384, -0.00000449, -0.00000026, -0.00000279, 0.00000473, -0.00000076, -0.00000962, -0.00000383, 0.00001041, -0.00000046, -0.00001052, 0.00000663, -0.00000371, 0.00000415, -0.00000576, 0.00000801, -0.00001106, 0.00000373, 0.00000738, },
		{ -0.00000238, -0.00000680, 0.00000428, 0.00000407, 0.00000072, -0.00000274, 0.00000060, 0.00000110, -0.00000198, 0.00000221, 0.00000035, -0.00000656, 0.00000241, -0.00000473, 0.00000542, 0.00000523, -0.00000425, 0.00000094, 0.00000187, -0.00000591, 0.00000426, 0.00000308, -0.00000056, 0.00000221, -0.00000861, 0.00001331, -0.00001444, 0.00001836, -0.00001655, 0.00000235, -0.00000283, -0.00001091, 0.00000318, -0.00000211, 0.00000168, -0.00000396, -0.00000469, -0.00000784, 0.00000045, -0.00000697, -0.00000603, 0.00000519, -0.00000017, -0.00000950, -0.00000014, -0.00000726, 0.00000033, -0.00001683, -0.00001064, -0.00000289, -0.00000221, -0.00002293, -0.00000698, -0.00003332, -0.00001021, -0.00006239, -0.00006278, -0.00027125, 0.00022523, 0.00019476, 0.00005809, 0.00011086, 0.00008156, 0.00016286, 0.00027783, -0.00025866, -0.00008978, -0.00002625, -0.00002836, -0.00002004, -0.00001579, -0.00000668, 0.00000948, -0.00002042, -0.00000515, 0.00000569, -0.00000430, -0.00000130, -0.00000497, -0.00000619, 0.00000837, -0.00000759, -0.00000102, 0.00000160, 0.00000734, -0.00000415, 0.00001794, -0.00000831, 0.00001067, -0.00001878, -0.00000932, 0.00000085, -0.00000895, 0.00000344, -0.00000829, 0.00000964, -0.00000716, -0.00000204, -0.00000560, 0.00000015, -0.00000238, 0.00000787, -0.00000596, -0.00000055, -0.00000202, -0.00000429, -0.00000210, 0.00000058, -0.00000276, 0.00000303, -0.00000498, 0.00000016, -0.00000246, 0.00000653, -0.00000514, -0.00000855, -0.00000146, 0.00001189, -0.00000687, -0.00000537, 0.00000484, -0.00000240, 0.00000221, -0.00000370, 0.00000684, -0.00001293, 0.00000956, 0.00000306, },
		{ -0.00000257, -0.00000649, 0.00000687, 0.00000210, 0.00000079, -0.00000361, 0.00000220, -0.00000008, -0.00000135, 0.00000262, -0.00000170, -0.00000494, 0.00000222, -0.00000519, 0.00000864, 0.00000193, -0.00000362, 0.00000101, 0.00000128, -0.00000600, 0.00000747, -0.00000104, 0.00000330, -0.00000273, -0.00000235, 0.00000703, -0.00000996, 0.00001796, -0.00002247, 0.00001101, -0.00001306, -0.00000138, -0.00000241, 0.00000137, -0.00000176, -0.00000041, -0.00000982, -0.00000239, -0.00000411, -0.00000281, -0.00000912, 0.00000932, -0.00000717, -0.00000192, -0.00000725, 0.00000023, -0.00000894, -0.00000818, -0.00001815, 0.00000574, -0.00001444, -0.00000895, -0.00002392, -0.00001544, -0.00003515, -0.00003331, -0.00013073, -0.00020080, 0.00029893, 0.00011028, 0.00008091, 0.00008023, 0.00009183, 0.00014005, 0.00027860, -0.00024959, -0.00008402, -0.00003114, -0.00002302, -0.00002277, -0.00001091, -0.00001053, 0.00001285, -0.00002308, -0.00000008, 0.00000085, -0.00000155, -0.00000258, -0.00000475, -0.00000612, 0.00001030, -0.00000870, -0.00000008, 0.00000203, 0.00000647, -0.00000159, 0.00001838, -0.00001205, 0.00001433, -0.00002788, 0.00000109, -0.00000840, -0.00000048, -0.00000395, -0.00000098, 0.00000368, -0.00000460, -0.00000364, -0.00000384, -0.00000198, 0.00000154, 0.00000481, -0.00000566, -0.00000083, -0.00000170, -0.00000480, -0.00000063, -0.00000148, -0.00000035, 0.00000074, -0.00000389, -0.00000074, -0.00000046, 0.00000617, -0.00000850, -0.00000713, 0.00000170, 0.00001076, -0.00001139, 0.00000013, 0.00000178, 0.00000002, -0.00000100, 0.00000036, 0.00000254, -0.00001080, 0.00001233, -0.00000028, },
		{ -0.00000369, -0.00000446, 0.00000774, 0.00000089, 0.00000019, -0.00000347, 0.00000303, -0.00000095, -0.00000060, 0.00000264, -0.00000378, -0.00000230, 0.00000041, -0.00000355, 0.00000998, -0.00000083, -0.00000281, 0.00000117, -0.00000001, -0.00000432, 0.00000854, -0.00000366, 0.00000613, -0.00000708, 0.00000449, -0.00000177, -0.00000058, 0.00001039, -0.00002062, 0.00001304, -0.00001863, 0.00000594, -0.00000720, 0.00000398, -0.00000388, 0.00000130, -0.00001289, 0.00000151, -0.00000748, 0.00000015, -0.00001043, 0.00001062, -0.00001116, 0.00000325, -0.00001215, 0.00000533, -0.00001609, -0.00000117, -0.00002354, 0.00001112, -0.00002295, 0.00000137, -0.00003740, -0.00000124, -0.00005623, -0.00000983, -0.00020250, -0.00008458, 0.00030782, 0.00005553, 0.00009488, 0.00005699, 0.00009764, 0.00012283, 0.00027525, -0.00023716, -0.00008178, -0.00003336, -0.00001981, -0.00002327, -0.00000834, -0.00001219, 0.00001382, -0.00002296, 0.00000249, -0.00000266, 0.00000061, -0.00000338, -0.00000530, -0.00000433, 0.00001014, -0.00000806, -0.00000062, 0.00000403, 0.00000361, 0.00000445, 0.00001361, -0.00001010, 0.00001116, -0.00003009, 0.00000653, -0.00001431, 0.00000581, -0.00000999, 0.00000590, -0.00000309, -0.00000079, -0.00000597, -0.00000191, -0.00000380, 0.00000542, 0.00000075, -0.00000418, -0.00000171, -0.00000118, -0.00000509, 0.00000037, -0.00000316, 0.00000199, -0.00000205, -0.00000198, -0.00000232, 0.00000269, 0.00000371, -0.00001037, -0.00000555, 0.00000527, 0.00000697, -0.00001283, 0.00000422, -0.00000099, 0.00000216, -0.00000396, 0.00000476, -0.00000342, -0.00000539, 0.00001171, -0.00000194, },
		{ -0.00000538, -0.00000113, 0.00000696, 0.00000053, -0.00000101, -0.00000239, 0.00000294, -0.00000130, 0.00000014, 0.00000205, -0.00000524, 0.00000047, -0.00000224, -0.00000018, 0.00000917, -0.00000246, -0.00000218, 0.00000137, -0.00000168, -0.00000111, 0.00000713, -0.00000379, 0.00000643, -0.00000866, 0.00000893, -0.00000952, 0.00001003, -0.00000146, -0.00001229, 0.00000803, -0.00001779, 0.00000878, -0.00000925, 0.00000429, -0.00000357, 0.00000007, -0.00001256, 0.00000228, -0.00000813, 0.00000058, -0.00000882, 0.00000785, -0.00001036, 0.00000378, -0.00001243, 0.00000532, -0.00001809, 0.00000108, -0.00002374, 0.00000987, -0.00002352, 0.00000276, -0.00004099, 0.00000135, -0.00006310, -0.00000798, -0.00025089, 0.00004684, 0.00026554, 0.00003465, 0.00009326, 0.00004580, 0.00009549, 0.00011354, 0.00026586, -0.00021962, -0.00008482, -0.00003116, -0.00002018, -0.00002044, -0.00000902, -0.00001090, 0.00001179, -0.00001937, 0.00000153, -0.00000362, 0.00000125, -0.00000328, -0.00000650, -0.00000122, 0.00000824, -0.00000594, -0.00000216, 0.00000672, 0.00000020, 0.00001203, 0.00000549, -0.00000399, 0.00000202, -0.00002486, 0.00000519, -0.00001449, 0.00000724, -0.00001189, 0.00000947, -0.00000799, 0.00000220, -0.00000754, -0.00000099, -0.00000408, 0.00000773, -0.00000282, -0.00000254, -0.00000253, -0.00000097, -0.00000470, 0.00000032, -0.00000364, 0.00000323, -0.00000422, -0.00000029, -0.00000359, 0.00000592, -0.00000017, -0.00001079, -0.00000371, 0.00000842, 0.00000133, -0.00001099, 0.00000584, -0.00000226, 0.00000282, -0.00000518, 0.00000759, -0.00000881, 0.00000146, 0.00000855, -0.00000212, },
		{ -0.00000681, 0.00000255, 0.00000515, 0.00000067, -0.00000238, -0.00000079, 0.00000216, -0.00000112, 0.00000079, 0.00000075, -0.00000550, 0.00000240, -0.00000461, 0.00000389, 0.00000667, -0.00000286, -0.00000185, 0.00000139, -0.00000309, 0.00000273, 0.00000384, -0.00000154, 0.00000382, -0.00000642, 0.00000888, -0.00001292, 0.00001757, -0.00001301, -0.00000133, -0.00000153, -0.00001155, 0.00000709, -0.00000830, 0.00000234, -0.00000115, -0.00000360, -0.00000920, 0.00000002, -0.00000610, -0.00000125, -0.00000498, 0.00000181, -0.00000530, -0.00000009, -0.00000820, 0.00000020, -0.00001460, -0.00000178, -0.00001872, 0.00000211, -0.00001604, -0.00000541, -0.00003352, -0.00000956, -0.00005277, -0.00003521, -0.00025507, 0.00016052, 0.00019718, 0.00003894, 0.00007852, 0.00004496, 0.00008670, 0.00011056, 0.00025199, -0.00019839, -0.00009196, -0.00002535, -0.00002338, -0.00001513, -0.00001206, -0.00000761, 0.00000787, -0.00001338, -0.00000235, -0.00000220, 0.00000048, -0.00000262, -0.00000757, 0.00000202, 0.00000577, -0.00000343, -0.00000354, 0.00000868, -0.00000164, 0.00001824, -0.00000272, 0.00000275, -0.00000956, -0.00001477, -0.00000159, -0.00000937, 0.00000366, -0.00000897, 0.00000854, -0.00000956, 0.00000312, -0.00000752, -0.00000159, -0.00000231, 0.00000766, -0.00000489, -0.00000153, -0.00000285, -0.00000127, -0.00000356, -0.00000088, -0.00000267, 0.00000283, -0.00000500, 0.00000040, -0.00000366, 0.00000808, -0.00000443, -0.00001011, -0.00000139, 0.00001017, -0.00000470, -0.00000672, 0.00000498, -0.00000170, 0.00000154, -0.00000393, 0.00000750, -0.00001163, 0.00000752, 0.00000442, -0.00000166, },
		{ -0.00000711, 0.00000553, 0.00000315, 0.00000077, -0.00000337, 0.00000076, 0.00000112, -0.00000059, 0.00000127, -0.00000114, -0.00000434, 0.00000278, -0.00000559, 0.00000741, 0.00000340, -0.00000238, -0.00000168, 0.00000094, -0.00000348, 0.00000597, 0.00000004, 0.00000189, -0.00000066, -0.00000097, 0.00000408, -0.00001042, 0.00001897, -0.00001997, 0.00000762, -0.00001153, -0.00000289, 0.00000268, -0.00000562, -0.00000061, 0.00000186, -0.00000797, -0.00000454, -0.00000372, -0.00000282, -0.00000378, -0.00000091, -0.00000503, 0.00000154, -0.00000603, -0.00000187, -0.00000760, -0.00000794, -0.00000742, -0.00001132, -0.00000865, -0.00000438, -0.00001902, -0.00001944, -0.00002897, -0.00003130, -0.00008638, -0.00020837, 0.00023202, 0.00012902, 0.00005362, 0.00005858, 0.00004877, 0.00007568, 0.00010983, 0.00023734, -0.00017691, -0.00010015, -0.00001843, -0.00002721, -0.00000947, -0.00001546, -0.00000428, 0.00000413, -0.00000708, -0.00000743, 0.00000046, -0.00000096, -0.00000206, -0.00000758, 0.00000415, 0.00000405, -0.00000170, -0.00000369, 0.00000872, -0.00000030, 0.00002059, -0.00000784, 0.00000626, -0.00001901, -0.00000421, -0.00001025, -0.00000172, -0.00000281, -0.00000277, 0.00000393, -0.00000798, 0.00000196, -0.00000607, -0.00000332, 0.00000101, 0.00000535, -0.00000519, -0.00000138, -0.00000266, -0.00000190, -0.00000206, -0.00000274, -0.00000064, 0.00000097, -0.00000429, -0.00000016, -0.00000214, 0.00000835, -0.00000808, -0.00000879, 0.00000152, 0.00000971, -0.00000946, -0.00000158, 0.00000255, 0.00000013, -0.00000114, -0.00000055, 0.00000431, -0.00001080, 0.00001105, 0.00000086, -0.00000150, },
		{ -0.00000580, 0.00000705, 0.00000163, 0.00000041, -0.00000357, 0.00000174, 0.00000024, 0.00000011, 0.00000142, -0.00000322, -0.00000205, 0.00000152, -0.00000459, 0.00000930, 0.00000039, -0.00000155, -0.00000150, -0.00000012, -0.00000237, 0.00000748, -0.00000271, 0.00000473, -0.00000500, 0.00000553, -0.00000364, -0.00000295, 0.00001365, -0.00002006, 0.00001108, -0.00001808, 0.00000475, -0.00000192, -0.00000303, -0.00000292, 0.00000372, -0.00001102, -0.00000072, -0.00000694, -0.00000013, -0.00000515, 0.00000112, -0.00000983, 0.00000705, -0.00001090, 0.00000324, -0.00001453, -0.00000163, -0.00001229, -0.00000563, -0.00001750, 0.00000568, -0.00003148, -0.00000635, -0.00004763, -0.00001123, -0.00014513, -0.00012047, 0.00025216, 0.00007978, 0.00006519, 0.00004175, 0.00005116, 0.00006700, 0.00010733, 0.00022552, -0.00015858, -0.00010626, -0.00001304, -0.00002943, -0.00000551, -0.00001732, -0.00000267, 0.00000246, -0.00000256, -0.00001179, 0.00000302, -0.00000224, -0.00000206, -0.00000609, 0.00000449, 0.00000381, -0.00000128, -0.00000229, 0.00000678, 0.00000446, 0.00001814, -0.00000836, 0.00000420, -0.00002275, 0.00000265, -0.00001679, 0.00000483, -0.00000892, 0.00000386, -0.00000212, -0.00000470, -0.00000031, -0.00000407, -0.00000514, 0.00000469, 0.00000174, -0.00000419, -0.00000184, -0.00000225, -0.00000241, -0.00000083, -0.00000447, 0.00000158, -0.00000159, -0.00000263, -0.00000154, 0.00000067, 0.00000650, -0.00001043, -0.00000716, 0.00000477, 0.00000683, -0.00001167, 0.00000276, -0.00000008, 0.00000205, -0.00000393, 0.00000360, -0.00000090, -0.00000661, 0.00001138, -0.00000119, -0.00000221, },
		{ -0.00000303, 0.00000692, 0.00000088, -0.00000050, -0.00000289, 0.00000194, -0.00000020, 0.00000082, 0.00000106, -0.00000490, 0.00000067, -0.00000085, -0.00000172, 0.00000907, -0.00000162, -0.00000084, -0.00000123, -0.00000165, 0.00000023, 0.00000674, -0.00000331, 0.00000544, -0.00000710, 0.00001032, -0.00001110, 0.00000643, 0.00000376, -0.00001388, 0.00000809, -0.00001907, 0.00000889, -0.00000476, -0.00000192, -0.00000350, 0.00000338, -0.00001148, 0.00000078, -0.00000818, 0.00000071, -0.00000427, -0.00000017, -0.00001078, 0.00000905, -0.00001236, 0.00000459, -0.00001775, 0.00000146, -0.00001360, -0.00000472, -0.00002057, 0.00000937, -0.00003712, -0.00000117, -0.00005674, -0.00000582, -0.00019000, -0.00001353, 0.00022768, 0.00005642, 0.00006628, 0.00003286, 0.00004865, 0.00006308, 0.00010106, 0.00021829, -0.00014512, -0.00010857, -0.00001060, -0.00002892, -0.00000424, -0.00001680, -0.00000346, 0.00000364, -0.00000093, -0.00001424, 0.00000459, -0.00000300, -0.00000250, -0.00000347, 0.00000328, 0.00000487, -0.00000181, -0.00000003, 0.00000399, 0.00001122, 0.00001202, -0.00000507, -0.00000314, -0.00001974, 0.00000362, -0.00001861, 0.00000756, -0.00001191, 0.00000815, -0.00000714, -0.00000160, -0.00000237, -0.00000258, -0.00000591, 0.00000737, -0.00000189, -0.00000270, -0.00000243, -0.00000204, -0.00000239, -0.00000044, -0.00000530, 0.00000303, -0.00000384, -0.00000089, -0.00000292, 0.00000394, 0.00000295, -0.00001127, -0.00000528, 0.00000778, 0.00000205, -0.00001087, 0.00000510, -0.00000170, 0.00000290, -0.00000544, 0.00000675, -0.00000619, -0.00000054, 0.00000903, -0.00000170, -0.00000367, },
		{ 0.00000046, 0.00000552, 0.00000076, -0.00000170, -0.00000160, 0.00000145, -0.00000016, 0.00000146, 0.00000006, -0.00000561, 0.00000288, -0.00000331, 0.00000220, 0.00000697, -0.00000234, -0.00000047, -0.00000098, -0.00000317, 0.00000368, 0.00000409, -0.00000164, 0.00000346, -0.00000572, 0.00001122, -0.00001513, 0.00001385, -0.00000677, -0.00000451, 0.00000033, -0.00001476, 0.00000892, -0.00000512, -0.00000261, -0.00000235, 0.00000097, -0.00000937, -0.00000025, -0.00000724, -0.00000030, -0.00000145, -0.00000443, -0.00000788, 0.00000721, -0.00001002, 0.00000161, -0.00001642, 0.00000044, -0.00001074, -0.00000914, -0.00001699, 0.00000528, -0.00003393, -0.00000661, -0.00005247, -0.00002278, -0.00020261, 0.00008599, 0.00017633, 0.00005494, 0.00005676, 0.00003197, 0.00004132, 0.00006337, 0.00009164, 0.00021493, -0.00013600, -0.00010734, -0.00001082, -0.00002609, -0.00000519, -0.00001444, -0.00000599, 0.00000706, -0.00000199, -0.00001461, 0.00000503, -0.00000336, -0.00000277, -0.00000066, 0.00000148, 0.00000635, -0.00000231, 0.00000179, 0.00000220, 0.00001750, 0.00000476, -0.00000062, -0.00001307, -0.00001178, -0.00000071, -0.00001546, 0.00000572, -0.00001066, 0.00000861, -0.00000947, -0.00000006, -0.00000325, -0.00000226, -0.00000504, 0.00000814, -0.00000452, -0.00000145, -0.00000276, -0.00000222, -0.00000172, -0.00000106, -0.00000485, 0.00000310, -0.00000499, 0.00000009, -0.00000339, 0.00000655, -0.00000140, -0.00001079, -0.00000310, 0.00000973, -0.00000346, -0.00000758, 0.00000516, -0.00000175, 0.00000209, -0.00000478, 0.00000746, -0.00000957, 0.00000537, 0.00000533, -0.00000131, -0.00000521, },
		{ 0.00000369, 0.00000357, 0.00000084, -0.00000276, -0.00000017, 0.00000061, 0.00000024, 0.00000195, -0.00000154, -0.00000500, 0.00000381, -0.00000476, 0.00000597, 0.00000378, -0.00000202, -0.00000038, -0.00000101, -0.00000400, 0.00000689, 0.00000066, 0.00000137, -0.00000049, -0.00000114, 0.00000763, -0.00001394, 0.00001622, -0.00001394, 0.00000396, -0.00000876, -0.00000739, 0.00000591, -0.00000366, -0.00000436, -0.00000047, -0.00000228, -0.00000587, -0.00000277, -0.00000504, -0.00000208, 0.00000181, -0.00000982, -0.00000288, 0.00000306, -0.00000540, -0.00000417, -0.00001187, -0.00000335, -0.00000545, -0.00001666, -0.00000918, -0.00000416, -0.00002440, -0.00001996, -0.00003803, -0.00006030, -0.00017460, 0.00015624, 0.00011911, 0.00006499, 0.00004160, 0.00003555, 0.00003194, 0.00006522, 0.00008160, 0.00021298, -0.00012903, -0.00010433, -0.00001216, -0.00002247, -0.00000693, -0.00001166, -0.00000869, 0.00001113, -0.00000453, -0.00001359, 0.00000476, -0.00000375, -0.00000217, 0.00000128, 0.00000021, 0.00000724, -0.00000177, 0.00000206, 0.00000307, 0.00002090, -0.00000086, 0.00000176, -0.00002168, -0.00000253, -0.00000759, -0.00000935, 0.00000073, -0.00000604, 0.00000550, -0.00000884, -0.00000046, -0.00000279, -0.00000300, -0.00000267, 0.00000686, -0.00000567, -0.00000080, -0.00000272, -0.00000270, -0.00000064, -0.00000239, -0.00000335, 0.00000178, -0.00000478, -0.00000005, -0.00000243, 0.00000758, -0.00000554, -0.00000940, -0.00000053, 0.00000987, -0.00000824, -0.00000303, 0.00000349, -0.00000052, -0.00000011, -0.00000201, 0.00000527, -0.00000978, 0.00000933, 0.00000178, -0.00000095, -0.00000597, },
		{ 0.00000579, 0.00000178, 0.00000071, -0.00000325, 0.00000091, -0.00000020, 0.00000083, 0.00000219, -0.00000346, -0.00000318, 0.00000314, -0.00000442, 0.00000843, 0.00000053, -0.00000117, -0.00000037, -0.00000152, -0.00000362, 0.00000878, -0.00000213, 0.00000417, -0.00000469, 0.00000486, 0.00000088, -0.00000792, 0.00001254, -0.00001530, 0.00000811, -0.00001556, 0.00000001, 0.00000188, -0.00000172, -0.00000597, 0.00000080, -0.00000481, -0.00000269, -0.00000524, -0.00000301, -0.00000316, 0.00000364, -0.00001398, 0.00000171, -0.00000093, -0.00000109, -0.00001002, -0.00000673, -0.00000732, -0.00000078, -0.00002354, -0.00000151, -0.00001408, -0.00001411, -0.00003455, -0.00002240, -0.00010713, -0.00011058, 0.00018638, 0.00007285, 0.00007560, 0.00002713, 0.00003911, 0.00002389, 0.00006562, 0.00007368, 0.00020975, -0.00012169, -0.00010164, -0.00001281, -0.00001974, -0.00000791, -0.00000991, -0.00000998, 0.00001413, -0.00000701, -0.00001218, 0.00000431, -0.00000444, -0.00000038, 0.00000177, 0.00000013, 0.00000703, 0.00000017, 0.00000048, 0.00000711, 0.00002016, -0.00000320, -0.00000011, -0.00002566, 0.00000418, -0.00001348, -0.00000336, -0.00000474, -0.00000032, 0.00000050, -0.00000625, -0.00000218, -0.00000154, -0.00000411, 0.00000033, 0.00000412, -0.00000551, -0.00000069, -0.00000250, -0.00000314, 0.00000033, -0.00000379, -0.00000147, -0.00000037, -0.00000353, -0.00000112, -0.00000008, 0.00000656, -0.00000864, -0.00000749, 0.00000228, 0.00000789, -0.00001103, 0.00000126, 0.00000124, 0.00000107, -0.00000269, 0.00000181, 0.00000088, -0.00000675, 0.00001042, -0.00000053, -0.00000130, -0.00000532, },
		{ 0.00000635, 0.00000060, 0.00000014, -0.00000301, 0.00000137, -0.00000069, 0.00000146, 0.00000199, -0.00000518, -0.00000073, 0.00000121, -0.00000218, 0.00000891, -0.00000190, -0.00000032, -0.00000027, -0.00000252, -0.00000188, 0.00000874, -0.00000317, 0.00000525, -0.00000714, 0.00000979, -0.00000629, 0.00000050, 0.00000428, -0.00001086, 0.00000655, -0.00001772, 0.00000495, -0.00000130, -0.00000052, -0.00000648, 0.00000053, -0.00000545, -0.00000113, -0.00000638, -0.00000223, -0.00000256, 0.00000282, -0.00001515, 0.00000383, -0.00000265, 0.00000064, -0.00001345, -0.00000350, -0.00000908, 0.00000061, -0.00002646, 0.00000197, -0.00001967, -0.00000877, -0.00004320, -0.00001619, -0.00014667, -0.00002615, 0.00017840, 0.00004583, 0.00007958, 0.00001781, 0.00003954, 0.00001929, 0.00006282, 0.00006943, 0.00020362, -0.00011238, -0.00010060, -0.00001169, -0.00001886, -0.00000729, -0.00000989, -0.00000905, 0.00001494, -0.00000830, -0.00001116, 0.00000400, -0.00000529, 0.00000231, 0.00000096, 0.00000115, 0.00000598, 0.00000301, -0.00000211, 0.00001343, 0.00001570, -0.00000244, -0.00000632, -0.00002372, 0.00000600, -0.00001572, -0.00000015, -0.00000810, 0.00000400, -0.00000422, -0.00000330, -0.00000413, -0.00000037, -0.00000464, 0.00000283, 0.00000094, -0.00000460, -0.00000083, -0.00000239, -0.00000319, 0.00000068, -0.00000457, -0.00000006, -0.00000248, -0.00000198, -0.00000242, 0.00000301, 0.00000368, -0.00001026, -0.00000532, 0.00000493, 0.00000409, -0.00001119, 0.00000409, -0.00000046, 0.00000197, -0.00000439, 0.00000513, -0.00000411, -0.00000156, 0.00000884, -0.00000133, -0.00000247, -0.00000318, },
		{ 0.00000551, 0.00000007, -0.00000075, -0.00000215, 0.00000122, -0.00000077, 0.00000206, 0.00000123, -0.00000613, 0.00000153, -0.00000113, 0.00000135, 0.00000741, -0.00000308, 0.00000021, -0.00000009, -0.00000368, 0.00000080, 0.00000687, -0.00000216, 0.00000394, -0.00000653, 0.00001147, -0.00001095, 0.00000794, -0.00000534, -0.00000293, 0.00000028, -0.00001493, 0.00000636, -0.00000266, -0.00000059, -0.00000571, -0.00000134, -0.00000403, -0.00000155, -0.00000584, -0.00000288, -0.00000035, -0.00000060, -0.00001301, 0.00000282, -0.00000136, -0.00000108, -0.00001330, -0.00000333, -0.00000771, -0.00000214, -0.00002417, -0.00000054, -0.00001861, -0.00001136, -0.00004194, -0.00002664, -0.00016326, 0.00005787, 0.00014429, 0.00003738, 0.00007522, 0.00001465, 0.00003630, 0.00001813, 0.00005705, 0.00006856, 0.00019472, -0.00010101, -0.00010124, -0.00000892, -0.00001970, -0.00000528, -0.00001125, -0.00000620, 0.00001348, -0.00000802, -0.00001081, 0.00000377, -0.00000583, 0.00000509, -0.00000035, 0.00000253, 0.00000494, 0.00000562, -0.00000411, 0.00002001, 0.00000934, -0.00000044, -0.00001486, -0.00001702, 0.00000290, -0.00001359, -0.00000082, -0.00000796, 0.00000528, -0.00000696, -0.00000137, -0.00000536, 0.00000006, -0.00000396, 0.00000394, -0.00000173, -0.00000357, -0.00000092, -0.00000258, -0.00000268, 0.00000018, -0.00000431, 0.00000026, -0.00000377, -0.00000093, -0.00000312, 0.00000584, -0.00000037, -0.00001039, -0.00000296, 0.00000679, -0.00000065, -0.00000895, 0.00000493, -0.00000095, 0.00000154, -0.00000427, 0.00000653, -0.00000783, 0.00000401, 0.00000563, -0.00000106, -0.00000399, -0.00000004, },
		{ 0.00000382, -0.00000006, -0.00000167, -0.00000103, 0.00000068, -0.00000051, 0.00000254, -0.00000012, -0.00000592, 0.00000279, -0.00000281, 0.00000509, 0.00000457, -0.00000310, 0.00000038, -0.00000002, -0.00000448, 0.00000356, 0.00000404, 0.00000027, 0.00000066, -0.00000286, 0.00000907, -0.00001118, 0.00001143, -0.00001270, 0.00000497, -0.00000794, -0.00000874, 0.00000471, -0.00000233, -0.00000160, -0.00000425, -0.00000402, -0.00000136, -0.00000332, -0.00000420, -0.00000423, 0.00000237, -0.00000530, -0.00000869, -0.00000042, 0.00000205, -0.00000540, -0.00001020, -0.00000563, -0.00000416, -0.00000776, -0.00001803, -0.00000780, -0.00001205, -0.00002070, -0.00003211, -0.00005397, -0.00014813, 0.00012262, 0.00010035, 0.00004072, 0.00006522, 0.00001573, 0.00003100, 0.00001867, 0.00005002, 0.00006930, 0.00018453, -0.00008874, -0.00010252, -0.00000553, -0.00002126, -0.00000289, -0.00001285, -0.00000256, 0.00001054, -0.00000651, -0.00001098, 0.00000342, -0.00000551, 0.00000708, -0.00000122, 0.00000342, 0.00000483, 0.00000691, -0.00000393, 0.00002464, 0.00000331, 0.00000033, -0.00002250, -0.00000855, -0.00000305, -0.00000840, -0.00000458, -0.00000465, 0.00000334, -0.00000713, -0.00000109, -0.00000546, -0.00000039, -0.00000201, 0.00000330, -0.00000330, -0.00000280, -0.00000084, -0.00000300, -0.00000177, -0.00000100, -0.00000308, -0.00000066, -0.00000387, -0.00000084, -0.00000258, 0.00000744, -0.00000463, -0.00000933, -0.00000044, 0.00000726, -0.00000512, -0.00000521, 0.00000406, -0.00000028, -0.00000016, -0.00000220, 0.00000532, -0.00000894, 0.00000821, 0.00000221, -0.00000058, -0.00000506, 0.00000323, },
		{ 0.00000195, -0.00000016, -0.00000227, -0.00000004, 0.00000007, -0.00000004, 0.00000281, -0.00000187, -0.00000452, 0.00000260, -0.00000299, 0.00000790, 0.00000134, -0.00000239, 0.00000036, -0.00000026, -0.00000440, 0.00000545, 0.00000146, 0.00000279, -0.00000316, 0.00000250, 0.00000343, -0.00000689, 0.00000966, -0.00001530, 0.00000961, -0.00001485, -0.00000168, 0.00000146, -0.00000119, -0.00000274, -0.00000309, -0.00000629, 0.00000129, -0.00000527, -0.00000251, -0.00000513, 0.00000408, -0.00000934, -0.00000417, -0.00000405, 0.00000566, -0.00001034, -0.00000603, -0.00000859, -0.00000060, -0.00001351, -0.00001120, -0.00001634, -0.00000391, -0.00003217, -0.00001992, -0.00009071, -0.00010251, 0.00015689, 0.00006113, 0.00004735, 0.00005413, 0.00001787, 0.00002604, 0.00001871, 0.00004389, 0.00006954, 0.00017501, -0.00007718, -0.00010305, -0.00000286, -0.00002232, -0.00000125, -0.00001342, 0.00000045, 0.00000732, -0.00000448, -0.00001136, 0.00000281, -0.00000411, 0.00000778, -0.00000106, 0.00000332, 0.00000607, 0.00000644, -0.00000081, 0.00002586, -0.00000079, -0.00000198, -0.00002635, -0.00000171, -0.00000884, -0.00000268, -0.00000935, 0.00000014, -0.00000067, -0.00000531, -0.00000218, -0.00000467, -0.00000130, 0.00000066, 0.00000124, -0.00000372, -0.00000240, -0.00000071, -0.00000338, -0.00000089, -0.00000236, -0.00000139, -0.00000243, -0.00000295, -0.00000169, -0.00000068, 0.00000718, -0.00000818, -0.00000750, 0.00000214, 0.00000599, -0.00000817, -0.00000125, 0.00000233, 0.00000089, -0.00000239, 0.00000108, 0.00000189, -0.00000705, 0.00000996, -0.00000030, -0.00000061, -0.00000497, 0.00000574, },
		{ 0.00000046, -0.00000049, -0.00000233, 0.00000054, -0.00000036, 0.00000051, 0.00000274, -0.00000357, -0.00000233, 0.00000110, -0.00000136, 0.00000896, -0.00000134, -0.00000150, 0.00000036, -0.00000090, -0.00000321, 0.00000582, 0.00000019, 0.00000404, -0.00000573, 0.00000738, -0.00000321, 0.00000012, 0.00000349, -0.00001269, 0.00000933, -0.00001802, 0.00000386, -0.00000171, -0.00000022, -0.00000326, -0.00000297, -0.00000720, 0.00000282, -0.00000635, -0.00000169, -0.00000472, 0.00000364, -0.00001110, -0.00000131, -0.00000625, 0.00000756, -0.00001377, -0.00000282, -0.00001030, 0.00000082, -0.00001667, -0.00000701, -0.00002228, 0.00000131, -0.00004016, -0.00001361, -0.00012429, -0.00003703, 0.00015954, 0.00003525, 0.00005077, 0.00004576, 0.00001853, 0.00002314, 0.00001677, 0.00004001, 0.00006784, 0.00016753, -0.00006746, -0.00010191, -0.00000176, -0.00002213, -0.00000100, -0.00001220, 0.00000179, 0.00000483, -0.00000255, -0.00001174, 0.00000211, -0.00000189, 0.00000730, 0.00000009, 0.00000243, 0.00000832, 0.00000477, 0.00000481, 0.00002351, -0.00000277, -0.00000759, -0.00002505, 0.00000113, -0.00001194, 0.00000113, -0.00001282, 0.00000425, -0.00000493, -0.00000278, -0.00000381, -0.00000365, -0.00000198, 0.00000317, -0.00000151, -0.00000330, -0.00000222, -0.00000071, -0.00000343, -0.00000046, -0.00000329, 0.00000004, -0.00000433, -0.00000159, -0.00000294, 0.00000217, 0.00000499, -0.00001039, -0.00000527, 0.00000450, 0.00000311, -0.00000910, 0.00000180, 0.00000072, 0.00000171, -0.00000405, 0.00000426, -0.00000250, -0.00000289, 0.00000917, -0.00000142, -0.00000145, -0.00000344, 0.00000691, },
		{ -0.00000045, -0.00000111, -0.00000186, 0.00000064, -0.00000050, 0.00000106, 0.00000215, -0.00000470, -0.00000009, -0.00000104, 0.00000171, 0.00000805, -0.00000290, -0.00000084, 0.00000046, -0.00000178, -0.00000114, 0.00000460, 0.00000063, 0.00000328, -0.00000574, 0.00000971, -0.00000821, 0.00000698, -0.00000454, -0.00000651, 0.00000455, -0.00001669, 0.00000650, -0.00000370, -0.00000001, -0.00000291, -0.00000406, -0.00000650, 0.00000281, -0.00000615, -0.00000198, -0.00000296, 0.00000093, -0.00001003, -0.00000095, -0.00000613, 0.00000675, -0.00001446, -0.00000181, -0.00000975, -0.00000090, -0.00001588, -0.00000729, -0.00002334, 0.00000076, -0.00004098, -0.00001954, -0.00014182, 0.00003237, 0.00013812, 0.00002424, 0.00004837, 0.00004160, 0.00001685, 0.00002256, 0.00001281, 0.00003836, 0.00006405, 0.00016233, -0.00005973, -0.00009903, -0.00000228, -0.00002071, -0.00000199, -0.00000926, 0.00000122, 0.00000356, -0.00000107, -0.00001211, 0.00000171, 0.00000046, 0.00000630, 0.00000162, 0.00000147, 0.00001062, 0.00000320, 0.00001141, 0.00001866, -0.00000371, -0.00001498, -0.00001935, -0.00000043, -0.00001131, 0.00000174, -0.00001351, 0.00000605, -0.00000783, -0.00000081, -0.00000510, -0.00000299, -0.00000181, 0.00000470, -0.00000414, -0.00000249, -0.00000210, -0.00000099, -0.00000299, -0.00000076, -0.00000338, 0.00000060, -0.00000565, -0.00000052, -0.00000384, 0.00000510, 0.00000134, -0.00001103, -0.00000290, 0.00000624, -0.00000074, -0.00000787, 0.00000330, -0.00000011, 0.00000155, -0.00000427, 0.00000600, -0.00000621, 0.00000206, 0.00000662, -0.00000140, -0.00000282, -0.00000075, 0.00000665, },
		{ -0.00000083, -0.00000185, -0.00000108, 0.00000040, -0.00000036, 0.00000156, 0.00000099, -0.00000484, 0.00000143, -0.00000285, 0.00000531, 0.00000560, -0.00000325, -0.00000055, 0.00000058, -0.00000255, 0.00000122, 0.00000239, 0.00000239, 0.00000070, -0.00000295, 0.00000846, -0.00000963, 0.00001096, -0.00001130, 0.00000032, -0.00000261, -0.00001180, 0.00000618, -0.00000425, -0.00000054, -0.00000200, -0.00000583, -0.00000470, 0.00000163, -0.00000501, -0.00000289, -0.00000064, -0.00000314, -0.00000680, -0.00000269, -0.00000408, 0.00000358, -0.00001255, -0.00000290, -0.00000734, -0.00000510, -0.00001178, -0.00001160, -0.00001981, -0.00000534, -0.00003485, -0.00003901, -0.00013490, 0.00008989, 0.00010484, 0.00002421, 0.00004117, 0.00004083, 0.00001363, 0.00002324, 0.00000795, 0.00003780, 0.00005919, 0.00015863, -0.00005328, -0.00009516, -0.00000374, -0.00001873, -0.00000341, -0.00000543, -0.00000079, 0.00000341, -0.00000005, -0.00001258, 0.00000204, 0.00000221, 0.00000556, 0.00000279, 0.00000128, 0.00001192, 0.00000323, 0.00001708, 0.00001292, -0.00000536, -0.00002159, -0.00001166, -0.00000494, -0.00000772, -0.00000056, -0.00001135, 0.00000507, -0.00000860, -0.00000017, -0.00000552, -0.00000297, -0.00000064, 0.00000482, -0.00000604, -0.00000166, -0.00000197, -0.00000146, -0.00000215, -0.00000169, -0.00000260, 0.00000009, -0.00000594, -0.00000025, -0.00000371, 0.00000715, -0.00000291, -0.00001023, -0.00000054, 0.00000690, -0.00000463, -0.00000510, 0.00000326, -0.00000002, 0.00000034, -0.00000277, 0.00000553, -0.00000786, 0.00000619, 0.00000352, -0.00000093, -0.00000404, 0.00000243, 0.00000530, },
		{ -0.00000094, -0.00000248, -0.00000028, 0.00000003, -0.00000003, 0.00000192, -0.00000060, -0.00000385, 0.00000170, -0.00000345, 0.00000832, 0.00000248, -0.00000270, -0.00000055, 0.00000056, -0.00000276, 0.00000307, 0.00000017, 0.00000443, -0.00000258, 0.00000163, 0.00000408, -0.00000701, 0.00001058, -0.00001436, 0.00000488, -0.00000930, -0.00000537, 0.00000389, -0.00000380, -0.00000133, -0.00000119, -0.00000742, -0.00000274, 0.00000015, -0.00000372, -0.00000352, 0.00000100, -0.00000699, -0.00000294, -0.00000519, -0.00000147, -0.00000054, -0.00000928, -0.00000493, -0.00000454, -0.00000986, -0.00000658, -0.00001757, -0.00001431, -0.00001392, -0.00002594, -0.00006724, -0.00010272, 0.00012486, 0.00007178, 0.00002893, 0.00003218, 0.00004144, 0.00001044, 0.00002364, 0.00000376, 0.00003674, 0.00005468, 0.00015520, -0.00004701, -0.00009135, -0.00000517, -0.00001708, -0.00000424, -0.00000183, -0.00000335, 0.00000396, 0.00000064, -0.00001318, 0.00000327, 0.00000292, 0.00000562, 0.00000314, 0.00000234, 0.00001169, 0.00000580, 0.00002025, 0.00000762, -0.00000901, -0.00002498, -0.00000494, -0.00000991, -0.00000312, -0.00000421, -0.00000751, 0.00000202, -0.00000747, -0.00000085, -0.00000507, -0.00000341, 0.00000124, 0.00000362, -0.00000700, -0.00000097, -0.00000190, -0.00000191, -0.00000126, -0.00000287, -0.00000133, -0.00000126, -0.00000526, -0.00000087, -0.00000229, 0.00000757, -0.00000681, -0.00000839, 0.00000166, 0.00000622, -0.00000760, -0.00000175, 0.00000223, 0.00000059, -0.00000141, -0.00000005, 0.00000296, -0.00000691, 0.00000835, 0.00000100, -0.00000077, -0.00000440, 0.00000522, 0.00000344, },
		{ -0.00000107, -0.00000279, 0.00000033, -0.00000027, 0.00000038, 0.00000199, -0.00000228, -0.00000200, 0.00000070, -0.00000240, 0.00000986, -0.00000040, -0.00000179, -0.00000062, 0.00000026, -0.00000217, 0.00000382, -0.00000117, 0.00000560, -0.00000504, 0.00000620, -0.00000170, -0.00000158, 0.00000619, -0.00001295, 0.00000536, -0.00001315, 0.00000041, 0.00000099, -0.00000306, -0.00000188, -0.00000105, -0.00000807, -0.00000152, -0.00000079, -0.00000302, -0.00000315, 0.00000095, -0.00000919, -0.00000003, -0.00000697, 0.00000015, -0.00000390, -0.00000626, -0.00000643, -0.00000300, -0.00001307, -0.00000288, -0.00002223, -0.00001026, -0.00002073, -0.00002047, -0.00009492, -0.00005218, 0.00013424, 0.00004720, 0.00003288, 0.00002437, 0.00004150, 0.00000854, 0.00002263, 0.00000131, 0.00003406, 0.00005164, 0.00015109, -0.00004008, -0.00008842, -0.00000589, -0.00001635, -0.00000379, 0.00000061, -0.00000558, 0.00000469, 0.00000111, -0.00001371, 0.00000515, 0.00000266, 0.00000647, 0.00000279, 0.00000444, 0.00001019, 0.00001087, 0.00002030, 0.00000319, -0.00001481, -0.00002394, -0.00000140, -0.00001302, 0.00000034, -0.00000727, -0.00000376, -0.00000167, -0.00000539, -0.00000229, -0.00000418, -0.00000381, 0.00000319, 0.00000158, -0.00000711, -0.00000043, -0.00000203, -0.00000207, -0.00000071, -0.00000378, -0.00000016, -0.00000287, -0.00000405, -0.00000202, 0.00000021, 0.00000607, -0.00000959, -0.00000596, 0.00000353, 0.00000425, -0.00000900, 0.00000119, 0.00000097, 0.00000108, -0.00000284, 0.00000286, -0.00000078, -0.00000374, 0.00000823, -0.00000034, -0.00000133, -0.00000345, 0.00000694, 0.00000167, },
		{ -0.00000135, -0.00000273, 0.00000063, -0.00000039, 0.00000083, 0.00000164, -0.00000357, 0.00000011, -0.00000109, 0.00000014, 0.00000953, -0.00000235, -0.00000097, -0.00000060, -0.00000030, -0.00000083, 0.00000331, -0.00000116, 0.00000514, -0.00000543, 0.00000888, -0.00000662, 0.00000436, -0.00000033, -0.00000813, 0.00000175, -0.00001307, 0.00000399, -0.00000138, -0.00000260, -0.00000192, -0.00000174, -0.00000752, -0.00000142, -0.00000082, -0.00000315, -0.00000167, -0.00000105, -0.00000903, 0.00000100, -0.00000712, -0.00000022, -0.00000531, -0.00000466, -0.00000646, -0.00000368, -0.00001345, -0.00000234, -0.00002353, -0.00001013, -0.00002264, -0.00002373, -0.00011169, 0.00000477, 0.00012225, 0.00003398, 0.00003314, 0.00001930, 0.00004015, 0.00000825, 0.00002006, 0.00000071, 0.00002955, 0.00005034, 0.00014603, -0.00003232, -0.00008656, -0.00000575, -0.00001652, -0.00000198, 0.00000152, -0.00000695, 0.00000528, 0.00000139, -0.00001385, 0.00000712, 0.00000196, 0.00000764, 0.00000233, 0.00000682, 0.00000845, 0.00001746, 0.00001756, -0.00000086, -0.00002151, -0.00001903, -0.00000167, -0.00001312, 0.00000133, -0.00000831, -0.00000161, -0.00000456, -0.00000349, -0.00000372, -0.00000332, -0.00000372, 0.00000454, -0.00000067, -0.00000665, 0.00000000, -0.00000242, -0.00000179, -0.00000075, -0.00000403, 0.00000037, -0.00000408, -0.00000296, -0.00000304, 0.00000307, 0.00000291, -0.00001079, -0.00000335, 0.00000482, 0.00000140, -0.00000863, 0.00000308, 0.00000007, 0.00000092, -0.00000317, 0.00000477, -0.00000431, 0.00000050, 0.00000631, -0.00000053, -0.00000250, -0.00000126, 0.00000730, 0.00000035, },
		{ -0.00000178, -0.00000237, 0.00000068, -0.00000033, 0.00000127, 0.00000076, -0.00000405, 0.00000175, -0.00000283, 0.00000345, 0.00000757, -0.00000310, -0.00000050, -0.00000047, -0.00000090, 0.00000087, 0.00000190, -0.00000002, 0.00000309, -0.00000337, 0.00000857, -0.00000880, 0.00000841, -0.00000636, -0.00000222, -0.00000442, -0.00000947, 0.00000491, -0.00000269, -0.00000258, -0.00000160, -0.00000294, -0.00000610, -0.00000223, -0.00000013, -0.00000377, 0.00000034, -0.00000437, -0.00000687, 0.00000020, -0.00000570, -0.00000253, -0.00000459, -0.00000468, -0.00000507, -0.00000632, -0.00001118, -0.00000499, -0.00002130, -0.00001422, -0.00001928, -0.00003748, -0.00011004, 0.00005531, 0.00009770, 0.00003040, 0.00002961, 0.00001690, 0.00003768, 0.00000893, 0.00001668, 0.00000116, 0.00002392, 0.00005024, 0.00014051, -0.00002419, -0.00008536, -0.00000515, -0.00001707, 0.00000065, 0.00000113, -0.00000736, 0.00000557, 0.00000145, -0.00001327, 0.00000857, 0.00000148, 0.00000848, 0.00000254, 0.00000852, 0.00000778, 0.00002401, 0.00001293, -0.00000550, -0.00002709, -0.00001223, -0.00000474, -0.00001056, -0.00000018, -0.00000699, -0.00000168, -0.00000582, -0.00000255, -0.00000460, -0.00000282, -0.00000290, 0.00000484, -0.00000257, -0.00000587, 0.00000031, -0.00000298, -0.00000116, -0.00000135, -0.00000354, -0.00000000, -0.00000447, -0.00000249, -0.00000325, 0.00000542, -0.00000122, -0.00001036, -0.00000091, 0.00000530, -0.00000168, -0.00000681, 0.00000374, -0.00000025, -0.00000000, -0.00000208, 0.00000490, -0.00000635, 0.00000438, 0.00000362, -0.00000013, -0.00000378, 0.00000167, 0.00000642, -0.00000040, },
		{ -0.00000226, -0.00000188, 0.00000059, -0.00000014, 0.00000163, -0.00000057, -0.00000351, 0.00000238, -0.00000368, 0.00000652, 0.00000465, -0.00000284, -0.00000038, -0.00000032, -0.00000127, 0.00000238, 0.00000028, 0.00000149, 0.00000030, 0.00000046, 0.00000536, -0.00000757, 0.00000906, -0.00000969, 0.00000222, -0.00001076, -0.00000389, 0.00000367, -0.00000300, -0.00000282, -0.00000131, -0.00000405, -0.00000450, -0.00000333, 0.00000072, -0.00000420, 0.00000191, -0.00000778, -0.00000379, -0.00000155, -0.00000360, -0.00000588, -0.00000247, -0.00000569, -0.00000317, -0.00000967, -0.00000772, -0.00000931, -0.00001717, -0.00002058, -0.00001336, -0.00005875, -0.00008806, 0.00008981, 0.00007024, 0.00003220, 0.00002399, 0.00001602, 0.00003503, 0.00000953, 0.00001362, 0.00000149, 0.00001827, 0.00005038, 0.00013536, -0.00001649, -0.00008414, -0.00000473, -0.00001724, 0.00000325, 0.00000004, -0.00000704, 0.00000558, 0.00000137, -0.00001187, 0.00000913, 0.00000167, 0.00000859, 0.00000390, 0.00000891, 0.00000921, 0.00002904, 0.00000730, -0.00001135, -0.00002964, -0.00000606, -0.00000861, -0.00000680, -0.00000312, -0.00000407, -0.00000361, -0.00000542, -0.00000276, -0.00000474, -0.00000266, -0.00000149, 0.00000404, -0.00000383, -0.00000498, 0.00000040, -0.00000349, -0.00000044, -0.00000221, -0.00000257, -0.00000113, -0.00000399, -0.00000283, -0.00000229, 0.00000643, -0.00000540, -0.00000859, 0.00000114, 0.00000485, -0.00000428, -0.00000422, 0.00000344, -0.00000016, -0.00000136, 0.00000014, 0.00000316, -0.00000624, 0.00000679, 0.00000119, 0.00000017, -0.00000447, 0.00000458, 0.00000474, -0.00000070, },
		{ -0.00000265, -0.00000140, 0.00000050, 0.00000014, 0.00000178, -0.00000208, -0.00000210, 0.00000184, -0.00000308, 0.00000842, 0.00000167, -0.00000201, -0.00000045, -0.00000028, -0.00000117, 0.00000322, -0.00000082, 0.00000244, -0.00000201, 0.00000463, 0.00000051, -0.00000361, 0.00000624, -0.00000937, 0.00000342, -0.00001506, 0.00000177, 0.00000130, -0.00000274, -0.00000302, -0.00000143, -0.00000451, -0.00000337, -0.00000410, 0.00000115, -0.00000385, 0.00000213, -0.00001001, -0.00000112, -0.00000310, -0.00000201, -0.00000893, -0.00000016, -0.00000664, -0.00000193, -0.00001216, -0.00000497, -0.00001314, -0.00001363, -0.00002614, -0.00000943, -0.00008078, -0.00004995, 0.00010427, 0.00004731, 0.00003503, 0.00001841, 0.00001534, 0.00003306, 0.00000923, 0.00001169, 0.00000082, 0.00001352, 0.00004995, 0.00013128, -0.00000991, -0.00008230, -0.00000496, -0.00001642, 0.00000502, -0.00000102, -0.00000631, 0.00000532, 0.00000134, -0.00000989, 0.00000887, 0.00000253, 0.00000803, 0.00000632, 0.00000805, 0.00001308, 0.00003156, 0.00000107, -0.00001820, -0.00002828, -0.00000247, -0.00001128, -0.00000361, -0.00000593, -0.00000091, -0.00000633, -0.00000402, -0.00000379, -0.00000433, -0.00000257, 0.00000009, 0.00000239, -0.00000437, -0.00000405, 0.00000018, -0.00000370, 0.00000000, -0.00000290, -0.00000161, -0.00000256, -0.00000296, -0.00000376, -0.00000022, 0.00000563, -0.00000874, -0.00000597, 0.00000264, 0.00000353, -0.00000587, -0.00000156, 0.00000265, -0.00000006, -0.00000253, 0.00000268, 0.00000021, -0.00000410, 0.00000725, -0.00000029, -0.00000012, -0.00000404, 0.00000674, 0.00000283, -0.00000076, },
		{ -0.00000285, -0.00000103, 0.00000047, 0.00000047, 0.00000158, -0.00000336, -0.00000029, 0.00000044, -0.00000103, 0.00000863, -0.00000061, -0.00000111, -0.00000050, -0.00000041, -0.00000057, 0.00000319, -0.00000099, 0.00000218, -0.00000273, 0.00000752, -0.00000412, 0.00000132, 0.00000126, -0.00000597, 0.00000100, -0.00001602, 0.00000591, -0.00000112, -0.00000235, -0.00000302, -0.00000207, -0.00000412, -0.00000305, -0.00000420, 0.00000094, -0.00000256, 0.00000068, -0.00001034, 0.00000023, -0.00000360, -0.00000182, -0.00001065, 0.00000136, -0.00000675, -0.00000214, -0.00001274, -0.00000434, -0.00001480, -0.00001266, -0.00002841, -0.00001171, -0.00009551, -0.00000444, 0.00010053, 0.00003247, 0.00003603, 0.00001429, 0.00001412, 0.00003204, 0.00000788, 0.00001101, -0.00000106, 0.00000996, 0.00004865, 0.00012851, -0.00000471, -0.00007965, -0.00000592, -0.00001448, 0.00000555, -0.00000155, -0.00000551, 0.00000480, 0.00000165, -0.00000779, 0.00000821, 0.00000369, 0.00000730, 0.00000922, 0.00000665, 0.00001889, 0.00003126, -0.00000582, -0.00002487, -0.00002359, -0.00000212, -0.00001161, -0.00000222, -0.00000735, 0.00000122, -0.00000867, -0.00000252, -0.00000507, -0.00000373, -0.00000224, 0.00000135, 0.00000033, -0.00000428, -0.00000318, -0.00000038, -0.00000348, -0.00000005, -0.00000308, -0.00000113, -0.00000373, -0.00000192, -0.00000473, 0.00000246, 0.00000305, -0.00001060, -0.00000306, 0.00000350, 0.00000158, -0.00000622, 0.00000057, 0.00000182, -0.00000031, -0.00000288, 0.00000459, -0.00000287, -0.00000077, 0.00000602, -0.00000071, -0.00000106, -0.00000238, 0.00000766, 0.00000118, -0.00000079, },
		{ -0.00000286, -0.00000076, 0.00000050, 0.00000085, 0.00000091, -0.00000401, 0.00000130, -0.00000114, 0.00000194, 0.00000718, -0.00000180, -0.00000048, -0.00000040, -0.00000069, 0.00000037, 0.00000248, -0.00000033, 0.00000067, -0.00000142, 0.00000802, -0.00000685, 0.00000523, -0.00000381, -0.00000126, -0.00000400, -0.00001363, 0.00000777, -0.00000290, -0.00000210, -0.00000285, -0.00000303, -0.00000310, -0.00000342, -0.00000375, 0.00000034, -0.00000075, -0.00000204, -0.00000883, 0.00000006, -0.00000291, -0.00000320, -0.00001066, 0.00000169, -0.00000587, -0.00000380, -0.00001136, -0.00000605, -0.00001392, -0.00001473, -0.00002682, -0.00002192, -0.00009654, 0.00003837, 0.00008459, 0.00002555, 0.00003438, 0.00001196, 0.00001238, 0.00003158, 0.00000597, 0.00001106, -0.00000367, 0.00000723, 0.00004678, 0.00012674, -0.00000067, -0.00007639, -0.00000725, -0.00001174, 0.00000489, -0.00000131, -0.00000483, 0.00000407, 0.00000255, -0.00000607, 0.00000769, 0.00000461, 0.00000708, 0.00001177, 0.00000580, 0.00002555, 0.00002826, -0.00001359, -0.00002967, -0.00001731, -0.00000430, -0.00000970, -0.00000285, -0.00000694, 0.00000166, -0.00000983, -0.00000165, -0.00000607, -0.00000320, -0.00000148, 0.00000191, -0.00000171, -0.00000369, -0.00000250, -0.00000112, -0.00000291, -0.00000060, -0.00000264, -0.00000138, -0.00000423, -0.00000136, -0.00000511, 0.00000496, -0.00000081, -0.00001070, -0.00000039, 0.00000371, -0.00000062, -0.00000541, 0.00000192, 0.00000117, -0.00000100, -0.00000214, 0.00000513, -0.00000498, 0.00000258, 0.00000386, -0.00000045, -0.00000227, 0.00000023, 0.00000726, 0.00000006, -0.00000094, },
		{ -0.00000270, -0.00000057, 0.00000059, 0.00000120, -0.00000020, -0.00000379, 0.00000211, -0.00000211, 0.00000498, 0.00000457, -0.00000190, -0.00000023, -0.00000013, -0.00000099, 0.00000132, 0.00000151, 0.00000066, -0.00000153, 0.00000152, 0.00000598, -0.00000685, 0.00000668, -0.00000707, 0.00000261, -0.00000965, -0.00000895, 0.00000743, -0.00000380, -0.00000202, -0.00000271, -0.00000389, -0.00000194, -0.00000402, -0.00000318, -0.00000013, 0.00000077, -0.00000508, -0.00000624, -0.00000112, -0.00000155, -0.00000557, -0.00000934, 0.00000110, -0.00000449, -0.00000614, -0.00000890, -0.00000919, -0.00001149, -0.00001864, -0.00002314, -0.00003827, -0.00008153, 0.00007022, 0.00006385, 0.00002401, 0.00003079, 0.00001091, 0.00001059, 0.00003102, 0.00000427, 0.00001099, -0.00000619, 0.00000465, 0.00004496, 0.00012535, 0.00000273, -0.00007296, -0.00000842, -0.00000882, 0.00000342, -0.00000034, -0.00000439, 0.00000325, 0.00000406, -0.00000502, 0.00000767, 0.00000494, 0.00000783, 0.00001331, 0.00000655, 0.00003179, 0.00002280, -0.00002210, -0.00003123, -0.00001160, -0.00000738, -0.00000661, -0.00000479, -0.00000515, 0.00000055, -0.00000966, -0.00000170, -0.00000652, -0.00000283, -0.00000032, 0.00000164, -0.00000342, -0.00000271, -0.00000216, -0.00000180, -0.00000224, -0.00000141, -0.00000179, -0.00000227, -0.00000398, -0.00000153, -0.00000447, 0.00000644, -0.00000513, -0.00000917, 0.00000164, 0.00000335, -0.00000265, -0.00000379, 0.00000250, 0.00000069, -0.00000194, -0.00000047, 0.00000414, -0.00000541, 0.00000492, 0.00000170, -0.00000011, -0.00000317, 0.00000314, 0.00000583, -0.00000045, -0.00000122, },
		{ -0.00000242, -0.00000042, 0.00000071, 0.00000143, -0.00000155, -0.00000274, 0.00000192, -0.00000189, 0.00000719, 0.00000160, -0.00000125, -0.00000027, 0.00000020, -0.00000116, 0.00000198, 0.00000077, 0.00000130, -0.00000346, 0.00000501, 0.00000223, -0.00000441, 0.00000534, -0.00000750, 0.00000400, -0.00001395, -0.00000359, 0.00000559, -0.00000395, -0.00000203, -0.00000282, -0.00000429, -0.00000112, -0.00000440, -0.00000292, 0.00000003, 0.00000125, -0.00000736, -0.00000361, -0.00000247, -0.00000038, -0.00000800, -0.00000750, 0.00000026, -0.00000337, -0.00000809, -0.00000670, -0.00001227, -0.00000920, -0.00002227, -0.00002062, -0.00005593, -0.00005289, 0.00008683, 0.00004464, 0.00002463, 0.00002658, 0.00001039, 0.00000923, 0.00002983, 0.00000333, 0.00001018, -0.00000793, 0.00000156, 0.00004377, 0.00012372, 0.00000600, -0.00006975, -0.00000894, -0.00000638, 0.00000168, 0.00000114, -0.00000426, 0.00000258, 0.00000597, -0.00000456, 0.00000816, 0.00000471, 0.00000960, 0.00001369, 0.00000950, 0.00003644, 0.00001505, -0.00003054, -0.00002915, -0.00000809, -0.00000972, -0.00000373, -0.00000687, -0.00000294, -0.00000137, -0.00000855, -0.00000255, -0.00000644, -0.00000254, 0.00000099, 0.00000062, -0.00000460, -0.00000151, -0.00000226, -0.00000217, -0.00000177, -0.00000212, -0.00000090, -0.00000344, -0.00000320, -0.00000229, -0.00000273, 0.00000630, -0.00000896, -0.00000648, 0.00000285, 0.00000257, -0.00000420, -0.00000177, 0.00000249, 0.00000022, -0.00000268, 0.00000154, 0.00000200, -0.00000412, 0.00000571, 0.00000021, -0.00000019, -0.00000319, 0.00000562, 0.00000385, -0.00000050, -0.00000162, },
		{ -0.00000204, -0.00000030, 0.00000089, 0.00000139, -0.00000282, -0.00000119, 0.00000087, -0.00000034, 0.00000793, -0.00000096, -0.00000033, -0.00000039, 0.00000047, -0.00000113, 0.00000221, 0.00000056, 0.00000110, -0.00000421, 0.00000771, -0.00000179, -0.00000073, 0.00000201, -0.00000533, 0.00000232, -0.00001554, 0.00000094, 0.00000315, -0.00000365, -0.00000207, -0.00000319, -0.00000413, -0.00000087, -0.00000430, -0.00000315, 0.00000097, 0.00000033, -0.00000816, -0.00000179, -0.00000329, -0.00000011, -0.00000960, -0.00000596, -0.00000025, -0.00000310, -0.00000888, -0.00000579, -0.00001405, -0.00000851, -0.00002379, -0.00002244, -0.00006873, -0.00001681, 0.00008845, 0.00003067, 0.00002493, 0.00002285, 0.00000990, 0.00000843, 0.00002793, 0.00000325, 0.00000846, -0.00000864, -0.00000229, 0.00004345, 0.00012154, 0.00000941, -0.00006688, -0.00000861, -0.00000484, 0.00000018, 0.00000288, -0.00000442, 0.00000229, 0.00000788, -0.00000436, 0.00000885, 0.00000433, 0.00001197, 0.00001332, 0.00001464, 0.00003858, 0.00000515, -0.00003748, -0.00002424, -0.00000733, -0.00001033, -0.00000208, -0.00000810, -0.00000131, -0.00000322, -0.00000716, -0.00000380, -0.00000598, -0.00000217, 0.00000215, -0.00000090, -0.00000517, -0.00000026, -0.00000279, -0.00000211, -0.00000172, -0.00000244, -0.00000038, -0.00000445, -0.00000234, -0.00000322, -0.00000022, 0.00000435, -0.00001147, -0.00000331, 0.00000325, 0.00000153, -0.00000509, 0.00000025, 0.00000208, -0.00000038, -0.00000282, 0.00000318, -0.00000046, -0.00000167, 0.00000501, -0.00000036, -0.00000088, -0.00000203, 0.00000708, 0.00000186, -0.00000029, -0.00000206, },
		{ -0.00000160, -0.00000022, 0.00000113, 0.00000097, -0.00000364, 0.00000032, -0.00000050, 0.00000223, 0.00000706, -0.00000256, 0.00000044, -0.00000036, 0.00000057, -0.00000090, 0.00000205, 0.00000083, -0.00000002, -0.00000334, 0.00000863, -0.00000467, 0.00000263, -0.00000178, -0.00000183, -0.00000180, -0.00001407, 0.00000365, 0.00000089, -0.00000317, -0.00000215, -0.00000366, -0.00000356, -0.00000109, -0.00000382, -0.00000367, 0.00000239, -0.00000177, -0.00000737, -0.00000112, -0.00000331, -0.00000100, -0.00000996, -0.00000514, -0.00000021, -0.00000373, -0.00000837, -0.00000642, -0.00001418, -0.00000983, -0.00002270, -0.00003008, -0.00007152, 0.00001900, 0.00007889, 0.00002279, 0.00002374, 0.00002010, 0.00000932, 0.00000796, 0.00002564, 0.00000365, 0.00000618, -0.00000858, -0.00000668, 0.00004382, 0.00011884, 0.00001295, -0.00006422, -0.00000758, -0.00000429, -0.00000078, 0.00000460, -0.00000480, 0.00000256, 0.00000941, -0.00000401, 0.00000933, 0.00000435, 0.00001426, 0.00001302, 0.00002145, 0.00003754, -0.00000647, -0.00004141, -0.00001825, -0.00000867, -0.00000922, -0.00000192, -0.00000804, -0.00000083, -0.00000433, -0.00000611, -0.00000503, -0.00000534, -0.00000164, 0.00000291, -0.00000266, -0.00000509, 0.00000077, -0.00000355, -0.00000170, -0.00000210, -0.00000228, -0.00000042, -0.00000493, -0.00000182, -0.00000379, 0.00000242, 0.00000086, -0.00001216, -0.00000039, 0.00000304, 0.00000040, -0.00000525, 0.00000201, 0.00000143, -0.00000111, -0.00000218, 0.00000384, -0.00000237, 0.00000107, 0.00000335, -0.00000023, -0.00000198, 0.00000017, 0.00000724, 0.00000031, -0.00000008, -0.00000245, },
		{ -0.00000114, -0.00000019, 0.00000140, 0.00000014, -0.00000374, 0.00000131, -0.00000154, 0.00000511, 0.00000488, -0.00000307, 0.00000082, -0.00000006, 0.00000046, -0.00000054, 0.00000172, 0.00000127, -0.00000165, -0.00000108, 0.00000745, -0.00000554, 0.00000442, -0.00000448, 0.00000130, -0.00000687, -0.00001018, 0.00000428, -0.00000072, -0.00000271, -0.00000235, -0.00000399, -0.00000294, -0.00000146, -0.00000329, -0.00000407, 0.00000366, -0.00000433, -0.00000545, -0.00000138, -0.00000279, -0.00000276, -0.00000917, -0.00000505, 0.00000020, -0.00000486, -0.00000711, -0.00000807, -0.00001324, -0.00001241, -0.00002018, -0.00004240, -0.00006197, 0.00004773, 0.00006353, 0.00001981, 0.00002105, 0.00001827, 0.00000883, 0.00000741, 0.00002348, 0.00000395, 0.00000396, -0.00000830, -0.00001113, 0.00004441, 0.00011593, 0.00001643, -0.00006144, -0.00000626, -0.00000455, -0.00000109, 0.00000612, -0.00000519, 0.00000334, 0.00001039, -0.00000325, 0.00000932, 0.00000522, 0.00001587, 0.00001373, 0.00002898, 0.00003274, -0.00001882, -0.00004141, -0.00001308, -0.00001077, -0.00000718, -0.00000278, -0.00000697, -0.00000145, -0.00000451, -0.00000569, -0.00000596, -0.00000462, -0.00000096, 0.00000313, -0.00000440, -0.00000439, 0.00000135, -0.00000424, -0.00000118, -0.00000269, -0.00000179, -0.00000097, -0.00000478, -0.00000191, -0.00000352, 0.00000438, -0.00000349, -0.00001094, 0.00000167, 0.00000256, -0.00000074, -0.00000473, 0.00000333, 0.00000060, -0.00000179, -0.00000088, 0.00000335, -0.00000310, 0.00000321, 0.00000147, 0.00000012, -0.00000298, 0.00000296, 0.00000617, -0.00000056, -0.00000005, -0.00000271, },
		{ -0.00000069, -0.00000018, 0.00000161, -0.00000097, -0.00000312, 0.00000148, -0.00000164, 0.00000746, 0.00000206, -0.00000266, 0.00000083, 0.00000049, 0.00000014, -0.00000014, 0.00000143, 0.00000147, -0.00000312, 0.00000180, 0.00000465, -0.00000439, 0.00000415, -0.00000516, 0.00000258, -0.00001117, -0.00000515, 0.00000318, -0.00000153, -0.00000235, -0.00000270, -0.00000396, -0.00000259, -0.00000164, -0.00000303, -0.00000394, 0.00000411, -0.00000647, -0.00000317, -0.00000205, -0.00000228, -0.00000473, -0.00000775, -0.00000534, 0.00000058, -0.00000586, -0.00000591, -0.00000978, -0.00001233, -0.00001482, -0.00001857, -0.00005591, -0.00004138, 0.00006533, 0.00004748, 0.00001962, 0.00001747, 0.00001707, 0.00000858, 0.00000647, 0.00002180, 0.00000372, 0.00000231, -0.00000831, -0.00001513, 0.00004467, 0.00011320, 0.00001959, -0.00005820, -0.00000512, -0.00000525, -0.00000082, 0.00000728, -0.00000536, 0.00000439, 0.00001092, -0.00000211, 0.00000885, 0.00000703, 0.00001649, 0.00001623, 0.00003607, 0.00002384, -0.00003024, -0.00003765, -0.00001006, -0.00001231, -0.00000523, -0.00000383, -0.00000553, -0.00000271, -0.00000396, -0.00000594, -0.00000645, -0.00000386, -0.00000024, 0.00000281, -0.00000589, -0.00000318, 0.00000131, -0.00000458, -0.00000085, -0.00000320, -0.00000128, -0.00000175, -0.00000416, -0.00000255, -0.00000224, 0.00000498, -0.00000775, -0.00000820, 0.00000258, 0.00000213, -0.00000187, -0.00000361, 0.00000408, -0.00000035, -0.00000216, 0.00000066, 0.00000192, -0.00000252, 0.00000423, 0.00000002, 0.00000020, -0.00000333, 0.00000564, 0.00000426, -0.00000076, -0.00000029, -0.00000274, },
		{ -0.00000034, -0.00000014, 0.00000164, -0.00000208, -0.00000197, 0.00000087, -0.00000057, 0.00000863, -0.00000063, -0.00000179, 0.00000066, 0.00000114, -0.00000033, 0.00000025, 0.00000129, 0.00000113, -0.00000374, 0.00000429, 0.00000125, -0.00000195, 0.00000226, -0.00000380, 0.00000137, -0.00001334, -0.00000036, 0.00000105, -0.00000167, -0.00000218, -0.00000314, -0.00000355, -0.00000264, -0.00000147, -0.00000318, -0.00000313, 0.00000341, -0.00000750, -0.00000128, -0.00000257, -0.00000230, -0.00000627, -0.00000626, -0.00000565, 0.00000059, -0.00000626, -0.00000542, -0.00001078, -0.00001237, -0.00001584, -0.00002014, -0.00006604, -0.00001402, 0.00007111, 0.00003421, 0.00002025, 0.00001371, 0.00001625, 0.00000858, 0.00000515, 0.00002063, 0.00000288, 0.00000140, -0.00000883, -0.00001839, 0.00004422, 0.00011093, 0.00002231, -0.00005433, -0.00000454, -0.00000596, -0.00000016, 0.00000807, -0.00000508, 0.00000538, 0.00001127, -0.00000084, 0.00000823, 0.00000955, 0.00001625, 0.00002098, 0.00004136, 0.00001093, -0.00003878, -0.00003144, -0.00000944, -0.00001256, -0.00000413, -0.00000437, -0.00000441, -0.00000400, -0.00000312, -0.00000665, -0.00000649, -0.00000310, 0.00000040, 0.00000197, -0.00000688, -0.00000175, 0.00000069, -0.00000445, -0.00000089, -0.00000338, -0.00000107, -0.00000240, -0.00000341, -0.00000346, -0.00000008, 0.00000382, -0.00001095, -0.00000470, 0.00000238, 0.00000193, -0.00000295, -0.00000199, 0.00000418, -0.00000130, -0.00000205, 0.00000196, 0.00000010, -0.00000091, 0.00000402, -0.00000071, -0.00000026, -0.00000264, 0.00000756, 0.00000207, -0.00000046, -0.00000079, -0.00000252, },
		{ -0.00000013, -0.00000004, 0.00000137, -0.00000291, -0.00000071, -0.00000015, 0.00000152, 0.00000828, -0.00000258, -0.00000090, 0.00000057, 0.00000169, -0.00000084, 0.00000066, 0.00000126, 0.00000022, -0.00000318, 0.00000555, -0.00000165, 0.00000068, -0.00000025, -0.00000127, -0.00000203, -0.00001282, 0.00000310, -0.00000132, -0.00000135, -0.00000223, -0.00000351, -0.00000295, -0.00000298, -0.00000106, -0.00000356, -0.00000187, 0.00000164, -0.00000720, -0.00000017, -0.00000269, -0.00000310, -0.00000696, -0.00000512, -0.00000579, 0.00000019, -0.00000601, -0.00000579, -0.00001087, -0.00001359, -0.00001522, -0.00002594, -0.00006879, 0.00001438, 0.00006722, 0.00002517, 0.00002046, 0.00001027, 0.00001570, 0.00000863, 0.00000370, 0.00001968, 0.00000172, 0.00000102, -0.00000975, -0.00002096, 0.00004295, 0.00010925, 0.00002464, -0.00004986, -0.00000472, -0.00000633, 0.00000060, 0.00000861, -0.00000427, 0.00000606, 0.00001175, 0.00000026, 0.00000791, 0.00001228, 0.00001564, 0.00002801, 0.00004334, -0.00000508, -0.00004279, -0.00002475, -0.00001047, -0.00001162, -0.00000399, -0.00000416, -0.00000395, -0.00000490, -0.00000240, -0.00000755, -0.00000611, -0.00000239, 0.00000086, 0.00000070, -0.00000716, -0.00000047, -0.00000026, -0.00000394, -0.00000129, -0.00000320, -0.00000129, -0.00000264, -0.00000290, -0.00000418, 0.00000245, 0.00000094, -0.00001234, -0.00000137, 0.00000149, 0.00000196, -0.00000389, -0.00000005, 0.00000360, -0.00000209, -0.00000142, 0.00000262, -0.00000149, 0.00000109, 0.00000293, -0.00000078, -0.00000118, -0.00000084, 0.00000824, 0.00000014, 0.00000001, -0.00000144, -0.00000204, },
		{ -0.00000008, 0.00000014, 0.00000075, -0.00000321, 0.00000025, -0.00000105, 0.00000412, 0.00000655, -0.00000347, -0.00000034, 0.00000082, 0.00000189, -0.00000124, 0.00000108, 0.00000115, -0.00000096, -0.00000153, 0.00000521, -0.00000327, 0.00000252, -0.00000222, 0.00000116, -0.00000652, -0.00000988, 0.00000468, -0.00000333, -0.00000084, -0.00000250, -0.00000363, -0.00000243, -0.00000333, -0.00000068, -0.00000383, -0.00000066, -0.00000066, -0.00000580, 0.00000014, -0.00000247, -0.00000455, -0.00000675, -0.00000448, -0.00000575, -0.00000043, -0.00000541, -0.00000666, -0.00001044, -0.00001542, -0.00001387, -0.00003514, -0.00006223, 0.00003853, 0.00005732, 0.00002017, 0.00001976, 0.00000737, 0.00001539, 0.00000850, 0.00000244, 0.00001855, 0.00000070, 0.00000076, -0.00001069, -0.00002317, 0.00004097, 0.00010813, 0.00002673, -0.00004502, -0.00000564, -0.00000612, 0.00000118, 0.00000915, -0.00000307, 0.00000637, 0.00001248, 0.00000098, 0.00000825, 0.00001471, 0.00001540, 0.00003693, 0.00004049, -0.00002231, -0.00004168, -0.00001946, -0.00001192, -0.00001013, -0.00000445, -0.00000339, -0.00000411, -0.00000525, -0.00000205, -0.00000841, -0.00000538, -0.00000181, 0.00000111, -0.00000087, -0.00000659, 0.00000028, -0.00000116, -0.00000332, -0.00000186, -0.00000278, -0.00000189, -0.00000237, -0.00000287, -0.00000428, 0.00000467, -0.00000319, -0.00001158, 0.00000095, 0.00000051, 0.00000196, -0.00000447, 0.00000192, 0.00000242, -0.00000249, -0.00000045, 0.00000250, -0.00000235, 0.00000284, 0.00000150, -0.00000054, -0.00000220, 0.00000175, 0.00000758, -0.00000109, 0.00000035, -0.00000205, -0.00000138, },
		{ -0.00000014, 0.00000032, -0.00000013, -0.00000291, 0.00000062, -0.00000130, 0.00000650, 0.00000393, -0.00000334, -0.00000023, 0.00000148, 0.00000160, -0.00000136, 0.00000149, 0.00000078, -0.00000195, 0.00000070, 0.00000348, -0.00000338, 0.00000308, -0.00000290, 0.00000227, -0.00001068, -0.00000545, 0.00000442, -0.00000459, -0.00000041, -0.00000291, -0.00000342, -0.00000218, -0.00000347, -0.00000059, -0.00000364, -0.00000006, -0.00000281, -0.00000384, -0.00000006, -0.00000219, -0.00000626, -0.00000589, -0.00000426, -0.00000564, -0.00000097, -0.00000491, -0.00000750, -0.00001019, -0.00001690, -0.00001348, -0.00004513, -0.00004702, 0.00005490, 0.00004525, 0.00001812, 0.00001817, 0.00000507, 0.00001527, 0.00000808, 0.00000152, 0.00001702, 0.00000016, 0.00000020, -0.00001122, -0.00002543, 0.00003854, 0.00010745, 0.00002879, -0.00004014, -0.00000708, -0.00000527, 0.00000139, 0.00000999, -0.00000175, 0.00000647, 0.00001338, 0.00000140, 0.00000941, 0.00001646, 0.00001638, 0.00004676, 0.00003165, -0.00003805, -0.00003624, -0.00001660, -0.00001273, -0.00000885, -0.00000495, -0.00000250, -0.00000460, -0.00000517, -0.00000216, -0.00000903, -0.00000442, -0.00000139, 0.00000106, -0.00000250, -0.00000524, 0.00000027, -0.00000169, -0.00000288, -0.00000233, -0.00000241, -0.00000261, -0.00000175, -0.00000331, -0.00000348, 0.00000584, -0.00000767, -0.00000892, 0.00000176, 0.00000002, 0.00000162, -0.00000439, 0.00000353, 0.00000084, -0.00000237, 0.00000057, 0.00000174, -0.00000233, 0.00000387, 0.00000026, -0.00000042, -0.00000281, 0.00000457, 0.00000580, -0.00000148, 0.00000029, -0.00000241, -0.00000068, },
		{ -0.00000022, 0.00000040, -0.00000109, -0.00000212, 0.00000034, -0.00000057, 0.00000798, 0.00000112, -0.00000252, -0.00000042, 0.00000244, 0.00000078, -0.00000108, 0.00000177, 0.00000007, -0.00000234, 0.00000279, 0.00000097, -0.00000232, 0.00000249, -0.00000216, 0.00000141, -0.00001324, -0.00000075, 0.00000278, -0.00000502, -0.00000024, -0.00000328, -0.00000299, -0.00000223, -0.00000332, -0.00000086, -0.00000287, -0.00000038, -0.00000419, -0.00000188, -0.00000046, -0.00000218, -0.00000779, -0.00000473, -0.00000433, -0.00000554, -0.00000129, -0.00000479, -0.00000794, -0.00001058, -0.00001732, -0.00001564, -0.00005254, -0.00002612, 0.00006237, 0.00003397, 0.00001766, 0.00001599, 0.00000340, 0.00001515, 0.00000749, 0.00000087, 0.00001510, 0.00000021, -0.00000085, -0.00001105, -0.00002813, 0.00003593, 0.00010710, 0.00003098, -0.00003554, -0.00000872, -0.00000391, 0.00000116, 0.00001130, -0.00000060, 0.00000659, 0.00001424, 0.00000174, 0.00001127, 0.00001741, 0.00001941, 0.00005586, 0.00001656, -0.00004946, -0.00002848, -0.00001604, -0.00001250, -0.00000819, -0.00000512, -0.00000185, -0.00000508, -0.00000488, -0.00000270, -0.00000921, -0.00000341, -0.00000109, 0.00000062, -0.00000385, -0.00000342, -0.00000048, -0.00000172, -0.00000279, -0.00000252, -0.00000231, -0.00000318, -0.00000102, -0.00000405, -0.00000180, 0.00000540, -0.00001143, -0.00000513, 0.00000108, 0.00000033, 0.00000071, -0.00000343, 0.00000436, -0.00000074, -0.00000177, 0.00000137, 0.00000066, -0.00000155, 0.00000401, -0.00000045, -0.00000072, -0.00000254, 0.00000689, 0.00000344, -0.00000113, -0.00000023, -0.00000237, -0.00000009, },
		},
		{ { -0.00000010, 0.00000007, -0.00000072, 0.00000024, -0.00000009, -0.00000009, 0.00000420, 0.00000159, -0.00000148, -0.00000331, 0.00001083, 0.00000010, -0.00000022, -0.00000026, -0.00000238, 0.00000052, -0.00000495, 0.00000152, -0.00000231, -0.00000508, -0.00000256, -0.00000251, 0.00000088, -0.00000118, 0.00000011, 0.00000055, -0.00001046, -0.00000097, 0.00000059, 0.00000017, -0.00000434, -0.00000460, -0.00000235, -0.00000168, -0.00000349, -0.00000330, -0.00000379, -0.00000278, -0.00000049, -0.00000454, -0.00000490, -0.00000590, -0.00001355, -0.00001097, -0.00001358, -0.00002082, -0.00005111, 0.00000046, 0.00005663, 0.00002818, 0.00001488, 0.00001212, 0.00001031, 0.00000915, 0.00001378, 0.00000112, 0.00000761, 0.00001437, -0.00000251, 0.00001077, 0.00001203, 0.00001963, 0.00005216, -0.00002177, -0.00009587, -0.00002177, 0.00005216, 0.00001963, 0.00001203, 0.00001077, -0.00000251, 0.00001437, 0.00000761, 0.00000112, 0.00001378, 0.00000915, 0.00001031, 0.00001212, 0.00001488, 0.00002818, 0.00005663, 0.00000046, -0.00005111, -0.00002082, -0.00001358, -0.00001097, -0.00001355, -0.00000590, -0.00000490, -0.00000454, -0.00000049, -0.00000278, -0.00000379, -0.00000330, -0.00000349, -0.00000168, -0.00000235, -0.00000460, -0.00000434, 0.00000017, 0.00000059, -0.00000097, -0.00001046, 0.00000055, 0.00000011, -0.00000118, 0.00000088, -0.00000251, -0.00000256, -0.00000508, -0.00000231, 0.00000152, -0.00000495, 0.00000052, -0.00000238, -0.00000026, -0.00000022, 0.00000010, 0.00001083, -0.00000331, -0.00000148, 0.00000159, 0.00000420, -0.00000009, -0.00000009, 0.00000024, -0.00000072, 0.00000007, },
		{ -0.00000006, -0.00000010, -0.00000060, -0.00000012, 0.00000048, -0.00000100, 0.00000370, 0.00000220, 0.00000033, -0.00000514, 0.00000812, 0.00000486, -0.00000268, 0.00000228, -0.00000448, 0.00000176, -0.00000479, -0.00000008, -0.00000084, -0.00000475, -0.00000334, -0.00000254, 0.00000003, 0.00000000, -0.00000133, 0.00000247, -0.00000983, -0.00000333, 0.00000076, 0.00000037, -0.00000276, -0.00000594, -0.00000182, -0.00000224, -0.00000337, -0.00000283, -0.00000387, -0.00000309, -0.00000079, -0.00000399, -0.00000493, -0.00000495, -0.00001365, -0.00001045, -0.00001394, -0.00001560, -0.00004729, -0.00002563, 0.00005712, 0.00003695, 0.00001551, 0.00001483, 0.00001073, 0.00000989, 0.00001505, 0.00000200, 0.00000724, 0.00001474, -0.00000079, 0.00001018, 0.00001440, 0.00001813, 0.00005542, -0.00001847, -0.00009621, -0.00002403, 0.00004754, 0.00002195, 0.00000970, 0.00001110, -0.00000393, 0.00001376, 0.00000797, 0.00000066, 0.00001237, 0.00000846, 0.00000983, 0.00001021, 0.00001330, 0.00002336, 0.00004924, 0.00002330, -0.00004561, -0.00002955, -0.00001250, -0.00001224, -0.00001281, -0.00000719, -0.00000459, -0.00000539, -0.00000018, -0.00000253, -0.00000354, -0.00000366, -0.00000389, -0.00000100, -0.00000281, -0.00000367, -0.00000491, -0.00000093, 0.00000078, 0.00000037, -0.00000922, -0.00000250, 0.00000176, -0.00000229, 0.00000154, -0.00000250, -0.00000183, -0.00000500, -0.00000369, 0.00000201, -0.00000356, -0.00000188, 0.00000009, -0.00000243, 0.00000159, -0.00000293, 0.00001049, 0.00000061, -0.00000355, 0.00000105, 0.00000393, 0.00000146, -0.00000082, 0.00000061, -0.00000083, 0.00000022, },
		{ 0.00000004, -0.00000023, -0.00000061, -0.00000028, 0.00000066, -0.00000115, 0.00000255, 0.00000290, 0.00000146, -0.00000477, 0.00000340, 0.00000958, -0.00000407, 0.00000356, -0.00000471, 0.00000083, -0.00000280, -0.00000258, 0.00000051, -0.00000421, -0.00000374, -0.00000293, -0.00000059, 0.00000063, -0.00000181, 0.00000271, -0.00000740, -0.00000630, 0.00000104, -0.00000014, -0.00000068, -0.00000711, -0.00000161, -0.00000249, -0.00000350, -0.00000247, -0.00000359, -0.00000362, -0.00000081, -0.00000402, -0.00000445, -0.00000468, -0.00001273, -0.00001111, -0.00001281, -0.00001443, -0.00003699, -0.00004866, 0.00004697, 0.00004941, 0.00001591, 0.00001800, 0.00001125, 0.00001077, 0.00001603, 0.00000330, 0.00000692, 0.00001493, 0.00000094, 0.00000981, 0.00001616, 0.00001795, 0.00005730, -0.00001442, -0.00009697, -0.00002549, 0.00004217, 0.00002416, 0.00000817, 0.00001072, -0.00000485, 0.00001284, 0.00000832, 0.00000041, 0.00001113, 0.00000764, 0.00000929, 0.00000905, 0.00001089, 0.00002131, 0.00003932, 0.00003851, -0.00003122, -0.00003896, -0.00001257, -0.00001318, -0.00001231, -0.00000811, -0.00000456, -0.00000602, -0.00000031, -0.00000202, -0.00000341, -0.00000370, -0.00000457, -0.00000037, -0.00000293, -0.00000338, -0.00000450, -0.00000262, 0.00000107, 0.00000086, -0.00000684, -0.00000551, 0.00000257, -0.00000259, 0.00000148, -0.00000205, -0.00000162, -0.00000426, -0.00000499, 0.00000166, -0.00000153, -0.00000404, 0.00000142, -0.00000301, 0.00000177, -0.00000356, 0.00000737, 0.00000552, -0.00000497, 0.00000028, 0.00000313, 0.00000318, -0.00000123, 0.00000069, -0.00000076, 0.00000022, },
		{ 0.00000022, -0.00000030, -0.00000074, -0.00000027, 0.00000054, -0.00000082, 0.00000120, 0.00000340, 0.00000204, -0.00000298, -0.00000154, 0.00001227, -0.00000296, 0.00000256, -0.00000253, -0.00000210, 0.00000028, -0.00000507, 0.00000121, -0.00000342, -0.00000373, -0.00000364, -0.00000095, 0.00000057, -0.00000127, 0.00000156, -0.00000405, -0.00000883, 0.00000077, -0.00000074, 0.00000113, -0.00000746, -0.00000209, -0.00000228, -0.00000384, -0.00000225, -0.00000303, -0.00000424, -0.00000067, -0.00000443, -0.00000370, -0.00000487, -0.00001111, -0.00001247, -0.00001081, -0.00001567, -0.00002506, -0.00006271, 0.00002516, 0.00006282, 0.00001795, 0.00002073, 0.00001241, 0.00001158, 0.00001692, 0.00000471, 0.00000692, 0.00001496, 0.00000251, 0.00000993, 0.00001707, 0.00001895, 0.00005845, -0.00001042, -0.00009756, -0.00002671, 0.00003687, 0.00002539, 0.00000795, 0.00000955, -0.00000534, 0.00001170, 0.00000867, 0.00000019, 0.00001024, 0.00000666, 0.00000863, 0.00000850, 0.00000811, 0.00002055, 0.00003019, 0.00004482, -0.00001144, -0.00004504, -0.00001560, -0.00001295, -0.00001271, -0.00000817, -0.00000519, -0.00000607, -0.00000113, -0.00000115, -0.00000351, -0.00000339, -0.00000540, 0.00000002, -0.00000264, -0.00000362, -0.00000354, -0.00000424, 0.00000095, 0.00000105, -0.00000433, -0.00000729, 0.00000178, -0.00000171, 0.00000053, -0.00000099, -0.00000209, -0.00000295, -0.00000609, 0.00000080, 0.00000027, -0.00000494, 0.00000097, -0.00000188, 0.00000051, -0.00000237, 0.00000285, 0.00000968, -0.00000470, -0.00000108, 0.00000222, 0.00000441, -0.00000082, 0.00000022, -0.00000036, 0.00000005, },
		{ 0.00000040, -0.00000031, -0.00000089, -0.00000025, 0.00000035, -0.00000041, 0.00000017, 0.00000338, 0.00000245, -0.00000087, -0.00000499, 0.00001173, 0.00000094, -0.00000039, 0.00000121, -0.00000560, 0.00000284, -0.00000626, 0.00000061, -0.00000217, -0.00000357, -0.00000423, -0.00000137, 0.00000011, -0.00000022, -0.00000003, -0.00000111, -0.00000983, -0.00000067, -0.00000085, 0.00000197, -0.00000656, -0.00000336, -0.00000164, -0.00000437, -0.00000212, -0.00000245, -0.00000460, -0.00000071, -0.00000476, -0.00000323, -0.00000493, -0.00000958, -0.00001346, -0.00000929, -0.00001686, -0.00001604, -0.00006481, -0.00000521, 0.00007206, 0.00002410, 0.00002199, 0.00001482, 0.00001193, 0.00001816, 0.00000575, 0.00000756, 0.00001477, 0.00000391, 0.00001048, 0.00001743, 0.00002045, 0.00005990, -0.00000734, -0.00009749, -0.00002818, 0.00003227, 0.00002522, 0.00000893, 0.00000799, -0.00000577, 0.00001060, 0.00000888, -0.00000001, 0.00000963, 0.00000578, 0.00000765, 0.00000846, 0.00000532, 0.00002001, 0.00002345, 0.00004374, 0.00000880, -0.00004452, -0.00002202, -0.00001170, -0.00001392, -0.00000748, -0.00000633, -0.00000565, -0.00000243, -0.00000020, -0.00000360, -0.00000297, -0.00000609, -0.00000000, -0.00000198, -0.00000412, -0.00000255, -0.00000521, 0.00000010, 0.00000131, -0.00000243, -0.00000727, -0.00000058, -0.00000002, -0.00000091, 0.00000034, -0.00000284, -0.00000160, -0.00000662, -0.00000040, 0.00000140, -0.00000445, -0.00000082, 0.00000002, -0.00000107, -0.00000073, -0.00000124, 0.00001154, -0.00000221, -0.00000304, 0.00000149, 0.00000473, 0.00000057, -0.00000071, 0.00000023, -0.00000020, },
		{ 0.00000046, -0.00000026, -0.00000101, -0.00000035, 0.00000029, -0.00000024, -0.00000029, 0.00000275, 0.00000291, 0.00000078, -0.00000613, 0.00000812, 0.00000640, -0.00000366, 0.00000456, -0.00000762, 0.00000322, -0.00000521, -0.00000157, -0.00000042, -0.00000352, -0.00000431, -0.00000216, -0.00000035, 0.00000066, -0.00000107, 0.00000036, -0.00000872, -0.00000334, -0.00000038, 0.00000173, -0.00000460, -0.00000505, -0.00000089, -0.00000489, -0.00000204, -0.00000211, -0.00000441, -0.00000120, -0.00000466, -0.00000342, -0.00000443, -0.00000879, -0.00001325, -0.00000929, -0.00001628, -0.00001215, -0.00005625, -0.00003732, 0.00007110, 0.00003608, 0.00002144, 0.00001875, 0.00001163, 0.00002008, 0.00000611, 0.00000892, 0.00001444, 0.00000516, 0.00001121, 0.00001774, 0.00002161, 0.00006250, -0.00000561, -0.00009671, -0.00002992, 0.00002845, 0.00002388, 0.00001046, 0.00000674, -0.00000664, 0.00000984, 0.00000880, -0.00000008, 0.00000903, 0.00000533, 0.00000617, 0.00000883, 0.00000277, 0.00001918, 0.00001924, 0.00003817, 0.00002514, -0.00003639, -0.00003028, -0.00001078, -0.00001514, -0.00000672, -0.00000739, -0.00000530, -0.00000362, 0.00000028, -0.00000331, -0.00000278, -0.00000635, -0.00000053, -0.00000108, -0.00000459, -0.00000195, -0.00000527, -0.00000143, 0.00000165, -0.00000135, -0.00000566, -0.00000369, 0.00000151, -0.00000199, 0.00000122, -0.00000318, -0.00000092, -0.00000618, -0.00000195, 0.00000190, -0.00000311, -0.00000283, 0.00000137, -0.00000177, 0.00000014, -0.00000359, 0.00001046, 0.00000210, -0.00000505, 0.00000085, 0.00000414, 0.00000259, -0.00000165, 0.00000066, -0.00000030, },
		{ 0.00000029, -0.00000012, -0.00000106, -0.00000058, 0.00000037, -0.00000030, -0.00000025, 0.00000175, 0.00000337, 0.00000170, -0.00000518, 0.00000289, 0.00001123, -0.00000510, 0.00000560, -0.00000668, 0.00000072, -0.00000203, -0.00000473, 0.00000134, -0.00000351, -0.00000386, -0.00000323, -0.00000073, 0.00000099, -0.00000109, 0.00000013, -0.00000590, -0.00000649, 0.00000014, 0.00000091, -0.00000234, -0.00000641, -0.00000056, -0.00000512, -0.00000208, -0.00000202, -0.00000373, -0.00000202, -0.00000422, -0.00000415, -0.00000348, -0.00000874, -0.00001184, -0.00001071, -0.00001393, -0.00001260, -0.00004203, -0.00006282, 0.00005565, 0.00005306, 0.00002034, 0.00002348, 0.00001115, 0.00002248, 0.00000597, 0.00001068, 0.00001429, 0.00000625, 0.00001190, 0.00001846, 0.00002190, 0.00006656, -0.00000498, -0.00009580, -0.00003148, 0.00002501, 0.00002206, 0.00001162, 0.00000650, -0.00000826, 0.00000955, 0.00000836, 0.00000006, 0.00000820, 0.00000552, 0.00000425, 0.00000932, 0.00000071, 0.00001784, 0.00001700, 0.00003094, 0.00003525, -0.00002231, -0.00003730, -0.00001196, -0.00001539, -0.00000665, -0.00000772, -0.00000557, -0.00000414, -0.00000014, -0.00000239, -0.00000301, -0.00000607, -0.00000147, -0.00000017, -0.00000478, -0.00000185, -0.00000460, -0.00000317, 0.00000172, -0.00000081, -0.00000328, -0.00000632, 0.00000182, -0.00000196, 0.00000106, -0.00000257, -0.00000134, -0.00000470, -0.00000367, 0.00000191, -0.00000165, -0.00000404, 0.00000130, -0.00000111, -0.00000014, -0.00000398, 0.00000708, 0.00000688, -0.00000609, -0.00000008, 0.00000307, 0.00000454, -0.00000194, 0.00000052, -0.00000001, },
		{ -0.00000009, 0.00000009, -0.00000107, -0.00000084, 0.00000044, -0.00000043, -0.00000003, 0.00000080, 0.00000357, 0.00000215, -0.00000312, -0.00000196, 0.00001337, -0.00000332, 0.00000359, -0.00000282, -0.00000374, 0.00000183, -0.00000747, 0.00000216, -0.00000305, -0.00000330, -0.00000410, -0.00000128, 0.00000088, -0.00000030, -0.00000118, -0.00000255, -0.00000884, -0.00000023, 0.00000037, -0.00000076, -0.00000663, -0.00000112, -0.00000477, -0.00000243, -0.00000197, -0.00000297, -0.00000264, -0.00000393, -0.00000485, -0.00000268, -0.00000883, -0.00001005, -0.00001241, -0.00001130, -0.00001452, -0.00002849, -0.00007519, 0.00002579, 0.00007065, 0.00002167, 0.00002742, 0.00001157, 0.00002463, 0.00000607, 0.00001208, 0.00001482, 0.00000700, 0.00001249, 0.00001968, 0.00002135, 0.00007162, -0.00000451, -0.00009575, -0.00003219, 0.00002138, 0.00002052, 0.00001169, 0.00000756, -0.00001056, 0.00000956, 0.00000771, 0.00000034, 0.00000715, 0.00000619, 0.00000231, 0.00000946, -0.00000055, 0.00001592, 0.00001610, 0.00002396, 0.00003912, -0.00000580, -0.00003989, -0.00001626, -0.00001443, -0.00000752, -0.00000714, -0.00000659, -0.00000390, -0.00000141, -0.00000106, -0.00000345, -0.00000550, -0.00000246, 0.00000042, -0.00000455, -0.00000220, -0.00000359, -0.00000454, 0.00000118, -0.00000043, -0.00000098, -0.00000748, 0.00000039, -0.00000069, -0.00000012, -0.00000108, -0.00000263, -0.00000270, -0.00000507, 0.00000143, -0.00000048, -0.00000407, -0.00000010, 0.00000041, -0.00000099, -0.00000321, 0.00000289, 0.00001044, -0.00000523, -0.00000164, 0.00000204, 0.00000567, -0.00000109, -0.00000036, 0.00000069, },
		{ -0.00000045, 0.00000023, -0.00000103, -0.00000104, 0.00000037, -0.00000043, 0.00000006, 0.00000024, 0.00000329, 0.00000248, -0.00000107, -0.00000487, 0.00001189, 0.00000149, -0.00000048, 0.00000226, -0.00000791, 0.00000420, -0.00000814, 0.00000113, -0.00000165, -0.00000319, -0.00000420, -0.00000238, 0.00000069, 0.00000065, -0.00000250, -0.00000010, -0.00000923, -0.00000212, 0.00000072, -0.00000049, -0.00000536, -0.00000264, -0.00000385, -0.00000314, -0.00000173, -0.00000259, -0.00000254, -0.00000434, -0.00000491, -0.00000268, -0.00000834, -0.00000891, -0.00001302, -0.00001016, -0.00001493, -0.00002036, -0.00007255, -0.00001290, 0.00008150, 0.00002904, 0.00002887, 0.00001406, 0.00002567, 0.00000725, 0.00001241, 0.00001646, 0.00000734, 0.00001306, 0.00002120, 0.00002043, 0.00007680, -0.00000303, -0.00009747, -0.00003162, 0.00001724, 0.00001968, 0.00001048, 0.00000964, -0.00001303, 0.00000946, 0.00000715, 0.00000054, 0.00000608, 0.00000691, 0.00000093, 0.00000877, -0.00000086, 0.00001346, 0.00001595, 0.00001816, 0.00003822, 0.00000941, -0.00003631, -0.00002303, -0.00001311, -0.00000871, -0.00000615, -0.00000789, -0.00000335, -0.00000292, 0.00000003, -0.00000358, -0.00000514, -0.00000303, 0.00000041, -0.00000385, -0.00000279, -0.00000265, -0.00000519, -0.00000006, -0.00000008, 0.00000073, -0.00000689, -0.00000243, 0.00000117, -0.00000164, 0.00000057, -0.00000395, -0.00000113, -0.00000544, 0.00000025, 0.00000037, -0.00000325, -0.00000196, 0.00000167, -0.00000140, -0.00000243, -0.00000054, 0.00001150, -0.00000217, -0.00000373, 0.00000129, 0.00000562, 0.00000090, -0.00000171, 0.00000152, },
		{ -0.00000045, 0.00000014, -0.00000091, -0.00000114, 0.00000010, -0.00000026, -0.00000007, 0.00000013, 0.00000259, 0.00000285, 0.00000034, -0.00000531, 0.00000745, 0.00000752, -0.00000427, 0.00000594, -0.00000936, 0.00000335, -0.00000591, -0.00000189, 0.00000062, -0.00000369, -0.00000333, -0.00000406, 0.00000059, 0.00000122, -0.00000296, 0.00000059, -0.00000733, -0.00000537, 0.00000182, -0.00000141, -0.00000304, -0.00000457, -0.00000278, -0.00000398, -0.00000129, -0.00000277, -0.00000155, -0.00000552, -0.00000421, -0.00000360, -0.00000704, -0.00000888, -0.00001187, -0.00001127, -0.00001256, -0.00001881, -0.00005851, -0.00005072, 0.00007790, 0.00004447, 0.00002753, 0.00001886, 0.00002541, 0.00000982, 0.00001146, 0.00001912, 0.00000747, 0.00001361, 0.00002272, 0.00001974, 0.00008124, 0.00000026, -0.00010123, -0.00002998, 0.00001281, 0.00001940, 0.00000838, 0.00001197, -0.00001494, 0.00000878, 0.00000697, 0.00000041, 0.00000525, 0.00000724, 0.00000054, 0.00000711, -0.00000040, 0.00001081, 0.00001600, 0.00001385, 0.00003444, 0.00002082, -0.00002707, -0.00002998, -0.00001313, -0.00000921, -0.00000563, -0.00000868, -0.00000324, -0.00000387, 0.00000017, -0.00000288, -0.00000539, -0.00000286, -0.00000031, -0.00000281, -0.00000339, -0.00000204, -0.00000507, -0.00000167, 0.00000002, 0.00000186, -0.00000501, -0.00000559, 0.00000246, -0.00000247, 0.00000145, -0.00000434, -0.00000090, -0.00000431, -0.00000165, 0.00000105, -0.00000225, -0.00000322, 0.00000168, -0.00000069, -0.00000234, -0.00000233, 0.00000983, 0.00000237, -0.00000561, 0.00000063, 0.00000460, 0.00000342, -0.00000283, 0.00000195, },
		{ 0.00000012, -0.00000027, -0.00000067, -0.00000118, -0.00000026, -0.00000005, -0.00000026, 0.00000024, 0.00000177, 0.00000314, 0.00000103, -0.00000387, 0.00000191, 0.00001220, -0.00000536, 0.00000622, -0.00000684, -0.00000088, -0.00000148, -0.00000583, 0.00000289, -0.00000429, -0.00000194, -0.00000578, 0.00000033, 0.00000134, -0.00000242, -0.00000036, -0.00000389, -0.00000888, 0.00000277, -0.00000265, -0.00000072, -0.00000590, -0.00000229, -0.00000447, -0.00000096, -0.00000325, -0.00000008, -0.00000696, -0.00000327, -0.00000489, -0.00000547, -0.00000954, -0.00000951, -0.00001384, -0.00000843, -0.00002141, -0.00004048, -0.00007734, 0.00005545, 0.00006610, 0.00002554, 0.00002466, 0.00002479, 0.00001312, 0.00000992, 0.00002200, 0.00000800, 0.00001394, 0.00002414, 0.00001961, 0.00008457, 0.00000540, -0.00010647, -0.00002816, 0.00000879, 0.00001911, 0.00000617, 0.00001368, -0.00001569, 0.00000729, 0.00000726, -0.00000010, 0.00000479, 0.00000699, 0.00000114, 0.00000487, 0.00000019, 0.00000858, 0.00001563, 0.00001116, 0.00002925, 0.00002774, -0.00001464, -0.00003409, -0.00001596, -0.00000829, -0.00000612, -0.00000845, -0.00000406, -0.00000378, -0.00000096, -0.00000129, -0.00000626, -0.00000205, -0.00000145, -0.00000176, -0.00000374, -0.00000183, -0.00000446, -0.00000314, -0.00000045, 0.00000268, -0.00000267, -0.00000787, 0.00000219, -0.00000193, 0.00000100, -0.00000332, -0.00000223, -0.00000189, -0.00000384, 0.00000146, -0.00000147, -0.00000335, 0.00000027, 0.00000094, -0.00000280, -0.00000266, 0.00000636, 0.00000688, -0.00000624, -0.00000039, 0.00000321, 0.00000555, -0.00000292, 0.00000152, },
		{ 0.00000116, -0.00000086, -0.00000041, -0.00000119, -0.00000056, -0.00000000, -0.00000030, 0.00000025, 0.00000119, 0.00000313, 0.00000134, -0.00000171, -0.00000263, 0.00001349, -0.00000255, 0.00000281, -0.00000116, -0.00000672, 0.00000299, -0.00000871, 0.00000376, -0.00000408, -0.00000093, -0.00000668, -0.00000061, 0.00000133, -0.00000135, -0.00000200, -0.00000042, -0.00001111, 0.00000242, -0.00000313, 0.00000044, -0.00000573, -0.00000295, -0.00000420, -0.00000111, -0.00000355, 0.00000115, -0.00000779, -0.00000297, -0.00000559, -0.00000458, -0.00000989, -0.00000730, -0.00001601, -0.00000496, -0.00002401, -0.00002612, -0.00008612, 0.00001607, 0.00008722, 0.00002737, 0.00002892, 0.00002566, 0.00001577, 0.00000912, 0.00002392, 0.00000972, 0.00001373, 0.00002558, 0.00001990, 0.00008712, 0.00001163, -0.00011196, -0.00002736, 0.00000597, 0.00001823, 0.00000453, 0.00001416, -0.00001514, 0.00000520, 0.00000779, -0.00000082, 0.00000456, 0.00000632, 0.00000223, 0.00000289, 0.00000006, 0.00000743, 0.00001433, 0.00001007, 0.00002354, 0.00003082, -0.00000212, -0.00003318, -0.00002175, -0.00000635, -0.00000738, -0.00000742, -0.00000561, -0.00000285, -0.00000293, 0.00000060, -0.00000719, -0.00000115, -0.00000243, -0.00000116, -0.00000362, -0.00000199, -0.00000370, -0.00000402, -0.00000167, 0.00000339, -0.00000055, -0.00000851, 0.00000011, -0.00000016, -0.00000053, -0.00000128, -0.00000447, 0.00000081, -0.00000543, 0.00000122, -0.00000089, -0.00000258, -0.00000183, 0.00000248, -0.00000300, -0.00000241, 0.00000260, 0.00000975, -0.00000479, -0.00000208, 0.00000202, 0.00000650, -0.00000155, 0.00000019, },
		{ 0.00000217, -0.00000129, -0.00000030, -0.00000115, -0.00000072, -0.00000025, -0.00000005, 0.00000003, 0.00000101, 0.00000272, 0.00000160, 0.00000014, -0.00000479, 0.00001088, 0.00000338, -0.00000249, 0.00000505, -0.00001115, 0.00000484, -0.00000869, 0.00000216, -0.00000239, -0.00000101, -0.00000612, -0.00000255, 0.00000152, -0.00000047, -0.00000317, 0.00000163, -0.00001094, 0.00000016, -0.00000232, -0.00000008, -0.00000385, -0.00000473, -0.00000319, -0.00000186, -0.00000331, 0.00000154, -0.00000736, -0.00000393, -0.00000507, -0.00000503, -0.00000912, -0.00000641, -0.00001617, -0.00000419, -0.00002339, -0.00001993, -0.00007703, -0.00003147, 0.00009787, 0.00003783, 0.00002936, 0.00002963, 0.00001652, 0.00001023, 0.00002397, 0.00001309, 0.00001284, 0.00002721, 0.00002025, 0.00008957, 0.00001789, -0.00011644, -0.00002847, 0.00000466, 0.00001661, 0.00000369, 0.00001341, -0.00001372, 0.00000307, 0.00000812, -0.00000142, 0.00000424, 0.00000567, 0.00000309, 0.00000198, -0.00000140, 0.00000763, 0.00001205, 0.00001024, 0.00001796, 0.00003118, 0.00000812, -0.00002699, -0.00002891, -0.00000486, -0.00000836, -0.00000642, -0.00000711, -0.00000191, -0.00000477, 0.00000181, -0.00000738, -0.00000085, -0.00000267, -0.00000140, -0.00000294, -0.00000239, -0.00000305, -0.00000417, -0.00000344, 0.00000386, 0.00000109, -0.00000754, -0.00000310, 0.00000190, -0.00000215, 0.00000072, -0.00000634, 0.00000246, -0.00000544, -0.00000010, -0.00000019, -0.00000161, -0.00000347, 0.00000282, -0.00000217, -0.00000242, -0.00000017, 0.00001000, -0.00000125, -0.00000417, 0.00000123, 0.00000605, 0.00000103, -0.00000151, },
		{ 0.00000249, -0.00000114, -0.00000054, -0.00000099, -0.00000076, -0.00000070, 0.00000035, -0.00000034, 0.00000109, 0.00000209, 0.00000188, 0.00000120, -0.00000445, 0.00000558, 0.00000997, -0.00000668, 0.00000866, -0.00001151, 0.00000246, -0.00000519, -0.00000176, 0.00000048, -0.00000210, -0.00000426, -0.00000515, 0.00000181, -0.00000010, -0.00000325, 0.00000167, -0.00000833, -0.00000354, -0.00000066, -0.00000180, -0.00000099, -0.00000683, -0.00000205, -0.00000282, -0.00000267, 0.00000109, -0.00000574, -0.00000594, -0.00000353, -0.00000660, -0.00000736, -0.00000701, -0.00001406, -0.00000636, -0.00001908, -0.00002149, -0.00005660, -0.00007418, 0.00008874, 0.00005893, 0.00002597, 0.00003674, 0.00001538, 0.00001335, 0.00002222, 0.00001773, 0.00001169, 0.00002895, 0.00002040, 0.00009250, 0.00002340, -0.00011931, -0.00003153, 0.00000441, 0.00001470, 0.00000330, 0.00001200, -0.00001228, 0.00000165, 0.00000783, -0.00000161, 0.00000355, 0.00000540, 0.00000319, 0.00000253, -0.00000414, 0.00000875, 0.00000937, 0.00001085, 0.00001322, 0.00002959, 0.00001525, -0.00001728, -0.00003460, -0.00000569, -0.00000794, -0.00000639, -0.00000767, -0.00000182, -0.00000554, 0.00000156, -0.00000633, -0.00000153, -0.00000191, -0.00000246, -0.00000186, -0.00000285, -0.00000264, -0.00000374, -0.00000523, 0.00000374, 0.00000241, -0.00000559, -0.00000621, 0.00000295, -0.00000272, 0.00000156, -0.00000671, 0.00000213, -0.00000348, -0.00000247, 0.00000073, -0.00000102, -0.00000386, 0.00000147, -0.00000022, -0.00000289, -0.00000151, 0.00000782, 0.00000339, -0.00000581, 0.00000051, 0.00000463, 0.00000394, -0.00000265, },
		{ 0.00000168, -0.00000026, -0.00000115, -0.00000072, -0.00000076, -0.00000118, 0.00000064, -0.00000056, 0.00000111, 0.00000158, 0.00000202, 0.00000159, -0.00000254, -0.00000018, 0.00001438, -0.00000717, 0.00000774, -0.00000705, -0.00000357, 0.00000036, -0.00000634, 0.00000317, -0.00000320, -0.00000209, -0.00000740, 0.00000158, -0.00000002, -0.00000241, 0.00000024, -0.00000449, -0.00000716, 0.00000054, -0.00000348, 0.00000146, -0.00000800, -0.00000171, -0.00000332, -0.00000220, 0.00000039, -0.00000376, -0.00000802, -0.00000201, -0.00000818, -0.00000564, -0.00000811, -0.00001093, -0.00000974, -0.00001342, -0.00002641, -0.00003483, -0.00009995, 0.00005594, 0.00008703, 0.00002229, 0.00004471, 0.00001403, 0.00001717, 0.00002003, 0.00002231, 0.00001126, 0.00003035, 0.00002045, 0.00009596, 0.00002815, -0.00012091, -0.00003561, 0.00000402, 0.00001343, 0.00000267, 0.00001075, -0.00001165, 0.00000142, 0.00000679, -0.00000135, 0.00000242, 0.00000563, 0.00000243, 0.00000425, -0.00000735, 0.00000975, 0.00000731, 0.00001085, 0.00001011, 0.00002635, 0.00001978, -0.00000684, -0.00003621, -0.00000983, -0.00000584, -0.00000757, -0.00000704, -0.00000290, -0.00000496, -0.00000021, -0.00000425, -0.00000290, -0.00000052, -0.00000387, -0.00000085, -0.00000304, -0.00000255, -0.00000307, -0.00000654, 0.00000275, 0.00000363, -0.00000346, -0.00000808, 0.00000214, -0.00000172, 0.00000081, -0.00000528, -0.00000012, -0.00000015, -0.00000511, 0.00000147, -0.00000084, -0.00000303, -0.00000104, 0.00000199, -0.00000323, -0.00000191, 0.00000435, 0.00000749, -0.00000589, -0.00000068, 0.00000302, 0.00000605, -0.00000236, },
		{ -0.00000011, 0.00000106, -0.00000188, -0.00000047, -0.00000074, -0.00000151, 0.00000060, -0.00000041, 0.00000082, 0.00000143, 0.00000184, 0.00000173, -0.00000031, -0.00000428, 0.00001472, -0.00000319, 0.00000273, 0.00000048, -0.00001051, 0.00000503, -0.00000907, 0.00000386, -0.00000297, -0.00000093, -0.00000810, 0.00000012, 0.00000029, -0.00000142, -0.00000140, -0.00000121, -0.00000894, -0.00000005, -0.00000387, 0.00000224, -0.00000725, -0.00000284, -0.00000279, -0.00000250, 0.00000023, -0.00000247, -0.00000894, -0.00000175, -0.00000850, -0.00000524, -0.00000828, -0.00000869, -0.00001173, -0.00000973, -0.00002933, -0.00002045, -0.00010271, 0.00000412, 0.00011244, 0.00002463, 0.00004986, 0.00001522, 0.00001954, 0.00001935, 0.00002515, 0.00001266, 0.00003085, 0.00002078, 0.00009954, 0.00003284, -0.00012239, -0.00003920, 0.00000213, 0.00001364, 0.00000120, 0.00001027, -0.00001232, 0.00000239, 0.00000527, -0.00000090, 0.00000106, 0.00000615, 0.00000119, 0.00000630, -0.00000988, 0.00000951, 0.00000676, 0.00000948, 0.00000902, 0.00002172, 0.00002261, 0.00000193, -0.00003268, -0.00001652, -0.00000307, -0.00000919, -0.00000583, -0.00000458, -0.00000362, -0.00000265, -0.00000213, -0.00000404, 0.00000066, -0.00000484, -0.00000050, -0.00000268, -0.00000277, -0.00000238, -0.00000708, 0.00000097, 0.00000469, -0.00000162, -0.00000822, -0.00000044, 0.00000042, -0.00000094, -0.00000282, -0.00000312, 0.00000305, -0.00000672, 0.00000127, -0.00000058, -0.00000174, -0.00000346, 0.00000316, -0.00000256, -0.00000220, 0.00000106, 0.00000957, -0.00000376, -0.00000255, 0.00000181, 0.00000664, -0.00000036, },
		{ -0.00000206, 0.00000212, -0.00000224, -0.00000049, -0.00000065, -0.00000163, 0.00000017, 0.00000010, 0.00000023, 0.00000164, 0.00000136, 0.00000188, 0.00000130, -0.00000568, 0.00001105, 0.00000377, -0.00000370, 0.00000753, -0.00001468, 0.00000585, -0.00000800, 0.00000143, -0.00000069, -0.00000153, -0.00000670, -0.00000274, 0.00000106, -0.00000096, -0.00000210, 0.00000013, -0.00000787, -0.00000292, -0.00000260, 0.00000100, -0.00000457, -0.00000526, -0.00000135, -0.00000365, 0.00000098, -0.00000251, -0.00000805, -0.00000332, -0.00000699, -0.00000673, -0.00000678, -0.00000856, -0.00001069, -0.00001010, -0.00002716, -0.00001708, -0.00008514, -0.00005388, 0.00012221, 0.00003902, 0.00004940, 0.00002093, 0.00001893, 0.00002160, 0.00002524, 0.00001634, 0.00003032, 0.00002165, 0.00010268, 0.00003836, -0.00012496, -0.00004104, -0.00000206, 0.00001550, -0.00000106, 0.00001060, -0.00001410, 0.00000400, 0.00000390, -0.00000070, -0.00000019, 0.00000657, 0.00000007, 0.00000776, -0.00001076, 0.00000744, 0.00000792, 0.00000676, 0.00000957, 0.00001642, 0.00002417, 0.00000812, -0.00002514, -0.00002328, -0.00000167, -0.00000984, -0.00000518, -0.00000579, -0.00000266, -0.00000448, -0.00000120, -0.00000399, 0.00000081, -0.00000470, -0.00000119, -0.00000166, -0.00000325, -0.00000183, -0.00000695, -0.00000113, 0.00000525, -0.00000009, -0.00000696, -0.00000382, 0.00000244, -0.00000246, -0.00000073, -0.00000521, 0.00000450, -0.00000615, -0.00000043, 0.00000025, -0.00000091, -0.00000458, 0.00000235, -0.00000046, -0.00000289, -0.00000112, 0.00000902, 0.00000029, -0.00000470, 0.00000107, 0.00000568, 0.00000281, },
		{ -0.00000304, 0.00000216, -0.00000180, -0.00000098, -0.00000043, -0.00000163, -0.00000045, 0.00000071, -0.00000037, 0.00000187, 0.00000088, 0.00000201, 0.00000205, -0.00000468, 0.00000512, 0.00001064, -0.00000798, 0.00001061, -0.00001344, 0.00000157, -0.00000305, -0.00000351, 0.00000299, -0.00000342, -0.00000382, -0.00000621, 0.00000186, -0.00000108, -0.00000154, -0.00000068, -0.00000436, -0.00000716, -0.00000053, -0.00000134, -0.00000107, -0.00000783, 0.00000014, -0.00000507, 0.00000227, -0.00000358, -0.00000580, -0.00000611, -0.00000434, -0.00000939, -0.00000419, -0.00001020, -0.00000696, -0.00001391, -0.00002071, -0.00002212, -0.00005733, -0.00010150, 0.00010565, 0.00006690, 0.00004407, 0.00003064, 0.00001580, 0.00002644, 0.00002315, 0.00002146, 0.00002944, 0.00002285, 0.00010521, 0.00004522, -0.00012923, -0.00004080, -0.00000822, 0.00001819, -0.00000337, 0.00001116, -0.00001626, 0.00000535, 0.00000340, -0.00000113, -0.00000106, 0.00000654, -0.00000047, 0.00000807, -0.00000978, 0.00000389, 0.00001001, 0.00000363, 0.00001055, 0.00001172, 0.00002410, 0.00001237, -0.00001626, -0.00002709, -0.00000346, -0.00000848, -0.00000595, -0.00000568, -0.00000295, -0.00000476, -0.00000221, -0.00000234, -0.00000029, -0.00000339, -0.00000279, -0.00000024, -0.00000379, -0.00000144, -0.00000647, -0.00000290, 0.00000488, 0.00000138, -0.00000506, -0.00000663, 0.00000302, -0.00000255, -0.00000021, -0.00000522, 0.00000335, -0.00000324, -0.00000340, 0.00000160, -0.00000090, -0.00000399, -0.00000033, 0.00000250, -0.00000366, -0.00000216, 0.00000638, 0.00000499, -0.00000608, 0.00000032, 0.00000386, 0.00000597, },
		{ -0.00000218, 0.00000083, -0.00000053, -0.00000185, -0.00000011, -0.00000159, -0.00000101, 0.00000105, -0.00000061, 0.00000177, 0.00000073, 0.00000194, 0.00000225, -0.00000250, -0.00000053, 0.00001441, -0.00000756, 0.00000816, -0.00000676, -0.00000633, 0.00000354, -0.00000865, 0.00000611, -0.00000501, -0.00000106, -0.00000874, 0.00000166, -0.00000121, -0.00000029, -0.00000263, -0.00000007, -0.00001082, 0.00000063, -0.00000316, 0.00000150, -0.00000903, 0.00000049, -0.00000577, 0.00000317, -0.00000464, -0.00000351, -0.00000856, -0.00000216, -0.00001151, -0.00000216, -0.00001192, -0.00000271, -0.00001824, -0.00001383, -0.00002896, -0.00003198, -0.00012520, 0.00006019, 0.00010209, 0.00003936, 0.00004077, 0.00001290, 0.00003165, 0.00002103, 0.00002607, 0.00002968, 0.00002358, 0.00010752, 0.00005315, -0.00013480, -0.00003935, -0.00001492, 0.00002016, -0.00000456, 0.00001109, -0.00001792, 0.00000560, 0.00000420, -0.00000226, -0.00000150, 0.00000600, -0.00000034, 0.00000727, -0.00000755, 0.00000007, 0.00001164, 0.00000146, 0.00001055, 0.00000893, 0.00002182, 0.00001594, -0.00000879, -0.00002593, -0.00000878, -0.00000535, -0.00000795, -0.00000434, -0.00000451, -0.00000351, -0.00000482, 0.00000031, -0.00000198, -0.00000159, -0.00000455, 0.00000096, -0.00000400, -0.00000129, -0.00000594, -0.00000389, 0.00000344, 0.00000285, -0.00000320, -0.00000789, 0.00000150, -0.00000093, -0.00000145, -0.00000316, 0.00000014, 0.00000091, -0.00000643, 0.00000268, -0.00000128, -0.00000232, -0.00000369, 0.00000494, -0.00000357, -0.00000277, 0.00000299, 0.00000865, -0.00000562, -0.00000101, 0.00000208, 0.00000791, },
		{ 0.00000056, -0.00000145, 0.00000107, -0.00000269, 0.00000010, -0.00000154, -0.00000134, 0.00000090, -0.00000029, 0.00000118, 0.00000104, 0.00000153, 0.00000234, -0.00000044, -0.00000391, 0.00001356, -0.00000227, 0.00000152, 0.00000249, -0.00001410, 0.00000807, -0.00001098, 0.00000648, -0.00000459, -0.00000009, -0.00000892, -0.00000036, -0.00000069, 0.00000064, -0.00000422, 0.00000297, -0.00001196, -0.00000061, -0.00000309, 0.00000183, -0.00000784, -0.00000101, -0.00000511, 0.00000292, -0.00000466, -0.00000255, -0.00000914, -0.00000204, -0.00001152, -0.00000228, -0.00001189, -0.00000040, -0.00001981, -0.00001076, -0.00003122, -0.00001843, -0.00012040, -0.00000565, 0.00013107, 0.00004382, 0.00004630, 0.00001397, 0.00003427, 0.00002154, 0.00002797, 0.00003245, 0.00002303, 0.00011028, 0.00006138, -0.00014057, -0.00003823, -0.00002036, 0.00001986, -0.00000373, 0.00000984, -0.00001852, 0.00000438, 0.00000620, -0.00000378, -0.00000176, 0.00000514, 0.00000018, 0.00000594, -0.00000518, -0.00000261, 0.00001150, 0.00000126, 0.00000867, 0.00000862, 0.00001732, 0.00001955, -0.00000410, -0.00002001, -0.00001591, -0.00000219, -0.00000987, -0.00000281, -0.00000637, -0.00000176, -0.00000775, 0.00000256, -0.00000303, -0.00000041, -0.00000550, 0.00000124, -0.00000354, -0.00000146, -0.00000550, -0.00000405, 0.00000129, 0.00000405, -0.00000162, -0.00000743, -0.00000165, 0.00000160, -0.00000352, -0.00000024, -0.00000345, 0.00000433, -0.00000786, 0.00000242, -0.00000118, -0.00000079, -0.00000609, 0.00000539, -0.00000183, -0.00000367, 0.00000015, 0.00000997, -0.00000287, -0.00000304, 0.00000091, 0.00000799, },
		{ 0.00000429, -0.00000356, 0.00000210, -0.00000292, -0.00000010, -0.00000139, -0.00000143, 0.00000034, 0.00000041, 0.00000030, 0.00000158, 0.00000097, 0.00000248, 0.00000076, -0.00000438, 0.00000877, 0.00000556, -0.00000573, 0.00000995, -0.00001752, 0.00000746, -0.00000861, 0.00000310, -0.00000153, -0.00000154, -0.00000642, -0.00000412, 0.00000054, 0.00000068, -0.00000437, 0.00000353, -0.00000981, -0.00000445, -0.00000103, -0.00000012, -0.00000453, -0.00000395, -0.00000334, 0.00000150, -0.00000331, -0.00000346, -0.00000734, -0.00000434, -0.00000908, -0.00000488, -0.00000955, -0.00000110, -0.00001715, -0.00001324, -0.00002641, -0.00001866, -0.00009349, -0.00007411, 0.00013758, 0.00006450, 0.00004415, 0.00002119, 0.00003264, 0.00002613, 0.00002625, 0.00003808, 0.00002112, 0.00011379, 0.00006928, -0.00014552, -0.00003878, -0.00002329, 0.00001668, -0.00000093, 0.00000765, -0.00001822, 0.00000203, 0.00000874, -0.00000502, -0.00000227, 0.00000430, 0.00000065, 0.00000481, -0.00000374, -0.00000319, 0.00000920, 0.00000299, 0.00000523, 0.00001014, 0.00001168, 0.00002271, -0.00000165, -0.00001173, -0.00002167, -0.00000140, -0.00001005, -0.00000245, -0.00000725, -0.00000088, -0.00000948, 0.00000300, -0.00000237, -0.00000070, -0.00000502, 0.00000026, -0.00000231, -0.00000192, -0.00000513, -0.00000369, -0.00000087, 0.00000441, -0.00000012, -0.00000584, -0.00000515, 0.00000353, -0.00000491, 0.00000184, -0.00000549, 0.00000522, -0.00000652, 0.00000022, -0.00000004, -0.00000035, -0.00000644, 0.00000328, 0.00000150, -0.00000492, -0.00000154, 0.00000873, 0.00000156, -0.00000514, 0.00000024, 0.00000647, },
		{ 0.00000758, -0.00000424, 0.00000176, -0.00000213, -0.00000086, -0.00000107, -0.00000141, -0.00000038, 0.00000108, -0.00000040, 0.00000190, 0.00000063, 0.00000255, 0.00000113, -0.00000271, 0.00000238, 0.00001235, -0.00000960, 0.00001197, -0.00001427, 0.00000104, -0.00000213, -0.00000285, 0.00000299, -0.00000438, -0.00000246, -0.00000825, 0.00000160, 0.00000015, -0.00000317, 0.00000190, -0.00000532, -0.00000948, 0.00000164, -0.00000295, -0.00000068, -0.00000680, -0.00000168, -0.00000017, -0.00000135, -0.00000551, -0.00000412, -0.00000787, -0.00000550, -0.00000855, -0.00000615, -0.00000373, -0.00001157, -0.00001939, -0.00001714, -0.00002723, -0.00005875, -0.00012486, 0.00011001, 0.00010160, 0.00003640, 0.00003304, 0.00002805, 0.00003375, 0.00002217, 0.00004513, 0.00001903, 0.00011751, 0.00007695, -0.00014942, -0.00004127, -0.00002381, 0.00001138, 0.00000260, 0.00000560, -0.00001784, -0.00000055, 0.00001083, -0.00000530, -0.00000336, 0.00000377, 0.00000069, 0.00000437, -0.00000373, -0.00000170, 0.00000546, 0.00000546, 0.00000157, 0.00001184, 0.00000677, 0.00002404, 0.00000040, -0.00000432, -0.00002304, -0.00000461, -0.00000768, -0.00000391, -0.00000647, -0.00000164, -0.00000921, 0.00000112, 0.00000015, -0.00000240, -0.00000329, -0.00000160, -0.00000073, -0.00000244, -0.00000483, -0.00000325, -0.00000230, 0.00000350, 0.00000154, -0.00000396, -0.00000761, 0.00000357, -0.00000451, 0.00000196, -0.00000488, 0.00000301, -0.00000262, -0.00000332, 0.00000178, -0.00000101, -0.00000484, -0.00000061, 0.00000524, -0.00000574, -0.00000247, 0.00000583, 0.00000615, -0.00000616, -0.00000056, 0.00000429, },
		{ 0.00000920, -0.00000272, -0.00000012, -0.00000047, -0.00000197, -0.00000069, -0.00000138, -0.00000093, 0.00000133, -0.00000052, 0.00000162, 0.00000081, 0.00000232, 0.00000114, -0.00000035, -0.00000291, 0.00001507, -0.00000782, 0.00000767, -0.00000531, -0.00000859, 0.00000523, -0.00000828, 0.00000639, -0.00000643, 0.00000076, -0.00001071, 0.00000110, -0.00000002, -0.00000166, -0.00000043, -0.00000065, -0.00001328, 0.00000279, -0.00000463, 0.00000163, -0.00000773, -0.00000159, -0.00000080, -0.00000008, -0.00000717, -0.00000136, -0.00001047, -0.00000308, -0.00001094, -0.00000405, -0.00000576, -0.00000629, -0.00002493, -0.00000904, -0.00003513, -0.00003150, -0.00014356, 0.00004824, 0.00014531, 0.00003120, 0.00004417, 0.00002464, 0.00004099, 0.00001893, 0.00005080, 0.00001878, 0.00012015, 0.00008526, -0.00015304, -0.00004479, -0.00002332, 0.00000590, 0.00000493, 0.00000510, -0.00001839, -0.00000240, 0.00001165, -0.00000422, -0.00000510, 0.00000360, 0.00000019, 0.00000470, -0.00000492, 0.00000091, 0.00000186, 0.00000695, -0.00000064, 0.00001200, 0.00000429, 0.00002236, 0.00000364, -0.00000013, -0.00001890, -0.00001131, -0.00000367, -0.00000645, -0.00000454, -0.00000363, -0.00000739, -0.00000229, 0.00000348, -0.00000446, -0.00000134, -0.00000336, 0.00000044, -0.00000255, -0.00000469, -0.00000296, -0.00000270, 0.00000141, 0.00000325, -0.00000228, -0.00000829, 0.00000138, -0.00000236, 0.00000025, -0.00000208, -0.00000121, 0.00000219, -0.00000654, 0.00000309, -0.00000185, -0.00000254, -0.00000450, 0.00000756, -0.00000500, -0.00000346, 0.00000270, 0.00000918, -0.00000506, -0.00000207, 0.00000243, },
		{ 0.00000871, 0.00000073, -0.00000278, 0.00000130, -0.00000284, -0.00000056, -0.00000131, -0.00000118, 0.00000097, 0.00000003, 0.00000075, 0.00000148, 0.00000178, 0.00000117, 0.00000146, -0.00000538, 0.00001277, -0.00000104, -0.00000057, 0.00000533, -0.00001670, 0.00000917, -0.00000980, 0.00000617, -0.00000567, 0.00000129, -0.00000996, -0.00000184, 0.00000092, -0.00000098, -0.00000170, 0.00000199, -0.00001383, 0.00000089, -0.00000378, 0.00000109, -0.00000577, -0.00000378, 0.00000034, -0.00000044, -0.00000716, -0.00000070, -0.00001035, -0.00000361, -0.00001024, -0.00000507, -0.00000502, -0.00000426, -0.00002599, -0.00000701, -0.00003529, -0.00002100, -0.00012827, -0.00003402, 0.00017736, 0.00003964, 0.00004812, 0.00002723, 0.00004404, 0.00001987, 0.00005243, 0.00002217, 0.00012057, 0.00009527, -0.00015755, -0.00004783, -0.00002374, 0.00000231, 0.00000441, 0.00000704, -0.00002037, -0.00000305, 0.00001092, -0.00000196, -0.00000714, 0.00000356, -0.00000066, 0.00000547, -0.00000658, 0.00000330, -0.00000004, 0.00000618, -0.00000046, 0.00000983, 0.00000474, 0.00001773, 0.00000826, 0.00000074, -0.00001080, -0.00001881, -0.00000048, -0.00000823, -0.00000296, -0.00000550, -0.00000548, -0.00000553, 0.00000584, -0.00000534, -0.00000049, -0.00000392, 0.00000043, -0.00000186, -0.00000488, -0.00000282, -0.00000226, -0.00000122, 0.00000445, -0.00000078, -0.00000735, -0.00000217, 0.00000034, -0.00000198, 0.00000126, -0.00000526, 0.00000551, -0.00000750, 0.00000257, -0.00000171, -0.00000105, -0.00000653, 0.00000704, -0.00000209, -0.00000502, 0.00000045, 0.00000966, -0.00000168, -0.00000425, 0.00000135, },
		{ 0.00000669, 0.00000486, -0.00000486, 0.00000213, -0.00000284, -0.00000098, -0.00000109, -0.00000119, 0.00000020, 0.00000093, -0.00000033, 0.00000220, 0.00000126, 0.00000127, 0.00000227, -0.00000490, 0.00000685, 0.00000756, -0.00000827, 0.00001258, -0.00001879, 0.00000667, -0.00000579, 0.00000164, -0.00000174, -0.00000118, -0.00000606, -0.00000672, 0.00000274, -0.00000150, -0.00000112, 0.00000170, -0.00001074, -0.00000387, -0.00000070, -0.00000191, -0.00000159, -0.00000748, 0.00000271, -0.00000223, -0.00000539, -0.00000242, -0.00000730, -0.00000712, -0.00000652, -0.00000910, -0.00000136, -0.00000615, -0.00002163, -0.00001202, -0.00002655, -0.00002664, -0.00009026, -0.00011317, 0.00017799, 0.00006955, 0.00004186, 0.00003790, 0.00004136, 0.00002630, 0.00004935, 0.00002925, 0.00011886, 0.00010726, -0.00016360, -0.00004941, -0.00002629, 0.00000163, 0.00000072, 0.00001104, -0.00002324, -0.00000289, 0.00000918, 0.00000071, -0.00000879, 0.00000326, -0.00000153, 0.00000616, -0.00000788, 0.00000436, 0.00000059, 0.00000302, 0.00000169, 0.00000603, 0.00000703, 0.00001180, 0.00001271, 0.00000022, -0.00000219, -0.00002340, -0.00000075, -0.00000745, -0.00000318, -0.00000584, -0.00000497, -0.00000696, 0.00000580, -0.00000401, -0.00000151, -0.00000277, -0.00000098, -0.00000036, -0.00000539, -0.00000269, -0.00000153, -0.00000345, 0.00000443, 0.00000090, -0.00000562, -0.00000557, 0.00000191, -0.00000307, 0.00000326, -0.00000703, 0.00000554, -0.00000516, -0.00000024, -0.00000014, -0.00000121, -0.00000593, 0.00000362, 0.00000235, -0.00000673, -0.00000077, 0.00000776, 0.00000304, -0.00000624, 0.00000072, },
		{ 0.00000429, 0.00000805, -0.00000500, 0.00000127, -0.00000165, -0.00000200, -0.00000071, -0.00000111, -0.00000064, 0.00000165, -0.00000101, 0.00000242, 0.00000118, 0.00000120, 0.00000236, -0.00000268, 0.00000013, 0.00001399, -0.00001125, 0.00001298, -0.00001313, -0.00000203, 0.00000225, -0.00000526, 0.00000351, -0.00000494, -0.00000096, -0.00001153, 0.00000402, -0.00000245, 0.00000074, -0.00000063, -0.00000560, -0.00000944, 0.00000262, -0.00000537, 0.00000268, -0.00001068, 0.00000466, -0.00000408, -0.00000309, -0.00000521, -0.00000297, -0.00001164, -0.00000194, -0.00001386, 0.00000306, -0.00000981, -0.00001450, -0.00002043, -0.00001384, -0.00003946, -0.00004856, -0.00016508, 0.00013539, 0.00011888, 0.00002961, 0.00005357, 0.00003552, 0.00003604, 0.00004383, 0.00003777, 0.00011677, 0.00012022, -0.00017068, -0.00004979, -0.00003069, 0.00000317, -0.00000476, 0.00001536, -0.00002554, -0.00000304, 0.00000751, 0.00000270, -0.00000935, 0.00000235, -0.00000208, 0.00000636, -0.00000830, 0.00000370, 0.00000339, -0.00000137, 0.00000423, 0.00000238, 0.00000906, 0.00000702, 0.00001475, 0.00000099, 0.00000346, -0.00002239, -0.00000562, -0.00000377, -0.00000544, -0.00000429, -0.00000637, -0.00000606, 0.00000319, -0.00000072, -0.00000396, -0.00000044, -0.00000315, 0.00000138, -0.00000590, -0.00000254, -0.00000098, -0.00000457, 0.00000284, 0.00000292, -0.00000390, -0.00000749, 0.00000118, -0.00000207, 0.00000295, -0.00000572, 0.00000215, -0.00000023, -0.00000429, 0.00000214, -0.00000254, -0.00000341, -0.00000124, 0.00000654, -0.00000742, -0.00000166, 0.00000470, 0.00000735, -0.00000677, -0.00000022, },
		{ 0.00000252, 0.00000913, -0.00000268, -0.00000118, 0.00000035, -0.00000327, -0.00000036, -0.00000102, -0.00000122, 0.00000177, -0.00000091, 0.00000184, 0.00000178, 0.00000075, 0.00000229, -0.00000029, -0.00000467, 0.00001547, -0.00000782, 0.00000658, -0.00000194, -0.00001313, 0.00001011, -0.00001067, 0.00000683, -0.00000716, 0.00000252, -0.00001374, 0.00000303, -0.00000251, 0.00000236, -0.00000298, -0.00000110, -0.00001294, 0.00000364, -0.00000689, 0.00000462, -0.00001127, 0.00000444, -0.00000435, -0.00000203, -0.00000698, 0.00000011, -0.00001432, 0.00000049, -0.00001626, 0.00000510, -0.00001175, -0.00000906, -0.00002645, -0.00000475, -0.00004800, -0.00002089, -0.00017554, 0.00005331, 0.00017285, 0.00002295, 0.00006663, 0.00003243, 0.00004426, 0.00004028, 0.00004400, 0.00011707, 0.00013228, -0.00017730, -0.00005040, -0.00003532, 0.00000483, -0.00000952, 0.00001765, -0.00002557, -0.00000468, 0.00000698, 0.00000314, -0.00000840, 0.00000073, -0.00000219, 0.00000587, -0.00000776, 0.00000169, 0.00000716, -0.00000512, 0.00000523, 0.00000069, 0.00000896, 0.00000527, 0.00001294, 0.00000458, 0.00000463, -0.00001578, -0.00001361, 0.00000109, -0.00000838, -0.00000191, -0.00000877, -0.00000383, -0.00000065, 0.00000298, -0.00000635, 0.00000167, -0.00000483, 0.00000242, -0.00000586, -0.00000255, -0.00000073, -0.00000448, 0.00000006, 0.00000492, -0.00000247, -0.00000749, -0.00000175, 0.00000056, 0.00000089, -0.00000228, -0.00000299, 0.00000504, -0.00000746, 0.00000348, -0.00000366, -0.00000074, -0.00000525, 0.00000837, -0.00000585, -0.00000307, 0.00000194, 0.00000964, -0.00000496, -0.00000200, },
		{ 0.00000163, 0.00000798, 0.00000146, -0.00000411, 0.00000214, -0.00000403, -0.00000041, -0.00000085, -0.00000146, 0.00000123, -0.00000008, 0.00000062, 0.00000280, 0.00000008, 0.00000235, 0.00000121, -0.00000625, 0.00001178, 0.00000039, -0.00000306, 0.00000959, -0.00002099, 0.00001297, -0.00001095, 0.00000556, -0.00000566, 0.00000234, -0.00001182, -0.00000105, -0.00000092, 0.00000252, -0.00000349, 0.00000051, -0.00001240, 0.00000090, -0.00000518, 0.00000306, -0.00000839, 0.00000143, -0.00000233, -0.00000322, -0.00000630, 0.00000015, -0.00001324, -0.00000113, -0.00001439, 0.00000281, -0.00000957, -0.00000858, -0.00002582, -0.00000451, -0.00004506, -0.00001543, -0.00014667, -0.00004779, 0.00020752, 0.00003586, 0.00006911, 0.00003797, 0.00004633, 0.00004267, 0.00004487, 0.00012181, 0.00014203, -0.00018205, -0.00005289, -0.00003824, 0.00000436, -0.00001129, 0.00001619, -0.00002248, -0.00000823, 0.00000800, 0.00000183, -0.00000617, -0.00000126, -0.00000202, 0.00000484, -0.00000662, -0.00000082, 0.00001039, -0.00000667, 0.00000356, 0.00000165, 0.00000625, 0.00000671, 0.00000776, 0.00001028, 0.00000258, -0.00000647, -0.00002107, 0.00000404, -0.00000965, -0.00000059, -0.00001042, -0.00000210, -0.00000362, 0.00000502, -0.00000693, 0.00000212, -0.00000485, 0.00000198, -0.00000485, -0.00000294, -0.00000060, -0.00000369, -0.00000294, 0.00000606, -0.00000100, -0.00000612, -0.00000563, 0.00000324, -0.00000122, 0.00000122, -0.00000721, 0.00000793, -0.00000764, 0.00000244, -0.00000325, 0.00000038, -0.00000655, 0.00000665, -0.00000179, -0.00000520, 0.00000027, 0.00000924, -0.00000094, -0.00000433, },
		{ 0.00000105, 0.00000549, 0.00000583, -0.00000593, 0.00000257, -0.00000364, -0.00000115, -0.00000048, -0.00000148, 0.00000032, 0.00000099, -0.00000064, 0.00000364, -0.00000036, 0.00000240, 0.00000171, -0.00000507, 0.00000506, 0.00000940, -0.00001071, 0.00001594, -0.00002112, 0.00000819, -0.00000499, -0.00000050, -0.00000054, -0.00000132, -0.00000644, -0.00000714, 0.00000163, 0.00000120, -0.00000175, -0.00000120, -0.00000797, -0.00000483, -0.00000111, -0.00000105, -0.00000326, -0.00000315, 0.00000105, -0.00000611, -0.00000346, -0.00000269, -0.00000872, -0.00000625, -0.00000898, -0.00000300, -0.00000382, -0.00001293, -0.00001863, -0.00001267, -0.00003158, -0.00002790, -0.00009589, -0.00013758, 0.00019955, 0.00007632, 0.00005869, 0.00005361, 0.00004130, 0.00005176, 0.00004028, 0.00013054, 0.00014994, -0.00018488, -0.00005779, -0.00003857, 0.00000076, -0.00000939, 0.00001111, -0.00001708, -0.00001283, 0.00000988, -0.00000049, -0.00000354, -0.00000294, -0.00000195, 0.00000363, -0.00000543, -0.00000293, 0.00001202, -0.00000545, -0.00000047, 0.00000439, 0.00000213, 0.00000968, 0.00000156, 0.00001549, 0.00000053, 0.00000126, -0.00002398, 0.00000238, -0.00000753, -0.00000171, -0.00000997, -0.00000238, -0.00000415, 0.00000408, -0.00000485, 0.00000036, -0.00000294, 0.00000004, -0.00000299, -0.00000370, -0.00000035, -0.00000290, -0.00000508, 0.00000556, 0.00000094, -0.00000435, -0.00000877, 0.00000418, -0.00000166, 0.00000280, -0.00000842, 0.00000683, -0.00000410, -0.00000111, -0.00000110, -0.00000054, -0.00000484, 0.00000194, 0.00000352, -0.00000716, -0.00000058, 0.00000677, 0.00000395, -0.00000615, },
		{ -0.00000005, 0.00000299, 0.00000870, -0.00000535, 0.00000104, -0.00000195, -0.00000250, 0.00000003, -0.00000145, -0.00000054, 0.00000170, -0.00000126, 0.00000369, -0.00000010, 0.00000213, 0.00000172, -0.00000269, -0.00000150, 0.00001500, -0.00001230, 0.00001417, -0.00001275, -0.00000285, 0.00000464, -0.00000848, 0.00000552, -0.00000592, -0.00000033, -0.00001255, 0.00000315, -0.00000026, 0.00000102, -0.00000461, -0.00000201, -0.00001079, 0.00000265, -0.00000506, 0.00000137, -0.00000671, 0.00000361, -0.00000877, -0.00000033, -0.00000633, -0.00000327, -0.00001193, -0.00000324, -0.00000897, 0.00000215, -0.00001858, -0.00000927, -0.00002339, -0.00001546, -0.00004520, -0.00004745, -0.00018836, 0.00013804, 0.00013869, 0.00004273, 0.00007399, 0.00003359, 0.00006379, 0.00003393, 0.00013985, 0.00015870, -0.00018754, -0.00006395, -0.00003724, -0.00000497, -0.00000529, 0.00000449, -0.00001168, -0.00001655, 0.00001119, -0.00000249, -0.00000176, -0.00000350, -0.00000240, 0.00000263, -0.00000473, -0.00000408, 0.00001179, -0.00000211, -0.00000532, 0.00000693, -0.00000126, 0.00001170, -0.00000275, 0.00001732, 0.00000161, 0.00000412, -0.00002032, -0.00000431, -0.00000232, -0.00000497, -0.00000752, -0.00000472, -0.00000218, 0.00000050, -0.00000084, -0.00000273, -0.00000003, -0.00000245, -0.00000105, -0.00000441, 0.00000002, -0.00000254, -0.00000574, 0.00000326, 0.00000336, -0.00000288, -0.00001002, 0.00000250, 0.00000020, 0.00000183, -0.00000624, 0.00000216, 0.00000185, -0.00000563, 0.00000157, -0.00000252, -0.00000148, -0.00000362, 0.00000777, -0.00000742, -0.00000157, 0.00000370, 0.00000785, -0.00000612, },
		{ -0.00000207, 0.00000138, 0.00000910, -0.00000214, -0.00000198, 0.00000037, -0.00000390, 0.00000039, -0.00000138, -0.00000108, 0.00000168, -0.00000093, 0.00000275, 0.00000094, 0.00000141, 0.00000180, -0.00000068, -0.00000537, 0.00001493, -0.00000698, 0.00000553, 0.00000051, -0.00001502, 0.00001270, -0.00001375, 0.00000863, -0.00000802, 0.00000308, -0.00001429, 0.00000159, -0.00000019, 0.00000287, -0.00000719, 0.00000231, -0.00001369, 0.00000319, -0.00000621, 0.00000280, -0.00000687, 0.00000330, -0.00000920, 0.00000085, -0.00000809, -0.00000008, -0.00001462, -0.00000090, -0.00001122, 0.00000435, -0.00002098, -0.00000353, -0.00002923, -0.00000621, -0.00005353, -0.00002089, -0.00018730, 0.00003246, 0.00020161, 0.00003719, 0.00008867, 0.00003128, 0.00007219, 0.00003167, 0.00014486, 0.00017198, -0.00019269, -0.00006916, -0.00003645, -0.00001039, -0.00000183, -0.00000055, -0.00000908, -0.00001743, 0.00001054, -0.00000295, -0.00000178, -0.00000249, -0.00000355, 0.00000211, -0.00000478, -0.00000419, 0.00001018, 0.00000191, -0.00000897, 0.00000733, -0.00000214, 0.00001097, -0.00000338, 0.00001445, 0.00000686, 0.00000171, -0.00001142, -0.00001340, 0.00000334, -0.00000824, -0.00000476, -0.00000765, 0.00000074, -0.00000381, 0.00000299, -0.00000520, 0.00000214, -0.00000401, -0.00000011, -0.00000443, 0.00000022, -0.00000259, -0.00000513, -0.00000013, 0.00000558, -0.00000167, -0.00000934, -0.00000124, 0.00000341, -0.00000063, -0.00000218, -0.00000373, 0.00000736, -0.00000853, 0.00000275, -0.00000375, 0.00000131, -0.00000738, 0.00000873, -0.00000482, -0.00000349, 0.00000140, 0.00000931, -0.00000358, },
		{ -0.00000446, 0.00000066, 0.00000728, 0.00000260, -0.00000506, 0.00000207, -0.00000449, 0.00000018, -0.00000116, -0.00000128, 0.00000099, 0.00000012, 0.00000122, 0.00000229, 0.00000060, 0.00000205, 0.00000022, -0.00000585, 0.00000983, 0.00000253, -0.00000520, 0.00001239, -0.00002201, 0.00001409, -0.00001263, 0.00000614, -0.00000546, 0.00000185, -0.00001104, -0.00000365, 0.00000203, 0.00000261, -0.00000713, 0.00000286, -0.00001176, -0.00000069, -0.00000350, 0.00000013, -0.00000304, -0.00000028, -0.00000679, -0.00000092, -0.00000655, -0.00000095, -0.00001243, -0.00000382, -0.00000794, 0.00000091, -0.00001775, -0.00000472, -0.00002597, -0.00000889, -0.00004639, -0.00002209, -0.00014238, -0.00008818, 0.00023442, 0.00005910, 0.00008830, 0.00004120, 0.00007165, 0.00003801, 0.00014219, 0.00019208, -0.00020197, -0.00007179, -0.00003813, -0.00001326, -0.00000148, -0.00000168, -0.00001089, -0.00001476, 0.00000764, -0.00000160, -0.00000369, -0.00000016, -0.00000510, 0.00000198, -0.00000551, -0.00000359, 0.00000810, 0.00000509, -0.00000998, 0.00000477, -0.00000022, 0.00000749, -0.00000071, 0.00000804, 0.00001446, -0.00000322, -0.00000145, -0.00002039, 0.00000579, -0.00000882, -0.00000386, -0.00000913, 0.00000244, -0.00000649, 0.00000442, -0.00000523, 0.00000211, -0.00000349, -0.00000090, -0.00000335, -0.00000003, -0.00000266, -0.00000403, -0.00000332, 0.00000653, -0.00000015, -0.00000761, -0.00000540, 0.00000606, -0.00000258, 0.00000130, -0.00000787, 0.00000956, -0.00000773, 0.00000105, -0.00000291, 0.00000188, -0.00000771, 0.00000563, 0.00000038, -0.00000612, 0.00000026, 0.00000809, 0.00000093, },
		{ -0.00000592, 0.00000005, 0.00000445, 0.00000700, -0.00000645, 0.00000199, -0.00000366, -0.00000086, -0.00000071, -0.00000131, 0.00000003, 0.00000127, -0.00000013, 0.00000316, 0.00000028, 0.00000217, 0.00000024, -0.00000395, 0.00000256, 0.00001154, -0.00001230, 0.00001722, -0.00001969, 0.00000689, -0.00000482, -0.00000140, 0.00000097, -0.00000310, -0.00000424, -0.00001078, 0.00000514, 0.00000074, -0.00000457, -0.00000011, -0.00000589, -0.00000748, 0.00000150, -0.00000499, 0.00000287, -0.00000528, -0.00000305, -0.00000454, -0.00000256, -0.00000505, -0.00000646, -0.00001053, -0.00000085, -0.00000631, -0.00001054, -0.00001143, -0.00001536, -0.00002083, -0.00002779, -0.00004184, -0.00007854, -0.00018592, 0.00021038, 0.00011575, 0.00007208, 0.00006358, 0.00006233, 0.00005269, 0.00013271, 0.00021778, -0.00021431, -0.00007233, -0.00004234, -0.00001318, -0.00000469, 0.00000115, -0.00001636, -0.00000991, 0.00000383, 0.00000052, -0.00000647, 0.00000239, -0.00000627, 0.00000183, -0.00000652, -0.00000293, 0.00000651, 0.00000644, -0.00000808, -0.00000001, 0.00000314, 0.00000297, 0.00000300, 0.00000116, 0.00002082, -0.00000643, 0.00000479, -0.00002128, 0.00000259, -0.00000529, -0.00000592, -0.00000803, 0.00000164, -0.00000616, 0.00000241, -0.00000229, -0.00000034, -0.00000097, -0.00000319, -0.00000145, -0.00000068, -0.00000246, -0.00000327, -0.00000514, 0.00000542, 0.00000211, -0.00000588, -0.00000823, 0.00000639, -0.00000237, 0.00000235, -0.00000832, 0.00000718, -0.00000296, -0.00000324, -0.00000019, 0.00000018, -0.00000489, -0.00000025, 0.00000632, -0.00000809, -0.00000038, 0.00000525, 0.00000577, },
		{ -0.00000516, -0.00000132, 0.00000199, 0.00000937, -0.00000503, -0.00000020, -0.00000153, -0.00000243, -0.00000016, -0.00000129, -0.00000079, 0.00000187, -0.00000061, 0.00000297, 0.00000088, 0.00000180, 0.00000006, -0.00000149, -0.00000354, 0.00001594, -0.00001214, 0.00001297, -0.00000857, -0.00000614, 0.00000588, -0.00001006, 0.00000764, -0.00000833, 0.00000245, -0.00001630, 0.00000650, -0.00000073, -0.00000148, -0.00000417, 0.00000070, -0.00001361, 0.00000534, -0.00000915, 0.00000743, -0.00000847, -0.00000079, -0.00000739, 0.00000119, -0.00000933, -0.00000031, -0.00001693, 0.00000557, -0.00001258, -0.00000419, -0.00001834, -0.00000397, -0.00003346, -0.00000917, -0.00006228, -0.00002560, -0.00022996, 0.00012099, 0.00019609, 0.00005185, 0.00008986, 0.00005120, 0.00006986, 0.00012191, 0.00024419, -0.00022580, -0.00007365, -0.00004691, -0.00001208, -0.00000944, 0.00000548, -0.00002249, -0.00000594, 0.00000158, 0.00000153, -0.00000844, 0.00000371, -0.00000611, 0.00000107, -0.00000721, -0.00000290, 0.00000604, 0.00000579, -0.00000420, -0.00000513, 0.00000570, -0.00000021, 0.00000507, -0.00000303, 0.00002275, -0.00000477, 0.00000460, -0.00001502, -0.00000569, 0.00000115, -0.00000988, -0.00000511, -0.00000115, -0.00000341, -0.00000205, 0.00000227, -0.00000376, 0.00000219, -0.00000566, 0.00000028, -0.00000117, -0.00000209, -0.00000318, -0.00000523, 0.00000239, 0.00000488, -0.00000463, -0.00000890, 0.00000394, 0.00000018, 0.00000077, -0.00000528, 0.00000138, 0.00000376, -0.00000798, 0.00000266, -0.00000221, -0.00000095, -0.00000616, 0.00001025, -0.00000764, -0.00000168, 0.00000241, 0.00000900, },
		{ -0.00000184, -0.00000361, 0.00000062, 0.00000904, -0.00000102, -0.00000364, 0.00000091, -0.00000377, 0.00000005, -0.00000119, -0.00000124, 0.00000166, -0.00000005, 0.00000171, 0.00000227, 0.00000102, 0.00000011, 0.00000016, -0.00000637, 0.00001424, -0.00000498, 0.00000233, 0.00000621, -0.00001858, 0.00001328, -0.00001454, 0.00001009, -0.00000983, 0.00000515, -0.00001691, 0.00000378, 0.00000016, -0.00000022, -0.00000632, 0.00000443, -0.00001548, 0.00000485, -0.00000935, 0.00000768, -0.00000724, -0.00000232, -0.00000710, 0.00000196, -0.00001053, 0.00000215, -0.00001874, 0.00000684, -0.00001325, -0.00000357, -0.00001981, 0.00000105, -0.00003774, -0.00000189, -0.00006763, -0.00000404, -0.00021058, -0.00001622, 0.00027024, 0.00004896, 0.00010629, 0.00004885, 0.00008099, 0.00011728, 0.00026522, -0.00023175, -0.00007941, -0.00004864, -0.00001310, -0.00001238, 0.00000763, -0.00002551, -0.00000608, 0.00000316, -0.00000018, -0.00000822, 0.00000268, -0.00000401, -0.00000065, -0.00000712, -0.00000395, 0.00000690, 0.00000364, 0.00000016, -0.00000862, 0.00000570, -0.00000055, 0.00000399, -0.00000312, 0.00001940, 0.00000188, -0.00000098, -0.00000443, -0.00001519, 0.00000692, -0.00001287, -0.00000267, -0.00000384, -0.00000037, -0.00000648, 0.00000587, -0.00000584, 0.00000394, -0.00000664, 0.00000061, -0.00000076, -0.00000195, -0.00000348, -0.00000413, -0.00000142, 0.00000710, -0.00000346, -0.00000779, -0.00000023, 0.00000362, -0.00000185, -0.00000085, -0.00000489, 0.00000908, -0.00001027, 0.00000337, -0.00000322, 0.00000161, -0.00000932, 0.00001004, -0.00000385, -0.00000422, 0.00000072, 0.00000949, },
		{ 0.00000309, -0.00000595, 0.00000002, 0.00000666, 0.00000406, -0.00000656, 0.00000225, -0.00000395, -0.00000052, -0.00000083, -0.00000142, 0.00000087, 0.00000111, 0.00000003, 0.00000371, 0.00000037, 0.00000025, 0.00000067, -0.00000590, 0.00000802, 0.00000532, -0.00000881, 0.00001750, -0.00002368, 0.00001229, -0.00001137, 0.00000593, -0.00000568, 0.00000221, -0.00001170, -0.00000317, 0.00000374, -0.00000178, -0.00000498, 0.00000348, -0.00001170, -0.00000075, -0.00000500, 0.00000316, -0.00000142, -0.00000769, -0.00000334, -0.00000106, -0.00000735, -0.00000069, -0.00001434, 0.00000142, -0.00000687, -0.00001015, -0.00001382, -0.00000321, -0.00003005, -0.00001003, -0.00005312, -0.00001592, -0.00014371, -0.00016163, 0.00029970, 0.00008355, 0.00010243, 0.00006276, 0.00008041, 0.00012350, 0.00027751, -0.00022987, -0.00009143, -0.00004568, -0.00001839, -0.00001107, 0.00000506, -0.00002317, -0.00001174, 0.00000903, -0.00000470, -0.00000567, -0.00000067, -0.00000030, -0.00000308, -0.00000623, -0.00000608, 0.00000873, 0.00000092, 0.00000353, -0.00000929, 0.00000278, 0.00000171, 0.00000038, -0.00000020, 0.00001282, 0.00001049, -0.00000775, 0.00000513, -0.00002073, 0.00000798, -0.00001198, -0.00000305, -0.00000425, 0.00000062, -0.00000839, 0.00000623, -0.00000478, 0.00000291, -0.00000514, -0.00000106, 0.00000091, -0.00000239, -0.00000362, -0.00000291, -0.00000447, 0.00000752, -0.00000161, -0.00000610, -0.00000414, 0.00000577, -0.00000330, 0.00000229, -0.00000848, 0.00001015, -0.00000818, 0.00000075, -0.00000168, 0.00000137, -0.00000847, 0.00000550, 0.00000244, -0.00000729, 0.00000015, 0.00000747, },
		{ 0.00000768, -0.00000689, -0.00000077, 0.00000371, 0.00000816, -0.00000720, 0.00000143, -0.00000250, -0.00000198, -0.00000020, -0.00000150, -0.00000006, 0.00000213, -0.00000118, 0.00000431, 0.00000055, 0.00000002, 0.00000060, -0.00000368, 0.00000073, 0.00001361, -0.00001454, 0.00001993, -0.00001810, 0.00000202, -0.00000131, -0.00000333, 0.00000239, -0.00000450, -0.00000307, -0.00001170, 0.00000801, -0.00000494, -0.00000106, -0.00000103, -0.00000412, -0.00000904, 0.00000141, -0.00000356, 0.00000622, -0.00001424, 0.00000166, -0.00000602, -0.00000150, -0.00000699, -0.00000594, -0.00000796, 0.00000345, -0.00002061, -0.00000363, -0.00001344, -0.00001451, -0.00002784, -0.00002699, -0.00004570, -0.00006326, -0.00026878, 0.00025480, 0.00016081, 0.00008032, 0.00009091, 0.00007024, 0.00013866, 0.00028348, -0.00022268, -0.00010771, -0.00003930, -0.00002728, -0.00000587, -0.00000188, -0.00001629, -0.00002131, 0.00001710, -0.00001012, -0.00000230, -0.00000494, 0.00000370, -0.00000529, -0.00000510, -0.00000875, 0.00001079, -0.00000136, 0.00000505, -0.00000722, -0.00000186, 0.00000474, -0.00000355, 0.00000294, 0.00000671, 0.00001666, -0.00001079, 0.00000876, -0.00001866, 0.00000243, -0.00000637, -0.00000682, -0.00000169, -0.00000131, -0.00000686, 0.00000282, -0.00000057, -0.00000061, -0.00000162, -0.00000414, 0.00000335, -0.00000326, -0.00000329, -0.00000243, -0.00000563, 0.00000548, 0.00000126, -0.00000484, -0.00000614, 0.00000507, -0.00000217, 0.00000253, -0.00000784, 0.00000627, -0.00000209, -0.00000434, 0.00000170, -0.00000114, -0.00000462, -0.00000135, 0.00000873, -0.00000909, -0.00000038, 0.00000439, },
		{ 0.00001010, -0.00000529, -0.00000252, 0.00000151, 0.00000975, -0.00000476, -0.00000152, 0.00000013, -0.00000384, 0.00000041, -0.00000154, -0.00000072, 0.00000244, -0.00000128, 0.00000358, 0.00000187, -0.00000085, 0.00000064, -0.00000155, -0.00000438, 0.00001620, -0.00001204, 0.00001273, -0.00000394, -0.00001315, 0.00001045, -0.00001260, 0.00000971, -0.00001051, 0.00000433, -0.00001749, 0.00000962, -0.00000699, 0.00000265, -0.00000577, 0.00000315, -0.00001556, 0.00000564, -0.00000833, 0.00001148, -0.00001802, 0.00000433, -0.00000946, 0.00000334, -0.00001258, 0.00000155, -0.00001580, 0.00001172, -0.00002859, 0.00000412, -0.00002216, -0.00000036, -0.00004348, -0.00000476, -0.00007027, -0.00000475, -0.00030461, 0.00013152, 0.00026186, 0.00005802, 0.00012041, 0.00006092, 0.00015405, 0.00029111, -0.00021702, -0.00012283, -0.00003374, -0.00003626, 0.00000001, -0.00000985, -0.00000858, -0.00003072, 0.00002343, -0.00001330, -0.00000052, -0.00000802, 0.00000615, -0.00000602, -0.00000457, -0.00001106, 0.00001211, -0.00000237, 0.00000451, -0.00000347, -0.00000623, 0.00000627, -0.00000544, 0.00000361, 0.00000421, 0.00001729, -0.00000739, 0.00000481, -0.00000928, -0.00000789, 0.00000167, -0.00001199, 0.00000231, -0.00000493, -0.00000318, -0.00000266, 0.00000478, -0.00000462, 0.00000204, -0.00000698, 0.00000529, -0.00000384, -0.00000274, -0.00000281, -0.00000486, 0.00000151, 0.00000456, -0.00000414, -0.00000580, 0.00000156, 0.00000117, 0.00000025, -0.00000383, -0.00000065, 0.00000524, -0.00000918, 0.00000464, -0.00000382, -0.00000036, -0.00000728, 0.00001204, -0.00000779, -0.00000207, 0.00000190, },
		{ 0.00000960, -0.00000121, -0.00000506, 0.00000047, 0.00000858, 0.00000004, -0.00000529, 0.00000265, -0.00000510, 0.00000047, -0.00000140, -0.00000104, 0.00000193, -0.00000034, 0.00000179, 0.00000387, -0.00000205, 0.00000098, -0.00000053, -0.00000588, 0.00001260, -0.00000295, 0.00000016, 0.00001214, -0.00002560, 0.00001693, -0.00001604, 0.00001139, -0.00001141, 0.00000631, -0.00001706, 0.00000613, -0.00000572, 0.00000347, -0.00000737, 0.00000621, -0.00001652, 0.00000434, -0.00000795, 0.00001126, -0.00001625, 0.00000214, -0.00000867, 0.00000395, -0.00001360, 0.00000365, -0.00001717, 0.00001277, -0.00002870, 0.00000372, -0.00002271, 0.00000403, -0.00004653, 0.00000071, -0.00007249, 0.00001157, -0.00026471, -0.00004163, 0.00034584, 0.00006356, 0.00013356, 0.00006600, 0.00015888, 0.00030976, -0.00022043, -0.00013106, -0.00003364, -0.00004116, 0.00000242, -0.00001459, -0.00000450, -0.00003559, 0.00002440, -0.00001168, -0.00000219, -0.00000830, 0.00000578, -0.00000451, -0.00000525, -0.00001229, 0.00001198, -0.00000175, 0.00000237, 0.00000051, -0.00000854, 0.00000490, -0.00000425, 0.00000087, 0.00000606, 0.00001236, 0.00000132, -0.00000407, 0.00000303, -0.00001790, 0.00000763, -0.00001501, 0.00000484, -0.00000759, -0.00000004, -0.00000721, 0.00000828, -0.00000649, 0.00000360, -0.00000778, 0.00000546, -0.00000328, -0.00000253, -0.00000350, -0.00000311, -0.00000280, 0.00000688, -0.00000326, -0.00000400, -0.00000312, 0.00000477, -0.00000248, 0.00000093, -0.00000712, 0.00001011, -0.00001067, 0.00000477, -0.00000442, 0.00000171, -0.00000955, 0.00001052, -0.00000285, -0.00000516, 0.00000083, },
		{ 0.00000696, 0.00000394, -0.00000723, -0.00000004, 0.00000572, 0.00000531, -0.00000788, 0.00000354, -0.00000484, -0.00000048, -0.00000089, -0.00000117, 0.00000096, 0.00000102, -0.00000016, 0.00000559, -0.00000277, 0.00000119, -0.00000048, -0.00000446, 0.00000537, 0.00000788, -0.00001098, 0.00002243, -0.00002856, 0.00001339, -0.00001069, 0.00000556, -0.00000581, 0.00000181, -0.00001015, -0.00000202, -0.00000131, 0.00000082, -0.00000471, 0.00000380, -0.00001118, -0.00000260, -0.00000250, 0.00000575, -0.00000940, -0.00000441, -0.00000378, -0.00000007, -0.00000919, -0.00000077, -0.00001103, 0.00000576, -0.00002029, -0.00000547, -0.00001394, -0.00000323, -0.00003474, -0.00001265, -0.00005070, -0.00001059, -0.00017546, -0.00021183, 0.00036506, 0.00012003, 0.00011941, 0.00009312, 0.00014760, 0.00034401, -0.00023596, -0.00013056, -0.00004054, -0.00004023, -0.00000068, -0.00001378, -0.00000638, -0.00003402, 0.00001902, -0.00000520, -0.00000707, -0.00000590, 0.00000279, -0.00000115, -0.00000690, -0.00001235, 0.00001036, 0.00000016, -0.00000045, 0.00000342, -0.00000799, 0.00000082, -0.00000083, -0.00000398, 0.00001037, 0.00000488, 0.00001108, -0.00001234, 0.00001201, -0.00002197, 0.00000736, -0.00001299, 0.00000365, -0.00000717, 0.00000025, -0.00000842, 0.00000781, -0.00000464, 0.00000192, -0.00000577, 0.00000344, -0.00000134, -0.00000305, -0.00000378, -0.00000168, -0.00000573, 0.00000683, -0.00000130, -0.00000213, -0.00000681, 0.00000631, -0.00000336, 0.00000366, -0.00001000, 0.00000989, -0.00000727, 0.00000126, -0.00000215, 0.00000057, -0.00000750, 0.00000469, 0.00000423, -0.00000843, 0.00000066, },
		{ 0.00000389, 0.00000804, -0.00000751, -0.00000107, 0.00000276, 0.00000893, -0.00000764, 0.00000197, -0.00000282, -0.00000233, -0.00000010, -0.00000128, 0.00000002, 0.00000199, -0.00000131, 0.00000608, -0.00000230, 0.00000075, -0.00000062, -0.00000204, -0.00000171, 0.00001514, -0.00001500, 0.00002231, -0.00001990, 0.00000038, 0.00000129, -0.00000501, 0.00000340, -0.00000614, -0.00000035, -0.00001111, 0.00000331, -0.00000315, 0.00000030, -0.00000185, -0.00000254, -0.00001161, 0.00000444, -0.00000140, -0.00000130, -0.00001162, 0.00000196, -0.00000582, -0.00000223, -0.00000851, -0.00000115, -0.00000488, -0.00000833, -0.00001807, -0.00000135, -0.00001616, -0.00001555, -0.00003492, -0.00001871, -0.00004777, -0.00008148, -0.00032320, 0.00028762, 0.00022866, 0.00008456, 0.00013677, 0.00012528, 0.00038949, -0.00025897, -0.00012589, -0.00005078, -0.00003618, -0.00000732, -0.00000904, -0.00001257, -0.00002810, 0.00001008, 0.00000302, -0.00001243, -0.00000299, -0.00000087, 0.00000230, -0.00000826, -0.00001206, 0.00000807, 0.00000234, -0.00000278, 0.00000437, -0.00000488, -0.00000445, 0.00000266, -0.00000842, 0.00001398, -0.00000103, 0.00001703, -0.00001489, 0.00001309, -0.00001728, -0.00000013, -0.00000599, -0.00000105, -0.00000365, -0.00000257, -0.00000600, 0.00000350, 0.00000030, -0.00000211, -0.00000194, 0.00000021, 0.00000121, -0.00000397, -0.00000345, -0.00000135, -0.00000633, 0.00000405, 0.00000186, -0.00000103, -0.00000810, 0.00000461, -0.00000140, 0.00000319, -0.00000830, 0.00000460, -0.00000013, -0.00000434, 0.00000167, -0.00000259, -0.00000287, -0.00000272, 0.00001042, -0.00000972, -0.00000000, },
		{ 0.00000197, 0.00000949, -0.00000503, -0.00000317, 0.00000084, 0.00000969, -0.00000425, -0.00000156, 0.00000010, -0.00000430, 0.00000048, -0.00000127, -0.00000062, 0.00000216, -0.00000124, 0.00000510, -0.00000060, -0.00000041, -0.00000041, -0.00000026, -0.00000582, 0.00001593, -0.00001033, 0.00001266, -0.00000358, -0.00001598, 0.00001328, -0.00001404, 0.00001047, -0.00001198, 0.00000671, -0.00001596, 0.00000405, -0.00000494, 0.00000398, -0.00000642, 0.00000431, -0.00001736, 0.00000775, -0.00000516, 0.00000300, -0.00001470, 0.00000408, -0.00000879, 0.00000237, -0.00001398, 0.00000602, -0.00001196, -0.00000064, -0.00002576, 0.00000617, -0.00002464, -0.00000138, -0.00005040, 0.00000330, -0.00007111, -0.00002405, -0.00034151, 0.00011665, 0.00035956, 0.00005553, 0.00017836, 0.00010691, 0.00043398, -0.00027845, -0.00012659, -0.00005669, -0.00003519, -0.00001219, -0.00000533, -0.00001810, -0.00002307, 0.00000296, 0.00000800, -0.00001429, -0.00000272, -0.00000244, 0.00000345, -0.00000761, -0.00001276, 0.00000651, 0.00000339, -0.00000344, 0.00000302, -0.00000019, -0.00000903, 0.00000419, -0.00001035, 0.00001453, -0.00000263, 0.00001658, -0.00000993, 0.00000611, -0.00000574, -0.00001149, 0.00000244, -0.00000623, 0.00000053, -0.00000639, -0.00000208, -0.00000214, 0.00000559, -0.00000587, 0.00000132, -0.00000218, 0.00000286, -0.00000435, -0.00000295, -0.00000193, -0.00000500, -0.00000043, 0.00000515, -0.00000041, -0.00000709, 0.00000034, 0.00000240, 0.00000055, -0.00000361, -0.00000304, 0.00000730, -0.00000879, 0.00000434, -0.00000521, 0.00000129, -0.00000813, 0.00001264, -0.00000729, -0.00000227, },
		{ 0.00000156, 0.00000807, -0.00000036, -0.00000579, -0.00000000, 0.00000784, 0.00000106, -0.00000529, 0.00000234, -0.00000522, 0.00000019, -0.00000089, -0.00000105, 0.00000163, -0.00000028, 0.00000327, 0.00000150, -0.00000161, -0.00000005, 0.00000040, -0.00000636, 0.00001098, -0.00000007, -0.00000069, 0.00001234, -0.00002711, 0.00001772, -0.00001535, 0.00001023, -0.00001110, 0.00000670, -0.00001310, -0.00000152, -0.00000226, 0.00000343, -0.00000634, 0.00000532, -0.00001596, 0.00000405, -0.00000231, 0.00000037, -0.00001080, -0.00000009, -0.00000600, 0.00000098, -0.00001280, 0.00000542, -0.00000998, -0.00000296, -0.00002256, 0.00000225, -0.00002115, -0.00000159, -0.00004766, 0.00000153, -0.00006298, -0.00002162, -0.00027034, -0.00010462, 0.00045767, 0.00006853, 0.00019570, 0.00010927, 0.00046441, -0.00028317, -0.00014189, -0.00005118, -0.00004310, -0.00000993, -0.00000789, -0.00001769, -0.00002421, 0.00000265, 0.00000569, -0.00000976, -0.00000723, -0.00000004, 0.00000077, -0.00000386, -0.00001541, 0.00000679, 0.00000223, -0.00000178, -0.00000042, 0.00000487, -0.00001157, 0.00000299, -0.00000944, 0.00001188, -0.00000011, 0.00001088, -0.00000010, -0.00000450, 0.00000664, -0.00002045, 0.00000692, -0.00000772, 0.00000200, -0.00000815, 0.00000019, -0.00000573, 0.00000796, -0.00000655, 0.00000168, -0.00000189, 0.00000226, -0.00000329, -0.00000306, -0.00000254, -0.00000313, -0.00000460, 0.00000683, 0.00000084, -0.00000512, -0.00000438, 0.00000563, -0.00000178, 0.00000098, -0.00000913, 0.00001111, -0.00000899, 0.00000352, -0.00000509, 0.00000248, -0.00000912, 0.00000955, -0.00000114, -0.00000594, },
		{ 0.00000165, 0.00000503, 0.00000468, -0.00000749, -0.00000067, 0.00000480, 0.00000607, -0.00000710, 0.00000240, -0.00000423, -0.00000134, 0.00000002, -0.00000149, 0.00000091, 0.00000075, 0.00000170, 0.00000280, -0.00000182, -0.00000022, 0.00000055, -0.00000475, 0.00000368, 0.00001009, -0.00001032, 0.00002000, -0.00002660, 0.00001064, -0.00000686, 0.00000167, -0.00000292, -0.00000057, -0.00000320, -0.00001209, 0.00000401, -0.00000125, -0.00000119, 0.00000005, -0.00000758, -0.00000585, 0.00000620, -0.00000815, -0.00000124, -0.00000926, 0.00000171, -0.00000614, -0.00000486, -0.00000320, 0.00000108, -0.00001496, -0.00000911, -0.00001236, -0.00000608, -0.00001624, -0.00002684, -0.00002284, -0.00002653, -0.00006242, -0.00014967, -0.00030701, 0.00046446, 0.00014926, 0.00017840, 0.00013931, 0.00047627, -0.00026945, -0.00017421, -0.00003304, -0.00006094, 0.00000076, -0.00001837, -0.00000946, -0.00003332, 0.00001041, -0.00000422, 0.00000060, -0.00001582, 0.00000585, -0.00000526, 0.00000238, -0.00001971, 0.00000900, -0.00000117, 0.00000183, -0.00000503, 0.00000916, -0.00001169, -0.00000020, -0.00000720, 0.00000800, 0.00000386, 0.00000386, 0.00000921, -0.00001218, 0.00001307, -0.00002134, 0.00000356, -0.00000299, -0.00000118, -0.00000597, -0.00000124, -0.00000512, 0.00000563, -0.00000295, -0.00000163, 0.00000149, -0.00000070, -0.00000068, -0.00000411, -0.00000238, -0.00000218, -0.00000665, 0.00000554, 0.00000361, -0.00000367, -0.00000730, 0.00000603, -0.00000160, 0.00000284, -0.00001081, 0.00000915, -0.00000398, -0.00000105, -0.00000194, 0.00000027, -0.00000575, 0.00000247, 0.00000655, -0.00000925, },
		{ 0.00000065, 0.00000229, 0.00000797, -0.00000682, -0.00000219, 0.00000217, 0.00000880, -0.00000563, -0.00000018, -0.00000148, -0.00000372, 0.00000120, -0.00000198, 0.00000043, 0.00000112, 0.00000127, 0.00000247, -0.00000046, -0.00000142, 0.00000102, -0.00000300, -0.00000216, 0.00001512, -0.00001129, 0.00001605, -0.00001402, -0.00000563, 0.00000762, -0.00001095, 0.00000826, -0.00001064, 0.00000878, -0.00002271, 0.00000982, -0.00000681, 0.00000590, -0.00000793, 0.00000337, -0.00001703, 0.00001543, -0.00001756, 0.00000884, -0.00001844, 0.00000979, -0.00001464, 0.00000535, -0.00001479, 0.00001539, -0.00002996, 0.00000714, -0.00002957, 0.00001200, -0.00003571, 0.00000021, -0.00005436, 0.00001728, -0.00011315, -0.00003684, -0.00042389, 0.00034669, 0.00029217, 0.00014028, 0.00018615, 0.00047938, -0.00024552, -0.00021589, -0.00000942, -0.00008277, 0.00001517, -0.00003292, 0.00000316, -0.00004698, 0.00002246, -0.00001742, 0.00001240, -0.00002468, 0.00001214, -0.00001190, 0.00000871, -0.00002386, 0.00001180, -0.00000548, 0.00000590, -0.00000928, 0.00001166, -0.00000972, -0.00000375, -0.00000581, 0.00000550, 0.00000600, -0.00000026, 0.00001310, -0.00001209, 0.00000986, -0.00001257, -0.00000734, 0.00000670, -0.00000773, -0.00000076, -0.00000577, -0.00000089, -0.00000042, 0.00000351, -0.00000701, 0.00000637, -0.00000458, 0.00000238, -0.00000556, -0.00000141, -0.00000269, -0.00000596, 0.00000130, 0.00000767, -0.00000327, -0.00000743, 0.00000294, 0.00000152, 0.00000137, -0.00000777, 0.00000227, 0.00000425, -0.00000711, 0.00000227, -0.00000349, -0.00000045, -0.00000523, 0.00001233, -0.00000982, },
		{ -0.00000228, 0.00000111, 0.00000828, -0.00000335, -0.00000476, 0.00000074, 0.00000851, -0.00000115, -0.00000434, 0.00000176, -0.00000586, 0.00000193, -0.00000224, 0.00000022, 0.00000065, 0.00000212, 0.00000064, 0.00000208, -0.00000329, 0.00000195, -0.00000228, -0.00000447, 0.00001331, -0.00000363, 0.00000332, 0.00000464, -0.00002318, 0.00002001, -0.00002015, 0.00001558, -0.00001691, 0.00001622, -0.00002739, 0.00001032, -0.00000890, 0.00001026, -0.00001325, 0.00001079, -0.00002318, 0.00001936, -0.00002189, 0.00001341, -0.00002183, 0.00001267, -0.00001879, 0.00001145, -0.00002210, 0.00002463, -0.00003875, 0.00001620, -0.00003868, 0.00002157, -0.00004677, 0.00001734, -0.00007294, 0.00004303, -0.00013869, 0.00002101, -0.00042126, 0.00011828, 0.00045177, 0.00011964, 0.00022343, 0.00049502, -0.00022847, -0.00025227, 0.00000704, -0.00009824, 0.00002468, -0.00004394, 0.00001306, -0.00005819, 0.00003172, -0.00002695, 0.00001937, -0.00002876, 0.00001474, -0.00001555, 0.00001199, -0.00002543, 0.00001314, -0.00000859, 0.00000836, -0.00001143, 0.00001161, -0.00000633, -0.00000621, -0.00000675, 0.00000581, 0.00000466, 0.00000045, 0.00001002, -0.00000392, -0.00000128, 0.00000199, -0.00002073, 0.00001639, -0.00001353, 0.00000410, -0.00001043, 0.00000398, -0.00000687, 0.00000939, -0.00001125, 0.00000990, -0.00000699, 0.00000408, -0.00000623, -0.00000039, -0.00000406, -0.00000350, -0.00000423, 0.00001144, -0.00000311, -0.00000549, -0.00000228, 0.00000590, -0.00000172, -0.00000234, -0.00000605, 0.00001165, -0.00001102, 0.00000454, -0.00000592, 0.00000344, -0.00000990, 0.00001327, -0.00000619, },
		{ -0.00000639, 0.00000132, 0.00000598, 0.00000180, -0.00000741, 0.00000012, 0.00000599, 0.00000457, -0.00000793, 0.00000364, -0.00000642, 0.00000142, -0.00000188, 0.00000000, -0.00000026, 0.00000354, -0.00000166, 0.00000455, -0.00000470, 0.00000268, -0.00000240, -0.00000367, 0.00000668, 0.00000807, -0.00001088, 0.00002013, -0.00003272, 0.00002249, -0.00001970, 0.00001388, -0.00001476, 0.00001489, -0.00002282, 0.00000326, -0.00000526, 0.00000896, -0.00001230, 0.00001065, -0.00002056, 0.00001473, -0.00001802, 0.00000941, -0.00001665, 0.00000763, -0.00001537, 0.00000945, -0.00002030, 0.00002333, -0.00003542, 0.00001195, -0.00003333, 0.00001579, -0.00004131, 0.00001455, -0.00006660, 0.00003669, -0.00012209, 0.00000922, -0.00031423, -0.00015766, 0.00055400, 0.00016334, 0.00022364, 0.00054409, -0.00023417, -0.00027041, 0.00000572, -0.00009900, 0.00002228, -0.00004497, 0.00001395, -0.00006072, 0.00003216, -0.00002731, 0.00001711, -0.00002489, 0.00001120, -0.00001398, 0.00001034, -0.00002295, 0.00001161, -0.00000889, 0.00000768, -0.00001048, 0.00000890, -0.00000227, -0.00000688, -0.00000997, 0.00000836, 0.00000070, 0.00000466, 0.00000244, 0.00000806, -0.00001414, 0.00001491, -0.00002935, 0.00002010, -0.00001397, 0.00000489, -0.00001182, 0.00000603, -0.00001008, 0.00001126, -0.00001147, 0.00000973, -0.00000608, 0.00000310, -0.00000512, -0.00000027, -0.00000510, -0.00000107, -0.00000868, 0.00001288, -0.00000178, -0.00000321, -0.00000708, 0.00000880, -0.00000368, 0.00000209, -0.00001171, 0.00001432, -0.00000990, 0.00000278, -0.00000508, 0.00000374, -0.00000959, 0.00000871, 0.00000098, },
		{ -0.00000945, 0.00000148, 0.00000277, 0.00000653, -0.00000850, -0.00000079, 0.00000289, 0.00000916, -0.00000884, 0.00000282, -0.00000475, -0.00000052, -0.00000081, -0.00000046, -0.00000103, 0.00000460, -0.00000318, 0.00000560, -0.00000452, 0.00000231, -0.00000238, -0.00000171, -0.00000074, 0.00001762, -0.00001899, 0.00002501, -0.00002880, 0.00001230, -0.00000876, 0.00000334, -0.00000481, 0.00000583, -0.00001096, -0.00000892, 0.00000221, 0.00000305, -0.00000570, 0.00000368, -0.00001056, 0.00000357, -0.00000811, -0.00000088, -0.00000538, -0.00000291, -0.00000632, 0.00000075, -0.00001054, 0.00001267, -0.00002160, -0.00000346, -0.00001620, -0.00000236, -0.00002223, -0.00000519, -0.00003932, 0.00000448, -0.00007357, -0.00004930, -0.00016065, -0.00039070, 0.00052641, 0.00029951, 0.00017823, 0.00063217, -0.00026535, -0.00026887, -0.00001386, -0.00008544, 0.00000826, -0.00003583, 0.00000521, -0.00005374, 0.00002308, -0.00001835, 0.00000638, -0.00001454, 0.00000298, -0.00000830, 0.00000477, -0.00001742, 0.00000777, -0.00000650, 0.00000405, -0.00000696, 0.00000455, 0.00000137, -0.00000569, -0.00001427, 0.00001119, -0.00000337, 0.00000905, -0.00000494, 0.00001764, -0.00002157, 0.00001930, -0.00002794, 0.00001455, -0.00000726, 0.00000037, -0.00000865, 0.00000373, -0.00000846, 0.00000791, -0.00000704, 0.00000564, -0.00000195, -0.00000032, -0.00000238, -0.00000128, -0.00000500, -0.00000013, -0.00001029, 0.00001077, 0.00000148, -0.00000202, -0.00000934, 0.00000819, -0.00000260, 0.00000322, -0.00001229, 0.00001078, -0.00000352, -0.00000259, -0.00000131, 0.00000071, -0.00000518, 0.00000077, 0.00000886, },
		{ -0.00000903, -0.00000003, 0.00000054, 0.00000885, -0.00000684, -0.00000280, 0.00000063, 0.00001097, -0.00000623, -0.00000060, -0.00000148, -0.00000314, 0.00000040, -0.00000097, -0.00000144, 0.00000479, -0.00000326, 0.00000471, -0.00000250, 0.00000067, -0.00000163, -0.00000038, -0.00000559, 0.00002082, -0.00001737, 0.00001773, -0.00001311, -0.00000606, 0.00000689, -0.00001002, 0.00000699, -0.00000484, 0.00000164, -0.00001976, 0.00000808, -0.00000277, 0.00000187, -0.00000488, 0.00000071, -0.00000751, 0.00000125, -0.00001079, 0.00000509, -0.00001220, 0.00000197, -0.00000831, 0.00000042, 0.00000023, -0.00000595, -0.00002020, 0.00000182, -0.00002101, -0.00000234, -0.00002727, -0.00000897, -0.00003073, -0.00002350, -0.00010861, -0.00003388, -0.00050092, 0.00033486, 0.00051228, 0.00011190, 0.00074096, -0.00030573, -0.00026192, -0.00003903, -0.00006885, -0.00000788, -0.00002438, -0.00000641, -0.00004340, 0.00001048, -0.00000630, -0.00000634, -0.00000398, -0.00000453, -0.00000301, -0.00000074, -0.00001232, 0.00000439, -0.00000369, -0.00000030, -0.00000322, 0.00000082, 0.00000310, -0.00000267, -0.00001833, 0.00001245, -0.00000534, 0.00001074, -0.00000832, 0.00002047, -0.00001968, 0.00001298, -0.00001668, 0.00000182, 0.00000379, -0.00000690, -0.00000297, -0.00000125, -0.00000362, 0.00000135, -0.00000033, 0.00000005, 0.00000312, -0.00000415, 0.00000047, -0.00000256, -0.00000403, -0.00000088, -0.00000898, 0.00000577, 0.00000590, -0.00000194, -0.00000866, 0.00000412, 0.00000124, 0.00000117, -0.00000833, 0.00000282, 0.00000525, -0.00000862, 0.00000283, -0.00000318, 0.00000028, -0.00000669, 0.00001374, },
		{ -0.00000412, -0.00000362, 0.00000010, 0.00000810, -0.00000259, -0.00000562, -0.00000049, 0.00000989, -0.00000106, -0.00000495, 0.00000160, -0.00000492, 0.00000072, -0.00000095, -0.00000182, 0.00000436, -0.00000232, 0.00000260, 0.00000032, -0.00000124, -0.00000066, -0.00000001, -0.00000696, 0.00001753, -0.00000804, 0.00000336, 0.00000617, -0.00002286, 0.00001771, -0.00001749, 0.00001261, -0.00000941, 0.00000733, -0.00002240, 0.00000664, -0.00000318, 0.00000462, -0.00000849, 0.00000597, -0.00001109, 0.00000289, -0.00001315, 0.00000750, -0.00001315, 0.00000249, -0.00001031, 0.00000427, -0.00000450, 0.00000081, -0.00002649, 0.00000798, -0.00002644, 0.00000340, -0.00003445, 0.00000362, -0.00004342, -0.00000369, -0.00012621, 0.00001235, -0.00045545, 0.00000840, 0.00073395, 0.00007972, 0.00083417, -0.00032600, -0.00027384, -0.00004939, -0.00006652, -0.00001184, -0.00002277, -0.00001007, -0.00003981, 0.00000412, -0.00000066, -0.00001197, -0.00000126, -0.00000471, -0.00000364, -0.00000125, -0.00001203, 0.00000508, -0.00000369, -0.00000209, -0.00000241, 0.00000038, 0.00000132, 0.00000234, -0.00002162, 0.00001174, -0.00000490, 0.00000925, -0.00000711, 0.00001657, -0.00001011, -0.00000006, -0.00000162, -0.00001122, 0.00001274, -0.00001155, 0.00000079, -0.00000499, 0.00000053, -0.00000424, 0.00000449, -0.00000322, 0.00000578, -0.00000563, 0.00000129, -0.00000257, -0.00000337, -0.00000215, -0.00000636, 0.00000017, 0.00000933, -0.00000161, -0.00000641, -0.00000133, 0.00000548, -0.00000171, -0.00000285, -0.00000542, 0.00001190, -0.00001140, 0.00000418, -0.00000493, 0.00000339, -0.00001014, 0.00001311, },
		{ 0.00000385, -0.00000794, 0.00000069, 0.00000528, 0.00000260, -0.00000784, -0.00000131, 0.00000723, 0.00000440, -0.00000772, 0.00000248, -0.00000444, -0.00000074, 0.00000017, -0.00000269, 0.00000406, -0.00000144, 0.00000078, 0.00000218, -0.00000178, -0.00000076, 0.00000040, -0.00000635, 0.00001112, 0.00000298, -0.00000948, 0.00001891, -0.00002844, 0.00001591, -0.00001303, 0.00000713, -0.00000348, 0.00000217, -0.00001392, -0.00000406, 0.00000393, -0.00000032, -0.00000358, 0.00000132, -0.00000360, -0.00000625, -0.00000499, -0.00000110, -0.00000299, -0.00000757, -0.00000180, -0.00000336, 0.00000366, -0.00000717, -0.00001610, -0.00000416, -0.00001198, -0.00001221, -0.00001820, -0.00001192, -0.00002149, -0.00002773, -0.00008687, -0.00003090, -0.00028519, -0.00036267, 0.00086363, 0.00014322, 0.00087851, -0.00030079, -0.00032460, -0.00002909, -0.00009119, 0.00000647, -0.00003960, 0.00000215, -0.00005057, 0.00001134, -0.00000836, -0.00000436, -0.00001128, 0.00000607, -0.00001310, 0.00000598, -0.00001896, 0.00001187, -0.00000857, 0.00000098, -0.00000672, 0.00000483, -0.00000463, 0.00000923, -0.00002434, 0.00001015, -0.00000372, 0.00000664, -0.00000398, 0.00000996, 0.00000111, -0.00001190, 0.00000849, -0.00001651, 0.00001317, -0.00000880, -0.00000120, -0.00000392, 0.00000034, -0.00000512, 0.00000396, -0.00000131, 0.00000375, -0.00000298, -0.00000121, -0.00000024, -0.00000418, -0.00000242, -0.00000465, -0.00000329, 0.00000950, 0.00000061, -0.00000468, -0.00000530, 0.00000719, -0.00000253, 0.00000065, -0.00000986, 0.00001279, -0.00000849, 0.00000110, -0.00000301, 0.00000250, -0.00000836, 0.00000710, },
		{ 0.00001149, -0.00001036, 0.00000053, 0.00000240, 0.00000646, -0.00000781, -0.00000297, 0.00000474, 0.00000781, -0.00000703, 0.00000012, -0.00000137, -0.00000392, 0.00000231, -0.00000421, 0.00000444, -0.00000159, 0.00000052, 0.00000174, 0.00000018, -0.00000286, 0.00000198, -0.00000602, 0.00000589, 0.00000947, -0.00001366, 0.00001865, -0.00001884, 0.00000044, 0.00000254, -0.00000772, 0.00001082, -0.00001124, 0.00000222, -0.00002016, 0.00001538, -0.00001062, 0.00000786, -0.00001100, 0.00001195, -0.00002258, 0.00000999, -0.00001688, 0.00001422, -0.00002423, 0.00001375, -0.00001943, 0.00002177, -0.00002662, 0.00000700, -0.00002979, 0.00001674, -0.00004289, 0.00001486, -0.00004815, 0.00002551, -0.00008217, -0.00001077, -0.00012545, -0.00007198, -0.00065876, 0.00080898, 0.00033339, 0.00087054, -0.00022825, -0.00041470, 0.00002089, -0.00014089, 0.00004452, -0.00007267, 0.00002871, -0.00007466, 0.00003140, -0.00002855, 0.00001507, -0.00003182, 0.00002507, -0.00002892, 0.00001895, -0.00003134, 0.00002318, -0.00001729, 0.00000835, -0.00001552, 0.00001324, -0.00001345, 0.00001687, -0.00002656, 0.00000915, -0.00000406, 0.00000574, -0.00000263, 0.00000576, 0.00000735, -0.00001528, 0.00000727, -0.00000972, 0.00000292, 0.00000207, -0.00000926, 0.00000243, -0.00000492, -0.00000046, -0.00000233, 0.00000563, -0.00000250, 0.00000311, -0.00000635, 0.00000397, -0.00000651, -0.00000098, -0.00000521, -0.00000311, 0.00000554, 0.00000522, -0.00000464, -0.00000598, 0.00000478, 0.00000014, 0.00000037, -0.00000872, 0.00000726, -0.00000054, -0.00000523, 0.00000152, -0.00000141, -0.00000300, -0.00000140, },
		{ 0.00001511, -0.00000863, -0.00000180, 0.00000110, 0.00000737, -0.00000479, -0.00000588, 0.00000342, 0.00000809, -0.00000275, -0.00000466, 0.00000304, -0.00000753, 0.00000445, -0.00000581, 0.00000531, -0.00000282, 0.00000195, -0.00000094, 0.00000421, -0.00000645, 0.00000471, -0.00000705, 0.00000422, 0.00000865, -0.00000749, 0.00000626, 0.00000160, -0.00002169, 0.00002115, -0.00002385, 0.00002552, -0.00002480, 0.00001751, -0.00003341, 0.00002399, -0.00001977, 0.00001915, -0.00002365, 0.00002732, -0.00003733, 0.00002302, -0.00003093, 0.00002926, -0.00003835, 0.00002735, -0.00003474, 0.00003983, -0.00004636, 0.00003016, -0.00005474, 0.00004408, -0.00007153, 0.00004595, -0.00008347, 0.00007132, -0.00013394, 0.00005888, -0.00020841, 0.00008919, -0.00078110, 0.00053366, 0.00061856, 0.00085394, -0.00013925, -0.00051927, 0.00007871, -0.00019632, 0.00008528, -0.00010750, 0.00005726, -0.00010120, 0.00005428, -0.00005158, 0.00003680, -0.00005344, 0.00004347, -0.00004344, 0.00003112, -0.00004333, 0.00003381, -0.00002552, 0.00001626, -0.00002513, 0.00002189, -0.00002167, 0.00002275, -0.00002746, 0.00000917, -0.00000679, 0.00000778, -0.00000498, 0.00000675, 0.00000553, -0.00000803, -0.00000532, 0.00000668, -0.00001379, 0.00001635, -0.00001918, 0.00001059, -0.00001231, 0.00000686, -0.00001118, 0.00001418, -0.00000960, 0.00000955, -0.00001144, 0.00000799, -0.00000902, 0.00000142, -0.00000761, 0.00000008, -0.00000115, 0.00001075, -0.00000570, -0.00000371, -0.00000089, 0.00000520, -0.00000275, -0.00000345, -0.00000195, 0.00000878, -0.00001123, 0.00000561, -0.00000538, 0.00000241, -0.00000826, },
		{ 0.00001280, -0.00000244, -0.00000613, 0.00000157, 0.00000542, 0.00000034, -0.00000912, 0.00000292, 0.00000590, 0.00000334, -0.00000944, 0.00000640, -0.00000957, 0.00000511, -0.00000640, 0.00000578, -0.00000420, 0.00000391, -0.00000425, 0.00000836, -0.00000960, 0.00000704, -0.00000854, 0.00000543, 0.00000240, 0.00000469, -0.00001070, 0.00002243, -0.00003887, 0.00003178, -0.00003124, 0.00003138, -0.00002956, 0.00002316, -0.00003589, 0.00002297, -0.00002117, 0.00002309, -0.00002856, 0.00003370, -0.00004158, 0.00002538, -0.00003440, 0.00003311, -0.00004091, 0.00002990, -0.00003944, 0.00004679, -0.00005388, 0.00003938, -0.00006368, 0.00005342, -0.00008018, 0.00005534, -0.00009514, 0.00008877, -0.00015019, 0.00008172, -0.00022732, 0.00013534, -0.00069739, 0.00008741, 0.00089664, 0.00091283, -0.00008674, -0.00059851, 0.00011138, -0.00022965, 0.00010521, -0.00012436, 0.00007077, -0.00011492, 0.00006581, -0.00006401, 0.00004799, -0.00006418, 0.00005067, -0.00004773, 0.00003481, -0.00004794, 0.00003749, -0.00002782, 0.00001966, -0.00003046, 0.00002593, -0.00002507, 0.00002387, -0.00002550, 0.00000923, -0.00001067, 0.00001138, -0.00000974, 0.00001159, -0.00000218, 0.00000573, -0.00002263, 0.00002421, -0.00002809, 0.00002603, -0.00002436, 0.00001505, -0.00001685, 0.00001194, -0.00001753, 0.00001938, -0.00001309, 0.00001246, -0.00001324, 0.00000927, -0.00000970, 0.00000303, -0.00001004, 0.00000388, -0.00000751, 0.00001444, -0.00000586, -0.00000057, -0.00000700, 0.00000966, -0.00000579, 0.00000237, -0.00001020, 0.00001468, -0.00001297, 0.00000615, -0.00000647, 0.00000467, -0.00001035, },
		{ 0.00000562, 0.00000602, -0.00001044, 0.00000242, 0.00000224, 0.00000542, -0.00001089, 0.00000200, 0.00000304, 0.00000853, -0.00001148, 0.00000657, -0.00000853, 0.00000330, -0.00000521, 0.00000501, -0.00000459, 0.00000481, -0.00000616, 0.00001028, -0.00001006, 0.00000701, -0.00000860, 0.00000686, -0.00000460, 0.00001555, -0.00002260, 0.00003296, -0.00004144, 0.00002696, -0.00002429, 0.00002384, -0.00002151, 0.00001566, -0.00002513, 0.00001072, -0.00001290, 0.00001688, -0.00002226, 0.00002738, -0.00003201, 0.00001419, -0.00002444, 0.00002286, -0.00002912, 0.00001838, -0.00002976, 0.00003784, -0.00004349, 0.00002838, -0.00005002, 0.00003802, -0.00006186, 0.00003549, -0.00007423, 0.00006725, -0.00011893, 0.00004507, -0.00016968, 0.00006553, -0.00046025, -0.00039954, 0.00102247, 0.00113734, -0.00011525, -0.00062238, 0.00009513, -0.00022190, 0.00008911, -0.00011103, 0.00005874, -0.00010615, 0.00005685, -0.00005718, 0.00004065, -0.00005705, 0.00004108, -0.00003751, 0.00002638, -0.00004170, 0.00003104, -0.00002134, 0.00001557, -0.00002833, 0.00002236, -0.00002128, 0.00001878, -0.00001985, 0.00000799, -0.00001334, 0.00001355, -0.00001347, 0.00001611, -0.00001012, 0.00001822, -0.00003517, 0.00003307, -0.00003139, 0.00002451, -0.00002002, 0.00001199, -0.00001496, 0.00001111, -0.00001762, 0.00001787, -0.00001024, 0.00000969, -0.00001009, 0.00000654, -0.00000740, 0.00000247, -0.00001062, 0.00000571, -0.00001057, 0.00001387, -0.00000333, 0.00000114, -0.00001052, 0.00001055, -0.00000590, 0.00000528, -0.00001361, 0.00001401, -0.00000861, 0.00000210, -0.00000375, 0.00000278, -0.00000728, },
		{ -0.00000293, 0.00001297, -0.00001194, 0.00000173, -0.00000009, 0.00000822, -0.00000971, -0.00000034, 0.00000111, 0.00001078, -0.00000937, 0.00000307, -0.00000460, -0.00000051, -0.00000266, 0.00000310, -0.00000365, 0.00000392, -0.00000562, 0.00000892, -0.00000705, 0.00000398, -0.00000629, 0.00000642, -0.00000835, 0.00001951, -0.00002345, 0.00002851, -0.00002762, 0.00000793, -0.00000602, 0.00000672, -0.00000481, -0.00000033, -0.00000664, -0.00000700, 0.00000006, 0.00000470, -0.00000874, 0.00001269, -0.00001379, -0.00000475, -0.00000695, 0.00000459, -0.00000935, -0.00000086, -0.00001161, 0.00001866, -0.00002113, 0.00000378, -0.00002148, 0.00000685, -0.00002682, -0.00000217, -0.00003328, 0.00002122, -0.00005843, -0.00002649, -0.00007067, -0.00005954, -0.00018735, -0.00076052, 0.00086450, 0.00157259, -0.00022320, -0.00059776, 0.00003624, -0.00017964, 0.00004380, -0.00007426, 0.00002702, -0.00007972, 0.00003145, -0.00003471, 0.00001843, -0.00003607, 0.00001930, -0.00001758, 0.00001012, -0.00002819, 0.00001764, -0.00000889, 0.00000613, -0.00002032, 0.00001277, -0.00001215, 0.00000958, -0.00001219, 0.00000569, -0.00001359, 0.00001232, -0.00001360, 0.00001678, -0.00001330, 0.00002311, -0.00003638, 0.00002812, -0.00002097, 0.00001134, -0.00000702, 0.00000247, -0.00000737, 0.00000472, -0.00001174, 0.00001028, -0.00000217, 0.00000259, -0.00000343, 0.00000109, -0.00000303, 0.00000003, -0.00000895, 0.00000456, -0.00000929, 0.00000874, 0.00000188, 0.00000067, -0.00001019, 0.00000694, -0.00000231, 0.00000419, -0.00001123, 0.00000704, 0.00000026, -0.00000455, 0.00000102, -0.00000155, -0.00000145, },
		{ -0.00000872, 0.00001502, -0.00000882, -0.00000145, -0.00000045, 0.00000784, -0.00000560, -0.00000379, 0.00000038, 0.00000991, -0.00000403, -0.00000223, 0.00000001, -0.00000414, -0.00000057, 0.00000143, -0.00000246, 0.00000216, -0.00000359, 0.00000556, -0.00000227, -0.00000020, -0.00000305, 0.00000458, -0.00000835, 0.00001620, -0.00001447, 0.00001328, -0.00000519, -0.00001504, 0.00001254, -0.00000905, 0.00000978, -0.00001390, 0.00000828, -0.00001925, 0.00000783, -0.00000404, 0.00000224, 0.00000030, 0.00000142, -0.00001921, 0.00000569, -0.00000897, 0.00000519, -0.00001444, 0.00000150, 0.00000344, -0.00000232, -0.00001712, 0.00000260, -0.00001861, 0.00000123, -0.00003157, -0.00000121, -0.00001591, -0.00000889, -0.00008305, 0.00000537, -0.00014912, -0.00001037, -0.00085910, 0.00036857, 0.00217416, -0.00033847, -0.00058123, -0.00002143, -0.00013969, 0.00000114, -0.00004173, -0.00000061, -0.00005627, 0.00000804, -0.00001354, -0.00000269, -0.00001662, 0.00000003, -0.00000147, -0.00000206, -0.00001789, 0.00000683, 0.00000091, -0.00000113, -0.00001324, 0.00000382, -0.00000420, 0.00000229, -0.00000734, 0.00000530, -0.00001313, 0.00000900, -0.00001106, 0.00001386, -0.00001123, 0.00001978, -0.00002694, 0.00001254, -0.00000251, -0.00000636, 0.00000738, -0.00000710, 0.00000049, -0.00000244, -0.00000448, 0.00000145, 0.00000621, -0.00000415, 0.00000248, -0.00000339, 0.00000046, -0.00000211, -0.00000663, 0.00000181, -0.00000537, 0.00000150, 0.00000738, -0.00000047, -0.00000730, 0.00000071, 0.00000292, 0.00000092, -0.00000547, -0.00000253, 0.00000921, -0.00000973, 0.00000442, -0.00000503, 0.00000340, },
		{ -0.00000929, 0.00001114, -0.00000167, -0.00000611, 0.00000046, 0.00000532, -0.00000042, -0.00000652, -0.00000045, 0.00000770, 0.00000163, -0.00000590, 0.00000201, -0.00000474, -0.00000134, 0.00000211, -0.00000304, 0.00000173, -0.00000267, 0.00000335, 0.00000063, -0.00000184, -0.00000230, 0.00000441, -0.00000788, 0.00001045, -0.00000328, -0.00000191, 0.00001238, -0.00002772, 0.00001805, -0.00001131, 0.00001082, -0.00001383, 0.00000861, -0.00001584, 0.00000119, -0.00000007, 0.00000053, 0.00000133, 0.00000163, -0.00001698, 0.00000129, -0.00000528, 0.00000150, -0.00000907, -0.00000426, 0.00000723, -0.00000386, -0.00001556, 0.00000146, -0.00001572, -0.00000226, -0.00002600, -0.00000762, -0.00001013, -0.00001043, -0.00007670, 0.00000035, -0.00012953, -0.00001552, -0.00064939, -0.00040072, 0.00279544, -0.00032572, -0.00066482, -0.00001115, -0.00015546, 0.00000546, -0.00005080, 0.00000746, -0.00006336, 0.00001136, -0.00001643, -0.00000148, -0.00001849, 0.00000148, -0.00000545, 0.00000398, -0.00002339, 0.00001020, -0.00000258, 0.00000342, -0.00001623, 0.00000447, -0.00000595, 0.00000456, -0.00001152, 0.00001157, -0.00001625, 0.00000818, -0.00001055, 0.00001195, -0.00000863, 0.00001398, -0.00001474, -0.00000331, 0.00001221, -0.00001695, 0.00001291, -0.00000821, 0.00000139, -0.00000374, -0.00000234, -0.00000209, 0.00000865, -0.00000492, 0.00000273, -0.00000271, -0.00000042, -0.00000098, -0.00000644, 0.00000041, -0.00000238, -0.00000372, 0.00000950, 0.00000055, -0.00000486, -0.00000447, 0.00000608, -0.00000096, -0.00000049, -0.00000963, 0.00001334, -0.00000963, 0.00000356, -0.00000502, 0.00000447, },
		{ -0.00000513, 0.00000331, 0.00000651, -0.00000955, 0.00000057, 0.00000301, 0.00000308, -0.00000623, -0.00000320, 0.00000660, 0.00000430, -0.00000476, -0.00000113, -0.00000040, -0.00000644, 0.00000652, -0.00000702, 0.00000471, -0.00000546, 0.00000547, -0.00000186, 0.00000252, -0.00000733, 0.00000934, -0.00001137, 0.00000888, 0.00000098, -0.00000587, 0.00001339, -0.00001995, 0.00000274, 0.00000591, -0.00000672, 0.00000445, -0.00000977, 0.00000637, -0.00002236, 0.00001967, -0.00001795, 0.00002063, -0.00001827, 0.00000673, -0.00002466, 0.00002037, -0.00002537, 0.00002029, -0.00003442, 0.00003666, -0.00003366, 0.00001741, -0.00003460, 0.00002570, -0.00004788, 0.00002570, -0.00006463, 0.00005254, -0.00007913, 0.00001026, -0.00010414, 0.00001656, -0.00020349, -0.00020192, -0.00126421, 0.00322043, -0.00003080, -0.00093467, 0.00012225, -0.00026820, 0.00008915, -0.00012746, 0.00007257, -0.00011942, 0.00005807, -0.00005868, 0.00003620, -0.00005452, 0.00003495, -0.00003904, 0.00003627, -0.00005193, 0.00003467, -0.00002583, 0.00002582, -0.00003536, 0.00002092, -0.00002321, 0.00002142, -0.00002877, 0.00002797, -0.00002705, 0.00001498, -0.00001777, 0.00001715, -0.00001237, 0.00001412, -0.00001024, -0.00000763, 0.00001171, -0.00001075, 0.00000223, 0.00000452, -0.00000910, 0.00000502, -0.00000961, 0.00000396, 0.00000130, 0.00000344, -0.00000524, 0.00000521, -0.00000742, 0.00000514, -0.00001044, 0.00000294, -0.00000357, -0.00000343, 0.00000547, 0.00000588, -0.00000554, -0.00000530, 0.00000412, 0.00000138, 0.00000034, -0.00001068, 0.00001002, -0.00000308, -0.00000193, -0.00000123, 0.00000151, },
		{ 0.00000075, -0.00000450, 0.00001176, -0.00000898, -0.00000203, 0.00000294, 0.00000285, -0.00000173, -0.00000862, 0.00000800, 0.00000229, 0.00000212, -0.00000936, 0.00000826, -0.00001509, 0.00001407, -0.00001420, 0.00001134, -0.00001261, 0.00001283, -0.00001061, 0.00001345, -0.00001860, 0.00002021, -0.00002081, 0.00001519, -0.00000658, 0.00000613, -0.00000509, 0.00000802, -0.00003016, 0.00003778, -0.00003740, 0.00003524, -0.00004068, 0.00004043, -0.00005576, 0.00004907, -0.00004784, 0.00005295, -0.00005254, 0.00004534, -0.00006501, 0.00006072, -0.00006797, 0.00006589, -0.00008124, 0.00008433, -0.00008437, 0.00007391, -0.00009665, 0.00009531, -0.00012367, 0.00010987, -0.00015694, 0.00015494, -0.00019467, 0.00015194, -0.00027313, 0.00023875, -0.00048822, 0.00031912, -0.00198184, 0.00323588, 0.00065093, -0.00141792, 0.00038289, -0.00047644, 0.00024811, -0.00026617, 0.00018870, -0.00021902, 0.00014350, -0.00013622, 0.00010646, -0.00012069, 0.00009600, -0.00009728, 0.00008979, -0.00009912, 0.00007665, -0.00006570, 0.00006335, -0.00006865, 0.00005171, -0.00005447, 0.00005109, -0.00005717, 0.00005316, -0.00004573, 0.00003105, -0.00003501, 0.00003233, -0.00002630, 0.00002538, -0.00001973, 0.00000562, -0.00000808, 0.00001361, -0.00002375, 0.00002900, -0.00002891, 0.00002230, -0.00002522, 0.00001861, -0.00001451, 0.00001921, -0.00001956, 0.00001854, -0.00001898, 0.00001521, -0.00001826, 0.00000974, -0.00000975, 0.00000304, -0.00000458, 0.00001513, -0.00000971, -0.00000101, -0.00000328, 0.00000808, -0.00000344, -0.00000549, 0.00000043, 0.00000749, -0.00000939, 0.00000400, -0.00000317, },
		{ 0.00000470, -0.00000853, 0.00001133, -0.00000335, -0.00000753, 0.00000538, -0.00000100, 0.00000593, -0.00001542, 0.00001111, -0.00000334, 0.00001248, -0.00001954, 0.00001785, -0.00002401, 0.00002183, -0.00002203, 0.00001930, -0.00002181, 0.00002293, -0.00002265, 0.00002744, -0.00003240, 0.00003359, -0.00003349, 0.00002721, -0.00002318, 0.00002922, -0.00003487, 0.00004445, -0.00006667, 0.00006983, -0.00006701, 0.00006448, -0.00006986, 0.00007170, -0.00008480, 0.00007483, -0.00007610, 0.00008456, -0.00008639, 0.00008286, -0.00010305, 0.00009872, -0.00010861, 0.00010925, -0.00012572, 0.00013062, -0.00013505, 0.00013095, -0.00015918, 0.00016471, -0.00019821, 0.00019180, -0.00024610, 0.00025396, -0.00030704, 0.00028827, -0.00043221, 0.00044118, -0.00073544, 0.00071916, -0.00233733, 0.00271613, 0.00171070, -0.00204349, 0.00070085, -0.00072141, 0.00043176, -0.00042272, 0.00031716, -0.00032850, 0.00023798, -0.00022226, 0.00018465, -0.00019382, 0.00016267, -0.00015952, 0.00014558, -0.00014809, 0.00012101, -0.00010829, 0.00010332, -0.00010469, 0.00008624, -0.00008942, 0.00008362, -0.00008751, 0.00007939, -0.00006652, 0.00005183, -0.00005808, 0.00005373, -0.00004730, 0.00004531, -0.00004063, 0.00003222, -0.00004023, 0.00004687, -0.00005472, 0.00005526, -0.00004934, 0.00004063, -0.00004243, 0.00003540, -0.00003221, 0.00003589, -0.00003417, 0.00003180, -0.00003032, 0.00002527, -0.00002675, 0.00001817, -0.00001845, 0.00001280, -0.00001704, 0.00002492, -0.00001495, 0.00000611, -0.00001324, 0.00001617, -0.00000916, 0.00000269, -0.00001097, 0.00001715, -0.00001452, 0.00000716, -0.00000626, },
		{ 0.00000465, -0.00000731, 0.00000537, 0.00000554, -0.00001373, 0.00000836, -0.00000606, 0.00001352, -0.00002045, 0.00001320, -0.00000921, 0.00002175, -0.00002664, 0.00002352, -0.00002873, 0.00002570, -0.00002653, 0.00002445, -0.00002853, 0.00003064, -0.00003210, 0.00003793, -0.00004197, 0.00004279, -0.00004276, 0.00003765, -0.00003954, 0.00005113, -0.00006047, 0.00007148, -0.00008856, 0.00008491, -0.00007952, 0.00007670, -0.00008190, 0.00008483, -0.00009478, 0.00008292, -0.00008822, 0.00009972, -0.00010273, 0.00010111, -0.00012007, 0.00011524, -0.00012728, 0.00012935, -0.00014594, 0.00015236, -0.00016046, 0.00016059, -0.00019133, 0.00020008, -0.00023460, 0.00023127, -0.00028796, 0.00030031, -0.00035960, 0.00035277, -0.00050222, 0.00052755, -0.00082301, 0.00085035, -0.00221944, 0.00168478, 0.00299827, -0.00264049, 0.00094242, -0.00090043, 0.00055669, -0.00052732, 0.00039863, -0.00039686, 0.00029662, -0.00027638, 0.00023376, -0.00023974, 0.00020334, -0.00019673, 0.00017748, -0.00017542, 0.00014642, -0.00013374, 0.00012725, -0.00012637, 0.00010822, -0.00011225, 0.00010400, -0.00010600, 0.00009462, -0.00007900, 0.00006740, -0.00007684, 0.00007120, -0.00006544, 0.00006388, -0.00006184, 0.00005888, -0.00006927, 0.00007271, -0.00007517, 0.00006972, -0.00005893, 0.00005002, -0.00005201, 0.00004536, -0.00004298, 0.00004516, -0.00004150, 0.00003828, -0.00003559, 0.00003022, -0.00003131, 0.00002379, -0.00002503, 0.00002062, -0.00002633, 0.00003033, -0.00001725, 0.00001191, -0.00002105, 0.00002117, -0.00001263, 0.00000915, -0.00001882, 0.00002104, -0.00001382, 0.00000579, -0.00000562, },
		{ 0.00000118, -0.00000235, -0.00000303, 0.00001371, -0.00001701, 0.00000883, -0.00000890, 0.00001721, -0.00002033, 0.00001125, -0.00001151, 0.00002526, -0.00002616, 0.00002140, -0.00002587, 0.00002248, -0.00002433, 0.00002295, -0.00002828, 0.00003076, -0.00003302, 0.00003853, -0.00004081, 0.00004126, -0.00004160, 0.00003820, -0.00004503, 0.00005874, -0.00006704, 0.00007417, -0.00008266, 0.00007220, -0.00006564, 0.00006325, -0.00006829, 0.00007155, -0.00007810, 0.00006592, -0.00007571, 0.00008855, -0.00009065, 0.00008860, -0.00010450, 0.00009851, -0.00011146, 0.00011288, -0.00012784, 0.00013424, -0.00014336, 0.00014355, -0.00017197, 0.00017859, -0.00020847, 0.00020217, -0.00025425, 0.00026270, -0.00031680, 0.00030492, -0.00043736, 0.00044665, -0.00069477, 0.00066874, -0.00166524, 0.00032395, 0.00424869, -0.00298357, 0.00095634, -0.00090743, 0.00054145, -0.00051457, 0.00037924, -0.00037864, 0.00027965, -0.00026292, 0.00022136, -0.00022883, 0.00019115, -0.00018479, 0.00016406, -0.00016194, 0.00013512, -0.00012517, 0.00011925, -0.00011868, 0.00010301, -0.00010865, 0.00009876, -0.00010039, 0.00008783, -0.00007287, 0.00006694, -0.00007965, 0.00007287, -0.00006868, 0.00006839, -0.00006930, 0.00006989, -0.00007871, 0.00007579, -0.00007221, 0.00006223, -0.00004970, 0.00004351, -0.00004722, 0.00004176, -0.00004024, 0.00004104, -0.00003642, 0.00003351, -0.00003083, 0.00002644, -0.00002835, 0.00002271, -0.00002511, 0.00002168, -0.00002764, 0.00002752, -0.00001348, 0.00001280, -0.00002274, 0.00001956, -0.00001070, 0.00001039, -0.00001963, 0.00001696, -0.00000672, 0.00000016, -0.00000167, },
		{ -0.00000297, 0.00000284, -0.00000952, 0.00001694, -0.00001440, 0.00000459, -0.00000718, 0.00001472, -0.00001357, 0.00000402, -0.00000826, 0.00002077, -0.00001664, 0.00001086, -0.00001513, 0.00001184, -0.00001469, 0.00001349, -0.00001917, 0.00002092, -0.00002272, 0.00002653, -0.00002645, 0.00002650, -0.00002700, 0.00002462, -0.00003382, 0.00004533, -0.00004836, 0.00004842, -0.00004805, 0.00003344, -0.00002830, 0.00002724, -0.00003212, 0.00003515, -0.00003849, 0.00002724, -0.00004072, 0.00005221, -0.00005102, 0.00004645, -0.00005800, 0.00005033, -0.00006267, 0.00006115, -0.00007258, 0.00007692, -0.00008363, 0.00007921, -0.00010059, 0.00010031, -0.00012069, 0.00010620, -0.00014740, 0.00014415, -0.00018231, 0.00015031, -0.00024746, 0.00021612, -0.00038345, 0.00024712, -0.00084634, -0.00107694, 0.00515994, -0.00287243, 0.00063988, -0.00068458, 0.00034765, -0.00035721, 0.00023878, -0.00025796, 0.00017354, -0.00016986, 0.00013678, -0.00015177, 0.00011819, -0.00011725, 0.00010013, -0.00010311, 0.00008244, -0.00007769, 0.00007448, -0.00007677, 0.00006542, -0.00007337, 0.00006315, -0.00006650, 0.00005525, -0.00004390, 0.00004484, -0.00005983, 0.00005172, -0.00004967, 0.00005084, -0.00005401, 0.00005561, -0.00005991, 0.00005007, -0.00004286, 0.00003216, -0.00002208, 0.00002134, -0.00002774, 0.00002388, -0.00002328, 0.00002325, -0.00001906, 0.00001780, -0.00001632, 0.00001394, -0.00001740, 0.00001385, -0.00001709, 0.00001409, -0.00001952, 0.00001600, -0.00000341, 0.00000786, -0.00001722, 0.00001093, -0.00000325, 0.00000602, -0.00001355, 0.00000657, 0.00000390, -0.00000674, 0.00000293, },
		{ -0.00000478, 0.00000497, -0.00001086, 0.00001320, -0.00000556, -0.00000396, -0.00000120, 0.00000672, -0.00000174, -0.00000683, -0.00000056, 0.00000972, -0.00000072, -0.00000476, 0.00000008, -0.00000312, -0.00000029, -0.00000164, -0.00000330, 0.00000325, -0.00000356, 0.00000483, -0.00000234, 0.00000214, -0.00000230, -0.00000030, -0.00000843, 0.00001439, -0.00001028, 0.00000338, 0.00000285, -0.00001723, 0.00001830, -0.00001766, 0.00001325, -0.00001092, 0.00001045, -0.00002015, 0.00000479, 0.00000247, 0.00000370, -0.00001184, 0.00000487, -0.00001419, 0.00000374, -0.00000992, 0.00000315, -0.00000245, 0.00000096, -0.00001323, 0.00000142, -0.00001079, 0.00000189, -0.00002676, -0.00000062, -0.00001824, 0.00000168, -0.00006116, 0.00000619, -0.00008605, 0.00000674, -0.00026040, -0.00000388, -0.00221485, 0.00549379, -0.00221485, -0.00000388, -0.00026040, 0.00000674, -0.00008605, 0.00000619, -0.00006116, 0.00000168, -0.00001824, -0.00000062, -0.00002676, 0.00000189, -0.00001079, 0.00000142, -0.00001323, 0.00000096, -0.00000245, 0.00000315, -0.00000992, 0.00000374, -0.00001419, 0.00000487, -0.00001184, 0.00000370, 0.00000247, 0.00000479, -0.00002015, 0.00001045, -0.00001092, 0.00001325, -0.00001766, 0.00001830, -0.00001723, 0.00000285, 0.00000338, -0.00001028, 0.00001439, -0.00000843, -0.00000030, -0.00000230, 0.00000214, -0.00000234, 0.00000483, -0.00000356, 0.00000325, -0.00000330, -0.00000164, -0.00000029, -0.00000312, 0.00000008, -0.00000476, -0.00000072, 0.00000972, -0.00000056, -0.00000683, -0.00000174, 0.00000672, -0.00000120, -0.00000396, -0.00000556, 0.00001320, -0.00001086, 0.00000497, },
		{ -0.00000297, 0.00000293, -0.00000674, 0.00000390, 0.00000657, -0.00001355, 0.00000602, -0.00000325, 0.00001093, -0.00001722, 0.00000786, -0.00000341, 0.00001600, -0.00001952, 0.00001409, -0.00001709, 0.00001385, -0.00001740, 0.00001394, -0.00001632, 0.00001780, -0.00001906, 0.00002325, -0.00002328, 0.00002388, -0.00002774, 0.00002134, -0.00002208, 0.00003216, -0.00004286, 0.00005007, -0.00005991, 0.00005561, -0.00005401, 0.00005084, -0.00004967, 0.00005172, -0.00005983, 0.00004484, -0.00004390, 0.00005525, -0.00006650, 0.00006315, -0.00007337, 0.00006542, -0.00007677, 0.00007448, -0.00007769, 0.00008244, -0.00010311, 0.00010013, -0.00011725, 0.00011819, -0.00015177, 0.00013678, -0.00016986, 0.00017354, -0.00025796, 0.00023878, -0.00035721, 0.00034765, -0.00068458, 0.00063988, -0.00287243, 0.00515994, -0.00107694, -0.00084634, 0.00024712, -0.00038345, 0.00021612, -0.00024746, 0.00015031, -0.00018231, 0.00014415, -0.00014740, 0.00010620, -0.00012069, 0.00010031, -0.00010059, 0.00007921, -0.00008363, 0.00007692, -0.00007258, 0.00006115, -0.00006267, 0.00005033, -0.00005800, 0.00004645, -0.00005102, 0.00005221, -0.00004072, 0.00002724, -0.00003849, 0.00003515, -0.00003212, 0.00002724, -0.00002830, 0.00003344, -0.00004805, 0.00004842, -0.00004836, 0.00004533, -0.00003382, 0.00002462, -0.00002700, 0.00002650, -0.00002645, 0.00002653, -0.00002272, 0.00002092, -0.00001917, 0.00001349, -0.00001469, 0.00001184, -0.00001513, 0.00001086, -0.00001664, 0.00002077, -0.00000826, 0.00000402, -0.00001357, 0.00001472, -0.00000718, 0.00000459, -0.00001440, 0.00001694, -0.00000952, 0.00000284, },
		{ 0.00000118, -0.00000167, 0.00000016, -0.00000672, 0.00001696, -0.00001963, 0.00001039, -0.00001070, 0.00001956, -0.00002274, 0.00001280, -0.00001348, 0.00002752, -0.00002764, 0.00002168, -0.00002511, 0.00002271, -0.00002835, 0.00002644, -0.00003083, 0.00003351, -0.00003642, 0.00004104, -0.00004024, 0.00004176, -0.00004722, 0.00004351, -0.00004970, 0.00006223, -0.00007221, 0.00007579, -0.00007871, 0.00006989, -0.00006930, 0.00006839, -0.00006868, 0.00007287, -0.00007965, 0.00006694, -0.00007287, 0.00008783, -0.00010039, 0.00009876, -0.00010865, 0.00010301, -0.00011868, 0.00011925, -0.00012517, 0.00013512, -0.00016194, 0.00016406, -0.00018479, 0.00019115, -0.00022883, 0.00022136, -0.00026292, 0.00027965, -0.00037864, 0.00037924, -0.00051457, 0.00054145, -0.00090743, 0.00095634, -0.00298357, 0.00424869, 0.00032395, -0.00166524, 0.00066874, -0.00069477, 0.00044665, -0.00043736, 0.00030492, -0.00031680, 0.00026270, -0.00025425, 0.00020217, -0.00020847, 0.00017859, -0.00017197, 0.00014355, -0.00014336, 0.00013424, -0.00012784, 0.00011288, -0.00011146, 0.00009851, -0.00010450, 0.00008860, -0.00009065, 0.00008855, -0.00007571, 0.00006592, -0.00007810, 0.00007155, -0.00006829, 0.00006325, -0.00006564, 0.00007220, -0.00008266, 0.00007417, -0.00006704, 0.00005874, -0.00004503, 0.00003820, -0.00004160, 0.00004126, -0.00004081, 0.00003853, -0.00003302, 0.00003076, -0.00002828, 0.00002295, -0.00002433, 0.00002248, -0.00002587, 0.00002140, -0.00002616, 0.00002526, -0.00001151, 0.00001125, -0.00002033, 0.00001721, -0.00000890, 0.00000883, -0.00001701, 0.00001371, -0.00000303, -0.00000235, },
		{ 0.00000465, -0.00000562, 0.00000579, -0.00001382, 0.00002104, -0.00001882, 0.00000915, -0.00001263, 0.00002117, -0.00002105, 0.00001191, -0.00001725, 0.00003033, -0.00002633, 0.00002062, -0.00002503, 0.00002379, -0.00003131, 0.00003022, -0.00003559, 0.00003828, -0.00004150, 0.00004516, -0.00004298, 0.00004536, -0.00005201, 0.00005002, -0.00005893, 0.00006972, -0.00007517, 0.00007271, -0.00006927, 0.00005888, -0.00006184, 0.00006388, -0.00006544, 0.00007120, -0.00007684, 0.00006740, -0.00007900, 0.00009462, -0.00010600, 0.00010400, -0.00011225, 0.00010822, -0.00012637, 0.00012725, -0.00013374, 0.00014642, -0.00017542, 0.00017748, -0.00019673, 0.00020334, -0.00023974, 0.00023376, -0.00027638, 0.00029662, -0.00039686, 0.00039863, -0.00052732, 0.00055669, -0.00090043, 0.00094242, -0.00264049, 0.00299827, 0.00168478, -0.00221944, 0.00085035, -0.00082301, 0.00052755, -0.00050222, 0.00035277, -0.00035960, 0.00030031, -0.00028796, 0.00023127, -0.00023460, 0.00020008, -0.00019133, 0.00016059, -0.00016046, 0.00015236, -0.00014594, 0.00012935, -0.00012728, 0.00011524, -0.00012007, 0.00010111, -0.00010273, 0.00009972, -0.00008822, 0.00008292, -0.00009478, 0.00008483, -0.00008190, 0.00007670, -0.00007952, 0.00008491, -0.00008856, 0.00007148, -0.00006047, 0.00005113, -0.00003954, 0.00003765, -0.00004276, 0.00004279, -0.00004197, 0.00003793, -0.00003210, 0.00003064, -0.00002853, 0.00002445, -0.00002653, 0.00002570, -0.00002873, 0.00002352, -0.00002664, 0.00002175, -0.00000921, 0.00001320, -0.00002045, 0.00001352, -0.00000606, 0.00000836, -0.00001373, 0.00000554, 0.00000537, -0.00000731, },
		{ 0.00000470, -0.00000626, 0.00000716, -0.00001452, 0.00001715, -0.00001097, 0.00000269, -0.00000916, 0.00001617, -0.00001324, 0.00000611, -0.00001495, 0.00002492, -0.00001704, 0.00001280, -0.00001845, 0.00001817, -0.00002675, 0.00002527, -0.00003032, 0.00003180, -0.00003417, 0.00003589, -0.00003221, 0.00003540, -0.00004243, 0.00004063, -0.00004934, 0.00005526, -0.00005472, 0.00004687, -0.00004023, 0.00003222, -0.00004063, 0.00004531, -0.00004730, 0.00005373, -0.00005808, 0.00005183, -0.00006652, 0.00007939, -0.00008751, 0.00008362, -0.00008942, 0.00008624, -0.00010469, 0.00010332, -0.00010829, 0.00012101, -0.00014809, 0.00014558, -0.00015952, 0.00016267, -0.00019382, 0.00018465, -0.00022226, 0.00023798, -0.00032850, 0.00031716, -0.00042272, 0.00043176, -0.00072141, 0.00070085, -0.00204349, 0.00171070, 0.00271613, -0.00233733, 0.00071916, -0.00073544, 0.00044118, -0.00043221, 0.00028827, -0.00030704, 0.00025396, -0.00024610, 0.00019180, -0.00019821, 0.00016471, -0.00015918, 0.00013095, -0.00013505, 0.00013062, -0.00012572, 0.00010925, -0.00010861, 0.00009872, -0.00010305, 0.00008286, -0.00008639, 0.00008456, -0.00007610, 0.00007483, -0.00008480, 0.00007170, -0.00006986, 0.00006448, -0.00006701, 0.00006983, -0.00006667, 0.00004445, -0.00003487, 0.00002922, -0.00002318, 0.00002721, -0.00003349, 0.00003359, -0.00003240, 0.00002744, -0.00002265, 0.00002293, -0.00002181, 0.00001930, -0.00002203, 0.00002183, -0.00002401, 0.00001785, -0.00001954, 0.00001248, -0.00000334, 0.00001111, -0.00001542, 0.00000593, -0.00000100, 0.00000538, -0.00000753, -0.00000335, 0.00001133, -0.00000853, },
		{ 0.00000075, -0.00000317, 0.00000400, -0.00000939, 0.00000749, 0.00000043, -0.00000549, -0.00000344, 0.00000808, -0.00000328, -0.00000101, -0.00000971, 0.00001513, -0.00000458, 0.00000304, -0.00000975, 0.00000974, -0.00001826, 0.00001521, -0.00001898, 0.00001854, -0.00001956, 0.00001921, -0.00001451, 0.00001861, -0.00002522, 0.00002230, -0.00002891, 0.00002900, -0.00002375, 0.00001361, -0.00000808, 0.00000562, -0.00001973, 0.00002538, -0.00002630, 0.00003233, -0.00003501, 0.00003105, -0.00004573, 0.00005316, -0.00005717, 0.00005109, -0.00005447, 0.00005171, -0.00006865, 0.00006335, -0.00006570, 0.00007665, -0.00009912, 0.00008979, -0.00009728, 0.00009600, -0.00012069, 0.00010646, -0.00013622, 0.00014350, -0.00021902, 0.00018870, -0.00026617, 0.00024811, -0.00047644, 0.00038289, -0.00141792, 0.00065093, 0.00323588, -0.00198184, 0.00031912, -0.00048822, 0.00023875, -0.00027313, 0.00015194, -0.00019467, 0.00015494, -0.00015694, 0.00010987, -0.00012367, 0.00009531, -0.00009665, 0.00007391, -0.00008437, 0.00008433, -0.00008124, 0.00006589, -0.00006797, 0.00006072, -0.00006501, 0.00004534, -0.00005254, 0.00005295, -0.00004784, 0.00004907, -0.00005576, 0.00004043, -0.00004068, 0.00003524, -0.00003740, 0.00003778, -0.00003016, 0.00000802, -0.00000509, 0.00000613, -0.00000658, 0.00001519, -0.00002081, 0.00002021, -0.00001860, 0.00001345, -0.00001061, 0.00001283, -0.00001261, 0.00001134, -0.00001420, 0.00001407, -0.00001509, 0.00000826, -0.00000936, 0.00000212, 0.00000229, 0.00000800, -0.00000862, -0.00000173, 0.00000285, 0.00000294, -0.00000203, -0.00000898, 0.00001176, -0.00000450, },
		{ -0.00000513, 0.00000151, -0.00000123, -0.00000193, -0.00000308, 0.00001002, -0.00001068, 0.00000034, 0.00000138, 0.00000412, -0.00000530, -0.00000554, 0.00000588, 0.00000547, -0.00000343, -0.00000357, 0.00000294, -0.00001044, 0.00000514, -0.00000742, 0.00000521, -0.00000524, 0.00000344, 0.00000130, 0.00000396, -0.00000961, 0.00000502, -0.00000910, 0.00000452, 0.00000223, -0.00001075, 0.00001171, -0.00000763, -0.00001024, 0.00001412, -0.00001237, 0.00001715, -0.00001777, 0.00001498, -0.00002705, 0.00002797, -0.00002877, 0.00002142, -0.00002321, 0.00002092, -0.00003536, 0.00002582, -0.00002583, 0.00003467, -0.00005193, 0.00003627, -0.00003904, 0.00003495, -0.00005452, 0.00003620, -0.00005868, 0.00005807, -0.00011942, 0.00007257, -0.00012746, 0.00008915, -0.00026820, 0.00012225, -0.00093467, -0.00003080, 0.00322043, -0.00126421, -0.00020192, -0.00020349, 0.00001656, -0.00010414, 0.00001026, -0.00007913, 0.00005254, -0.00006463, 0.00002570, -0.00004788, 0.00002570, -0.00003460, 0.00001741, -0.00003366, 0.00003666, -0.00003442, 0.00002029, -0.00002537, 0.00002037, -0.00002466, 0.00000673, -0.00001827, 0.00002063, -0.00001795, 0.00001967, -0.00002236, 0.00000637, -0.00000977, 0.00000445, -0.00000672, 0.00000591, 0.00000274, -0.00001995, 0.00001339, -0.00000587, 0.00000098, 0.00000888, -0.00001137, 0.00000934, -0.00000733, 0.00000252, -0.00000186, 0.00000547, -0.00000546, 0.00000471, -0.00000702, 0.00000652, -0.00000644, -0.00000040, -0.00000113, -0.00000476, 0.00000430, 0.00000660, -0.00000320, -0.00000623, 0.00000308, 0.00000301, 0.00000057, -0.00000955, 0.00000651, 0.00000331, },
		{ -0.00000929, 0.00000447, -0.00000502, 0.00000356, -0.00000963, 0.00001334, -0.00000963, -0.00000049, -0.00000096, 0.00000608, -0.00000447, -0.00000486, 0.00000055, 0.00000950, -0.00000372, -0.00000238, 0.00000041, -0.00000644, -0.00000098, -0.00000042, -0.00000271, 0.00000273, -0.00000492, 0.00000865, -0.00000209, -0.00000234, -0.00000374, 0.00000139, -0.00000821, 0.00001291, -0.00001695, 0.00001221, -0.00000331, -0.00001474, 0.00001398, -0.00000863, 0.00001195, -0.00001055, 0.00000818, -0.00001625, 0.00001157, -0.00001152, 0.00000456, -0.00000595, 0.00000447, -0.00001623, 0.00000342, -0.00000258, 0.00001020, -0.00002339, 0.00000398, -0.00000545, 0.00000148, -0.00001849, -0.00000148, -0.00001643, 0.00001136, -0.00006336, 0.00000746, -0.00005080, 0.00000546, -0.00015546, -0.00001115, -0.00066482, -0.00032572, 0.00279544, -0.00040072, -0.00064939, -0.00001552, -0.00012953, 0.00000035, -0.00007670, -0.00001043, -0.00001013, -0.00000762, -0.00002600, -0.00000226, -0.00001572, 0.00000146, -0.00001556, -0.00000386, 0.00000723, -0.00000426, -0.00000907, 0.00000150, -0.00000528, 0.00000129, -0.00001698, 0.00000163, 0.00000133, 0.00000053, -0.00000007, 0.00000119, -0.00001584, 0.00000861, -0.00001383, 0.00001082, -0.00001131, 0.00001805, -0.00002772, 0.00001238, -0.00000191, -0.00000328, 0.00001045, -0.00000788, 0.00000441, -0.00000230, -0.00000184, 0.00000063, 0.00000335, -0.00000267, 0.00000173, -0.00000304, 0.00000211, -0.00000134, -0.00000474, 0.00000201, -0.00000590, 0.00000163, 0.00000770, -0.00000045, -0.00000652, -0.00000042, 0.00000532, 0.00000046, -0.00000611, -0.00000167, 0.00001114, },
		{ -0.00000872, 0.00000340, -0.00000503, 0.00000442, -0.00000973, 0.00000921, -0.00000253, -0.00000547, 0.00000092, 0.00000292, 0.00000071, -0.00000730, -0.00000047, 0.00000738, 0.00000150, -0.00000537, 0.00000181, -0.00000663, -0.00000211, 0.00000046, -0.00000339, 0.00000248, -0.00000415, 0.00000621, 0.00000145, -0.00000448, -0.00000244, 0.00000049, -0.00000710, 0.00000738, -0.00000636, -0.00000251, 0.00001254, -0.00002694, 0.00001978, -0.00001123, 0.00001386, -0.00001106, 0.00000900, -0.00001313, 0.00000530, -0.00000734, 0.00000229, -0.00000420, 0.00000382, -0.00001324, -0.00000113, 0.00000091, 0.00000683, -0.00001789, -0.00000206, -0.00000147, 0.00000003, -0.00001662, -0.00000269, -0.00001354, 0.00000804, -0.00005627, -0.00000061, -0.00004173, 0.00000114, -0.00013969, -0.00002143, -0.00058123, -0.00033847, 0.00217416, 0.00036857, -0.00085910, -0.00001037, -0.00014912, 0.00000537, -0.00008305, -0.00000889, -0.00001591, -0.00000121, -0.00003157, 0.00000123, -0.00001861, 0.00000260, -0.00001712, -0.00000232, 0.00000344, 0.00000150, -0.00001444, 0.00000519, -0.00000897, 0.00000569, -0.00001921, 0.00000142, 0.00000030, 0.00000224, -0.00000404, 0.00000783, -0.00001925, 0.00000828, -0.00001390, 0.00000978, -0.00000905, 0.00001254, -0.00001504, -0.00000519, 0.00001328, -0.00001447, 0.00001620, -0.00000835, 0.00000458, -0.00000305, -0.00000020, -0.00000227, 0.00000556, -0.00000359, 0.00000216, -0.00000246, 0.00000143, -0.00000057, -0.00000414, 0.00000001, -0.00000223, -0.00000403, 0.00000991, 0.00000038, -0.00000379, -0.00000560, 0.00000784, -0.00000045, -0.00000145, -0.00000882, 0.00001502, },
		{ -0.00000293, -0.00000145, -0.00000155, 0.00000102, -0.00000455, 0.00000026, 0.00000704, -0.00001123, 0.00000419, -0.00000231, 0.00000694, -0.00001019, 0.00000067, 0.00000188, 0.00000874, -0.00000929, 0.00000456, -0.00000895, 0.00000003, -0.00000303, 0.00000109, -0.00000343, 0.00000259, -0.00000217, 0.00001028, -0.00001174, 0.00000472, -0.00000737, 0.00000247, -0.00000702, 0.00001134, -0.00002097, 0.00002812, -0.00003638, 0.00002311, -0.00001330, 0.00001678, -0.00001360, 0.00001232, -0.00001359, 0.00000569, -0.00001219, 0.00000958, -0.00001215, 0.00001277, -0.00002032, 0.00000613, -0.00000889, 0.00001764, -0.00002819, 0.00001012, -0.00001758, 0.00001930, -0.00003607, 0.00001843, -0.00003471, 0.00003145, -0.00007972, 0.00002702, -0.00007426, 0.00004380, -0.00017964, 0.00003624, -0.00059776, -0.00022320, 0.00157259, 0.00086450, -0.00076052, -0.00018735, -0.00005954, -0.00007067, -0.00002649, -0.00005843, 0.00002122, -0.00003328, -0.00000217, -0.00002682, 0.00000685, -0.00002148, 0.00000378, -0.00002113, 0.00001866, -0.00001161, -0.00000086, -0.00000935, 0.00000459, -0.00000695, -0.00000475, -0.00001379, 0.00001269, -0.00000874, 0.00000470, 0.00000006, -0.00000700, -0.00000664, -0.00000033, -0.00000481, 0.00000672, -0.00000602, 0.00000793, -0.00002762, 0.00002851, -0.00002345, 0.00001951, -0.00000835, 0.00000642, -0.00000629, 0.00000398, -0.00000705, 0.00000892, -0.00000562, 0.00000392, -0.00000365, 0.00000310, -0.00000266, -0.00000051, -0.00000460, 0.00000307, -0.00000937, 0.00001078, 0.00000111, -0.00000034, -0.00000971, 0.00000822, -0.00000009, 0.00000173, -0.00001194, 0.00001297, },
		{ 0.00000562, -0.00000728, 0.00000278, -0.00000375, 0.00000210, -0.00000861, 0.00001401, -0.00001361, 0.00000528, -0.00000590, 0.00001055, -0.00001052, 0.00000114, -0.00000333, 0.00001387, -0.00001057, 0.00000571, -0.00001062, 0.00000247, -0.00000740, 0.00000654, -0.00001009, 0.00000969, -0.00001024, 0.00001787, -0.00001762, 0.00001111, -0.00001496, 0.00001199, -0.00002002, 0.00002451, -0.00003139, 0.00003307, -0.00003517, 0.00001822, -0.00001012, 0.00001611, -0.00001347, 0.00001355, -0.00001334, 0.00000799, -0.00001985, 0.00001878, -0.00002128, 0.00002236, -0.00002833, 0.00001557, -0.00002134, 0.00003104, -0.00004170, 0.00002638, -0.00003751, 0.00004108, -0.00005705, 0.00004065, -0.00005718, 0.00005685, -0.00010615, 0.00005874, -0.00011103, 0.00008911, -0.00022190, 0.00009513, -0.00062238, -0.00011525, 0.00113734, 0.00102247, -0.00039954, -0.00046025, 0.00006553, -0.00016968, 0.00004507, -0.00011893, 0.00006725, -0.00007423, 0.00003549, -0.00006186, 0.00003802, -0.00005002, 0.00002838, -0.00004349, 0.00003784, -0.00002976, 0.00001838, -0.00002912, 0.00002286, -0.00002444, 0.00001419, -0.00003201, 0.00002738, -0.00002226, 0.00001688, -0.00001290, 0.00001072, -0.00002513, 0.00001566, -0.00002151, 0.00002384, -0.00002429, 0.00002696, -0.00004144, 0.00003296, -0.00002260, 0.00001555, -0.00000460, 0.00000686, -0.00000860, 0.00000701, -0.00001006, 0.00001028, -0.00000616, 0.00000481, -0.00000459, 0.00000501, -0.00000521, 0.00000330, -0.00000853, 0.00000657, -0.00001148, 0.00000853, 0.00000304, 0.00000200, -0.00001089, 0.00000542, 0.00000224, 0.00000242, -0.00001044, 0.00000602, },
		{ 0.00001280, -0.00001035, 0.00000467, -0.00000647, 0.00000615, -0.00001297, 0.00001468, -0.00001020, 0.00000237, -0.00000579, 0.00000966, -0.00000700, -0.00000057, -0.00000586, 0.00001444, -0.00000751, 0.00000388, -0.00001004, 0.00000303, -0.00000970, 0.00000927, -0.00001324, 0.00001246, -0.00001309, 0.00001938, -0.00001753, 0.00001194, -0.00001685, 0.00001505, -0.00002436, 0.00002603, -0.00002809, 0.00002421, -0.00002263, 0.00000573, -0.00000218, 0.00001159, -0.00000974, 0.00001138, -0.00001067, 0.00000923, -0.00002550, 0.00002387, -0.00002507, 0.00002593, -0.00003046, 0.00001966, -0.00002782, 0.00003749, -0.00004794, 0.00003481, -0.00004773, 0.00005067, -0.00006418, 0.00004799, -0.00006401, 0.00006581, -0.00011492, 0.00007077, -0.00012436, 0.00010521, -0.00022965, 0.00011138, -0.00059851, -0.00008674, 0.00091283, 0.00089664, 0.00008741, -0.00069739, 0.00013534, -0.00022732, 0.00008172, -0.00015019, 0.00008877, -0.00009514, 0.00005534, -0.00008018, 0.00005342, -0.00006368, 0.00003938, -0.00005388, 0.00004679, -0.00003944, 0.00002990, -0.00004091, 0.00003311, -0.00003440, 0.00002538, -0.00004158, 0.00003370, -0.00002856, 0.00002309, -0.00002117, 0.00002297, -0.00003589, 0.00002316, -0.00002956, 0.00003138, -0.00003124, 0.00003178, -0.00003887, 0.00002243, -0.00001070, 0.00000469, 0.00000240, 0.00000543, -0.00000854, 0.00000704, -0.00000960, 0.00000836, -0.00000425, 0.00000391, -0.00000420, 0.00000578, -0.00000640, 0.00000511, -0.00000957, 0.00000640, -0.00000944, 0.00000334, 0.00000590, 0.00000292, -0.00000912, 0.00000034, 0.00000542, 0.00000157, -0.00000613, -0.00000244, },
		{ 0.00001511, -0.00000826, 0.00000241, -0.00000538, 0.00000561, -0.00001123, 0.00000878, -0.00000195, -0.00000345, -0.00000275, 0.00000520, -0.00000089, -0.00000371, -0.00000570, 0.00001075, -0.00000115, 0.00000008, -0.00000761, 0.00000142, -0.00000902, 0.00000799, -0.00001144, 0.00000955, -0.00000960, 0.00001418, -0.00001118, 0.00000686, -0.00001231, 0.00001059, -0.00001918, 0.00001635, -0.00001379, 0.00000668, -0.00000532, -0.00000803, 0.00000553, 0.00000675, -0.00000498, 0.00000778, -0.00000679, 0.00000917, -0.00002746, 0.00002275, -0.00002167, 0.00002189, -0.00002513, 0.00001626, -0.00002552, 0.00003381, -0.00004333, 0.00003112, -0.00004344, 0.00004347, -0.00005344, 0.00003680, -0.00005158, 0.00005428, -0.00010120, 0.00005726, -0.00010750, 0.00008528, -0.00019632, 0.00007871, -0.00051927, -0.00013925, 0.00085394, 0.00061856, 0.00053366, -0.00078110, 0.00008919, -0.00020841, 0.00005888, -0.00013394, 0.00007132, -0.00008347, 0.00004595, -0.00007153, 0.00004408, -0.00005474, 0.00003016, -0.00004636, 0.00003983, -0.00003474, 0.00002735, -0.00003835, 0.00002926, -0.00003093, 0.00002302, -0.00003733, 0.00002732, -0.00002365, 0.00001915, -0.00001977, 0.00002399, -0.00003341, 0.00001751, -0.00002480, 0.00002552, -0.00002385, 0.00002115, -0.00002169, 0.00000160, 0.00000626, -0.00000749, 0.00000865, 0.00000422, -0.00000705, 0.00000471, -0.00000645, 0.00000421, -0.00000094, 0.00000195, -0.00000282, 0.00000531, -0.00000581, 0.00000445, -0.00000753, 0.00000304, -0.00000466, -0.00000275, 0.00000809, 0.00000342, -0.00000588, -0.00000479, 0.00000737, 0.00000110, -0.00000180, -0.00000863, },
		{ 0.00001149, -0.00000140, -0.00000300, -0.00000141, 0.00000152, -0.00000523, -0.00000054, 0.00000726, -0.00000872, 0.00000037, 0.00000014, 0.00000478, -0.00000598, -0.00000464, 0.00000522, 0.00000554, -0.00000311, -0.00000521, -0.00000098, -0.00000651, 0.00000397, -0.00000635, 0.00000311, -0.00000250, 0.00000563, -0.00000233, -0.00000046, -0.00000492, 0.00000243, -0.00000926, 0.00000207, 0.00000292, -0.00000972, 0.00000727, -0.00001528, 0.00000735, 0.00000576, -0.00000263, 0.00000574, -0.00000406, 0.00000915, -0.00002656, 0.00001687, -0.00001345, 0.00001324, -0.00001552, 0.00000835, -0.00001729, 0.00002318, -0.00003134, 0.00001895, -0.00002892, 0.00002507, -0.00003182, 0.00001507, -0.00002855, 0.00003140, -0.00007466, 0.00002871, -0.00007267, 0.00004452, -0.00014089, 0.00002089, -0.00041470, -0.00022825, 0.00087054, 0.00033339, 0.00080898, -0.00065876, -0.00007198, -0.00012545, -0.00001077, -0.00008217, 0.00002551, -0.00004815, 0.00001486, -0.00004289, 0.00001674, -0.00002979, 0.00000700, -0.00002662, 0.00002177, -0.00001943, 0.00001375, -0.00002423, 0.00001422, -0.00001688, 0.00000999, -0.00002258, 0.00001195, -0.00001100, 0.00000786, -0.00001062, 0.00001538, -0.00002016, 0.00000222, -0.00001124, 0.00001082, -0.00000772, 0.00000254, 0.00000044, -0.00001884, 0.00001865, -0.00001366, 0.00000947, 0.00000589, -0.00000602, 0.00000198, -0.00000286, 0.00000018, 0.00000174, 0.00000052, -0.00000159, 0.00000444, -0.00000421, 0.00000231, -0.00000392, -0.00000137, 0.00000012, -0.00000703, 0.00000781, 0.00000474, -0.00000297, -0.00000781, 0.00000646, 0.00000240, 0.00000053, -0.00001036, },
		{ 0.00000385, 0.00000710, -0.00000836, 0.00000250, -0.00000301, 0.00000110, -0.00000849, 0.00001279, -0.00000986, 0.00000065, -0.00000253, 0.00000719, -0.00000530, -0.00000468, 0.00000061, 0.00000950, -0.00000329, -0.00000465, -0.00000242, -0.00000418, -0.00000024, -0.00000121, -0.00000298, 0.00000375, -0.00000131, 0.00000396, -0.00000512, 0.00000034, -0.00000392, -0.00000120, -0.00000880, 0.00001317, -0.00001651, 0.00000849, -0.00001190, 0.00000111, 0.00000996, -0.00000398, 0.00000664, -0.00000372, 0.00001015, -0.00002434, 0.00000923, -0.00000463, 0.00000483, -0.00000672, 0.00000098, -0.00000857, 0.00001187, -0.00001896, 0.00000598, -0.00001310, 0.00000607, -0.00001128, -0.00000436, -0.00000836, 0.00001134, -0.00005057, 0.00000215, -0.00003960, 0.00000647, -0.00009119, -0.00002909, -0.00032460, -0.00030079, 0.00087851, 0.00014322, 0.00086363, -0.00036267, -0.00028519, -0.00003090, -0.00008687, -0.00002773, -0.00002149, -0.00001192, -0.00001820, -0.00001221, -0.00001198, -0.00000416, -0.00001610, -0.00000717, 0.00000366, -0.00000336, -0.00000180, -0.00000757, -0.00000299, -0.00000110, -0.00000499, -0.00000625, -0.00000360, 0.00000132, -0.00000358, -0.00000032, 0.00000393, -0.00000406, -0.00001392, 0.00000217, -0.00000348, 0.00000713, -0.00001303, 0.00001591, -0.00002844, 0.00001891, -0.00000948, 0.00000298, 0.00001112, -0.00000635, 0.00000040, -0.00000076, -0.00000178, 0.00000218, 0.00000078, -0.00000144, 0.00000406, -0.00000269, 0.00000017, -0.00000074, -0.00000444, 0.00000248, -0.00000772, 0.00000440, 0.00000723, -0.00000131, -0.00000784, 0.00000260, 0.00000528, 0.00000069, -0.00000794, },
		{ -0.00000412, 0.00001311, -0.00001014, 0.00000339, -0.00000493, 0.00000418, -0.00001140, 0.00001190, -0.00000542, -0.00000285, -0.00000171, 0.00000548, -0.00000133, -0.00000641, -0.00000161, 0.00000933, 0.00000017, -0.00000636, -0.00000215, -0.00000337, -0.00000257, 0.00000129, -0.00000563, 0.00000578, -0.00000322, 0.00000449, -0.00000424, 0.00000053, -0.00000499, 0.00000079, -0.00001155, 0.00001274, -0.00001122, -0.00000162, -0.00000006, -0.00001011, 0.00001657, -0.00000711, 0.00000925, -0.00000490, 0.00001174, -0.00002162, 0.00000234, 0.00000132, 0.00000038, -0.00000241, -0.00000209, -0.00000369, 0.00000508, -0.00001203, -0.00000125, -0.00000364, -0.00000471, -0.00000126, -0.00001197, -0.00000066, 0.00000412, -0.00003981, -0.00001007, -0.00002277, -0.00001184, -0.00006652, -0.00004939, -0.00027384, -0.00032600, 0.00083417, 0.00007972, 0.00073395, 0.00000840, -0.00045545, 0.00001235, -0.00012621, -0.00000369, -0.00004342, 0.00000362, -0.00003445, 0.00000340, -0.00002644, 0.00000798, -0.00002649, 0.00000081, -0.00000450, 0.00000427, -0.00001031, 0.00000249, -0.00001315, 0.00000750, -0.00001315, 0.00000289, -0.00001109, 0.00000597, -0.00000849, 0.00000462, -0.00000318, 0.00000664, -0.00002240, 0.00000733, -0.00000941, 0.00001261, -0.00001749, 0.00001771, -0.00002286, 0.00000617, 0.00000336, -0.00000804, 0.00001753, -0.00000696, -0.00000001, -0.00000066, -0.00000124, 0.00000032, 0.00000260, -0.00000232, 0.00000436, -0.00000182, -0.00000095, 0.00000072, -0.00000492, 0.00000160, -0.00000495, -0.00000106, 0.00000989, -0.00000049, -0.00000562, -0.00000259, 0.00000810, 0.00000010, -0.00000362, },
		{ -0.00000903, 0.00001374, -0.00000669, 0.00000028, -0.00000318, 0.00000283, -0.00000862, 0.00000525, 0.00000282, -0.00000833, 0.00000117, 0.00000124, 0.00000412, -0.00000866, -0.00000194, 0.00000590, 0.00000577, -0.00000898, -0.00000088, -0.00000403, -0.00000256, 0.00000047, -0.00000415, 0.00000312, 0.00000005, -0.00000033, 0.00000135, -0.00000362, -0.00000125, -0.00000297, -0.00000690, 0.00000379, 0.00000182, -0.00001668, 0.00001298, -0.00001968, 0.00002047, -0.00000832, 0.00001074, -0.00000534, 0.00001245, -0.00001833, -0.00000267, 0.00000310, 0.00000082, -0.00000322, -0.00000030, -0.00000369, 0.00000439, -0.00001232, -0.00000074, -0.00000301, -0.00000453, -0.00000398, -0.00000634, -0.00000630, 0.00001048, -0.00004340, -0.00000641, -0.00002438, -0.00000788, -0.00006885, -0.00003903, -0.00026192, -0.00030573, 0.00074096, 0.00011190, 0.00051228, 0.00033486, -0.00050092, -0.00003388, -0.00010861, -0.00002350, -0.00003073, -0.00000897, -0.00002727, -0.00000234, -0.00002101, 0.00000182, -0.00002020, -0.00000595, 0.00000023, 0.00000042, -0.00000831, 0.00000197, -0.00001220, 0.00000509, -0.00001079, 0.00000125, -0.00000751, 0.00000071, -0.00000488, 0.00000187, -0.00000277, 0.00000808, -0.00001976, 0.00000164, -0.00000484, 0.00000699, -0.00001002, 0.00000689, -0.00000606, -0.00001311, 0.00001773, -0.00001737, 0.00002082, -0.00000559, -0.00000038, -0.00000163, 0.00000067, -0.00000250, 0.00000471, -0.00000326, 0.00000479, -0.00000144, -0.00000097, 0.00000040, -0.00000314, -0.00000148, -0.00000060, -0.00000623, 0.00001097, 0.00000063, -0.00000280, -0.00000684, 0.00000885, 0.00000054, -0.00000003, },
		{ -0.00000945, 0.00000886, 0.00000077, -0.00000518, 0.00000071, -0.00000131, -0.00000259, -0.00000352, 0.00001078, -0.00001229, 0.00000322, -0.00000260, 0.00000819, -0.00000934, -0.00000202, 0.00000148, 0.00001077, -0.00001029, -0.00000013, -0.00000500, -0.00000128, -0.00000238, -0.00000032, -0.00000195, 0.00000564, -0.00000704, 0.00000791, -0.00000846, 0.00000373, -0.00000865, 0.00000037, -0.00000726, 0.00001455, -0.00002794, 0.00001930, -0.00002157, 0.00001764, -0.00000494, 0.00000905, -0.00000337, 0.00001119, -0.00001427, -0.00000569, 0.00000137, 0.00000455, -0.00000696, 0.00000405, -0.00000650, 0.00000777, -0.00001742, 0.00000477, -0.00000830, 0.00000298, -0.00001454, 0.00000638, -0.00001835, 0.00002308, -0.00005374, 0.00000521, -0.00003583, 0.00000826, -0.00008544, -0.00001386, -0.00026887, -0.00026535, 0.00063217, 0.00017823, 0.00029951, 0.00052641, -0.00039070, -0.00016065, -0.00004930, -0.00007357, 0.00000448, -0.00003932, -0.00000519, -0.00002223, -0.00000236, -0.00001620, -0.00000346, -0.00002160, 0.00001267, -0.00001054, 0.00000075, -0.00000632, -0.00000291, -0.00000538, -0.00000088, -0.00000811, 0.00000357, -0.00001056, 0.00000368, -0.00000570, 0.00000305, 0.00000221, -0.00000892, -0.00001096, 0.00000583, -0.00000481, 0.00000334, -0.00000876, 0.00001230, -0.00002880, 0.00002501, -0.00001899, 0.00001762, -0.00000074, -0.00000171, -0.00000238, 0.00000231, -0.00000452, 0.00000560, -0.00000318, 0.00000460, -0.00000103, -0.00000046, -0.00000081, -0.00000052, -0.00000475, 0.00000282, -0.00000884, 0.00000916, 0.00000289, -0.00000079, -0.00000850, 0.00000653, 0.00000277, 0.00000148, },
		{ -0.00000639, 0.00000098, 0.00000871, -0.00000959, 0.00000374, -0.00000508, 0.00000278, -0.00000990, 0.00001432, -0.00001171, 0.00000209, -0.00000368, 0.00000880, -0.00000708, -0.00000321, -0.00000178, 0.00001288, -0.00000868, -0.00000107, -0.00000510, -0.00000027, -0.00000512, 0.00000310, -0.00000608, 0.00000973, -0.00001147, 0.00001126, -0.00001008, 0.00000603, -0.00001182, 0.00000489, -0.00001397, 0.00002010, -0.00002935, 0.00001491, -0.00001414, 0.00000806, 0.00000244, 0.00000466, 0.00000070, 0.00000836, -0.00000997, -0.00000688, -0.00000227, 0.00000890, -0.00001048, 0.00000768, -0.00000889, 0.00001161, -0.00002295, 0.00001034, -0.00001398, 0.00001120, -0.00002489, 0.00001711, -0.00002731, 0.00003216, -0.00006072, 0.00001395, -0.00004497, 0.00002228, -0.00009900, 0.00000572, -0.00027041, -0.00023417, 0.00054409, 0.00022364, 0.00016334, 0.00055400, -0.00015766, -0.00031423, 0.00000922, -0.00012209, 0.00003669, -0.00006660, 0.00001455, -0.00004131, 0.00001579, -0.00003333, 0.00001195, -0.00003542, 0.00002333, -0.00002030, 0.00000945, -0.00001537, 0.00000763, -0.00001665, 0.00000941, -0.00001802, 0.00001473, -0.00002056, 0.00001065, -0.00001230, 0.00000896, -0.00000526, 0.00000326, -0.00002282, 0.00001489, -0.00001476, 0.00001388, -0.00001970, 0.00002249, -0.00003272, 0.00002013, -0.00001088, 0.00000807, 0.00000668, -0.00000367, -0.00000240, 0.00000268, -0.00000470, 0.00000455, -0.00000166, 0.00000354, -0.00000026, 0.00000000, -0.00000188, 0.00000142, -0.00000642, 0.00000364, -0.00000793, 0.00000457, 0.00000599, 0.00000012, -0.00000741, 0.00000180, 0.00000598, 0.00000132, },
		{ -0.00000228, -0.00000619, 0.00001327, -0.00000990, 0.00000344, -0.00000592, 0.00000454, -0.00001102, 0.00001165, -0.00000605, -0.00000234, -0.00000172, 0.00000590, -0.00000228, -0.00000549, -0.00000311, 0.00001144, -0.00000423, -0.00000350, -0.00000406, -0.00000039, -0.00000623, 0.00000408, -0.00000699, 0.00000990, -0.00001125, 0.00000939, -0.00000687, 0.00000398, -0.00001043, 0.00000410, -0.00001353, 0.00001639, -0.00002073, 0.00000199, -0.00000128, -0.00000392, 0.00001002, 0.00000045, 0.00000466, 0.00000581, -0.00000675, -0.00000621, -0.00000633, 0.00001161, -0.00001143, 0.00000836, -0.00000859, 0.00001314, -0.00002543, 0.00001199, -0.00001555, 0.00001474, -0.00002876, 0.00001937, -0.00002695, 0.00003172, -0.00005819, 0.00001306, -0.00004394, 0.00002468, -0.00009824, 0.00000704, -0.00025227, -0.00022847, 0.00049502, 0.00022343, 0.00011964, 0.00045177, 0.00011828, -0.00042126, 0.00002101, -0.00013869, 0.00004303, -0.00007294, 0.00001734, -0.00004677, 0.00002157, -0.00003868, 0.00001620, -0.00003875, 0.00002463, -0.00002210, 0.00001145, -0.00001879, 0.00001267, -0.00002183, 0.00001341, -0.00002189, 0.00001936, -0.00002318, 0.00001079, -0.00001325, 0.00001026, -0.00000890, 0.00001032, -0.00002739, 0.00001622, -0.00001691, 0.00001558, -0.00002015, 0.00002001, -0.00002318, 0.00000464, 0.00000332, -0.00000363, 0.00001331, -0.00000447, -0.00000228, 0.00000195, -0.00000329, 0.00000208, 0.00000064, 0.00000212, 0.00000065, 0.00000022, -0.00000224, 0.00000193, -0.00000586, 0.00000176, -0.00000434, -0.00000115, 0.00000851, 0.00000074, -0.00000476, -0.00000335, 0.00000828, 0.00000111, },
		{ 0.00000065, -0.00000982, 0.00001233, -0.00000523, -0.00000045, -0.00000349, 0.00000227, -0.00000711, 0.00000425, 0.00000227, -0.00000777, 0.00000137, 0.00000152, 0.00000294, -0.00000743, -0.00000327, 0.00000767, 0.00000130, -0.00000596, -0.00000269, -0.00000141, -0.00000556, 0.00000238, -0.00000458, 0.00000637, -0.00000701, 0.00000351, -0.00000042, -0.00000089, -0.00000577, -0.00000076, -0.00000773, 0.00000670, -0.00000734, -0.00001257, 0.00000986, -0.00001209, 0.00001310, -0.00000026, 0.00000600, 0.00000550, -0.00000581, -0.00000375, -0.00000972, 0.00001166, -0.00000928, 0.00000590, -0.00000548, 0.00001180, -0.00002386, 0.00000871, -0.00001190, 0.00001214, -0.00002468, 0.00001240, -0.00001742, 0.00002246, -0.00004698, 0.00000316, -0.00003292, 0.00001517, -0.00008277, -0.00000942, -0.00021589, -0.00024552, 0.00047938, 0.00018615, 0.00014028, 0.00029217, 0.00034669, -0.00042389, -0.00003684, -0.00011315, 0.00001728, -0.00005436, 0.00000021, -0.00003571, 0.00001200, -0.00002957, 0.00000714, -0.00002996, 0.00001539, -0.00001479, 0.00000535, -0.00001464, 0.00000979, -0.00001844, 0.00000884, -0.00001756, 0.00001543, -0.00001703, 0.00000337, -0.00000793, 0.00000590, -0.00000681, 0.00000982, -0.00002271, 0.00000878, -0.00001064, 0.00000826, -0.00001095, 0.00000762, -0.00000563, -0.00001402, 0.00001605, -0.00001129, 0.00001512, -0.00000216, -0.00000300, 0.00000102, -0.00000142, -0.00000046, 0.00000247, 0.00000127, 0.00000112, 0.00000043, -0.00000198, 0.00000120, -0.00000372, -0.00000148, -0.00000018, -0.00000563, 0.00000880, 0.00000217, -0.00000219, -0.00000682, 0.00000797, 0.00000229, },
		{ 0.00000165, -0.00000925, 0.00000655, 0.00000247, -0.00000575, 0.00000027, -0.00000194, -0.00000105, -0.00000398, 0.00000915, -0.00001081, 0.00000284, -0.00000160, 0.00000603, -0.00000730, -0.00000367, 0.00000361, 0.00000554, -0.00000665, -0.00000218, -0.00000238, -0.00000411, -0.00000068, -0.00000070, 0.00000149, -0.00000163, -0.00000295, 0.00000563, -0.00000512, -0.00000124, -0.00000597, -0.00000118, -0.00000299, 0.00000356, -0.00002134, 0.00001307, -0.00001218, 0.00000921, 0.00000386, 0.00000386, 0.00000800, -0.00000720, -0.00000020, -0.00001169, 0.00000916, -0.00000503, 0.00000183, -0.00000117, 0.00000900, -0.00001971, 0.00000238, -0.00000526, 0.00000585, -0.00001582, 0.00000060, -0.00000422, 0.00001041, -0.00003332, -0.00000946, -0.00001837, 0.00000076, -0.00006094, -0.00003304, -0.00017421, -0.00026945, 0.00047627, 0.00013931, 0.00017840, 0.00014926, 0.00046446, -0.00030701, -0.00014967, -0.00006242, -0.00002653, -0.00002284, -0.00002684, -0.00001624, -0.00000608, -0.00001236, -0.00000911, -0.00001496, 0.00000108, -0.00000320, -0.00000486, -0.00000614, 0.00000171, -0.00000926, -0.00000124, -0.00000815, 0.00000620, -0.00000585, -0.00000758, 0.00000005, -0.00000119, -0.00000125, 0.00000401, -0.00001209, -0.00000320, -0.00000057, -0.00000292, 0.00000167, -0.00000686, 0.00001064, -0.00002660, 0.00002000, -0.00001032, 0.00001009, 0.00000368, -0.00000475, 0.00000055, -0.00000022, -0.00000182, 0.00000280, 0.00000170, 0.00000075, 0.00000091, -0.00000149, 0.00000002, -0.00000134, -0.00000423, 0.00000240, -0.00000710, 0.00000607, 0.00000480, -0.00000067, -0.00000749, 0.00000468, 0.00000503, },
		{ 0.00000156, -0.00000594, -0.00000114, 0.00000955, -0.00000912, 0.00000248, -0.00000509, 0.00000352, -0.00000899, 0.00001111, -0.00000913, 0.00000098, -0.00000178, 0.00000563, -0.00000438, -0.00000512, 0.00000084, 0.00000683, -0.00000460, -0.00000313, -0.00000254, -0.00000306, -0.00000329, 0.00000226, -0.00000189, 0.00000168, -0.00000655, 0.00000796, -0.00000573, 0.00000019, -0.00000815, 0.00000200, -0.00000772, 0.00000692, -0.00002045, 0.00000664, -0.00000450, -0.00000010, 0.00001088, -0.00000011, 0.00001188, -0.00000944, 0.00000299, -0.00001157, 0.00000487, -0.00000042, -0.00000178, 0.00000223, 0.00000679, -0.00001541, -0.00000386, 0.00000077, -0.00000004, -0.00000723, -0.00000976, 0.00000569, 0.00000265, -0.00002421, -0.00001769, -0.00000789, -0.00000993, -0.00004310, -0.00005118, -0.00014189, -0.00028317, 0.00046441, 0.00010927, 0.00019570, 0.00006853, 0.00045767, -0.00010462, -0.00027034, -0.00002162, -0.00006298, 0.00000153, -0.00004766, -0.00000159, -0.00002115, 0.00000225, -0.00002256, -0.00000296, -0.00000998, 0.00000542, -0.00001280, 0.00000098, -0.00000600, -0.00000009, -0.00001080, 0.00000037, -0.00000231, 0.00000405, -0.00001596, 0.00000532, -0.00000634, 0.00000343, -0.00000226, -0.00000152, -0.00001310, 0.00000670, -0.00001110, 0.00001023, -0.00001535, 0.00001772, -0.00002711, 0.00001234, -0.00000069, -0.00000007, 0.00001098, -0.00000636, 0.00000040, -0.00000005, -0.00000161, 0.00000150, 0.00000327, -0.00000028, 0.00000163, -0.00000105, -0.00000089, 0.00000019, -0.00000522, 0.00000234, -0.00000529, 0.00000106, 0.00000784, -0.00000000, -0.00000579, -0.00000036, 0.00000807, },
		{ 0.00000197, -0.00000227, -0.00000729, 0.00001264, -0.00000813, 0.00000129, -0.00000521, 0.00000434, -0.00000879, 0.00000730, -0.00000304, -0.00000361, 0.00000055, 0.00000240, 0.00000034, -0.00000709, -0.00000041, 0.00000515, -0.00000043, -0.00000500, -0.00000193, -0.00000295, -0.00000435, 0.00000286, -0.00000218, 0.00000132, -0.00000587, 0.00000559, -0.00000214, -0.00000208, -0.00000639, 0.00000053, -0.00000623, 0.00000244, -0.00001149, -0.00000574, 0.00000611, -0.00000993, 0.00001658, -0.00000263, 0.00001453, -0.00001035, 0.00000419, -0.00000903, -0.00000019, 0.00000302, -0.00000344, 0.00000339, 0.00000651, -0.00001276, -0.00000761, 0.00000345, -0.00000244, -0.00000272, -0.00001429, 0.00000800, 0.00000296, -0.00002307, -0.00001810, -0.00000533, -0.00001219, -0.00003519, -0.00005669, -0.00012659, -0.00027845, 0.00043398, 0.00010691, 0.00017836, 0.00005553, 0.00035956, 0.00011665, -0.00034151, -0.00002405, -0.00007111, 0.00000330, -0.00005040, -0.00000138, -0.00002464, 0.00000617, -0.00002576, -0.00000064, -0.00001196, 0.00000602, -0.00001398, 0.00000237, -0.00000879, 0.00000408, -0.00001470, 0.00000300, -0.00000516, 0.00000775, -0.00001736, 0.00000431, -0.00000642, 0.00000398, -0.00000494, 0.00000405, -0.00001596, 0.00000671, -0.00001198, 0.00001047, -0.00001404, 0.00001328, -0.00001598, -0.00000358, 0.00001266, -0.00001033, 0.00001593, -0.00000582, -0.00000026, -0.00000041, -0.00000041, -0.00000060, 0.00000510, -0.00000124, 0.00000216, -0.00000062, -0.00000127, 0.00000048, -0.00000430, 0.00000010, -0.00000156, -0.00000425, 0.00000969, 0.00000084, -0.00000317, -0.00000503, 0.00000949, },
		{ 0.00000389, -0.00000000, -0.00000972, 0.00001042, -0.00000272, -0.00000287, -0.00000259, 0.00000167, -0.00000434, -0.00000013, 0.00000460, -0.00000830, 0.00000319, -0.00000140, 0.00000461, -0.00000810, -0.00000103, 0.00000186, 0.00000405, -0.00000633, -0.00000135, -0.00000345, -0.00000397, 0.00000121, 0.00000021, -0.00000194, -0.00000211, 0.00000030, 0.00000350, -0.00000600, -0.00000257, -0.00000365, -0.00000105, -0.00000599, -0.00000013, -0.00001728, 0.00001309, -0.00001489, 0.00001703, -0.00000103, 0.00001398, -0.00000842, 0.00000266, -0.00000445, -0.00000488, 0.00000437, -0.00000278, 0.00000234, 0.00000807, -0.00001206, -0.00000826, 0.00000230, -0.00000087, -0.00000299, -0.00001243, 0.00000302, 0.00001008, -0.00002810, -0.00001257, -0.00000904, -0.00000732, -0.00003618, -0.00005078, -0.00012589, -0.00025897, 0.00038949, 0.00012528, 0.00013677, 0.00008456, 0.00022866, 0.00028762, -0.00032320, -0.00008148, -0.00004777, -0.00001871, -0.00003492, -0.00001555, -0.00001616, -0.00000135, -0.00001807, -0.00000833, -0.00000488, -0.00000115, -0.00000851, -0.00000223, -0.00000582, 0.00000196, -0.00001162, -0.00000130, -0.00000140, 0.00000444, -0.00001161, -0.00000254, -0.00000185, 0.00000030, -0.00000315, 0.00000331, -0.00001111, -0.00000035, -0.00000614, 0.00000340, -0.00000501, 0.00000129, 0.00000038, -0.00001990, 0.00002231, -0.00001500, 0.00001514, -0.00000171, -0.00000204, -0.00000062, 0.00000075, -0.00000230, 0.00000608, -0.00000131, 0.00000199, 0.00000002, -0.00000128, -0.00000010, -0.00000233, -0.00000282, 0.00000197, -0.00000764, 0.00000893, 0.00000276, -0.00000107, -0.00000751, 0.00000804, },
		{ 0.00000696, 0.00000066, -0.00000843, 0.00000423, 0.00000469, -0.00000750, 0.00000057, -0.00000215, 0.00000126, -0.00000727, 0.00000989, -0.00001000, 0.00000366, -0.00000336, 0.00000631, -0.00000681, -0.00000213, -0.00000130, 0.00000683, -0.00000573, -0.00000168, -0.00000378, -0.00000305, -0.00000134, 0.00000344, -0.00000577, 0.00000192, -0.00000464, 0.00000781, -0.00000842, 0.00000025, -0.00000717, 0.00000365, -0.00001299, 0.00000736, -0.00002197, 0.00001201, -0.00001234, 0.00001108, 0.00000488, 0.00001037, -0.00000398, -0.00000083, 0.00000082, -0.00000799, 0.00000342, -0.00000045, 0.00000016, 0.00001036, -0.00001235, -0.00000690, -0.00000115, 0.00000279, -0.00000590, -0.00000707, -0.00000520, 0.00001902, -0.00003402, -0.00000638, -0.00001378, -0.00000068, -0.00004023, -0.00004054, -0.00013056, -0.00023596, 0.00034401, 0.00014760, 0.00009312, 0.00011941, 0.00012003, 0.00036506, -0.00021183, -0.00017546, -0.00001059, -0.00005070, -0.00001265, -0.00003474, -0.00000323, -0.00001394, -0.00000547, -0.00002029, 0.00000576, -0.00001103, -0.00000077, -0.00000919, -0.00000007, -0.00000378, -0.00000441, -0.00000940, 0.00000575, -0.00000250, -0.00000260, -0.00001118, 0.00000380, -0.00000471, 0.00000082, -0.00000131, -0.00000202, -0.00001015, 0.00000181, -0.00000581, 0.00000556, -0.00001069, 0.00001339, -0.00002856, 0.00002243, -0.00001098, 0.00000788, 0.00000537, -0.00000446, -0.00000048, 0.00000119, -0.00000277, 0.00000559, -0.00000016, 0.00000102, 0.00000096, -0.00000117, -0.00000089, -0.00000048, -0.00000484, 0.00000354, -0.00000788, 0.00000531, 0.00000572, -0.00000004, -0.00000723, 0.00000394, },
		{ 0.00000960, 0.00000083, -0.00000516, -0.00000285, 0.00001052, -0.00000955, 0.00000171, -0.00000442, 0.00000477, -0.00001067, 0.00001011, -0.00000712, 0.00000093, -0.00000248, 0.00000477, -0.00000312, -0.00000400, -0.00000326, 0.00000688, -0.00000280, -0.00000311, -0.00000350, -0.00000253, -0.00000328, 0.00000546, -0.00000778, 0.00000360, -0.00000649, 0.00000828, -0.00000721, -0.00000004, -0.00000759, 0.00000484, -0.00001501, 0.00000763, -0.00001790, 0.00000303, -0.00000407, 0.00000132, 0.00001236, 0.00000606, 0.00000087, -0.00000425, 0.00000490, -0.00000854, 0.00000051, 0.00000237, -0.00000175, 0.00001198, -0.00001229, -0.00000525, -0.00000451, 0.00000578, -0.00000830, -0.00000219, -0.00001168, 0.00002440, -0.00003559, -0.00000450, -0.00001459, 0.00000242, -0.00004116, -0.00003364, -0.00013106, -0.00022043, 0.00030976, 0.00015888, 0.00006600, 0.00013356, 0.00006356, 0.00034584, -0.00004163, -0.00026471, 0.00001157, -0.00007249, 0.00000071, -0.00004653, 0.00000403, -0.00002271, 0.00000372, -0.00002870, 0.00001277, -0.00001717, 0.00000365, -0.00001360, 0.00000395, -0.00000867, 0.00000214, -0.00001625, 0.00001126, -0.00000795, 0.00000434, -0.00001652, 0.00000621, -0.00000737, 0.00000347, -0.00000572, 0.00000613, -0.00001706, 0.00000631, -0.00001141, 0.00001139, -0.00001604, 0.00001693, -0.00002560, 0.00001214, 0.00000016, -0.00000295, 0.00001260, -0.00000588, -0.00000053, 0.00000098, -0.00000205, 0.00000387, 0.00000179, -0.00000034, 0.00000193, -0.00000104, -0.00000140, 0.00000047, -0.00000510, 0.00000265, -0.00000529, 0.00000004, 0.00000858, 0.00000047, -0.00000506, -0.00000121, },
		{ 0.00001010, 0.00000190, -0.00000207, -0.00000779, 0.00001204, -0.00000728, -0.00000036, -0.00000382, 0.00000464, -0.00000918, 0.00000524, -0.00000065, -0.00000383, 0.00000025, 0.00000117, 0.00000156, -0.00000580, -0.00000414, 0.00000456, 0.00000151, -0.00000486, -0.00000281, -0.00000274, -0.00000384, 0.00000529, -0.00000698, 0.00000204, -0.00000462, 0.00000478, -0.00000266, -0.00000318, -0.00000493, 0.00000231, -0.00001199, 0.00000167, -0.00000789, -0.00000928, 0.00000481, -0.00000739, 0.00001729, 0.00000421, 0.00000361, -0.00000544, 0.00000627, -0.00000623, -0.00000347, 0.00000451, -0.00000237, 0.00001211, -0.00001106, -0.00000457, -0.00000602, 0.00000615, -0.00000802, -0.00000052, -0.00001330, 0.00002343, -0.00003072, -0.00000858, -0.00000985, 0.00000001, -0.00003626, -0.00003374, -0.00012283, -0.00021702, 0.00029111, 0.00015405, 0.00006092, 0.00012041, 0.00005802, 0.00026186, 0.00013152, -0.00030461, -0.00000475, -0.00007027, -0.00000476, -0.00004348, -0.00000036, -0.00002216, 0.00000412, -0.00002859, 0.00001172, -0.00001580, 0.00000155, -0.00001258, 0.00000334, -0.00000946, 0.00000433, -0.00001802, 0.00001148, -0.00000833, 0.00000564, -0.00001556, 0.00000315, -0.00000577, 0.00000265, -0.00000699, 0.00000962, -0.00001749, 0.00000433, -0.00001051, 0.00000971, -0.00001260, 0.00001045, -0.00001315, -0.00000394, 0.00001273, -0.00001204, 0.00001620, -0.00000438, -0.00000155, 0.00000064, -0.00000085, 0.00000187, 0.00000358, -0.00000128, 0.00000244, -0.00000072, -0.00000154, 0.00000041, -0.00000384, 0.00000013, -0.00000152, -0.00000476, 0.00000975, 0.00000151, -0.00000252, -0.00000529, },
		{ 0.00000768, 0.00000439, -0.00000038, -0.00000909, 0.00000873, -0.00000135, -0.00000462, -0.00000114, 0.00000170, -0.00000434, -0.00000209, 0.00000627, -0.00000784, 0.00000253, -0.00000217, 0.00000507, -0.00000614, -0.00000484, 0.00000126, 0.00000548, -0.00000563, -0.00000243, -0.00000329, -0.00000326, 0.00000335, -0.00000414, -0.00000162, -0.00000061, -0.00000057, 0.00000282, -0.00000686, -0.00000131, -0.00000169, -0.00000682, -0.00000637, 0.00000243, -0.00001866, 0.00000876, -0.00001079, 0.00001666, 0.00000671, 0.00000294, -0.00000355, 0.00000474, -0.00000186, -0.00000722, 0.00000505, -0.00000136, 0.00001079, -0.00000875, -0.00000510, -0.00000529, 0.00000370, -0.00000494, -0.00000230, -0.00001012, 0.00001710, -0.00002131, -0.00001629, -0.00000188, -0.00000587, -0.00002728, -0.00003930, -0.00010771, -0.00022268, 0.00028348, 0.00013866, 0.00007024, 0.00009091, 0.00008032, 0.00016081, 0.00025480, -0.00026878, -0.00006326, -0.00004570, -0.00002699, -0.00002784, -0.00001451, -0.00001344, -0.00000363, -0.00002061, 0.00000345, -0.00000796, -0.00000594, -0.00000699, -0.00000150, -0.00000602, 0.00000166, -0.00001424, 0.00000622, -0.00000356, 0.00000141, -0.00000904, -0.00000412, -0.00000103, -0.00000106, -0.00000494, 0.00000801, -0.00001170, -0.00000307, -0.00000450, 0.00000239, -0.00000333, -0.00000131, 0.00000202, -0.00001810, 0.00001993, -0.00001454, 0.00001361, 0.00000073, -0.00000368, 0.00000060, 0.00000002, 0.00000055, 0.00000431, -0.00000118, 0.00000213, -0.00000006, -0.00000150, -0.00000020, -0.00000198, -0.00000250, 0.00000143, -0.00000720, 0.00000816, 0.00000371, -0.00000077, -0.00000689, },
		{ 0.00000309, 0.00000747, 0.00000015, -0.00000729, 0.00000244, 0.00000550, -0.00000847, 0.00000137, -0.00000168, 0.00000075, -0.00000818, 0.00001015, -0.00000848, 0.00000229, -0.00000330, 0.00000577, -0.00000414, -0.00000610, -0.00000161, 0.00000752, -0.00000447, -0.00000291, -0.00000362, -0.00000239, 0.00000091, -0.00000106, -0.00000514, 0.00000291, -0.00000478, 0.00000623, -0.00000839, 0.00000062, -0.00000425, -0.00000305, -0.00001198, 0.00000798, -0.00002073, 0.00000513, -0.00000775, 0.00001049, 0.00001282, -0.00000020, 0.00000038, 0.00000171, 0.00000278, -0.00000929, 0.00000353, 0.00000092, 0.00000873, -0.00000608, -0.00000623, -0.00000308, -0.00000030, -0.00000067, -0.00000567, -0.00000470, 0.00000903, -0.00001174, -0.00002317, 0.00000506, -0.00001107, -0.00001839, -0.00004568, -0.00009143, -0.00022987, 0.00027751, 0.00012350, 0.00008041, 0.00006276, 0.00010243, 0.00008355, 0.00029970, -0.00016163, -0.00014371, -0.00001592, -0.00005312, -0.00001003, -0.00003005, -0.00000321, -0.00001382, -0.00001015, -0.00000687, 0.00000142, -0.00001434, -0.00000069, -0.00000735, -0.00000106, -0.00000334, -0.00000769, -0.00000142, 0.00000316, -0.00000500, -0.00000075, -0.00001170, 0.00000348, -0.00000498, -0.00000178, 0.00000374, -0.00000317, -0.00001170, 0.00000221, -0.00000568, 0.00000593, -0.00001137, 0.00001229, -0.00002368, 0.00001750, -0.00000881, 0.00000532, 0.00000802, -0.00000590, 0.00000067, 0.00000025, 0.00000037, 0.00000371, 0.00000003, 0.00000111, 0.00000087, -0.00000142, -0.00000083, -0.00000052, -0.00000395, 0.00000225, -0.00000656, 0.00000406, 0.00000666, 0.00000002, -0.00000595, },
		{ -0.00000184, 0.00000949, 0.00000072, -0.00000422, -0.00000385, 0.00001004, -0.00000932, 0.00000161, -0.00000322, 0.00000337, -0.00001027, 0.00000908, -0.00000489, -0.00000085, -0.00000185, 0.00000362, -0.00000023, -0.00000779, -0.00000346, 0.00000710, -0.00000142, -0.00000413, -0.00000348, -0.00000195, -0.00000076, 0.00000061, -0.00000664, 0.00000394, -0.00000584, 0.00000587, -0.00000648, -0.00000037, -0.00000384, -0.00000267, -0.00001287, 0.00000692, -0.00001519, -0.00000443, -0.00000098, 0.00000188, 0.00001940, -0.00000312, 0.00000399, -0.00000055, 0.00000570, -0.00000862, 0.00000016, 0.00000364, 0.00000690, -0.00000395, -0.00000712, -0.00000065, -0.00000401, 0.00000268, -0.00000822, -0.00000018, 0.00000316, -0.00000608, -0.00002551, 0.00000763, -0.00001238, -0.00001310, -0.00004864, -0.00007941, -0.00023175, 0.00026522, 0.00011728, 0.00008099, 0.00004885, 0.00010629, 0.00004896, 0.00027024, -0.00001622, -0.00021058, -0.00000404, -0.00006763, -0.00000189, -0.00003774, 0.00000105, -0.00001981, -0.00000357, -0.00001325, 0.00000684, -0.00001874, 0.00000215, -0.00001053, 0.00000196, -0.00000710, -0.00000232, -0.00000724, 0.00000768, -0.00000935, 0.00000485, -0.00001548, 0.00000443, -0.00000632, -0.00000022, 0.00000016, 0.00000378, -0.00001691, 0.00000515, -0.00000983, 0.00001009, -0.00001454, 0.00001328, -0.00001858, 0.00000621, 0.00000233, -0.00000498, 0.00001424, -0.00000637, 0.00000016, 0.00000011, 0.00000102, 0.00000227, 0.00000171, -0.00000005, 0.00000166, -0.00000124, -0.00000119, 0.00000005, -0.00000377, 0.00000091, -0.00000364, -0.00000102, 0.00000904, 0.00000062, -0.00000361, },
		{ -0.00000516, 0.00000900, 0.00000241, -0.00000168, -0.00000764, 0.00001025, -0.00000616, -0.00000095, -0.00000221, 0.00000266, -0.00000798, 0.00000376, 0.00000138, -0.00000528, 0.00000077, 0.00000018, 0.00000394, -0.00000890, -0.00000463, 0.00000488, 0.00000239, -0.00000523, -0.00000318, -0.00000209, -0.00000117, 0.00000028, -0.00000566, 0.00000219, -0.00000376, 0.00000227, -0.00000205, -0.00000341, -0.00000115, -0.00000511, -0.00000988, 0.00000115, -0.00000569, -0.00001502, 0.00000460, -0.00000477, 0.00002275, -0.00000303, 0.00000507, -0.00000021, 0.00000570, -0.00000513, -0.00000420, 0.00000579, 0.00000604, -0.00000290, -0.00000721, 0.00000107, -0.00000611, 0.00000371, -0.00000844, 0.00000153, 0.00000158, -0.00000594, -0.00002249, 0.00000548, -0.00000944, -0.00001208, -0.00004691, -0.00007365, -0.00022580, 0.00024419, 0.00012191, 0.00006986, 0.00005120, 0.00008986, 0.00005185, 0.00019609, 0.00012099, -0.00022996, -0.00002560, -0.00006228, -0.00000917, -0.00003346, -0.00000397, -0.00001834, -0.00000419, -0.00001258, 0.00000557, -0.00001693, -0.00000031, -0.00000933, 0.00000119, -0.00000739, -0.00000079, -0.00000847, 0.00000743, -0.00000915, 0.00000534, -0.00001361, 0.00000070, -0.00000417, -0.00000148, -0.00000073, 0.00000650, -0.00001630, 0.00000245, -0.00000833, 0.00000764, -0.00001006, 0.00000588, -0.00000614, -0.00000857, 0.00001297, -0.00001214, 0.00001594, -0.00000354, -0.00000149, 0.00000006, 0.00000180, 0.00000088, 0.00000297, -0.00000061, 0.00000187, -0.00000079, -0.00000129, -0.00000016, -0.00000243, -0.00000153, -0.00000020, -0.00000503, 0.00000937, 0.00000199, -0.00000132, },
		{ -0.00000592, 0.00000577, 0.00000525, -0.00000038, -0.00000809, 0.00000632, -0.00000025, -0.00000489, 0.00000018, -0.00000019, -0.00000324, -0.00000296, 0.00000718, -0.00000832, 0.00000235, -0.00000237, 0.00000639, -0.00000823, -0.00000588, 0.00000211, 0.00000542, -0.00000514, -0.00000327, -0.00000246, -0.00000068, -0.00000145, -0.00000319, -0.00000097, -0.00000034, -0.00000229, 0.00000241, -0.00000616, 0.00000164, -0.00000803, -0.00000592, -0.00000529, 0.00000259, -0.00002128, 0.00000479, -0.00000643, 0.00002082, 0.00000116, 0.00000300, 0.00000297, 0.00000314, -0.00000001, -0.00000808, 0.00000644, 0.00000651, -0.00000293, -0.00000652, 0.00000183, -0.00000627, 0.00000239, -0.00000647, 0.00000052, 0.00000383, -0.00000991, -0.00001636, 0.00000115, -0.00000469, -0.00001318, -0.00004234, -0.00007233, -0.00021431, 0.00021778, 0.00013271, 0.00005269, 0.00006233, 0.00006358, 0.00007208, 0.00011575, 0.00021038, -0.00018592, -0.00007854, -0.00004184, -0.00002779, -0.00002083, -0.00001536, -0.00001143, -0.00001054, -0.00000631, -0.00000085, -0.00001053, -0.00000646, -0.00000505, -0.00000256, -0.00000454, -0.00000305, -0.00000528, 0.00000287, -0.00000499, 0.00000150, -0.00000748, -0.00000589, -0.00000011, -0.00000457, 0.00000074, 0.00000514, -0.00001078, -0.00000424, -0.00000310, 0.00000097, -0.00000140, -0.00000482, 0.00000689, -0.00001969, 0.00001722, -0.00001230, 0.00001154, 0.00000256, -0.00000395, 0.00000024, 0.00000217, 0.00000028, 0.00000316, -0.00000013, 0.00000127, 0.00000003, -0.00000131, -0.00000071, -0.00000086, -0.00000366, 0.00000199, -0.00000645, 0.00000700, 0.00000445, 0.00000005, },
		{ -0.00000446, 0.00000093, 0.00000809, 0.00000026, -0.00000612, 0.00000038, 0.00000563, -0.00000771, 0.00000188, -0.00000291, 0.00000105, -0.00000773, 0.00000956, -0.00000787, 0.00000130, -0.00000258, 0.00000606, -0.00000540, -0.00000761, -0.00000015, 0.00000653, -0.00000332, -0.00000403, -0.00000266, -0.00000003, -0.00000335, -0.00000090, -0.00000349, 0.00000211, -0.00000523, 0.00000442, -0.00000649, 0.00000244, -0.00000913, -0.00000386, -0.00000882, 0.00000579, -0.00002039, -0.00000145, -0.00000322, 0.00001446, 0.00000804, -0.00000071, 0.00000749, -0.00000022, 0.00000477, -0.00000998, 0.00000509, 0.00000810, -0.00000359, -0.00000551, 0.00000198, -0.00000510, -0.00000016, -0.00000369, -0.00000160, 0.00000764, -0.00001476, -0.00001089, -0.00000168, -0.00000148, -0.00001326, -0.00003813, -0.00007179, -0.00020197, 0.00019208, 0.00014219, 0.00003801, 0.00007165, 0.00004120, 0.00008830, 0.00005910, 0.00023442, -0.00008818, -0.00014238, -0.00002209, -0.00004639, -0.00000889, -0.00002597, -0.00000472, -0.00001775, 0.00000091, -0.00000794, -0.00000382, -0.00001243, -0.00000095, -0.00000655, -0.00000092, -0.00000679, -0.00000028, -0.00000304, 0.00000013, -0.00000350, -0.00000069, -0.00001176, 0.00000286, -0.00000713, 0.00000261, 0.00000203, -0.00000365, -0.00001104, 0.00000185, -0.00000546, 0.00000614, -0.00001263, 0.00001409, -0.00002201, 0.00001239, -0.00000520, 0.00000253, 0.00000983, -0.00000585, 0.00000022, 0.00000205, 0.00000060, 0.00000229, 0.00000122, 0.00000012, 0.00000099, -0.00000128, -0.00000116, 0.00000018, -0.00000449, 0.00000207, -0.00000506, 0.00000260, 0.00000728, 0.00000066, },
		{ -0.00000207, -0.00000358, 0.00000931, 0.00000140, -0.00000349, -0.00000482, 0.00000873, -0.00000738, 0.00000131, -0.00000375, 0.00000275, -0.00000853, 0.00000736, -0.00000373, -0.00000218, -0.00000063, 0.00000341, -0.00000124, -0.00000934, -0.00000167, 0.00000558, -0.00000013, -0.00000513, -0.00000259, 0.00000022, -0.00000443, -0.00000011, -0.00000401, 0.00000214, -0.00000520, 0.00000299, -0.00000381, 0.00000074, -0.00000765, -0.00000476, -0.00000824, 0.00000334, -0.00001340, -0.00001142, 0.00000171, 0.00000686, 0.00001445, -0.00000338, 0.00001097, -0.00000214, 0.00000733, -0.00000897, 0.00000191, 0.00001018, -0.00000419, -0.00000478, 0.00000211, -0.00000355, -0.00000249, -0.00000178, -0.00000295, 0.00001054, -0.00001743, -0.00000908, -0.00000055, -0.00000183, -0.00001039, -0.00003645, -0.00006916, -0.00019269, 0.00017198, 0.00014486, 0.00003167, 0.00007219, 0.00003128, 0.00008867, 0.00003719, 0.00020161, 0.00003246, -0.00018730, -0.00002089, -0.00005353, -0.00000621, -0.00002923, -0.00000353, -0.00002098, 0.00000435, -0.00001122, -0.00000090, -0.00001462, -0.00000008, -0.00000809, 0.00000085, -0.00000920, 0.00000330, -0.00000687, 0.00000280, -0.00000621, 0.00000319, -0.00001369, 0.00000231, -0.00000719, 0.00000287, -0.00000019, 0.00000159, -0.00001429, 0.00000308, -0.00000802, 0.00000863, -0.00001375, 0.00001270, -0.00001502, 0.00000051, 0.00000553, -0.00000698, 0.00001493, -0.00000537, -0.00000068, 0.00000180, 0.00000141, 0.00000094, 0.00000275, -0.00000093, 0.00000168, -0.00000108, -0.00000138, 0.00000039, -0.00000390, 0.00000037, -0.00000198, -0.00000214, 0.00000910, 0.00000138, },
		{ -0.00000005, -0.00000612, 0.00000785, 0.00000370, -0.00000157, -0.00000742, 0.00000777, -0.00000362, -0.00000148, -0.00000252, 0.00000157, -0.00000563, 0.00000185, 0.00000216, -0.00000624, 0.00000183, 0.00000020, 0.00000250, -0.00001002, -0.00000288, 0.00000336, 0.00000326, -0.00000574, -0.00000254, 0.00000002, -0.00000441, -0.00000105, -0.00000245, -0.00000003, -0.00000273, -0.00000084, 0.00000050, -0.00000218, -0.00000472, -0.00000752, -0.00000497, -0.00000232, -0.00000431, -0.00002032, 0.00000412, 0.00000161, 0.00001732, -0.00000275, 0.00001170, -0.00000126, 0.00000693, -0.00000532, -0.00000211, 0.00001179, -0.00000408, -0.00000473, 0.00000263, -0.00000240, -0.00000350, -0.00000176, -0.00000249, 0.00001119, -0.00001655, -0.00001168, 0.00000449, -0.00000529, -0.00000497, -0.00003724, -0.00006395, -0.00018754, 0.00015870, 0.00013985, 0.00003393, 0.00006379, 0.00003359, 0.00007399, 0.00004273, 0.00013869, 0.00013804, -0.00018836, -0.00004745, -0.00004520, -0.00001546, -0.00002339, -0.00000927, -0.00001858, 0.00000215, -0.00000897, -0.00000324, -0.00001193, -0.00000327, -0.00000633, -0.00000033, -0.00000877, 0.00000361, -0.00000671, 0.00000137, -0.00000506, 0.00000265, -0.00001079, -0.00000201, -0.00000461, 0.00000102, -0.00000026, 0.00000315, -0.00001255, -0.00000033, -0.00000592, 0.00000552, -0.00000848, 0.00000464, -0.00000285, -0.00001275, 0.00001417, -0.00001230, 0.00001500, -0.00000150, -0.00000269, 0.00000172, 0.00000213, -0.00000010, 0.00000369, -0.00000126, 0.00000170, -0.00000054, -0.00000145, 0.00000003, -0.00000250, -0.00000195, 0.00000104, -0.00000535, 0.00000870, 0.00000299, },
		{ 0.00000105, -0.00000615, 0.00000395, 0.00000677, -0.00000058, -0.00000716, 0.00000352, 0.00000194, -0.00000484, -0.00000054, -0.00000110, -0.00000111, -0.00000410, 0.00000683, -0.00000842, 0.00000280, -0.00000166, 0.00000418, -0.00000877, -0.00000435, 0.00000094, 0.00000556, -0.00000508, -0.00000290, -0.00000035, -0.00000370, -0.00000299, 0.00000004, -0.00000294, 0.00000036, -0.00000485, 0.00000408, -0.00000415, -0.00000238, -0.00000997, -0.00000171, -0.00000753, 0.00000238, -0.00002398, 0.00000126, 0.00000053, 0.00001549, 0.00000156, 0.00000968, 0.00000213, 0.00000439, -0.00000047, -0.00000545, 0.00001202, -0.00000293, -0.00000543, 0.00000363, -0.00000195, -0.00000294, -0.00000354, -0.00000049, 0.00000988, -0.00001283, -0.00001708, 0.00001111, -0.00000939, 0.00000076, -0.00003857, -0.00005779, -0.00018488, 0.00014994, 0.00013054, 0.00004028, 0.00005176, 0.00004130, 0.00005361, 0.00005869, 0.00007632, 0.00019955, -0.00013758, -0.00009589, -0.00002790, -0.00003158, -0.00001267, -0.00001863, -0.00001293, -0.00000382, -0.00000300, -0.00000898, -0.00000625, -0.00000872, -0.00000269, -0.00000346, -0.00000611, 0.00000105, -0.00000315, -0.00000326, -0.00000105, -0.00000111, -0.00000483, -0.00000797, -0.00000120, -0.00000175, 0.00000120, 0.00000163, -0.00000714, -0.00000644, -0.00000132, -0.00000054, -0.00000050, -0.00000499, 0.00000819, -0.00002112, 0.00001594, -0.00001071, 0.00000940, 0.00000506, -0.00000507, 0.00000171, 0.00000240, -0.00000036, 0.00000364, -0.00000064, 0.00000099, 0.00000032, -0.00000148, -0.00000048, -0.00000115, -0.00000364, 0.00000257, -0.00000593, 0.00000583, 0.00000549, },
		{ 0.00000163, -0.00000433, -0.00000094, 0.00000924, 0.00000027, -0.00000520, -0.00000179, 0.00000665, -0.00000655, 0.00000038, -0.00000325, 0.00000244, -0.00000764, 0.00000793, -0.00000721, 0.00000122, -0.00000122, 0.00000324, -0.00000563, -0.00000612, -0.00000100, 0.00000606, -0.00000294, -0.00000369, -0.00000060, -0.00000294, -0.00000485, 0.00000198, -0.00000485, 0.00000212, -0.00000693, 0.00000502, -0.00000362, -0.00000210, -0.00001042, -0.00000059, -0.00000965, 0.00000404, -0.00002107, -0.00000647, 0.00000258, 0.00001028, 0.00000776, 0.00000671, 0.00000625, 0.00000165, 0.00000356, -0.00000667, 0.00001039, -0.00000082, -0.00000662, 0.00000484, -0.00000202, -0.00000126, -0.00000617, 0.00000183, 0.00000800, -0.00000823, -0.00002248, 0.00001619, -0.00001129, 0.00000436, -0.00003824, -0.00005289, -0.00018205, 0.00014203, 0.00012181, 0.00004487, 0.00004267, 0.00004633, 0.00003797, 0.00006911, 0.00003586, 0.00020752, -0.00004779, -0.00014667, -0.00001543, -0.00004506, -0.00000451, -0.00002582, -0.00000858, -0.00000957, 0.00000281, -0.00001439, -0.00000113, -0.00001324, 0.00000015, -0.00000630, -0.00000322, -0.00000233, 0.00000143, -0.00000839, 0.00000306, -0.00000518, 0.00000090, -0.00001240, 0.00000051, -0.00000349, 0.00000252, -0.00000092, -0.00000105, -0.00001182, 0.00000234, -0.00000566, 0.00000556, -0.00001095, 0.00001297, -0.00002099, 0.00000959, -0.00000306, 0.00000039, 0.00001178, -0.00000625, 0.00000121, 0.00000235, 0.00000008, 0.00000280, 0.00000062, -0.00000008, 0.00000123, -0.00000146, -0.00000085, -0.00000041, -0.00000403, 0.00000214, -0.00000411, 0.00000146, 0.00000798, },
		{ 0.00000252, -0.00000200, -0.00000496, 0.00000964, 0.00000194, -0.00000307, -0.00000585, 0.00000837, -0.00000525, -0.00000074, -0.00000366, 0.00000348, -0.00000746, 0.00000504, -0.00000299, -0.00000228, 0.00000089, 0.00000056, -0.00000175, -0.00000749, -0.00000247, 0.00000492, 0.00000006, -0.00000448, -0.00000073, -0.00000255, -0.00000586, 0.00000242, -0.00000483, 0.00000167, -0.00000635, 0.00000298, -0.00000065, -0.00000383, -0.00000877, -0.00000191, -0.00000838, 0.00000109, -0.00001361, -0.00001578, 0.00000463, 0.00000458, 0.00001294, 0.00000527, 0.00000896, 0.00000069, 0.00000523, -0.00000512, 0.00000716, 0.00000169, -0.00000776, 0.00000587, -0.00000219, 0.00000073, -0.00000840, 0.00000314, 0.00000698, -0.00000468, -0.00002557, 0.00001765, -0.00000952, 0.00000483, -0.00003532, -0.00005040, -0.00017730, 0.00013228, 0.00011707, 0.00004400, 0.00004028, 0.00004426, 0.00003243, 0.00006663, 0.00002295, 0.00017285, 0.00005331, -0.00017554, -0.00002089, -0.00004800, -0.00000475, -0.00002645, -0.00000906, -0.00001175, 0.00000510, -0.00001626, 0.00000049, -0.00001432, 0.00000011, -0.00000698, -0.00000203, -0.00000435, 0.00000444, -0.00001127, 0.00000462, -0.00000689, 0.00000364, -0.00001294, -0.00000110, -0.00000298, 0.00000236, -0.00000251, 0.00000303, -0.00001374, 0.00000252, -0.00000716, 0.00000683, -0.00001067, 0.00001011, -0.00001313, -0.00000194, 0.00000658, -0.00000782, 0.00001547, -0.00000467, -0.00000029, 0.00000229, 0.00000075, 0.00000178, 0.00000184, -0.00000091, 0.00000177, -0.00000122, -0.00000102, -0.00000036, -0.00000327, 0.00000035, -0.00000118, -0.00000268, 0.00000913, },
		{ 0.00000429, -0.00000022, -0.00000677, 0.00000735, 0.00000470, -0.00000166, -0.00000742, 0.00000654, -0.00000124, -0.00000341, -0.00000254, 0.00000214, -0.00000429, -0.00000023, 0.00000215, -0.00000572, 0.00000295, -0.00000207, 0.00000118, -0.00000749, -0.00000390, 0.00000292, 0.00000284, -0.00000457, -0.00000098, -0.00000254, -0.00000590, 0.00000138, -0.00000315, -0.00000044, -0.00000396, -0.00000072, 0.00000319, -0.00000606, -0.00000637, -0.00000429, -0.00000544, -0.00000377, -0.00000562, -0.00002239, 0.00000346, 0.00000099, 0.00001475, 0.00000702, 0.00000906, 0.00000238, 0.00000423, -0.00000137, 0.00000339, 0.00000370, -0.00000830, 0.00000636, -0.00000208, 0.00000235, -0.00000935, 0.00000270, 0.00000751, -0.00000304, -0.00002554, 0.00001536, -0.00000476, 0.00000317, -0.00003069, -0.00004979, -0.00017068, 0.00012022, 0.00011677, 0.00003777, 0.00004383, 0.00003604, 0.00003552, 0.00005357, 0.00002961, 0.00011888, 0.00013539, -0.00016508, -0.00004856, -0.00003946, -0.00001384, -0.00002043, -0.00001450, -0.00000981, 0.00000306, -0.00001386, -0.00000194, -0.00001164, -0.00000297, -0.00000521, -0.00000309, -0.00000408, 0.00000466, -0.00001068, 0.00000268, -0.00000537, 0.00000262, -0.00000944, -0.00000560, -0.00000063, 0.00000074, -0.00000245, 0.00000402, -0.00001153, -0.00000096, -0.00000494, 0.00000351, -0.00000526, 0.00000225, -0.00000203, -0.00001313, 0.00001298, -0.00001125, 0.00001399, 0.00000013, -0.00000268, 0.00000236, 0.00000120, 0.00000118, 0.00000242, -0.00000101, 0.00000165, -0.00000064, -0.00000111, -0.00000071, -0.00000200, -0.00000165, 0.00000127, -0.00000500, 0.00000805, },
		{ 0.00000669, 0.00000072, -0.00000624, 0.00000304, 0.00000776, -0.00000077, -0.00000673, 0.00000235, 0.00000362, -0.00000593, -0.00000121, -0.00000014, -0.00000024, -0.00000516, 0.00000554, -0.00000703, 0.00000326, -0.00000307, 0.00000191, -0.00000557, -0.00000562, 0.00000090, 0.00000443, -0.00000345, -0.00000153, -0.00000269, -0.00000539, -0.00000036, -0.00000098, -0.00000277, -0.00000151, -0.00000401, 0.00000580, -0.00000696, -0.00000497, -0.00000584, -0.00000318, -0.00000745, -0.00000075, -0.00002340, -0.00000219, 0.00000022, 0.00001271, 0.00001180, 0.00000703, 0.00000603, 0.00000169, 0.00000302, 0.00000059, 0.00000436, -0.00000788, 0.00000616, -0.00000153, 0.00000326, -0.00000879, 0.00000071, 0.00000918, -0.00000289, -0.00002324, 0.00001104, 0.00000072, 0.00000163, -0.00002629, -0.00004941, -0.00016360, 0.00010726, 0.00011886, 0.00002925, 0.00004935, 0.00002630, 0.00004136, 0.00003790, 0.00004186, 0.00006955, 0.00017799, -0.00011317, -0.00009026, -0.00002664, -0.00002655, -0.00001202, -0.00002163, -0.00000615, -0.00000136, -0.00000910, -0.00000652, -0.00000712, -0.00000730, -0.00000242, -0.00000539, -0.00000223, 0.00000271, -0.00000748, -0.00000159, -0.00000191, -0.00000070, -0.00000387, -0.00001074, 0.00000170, -0.00000112, -0.00000150, 0.00000274, -0.00000672, -0.00000606, -0.00000118, -0.00000174, 0.00000164, -0.00000579, 0.00000667, -0.00001879, 0.00001258, -0.00000827, 0.00000756, 0.00000685, -0.00000490, 0.00000227, 0.00000127, 0.00000126, 0.00000220, -0.00000033, 0.00000093, 0.00000020, -0.00000119, -0.00000109, -0.00000098, -0.00000284, 0.00000213, -0.00000486, 0.00000486, },
		{ 0.00000871, 0.00000135, -0.00000425, -0.00000168, 0.00000966, 0.00000045, -0.00000502, -0.00000209, 0.00000704, -0.00000653, -0.00000105, -0.00000171, 0.00000257, -0.00000750, 0.00000551, -0.00000526, 0.00000126, -0.00000198, 0.00000034, -0.00000217, -0.00000735, -0.00000078, 0.00000445, -0.00000122, -0.00000226, -0.00000282, -0.00000488, -0.00000186, 0.00000043, -0.00000392, -0.00000049, -0.00000534, 0.00000584, -0.00000553, -0.00000548, -0.00000550, -0.00000296, -0.00000823, -0.00000048, -0.00001881, -0.00001080, 0.00000074, 0.00000826, 0.00001773, 0.00000474, 0.00000983, -0.00000046, 0.00000618, -0.00000004, 0.00000330, -0.00000658, 0.00000547, -0.00000066, 0.00000356, -0.00000714, -0.00000196, 0.00001092, -0.00000305, -0.00002037, 0.00000704, 0.00000441, 0.00000231, -0.00002374, -0.00004783, -0.00015755, 0.00009527, 0.00012057, 0.00002217, 0.00005243, 0.00001987, 0.00004404, 0.00002723, 0.00004812, 0.00003964, 0.00017736, -0.00003402, -0.00012827, -0.00002100, -0.00003529, -0.00000701, -0.00002599, -0.00000426, -0.00000502, -0.00000507, -0.00001024, -0.00000361, -0.00001035, -0.00000070, -0.00000716, -0.00000044, 0.00000034, -0.00000378, -0.00000577, 0.00000109, -0.00000378, 0.00000089, -0.00001383, 0.00000199, -0.00000170, -0.00000098, 0.00000092, -0.00000184, -0.00000996, 0.00000129, -0.00000567, 0.00000617, -0.00000980, 0.00000917, -0.00001670, 0.00000533, -0.00000057, -0.00000104, 0.00001277, -0.00000538, 0.00000146, 0.00000117, 0.00000178, 0.00000148, 0.00000075, 0.00000003, 0.00000097, -0.00000118, -0.00000131, -0.00000056, -0.00000284, 0.00000130, -0.00000278, 0.00000073, },
		{ 0.00000920, 0.00000243, -0.00000207, -0.00000506, 0.00000918, 0.00000270, -0.00000346, -0.00000500, 0.00000756, -0.00000450, -0.00000254, -0.00000185, 0.00000309, -0.00000654, 0.00000219, -0.00000121, -0.00000208, 0.00000025, -0.00000236, 0.00000138, -0.00000829, -0.00000228, 0.00000325, 0.00000141, -0.00000270, -0.00000296, -0.00000469, -0.00000255, 0.00000044, -0.00000336, -0.00000134, -0.00000446, 0.00000348, -0.00000229, -0.00000739, -0.00000363, -0.00000454, -0.00000645, -0.00000367, -0.00001131, -0.00001890, -0.00000013, 0.00000364, 0.00002236, 0.00000429, 0.00001200, -0.00000064, 0.00000695, 0.00000186, 0.00000091, -0.00000492, 0.00000470, 0.00000019, 0.00000360, -0.00000510, -0.00000422, 0.00001165, -0.00000240, -0.00001839, 0.00000510, 0.00000493, 0.00000590, -0.00002332, -0.00004479, -0.00015304, 0.00008526, 0.00012015, 0.00001878, 0.00005080, 0.00001893, 0.00004099, 0.00002464, 0.00004417, 0.00003120, 0.00014531, 0.00004824, -0.00014356, -0.00003150, -0.00003513, -0.00000904, -0.00002493, -0.00000629, -0.00000576, -0.00000405, -0.00001094, -0.00000308, -0.00001047, -0.00000136, -0.00000717, -0.00000008, -0.00000080, -0.00000159, -0.00000773, 0.00000163, -0.00000463, 0.00000279, -0.00001328, -0.00000065, -0.00000043, -0.00000166, -0.00000002, 0.00000110, -0.00001071, 0.00000076, -0.00000643, 0.00000639, -0.00000828, 0.00000523, -0.00000859, -0.00000531, 0.00000767, -0.00000782, 0.00001507, -0.00000291, -0.00000035, 0.00000114, 0.00000232, 0.00000081, 0.00000162, -0.00000052, 0.00000133, -0.00000093, -0.00000138, -0.00000069, -0.00000197, -0.00000047, -0.00000012, -0.00000272, },
		{ 0.00000758, 0.00000429, -0.00000056, -0.00000616, 0.00000615, 0.00000583, -0.00000247, -0.00000574, 0.00000524, -0.00000061, -0.00000484, -0.00000101, 0.00000178, -0.00000332, -0.00000262, 0.00000301, -0.00000488, 0.00000196, -0.00000451, 0.00000357, -0.00000761, -0.00000396, 0.00000154, 0.00000350, -0.00000230, -0.00000325, -0.00000483, -0.00000244, -0.00000073, -0.00000160, -0.00000329, -0.00000240, 0.00000015, 0.00000112, -0.00000921, -0.00000164, -0.00000647, -0.00000391, -0.00000768, -0.00000461, -0.00002304, -0.00000432, 0.00000040, 0.00002404, 0.00000677, 0.00001184, 0.00000157, 0.00000546, 0.00000546, -0.00000170, -0.00000373, 0.00000437, 0.00000069, 0.00000377, -0.00000336, -0.00000530, 0.00001083, -0.00000055, -0.00001784, 0.00000560, 0.00000260, 0.00001138, -0.00002381, -0.00004127, -0.00014942, 0.00007695, 0.00011751, 0.00001903, 0.00004513, 0.00002217, 0.00003375, 0.00002805, 0.00003304, 0.00003640, 0.00010160, 0.00011001, -0.00012486, -0.00005875, -0.00002723, -0.00001714, -0.00001939, -0.00001157, -0.00000373, -0.00000615, -0.00000855, -0.00000550, -0.00000787, -0.00000412, -0.00000551, -0.00000135, -0.00000017, -0.00000168, -0.00000680, -0.00000068, -0.00000295, 0.00000164, -0.00000948, -0.00000532, 0.00000190, -0.00000317, 0.00000015, 0.00000160, -0.00000825, -0.00000246, -0.00000438, 0.00000299, -0.00000285, -0.00000213, 0.00000104, -0.00001427, 0.00001197, -0.00000960, 0.00001235, 0.00000238, -0.00000271, 0.00000113, 0.00000255, 0.00000063, 0.00000190, -0.00000040, 0.00000108, -0.00000038, -0.00000141, -0.00000107, -0.00000086, -0.00000213, 0.00000176, -0.00000424, },
		{ 0.00000429, 0.00000647, 0.00000024, -0.00000514, 0.00000156, 0.00000873, -0.00000154, -0.00000492, 0.00000150, 0.00000328, -0.00000644, -0.00000035, -0.00000004, 0.00000022, -0.00000652, 0.00000522, -0.00000549, 0.00000184, -0.00000491, 0.00000353, -0.00000515, -0.00000584, -0.00000012, 0.00000441, -0.00000087, -0.00000369, -0.00000513, -0.00000192, -0.00000231, 0.00000026, -0.00000502, -0.00000070, -0.00000237, 0.00000300, -0.00000948, -0.00000088, -0.00000725, -0.00000245, -0.00001005, -0.00000140, -0.00002167, -0.00001173, -0.00000165, 0.00002271, 0.00001168, 0.00001014, 0.00000523, 0.00000299, 0.00000920, -0.00000319, -0.00000374, 0.00000481, 0.00000065, 0.00000430, -0.00000227, -0.00000502, 0.00000874, 0.00000203, -0.00001822, 0.00000765, -0.00000093, 0.00001668, -0.00002329, -0.00003878, -0.00014552, 0.00006928, 0.00011379, 0.00002112, 0.00003808, 0.00002625, 0.00002613, 0.00003264, 0.00002119, 0.00004415, 0.00006450, 0.00013758, -0.00007411, -0.00009349, -0.00001866, -0.00002641, -0.00001324, -0.00001715, -0.00000110, -0.00000955, -0.00000488, -0.00000908, -0.00000434, -0.00000734, -0.00000346, -0.00000331, 0.00000150, -0.00000334, -0.00000395, -0.00000453, -0.00000012, -0.00000103, -0.00000445, -0.00000981, 0.00000353, -0.00000437, 0.00000068, 0.00000054, -0.00000412, -0.00000642, -0.00000154, -0.00000153, 0.00000310, -0.00000861, 0.00000746, -0.00001752, 0.00000995, -0.00000573, 0.00000556, 0.00000877, -0.00000438, 0.00000076, 0.00000248, 0.00000097, 0.00000158, 0.00000030, 0.00000041, 0.00000034, -0.00000143, -0.00000139, -0.00000010, -0.00000292, 0.00000210, -0.00000356, },
		{ 0.00000056, 0.00000799, 0.00000091, -0.00000304, -0.00000287, 0.00000997, 0.00000015, -0.00000367, -0.00000183, 0.00000539, -0.00000609, -0.00000079, -0.00000118, 0.00000242, -0.00000786, 0.00000433, -0.00000345, -0.00000024, -0.00000352, 0.00000160, -0.00000165, -0.00000743, -0.00000162, 0.00000405, 0.00000129, -0.00000405, -0.00000550, -0.00000146, -0.00000354, 0.00000124, -0.00000550, -0.00000041, -0.00000303, 0.00000256, -0.00000775, -0.00000176, -0.00000637, -0.00000281, -0.00000987, -0.00000219, -0.00001591, -0.00002001, -0.00000410, 0.00001955, 0.00001732, 0.00000862, 0.00000867, 0.00000126, 0.00001150, -0.00000261, -0.00000518, 0.00000594, 0.00000018, 0.00000514, -0.00000176, -0.00000378, 0.00000620, 0.00000438, -0.00001852, 0.00000984, -0.00000373, 0.00001986, -0.00002036, -0.00003823, -0.00014057, 0.00006138, 0.00011028, 0.00002303, 0.00003245, 0.00002797, 0.00002154, 0.00003427, 0.00001397, 0.00004630, 0.00004382, 0.00013107, -0.00000565, -0.00012040, -0.00001843, -0.00003122, -0.00001076, -0.00001981, -0.00000040, -0.00001189, -0.00000228, -0.00001152, -0.00000204, -0.00000914, -0.00000255, -0.00000466, 0.00000292, -0.00000511, -0.00000101, -0.00000784, 0.00000183, -0.00000309, -0.00000061, -0.00001196, 0.00000297, -0.00000422, 0.00000064, -0.00000069, -0.00000036, -0.00000892, -0.00000009, -0.00000459, 0.00000648, -0.00001098, 0.00000807, -0.00001410, 0.00000249, 0.00000152, -0.00000227, 0.00001356, -0.00000391, -0.00000044, 0.00000234, 0.00000153, 0.00000104, 0.00000118, -0.00000029, 0.00000090, -0.00000134, -0.00000154, 0.00000010, -0.00000269, 0.00000107, -0.00000145, },
		{ -0.00000218, 0.00000791, 0.00000208, -0.00000101, -0.00000562, 0.00000865, 0.00000299, -0.00000277, -0.00000357, 0.00000494, -0.00000369, -0.00000232, -0.00000128, 0.00000268, -0.00000643, 0.00000091, 0.00000014, -0.00000316, -0.00000145, -0.00000093, 0.00000150, -0.00000789, -0.00000320, 0.00000285, 0.00000344, -0.00000389, -0.00000594, -0.00000129, -0.00000400, 0.00000096, -0.00000455, -0.00000159, -0.00000198, 0.00000031, -0.00000482, -0.00000351, -0.00000451, -0.00000434, -0.00000795, -0.00000535, -0.00000878, -0.00002593, -0.00000879, 0.00001594, 0.00002182, 0.00000893, 0.00001055, 0.00000146, 0.00001164, 0.00000007, -0.00000755, 0.00000727, -0.00000034, 0.00000600, -0.00000150, -0.00000226, 0.00000420, 0.00000560, -0.00001792, 0.00001109, -0.00000456, 0.00002016, -0.00001492, -0.00003935, -0.00013480, 0.00005315, 0.00010752, 0.00002358, 0.00002968, 0.00002607, 0.00002103, 0.00003165, 0.00001290, 0.00004077, 0.00003936, 0.00010209, 0.00006019, -0.00012520, -0.00003198, -0.00002896, -0.00001383, -0.00001824, -0.00000271, -0.00001192, -0.00000216, -0.00001151, -0.00000216, -0.00000856, -0.00000351, -0.00000464, 0.00000317, -0.00000577, 0.00000049, -0.00000903, 0.00000150, -0.00000316, 0.00000063, -0.00001082, -0.00000007, -0.00000263, -0.00000029, -0.00000121, 0.00000166, -0.00000874, -0.00000106, -0.00000501, 0.00000611, -0.00000865, 0.00000354, -0.00000633, -0.00000676, 0.00000816, -0.00000756, 0.00001441, -0.00000053, -0.00000250, 0.00000225, 0.00000194, 0.00000073, 0.00000177, -0.00000061, 0.00000105, -0.00000101, -0.00000159, -0.00000011, -0.00000185, -0.00000053, 0.00000083, },
		{ -0.00000304, 0.00000597, 0.00000386, 0.00000032, -0.00000608, 0.00000499, 0.00000638, -0.00000216, -0.00000366, 0.00000250, -0.00000033, -0.00000399, -0.00000090, 0.00000160, -0.00000340, -0.00000324, 0.00000335, -0.00000522, -0.00000021, -0.00000255, 0.00000302, -0.00000663, -0.00000506, 0.00000138, 0.00000488, -0.00000290, -0.00000647, -0.00000144, -0.00000379, -0.00000024, -0.00000279, -0.00000339, -0.00000029, -0.00000234, -0.00000221, -0.00000476, -0.00000295, -0.00000568, -0.00000595, -0.00000848, -0.00000346, -0.00002709, -0.00001626, 0.00001237, 0.00002410, 0.00001172, 0.00001055, 0.00000363, 0.00001001, 0.00000389, -0.00000978, 0.00000807, -0.00000047, 0.00000654, -0.00000106, -0.00000113, 0.00000340, 0.00000535, -0.00001626, 0.00001116, -0.00000337, 0.00001819, -0.00000822, -0.00004080, -0.00012923, 0.00004522, 0.00010521, 0.00002285, 0.00002944, 0.00002146, 0.00002315, 0.00002644, 0.00001580, 0.00003064, 0.00004407, 0.00006690, 0.00010565, -0.00010150, -0.00005733, -0.00002212, -0.00002071, -0.00001391, -0.00000696, -0.00001020, -0.00000419, -0.00000939, -0.00000434, -0.00000611, -0.00000580, -0.00000358, 0.00000227, -0.00000507, 0.00000014, -0.00000783, -0.00000107, -0.00000134, -0.00000053, -0.00000716, -0.00000436, -0.00000068, -0.00000154, -0.00000108, 0.00000186, -0.00000621, -0.00000382, -0.00000342, 0.00000299, -0.00000351, -0.00000305, 0.00000157, -0.00001344, 0.00001061, -0.00000798, 0.00001064, 0.00000512, -0.00000468, 0.00000205, 0.00000201, 0.00000088, 0.00000187, -0.00000037, 0.00000071, -0.00000045, -0.00000163, -0.00000043, -0.00000098, -0.00000180, 0.00000216, },
		{ -0.00000206, 0.00000281, 0.00000568, 0.00000107, -0.00000470, 0.00000029, 0.00000902, -0.00000112, -0.00000289, -0.00000046, 0.00000235, -0.00000458, -0.00000091, 0.00000025, -0.00000043, -0.00000615, 0.00000450, -0.00000521, -0.00000073, -0.00000246, 0.00000244, -0.00000382, -0.00000696, -0.00000009, 0.00000525, -0.00000113, -0.00000695, -0.00000183, -0.00000325, -0.00000166, -0.00000119, -0.00000470, 0.00000081, -0.00000399, -0.00000120, -0.00000448, -0.00000266, -0.00000579, -0.00000518, -0.00000984, -0.00000167, -0.00002328, -0.00002514, 0.00000812, 0.00002417, 0.00001642, 0.00000957, 0.00000676, 0.00000792, 0.00000744, -0.00001076, 0.00000776, 0.00000007, 0.00000657, -0.00000019, -0.00000070, 0.00000390, 0.00000400, -0.00001410, 0.00001060, -0.00000106, 0.00001550, -0.00000206, -0.00004104, -0.00012496, 0.00003836, 0.00010268, 0.00002165, 0.00003032, 0.00001634, 0.00002524, 0.00002160, 0.00001893, 0.00002093, 0.00004940, 0.00003902, 0.00012221, -0.00005388, -0.00008514, -0.00001708, -0.00002716, -0.00001010, -0.00001069, -0.00000856, -0.00000678, -0.00000673, -0.00000699, -0.00000332, -0.00000805, -0.00000251, 0.00000098, -0.00000365, -0.00000135, -0.00000526, -0.00000457, 0.00000100, -0.00000260, -0.00000292, -0.00000787, 0.00000013, -0.00000210, -0.00000096, 0.00000106, -0.00000274, -0.00000670, -0.00000153, -0.00000069, 0.00000143, -0.00000800, 0.00000585, -0.00001468, 0.00000753, -0.00000370, 0.00000377, 0.00001105, -0.00000568, 0.00000130, 0.00000188, 0.00000136, 0.00000164, 0.00000023, 0.00000010, 0.00000017, -0.00000163, -0.00000065, -0.00000049, -0.00000224, 0.00000212, },
		{ -0.00000011, -0.00000036, 0.00000664, 0.00000181, -0.00000255, -0.00000376, 0.00000957, 0.00000106, -0.00000220, -0.00000256, 0.00000316, -0.00000346, -0.00000174, -0.00000058, 0.00000127, -0.00000672, 0.00000305, -0.00000312, -0.00000282, -0.00000094, 0.00000042, -0.00000044, -0.00000822, -0.00000162, 0.00000469, 0.00000097, -0.00000708, -0.00000238, -0.00000277, -0.00000268, -0.00000050, -0.00000484, 0.00000066, -0.00000404, -0.00000213, -0.00000265, -0.00000362, -0.00000458, -0.00000583, -0.00000919, -0.00000307, -0.00001652, -0.00003268, 0.00000193, 0.00002261, 0.00002172, 0.00000902, 0.00000948, 0.00000676, 0.00000951, -0.00000988, 0.00000630, 0.00000119, 0.00000615, 0.00000106, -0.00000090, 0.00000527, 0.00000239, -0.00001232, 0.00001027, 0.00000120, 0.00001364, 0.00000213, -0.00003920, -0.00012239, 0.00003284, 0.00009954, 0.00002078, 0.00003085, 0.00001266, 0.00002515, 0.00001935, 0.00001954, 0.00001522, 0.00004986, 0.00002463, 0.00011244, 0.00000412, -0.00010271, -0.00002045, -0.00002933, -0.00000973, -0.00001173, -0.00000869, -0.00000828, -0.00000524, -0.00000850, -0.00000175, -0.00000894, -0.00000247, 0.00000023, -0.00000250, -0.00000279, -0.00000284, -0.00000725, 0.00000224, -0.00000387, -0.00000005, -0.00000894, -0.00000121, -0.00000140, -0.00000142, 0.00000029, 0.00000012, -0.00000810, -0.00000093, -0.00000297, 0.00000386, -0.00000907, 0.00000503, -0.00001051, 0.00000048, 0.00000273, -0.00000319, 0.00001472, -0.00000428, -0.00000031, 0.00000173, 0.00000184, 0.00000143, 0.00000082, -0.00000041, 0.00000060, -0.00000151, -0.00000074, -0.00000047, -0.00000188, 0.00000106, },
		{ 0.00000168, -0.00000236, 0.00000605, 0.00000302, -0.00000068, -0.00000589, 0.00000749, 0.00000435, -0.00000191, -0.00000323, 0.00000199, -0.00000104, -0.00000303, -0.00000084, 0.00000147, -0.00000511, -0.00000015, -0.00000012, -0.00000528, 0.00000081, -0.00000172, 0.00000214, -0.00000808, -0.00000346, 0.00000363, 0.00000275, -0.00000654, -0.00000307, -0.00000255, -0.00000304, -0.00000085, -0.00000387, -0.00000052, -0.00000290, -0.00000425, -0.00000021, -0.00000496, -0.00000290, -0.00000704, -0.00000757, -0.00000584, -0.00000983, -0.00003621, -0.00000684, 0.00001978, 0.00002635, 0.00001011, 0.00001085, 0.00000731, 0.00000975, -0.00000735, 0.00000425, 0.00000243, 0.00000563, 0.00000242, -0.00000135, 0.00000679, 0.00000142, -0.00001165, 0.00001075, 0.00000267, 0.00001343, 0.00000402, -0.00003561, -0.00012091, 0.00002815, 0.00009596, 0.00002045, 0.00003035, 0.00001126, 0.00002231, 0.00002003, 0.00001717, 0.00001403, 0.00004471, 0.00002229, 0.00008703, 0.00005594, -0.00009995, -0.00003483, -0.00002641, -0.00001342, -0.00000974, -0.00001093, -0.00000811, -0.00000564, -0.00000818, -0.00000201, -0.00000802, -0.00000376, 0.00000039, -0.00000220, -0.00000332, -0.00000171, -0.00000800, 0.00000146, -0.00000348, 0.00000054, -0.00000716, -0.00000449, 0.00000024, -0.00000241, -0.00000002, 0.00000158, -0.00000740, -0.00000209, -0.00000320, 0.00000317, -0.00000634, 0.00000036, -0.00000357, -0.00000705, 0.00000774, -0.00000717, 0.00001438, -0.00000018, -0.00000254, 0.00000159, 0.00000202, 0.00000158, 0.00000111, -0.00000056, 0.00000064, -0.00000118, -0.00000076, -0.00000072, -0.00000115, -0.00000026, },
		{ 0.00000249, -0.00000265, 0.00000394, 0.00000463, 0.00000051, -0.00000581, 0.00000339, 0.00000782, -0.00000151, -0.00000289, -0.00000022, 0.00000147, -0.00000386, -0.00000102, 0.00000073, -0.00000247, -0.00000348, 0.00000213, -0.00000671, 0.00000156, -0.00000272, 0.00000295, -0.00000621, -0.00000559, 0.00000241, 0.00000374, -0.00000523, -0.00000374, -0.00000264, -0.00000285, -0.00000186, -0.00000246, -0.00000191, -0.00000153, -0.00000633, 0.00000156, -0.00000554, -0.00000182, -0.00000767, -0.00000639, -0.00000794, -0.00000569, -0.00003460, -0.00001728, 0.00001525, 0.00002959, 0.00001322, 0.00001085, 0.00000937, 0.00000875, -0.00000414, 0.00000253, 0.00000319, 0.00000540, 0.00000355, -0.00000161, 0.00000783, 0.00000165, -0.00001228, 0.00001200, 0.00000330, 0.00001470, 0.00000441, -0.00003153, -0.00011931, 0.00002340, 0.00009250, 0.00002040, 0.00002895, 0.00001169, 0.00001773, 0.00002222, 0.00001335, 0.00001538, 0.00003674, 0.00002597, 0.00005893, 0.00008874, -0.00007418, -0.00005660, -0.00002149, -0.00001908, -0.00000636, -0.00001406, -0.00000701, -0.00000736, -0.00000660, -0.00000353, -0.00000594, -0.00000574, 0.00000109, -0.00000267, -0.00000282, -0.00000205, -0.00000683, -0.00000099, -0.00000180, -0.00000066, -0.00000354, -0.00000833, 0.00000167, -0.00000325, -0.00000010, 0.00000181, -0.00000515, -0.00000426, -0.00000210, 0.00000048, -0.00000176, -0.00000519, 0.00000246, -0.00001151, 0.00000866, -0.00000668, 0.00000997, 0.00000558, -0.00000445, 0.00000120, 0.00000188, 0.00000209, 0.00000109, -0.00000034, 0.00000035, -0.00000070, -0.00000076, -0.00000099, -0.00000054, -0.00000114, },
		{ 0.00000217, -0.00000151, 0.00000103, 0.00000605, 0.00000123, -0.00000417, -0.00000125, 0.00001000, -0.00000017, -0.00000242, -0.00000217, 0.00000282, -0.00000347, -0.00000161, -0.00000019, -0.00000010, -0.00000544, 0.00000246, -0.00000634, 0.00000072, -0.00000215, 0.00000190, -0.00000310, -0.00000754, 0.00000109, 0.00000386, -0.00000344, -0.00000417, -0.00000305, -0.00000239, -0.00000294, -0.00000140, -0.00000267, -0.00000085, -0.00000738, 0.00000181, -0.00000477, -0.00000191, -0.00000711, -0.00000642, -0.00000836, -0.00000486, -0.00002891, -0.00002699, 0.00000812, 0.00003118, 0.00001796, 0.00001024, 0.00001205, 0.00000763, -0.00000140, 0.00000198, 0.00000309, 0.00000567, 0.00000424, -0.00000142, 0.00000812, 0.00000307, -0.00001372, 0.00001341, 0.00000369, 0.00001661, 0.00000466, -0.00002847, -0.00011644, 0.00001789, 0.00008957, 0.00002025, 0.00002721, 0.00001284, 0.00001309, 0.00002397, 0.00001023, 0.00001652, 0.00002963, 0.00002936, 0.00003783, 0.00009787, -0.00003147, -0.00007703, -0.00001993, -0.00002339, -0.00000419, -0.00001617, -0.00000641, -0.00000912, -0.00000503, -0.00000507, -0.00000393, -0.00000736, 0.00000154, -0.00000331, -0.00000186, -0.00000319, -0.00000473, -0.00000385, -0.00000008, -0.00000232, 0.00000016, -0.00001094, 0.00000163, -0.00000317, -0.00000047, 0.00000152, -0.00000255, -0.00000612, -0.00000101, -0.00000239, 0.00000216, -0.00000869, 0.00000484, -0.00001115, 0.00000505, -0.00000249, 0.00000338, 0.00001088, -0.00000479, 0.00000014, 0.00000160, 0.00000272, 0.00000101, 0.00000003, -0.00000005, -0.00000025, -0.00000072, -0.00000115, -0.00000030, -0.00000129, },
		{ 0.00000116, 0.00000019, -0.00000155, 0.00000650, 0.00000202, -0.00000208, -0.00000479, 0.00000975, 0.00000260, -0.00000241, -0.00000300, 0.00000248, -0.00000183, -0.00000258, -0.00000089, 0.00000122, -0.00000543, 0.00000081, -0.00000447, -0.00000128, -0.00000053, -0.00000016, 0.00000011, -0.00000851, -0.00000055, 0.00000339, -0.00000167, -0.00000402, -0.00000370, -0.00000199, -0.00000362, -0.00000116, -0.00000243, -0.00000115, -0.00000719, 0.00000060, -0.00000293, -0.00000285, -0.00000561, -0.00000742, -0.00000738, -0.00000635, -0.00002175, -0.00003318, -0.00000212, 0.00003082, 0.00002354, 0.00001007, 0.00001433, 0.00000743, 0.00000006, 0.00000289, 0.00000223, 0.00000632, 0.00000456, -0.00000082, 0.00000779, 0.00000520, -0.00001514, 0.00001416, 0.00000453, 0.00001823, 0.00000597, -0.00002736, -0.00011196, 0.00001163, 0.00008712, 0.00001990, 0.00002558, 0.00001373, 0.00000972, 0.00002392, 0.00000912, 0.00001577, 0.00002566, 0.00002892, 0.00002737, 0.00008722, 0.00001607, -0.00008612, -0.00002612, -0.00002401, -0.00000496, -0.00001601, -0.00000730, -0.00000989, -0.00000458, -0.00000559, -0.00000297, -0.00000779, 0.00000115, -0.00000355, -0.00000111, -0.00000420, -0.00000295, -0.00000573, 0.00000044, -0.00000313, 0.00000242, -0.00001111, -0.00000042, -0.00000200, -0.00000135, 0.00000133, -0.00000061, -0.00000668, -0.00000093, -0.00000408, 0.00000376, -0.00000871, 0.00000299, -0.00000672, -0.00000116, 0.00000281, -0.00000255, 0.00001349, -0.00000263, -0.00000171, 0.00000134, 0.00000313, 0.00000119, 0.00000025, -0.00000030, -0.00000000, -0.00000056, -0.00000119, -0.00000041, -0.00000086, },
		{ 0.00000012, 0.00000152, -0.00000292, 0.00000555, 0.00000321, -0.00000039, -0.00000624, 0.00000688, 0.00000636, -0.00000266, -0.00000280, 0.00000094, 0.00000027, -0.00000335, -0.00000147, 0.00000146, -0.00000384, -0.00000189, -0.00000223, -0.00000332, 0.00000100, -0.00000193, 0.00000219, -0.00000787, -0.00000267, 0.00000268, -0.00000045, -0.00000314, -0.00000446, -0.00000183, -0.00000374, -0.00000176, -0.00000145, -0.00000205, -0.00000626, -0.00000129, -0.00000096, -0.00000378, -0.00000406, -0.00000845, -0.00000612, -0.00000829, -0.00001596, -0.00003409, -0.00001464, 0.00002774, 0.00002925, 0.00001116, 0.00001563, 0.00000858, 0.00000019, 0.00000487, 0.00000114, 0.00000699, 0.00000479, -0.00000010, 0.00000726, 0.00000729, -0.00001569, 0.00001368, 0.00000617, 0.00001911, 0.00000879, -0.00002816, -0.00010647, 0.00000540, 0.00008457, 0.00001961, 0.00002414, 0.00001394, 0.00000800, 0.00002200, 0.00000992, 0.00001312, 0.00002479, 0.00002466, 0.00002554, 0.00006610, 0.00005545, -0.00007734, -0.00004048, -0.00002141, -0.00000843, -0.00001384, -0.00000951, -0.00000954, -0.00000547, -0.00000489, -0.00000327, -0.00000696, -0.00000008, -0.00000325, -0.00000096, -0.00000447, -0.00000229, -0.00000590, -0.00000072, -0.00000265, 0.00000277, -0.00000888, -0.00000389, -0.00000036, -0.00000242, 0.00000134, 0.00000033, -0.00000578, -0.00000194, -0.00000429, 0.00000289, -0.00000583, -0.00000148, -0.00000088, -0.00000684, 0.00000622, -0.00000536, 0.00001220, 0.00000191, -0.00000387, 0.00000103, 0.00000314, 0.00000177, 0.00000024, -0.00000026, -0.00000005, -0.00000026, -0.00000118, -0.00000067, -0.00000027, },
		{ -0.00000045, 0.00000195, -0.00000283, 0.00000342, 0.00000460, 0.00000063, -0.00000561, 0.00000237, 0.00000983, -0.00000233, -0.00000234, -0.00000069, 0.00000168, -0.00000322, -0.00000225, 0.00000105, -0.00000165, -0.00000431, -0.00000090, -0.00000434, 0.00000145, -0.00000247, 0.00000246, -0.00000559, -0.00000501, 0.00000186, 0.00000002, -0.00000167, -0.00000507, -0.00000204, -0.00000339, -0.00000281, -0.00000031, -0.00000286, -0.00000539, -0.00000288, 0.00000017, -0.00000387, -0.00000324, -0.00000868, -0.00000563, -0.00000921, -0.00001313, -0.00002998, -0.00002707, 0.00002082, 0.00003444, 0.00001385, 0.00001600, 0.00001081, -0.00000040, 0.00000711, 0.00000054, 0.00000724, 0.00000525, 0.00000041, 0.00000697, 0.00000878, -0.00001494, 0.00001197, 0.00000838, 0.00001940, 0.00001281, -0.00002998, -0.00010123, 0.00000026, 0.00008124, 0.00001974, 0.00002272, 0.00001361, 0.00000747, 0.00001912, 0.00001146, 0.00000982, 0.00002541, 0.00001886, 0.00002753, 0.00004447, 0.00007790, -0.00005072, -0.00005851, -0.00001881, -0.00001256, -0.00001127, -0.00001187, -0.00000888, -0.00000704, -0.00000360, -0.00000421, -0.00000552, -0.00000155, -0.00000277, -0.00000129, -0.00000398, -0.00000278, -0.00000457, -0.00000304, -0.00000141, 0.00000182, -0.00000537, -0.00000733, 0.00000059, -0.00000296, 0.00000122, 0.00000059, -0.00000406, -0.00000333, -0.00000369, 0.00000062, -0.00000189, -0.00000591, 0.00000335, -0.00000936, 0.00000594, -0.00000427, 0.00000752, 0.00000745, -0.00000531, 0.00000034, 0.00000285, 0.00000259, 0.00000013, -0.00000007, -0.00000026, 0.00000010, -0.00000114, -0.00000091, 0.00000014, },
		{ -0.00000045, 0.00000152, -0.00000171, 0.00000090, 0.00000562, 0.00000129, -0.00000373, -0.00000217, 0.00001150, -0.00000054, -0.00000243, -0.00000140, 0.00000167, -0.00000196, -0.00000325, 0.00000037, 0.00000025, -0.00000544, -0.00000113, -0.00000395, 0.00000057, -0.00000164, 0.00000117, -0.00000243, -0.00000689, 0.00000073, -0.00000008, -0.00000006, -0.00000519, -0.00000265, -0.00000279, -0.00000385, 0.00000041, -0.00000303, -0.00000514, -0.00000358, 0.00000003, -0.00000292, -0.00000335, -0.00000789, -0.00000615, -0.00000871, -0.00001311, -0.00002303, -0.00003631, 0.00000941, 0.00003822, 0.00001816, 0.00001595, 0.00001346, -0.00000086, 0.00000877, 0.00000093, 0.00000691, 0.00000608, 0.00000054, 0.00000715, 0.00000946, -0.00001303, 0.00000964, 0.00001048, 0.00001968, 0.00001724, -0.00003162, -0.00009747, -0.00000303, 0.00007680, 0.00002043, 0.00002120, 0.00001306, 0.00000734, 0.00001646, 0.00001241, 0.00000725, 0.00002567, 0.00001406, 0.00002887, 0.00002904, 0.00008150, -0.00001290, -0.00007255, -0.00002036, -0.00001493, -0.00001016, -0.00001302, -0.00000891, -0.00000834, -0.00000268, -0.00000491, -0.00000434, -0.00000254, -0.00000259, -0.00000173, -0.00000314, -0.00000385, -0.00000264, -0.00000536, -0.00000049, 0.00000072, -0.00000212, -0.00000923, -0.00000010, -0.00000250, 0.00000065, 0.00000069, -0.00000238, -0.00000420, -0.00000319, -0.00000165, 0.00000113, -0.00000814, 0.00000420, -0.00000791, 0.00000226, -0.00000048, 0.00000149, 0.00001189, -0.00000487, -0.00000107, 0.00000248, 0.00000329, 0.00000024, 0.00000006, -0.00000043, 0.00000037, -0.00000104, -0.00000103, 0.00000023, },
		{ -0.00000009, 0.00000069, -0.00000036, -0.00000109, 0.00000567, 0.00000204, -0.00000164, -0.00000523, 0.00001044, 0.00000289, -0.00000321, -0.00000099, 0.00000041, -0.00000010, -0.00000407, -0.00000048, 0.00000143, -0.00000507, -0.00000270, -0.00000263, -0.00000108, -0.00000012, -0.00000069, 0.00000039, -0.00000748, -0.00000098, -0.00000043, 0.00000118, -0.00000454, -0.00000359, -0.00000220, -0.00000455, 0.00000042, -0.00000246, -0.00000550, -0.00000345, -0.00000106, -0.00000141, -0.00000390, -0.00000659, -0.00000714, -0.00000752, -0.00001443, -0.00001626, -0.00003989, -0.00000580, 0.00003912, 0.00002396, 0.00001610, 0.00001592, -0.00000055, 0.00000946, 0.00000231, 0.00000619, 0.00000715, 0.00000034, 0.00000771, 0.00000956, -0.00001056, 0.00000756, 0.00001169, 0.00002052, 0.00002138, -0.00003219, -0.00009575, -0.00000451, 0.00007162, 0.00002135, 0.00001968, 0.00001249, 0.00000700, 0.00001482, 0.00001208, 0.00000607, 0.00002463, 0.00001157, 0.00002742, 0.00002167, 0.00007065, 0.00002579, -0.00007519, -0.00002849, -0.00001452, -0.00001130, -0.00001241, -0.00001005, -0.00000883, -0.00000268, -0.00000485, -0.00000393, -0.00000264, -0.00000297, -0.00000197, -0.00000243, -0.00000477, -0.00000112, -0.00000663, -0.00000076, 0.00000037, -0.00000023, -0.00000884, -0.00000255, -0.00000118, -0.00000030, 0.00000088, -0.00000128, -0.00000410, -0.00000330, -0.00000305, 0.00000216, -0.00000747, 0.00000183, -0.00000374, -0.00000282, 0.00000359, -0.00000332, 0.00001337, -0.00000196, -0.00000312, 0.00000215, 0.00000357, 0.00000080, -0.00000003, -0.00000043, 0.00000044, -0.00000084, -0.00000107, 0.00000009, },
		{ 0.00000029, -0.00000001, 0.00000052, -0.00000194, 0.00000454, 0.00000307, -0.00000008, -0.00000609, 0.00000688, 0.00000708, -0.00000398, -0.00000014, -0.00000111, 0.00000130, -0.00000404, -0.00000165, 0.00000191, -0.00000367, -0.00000470, -0.00000134, -0.00000257, 0.00000106, -0.00000196, 0.00000182, -0.00000632, -0.00000328, -0.00000081, 0.00000172, -0.00000317, -0.00000460, -0.00000185, -0.00000478, -0.00000017, -0.00000147, -0.00000607, -0.00000301, -0.00000239, -0.00000014, -0.00000414, -0.00000557, -0.00000772, -0.00000665, -0.00001539, -0.00001196, -0.00003730, -0.00002231, 0.00003525, 0.00003094, 0.00001700, 0.00001784, 0.00000071, 0.00000932, 0.00000425, 0.00000552, 0.00000820, 0.00000006, 0.00000836, 0.00000955, -0.00000826, 0.00000650, 0.00001162, 0.00002206, 0.00002501, -0.00003148, -0.00009580, -0.00000498, 0.00006656, 0.00002190, 0.00001846, 0.00001190, 0.00000625, 0.00001429, 0.00001068, 0.00000597, 0.00002248, 0.00001115, 0.00002348, 0.00002034, 0.00005306, 0.00005565, -0.00006282, -0.00004203, -0.00001260, -0.00001393, -0.00001071, -0.00001184, -0.00000874, -0.00000348, -0.00000415, -0.00000422, -0.00000202, -0.00000373, -0.00000202, -0.00000208, -0.00000512, -0.00000056, -0.00000641, -0.00000234, 0.00000091, 0.00000014, -0.00000649, -0.00000590, 0.00000013, -0.00000109, 0.00000099, -0.00000073, -0.00000323, -0.00000386, -0.00000351, 0.00000134, -0.00000473, -0.00000203, 0.00000072, -0.00000668, 0.00000560, -0.00000510, 0.00001123, 0.00000289, -0.00000518, 0.00000170, 0.00000337, 0.00000175, -0.00000025, -0.00000030, 0.00000037, -0.00000058, -0.00000106, -0.00000012, },
		{ 0.00000046, -0.00000030, 0.00000066, -0.00000165, 0.00000259, 0.00000414, 0.00000085, -0.00000505, 0.00000210, 0.00001046, -0.00000359, 0.00000014, -0.00000177, 0.00000137, -0.00000283, -0.00000311, 0.00000190, -0.00000195, -0.00000618, -0.00000092, -0.00000318, 0.00000122, -0.00000199, 0.00000151, -0.00000369, -0.00000566, -0.00000135, 0.00000165, -0.00000143, -0.00000527, -0.00000195, -0.00000459, -0.00000108, -0.00000053, -0.00000635, -0.00000278, -0.00000331, 0.00000028, -0.00000362, -0.00000530, -0.00000739, -0.00000672, -0.00001514, -0.00001078, -0.00003028, -0.00003639, 0.00002514, 0.00003817, 0.00001924, 0.00001918, 0.00000277, 0.00000883, 0.00000617, 0.00000533, 0.00000903, -0.00000008, 0.00000880, 0.00000984, -0.00000664, 0.00000674, 0.00001046, 0.00002388, 0.00002845, -0.00002992, -0.00009671, -0.00000561, 0.00006250, 0.00002161, 0.00001774, 0.00001121, 0.00000516, 0.00001444, 0.00000892, 0.00000611, 0.00002008, 0.00001163, 0.00001875, 0.00002144, 0.00003608, 0.00007110, -0.00003732, -0.00005625, -0.00001215, -0.00001628, -0.00000929, -0.00001325, -0.00000879, -0.00000443, -0.00000342, -0.00000466, -0.00000120, -0.00000441, -0.00000211, -0.00000204, -0.00000489, -0.00000089, -0.00000505, -0.00000460, 0.00000173, -0.00000038, -0.00000334, -0.00000872, 0.00000036, -0.00000107, 0.00000066, -0.00000035, -0.00000216, -0.00000431, -0.00000352, -0.00000042, -0.00000157, -0.00000521, 0.00000322, -0.00000762, 0.00000456, -0.00000366, 0.00000640, 0.00000812, -0.00000613, 0.00000078, 0.00000291, 0.00000275, -0.00000029, -0.00000024, 0.00000029, -0.00000035, -0.00000101, -0.00000026, },
		{ 0.00000040, -0.00000020, 0.00000023, -0.00000071, 0.00000057, 0.00000473, 0.00000149, -0.00000304, -0.00000221, 0.00001154, -0.00000124, -0.00000073, -0.00000107, 0.00000002, -0.00000082, -0.00000445, 0.00000140, -0.00000040, -0.00000662, -0.00000160, -0.00000284, 0.00000034, -0.00000091, -0.00000002, -0.00000058, -0.00000727, -0.00000243, 0.00000131, 0.00000010, -0.00000521, -0.00000255, -0.00000412, -0.00000198, -0.00000000, -0.00000609, -0.00000297, -0.00000360, -0.00000020, -0.00000243, -0.00000565, -0.00000633, -0.00000748, -0.00001392, -0.00001170, -0.00002202, -0.00004452, 0.00000880, 0.00004374, 0.00002345, 0.00002001, 0.00000532, 0.00000846, 0.00000765, 0.00000578, 0.00000963, -0.00000001, 0.00000888, 0.00001060, -0.00000577, 0.00000799, 0.00000893, 0.00002522, 0.00003227, -0.00002818, -0.00009749, -0.00000734, 0.00005990, 0.00002045, 0.00001743, 0.00001048, 0.00000391, 0.00001477, 0.00000756, 0.00000575, 0.00001816, 0.00001193, 0.00001482, 0.00002199, 0.00002410, 0.00007206, -0.00000521, -0.00006481, -0.00001604, -0.00001686, -0.00000929, -0.00001346, -0.00000958, -0.00000493, -0.00000323, -0.00000476, -0.00000071, -0.00000460, -0.00000245, -0.00000212, -0.00000437, -0.00000164, -0.00000336, -0.00000656, 0.00000197, -0.00000085, -0.00000067, -0.00000983, -0.00000111, -0.00000003, -0.00000022, 0.00000011, -0.00000137, -0.00000423, -0.00000357, -0.00000217, 0.00000061, -0.00000626, 0.00000284, -0.00000560, 0.00000121, -0.00000039, 0.00000094, 0.00001173, -0.00000499, -0.00000087, 0.00000245, 0.00000338, 0.00000017, -0.00000041, 0.00000035, -0.00000025, -0.00000089, -0.00000031, },
		{ 0.00000022, 0.00000005, -0.00000036, 0.00000022, -0.00000082, 0.00000441, 0.00000222, -0.00000108, -0.00000470, 0.00000968, 0.00000285, -0.00000237, 0.00000051, -0.00000188, 0.00000097, -0.00000494, 0.00000027, 0.00000080, -0.00000609, -0.00000295, -0.00000209, -0.00000099, 0.00000053, -0.00000171, 0.00000178, -0.00000729, -0.00000433, 0.00000105, 0.00000095, -0.00000424, -0.00000354, -0.00000362, -0.00000264, 0.00000002, -0.00000540, -0.00000339, -0.00000351, -0.00000115, -0.00000113, -0.00000607, -0.00000519, -0.00000817, -0.00001271, -0.00001295, -0.00001560, -0.00004504, -0.00001144, 0.00004482, 0.00003019, 0.00002055, 0.00000811, 0.00000850, 0.00000863, 0.00000666, 0.00001024, 0.00000019, 0.00000867, 0.00001170, -0.00000534, 0.00000955, 0.00000795, 0.00002539, 0.00003687, -0.00002671, -0.00009756, -0.00001042, 0.00005845, 0.00001895, 0.00001707, 0.00000993, 0.00000251, 0.00001496, 0.00000692, 0.00000471, 0.00001692, 0.00001158, 0.00001241, 0.00002073, 0.00001795, 0.00006282, 0.00002516, -0.00006271, -0.00002506, -0.00001567, -0.00001081, -0.00001247, -0.00001111, -0.00000487, -0.00000370, -0.00000443, -0.00000067, -0.00000424, -0.00000303, -0.00000225, -0.00000384, -0.00000228, -0.00000209, -0.00000746, 0.00000113, -0.00000074, 0.00000077, -0.00000883, -0.00000405, 0.00000156, -0.00000127, 0.00000057, -0.00000095, -0.00000364, -0.00000373, -0.00000342, 0.00000121, -0.00000507, 0.00000028, -0.00000210, -0.00000253, 0.00000256, -0.00000296, 0.00001227, -0.00000154, -0.00000298, 0.00000204, 0.00000340, 0.00000120, -0.00000082, 0.00000054, -0.00000027, -0.00000074, -0.00000030, },
		{ 0.00000004, 0.00000022, -0.00000076, 0.00000069, -0.00000123, 0.00000318, 0.00000313, 0.00000028, -0.00000497, 0.00000552, 0.00000737, -0.00000356, 0.00000177, -0.00000301, 0.00000142, -0.00000404, -0.00000153, 0.00000166, -0.00000499, -0.00000426, -0.00000162, -0.00000205, 0.00000148, -0.00000259, 0.00000257, -0.00000551, -0.00000684, 0.00000086, 0.00000107, -0.00000262, -0.00000450, -0.00000338, -0.00000293, -0.00000037, -0.00000457, -0.00000370, -0.00000341, -0.00000202, -0.00000031, -0.00000602, -0.00000456, -0.00000811, -0.00001231, -0.00001318, -0.00001257, -0.00003896, -0.00003122, 0.00003851, 0.00003932, 0.00002131, 0.00001089, 0.00000905, 0.00000929, 0.00000764, 0.00001113, 0.00000041, 0.00000832, 0.00001284, -0.00000485, 0.00001072, 0.00000817, 0.00002416, 0.00004217, -0.00002549, -0.00009697, -0.00001442, 0.00005730, 0.00001795, 0.00001616, 0.00000981, 0.00000094, 0.00001493, 0.00000692, 0.00000330, 0.00001603, 0.00001077, 0.00001125, 0.00001800, 0.00001591, 0.00004941, 0.00004697, -0.00004866, -0.00003699, -0.00001443, -0.00001281, -0.00001111, -0.00001273, -0.00000468, -0.00000445, -0.00000402, -0.00000081, -0.00000362, -0.00000359, -0.00000247, -0.00000350, -0.00000249, -0.00000161, -0.00000711, -0.00000068, -0.00000014, 0.00000104, -0.00000630, -0.00000740, 0.00000271, -0.00000181, 0.00000063, -0.00000059, -0.00000293, -0.00000374, -0.00000421, 0.00000051, -0.00000258, -0.00000280, 0.00000083, -0.00000471, 0.00000356, -0.00000407, 0.00000958, 0.00000340, -0.00000477, 0.00000146, 0.00000290, 0.00000255, -0.00000115, 0.00000066, -0.00000028, -0.00000061, -0.00000023, },
		{ -0.00000006, 0.00000022, -0.00000083, 0.00000061, -0.00000082, 0.00000146, 0.00000393, 0.00000105, -0.00000355, 0.00000061, 0.00001049, -0.00000293, 0.00000159, -0.00000243, 0.00000009, -0.00000188, -0.00000356, 0.00000201, -0.00000369, -0.00000500, -0.00000183, -0.00000250, 0.00000154, -0.00000229, 0.00000176, -0.00000250, -0.00000922, 0.00000037, 0.00000078, -0.00000093, -0.00000491, -0.00000367, -0.00000281, -0.00000100, -0.00000389, -0.00000366, -0.00000354, -0.00000253, -0.00000018, -0.00000539, -0.00000459, -0.00000719, -0.00001281, -0.00001224, -0.00001250, -0.00002955, -0.00004561, 0.00002330, 0.00004924, 0.00002336, 0.00001330, 0.00001021, 0.00000983, 0.00000846, 0.00001237, 0.00000066, 0.00000797, 0.00001376, -0.00000393, 0.00001110, 0.00000970, 0.00002195, 0.00004754, -0.00002403, -0.00009621, -0.00001847, 0.00005542, 0.00001813, 0.00001440, 0.00001018, -0.00000079, 0.00001474, 0.00000724, 0.00000200, 0.00001505, 0.00000989, 0.00001073, 0.00001483, 0.00001551, 0.00003695, 0.00005712, -0.00002563, -0.00004729, -0.00001560, -0.00001394, -0.00001045, -0.00001365, -0.00000495, -0.00000493, -0.00000399, -0.00000079, -0.00000309, -0.00000387, -0.00000283, -0.00000337, -0.00000224, -0.00000182, -0.00000594, -0.00000276, 0.00000037, 0.00000076, -0.00000333, -0.00000983, 0.00000247, -0.00000133, 0.00000000, 0.00000003, -0.00000254, -0.00000334, -0.00000475, -0.00000084, -0.00000008, -0.00000479, 0.00000176, -0.00000448, 0.00000228, -0.00000268, 0.00000486, 0.00000812, -0.00000514, 0.00000033, 0.00000220, 0.00000370, -0.00000100, 0.00000048, -0.00000012, -0.00000060, -0.00000010, },
		},
		{ { -0.00000087, -0.00000094, -0.00000226, -0.00000137, -0.00000166, -0.00000322, -0.00000912, -0.00000400, 0.00000046, 0.00000223, -0.00001423, -0.00000224, -0.00000230, -0.00000192, -0.00000286, -0.00000165, 0.00000281, -0.00000654, -0.00000326, 0.00000134, 0.00000009, -0.00000045, -0.00001128, -0.00000255, -0.00000557, -0.00000833, -0.00000011, -0.00000945, -0.00001554, -0.00002598, -0.00005141, 0.00000497, 0.00005336, 0.00002867, 0.00001802, 0.00001470, 0.00001284, 0.00001123, 0.00002133, 0.00000739, 0.00001101, 0.00001679, 0.00001703, 0.00001365, 0.00001721, 0.00002562, 0.00006434, 0.00000102, -0.00005314, -0.00002081, -0.00001461, -0.00000829, -0.00000626, -0.00000403, -0.00000077, -0.00000450, -0.00000455, -0.00000433, -0.00000015, -0.00000280, -0.00000169, -0.00000141, -0.00000527, -0.00000203, -0.00000249, -0.00000203, -0.00000527, -0.00000141, -0.00000169, -0.00000280, -0.00000015, -0.00000433, -0.00000455, -0.00000450, -0.00000077, -0.00000403, -0.00000626, -0.00000829, -0.00001461, -0.00002081, -0.00005314, 0.00000102, 0.00006434, 0.00002562, 0.00001721, 0.00001365, 0.00001703, 0.00001679, 0.00001101, 0.00000739, 0.00002133, 0.00001123, 0.00001284, 0.00001470, 0.00001802, 0.00002867, 0.00005336, 0.00000497, -0.00005141, -0.00002598, -0.00001554, -0.00000945, -0.00000011, -0.00000833, -0.00000557, -0.00000255, -0.00001128, -0.00000045, 0.00000009, 0.00000134, -0.00000326, -0.00000654, 0.00000281, -0.00000165, -0.00000286, -0.00000192, -0.00000230, -0.00000224, -0.00001423, 0.00000223, 0.00000046, -0.00000400, -0.00000912, -0.00000322, -0.00000166, -0.00000137, -0.00000226, -0.00000094, },
		{ -0.00000059, -0.00000086, -0.00000184, -0.00000202, -0.00000148, -0.00000176, -0.00000799, -0.00000633, -0.00000127, 0.00000408, -0.00001086, -0.00000722, 0.00000035, -0.00000425, 0.00000041, -0.00000594, 0.00000479, -0.00000432, -0.00000549, 0.00000053, 0.00000012, 0.00000102, -0.00000708, -0.00000900, -0.00000069, -0.00001016, -0.00000060, -0.00000656, -0.00001363, -0.00001784, -0.00004092, -0.00003941, 0.00004405, 0.00004390, 0.00002221, 0.00001647, 0.00001585, 0.00000929, 0.00001870, 0.00001607, 0.00000725, 0.00001526, 0.00001950, 0.00001428, 0.00001825, 0.00001878, 0.00005919, 0.00003163, -0.00005471, -0.00002957, -0.00001545, -0.00001041, -0.00000621, -0.00000576, -0.00000138, -0.00000227, -0.00000494, -0.00000511, -0.00000178, -0.00000090, -0.00000299, -0.00000024, -0.00000600, -0.00000198, -0.00000221, -0.00000221, -0.00000501, -0.00000218, -0.00000064, -0.00000366, -0.00000047, -0.00000262, -0.00000424, -0.00000496, -0.00000265, -0.00000165, -0.00000621, -0.00000652, -0.00001375, -0.00001549, -0.00004505, -0.00002320, 0.00005531, 0.00003825, 0.00001515, 0.00001506, 0.00001371, 0.00001680, 0.00001406, 0.00000636, 0.00001363, 0.00001907, 0.00000829, 0.00001544, 0.00001332, 0.00002170, 0.00004058, 0.00004078, -0.00003145, -0.00004318, -0.00001687, -0.00001363, -0.00000097, -0.00000495, -0.00000971, -0.00000083, -0.00000891, -0.00000552, 0.00000075, 0.00000100, -0.00000080, -0.00000673, -0.00000054, 0.00000107, -0.00000327, -0.00000170, -0.00000341, 0.00000023, -0.00001362, -0.00000250, 0.00000274, -0.00000289, -0.00000767, -0.00000599, -0.00000129, -0.00000193, -0.00000132, -0.00000189, },
		{ -0.00000075, -0.00000093, -0.00000111, -0.00000256, -0.00000161, -0.00000118, -0.00000563, -0.00000810, -0.00000315, 0.00000361, -0.00000538, -0.00001216, 0.00000223, -0.00000617, 0.00000263, -0.00000691, 0.00000211, 0.00000030, -0.00000695, -0.00000123, 0.00000076, -0.00000001, -0.00000134, -0.00001234, -0.00000147, -0.00000752, -0.00000311, -0.00000410, -0.00001125, -0.00001615, -0.00002372, -0.00005540, 0.00000000, 0.00006056, 0.00002800, 0.00002187, 0.00001557, 0.00001401, 0.00001089, 0.00002292, 0.00000938, 0.00001144, 0.00001980, 0.00001744, 0.00001725, 0.00001845, 0.00004532, 0.00005781, -0.00004301, -0.00004272, -0.00001594, -0.00001344, -0.00000611, -0.00000664, -0.00000333, -0.00000077, -0.00000380, -0.00000563, -0.00000387, 0.00000009, -0.00000324, 0.00000018, -0.00000561, -0.00000345, -0.00000100, -0.00000244, -0.00000472, -0.00000341, 0.00000050, -0.00000364, -0.00000170, -0.00000134, -0.00000360, -0.00000436, -0.00000468, -0.00000102, -0.00000492, -0.00000558, -0.00001228, -0.00001307, -0.00003487, -0.00003765, 0.00003664, 0.00004877, 0.00001711, 0.00001512, 0.00001234, 0.00001522, 0.00001462, 0.00001065, 0.00000528, 0.00002059, 0.00001134, 0.00001150, 0.00001373, 0.00001513, 0.00002911, 0.00004676, 0.00000985, -0.00004970, -0.00002649, -0.00001529, -0.00000493, -0.00000134, -0.00001067, -0.00000378, -0.00000427, -0.00000849, -0.00000141, 0.00000141, 0.00000050, -0.00000528, -0.00000317, 0.00000113, -0.00000194, -0.00000184, -0.00000406, 0.00000090, -0.00001051, -0.00000786, 0.00000401, -0.00000211, -0.00000533, -0.00000810, -0.00000156, -0.00000266, -0.00000056, -0.00000224, },
		{ -0.00000184, -0.00000046, -0.00000118, -0.00000191, -0.00000260, -0.00000071, -0.00000380, -0.00000778, -0.00000564, 0.00000204, -0.00000057, -0.00001401, 0.00000069, -0.00000480, 0.00000050, -0.00000251, -0.00000434, 0.00000495, -0.00000695, -0.00000256, -0.00000011, 0.00000029, 0.00000004, -0.00000822, -0.00000796, -0.00000268, -0.00000634, -0.00000091, -0.00001124, -0.00001275, -0.00001802, -0.00003920, -0.00004900, 0.00004846, 0.00004660, 0.00002384, 0.00001955, 0.00001626, 0.00000984, 0.00001907, 0.00001871, 0.00000781, 0.00001781, 0.00002046, 0.00001717, 0.00001987, 0.00003251, 0.00006985, -0.00001646, -0.00005719, -0.00001746, -0.00001711, -0.00000607, -0.00000753, -0.00000429, -0.00000220, -0.00000077, -0.00000610, -0.00000503, -0.00000111, -0.00000185, -0.00000087, -0.00000326, -0.00000648, 0.00000050, -0.00000261, -0.00000371, -0.00000556, 0.00000157, -0.00000301, -0.00000263, -0.00000143, -0.00000233, -0.00000385, -0.00000499, -0.00000264, -0.00000301, -0.00000507, -0.00001009, -0.00001274, -0.00002502, -0.00004360, 0.00001559, 0.00005148, 0.00002386, 0.00001411, 0.00001261, 0.00001316, 0.00001374, 0.00001378, 0.00000476, 0.00001229, 0.00001931, 0.00000708, 0.00001484, 0.00001081, 0.00002161, 0.00003683, 0.00003907, -0.00002674, -0.00004383, -0.00001643, -0.00001064, 0.00000084, -0.00000929, -0.00000687, -0.00000328, -0.00000620, -0.00000598, 0.00000147, 0.00000126, -0.00000339, -0.00000451, 0.00000024, -0.00000159, -0.00000066, -0.00000480, 0.00000050, -0.00000666, -0.00001169, 0.00000293, -0.00000053, -0.00000399, -0.00000801, -0.00000326, -0.00000242, -0.00000134, -0.00000109, },
		{ -0.00000218, -0.00000066, -0.00000107, -0.00000137, -0.00000292, -0.00000127, -0.00000220, -0.00000659, -0.00000701, -0.00000096, 0.00000287, -0.00001281, -0.00000329, -0.00000153, -0.00000368, 0.00000283, -0.00000855, 0.00000508, -0.00000401, -0.00000387, -0.00000149, 0.00000097, -0.00000085, -0.00000275, -0.00001133, -0.00000298, -0.00000590, -0.00000003, -0.00001007, -0.00001049, -0.00001729, -0.00002027, -0.00006210, -0.00000351, 0.00006663, 0.00002885, 0.00002539, 0.00001698, 0.00001422, 0.00001152, 0.00002451, 0.00001136, 0.00001271, 0.00002215, 0.00001842, 0.00002124, 0.00002459, 0.00006828, 0.00001661, -0.00006437, -0.00002409, -0.00001893, -0.00000804, -0.00000755, -0.00000484, -0.00000455, 0.00000088, -0.00000444, -0.00000614, -0.00000293, -0.00000087, -0.00000175, -0.00000086, -0.00000837, -0.00000012, -0.00000147, -0.00000296, -0.00000685, 0.00000064, -0.00000082, -0.00000370, -0.00000175, -0.00000189, -0.00000277, -0.00000467, -0.00000395, -0.00000290, -0.00000373, -0.00000818, -0.00001262, -0.00001785, -0.00004232, -0.00000425, 0.00004743, 0.00003076, 0.00001532, 0.00001229, 0.00001234, 0.00001216, 0.00001383, 0.00000938, 0.00000404, 0.00002020, 0.00001094, 0.00001026, 0.00001264, 0.00001361, 0.00002811, 0.00004250, 0.00001206, -0.00004779, -0.00002672, -0.00001305, -0.00000155, -0.00000568, -0.00000884, -0.00000486, -0.00000352, -0.00000706, -0.00000226, 0.00000319, -0.00000249, -0.00000431, -0.00000082, -0.00000190, -0.00000017, -0.00000360, -0.00000183, -0.00000232, -0.00001379, 0.00000024, 0.00000111, -0.00000317, -0.00000693, -0.00000487, -0.00000223, -0.00000251, -0.00000013, },
		{ -0.00000112, -0.00000173, -0.00000044, -0.00000162, -0.00000202, -0.00000271, -0.00000087, -0.00000557, -0.00000649, -0.00000465, 0.00000433, -0.00000952, -0.00000767, 0.00000105, -0.00000637, 0.00000463, -0.00000709, 0.00000037, 0.00000040, -0.00000499, -0.00000172, -0.00000026, -0.00000002, -0.00000183, -0.00000744, -0.00000846, -0.00000338, -0.00000129, -0.00000592, -0.00001224, -0.00001327, -0.00001591, -0.00004061, -0.00005707, 0.00005135, 0.00005090, 0.00002518, 0.00002350, 0.00001545, 0.00001149, 0.00001916, 0.00002187, 0.00000827, 0.00002132, 0.00001976, 0.00002308, 0.00002081, 0.00005914, 0.00004598, -0.00005816, -0.00003680, -0.00001831, -0.00001271, -0.00000626, -0.00000626, -0.00000508, -0.00000123, -0.00000040, -0.00000717, -0.00000412, -0.00000159, -0.00000154, 0.00000009, -0.00000742, -0.00000346, 0.00000072, -0.00000282, -0.00000618, -0.00000269, 0.00000224, -0.00000443, -0.00000180, -0.00000263, -0.00000126, -0.00000439, -0.00000377, -0.00000442, -0.00000252, -0.00000621, -0.00001205, -0.00001385, -0.00003616, -0.00002058, 0.00003860, 0.00003541, 0.00001823, 0.00001250, 0.00001164, 0.00001148, 0.00001165, 0.00001320, 0.00000311, 0.00001219, 0.00001851, 0.00000647, 0.00001337, 0.00001020, 0.00002004, 0.00003472, 0.00003661, -0.00002373, -0.00004329, -0.00001522, -0.00000640, -0.00000226, -0.00000959, -0.00000576, -0.00000475, -0.00000319, -0.00000781, 0.00000394, -0.00000118, -0.00000390, -0.00000155, -0.00000160, -0.00000188, -0.00000047, -0.00000468, 0.00000084, -0.00001343, -0.00000367, 0.00000228, -0.00000229, -0.00000598, -0.00000551, -0.00000257, -0.00000311, -0.00000056, },
		{ -0.00000042, -0.00000186, -0.00000073, -0.00000135, -0.00000165, -0.00000299, -0.00000123, -0.00000389, -0.00000588, -0.00000635, 0.00000233, -0.00000424, -0.00001192, 0.00000250, -0.00000717, 0.00000333, -0.00000287, -0.00000416, 0.00000216, -0.00000433, -0.00000166, -0.00000192, 0.00000106, -0.00000270, -0.00000295, -0.00001022, -0.00000563, -0.00000048, -0.00000273, -0.00001288, -0.00001065, -0.00001634, -0.00001916, -0.00006819, -0.00000779, 0.00007310, 0.00003035, 0.00002935, 0.00001760, 0.00001500, 0.00001217, 0.00002662, 0.00001285, 0.00001583, 0.00002180, 0.00002343, 0.00002200, 0.00004580, 0.00006772, -0.00004024, -0.00005086, -0.00001842, -0.00001757, -0.00000581, -0.00000749, -0.00000455, -0.00000473, 0.00000226, -0.00000571, -0.00000564, -0.00000262, -0.00000176, 0.00000052, -0.00000506, -0.00000690, 0.00000140, -0.00000190, -0.00000487, -0.00000588, 0.00000321, -0.00000297, -0.00000267, -0.00000293, -0.00000112, -0.00000312, -0.00000361, -0.00000479, -0.00000374, -0.00000322, -0.00001149, -0.00001159, -0.00002883, -0.00003060, 0.00002505, 0.00003920, 0.00001958, 0.00001528, 0.00000987, 0.00001228, 0.00000894, 0.00001389, 0.00000693, 0.00000483, 0.00001852, 0.00001131, 0.00000797, 0.00001308, 0.00001160, 0.00002720, 0.00003941, 0.00001157, -0.00004477, -0.00002666, -0.00000843, -0.00000281, -0.00000815, -0.00000618, -0.00000723, -0.00000050, -0.00000848, -0.00000047, 0.00000234, -0.00000455, -0.00000149, -0.00000097, -0.00000376, 0.00000088, -0.00000439, 0.00000029, -0.00000957, -0.00000889, 0.00000330, -0.00000186, -0.00000466, -0.00000626, -0.00000228, -0.00000385, -0.00000131, },
		{ -0.00000070, -0.00000085, -0.00000179, -0.00000052, -0.00000216, -0.00000190, -0.00000279, -0.00000201, -0.00000581, -0.00000557, -0.00000193, 0.00000101, -0.00001451, 0.00000196, -0.00000597, 0.00000028, 0.00000112, -0.00000582, 0.00000102, -0.00000298, -0.00000150, -0.00000234, 0.00000025, -0.00000173, -0.00000298, -0.00000547, -0.00001128, 0.00000019, -0.00000057, -0.00001039, -0.00001215, -0.00001334, -0.00001396, -0.00004355, -0.00006390, 0.00005254, 0.00005715, 0.00002623, 0.00002757, 0.00001417, 0.00001402, 0.00001912, 0.00002503, 0.00001009, 0.00002307, 0.00002147, 0.00002731, 0.00003188, 0.00007974, -0.00001358, -0.00006237, -0.00002110, -0.00002098, -0.00000758, -0.00000774, -0.00000450, -0.00000659, 0.00000100, -0.00000166, -0.00000706, -0.00000333, -0.00000292, 0.00000066, -0.00000253, -0.00000851, -0.00000038, -0.00000018, -0.00000373, -0.00000711, 0.00000103, 0.00000030, -0.00000390, -0.00000248, -0.00000245, -0.00000120, -0.00000391, -0.00000332, -0.00000650, -0.00000095, -0.00000993, -0.00001062, -0.00002222, -0.00003411, 0.00000874, 0.00004120, 0.00002041, 0.00001770, 0.00001002, 0.00001178, 0.00000857, 0.00001132, 0.00001080, 0.00000357, 0.00001125, 0.00001804, 0.00000531, 0.00001244, 0.00001035, 0.00001738, 0.00003450, 0.00003200, -0.00002073, -0.00004218, -0.00001180, -0.00000483, -0.00000673, -0.00000644, -0.00000769, -0.00000276, -0.00000350, -0.00000726, 0.00000484, -0.00000420, -0.00000190, -0.00000026, -0.00000383, -0.00000131, -0.00000100, -0.00000244, -0.00000455, -0.00001314, 0.00000284, -0.00000116, -0.00000333, -0.00000717, -0.00000163, -0.00000429, -0.00000208, },
		{ -0.00000067, -0.00000066, -0.00000158, -0.00000088, -0.00000186, -0.00000162, -0.00000302, -0.00000196, -0.00000479, -0.00000486, -0.00000457, 0.00000261, -0.00001282, -0.00000243, -0.00000165, -0.00000478, 0.00000508, -0.00000660, 0.00000073, -0.00000385, -0.00000030, -0.00000232, -0.00000070, -0.00000088, -0.00000394, -0.00000182, -0.00001100, -0.00000570, 0.00000401, -0.00000856, -0.00001213, -0.00001245, -0.00001307, -0.00002175, -0.00007131, -0.00001503, 0.00008175, 0.00003144, 0.00003421, 0.00001675, 0.00001761, 0.00001180, 0.00002975, 0.00001430, 0.00001943, 0.00002065, 0.00003069, 0.00002472, 0.00007762, 0.00002158, -0.00007065, -0.00002560, -0.00002352, -0.00001048, -0.00000827, -0.00000451, -0.00000708, -0.00000200, 0.00000114, -0.00000574, -0.00000498, -0.00000336, -0.00000095, 0.00000051, -0.00000880, -0.00000281, 0.00000042, -0.00000202, -0.00000705, -0.00000213, 0.00000237, -0.00000334, -0.00000270, -0.00000335, -0.00000065, -0.00000330, -0.00000207, -0.00000741, -0.00000210, -0.00000622, -0.00001111, -0.00001632, -0.00003351, -0.00000597, 0.00003782, 0.00002459, 0.00001577, 0.00001425, 0.00000887, 0.00001079, 0.00000743, 0.00001254, 0.00000561, 0.00000599, 0.00001623, 0.00001199, 0.00000515, 0.00001513, 0.00000821, 0.00002776, 0.00003548, 0.00000993, -0.00004097, -0.00002501, -0.00000418, -0.00000797, -0.00000609, -0.00000656, -0.00000676, 0.00000052, -0.00000931, 0.00000191, -0.00000091, -0.00000369, 0.00000057, -0.00000246, -0.00000443, 0.00000106, -0.00000287, -0.00000231, -0.00001345, -0.00000088, 0.00000107, -0.00000319, -0.00000680, -0.00000257, -0.00000277, -0.00000390, },
		{ -0.00000063, -0.00000139, -0.00000035, -0.00000194, -0.00000095, -0.00000234, -0.00000180, -0.00000328, -0.00000315, -0.00000503, -0.00000450, 0.00000064, -0.00000785, -0.00000913, 0.00000369, -0.00000953, 0.00000751, -0.00000643, 0.00000112, -0.00000500, -0.00000041, -0.00000128, -0.00000096, -0.00000138, -0.00000309, -0.00000308, -0.00000409, -0.00001403, 0.00000632, -0.00000572, -0.00001017, -0.00001450, -0.00001007, -0.00001611, -0.00004383, -0.00007277, 0.00005456, 0.00006352, 0.00002861, 0.00002961, 0.00001511, 0.00001511, 0.00002076, 0.00002698, 0.00001441, 0.00002131, 0.00002872, 0.00002717, 0.00006263, 0.00005955, -0.00007030, -0.00003510, -0.00002326, -0.00001504, -0.00000890, -0.00000505, -0.00000666, -0.00000456, 0.00000081, -0.00000216, -0.00000654, -0.00000324, -0.00000365, 0.00000273, -0.00000749, -0.00000508, -0.00000027, -0.00000042, -0.00000590, -0.00000468, 0.00000208, -0.00000122, -0.00000322, -0.00000349, -0.00000157, -0.00000202, -0.00000171, -0.00000571, -0.00000564, -0.00000249, -0.00001122, -0.00001205, -0.00003032, -0.00001673, 0.00002854, 0.00003135, 0.00001200, 0.00001810, 0.00000789, 0.00001152, 0.00000612, 0.00001108, 0.00000775, 0.00000599, 0.00000865, 0.00001877, 0.00000277, 0.00001346, 0.00000938, 0.00001550, 0.00003439, 0.00002592, -0.00001730, -0.00003992, -0.00000716, -0.00000795, -0.00000780, -0.00000432, -0.00000909, -0.00000014, -0.00000591, -0.00000373, 0.00000191, -0.00000438, 0.00000006, -0.00000031, -0.00000576, -0.00000059, -0.00000036, -0.00000243, -0.00001094, -0.00000611, 0.00000331, -0.00000334, -0.00000551, -0.00000493, -0.00000010, -0.00000567, },
		{ -0.00000230, -0.00000118, -0.00000043, -0.00000147, -0.00000131, -0.00000219, -0.00000147, -0.00000330, -0.00000309, -0.00000444, -0.00000421, -0.00000121, -0.00000381, -0.00001351, 0.00000512, -0.00000939, 0.00000476, -0.00000222, -0.00000183, -0.00000171, -0.00000502, 0.00000206, -0.00000210, -0.00000058, -0.00000343, -0.00000413, -0.00000021, -0.00001354, -0.00000133, 0.00000238, -0.00001112, -0.00001244, -0.00001209, -0.00001198, -0.00002411, -0.00007404, -0.00002327, 0.00009097, 0.00003396, 0.00003732, 0.00001743, 0.00001948, 0.00001261, 0.00003177, 0.00001853, 0.00001921, 0.00002511, 0.00003256, 0.00004575, 0.00008653, -0.00005067, -0.00005509, -0.00001759, -0.00002254, -0.00000798, -0.00000804, -0.00000433, -0.00000740, -0.00000011, -0.00000054, -0.00000488, -0.00000461, -0.00000466, 0.00000149, -0.00000342, -0.00000795, -0.00000049, -0.00000077, -0.00000311, -0.00000681, 0.00000098, -0.00000001, -0.00000226, -0.00000403, -0.00000229, -0.00000176, -0.00000115, -0.00000338, -0.00000800, -0.00000184, -0.00000914, -0.00001028, -0.00002509, -0.00002382, 0.00001661, 0.00003568, 0.00001237, 0.00001624, 0.00001187, 0.00000904, 0.00000809, 0.00000773, 0.00000889, 0.00000738, 0.00000497, 0.00001550, 0.00001104, 0.00000405, 0.00001660, 0.00000501, 0.00002885, 0.00003033, 0.00000818, -0.00003625, -0.00002151, -0.00000368, -0.00001247, -0.00000213, -0.00000909, -0.00000310, -0.00000265, -0.00000593, 0.00000044, -0.00000238, -0.00000200, 0.00000164, -0.00000480, -0.00000394, 0.00000134, -0.00000137, -0.00000885, -0.00000965, 0.00000303, -0.00000193, -0.00000491, -0.00000660, 0.00000084, -0.00000487, },
		{ -0.00000469, -0.00000050, -0.00000159, -0.00000004, -0.00000221, -0.00000154, -0.00000206, -0.00000209, -0.00000407, -0.00000344, -0.00000440, -0.00000172, -0.00000182, -0.00001424, 0.00000157, -0.00000427, -0.00000222, 0.00000479, -0.00000714, 0.00000411, -0.00001028, 0.00000395, -0.00000256, 0.00000135, -0.00000518, -0.00000298, -0.00000237, -0.00000484, -0.00001285, 0.00000883, -0.00001122, -0.00000807, -0.00001729, -0.00000586, -0.00002103, -0.00004151, -0.00008378, 0.00005731, 0.00007068, 0.00003081, 0.00003119, 0.00001728, 0.00001560, 0.00002296, 0.00002992, 0.00001737, 0.00002164, 0.00003490, 0.00003584, 0.00009403, -0.00001018, -0.00008047, -0.00001284, -0.00002829, -0.00000791, -0.00001179, -0.00000187, -0.00000973, -0.00000060, -0.00000178, -0.00000110, -0.00000619, -0.00000419, -0.00000227, 0.00000129, -0.00000989, -0.00000077, -0.00000213, -0.00000048, -0.00000748, -0.00000055, 0.00000003, -0.00000067, -0.00000410, -0.00000295, -0.00000216, -0.00000093, -0.00000109, -0.00000814, -0.00000364, -0.00000652, -0.00000937, -0.00001967, -0.00002715, 0.00000460, 0.00003487, 0.00001784, 0.00001066, 0.00001641, 0.00000773, 0.00000922, 0.00000621, 0.00000811, 0.00000758, 0.00000694, 0.00000690, 0.00001859, 0.00000092, 0.00001494, 0.00000780, 0.00001458, 0.00003323, 0.00001942, -0.00001313, -0.00003589, -0.00000436, -0.00001340, -0.00000467, -0.00000583, -0.00000634, -0.00000181, -0.00000498, -0.00000219, -0.00000143, -0.00000251, 0.00000121, -0.00000167, -0.00000664, 0.00000034, 0.00000099, -0.00000689, -0.00001169, 0.00000089, 0.00000012, -0.00000452, -0.00000725, -0.00000047, -0.00000192, },
		{ -0.00000510, -0.00000134, -0.00000171, -0.00000017, -0.00000137, -0.00000223, -0.00000177, -0.00000187, -0.00000364, -0.00000398, -0.00000362, -0.00000273, 0.00000021, -0.00001345, -0.00000381, 0.00000116, -0.00000782, 0.00000876, -0.00000874, 0.00000532, -0.00000891, -0.00000059, 0.00000052, 0.00000104, -0.00000400, -0.00000414, -0.00000376, 0.00000024, -0.00001446, 0.00000343, -0.00000371, -0.00000903, -0.00001548, -0.00000825, -0.00001557, -0.00002302, -0.00007905, -0.00003049, 0.00009939, 0.00003873, 0.00003753, 0.00002153, 0.00001873, 0.00001609, 0.00003225, 0.00002501, 0.00001616, 0.00003474, 0.00003200, 0.00008660, 0.00003796, -0.00009411, -0.00002189, -0.00002485, -0.00001327, -0.00001204, -0.00000344, -0.00000849, -0.00000313, -0.00000206, -0.00000023, -0.00000430, -0.00000493, -0.00000474, 0.00000251, -0.00000809, -0.00000324, -0.00000170, -0.00000079, -0.00000521, -0.00000336, 0.00000067, -0.00000087, -0.00000207, -0.00000454, -0.00000182, -0.00000227, 0.00000147, -0.00000713, -0.00000527, -0.00000569, -0.00000783, -0.00001542, -0.00002697, -0.00000610, 0.00002952, 0.00002470, 0.00000723, 0.00001613, 0.00001053, 0.00000797, 0.00000697, 0.00000659, 0.00000630, 0.00000972, 0.00000307, 0.00001547, 0.00000912, 0.00000475, 0.00001661, 0.00000316, 0.00002906, 0.00002448, 0.00000711, -0.00003023, -0.00001793, -0.00000745, -0.00001239, -0.00000112, -0.00000876, -0.00000162, -0.00000533, -0.00000131, -0.00000415, -0.00000059, -0.00000126, 0.00000190, -0.00000738, -0.00000208, 0.00000183, -0.00000300, -0.00001356, -0.00000149, 0.00000115, -0.00000319, -0.00000772, -0.00000250, 0.00000046, },
		{ -0.00000372, -0.00000286, -0.00000128, -0.00000135, 0.00000016, -0.00000301, -0.00000145, -0.00000218, -0.00000243, -0.00000496, -0.00000286, -0.00000381, 0.00000178, -0.00001090, -0.00000969, 0.00000444, -0.00000955, 0.00000765, -0.00000497, 0.00000071, -0.00000170, -0.00000811, 0.00000374, -0.00000037, 0.00000002, -0.00000761, -0.00000247, -0.00000147, -0.00000664, -0.00000775, 0.00000436, -0.00001128, -0.00000916, -0.00001601, -0.00000611, -0.00002344, -0.00004071, -0.00009352, 0.00005789, 0.00008145, 0.00002929, 0.00003660, 0.00001651, 0.00001929, 0.00002252, 0.00003715, 0.00001460, 0.00003064, 0.00003239, 0.00007131, 0.00007952, -0.00008385, -0.00004751, -0.00001489, -0.00002136, -0.00000975, -0.00000851, -0.00000460, -0.00000740, -0.00000031, -0.00000296, -0.00000018, -0.00000590, -0.00000542, 0.00000019, -0.00000379, -0.00000675, -0.00000004, -0.00000297, -0.00000197, -0.00000587, 0.00000141, -0.00000213, 0.00000030, -0.00000533, -0.00000176, -0.00000372, 0.00000255, -0.00000450, -0.00000656, -0.00000554, -0.00000733, -0.00001111, -0.00002539, -0.00001354, 0.00002049, 0.00003015, 0.00000793, 0.00001231, 0.00001352, 0.00000831, 0.00000642, 0.00000718, 0.00000360, 0.00001035, 0.00000520, 0.00000736, 0.00001614, 0.00000166, 0.00001472, 0.00000738, 0.00001336, 0.00003093, 0.00001351, -0.00000794, -0.00003068, -0.00000598, -0.00001607, -0.00000187, -0.00000741, -0.00000238, -0.00000655, 0.00000030, -0.00000567, -0.00000112, -0.00000154, 0.00000241, -0.00000486, -0.00000504, 0.00000087, 0.00000098, -0.00001306, -0.00000540, 0.00000220, -0.00000218, -0.00000703, -0.00000512, 0.00000158, },
		{ -0.00000239, -0.00000295, -0.00000206, -0.00000147, -0.00000015, -0.00000161, -0.00000288, -0.00000122, -0.00000269, -0.00000393, -0.00000415, -0.00000308, 0.00000061, -0.00000505, -0.00001613, 0.00000567, -0.00000875, 0.00000408, 0.00000037, -0.00000438, 0.00000387, -0.00001060, 0.00000155, 0.00000065, 0.00000242, -0.00000785, -0.00000382, -0.00000257, -0.00000174, -0.00001029, 0.00000184, -0.00000662, -0.00000804, -0.00001636, -0.00000630, -0.00001931, -0.00002159, -0.00008372, -0.00003921, 0.00010994, 0.00004181, 0.00003938, 0.00002482, 0.00001930, 0.00001842, 0.00003596, 0.00002692, 0.00001990, 0.00003758, 0.00005155, 0.00010803, -0.00005118, -0.00007800, -0.00001067, -0.00002438, -0.00000989, -0.00001231, -0.00000283, -0.00000952, -0.00000002, -0.00000457, 0.00000077, -0.00000357, -0.00000652, -0.00000232, -0.00000107, -0.00000795, -0.00000010, -0.00000352, -0.00000129, -0.00000559, 0.00000023, -0.00000173, -0.00000007, -0.00000321, -0.00000345, -0.00000333, 0.00000049, 0.00000016, -0.00000818, -0.00000423, -0.00000912, -0.00000645, -0.00002338, -0.00001744, 0.00000966, 0.00003245, 0.00001162, 0.00000885, 0.00001300, 0.00001157, 0.00000465, 0.00000860, 0.00000208, 0.00000791, 0.00000885, 0.00000356, 0.00001323, 0.00000924, 0.00000467, 0.00001676, 0.00000192, 0.00002790, 0.00001869, 0.00000738, -0.00002302, -0.00001726, -0.00001166, -0.00000930, -0.00000318, -0.00000433, -0.00000558, -0.00000234, -0.00000156, -0.00000629, 0.00000117, -0.00000014, -0.00000033, -0.00000763, -0.00000021, 0.00000181, -0.00000811, -0.00001139, 0.00000373, -0.00000228, -0.00000468, -0.00000832, 0.00000195, },
		{ -0.00000067, -0.00000254, -0.00000269, -0.00000149, -0.00000128, 0.00000026, -0.00000410, -0.00000059, -0.00000317, -0.00000238, -0.00000560, -0.00000219, -0.00000171, 0.00000114, -0.00001931, 0.00000246, -0.00000484, -0.00000137, 0.00000571, -0.00000753, 0.00000528, -0.00000768, -0.00000344, 0.00000157, 0.00000340, -0.00000452, -0.00000764, -0.00000164, -0.00000239, -0.00000505, -0.00000525, -0.00000160, -0.00000857, -0.00001161, -0.00001269, -0.00001002, -0.00002293, -0.00004173, -0.00010213, 0.00005716, 0.00009397, 0.00002605, 0.00004452, 0.00001386, 0.00002494, 0.00002202, 0.00004418, 0.00001256, 0.00004070, 0.00003531, 0.00011712, -0.00000123, -0.00010159, -0.00001835, -0.00002223, -0.00001184, -0.00001413, -0.00000409, -0.00000902, -0.00000202, -0.00000345, -0.00000203, 0.00000035, -0.00000662, -0.00000456, -0.00000047, -0.00000715, -0.00000164, -0.00000232, -0.00000270, -0.00000378, -0.00000181, -0.00000010, -0.00000219, -0.00000029, -0.00000483, -0.00000251, -0.00000263, 0.00000397, -0.00000811, -0.00000337, -0.00001031, -0.00000463, -0.00001889, -0.00002042, 0.00000087, 0.00002930, 0.00001804, 0.00000622, 0.00001130, 0.00001293, 0.00000651, 0.00000650, 0.00000488, 0.00000248, 0.00001141, 0.00000431, 0.00000718, 0.00001397, 0.00000326, 0.00001346, 0.00000812, 0.00001118, 0.00002764, 0.00000951, -0.00000198, -0.00002641, -0.00001098, -0.00001369, -0.00000329, -0.00000427, -0.00000357, -0.00000719, 0.00000346, -0.00000967, 0.00000104, -0.00000088, 0.00000199, -0.00000751, -0.00000175, 0.00000053, -0.00000221, -0.00001505, 0.00000229, -0.00000094, -0.00000323, -0.00000959, 0.00000019, },
		{ 0.00000209, -0.00000331, -0.00000128, -0.00000305, -0.00000102, -0.00000017, -0.00000251, -0.00000243, -0.00000169, -0.00000301, -0.00000433, -0.00000378, -0.00000200, 0.00000308, -0.00001517, -0.00000687, 0.00000249, -0.00000876, 0.00001112, -0.00000939, 0.00000495, -0.00000459, -0.00000494, -0.00000166, 0.00000535, -0.00000162, -0.00000836, -0.00000401, -0.00000146, -0.00000358, -0.00000390, -0.00000589, -0.00000350, -0.00001124, -0.00001218, -0.00001028, -0.00001865, -0.00002353, -0.00008513, -0.00005157, 0.00012418, 0.00004111, 0.00004596, 0.00002397, 0.00002426, 0.00001762, 0.00004399, 0.00002325, 0.00003277, 0.00003236, 0.00010161, 0.00005982, -0.00011169, -0.00003494, -0.00002147, -0.00001169, -0.00001601, -0.00000589, -0.00000884, -0.00000371, -0.00000250, -0.00000399, 0.00000065, -0.00000280, -0.00000779, -0.00000002, -0.00000694, -0.00000247, -0.00000162, -0.00000329, -0.00000349, -0.00000236, 0.00000043, -0.00000283, -0.00000009, -0.00000340, -0.00000324, -0.00000398, 0.00000402, -0.00000454, -0.00000489, -0.00000816, -0.00000745, -0.00001165, -0.00002347, -0.00000428, 0.00002097, 0.00002564, 0.00000432, 0.00001101, 0.00001000, 0.00001201, 0.00000244, 0.00000905, -0.00000180, 0.00001080, 0.00000624, 0.00000563, 0.00000926, 0.00001169, 0.00000284, 0.00001804, 0.00000044, 0.00002502, 0.00001455, 0.00000872, -0.00001562, -0.00002114, -0.00001135, -0.00000882, -0.00000269, -0.00000197, -0.00000951, 0.00000293, -0.00000599, -0.00000451, 0.00000140, 0.00000120, -0.00000462, -0.00000421, 0.00000002, -0.00000001, -0.00001270, -0.00000382, 0.00000285, -0.00000395, -0.00000773, -0.00000421, },
		{ 0.00000330, -0.00000349, 0.00000009, -0.00000393, -0.00000097, -0.00000134, -0.00000031, -0.00000427, -0.00000035, -0.00000406, -0.00000232, -0.00000564, -0.00000168, 0.00000182, -0.00000687, -0.00001670, 0.00000753, -0.00001305, 0.00001218, -0.00000689, 0.00000166, -0.00000122, -0.00000462, -0.00000468, 0.00000441, 0.00000198, -0.00000666, -0.00000794, 0.00000028, -0.00000565, 0.00000186, -0.00001079, -0.00000136, -0.00000997, -0.00000909, -0.00001437, -0.00001095, -0.00002476, -0.00004013, -0.00011361, 0.00005926, 0.00010291, 0.00002752, 0.00004844, 0.00001560, 0.00002685, 0.00002756, 0.00004310, 0.00002243, 0.00003788, 0.00007170, 0.00011409, -0.00009714, -0.00006078, -0.00002161, -0.00001350, -0.00001436, -0.00000985, -0.00000796, -0.00000599, -0.00000170, -0.00000431, -0.00000202, 0.00000169, -0.00000876, -0.00000121, -0.00000616, -0.00000395, -0.00000065, -0.00000333, -0.00000393, -0.00000263, 0.00000055, -0.00000217, -0.00000167, -0.00000123, -0.00000368, -0.00000446, 0.00000172, 0.00000011, -0.00000671, -0.00000476, -0.00001120, -0.00000610, -0.00002359, -0.00000850, 0.00001194, 0.00002975, 0.00000602, 0.00000979, 0.00000745, 0.00001429, 0.00000320, 0.00000834, -0.00000006, 0.00000556, 0.00000863, 0.00000652, 0.00000410, 0.00001467, 0.00000320, 0.00001292, 0.00000904, 0.00000759, 0.00002454, 0.00000773, 0.00000428, -0.00002590, -0.00001429, -0.00001037, -0.00000583, 0.00000102, -0.00000920, -0.00000170, 0.00000013, -0.00000901, 0.00000120, 0.00000139, -0.00000230, -0.00000577, 0.00000008, -0.00000068, -0.00000759, -0.00000990, 0.00000480, -0.00000361, -0.00000574, -0.00000801, },
		{ 0.00000110, -0.00000125, -0.00000108, -0.00000178, -0.00000307, -0.00000080, -0.00000058, -0.00000315, -0.00000173, -0.00000286, -0.00000277, -0.00000454, -0.00000355, 0.00000107, -0.00000081, -0.00001963, 0.00000431, -0.00000941, 0.00000493, 0.00000257, -0.00000631, 0.00000494, -0.00000735, -0.00000172, -0.00000283, 0.00000811, -0.00000647, -0.00000736, -0.00000345, -0.00000299, -0.00000026, -0.00000347, -0.00001103, -0.00000106, -0.00001285, -0.00000899, -0.00001527, -0.00001703, -0.00002577, -0.00008656, -0.00006352, 0.00013658, 0.00004289, 0.00005138, 0.00002458, 0.00002812, 0.00002052, 0.00004690, 0.00002702, 0.00003845, 0.00004751, 0.00013897, -0.00004772, -0.00009697, -0.00001826, -0.00002363, -0.00000569, -0.00001767, -0.00000419, -0.00001128, 0.00000103, -0.00000581, -0.00000268, 0.00000094, -0.00000382, -0.00000588, -0.00000273, -0.00000811, 0.00000195, -0.00000438, -0.00000287, -0.00000475, 0.00000179, -0.00000201, -0.00000210, -0.00000151, -0.00000180, -0.00000548, -0.00000045, 0.00000272, -0.00000622, -0.00000320, -0.00001176, -0.00000556, -0.00001933, -0.00001355, 0.00000551, 0.00002761, 0.00001253, 0.00000591, 0.00000875, 0.00001040, 0.00000930, 0.00000386, 0.00000502, -0.00000038, 0.00000974, 0.00000645, 0.00000527, 0.00000733, 0.00001347, 0.00000106, 0.00001961, -0.00000154, 0.00002065, 0.00001344, 0.00001041, -0.00001038, -0.00002672, -0.00000682, -0.00001232, 0.00000283, -0.00000639, -0.00000636, 0.00000254, -0.00000758, -0.00000339, 0.00000302, -0.00000102, -0.00000610, -0.00000031, -0.00000103, -0.00000464, -0.00001138, 0.00000204, -0.00000041, -0.00000560, -0.00000890, },
		{ -0.00000217, 0.00000120, -0.00000285, 0.00000089, -0.00000463, -0.00000055, -0.00000148, -0.00000138, -0.00000330, -0.00000153, -0.00000389, -0.00000258, -0.00000537, 0.00000004, 0.00000224, -0.00001617, -0.00000412, -0.00000196, -0.00000562, 0.00001279, -0.00001278, 0.00000905, -0.00000945, 0.00000221, -0.00000891, 0.00001013, -0.00000433, -0.00000498, -0.00000885, 0.00000121, -0.00000666, 0.00000770, -0.00001935, 0.00000301, -0.00001429, -0.00000280, -0.00002122, -0.00000767, -0.00002972, -0.00003591, -0.00012694, 0.00006202, 0.00011035, 0.00003258, 0.00004908, 0.00002092, 0.00002737, 0.00003473, 0.00004033, 0.00003690, 0.00003523, 0.00013108, 0.00002552, -0.00012841, -0.00001995, -0.00003462, 0.00000057, -0.00002152, -0.00000276, -0.00001522, 0.00000194, -0.00000636, -0.00000209, -0.00000300, 0.00000238, -0.00000925, -0.00000013, -0.00001142, 0.00000261, -0.00000389, -0.00000219, -0.00000646, 0.00000173, -0.00000090, -0.00000213, -0.00000272, -0.00000025, -0.00000487, -0.00000320, 0.00000386, -0.00000455, -0.00000281, -0.00000985, -0.00000769, -0.00001417, -0.00001712, 0.00000060, 0.00002159, 0.00002011, 0.00000262, 0.00001066, 0.00000584, 0.00001322, 0.00000296, 0.00000655, -0.00000147, 0.00000715, 0.00000632, 0.00000845, 0.00000094, 0.00001647, 0.00000215, 0.00001296, 0.00000996, 0.00000190, 0.00002389, 0.00000706, 0.00000972, -0.00002944, -0.00001217, -0.00001131, -0.00000347, 0.00000069, -0.00001056, 0.00000221, -0.00000444, -0.00000618, 0.00000095, 0.00000243, -0.00000719, -0.00000051, -0.00000068, -0.00000369, -0.00001031, -0.00000191, 0.00000207, -0.00000470, -0.00000893, },
		{ -0.00000466, 0.00000181, -0.00000283, 0.00000101, -0.00000290, -0.00000268, -0.00000060, -0.00000187, -0.00000231, -0.00000250, -0.00000307, -0.00000289, -0.00000417, -0.00000314, 0.00000438, -0.00001117, -0.00001079, 0.00000211, -0.00001156, 0.00001531, -0.00001023, 0.00000486, -0.00000464, -0.00000084, -0.00000554, 0.00000251, 0.00000343, -0.00000652, -0.00000785, -0.00000186, -0.00000586, 0.00000701, -0.00001076, -0.00000877, -0.00000430, -0.00000803, -0.00001272, -0.00001511, -0.00001940, -0.00002436, -0.00009119, -0.00007216, 0.00014321, 0.00005240, 0.00004998, 0.00003223, 0.00002641, 0.00003098, 0.00004026, 0.00004511, 0.00002905, 0.00010551, 0.00009452, -0.00013022, -0.00004197, -0.00003265, -0.00000827, -0.00001319, -0.00000940, -0.00001210, -0.00000417, -0.00000186, -0.00000467, -0.00000355, 0.00000230, -0.00000580, -0.00000246, -0.00000951, -0.00000238, 0.00000057, -0.00000426, -0.00000480, -0.00000211, 0.00000262, -0.00000340, -0.00000208, -0.00000188, -0.00000127, -0.00000679, 0.00000444, -0.00000316, -0.00000235, -0.00000746, -0.00000930, -0.00001105, -0.00001799, -0.00000367, 0.00001445, 0.00002500, 0.00000311, 0.00000971, 0.00000526, 0.00001122, 0.00000687, 0.00000482, 0.00000064, 0.00000348, 0.00000587, 0.00000914, 0.00000240, 0.00000828, 0.00001326, 0.00000032, 0.00002102, -0.00000505, 0.00001703, 0.00001460, 0.00001208, -0.00000975, -0.00002832, -0.00000410, -0.00001383, 0.00000676, -0.00001181, 0.00000070, -0.00000344, -0.00000430, -0.00000471, 0.00000621, -0.00000730, -0.00000170, 0.00000028, -0.00000274, -0.00000992, -0.00000384, 0.00000167, -0.00000194, -0.00000924, },
		{ -0.00000800, 0.00000201, -0.00000233, 0.00000001, -0.00000031, -0.00000430, -0.00000032, -0.00000237, -0.00000141, -0.00000303, -0.00000263, -0.00000343, -0.00000258, -0.00000531, 0.00000376, -0.00000517, -0.00001483, 0.00000279, -0.00001274, 0.00001066, -0.00000123, -0.00000370, 0.00000323, -0.00000724, 0.00000187, -0.00000667, 0.00000941, -0.00000768, -0.00000392, -0.00000794, -0.00000130, -0.00000122, 0.00000470, -0.00002114, 0.00000269, -0.00001327, -0.00000154, -0.00002513, -0.00000670, -0.00003207, -0.00003472, -0.00013585, 0.00005753, 0.00012539, 0.00003195, 0.00005565, 0.00002106, 0.00003616, 0.00003119, 0.00005279, 0.00003436, 0.00007218, 0.00014149, -0.00009474, -0.00008163, -0.00002108, -0.00002688, -0.00000089, -0.00001726, -0.00000644, -0.00001301, 0.00000352, -0.00000793, -0.00000161, -0.00000216, 0.00000033, -0.00000572, -0.00000543, -0.00000922, 0.00000468, -0.00000576, -0.00000249, -0.00000631, 0.00000488, -0.00000332, -0.00000139, -0.00000406, 0.00000137, -0.00000755, 0.00000221, -0.00000040, -0.00000343, -0.00000385, -0.00001076, -0.00000853, -0.00001863, -0.00000574, 0.00000662, 0.00002668, 0.00000662, 0.00000711, 0.00000675, 0.00000767, 0.00000935, 0.00000570, 0.00000068, 0.00000276, 0.00000376, 0.00000807, 0.00000666, 0.00000122, 0.00001627, 0.00000207, 0.00001271, 0.00001022, -0.00000476, 0.00002628, 0.00000680, 0.00001163, -0.00003235, -0.00000798, -0.00001426, 0.00000203, -0.00000565, -0.00000399, -0.00000160, -0.00000357, -0.00000625, 0.00000368, -0.00000209, -0.00000630, 0.00000277, -0.00000228, -0.00000893, -0.00000620, 0.00000171, -0.00000084, -0.00000720, },
		{ -0.00001157, 0.00000186, -0.00000210, -0.00000046, 0.00000044, -0.00000286, -0.00000258, -0.00000073, -0.00000304, -0.00000092, -0.00000442, -0.00000207, -0.00000345, -0.00000369, -0.00000085, 0.00000206, -0.00001770, 0.00000265, -0.00001201, 0.00000354, 0.00000713, -0.00000873, 0.00000645, -0.00000875, 0.00000338, -0.00000727, 0.00000620, -0.00000287, -0.00000440, -0.00000773, -0.00000300, -0.00000364, 0.00000879, -0.00001610, -0.00000655, -0.00000674, -0.00000325, -0.00001821, -0.00001350, -0.00002217, -0.00002353, -0.00009363, -0.00008448, 0.00015160, 0.00006247, 0.00004874, 0.00003999, 0.00002782, 0.00003710, 0.00004025, 0.00005736, 0.00003771, 0.00016210, -0.00003335, -0.00011954, -0.00001714, -0.00003834, 0.00000022, -0.00001614, -0.00000601, -0.00001653, 0.00000238, -0.00000611, -0.00000268, -0.00000379, 0.00000136, -0.00000411, -0.00000414, -0.00001201, 0.00000307, -0.00000286, -0.00000319, -0.00000661, 0.00000222, 0.00000039, -0.00000288, -0.00000346, -0.00000034, -0.00000363, -0.00000313, 0.00000393, -0.00000614, 0.00000046, -0.00001198, -0.00000588, -0.00001953, -0.00000666, 0.00000026, 0.00002417, 0.00001227, 0.00000438, 0.00000798, 0.00000593, 0.00000817, 0.00000880, 0.00000041, 0.00000227, 0.00000300, 0.00000509, 0.00000924, 0.00000147, 0.00000875, 0.00001301, -0.00000032, 0.00002231, -0.00001060, 0.00001653, 0.00001638, 0.00001227, -0.00001218, -0.00002467, -0.00000610, -0.00000940, 0.00000262, -0.00000909, 0.00000180, -0.00000661, -0.00000196, -0.00000403, 0.00000533, -0.00001091, 0.00000421, -0.00000164, -0.00000627, -0.00001053, 0.00000343, -0.00000202, -0.00000308, },
		{ -0.00001102, -0.00000176, 0.00000020, -0.00000213, 0.00000106, -0.00000113, -0.00000393, -0.00000015, -0.00000394, 0.00000009, -0.00000468, -0.00000205, -0.00000379, -0.00000205, -0.00000444, 0.00000562, -0.00001599, -0.00000116, -0.00000701, -0.00000628, 0.00001375, -0.00000997, 0.00000582, -0.00000645, 0.00000046, -0.00000299, 0.00000027, 0.00000177, -0.00000563, -0.00000453, -0.00000697, -0.00000295, 0.00000495, -0.00000322, -0.00001703, -0.00000233, -0.00000675, -0.00000689, -0.00002368, -0.00000970, -0.00003118, -0.00003605, -0.00014176, 0.00004680, 0.00014801, 0.00002523, 0.00006860, 0.00001725, 0.00004993, 0.00002197, 0.00007533, 0.00002405, 0.00014855, 0.00004442, -0.00014311, -0.00002755, -0.00003895, -0.00000930, -0.00001052, -0.00000737, -0.00001617, -0.00000335, -0.00000193, -0.00000505, -0.00000308, -0.00000113, -0.00000106, -0.00000306, -0.00001202, -0.00000191, 0.00000107, -0.00000429, -0.00000511, -0.00000222, 0.00000391, -0.00000362, -0.00000227, -0.00000324, 0.00000022, -0.00000657, 0.00000558, -0.00000640, 0.00000136, -0.00000970, -0.00000610, -0.00001704, -0.00001080, -0.00000075, 0.00001584, 0.00002016, 0.00000128, 0.00000959, 0.00000438, 0.00000756, 0.00000819, 0.00000504, -0.00000212, 0.00000617, 0.00000014, 0.00001030, 0.00000391, 0.00000334, 0.00001462, 0.00000270, 0.00001253, 0.00000816, -0.00000905, 0.00002972, 0.00000676, 0.00000844, -0.00002950, -0.00000814, -0.00001208, 0.00000164, -0.00000799, 0.00000197, -0.00000763, 0.00000005, -0.00000721, 0.00000600, -0.00000875, 0.00000020, 0.00000166, -0.00000435, -0.00001294, 0.00000206, -0.00000055, -0.00000190, },
		{ -0.00000653, -0.00000800, 0.00000381, -0.00000474, 0.00000237, -0.00000129, -0.00000205, -0.00000258, -0.00000184, -0.00000251, -0.00000126, -0.00000514, -0.00000146, -0.00000355, -0.00000328, 0.00000263, -0.00000798, -0.00001018, 0.00000285, -0.00001758, 0.00001801, -0.00000916, 0.00000465, -0.00000453, -0.00000123, -0.00000158, -0.00000017, 0.00000024, -0.00000277, -0.00000506, -0.00000504, -0.00000742, 0.00000472, 0.00000078, -0.00001224, -0.00001175, -0.00000110, -0.00000684, -0.00001722, -0.00001721, -0.00001960, -0.00002843, -0.00008869, -0.00010704, 0.00016970, 0.00006450, 0.00005654, 0.00003993, 0.00004019, 0.00002950, 0.00006035, 0.00004659, 0.00010243, 0.00012624, -0.00014518, -0.00004971, -0.00003562, -0.00001924, -0.00001032, -0.00000478, -0.00001614, -0.00000801, -0.00000129, -0.00000404, -0.00000451, -0.00000128, -0.00000243, 0.00000124, -0.00001206, -0.00000578, 0.00000142, -0.00000250, -0.00000516, -0.00000412, 0.00000321, -0.00000088, -0.00000318, -0.00000350, -0.00000014, -0.00000506, 0.00000259, -0.00000240, -0.00000247, -0.00000352, -0.00000963, -0.00001057, -0.00001809, 0.00000201, 0.00000521, 0.00002649, 0.00000028, 0.00001023, 0.00000312, 0.00000846, 0.00000421, 0.00001140, -0.00000567, 0.00000774, -0.00000156, 0.00000802, 0.00000572, 0.00000416, 0.00000690, 0.00001381, -0.00000097, 0.00002187, -0.00001551, 0.00001824, 0.00001865, 0.00000804, -0.00001157, -0.00002183, -0.00000698, -0.00000555, -0.00000407, -0.00000055, -0.00000428, -0.00000378, -0.00000323, -0.00000020, -0.00000170, -0.00000658, 0.00000552, -0.00000346, -0.00001122, -0.00000438, 0.00000458, -0.00000418, },
		{ -0.00000421, -0.00001096, 0.00000367, -0.00000414, 0.00000120, -0.00000053, -0.00000060, -0.00000382, -0.00000076, -0.00000445, 0.00000096, -0.00000645, -0.00000075, -0.00000437, -0.00000167, -0.00000128, -0.00000031, -0.00001726, 0.00000899, -0.00002121, 0.00001309, -0.00000201, -0.00000058, 0.00000023, -0.00000443, -0.00000014, -0.00000015, -0.00000014, -0.00000256, -0.00000451, -0.00000261, -0.00001044, 0.00000290, 0.00000157, -0.00000401, -0.00001791, -0.00000179, -0.00000536, -0.00000921, -0.00002477, -0.00000934, -0.00003465, -0.00003134, -0.00015512, 0.00004159, 0.00016589, 0.00002501, 0.00007564, 0.00002338, 0.00005107, 0.00003074, 0.00007760, 0.00005770, 0.00018034, -0.00010786, -0.00008913, -0.00002647, -0.00003093, -0.00001100, -0.00000587, -0.00001078, -0.00001440, -0.00000132, -0.00000341, -0.00000492, -0.00000151, -0.00000428, 0.00000308, -0.00000804, -0.00001043, 0.00000070, -0.00000123, -0.00000414, -0.00000537, 0.00000113, 0.00000164, -0.00000322, -0.00000334, -0.00000176, -0.00000278, -0.00000055, 0.00000155, -0.00000618, 0.00000121, -0.00001146, -0.00000550, -0.00002213, 0.00000130, -0.00000015, 0.00002503, 0.00000564, 0.00000651, 0.00000537, 0.00000683, 0.00000343, 0.00001123, -0.00000107, 0.00000202, 0.00000353, 0.00000148, 0.00000822, 0.00000474, 0.00000358, 0.00001377, 0.00000312, 0.00001189, 0.00000441, -0.00000957, 0.00003286, 0.00000474, 0.00000405, -0.00002305, -0.00001167, -0.00000526, -0.00000663, -0.00000060, -0.00000099, -0.00000744, -0.00000018, -0.00000389, 0.00000139, -0.00000880, 0.00000480, -0.00000071, -0.00000842, -0.00001064, 0.00000690, -0.00000370, },
		{ -0.00000549, -0.00000941, -0.00000091, 0.00000002, -0.00000276, 0.00000224, -0.00000183, -0.00000146, -0.00000293, -0.00000297, -0.00000112, -0.00000323, -0.00000401, -0.00000186, -0.00000330, -0.00000124, 0.00000151, -0.00001633, 0.00000479, -0.00001208, -0.00000270, 0.00001165, -0.00001057, 0.00000909, -0.00001103, 0.00000484, -0.00000524, 0.00000639, -0.00000881, 0.00000071, -0.00000588, -0.00000444, -0.00000760, 0.00000893, -0.00000691, -0.00000686, -0.00001772, 0.00000615, -0.00001382, -0.00001331, -0.00002245, -0.00001649, -0.00003285, -0.00008425, -0.00013133, 0.00018909, 0.00006671, 0.00006529, 0.00004234, 0.00004954, 0.00002691, 0.00007758, 0.00004878, 0.00018192, -0.00002240, -0.00014470, -0.00001181, -0.00004746, -0.00000552, -0.00001758, 0.00000266, -0.00002390, 0.00000133, -0.00000754, -0.00000043, -0.00000602, -0.00000093, -0.00000272, 0.00000202, -0.00001671, 0.00000146, -0.00000350, -0.00000013, -0.00000801, 0.00000093, 0.00000062, -0.00000011, -0.00000498, -0.00000135, -0.00000336, -0.00000078, 0.00000261, -0.00000634, 0.00000107, -0.00000874, -0.00000514, -0.00001944, -0.00000563, 0.00000145, 0.00001612, 0.00001550, 0.00000024, 0.00000994, 0.00000274, 0.00000675, 0.00000508, 0.00000801, -0.00000531, 0.00000872, -0.00000308, 0.00000869, 0.00000361, 0.00000677, 0.00000488, 0.00001447, -0.00000101, 0.00001816, -0.00001621, 0.00001955, 0.00002089, -0.00000045, -0.00000494, -0.00002438, -0.00000117, -0.00001066, -0.00000211, 0.00000219, -0.00000792, -0.00000178, -0.00000242, -0.00000028, -0.00000694, 0.00000084, 0.00000284, -0.00000601, -0.00001282, 0.00000296, 0.00000147, },
		{ -0.00000474, -0.00000855, -0.00000461, 0.00000254, -0.00000473, 0.00000299, -0.00000208, 0.00000022, -0.00000379, -0.00000225, -0.00000307, -0.00000057, -0.00000602, -0.00000064, -0.00000426, -0.00000092, 0.00000132, -0.00001167, -0.00000415, 0.00000088, -0.00001747, 0.00002015, -0.00001503, 0.00001236, -0.00001206, 0.00000544, -0.00000746, 0.00000982, -0.00000995, 0.00000049, -0.00000664, 0.00000101, -0.00001503, 0.00001187, -0.00000918, 0.00000331, -0.00002567, 0.00000649, -0.00001271, -0.00000427, -0.00003153, -0.00000383, -0.00004308, -0.00002078, -0.00017633, 0.00004338, 0.00017660, 0.00003415, 0.00007564, 0.00003878, 0.00004132, 0.00005708, 0.00006151, 0.00015059, 0.00008096, -0.00018415, -0.00001544, -0.00005278, -0.00000610, -0.00002626, 0.00000874, -0.00002422, -0.00000153, -0.00000923, 0.00000168, -0.00000848, 0.00000186, -0.00000850, 0.00000840, -0.00001661, -0.00000154, -0.00000413, 0.00000119, -0.00000749, -0.00000083, -0.00000006, 0.00000164, -0.00000467, -0.00000185, -0.00000351, -0.00000183, 0.00000387, -0.00000536, -0.00000098, -0.00000495, -0.00000632, -0.00001392, -0.00001351, 0.00000380, 0.00000710, 0.00002231, -0.00000208, 0.00001115, 0.00000146, 0.00000825, 0.00000130, 0.00001149, -0.00000496, 0.00000641, -0.00000045, 0.00000446, 0.00000417, 0.00000799, 0.00000220, 0.00001340, 0.00000418, 0.00000826, 0.00000316, -0.00000890, 0.00003619, -0.00000197, 0.00000478, -0.00002033, -0.00001063, -0.00000376, -0.00001193, 0.00000777, -0.00000771, -0.00000334, -0.00000236, 0.00000044, -0.00000742, -0.00000095, 0.00000332, -0.00000235, -0.00001262, -0.00000291, 0.00000529, },
		{ -0.00000048, -0.00000929, -0.00000618, 0.00000201, -0.00000333, 0.00000049, 0.00000031, -0.00000126, -0.00000103, -0.00000438, -0.00000220, -0.00000164, -0.00000389, -0.00000316, -0.00000199, -0.00000335, 0.00000285, -0.00000826, -0.00001042, 0.00000737, -0.00002052, 0.00001513, -0.00000727, 0.00000389, -0.00000210, -0.00000352, -0.00000043, 0.00000218, 0.00000129, -0.00001006, 0.00000028, -0.00000277, -0.00000885, 0.00000034, 0.00000158, -0.00000402, -0.00001026, -0.00001425, 0.00000392, -0.00001351, -0.00001548, -0.00002129, -0.00002039, -0.00002877, -0.00009014, -0.00014570, 0.00019672, 0.00008274, 0.00006221, 0.00006073, 0.00003952, 0.00005189, 0.00006250, 0.00012106, 0.00016045, -0.00017050, -0.00005998, -0.00003262, -0.00002560, -0.00001742, -0.00000532, -0.00000800, -0.00001483, -0.00000241, -0.00000530, -0.00000263, -0.00000276, -0.00000563, 0.00000304, -0.00000576, -0.00001077, 0.00000037, -0.00000378, -0.00000146, -0.00000623, 0.00000263, -0.00000106, -0.00000031, -0.00000516, -0.00000077, -0.00000577, 0.00000638, -0.00000438, -0.00000268, -0.00000277, -0.00000650, -0.00000920, -0.00001794, 0.00000271, 0.00000217, 0.00002270, 0.00000155, 0.00000765, 0.00000428, 0.00000595, 0.00000260, 0.00000811, 0.00000127, 0.00000063, 0.00000403, 0.00000047, 0.00000472, 0.00000567, 0.00000677, 0.00000406, 0.00001496, -0.00000225, 0.00001385, -0.00001335, 0.00002024, 0.00001994, -0.00000711, 0.00000206, -0.00002701, 0.00000536, -0.00002007, 0.00000776, -0.00000397, -0.00000432, -0.00000462, 0.00000317, -0.00000845, -0.00000260, 0.00000289, 0.00000063, -0.00001058, -0.00000737, 0.00000454, },
		{ 0.00000181, -0.00000661, -0.00000960, 0.00000210, -0.00000239, -0.00000124, 0.00000153, -0.00000202, 0.00000069, -0.00000450, -0.00000280, -0.00000207, -0.00000261, -0.00000428, -0.00000131, -0.00000410, 0.00000215, -0.00000337, -0.00001513, 0.00000821, -0.00001537, 0.00000361, 0.00000423, -0.00000717, 0.00000876, -0.00001183, 0.00000630, -0.00000635, 0.00001205, -0.00001658, 0.00000276, -0.00000600, -0.00000040, -0.00001154, 0.00001114, -0.00001300, 0.00000578, -0.00002802, 0.00001008, -0.00001760, -0.00000151, -0.00003487, -0.00000293, -0.00004522, -0.00001794, -0.00019010, 0.00003591, 0.00019659, 0.00003644, 0.00008731, 0.00003991, 0.00005184, 0.00006126, 0.00009126, 0.00020878, -0.00010514, -0.00012878, -0.00000667, -0.00004832, -0.00000397, -0.00002463, 0.00000779, -0.00002454, 0.00000283, -0.00001255, 0.00000315, -0.00000790, -0.00000024, -0.00000583, 0.00000504, -0.00001658, 0.00000279, -0.00000793, 0.00000259, -0.00000890, 0.00000348, -0.00000284, 0.00000230, -0.00000617, 0.00000029, -0.00000745, 0.00000494, 0.00000024, -0.00000607, -0.00000024, -0.00000762, -0.00000439, -0.00001991, 0.00000005, -0.00000071, 0.00001947, 0.00000731, 0.00000344, 0.00000725, 0.00000373, 0.00000438, 0.00000468, 0.00000486, -0.00000060, 0.00000376, 0.00000118, 0.00000279, 0.00000381, 0.00000924, 0.00000184, 0.00001234, 0.00000577, 0.00000151, 0.00000679, -0.00000919, 0.00003825, -0.00001067, 0.00001009, -0.00002205, -0.00000380, -0.00001163, -0.00000601, 0.00000639, -0.00000786, -0.00000357, 0.00000051, -0.00000288, -0.00000952, 0.00000575, -0.00000026, -0.00000545, -0.00001140, 0.00000272, },
		{ 0.00000088, -0.00000108, -0.00001338, 0.00000187, -0.00000187, -0.00000153, 0.00000061, -0.00000048, -0.00000084, -0.00000087, -0.00000610, -0.00000014, -0.00000437, -0.00000197, -0.00000395, -0.00000141, -0.00000229, 0.00000371, -0.00001881, 0.00000604, -0.00000844, -0.00000565, 0.00001060, -0.00001163, 0.00001082, -0.00001035, 0.00000391, -0.00000566, 0.00001065, -0.00000876, -0.00000636, -0.00000134, -0.00000090, -0.00001011, 0.00000573, -0.00000724, 0.00000200, -0.00001606, -0.00000734, -0.00000340, -0.00000901, -0.00002001, -0.00001887, -0.00002446, -0.00002482, -0.00009713, -0.00015912, 0.00020088, 0.00010450, 0.00005822, 0.00008018, 0.00002997, 0.00008072, 0.00005073, 0.00023084, -0.00000969, -0.00018588, -0.00000628, -0.00005136, -0.00000589, -0.00002866, 0.00000516, -0.00001852, -0.00000267, -0.00001054, -0.00000039, -0.00000527, -0.00000112, -0.00000747, 0.00000528, -0.00001183, -0.00000209, -0.00000547, -0.00000085, -0.00000478, -0.00000098, 0.00000026, -0.00000057, -0.00000214, -0.00000287, -0.00000407, -0.00000196, 0.00000789, -0.00000987, 0.00000211, -0.00000920, -0.00000054, -0.00001921, -0.00000364, -0.00000181, 0.00001355, 0.00001357, 0.00000008, 0.00000896, 0.00000290, 0.00000498, 0.00000334, 0.00000503, 0.00000085, 0.00000252, 0.00000174, 0.00000249, 0.00000151, 0.00000810, 0.00000636, 0.00000300, 0.00001657, -0.00000734, 0.00001372, -0.00001055, 0.00002092, 0.00001438, -0.00000711, 0.00000316, -0.00002250, 0.00000234, -0.00002063, 0.00001244, -0.00000860, -0.00000093, -0.00000581, 0.00000556, -0.00001625, 0.00000677, -0.00000068, -0.00000107, -0.00001331, 0.00000090, },
		{ 0.00000165, 0.00000047, -0.00001105, -0.00000364, 0.00000208, -0.00000393, 0.00000126, -0.00000043, -0.00000131, 0.00000100, -0.00000622, -0.00000120, -0.00000379, -0.00000209, -0.00000389, -0.00000125, -0.00000400, 0.00000589, -0.00001528, -0.00000262, 0.00000146, -0.00001405, 0.00001451, -0.00001257, 0.00000813, -0.00000494, -0.00000087, -0.00000226, 0.00000478, 0.00000203, -0.00001354, -0.00000004, -0.00000176, -0.00000527, -0.00000212, 0.00000028, -0.00000623, -0.00000116, -0.00002006, 0.00000265, -0.00001335, -0.00000620, -0.00003099, -0.00000933, -0.00003883, -0.00002612, -0.00019152, 0.00001269, 0.00023279, 0.00002682, 0.00011462, 0.00002270, 0.00009018, 0.00003355, 0.00020382, 0.00011059, -0.00021586, -0.00003451, -0.00004052, -0.00001717, -0.00002301, -0.00000686, -0.00000873, -0.00000894, -0.00000652, -0.00000639, -0.00000158, -0.00000312, -0.00000596, 0.00000086, -0.00000487, -0.00000661, -0.00000237, -0.00000541, -0.00000088, -0.00000448, 0.00000269, -0.00000304, 0.00000069, -0.00000457, -0.00000134, -0.00000716, 0.00001061, -0.00000748, -0.00000055, -0.00000663, -0.00000216, -0.00001274, -0.00001094, 0.00000156, 0.00000342, 0.00002100, -0.00000330, 0.00001064, 0.00000172, 0.00000651, 0.00000118, 0.00000655, -0.00000102, 0.00000565, -0.00000182, 0.00000587, -0.00000202, 0.00000675, 0.00000824, 0.00000226, 0.00001187, 0.00000459, -0.00000353, 0.00001228, -0.00000935, 0.00003538, -0.00001410, 0.00001208, -0.00001980, -0.00000166, -0.00001779, 0.00000283, -0.00000060, -0.00000276, -0.00000696, 0.00000603, -0.00001255, -0.00000159, 0.00000491, -0.00000209, -0.00000990, -0.00000377, },
		{ 0.00000416, -0.00000223, -0.00000409, -0.00001151, 0.00000653, -0.00000644, 0.00000257, -0.00000186, 0.00000031, -0.00000077, -0.00000176, -0.00000598, 0.00000011, -0.00000599, 0.00000007, -0.00000492, -0.00000107, 0.00000123, -0.00000435, -0.00001577, 0.00001206, -0.00002107, 0.00001683, -0.00001164, 0.00000427, -0.00000116, -0.00000203, -0.00000239, 0.00000264, 0.00000528, -0.00000949, -0.00000788, 0.00000262, -0.00000621, -0.00000087, -0.00000252, -0.00000372, -0.00000255, -0.00001068, -0.00001134, -0.00000210, -0.00001221, -0.00001595, -0.00002618, -0.00001664, -0.00003542, -0.00008847, -0.00019012, 0.00022057, 0.00011340, 0.00007403, 0.00007544, 0.00005498, 0.00006876, 0.00012206, 0.00023571, -0.00020466, -0.00008318, -0.00003105, -0.00002514, -0.00002017, -0.00001374, -0.00000818, -0.00000758, -0.00000714, -0.00000733, -0.00000401, -0.00000037, -0.00000743, -0.00000023, -0.00000378, -0.00000523, -0.00000312, -0.00000567, -0.00000244, -0.00000294, 0.00000070, -0.00000097, -0.00000154, -0.00000177, -0.00000259, -0.00000652, 0.00000539, 0.00000147, -0.00000725, -0.00000061, -0.00000811, -0.00000299, -0.00001919, 0.00000644, -0.00000647, 0.00002486, -0.00000290, 0.00000941, 0.00000231, 0.00000712, -0.00000025, 0.00000799, -0.00000331, 0.00000787, -0.00000283, 0.00000580, -0.00000109, 0.00000312, 0.00000750, 0.00000735, 0.00000161, 0.00001747, -0.00001394, 0.00001777, -0.00000853, 0.00001874, 0.00000966, -0.00000507, 0.00000251, -0.00001562, -0.00000720, -0.00001083, 0.00000640, -0.00000483, -0.00000357, -0.00000065, -0.00000183, -0.00001329, 0.00001045, -0.00000377, -0.00000499, -0.00000916, },
		{ 0.00000324, -0.00000229, -0.00000120, -0.00001298, 0.00000404, -0.00000310, -0.00000045, -0.00000003, -0.00000056, -0.00000098, 0.00000063, -0.00000717, 0.00000037, -0.00000654, 0.00000043, -0.00000497, -0.00000079, -0.00000189, 0.00000334, -0.00002154, 0.00001252, -0.00001759, 0.00000909, -0.00000128, -0.00000642, 0.00000612, -0.00000588, 0.00000088, -0.00000217, 0.00000836, -0.00000551, -0.00001072, -0.00000010, -0.00000286, -0.00000189, -0.00000206, -0.00000431, -0.00000372, -0.00000341, -0.00001583, -0.00000314, -0.00001088, -0.00000742, -0.00003202, -0.00000894, -0.00004064, -0.00002672, -0.00020017, -0.00000664, 0.00026503, 0.00002924, 0.00012776, 0.00002609, 0.00010732, 0.00004981, 0.00029904, -0.00011621, -0.00015984, -0.00001738, -0.00003960, -0.00001143, -0.00002323, -0.00000622, -0.00001122, -0.00000224, -0.00001093, -0.00000450, -0.00000127, -0.00000490, -0.00000375, -0.00000167, -0.00000631, -0.00000066, -0.00000707, -0.00000343, -0.00000297, 0.00000025, 0.00000006, -0.00000268, -0.00000082, -0.00000227, -0.00000551, -0.00000035, 0.00000785, -0.00000942, 0.00000143, -0.00001059, 0.00000150, -0.00002040, 0.00000477, -0.00000752, 0.00001802, 0.00000625, 0.00000165, 0.00000781, 0.00000340, 0.00000302, 0.00000447, 0.00000043, 0.00000223, 0.00000433, -0.00000149, 0.00000584, -0.00000325, 0.00000787, 0.00000810, 0.00000186, 0.00001256, -0.00000067, -0.00000301, 0.00001626, -0.00000964, 0.00002865, -0.00001087, 0.00000807, -0.00001033, -0.00001064, -0.00001131, 0.00000145, -0.00000106, -0.00000219, -0.00000483, 0.00000365, -0.00001602, 0.00000642, 0.00000079, -0.00000450, -0.00001012, },
		{ -0.00000071, 0.00000139, -0.00000353, -0.00000829, -0.00000360, 0.00000357, -0.00000563, 0.00000370, -0.00000349, 0.00000124, -0.00000107, -0.00000304, -0.00000423, -0.00000230, -0.00000446, 0.00000004, -0.00000493, -0.00000013, 0.00000292, -0.00001556, 0.00000062, -0.00000290, -0.00000863, 0.00001682, -0.00002101, 0.00001523, -0.00001285, 0.00000861, -0.00001096, 0.00001465, -0.00000789, -0.00000287, -0.00001244, 0.00000716, -0.00001063, 0.00000863, -0.00001520, 0.00000575, -0.00001246, 0.00000117, -0.00002322, 0.00000588, -0.00002019, -0.00000801, -0.00003664, -0.00000522, -0.00005195, -0.00007226, -0.00023151, 0.00024820, 0.00012020, 0.00009523, 0.00006553, 0.00009299, 0.00004695, 0.00026006, 0.00004985, -0.00024766, -0.00000480, -0.00006319, 0.00000567, -0.00003932, 0.00000406, -0.00002570, 0.00001041, -0.00001920, 0.00000102, -0.00001009, 0.00000470, -0.00001276, 0.00000577, -0.00001374, 0.00000664, -0.00001104, -0.00000088, -0.00000762, 0.00000369, -0.00000253, 0.00000046, -0.00000459, 0.00000168, -0.00000691, -0.00000214, 0.00000724, -0.00000448, -0.00000215, -0.00000731, -0.00000160, -0.00001337, -0.00000428, 0.00000019, 0.00000330, 0.00001920, -0.00000732, 0.00001350, -0.00000097, 0.00000787, -0.00000110, 0.00000695, -0.00000596, 0.00001174, -0.00000674, 0.00000931, -0.00000488, 0.00000590, 0.00000605, 0.00000799, 0.00000193, 0.00001381, -0.00001597, 0.00002132, -0.00000560, 0.00001076, 0.00001228, -0.00000877, 0.00000882, -0.00001827, -0.00000854, -0.00000471, 0.00000030, -0.00000019, -0.00000549, 0.00000267, -0.00001090, -0.00000300, 0.00000675, -0.00000501, -0.00000861, },
		{ -0.00000324, 0.00000353, -0.00000415, -0.00000585, -0.00000745, 0.00000516, -0.00000563, 0.00000287, -0.00000256, 0.00000043, -0.00000101, -0.00000078, -0.00000582, -0.00000098, -0.00000658, 0.00000165, -0.00000545, -0.00000064, 0.00000187, -0.00000775, -0.00001098, 0.00000879, -0.00002077, 0.00002548, -0.00002263, 0.00001186, -0.00001000, 0.00000735, -0.00001003, 0.00001136, -0.00000356, -0.00000084, -0.00001514, 0.00000588, -0.00001038, 0.00001072, -0.00001531, 0.00000462, -0.00001355, 0.00000755, -0.00002523, 0.00000272, -0.00001852, 0.00000157, -0.00004383, 0.00000414, -0.00005811, -0.00001014, -0.00022756, -0.00001016, 0.00028239, 0.00005510, 0.00011356, 0.00006989, 0.00007642, 0.00017612, 0.00021244, -0.00027618, -0.00003596, -0.00006420, 0.00000171, -0.00003743, 0.00000126, -0.00002854, 0.00000849, -0.00001431, -0.00000173, -0.00001253, 0.00000677, -0.00001389, 0.00000625, -0.00001442, 0.00000534, -0.00000666, -0.00000348, -0.00000800, 0.00000208, -0.00000096, 0.00000065, -0.00000535, 0.00000160, -0.00000452, -0.00000510, 0.00000584, -0.00000008, -0.00000412, -0.00000450, -0.00000554, -0.00000604, -0.00001232, 0.00000700, -0.00000755, 0.00002424, -0.00000753, 0.00001134, 0.00000077, 0.00000766, -0.00000160, 0.00000797, -0.00000684, 0.00000936, -0.00000176, 0.00000354, 0.00000125, -0.00000017, 0.00000623, 0.00000880, 0.00000214, 0.00001064, -0.00000487, 0.00000009, 0.00001988, -0.00001416, 0.00002567, -0.00000892, 0.00000638, -0.00000403, -0.00002013, 0.00000075, -0.00000772, 0.00000529, -0.00000636, 0.00000198, -0.00000743, -0.00000743, 0.00000573, -0.00000082, -0.00000920, },
		{ -0.00000587, 0.00000375, -0.00000167, -0.00000690, -0.00000706, 0.00000212, -0.00000150, -0.00000137, 0.00000143, -0.00000313, 0.00000166, -0.00000235, -0.00000257, -0.00000410, -0.00000395, -0.00000226, -0.00000074, -0.00000498, 0.00000340, -0.00000410, -0.00001455, 0.00000931, -0.00001878, 0.00001654, -0.00000609, -0.00000600, 0.00000440, -0.00000576, 0.00000353, -0.00000402, 0.00001127, -0.00001087, -0.00000229, -0.00001013, 0.00000284, -0.00000283, 0.00000316, -0.00001454, 0.00000369, -0.00001033, 0.00000150, -0.00002584, 0.00000531, -0.00001842, -0.00001131, -0.00003342, -0.00000995, -0.00005178, -0.00007329, -0.00025751, 0.00025697, 0.00015543, 0.00008515, 0.00009901, 0.00008089, 0.00012123, 0.00029453, -0.00019823, -0.00012918, -0.00002820, -0.00003784, -0.00000665, -0.00002522, -0.00000783, -0.00001706, 0.00000816, -0.00001694, -0.00000201, -0.00000577, -0.00000114, -0.00000610, -0.00000203, -0.00000923, 0.00000764, -0.00001248, -0.00000129, -0.00000720, 0.00000653, -0.00000432, -0.00000012, -0.00000486, 0.00000271, -0.00001041, 0.00000640, 0.00000050, -0.00000225, -0.00000417, -0.00000698, -0.00000278, -0.00001485, 0.00000758, -0.00000903, 0.00001815, 0.00000139, 0.00000267, 0.00000690, 0.00000348, 0.00000244, 0.00000384, -0.00000136, 0.00000073, 0.00000678, -0.00000302, 0.00000624, -0.00000328, 0.00000538, 0.00000619, 0.00000781, 0.00000261, 0.00000705, -0.00001229, 0.00002274, -0.00000387, 0.00000143, 0.00001906, -0.00001566, 0.00001902, -0.00002958, 0.00000256, -0.00001059, 0.00000340, -0.00000255, 0.00000023, -0.00000498, -0.00000870, 0.00000197, 0.00000301, -0.00000825, },
		{ -0.00001110, 0.00000468, -0.00000023, -0.00000560, -0.00000907, 0.00000135, -0.00000023, -0.00000233, 0.00000207, -0.00000320, 0.00000107, -0.00000162, -0.00000170, -0.00000404, -0.00000398, -0.00000381, 0.00000066, -0.00000503, 0.00000096, 0.00000072, -0.00001585, 0.00000588, -0.00001146, 0.00000194, 0.00001223, -0.00001986, 0.00001295, -0.00001394, 0.00001194, -0.00001289, 0.00001839, -0.00001447, 0.00000478, -0.00001741, 0.00000695, -0.00001032, 0.00001513, -0.00002370, 0.00001067, -0.00002015, 0.00001631, -0.00003441, 0.00000796, -0.00002356, 0.00000632, -0.00004949, 0.00001001, -0.00006924, 0.00000044, -0.00025006, -0.00002297, 0.00031423, 0.00006869, 0.00011834, 0.00009687, 0.00008278, 0.00031234, -0.00005748, -0.00022868, -0.00000022, -0.00007441, 0.00001662, -0.00004421, 0.00000843, -0.00003708, 0.00001992, -0.00002196, 0.00000331, -0.00001529, 0.00000785, -0.00001470, 0.00000723, -0.00001945, 0.00001404, -0.00001323, 0.00000046, -0.00001202, 0.00000818, -0.00000441, 0.00000182, -0.00000788, 0.00000486, -0.00000972, 0.00000238, 0.00000359, -0.00000309, -0.00000105, -0.00000946, -0.00000046, -0.00001525, 0.00000515, -0.00000554, 0.00000813, 0.00001065, -0.00000392, 0.00001004, 0.00000151, 0.00000507, 0.00000098, 0.00000271, -0.00000442, 0.00000873, -0.00000200, 0.00000411, -0.00000020, 0.00000146, 0.00000542, 0.00000816, 0.00000464, 0.00000336, -0.00000259, 0.00000103, 0.00002418, -0.00002360, 0.00003002, -0.00001311, 0.00001248, -0.00000984, -0.00001646, 0.00000294, -0.00001094, 0.00000739, -0.00000518, 0.00000145, -0.00001363, 0.00000310, 0.00000020, -0.00000187, },
		{ -0.00001263, 0.00000254, 0.00000146, -0.00000249, -0.00001229, 0.00000124, -0.00000029, -0.00000135, 0.00000035, -0.00000049, -0.00000229, 0.00000166, -0.00000421, -0.00000013, -0.00000709, -0.00000196, -0.00000251, -0.00000033, -0.00000502, 0.00000658, -0.00001660, 0.00000231, -0.00000492, -0.00000880, 0.00002027, -0.00001895, 0.00000761, -0.00000949, 0.00000664, -0.00000676, 0.00000985, -0.00000315, -0.00000388, -0.00000690, -0.00000499, -0.00000323, 0.00000830, -0.00000991, -0.00000490, -0.00000673, 0.00000128, -0.00000890, -0.00001916, -0.00000244, -0.00001037, -0.00002041, -0.00002347, -0.00002384, -0.00004103, -0.00008656, -0.00027339, 0.00025432, 0.00020872, 0.00005831, 0.00016134, 0.00003852, 0.00030161, 0.00008771, -0.00027692, -0.00002111, -0.00007387, 0.00000256, -0.00003227, -0.00000130, -0.00003012, 0.00000413, -0.00000623, -0.00000594, -0.00001053, 0.00000153, -0.00000889, 0.00000251, -0.00001408, 0.00000370, -0.00000089, -0.00000678, -0.00000717, -0.00000047, 0.00000345, -0.00000281, -0.00000346, -0.00000126, -0.00000169, -0.00000652, 0.00000974, -0.00000690, 0.00000380, -0.00001153, 0.00000044, -0.00001388, 0.00000067, 0.00000014, -0.00000123, 0.00001602, -0.00000589, 0.00000926, 0.00000191, 0.00000584, 0.00000025, 0.00000417, -0.00000532, 0.00000583, 0.00000167, 0.00000169, 0.00000157, 0.00000047, 0.00000355, 0.00000601, 0.00000954, -0.00000133, 0.00000435, -0.00000832, 0.00002463, -0.00000768, -0.00000121, 0.00002088, -0.00001582, 0.00001994, -0.00003378, 0.00001178, -0.00001824, 0.00000870, -0.00000558, 0.00000594, -0.00001662, 0.00000313, -0.00000207, 0.00000230, },
		{ -0.00000579, -0.00000560, 0.00000625, -0.00000236, -0.00001005, -0.00000487, 0.00000431, -0.00000414, 0.00000197, -0.00000093, -0.00000210, 0.00000106, -0.00000315, -0.00000041, -0.00000504, -0.00000449, -0.00000195, -0.00000018, -0.00000488, 0.00000475, -0.00000962, -0.00000728, 0.00000533, -0.00001891, 0.00002267, -0.00001230, 0.00000043, -0.00000453, 0.00000014, 0.00000017, 0.00000144, 0.00000648, -0.00001040, 0.00000061, -0.00001076, -0.00000172, 0.00000258, 0.00000272, -0.00001531, 0.00000114, -0.00001029, 0.00000833, -0.00002892, 0.00000004, -0.00001493, -0.00000385, -0.00003760, -0.00000502, -0.00005744, -0.00001419, -0.00024714, -0.00006627, 0.00038325, 0.00004384, 0.00017935, 0.00005670, 0.00023034, 0.00024263, -0.00028096, -0.00007141, -0.00005736, -0.00002579, -0.00001433, -0.00001502, -0.00001699, -0.00001529, 0.00000592, -0.00001050, -0.00000653, -0.00000558, -0.00000304, -0.00000294, -0.00000648, -0.00000758, 0.00000776, -0.00000907, -0.00000427, -0.00000719, 0.00000705, -0.00000354, -0.00000126, -0.00000519, 0.00000247, -0.00000970, 0.00000946, -0.00000448, 0.00000154, -0.00000607, -0.00000514, -0.00000681, -0.00000873, 0.00000910, -0.00001038, 0.00001936, -0.00000649, 0.00000838, 0.00000124, 0.00000745, -0.00000096, 0.00000603, -0.00000615, 0.00000429, 0.00000116, 0.00000436, -0.00000139, 0.00000367, -0.00000028, 0.00000592, 0.00000766, 0.00000545, -0.00000428, 0.00000363, 0.00000079, 0.00002496, -0.00003040, 0.00003423, -0.00001585, 0.00001680, -0.00001842, -0.00000466, -0.00000476, -0.00000574, 0.00000309, 0.00000260, -0.00000979, -0.00000624, 0.00000429, -0.00000250, },
		{ 0.00000016, -0.00001159, 0.00000840, -0.00000228, -0.00000475, -0.00001313, 0.00000886, -0.00000642, 0.00000352, -0.00000216, 0.00000008, -0.00000225, 0.00000096, -0.00000473, 0.00000127, -0.00001005, 0.00000166, -0.00000445, 0.00000070, -0.00000296, 0.00000215, -0.00001929, 0.00001532, -0.00002516, 0.00001972, -0.00000436, -0.00000338, -0.00000330, -0.00000252, 0.00000133, 0.00000040, 0.00000690, -0.00000649, -0.00000391, -0.00000315, -0.00001055, 0.00000482, 0.00000287, -0.00000836, -0.00000809, -0.00000382, 0.00000014, -0.00000878, -0.00002258, 0.00000175, -0.00001650, -0.00001237, -0.00003341, -0.00001675, -0.00005264, -0.00007733, -0.00031610, 0.00028147, 0.00023069, 0.00008014, 0.00016848, 0.00010077, 0.00039095, -0.00023740, -0.00013697, -0.00004360, -0.00004687, -0.00001115, -0.00001628, -0.00001400, -0.00002069, 0.00000077, -0.00000266, -0.00000989, -0.00000506, -0.00000570, -0.00000122, -0.00000556, -0.00000959, 0.00000442, -0.00000193, -0.00000722, -0.00000667, 0.00000189, 0.00000249, -0.00000428, -0.00000299, -0.00000087, -0.00000492, 0.00000168, 0.00000449, -0.00000680, 0.00000383, -0.00001307, 0.00000173, -0.00001823, 0.00001570, -0.00001447, 0.00001756, -0.00000339, 0.00000496, 0.00000245, 0.00000671, 0.00000015, 0.00000576, -0.00000473, 0.00000180, 0.00000077, 0.00000519, -0.00000081, 0.00000260, 0.00000069, 0.00000367, 0.00000436, 0.00001268, -0.00000974, 0.00000843, -0.00000707, 0.00002607, -0.00001424, 0.00000180, 0.00001805, -0.00000999, 0.00000968, -0.00002338, 0.00000766, -0.00001592, 0.00000620, -0.00000008, 0.00000005, -0.00001712, 0.00001053, -0.00000702, },
		{ -0.00000333, -0.00000854, 0.00000269, 0.00000263, -0.00000342, -0.00001411, 0.00000405, -0.00000038, -0.00000171, 0.00000203, -0.00000215, -0.00000121, 0.00000049, -0.00000443, 0.00000176, -0.00000831, -0.00000144, -0.00000297, 0.00000037, -0.00000399, 0.00000527, -0.00001984, 0.00001129, -0.00001561, 0.00000285, 0.00001197, -0.00001287, 0.00000486, -0.00001143, 0.00000707, -0.00000511, 0.00001133, -0.00000673, -0.00000350, -0.00000281, -0.00000773, -0.00000433, 0.00000930, -0.00000721, -0.00000747, -0.00000749, -0.00000018, -0.00000183, -0.00002197, -0.00000689, -0.00000915, -0.00000950, -0.00003081, -0.00001614, -0.00004871, -0.00002750, -0.00024691, -0.00010788, 0.00044945, 0.00003303, 0.00023364, 0.00004013, 0.00042846, -0.00008139, -0.00025047, -0.00001020, -0.00007990, -0.00000148, -0.00002831, -0.00000109, -0.00003086, 0.00000054, -0.00000557, -0.00000196, -0.00001194, -0.00000212, -0.00000659, 0.00000110, -0.00001429, 0.00000336, -0.00000084, -0.00000317, -0.00000970, -0.00000022, 0.00000305, -0.00000184, -0.00000442, -0.00000129, -0.00000345, -0.00000204, 0.00000822, -0.00000868, 0.00000543, -0.00001177, 0.00000130, -0.00001753, 0.00001026, -0.00000600, 0.00000513, 0.00000814, -0.00000532, 0.00000960, 0.00000003, 0.00000660, 0.00000028, 0.00000219, -0.00000548, 0.00000572, -0.00000121, 0.00000702, -0.00000451, 0.00000659, -0.00000135, 0.00000486, 0.00000999, 0.00000010, -0.00000503, 0.00000755, 0.00000196, 0.00001945, -0.00002863, 0.00003120, -0.00000947, 0.00000796, -0.00001364, -0.00000205, -0.00000754, -0.00000393, 0.00000334, 0.00000420, -0.00001797, 0.00000524, -0.00000109, },
		{ -0.00000786, -0.00000442, -0.00000377, 0.00000718, -0.00000373, -0.00000953, -0.00000626, 0.00000836, -0.00000799, 0.00000671, -0.00000519, 0.00000180, -0.00000300, -0.00000017, -0.00000330, -0.00000050, -0.00000931, 0.00000295, -0.00000540, 0.00000158, -0.00000028, -0.00000935, -0.00000477, 0.00000493, -0.00002060, 0.00002987, -0.00002365, 0.00001589, -0.00002195, 0.00001408, -0.00001348, 0.00001906, -0.00001159, 0.00000401, -0.00001267, 0.00000713, -0.00002200, 0.00002089, -0.00001506, 0.00000690, -0.00002439, 0.00001347, -0.00001656, 0.00000528, -0.00003778, 0.00001732, -0.00003443, 0.00000938, -0.00006035, 0.00000999, -0.00008703, -0.00004469, -0.00038510, 0.00033456, 0.00022789, 0.00014913, 0.00012099, 0.00031748, 0.00018190, -0.00038324, 0.00002011, -0.00011358, 0.00001306, -0.00005248, 0.00002134, -0.00004747, 0.00001035, -0.00002228, 0.00001591, -0.00002489, 0.00000863, -0.00002002, 0.00001368, -0.00002310, 0.00000808, -0.00000836, 0.00000748, -0.00001608, 0.00000268, -0.00000363, 0.00000644, -0.00001011, 0.00000300, -0.00000747, 0.00000037, 0.00000450, -0.00000202, -0.00000211, -0.00000223, -0.00000648, -0.00000827, -0.00000420, 0.00000967, -0.00001122, 0.00002107, -0.00001524, 0.00001551, -0.00000551, 0.00001164, -0.00000425, 0.00000854, -0.00001136, 0.00000950, -0.00000650, 0.00001158, -0.00000651, 0.00000691, -0.00000111, 0.00000384, 0.00000456, 0.00001129, -0.00001465, 0.00001339, -0.00000531, 0.00002169, -0.00001447, 0.00000065, 0.00002102, -0.00001195, 0.00000429, -0.00001212, -0.00000079, -0.00000933, 0.00000115, 0.00000758, -0.00001239, -0.00000564, 0.00000741, },
		{ -0.00000371, -0.00000719, -0.00000409, 0.00000551, -0.00000016, -0.00000720, -0.00001117, 0.00000834, -0.00000517, 0.00000315, -0.00000124, -0.00000085, -0.00000145, -0.00000063, -0.00000386, 0.00000157, -0.00000965, 0.00000150, -0.00000470, 0.00000116, -0.00000098, -0.00000315, -0.00001468, 0.00001528, -0.00002829, 0.00002865, -0.00001655, 0.00000987, -0.00001402, 0.00000384, -0.00000637, 0.00001085, -0.00000136, -0.00000201, -0.00000919, 0.00000589, -0.00001814, 0.00000995, -0.00000470, 0.00000322, -0.00001877, 0.00000397, -0.00001033, 0.00000610, -0.00003188, 0.00000432, -0.00002362, 0.00000839, -0.00005140, 0.00000274, -0.00007265, -0.00000753, -0.00028298, -0.00011183, 0.00046909, 0.00009799, 0.00019938, 0.00020007, 0.00040004, -0.00039750, -0.00004995, -0.00008117, -0.00001906, -0.00004148, 0.00000834, -0.00003214, -0.00000406, -0.00001747, 0.00000924, -0.00001494, 0.00000148, -0.00001680, 0.00000838, -0.00001499, -0.00000065, -0.00000426, 0.00000428, -0.00000867, -0.00000393, -0.00000235, 0.00000509, -0.00000627, -0.00000047, -0.00000467, -0.00000311, 0.00000496, 0.00000112, -0.00000555, 0.00000183, -0.00000879, -0.00000298, -0.00001371, 0.00001728, -0.00001653, 0.00002151, -0.00001210, 0.00000919, -0.00000044, 0.00000705, -0.00000044, 0.00000670, -0.00000801, 0.00000411, -0.00000209, 0.00000499, 0.00000203, -0.00000119, 0.00000531, -0.00000046, 0.00000331, 0.00001162, -0.00000716, -0.00000035, 0.00000922, 0.00000094, 0.00001432, -0.00002443, 0.00002709, -0.00000203, -0.00000672, 0.00000225, -0.00001162, -0.00000031, -0.00001000, 0.00001194, -0.00000668, -0.00001145, 0.00000721, },
		{ 0.00000196, -0.00000909, -0.00000497, 0.00000322, 0.00000325, -0.00000584, -0.00001090, 0.00000168, 0.00000246, -0.00000368, 0.00000499, -0.00000541, 0.00000258, -0.00000424, -0.00000032, -0.00000235, -0.00000302, -0.00000596, 0.00000156, -0.00000502, 0.00000383, -0.00000373, -0.00001390, 0.00001014, -0.00001669, 0.00000799, 0.00000702, -0.00001154, 0.00000902, -0.00001864, 0.00001200, -0.00001001, 0.00002063, -0.00001923, 0.00000756, -0.00001263, 0.00000593, -0.00001859, 0.00002015, -0.00001854, 0.00000949, -0.00002751, 0.00001835, -0.00002310, 0.00000920, -0.00004038, 0.00001772, -0.00003425, 0.00001050, -0.00006579, 0.00001447, -0.00009974, -0.00003618, -0.00042702, 0.00034724, 0.00028723, 0.00015713, 0.00018598, 0.00046735, -0.00022447, -0.00023821, 0.00001444, -0.00009599, 0.00000877, -0.00004481, 0.00001642, -0.00004478, 0.00001355, -0.00002447, 0.00001658, -0.00002298, 0.00000496, -0.00001636, 0.00000994, -0.00002295, 0.00001298, -0.00001322, 0.00001023, -0.00001817, 0.00000741, -0.00000654, 0.00000657, -0.00001129, 0.00000520, -0.00001271, 0.00001010, -0.00000103, -0.00000208, -0.00000193, -0.00000418, -0.00000390, -0.00001436, 0.00001283, -0.00000851, 0.00000910, 0.00000239, -0.00000559, 0.00001036, -0.00000200, 0.00000694, 0.00000065, 0.00000046, -0.00000553, 0.00000601, -0.00000440, 0.00001032, -0.00000629, 0.00000750, -0.00000062, 0.00000187, 0.00000745, 0.00000367, -0.00001124, 0.00001442, -0.00000227, 0.00001220, -0.00000692, -0.00000686, 0.00003162, -0.00002662, 0.00001466, -0.00001453, 0.00000053, -0.00001051, 0.00000484, 0.00000369, -0.00001457, 0.00000301, },
		{ 0.00000077, -0.00000308, -0.00001152, 0.00000480, 0.00000219, -0.00000127, -0.00001166, -0.00000215, 0.00000414, -0.00000350, 0.00000404, -0.00000318, 0.00000061, -0.00000235, -0.00000167, -0.00000205, -0.00000143, -0.00000678, 0.00000103, -0.00000464, 0.00000202, 0.00000037, -0.00001414, 0.00000372, -0.00000422, -0.00000971, 0.00002282, -0.00002284, 0.00001947, -0.00002512, 0.00001469, -0.00001616, 0.00002562, -0.00001876, 0.00000847, -0.00001657, 0.00001333, -0.00002519, 0.00002139, -0.00001983, 0.00001656, -0.00003265, 0.00001934, -0.00002622, 0.00001868, -0.00004374, 0.00001583, -0.00003720, 0.00002535, -0.00007274, 0.00002393, -0.00010099, 0.00001565, -0.00032020, -0.00012349, 0.00049894, 0.00017531, 0.00016593, 0.00047056, 0.00002361, -0.00040724, 0.00005328, -0.00012665, 0.00002498, -0.00007108, 0.00003537, -0.00005707, 0.00002178, -0.00003710, 0.00002439, -0.00002630, 0.00001122, -0.00002676, 0.00001857, -0.00002844, 0.00001527, -0.00001694, 0.00001340, -0.00001694, 0.00000579, -0.00000862, 0.00000877, -0.00001157, 0.00000561, -0.00001258, 0.00000567, 0.00000369, -0.00000310, -0.00000179, -0.00000397, -0.00000170, -0.00001425, 0.00000565, 0.00000141, -0.00000315, 0.00001407, -0.00001434, 0.00001364, -0.00000405, 0.00000817, -0.00000090, 0.00000489, -0.00000955, 0.00000753, -0.00000591, 0.00000874, -0.00000175, 0.00000205, 0.00000425, -0.00000119, 0.00000521, 0.00000634, -0.00000733, 0.00000212, 0.00001270, -0.00000695, 0.00001767, -0.00002690, 0.00003154, -0.00000716, -0.00000900, 0.00001062, -0.00001809, 0.00000492, -0.00001270, 0.00001599, -0.00001687, 0.00000169, },
		{ 0.00000009, 0.00000126, -0.00001364, 0.00000189, 0.00000323, 0.00000137, -0.00001073, -0.00000436, 0.00000210, -0.00000006, 0.00000023, 0.00000161, -0.00000366, 0.00000208, -0.00000565, 0.00000133, -0.00000444, -0.00000206, -0.00000452, 0.00000066, -0.00000465, 0.00000729, -0.00001519, -0.00000078, 0.00000243, -0.00001610, 0.00002278, -0.00001635, 0.00001018, -0.00001098, -0.00000072, -0.00000450, 0.00001004, 0.00000183, -0.00000888, -0.00000144, -0.00000200, -0.00000578, -0.00000201, 0.00000177, -0.00000345, -0.00000595, -0.00001116, 0.00000274, -0.00001094, -0.00000358, -0.00002607, 0.00000010, -0.00001348, -0.00001289, -0.00004344, -0.00001220, -0.00008058, -0.00006119, -0.00043436, 0.00030834, 0.00043241, 0.00007236, 0.00047379, 0.00024831, -0.00045122, -0.00003068, -0.00007614, -0.00001982, -0.00004315, 0.00000078, -0.00002336, -0.00000848, -0.00001294, -0.00000312, -0.00000153, -0.00000565, -0.00001272, 0.00000137, -0.00000979, -0.00000383, -0.00000046, -0.00000348, 0.00000080, -0.00000817, 0.00000166, -0.00000231, 0.00000008, -0.00000469, -0.00000166, -0.00000793, 0.00001392, -0.00000799, 0.00000259, -0.00000792, 0.00000233, -0.00001320, -0.00000158, 0.00000844, -0.00000978, 0.00001755, -0.00001311, 0.00000801, 0.00000034, 0.00000474, 0.00000077, 0.00000550, -0.00000812, 0.00000367, -0.00000230, 0.00000294, 0.00000383, -0.00000127, 0.00000501, 0.00000053, 0.00000132, 0.00000658, -0.00000194, -0.00000536, 0.00001487, -0.00000388, 0.00000719, -0.00000157, -0.00001023, 0.00003606, -0.00003843, 0.00002846, -0.00002362, 0.00000786, -0.00001594, 0.00001289, -0.00000927, -0.00000275, },
		{ 0.00000592, -0.00000360, -0.00000509, -0.00000940, 0.00000954, -0.00000146, -0.00000440, -0.00001015, 0.00000404, -0.00000218, 0.00000230, 0.00000030, -0.00000167, 0.00000028, -0.00000316, -0.00000159, -0.00000169, -0.00000362, -0.00000265, -0.00000157, -0.00000320, 0.00000409, -0.00000590, -0.00001237, 0.00001179, -0.00002140, 0.00002002, -0.00000799, 0.00000034, 0.00000070, -0.00000959, 0.00000086, -0.00000046, 0.00001404, -0.00001507, 0.00000317, -0.00000813, 0.00000308, -0.00001085, 0.00000788, -0.00001067, 0.00000664, -0.00002216, 0.00001026, -0.00001971, 0.00000955, -0.00003176, 0.00000062, -0.00002004, 0.00000702, -0.00005544, 0.00000319, -0.00008557, -0.00000974, -0.00030056, -0.00021627, 0.00063597, 0.00014185, 0.00035000, 0.00050188, -0.00041373, -0.00015880, -0.00002949, -0.00005610, -0.00001956, -0.00003088, 0.00000141, -0.00002810, 0.00000279, -0.00002086, 0.00000922, -0.00001063, -0.00000498, -0.00001081, 0.00000201, -0.00001488, 0.00000805, -0.00001189, 0.00000744, -0.00001081, 0.00000366, -0.00000617, 0.00000351, -0.00000703, 0.00000200, -0.00001241, 0.00001270, -0.00000245, -0.00000146, -0.00000372, -0.00000287, -0.00000342, -0.00001352, 0.00001628, -0.00001480, 0.00001777, -0.00000880, 0.00000208, 0.00000259, 0.00000402, -0.00000002, 0.00000682, -0.00000633, 0.00000067, -0.00000074, 0.00000092, 0.00000320, 0.00000209, 0.00000019, 0.00000595, -0.00000345, 0.00000808, -0.00000308, -0.00000012, 0.00000098, 0.00001559, -0.00001652, 0.00002529, -0.00003198, 0.00003872, -0.00002033, 0.00000234, 0.00000495, -0.00001336, 0.00000099, -0.00000550, 0.00000626, -0.00001203, },
		{ 0.00000834, -0.00000684, 0.00000146, -0.00001581, 0.00000985, -0.00000092, -0.00000013, -0.00001368, 0.00000519, -0.00000456, 0.00000436, -0.00000141, 0.00000139, -0.00000314, 0.00000159, -0.00000682, 0.00000346, -0.00000848, 0.00000294, -0.00000715, 0.00000233, -0.00000433, 0.00000693, -0.00002345, 0.00001772, -0.00002223, 0.00001404, 0.00000033, -0.00000586, 0.00000455, -0.00000825, -0.00000218, -0.00000217, 0.00001293, -0.00000642, -0.00000554, -0.00000099, -0.00000382, -0.00000178, -0.00000261, -0.00000203, -0.00000136, -0.00000781, -0.00000756, -0.00000214, -0.00001008, -0.00000198, -0.00002916, 0.00000315, -0.00001616, -0.00001177, -0.00004658, -0.00001301, -0.00009126, -0.00004904, -0.00049481, 0.00032563, 0.00053304, 0.00009517, 0.00074830, -0.00029059, -0.00028851, -0.00002527, -0.00006221, -0.00001939, -0.00003762, -0.00000284, -0.00002198, -0.00000393, -0.00001503, -0.00000316, 0.00000066, -0.00000895, -0.00000948, -0.00000176, -0.00001046, 0.00000166, -0.00000517, -0.00000126, -0.00000112, -0.00000354, -0.00000041, -0.00000328, -0.00000041, -0.00000296, -0.00000643, 0.00000077, 0.00001044, -0.00001039, 0.00000508, -0.00001227, 0.00000775, -0.00002237, 0.00001874, -0.00001387, 0.00001227, -0.00000088, -0.00000415, 0.00000408, 0.00000306, 0.00000063, 0.00000519, -0.00000181, -0.00000323, 0.00000107, -0.00000093, 0.00000257, 0.00000315, 0.00000004, 0.00000511, -0.00000127, 0.00000488, -0.00000210, 0.00000175, -0.00000394, 0.00001757, -0.00001196, 0.00001103, -0.00000354, -0.00000464, 0.00002607, -0.00003302, 0.00002823, -0.00002320, 0.00000751, -0.00001269, 0.00001007, -0.00001361, },
		{ 0.00000199, 0.00000006, -0.00000347, -0.00000909, -0.00000143, 0.00000747, -0.00000311, -0.00000895, -0.00000171, 0.00000154, -0.00000253, 0.00000495, -0.00000298, 0.00000071, -0.00000112, -0.00000372, -0.00000027, -0.00000424, -0.00000149, -0.00000192, -0.00000250, -0.00000194, 0.00000564, -0.00001674, 0.00000488, -0.00000523, -0.00000721, 0.00002009, -0.00002035, 0.00001493, -0.00001484, 0.00000592, -0.00001380, 0.00001892, -0.00000718, -0.00000123, -0.00000777, 0.00000312, -0.00000841, 0.00000586, -0.00001219, 0.00000673, -0.00001135, -0.00000202, -0.00001022, -0.00000306, -0.00000705, -0.00001359, -0.00002019, 0.00000356, -0.00002052, -0.00002625, -0.00003246, -0.00005503, -0.00004996, -0.00026661, -0.00034231, 0.00082797, 0.00007495, 0.00073599, 0.00005784, -0.00048865, 0.00000541, -0.00010869, 0.00001455, -0.00006792, 0.00001189, -0.00003741, 0.00000961, -0.00002607, 0.00000276, -0.00001056, 0.00000706, -0.00002200, 0.00000609, -0.00001770, 0.00000635, -0.00000890, 0.00000120, -0.00000495, 0.00000244, -0.00000511, -0.00000003, -0.00000462, 0.00000169, -0.00000814, -0.00000217, 0.00001084, -0.00000603, 0.00000142, -0.00000862, 0.00000358, -0.00001406, 0.00000504, 0.00000139, -0.00000527, 0.00001537, -0.00001570, 0.00001101, -0.00000451, 0.00000824, -0.00000337, 0.00000798, -0.00001074, 0.00000676, -0.00000706, 0.00000722, -0.00000282, 0.00000709, -0.00000214, 0.00000619, -0.00000207, 0.00000403, -0.00000602, 0.00000557, 0.00000160, 0.00001135, -0.00001771, 0.00002639, -0.00002841, 0.00003335, -0.00002197, 0.00000823, 0.00000013, -0.00000886, -0.00000137, -0.00000050, -0.00000560, },
		{ -0.00000247, 0.00000739, -0.00000884, -0.00000203, -0.00001057, 0.00001162, -0.00000393, -0.00000389, -0.00000944, 0.00000875, -0.00000986, 0.00001041, -0.00000654, 0.00000467, -0.00000515, 0.00000207, -0.00000734, 0.00000382, -0.00001046, 0.00000770, -0.00001107, 0.00000548, -0.00000380, -0.00000044, -0.00001590, 0.00001705, -0.00003018, 0.00003775, -0.00003102, 0.00002244, -0.00002173, 0.00001619, -0.00002514, 0.00002404, -0.00001161, 0.00000979, -0.00002113, 0.00001693, -0.00002373, 0.00002352, -0.00002906, 0.00002097, -0.00002586, 0.00001869, -0.00003346, 0.00002096, -0.00003585, 0.00002749, -0.00006257, 0.00004068, -0.00005891, 0.00003445, -0.00010010, 0.00004018, -0.00016173, 0.00003318, -0.00065253, 0.00045869, 0.00051180, 0.00042650, 0.00057749, -0.00068042, 0.00002047, -0.00016818, 0.00005384, -0.00010075, 0.00003381, -0.00006517, 0.00003338, -0.00004650, 0.00002166, -0.00003574, 0.00003097, -0.00003834, 0.00001937, -0.00003163, 0.00001742, -0.00001953, 0.00001211, -0.00001798, 0.00001511, -0.00001469, 0.00000956, -0.00001571, 0.00001155, -0.00001460, 0.00000242, 0.00000105, 0.00000745, -0.00001003, 0.00000343, -0.00000987, 0.00000275, -0.00001440, 0.00001988, -0.00002307, 0.00002838, -0.00002268, 0.00001484, -0.00000996, 0.00001342, -0.00000891, 0.00001337, -0.00001350, 0.00000879, -0.00000979, 0.00000872, -0.00000557, 0.00000924, -0.00000322, 0.00000705, -0.00000280, 0.00000626, -0.00001082, 0.00000971, -0.00000427, 0.00001682, -0.00001729, 0.00001457, -0.00000331, -0.00000285, 0.00001565, -0.00002032, 0.00001816, -0.00001443, 0.00000067, -0.00000262, -0.00000413, },
		{ -0.00000298, 0.00000837, -0.00000493, -0.00000505, -0.00000801, 0.00000431, 0.00000340, -0.00000556, -0.00000910, 0.00000613, -0.00000538, 0.00000320, 0.00000142, -0.00000163, 0.00000028, -0.00000138, -0.00000440, 0.00000106, -0.00000771, 0.00000392, -0.00000539, -0.00000015, -0.00000133, 0.00000211, -0.00001881, 0.00001751, -0.00002717, 0.00002612, -0.00001241, 0.00000305, -0.00000395, 0.00000080, -0.00000772, 0.00000147, 0.00000803, -0.00000392, -0.00000674, 0.00000129, -0.00000803, 0.00000638, -0.00000708, -0.00000339, -0.00000388, -0.00000005, -0.00001132, -0.00000301, -0.00001408, 0.00000708, -0.00003010, -0.00000132, -0.00001782, 0.00000093, -0.00005008, -0.00001202, -0.00008694, -0.00001656, -0.00032713, -0.00036759, 0.00090070, 0.00027912, 0.00088837, -0.00057049, -0.00017307, -0.00009111, -0.00002784, -0.00003580, -0.00002121, -0.00002706, -0.00000524, -0.00001190, -0.00000648, -0.00001539, 0.00000735, -0.00001125, -0.00000381, -0.00001200, -0.00000344, -0.00000143, -0.00000287, -0.00000567, 0.00000151, -0.00000021, -0.00000118, -0.00000691, 0.00000102, -0.00000280, -0.00000732, 0.00000407, 0.00000638, -0.00000659, 0.00000177, -0.00000979, 0.00000466, -0.00001691, 0.00002003, -0.00002052, 0.00001971, -0.00000965, 0.00000186, -0.00000010, 0.00000375, -0.00000023, 0.00000480, -0.00000359, -0.00000037, -0.00000141, -0.00000060, 0.00000252, 0.00000012, 0.00000579, -0.00000045, 0.00000331, 0.00000238, -0.00000655, 0.00000043, 0.00000604, 0.00000367, 0.00000276, -0.00001186, 0.00002213, -0.00001911, 0.00001654, -0.00000700, -0.00000154, 0.00000731, -0.00001431, 0.00000701, -0.00000966, },
		{ -0.00001194, 0.00001032, 0.00000056, -0.00000844, -0.00000494, -0.00000234, 0.00000798, -0.00000577, -0.00000690, -0.00000064, 0.00000408, -0.00000774, 0.00001092, -0.00000876, 0.00000697, -0.00000721, 0.00000261, -0.00000696, 0.00000158, -0.00000763, 0.00000765, -0.00001179, 0.00000773, -0.00000531, -0.00000769, 0.00000107, -0.00000486, -0.00000508, 0.00002240, -0.00002860, 0.00002467, -0.00002695, 0.00002295, -0.00003116, 0.00003561, -0.00002847, 0.00002147, -0.00002925, 0.00002428, -0.00003010, 0.00003437, -0.00004400, 0.00003406, -0.00003999, 0.00003588, -0.00005201, 0.00003589, -0.00005006, 0.00004235, -0.00007842, 0.00005870, -0.00008206, 0.00006110, -0.00013472, 0.00007207, -0.00020577, 0.00008123, -0.00078305, 0.00054777, 0.00061283, 0.00083368, -0.00009609, -0.00055550, 0.00009055, -0.00019740, 0.00008979, -0.00012297, 0.00005552, -0.00008684, 0.00005793, -0.00006553, 0.00003749, -0.00004831, 0.00004087, -0.00004732, 0.00002845, -0.00004410, 0.00003352, -0.00003360, 0.00002316, -0.00002833, 0.00002694, -0.00002272, 0.00001452, -0.00002227, 0.00001896, -0.00002470, 0.00001569, -0.00000573, 0.00000781, -0.00001022, 0.00000154, -0.00000615, -0.00000546, 0.00000527, -0.00000191, -0.00000402, 0.00001453, -0.00001864, 0.00001598, -0.00001134, 0.00001288, -0.00000764, 0.00000913, -0.00001174, 0.00000935, -0.00001148, 0.00001127, -0.00000849, 0.00001196, -0.00000382, 0.00000546, 0.00000083, -0.00000136, -0.00000947, 0.00001260, -0.00000114, 0.00000799, -0.00001245, 0.00000980, 0.00000677, -0.00001626, 0.00002227, -0.00002020, 0.00001596, -0.00001212, 0.00000137, -0.00000184, },
		{ -0.00002351, 0.00001101, 0.00000330, -0.00000449, -0.00000967, -0.00000031, 0.00000327, 0.00000101, -0.00000951, -0.00000221, 0.00000594, -0.00000770, 0.00000743, -0.00000332, 0.00000204, -0.00000237, -0.00000044, -0.00000465, 0.00000026, -0.00000767, 0.00000695, -0.00000835, 0.00000290, -0.00000137, -0.00000646, -0.00000515, 0.00000583, -0.00002042, 0.00003464, -0.00003298, 0.00002495, -0.00002648, 0.00002312, -0.00002907, 0.00002775, -0.00001983, 0.00001625, -0.00002440, 0.00001997, -0.00002784, 0.00003164, -0.00003499, 0.00002217, -0.00003137, 0.00003002, -0.00004128, 0.00002373, -0.00004229, 0.00003791, -0.00006399, 0.00003737, -0.00006332, 0.00005073, -0.00010700, 0.00004396, -0.00015681, 0.00005967, -0.00044829, -0.00033441, 0.00090639, 0.00084262, 0.00038074, -0.00078362, 0.00010124, -0.00023049, 0.00008805, -0.00011876, 0.00005223, -0.00009192, 0.00005399, -0.00005912, 0.00003429, -0.00005053, 0.00003871, -0.00004133, 0.00002597, -0.00004337, 0.00002777, -0.00002714, 0.00001846, -0.00002561, 0.00002109, -0.00001472, 0.00001110, -0.00002113, 0.00001522, -0.00001797, 0.00000708, -0.00000093, 0.00000594, -0.00000710, 0.00000014, -0.00000608, -0.00000335, -0.00000155, 0.00000920, -0.00001758, 0.00002449, -0.00002311, 0.00001660, -0.00001159, 0.00001186, -0.00000715, 0.00000957, -0.00001164, 0.00000918, -0.00001075, 0.00000821, -0.00000519, 0.00000622, 0.00000257, 0.00000075, 0.00000340, 0.00000074, -0.00001197, 0.00000672, 0.00000745, -0.00000059, 0.00000114, -0.00001070, 0.00002461, -0.00002030, 0.00000844, 0.00000589, -0.00001320, 0.00001632, -0.00002017, 0.00001551, },
		{ -0.00001409, -0.00000341, 0.00001118, -0.00000177, -0.00001144, -0.00000121, 0.00000119, 0.00000402, -0.00000885, -0.00000403, 0.00000434, -0.00000184, -0.00000148, 0.00000570, -0.00000562, 0.00000515, -0.00000676, 0.00000216, -0.00000653, -0.00000065, -0.00000263, 0.00000376, -0.00000869, 0.00000814, -0.00001154, -0.00000260, 0.00000414, -0.00001742, 0.00002258, -0.00001140, 0.00000067, -0.00000193, -0.00000268, 0.00000034, -0.00000530, 0.00001213, -0.00001383, 0.00000713, -0.00001223, 0.00000518, -0.00000628, 0.00001033, -0.00002311, 0.00001126, -0.00001614, 0.00001365, -0.00003232, 0.00001322, -0.00002423, 0.00001319, -0.00004460, 0.00002021, -0.00004240, 0.00001821, -0.00009436, 0.00002109, -0.00015343, 0.00000773, -0.00077521, 0.00042565, 0.00120537, 0.00068882, -0.00070003, -0.00014408, -0.00008461, -0.00006021, -0.00000516, -0.00004006, -0.00001428, -0.00002677, 0.00001163, -0.00002506, 0.00000210, -0.00001550, 0.00000854, -0.00001562, -0.00000296, -0.00001577, 0.00001218, -0.00001601, 0.00000633, -0.00001234, 0.00001562, -0.00001183, 0.00000130, -0.00000990, 0.00000702, -0.00001607, 0.00001647, -0.00000843, 0.00000717, -0.00001075, 0.00000286, -0.00000945, -0.00000038, 0.00001041, -0.00001687, 0.00001717, -0.00001094, 0.00000288, 0.00000063, 0.00000035, 0.00000196, 0.00000218, -0.00000422, 0.00000163, -0.00000238, -0.00000129, 0.00000295, -0.00000211, 0.00000799, -0.00000069, 0.00000220, 0.00000462, -0.00001185, -0.00000113, 0.00001151, 0.00000096, -0.00000110, -0.00000479, 0.00000440, 0.00001582, -0.00003325, 0.00003921, -0.00003197, 0.00002357, -0.00001784, 0.00001066, },
		{ 0.00000987, -0.00002392, 0.00001918, -0.00000268, -0.00000423, -0.00001116, 0.00000752, -0.00000174, -0.00000060, -0.00001112, 0.00000681, -0.00000140, -0.00000173, 0.00000371, -0.00000195, 0.00000150, -0.00000228, -0.00000151, -0.00000266, -0.00000404, -0.00000144, 0.00000275, -0.00000506, 0.00000221, -0.00000252, -0.00001180, 0.00001112, -0.00001929, 0.00001428, 0.00000258, -0.00001122, 0.00000798, -0.00001282, 0.00001126, -0.00001555, 0.00001936, -0.00001917, 0.00001259, -0.00001685, 0.00001047, -0.00001535, 0.00002136, -0.00002822, 0.00001360, -0.00002306, 0.00002382, -0.00003622, 0.00001457, -0.00003026, 0.00002332, -0.00004646, 0.00001858, -0.00004634, 0.00003273, -0.00009125, 0.00001852, -0.00013014, 0.00001530, -0.00042620, -0.00052612, 0.00129684, 0.00123917, -0.00055632, -0.00039055, 0.00000303, -0.00014596, 0.00003698, -0.00007138, 0.00001520, -0.00006198, 0.00003390, -0.00004028, 0.00001728, -0.00003320, 0.00002045, -0.00002382, 0.00000978, -0.00003126, 0.00002165, -0.00002224, 0.00001260, -0.00001998, 0.00001858, -0.00001049, 0.00000427, -0.00001570, 0.00001085, -0.00001755, 0.00001342, -0.00000426, 0.00000364, -0.00000505, -0.00000283, -0.00000239, -0.00000979, 0.00001745, -0.00001715, 0.00000994, -0.00000131, -0.00000580, 0.00000638, -0.00000367, 0.00000351, 0.00000116, -0.00000196, -0.00000172, 0.00000175, -0.00000498, 0.00000410, -0.00000235, 0.00000454, 0.00000499, -0.00000311, 0.00000903, -0.00001024, -0.00000556, 0.00000585, 0.00001247, -0.00001175, 0.00000884, -0.00001664, 0.00003450, -0.00003488, 0.00001845, 0.00000253, -0.00001256, 0.00001492, -0.00001481, },
		{ 0.00001245, -0.00002254, 0.00000954, 0.00000403, -0.00000006, -0.00001570, 0.00000738, -0.00000189, 0.00000208, -0.00001251, 0.00000502, 0.00000003, -0.00000003, -0.00000067, 0.00000320, -0.00000341, 0.00000356, -0.00000687, 0.00000351, -0.00000981, 0.00000317, -0.00000362, 0.00000417, -0.00000793, 0.00000894, -0.00002125, 0.00001621, -0.00001822, 0.00000606, 0.00001012, -0.00001222, 0.00000567, -0.00000939, 0.00000578, -0.00000740, 0.00000850, -0.00000632, -0.00000168, -0.00000123, -0.00000439, -0.00000203, 0.00000523, -0.00000457, -0.00000990, -0.00000327, 0.00000048, -0.00000311, -0.00001937, 0.00000120, -0.00001229, 0.00000060, -0.00002893, -0.00000021, -0.00002071, -0.00000725, -0.00007434, -0.00000308, -0.00013662, -0.00002730, -0.00083746, 0.00035909, 0.00216000, -0.00031861, -0.00059201, -0.00001903, -0.00013699, -0.00000687, -0.00003991, -0.00000707, -0.00004406, 0.00000349, -0.00001191, -0.00000510, -0.00001342, -0.00000291, -0.00000392, -0.00000266, -0.00001844, 0.00000235, -0.00000328, -0.00000381, -0.00000483, -0.00000023, 0.00000736, -0.00000635, -0.00000568, -0.00000225, -0.00000340, -0.00000318, 0.00001124, -0.00000968, 0.00000822, -0.00001393, 0.00000846, -0.00001981, 0.00002093, -0.00001237, 0.00000075, 0.00000614, -0.00001007, 0.00000703, -0.00000281, 0.00000164, 0.00000157, -0.00000012, -0.00000465, 0.00000430, -0.00000580, 0.00000272, -0.00000102, 0.00000143, 0.00000663, -0.00000157, 0.00000507, -0.00000157, -0.00001189, 0.00000213, 0.00001427, -0.00000569, 0.00000022, -0.00000530, 0.00000893, 0.00000934, -0.00003349, 0.00004465, -0.00003701, 0.00002577, -0.00001720, },
		{ -0.00000677, -0.00000313, -0.00001104, 0.00001521, -0.00000317, -0.00000710, -0.00000634, 0.00000977, -0.00000657, -0.00000255, -0.00000712, 0.00001022, -0.00000732, 0.00000615, -0.00000456, 0.00000489, -0.00000419, 0.00000188, -0.00000510, 0.00000031, -0.00000821, 0.00000619, -0.00000497, 0.00000278, -0.00000251, -0.00000551, -0.00000529, 0.00000855, -0.00002349, 0.00003425, -0.00002913, 0.00002104, -0.00002446, 0.00001932, -0.00002052, 0.00002174, -0.00001877, 0.00001085, -0.00001579, 0.00001360, -0.00002195, 0.00002293, -0.00001988, 0.00001172, -0.00002895, 0.00002204, -0.00002191, 0.00000806, -0.00003266, 0.00002123, -0.00003390, 0.00001881, -0.00005512, 0.00003322, -0.00005669, 0.00000343, -0.00009505, 0.00000034, -0.00016421, -0.00023183, -0.00103758, 0.00244202, 0.00069357, -0.00107467, 0.00008850, -0.00024750, 0.00005448, -0.00011165, 0.00005649, -0.00009396, 0.00003521, -0.00004696, 0.00003070, -0.00004471, 0.00002333, -0.00003357, 0.00002779, -0.00004073, 0.00001715, -0.00001992, 0.00001393, -0.00002116, 0.00001244, -0.00000819, 0.00001351, -0.00002087, 0.00000884, -0.00001414, 0.00000538, 0.00000100, 0.00000293, -0.00000476, 0.00000083, -0.00000607, -0.00000339, -0.00000220, 0.00001512, -0.00002452, 0.00002514, -0.00002429, 0.00001731, -0.00001209, 0.00001104, -0.00000962, 0.00001209, -0.00001592, 0.00001310, -0.00001221, 0.00000807, -0.00000742, 0.00000732, -0.00000206, 0.00000935, -0.00000616, 0.00001093, -0.00001898, 0.00000355, 0.00000377, 0.00001306, -0.00001819, 0.00001414, -0.00001771, 0.00003567, -0.00004329, 0.00002944, -0.00000548, -0.00000767, 0.00001079, },
		{ -0.00000643, 0.00000029, -0.00001509, 0.00001231, -0.00000074, 0.00000034, -0.00001602, 0.00001482, -0.00000895, 0.00000238, -0.00001439, 0.00001541, -0.00001116, 0.00001146, -0.00001096, 0.00001134, -0.00001064, 0.00001010, -0.00001402, 0.00001145, -0.00002022, 0.00001691, -0.00001662, 0.00001669, -0.00001747, 0.00001430, -0.00002962, 0.00003439, -0.00004682, 0.00004928, -0.00003844, 0.00003101, -0.00003486, 0.00002937, -0.00003233, 0.00003444, -0.00003050, 0.00002456, -0.00003417, 0.00003547, -0.00004431, 0.00004351, -0.00004157, 0.00004063, -0.00005821, 0.00004744, -0.00005113, 0.00004898, -0.00007710, 0.00006557, -0.00008439, 0.00008162, -0.00011810, 0.00009586, -0.00012868, 0.00011061, -0.00021678, 0.00016542, -0.00036112, 0.00024498, -0.00136893, 0.00101438, 0.00262164, -0.00159759, 0.00014764, -0.00036387, 0.00012960, -0.00020181, 0.00012543, -0.00014565, 0.00007654, -0.00009677, 0.00007696, -0.00008424, 0.00006020, -0.00007285, 0.00006155, -0.00006451, 0.00003818, -0.00004550, 0.00003928, -0.00004415, 0.00003331, -0.00003260, 0.00003820, -0.00003848, 0.00002483, -0.00003131, 0.00002083, -0.00001724, 0.00002379, -0.00002547, 0.00002210, -0.00002646, 0.00001916, -0.00002925, 0.00004084, -0.00004277, 0.00003603, -0.00003145, 0.00002184, -0.00001726, 0.00001745, -0.00001731, 0.00001916, -0.00002062, 0.00001503, -0.00001315, 0.00000977, -0.00001050, 0.00000990, -0.00000672, 0.00001369, -0.00000877, 0.00001310, -0.00001592, -0.00000020, -0.00000245, 0.00001976, -0.00001672, 0.00000830, -0.00000987, 0.00001597, -0.00000491, -0.00001936, 0.00003508, -0.00002994, 0.00001815, },
		{ 0.00001732, -0.00001167, -0.00000498, -0.00000113, 0.00000790, -0.00000178, -0.00000933, 0.00000147, 0.00000559, -0.00000848, -0.00000481, 0.00000225, 0.00000347, -0.00000248, 0.00000398, -0.00000476, 0.00000547, -0.00000479, 0.00000127, -0.00000330, -0.00000377, -0.00000228, 0.00000253, -0.00000236, 0.00000245, -0.00000284, -0.00001237, 0.00001246, -0.00001711, 0.00000972, 0.00000538, -0.00001182, 0.00000876, -0.00001527, 0.00001176, -0.00001197, 0.00001926, -0.00002339, 0.00001143, -0.00001254, 0.00000885, -0.00001479, 0.00001847, -0.00001917, 0.00000968, -0.00002724, 0.00002199, -0.00002306, 0.00000569, -0.00002741, 0.00001374, -0.00002507, 0.00001087, -0.00004931, 0.00002656, -0.00005322, -0.00000622, -0.00008548, -0.00001935, -0.00016467, -0.00027835, -0.00133291, 0.00391633, -0.00088836, -0.00058947, 0.00001467, -0.00017699, 0.00002019, -0.00007876, 0.00003263, -0.00007448, 0.00002456, -0.00003678, 0.00002137, -0.00003379, 0.00001161, -0.00002310, 0.00001836, -0.00003453, 0.00001644, -0.00001990, 0.00001248, -0.00001959, 0.00001363, -0.00000724, 0.00000947, -0.00001787, 0.00000704, -0.00001651, 0.00001373, -0.00000298, 0.00000095, -0.00000314, -0.00000203, -0.00000166, -0.00001133, 0.00001854, -0.00001314, 0.00000308, 0.00000102, -0.00000963, 0.00001007, -0.00000662, 0.00000496, -0.00000325, 0.00000262, -0.00000821, 0.00000772, -0.00000828, 0.00000582, -0.00000629, 0.00000709, -0.00000066, 0.00000588, -0.00000040, 0.00000013, -0.00001319, 0.00000338, 0.00000783, 0.00000372, -0.00001256, 0.00000997, -0.00000870, 0.00002053, -0.00003149, 0.00002342, -0.00000126, -0.00001277, },
		{ 0.00001510, -0.00000131, -0.00001008, -0.00000185, 0.00000700, -0.00000219, 0.00000022, -0.00001264, 0.00001790, -0.00001690, 0.00000493, -0.00001319, 0.00002064, -0.00001982, 0.00002333, -0.00002474, 0.00002492, -0.00002407, 0.00002260, -0.00002608, 0.00002266, -0.00003191, 0.00003255, -0.00003370, 0.00003538, -0.00003473, 0.00002389, -0.00003220, 0.00003619, -0.00005073, 0.00006645, -0.00007046, 0.00006783, -0.00007480, 0.00007156, -0.00007583, 0.00008574, -0.00008686, 0.00007573, -0.00008412, 0.00008756, -0.00009853, 0.00010482, -0.00010898, 0.00010933, -0.00013121, 0.00012607, -0.00013459, 0.00013381, -0.00016634, 0.00016186, -0.00019032, 0.00020036, -0.00025159, 0.00024443, -0.00029956, 0.00030107, -0.00044295, 0.00043632, -0.00072781, 0.00072277, -0.00233247, 0.00268883, 0.00174493, -0.00208014, 0.00072494, -0.00072053, 0.00042348, -0.00043096, 0.00032215, -0.00031601, 0.00023117, -0.00023114, 0.00019636, -0.00018997, 0.00015636, -0.00016267, 0.00014568, -0.00014574, 0.00011762, -0.00011802, 0.00010472, -0.00010530, 0.00009233, -0.00008475, 0.00008489, -0.00008460, 0.00007002, -0.00007734, 0.00006655, -0.00005277, 0.00005159, -0.00005234, 0.00004523, -0.00004542, 0.00003203, -0.00003057, 0.00004036, -0.00004873, 0.00004890, -0.00005362, 0.00004758, -0.00004038, 0.00003715, -0.00003481, 0.00003322, -0.00003657, 0.00003213, -0.00003004, 0.00002647, -0.00002604, 0.00002381, -0.00001720, 0.00002081, -0.00001324, 0.00001437, -0.00002341, 0.00000983, -0.00000556, 0.00001827, -0.00001932, 0.00001025, -0.00000609, 0.00000899, -0.00000483, -0.00001356, 0.00002921, -0.00002588, },
		{ -0.00002219, 0.00002786, -0.00002474, 0.00000589, -0.00000121, 0.00000315, -0.00000033, -0.00000938, 0.00000920, -0.00000456, -0.00000445, -0.00000855, 0.00001449, -0.00001140, 0.00001432, -0.00001314, 0.00001119, -0.00001010, 0.00001016, -0.00001401, 0.00001236, -0.00002148, 0.00001979, -0.00002022, 0.00002089, -0.00001833, 0.00001220, -0.00002629, 0.00003280, -0.00004645, 0.00005565, -0.00005249, 0.00004624, -0.00004976, 0.00004403, -0.00004867, 0.00005476, -0.00004841, 0.00003769, -0.00005054, 0.00005354, -0.00006122, 0.00006312, -0.00006421, 0.00006428, -0.00007879, 0.00006704, -0.00007673, 0.00007788, -0.00010230, 0.00009071, -0.00011659, 0.00011956, -0.00014915, 0.00012709, -0.00017247, 0.00016649, -0.00026324, 0.00021819, -0.00040799, 0.00031446, -0.00104209, -0.00057323, 0.00456578, -0.00255052, 0.00050498, -0.00058784, 0.00028324, -0.00032805, 0.00021291, -0.00021517, 0.00014369, -0.00016223, 0.00012832, -0.00012582, 0.00010107, -0.00011388, 0.00009378, -0.00009249, 0.00006994, -0.00007688, 0.00006447, -0.00006654, 0.00005523, -0.00005282, 0.00005435, -0.00005178, 0.00004123, -0.00005030, 0.00003620, -0.00002673, 0.00003166, -0.00003396, 0.00002791, -0.00002923, 0.00002068, -0.00002709, 0.00003975, -0.00004505, 0.00004206, -0.00004380, 0.00003320, -0.00002610, 0.00002464, -0.00002365, 0.00002264, -0.00002548, 0.00001978, -0.00001858, 0.00001632, -0.00001606, 0.00001270, -0.00000801, 0.00001176, -0.00000526, 0.00001001, -0.00001686, 0.00000298, -0.00000445, 0.00001348, -0.00000714, -0.00000333, 0.00000539, -0.00000327, 0.00000600, -0.00001351, 0.00000878, 0.00000977, },
		{ -0.00003243, 0.00002320, -0.00001075, -0.00000512, 0.00000263, 0.00000265, -0.00000364, 0.00000209, -0.00000777, 0.00001452, -0.00002046, 0.00000658, -0.00000715, 0.00001534, -0.00001541, 0.00002100, -0.00002497, 0.00002622, -0.00002633, 0.00002465, -0.00002774, 0.00002246, -0.00002889, 0.00003114, -0.00003358, 0.00003849, -0.00004129, 0.00002705, -0.00002566, 0.00002008, -0.00002235, 0.00003526, -0.00004713, 0.00004871, -0.00005835, 0.00005685, -0.00005934, 0.00007414, -0.00008405, 0.00007198, -0.00007788, 0.00007991, -0.00008742, 0.00009509, -0.00010335, 0.00010293, -0.00012470, 0.00012175, -0.00013377, 0.00013127, -0.00016086, 0.00015418, -0.00018124, 0.00019152, -0.00024467, 0.00023853, -0.00030139, 0.00030238, -0.00044960, 0.00044384, -0.00077921, 0.00079882, -0.00292679, 0.00464001, -0.00028845, -0.00132516, 0.00051874, -0.00057474, 0.00034019, -0.00036236, 0.00026902, -0.00027170, 0.00020109, -0.00020609, 0.00017495, -0.00016910, 0.00013592, -0.00014347, 0.00012686, -0.00012968, 0.00010843, -0.00011169, 0.00009820, -0.00009912, 0.00009017, -0.00008190, 0.00007862, -0.00007869, 0.00006500, -0.00007643, 0.00007417, -0.00005975, 0.00005401, -0.00005569, 0.00004871, -0.00004910, 0.00003504, -0.00002144, 0.00001914, -0.00002323, 0.00002231, -0.00003343, 0.00003516, -0.00003130, 0.00002865, -0.00002665, 0.00002169, -0.00002549, 0.00002284, -0.00002229, 0.00002090, -0.00002222, 0.00002185, -0.00001517, 0.00001734, -0.00000784, 0.00000254, -0.00001480, 0.00000833, -0.00000195, 0.00000788, -0.00001090, 0.00000442, 0.00000402, -0.00000977, 0.00001284, -0.00002468, 0.00003643, },
		{ 0.00001111, -0.00002028, 0.00002476, -0.00002629, 0.00001303, -0.00000275, -0.00000246, 0.00000473, -0.00000996, 0.00001534, -0.00001771, 0.00000539, -0.00001335, 0.00002412, -0.00002442, 0.00003113, -0.00003320, 0.00003260, -0.00003275, 0.00003296, -0.00003699, 0.00003459, -0.00004332, 0.00004570, -0.00004872, 0.00005269, -0.00005147, 0.00003978, -0.00004531, 0.00004710, -0.00005744, 0.00007471, -0.00008630, 0.00008716, -0.00009577, 0.00009433, -0.00010131, 0.00011654, -0.00011945, 0.00010769, -0.00012074, 0.00012627, -0.00013612, 0.00014581, -0.00015555, 0.00015937, -0.00018009, 0.00017730, -0.00019620, 0.00020189, -0.00023266, 0.00023107, -0.00026992, 0.00028955, -0.00033991, 0.00034195, -0.00042112, 0.00044748, -0.00059472, 0.00061751, -0.00093894, 0.00101989, -0.00231699, 0.00139867, 0.00357349, -0.00310802, 0.00121749, -0.00106720, 0.00068124, -0.00064501, 0.00048448, -0.00045154, 0.00035907, -0.00035250, 0.00029872, -0.00027831, 0.00023961, -0.00024161, 0.00021222, -0.00020568, 0.00018082, -0.00018054, 0.00015979, -0.00015600, 0.00014191, -0.00013185, 0.00012567, -0.00012072, 0.00010820, -0.00011778, 0.00010544, -0.00008761, 0.00008366, -0.00008364, 0.00007347, -0.00007026, 0.00005527, -0.00004668, 0.00005022, -0.00005656, 0.00005703, -0.00006708, 0.00006214, -0.00005443, 0.00005024, -0.00004690, 0.00004148, -0.00004474, 0.00003994, -0.00003884, 0.00003682, -0.00003578, 0.00003132, -0.00002355, 0.00002360, -0.00001367, 0.00001306, -0.00002459, 0.00001479, -0.00000987, 0.00001198, -0.00000863, -0.00000105, 0.00000912, -0.00001296, 0.00001019, -0.00001078, 0.00000503, },
		{ 0.00004268, -0.00003556, 0.00002674, -0.00001922, 0.00000521, 0.00000115, -0.00000168, -0.00000162, 0.00000243, -0.00000031, 0.00000219, -0.00001330, 0.00000194, 0.00000479, -0.00000007, 0.00000315, 0.00000051, -0.00000439, 0.00000536, -0.00000593, 0.00000423, -0.00000761, 0.00000133, -0.00000300, 0.00000343, -0.00000419, 0.00001027, -0.00001937, 0.00001036, -0.00000744, -0.00000288, 0.00001519, -0.00001763, 0.00001023, -0.00001187, 0.00000530, -0.00000918, 0.00001500, -0.00000528, -0.00000880, -0.00000349, 0.00000085, -0.00000251, 0.00000365, -0.00000306, -0.00000277, -0.00000386, -0.00000982, 0.00000017, -0.00000939, -0.00000008, -0.00001830, -0.00000177, -0.00000914, 0.00000035, -0.00003230, -0.00000102, -0.00003766, -0.00000197, -0.00009255, -0.00000504, -0.00023338, -0.00001065, -0.00223463, 0.00551820, -0.00223463, -0.00001065, -0.00023338, -0.00000504, -0.00009255, -0.00000197, -0.00003766, -0.00000102, -0.00003230, 0.00000035, -0.00000914, -0.00000177, -0.00001830, -0.00000008, -0.00000939, 0.00000017, -0.00000982, -0.00000386, -0.00000277, -0.00000306, 0.00000365, -0.00000251, 0.00000085, -0.00000349, -0.00000880, -0.00000528, 0.00001500, -0.00000918, 0.00000530, -0.00001187, 0.00001023, -0.00001763, 0.00001519, -0.00000288, -0.00000744, 0.00001036, -0.00001937, 0.00001027, -0.00000419, 0.00000343, -0.00000300, 0.00000133, -0.00000761, 0.00000423, -0.00000593, 0.00000536, -0.00000439, 0.00000051, 0.00000315, -0.00000007, 0.00000479, 0.00000194, -0.00001330, 0.00000219, -0.00000031, 0.00000243, -0.00000162, -0.00000168, 0.00000115, 0.00000521, -0.00001922, 0.00002674, -0.00003556, },
		{ 0.00001111, 0.00000503, -0.00001078, 0.00001019, -0.00001296, 0.00000912, -0.00000105, -0.00000863, 0.00001198, -0.00000987, 0.00001479, -0.00002459, 0.00001306, -0.00001367, 0.00002360, -0.00002355, 0.00003132, -0.00003578, 0.00003682, -0.00003884, 0.00003994, -0.00004474, 0.00004148, -0.00004690, 0.00005024, -0.00005443, 0.00006214, -0.00006708, 0.00005703, -0.00005656, 0.00005022, -0.00004668, 0.00005527, -0.00007026, 0.00007347, -0.00008364, 0.00008366, -0.00008761, 0.00010544, -0.00011778, 0.00010820, -0.00012072, 0.00012567, -0.00013185, 0.00014191, -0.00015600, 0.00015979, -0.00018054, 0.00018082, -0.00020568, 0.00021222, -0.00024161, 0.00023961, -0.00027831, 0.00029872, -0.00035250, 0.00035907, -0.00045154, 0.00048448, -0.00064501, 0.00068124, -0.00106720, 0.00121749, -0.00310802, 0.00357349, 0.00139867, -0.00231699, 0.00101989, -0.00093894, 0.00061751, -0.00059472, 0.00044748, -0.00042112, 0.00034195, -0.00033991, 0.00028955, -0.00026992, 0.00023107, -0.00023266, 0.00020189, -0.00019620, 0.00017730, -0.00018009, 0.00015937, -0.00015555, 0.00014581, -0.00013612, 0.00012627, -0.00012074, 0.00010769, -0.00011945, 0.00011654, -0.00010131, 0.00009433, -0.00009577, 0.00008716, -0.00008630, 0.00007471, -0.00005744, 0.00004710, -0.00004531, 0.00003978, -0.00005147, 0.00005269, -0.00004872, 0.00004570, -0.00004332, 0.00003459, -0.00003699, 0.00003296, -0.00003275, 0.00003260, -0.00003320, 0.00003113, -0.00002442, 0.00002412, -0.00001335, 0.00000539, -0.00001771, 0.00001534, -0.00000996, 0.00000473, -0.00000246, -0.00000275, 0.00001303, -0.00002629, 0.00002476, -0.00002028, },
		{ -0.00003243, 0.00003643, -0.00002468, 0.00001284, -0.00000977, 0.00000402, 0.00000442, -0.00001090, 0.00000788, -0.00000195, 0.00000833, -0.00001480, 0.00000254, -0.00000784, 0.00001734, -0.00001517, 0.00002185, -0.00002222, 0.00002090, -0.00002229, 0.00002284, -0.00002549, 0.00002169, -0.00002665, 0.00002865, -0.00003130, 0.00003516, -0.00003343, 0.00002231, -0.00002323, 0.00001914, -0.00002144, 0.00003504, -0.00004910, 0.00004871, -0.00005569, 0.00005401, -0.00005975, 0.00007417, -0.00007643, 0.00006500, -0.00007869, 0.00007862, -0.00008190, 0.00009017, -0.00009912, 0.00009820, -0.00011169, 0.00010843, -0.00012968, 0.00012686, -0.00014347, 0.00013592, -0.00016910, 0.00017495, -0.00020609, 0.00020109, -0.00027170, 0.00026902, -0.00036236, 0.00034019, -0.00057474, 0.00051874, -0.00132516, -0.00028845, 0.00464001, -0.00292679, 0.00079882, -0.00077921, 0.00044384, -0.00044960, 0.00030238, -0.00030139, 0.00023853, -0.00024467, 0.00019152, -0.00018124, 0.00015418, -0.00016086, 0.00013127, -0.00013377, 0.00012175, -0.00012470, 0.00010293, -0.00010335, 0.00009509, -0.00008742, 0.00007991, -0.00007788, 0.00007198, -0.00008405, 0.00007414, -0.00005934, 0.00005685, -0.00005835, 0.00004871, -0.00004713, 0.00003526, -0.00002235, 0.00002008, -0.00002566, 0.00002705, -0.00004129, 0.00003849, -0.00003358, 0.00003114, -0.00002889, 0.00002246, -0.00002774, 0.00002465, -0.00002633, 0.00002622, -0.00002497, 0.00002100, -0.00001541, 0.00001534, -0.00000715, 0.00000658, -0.00002046, 0.00001452, -0.00000777, 0.00000209, -0.00000364, 0.00000265, 0.00000263, -0.00000512, -0.00001075, 0.00002320, },
		{ -0.00002219, 0.00000977, 0.00000878, -0.00001351, 0.00000600, -0.00000327, 0.00000539, -0.00000333, -0.00000714, 0.00001348, -0.00000445, 0.00000298, -0.00001686, 0.00001001, -0.00000526, 0.00001176, -0.00000801, 0.00001270, -0.00001606, 0.00001632, -0.00001858, 0.00001978, -0.00002548, 0.00002264, -0.00002365, 0.00002464, -0.00002610, 0.00003320, -0.00004380, 0.00004206, -0.00004505, 0.00003975, -0.00002709, 0.00002068, -0.00002923, 0.00002791, -0.00003396, 0.00003166, -0.00002673, 0.00003620, -0.00005030, 0.00004123, -0.00005178, 0.00005435, -0.00005282, 0.00005523, -0.00006654, 0.00006447, -0.00007688, 0.00006994, -0.00009249, 0.00009378, -0.00011388, 0.00010107, -0.00012582, 0.00012832, -0.00016223, 0.00014369, -0.00021517, 0.00021291, -0.00032805, 0.00028324, -0.00058784, 0.00050498, -0.00255052, 0.00456578, -0.00057323, -0.00104209, 0.00031446, -0.00040799, 0.00021819, -0.00026324, 0.00016649, -0.00017247, 0.00012709, -0.00014915, 0.00011956, -0.00011659, 0.00009071, -0.00010230, 0.00007788, -0.00007673, 0.00006704, -0.00007879, 0.00006428, -0.00006421, 0.00006312, -0.00006122, 0.00005354, -0.00005054, 0.00003769, -0.00004841, 0.00005476, -0.00004867, 0.00004403, -0.00004976, 0.00004624, -0.00005249, 0.00005565, -0.00004645, 0.00003280, -0.00002629, 0.00001220, -0.00001833, 0.00002089, -0.00002022, 0.00001979, -0.00002148, 0.00001236, -0.00001401, 0.00001016, -0.00001010, 0.00001119, -0.00001314, 0.00001432, -0.00001140, 0.00001449, -0.00000855, -0.00000445, -0.00000456, 0.00000920, -0.00000938, -0.00000033, 0.00000315, -0.00000121, 0.00000589, -0.00002474, 0.00002786, },
		{ 0.00001510, -0.00002588, 0.00002921, -0.00001356, -0.00000483, 0.00000899, -0.00000609, 0.00001025, -0.00001932, 0.00001827, -0.00000556, 0.00000983, -0.00002341, 0.00001437, -0.00001324, 0.00002081, -0.00001720, 0.00002381, -0.00002604, 0.00002647, -0.00003004, 0.00003213, -0.00003657, 0.00003322, -0.00003481, 0.00003715, -0.00004038, 0.00004758, -0.00005362, 0.00004890, -0.00004873, 0.00004036, -0.00003057, 0.00003203, -0.00004542, 0.00004523, -0.00005234, 0.00005159, -0.00005277, 0.00006655, -0.00007734, 0.00007002, -0.00008460, 0.00008489, -0.00008475, 0.00009233, -0.00010530, 0.00010472, -0.00011802, 0.00011762, -0.00014574, 0.00014568, -0.00016267, 0.00015636, -0.00018997, 0.00019636, -0.00023114, 0.00023117, -0.00031601, 0.00032215, -0.00043096, 0.00042348, -0.00072053, 0.00072494, -0.00208014, 0.00174493, 0.00268883, -0.00233247, 0.00072277, -0.00072781, 0.00043632, -0.00044295, 0.00030107, -0.00029956, 0.00024443, -0.00025159, 0.00020036, -0.00019032, 0.00016186, -0.00016634, 0.00013381, -0.00013459, 0.00012607, -0.00013121, 0.00010933, -0.00010898, 0.00010482, -0.00009853, 0.00008756, -0.00008412, 0.00007573, -0.00008686, 0.00008574, -0.00007583, 0.00007156, -0.00007480, 0.00006783, -0.00007046, 0.00006645, -0.00005073, 0.00003619, -0.00003220, 0.00002389, -0.00003473, 0.00003538, -0.00003370, 0.00003255, -0.00003191, 0.00002266, -0.00002608, 0.00002260, -0.00002407, 0.00002492, -0.00002474, 0.00002333, -0.00001982, 0.00002064, -0.00001319, 0.00000493, -0.00001690, 0.00001790, -0.00001264, 0.00000022, -0.00000219, 0.00000700, -0.00000185, -0.00001008, -0.00000131, },
		{ 0.00001732, -0.00001277, -0.00000126, 0.00002342, -0.00003149, 0.00002053, -0.00000870, 0.00000997, -0.00001256, 0.00000372, 0.00000783, 0.00000338, -0.00001319, 0.00000013, -0.00000040, 0.00000588, -0.00000066, 0.00000709, -0.00000629, 0.00000582, -0.00000828, 0.00000772, -0.00000821, 0.00000262, -0.00000325, 0.00000496, -0.00000662, 0.00001007, -0.00000963, 0.00000102, 0.00000308, -0.00001314, 0.00001854, -0.00001133, -0.00000166, -0.00000203, -0.00000314, 0.00000095, -0.00000298, 0.00001373, -0.00001651, 0.00000704, -0.00001787, 0.00000947, -0.00000724, 0.00001363, -0.00001959, 0.00001248, -0.00001990, 0.00001644, -0.00003453, 0.00001836, -0.00002310, 0.00001161, -0.00003379, 0.00002137, -0.00003678, 0.00002456, -0.00007448, 0.00003263, -0.00007876, 0.00002019, -0.00017699, 0.00001467, -0.00058947, -0.00088836, 0.00391633, -0.00133291, -0.00027835, -0.00016467, -0.00001935, -0.00008548, -0.00000622, -0.00005322, 0.00002656, -0.00004931, 0.00001087, -0.00002507, 0.00001374, -0.00002741, 0.00000569, -0.00002306, 0.00002199, -0.00002724, 0.00000968, -0.00001917, 0.00001847, -0.00001479, 0.00000885, -0.00001254, 0.00001143, -0.00002339, 0.00001926, -0.00001197, 0.00001176, -0.00001527, 0.00000876, -0.00001182, 0.00000538, 0.00000972, -0.00001711, 0.00001246, -0.00001237, -0.00000284, 0.00000245, -0.00000236, 0.00000253, -0.00000228, -0.00000377, -0.00000330, 0.00000127, -0.00000479, 0.00000547, -0.00000476, 0.00000398, -0.00000248, 0.00000347, 0.00000225, -0.00000481, -0.00000848, 0.00000559, 0.00000147, -0.00000933, -0.00000178, 0.00000790, -0.00000113, -0.00000498, -0.00001167, },
		{ -0.00000643, 0.00001815, -0.00002994, 0.00003508, -0.00001936, -0.00000491, 0.00001597, -0.00000987, 0.00000830, -0.00001672, 0.00001976, -0.00000245, -0.00000020, -0.00001592, 0.00001310, -0.00000877, 0.00001369, -0.00000672, 0.00000990, -0.00001050, 0.00000977, -0.00001315, 0.00001503, -0.00002062, 0.00001916, -0.00001731, 0.00001745, -0.00001726, 0.00002184, -0.00003145, 0.00003603, -0.00004277, 0.00004084, -0.00002925, 0.00001916, -0.00002646, 0.00002210, -0.00002547, 0.00002379, -0.00001724, 0.00002083, -0.00003131, 0.00002483, -0.00003848, 0.00003820, -0.00003260, 0.00003331, -0.00004415, 0.00003928, -0.00004550, 0.00003818, -0.00006451, 0.00006155, -0.00007285, 0.00006020, -0.00008424, 0.00007696, -0.00009677, 0.00007654, -0.00014565, 0.00012543, -0.00020181, 0.00012960, -0.00036387, 0.00014764, -0.00159759, 0.00262164, 0.00101438, -0.00136893, 0.00024498, -0.00036112, 0.00016542, -0.00021678, 0.00011061, -0.00012868, 0.00009586, -0.00011810, 0.00008162, -0.00008439, 0.00006557, -0.00007710, 0.00004898, -0.00005113, 0.00004744, -0.00005821, 0.00004063, -0.00004157, 0.00004351, -0.00004431, 0.00003547, -0.00003417, 0.00002456, -0.00003050, 0.00003444, -0.00003233, 0.00002937, -0.00003486, 0.00003101, -0.00003844, 0.00004928, -0.00004682, 0.00003439, -0.00002962, 0.00001430, -0.00001747, 0.00001669, -0.00001662, 0.00001691, -0.00002022, 0.00001145, -0.00001402, 0.00001010, -0.00001064, 0.00001134, -0.00001096, 0.00001146, -0.00001116, 0.00001541, -0.00001439, 0.00000238, -0.00000895, 0.00001482, -0.00001602, 0.00000034, -0.00000074, 0.00001231, -0.00001509, 0.00000029, },
		{ -0.00000677, 0.00001079, -0.00000767, -0.00000548, 0.00002944, -0.00004329, 0.00003567, -0.00001771, 0.00001414, -0.00001819, 0.00001306, 0.00000377, 0.00000355, -0.00001898, 0.00001093, -0.00000616, 0.00000935, -0.00000206, 0.00000732, -0.00000742, 0.00000807, -0.00001221, 0.00001310, -0.00001592, 0.00001209, -0.00000962, 0.00001104, -0.00001209, 0.00001731, -0.00002429, 0.00002514, -0.00002452, 0.00001512, -0.00000220, -0.00000339, -0.00000607, 0.00000083, -0.00000476, 0.00000293, 0.00000100, 0.00000538, -0.00001414, 0.00000884, -0.00002087, 0.00001351, -0.00000819, 0.00001244, -0.00002116, 0.00001393, -0.00001992, 0.00001715, -0.00004073, 0.00002779, -0.00003357, 0.00002333, -0.00004471, 0.00003070, -0.00004696, 0.00003521, -0.00009396, 0.00005649, -0.00011165, 0.00005448, -0.00024750, 0.00008850, -0.00107467, 0.00069357, 0.00244202, -0.00103758, -0.00023183, -0.00016421, 0.00000034, -0.00009505, 0.00000343, -0.00005669, 0.00003322, -0.00005512, 0.00001881, -0.00003390, 0.00002123, -0.00003266, 0.00000806, -0.00002191, 0.00002204, -0.00002895, 0.00001172, -0.00001988, 0.00002293, -0.00002195, 0.00001360, -0.00001579, 0.00001085, -0.00001877, 0.00002174, -0.00002052, 0.00001932, -0.00002446, 0.00002104, -0.00002913, 0.00003425, -0.00002349, 0.00000855, -0.00000529, -0.00000551, -0.00000251, 0.00000278, -0.00000497, 0.00000619, -0.00000821, 0.00000031, -0.00000510, 0.00000188, -0.00000419, 0.00000489, -0.00000456, 0.00000615, -0.00000732, 0.00001022, -0.00000712, -0.00000255, -0.00000657, 0.00000977, -0.00000634, -0.00000710, -0.00000317, 0.00001521, -0.00001104, -0.00000313, },
		{ 0.00001245, -0.00001720, 0.00002577, -0.00003701, 0.00004465, -0.00003349, 0.00000934, 0.00000893, -0.00000530, 0.00000022, -0.00000569, 0.00001427, 0.00000213, -0.00001189, -0.00000157, 0.00000507, -0.00000157, 0.00000663, 0.00000143, -0.00000102, 0.00000272, -0.00000580, 0.00000430, -0.00000465, -0.00000012, 0.00000157, 0.00000164, -0.00000281, 0.00000703, -0.00001007, 0.00000614, 0.00000075, -0.00001237, 0.00002093, -0.00001981, 0.00000846, -0.00001393, 0.00000822, -0.00000968, 0.00001124, -0.00000318, -0.00000340, -0.00000225, -0.00000568, -0.00000635, 0.00000736, -0.00000023, -0.00000483, -0.00000381, -0.00000328, 0.00000235, -0.00001844, -0.00000266, -0.00000392, -0.00000291, -0.00001342, -0.00000510, -0.00001191, 0.00000349, -0.00004406, -0.00000707, -0.00003991, -0.00000687, -0.00013699, -0.00001903, -0.00059201, -0.00031861, 0.00216000, 0.00035909, -0.00083746, -0.00002730, -0.00013662, -0.00000308, -0.00007434, -0.00000725, -0.00002071, -0.00000021, -0.00002893, 0.00000060, -0.00001229, 0.00000120, -0.00001937, -0.00000311, 0.00000048, -0.00000327, -0.00000990, -0.00000457, 0.00000523, -0.00000203, -0.00000439, -0.00000123, -0.00000168, -0.00000632, 0.00000850, -0.00000740, 0.00000578, -0.00000939, 0.00000567, -0.00001222, 0.00001012, 0.00000606, -0.00001822, 0.00001621, -0.00002125, 0.00000894, -0.00000793, 0.00000417, -0.00000362, 0.00000317, -0.00000981, 0.00000351, -0.00000687, 0.00000356, -0.00000341, 0.00000320, -0.00000067, -0.00000003, 0.00000003, 0.00000502, -0.00001251, 0.00000208, -0.00000189, 0.00000738, -0.00001570, -0.00000006, 0.00000403, 0.00000954, -0.00002254, },
		{ 0.00000987, -0.00001481, 0.00001492, -0.00001256, 0.00000253, 0.00001845, -0.00003488, 0.00003450, -0.00001664, 0.00000884, -0.00001175, 0.00001247, 0.00000585, -0.00000556, -0.00001024, 0.00000903, -0.00000311, 0.00000499, 0.00000454, -0.00000235, 0.00000410, -0.00000498, 0.00000175, -0.00000172, -0.00000196, 0.00000116, 0.00000351, -0.00000367, 0.00000638, -0.00000580, -0.00000131, 0.00000994, -0.00001715, 0.00001745, -0.00000979, -0.00000239, -0.00000283, -0.00000505, 0.00000364, -0.00000426, 0.00001342, -0.00001755, 0.00001085, -0.00001570, 0.00000427, -0.00001049, 0.00001858, -0.00001998, 0.00001260, -0.00002224, 0.00002165, -0.00003126, 0.00000978, -0.00002382, 0.00002045, -0.00003320, 0.00001728, -0.00004028, 0.00003390, -0.00006198, 0.00001520, -0.00007138, 0.00003698, -0.00014596, 0.00000303, -0.00039055, -0.00055632, 0.00123917, 0.00129684, -0.00052612, -0.00042620, 0.00001530, -0.00013014, 0.00001852, -0.00009125, 0.00003273, -0.00004634, 0.00001858, -0.00004646, 0.00002332, -0.00003026, 0.00001457, -0.00003622, 0.00002382, -0.00002306, 0.00001360, -0.00002822, 0.00002136, -0.00001535, 0.00001047, -0.00001685, 0.00001259, -0.00001917, 0.00001936, -0.00001555, 0.00001126, -0.00001282, 0.00000798, -0.00001122, 0.00000258, 0.00001428, -0.00001929, 0.00001112, -0.00001180, -0.00000252, 0.00000221, -0.00000506, 0.00000275, -0.00000144, -0.00000404, -0.00000266, -0.00000151, -0.00000228, 0.00000150, -0.00000195, 0.00000371, -0.00000173, -0.00000140, 0.00000681, -0.00001112, -0.00000060, -0.00000174, 0.00000752, -0.00001116, -0.00000423, -0.00000268, 0.00001918, -0.00002392, },
		{ -0.00001409, 0.00001066, -0.00001784, 0.00002357, -0.00003197, 0.00003921, -0.00003325, 0.00001582, 0.00000440, -0.00000479, -0.00000110, 0.00000096, 0.00001151, -0.00000113, -0.00001185, 0.00000462, 0.00000220, -0.00000069, 0.00000799, -0.00000211, 0.00000295, -0.00000129, -0.00000238, 0.00000163, -0.00000422, 0.00000218, 0.00000196, 0.00000035, 0.00000063, 0.00000288, -0.00001094, 0.00001717, -0.00001687, 0.00001041, -0.00000038, -0.00000945, 0.00000286, -0.00001075, 0.00000717, -0.00000843, 0.00001647, -0.00001607, 0.00000702, -0.00000990, 0.00000130, -0.00001183, 0.00001562, -0.00001234, 0.00000633, -0.00001601, 0.00001218, -0.00001577, -0.00000296, -0.00001562, 0.00000854, -0.00001550, 0.00000210, -0.00002506, 0.00001163, -0.00002677, -0.00001428, -0.00004006, -0.00000516, -0.00006021, -0.00008461, -0.00014408, -0.00070003, 0.00068882, 0.00120537, 0.00042565, -0.00077521, 0.00000773, -0.00015343, 0.00002109, -0.00009436, 0.00001821, -0.00004240, 0.00002021, -0.00004460, 0.00001319, -0.00002423, 0.00001322, -0.00003232, 0.00001365, -0.00001614, 0.00001126, -0.00002311, 0.00001033, -0.00000628, 0.00000518, -0.00001223, 0.00000713, -0.00001383, 0.00001213, -0.00000530, 0.00000034, -0.00000268, -0.00000193, 0.00000067, -0.00001140, 0.00002258, -0.00001742, 0.00000414, -0.00000260, -0.00001154, 0.00000814, -0.00000869, 0.00000376, -0.00000263, -0.00000065, -0.00000653, 0.00000216, -0.00000676, 0.00000515, -0.00000562, 0.00000570, -0.00000148, -0.00000184, 0.00000434, -0.00000403, -0.00000885, 0.00000402, 0.00000119, -0.00000121, -0.00001144, -0.00000177, 0.00001118, -0.00000341, },
		{ -0.00002351, 0.00001551, -0.00002017, 0.00001632, -0.00001320, 0.00000589, 0.00000844, -0.00002030, 0.00002461, -0.00001070, 0.00000114, -0.00000059, 0.00000745, 0.00000672, -0.00001197, 0.00000074, 0.00000340, 0.00000075, 0.00000257, 0.00000622, -0.00000519, 0.00000821, -0.00001075, 0.00000918, -0.00001164, 0.00000957, -0.00000715, 0.00001186, -0.00001159, 0.00001660, -0.00002311, 0.00002449, -0.00001758, 0.00000920, -0.00000155, -0.00000335, -0.00000608, 0.00000014, -0.00000710, 0.00000594, -0.00000093, 0.00000708, -0.00001797, 0.00001522, -0.00002113, 0.00001110, -0.00001472, 0.00002109, -0.00002561, 0.00001846, -0.00002714, 0.00002777, -0.00004337, 0.00002597, -0.00004133, 0.00003871, -0.00005053, 0.00003429, -0.00005912, 0.00005399, -0.00009192, 0.00005223, -0.00011876, 0.00008805, -0.00023049, 0.00010124, -0.00078362, 0.00038074, 0.00084262, 0.00090639, -0.00033441, -0.00044829, 0.00005967, -0.00015681, 0.00004396, -0.00010700, 0.00005073, -0.00006332, 0.00003737, -0.00006399, 0.00003791, -0.00004229, 0.00002373, -0.00004128, 0.00003002, -0.00003137, 0.00002217, -0.00003499, 0.00003164, -0.00002784, 0.00001997, -0.00002440, 0.00001625, -0.00001983, 0.00002775, -0.00002907, 0.00002312, -0.00002648, 0.00002495, -0.00003298, 0.00003464, -0.00002042, 0.00000583, -0.00000515, -0.00000646, -0.00000137, 0.00000290, -0.00000835, 0.00000695, -0.00000767, 0.00000026, -0.00000465, -0.00000044, -0.00000237, 0.00000204, -0.00000332, 0.00000743, -0.00000770, 0.00000594, -0.00000221, -0.00000951, 0.00000101, 0.00000327, -0.00000031, -0.00000967, -0.00000449, 0.00000330, 0.00001101, },
		{ -0.00001194, -0.00000184, 0.00000137, -0.00001212, 0.00001596, -0.00002020, 0.00002227, -0.00001626, 0.00000677, 0.00000980, -0.00001245, 0.00000799, -0.00000114, 0.00001260, -0.00000947, -0.00000136, 0.00000083, 0.00000546, -0.00000382, 0.00001196, -0.00000849, 0.00001127, -0.00001148, 0.00000935, -0.00001174, 0.00000913, -0.00000764, 0.00001288, -0.00001134, 0.00001598, -0.00001864, 0.00001453, -0.00000402, -0.00000191, 0.00000527, -0.00000546, -0.00000615, 0.00000154, -0.00001022, 0.00000781, -0.00000573, 0.00001569, -0.00002470, 0.00001896, -0.00002227, 0.00001452, -0.00002272, 0.00002694, -0.00002833, 0.00002316, -0.00003360, 0.00003352, -0.00004410, 0.00002845, -0.00004732, 0.00004087, -0.00004831, 0.00003749, -0.00006553, 0.00005793, -0.00008684, 0.00005552, -0.00012297, 0.00008979, -0.00019740, 0.00009055, -0.00055550, -0.00009609, 0.00083368, 0.00061283, 0.00054777, -0.00078305, 0.00008123, -0.00020577, 0.00007207, -0.00013472, 0.00006110, -0.00008206, 0.00005870, -0.00007842, 0.00004235, -0.00005006, 0.00003589, -0.00005201, 0.00003588, -0.00003999, 0.00003406, -0.00004400, 0.00003437, -0.00003010, 0.00002428, -0.00002925, 0.00002147, -0.00002847, 0.00003561, -0.00003116, 0.00002295, -0.00002695, 0.00002467, -0.00002860, 0.00002240, -0.00000508, -0.00000486, 0.00000107, -0.00000769, -0.00000531, 0.00000773, -0.00001179, 0.00000765, -0.00000763, 0.00000158, -0.00000696, 0.00000261, -0.00000721, 0.00000697, -0.00000876, 0.00001092, -0.00000774, 0.00000408, -0.00000064, -0.00000690, -0.00000577, 0.00000798, -0.00000234, -0.00000494, -0.00000844, 0.00000056, 0.00001032, },
		{ -0.00000298, -0.00000966, 0.00000701, -0.00001431, 0.00000731, -0.00000154, -0.00000700, 0.00001654, -0.00001911, 0.00002213, -0.00001186, 0.00000276, 0.00000367, 0.00000604, 0.00000043, -0.00000655, 0.00000238, 0.00000331, -0.00000045, 0.00000579, 0.00000012, 0.00000252, -0.00000060, -0.00000141, -0.00000037, -0.00000359, 0.00000480, -0.00000023, 0.00000375, -0.00000010, 0.00000186, -0.00000965, 0.00001971, -0.00002052, 0.00002003, -0.00001691, 0.00000466, -0.00000979, 0.00000177, -0.00000659, 0.00000638, 0.00000407, -0.00000732, -0.00000280, 0.00000102, -0.00000691, -0.00000118, -0.00000021, 0.00000151, -0.00000567, -0.00000287, -0.00000143, -0.00000344, -0.00001200, -0.00000381, -0.00001125, 0.00000735, -0.00001539, -0.00000648, -0.00001190, -0.00000524, -0.00002706, -0.00002121, -0.00003580, -0.00002784, -0.00009111, -0.00017307, -0.00057049, 0.00088837, 0.00027912, 0.00090070, -0.00036759, -0.00032713, -0.00001656, -0.00008694, -0.00001202, -0.00005008, 0.00000093, -0.00001782, -0.00000132, -0.00003010, 0.00000708, -0.00001408, -0.00000301, -0.00001132, -0.00000005, -0.00000388, -0.00000339, -0.00000708, 0.00000638, -0.00000803, 0.00000129, -0.00000674, -0.00000392, 0.00000803, 0.00000147, -0.00000772, 0.00000080, -0.00000395, 0.00000305, -0.00001241, 0.00002612, -0.00002717, 0.00001751, -0.00001881, 0.00000211, -0.00000133, -0.00000015, -0.00000539, 0.00000392, -0.00000771, 0.00000106, -0.00000440, -0.00000138, 0.00000028, -0.00000163, 0.00000142, 0.00000320, -0.00000538, 0.00000613, -0.00000910, -0.00000556, 0.00000340, 0.00000431, -0.00000801, -0.00000505, -0.00000493, 0.00000837, },
		{ -0.00000247, -0.00000413, -0.00000262, 0.00000067, -0.00001443, 0.00001816, -0.00002032, 0.00001565, -0.00000285, -0.00000331, 0.00001457, -0.00001729, 0.00001682, -0.00000427, 0.00000971, -0.00001082, 0.00000626, -0.00000280, 0.00000705, -0.00000322, 0.00000924, -0.00000557, 0.00000872, -0.00000979, 0.00000879, -0.00001350, 0.00001337, -0.00000891, 0.00001342, -0.00000996, 0.00001484, -0.00002268, 0.00002838, -0.00002307, 0.00001988, -0.00001440, 0.00000275, -0.00000987, 0.00000343, -0.00001003, 0.00000745, 0.00000105, 0.00000242, -0.00001460, 0.00001155, -0.00001571, 0.00000956, -0.00001469, 0.00001511, -0.00001798, 0.00001211, -0.00001953, 0.00001742, -0.00003163, 0.00001937, -0.00003834, 0.00003097, -0.00003574, 0.00002166, -0.00004650, 0.00003338, -0.00006517, 0.00003381, -0.00010075, 0.00005384, -0.00016818, 0.00002047, -0.00068042, 0.00057749, 0.00042650, 0.00051180, 0.00045869, -0.00065253, 0.00003318, -0.00016173, 0.00004018, -0.00010010, 0.00003445, -0.00005891, 0.00004068, -0.00006257, 0.00002749, -0.00003585, 0.00002096, -0.00003346, 0.00001869, -0.00002586, 0.00002097, -0.00002906, 0.00002352, -0.00002373, 0.00001693, -0.00002113, 0.00000979, -0.00001161, 0.00002404, -0.00002514, 0.00001619, -0.00002173, 0.00002244, -0.00003102, 0.00003775, -0.00003018, 0.00001705, -0.00001590, -0.00000044, -0.00000380, 0.00000548, -0.00001107, 0.00000770, -0.00001046, 0.00000382, -0.00000734, 0.00000207, -0.00000515, 0.00000467, -0.00000654, 0.00001041, -0.00000986, 0.00000875, -0.00000944, -0.00000389, -0.00000393, 0.00001162, -0.00001057, -0.00000203, -0.00000884, 0.00000739, },
		{ 0.00000199, -0.00000560, -0.00000050, -0.00000137, -0.00000886, 0.00000013, 0.00000823, -0.00002197, 0.00003335, -0.00002841, 0.00002639, -0.00001771, 0.00001135, 0.00000160, 0.00000557, -0.00000602, 0.00000403, -0.00000207, 0.00000619, -0.00000214, 0.00000709, -0.00000282, 0.00000722, -0.00000706, 0.00000676, -0.00001074, 0.00000798, -0.00000337, 0.00000824, -0.00000451, 0.00001101, -0.00001570, 0.00001537, -0.00000527, 0.00000139, 0.00000504, -0.00001406, 0.00000358, -0.00000862, 0.00000142, -0.00000603, 0.00001084, -0.00000217, -0.00000814, 0.00000169, -0.00000462, -0.00000003, -0.00000511, 0.00000244, -0.00000495, 0.00000120, -0.00000890, 0.00000635, -0.00001770, 0.00000609, -0.00002200, 0.00000706, -0.00001056, 0.00000276, -0.00002607, 0.00000961, -0.00003741, 0.00001189, -0.00006792, 0.00001455, -0.00010869, 0.00000541, -0.00048865, 0.00005784, 0.00073599, 0.00007495, 0.00082797, -0.00034231, -0.00026661, -0.00004996, -0.00005503, -0.00003246, -0.00002625, -0.00002052, 0.00000356, -0.00002019, -0.00001359, -0.00000705, -0.00000306, -0.00001022, -0.00000202, -0.00001135, 0.00000673, -0.00001219, 0.00000586, -0.00000841, 0.00000312, -0.00000777, -0.00000123, -0.00000718, 0.00001892, -0.00001380, 0.00000592, -0.00001484, 0.00001493, -0.00002035, 0.00002009, -0.00000721, -0.00000523, 0.00000488, -0.00001674, 0.00000564, -0.00000194, -0.00000250, -0.00000192, -0.00000149, -0.00000424, -0.00000027, -0.00000372, -0.00000112, 0.00000071, -0.00000298, 0.00000495, -0.00000253, 0.00000154, -0.00000171, -0.00000895, -0.00000311, 0.00000747, -0.00000143, -0.00000909, -0.00000347, 0.00000006, },
		{ 0.00000834, -0.00001361, 0.00001007, -0.00001269, 0.00000751, -0.00002320, 0.00002823, -0.00003302, 0.00002607, -0.00000464, -0.00000354, 0.00001103, -0.00001196, 0.00001757, -0.00000394, 0.00000175, -0.00000210, 0.00000488, -0.00000127, 0.00000511, 0.00000004, 0.00000315, 0.00000257, -0.00000093, 0.00000107, -0.00000323, -0.00000181, 0.00000519, 0.00000063, 0.00000306, 0.00000408, -0.00000415, -0.00000088, 0.00001227, -0.00001387, 0.00001874, -0.00002237, 0.00000775, -0.00001227, 0.00000508, -0.00001039, 0.00001044, 0.00000077, -0.00000643, -0.00000296, -0.00000041, -0.00000328, -0.00000041, -0.00000354, -0.00000112, -0.00000126, -0.00000517, 0.00000166, -0.00001046, -0.00000176, -0.00000948, -0.00000895, 0.00000066, -0.00000316, -0.00001503, -0.00000393, -0.00002198, -0.00000284, -0.00003762, -0.00001939, -0.00006221, -0.00002527, -0.00028851, -0.00029059, 0.00074830, 0.00009517, 0.00053304, 0.00032563, -0.00049481, -0.00004904, -0.00009126, -0.00001301, -0.00004658, -0.00001177, -0.00001616, 0.00000315, -0.00002916, -0.00000198, -0.00001008, -0.00000214, -0.00000756, -0.00000781, -0.00000136, -0.00000203, -0.00000261, -0.00000178, -0.00000382, -0.00000099, -0.00000554, -0.00000642, 0.00001293, -0.00000217, -0.00000218, -0.00000825, 0.00000455, -0.00000586, 0.00000033, 0.00001404, -0.00002223, 0.00001772, -0.00002345, 0.00000693, -0.00000433, 0.00000233, -0.00000715, 0.00000294, -0.00000848, 0.00000346, -0.00000682, 0.00000159, -0.00000314, 0.00000139, -0.00000141, 0.00000436, -0.00000456, 0.00000519, -0.00001368, -0.00000013, -0.00000092, 0.00000985, -0.00001581, 0.00000146, -0.00000684, },
		{ 0.00000592, -0.00001203, 0.00000626, -0.00000550, 0.00000099, -0.00001336, 0.00000495, 0.00000234, -0.00002033, 0.00003872, -0.00003198, 0.00002529, -0.00001652, 0.00001559, 0.00000098, -0.00000012, -0.00000308, 0.00000808, -0.00000345, 0.00000595, 0.00000019, 0.00000209, 0.00000320, 0.00000092, -0.00000074, 0.00000067, -0.00000633, 0.00000682, -0.00000002, 0.00000402, 0.00000259, 0.00000208, -0.00000880, 0.00001777, -0.00001480, 0.00001628, -0.00001352, -0.00000342, -0.00000287, -0.00000372, -0.00000146, -0.00000245, 0.00001270, -0.00001241, 0.00000200, -0.00000703, 0.00000351, -0.00000617, 0.00000366, -0.00001081, 0.00000744, -0.00001189, 0.00000805, -0.00001488, 0.00000201, -0.00001081, -0.00000498, -0.00001063, 0.00000922, -0.00002086, 0.00000279, -0.00002810, 0.00000141, -0.00003088, -0.00001956, -0.00005610, -0.00002949, -0.00015880, -0.00041373, 0.00050188, 0.00035000, 0.00014185, 0.00063597, -0.00021627, -0.00030056, -0.00000974, -0.00008557, 0.00000319, -0.00005544, 0.00000702, -0.00002004, 0.00000062, -0.00003176, 0.00000955, -0.00001971, 0.00001026, -0.00002216, 0.00000664, -0.00001067, 0.00000788, -0.00001085, 0.00000308, -0.00000813, 0.00000317, -0.00001507, 0.00001404, -0.00000046, 0.00000086, -0.00000959, 0.00000070, 0.00000034, -0.00000799, 0.00002002, -0.00002140, 0.00001179, -0.00001237, -0.00000590, 0.00000409, -0.00000320, -0.00000157, -0.00000265, -0.00000362, -0.00000169, -0.00000159, -0.00000316, 0.00000028, -0.00000167, 0.00000030, 0.00000230, -0.00000218, 0.00000404, -0.00001015, -0.00000440, -0.00000146, 0.00000954, -0.00000940, -0.00000509, -0.00000360, },
		{ 0.00000009, -0.00000275, -0.00000927, 0.00001289, -0.00001594, 0.00000786, -0.00002362, 0.00002846, -0.00003843, 0.00003606, -0.00001023, -0.00000157, 0.00000719, -0.00000388, 0.00001487, -0.00000536, -0.00000194, 0.00000658, 0.00000132, 0.00000053, 0.00000501, -0.00000127, 0.00000383, 0.00000294, -0.00000230, 0.00000367, -0.00000812, 0.00000550, 0.00000077, 0.00000474, 0.00000034, 0.00000801, -0.00001311, 0.00001755, -0.00000978, 0.00000844, -0.00000158, -0.00001320, 0.00000233, -0.00000792, 0.00000259, -0.00000799, 0.00001392, -0.00000793, -0.00000166, -0.00000469, 0.00000008, -0.00000231, 0.00000166, -0.00000817, 0.00000080, -0.00000348, -0.00000046, -0.00000383, -0.00000979, 0.00000137, -0.00001272, -0.00000565, -0.00000153, -0.00000312, -0.00001294, -0.00000848, -0.00002336, 0.00000078, -0.00004315, -0.00001982, -0.00007614, -0.00003068, -0.00045122, 0.00024831, 0.00047379, 0.00007236, 0.00043241, 0.00030834, -0.00043436, -0.00006119, -0.00008058, -0.00001220, -0.00004344, -0.00001289, -0.00001348, 0.00000010, -0.00002607, -0.00000358, -0.00001094, 0.00000274, -0.00001116, -0.00000595, -0.00000345, 0.00000177, -0.00000201, -0.00000578, -0.00000200, -0.00000144, -0.00000888, 0.00000183, 0.00001004, -0.00000450, -0.00000072, -0.00001098, 0.00001018, -0.00001635, 0.00002278, -0.00001610, 0.00000243, -0.00000078, -0.00001519, 0.00000729, -0.00000465, 0.00000066, -0.00000452, -0.00000206, -0.00000444, 0.00000133, -0.00000565, 0.00000208, -0.00000366, 0.00000161, 0.00000023, -0.00000006, 0.00000210, -0.00000436, -0.00001073, 0.00000137, 0.00000323, 0.00000189, -0.00001364, 0.00000126, },
		{ 0.00000077, 0.00000169, -0.00001687, 0.00001599, -0.00001270, 0.00000492, -0.00001809, 0.00001062, -0.00000900, -0.00000716, 0.00003154, -0.00002690, 0.00001767, -0.00000695, 0.00001270, 0.00000212, -0.00000733, 0.00000634, 0.00000521, -0.00000119, 0.00000425, 0.00000205, -0.00000175, 0.00000874, -0.00000591, 0.00000753, -0.00000955, 0.00000489, -0.00000090, 0.00000817, -0.00000405, 0.00001364, -0.00001434, 0.00001407, -0.00000315, 0.00000141, 0.00000565, -0.00001425, -0.00000170, -0.00000397, -0.00000179, -0.00000310, 0.00000369, 0.00000567, -0.00001258, 0.00000561, -0.00001157, 0.00000877, -0.00000862, 0.00000579, -0.00001694, 0.00001340, -0.00001694, 0.00001527, -0.00002844, 0.00001857, -0.00002676, 0.00001122, -0.00002630, 0.00002439, -0.00003710, 0.00002178, -0.00005707, 0.00003537, -0.00007108, 0.00002498, -0.00012665, 0.00005328, -0.00040724, 0.00002361, 0.00047056, 0.00016593, 0.00017531, 0.00049894, -0.00012349, -0.00032020, 0.00001565, -0.00010099, 0.00002393, -0.00007274, 0.00002535, -0.00003720, 0.00001583, -0.00004374, 0.00001868, -0.00002622, 0.00001934, -0.00003265, 0.00001656, -0.00001983, 0.00002139, -0.00002519, 0.00001333, -0.00001657, 0.00000847, -0.00001876, 0.00002562, -0.00001616, 0.00001469, -0.00002512, 0.00001947, -0.00002284, 0.00002282, -0.00000971, -0.00000422, 0.00000372, -0.00001414, 0.00000037, 0.00000202, -0.00000464, 0.00000103, -0.00000678, -0.00000143, -0.00000205, -0.00000167, -0.00000235, 0.00000061, -0.00000318, 0.00000404, -0.00000350, 0.00000414, -0.00000215, -0.00001166, -0.00000127, 0.00000219, 0.00000480, -0.00001152, -0.00000308, },
		{ 0.00000196, 0.00000301, -0.00001457, 0.00000369, 0.00000484, -0.00001051, 0.00000053, -0.00001453, 0.00001466, -0.00002662, 0.00003162, -0.00000686, -0.00000692, 0.00001220, -0.00000227, 0.00001442, -0.00001124, 0.00000367, 0.00000745, 0.00000187, -0.00000062, 0.00000750, -0.00000629, 0.00001032, -0.00000440, 0.00000601, -0.00000553, 0.00000046, 0.00000065, 0.00000694, -0.00000200, 0.00001036, -0.00000559, 0.00000239, 0.00000910, -0.00000851, 0.00001283, -0.00001436, -0.00000390, -0.00000418, -0.00000193, -0.00000208, -0.00000103, 0.00001010, -0.00001271, 0.00000520, -0.00001129, 0.00000657, -0.00000654, 0.00000741, -0.00001817, 0.00001023, -0.00001322, 0.00001298, -0.00002295, 0.00000994, -0.00001636, 0.00000496, -0.00002298, 0.00001658, -0.00002447, 0.00001355, -0.00004478, 0.00001642, -0.00004481, 0.00000877, -0.00009599, 0.00001444, -0.00023821, -0.00022447, 0.00046735, 0.00018598, 0.00015713, 0.00028723, 0.00034724, -0.00042702, -0.00003618, -0.00009974, 0.00001447, -0.00006579, 0.00001050, -0.00003425, 0.00001772, -0.00004038, 0.00000920, -0.00002310, 0.00001835, -0.00002751, 0.00000949, -0.00001854, 0.00002015, -0.00001859, 0.00000593, -0.00001263, 0.00000756, -0.00001923, 0.00002063, -0.00001001, 0.00001200, -0.00001864, 0.00000902, -0.00001154, 0.00000702, 0.00000799, -0.00001669, 0.00001014, -0.00001390, -0.00000373, 0.00000383, -0.00000502, 0.00000156, -0.00000596, -0.00000302, -0.00000235, -0.00000032, -0.00000424, 0.00000258, -0.00000541, 0.00000499, -0.00000368, 0.00000246, 0.00000168, -0.00001090, -0.00000584, 0.00000325, 0.00000322, -0.00000497, -0.00000909, },
		{ -0.00000371, 0.00000721, -0.00001145, -0.00000668, 0.00001194, -0.00001000, -0.00000031, -0.00001162, 0.00000225, -0.00000672, -0.00000203, 0.00002709, -0.00002443, 0.00001432, 0.00000094, 0.00000922, -0.00000035, -0.00000716, 0.00001162, 0.00000331, -0.00000046, 0.00000531, -0.00000119, 0.00000203, 0.00000499, -0.00000209, 0.00000411, -0.00000801, 0.00000670, -0.00000044, 0.00000705, -0.00000044, 0.00000919, -0.00001210, 0.00002151, -0.00001653, 0.00001728, -0.00001371, -0.00000298, -0.00000879, 0.00000183, -0.00000555, 0.00000112, 0.00000496, -0.00000311, -0.00000467, -0.00000047, -0.00000627, 0.00000509, -0.00000235, -0.00000393, -0.00000867, 0.00000428, -0.00000426, -0.00000065, -0.00001499, 0.00000838, -0.00001680, 0.00000148, -0.00001494, 0.00000924, -0.00001747, -0.00000406, -0.00003214, 0.00000834, -0.00004148, -0.00001906, -0.00008117, -0.00004995, -0.00039750, 0.00040004, 0.00020007, 0.00019938, 0.00009799, 0.00046909, -0.00011183, -0.00028298, -0.00000753, -0.00007265, 0.00000274, -0.00005140, 0.00000839, -0.00002362, 0.00000432, -0.00003188, 0.00000610, -0.00001033, 0.00000397, -0.00001877, 0.00000322, -0.00000470, 0.00000995, -0.00001814, 0.00000589, -0.00000919, -0.00000201, -0.00000136, 0.00001085, -0.00000637, 0.00000384, -0.00001402, 0.00000987, -0.00001655, 0.00002865, -0.00002829, 0.00001528, -0.00001468, -0.00000315, -0.00000098, 0.00000116, -0.00000470, 0.00000150, -0.00000965, 0.00000157, -0.00000386, -0.00000063, -0.00000145, -0.00000085, -0.00000124, 0.00000315, -0.00000517, 0.00000834, -0.00001117, -0.00000720, -0.00000016, 0.00000551, -0.00000409, -0.00000719, },
		{ -0.00000786, 0.00000741, -0.00000564, -0.00001239, 0.00000758, 0.00000115, -0.00000933, -0.00000079, -0.00001212, 0.00000429, -0.00001195, 0.00002102, 0.00000065, -0.00001447, 0.00002169, -0.00000531, 0.00001339, -0.00001465, 0.00001129, 0.00000456, 0.00000384, -0.00000111, 0.00000691, -0.00000651, 0.00001158, -0.00000650, 0.00000950, -0.00001136, 0.00000854, -0.00000425, 0.00001164, -0.00000551, 0.00001551, -0.00001524, 0.00002107, -0.00001122, 0.00000967, -0.00000420, -0.00000827, -0.00000648, -0.00000223, -0.00000211, -0.00000202, 0.00000450, 0.00000037, -0.00000747, 0.00000300, -0.00001011, 0.00000644, -0.00000363, 0.00000268, -0.00001608, 0.00000748, -0.00000836, 0.00000808, -0.00002310, 0.00001368, -0.00002002, 0.00000863, -0.00002489, 0.00001591, -0.00002228, 0.00001035, -0.00004747, 0.00002134, -0.00005248, 0.00001306, -0.00011358, 0.00002011, -0.00038324, 0.00018190, 0.00031748, 0.00012099, 0.00014913, 0.00022789, 0.00033456, -0.00038510, -0.00004469, -0.00008703, 0.00000999, -0.00006035, 0.00000938, -0.00003443, 0.00001732, -0.00003778, 0.00000528, -0.00001656, 0.00001347, -0.00002439, 0.00000690, -0.00001506, 0.00002089, -0.00002200, 0.00000713, -0.00001267, 0.00000401, -0.00001159, 0.00001906, -0.00001348, 0.00001408, -0.00002195, 0.00001589, -0.00002365, 0.00002987, -0.00002060, 0.00000493, -0.00000477, -0.00000935, -0.00000028, 0.00000158, -0.00000540, 0.00000295, -0.00000931, -0.00000050, -0.00000330, -0.00000017, -0.00000300, 0.00000180, -0.00000519, 0.00000671, -0.00000799, 0.00000836, -0.00000626, -0.00000953, -0.00000373, 0.00000718, -0.00000377, -0.00000442, },
		{ -0.00000333, -0.00000109, 0.00000524, -0.00001797, 0.00000420, 0.00000334, -0.00000393, -0.00000754, -0.00000205, -0.00001364, 0.00000796, -0.00000947, 0.00003120, -0.00002863, 0.00001945, 0.00000196, 0.00000755, -0.00000503, 0.00000010, 0.00000999, 0.00000486, -0.00000135, 0.00000659, -0.00000451, 0.00000702, -0.00000121, 0.00000572, -0.00000548, 0.00000219, 0.00000028, 0.00000660, 0.00000003, 0.00000960, -0.00000532, 0.00000814, 0.00000513, -0.00000600, 0.00001026, -0.00001753, 0.00000130, -0.00001177, 0.00000543, -0.00000868, 0.00000822, -0.00000204, -0.00000345, -0.00000129, -0.00000442, -0.00000184, 0.00000305, -0.00000022, -0.00000970, -0.00000317, -0.00000084, 0.00000336, -0.00001429, 0.00000110, -0.00000659, -0.00000212, -0.00001194, -0.00000196, -0.00000557, 0.00000054, -0.00003086, -0.00000109, -0.00002831, -0.00000148, -0.00007990, -0.00001020, -0.00025047, -0.00008139, 0.00042846, 0.00004013, 0.00023364, 0.00003303, 0.00044945, -0.00010788, -0.00024691, -0.00002750, -0.00004871, -0.00001614, -0.00003081, -0.00000950, -0.00000915, -0.00000689, -0.00002197, -0.00000183, -0.00000018, -0.00000749, -0.00000747, -0.00000721, 0.00000930, -0.00000433, -0.00000773, -0.00000281, -0.00000350, -0.00000673, 0.00001133, -0.00000511, 0.00000707, -0.00001143, 0.00000486, -0.00001287, 0.00001197, 0.00000285, -0.00001561, 0.00001129, -0.00001984, 0.00000527, -0.00000399, 0.00000037, -0.00000297, -0.00000144, -0.00000831, 0.00000176, -0.00000443, 0.00000049, -0.00000121, -0.00000215, 0.00000203, -0.00000171, -0.00000038, 0.00000405, -0.00001411, -0.00000342, 0.00000263, 0.00000269, -0.00000854, },
		{ 0.00000016, -0.00000702, 0.00001053, -0.00001712, 0.00000005, -0.00000008, 0.00000620, -0.00001592, 0.00000766, -0.00002338, 0.00000968, -0.00000999, 0.00001805, 0.00000180, -0.00001424, 0.00002607, -0.00000707, 0.00000843, -0.00000974, 0.00001268, 0.00000436, 0.00000367, 0.00000069, 0.00000260, -0.00000081, 0.00000519, 0.00000077, 0.00000180, -0.00000473, 0.00000576, 0.00000015, 0.00000671, 0.00000245, 0.00000496, -0.00000339, 0.00001756, -0.00001447, 0.00001570, -0.00001823, 0.00000173, -0.00001307, 0.00000383, -0.00000680, 0.00000449, 0.00000168, -0.00000492, -0.00000087, -0.00000299, -0.00000428, 0.00000249, 0.00000189, -0.00000667, -0.00000722, -0.00000193, 0.00000442, -0.00000959, -0.00000556, -0.00000122, -0.00000570, -0.00000506, -0.00000989, -0.00000266, 0.00000077, -0.00002069, -0.00001400, -0.00001628, -0.00001115, -0.00004687, -0.00004360, -0.00013697, -0.00023740, 0.00039095, 0.00010077, 0.00016848, 0.00008014, 0.00023069, 0.00028147, -0.00031610, -0.00007733, -0.00005264, -0.00001675, -0.00003341, -0.00001237, -0.00001650, 0.00000175, -0.00002258, -0.00000878, 0.00000014, -0.00000382, -0.00000809, -0.00000836, 0.00000287, 0.00000482, -0.00001055, -0.00000315, -0.00000391, -0.00000649, 0.00000690, 0.00000040, 0.00000133, -0.00000252, -0.00000330, -0.00000338, -0.00000436, 0.00001972, -0.00002516, 0.00001532, -0.00001929, 0.00000215, -0.00000296, 0.00000070, -0.00000445, 0.00000166, -0.00001005, 0.00000127, -0.00000473, 0.00000096, -0.00000225, 0.00000008, -0.00000216, 0.00000352, -0.00000642, 0.00000886, -0.00001313, -0.00000475, -0.00000228, 0.00000840, -0.00001159, },
		{ -0.00000579, -0.00000250, 0.00000429, -0.00000624, -0.00000979, 0.00000260, 0.00000309, -0.00000574, -0.00000476, -0.00000466, -0.00001842, 0.00001680, -0.00001585, 0.00003423, -0.00003040, 0.00002496, 0.00000079, 0.00000363, -0.00000428, 0.00000545, 0.00000766, 0.00000592, -0.00000028, 0.00000367, -0.00000139, 0.00000436, 0.00000116, 0.00000429, -0.00000615, 0.00000603, -0.00000096, 0.00000745, 0.00000124, 0.00000838, -0.00000649, 0.00001936, -0.00001038, 0.00000910, -0.00000873, -0.00000681, -0.00000514, -0.00000607, 0.00000154, -0.00000448, 0.00000946, -0.00000970, 0.00000247, -0.00000519, -0.00000126, -0.00000354, 0.00000705, -0.00000719, -0.00000427, -0.00000907, 0.00000776, -0.00000758, -0.00000648, -0.00000294, -0.00000304, -0.00000558, -0.00000653, -0.00001050, 0.00000592, -0.00001529, -0.00001699, -0.00001502, -0.00001433, -0.00002579, -0.00005736, -0.00007141, -0.00028096, 0.00024263, 0.00023034, 0.00005670, 0.00017935, 0.00004384, 0.00038325, -0.00006627, -0.00024714, -0.00001419, -0.00005744, -0.00000502, -0.00003760, -0.00000385, -0.00001493, 0.00000004, -0.00002892, 0.00000833, -0.00001029, 0.00000114, -0.00001531, 0.00000272, 0.00000258, -0.00000172, -0.00001076, 0.00000061, -0.00001040, 0.00000648, 0.00000144, 0.00000017, 0.00000014, -0.00000453, 0.00000043, -0.00001230, 0.00002267, -0.00001891, 0.00000533, -0.00000728, -0.00000962, 0.00000475, -0.00000488, -0.00000018, -0.00000195, -0.00000449, -0.00000504, -0.00000041, -0.00000315, 0.00000106, -0.00000210, -0.00000093, 0.00000197, -0.00000414, 0.00000431, -0.00000487, -0.00001005, -0.00000236, 0.00000625, -0.00000560, },
		{ -0.00001263, 0.00000230, -0.00000207, 0.00000313, -0.00001662, 0.00000594, -0.00000558, 0.00000870, -0.00001824, 0.00001178, -0.00003378, 0.00001994, -0.00001582, 0.00002088, -0.00000121, -0.00000768, 0.00002463, -0.00000832, 0.00000435, -0.00000133, 0.00000954, 0.00000601, 0.00000355, 0.00000047, 0.00000157, 0.00000169, 0.00000167, 0.00000583, -0.00000532, 0.00000417, 0.00000025, 0.00000584, 0.00000191, 0.00000926, -0.00000589, 0.00001602, -0.00000123, 0.00000014, 0.00000067, -0.00001388, 0.00000044, -0.00001153, 0.00000380, -0.00000690, 0.00000974, -0.00000652, -0.00000169, -0.00000126, -0.00000346, -0.00000281, 0.00000345, -0.00000047, -0.00000717, -0.00000678, -0.00000089, 0.00000370, -0.00001408, 0.00000251, -0.00000889, 0.00000153, -0.00001053, -0.00000594, -0.00000623, 0.00000413, -0.00003012, -0.00000130, -0.00003227, 0.00000256, -0.00007387, -0.00002111, -0.00027692, 0.00008771, 0.00030161, 0.00003852, 0.00016134, 0.00005831, 0.00020872, 0.00025432, -0.00027339, -0.00008656, -0.00004103, -0.00002384, -0.00002347, -0.00002041, -0.00001037, -0.00000244, -0.00001916, -0.00000890, 0.00000128, -0.00000673, -0.00000490, -0.00000991, 0.00000830, -0.00000323, -0.00000499, -0.00000690, -0.00000388, -0.00000315, 0.00000985, -0.00000676, 0.00000664, -0.00000949, 0.00000761, -0.00001895, 0.00002027, -0.00000880, -0.00000492, 0.00000231, -0.00001660, 0.00000658, -0.00000502, -0.00000033, -0.00000251, -0.00000196, -0.00000709, -0.00000013, -0.00000421, 0.00000166, -0.00000229, -0.00000049, 0.00000035, -0.00000135, -0.00000029, 0.00000124, -0.00001229, -0.00000249, 0.00000146, 0.00000254, },
		{ -0.00001110, -0.00000187, 0.00000020, 0.00000310, -0.00001363, 0.00000145, -0.00000518, 0.00000739, -0.00001094, 0.00000294, -0.00001646, -0.00000984, 0.00001248, -0.00001311, 0.00003002, -0.00002360, 0.00002418, 0.00000103, -0.00000259, 0.00000336, 0.00000464, 0.00000816, 0.00000542, 0.00000146, -0.00000020, 0.00000411, -0.00000200, 0.00000873, -0.00000442, 0.00000271, 0.00000098, 0.00000507, 0.00000151, 0.00001004, -0.00000392, 0.00001065, 0.00000813, -0.00000554, 0.00000515, -0.00001525, -0.00000046, -0.00000946, -0.00000105, -0.00000309, 0.00000359, 0.00000238, -0.00000972, 0.00000486, -0.00000788, 0.00000182, -0.00000441, 0.00000818, -0.00001202, 0.00000046, -0.00001323, 0.00001404, -0.00001945, 0.00000723, -0.00001470, 0.00000785, -0.00001529, 0.00000331, -0.00002196, 0.00001992, -0.00003708, 0.00000843, -0.00004421, 0.00001662, -0.00007441, -0.00000022, -0.00022868, -0.00005748, 0.00031234, 0.00008278, 0.00009687, 0.00011834, 0.00006869, 0.00031423, -0.00002297, -0.00025006, 0.00000044, -0.00006924, 0.00001001, -0.00004949, 0.00000632, -0.00002356, 0.00000796, -0.00003441, 0.00001631, -0.00002015, 0.00001067, -0.00002370, 0.00001513, -0.00001032, 0.00000695, -0.00001741, 0.00000478, -0.00001447, 0.00001839, -0.00001289, 0.00001194, -0.00001394, 0.00001295, -0.00001986, 0.00001223, 0.00000194, -0.00001146, 0.00000588, -0.00001585, 0.00000072, 0.00000096, -0.00000503, 0.00000066, -0.00000381, -0.00000398, -0.00000404, -0.00000170, -0.00000162, 0.00000107, -0.00000320, 0.00000207, -0.00000233, -0.00000023, 0.00000135, -0.00000907, -0.00000560, -0.00000023, 0.00000468, },
		{ -0.00000587, -0.00000825, 0.00000301, 0.00000197, -0.00000870, -0.00000498, 0.00000023, -0.00000255, 0.00000340, -0.00001059, 0.00000256, -0.00002958, 0.00001902, -0.00001566, 0.00001906, 0.00000143, -0.00000387, 0.00002274, -0.00001229, 0.00000705, 0.00000261, 0.00000781, 0.00000619, 0.00000538, -0.00000328, 0.00000624, -0.00000302, 0.00000678, 0.00000073, -0.00000136, 0.00000384, 0.00000244, 0.00000348, 0.00000690, 0.00000267, 0.00000139, 0.00001815, -0.00000903, 0.00000758, -0.00001485, -0.00000278, -0.00000698, -0.00000417, -0.00000225, 0.00000050, 0.00000640, -0.00001041, 0.00000271, -0.00000486, -0.00000012, -0.00000432, 0.00000653, -0.00000720, -0.00000129, -0.00001248, 0.00000764, -0.00000923, -0.00000203, -0.00000610, -0.00000114, -0.00000577, -0.00000201, -0.00001694, 0.00000816, -0.00001706, -0.00000783, -0.00002522, -0.00000665, -0.00003784, -0.00002820, -0.00012918, -0.00019823, 0.00029453, 0.00012123, 0.00008089, 0.00009901, 0.00008515, 0.00015543, 0.00025697, -0.00025751, -0.00007329, -0.00005178, -0.00000995, -0.00003342, -0.00001131, -0.00001842, 0.00000531, -0.00002584, 0.00000150, -0.00001033, 0.00000369, -0.00001454, 0.00000316, -0.00000283, 0.00000284, -0.00001013, -0.00000229, -0.00001087, 0.00001127, -0.00000402, 0.00000353, -0.00000576, 0.00000440, -0.00000600, -0.00000609, 0.00001654, -0.00001878, 0.00000931, -0.00001455, -0.00000410, 0.00000340, -0.00000498, -0.00000074, -0.00000226, -0.00000395, -0.00000410, -0.00000257, -0.00000235, 0.00000166, -0.00000313, 0.00000143, -0.00000137, -0.00000150, 0.00000212, -0.00000706, -0.00000690, -0.00000167, 0.00000375, },
		{ -0.00000324, -0.00000920, -0.00000082, 0.00000573, -0.00000743, -0.00000743, 0.00000198, -0.00000636, 0.00000529, -0.00000772, 0.00000075, -0.00002013, -0.00000403, 0.00000638, -0.00000892, 0.00002567, -0.00001416, 0.00001988, 0.00000009, -0.00000487, 0.00001064, 0.00000214, 0.00000880, 0.00000623, -0.00000017, 0.00000125, 0.00000354, -0.00000176, 0.00000936, -0.00000684, 0.00000797, -0.00000160, 0.00000766, 0.00000077, 0.00001134, -0.00000753, 0.00002424, -0.00000755, 0.00000700, -0.00001232, -0.00000604, -0.00000554, -0.00000450, -0.00000412, -0.00000008, 0.00000584, -0.00000510, -0.00000452, 0.00000160, -0.00000535, 0.00000065, -0.00000096, 0.00000208, -0.00000800, -0.00000348, -0.00000666, 0.00000534, -0.00001442, 0.00000625, -0.00001389, 0.00000677, -0.00001253, -0.00000173, -0.00001431, 0.00000849, -0.00002854, 0.00000126, -0.00003743, 0.00000171, -0.00006420, -0.00003596, -0.00027618, 0.00021244, 0.00017612, 0.00007642, 0.00006989, 0.00011356, 0.00005510, 0.00028239, -0.00001016, -0.00022756, -0.00001014, -0.00005811, 0.00000414, -0.00004383, 0.00000157, -0.00001852, 0.00000272, -0.00002523, 0.00000755, -0.00001355, 0.00000462, -0.00001531, 0.00001072, -0.00001038, 0.00000588, -0.00001514, -0.00000084, -0.00000356, 0.00001136, -0.00001003, 0.00000735, -0.00001000, 0.00001186, -0.00002263, 0.00002548, -0.00002077, 0.00000879, -0.00001098, -0.00000775, 0.00000187, -0.00000064, -0.00000545, 0.00000165, -0.00000658, -0.00000098, -0.00000582, -0.00000078, -0.00000101, 0.00000043, -0.00000256, 0.00000287, -0.00000563, 0.00000516, -0.00000745, -0.00000585, -0.00000415, 0.00000353, },
		{ -0.00000071, -0.00000861, -0.00000501, 0.00000675, -0.00000300, -0.00001090, 0.00000267, -0.00000549, -0.00000019, 0.00000030, -0.00000471, -0.00000854, -0.00001827, 0.00000882, -0.00000877, 0.00001228, 0.00001076, -0.00000560, 0.00002132, -0.00001597, 0.00001381, 0.00000193, 0.00000799, 0.00000605, 0.00000590, -0.00000488, 0.00000931, -0.00000674, 0.00001174, -0.00000596, 0.00000695, -0.00000110, 0.00000787, -0.00000097, 0.00001350, -0.00000732, 0.00001920, 0.00000330, 0.00000019, -0.00000428, -0.00001337, -0.00000160, -0.00000731, -0.00000215, -0.00000448, 0.00000724, -0.00000214, -0.00000691, 0.00000168, -0.00000459, 0.00000046, -0.00000253, 0.00000369, -0.00000762, -0.00000088, -0.00001104, 0.00000664, -0.00001374, 0.00000577, -0.00001276, 0.00000470, -0.00001009, 0.00000102, -0.00001920, 0.00001041, -0.00002570, 0.00000406, -0.00003932, 0.00000567, -0.00006319, -0.00000480, -0.00024766, 0.00004985, 0.00026006, 0.00004695, 0.00009299, 0.00006553, 0.00009523, 0.00012020, 0.00024820, -0.00023151, -0.00007226, -0.00005195, -0.00000522, -0.00003664, -0.00000801, -0.00002019, 0.00000588, -0.00002322, 0.00000117, -0.00001246, 0.00000575, -0.00001520, 0.00000863, -0.00001063, 0.00000716, -0.00001244, -0.00000287, -0.00000789, 0.00001465, -0.00001096, 0.00000861, -0.00001285, 0.00001523, -0.00002101, 0.00001682, -0.00000863, -0.00000290, 0.00000062, -0.00001556, 0.00000292, -0.00000013, -0.00000493, 0.00000004, -0.00000446, -0.00000230, -0.00000423, -0.00000304, -0.00000107, 0.00000124, -0.00000349, 0.00000370, -0.00000563, 0.00000357, -0.00000360, -0.00000829, -0.00000353, 0.00000139, },
		{ 0.00000324, -0.00001012, -0.00000450, 0.00000079, 0.00000642, -0.00001602, 0.00000365, -0.00000483, -0.00000219, -0.00000106, 0.00000145, -0.00001131, -0.00001064, -0.00001033, 0.00000807, -0.00001087, 0.00002865, -0.00000964, 0.00001626, -0.00000301, -0.00000067, 0.00001256, 0.00000186, 0.00000810, 0.00000787, -0.00000325, 0.00000584, -0.00000149, 0.00000433, 0.00000223, 0.00000043, 0.00000447, 0.00000302, 0.00000340, 0.00000781, 0.00000165, 0.00000625, 0.00001802, -0.00000752, 0.00000477, -0.00002040, 0.00000150, -0.00001059, 0.00000143, -0.00000942, 0.00000785, -0.00000035, -0.00000551, -0.00000227, -0.00000082, -0.00000268, 0.00000006, 0.00000025, -0.00000297, -0.00000343, -0.00000707, -0.00000066, -0.00000631, -0.00000167, -0.00000375, -0.00000490, -0.00000127, -0.00000450, -0.00001093, -0.00000224, -0.00001122, -0.00000622, -0.00002323, -0.00001143, -0.00003960, -0.00001738, -0.00015984, -0.00011621, 0.00029904, 0.00004981, 0.00010732, 0.00002609, 0.00012776, 0.00002924, 0.00026503, -0.00000664, -0.00020017, -0.00002672, -0.00004064, -0.00000894, -0.00003202, -0.00000742, -0.00001088, -0.00000314, -0.00001583, -0.00000341, -0.00000372, -0.00000431, -0.00000206, -0.00000189, -0.00000286, -0.00000010, -0.00001072, -0.00000551, 0.00000836, -0.00000217, 0.00000088, -0.00000588, 0.00000612, -0.00000642, -0.00000128, 0.00000909, -0.00001759, 0.00001252, -0.00002154, 0.00000334, -0.00000189, -0.00000079, -0.00000497, 0.00000043, -0.00000654, 0.00000037, -0.00000717, 0.00000063, -0.00000098, -0.00000056, -0.00000003, -0.00000045, -0.00000310, 0.00000404, -0.00001298, -0.00000120, -0.00000229, },
		{ 0.00000416, -0.00000916, -0.00000499, -0.00000377, 0.00001045, -0.00001329, -0.00000183, -0.00000065, -0.00000357, -0.00000483, 0.00000640, -0.00001083, -0.00000720, -0.00001562, 0.00000251, -0.00000507, 0.00000966, 0.00001874, -0.00000853, 0.00001777, -0.00001394, 0.00001747, 0.00000161, 0.00000735, 0.00000750, 0.00000312, -0.00000109, 0.00000580, -0.00000283, 0.00000787, -0.00000331, 0.00000799, -0.00000025, 0.00000712, 0.00000231, 0.00000941, -0.00000290, 0.00002486, -0.00000647, 0.00000644, -0.00001919, -0.00000299, -0.00000811, -0.00000061, -0.00000725, 0.00000147, 0.00000539, -0.00000652, -0.00000259, -0.00000177, -0.00000154, -0.00000097, 0.00000070, -0.00000294, -0.00000244, -0.00000567, -0.00000312, -0.00000523, -0.00000378, -0.00000023, -0.00000743, -0.00000037, -0.00000401, -0.00000733, -0.00000714, -0.00000758, -0.00000818, -0.00001374, -0.00002017, -0.00002514, -0.00003105, -0.00008318, -0.00020466, 0.00023571, 0.00012206, 0.00006876, 0.00005498, 0.00007544, 0.00007403, 0.00011340, 0.00022057, -0.00019012, -0.00008847, -0.00003542, -0.00001664, -0.00002618, -0.00001595, -0.00001221, -0.00000210, -0.00001134, -0.00001068, -0.00000255, -0.00000372, -0.00000252, -0.00000087, -0.00000621, 0.00000262, -0.00000788, -0.00000949, 0.00000528, 0.00000264, -0.00000239, -0.00000203, -0.00000116, 0.00000427, -0.00001164, 0.00001683, -0.00002107, 0.00001206, -0.00001577, -0.00000435, 0.00000123, -0.00000107, -0.00000492, 0.00000007, -0.00000599, 0.00000011, -0.00000598, -0.00000176, -0.00000077, 0.00000031, -0.00000186, 0.00000257, -0.00000644, 0.00000653, -0.00001151, -0.00000409, -0.00000223, },
		{ 0.00000165, -0.00000377, -0.00000990, -0.00000209, 0.00000491, -0.00000159, -0.00001255, 0.00000603, -0.00000696, -0.00000276, -0.00000060, 0.00000283, -0.00001779, -0.00000166, -0.00001980, 0.00001208, -0.00001410, 0.00003538, -0.00000935, 0.00001228, -0.00000353, 0.00000459, 0.00001187, 0.00000226, 0.00000824, 0.00000675, -0.00000202, 0.00000587, -0.00000182, 0.00000565, -0.00000102, 0.00000655, 0.00000118, 0.00000651, 0.00000172, 0.00001064, -0.00000330, 0.00002100, 0.00000342, 0.00000156, -0.00001094, -0.00001274, -0.00000216, -0.00000663, -0.00000055, -0.00000748, 0.00001061, -0.00000716, -0.00000134, -0.00000457, 0.00000069, -0.00000304, 0.00000269, -0.00000448, -0.00000088, -0.00000541, -0.00000237, -0.00000661, -0.00000487, 0.00000086, -0.00000596, -0.00000312, -0.00000158, -0.00000639, -0.00000652, -0.00000894, -0.00000873, -0.00000686, -0.00002301, -0.00001717, -0.00004052, -0.00003451, -0.00021586, 0.00011059, 0.00020382, 0.00003355, 0.00009018, 0.00002270, 0.00011462, 0.00002682, 0.00023279, 0.00001269, -0.00019152, -0.00002612, -0.00003883, -0.00000933, -0.00003099, -0.00000620, -0.00001335, 0.00000265, -0.00002006, -0.00000116, -0.00000623, 0.00000028, -0.00000212, -0.00000527, -0.00000176, -0.00000004, -0.00001354, 0.00000203, 0.00000478, -0.00000226, -0.00000087, -0.00000494, 0.00000813, -0.00001257, 0.00001451, -0.00001405, 0.00000146, -0.00000262, -0.00001528, 0.00000589, -0.00000400, -0.00000125, -0.00000389, -0.00000209, -0.00000379, -0.00000120, -0.00000622, 0.00000100, -0.00000131, -0.00000043, 0.00000126, -0.00000393, 0.00000208, -0.00000364, -0.00001105, 0.00000047, },
		{ 0.00000088, 0.00000090, -0.00001331, -0.00000107, -0.00000068, 0.00000677, -0.00001625, 0.00000556, -0.00000581, -0.00000093, -0.00000860, 0.00001244, -0.00002063, 0.00000234, -0.00002250, 0.00000316, -0.00000711, 0.00001438, 0.00002092, -0.00001055, 0.00001372, -0.00000734, 0.00001657, 0.00000300, 0.00000636, 0.00000810, 0.00000151, 0.00000249, 0.00000174, 0.00000252, 0.00000085, 0.00000503, 0.00000334, 0.00000498, 0.00000290, 0.00000896, 0.00000008, 0.00001357, 0.00001355, -0.00000181, -0.00000364, -0.00001921, -0.00000054, -0.00000920, 0.00000211, -0.00000987, 0.00000789, -0.00000196, -0.00000407, -0.00000287, -0.00000214, -0.00000057, 0.00000026, -0.00000098, -0.00000478, -0.00000085, -0.00000547, -0.00000209, -0.00001183, 0.00000528, -0.00000747, -0.00000112, -0.00000527, -0.00000039, -0.00001054, -0.00000267, -0.00001852, 0.00000516, -0.00002866, -0.00000589, -0.00005136, -0.00000628, -0.00018588, -0.00000969, 0.00023084, 0.00005073, 0.00008072, 0.00002997, 0.00008018, 0.00005822, 0.00010450, 0.00020088, -0.00015912, -0.00009713, -0.00002482, -0.00002446, -0.00001887, -0.00002001, -0.00000901, -0.00000340, -0.00000734, -0.00001606, 0.00000200, -0.00000724, 0.00000573, -0.00001011, -0.00000090, -0.00000134, -0.00000636, -0.00000876, 0.00001065, -0.00000566, 0.00000391, -0.00001035, 0.00001082, -0.00001163, 0.00001060, -0.00000565, -0.00000844, 0.00000604, -0.00001881, 0.00000371, -0.00000229, -0.00000141, -0.00000395, -0.00000197, -0.00000437, -0.00000014, -0.00000610, -0.00000087, -0.00000084, -0.00000048, 0.00000061, -0.00000153, -0.00000187, 0.00000187, -0.00001338, -0.00000108, },
		{ 0.00000181, 0.00000272, -0.00001140, -0.00000545, -0.00000026, 0.00000575, -0.00000952, -0.00000288, 0.00000051, -0.00000357, -0.00000786, 0.00000639, -0.00000601, -0.00001163, -0.00000380, -0.00002205, 0.00001009, -0.00001067, 0.00003825, -0.00000919, 0.00000679, 0.00000151, 0.00000577, 0.00001234, 0.00000184, 0.00000924, 0.00000381, 0.00000279, 0.00000118, 0.00000376, -0.00000060, 0.00000486, 0.00000468, 0.00000438, 0.00000373, 0.00000725, 0.00000344, 0.00000731, 0.00001947, -0.00000071, 0.00000005, -0.00001991, -0.00000439, -0.00000762, -0.00000024, -0.00000607, 0.00000024, 0.00000494, -0.00000745, 0.00000029, -0.00000617, 0.00000230, -0.00000284, 0.00000348, -0.00000890, 0.00000259, -0.00000793, 0.00000279, -0.00001658, 0.00000504, -0.00000583, -0.00000024, -0.00000790, 0.00000315, -0.00001255, 0.00000283, -0.00002454, 0.00000779, -0.00002463, -0.00000397, -0.00004832, -0.00000667, -0.00012878, -0.00010514, 0.00020878, 0.00009126, 0.00006126, 0.00005184, 0.00003991, 0.00008731, 0.00003644, 0.00019659, 0.00003591, -0.00019010, -0.00001794, -0.00004522, -0.00000293, -0.00003487, -0.00000151, -0.00001760, 0.00001008, -0.00002802, 0.00000578, -0.00001300, 0.00001114, -0.00001154, -0.00000040, -0.00000600, 0.00000276, -0.00001658, 0.00001205, -0.00000635, 0.00000630, -0.00001183, 0.00000876, -0.00000717, 0.00000423, 0.00000361, -0.00001537, 0.00000821, -0.00001513, -0.00000337, 0.00000215, -0.00000410, -0.00000131, -0.00000428, -0.00000261, -0.00000207, -0.00000280, -0.00000450, 0.00000069, -0.00000202, 0.00000153, -0.00000124, -0.00000239, 0.00000210, -0.00000960, -0.00000661, },
		{ -0.00000048, 0.00000454, -0.00000737, -0.00001058, 0.00000063, 0.00000289, -0.00000260, -0.00000845, 0.00000317, -0.00000462, -0.00000432, -0.00000397, 0.00000776, -0.00002007, 0.00000536, -0.00002701, 0.00000206, -0.00000711, 0.00001994, 0.00002024, -0.00001335, 0.00001385, -0.00000225, 0.00001496, 0.00000406, 0.00000677, 0.00000567, 0.00000472, 0.00000047, 0.00000403, 0.00000063, 0.00000127, 0.00000811, 0.00000260, 0.00000595, 0.00000428, 0.00000765, 0.00000155, 0.00002270, 0.00000217, 0.00000271, -0.00001794, -0.00000920, -0.00000650, -0.00000277, -0.00000268, -0.00000438, 0.00000638, -0.00000577, -0.00000077, -0.00000516, -0.00000031, -0.00000106, 0.00000263, -0.00000623, -0.00000146, -0.00000378, 0.00000037, -0.00001077, -0.00000576, 0.00000304, -0.00000563, -0.00000276, -0.00000263, -0.00000530, -0.00000241, -0.00001483, -0.00000800, -0.00000532, -0.00001742, -0.00002560, -0.00003262, -0.00005998, -0.00017050, 0.00016045, 0.00012106, 0.00006250, 0.00005189, 0.00003952, 0.00006073, 0.00006221, 0.00008274, 0.00019672, -0.00014570, -0.00009014, -0.00002877, -0.00002039, -0.00002129, -0.00001548, -0.00001351, 0.00000392, -0.00001425, -0.00001026, -0.00000402, 0.00000158, 0.00000034, -0.00000885, -0.00000277, 0.00000028, -0.00001006, 0.00000129, 0.00000218, -0.00000043, -0.00000352, -0.00000210, 0.00000389, -0.00000727, 0.00001513, -0.00002052, 0.00000737, -0.00001042, -0.00000826, 0.00000285, -0.00000335, -0.00000199, -0.00000316, -0.00000389, -0.00000164, -0.00000220, -0.00000438, -0.00000103, -0.00000126, 0.00000031, 0.00000049, -0.00000333, 0.00000201, -0.00000618, -0.00000929, },
		{ -0.00000474, 0.00000529, -0.00000291, -0.00001262, -0.00000235, 0.00000332, -0.00000095, -0.00000742, 0.00000044, -0.00000236, -0.00000334, -0.00000771, 0.00000777, -0.00001193, -0.00000376, -0.00001063, -0.00002033, 0.00000478, -0.00000197, 0.00003619, -0.00000890, 0.00000316, 0.00000826, 0.00000418, 0.00001340, 0.00000220, 0.00000799, 0.00000417, 0.00000446, -0.00000045, 0.00000641, -0.00000496, 0.00001149, 0.00000130, 0.00000825, 0.00000146, 0.00001115, -0.00000208, 0.00002231, 0.00000710, 0.00000380, -0.00001351, -0.00001392, -0.00000632, -0.00000495, -0.00000098, -0.00000536, 0.00000387, -0.00000183, -0.00000351, -0.00000185, -0.00000467, 0.00000164, -0.00000006, -0.00000083, -0.00000749, 0.00000119, -0.00000413, -0.00000154, -0.00001661, 0.00000840, -0.00000850, 0.00000186, -0.00000848, 0.00000168, -0.00000923, -0.00000153, -0.00002422, 0.00000874, -0.00002626, -0.00000610, -0.00005278, -0.00001544, -0.00018415, 0.00008096, 0.00015059, 0.00006151, 0.00005708, 0.00004132, 0.00003878, 0.00007564, 0.00003415, 0.00017660, 0.00004338, -0.00017633, -0.00002078, -0.00004308, -0.00000383, -0.00003153, -0.00000427, -0.00001271, 0.00000649, -0.00002567, 0.00000331, -0.00000918, 0.00001187, -0.00001503, 0.00000101, -0.00000664, 0.00000049, -0.00000995, 0.00000982, -0.00000746, 0.00000544, -0.00001206, 0.00001236, -0.00001503, 0.00002015, -0.00001747, 0.00000088, -0.00000415, -0.00001167, 0.00000132, -0.00000092, -0.00000426, -0.00000064, -0.00000602, -0.00000057, -0.00000307, -0.00000225, -0.00000379, 0.00000022, -0.00000208, 0.00000299, -0.00000473, 0.00000254, -0.00000461, -0.00000855, },
		{ -0.00000549, 0.00000147, 0.00000296, -0.00001282, -0.00000601, 0.00000284, 0.00000084, -0.00000694, -0.00000028, -0.00000242, -0.00000178, -0.00000792, 0.00000219, -0.00000211, -0.00001066, -0.00000117, -0.00002438, -0.00000494, -0.00000045, 0.00002089, 0.00001955, -0.00001621, 0.00001816, -0.00000101, 0.00001447, 0.00000488, 0.00000677, 0.00000361, 0.00000869, -0.00000308, 0.00000872, -0.00000531, 0.00000801, 0.00000508, 0.00000675, 0.00000274, 0.00000994, 0.00000024, 0.00001550, 0.00001612, 0.00000145, -0.00000563, -0.00001944, -0.00000514, -0.00000874, 0.00000107, -0.00000634, 0.00000261, -0.00000078, -0.00000336, -0.00000135, -0.00000498, -0.00000011, 0.00000062, 0.00000093, -0.00000801, -0.00000013, -0.00000350, 0.00000146, -0.00001671, 0.00000202, -0.00000272, -0.00000093, -0.00000602, -0.00000043, -0.00000754, 0.00000133, -0.00002390, 0.00000266, -0.00001758, -0.00000552, -0.00004746, -0.00001181, -0.00014470, -0.00002240, 0.00018192, 0.00004878, 0.00007758, 0.00002691, 0.00004954, 0.00004234, 0.00006529, 0.00006671, 0.00018909, -0.00013133, -0.00008425, -0.00003285, -0.00001649, -0.00002245, -0.00001331, -0.00001382, 0.00000615, -0.00001772, -0.00000686, -0.00000691, 0.00000893, -0.00000760, -0.00000444, -0.00000588, 0.00000071, -0.00000881, 0.00000639, -0.00000524, 0.00000484, -0.00001103, 0.00000909, -0.00001057, 0.00001165, -0.00000270, -0.00001208, 0.00000479, -0.00001633, 0.00000151, -0.00000124, -0.00000330, -0.00000186, -0.00000401, -0.00000323, -0.00000112, -0.00000297, -0.00000293, -0.00000146, -0.00000183, 0.00000224, -0.00000276, 0.00000002, -0.00000091, -0.00000941, },
		{ -0.00000421, -0.00000370, 0.00000690, -0.00001064, -0.00000842, -0.00000071, 0.00000480, -0.00000880, 0.00000139, -0.00000389, -0.00000018, -0.00000744, -0.00000099, -0.00000060, -0.00000663, -0.00000526, -0.00001167, -0.00002305, 0.00000405, 0.00000474, 0.00003286, -0.00000957, 0.00000441, 0.00001189, 0.00000312, 0.00001377, 0.00000358, 0.00000474, 0.00000822, 0.00000148, 0.00000353, 0.00000202, -0.00000107, 0.00001123, 0.00000343, 0.00000683, 0.00000537, 0.00000651, 0.00000564, 0.00002503, -0.00000015, 0.00000130, -0.00002213, -0.00000550, -0.00001146, 0.00000121, -0.00000618, 0.00000155, -0.00000055, -0.00000278, -0.00000176, -0.00000334, -0.00000322, 0.00000164, 0.00000113, -0.00000537, -0.00000414, -0.00000123, 0.00000070, -0.00001043, -0.00000804, 0.00000308, -0.00000428, -0.00000151, -0.00000492, -0.00000341, -0.00000132, -0.00001440, -0.00001078, -0.00000587, -0.00001100, -0.00003093, -0.00002647, -0.00008913, -0.00010786, 0.00018034, 0.00005770, 0.00007760, 0.00003074, 0.00005107, 0.00002338, 0.00007564, 0.00002501, 0.00016589, 0.00004159, -0.00015512, -0.00003134, -0.00003465, -0.00000934, -0.00002477, -0.00000921, -0.00000536, -0.00000179, -0.00001791, -0.00000401, 0.00000157, 0.00000290, -0.00001044, -0.00000261, -0.00000451, -0.00000256, -0.00000014, -0.00000015, -0.00000014, -0.00000443, 0.00000023, -0.00000058, -0.00000201, 0.00001309, -0.00002121, 0.00000899, -0.00001726, -0.00000031, -0.00000128, -0.00000167, -0.00000437, -0.00000075, -0.00000645, 0.00000096, -0.00000445, -0.00000076, -0.00000382, -0.00000060, -0.00000053, 0.00000120, -0.00000414, 0.00000367, -0.00001096, },
		{ -0.00000653, -0.00000418, 0.00000458, -0.00000438, -0.00001122, -0.00000346, 0.00000552, -0.00000658, -0.00000170, -0.00000020, -0.00000323, -0.00000378, -0.00000428, -0.00000055, -0.00000407, -0.00000555, -0.00000698, -0.00002183, -0.00001157, 0.00000804, 0.00001865, 0.00001824, -0.00001551, 0.00002187, -0.00000097, 0.00001381, 0.00000690, 0.00000416, 0.00000572, 0.00000802, -0.00000156, 0.00000774, -0.00000567, 0.00001140, 0.00000421, 0.00000846, 0.00000312, 0.00001023, 0.00000028, 0.00002649, 0.00000521, 0.00000201, -0.00001809, -0.00001057, -0.00000963, -0.00000352, -0.00000247, -0.00000240, 0.00000259, -0.00000506, -0.00000014, -0.00000350, -0.00000318, -0.00000088, 0.00000321, -0.00000412, -0.00000516, -0.00000250, 0.00000142, -0.00000578, -0.00001206, 0.00000124, -0.00000243, -0.00000128, -0.00000451, -0.00000404, -0.00000129, -0.00000801, -0.00001614, -0.00000478, -0.00001032, -0.00001924, -0.00003562, -0.00004971, -0.00014518, 0.00012624, 0.00010243, 0.00004659, 0.00006035, 0.00002950, 0.00004019, 0.00003993, 0.00005654, 0.00006450, 0.00016970, -0.00010704, -0.00008869, -0.00002843, -0.00001960, -0.00001721, -0.00001722, -0.00000684, -0.00000110, -0.00001175, -0.00001224, 0.00000078, 0.00000472, -0.00000742, -0.00000504, -0.00000506, -0.00000277, 0.00000024, -0.00000017, -0.00000158, -0.00000123, -0.00000453, 0.00000465, -0.00000916, 0.00001801, -0.00001758, 0.00000285, -0.00001018, -0.00000798, 0.00000263, -0.00000328, -0.00000355, -0.00000146, -0.00000514, -0.00000126, -0.00000251, -0.00000184, -0.00000258, -0.00000205, -0.00000129, 0.00000237, -0.00000474, 0.00000381, -0.00000800, },
		{ -0.00001102, -0.00000190, -0.00000055, 0.00000206, -0.00001294, -0.00000435, 0.00000166, 0.00000020, -0.00000875, 0.00000600, -0.00000721, 0.00000005, -0.00000763, 0.00000197, -0.00000799, 0.00000164, -0.00001208, -0.00000814, -0.00002950, 0.00000844, 0.00000676, 0.00002972, -0.00000905, 0.00000816, 0.00001253, 0.00000270, 0.00001462, 0.00000334, 0.00000391, 0.00001030, 0.00000014, 0.00000617, -0.00000212, 0.00000504, 0.00000819, 0.00000756, 0.00000438, 0.00000959, 0.00000128, 0.00002016, 0.00001584, -0.00000075, -0.00001080, -0.00001704, -0.00000610, -0.00000970, 0.00000136, -0.00000640, 0.00000558, -0.00000657, 0.00000022, -0.00000324, -0.00000227, -0.00000362, 0.00000391, -0.00000222, -0.00000511, -0.00000429, 0.00000107, -0.00000191, -0.00001202, -0.00000306, -0.00000106, -0.00000113, -0.00000308, -0.00000505, -0.00000193, -0.00000335, -0.00001617, -0.00000737, -0.00001052, -0.00000930, -0.00003895, -0.00002755, -0.00014311, 0.00004442, 0.00014855, 0.00002405, 0.00007533, 0.00002197, 0.00004993, 0.00001725, 0.00006860, 0.00002523, 0.00014801, 0.00004680, -0.00014176, -0.00003605, -0.00003118, -0.00000970, -0.00002368, -0.00000689, -0.00000675, -0.00000233, -0.00001703, -0.00000322, 0.00000495, -0.00000295, -0.00000697, -0.00000453, -0.00000563, 0.00000177, 0.00000027, -0.00000299, 0.00000046, -0.00000645, 0.00000582, -0.00000997, 0.00001375, -0.00000628, -0.00000701, -0.00000116, -0.00001599, 0.00000562, -0.00000444, -0.00000205, -0.00000379, -0.00000205, -0.00000468, 0.00000009, -0.00000394, -0.00000015, -0.00000393, -0.00000113, 0.00000106, -0.00000213, 0.00000020, -0.00000176, },
		{ -0.00001157, -0.00000308, -0.00000202, 0.00000343, -0.00001053, -0.00000627, -0.00000164, 0.00000421, -0.00001091, 0.00000533, -0.00000403, -0.00000196, -0.00000661, 0.00000180, -0.00000909, 0.00000262, -0.00000940, -0.00000610, -0.00002467, -0.00001218, 0.00001227, 0.00001638, 0.00001653, -0.00001060, 0.00002231, -0.00000032, 0.00001301, 0.00000875, 0.00000147, 0.00000924, 0.00000509, 0.00000300, 0.00000227, 0.00000041, 0.00000880, 0.00000817, 0.00000593, 0.00000798, 0.00000438, 0.00001227, 0.00002417, 0.00000026, -0.00000666, -0.00001953, -0.00000588, -0.00001198, 0.00000046, -0.00000614, 0.00000393, -0.00000313, -0.00000363, -0.00000034, -0.00000346, -0.00000288, 0.00000039, 0.00000222, -0.00000661, -0.00000319, -0.00000286, 0.00000307, -0.00001201, -0.00000414, -0.00000411, 0.00000136, -0.00000379, -0.00000268, -0.00000611, 0.00000238, -0.00001653, -0.00000601, -0.00001614, 0.00000022, -0.00003834, -0.00001714, -0.00011954, -0.00003335, 0.00016210, 0.00003771, 0.00005736, 0.00004025, 0.00003710, 0.00002782, 0.00003999, 0.00004874, 0.00006247, 0.00015160, -0.00008448, -0.00009363, -0.00002353, -0.00002217, -0.00001350, -0.00001821, -0.00000325, -0.00000674, -0.00000655, -0.00001610, 0.00000879, -0.00000364, -0.00000300, -0.00000773, -0.00000440, -0.00000287, 0.00000620, -0.00000727, 0.00000338, -0.00000875, 0.00000645, -0.00000873, 0.00000713, 0.00000354, -0.00001201, 0.00000265, -0.00001770, 0.00000206, -0.00000085, -0.00000369, -0.00000345, -0.00000207, -0.00000442, -0.00000092, -0.00000304, -0.00000073, -0.00000258, -0.00000286, 0.00000044, -0.00000046, -0.00000210, 0.00000186, },
		{ -0.00000800, -0.00000720, -0.00000084, 0.00000171, -0.00000620, -0.00000893, -0.00000228, 0.00000277, -0.00000630, -0.00000209, 0.00000368, -0.00000625, -0.00000357, -0.00000160, -0.00000399, -0.00000565, 0.00000203, -0.00001426, -0.00000798, -0.00003235, 0.00001163, 0.00000680, 0.00002628, -0.00000476, 0.00001022, 0.00001271, 0.00000207, 0.00001627, 0.00000122, 0.00000666, 0.00000807, 0.00000376, 0.00000276, 0.00000068, 0.00000570, 0.00000935, 0.00000767, 0.00000675, 0.00000711, 0.00000662, 0.00002668, 0.00000662, -0.00000574, -0.00001863, -0.00000853, -0.00001076, -0.00000385, -0.00000343, -0.00000040, 0.00000221, -0.00000755, 0.00000137, -0.00000406, -0.00000139, -0.00000332, 0.00000488, -0.00000631, -0.00000249, -0.00000576, 0.00000468, -0.00000922, -0.00000543, -0.00000572, 0.00000033, -0.00000216, -0.00000161, -0.00000793, 0.00000352, -0.00001301, -0.00000644, -0.00001726, -0.00000089, -0.00002688, -0.00002108, -0.00008163, -0.00009474, 0.00014149, 0.00007218, 0.00003436, 0.00005279, 0.00003119, 0.00003616, 0.00002106, 0.00005565, 0.00003195, 0.00012539, 0.00005753, -0.00013585, -0.00003472, -0.00003207, -0.00000670, -0.00002513, -0.00000154, -0.00001327, 0.00000269, -0.00002114, 0.00000470, -0.00000122, -0.00000130, -0.00000794, -0.00000392, -0.00000768, 0.00000941, -0.00000667, 0.00000187, -0.00000724, 0.00000323, -0.00000370, -0.00000123, 0.00001066, -0.00001274, 0.00000279, -0.00001483, -0.00000517, 0.00000376, -0.00000531, -0.00000258, -0.00000343, -0.00000263, -0.00000303, -0.00000141, -0.00000237, -0.00000032, -0.00000430, -0.00000031, 0.00000001, -0.00000233, 0.00000201, },
		{ -0.00000466, -0.00000924, -0.00000194, 0.00000167, -0.00000384, -0.00000992, -0.00000274, 0.00000028, -0.00000170, -0.00000730, 0.00000621, -0.00000471, -0.00000430, -0.00000344, 0.00000070, -0.00001181, 0.00000676, -0.00001383, -0.00000410, -0.00002832, -0.00000975, 0.00001208, 0.00001460, 0.00001703, -0.00000505, 0.00002102, 0.00000032, 0.00001326, 0.00000828, 0.00000240, 0.00000914, 0.00000587, 0.00000348, 0.00000064, 0.00000482, 0.00000687, 0.00001122, 0.00000526, 0.00000971, 0.00000311, 0.00002500, 0.00001445, -0.00000367, -0.00001799, -0.00001105, -0.00000930, -0.00000746, -0.00000235, -0.00000316, 0.00000444, -0.00000679, -0.00000127, -0.00000188, -0.00000208, -0.00000340, 0.00000262, -0.00000211, -0.00000480, -0.00000426, 0.00000057, -0.00000238, -0.00000951, -0.00000246, -0.00000580, 0.00000230, -0.00000355, -0.00000467, -0.00000186, -0.00000417, -0.00001210, -0.00000940, -0.00001319, -0.00000827, -0.00003265, -0.00004197, -0.00013022, 0.00009452, 0.00010551, 0.00002905, 0.00004511, 0.00004026, 0.00003098, 0.00002641, 0.00003223, 0.00004998, 0.00005240, 0.00014321, -0.00007216, -0.00009119, -0.00002436, -0.00001940, -0.00001511, -0.00001272, -0.00000803, -0.00000430, -0.00000877, -0.00001076, 0.00000701, -0.00000586, -0.00000186, -0.00000785, -0.00000652, 0.00000343, 0.00000251, -0.00000554, -0.00000084, -0.00000464, 0.00000486, -0.00001023, 0.00001531, -0.00001156, 0.00000211, -0.00001079, -0.00001117, 0.00000438, -0.00000314, -0.00000417, -0.00000289, -0.00000307, -0.00000250, -0.00000231, -0.00000187, -0.00000060, -0.00000268, -0.00000290, 0.00000101, -0.00000283, 0.00000181, },
		{ -0.00000217, -0.00000893, -0.00000470, 0.00000207, -0.00000191, -0.00001031, -0.00000369, -0.00000068, -0.00000051, -0.00000719, 0.00000243, 0.00000095, -0.00000618, -0.00000444, 0.00000221, -0.00001056, 0.00000069, -0.00000347, -0.00001131, -0.00001217, -0.00002944, 0.00000972, 0.00000706, 0.00002389, 0.00000190, 0.00000996, 0.00001296, 0.00000215, 0.00001647, 0.00000094, 0.00000845, 0.00000632, 0.00000715, -0.00000147, 0.00000655, 0.00000296, 0.00001322, 0.00000584, 0.00001066, 0.00000262, 0.00002011, 0.00002159, 0.00000060, -0.00001712, -0.00001417, -0.00000769, -0.00000985, -0.00000281, -0.00000455, 0.00000386, -0.00000320, -0.00000487, -0.00000025, -0.00000272, -0.00000213, -0.00000090, 0.00000173, -0.00000646, -0.00000219, -0.00000389, 0.00000261, -0.00001142, -0.00000013, -0.00000925, 0.00000238, -0.00000300, -0.00000209, -0.00000636, 0.00000194, -0.00001522, -0.00000276, -0.00002152, 0.00000057, -0.00003462, -0.00001995, -0.00012841, 0.00002552, 0.00013108, 0.00003523, 0.00003690, 0.00004033, 0.00003473, 0.00002737, 0.00002092, 0.00004908, 0.00003258, 0.00011035, 0.00006202, -0.00012694, -0.00003591, -0.00002972, -0.00000767, -0.00002122, -0.00000280, -0.00001429, 0.00000301, -0.00001935, 0.00000770, -0.00000666, 0.00000121, -0.00000885, -0.00000498, -0.00000433, 0.00001013, -0.00000891, 0.00000221, -0.00000945, 0.00000905, -0.00001278, 0.00001279, -0.00000562, -0.00000196, -0.00000412, -0.00001617, 0.00000224, 0.00000004, -0.00000537, -0.00000258, -0.00000389, -0.00000153, -0.00000330, -0.00000138, -0.00000148, -0.00000055, -0.00000463, 0.00000089, -0.00000285, 0.00000120, },
		{ 0.00000110, -0.00000890, -0.00000560, -0.00000041, 0.00000204, -0.00001138, -0.00000464, -0.00000103, -0.00000031, -0.00000610, -0.00000102, 0.00000302, -0.00000339, -0.00000758, 0.00000254, -0.00000636, -0.00000639, 0.00000283, -0.00001232, -0.00000682, -0.00002672, -0.00001038, 0.00001041, 0.00001344, 0.00002065, -0.00000154, 0.00001961, 0.00000106, 0.00001347, 0.00000733, 0.00000527, 0.00000645, 0.00000974, -0.00000038, 0.00000502, 0.00000386, 0.00000930, 0.00001040, 0.00000875, 0.00000591, 0.00001253, 0.00002761, 0.00000551, -0.00001355, -0.00001933, -0.00000556, -0.00001176, -0.00000320, -0.00000622, 0.00000272, -0.00000045, -0.00000548, -0.00000180, -0.00000151, -0.00000210, -0.00000201, 0.00000179, -0.00000475, -0.00000287, -0.00000438, 0.00000195, -0.00000811, -0.00000273, -0.00000588, -0.00000382, 0.00000094, -0.00000268, -0.00000581, 0.00000103, -0.00001128, -0.00000419, -0.00001767, -0.00000569, -0.00002363, -0.00001826, -0.00009697, -0.00004772, 0.00013897, 0.00004751, 0.00003845, 0.00002702, 0.00004690, 0.00002052, 0.00002812, 0.00002458, 0.00005138, 0.00004289, 0.00013658, -0.00006352, -0.00008656, -0.00002577, -0.00001703, -0.00001527, -0.00000899, -0.00001285, -0.00000106, -0.00001103, -0.00000347, -0.00000026, -0.00000299, -0.00000345, -0.00000736, -0.00000647, 0.00000811, -0.00000283, -0.00000172, -0.00000735, 0.00000494, -0.00000631, 0.00000257, 0.00000493, -0.00000941, 0.00000431, -0.00001963, -0.00000081, 0.00000107, -0.00000355, -0.00000454, -0.00000277, -0.00000286, -0.00000173, -0.00000315, -0.00000058, -0.00000080, -0.00000307, -0.00000178, -0.00000108, -0.00000125, },
		{ 0.00000330, -0.00000801, -0.00000574, -0.00000361, 0.00000480, -0.00000990, -0.00000759, -0.00000068, 0.00000008, -0.00000577, -0.00000230, 0.00000139, 0.00000120, -0.00000901, 0.00000013, -0.00000170, -0.00000920, 0.00000102, -0.00000583, -0.00001037, -0.00001429, -0.00002590, 0.00000428, 0.00000773, 0.00002454, 0.00000759, 0.00000904, 0.00001292, 0.00000320, 0.00001467, 0.00000410, 0.00000652, 0.00000863, 0.00000556, -0.00000006, 0.00000834, 0.00000320, 0.00001429, 0.00000745, 0.00000979, 0.00000602, 0.00002975, 0.00001194, -0.00000850, -0.00002359, -0.00000610, -0.00001120, -0.00000476, -0.00000671, 0.00000011, 0.00000172, -0.00000446, -0.00000368, -0.00000123, -0.00000167, -0.00000217, 0.00000055, -0.00000263, -0.00000393, -0.00000333, -0.00000065, -0.00000395, -0.00000616, -0.00000121, -0.00000876, 0.00000169, -0.00000202, -0.00000431, -0.00000170, -0.00000599, -0.00000796, -0.00000985, -0.00001436, -0.00001350, -0.00002161, -0.00006078, -0.00009714, 0.00011409, 0.00007170, 0.00003788, 0.00002243, 0.00004310, 0.00002756, 0.00002685, 0.00001560, 0.00004844, 0.00002752, 0.00010291, 0.00005926, -0.00011361, -0.00004013, -0.00002476, -0.00001095, -0.00001437, -0.00000909, -0.00000997, -0.00000136, -0.00001079, 0.00000186, -0.00000565, 0.00000028, -0.00000794, -0.00000666, 0.00000198, 0.00000441, -0.00000468, -0.00000462, -0.00000122, 0.00000166, -0.00000689, 0.00001218, -0.00001305, 0.00000753, -0.00001670, -0.00000687, 0.00000182, -0.00000168, -0.00000564, -0.00000232, -0.00000406, -0.00000035, -0.00000427, -0.00000031, -0.00000134, -0.00000097, -0.00000393, 0.00000009, -0.00000349, },
		{ 0.00000209, -0.00000421, -0.00000773, -0.00000395, 0.00000285, -0.00000382, -0.00001270, -0.00000001, 0.00000002, -0.00000421, -0.00000462, 0.00000120, 0.00000140, -0.00000451, -0.00000599, 0.00000293, -0.00000951, -0.00000197, -0.00000269, -0.00000882, -0.00001135, -0.00002114, -0.00001562, 0.00000872, 0.00001455, 0.00002502, 0.00000044, 0.00001804, 0.00000284, 0.00001169, 0.00000926, 0.00000563, 0.00000624, 0.00001080, -0.00000180, 0.00000905, 0.00000244, 0.00001201, 0.00001000, 0.00001101, 0.00000432, 0.00002564, 0.00002097, -0.00000428, -0.00002347, -0.00001165, -0.00000745, -0.00000816, -0.00000489, -0.00000454, 0.00000402, -0.00000398, -0.00000324, -0.00000340, -0.00000009, -0.00000283, 0.00000043, -0.00000236, -0.00000349, -0.00000329, -0.00000162, -0.00000247, -0.00000694, -0.00000002, -0.00000779, -0.00000280, 0.00000065, -0.00000399, -0.00000250, -0.00000371, -0.00000884, -0.00000589, -0.00001601, -0.00001169, -0.00002147, -0.00003494, -0.00011169, 0.00005982, 0.00010161, 0.00003236, 0.00003277, 0.00002325, 0.00004399, 0.00001762, 0.00002426, 0.00002397, 0.00004596, 0.00004111, 0.00012418, -0.00005157, -0.00008513, -0.00002353, -0.00001865, -0.00001028, -0.00001218, -0.00001124, -0.00000350, -0.00000589, -0.00000390, -0.00000358, -0.00000146, -0.00000401, -0.00000836, -0.00000162, 0.00000535, -0.00000166, -0.00000494, -0.00000459, 0.00000495, -0.00000939, 0.00001112, -0.00000876, 0.00000249, -0.00000687, -0.00001517, 0.00000308, -0.00000200, -0.00000378, -0.00000433, -0.00000301, -0.00000169, -0.00000243, -0.00000251, -0.00000017, -0.00000102, -0.00000305, -0.00000128, -0.00000331, },
		{ -0.00000067, 0.00000019, -0.00000959, -0.00000323, -0.00000094, 0.00000229, -0.00001505, -0.00000221, 0.00000053, -0.00000175, -0.00000751, 0.00000199, -0.00000088, 0.00000104, -0.00000967, 0.00000346, -0.00000719, -0.00000357, -0.00000427, -0.00000329, -0.00001369, -0.00001098, -0.00002641, -0.00000198, 0.00000951, 0.00002764, 0.00001118, 0.00000812, 0.00001346, 0.00000326, 0.00001397, 0.00000718, 0.00000431, 0.00001141, 0.00000248, 0.00000488, 0.00000650, 0.00000651, 0.00001293, 0.00001130, 0.00000622, 0.00001804, 0.00002930, 0.00000087, -0.00002042, -0.00001889, -0.00000463, -0.00001031, -0.00000337, -0.00000811, 0.00000397, -0.00000263, -0.00000251, -0.00000483, -0.00000029, -0.00000219, -0.00000010, -0.00000181, -0.00000378, -0.00000270, -0.00000232, -0.00000164, -0.00000715, -0.00000047, -0.00000456, -0.00000662, 0.00000035, -0.00000203, -0.00000345, -0.00000202, -0.00000902, -0.00000409, -0.00001413, -0.00001184, -0.00002223, -0.00001835, -0.00010159, -0.00000123, 0.00011712, 0.00003531, 0.00004070, 0.00001256, 0.00004418, 0.00002202, 0.00002494, 0.00001386, 0.00004452, 0.00002605, 0.00009397, 0.00005716, -0.00010213, -0.00004173, -0.00002293, -0.00001002, -0.00001269, -0.00001161, -0.00000857, -0.00000160, -0.00000525, -0.00000505, -0.00000239, -0.00000164, -0.00000764, -0.00000452, 0.00000340, 0.00000157, -0.00000344, -0.00000768, 0.00000528, -0.00000753, 0.00000571, -0.00000137, -0.00000484, 0.00000246, -0.00001931, 0.00000114, -0.00000171, -0.00000219, -0.00000560, -0.00000238, -0.00000317, -0.00000059, -0.00000410, 0.00000026, -0.00000128, -0.00000149, -0.00000269, -0.00000254, },
		{ -0.00000239, 0.00000195, -0.00000832, -0.00000468, -0.00000228, 0.00000373, -0.00001139, -0.00000811, 0.00000181, -0.00000021, -0.00000763, -0.00000033, -0.00000014, 0.00000117, -0.00000629, -0.00000156, -0.00000234, -0.00000558, -0.00000433, -0.00000318, -0.00000930, -0.00001166, -0.00001726, -0.00002302, 0.00000738, 0.00001869, 0.00002790, 0.00000192, 0.00001676, 0.00000467, 0.00000924, 0.00001323, 0.00000356, 0.00000885, 0.00000791, 0.00000208, 0.00000860, 0.00000465, 0.00001157, 0.00001300, 0.00000885, 0.00001162, 0.00003245, 0.00000966, -0.00001744, -0.00002338, -0.00000645, -0.00000912, -0.00000423, -0.00000818, 0.00000016, 0.00000049, -0.00000333, -0.00000345, -0.00000321, -0.00000007, -0.00000173, 0.00000023, -0.00000559, -0.00000129, -0.00000352, -0.00000010, -0.00000795, -0.00000107, -0.00000232, -0.00000652, -0.00000357, 0.00000077, -0.00000457, -0.00000002, -0.00000952, -0.00000283, -0.00001231, -0.00000989, -0.00002438, -0.00001067, -0.00007800, -0.00005118, 0.00010803, 0.00005155, 0.00003758, 0.00001990, 0.00002692, 0.00003596, 0.00001842, 0.00001930, 0.00002482, 0.00003938, 0.00004181, 0.00010994, -0.00003921, -0.00008372, -0.00002159, -0.00001931, -0.00000630, -0.00001636, -0.00000804, -0.00000662, 0.00000184, -0.00001029, -0.00000174, -0.00000257, -0.00000382, -0.00000785, 0.00000242, 0.00000065, 0.00000155, -0.00001060, 0.00000387, -0.00000438, 0.00000037, 0.00000408, -0.00000875, 0.00000567, -0.00001613, -0.00000505, 0.00000061, -0.00000308, -0.00000415, -0.00000393, -0.00000269, -0.00000122, -0.00000288, -0.00000161, -0.00000015, -0.00000147, -0.00000206, -0.00000295, },
		{ -0.00000372, 0.00000158, -0.00000512, -0.00000703, -0.00000218, 0.00000220, -0.00000540, -0.00001306, 0.00000098, 0.00000087, -0.00000504, -0.00000486, 0.00000241, -0.00000154, -0.00000112, -0.00000567, 0.00000030, -0.00000655, -0.00000238, -0.00000741, -0.00000187, -0.00001607, -0.00000598, -0.00003068, -0.00000794, 0.00001351, 0.00003093, 0.00001336, 0.00000738, 0.00001472, 0.00000166, 0.00001614, 0.00000736, 0.00000520, 0.00001035, 0.00000360, 0.00000718, 0.00000642, 0.00000831, 0.00001352, 0.00001231, 0.00000793, 0.00003015, 0.00002049, -0.00001354, -0.00002539, -0.00001111, -0.00000733, -0.00000554, -0.00000656, -0.00000450, 0.00000255, -0.00000372, -0.00000176, -0.00000533, 0.00000030, -0.00000213, 0.00000141, -0.00000587, -0.00000197, -0.00000297, -0.00000004, -0.00000675, -0.00000379, 0.00000019, -0.00000542, -0.00000590, -0.00000018, -0.00000296, -0.00000031, -0.00000740, -0.00000460, -0.00000851, -0.00000975, -0.00002136, -0.00001489, -0.00004751, -0.00008385, 0.00007952, 0.00007131, 0.00003239, 0.00003064, 0.00001460, 0.00003715, 0.00002252, 0.00001929, 0.00001651, 0.00003660, 0.00002929, 0.00008145, 0.00005789, -0.00009352, -0.00004071, -0.00002344, -0.00000611, -0.00001601, -0.00000916, -0.00001128, 0.00000436, -0.00000775, -0.00000664, -0.00000147, -0.00000247, -0.00000761, 0.00000002, -0.00000037, 0.00000374, -0.00000811, -0.00000170, 0.00000071, -0.00000497, 0.00000765, -0.00000955, 0.00000444, -0.00000969, -0.00001090, 0.00000178, -0.00000381, -0.00000286, -0.00000496, -0.00000243, -0.00000218, -0.00000145, -0.00000301, 0.00000016, -0.00000135, -0.00000128, -0.00000286, },
		{ -0.00000510, 0.00000046, -0.00000250, -0.00000772, -0.00000319, 0.00000115, -0.00000149, -0.00001356, -0.00000300, 0.00000183, -0.00000208, -0.00000738, 0.00000190, -0.00000126, -0.00000059, -0.00000415, -0.00000131, -0.00000533, -0.00000162, -0.00000876, -0.00000112, -0.00001239, -0.00000745, -0.00001793, -0.00003023, 0.00000711, 0.00002448, 0.00002906, 0.00000316, 0.00001661, 0.00000475, 0.00000912, 0.00001547, 0.00000307, 0.00000972, 0.00000630, 0.00000659, 0.00000697, 0.00000797, 0.00001053, 0.00001613, 0.00000723, 0.00002470, 0.00002952, -0.00000610, -0.00002697, -0.00001542, -0.00000783, -0.00000569, -0.00000527, -0.00000713, 0.00000147, -0.00000227, -0.00000182, -0.00000454, -0.00000207, -0.00000087, 0.00000067, -0.00000336, -0.00000521, -0.00000079, -0.00000170, -0.00000324, -0.00000809, 0.00000251, -0.00000474, -0.00000493, -0.00000430, -0.00000023, -0.00000206, -0.00000313, -0.00000849, -0.00000344, -0.00001204, -0.00001327, -0.00002485, -0.00002189, -0.00009411, 0.00003796, 0.00008660, 0.00003200, 0.00003474, 0.00001616, 0.00002501, 0.00003225, 0.00001609, 0.00001873, 0.00002153, 0.00003753, 0.00003873, 0.00009939, -0.00003049, -0.00007905, -0.00002302, -0.00001557, -0.00000825, -0.00001548, -0.00000903, -0.00000371, 0.00000343, -0.00001446, 0.00000024, -0.00000376, -0.00000414, -0.00000400, 0.00000104, 0.00000052, -0.00000059, -0.00000891, 0.00000532, -0.00000874, 0.00000876, -0.00000782, 0.00000116, -0.00000381, -0.00001345, 0.00000021, -0.00000273, -0.00000362, -0.00000398, -0.00000364, -0.00000187, -0.00000177, -0.00000223, -0.00000137, -0.00000017, -0.00000171, -0.00000134, },
		{ -0.00000469, -0.00000192, -0.00000047, -0.00000725, -0.00000452, 0.00000012, 0.00000089, -0.00001169, -0.00000689, 0.00000099, 0.00000034, -0.00000664, -0.00000167, 0.00000121, -0.00000251, -0.00000143, -0.00000219, -0.00000498, -0.00000181, -0.00000634, -0.00000583, -0.00000467, -0.00001340, -0.00000436, -0.00003589, -0.00001313, 0.00001942, 0.00003323, 0.00001458, 0.00000780, 0.00001494, 0.00000092, 0.00001859, 0.00000690, 0.00000694, 0.00000758, 0.00000811, 0.00000621, 0.00000922, 0.00000773, 0.00001641, 0.00001066, 0.00001784, 0.00003487, 0.00000460, -0.00002715, -0.00001967, -0.00000937, -0.00000652, -0.00000364, -0.00000814, -0.00000109, -0.00000093, -0.00000216, -0.00000295, -0.00000410, -0.00000067, 0.00000003, -0.00000055, -0.00000748, -0.00000048, -0.00000213, -0.00000077, -0.00000989, 0.00000129, -0.00000227, -0.00000419, -0.00000619, -0.00000110, -0.00000178, -0.00000060, -0.00000973, -0.00000187, -0.00001179, -0.00000791, -0.00002829, -0.00001284, -0.00008047, -0.00001018, 0.00009403, 0.00003584, 0.00003490, 0.00002164, 0.00001737, 0.00002992, 0.00002296, 0.00001560, 0.00001728, 0.00003119, 0.00003081, 0.00007068, 0.00005731, -0.00008378, -0.00004151, -0.00002103, -0.00000586, -0.00001729, -0.00000807, -0.00001122, 0.00000883, -0.00001285, -0.00000484, -0.00000237, -0.00000298, -0.00000518, 0.00000135, -0.00000256, 0.00000395, -0.00001028, 0.00000411, -0.00000714, 0.00000479, -0.00000222, -0.00000427, 0.00000157, -0.00001424, -0.00000182, -0.00000172, -0.00000440, -0.00000344, -0.00000407, -0.00000209, -0.00000206, -0.00000154, -0.00000221, -0.00000004, -0.00000159, -0.00000050, },
		{ -0.00000230, -0.00000487, 0.00000084, -0.00000660, -0.00000491, -0.00000193, 0.00000303, -0.00000965, -0.00000885, -0.00000137, 0.00000134, -0.00000394, -0.00000480, 0.00000164, -0.00000200, -0.00000238, 0.00000044, -0.00000593, -0.00000265, -0.00000310, -0.00000909, -0.00000213, -0.00001247, -0.00000368, -0.00002151, -0.00003625, 0.00000818, 0.00003033, 0.00002885, 0.00000501, 0.00001660, 0.00000405, 0.00001104, 0.00001550, 0.00000497, 0.00000738, 0.00000889, 0.00000773, 0.00000809, 0.00000904, 0.00001187, 0.00001624, 0.00001237, 0.00003568, 0.00001661, -0.00002382, -0.00002509, -0.00001028, -0.00000914, -0.00000184, -0.00000800, -0.00000338, -0.00000115, -0.00000176, -0.00000229, -0.00000403, -0.00000226, -0.00000001, 0.00000098, -0.00000681, -0.00000311, -0.00000077, -0.00000049, -0.00000795, -0.00000342, 0.00000149, -0.00000466, -0.00000461, -0.00000488, -0.00000054, -0.00000011, -0.00000740, -0.00000433, -0.00000804, -0.00000798, -0.00002254, -0.00001759, -0.00005509, -0.00005067, 0.00008653, 0.00004575, 0.00003256, 0.00002511, 0.00001921, 0.00001853, 0.00003177, 0.00001261, 0.00001948, 0.00001743, 0.00003732, 0.00003396, 0.00009097, -0.00002327, -0.00007404, -0.00002411, -0.00001198, -0.00001209, -0.00001244, -0.00001112, 0.00000238, -0.00000133, -0.00001354, -0.00000021, -0.00000413, -0.00000343, -0.00000058, -0.00000210, 0.00000206, -0.00000502, -0.00000171, -0.00000183, -0.00000222, 0.00000476, -0.00000939, 0.00000512, -0.00001351, -0.00000381, -0.00000121, -0.00000421, -0.00000444, -0.00000309, -0.00000330, -0.00000147, -0.00000219, -0.00000131, -0.00000147, -0.00000043, -0.00000118, },
		{ -0.00000063, -0.00000567, -0.00000010, -0.00000493, -0.00000551, -0.00000334, 0.00000331, -0.00000611, -0.00001094, -0.00000243, -0.00000036, -0.00000059, -0.00000576, -0.00000031, 0.00000006, -0.00000438, 0.00000191, -0.00000373, -0.00000591, -0.00000014, -0.00000909, -0.00000432, -0.00000780, -0.00000795, -0.00000716, -0.00003992, -0.00001730, 0.00002592, 0.00003439, 0.00001550, 0.00000938, 0.00001346, 0.00000277, 0.00001877, 0.00000865, 0.00000599, 0.00000775, 0.00001108, 0.00000612, 0.00001152, 0.00000789, 0.00001810, 0.00001200, 0.00003135, 0.00002854, -0.00001673, -0.00003032, -0.00001205, -0.00001122, -0.00000249, -0.00000564, -0.00000571, -0.00000171, -0.00000202, -0.00000157, -0.00000349, -0.00000322, -0.00000122, 0.00000208, -0.00000468, -0.00000590, -0.00000042, -0.00000027, -0.00000508, -0.00000749, 0.00000273, -0.00000365, -0.00000324, -0.00000654, -0.00000216, 0.00000081, -0.00000456, -0.00000666, -0.00000505, -0.00000890, -0.00001504, -0.00002326, -0.00003510, -0.00007030, 0.00005955, 0.00006263, 0.00002717, 0.00002872, 0.00002131, 0.00001441, 0.00002698, 0.00002076, 0.00001511, 0.00001511, 0.00002961, 0.00002861, 0.00006352, 0.00005456, -0.00007277, -0.00004383, -0.00001611, -0.00001007, -0.00001450, -0.00001017, -0.00000572, 0.00000632, -0.00001403, -0.00000409, -0.00000308, -0.00000309, -0.00000138, -0.00000096, -0.00000128, -0.00000041, -0.00000500, 0.00000112, -0.00000643, 0.00000751, -0.00000953, 0.00000369, -0.00000913, -0.00000785, 0.00000064, -0.00000450, -0.00000503, -0.00000315, -0.00000328, -0.00000180, -0.00000234, -0.00000095, -0.00000194, -0.00000035, -0.00000139, },
		{ -0.00000067, -0.00000390, -0.00000277, -0.00000257, -0.00000680, -0.00000319, 0.00000107, -0.00000088, -0.00001345, -0.00000231, -0.00000287, 0.00000106, -0.00000443, -0.00000246, 0.00000057, -0.00000369, -0.00000091, 0.00000191, -0.00000931, 0.00000052, -0.00000676, -0.00000656, -0.00000609, -0.00000797, -0.00000418, -0.00002501, -0.00004097, 0.00000993, 0.00003548, 0.00002776, 0.00000821, 0.00001513, 0.00000515, 0.00001199, 0.00001623, 0.00000599, 0.00000561, 0.00001254, 0.00000743, 0.00001079, 0.00000887, 0.00001425, 0.00001577, 0.00002459, 0.00003782, -0.00000597, -0.00003351, -0.00001632, -0.00001111, -0.00000622, -0.00000210, -0.00000741, -0.00000207, -0.00000330, -0.00000065, -0.00000335, -0.00000270, -0.00000334, 0.00000237, -0.00000213, -0.00000705, -0.00000202, 0.00000042, -0.00000281, -0.00000880, 0.00000051, -0.00000095, -0.00000336, -0.00000498, -0.00000574, 0.00000114, -0.00000200, -0.00000708, -0.00000451, -0.00000827, -0.00001048, -0.00002352, -0.00002560, -0.00007065, 0.00002158, 0.00007762, 0.00002472, 0.00003069, 0.00002065, 0.00001943, 0.00001430, 0.00002975, 0.00001180, 0.00001761, 0.00001675, 0.00003421, 0.00003144, 0.00008175, -0.00001503, -0.00007131, -0.00002175, -0.00001307, -0.00001245, -0.00001213, -0.00000856, 0.00000401, -0.00000570, -0.00001100, -0.00000182, -0.00000394, -0.00000088, -0.00000070, -0.00000232, -0.00000030, -0.00000385, 0.00000073, -0.00000660, 0.00000508, -0.00000478, -0.00000165, -0.00000243, -0.00001282, 0.00000261, -0.00000457, -0.00000486, -0.00000479, -0.00000196, -0.00000302, -0.00000162, -0.00000186, -0.00000088, -0.00000158, -0.00000066, },
		{ -0.00000070, -0.00000208, -0.00000429, -0.00000163, -0.00000717, -0.00000333, -0.00000116, 0.00000284, -0.00001314, -0.00000455, -0.00000244, -0.00000100, -0.00000131, -0.00000383, -0.00000026, -0.00000190, -0.00000420, 0.00000484, -0.00000726, -0.00000350, -0.00000276, -0.00000769, -0.00000644, -0.00000673, -0.00000483, -0.00001180, -0.00004218, -0.00002073, 0.00003200, 0.00003450, 0.00001738, 0.00001035, 0.00001244, 0.00000531, 0.00001804, 0.00001125, 0.00000357, 0.00001080, 0.00001132, 0.00000857, 0.00001178, 0.00001002, 0.00001770, 0.00002041, 0.00004120, 0.00000874, -0.00003411, -0.00002222, -0.00001062, -0.00000993, -0.00000095, -0.00000650, -0.00000332, -0.00000391, -0.00000120, -0.00000245, -0.00000248, -0.00000390, 0.00000030, 0.00000103, -0.00000711, -0.00000373, -0.00000018, -0.00000038, -0.00000851, -0.00000253, 0.00000066, -0.00000292, -0.00000333, -0.00000706, -0.00000166, 0.00000100, -0.00000659, -0.00000450, -0.00000774, -0.00000758, -0.00002098, -0.00002110, -0.00006237, -0.00001358, 0.00007974, 0.00003188, 0.00002731, 0.00002147, 0.00002307, 0.00001009, 0.00002503, 0.00001912, 0.00001402, 0.00001417, 0.00002757, 0.00002623, 0.00005715, 0.00005254, -0.00006390, -0.00004355, -0.00001396, -0.00001334, -0.00001215, -0.00001039, -0.00000057, 0.00000019, -0.00001128, -0.00000547, -0.00000298, -0.00000173, 0.00000025, -0.00000234, -0.00000150, -0.00000298, 0.00000102, -0.00000582, 0.00000112, 0.00000028, -0.00000597, 0.00000196, -0.00001451, 0.00000101, -0.00000193, -0.00000557, -0.00000581, -0.00000201, -0.00000279, -0.00000190, -0.00000216, -0.00000052, -0.00000179, -0.00000085, },
		{ -0.00000042, -0.00000131, -0.00000385, -0.00000228, -0.00000626, -0.00000466, -0.00000186, 0.00000330, -0.00000889, -0.00000957, 0.00000029, -0.00000439, 0.00000088, -0.00000376, -0.00000097, -0.00000149, -0.00000455, 0.00000234, -0.00000047, -0.00000848, -0.00000050, -0.00000723, -0.00000618, -0.00000815, -0.00000281, -0.00000843, -0.00002666, -0.00004477, 0.00001157, 0.00003941, 0.00002720, 0.00001160, 0.00001308, 0.00000797, 0.00001131, 0.00001852, 0.00000483, 0.00000693, 0.00001389, 0.00000894, 0.00001228, 0.00000987, 0.00001528, 0.00001958, 0.00003920, 0.00002505, -0.00003060, -0.00002883, -0.00001159, -0.00001149, -0.00000322, -0.00000374, -0.00000479, -0.00000361, -0.00000312, -0.00000112, -0.00000293, -0.00000267, -0.00000297, 0.00000321, -0.00000588, -0.00000487, -0.00000190, 0.00000140, -0.00000690, -0.00000506, 0.00000052, -0.00000176, -0.00000262, -0.00000564, -0.00000571, 0.00000226, -0.00000473, -0.00000455, -0.00000749, -0.00000581, -0.00001757, -0.00001842, -0.00005086, -0.00004024, 0.00006772, 0.00004580, 0.00002200, 0.00002343, 0.00002180, 0.00001583, 0.00001285, 0.00002662, 0.00001217, 0.00001500, 0.00001760, 0.00002935, 0.00003035, 0.00007310, -0.00000779, -0.00006819, -0.00001916, -0.00001634, -0.00001065, -0.00001288, -0.00000273, -0.00000048, -0.00000563, -0.00001022, -0.00000295, -0.00000270, 0.00000106, -0.00000192, -0.00000166, -0.00000433, 0.00000216, -0.00000416, -0.00000287, 0.00000333, -0.00000717, 0.00000250, -0.00001192, -0.00000424, 0.00000233, -0.00000635, -0.00000588, -0.00000389, -0.00000123, -0.00000299, -0.00000165, -0.00000135, -0.00000073, -0.00000186, },
		{ -0.00000112, -0.00000056, -0.00000311, -0.00000257, -0.00000551, -0.00000598, -0.00000229, 0.00000228, -0.00000367, -0.00001343, 0.00000084, -0.00000468, -0.00000047, -0.00000188, -0.00000160, -0.00000155, -0.00000390, -0.00000118, 0.00000394, -0.00000781, -0.00000319, -0.00000475, -0.00000576, -0.00000959, -0.00000226, -0.00000640, -0.00001522, -0.00004329, -0.00002373, 0.00003661, 0.00003472, 0.00002004, 0.00001020, 0.00001337, 0.00000647, 0.00001851, 0.00001219, 0.00000311, 0.00001320, 0.00001165, 0.00001148, 0.00001164, 0.00001250, 0.00001823, 0.00003541, 0.00003860, -0.00002058, -0.00003616, -0.00001385, -0.00001205, -0.00000621, -0.00000252, -0.00000442, -0.00000377, -0.00000439, -0.00000126, -0.00000263, -0.00000180, -0.00000443, 0.00000224, -0.00000269, -0.00000618, -0.00000282, 0.00000072, -0.00000346, -0.00000742, 0.00000009, -0.00000154, -0.00000159, -0.00000412, -0.00000717, -0.00000040, -0.00000123, -0.00000508, -0.00000626, -0.00000626, -0.00001271, -0.00001831, -0.00003680, -0.00005816, 0.00004598, 0.00005914, 0.00002081, 0.00002308, 0.00001976, 0.00002132, 0.00000827, 0.00002187, 0.00001916, 0.00001149, 0.00001545, 0.00002350, 0.00002518, 0.00005090, 0.00005135, -0.00005707, -0.00004061, -0.00001591, -0.00001327, -0.00001224, -0.00000592, -0.00000129, -0.00000338, -0.00000846, -0.00000744, -0.00000183, -0.00000002, -0.00000026, -0.00000172, -0.00000499, 0.00000040, 0.00000037, -0.00000709, 0.00000463, -0.00000637, 0.00000105, -0.00000767, -0.00000952, 0.00000433, -0.00000465, -0.00000649, -0.00000557, -0.00000087, -0.00000271, -0.00000202, -0.00000162, -0.00000044, -0.00000173, },
		{ -0.00000218, -0.00000013, -0.00000251, -0.00000223, -0.00000487, -0.00000693, -0.00000317, 0.00000111, 0.00000024, -0.00001379, -0.00000232, -0.00000183, -0.00000360, -0.00000017, -0.00000190, -0.00000082, -0.00000431, -0.00000249, 0.00000319, -0.00000226, -0.00000706, -0.00000352, -0.00000486, -0.00000884, -0.00000568, -0.00000155, -0.00001305, -0.00002672, -0.00004779, 0.00001206, 0.00004250, 0.00002811, 0.00001361, 0.00001264, 0.00001026, 0.00001094, 0.00002020, 0.00000404, 0.00000938, 0.00001383, 0.00001216, 0.00001234, 0.00001229, 0.00001532, 0.00003076, 0.00004743, -0.00000425, -0.00004232, -0.00001785, -0.00001262, -0.00000818, -0.00000373, -0.00000290, -0.00000395, -0.00000467, -0.00000277, -0.00000189, -0.00000175, -0.00000370, -0.00000082, 0.00000064, -0.00000685, -0.00000296, -0.00000147, -0.00000012, -0.00000837, -0.00000086, -0.00000175, -0.00000087, -0.00000293, -0.00000614, -0.00000444, 0.00000088, -0.00000455, -0.00000484, -0.00000755, -0.00000804, -0.00001893, -0.00002409, -0.00006437, 0.00001661, 0.00006828, 0.00002459, 0.00002124, 0.00001842, 0.00002215, 0.00001271, 0.00001136, 0.00002451, 0.00001152, 0.00001422, 0.00001698, 0.00002539, 0.00002885, 0.00006663, -0.00000351, -0.00006210, -0.00002027, -0.00001729, -0.00001049, -0.00001007, -0.00000003, -0.00000590, -0.00000298, -0.00001133, -0.00000275, -0.00000085, 0.00000097, -0.00000149, -0.00000387, -0.00000401, 0.00000508, -0.00000855, 0.00000283, -0.00000368, -0.00000153, -0.00000329, -0.00001281, 0.00000287, -0.00000096, -0.00000701, -0.00000659, -0.00000220, -0.00000127, -0.00000292, -0.00000137, -0.00000107, -0.00000066, },
		{ -0.00000184, -0.00000109, -0.00000134, -0.00000242, -0.00000326, -0.00000801, -0.00000399, -0.00000053, 0.00000293, -0.00001169, -0.00000666, 0.00000050, -0.00000480, -0.00000066, -0.00000159, 0.00000024, -0.00000451, -0.00000339, 0.00000126, 0.00000147, -0.00000598, -0.00000620, -0.00000328, -0.00000687, -0.00000929, 0.00000084, -0.00001064, -0.00001643, -0.00004383, -0.00002674, 0.00003907, 0.00003683, 0.00002161, 0.00001081, 0.00001484, 0.00000708, 0.00001931, 0.00001229, 0.00000476, 0.00001378, 0.00001374, 0.00001316, 0.00001261, 0.00001411, 0.00002386, 0.00005148, 0.00001559, -0.00004360, -0.00002502, -0.00001274, -0.00001009, -0.00000507, -0.00000301, -0.00000264, -0.00000499, -0.00000385, -0.00000233, -0.00000143, -0.00000263, -0.00000301, 0.00000157, -0.00000556, -0.00000371, -0.00000261, 0.00000050, -0.00000648, -0.00000326, -0.00000087, -0.00000185, -0.00000111, -0.00000503, -0.00000610, -0.00000077, -0.00000220, -0.00000429, -0.00000753, -0.00000607, -0.00001711, -0.00001746, -0.00005719, -0.00001646, 0.00006985, 0.00003251, 0.00001987, 0.00001717, 0.00002046, 0.00001781, 0.00000781, 0.00001871, 0.00001907, 0.00000984, 0.00001626, 0.00001955, 0.00002384, 0.00004660, 0.00004846, -0.00004900, -0.00003920, -0.00001802, -0.00001275, -0.00001124, -0.00000091, -0.00000634, -0.00000268, -0.00000796, -0.00000822, 0.00000004, 0.00000029, -0.00000011, -0.00000256, -0.00000695, 0.00000495, -0.00000434, -0.00000251, 0.00000050, -0.00000480, 0.00000069, -0.00001401, -0.00000057, 0.00000204, -0.00000564, -0.00000778, -0.00000380, -0.00000071, -0.00000260, -0.00000191, -0.00000118, -0.00000046, },
		{ -0.00000075, -0.00000224, -0.00000056, -0.00000266, -0.00000156, -0.00000810, -0.00000533, -0.00000211, 0.00000401, -0.00000786, -0.00001051, 0.00000090, -0.00000406, -0.00000184, -0.00000194, 0.00000113, -0.00000317, -0.00000528, 0.00000050, 0.00000141, -0.00000141, -0.00000849, -0.00000427, -0.00000378, -0.00001067, -0.00000134, -0.00000493, -0.00001529, -0.00002649, -0.00004970, 0.00000985, 0.00004676, 0.00002911, 0.00001513, 0.00001373, 0.00001150, 0.00001134, 0.00002059, 0.00000528, 0.00001065, 0.00001462, 0.00001522, 0.00001234, 0.00001512, 0.00001711, 0.00004877, 0.00003664, -0.00003765, -0.00003487, -0.00001307, -0.00001228, -0.00000558, -0.00000492, -0.00000102, -0.00000468, -0.00000436, -0.00000360, -0.00000134, -0.00000170, -0.00000364, 0.00000050, -0.00000341, -0.00000472, -0.00000244, -0.00000100, -0.00000345, -0.00000561, 0.00000018, -0.00000324, 0.00000009, -0.00000387, -0.00000563, -0.00000380, -0.00000077, -0.00000333, -0.00000664, -0.00000611, -0.00001344, -0.00001594, -0.00004272, -0.00004301, 0.00005781, 0.00004532, 0.00001845, 0.00001725, 0.00001744, 0.00001980, 0.00001144, 0.00000938, 0.00002292, 0.00001089, 0.00001401, 0.00001557, 0.00002187, 0.00002800, 0.00006056, 0.00000000, -0.00005540, -0.00002372, -0.00001615, -0.00001125, -0.00000410, -0.00000311, -0.00000752, -0.00000147, -0.00001234, -0.00000134, -0.00000001, 0.00000076, -0.00000123, -0.00000695, 0.00000030, 0.00000211, -0.00000691, 0.00000263, -0.00000617, 0.00000223, -0.00001216, -0.00000538, 0.00000361, -0.00000315, -0.00000810, -0.00000563, -0.00000118, -0.00000161, -0.00000256, -0.00000111, -0.00000093, },
		{ -0.00000059, -0.00000189, -0.00000132, -0.00000193, -0.00000129, -0.00000599, -0.00000767, -0.00000289, 0.00000274, -0.00000250, -0.00001362, 0.00000023, -0.00000341, -0.00000170, -0.00000327, 0.00000107, -0.00000054, -0.00000673, -0.00000080, 0.00000100, 0.00000075, -0.00000552, -0.00000891, -0.00000083, -0.00000971, -0.00000495, -0.00000097, -0.00001363, -0.00001687, -0.00004318, -0.00003145, 0.00004078, 0.00004058, 0.00002170, 0.00001332, 0.00001544, 0.00000829, 0.00001907, 0.00001363, 0.00000636, 0.00001406, 0.00001680, 0.00001371, 0.00001506, 0.00001515, 0.00003825, 0.00005531, -0.00002320, -0.00004505, -0.00001549, -0.00001375, -0.00000652, -0.00000621, -0.00000165, -0.00000265, -0.00000496, -0.00000424, -0.00000262, -0.00000047, -0.00000366, -0.00000064, -0.00000218, -0.00000501, -0.00000221, -0.00000221, -0.00000198, -0.00000600, -0.00000024, -0.00000299, -0.00000090, -0.00000178, -0.00000511, -0.00000494, -0.00000227, -0.00000138, -0.00000576, -0.00000621, -0.00001041, -0.00001545, -0.00002957, -0.00005471, 0.00003163, 0.00005919, 0.00001878, 0.00001825, 0.00001428, 0.00001950, 0.00001526, 0.00000725, 0.00001607, 0.00001870, 0.00000929, 0.00001585, 0.00001647, 0.00002221, 0.00004390, 0.00004405, -0.00003941, -0.00004092, -0.00001784, -0.00001363, -0.00000656, -0.00000060, -0.00001016, -0.00000069, -0.00000900, -0.00000708, 0.00000102, 0.00000012, 0.00000053, -0.00000549, -0.00000432, 0.00000479, -0.00000594, 0.00000041, -0.00000425, 0.00000035, -0.00000722, -0.00001086, 0.00000408, -0.00000127, -0.00000633, -0.00000799, -0.00000176, -0.00000148, -0.00000202, -0.00000184, -0.00000086, },
		},
		{ { -0.00000315, -0.00000352, -0.00000615, -0.00000305, -0.00000219, -0.00000104, 0.00000069, -0.00000423, -0.00000596, -0.00000738, -0.00000487, -0.00001021, -0.00001357, -0.00002079, -0.00005037, 0.00000203, 0.00005748, 0.00002891, 0.00001897, 0.00001554, 0.00001158, 0.00001031, 0.00001711, 0.00000898, 0.00001139, 0.00001489, 0.00001715, 0.00001339, 0.00001700, 0.00002727, 0.00006462, 0.00000013, -0.00004798, -0.00002184, -0.00001318, -0.00000740, -0.00000506, -0.00000271, -0.00001546, 0.00000107, -0.00000525, -0.00001117, 0.00000750, -0.00000106, -0.00000149, -0.00000368, -0.00000647, -0.00000021, -0.00000580, -0.00000005, -0.00000305, -0.00000512, -0.00000257, -0.00000327, -0.00000779, 0.00000030, 0.00000117, 0.00000024, -0.00001164, -0.00000270, -0.00000101, -0.00000008, -0.00000168, -0.00000276, -0.00000155, -0.00000276, -0.00000168, -0.00000008, -0.00000101, -0.00000270, -0.00001164, 0.00000024, 0.00000117, 0.00000030, -0.00000779, -0.00000327, -0.00000257, -0.00000512, -0.00000305, -0.00000005, -0.00000580, -0.00000021, -0.00000647, -0.00000368, -0.00000149, -0.00000106, 0.00000750, -0.00001117, -0.00000525, 0.00000107, -0.00001546, -0.00000271, -0.00000506, -0.00000740, -0.00001318, -0.00002184, -0.00004798, 0.00000013, 0.00006462, 0.00002727, 0.00001700, 0.00001339, 0.00001715, 0.00001489, 0.00001139, 0.00000898, 0.00001711, 0.00001031, 0.00001158, 0.00001554, 0.00001897, 0.00002891, 0.00005748, 0.00000203, -0.00005037, -0.00002079, -0.00001357, -0.00001021, -0.00000487, -0.00000738, -0.00000596, -0.00000423, 0.00000069, -0.00000104, -0.00000219, -0.00000305, -0.00000615, -0.00000352, },
		{ -0.00000302, -0.00000333, -0.00000369, -0.00000648, -0.00000205, -0.00000239, 0.00000060, -0.00000123, -0.00000508, -0.00000666, -0.00000673, -0.00000532, -0.00001222, -0.00001372, -0.00002772, -0.00005191, 0.00002422, 0.00005753, 0.00002367, 0.00001915, 0.00001485, 0.00001055, 0.00001505, 0.00001503, 0.00000833, 0.00001404, 0.00001616, 0.00001697, 0.00001583, 0.00001877, 0.00004802, 0.00005179, -0.00004111, -0.00003673, -0.00001707, -0.00000972, -0.00000771, -0.00000116, -0.00001078, -0.00000973, 0.00000347, -0.00001103, -0.00000646, 0.00000907, -0.00000380, 0.00000074, -0.00000864, -0.00000220, -0.00000133, -0.00000527, -0.00000028, -0.00000393, -0.00000489, -0.00000186, -0.00000605, -0.00000457, 0.00000094, 0.00000151, -0.00000345, -0.00001096, -0.00000044, -0.00000143, -0.00000050, -0.00000194, -0.00000210, -0.00000172, -0.00000372, -0.00000021, -0.00000051, -0.00000055, -0.00000609, -0.00000955, 0.00000246, 0.00000096, -0.00000235, -0.00000838, -0.00000105, -0.00000380, -0.00000511, -0.00000147, -0.00000095, -0.00000515, -0.00000211, -0.00000646, -0.00000078, -0.00000376, 0.00000215, 0.00000368, -0.00001275, 0.00000058, -0.00000746, -0.00001203, 0.00000016, -0.00000848, -0.00000757, -0.00001666, -0.00003294, -0.00003711, 0.00004076, 0.00005099, 0.00001696, 0.00001517, 0.00001255, 0.00001686, 0.00001349, 0.00000846, 0.00001297, 0.00001403, 0.00000869, 0.00001237, 0.00001653, 0.00001859, 0.00003563, 0.00005054, -0.00001573, -0.00004584, -0.00001540, -0.00001372, -0.00000782, -0.00000554, -0.00000717, -0.00000558, -0.00000229, 0.00000088, -0.00000221, -0.00000209, -0.00000362, -0.00000617, },
		{ -0.00000602, -0.00000259, -0.00000339, -0.00000419, -0.00000643, -0.00000114, -0.00000206, 0.00000140, -0.00000373, -0.00000515, -0.00000734, -0.00000586, -0.00000649, -0.00001416, -0.00001362, -0.00003766, -0.00004677, 0.00004732, 0.00005076, 0.00002128, 0.00001866, 0.00001405, 0.00001158, 0.00001900, 0.00001005, 0.00001198, 0.00001434, 0.00001845, 0.00001656, 0.00001866, 0.00002689, 0.00006645, 0.00000853, -0.00005635, -0.00002169, -0.00001593, -0.00000718, -0.00000580, -0.00000270, -0.00001531, -0.00000079, -0.00000057, -0.00001346, -0.00000040, 0.00000773, -0.00000438, -0.00000000, -0.00001127, 0.00000309, -0.00000485, -0.00000336, -0.00000158, -0.00000397, -0.00000458, -0.00000246, -0.00000718, -0.00000085, 0.00000011, 0.00000155, -0.00000729, -0.00000823, 0.00000042, -0.00000199, -0.00000088, -0.00000147, -0.00000195, -0.00000247, -0.00000381, 0.00000134, -0.00000139, -0.00000070, -0.00000905, -0.00000579, 0.00000327, 0.00000006, -0.00000539, -0.00000668, -0.00000027, -0.00000506, -0.00000433, -0.00000023, -0.00000188, -0.00000539, -0.00000394, -0.00000367, -0.00000103, -0.00000437, 0.00000439, -0.00000123, -0.00001003, 0.00000087, -0.00001438, -0.00000356, -0.00000373, -0.00000825, -0.00000991, -0.00002273, -0.00004057, -0.00000402, 0.00005749, 0.00003106, 0.00001464, 0.00001227, 0.00001301, 0.00001653, 0.00001035, 0.00000909, 0.00001492, 0.00000949, 0.00000959, 0.00001262, 0.00001744, 0.00001905, 0.00004137, 0.00003929, -0.00002913, -0.00003919, -0.00001173, -0.00001366, -0.00000572, -0.00000663, -0.00000648, -0.00000486, -0.00000042, -0.00000008, -0.00000280, -0.00000185, -0.00000436, },
		{ -0.00000590, -0.00000494, -0.00000285, -0.00000284, -0.00000570, -0.00000508, -0.00000106, -0.00000065, 0.00000007, -0.00000502, -0.00000561, -0.00000716, -0.00000577, -0.00000768, -0.00001620, -0.00001350, -0.00005115, -0.00003072, 0.00006488, 0.00004073, 0.00002104, 0.00001770, 0.00001328, 0.00001577, 0.00001771, 0.00000835, 0.00001542, 0.00001477, 0.00002009, 0.00001766, 0.00002120, 0.00004660, 0.00005860, -0.00004029, -0.00004311, -0.00001685, -0.00001226, -0.00000694, -0.00000229, -0.00001027, -0.00001040, 0.00000260, -0.00000552, -0.00001316, 0.00000593, 0.00000375, -0.00000273, -0.00000441, -0.00000835, 0.00000477, -0.00000757, -0.00000175, -0.00000265, -0.00000395, -0.00000425, -0.00000419, -0.00000537, 0.00000086, -0.00000053, 0.00000035, -0.00000953, -0.00000512, 0.00000033, -0.00000288, -0.00000021, -0.00000143, -0.00000196, -0.00000379, -0.00000222, 0.00000157, -0.00000184, -0.00000203, -0.00001013, -0.00000200, 0.00000309, -0.00000200, -0.00000689, -0.00000448, -0.00000056, -0.00000605, -0.00000272, 0.00000018, -0.00000271, -0.00000718, -0.00000284, -0.00000181, -0.00000249, -0.00000368, 0.00000506, -0.00000511, -0.00000550, -0.00000529, -0.00001293, 0.00000097, -0.00000865, -0.00000640, -0.00001500, -0.00003148, -0.00003140, 0.00003153, 0.00004950, 0.00001910, 0.00001337, 0.00001012, 0.00001415, 0.00001516, 0.00000720, 0.00001295, 0.00001195, 0.00000785, 0.00001021, 0.00001347, 0.00001750, 0.00002093, 0.00004471, 0.00002582, -0.00003876, -0.00003053, -0.00001058, -0.00001242, -0.00000485, -0.00000717, -0.00000584, -0.00000336, 0.00000022, -0.00000105, -0.00000320, -0.00000121, },
		{ -0.00000096, -0.00000752, -0.00000333, -0.00000343, -0.00000235, -0.00000743, -0.00000284, -0.00000140, 0.00000085, -0.00000273, -0.00000488, -0.00000657, -0.00000638, -0.00000631, -0.00000935, -0.00001694, -0.00001640, -0.00006421, -0.00000404, 0.00007233, 0.00003143, 0.00002172, 0.00001615, 0.00001457, 0.00001962, 0.00001280, 0.00001171, 0.00001583, 0.00001719, 0.00002040, 0.00002011, 0.00002885, 0.00006731, 0.00001510, -0.00006116, -0.00002418, -0.00001773, -0.00000759, -0.00000629, -0.00000282, -0.00001556, -0.00000123, -0.00000051, -0.00000825, -0.00001070, 0.00001061, -0.00000074, -0.00000142, -0.00000889, -0.00000200, 0.00000209, -0.00000797, -0.00000119, -0.00000305, -0.00000410, -0.00000454, -0.00000519, -0.00000216, 0.00000037, -0.00000060, -0.00000176, -0.00000991, -0.00000246, -0.00000074, -0.00000316, 0.00000096, -0.00000201, -0.00000234, -0.00000457, 0.00000010, 0.00000048, -0.00000189, -0.00000399, -0.00000930, 0.00000106, 0.00000162, -0.00000410, -0.00000680, -0.00000278, -0.00000152, -0.00000626, -0.00000101, 0.00000022, -0.00000472, -0.00000814, 0.00000029, -0.00000240, -0.00000338, -0.00000270, 0.00000449, -0.00000636, -0.00000409, -0.00001179, -0.00000519, -0.00000211, -0.00000911, -0.00000719, -0.00002264, -0.00003593, -0.00000431, 0.00004910, 0.00003256, 0.00001557, 0.00001039, 0.00000996, 0.00001521, 0.00001184, 0.00000737, 0.00001533, 0.00000690, 0.00000934, 0.00000948, 0.00001542, 0.00001641, 0.00002420, 0.00004546, 0.00001059, -0.00004364, -0.00002149, -0.00001139, -0.00001010, -0.00000520, -0.00000692, -0.00000508, -0.00000178, -0.00000013, -0.00000181, -0.00000308, },
		{ -0.00000188, -0.00000227, -0.00000782, -0.00000241, -0.00000332, -0.00000317, -0.00000775, -0.00000115, -0.00000104, 0.00000064, -0.00000483, -0.00000472, -0.00000695, -0.00000624, -0.00000646, -0.00001232, -0.00001501, -0.00002620, -0.00006933, 0.00002717, 0.00006989, 0.00002485, 0.00002252, 0.00001432, 0.00001929, 0.00001816, 0.00001094, 0.00001524, 0.00001606, 0.00001931, 0.00002193, 0.00002216, 0.00004895, 0.00006212, -0.00003872, -0.00004793, -0.00001831, -0.00001431, -0.00000642, -0.00000339, -0.00000992, -0.00001106, 0.00000232, -0.00000406, -0.00001019, -0.00000535, 0.00001146, -0.00000362, -0.00000256, -0.00000929, 0.00000271, -0.00000191, -0.00000725, -0.00000079, -0.00000351, -0.00000418, -0.00000576, -0.00000373, -0.00000035, -0.00000059, -0.00000092, -0.00000350, -0.00000940, -0.00000029, -0.00000291, -0.00000166, 0.00000114, -0.00000253, -0.00000339, -0.00000364, 0.00000137, -0.00000075, -0.00000218, -0.00000542, -0.00000748, 0.00000316, -0.00000126, -0.00000466, -0.00000658, -0.00000127, -0.00000312, -0.00000522, 0.00000011, -0.00000005, -0.00000831, -0.00000564, 0.00000198, -0.00000375, -0.00000389, -0.00000164, 0.00000369, -0.00000609, -0.00000771, -0.00001166, 0.00000028, -0.00000778, -0.00000601, -0.00001303, -0.00003015, -0.00002741, 0.00002665, 0.00004449, 0.00002134, 0.00001371, 0.00000760, 0.00001127, 0.00001504, 0.00000801, 0.00001119, 0.00001269, 0.00000465, 0.00001071, 0.00000899, 0.00001732, 0.00001475, 0.00002871, 0.00004317, -0.00000572, -0.00004261, -0.00001463, -0.00001236, -0.00000786, -0.00000586, -0.00000630, -0.00000384, -0.00000115, -0.00000044, -0.00000280, },
		{ -0.00000364, -0.00000026, -0.00000485, -0.00000655, -0.00000241, -0.00000265, -0.00000515, -0.00000628, -0.00000042, -0.00000017, -0.00000138, -0.00000531, -0.00000522, -0.00000657, -0.00000688, -0.00000645, -0.00001553, -0.00001229, -0.00004249, -0.00006153, 0.00005565, 0.00006027, 0.00002204, 0.00002192, 0.00001497, 0.00002363, 0.00001282, 0.00001436, 0.00001571, 0.00001834, 0.00002006, 0.00002493, 0.00002870, 0.00007271, 0.00001667, -0.00006303, -0.00002698, -0.00002012, -0.00000811, -0.00000653, -0.00000325, -0.00001594, -0.00000117, -0.00000135, -0.00000481, -0.00001183, 0.00000171, 0.00000886, -0.00000546, -0.00000424, -0.00000628, 0.00000389, -0.00000524, -0.00000602, -0.00000043, -0.00000411, -0.00000478, -0.00000629, -0.00000093, -0.00000053, -0.00000129, -0.00000139, -0.00000473, -0.00000831, 0.00000077, -0.00000490, 0.00000104, -0.00000003, -0.00000293, -0.00000433, -0.00000153, 0.00000109, -0.00000159, -0.00000266, -0.00000612, -0.00000515, 0.00000356, -0.00000413, -0.00000390, -0.00000682, 0.00000026, -0.00000501, -0.00000322, 0.00000063, -0.00000165, -0.00001125, -0.00000041, 0.00000046, -0.00000396, -0.00000513, 0.00000024, 0.00000320, -0.00000751, -0.00001217, -0.00000509, -0.00000192, -0.00000874, -0.00000564, -0.00002167, -0.00003248, -0.00000461, 0.00004407, 0.00002959, 0.00001851, 0.00000945, 0.00000743, 0.00001247, 0.00001308, 0.00000658, 0.00001485, 0.00000670, 0.00000663, 0.00000969, 0.00001056, 0.00001744, 0.00001383, 0.00003414, 0.00003654, -0.00002054, -0.00003670, -0.00001104, -0.00001233, -0.00000645, -0.00000620, -0.00000537, -0.00000262, -0.00000131, -0.00000057, },
		{ -0.00000131, -0.00000323, 0.00000006, -0.00000695, -0.00000499, -0.00000236, -0.00000277, -0.00000660, -0.00000428, 0.00000021, -0.00000034, -0.00000334, -0.00000532, -0.00000546, -0.00000666, -0.00000707, -0.00000766, -0.00001671, -0.00001327, -0.00005977, -0.00004149, 0.00007686, 0.00004637, 0.00002348, 0.00001879, 0.00002082, 0.00002128, 0.00001139, 0.00001742, 0.00001664, 0.00001947, 0.00002318, 0.00002605, 0.00004982, 0.00006973, -0.00004287, -0.00004828, -0.00002162, -0.00001607, -0.00000640, -0.00000395, -0.00001013, -0.00001183, 0.00000307, -0.00000530, -0.00000529, -0.00001096, 0.00000727, 0.00000490, -0.00000764, -0.00000319, -0.00000400, 0.00000363, -0.00000841, -0.00000340, -0.00000131, -0.00000410, -0.00000659, -0.00000412, 0.00000030, -0.00000103, -0.00000239, -0.00000090, -0.00000656, -0.00000604, -0.00000029, -0.00000469, 0.00000284, -0.00000149, -0.00000375, -0.00000375, -0.00000013, 0.00000034, -0.00000248, -0.00000252, -0.00000679, -0.00000227, 0.00000153, -0.00000483, -0.00000386, -0.00000642, 0.00000098, -0.00000621, -0.00000112, 0.00000068, -0.00000526, -0.00001043, 0.00000369, -0.00000231, -0.00000364, -0.00000668, 0.00000343, 0.00000200, -0.00001221, -0.00001115, 0.00000014, -0.00000765, -0.00000476, -0.00001211, -0.00002791, -0.00002501, 0.00002322, 0.00004117, 0.00001926, 0.00001735, 0.00000493, 0.00000936, 0.00001247, 0.00001024, 0.00000869, 0.00001422, 0.00000237, 0.00000977, 0.00000745, 0.00001375, 0.00001517, 0.00001509, 0.00003917, 0.00002493, -0.00003051, -0.00002909, -0.00000965, -0.00001167, -0.00000552, -0.00000633, -0.00000384, -0.00000248, -0.00000110, },
		{ -0.00000028, -0.00000260, -0.00000166, -0.00000140, -0.00000761, -0.00000384, -0.00000205, -0.00000395, -0.00000668, -0.00000249, 0.00000068, -0.00000175, -0.00000429, -0.00000547, -0.00000529, -0.00000743, -0.00000670, -0.00000980, -0.00001620, -0.00001983, -0.00007271, -0.00001216, 0.00008747, 0.00003274, 0.00002664, 0.00001616, 0.00002814, 0.00001320, 0.00001683, 0.00001631, 0.00002056, 0.00001815, 0.00002934, 0.00002948, 0.00007894, 0.00001994, -0.00006962, -0.00002497, -0.00002523, -0.00000784, -0.00000731, -0.00000319, -0.00001692, -0.00000115, -0.00000092, -0.00000551, -0.00000688, -0.00000739, 0.00001032, -0.00000045, -0.00000799, -0.00000068, -0.00000350, 0.00000283, -0.00001098, 0.00000006, -0.00000335, -0.00000400, -0.00000810, -0.00000026, -0.00000071, -0.00000116, -0.00000353, 0.00000021, -0.00000892, -0.00000309, -0.00000275, -0.00000212, 0.00000266, -0.00000265, -0.00000458, -0.00000185, -0.00000024, -0.00000016, -0.00000331, -0.00000186, -0.00000751, 0.00000037, -0.00000188, -0.00000311, -0.00000574, -0.00000431, 0.00000014, -0.00000597, 0.00000041, -0.00000036, -0.00000923, -0.00000566, 0.00000390, -0.00000350, -0.00000489, -0.00000649, 0.00000708, -0.00000202, -0.00001694, -0.00000381, -0.00000252, -0.00000815, -0.00000385, -0.00002145, -0.00002830, -0.00000642, 0.00004144, 0.00002648, 0.00001815, 0.00001239, 0.00000394, 0.00001086, 0.00001136, 0.00000853, 0.00001189, 0.00000943, 0.00000250, 0.00001071, 0.00000665, 0.00001639, 0.00001154, 0.00001953, 0.00004106, 0.00001048, -0.00003460, -0.00002223, -0.00000921, -0.00001083, -0.00000475, -0.00000607, -0.00000232, -0.00000324, },
		{ -0.00000326, -0.00000010, -0.00000303, -0.00000068, -0.00000314, -0.00000770, -0.00000239, -0.00000271, -0.00000469, -0.00000632, -0.00000029, -0.00000045, -0.00000259, -0.00000537, -0.00000473, -0.00000615, -0.00000740, -0.00000723, -0.00001098, -0.00001706, -0.00002966, -0.00007927, 0.00002383, 0.00008462, 0.00002502, 0.00002706, 0.00001926, 0.00002836, 0.00000906, 0.00002208, 0.00001587, 0.00002265, 0.00002017, 0.00003266, 0.00004886, 0.00008154, -0.00004954, -0.00005010, -0.00002163, -0.00002018, -0.00000581, -0.00000456, -0.00001037, -0.00001290, 0.00000380, -0.00000550, -0.00000494, -0.00000724, -0.00000372, 0.00001188, -0.00000790, -0.00000309, -0.00000151, -0.00000172, -0.00000042, -0.00001041, 0.00000168, -0.00000474, -0.00000525, -0.00000666, 0.00000182, -0.00000180, -0.00000214, -0.00000286, -0.00000020, -0.00001003, -0.00000134, -0.00000421, 0.00000066, 0.00000143, -0.00000402, -0.00000396, -0.00000060, -0.00000069, -0.00000100, -0.00000299, -0.00000201, -0.00000731, 0.00000118, -0.00000398, -0.00000141, -0.00000813, -0.00000123, -0.00000173, -0.00000461, 0.00000124, -0.00000326, -0.00001031, -0.00000063, 0.00000181, -0.00000341, -0.00000776, -0.00000323, 0.00000921, -0.00000983, -0.00001562, 0.00000231, -0.00000935, -0.00000227, -0.00001181, -0.00002592, -0.00002212, 0.00001855, 0.00004078, 0.00001524, 0.00001942, 0.00000584, 0.00000622, 0.00001069, 0.00000981, 0.00000930, 0.00001230, 0.00000475, 0.00000503, 0.00000955, 0.00000797, 0.00001690, 0.00000843, 0.00002675, 0.00003721, -0.00000268, -0.00003484, -0.00001624, -0.00000980, -0.00000926, -0.00000466, -0.00000470, -0.00000228, },
		{ -0.00000391, -0.00000186, -0.00000106, -0.00000214, -0.00000113, -0.00000422, -0.00000763, -0.00000072, -0.00000448, -0.00000434, -0.00000585, 0.00000179, -0.00000301, -0.00000226, -0.00000686, -0.00000322, -0.00000811, -0.00000649, -0.00000834, -0.00001183, -0.00001979, -0.00004153, -0.00007672, 0.00006090, 0.00007022, 0.00002458, 0.00002485, 0.00002678, 0.00002001, 0.00001415, 0.00002143, 0.00002011, 0.00002005, 0.00002812, 0.00003496, 0.00008169, 0.00002978, -0.00008292, -0.00002007, -0.00003003, -0.00000835, -0.00000848, -0.00000224, -0.00001888, -0.00000082, -0.00000049, -0.00000601, -0.00000531, -0.00000609, -0.00000029, 0.00001038, -0.00001363, 0.00000467, -0.00000621, 0.00000218, -0.00000613, -0.00000626, 0.00000039, -0.00000508, -0.00000703, -0.00000264, 0.00000090, -0.00000186, -0.00000395, -0.00000022, -0.00000299, -0.00000911, -0.00000135, -0.00000358, 0.00000205, -0.00000003, -0.00000531, -0.00000182, -0.00000108, -0.00000055, -0.00000232, -0.00000127, -0.00000368, -0.00000589, -0.00000006, -0.00000336, -0.00000202, -0.00000882, 0.00000117, -0.00000351, -0.00000279, 0.00000085, -0.00000692, -0.00000748, 0.00000123, 0.00000075, -0.00000479, -0.00000955, 0.00000213, 0.00000746, -0.00001836, -0.00000649, -0.00000078, -0.00000980, -0.00000003, -0.00002286, -0.00002343, -0.00000836, 0.00003815, 0.00002629, 0.00001439, 0.00001587, 0.00000293, 0.00000835, 0.00000952, 0.00000905, 0.00001113, 0.00000921, 0.00000334, 0.00000670, 0.00000846, 0.00001012, 0.00001500, 0.00000810, 0.00003391, 0.00002817, -0.00001214, -0.00003330, -0.00001087, -0.00001143, -0.00000669, -0.00000525, -0.00000253, },
		{ -0.00000154, -0.00000516, -0.00000055, -0.00000167, -0.00000161, -0.00000158, -0.00000565, -0.00000616, -0.00000072, -0.00000530, -0.00000439, -0.00000390, 0.00000138, -0.00000421, -0.00000289, -0.00000667, -0.00000344, -0.00000873, -0.00000692, -0.00000753, -0.00001602, -0.00002082, -0.00005782, -0.00005954, 0.00008922, 0.00005228, 0.00002709, 0.00002516, 0.00002992, 0.00001290, 0.00002153, 0.00001977, 0.00002355, 0.00001980, 0.00003453, 0.00005233, 0.00009071, -0.00005029, -0.00006024, -0.00001564, -0.00002672, -0.00000441, -0.00000642, -0.00000902, -0.00001571, 0.00000538, -0.00000551, -0.00000586, -0.00000388, -0.00000655, 0.00000503, 0.00000294, -0.00001138, 0.00000775, -0.00000940, 0.00000443, -0.00001048, -0.00000178, -0.00000190, -0.00000585, -0.00000649, 0.00000025, -0.00000065, -0.00000239, -0.00000431, 0.00000179, -0.00000632, -0.00000755, -0.00000144, -0.00000231, 0.00000256, -0.00000216, -0.00000491, -0.00000031, -0.00000197, -0.00000057, -0.00000276, 0.00000018, -0.00000574, -0.00000455, -0.00000130, -0.00000173, -0.00000443, -0.00000747, 0.00000237, -0.00000488, -0.00000076, -0.00000173, -0.00000817, -0.00000395, 0.00000100, 0.00000079, -0.00000826, -0.00000819, 0.00000739, 0.00000020, -0.00002071, 0.00000273, -0.00000989, -0.00000138, -0.00000950, -0.00002615, -0.00001753, 0.00001275, 0.00004080, 0.00001334, 0.00001740, 0.00000925, 0.00000400, 0.00000876, 0.00000805, 0.00001037, 0.00001052, 0.00000631, 0.00000358, 0.00000749, 0.00000791, 0.00001221, 0.00001109, 0.00001233, 0.00003673, 0.00001792, -0.00001912, -0.00002963, -0.00000751, -0.00001241, -0.00000446, -0.00000513, },
		{ -0.00000361, -0.00000281, -0.00000451, -0.00000050, -0.00000107, -0.00000245, -0.00000109, -0.00000792, -0.00000305, -0.00000304, -0.00000405, -0.00000551, -0.00000055, -0.00000199, -0.00000276, -0.00000549, -0.00000414, -0.00000625, -0.00000717, -0.00000883, -0.00000542, -0.00002353, -0.00001957, -0.00007796, -0.00002593, 0.00010181, 0.00003774, 0.00002997, 0.00002866, 0.00002512, 0.00001431, 0.00002389, 0.00002243, 0.00002175, 0.00002664, 0.00003857, 0.00008671, 0.00003989, -0.00009405, -0.00002143, -0.00002864, -0.00001222, -0.00000842, -0.00000234, -0.00001958, -0.00000211, 0.00000074, -0.00000578, -0.00000695, -0.00000071, -0.00000832, 0.00001033, -0.00000730, -0.00000169, 0.00000315, -0.00000768, 0.00000261, -0.00001097, 0.00000051, -0.00000398, -0.00000674, -0.00000352, 0.00000026, -0.00000118, -0.00000397, -0.00000227, 0.00000137, -0.00000863, -0.00000616, -0.00000073, -0.00000165, 0.00000267, -0.00000462, -0.00000254, -0.00000085, -0.00000197, -0.00000135, -0.00000152, -0.00000008, -0.00000712, -0.00000400, -0.00000098, -0.00000139, -0.00000662, -0.00000511, 0.00000261, -0.00000581, 0.00000082, -0.00000572, -0.00000555, -0.00000337, 0.00000221, -0.00000119, -0.00001102, -0.00000425, 0.00001008, -0.00001058, -0.00001308, 0.00000180, -0.00001227, 0.00000389, -0.00002351, -0.00002022, -0.00000832, 0.00003300, 0.00002769, 0.00001138, 0.00001550, 0.00000580, 0.00000521, 0.00000839, 0.00000708, 0.00001279, 0.00000652, 0.00000662, 0.00000253, 0.00000927, 0.00000684, 0.00001394, 0.00000676, 0.00002018, 0.00003346, 0.00000976, -0.00002541, -0.00002318, -0.00000747, -0.00001116, -0.00000372, },
		{ -0.00000354, -0.00000232, -0.00000464, -0.00000315, -0.00000046, -0.00000090, -0.00000294, -0.00000131, -0.00000899, -0.00000082, -0.00000519, -0.00000272, -0.00000567, 0.00000112, -0.00000487, -0.00000139, -0.00000738, -0.00000230, -0.00000868, -0.00000628, -0.00000875, -0.00000732, -0.00002800, -0.00002266, -0.00009359, 0.00001721, 0.00009972, 0.00002789, 0.00003548, 0.00002819, 0.00002078, 0.00001865, 0.00002526, 0.00002406, 0.00002206, 0.00003257, 0.00005771, 0.00009936, -0.00004757, -0.00007311, -0.00001146, -0.00002950, -0.00000531, -0.00000797, -0.00000742, -0.00001868, 0.00000642, -0.00000536, -0.00000548, -0.00000532, -0.00000072, -0.00000681, 0.00001003, -0.00001201, 0.00000699, -0.00000397, -0.00000383, -0.00000087, -0.00000936, 0.00000135, -0.00000673, -0.00000511, -0.00000170, -0.00000034, -0.00000215, -0.00000441, -0.00000008, -0.00000035, -0.00001051, -0.00000368, -0.00000073, -0.00000068, 0.00000137, -0.00000531, -0.00000057, -0.00000205, -0.00000182, -0.00000162, 0.00000001, -0.00000144, -0.00000838, -0.00000285, -0.00000035, -0.00000217, -0.00000807, -0.00000198, 0.00000142, -0.00000549, 0.00000017, -0.00000768, -0.00000229, -0.00000455, 0.00000488, -0.00000611, -0.00001074, 0.00000067, 0.00000747, -0.00001764, -0.00000226, -0.00000781, -0.00000305, -0.00000515, -0.00002760, -0.00001288, 0.00000798, 0.00003914, 0.00001372, 0.00001466, 0.00000982, 0.00000626, 0.00000502, 0.00000758, 0.00000833, 0.00001245, 0.00000328, 0.00000783, 0.00000118, 0.00001136, 0.00000568, 0.00001425, 0.00000499, 0.00002742, 0.00002691, 0.00000297, -0.00002985, -0.00001593, -0.00000930, -0.00000851, },
		{ -0.00000613, -0.00000280, -0.00000274, -0.00000514, -0.00000244, 0.00000059, -0.00000246, -0.00000162, -0.00000358, -0.00000744, -0.00000133, -0.00000500, -0.00000300, -0.00000378, -0.00000066, -0.00000464, -0.00000248, -0.00000650, -0.00000343, -0.00000856, -0.00000750, -0.00000573, -0.00001554, -0.00002495, -0.00003630, -0.00009533, 0.00006023, 0.00008682, 0.00002431, 0.00004175, 0.00002148, 0.00002343, 0.00001935, 0.00003041, 0.00002012, 0.00003002, 0.00003547, 0.00009716, 0.00004700, -0.00009960, -0.00002942, -0.00002445, -0.00001564, -0.00000878, -0.00000349, -0.00001851, -0.00000527, 0.00000285, -0.00000605, -0.00000735, -0.00000016, -0.00000500, -0.00000164, 0.00000411, -0.00000841, 0.00000895, -0.00000851, -0.00000096, -0.00000368, -0.00000737, 0.00000100, -0.00000915, -0.00000114, -0.00000275, 0.00000014, -0.00000426, -0.00000257, 0.00000042, -0.00000226, -0.00001199, 0.00000011, -0.00000238, 0.00000117, -0.00000140, -0.00000342, -0.00000087, -0.00000227, -0.00000237, -0.00000034, 0.00000024, -0.00000298, -0.00000976, -0.00000006, -0.00000126, -0.00000261, -0.00000928, 0.00000200, -0.00000161, -0.00000364, -0.00000314, -0.00000513, -0.00000281, -0.00000306, 0.00000495, -0.00001068, -0.00000817, 0.00000503, -0.00000060, -0.00001474, -0.00000035, -0.00001200, 0.00000486, -0.00002121, -0.00001964, -0.00000645, 0.00002731, 0.00002835, 0.00001069, 0.00001306, 0.00000750, 0.00000613, 0.00000505, 0.00000628, 0.00001180, 0.00000739, 0.00000490, 0.00000571, 0.00000291, 0.00001099, 0.00000609, 0.00001231, 0.00000750, 0.00003039, 0.00002082, -0.00000467, -0.00003027, -0.00001077, -0.00001049, },
		{ -0.00001062, -0.00000389, -0.00000258, -0.00000366, -0.00000510, -0.00000134, 0.00000092, -0.00000387, -0.00000021, -0.00000610, -0.00000525, -0.00000277, -0.00000384, -0.00000337, -0.00000235, -0.00000315, -0.00000325, -0.00000420, -0.00000505, -0.00000537, -0.00000800, -0.00000786, -0.00000471, -0.00002442, -0.00001878, -0.00005836, -0.00008037, 0.00009715, 0.00006522, 0.00003158, 0.00003923, 0.00001872, 0.00002601, 0.00002206, 0.00003201, 0.00001999, 0.00003557, 0.00005575, 0.00011530, -0.00004925, -0.00007992, -0.00001570, -0.00002627, -0.00000850, -0.00000843, -0.00000770, -0.00001911, 0.00000456, -0.00000337, -0.00000578, -0.00000665, 0.00000254, -0.00000818, 0.00000169, -0.00000038, -0.00000273, 0.00000687, -0.00001114, 0.00000198, -0.00000709, -0.00000395, -0.00000233, -0.00000755, 0.00000069, -0.00000371, -0.00000024, -0.00000500, -0.00000085, 0.00000052, -0.00000549, -0.00001077, 0.00000246, -0.00000371, 0.00000216, -0.00000321, -0.00000141, -0.00000202, -0.00000231, -0.00000233, 0.00000106, -0.00000019, -0.00000551, -0.00000927, 0.00000233, -0.00000276, -0.00000317, -0.00000924, 0.00000519, -0.00000497, -0.00000215, -0.00000592, -0.00000091, -0.00000628, 0.00000171, 0.00000059, -0.00001207, -0.00000468, 0.00000602, -0.00000804, -0.00000655, -0.00000820, -0.00000337, -0.00000265, -0.00002670, -0.00001082, 0.00000574, 0.00003581, 0.00001441, 0.00001390, 0.00000712, 0.00000958, 0.00000377, 0.00000557, 0.00000634, 0.00001350, 0.00000178, 0.00000909, 0.00000100, 0.00000766, 0.00000773, 0.00000808, 0.00000949, 0.00001243, 0.00002965, 0.00001542, -0.00001249, -0.00002727, -0.00000803, },
		{ -0.00000642, -0.00001029, -0.00000143, -0.00000388, -0.00000356, -0.00000557, 0.00000111, -0.00000101, -0.00000313, -0.00000071, -0.00000692, -0.00000450, -0.00000296, -0.00000345, -0.00000263, -0.00000328, -0.00000362, -0.00000312, -0.00000451, -0.00000519, -0.00000604, -0.00000850, -0.00000597, -0.00000903, -0.00002801, -0.00001657, -0.00008277, -0.00004890, 0.00012161, 0.00004186, 0.00004661, 0.00002540, 0.00002707, 0.00002135, 0.00003236, 0.00002410, 0.00003114, 0.00003427, 0.00010319, 0.00006006, -0.00010874, -0.00003301, -0.00002789, -0.00001213, -0.00001270, -0.00000311, -0.00001902, -0.00000636, 0.00000224, -0.00000449, -0.00000788, -0.00000158, 0.00000013, -0.00000806, 0.00000180, -0.00000035, 0.00000024, 0.00000398, -0.00001269, 0.00000521, -0.00001130, 0.00000040, -0.00000757, -0.00000167, -0.00000195, -0.00000229, -0.00000269, -0.00000286, -0.00000119, 0.00000123, -0.00001018, -0.00000590, 0.00000134, -0.00000299, 0.00000115, -0.00000242, -0.00000165, -0.00000202, -0.00000335, -0.00000036, 0.00000082, -0.00000029, -0.00000944, -0.00000535, 0.00000170, -0.00000252, -0.00000546, -0.00000628, 0.00000534, -0.00000667, -0.00000280, -0.00000473, -0.00000030, -0.00000743, 0.00000514, -0.00000493, -0.00001103, -0.00000087, 0.00000190, -0.00000748, -0.00000552, -0.00001155, 0.00000517, -0.00001889, -0.00001829, -0.00000591, 0.00002358, 0.00002714, 0.00001099, 0.00001208, 0.00000510, 0.00001047, 0.00000189, 0.00000543, 0.00000905, 0.00001003, 0.00000124, 0.00001014, -0.00000152, 0.00001155, 0.00000399, 0.00001031, 0.00000758, 0.00001706, 0.00002746, 0.00000916, -0.00001825, -0.00002314, },
		{ -0.00001817, -0.00000626, -0.00000847, -0.00000057, -0.00000486, -0.00000381, -0.00000487, 0.00000264, -0.00000313, -0.00000174, -0.00000180, -0.00000742, -0.00000374, -0.00000310, -0.00000244, -0.00000311, -0.00000404, -0.00000377, -0.00000266, -0.00000543, -0.00000486, -0.00000750, -0.00000716, -0.00000628, -0.00001441, -0.00002790, -0.00002010, -0.00010579, -0.00000019, 0.00012448, 0.00003058, 0.00005396, 0.00001447, 0.00003633, 0.00001796, 0.00004005, 0.00001883, 0.00004067, 0.00005030, 0.00013291, -0.00004993, -0.00008654, -0.00001976, -0.00002705, -0.00000650, -0.00001175, -0.00000736, -0.00001960, 0.00000296, -0.00000260, -0.00000525, -0.00000704, 0.00000150, -0.00000235, -0.00000907, 0.00000449, -0.00000128, 0.00000338, -0.00000171, -0.00000987, 0.00000472, -0.00001251, 0.00000160, -0.00000937, 0.00000286, -0.00000540, -0.00000091, -0.00000432, -0.00000069, -0.00000166, 0.00000042, -0.00001307, -0.00000063, -0.00000134, -0.00000161, 0.00000017, -0.00000122, -0.00000281, -0.00000174, -0.00000394, 0.00000196, -0.00000017, -0.00000137, -0.00001222, -0.00000032, -0.00000094, -0.00000101, -0.00000868, -0.00000127, 0.00000269, -0.00000716, -0.00000375, -0.00000127, -0.00000356, -0.00000436, 0.00000424, -0.00000828, -0.00000886, 0.00000090, -0.00000264, -0.00000200, -0.00001380, -0.00000148, -0.00000254, -0.00002401, -0.00000964, 0.00000378, 0.00003319, 0.00001369, 0.00001452, 0.00000503, 0.00000880, 0.00000769, 0.00000152, 0.00000582, 0.00001123, 0.00000438, 0.00000502, 0.00000693, 0.00000009, 0.00001221, 0.00000153, 0.00001253, 0.00000604, 0.00002157, 0.00002342, 0.00000300, -0.00002216, },
		{ -0.00002562, -0.00001177, -0.00000814, -0.00000446, -0.00000274, -0.00000365, -0.00000582, -0.00000151, 0.00000107, -0.00000297, -0.00000183, -0.00000186, -0.00000912, -0.00000137, -0.00000449, 0.00000003, -0.00000631, -0.00000235, -0.00000547, -0.00000053, -0.00000842, -0.00000263, -0.00001062, -0.00000318, -0.00001132, -0.00001664, -0.00002728, -0.00003049, -0.00012064, 0.00005948, 0.00010542, 0.00003682, 0.00004420, 0.00001927, 0.00003448, 0.00002592, 0.00003466, 0.00002827, 0.00003913, 0.00010106, 0.00008093, -0.00012309, -0.00003100, -0.00003693, -0.00000702, -0.00001528, -0.00000325, -0.00002089, -0.00000528, -0.00000066, -0.00000182, -0.00000917, -0.00000110, -0.00000045, -0.00000290, -0.00001147, 0.00001107, -0.00000618, 0.00000869, -0.00001071, -0.00000168, -0.00000138, -0.00000902, -0.00000163, -0.00000531, 0.00000193, -0.00000595, -0.00000182, -0.00000282, -0.00000115, -0.00000032, -0.00000338, -0.00001158, 0.00000162, -0.00000292, -0.00000118, 0.00000104, -0.00000214, -0.00000258, -0.00000280, -0.00000223, 0.00000229, -0.00000024, -0.00000483, -0.00001098, 0.00000197, -0.00000222, -0.00000098, -0.00001009, 0.00000294, -0.00000077, -0.00000792, -0.00000188, -0.00000057, -0.00000565, -0.00000115, 0.00000150, -0.00000995, -0.00000560, -0.00000190, 0.00000006, -0.00000420, -0.00001656, 0.00000854, -0.00001987, -0.00001390, -0.00000751, 0.00002148, 0.00002533, 0.00001062, 0.00001275, 0.00000196, 0.00001192, 0.00000425, 0.00000133, 0.00000841, 0.00000903, 0.00000236, 0.00000767, 0.00000339, 0.00000340, 0.00001027, 0.00000093, 0.00001450, 0.00000427, 0.00002673, 0.00001631, -0.00000105, },
		{ -0.00000440, -0.00002716, -0.00000608, -0.00000975, -0.00000099, -0.00000540, -0.00000227, -0.00000729, 0.00000186, -0.00000142, -0.00000210, -0.00000181, -0.00000297, -0.00000949, 0.00000020, -0.00000454, 0.00000038, -0.00000840, -0.00000126, -0.00000570, -0.00000040, -0.00000962, -0.00000243, -0.00001065, -0.00000283, -0.00001503, -0.00001960, -0.00002344, -0.00005364, -0.00011320, 0.00011181, 0.00007970, 0.00004731, 0.00002921, 0.00003167, 0.00002848, 0.00003603, 0.00002767, 0.00003961, 0.00005172, 0.00014155, -0.00004058, -0.00010171, -0.00001477, -0.00003784, 0.00000219, -0.00001771, -0.00000513, -0.00002314, 0.00000524, -0.00000610, -0.00000167, -0.00000934, 0.00000230, -0.00000144, -0.00000712, -0.00000740, 0.00001303, -0.00000866, 0.00001066, -0.00001674, 0.00000584, -0.00000835, -0.00000495, -0.00000437, -0.00000030, -0.00000163, -0.00000530, -0.00000260, -0.00000080, -0.00000236, 0.00000078, -0.00000742, -0.00000788, 0.00000159, -0.00000395, -0.00000016, 0.00000175, -0.00000388, -0.00000175, -0.00000394, 0.00000046, 0.00000130, -0.00000044, -0.00000868, -0.00000703, 0.00000137, -0.00000147, -0.00000300, -0.00000892, 0.00000524, -0.00000475, -0.00000737, 0.00000124, -0.00000316, -0.00000439, -0.00000077, -0.00000029, -0.00001067, -0.00000349, -0.00000517, 0.00000742, -0.00001687, -0.00000339, -0.00000006, -0.00002432, -0.00000509, -0.00000092, 0.00003297, 0.00001138, 0.00001525, 0.00000473, 0.00000564, 0.00001109, 0.00000168, 0.00000233, 0.00001033, 0.00000480, 0.00000398, 0.00000752, 0.00000159, 0.00000659, 0.00000623, 0.00000371, 0.00001344, 0.00000521, 0.00002950, 0.00000871, },
		{ 0.00000487, -0.00001017, -0.00002404, -0.00000404, -0.00000835, -0.00000116, -0.00000525, -0.00000351, -0.00000536, 0.00000191, -0.00000162, -0.00000302, 0.00000027, -0.00000721, -0.00000596, -0.00000148, -0.00000129, -0.00000406, -0.00000586, -0.00000387, -0.00000182, -0.00000516, -0.00000594, -0.00000684, -0.00000505, -0.00001020, -0.00001199, -0.00002740, -0.00001539, -0.00009053, -0.00007476, 0.00014126, 0.00006175, 0.00004812, 0.00002496, 0.00003708, 0.00003070, 0.00003528, 0.00003415, 0.00003948, 0.00010288, 0.00009649, -0.00013057, -0.00003556, -0.00003827, -0.00001208, -0.00000933, -0.00000783, -0.00002021, -0.00000710, 0.00000005, -0.00000349, -0.00000714, -0.00000218, -0.00000107, 0.00000120, -0.00001576, 0.00000564, 0.00000502, -0.00000339, 0.00000487, -0.00001475, 0.00000718, -0.00001172, -0.00000342, -0.00000309, 0.00000084, -0.00000378, -0.00000582, -0.00000075, -0.00000132, -0.00000175, -0.00000059, -0.00000872, -0.00000520, 0.00000130, -0.00000541, 0.00000295, -0.00000031, -0.00000379, -0.00000234, -0.00000295, 0.00000153, 0.00000093, -0.00000263, -0.00000976, -0.00000408, 0.00000083, -0.00000096, -0.00000511, -0.00000673, 0.00000620, -0.00000980, -0.00000274, 0.00000061, -0.00000403, -0.00000369, -0.00000068, -0.00000227, -0.00000935, -0.00000612, -0.00000030, 0.00000549, -0.00002350, 0.00001115, -0.00002073, -0.00001066, -0.00000678, 0.00001708, 0.00002584, 0.00000830, 0.00001410, 0.00000056, 0.00000947, 0.00000909, -0.00000123, 0.00000645, 0.00000746, 0.00000420, 0.00000436, 0.00000765, -0.00000006, 0.00000974, 0.00000064, 0.00001031, 0.00000737, 0.00001196, 0.00002582, },
		{ 0.00001723, 0.00000437, -0.00001748, -0.00001743, -0.00000472, -0.00000543, -0.00000343, -0.00000353, -0.00000571, -0.00000198, 0.00000020, -0.00000086, -0.00000391, 0.00000171, -0.00001139, -0.00000136, -0.00000354, 0.00000135, -0.00000908, -0.00000209, -0.00000668, 0.00000188, -0.00001036, -0.00000186, -0.00001052, -0.00000085, -0.00001812, -0.00000797, -0.00003398, -0.00001330, -0.00012705, -0.00001436, 0.00014933, 0.00004880, 0.00004478, 0.00002777, 0.00003918, 0.00003419, 0.00003374, 0.00004200, 0.00005326, 0.00014901, -0.00003061, -0.00011642, -0.00001026, -0.00004673, 0.00000511, -0.00001730, -0.00000614, -0.00002582, 0.00000694, -0.00000857, -0.00000025, -0.00001000, 0.00000339, -0.00000382, -0.00000002, -0.00001857, 0.00001703, -0.00000559, 0.00000449, -0.00000395, -0.00000837, 0.00000431, -0.00001312, -0.00000138, -0.00000139, 0.00000008, -0.00000575, -0.00000518, 0.00000106, -0.00000257, -0.00000100, -0.00000237, -0.00000879, -0.00000290, -0.00000026, -0.00000500, 0.00000543, -0.00000345, -0.00000271, -0.00000324, -0.00000094, 0.00000143, 0.00000040, -0.00000531, -0.00000874, -0.00000262, 0.00000108, -0.00000129, -0.00000682, -0.00000365, 0.00000461, -0.00001293, 0.00000353, -0.00000385, -0.00000127, -0.00000620, 0.00000158, -0.00000526, -0.00000766, -0.00001013, 0.00001109, -0.00001033, -0.00001061, 0.00000476, -0.00002735, 0.00000124, -0.00000564, 0.00003211, 0.00001027, 0.00001449, 0.00000531, 0.00000389, 0.00000967, 0.00000713, -0.00000279, 0.00001043, 0.00000230, 0.00000685, 0.00000272, 0.00000868, -0.00000150, 0.00001132, -0.00000301, 0.00001634, 0.00000024, 0.00002149, },
		{ 0.00002701, 0.00000949, 0.00000296, -0.00002231, -0.00001116, -0.00000504, -0.00000432, -0.00000420, -0.00000347, -0.00000555, -0.00000085, 0.00000025, -0.00000218, -0.00000182, -0.00000108, -0.00001116, 0.00000018, -0.00000245, -0.00000096, -0.00000933, -0.00000209, -0.00000507, 0.00000023, -0.00001047, -0.00000246, -0.00000831, -0.00000509, -0.00001767, -0.00001155, -0.00003156, -0.00002839, -0.00014529, 0.00005226, 0.00014269, 0.00003488, 0.00004809, 0.00002668, 0.00004774, 0.00002733, 0.00004639, 0.00003625, 0.00011055, 0.00010546, -0.00013193, -0.00004576, -0.00003351, -0.00002244, -0.00000250, -0.00001058, -0.00001979, -0.00001094, 0.00000352, -0.00000777, -0.00000390, -0.00000424, 0.00000081, -0.00000099, -0.00000875, -0.00000920, 0.00001712, -0.00000969, 0.00000790, -0.00000948, -0.00000299, 0.00000058, -0.00001412, 0.00000330, -0.00000302, 0.00000060, -0.00000935, -0.00000103, -0.00000037, -0.00000182, -0.00000226, -0.00000209, -0.00001006, 0.00000052, -0.00000415, -0.00000065, 0.00000387, -0.00000433, -0.00000304, -0.00000209, -0.00000054, 0.00000214, -0.00000215, -0.00000559, -0.00000875, -0.00000021, 0.00000019, -0.00000095, -0.00000934, 0.00000130, -0.00000138, -0.00000966, 0.00000483, -0.00000587, 0.00000008, -0.00000817, 0.00000364, -0.00000704, -0.00001021, -0.00000570, 0.00001359, -0.00002253, 0.00000752, -0.00001777, -0.00001191, -0.00000174, 0.00001007, 0.00002787, 0.00000546, 0.00001498, 0.00000002, 0.00000761, 0.00000946, 0.00000303, 0.00000042, 0.00000865, 0.00000170, 0.00000682, 0.00000348, 0.00000720, -0.00000020, 0.00000882, -0.00000087, 0.00001617, -0.00000130, },
		{ 0.00000372, 0.00002614, 0.00000511, -0.00000078, -0.00002347, -0.00000620, -0.00000524, -0.00000439, -0.00000391, -0.00000420, -0.00000419, -0.00000099, 0.00000103, -0.00000396, 0.00000100, -0.00000588, -0.00000796, 0.00000009, -0.00000086, -0.00000531, -0.00000676, -0.00000384, -0.00000175, -0.00000411, -0.00000796, -0.00000442, -0.00000516, -0.00001287, -0.00001194, -0.00002028, -0.00002416, -0.00005828, -0.00013948, 0.00011823, 0.00011722, 0.00003053, 0.00005001, 0.00002924, 0.00005193, 0.00002399, 0.00005746, 0.00004593, 0.00016657, -0.00003114, -0.00012162, -0.00001481, -0.00004551, -0.00000258, -0.00001013, -0.00000976, -0.00002603, 0.00000446, -0.00000582, -0.00000433, -0.00000665, 0.00000152, -0.00000277, 0.00000011, -0.00001531, 0.00000360, 0.00001007, -0.00000869, 0.00000817, -0.00001284, 0.00000211, -0.00000575, -0.00001059, 0.00000558, -0.00000462, -0.00000019, -0.00001097, 0.00000302, -0.00000266, -0.00000109, -0.00000306, -0.00000215, -0.00001053, 0.00000265, -0.00000706, 0.00000447, -0.00000007, -0.00000360, -0.00000374, -0.00000011, -0.00000100, 0.00000253, -0.00000496, -0.00000460, -0.00000946, 0.00000296, -0.00000160, -0.00000054, -0.00001142, 0.00000591, -0.00000878, -0.00000156, -0.00000013, -0.00000235, -0.00000382, -0.00000580, 0.00000323, -0.00000798, -0.00001493, 0.00000749, -0.00000038, -0.00001398, 0.00000509, -0.00002751, 0.00000363, -0.00000549, 0.00002732, 0.00001211, 0.00001189, 0.00000627, 0.00000318, 0.00000726, 0.00001015, -0.00000175, 0.00000591, 0.00000267, 0.00000604, 0.00000275, 0.00000786, 0.00000239, 0.00000364, 0.00000380, 0.00000536, 0.00001036, },
		{ 0.00000449, 0.00001067, 0.00002183, 0.00000237, -0.00000498, -0.00002254, -0.00000152, -0.00000726, -0.00000298, -0.00000511, -0.00000300, -0.00000477, 0.00000080, -0.00000057, -0.00000287, 0.00000049, -0.00000814, -0.00000562, 0.00000146, -0.00000285, -0.00000653, -0.00000611, -0.00000297, -0.00000167, -0.00000600, -0.00000732, -0.00000315, -0.00000776, -0.00001544, -0.00000928, -0.00002683, -0.00002070, -0.00009410, -0.00010877, 0.00017539, 0.00007277, 0.00004669, 0.00003804, 0.00004735, 0.00003700, 0.00004340, 0.00004632, 0.00010958, 0.00012405, -0.00014394, -0.00004654, -0.00003734, -0.00002289, -0.00000760, -0.00000372, -0.00002437, -0.00001140, 0.00000273, -0.00000673, -0.00000679, -0.00000146, -0.00000039, -0.00000109, -0.00000577, -0.00001176, 0.00000955, 0.00000393, -0.00000716, 0.00000841, -0.00001625, 0.00000780, -0.00001491, -0.00000051, 0.00000105, -0.00000204, -0.00000529, -0.00000691, 0.00000254, -0.00000223, -0.00000309, -0.00000048, -0.00000574, -0.00000731, 0.00000067, -0.00000516, 0.00000557, -0.00000221, -0.00000414, -0.00000206, -0.00000039, 0.00000024, 0.00000010, -0.00000455, -0.00000614, -0.00000734, 0.00000369, -0.00000145, -0.00000293, -0.00000976, 0.00000583, -0.00001111, 0.00000361, -0.00000453, 0.00000140, -0.00000915, -0.00000122, 0.00000221, -0.00001272, -0.00001147, 0.00001371, -0.00001403, 0.00000304, -0.00001606, -0.00001297, 0.00000200, 0.00000619, 0.00002665, 0.00000549, 0.00001379, 0.00000004, 0.00000732, 0.00000665, 0.00000837, -0.00000233, 0.00000692, 0.00000012, 0.00000852, 0.00000016, 0.00001122, -0.00000212, 0.00000665, 0.00000058, 0.00001020, },
		{ 0.00001214, 0.00000099, 0.00001696, 0.00001558, 0.00000112, -0.00000969, -0.00001879, 0.00000077, -0.00000900, -0.00000162, -0.00000609, -0.00000175, -0.00000558, 0.00000289, -0.00000273, -0.00000089, -0.00000173, -0.00000877, -0.00000288, 0.00000152, -0.00000548, -0.00000646, -0.00000551, -0.00000203, -0.00000246, -0.00000763, -0.00000534, -0.00000351, -0.00001162, -0.00001490, -0.00000885, -0.00003282, -0.00002103, -0.00013398, -0.00004505, 0.00020040, 0.00003366, 0.00006752, 0.00002520, 0.00006549, 0.00002079, 0.00006653, 0.00004490, 0.00018240, -0.00002907, -0.00013196, -0.00001465, -0.00004803, -0.00000525, -0.00001140, -0.00000511, -0.00003053, 0.00000371, -0.00000478, -0.00000582, -0.00000721, 0.00000282, -0.00000278, -0.00000125, -0.00000980, -0.00000529, 0.00001138, -0.00000254, -0.00000227, 0.00000455, -0.00001514, 0.00000831, -0.00001926, 0.00000856, -0.00000538, 0.00000118, -0.00001057, -0.00000083, -0.00000024, -0.00000113, -0.00000494, 0.00000244, -0.00001037, -0.00000275, -0.00000279, -0.00000066, 0.00000335, -0.00000261, -0.00000510, 0.00000078, -0.00000217, 0.00000211, -0.00000353, -0.00000202, -0.00000990, -0.00000225, 0.00000157, 0.00000065, -0.00000784, -0.00000481, 0.00000150, -0.00000733, 0.00000233, -0.00000347, -0.00000004, -0.00001080, 0.00000260, 0.00000046, -0.00001979, 0.00000210, 0.00000299, -0.00000790, 0.00000093, -0.00002591, 0.00000340, -0.00000335, 0.00002266, 0.00001282, 0.00001078, 0.00000590, 0.00000268, 0.00000695, 0.00000801, 0.00000427, 0.00000056, 0.00000320, 0.00000307, 0.00000608, 0.00000224, 0.00001046, -0.00000412, 0.00000801, -0.00000021, },
		{ -0.00000119, 0.00001363, -0.00000169, 0.00002317, 0.00000676, 0.00000285, -0.00001633, -0.00001071, -0.00000229, -0.00000677, -0.00000367, -0.00000345, -0.00000421, -0.00000270, 0.00000149, -0.00000168, -0.00000205, -0.00000167, -0.00001077, 0.00000274, -0.00000351, -0.00000385, -0.00000935, -0.00000113, -0.00000529, 0.00000011, -0.00001247, 0.00000165, -0.00001131, -0.00000924, -0.00001750, -0.00000678, -0.00004147, -0.00002464, -0.00017196, 0.00005052, 0.00017639, 0.00002720, 0.00006826, 0.00003351, 0.00005928, 0.00003224, 0.00006245, 0.00010478, 0.00015078, -0.00016423, -0.00004270, -0.00004481, -0.00001902, -0.00001603, 0.00000212, -0.00002572, -0.00001268, 0.00000003, -0.00000301, -0.00001191, 0.00000182, -0.00000201, 0.00000109, -0.00000871, -0.00000440, -0.00000477, 0.00001495, -0.00001202, 0.00000845, -0.00000661, -0.00000616, -0.00000012, -0.00001285, 0.00000861, -0.00000661, -0.00000039, -0.00001036, 0.00000142, -0.00000075, -0.00000295, -0.00000260, 0.00000095, -0.00001145, -0.00000103, -0.00000317, 0.00000146, 0.00000192, -0.00000416, -0.00000339, 0.00000109, -0.00000211, 0.00000130, -0.00000411, -0.00000209, -0.00001129, 0.00000170, 0.00000013, 0.00000107, -0.00001122, -0.00000095, -0.00000149, -0.00000371, -0.00000011, -0.00000087, -0.00000429, -0.00000949, 0.00000673, -0.00000693, -0.00001809, 0.00000998, -0.00000881, 0.00000733, -0.00002050, -0.00001054, 0.00000143, 0.00000681, 0.00002268, 0.00000686, 0.00001226, -0.00000020, 0.00000698, 0.00000574, 0.00000791, 0.00000261, 0.00000075, 0.00000160, 0.00000553, 0.00000324, 0.00000579, 0.00000726, -0.00000460, 0.00000943, },
		{ 0.00000924, -0.00000104, 0.00001361, -0.00000206, 0.00002674, -0.00000157, 0.00000494, -0.00002191, -0.00000202, -0.00000772, -0.00000270, -0.00000683, 0.00000023, -0.00000800, 0.00000174, -0.00000132, 0.00000024, -0.00000403, -0.00000152, -0.00001101, 0.00000651, -0.00000896, -0.00000147, -0.00001149, 0.00000247, -0.00000818, 0.00000129, -0.00001434, 0.00000565, -0.00001897, -0.00000598, -0.00001818, -0.00001062, -0.00004324, -0.00004521, -0.00018227, 0.00014627, 0.00012363, 0.00004602, 0.00005482, 0.00005295, 0.00004175, 0.00005670, 0.00005918, 0.00018771, -0.00001288, -0.00015864, -0.00000136, -0.00006179, 0.00000377, -0.00002410, 0.00000662, -0.00003795, 0.00000629, -0.00000927, -0.00000056, -0.00001467, 0.00000914, -0.00000678, 0.00000238, -0.00001306, 0.00000141, -0.00000463, 0.00001500, -0.00001711, 0.00001678, -0.00001756, 0.00000352, -0.00000998, -0.00000191, 0.00000267, -0.00000434, -0.00000430, -0.00000681, 0.00000098, -0.00000042, -0.00000523, 0.00000147, -0.00000350, -0.00000983, -0.00000092, -0.00000157, 0.00000130, 0.00000138, -0.00000605, 0.00000001, -0.00000081, -0.00000064, -0.00000134, -0.00000190, -0.00000549, -0.00000874, 0.00000269, 0.00000070, -0.00000144, -0.00001173, 0.00000075, -0.00000137, -0.00000379, 0.00000093, -0.00000179, -0.00000692, -0.00000701, 0.00001002, -0.00001868, -0.00000423, 0.00000040, -0.00000064, 0.00000257, -0.00002986, 0.00000665, -0.00000551, 0.00002331, 0.00000914, 0.00001280, 0.00000375, 0.00000268, 0.00000683, 0.00000640, 0.00000593, 0.00000317, -0.00000226, 0.00000423, 0.00000438, 0.00000319, 0.00000811, 0.00000271, -0.00000262, },
		{ 0.00000413, 0.00000366, 0.00000446, 0.00000792, 0.00000416, 0.00002264, -0.00000369, 0.00000216, -0.00002115, 0.00000099, -0.00000927, -0.00000223, -0.00000580, -0.00000069, -0.00000737, 0.00000237, -0.00000052, -0.00000204, -0.00000173, -0.00000579, -0.00000490, 0.00000217, -0.00000752, -0.00000491, -0.00000667, -0.00000149, -0.00000383, -0.00000537, -0.00000612, -0.00000291, -0.00001520, -0.00001173, -0.00000909, -0.00002882, -0.00002892, -0.00009407, -0.00014045, 0.00020304, 0.00008033, 0.00006168, 0.00005164, 0.00005618, 0.00004565, 0.00005817, 0.00011702, 0.00016749, -0.00017277, -0.00005480, -0.00003768, -0.00002806, -0.00001082, -0.00000680, -0.00001549, -0.00002052, 0.00000248, -0.00000568, -0.00000924, -0.00000355, 0.00000341, -0.00000194, -0.00000685, -0.00000453, -0.00000332, 0.00000365, 0.00000479, -0.00000999, 0.00001361, -0.00001936, 0.00000559, -0.00001213, 0.00000372, -0.00000144, -0.00000415, -0.00000504, -0.00000540, 0.00000215, -0.00000297, -0.00000344, 0.00000202, -0.00000627, -0.00000915, 0.00000128, -0.00000225, 0.00000256, -0.00000124, -0.00000483, 0.00000107, -0.00000144, -0.00000109, -0.00000166, -0.00000148, -0.00000797, -0.00000565, 0.00000327, 0.00000062, -0.00000410, -0.00001200, 0.00000406, -0.00000309, -0.00000293, 0.00000177, -0.00000403, -0.00000884, -0.00000106, 0.00000550, -0.00002204, 0.00000553, -0.00001112, 0.00001747, -0.00002369, -0.00001105, 0.00000250, 0.00000494, 0.00002288, 0.00000400, 0.00001427, -0.00000273, 0.00000721, 0.00000592, 0.00000551, 0.00000669, -0.00000003, -0.00000172, 0.00000521, 0.00000381, 0.00000331, 0.00001027, -0.00000404, },
		{ -0.00000996, 0.00001221, -0.00000467, 0.00001334, -0.00000170, 0.00001534, 0.00001178, 0.00000034, -0.00000507, -0.00001496, -0.00000143, -0.00000694, -0.00000471, -0.00000179, -0.00000512, -0.00000315, 0.00000028, 0.00000202, -0.00000616, 0.00000233, -0.00001107, 0.00000292, -0.00000609, -0.00000219, -0.00001077, 0.00000075, -0.00000853, 0.00000294, -0.00001364, 0.00000436, -0.00001684, -0.00000517, -0.00002097, 0.00000092, -0.00005001, -0.00001260, -0.00015347, -0.00005715, 0.00021613, 0.00005665, 0.00007046, 0.00005475, 0.00005153, 0.00006023, 0.00006129, 0.00020435, -0.00000030, -0.00018286, 0.00000470, -0.00006764, 0.00000572, -0.00002847, 0.00000716, -0.00003516, 0.00000322, -0.00001067, 0.00000139, -0.00001758, 0.00000934, -0.00000560, 0.00000308, -0.00001536, 0.00000680, -0.00001178, 0.00001393, -0.00000857, 0.00000314, 0.00000295, -0.00001489, 0.00000329, -0.00000912, 0.00000441, -0.00000377, -0.00000449, -0.00000444, -0.00000484, 0.00000314, -0.00000581, 0.00000022, -0.00000003, -0.00000791, -0.00000792, 0.00000370, -0.00000421, 0.00000459, -0.00000477, -0.00000140, -0.00000015, -0.00000097, -0.00000279, 0.00000024, -0.00000383, -0.00000777, -0.00000368, 0.00000436, -0.00000091, -0.00000631, -0.00001170, 0.00000896, -0.00000860, 0.00000244, -0.00000200, -0.00000380, -0.00001129, 0.00000761, -0.00000751, -0.00001081, -0.00000320, -0.00000099, 0.00001357, -0.00003741, 0.00001076, -0.00000921, 0.00002479, 0.00000670, 0.00001261, 0.00000395, 0.00000103, 0.00000670, 0.00000709, 0.00000279, 0.00000962, -0.00000697, 0.00000406, 0.00000188, 0.00000620, 0.00000232, 0.00001140, },
		{ 0.00000751, -0.00000890, 0.00001376, -0.00000742, 0.00001735, -0.00000714, 0.00002348, 0.00000151, 0.00000453, -0.00001112, -0.00000945, -0.00000351, -0.00000587, -0.00000471, -0.00000086, -0.00000682, -0.00000114, 0.00000098, 0.00000030, -0.00000551, 0.00000152, -0.00001029, 0.00000376, -0.00000875, -0.00000172, -0.00001045, 0.00000132, -0.00000881, 0.00000229, -0.00001261, 0.00000352, -0.00002076, -0.00000316, -0.00002034, -0.00000402, -0.00005480, -0.00001825, -0.00019141, 0.00003543, 0.00020651, 0.00003968, 0.00008727, 0.00004206, 0.00006935, 0.00004896, 0.00013189, 0.00018511, -0.00017642, -0.00007545, -0.00002580, -0.00004000, -0.00000252, -0.00001737, -0.00000783, -0.00002418, 0.00000382, -0.00001016, -0.00000410, -0.00001063, 0.00000844, -0.00000458, -0.00000324, -0.00000971, 0.00000598, -0.00001097, 0.00001487, -0.00001341, 0.00001081, -0.00000586, -0.00001127, 0.00000367, -0.00000840, 0.00000581, -0.00000832, -0.00000087, -0.00000744, -0.00000078, -0.00000024, -0.00000391, 0.00000037, -0.00000036, -0.00001168, -0.00000231, 0.00000144, -0.00000272, 0.00000332, -0.00000490, -0.00000023, -0.00000001, -0.00000277, -0.00000154, -0.00000023, -0.00000486, -0.00000795, -0.00000042, 0.00000319, -0.00000101, -0.00001107, -0.00000592, 0.00000862, -0.00001117, 0.00000686, -0.00000672, -0.00000360, -0.00000987, 0.00001013, -0.00001561, -0.00000065, -0.00001589, 0.00002227, -0.00001671, -0.00001835, 0.00000767, -0.00000156, 0.00002746, -0.00000129, 0.00001718, -0.00000515, 0.00000714, 0.00000533, 0.00000613, 0.00000401, 0.00000721, -0.00000924, 0.00000763, -0.00000012, 0.00000732, 0.00000351, },
		{ 0.00000745, -0.00000059, -0.00000107, 0.00000741, -0.00000202, 0.00001316, -0.00000454, 0.00002425, -0.00000401, 0.00000611, -0.00001438, -0.00000610, -0.00000418, -0.00000631, -0.00000243, -0.00000310, -0.00000542, -0.00000120, 0.00000338, -0.00000428, -0.00000119, -0.00000271, -0.00000562, -0.00000082, -0.00000626, -0.00000520, -0.00000571, -0.00000321, -0.00000410, -0.00000299, -0.00000615, -0.00000585, -0.00001511, -0.00000863, -0.00001275, -0.00002066, -0.00004490, -0.00004696, -0.00019907, 0.00012550, 0.00017710, 0.00003747, 0.00009820, 0.00002973, 0.00009264, 0.00004134, 0.00023898, 0.00000192, -0.00019499, -0.00000585, -0.00005916, -0.00000457, -0.00002048, -0.00000374, -0.00002557, -0.00000286, -0.00000799, -0.00000326, -0.00001257, 0.00000164, 0.00000124, -0.00000088, -0.00001160, 0.00000223, -0.00000298, -0.00000333, 0.00000899, -0.00001112, 0.00001299, -0.00001317, -0.00000597, 0.00000342, -0.00000794, 0.00000576, -0.00001142, 0.00000298, -0.00001116, 0.00000392, -0.00000485, 0.00000024, -0.00000163, -0.00000049, -0.00001477, 0.00000473, -0.00000402, 0.00000159, -0.00000050, -0.00000212, -0.00000138, 0.00000114, -0.00000587, 0.00000223, -0.00000372, -0.00000335, -0.00000941, 0.00000431, -0.00000035, 0.00000012, -0.00001729, 0.00000530, 0.00000005, -0.00000548, 0.00000404, -0.00000697, -0.00000638, -0.00000280, 0.00000251, -0.00000952, -0.00000846, -0.00000633, 0.00002382, -0.00003738, 0.00000808, -0.00000800, 0.00002051, 0.00001063, 0.00000732, 0.00000772, -0.00000261, 0.00000741, 0.00000692, 0.00000194, 0.00000961, -0.00000196, -0.00000335, 0.00000462, 0.00000297, 0.00000467, },
		{ 0.00000211, 0.00000935, -0.00000628, 0.00000602, 0.00000027, 0.00000536, 0.00000588, 0.00000212, 0.00002025, -0.00000631, 0.00000692, -0.00001771, -0.00000160, -0.00000707, -0.00000385, -0.00000308, -0.00000317, -0.00000584, 0.00000163, 0.00000155, -0.00000496, 0.00000017, -0.00000368, -0.00000492, -0.00000240, -0.00000633, -0.00000496, -0.00000509, -0.00000395, -0.00000326, -0.00000327, -0.00000628, -0.00001027, -0.00001204, -0.00001041, -0.00001156, -0.00003358, -0.00003596, -0.00008462, -0.00018334, 0.00021543, 0.00011886, 0.00007002, 0.00007347, 0.00005707, 0.00007400, 0.00011820, 0.00022860, -0.00019714, -0.00008138, -0.00003257, -0.00003425, -0.00001013, -0.00001157, -0.00001688, -0.00001502, -0.00000258, -0.00000776, -0.00000749, -0.00000759, 0.00000250, 0.00000185, -0.00000683, -0.00000755, 0.00000464, -0.00000686, 0.00000075, 0.00000551, -0.00000944, 0.00001483, -0.00002256, 0.00000592, -0.00000424, -0.00000152, -0.00000116, -0.00000668, -0.00000005, -0.00000900, 0.00000307, -0.00000421, 0.00000023, -0.00000043, -0.00000512, -0.00001131, 0.00000612, -0.00000612, 0.00000336, -0.00000181, -0.00000133, -0.00000067, -0.00000087, -0.00000510, 0.00000291, -0.00000516, -0.00000347, -0.00000826, 0.00000612, -0.00000175, -0.00000260, -0.00001759, 0.00001297, -0.00000885, 0.00000283, -0.00000277, -0.00000487, -0.00000800, 0.00000181, -0.00000358, -0.00000282, -0.00002169, 0.00001977, -0.00000505, -0.00002309, 0.00000910, -0.00000533, 0.00002806, -0.00000102, 0.00001471, -0.00000314, 0.00000426, 0.00000571, 0.00000697, 0.00000104, 0.00001124, -0.00000890, 0.00000315, 0.00000059, 0.00000679, },
		{ 0.00000726, 0.00000215, 0.00000751, -0.00000699, 0.00000853, -0.00000320, 0.00001021, -0.00000043, 0.00000972, 0.00001307, -0.00000494, 0.00000531, -0.00001943, 0.00000245, -0.00001046, -0.00000002, -0.00000574, -0.00000167, -0.00000692, 0.00000548, -0.00000288, -0.00000259, -0.00000040, -0.00000331, -0.00000652, -0.00000180, -0.00000762, -0.00000308, -0.00000665, -0.00000277, -0.00000348, -0.00000244, -0.00001006, -0.00001046, -0.00001065, -0.00001155, -0.00001386, -0.00004415, -0.00002658, -0.00013573, -0.00012859, 0.00027894, 0.00005848, 0.00011426, 0.00003408, 0.00010450, 0.00004339, 0.00025457, 0.00002410, -0.00022089, -0.00000683, -0.00006339, -0.00000193, -0.00002426, -0.00000188, -0.00003079, 0.00000386, -0.00001297, -0.00000259, -0.00001346, 0.00000134, -0.00000101, 0.00000308, -0.00001338, 0.00000175, 0.00000035, -0.00000697, 0.00000350, 0.00000181, -0.00000539, 0.00001145, -0.00002643, 0.00001658, -0.00001390, 0.00000648, -0.00001008, 0.00000179, -0.00000732, -0.00000310, -0.00000105, -0.00000018, -0.00000276, 0.00000240, -0.00001175, -0.00000321, 0.00000178, -0.00000387, 0.00000205, -0.00000030, -0.00000296, 0.00000176, -0.00000545, -0.00000026, -0.00000088, -0.00000263, -0.00000691, -0.00000321, 0.00000356, 0.00000025, -0.00001013, -0.00000898, 0.00001067, -0.00000970, 0.00000526, -0.00000667, -0.00000461, -0.00000517, -0.00000146, 0.00000209, -0.00001330, -0.00001262, 0.00002613, -0.00002855, 0.00000271, -0.00000623, 0.00001503, 0.00001449, 0.00000431, 0.00000833, -0.00000312, 0.00000607, 0.00000674, 0.00000324, 0.00000559, 0.00000527, -0.00000794, 0.00000409, 0.00000080, },
		{ 0.00000251, 0.00000661, 0.00000201, 0.00000590, -0.00000692, 0.00000991, -0.00000549, 0.00001408, -0.00000664, 0.00001886, 0.00000157, 0.00000286, -0.00000323, -0.00001406, 0.00000015, -0.00000809, -0.00000141, -0.00000403, -0.00000445, -0.00000263, 0.00000338, -0.00000256, -0.00000378, 0.00000332, -0.00000896, -0.00000229, -0.00000663, -0.00000255, -0.00000796, -0.00000174, -0.00000804, 0.00000420, -0.00001174, -0.00000492, -0.00001741, -0.00000120, -0.00002328, -0.00000771, -0.00006240, -0.00000896, -0.00020972, -0.00001158, 0.00027531, 0.00004430, 0.00011862, 0.00003812, 0.00010620, 0.00010266, 0.00027437, -0.00021599, -0.00008740, -0.00004290, -0.00002791, -0.00001682, -0.00000633, -0.00002382, -0.00001177, -0.00000250, -0.00000790, -0.00001133, -0.00000365, -0.00000328, 0.00000576, -0.00000802, -0.00000544, 0.00000040, 0.00000179, -0.00001211, 0.00001254, -0.00000920, 0.00000727, -0.00000388, -0.00001507, 0.00001383, -0.00001424, 0.00000626, -0.00001103, 0.00000401, -0.00001014, -0.00000099, -0.00000110, -0.00000003, -0.00000224, 0.00000058, -0.00001300, 0.00000157, -0.00000163, -0.00000229, 0.00000234, -0.00000078, -0.00000242, 0.00000104, -0.00000670, 0.00000246, -0.00000360, -0.00000082, -0.00000930, 0.00000113, 0.00000123, 0.00000040, -0.00001532, 0.00000093, 0.00000385, -0.00000535, 0.00000252, -0.00000786, -0.00000346, -0.00000466, -0.00000229, 0.00000688, -0.00002942, 0.00001534, 0.00000034, -0.00001759, 0.00000438, -0.00000512, 0.00002385, 0.00000341, 0.00001032, -0.00000077, 0.00000191, 0.00000571, 0.00000653, 0.00000187, 0.00000799, -0.00000079, -0.00000470, 0.00000317, },
		{ 0.00000455, 0.00000277, 0.00000620, 0.00000136, 0.00000501, -0.00000691, 0.00001085, -0.00000670, 0.00001602, -0.00001050, 0.00002601, -0.00000980, 0.00001273, -0.00001540, -0.00000385, -0.00000671, -0.00000155, -0.00000680, 0.00000085, -0.00000964, 0.00000458, -0.00000259, 0.00000096, -0.00000617, 0.00000744, -0.00001660, 0.00000456, -0.00001327, 0.00000446, -0.00001511, 0.00000497, -0.00001374, 0.00001215, -0.00002412, 0.00000394, -0.00002515, 0.00000735, -0.00003546, -0.00000423, -0.00007252, -0.00000811, -0.00027304, 0.00014082, 0.00021498, 0.00007315, 0.00008727, 0.00008068, 0.00007780, 0.00024642, 0.00006885, -0.00026431, 0.00000813, -0.00008603, 0.00001619, -0.00004104, 0.00001273, -0.00004715, 0.00001676, -0.00002087, 0.00000293, -0.00002202, 0.00000988, -0.00001178, 0.00001297, -0.00002007, 0.00000788, -0.00000674, 0.00000481, -0.00001608, 0.00002039, -0.00001970, 0.00001987, -0.00002169, 0.00000413, 0.00000097, -0.00000675, -0.00000037, -0.00000525, 0.00000004, -0.00000848, -0.00000189, 0.00000194, -0.00000322, 0.00000110, -0.00000481, -0.00000893, 0.00000140, -0.00000222, -0.00000221, 0.00000444, -0.00000367, 0.00000052, -0.00000306, -0.00000336, 0.00000063, -0.00000229, -0.00000265, -0.00000794, 0.00000203, 0.00000162, -0.00000387, -0.00001342, 0.00000475, 0.00000062, -0.00000320, 0.00000006, -0.00000921, 0.00000094, -0.00001133, 0.00000945, -0.00000791, -0.00002170, 0.00002475, -0.00002234, 0.00000610, -0.00001138, 0.00001470, 0.00001192, 0.00000815, 0.00000371, -0.00000005, 0.00000271, 0.00000772, 0.00000269, 0.00000544, 0.00000426, -0.00000145, -0.00000447, },
		{ -0.00000326, 0.00000487, 0.00000440, 0.00000269, 0.00000486, -0.00000047, -0.00000159, 0.00000603, -0.00000153, 0.00001040, -0.00000547, 0.00002337, -0.00001179, 0.00001439, -0.00002081, 0.00000147, -0.00000897, 0.00000009, -0.00000737, 0.00000045, -0.00000849, 0.00000522, -0.00000289, -0.00000110, -0.00000091, 0.00000123, -0.00001413, 0.00000225, -0.00000991, 0.00000070, -0.00001138, 0.00000033, -0.00000567, 0.00000318, -0.00002047, -0.00000150, -0.00001622, -0.00000490, -0.00002595, -0.00002648, -0.00004723, -0.00006474, -0.00026432, 0.00026070, 0.00015552, 0.00009254, 0.00007954, 0.00009211, 0.00012967, 0.00028793, -0.00020508, -0.00011916, -0.00003014, -0.00004693, -0.00000020, -0.00002186, -0.00001016, -0.00002846, 0.00001271, -0.00001797, -0.00000551, -0.00001144, 0.00000429, -0.00000431, 0.00000247, -0.00001307, 0.00000641, -0.00000346, -0.00000349, -0.00000602, 0.00001334, -0.00001582, 0.00001790, -0.00002555, 0.00001420, -0.00000802, -0.00000233, -0.00000325, -0.00000235, -0.00000212, -0.00000911, 0.00000110, 0.00000074, -0.00000283, 0.00000060, -0.00000652, -0.00000672, 0.00000190, -0.00000427, 0.00000084, 0.00000323, -0.00000385, 0.00000060, -0.00000463, -0.00000124, -0.00000079, -0.00000103, -0.00000457, -0.00000628, 0.00000294, 0.00000060, -0.00000707, -0.00000974, 0.00000577, -0.00000056, -0.00000286, -0.00000276, -0.00000704, 0.00000068, -0.00001357, 0.00002010, -0.00003096, 0.00000861, -0.00000006, -0.00000992, 0.00000457, -0.00000993, 0.00002336, 0.00000250, 0.00001279, -0.00000470, 0.00000464, 0.00000241, 0.00000764, 0.00000171, 0.00000644, 0.00000154, -0.00000215, },
		{ -0.00000748, 0.00000254, 0.00000166, 0.00000808, -0.00000324, 0.00001175, -0.00001021, 0.00000897, -0.00000449, 0.00000963, -0.00000213, 0.00000788, 0.00001102, -0.00000362, 0.00000599, -0.00001690, -0.00000038, -0.00000507, -0.00000443, -0.00000217, -0.00000539, -0.00000138, -0.00000050, 0.00000223, -0.00000717, 0.00000880, -0.00001244, -0.00000298, -0.00000762, 0.00000099, -0.00001123, 0.00000040, -0.00001141, 0.00001040, -0.00001795, -0.00000325, -0.00001803, 0.00000323, -0.00003077, -0.00000258, -0.00006315, -0.00000195, -0.00016946, -0.00017013, 0.00031923, 0.00011793, 0.00009386, 0.00009980, 0.00007868, 0.00027218, 0.00008874, -0.00028645, 0.00000460, -0.00009253, 0.00001524, -0.00003899, 0.00001012, -0.00004562, 0.00001052, -0.00001276, -0.00000184, -0.00002230, 0.00000911, -0.00001148, 0.00001010, -0.00001546, 0.00000515, -0.00000594, 0.00000652, -0.00001716, 0.00001171, -0.00000362, -0.00000159, 0.00000472, -0.00001625, 0.00001264, -0.00001033, -0.00000191, -0.00000204, -0.00000301, -0.00000244, -0.00001071, 0.00000643, -0.00000396, 0.00000067, -0.00000311, -0.00000449, -0.00000741, 0.00000390, -0.00000753, 0.00000637, -0.00000148, -0.00000083, -0.00000279, -0.00000217, -0.00000278, 0.00000103, -0.00000256, -0.00000426, -0.00000633, 0.00000553, -0.00000375, -0.00000550, -0.00000976, 0.00000978, -0.00000518, -0.00000014, -0.00000827, 0.00000133, -0.00001039, 0.00000110, 0.00000764, -0.00002950, 0.00002221, -0.00002367, 0.00001659, -0.00001685, 0.00001320, 0.00000948, 0.00001027, 0.00000310, -0.00000159, 0.00000356, 0.00000595, 0.00000263, 0.00000660, 0.00000079, 0.00000484, },
		{ 0.00000707, -0.00001138, 0.00000751, -0.00000016, 0.00000800, -0.00000388, 0.00001296, -0.00001460, 0.00001512, -0.00001118, 0.00001711, -0.00001152, 0.00001940, -0.00000104, 0.00000497, -0.00000346, -0.00001166, -0.00000168, -0.00000335, -0.00000638, -0.00000002, -0.00000736, 0.00000154, -0.00000214, 0.00000252, -0.00000610, 0.00000932, -0.00001844, 0.00000230, -0.00001067, 0.00000393, -0.00001506, 0.00000360, -0.00001198, 0.00001264, -0.00002728, 0.00000460, -0.00002284, 0.00000693, -0.00004045, 0.00000219, -0.00007527, 0.00001127, -0.00026196, -0.00003539, 0.00034132, 0.00007369, 0.00012460, 0.00008226, 0.00015437, 0.00031172, -0.00020085, -0.00014759, -0.00001866, -0.00006671, 0.00001367, -0.00003304, -0.00000048, -0.00003950, 0.00001879, -0.00001925, -0.00000369, -0.00001822, 0.00001087, -0.00001192, 0.00000792, -0.00001593, 0.00001009, -0.00000878, 0.00000572, -0.00001833, 0.00001867, -0.00001317, 0.00000709, -0.00000410, -0.00000802, 0.00000989, -0.00001368, 0.00000235, -0.00000492, 0.00000007, -0.00000775, -0.00000553, 0.00000549, -0.00000411, -0.00000005, -0.00000254, -0.00000610, -0.00000468, 0.00000175, -0.00000583, 0.00000765, -0.00000344, -0.00000054, -0.00000347, -0.00000166, -0.00000276, 0.00000169, -0.00000313, -0.00000541, -0.00000400, 0.00000490, -0.00000538, -0.00000515, -0.00000860, 0.00001289, -0.00001036, 0.00000202, -0.00001066, 0.00000622, -0.00001968, 0.00001948, -0.00001911, -0.00000065, -0.00000113, -0.00000997, 0.00001531, -0.00002004, 0.00002603, -0.00000266, 0.00001849, -0.00000931, 0.00000657, 0.00000051, 0.00000804, 0.00000010, 0.00000888, -0.00000317, },
		{ 0.00000140, 0.00000145, -0.00000769, 0.00000660, 0.00000255, 0.00000221, 0.00000328, 0.00000469, -0.00000866, 0.00001128, -0.00000834, 0.00001512, -0.00001171, 0.00002300, -0.00000712, 0.00000868, -0.00000917, -0.00000858, -0.00000013, -0.00000600, -0.00000348, -0.00000276, -0.00000399, -0.00000128, 0.00000177, -0.00000255, 0.00000182, -0.00000032, -0.00001389, -0.00000095, -0.00000522, -0.00000275, -0.00000907, -0.00000246, -0.00000156, -0.00000075, -0.00002002, -0.00000134, -0.00001383, -0.00000744, -0.00003018, -0.00001516, -0.00005783, -0.00002500, -0.00030760, 0.00011286, 0.00032682, 0.00003670, 0.00017386, 0.00004158, 0.00033259, 0.00008868, -0.00029346, -0.00001455, -0.00008241, -0.00000446, -0.00002158, -0.00000507, -0.00003196, -0.00000585, 0.00000038, -0.00000893, -0.00001747, 0.00000011, -0.00000200, -0.00000120, -0.00000537, -0.00000210, 0.00000127, -0.00000167, -0.00000453, -0.00000781, 0.00001370, -0.00001354, 0.00000909, -0.00000694, -0.00000348, 0.00000707, -0.00001647, 0.00000824, -0.00001032, 0.00000501, -0.00001535, 0.00000427, -0.00000103, 0.00000010, -0.00000442, 0.00000197, -0.00001172, 0.00000192, -0.00000450, 0.00000106, 0.00000348, -0.00000128, -0.00000406, 0.00000016, -0.00000541, 0.00000156, -0.00000146, -0.00000070, -0.00000962, 0.00000226, -0.00000110, -0.00000052, -0.00001155, 0.00000038, 0.00000774, -0.00000953, -0.00000131, -0.00000426, -0.00000212, -0.00001065, 0.00001582, -0.00002502, 0.00001582, -0.00002695, 0.00002191, -0.00001151, 0.00000383, 0.00001287, 0.00000512, 0.00000983, -0.00000838, 0.00000777, 0.00000231, 0.00000406, 0.00000485, 0.00000272, },
		{ -0.00000384, 0.00000770, -0.00000653, -0.00000087, 0.00000377, 0.00000457, -0.00000198, 0.00000983, -0.00000492, 0.00000044, 0.00000352, -0.00000138, 0.00000852, -0.00000680, 0.00002222, -0.00001093, 0.00001158, -0.00001543, -0.00000233, -0.00000266, -0.00000532, -0.00000349, -0.00000223, -0.00000428, -0.00000031, 0.00000154, -0.00000224, 0.00000365, -0.00000646, -0.00001111, -0.00000067, -0.00000462, -0.00000485, -0.00000775, -0.00000170, 0.00000053, -0.00000908, -0.00001498, -0.00000273, -0.00001249, -0.00001566, -0.00002512, -0.00002619, -0.00004519, -0.00008084, -0.00031630, 0.00028244, 0.00023921, 0.00007465, 0.00015311, 0.00011761, 0.00039456, -0.00024028, -0.00014391, -0.00003543, -0.00005916, -0.00000200, -0.00001720, -0.00001477, -0.00002715, 0.00000279, -0.00000388, -0.00001341, -0.00001306, 0.00000299, -0.00000341, -0.00000304, -0.00000510, 0.00000201, -0.00000148, -0.00000142, -0.00000750, -0.00000167, 0.00001025, -0.00001585, 0.00001527, -0.00001444, 0.00000615, -0.00000389, -0.00000897, 0.00000491, -0.00000798, 0.00000184, -0.00001443, 0.00000758, -0.00000330, -0.00000004, -0.00000357, 0.00000128, -0.00001286, 0.00000458, -0.00000700, 0.00000538, 0.00000069, -0.00000133, -0.00000465, 0.00000127, -0.00000709, 0.00000460, -0.00000356, -0.00000028, -0.00001119, 0.00000598, -0.00000507, 0.00000330, -0.00001810, 0.00001208, -0.00000272, -0.00000498, -0.00000523, 0.00000259, -0.00001342, 0.00000719, -0.00000593, -0.00000205, -0.00000508, -0.00001479, 0.00002607, -0.00002286, 0.00002191, -0.00000257, 0.00001708, -0.00000544, 0.00000102, 0.00000446, 0.00000457, 0.00000087, 0.00000879, },
		{ 0.00000466, -0.00000233, 0.00000716, -0.00000912, 0.00000287, 0.00000312, 0.00000266, -0.00000022, 0.00000978, -0.00000896, 0.00000595, -0.00000220, 0.00000460, 0.00000151, 0.00000051, 0.00001801, -0.00001250, 0.00001335, -0.00002131, 0.00000607, -0.00000926, -0.00000058, -0.00000698, 0.00000177, -0.00000853, 0.00000497, -0.00000292, 0.00000318, -0.00000173, -0.00000645, -0.00001186, 0.00000352, -0.00000934, -0.00000171, -0.00001078, 0.00000488, -0.00000683, -0.00000866, -0.00001476, 0.00000029, -0.00001883, -0.00001641, -0.00002558, -0.00003098, -0.00004302, -0.00015068, -0.00026914, 0.00043297, 0.00011057, 0.00017524, 0.00007015, 0.00034266, 0.00014281, -0.00034678, 0.00000045, -0.00009994, 0.00000088, -0.00003155, 0.00000609, -0.00004227, 0.00000176, -0.00001083, 0.00000418, -0.00002665, 0.00000341, -0.00000627, 0.00000320, -0.00001251, 0.00000599, -0.00000434, 0.00000252, -0.00000702, -0.00000387, -0.00000015, 0.00000740, -0.00001682, 0.00002173, -0.00002360, 0.00001734, -0.00001837, 0.00000523, -0.00000592, 0.00000092, -0.00000813, -0.00000524, 0.00000319, -0.00000011, -0.00000489, 0.00000300, -0.00000623, -0.00000725, 0.00000121, -0.00000347, 0.00000434, 0.00000163, -0.00000535, -0.00000001, -0.00000338, -0.00000278, 0.00000239, -0.00000127, -0.00000493, -0.00000660, 0.00000282, -0.00000104, -0.00000207, -0.00001483, 0.00001585, -0.00000942, -0.00000337, -0.00000245, -0.00000167, -0.00000962, 0.00000759, -0.00001147, 0.00001190, -0.00003145, 0.00001975, -0.00000017, -0.00000418, 0.00001364, 0.00000148, 0.00001310, -0.00000925, 0.00000570, 0.00000415, 0.00000191, 0.00000436, },
		{ 0.00000742, -0.00000084, 0.00000226, 0.00000306, -0.00000845, 0.00000555, 0.00000105, 0.00000235, 0.00000156, 0.00000801, -0.00001091, 0.00001040, -0.00000813, 0.00001191, -0.00000834, 0.00001358, 0.00000494, -0.00000453, 0.00000569, -0.00001725, 0.00000595, -0.00000935, -0.00000170, -0.00000394, -0.00000153, -0.00000539, 0.00000316, 0.00000037, 0.00000014, -0.00000053, -0.00001276, -0.00000345, -0.00000280, -0.00000407, -0.00000887, -0.00000107, -0.00000283, -0.00000261, -0.00001855, -0.00000038, -0.00001475, -0.00000792, -0.00003501, -0.00000354, -0.00006298, -0.00001913, -0.00026455, -0.00011060, 0.00046955, 0.00005646, 0.00021393, 0.00009810, 0.00046930, -0.00026150, -0.00016240, -0.00003623, -0.00006223, -0.00000984, -0.00001217, -0.00001724, -0.00002567, -0.00000119, -0.00000320, -0.00000988, -0.00001555, -0.00000025, 0.00000014, -0.00000720, -0.00000370, 0.00000274, -0.00000043, -0.00000418, -0.00000158, -0.00000950, 0.00001032, -0.00000747, -0.00000200, 0.00001231, -0.00001887, 0.00001502, -0.00002039, 0.00001027, -0.00001015, 0.00000351, -0.00001212, 0.00000011, 0.00000144, -0.00000077, -0.00000462, 0.00000499, -0.00001115, -0.00000288, -0.00000077, -0.00000099, 0.00000423, 0.00000007, -0.00000626, 0.00000251, -0.00000707, 0.00000188, -0.00000032, 0.00000049, -0.00000913, -0.00000175, -0.00000077, 0.00000455, -0.00001166, -0.00000406, 0.00001112, -0.00001147, -0.00000227, 0.00000089, -0.00000940, 0.00000295, -0.00000867, 0.00001004, -0.00001083, -0.00002010, 0.00002857, -0.00001532, 0.00001334, -0.00000009, 0.00001212, 0.00000060, -0.00000383, 0.00000619, 0.00000367, 0.00000030, },
		{ 0.00000539, 0.00000294, 0.00000044, 0.00000223, 0.00000155, -0.00000856, 0.00000913, -0.00000387, 0.00000621, -0.00000027, 0.00000797, -0.00001300, 0.00001453, -0.00001380, 0.00001889, -0.00001799, 0.00002843, -0.00001369, 0.00001111, -0.00001034, -0.00000344, -0.00000381, -0.00000148, -0.00000936, 0.00000548, -0.00001200, 0.00000528, -0.00000537, 0.00001034, -0.00001104, 0.00000768, -0.00002440, 0.00001180, -0.00001758, 0.00000927, -0.00002314, 0.00001727, -0.00002240, 0.00001252, -0.00003605, 0.00002221, -0.00004162, 0.00001430, -0.00006252, 0.00002875, -0.00011103, 0.00001094, -0.00039359, 0.00013959, 0.00038007, 0.00011066, 0.00015761, 0.00031565, 0.00023771, -0.00042964, 0.00003192, -0.00013194, 0.00002221, -0.00005923, 0.00003116, -0.00006342, 0.00001957, -0.00003003, 0.00002040, -0.00003698, 0.00001142, -0.00001855, 0.00001630, -0.00002706, 0.00001813, -0.00001319, 0.00001199, -0.00001784, 0.00001093, -0.00002051, 0.00002431, -0.00002727, 0.00002263, -0.00001030, -0.00000130, -0.00000013, -0.00000904, 0.00000356, -0.00000526, -0.00000209, -0.00000785, -0.00000123, 0.00000445, -0.00000640, 0.00000235, -0.00000075, -0.00000965, -0.00000291, 0.00000127, -0.00000218, 0.00000709, -0.00000576, -0.00000123, -0.00000124, -0.00000481, 0.00000196, 0.00000060, -0.00000228, -0.00000826, -0.00000192, 0.00000199, 0.00000262, -0.00001488, 0.00000409, 0.00000512, -0.00001337, 0.00000409, -0.00000586, -0.00000332, -0.00000086, -0.00000854, 0.00002181, -0.00004001, 0.00001636, 0.00000483, -0.00000074, 0.00000648, 0.00000374, 0.00000888, -0.00000281, -0.00000125, 0.00000880, -0.00000164, },
		{ -0.00000464, 0.00000928, -0.00000202, 0.00000325, 0.00000131, -0.00000066, -0.00000550, 0.00000768, -0.00000393, 0.00000633, 0.00000137, 0.00000313, -0.00000837, 0.00001059, -0.00001038, 0.00001552, -0.00001544, 0.00003058, -0.00002220, 0.00001909, -0.00001964, 0.00000524, -0.00000924, 0.00000114, -0.00000990, 0.00000643, -0.00001396, 0.00000784, -0.00000439, 0.00000887, -0.00001147, 0.00000445, -0.00002179, 0.00001211, -0.00001800, 0.00000730, -0.00001901, 0.00001512, -0.00002315, 0.00000869, -0.00002984, 0.00001677, -0.00004017, 0.00000530, -0.00005031, 0.00001201, -0.00010773, -0.00003229, -0.00043013, 0.00035785, 0.00028138, 0.00015574, 0.00017859, 0.00047559, -0.00021523, -0.00024153, 0.00000849, -0.00010342, 0.00001852, -0.00004426, 0.00001372, -0.00005359, 0.00002273, -0.00002814, 0.00001332, -0.00003276, 0.00001218, -0.00001473, 0.00000852, -0.00002193, 0.00001977, -0.00001362, 0.00000841, -0.00001459, 0.00000730, -0.00001501, 0.00001974, -0.00002769, 0.00003192, -0.00002355, 0.00000955, -0.00000986, -0.00000084, -0.00000097, -0.00000357, -0.00000380, -0.00000677, 0.00000094, 0.00000233, -0.00000630, 0.00000515, -0.00000489, -0.00000912, -0.00000045, 0.00000028, -0.00000019, 0.00000573, -0.00000789, 0.00000195, -0.00000464, -0.00000175, 0.00000157, 0.00000086, -0.00000482, -0.00000716, -0.00000183, 0.00000598, -0.00000371, -0.00001182, 0.00000755, -0.00000117, -0.00001166, 0.00000852, -0.00001464, 0.00001034, -0.00002069, 0.00001988, -0.00000603, -0.00002963, 0.00002792, -0.00000993, 0.00001383, -0.00000555, 0.00001350, -0.00000176, 0.00000098, -0.00000046, 0.00000925, },
		{ 0.00000184, 0.00000160, 0.00000374, 0.00000130, -0.00000019, 0.00000542, -0.00000751, 0.00000318, -0.00000113, 0.00000429, -0.00000136, 0.00001007, -0.00000921, 0.00000502, -0.00000327, 0.00000395, 0.00000007, 0.00000175, 0.00001544, -0.00001398, 0.00001242, -0.00001534, 0.00000196, -0.00000439, -0.00000600, 0.00000024, -0.00000483, -0.00000325, -0.00000119, 0.00000851, -0.00000712, 0.00000291, -0.00001341, -0.00000229, -0.00000595, -0.00000038, -0.00001249, 0.00000546, -0.00001155, 0.00000109, -0.00001888, 0.00000272, -0.00002034, -0.00000537, -0.00003838, 0.00000448, -0.00006257, -0.00003803, -0.00017240, -0.00032358, 0.00047206, 0.00024168, 0.00013757, 0.00038524, 0.00026781, -0.00045285, 0.00000447, -0.00011615, 0.00000394, -0.00005047, 0.00001838, -0.00004874, 0.00000482, -0.00001763, 0.00000588, -0.00002172, 0.00000033, -0.00001309, 0.00000987, -0.00002006, 0.00000754, -0.00000179, 0.00000394, -0.00001123, 0.00000518, -0.00001244, 0.00000648, -0.00000294, -0.00000633, 0.00001939, -0.00001988, 0.00000740, -0.00000834, -0.00000143, 0.00000085, -0.00000771, 0.00000074, -0.00001132, 0.00000839, -0.00000620, 0.00000177, -0.00000006, -0.00000348, -0.00001211, 0.00000624, -0.00000564, 0.00000683, -0.00000207, -0.00000274, -0.00000151, -0.00000244, -0.00000262, 0.00000440, -0.00000308, -0.00000310, -0.00001026, 0.00000444, 0.00000302, -0.00000543, -0.00001009, 0.00001098, -0.00001085, -0.00000049, -0.00000080, -0.00000672, 0.00000484, -0.00002170, 0.00003847, -0.00004363, 0.00001026, 0.00000415, 0.00000607, 0.00000323, 0.00000134, 0.00000806, -0.00000192, -0.00000021, 0.00000534, },
		{ 0.00001072, -0.00000542, 0.00000867, -0.00000363, 0.00000679, -0.00000461, 0.00000903, -0.00001244, 0.00000906, -0.00000734, 0.00001051, -0.00000684, 0.00001560, -0.00001803, 0.00001537, -0.00001457, 0.00001602, -0.00001326, 0.00001842, -0.00000261, -0.00000157, 0.00000282, -0.00000916, -0.00000161, -0.00000138, -0.00000933, 0.00000568, -0.00001193, 0.00000364, -0.00000429, 0.00001323, -0.00001608, 0.00001057, -0.00002317, 0.00000941, -0.00001678, 0.00000886, -0.00002108, 0.00001724, -0.00002776, 0.00001569, -0.00003360, 0.00001985, -0.00004317, 0.00001381, -0.00006063, 0.00003437, -0.00011989, 0.00001681, -0.00032259, -0.00014733, 0.00054526, 0.00017515, 0.00022076, 0.00053273, -0.00020035, -0.00029578, 0.00002406, -0.00012027, 0.00002832, -0.00006040, 0.00002596, -0.00006216, 0.00002860, -0.00003439, 0.00001611, -0.00003214, 0.00001295, -0.00002077, 0.00001470, -0.00002844, 0.00002238, -0.00001339, 0.00001032, -0.00001830, 0.00001334, -0.00002154, 0.00001783, -0.00001726, 0.00001206, 0.00000481, -0.00001480, 0.00000651, -0.00000904, 0.00000131, -0.00000268, -0.00000530, -0.00000112, -0.00000933, 0.00000897, -0.00000855, 0.00000566, -0.00000307, -0.00000501, -0.00000997, 0.00000778, -0.00000724, 0.00000962, -0.00000690, 0.00000114, -0.00000451, -0.00000029, -0.00000259, 0.00000521, -0.00000552, -0.00000294, -0.00001114, 0.00001066, -0.00000294, -0.00000388, -0.00000939, 0.00001211, -0.00001748, 0.00001114, -0.00001541, 0.00001247, -0.00002141, 0.00001158, 0.00001234, -0.00004021, 0.00002669, -0.00001236, 0.00002307, -0.00001345, 0.00001555, -0.00000602, 0.00000601, -0.00000437, },
		{ -0.00000208, 0.00000720, -0.00000284, 0.00000585, -0.00000188, 0.00000444, -0.00000150, 0.00000488, -0.00000908, 0.00000568, -0.00000346, 0.00000648, -0.00000177, 0.00000953, -0.00001461, 0.00001356, -0.00001396, 0.00001641, -0.00001445, 0.00002325, -0.00001193, 0.00000501, -0.00000128, -0.00000871, 0.00000135, -0.00000626, -0.00000286, -0.00000043, -0.00000727, -0.00000027, 0.00000480, 0.00000201, -0.00000836, 0.00000194, -0.00001528, 0.00000284, -0.00000963, -0.00000029, -0.00000830, 0.00000319, -0.00001739, 0.00000485, -0.00002064, 0.00000448, -0.00003190, -0.00000286, -0.00003853, 0.00000612, -0.00011489, 0.00000056, -0.00041681, 0.00004768, 0.00060462, 0.00005755, 0.00051437, 0.00027636, -0.00045705, -0.00004424, -0.00008761, -0.00002550, -0.00002841, -0.00000913, -0.00002333, -0.00001760, 0.00000162, -0.00001318, -0.00000636, -0.00000988, -0.00000418, -0.00000330, -0.00000599, -0.00000835, 0.00001118, -0.00000495, -0.00000158, -0.00000615, 0.00000249, -0.00001257, 0.00001148, -0.00001378, 0.00001613, -0.00000231, -0.00001402, 0.00001073, -0.00001523, 0.00000954, -0.00001282, 0.00000528, -0.00001177, 0.00000185, -0.00000024, -0.00000074, 0.00000030, 0.00000070, -0.00001251, 0.00000002, 0.00000067, -0.00000059, 0.00000380, -0.00000342, -0.00000219, -0.00000114, -0.00000337, 0.00000253, -0.00000013, -0.00000179, -0.00000951, -0.00000285, 0.00000764, -0.00000225, -0.00000735, -0.00000332, 0.00000475, -0.00001118, 0.00000734, -0.00001294, 0.00001317, -0.00003144, 0.00004088, -0.00003142, -0.00000088, 0.00000391, 0.00000483, 0.00001197, -0.00000871, 0.00001290, -0.00000827, 0.00000747, },
		{ 0.00000245, 0.00000316, 0.00000079, 0.00000316, -0.00000105, 0.00000436, -0.00000172, 0.00000487, -0.00000209, -0.00000367, 0.00000062, 0.00000258, -0.00000005, 0.00000619, -0.00000059, -0.00000619, 0.00000654, -0.00000824, 0.00001218, -0.00001080, 0.00002260, -0.00001801, 0.00001214, -0.00000814, -0.00000428, -0.00000020, -0.00000643, -0.00000071, -0.00000336, -0.00000552, 0.00000096, 0.00000743, -0.00000538, -0.00000298, -0.00000357, -0.00001055, -0.00000042, -0.00000700, -0.00000271, -0.00000290, -0.00000660, -0.00000940, -0.00000132, -0.00001433, -0.00000598, -0.00002572, -0.00001085, -0.00002454, -0.00002611, -0.00009774, -0.00004385, -0.00049121, 0.00032619, 0.00052403, 0.00010878, 0.00072953, -0.00027692, -0.00028177, -0.00002283, -0.00008716, -0.00000473, -0.00003533, -0.00000330, -0.00003360, 0.00000180, -0.00001006, -0.00000861, -0.00001028, -0.00000308, -0.00000835, -0.00000046, -0.00001240, 0.00000322, 0.00000469, -0.00000342, -0.00000512, -0.00000076, -0.00000347, -0.00000767, 0.00000867, -0.00001146, 0.00002063, -0.00001455, -0.00000358, 0.00000476, -0.00001169, 0.00000783, -0.00001338, 0.00000773, -0.00001577, 0.00000761, -0.00000503, 0.00000398, -0.00000298, 0.00000110, -0.00001526, 0.00000669, -0.00000470, 0.00000450, -0.00000118, -0.00000018, -0.00000535, 0.00000226, -0.00000584, 0.00000642, -0.00000490, 0.00000121, -0.00001534, 0.00000776, -0.00000030, 0.00000305, -0.00001452, 0.00000544, -0.00000562, 0.00000136, -0.00000688, 0.00000549, -0.00001147, -0.00000357, 0.00002516, -0.00003784, 0.00001982, -0.00001566, 0.00002716, -0.00000999, 0.00000814, -0.00000247, 0.00000117, },
		{ 0.00000698, 0.00000089, 0.00000244, 0.00000147, 0.00000161, -0.00000075, 0.00000456, -0.00000237, 0.00000629, -0.00000478, -0.00000304, 0.00000155, 0.00000220, -0.00000016, 0.00000801, -0.00000593, -0.00000095, 0.00000168, -0.00000377, 0.00000907, -0.00000792, 0.00002113, -0.00002353, 0.00002169, -0.00002026, 0.00000744, -0.00001009, 0.00000261, -0.00000839, 0.00000299, -0.00001211, 0.00001240, -0.00000357, 0.00000024, -0.00000869, 0.00000212, -0.00001732, 0.00000796, -0.00001655, 0.00000962, -0.00001593, 0.00000166, -0.00001673, 0.00000841, -0.00002737, 0.00000312, -0.00003880, 0.00000550, -0.00004205, -0.00003293, -0.00009580, -0.00010362, -0.00053137, 0.00067823, 0.00027031, 0.00048853, 0.00043759, -0.00058036, -0.00000311, -0.00013860, 0.00001073, -0.00006184, 0.00001465, -0.00004911, 0.00000771, -0.00002251, 0.00000913, -0.00003010, 0.00001270, -0.00002137, 0.00000861, -0.00001890, 0.00000459, -0.00000538, 0.00001251, -0.00001525, 0.00000589, -0.00000909, 0.00000193, -0.00001222, 0.00001500, -0.00001610, 0.00002864, -0.00003188, 0.00001654, -0.00001305, 0.00000394, -0.00000648, -0.00000050, -0.00000317, -0.00000735, 0.00000215, 0.00000073, -0.00000101, 0.00000181, -0.00000672, -0.00000814, 0.00000391, -0.00000211, 0.00000197, 0.00000113, -0.00000392, -0.00000149, -0.00000074, -0.00000198, 0.00000320, -0.00000257, -0.00000383, -0.00001110, 0.00000927, -0.00000160, 0.00000238, -0.00001609, 0.00000773, -0.00000679, 0.00000279, -0.00000854, 0.00001042, -0.00002621, 0.00002785, -0.00001459, -0.00000528, 0.00000021, -0.00000057, 0.00002126, -0.00001206, 0.00001081, -0.00000887, },
		{ -0.00000353, 0.00000695, 0.00000097, 0.00000049, 0.00000388, -0.00000260, 0.00000308, 0.00000159, -0.00000036, 0.00000601, -0.00000779, -0.00000088, 0.00000159, 0.00000190, 0.00000020, 0.00000900, -0.00001150, 0.00000637, -0.00000698, 0.00000663, -0.00000158, 0.00000367, 0.00000821, -0.00001533, 0.00001805, -0.00002093, 0.00000945, -0.00001090, 0.00000270, -0.00000754, 0.00000026, -0.00000720, 0.00001162, -0.00000525, -0.00000281, -0.00000339, -0.00000499, -0.00001005, 0.00000093, -0.00000841, 0.00000313, -0.00001256, -0.00000576, -0.00000410, -0.00000620, -0.00001553, -0.00001514, -0.00001795, -0.00001571, -0.00002597, -0.00007945, -0.00003773, -0.00027847, -0.00037072, 0.00087517, 0.00014022, 0.00086659, -0.00027406, -0.00034565, -0.00000881, -0.00010964, 0.00000856, -0.00004791, 0.00000101, -0.00003981, 0.00000879, -0.00001705, -0.00000237, -0.00001960, 0.00000792, -0.00001602, 0.00000170, -0.00001454, 0.00000441, -0.00000000, 0.00000405, -0.00001009, 0.00000217, -0.00000444, -0.00000612, -0.00000172, 0.00000676, -0.00000622, 0.00001892, -0.00003084, 0.00002220, -0.00002013, 0.00001057, -0.00001305, 0.00000659, -0.00001020, -0.00000196, 0.00000004, 0.00000330, -0.00000349, 0.00000339, -0.00001126, -0.00000231, 0.00000156, -0.00000069, 0.00000079, 0.00000224, -0.00000691, 0.00000262, -0.00000413, 0.00000209, -0.00000084, 0.00000056, -0.00001046, -0.00000198, 0.00000413, 0.00000350, -0.00000513, -0.00001162, 0.00000536, -0.00000166, -0.00000562, 0.00000503, -0.00000904, -0.00000536, 0.00001901, -0.00002288, 0.00001312, -0.00001958, 0.00002373, 0.00000052, 0.00000046, -0.00000016, },
		{ -0.00000446, -0.00000005, 0.00000927, -0.00000471, 0.00000616, -0.00000208, 0.00000107, 0.00000074, 0.00000366, -0.00000231, 0.00000904, -0.00001558, 0.00000826, -0.00000578, 0.00000836, -0.00000504, 0.00001403, -0.00002072, 0.00001806, -0.00002078, 0.00002392, -0.00002121, 0.00002566, -0.00001792, 0.00000972, -0.00000412, -0.00000335, -0.00000533, 0.00000433, -0.00001335, 0.00000909, -0.00001792, 0.00001539, -0.00000959, 0.00001201, -0.00002284, 0.00001992, -0.00003205, 0.00001873, -0.00002866, 0.00002508, -0.00003268, 0.00002004, -0.00004013, 0.00003750, -0.00005542, 0.00003329, -0.00007214, 0.00004919, -0.00009331, 0.00004633, -0.00018370, 0.00008458, -0.00058360, 0.00006750, 0.00076526, 0.00042320, 0.00067257, -0.00074643, 0.00005299, -0.00020093, 0.00005159, -0.00009945, 0.00004556, -0.00008330, 0.00003750, -0.00004848, 0.00003274, -0.00005292, 0.00003056, -0.00003400, 0.00002108, -0.00003558, 0.00002120, -0.00002344, 0.00002753, -0.00002593, 0.00001726, -0.00002204, 0.00001793, -0.00003077, 0.00002730, -0.00002067, 0.00002207, -0.00001189, -0.00000574, 0.00000511, -0.00000767, -0.00000020, -0.00000320, -0.00000127, -0.00000463, -0.00000721, 0.00000848, -0.00000482, 0.00000365, -0.00000454, -0.00000597, -0.00000471, 0.00000595, -0.00000630, 0.00000719, -0.00000477, -0.00000172, -0.00000025, -0.00000132, -0.00000020, 0.00000139, -0.00000344, -0.00000886, 0.00000067, 0.00000348, 0.00000480, -0.00001086, -0.00000871, 0.00000830, -0.00000528, -0.00000317, 0.00000691, -0.00002092, 0.00002118, -0.00001419, 0.00000642, -0.00000796, -0.00000496, 0.00002159, -0.00000363, 0.00000192, },
		{ -0.00000816, 0.00000065, 0.00000030, 0.00001041, -0.00000933, 0.00001195, -0.00000958, 0.00000716, -0.00000308, 0.00000580, -0.00000227, 0.00000718, -0.00001760, 0.00001310, -0.00000975, 0.00001118, -0.00000544, 0.00001167, -0.00002043, 0.00001909, -0.00002246, 0.00002852, -0.00002883, 0.00003592, -0.00003361, 0.00002731, -0.00002127, 0.00000987, -0.00001420, 0.00001175, -0.00002068, 0.00001583, -0.00002381, 0.00002513, -0.00002006, 0.00001763, -0.00002758, 0.00002565, -0.00004083, 0.00002857, -0.00003697, 0.00003567, -0.00004747, 0.00003131, -0.00004744, 0.00004676, -0.00007164, 0.00004630, -0.00008578, 0.00006790, -0.00012376, 0.00006247, -0.00021254, 0.00009367, -0.00078822, 0.00054262, 0.00061980, 0.00083622, -0.00010709, -0.00054924, 0.00010669, -0.00021607, 0.00008901, -0.00011799, 0.00005954, -0.00009844, 0.00006207, -0.00006474, 0.00004147, -0.00006236, 0.00004552, -0.00004615, 0.00002839, -0.00004382, 0.00003326, -0.00003075, 0.00003293, -0.00003401, 0.00002519, -0.00002814, 0.00002113, -0.00003491, 0.00003825, -0.00003337, 0.00003650, -0.00003191, 0.00001389, -0.00000924, 0.00000173, -0.00000711, 0.00000315, -0.00000616, -0.00000286, -0.00000694, 0.00001150, -0.00000899, 0.00000703, -0.00000900, -0.00000309, -0.00000421, 0.00000620, -0.00000783, 0.00001047, -0.00001023, 0.00000365, -0.00000345, 0.00000118, -0.00000230, 0.00000346, -0.00000838, -0.00000360, -0.00000125, 0.00000751, 0.00000067, -0.00001363, -0.00000554, 0.00001291, -0.00001571, 0.00001273, -0.00001433, 0.00000117, 0.00000947, -0.00001706, 0.00002134, -0.00002979, 0.00002085, 0.00000364, 0.00000570, },
		{ 0.00000276, -0.00000934, 0.00000117, 0.00000661, 0.00000127, -0.00000165, 0.00000510, -0.00000551, 0.00000371, 0.00000177, -0.00000049, 0.00000690, -0.00000683, -0.00000473, 0.00000310, 0.00000025, 0.00000063, 0.00000767, -0.00000617, -0.00000227, 0.00000060, -0.00000209, 0.00000847, -0.00001063, 0.00001933, -0.00002209, 0.00001934, -0.00001536, 0.00000200, -0.00000155, -0.00000431, -0.00000324, -0.00000309, -0.00000123, 0.00000352, 0.00000011, -0.00000728, 0.00000146, -0.00000621, -0.00000962, -0.00000295, -0.00000024, -0.00000422, -0.00001033, -0.00000922, 0.00000239, -0.00001031, -0.00001605, -0.00001780, -0.00000968, -0.00001995, -0.00003770, -0.00005420, -0.00007633, -0.00014455, -0.00067983, 0.00079920, 0.00081186, 0.00070970, -0.00074517, -0.00004908, -0.00013473, -0.00001836, -0.00004448, -0.00000291, -0.00004427, -0.00000356, -0.00000937, -0.00000268, -0.00002021, -0.00000318, -0.00000269, -0.00000336, -0.00001615, -0.00000003, -0.00000242, 0.00000365, -0.00000180, -0.00000272, -0.00000368, -0.00000020, -0.00000937, -0.00000286, 0.00001317, -0.00001335, 0.00001971, -0.00002288, 0.00000974, -0.00000469, -0.00000559, 0.00000237, -0.00000604, 0.00000276, -0.00001424, 0.00000798, -0.00000081, 0.00000057, -0.00000218, -0.00000154, -0.00001072, 0.00000652, -0.00000490, 0.00000278, 0.00000157, -0.00000464, 0.00000027, 0.00000069, -0.00000377, 0.00000330, -0.00000292, -0.00000443, -0.00000556, 0.00000276, 0.00000679, -0.00000160, -0.00001932, 0.00000686, 0.00000395, -0.00001144, 0.00001457, -0.00002692, 0.00002853, -0.00002647, 0.00001873, -0.00000614, -0.00001465, 0.00002001, 0.00000207, },
		{ 0.00001487, -0.00001313, 0.00000083, -0.00000342, 0.00001430, -0.00001045, 0.00001047, -0.00000762, 0.00000474, -0.00000418, 0.00000925, -0.00000746, 0.00001477, -0.00002029, 0.00001007, -0.00000950, 0.00001210, -0.00001043, 0.00001975, -0.00002312, 0.00001601, -0.00001823, 0.00001981, -0.00001566, 0.00001343, -0.00000450, -0.00000234, 0.00000392, -0.00000452, -0.00000716, 0.00001010, -0.00001923, 0.00001302, -0.00001947, 0.00001881, -0.00001755, 0.00001892, -0.00002824, 0.00002616, -0.00003515, 0.00001932, -0.00003029, 0.00003147, -0.00004226, 0.00002592, -0.00004472, 0.00004483, -0.00006317, 0.00003615, -0.00007489, 0.00006005, -0.00011063, 0.00005457, -0.00017786, 0.00007240, -0.00046876, -0.00039248, 0.00102737, 0.00111521, -0.00007978, -0.00065603, 0.00012172, -0.00023514, 0.00009022, -0.00012048, 0.00006099, -0.00010499, 0.00006330, -0.00006295, 0.00003963, -0.00006094, 0.00004017, -0.00003895, 0.00002535, -0.00004594, 0.00003391, -0.00003129, 0.00002970, -0.00002813, 0.00002180, -0.00002637, 0.00002062, -0.00003273, 0.00002650, -0.00001381, 0.00000918, -0.00000031, -0.00000967, 0.00000341, -0.00000157, -0.00000899, 0.00000764, -0.00001121, 0.00000655, -0.00001940, 0.00001832, -0.00001109, 0.00000831, -0.00000906, 0.00000340, -0.00001497, 0.00001350, -0.00001340, 0.00001257, -0.00000834, 0.00000301, -0.00000426, 0.00000447, -0.00000811, 0.00000856, -0.00001014, 0.00000244, -0.00001118, 0.00001121, 0.00000130, -0.00000302, -0.00002311, 0.00002218, -0.00001548, 0.00001061, -0.00001080, -0.00000212, 0.00001486, -0.00002985, 0.00003967, -0.00003704, 0.00001566, 0.00000113, },
		{ -0.00000122, 0.00001414, -0.00001913, 0.00000579, -0.00000103, 0.00001064, -0.00000957, 0.00001047, -0.00000966, 0.00000605, -0.00000310, 0.00000705, -0.00000330, 0.00000839, -0.00001789, 0.00001017, -0.00000825, 0.00000988, -0.00000630, 0.00001442, -0.00002154, 0.00001586, -0.00001764, 0.00002153, -0.00002091, 0.00002039, -0.00001289, 0.00000359, 0.00000183, -0.00000816, 0.00000162, -0.00000004, -0.00001051, 0.00000483, -0.00000888, 0.00000907, -0.00000886, 0.00000719, -0.00001516, 0.00001364, -0.00002580, 0.00000946, -0.00001494, 0.00001477, -0.00003062, 0.00001257, -0.00002493, 0.00002345, -0.00004847, 0.00001700, -0.00004910, 0.00003223, -0.00009750, 0.00003411, -0.00017072, 0.00005490, -0.00066311, -0.00013281, 0.00158358, 0.00079734, -0.00075869, -0.00016509, -0.00006925, -0.00008436, 0.00000194, -0.00004410, -0.00000834, -0.00004298, 0.00002443, -0.00003081, 0.00000530, -0.00002801, 0.00001681, -0.00001743, -0.00000118, -0.00001987, 0.00001666, -0.00001576, 0.00001395, -0.00001414, 0.00000860, -0.00001386, 0.00000677, -0.00002047, 0.00002339, -0.00001487, 0.00000974, -0.00000118, -0.00001235, 0.00001171, -0.00001434, 0.00000557, -0.00000562, 0.00000145, -0.00000790, -0.00000459, 0.00000914, -0.00000483, 0.00000131, -0.00000160, -0.00000578, -0.00000450, 0.00000501, -0.00000664, 0.00000828, -0.00000626, 0.00000111, -0.00000009, -0.00000137, -0.00000203, 0.00000280, -0.00000613, -0.00000069, -0.00000781, 0.00001264, 0.00000005, -0.00001177, -0.00001233, 0.00002067, -0.00001996, 0.00002043, -0.00003082, 0.00003237, -0.00002858, 0.00001200, 0.00001218, -0.00002816, 0.00001946, },
		{ 0.00000130, 0.00001414, -0.00000366, -0.00000999, 0.00000251, 0.00000616, 0.00000056, -0.00000074, 0.00000185, -0.00000351, 0.00000112, 0.00000293, 0.00000086, 0.00000405, -0.00000281, -0.00000857, 0.00000367, -0.00000136, 0.00000268, 0.00000271, 0.00000278, -0.00001266, 0.00000902, -0.00000997, 0.00001487, -0.00001814, 0.00002125, -0.00001777, 0.00000901, -0.00000246, -0.00000761, 0.00000657, -0.00000857, -0.00000212, -0.00000216, 0.00000061, -0.00000169, 0.00000083, -0.00000428, -0.00000160, -0.00000112, -0.00001427, 0.00000002, -0.00000046, -0.00000508, -0.00001445, -0.00000180, -0.00000505, -0.00000227, -0.00002862, -0.00000379, -0.00001976, -0.00000981, -0.00006550, -0.00000556, -0.00014156, -0.00001869, -0.00085418, 0.00037502, 0.00215186, -0.00030533, -0.00061086, -0.00000360, -0.00014378, -0.00000062, -0.00004804, -0.00000256, -0.00005006, 0.00000591, -0.00000979, -0.00000728, -0.00001812, -0.00000181, -0.00000210, -0.00000375, -0.00001939, 0.00000351, -0.00000098, -0.00000191, 0.00000089, -0.00000212, -0.00000304, -0.00000344, -0.00000546, -0.00000583, 0.00001572, -0.00001450, 0.00001219, -0.00000619, -0.00000812, 0.00001107, -0.00001822, 0.00001331, -0.00001395, 0.00000953, -0.00001844, 0.00000930, -0.00000143, 0.00000196, -0.00000480, 0.00000456, -0.00001383, 0.00000582, -0.00000489, 0.00000292, 0.00000028, -0.00000062, -0.00000274, 0.00000458, -0.00000805, 0.00000590, -0.00000619, 0.00000303, -0.00001026, 0.00000329, 0.00000677, 0.00000184, -0.00002302, 0.00000841, 0.00000165, -0.00000141, -0.00000051, -0.00001149, 0.00002570, -0.00004186, 0.00004465, -0.00002665, 0.00000264, },
		{ 0.00001100, -0.00000307, 0.00001841, -0.00001512, -0.00000232, 0.00000312, 0.00000515, -0.00000067, -0.00000012, 0.00000044, -0.00000357, 0.00000306, 0.00000137, 0.00000316, 0.00000095, -0.00000333, -0.00000843, 0.00000620, -0.00000414, 0.00000594, 0.00000049, 0.00000148, -0.00001290, 0.00001202, -0.00001264, 0.00001721, -0.00002353, 0.00003112, -0.00003400, 0.00002906, -0.00002483, 0.00001493, -0.00001307, 0.00000752, -0.00001774, 0.00001669, -0.00001841, 0.00001609, -0.00001827, 0.00001532, -0.00002087, 0.00001699, -0.00003624, 0.00002872, -0.00002970, 0.00001915, -0.00004159, 0.00003217, -0.00004037, 0.00002897, -0.00006841, 0.00004559, -0.00007264, 0.00003238, -0.00012045, 0.00005204, -0.00023022, 0.00000363, -0.00122158, 0.00161700, 0.00174244, -0.00123242, 0.00000728, -0.00024242, 0.00004462, -0.00011747, 0.00005462, -0.00009359, 0.00002860, -0.00004654, 0.00003650, -0.00005565, 0.00002761, -0.00003781, 0.00003309, -0.00004287, 0.00001578, -0.00002075, 0.00002136, -0.00002518, 0.00002384, -0.00002420, 0.00001701, -0.00002356, 0.00001251, -0.00001716, 0.00002806, -0.00003280, 0.00003461, -0.00003247, 0.00001926, -0.00001478, 0.00000438, -0.00000459, 0.00000142, -0.00000535, -0.00000614, 0.00000285, 0.00000455, -0.00000645, 0.00000477, -0.00000537, -0.00000580, 0.00000078, -0.00000070, -0.00000006, 0.00000306, -0.00000429, 0.00000278, -0.00000149, -0.00000338, 0.00000252, -0.00000401, 0.00000220, -0.00001148, 0.00000892, 0.00000423, -0.00000381, -0.00002112, 0.00001711, -0.00000923, 0.00001186, -0.00002299, 0.00002455, -0.00001750, -0.00000464, 0.00002713, -0.00002823, },
		{ -0.00000450, -0.00000314, 0.00001080, 0.00000527, -0.00001290, -0.00000021, 0.00000584, 0.00000063, 0.00000229, -0.00000366, 0.00000252, -0.00000544, 0.00000603, -0.00000042, 0.00000458, -0.00000195, -0.00000365, -0.00000705, 0.00000659, -0.00000439, 0.00000639, 0.00000093, -0.00000395, -0.00000619, 0.00000692, -0.00000689, 0.00000895, -0.00001511, 0.00002488, -0.00003352, 0.00003319, -0.00003240, 0.00002476, -0.00002218, 0.00001299, -0.00002029, 0.00002141, -0.00002478, 0.00002088, -0.00002235, 0.00001902, -0.00002374, 0.00001624, -0.00003526, 0.00003348, -0.00003797, 0.00002168, -0.00004115, 0.00003519, -0.00004532, 0.00002570, -0.00006525, 0.00004877, -0.00008235, 0.00002443, -0.00010762, 0.00001791, -0.00020687, -0.00020169, -0.00125512, 0.00319696, 0.00000201, -0.00096329, 0.00013765, -0.00027325, 0.00009286, -0.00013525, 0.00007038, -0.00011276, 0.00005508, -0.00005793, 0.00003987, -0.00006157, 0.00003749, -0.00004196, 0.00003527, -0.00005144, 0.00002807, -0.00002570, 0.00002231, -0.00002624, 0.00002612, -0.00002728, 0.00001848, -0.00002582, 0.00001568, -0.00001422, 0.00002215, -0.00003056, 0.00003743, -0.00004127, 0.00003196, -0.00002920, 0.00001917, -0.00001704, 0.00001188, -0.00001616, 0.00000434, -0.00000241, 0.00000737, -0.00001039, 0.00001047, -0.00001286, 0.00000112, -0.00000389, 0.00000358, -0.00000337, 0.00000598, -0.00000726, 0.00000725, -0.00000784, 0.00000312, -0.00000402, 0.00000332, -0.00000554, -0.00000469, 0.00000790, 0.00000460, -0.00001383, -0.00000720, 0.00000867, 0.00000030, -0.00000216, -0.00001032, 0.00002498, -0.00003794, 0.00003193, -0.00000889, },
		{ -0.00001261, -0.00000007, -0.00000428, 0.00001428, -0.00000244, -0.00001287, 0.00000834, -0.00000286, 0.00000872, -0.00000779, 0.00000588, -0.00000826, 0.00000651, -0.00000535, 0.00001287, -0.00001113, 0.00001319, -0.00002220, 0.00001404, -0.00001391, 0.00001697, -0.00001546, 0.00002356, -0.00003265, 0.00002722, -0.00002769, 0.00002959, -0.00003236, 0.00003070, -0.00002363, 0.00001281, -0.00001031, 0.00000915, -0.00001512, 0.00001780, -0.00003093, 0.00003014, -0.00003133, 0.00002766, -0.00003427, 0.00003696, -0.00004443, 0.00004353, -0.00005909, 0.00004841, -0.00005248, 0.00004736, -0.00007237, 0.00006540, -0.00007962, 0.00007566, -0.00011480, 0.00009565, -0.00012885, 0.00010521, -0.00020447, 0.00016941, -0.00035307, 0.00024447, -0.00112823, 0.00004599, 0.00359533, -0.00184215, 0.00015813, -0.00039903, 0.00014718, -0.00020942, 0.00012424, -0.00015335, 0.00008059, -0.00010025, 0.00008378, -0.00009566, 0.00006333, -0.00007173, 0.00006129, -0.00006510, 0.00003868, -0.00004843, 0.00004791, -0.00005101, 0.00004485, -0.00004044, 0.00003408, -0.00004142, 0.00003159, -0.00003641, 0.00003941, -0.00003492, 0.00002494, -0.00001302, 0.00000132, -0.00000382, 0.00000163, -0.00000718, 0.00000751, -0.00001215, 0.00000566, -0.00001442, 0.00001834, -0.00001528, 0.00001159, -0.00000917, 0.00000321, -0.00001347, 0.00001156, -0.00001110, 0.00001099, -0.00000748, 0.00000568, -0.00000446, 0.00000132, -0.00000475, 0.00000307, -0.00000122, -0.00000419, -0.00000354, 0.00001107, -0.00000308, -0.00001251, -0.00000168, 0.00000548, 0.00000607, -0.00001909, 0.00002065, -0.00001218, -0.00000883, 0.00002205, },
		{ -0.00000800, 0.00000816, -0.00001723, 0.00001320, 0.00000055, 0.00000184, -0.00001684, 0.00001855, -0.00001504, 0.00002108, -0.00002267, 0.00002031, -0.00002282, 0.00002163, -0.00001864, 0.00002632, -0.00002701, 0.00002853, -0.00004029, 0.00003496, -0.00003417, 0.00003729, -0.00003535, 0.00004204, -0.00005509, 0.00005479, -0.00005763, 0.00006098, -0.00006806, 0.00007232, -0.00007122, 0.00006125, -0.00005696, 0.00005420, -0.00005827, 0.00005901, -0.00007294, 0.00007801, -0.00008380, 0.00008001, -0.00008704, 0.00009262, -0.00010387, 0.00010532, -0.00012693, 0.00012590, -0.00013651, 0.00013069, -0.00015975, 0.00016375, -0.00018859, 0.00018922, -0.00024383, 0.00024759, -0.00030658, 0.00029342, -0.00042791, 0.00043895, -0.00073648, 0.00071681, -0.00232624, 0.00268994, 0.00174674, -0.00207581, 0.00072018, -0.00073291, 0.00044146, -0.00043054, 0.00031561, -0.00032440, 0.00023940, -0.00022970, 0.00019635, -0.00020243, 0.00016579, -0.00016197, 0.00014439, -0.00014788, 0.00011762, -0.00011531, 0.00010851, -0.00011029, 0.00010388, -0.00009694, 0.00008539, -0.00009121, 0.00008126, -0.00008135, 0.00008374, -0.00008365, 0.00007649, -0.00006383, 0.00004671, -0.00004313, 0.00003643, -0.00003700, 0.00003397, -0.00003722, 0.00002873, -0.00003352, 0.00003746, -0.00003588, 0.00003281, -0.00002936, 0.00001988, -0.00002806, 0.00002707, -0.00002640, 0.00002648, -0.00002202, 0.00001963, -0.00001820, 0.00001337, -0.00001607, 0.00001530, -0.00001236, 0.00000384, -0.00000687, 0.00001571, -0.00001431, -0.00000162, -0.00000874, 0.00001820, -0.00001217, -0.00000393, 0.00001994, -0.00003133, 0.00002379, },
		{ 0.00001207, -0.00000607, 0.00000719, -0.00001783, 0.00001727, -0.00000381, -0.00000263, -0.00000612, 0.00000857, -0.00000511, 0.00000993, -0.00001342, 0.00001059, -0.00001206, 0.00001036, -0.00000362, 0.00000776, -0.00000804, 0.00000690, -0.00001896, 0.00001517, -0.00001239, 0.00001335, -0.00000854, 0.00000975, -0.00002137, 0.00002239, -0.00002541, 0.00002709, -0.00003423, 0.00004114, -0.00004527, 0.00003665, -0.00002943, 0.00002271, -0.00002218, 0.00001675, -0.00002437, 0.00002825, -0.00003529, 0.00002805, -0.00002879, 0.00002941, -0.00003717, 0.00003246, -0.00004873, 0.00004635, -0.00005648, 0.00004130, -0.00005833, 0.00005436, -0.00007296, 0.00005656, -0.00009500, 0.00008599, -0.00013771, 0.00008888, -0.00018474, 0.00013038, -0.00039566, 0.00016156, -0.00219772, 0.00483099, -0.00128760, -0.00055906, 0.00004972, -0.00023011, 0.00008951, -0.00013592, 0.00006804, -0.00011286, 0.00006786, -0.00007314, 0.00004964, -0.00006932, 0.00004981, -0.00005354, 0.00004119, -0.00005533, 0.00003652, -0.00003589, 0.00003011, -0.00003599, 0.00003653, -0.00003526, 0.00002446, -0.00003257, 0.00002850, -0.00003007, 0.00003430, -0.00004128, 0.00004301, -0.00003751, 0.00002205, -0.00001732, 0.00001045, -0.00000989, 0.00000585, -0.00000951, 0.00000179, -0.00000501, 0.00000986, -0.00001084, 0.00001087, -0.00000964, -0.00000108, -0.00000600, 0.00000726, -0.00000849, 0.00001145, -0.00000831, 0.00000657, -0.00000667, 0.00000186, -0.00000493, 0.00000755, -0.00000721, -0.00000101, 0.00000161, 0.00000580, -0.00000944, -0.00000404, -0.00000428, 0.00002120, -0.00002955, 0.00002756, -0.00001688, -0.00000292, },
		{ 0.00003017, -0.00002135, 0.00002344, -0.00002133, 0.00000782, -0.00000305, 0.00001124, -0.00002083, 0.00001828, -0.00001828, 0.00002394, -0.00002255, 0.00001851, -0.00002151, 0.00002099, -0.00002251, 0.00003243, -0.00003363, 0.00003612, -0.00004202, 0.00003192, -0.00003500, 0.00003990, -0.00004178, 0.00005007, -0.00005617, 0.00005006, -0.00005179, 0.00005103, -0.00005348, 0.00005020, -0.00004484, 0.00003739, -0.00004375, 0.00005330, -0.00006462, 0.00007050, -0.00008315, 0.00008581, -0.00008968, 0.00008512, -0.00009701, 0.00010580, -0.00011558, 0.00011647, -0.00013248, 0.00012974, -0.00014430, 0.00014215, -0.00017153, 0.00017634, -0.00020376, 0.00020578, -0.00025475, 0.00025821, -0.00031684, 0.00030897, -0.00043705, 0.00045069, -0.00069988, 0.00066820, -0.00165471, 0.00029669, 0.00428589, -0.00301609, 0.00097466, -0.00091716, 0.00054577, -0.00051357, 0.00037340, -0.00037267, 0.00028014, -0.00026995, 0.00022812, -0.00022751, 0.00018754, -0.00018190, 0.00015971, -0.00016045, 0.00013330, -0.00013464, 0.00012461, -0.00012404, 0.00011200, -0.00010203, 0.00009295, -0.00009903, 0.00008801, -0.00008474, 0.00007808, -0.00007097, 0.00005762, -0.00004553, 0.00003921, -0.00004730, 0.00004908, -0.00005285, 0.00005125, -0.00005396, 0.00004875, -0.00005428, 0.00005095, -0.00004448, 0.00004079, -0.00003675, 0.00003355, -0.00004371, 0.00003763, -0.00003442, 0.00003108, -0.00002385, 0.00002365, -0.00002371, 0.00002124, -0.00002540, 0.00002282, -0.00001709, 0.00001330, -0.00001825, 0.00001960, -0.00001476, 0.00001007, -0.00002368, 0.00002208, -0.00000618, -0.00001037, 0.00002559, -0.00003602, },
		{ -0.00000558, 0.00001742, -0.00002115, 0.00003038, -0.00003480, 0.00002330, -0.00001641, 0.00001873, -0.00002599, 0.00002517, -0.00002612, 0.00003312, -0.00003552, 0.00003157, -0.00003370, 0.00003264, -0.00003127, 0.00004020, -0.00004401, 0.00004786, -0.00005766, 0.00004946, -0.00005048, 0.00005470, -0.00005659, 0.00006492, -0.00007503, 0.00007309, -0.00007744, 0.00007799, -0.00008291, 0.00008290, -0.00007993, 0.00006884, -0.00006893, 0.00007514, -0.00008590, 0.00009190, -0.00010523, 0.00011261, -0.00012243, 0.00011785, -0.00012779, 0.00013876, -0.00015316, 0.00015649, -0.00017533, 0.00017780, -0.00019939, 0.00019835, -0.00022843, 0.00023971, -0.00027815, 0.00028517, -0.00034338, 0.00036378, -0.00045228, 0.00046043, -0.00062198, 0.00068319, -0.00107185, 0.00119205, -0.00314577, 0.00383989, 0.00107299, -0.00215107, 0.00095183, -0.00091667, 0.00061064, -0.00056851, 0.00042784, -0.00042463, 0.00034177, -0.00032334, 0.00027385, -0.00027071, 0.00023328, -0.00022293, 0.00019524, -0.00019556, 0.00016988, -0.00016643, 0.00015283, -0.00015203, 0.00014206, -0.00013122, 0.00011730, -0.00012219, 0.00011465, -0.00011140, 0.00010425, -0.00009926, 0.00008681, -0.00007122, 0.00005664, -0.00005786, 0.00005743, -0.00005960, 0.00005689, -0.00005923, 0.00005443, -0.00006018, 0.00005885, -0.00005312, 0.00004962, -0.00004427, 0.00003766, -0.00004773, 0.00004465, -0.00004260, 0.00004069, -0.00003179, 0.00002906, -0.00002836, 0.00002442, -0.00002861, 0.00002917, -0.00002395, 0.00001866, -0.00002110, 0.00002127, -0.00001758, 0.00001348, -0.00002906, 0.00003855, -0.00003538, 0.00002879, -0.00001558, },
		{ -0.00004140, 0.00003395, -0.00002346, 0.00001551, -0.00000056, -0.00001241, 0.00000593, -0.00000071, 0.00000038, -0.00000391, 0.00000076, 0.00000089, 0.00000401, -0.00000824, 0.00000416, -0.00000454, 0.00000195, 0.00000415, -0.00000058, -0.00000112, 0.00000250, -0.00001324, 0.00000561, -0.00000284, 0.00000312, -0.00000132, 0.00000517, -0.00001394, 0.00001159, -0.00001550, 0.00001450, -0.00001826, 0.00001830, -0.00001663, 0.00000159, 0.00000796, -0.00001143, 0.00000818, -0.00000881, 0.00000348, -0.00000172, -0.00000686, -0.00000308, 0.00000387, -0.00000232, -0.00000554, 0.00000043, -0.00000824, -0.00000128, -0.00001180, -0.00000266, -0.00000731, -0.00000283, -0.00001793, -0.00000194, -0.00002116, -0.00000043, -0.00005372, -0.00000094, -0.00007760, 0.00000174, -0.00026271, 0.00000871, -0.00224382, 0.00553042, -0.00224382, 0.00000871, -0.00026271, 0.00000174, -0.00007760, -0.00000094, -0.00005372, -0.00000043, -0.00002116, -0.00000194, -0.00001793, -0.00000283, -0.00000731, -0.00000266, -0.00001180, -0.00000128, -0.00000824, 0.00000043, -0.00000554, -0.00000232, 0.00000387, -0.00000308, -0.00000686, -0.00000172, 0.00000348, -0.00000881, 0.00000818, -0.00001143, 0.00000796, 0.00000159, -0.00001663, 0.00001830, -0.00001826, 0.00001450, -0.00001550, 0.00001159, -0.00001394, 0.00000517, -0.00000132, 0.00000312, -0.00000284, 0.00000561, -0.00001324, 0.00000250, -0.00000112, -0.00000058, 0.00000415, 0.00000195, -0.00000454, 0.00000416, -0.00000824, 0.00000401, 0.00000089, 0.00000076, -0.00000391, 0.00000038, -0.00000071, 0.00000593, -0.00001241, -0.00000056, 0.00001551, -0.00002346, 0.00003395, },
		{ -0.00000558, -0.00001558, 0.00002879, -0.00003538, 0.00003855, -0.00002906, 0.00001348, -0.00001758, 0.00002127, -0.00002110, 0.00001866, -0.00002395, 0.00002917, -0.00002861, 0.00002442, -0.00002836, 0.00002906, -0.00003179, 0.00004069, -0.00004260, 0.00004465, -0.00004773, 0.00003766, -0.00004427, 0.00004962, -0.00005312, 0.00005885, -0.00006018, 0.00005443, -0.00005923, 0.00005689, -0.00005960, 0.00005743, -0.00005786, 0.00005664, -0.00007122, 0.00008681, -0.00009926, 0.00010425, -0.00011140, 0.00011465, -0.00012219, 0.00011730, -0.00013122, 0.00014206, -0.00015203, 0.00015283, -0.00016643, 0.00016988, -0.00019556, 0.00019524, -0.00022293, 0.00023328, -0.00027071, 0.00027385, -0.00032334, 0.00034177, -0.00042463, 0.00042784, -0.00056851, 0.00061064, -0.00091667, 0.00095183, -0.00215107, 0.00107299, 0.00383989, -0.00314577, 0.00119205, -0.00107185, 0.00068319, -0.00062198, 0.00046043, -0.00045228, 0.00036378, -0.00034338, 0.00028517, -0.00027815, 0.00023971, -0.00022843, 0.00019835, -0.00019939, 0.00017780, -0.00017533, 0.00015649, -0.00015316, 0.00013876, -0.00012779, 0.00011785, -0.00012243, 0.00011261, -0.00010523, 0.00009190, -0.00008590, 0.00007514, -0.00006893, 0.00006884, -0.00007993, 0.00008290, -0.00008291, 0.00007799, -0.00007744, 0.00007309, -0.00007503, 0.00006492, -0.00005659, 0.00005470, -0.00005048, 0.00004946, -0.00005766, 0.00004786, -0.00004401, 0.00004020, -0.00003127, 0.00003264, -0.00003370, 0.00003157, -0.00003552, 0.00003312, -0.00002612, 0.00002517, -0.00002599, 0.00001873, -0.00001641, 0.00002330, -0.00003480, 0.00003038, -0.00002115, 0.00001742, },
		{ 0.00003017, -0.00003602, 0.00002559, -0.00001037, -0.00000618, 0.00002208, -0.00002368, 0.00001007, -0.00001476, 0.00001960, -0.00001825, 0.00001330, -0.00001709, 0.00002282, -0.00002540, 0.00002124, -0.00002371, 0.00002365, -0.00002385, 0.00003108, -0.00003442, 0.00003763, -0.00004371, 0.00003355, -0.00003675, 0.00004079, -0.00004448, 0.00005095, -0.00005428, 0.00004875, -0.00005396, 0.00005125, -0.00005285, 0.00004908, -0.00004730, 0.00003921, -0.00004553, 0.00005762, -0.00007097, 0.00007808, -0.00008474, 0.00008801, -0.00009903, 0.00009295, -0.00010203, 0.00011200, -0.00012404, 0.00012461, -0.00013464, 0.00013330, -0.00016045, 0.00015971, -0.00018190, 0.00018754, -0.00022751, 0.00022812, -0.00026995, 0.00028014, -0.00037267, 0.00037340, -0.00051357, 0.00054577, -0.00091716, 0.00097466, -0.00301609, 0.00428589, 0.00029669, -0.00165471, 0.00066820, -0.00069988, 0.00045069, -0.00043705, 0.00030897, -0.00031684, 0.00025821, -0.00025475, 0.00020578, -0.00020376, 0.00017634, -0.00017153, 0.00014215, -0.00014430, 0.00012974, -0.00013248, 0.00011647, -0.00011558, 0.00010580, -0.00009701, 0.00008512, -0.00008968, 0.00008581, -0.00008315, 0.00007050, -0.00006462, 0.00005330, -0.00004375, 0.00003739, -0.00004484, 0.00005020, -0.00005348, 0.00005103, -0.00005179, 0.00005006, -0.00005617, 0.00005007, -0.00004178, 0.00003990, -0.00003500, 0.00003192, -0.00004202, 0.00003612, -0.00003363, 0.00003243, -0.00002251, 0.00002099, -0.00002151, 0.00001851, -0.00002255, 0.00002394, -0.00001828, 0.00001828, -0.00002083, 0.00001124, -0.00000305, 0.00000782, -0.00002133, 0.00002344, -0.00002135, },
		{ 0.00001207, -0.00000292, -0.00001688, 0.00002756, -0.00002955, 0.00002120, -0.00000428, -0.00000404, -0.00000944, 0.00000580, 0.00000161, -0.00000101, -0.00000721, 0.00000755, -0.00000493, 0.00000186, -0.00000667, 0.00000657, -0.00000831, 0.00001145, -0.00000849, 0.00000726, -0.00000600, -0.00000108, -0.00000964, 0.00001087, -0.00001084, 0.00000986, -0.00000501, 0.00000179, -0.00000951, 0.00000585, -0.00000989, 0.00001045, -0.00001732, 0.00002205, -0.00003751, 0.00004301, -0.00004128, 0.00003430, -0.00003007, 0.00002850, -0.00003257, 0.00002446, -0.00003526, 0.00003653, -0.00003599, 0.00003011, -0.00003589, 0.00003652, -0.00005533, 0.00004119, -0.00005354, 0.00004981, -0.00006932, 0.00004964, -0.00007314, 0.00006786, -0.00011286, 0.00006804, -0.00013592, 0.00008951, -0.00023011, 0.00004972, -0.00055906, -0.00128760, 0.00483099, -0.00219772, 0.00016156, -0.00039566, 0.00013038, -0.00018474, 0.00008888, -0.00013771, 0.00008599, -0.00009500, 0.00005656, -0.00007296, 0.00005436, -0.00005833, 0.00004130, -0.00005648, 0.00004635, -0.00004873, 0.00003246, -0.00003717, 0.00002941, -0.00002879, 0.00002805, -0.00003529, 0.00002825, -0.00002437, 0.00001675, -0.00002218, 0.00002271, -0.00002943, 0.00003665, -0.00004527, 0.00004114, -0.00003423, 0.00002709, -0.00002541, 0.00002239, -0.00002137, 0.00000975, -0.00000854, 0.00001335, -0.00001239, 0.00001517, -0.00001896, 0.00000690, -0.00000804, 0.00000776, -0.00000362, 0.00001036, -0.00001206, 0.00001059, -0.00001342, 0.00000993, -0.00000511, 0.00000857, -0.00000612, -0.00000263, -0.00000381, 0.00001727, -0.00001783, 0.00000719, -0.00000607, },
		{ -0.00000800, 0.00002379, -0.00003133, 0.00001994, -0.00000393, -0.00001217, 0.00001820, -0.00000874, -0.00000162, -0.00001431, 0.00001571, -0.00000687, 0.00000384, -0.00001236, 0.00001530, -0.00001607, 0.00001337, -0.00001820, 0.00001963, -0.00002202, 0.00002648, -0.00002640, 0.00002707, -0.00002806, 0.00001988, -0.00002936, 0.00003281, -0.00003588, 0.00003746, -0.00003352, 0.00002873, -0.00003722, 0.00003397, -0.00003700, 0.00003643, -0.00004313, 0.00004671, -0.00006383, 0.00007649, -0.00008365, 0.00008374, -0.00008135, 0.00008126, -0.00009121, 0.00008539, -0.00009694, 0.00010388, -0.00011029, 0.00010851, -0.00011531, 0.00011762, -0.00014788, 0.00014439, -0.00016197, 0.00016579, -0.00020243, 0.00019635, -0.00022970, 0.00023940, -0.00032440, 0.00031561, -0.00043054, 0.00044146, -0.00073291, 0.00072018, -0.00207581, 0.00174674, 0.00268994, -0.00232624, 0.00071681, -0.00073648, 0.00043895, -0.00042791, 0.00029342, -0.00030658, 0.00024759, -0.00024383, 0.00018922, -0.00018859, 0.00016375, -0.00015975, 0.00013069, -0.00013651, 0.00012590, -0.00012693, 0.00010532, -0.00010387, 0.00009262, -0.00008704, 0.00008001, -0.00008380, 0.00007801, -0.00007294, 0.00005901, -0.00005827, 0.00005420, -0.00005696, 0.00006125, -0.00007122, 0.00007232, -0.00006806, 0.00006098, -0.00005763, 0.00005479, -0.00005509, 0.00004204, -0.00003535, 0.00003729, -0.00003417, 0.00003496, -0.00004029, 0.00002853, -0.00002701, 0.00002632, -0.00001864, 0.00002163, -0.00002282, 0.00002031, -0.00002267, 0.00002108, -0.00001504, 0.00001855, -0.00001684, 0.00000184, 0.00000055, 0.00001320, -0.00001723, 0.00000816, },
		{ -0.00001261, 0.00002205, -0.00000883, -0.00001218, 0.00002065, -0.00001909, 0.00000607, 0.00000548, -0.00000168, -0.00001251, -0.00000308, 0.00001107, -0.00000354, -0.00000419, -0.00000122, 0.00000307, -0.00000475, 0.00000132, -0.00000446, 0.00000568, -0.00000748, 0.00001099, -0.00001110, 0.00001156, -0.00001347, 0.00000321, -0.00000917, 0.00001159, -0.00001528, 0.00001834, -0.00001442, 0.00000566, -0.00001215, 0.00000751, -0.00000718, 0.00000163, -0.00000382, 0.00000132, -0.00001302, 0.00002494, -0.00003492, 0.00003941, -0.00003641, 0.00003159, -0.00004142, 0.00003408, -0.00004044, 0.00004485, -0.00005101, 0.00004791, -0.00004843, 0.00003868, -0.00006510, 0.00006129, -0.00007173, 0.00006333, -0.00009566, 0.00008378, -0.00010025, 0.00008059, -0.00015335, 0.00012424, -0.00020942, 0.00014718, -0.00039903, 0.00015813, -0.00184215, 0.00359533, 0.00004599, -0.00112823, 0.00024447, -0.00035307, 0.00016941, -0.00020447, 0.00010521, -0.00012885, 0.00009565, -0.00011480, 0.00007566, -0.00007962, 0.00006540, -0.00007237, 0.00004736, -0.00005248, 0.00004841, -0.00005909, 0.00004353, -0.00004443, 0.00003696, -0.00003427, 0.00002766, -0.00003133, 0.00003014, -0.00003093, 0.00001780, -0.00001512, 0.00000915, -0.00001031, 0.00001281, -0.00002363, 0.00003070, -0.00003236, 0.00002959, -0.00002769, 0.00002722, -0.00003265, 0.00002356, -0.00001546, 0.00001697, -0.00001391, 0.00001404, -0.00002220, 0.00001319, -0.00001113, 0.00001287, -0.00000535, 0.00000651, -0.00000826, 0.00000588, -0.00000779, 0.00000872, -0.00000286, 0.00000834, -0.00001287, -0.00000244, 0.00001428, -0.00000428, -0.00000007, },
		{ -0.00000450, -0.00000889, 0.00003193, -0.00003794, 0.00002498, -0.00001032, -0.00000216, 0.00000030, 0.00000867, -0.00000720, -0.00001383, 0.00000460, 0.00000790, -0.00000469, -0.00000554, 0.00000332, -0.00000402, 0.00000312, -0.00000784, 0.00000725, -0.00000726, 0.00000598, -0.00000337, 0.00000358, -0.00000389, 0.00000112, -0.00001286, 0.00001047, -0.00001039, 0.00000737, -0.00000241, 0.00000434, -0.00001616, 0.00001188, -0.00001704, 0.00001917, -0.00002920, 0.00003196, -0.00004127, 0.00003743, -0.00003056, 0.00002215, -0.00001422, 0.00001568, -0.00002582, 0.00001848, -0.00002728, 0.00002612, -0.00002624, 0.00002231, -0.00002570, 0.00002807, -0.00005144, 0.00003527, -0.00004196, 0.00003749, -0.00006157, 0.00003987, -0.00005793, 0.00005508, -0.00011276, 0.00007038, -0.00013525, 0.00009286, -0.00027325, 0.00013765, -0.00096329, 0.00000201, 0.00319696, -0.00125512, -0.00020169, -0.00020687, 0.00001791, -0.00010762, 0.00002443, -0.00008235, 0.00004877, -0.00006525, 0.00002570, -0.00004532, 0.00003519, -0.00004115, 0.00002168, -0.00003797, 0.00003348, -0.00003526, 0.00001624, -0.00002374, 0.00001902, -0.00002235, 0.00002088, -0.00002478, 0.00002141, -0.00002029, 0.00001299, -0.00002218, 0.00002476, -0.00003240, 0.00003319, -0.00003352, 0.00002488, -0.00001511, 0.00000895, -0.00000689, 0.00000692, -0.00000619, -0.00000395, 0.00000093, 0.00000639, -0.00000439, 0.00000659, -0.00000705, -0.00000365, -0.00000195, 0.00000458, -0.00000042, 0.00000603, -0.00000544, 0.00000252, -0.00000366, 0.00000229, 0.00000063, 0.00000584, -0.00000021, -0.00001290, 0.00000527, 0.00001080, -0.00000314, },
		{ 0.00001100, -0.00002823, 0.00002713, -0.00000464, -0.00001750, 0.00002455, -0.00002299, 0.00001186, -0.00000923, 0.00001711, -0.00002112, -0.00000381, 0.00000423, 0.00000892, -0.00001148, 0.00000220, -0.00000401, 0.00000252, -0.00000338, -0.00000149, 0.00000278, -0.00000429, 0.00000306, -0.00000006, -0.00000070, 0.00000078, -0.00000580, -0.00000537, 0.00000477, -0.00000645, 0.00000455, 0.00000285, -0.00000614, -0.00000535, 0.00000142, -0.00000459, 0.00000438, -0.00001478, 0.00001926, -0.00003247, 0.00003461, -0.00003280, 0.00002806, -0.00001716, 0.00001251, -0.00002356, 0.00001701, -0.00002420, 0.00002384, -0.00002518, 0.00002136, -0.00002075, 0.00001578, -0.00004287, 0.00003309, -0.00003781, 0.00002761, -0.00005565, 0.00003650, -0.00004654, 0.00002860, -0.00009359, 0.00005462, -0.00011747, 0.00004462, -0.00024242, 0.00000728, -0.00123242, 0.00174244, 0.00161700, -0.00122158, 0.00000363, -0.00023022, 0.00005204, -0.00012045, 0.00003238, -0.00007264, 0.00004559, -0.00006841, 0.00002897, -0.00004037, 0.00003217, -0.00004159, 0.00001915, -0.00002970, 0.00002872, -0.00003624, 0.00001699, -0.00002087, 0.00001532, -0.00001827, 0.00001609, -0.00001841, 0.00001669, -0.00001774, 0.00000752, -0.00001307, 0.00001493, -0.00002483, 0.00002906, -0.00003400, 0.00003112, -0.00002353, 0.00001721, -0.00001264, 0.00001202, -0.00001290, 0.00000148, 0.00000049, 0.00000594, -0.00000414, 0.00000620, -0.00000843, -0.00000333, 0.00000095, 0.00000316, 0.00000137, 0.00000306, -0.00000357, 0.00000044, -0.00000012, -0.00000067, 0.00000515, 0.00000312, -0.00000232, -0.00001512, 0.00001841, -0.00000307, },
		{ 0.00000130, 0.00000264, -0.00002665, 0.00004465, -0.00004186, 0.00002570, -0.00001149, -0.00000051, -0.00000141, 0.00000165, 0.00000841, -0.00002302, 0.00000184, 0.00000677, 0.00000329, -0.00001026, 0.00000303, -0.00000619, 0.00000590, -0.00000805, 0.00000458, -0.00000274, -0.00000062, 0.00000028, 0.00000292, -0.00000489, 0.00000582, -0.00001383, 0.00000456, -0.00000480, 0.00000196, -0.00000143, 0.00000930, -0.00001844, 0.00000953, -0.00001395, 0.00001331, -0.00001822, 0.00001107, -0.00000812, -0.00000619, 0.00001219, -0.00001450, 0.00001572, -0.00000583, -0.00000546, -0.00000344, -0.00000304, -0.00000212, 0.00000089, -0.00000191, -0.00000098, 0.00000351, -0.00001939, -0.00000375, -0.00000210, -0.00000181, -0.00001812, -0.00000728, -0.00000979, 0.00000591, -0.00005006, -0.00000256, -0.00004804, -0.00000062, -0.00014378, -0.00000360, -0.00061086, -0.00030533, 0.00215186, 0.00037502, -0.00085418, -0.00001869, -0.00014156, -0.00000556, -0.00006550, -0.00000981, -0.00001976, -0.00000379, -0.00002862, -0.00000227, -0.00000505, -0.00000180, -0.00001445, -0.00000508, -0.00000046, 0.00000002, -0.00001427, -0.00000112, -0.00000160, -0.00000428, 0.00000083, -0.00000169, 0.00000061, -0.00000216, -0.00000212, -0.00000857, 0.00000657, -0.00000761, -0.00000246, 0.00000901, -0.00001777, 0.00002125, -0.00001814, 0.00001487, -0.00000997, 0.00000902, -0.00001266, 0.00000278, 0.00000271, 0.00000268, -0.00000136, 0.00000367, -0.00000857, -0.00000281, 0.00000405, 0.00000086, 0.00000293, 0.00000112, -0.00000351, 0.00000185, -0.00000074, 0.00000056, 0.00000616, 0.00000251, -0.00000999, -0.00000366, 0.00001414, },
		{ -0.00000122, 0.00001946, -0.00002816, 0.00001218, 0.00001200, -0.00002858, 0.00003237, -0.00003082, 0.00002043, -0.00001996, 0.00002067, -0.00001233, -0.00001177, 0.00000005, 0.00001264, -0.00000781, -0.00000069, -0.00000613, 0.00000280, -0.00000203, -0.00000137, -0.00000009, 0.00000111, -0.00000626, 0.00000828, -0.00000664, 0.00000501, -0.00000450, -0.00000578, -0.00000160, 0.00000131, -0.00000483, 0.00000914, -0.00000459, -0.00000790, 0.00000145, -0.00000562, 0.00000557, -0.00001434, 0.00001171, -0.00001235, -0.00000118, 0.00000974, -0.00001487, 0.00002339, -0.00002047, 0.00000677, -0.00001386, 0.00000860, -0.00001414, 0.00001395, -0.00001576, 0.00001666, -0.00001987, -0.00000118, -0.00001743, 0.00001681, -0.00002801, 0.00000530, -0.00003081, 0.00002443, -0.00004298, -0.00000834, -0.00004410, 0.00000194, -0.00008436, -0.00006925, -0.00016509, -0.00075869, 0.00079734, 0.00158358, -0.00013281, -0.00066311, 0.00005490, -0.00017072, 0.00003411, -0.00009750, 0.00003223, -0.00004910, 0.00001700, -0.00004847, 0.00002345, -0.00002493, 0.00001257, -0.00003062, 0.00001477, -0.00001494, 0.00000946, -0.00002580, 0.00001364, -0.00001516, 0.00000719, -0.00000886, 0.00000907, -0.00000888, 0.00000483, -0.00001051, -0.00000004, 0.00000162, -0.00000816, 0.00000183, 0.00000359, -0.00001289, 0.00002039, -0.00002091, 0.00002153, -0.00001764, 0.00001586, -0.00002154, 0.00001442, -0.00000630, 0.00000988, -0.00000825, 0.00001017, -0.00001789, 0.00000839, -0.00000330, 0.00000705, -0.00000310, 0.00000605, -0.00000966, 0.00001047, -0.00000957, 0.00001064, -0.00000103, 0.00000579, -0.00001913, 0.00001414, },
		{ 0.00001487, 0.00000113, 0.00001566, -0.00003704, 0.00003967, -0.00002985, 0.00001486, -0.00000212, -0.00001080, 0.00001061, -0.00001548, 0.00002218, -0.00002311, -0.00000302, 0.00000130, 0.00001121, -0.00001118, 0.00000244, -0.00001014, 0.00000856, -0.00000811, 0.00000447, -0.00000426, 0.00000301, -0.00000834, 0.00001257, -0.00001340, 0.00001350, -0.00001497, 0.00000340, -0.00000906, 0.00000831, -0.00001109, 0.00001832, -0.00001940, 0.00000655, -0.00001121, 0.00000764, -0.00000899, -0.00000157, 0.00000341, -0.00000967, -0.00000031, 0.00000918, -0.00001381, 0.00002650, -0.00003273, 0.00002062, -0.00002637, 0.00002180, -0.00002813, 0.00002970, -0.00003129, 0.00003391, -0.00004594, 0.00002535, -0.00003895, 0.00004017, -0.00006094, 0.00003963, -0.00006295, 0.00006330, -0.00010499, 0.00006099, -0.00012048, 0.00009022, -0.00023514, 0.00012172, -0.00065603, -0.00007978, 0.00111521, 0.00102737, -0.00039248, -0.00046876, 0.00007240, -0.00017786, 0.00005457, -0.00011063, 0.00006005, -0.00007489, 0.00003615, -0.00006317, 0.00004483, -0.00004472, 0.00002592, -0.00004226, 0.00003147, -0.00003029, 0.00001932, -0.00003515, 0.00002616, -0.00002824, 0.00001892, -0.00001755, 0.00001881, -0.00001947, 0.00001302, -0.00001923, 0.00001010, -0.00000716, -0.00000452, 0.00000392, -0.00000234, -0.00000450, 0.00001343, -0.00001566, 0.00001981, -0.00001823, 0.00001601, -0.00002312, 0.00001975, -0.00001043, 0.00001210, -0.00000950, 0.00001007, -0.00002029, 0.00001477, -0.00000746, 0.00000925, -0.00000418, 0.00000474, -0.00000762, 0.00001047, -0.00001045, 0.00001430, -0.00000342, 0.00000083, -0.00001313, },
		{ 0.00000276, 0.00000207, 0.00002001, -0.00001465, -0.00000614, 0.00001873, -0.00002647, 0.00002853, -0.00002692, 0.00001457, -0.00001144, 0.00000395, 0.00000686, -0.00001932, -0.00000160, 0.00000679, 0.00000276, -0.00000556, -0.00000443, -0.00000292, 0.00000330, -0.00000377, 0.00000069, 0.00000027, -0.00000464, 0.00000157, 0.00000278, -0.00000490, 0.00000652, -0.00001072, -0.00000154, -0.00000218, 0.00000057, -0.00000081, 0.00000798, -0.00001424, 0.00000276, -0.00000604, 0.00000237, -0.00000559, -0.00000469, 0.00000974, -0.00002288, 0.00001971, -0.00001335, 0.00001317, -0.00000286, -0.00000937, -0.00000020, -0.00000368, -0.00000272, -0.00000180, 0.00000365, -0.00000242, -0.00000003, -0.00001615, -0.00000336, -0.00000269, -0.00000318, -0.00002021, -0.00000268, -0.00000937, -0.00000356, -0.00004427, -0.00000291, -0.00004448, -0.00001836, -0.00013473, -0.00004908, -0.00074517, 0.00070970, 0.00081186, 0.00079920, -0.00067983, -0.00014455, -0.00007633, -0.00005420, -0.00003770, -0.00001995, -0.00000968, -0.00001780, -0.00001605, -0.00001031, 0.00000239, -0.00000922, -0.00001033, -0.00000422, -0.00000024, -0.00000295, -0.00000962, -0.00000621, 0.00000146, -0.00000728, 0.00000011, 0.00000352, -0.00000123, -0.00000309, -0.00000324, -0.00000431, -0.00000155, 0.00000200, -0.00001536, 0.00001934, -0.00002209, 0.00001933, -0.00001063, 0.00000847, -0.00000209, 0.00000060, -0.00000227, -0.00000617, 0.00000767, 0.00000063, 0.00000025, 0.00000310, -0.00000473, -0.00000683, 0.00000690, -0.00000049, 0.00000177, 0.00000371, -0.00000551, 0.00000510, -0.00000165, 0.00000127, 0.00000661, 0.00000117, -0.00000934, },
		{ -0.00000816, 0.00000570, 0.00000364, 0.00002085, -0.00002979, 0.00002134, -0.00001706, 0.00000947, 0.00000117, -0.00001433, 0.00001273, -0.00001571, 0.00001291, -0.00000554, -0.00001363, 0.00000067, 0.00000751, -0.00000125, -0.00000360, -0.00000838, 0.00000346, -0.00000230, 0.00000118, -0.00000345, 0.00000365, -0.00001023, 0.00001047, -0.00000783, 0.00000620, -0.00000421, -0.00000309, -0.00000900, 0.00000703, -0.00000899, 0.00001150, -0.00000694, -0.00000286, -0.00000616, 0.00000315, -0.00000711, 0.00000173, -0.00000924, 0.00001389, -0.00003191, 0.00003650, -0.00003337, 0.00003825, -0.00003491, 0.00002113, -0.00002814, 0.00002519, -0.00003401, 0.00003293, -0.00003075, 0.00003326, -0.00004382, 0.00002839, -0.00004615, 0.00004552, -0.00006236, 0.00004147, -0.00006474, 0.00006207, -0.00009844, 0.00005954, -0.00011799, 0.00008901, -0.00021607, 0.00010669, -0.00054924, -0.00010709, 0.00083622, 0.00061980, 0.00054262, -0.00078822, 0.00009367, -0.00021254, 0.00006247, -0.00012376, 0.00006790, -0.00008578, 0.00004630, -0.00007164, 0.00004676, -0.00004744, 0.00003131, -0.00004747, 0.00003567, -0.00003697, 0.00002857, -0.00004083, 0.00002565, -0.00002758, 0.00001763, -0.00002006, 0.00002513, -0.00002381, 0.00001583, -0.00002068, 0.00001175, -0.00001420, 0.00000987, -0.00002127, 0.00002731, -0.00003361, 0.00003592, -0.00002883, 0.00002852, -0.00002246, 0.00001909, -0.00002043, 0.00001167, -0.00000544, 0.00001118, -0.00000975, 0.00001310, -0.00001760, 0.00000718, -0.00000227, 0.00000580, -0.00000308, 0.00000716, -0.00000958, 0.00001195, -0.00000933, 0.00001041, 0.00000030, 0.00000065, },
		{ -0.00000446, 0.00000192, -0.00000363, 0.00002159, -0.00000496, -0.00000796, 0.00000642, -0.00001419, 0.00002118, -0.00002092, 0.00000691, -0.00000317, -0.00000528, 0.00000830, -0.00000871, -0.00001086, 0.00000480, 0.00000348, 0.00000067, -0.00000886, -0.00000344, 0.00000139, -0.00000020, -0.00000132, -0.00000025, -0.00000172, -0.00000477, 0.00000719, -0.00000630, 0.00000595, -0.00000471, -0.00000597, -0.00000454, 0.00000365, -0.00000482, 0.00000848, -0.00000721, -0.00000463, -0.00000127, -0.00000320, -0.00000020, -0.00000767, 0.00000511, -0.00000574, -0.00001189, 0.00002207, -0.00002067, 0.00002730, -0.00003077, 0.00001793, -0.00002204, 0.00001726, -0.00002593, 0.00002753, -0.00002344, 0.00002120, -0.00003558, 0.00002108, -0.00003400, 0.00003056, -0.00005292, 0.00003274, -0.00004848, 0.00003750, -0.00008330, 0.00004556, -0.00009945, 0.00005159, -0.00020093, 0.00005299, -0.00074643, 0.00067257, 0.00042320, 0.00076526, 0.00006750, -0.00058360, 0.00008458, -0.00018370, 0.00004633, -0.00009331, 0.00004919, -0.00007214, 0.00003329, -0.00005542, 0.00003750, -0.00004013, 0.00002004, -0.00003268, 0.00002508, -0.00002866, 0.00001873, -0.00003205, 0.00001992, -0.00002284, 0.00001201, -0.00000959, 0.00001539, -0.00001792, 0.00000909, -0.00001335, 0.00000433, -0.00000533, -0.00000335, -0.00000412, 0.00000972, -0.00001792, 0.00002566, -0.00002121, 0.00002392, -0.00002078, 0.00001806, -0.00002072, 0.00001403, -0.00000504, 0.00000836, -0.00000578, 0.00000826, -0.00001558, 0.00000904, -0.00000231, 0.00000366, 0.00000074, 0.00000107, -0.00000208, 0.00000616, -0.00000471, 0.00000927, -0.00000005, },
		{ -0.00000353, -0.00000016, 0.00000046, 0.00000052, 0.00002373, -0.00001958, 0.00001312, -0.00002288, 0.00001901, -0.00000536, -0.00000904, 0.00000503, -0.00000562, -0.00000166, 0.00000536, -0.00001162, -0.00000513, 0.00000350, 0.00000413, -0.00000198, -0.00001046, 0.00000056, -0.00000084, 0.00000209, -0.00000413, 0.00000262, -0.00000691, 0.00000224, 0.00000079, -0.00000069, 0.00000156, -0.00000231, -0.00001126, 0.00000339, -0.00000349, 0.00000330, 0.00000004, -0.00000196, -0.00001020, 0.00000659, -0.00001305, 0.00001057, -0.00002013, 0.00002220, -0.00003084, 0.00001892, -0.00000622, 0.00000676, -0.00000172, -0.00000612, -0.00000444, 0.00000217, -0.00001009, 0.00000405, -0.00000000, 0.00000441, -0.00001454, 0.00000170, -0.00001602, 0.00000792, -0.00001960, -0.00000237, -0.00001705, 0.00000879, -0.00003981, 0.00000101, -0.00004791, 0.00000856, -0.00010964, -0.00000881, -0.00034565, -0.00027406, 0.00086659, 0.00014022, 0.00087517, -0.00037072, -0.00027847, -0.00003773, -0.00007945, -0.00002597, -0.00001571, -0.00001795, -0.00001514, -0.00001553, -0.00000620, -0.00000410, -0.00000576, -0.00001256, 0.00000313, -0.00000841, 0.00000093, -0.00001005, -0.00000499, -0.00000339, -0.00000281, -0.00000525, 0.00001162, -0.00000720, 0.00000026, -0.00000754, 0.00000270, -0.00001090, 0.00000945, -0.00002093, 0.00001805, -0.00001533, 0.00000821, 0.00000367, -0.00000158, 0.00000663, -0.00000698, 0.00000637, -0.00001150, 0.00000900, 0.00000020, 0.00000190, 0.00000159, -0.00000088, -0.00000779, 0.00000601, -0.00000036, 0.00000159, 0.00000308, -0.00000260, 0.00000388, 0.00000049, 0.00000097, 0.00000695, },
		{ 0.00000698, -0.00000887, 0.00001081, -0.00001206, 0.00002126, -0.00000057, 0.00000021, -0.00000528, -0.00001459, 0.00002785, -0.00002621, 0.00001042, -0.00000854, 0.00000279, -0.00000679, 0.00000773, -0.00001609, 0.00000238, -0.00000160, 0.00000927, -0.00001110, -0.00000383, -0.00000257, 0.00000320, -0.00000198, -0.00000074, -0.00000149, -0.00000392, 0.00000113, 0.00000197, -0.00000211, 0.00000391, -0.00000814, -0.00000672, 0.00000181, -0.00000101, 0.00000073, 0.00000215, -0.00000735, -0.00000317, -0.00000050, -0.00000648, 0.00000394, -0.00001305, 0.00001654, -0.00003188, 0.00002864, -0.00001610, 0.00001500, -0.00001222, 0.00000193, -0.00000909, 0.00000589, -0.00001525, 0.00001251, -0.00000538, 0.00000459, -0.00001890, 0.00000861, -0.00002137, 0.00001270, -0.00003010, 0.00000913, -0.00002251, 0.00000771, -0.00004911, 0.00001465, -0.00006184, 0.00001073, -0.00013860, -0.00000311, -0.00058036, 0.00043759, 0.00048853, 0.00027031, 0.00067823, -0.00053137, -0.00010362, -0.00009580, -0.00003293, -0.00004205, 0.00000550, -0.00003880, 0.00000312, -0.00002737, 0.00000841, -0.00001673, 0.00000166, -0.00001593, 0.00000962, -0.00001655, 0.00000796, -0.00001732, 0.00000212, -0.00000869, 0.00000024, -0.00000357, 0.00001240, -0.00001211, 0.00000299, -0.00000839, 0.00000261, -0.00001009, 0.00000744, -0.00002026, 0.00002169, -0.00002353, 0.00002113, -0.00000792, 0.00000907, -0.00000377, 0.00000168, -0.00000095, -0.00000593, 0.00000801, -0.00000016, 0.00000220, 0.00000155, -0.00000304, -0.00000478, 0.00000629, -0.00000237, 0.00000456, -0.00000075, 0.00000161, 0.00000147, 0.00000244, 0.00000089, },
		{ 0.00000245, 0.00000117, -0.00000247, 0.00000814, -0.00000999, 0.00002716, -0.00001566, 0.00001982, -0.00003784, 0.00002516, -0.00000357, -0.00001147, 0.00000549, -0.00000688, 0.00000136, -0.00000562, 0.00000544, -0.00001452, 0.00000305, -0.00000030, 0.00000776, -0.00001534, 0.00000121, -0.00000490, 0.00000642, -0.00000584, 0.00000226, -0.00000535, -0.00000018, -0.00000118, 0.00000450, -0.00000470, 0.00000669, -0.00001526, 0.00000110, -0.00000298, 0.00000398, -0.00000503, 0.00000761, -0.00001577, 0.00000773, -0.00001338, 0.00000783, -0.00001169, 0.00000476, -0.00000358, -0.00001455, 0.00002063, -0.00001146, 0.00000867, -0.00000767, -0.00000347, -0.00000076, -0.00000512, -0.00000342, 0.00000469, 0.00000322, -0.00001240, -0.00000046, -0.00000835, -0.00000308, -0.00001028, -0.00000861, -0.00001006, 0.00000180, -0.00003360, -0.00000330, -0.00003533, -0.00000473, -0.00008716, -0.00002283, -0.00028177, -0.00027692, 0.00072953, 0.00010878, 0.00052403, 0.00032619, -0.00049121, -0.00004385, -0.00009774, -0.00002611, -0.00002454, -0.00001085, -0.00002572, -0.00000598, -0.00001433, -0.00000132, -0.00000940, -0.00000660, -0.00000290, -0.00000271, -0.00000700, -0.00000042, -0.00001055, -0.00000357, -0.00000298, -0.00000538, 0.00000743, 0.00000096, -0.00000552, -0.00000336, -0.00000071, -0.00000643, -0.00000020, -0.00000428, -0.00000814, 0.00001214, -0.00001801, 0.00002260, -0.00001080, 0.00001218, -0.00000824, 0.00000654, -0.00000619, -0.00000059, 0.00000619, -0.00000005, 0.00000258, 0.00000062, -0.00000367, -0.00000209, 0.00000487, -0.00000172, 0.00000436, -0.00000105, 0.00000316, 0.00000079, 0.00000316, },
		{ -0.00000208, 0.00000747, -0.00000827, 0.00001290, -0.00000871, 0.00001197, 0.00000483, 0.00000391, -0.00000088, -0.00003142, 0.00004088, -0.00003144, 0.00001317, -0.00001294, 0.00000734, -0.00001118, 0.00000475, -0.00000332, -0.00000735, -0.00000225, 0.00000764, -0.00000285, -0.00000951, -0.00000179, -0.00000013, 0.00000253, -0.00000337, -0.00000114, -0.00000219, -0.00000342, 0.00000380, -0.00000059, 0.00000067, 0.00000002, -0.00001251, 0.00000070, 0.00000030, -0.00000074, -0.00000024, 0.00000185, -0.00001177, 0.00000528, -0.00001282, 0.00000954, -0.00001523, 0.00001073, -0.00001402, -0.00000231, 0.00001613, -0.00001378, 0.00001148, -0.00001257, 0.00000249, -0.00000615, -0.00000158, -0.00000495, 0.00001118, -0.00000835, -0.00000599, -0.00000330, -0.00000418, -0.00000988, -0.00000636, -0.00001318, 0.00000162, -0.00001760, -0.00002333, -0.00000913, -0.00002841, -0.00002550, -0.00008761, -0.00004424, -0.00045705, 0.00027636, 0.00051437, 0.00005755, 0.00060462, 0.00004768, -0.00041681, 0.00000056, -0.00011489, 0.00000612, -0.00003853, -0.00000286, -0.00003190, 0.00000448, -0.00002064, 0.00000485, -0.00001739, 0.00000319, -0.00000830, -0.00000029, -0.00000963, 0.00000284, -0.00001528, 0.00000194, -0.00000836, 0.00000201, 0.00000480, -0.00000027, -0.00000727, -0.00000043, -0.00000286, -0.00000626, 0.00000135, -0.00000871, -0.00000128, 0.00000501, -0.00001193, 0.00002325, -0.00001445, 0.00001641, -0.00001396, 0.00001356, -0.00001461, 0.00000953, -0.00000177, 0.00000648, -0.00000346, 0.00000568, -0.00000908, 0.00000488, -0.00000150, 0.00000444, -0.00000188, 0.00000585, -0.00000284, 0.00000720, },
		{ 0.00001072, -0.00000437, 0.00000601, -0.00000602, 0.00001555, -0.00001345, 0.00002307, -0.00001236, 0.00002669, -0.00004021, 0.00001234, 0.00001158, -0.00002141, 0.00001247, -0.00001541, 0.00001114, -0.00001748, 0.00001211, -0.00000939, -0.00000388, -0.00000294, 0.00001066, -0.00001114, -0.00000294, -0.00000552, 0.00000521, -0.00000259, -0.00000029, -0.00000451, 0.00000114, -0.00000690, 0.00000962, -0.00000724, 0.00000778, -0.00000997, -0.00000501, -0.00000307, 0.00000566, -0.00000855, 0.00000897, -0.00000933, -0.00000112, -0.00000530, -0.00000268, 0.00000131, -0.00000904, 0.00000651, -0.00001480, 0.00000481, 0.00001206, -0.00001726, 0.00001783, -0.00002154, 0.00001334, -0.00001830, 0.00001032, -0.00001339, 0.00002238, -0.00002844, 0.00001470, -0.00002077, 0.00001295, -0.00003214, 0.00001611, -0.00003439, 0.00002860, -0.00006216, 0.00002596, -0.00006040, 0.00002832, -0.00012027, 0.00002406, -0.00029578, -0.00020035, 0.00053273, 0.00022076, 0.00017515, 0.00054526, -0.00014733, -0.00032259, 0.00001681, -0.00011989, 0.00003437, -0.00006063, 0.00001381, -0.00004317, 0.00001985, -0.00003360, 0.00001569, -0.00002776, 0.00001724, -0.00002108, 0.00000886, -0.00001678, 0.00000941, -0.00002317, 0.00001057, -0.00001608, 0.00001323, -0.00000429, 0.00000364, -0.00001193, 0.00000568, -0.00000933, -0.00000138, -0.00000161, -0.00000916, 0.00000282, -0.00000157, -0.00000261, 0.00001842, -0.00001326, 0.00001602, -0.00001457, 0.00001537, -0.00001803, 0.00001560, -0.00000684, 0.00001051, -0.00000734, 0.00000906, -0.00001244, 0.00000903, -0.00000461, 0.00000679, -0.00000363, 0.00000867, -0.00000542, },
		{ 0.00000184, 0.00000534, -0.00000021, -0.00000192, 0.00000806, 0.00000134, 0.00000323, 0.00000607, 0.00000415, 0.00001026, -0.00004363, 0.00003847, -0.00002170, 0.00000484, -0.00000672, -0.00000080, -0.00000049, -0.00001085, 0.00001098, -0.00001009, -0.00000543, 0.00000302, 0.00000444, -0.00001026, -0.00000310, -0.00000308, 0.00000440, -0.00000262, -0.00000244, -0.00000151, -0.00000274, -0.00000207, 0.00000683, -0.00000564, 0.00000624, -0.00001211, -0.00000348, -0.00000006, 0.00000177, -0.00000620, 0.00000839, -0.00001132, 0.00000074, -0.00000771, 0.00000085, -0.00000143, -0.00000834, 0.00000740, -0.00001988, 0.00001939, -0.00000633, -0.00000294, 0.00000648, -0.00001244, 0.00000518, -0.00001123, 0.00000394, -0.00000179, 0.00000754, -0.00002006, 0.00000987, -0.00001309, 0.00000033, -0.00002172, 0.00000588, -0.00001763, 0.00000482, -0.00004874, 0.00001838, -0.00005047, 0.00000394, -0.00011615, 0.00000447, -0.00045285, 0.00026781, 0.00038524, 0.00013757, 0.00024168, 0.00047206, -0.00032358, -0.00017240, -0.00003803, -0.00006257, 0.00000448, -0.00003838, -0.00000537, -0.00002034, 0.00000272, -0.00001888, 0.00000109, -0.00001155, 0.00000546, -0.00001249, -0.00000038, -0.00000595, -0.00000229, -0.00001341, 0.00000291, -0.00000712, 0.00000851, -0.00000119, -0.00000325, -0.00000483, 0.00000024, -0.00000600, -0.00000439, 0.00000196, -0.00001534, 0.00001242, -0.00001398, 0.00001544, 0.00000175, 0.00000007, 0.00000395, -0.00000327, 0.00000502, -0.00000921, 0.00001007, -0.00000136, 0.00000429, -0.00000113, 0.00000318, -0.00000751, 0.00000542, -0.00000019, 0.00000130, 0.00000374, 0.00000160, },
		{ -0.00000464, 0.00000925, -0.00000046, 0.00000098, -0.00000176, 0.00001350, -0.00000555, 0.00001383, -0.00000993, 0.00002792, -0.00002963, -0.00000603, 0.00001988, -0.00002069, 0.00001034, -0.00001464, 0.00000852, -0.00001166, -0.00000117, 0.00000755, -0.00001182, -0.00000371, 0.00000598, -0.00000183, -0.00000716, -0.00000482, 0.00000086, 0.00000157, -0.00000175, -0.00000464, 0.00000195, -0.00000789, 0.00000573, -0.00000019, 0.00000028, -0.00000045, -0.00000912, -0.00000489, 0.00000515, -0.00000630, 0.00000233, 0.00000094, -0.00000677, -0.00000380, -0.00000357, -0.00000097, -0.00000084, -0.00000986, 0.00000955, -0.00002355, 0.00003192, -0.00002769, 0.00001974, -0.00001501, 0.00000730, -0.00001459, 0.00000841, -0.00001362, 0.00001977, -0.00002193, 0.00000852, -0.00001473, 0.00001218, -0.00003276, 0.00001332, -0.00002814, 0.00002273, -0.00005359, 0.00001372, -0.00004426, 0.00001852, -0.00010342, 0.00000849, -0.00024153, -0.00021523, 0.00047559, 0.00017859, 0.00015574, 0.00028138, 0.00035785, -0.00043013, -0.00003229, -0.00010773, 0.00001201, -0.00005031, 0.00000530, -0.00004017, 0.00001677, -0.00002984, 0.00000869, -0.00002315, 0.00001512, -0.00001901, 0.00000730, -0.00001800, 0.00001211, -0.00002179, 0.00000445, -0.00001147, 0.00000887, -0.00000439, 0.00000784, -0.00001396, 0.00000643, -0.00000990, 0.00000114, -0.00000924, 0.00000524, -0.00001964, 0.00001909, -0.00002220, 0.00003058, -0.00001544, 0.00001552, -0.00001038, 0.00001059, -0.00000837, 0.00000313, 0.00000137, 0.00000633, -0.00000393, 0.00000768, -0.00000550, -0.00000066, 0.00000131, 0.00000325, -0.00000202, 0.00000928, },
		{ 0.00000539, -0.00000164, 0.00000880, -0.00000125, -0.00000281, 0.00000888, 0.00000374, 0.00000648, -0.00000074, 0.00000483, 0.00001636, -0.00004001, 0.00002181, -0.00000854, -0.00000086, -0.00000332, -0.00000586, 0.00000409, -0.00001337, 0.00000512, 0.00000409, -0.00001488, 0.00000262, 0.00000199, -0.00000192, -0.00000826, -0.00000228, 0.00000060, 0.00000196, -0.00000481, -0.00000124, -0.00000123, -0.00000576, 0.00000709, -0.00000218, 0.00000127, -0.00000291, -0.00000965, -0.00000075, 0.00000235, -0.00000640, 0.00000445, -0.00000123, -0.00000785, -0.00000209, -0.00000526, 0.00000356, -0.00000904, -0.00000013, -0.00000130, -0.00001030, 0.00002263, -0.00002727, 0.00002431, -0.00002051, 0.00001093, -0.00001784, 0.00001199, -0.00001319, 0.00001813, -0.00002706, 0.00001630, -0.00001855, 0.00001142, -0.00003698, 0.00002040, -0.00003003, 0.00001957, -0.00006342, 0.00003116, -0.00005923, 0.00002221, -0.00013194, 0.00003192, -0.00042964, 0.00023771, 0.00031565, 0.00015761, 0.00011066, 0.00038007, 0.00013959, -0.00039359, 0.00001094, -0.00011103, 0.00002875, -0.00006252, 0.00001430, -0.00004162, 0.00002221, -0.00003605, 0.00001252, -0.00002240, 0.00001727, -0.00002314, 0.00000927, -0.00001758, 0.00001180, -0.00002440, 0.00000768, -0.00001104, 0.00001034, -0.00000537, 0.00000528, -0.00001200, 0.00000548, -0.00000936, -0.00000148, -0.00000381, -0.00000344, -0.00001034, 0.00001111, -0.00001369, 0.00002843, -0.00001799, 0.00001889, -0.00001380, 0.00001453, -0.00001300, 0.00000797, -0.00000027, 0.00000621, -0.00000387, 0.00000913, -0.00000856, 0.00000155, 0.00000223, 0.00000044, 0.00000294, },
		{ 0.00000742, 0.00000030, 0.00000367, 0.00000619, -0.00000383, 0.00000060, 0.00001212, -0.00000009, 0.00001334, -0.00001532, 0.00002857, -0.00002010, -0.00001083, 0.00001004, -0.00000867, 0.00000295, -0.00000940, 0.00000089, -0.00000227, -0.00001147, 0.00001112, -0.00000406, -0.00001166, 0.00000455, -0.00000077, -0.00000175, -0.00000913, 0.00000049, -0.00000032, 0.00000188, -0.00000707, 0.00000251, -0.00000626, 0.00000007, 0.00000423, -0.00000099, -0.00000077, -0.00000288, -0.00001115, 0.00000499, -0.00000462, -0.00000077, 0.00000144, 0.00000011, -0.00001212, 0.00000351, -0.00001015, 0.00001027, -0.00002039, 0.00001502, -0.00001887, 0.00001231, -0.00000200, -0.00000747, 0.00001032, -0.00000950, -0.00000158, -0.00000418, -0.00000043, 0.00000274, -0.00000370, -0.00000720, 0.00000014, -0.00000025, -0.00001555, -0.00000988, -0.00000320, -0.00000119, -0.00002567, -0.00001724, -0.00001217, -0.00000984, -0.00006223, -0.00003623, -0.00016240, -0.00026150, 0.00046930, 0.00009810, 0.00021393, 0.00005646, 0.00046955, -0.00011060, -0.00026455, -0.00001913, -0.00006298, -0.00000354, -0.00003501, -0.00000792, -0.00001475, -0.00000038, -0.00001855, -0.00000261, -0.00000283, -0.00000107, -0.00000887, -0.00000407, -0.00000280, -0.00000345, -0.00001276, -0.00000053, 0.00000014, 0.00000037, 0.00000316, -0.00000539, -0.00000153, -0.00000394, -0.00000170, -0.00000935, 0.00000595, -0.00001725, 0.00000569, -0.00000453, 0.00000494, 0.00001358, -0.00000834, 0.00001191, -0.00000813, 0.00001040, -0.00001091, 0.00000801, 0.00000156, 0.00000235, 0.00000105, 0.00000555, -0.00000845, 0.00000306, 0.00000226, -0.00000084, },
		{ 0.00000466, 0.00000436, 0.00000191, 0.00000415, 0.00000570, -0.00000925, 0.00001310, 0.00000148, 0.00001364, -0.00000418, -0.00000017, 0.00001975, -0.00003145, 0.00001190, -0.00001147, 0.00000759, -0.00000962, -0.00000167, -0.00000245, -0.00000337, -0.00000942, 0.00001585, -0.00001483, -0.00000207, -0.00000104, 0.00000282, -0.00000660, -0.00000493, -0.00000127, 0.00000239, -0.00000278, -0.00000338, -0.00000001, -0.00000535, 0.00000163, 0.00000434, -0.00000347, 0.00000121, -0.00000725, -0.00000623, 0.00000300, -0.00000489, -0.00000011, 0.00000319, -0.00000524, -0.00000813, 0.00000092, -0.00000592, 0.00000523, -0.00001837, 0.00001734, -0.00002360, 0.00002173, -0.00001682, 0.00000740, -0.00000015, -0.00000387, -0.00000702, 0.00000252, -0.00000434, 0.00000599, -0.00001251, 0.00000320, -0.00000627, 0.00000341, -0.00002665, 0.00000418, -0.00001083, 0.00000176, -0.00004227, 0.00000609, -0.00003155, 0.00000088, -0.00009994, 0.00000045, -0.00034678, 0.00014281, 0.00034266, 0.00007015, 0.00017524, 0.00011057, 0.00043297, -0.00026914, -0.00015068, -0.00004302, -0.00003098, -0.00002558, -0.00001641, -0.00001883, 0.00000029, -0.00001476, -0.00000866, -0.00000683, 0.00000488, -0.00001078, -0.00000171, -0.00000934, 0.00000352, -0.00001186, -0.00000645, -0.00000173, 0.00000318, -0.00000292, 0.00000497, -0.00000853, 0.00000177, -0.00000698, -0.00000058, -0.00000926, 0.00000607, -0.00002131, 0.00001335, -0.00001250, 0.00001801, 0.00000051, 0.00000151, 0.00000460, -0.00000220, 0.00000595, -0.00000896, 0.00000978, -0.00000022, 0.00000266, 0.00000312, 0.00000287, -0.00000912, 0.00000716, -0.00000233, },
		{ -0.00000384, 0.00000879, 0.00000087, 0.00000457, 0.00000446, 0.00000102, -0.00000544, 0.00001708, -0.00000257, 0.00002191, -0.00002286, 0.00002607, -0.00001479, -0.00000508, -0.00000205, -0.00000593, 0.00000719, -0.00001342, 0.00000259, -0.00000523, -0.00000498, -0.00000272, 0.00001208, -0.00001810, 0.00000330, -0.00000507, 0.00000598, -0.00001119, -0.00000028, -0.00000356, 0.00000460, -0.00000709, 0.00000127, -0.00000465, -0.00000133, 0.00000069, 0.00000538, -0.00000700, 0.00000458, -0.00001286, 0.00000128, -0.00000357, -0.00000004, -0.00000330, 0.00000758, -0.00001443, 0.00000184, -0.00000798, 0.00000491, -0.00000897, -0.00000389, 0.00000615, -0.00001444, 0.00001527, -0.00001585, 0.00001025, -0.00000167, -0.00000750, -0.00000142, -0.00000148, 0.00000201, -0.00000510, -0.00000304, -0.00000341, 0.00000299, -0.00001306, -0.00001341, -0.00000388, 0.00000279, -0.00002715, -0.00001477, -0.00001720, -0.00000200, -0.00005916, -0.00003543, -0.00014391, -0.00024028, 0.00039456, 0.00011761, 0.00015311, 0.00007465, 0.00023921, 0.00028244, -0.00031630, -0.00008084, -0.00004519, -0.00002619, -0.00002512, -0.00001566, -0.00001249, -0.00000273, -0.00001498, -0.00000908, 0.00000053, -0.00000170, -0.00000775, -0.00000485, -0.00000462, -0.00000067, -0.00001111, -0.00000646, 0.00000365, -0.00000224, 0.00000154, -0.00000031, -0.00000428, -0.00000223, -0.00000349, -0.00000532, -0.00000266, -0.00000233, -0.00001543, 0.00001158, -0.00001093, 0.00002222, -0.00000680, 0.00000852, -0.00000138, 0.00000352, 0.00000044, -0.00000492, 0.00000983, -0.00000198, 0.00000457, 0.00000377, -0.00000087, -0.00000653, 0.00000770, },
		{ 0.00000140, 0.00000272, 0.00000485, 0.00000406, 0.00000231, 0.00000777, -0.00000838, 0.00000983, 0.00000512, 0.00001287, 0.00000383, -0.00001151, 0.00002191, -0.00002695, 0.00001582, -0.00002502, 0.00001582, -0.00001065, -0.00000212, -0.00000426, -0.00000131, -0.00000953, 0.00000774, 0.00000038, -0.00001155, -0.00000052, -0.00000110, 0.00000226, -0.00000962, -0.00000070, -0.00000146, 0.00000156, -0.00000541, 0.00000016, -0.00000406, -0.00000128, 0.00000348, 0.00000106, -0.00000450, 0.00000192, -0.00001172, 0.00000197, -0.00000442, 0.00000010, -0.00000103, 0.00000427, -0.00001535, 0.00000501, -0.00001032, 0.00000824, -0.00001647, 0.00000707, -0.00000348, -0.00000694, 0.00000909, -0.00001354, 0.00001370, -0.00000781, -0.00000453, -0.00000167, 0.00000127, -0.00000210, -0.00000537, -0.00000120, -0.00000200, 0.00000011, -0.00001747, -0.00000893, 0.00000038, -0.00000585, -0.00003196, -0.00000507, -0.00002158, -0.00000446, -0.00008241, -0.00001455, -0.00029346, 0.00008868, 0.00033259, 0.00004158, 0.00017386, 0.00003670, 0.00032682, 0.00011286, -0.00030760, -0.00002500, -0.00005783, -0.00001516, -0.00003018, -0.00000744, -0.00001383, -0.00000134, -0.00002002, -0.00000075, -0.00000156, -0.00000246, -0.00000907, -0.00000275, -0.00000522, -0.00000095, -0.00001389, -0.00000032, 0.00000182, -0.00000255, 0.00000177, -0.00000128, -0.00000399, -0.00000276, -0.00000348, -0.00000600, -0.00000013, -0.00000858, -0.00000917, 0.00000868, -0.00000712, 0.00002300, -0.00001171, 0.00001512, -0.00000834, 0.00001128, -0.00000866, 0.00000469, 0.00000328, 0.00000221, 0.00000255, 0.00000660, -0.00000769, 0.00000145, },
		{ 0.00000707, -0.00000317, 0.00000888, 0.00000010, 0.00000804, 0.00000051, 0.00000657, -0.00000931, 0.00001849, -0.00000266, 0.00002603, -0.00002004, 0.00001531, -0.00000997, -0.00000113, -0.00000065, -0.00001911, 0.00001948, -0.00001968, 0.00000622, -0.00001066, 0.00000202, -0.00001036, 0.00001289, -0.00000860, -0.00000515, -0.00000538, 0.00000490, -0.00000400, -0.00000541, -0.00000313, 0.00000169, -0.00000276, -0.00000166, -0.00000347, -0.00000054, -0.00000344, 0.00000765, -0.00000583, 0.00000175, -0.00000468, -0.00000610, -0.00000254, -0.00000005, -0.00000411, 0.00000549, -0.00000553, -0.00000775, 0.00000007, -0.00000492, 0.00000235, -0.00001368, 0.00000989, -0.00000802, -0.00000410, 0.00000709, -0.00001317, 0.00001867, -0.00001833, 0.00000572, -0.00000878, 0.00001009, -0.00001593, 0.00000792, -0.00001192, 0.00001087, -0.00001822, -0.00000369, -0.00001925, 0.00001879, -0.00003950, -0.00000048, -0.00003304, 0.00001367, -0.00006671, -0.00001866, -0.00014759, -0.00020085, 0.00031172, 0.00015437, 0.00008226, 0.00012460, 0.00007369, 0.00034132, -0.00003539, -0.00026196, 0.00001127, -0.00007527, 0.00000219, -0.00004045, 0.00000693, -0.00002284, 0.00000460, -0.00002728, 0.00001264, -0.00001198, 0.00000360, -0.00001506, 0.00000393, -0.00001067, 0.00000230, -0.00001844, 0.00000932, -0.00000610, 0.00000252, -0.00000214, 0.00000154, -0.00000736, -0.00000002, -0.00000638, -0.00000335, -0.00000168, -0.00001166, -0.00000346, 0.00000497, -0.00000104, 0.00001940, -0.00001152, 0.00001711, -0.00001118, 0.00001512, -0.00001460, 0.00001296, -0.00000388, 0.00000800, -0.00000016, 0.00000751, -0.00001138, },
		{ -0.00000748, 0.00000484, 0.00000079, 0.00000660, 0.00000263, 0.00000595, 0.00000356, -0.00000159, 0.00000310, 0.00001027, 0.00000948, 0.00001320, -0.00001685, 0.00001659, -0.00002367, 0.00002221, -0.00002950, 0.00000764, 0.00000110, -0.00001039, 0.00000133, -0.00000827, -0.00000014, -0.00000518, 0.00000978, -0.00000976, -0.00000550, -0.00000375, 0.00000553, -0.00000633, -0.00000426, -0.00000256, 0.00000103, -0.00000278, -0.00000217, -0.00000279, -0.00000083, -0.00000148, 0.00000637, -0.00000753, 0.00000390, -0.00000741, -0.00000449, -0.00000311, 0.00000067, -0.00000396, 0.00000643, -0.00001071, -0.00000244, -0.00000301, -0.00000204, -0.00000191, -0.00001033, 0.00001264, -0.00001625, 0.00000472, -0.00000159, -0.00000362, 0.00001171, -0.00001716, 0.00000652, -0.00000594, 0.00000515, -0.00001546, 0.00001010, -0.00001148, 0.00000911, -0.00002230, -0.00000184, -0.00001276, 0.00001052, -0.00004562, 0.00001012, -0.00003899, 0.00001524, -0.00009253, 0.00000460, -0.00028645, 0.00008874, 0.00027218, 0.00007868, 0.00009980, 0.00009386, 0.00011793, 0.00031923, -0.00017013, -0.00016946, -0.00000195, -0.00006315, -0.00000258, -0.00003077, 0.00000323, -0.00001803, -0.00000325, -0.00001795, 0.00001040, -0.00001141, 0.00000040, -0.00001123, 0.00000099, -0.00000762, -0.00000298, -0.00001244, 0.00000880, -0.00000717, 0.00000223, -0.00000050, -0.00000138, -0.00000539, -0.00000217, -0.00000443, -0.00000507, -0.00000038, -0.00001690, 0.00000599, -0.00000362, 0.00001102, 0.00000788, -0.00000213, 0.00000963, -0.00000449, 0.00000897, -0.00001021, 0.00001175, -0.00000324, 0.00000808, 0.00000166, 0.00000254, },
		{ -0.00000326, -0.00000215, 0.00000154, 0.00000644, 0.00000171, 0.00000764, 0.00000241, 0.00000464, -0.00000470, 0.00001279, 0.00000250, 0.00002336, -0.00000993, 0.00000457, -0.00000992, -0.00000006, 0.00000861, -0.00003096, 0.00002010, -0.00001357, 0.00000068, -0.00000704, -0.00000276, -0.00000286, -0.00000056, 0.00000577, -0.00000974, -0.00000707, 0.00000060, 0.00000294, -0.00000628, -0.00000457, -0.00000103, -0.00000079, -0.00000124, -0.00000463, 0.00000060, -0.00000385, 0.00000323, 0.00000084, -0.00000427, 0.00000190, -0.00000672, -0.00000652, 0.00000060, -0.00000283, 0.00000074, 0.00000110, -0.00000911, -0.00000212, -0.00000235, -0.00000325, -0.00000233, -0.00000802, 0.00001420, -0.00002555, 0.00001790, -0.00001582, 0.00001334, -0.00000602, -0.00000349, -0.00000346, 0.00000641, -0.00001307, 0.00000247, -0.00000431, 0.00000429, -0.00001144, -0.00000551, -0.00001797, 0.00001271, -0.00002846, -0.00001016, -0.00002186, -0.00000020, -0.00004693, -0.00003014, -0.00011916, -0.00020508, 0.00028793, 0.00012967, 0.00009211, 0.00007954, 0.00009254, 0.00015552, 0.00026070, -0.00026432, -0.00006474, -0.00004723, -0.00002648, -0.00002595, -0.00000490, -0.00001622, -0.00000150, -0.00002047, 0.00000318, -0.00000567, 0.00000033, -0.00001138, 0.00000070, -0.00000991, 0.00000225, -0.00001413, 0.00000123, -0.00000091, -0.00000110, -0.00000289, 0.00000522, -0.00000849, 0.00000045, -0.00000737, 0.00000009, -0.00000897, 0.00000147, -0.00002081, 0.00001439, -0.00001179, 0.00002337, -0.00000547, 0.00001040, -0.00000153, 0.00000603, -0.00000159, -0.00000047, 0.00000486, 0.00000269, 0.00000440, 0.00000487, },
		{ 0.00000455, -0.00000447, -0.00000145, 0.00000426, 0.00000544, 0.00000269, 0.00000772, 0.00000271, -0.00000005, 0.00000371, 0.00000815, 0.00001192, 0.00001470, -0.00001138, 0.00000610, -0.00002234, 0.00002475, -0.00002170, -0.00000791, 0.00000945, -0.00001133, 0.00000094, -0.00000921, 0.00000006, -0.00000320, 0.00000062, 0.00000475, -0.00001342, -0.00000387, 0.00000162, 0.00000203, -0.00000794, -0.00000265, -0.00000229, 0.00000063, -0.00000336, -0.00000306, 0.00000052, -0.00000367, 0.00000444, -0.00000221, -0.00000222, 0.00000140, -0.00000893, -0.00000481, 0.00000110, -0.00000322, 0.00000194, -0.00000189, -0.00000848, 0.00000004, -0.00000525, -0.00000037, -0.00000675, 0.00000097, 0.00000413, -0.00002169, 0.00001987, -0.00001970, 0.00002039, -0.00001608, 0.00000481, -0.00000674, 0.00000788, -0.00002007, 0.00001297, -0.00001178, 0.00000988, -0.00002202, 0.00000293, -0.00002087, 0.00001676, -0.00004715, 0.00001273, -0.00004104, 0.00001619, -0.00008603, 0.00000813, -0.00026431, 0.00006885, 0.00024642, 0.00007780, 0.00008068, 0.00008727, 0.00007315, 0.00021498, 0.00014082, -0.00027304, -0.00000811, -0.00007252, -0.00000423, -0.00003546, 0.00000735, -0.00002515, 0.00000394, -0.00002412, 0.00001215, -0.00001374, 0.00000497, -0.00001511, 0.00000446, -0.00001327, 0.00000456, -0.00001660, 0.00000744, -0.00000617, 0.00000096, -0.00000259, 0.00000458, -0.00000964, 0.00000085, -0.00000680, -0.00000155, -0.00000671, -0.00000385, -0.00001540, 0.00001273, -0.00000980, 0.00002601, -0.00001050, 0.00001602, -0.00000670, 0.00001085, -0.00000691, 0.00000501, 0.00000136, 0.00000620, 0.00000277, },
		{ 0.00000251, 0.00000317, -0.00000470, -0.00000079, 0.00000799, 0.00000187, 0.00000653, 0.00000571, 0.00000191, -0.00000077, 0.00001032, 0.00000341, 0.00002385, -0.00000512, 0.00000438, -0.00001759, 0.00000034, 0.00001534, -0.00002942, 0.00000688, -0.00000229, -0.00000466, -0.00000346, -0.00000786, 0.00000252, -0.00000535, 0.00000385, 0.00000093, -0.00001532, 0.00000040, 0.00000123, 0.00000113, -0.00000930, -0.00000082, -0.00000360, 0.00000246, -0.00000670, 0.00000104, -0.00000242, -0.00000078, 0.00000234, -0.00000229, -0.00000163, 0.00000157, -0.00001300, 0.00000058, -0.00000224, -0.00000003, -0.00000110, -0.00000099, -0.00001014, 0.00000401, -0.00001103, 0.00000626, -0.00001424, 0.00001383, -0.00001507, -0.00000388, 0.00000727, -0.00000920, 0.00001254, -0.00001211, 0.00000179, 0.00000040, -0.00000544, -0.00000802, 0.00000576, -0.00000328, -0.00000365, -0.00001133, -0.00000790, -0.00000250, -0.00001177, -0.00002382, -0.00000633, -0.00001682, -0.00002791, -0.00004290, -0.00008740, -0.00021599, 0.00027437, 0.00010266, 0.00010620, 0.00003812, 0.00011862, 0.00004430, 0.00027531, -0.00001158, -0.00020972, -0.00000896, -0.00006240, -0.00000771, -0.00002328, -0.00000120, -0.00001741, -0.00000492, -0.00001174, 0.00000420, -0.00000804, -0.00000174, -0.00000796, -0.00000255, -0.00000663, -0.00000229, -0.00000896, 0.00000332, -0.00000378, -0.00000256, 0.00000338, -0.00000263, -0.00000445, -0.00000403, -0.00000141, -0.00000809, 0.00000015, -0.00001406, -0.00000323, 0.00000286, 0.00000157, 0.00001886, -0.00000664, 0.00001408, -0.00000549, 0.00000991, -0.00000692, 0.00000590, 0.00000201, 0.00000661, },
		{ 0.00000726, 0.00000080, 0.00000409, -0.00000794, 0.00000527, 0.00000559, 0.00000324, 0.00000674, 0.00000607, -0.00000312, 0.00000833, 0.00000431, 0.00001449, 0.00001503, -0.00000623, 0.00000271, -0.00002855, 0.00002613, -0.00001262, -0.00001330, 0.00000209, -0.00000146, -0.00000517, -0.00000461, -0.00000667, 0.00000526, -0.00000970, 0.00001067, -0.00000898, -0.00001013, 0.00000025, 0.00000356, -0.00000321, -0.00000691, -0.00000263, -0.00000088, -0.00000026, -0.00000545, 0.00000176, -0.00000296, -0.00000030, 0.00000205, -0.00000387, 0.00000178, -0.00000321, -0.00001175, 0.00000240, -0.00000276, -0.00000018, -0.00000105, -0.00000310, -0.00000732, 0.00000179, -0.00001008, 0.00000648, -0.00001390, 0.00001658, -0.00002643, 0.00001145, -0.00000539, 0.00000181, 0.00000350, -0.00000697, 0.00000035, 0.00000175, -0.00001338, 0.00000308, -0.00000101, 0.00000134, -0.00001346, -0.00000259, -0.00001297, 0.00000386, -0.00003079, -0.00000188, -0.00002426, -0.00000193, -0.00006339, -0.00000683, -0.00022089, 0.00002410, 0.00025457, 0.00004339, 0.00010450, 0.00003408, 0.00011426, 0.00005848, 0.00027894, -0.00012859, -0.00013573, -0.00002658, -0.00004415, -0.00001386, -0.00001155, -0.00001065, -0.00001046, -0.00001006, -0.00000244, -0.00000348, -0.00000277, -0.00000665, -0.00000308, -0.00000762, -0.00000180, -0.00000652, -0.00000331, -0.00000040, -0.00000259, -0.00000288, 0.00000548, -0.00000692, -0.00000167, -0.00000574, -0.00000002, -0.00001046, 0.00000245, -0.00001943, 0.00000531, -0.00000494, 0.00001307, 0.00000972, -0.00000043, 0.00001021, -0.00000320, 0.00000853, -0.00000699, 0.00000751, 0.00000215, },
		{ 0.00000211, 0.00000679, 0.00000059, 0.00000315, -0.00000890, 0.00001124, 0.00000104, 0.00000697, 0.00000571, 0.00000426, -0.00000314, 0.00001471, -0.00000102, 0.00002806, -0.00000533, 0.00000910, -0.00002309, -0.00000505, 0.00001977, -0.00002169, -0.00000282, -0.00000358, 0.00000181, -0.00000800, -0.00000487, -0.00000277, 0.00000283, -0.00000885, 0.00001297, -0.00001759, -0.00000260, -0.00000175, 0.00000612, -0.00000826, -0.00000347, -0.00000516, 0.00000291, -0.00000510, -0.00000087, -0.00000067, -0.00000133, -0.00000181, 0.00000336, -0.00000612, 0.00000612, -0.00001131, -0.00000512, -0.00000043, 0.00000023, -0.00000421, 0.00000307, -0.00000900, -0.00000005, -0.00000668, -0.00000116, -0.00000152, -0.00000424, 0.00000592, -0.00002256, 0.00001483, -0.00000944, 0.00000551, 0.00000075, -0.00000686, 0.00000464, -0.00000755, -0.00000683, 0.00000185, 0.00000250, -0.00000759, -0.00000749, -0.00000776, -0.00000258, -0.00001502, -0.00001688, -0.00001157, -0.00001013, -0.00003425, -0.00003257, -0.00008138, -0.00019714, 0.00022860, 0.00011820, 0.00007400, 0.00005707, 0.00007347, 0.00007002, 0.00011886, 0.00021543, -0.00018334, -0.00008462, -0.00003596, -0.00003358, -0.00001156, -0.00001041, -0.00001204, -0.00001027, -0.00000628, -0.00000327, -0.00000326, -0.00000395, -0.00000509, -0.00000496, -0.00000633, -0.00000240, -0.00000492, -0.00000368, 0.00000017, -0.00000496, 0.00000155, 0.00000163, -0.00000584, -0.00000317, -0.00000308, -0.00000385, -0.00000707, -0.00000160, -0.00001771, 0.00000692, -0.00000631, 0.00002025, 0.00000212, 0.00000588, 0.00000536, 0.00000027, 0.00000602, -0.00000628, 0.00000935, },
		{ 0.00000745, 0.00000467, 0.00000297, 0.00000462, -0.00000335, -0.00000196, 0.00000961, 0.00000194, 0.00000692, 0.00000741, -0.00000261, 0.00000772, 0.00000732, 0.00001063, 0.00002051, -0.00000800, 0.00000808, -0.00003738, 0.00002382, -0.00000633, -0.00000846, -0.00000952, 0.00000251, -0.00000280, -0.00000638, -0.00000697, 0.00000404, -0.00000548, 0.00000005, 0.00000530, -0.00001729, 0.00000012, -0.00000035, 0.00000431, -0.00000941, -0.00000335, -0.00000372, 0.00000223, -0.00000587, 0.00000114, -0.00000138, -0.00000212, -0.00000050, 0.00000159, -0.00000402, 0.00000473, -0.00001477, -0.00000049, -0.00000163, 0.00000024, -0.00000485, 0.00000392, -0.00001116, 0.00000298, -0.00001142, 0.00000576, -0.00000794, 0.00000342, -0.00000597, -0.00001317, 0.00001299, -0.00001112, 0.00000899, -0.00000333, -0.00000298, 0.00000223, -0.00001160, -0.00000088, 0.00000124, 0.00000164, -0.00001257, -0.00000326, -0.00000799, -0.00000286, -0.00002557, -0.00000374, -0.00002048, -0.00000457, -0.00005916, -0.00000585, -0.00019499, 0.00000192, 0.00023898, 0.00004134, 0.00009264, 0.00002973, 0.00009820, 0.00003747, 0.00017710, 0.00012550, -0.00019907, -0.00004696, -0.00004490, -0.00002066, -0.00001275, -0.00000863, -0.00001511, -0.00000585, -0.00000615, -0.00000299, -0.00000410, -0.00000321, -0.00000571, -0.00000520, -0.00000626, -0.00000082, -0.00000562, -0.00000271, -0.00000119, -0.00000428, 0.00000338, -0.00000120, -0.00000542, -0.00000310, -0.00000243, -0.00000631, -0.00000418, -0.00000610, -0.00001438, 0.00000611, -0.00000401, 0.00002425, -0.00000454, 0.00001316, -0.00000202, 0.00000741, -0.00000107, -0.00000059, },
		{ 0.00000751, 0.00000351, 0.00000732, -0.00000012, 0.00000763, -0.00000924, 0.00000721, 0.00000401, 0.00000613, 0.00000533, 0.00000714, -0.00000515, 0.00001718, -0.00000129, 0.00002746, -0.00000156, 0.00000767, -0.00001835, -0.00001671, 0.00002227, -0.00001589, -0.00000065, -0.00001561, 0.00001013, -0.00000987, -0.00000360, -0.00000672, 0.00000686, -0.00001117, 0.00000862, -0.00000592, -0.00001107, -0.00000101, 0.00000319, -0.00000042, -0.00000795, -0.00000486, -0.00000023, -0.00000154, -0.00000277, -0.00000001, -0.00000023, -0.00000490, 0.00000332, -0.00000272, 0.00000144, -0.00000231, -0.00001168, -0.00000036, 0.00000037, -0.00000391, -0.00000024, -0.00000078, -0.00000744, -0.00000087, -0.00000832, 0.00000581, -0.00000840, 0.00000367, -0.00001127, -0.00000586, 0.00001081, -0.00001341, 0.00001487, -0.00001097, 0.00000598, -0.00000971, -0.00000324, -0.00000458, 0.00000844, -0.00001063, -0.00000410, -0.00001016, 0.00000382, -0.00002418, -0.00000783, -0.00001737, -0.00000252, -0.00004000, -0.00002580, -0.00007545, -0.00017642, 0.00018511, 0.00013189, 0.00004896, 0.00006935, 0.00004206, 0.00008727, 0.00003968, 0.00020651, 0.00003543, -0.00019141, -0.00001825, -0.00005480, -0.00000402, -0.00002034, -0.00000316, -0.00002076, 0.00000352, -0.00001261, 0.00000229, -0.00000881, 0.00000132, -0.00001045, -0.00000172, -0.00000875, 0.00000376, -0.00001029, 0.00000152, -0.00000551, 0.00000030, 0.00000098, -0.00000114, -0.00000682, -0.00000086, -0.00000471, -0.00000587, -0.00000351, -0.00000945, -0.00001112, 0.00000453, 0.00000151, 0.00002348, -0.00000714, 0.00001735, -0.00000742, 0.00001376, -0.00000890, },
		{ -0.00000996, 0.00001140, 0.00000232, 0.00000620, 0.00000188, 0.00000406, -0.00000697, 0.00000962, 0.00000279, 0.00000709, 0.00000670, 0.00000103, 0.00000395, 0.00001261, 0.00000670, 0.00002479, -0.00000921, 0.00001076, -0.00003741, 0.00001357, -0.00000099, -0.00000320, -0.00001081, -0.00000751, 0.00000761, -0.00001129, -0.00000380, -0.00000200, 0.00000244, -0.00000860, 0.00000896, -0.00001170, -0.00000631, -0.00000091, 0.00000436, -0.00000368, -0.00000777, -0.00000383, 0.00000024, -0.00000279, -0.00000097, -0.00000015, -0.00000140, -0.00000477, 0.00000459, -0.00000421, 0.00000370, -0.00000792, -0.00000791, -0.00000003, 0.00000022, -0.00000581, 0.00000314, -0.00000484, -0.00000444, -0.00000449, -0.00000377, 0.00000441, -0.00000912, 0.00000329, -0.00001489, 0.00000295, 0.00000314, -0.00000857, 0.00001393, -0.00001178, 0.00000680, -0.00001536, 0.00000308, -0.00000560, 0.00000934, -0.00001758, 0.00000139, -0.00001067, 0.00000322, -0.00003516, 0.00000716, -0.00002847, 0.00000572, -0.00006764, 0.00000470, -0.00018286, -0.00000030, 0.00020435, 0.00006129, 0.00006023, 0.00005153, 0.00005475, 0.00007046, 0.00005665, 0.00021613, -0.00005715, -0.00015347, -0.00001260, -0.00005001, 0.00000092, -0.00002097, -0.00000517, -0.00001684, 0.00000436, -0.00001364, 0.00000294, -0.00000853, 0.00000075, -0.00001077, -0.00000219, -0.00000609, 0.00000292, -0.00001107, 0.00000233, -0.00000616, 0.00000202, 0.00000028, -0.00000315, -0.00000512, -0.00000179, -0.00000471, -0.00000694, -0.00000143, -0.00001496, -0.00000507, 0.00000034, 0.00001178, 0.00001534, -0.00000170, 0.00001334, -0.00000467, 0.00001221, },
		{ 0.00000413, -0.00000404, 0.00001027, 0.00000331, 0.00000381, 0.00000521, -0.00000172, -0.00000003, 0.00000669, 0.00000551, 0.00000592, 0.00000721, -0.00000273, 0.00001427, 0.00000400, 0.00002288, 0.00000494, 0.00000250, -0.00001105, -0.00002369, 0.00001747, -0.00001112, 0.00000553, -0.00002204, 0.00000550, -0.00000106, -0.00000884, -0.00000403, 0.00000177, -0.00000293, -0.00000309, 0.00000406, -0.00001200, -0.00000410, 0.00000062, 0.00000327, -0.00000565, -0.00000797, -0.00000148, -0.00000166, -0.00000109, -0.00000144, 0.00000107, -0.00000483, -0.00000124, 0.00000256, -0.00000225, 0.00000128, -0.00000915, -0.00000627, 0.00000202, -0.00000344, -0.00000297, 0.00000215, -0.00000540, -0.00000504, -0.00000415, -0.00000144, 0.00000372, -0.00001213, 0.00000559, -0.00001936, 0.00001361, -0.00000999, 0.00000479, 0.00000365, -0.00000332, -0.00000453, -0.00000685, -0.00000194, 0.00000341, -0.00000355, -0.00000924, -0.00000568, 0.00000248, -0.00002052, -0.00001549, -0.00000680, -0.00001082, -0.00002806, -0.00003768, -0.00005480, -0.00017277, 0.00016749, 0.00011702, 0.00005817, 0.00004565, 0.00005618, 0.00005164, 0.00006168, 0.00008033, 0.00020304, -0.00014045, -0.00009407, -0.00002892, -0.00002882, -0.00000909, -0.00001173, -0.00001520, -0.00000291, -0.00000612, -0.00000537, -0.00000383, -0.00000149, -0.00000667, -0.00000491, -0.00000752, 0.00000217, -0.00000490, -0.00000579, -0.00000173, -0.00000204, -0.00000052, 0.00000237, -0.00000737, -0.00000069, -0.00000580, -0.00000223, -0.00000927, 0.00000099, -0.00002115, 0.00000216, -0.00000369, 0.00002264, 0.00000416, 0.00000792, 0.00000446, 0.00000366, },
		{ 0.00000924, -0.00000262, 0.00000271, 0.00000811, 0.00000319, 0.00000438, 0.00000423, -0.00000226, 0.00000317, 0.00000593, 0.00000640, 0.00000683, 0.00000268, 0.00000375, 0.00001280, 0.00000914, 0.00002331, -0.00000551, 0.00000665, -0.00002986, 0.00000257, -0.00000064, 0.00000040, -0.00000423, -0.00001868, 0.00001002, -0.00000701, -0.00000692, -0.00000179, 0.00000093, -0.00000379, -0.00000137, 0.00000075, -0.00001173, -0.00000144, 0.00000070, 0.00000269, -0.00000874, -0.00000549, -0.00000190, -0.00000134, -0.00000064, -0.00000081, 0.00000001, -0.00000605, 0.00000138, 0.00000130, -0.00000157, -0.00000092, -0.00000983, -0.00000350, 0.00000147, -0.00000523, -0.00000042, 0.00000098, -0.00000681, -0.00000430, -0.00000434, 0.00000267, -0.00000191, -0.00000998, 0.00000352, -0.00001756, 0.00001678, -0.00001711, 0.00001500, -0.00000463, 0.00000141, -0.00001306, 0.00000238, -0.00000678, 0.00000914, -0.00001467, -0.00000056, -0.00000927, 0.00000629, -0.00003795, 0.00000662, -0.00002410, 0.00000377, -0.00006179, -0.00000136, -0.00015864, -0.00001288, 0.00018771, 0.00005918, 0.00005670, 0.00004175, 0.00005295, 0.00005482, 0.00004602, 0.00012363, 0.00014627, -0.00018227, -0.00004521, -0.00004324, -0.00001062, -0.00001818, -0.00000598, -0.00001897, 0.00000565, -0.00001434, 0.00000129, -0.00000818, 0.00000247, -0.00001149, -0.00000147, -0.00000896, 0.00000651, -0.00001101, -0.00000152, -0.00000403, 0.00000024, -0.00000132, 0.00000174, -0.00000800, 0.00000023, -0.00000683, -0.00000270, -0.00000772, -0.00000202, -0.00002191, 0.00000494, -0.00000157, 0.00002674, -0.00000206, 0.00001361, -0.00000104, },
		{ -0.00000119, 0.00000943, -0.00000460, 0.00000726, 0.00000579, 0.00000324, 0.00000553, 0.00000160, 0.00000075, 0.00000261, 0.00000791, 0.00000574, 0.00000698, -0.00000020, 0.00001226, 0.00000686, 0.00002268, 0.00000681, 0.00000143, -0.00001054, -0.00002050, 0.00000733, -0.00000881, 0.00000998, -0.00001809, -0.00000693, 0.00000673, -0.00000949, -0.00000429, -0.00000087, -0.00000011, -0.00000371, -0.00000149, -0.00000095, -0.00001122, 0.00000107, 0.00000013, 0.00000170, -0.00001129, -0.00000209, -0.00000411, 0.00000130, -0.00000211, 0.00000109, -0.00000339, -0.00000416, 0.00000192, 0.00000146, -0.00000317, -0.00000103, -0.00001145, 0.00000095, -0.00000260, -0.00000295, -0.00000075, 0.00000142, -0.00001036, -0.00000039, -0.00000661, 0.00000861, -0.00001285, -0.00000012, -0.00000616, -0.00000661, 0.00000845, -0.00001202, 0.00001495, -0.00000477, -0.00000440, -0.00000871, 0.00000109, -0.00000201, 0.00000182, -0.00001191, -0.00000301, 0.00000003, -0.00001268, -0.00002572, 0.00000212, -0.00001603, -0.00001902, -0.00004481, -0.00004270, -0.00016423, 0.00015078, 0.00010478, 0.00006245, 0.00003224, 0.00005928, 0.00003351, 0.00006826, 0.00002720, 0.00017639, 0.00005052, -0.00017196, -0.00002464, -0.00004147, -0.00000678, -0.00001750, -0.00000924, -0.00001131, 0.00000165, -0.00001247, 0.00000011, -0.00000529, -0.00000113, -0.00000935, -0.00000385, -0.00000351, 0.00000274, -0.00001077, -0.00000167, -0.00000205, -0.00000168, 0.00000149, -0.00000270, -0.00000421, -0.00000345, -0.00000367, -0.00000677, -0.00000229, -0.00001071, -0.00001633, 0.00000285, 0.00000676, 0.00002317, -0.00000169, 0.00001363, },
		{ 0.00001214, -0.00000021, 0.00000801, -0.00000412, 0.00001046, 0.00000224, 0.00000608, 0.00000307, 0.00000320, 0.00000056, 0.00000427, 0.00000801, 0.00000695, 0.00000268, 0.00000590, 0.00001078, 0.00001282, 0.00002266, -0.00000335, 0.00000340, -0.00002591, 0.00000093, -0.00000790, 0.00000299, 0.00000210, -0.00001979, 0.00000046, 0.00000260, -0.00001080, -0.00000004, -0.00000347, 0.00000233, -0.00000733, 0.00000150, -0.00000481, -0.00000784, 0.00000065, 0.00000157, -0.00000225, -0.00000990, -0.00000202, -0.00000353, 0.00000211, -0.00000217, 0.00000078, -0.00000510, -0.00000261, 0.00000335, -0.00000066, -0.00000279, -0.00000275, -0.00001037, 0.00000244, -0.00000494, -0.00000113, -0.00000024, -0.00000083, -0.00001057, 0.00000118, -0.00000538, 0.00000856, -0.00001926, 0.00000831, -0.00001514, 0.00000455, -0.00000227, -0.00000254, 0.00001138, -0.00000529, -0.00000980, -0.00000125, -0.00000278, 0.00000282, -0.00000721, -0.00000582, -0.00000478, 0.00000371, -0.00003053, -0.00000511, -0.00001140, -0.00000525, -0.00004803, -0.00001465, -0.00013196, -0.00002907, 0.00018240, 0.00004490, 0.00006653, 0.00002079, 0.00006549, 0.00002520, 0.00006752, 0.00003366, 0.00020040, -0.00004505, -0.00013398, -0.00002103, -0.00003282, -0.00000885, -0.00001490, -0.00001162, -0.00000351, -0.00000534, -0.00000763, -0.00000246, -0.00000203, -0.00000551, -0.00000646, -0.00000548, 0.00000152, -0.00000288, -0.00000877, -0.00000173, -0.00000089, -0.00000273, 0.00000289, -0.00000558, -0.00000175, -0.00000609, -0.00000162, -0.00000900, 0.00000077, -0.00001879, -0.00000969, 0.00000112, 0.00001558, 0.00001696, 0.00000099, },
		{ 0.00000449, 0.00001020, 0.00000058, 0.00000665, -0.00000212, 0.00001122, 0.00000016, 0.00000852, 0.00000012, 0.00000692, -0.00000233, 0.00000837, 0.00000665, 0.00000732, 0.00000004, 0.00001379, 0.00000549, 0.00002665, 0.00000619, 0.00000200, -0.00001297, -0.00001606, 0.00000304, -0.00001403, 0.00001371, -0.00001147, -0.00001272, 0.00000221, -0.00000122, -0.00000915, 0.00000140, -0.00000453, 0.00000361, -0.00001111, 0.00000583, -0.00000976, -0.00000293, -0.00000145, 0.00000369, -0.00000734, -0.00000614, -0.00000455, 0.00000010, 0.00000024, -0.00000039, -0.00000206, -0.00000414, -0.00000221, 0.00000557, -0.00000516, 0.00000067, -0.00000731, -0.00000574, -0.00000048, -0.00000309, -0.00000223, 0.00000254, -0.00000691, -0.00000529, -0.00000204, 0.00000105, -0.00000051, -0.00001491, 0.00000780, -0.00001625, 0.00000841, -0.00000716, 0.00000393, 0.00000955, -0.00001176, -0.00000577, -0.00000109, -0.00000039, -0.00000146, -0.00000679, -0.00000673, 0.00000273, -0.00001140, -0.00002437, -0.00000372, -0.00000760, -0.00002289, -0.00003734, -0.00004654, -0.00014394, 0.00012405, 0.00010958, 0.00004632, 0.00004340, 0.00003700, 0.00004735, 0.00003804, 0.00004669, 0.00007277, 0.00017539, -0.00010877, -0.00009410, -0.00002070, -0.00002683, -0.00000928, -0.00001544, -0.00000776, -0.00000315, -0.00000732, -0.00000600, -0.00000167, -0.00000297, -0.00000611, -0.00000653, -0.00000285, 0.00000146, -0.00000562, -0.00000814, 0.00000049, -0.00000287, -0.00000057, 0.00000080, -0.00000477, -0.00000300, -0.00000511, -0.00000298, -0.00000726, -0.00000152, -0.00002254, -0.00000498, 0.00000237, 0.00002183, 0.00001067, },
		{ 0.00000372, 0.00001036, 0.00000536, 0.00000380, 0.00000364, 0.00000239, 0.00000786, 0.00000275, 0.00000604, 0.00000267, 0.00000591, -0.00000175, 0.00001015, 0.00000726, 0.00000318, 0.00000627, 0.00001189, 0.00001211, 0.00002732, -0.00000549, 0.00000363, -0.00002751, 0.00000509, -0.00001398, -0.00000038, 0.00000749, -0.00001493, -0.00000798, 0.00000323, -0.00000580, -0.00000382, -0.00000235, -0.00000013, -0.00000156, -0.00000878, 0.00000591, -0.00001142, -0.00000054, -0.00000160, 0.00000296, -0.00000946, -0.00000460, -0.00000496, 0.00000253, -0.00000100, -0.00000011, -0.00000374, -0.00000360, -0.00000007, 0.00000447, -0.00000706, 0.00000265, -0.00001053, -0.00000215, -0.00000306, -0.00000109, -0.00000266, 0.00000302, -0.00001097, -0.00000019, -0.00000462, 0.00000558, -0.00001059, -0.00000575, 0.00000211, -0.00001284, 0.00000817, -0.00000869, 0.00001007, 0.00000360, -0.00001531, 0.00000011, -0.00000277, 0.00000152, -0.00000665, -0.00000433, -0.00000582, 0.00000446, -0.00002603, -0.00000976, -0.00001013, -0.00000258, -0.00004551, -0.00001481, -0.00012162, -0.00003114, 0.00016657, 0.00004593, 0.00005746, 0.00002399, 0.00005193, 0.00002924, 0.00005001, 0.00003053, 0.00011722, 0.00011823, -0.00013948, -0.00005828, -0.00002416, -0.00002028, -0.00001194, -0.00001287, -0.00000516, -0.00000442, -0.00000796, -0.00000411, -0.00000175, -0.00000384, -0.00000676, -0.00000531, -0.00000086, 0.00000009, -0.00000796, -0.00000588, 0.00000100, -0.00000396, 0.00000103, -0.00000099, -0.00000419, -0.00000420, -0.00000391, -0.00000439, -0.00000524, -0.00000620, -0.00002347, -0.00000078, 0.00000511, 0.00002614, },
		{ 0.00002701, -0.00000130, 0.00001617, -0.00000087, 0.00000882, -0.00000020, 0.00000720, 0.00000348, 0.00000682, 0.00000170, 0.00000865, 0.00000042, 0.00000303, 0.00000946, 0.00000761, 0.00000002, 0.00001498, 0.00000546, 0.00002787, 0.00001007, -0.00000174, -0.00001191, -0.00001777, 0.00000752, -0.00002253, 0.00001359, -0.00000570, -0.00001021, -0.00000704, 0.00000364, -0.00000817, 0.00000008, -0.00000587, 0.00000483, -0.00000966, -0.00000138, 0.00000130, -0.00000934, -0.00000095, 0.00000019, -0.00000021, -0.00000875, -0.00000559, -0.00000215, 0.00000214, -0.00000054, -0.00000209, -0.00000304, -0.00000433, 0.00000387, -0.00000065, -0.00000415, 0.00000052, -0.00001006, -0.00000209, -0.00000226, -0.00000182, -0.00000037, -0.00000103, -0.00000935, 0.00000060, -0.00000302, 0.00000330, -0.00001412, 0.00000058, -0.00000299, -0.00000948, 0.00000790, -0.00000969, 0.00001712, -0.00000920, -0.00000875, -0.00000099, 0.00000081, -0.00000424, -0.00000390, -0.00000777, 0.00000352, -0.00001094, -0.00001979, -0.00001058, -0.00000250, -0.00002244, -0.00003351, -0.00004576, -0.00013193, 0.00010546, 0.00011055, 0.00003625, 0.00004639, 0.00002733, 0.00004774, 0.00002668, 0.00004809, 0.00003488, 0.00014269, 0.00005226, -0.00014529, -0.00002839, -0.00003156, -0.00001155, -0.00001767, -0.00000509, -0.00000831, -0.00000246, -0.00001047, 0.00000023, -0.00000507, -0.00000209, -0.00000933, -0.00000096, -0.00000245, 0.00000018, -0.00001116, -0.00000108, -0.00000182, -0.00000218, 0.00000025, -0.00000085, -0.00000555, -0.00000347, -0.00000420, -0.00000432, -0.00000504, -0.00001116, -0.00002231, 0.00000296, 0.00000949, },
		{ 0.00001723, 0.00002149, 0.00000024, 0.00001634, -0.00000301, 0.00001132, -0.00000150, 0.00000868, 0.00000272, 0.00000685, 0.00000230, 0.00001043, -0.00000279, 0.00000713, 0.00000967, 0.00000389, 0.00000531, 0.00001449, 0.00001027, 0.00003211, -0.00000564, 0.00000124, -0.00002735, 0.00000476, -0.00001061, -0.00001033, 0.00001109, -0.00001013, -0.00000766, -0.00000526, 0.00000158, -0.00000620, -0.00000127, -0.00000385, 0.00000353, -0.00001293, 0.00000461, -0.00000365, -0.00000682, -0.00000129, 0.00000108, -0.00000262, -0.00000874, -0.00000531, 0.00000040, 0.00000143, -0.00000094, -0.00000324, -0.00000271, -0.00000345, 0.00000543, -0.00000500, -0.00000026, -0.00000290, -0.00000879, -0.00000237, -0.00000100, -0.00000257, 0.00000106, -0.00000518, -0.00000575, 0.00000008, -0.00000139, -0.00000138, -0.00001312, 0.00000431, -0.00000837, -0.00000395, 0.00000449, -0.00000559, 0.00001703, -0.00001857, -0.00000002, -0.00000382, 0.00000339, -0.00001000, -0.00000025, -0.00000857, 0.00000694, -0.00002582, -0.00000614, -0.00001730, 0.00000511, -0.00004673, -0.00001026, -0.00011642, -0.00003061, 0.00014901, 0.00005326, 0.00004200, 0.00003374, 0.00003419, 0.00003918, 0.00002777, 0.00004478, 0.00004880, 0.00014933, -0.00001436, -0.00012705, -0.00001330, -0.00003398, -0.00000797, -0.00001812, -0.00000085, -0.00001052, -0.00000186, -0.00001036, 0.00000188, -0.00000668, -0.00000209, -0.00000908, 0.00000135, -0.00000354, -0.00000136, -0.00001139, 0.00000171, -0.00000391, -0.00000086, 0.00000020, -0.00000198, -0.00000571, -0.00000353, -0.00000343, -0.00000543, -0.00000472, -0.00001743, -0.00001748, 0.00000437, },
		{ 0.00000487, 0.00002582, 0.00001196, 0.00000737, 0.00001031, 0.00000064, 0.00000974, -0.00000006, 0.00000765, 0.00000436, 0.00000420, 0.00000746, 0.00000645, -0.00000123, 0.00000909, 0.00000947, 0.00000056, 0.00001410, 0.00000830, 0.00002584, 0.00001708, -0.00000678, -0.00001066, -0.00002073, 0.00001115, -0.00002350, 0.00000549, -0.00000030, -0.00000612, -0.00000935, -0.00000227, -0.00000068, -0.00000369, -0.00000403, 0.00000061, -0.00000274, -0.00000980, 0.00000620, -0.00000673, -0.00000511, -0.00000096, 0.00000083, -0.00000408, -0.00000976, -0.00000263, 0.00000093, 0.00000153, -0.00000295, -0.00000234, -0.00000379, -0.00000031, 0.00000295, -0.00000541, 0.00000130, -0.00000520, -0.00000872, -0.00000059, -0.00000175, -0.00000132, -0.00000075, -0.00000582, -0.00000378, 0.00000084, -0.00000309, -0.00000342, -0.00001172, 0.00000718, -0.00001475, 0.00000487, -0.00000339, 0.00000502, 0.00000564, -0.00001576, 0.00000120, -0.00000107, -0.00000218, -0.00000714, -0.00000349, 0.00000005, -0.00000710, -0.00002021, -0.00000783, -0.00000933, -0.00001208, -0.00003827, -0.00003556, -0.00013057, 0.00009649, 0.00010288, 0.00003948, 0.00003415, 0.00003528, 0.00003070, 0.00003708, 0.00002496, 0.00004812, 0.00006175, 0.00014126, -0.00007476, -0.00009053, -0.00001539, -0.00002740, -0.00001199, -0.00001020, -0.00000505, -0.00000684, -0.00000594, -0.00000516, -0.00000182, -0.00000387, -0.00000586, -0.00000406, -0.00000129, -0.00000148, -0.00000596, -0.00000721, 0.00000027, -0.00000302, -0.00000162, 0.00000191, -0.00000536, -0.00000351, -0.00000525, -0.00000116, -0.00000835, -0.00000404, -0.00002404, -0.00001017, },
		{ -0.00000440, 0.00000871, 0.00002950, 0.00000521, 0.00001344, 0.00000371, 0.00000623, 0.00000659, 0.00000159, 0.00000752, 0.00000398, 0.00000480, 0.00001033, 0.00000233, 0.00000168, 0.00001109, 0.00000564, 0.00000473, 0.00001525, 0.00001138, 0.00003297, -0.00000092, -0.00000509, -0.00002432, -0.00000006, -0.00000339, -0.00001687, 0.00000742, -0.00000517, -0.00000349, -0.00001067, -0.00000029, -0.00000077, -0.00000439, -0.00000316, 0.00000124, -0.00000737, -0.00000475, 0.00000524, -0.00000892, -0.00000300, -0.00000147, 0.00000137, -0.00000703, -0.00000868, -0.00000044, 0.00000130, 0.00000046, -0.00000394, -0.00000175, -0.00000388, 0.00000175, -0.00000016, -0.00000395, 0.00000159, -0.00000788, -0.00000742, 0.00000078, -0.00000236, -0.00000080, -0.00000260, -0.00000530, -0.00000163, -0.00000030, -0.00000437, -0.00000495, -0.00000835, 0.00000584, -0.00001674, 0.00001066, -0.00000866, 0.00001303, -0.00000740, -0.00000712, -0.00000144, 0.00000230, -0.00000934, -0.00000167, -0.00000610, 0.00000524, -0.00002314, -0.00000513, -0.00001771, 0.00000219, -0.00003784, -0.00001477, -0.00010171, -0.00004058, 0.00014155, 0.00005172, 0.00003961, 0.00002767, 0.00003603, 0.00002848, 0.00003167, 0.00002921, 0.00004731, 0.00007970, 0.00011181, -0.00011320, -0.00005364, -0.00002344, -0.00001960, -0.00001503, -0.00000283, -0.00001065, -0.00000243, -0.00000962, -0.00000040, -0.00000570, -0.00000126, -0.00000840, 0.00000038, -0.00000454, 0.00000020, -0.00000949, -0.00000297, -0.00000181, -0.00000210, -0.00000142, 0.00000186, -0.00000729, -0.00000227, -0.00000540, -0.00000099, -0.00000975, -0.00000608, -0.00002716, },
		{ -0.00002562, -0.00000105, 0.00001631, 0.00002673, 0.00000427, 0.00001450, 0.00000093, 0.00001027, 0.00000340, 0.00000339, 0.00000767, 0.00000236, 0.00000903, 0.00000841, 0.00000133, 0.00000425, 0.00001192, 0.00000196, 0.00001275, 0.00001062, 0.00002533, 0.00002148, -0.00000751, -0.00001390, -0.00001987, 0.00000854, -0.00001656, -0.00000420, 0.00000006, -0.00000190, -0.00000560, -0.00000995, 0.00000150, -0.00000115, -0.00000565, -0.00000057, -0.00000188, -0.00000792, -0.00000077, 0.00000294, -0.00001009, -0.00000098, -0.00000222, 0.00000197, -0.00001098, -0.00000483, -0.00000024, 0.00000229, -0.00000223, -0.00000280, -0.00000258, -0.00000214, 0.00000104, -0.00000118, -0.00000292, 0.00000162, -0.00001158, -0.00000338, -0.00000032, -0.00000115, -0.00000282, -0.00000182, -0.00000595, 0.00000193, -0.00000531, -0.00000163, -0.00000902, -0.00000138, -0.00000168, -0.00001071, 0.00000869, -0.00000618, 0.00001107, -0.00001147, -0.00000290, -0.00000045, -0.00000110, -0.00000917, -0.00000182, -0.00000066, -0.00000528, -0.00002089, -0.00000325, -0.00001528, -0.00000702, -0.00003693, -0.00003100, -0.00012309, 0.00008093, 0.00010106, 0.00003913, 0.00002827, 0.00003466, 0.00002592, 0.00003448, 0.00001927, 0.00004420, 0.00003682, 0.00010542, 0.00005948, -0.00012064, -0.00003049, -0.00002728, -0.00001664, -0.00001132, -0.00000318, -0.00001062, -0.00000263, -0.00000842, -0.00000053, -0.00000547, -0.00000235, -0.00000631, 0.00000003, -0.00000449, -0.00000137, -0.00000912, -0.00000186, -0.00000183, -0.00000297, 0.00000107, -0.00000151, -0.00000582, -0.00000365, -0.00000274, -0.00000446, -0.00000814, -0.00001177, },
		{ -0.00001817, -0.00002216, 0.00000300, 0.00002342, 0.00002157, 0.00000604, 0.00001253, 0.00000153, 0.00001221, 0.00000009, 0.00000693, 0.00000502, 0.00000438, 0.00001123, 0.00000582, 0.00000152, 0.00000769, 0.00000880, 0.00000503, 0.00001452, 0.00001369, 0.00003319, 0.00000378, -0.00000964, -0.00002401, -0.00000254, -0.00000148, -0.00001380, -0.00000200, -0.00000264, 0.00000090, -0.00000886, -0.00000828, 0.00000424, -0.00000436, -0.00000356, -0.00000127, -0.00000375, -0.00000716, 0.00000269, -0.00000127, -0.00000868, -0.00000101, -0.00000094, -0.00000032, -0.00001222, -0.00000137, -0.00000017, 0.00000196, -0.00000394, -0.00000174, -0.00000281, -0.00000122, 0.00000017, -0.00000161, -0.00000134, -0.00000063, -0.00001307, 0.00000042, -0.00000166, -0.00000069, -0.00000432, -0.00000091, -0.00000540, 0.00000286, -0.00000937, 0.00000160, -0.00001251, 0.00000472, -0.00000987, -0.00000171, 0.00000338, -0.00000128, 0.00000449, -0.00000907, -0.00000235, 0.00000150, -0.00000704, -0.00000525, -0.00000260, 0.00000296, -0.00001960, -0.00000736, -0.00001175, -0.00000650, -0.00002705, -0.00001976, -0.00008654, -0.00004993, 0.00013291, 0.00005030, 0.00004067, 0.00001883, 0.00004005, 0.00001796, 0.00003633, 0.00001447, 0.00005396, 0.00003058, 0.00012448, -0.00000019, -0.00010579, -0.00002010, -0.00002790, -0.00001441, -0.00000628, -0.00000716, -0.00000750, -0.00000486, -0.00000543, -0.00000266, -0.00000377, -0.00000404, -0.00000311, -0.00000244, -0.00000310, -0.00000374, -0.00000742, -0.00000180, -0.00000174, -0.00000313, 0.00000264, -0.00000487, -0.00000381, -0.00000486, -0.00000057, -0.00000847, -0.00000626, },
		{ -0.00000642, -0.00002314, -0.00001825, 0.00000916, 0.00002746, 0.00001706, 0.00000758, 0.00001031, 0.00000399, 0.00001155, -0.00000152, 0.00001014, 0.00000124, 0.00001003, 0.00000905, 0.00000543, 0.00000189, 0.00001047, 0.00000510, 0.00001208, 0.00001099, 0.00002714, 0.00002358, -0.00000591, -0.00001829, -0.00001889, 0.00000517, -0.00001155, -0.00000552, -0.00000748, 0.00000190, -0.00000087, -0.00001103, -0.00000493, 0.00000514, -0.00000743, -0.00000030, -0.00000473, -0.00000280, -0.00000667, 0.00000534, -0.00000628, -0.00000546, -0.00000252, 0.00000170, -0.00000535, -0.00000944, -0.00000029, 0.00000082, -0.00000036, -0.00000335, -0.00000202, -0.00000165, -0.00000242, 0.00000115, -0.00000299, 0.00000134, -0.00000590, -0.00001018, 0.00000123, -0.00000119, -0.00000286, -0.00000269, -0.00000229, -0.00000195, -0.00000167, -0.00000757, 0.00000040, -0.00001130, 0.00000521, -0.00001269, 0.00000398, 0.00000024, -0.00000035, 0.00000180, -0.00000806, 0.00000013, -0.00000158, -0.00000788, -0.00000449, 0.00000224, -0.00000636, -0.00001902, -0.00000311, -0.00001270, -0.00001213, -0.00002789, -0.00003301, -0.00010874, 0.00006006, 0.00010319, 0.00003427, 0.00003114, 0.00002410, 0.00003236, 0.00002135, 0.00002707, 0.00002540, 0.00004661, 0.00004186, 0.00012161, -0.00004890, -0.00008277, -0.00001657, -0.00002801, -0.00000903, -0.00000597, -0.00000850, -0.00000604, -0.00000519, -0.00000451, -0.00000312, -0.00000362, -0.00000328, -0.00000263, -0.00000345, -0.00000296, -0.00000450, -0.00000692, -0.00000071, -0.00000313, -0.00000101, 0.00000111, -0.00000557, -0.00000356, -0.00000388, -0.00000143, -0.00001029, },
		{ -0.00001062, -0.00000803, -0.00002727, -0.00001249, 0.00001542, 0.00002965, 0.00001243, 0.00000949, 0.00000808, 0.00000773, 0.00000766, 0.00000100, 0.00000909, 0.00000178, 0.00001350, 0.00000634, 0.00000557, 0.00000377, 0.00000958, 0.00000712, 0.00001390, 0.00001441, 0.00003581, 0.00000574, -0.00001082, -0.00002670, -0.00000265, -0.00000337, -0.00000820, -0.00000655, -0.00000804, 0.00000602, -0.00000468, -0.00001207, 0.00000059, 0.00000171, -0.00000628, -0.00000091, -0.00000592, -0.00000215, -0.00000497, 0.00000519, -0.00000924, -0.00000317, -0.00000276, 0.00000233, -0.00000927, -0.00000551, -0.00000019, 0.00000106, -0.00000233, -0.00000231, -0.00000202, -0.00000141, -0.00000321, 0.00000216, -0.00000371, 0.00000246, -0.00001077, -0.00000549, 0.00000052, -0.00000085, -0.00000500, -0.00000024, -0.00000371, 0.00000069, -0.00000755, -0.00000233, -0.00000395, -0.00000709, 0.00000198, -0.00001114, 0.00000687, -0.00000273, -0.00000038, 0.00000169, -0.00000818, 0.00000254, -0.00000665, -0.00000578, -0.00000337, 0.00000456, -0.00001911, -0.00000770, -0.00000843, -0.00000850, -0.00002627, -0.00001570, -0.00007992, -0.00004925, 0.00011530, 0.00005575, 0.00003557, 0.00001999, 0.00003201, 0.00002206, 0.00002601, 0.00001872, 0.00003923, 0.00003158, 0.00006522, 0.00009715, -0.00008037, -0.00005836, -0.00001878, -0.00002442, -0.00000471, -0.00000786, -0.00000800, -0.00000537, -0.00000505, -0.00000420, -0.00000325, -0.00000315, -0.00000235, -0.00000337, -0.00000384, -0.00000277, -0.00000525, -0.00000610, -0.00000021, -0.00000387, 0.00000092, -0.00000134, -0.00000510, -0.00000366, -0.00000258, -0.00000389, },
		{ -0.00000613, -0.00001049, -0.00001077, -0.00003027, -0.00000467, 0.00002082, 0.00003039, 0.00000750, 0.00001231, 0.00000609, 0.00001099, 0.00000291, 0.00000571, 0.00000490, 0.00000739, 0.00001180, 0.00000628, 0.00000505, 0.00000613, 0.00000750, 0.00001306, 0.00001069, 0.00002835, 0.00002731, -0.00000645, -0.00001964, -0.00002121, 0.00000486, -0.00001200, -0.00000035, -0.00001474, -0.00000060, 0.00000503, -0.00000817, -0.00001068, 0.00000495, -0.00000306, -0.00000281, -0.00000513, -0.00000314, -0.00000364, -0.00000161, 0.00000200, -0.00000928, -0.00000261, -0.00000126, -0.00000006, -0.00000976, -0.00000298, 0.00000024, -0.00000034, -0.00000237, -0.00000227, -0.00000087, -0.00000342, -0.00000140, 0.00000117, -0.00000238, 0.00000011, -0.00001199, -0.00000226, 0.00000042, -0.00000257, -0.00000426, 0.00000014, -0.00000275, -0.00000114, -0.00000915, 0.00000100, -0.00000737, -0.00000368, -0.00000096, -0.00000851, 0.00000895, -0.00000841, 0.00000411, -0.00000164, -0.00000500, -0.00000016, -0.00000735, -0.00000605, 0.00000285, -0.00000527, -0.00001851, -0.00000349, -0.00000878, -0.00001564, -0.00002445, -0.00002942, -0.00009960, 0.00004700, 0.00009716, 0.00003547, 0.00003002, 0.00002012, 0.00003041, 0.00001935, 0.00002343, 0.00002148, 0.00004175, 0.00002431, 0.00008682, 0.00006023, -0.00009533, -0.00003630, -0.00002495, -0.00001554, -0.00000573, -0.00000750, -0.00000856, -0.00000343, -0.00000650, -0.00000248, -0.00000464, -0.00000066, -0.00000378, -0.00000300, -0.00000500, -0.00000133, -0.00000744, -0.00000358, -0.00000162, -0.00000246, 0.00000059, -0.00000244, -0.00000514, -0.00000274, -0.00000280, },
		{ -0.00000354, -0.00000851, -0.00000930, -0.00001593, -0.00002985, 0.00000297, 0.00002691, 0.00002742, 0.00000499, 0.00001425, 0.00000568, 0.00001136, 0.00000118, 0.00000783, 0.00000328, 0.00001245, 0.00000833, 0.00000758, 0.00000502, 0.00000626, 0.00000982, 0.00001466, 0.00001372, 0.00003914, 0.00000798, -0.00001288, -0.00002760, -0.00000515, -0.00000305, -0.00000781, -0.00000226, -0.00001764, 0.00000747, 0.00000067, -0.00001074, -0.00000611, 0.00000488, -0.00000455, -0.00000229, -0.00000768, 0.00000017, -0.00000549, 0.00000142, -0.00000198, -0.00000807, -0.00000217, -0.00000035, -0.00000285, -0.00000838, -0.00000144, 0.00000001, -0.00000162, -0.00000182, -0.00000205, -0.00000057, -0.00000531, 0.00000137, -0.00000068, -0.00000073, -0.00000368, -0.00001051, -0.00000035, -0.00000008, -0.00000441, -0.00000215, -0.00000034, -0.00000170, -0.00000511, -0.00000673, 0.00000135, -0.00000936, -0.00000087, -0.00000383, -0.00000397, 0.00000699, -0.00001201, 0.00001003, -0.00000681, -0.00000072, -0.00000532, -0.00000548, -0.00000536, 0.00000642, -0.00001868, -0.00000742, -0.00000797, -0.00000531, -0.00002950, -0.00001146, -0.00007311, -0.00004757, 0.00009936, 0.00005771, 0.00003257, 0.00002206, 0.00002406, 0.00002526, 0.00001865, 0.00002078, 0.00002819, 0.00003548, 0.00002789, 0.00009972, 0.00001721, -0.00009359, -0.00002266, -0.00002800, -0.00000732, -0.00000875, -0.00000628, -0.00000868, -0.00000230, -0.00000738, -0.00000139, -0.00000487, 0.00000112, -0.00000567, -0.00000272, -0.00000519, -0.00000082, -0.00000899, -0.00000131, -0.00000294, -0.00000090, -0.00000046, -0.00000315, -0.00000464, -0.00000232, },
		{ -0.00000361, -0.00000372, -0.00001116, -0.00000747, -0.00002318, -0.00002541, 0.00000976, 0.00003346, 0.00002018, 0.00000676, 0.00001394, 0.00000684, 0.00000927, 0.00000253, 0.00000662, 0.00000652, 0.00001279, 0.00000708, 0.00000839, 0.00000521, 0.00000580, 0.00001550, 0.00001138, 0.00002769, 0.00003300, -0.00000832, -0.00002022, -0.00002351, 0.00000389, -0.00001227, 0.00000180, -0.00001308, -0.00001058, 0.00001008, -0.00000425, -0.00001102, -0.00000119, 0.00000221, -0.00000337, -0.00000555, -0.00000572, 0.00000082, -0.00000581, 0.00000261, -0.00000511, -0.00000662, -0.00000139, -0.00000098, -0.00000400, -0.00000712, -0.00000008, -0.00000152, -0.00000135, -0.00000197, -0.00000085, -0.00000254, -0.00000462, 0.00000267, -0.00000165, -0.00000073, -0.00000616, -0.00000863, 0.00000137, -0.00000227, -0.00000397, -0.00000118, 0.00000026, -0.00000352, -0.00000674, -0.00000398, 0.00000051, -0.00001097, 0.00000261, -0.00000768, 0.00000315, -0.00000169, -0.00000730, 0.00001033, -0.00000832, -0.00000071, -0.00000695, -0.00000578, 0.00000074, -0.00000211, -0.00001958, -0.00000234, -0.00000842, -0.00001222, -0.00002864, -0.00002143, -0.00009405, 0.00003989, 0.00008671, 0.00003857, 0.00002664, 0.00002175, 0.00002243, 0.00002389, 0.00001431, 0.00002512, 0.00002866, 0.00002997, 0.00003774, 0.00010181, -0.00002593, -0.00007796, -0.00001957, -0.00002353, -0.00000542, -0.00000883, -0.00000717, -0.00000625, -0.00000414, -0.00000549, -0.00000276, -0.00000199, -0.00000055, -0.00000551, -0.00000405, -0.00000304, -0.00000305, -0.00000792, -0.00000109, -0.00000245, -0.00000107, -0.00000050, -0.00000451, -0.00000281, },
		{ -0.00000154, -0.00000513, -0.00000446, -0.00001241, -0.00000751, -0.00002963, -0.00001912, 0.00001792, 0.00003673, 0.00001233, 0.00001109, 0.00001221, 0.00000791, 0.00000749, 0.00000358, 0.00000631, 0.00001052, 0.00001037, 0.00000805, 0.00000876, 0.00000400, 0.00000925, 0.00001740, 0.00001334, 0.00004080, 0.00001275, -0.00001753, -0.00002615, -0.00000950, -0.00000138, -0.00000989, 0.00000273, -0.00002071, 0.00000020, 0.00000739, -0.00000819, -0.00000826, 0.00000079, 0.00000100, -0.00000395, -0.00000817, -0.00000173, -0.00000076, -0.00000488, 0.00000237, -0.00000747, -0.00000443, -0.00000173, -0.00000130, -0.00000455, -0.00000574, 0.00000018, -0.00000276, -0.00000057, -0.00000197, -0.00000031, -0.00000491, -0.00000216, 0.00000256, -0.00000231, -0.00000144, -0.00000755, -0.00000632, 0.00000179, -0.00000431, -0.00000239, -0.00000065, 0.00000025, -0.00000649, -0.00000585, -0.00000190, -0.00000178, -0.00001048, 0.00000443, -0.00000940, 0.00000775, -0.00001138, 0.00000294, 0.00000503, -0.00000655, -0.00000388, -0.00000586, -0.00000551, 0.00000538, -0.00001571, -0.00000902, -0.00000642, -0.00000441, -0.00002672, -0.00001564, -0.00006024, -0.00005029, 0.00009071, 0.00005233, 0.00003453, 0.00001980, 0.00002355, 0.00001977, 0.00002153, 0.00001290, 0.00002992, 0.00002516, 0.00002709, 0.00005228, 0.00008922, -0.00005954, -0.00005782, -0.00002082, -0.00001602, -0.00000753, -0.00000692, -0.00000873, -0.00000344, -0.00000667, -0.00000289, -0.00000421, 0.00000138, -0.00000390, -0.00000439, -0.00000530, -0.00000072, -0.00000616, -0.00000565, -0.00000158, -0.00000161, -0.00000167, -0.00000055, -0.00000516, },
		{ -0.00000391, -0.00000253, -0.00000525, -0.00000669, -0.00001143, -0.00001087, -0.00003330, -0.00001214, 0.00002817, 0.00003391, 0.00000810, 0.00001500, 0.00001012, 0.00000846, 0.00000670, 0.00000334, 0.00000921, 0.00001113, 0.00000905, 0.00000952, 0.00000835, 0.00000293, 0.00001587, 0.00001439, 0.00002629, 0.00003815, -0.00000836, -0.00002343, -0.00002286, -0.00000003, -0.00000980, -0.00000078, -0.00000649, -0.00001836, 0.00000746, 0.00000213, -0.00000955, -0.00000479, 0.00000075, 0.00000123, -0.00000748, -0.00000692, 0.00000085, -0.00000279, -0.00000351, 0.00000117, -0.00000882, -0.00000202, -0.00000336, -0.00000006, -0.00000589, -0.00000368, -0.00000127, -0.00000232, -0.00000055, -0.00000108, -0.00000182, -0.00000531, -0.00000003, 0.00000205, -0.00000358, -0.00000135, -0.00000911, -0.00000299, -0.00000022, -0.00000395, -0.00000186, 0.00000090, -0.00000264, -0.00000703, -0.00000508, 0.00000039, -0.00000626, -0.00000613, 0.00000218, -0.00000621, 0.00000467, -0.00001363, 0.00001038, -0.00000029, -0.00000609, -0.00000531, -0.00000601, -0.00000049, -0.00000082, -0.00001888, -0.00000224, -0.00000848, -0.00000835, -0.00003003, -0.00002007, -0.00008292, 0.00002978, 0.00008169, 0.00003496, 0.00002812, 0.00002005, 0.00002011, 0.00002143, 0.00001415, 0.00002001, 0.00002678, 0.00002485, 0.00002458, 0.00007022, 0.00006090, -0.00007672, -0.00004153, -0.00001979, -0.00001183, -0.00000834, -0.00000649, -0.00000811, -0.00000322, -0.00000686, -0.00000226, -0.00000301, 0.00000179, -0.00000585, -0.00000434, -0.00000448, -0.00000072, -0.00000763, -0.00000422, -0.00000113, -0.00000214, -0.00000106, -0.00000186, },
		{ -0.00000326, -0.00000228, -0.00000470, -0.00000466, -0.00000926, -0.00000980, -0.00001624, -0.00003484, -0.00000268, 0.00003721, 0.00002675, 0.00000843, 0.00001690, 0.00000797, 0.00000955, 0.00000503, 0.00000475, 0.00001230, 0.00000930, 0.00000981, 0.00001069, 0.00000622, 0.00000584, 0.00001942, 0.00001524, 0.00004078, 0.00001855, -0.00002212, -0.00002592, -0.00001181, -0.00000227, -0.00000935, 0.00000231, -0.00001562, -0.00000983, 0.00000921, -0.00000323, -0.00000776, -0.00000341, 0.00000181, -0.00000063, -0.00001031, -0.00000326, 0.00000124, -0.00000461, -0.00000173, -0.00000123, -0.00000813, -0.00000141, -0.00000398, 0.00000118, -0.00000731, -0.00000201, -0.00000299, -0.00000100, -0.00000069, -0.00000060, -0.00000396, -0.00000402, 0.00000143, 0.00000066, -0.00000421, -0.00000134, -0.00001003, -0.00000020, -0.00000286, -0.00000214, -0.00000180, 0.00000182, -0.00000666, -0.00000525, -0.00000474, 0.00000168, -0.00001041, -0.00000042, -0.00000172, -0.00000151, -0.00000309, -0.00000790, 0.00001188, -0.00000372, -0.00000724, -0.00000494, -0.00000550, 0.00000380, -0.00001290, -0.00001037, -0.00000456, -0.00000581, -0.00002018, -0.00002163, -0.00005010, -0.00004954, 0.00008154, 0.00004886, 0.00003266, 0.00002017, 0.00002265, 0.00001587, 0.00002208, 0.00000906, 0.00002836, 0.00001926, 0.00002706, 0.00002502, 0.00008462, 0.00002383, -0.00007927, -0.00002966, -0.00001706, -0.00001098, -0.00000723, -0.00000740, -0.00000615, -0.00000473, -0.00000537, -0.00000259, -0.00000045, -0.00000029, -0.00000632, -0.00000469, -0.00000271, -0.00000239, -0.00000770, -0.00000314, -0.00000068, -0.00000303, -0.00000010, },
		{ -0.00000028, -0.00000324, -0.00000232, -0.00000607, -0.00000475, -0.00001083, -0.00000921, -0.00002223, -0.00003460, 0.00001048, 0.00004106, 0.00001953, 0.00001154, 0.00001639, 0.00000665, 0.00001071, 0.00000250, 0.00000943, 0.00001189, 0.00000853, 0.00001136, 0.00001086, 0.00000394, 0.00001239, 0.00001815, 0.00002648, 0.00004144, -0.00000642, -0.00002830, -0.00002145, -0.00000385, -0.00000815, -0.00000252, -0.00000381, -0.00001694, -0.00000202, 0.00000708, -0.00000649, -0.00000489, -0.00000350, 0.00000390, -0.00000566, -0.00000923, -0.00000036, 0.00000041, -0.00000597, 0.00000014, -0.00000431, -0.00000574, -0.00000311, -0.00000188, 0.00000037, -0.00000751, -0.00000186, -0.00000331, -0.00000016, -0.00000024, -0.00000185, -0.00000458, -0.00000265, 0.00000266, -0.00000212, -0.00000275, -0.00000309, -0.00000892, 0.00000021, -0.00000353, -0.00000116, -0.00000071, -0.00000026, -0.00000810, -0.00000400, -0.00000335, 0.00000006, -0.00001098, 0.00000283, -0.00000350, -0.00000068, -0.00000799, -0.00000045, 0.00001032, -0.00000739, -0.00000688, -0.00000551, -0.00000092, -0.00000115, -0.00001692, -0.00000319, -0.00000731, -0.00000784, -0.00002523, -0.00002497, -0.00006962, 0.00001994, 0.00007894, 0.00002948, 0.00002934, 0.00001815, 0.00002056, 0.00001631, 0.00001683, 0.00001320, 0.00002814, 0.00001616, 0.00002664, 0.00003274, 0.00008747, -0.00001216, -0.00007271, -0.00001983, -0.00001620, -0.00000980, -0.00000670, -0.00000743, -0.00000529, -0.00000547, -0.00000429, -0.00000175, 0.00000068, -0.00000249, -0.00000668, -0.00000395, -0.00000205, -0.00000384, -0.00000761, -0.00000140, -0.00000166, -0.00000260, },
		{ -0.00000131, -0.00000110, -0.00000248, -0.00000384, -0.00000633, -0.00000552, -0.00001167, -0.00000965, -0.00002909, -0.00003051, 0.00002493, 0.00003917, 0.00001509, 0.00001517, 0.00001375, 0.00000745, 0.00000977, 0.00000237, 0.00001422, 0.00000869, 0.00001024, 0.00001247, 0.00000936, 0.00000493, 0.00001735, 0.00001926, 0.00004117, 0.00002322, -0.00002501, -0.00002791, -0.00001211, -0.00000476, -0.00000765, 0.00000014, -0.00001115, -0.00001221, 0.00000200, 0.00000343, -0.00000668, -0.00000364, -0.00000231, 0.00000369, -0.00001043, -0.00000526, 0.00000068, -0.00000112, -0.00000621, 0.00000098, -0.00000642, -0.00000386, -0.00000483, 0.00000153, -0.00000227, -0.00000679, -0.00000252, -0.00000248, 0.00000034, -0.00000013, -0.00000375, -0.00000375, -0.00000149, 0.00000284, -0.00000469, -0.00000029, -0.00000604, -0.00000656, -0.00000090, -0.00000239, -0.00000103, 0.00000030, -0.00000412, -0.00000659, -0.00000410, -0.00000131, -0.00000340, -0.00000841, 0.00000363, -0.00000400, -0.00000319, -0.00000764, 0.00000490, 0.00000727, -0.00001096, -0.00000529, -0.00000530, 0.00000307, -0.00001183, -0.00001013, -0.00000395, -0.00000640, -0.00001607, -0.00002162, -0.00004828, -0.00004287, 0.00006973, 0.00004982, 0.00002605, 0.00002318, 0.00001947, 0.00001664, 0.00001742, 0.00001139, 0.00002128, 0.00002082, 0.00001879, 0.00002348, 0.00004637, 0.00007686, -0.00004149, -0.00005977, -0.00001327, -0.00001671, -0.00000766, -0.00000707, -0.00000666, -0.00000546, -0.00000532, -0.00000334, -0.00000034, 0.00000021, -0.00000428, -0.00000660, -0.00000277, -0.00000236, -0.00000499, -0.00000695, 0.00000006, -0.00000323, },
		{ -0.00000364, -0.00000057, -0.00000131, -0.00000262, -0.00000537, -0.00000620, -0.00000645, -0.00001233, -0.00001104, -0.00003670, -0.00002054, 0.00003654, 0.00003414, 0.00001383, 0.00001744, 0.00001056, 0.00000969, 0.00000663, 0.00000670, 0.00001485, 0.00000658, 0.00001308, 0.00001247, 0.00000743, 0.00000945, 0.00001851, 0.00002959, 0.00004407, -0.00000461, -0.00003248, -0.00002167, -0.00000564, -0.00000874, -0.00000192, -0.00000509, -0.00001217, -0.00000751, 0.00000320, 0.00000024, -0.00000513, -0.00000396, 0.00000046, -0.00000041, -0.00001125, -0.00000165, 0.00000063, -0.00000322, -0.00000501, 0.00000026, -0.00000682, -0.00000390, -0.00000413, 0.00000356, -0.00000515, -0.00000612, -0.00000266, -0.00000159, 0.00000109, -0.00000153, -0.00000433, -0.00000293, -0.00000003, 0.00000104, -0.00000490, 0.00000077, -0.00000831, -0.00000473, -0.00000139, -0.00000129, -0.00000053, -0.00000093, -0.00000629, -0.00000478, -0.00000411, -0.00000043, -0.00000602, -0.00000524, 0.00000389, -0.00000628, -0.00000424, -0.00000546, 0.00000886, 0.00000171, -0.00001183, -0.00000481, -0.00000135, -0.00000117, -0.00001594, -0.00000325, -0.00000653, -0.00000811, -0.00002012, -0.00002698, -0.00006303, 0.00001667, 0.00007271, 0.00002870, 0.00002493, 0.00002006, 0.00001834, 0.00001571, 0.00001436, 0.00001282, 0.00002363, 0.00001497, 0.00002192, 0.00002204, 0.00006027, 0.00005565, -0.00006153, -0.00004249, -0.00001229, -0.00001553, -0.00000645, -0.00000688, -0.00000657, -0.00000522, -0.00000531, -0.00000138, -0.00000017, -0.00000042, -0.00000628, -0.00000515, -0.00000265, -0.00000241, -0.00000655, -0.00000485, -0.00000026, },
		{ -0.00000188, -0.00000280, -0.00000044, -0.00000115, -0.00000384, -0.00000630, -0.00000586, -0.00000786, -0.00001236, -0.00001463, -0.00004261, -0.00000572, 0.00004317, 0.00002871, 0.00001475, 0.00001732, 0.00000899, 0.00001071, 0.00000465, 0.00001269, 0.00001119, 0.00000801, 0.00001504, 0.00001127, 0.00000760, 0.00001371, 0.00002134, 0.00004449, 0.00002665, -0.00002741, -0.00003015, -0.00001303, -0.00000601, -0.00000778, 0.00000028, -0.00001166, -0.00000771, -0.00000609, 0.00000369, -0.00000164, -0.00000389, -0.00000375, 0.00000198, -0.00000564, -0.00000831, -0.00000005, 0.00000011, -0.00000522, -0.00000312, -0.00000127, -0.00000658, -0.00000466, -0.00000126, 0.00000316, -0.00000748, -0.00000542, -0.00000218, -0.00000075, 0.00000137, -0.00000364, -0.00000339, -0.00000253, 0.00000114, -0.00000166, -0.00000291, -0.00000029, -0.00000940, -0.00000350, -0.00000092, -0.00000059, -0.00000035, -0.00000373, -0.00000576, -0.00000418, -0.00000351, -0.00000079, -0.00000725, -0.00000191, 0.00000271, -0.00000929, -0.00000256, -0.00000362, 0.00001146, -0.00000535, -0.00001019, -0.00000406, 0.00000232, -0.00001106, -0.00000992, -0.00000339, -0.00000642, -0.00001431, -0.00001831, -0.00004793, -0.00003872, 0.00006212, 0.00004895, 0.00002216, 0.00002193, 0.00001931, 0.00001606, 0.00001524, 0.00001094, 0.00001816, 0.00001929, 0.00001432, 0.00002252, 0.00002485, 0.00006989, 0.00002717, -0.00006933, -0.00002620, -0.00001501, -0.00001232, -0.00000646, -0.00000624, -0.00000695, -0.00000472, -0.00000483, 0.00000064, -0.00000104, -0.00000115, -0.00000775, -0.00000317, -0.00000332, -0.00000241, -0.00000782, -0.00000227, },
		{ -0.00000096, -0.00000308, -0.00000181, -0.00000013, -0.00000178, -0.00000508, -0.00000692, -0.00000520, -0.00001010, -0.00001139, -0.00002149, -0.00004364, 0.00001059, 0.00004546, 0.00002420, 0.00001641, 0.00001542, 0.00000948, 0.00000934, 0.00000690, 0.00001533, 0.00000737, 0.00001184, 0.00001521, 0.00000996, 0.00001039, 0.00001557, 0.00003256, 0.00004910, -0.00000431, -0.00003593, -0.00002264, -0.00000719, -0.00000911, -0.00000211, -0.00000519, -0.00001179, -0.00000409, -0.00000636, 0.00000449, -0.00000270, -0.00000338, -0.00000240, 0.00000029, -0.00000814, -0.00000472, 0.00000022, -0.00000101, -0.00000626, -0.00000152, -0.00000278, -0.00000680, -0.00000410, 0.00000162, 0.00000106, -0.00000930, -0.00000399, -0.00000189, 0.00000048, 0.00000010, -0.00000457, -0.00000234, -0.00000201, 0.00000096, -0.00000316, -0.00000074, -0.00000246, -0.00000991, -0.00000176, -0.00000060, 0.00000037, -0.00000216, -0.00000519, -0.00000454, -0.00000410, -0.00000305, -0.00000119, -0.00000797, 0.00000209, -0.00000200, -0.00000889, -0.00000142, -0.00000074, 0.00001061, -0.00001070, -0.00000825, -0.00000051, -0.00000123, -0.00001556, -0.00000282, -0.00000629, -0.00000759, -0.00001773, -0.00002418, -0.00006116, 0.00001510, 0.00006731, 0.00002885, 0.00002011, 0.00002040, 0.00001719, 0.00001583, 0.00001171, 0.00001280, 0.00001962, 0.00001457, 0.00001615, 0.00002172, 0.00003143, 0.00007233, -0.00000404, -0.00006421, -0.00001640, -0.00001694, -0.00000935, -0.00000631, -0.00000638, -0.00000657, -0.00000488, -0.00000273, 0.00000085, -0.00000140, -0.00000284, -0.00000743, -0.00000235, -0.00000343, -0.00000333, -0.00000752, },
		{ -0.00000590, -0.00000121, -0.00000320, -0.00000105, 0.00000022, -0.00000336, -0.00000584, -0.00000717, -0.00000485, -0.00001242, -0.00001058, -0.00003053, -0.00003876, 0.00002582, 0.00004471, 0.00002093, 0.00001750, 0.00001347, 0.00001021, 0.00000785, 0.00001195, 0.00001295, 0.00000720, 0.00001516, 0.00001415, 0.00001012, 0.00001337, 0.00001910, 0.00004950, 0.00003153, -0.00003140, -0.00003148, -0.00001500, -0.00000640, -0.00000865, 0.00000097, -0.00001293, -0.00000529, -0.00000550, -0.00000511, 0.00000506, -0.00000368, -0.00000249, -0.00000181, -0.00000284, -0.00000718, -0.00000271, 0.00000018, -0.00000272, -0.00000605, -0.00000056, -0.00000448, -0.00000689, -0.00000200, 0.00000309, -0.00000200, -0.00001013, -0.00000203, -0.00000184, 0.00000157, -0.00000222, -0.00000379, -0.00000196, -0.00000143, -0.00000021, -0.00000288, 0.00000033, -0.00000512, -0.00000953, 0.00000035, -0.00000053, 0.00000086, -0.00000537, -0.00000419, -0.00000425, -0.00000395, -0.00000265, -0.00000175, -0.00000757, 0.00000477, -0.00000835, -0.00000441, -0.00000273, 0.00000375, 0.00000593, -0.00001316, -0.00000552, 0.00000260, -0.00001040, -0.00001027, -0.00000229, -0.00000694, -0.00001226, -0.00001685, -0.00004311, -0.00004029, 0.00005860, 0.00004660, 0.00002120, 0.00001766, 0.00002009, 0.00001477, 0.00001542, 0.00000835, 0.00001771, 0.00001577, 0.00001328, 0.00001770, 0.00002104, 0.00004073, 0.00006488, -0.00003072, -0.00005115, -0.00001350, -0.00001620, -0.00000768, -0.00000577, -0.00000716, -0.00000561, -0.00000502, 0.00000007, -0.00000065, -0.00000106, -0.00000508, -0.00000570, -0.00000284, -0.00000285, -0.00000494, },
		{ -0.00000602, -0.00000436, -0.00000185, -0.00000280, -0.00000008, -0.00000042, -0.00000486, -0.00000648, -0.00000663, -0.00000572, -0.00001366, -0.00001173, -0.00003919, -0.00002913, 0.00003929, 0.00004137, 0.00001905, 0.00001744, 0.00001262, 0.00000959, 0.00000949, 0.00001492, 0.00000909, 0.00001035, 0.00001653, 0.00001301, 0.00001227, 0.00001464, 0.00003106, 0.00005749, -0.00000402, -0.00004057, -0.00002273, -0.00000991, -0.00000825, -0.00000373, -0.00000356, -0.00001438, 0.00000087, -0.00001003, -0.00000123, 0.00000439, -0.00000437, -0.00000103, -0.00000367, -0.00000394, -0.00000539, -0.00000188, -0.00000023, -0.00000433, -0.00000506, -0.00000027, -0.00000668, -0.00000539, 0.00000006, 0.00000327, -0.00000579, -0.00000905, -0.00000070, -0.00000139, 0.00000134, -0.00000381, -0.00000247, -0.00000195, -0.00000147, -0.00000088, -0.00000199, 0.00000042, -0.00000823, -0.00000729, 0.00000155, 0.00000011, -0.00000085, -0.00000718, -0.00000246, -0.00000458, -0.00000397, -0.00000158, -0.00000336, -0.00000485, 0.00000309, -0.00001127, -0.00000000, -0.00000438, 0.00000773, -0.00000040, -0.00001346, -0.00000057, -0.00000079, -0.00001531, -0.00000270, -0.00000580, -0.00000718, -0.00001593, -0.00002169, -0.00005635, 0.00000853, 0.00006645, 0.00002689, 0.00001866, 0.00001656, 0.00001845, 0.00001434, 0.00001198, 0.00001005, 0.00001900, 0.00001158, 0.00001405, 0.00001866, 0.00002128, 0.00005076, 0.00004732, -0.00004677, -0.00003766, -0.00001362, -0.00001416, -0.00000649, -0.00000586, -0.00000734, -0.00000515, -0.00000373, 0.00000140, -0.00000206, -0.00000114, -0.00000643, -0.00000419, -0.00000339, -0.00000259, },
		{ -0.00000302, -0.00000617, -0.00000362, -0.00000209, -0.00000221, 0.00000088, -0.00000229, -0.00000558, -0.00000717, -0.00000554, -0.00000782, -0.00001372, -0.00001540, -0.00004584, -0.00001573, 0.00005054, 0.00003563, 0.00001859, 0.00001653, 0.00001237, 0.00000869, 0.00001403, 0.00001297, 0.00000846, 0.00001349, 0.00001686, 0.00001255, 0.00001517, 0.00001696, 0.00005099, 0.00004076, -0.00003711, -0.00003294, -0.00001666, -0.00000757, -0.00000848, 0.00000016, -0.00001203, -0.00000746, 0.00000058, -0.00001275, 0.00000368, 0.00000215, -0.00000376, -0.00000078, -0.00000646, -0.00000211, -0.00000515, -0.00000095, -0.00000147, -0.00000511, -0.00000380, -0.00000105, -0.00000838, -0.00000235, 0.00000096, 0.00000246, -0.00000955, -0.00000609, -0.00000055, -0.00000051, -0.00000021, -0.00000372, -0.00000172, -0.00000210, -0.00000194, -0.00000050, -0.00000143, -0.00000044, -0.00001096, -0.00000345, 0.00000151, 0.00000094, -0.00000457, -0.00000605, -0.00000186, -0.00000489, -0.00000393, -0.00000028, -0.00000527, -0.00000133, -0.00000220, -0.00000864, 0.00000074, -0.00000380, 0.00000907, -0.00000646, -0.00001103, 0.00000347, -0.00000973, -0.00001078, -0.00000116, -0.00000771, -0.00000972, -0.00001707, -0.00003673, -0.00004111, 0.00005179, 0.00004802, 0.00001877, 0.00001583, 0.00001697, 0.00001616, 0.00001404, 0.00000833, 0.00001503, 0.00001505, 0.00001055, 0.00001485, 0.00001915, 0.00002367, 0.00005753, 0.00002422, -0.00005191, -0.00002772, -0.00001372, -0.00001222, -0.00000532, -0.00000673, -0.00000666, -0.00000508, -0.00000123, 0.00000060, -0.00000239, -0.00000205, -0.00000648, -0.00000369, -0.00000333, },
		},
		{ { 0.00010199, 0.00003467, -0.00002911, -0.00001041, -0.00000394, 0.00000034, 0.00001421, 0.00000318, 0.00000157, 0.00001592, 0.00000479, 0.00001304, 0.00001670, 0.00002421, 0.00006297, 0.00000011, -0.00006015, -0.00001967, -0.00001542, -0.00001068, -0.00000949, -0.00000577, -0.00000120, -0.00000146, -0.00000926, -0.00000279, -0.00000828, -0.00000083, -0.00000333, 0.00000046, -0.00000601, 0.00000344, -0.00000638, -0.00000068, -0.00000311, -0.00000010, -0.00000519, -0.00000160, -0.00000369, 0.00000120, -0.00000468, 0.00000513, -0.00001528, 0.00000209, -0.00000391, 0.00000411, -0.00000477, 0.00000433, -0.00000092, 0.00000119, -0.00000431, 0.00000645, -0.00000485, 0.00000375, -0.00000155, 0.00000498, -0.00000635, 0.00000291, 0.00000373, 0.00000611, -0.00000539, 0.00000376, -0.00000702, 0.00000542, -0.00000521, 0.00000542, -0.00000702, 0.00000376, -0.00000539, 0.00000611, 0.00000373, 0.00000291, -0.00000635, 0.00000498, -0.00000155, 0.00000375, -0.00000485, 0.00000645, -0.00000431, 0.00000119, -0.00000092, 0.00000433, -0.00000477, 0.00000411, -0.00000391, 0.00000209, -0.00001528, 0.00000513, -0.00000468, 0.00000120, -0.00000369, -0.00000160, -0.00000519, -0.00000010, -0.00000311, -0.00000068, -0.00000638, 0.00000344, -0.00000601, 0.00000046, -0.00000333, -0.00000083, -0.00000828, -0.00000279, -0.00000926, -0.00000146, -0.00000120, -0.00000577, -0.00000949, -0.00001068, -0.00001542, -0.00001967, -0.00006015, 0.00000011, 0.00006297, 0.00002421, 0.00001670, 0.00001304, 0.00000479, 0.00001592, 0.00000157, 0.00000318, 0.00001421, 0.00000034, -0.00000394, -0.00001041, -0.00002911, 0.00003467, },
		{ 0.00003245, 0.00010210, 0.00003708, -0.00002576, -0.00001117, -0.00000250, -0.00000015, 0.00001615, 0.00000395, 0.00000219, 0.00001602, 0.00000590, 0.00001555, 0.00001666, 0.00003300, 0.00006322, -0.00002216, -0.00005822, -0.00001453, -0.00001601, -0.00000966, -0.00000936, -0.00000561, -0.00000117, -0.00000141, -0.00000903, -0.00000397, -0.00000735, -0.00000028, -0.00000301, -0.00000104, -0.00000585, 0.00000482, -0.00000752, -0.00000025, -0.00000354, 0.00000025, -0.00000574, -0.00000132, -0.00000369, 0.00000107, -0.00000234, 0.00000124, -0.00001512, 0.00000491, -0.00000562, 0.00000539, -0.00000653, 0.00000696, -0.00000227, 0.00000079, -0.00000348, 0.00000653, -0.00000561, 0.00000499, -0.00000262, 0.00000521, -0.00000737, 0.00000525, 0.00000352, 0.00000447, -0.00000471, 0.00000278, -0.00000623, 0.00000554, -0.00000517, 0.00000504, -0.00000747, 0.00000462, -0.00000612, 0.00000838, 0.00000254, 0.00000153, -0.00000557, 0.00000510, -0.00000087, 0.00000251, -0.00000377, 0.00000596, -0.00000511, 0.00000232, -0.00000066, 0.00000232, -0.00000315, 0.00000300, -0.00000223, -0.00000184, -0.00001314, 0.00000744, -0.00000698, 0.00000175, -0.00000403, -0.00000150, -0.00000486, -0.00000034, -0.00000279, -0.00000099, -0.00000524, 0.00000150, -0.00000518, 0.00000140, -0.00000358, -0.00000178, -0.00000858, -0.00000184, -0.00000957, -0.00000135, -0.00000139, -0.00000590, -0.00000969, -0.00001169, -0.00001475, -0.00002707, -0.00005517, 0.00002005, 0.00005706, 0.00001844, 0.00001645, 0.00001065, 0.00000414, 0.00001553, 0.00000116, 0.00000219, 0.00001250, 0.00000089, -0.00000550, -0.00000911, -0.00003316, },
		{ -0.00003778, 0.00003044, 0.00010246, 0.00003954, -0.00002308, -0.00001120, -0.00000147, -0.00000025, 0.00001798, 0.00000475, 0.00000285, 0.00001593, 0.00000760, 0.00001783, 0.00001709, 0.00004430, 0.00005555, -0.00004283, -0.00005079, -0.00001252, -0.00001559, -0.00000921, -0.00000889, -0.00000573, -0.00000103, -0.00000130, -0.00000898, -0.00000516, -0.00000593, -0.00000015, -0.00000266, -0.00000295, -0.00000464, 0.00000529, -0.00000825, -0.00000001, -0.00000383, 0.00000047, -0.00000628, -0.00000081, -0.00000407, 0.00000163, -0.00000068, -0.00000348, -0.00001266, 0.00000588, -0.00000644, 0.00000611, -0.00000778, 0.00000959, -0.00000438, 0.00000121, -0.00000288, 0.00000634, -0.00000594, 0.00000587, -0.00000368, 0.00000545, -0.00000820, 0.00000809, 0.00000197, 0.00000380, -0.00000445, 0.00000202, -0.00000529, 0.00000546, -0.00000509, 0.00000445, -0.00000749, 0.00000512, -0.00000645, 0.00001069, 0.00000027, 0.00000124, -0.00000526, 0.00000577, -0.00000093, 0.00000166, -0.00000264, 0.00000508, -0.00000562, 0.00000388, -0.00000148, 0.00000123, -0.00000206, 0.00000258, -0.00000149, -0.00000573, -0.00000930, 0.00000793, -0.00000868, 0.00000234, -0.00000456, -0.00000105, -0.00000488, -0.00000029, -0.00000279, -0.00000090, -0.00000455, -0.00000045, -0.00000368, 0.00000182, -0.00000384, -0.00000291, -0.00000832, -0.00000124, -0.00000985, -0.00000107, -0.00000189, -0.00000569, -0.00001041, -0.00001198, -0.00001532, -0.00003468, -0.00004376, 0.00003491, 0.00004812, 0.00001542, 0.00001551, 0.00000871, 0.00000373, 0.00001489, 0.00000098, 0.00000088, 0.00001126, 0.00000117, -0.00000683, -0.00000761, },
		{ -0.00000629, -0.00004270, 0.00002856, 0.00010314, 0.00004183, -0.00002084, -0.00001053, -0.00000093, 0.00000024, 0.00001942, 0.00000581, 0.00000334, 0.00001586, 0.00000991, 0.00001952, 0.00001912, 0.00005640, 0.00003897, -0.00005780, -0.00004077, -0.00001304, -0.00001413, -0.00000939, -0.00000803, -0.00000621, -0.00000070, -0.00000120, -0.00000921, -0.00000605, -0.00000433, -0.00000029, -0.00000249, -0.00000486, -0.00000268, 0.00000483, -0.00000839, -0.00000013, -0.00000383, 0.00000039, -0.00000660, -0.00000025, -0.00000473, 0.00000303, -0.00000039, -0.00000792, -0.00000861, 0.00000489, -0.00000585, 0.00000569, -0.00000786, 0.00001144, -0.00000661, 0.00000225, -0.00000262, 0.00000602, -0.00000587, 0.00000625, -0.00000443, 0.00000538, -0.00000834, 0.00001075, -0.00000048, 0.00000409, -0.00000484, 0.00000174, -0.00000443, 0.00000528, -0.00000503, 0.00000375, -0.00000710, 0.00000515, -0.00000603, 0.00001242, -0.00000249, 0.00000183, -0.00000538, 0.00000696, -0.00000187, 0.00000158, -0.00000184, 0.00000397, -0.00000564, 0.00000543, -0.00000308, 0.00000103, -0.00000156, 0.00000297, -0.00000225, -0.00000847, -0.00000475, 0.00000680, -0.00000950, 0.00000263, -0.00000498, -0.00000043, -0.00000517, 0.00000005, -0.00000315, -0.00000028, -0.00000459, -0.00000193, -0.00000190, 0.00000185, -0.00000423, -0.00000401, -0.00000764, -0.00000107, -0.00000993, -0.00000076, -0.00000265, -0.00000514, -0.00001173, -0.00001122, -0.00001821, -0.00003997, -0.00002801, 0.00004352, 0.00003845, 0.00001447, 0.00001376, 0.00000749, 0.00000323, 0.00001421, 0.00000083, -0.00000059, 0.00001050, 0.00000102, -0.00000763, },
		{ -0.00000780, -0.00000550, -0.00004758, 0.00002666, 0.00010426, 0.00004367, -0.00001868, -0.00000946, -0.00000068, 0.00000122, 0.00002044, 0.00000723, 0.00000355, 0.00001598, 0.00001281, 0.00002034, 0.00002408, 0.00006634, 0.00001461, -0.00006414, -0.00003155, -0.00001453, -0.00001232, -0.00000975, -0.00000713, -0.00000677, -0.00000032, -0.00000102, -0.00000986, -0.00000628, -0.00000300, -0.00000037, -0.00000285, -0.00000619, -0.00000058, 0.00000377, -0.00000808, -0.00000051, -0.00000364, 0.00000004, -0.00000661, 0.00000019, -0.00000547, 0.00000516, -0.00000194, -0.00001086, -0.00000424, 0.00000259, -0.00000397, 0.00000400, -0.00000644, 0.00001190, -0.00000828, 0.00000347, -0.00000262, 0.00000569, -0.00000553, 0.00000613, -0.00000477, 0.00000487, -0.00000747, 0.00001252, -0.00000314, 0.00000504, -0.00000583, 0.00000207, -0.00000383, 0.00000511, -0.00000509, 0.00000309, -0.00000647, 0.00000480, -0.00000477, 0.00001304, -0.00000503, 0.00000283, -0.00000562, 0.00000835, -0.00000354, 0.00000238, -0.00000168, 0.00000291, -0.00000514, 0.00000660, -0.00000503, 0.00000148, -0.00000140, 0.00000387, -0.00000452, -0.00000931, -0.00000056, 0.00000463, -0.00000942, 0.00000245, -0.00000505, 0.00000006, -0.00000551, 0.00000050, -0.00000370, 0.00000071, -0.00000541, -0.00000263, -0.00000020, 0.00000160, -0.00000478, -0.00000486, -0.00000671, -0.00000134, -0.00000958, -0.00000077, -0.00000331, -0.00000462, -0.00001327, -0.00000982, -0.00002352, -0.00004094, -0.00001077, 0.00004614, 0.00002967, 0.00001473, 0.00001134, 0.00000708, 0.00000241, 0.00001371, 0.00000047, -0.00000194, 0.00001001, 0.00000048, },
		{ -0.00000024, -0.00000738, -0.00000543, -0.00005211, 0.00002452, 0.00010593, 0.00004482, -0.00001615, -0.00000840, -0.00000031, 0.00000234, 0.00002129, 0.00000883, 0.00000358, 0.00001635, 0.00001624, 0.00002026, 0.00003307, 0.00007029, -0.00001403, -0.00006116, -0.00002570, -0.00001529, -0.00001112, -0.00000958, -0.00000676, -0.00000697, -0.00000016, -0.00000066, -0.00001101, -0.00000561, -0.00000234, -0.00000010, -0.00000401, -0.00000641, 0.00000098, 0.00000266, -0.00000766, -0.00000092, -0.00000347, -0.00000038, -0.00000637, 0.00000041, -0.00000602, 0.00000767, -0.00000536, -0.00001138, -0.00000096, 0.00000019, -0.00000157, 0.00000152, -0.00000374, 0.00001075, -0.00000886, 0.00000443, -0.00000271, 0.00000538, -0.00000508, 0.00000567, -0.00000478, 0.00000402, -0.00000553, 0.00001291, -0.00000530, 0.00000611, -0.00000712, 0.00000292, -0.00000358, 0.00000500, -0.00000529, 0.00000261, -0.00000580, 0.00000430, -0.00000282, 0.00001230, -0.00000678, 0.00000374, -0.00000555, 0.00000940, -0.00000546, 0.00000391, -0.00000230, 0.00000220, -0.00000427, 0.00000716, -0.00000696, 0.00000230, -0.00000127, 0.00000469, -0.00000781, -0.00000801, 0.00000238, 0.00000215, -0.00000872, 0.00000179, -0.00000463, 0.00000020, -0.00000563, 0.00000079, -0.00000415, 0.00000176, -0.00000675, -0.00000249, 0.00000119, 0.00000113, -0.00000546, -0.00000536, -0.00000575, -0.00000198, -0.00000869, -0.00000140, -0.00000342, -0.00000468, -0.00001427, -0.00000884, -0.00003024, -0.00003677, 0.00000526, 0.00004392, 0.00002276, 0.00001538, 0.00000861, 0.00000734, 0.00000122, 0.00001346, -0.00000024, -0.00000290, 0.00000953, },
		{ 0.00000879, -0.00000084, -0.00000662, -0.00000612, -0.00005605, 0.00002200, 0.00010819, 0.00004514, -0.00001295, -0.00000773, 0.00000060, 0.00000314, 0.00002235, 0.00001029, 0.00000366, 0.00001701, 0.00001996, 0.00001975, 0.00004630, 0.00006457, -0.00004163, -0.00005080, -0.00002409, -0.00001440, -0.00001112, -0.00000850, -0.00000725, -0.00000650, -0.00000043, -0.00000012, -0.00001252, -0.00000408, -0.00000245, 0.00000057, -0.00000596, -0.00000528, 0.00000147, 0.00000200, -0.00000749, -0.00000107, -0.00000360, -0.00000059, -0.00000607, 0.00000042, -0.00000611, 0.00000992, -0.00001012, -0.00000919, 0.00000011, -0.00000102, 0.00000021, -0.00000076, -0.00000054, 0.00000838, -0.00000825, 0.00000483, -0.00000270, 0.00000507, -0.00000467, 0.00000510, -0.00000470, 0.00000319, -0.00000285, 0.00001179, -0.00000644, 0.00000677, -0.00000829, 0.00000406, -0.00000363, 0.00000497, -0.00000565, 0.00000238, -0.00000528, 0.00000398, -0.00000060, 0.00001034, -0.00000743, 0.00000419, -0.00000483, 0.00000956, -0.00000698, 0.00000568, -0.00000356, 0.00000201, -0.00000323, 0.00000705, -0.00000857, 0.00000326, -0.00000093, 0.00000482, -0.00001128, -0.00000488, 0.00000358, 0.00000005, -0.00000783, 0.00000089, -0.00000380, -0.00000010, -0.00000539, 0.00000072, -0.00000424, 0.00000249, -0.00000824, -0.00000163, 0.00000214, 0.00000050, -0.00000618, -0.00000544, -0.00000500, -0.00000273, -0.00000739, -0.00000269, -0.00000279, -0.00000571, -0.00001407, -0.00000951, -0.00003649, -0.00002809, 0.00001820, 0.00003841, 0.00001813, 0.00001567, 0.00000609, 0.00000792, -0.00000017, 0.00001334, -0.00000127, -0.00000334, },
		{ -0.00000333, 0.00000768, -0.00000107, -0.00000587, -0.00000736, -0.00005938, 0.00001904, 0.00011093, 0.00004477, -0.00000902, -0.00000759, 0.00000225, 0.00000334, 0.00002390, 0.00001138, 0.00000398, 0.00001805, 0.00002350, 0.00002009, 0.00006230, 0.00004705, -0.00006248, -0.00003706, -0.00002565, -0.00001211, -0.00001208, -0.00000669, -0.00000846, -0.00000540, -0.00000108, 0.00000041, -0.00001396, -0.00000210, -0.00000311, 0.00000139, -0.00000829, -0.00000304, 0.00000078, 0.00000203, -0.00000772, -0.00000086, -0.00000415, -0.00000043, -0.00000586, 0.00000025, -0.00000553, 0.00001114, -0.00001513, -0.00000487, -0.00000132, -0.00000020, 0.00000035, -0.00000174, 0.00000205, 0.00000561, -0.00000680, 0.00000468, -0.00000253, 0.00000476, -0.00000440, 0.00000464, -0.00000480, 0.00000283, -0.00000004, 0.00000948, -0.00000639, 0.00000665, -0.00000894, 0.00000519, -0.00000386, 0.00000496, -0.00000610, 0.00000241, -0.00000502, 0.00000416, 0.00000134, 0.00000762, -0.00000703, 0.00000410, -0.00000338, 0.00000850, -0.00000749, 0.00000706, -0.00000507, 0.00000230, -0.00000220, 0.00000629, -0.00000968, 0.00000427, -0.00000036, 0.00000385, -0.00001399, -0.00000069, 0.00000302, -0.00000118, -0.00000720, 0.00000011, -0.00000284, -0.00000068, -0.00000483, 0.00000029, -0.00000387, 0.00000265, -0.00000951, -0.00000028, 0.00000258, -0.00000020, -0.00000691, -0.00000507, -0.00000471, -0.00000318, -0.00000614, -0.00000429, -0.00000167, -0.00000756, -0.00001258, -0.00001258, -0.00004030, -0.00001652, 0.00002714, 0.00003109, 0.00001572, 0.00001511, 0.00000427, 0.00000835, -0.00000145, 0.00001305, -0.00000234, },
		{ -0.00000313, -0.00000309, 0.00000621, -0.00000071, -0.00000555, -0.00000878, -0.00006227, 0.00001579, 0.00011384, 0.00004408, -0.00000464, -0.00000774, 0.00000448, 0.00000301, 0.00002592, 0.00001209, 0.00000456, 0.00001973, 0.00002613, 0.00002322, 0.00007759, 0.00001841, -0.00007228, -0.00002452, -0.00002811, -0.00000965, -0.00001308, -0.00000495, -0.00000975, -0.00000416, -0.00000178, 0.00000055, -0.00001476, -0.00000033, -0.00000378, 0.00000181, -0.00001029, -0.00000037, -0.00000073, 0.00000258, -0.00000823, -0.00000044, -0.00000498, 0.00000010, -0.00000581, -0.00000003, -0.00000416, 0.00001063, -0.00001899, 0.00000023, -0.00000460, 0.00000265, -0.00000156, -0.00000069, 0.00000303, 0.00000345, -0.00000517, 0.00000430, -0.00000231, 0.00000450, -0.00000435, 0.00000444, -0.00000528, 0.00000337, 0.00000212, 0.00000665, -0.00000543, 0.00000571, -0.00000886, 0.00000605, -0.00000415, 0.00000493, -0.00000657, 0.00000267, -0.00000503, 0.00000501, 0.00000243, 0.00000479, -0.00000599, 0.00000375, -0.00000151, 0.00000631, -0.00000668, 0.00000751, -0.00000632, 0.00000290, -0.00000130, 0.00000498, -0.00001018, 0.00000530, 0.00000025, 0.00000174, -0.00001518, 0.00000359, 0.00000108, -0.00000138, -0.00000712, -0.00000020, -0.00000212, -0.00000123, -0.00000419, -0.00000032, -0.00000316, 0.00000221, -0.00001027, 0.00000134, 0.00000245, -0.00000087, -0.00000759, -0.00000426, -0.00000510, -0.00000290, -0.00000552, -0.00000551, -0.00000076, -0.00000950, -0.00001053, -0.00001780, -0.00004034, -0.00000402, 0.00003197, 0.00002334, 0.00001510, 0.00001357, 0.00000339, 0.00000833, -0.00000234, 0.00001230, },
		{ 0.00001100, -0.00000341, -0.00000288, 0.00000452, 0.00000030, -0.00000604, -0.00000989, -0.00006505, 0.00001247, 0.00011653, 0.00004361, -0.00000032, -0.00000766, 0.00000676, 0.00000259, 0.00002806, 0.00001272, 0.00000522, 0.00002235, 0.00002725, 0.00003126, 0.00008708, -0.00001730, -0.00006977, -0.00001667, -0.00002914, -0.00000844, -0.00001302, -0.00000421, -0.00001030, -0.00000338, -0.00000213, -0.00000003, -0.00001443, 0.00000058, -0.00000392, 0.00000130, -0.00001116, 0.00000186, -0.00000240, 0.00000327, -0.00000870, -0.00000016, -0.00000576, 0.00000079, -0.00000588, -0.00000031, -0.00000214, 0.00000803, -0.00002043, 0.00000439, -0.00000826, 0.00000660, -0.00000505, 0.00000237, 0.00000191, 0.00000269, -0.00000407, 0.00000409, -0.00000228, 0.00000436, -0.00000452, 0.00000450, -0.00000615, 0.00000501, 0.00000298, 0.00000413, -0.00000413, 0.00000424, -0.00000807, 0.00000650, -0.00000441, 0.00000485, -0.00000701, 0.00000305, -0.00000522, 0.00000654, 0.00000230, 0.00000254, -0.00000486, 0.00000362, 0.00000015, 0.00000353, -0.00000470, 0.00000679, -0.00000693, 0.00000355, -0.00000057, 0.00000322, -0.00001001, 0.00000631, 0.00000058, -0.00000121, -0.00001448, 0.00000703, -0.00000153, -0.00000074, -0.00000764, 0.00000012, -0.00000192, -0.00000145, -0.00000378, -0.00000081, -0.00000240, 0.00000131, -0.00001041, 0.00000297, 0.00000179, -0.00000138, -0.00000820, -0.00000314, -0.00000614, -0.00000168, -0.00000602, -0.00000571, -0.00000087, -0.00001051, -0.00000924, -0.00002384, -0.00003646, 0.00000775, 0.00003306, 0.00001632, 0.00001548, 0.00001135, 0.00000336, 0.00000783, -0.00000284, },
		{ -0.00000321, 0.00000933, -0.00000313, -0.00000287, 0.00000283, 0.00000185, -0.00000754, -0.00001029, -0.00006806, 0.00000929, 0.00011871, 0.00004383, 0.00000343, -0.00000682, 0.00000849, 0.00000266, 0.00002981, 0.00001367, 0.00000578, 0.00002596, 0.00002685, 0.00004540, 0.00008538, -0.00005326, -0.00005726, -0.00001463, -0.00002756, -0.00000923, -0.00001132, -0.00000500, -0.00000957, -0.00000351, -0.00000194, -0.00000133, -0.00001284, 0.00000034, -0.00000328, -0.00000036, -0.00001044, 0.00000293, -0.00000360, 0.00000366, -0.00000879, -0.00000033, -0.00000616, 0.00000137, -0.00000600, -0.00000041, 0.00000006, 0.00000361, -0.00001881, 0.00000612, -0.00001060, 0.00001009, -0.00000887, 0.00000658, -0.00000100, 0.00000354, -0.00000387, 0.00000438, -0.00000264, 0.00000442, -0.00000486, 0.00000471, -0.00000716, 0.00000761, 0.00000222, 0.00000263, -0.00000315, 0.00000275, -0.00000683, 0.00000657, -0.00000460, 0.00000473, -0.00000739, 0.00000349, -0.00000541, 0.00000851, 0.00000086, 0.00000134, -0.00000414, 0.00000419, 0.00000085, 0.00000104, -0.00000216, 0.00000512, -0.00000676, 0.00000407, -0.00000010, 0.00000121, -0.00000912, 0.00000723, 0.00000030, -0.00000441, -0.00001196, 0.00000893, -0.00000402, 0.00000027, -0.00000851, 0.00000097, -0.00000229, -0.00000118, -0.00000379, -0.00000095, -0.00000188, 0.00000021, -0.00000990, 0.00000436, 0.00000072, -0.00000173, -0.00000862, -0.00000200, -0.00000752, 0.00000029, -0.00000766, -0.00000463, -0.00000248, -0.00000979, -0.00001004, -0.00002880, -0.00002957, 0.00001779, 0.00003097, 0.00001096, 0.00001591, 0.00000906, 0.00000373, 0.00000713, },
		{ 0.00000663, -0.00000384, 0.00000766, -0.00000244, -0.00000317, 0.00000137, 0.00000363, -0.00000993, -0.00000983, -0.00007143, 0.00000626, 0.00012033, 0.00004494, 0.00000642, -0.00000494, 0.00000931, 0.00000360, 0.00003083, 0.00001520, 0.00000637, 0.00003007, 0.00002612, 0.00006479, 0.00006868, -0.00008176, -0.00003989, -0.00001693, -0.00002387, -0.00001161, -0.00000834, -0.00000707, -0.00000769, -0.00000445, -0.00000144, -0.00000292, -0.00001044, -0.00000080, -0.00000211, -0.00000282, -0.00000825, 0.00000261, -0.00000403, 0.00000356, -0.00000841, -0.00000107, -0.00000600, 0.00000169, -0.00000615, -0.00000004, 0.00000167, -0.00000165, -0.00001449, 0.00000480, -0.00001035, 0.00001160, -0.00001145, 0.00001052, -0.00000469, 0.00000556, -0.00000452, 0.00000518, -0.00000343, 0.00000468, -0.00000526, 0.00000486, -0.00000790, 0.00001062, -0.00000000, 0.00000246, -0.00000302, 0.00000174, -0.00000550, 0.00000636, -0.00000475, 0.00000458, -0.00000768, 0.00000389, -0.00000534, 0.00001047, -0.00000160, 0.00000132, -0.00000402, 0.00000569, 0.00000004, -0.00000029, 0.00000004, 0.00000307, -0.00000598, 0.00000447, -0.00000003, -0.00000079, -0.00000756, 0.00000791, -0.00000080, -0.00000712, -0.00000813, 0.00000901, -0.00000575, 0.00000110, -0.00000929, 0.00000199, -0.00000304, -0.00000050, -0.00000423, -0.00000065, -0.00000179, -0.00000084, -0.00000877, 0.00000524, -0.00000051, -0.00000200, -0.00000867, -0.00000126, -0.00000863, 0.00000234, -0.00000988, -0.00000267, -0.00000534, -0.00000734, -0.00001359, -0.00003093, -0.00002120, 0.00002569, 0.00002638, 0.00000780, 0.00001559, 0.00000738, 0.00000394, },
		{ 0.00000358, 0.00000666, -0.00000507, 0.00000638, -0.00000161, -0.00000376, 0.00000043, 0.00000515, -0.00001279, -0.00000870, -0.00007504, 0.00000311, 0.00012172, 0.00004668, 0.00000890, -0.00000229, 0.00000937, 0.00000532, 0.00003120, 0.00001716, 0.00000749, 0.00003356, 0.00002754, 0.00008572, 0.00003674, -0.00009662, -0.00002384, -0.00002040, -0.00001989, -0.00001412, -0.00000532, -0.00000938, -0.00000544, -0.00000563, -0.00000120, -0.00000400, -0.00000808, -0.00000211, -0.00000112, -0.00000523, -0.00000531, 0.00000125, -0.00000384, 0.00000316, -0.00000781, -0.00000213, -0.00000539, 0.00000179, -0.00000640, 0.00000108, 0.00000177, -0.00000623, -0.00000882, 0.00000099, -0.00000728, 0.00001032, -0.00001160, 0.00001275, -0.00000783, 0.00000784, -0.00000552, 0.00000622, -0.00000451, 0.00000507, -0.00000559, 0.00000471, -0.00000790, 0.00001326, -0.00000306, 0.00000349, -0.00000391, 0.00000156, -0.00000441, 0.00000603, -0.00000485, 0.00000439, -0.00000784, 0.00000417, -0.00000481, 0.00001189, -0.00000449, 0.00000218, -0.00000434, 0.00000786, -0.00000231, 0.00000010, 0.00000110, 0.00000137, -0.00000502, 0.00000488, -0.00000055, -0.00000240, -0.00000552, 0.00000816, -0.00000274, -0.00000867, -0.00000380, 0.00000748, -0.00000638, 0.00000135, -0.00000954, 0.00000273, -0.00000375, 0.00000028, -0.00000489, -0.00000002, -0.00000217, -0.00000166, -0.00000711, 0.00000538, -0.00000159, -0.00000234, -0.00000817, -0.00000131, -0.00000882, 0.00000358, -0.00001168, -0.00000086, -0.00000842, -0.00000414, -0.00001934, -0.00002942, -0.00001276, 0.00003130, 0.00002020, 0.00000685, 0.00001408, 0.00000672, },
		{ 0.00000705, 0.00000263, 0.00000720, -0.00000698, 0.00000571, -0.00000086, -0.00000460, 0.00000020, 0.00000587, -0.00001549, -0.00000739, -0.00007849, -0.00000068, 0.00012349, 0.00004848, 0.00001151, 0.00000045, 0.00000933, 0.00000725, 0.00003146, 0.00001907, 0.00000989, 0.00003514, 0.00003431, 0.00010186, -0.00000619, -0.00009542, -0.00001394, -0.00002178, -0.00001775, -0.00001507, -0.00000374, -0.00001061, -0.00000388, -0.00000629, -0.00000189, -0.00000375, -0.00000673, -0.00000275, -0.00000114, -0.00000656, -0.00000270, -0.00000036, -0.00000353, 0.00000288, -0.00000746, -0.00000303, -0.00000470, 0.00000185, -0.00000683, 0.00000308, -0.00000033, -0.00000853, -0.00000380, -0.00000364, -0.00000241, 0.00000659, -0.00000903, 0.00001245, -0.00000934, 0.00000944, -0.00000619, 0.00000701, -0.00000557, 0.00000544, -0.00000576, 0.00000418, -0.00000680, 0.00001475, -0.00000606, 0.00000512, -0.00000556, 0.00000225, -0.00000376, 0.00000568, -0.00000490, 0.00000413, -0.00000784, 0.00000431, -0.00000367, 0.00001227, -0.00000708, 0.00000336, -0.00000458, 0.00001003, -0.00000567, 0.00000216, 0.00000062, 0.00000059, -0.00000429, 0.00000545, -0.00000185, -0.00000318, -0.00000331, 0.00000784, -0.00000532, -0.00000857, 0.00000006, 0.00000494, -0.00000599, 0.00000087, -0.00000903, 0.00000282, -0.00000406, 0.00000082, -0.00000545, 0.00000067, -0.00000289, -0.00000209, -0.00000512, 0.00000469, -0.00000228, -0.00000284, -0.00000713, -0.00000228, -0.00000769, 0.00000328, -0.00001211, -0.00000034, -0.00001033, -0.00000184, -0.00002565, -0.00002467, -0.00000506, 0.00003445, 0.00001360, 0.00000750, 0.00001156, },
		{ 0.00000867, 0.00000789, 0.00000145, 0.00000788, -0.00000928, 0.00000563, -0.00000031, -0.00000557, 0.00000074, 0.00000540, -0.00001740, -0.00000645, -0.00008136, -0.00000562, 0.00012626, 0.00004971, 0.00001495, 0.00000246, 0.00001000, 0.00000864, 0.00003228, 0.00002046, 0.00001407, 0.00003408, 0.00004898, 0.00010581, -0.00005182, -0.00008052, -0.00001199, -0.00001930, -0.00001865, -0.00001346, -0.00000452, -0.00000986, -0.00000378, -0.00000594, -0.00000378, -0.00000179, -0.00000695, -0.00000218, -0.00000269, -0.00000605, -0.00000137, -0.00000145, -0.00000362, 0.00000315, -0.00000779, -0.00000329, -0.00000433, 0.00000207, -0.00000739, 0.00000572, -0.00000471, -0.00000747, -0.00000137, -0.00000695, 0.00000236, 0.00000184, -0.00000455, 0.00000976, -0.00000881, 0.00000977, -0.00000602, 0.00000712, -0.00000631, 0.00000566, -0.00000579, 0.00000338, -0.00000461, 0.00001457, -0.00000813, 0.00000658, -0.00000744, 0.00000358, -0.00000358, 0.00000537, -0.00000488, 0.00000376, -0.00000764, 0.00000435, -0.00000200, 0.00001134, -0.00000873, 0.00000427, -0.00000414, 0.00001128, -0.00000901, 0.00000520, -0.00000122, 0.00000092, -0.00000402, 0.00000622, -0.00000394, -0.00000281, -0.00000136, 0.00000688, -0.00000814, -0.00000671, 0.00000263, 0.00000222, -0.00000503, -0.00000009, -0.00000785, 0.00000218, -0.00000377, 0.00000086, -0.00000566, 0.00000118, -0.00000381, -0.00000201, -0.00000310, 0.00000330, -0.00000255, -0.00000342, -0.00000582, -0.00000387, -0.00000539, 0.00000130, -0.00001070, -0.00000185, -0.00001000, -0.00000206, -0.00003037, -0.00001806, 0.00000175, 0.00003489, 0.00000786, 0.00000872, },
		{ 0.00000941, 0.00000624, 0.00000857, 0.00000055, 0.00000818, -0.00001148, 0.00000589, -0.00000001, -0.00000644, 0.00000181, 0.00000369, -0.00001819, -0.00000621, -0.00008344, -0.00001194, 0.00013033, 0.00005009, 0.00001957, 0.00000326, 0.00001193, 0.00000899, 0.00003402, 0.00002133, 0.00001970, 0.00003121, 0.00007174, 0.00009157, -0.00009005, -0.00005829, -0.00001625, -0.00001354, -0.00002212, -0.00000968, -0.00000742, -0.00000719, -0.00000516, -0.00000472, -0.00000649, 0.00000151, -0.00000856, -0.00000056, -0.00000558, -0.00000362, -0.00000172, -0.00000159, -0.00000435, 0.00000408, -0.00000894, -0.00000269, -0.00000448, 0.00000251, -0.00000784, 0.00000831, -0.00001059, -0.00000309, -0.00000255, -0.00000717, 0.00000497, -0.00000193, 0.00000014, 0.00000580, -0.00000669, 0.00000889, -0.00000493, 0.00000636, -0.00000654, 0.00000571, -0.00000579, 0.00000265, -0.00000172, 0.00001270, -0.00000874, 0.00000718, -0.00000892, 0.00000513, -0.00000376, 0.00000512, -0.00000480, 0.00000329, -0.00000725, 0.00000440, -0.00000008, 0.00000917, -0.00000914, 0.00000452, -0.00000264, 0.00001084, -0.00001116, 0.00000811, -0.00000371, 0.00000206, -0.00000409, 0.00000693, -0.00000655, -0.00000119, -0.00000003, 0.00000529, -0.00001069, -0.00000336, 0.00000342, 0.00000010, -0.00000411, -0.00000103, -0.00000641, 0.00000107, -0.00000297, 0.00000042, -0.00000544, 0.00000137, -0.00000480, -0.00000130, -0.00000149, 0.00000159, -0.00000257, -0.00000379, -0.00000474, -0.00000535, -0.00000271, -0.00000170, -0.00000791, -0.00000515, -0.00000736, -0.00000553, -0.00003175, -0.00001125, 0.00000789, 0.00003251, 0.00000406, },
		{ 0.00000269, 0.00000884, 0.00000487, 0.00000859, 0.00000032, 0.00000767, -0.00001310, 0.00000620, 0.00000003, -0.00000692, 0.00000288, 0.00000116, -0.00001801, -0.00000660, -0.00008492, -0.00001936, 0.00013548, 0.00004991, 0.00002508, 0.00000307, 0.00001498, 0.00000846, 0.00003642, 0.00002240, 0.00002536, 0.00002938, 0.00009900, 0.00005730, -0.00011218, -0.00003681, -0.00002251, -0.00000699, -0.00002618, -0.00000538, -0.00001106, -0.00000364, -0.00000723, -0.00000344, -0.00000900, 0.00000509, -0.00001069, 0.00000127, -0.00000888, -0.00000001, -0.00000343, -0.00000093, -0.00000551, 0.00000536, -0.00001057, -0.00000143, -0.00000506, 0.00000299, -0.00000771, 0.00000979, -0.00001641, 0.00000334, -0.00000697, -0.00000373, 0.00000410, -0.00000298, 0.00000318, 0.00000217, -0.00000408, 0.00000744, -0.00000332, 0.00000494, -0.00000630, 0.00000564, -0.00000598, 0.00000250, 0.00000111, 0.00000966, -0.00000792, 0.00000661, -0.00000952, 0.00000649, -0.00000414, 0.00000495, -0.00000473, 0.00000279, -0.00000678, 0.00000470, 0.00000162, 0.00000625, -0.00000843, 0.00000419, -0.00000020, 0.00000845, -0.00001130, 0.00000981, -0.00000586, 0.00000340, -0.00000416, 0.00000717, -0.00000911, 0.00000146, 0.00000042, 0.00000315, -0.00001240, 0.00000084, 0.00000240, -0.00000089, -0.00000378, -0.00000143, -0.00000527, -0.00000004, -0.00000201, -0.00000029, -0.00000494, 0.00000125, -0.00000577, 0.00000013, -0.00000072, 0.00000005, -0.00000268, -0.00000361, -0.00000449, -0.00000583, -0.00000085, -0.00000446, -0.00000499, -0.00000901, -0.00000351, -0.00001158, -0.00002922, -0.00000544, 0.00001342, 0.00002763, },
		{ 0.00002123, 0.00000344, 0.00000698, 0.00000463, 0.00000788, 0.00000075, 0.00000628, -0.00001392, 0.00000636, -0.00000022, -0.00000666, 0.00000327, -0.00000139, -0.00001746, -0.00000714, -0.00008639, -0.00002722, 0.00014099, 0.00004997, 0.00003057, 0.00000276, 0.00001834, 0.00000786, 0.00003863, 0.00002499, 0.00002893, 0.00003296, 0.00012334, 0.00000700, -0.00011398, -0.00002284, -0.00002622, -0.00000270, -0.00002835, -0.00000274, -0.00001353, -0.00000079, -0.00000879, -0.00000311, -0.00001012, 0.00000764, -0.00001213, 0.00000216, -0.00001128, 0.00000351, -0.00000559, -0.00000007, -0.00000653, 0.00000634, -0.00001203, -0.00000007, -0.00000567, 0.00000320, -0.00000659, 0.00000910, -0.00002024, 0.00000958, -0.00001288, 0.00000242, -0.00000017, -0.00000060, 0.00000333, 0.00000031, -0.00000222, 0.00000636, -0.00000191, 0.00000332, -0.00000583, 0.00000560, -0.00000653, 0.00000336, 0.00000299, 0.00000638, -0.00000620, 0.00000502, -0.00000911, 0.00000740, -0.00000459, 0.00000491, -0.00000480, 0.00000244, -0.00000635, 0.00000545, 0.00000252, 0.00000331, -0.00000714, 0.00000377, 0.00000250, 0.00000459, -0.00000927, 0.00000964, -0.00000690, 0.00000430, -0.00000382, 0.00000645, -0.00001090, 0.00000458, -0.00000007, 0.00000062, -0.00001277, 0.00000499, 0.00000005, -0.00000070, -0.00000427, -0.00000096, -0.00000489, -0.00000068, -0.00000132, -0.00000091, -0.00000438, 0.00000092, -0.00000662, 0.00000222, -0.00000106, -0.00000090, -0.00000313, -0.00000270, -0.00000540, -0.00000464, -0.00000085, -0.00000561, -0.00000349, -0.00001166, -0.00000048, -0.00001825, -0.00002366, -0.00000085, 0.00001794, },
		{ 0.00002057, 0.00001480, 0.00000525, 0.00000445, 0.00000503, 0.00000686, 0.00000140, 0.00000438, -0.00001403, 0.00000630, -0.00000075, -0.00000543, 0.00000233, -0.00000312, -0.00001728, -0.00000721, -0.00008855, -0.00003470, 0.00014604, 0.00005124, 0.00003492, 0.00000351, 0.00002088, 0.00000826, 0.00003971, 0.00003026, 0.00002870, 0.00004619, 0.00013509, -0.00004970, -0.00009738, -0.00001933, -0.00002466, -0.00000263, -0.00002699, -0.00000328, -0.00001337, 0.00000012, -0.00000895, -0.00000440, -0.00000912, 0.00000824, -0.00001201, 0.00000123, -0.00001168, 0.00000573, -0.00000724, 0.00000031, -0.00000682, 0.00000641, -0.00001261, 0.00000080, -0.00000593, 0.00000298, -0.00000440, 0.00000575, -0.00002054, 0.00001326, -0.00001770, 0.00000907, -0.00000635, 0.00000451, 0.00000048, 0.00000090, -0.00000193, 0.00000641, -0.00000140, 0.00000207, -0.00000541, 0.00000567, -0.00000742, 0.00000541, 0.00000324, 0.00000386, -0.00000443, 0.00000297, -0.00000787, 0.00000779, -0.00000506, 0.00000505, -0.00000511, 0.00000237, -0.00000608, 0.00000676, 0.00000217, 0.00000111, -0.00000595, 0.00000394, 0.00000442, 0.00000038, -0.00000576, 0.00000769, -0.00000652, 0.00000444, -0.00000290, 0.00000450, -0.00001125, 0.00000748, -0.00000136, -0.00000207, -0.00001152, 0.00000818, -0.00000284, 0.00000028, -0.00000539, 0.00000034, -0.00000539, -0.00000058, -0.00000120, -0.00000116, -0.00000401, 0.00000047, -0.00000718, 0.00000464, -0.00000243, -0.00000113, -0.00000395, -0.00000123, -0.00000726, -0.00000181, -0.00000310, -0.00000439, -0.00000451, -0.00001165, -0.00000025, -0.00002303, -0.00001697, 0.00000318, },
		{ 0.00000757, 0.00002040, 0.00000988, 0.00000682, 0.00000221, 0.00000533, 0.00000615, 0.00000160, 0.00000258, -0.00001375, 0.00000605, -0.00000142, -0.00000333, -0.00000015, -0.00000348, -0.00001804, -0.00000640, -0.00009179, -0.00004134, 0.00015015, 0.00005426, 0.00003744, 0.00000610, 0.00002185, 0.00001028, 0.00003936, 0.00003828, 0.00002477, 0.00007081, 0.00012546, -0.00009997, -0.00006988, -0.00002455, -0.00001821, -0.00000649, -0.00002227, -0.00000697, -0.00001040, -0.00000118, -0.00000766, -0.00000708, -0.00000618, 0.00000685, -0.00001025, -0.00000157, -0.00000973, 0.00000599, -0.00000777, -0.00000013, -0.00000621, 0.00000535, -0.00001200, 0.00000084, -0.00000569, 0.00000248, -0.00000169, 0.00000021, -0.00001690, 0.00001284, -0.00001910, 0.00001360, -0.00001201, 0.00001048, -0.00000426, 0.00000356, -0.00000326, 0.00000778, -0.00000215, 0.00000157, -0.00000524, 0.00000582, -0.00000841, 0.00000839, 0.00000162, 0.00000283, -0.00000345, 0.00000118, -0.00000626, 0.00000775, -0.00000549, 0.00000538, -0.00000573, 0.00000269, -0.00000597, 0.00000854, 0.00000046, 0.00000016, -0.00000539, 0.00000525, 0.00000455, -0.00000276, -0.00000203, 0.00000475, -0.00000504, 0.00000394, -0.00000161, 0.00000150, -0.00000984, 0.00000950, -0.00000323, -0.00000459, -0.00000870, 0.00000971, -0.00000535, 0.00000140, -0.00000662, 0.00000204, -0.00000654, 0.00000019, -0.00000168, -0.00000096, -0.00000390, -0.00000008, -0.00000717, 0.00000686, -0.00000439, -0.00000085, -0.00000481, 0.00000025, -0.00000930, 0.00000182, -0.00000695, -0.00000115, -0.00000799, -0.00000862, -0.00000384, -0.00002394, -0.00001119, },
		{ -0.00000755, 0.00001285, 0.00001710, 0.00000746, 0.00000712, 0.00000103, 0.00000490, 0.00000625, 0.00000083, 0.00000145, -0.00001342, 0.00000558, -0.00000190, -0.00000090, -0.00000381, -0.00000254, -0.00001972, -0.00000484, -0.00009593, -0.00004733, 0.00015357, 0.00005876, 0.00003832, 0.00001045, 0.00002140, 0.00001362, 0.00003844, 0.00004748, 0.00002007, 0.00010391, 0.00009022, -0.00013188, -0.00004183, -0.00003319, -0.00001001, -0.00001181, -0.00001624, -0.00001213, -0.00000588, -0.00000372, -0.00000586, -0.00001000, -0.00000249, 0.00000446, -0.00000777, -0.00000532, -0.00000614, 0.00000457, -0.00000732, -0.00000113, -0.00000506, 0.00000355, -0.00001046, 0.00000023, -0.00000520, 0.00000227, 0.00000041, -0.00000599, -0.00001042, 0.00000835, -0.00001604, 0.00001412, -0.00001486, 0.00001505, -0.00000901, 0.00000704, -0.00000546, 0.00001004, -0.00000401, 0.00000188, -0.00000535, 0.00000590, -0.00000903, 0.00001162, -0.00000147, 0.00000345, -0.00000374, 0.00000024, -0.00000473, 0.00000742, -0.00000583, 0.00000577, -0.00000655, 0.00000331, -0.00000586, 0.00001039, -0.00000231, 0.00000051, -0.00000554, 0.00000775, 0.00000238, -0.00000368, 0.00000053, 0.00000198, -0.00000325, 0.00000332, -0.00000057, -0.00000190, -0.00000685, 0.00001023, -0.00000544, -0.00000651, -0.00000476, 0.00000933, -0.00000679, 0.00000200, -0.00000734, 0.00000348, -0.00000779, 0.00000123, -0.00000251, -0.00000047, -0.00000399, -0.00000076, -0.00000632, 0.00000825, -0.00000629, -0.00000055, -0.00000522, 0.00000098, -0.00001040, 0.00000484, -0.00001090, 0.00000269, -0.00001261, -0.00000373, -0.00001062, -0.00002042, },
		{ -0.00001369, -0.00000591, 0.00001857, 0.00001129, 0.00000760, 0.00000588, 0.00000110, 0.00000363, 0.00000715, -0.00000099, 0.00000123, -0.00001322, 0.00000478, -0.00000168, 0.00000099, -0.00000770, -0.00000101, -0.00002174, -0.00000325, -0.00010016, -0.00005348, 0.00015727, 0.00006363, 0.00003870, 0.00001546, 0.00002065, 0.00001701, 0.00003888, 0.00005489, 0.00002010, 0.00013726, 0.00003241, -0.00013852, -0.00002256, -0.00003903, -0.00000428, -0.00001516, -0.00001187, -0.00001616, -0.00000196, -0.00000586, -0.00000501, -0.00001154, 0.00000026, 0.00000263, -0.00000606, -0.00000842, -0.00000234, 0.00000248, -0.00000661, -0.00000198, -0.00000422, 0.00000191, -0.00000874, -0.00000047, -0.00000499, 0.00000306, 0.00000053, -0.00001065, -0.00000348, 0.00000157, -0.00000930, 0.00001029, -0.00001369, 0.00001649, -0.00001196, 0.00000981, -0.00000737, 0.00001228, -0.00000636, 0.00000273, -0.00000560, 0.00000570, -0.00000876, 0.00001416, -0.00000514, 0.00000527, -0.00000528, 0.00000044, -0.00000363, 0.00000695, -0.00000599, 0.00000603, -0.00000734, 0.00000406, -0.00000548, 0.00001173, -0.00000547, 0.00000173, -0.00000599, 0.00001084, -0.00000175, -0.00000198, 0.00000098, 0.00000044, -0.00000199, 0.00000322, -0.00000051, -0.00000472, -0.00000296, 0.00000963, -0.00000780, -0.00000737, -0.00000048, 0.00000732, -0.00000694, 0.00000171, -0.00000708, 0.00000409, -0.00000855, 0.00000203, -0.00000330, 0.00000006, -0.00000416, -0.00000159, -0.00000454, 0.00000831, -0.00000748, -0.00000069, -0.00000478, 0.00000039, -0.00000960, 0.00000579, -0.00001320, 0.00000518, -0.00001622, 0.00000071, -0.00001837, },
		{ -0.00002415, -0.00000616, -0.00000497, 0.00002341, 0.00000453, 0.00000932, 0.00000365, 0.00000196, 0.00000199, 0.00000834, -0.00000343, 0.00000175, -0.00001314, 0.00000351, -0.00000034, 0.00000145, -0.00001072, 0.00000012, -0.00002323, -0.00000256, -0.00010341, -0.00006092, 0.00016246, 0.00006743, 0.00004015, 0.00001960, 0.00002116, 0.00001888, 0.00004271, 0.00005752, 0.00003111, 0.00015877, -0.00003729, -0.00012070, -0.00001684, -0.00003791, -0.00000410, -0.00001392, -0.00001154, -0.00001688, -0.00000053, -0.00000613, -0.00000628, -0.00001047, 0.00000065, 0.00000271, -0.00000644, -0.00000936, 0.00000010, 0.00000096, -0.00000652, -0.00000188, -0.00000455, 0.00000137, -0.00000771, -0.00000061, -0.00000551, 0.00000528, -0.00000229, -0.00001175, 0.00000110, -0.00000456, -0.00000127, 0.00000359, -0.00000894, 0.00001435, -0.00001212, 0.00001077, -0.00000796, 0.00001351, -0.00000843, 0.00000372, -0.00000583, 0.00000512, -0.00000727, 0.00001517, -0.00000827, 0.00000742, -0.00000754, 0.00000163, -0.00000310, 0.00000643, -0.00000587, 0.00000595, -0.00000782, 0.00000468, -0.00000459, 0.00001199, -0.00000816, 0.00000309, -0.00000601, 0.00001338, -0.00000671, 0.00000176, -0.00000078, 0.00000062, -0.00000179, 0.00000405, -0.00000199, -0.00000599, 0.00000089, 0.00000797, -0.00001018, -0.00000680, 0.00000319, 0.00000444, -0.00000606, 0.00000061, -0.00000581, 0.00000362, -0.00000843, 0.00000220, -0.00000371, 0.00000039, -0.00000431, -0.00000238, -0.00000204, 0.00000696, -0.00000764, -0.00000148, -0.00000347, -0.00000153, -0.00000668, 0.00000397, -0.00001265, 0.00000477, -0.00001687, 0.00000221, },
		{ -0.00000070, -0.00002553, -0.00000042, -0.00000301, 0.00002576, -0.00000126, 0.00001117, 0.00000146, 0.00000277, 0.00000072, 0.00000900, -0.00000574, 0.00000259, -0.00001306, 0.00000179, 0.00000222, 0.00000001, -0.00001204, 0.00000001, -0.00002351, -0.00000347, -0.00010495, -0.00007035, 0.00016994, 0.00006915, 0.00004383, 0.00002171, 0.00002397, 0.00001843, 0.00005072, 0.00005433, 0.00005698, 0.00015623, -0.00010309, -0.00008681, -0.00002337, -0.00002962, -0.00000970, -0.00000781, -0.00001569, -0.00001363, -0.00000225, -0.00000413, -0.00000976, -0.00000665, -0.00000168, 0.00000510, -0.00000932, -0.00000745, 0.00000025, 0.00000080, -0.00000755, -0.00000050, -0.00000641, 0.00000243, -0.00000788, 0.00000013, -0.00000679, 0.00000869, -0.00000802, -0.00000834, 0.00000118, -0.00000712, 0.00000489, -0.00000325, -0.00000255, 0.00000967, -0.00000971, 0.00000969, -0.00000682, 0.00001316, -0.00000961, 0.00000446, -0.00000595, 0.00000431, -0.00000467, 0.00001422, -0.00000997, 0.00000893, -0.00000974, 0.00000339, -0.00000307, 0.00000590, -0.00000545, 0.00000539, -0.00000777, 0.00000501, -0.00000310, 0.00001082, -0.00000967, 0.00000393, -0.00000489, 0.00001416, -0.00001088, 0.00000619, -0.00000397, 0.00000225, -0.00000257, 0.00000566, -0.00000503, -0.00000511, 0.00000375, 0.00000567, -0.00001242, -0.00000466, 0.00000539, 0.00000165, -0.00000483, -0.00000076, -0.00000396, 0.00000228, -0.00000747, 0.00000165, -0.00000363, 0.00000047, -0.00000453, -0.00000278, 0.00000058, 0.00000464, -0.00000696, -0.00000263, -0.00000183, -0.00000401, -0.00000239, -0.00000014, -0.00000935, 0.00000117, -0.00001380, },
		{ -0.00000803, -0.00000759, -0.00002177, 0.00000201, 0.00000109, 0.00002455, -0.00000456, 0.00001180, 0.00000027, 0.00000279, 0.00000042, 0.00000849, -0.00000724, 0.00000335, -0.00001298, -0.00000004, 0.00000552, -0.00000308, -0.00001158, -0.00000148, -0.00002255, -0.00000593, -0.00010482, -0.00008162, 0.00017945, 0.00006895, 0.00004974, 0.00002185, 0.00002883, 0.00001652, 0.00006143, 0.00004794, 0.00009618, 0.00012208, -0.00014904, -0.00004976, -0.00003579, -0.00001782, -0.00001824, 0.00000090, -0.00002248, -0.00000776, -0.00000614, -0.00000085, -0.00001422, -0.00000130, -0.00000572, 0.00000892, -0.00001386, -0.00000325, -0.00000175, 0.00000195, -0.00000945, 0.00000177, -0.00000940, 0.00000485, -0.00000911, 0.00000150, -0.00000826, 0.00001216, -0.00001526, -0.00000120, -0.00000364, -0.00000449, 0.00000664, -0.00000736, 0.00000284, 0.00000449, -0.00000605, 0.00000735, -0.00000436, 0.00001127, -0.00000969, 0.00000480, -0.00000606, 0.00000367, -0.00000151, 0.00001150, -0.00000985, 0.00000912, -0.00001116, 0.00000518, -0.00000340, 0.00000545, -0.00000485, 0.00000442, -0.00000718, 0.00000511, -0.00000124, 0.00000831, -0.00000975, 0.00000397, -0.00000237, 0.00001240, -0.00001278, 0.00000966, -0.00000727, 0.00000443, -0.00000371, 0.00000739, -0.00000898, -0.00000210, 0.00000500, 0.00000317, -0.00001427, -0.00000117, 0.00000566, -0.00000020, -0.00000401, -0.00000170, -0.00000224, 0.00000069, -0.00000608, 0.00000065, -0.00000323, 0.00000041, -0.00000498, -0.00000233, 0.00000250, 0.00000215, -0.00000604, -0.00000354, -0.00000073, -0.00000577, 0.00000164, -0.00000493, -0.00000475, -0.00000434, },
		{ -0.00000930, -0.00000210, -0.00001611, -0.00001422, 0.00000137, 0.00000710, 0.00001995, -0.00000493, 0.00001069, 0.00000044, 0.00000185, 0.00000113, 0.00000670, -0.00000766, 0.00000387, -0.00001312, -0.00000134, 0.00000853, -0.00000679, -0.00001011, -0.00000370, -0.00002111, -0.00000907, -0.00010396, -0.00009359, 0.00018963, 0.00006825, 0.00005653, 0.00002149, 0.00003401, 0.00001569, 0.00007105, 0.00004483, 0.00014009, 0.00005748, -0.00016453, -0.00002210, -0.00004586, -0.00000815, -0.00002518, 0.00000832, -0.00002859, -0.00000201, -0.00001004, 0.00000177, -0.00001762, 0.00000344, -0.00000954, 0.00001234, -0.00001820, 0.00000161, -0.00000477, 0.00000355, -0.00001132, 0.00000386, -0.00001241, 0.00000766, -0.00001063, 0.00000274, -0.00000890, 0.00001404, -0.00002164, 0.00000724, -0.00001172, 0.00000290, 0.00000313, -0.00000695, 0.00000500, 0.00000101, -0.00000292, 0.00000510, -0.00000164, 0.00000850, -0.00000888, 0.00000482, -0.00000636, 0.00000376, 0.00000127, 0.00000785, -0.00000822, 0.00000787, -0.00001139, 0.00000659, -0.00000393, 0.00000521, -0.00000432, 0.00000332, -0.00000627, 0.00000525, 0.00000044, 0.00000504, -0.00000867, 0.00000348, 0.00000105, 0.00000819, -0.00001170, 0.00001091, -0.00000935, 0.00000610, -0.00000434, 0.00000826, -0.00001271, 0.00000234, 0.00000451, 0.00000073, -0.00001534, 0.00000303, 0.00000406, -0.00000067, -0.00000412, -0.00000161, -0.00000138, -0.00000044, -0.00000487, -0.00000037, -0.00000281, 0.00000042, -0.00000580, -0.00000071, 0.00000298, 0.00000038, -0.00000557, -0.00000357, -0.00000102, -0.00000554, 0.00000359, -0.00000821, -0.00000118, },
		{ -0.00000071, -0.00001121, 0.00000098, -0.00002285, -0.00000579, -0.00000050, 0.00001345, 0.00001340, -0.00000314, 0.00000826, 0.00000152, 0.00000043, 0.00000227, 0.00000413, -0.00000724, 0.00000427, -0.00001380, -0.00000142, 0.00001007, -0.00000971, -0.00000890, -0.00000551, -0.00002042, -0.00001158, -0.00010376, -0.00010466, 0.00019854, 0.00006918, 0.00006207, 0.00002286, 0.00003711, 0.00001912, 0.00007501, 0.00005353, 0.00017416, -0.00002638, -0.00014830, -0.00001133, -0.00004738, -0.00000522, -0.00002663, 0.00001099, -0.00003085, 0.00000090, -0.00001177, 0.00000194, -0.00001808, 0.00000558, -0.00001124, 0.00001351, -0.00002037, 0.00000520, -0.00000728, 0.00000446, -0.00001209, 0.00000462, -0.00001415, 0.00000967, -0.00001146, 0.00000305, -0.00000785, 0.00001285, -0.00002461, 0.00001373, -0.00001982, 0.00001255, -0.00000433, -0.00000210, 0.00000307, 0.00000057, -0.00000176, 0.00000425, 0.00000012, 0.00000581, -0.00000772, 0.00000475, -0.00000701, 0.00000501, 0.00000269, 0.00000440, -0.00000593, 0.00000561, -0.00001045, 0.00000741, -0.00000457, 0.00000531, -0.00000418, 0.00000249, -0.00000538, 0.00000578, 0.00000127, 0.00000189, -0.00000715, 0.00000322, 0.00000426, 0.00000262, -0.00000800, 0.00000956, -0.00000949, 0.00000649, -0.00000382, 0.00000743, -0.00001495, 0.00000706, 0.00000262, -0.00000154, -0.00001522, 0.00000703, 0.00000121, 0.00000009, -0.00000520, -0.00000033, -0.00000177, -0.00000060, -0.00000434, -0.00000098, -0.00000263, 0.00000060, -0.00000688, 0.00000200, 0.00000172, -0.00000012, -0.00000599, -0.00000246, -0.00000296, -0.00000273, 0.00000232, -0.00000832, },
		{ -0.00000491, -0.00000415, -0.00000875, -0.00000086, -0.00002483, 0.00000032, -0.00000112, 0.00001789, 0.00000714, -0.00000078, 0.00000561, 0.00000262, -0.00000061, 0.00000288, 0.00000166, -0.00000651, 0.00000478, -0.00001515, 0.00000006, 0.00000933, -0.00001068, -0.00000909, -0.00000593, -0.00002146, -0.00001241, -0.00010525, -0.00011368, 0.00020466, 0.00007349, 0.00006461, 0.00002767, 0.00003651, 0.00002874, 0.00007061, 0.00008079, 0.00018210, -0.00011063, -0.00010914, -0.00001721, -0.00003898, -0.00001021, -0.00002142, 0.00000767, -0.00002791, -0.00000033, -0.00001039, -0.00000094, -0.00001498, 0.00000428, -0.00000993, 0.00001154, -0.00001921, 0.00000621, -0.00000820, 0.00000394, -0.00001119, 0.00000339, -0.00001383, 0.00001006, -0.00001102, 0.00000210, -0.00000498, 0.00000812, -0.00002263, 0.00001550, -0.00002443, 0.00002083, -0.00001279, 0.00000527, -0.00000211, 0.00000319, -0.00000299, 0.00000545, 0.00000007, 0.00000405, -0.00000671, 0.00000475, -0.00000798, 0.00000751, 0.00000214, 0.00000226, -0.00000404, 0.00000315, -0.00000874, 0.00000765, -0.00000525, 0.00000581, -0.00000465, 0.00000226, -0.00000483, 0.00000696, 0.00000070, -0.00000025, -0.00000597, 0.00000403, 0.00000584, -0.00000257, -0.00000304, 0.00000629, -0.00000780, 0.00000555, -0.00000209, 0.00000461, -0.00001477, 0.00001084, -0.00000007, -0.00000363, -0.00001361, 0.00000986, -0.00000189, 0.00000146, -0.00000677, 0.00000178, -0.00000325, 0.00000028, -0.00000466, -0.00000103, -0.00000275, 0.00000084, -0.00000785, 0.00000524, -0.00000095, 0.00000060, -0.00000714, -0.00000062, -0.00000593, 0.00000206, -0.00000201, },
		{ -0.00000775, 0.00000063, -0.00001037, -0.00000263, -0.00000760, -0.00002087, 0.00000207, 0.00000135, 0.00001860, 0.00000316, 0.00000056, 0.00000386, 0.00000287, -0.00000055, 0.00000219, 0.00000010, -0.00000596, 0.00000544, -0.00001690, 0.00000274, 0.00000641, -0.00000941, -0.00001103, -0.00000473, -0.00002437, -0.00001145, -0.00010840, -0.00012073, 0.00020788, 0.00008149, 0.00006395, 0.00003594, 0.00003278, 0.00004352, 0.00005958, 0.00012729, 0.00015190, -0.00017445, -0.00006271, -0.00003250, -0.00002467, -0.00002013, -0.00001190, 0.00000014, -0.00002102, -0.00000480, -0.00000681, -0.00000573, -0.00000946, 0.00000040, -0.00000634, 0.00000709, -0.00001507, 0.00000456, -0.00000743, 0.00000217, -0.00000897, 0.00000054, -0.00001163, 0.00000890, -0.00000948, 0.00000040, -0.00000125, 0.00000092, -0.00001600, 0.00001156, -0.00002322, 0.00002452, -0.00001883, 0.00001221, -0.00000835, 0.00000755, -0.00000592, 0.00000839, -0.00000191, 0.00000364, -0.00000619, 0.00000484, -0.00000895, 0.00001082, -0.00000038, 0.00000200, -0.00000341, 0.00000132, -0.00000684, 0.00000747, -0.00000588, 0.00000663, -0.00000574, 0.00000277, -0.00000469, 0.00000875, -0.00000141, -0.00000090, -0.00000561, 0.00000635, 0.00000474, -0.00000560, 0.00000133, 0.00000251, -0.00000523, 0.00000392, 0.00000019, 0.00000033, -0.00001198, 0.00001279, -0.00000293, -0.00000544, -0.00001053, 0.00001083, -0.00000426, 0.00000260, -0.00000804, 0.00000392, -0.00000520, 0.00000178, -0.00000561, -0.00000068, -0.00000300, 0.00000084, -0.00000807, 0.00000803, -0.00000412, 0.00000188, -0.00000836, 0.00000099, -0.00000855, 0.00000715, },
		{ 0.00001034, -0.00001239, 0.00000571, -0.00001673, 0.00000448, -0.00001687, -0.00001229, -0.00000044, 0.00000711, 0.00001518, 0.00000237, 0.00000005, 0.00000352, 0.00000200, 0.00000073, 0.00000003, -0.00000026, -0.00000572, 0.00000594, -0.00001827, 0.00000547, 0.00000240, -0.00000673, -0.00001401, -0.00000275, -0.00002819, -0.00000978, -0.00011189, -0.00012742, 0.00020982, 0.00009162, 0.00006182, 0.00004555, 0.00002901, 0.00005903, 0.00004909, 0.00018476, 0.00008159, -0.00020224, -0.00002560, -0.00004658, -0.00001160, -0.00002929, -0.00000284, -0.00000753, -0.00001350, -0.00000988, -0.00000343, -0.00000991, -0.00000402, -0.00000378, -0.00000258, 0.00000222, -0.00000970, 0.00000150, -0.00000597, 0.00000023, -0.00000669, -0.00000261, -0.00000869, 0.00000713, -0.00000779, -0.00000081, 0.00000145, -0.00000624, -0.00000708, 0.00000329, -0.00001619, 0.00002215, -0.00002005, 0.00001600, -0.00001317, 0.00001167, -0.00000907, 0.00001196, -0.00000519, 0.00000444, -0.00000616, 0.00000486, -0.00000938, 0.00001407, -0.00000421, 0.00000352, -0.00000437, 0.00000070, -0.00000524, 0.00000702, -0.00000632, 0.00000748, -0.00000717, 0.00000384, -0.00000478, 0.00001070, -0.00000464, -0.00000014, -0.00000597, 0.00000990, 0.00000078, -0.00000539, 0.00000349, -0.00000021, -0.00000304, 0.00000266, 0.00000187, -0.00000420, -0.00000720, 0.00001263, -0.00000555, -0.00000674, -0.00000641, 0.00000977, -0.00000525, 0.00000281, -0.00000828, 0.00000524, -0.00000678, 0.00000322, -0.00000666, -0.00000028, -0.00000307, 0.00000030, -0.00000699, 0.00000939, -0.00000667, 0.00000280, -0.00000879, 0.00000128, -0.00000922, },
		{ -0.00000694, 0.00000997, -0.00001367, 0.00000764, -0.00002014, 0.00000905, -0.00002485, -0.00000238, -0.00000502, 0.00001451, 0.00000886, 0.00000425, -0.00000205, 0.00000422, 0.00000054, 0.00000256, -0.00000302, 0.00000026, -0.00000551, 0.00000573, -0.00001826, 0.00000672, -0.00000095, -0.00000419, -0.00001660, -0.00000154, -0.00003124, -0.00000920, -0.00011360, -0.00013626, 0.00021318, 0.00010115, 0.00006124, 0.00005312, 0.00002969, 0.00006891, 0.00005013, 0.00023636, -0.00001800, -0.00018940, -0.00000905, -0.00005050, -0.00000644, -0.00003210, 0.00000087, -0.00001097, -0.00000909, -0.00001251, -0.00000283, -0.00001088, -0.00000139, -0.00000569, -0.00000112, -0.00000056, -0.00000546, -0.00000111, -0.00000529, -0.00000044, -0.00000592, -0.00000439, -0.00000658, 0.00000608, -0.00000717, -0.00000019, 0.00000113, -0.00001036, 0.00000052, -0.00000588, -0.00000579, 0.00001474, -0.00001605, 0.00001531, -0.00001479, 0.00001380, -0.00001086, 0.00001474, -0.00000869, 0.00000592, -0.00000640, 0.00000459, -0.00000877, 0.00001623, -0.00000819, 0.00000606, -0.00000660, 0.00000139, -0.00000425, 0.00000643, -0.00000639, 0.00000797, -0.00000847, 0.00000506, -0.00000466, 0.00001210, -0.00000814, 0.00000140, -0.00000636, 0.00001361, -0.00000514, -0.00000200, 0.00000274, -0.00000082, -0.00000224, 0.00000268, 0.00000184, -0.00000748, -0.00000170, 0.00001068, -0.00000779, -0.00000711, -0.00000207, 0.00000712, -0.00000480, 0.00000190, -0.00000715, 0.00000518, -0.00000732, 0.00000394, -0.00000730, -0.00000017, -0.00000280, -0.00000083, -0.00000444, 0.00000869, -0.00000773, 0.00000265, -0.00000787, -0.00000036, },
		{ -0.00000355, -0.00000198, 0.00000578, -0.00001073, 0.00000504, -0.00001860, 0.00000831, -0.00002788, 0.00000495, -0.00000845, 0.00002078, 0.00000205, 0.00000716, -0.00000459, 0.00000496, -0.00000050, 0.00000381, -0.00000591, 0.00000095, -0.00000493, 0.00000442, -0.00001617, 0.00000528, -0.00000207, -0.00000337, -0.00001737, -0.00000254, -0.00003200, -0.00001126, -0.00011179, -0.00014945, 0.00022040, 0.00010754, 0.00006486, 0.00005589, 0.00003835, 0.00006820, 0.00007303, 0.00026109, -0.00012502, -0.00014429, -0.00001489, -0.00004103, -0.00001192, -0.00002608, -0.00000321, -0.00000777, -0.00001001, -0.00001095, -0.00000635, -0.00000728, -0.00000309, -0.00000379, -0.00000348, 0.00000048, -0.00000418, -0.00000178, -0.00000646, 0.00000107, -0.00000767, -0.00000359, -0.00000653, 0.00000674, -0.00000834, 0.00000289, -0.00000331, -0.00000920, 0.00000336, -0.00001186, 0.00000397, 0.00000534, -0.00000863, 0.00001072, -0.00001292, 0.00001322, -0.00001037, 0.00001560, -0.00001133, 0.00000740, -0.00000666, 0.00000399, -0.00000687, 0.00001650, -0.00001111, 0.00000849, -0.00000934, 0.00000305, -0.00000390, 0.00000576, -0.00000598, 0.00000778, -0.00000916, 0.00000599, -0.00000392, 0.00001223, -0.00001092, 0.00000285, -0.00000585, 0.00001597, -0.00001121, 0.00000330, -0.00000044, 0.00000074, -0.00000310, 0.00000426, -0.00000050, -0.00000825, 0.00000308, 0.00000767, -0.00000973, -0.00000617, 0.00000146, 0.00000382, -0.00000346, 0.00000025, -0.00000493, 0.00000374, -0.00000659, 0.00000366, -0.00000729, -0.00000045, -0.00000227, -0.00000220, -0.00000088, 0.00000609, -0.00000715, 0.00000134, -0.00000577, },
		{ -0.00000346, -0.00000692, 0.00000400, -0.00000073, -0.00000467, -0.00000138, -0.00001234, 0.00000170, -0.00002416, 0.00000700, -0.00000808, 0.00002338, -0.00000277, 0.00000924, -0.00000628, 0.00000475, -0.00000026, 0.00000354, -0.00000766, 0.00000122, -0.00000382, 0.00000213, -0.00001217, 0.00000094, -0.00000034, -0.00000494, -0.00001579, -0.00000620, -0.00003005, -0.00001625, -0.00010618, -0.00016751, 0.00023221, 0.00011017, 0.00007322, 0.00005376, 0.00005494, 0.00005703, 0.00012180, 0.00024119, -0.00021300, -0.00008527, -0.00003539, -0.00002202, -0.00002529, -0.00001323, -0.00001363, 0.00000119, -0.00001572, -0.00000582, -0.00001311, 0.00000003, -0.00000855, 0.00000150, -0.00000934, 0.00000537, -0.00000622, -0.00000019, -0.00000948, 0.00000455, -0.00001174, -0.00000021, -0.00000867, 0.00000911, -0.00001099, 0.00000775, -0.00001122, -0.00000248, -0.00000025, -0.00001159, 0.00000917, -0.00000213, -0.00000095, 0.00000442, -0.00000880, 0.00001051, -0.00000772, 0.00001420, -0.00001248, 0.00000840, -0.00000683, 0.00000327, -0.00000396, 0.00001466, -0.00001217, 0.00000982, -0.00001166, 0.00000511, -0.00000406, 0.00000511, -0.00000513, 0.00000680, -0.00000896, 0.00000636, -0.00000241, 0.00001071, -0.00001224, 0.00000356, -0.00000380, 0.00001566, -0.00001541, 0.00000851, -0.00000463, 0.00000358, -0.00000504, 0.00000685, -0.00000482, -0.00000601, 0.00000596, 0.00000441, -0.00001144, -0.00000372, 0.00000329, 0.00000097, -0.00000211, -0.00000134, -0.00000239, 0.00000154, -0.00000496, 0.00000254, -0.00000676, -0.00000091, -0.00000188, -0.00000306, 0.00000264, 0.00000251, -0.00000553, -0.00000049, },
		{ -0.00000177, -0.00000225, -0.00000858, 0.00000852, -0.00000681, 0.00000173, -0.00000887, -0.00000398, -0.00000862, -0.00001473, 0.00000350, -0.00000311, 0.00002119, -0.00000418, 0.00000934, -0.00000645, 0.00000323, 0.00000146, 0.00000149, -0.00000794, 0.00000093, -0.00000257, -0.00000025, -0.00000748, -0.00000517, 0.00000345, -0.00000820, -0.00001268, -0.00001156, -0.00002650, -0.00002276, -0.00009848, -0.00018874, 0.00024686, 0.00011104, 0.00008392, 0.00005023, 0.00007462, 0.00004276, 0.00018950, 0.00016993, -0.00025978, -0.00003383, -0.00005731, -0.00000225, -0.00003968, 0.00000077, -0.00002566, 0.00001207, -0.00002314, 0.00000010, -0.00002022, 0.00000821, -0.00001523, 0.00000786, -0.00001650, 0.00001227, -0.00001031, 0.00000256, -0.00001309, 0.00000856, -0.00001665, 0.00000449, -0.00001194, 0.00001212, -0.00001373, 0.00001241, -0.00002016, 0.00000768, -0.00000934, -0.00000447, 0.00000768, -0.00000465, 0.00000381, -0.00000081, -0.00000453, 0.00000720, -0.00000402, 0.00001106, -0.00001211, 0.00000878, -0.00000699, 0.00000289, -0.00000078, 0.00001119, -0.00001123, 0.00000949, -0.00001282, 0.00000697, -0.00000454, 0.00000462, -0.00000411, 0.00000526, -0.00000797, 0.00000627, -0.00000041, 0.00000773, -0.00001195, 0.00000340, -0.00000024, 0.00001214, -0.00001625, 0.00001172, -0.00000805, 0.00000631, -0.00000687, 0.00000925, -0.00000995, -0.00000123, 0.00000640, 0.00000152, -0.00001284, 0.00000004, 0.00000301, -0.00000057, -0.00000161, -0.00000197, -0.00000056, -0.00000047, -0.00000320, 0.00000115, -0.00000611, -0.00000115, -0.00000211, -0.00000262, 0.00000487, -0.00000072, -0.00000401, },
		{ -0.00000361, -0.00000162, -0.00000312, -0.00000709, 0.00000936, -0.00000960, 0.00000523, -0.00001382, 0.00000248, -0.00001848, -0.00000316, -0.00000323, 0.00000493, 0.00001514, -0.00000241, 0.00000756, -0.00000539, 0.00000094, 0.00000393, -0.00000166, -0.00000717, 0.00000055, -0.00000200, -0.00000137, -0.00000397, -0.00001099, 0.00000742, -0.00001140, -0.00000995, -0.00001653, -0.00002360, -0.00002813, -0.00009185, -0.00020971, 0.00026072, 0.00011418, 0.00009236, 0.00005136, 0.00008908, 0.00003891, 0.00025740, 0.00005649, -0.00025570, -0.00000631, -0.00006809, 0.00000903, -0.00004745, 0.00000933, -0.00003350, 0.00001993, -0.00002820, 0.00000326, -0.00002416, 0.00001371, -0.00001981, 0.00001215, -0.00002188, 0.00001836, -0.00001417, 0.00000463, -0.00001546, 0.00001108, -0.00002026, 0.00000848, -0.00001460, 0.00001412, -0.00001473, 0.00001434, -0.00002666, 0.00001739, -0.00002043, 0.00000725, 0.00000005, -0.00000128, 0.00000382, -0.00000285, -0.00000213, 0.00000506, -0.00000082, 0.00000731, -0.00001072, 0.00000869, -0.00000732, 0.00000339, 0.00000162, 0.00000715, -0.00000891, 0.00000762, -0.00001256, 0.00000824, -0.00000520, 0.00000448, -0.00000336, 0.00000365, -0.00000661, 0.00000614, 0.00000140, 0.00000405, -0.00001052, 0.00000294, 0.00000389, 0.00000610, -0.00001345, 0.00001182, -0.00000935, 0.00000764, -0.00000745, 0.00001008, -0.00001422, 0.00000483, 0.00000464, -0.00000079, -0.00001361, 0.00000442, 0.00000083, -0.00000052, -0.00000235, -0.00000110, -0.00000019, -0.00000142, -0.00000212, 0.00000013, -0.00000579, -0.00000087, -0.00000322, -0.00000040, 0.00000485, -0.00000238, },
		{ -0.00000199, -0.00000475, 0.00000011, -0.00000606, -0.00000224, 0.00000568, -0.00000756, 0.00000370, -0.00001346, 0.00000349, -0.00002347, 0.00000596, -0.00000932, 0.00001288, 0.00000773, 0.00000091, 0.00000506, -0.00000420, -0.00000095, 0.00000581, -0.00000465, -0.00000628, 0.00000074, -0.00000286, -0.00000010, -0.00000340, -0.00001440, 0.00000948, -0.00001258, -0.00000960, -0.00001903, -0.00002355, -0.00002989, -0.00008932, -0.00022702, 0.00027010, 0.00012365, 0.00009416, 0.00006285, 0.00009041, 0.00006023, 0.00029920, -0.00007470, -0.00020741, -0.00000785, -0.00006167, 0.00000680, -0.00004416, 0.00000833, -0.00003328, 0.00002129, -0.00002804, 0.00000133, -0.00002263, 0.00001409, -0.00001988, 0.00001202, -0.00002297, 0.00002117, -0.00001578, 0.00000443, -0.00001515, 0.00001058, -0.00002084, 0.00001003, -0.00001515, 0.00001384, -0.00001274, 0.00001175, -0.00002775, 0.00002254, -0.00002888, 0.00001925, -0.00001064, 0.00000643, -0.00000067, -0.00000106, -0.00000263, 0.00000529, 0.00000052, 0.00000418, -0.00000904, 0.00000842, -0.00000794, 0.00000512, 0.00000234, 0.00000381, -0.00000630, 0.00000490, -0.00001110, 0.00000879, -0.00000595, 0.00000486, -0.00000330, 0.00000257, -0.00000541, 0.00000651, 0.00000218, 0.00000072, -0.00000884, 0.00000312, 0.00000704, -0.00000073, -0.00000806, 0.00000901, -0.00000816, 0.00000704, -0.00000617, 0.00000843, -0.00001606, 0.00001051, 0.00000147, -0.00000261, -0.00001329, 0.00000841, -0.00000242, 0.00000074, -0.00000410, 0.00000115, -0.00000147, -0.00000090, -0.00000219, -0.00000015, -0.00000598, -0.00000008, -0.00000496, 0.00000332, 0.00000244, },
		{ -0.00000154, -0.00000004, -0.00000691, 0.00000260, -0.00000983, 0.00000448, -0.00000133, -0.00000137, -0.00000269, -0.00000734, -0.00000229, -0.00002102, 0.00000911, -0.00001124, 0.00001746, 0.00000187, 0.00000367, 0.00000327, -0.00000402, -0.00000136, 0.00000595, -0.00000631, -0.00000606, 0.00000184, -0.00000524, 0.00000373, -0.00000637, -0.00001435, 0.00000854, -0.00001079, -0.00001253, -0.00001816, -0.00002709, -0.00002720, -0.00009207, -0.00023922, 0.00027344, 0.00014105, 0.00008798, 0.00008640, 0.00007613, 0.00011544, 0.00028984, -0.00019070, -0.00013574, -0.00003103, -0.00004105, -0.00000704, -0.00003098, -0.00000160, -0.00002511, 0.00001599, -0.00002270, -0.00000537, -0.00001595, 0.00000938, -0.00001532, 0.00000725, -0.00001925, 0.00001988, -0.00001448, 0.00000171, -0.00001211, 0.00000693, -0.00001801, 0.00000862, -0.00001324, 0.00001124, -0.00000799, 0.00000467, -0.00002242, 0.00002063, -0.00003085, 0.00002696, -0.00002008, 0.00001519, -0.00000755, 0.00000348, -0.00000565, 0.00000800, -0.00000061, 0.00000257, -0.00000774, 0.00000818, -0.00000874, 0.00000799, 0.00000092, 0.00000220, -0.00000452, 0.00000230, -0.00000900, 0.00000874, -0.00000670, 0.00000577, -0.00000413, 0.00000243, -0.00000481, 0.00000773, 0.00000131, -0.00000131, -0.00000774, 0.00000478, 0.00000766, -0.00000612, -0.00000213, 0.00000463, -0.00000525, 0.00000495, -0.00000341, 0.00000431, -0.00001462, 0.00001434, -0.00000211, -0.00000416, -0.00001151, 0.00001099, -0.00000559, 0.00000230, -0.00000606, 0.00000403, -0.00000384, 0.00000081, -0.00000332, 0.00000026, -0.00000648, 0.00000078, -0.00000656, 0.00000754, },
		{ 0.00001075, -0.00000557, 0.00000222, -0.00000889, 0.00000435, -0.00001233, 0.00001040, -0.00000877, 0.00000618, -0.00001128, 0.00000213, -0.00001304, -0.00001169, 0.00000544, -0.00000746, 0.00001683, -0.00000051, 0.00000442, 0.00000307, -0.00000533, -0.00000002, 0.00000399, -0.00000624, -0.00000663, 0.00000342, -0.00000822, 0.00000890, -0.00001183, -0.00001152, 0.00000520, -0.00000680, -0.00001778, -0.00001506, -0.00003278, -0.00002185, -0.00009832, -0.00024808, 0.00027267, 0.00016406, 0.00007733, 0.00011735, 0.00005298, 0.00020038, 0.00021567, -0.00026094, -0.00006789, -0.00005978, -0.00001657, -0.00002457, -0.00001435, -0.00001519, -0.00001321, 0.00000744, -0.00001524, -0.00001367, -0.00000715, 0.00000225, -0.00000855, 0.00000005, -0.00001249, 0.00001575, -0.00001140, -0.00000220, -0.00000788, 0.00000172, -0.00001312, 0.00000537, -0.00001000, 0.00000776, -0.00000245, -0.00000463, -0.00001242, 0.00001209, -0.00002503, 0.00002749, -0.00002461, 0.00002142, -0.00001385, 0.00000862, -0.00000968, 0.00001219, -0.00000388, 0.00000271, -0.00000713, 0.00000801, -0.00000937, 0.00001139, -0.00000238, 0.00000268, -0.00000437, 0.00000070, -0.00000694, 0.00000828, -0.00000733, 0.00000698, -0.00000575, 0.00000329, -0.00000488, 0.00000971, -0.00000128, -0.00000162, -0.00000755, 0.00000812, 0.00000488, -0.00000823, 0.00000216, 0.00000059, -0.00000209, 0.00000258, -0.00000036, -0.00000122, -0.00001010, 0.00001551, -0.00000530, -0.00000555, -0.00000826, 0.00001146, -0.00000761, 0.00000319, -0.00000722, 0.00000641, -0.00000628, 0.00000292, -0.00000492, 0.00000089, -0.00000679, 0.00000105, -0.00000699, },
		{ -0.00000542, 0.00001163, -0.00000807, 0.00000339, -0.00000944, 0.00000398, -0.00001163, 0.00001272, -0.00001323, 0.00001132, -0.00001800, 0.00001059, -0.00002439, 0.00000086, -0.00000277, 0.00000078, 0.00001148, 0.00000070, 0.00000307, 0.00000419, -0.00000752, 0.00000224, 0.00000073, -0.00000502, -0.00000730, 0.00000433, -0.00001005, 0.00001313, -0.00001733, -0.00000816, 0.00000161, -0.00000294, -0.00002282, -0.00001254, -0.00003730, -0.00001774, -0.00010383, -0.00025816, 0.00027278, 0.00018692, 0.00006991, 0.00014544, 0.00003685, 0.00029522, 0.00008224, -0.00026827, -0.00002701, -0.00007703, -0.00000067, -0.00003565, -0.00000284, -0.00002500, -0.00000389, 0.00000089, -0.00001030, -0.00001906, -0.00000069, -0.00000314, -0.00000342, -0.00000585, -0.00000607, 0.00001149, -0.00000883, -0.00000497, -0.00000500, -0.00000241, -0.00000865, 0.00000245, -0.00000752, 0.00000572, 0.00000090, -0.00001234, -0.00000179, 0.00000030, -0.00001336, 0.00002089, -0.00002274, 0.00002281, -0.00001711, 0.00001213, -0.00001283, 0.00001626, -0.00000820, 0.00000422, -0.00000719, 0.00000781, -0.00000932, 0.00001438, -0.00000666, 0.00000488, -0.00000596, 0.00000058, -0.00000543, 0.00000763, -0.00000765, 0.00000808, -0.00000768, 0.00000480, -0.00000525, 0.00001187, -0.00000500, -0.00000047, -0.00000792, 0.00001244, -0.00000096, -0.00000634, 0.00000331, -0.00000144, -0.00000020, 0.00000132, 0.00000143, -0.00000640, -0.00000371, 0.00001407, -0.00000770, -0.00000658, -0.00000399, 0.00000978, -0.00000796, 0.00000276, -0.00000680, 0.00000730, -0.00000771, 0.00000443, -0.00000624, 0.00000118, -0.00000643, 0.00000021, },
		{ -0.00000172, -0.00000180, 0.00000963, -0.00000814, 0.00000267, -0.00000799, 0.00000102, -0.00000696, 0.00000993, -0.00001240, 0.00001111, -0.00001922, 0.00001364, -0.00003118, 0.00001140, -0.00001112, 0.00001005, 0.00000394, 0.00000391, 0.00000082, 0.00000542, -0.00000922, 0.00000384, -0.00000227, -0.00000391, -0.00000700, 0.00000332, -0.00000898, 0.00001399, -0.00002009, -0.00000701, 0.00000043, -0.00000198, -0.00002470, -0.00001371, -0.00003707, -0.00001914, -0.00010379, -0.00027474, 0.00027948, 0.00020323, 0.00007415, 0.00015911, 0.00004774, 0.00036799, -0.00008437, -0.00021571, -0.00002347, -0.00007251, -0.00000162, -0.00003312, -0.00000290, -0.00002517, -0.00000235, 0.00000069, -0.00001159, -0.00001794, -0.00000023, -0.00000328, -0.00000334, -0.00000703, -0.00000323, 0.00000982, -0.00000901, -0.00000450, -0.00000569, -0.00000303, -0.00000701, 0.00000201, -0.00000773, 0.00000710, -0.00000053, -0.00001470, 0.00000470, -0.00000961, -0.00000035, 0.00001014, -0.00001568, 0.00001924, -0.00001635, 0.00001271, -0.00001367, 0.00001867, -0.00001224, 0.00000638, -0.00000767, 0.00000746, -0.00000822, 0.00001599, -0.00001068, 0.00000784, -0.00000877, 0.00000189, -0.00000471, 0.00000692, -0.00000749, 0.00000858, -0.00000926, 0.00000634, -0.00000534, 0.00001333, -0.00000882, 0.00000132, -0.00000797, 0.00001624, -0.00000827, -0.00000120, 0.00000113, -0.00000077, -0.00000041, 0.00000206, 0.00000071, -0.00000942, 0.00000280, 0.00001079, -0.00000938, -0.00000683, 0.00000036, 0.00000658, -0.00000679, 0.00000104, -0.00000470, 0.00000629, -0.00000748, 0.00000465, -0.00000669, 0.00000080, -0.00000527, },
		{ -0.00000372, -0.00000401, 0.00000297, 0.00000535, -0.00000601, 0.00000029, -0.00000511, -0.00000357, 0.00000058, 0.00000270, -0.00000631, 0.00000496, -0.00001366, 0.00000903, -0.00002989, 0.00001545, -0.00001517, 0.00001637, -0.00000247, 0.00000677, -0.00000070, 0.00000537, -0.00000905, 0.00000328, -0.00000352, -0.00000399, -0.00000506, -0.00000013, -0.00000432, 0.00001025, -0.00001840, -0.00000983, 0.00000329, -0.00000554, -0.00002178, -0.00002016, -0.00003033, -0.00002832, -0.00009549, -0.00030081, 0.00029587, 0.00020984, 0.00009442, 0.00015213, 0.00010078, 0.00038440, -0.00024407, -0.00012574, -0.00005136, -0.00004724, -0.00001932, -0.00001656, -0.00001537, -0.00001448, -0.00000990, 0.00000783, -0.00001970, -0.00000970, -0.00000662, 0.00000282, -0.00000936, -0.00000217, -0.00000547, 0.00001197, -0.00001269, -0.00000024, -0.00001058, 0.00000076, -0.00000924, 0.00000490, -0.00001121, 0.00001231, -0.00000750, -0.00000983, 0.00000366, -0.00001301, 0.00000881, -0.00000016, -0.00000678, 0.00001267, -0.00001245, 0.00001051, -0.00001185, 0.00001858, -0.00001493, 0.00000841, -0.00000829, 0.00000699, -0.00000601, 0.00001560, -0.00001328, 0.00001034, -0.00001183, 0.00000411, -0.00000473, 0.00000623, -0.00000678, 0.00000815, -0.00000994, 0.00000736, -0.00000463, 0.00001327, -0.00001163, 0.00000283, -0.00000672, 0.00001779, -0.00001473, 0.00000523, -0.00000324, 0.00000206, -0.00000250, 0.00000471, -0.00000282, -0.00000906, 0.00000769, 0.00000681, -0.00001066, -0.00000580, 0.00000369, 0.00000302, -0.00000493, -0.00000126, -0.00000156, 0.00000375, -0.00000568, 0.00000352, -0.00000622, -0.00000012, },
		{ -0.00000101, -0.00000257, -0.00000544, 0.00000727, 0.00000038, -0.00000295, -0.00000250, -0.00000233, -0.00000760, 0.00000825, -0.00000613, 0.00000243, -0.00000476, -0.00000329, -0.00000205, -0.00002038, 0.00001138, -0.00001251, 0.00001711, -0.00000529, 0.00000750, -0.00000041, 0.00000343, -0.00000659, 0.00000006, -0.00000249, -0.00000549, -0.00000186, -0.00000517, 0.00000288, 0.00000274, -0.00001269, -0.00001626, 0.00000965, -0.00001290, -0.00001498, -0.00003066, -0.00001859, -0.00004386, -0.00008026, -0.00033493, 0.00031995, 0.00020989, 0.00012703, 0.00012946, 0.00019661, 0.00032104, -0.00035560, -0.00003204, -0.00009198, -0.00001280, -0.00004513, 0.00000722, -0.00003478, 0.00000276, -0.00002311, 0.00001922, -0.00003147, 0.00000263, -0.00001725, 0.00001283, -0.00001940, 0.00000712, -0.00001170, 0.00001689, -0.00001857, 0.00000628, -0.00001814, 0.00000772, -0.00001435, 0.00001012, -0.00001664, 0.00001958, -0.00001811, 0.00000108, -0.00000520, -0.00000777, 0.00001047, -0.00000574, 0.00000006, 0.00000618, -0.00000753, 0.00000692, -0.00000818, 0.00001606, -0.00001576, 0.00000979, -0.00000888, 0.00000662, -0.00000310, 0.00001320, -0.00001381, 0.00001139, -0.00001415, 0.00000653, -0.00000527, 0.00000565, -0.00000567, 0.00000678, -0.00000949, 0.00000759, -0.00000297, 0.00001135, -0.00001272, 0.00000346, -0.00000371, 0.00001589, -0.00001817, 0.00001054, -0.00000780, 0.00000560, -0.00000532, 0.00000818, -0.00000836, -0.00000520, 0.00000983, 0.00000315, -0.00001182, -0.00000323, 0.00000508, 0.00000030, -0.00000348, -0.00000302, 0.00000143, 0.00000075, -0.00000312, 0.00000159, -0.00000530, },
		{ -0.00000464, -0.00000118, -0.00000263, -0.00000484, 0.00000936, -0.00000332, -0.00000070, -0.00000411, -0.00000148, -0.00000867, 0.00001274, -0.00001264, 0.00000961, -0.00001363, 0.00000733, -0.00001515, -0.00000621, 0.00000122, -0.00000393, 0.00001224, -0.00000402, 0.00000582, 0.00000145, 0.00000024, -0.00000277, -0.00000489, 0.00000006, -0.00000744, 0.00000111, -0.00000964, 0.00000990, -0.00000575, -0.00000554, -0.00002377, 0.00001676, -0.00002103, -0.00000764, -0.00004123, -0.00000649, -0.00006071, -0.00006352, -0.00037116, 0.00034454, 0.00021288, 0.00015984, 0.00010905, 0.00031565, 0.00017690, -0.00039111, 0.00003331, -0.00012264, 0.00001460, -0.00006605, 0.00002736, -0.00005186, 0.00001873, -0.00003542, 0.00002902, -0.00004129, 0.00001361, -0.00002711, 0.00002214, -0.00002909, 0.00001696, -0.00001873, 0.00002179, -0.00002377, 0.00001197, -0.00002521, 0.00001489, -0.00001973, 0.00001513, -0.00002125, 0.00002552, -0.00002828, 0.00001395, -0.00001869, 0.00000460, 0.00000409, -0.00000453, 0.00000211, 0.00000250, -0.00000394, 0.00000386, -0.00000421, 0.00001205, -0.00001490, 0.00001034, -0.00000939, 0.00000674, -0.00000026, 0.00000944, -0.00001235, 0.00001059, -0.00001501, 0.00000848, -0.00000603, 0.00000532, -0.00000451, 0.00000487, -0.00000815, 0.00000726, -0.00000076, 0.00000789, -0.00001209, 0.00000331, 0.00000064, 0.00001051, -0.00001741, 0.00001284, -0.00001062, 0.00000814, -0.00000730, 0.00001081, -0.00001416, 0.00000116, 0.00000901, 0.00000042, -0.00001287, 0.00000067, 0.00000416, -0.00000076, -0.00000331, -0.00000325, 0.00000303, -0.00000142, -0.00000094, -0.00000022, },
		{ -0.00000112, -0.00000477, -0.00000029, -0.00000419, -0.00000172, 0.00000820, -0.00000434, -0.00000058, -0.00000345, -0.00000356, -0.00000534, 0.00001173, -0.00001366, 0.00001143, -0.00001716, 0.00001295, -0.00002441, 0.00000681, -0.00001022, 0.00000699, 0.00000424, -0.00000023, 0.00000297, 0.00000349, -0.00000248, 0.00000045, -0.00000961, 0.00000242, -0.00000825, 0.00000193, -0.00001103, 0.00001352, -0.00001159, -0.00000052, -0.00002883, 0.00002089, -0.00002596, -0.00000406, -0.00004698, 0.00000021, -0.00007238, -0.00005238, -0.00040164, 0.00036031, 0.00023105, 0.00017678, 0.00011717, 0.00041992, -0.00002205, -0.00034695, 0.00005211, -0.00012706, 0.00002211, -0.00007120, 0.00003432, -0.00005810, 0.00002598, -0.00004050, 0.00003175, -0.00004388, 0.00001806, -0.00003127, 0.00002606, -0.00003385, 0.00002306, -0.00002295, 0.00002360, -0.00002534, 0.00001373, -0.00002848, 0.00001902, -0.00002244, 0.00001720, -0.00002221, 0.00002667, -0.00003347, 0.00002338, -0.00003133, 0.00001950, -0.00000755, 0.00000260, -0.00000113, 0.00000287, -0.00000317, 0.00000289, -0.00000153, 0.00000784, -0.00001297, 0.00001018, -0.00000985, 0.00000768, 0.00000158, 0.00000544, -0.00000965, 0.00000820, -0.00001424, 0.00000956, -0.00000682, 0.00000539, -0.00000379, 0.00000303, -0.00000649, 0.00000691, 0.00000115, 0.00000383, -0.00001038, 0.00000314, 0.00000496, 0.00000297, -0.00001280, 0.00001151, -0.00001065, 0.00000851, -0.00000726, 0.00001107, -0.00001820, 0.00000822, 0.00000591, -0.00000136, -0.00001345, 0.00000514, 0.00000133, -0.00000005, -0.00000467, -0.00000155, 0.00000254, -0.00000181, -0.00000011, },
		{ -0.00000094, -0.00000086, -0.00000570, 0.00000140, -0.00000671, 0.00000328, 0.00000402, -0.00000251, -0.00000274, -0.00000067, -0.00000806, 0.00000187, 0.00000520, -0.00000838, 0.00000636, -0.00001305, 0.00001024, -0.00002524, 0.00001321, -0.00001758, 0.00001559, -0.00000318, 0.00000348, 0.00000087, 0.00000400, -0.00000301, 0.00000113, -0.00001216, 0.00000303, -0.00000671, -0.00000056, -0.00000787, 0.00001163, -0.00001217, -0.00000032, -0.00002884, 0.00001940, -0.00002495, -0.00000704, -0.00004477, -0.00000235, -0.00007439, -0.00005182, -0.00042103, 0.00036097, 0.00027306, 0.00016603, 0.00017783, 0.00046352, -0.00022812, -0.00024564, 0.00002665, -0.00010264, 0.00000675, -0.00005747, 0.00002487, -0.00005018, 0.00002134, -0.00003573, 0.00002530, -0.00003723, 0.00001377, -0.00002747, 0.00002232, -0.00003126, 0.00002296, -0.00002229, 0.00002077, -0.00002196, 0.00001013, -0.00002625, 0.00001830, -0.00002086, 0.00001500, -0.00001826, 0.00002132, -0.00003077, 0.00002501, -0.00003760, 0.00003117, -0.00001958, 0.00001232, -0.00000786, 0.00000665, -0.00000530, 0.00000458, -0.00000120, 0.00000465, -0.00001077, 0.00000959, -0.00001026, 0.00000957, 0.00000171, 0.00000240, -0.00000689, 0.00000507, -0.00001221, 0.00000972, -0.00000750, 0.00000596, -0.00000391, 0.00000194, -0.00000517, 0.00000721, 0.00000182, 0.00000036, -0.00000857, 0.00000398, 0.00000747, -0.00000444, -0.00000612, 0.00000740, -0.00000809, 0.00000664, -0.00000496, 0.00000826, -0.00001894, 0.00001404, 0.00000171, -0.00000253, -0.00001302, 0.00000906, -0.00000238, 0.00000177, -0.00000701, 0.00000164, 0.00000010, -0.00000027, },
		{ 0.00000246, -0.00000293, 0.00000015, -0.00000685, 0.00000300, -0.00000886, 0.00000851, -0.00000161, 0.00000099, -0.00000604, 0.00000276, -0.00001290, 0.00001032, -0.00000425, 0.00000095, -0.00000374, -0.00000259, -0.00000042, -0.00001658, 0.00001037, -0.00001747, 0.00001842, -0.00000699, 0.00000498, 0.00000082, 0.00000216, -0.00000066, -0.00000146, -0.00001176, 0.00000149, -0.00000300, -0.00000573, -0.00000089, 0.00000464, -0.00000747, -0.00000503, -0.00002389, 0.00001254, -0.00001842, -0.00001584, -0.00003550, -0.00001344, -0.00006726, -0.00006125, -0.00043037, 0.00034793, 0.00033789, 0.00012890, 0.00029966, 0.00040905, -0.00038640, -0.00012802, -0.00002226, -0.00006142, -0.00002284, -0.00003140, 0.00000411, -0.00003199, 0.00000784, -0.00002385, 0.00001249, -0.00002399, 0.00000297, -0.00001762, 0.00001258, -0.00002249, 0.00001738, -0.00001748, 0.00001440, -0.00001495, 0.00000247, -0.00001947, 0.00001339, -0.00001571, 0.00000950, -0.00001075, 0.00001079, -0.00002062, 0.00001767, -0.00003453, 0.00003517, -0.00002719, 0.00002049, -0.00001498, 0.00001180, -0.00000910, 0.00000834, -0.00000337, 0.00000321, -0.00000900, 0.00000885, -0.00001051, 0.00001221, -0.00000014, 0.00000122, -0.00000522, 0.00000232, -0.00000966, 0.00000919, -0.00000802, 0.00000697, -0.00000504, 0.00000199, -0.00000464, 0.00000853, 0.00000061, -0.00000155, -0.00000746, 0.00000653, 0.00000665, -0.00000928, 0.00000012, 0.00000246, -0.00000433, 0.00000357, -0.00000128, 0.00000292, -0.00001585, 0.00001715, -0.00000233, -0.00000353, -0.00001111, 0.00001132, -0.00000565, 0.00000354, -0.00000919, 0.00000511, -0.00000326, },
		{ -0.00000607, 0.00000509, -0.00000505, 0.00000105, -0.00000732, 0.00000341, -0.00000911, 0.00001183, -0.00000642, 0.00000411, -0.00000855, 0.00000465, -0.00001523, 0.00001627, -0.00001233, 0.00000985, -0.00001426, 0.00000957, -0.00001471, -0.00000170, -0.00000009, -0.00001010, 0.00001489, -0.00000635, 0.00000383, 0.00000259, -0.00000128, 0.00000347, -0.00000623, -0.00000943, -0.00000096, 0.00000099, -0.00001102, 0.00000680, -0.00000430, -0.00000040, -0.00001172, -0.00001720, 0.00000387, -0.00001035, -0.00002580, -0.00002461, -0.00002716, -0.00005734, -0.00007357, -0.00043829, 0.00033189, 0.00041251, 0.00008436, 0.00046589, 0.00024410, -0.00045359, -0.00003793, -0.00006500, -0.00002408, -0.00005040, -0.00000639, -0.00001669, -0.00001299, -0.00000656, -0.00001194, 0.00000010, -0.00001059, -0.00000847, -0.00000711, 0.00000183, -0.00001194, 0.00000998, -0.00001186, 0.00000791, -0.00000794, -0.00000559, -0.00001155, 0.00000732, -0.00000989, 0.00000387, -0.00000338, -0.00000074, -0.00000699, 0.00000422, -0.00002308, 0.00003033, -0.00002778, 0.00002402, -0.00001960, 0.00001598, -0.00001274, 0.00001281, -0.00000742, 0.00000369, -0.00000814, 0.00000824, -0.00001042, 0.00001504, -0.00000364, 0.00000218, -0.00000541, 0.00000087, -0.00000742, 0.00000835, -0.00000833, 0.00000819, -0.00000695, 0.00000315, -0.00000491, 0.00001070, -0.00000244, -0.00000158, -0.00000724, 0.00001067, 0.00000207, -0.00000991, 0.00000367, -0.00000116, -0.00000119, 0.00000093, 0.00000209, -0.00000337, -0.00000967, 0.00001702, -0.00000538, -0.00000461, -0.00000764, 0.00001125, -0.00000737, 0.00000415, -0.00000998, 0.00000742, },
		{ 0.00000747, -0.00000703, 0.00000634, -0.00000622, 0.00000103, -0.00000643, 0.00000191, -0.00000643, 0.00001160, -0.00000839, 0.00000495, -0.00000859, 0.00000332, -0.00001280, 0.00001652, -0.00001493, 0.00001357, -0.00001982, 0.00001737, -0.00002608, 0.00001307, -0.00001304, 0.00000083, 0.00000746, -0.00000284, 0.00000146, 0.00000441, -0.00000413, 0.00000688, -0.00001061, -0.00000751, -0.00000190, 0.00000241, -0.00001287, 0.00001087, -0.00001046, 0.00000432, -0.00001565, -0.00001382, -0.00000115, -0.00000668, -0.00003013, -0.00001999, -0.00003465, -0.00005439, -0.00007776, -0.00045804, 0.00032931, 0.00047588, 0.00006593, 0.00063295, -0.00000901, -0.00041428, -0.00000545, -0.00007663, -0.00000988, -0.00005992, 0.00000380, -0.00002548, -0.00000366, -0.00001289, -0.00000790, -0.00000442, -0.00000426, -0.00001377, -0.00000229, -0.00000384, -0.00000523, 0.00000560, -0.00000965, 0.00000542, -0.00000520, -0.00000962, -0.00000677, 0.00000400, -0.00000712, 0.00000193, -0.00000058, -0.00000784, 0.00000405, -0.00000949, -0.00000790, 0.00001927, -0.00002192, 0.00002216, -0.00002034, 0.00001764, -0.00001469, 0.00001650, -0.00001225, 0.00000574, -0.00000833, 0.00000794, -0.00000984, 0.00001737, -0.00000795, 0.00000488, -0.00000751, 0.00000125, -0.00000613, 0.00000757, -0.00000839, 0.00000920, -0.00000911, 0.00000498, -0.00000551, 0.00001303, -0.00000662, -0.00000012, -0.00000737, 0.00001529, -0.00000533, -0.00000622, 0.00000345, -0.00000204, -0.00000013, 0.00000018, 0.00000340, -0.00000848, -0.00000211, 0.00001414, -0.00000720, -0.00000558, -0.00000313, 0.00000888, -0.00000709, 0.00000302, -0.00000861, },
		{ -0.00000523, 0.00000508, -0.00000561, 0.00000556, -0.00000587, -0.00000021, -0.00000418, -0.00000130, -0.00000098, 0.00000751, -0.00000672, 0.00000275, -0.00000568, -0.00000142, -0.00000537, 0.00001008, -0.00001018, 0.00000946, -0.00001688, 0.00001620, -0.00002870, 0.00002111, -0.00002214, 0.00000996, 0.00000032, 0.00000051, 0.00000025, 0.00000405, -0.00000399, 0.00000689, -0.00001188, -0.00000844, 0.00000079, -0.00000097, -0.00000856, 0.00000787, -0.00000981, 0.00000254, -0.00001266, -0.00001808, 0.00000203, -0.00001225, -0.00002339, -0.00002807, -0.00002850, -0.00006677, -0.00006440, -0.00050105, 0.00035462, 0.00050891, 0.00010979, 0.00074042, -0.00029554, -0.00028767, -0.00003557, -0.00004805, -0.00002713, -0.00004376, -0.00000796, -0.00001553, -0.00001012, -0.00000588, -0.00001622, 0.00000321, -0.00000927, -0.00000876, -0.00000722, -0.00000034, -0.00000630, 0.00000761, -0.00001364, 0.00000947, -0.00000936, -0.00000671, -0.00000813, 0.00000618, -0.00000986, 0.00000611, -0.00000531, -0.00000638, 0.00000704, -0.00001705, 0.00000462, 0.00000714, -0.00001295, 0.00001661, -0.00001772, 0.00001661, -0.00001431, 0.00001833, -0.00001668, 0.00000869, -0.00000948, 0.00000810, -0.00000870, 0.00001851, -0.00001197, 0.00000832, -0.00001095, 0.00000331, -0.00000603, 0.00000709, -0.00000814, 0.00000961, -0.00001085, 0.00000679, -0.00000579, 0.00001453, -0.00001076, 0.00000193, -0.00000689, 0.00001862, -0.00001338, 0.00000035, -0.00000009, -0.00000001, -0.00000156, 0.00000192, 0.00000160, -0.00001056, 0.00000475, 0.00000966, -0.00000810, -0.00000593, 0.00000144, 0.00000504, -0.00000521, 0.00000041, },
		{ -0.00000261, -0.00000092, 0.00000109, -0.00000237, 0.00000302, -0.00000427, -0.00000218, -0.00000138, -0.00000498, 0.00000572, 0.00000084, -0.00000231, -0.00000166, -0.00000100, -0.00000779, 0.00000486, -0.00000101, 0.00000039, -0.00000150, -0.00000575, 0.00000532, -0.00002031, 0.00001863, -0.00002285, 0.00001288, -0.00000285, 0.00000108, 0.00000190, 0.00000023, 0.00000030, 0.00000223, -0.00000873, -0.00001313, 0.00000752, -0.00000921, 0.00000196, -0.00000273, -0.00000134, -0.00000681, -0.00000182, -0.00003084, 0.00001420, -0.00002763, -0.00000500, -0.00004989, -0.00000719, -0.00009624, -0.00003161, -0.00056978, 0.00041113, 0.00050700, 0.00023745, 0.00073064, -0.00054339, -0.00012160, -0.00010706, 0.00000998, -0.00006872, -0.00000691, -0.00003810, 0.00001070, -0.00003065, 0.00001289, -0.00003518, 0.00002139, -0.00002445, 0.00000567, -0.00002125, 0.00001208, -0.00001527, 0.00001597, -0.00002336, 0.00001936, -0.00001979, 0.00000285, -0.00001562, 0.00001379, -0.00001778, 0.00001582, -0.00001716, 0.00000425, -0.00000030, -0.00001443, 0.00000920, -0.00000074, -0.00000526, 0.00001050, -0.00001370, 0.00001400, -0.00001202, 0.00001793, -0.00001982, 0.00001174, -0.00001124, 0.00000876, -0.00000709, 0.00001804, -0.00001466, 0.00001128, -0.00001463, 0.00000640, -0.00000695, 0.00000698, -0.00000758, 0.00000914, -0.00001163, 0.00000799, -0.00000517, 0.00001436, -0.00001371, 0.00000362, -0.00000490, 0.00001895, -0.00001950, 0.00000728, -0.00000516, 0.00000375, -0.00000466, 0.00000552, -0.00000316, -0.00000870, 0.00000917, 0.00000499, -0.00000863, -0.00000505, 0.00000484, 0.00000106, -0.00000282, },
		{ -0.00000125, -0.00000461, 0.00000270, -0.00000288, 0.00000133, -0.00000013, -0.00000237, -0.00000380, 0.00000059, -0.00000726, 0.00001123, -0.00000586, 0.00000259, -0.00000615, 0.00000303, -0.00001264, 0.00001386, -0.00001224, 0.00001222, -0.00001472, 0.00000912, -0.00001133, -0.00000363, 0.00000653, -0.00001462, 0.00000834, -0.00000075, -0.00000176, 0.00000623, -0.00000624, 0.00000760, -0.00000592, -0.00000230, -0.00002014, 0.00001613, -0.00001946, 0.00001537, -0.00001761, 0.00001188, -0.00002066, 0.00001344, -0.00004827, 0.00003107, -0.00004781, 0.00001926, -0.00007921, 0.00002262, -0.00013539, 0.00001183, -0.00065379, 0.00048548, 0.00048936, 0.00043951, 0.00057173, -0.00068672, 0.00002480, -0.00018221, 0.00007195, -0.00011505, 0.00003477, -0.00007346, 0.00004227, -0.00005605, 0.00003522, -0.00005703, 0.00004284, -0.00004320, 0.00002352, -0.00003886, 0.00002860, -0.00002808, 0.00002694, -0.00003492, 0.00003102, -0.00003242, 0.00001534, -0.00002596, 0.00002367, -0.00002749, 0.00002724, -0.00003198, 0.00002035, -0.00001560, -0.00000196, 0.00000405, -0.00000124, -0.00000221, 0.00000675, -0.00001053, 0.00001142, -0.00000887, 0.00001558, -0.00002113, 0.00001413, -0.00001309, 0.00000985, -0.00000525, 0.00001591, -0.00001534, 0.00001269, -0.00001736, 0.00000951, -0.00000840, 0.00000716, -0.00000680, 0.00000781, -0.00001122, 0.00000832, -0.00000353, 0.00001219, -0.00001476, 0.00000445, -0.00000122, 0.00001540, -0.00002163, 0.00001204, -0.00000948, 0.00000737, -0.00000780, 0.00000939, -0.00000949, -0.00000329, 0.00001025, 0.00000122, -0.00000926, -0.00000259, 0.00000610, -0.00000173, },
		{ -0.00000245, -0.00000147, -0.00000444, 0.00000411, -0.00000514, 0.00000382, -0.00000250, -0.00000134, -0.00000399, 0.00000050, -0.00000649, 0.00001325, -0.00000992, 0.00000544, -0.00000833, 0.00000395, -0.00001288, 0.00001747, -0.00001856, 0.00001963, -0.00002394, 0.00002086, -0.00002658, 0.00001456, -0.00000979, -0.00000121, -0.00000119, 0.00000487, -0.00000626, 0.00001098, -0.00001252, 0.00001464, -0.00001418, 0.00000416, -0.00002603, 0.00002261, -0.00002687, 0.00002600, -0.00003068, 0.00002373, -0.00003270, 0.00002639, -0.00006309, 0.00004467, -0.00006372, 0.00003897, -0.00010440, 0.00004807, -0.00016966, 0.00004868, -0.00073212, 0.00054973, 0.00049936, 0.00066817, 0.00027524, -0.00068974, 0.00010329, -0.00022358, 0.00010966, -0.00014313, 0.00006187, -0.00009727, 0.00006469, -0.00007384, 0.00005009, -0.00007143, 0.00005776, -0.00005640, 0.00003631, -0.00005206, 0.00004191, -0.00003834, 0.00003479, -0.00004271, 0.00003871, -0.00004144, 0.00002519, -0.00003402, 0.00003095, -0.00003401, 0.00003482, -0.00004345, 0.00003520, -0.00003264, 0.00001563, -0.00000837, 0.00000526, -0.00000466, 0.00000674, -0.00000960, 0.00001018, -0.00000605, 0.00001196, -0.00002049, 0.00001522, -0.00001446, 0.00001120, -0.00000354, 0.00001252, -0.00001394, 0.00001196, -0.00001822, 0.00001170, -0.00000977, 0.00000748, -0.00000600, 0.00000593, -0.00000986, 0.00000800, -0.00000130, 0.00000841, -0.00001398, 0.00000464, 0.00000340, 0.00000849, -0.00001910, 0.00001304, -0.00001122, 0.00000910, -0.00000929, 0.00001161, -0.00001530, 0.00000416, 0.00000818, -0.00000114, -0.00001004, 0.00000129, 0.00000484, },
		{ 0.00000154, -0.00000113, -0.00000359, -0.00000178, 0.00000265, -0.00000468, 0.00000399, -0.00000307, -0.00000190, -0.00000226, -0.00000202, -0.00000214, 0.00001063, -0.00000971, 0.00000462, -0.00000687, 0.00000064, -0.00000717, 0.00001347, -0.00001674, 0.00001853, -0.00002414, 0.00002339, -0.00003317, 0.00002639, -0.00002289, 0.00001116, -0.00001087, 0.00001035, -0.00000930, 0.00001296, -0.00001514, 0.00001758, -0.00001858, 0.00000686, -0.00002717, 0.00002303, -0.00002689, 0.00002839, -0.00003581, 0.00002789, -0.00003649, 0.00003018, -0.00006799, 0.00004714, -0.00006654, 0.00004388, -0.00011376, 0.00005597, -0.00018400, 0.00006095, -0.00078250, 0.00057247, 0.00059295, 0.00084397, -0.00009967, -0.00056140, 0.00009503, -0.00021073, 0.00010581, -0.00013791, 0.00006098, -0.00009742, 0.00006713, -0.00007461, 0.00004922, -0.00007066, 0.00005869, -0.00005695, 0.00003728, -0.00005428, 0.00004581, -0.00004061, 0.00003478, -0.00004231, 0.00003795, -0.00004214, 0.00002766, -0.00003541, 0.00003154, -0.00003331, 0.00003400, -0.00004586, 0.00004194, -0.00004393, 0.00003122, -0.00002250, 0.00001521, -0.00001079, 0.00000981, -0.00001093, 0.00001072, -0.00000440, 0.00000791, -0.00001816, 0.00001473, -0.00001485, 0.00001258, -0.00000233, 0.00000863, -0.00001102, 0.00000926, -0.00001692, 0.00001238, -0.00001057, 0.00000782, -0.00000545, 0.00000403, -0.00000807, 0.00000759, 0.00000064, 0.00000402, -0.00001205, 0.00000500, 0.00000736, 0.00000007, -0.00001294, 0.00001027, -0.00000985, 0.00000819, -0.00000819, 0.00001082, -0.00001849, 0.00001156, 0.00000399, -0.00000225, -0.00001063, 0.00000583, },
		{ 0.00000982, -0.00000260, 0.00000134, -0.00000680, 0.00000253, -0.00000115, -0.00000167, 0.00000176, -0.00000177, -0.00000383, 0.00000081, -0.00000603, 0.00000466, 0.00000417, -0.00000553, 0.00000039, -0.00000235, -0.00000578, 0.00000306, 0.00000299, -0.00000723, 0.00000872, -0.00001430, 0.00001447, -0.00002732, 0.00002653, -0.00002683, 0.00001692, -0.00001608, 0.00001230, -0.00000839, 0.00001012, -0.00001203, 0.00001418, -0.00001679, 0.00000382, -0.00002200, 0.00001604, -0.00001797, 0.00002039, -0.00003029, 0.00002148, -0.00002916, 0.00002189, -0.00005994, 0.00003542, -0.00005291, 0.00002980, -0.00010217, 0.00004044, -0.00017166, 0.00004042, -0.00079448, 0.00053616, 0.00081700, 0.00087717, -0.00045916, -0.00035459, 0.00001778, -0.00014963, 0.00006308, -0.00010034, 0.00003202, -0.00007307, 0.00004838, -0.00005731, 0.00003194, -0.00005405, 0.00004466, -0.00004373, 0.00002516, -0.00004398, 0.00003858, -0.00003351, 0.00002607, -0.00003312, 0.00002809, -0.00003353, 0.00002153, -0.00002900, 0.00002460, -0.00002469, 0.00002381, -0.00003719, 0.00003690, -0.00004414, 0.00003850, -0.00003232, 0.00002388, -0.00001730, 0.00001393, -0.00001338, 0.00001263, -0.00000428, 0.00000430, -0.00001484, 0.00001283, -0.00001410, 0.00001383, -0.00000206, 0.00000524, -0.00000768, 0.00000547, -0.00001390, 0.00001153, -0.00001062, 0.00000815, -0.00000546, 0.00000273, -0.00000652, 0.00000774, 0.00000132, 0.00000028, -0.00000997, 0.00000646, 0.00000883, -0.00000722, -0.00000551, 0.00000526, -0.00000623, 0.00000517, -0.00000480, 0.00000677, -0.00001779, 0.00001692, -0.00000087, -0.00000267, -0.00001039, },
		{ -0.00000868, 0.00001204, -0.00000606, 0.00000351, -0.00000953, 0.00000680, -0.00000564, 0.00000248, -0.00000174, 0.00000040, -0.00000597, 0.00000366, -0.00000944, 0.00001133, -0.00000351, 0.00000031, -0.00000507, 0.00000272, -0.00001203, 0.00001368, -0.00000953, 0.00000559, -0.00000554, 0.00000160, -0.00000261, -0.00001093, 0.00001501, -0.00002017, 0.00001406, -0.00001506, 0.00000981, -0.00000363, 0.00000319, -0.00000430, 0.00000557, -0.00000993, -0.00000346, -0.00001274, 0.00000453, -0.00000337, 0.00000517, -0.00001701, 0.00000746, -0.00001405, 0.00000525, -0.00004316, 0.00001447, -0.00002847, 0.00000274, -0.00007599, 0.00000852, -0.00014096, -0.00000300, -0.00078080, 0.00045387, 0.00118438, 0.00069822, -0.00070210, -0.00014897, -0.00008140, -0.00007066, 0.00000417, -0.00004836, -0.00001037, -0.00003627, 0.00001855, -0.00003086, 0.00000662, -0.00002936, 0.00002265, -0.00002302, 0.00000568, -0.00002612, 0.00002443, -0.00002098, 0.00001279, -0.00001946, 0.00001336, -0.00001942, 0.00001012, -0.00001798, 0.00001350, -0.00001189, 0.00000809, -0.00002060, 0.00002175, -0.00003286, 0.00003499, -0.00003424, 0.00002770, -0.00002130, 0.00001701, -0.00001560, 0.00001519, -0.00000582, 0.00000204, -0.00001160, 0.00001031, -0.00001251, 0.00001504, -0.00000314, 0.00000339, -0.00000531, 0.00000200, -0.00001029, 0.00000975, -0.00001012, 0.00000855, -0.00000624, 0.00000247, -0.00000573, 0.00000888, 0.00000006, -0.00000181, -0.00000851, 0.00000955, 0.00000654, -0.00001096, 0.00000045, 0.00000032, -0.00000220, 0.00000158, -0.00000054, 0.00000054, -0.00001319, 0.00001899, -0.00000502, -0.00000308, },
		{ -0.00000391, -0.00000525, 0.00001162, -0.00000749, 0.00000393, -0.00001013, 0.00000901, -0.00000864, 0.00000567, -0.00000466, 0.00000178, -0.00000663, 0.00000438, -0.00000980, 0.00001468, -0.00000889, 0.00000444, -0.00000851, 0.00000494, -0.00001400, 0.00001945, -0.00001791, 0.00001496, -0.00001698, 0.00001580, -0.00001992, 0.00000869, -0.00000238, -0.00000695, 0.00000537, -0.00001004, 0.00000541, 0.00000175, -0.00000383, 0.00000354, -0.00000353, -0.00000275, -0.00000992, -0.00000522, -0.00000471, 0.00000918, -0.00000906, -0.00000430, -0.00000550, -0.00000049, -0.00000960, -0.00002878, -0.00000316, -0.00000752, -0.00002138, -0.00005291, -0.00001965, -0.00011574, -0.00003995, -0.00078050, 0.00037765, 0.00165576, 0.00028705, -0.00075329, -0.00002317, -0.00014542, -0.00001579, -0.00003744, -0.00000982, -0.00004244, -0.00000742, -0.00000475, -0.00001094, -0.00001224, -0.00001015, 0.00000517, -0.00000639, -0.00001039, -0.00001054, 0.00001209, -0.00001101, 0.00000272, -0.00000914, 0.00000154, -0.00000722, 0.00000025, -0.00000885, 0.00000478, -0.00000186, -0.00000561, -0.00000374, 0.00000331, -0.00001515, 0.00002334, -0.00002882, 0.00002609, -0.00002186, 0.00001822, -0.00001690, 0.00001792, -0.00000910, 0.00000206, -0.00000983, 0.00000842, -0.00001090, 0.00001649, -0.00000588, 0.00000391, -0.00000516, 0.00000034, -0.00000753, 0.00000804, -0.00000959, 0.00000918, -0.00000785, 0.00000346, -0.00000587, 0.00001100, -0.00000321, -0.00000181, -0.00000788, 0.00001393, 0.00000048, -0.00000992, 0.00000289, -0.00000242, 0.00000031, -0.00000075, 0.00000263, -0.00000589, -0.00000587, 0.00001755, -0.00000762, },
		{ -0.00000847, -0.00000515, -0.00000038, 0.00000843, -0.00000625, 0.00000179, -0.00000758, 0.00000770, -0.00000833, 0.00000597, -0.00000516, 0.00000084, -0.00000443, 0.00000155, -0.00000538, 0.00001228, -0.00000897, 0.00000386, -0.00000716, 0.00000162, -0.00000858, 0.00001623, -0.00001692, 0.00001473, -0.00001844, 0.00002004, -0.00002813, 0.00002167, -0.00001642, 0.00000502, -0.00000278, -0.00000633, 0.00000403, 0.00000249, -0.00000516, 0.00000517, -0.00000650, -0.00000184, -0.00000885, -0.00000671, -0.00000349, 0.00001035, -0.00001212, -0.00000280, -0.00000630, -0.00000039, -0.00000982, -0.00003074, -0.00000184, -0.00000792, -0.00002228, -0.00005589, -0.00001752, -0.00012789, -0.00003020, -0.00084997, 0.00039197, 0.00213695, -0.00031018, -0.00059074, -0.00002632, -0.00013074, -0.00002002, -0.00003227, -0.00001020, -0.00004177, -0.00000632, -0.00000416, -0.00001294, -0.00001040, -0.00000986, 0.00000480, -0.00000566, -0.00001183, -0.00000768, 0.00001091, -0.00001202, 0.00000387, -0.00001008, 0.00000062, -0.00000473, -0.00000074, -0.00000849, 0.00000523, -0.00000186, -0.00000910, 0.00000431, -0.00000905, 0.00000040, 0.00001023, -0.00002048, 0.00002172, -0.00002049, 0.00001844, -0.00001778, 0.00002095, -0.00001432, 0.00000514, -0.00001085, 0.00000857, -0.00001034, 0.00001862, -0.00001038, 0.00000712, -0.00000798, 0.00000164, -0.00000691, 0.00000749, -0.00000965, 0.00001022, -0.00001020, 0.00000557, -0.00000676, 0.00001360, -0.00000791, 0.00000001, -0.00000766, 0.00001836, -0.00000794, -0.00000457, 0.00000123, -0.00000196, 0.00000018, -0.00000058, 0.00000311, -0.00001037, 0.00000223, 0.00001331, },
		{ 0.00000757, -0.00000792, -0.00000642, 0.00000509, 0.00000317, -0.00000267, -0.00000271, -0.00000196, 0.00000261, -0.00000403, 0.00000243, -0.00000232, -0.00000307, 0.00000093, -0.00000489, 0.00000386, 0.00000360, -0.00000267, -0.00000262, -0.00000009, -0.00000772, 0.00000454, 0.00000315, -0.00000482, 0.00000234, -0.00000639, 0.00000944, -0.00002072, 0.00002001, -0.00001858, 0.00000776, -0.00000354, -0.00000953, 0.00001046, -0.00000597, 0.00000394, -0.00000454, 0.00000204, -0.00001243, 0.00000477, -0.00002236, 0.00001388, -0.00000646, 0.00000347, -0.00002043, 0.00001335, -0.00002255, 0.00001404, -0.00005921, 0.00002976, -0.00004288, 0.00001542, -0.00010277, 0.00003590, -0.00020341, 0.00006065, -0.00104248, 0.00059140, 0.00249629, -0.00098487, -0.00025655, -0.00015974, -0.00002600, -0.00009528, 0.00003119, -0.00006042, 0.00000202, -0.00004262, 0.00002893, -0.00004443, 0.00001916, -0.00003531, 0.00002812, -0.00002714, 0.00000760, -0.00002360, 0.00002630, -0.00002863, 0.00002041, -0.00002641, 0.00001493, -0.00001641, 0.00001132, -0.00002072, 0.00001845, -0.00001575, 0.00000254, -0.00000285, -0.00000758, 0.00000543, 0.00000335, -0.00001532, 0.00001898, -0.00002022, 0.00001973, -0.00001950, 0.00002475, -0.00002149, 0.00001153, -0.00001538, 0.00001180, -0.00001171, 0.00002169, -0.00001632, 0.00001266, -0.00001365, 0.00000616, -0.00000910, 0.00000882, -0.00001078, 0.00001175, -0.00001300, 0.00000842, -0.00000798, 0.00001586, -0.00001298, 0.00000288, -0.00000704, 0.00002111, -0.00001629, 0.00000309, -0.00000342, 0.00000120, -0.00000239, 0.00000211, 0.00000034, -0.00001140, 0.00000910, },
		{ 0.00001327, 0.00000173, -0.00000660, -0.00000710, 0.00000998, -0.00000281, 0.00000207, -0.00000831, 0.00000538, -0.00000506, 0.00000336, -0.00000437, 0.00000340, -0.00000928, 0.00000838, -0.00001333, 0.00001589, -0.00000945, 0.00000851, -0.00001365, 0.00001104, -0.00002070, 0.00002226, -0.00001651, 0.00001536, -0.00001949, 0.00001696, -0.00001471, 0.00000258, 0.00000146, -0.00000515, -0.00000285, 0.00000675, -0.00002229, 0.00002634, -0.00002464, 0.00002429, -0.00002648, 0.00002308, -0.00003523, 0.00003108, -0.00005180, 0.00004692, -0.00004114, 0.00003811, -0.00005776, 0.00005391, -0.00006738, 0.00006230, -0.00011429, 0.00009173, -0.00011299, 0.00009309, -0.00019566, 0.00014368, -0.00034725, 0.00024176, -0.00138246, 0.00104769, 0.00259904, -0.00159137, 0.00015348, -0.00037576, 0.00014203, -0.00022331, 0.00013955, -0.00015022, 0.00008100, -0.00011004, 0.00008910, -0.00010031, 0.00007177, -0.00008247, 0.00007161, -0.00006778, 0.00004546, -0.00005642, 0.00005642, -0.00005867, 0.00004990, -0.00005572, 0.00004248, -0.00004069, 0.00003493, -0.00004380, 0.00004243, -0.00004152, 0.00002804, -0.00002554, 0.00001005, -0.00000425, 0.00000767, -0.00001795, 0.00002159, -0.00002376, 0.00002396, -0.00002322, 0.00002956, -0.00003002, 0.00002051, -0.00002305, 0.00001815, -0.00001523, 0.00002546, -0.00002285, 0.00001938, -0.00002106, 0.00001312, -0.00001381, 0.00001207, -0.00001304, 0.00001360, -0.00001581, 0.00001149, -0.00000901, 0.00001697, -0.00001726, 0.00000590, -0.00000534, 0.00002068, -0.00002192, 0.00001023, -0.00000878, 0.00000534, -0.00000597, 0.00000608, -0.00000489, -0.00000864, },
		{ -0.00000298, 0.00001420, -0.00000312, -0.00000516, -0.00000666, 0.00001312, -0.00000788, 0.00000628, -0.00001314, 0.00001221, -0.00001296, 0.00001161, -0.00001243, 0.00001018, -0.00001586, 0.00001564, -0.00002091, 0.00002713, -0.00002298, 0.00002090, -0.00002566, 0.00002243, -0.00003275, 0.00003885, -0.00003617, 0.00003658, -0.00004330, 0.00004378, -0.00004464, 0.00003483, -0.00002872, 0.00002047, -0.00002460, 0.00002622, -0.00004223, 0.00004837, -0.00004940, 0.00005118, -0.00005570, 0.00005160, -0.00006482, 0.00006381, -0.00008783, 0.00008754, -0.00008522, 0.00008322, -0.00010583, 0.00010572, -0.00012443, 0.00012353, -0.00018316, 0.00016961, -0.00020239, 0.00019343, -0.00031534, 0.00028383, -0.00053392, 0.00048355, -0.00184417, 0.00177712, 0.00234904, -0.00198713, 0.00051886, -0.00059464, 0.00031767, -0.00036110, 0.00025790, -0.00025040, 0.00017038, -0.00018728, 0.00015753, -0.00016331, 0.00013146, -0.00013671, 0.00012185, -0.00011516, 0.00009047, -0.00009610, 0.00009200, -0.00009308, 0.00008333, -0.00008914, 0.00007487, -0.00006983, 0.00006276, -0.00007042, 0.00006952, -0.00007115, 0.00005957, -0.00005699, 0.00003909, -0.00002635, 0.00002290, -0.00002901, 0.00003033, -0.00003167, 0.00003143, -0.00002892, 0.00003472, -0.00003843, 0.00003019, -0.00003216, 0.00002638, -0.00002012, 0.00002909, -0.00002858, 0.00002550, -0.00002827, 0.00002076, -0.00001976, 0.00001645, -0.00001596, 0.00001537, -0.00001808, 0.00001422, -0.00000949, 0.00001641, -0.00001987, 0.00000845, -0.00000248, 0.00001651, -0.00002304, 0.00001436, -0.00001248, 0.00000835, -0.00000862, 0.00000935, -0.00001073, },
		{ -0.00001504, 0.00000389, 0.00001220, -0.00000633, -0.00000406, -0.00000480, 0.00001369, -0.00001069, 0.00000839, -0.00001531, 0.00001616, -0.00001833, 0.00001784, -0.00001900, 0.00001550, -0.00002036, 0.00002006, -0.00002450, 0.00003362, -0.00003243, 0.00002987, -0.00003418, 0.00002958, -0.00003876, 0.00004805, -0.00004834, 0.00005037, -0.00005959, 0.00006348, -0.00006876, 0.00006405, -0.00005937, 0.00004862, -0.00004928, 0.00004755, -0.00006214, 0.00006875, -0.00007164, 0.00007526, -0.00008237, 0.00007757, -0.00009076, 0.00009169, -0.00011793, 0.00012184, -0.00012370, 0.00012311, -0.00014813, 0.00015106, -0.00017452, 0.00017689, -0.00024263, 0.00023726, -0.00028170, 0.00028341, -0.00042401, 0.00041185, -0.00070932, 0.00071719, -0.00234467, 0.00272049, 0.00172365, -0.00207481, 0.00073221, -0.00073298, 0.00043694, -0.00045663, 0.00034245, -0.00032316, 0.00023719, -0.00024547, 0.00020854, -0.00020993, 0.00017639, -0.00017781, 0.00016003, -0.00015166, 0.00012629, -0.00012781, 0.00011947, -0.00011896, 0.00010811, -0.00011425, 0.00010013, -0.00009260, 0.00008417, -0.00009017, 0.00008904, -0.00009329, 0.00008523, -0.00008547, 0.00006894, -0.00005252, 0.00004312, -0.00004441, 0.00004222, -0.00004159, 0.00004014, -0.00003495, 0.00003847, -0.00004434, 0.00003773, -0.00003986, 0.00003411, -0.00002469, 0.00003117, -0.00003186, 0.00002902, -0.00003301, 0.00002676, -0.00002498, 0.00002055, -0.00001864, 0.00001644, -0.00001922, 0.00001614, -0.00000927, 0.00001409, -0.00002043, 0.00001040, 0.00000088, 0.00000936, -0.00001940, 0.00001428, -0.00001300, 0.00000874, -0.00000879, 0.00001011, },
		{ 0.00000756, -0.00001622, 0.00001014, 0.00000815, -0.00000773, -0.00000352, -0.00000159, 0.00001145, -0.00001046, 0.00000742, -0.00001365, 0.00001558, -0.00001901, 0.00001956, -0.00002154, 0.00001704, -0.00002067, 0.00001964, -0.00002188, 0.00003245, -0.00003420, 0.00003162, -0.00003555, 0.00002909, -0.00003526, 0.00004553, -0.00004763, 0.00005027, -0.00006071, 0.00006686, -0.00007613, 0.00007783, -0.00007760, 0.00006709, -0.00006582, 0.00006082, -0.00007264, 0.00007798, -0.00008131, 0.00008575, -0.00009516, 0.00008942, -0.00010127, 0.00010229, -0.00012860, 0.00013487, -0.00014023, 0.00014038, -0.00016622, 0.00017013, -0.00019626, 0.00019911, -0.00026688, 0.00026534, -0.00031746, 0.00032467, -0.00047696, 0.00047383, -0.00080547, 0.00085043, -0.00275685, 0.00374521, 0.00078934, -0.00183218, 0.00073287, -0.00073327, 0.00045173, -0.00046873, 0.00035726, -0.00033661, 0.00025296, -0.00025930, 0.00021953, -0.00021947, 0.00018721, -0.00018768, 0.00016916, -0.00016116, 0.00013776, -0.00013769, 0.00012627, -0.00012460, 0.00011292, -0.00011988, 0.00010733, -0.00009866, 0.00008946, -0.00009376, 0.00009161, -0.00009773, 0.00009367, -0.00009870, 0.00008726, -0.00007157, 0.00005915, -0.00005689, 0.00005155, -0.00004887, 0.00004635, -0.00003843, 0.00003841, -0.00004516, 0.00004015, -0.00004304, 0.00003850, -0.00002683, 0.00003022, -0.00003127, 0.00002830, -0.00003334, 0.00002895, -0.00002744, 0.00002272, -0.00001995, 0.00001612, -0.00001872, 0.00001691, -0.00000846, 0.00001038, -0.00001904, 0.00001198, 0.00000346, 0.00000116, -0.00001240, 0.00001047, -0.00001035, 0.00000632, -0.00000615, },
		{ -0.00000178, 0.00000233, -0.00001387, 0.00001446, 0.00000304, -0.00000743, -0.00000353, 0.00000254, 0.00000677, -0.00000730, 0.00000338, -0.00000811, 0.00001010, -0.00001419, 0.00001555, -0.00001865, 0.00001357, -0.00001590, 0.00001384, -0.00001272, 0.00002293, -0.00002700, 0.00002459, -0.00002838, 0.00002013, -0.00002189, 0.00003081, -0.00003296, 0.00003461, -0.00004417, 0.00005018, -0.00006116, 0.00006841, -0.00007410, 0.00006617, -0.00006500, 0.00005770, -0.00006624, 0.00006888, -0.00007109, 0.00007488, -0.00008578, 0.00007877, -0.00008806, 0.00008721, -0.00011090, 0.00011659, -0.00012356, 0.00012286, -0.00014712, 0.00014892, -0.00017448, 0.00017376, -0.00023765, 0.00023281, -0.00028503, 0.00028822, -0.00043940, 0.00042622, -0.00076433, 0.00079631, -0.00294120, 0.00467183, -0.00030921, -0.00131931, 0.00052459, -0.00058360, 0.00034774, -0.00038299, 0.00028846, -0.00027742, 0.00020507, -0.00021722, 0.00018020, -0.00018246, 0.00015493, -0.00015773, 0.00014107, -0.00013563, 0.00011698, -0.00011850, 0.00010607, -0.00010439, 0.00009251, -0.00010071, 0.00009107, -0.00008289, 0.00007403, -0.00007707, 0.00007322, -0.00007984, 0.00007887, -0.00008896, 0.00008490, -0.00007406, 0.00006240, -0.00005920, 0.00005233, -0.00004858, 0.00004607, -0.00003644, 0.00003255, -0.00003909, 0.00003537, -0.00003936, 0.00003725, -0.00002484, 0.00002525, -0.00002614, 0.00002269, -0.00002836, 0.00002607, -0.00002569, 0.00002167, -0.00001898, 0.00001393, -0.00001628, 0.00001631, -0.00000728, 0.00000591, -0.00001612, 0.00001354, 0.00000397, -0.00000564, -0.00000458, 0.00000486, -0.00000595, 0.00000232, },
		{ -0.00000120, 0.00000226, -0.00000370, -0.00000888, 0.00001641, -0.00000233, -0.00000564, -0.00000402, 0.00000699, 0.00000057, -0.00000201, -0.00000279, 0.00000018, 0.00000087, -0.00000473, 0.00000630, -0.00001059, 0.00000543, -0.00000669, 0.00000384, 0.00000121, 0.00000696, -0.00001233, 0.00001005, -0.00001411, 0.00000480, -0.00000174, 0.00000758, -0.00000825, 0.00000744, -0.00001402, 0.00001705, -0.00002623, 0.00003614, -0.00004715, 0.00004288, -0.00004351, 0.00003542, -0.00004097, 0.00004013, -0.00003991, 0.00004150, -0.00005285, 0.00004430, -0.00005024, 0.00004604, -0.00006468, 0.00006660, -0.00007267, 0.00006917, -0.00008931, 0.00008573, -0.00010726, 0.00009885, -0.00015276, 0.00013659, -0.00017979, 0.00016747, -0.00030199, 0.00025490, -0.00056227, 0.00050587, -0.00278757, 0.00531861, -0.00138413, -0.00066054, 0.00017149, -0.00032047, 0.00014963, -0.00021746, 0.00014990, -0.00015641, 0.00010209, -0.00012642, 0.00009712, -0.00010496, 0.00008492, -0.00009275, 0.00008003, -0.00007867, 0.00006685, -0.00007283, 0.00006170, -0.00006147, 0.00005015, -0.00005979, 0.00005394, -0.00004773, 0.00004054, -0.00004321, 0.00003738, -0.00004292, 0.00004305, -0.00005655, 0.00005993, -0.00005642, 0.00004875, -0.00004740, 0.00004116, -0.00003788, 0.00003699, -0.00002753, 0.00002040, -0.00002610, 0.00002325, -0.00002830, 0.00002955, -0.00001817, 0.00001637, -0.00001701, 0.00001287, -0.00001861, 0.00001824, -0.00001947, 0.00001696, -0.00001538, 0.00000978, -0.00001191, 0.00001432, -0.00000602, 0.00000138, -0.00001215, 0.00001513, 0.00000170, -0.00000901, 0.00000139, -0.00000004, -0.00000202, },
		{ -0.00000046, -0.00000228, 0.00000375, -0.00000813, -0.00000303, 0.00001635, -0.00000748, -0.00000261, -0.00000489, 0.00001109, -0.00000602, 0.00000413, -0.00000958, 0.00000925, -0.00000988, 0.00000707, -0.00000608, 0.00000084, -0.00000564, 0.00000494, -0.00000782, 0.00001656, -0.00001162, 0.00000599, -0.00000838, 0.00000363, -0.00001265, 0.00001978, -0.00001761, 0.00001913, -0.00002299, 0.00002056, -0.00002265, 0.00001873, -0.00001019, -0.00000373, 0.00000257, -0.00000582, -0.00000164, -0.00000186, -0.00000250, 0.00000579, -0.00000763, -0.00000319, -0.00000691, 0.00000440, -0.00001244, 0.00000022, -0.00000461, 0.00000166, -0.00000953, -0.00000465, -0.00000675, -0.00000824, -0.00001056, -0.00002903, -0.00000491, -0.00002163, -0.00001604, -0.00008833, -0.00001451, -0.00022543, -0.00000724, -0.00225331, 0.00555077, -0.00225331, -0.00000724, -0.00022543, -0.00001451, -0.00008833, -0.00001604, -0.00002163, -0.00000491, -0.00002903, -0.00001056, -0.00000824, -0.00000675, -0.00000465, -0.00000953, 0.00000166, -0.00000461, 0.00000022, -0.00001244, 0.00000440, -0.00000691, -0.00000319, -0.00000763, 0.00000579, -0.00000250, -0.00000186, -0.00000164, -0.00000582, 0.00000257, -0.00000373, -0.00001019, 0.00001873, -0.00002265, 0.00002056, -0.00002299, 0.00001913, -0.00001761, 0.00001978, -0.00001265, 0.00000363, -0.00000838, 0.00000599, -0.00001162, 0.00001656, -0.00000782, 0.00000494, -0.00000564, 0.00000084, -0.00000608, 0.00000707, -0.00000988, 0.00000925, -0.00000958, 0.00000413, -0.00000602, 0.00001109, -0.00000489, -0.00000261, -0.00000748, 0.00001635, -0.00000303, -0.00000813, 0.00000375, -0.00000228, },
		{ -0.00000120, -0.00000202, -0.00000004, 0.00000139, -0.00000901, 0.00000170, 0.00001513, -0.00001215, 0.00000138, -0.00000602, 0.00001432, -0.00001191, 0.00000978, -0.00001538, 0.00001696, -0.00001947, 0.00001824, -0.00001861, 0.00001287, -0.00001701, 0.00001637, -0.00001817, 0.00002955, -0.00002830, 0.00002325, -0.00002610, 0.00002040, -0.00002753, 0.00003699, -0.00003788, 0.00004116, -0.00004740, 0.00004875, -0.00005642, 0.00005993, -0.00005655, 0.00004305, -0.00004292, 0.00003738, -0.00004321, 0.00004054, -0.00004773, 0.00005394, -0.00005979, 0.00005015, -0.00006147, 0.00006170, -0.00007283, 0.00006685, -0.00007867, 0.00008003, -0.00009275, 0.00008492, -0.00010496, 0.00009712, -0.00012642, 0.00010209, -0.00015641, 0.00014990, -0.00021746, 0.00014963, -0.00032047, 0.00017149, -0.00066054, -0.00138413, 0.00531861, -0.00278757, 0.00050587, -0.00056227, 0.00025490, -0.00030199, 0.00016747, -0.00017979, 0.00013659, -0.00015276, 0.00009885, -0.00010726, 0.00008573, -0.00008931, 0.00006917, -0.00007267, 0.00006660, -0.00006468, 0.00004604, -0.00005024, 0.00004430, -0.00005285, 0.00004150, -0.00003991, 0.00004013, -0.00004097, 0.00003542, -0.00004351, 0.00004288, -0.00004715, 0.00003614, -0.00002623, 0.00001705, -0.00001402, 0.00000744, -0.00000825, 0.00000758, -0.00000174, 0.00000480, -0.00001411, 0.00001005, -0.00001233, 0.00000696, 0.00000121, 0.00000384, -0.00000669, 0.00000543, -0.00001059, 0.00000630, -0.00000473, 0.00000087, 0.00000018, -0.00000279, -0.00000201, 0.00000057, 0.00000699, -0.00000402, -0.00000564, -0.00000233, 0.00001641, -0.00000888, -0.00000370, 0.00000226, },
		{ -0.00000178, 0.00000232, -0.00000595, 0.00000486, -0.00000458, -0.00000564, 0.00000397, 0.00001354, -0.00001612, 0.00000591, -0.00000728, 0.00001631, -0.00001628, 0.00001393, -0.00001898, 0.00002167, -0.00002569, 0.00002607, -0.00002836, 0.00002269, -0.00002614, 0.00002525, -0.00002484, 0.00003725, -0.00003936, 0.00003537, -0.00003909, 0.00003255, -0.00003644, 0.00004607, -0.00004858, 0.00005233, -0.00005920, 0.00006240, -0.00007406, 0.00008490, -0.00008896, 0.00007887, -0.00007984, 0.00007322, -0.00007707, 0.00007403, -0.00008289, 0.00009107, -0.00010071, 0.00009251, -0.00010439, 0.00010607, -0.00011850, 0.00011698, -0.00013563, 0.00014107, -0.00015773, 0.00015493, -0.00018246, 0.00018020, -0.00021722, 0.00020507, -0.00027742, 0.00028846, -0.00038299, 0.00034774, -0.00058360, 0.00052459, -0.00131931, -0.00030921, 0.00467183, -0.00294120, 0.00079631, -0.00076433, 0.00042622, -0.00043940, 0.00028822, -0.00028503, 0.00023281, -0.00023765, 0.00017376, -0.00017448, 0.00014892, -0.00014712, 0.00012286, -0.00012356, 0.00011659, -0.00011090, 0.00008721, -0.00008806, 0.00007877, -0.00008578, 0.00007488, -0.00007109, 0.00006888, -0.00006624, 0.00005770, -0.00006500, 0.00006617, -0.00007410, 0.00006841, -0.00006116, 0.00005018, -0.00004417, 0.00003461, -0.00003296, 0.00003081, -0.00002189, 0.00002013, -0.00002838, 0.00002459, -0.00002700, 0.00002293, -0.00001272, 0.00001384, -0.00001590, 0.00001357, -0.00001865, 0.00001555, -0.00001419, 0.00001010, -0.00000811, 0.00000338, -0.00000730, 0.00000677, 0.00000254, -0.00000353, -0.00000743, 0.00000304, 0.00001446, -0.00001387, 0.00000233, },
		{ 0.00000756, -0.00000615, 0.00000632, -0.00001035, 0.00001047, -0.00001240, 0.00000116, 0.00000346, 0.00001198, -0.00001904, 0.00001038, -0.00000846, 0.00001691, -0.00001872, 0.00001612, -0.00001995, 0.00002272, -0.00002744, 0.00002895, -0.00003334, 0.00002830, -0.00003127, 0.00003022, -0.00002683, 0.00003850, -0.00004304, 0.00004015, -0.00004516, 0.00003841, -0.00003843, 0.00004635, -0.00004887, 0.00005155, -0.00005689, 0.00005915, -0.00007157, 0.00008726, -0.00009870, 0.00009367, -0.00009773, 0.00009161, -0.00009376, 0.00008946, -0.00009866, 0.00010733, -0.00011988, 0.00011292, -0.00012460, 0.00012627, -0.00013769, 0.00013776, -0.00016116, 0.00016916, -0.00018768, 0.00018721, -0.00021947, 0.00021953, -0.00025930, 0.00025296, -0.00033661, 0.00035726, -0.00046873, 0.00045173, -0.00073327, 0.00073287, -0.00183218, 0.00078934, 0.00374521, -0.00275685, 0.00085043, -0.00080547, 0.00047383, -0.00047696, 0.00032467, -0.00031746, 0.00026534, -0.00026688, 0.00019911, -0.00019626, 0.00017013, -0.00016622, 0.00014038, -0.00014023, 0.00013487, -0.00012860, 0.00010229, -0.00010127, 0.00008942, -0.00009516, 0.00008575, -0.00008131, 0.00007798, -0.00007264, 0.00006082, -0.00006582, 0.00006709, -0.00007760, 0.00007783, -0.00007613, 0.00006686, -0.00006071, 0.00005027, -0.00004763, 0.00004553, -0.00003526, 0.00002909, -0.00003555, 0.00003162, -0.00003420, 0.00003245, -0.00002188, 0.00001964, -0.00002067, 0.00001704, -0.00002154, 0.00001956, -0.00001901, 0.00001558, -0.00001365, 0.00000742, -0.00001046, 0.00001145, -0.00000159, -0.00000352, -0.00000773, 0.00000815, 0.00001014, -0.00001622, },
		{ -0.00001504, 0.00001011, -0.00000879, 0.00000874, -0.00001300, 0.00001428, -0.00001940, 0.00000936, 0.00000088, 0.00001040, -0.00002043, 0.00001409, -0.00000927, 0.00001614, -0.00001922, 0.00001644, -0.00001864, 0.00002055, -0.00002498, 0.00002676, -0.00003301, 0.00002902, -0.00003186, 0.00003117, -0.00002469, 0.00003411, -0.00003986, 0.00003773, -0.00004434, 0.00003847, -0.00003495, 0.00004014, -0.00004159, 0.00004222, -0.00004441, 0.00004312, -0.00005252, 0.00006894, -0.00008547, 0.00008523, -0.00009329, 0.00008904, -0.00009017, 0.00008417, -0.00009260, 0.00010013, -0.00011425, 0.00010811, -0.00011896, 0.00011947, -0.00012781, 0.00012629, -0.00015166, 0.00016003, -0.00017781, 0.00017639, -0.00020993, 0.00020854, -0.00024547, 0.00023719, -0.00032316, 0.00034245, -0.00045663, 0.00043694, -0.00073298, 0.00073221, -0.00207481, 0.00172365, 0.00272049, -0.00234467, 0.00071719, -0.00070932, 0.00041185, -0.00042401, 0.00028341, -0.00028170, 0.00023726, -0.00024263, 0.00017689, -0.00017452, 0.00015106, -0.00014813, 0.00012311, -0.00012370, 0.00012184, -0.00011793, 0.00009169, -0.00009076, 0.00007757, -0.00008237, 0.00007526, -0.00007164, 0.00006875, -0.00006214, 0.00004755, -0.00004928, 0.00004862, -0.00005937, 0.00006405, -0.00006876, 0.00006348, -0.00005959, 0.00005037, -0.00004834, 0.00004805, -0.00003876, 0.00002958, -0.00003418, 0.00002987, -0.00003243, 0.00003362, -0.00002450, 0.00002006, -0.00002036, 0.00001550, -0.00001900, 0.00001784, -0.00001833, 0.00001616, -0.00001531, 0.00000839, -0.00001069, 0.00001369, -0.00000480, -0.00000406, -0.00000633, 0.00001220, 0.00000389, },
		{ -0.00000298, -0.00001073, 0.00000935, -0.00000862, 0.00000835, -0.00001248, 0.00001436, -0.00002304, 0.00001651, -0.00000248, 0.00000845, -0.00001987, 0.00001641, -0.00000949, 0.00001422, -0.00001808, 0.00001537, -0.00001596, 0.00001645, -0.00001976, 0.00002076, -0.00002827, 0.00002550, -0.00002858, 0.00002909, -0.00002012, 0.00002638, -0.00003216, 0.00003019, -0.00003843, 0.00003472, -0.00002892, 0.00003143, -0.00003167, 0.00003033, -0.00002901, 0.00002290, -0.00002635, 0.00003909, -0.00005699, 0.00005957, -0.00007115, 0.00006952, -0.00007042, 0.00006276, -0.00006983, 0.00007487, -0.00008914, 0.00008333, -0.00009308, 0.00009200, -0.00009610, 0.00009047, -0.00011516, 0.00012185, -0.00013671, 0.00013146, -0.00016331, 0.00015753, -0.00018728, 0.00017038, -0.00025040, 0.00025790, -0.00036110, 0.00031767, -0.00059464, 0.00051886, -0.00198713, 0.00234904, 0.00177712, -0.00184417, 0.00048355, -0.00053392, 0.00028383, -0.00031534, 0.00019343, -0.00020239, 0.00016961, -0.00018316, 0.00012353, -0.00012443, 0.00010572, -0.00010583, 0.00008322, -0.00008522, 0.00008754, -0.00008783, 0.00006381, -0.00006482, 0.00005160, -0.00005570, 0.00005118, -0.00004940, 0.00004837, -0.00004223, 0.00002622, -0.00002460, 0.00002047, -0.00002872, 0.00003483, -0.00004464, 0.00004378, -0.00004330, 0.00003658, -0.00003617, 0.00003885, -0.00003275, 0.00002243, -0.00002566, 0.00002090, -0.00002298, 0.00002713, -0.00002091, 0.00001564, -0.00001586, 0.00001018, -0.00001243, 0.00001161, -0.00001296, 0.00001221, -0.00001314, 0.00000628, -0.00000788, 0.00001312, -0.00000666, -0.00000516, -0.00000312, 0.00001420, },
		{ 0.00001327, -0.00000864, -0.00000489, 0.00000608, -0.00000597, 0.00000534, -0.00000878, 0.00001023, -0.00002192, 0.00002068, -0.00000534, 0.00000590, -0.00001726, 0.00001697, -0.00000901, 0.00001149, -0.00001581, 0.00001360, -0.00001304, 0.00001207, -0.00001381, 0.00001312, -0.00002106, 0.00001938, -0.00002285, 0.00002546, -0.00001523, 0.00001815, -0.00002305, 0.00002051, -0.00003002, 0.00002956, -0.00002322, 0.00002396, -0.00002376, 0.00002159, -0.00001795, 0.00000767, -0.00000425, 0.00001005, -0.00002554, 0.00002804, -0.00004152, 0.00004243, -0.00004380, 0.00003493, -0.00004069, 0.00004248, -0.00005572, 0.00004990, -0.00005867, 0.00005642, -0.00005642, 0.00004546, -0.00006778, 0.00007161, -0.00008247, 0.00007177, -0.00010031, 0.00008910, -0.00011004, 0.00008100, -0.00015022, 0.00013955, -0.00022331, 0.00014203, -0.00037576, 0.00015348, -0.00159137, 0.00259904, 0.00104769, -0.00138246, 0.00024176, -0.00034725, 0.00014368, -0.00019566, 0.00009309, -0.00011299, 0.00009173, -0.00011429, 0.00006230, -0.00006738, 0.00005391, -0.00005776, 0.00003811, -0.00004114, 0.00004692, -0.00005180, 0.00003108, -0.00003523, 0.00002308, -0.00002648, 0.00002429, -0.00002464, 0.00002634, -0.00002229, 0.00000675, -0.00000285, -0.00000515, 0.00000146, 0.00000258, -0.00001471, 0.00001696, -0.00001949, 0.00001536, -0.00001651, 0.00002226, -0.00002070, 0.00001104, -0.00001365, 0.00000851, -0.00000945, 0.00001589, -0.00001333, 0.00000838, -0.00000928, 0.00000340, -0.00000437, 0.00000336, -0.00000506, 0.00000538, -0.00000831, 0.00000207, -0.00000281, 0.00000998, -0.00000710, -0.00000660, 0.00000173, },
		{ 0.00000757, 0.00000910, -0.00001140, 0.00000034, 0.00000211, -0.00000239, 0.00000120, -0.00000342, 0.00000309, -0.00001629, 0.00002111, -0.00000704, 0.00000288, -0.00001298, 0.00001586, -0.00000798, 0.00000842, -0.00001300, 0.00001175, -0.00001078, 0.00000882, -0.00000910, 0.00000616, -0.00001365, 0.00001266, -0.00001632, 0.00002169, -0.00001171, 0.00001180, -0.00001538, 0.00001153, -0.00002149, 0.00002475, -0.00001950, 0.00001973, -0.00002022, 0.00001898, -0.00001532, 0.00000335, 0.00000543, -0.00000758, -0.00000285, 0.00000254, -0.00001575, 0.00001845, -0.00002072, 0.00001132, -0.00001641, 0.00001493, -0.00002641, 0.00002041, -0.00002863, 0.00002630, -0.00002360, 0.00000760, -0.00002714, 0.00002812, -0.00003531, 0.00001916, -0.00004443, 0.00002893, -0.00004262, 0.00000202, -0.00006042, 0.00003119, -0.00009528, -0.00002600, -0.00015974, -0.00025655, -0.00098487, 0.00249629, 0.00059140, -0.00104248, 0.00006065, -0.00020341, 0.00003590, -0.00010277, 0.00001542, -0.00004288, 0.00002976, -0.00005921, 0.00001404, -0.00002255, 0.00001335, -0.00002043, 0.00000347, -0.00000646, 0.00001388, -0.00002236, 0.00000477, -0.00001243, 0.00000204, -0.00000454, 0.00000394, -0.00000597, 0.00001046, -0.00000953, -0.00000354, 0.00000776, -0.00001858, 0.00002001, -0.00002072, 0.00000944, -0.00000639, 0.00000234, -0.00000482, 0.00000315, 0.00000454, -0.00000772, -0.00000009, -0.00000262, -0.00000267, 0.00000360, 0.00000386, -0.00000489, 0.00000093, -0.00000307, -0.00000232, 0.00000243, -0.00000403, 0.00000261, -0.00000196, -0.00000271, -0.00000267, 0.00000317, 0.00000509, -0.00000642, -0.00000792, },
		{ -0.00000847, 0.00001331, 0.00000223, -0.00001037, 0.00000311, -0.00000058, 0.00000018, -0.00000196, 0.00000123, -0.00000457, -0.00000794, 0.00001836, -0.00000766, 0.00000001, -0.00000791, 0.00001360, -0.00000676, 0.00000557, -0.00001020, 0.00001022, -0.00000965, 0.00000749, -0.00000691, 0.00000164, -0.00000798, 0.00000712, -0.00001038, 0.00001862, -0.00001034, 0.00000857, -0.00001085, 0.00000514, -0.00001432, 0.00002095, -0.00001778, 0.00001844, -0.00002049, 0.00002172, -0.00002048, 0.00001023, 0.00000040, -0.00000905, 0.00000431, -0.00000910, -0.00000186, 0.00000523, -0.00000849, -0.00000074, -0.00000473, 0.00000062, -0.00001008, 0.00000387, -0.00001202, 0.00001091, -0.00000768, -0.00001183, -0.00000566, 0.00000480, -0.00000986, -0.00001040, -0.00001294, -0.00000416, -0.00000632, -0.00004177, -0.00001020, -0.00003227, -0.00002002, -0.00013074, -0.00002632, -0.00059074, -0.00031018, 0.00213695, 0.00039197, -0.00084997, -0.00003020, -0.00012789, -0.00001752, -0.00005589, -0.00002228, -0.00000792, -0.00000184, -0.00003074, -0.00000982, -0.00000039, -0.00000630, -0.00000280, -0.00001212, 0.00001035, -0.00000349, -0.00000671, -0.00000885, -0.00000184, -0.00000650, 0.00000517, -0.00000516, 0.00000249, 0.00000403, -0.00000633, -0.00000278, 0.00000502, -0.00001642, 0.00002167, -0.00002813, 0.00002004, -0.00001844, 0.00001473, -0.00001692, 0.00001623, -0.00000858, 0.00000162, -0.00000716, 0.00000386, -0.00000897, 0.00001228, -0.00000538, 0.00000155, -0.00000443, 0.00000084, -0.00000516, 0.00000597, -0.00000833, 0.00000770, -0.00000758, 0.00000179, -0.00000625, 0.00000843, -0.00000038, -0.00000515, },
		{ -0.00000391, -0.00000762, 0.00001755, -0.00000587, -0.00000589, 0.00000263, -0.00000075, 0.00000031, -0.00000242, 0.00000289, -0.00000992, 0.00000048, 0.00001393, -0.00000788, -0.00000181, -0.00000321, 0.00001100, -0.00000587, 0.00000346, -0.00000785, 0.00000918, -0.00000959, 0.00000804, -0.00000753, 0.00000034, -0.00000516, 0.00000391, -0.00000588, 0.00001649, -0.00001090, 0.00000842, -0.00000983, 0.00000206, -0.00000910, 0.00001792, -0.00001690, 0.00001822, -0.00002186, 0.00002609, -0.00002882, 0.00002334, -0.00001515, 0.00000331, -0.00000374, -0.00000561, -0.00000186, 0.00000478, -0.00000885, 0.00000025, -0.00000722, 0.00000154, -0.00000914, 0.00000272, -0.00001101, 0.00001209, -0.00001054, -0.00001039, -0.00000639, 0.00000517, -0.00001015, -0.00001224, -0.00001094, -0.00000475, -0.00000742, -0.00004244, -0.00000982, -0.00003744, -0.00001579, -0.00014542, -0.00002317, -0.00075329, 0.00028705, 0.00165576, 0.00037765, -0.00078050, -0.00003995, -0.00011574, -0.00001965, -0.00005291, -0.00002138, -0.00000752, -0.00000316, -0.00002878, -0.00000960, -0.00000049, -0.00000550, -0.00000430, -0.00000906, 0.00000918, -0.00000471, -0.00000522, -0.00000992, -0.00000275, -0.00000353, 0.00000354, -0.00000383, 0.00000175, 0.00000541, -0.00001004, 0.00000537, -0.00000695, -0.00000238, 0.00000869, -0.00001992, 0.00001580, -0.00001698, 0.00001496, -0.00001791, 0.00001945, -0.00001400, 0.00000494, -0.00000851, 0.00000444, -0.00000889, 0.00001468, -0.00000980, 0.00000438, -0.00000663, 0.00000178, -0.00000466, 0.00000567, -0.00000864, 0.00000901, -0.00001013, 0.00000393, -0.00000749, 0.00001162, -0.00000525, },
		{ -0.00000868, -0.00000308, -0.00000502, 0.00001899, -0.00001319, 0.00000054, -0.00000054, 0.00000158, -0.00000220, 0.00000032, 0.00000045, -0.00001096, 0.00000654, 0.00000955, -0.00000851, -0.00000181, 0.00000006, 0.00000888, -0.00000573, 0.00000247, -0.00000624, 0.00000855, -0.00001012, 0.00000975, -0.00001029, 0.00000200, -0.00000531, 0.00000339, -0.00000314, 0.00001504, -0.00001251, 0.00001031, -0.00001160, 0.00000204, -0.00000582, 0.00001519, -0.00001560, 0.00001701, -0.00002130, 0.00002770, -0.00003424, 0.00003499, -0.00003286, 0.00002175, -0.00002060, 0.00000809, -0.00001189, 0.00001350, -0.00001798, 0.00001012, -0.00001942, 0.00001336, -0.00001946, 0.00001279, -0.00002098, 0.00002443, -0.00002612, 0.00000568, -0.00002302, 0.00002265, -0.00002936, 0.00000662, -0.00003086, 0.00001855, -0.00003627, -0.00001037, -0.00004836, 0.00000417, -0.00007066, -0.00008140, -0.00014897, -0.00070210, 0.00069822, 0.00118438, 0.00045387, -0.00078080, -0.00000300, -0.00014096, 0.00000852, -0.00007599, 0.00000274, -0.00002847, 0.00001447, -0.00004316, 0.00000525, -0.00001405, 0.00000746, -0.00001701, 0.00000517, -0.00000337, 0.00000453, -0.00001274, -0.00000346, -0.00000993, 0.00000557, -0.00000430, 0.00000319, -0.00000363, 0.00000981, -0.00001506, 0.00001406, -0.00002017, 0.00001501, -0.00001093, -0.00000261, 0.00000160, -0.00000554, 0.00000559, -0.00000953, 0.00001368, -0.00001203, 0.00000272, -0.00000507, 0.00000031, -0.00000351, 0.00001133, -0.00000944, 0.00000366, -0.00000597, 0.00000040, -0.00000174, 0.00000248, -0.00000564, 0.00000680, -0.00000953, 0.00000351, -0.00000606, 0.00001204, },
		{ 0.00000982, -0.00001039, -0.00000267, -0.00000087, 0.00001692, -0.00001779, 0.00000677, -0.00000480, 0.00000517, -0.00000623, 0.00000526, -0.00000551, -0.00000722, 0.00000883, 0.00000646, -0.00000997, 0.00000028, 0.00000132, 0.00000774, -0.00000652, 0.00000273, -0.00000546, 0.00000815, -0.00001062, 0.00001153, -0.00001390, 0.00000547, -0.00000768, 0.00000524, -0.00000206, 0.00001383, -0.00001410, 0.00001283, -0.00001484, 0.00000430, -0.00000428, 0.00001263, -0.00001338, 0.00001393, -0.00001730, 0.00002388, -0.00003232, 0.00003850, -0.00004414, 0.00003690, -0.00003719, 0.00002381, -0.00002469, 0.00002460, -0.00002900, 0.00002153, -0.00003353, 0.00002809, -0.00003312, 0.00002607, -0.00003351, 0.00003858, -0.00004398, 0.00002516, -0.00004373, 0.00004466, -0.00005405, 0.00003194, -0.00005731, 0.00004838, -0.00007307, 0.00003202, -0.00010034, 0.00006308, -0.00014963, 0.00001778, -0.00035459, -0.00045916, 0.00087717, 0.00081700, 0.00053616, -0.00079448, 0.00004042, -0.00017166, 0.00004044, -0.00010217, 0.00002980, -0.00005291, 0.00003542, -0.00005994, 0.00002189, -0.00002916, 0.00002148, -0.00003029, 0.00002039, -0.00001797, 0.00001604, -0.00002200, 0.00000382, -0.00001679, 0.00001418, -0.00001203, 0.00001012, -0.00000839, 0.00001230, -0.00001608, 0.00001692, -0.00002683, 0.00002653, -0.00002732, 0.00001447, -0.00001430, 0.00000872, -0.00000723, 0.00000299, 0.00000306, -0.00000578, -0.00000235, 0.00000039, -0.00000553, 0.00000417, 0.00000466, -0.00000603, 0.00000081, -0.00000383, -0.00000177, 0.00000176, -0.00000167, -0.00000115, 0.00000253, -0.00000680, 0.00000134, -0.00000260, },
		{ 0.00000154, 0.00000583, -0.00001063, -0.00000225, 0.00000399, 0.00001156, -0.00001849, 0.00001082, -0.00000819, 0.00000819, -0.00000985, 0.00001027, -0.00001294, 0.00000007, 0.00000736, 0.00000500, -0.00001205, 0.00000402, 0.00000064, 0.00000759, -0.00000807, 0.00000403, -0.00000545, 0.00000782, -0.00001057, 0.00001238, -0.00001692, 0.00000926, -0.00001102, 0.00000863, -0.00000233, 0.00001258, -0.00001485, 0.00001473, -0.00001816, 0.00000791, -0.00000440, 0.00001072, -0.00001093, 0.00000981, -0.00001079, 0.00001521, -0.00002250, 0.00003122, -0.00004393, 0.00004194, -0.00004586, 0.00003400, -0.00003331, 0.00003154, -0.00003541, 0.00002766, -0.00004214, 0.00003795, -0.00004231, 0.00003478, -0.00004061, 0.00004581, -0.00005428, 0.00003728, -0.00005695, 0.00005869, -0.00007066, 0.00004922, -0.00007461, 0.00006713, -0.00009742, 0.00006098, -0.00013791, 0.00010581, -0.00021073, 0.00009503, -0.00056140, -0.00009967, 0.00084397, 0.00059295, 0.00057247, -0.00078250, 0.00006095, -0.00018400, 0.00005597, -0.00011376, 0.00004388, -0.00006654, 0.00004714, -0.00006799, 0.00003018, -0.00003649, 0.00002789, -0.00003581, 0.00002839, -0.00002689, 0.00002303, -0.00002717, 0.00000686, -0.00001858, 0.00001758, -0.00001514, 0.00001296, -0.00000930, 0.00001035, -0.00001087, 0.00001116, -0.00002289, 0.00002639, -0.00003317, 0.00002339, -0.00002414, 0.00001853, -0.00001674, 0.00001347, -0.00000717, 0.00000064, -0.00000687, 0.00000462, -0.00000971, 0.00001063, -0.00000214, -0.00000202, -0.00000226, -0.00000190, -0.00000307, 0.00000399, -0.00000468, 0.00000265, -0.00000178, -0.00000359, -0.00000113, },
		{ -0.00000245, 0.00000484, 0.00000129, -0.00001004, -0.00000114, 0.00000818, 0.00000416, -0.00001530, 0.00001161, -0.00000929, 0.00000910, -0.00001122, 0.00001304, -0.00001910, 0.00000849, 0.00000340, 0.00000464, -0.00001398, 0.00000841, -0.00000130, 0.00000800, -0.00000986, 0.00000593, -0.00000600, 0.00000748, -0.00000977, 0.00001170, -0.00001822, 0.00001196, -0.00001394, 0.00001252, -0.00000354, 0.00001120, -0.00001446, 0.00001522, -0.00002049, 0.00001196, -0.00000605, 0.00001018, -0.00000960, 0.00000674, -0.00000466, 0.00000526, -0.00000837, 0.00001563, -0.00003264, 0.00003520, -0.00004345, 0.00003482, -0.00003401, 0.00003095, -0.00003402, 0.00002519, -0.00004144, 0.00003871, -0.00004271, 0.00003479, -0.00003834, 0.00004191, -0.00005206, 0.00003631, -0.00005640, 0.00005776, -0.00007143, 0.00005009, -0.00007384, 0.00006469, -0.00009727, 0.00006187, -0.00014313, 0.00010966, -0.00022358, 0.00010329, -0.00068974, 0.00027524, 0.00066817, 0.00049936, 0.00054973, -0.00073212, 0.00004868, -0.00016966, 0.00004807, -0.00010440, 0.00003897, -0.00006372, 0.00004467, -0.00006309, 0.00002639, -0.00003270, 0.00002373, -0.00003068, 0.00002600, -0.00002687, 0.00002261, -0.00002603, 0.00000416, -0.00001418, 0.00001464, -0.00001252, 0.00001098, -0.00000626, 0.00000487, -0.00000119, -0.00000121, -0.00000979, 0.00001456, -0.00002658, 0.00002086, -0.00002394, 0.00001963, -0.00001856, 0.00001747, -0.00001288, 0.00000395, -0.00000833, 0.00000544, -0.00000992, 0.00001325, -0.00000649, 0.00000050, -0.00000399, -0.00000134, -0.00000250, 0.00000382, -0.00000514, 0.00000411, -0.00000444, -0.00000147, },
		{ -0.00000125, -0.00000173, 0.00000610, -0.00000259, -0.00000926, 0.00000122, 0.00001025, -0.00000329, -0.00000949, 0.00000939, -0.00000780, 0.00000737, -0.00000948, 0.00001204, -0.00002163, 0.00001540, -0.00000122, 0.00000445, -0.00001476, 0.00001219, -0.00000353, 0.00000832, -0.00001122, 0.00000781, -0.00000680, 0.00000716, -0.00000840, 0.00000951, -0.00001736, 0.00001269, -0.00001534, 0.00001591, -0.00000525, 0.00000985, -0.00001309, 0.00001413, -0.00002113, 0.00001558, -0.00000887, 0.00001142, -0.00001053, 0.00000675, -0.00000221, -0.00000124, 0.00000405, -0.00000196, -0.00001560, 0.00002035, -0.00003198, 0.00002724, -0.00002749, 0.00002367, -0.00002596, 0.00001534, -0.00003242, 0.00003102, -0.00003492, 0.00002694, -0.00002808, 0.00002860, -0.00003886, 0.00002352, -0.00004320, 0.00004284, -0.00005703, 0.00003522, -0.00005605, 0.00004227, -0.00007346, 0.00003477, -0.00011505, 0.00007195, -0.00018221, 0.00002480, -0.00068672, 0.00057173, 0.00043951, 0.00048936, 0.00048548, -0.00065379, 0.00001183, -0.00013539, 0.00002262, -0.00007921, 0.00001926, -0.00004781, 0.00003107, -0.00004827, 0.00001344, -0.00002066, 0.00001188, -0.00001761, 0.00001537, -0.00001946, 0.00001613, -0.00002014, -0.00000230, -0.00000592, 0.00000760, -0.00000624, 0.00000623, -0.00000176, -0.00000075, 0.00000834, -0.00001462, 0.00000653, -0.00000363, -0.00001133, 0.00000912, -0.00001472, 0.00001222, -0.00001224, 0.00001386, -0.00001264, 0.00000303, -0.00000615, 0.00000259, -0.00000586, 0.00001123, -0.00000726, 0.00000059, -0.00000380, -0.00000237, -0.00000013, 0.00000133, -0.00000288, 0.00000270, -0.00000461, },
		{ -0.00000261, -0.00000282, 0.00000106, 0.00000484, -0.00000505, -0.00000863, 0.00000499, 0.00000917, -0.00000870, -0.00000316, 0.00000552, -0.00000466, 0.00000375, -0.00000516, 0.00000728, -0.00001950, 0.00001895, -0.00000490, 0.00000362, -0.00001371, 0.00001436, -0.00000517, 0.00000799, -0.00001163, 0.00000914, -0.00000758, 0.00000698, -0.00000695, 0.00000640, -0.00001463, 0.00001128, -0.00001466, 0.00001804, -0.00000709, 0.00000876, -0.00001124, 0.00001174, -0.00001982, 0.00001793, -0.00001202, 0.00001400, -0.00001370, 0.00001050, -0.00000526, -0.00000074, 0.00000920, -0.00001443, -0.00000030, 0.00000425, -0.00001716, 0.00001582, -0.00001778, 0.00001379, -0.00001562, 0.00000285, -0.00001979, 0.00001936, -0.00002336, 0.00001597, -0.00001527, 0.00001208, -0.00002125, 0.00000567, -0.00002445, 0.00002139, -0.00003518, 0.00001289, -0.00003065, 0.00001070, -0.00003810, -0.00000691, -0.00006872, 0.00000998, -0.00010706, -0.00012160, -0.00054339, 0.00073064, 0.00023745, 0.00050700, 0.00041113, -0.00056978, -0.00003161, -0.00009624, -0.00000719, -0.00004989, -0.00000500, -0.00002763, 0.00001420, -0.00003084, -0.00000182, -0.00000681, -0.00000134, -0.00000273, 0.00000196, -0.00000921, 0.00000752, -0.00001313, -0.00000873, 0.00000223, 0.00000030, 0.00000023, 0.00000190, 0.00000108, -0.00000285, 0.00001288, -0.00002285, 0.00001863, -0.00002031, 0.00000532, -0.00000575, -0.00000150, 0.00000039, -0.00000101, 0.00000486, -0.00000779, -0.00000100, -0.00000166, -0.00000231, 0.00000084, 0.00000572, -0.00000498, -0.00000138, -0.00000218, -0.00000427, 0.00000302, -0.00000237, 0.00000109, -0.00000092, },
		{ -0.00000523, 0.00000041, -0.00000521, 0.00000504, 0.00000144, -0.00000593, -0.00000810, 0.00000966, 0.00000475, -0.00001056, 0.00000160, 0.00000192, -0.00000156, -0.00000001, -0.00000009, 0.00000035, -0.00001338, 0.00001862, -0.00000689, 0.00000193, -0.00001076, 0.00001453, -0.00000579, 0.00000679, -0.00001085, 0.00000961, -0.00000814, 0.00000709, -0.00000603, 0.00000331, -0.00001095, 0.00000832, -0.00001197, 0.00001851, -0.00000870, 0.00000810, -0.00000948, 0.00000869, -0.00001668, 0.00001833, -0.00001431, 0.00001661, -0.00001772, 0.00001661, -0.00001295, 0.00000714, 0.00000462, -0.00001705, 0.00000704, -0.00000638, -0.00000531, 0.00000611, -0.00000986, 0.00000618, -0.00000813, -0.00000671, -0.00000936, 0.00000947, -0.00001364, 0.00000761, -0.00000630, -0.00000034, -0.00000722, -0.00000876, -0.00000927, 0.00000321, -0.00001622, -0.00000588, -0.00001012, -0.00001553, -0.00000796, -0.00004376, -0.00002713, -0.00004805, -0.00003557, -0.00028767, -0.00029554, 0.00074042, 0.00010979, 0.00050891, 0.00035462, -0.00050105, -0.00006440, -0.00006677, -0.00002850, -0.00002807, -0.00002339, -0.00001225, 0.00000203, -0.00001808, -0.00001266, 0.00000254, -0.00000981, 0.00000787, -0.00000856, -0.00000097, 0.00000079, -0.00000844, -0.00001188, 0.00000689, -0.00000399, 0.00000405, 0.00000025, 0.00000051, 0.00000032, 0.00000996, -0.00002214, 0.00002111, -0.00002870, 0.00001620, -0.00001688, 0.00000946, -0.00001018, 0.00001008, -0.00000537, -0.00000142, -0.00000568, 0.00000275, -0.00000672, 0.00000751, -0.00000098, -0.00000130, -0.00000418, -0.00000021, -0.00000587, 0.00000556, -0.00000561, 0.00000508, },
		{ 0.00000747, -0.00000861, 0.00000302, -0.00000709, 0.00000888, -0.00000313, -0.00000558, -0.00000720, 0.00001414, -0.00000211, -0.00000848, 0.00000340, 0.00000018, -0.00000013, -0.00000204, 0.00000345, -0.00000622, -0.00000533, 0.00001529, -0.00000737, -0.00000012, -0.00000662, 0.00001303, -0.00000551, 0.00000498, -0.00000911, 0.00000920, -0.00000839, 0.00000757, -0.00000613, 0.00000125, -0.00000751, 0.00000488, -0.00000795, 0.00001737, -0.00000984, 0.00000794, -0.00000833, 0.00000574, -0.00001225, 0.00001650, -0.00001469, 0.00001764, -0.00002034, 0.00002216, -0.00002192, 0.00001927, -0.00000790, -0.00000949, 0.00000405, -0.00000784, -0.00000058, 0.00000193, -0.00000712, 0.00000400, -0.00000677, -0.00000962, -0.00000520, 0.00000542, -0.00000965, 0.00000560, -0.00000523, -0.00000384, -0.00000229, -0.00001377, -0.00000426, -0.00000442, -0.00000790, -0.00001289, -0.00000366, -0.00002548, 0.00000380, -0.00005992, -0.00000988, -0.00007663, -0.00000545, -0.00041428, -0.00000901, 0.00063295, 0.00006593, 0.00047588, 0.00032931, -0.00045804, -0.00007776, -0.00005439, -0.00003465, -0.00001999, -0.00003013, -0.00000668, -0.00000115, -0.00001382, -0.00001565, 0.00000432, -0.00001046, 0.00001087, -0.00001287, 0.00000241, -0.00000190, -0.00000751, -0.00001061, 0.00000688, -0.00000413, 0.00000441, 0.00000146, -0.00000284, 0.00000746, 0.00000083, -0.00001304, 0.00001307, -0.00002608, 0.00001737, -0.00001982, 0.00001357, -0.00001493, 0.00001652, -0.00001280, 0.00000332, -0.00000859, 0.00000495, -0.00000839, 0.00001160, -0.00000643, 0.00000191, -0.00000643, 0.00000103, -0.00000622, 0.00000634, -0.00000703, },
		{ -0.00000607, 0.00000742, -0.00000998, 0.00000415, -0.00000737, 0.00001125, -0.00000764, -0.00000461, -0.00000538, 0.00001702, -0.00000967, -0.00000337, 0.00000209, 0.00000093, -0.00000119, -0.00000116, 0.00000367, -0.00000991, 0.00000207, 0.00001067, -0.00000724, -0.00000158, -0.00000244, 0.00001070, -0.00000491, 0.00000315, -0.00000695, 0.00000819, -0.00000833, 0.00000835, -0.00000742, 0.00000087, -0.00000541, 0.00000218, -0.00000364, 0.00001504, -0.00001042, 0.00000824, -0.00000814, 0.00000369, -0.00000742, 0.00001281, -0.00001274, 0.00001598, -0.00001960, 0.00002402, -0.00002778, 0.00003033, -0.00002308, 0.00000422, -0.00000699, -0.00000074, -0.00000338, 0.00000387, -0.00000989, 0.00000732, -0.00001155, -0.00000559, -0.00000794, 0.00000791, -0.00001186, 0.00000998, -0.00001194, 0.00000183, -0.00000711, -0.00000847, -0.00001059, 0.00000010, -0.00001194, -0.00000656, -0.00001299, -0.00001669, -0.00000639, -0.00005040, -0.00002408, -0.00006500, -0.00003793, -0.00045359, 0.00024410, 0.00046589, 0.00008436, 0.00041251, 0.00033189, -0.00043829, -0.00007357, -0.00005734, -0.00002716, -0.00002461, -0.00002580, -0.00001035, 0.00000387, -0.00001720, -0.00001172, -0.00000040, -0.00000430, 0.00000680, -0.00001102, 0.00000099, -0.00000096, -0.00000943, -0.00000623, 0.00000347, -0.00000128, 0.00000259, 0.00000383, -0.00000635, 0.00001489, -0.00001010, -0.00000009, -0.00000170, -0.00001471, 0.00000957, -0.00001426, 0.00000985, -0.00001233, 0.00001627, -0.00001523, 0.00000465, -0.00000855, 0.00000411, -0.00000642, 0.00001183, -0.00000911, 0.00000341, -0.00000732, 0.00000105, -0.00000505, 0.00000509, },
		{ 0.00000246, -0.00000326, 0.00000511, -0.00000919, 0.00000354, -0.00000565, 0.00001132, -0.00001111, -0.00000353, -0.00000233, 0.00001715, -0.00001585, 0.00000292, -0.00000128, 0.00000357, -0.00000433, 0.00000246, 0.00000012, -0.00000928, 0.00000665, 0.00000653, -0.00000746, -0.00000155, 0.00000061, 0.00000853, -0.00000464, 0.00000199, -0.00000504, 0.00000697, -0.00000802, 0.00000919, -0.00000966, 0.00000232, -0.00000522, 0.00000122, -0.00000014, 0.00001221, -0.00001051, 0.00000885, -0.00000900, 0.00000321, -0.00000337, 0.00000834, -0.00000910, 0.00001180, -0.00001498, 0.00002049, -0.00002719, 0.00003517, -0.00003453, 0.00001767, -0.00002062, 0.00001079, -0.00001075, 0.00000950, -0.00001571, 0.00001339, -0.00001947, 0.00000247, -0.00001495, 0.00001440, -0.00001748, 0.00001738, -0.00002249, 0.00001258, -0.00001762, 0.00000297, -0.00002399, 0.00001249, -0.00002385, 0.00000784, -0.00003199, 0.00000411, -0.00003140, -0.00002284, -0.00006142, -0.00002226, -0.00012802, -0.00038640, 0.00040905, 0.00029966, 0.00012890, 0.00033789, 0.00034793, -0.00043037, -0.00006125, -0.00006726, -0.00001344, -0.00003550, -0.00001584, -0.00001842, 0.00001254, -0.00002389, -0.00000503, -0.00000747, 0.00000464, -0.00000089, -0.00000573, -0.00000300, 0.00000149, -0.00001176, -0.00000146, -0.00000066, 0.00000216, 0.00000082, 0.00000498, -0.00000699, 0.00001842, -0.00001747, 0.00001037, -0.00001658, -0.00000042, -0.00000259, -0.00000374, 0.00000095, -0.00000425, 0.00001032, -0.00001290, 0.00000276, -0.00000604, 0.00000099, -0.00000161, 0.00000851, -0.00000886, 0.00000300, -0.00000685, 0.00000015, -0.00000293, },
		{ -0.00000094, -0.00000027, 0.00000010, 0.00000164, -0.00000701, 0.00000177, -0.00000238, 0.00000906, -0.00001302, -0.00000253, 0.00000171, 0.00001404, -0.00001894, 0.00000826, -0.00000496, 0.00000664, -0.00000809, 0.00000740, -0.00000612, -0.00000444, 0.00000747, 0.00000398, -0.00000857, 0.00000036, 0.00000182, 0.00000721, -0.00000517, 0.00000194, -0.00000391, 0.00000596, -0.00000750, 0.00000972, -0.00001221, 0.00000507, -0.00000689, 0.00000240, 0.00000171, 0.00000957, -0.00001026, 0.00000959, -0.00001077, 0.00000465, -0.00000120, 0.00000458, -0.00000530, 0.00000665, -0.00000786, 0.00001232, -0.00001958, 0.00003117, -0.00003760, 0.00002501, -0.00003077, 0.00002132, -0.00001826, 0.00001500, -0.00002086, 0.00001830, -0.00002625, 0.00001013, -0.00002196, 0.00002077, -0.00002229, 0.00002296, -0.00003126, 0.00002232, -0.00002747, 0.00001377, -0.00003723, 0.00002530, -0.00003573, 0.00002134, -0.00005018, 0.00002487, -0.00005747, 0.00000675, -0.00010264, 0.00002665, -0.00024564, -0.00022812, 0.00046352, 0.00017783, 0.00016603, 0.00027306, 0.00036097, -0.00042103, -0.00005182, -0.00007439, -0.00000235, -0.00004477, -0.00000704, -0.00002495, 0.00001940, -0.00002884, -0.00000032, -0.00001217, 0.00001163, -0.00000787, -0.00000056, -0.00000671, 0.00000303, -0.00001216, 0.00000113, -0.00000301, 0.00000400, 0.00000087, 0.00000348, -0.00000318, 0.00001559, -0.00001758, 0.00001321, -0.00002524, 0.00001024, -0.00001305, 0.00000636, -0.00000838, 0.00000520, 0.00000187, -0.00000806, -0.00000067, -0.00000274, -0.00000251, 0.00000402, 0.00000328, -0.00000671, 0.00000140, -0.00000570, -0.00000086, },
		{ -0.00000112, -0.00000011, -0.00000181, 0.00000254, -0.00000155, -0.00000467, -0.00000005, 0.00000133, 0.00000514, -0.00001345, -0.00000136, 0.00000591, 0.00000822, -0.00001820, 0.00001107, -0.00000726, 0.00000851, -0.00001065, 0.00001151, -0.00001280, 0.00000297, 0.00000496, 0.00000314, -0.00001038, 0.00000383, 0.00000115, 0.00000691, -0.00000649, 0.00000303, -0.00000379, 0.00000539, -0.00000682, 0.00000956, -0.00001424, 0.00000820, -0.00000965, 0.00000544, 0.00000158, 0.00000768, -0.00000985, 0.00001018, -0.00001297, 0.00000784, -0.00000153, 0.00000289, -0.00000317, 0.00000287, -0.00000113, 0.00000260, -0.00000755, 0.00001950, -0.00003133, 0.00002338, -0.00003347, 0.00002667, -0.00002221, 0.00001720, -0.00002244, 0.00001902, -0.00002848, 0.00001373, -0.00002534, 0.00002360, -0.00002295, 0.00002306, -0.00003385, 0.00002606, -0.00003127, 0.00001806, -0.00004388, 0.00003175, -0.00004050, 0.00002598, -0.00005810, 0.00003432, -0.00007120, 0.00002211, -0.00012706, 0.00005211, -0.00034695, -0.00002205, 0.00041992, 0.00011717, 0.00017678, 0.00023105, 0.00036031, -0.00040164, -0.00005238, -0.00007238, 0.00000021, -0.00004698, -0.00000406, -0.00002596, 0.00002089, -0.00002883, -0.00000052, -0.00001159, 0.00001352, -0.00001103, 0.00000193, -0.00000825, 0.00000242, -0.00000961, 0.00000045, -0.00000248, 0.00000349, 0.00000297, -0.00000023, 0.00000424, 0.00000699, -0.00001022, 0.00000681, -0.00002441, 0.00001295, -0.00001716, 0.00001143, -0.00001366, 0.00001173, -0.00000534, -0.00000356, -0.00000345, -0.00000058, -0.00000434, 0.00000820, -0.00000172, -0.00000419, -0.00000029, -0.00000477, },
		{ -0.00000464, -0.00000022, -0.00000094, -0.00000142, 0.00000303, -0.00000325, -0.00000331, -0.00000076, 0.00000416, 0.00000067, -0.00001287, 0.00000042, 0.00000901, 0.00000116, -0.00001416, 0.00001081, -0.00000730, 0.00000814, -0.00001062, 0.00001284, -0.00001741, 0.00001051, 0.00000064, 0.00000331, -0.00001209, 0.00000789, -0.00000076, 0.00000726, -0.00000815, 0.00000487, -0.00000451, 0.00000532, -0.00000603, 0.00000848, -0.00001501, 0.00001059, -0.00001235, 0.00000944, -0.00000026, 0.00000674, -0.00000939, 0.00001034, -0.00001490, 0.00001205, -0.00000421, 0.00000386, -0.00000394, 0.00000250, 0.00000211, -0.00000453, 0.00000409, 0.00000460, -0.00001869, 0.00001395, -0.00002828, 0.00002552, -0.00002125, 0.00001513, -0.00001973, 0.00001489, -0.00002521, 0.00001197, -0.00002377, 0.00002179, -0.00001873, 0.00001696, -0.00002909, 0.00002214, -0.00002711, 0.00001361, -0.00004129, 0.00002902, -0.00003542, 0.00001873, -0.00005186, 0.00002736, -0.00006605, 0.00001460, -0.00012264, 0.00003331, -0.00039111, 0.00017690, 0.00031565, 0.00010905, 0.00015984, 0.00021288, 0.00034454, -0.00037116, -0.00006352, -0.00006071, -0.00000649, -0.00004123, -0.00000764, -0.00002103, 0.00001676, -0.00002377, -0.00000554, -0.00000575, 0.00000990, -0.00000964, 0.00000111, -0.00000744, 0.00000006, -0.00000489, -0.00000277, 0.00000024, 0.00000145, 0.00000582, -0.00000402, 0.00001224, -0.00000393, 0.00000122, -0.00000621, -0.00001515, 0.00000733, -0.00001363, 0.00000961, -0.00001264, 0.00001274, -0.00000867, -0.00000148, -0.00000411, -0.00000070, -0.00000332, 0.00000936, -0.00000484, -0.00000263, -0.00000118, },
		{ -0.00000101, -0.00000530, 0.00000159, -0.00000312, 0.00000075, 0.00000143, -0.00000302, -0.00000348, 0.00000030, 0.00000508, -0.00000323, -0.00001182, 0.00000315, 0.00000983, -0.00000520, -0.00000836, 0.00000818, -0.00000532, 0.00000560, -0.00000780, 0.00001054, -0.00001817, 0.00001589, -0.00000371, 0.00000346, -0.00001272, 0.00001135, -0.00000297, 0.00000759, -0.00000949, 0.00000678, -0.00000567, 0.00000565, -0.00000527, 0.00000653, -0.00001415, 0.00001139, -0.00001381, 0.00001320, -0.00000310, 0.00000662, -0.00000888, 0.00000979, -0.00001576, 0.00001606, -0.00000818, 0.00000692, -0.00000753, 0.00000618, 0.00000006, -0.00000574, 0.00001047, -0.00000777, -0.00000520, 0.00000108, -0.00001811, 0.00001958, -0.00001664, 0.00001012, -0.00001435, 0.00000772, -0.00001814, 0.00000628, -0.00001857, 0.00001689, -0.00001170, 0.00000712, -0.00001940, 0.00001283, -0.00001725, 0.00000263, -0.00003147, 0.00001922, -0.00002311, 0.00000276, -0.00003478, 0.00000722, -0.00004513, -0.00001280, -0.00009198, -0.00003204, -0.00035560, 0.00032104, 0.00019661, 0.00012946, 0.00012703, 0.00020989, 0.00031995, -0.00033493, -0.00008026, -0.00004386, -0.00001859, -0.00003066, -0.00001498, -0.00001290, 0.00000965, -0.00001626, -0.00001269, 0.00000274, 0.00000288, -0.00000517, -0.00000186, -0.00000549, -0.00000249, 0.00000006, -0.00000659, 0.00000343, -0.00000041, 0.00000750, -0.00000529, 0.00001711, -0.00001251, 0.00001138, -0.00002038, -0.00000205, -0.00000329, -0.00000476, 0.00000243, -0.00000613, 0.00000825, -0.00000760, -0.00000233, -0.00000250, -0.00000295, 0.00000038, 0.00000727, -0.00000544, -0.00000257, },
		{ -0.00000372, -0.00000012, -0.00000622, 0.00000352, -0.00000568, 0.00000375, -0.00000156, -0.00000126, -0.00000493, 0.00000302, 0.00000369, -0.00000580, -0.00001066, 0.00000681, 0.00000769, -0.00000906, -0.00000282, 0.00000471, -0.00000250, 0.00000206, -0.00000324, 0.00000523, -0.00001473, 0.00001779, -0.00000672, 0.00000283, -0.00001163, 0.00001327, -0.00000463, 0.00000736, -0.00000994, 0.00000815, -0.00000678, 0.00000623, -0.00000473, 0.00000411, -0.00001183, 0.00001034, -0.00001328, 0.00001560, -0.00000601, 0.00000699, -0.00000829, 0.00000841, -0.00001493, 0.00001858, -0.00001185, 0.00001051, -0.00001245, 0.00001267, -0.00000678, -0.00000016, 0.00000881, -0.00001301, 0.00000366, -0.00000983, -0.00000750, 0.00001231, -0.00001121, 0.00000490, -0.00000924, 0.00000076, -0.00001058, -0.00000024, -0.00001269, 0.00001197, -0.00000547, -0.00000217, -0.00000936, 0.00000282, -0.00000662, -0.00000970, -0.00001970, 0.00000783, -0.00000990, -0.00001448, -0.00001537, -0.00001656, -0.00001932, -0.00004724, -0.00005136, -0.00012574, -0.00024407, 0.00038440, 0.00010078, 0.00015213, 0.00009442, 0.00020984, 0.00029587, -0.00030081, -0.00009549, -0.00002832, -0.00003033, -0.00002016, -0.00002178, -0.00000554, 0.00000329, -0.00000983, -0.00001840, 0.00001025, -0.00000432, -0.00000013, -0.00000506, -0.00000399, -0.00000352, 0.00000328, -0.00000905, 0.00000537, -0.00000070, 0.00000677, -0.00000247, 0.00001637, -0.00001517, 0.00001545, -0.00002989, 0.00000903, -0.00001366, 0.00000496, -0.00000631, 0.00000270, 0.00000058, -0.00000357, -0.00000511, 0.00000029, -0.00000601, 0.00000535, 0.00000297, -0.00000401, },
		{ -0.00000172, -0.00000527, 0.00000080, -0.00000669, 0.00000465, -0.00000748, 0.00000629, -0.00000470, 0.00000104, -0.00000679, 0.00000658, 0.00000036, -0.00000683, -0.00000938, 0.00001079, 0.00000280, -0.00000942, 0.00000071, 0.00000206, -0.00000041, -0.00000077, 0.00000113, -0.00000120, -0.00000827, 0.00001624, -0.00000797, 0.00000132, -0.00000882, 0.00001333, -0.00000534, 0.00000634, -0.00000926, 0.00000858, -0.00000749, 0.00000692, -0.00000471, 0.00000189, -0.00000877, 0.00000784, -0.00001068, 0.00001599, -0.00000822, 0.00000746, -0.00000767, 0.00000638, -0.00001224, 0.00001867, -0.00001367, 0.00001271, -0.00001635, 0.00001924, -0.00001568, 0.00001014, -0.00000035, -0.00000961, 0.00000470, -0.00001470, -0.00000053, 0.00000710, -0.00000773, 0.00000201, -0.00000701, -0.00000303, -0.00000569, -0.00000450, -0.00000901, 0.00000982, -0.00000323, -0.00000703, -0.00000334, -0.00000328, -0.00000023, -0.00001794, -0.00001159, 0.00000069, -0.00000235, -0.00002517, -0.00000290, -0.00003312, -0.00000162, -0.00007251, -0.00002347, -0.00021571, -0.00008437, 0.00036799, 0.00004774, 0.00015911, 0.00007415, 0.00020323, 0.00027948, -0.00027474, -0.00010379, -0.00001914, -0.00003707, -0.00001371, -0.00002470, -0.00000198, 0.00000043, -0.00000701, -0.00002009, 0.00001399, -0.00000898, 0.00000332, -0.00000700, -0.00000391, -0.00000227, 0.00000384, -0.00000922, 0.00000542, 0.00000082, 0.00000391, 0.00000394, 0.00001005, -0.00001112, 0.00001140, -0.00003118, 0.00001364, -0.00001922, 0.00001111, -0.00001240, 0.00000993, -0.00000696, 0.00000102, -0.00000799, 0.00000267, -0.00000814, 0.00000963, -0.00000180, },
		{ -0.00000542, 0.00000021, -0.00000643, 0.00000118, -0.00000624, 0.00000443, -0.00000771, 0.00000730, -0.00000680, 0.00000276, -0.00000796, 0.00000978, -0.00000399, -0.00000658, -0.00000770, 0.00001407, -0.00000371, -0.00000640, 0.00000143, 0.00000132, -0.00000020, -0.00000144, 0.00000331, -0.00000634, -0.00000096, 0.00001244, -0.00000792, -0.00000047, -0.00000500, 0.00001187, -0.00000525, 0.00000480, -0.00000768, 0.00000808, -0.00000765, 0.00000763, -0.00000543, 0.00000058, -0.00000596, 0.00000488, -0.00000666, 0.00001438, -0.00000932, 0.00000781, -0.00000719, 0.00000422, -0.00000820, 0.00001626, -0.00001283, 0.00001213, -0.00001711, 0.00002281, -0.00002274, 0.00002089, -0.00001336, 0.00000030, -0.00000179, -0.00001234, 0.00000090, 0.00000572, -0.00000752, 0.00000245, -0.00000865, -0.00000241, -0.00000500, -0.00000497, -0.00000883, 0.00001149, -0.00000607, -0.00000585, -0.00000342, -0.00000314, -0.00000069, -0.00001906, -0.00001030, 0.00000089, -0.00000389, -0.00002500, -0.00000284, -0.00003565, -0.00000067, -0.00007703, -0.00002701, -0.00026827, 0.00008224, 0.00029522, 0.00003685, 0.00014544, 0.00006991, 0.00018692, 0.00027278, -0.00025816, -0.00010383, -0.00001774, -0.00003730, -0.00001254, -0.00002282, -0.00000294, 0.00000161, -0.00000816, -0.00001733, 0.00001313, -0.00001005, 0.00000433, -0.00000730, -0.00000502, 0.00000073, 0.00000224, -0.00000752, 0.00000419, 0.00000307, 0.00000070, 0.00001148, 0.00000078, -0.00000277, 0.00000086, -0.00002439, 0.00001059, -0.00001800, 0.00001132, -0.00001323, 0.00001272, -0.00001163, 0.00000398, -0.00000944, 0.00000339, -0.00000807, 0.00001163, },
		{ 0.00001075, -0.00000699, 0.00000105, -0.00000679, 0.00000089, -0.00000492, 0.00000292, -0.00000628, 0.00000641, -0.00000722, 0.00000319, -0.00000761, 0.00001146, -0.00000826, -0.00000555, -0.00000530, 0.00001551, -0.00001010, -0.00000122, -0.00000036, 0.00000258, -0.00000209, 0.00000059, 0.00000216, -0.00000823, 0.00000488, 0.00000812, -0.00000755, -0.00000162, -0.00000128, 0.00000971, -0.00000488, 0.00000329, -0.00000575, 0.00000698, -0.00000733, 0.00000828, -0.00000694, 0.00000070, -0.00000437, 0.00000268, -0.00000238, 0.00001139, -0.00000937, 0.00000801, -0.00000713, 0.00000271, -0.00000388, 0.00001219, -0.00000968, 0.00000862, -0.00001385, 0.00002142, -0.00002461, 0.00002749, -0.00002503, 0.00001209, -0.00001242, -0.00000463, -0.00000245, 0.00000776, -0.00001000, 0.00000537, -0.00001312, 0.00000172, -0.00000788, -0.00000220, -0.00001140, 0.00001575, -0.00001249, 0.00000005, -0.00000855, 0.00000225, -0.00000715, -0.00001367, -0.00001524, 0.00000744, -0.00001321, -0.00001519, -0.00001435, -0.00002457, -0.00001657, -0.00005978, -0.00006789, -0.00026094, 0.00021567, 0.00020038, 0.00005298, 0.00011735, 0.00007733, 0.00016406, 0.00027267, -0.00024808, -0.00009832, -0.00002185, -0.00003278, -0.00001506, -0.00001778, -0.00000680, 0.00000520, -0.00001152, -0.00001183, 0.00000890, -0.00000822, 0.00000342, -0.00000663, -0.00000624, 0.00000399, -0.00000002, -0.00000533, 0.00000307, 0.00000442, -0.00000051, 0.00001683, -0.00000746, 0.00000544, -0.00001169, -0.00001304, 0.00000213, -0.00001128, 0.00000618, -0.00000877, 0.00001040, -0.00001233, 0.00000435, -0.00000889, 0.00000222, -0.00000557, },
		{ -0.00000154, 0.00000754, -0.00000656, 0.00000078, -0.00000648, 0.00000026, -0.00000332, 0.00000081, -0.00000384, 0.00000403, -0.00000606, 0.00000230, -0.00000559, 0.00001099, -0.00001151, -0.00000416, -0.00000211, 0.00001434, -0.00001462, 0.00000431, -0.00000341, 0.00000495, -0.00000525, 0.00000463, -0.00000213, -0.00000612, 0.00000766, 0.00000478, -0.00000774, -0.00000131, 0.00000131, 0.00000773, -0.00000481, 0.00000243, -0.00000413, 0.00000577, -0.00000670, 0.00000874, -0.00000900, 0.00000230, -0.00000452, 0.00000220, 0.00000092, 0.00000799, -0.00000874, 0.00000818, -0.00000774, 0.00000257, -0.00000061, 0.00000800, -0.00000565, 0.00000348, -0.00000755, 0.00001519, -0.00002008, 0.00002696, -0.00003085, 0.00002063, -0.00002242, 0.00000467, -0.00000799, 0.00001124, -0.00001324, 0.00000862, -0.00001801, 0.00000693, -0.00001211, 0.00000171, -0.00001448, 0.00001988, -0.00001925, 0.00000725, -0.00001532, 0.00000938, -0.00001595, -0.00000537, -0.00002270, 0.00001599, -0.00002511, -0.00000160, -0.00003098, -0.00000704, -0.00004105, -0.00003103, -0.00013574, -0.00019070, 0.00028984, 0.00011544, 0.00007613, 0.00008640, 0.00008798, 0.00014105, 0.00027344, -0.00023922, -0.00009207, -0.00002720, -0.00002709, -0.00001816, -0.00001253, -0.00001079, 0.00000854, -0.00001435, -0.00000637, 0.00000373, -0.00000524, 0.00000184, -0.00000606, -0.00000631, 0.00000595, -0.00000136, -0.00000402, 0.00000327, 0.00000367, 0.00000187, 0.00001746, -0.00001124, 0.00000911, -0.00002102, -0.00000229, -0.00000734, -0.00000269, -0.00000137, -0.00000133, 0.00000448, -0.00000983, 0.00000260, -0.00000691, -0.00000004, },
		{ -0.00000199, 0.00000244, 0.00000332, -0.00000496, -0.00000008, -0.00000598, -0.00000015, -0.00000219, -0.00000090, -0.00000147, 0.00000115, -0.00000410, 0.00000074, -0.00000242, 0.00000841, -0.00001329, -0.00000261, 0.00000147, 0.00001051, -0.00001606, 0.00000843, -0.00000617, 0.00000704, -0.00000816, 0.00000901, -0.00000806, -0.00000073, 0.00000704, 0.00000312, -0.00000884, 0.00000072, 0.00000218, 0.00000651, -0.00000541, 0.00000257, -0.00000330, 0.00000486, -0.00000595, 0.00000879, -0.00001110, 0.00000490, -0.00000630, 0.00000381, 0.00000234, 0.00000512, -0.00000794, 0.00000842, -0.00000904, 0.00000418, 0.00000052, 0.00000529, -0.00000263, -0.00000106, -0.00000067, 0.00000643, -0.00001064, 0.00001925, -0.00002888, 0.00002254, -0.00002775, 0.00001175, -0.00001274, 0.00001384, -0.00001515, 0.00001003, -0.00002084, 0.00001058, -0.00001515, 0.00000443, -0.00001578, 0.00002117, -0.00002297, 0.00001202, -0.00001988, 0.00001409, -0.00002263, 0.00000133, -0.00002804, 0.00002129, -0.00003328, 0.00000833, -0.00004416, 0.00000680, -0.00006167, -0.00000785, -0.00020741, -0.00007470, 0.00029920, 0.00006023, 0.00009041, 0.00006285, 0.00009416, 0.00012365, 0.00027010, -0.00022702, -0.00008932, -0.00002989, -0.00002355, -0.00001903, -0.00000960, -0.00001258, 0.00000948, -0.00001440, -0.00000340, -0.00000010, -0.00000286, 0.00000074, -0.00000628, -0.00000465, 0.00000581, -0.00000095, -0.00000420, 0.00000506, 0.00000091, 0.00000773, 0.00001288, -0.00000932, 0.00000596, -0.00002347, 0.00000349, -0.00001346, 0.00000370, -0.00000756, 0.00000568, -0.00000224, -0.00000606, 0.00000011, -0.00000475, },
		{ -0.00000361, -0.00000238, 0.00000485, -0.00000040, -0.00000322, -0.00000087, -0.00000579, 0.00000013, -0.00000212, -0.00000142, -0.00000019, -0.00000110, -0.00000235, -0.00000052, 0.00000083, 0.00000442, -0.00001361, -0.00000079, 0.00000464, 0.00000483, -0.00001422, 0.00001008, -0.00000745, 0.00000764, -0.00000935, 0.00001182, -0.00001345, 0.00000610, 0.00000389, 0.00000294, -0.00001052, 0.00000405, 0.00000140, 0.00000614, -0.00000661, 0.00000365, -0.00000336, 0.00000448, -0.00000520, 0.00000824, -0.00001256, 0.00000762, -0.00000891, 0.00000715, 0.00000162, 0.00000339, -0.00000732, 0.00000869, -0.00001072, 0.00000731, -0.00000082, 0.00000506, -0.00000213, -0.00000285, 0.00000382, -0.00000128, 0.00000005, 0.00000725, -0.00002043, 0.00001739, -0.00002666, 0.00001434, -0.00001473, 0.00001412, -0.00001460, 0.00000848, -0.00002026, 0.00001108, -0.00001546, 0.00000463, -0.00001417, 0.00001836, -0.00002188, 0.00001215, -0.00001981, 0.00001371, -0.00002416, 0.00000326, -0.00002820, 0.00001993, -0.00003350, 0.00000933, -0.00004745, 0.00000903, -0.00006809, -0.00000631, -0.00025570, 0.00005649, 0.00025740, 0.00003891, 0.00008908, 0.00005136, 0.00009236, 0.00011418, 0.00026072, -0.00020971, -0.00009185, -0.00002813, -0.00002360, -0.00001653, -0.00000995, -0.00001140, 0.00000742, -0.00001099, -0.00000397, -0.00000137, -0.00000200, 0.00000055, -0.00000717, -0.00000166, 0.00000393, 0.00000094, -0.00000539, 0.00000756, -0.00000241, 0.00001514, 0.00000493, -0.00000323, -0.00000316, -0.00001848, 0.00000248, -0.00001382, 0.00000523, -0.00000960, 0.00000936, -0.00000709, -0.00000312, -0.00000162, },
		{ -0.00000177, -0.00000401, -0.00000072, 0.00000487, -0.00000262, -0.00000211, -0.00000115, -0.00000611, 0.00000115, -0.00000320, -0.00000047, -0.00000056, -0.00000197, -0.00000161, -0.00000057, 0.00000301, 0.00000004, -0.00001284, 0.00000152, 0.00000640, -0.00000123, -0.00000995, 0.00000925, -0.00000687, 0.00000631, -0.00000805, 0.00001172, -0.00001625, 0.00001214, -0.00000024, 0.00000340, -0.00001195, 0.00000773, -0.00000041, 0.00000627, -0.00000797, 0.00000526, -0.00000411, 0.00000462, -0.00000454, 0.00000697, -0.00001282, 0.00000949, -0.00001123, 0.00001119, -0.00000078, 0.00000289, -0.00000699, 0.00000878, -0.00001211, 0.00001106, -0.00000402, 0.00000720, -0.00000453, -0.00000081, 0.00000381, -0.00000465, 0.00000768, -0.00000447, -0.00000934, 0.00000768, -0.00002016, 0.00001241, -0.00001373, 0.00001212, -0.00001194, 0.00000449, -0.00001665, 0.00000856, -0.00001309, 0.00000256, -0.00001031, 0.00001227, -0.00001650, 0.00000786, -0.00001523, 0.00000821, -0.00002022, 0.00000010, -0.00002314, 0.00001207, -0.00002566, 0.00000077, -0.00003968, -0.00000225, -0.00005731, -0.00003383, -0.00025978, 0.00016993, 0.00018950, 0.00004276, 0.00007462, 0.00005023, 0.00008392, 0.00011104, 0.00024686, -0.00018874, -0.00009848, -0.00002276, -0.00002650, -0.00001156, -0.00001268, -0.00000820, 0.00000345, -0.00000517, -0.00000748, -0.00000025, -0.00000257, 0.00000093, -0.00000794, 0.00000149, 0.00000146, 0.00000323, -0.00000645, 0.00000934, -0.00000418, 0.00002119, -0.00000311, 0.00000350, -0.00001473, -0.00000862, -0.00000398, -0.00000887, 0.00000173, -0.00000681, 0.00000852, -0.00000858, -0.00000225, },
		{ -0.00000346, -0.00000049, -0.00000553, 0.00000251, 0.00000264, -0.00000306, -0.00000188, -0.00000091, -0.00000676, 0.00000254, -0.00000496, 0.00000154, -0.00000239, -0.00000134, -0.00000211, 0.00000097, 0.00000329, -0.00000372, -0.00001144, 0.00000441, 0.00000596, -0.00000601, -0.00000482, 0.00000685, -0.00000504, 0.00000358, -0.00000463, 0.00000851, -0.00001541, 0.00001566, -0.00000380, 0.00000356, -0.00001224, 0.00001071, -0.00000241, 0.00000636, -0.00000896, 0.00000680, -0.00000513, 0.00000511, -0.00000406, 0.00000511, -0.00001166, 0.00000982, -0.00001217, 0.00001466, -0.00000396, 0.00000327, -0.00000683, 0.00000840, -0.00001248, 0.00001420, -0.00000772, 0.00001051, -0.00000880, 0.00000442, -0.00000095, -0.00000213, 0.00000917, -0.00001159, -0.00000025, -0.00000248, -0.00001122, 0.00000775, -0.00001099, 0.00000911, -0.00000867, -0.00000021, -0.00001174, 0.00000455, -0.00000948, -0.00000019, -0.00000622, 0.00000537, -0.00000934, 0.00000150, -0.00000855, 0.00000003, -0.00001311, -0.00000582, -0.00001572, 0.00000119, -0.00001363, -0.00001323, -0.00002529, -0.00002202, -0.00003539, -0.00008527, -0.00021300, 0.00024119, 0.00012180, 0.00005703, 0.00005494, 0.00005376, 0.00007322, 0.00011017, 0.00023221, -0.00016751, -0.00010618, -0.00001625, -0.00003005, -0.00000620, -0.00001579, -0.00000494, -0.00000034, 0.00000094, -0.00001217, 0.00000213, -0.00000382, 0.00000122, -0.00000766, 0.00000354, -0.00000026, 0.00000475, -0.00000628, 0.00000924, -0.00000277, 0.00002338, -0.00000808, 0.00000700, -0.00002416, 0.00000170, -0.00001234, -0.00000138, -0.00000467, -0.00000073, 0.00000400, -0.00000692, },
		{ -0.00000355, -0.00000577, 0.00000134, -0.00000715, 0.00000609, -0.00000088, -0.00000220, -0.00000227, -0.00000045, -0.00000729, 0.00000366, -0.00000659, 0.00000374, -0.00000493, 0.00000025, -0.00000346, 0.00000382, 0.00000146, -0.00000617, -0.00000973, 0.00000767, 0.00000308, -0.00000825, -0.00000050, 0.00000426, -0.00000310, 0.00000074, -0.00000044, 0.00000330, -0.00001121, 0.00001597, -0.00000585, 0.00000285, -0.00001092, 0.00001223, -0.00000392, 0.00000599, -0.00000916, 0.00000778, -0.00000598, 0.00000576, -0.00000390, 0.00000305, -0.00000934, 0.00000849, -0.00001111, 0.00001650, -0.00000687, 0.00000399, -0.00000666, 0.00000740, -0.00001133, 0.00001560, -0.00001037, 0.00001322, -0.00001292, 0.00001072, -0.00000863, 0.00000534, 0.00000397, -0.00001186, 0.00000336, -0.00000920, -0.00000331, 0.00000289, -0.00000834, 0.00000674, -0.00000653, -0.00000359, -0.00000767, 0.00000107, -0.00000646, -0.00000178, -0.00000418, 0.00000048, -0.00000348, -0.00000379, -0.00000309, -0.00000728, -0.00000635, -0.00001095, -0.00001001, -0.00000777, -0.00000321, -0.00002608, -0.00001192, -0.00004103, -0.00001489, -0.00014429, -0.00012502, 0.00026109, 0.00007303, 0.00006820, 0.00003835, 0.00005589, 0.00006486, 0.00010754, 0.00022040, -0.00014945, -0.00011179, -0.00001126, -0.00003200, -0.00000254, -0.00001737, -0.00000337, -0.00000207, 0.00000528, -0.00001617, 0.00000442, -0.00000493, 0.00000095, -0.00000591, 0.00000381, -0.00000050, 0.00000496, -0.00000459, 0.00000716, 0.00000205, 0.00002078, -0.00000845, 0.00000495, -0.00002788, 0.00000831, -0.00001860, 0.00000504, -0.00001073, 0.00000578, -0.00000198, },
		{ -0.00000694, -0.00000036, -0.00000787, 0.00000265, -0.00000773, 0.00000869, -0.00000444, -0.00000083, -0.00000280, -0.00000017, -0.00000730, 0.00000394, -0.00000732, 0.00000518, -0.00000715, 0.00000190, -0.00000480, 0.00000712, -0.00000207, -0.00000711, -0.00000779, 0.00001068, -0.00000170, -0.00000748, 0.00000184, 0.00000268, -0.00000224, -0.00000082, 0.00000274, -0.00000200, -0.00000514, 0.00001361, -0.00000636, 0.00000140, -0.00000814, 0.00001210, -0.00000466, 0.00000506, -0.00000847, 0.00000797, -0.00000639, 0.00000643, -0.00000425, 0.00000139, -0.00000660, 0.00000606, -0.00000819, 0.00001623, -0.00000877, 0.00000459, -0.00000640, 0.00000592, -0.00000869, 0.00001474, -0.00001086, 0.00001380, -0.00001479, 0.00001531, -0.00001605, 0.00001474, -0.00000579, -0.00000588, 0.00000052, -0.00001036, 0.00000113, -0.00000019, -0.00000717, 0.00000608, -0.00000658, -0.00000439, -0.00000592, -0.00000044, -0.00000529, -0.00000111, -0.00000546, -0.00000056, -0.00000112, -0.00000569, -0.00000139, -0.00001088, -0.00000283, -0.00001251, -0.00000909, -0.00001097, 0.00000087, -0.00003210, -0.00000644, -0.00005050, -0.00000905, -0.00018940, -0.00001800, 0.00023636, 0.00005013, 0.00006891, 0.00002969, 0.00005312, 0.00006124, 0.00010115, 0.00021318, -0.00013626, -0.00011360, -0.00000920, -0.00003124, -0.00000154, -0.00001660, -0.00000419, -0.00000095, 0.00000672, -0.00001826, 0.00000573, -0.00000551, 0.00000026, -0.00000302, 0.00000256, 0.00000054, 0.00000422, -0.00000205, 0.00000425, 0.00000886, 0.00001451, -0.00000502, -0.00000238, -0.00002485, 0.00000905, -0.00002014, 0.00000764, -0.00001367, 0.00000997, },
		{ 0.00001034, -0.00000922, 0.00000128, -0.00000879, 0.00000280, -0.00000667, 0.00000939, -0.00000699, 0.00000030, -0.00000307, -0.00000028, -0.00000666, 0.00000322, -0.00000678, 0.00000524, -0.00000828, 0.00000281, -0.00000525, 0.00000977, -0.00000641, -0.00000674, -0.00000555, 0.00001263, -0.00000720, -0.00000420, 0.00000187, 0.00000266, -0.00000304, -0.00000021, 0.00000349, -0.00000539, 0.00000078, 0.00000990, -0.00000597, -0.00000014, -0.00000464, 0.00001070, -0.00000478, 0.00000384, -0.00000717, 0.00000748, -0.00000632, 0.00000702, -0.00000524, 0.00000070, -0.00000437, 0.00000352, -0.00000421, 0.00001407, -0.00000938, 0.00000486, -0.00000616, 0.00000444, -0.00000519, 0.00001196, -0.00000907, 0.00001167, -0.00001317, 0.00001600, -0.00002005, 0.00002215, -0.00001619, 0.00000329, -0.00000708, -0.00000624, 0.00000145, -0.00000081, -0.00000779, 0.00000713, -0.00000869, -0.00000261, -0.00000669, 0.00000023, -0.00000597, 0.00000150, -0.00000970, 0.00000222, -0.00000258, -0.00000378, -0.00000402, -0.00000991, -0.00000343, -0.00000988, -0.00001350, -0.00000753, -0.00000284, -0.00002929, -0.00001160, -0.00004658, -0.00002560, -0.00020224, 0.00008159, 0.00018476, 0.00004909, 0.00005903, 0.00002901, 0.00004555, 0.00006182, 0.00009162, 0.00020982, -0.00012742, -0.00011189, -0.00000978, -0.00002819, -0.00000275, -0.00001401, -0.00000673, 0.00000240, 0.00000547, -0.00001827, 0.00000594, -0.00000572, -0.00000026, 0.00000003, 0.00000073, 0.00000200, 0.00000352, 0.00000005, 0.00000237, 0.00001518, 0.00000711, -0.00000044, -0.00001229, -0.00001687, 0.00000448, -0.00001673, 0.00000571, -0.00001239, },
		{ -0.00000775, 0.00000715, -0.00000855, 0.00000099, -0.00000836, 0.00000188, -0.00000412, 0.00000803, -0.00000807, 0.00000084, -0.00000300, -0.00000068, -0.00000561, 0.00000178, -0.00000520, 0.00000392, -0.00000804, 0.00000260, -0.00000426, 0.00001083, -0.00001053, -0.00000544, -0.00000293, 0.00001279, -0.00001198, 0.00000033, 0.00000019, 0.00000392, -0.00000523, 0.00000251, 0.00000133, -0.00000560, 0.00000474, 0.00000635, -0.00000561, -0.00000090, -0.00000141, 0.00000875, -0.00000469, 0.00000277, -0.00000574, 0.00000663, -0.00000588, 0.00000747, -0.00000684, 0.00000132, -0.00000341, 0.00000200, -0.00000038, 0.00001082, -0.00000895, 0.00000484, -0.00000619, 0.00000364, -0.00000191, 0.00000839, -0.00000592, 0.00000755, -0.00000835, 0.00001221, -0.00001883, 0.00002452, -0.00002322, 0.00001156, -0.00001600, 0.00000092, -0.00000125, 0.00000040, -0.00000948, 0.00000890, -0.00001163, 0.00000054, -0.00000897, 0.00000217, -0.00000743, 0.00000456, -0.00001507, 0.00000709, -0.00000634, 0.00000040, -0.00000946, -0.00000573, -0.00000681, -0.00000480, -0.00002102, 0.00000014, -0.00001190, -0.00002013, -0.00002467, -0.00003250, -0.00006271, -0.00017445, 0.00015190, 0.00012729, 0.00005958, 0.00004352, 0.00003278, 0.00003594, 0.00006395, 0.00008149, 0.00020788, -0.00012073, -0.00010840, -0.00001145, -0.00002437, -0.00000473, -0.00001103, -0.00000941, 0.00000641, 0.00000274, -0.00001690, 0.00000544, -0.00000596, 0.00000010, 0.00000219, -0.00000055, 0.00000287, 0.00000386, 0.00000056, 0.00000316, 0.00001860, 0.00000135, 0.00000207, -0.00002087, -0.00000760, -0.00000263, -0.00001037, 0.00000063, },
		{ -0.00000491, -0.00000201, 0.00000206, -0.00000593, -0.00000062, -0.00000714, 0.00000060, -0.00000095, 0.00000524, -0.00000785, 0.00000084, -0.00000275, -0.00000103, -0.00000466, 0.00000028, -0.00000325, 0.00000178, -0.00000677, 0.00000146, -0.00000189, 0.00000986, -0.00001361, -0.00000363, -0.00000007, 0.00001084, -0.00001477, 0.00000461, -0.00000209, 0.00000555, -0.00000780, 0.00000629, -0.00000304, -0.00000257, 0.00000584, 0.00000403, -0.00000597, -0.00000025, 0.00000070, 0.00000696, -0.00000483, 0.00000226, -0.00000465, 0.00000581, -0.00000525, 0.00000765, -0.00000874, 0.00000315, -0.00000404, 0.00000226, 0.00000214, 0.00000751, -0.00000798, 0.00000475, -0.00000671, 0.00000405, 0.00000007, 0.00000545, -0.00000299, 0.00000319, -0.00000211, 0.00000527, -0.00001279, 0.00002083, -0.00002443, 0.00001550, -0.00002263, 0.00000812, -0.00000498, 0.00000210, -0.00001102, 0.00001006, -0.00001383, 0.00000339, -0.00001119, 0.00000394, -0.00000820, 0.00000621, -0.00001921, 0.00001154, -0.00000993, 0.00000428, -0.00001498, -0.00000094, -0.00001039, -0.00000033, -0.00002791, 0.00000767, -0.00002142, -0.00001021, -0.00003898, -0.00001721, -0.00010914, -0.00011063, 0.00018210, 0.00008079, 0.00007061, 0.00002874, 0.00003651, 0.00002767, 0.00006461, 0.00007349, 0.00020466, -0.00011368, -0.00010525, -0.00001241, -0.00002146, -0.00000593, -0.00000909, -0.00001068, 0.00000933, 0.00000006, -0.00001515, 0.00000478, -0.00000651, 0.00000166, 0.00000288, -0.00000061, 0.00000262, 0.00000561, -0.00000078, 0.00000714, 0.00001789, -0.00000112, 0.00000032, -0.00002483, -0.00000086, -0.00000875, -0.00000415, },
		{ -0.00000071, -0.00000832, 0.00000232, -0.00000273, -0.00000296, -0.00000246, -0.00000599, -0.00000012, 0.00000172, 0.00000200, -0.00000688, 0.00000060, -0.00000263, -0.00000098, -0.00000434, -0.00000060, -0.00000177, -0.00000033, -0.00000520, 0.00000009, 0.00000121, 0.00000703, -0.00001522, -0.00000154, 0.00000262, 0.00000706, -0.00001495, 0.00000743, -0.00000382, 0.00000649, -0.00000949, 0.00000956, -0.00000800, 0.00000262, 0.00000426, 0.00000322, -0.00000715, 0.00000189, 0.00000127, 0.00000578, -0.00000538, 0.00000249, -0.00000418, 0.00000531, -0.00000457, 0.00000741, -0.00001045, 0.00000561, -0.00000593, 0.00000440, 0.00000269, 0.00000501, -0.00000701, 0.00000475, -0.00000772, 0.00000581, 0.00000012, 0.00000425, -0.00000176, 0.00000057, 0.00000307, -0.00000210, -0.00000433, 0.00001255, -0.00001982, 0.00001373, -0.00002461, 0.00001285, -0.00000785, 0.00000305, -0.00001146, 0.00000967, -0.00001415, 0.00000462, -0.00001209, 0.00000446, -0.00000728, 0.00000520, -0.00002037, 0.00001351, -0.00001124, 0.00000558, -0.00001808, 0.00000194, -0.00001177, 0.00000090, -0.00003085, 0.00001099, -0.00002663, -0.00000522, -0.00004738, -0.00001133, -0.00014830, -0.00002638, 0.00017416, 0.00005353, 0.00007501, 0.00001912, 0.00003711, 0.00002286, 0.00006207, 0.00006918, 0.00019854, -0.00010466, -0.00010376, -0.00001158, -0.00002042, -0.00000551, -0.00000890, -0.00000971, 0.00001007, -0.00000142, -0.00001380, 0.00000427, -0.00000724, 0.00000413, 0.00000227, 0.00000043, 0.00000152, 0.00000826, -0.00000314, 0.00001340, 0.00001345, -0.00000050, -0.00000579, -0.00002285, 0.00000098, -0.00001121, },
		{ -0.00000930, -0.00000118, -0.00000821, 0.00000359, -0.00000554, -0.00000102, -0.00000357, -0.00000557, 0.00000038, 0.00000298, -0.00000071, -0.00000580, 0.00000042, -0.00000281, -0.00000037, -0.00000487, -0.00000044, -0.00000138, -0.00000161, -0.00000412, -0.00000067, 0.00000406, 0.00000303, -0.00001534, 0.00000073, 0.00000451, 0.00000234, -0.00001271, 0.00000826, -0.00000434, 0.00000610, -0.00000935, 0.00001091, -0.00001170, 0.00000819, 0.00000105, 0.00000348, -0.00000867, 0.00000504, 0.00000044, 0.00000525, -0.00000627, 0.00000332, -0.00000432, 0.00000521, -0.00000393, 0.00000659, -0.00001139, 0.00000787, -0.00000822, 0.00000785, 0.00000127, 0.00000376, -0.00000636, 0.00000482, -0.00000888, 0.00000850, -0.00000164, 0.00000510, -0.00000292, 0.00000101, 0.00000500, -0.00000695, 0.00000313, 0.00000290, -0.00001172, 0.00000724, -0.00002164, 0.00001404, -0.00000890, 0.00000274, -0.00001063, 0.00000766, -0.00001241, 0.00000386, -0.00001132, 0.00000355, -0.00000477, 0.00000161, -0.00001820, 0.00001234, -0.00000954, 0.00000344, -0.00001762, 0.00000177, -0.00001004, -0.00000201, -0.00002859, 0.00000832, -0.00002518, -0.00000815, -0.00004586, -0.00002210, -0.00016453, 0.00005748, 0.00014009, 0.00004483, 0.00007105, 0.00001569, 0.00003401, 0.00002149, 0.00005653, 0.00006825, 0.00018963, -0.00009359, -0.00010396, -0.00000907, -0.00002111, -0.00000370, -0.00001011, -0.00000679, 0.00000853, -0.00000134, -0.00001312, 0.00000387, -0.00000766, 0.00000670, 0.00000113, 0.00000185, 0.00000044, 0.00001069, -0.00000493, 0.00001995, 0.00000710, 0.00000137, -0.00001422, -0.00001611, -0.00000210, },
		{ -0.00000803, -0.00000434, -0.00000475, -0.00000493, 0.00000164, -0.00000577, -0.00000073, -0.00000354, -0.00000604, 0.00000215, 0.00000250, -0.00000233, -0.00000498, 0.00000041, -0.00000323, 0.00000065, -0.00000608, 0.00000069, -0.00000224, -0.00000170, -0.00000401, -0.00000020, 0.00000566, -0.00000117, -0.00001427, 0.00000317, 0.00000500, -0.00000210, -0.00000898, 0.00000739, -0.00000371, 0.00000443, -0.00000727, 0.00000966, -0.00001278, 0.00001240, -0.00000237, 0.00000397, -0.00000975, 0.00000831, -0.00000124, 0.00000511, -0.00000718, 0.00000442, -0.00000485, 0.00000545, -0.00000340, 0.00000518, -0.00001116, 0.00000912, -0.00000985, 0.00001150, -0.00000151, 0.00000367, -0.00000606, 0.00000480, -0.00000969, 0.00001127, -0.00000436, 0.00000735, -0.00000605, 0.00000449, 0.00000284, -0.00000736, 0.00000664, -0.00000449, -0.00000364, -0.00000120, -0.00001526, 0.00001216, -0.00000826, 0.00000150, -0.00000911, 0.00000485, -0.00000940, 0.00000177, -0.00000945, 0.00000195, -0.00000175, -0.00000325, -0.00001386, 0.00000892, -0.00000572, -0.00000130, -0.00001422, -0.00000085, -0.00000614, -0.00000776, -0.00002248, 0.00000090, -0.00001824, -0.00001782, -0.00003579, -0.00004976, -0.00014904, 0.00012208, 0.00009618, 0.00004794, 0.00006143, 0.00001652, 0.00002883, 0.00002185, 0.00004974, 0.00006895, 0.00017945, -0.00008162, -0.00010482, -0.00000593, -0.00002255, -0.00000148, -0.00001158, -0.00000308, 0.00000552, -0.00000004, -0.00001298, 0.00000335, -0.00000724, 0.00000849, 0.00000042, 0.00000279, 0.00000027, 0.00001180, -0.00000456, 0.00002455, 0.00000109, 0.00000201, -0.00002177, -0.00000759, },
		{ -0.00000070, -0.00001380, 0.00000117, -0.00000935, -0.00000014, -0.00000239, -0.00000401, -0.00000183, -0.00000263, -0.00000696, 0.00000464, 0.00000058, -0.00000278, -0.00000453, 0.00000047, -0.00000363, 0.00000165, -0.00000747, 0.00000228, -0.00000396, -0.00000076, -0.00000483, 0.00000165, 0.00000539, -0.00000466, -0.00001242, 0.00000567, 0.00000375, -0.00000511, -0.00000503, 0.00000566, -0.00000257, 0.00000225, -0.00000397, 0.00000619, -0.00001088, 0.00001416, -0.00000489, 0.00000393, -0.00000967, 0.00001082, -0.00000310, 0.00000501, -0.00000777, 0.00000539, -0.00000545, 0.00000590, -0.00000307, 0.00000339, -0.00000974, 0.00000893, -0.00000997, 0.00001422, -0.00000467, 0.00000431, -0.00000595, 0.00000446, -0.00000961, 0.00001316, -0.00000682, 0.00000969, -0.00000971, 0.00000967, -0.00000255, -0.00000325, 0.00000489, -0.00000712, 0.00000118, -0.00000834, -0.00000802, 0.00000869, -0.00000679, 0.00000013, -0.00000788, 0.00000243, -0.00000641, -0.00000050, -0.00000755, 0.00000080, 0.00000025, -0.00000745, -0.00000932, 0.00000510, -0.00000168, -0.00000665, -0.00000976, -0.00000413, -0.00000225, -0.00001363, -0.00001569, -0.00000781, -0.00000970, -0.00002962, -0.00002337, -0.00008681, -0.00010309, 0.00015623, 0.00005698, 0.00005433, 0.00005072, 0.00001843, 0.00002397, 0.00002171, 0.00004383, 0.00006915, 0.00016994, -0.00007035, -0.00010495, -0.00000347, -0.00002351, 0.00000001, -0.00001204, 0.00000001, 0.00000222, 0.00000179, -0.00001306, 0.00000259, -0.00000574, 0.00000900, 0.00000072, 0.00000277, 0.00000146, 0.00001117, -0.00000126, 0.00002576, -0.00000301, -0.00000042, -0.00002553, },
		{ -0.00002415, 0.00000221, -0.00001687, 0.00000477, -0.00001265, 0.00000397, -0.00000668, -0.00000153, -0.00000347, -0.00000148, -0.00000764, 0.00000696, -0.00000204, -0.00000238, -0.00000431, 0.00000039, -0.00000371, 0.00000220, -0.00000843, 0.00000362, -0.00000581, 0.00000061, -0.00000606, 0.00000444, 0.00000319, -0.00000680, -0.00001018, 0.00000797, 0.00000089, -0.00000599, -0.00000199, 0.00000405, -0.00000179, 0.00000062, -0.00000078, 0.00000176, -0.00000671, 0.00001338, -0.00000601, 0.00000309, -0.00000816, 0.00001199, -0.00000459, 0.00000468, -0.00000782, 0.00000595, -0.00000587, 0.00000643, -0.00000310, 0.00000163, -0.00000754, 0.00000742, -0.00000827, 0.00001517, -0.00000727, 0.00000512, -0.00000583, 0.00000372, -0.00000843, 0.00001351, -0.00000796, 0.00001077, -0.00001212, 0.00001435, -0.00000894, 0.00000359, -0.00000127, -0.00000456, 0.00000110, -0.00001175, -0.00000229, 0.00000528, -0.00000551, -0.00000061, -0.00000771, 0.00000137, -0.00000455, -0.00000188, -0.00000652, 0.00000096, 0.00000010, -0.00000936, -0.00000644, 0.00000271, 0.00000065, -0.00001047, -0.00000628, -0.00000613, -0.00000053, -0.00001688, -0.00001154, -0.00001392, -0.00000410, -0.00003791, -0.00001684, -0.00012070, -0.00003729, 0.00015877, 0.00003111, 0.00005752, 0.00004271, 0.00001888, 0.00002116, 0.00001960, 0.00004015, 0.00006743, 0.00016246, -0.00006092, -0.00010341, -0.00000256, -0.00002323, 0.00000012, -0.00001072, 0.00000145, -0.00000034, 0.00000351, -0.00001314, 0.00000175, -0.00000343, 0.00000834, 0.00000199, 0.00000196, 0.00000365, 0.00000932, 0.00000453, 0.00002341, -0.00000497, -0.00000616, },
		{ -0.00001369, -0.00001837, 0.00000071, -0.00001622, 0.00000518, -0.00001320, 0.00000579, -0.00000960, 0.00000039, -0.00000478, -0.00000069, -0.00000748, 0.00000831, -0.00000454, -0.00000159, -0.00000416, 0.00000006, -0.00000330, 0.00000203, -0.00000855, 0.00000409, -0.00000708, 0.00000171, -0.00000694, 0.00000732, -0.00000048, -0.00000737, -0.00000780, 0.00000963, -0.00000296, -0.00000472, -0.00000051, 0.00000322, -0.00000199, 0.00000044, 0.00000098, -0.00000198, -0.00000175, 0.00001084, -0.00000599, 0.00000173, -0.00000547, 0.00001173, -0.00000548, 0.00000406, -0.00000734, 0.00000603, -0.00000599, 0.00000695, -0.00000363, 0.00000044, -0.00000528, 0.00000527, -0.00000514, 0.00001416, -0.00000876, 0.00000570, -0.00000560, 0.00000273, -0.00000636, 0.00001228, -0.00000737, 0.00000981, -0.00001196, 0.00001649, -0.00001369, 0.00001029, -0.00000930, 0.00000157, -0.00000348, -0.00001065, 0.00000053, 0.00000306, -0.00000499, -0.00000047, -0.00000874, 0.00000191, -0.00000422, -0.00000198, -0.00000661, 0.00000248, -0.00000234, -0.00000842, -0.00000606, 0.00000263, 0.00000026, -0.00001154, -0.00000501, -0.00000586, -0.00000196, -0.00001616, -0.00001187, -0.00001516, -0.00000428, -0.00003903, -0.00002256, -0.00013852, 0.00003241, 0.00013726, 0.00002010, 0.00005489, 0.00003888, 0.00001701, 0.00002065, 0.00001546, 0.00003870, 0.00006363, 0.00015727, -0.00005348, -0.00010016, -0.00000325, -0.00002174, -0.00000101, -0.00000770, 0.00000099, -0.00000168, 0.00000478, -0.00001322, 0.00000123, -0.00000099, 0.00000715, 0.00000363, 0.00000110, 0.00000588, 0.00000760, 0.00001129, 0.00001857, -0.00000591, },
		{ -0.00000755, -0.00002042, -0.00001062, -0.00000373, -0.00001261, 0.00000269, -0.00001090, 0.00000484, -0.00001040, 0.00000098, -0.00000522, -0.00000055, -0.00000629, 0.00000825, -0.00000632, -0.00000076, -0.00000399, -0.00000047, -0.00000251, 0.00000123, -0.00000779, 0.00000348, -0.00000734, 0.00000200, -0.00000679, 0.00000933, -0.00000476, -0.00000651, -0.00000544, 0.00001023, -0.00000685, -0.00000190, -0.00000057, 0.00000332, -0.00000325, 0.00000198, 0.00000053, -0.00000368, 0.00000238, 0.00000775, -0.00000554, 0.00000051, -0.00000231, 0.00001039, -0.00000586, 0.00000331, -0.00000655, 0.00000577, -0.00000583, 0.00000742, -0.00000473, 0.00000024, -0.00000374, 0.00000345, -0.00000147, 0.00001162, -0.00000903, 0.00000590, -0.00000535, 0.00000188, -0.00000401, 0.00001004, -0.00000546, 0.00000704, -0.00000901, 0.00001505, -0.00001486, 0.00001412, -0.00001604, 0.00000835, -0.00001042, -0.00000599, 0.00000041, 0.00000227, -0.00000520, 0.00000023, -0.00001046, 0.00000355, -0.00000506, -0.00000113, -0.00000732, 0.00000457, -0.00000614, -0.00000532, -0.00000777, 0.00000446, -0.00000249, -0.00001000, -0.00000586, -0.00000372, -0.00000588, -0.00001213, -0.00001624, -0.00001181, -0.00001001, -0.00003319, -0.00004183, -0.00013188, 0.00009022, 0.00010391, 0.00002007, 0.00004748, 0.00003844, 0.00001362, 0.00002140, 0.00001045, 0.00003832, 0.00005876, 0.00015357, -0.00004733, -0.00009593, -0.00000484, -0.00001972, -0.00000254, -0.00000381, -0.00000090, -0.00000190, 0.00000558, -0.00001342, 0.00000145, 0.00000083, 0.00000625, 0.00000490, 0.00000103, 0.00000712, 0.00000746, 0.00001710, 0.00001285, },
		{ 0.00000757, -0.00001119, -0.00002394, -0.00000384, -0.00000862, -0.00000799, -0.00000115, -0.00000695, 0.00000182, -0.00000930, 0.00000025, -0.00000481, -0.00000085, -0.00000439, 0.00000686, -0.00000717, -0.00000008, -0.00000390, -0.00000096, -0.00000168, 0.00000019, -0.00000654, 0.00000204, -0.00000662, 0.00000140, -0.00000535, 0.00000971, -0.00000870, -0.00000459, -0.00000323, 0.00000950, -0.00000984, 0.00000150, -0.00000161, 0.00000394, -0.00000504, 0.00000475, -0.00000203, -0.00000276, 0.00000455, 0.00000525, -0.00000539, 0.00000016, 0.00000046, 0.00000854, -0.00000597, 0.00000269, -0.00000573, 0.00000538, -0.00000549, 0.00000775, -0.00000626, 0.00000118, -0.00000345, 0.00000283, 0.00000162, 0.00000839, -0.00000841, 0.00000582, -0.00000524, 0.00000157, -0.00000215, 0.00000778, -0.00000326, 0.00000356, -0.00000426, 0.00001048, -0.00001201, 0.00001360, -0.00001910, 0.00001284, -0.00001690, 0.00000021, -0.00000169, 0.00000248, -0.00000569, 0.00000084, -0.00001200, 0.00000535, -0.00000621, -0.00000013, -0.00000777, 0.00000599, -0.00000973, -0.00000157, -0.00001025, 0.00000685, -0.00000618, -0.00000708, -0.00000766, -0.00000118, -0.00001040, -0.00000697, -0.00002227, -0.00000649, -0.00001821, -0.00002455, -0.00006988, -0.00009997, 0.00012546, 0.00007081, 0.00002477, 0.00003828, 0.00003936, 0.00001028, 0.00002185, 0.00000610, 0.00003744, 0.00005426, 0.00015015, -0.00004134, -0.00009179, -0.00000640, -0.00001804, -0.00000348, -0.00000015, -0.00000333, -0.00000142, 0.00000605, -0.00001375, 0.00000258, 0.00000160, 0.00000615, 0.00000533, 0.00000221, 0.00000682, 0.00000988, 0.00002040, },
		{ 0.00002057, 0.00000318, -0.00001697, -0.00002303, -0.00000025, -0.00001165, -0.00000451, -0.00000439, -0.00000310, -0.00000181, -0.00000726, -0.00000123, -0.00000395, -0.00000113, -0.00000243, 0.00000464, -0.00000718, 0.00000047, -0.00000401, -0.00000116, -0.00000120, -0.00000058, -0.00000539, 0.00000034, -0.00000539, 0.00000028, -0.00000284, 0.00000818, -0.00001152, -0.00000207, -0.00000136, 0.00000748, -0.00001125, 0.00000450, -0.00000290, 0.00000444, -0.00000652, 0.00000769, -0.00000576, 0.00000038, 0.00000442, 0.00000394, -0.00000595, 0.00000111, 0.00000217, 0.00000676, -0.00000608, 0.00000237, -0.00000511, 0.00000505, -0.00000506, 0.00000779, -0.00000787, 0.00000297, -0.00000443, 0.00000386, 0.00000324, 0.00000541, -0.00000742, 0.00000567, -0.00000541, 0.00000207, -0.00000140, 0.00000641, -0.00000193, 0.00000090, 0.00000048, 0.00000451, -0.00000635, 0.00000907, -0.00001770, 0.00001326, -0.00002054, 0.00000575, -0.00000440, 0.00000298, -0.00000593, 0.00000080, -0.00001261, 0.00000641, -0.00000682, 0.00000031, -0.00000724, 0.00000573, -0.00001168, 0.00000123, -0.00001201, 0.00000824, -0.00000912, -0.00000440, -0.00000895, 0.00000012, -0.00001337, -0.00000328, -0.00002699, -0.00000263, -0.00002466, -0.00001933, -0.00009738, -0.00004970, 0.00013509, 0.00004619, 0.00002870, 0.00003026, 0.00003971, 0.00000826, 0.00002088, 0.00000351, 0.00003492, 0.00005124, 0.00014604, -0.00003470, -0.00008855, -0.00000721, -0.00001728, -0.00000312, 0.00000233, -0.00000543, -0.00000075, 0.00000630, -0.00001403, 0.00000438, 0.00000140, 0.00000686, 0.00000503, 0.00000445, 0.00000525, 0.00001480, },
		{ 0.00002123, 0.00001794, -0.00000085, -0.00002366, -0.00001825, -0.00000048, -0.00001166, -0.00000349, -0.00000561, -0.00000085, -0.00000464, -0.00000540, -0.00000270, -0.00000313, -0.00000090, -0.00000106, 0.00000222, -0.00000662, 0.00000092, -0.00000438, -0.00000091, -0.00000132, -0.00000068, -0.00000489, -0.00000096, -0.00000427, -0.00000070, 0.00000005, 0.00000499, -0.00001277, 0.00000062, -0.00000007, 0.00000458, -0.00001090, 0.00000645, -0.00000382, 0.00000430, -0.00000690, 0.00000964, -0.00000927, 0.00000459, 0.00000250, 0.00000377, -0.00000714, 0.00000331, 0.00000252, 0.00000545, -0.00000635, 0.00000244, -0.00000480, 0.00000491, -0.00000459, 0.00000740, -0.00000911, 0.00000502, -0.00000620, 0.00000638, 0.00000299, 0.00000336, -0.00000653, 0.00000560, -0.00000583, 0.00000332, -0.00000191, 0.00000636, -0.00000222, 0.00000031, 0.00000333, -0.00000060, -0.00000017, 0.00000242, -0.00001288, 0.00000958, -0.00002024, 0.00000910, -0.00000659, 0.00000320, -0.00000567, -0.00000007, -0.00001203, 0.00000634, -0.00000653, -0.00000007, -0.00000559, 0.00000351, -0.00001128, 0.00000216, -0.00001213, 0.00000764, -0.00001012, -0.00000311, -0.00000879, -0.00000079, -0.00001353, -0.00000274, -0.00002835, -0.00000270, -0.00002622, -0.00002284, -0.00011398, 0.00000700, 0.00012334, 0.00003296, 0.00002893, 0.00002499, 0.00003863, 0.00000786, 0.00001834, 0.00000276, 0.00003057, 0.00004997, 0.00014099, -0.00002722, -0.00008639, -0.00000714, -0.00001746, -0.00000139, 0.00000327, -0.00000666, -0.00000022, 0.00000636, -0.00001392, 0.00000628, 0.00000075, 0.00000788, 0.00000463, 0.00000698, 0.00000344, },
		{ 0.00000269, 0.00002763, 0.00001342, -0.00000544, -0.00002922, -0.00001158, -0.00000351, -0.00000901, -0.00000499, -0.00000446, -0.00000085, -0.00000583, -0.00000449, -0.00000361, -0.00000268, 0.00000005, -0.00000072, 0.00000013, -0.00000577, 0.00000125, -0.00000494, -0.00000029, -0.00000201, -0.00000004, -0.00000527, -0.00000143, -0.00000378, -0.00000089, 0.00000240, 0.00000084, -0.00001240, 0.00000315, 0.00000042, 0.00000146, -0.00000911, 0.00000717, -0.00000416, 0.00000340, -0.00000586, 0.00000981, -0.00001130, 0.00000845, -0.00000020, 0.00000419, -0.00000843, 0.00000625, 0.00000162, 0.00000470, -0.00000678, 0.00000279, -0.00000473, 0.00000495, -0.00000414, 0.00000649, -0.00000952, 0.00000661, -0.00000792, 0.00000966, 0.00000111, 0.00000250, -0.00000598, 0.00000564, -0.00000630, 0.00000494, -0.00000332, 0.00000744, -0.00000408, 0.00000217, 0.00000318, -0.00000298, 0.00000410, -0.00000373, -0.00000697, 0.00000334, -0.00001641, 0.00000979, -0.00000771, 0.00000299, -0.00000506, -0.00000143, -0.00001057, 0.00000536, -0.00000551, -0.00000093, -0.00000343, -0.00000001, -0.00000888, 0.00000127, -0.00001069, 0.00000509, -0.00000900, -0.00000344, -0.00000723, -0.00000364, -0.00001106, -0.00000538, -0.00002618, -0.00000699, -0.00002251, -0.00003681, -0.00011218, 0.00005730, 0.00009900, 0.00002938, 0.00002536, 0.00002240, 0.00003642, 0.00000846, 0.00001498, 0.00000307, 0.00002508, 0.00004991, 0.00013548, -0.00001936, -0.00008492, -0.00000660, -0.00001801, 0.00000116, 0.00000288, -0.00000692, 0.00000003, 0.00000620, -0.00001310, 0.00000767, 0.00000032, 0.00000859, 0.00000487, 0.00000884, },
		{ 0.00000941, 0.00000406, 0.00003251, 0.00000789, -0.00001125, -0.00003175, -0.00000553, -0.00000736, -0.00000515, -0.00000791, -0.00000170, -0.00000271, -0.00000535, -0.00000474, -0.00000379, -0.00000257, 0.00000159, -0.00000149, -0.00000130, -0.00000480, 0.00000137, -0.00000544, 0.00000042, -0.00000297, 0.00000107, -0.00000641, -0.00000103, -0.00000411, 0.00000010, 0.00000342, -0.00000336, -0.00001069, 0.00000529, -0.00000003, -0.00000119, -0.00000655, 0.00000693, -0.00000409, 0.00000206, -0.00000371, 0.00000811, -0.00001116, 0.00001084, -0.00000264, 0.00000452, -0.00000914, 0.00000917, -0.00000008, 0.00000440, -0.00000725, 0.00000329, -0.00000480, 0.00000512, -0.00000376, 0.00000513, -0.00000892, 0.00000718, -0.00000874, 0.00001270, -0.00000172, 0.00000265, -0.00000579, 0.00000571, -0.00000654, 0.00000636, -0.00000493, 0.00000889, -0.00000669, 0.00000580, 0.00000014, -0.00000193, 0.00000497, -0.00000717, -0.00000255, -0.00000309, -0.00001059, 0.00000831, -0.00000784, 0.00000251, -0.00000448, -0.00000269, -0.00000894, 0.00000408, -0.00000435, -0.00000159, -0.00000172, -0.00000362, -0.00000558, -0.00000056, -0.00000856, 0.00000151, -0.00000649, -0.00000472, -0.00000516, -0.00000719, -0.00000742, -0.00000968, -0.00002212, -0.00001354, -0.00001625, -0.00005829, -0.00009005, 0.00009157, 0.00007174, 0.00003121, 0.00001970, 0.00002133, 0.00003402, 0.00000899, 0.00001193, 0.00000326, 0.00001957, 0.00005009, 0.00013033, -0.00001194, -0.00008344, -0.00000621, -0.00001819, 0.00000369, 0.00000181, -0.00000644, -0.00000001, 0.00000589, -0.00001148, 0.00000818, 0.00000055, 0.00000857, 0.00000624, },
		{ 0.00000867, 0.00000872, 0.00000786, 0.00003489, 0.00000175, -0.00001806, -0.00003037, -0.00000206, -0.00001000, -0.00000185, -0.00001070, 0.00000130, -0.00000539, -0.00000387, -0.00000582, -0.00000342, -0.00000255, 0.00000330, -0.00000310, -0.00000201, -0.00000381, 0.00000118, -0.00000566, 0.00000086, -0.00000377, 0.00000218, -0.00000785, -0.00000009, -0.00000503, 0.00000222, 0.00000263, -0.00000671, -0.00000814, 0.00000688, -0.00000136, -0.00000281, -0.00000394, 0.00000622, -0.00000402, 0.00000092, -0.00000122, 0.00000520, -0.00000901, 0.00001128, -0.00000414, 0.00000427, -0.00000873, 0.00001134, -0.00000200, 0.00000435, -0.00000764, 0.00000376, -0.00000488, 0.00000537, -0.00000358, 0.00000358, -0.00000744, 0.00000658, -0.00000813, 0.00001457, -0.00000461, 0.00000338, -0.00000579, 0.00000566, -0.00000631, 0.00000712, -0.00000602, 0.00000977, -0.00000881, 0.00000976, -0.00000455, 0.00000184, 0.00000236, -0.00000695, -0.00000137, -0.00000747, -0.00000471, 0.00000572, -0.00000739, 0.00000207, -0.00000433, -0.00000329, -0.00000779, 0.00000315, -0.00000362, -0.00000145, -0.00000137, -0.00000605, -0.00000269, -0.00000218, -0.00000695, -0.00000179, -0.00000378, -0.00000594, -0.00000378, -0.00000986, -0.00000452, -0.00001346, -0.00001865, -0.00001930, -0.00001199, -0.00008052, -0.00005182, 0.00010581, 0.00004898, 0.00003408, 0.00001407, 0.00002046, 0.00003228, 0.00000864, 0.00001000, 0.00000246, 0.00001495, 0.00004971, 0.00012626, -0.00000562, -0.00008136, -0.00000645, -0.00001740, 0.00000540, 0.00000074, -0.00000557, -0.00000031, 0.00000563, -0.00000928, 0.00000788, 0.00000145, 0.00000789, },
		{ 0.00000705, 0.00001156, 0.00000750, 0.00001360, 0.00003445, -0.00000506, -0.00002467, -0.00002565, -0.00000184, -0.00001033, -0.00000034, -0.00001211, 0.00000328, -0.00000769, -0.00000228, -0.00000713, -0.00000284, -0.00000228, 0.00000469, -0.00000512, -0.00000209, -0.00000289, 0.00000067, -0.00000545, 0.00000082, -0.00000406, 0.00000282, -0.00000903, 0.00000087, -0.00000599, 0.00000494, 0.00000006, -0.00000857, -0.00000532, 0.00000784, -0.00000331, -0.00000318, -0.00000185, 0.00000545, -0.00000429, 0.00000059, 0.00000062, 0.00000216, -0.00000567, 0.00001003, -0.00000458, 0.00000336, -0.00000708, 0.00001227, -0.00000367, 0.00000431, -0.00000784, 0.00000413, -0.00000490, 0.00000568, -0.00000376, 0.00000225, -0.00000556, 0.00000512, -0.00000606, 0.00001475, -0.00000680, 0.00000418, -0.00000576, 0.00000544, -0.00000557, 0.00000701, -0.00000619, 0.00000944, -0.00000934, 0.00001245, -0.00000903, 0.00000659, -0.00000241, -0.00000364, -0.00000380, -0.00000853, -0.00000033, 0.00000308, -0.00000683, 0.00000185, -0.00000470, -0.00000303, -0.00000746, 0.00000288, -0.00000353, -0.00000036, -0.00000270, -0.00000656, -0.00000114, -0.00000275, -0.00000673, -0.00000375, -0.00000189, -0.00000629, -0.00000388, -0.00001061, -0.00000374, -0.00001507, -0.00001775, -0.00002178, -0.00001394, -0.00009542, -0.00000619, 0.00010186, 0.00003431, 0.00003514, 0.00000989, 0.00001907, 0.00003146, 0.00000725, 0.00000933, 0.00000045, 0.00001151, 0.00004848, 0.00012349, -0.00000068, -0.00007849, -0.00000739, -0.00001549, 0.00000587, 0.00000020, -0.00000460, -0.00000086, 0.00000571, -0.00000698, 0.00000720, 0.00000263, },
		{ 0.00000358, 0.00000672, 0.00001408, 0.00000685, 0.00002020, 0.00003130, -0.00001276, -0.00002942, -0.00001934, -0.00000414, -0.00000842, -0.00000086, -0.00001168, 0.00000358, -0.00000882, -0.00000131, -0.00000817, -0.00000234, -0.00000159, 0.00000538, -0.00000711, -0.00000166, -0.00000217, -0.00000002, -0.00000489, 0.00000028, -0.00000375, 0.00000273, -0.00000954, 0.00000135, -0.00000638, 0.00000748, -0.00000380, -0.00000867, -0.00000274, 0.00000816, -0.00000552, -0.00000240, -0.00000055, 0.00000488, -0.00000502, 0.00000137, 0.00000110, 0.00000010, -0.00000231, 0.00000786, -0.00000434, 0.00000218, -0.00000449, 0.00001189, -0.00000481, 0.00000417, -0.00000784, 0.00000439, -0.00000485, 0.00000603, -0.00000441, 0.00000156, -0.00000391, 0.00000349, -0.00000306, 0.00001326, -0.00000790, 0.00000471, -0.00000559, 0.00000507, -0.00000451, 0.00000622, -0.00000552, 0.00000784, -0.00000783, 0.00001275, -0.00001160, 0.00001032, -0.00000728, 0.00000099, -0.00000882, -0.00000623, 0.00000177, 0.00000108, -0.00000640, 0.00000179, -0.00000539, -0.00000213, -0.00000781, 0.00000316, -0.00000384, 0.00000125, -0.00000531, -0.00000523, -0.00000112, -0.00000211, -0.00000808, -0.00000400, -0.00000120, -0.00000563, -0.00000544, -0.00000938, -0.00000532, -0.00001412, -0.00001989, -0.00002040, -0.00002384, -0.00009662, 0.00003674, 0.00008572, 0.00002754, 0.00003356, 0.00000749, 0.00001716, 0.00003120, 0.00000532, 0.00000937, -0.00000229, 0.00000890, 0.00004668, 0.00012172, 0.00000311, -0.00007504, -0.00000870, -0.00001279, 0.00000515, 0.00000043, -0.00000376, -0.00000161, 0.00000638, -0.00000507, 0.00000666, },
		{ 0.00000663, 0.00000394, 0.00000738, 0.00001559, 0.00000780, 0.00002638, 0.00002569, -0.00002120, -0.00003093, -0.00001359, -0.00000734, -0.00000534, -0.00000267, -0.00000988, 0.00000234, -0.00000863, -0.00000126, -0.00000867, -0.00000200, -0.00000051, 0.00000524, -0.00000877, -0.00000084, -0.00000179, -0.00000065, -0.00000423, -0.00000050, -0.00000304, 0.00000199, -0.00000929, 0.00000110, -0.00000575, 0.00000901, -0.00000813, -0.00000712, -0.00000080, 0.00000791, -0.00000756, -0.00000079, -0.00000003, 0.00000447, -0.00000598, 0.00000307, 0.00000004, -0.00000029, 0.00000004, 0.00000569, -0.00000402, 0.00000132, -0.00000160, 0.00001047, -0.00000534, 0.00000389, -0.00000768, 0.00000458, -0.00000475, 0.00000636, -0.00000550, 0.00000174, -0.00000302, 0.00000246, -0.00000000, 0.00001062, -0.00000790, 0.00000486, -0.00000526, 0.00000468, -0.00000343, 0.00000518, -0.00000452, 0.00000556, -0.00000469, 0.00001052, -0.00001145, 0.00001160, -0.00001035, 0.00000480, -0.00001449, -0.00000165, 0.00000167, -0.00000004, -0.00000615, 0.00000169, -0.00000600, -0.00000107, -0.00000841, 0.00000356, -0.00000403, 0.00000261, -0.00000825, -0.00000282, -0.00000211, -0.00000080, -0.00001044, -0.00000292, -0.00000144, -0.00000445, -0.00000769, -0.00000707, -0.00000834, -0.00001161, -0.00002387, -0.00001693, -0.00003989, -0.00008176, 0.00006868, 0.00006479, 0.00002612, 0.00003007, 0.00000637, 0.00001520, 0.00003083, 0.00000360, 0.00000931, -0.00000494, 0.00000642, 0.00004494, 0.00012033, 0.00000626, -0.00007143, -0.00000983, -0.00000993, 0.00000363, 0.00000137, -0.00000317, -0.00000244, 0.00000766, -0.00000384, },
		{ -0.00000321, 0.00000713, 0.00000373, 0.00000906, 0.00001591, 0.00001096, 0.00003097, 0.00001779, -0.00002957, -0.00002880, -0.00001004, -0.00000979, -0.00000248, -0.00000463, -0.00000766, 0.00000029, -0.00000752, -0.00000200, -0.00000862, -0.00000173, 0.00000072, 0.00000436, -0.00000990, 0.00000021, -0.00000188, -0.00000095, -0.00000379, -0.00000118, -0.00000229, 0.00000097, -0.00000851, 0.00000027, -0.00000402, 0.00000893, -0.00001196, -0.00000441, 0.00000030, 0.00000723, -0.00000912, 0.00000121, -0.00000010, 0.00000407, -0.00000676, 0.00000512, -0.00000216, 0.00000104, 0.00000085, 0.00000419, -0.00000414, 0.00000134, 0.00000086, 0.00000851, -0.00000541, 0.00000349, -0.00000739, 0.00000473, -0.00000460, 0.00000657, -0.00000683, 0.00000275, -0.00000315, 0.00000263, 0.00000222, 0.00000761, -0.00000716, 0.00000471, -0.00000486, 0.00000442, -0.00000264, 0.00000438, -0.00000387, 0.00000354, -0.00000100, 0.00000658, -0.00000887, 0.00001009, -0.00001060, 0.00000612, -0.00001881, 0.00000361, 0.00000006, -0.00000041, -0.00000600, 0.00000137, -0.00000616, -0.00000033, -0.00000879, 0.00000366, -0.00000360, 0.00000293, -0.00001044, -0.00000036, -0.00000328, 0.00000034, -0.00001284, -0.00000133, -0.00000194, -0.00000351, -0.00000957, -0.00000500, -0.00001132, -0.00000923, -0.00002756, -0.00001463, -0.00005726, -0.00005326, 0.00008538, 0.00004540, 0.00002685, 0.00002596, 0.00000578, 0.00001367, 0.00002981, 0.00000266, 0.00000849, -0.00000682, 0.00000343, 0.00004383, 0.00011871, 0.00000929, -0.00006806, -0.00001029, -0.00000754, 0.00000185, 0.00000283, -0.00000287, -0.00000313, 0.00000933, },
		{ 0.00001100, -0.00000284, 0.00000783, 0.00000336, 0.00001135, 0.00001548, 0.00001632, 0.00003306, 0.00000775, -0.00003646, -0.00002384, -0.00000924, -0.00001051, -0.00000087, -0.00000571, -0.00000602, -0.00000168, -0.00000614, -0.00000314, -0.00000820, -0.00000138, 0.00000179, 0.00000297, -0.00001041, 0.00000131, -0.00000240, -0.00000081, -0.00000378, -0.00000145, -0.00000192, 0.00000012, -0.00000764, -0.00000074, -0.00000153, 0.00000703, -0.00001448, -0.00000121, 0.00000058, 0.00000631, -0.00001001, 0.00000322, -0.00000057, 0.00000355, -0.00000693, 0.00000679, -0.00000470, 0.00000353, 0.00000015, 0.00000362, -0.00000486, 0.00000254, 0.00000230, 0.00000654, -0.00000522, 0.00000305, -0.00000701, 0.00000485, -0.00000441, 0.00000650, -0.00000807, 0.00000424, -0.00000413, 0.00000413, 0.00000298, 0.00000501, -0.00000615, 0.00000450, -0.00000452, 0.00000436, -0.00000228, 0.00000409, -0.00000407, 0.00000269, 0.00000191, 0.00000237, -0.00000505, 0.00000660, -0.00000826, 0.00000439, -0.00002043, 0.00000803, -0.00000214, -0.00000031, -0.00000588, 0.00000079, -0.00000576, -0.00000016, -0.00000870, 0.00000327, -0.00000240, 0.00000186, -0.00001116, 0.00000130, -0.00000392, 0.00000058, -0.00001443, -0.00000003, -0.00000213, -0.00000338, -0.00001030, -0.00000421, -0.00001302, -0.00000844, -0.00002914, -0.00001667, -0.00006977, -0.00001730, 0.00008708, 0.00003126, 0.00002725, 0.00002235, 0.00000522, 0.00001272, 0.00002806, 0.00000259, 0.00000676, -0.00000766, -0.00000032, 0.00004361, 0.00011653, 0.00001247, -0.00006505, -0.00000989, -0.00000604, 0.00000030, 0.00000452, -0.00000288, -0.00000341, },
		{ -0.00000313, 0.00001230, -0.00000234, 0.00000833, 0.00000339, 0.00001357, 0.00001510, 0.00002334, 0.00003197, -0.00000402, -0.00004034, -0.00001780, -0.00001053, -0.00000950, -0.00000076, -0.00000551, -0.00000552, -0.00000290, -0.00000510, -0.00000426, -0.00000759, -0.00000087, 0.00000245, 0.00000134, -0.00001027, 0.00000221, -0.00000316, -0.00000032, -0.00000419, -0.00000123, -0.00000212, -0.00000020, -0.00000712, -0.00000138, 0.00000108, 0.00000359, -0.00001518, 0.00000174, 0.00000025, 0.00000530, -0.00001018, 0.00000498, -0.00000130, 0.00000290, -0.00000632, 0.00000751, -0.00000668, 0.00000631, -0.00000151, 0.00000375, -0.00000599, 0.00000479, 0.00000243, 0.00000501, -0.00000503, 0.00000267, -0.00000657, 0.00000493, -0.00000415, 0.00000605, -0.00000886, 0.00000571, -0.00000543, 0.00000665, 0.00000212, 0.00000337, -0.00000528, 0.00000444, -0.00000435, 0.00000450, -0.00000231, 0.00000430, -0.00000517, 0.00000345, 0.00000303, -0.00000069, -0.00000156, 0.00000265, -0.00000460, 0.00000023, -0.00001899, 0.00001063, -0.00000416, -0.00000003, -0.00000581, 0.00000010, -0.00000498, -0.00000044, -0.00000823, 0.00000258, -0.00000073, -0.00000037, -0.00001029, 0.00000181, -0.00000378, -0.00000033, -0.00001476, 0.00000055, -0.00000178, -0.00000416, -0.00000975, -0.00000495, -0.00001308, -0.00000965, -0.00002811, -0.00002452, -0.00007228, 0.00001841, 0.00007759, 0.00002322, 0.00002613, 0.00001973, 0.00000456, 0.00001209, 0.00002592, 0.00000301, 0.00000448, -0.00000774, -0.00000464, 0.00004408, 0.00011384, 0.00001579, -0.00006227, -0.00000878, -0.00000555, -0.00000071, 0.00000621, -0.00000309, },
		{ -0.00000333, -0.00000234, 0.00001305, -0.00000145, 0.00000835, 0.00000427, 0.00001511, 0.00001572, 0.00003109, 0.00002714, -0.00001652, -0.00004030, -0.00001258, -0.00001258, -0.00000756, -0.00000167, -0.00000429, -0.00000614, -0.00000318, -0.00000471, -0.00000507, -0.00000691, -0.00000020, 0.00000258, -0.00000028, -0.00000951, 0.00000265, -0.00000387, 0.00000029, -0.00000483, -0.00000068, -0.00000284, 0.00000011, -0.00000720, -0.00000118, 0.00000302, -0.00000069, -0.00001399, 0.00000385, -0.00000036, 0.00000427, -0.00000968, 0.00000629, -0.00000220, 0.00000230, -0.00000507, 0.00000706, -0.00000749, 0.00000850, -0.00000338, 0.00000410, -0.00000703, 0.00000762, 0.00000134, 0.00000416, -0.00000502, 0.00000241, -0.00000610, 0.00000496, -0.00000386, 0.00000519, -0.00000894, 0.00000665, -0.00000639, 0.00000948, -0.00000004, 0.00000283, -0.00000480, 0.00000464, -0.00000440, 0.00000476, -0.00000253, 0.00000468, -0.00000680, 0.00000561, 0.00000205, -0.00000174, 0.00000035, -0.00000020, -0.00000132, -0.00000487, -0.00001513, 0.00001114, -0.00000553, 0.00000025, -0.00000586, -0.00000043, -0.00000415, -0.00000086, -0.00000772, 0.00000203, 0.00000078, -0.00000304, -0.00000829, 0.00000139, -0.00000311, -0.00000210, -0.00001396, 0.00000041, -0.00000108, -0.00000540, -0.00000846, -0.00000669, -0.00001208, -0.00001211, -0.00002565, -0.00003706, -0.00006248, 0.00004705, 0.00006230, 0.00002009, 0.00002350, 0.00001805, 0.00000398, 0.00001138, 0.00002390, 0.00000334, 0.00000225, -0.00000759, -0.00000902, 0.00004477, 0.00011093, 0.00001904, -0.00005938, -0.00000736, -0.00000587, -0.00000107, 0.00000768, },
		{ 0.00000879, -0.00000334, -0.00000127, 0.00001334, -0.00000017, 0.00000792, 0.00000609, 0.00001567, 0.00001813, 0.00003841, 0.00001820, -0.00002809, -0.00003649, -0.00000951, -0.00001407, -0.00000571, -0.00000279, -0.00000269, -0.00000739, -0.00000273, -0.00000500, -0.00000544, -0.00000618, 0.00000050, 0.00000214, -0.00000163, -0.00000824, 0.00000249, -0.00000424, 0.00000072, -0.00000539, -0.00000010, -0.00000380, 0.00000089, -0.00000783, 0.00000005, 0.00000358, -0.00000488, -0.00001128, 0.00000482, -0.00000093, 0.00000326, -0.00000857, 0.00000705, -0.00000323, 0.00000201, -0.00000356, 0.00000568, -0.00000698, 0.00000956, -0.00000483, 0.00000419, -0.00000743, 0.00001034, -0.00000060, 0.00000398, -0.00000528, 0.00000238, -0.00000565, 0.00000497, -0.00000363, 0.00000406, -0.00000829, 0.00000677, -0.00000644, 0.00001179, -0.00000285, 0.00000319, -0.00000470, 0.00000510, -0.00000467, 0.00000507, -0.00000270, 0.00000483, -0.00000825, 0.00000838, -0.00000054, -0.00000076, 0.00000021, -0.00000102, 0.00000011, -0.00000919, -0.00001012, 0.00000992, -0.00000611, 0.00000042, -0.00000607, -0.00000059, -0.00000360, -0.00000107, -0.00000749, 0.00000200, 0.00000147, -0.00000528, -0.00000596, 0.00000057, -0.00000245, -0.00000408, -0.00001252, -0.00000012, -0.00000043, -0.00000650, -0.00000725, -0.00000850, -0.00001112, -0.00001440, -0.00002409, -0.00005080, -0.00004163, 0.00006457, 0.00004630, 0.00001975, 0.00001996, 0.00001701, 0.00000366, 0.00001029, 0.00002235, 0.00000314, 0.00000060, -0.00000773, -0.00001295, 0.00004514, 0.00010819, 0.00002200, -0.00005605, -0.00000612, -0.00000662, -0.00000084, },
		{ -0.00000024, 0.00000953, -0.00000290, -0.00000024, 0.00001346, 0.00000122, 0.00000734, 0.00000861, 0.00001538, 0.00002276, 0.00004392, 0.00000526, -0.00003677, -0.00003024, -0.00000884, -0.00001427, -0.00000468, -0.00000342, -0.00000140, -0.00000869, -0.00000198, -0.00000575, -0.00000536, -0.00000546, 0.00000113, 0.00000119, -0.00000249, -0.00000675, 0.00000176, -0.00000415, 0.00000079, -0.00000563, 0.00000020, -0.00000463, 0.00000179, -0.00000872, 0.00000215, 0.00000238, -0.00000801, -0.00000781, 0.00000469, -0.00000127, 0.00000230, -0.00000696, 0.00000716, -0.00000427, 0.00000220, -0.00000230, 0.00000391, -0.00000546, 0.00000940, -0.00000555, 0.00000374, -0.00000678, 0.00001230, -0.00000282, 0.00000430, -0.00000580, 0.00000261, -0.00000529, 0.00000500, -0.00000358, 0.00000292, -0.00000712, 0.00000611, -0.00000530, 0.00001291, -0.00000553, 0.00000402, -0.00000478, 0.00000567, -0.00000508, 0.00000538, -0.00000271, 0.00000443, -0.00000886, 0.00001075, -0.00000374, 0.00000152, -0.00000157, 0.00000019, -0.00000096, -0.00001138, -0.00000536, 0.00000767, -0.00000602, 0.00000041, -0.00000637, -0.00000038, -0.00000347, -0.00000092, -0.00000766, 0.00000266, 0.00000098, -0.00000641, -0.00000401, -0.00000010, -0.00000234, -0.00000561, -0.00001101, -0.00000066, -0.00000016, -0.00000697, -0.00000676, -0.00000958, -0.00001112, -0.00001529, -0.00002570, -0.00006116, -0.00001403, 0.00007029, 0.00003307, 0.00002026, 0.00001624, 0.00001635, 0.00000358, 0.00000883, 0.00002129, 0.00000234, -0.00000031, -0.00000840, -0.00001615, 0.00004482, 0.00010593, 0.00002452, -0.00005211, -0.00000543, -0.00000738, },
		{ -0.00000780, 0.00000048, 0.00001001, -0.00000194, 0.00000047, 0.00001371, 0.00000241, 0.00000708, 0.00001134, 0.00001473, 0.00002967, 0.00004614, -0.00001077, -0.00004094, -0.00002352, -0.00000982, -0.00001327, -0.00000462, -0.00000331, -0.00000077, -0.00000958, -0.00000134, -0.00000671, -0.00000486, -0.00000478, 0.00000160, -0.00000020, -0.00000263, -0.00000541, 0.00000071, -0.00000370, 0.00000050, -0.00000551, 0.00000006, -0.00000505, 0.00000245, -0.00000942, 0.00000463, -0.00000056, -0.00000931, -0.00000452, 0.00000387, -0.00000140, 0.00000148, -0.00000503, 0.00000660, -0.00000514, 0.00000291, -0.00000168, 0.00000238, -0.00000354, 0.00000835, -0.00000562, 0.00000283, -0.00000503, 0.00001304, -0.00000477, 0.00000480, -0.00000647, 0.00000309, -0.00000509, 0.00000511, -0.00000383, 0.00000207, -0.00000583, 0.00000504, -0.00000314, 0.00001252, -0.00000747, 0.00000487, -0.00000477, 0.00000613, -0.00000553, 0.00000569, -0.00000262, 0.00000347, -0.00000828, 0.00001190, -0.00000644, 0.00000400, -0.00000397, 0.00000259, -0.00000424, -0.00001086, -0.00000194, 0.00000516, -0.00000547, 0.00000019, -0.00000661, 0.00000004, -0.00000364, -0.00000051, -0.00000808, 0.00000377, -0.00000058, -0.00000619, -0.00000285, -0.00000037, -0.00000300, -0.00000628, -0.00000986, -0.00000102, -0.00000032, -0.00000677, -0.00000713, -0.00000975, -0.00001232, -0.00001453, -0.00003155, -0.00006414, 0.00001461, 0.00006634, 0.00002408, 0.00002034, 0.00001281, 0.00001598, 0.00000355, 0.00000723, 0.00002044, 0.00000122, -0.00000068, -0.00000946, -0.00001868, 0.00004367, 0.00010426, 0.00002666, -0.00004758, -0.00000550, },
		{ -0.00000629, -0.00000763, 0.00000102, 0.00001050, -0.00000059, 0.00000083, 0.00001421, 0.00000323, 0.00000749, 0.00001376, 0.00001447, 0.00003845, 0.00004352, -0.00002801, -0.00003997, -0.00001821, -0.00001122, -0.00001173, -0.00000514, -0.00000265, -0.00000076, -0.00000993, -0.00000107, -0.00000764, -0.00000401, -0.00000423, 0.00000185, -0.00000190, -0.00000193, -0.00000459, -0.00000028, -0.00000315, 0.00000005, -0.00000517, -0.00000043, -0.00000498, 0.00000263, -0.00000950, 0.00000680, -0.00000475, -0.00000847, -0.00000225, 0.00000297, -0.00000156, 0.00000103, -0.00000308, 0.00000543, -0.00000564, 0.00000397, -0.00000184, 0.00000158, -0.00000187, 0.00000696, -0.00000538, 0.00000183, -0.00000249, 0.00001242, -0.00000603, 0.00000515, -0.00000710, 0.00000375, -0.00000503, 0.00000528, -0.00000443, 0.00000174, -0.00000484, 0.00000409, -0.00000048, 0.00001075, -0.00000834, 0.00000538, -0.00000443, 0.00000625, -0.00000587, 0.00000602, -0.00000262, 0.00000225, -0.00000661, 0.00001144, -0.00000786, 0.00000569, -0.00000585, 0.00000489, -0.00000861, -0.00000792, -0.00000039, 0.00000303, -0.00000473, -0.00000025, -0.00000660, 0.00000039, -0.00000383, -0.00000013, -0.00000839, 0.00000483, -0.00000268, -0.00000486, -0.00000249, -0.00000029, -0.00000433, -0.00000605, -0.00000921, -0.00000120, -0.00000070, -0.00000621, -0.00000803, -0.00000939, -0.00001413, -0.00001304, -0.00004077, -0.00005780, 0.00003897, 0.00005640, 0.00001912, 0.00001952, 0.00000991, 0.00001586, 0.00000334, 0.00000581, 0.00001942, 0.00000024, -0.00000093, -0.00001053, -0.00002084, 0.00004183, 0.00010314, 0.00002856, -0.00004270, },
		{ -0.00003778, -0.00000761, -0.00000683, 0.00000117, 0.00001126, 0.00000088, 0.00000098, 0.00001489, 0.00000373, 0.00000871, 0.00001551, 0.00001542, 0.00004812, 0.00003491, -0.00004376, -0.00003468, -0.00001532, -0.00001198, -0.00001041, -0.00000569, -0.00000189, -0.00000107, -0.00000985, -0.00000124, -0.00000832, -0.00000291, -0.00000384, 0.00000182, -0.00000368, -0.00000045, -0.00000455, -0.00000090, -0.00000279, -0.00000029, -0.00000488, -0.00000105, -0.00000456, 0.00000234, -0.00000868, 0.00000793, -0.00000930, -0.00000573, -0.00000149, 0.00000258, -0.00000206, 0.00000123, -0.00000148, 0.00000388, -0.00000562, 0.00000508, -0.00000264, 0.00000166, -0.00000093, 0.00000577, -0.00000526, 0.00000124, 0.00000027, 0.00001069, -0.00000645, 0.00000512, -0.00000749, 0.00000445, -0.00000509, 0.00000546, -0.00000529, 0.00000202, -0.00000445, 0.00000380, 0.00000197, 0.00000809, -0.00000820, 0.00000545, -0.00000368, 0.00000587, -0.00000594, 0.00000634, -0.00000288, 0.00000121, -0.00000438, 0.00000959, -0.00000778, 0.00000611, -0.00000644, 0.00000588, -0.00001266, -0.00000348, -0.00000068, 0.00000163, -0.00000407, -0.00000081, -0.00000628, 0.00000047, -0.00000383, -0.00000001, -0.00000825, 0.00000529, -0.00000464, -0.00000295, -0.00000266, -0.00000015, -0.00000593, -0.00000516, -0.00000898, -0.00000130, -0.00000103, -0.00000573, -0.00000889, -0.00000921, -0.00001559, -0.00001252, -0.00005079, -0.00004283, 0.00005555, 0.00004430, 0.00001709, 0.00001783, 0.00000760, 0.00001593, 0.00000285, 0.00000475, 0.00001798, -0.00000025, -0.00000147, -0.00001120, -0.00002308, 0.00003954, 0.00010246, 0.00003044, },
		{ 0.00003245, -0.00003316, -0.00000911, -0.00000550, 0.00000089, 0.00001250, 0.00000219, 0.00000116, 0.00001553, 0.00000414, 0.00001065, 0.00001645, 0.00001844, 0.00005706, 0.00002005, -0.00005517, -0.00002707, -0.00001475, -0.00001169, -0.00000969, -0.00000590, -0.00000139, -0.00000135, -0.00000957, -0.00000184, -0.00000858, -0.00000178, -0.00000358, 0.00000140, -0.00000518, 0.00000150, -0.00000524, -0.00000099, -0.00000279, -0.00000034, -0.00000486, -0.00000150, -0.00000403, 0.00000175, -0.00000698, 0.00000744, -0.00001314, -0.00000184, -0.00000223, 0.00000300, -0.00000315, 0.00000232, -0.00000066, 0.00000232, -0.00000511, 0.00000596, -0.00000377, 0.00000251, -0.00000087, 0.00000510, -0.00000557, 0.00000153, 0.00000254, 0.00000838, -0.00000612, 0.00000462, -0.00000747, 0.00000504, -0.00000517, 0.00000554, -0.00000623, 0.00000278, -0.00000471, 0.00000447, 0.00000352, 0.00000525, -0.00000737, 0.00000521, -0.00000262, 0.00000499, -0.00000561, 0.00000653, -0.00000348, 0.00000079, -0.00000227, 0.00000696, -0.00000653, 0.00000539, -0.00000562, 0.00000491, -0.00001512, 0.00000124, -0.00000234, 0.00000107, -0.00000369, -0.00000132, -0.00000574, 0.00000025, -0.00000354, -0.00000025, -0.00000752, 0.00000482, -0.00000585, -0.00000104, -0.00000301, -0.00000028, -0.00000735, -0.00000397, -0.00000903, -0.00000141, -0.00000117, -0.00000561, -0.00000936, -0.00000966, -0.00001601, -0.00001453, -0.00005822, -0.00002216, 0.00006322, 0.00003300, 0.00001666, 0.00001555, 0.00000590, 0.00001602, 0.00000219, 0.00000395, 0.00001615, -0.00000015, -0.00000250, -0.00001117, -0.00002576, 0.00003708, 0.00010210, },
		},
	}
	;

	float f64_4[16][64][64] = {
		{ { -0.00030959, -0.00004182, 0.00013342, 0.00006017, 0.00002491, 0.00005387, -0.00005337, 0.00003767, 0.00005464, 0.00002606, 0.00005868, 0.00006603, -0.00000795, 0.00001538, -0.00007084, -0.00009570, 0.00003627, 0.00002447, -0.00004196, -0.00000980, -0.00001628, -0.00003032, 0.00001883, -0.00003884, 0.00001935, 0.00001841, -0.00003233, 0.00000483, 0.00006797, -0.00005860, 0.00000280, -0.00000128, -0.00001770, -0.00000128, 0.00000280, -0.00005860, 0.00006797, 0.00000483, -0.00003233, 0.00001841, 0.00001935, -0.00003884, 0.00001883, -0.00003032, -0.00001628, -0.00000980, -0.00004196, 0.00002447, 0.00003627, -0.00009570, -0.00007084, 0.00001538, -0.00000795, 0.00006603, 0.00005868, 0.00002606, 0.00005464, 0.00003767, -0.00005337, 0.00005387, 0.00002491, 0.00006017, 0.00013342, -0.00004182, },
		{ -0.00003248, -0.00031173, -0.00004876, 0.00011149, 0.00006182, 0.00001330, 0.00005917, -0.00005563, 0.00001979, 0.00005577, 0.00002170, 0.00003736, 0.00007702, -0.00000335, 0.00001133, -0.00002558, -0.00010173, -0.00000619, 0.00004813, -0.00002778, -0.00002837, 0.00000449, -0.00004476, 0.00001296, -0.00002053, -0.00000124, 0.00002307, -0.00000706, -0.00004297, 0.00008955, -0.00003418, -0.00002212, 0.00000683, -0.00001204, -0.00001680, 0.00002045, -0.00005609, 0.00002554, 0.00004047, -0.00002775, -0.00000642, 0.00003995, -0.00004144, 0.00000557, -0.00000676, -0.00003103, -0.00000664, -0.00003485, -0.00000775, 0.00005413, -0.00005912, -0.00010774, -0.00000471, -0.00000704, 0.00005321, 0.00007350, 0.00003888, 0.00005087, 0.00005315, -0.00004339, 0.00004634, 0.00003449, 0.00005970, 0.00015609, },
		{ 0.00017918, -0.00001953, -0.00031771, -0.00005456, 0.00009002, 0.00006501, 0.00000204, 0.00005933, -0.00005214, 0.00000469, 0.00005043, 0.00002496, 0.00001564, 0.00007757, 0.00001433, -0.00000668, 0.00000278, -0.00007252, -0.00005017, 0.00004243, 0.00000808, -0.00004893, 0.00001463, -0.00003975, -0.00000802, -0.00000360, -0.00000256, 0.00000036, 0.00003467, -0.00007737, 0.00007708, 0.00000767, -0.00003861, 0.00000130, 0.00000029, -0.00002819, 0.00002092, -0.00003252, -0.00001489, 0.00004674, 0.00000124, -0.00003141, 0.00004320, -0.00002483, -0.00001603, 0.00000961, -0.00002930, -0.00001768, -0.00002068, -0.00002810, 0.00004287, -0.00001393, -0.00011627, -0.00004169, -0.00001473, 0.00004275, 0.00008127, 0.00005703, 0.00004930, 0.00006335, -0.00002716, 0.00004034, 0.00004130, 0.00005926, },
		{ 0.00005810, 0.00020155, -0.00000236, -0.00032629, -0.00006019, 0.00006848, 0.00006908, -0.00000635, 0.00005436, -0.00004778, -0.00000331, 0.00003928, 0.00002992, 0.00000179, 0.00006337, 0.00003982, -0.00001899, 0.00000068, -0.00002589, -0.00006710, 0.00000228, 0.00004777, -0.00005179, 0.00000478, -0.00002038, -0.00002643, -0.00000720, 0.00002253, -0.00003825, 0.00006824, -0.00007937, 0.00003511, 0.00004700, -0.00003539, -0.00001379, 0.00000884, -0.00002724, 0.00000606, -0.00000451, -0.00003428, 0.00002411, 0.00003462, -0.00003756, 0.00002542, -0.00000028, -0.00002966, 0.00000916, -0.00001404, -0.00003048, -0.00001454, -0.00002943, 0.00001575, 0.00001625, -0.00009423, -0.00007558, -0.00003959, 0.00003002, 0.00008437, 0.00007727, 0.00005303, 0.00006831, -0.00000775, 0.00003845, 0.00004629, },
		{ 0.00005082, 0.00005656, 0.00022217, 0.00001825, -0.00033574, -0.00006593, 0.00004622, 0.00007302, -0.00001085, 0.00004709, -0.00004726, -0.00000432, 0.00002791, 0.00002928, -0.00000027, 0.00003950, 0.00005920, -0.00001134, -0.00002413, 0.00000798, -0.00004209, -0.00005417, 0.00006462, -0.00002669, -0.00001716, -0.00000551, -0.00002286, -0.00003820, 0.00006250, -0.00006899, 0.00007277, -0.00004753, -0.00001644, 0.00006497, -0.00001322, -0.00002646, 0.00000538, -0.00001504, -0.00001273, 0.00001155, -0.00002781, -0.00001096, 0.00005097, -0.00001932, -0.00000172, 0.00001528, -0.00002585, -0.00000452, 0.00000194, -0.00003256, -0.00002130, -0.00001935, -0.00000815, 0.00002074, -0.00005710, -0.00008735, -0.00007647, 0.00000557, 0.00008203, 0.00009832, 0.00006369, 0.00006950, 0.00001192, 0.00004171, },
		{ 0.00005049, 0.00005580, 0.00005533, 0.00024150, 0.00004023, -0.00034514, -0.00007059, 0.00002225, 0.00007602, -0.00001200, 0.00004089, -0.00005181, -0.00000329, 0.00002307, 0.00002011, 0.00000528, 0.00001758, 0.00006123, 0.00001429, -0.00004794, 0.00000625, 0.00001388, -0.00009410, 0.00004183, 0.00001558, -0.00002838, -0.00001490, 0.00000906, -0.00008427, 0.00009298, -0.00007054, 0.00004381, 0.00000159, -0.00005393, 0.00005484, 0.00001525, -0.00002503, -0.00000945, -0.00000118, -0.00002288, 0.00000927, -0.00000580, -0.00003716, 0.00004061, 0.00001244, -0.00001999, 0.00001176, -0.00000851, -0.00001834, 0.00000651, -0.00002107, -0.00003385, -0.00001141, -0.00001741, 0.00000582, -0.00002641, -0.00007167, -0.00010861, -0.00003519, 0.00006741, 0.00011841, 0.00008339, 0.00006892, 0.00002923, },
		{ 0.00004109, 0.00006476, 0.00006246, 0.00005387, 0.00026173, 0.00006201, -0.00035602, -0.00007158, -0.00000397, 0.00007699, -0.00001057, 0.00003721, -0.00005854, -0.00000655, 0.00002728, 0.00000662, 0.00000929, 0.00000679, 0.00004672, 0.00004119, -0.00004785, -0.00003066, 0.00006873, -0.00009084, -0.00001218, 0.00004549, -0.00000617, -0.00005321, 0.00005558, -0.00011977, 0.00009171, -0.00003765, -0.00000423, 0.00004374, -0.00006317, 0.00002520, 0.00003340, -0.00000653, -0.00002583, 0.00000334, -0.00001937, -0.00000519, 0.00001453, -0.00004135, 0.00001058, 0.00003828, -0.00001744, -0.00000719, 0.00000867, -0.00002000, -0.00000279, -0.00000403, -0.00004048, -0.00001377, -0.00001390, -0.00001283, -0.00001564, -0.00003968, -0.00011888, -0.00008473, 0.00003161, 0.00013056, 0.00011463, 0.00007057, },
		{ 0.00008229, 0.00004490, 0.00008242, 0.00007384, 0.00005018, 0.00028480, 0.00008511, -0.00037267, -0.00006669, -0.00003089, 0.00007365, -0.00000658, 0.00003528, -0.00006321, -0.00001695, 0.00003647, -0.00000303, 0.00000525, 0.00000655, 0.00002823, 0.00005071, -0.00001826, -0.00007862, 0.00008939, -0.00004385, -0.00006392, 0.00003220, 0.00005370, -0.00010282, 0.00008939, -0.00012132, 0.00005291, 0.00001532, -0.00004720, 0.00006121, -0.00004628, -0.00000604, 0.00003107, 0.00002063, -0.00003111, -0.00000564, -0.00000757, -0.00001900, 0.00002049, -0.00002523, -0.00002115, 0.00004314, 0.00000373, -0.00002730, 0.00001260, -0.00000706, -0.00001758, 0.00000667, -0.00003486, -0.00002393, -0.00000805, -0.00002255, -0.00002339, -0.00001043, -0.00010191, -0.00012504, -0.00002708, 0.00012121, 0.00015593, },
		{ 0.00019590, 0.00011416, 0.00004249, 0.00009733, 0.00009434, 0.00004329, 0.00030927, 0.00011488, -0.00039958, -0.00005720, -0.00005383, 0.00006312, -0.00000016, 0.00003413, -0.00006382, -0.00003191, 0.00004319, -0.00000382, -0.00000475, 0.00000712, 0.00002008, 0.00003660, 0.00002506, -0.00010621, 0.00006482, 0.00001671, -0.00007315, -0.00003400, 0.00012836, -0.00013172, 0.00008659, -0.00008296, -0.00000852, 0.00006303, -0.00006587, 0.00005198, -0.00001829, -0.00002293, 0.00001058, 0.00004167, -0.00001868, -0.00002209, 0.00000132, -0.00002221, 0.00001103, -0.00000210, -0.00003770, 0.00002553, 0.00002924, -0.00003404, 0.00000008, 0.00001188, -0.00002572, 0.00000471, -0.00002057, -0.00003245, -0.00000884, -0.00002138, -0.00003759, 0.00000216, -0.00006799, -0.00013842, -0.00009627, 0.00007620, },
		{ -0.00000782, 0.00021189, 0.00017118, 0.00004445, 0.00010064, 0.00012579, 0.00003692, 0.00032941, 0.00015756, -0.00043699, -0.00004971, -0.00006663, 0.00004454, 0.00000634, 0.00003466, -0.00006169, -0.00004629, 0.00004302, 0.00000121, -0.00001145, -0.00000166, 0.00002672, 0.00000972, 0.00005699, -0.00009729, 0.00001579, 0.00004630, -0.00001957, -0.00012941, 0.00017692, -0.00011317, 0.00004294, -0.00002061, -0.00006561, 0.00008429, -0.00005648, 0.00002855, 0.00000346, -0.00002041, -0.00001538, 0.00004477, 0.00000706, -0.00003365, -0.00000109, -0.00001426, -0.00000466, 0.00001311, -0.00003335, -0.00000261, 0.00004268, -0.00002258, -0.00002048, 0.00002401, -0.00002057, -0.00000631, -0.00000753, -0.00003162, -0.00001687, -0.00001661, -0.00004624, -0.00000259, -0.00003600, -0.00011981, -0.00015188, },
		{ -0.00017045, -0.00011457, 0.00017818, 0.00024324, 0.00006937, 0.00008743, 0.00016237, 0.00004022, 0.00033877, 0.00021472, -0.00047801, -0.00005379, -0.00006635, 0.00002220, 0.00000775, 0.00003876, -0.00005860, -0.00005761, 0.00003862, 0.00000293, -0.00000664, -0.00002079, 0.00003909, -0.00001114, 0.00006264, -0.00006378, -0.00001680, 0.00001452, 0.00007954, -0.00020525, 0.00016380, -0.00004449, -0.00002166, 0.00003730, -0.00009528, 0.00007382, -0.00003099, 0.00000819, 0.00000991, -0.00000553, -0.00003228, 0.00002849, 0.00003294, -0.00003015, -0.00001413, -0.00000324, -0.00001510, 0.00001335, -0.00001536, -0.00002461, 0.00003626, -0.00000050, -0.00003557, 0.00002171, -0.00000551, -0.00001633, -0.00000385, -0.00002168, -0.00002517, -0.00001651, -0.00004596, -0.00001438, -0.00002048, -0.00008148, },
		{ -0.00004588, -0.00014415, -0.00021030, 0.00008182, 0.00030063, 0.00013409, 0.00006568, 0.00018960, 0.00006263, 0.00033707, 0.00027949, -0.00051114, -0.00007556, -0.00005795, 0.00000529, -0.00000073, 0.00004555, -0.00005289, -0.00006875, 0.00003743, -0.00000587, 0.00000841, -0.00004000, 0.00004298, -0.00001561, 0.00005000, -0.00003867, 0.00000106, -0.00007201, 0.00017372, -0.00021384, 0.00008222, 0.00004770, -0.00007440, 0.00006715, -0.00009103, 0.00004340, -0.00000777, 0.00000161, 0.00000400, 0.00000896, -0.00003278, 0.00000236, 0.00004521, -0.00001081, -0.00002848, 0.00000106, -0.00001483, 0.00000248, 0.00000258, -0.00002980, 0.00001493, 0.00001809, -0.00003630, 0.00000760, 0.00000918, -0.00001685, -0.00000974, -0.00001011, -0.00002612, -0.00002289, -0.00004243, -0.00002170, -0.00002254, },
		{ -0.00003105, -0.00003112, -0.00008764, -0.00025891, -0.00006305, 0.00030227, 0.00023796, 0.00005979, 0.00019197, 0.00010447, 0.00033482, 0.00033944, -0.00052829, -0.00011100, -0.00005413, 0.00000252, -0.00001777, 0.00004823, -0.00003900, -0.00008477, 0.00004423, -0.00002306, 0.00002223, -0.00004651, 0.00003330, -0.00001243, 0.00004528, -0.00005262, 0.00007131, -0.00016860, 0.00020591, -0.00013786, -0.00003761, 0.00012085, -0.00008956, 0.00006377, -0.00006443, 0.00001262, 0.00000046, 0.00000758, -0.00000386, 0.00001389, -0.00002023, -0.00001954, 0.00003807, 0.00001491, -0.00003292, -0.00000520, -0.00000736, -0.00000950, 0.00000986, -0.00001903, -0.00000801, 0.00002237, -0.00002405, -0.00000821, 0.00001421, -0.00000706, -0.00001804, -0.00000512, -0.00001822, -0.00003008, -0.00004340, -0.00002103, },
		{ -0.00001842, -0.00003282, -0.00003937, -0.00003086, -0.00024241, -0.00021389, 0.00021666, 0.00035094, 0.00010156, 0.00016725, 0.00015059, 0.00034951, 0.00038644, -0.00053332, -0.00014550, -0.00006800, 0.00001362, -0.00003248, 0.00003606, -0.00001258, -0.00010602, 0.00005513, -0.00003746, 0.00002242, -0.00003835, 0.00002258, -0.00002475, 0.00007285, -0.00010975, 0.00015799, -0.00021589, 0.00014605, -0.00000411, -0.00014250, 0.00014094, -0.00006447, 0.00004168, -0.00003537, -0.00000392, -0.00000670, 0.00001612, -0.00000454, 0.00000899, -0.00000498, -0.00002740, 0.00001683, 0.00003334, -0.00002232, -0.00001707, -0.00000118, -0.00001446, 0.00000500, -0.00000222, -0.00002023, 0.00001148, -0.00000862, -0.00001584, 0.00000775, 0.00000620, -0.00002030, -0.00000914, -0.00000722, -0.00003058, -0.00005088, },
		{ -0.00005902, -0.00002281, -0.00002356, -0.00005668, -0.00000112, -0.00017346, -0.00031647, 0.00004520, 0.00041738, 0.00020894, 0.00013807, 0.00017516, 0.00039277, 0.00042730, -0.00054317, -0.00016268, -0.00010178, 0.00002470, -0.00002836, 0.00000247, 0.00002208, -0.00012401, 0.00005940, -0.00003845, 0.00000860, -0.00003127, 0.00003486, -0.00007157, 0.00013344, -0.00017978, 0.00020620, -0.00017488, 0.00001135, 0.00012817, -0.00018521, 0.00010289, -0.00002064, 0.00002344, -0.00001820, -0.00000415, -0.00001965, 0.00001714, 0.00000330, 0.00000151, 0.00000346, -0.00002103, -0.00000603, 0.00003530, -0.00000145, -0.00002470, -0.00000257, -0.00001148, -0.00000555, 0.00000951, -0.00001776, -0.00000609, -0.00000067, -0.00001246, -0.00000421, 0.00001424, -0.00001337, -0.00001735, -0.00000108, -0.00002206, },
		{ -0.00000968, -0.00005925, -0.00003699, -0.00001103, -0.00006438, -0.00000626, -0.00009039, -0.00033271, -0.00016764, 0.00037968, 0.00036338, 0.00014868, 0.00015962, 0.00045603, 0.00048549, -0.00057748, -0.00015860, -0.00013998, 0.00001332, 0.00000515, -0.00004577, 0.00004977, -0.00012600, 0.00004983, -0.00002835, -0.00000177, -0.00005122, 0.00008715, -0.00014806, 0.00019702, -0.00021383, 0.00017532, -0.00005196, -0.00014028, 0.00019876, -0.00015129, 0.00003364, 0.00001075, 0.00002342, -0.00001359, 0.00000336, -0.00002735, 0.00000777, 0.00001324, -0.00000052, 0.00000270, -0.00000842, -0.00001907, 0.00002152, 0.00001821, -0.00002075, -0.00001068, -0.00000603, -0.00001364, 0.00001098, -0.00000589, -0.00001924, -0.00000447, -0.00000328, -0.00001322, 0.00001299, -0.00000119, -0.00002212, -0.00000336, },
		{ -0.00001056, -0.00000183, -0.00004805, -0.00005405, -0.00000810, -0.00005382, -0.00003057, -0.00003566, -0.00026360, -0.00034824, 0.00021248, 0.00050276, 0.00024186, 0.00011452, 0.00050750, 0.00058828, -0.00064113, -0.00015074, -0.00015518, -0.00003645, 0.00006054, -0.00008695, 0.00005231, -0.00010728, 0.00003421, -0.00002660, 0.00001784, -0.00011494, 0.00017057, -0.00022015, 0.00021759, -0.00017487, 0.00006572, 0.00010203, -0.00024090, 0.00017950, -0.00006702, -0.00002428, 0.00000937, 0.00003840, -0.00001120, 0.00000844, -0.00002531, -0.00000658, 0.00001692, 0.00000540, -0.00000212, 0.00000083, -0.00001867, 0.00000144, 0.00002618, -0.00000586, -0.00001853, -0.00000442, -0.00001551, 0.00000466, 0.00000612, -0.00002167, -0.00001598, 0.00000334, -0.00001480, 0.00000517, 0.00000905, -0.00001933, },
		{ -0.00001124, -0.00001546, -0.00000316, -0.00003065, -0.00006233, -0.00002118, -0.00003315, -0.00004691, -0.00003076, -0.00015147, -0.00042830, -0.00005225, 0.00054154, 0.00042449, 0.00008849, 0.00050866, 0.00074424, -0.00071256, -0.00017206, -0.00012613, -0.00011824, 0.00011021, -0.00009586, 0.00002429, -0.00008256, 0.00003716, -0.00005939, 0.00008299, -0.00021137, 0.00024647, -0.00024025, 0.00016213, -0.00006364, -0.00007848, 0.00021900, -0.00024937, 0.00008906, 0.00001648, -0.00003874, -0.00002121, 0.00005115, -0.00000027, 0.00000780, -0.00001740, -0.00001732, 0.00001071, 0.00001452, -0.00000336, 0.00000231, -0.00000985, -0.00001360, 0.00001956, 0.00001181, -0.00001877, -0.00000836, -0.00001340, -0.00000318, 0.00001173, -0.00001481, -0.00002692, 0.00000240, -0.00001101, -0.00000296, 0.00001282, },
		{ 0.00001125, -0.00000392, -0.00001346, -0.00001064, -0.00001735, -0.00005498, -0.00004360, -0.00002097, -0.00003669, -0.00006385, -0.00005772, -0.00038213, -0.00032839, 0.00041398, 0.00064299, 0.00015059, 0.00044559, 0.00092711, -0.00075046, -0.00024925, -0.00005779, -0.00019988, 0.00012087, -0.00006412, -0.00001400, -0.00008520, 0.00008671, -0.00013873, 0.00017991, -0.00029858, 0.00026362, -0.00017636, 0.00003528, 0.00007752, -0.00019310, 0.00025109, -0.00017263, -0.00001918, 0.00005335, -0.00000693, -0.00005665, 0.00004425, 0.00002017, 0.00000683, -0.00001089, -0.00001959, -0.00000176, 0.00001897, 0.00000272, -0.00000097, -0.00000137, -0.00001761, 0.00000431, 0.00002239, -0.00000955, -0.00001376, -0.00001207, -0.00000809, 0.00001044, -0.00000507, -0.00003111, -0.00000464, -0.00000755, -0.00000726, },
		{ -0.00000870, 0.00000900, -0.00000160, -0.00000637, -0.00001625, -0.00001530, -0.00003630, -0.00005949, -0.00003139, -0.00000378, -0.00009899, -0.00002786, -0.00024374, -0.00051160, 0.00011897, 0.00079014, 0.00034756, 0.00035577, 0.00108280, -0.00072075, -0.00037778, 0.00001316, -0.00024016, 0.00007906, -0.00001607, -0.00001939, -0.00014940, 0.00019129, -0.00024487, 0.00026379, -0.00032243, 0.00018854, -0.00003635, -0.00011756, 0.00018671, -0.00022989, 0.00019573, -0.00005908, -0.00008855, 0.00002886, 0.00004443, -0.00006885, 0.00001408, 0.00003934, 0.00001365, -0.00000930, -0.00001541, -0.00001273, 0.00001394, 0.00001301, -0.00000207, 0.00000113, -0.00001281, -0.00000966, 0.00002110, 0.00000396, -0.00001463, -0.00001383, -0.00001014, 0.00000597, 0.00000179, -0.00002821, -0.00001212, -0.00000833, },
		{ -0.00001225, -0.00001146, 0.00000919, -0.00000316, -0.00000047, -0.00001393, -0.00002249, -0.00001926, -0.00005569, -0.00006176, 0.00002813, -0.00010143, -0.00006535, -0.00009073, -0.00053209, -0.00026201, 0.00074976, 0.00065729, 0.00032656, 0.00116450, -0.00062928, -0.00051384, 0.00003745, -0.00022003, 0.00001153, -0.00000533, 0.00005146, -0.00028317, 0.00032231, -0.00032680, 0.00027895, -0.00024801, 0.00003028, 0.00012975, -0.00023248, 0.00021726, -0.00018282, 0.00009395, 0.00003092, -0.00009043, -0.00003445, 0.00007728, -0.00004524, -0.00002519, 0.00004256, 0.00003061, -0.00000909, -0.00001101, -0.00001642, 0.00000132, 0.00002010, 0.00000373, -0.00000118, -0.00000649, -0.00001547, 0.00001111, 0.00001353, -0.00000846, -0.00001639, -0.00001215, 0.00000178, 0.00000426, -0.00002230, -0.00001539, },
		{ -0.00001433, -0.00001473, -0.00001809, 0.00001019, -0.00000376, -0.00000055, -0.00000497, -0.00002925, -0.00001572, -0.00003256, -0.00009292, 0.00003311, -0.00006211, -0.00013133, -0.00000176, -0.00039466, -0.00059128, 0.00046182, 0.00096988, 0.00045385, 0.00117981, -0.00053505, -0.00059676, -0.00000898, -0.00016853, -0.00001774, -0.00008693, 0.00020915, -0.00044954, 0.00041770, -0.00032553, 0.00019157, -0.00008504, -0.00015755, 0.00025322, -0.00026220, 0.00016395, -0.00008888, 0.00000401, 0.00005904, -0.00003721, -0.00009321, 0.00006681, 0.00000241, -0.00005043, 0.00002299, 0.00005014, -0.00000289, -0.00001079, -0.00001307, -0.00001190, 0.00001801, 0.00001488, -0.00000222, -0.00000465, -0.00001341, 0.00000049, 0.00001409, 0.00000135, -0.00001526, -0.00001564, -0.00000185, 0.00000435, -0.00001749, },
		{ -0.00001454, -0.00001253, -0.00001238, -0.00002624, 0.00000718, -0.00000019, -0.00000529, 0.00000296, -0.00002618, -0.00002694, -0.00000488, -0.00010250, 0.00000312, -0.00000416, -0.00017148, -0.00001189, -0.00018028, -0.00073845, -0.00002774, 0.00111924, 0.00077453, 0.00121841, -0.00052430, -0.00059654, -0.00009171, -0.00016125, 0.00005969, -0.00027712, 0.00040931, -0.00057266, 0.00040989, -0.00020909, 0.00001513, 0.00011046, -0.00030142, 0.00028323, -0.00020260, 0.00006442, -0.00000393, -0.00003255, 0.00002709, 0.00002971, -0.00011079, 0.00001656, 0.00004563, -0.00004534, -0.00001148, 0.00005887, 0.00001348, -0.00001309, -0.00000793, -0.00001899, 0.00000678, 0.00002455, 0.00000342, -0.00000740, -0.00000942, -0.00000405, 0.00000736, 0.00000801, -0.00000839, -0.00001848, -0.00000680, 0.00000402, },
		{ 0.00000244, -0.00001134, -0.00001308, -0.00000657, -0.00003010, -0.00000258, 0.00000501, -0.00000813, 0.00000256, -0.00001239, -0.00004201, 0.00000852, -0.00008172, -0.00004432, 0.00003342, -0.00015284, -0.00009354, -0.00000418, -0.00064913, -0.00056692, 0.00096067, 0.00120957, 0.00143065, -0.00064842, -0.00055902, -0.00011690, -0.00028245, 0.00026794, -0.00052566, 0.00056031, -0.00057193, 0.00026544, 0.00000178, -0.00018938, 0.00026199, -0.00034605, 0.00021469, -0.00009417, -0.00002458, 0.00003054, -0.00001164, -0.00003002, 0.00006753, -0.00007864, -0.00004502, 0.00005776, -0.00001210, -0.00004179, 0.00004711, 0.00003571, -0.00001138, -0.00000625, -0.00001826, -0.00000761, 0.00002576, 0.00001558, -0.00000930, -0.00000946, -0.00000198, 0.00000020, 0.00000689, 0.00000105, -0.00001658, -0.00001363, },
		{ -0.00001921, -0.00000302, -0.00000651, -0.00001517, -0.00000248, -0.00002564, -0.00001561, 0.00000472, -0.00000356, -0.00000606, 0.00000310, -0.00004624, -0.00000093, -0.00004312, -0.00007725, 0.00002421, -0.00008647, -0.00017586, 0.00004756, -0.00037898, -0.00095753, 0.00045921, 0.00155941, 0.00195224, -0.00085630, -0.00061354, 0.00002431, -0.00056553, 0.00055139, -0.00072345, 0.00056393, -0.00040830, 0.00001233, 0.00023577, -0.00034154, 0.00031216, -0.00028373, 0.00009016, 0.00000238, -0.00006067, 0.00000908, 0.00003699, -0.00006968, 0.00005726, -0.00001869, -0.00008348, 0.00003121, 0.00002917, -0.00004956, 0.00001696, 0.00005229, 0.00000000, -0.00000841, -0.00001371, -0.00001771, 0.00001661, 0.00002813, -0.00000434, -0.00001433, 0.00000135, -0.00000105, -0.00000043, 0.00000650, -0.00000826, },
		{ 0.00000271, -0.00001831, -0.00001307, -0.00000234, -0.00001475, -0.00000428, -0.00001511, -0.00002340, -0.00000648, 0.00000660, -0.00001443, 0.00000840, -0.00003308, -0.00002438, -0.00001245, -0.00007211, -0.00002506, -0.00002008, -0.00019140, -0.00003383, -0.00006921, -0.00105081, -0.00026173, 0.00157126, 0.00280477, -0.00096671, -0.00092212, 0.00038113, -0.00094688, 0.00078069, -0.00075014, 0.00037075, -0.00011574, -0.00026447, 0.00040338, -0.00038332, 0.00025288, -0.00015703, -0.00002514, 0.00004105, -0.00003829, -0.00003987, 0.00007301, -0.00006814, 0.00001165, 0.00003302, -0.00007987, -0.00001806, 0.00005292, -0.00002941, -0.00001729, 0.00005186, 0.00001997, -0.00000920, -0.00001108, -0.00001989, 0.00000171, 0.00003321, 0.00000869, -0.00001885, -0.00000028, 0.00000424, -0.00000708, 0.00000334, },
		{ -0.00000665, 0.00000862, -0.00000836, -0.00002359, -0.00000376, -0.00000893, -0.00001059, -0.00000642, -0.00001865, -0.00002606, 0.00001273, -0.00001241, -0.00000088, -0.00001032, -0.00004164, -0.00001008, -0.00003177, -0.00007713, -0.00000489, -0.00012081, -0.00017689, 0.00012736, -0.00082886, -0.00094849, 0.00105858, 0.00382674, -0.00071413, -0.00158312, 0.00087925, -0.00126809, 0.00080757, -0.00054062, 0.00001838, 0.00020461, -0.00046123, 0.00044288, -0.00030997, 0.00012790, -0.00003332, -0.00007995, 0.00001455, 0.00001281, -0.00007641, 0.00007196, -0.00003175, -0.00003719, 0.00004958, -0.00004118, -0.00006110, 0.00004433, 0.00000696, -0.00003745, 0.00003165, 0.00003959, -0.00000215, -0.00001269, -0.00001644, -0.00001116, 0.00002697, 0.00002441, -0.00001606, -0.00000831, 0.00001037, -0.00000623, },
		{ 0.00000309, -0.00001545, 0.00000342, 0.00000666, -0.00002703, -0.00001369, -0.00000018, -0.00001472, -0.00000725, -0.00000211, -0.00004289, 0.00000425, 0.00000231, -0.00001660, 0.00000511, -0.00003668, -0.00003519, 0.00001477, -0.00008924, -0.00005871, -0.00000626, -0.00027574, 0.00013019, -0.00042426, -0.00132010, 0.00001657, 0.00468573, 0.00013549, -0.00253607, 0.00131309, -0.00133907, 0.00054541, -0.00013317, -0.00037405, 0.00042949, -0.00050809, 0.00035310, -0.00017120, 0.00000625, 0.00003299, -0.00006291, -0.00004629, 0.00005027, -0.00007572, 0.00003794, 0.00001198, -0.00006032, 0.00002832, 0.00000611, -0.00007426, 0.00000845, 0.00003783, -0.00003291, 0.00000067, 0.00004721, 0.00001450, -0.00001518, -0.00001303, -0.00001655, 0.00001234, 0.00003419, -0.00000312, -0.00001752, 0.00000993, },
		{ 0.00000010, 0.00001438, -0.00001432, -0.00001170, 0.00001704, -0.00001800, -0.00002854, 0.00000397, -0.00001004, -0.00001820, 0.00001621, -0.00004423, -0.00002083, 0.00001995, -0.00002328, 0.00000057, -0.00001098, -0.00006515, 0.00003208, -0.00004384, -0.00015090, 0.00007626, -0.00025798, -0.00002342, -0.00006044, -0.00120135, -0.00133091, 0.00498599, 0.00164670, -0.00352642, 0.00141558, -0.00103770, 0.00003139, 0.00033335, -0.00065793, 0.00047526, -0.00040595, 0.00019301, -0.00004177, -0.00005703, 0.00002777, -0.00000329, -0.00009555, 0.00004689, -0.00004184, -0.00000351, 0.00003538, -0.00004938, -0.00001084, 0.00003383, -0.00005313, -0.00003359, 0.00004509, -0.00000768, -0.00002494, 0.00003636, 0.00003490, -0.00001213, -0.00001362, -0.00001480, -0.00000301, 0.00003207, 0.00001563, -0.00001949, },
		{ -0.00000907, -0.00001412, 0.00001799, -0.00000070, -0.00002735, 0.00001356, 0.00000105, -0.00003854, -0.00000349, 0.00000338, -0.00003096, 0.00002332, -0.00002559, -0.00005156, 0.00002413, -0.00001061, -0.00002081, 0.00001588, -0.00007152, 0.00000415, 0.00003685, -0.00022077, 0.00006876, -0.00013284, -0.00019951, 0.00006030, -0.00061786, -0.00259021, 0.00443451, 0.00362565, -0.00416700, 0.00098355, -0.00039438, -0.00057023, 0.00067875, -0.00072876, 0.00034185, -0.00022514, 0.00004760, 0.00002081, -0.00004889, -0.00002145, 0.00005185, -0.00010150, 0.00000578, -0.00000110, -0.00002556, 0.00002848, -0.00001845, -0.00004073, 0.00002917, -0.00001355, -0.00005821, 0.00002533, 0.00002204, -0.00003182, 0.00001097, 0.00004836, 0.00000036, -0.00001688, -0.00001092, -0.00001205, 0.00001892, 0.00003088, },
		{ 0.00003439, 0.00001114, -0.00002263, 0.00000840, 0.00001864, -0.00003151, -0.00000515, 0.00001905, -0.00003390, -0.00002216, 0.00001680, -0.00003385, 0.00001312, 0.00000401, -0.00006944, 0.00000439, 0.00001619, -0.00004031, 0.00002019, -0.00004189, -0.00005235, 0.00010393, -0.00021865, -0.00002625, 0.00001106, -0.00024564, -0.00013703, 0.00020011, -0.00334669, 0.00298970, 0.00563857, -0.00408468, -0.00000040, 0.00039081, -0.00103490, 0.00076977, -0.00058044, 0.00011220, -0.00006564, -0.00001957, 0.00000561, 0.00000085, -0.00006840, 0.00006798, -0.00006370, -0.00004534, 0.00001813, -0.00001825, 0.00000277, 0.00000851, -0.00004467, 0.00000027, 0.00001986, -0.00005380, -0.00000868, 0.00003811, -0.00001771, -0.00001626, 0.00004623, 0.00001950, -0.00001718, -0.00001003, -0.00001292, 0.00000180, },
		{ -0.00001093, 0.00002442, 0.00003160, -0.00001709, -0.00001077, 0.00003126, -0.00001801, -0.00002958, 0.00002245, -0.00001295, -0.00004178, 0.00001819, -0.00002126, -0.00000774, 0.00002656, -0.00006122, -0.00003343, 0.00003720, -0.00003676, -0.00000701, 0.00000881, -0.00009804, 0.00011531, -0.00014253, -0.00014148, 0.00005910, -0.00008214, -0.00055783, 0.00091381, -0.00333834, 0.00091583, 0.00714663, -0.00309088, -0.00131738, 0.00101947, -0.00118540, 0.00058931, -0.00030818, -0.00009362, -0.00000187, 0.00000124, -0.00005211, 0.00004408, -0.00007906, 0.00004144, -0.00000862, -0.00007486, 0.00000465, 0.00000691, -0.00001981, 0.00001471, -0.00002497, -0.00003179, 0.00002944, -0.00002661, -0.00003673, 0.00003164, 0.00000727, -0.00003145, 0.00002810, 0.00003643, -0.00000927, -0.00001286, -0.00000947, },
		{ -0.00000743, -0.00001510, 0.00000678, 0.00004169, 0.00000275, -0.00002809, 0.00002737, 0.00000792, -0.00004403, 0.00000486, 0.00001401, -0.00004779, 0.00000185, 0.00000019, -0.00002329, 0.00002745, -0.00002993, -0.00006785, 0.00003092, -0.00000389, -0.00005043, 0.00004687, -0.00009869, 0.00006719, -0.00004472, -0.00018379, -0.00005645, 0.00024631, -0.00098413, 0.00122733, -0.00257096, -0.00129282, 0.00770637, -0.00129282, -0.00257096, 0.00122733, -0.00098413, 0.00024631, -0.00005645, -0.00018379, -0.00004472, 0.00006719, -0.00009869, 0.00004687, -0.00005043, -0.00000389, 0.00003092, -0.00006785, -0.00002993, 0.00002745, -0.00002329, 0.00000019, 0.00000185, -0.00004779, 0.00001401, 0.00000486, -0.00004403, 0.00000792, 0.00002737, -0.00002809, 0.00000275, 0.00004169, 0.00000678, -0.00001510, },
		{ -0.00001093, -0.00000947, -0.00001286, -0.00000927, 0.00003643, 0.00002810, -0.00003145, 0.00000727, 0.00003164, -0.00003673, -0.00002661, 0.00002944, -0.00003179, -0.00002497, 0.00001471, -0.00001981, 0.00000691, 0.00000465, -0.00007486, -0.00000862, 0.00004144, -0.00007906, 0.00004408, -0.00005211, 0.00000124, -0.00000187, -0.00009362, -0.00030818, 0.00058931, -0.00118540, 0.00101947, -0.00131738, -0.00309088, 0.00714663, 0.00091583, -0.00333834, 0.00091381, -0.00055783, -0.00008214, 0.00005910, -0.00014148, -0.00014253, 0.00011531, -0.00009804, 0.00000881, -0.00000701, -0.00003676, 0.00003720, -0.00003343, -0.00006122, 0.00002656, -0.00000774, -0.00002126, 0.00001819, -0.00004178, -0.00001295, 0.00002245, -0.00002958, -0.00001801, 0.00003126, -0.00001077, -0.00001709, 0.00003160, 0.00002442, },
		{ 0.00003439, 0.00000180, -0.00001292, -0.00001003, -0.00001718, 0.00001950, 0.00004623, -0.00001626, -0.00001771, 0.00003811, -0.00000868, -0.00005380, 0.00001986, 0.00000027, -0.00004467, 0.00000851, 0.00000277, -0.00001825, 0.00001813, -0.00004534, -0.00006370, 0.00006798, -0.00006840, 0.00000085, 0.00000561, -0.00001957, -0.00006564, 0.00011220, -0.00058044, 0.00076977, -0.00103490, 0.00039081, -0.00000040, -0.00408468, 0.00563857, 0.00298970, -0.00334669, 0.00020011, -0.00013703, -0.00024564, 0.00001106, -0.00002625, -0.00021865, 0.00010393, -0.00005235, -0.00004189, 0.00002019, -0.00004031, 0.00001619, 0.00000439, -0.00006944, 0.00000401, 0.00001312, -0.00003385, 0.00001680, -0.00002216, -0.00003390, 0.00001905, -0.00000515, -0.00003151, 0.00001864, 0.00000840, -0.00002263, 0.00001114, },
		{ -0.00000907, 0.00003088, 0.00001892, -0.00001205, -0.00001092, -0.00001688, 0.00000036, 0.00004836, 0.00001097, -0.00003182, 0.00002204, 0.00002533, -0.00005821, -0.00001355, 0.00002917, -0.00004073, -0.00001845, 0.00002848, -0.00002556, -0.00000110, 0.00000578, -0.00010150, 0.00005185, -0.00002145, -0.00004889, 0.00002081, 0.00004760, -0.00022514, 0.00034185, -0.00072876, 0.00067875, -0.00057023, -0.00039438, 0.00098355, -0.00416700, 0.00362565, 0.00443451, -0.00259021, -0.00061786, 0.00006030, -0.00019951, -0.00013284, 0.00006876, -0.00022077, 0.00003685, 0.00000415, -0.00007152, 0.00001588, -0.00002081, -0.00001061, 0.00002413, -0.00005156, -0.00002559, 0.00002332, -0.00003096, 0.00000338, -0.00000349, -0.00003854, 0.00000105, 0.00001356, -0.00002735, -0.00000070, 0.00001799, -0.00001412, },
		{ 0.00000010, -0.00001949, 0.00001563, 0.00003207, -0.00000301, -0.00001480, -0.00001362, -0.00001213, 0.00003490, 0.00003636, -0.00002494, -0.00000768, 0.00004509, -0.00003359, -0.00005313, 0.00003383, -0.00001084, -0.00004938, 0.00003538, -0.00000351, -0.00004184, 0.00004689, -0.00009555, -0.00000329, 0.00002777, -0.00005703, -0.00004177, 0.00019301, -0.00040595, 0.00047526, -0.00065793, 0.00033335, 0.00003139, -0.00103770, 0.00141558, -0.00352642, 0.00164670, 0.00498599, -0.00133091, -0.00120135, -0.00006044, -0.00002342, -0.00025798, 0.00007626, -0.00015090, -0.00004384, 0.00003208, -0.00006515, -0.00001098, 0.00000057, -0.00002328, 0.00001995, -0.00002083, -0.00004423, 0.00001621, -0.00001820, -0.00001004, 0.00000397, -0.00002854, -0.00001800, 0.00001704, -0.00001170, -0.00001432, 0.00001438, },
		{ 0.00000309, 0.00000993, -0.00001752, -0.00000312, 0.00003419, 0.00001234, -0.00001655, -0.00001303, -0.00001518, 0.00001450, 0.00004721, 0.00000067, -0.00003291, 0.00003783, 0.00000845, -0.00007426, 0.00000611, 0.00002832, -0.00006032, 0.00001198, 0.00003794, -0.00007572, 0.00005027, -0.00004629, -0.00006291, 0.00003299, 0.00000625, -0.00017120, 0.00035310, -0.00050809, 0.00042949, -0.00037405, -0.00013317, 0.00054541, -0.00133907, 0.00131309, -0.00253607, 0.00013549, 0.00468573, 0.00001657, -0.00132010, -0.00042426, 0.00013019, -0.00027574, -0.00000626, -0.00005871, -0.00008924, 0.00001477, -0.00003519, -0.00003668, 0.00000511, -0.00001660, 0.00000231, 0.00000425, -0.00004289, -0.00000211, -0.00000725, -0.00001472, -0.00000018, -0.00001369, -0.00002703, 0.00000666, 0.00000342, -0.00001545, },
		{ -0.00000665, -0.00000623, 0.00001037, -0.00000831, -0.00001606, 0.00002441, 0.00002697, -0.00001116, -0.00001644, -0.00001269, -0.00000215, 0.00003959, 0.00003165, -0.00003745, 0.00000696, 0.00004433, -0.00006110, -0.00004118, 0.00004958, -0.00003719, -0.00003175, 0.00007196, -0.00007641, 0.00001281, 0.00001455, -0.00007995, -0.00003332, 0.00012790, -0.00030997, 0.00044288, -0.00046123, 0.00020461, 0.00001838, -0.00054062, 0.00080757, -0.00126809, 0.00087925, -0.00158312, -0.00071413, 0.00382674, 0.00105858, -0.00094849, -0.00082886, 0.00012736, -0.00017689, -0.00012081, -0.00000489, -0.00007713, -0.00003177, -0.00001008, -0.00004164, -0.00001032, -0.00000088, -0.00001241, 0.00001273, -0.00002606, -0.00001865, -0.00000642, -0.00001059, -0.00000893, -0.00000376, -0.00002359, -0.00000836, 0.00000862, },
		{ 0.00000271, 0.00000334, -0.00000708, 0.00000424, -0.00000028, -0.00001885, 0.00000869, 0.00003321, 0.00000171, -0.00001989, -0.00001108, -0.00000920, 0.00001997, 0.00005186, -0.00001729, -0.00002941, 0.00005292, -0.00001806, -0.00007987, 0.00003302, 0.00001165, -0.00006814, 0.00007301, -0.00003987, -0.00003829, 0.00004105, -0.00002514, -0.00015703, 0.00025288, -0.00038332, 0.00040338, -0.00026447, -0.00011574, 0.00037075, -0.00075014, 0.00078069, -0.00094688, 0.00038113, -0.00092212, -0.00096671, 0.00280477, 0.00157126, -0.00026173, -0.00105081, -0.00006921, -0.00003383, -0.00019140, -0.00002008, -0.00002506, -0.00007211, -0.00001245, -0.00002438, -0.00003308, 0.00000840, -0.00001443, 0.00000660, -0.00000648, -0.00002340, -0.00001511, -0.00000428, -0.00001475, -0.00000234, -0.00001307, -0.00001831, },
		{ -0.00001921, -0.00000826, 0.00000650, -0.00000043, -0.00000105, 0.00000135, -0.00001433, -0.00000434, 0.00002813, 0.00001661, -0.00001771, -0.00001371, -0.00000841, 0.00000000, 0.00005229, 0.00001696, -0.00004956, 0.00002917, 0.00003121, -0.00008348, -0.00001869, 0.00005726, -0.00006968, 0.00003699, 0.00000908, -0.00006067, 0.00000238, 0.00009016, -0.00028373, 0.00031216, -0.00034154, 0.00023577, 0.00001233, -0.00040830, 0.00056393, -0.00072345, 0.00055139, -0.00056553, 0.00002431, -0.00061354, -0.00085630, 0.00195224, 0.00155941, 0.00045921, -0.00095753, -0.00037898, 0.00004756, -0.00017586, -0.00008647, 0.00002421, -0.00007725, -0.00004312, -0.00000093, -0.00004624, 0.00000310, -0.00000606, -0.00000356, 0.00000472, -0.00001561, -0.00002564, -0.00000248, -0.00001517, -0.00000651, -0.00000302, },
		{ 0.00000244, -0.00001363, -0.00001658, 0.00000105, 0.00000689, 0.00000020, -0.00000198, -0.00000946, -0.00000930, 0.00001558, 0.00002576, -0.00000761, -0.00001826, -0.00000625, -0.00001138, 0.00003571, 0.00004711, -0.00004179, -0.00001210, 0.00005776, -0.00004502, -0.00007864, 0.00006753, -0.00003002, -0.00001164, 0.00003054, -0.00002458, -0.00009417, 0.00021469, -0.00034605, 0.00026199, -0.00018938, 0.00000178, 0.00026544, -0.00057193, 0.00056031, -0.00052566, 0.00026794, -0.00028245, -0.00011690, -0.00055902, -0.00064842, 0.00143065, 0.00120957, 0.00096067, -0.00056692, -0.00064913, -0.00000418, -0.00009354, -0.00015284, 0.00003342, -0.00004432, -0.00008172, 0.00000852, -0.00004201, -0.00001239, 0.00000256, -0.00000813, 0.00000501, -0.00000258, -0.00003010, -0.00000657, -0.00001308, -0.00001134, },
		{ -0.00001454, 0.00000402, -0.00000680, -0.00001848, -0.00000839, 0.00000801, 0.00000736, -0.00000405, -0.00000942, -0.00000740, 0.00000342, 0.00002455, 0.00000678, -0.00001899, -0.00000793, -0.00001309, 0.00001348, 0.00005887, -0.00001148, -0.00004534, 0.00004563, 0.00001656, -0.00011079, 0.00002971, 0.00002709, -0.00003255, -0.00000393, 0.00006442, -0.00020260, 0.00028323, -0.00030142, 0.00011046, 0.00001513, -0.00020909, 0.00040989, -0.00057266, 0.00040931, -0.00027712, 0.00005969, -0.00016125, -0.00009171, -0.00059654, -0.00052430, 0.00121841, 0.00077453, 0.00111924, -0.00002774, -0.00073845, -0.00018028, -0.00001189, -0.00017148, -0.00000416, 0.00000312, -0.00010250, -0.00000488, -0.00002694, -0.00002618, 0.00000296, -0.00000529, -0.00000019, 0.00000718, -0.00002624, -0.00001238, -0.00001253, },
		{ -0.00001433, -0.00001749, 0.00000435, -0.00000185, -0.00001564, -0.00001526, 0.00000135, 0.00001409, 0.00000049, -0.00001341, -0.00000465, -0.00000222, 0.00001488, 0.00001801, -0.00001190, -0.00001307, -0.00001079, -0.00000289, 0.00005014, 0.00002299, -0.00005043, 0.00000241, 0.00006681, -0.00009321, -0.00003721, 0.00005904, 0.00000401, -0.00008888, 0.00016395, -0.00026220, 0.00025322, -0.00015755, -0.00008504, 0.00019157, -0.00032553, 0.00041770, -0.00044954, 0.00020915, -0.00008693, -0.00001774, -0.00016853, -0.00000898, -0.00059676, -0.00053505, 0.00117981, 0.00045385, 0.00096988, 0.00046182, -0.00059128, -0.00039466, -0.00000176, -0.00013133, -0.00006211, 0.00003311, -0.00009292, -0.00003256, -0.00001572, -0.00002925, -0.00000497, -0.00000055, -0.00000376, 0.00001019, -0.00001809, -0.00001473, },
		{ -0.00001225, -0.00001539, -0.00002230, 0.00000426, 0.00000178, -0.00001215, -0.00001639, -0.00000846, 0.00001353, 0.00001111, -0.00001547, -0.00000649, -0.00000118, 0.00000373, 0.00002010, 0.00000132, -0.00001642, -0.00001101, -0.00000909, 0.00003061, 0.00004256, -0.00002519, -0.00004524, 0.00007728, -0.00003445, -0.00009043, 0.00003092, 0.00009395, -0.00018282, 0.00021726, -0.00023248, 0.00012975, 0.00003028, -0.00024801, 0.00027895, -0.00032680, 0.00032231, -0.00028317, 0.00005146, -0.00000533, 0.00001153, -0.00022003, 0.00003745, -0.00051384, -0.00062928, 0.00116450, 0.00032656, 0.00065729, 0.00074976, -0.00026201, -0.00053209, -0.00009073, -0.00006535, -0.00010143, 0.00002813, -0.00006176, -0.00005569, -0.00001926, -0.00002249, -0.00001393, -0.00000047, -0.00000316, 0.00000919, -0.00001146, },
		{ -0.00000870, -0.00000833, -0.00001212, -0.00002821, 0.00000179, 0.00000597, -0.00001014, -0.00001383, -0.00001463, 0.00000396, 0.00002110, -0.00000966, -0.00001281, 0.00000113, -0.00000207, 0.00001301, 0.00001394, -0.00001273, -0.00001541, -0.00000930, 0.00001365, 0.00003934, 0.00001408, -0.00006885, 0.00004443, 0.00002886, -0.00008855, -0.00005908, 0.00019573, -0.00022989, 0.00018671, -0.00011756, -0.00003635, 0.00018854, -0.00032243, 0.00026379, -0.00024487, 0.00019129, -0.00014940, -0.00001939, -0.00001607, 0.00007906, -0.00024016, 0.00001316, -0.00037778, -0.00072075, 0.00108280, 0.00035577, 0.00034756, 0.00079014, 0.00011897, -0.00051160, -0.00024374, -0.00002786, -0.00009899, -0.00000378, -0.00003139, -0.00005949, -0.00003630, -0.00001530, -0.00001625, -0.00000637, -0.00000160, 0.00000900, },
		{ 0.00001125, -0.00000726, -0.00000755, -0.00000464, -0.00003111, -0.00000507, 0.00001044, -0.00000809, -0.00001207, -0.00001376, -0.00000955, 0.00002239, 0.00000431, -0.00001761, -0.00000137, -0.00000097, 0.00000272, 0.00001897, -0.00000176, -0.00001959, -0.00001089, 0.00000683, 0.00002017, 0.00004425, -0.00005665, -0.00000693, 0.00005335, -0.00001918, -0.00017263, 0.00025109, -0.00019310, 0.00007752, 0.00003528, -0.00017636, 0.00026362, -0.00029858, 0.00017991, -0.00013873, 0.00008671, -0.00008520, -0.00001400, -0.00006412, 0.00012087, -0.00019988, -0.00005779, -0.00024925, -0.00075046, 0.00092711, 0.00044559, 0.00015059, 0.00064299, 0.00041398, -0.00032839, -0.00038213, -0.00005772, -0.00006385, -0.00003669, -0.00002097, -0.00004360, -0.00005498, -0.00001735, -0.00001064, -0.00001346, -0.00000392, },
		{ -0.00001124, 0.00001282, -0.00000296, -0.00001101, 0.00000240, -0.00002692, -0.00001481, 0.00001173, -0.00000318, -0.00001340, -0.00000836, -0.00001877, 0.00001181, 0.00001956, -0.00001360, -0.00000985, 0.00000231, -0.00000336, 0.00001452, 0.00001071, -0.00001732, -0.00001740, 0.00000780, -0.00000027, 0.00005115, -0.00002121, -0.00003874, 0.00001648, 0.00008906, -0.00024937, 0.00021900, -0.00007848, -0.00006364, 0.00016213, -0.00024025, 0.00024647, -0.00021137, 0.00008299, -0.00005939, 0.00003716, -0.00008256, 0.00002429, -0.00009586, 0.00011021, -0.00011824, -0.00012613, -0.00017206, -0.00071256, 0.00074424, 0.00050866, 0.00008849, 0.00042449, 0.00054154, -0.00005225, -0.00042830, -0.00015147, -0.00003076, -0.00004691, -0.00003315, -0.00002118, -0.00006233, -0.00003065, -0.00000316, -0.00001546, },
		{ -0.00001056, -0.00001933, 0.00000905, 0.00000517, -0.00001480, 0.00000334, -0.00001598, -0.00002167, 0.00000612, 0.00000466, -0.00001551, -0.00000442, -0.00001853, -0.00000586, 0.00002618, 0.00000144, -0.00001867, 0.00000083, -0.00000212, 0.00000540, 0.00001692, -0.00000658, -0.00002531, 0.00000844, -0.00001120, 0.00003840, 0.00000937, -0.00002428, -0.00006702, 0.00017950, -0.00024090, 0.00010203, 0.00006572, -0.00017487, 0.00021759, -0.00022015, 0.00017057, -0.00011494, 0.00001784, -0.00002660, 0.00003421, -0.00010728, 0.00005231, -0.00008695, 0.00006054, -0.00003645, -0.00015518, -0.00015074, -0.00064113, 0.00058828, 0.00050750, 0.00011452, 0.00024186, 0.00050276, 0.00021248, -0.00034824, -0.00026360, -0.00003566, -0.00003057, -0.00005382, -0.00000810, -0.00005405, -0.00004805, -0.00000183, },
		{ -0.00000968, -0.00000336, -0.00002212, -0.00000119, 0.00001299, -0.00001322, -0.00000328, -0.00000447, -0.00001924, -0.00000589, 0.00001098, -0.00001364, -0.00000603, -0.00001068, -0.00002075, 0.00001821, 0.00002152, -0.00001907, -0.00000842, 0.00000270, -0.00000052, 0.00001324, 0.00000777, -0.00002735, 0.00000336, -0.00001359, 0.00002342, 0.00001075, 0.00003364, -0.00015129, 0.00019876, -0.00014028, -0.00005196, 0.00017532, -0.00021383, 0.00019702, -0.00014806, 0.00008715, -0.00005122, -0.00000177, -0.00002835, 0.00004983, -0.00012600, 0.00004977, -0.00004577, 0.00000515, 0.00001332, -0.00013998, -0.00015860, -0.00057748, 0.00048549, 0.00045603, 0.00015962, 0.00014868, 0.00036338, 0.00037968, -0.00016764, -0.00033271, -0.00009039, -0.00000626, -0.00006438, -0.00001103, -0.00003699, -0.00005925, },
		{ -0.00005902, -0.00002206, -0.00000108, -0.00001735, -0.00001337, 0.00001424, -0.00000421, -0.00001246, -0.00000067, -0.00000609, -0.00001776, 0.00000951, -0.00000555, -0.00001148, -0.00000257, -0.00002470, -0.00000145, 0.00003530, -0.00000603, -0.00002103, 0.00000346, 0.00000151, 0.00000330, 0.00001714, -0.00001965, -0.00000415, -0.00001820, 0.00002344, -0.00002064, 0.00010289, -0.00018521, 0.00012817, 0.00001135, -0.00017488, 0.00020620, -0.00017978, 0.00013344, -0.00007157, 0.00003486, -0.00003127, 0.00000860, -0.00003845, 0.00005940, -0.00012401, 0.00002208, 0.00000247, -0.00002836, 0.00002470, -0.00010178, -0.00016268, -0.00054317, 0.00042730, 0.00039277, 0.00017516, 0.00013807, 0.00020894, 0.00041738, 0.00004520, -0.00031647, -0.00017346, -0.00000112, -0.00005668, -0.00002356, -0.00002281, },
		{ -0.00001842, -0.00005088, -0.00003058, -0.00000722, -0.00000914, -0.00002030, 0.00000620, 0.00000775, -0.00001584, -0.00000862, 0.00001148, -0.00002023, -0.00000222, 0.00000500, -0.00001446, -0.00000118, -0.00001707, -0.00002232, 0.00003334, 0.00001683, -0.00002740, -0.00000498, 0.00000899, -0.00000454, 0.00001612, -0.00000670, -0.00000392, -0.00003537, 0.00004168, -0.00006447, 0.00014094, -0.00014250, -0.00000411, 0.00014605, -0.00021589, 0.00015799, -0.00010975, 0.00007285, -0.00002475, 0.00002258, -0.00003835, 0.00002242, -0.00003746, 0.00005513, -0.00010602, -0.00001258, 0.00003606, -0.00003248, 0.00001362, -0.00006800, -0.00014550, -0.00053332, 0.00038644, 0.00034951, 0.00015059, 0.00016725, 0.00010156, 0.00035094, 0.00021666, -0.00021389, -0.00024241, -0.00003086, -0.00003937, -0.00003282, },
		{ -0.00003105, -0.00002103, -0.00004340, -0.00003008, -0.00001822, -0.00000512, -0.00001804, -0.00000706, 0.00001421, -0.00000821, -0.00002405, 0.00002237, -0.00000801, -0.00001903, 0.00000986, -0.00000950, -0.00000736, -0.00000520, -0.00003292, 0.00001491, 0.00003807, -0.00001954, -0.00002023, 0.00001389, -0.00000386, 0.00000758, 0.00000046, 0.00001262, -0.00006443, 0.00006377, -0.00008956, 0.00012085, -0.00003761, -0.00013786, 0.00020591, -0.00016860, 0.00007131, -0.00005262, 0.00004528, -0.00001243, 0.00003330, -0.00004651, 0.00002223, -0.00002306, 0.00004423, -0.00008477, -0.00003900, 0.00004823, -0.00001777, 0.00000252, -0.00005413, -0.00011100, -0.00052829, 0.00033944, 0.00033482, 0.00010447, 0.00019197, 0.00005979, 0.00023796, 0.00030227, -0.00006305, -0.00025891, -0.00008764, -0.00003112, },
		{ -0.00004588, -0.00002254, -0.00002170, -0.00004243, -0.00002289, -0.00002612, -0.00001011, -0.00000974, -0.00001685, 0.00000918, 0.00000760, -0.00003630, 0.00001809, 0.00001493, -0.00002980, 0.00000258, 0.00000248, -0.00001483, 0.00000106, -0.00002848, -0.00001081, 0.00004521, 0.00000236, -0.00003278, 0.00000896, 0.00000400, 0.00000161, -0.00000777, 0.00004340, -0.00009103, 0.00006715, -0.00007440, 0.00004770, 0.00008222, -0.00021384, 0.00017372, -0.00007201, 0.00000106, -0.00003867, 0.00005000, -0.00001561, 0.00004298, -0.00004000, 0.00000841, -0.00000587, 0.00003743, -0.00006875, -0.00005289, 0.00004555, -0.00000073, 0.00000529, -0.00005795, -0.00007556, -0.00051114, 0.00027949, 0.00033707, 0.00006263, 0.00018960, 0.00006568, 0.00013409, 0.00030063, 0.00008182, -0.00021030, -0.00014415, },
		{ -0.00017045, -0.00008148, -0.00002048, -0.00001438, -0.00004596, -0.00001651, -0.00002517, -0.00002168, -0.00000385, -0.00001633, -0.00000551, 0.00002171, -0.00003557, -0.00000050, 0.00003626, -0.00002461, -0.00001536, 0.00001335, -0.00001510, -0.00000324, -0.00001413, -0.00003015, 0.00003294, 0.00002849, -0.00003228, -0.00000553, 0.00000991, 0.00000819, -0.00003099, 0.00007382, -0.00009528, 0.00003730, -0.00002166, -0.00004449, 0.00016380, -0.00020525, 0.00007954, 0.00001452, -0.00001680, -0.00006378, 0.00006264, -0.00001114, 0.00003909, -0.00002079, -0.00000664, 0.00000293, 0.00003862, -0.00005761, -0.00005860, 0.00003876, 0.00000775, 0.00002220, -0.00006635, -0.00005379, -0.00047801, 0.00021472, 0.00033877, 0.00004022, 0.00016237, 0.00008743, 0.00006937, 0.00024324, 0.00017818, -0.00011457, },
		{ -0.00000782, -0.00015188, -0.00011981, -0.00003600, -0.00000259, -0.00004624, -0.00001661, -0.00001687, -0.00003162, -0.00000753, -0.00000631, -0.00002057, 0.00002401, -0.00002048, -0.00002258, 0.00004268, -0.00000261, -0.00003335, 0.00001311, -0.00000466, -0.00001426, -0.00000109, -0.00003365, 0.00000706, 0.00004477, -0.00001538, -0.00002041, 0.00000346, 0.00002855, -0.00005648, 0.00008429, -0.00006561, -0.00002061, 0.00004294, -0.00011317, 0.00017692, -0.00012941, -0.00001957, 0.00004630, 0.00001579, -0.00009729, 0.00005699, 0.00000972, 0.00002672, -0.00000166, -0.00001145, 0.00000121, 0.00004302, -0.00004629, -0.00006169, 0.00003466, 0.00000634, 0.00004454, -0.00006663, -0.00004971, -0.00043699, 0.00015756, 0.00032941, 0.00003692, 0.00012579, 0.00010064, 0.00004445, 0.00017118, 0.00021189, },
		{ 0.00019590, 0.00007620, -0.00009627, -0.00013842, -0.00006799, 0.00000216, -0.00003759, -0.00002138, -0.00000884, -0.00003245, -0.00002057, 0.00000471, -0.00002572, 0.00001188, 0.00000008, -0.00003404, 0.00002924, 0.00002553, -0.00003770, -0.00000210, 0.00001103, -0.00002221, 0.00000132, -0.00002209, -0.00001868, 0.00004167, 0.00001058, -0.00002293, -0.00001829, 0.00005198, -0.00006587, 0.00006303, -0.00000852, -0.00008296, 0.00008659, -0.00013172, 0.00012836, -0.00003400, -0.00007315, 0.00001671, 0.00006482, -0.00010621, 0.00002506, 0.00003660, 0.00002008, 0.00000712, -0.00000475, -0.00000382, 0.00004319, -0.00003191, -0.00006382, 0.00003413, -0.00000016, 0.00006312, -0.00005383, -0.00005720, -0.00039958, 0.00011488, 0.00030927, 0.00004329, 0.00009434, 0.00009733, 0.00004249, 0.00011416, },
		{ 0.00008229, 0.00015593, 0.00012121, -0.00002708, -0.00012504, -0.00010191, -0.00001043, -0.00002339, -0.00002255, -0.00000805, -0.00002393, -0.00003486, 0.00000667, -0.00001758, -0.00000706, 0.00001260, -0.00002730, 0.00000373, 0.00004314, -0.00002115, -0.00002523, 0.00002049, -0.00001900, -0.00000757, -0.00000564, -0.00003111, 0.00002063, 0.00003107, -0.00000604, -0.00004628, 0.00006121, -0.00004720, 0.00001532, 0.00005291, -0.00012132, 0.00008939, -0.00010282, 0.00005370, 0.00003220, -0.00006392, -0.00004385, 0.00008939, -0.00007862, -0.00001826, 0.00005071, 0.00002823, 0.00000655, 0.00000525, -0.00000303, 0.00003647, -0.00001695, -0.00006321, 0.00003528, -0.00000658, 0.00007365, -0.00003089, -0.00006669, -0.00037267, 0.00008511, 0.00028480, 0.00005018, 0.00007384, 0.00008242, 0.00004490, },
		{ 0.00004109, 0.00007057, 0.00011463, 0.00013056, 0.00003161, -0.00008473, -0.00011888, -0.00003968, -0.00001564, -0.00001283, -0.00001390, -0.00001377, -0.00004048, -0.00000403, -0.00000279, -0.00002000, 0.00000867, -0.00000719, -0.00001744, 0.00003828, 0.00001058, -0.00004135, 0.00001453, -0.00000519, -0.00001937, 0.00000334, -0.00002583, -0.00000653, 0.00003340, 0.00002520, -0.00006317, 0.00004374, -0.00000423, -0.00003765, 0.00009171, -0.00011977, 0.00005558, -0.00005321, -0.00000617, 0.00004549, -0.00001218, -0.00009084, 0.00006873, -0.00003066, -0.00004785, 0.00004119, 0.00004672, 0.00000679, 0.00000929, 0.00000662, 0.00002728, -0.00000655, -0.00005854, 0.00003721, -0.00001057, 0.00007699, -0.00000397, -0.00007158, -0.00035602, 0.00006201, 0.00026173, 0.00005387, 0.00006246, 0.00006476, },
		{ 0.00005049, 0.00002923, 0.00006892, 0.00008339, 0.00011841, 0.00006741, -0.00003519, -0.00010861, -0.00007167, -0.00002641, 0.00000582, -0.00001741, -0.00001141, -0.00003385, -0.00002107, 0.00000651, -0.00001834, -0.00000851, 0.00001176, -0.00001999, 0.00001244, 0.00004061, -0.00003716, -0.00000580, 0.00000927, -0.00002288, -0.00000118, -0.00000945, -0.00002503, 0.00001525, 0.00005484, -0.00005393, 0.00000159, 0.00004381, -0.00007054, 0.00009298, -0.00008427, 0.00000906, -0.00001490, -0.00002838, 0.00001558, 0.00004183, -0.00009410, 0.00001388, 0.00000625, -0.00004794, 0.00001429, 0.00006123, 0.00001758, 0.00000528, 0.00002011, 0.00002307, -0.00000329, -0.00005181, 0.00004089, -0.00001200, 0.00007602, 0.00002225, -0.00007059, -0.00034514, 0.00004023, 0.00024150, 0.00005533, 0.00005580, },
		{ 0.00005082, 0.00004171, 0.00001192, 0.00006950, 0.00006369, 0.00009832, 0.00008203, 0.00000557, -0.00007647, -0.00008735, -0.00005710, 0.00002074, -0.00000815, -0.00001935, -0.00002130, -0.00003256, 0.00000194, -0.00000452, -0.00002585, 0.00001528, -0.00000172, -0.00001932, 0.00005097, -0.00001096, -0.00002781, 0.00001155, -0.00001273, -0.00001504, 0.00000538, -0.00002646, -0.00001322, 0.00006497, -0.00001644, -0.00004753, 0.00007277, -0.00006899, 0.00006250, -0.00003820, -0.00002286, -0.00000551, -0.00001716, -0.00002669, 0.00006462, -0.00005417, -0.00004209, 0.00000798, -0.00002413, -0.00001134, 0.00005920, 0.00003950, -0.00000027, 0.00002928, 0.00002791, -0.00000432, -0.00004726, 0.00004709, -0.00001085, 0.00007302, 0.00004622, -0.00006593, -0.00033574, 0.00001825, 0.00022217, 0.00005656, },
		{ 0.00005810, 0.00004629, 0.00003845, -0.00000775, 0.00006831, 0.00005303, 0.00007727, 0.00008437, 0.00003002, -0.00003959, -0.00007558, -0.00009423, 0.00001625, 0.00001575, -0.00002943, -0.00001454, -0.00003048, -0.00001404, 0.00000916, -0.00002966, -0.00000028, 0.00002542, -0.00003756, 0.00003462, 0.00002411, -0.00003428, -0.00000451, 0.00000606, -0.00002724, 0.00000884, -0.00001379, -0.00003539, 0.00004700, 0.00003511, -0.00007937, 0.00006824, -0.00003825, 0.00002253, -0.00000720, -0.00002643, -0.00002038, 0.00000478, -0.00005179, 0.00004777, 0.00000228, -0.00006710, -0.00002589, 0.00000068, -0.00001899, 0.00003982, 0.00006337, 0.00000179, 0.00002992, 0.00003928, -0.00000331, -0.00004778, 0.00005436, -0.00000635, 0.00006908, 0.00006848, -0.00006019, -0.00032629, -0.00000236, 0.00020155, },
		{ 0.00017918, 0.00005926, 0.00004130, 0.00004034, -0.00002716, 0.00006335, 0.00004930, 0.00005703, 0.00008127, 0.00004275, -0.00001473, -0.00004169, -0.00011627, -0.00001393, 0.00004287, -0.00002810, -0.00002068, -0.00001768, -0.00002930, 0.00000961, -0.00001603, -0.00002483, 0.00004320, -0.00003141, 0.00000124, 0.00004674, -0.00001489, -0.00003252, 0.00002092, -0.00002819, 0.00000029, 0.00000130, -0.00003861, 0.00000767, 0.00007708, -0.00007737, 0.00003467, 0.00000036, -0.00000256, -0.00000360, -0.00000802, -0.00003975, 0.00001463, -0.00004893, 0.00000808, 0.00004243, -0.00005017, -0.00007252, 0.00000278, -0.00000668, 0.00001433, 0.00007757, 0.00001564, 0.00002496, 0.00005043, 0.00000469, -0.00005214, 0.00005933, 0.00000204, 0.00006501, 0.00009002, -0.00005456, -0.00031771, -0.00001953, },
		{ -0.00003248, 0.00015609, 0.00005970, 0.00003449, 0.00004634, -0.00004339, 0.00005315, 0.00005087, 0.00003888, 0.00007350, 0.00005321, -0.00000704, -0.00000471, -0.00010774, -0.00005912, 0.00005413, -0.00000775, -0.00003485, -0.00000664, -0.00003103, -0.00000676, 0.00000557, -0.00004144, 0.00003995, -0.00000642, -0.00002775, 0.00004047, 0.00002554, -0.00005609, 0.00002045, -0.00001680, -0.00001204, 0.00000683, -0.00002212, -0.00003418, 0.00008955, -0.00004297, -0.00000706, 0.00002307, -0.00000124, -0.00002053, 0.00001296, -0.00004476, 0.00000449, -0.00002837, -0.00002778, 0.00004813, -0.00000619, -0.00010173, -0.00002558, 0.00001133, -0.00000335, 0.00007702, 0.00003736, 0.00002170, 0.00005577, 0.00001979, -0.00005563, 0.00005917, 0.00001330, 0.00006182, 0.00011149, -0.00004876, -0.00031173, },
		},
		{ { -0.00001772, -0.00002039, -0.00001632, -0.00000083, 0.00000158, 0.00000253, 0.00000327, -0.00003333, -0.00004232, -0.00002255, -0.00005992, -0.00005183, 0.00000652, 0.00000987, 0.00006445, 0.00008941, -0.00001815, -0.00001366, 0.00004893, -0.00000199, 0.00000179, 0.00002756, 0.00001482, 0.00003674, -0.00003041, -0.00005404, 0.00007738, 0.00005751, -0.00004041, -0.00004279, -0.00007159, 0.00001779, 0.00013847, 0.00001779, -0.00007159, -0.00004279, -0.00004041, 0.00005751, 0.00007738, -0.00005404, -0.00003041, 0.00003674, 0.00001482, 0.00002756, 0.00000179, -0.00000199, 0.00004893, -0.00001366, -0.00001815, 0.00008941, 0.00006445, 0.00000987, 0.00000652, -0.00005183, -0.00005992, -0.00002255, -0.00004232, -0.00003333, 0.00000327, 0.00000253, 0.00000158, -0.00000083, -0.00001632, -0.00002039, },
		{ -0.00001810, -0.00001941, -0.00002028, -0.00001841, -0.00000869, -0.00000608, -0.00000337, 0.00001604, -0.00000510, -0.00004481, -0.00003093, -0.00004550, -0.00005860, 0.00000095, 0.00000114, 0.00001688, 0.00010200, 0.00002941, -0.00004996, 0.00003269, 0.00003550, -0.00000227, 0.00001300, -0.00001356, 0.00002801, 0.00004158, -0.00006140, -0.00001704, 0.00005376, 0.00000732, 0.00002787, -0.00002834, -0.00011305, 0.00003135, 0.00011733, 0.00003495, -0.00002664, -0.00010527, -0.00004977, 0.00010819, 0.00004666, -0.00004619, -0.00000503, 0.00000773, 0.00003863, 0.00004421, -0.00001094, 0.00002192, 0.00001941, -0.00002801, 0.00005264, 0.00009086, 0.00003311, 0.00001618, -0.00003477, -0.00007550, -0.00003238, -0.00002816, -0.00004142, -0.00001415, -0.00000661, -0.00000276, 0.00000826, -0.00000450, },
		{ 0.00000508, -0.00000575, -0.00001427, -0.00002330, -0.00002226, -0.00000916, -0.00000648, -0.00001632, 0.00000371, 0.00001837, -0.00002100, -0.00003457, -0.00004062, -0.00006686, -0.00000866, 0.00001837, -0.00002186, 0.00006368, 0.00008554, -0.00004068, -0.00001608, 0.00004627, 0.00001339, 0.00002622, -0.00001266, -0.00003347, 0.00005445, 0.00000790, -0.00004943, 0.00000980, -0.00001634, 0.00001805, 0.00008634, -0.00005364, -0.00010769, 0.00002330, 0.00008820, 0.00008800, -0.00003359, -0.00014937, -0.00000669, 0.00009946, 0.00002213, -0.00000520, -0.00001660, 0.00000080, 0.00006796, 0.00003347, 0.00000609, 0.00002426, -0.00001995, 0.00001671, 0.00009567, 0.00006191, 0.00002477, -0.00001446, -0.00007512, -0.00005213, -0.00002464, -0.00003914, -0.00002093, -0.00001175, -0.00001730, 0.00000041, },
		{ -0.00001610, -0.00000321, -0.00000069, 0.00000081, -0.00001461, -0.00002875, -0.00001658, -0.00000124, -0.00001364, -0.00001714, 0.00001138, 0.00000316, -0.00001934, -0.00002694, -0.00007147, -0.00004066, 0.00003781, -0.00001429, -0.00000019, 0.00009704, 0.00000994, -0.00004762, 0.00002804, 0.00001518, 0.00002858, 0.00003753, -0.00005578, -0.00001084, 0.00004842, -0.00002162, -0.00000466, -0.00001837, -0.00006212, 0.00006721, 0.00008770, -0.00005856, -0.00008422, -0.00001689, 0.00009741, 0.00011851, -0.00007324, -0.00012475, 0.00001708, 0.00005209, 0.00004255, 0.00001662, -0.00003657, 0.00002343, 0.00006819, 0.00003223, 0.00003151, -0.00000558, -0.00001494, 0.00007685, 0.00009331, 0.00004212, -0.00000260, -0.00006730, -0.00006617, -0.00002752, -0.00003888, -0.00002783, -0.00000746, -0.00002104, },
		{ -0.00001361, -0.00002112, -0.00001516, -0.00001438, 0.00000262, 0.00000490, -0.00001973, -0.00002631, -0.00000815, -0.00000365, -0.00001862, -0.00000781, 0.00000253, -0.00000881, -0.00000312, -0.00004681, -0.00007683, 0.00001832, 0.00002445, -0.00003307, 0.00005936, 0.00005568, -0.00004771, 0.00000683, 0.00001907, -0.00000664, 0.00006750, 0.00000400, -0.00006741, 0.00001978, 0.00001072, -0.00001186, 0.00001518, -0.00006918, -0.00004754, 0.00008569, 0.00006421, -0.00002544, -0.00008809, -0.00004089, 0.00012542, 0.00009164, -0.00007528, -0.00006770, -0.00001389, 0.00002955, 0.00007654, 0.00000639, -0.00002716, 0.00003959, 0.00005205, 0.00005361, 0.00002984, -0.00002404, 0.00003444, 0.00010593, 0.00007705, 0.00001447, -0.00006417, -0.00008211, -0.00003100, -0.00003321, -0.00003887, -0.00000867, },
		{ -0.00001664, -0.00000915, -0.00001953, -0.00001395, -0.00002539, -0.00001723, 0.00000815, -0.00000049, -0.00001815, -0.00001682, -0.00000794, -0.00001148, -0.00001153, -0.00000471, -0.00001717, -0.00000127, -0.00000280, -0.00007091, -0.00003033, 0.00004032, -0.00002292, 0.00001329, 0.00007650, -0.00003228, -0.00002272, 0.00004554, -0.00003033, 0.00001985, 0.00007724, -0.00004274, -0.00003798, 0.00002091, -0.00001608, 0.00001101, -0.00000979, -0.00007919, -0.00001678, 0.00005553, 0.00006654, 0.00000244, -0.00010581, -0.00002175, 0.00011436, 0.00005199, -0.00001943, -0.00004672, -0.00005587, 0.00004678, 0.00007983, 0.00000088, 0.00000216, 0.00003023, 0.00005081, 0.00006862, 0.00001135, 0.00000277, 0.00008457, 0.00010831, 0.00005182, -0.00005216, -0.00010667, -0.00004833, -0.00002226, -0.00004208, },
		{ -0.00003873, -0.00002240, -0.00000384, -0.00002170, -0.00001247, -0.00001846, -0.00003266, -0.00000949, 0.00000107, -0.00000425, -0.00000529, -0.00001163, -0.00001719, -0.00001266, 0.00000044, -0.00001678, -0.00002609, 0.00000923, -0.00002582, -0.00005181, 0.00002210, -0.00000495, -0.00002305, 0.00008046, 0.00000608, -0.00007212, 0.00005256, 0.00001638, -0.00005550, 0.00006617, 0.00004056, -0.00005814, -0.00000365, 0.00000341, -0.00001644, 0.00000220, -0.00003279, -0.00004393, -0.00002377, 0.00002325, 0.00008645, -0.00000192, -0.00009033, 0.00000014, 0.00005938, 0.00005971, 0.00003744, -0.00006778, -0.00005716, 0.00006134, 0.00005996, 0.00002304, 0.00001768, 0.00001728, 0.00006178, 0.00005866, 0.00001603, 0.00005739, 0.00011515, 0.00009608, -0.00001619, -0.00012406, -0.00008544, -0.00002236, },
		{ -0.00003893, -0.00003881, -0.00003469, 0.00000610, -0.00001409, -0.00002050, -0.00001126, -0.00003115, -0.00002163, -0.00001079, -0.00001363, 0.00000596, 0.00000781, -0.00001615, -0.00002663, -0.00000487, 0.00000353, -0.00003157, -0.00002031, -0.00000254, -0.00003401, 0.00000966, 0.00001224, -0.00005739, 0.00004928, 0.00007765, -0.00009226, -0.00001848, 0.00009120, -0.00004785, -0.00002825, 0.00008121, -0.00000695, -0.00004127, 0.00001156, -0.00000252, -0.00002263, -0.00003005, -0.00001220, -0.00001746, -0.00005751, 0.00001954, 0.00008096, -0.00000637, -0.00004223, -0.00001981, 0.00002006, 0.00010438, 0.00004506, -0.00007729, -0.00002808, 0.00004589, 0.00005403, 0.00005161, 0.00001517, 0.00002028, 0.00005975, 0.00004623, 0.00005499, 0.00011200, 0.00013276, 0.00003906, -0.00011925, -0.00013087, },
		{ -0.00017726, -0.00007184, -0.00003529, -0.00005849, -0.00000169, 0.00000783, -0.00001929, -0.00001166, -0.00002987, -0.00002624, -0.00000745, -0.00002964, -0.00001555, 0.00002232, 0.00001138, -0.00002400, -0.00003012, 0.00000315, -0.00000324, -0.00003391, -0.00002096, -0.00002693, 0.00000756, 0.00005025, -0.00005843, -0.00003822, 0.00012133, -0.00001484, -0.00012155, 0.00007309, 0.00005693, -0.00008258, 0.00001663, 0.00005683, -0.00003128, -0.00001997, 0.00002101, -0.00000186, -0.00005574, -0.00003727, 0.00002783, -0.00002890, -0.00007185, 0.00001716, 0.00004844, 0.00002255, -0.00001289, -0.00005894, 0.00003424, 0.00012529, 0.00002507, -0.00004616, -0.00001196, 0.00002340, 0.00006939, 0.00006196, 0.00001620, 0.00002753, 0.00004419, 0.00005654, 0.00010976, 0.00016518, 0.00011233, -0.00008378, },
		{ 0.00000513, -0.00020570, -0.00013605, -0.00002983, -0.00006884, -0.00003737, 0.00001563, -0.00000505, -0.00000237, -0.00002459, -0.00004410, -0.00000185, -0.00001540, -0.00004388, -0.00000229, 0.00002834, 0.00001058, -0.00002950, -0.00002969, 0.00000439, -0.00000892, -0.00002759, -0.00004183, -0.00002506, 0.00007577, 0.00001858, -0.00011290, 0.00004868, 0.00010970, -0.00011824, -0.00005913, 0.00011925, -0.00000779, -0.00007017, 0.00003773, 0.00002312, -0.00003518, -0.00000326, 0.00003536, -0.00002259, -0.00008519, -0.00000645, 0.00003802, -0.00004702, -0.00006131, -0.00001397, 0.00003012, 0.00005994, -0.00002514, -0.00006003, 0.00006738, 0.00010087, 0.00002776, -0.00000373, -0.00001749, 0.00002304, 0.00008219, 0.00005994, 0.00002698, 0.00002596, 0.00003700, 0.00008647, 0.00017657, 0.00019961, },
		{ 0.00025224, 0.00014818, -0.00016261, -0.00022151, -0.00005746, -0.00005452, -0.00007178, -0.00000302, -0.00000324, 0.00000640, 0.00000664, -0.00005419, -0.00002756, 0.00000770, -0.00003565, -0.00003346, 0.00000408, 0.00002534, 0.00001263, -0.00003382, -0.00002803, -0.00000149, -0.00000582, -0.00002723, -0.00006680, 0.00001965, 0.00010979, -0.00005957, -0.00009084, 0.00012085, 0.00001934, -0.00014841, 0.00003033, 0.00009901, -0.00004853, -0.00004159, 0.00003262, 0.00000025, -0.00003222, 0.00001804, 0.00003490, -0.00005962, -0.00007379, 0.00002071, 0.00001674, -0.00003870, -0.00006154, -0.00005315, 0.00003941, 0.00005905, -0.00003876, -0.00001523, 0.00006985, 0.00007151, 0.00005835, 0.00001580, -0.00001673, 0.00003975, 0.00008015, 0.00005981, 0.00003657, 0.00002553, 0.00004484, 0.00013508, },
		{ 0.00006772, 0.00021794, 0.00027233, -0.00002382, -0.00026220, -0.00012979, -0.00004625, -0.00008900, -0.00002657, -0.00001001, -0.00001647, 0.00003457, -0.00001812, -0.00005720, -0.00000349, -0.00001471, -0.00003166, -0.00001572, -0.00000726, 0.00002709, 0.00001233, -0.00003384, -0.00002836, -0.00001305, 0.00000696, -0.00003704, -0.00007162, 0.00007738, 0.00008253, -0.00011305, -0.00001677, 0.00011651, -0.00007035, -0.00010203, 0.00007900, 0.00005225, -0.00005557, -0.00002352, 0.00001964, -0.00001398, -0.00001989, 0.00003531, 0.00000485, -0.00007225, -0.00003600, 0.00001143, 0.00000312, -0.00002004, -0.00008377, -0.00006381, 0.00003837, 0.00002318, -0.00001364, 0.00002788, 0.00004781, 0.00007371, 0.00008179, 0.00001917, 0.00000591, 0.00005429, 0.00007187, 0.00006036, 0.00004605, 0.00003407, },
		{ 0.00006204, 0.00004596, 0.00013405, 0.00030060, 0.00014281, -0.00021426, -0.00020184, -0.00005553, -0.00010283, -0.00006052, 0.00000200, -0.00004510, 0.00000770, 0.00002992, -0.00004166, -0.00002326, -0.00001761, -0.00003285, 0.00000120, -0.00002004, -0.00001732, 0.00002948, 0.00000766, -0.00002071, -0.00004341, -0.00002337, 0.00003098, -0.00005200, -0.00005542, 0.00010954, 0.00002689, -0.00010864, 0.00003908, 0.00005671, -0.00009960, -0.00003736, 0.00007762, 0.00001659, -0.00004874, -0.00001604, 0.00000602, -0.00001908, 0.00000214, 0.00002453, -0.00002851, -0.00004327, -0.00001721, -0.00001413, 0.00001132, -0.00002222, -0.00009603, -0.00005427, -0.00000175, -0.00000194, 0.00003142, 0.00003938, 0.00004456, 0.00009475, 0.00008074, 0.00003028, 0.00003886, 0.00006274, 0.00006252, 0.00006178, },
		{ 0.00005435, 0.00007394, 0.00007179, 0.00008484, 0.00025710, 0.00026136, -0.00011386, -0.00025245, -0.00006226, -0.00008852, -0.00012471, 0.00000051, -0.00002668, -0.00004449, 0.00002643, -0.00001103, -0.00001368, -0.00000578, -0.00006154, -0.00000851, 0.00002192, -0.00003646, -0.00001446, 0.00001310, 0.00002056, -0.00000406, -0.00008643, -0.00001049, 0.00006271, -0.00007460, -0.00003401, 0.00011219, -0.00001292, -0.00007635, 0.00004449, 0.00000354, -0.00007758, 0.00000057, 0.00005450, -0.00000280, -0.00003963, -0.00001378, -0.00000718, -0.00000794, 0.00001226, -0.00001164, -0.00002764, -0.00000582, -0.00002910, -0.00002218, 0.00001462, -0.00003728, -0.00007817, -0.00006337, -0.00005600, 0.00000100, 0.00005160, 0.00004283, 0.00007536, 0.00010607, 0.00007152, 0.00005333, 0.00006996, 0.00006759, },
		{ 0.00009478, 0.00006833, 0.00006319, 0.00008610, 0.00007583, 0.00020513, 0.00033957, 0.00001612, -0.00030695, -0.00010779, -0.00001059, -0.00015828, -0.00006649, 0.00000547, -0.00005495, -0.00000548, -0.00001368, -0.00002126, 0.00004203, -0.00004416, -0.00007313, 0.00003090, 0.00000992, -0.00002804, -0.00002891, -0.00001157, 0.00007309, -0.00002382, -0.00013953, 0.00004867, 0.00007802, -0.00011053, -0.00000985, 0.00010573, -0.00003148, -0.00005654, 0.00002527, -0.00001287, -0.00004521, 0.00000968, 0.00002939, -0.00001236, -0.00002865, -0.00001800, -0.00001029, 0.00000990, -0.00000767, -0.00003631, 0.00000345, 0.00000411, -0.00004171, -0.00001701, -0.00000328, -0.00003413, -0.00005524, -0.00009782, -0.00008640, -0.00000004, 0.00004045, 0.00006939, 0.00011753, 0.00010648, 0.00007062, 0.00007915, },
		{ 0.00007951, 0.00010161, 0.00010523, 0.00008059, 0.00009356, 0.00007165, 0.00013839, 0.00038082, 0.00021291, -0.00031274, -0.00026592, 0.00004514, -0.00007946, -0.00014057, -0.00001607, -0.00004577, -0.00001731, -0.00000602, -0.00007512, 0.00003501, 0.00004718, -0.00008834, -0.00003373, 0.00002519, 0.00001138, -0.00000255, -0.00009277, 0.00001001, 0.00013588, -0.00010890, -0.00014590, 0.00013962, 0.00004288, -0.00014408, 0.00002854, 0.00009493, -0.00004936, -0.00005165, 0.00001521, -0.00000678, -0.00002416, 0.00000345, 0.00000620, -0.00001217, -0.00002041, -0.00002390, 0.00000165, 0.00000932, -0.00003973, -0.00002561, 0.00002637, -0.00000639, -0.00003628, -0.00002403, -0.00002090, -0.00001225, -0.00005596, -0.00012446, -0.00009738, -0.00003242, 0.00002701, 0.00011388, 0.00015361, 0.00011245, },
		{ 0.00018877, 0.00013290, 0.00009807, 0.00010858, 0.00010551, 0.00012910, 0.00010780, 0.00006462, 0.00031806, 0.00042936, -0.00014447, -0.00045795, -0.00006600, 0.00003837, -0.00012402, -0.00004987, -0.00005833, -0.00005391, 0.00003980, -0.00007826, -0.00006817, 0.00008530, -0.00000640, -0.00006055, -0.00001706, -0.00001578, 0.00007556, -0.00002942, -0.00016802, 0.00011100, 0.00013131, -0.00021929, -0.00006191, 0.00019833, -0.00004361, -0.00013855, 0.00007486, 0.00006396, -0.00006967, -0.00004113, 0.00001739, 0.00000187, -0.00001395, -0.00001234, -0.00001011, -0.00000456, -0.00001953, -0.00001952, 0.00001230, -0.00001542, -0.00004913, 0.00000054, 0.00002260, -0.00000980, -0.00002913, -0.00004219, -0.00001785, -0.00000131, -0.00006901, -0.00012440, -0.00011953, -0.00008249, 0.00002491, 0.00015387, },
		{ 0.00001991, 0.00018224, 0.00023025, 0.00017396, 0.00011996, 0.00008737, 0.00013072, 0.00019295, 0.00008576, 0.00017444, 0.00050642, 0.00017090, -0.00048176, -0.00030786, 0.00003839, -0.00006675, -0.00003758, -0.00002361, -0.00013053, 0.00001809, 0.00001635, -0.00013032, 0.00000110, 0.00004196, -0.00001567, 0.00001349, -0.00009135, 0.00000194, 0.00015790, -0.00014845, -0.00015965, 0.00023196, 0.00001873, -0.00025880, 0.00007757, 0.00016426, -0.00012685, -0.00007837, 0.00009475, 0.00001156, -0.00007269, -0.00002084, 0.00001898, 0.00000931, -0.00001414, -0.00003240, -0.00001360, 0.00000079, -0.00001902, -0.00000584, 0.00000233, -0.00003831, -0.00003010, 0.00001067, 0.00001028, -0.00000387, -0.00003539, -0.00004777, -0.00000836, -0.00001216, -0.00006747, -0.00011690, -0.00015824, -0.00012661, },
		{ -0.00019261, -0.00016549, 0.00000043, 0.00020037, 0.00028295, 0.00024111, 0.00012177, 0.00006315, 0.00019836, 0.00020883, 0.00012102, 0.00042271, 0.00043520, -0.00029452, -0.00049699, -0.00007302, -0.00006618, -0.00007021, 0.00008120, -0.00010546, -0.00010988, 0.00006165, -0.00006866, -0.00006957, 0.00000365, -0.00004579, 0.00009617, 0.00000571, -0.00019694, 0.00013104, 0.00014418, -0.00028446, -0.00002915, 0.00025772, -0.00013402, -0.00017177, 0.00017044, 0.00005145, -0.00014052, -0.00000121, 0.00006323, -0.00003262, -0.00004989, -0.00000577, 0.00001819, 0.00001477, -0.00002645, -0.00004482, -0.00000881, -0.00000024, -0.00001030, -0.00000035, -0.00002116, -0.00003619, -0.00000887, 0.00000345, 0.00000611, -0.00000613, -0.00004192, -0.00003664, -0.00001413, -0.00002435, -0.00005348, -0.00012043, },
		{ -0.00004772, -0.00012471, -0.00021384, -0.00020870, -0.00003411, 0.00020605, 0.00036037, 0.00031485, 0.00010027, 0.00008954, 0.00024627, 0.00019471, 0.00034983, 0.00058022, -0.00002522, -0.00058456, -0.00020299, -0.00004649, -0.00020263, 0.00006936, 0.00007179, -0.00015730, -0.00002190, -0.00002164, -0.00006366, 0.00002976, -0.00013688, -0.00001842, 0.00023969, -0.00014688, -0.00018999, 0.00028095, -0.00000564, -0.00030562, 0.00013181, 0.00014503, -0.00020001, -0.00001759, 0.00014437, -0.00005188, -0.00007738, 0.00003203, 0.00000150, -0.00003944, -0.00002013, -0.00000277, 0.00002068, 0.00001045, -0.00004268, -0.00004273, -0.00000637, -0.00000102, 0.00000037, -0.00001149, -0.00003331, -0.00001796, -0.00000311, -0.00000298, 0.00000177, -0.00001715, -0.00003382, -0.00002863, -0.00002859, -0.00002623, },
		{ -0.00003485, -0.00003239, -0.00005038, -0.00011722, -0.00022974, -0.00025427, -0.00009060, 0.00021029, 0.00046121, 0.00036109, 0.00011427, 0.00015535, 0.00020613, 0.00031798, 0.00069857, 0.00032226, -0.00059876, -0.00043076, 0.00004884, -0.00022815, -0.00015643, 0.00015924, -0.00002799, -0.00006253, -0.00004123, -0.00013440, 0.00010362, -0.00001749, -0.00029624, 0.00018353, 0.00023523, -0.00032696, -0.00002412, 0.00029971, -0.00018000, -0.00017726, 0.00018349, -0.00001237, -0.00012630, 0.00007793, 0.00003494, -0.00006885, -0.00000148, 0.00000123, -0.00004091, -0.00001516, -0.00000499, -0.00000194, 0.00002214, -0.00000638, -0.00004833, -0.00003277, -0.00000904, 0.00000187, 0.00000044, -0.00002502, -0.00002602, -0.00000535, -0.00000438, -0.00000561, -0.00001135, -0.00002087, -0.00002134, -0.00003033, },
		{ -0.00001840, -0.00002721, -0.00003916, -0.00005033, -0.00004683, -0.00010580, -0.00023784, -0.00030571, -0.00017032, 0.00023353, 0.00054097, 0.00040212, 0.00022061, 0.00015563, 0.00017425, 0.00069597, 0.00079336, -0.00035033, -0.00079889, -0.00003033, -0.00001017, -0.00032425, 0.00000094, 0.00005512, -0.00000319, 0.00006072, -0.00023853, -0.00006842, 0.00027528, -0.00025880, -0.00026615, 0.00040252, 0.00003686, -0.00035241, 0.00016769, 0.00015638, -0.00022684, -0.00001646, 0.00009306, -0.00007650, 0.00000487, 0.00005086, -0.00005094, -0.00001078, 0.00001887, -0.00005280, -0.00003880, 0.00000885, -0.00000320, 0.00000361, 0.00001168, -0.00002405, -0.00003866, -0.00002509, -0.00001268, 0.00000173, -0.00000829, -0.00002557, -0.00001370, -0.00000174, -0.00000396, -0.00001395, -0.00002210, -0.00001667, },
		{ -0.00002183, -0.00001964, -0.00001627, -0.00001787, -0.00005528, -0.00006377, -0.00004171, -0.00009436, -0.00023179, -0.00037818, -0.00024438, 0.00025381, 0.00056518, 0.00052944, 0.00035527, 0.00005970, 0.00038735, 0.00109313, 0.00031225, -0.00094751, -0.00047525, 0.00014598, -0.00021415, -0.00015960, -0.00001283, -0.00008968, 0.00022555, -0.00002324, -0.00042372, 0.00021060, 0.00022118, -0.00049546, -0.00001617, 0.00042443, -0.00018840, -0.00019299, 0.00021606, -0.00001220, -0.00012660, 0.00003519, -0.00002818, -0.00001397, 0.00006575, -0.00004002, -0.00004753, 0.00004322, -0.00002013, -0.00007619, -0.00001231, 0.00001444, -0.00000016, 0.00000533, -0.00001019, -0.00002802, -0.00002208, -0.00002295, -0.00001679, -0.00000511, -0.00001330, -0.00001704, -0.00000605, 0.00000085, -0.00000852, -0.00002336, },
		{ -0.00001930, -0.00002054, -0.00002290, -0.00002827, -0.00000545, -0.00002120, -0.00007032, -0.00006686, -0.00004968, -0.00006799, -0.00023013, -0.00045845, -0.00029248, 0.00020134, 0.00059729, 0.00077085, 0.00038749, 0.00008314, 0.00087402, 0.00098986, -0.00052683, -0.00093399, -0.00005331, -0.00007344, -0.00016986, 0.00004417, -0.00028687, 0.00000844, 0.00043297, -0.00033941, -0.00034512, 0.00046603, -0.00006789, -0.00049181, 0.00024247, 0.00022962, -0.00024508, -0.00000850, 0.00010909, -0.00005891, -0.00000854, -0.00003163, -0.00005532, 0.00007682, 0.00001919, -0.00008791, 0.00001052, 0.00003592, -0.00006503, -0.00005723, 0.00000498, 0.00001069, 0.00000705, -0.00000706, -0.00002860, -0.00001617, -0.00000937, -0.00002533, -0.00002274, -0.00001204, -0.00001223, -0.00000901, 0.00000174, -0.00000137, },
		{ 0.00000819, -0.00001526, -0.00002105, -0.00001179, -0.00003312, -0.00002552, -0.00000410, -0.00003672, -0.00006368, -0.00008100, -0.00005815, -0.00003054, -0.00025067, -0.00049720, -0.00036930, 0.00006493, 0.00074407, 0.00098868, 0.00038403, 0.00050409, 0.00117285, 0.00014888, -0.00100475, -0.00036312, -0.00008253, -0.00032130, 0.00020850, -0.00007014, -0.00053734, 0.00037181, 0.00033725, -0.00061321, -0.00001282, 0.00044647, -0.00034007, -0.00025264, 0.00029082, 0.00001326, -0.00012777, 0.00004226, -0.00000849, 0.00001053, 0.00000864, -0.00010011, 0.00001285, 0.00008759, -0.00005691, -0.00005476, 0.00004258, -0.00001510, -0.00007052, -0.00002602, 0.00000392, 0.00001168, 0.00000648, -0.00002850, -0.00002985, 0.00000125, -0.00000637, -0.00002994, -0.00002605, -0.00001596, -0.00001245, 0.00000158, },
		{ -0.00001124, 0.00001734, 0.00000081, -0.00002785, -0.00000957, -0.00001473, -0.00003389, -0.00001842, -0.00002992, -0.00003421, -0.00005045, -0.00011136, -0.00004602, -0.00002045, -0.00025430, -0.00049233, -0.00055573, -0.00003952, 0.00096048, 0.00106438, 0.00071277, 0.00117703, 0.00064004, -0.00091676, -0.00060613, 0.00009042, -0.00057626, -0.00013181, 0.00053574, -0.00051322, -0.00043016, 0.00067905, -0.00003665, -0.00057427, 0.00028500, 0.00018225, -0.00036504, -0.00001392, 0.00014151, -0.00004733, -0.00000114, -0.00002564, -0.00001988, 0.00007000, -0.00005298, -0.00007842, 0.00007230, 0.00001784, -0.00007284, 0.00000414, 0.00001460, -0.00005270, -0.00004152, -0.00001112, -0.00000103, 0.00001477, -0.00000947, -0.00004184, -0.00001425, 0.00001089, -0.00001268, -0.00002970, -0.00002388, -0.00002426, },
		{ -0.00002587, -0.00003704, 0.00000118, 0.00002316, -0.00001683, -0.00002182, -0.00000634, -0.00002037, -0.00001341, -0.00003871, -0.00005829, -0.00000635, -0.00006016, -0.00012316, -0.00004025, -0.00004320, -0.00018204, -0.00053930, -0.00080772, -0.00002480, 0.00102494, 0.00123247, 0.00164634, 0.00136278, -0.00079253, -0.00126580, 0.00034968, -0.00015678, -0.00096008, 0.00046453, 0.00040364, -0.00088595, -0.00000288, 0.00062883, -0.00036159, -0.00027350, 0.00030928, -0.00005769, -0.00018390, 0.00003497, -0.00000986, 0.00002290, 0.00001312, -0.00007772, 0.00003927, 0.00004172, -0.00008407, -0.00000033, 0.00004902, -0.00004546, -0.00002448, 0.00001921, -0.00003712, -0.00004758, -0.00001237, -0.00001191, 0.00000059, 0.00000666, -0.00003135, -0.00003458, 0.00000475, 0.00000468, -0.00001890, -0.00001809, },
		{ -0.00001079, -0.00000417, -0.00004052, -0.00003791, 0.00001539, 0.00000723, -0.00001904, -0.00001024, -0.00002145, -0.00000053, -0.00000191, -0.00008573, -0.00005338, 0.00000663, -0.00007365, -0.00010393, -0.00009123, -0.00002640, -0.00007550, -0.00070204, -0.00095686, -0.00005896, 0.00088590, 0.00203594, 0.00278416, 0.00032691, -0.00227011, -0.00046137, 0.00095068, -0.00093381, -0.00072815, 0.00094172, -0.00011937, -0.00083388, 0.00036434, 0.00028860, -0.00038250, -0.00000888, 0.00012530, -0.00008366, -0.00002342, -0.00005945, -0.00003219, 0.00007568, -0.00004084, -0.00004395, 0.00005986, -0.00002258, -0.00003586, 0.00003747, -0.00002298, -0.00004105, 0.00002181, -0.00001703, -0.00006081, -0.00001856, -0.00000282, -0.00001214, 0.00000138, -0.00001643, -0.00004040, -0.00001040, 0.00000983, -0.00001220, },
		{ -0.00000596, -0.00001985, 0.00001121, -0.00000750, -0.00005271, -0.00002097, 0.00000603, -0.00000964, 0.00000137, -0.00002147, -0.00002518, 0.00003607, -0.00003098, -0.00010982, -0.00003618, -0.00001611, -0.00004096, -0.00011186, -0.00016988, 0.00007563, -0.00006901, -0.00084647, -0.00098547, -0.00044879, 0.00101201, 0.00369449, 0.00308489, -0.00177323, -0.00266145, 0.00089905, 0.00061973, -0.00149561, -0.00002250, 0.00087555, -0.00056462, -0.00044329, 0.00038644, -0.00000860, -0.00018153, 0.00002782, -0.00001990, 0.00003043, -0.00001512, -0.00012645, 0.00001987, 0.00004201, -0.00005315, 0.00001472, 0.00001082, -0.00003327, 0.00003150, 0.00000075, -0.00006146, 0.00000540, 0.00001661, -0.00005775, -0.00004492, 0.00000177, -0.00000479, -0.00000633, -0.00001140, -0.00004085, -0.00002588, 0.00001162, },
		{ 0.00000283, 0.00001072, -0.00002673, -0.00000307, 0.00002250, -0.00002658, -0.00003863, -0.00001268, -0.00002609, 0.00000192, 0.00001212, -0.00004502, 0.00000605, 0.00003851, -0.00006860, -0.00008817, -0.00006338, -0.00002236, 0.00002038, -0.00018927, -0.00017166, 0.00015094, -0.00016886, -0.00076135, -0.00123360, -0.00112777, 0.00212952, 0.00550898, 0.00176211, -0.00375325, -0.00164233, 0.00172213, -0.00030718, -0.00133600, 0.00054876, 0.00039602, -0.00059701, -0.00009981, 0.00015995, -0.00005810, -0.00002379, -0.00007132, -0.00000401, 0.00009571, -0.00008324, -0.00009578, 0.00002976, -0.00001173, -0.00000319, 0.00001683, -0.00003829, 0.00001648, 0.00004162, -0.00005393, -0.00003791, 0.00002927, -0.00002479, -0.00006348, -0.00001639, 0.00000321, -0.00000333, -0.00000525, -0.00003714, -0.00004638, },
		{ -0.00005779, -0.00002558, 0.00001931, -0.00001298, -0.00002356, 0.00002228, 0.00000309, -0.00002938, -0.00000788, -0.00003849, -0.00004275, 0.00002780, -0.00000816, -0.00003518, 0.00003313, -0.00000614, -0.00005181, -0.00007749, -0.00012527, 0.00003373, 0.00002142, -0.00024914, -0.00009755, 0.00003821, -0.00011703, -0.00054688, -0.00209051, -0.00135949, 0.00437414, 0.00616287, -0.00079294, -0.00435821, -0.00003970, 0.00147385, -0.00099225, -0.00067837, 0.00058971, -0.00002266, -0.00033373, -0.00003513, -0.00002272, 0.00003801, -0.00002866, -0.00011111, 0.00007520, 0.00005066, -0.00009849, -0.00004251, -0.00001894, -0.00000946, 0.00003645, -0.00002795, -0.00002702, 0.00005625, -0.00000334, -0.00006210, 0.00000332, 0.00000516, -0.00005489, -0.00003796, -0.00000187, -0.00000179, 0.00000200, -0.00001991, },
		{ -0.00000171, -0.00004470, -0.00005232, 0.00000346, 0.00000335, -0.00002946, 0.00000974, 0.00001936, -0.00002592, 0.00000327, -0.00000195, -0.00007535, -0.00002447, 0.00002281, -0.00002107, 0.00001014, -0.00000186, -0.00003061, 0.00000763, -0.00013027, -0.00013748, 0.00008259, -0.00004251, -0.00016560, -0.00015467, -0.00016974, 0.00033024, -0.00074219, -0.00334733, -0.00025783, 0.00676633, 0.00486752, -0.00314822, -0.00338845, 0.00097034, 0.00060008, -0.00097489, -0.00013270, 0.00026598, -0.00010596, -0.00011329, -0.00013712, -0.00003231, 0.00007761, -0.00008862, -0.00004488, 0.00008313, -0.00001231, -0.00004843, -0.00004350, -0.00005271, 0.00002968, 0.00001820, -0.00004922, 0.00001938, 0.00003882, -0.00004051, -0.00002501, 0.00001045, -0.00003703, -0.00004946, -0.00001030, -0.00000325, -0.00000132, },
		{ -0.00000644, 0.00000052, -0.00002276, -0.00005695, -0.00001983, 0.00000782, -0.00003096, -0.00000781, 0.00003718, -0.00001936, -0.00002124, 0.00003857, -0.00003291, -0.00007553, -0.00001178, -0.00001203, 0.00002097, 0.00002105, -0.00007418, 0.00001894, 0.00002069, -0.00018523, -0.00008676, 0.00001461, -0.00000567, -0.00002107, -0.00051807, -0.00007074, 0.00093229, -0.00180425, -0.00402544, 0.00221250, 0.00780548, 0.00221250, -0.00402544, -0.00180425, 0.00093229, -0.00007074, -0.00051807, -0.00002107, -0.00000567, 0.00001461, -0.00008676, -0.00018523, 0.00002069, 0.00001894, -0.00007418, 0.00002105, 0.00002097, -0.00001203, -0.00001178, -0.00007553, -0.00003291, 0.00003857, -0.00002124, -0.00001936, 0.00003718, -0.00000781, -0.00003096, 0.00000782, -0.00001983, -0.00005695, -0.00002276, 0.00000052, },
		{ -0.00000171, -0.00000132, -0.00000325, -0.00001030, -0.00004946, -0.00003703, 0.00001045, -0.00002501, -0.00004051, 0.00003882, 0.00001938, -0.00004922, 0.00001820, 0.00002968, -0.00005271, -0.00004350, -0.00004843, -0.00001231, 0.00008313, -0.00004488, -0.00008862, 0.00007761, -0.00003231, -0.00013712, -0.00011329, -0.00010596, 0.00026598, -0.00013270, -0.00097489, 0.00060008, 0.00097034, -0.00338845, -0.00314822, 0.00486752, 0.00676633, -0.00025783, -0.00334733, -0.00074219, 0.00033024, -0.00016974, -0.00015467, -0.00016560, -0.00004251, 0.00008259, -0.00013748, -0.00013027, 0.00000763, -0.00003061, -0.00000186, 0.00001014, -0.00002107, 0.00002281, -0.00002447, -0.00007535, -0.00000195, 0.00000327, -0.00002592, 0.00001936, 0.00000974, -0.00002946, 0.00000335, 0.00000346, -0.00005232, -0.00004470, },
		{ -0.00005779, -0.00001991, 0.00000200, -0.00000179, -0.00000187, -0.00003796, -0.00005489, 0.00000516, 0.00000332, -0.00006210, -0.00000334, 0.00005625, -0.00002702, -0.00002795, 0.00003645, -0.00000946, -0.00001894, -0.00004251, -0.00009849, 0.00005066, 0.00007520, -0.00011111, -0.00002866, 0.00003801, -0.00002272, -0.00003513, -0.00033373, -0.00002266, 0.00058971, -0.00067837, -0.00099225, 0.00147385, -0.00003970, -0.00435821, -0.00079294, 0.00616287, 0.00437414, -0.00135949, -0.00209051, -0.00054688, -0.00011703, 0.00003821, -0.00009755, -0.00024914, 0.00002142, 0.00003373, -0.00012527, -0.00007749, -0.00005181, -0.00000614, 0.00003313, -0.00003518, -0.00000816, 0.00002780, -0.00004275, -0.00003849, -0.00000788, -0.00002938, 0.00000309, 0.00002228, -0.00002356, -0.00001298, 0.00001931, -0.00002558, },
		{ 0.00000283, -0.00004638, -0.00003714, -0.00000525, -0.00000333, 0.00000321, -0.00001639, -0.00006348, -0.00002479, 0.00002927, -0.00003791, -0.00005393, 0.00004162, 0.00001648, -0.00003829, 0.00001683, -0.00000319, -0.00001173, 0.00002976, -0.00009578, -0.00008324, 0.00009571, -0.00000401, -0.00007132, -0.00002379, -0.00005810, 0.00015995, -0.00009981, -0.00059701, 0.00039602, 0.00054876, -0.00133600, -0.00030718, 0.00172213, -0.00164233, -0.00375325, 0.00176211, 0.00550898, 0.00212952, -0.00112777, -0.00123360, -0.00076135, -0.00016886, 0.00015094, -0.00017166, -0.00018927, 0.00002038, -0.00002236, -0.00006338, -0.00008817, -0.00006860, 0.00003851, 0.00000605, -0.00004502, 0.00001212, 0.00000192, -0.00002609, -0.00001268, -0.00003863, -0.00002658, 0.00002250, -0.00000307, -0.00002673, 0.00001072, },
		{ -0.00000596, 0.00001162, -0.00002588, -0.00004085, -0.00001140, -0.00000633, -0.00000479, 0.00000177, -0.00004492, -0.00005775, 0.00001661, 0.00000540, -0.00006146, 0.00000075, 0.00003150, -0.00003327, 0.00001082, 0.00001472, -0.00005315, 0.00004201, 0.00001987, -0.00012645, -0.00001512, 0.00003043, -0.00001990, 0.00002782, -0.00018153, -0.00000860, 0.00038644, -0.00044329, -0.00056462, 0.00087555, -0.00002250, -0.00149561, 0.00061973, 0.00089905, -0.00266145, -0.00177323, 0.00308489, 0.00369449, 0.00101201, -0.00044879, -0.00098547, -0.00084647, -0.00006901, 0.00007563, -0.00016988, -0.00011186, -0.00004096, -0.00001611, -0.00003618, -0.00010982, -0.00003098, 0.00003607, -0.00002518, -0.00002147, 0.00000137, -0.00000964, 0.00000603, -0.00002097, -0.00005271, -0.00000750, 0.00001121, -0.00001985, },
		{ -0.00001079, -0.00001220, 0.00000983, -0.00001040, -0.00004040, -0.00001643, 0.00000138, -0.00001214, -0.00000282, -0.00001856, -0.00006081, -0.00001703, 0.00002181, -0.00004105, -0.00002298, 0.00003747, -0.00003586, -0.00002258, 0.00005986, -0.00004395, -0.00004084, 0.00007568, -0.00003219, -0.00005945, -0.00002342, -0.00008366, 0.00012530, -0.00000888, -0.00038250, 0.00028860, 0.00036434, -0.00083388, -0.00011937, 0.00094172, -0.00072815, -0.00093381, 0.00095068, -0.00046137, -0.00227011, 0.00032691, 0.00278416, 0.00203594, 0.00088590, -0.00005896, -0.00095686, -0.00070204, -0.00007550, -0.00002640, -0.00009123, -0.00010393, -0.00007365, 0.00000663, -0.00005338, -0.00008573, -0.00000191, -0.00000053, -0.00002145, -0.00001024, -0.00001904, 0.00000723, 0.00001539, -0.00003791, -0.00004052, -0.00000417, },
		{ -0.00002587, -0.00001809, -0.00001890, 0.00000468, 0.00000475, -0.00003458, -0.00003135, 0.00000666, 0.00000059, -0.00001191, -0.00001237, -0.00004758, -0.00003712, 0.00001921, -0.00002448, -0.00004546, 0.00004902, -0.00000033, -0.00008407, 0.00004172, 0.00003927, -0.00007772, 0.00001312, 0.00002290, -0.00000986, 0.00003497, -0.00018390, -0.00005769, 0.00030928, -0.00027350, -0.00036159, 0.00062883, -0.00000288, -0.00088595, 0.00040364, 0.00046453, -0.00096008, -0.00015678, 0.00034968, -0.00126580, -0.00079253, 0.00136278, 0.00164634, 0.00123247, 0.00102494, -0.00002480, -0.00080772, -0.00053930, -0.00018204, -0.00004320, -0.00004025, -0.00012316, -0.00006016, -0.00000635, -0.00005829, -0.00003871, -0.00001341, -0.00002037, -0.00000634, -0.00002182, -0.00001683, 0.00002316, 0.00000118, -0.00003704, },
		{ -0.00001124, -0.00002426, -0.00002388, -0.00002970, -0.00001268, 0.00001089, -0.00001425, -0.00004184, -0.00000947, 0.00001477, -0.00000103, -0.00001112, -0.00004152, -0.00005270, 0.00001460, 0.00000414, -0.00007284, 0.00001784, 0.00007230, -0.00007842, -0.00005298, 0.00007000, -0.00001988, -0.00002564, -0.00000114, -0.00004733, 0.00014151, -0.00001392, -0.00036504, 0.00018225, 0.00028500, -0.00057427, -0.00003665, 0.00067905, -0.00043016, -0.00051322, 0.00053574, -0.00013181, -0.00057626, 0.00009042, -0.00060613, -0.00091676, 0.00064004, 0.00117703, 0.00071277, 0.00106438, 0.00096048, -0.00003952, -0.00055573, -0.00049233, -0.00025430, -0.00002045, -0.00004602, -0.00011136, -0.00005045, -0.00003421, -0.00002992, -0.00001842, -0.00003389, -0.00001473, -0.00000957, -0.00002785, 0.00000081, 0.00001734, },
		{ 0.00000819, 0.00000158, -0.00001245, -0.00001596, -0.00002605, -0.00002994, -0.00000637, 0.00000125, -0.00002985, -0.00002850, 0.00000648, 0.00001168, 0.00000392, -0.00002602, -0.00007052, -0.00001510, 0.00004258, -0.00005476, -0.00005691, 0.00008759, 0.00001285, -0.00010011, 0.00000864, 0.00001053, -0.00000849, 0.00004226, -0.00012777, 0.00001326, 0.00029082, -0.00025264, -0.00034007, 0.00044647, -0.00001282, -0.00061321, 0.00033725, 0.00037181, -0.00053734, -0.00007014, 0.00020850, -0.00032130, -0.00008253, -0.00036312, -0.00100475, 0.00014888, 0.00117285, 0.00050409, 0.00038403, 0.00098868, 0.00074407, 0.00006493, -0.00036930, -0.00049720, -0.00025067, -0.00003054, -0.00005815, -0.00008100, -0.00006368, -0.00003672, -0.00000410, -0.00002552, -0.00003312, -0.00001179, -0.00002105, -0.00001526, },
		{ -0.00001930, -0.00000137, 0.00000174, -0.00000901, -0.00001223, -0.00001204, -0.00002274, -0.00002533, -0.00000937, -0.00001617, -0.00002860, -0.00000706, 0.00000705, 0.00001069, 0.00000498, -0.00005723, -0.00006503, 0.00003592, 0.00001052, -0.00008791, 0.00001919, 0.00007682, -0.00005532, -0.00003163, -0.00000854, -0.00005891, 0.00010909, -0.00000850, -0.00024508, 0.00022962, 0.00024247, -0.00049181, -0.00006789, 0.00046603, -0.00034512, -0.00033941, 0.00043297, 0.00000844, -0.00028687, 0.00004417, -0.00016986, -0.00007344, -0.00005331, -0.00093399, -0.00052683, 0.00098986, 0.00087402, 0.00008314, 0.00038749, 0.00077085, 0.00059729, 0.00020134, -0.00029248, -0.00045845, -0.00023013, -0.00006799, -0.00004968, -0.00006686, -0.00007032, -0.00002120, -0.00000545, -0.00002827, -0.00002290, -0.00002054, },
		{ -0.00002183, -0.00002336, -0.00000852, 0.00000085, -0.00000605, -0.00001704, -0.00001330, -0.00000511, -0.00001679, -0.00002295, -0.00002208, -0.00002802, -0.00001019, 0.00000533, -0.00000016, 0.00001444, -0.00001231, -0.00007619, -0.00002013, 0.00004322, -0.00004753, -0.00004002, 0.00006575, -0.00001397, -0.00002818, 0.00003519, -0.00012660, -0.00001220, 0.00021606, -0.00019299, -0.00018840, 0.00042443, -0.00001617, -0.00049546, 0.00022118, 0.00021060, -0.00042372, -0.00002324, 0.00022555, -0.00008968, -0.00001283, -0.00015960, -0.00021415, 0.00014598, -0.00047525, -0.00094751, 0.00031225, 0.00109313, 0.00038735, 0.00005970, 0.00035527, 0.00052944, 0.00056518, 0.00025381, -0.00024438, -0.00037818, -0.00023179, -0.00009436, -0.00004171, -0.00006377, -0.00005528, -0.00001787, -0.00001627, -0.00001964, },
		{ -0.00001840, -0.00001667, -0.00002210, -0.00001395, -0.00000396, -0.00000174, -0.00001370, -0.00002557, -0.00000829, 0.00000173, -0.00001268, -0.00002509, -0.00003866, -0.00002405, 0.00001168, 0.00000361, -0.00000320, 0.00000885, -0.00003880, -0.00005280, 0.00001887, -0.00001078, -0.00005094, 0.00005086, 0.00000487, -0.00007650, 0.00009306, -0.00001646, -0.00022684, 0.00015638, 0.00016769, -0.00035241, 0.00003686, 0.00040252, -0.00026615, -0.00025880, 0.00027528, -0.00006842, -0.00023853, 0.00006072, -0.00000319, 0.00005512, 0.00000094, -0.00032425, -0.00001017, -0.00003033, -0.00079889, -0.00035033, 0.00079336, 0.00069597, 0.00017425, 0.00015563, 0.00022061, 0.00040212, 0.00054097, 0.00023353, -0.00017032, -0.00030571, -0.00023784, -0.00010580, -0.00004683, -0.00005033, -0.00003916, -0.00002721, },
		{ -0.00003485, -0.00003033, -0.00002134, -0.00002087, -0.00001135, -0.00000561, -0.00000438, -0.00000535, -0.00002602, -0.00002502, 0.00000044, 0.00000187, -0.00000904, -0.00003277, -0.00004833, -0.00000638, 0.00002214, -0.00000194, -0.00000499, -0.00001516, -0.00004091, 0.00000123, -0.00000148, -0.00006885, 0.00003494, 0.00007793, -0.00012630, -0.00001237, 0.00018349, -0.00017726, -0.00018000, 0.00029971, -0.00002412, -0.00032696, 0.00023523, 0.00018353, -0.00029624, -0.00001749, 0.00010362, -0.00013440, -0.00004123, -0.00006253, -0.00002799, 0.00015924, -0.00015643, -0.00022815, 0.00004884, -0.00043076, -0.00059876, 0.00032226, 0.00069857, 0.00031798, 0.00020613, 0.00015535, 0.00011427, 0.00036109, 0.00046121, 0.00021029, -0.00009060, -0.00025427, -0.00022974, -0.00011722, -0.00005038, -0.00003239, },
		{ -0.00004772, -0.00002623, -0.00002859, -0.00002863, -0.00003382, -0.00001715, 0.00000177, -0.00000298, -0.00000311, -0.00001796, -0.00003331, -0.00001149, 0.00000037, -0.00000102, -0.00000637, -0.00004273, -0.00004268, 0.00001045, 0.00002068, -0.00000277, -0.00002013, -0.00003944, 0.00000150, 0.00003203, -0.00007738, -0.00005188, 0.00014437, -0.00001759, -0.00020001, 0.00014503, 0.00013181, -0.00030562, -0.00000564, 0.00028095, -0.00018999, -0.00014688, 0.00023969, -0.00001842, -0.00013688, 0.00002976, -0.00006366, -0.00002164, -0.00002190, -0.00015730, 0.00007179, 0.00006936, -0.00020263, -0.00004649, -0.00020299, -0.00058456, -0.00002522, 0.00058022, 0.00034983, 0.00019471, 0.00024627, 0.00008954, 0.00010027, 0.00031485, 0.00036037, 0.00020605, -0.00003411, -0.00020870, -0.00021384, -0.00012471, },
		{ -0.00019261, -0.00012043, -0.00005348, -0.00002435, -0.00001413, -0.00003664, -0.00004192, -0.00000613, 0.00000611, 0.00000345, -0.00000887, -0.00003619, -0.00002116, -0.00000035, -0.00001030, -0.00000024, -0.00000881, -0.00004482, -0.00002645, 0.00001477, 0.00001819, -0.00000577, -0.00004989, -0.00003262, 0.00006323, -0.00000121, -0.00014052, 0.00005145, 0.00017044, -0.00017177, -0.00013402, 0.00025772, -0.00002915, -0.00028446, 0.00014418, 0.00013104, -0.00019694, 0.00000571, 0.00009617, -0.00004579, 0.00000365, -0.00006957, -0.00006866, 0.00006165, -0.00010988, -0.00010546, 0.00008120, -0.00007021, -0.00006618, -0.00007302, -0.00049699, -0.00029452, 0.00043520, 0.00042271, 0.00012102, 0.00020883, 0.00019836, 0.00006315, 0.00012177, 0.00024111, 0.00028295, 0.00020037, 0.00000043, -0.00016549, },
		{ 0.00001991, -0.00012661, -0.00015824, -0.00011690, -0.00006747, -0.00001216, -0.00000836, -0.00004777, -0.00003539, -0.00000387, 0.00001028, 0.00001067, -0.00003010, -0.00003831, 0.00000233, -0.00000584, -0.00001902, 0.00000079, -0.00001360, -0.00003240, -0.00001414, 0.00000931, 0.00001898, -0.00002084, -0.00007269, 0.00001156, 0.00009475, -0.00007837, -0.00012685, 0.00016426, 0.00007757, -0.00025880, 0.00001873, 0.00023196, -0.00015965, -0.00014845, 0.00015790, 0.00000194, -0.00009135, 0.00001349, -0.00001567, 0.00004196, 0.00000110, -0.00013032, 0.00001635, 0.00001809, -0.00013053, -0.00002361, -0.00003758, -0.00006675, 0.00003839, -0.00030786, -0.00048176, 0.00017090, 0.00050642, 0.00017444, 0.00008576, 0.00019295, 0.00013072, 0.00008737, 0.00011996, 0.00017396, 0.00023025, 0.00018224, },
		{ 0.00018877, 0.00015387, 0.00002491, -0.00008249, -0.00011953, -0.00012440, -0.00006901, -0.00000131, -0.00001785, -0.00004219, -0.00002913, -0.00000980, 0.00002260, 0.00000054, -0.00004913, -0.00001542, 0.00001230, -0.00001952, -0.00001953, -0.00000456, -0.00001011, -0.00001234, -0.00001395, 0.00000187, 0.00001739, -0.00004113, -0.00006967, 0.00006396, 0.00007486, -0.00013855, -0.00004361, 0.00019833, -0.00006191, -0.00021929, 0.00013131, 0.00011100, -0.00016802, -0.00002942, 0.00007556, -0.00001578, -0.00001706, -0.00006055, -0.00000640, 0.00008530, -0.00006817, -0.00007826, 0.00003980, -0.00005391, -0.00005833, -0.00004987, -0.00012402, 0.00003837, -0.00006600, -0.00045795, -0.00014447, 0.00042936, 0.00031806, 0.00006462, 0.00010780, 0.00012910, 0.00010551, 0.00010858, 0.00009807, 0.00013290, },
		{ 0.00007951, 0.00011245, 0.00015361, 0.00011388, 0.00002701, -0.00003242, -0.00009738, -0.00012446, -0.00005596, -0.00001225, -0.00002090, -0.00002403, -0.00003628, -0.00000639, 0.00002637, -0.00002561, -0.00003973, 0.00000932, 0.00000165, -0.00002390, -0.00002041, -0.00001217, 0.00000620, 0.00000345, -0.00002416, -0.00000678, 0.00001521, -0.00005165, -0.00004936, 0.00009493, 0.00002854, -0.00014408, 0.00004288, 0.00013962, -0.00014590, -0.00010890, 0.00013588, 0.00001001, -0.00009277, -0.00000255, 0.00001138, 0.00002519, -0.00003373, -0.00008834, 0.00004718, 0.00003501, -0.00007512, -0.00000602, -0.00001731, -0.00004577, -0.00001607, -0.00014057, -0.00007946, 0.00004514, -0.00026592, -0.00031274, 0.00021291, 0.00038082, 0.00013839, 0.00007165, 0.00009356, 0.00008059, 0.00010523, 0.00010161, },
		{ 0.00009478, 0.00007915, 0.00007062, 0.00010648, 0.00011753, 0.00006939, 0.00004045, -0.00000004, -0.00008640, -0.00009782, -0.00005524, -0.00003413, -0.00000328, -0.00001701, -0.00004171, 0.00000411, 0.00000345, -0.00003631, -0.00000767, 0.00000990, -0.00001029, -0.00001800, -0.00002865, -0.00001236, 0.00002939, 0.00000968, -0.00004521, -0.00001287, 0.00002527, -0.00005654, -0.00003148, 0.00010573, -0.00000985, -0.00011053, 0.00007802, 0.00004867, -0.00013953, -0.00002382, 0.00007309, -0.00001157, -0.00002891, -0.00002804, 0.00000992, 0.00003090, -0.00007313, -0.00004416, 0.00004203, -0.00002126, -0.00001368, -0.00000548, -0.00005495, 0.00000547, -0.00006649, -0.00015828, -0.00001059, -0.00010779, -0.00030695, 0.00001612, 0.00033957, 0.00020513, 0.00007583, 0.00008610, 0.00006319, 0.00006833, },
		{ 0.00005435, 0.00006759, 0.00006996, 0.00005333, 0.00007152, 0.00010607, 0.00007536, 0.00004283, 0.00005160, 0.00000100, -0.00005600, -0.00006337, -0.00007817, -0.00003728, 0.00001462, -0.00002218, -0.00002910, -0.00000582, -0.00002764, -0.00001164, 0.00001226, -0.00000794, -0.00000718, -0.00001378, -0.00003963, -0.00000280, 0.00005450, 0.00000057, -0.00007758, 0.00000354, 0.00004449, -0.00007635, -0.00001292, 0.00011219, -0.00003401, -0.00007460, 0.00006271, -0.00001049, -0.00008643, -0.00000406, 0.00002056, 0.00001310, -0.00001446, -0.00003646, 0.00002192, -0.00000851, -0.00006154, -0.00000578, -0.00001368, -0.00001103, 0.00002643, -0.00004449, -0.00002668, 0.00000051, -0.00012471, -0.00008852, -0.00006226, -0.00025245, -0.00011386, 0.00026136, 0.00025710, 0.00008484, 0.00007179, 0.00007394, },
		{ 0.00006204, 0.00006178, 0.00006252, 0.00006274, 0.00003886, 0.00003028, 0.00008074, 0.00009475, 0.00004456, 0.00003938, 0.00003142, -0.00000194, -0.00000175, -0.00005427, -0.00009603, -0.00002222, 0.00001132, -0.00001413, -0.00001721, -0.00004327, -0.00002851, 0.00002453, 0.00000214, -0.00001908, 0.00000602, -0.00001604, -0.00004874, 0.00001659, 0.00007762, -0.00003736, -0.00009960, 0.00005671, 0.00003908, -0.00010864, 0.00002689, 0.00010954, -0.00005542, -0.00005200, 0.00003098, -0.00002337, -0.00004341, -0.00002071, 0.00000766, 0.00002948, -0.00001732, -0.00002004, 0.00000120, -0.00003285, -0.00001761, -0.00002326, -0.00004166, 0.00002992, 0.00000770, -0.00004510, 0.00000200, -0.00006052, -0.00010283, -0.00005553, -0.00020184, -0.00021426, 0.00014281, 0.00030060, 0.00013405, 0.00004596, },
		{ 0.00006772, 0.00003407, 0.00004605, 0.00006036, 0.00007187, 0.00005429, 0.00000591, 0.00001917, 0.00008179, 0.00007371, 0.00004781, 0.00002788, -0.00001364, 0.00002318, 0.00003837, -0.00006381, -0.00008377, -0.00002004, 0.00000312, 0.00001143, -0.00003600, -0.00007225, 0.00000485, 0.00003531, -0.00001989, -0.00001398, 0.00001964, -0.00002352, -0.00005557, 0.00005225, 0.00007900, -0.00010203, -0.00007035, 0.00011651, -0.00001677, -0.00011305, 0.00008253, 0.00007738, -0.00007162, -0.00003704, 0.00000696, -0.00001305, -0.00002836, -0.00003384, 0.00001233, 0.00002709, -0.00000726, -0.00001572, -0.00003166, -0.00001471, -0.00000349, -0.00005720, -0.00001812, 0.00003457, -0.00001647, -0.00001001, -0.00002657, -0.00008900, -0.00004625, -0.00012979, -0.00026220, -0.00002382, 0.00027233, 0.00021794, },
		{ 0.00025224, 0.00013508, 0.00004484, 0.00002553, 0.00003657, 0.00005981, 0.00008015, 0.00003975, -0.00001673, 0.00001580, 0.00005835, 0.00007151, 0.00006985, -0.00001523, -0.00003876, 0.00005905, 0.00003941, -0.00005315, -0.00006154, -0.00003870, 0.00001674, 0.00002071, -0.00007379, -0.00005962, 0.00003490, 0.00001804, -0.00003222, 0.00000025, 0.00003262, -0.00004159, -0.00004853, 0.00009901, 0.00003033, -0.00014841, 0.00001934, 0.00012085, -0.00009084, -0.00005957, 0.00010979, 0.00001965, -0.00006680, -0.00002723, -0.00000582, -0.00000149, -0.00002803, -0.00003382, 0.00001263, 0.00002534, 0.00000408, -0.00003346, -0.00003565, 0.00000770, -0.00002756, -0.00005419, 0.00000664, 0.00000640, -0.00000324, -0.00000302, -0.00007178, -0.00005452, -0.00005746, -0.00022151, -0.00016261, 0.00014818, },
		{ 0.00000513, 0.00019961, 0.00017657, 0.00008647, 0.00003700, 0.00002596, 0.00002698, 0.00005994, 0.00008219, 0.00002304, -0.00001749, -0.00000373, 0.00002776, 0.00010087, 0.00006738, -0.00006003, -0.00002514, 0.00005994, 0.00003012, -0.00001397, -0.00006131, -0.00004702, 0.00003802, -0.00000645, -0.00008519, -0.00002259, 0.00003536, -0.00000326, -0.00003518, 0.00002312, 0.00003773, -0.00007017, -0.00000779, 0.00011925, -0.00005913, -0.00011824, 0.00010970, 0.00004868, -0.00011290, 0.00001858, 0.00007577, -0.00002506, -0.00004183, -0.00002759, -0.00000892, 0.00000439, -0.00002969, -0.00002950, 0.00001058, 0.00002834, -0.00000229, -0.00004388, -0.00001540, -0.00000185, -0.00004410, -0.00002459, -0.00000237, -0.00000505, 0.00001563, -0.00003737, -0.00006884, -0.00002983, -0.00013605, -0.00020570, },
		{ -0.00017726, -0.00008378, 0.00011233, 0.00016518, 0.00010976, 0.00005654, 0.00004419, 0.00002753, 0.00001620, 0.00006196, 0.00006939, 0.00002340, -0.00001196, -0.00004616, 0.00002507, 0.00012529, 0.00003424, -0.00005894, -0.00001289, 0.00002255, 0.00004844, 0.00001716, -0.00007185, -0.00002890, 0.00002783, -0.00003727, -0.00005574, -0.00000186, 0.00002101, -0.00001997, -0.00003128, 0.00005683, 0.00001663, -0.00008258, 0.00005693, 0.00007309, -0.00012155, -0.00001484, 0.00012133, -0.00003822, -0.00005843, 0.00005025, 0.00000756, -0.00002693, -0.00002096, -0.00003391, -0.00000324, 0.00000315, -0.00003012, -0.00002400, 0.00001138, 0.00002232, -0.00001555, -0.00002964, -0.00000745, -0.00002624, -0.00002987, -0.00001166, -0.00001929, 0.00000783, -0.00000169, -0.00005849, -0.00003529, -0.00007184, },
		{ -0.00003893, -0.00013087, -0.00011925, 0.00003906, 0.00013276, 0.00011200, 0.00005499, 0.00004623, 0.00005975, 0.00002028, 0.00001517, 0.00005161, 0.00005403, 0.00004589, -0.00002808, -0.00007729, 0.00004506, 0.00010438, 0.00002006, -0.00001981, -0.00004223, -0.00000637, 0.00008096, 0.00001954, -0.00005751, -0.00001746, -0.00001220, -0.00003005, -0.00002263, -0.00000252, 0.00001156, -0.00004127, -0.00000695, 0.00008121, -0.00002825, -0.00004785, 0.00009120, -0.00001848, -0.00009226, 0.00007765, 0.00004928, -0.00005739, 0.00001224, 0.00000966, -0.00003401, -0.00000254, -0.00002031, -0.00003157, 0.00000353, -0.00000487, -0.00002663, -0.00001615, 0.00000781, 0.00000596, -0.00001363, -0.00001079, -0.00002163, -0.00003115, -0.00001126, -0.00002050, -0.00001409, 0.00000610, -0.00003469, -0.00003881, },
		{ -0.00003873, -0.00002236, -0.00008544, -0.00012406, -0.00001619, 0.00009608, 0.00011515, 0.00005739, 0.00001603, 0.00005866, 0.00006178, 0.00001728, 0.00001768, 0.00002304, 0.00005996, 0.00006134, -0.00005716, -0.00006778, 0.00003744, 0.00005971, 0.00005938, 0.00000014, -0.00009033, -0.00000192, 0.00008645, 0.00002325, -0.00002377, -0.00004393, -0.00003279, 0.00000220, -0.00001644, 0.00000341, -0.00000365, -0.00005814, 0.00004056, 0.00006617, -0.00005550, 0.00001638, 0.00005256, -0.00007212, 0.00000608, 0.00008046, -0.00002305, -0.00000495, 0.00002210, -0.00005181, -0.00002582, 0.00000923, -0.00002609, -0.00001678, 0.00000044, -0.00001266, -0.00001719, -0.00001163, -0.00000529, -0.00000425, 0.00000107, -0.00000949, -0.00003266, -0.00001846, -0.00001247, -0.00002170, -0.00000384, -0.00002240, },
		{ -0.00001664, -0.00004208, -0.00002226, -0.00004833, -0.00010667, -0.00005216, 0.00005182, 0.00010831, 0.00008457, 0.00000277, 0.00001135, 0.00006862, 0.00005081, 0.00003023, 0.00000216, 0.00000088, 0.00007983, 0.00004678, -0.00005587, -0.00004672, -0.00001943, 0.00005199, 0.00011436, -0.00002175, -0.00010581, 0.00000244, 0.00006654, 0.00005553, -0.00001678, -0.00007919, -0.00000979, 0.00001101, -0.00001608, 0.00002091, -0.00003798, -0.00004274, 0.00007724, 0.00001985, -0.00003033, 0.00004554, -0.00002272, -0.00003228, 0.00007650, 0.00001329, -0.00002292, 0.00004032, -0.00003033, -0.00007091, -0.00000280, -0.00000127, -0.00001717, -0.00000471, -0.00001153, -0.00001148, -0.00000794, -0.00001682, -0.00001815, -0.00000049, 0.00000815, -0.00001723, -0.00002539, -0.00001395, -0.00001953, -0.00000915, },
		{ -0.00001361, -0.00000867, -0.00003887, -0.00003321, -0.00003100, -0.00008211, -0.00006417, 0.00001447, 0.00007705, 0.00010593, 0.00003444, -0.00002404, 0.00002984, 0.00005361, 0.00005205, 0.00003959, -0.00002716, 0.00000639, 0.00007654, 0.00002955, -0.00001389, -0.00006770, -0.00007528, 0.00009164, 0.00012542, -0.00004089, -0.00008809, -0.00002544, 0.00006421, 0.00008569, -0.00004754, -0.00006918, 0.00001518, -0.00001186, 0.00001072, 0.00001978, -0.00006741, 0.00000400, 0.00006750, -0.00000664, 0.00001907, 0.00000683, -0.00004771, 0.00005568, 0.00005936, -0.00003307, 0.00002445, 0.00001832, -0.00007683, -0.00004681, -0.00000312, -0.00000881, 0.00000253, -0.00000781, -0.00001862, -0.00000365, -0.00000815, -0.00002631, -0.00001973, 0.00000490, 0.00000262, -0.00001438, -0.00001516, -0.00002112, },
		{ -0.00001610, -0.00002104, -0.00000746, -0.00002783, -0.00003888, -0.00002752, -0.00006617, -0.00006730, -0.00000260, 0.00004212, 0.00009331, 0.00007685, -0.00001494, -0.00000558, 0.00003151, 0.00003223, 0.00006819, 0.00002343, -0.00003657, 0.00001662, 0.00004255, 0.00005209, 0.00001708, -0.00012475, -0.00007324, 0.00011851, 0.00009741, -0.00001689, -0.00008422, -0.00005856, 0.00008770, 0.00006721, -0.00006212, -0.00001837, -0.00000466, -0.00002162, 0.00004842, -0.00001084, -0.00005578, 0.00003753, 0.00002858, 0.00001518, 0.00002804, -0.00004762, 0.00000994, 0.00009704, -0.00000019, -0.00001429, 0.00003781, -0.00004066, -0.00007147, -0.00002694, -0.00001934, 0.00000316, 0.00001138, -0.00001714, -0.00001364, -0.00000124, -0.00001658, -0.00002875, -0.00001461, 0.00000081, -0.00000069, -0.00000321, },
		{ 0.00000508, 0.00000041, -0.00001730, -0.00001175, -0.00002093, -0.00003914, -0.00002464, -0.00005213, -0.00007512, -0.00001446, 0.00002477, 0.00006191, 0.00009567, 0.00001671, -0.00001995, 0.00002426, 0.00000609, 0.00003347, 0.00006796, 0.00000080, -0.00001660, -0.00000520, 0.00002213, 0.00009946, -0.00000669, -0.00014937, -0.00003359, 0.00008800, 0.00008820, 0.00002330, -0.00010769, -0.00005364, 0.00008634, 0.00001805, -0.00001634, 0.00000980, -0.00004943, 0.00000790, 0.00005445, -0.00003347, -0.00001266, 0.00002622, 0.00001339, 0.00004627, -0.00001608, -0.00004068, 0.00008554, 0.00006368, -0.00002186, 0.00001837, -0.00000866, -0.00006686, -0.00004062, -0.00003457, -0.00002100, 0.00001837, 0.00000371, -0.00001632, -0.00000648, -0.00000916, -0.00002226, -0.00002330, -0.00001427, -0.00000575, },
		{ -0.00001810, -0.00000450, 0.00000826, -0.00000276, -0.00000661, -0.00001415, -0.00004142, -0.00002816, -0.00003238, -0.00007550, -0.00003477, 0.00001618, 0.00003311, 0.00009086, 0.00005264, -0.00002801, 0.00001941, 0.00002192, -0.00001094, 0.00004421, 0.00003863, 0.00000773, -0.00000503, -0.00004619, 0.00004666, 0.00010819, -0.00004977, -0.00010527, -0.00002664, 0.00003495, 0.00011733, 0.00003135, -0.00011305, -0.00002834, 0.00002787, 0.00000732, 0.00005376, -0.00001704, -0.00006140, 0.00004158, 0.00002801, -0.00001356, 0.00001300, -0.00000227, 0.00003550, 0.00003269, -0.00004996, 0.00002941, 0.00010200, 0.00001688, 0.00000114, 0.00000095, -0.00005860, -0.00004550, -0.00003093, -0.00004481, -0.00000510, 0.00001604, -0.00000337, -0.00000608, -0.00000869, -0.00001841, -0.00002028, -0.00001941, },
		},
		{ { 0.00000139, 0.00000257, 0.00000236, -0.00000445, -0.00001437, -0.00001684, -0.00000940, -0.00000536, -0.00001769, -0.00003897, -0.00004960, -0.00004462, -0.00003647, -0.00002926, -0.00001116, 0.00001833, 0.00003630, 0.00002949, 0.00001776, 0.00001904, 0.00001211, -0.00002678, -0.00006520, -0.00004740, 0.00001134, 0.00002755, -0.00002543, -0.00005979, 0.00000058, 0.00008640, 0.00006437, -0.00006249, -0.00013756, -0.00006249, 0.00006437, 0.00008640, 0.00000058, -0.00005979, -0.00002543, 0.00002755, 0.00001134, -0.00004740, -0.00006520, -0.00002678, 0.00001211, 0.00001904, 0.00001776, 0.00002949, 0.00003630, 0.00001833, -0.00001116, -0.00002926, -0.00003647, -0.00004462, -0.00004960, -0.00003897, -0.00001769, -0.00000536, -0.00000940, -0.00001684, -0.00001437, -0.00000445, 0.00000236, 0.00000257, },
		{ 0.00000153, 0.00000711, 0.00000510, 0.00000279, 0.00000126, -0.00000523, -0.00001578, -0.00001764, -0.00000355, 0.00001213, 0.00000898, -0.00001101, -0.00002760, -0.00003426, -0.00004430, -0.00005987, -0.00006042, -0.00003636, -0.00000930, 0.00000303, 0.00001742, 0.00004904, 0.00006719, 0.00003437, -0.00002154, -0.00003356, -0.00000098, -0.00000059, -0.00006049, -0.00009162, -0.00001315, 0.00010192, 0.00010318, -0.00002314, -0.00011933, -0.00006922, 0.00004419, 0.00006384, -0.00002216, -0.00008193, -0.00003802, 0.00003682, 0.00004322, -0.00000946, -0.00004291, -0.00003063, -0.00001392, -0.00001985, -0.00002584, -0.00000948, 0.00001412, 0.00002164, 0.00001579, 0.00001099, 0.00000614, -0.00000779, -0.00002577, -0.00003349, -0.00002920, -0.00002549, -0.00002986, -0.00003473, -0.00002918, -0.00001348, },
		{ -0.00004479, -0.00003390, -0.00001564, -0.00000340, -0.00000058, 0.00000070, 0.00000405, 0.00000126, -0.00001138, -0.00001955, -0.00000893, 0.00000975, 0.00001314, 0.00000050, -0.00000681, -0.00000328, -0.00000939, -0.00003377, -0.00005401, -0.00005371, -0.00005014, -0.00005742, -0.00004981, -0.00000294, 0.00004934, 0.00005376, 0.00002491, 0.00002788, 0.00006326, 0.00004675, -0.00005310, -0.00012917, -0.00006954, 0.00006533, 0.00010342, -0.00000147, -0.00010047, -0.00005927, 0.00005567, 0.00008089, -0.00001295, -0.00009619, -0.00007305, 0.00000175, 0.00002728, -0.00000347, -0.00002417, -0.00000770, 0.00000869, -0.00000387, -0.00002611, -0.00003107, -0.00002255, -0.00001675, -0.00001314, -0.00000279, 0.00001034, 0.00001445, 0.00000840, 0.00000135, -0.00000277, -0.00000982, -0.00002424, -0.00003989, },
		{ 0.00000220, -0.00001610, -0.00003100, -0.00003143, -0.00002041, -0.00001260, -0.00001447, -0.00001577, -0.00000862, -0.00000317, -0.00000997, -0.00001711, -0.00000648, 0.00001316, 0.00001541, -0.00000116, -0.00000921, 0.00000185, 0.00000825, -0.00000548, -0.00001929, -0.00001759, -0.00002333, -0.00005539, -0.00008060, -0.00005887, -0.00001670, -0.00001176, -0.00002909, 0.00000779, 0.00009999, 0.00013329, 0.00003710, -0.00008523, -0.00008074, 0.00003262, 0.00008291, -0.00000878, -0.00011294, -0.00008022, 0.00004555, 0.00009750, 0.00002075, -0.00007091, -0.00007136, -0.00001487, 0.00000508, -0.00002528, -0.00004471, -0.00002185, 0.00001005, 0.00001556, 0.00000177, -0.00000912, -0.00001498, -0.00002422, -0.00003335, -0.00003209, -0.00002044, -0.00000758, 0.00000121, 0.00000761, 0.00001271, 0.00001251, },
		{ -0.00000333, 0.00000321, 0.00000398, -0.00000486, -0.00001614, -0.00001605, -0.00000380, 0.00000346, -0.00000510, -0.00001699, -0.00001740, -0.00001452, -0.00002278, -0.00003115, -0.00001705, 0.00001166, 0.00002171, 0.00000593, -0.00000522, 0.00000450, 0.00001098, -0.00000302, -0.00001298, 0.00000128, 0.00001003, -0.00001780, -0.00004974, -0.00003643, -0.00000645, -0.00003392, -0.00010417, -0.00010382, 0.00001040, 0.00011812, 0.00009189, -0.00001293, -0.00003389, 0.00005552, 0.00010414, 0.00001271, -0.00011162, -0.00010870, 0.00000673, 0.00007652, 0.00002538, -0.00005471, -0.00005781, -0.00000526, 0.00001379, -0.00002206, -0.00005563, -0.00004653, -0.00001595, 0.00000147, 0.00000279, 0.00000192, 0.00000073, -0.00000723, -0.00002010, -0.00002901, -0.00002962, -0.00002446, -0.00001753, -0.00001057, },
		{ -0.00001406, -0.00001116, -0.00001096, -0.00000936, -0.00000835, -0.00001533, -0.00002668, -0.00002523, -0.00000547, 0.00001264, 0.00001084, -0.00000016, -0.00000139, -0.00000045, -0.00001637, -0.00003798, -0.00003413, -0.00000645, 0.00000805, -0.00000344, -0.00000890, 0.00001030, 0.00002412, 0.00000844, -0.00000729, 0.00000700, 0.00001897, -0.00001297, -0.00005000, -0.00001941, 0.00005238, 0.00004989, -0.00005686, -0.00014189, -0.00009506, 0.00001276, 0.00003537, -0.00002607, -0.00003185, 0.00006659, 0.00013803, 0.00006810, -0.00006226, -0.00009068, -0.00000334, 0.00006028, 0.00001562, -0.00006264, -0.00006933, -0.00001391, 0.00001772, -0.00000772, -0.00004422, -0.00004782, -0.00002724, -0.00001133, -0.00000755, -0.00000686, -0.00000566, -0.00000749, -0.00001342, -0.00001986, -0.00002249, -0.00001955, },
		{ -0.00002704, -0.00002268, -0.00001218, -0.00000543, -0.00000552, -0.00000496, -0.00000293, -0.00001088, -0.00002975, -0.00003850, -0.00002345, -0.00000328, -0.00000133, -0.00000783, 0.00000058, 0.00001524, 0.00000766, -0.00001667, -0.00002114, -0.00000091, 0.00000475, -0.00001951, -0.00003400, -0.00001136, 0.00001350, 0.00000580, -0.00000199, 0.00003029, 0.00005961, 0.00001485, -0.00006881, -0.00007198, 0.00002894, 0.00010422, 0.00005134, -0.00005399, -0.00007631, -0.00002502, -0.00002128, -0.00008622, -0.00009937, 0.00000151, 0.00010317, 0.00007796, -0.00002511, -0.00005479, 0.00002140, 0.00008254, 0.00003949, -0.00004767, -0.00007338, -0.00003113, 0.00000311, -0.00001184, -0.00004355, -0.00005013, -0.00003188, -0.00001289, -0.00000560, -0.00000659, -0.00000985, -0.00001313, -0.00001733, -0.00002326, },
		{ -0.00001367, -0.00001932, -0.00002760, -0.00002754, -0.00001770, -0.00001095, -0.00001271, -0.00001115, -0.00000133, 0.00000035, -0.00001561, -0.00002860, -0.00001872, -0.00000337, -0.00001047, -0.00002883, -0.00002519, -0.00000348, 0.00000049, -0.00001496, -0.00001009, 0.00002286, 0.00003569, 0.00000353, -0.00002744, -0.00001695, -0.00000290, -0.00003084, -0.00005529, -0.00000200, 0.00008994, 0.00009580, -0.00000720, -0.00008639, -0.00004017, 0.00005585, 0.00007167, 0.00001679, -0.00000067, 0.00003285, 0.00001815, -0.00007342, -0.00012978, -0.00006892, 0.00002425, 0.00002174, -0.00005459, -0.00007116, 0.00001382, 0.00009367, 0.00007336, -0.00000562, -0.00004324, -0.00001868, 0.00000703, -0.00000860, -0.00004381, -0.00005940, -0.00004724, -0.00002522, -0.00001054, -0.00000775, -0.00001103, -0.00001288, },
		{ -0.00002509, -0.00002748, -0.00002172, -0.00001957, -0.00002373, -0.00002117, -0.00000852, -0.00000313, -0.00001391, -0.00002286, -0.00001580, -0.00000925, -0.00001985, -0.00002837, -0.00001016, 0.00001547, 0.00001019, -0.00001859, -0.00002831, -0.00001489, -0.00001815, -0.00004334, -0.00004123, 0.00000526, 0.00003881, 0.00001935, -0.00000282, 0.00002607, 0.00005444, -0.00000118, -0.00009944, -0.00010530, 0.00000866, 0.00010487, 0.00007422, -0.00001316, -0.00002952, 0.00002003, 0.00003657, 0.00000659, 0.00001234, 0.00006867, 0.00007973, -0.00000130, -0.00008059, -0.00006384, -0.00000454, -0.00001790, -0.00009716, -0.00012264, -0.00004355, 0.00005001, 0.00006039, 0.00000809, -0.00002081, 0.00000147, 0.00002779, 0.00001701, -0.00001914, -0.00004585, -0.00004673, -0.00003065, -0.00001664, -0.00001665, },
		{ 0.00000526, 0.00000068, -0.00002185, -0.00003901, -0.00004101, -0.00004101, -0.00004439, -0.00003495, -0.00000759, 0.00001214, 0.00000498, -0.00001029, -0.00001081, -0.00001073, -0.00003084, -0.00004914, -0.00003020, 0.00000894, 0.00001801, -0.00000466, -0.00001122, 0.00000752, 0.00000262, -0.00003953, -0.00005803, -0.00001939, 0.00001200, -0.00002182, -0.00005666, 0.00000371, 0.00011311, 0.00012067, -0.00000674, -0.00012010, -0.00009457, 0.00000561, 0.00004622, 0.00001951, 0.00001449, 0.00004704, 0.00004718, 0.00000426, -0.00000224, 0.00005529, 0.00009063, 0.00004010, -0.00002924, -0.00002587, 0.00001781, -0.00000101, -0.00008752, -0.00013586, -0.00008612, -0.00000573, 0.00001528, -0.00001801, -0.00003654, -0.00001154, 0.00002192, 0.00002574, 0.00000153, -0.00002364, -0.00002801, -0.00001145, },
		{ -0.00003590, -0.00000950, 0.00002613, 0.00003639, 0.00001527, -0.00001041, -0.00002450, -0.00004054, -0.00006385, -0.00006875, -0.00003995, -0.00000614, 0.00000151, -0.00000271, 0.00000754, 0.00001525, -0.00001030, -0.00004675, -0.00004355, -0.00000787, 0.00000247, -0.00002184, -0.00002471, 0.00001079, 0.00002257, -0.00001924, -0.00004237, 0.00000933, 0.00005650, -0.00000741, -0.00012615, -0.00012988, 0.00001746, 0.00014906, 0.00012218, 0.00000205, -0.00005973, -0.00003814, -0.00001841, -0.00002734, -0.00000955, 0.00004297, 0.00006220, 0.00002769, 0.00001180, 0.00005612, 0.00009317, 0.00005563, -0.00000895, -0.00001369, 0.00002627, 0.00001967, -0.00005257, -0.00010624, -0.00008154, -0.00002432, -0.00000884, -0.00003690, -0.00005516, -0.00003707, -0.00000633, 0.00000499, -0.00000968, -0.00003216, },
		{ -0.00001735, -0.00004770, -0.00006653, -0.00004609, 0.00000081, 0.00003107, 0.00002804, 0.00001802, 0.00001806, 0.00000625, -0.00003106, -0.00006161, -0.00005459, -0.00003276, -0.00003166, -0.00003589, -0.00001148, 0.00002089, 0.00001242, -0.00002423, -0.00002760, 0.00000595, 0.00001110, -0.00002920, -0.00004493, -0.00000367, 0.00001665, -0.00004073, -0.00008642, -0.00000832, 0.00012521, 0.00012752, -0.00003548, -0.00017596, -0.00013605, 0.00001053, 0.00009016, 0.00006531, 0.00002852, 0.00002457, 0.00000954, -0.00002899, -0.00003606, 0.00000460, 0.00003187, 0.00001426, 0.00000842, 0.00005764, 0.00010593, 0.00008257, 0.00001757, -0.00000616, 0.00001689, 0.00001571, -0.00003557, -0.00007770, -0.00006226, -0.00002126, -0.00001105, -0.00003425, -0.00005166, -0.00004186, -0.00001841, -0.00000548, },
		{ -0.00002408, -0.00000346, -0.00000132, -0.00002965, -0.00006397, -0.00006332, -0.00002822, -0.00000364, -0.00001013, -0.00001417, 0.00000889, 0.00002808, 0.00001153, -0.00001455, -0.00001240, -0.00000278, -0.00002607, -0.00006024, -0.00005111, -0.00001096, -0.00000396, -0.00003209, -0.00002726, 0.00002213, 0.00003631, -0.00001622, -0.00004479, 0.00001439, 0.00006382, -0.00001713, -0.00015332, -0.00014583, 0.00003580, 0.00018493, 0.00013373, -0.00003130, -0.00011581, -0.00007685, -0.00002052, -0.00000749, 0.00000411, 0.00003823, 0.00005090, 0.00002347, 0.00000459, 0.00002081, 0.00002778, -0.00000265, -0.00001908, 0.00002648, 0.00008889, 0.00009191, 0.00004187, 0.00000779, 0.00001054, 0.00000519, -0.00002986, -0.00005613, -0.00004266, -0.00001485, -0.00001109, -0.00003012, -0.00004521, -0.00004212, },
		{ -0.00002982, -0.00004133, -0.00004207, -0.00002406, -0.00000551, -0.00001559, -0.00004239, -0.00004021, -0.00000407, 0.00001252, -0.00001917, -0.00005392, -0.00004554, -0.00001913, -0.00001723, -0.00001972, 0.00001318, 0.00005181, 0.00003466, -0.00002181, -0.00004208, -0.00001891, -0.00002214, -0.00006260, -0.00005830, 0.00001438, 0.00004879, -0.00002322, -0.00008850, -0.00001018, 0.00013291, 0.00012689, -0.00006078, -0.00020976, -0.00014433, 0.00003783, 0.00012634, 0.00007758, 0.00001148, -0.00000092, -0.00000527, -0.00003603, -0.00005193, -0.00002506, 0.00000614, 0.00001228, 0.00002255, 0.00005544, 0.00006901, 0.00003321, -0.00000419, 0.00001519, 0.00006696, 0.00008064, 0.00004310, 0.00000304, -0.00001263, -0.00002391, -0.00004132, -0.00004306, -0.00002026, 0.00000128, 0.00000021, -0.00001491, },
		{ 0.00000600, 0.00000932, 0.00001163, -0.00000517, -0.00003205, -0.00003849, -0.00002921, -0.00004048, -0.00006674, -0.00005261, 0.00001389, 0.00006253, 0.00003889, -0.00001583, -0.00003840, -0.00004229, -0.00006980, -0.00009145, -0.00004763, 0.00003265, 0.00005896, 0.00002315, 0.00000690, 0.00002960, 0.00001057, -0.00006867, -0.00009197, 0.00000573, 0.00009031, 0.00000856, -0.00015124, -0.00015230, 0.00004448, 0.00020347, 0.00013588, -0.00005448, -0.00014344, -0.00008547, -0.00001036, 0.00000467, 0.00000965, 0.00004480, 0.00006601, 0.00003705, -0.00000536, -0.00002102, -0.00002517, -0.00003515, -0.00002367, 0.00002524, 0.00006730, 0.00005999, 0.00003368, 0.00004483, 0.00008637, 0.00010052, 0.00006354, 0.00000550, -0.00004377, -0.00007793, -0.00008848, -0.00006482, -0.00002378, 0.00000154, },
		{ -0.00005257, -0.00003396, -0.00003347, -0.00000966, 0.00002995, 0.00003915, 0.00001452, -0.00000105, -0.00000255, -0.00003511, -0.00009766, -0.00011316, -0.00004251, 0.00004197, 0.00005950, 0.00003471, 0.00003159, 0.00003051, -0.00002689, -0.00010581, -0.00011054, -0.00004245, -0.00000192, -0.00001595, 0.00000287, 0.00007687, 0.00008513, -0.00003460, -0.00013043, -0.00003730, 0.00014260, 0.00014793, -0.00006380, -0.00023543, -0.00016073, 0.00004929, 0.00015149, 0.00009328, 0.00001214, -0.00000699, -0.00001416, -0.00005081, -0.00006919, -0.00002946, 0.00002699, 0.00004880, 0.00004531, 0.00004194, 0.00002710, -0.00000844, -0.00002998, -0.00000825, 0.00002713, 0.00003357, 0.00002572, 0.00004923, 0.00010346, 0.00013835, 0.00011836, 0.00005080, -0.00003734, -0.00011232, -0.00013669, -0.00010222, },
		{ -0.00015851, -0.00011244, -0.00005402, -0.00005346, -0.00008335, -0.00006955, -0.00001187, 0.00002236, 0.00001620, 0.00002294, 0.00005274, 0.00003744, -0.00004671, -0.00011209, -0.00008498, -0.00002068, -0.00000681, -0.00001843, 0.00002122, 0.00008321, 0.00006834, -0.00001635, -0.00005840, -0.00003180, -0.00003794, -0.00010416, -0.00009951, 0.00004159, 0.00015126, 0.00005056, -0.00014697, -0.00015688, 0.00006572, 0.00024228, 0.00015213, -0.00008088, -0.00019044, -0.00011778, -0.00001379, 0.00002226, 0.00003679, 0.00007334, 0.00008520, 0.00003480, -0.00002936, -0.00004829, -0.00003211, -0.00001655, 0.00000045, 0.00003043, 0.00005074, 0.00003950, 0.00001982, 0.00002305, 0.00003535, 0.00002604, 0.00000786, 0.00002341, 0.00008130, 0.00014140, 0.00015620, 0.00010260, -0.00000761, -0.00011889, },
		{ 0.00000916, -0.00013908, -0.00019393, -0.00012446, -0.00003378, -0.00003033, -0.00008575, -0.00009446, -0.00004262, -0.00001516, -0.00003844, -0.00003531, 0.00002810, 0.00006687, 0.00001846, -0.00004229, -0.00002629, 0.00001529, -0.00001817, -0.00009132, -0.00008329, 0.00000386, 0.00004619, 0.00001478, 0.00002024, 0.00008924, 0.00007608, -0.00008736, -0.00021050, -0.00009389, 0.00013742, 0.00016950, -0.00005181, -0.00022980, -0.00012950, 0.00011431, 0.00021640, 0.00011856, -0.00000893, -0.00005124, -0.00005569, -0.00007514, -0.00006977, -0.00000589, 0.00006385, 0.00007862, 0.00005335, 0.00003258, 0.00001830, -0.00000607, -0.00002526, -0.00001656, 0.00000631, 0.00001799, 0.00002547, 0.00004687, 0.00006787, 0.00006055, 0.00003375, 0.00002862, 0.00006829, 0.00013306, 0.00017367, 0.00013683, },
		{ 0.00020757, 0.00018810, 0.00004034, -0.00015914, -0.00025027, -0.00017020, -0.00004338, -0.00002170, -0.00008394, -0.00009366, -0.00002380, 0.00001109, -0.00004224, -0.00008580, -0.00004402, 0.00000588, -0.00002146, -0.00005640, 0.00000656, 0.00010566, 0.00009216, -0.00002446, -0.00008940, -0.00006004, -0.00005770, -0.00011543, -0.00007960, 0.00011397, 0.00024700, 0.00010104, -0.00017737, -0.00023288, 0.00000216, 0.00020696, 0.00012205, -0.00011448, -0.00020180, -0.00008229, 0.00005730, 0.00009269, 0.00007917, 0.00008304, 0.00007034, 0.00000638, -0.00005755, -0.00006193, -0.00002570, 0.00000176, 0.00001800, 0.00004391, 0.00006695, 0.00006087, 0.00003411, 0.00001420, 0.00000484, -0.00000404, -0.00000311, 0.00002214, 0.00005693, 0.00007074, 0.00005972, 0.00005255, 0.00007966, 0.00014497, },
		{ 0.00007884, 0.00014738, 0.00023760, 0.00024864, 0.00009468, -0.00015247, -0.00028772, -0.00020721, -0.00005236, -0.00002515, -0.00011464, -0.00013986, -0.00004069, 0.00004079, 0.00000403, -0.00005610, -0.00003425, -0.00000343, -0.00007064, -0.00015515, -0.00009851, 0.00006147, 0.00013150, 0.00006856, 0.00002806, 0.00006437, 0.00001574, -0.00018625, -0.00030407, -0.00010880, 0.00022274, 0.00029035, 0.00002032, -0.00022790, -0.00015947, 0.00008456, 0.00018043, 0.00006306, -0.00006952, -0.00008619, -0.00005064, -0.00004025, -0.00002248, 0.00004210, 0.00010444, 0.00010428, 0.00006164, 0.00002906, 0.00001050, -0.00001559, -0.00003617, -0.00002242, 0.00001725, 0.00004898, 0.00006093, 0.00006285, 0.00005550, 0.00003513, 0.00001640, 0.00001957, 0.00004108, 0.00005872, 0.00006075, 0.00005814, },
		{ 0.00007501, 0.00007296, 0.00008302, 0.00014667, 0.00025677, 0.00029616, 0.00013645, -0.00016110, -0.00034158, -0.00024690, -0.00003449, 0.00002204, -0.00010143, -0.00017972, -0.00009550, 0.00000568, -0.00001510, -0.00006430, -0.00000714, 0.00007186, 0.00000660, -0.00014476, -0.00016911, -0.00004712, 0.00001890, -0.00003582, -0.00001544, 0.00017270, 0.00027509, 0.00004564, -0.00031521, -0.00036843, -0.00004312, 0.00025406, 0.00019269, -0.00007366, -0.00018688, -0.00006960, 0.00006650, 0.00007913, 0.00003760, 0.00002910, 0.00002165, -0.00002909, -0.00007563, -0.00005676, 0.00000507, 0.00005146, 0.00007443, 0.00009580, 0.00010494, 0.00007634, 0.00002397, -0.00001112, -0.00001381, 0.00000062, 0.00002075, 0.00004488, 0.00006336, 0.00006488, 0.00005547, 0.00005119, 0.00005776, 0.00006879, },
		{ 0.00007093, 0.00007906, 0.00009125, 0.00009480, 0.00010155, 0.00016765, 0.00030600, 0.00038130, 0.00021051, -0.00016760, -0.00043902, -0.00035975, -0.00007758, 0.00005512, -0.00005160, -0.00015235, -0.00007837, 0.00001952, -0.00003853, -0.00014500, -0.00009496, 0.00005107, 0.00006306, -0.00006090, -0.00008874, 0.00002814, 0.00004000, -0.00016547, -0.00029335, -0.00005687, 0.00032545, 0.00036853, -0.00000187, -0.00033343, -0.00026283, 0.00003766, 0.00017169, 0.00005271, -0.00008741, -0.00009185, -0.00003722, -0.00002324, -0.00001877, 0.00002742, 0.00007179, 0.00005379, -0.00000117, -0.00003035, -0.00002597, -0.00001619, 0.00000132, 0.00004583, 0.00009982, 0.00012225, 0.00010376, 0.00007051, 0.00004326, 0.00002436, 0.00001612, 0.00002322, 0.00004015, 0.00005534, 0.00006413, 0.00006809, },
		{ 0.00007035, 0.00008263, 0.00008501, 0.00008886, 0.00009963, 0.00009785, 0.00009009, 0.00015487, 0.00033820, 0.00049116, 0.00035637, -0.00008477, -0.00047690, -0.00047192, -0.00017385, 0.00000499, -0.00009010, -0.00018938, -0.00006533, 0.00010931, 0.00006566, -0.00011506, -0.00015030, -0.00003419, -0.00002409, -0.00015507, -0.00013491, 0.00014332, 0.00031762, 0.00005669, -0.00038056, -0.00043625, -0.00002399, 0.00034887, 0.00027497, -0.00005745, -0.00021187, -0.00009212, 0.00004942, 0.00004400, -0.00002020, -0.00002992, -0.00001913, -0.00005065, -0.00008640, -0.00006419, -0.00000791, 0.00002012, 0.00001585, 0.00001452, 0.00001849, 0.00000727, -0.00000760, 0.00000688, 0.00005136, 0.00009309, 0.00011123, 0.00011159, 0.00010281, 0.00008599, 0.00006547, 0.00005031, 0.00004620, 0.00005429, },
		{ 0.00007973, 0.00006729, 0.00007701, 0.00009457, 0.00010600, 0.00012253, 0.00014781, 0.00014366, 0.00009965, 0.00011860, 0.00030408, 0.00052530, 0.00045975, 0.00000944, -0.00045702, -0.00049762, -0.00017410, 0.00002704, -0.00012806, -0.00032274, -0.00020940, 0.00006390, 0.00012642, -0.00001752, -0.00004713, 0.00008180, 0.00003744, -0.00028263, -0.00045236, -0.00011299, 0.00040919, 0.00047048, -0.00001210, -0.00044722, -0.00036807, 0.00001415, 0.00020671, 0.00009418, -0.00005149, -0.00004721, 0.00001070, 0.00000091, -0.00003354, -0.00001555, 0.00001928, 0.00000155, -0.00005004, -0.00007334, -0.00006462, -0.00006080, -0.00006329, -0.00004625, -0.00001517, -0.00000013, -0.00000291, 0.00000445, 0.00003411, 0.00007016, 0.00009789, 0.00011790, 0.00013238, 0.00013721, 0.00012813, 0.00010585, },
		{ 0.00020178, 0.00020247, 0.00017123, 0.00013179, 0.00010771, 0.00009440, 0.00008617, 0.00011014, 0.00017771, 0.00022393, 0.00019250, 0.00017909, 0.00033669, 0.00057074, 0.00051543, 0.00000507, -0.00056263, -0.00062141, -0.00017594, 0.00017169, 0.00002627, -0.00029905, -0.00031091, -0.00006938, -0.00000466, -0.00016016, -0.00012189, 0.00023220, 0.00040784, 0.00001144, -0.00056186, -0.00058716, -0.00000485, 0.00049162, 0.00038150, -0.00007538, -0.00030882, -0.00018367, -0.00000583, 0.00001334, -0.00002820, -0.00000217, 0.00004128, 0.00001608, -0.00003747, -0.00003777, 0.00000159, 0.00001511, -0.00000384, -0.00001721, -0.00002241, -0.00004623, -0.00008330, -0.00010037, -0.00009075, -0.00007820, -0.00007260, -0.00005866, -0.00002686, 0.00001329, 0.00005158, 0.00008903, 0.00012985, 0.00017169, },
		{ 0.00000828, 0.00008653, 0.00018241, 0.00026130, 0.00029553, 0.00029258, 0.00026882, 0.00021440, 0.00013073, 0.00008066, 0.00012150, 0.00019712, 0.00021467, 0.00024417, 0.00046031, 0.00077751, 0.00075193, 0.00011487, -0.00070189, -0.00093405, -0.00042944, 0.00013342, 0.00014701, -0.00015638, -0.00017772, 0.00008295, 0.00007027, -0.00035246, -0.00056878, -0.00011432, 0.00053823, 0.00055967, -0.00009323, -0.00062214, -0.00045945, 0.00007227, 0.00032412, 0.00015917, -0.00006173, -0.00009260, -0.00004184, -0.00005716, -0.00008946, -0.00004741, 0.00002315, 0.00003068, -0.00001152, -0.00003049, -0.00001599, -0.00000926, -0.00001609, -0.00000880, 0.00000938, 0.00000591, -0.00002527, -0.00005780, -0.00007759, -0.00009549, -0.00011676, -0.00012887, -0.00012210, -0.00010145, -0.00007517, -0.00004225, },
		{ -0.00015795, -0.00018034, -0.00018862, -0.00015036, -0.00005142, 0.00008145, 0.00021212, 0.00033334, 0.00043720, 0.00047191, 0.00039856, 0.00028067, 0.00022022, 0.00020166, 0.00014111, 0.00012897, 0.00040465, 0.00091291, 0.00109474, 0.00047854, -0.00056610, -0.00108651, -0.00069757, -0.00007843, -0.00002463, -0.00035632, -0.00028159, 0.00031454, 0.00060942, 0.00005305, -0.00074027, -0.00077367, -0.00001478, 0.00059934, 0.00042776, -0.00014897, -0.00040305, -0.00020068, 0.00004238, 0.00005857, -0.00001589, -0.00000646, 0.00003331, -0.00000319, -0.00007146, -0.00007318, -0.00002228, 0.00000180, -0.00001214, -0.00001749, -0.00000588, -0.00000872, -0.00002765, -0.00003153, -0.00001362, -0.00000002, -0.00000454, -0.00001474, -0.00002263, -0.00003744, -0.00006495, -0.00009619, -0.00012072, -0.00013877, },
		{ -0.00004615, -0.00006006, -0.00008744, -0.00014531, -0.00022720, -0.00029181, -0.00030037, -0.00025381, -0.00016329, -0.00000701, 0.00022423, 0.00045390, 0.00057972, 0.00060917, 0.00062271, 0.00058939, 0.00039982, 0.00017137, 0.00027803, 0.00080883, 0.00117233, 0.00068257, -0.00042909, -0.00107535, -0.00065228, 0.00010161, 0.00006674, -0.00069251, -0.00098829, -0.00018883, 0.00083946, 0.00086910, -0.00007472, -0.00082845, -0.00063754, 0.00003817, 0.00035359, 0.00015672, -0.00008717, -0.00009002, -0.00001200, -0.00004136, -0.00010398, -0.00007064, 0.00001094, 0.00002616, -0.00001715, -0.00003563, -0.00001717, -0.00001198, -0.00002821, -0.00002867, -0.00000772, 0.00000156, -0.00001201, -0.00002478, -0.00002356, -0.00002131, -0.00002652, -0.00002970, -0.00002434, -0.00001862, -0.00002189, -0.00003320, },
		{ -0.00003997, -0.00003897, -0.00003397, -0.00001815, -0.00000262, -0.00001920, -0.00008511, -0.00017540, -0.00025821, -0.00033878, -0.00042011, -0.00043866, -0.00031799, -0.00008589, 0.00016013, 0.00041169, 0.00072862, 0.00101443, 0.00102874, 0.00079039, 0.00073502, 0.00111790, 0.00139027, 0.00071196, -0.00076335, -0.00167329, -0.00100563, 0.00046088, 0.00094654, -0.00012069, -0.00137612, -0.00127499, 0.00000456, 0.00095660, 0.00070354, -0.00013869, -0.00053931, -0.00031879, -0.00002081, 0.00001726, -0.00003540, 0.00001817, 0.00008684, 0.00003191, -0.00007936, -0.00010745, -0.00005715, -0.00002475, -0.00003074, -0.00002465, 0.00000109, 0.00000528, -0.00001857, -0.00003261, -0.00002096, -0.00000720, -0.00000729, -0.00001017, -0.00000780, -0.00000982, -0.00002264, -0.00003733, -0.00004377, -0.00004250, },
		{ -0.00002932, -0.00003766, -0.00004380, -0.00005382, -0.00006320, -0.00005487, -0.00002610, -0.00000542, -0.00001666, -0.00004156, -0.00006050, -0.00011270, -0.00024774, -0.00042205, -0.00053227, -0.00055711, -0.00054409, -0.00042811, -0.00004792, 0.00054578, 0.00101776, 0.00122354, 0.00153999, 0.00226614, 0.00275356, 0.00181170, -0.00062701, -0.00276982, -0.00264402, -0.00044887, 0.00148311, 0.00129897, -0.00038450, -0.00150289, -0.00105782, 0.00006270, 0.00057543, 0.00030255, -0.00007207, -0.00013259, -0.00006730, -0.00010434, -0.00015212, -0.00007024, 0.00005586, 0.00007291, -0.00000201, -0.00005027, -0.00004433, -0.00004274, -0.00005983, -0.00005433, -0.00001993, 0.00000172, -0.00000749, -0.00002141, -0.00002044, -0.00001525, -0.00001599, -0.00001487, -0.00000625, 0.00000081, -0.00000320, -0.00001610, },
		{ -0.00000298, -0.00000710, -0.00001571, -0.00002020, -0.00002013, -0.00002716, -0.00004483, -0.00005510, -0.00004769, -0.00004752, -0.00007437, -0.00009086, -0.00005080, 0.00000733, -0.00000545, -0.00009737, -0.00020432, -0.00034555, -0.00061187, -0.00091978, -0.00097378, -0.00065021, -0.00020768, 0.00020536, 0.00105217, 0.00279248, 0.00462887, 0.00461422, 0.00175256, -0.00226428, -0.00428421, -0.00291687, 0.00003053, 0.00166291, 0.00103839, -0.00039694, -0.00097941, -0.00057159, -0.00006658, 0.00003146, -0.00003704, 0.00001133, 0.00008100, 0.00001320, -0.00010247, -0.00010556, -0.00002347, 0.00001373, -0.00001370, -0.00002945, -0.00001271, -0.00001176, -0.00003978, -0.00005536, -0.00003940, -0.00002007, -0.00001865, -0.00002271, -0.00001913, -0.00001512, -0.00001810, -0.00002119, -0.00001654, -0.00000755, },
		{ -0.00002246, -0.00001654, -0.00000982, -0.00001066, -0.00001949, -0.00002437, -0.00001861, -0.00001471, -0.00002147, -0.00002092, -0.00000205, -0.00000207, -0.00005982, -0.00013551, -0.00014933, -0.00010315, -0.00007032, -0.00005142, 0.00002378, 0.00009468, -0.00005153, -0.00046024, -0.00085755, -0.00105274, -0.00127282, -0.00162619, -0.00132146, 0.00078614, 0.00436308, 0.00688688, 0.00575820, 0.00127995, -0.00318724, -0.00446567, -0.00255499, -0.00005009, 0.00091820, 0.00045462, -0.00015092, -0.00025086, -0.00015387, -0.00020209, -0.00027419, -0.00017711, -0.00001124, 0.00003659, -0.00001302, -0.00002709, 0.00000806, 0.00000989, -0.00003287, -0.00005344, -0.00003012, -0.00000980, -0.00001937, -0.00003292, -0.00002902, -0.00002150, -0.00002474, -0.00002994, -0.00002666, -0.00002038, -0.00001964, -0.00002264, },
		{ -0.00002250, -0.00002462, -0.00002765, -0.00002491, -0.00001635, -0.00001288, -0.00002071, -0.00002779, -0.00002377, -0.00002264, -0.00003573, -0.00003583, 0.00000402, 0.00004500, 0.00002200, -0.00004984, -0.00009353, -0.00010745, -0.00016603, -0.00024086, -0.00017841, 0.00003443, 0.00015822, 0.00004562, -0.00009959, -0.00015507, -0.00059695, -0.00191919, -0.00332635, -0.00277054, 0.00083466, 0.00560226, 0.00782624, 0.00560226, 0.00083466, -0.00277054, -0.00332635, -0.00191919, -0.00059695, -0.00015507, -0.00009959, 0.00004562, 0.00015822, 0.00003443, -0.00017841, -0.00024086, -0.00016603, -0.00010745, -0.00009353, -0.00004984, 0.00002200, 0.00004500, 0.00000402, -0.00003583, -0.00003573, -0.00002264, -0.00002377, -0.00002779, -0.00002071, -0.00001288, -0.00001635, -0.00002491, -0.00002765, -0.00002462, },
		{ -0.00002246, -0.00002264, -0.00001964, -0.00002038, -0.00002666, -0.00002994, -0.00002474, -0.00002150, -0.00002902, -0.00003292, -0.00001937, -0.00000980, -0.00003012, -0.00005344, -0.00003287, 0.00000989, 0.00000806, -0.00002709, -0.00001302, 0.00003659, -0.00001124, -0.00017711, -0.00027419, -0.00020209, -0.00015387, -0.00025086, -0.00015092, 0.00045462, 0.00091820, -0.00005009, -0.00255499, -0.00446567, -0.00318724, 0.00127995, 0.00575820, 0.00688688, 0.00436308, 0.00078614, -0.00132146, -0.00162619, -0.00127282, -0.00105274, -0.00085755, -0.00046024, -0.00005153, 0.00009468, 0.00002378, -0.00005142, -0.00007032, -0.00010315, -0.00014933, -0.00013551, -0.00005982, -0.00000207, -0.00000205, -0.00002092, -0.00002147, -0.00001471, -0.00001861, -0.00002437, -0.00001949, -0.00001066, -0.00000982, -0.00001654, },
		{ -0.00000298, -0.00000755, -0.00001654, -0.00002119, -0.00001810, -0.00001512, -0.00001913, -0.00002271, -0.00001865, -0.00002007, -0.00003940, -0.00005536, -0.00003978, -0.00001176, -0.00001271, -0.00002945, -0.00001370, 0.00001373, -0.00002347, -0.00010556, -0.00010247, 0.00001320, 0.00008100, 0.00001133, -0.00003704, 0.00003146, -0.00006658, -0.00057159, -0.00097941, -0.00039694, 0.00103839, 0.00166291, 0.00003053, -0.00291687, -0.00428421, -0.00226428, 0.00175256, 0.00461422, 0.00462887, 0.00279248, 0.00105217, 0.00020536, -0.00020768, -0.00065021, -0.00097378, -0.00091978, -0.00061187, -0.00034555, -0.00020432, -0.00009737, -0.00000545, 0.00000733, -0.00005080, -0.00009086, -0.00007437, -0.00004752, -0.00004769, -0.00005510, -0.00004483, -0.00002716, -0.00002013, -0.00002020, -0.00001571, -0.00000710, },
		{ -0.00002932, -0.00001610, -0.00000320, 0.00000081, -0.00000625, -0.00001487, -0.00001599, -0.00001525, -0.00002044, -0.00002141, -0.00000749, 0.00000172, -0.00001993, -0.00005433, -0.00005983, -0.00004274, -0.00004433, -0.00005027, -0.00000201, 0.00007291, 0.00005586, -0.00007024, -0.00015212, -0.00010434, -0.00006730, -0.00013259, -0.00007207, 0.00030255, 0.00057543, 0.00006270, -0.00105782, -0.00150289, -0.00038450, 0.00129897, 0.00148311, -0.00044887, -0.00264402, -0.00276982, -0.00062701, 0.00181170, 0.00275356, 0.00226614, 0.00153999, 0.00122354, 0.00101776, 0.00054578, -0.00004792, -0.00042811, -0.00054409, -0.00055711, -0.00053227, -0.00042205, -0.00024774, -0.00011270, -0.00006050, -0.00004156, -0.00001666, -0.00000542, -0.00002610, -0.00005487, -0.00006320, -0.00005382, -0.00004380, -0.00003766, },
		{ -0.00003997, -0.00004250, -0.00004377, -0.00003733, -0.00002264, -0.00000982, -0.00000780, -0.00001017, -0.00000729, -0.00000720, -0.00002096, -0.00003261, -0.00001857, 0.00000528, 0.00000109, -0.00002465, -0.00003074, -0.00002475, -0.00005715, -0.00010745, -0.00007936, 0.00003191, 0.00008684, 0.00001817, -0.00003540, 0.00001726, -0.00002081, -0.00031879, -0.00053931, -0.00013869, 0.00070354, 0.00095660, 0.00000456, -0.00127499, -0.00137612, -0.00012069, 0.00094654, 0.00046088, -0.00100563, -0.00167329, -0.00076335, 0.00071196, 0.00139027, 0.00111790, 0.00073502, 0.00079039, 0.00102874, 0.00101443, 0.00072862, 0.00041169, 0.00016013, -0.00008589, -0.00031799, -0.00043866, -0.00042011, -0.00033878, -0.00025821, -0.00017540, -0.00008511, -0.00001920, -0.00000262, -0.00001815, -0.00003397, -0.00003897, },
		{ -0.00004615, -0.00003320, -0.00002189, -0.00001862, -0.00002434, -0.00002970, -0.00002652, -0.00002131, -0.00002356, -0.00002478, -0.00001201, 0.00000156, -0.00000772, -0.00002867, -0.00002821, -0.00001198, -0.00001717, -0.00003563, -0.00001715, 0.00002616, 0.00001094, -0.00007064, -0.00010398, -0.00004136, -0.00001200, -0.00009002, -0.00008717, 0.00015672, 0.00035359, 0.00003817, -0.00063754, -0.00082845, -0.00007472, 0.00086910, 0.00083946, -0.00018883, -0.00098829, -0.00069251, 0.00006674, 0.00010161, -0.00065228, -0.00107535, -0.00042909, 0.00068257, 0.00117233, 0.00080883, 0.00027803, 0.00017137, 0.00039982, 0.00058939, 0.00062271, 0.00060917, 0.00057972, 0.00045390, 0.00022423, -0.00000701, -0.00016329, -0.00025381, -0.00030037, -0.00029181, -0.00022720, -0.00014531, -0.00008744, -0.00006006, },
		{ -0.00015795, -0.00013877, -0.00012072, -0.00009619, -0.00006495, -0.00003744, -0.00002263, -0.00001474, -0.00000454, -0.00000002, -0.00001362, -0.00003153, -0.00002765, -0.00000872, -0.00000588, -0.00001749, -0.00001214, 0.00000180, -0.00002228, -0.00007318, -0.00007146, -0.00000319, 0.00003331, -0.00000646, -0.00001589, 0.00005857, 0.00004238, -0.00020068, -0.00040305, -0.00014897, 0.00042776, 0.00059934, -0.00001478, -0.00077367, -0.00074027, 0.00005305, 0.00060942, 0.00031454, -0.00028159, -0.00035632, -0.00002463, -0.00007843, -0.00069757, -0.00108651, -0.00056610, 0.00047854, 0.00109474, 0.00091291, 0.00040465, 0.00012897, 0.00014111, 0.00020166, 0.00022022, 0.00028067, 0.00039856, 0.00047191, 0.00043720, 0.00033334, 0.00021212, 0.00008145, -0.00005142, -0.00015036, -0.00018862, -0.00018034, },
		{ 0.00000828, -0.00004225, -0.00007517, -0.00010145, -0.00012210, -0.00012887, -0.00011676, -0.00009549, -0.00007759, -0.00005780, -0.00002527, 0.00000591, 0.00000938, -0.00000880, -0.00001609, -0.00000926, -0.00001599, -0.00003049, -0.00001152, 0.00003068, 0.00002315, -0.00004741, -0.00008946, -0.00005716, -0.00004184, -0.00009260, -0.00006173, 0.00015917, 0.00032412, 0.00007227, -0.00045945, -0.00062214, -0.00009323, 0.00055967, 0.00053823, -0.00011432, -0.00056878, -0.00035246, 0.00007027, 0.00008295, -0.00017772, -0.00015638, 0.00014701, 0.00013342, -0.00042944, -0.00093405, -0.00070189, 0.00011487, 0.00075193, 0.00077751, 0.00046031, 0.00024417, 0.00021467, 0.00019712, 0.00012150, 0.00008066, 0.00013073, 0.00021440, 0.00026882, 0.00029258, 0.00029553, 0.00026130, 0.00018241, 0.00008653, },
		{ 0.00020178, 0.00017169, 0.00012985, 0.00008903, 0.00005158, 0.00001329, -0.00002686, -0.00005866, -0.00007260, -0.00007820, -0.00009075, -0.00010037, -0.00008330, -0.00004623, -0.00002241, -0.00001721, -0.00000384, 0.00001511, 0.00000159, -0.00003777, -0.00003747, 0.00001608, 0.00004128, -0.00000217, -0.00002820, 0.00001334, -0.00000583, -0.00018367, -0.00030882, -0.00007538, 0.00038150, 0.00049162, -0.00000485, -0.00058716, -0.00056186, 0.00001144, 0.00040784, 0.00023220, -0.00012189, -0.00016016, -0.00000466, -0.00006938, -0.00031091, -0.00029905, 0.00002627, 0.00017169, -0.00017594, -0.00062141, -0.00056263, 0.00000507, 0.00051543, 0.00057074, 0.00033669, 0.00017909, 0.00019250, 0.00022393, 0.00017771, 0.00011014, 0.00008617, 0.00009440, 0.00010771, 0.00013179, 0.00017123, 0.00020247, },
		{ 0.00007973, 0.00010585, 0.00012813, 0.00013721, 0.00013238, 0.00011790, 0.00009789, 0.00007016, 0.00003411, 0.00000445, -0.00000291, -0.00000013, -0.00001517, -0.00004625, -0.00006329, -0.00006080, -0.00006462, -0.00007334, -0.00005004, 0.00000155, 0.00001928, -0.00001555, -0.00003354, 0.00000091, 0.00001070, -0.00004721, -0.00005149, 0.00009418, 0.00020671, 0.00001415, -0.00036807, -0.00044722, -0.00001210, 0.00047048, 0.00040919, -0.00011299, -0.00045236, -0.00028263, 0.00003744, 0.00008180, -0.00004713, -0.00001752, 0.00012642, 0.00006390, -0.00020940, -0.00032274, -0.00012806, 0.00002704, -0.00017410, -0.00049762, -0.00045702, 0.00000944, 0.00045975, 0.00052530, 0.00030408, 0.00011860, 0.00009965, 0.00014366, 0.00014781, 0.00012253, 0.00010600, 0.00009457, 0.00007701, 0.00006729, },
		{ 0.00007035, 0.00005429, 0.00004620, 0.00005031, 0.00006547, 0.00008599, 0.00010281, 0.00011159, 0.00011123, 0.00009309, 0.00005136, 0.00000688, -0.00000760, 0.00000727, 0.00001849, 0.00001452, 0.00001585, 0.00002012, -0.00000791, -0.00006419, -0.00008640, -0.00005065, -0.00001913, -0.00002992, -0.00002020, 0.00004400, 0.00004942, -0.00009212, -0.00021187, -0.00005745, 0.00027497, 0.00034887, -0.00002399, -0.00043625, -0.00038056, 0.00005669, 0.00031762, 0.00014332, -0.00013491, -0.00015507, -0.00002409, -0.00003419, -0.00015030, -0.00011506, 0.00006566, 0.00010931, -0.00006533, -0.00018938, -0.00009010, 0.00000499, -0.00017385, -0.00047192, -0.00047690, -0.00008477, 0.00035637, 0.00049116, 0.00033820, 0.00015487, 0.00009009, 0.00009785, 0.00009963, 0.00008886, 0.00008501, 0.00008263, },
		{ 0.00007093, 0.00006809, 0.00006413, 0.00005534, 0.00004015, 0.00002322, 0.00001612, 0.00002436, 0.00004326, 0.00007051, 0.00010376, 0.00012225, 0.00009982, 0.00004583, 0.00000132, -0.00001619, -0.00002597, -0.00003035, -0.00000117, 0.00005379, 0.00007179, 0.00002742, -0.00001877, -0.00002324, -0.00003722, -0.00009185, -0.00008741, 0.00005271, 0.00017169, 0.00003766, -0.00026283, -0.00033343, -0.00000187, 0.00036853, 0.00032545, -0.00005687, -0.00029335, -0.00016547, 0.00004000, 0.00002814, -0.00008874, -0.00006090, 0.00006306, 0.00005107, -0.00009496, -0.00014500, -0.00003853, 0.00001952, -0.00007837, -0.00015235, -0.00005160, 0.00005512, -0.00007758, -0.00035975, -0.00043902, -0.00016760, 0.00021051, 0.00038130, 0.00030600, 0.00016765, 0.00010155, 0.00009480, 0.00009125, 0.00007906, },
		{ 0.00007501, 0.00006879, 0.00005776, 0.00005119, 0.00005547, 0.00006488, 0.00006336, 0.00004488, 0.00002075, 0.00000062, -0.00001381, -0.00001112, 0.00002397, 0.00007634, 0.00010494, 0.00009580, 0.00007443, 0.00005146, 0.00000507, -0.00005676, -0.00007563, -0.00002909, 0.00002165, 0.00002910, 0.00003760, 0.00007913, 0.00006650, -0.00006960, -0.00018688, -0.00007366, 0.00019269, 0.00025406, -0.00004312, -0.00036843, -0.00031521, 0.00004564, 0.00027509, 0.00017270, -0.00001544, -0.00003582, 0.00001890, -0.00004712, -0.00016911, -0.00014476, 0.00000660, 0.00007186, -0.00000714, -0.00006430, -0.00001510, 0.00000568, -0.00009550, -0.00017972, -0.00010143, 0.00002204, -0.00003449, -0.00024690, -0.00034158, -0.00016110, 0.00013645, 0.00029616, 0.00025677, 0.00014667, 0.00008302, 0.00007296, },
		{ 0.00007884, 0.00005814, 0.00006075, 0.00005872, 0.00004108, 0.00001957, 0.00001640, 0.00003513, 0.00005550, 0.00006285, 0.00006093, 0.00004898, 0.00001725, -0.00002242, -0.00003617, -0.00001559, 0.00001050, 0.00002906, 0.00006164, 0.00010428, 0.00010444, 0.00004210, -0.00002248, -0.00004025, -0.00005064, -0.00008619, -0.00006952, 0.00006306, 0.00018043, 0.00008456, -0.00015947, -0.00022790, 0.00002032, 0.00029035, 0.00022274, -0.00010880, -0.00030407, -0.00018625, 0.00001574, 0.00006437, 0.00002806, 0.00006856, 0.00013150, 0.00006147, -0.00009851, -0.00015515, -0.00007064, -0.00000343, -0.00003425, -0.00005610, 0.00000403, 0.00004079, -0.00004069, -0.00013986, -0.00011464, -0.00002515, -0.00005236, -0.00020721, -0.00028772, -0.00015247, 0.00009468, 0.00024864, 0.00023760, 0.00014738, },
		{ 0.00020757, 0.00014497, 0.00007966, 0.00005255, 0.00005972, 0.00007074, 0.00005693, 0.00002214, -0.00000311, -0.00000404, 0.00000484, 0.00001420, 0.00003411, 0.00006087, 0.00006695, 0.00004391, 0.00001800, 0.00000176, -0.00002570, -0.00006193, -0.00005755, 0.00000638, 0.00007034, 0.00008304, 0.00007917, 0.00009269, 0.00005730, -0.00008229, -0.00020180, -0.00011448, 0.00012205, 0.00020696, 0.00000216, -0.00023288, -0.00017737, 0.00010104, 0.00024700, 0.00011397, -0.00007960, -0.00011543, -0.00005770, -0.00006004, -0.00008940, -0.00002446, 0.00009216, 0.00010566, 0.00000656, -0.00005640, -0.00002146, 0.00000588, -0.00004402, -0.00008580, -0.00004224, 0.00001109, -0.00002380, -0.00009366, -0.00008394, -0.00002170, -0.00004338, -0.00017020, -0.00025027, -0.00015914, 0.00004034, 0.00018810, },
		{ 0.00000916, 0.00013683, 0.00017367, 0.00013306, 0.00006829, 0.00002862, 0.00003375, 0.00006055, 0.00006787, 0.00004687, 0.00002547, 0.00001799, 0.00000631, -0.00001656, -0.00002526, -0.00000607, 0.00001830, 0.00003258, 0.00005335, 0.00007862, 0.00006385, -0.00000589, -0.00006977, -0.00007514, -0.00005569, -0.00005124, -0.00000893, 0.00011856, 0.00021640, 0.00011431, -0.00012950, -0.00022980, -0.00005181, 0.00016950, 0.00013742, -0.00009389, -0.00021050, -0.00008736, 0.00007608, 0.00008924, 0.00002024, 0.00001478, 0.00004619, 0.00000386, -0.00008329, -0.00009132, -0.00001817, 0.00001529, -0.00002629, -0.00004229, 0.00001846, 0.00006687, 0.00002810, -0.00003531, -0.00003844, -0.00001516, -0.00004262, -0.00009446, -0.00008575, -0.00003033, -0.00003378, -0.00012446, -0.00019393, -0.00013908, },
		{ -0.00015851, -0.00011889, -0.00000761, 0.00010260, 0.00015620, 0.00014140, 0.00008130, 0.00002341, 0.00000786, 0.00002604, 0.00003535, 0.00002305, 0.00001982, 0.00003950, 0.00005074, 0.00003043, 0.00000045, -0.00001655, -0.00003211, -0.00004829, -0.00002936, 0.00003480, 0.00008520, 0.00007334, 0.00003679, 0.00002226, -0.00001379, -0.00011778, -0.00019044, -0.00008088, 0.00015213, 0.00024228, 0.00006572, -0.00015688, -0.00014697, 0.00005056, 0.00015126, 0.00004159, -0.00009951, -0.00010416, -0.00003794, -0.00003180, -0.00005840, -0.00001635, 0.00006834, 0.00008321, 0.00002122, -0.00001843, -0.00000681, -0.00002068, -0.00008498, -0.00011209, -0.00004671, 0.00003744, 0.00005274, 0.00002294, 0.00001620, 0.00002236, -0.00001187, -0.00006955, -0.00008335, -0.00005346, -0.00005402, -0.00011244, },
		{ -0.00005257, -0.00010222, -0.00013669, -0.00011232, -0.00003734, 0.00005080, 0.00011836, 0.00013835, 0.00010346, 0.00004923, 0.00002572, 0.00003357, 0.00002713, -0.00000825, -0.00002998, -0.00000844, 0.00002710, 0.00004194, 0.00004531, 0.00004880, 0.00002699, -0.00002946, -0.00006919, -0.00005081, -0.00001416, -0.00000699, 0.00001214, 0.00009328, 0.00015149, 0.00004929, -0.00016073, -0.00023543, -0.00006380, 0.00014793, 0.00014260, -0.00003730, -0.00013043, -0.00003460, 0.00008513, 0.00007687, 0.00000287, -0.00001595, -0.00000192, -0.00004245, -0.00011054, -0.00010581, -0.00002689, 0.00003051, 0.00003159, 0.00003471, 0.00005950, 0.00004197, -0.00004251, -0.00011316, -0.00009766, -0.00003511, -0.00000255, -0.00000105, 0.00001452, 0.00003915, 0.00002995, -0.00000966, -0.00003347, -0.00003396, },
		{ 0.00000600, 0.00000154, -0.00002378, -0.00006482, -0.00008848, -0.00007793, -0.00004377, 0.00000550, 0.00006354, 0.00010052, 0.00008637, 0.00004483, 0.00003368, 0.00005999, 0.00006730, 0.00002524, -0.00002367, -0.00003515, -0.00002517, -0.00002102, -0.00000536, 0.00003705, 0.00006601, 0.00004480, 0.00000965, 0.00000467, -0.00001036, -0.00008547, -0.00014344, -0.00005448, 0.00013588, 0.00020347, 0.00004448, -0.00015230, -0.00015124, 0.00000856, 0.00009031, 0.00000573, -0.00009197, -0.00006867, 0.00001057, 0.00002960, 0.00000690, 0.00002315, 0.00005896, 0.00003265, -0.00004763, -0.00009145, -0.00006980, -0.00004229, -0.00003840, -0.00001583, 0.00003889, 0.00006253, 0.00001389, -0.00005261, -0.00006674, -0.00004048, -0.00002921, -0.00003849, -0.00003205, -0.00000517, 0.00001163, 0.00000932, },
		{ -0.00002982, -0.00001491, 0.00000021, 0.00000128, -0.00002026, -0.00004306, -0.00004132, -0.00002391, -0.00001263, 0.00000304, 0.00004310, 0.00008064, 0.00006696, 0.00001519, -0.00000419, 0.00003321, 0.00006901, 0.00005544, 0.00002255, 0.00001228, 0.00000614, -0.00002506, -0.00005193, -0.00003603, -0.00000527, -0.00000092, 0.00001148, 0.00007758, 0.00012634, 0.00003783, -0.00014433, -0.00020976, -0.00006078, 0.00012689, 0.00013291, -0.00001018, -0.00008850, -0.00002322, 0.00004879, 0.00001438, -0.00005830, -0.00006260, -0.00002214, -0.00001891, -0.00004208, -0.00002181, 0.00003466, 0.00005181, 0.00001318, -0.00001972, -0.00001723, -0.00001913, -0.00004554, -0.00005392, -0.00001917, 0.00001252, -0.00000407, -0.00004021, -0.00004239, -0.00001559, -0.00000551, -0.00002406, -0.00004207, -0.00004133, },
		{ -0.00002408, -0.00004212, -0.00004521, -0.00003012, -0.00001109, -0.00001485, -0.00004266, -0.00005613, -0.00002986, 0.00000519, 0.00001054, 0.00000779, 0.00004187, 0.00009191, 0.00008889, 0.00002648, -0.00001908, -0.00000265, 0.00002778, 0.00002081, 0.00000459, 0.00002347, 0.00005090, 0.00003823, 0.00000411, -0.00000749, -0.00002052, -0.00007685, -0.00011581, -0.00003130, 0.00013373, 0.00018493, 0.00003580, -0.00014583, -0.00015332, -0.00001713, 0.00006382, 0.00001439, -0.00004479, -0.00001622, 0.00003631, 0.00002213, -0.00002726, -0.00003209, -0.00000396, -0.00001096, -0.00005111, -0.00006024, -0.00002607, -0.00000278, -0.00001240, -0.00001455, 0.00001153, 0.00002808, 0.00000889, -0.00001417, -0.00001013, -0.00000364, -0.00002822, -0.00006332, -0.00006397, -0.00002965, -0.00000132, -0.00000346, },
		{ -0.00001735, -0.00000548, -0.00001841, -0.00004186, -0.00005166, -0.00003425, -0.00001105, -0.00002126, -0.00006226, -0.00007770, -0.00003557, 0.00001571, 0.00001689, -0.00000616, 0.00001757, 0.00008257, 0.00010593, 0.00005764, 0.00000842, 0.00001426, 0.00003187, 0.00000460, -0.00003606, -0.00002899, 0.00000954, 0.00002457, 0.00002852, 0.00006531, 0.00009016, 0.00001053, -0.00013605, -0.00017596, -0.00003548, 0.00012752, 0.00012521, -0.00000832, -0.00008642, -0.00004073, 0.00001665, -0.00000367, -0.00004493, -0.00002920, 0.00001110, 0.00000595, -0.00002760, -0.00002423, 0.00001242, 0.00002089, -0.00001148, -0.00003589, -0.00003166, -0.00003276, -0.00005459, -0.00006161, -0.00003106, 0.00000625, 0.00001806, 0.00001802, 0.00002804, 0.00003107, 0.00000081, -0.00004609, -0.00006653, -0.00004770, },
		{ -0.00003590, -0.00003216, -0.00000968, 0.00000499, -0.00000633, -0.00003707, -0.00005516, -0.00003690, -0.00000884, -0.00002432, -0.00008154, -0.00010624, -0.00005257, 0.00001967, 0.00002627, -0.00001369, -0.00000895, 0.00005563, 0.00009317, 0.00005612, 0.00001180, 0.00002769, 0.00006220, 0.00004297, -0.00000955, -0.00002734, -0.00001841, -0.00003814, -0.00005973, 0.00000205, 0.00012218, 0.00014906, 0.00001746, -0.00012988, -0.00012615, -0.00000741, 0.00005650, 0.00000933, -0.00004237, -0.00001924, 0.00002257, 0.00001079, -0.00002471, -0.00002184, 0.00000247, -0.00000787, -0.00004355, -0.00004675, -0.00001030, 0.00001525, 0.00000754, -0.00000271, 0.00000151, -0.00000614, -0.00003995, -0.00006875, -0.00006385, -0.00004054, -0.00002450, -0.00001041, 0.00001527, 0.00003639, 0.00002613, -0.00000950, },
		{ 0.00000526, -0.00001145, -0.00002801, -0.00002364, 0.00000153, 0.00002574, 0.00002192, -0.00001154, -0.00003654, -0.00001801, 0.00001528, -0.00000573, -0.00008612, -0.00013586, -0.00008752, -0.00000101, 0.00001781, -0.00002587, -0.00002924, 0.00004010, 0.00009063, 0.00005529, -0.00000224, 0.00000426, 0.00004718, 0.00004704, 0.00001449, 0.00001951, 0.00004622, 0.00000561, -0.00009457, -0.00012010, -0.00000674, 0.00012067, 0.00011311, 0.00000371, -0.00005666, -0.00002182, 0.00001200, -0.00001939, -0.00005803, -0.00003953, 0.00000262, 0.00000752, -0.00001122, -0.00000466, 0.00001801, 0.00000894, -0.00003020, -0.00004914, -0.00003084, -0.00001073, -0.00001081, -0.00001029, 0.00000498, 0.00001214, -0.00000759, -0.00003495, -0.00004439, -0.00004101, -0.00004101, -0.00003901, -0.00002185, 0.00000068, },
		{ -0.00002509, -0.00001665, -0.00001664, -0.00003065, -0.00004673, -0.00004585, -0.00001914, 0.00001701, 0.00002779, 0.00000147, -0.00002081, 0.00000809, 0.00006039, 0.00005001, -0.00004355, -0.00012264, -0.00009716, -0.00001790, -0.00000454, -0.00006384, -0.00008059, -0.00000130, 0.00007973, 0.00006867, 0.00001234, 0.00000659, 0.00003657, 0.00002003, -0.00002952, -0.00001316, 0.00007422, 0.00010487, 0.00000866, -0.00010530, -0.00009944, -0.00000118, 0.00005444, 0.00002607, -0.00000282, 0.00001935, 0.00003881, 0.00000526, -0.00004123, -0.00004334, -0.00001815, -0.00001489, -0.00002831, -0.00001859, 0.00001019, 0.00001547, -0.00001016, -0.00002837, -0.00001985, -0.00000925, -0.00001580, -0.00002286, -0.00001391, -0.00000313, -0.00000852, -0.00002117, -0.00002373, -0.00001957, -0.00002172, -0.00002748, },
		{ -0.00001367, -0.00001288, -0.00001103, -0.00000775, -0.00001054, -0.00002522, -0.00004724, -0.00005940, -0.00004381, -0.00000860, 0.00000703, -0.00001868, -0.00004324, -0.00000562, 0.00007336, 0.00009367, 0.00001382, -0.00007116, -0.00005459, 0.00002174, 0.00002425, -0.00006892, -0.00012978, -0.00007342, 0.00001815, 0.00003285, -0.00000067, 0.00001679, 0.00007167, 0.00005585, -0.00004017, -0.00008639, -0.00000720, 0.00009580, 0.00008994, -0.00000200, -0.00005529, -0.00003084, -0.00000290, -0.00001695, -0.00002744, 0.00000353, 0.00003569, 0.00002286, -0.00001009, -0.00001496, 0.00000049, -0.00000348, -0.00002519, -0.00002883, -0.00001047, -0.00000337, -0.00001872, -0.00002860, -0.00001561, 0.00000035, -0.00000133, -0.00001115, -0.00001271, -0.00001095, -0.00001770, -0.00002754, -0.00002760, -0.00001932, },
		{ -0.00002704, -0.00002326, -0.00001733, -0.00001313, -0.00000985, -0.00000659, -0.00000560, -0.00001289, -0.00003188, -0.00005013, -0.00004355, -0.00001184, 0.00000311, -0.00003113, -0.00007338, -0.00004767, 0.00003949, 0.00008254, 0.00002140, -0.00005479, -0.00002511, 0.00007796, 0.00010317, 0.00000151, -0.00009937, -0.00008622, -0.00002128, -0.00002502, -0.00007631, -0.00005399, 0.00005134, 0.00010422, 0.00002894, -0.00007198, -0.00006881, 0.00001485, 0.00005961, 0.00003029, -0.00000199, 0.00000580, 0.00001350, -0.00001136, -0.00003400, -0.00001951, 0.00000475, -0.00000091, -0.00002114, -0.00001667, 0.00000766, 0.00001524, 0.00000058, -0.00000783, -0.00000133, -0.00000328, -0.00002345, -0.00003850, -0.00002975, -0.00001088, -0.00000293, -0.00000496, -0.00000552, -0.00000543, -0.00001218, -0.00002268, },
		{ -0.00001406, -0.00001955, -0.00002249, -0.00001986, -0.00001342, -0.00000749, -0.00000566, -0.00000686, -0.00000755, -0.00001133, -0.00002724, -0.00004782, -0.00004422, -0.00000772, 0.00001772, -0.00001391, -0.00006933, -0.00006264, 0.00001562, 0.00006028, -0.00000334, -0.00009068, -0.00006226, 0.00006810, 0.00013803, 0.00006659, -0.00003185, -0.00002607, 0.00003537, 0.00001276, -0.00009506, -0.00014189, -0.00005686, 0.00004989, 0.00005238, -0.00001941, -0.00005000, -0.00001297, 0.00001897, 0.00000700, -0.00000729, 0.00000844, 0.00002412, 0.00001030, -0.00000890, -0.00000344, 0.00000805, -0.00000645, -0.00003413, -0.00003798, -0.00001637, -0.00000045, -0.00000139, -0.00000016, 0.00001084, 0.00001264, -0.00000547, -0.00002523, -0.00002668, -0.00001533, -0.00000835, -0.00000936, -0.00001096, -0.00001116, },
		{ -0.00000333, -0.00001057, -0.00001753, -0.00002446, -0.00002962, -0.00002901, -0.00002010, -0.00000723, 0.00000073, 0.00000192, 0.00000279, 0.00000147, -0.00001595, -0.00004653, -0.00005563, -0.00002206, 0.00001379, -0.00000526, -0.00005781, -0.00005471, 0.00002538, 0.00007652, 0.00000673, -0.00010870, -0.00011162, 0.00001271, 0.00010414, 0.00005552, -0.00003389, -0.00001293, 0.00009189, 0.00011812, 0.00001040, -0.00010382, -0.00010417, -0.00003392, -0.00000645, -0.00003643, -0.00004974, -0.00001780, 0.00001003, 0.00000128, -0.00001298, -0.00000302, 0.00001098, 0.00000450, -0.00000522, 0.00000593, 0.00002171, 0.00001166, -0.00001705, -0.00003115, -0.00002278, -0.00001452, -0.00001740, -0.00001699, -0.00000510, 0.00000346, -0.00000380, -0.00001605, -0.00001614, -0.00000486, 0.00000398, 0.00000321, },
		{ 0.00000220, 0.00001251, 0.00001271, 0.00000761, 0.00000121, -0.00000758, -0.00002044, -0.00003209, -0.00003335, -0.00002422, -0.00001498, -0.00000912, 0.00000177, 0.00001556, 0.00001005, -0.00002185, -0.00004471, -0.00002528, 0.00000508, -0.00001487, -0.00007136, -0.00007091, 0.00002075, 0.00009750, 0.00004555, -0.00008022, -0.00011294, -0.00000878, 0.00008291, 0.00003262, -0.00008074, -0.00008523, 0.00003710, 0.00013329, 0.00009999, 0.00000779, -0.00002909, -0.00001176, -0.00001670, -0.00005887, -0.00008060, -0.00005539, -0.00002333, -0.00001759, -0.00001929, -0.00000548, 0.00000825, 0.00000185, -0.00000921, -0.00000116, 0.00001541, 0.00001316, -0.00000648, -0.00001711, -0.00000997, -0.00000317, -0.00000862, -0.00001577, -0.00001447, -0.00001260, -0.00002041, -0.00003143, -0.00003100, -0.00001610, },
		{ -0.00004479, -0.00003989, -0.00002424, -0.00000982, -0.00000277, 0.00000135, 0.00000840, 0.00001445, 0.00001034, -0.00000279, -0.00001314, -0.00001675, -0.00002255, -0.00003107, -0.00002611, -0.00000387, 0.00000869, -0.00000770, -0.00002417, -0.00000347, 0.00002728, 0.00000175, -0.00007305, -0.00009619, -0.00001295, 0.00008089, 0.00005567, -0.00005927, -0.00010047, -0.00000147, 0.00010342, 0.00006533, -0.00006954, -0.00012917, -0.00005310, 0.00004675, 0.00006326, 0.00002788, 0.00002491, 0.00005376, 0.00004934, -0.00000294, -0.00004981, -0.00005742, -0.00005014, -0.00005371, -0.00005401, -0.00003377, -0.00000939, -0.00000328, -0.00000681, 0.00000050, 0.00001314, 0.00000975, -0.00000893, -0.00001955, -0.00001138, 0.00000126, 0.00000405, 0.00000070, -0.00000058, -0.00000340, -0.00001564, -0.00003390, },
		{ 0.00000153, -0.00001348, -0.00002918, -0.00003473, -0.00002986, -0.00002549, -0.00002920, -0.00003349, -0.00002577, -0.00000779, 0.00000614, 0.00001099, 0.00001579, 0.00002164, 0.00001412, -0.00000948, -0.00002584, -0.00001985, -0.00001392, -0.00003063, -0.00004291, -0.00000946, 0.00004322, 0.00003682, -0.00003802, -0.00008193, -0.00002216, 0.00006384, 0.00004419, -0.00006922, -0.00011933, -0.00002314, 0.00010318, 0.00010192, -0.00001315, -0.00009162, -0.00006049, -0.00000059, -0.00000098, -0.00003356, -0.00002154, 0.00003437, 0.00006719, 0.00004904, 0.00001742, 0.00000303, -0.00000930, -0.00003636, -0.00006042, -0.00005987, -0.00004430, -0.00003426, -0.00002760, -0.00001101, 0.00000898, 0.00001213, -0.00000355, -0.00001764, -0.00001578, -0.00000523, 0.00000126, 0.00000279, 0.00000510, 0.00000711, },
		},
		{ { -0.00000794, -0.00000853, -0.00000999, -0.00001147, -0.00001174, -0.00000951, -0.00000390, 0.00000504, 0.00001601, 0.00002648, 0.00003321, 0.00003321, 0.00002495, 0.00000917, -0.00001087, -0.00003011, -0.00004313, -0.00004600, -0.00003776, -0.00002094, -0.00000091, 0.00001588, 0.00002414, 0.00002167, 0.00001013, -0.00000556, -0.00001904, -0.00002497, -0.00002104, -0.00000886, 0.00000672, 0.00001950, 0.00002441, 0.00001950, 0.00000672, -0.00000886, -0.00002104, -0.00002497, -0.00001904, -0.00000556, 0.00001013, 0.00002167, 0.00002414, 0.00001588, -0.00000091, -0.00002094, -0.00003776, -0.00004600, -0.00004313, -0.00003011, -0.00001087, 0.00000917, 0.00002495, 0.00003321, 0.00003321, 0.00002648, 0.00001601, 0.00000504, -0.00000390, -0.00000951, -0.00001174, -0.00001147, -0.00000999, -0.00000853, },
		{ -0.00000720, -0.00000921, -0.00000916, -0.00000792, -0.00000651, -0.00000583, -0.00000636, -0.00000792, -0.00000951, -0.00000953, -0.00000627, 0.00000131, 0.00001290, 0.00002642, 0.00003831, 0.00004445, 0.00004157, 0.00002866, 0.00000772, -0.00001638, -0.00003721, -0.00004877, -0.00004757, -0.00003394, -0.00001214, 0.00001101, 0.00002839, 0.00003504, 0.00002981, 0.00001571, -0.00000136, -0.00001482, -0.00001992, -0.00001539, -0.00000383, 0.00000953, 0.00001896, 0.00002054, 0.00001367, 0.00000126, -0.00001158, -0.00001955, -0.00001936, -0.00001098, 0.00000236, 0.00001543, 0.00002304, 0.00002198, 0.00001219, -0.00000330, -0.00001948, -0.00003124, -0.00003502, -0.00002990, -0.00001766, -0.00000197, 0.00001296, 0.00002368, 0.00002836, 0.00002700, 0.00002107, 0.00001277, 0.00000431, -0.00000265, },
		{ 0.00002317, 0.00001927, 0.00001351, 0.00000719, 0.00000147, -0.00000275, -0.00000498, -0.00000525, -0.00000416, -0.00000271, -0.00000197, -0.00000255, -0.00000424, -0.00000576, -0.00000514, -0.00000045, 0.00000919, 0.00002287, 0.00003760, 0.00004896, 0.00005241, 0.00004495, 0.00002650, 0.00000040, -0.00002725, -0.00004938, -0.00006015, -0.00005691, -0.00004116, -0.00001805, 0.00000536, 0.00002249, 0.00002943, 0.00002612, 0.00001604, 0.00000466, -0.00000275, -0.00000319, 0.00000311, 0.00001297, 0.00002170, 0.00002517, 0.00002146, 0.00001166, -0.00000064, -0.00001083, -0.00001512, -0.00001203, -0.00000292, 0.00000858, 0.00001802, 0.00002179, 0.00001838, 0.00000888, -0.00000359, -0.00001506, -0.00002210, -0.00002284, -0.00001744, -0.00000771, 0.00000362, 0.00001382, 0.00002092, 0.00002400, },
		{ -0.00000751, -0.00000114, 0.00000488, 0.00000956, 0.00001226, 0.00001264, 0.00001064, 0.00000656, 0.00000112, -0.00000460, -0.00000941, -0.00001236, -0.00001311, -0.00001211, -0.00001045, -0.00000932, -0.00000938, -0.00001027, -0.00001050, -0.00000794, -0.00000080, 0.00001140, 0.00002711, 0.00004282, 0.00005394, 0.00005628, 0.00004767, 0.00002899, 0.00000417, -0.00002080, -0.00003976, -0.00004839, -0.00004554, -0.00003361, -0.00001755, -0.00000304, 0.00000552, 0.00000647, 0.00000124, -0.00000646, -0.00001229, -0.00001309, -0.00000806, 0.00000099, 0.00001052, 0.00001678, 0.00001734, 0.00001205, 0.00000304, -0.00000616, -0.00001208, -0.00001260, -0.00000772, 0.00000057, 0.00000910, 0.00001479, 0.00001571, 0.00001167, 0.00000412, -0.00000453, -0.00001178, -0.00001587, -0.00001613, -0.00001297, },
		{ -0.00000480, -0.00000842, -0.00001092, -0.00001212, -0.00001195, -0.00001040, -0.00000749, -0.00000346, 0.00000112, 0.00000527, 0.00000787, 0.00000794, 0.00000513, -0.00000006, -0.00000641, -0.00001230, -0.00001638, -0.00001802, -0.00001758, -0.00001610, -0.00001481, -0.00001438, -0.00001455, -0.00001412, -0.00001153, -0.00000569, 0.00000324, 0.00001355, 0.00002229, 0.00002628, 0.00002336, 0.00001336, -0.00000150, -0.00001721, -0.00002918, -0.00003383, -0.00002989, -0.00001880, -0.00000427, 0.00000908, 0.00001734, 0.00001865, 0.00001370, 0.00000530, -0.00000281, -0.00000747, -0.00000721, -0.00000271, 0.00000367, 0.00000892, 0.00001067, 0.00000809, 0.00000210, -0.00000510, -0.00001092, -0.00001346, -0.00001211, -0.00000769, -0.00000199, 0.00000300, 0.00000576, 0.00000573, 0.00000329, -0.00000060, },
		{ 0.00000180, 0.00000255, 0.00000284, 0.00000279, 0.00000231, 0.00000108, -0.00000125, -0.00000471, -0.00000885, -0.00001262, -0.00001477, -0.00001428, -0.00001087, -0.00000524, 0.00000102, 0.00000599, 0.00000802, 0.00000637, 0.00000148, -0.00000519, -0.00001170, -0.00001624, -0.00001771, -0.00001587, -0.00001121, -0.00000452, 0.00000351, 0.00001245, 0.00002200, 0.00003163, 0.00004020, 0.00004591, 0.00004656, 0.00004033, 0.00002659, 0.00000656, -0.00001661, -0.00003848, -0.00005445, -0.00006115, -0.00005747, -0.00004495, -0.00002730, -0.00000921, 0.00000511, 0.00001314, 0.00001462, 0.00001136, 0.00000625, 0.00000214, 0.00000079, 0.00000239, 0.00000578, 0.00000910, 0.00001070, 0.00000977, 0.00000663, 0.00000242, -0.00000143, -0.00000381, -0.00000431, -0.00000325, -0.00000140, 0.00000045, },
		{ -0.00000210, -0.00000142, -0.00000141, -0.00000202, -0.00000266, -0.00000246, -0.00000076, 0.00000249, 0.00000650, 0.00000986, 0.00001104, 0.00000905, 0.00000393, -0.00000312, -0.00001005, -0.00001467, -0.00001545, -0.00001204, -0.00000550, 0.00000204, 0.00000808, 0.00001044, 0.00000792, 0.00000052, -0.00001069, -0.00002388, -0.00003689, -0.00004760, -0.00005405, -0.00005462, -0.00004816, -0.00003423, -0.00001344, 0.00001221, 0.00003931, 0.00006345, 0.00008002, 0.00008535, 0.00007783, 0.00005854, 0.00003122, 0.00000144, -0.00002485, -0.00004284, -0.00005005, -0.00004686, -0.00003609, -0.00002189, -0.00000842, 0.00000136, 0.00000633, 0.00000708, 0.00000530, 0.00000288, 0.00000116, 0.00000054, 0.00000065, 0.00000076, 0.00000028, -0.00000086, -0.00000227, -0.00000333, -0.00000358, -0.00000303, },
		{ -0.00000313, -0.00000357, -0.00000181, 0.00000123, 0.00000404, 0.00000510, 0.00000361, -0.00000013, -0.00000475, -0.00000830, -0.00000910, -0.00000645, -0.00000102, 0.00000531, 0.00001011, 0.00001129, 0.00000789, 0.00000049, -0.00000899, -0.00001786, -0.00002352, -0.00002424, -0.00001953, -0.00001024, 0.00000183, 0.00001434, 0.00002489, 0.00003134, 0.00003206, 0.00002603, 0.00001314, -0.00000563, -0.00002797, -0.00005037, -0.00006855, -0.00007836, -0.00007690, -0.00006350, -0.00004029, -0.00001194, 0.00001535, 0.00003548, 0.00004406, 0.00003968, 0.00002421, 0.00000217, -0.00002066, -0.00003894, -0.00004916, -0.00005039, -0.00004405, -0.00003310, -0.00002080, -0.00000971, -0.00000114, 0.00000476, 0.00000858, 0.00001095, 0.00001218, 0.00001214, 0.00001057, 0.00000749, 0.00000344, -0.00000051, },
		{ 0.00001429, 0.00000899, 0.00000119, -0.00000684, -0.00001263, -0.00001442, -0.00001192, -0.00000645, -0.00000052, 0.00000321, 0.00000298, -0.00000138, -0.00000835, -0.00001534, -0.00001969, -0.00001973, -0.00001542, -0.00000838, -0.00000123, 0.00000330, 0.00000328, -0.00000190, -0.00001148, -0.00002377, -0.00003671, -0.00004837, -0.00005719, -0.00006197, -0.00006182, -0.00005612, -0.00004475, -0.00002845, -0.00000916, 0.00001009, 0.00002559, 0.00003401, 0.00003345, 0.00002435, 0.00000969, -0.00000574, -0.00001666, -0.00001902, -0.00001144, 0.00000432, 0.00002379, 0.00004123, 0.00005145, 0.00005148, 0.00004131, 0.00002376, 0.00000332, -0.00001534, -0.00002879, -0.00003562, -0.00003633, -0.00003267, -0.00002666, -0.00001979, -0.00001273, -0.00000560, 0.00000158, 0.00000829, 0.00001348, 0.00001580, },
		{ -0.00000873, 0.00000253, 0.00001347, 0.00002118, 0.00002330, 0.00001908, 0.00000987, -0.00000120, -0.00001030, -0.00001432, -0.00001202, -0.00000450, 0.00000529, 0.00001371, 0.00001790, 0.00001682, 0.00001153, 0.00000469, -0.00000058, -0.00000193, 0.00000148, 0.00000888, 0.00001843, 0.00002813, 0.00003655, 0.00004309, 0.00004773, 0.00005045, 0.00005069, 0.00004727, 0.00003882, 0.00002463, 0.00000543, -0.00001628, -0.00003662, -0.00005149, -0.00005795, -0.00005535, -0.00004571, -0.00003302, -0.00002185, -0.00001560, -0.00001530, -0.00001923, -0.00002370, -0.00002466, -0.00001935, -0.00000751, 0.00000846, 0.00002437, 0.00003581, 0.00003970, 0.00003532, 0.00002437, 0.00001011, -0.00000392, -0.00001515, -0.00002257, -0.00002658, -0.00002827, -0.00002850, -0.00002734, -0.00002412, -0.00001798, },
		{ -0.00001038, -0.00001513, -0.00001756, -0.00001515, -0.00000656, 0.00000738, 0.00002353, 0.00003744, 0.00004491, 0.00004366, 0.00003430, 0.00002020, 0.00000623, -0.00000308, -0.00000522, -0.00000051, 0.00000814, 0.00001661, 0.00002125, 0.00002024, 0.00001408, 0.00000497, -0.00000437, -0.00001210, -0.00001794, -0.00002300, -0.00002881, -0.00003602, -0.00004353, -0.00004849, -0.00004736, -0.00003745, -0.00001850, 0.00000670, 0.00003278, 0.00005346, 0.00006365, 0.00006119, 0.00004761, 0.00002748, 0.00000666, -0.00000974, -0.00001905, -0.00002157, -0.00001995, -0.00001755, -0.00001690, -0.00001848, -0.00002079, -0.00002120, -0.00001743, -0.00000879, 0.00000331, 0.00001588, 0.00002556, 0.00002994, 0.00002853, 0.00002275, 0.00001518, 0.00000829, 0.00000341, 0.00000032, -0.00000226, -0.00000570, },
		{ 0.00000409, 0.00000569, 0.00000570, 0.00000147, -0.00000781, -0.00002031, -0.00003200, -0.00003818, -0.00003545, -0.00002343, -0.00000515, 0.00001379, 0.00002736, 0.00003135, 0.00002504, 0.00001145, -0.00000395, -0.00001544, -0.00001926, -0.00001491, -0.00000498, 0.00000632, 0.00001528, 0.00002035, 0.00002275, 0.00002550, 0.00003158, 0.00004192, 0.00005426, 0.00006358, 0.00006394, 0.00005099, 0.00002429, -0.00001196, -0.00004979, -0.00008000, -0.00009522, -0.00009246, -0.00007396, -0.00004635, -0.00001817, 0.00000309, 0.00001320, 0.00001216, 0.00000342, -0.00000800, -0.00001766, -0.00002317, -0.00002455, -0.00002345, -0.00002179, -0.00002053, -0.00001926, -0.00001658, -0.00001118, -0.00000285, 0.00000706, 0.00001604, 0.00002154, 0.00002214, 0.00001830, 0.00001218, 0.00000664, 0.00000384, },
		{ 0.00000273, -0.00000487, -0.00000922, -0.00000752, 0.00000025, 0.00001100, 0.00001967, 0.00002136, 0.00001365, -0.00000203, -0.00002062, -0.00003522, -0.00003990, -0.00003217, -0.00001422, 0.00000781, 0.00002619, 0.00003462, 0.00003072, 0.00001678, -0.00000141, -0.00001722, -0.00002615, -0.00002772, -0.00002549, -0.00002511, -0.00003133, -0.00004521, -0.00006301, -0.00007716, -0.00007918, -0.00006338, -0.00002964, 0.00001577, 0.00006184, 0.00009648, 0.00011064, 0.00010141, 0.00007294, 0.00003478, -0.00000156, -0.00002653, -0.00003548, -0.00002954, -0.00001444, 0.00000220, 0.00001380, 0.00001689, 0.00001166, 0.00000102, -0.00001120, -0.00002190, -0.00002960, -0.00003435, -0.00003681, -0.00003737, -0.00003569, -0.00003098, -0.00002281, -0.00001198, -0.00000066, 0.00000813, 0.00001179, 0.00000947, },
		{ 0.00000410, 0.00001353, 0.00001538, 0.00000851, -0.00000479, -0.00001928, -0.00002870, -0.00002837, -0.00001740, 0.00000056, 0.00001842, 0.00002847, 0.00002564, 0.00000984, -0.00001362, -0.00003594, -0.00004825, -0.00004529, -0.00002758, -0.00000129, 0.00002433, 0.00004087, 0.00004438, 0.00003700, 0.00002604, 0.00002055, 0.00002707, 0.00004622, 0.00007179, 0.00009277, 0.00009765, 0.00007925, 0.00003815, -0.00001685, -0.00007140, -0.00011050, -0.00012378, -0.00010905, -0.00007285, -0.00002788, 0.00001167, 0.00003499, 0.00003798, 0.00002385, 0.00000109, -0.00002025, -0.00003235, -0.00003205, -0.00002118, -0.00000504, 0.00001019, 0.00001985, 0.00002202, 0.00001751, 0.00000867, -0.00000190, -0.00001235, -0.00002170, -0.00002933, -0.00003436, -0.00003554, -0.00003162, -0.00002237, -0.00000932, },
		{ -0.00005556, -0.00005203, -0.00004012, -0.00002148, -0.00000064, 0.00001610, 0.00002298, 0.00001733, 0.00000113, -0.00001914, -0.00003472, -0.00003804, -0.00002599, -0.00000174, 0.00002615, 0.00004683, 0.00005157, 0.00003750, 0.00000911, -0.00002332, -0.00004782, -0.00005573, -0.00004548, -0.00002355, -0.00000187, 0.00000714, -0.00000388, -0.00003354, -0.00007158, -0.00010236, -0.00011088, -0.00008884, -0.00003835, 0.00002824, 0.00009242, 0.00013597, 0.00014737, 0.00012556, 0.00007997, 0.00002666, -0.00001760, -0.00004117, -0.00004100, -0.00002269, 0.00000262, 0.00002302, 0.00003030, 0.00002249, 0.00000366, -0.00001856, -0.00003635, -0.00004469, -0.00004274, -0.00003330, -0.00002103, -0.00001040, -0.00000419, -0.00000316, -0.00000662, -0.00001339, -0.00002242, -0.00003279, -0.00004325, -0.00005176, },
		{ 0.00000199, -0.00000908, -0.00002379, -0.00003945, -0.00005170, -0.00005575, -0.00004843, -0.00003020, -0.00000586, 0.00001647, 0.00002821, 0.00002383, 0.00000367, -0.00002543, -0.00005226, -0.00006539, -0.00005802, -0.00003114, 0.00000620, 0.00004014, 0.00005747, 0.00005129, 0.00002434, -0.00001149, -0.00003959, -0.00004534, -0.00002250, 0.00002348, 0.00007667, 0.00011616, 0.00012388, 0.00009169, 0.00002493, -0.00005900, -0.00013653, -0.00018614, -0.00019569, -0.00016627, -0.00011107, -0.00005011, -0.00000277, 0.00001879, 0.00001306, -0.00001160, -0.00004130, -0.00006250, -0.00006704, -0.00005447, -0.00003110, -0.00000661, 0.00001029, 0.00001503, 0.00000837, -0.00000483, -0.00001818, -0.00002649, -0.00002752, -0.00002207, -0.00001293, -0.00000327, 0.00000458, 0.00000943, 0.00001087, 0.00000855, },
		{ -0.00001208, -0.00000719, -0.00000261, 0.00000029, -0.00000084, -0.00000841, -0.00002333, -0.00004352, -0.00006369, -0.00007668, -0.00007626, -0.00006019, -0.00003222, -0.00000163, 0.00001976, 0.00002223, 0.00000258, -0.00003369, -0.00007364, -0.00010161, -0.00010552, -0.00008239, -0.00004054, 0.00000279, 0.00002776, 0.00001995, -0.00002317, -0.00009002, -0.00015806, -0.00020161, -0.00020118, -0.00015104, -0.00006191, 0.00004236, 0.00013285, 0.00018547, 0.00018896, 0.00014811, 0.00008112, 0.00001241, -0.00003624, -0.00005296, -0.00003876, -0.00000547, 0.00002996, 0.00005251, 0.00005447, 0.00003732, 0.00000988, -0.00001628, -0.00003179, -0.00003290, -0.00002208, -0.00000609, 0.00000750, 0.00001348, 0.00001063, 0.00000146, -0.00000965, -0.00001863, -0.00002329, -0.00002361, -0.00002093, -0.00001677, },
		{ -0.00002180, -0.00002269, -0.00002314, -0.00002340, -0.00002252, -0.00001888, -0.00001160, -0.00000188, 0.00000654, 0.00000864, 0.00000052, -0.00001802, -0.00004207, -0.00006255, -0.00006943, -0.00005615, -0.00002335, 0.00002011, 0.00005954, 0.00007996, 0.00007250, 0.00003909, -0.00000688, -0.00004469, -0.00005405, -0.00002353, 0.00004364, 0.00012909, 0.00020472, 0.00024240, 0.00022423, 0.00014971, 0.00003668, -0.00008422, -0.00017987, -0.00022578, -0.00021397, -0.00015492, -0.00007281, 0.00000364, 0.00005142, 0.00006044, 0.00003558, -0.00000698, -0.00004721, -0.00006913, -0.00006636, -0.00004316, -0.00001129, 0.00001573, 0.00002822, 0.00002365, 0.00000661, -0.00001422, -0.00003022, -0.00003634, -0.00003257, -0.00002292, -0.00001294, -0.00000695, -0.00000645, -0.00001014, -0.00001529, -0.00001950, },
		{ 0.00000284, 0.00000367, 0.00000543, 0.00000907, 0.00001364, 0.00001672, 0.00001615, 0.00001171, 0.00000585, 0.00000284, 0.00000637, 0.00001707, 0.00003112, 0.00004109, 0.00003903, 0.00002059, -0.00001189, -0.00004881, -0.00007636, -0.00008207, -0.00006085, -0.00001867, 0.00002804, 0.00005769, 0.00005192, 0.00000372, -0.00007800, -0.00016970, -0.00024054, -0.00026304, -0.00022352, -0.00012789, -0.00000066, 0.00012293, 0.00020869, 0.00023498, 0.00019964, 0.00011988, 0.00002532, -0.00005295, -0.00009311, -0.00008925, -0.00005167, -0.00000119, 0.00003971, 0.00005558, 0.00004288, 0.00000976, -0.00002864, -0.00005708, -0.00006630, -0.00005584, -0.00003293, -0.00000843, 0.00000819, 0.00001246, 0.00000591, -0.00000576, -0.00001590, -0.00002008, -0.00001761, -0.00001098, -0.00000393, 0.00000082, },
		{ -0.00001737, -0.00001696, -0.00001868, -0.00002364, -0.00003025, -0.00003502, -0.00003468, -0.00002839, -0.00001875, -0.00001084, -0.00000964, -0.00001709, -0.00003038, -0.00004263, -0.00004600, -0.00003579, -0.00001368, 0.00001186, 0.00002856, 0.00002588, 0.00000070, -0.00003951, -0.00007794, -0.00009427, -0.00007298, -0.00001103, 0.00007875, 0.00017024, 0.00023233, 0.00024001, 0.00018409, 0.00007567, -0.00005671, -0.00017633, -0.00025076, -0.00026286, -0.00021610, -0.00013249, -0.00004387, 0.00001987, 0.00004119, 0.00001997, -0.00002817, -0.00007935, -0.00011133, -0.00011185, -0.00008231, -0.00003574, 0.00000959, 0.00003795, 0.00004200, 0.00002468, -0.00000336, -0.00002911, -0.00004275, -0.00004121, -0.00002822, -0.00001157, 0.00000101, 0.00000524, 0.00000153, -0.00000624, -0.00001337, -0.00001702, },
		{ -0.00002230, -0.00002349, -0.00002149, -0.00001522, -0.00000705, -0.00000154, -0.00000269, -0.00001134, -0.00002403, -0.00003431, -0.00003589, -0.00002615, -0.00000815, 0.00001016, 0.00001939, 0.00001338, -0.00000714, -0.00003399, -0.00005466, -0.00005806, -0.00004051, -0.00000888, 0.00002073, 0.00002902, 0.00000189, -0.00006216, -0.00014908, -0.00023248, -0.00028231, -0.00027595, -0.00020734, -0.00009045, 0.00004459, 0.00016147, 0.00023011, 0.00023679, 0.00018824, 0.00010811, 0.00002752, -0.00002673, -0.00004142, -0.00002045, 0.00001850, 0.00005243, 0.00006278, 0.00004268, -0.00000118, -0.00005237, -0.00009220, -0.00010750, -0.00009554, -0.00006419, -0.00002773, -0.00000045, 0.00000906, 0.00000061, -0.00001879, -0.00003895, -0.00005135, -0.00005254, -0.00004474, -0.00003364, -0.00002497, -0.00002156, },
		{ -0.00003152, -0.00002812, -0.00002938, -0.00003616, -0.00004507, -0.00005032, -0.00004720, -0.00003516, -0.00001879, -0.00000604, -0.00000426, -0.00001613, -0.00003751, -0.00005863, -0.00006820, -0.00005898, -0.00003181, 0.00000379, 0.00003325, 0.00004351, 0.00002962, -0.00000170, -0.00003355, -0.00004532, -0.00002166, 0.00003969, 0.00012493, 0.00020753, 0.00025734, 0.00025188, 0.00018542, 0.00007221, -0.00005778, -0.00016908, -0.00023300, -0.00023732, -0.00018964, -0.00011335, -0.00003801, 0.00001190, 0.00002539, 0.00000741, -0.00002521, -0.00005218, -0.00005879, -0.00004170, -0.00000944, 0.00002238, 0.00003857, 0.00003118, 0.00000275, -0.00003525, -0.00006776, -0.00008274, -0.00007602, -0.00005239, -0.00002285, 0.00000069, 0.00001045, 0.00000558, -0.00000857, -0.00002379, -0.00003336, -0.00003505, },
		{ -0.00005422, -0.00005923, -0.00005631, -0.00004523, -0.00003115, -0.00002163, -0.00002229, -0.00003329, -0.00004856, -0.00005849, -0.00005471, -0.00003485, -0.00000465, 0.00002384, 0.00003748, 0.00002836, -0.00000170, -0.00004086, -0.00007182, -0.00007946, -0.00005834, -0.00001653, 0.00002635, 0.00004657, 0.00002637, -0.00003739, -0.00012999, -0.00022263, -0.00028228, -0.00028384, -0.00021993, -0.00010429, 0.00003265, 0.00015433, 0.00023069, 0.00024812, 0.00021280, 0.00014671, 0.00007806, 0.00003032, 0.00001416, 0.00002533, 0.00004868, 0.00006618, 0.00006516, 0.00004326, 0.00000832, -0.00002640, -0.00004894, -0.00005431, -0.00004635, -0.00003534, -0.00003262, -0.00004489, -0.00007102, -0.00010258, -0.00012791, -0.00013751, -0.00012834, -0.00010497, -0.00007716, -0.00005519, -0.00004522, -0.00004689, },
		{ -0.00004144, -0.00002449, -0.00002362, -0.00003603, -0.00005258, -0.00006240, -0.00005850, -0.00004150, -0.00001972, -0.00000521, -0.00000767, -0.00002907, -0.00006163, -0.00009067, -0.00010106, -0.00008459, -0.00004479, 0.00000344, 0.00003963, 0.00004675, 0.00001964, -0.00003111, -0.00008200, -0.00010546, -0.00008144, -0.00000695, 0.00010059, 0.00020802, 0.00027767, 0.00028098, 0.00020928, 0.00007741, -0.00008110, -0.00022558, -0.00032196, -0.00035347, -0.00032452, -0.00025672, -0.00017908, -0.00011676, -0.00008270, -0.00007511, -0.00008125, -0.00008500, -0.00007478, -0.00004829, -0.00001237, 0.00002125, 0.00004205, 0.00004555, 0.00003477, 0.00001784, 0.00000335, -0.00000410, -0.00000597, -0.00000842, -0.00001838, -0.00003919, -0.00006801, -0.00009652, -0.00011458, -0.00011511, -0.00009773, -0.00006940, },
		{ -0.00018336, -0.00017784, -0.00014778, -0.00010471, -0.00006509, -0.00004368, -0.00004722, -0.00007129, -0.00010219, -0.00012305, -0.00012169, -0.00009674, -0.00005885, -0.00002639, -0.00001706, -0.00003931, -0.00008734, -0.00014240, -0.00018033, -0.00018225, -0.00014396, -0.00007962, -0.00001764, 0.00001008, -0.00001855, -0.00010530, -0.00022858, -0.00034939, -0.00042455, -0.00042206, -0.00033298, -0.00017509, 0.00001297, 0.00018478, 0.00030150, 0.00034368, 0.00031561, 0.00024093, 0.00015216, 0.00007835, 0.00003588, 0.00002545, 0.00003551, 0.00004988, 0.00005584, 0.00004915, 0.00003440, 0.00002114, 0.00001824, 0.00002931, 0.00005126, 0.00007627, 0.00009586, 0.00010472, 0.00010245, 0.00009248, 0.00007895, 0.00006366, 0.00004469, 0.00001778, -0.00002038, -0.00006859, -0.00011963, -0.00016180, },
		{ -0.00001105, -0.00009116, -0.00017457, -0.00024241, -0.00027673, -0.00026764, -0.00021840, -0.00014560, -0.00007395, -0.00002710, -0.00001831, -0.00004484, -0.00008876, -0.00012433, -0.00012909, -0.00009400, -0.00002815, 0.00004431, 0.00009420, 0.00010031, 0.00005984, -0.00000838, -0.00006991, -0.00008788, -0.00003843, 0.00007787, 0.00023362, 0.00038246, 0.00047469, 0.00047482, 0.00037452, 0.00019594, -0.00001563, -0.00020696, -0.00033416, -0.00037581, -0.00033733, -0.00024598, -0.00013892, -0.00004951, 0.00000296, 0.00001677, 0.00000426, -0.00001632, -0.00002960, -0.00002873, -0.00001626, -0.00000058, 0.00000966, 0.00001024, 0.00000307, -0.00000540, -0.00000787, 0.00000011, 0.00001836, 0.00004279, 0.00006787, 0.00008930, 0.00010520, 0.00011537, 0.00011916, 0.00011355, 0.00009309, 0.00005217, },
		{ 0.00023251, 0.00025598, 0.00025381, 0.00021462, 0.00013374, 0.00001811, -0.00011237, -0.00022873, -0.00030149, -0.00031121, -0.00025680, -0.00015775, -0.00004870, 0.00003243, 0.00005844, 0.00002348, -0.00005430, -0.00013961, -0.00019418, -0.00019283, -0.00013489, -0.00004606, 0.00003064, 0.00005162, -0.00000915, -0.00014782, -0.00032934, -0.00049801, -0.00059635, -0.00058525, -0.00045791, -0.00024259, 0.00000691, 0.00022912, 0.00037453, 0.00041998, 0.00037308, 0.00026584, 0.00014106, 0.00003686, -0.00002469, -0.00004146, -0.00002723, -0.00000266, 0.00001440, 0.00001538, 0.00000220, -0.00001633, -0.00003027, -0.00003375, -0.00002704, -0.00001531, -0.00000507, -0.00000051, -0.00000160, -0.00000449, -0.00000388, 0.00000440, 0.00002185, 0.00004749, 0.00007930, 0.00011556, 0.00015508, 0.00019577, },
		{ 0.00004752, 0.00007386, 0.00011411, 0.00016771, 0.00022758, 0.00027852, 0.00029908, 0.00026759, 0.00017089, 0.00001236, -0.00018430, -0.00037856, -0.00052382, -0.00058238, -0.00053949, -0.00041062, -0.00023834, -0.00007896, 0.00001701, 0.00002336, -0.00005128, -0.00016569, -0.00026064, -0.00028106, -0.00019758, -0.00001931, 0.00020722, 0.00041296, 0.00052972, 0.00051371, 0.00036061, 0.00010702, -0.00018201, -0.00043450, -0.00059411, -0.00063592, -0.00057060, -0.00043632, -0.00028279, -0.00015371, -0.00007407, -0.00004616, -0.00005423, -0.00007472, -0.00008716, -0.00008133, -0.00005879, -0.00002905, -0.00000322, 0.00001159, 0.00001485, 0.00001147, 0.00000830, 0.00001029, 0.00001826, 0.00002925, 0.00003867, 0.00004297, 0.00004134, 0.00003566, 0.00002924, 0.00002521, 0.00002573, 0.00003247, },
		{ 0.00003200, 0.00003643, 0.00003991, 0.00004467, 0.00005594, 0.00008205, 0.00013201, 0.00021056, 0.00031232, 0.00041818, 0.00049644, 0.00051022, 0.00042983, 0.00024597, -0.00002128, -0.00032274, -0.00059110, -0.00076049, -0.00078856, -0.00067296, -0.00045570, -0.00021239, -0.00002899, 0.00002656, -0.00007205, -0.00029917, -0.00058420, -0.00083397, -0.00096299, -0.00092103, -0.00070876, -0.00037630, -0.00000582, 0.00031524, 0.00052075, 0.00058339, 0.00051799, 0.00037102, 0.00020134, 0.00005989, -0.00002454, -0.00004946, -0.00003299, -0.00000213, 0.00001963, 0.00002035, 0.00000139, -0.00002683, -0.00005174, -0.00006493, -0.00006508, -0.00005693, -0.00004761, -0.00004257, -0.00004315, -0.00004666, -0.00004860, -0.00004532, -0.00003585, -0.00002194, -0.00000676, 0.00000693, 0.00001781, 0.00002596, },
		{ 0.00001414, 0.00001962, 0.00002584, 0.00003241, 0.00003787, 0.00003948, 0.00003466, 0.00002396, 0.00001409, 0.00001895, 0.00005668, 0.00014252, 0.00027938, 0.00044993, 0.00061462, 0.00071847, 0.00070655, 0.00054393, 0.00023337, -0.00017670, -0.00059866, -0.00092698, -0.00106854, -0.00097281, -0.00065135, -0.00017899, 0.00032467, 0.00072868, 0.00092838, 0.00087533, 0.00058975, 0.00015199, -0.00032376, -0.00072426, -0.00096882, -0.00102814, -0.00092530, -0.00072083, -0.00048837, -0.00029041, -0.00016195, -0.00010634, -0.00010253, -0.00011905, -0.00012831, -0.00011597, -0.00008284, -0.00004020, -0.00000202, 0.00002206, 0.00003017, 0.00002723, 0.00002112, 0.00001819, 0.00002064, 0.00002642, 0.00003139, 0.00003213, 0.00002780, 0.00002027, 0.00001279, 0.00000810, 0.00000727, 0.00000971, },
		{ -0.00012100, -0.00009890, -0.00007393, -0.00004678, -0.00001764, 0.00001439, 0.00005078, 0.00009170, 0.00013405, 0.00017094, 0.00019386, 0.00019757, 0.00018599, 0.00017576, 0.00019459, 0.00027282, 0.00042968, 0.00065883, 0.00091926, 0.00113731, 0.00122193, 0.00109076, 0.00069954, 0.00006505, -0.00072778, -0.00153923, -0.00220669, -0.00258671, -0.00259449, -0.00222825, -0.00157067, -0.00076639, 0.00001720, 0.00063373, 0.00099285, 0.00107646, 0.00093304, 0.00065385, 0.00034062, 0.00007567, -0.00009646, -0.00017216, -0.00017828, -0.00015439, -0.00013531, -0.00013993, -0.00016898, -0.00021039, -0.00024839, -0.00027163, -0.00027725, -0.00027005, -0.00025840, -0.00024936, -0.00024565, -0.00024541, -0.00024436, -0.00023865, -0.00022695, -0.00021061, -0.00019235, -0.00017443, -0.00015740, -0.00014015, },
		{ 0.00010662, 0.00011412, 0.00011868, 0.00012213, 0.00012590, 0.00013014, 0.00013386, 0.00013632, 0.00013891, 0.00014591, 0.00016336, 0.00019572, 0.00024199, 0.00029354, 0.00033599, 0.00035587, 0.00035017, 0.00033497, 0.00034809, 0.00044224, 0.00066819, 0.00105226, 0.00157537, 0.00216266, 0.00268993, 0.00300850, 0.00298312, 0.00253240, 0.00165894, 0.00045842, -0.00089680, -0.00219629, -0.00324136, -0.00389177, -0.00409567, -0.00389464, -0.00340378, -0.00277454, -0.00215249, -0.00164317, -0.00129470, -0.00109978, -0.00101319, -0.00097623, -0.00093909, -0.00087414, -0.00077800, -0.00066428, -0.00055225, -0.00045665, -0.00038254, -0.00032578, -0.00027762, -0.00023012, -0.00017976, -0.00012797, -0.00007903, -0.00003699, -0.00000350, 0.00002258, 0.00004392, 0.00006287, 0.00008021, 0.00009519, },
		{ 0.00038548, 0.00038775, 0.00039386, 0.00040235, 0.00041222, 0.00042398, 0.00043963, 0.00046137, 0.00048976, 0.00052259, 0.00055557, 0.00058480, 0.00061009, 0.00063695, 0.00067554, 0.00073595, 0.00082144, 0.00092291, 0.00101830, 0.00107939, 0.00108523, 0.00103822, 0.00097611, 0.00097326, 0.00112730, 0.00153240, 0.00224596, 0.00325930, 0.00448338, 0.00575687, 0.00687694, 0.00764575, 0.00791953, 0.00764575, 0.00687694, 0.00575687, 0.00448338, 0.00325930, 0.00224596, 0.00153240, 0.00112730, 0.00097326, 0.00097611, 0.00103822, 0.00108523, 0.00107939, 0.00101830, 0.00092291, 0.00082144, 0.00073595, 0.00067554, 0.00063695, 0.00061009, 0.00058480, 0.00055557, 0.00052259, 0.00048976, 0.00046137, 0.00043963, 0.00042398, 0.00041222, 0.00040235, 0.00039386, 0.00038775, },
		{ 0.00010662, 0.00009519, 0.00008021, 0.00006287, 0.00004392, 0.00002258, -0.00000350, -0.00003699, -0.00007903, -0.00012797, -0.00017976, -0.00023012, -0.00027762, -0.00032578, -0.00038254, -0.00045665, -0.00055225, -0.00066428, -0.00077800, -0.00087414, -0.00093909, -0.00097623, -0.00101319, -0.00109978, -0.00129470, -0.00164317, -0.00215249, -0.00277454, -0.00340378, -0.00389464, -0.00409567, -0.00389177, -0.00324136, -0.00219629, -0.00089680, 0.00045842, 0.00165894, 0.00253240, 0.00298312, 0.00300850, 0.00268993, 0.00216266, 0.00157537, 0.00105226, 0.00066819, 0.00044224, 0.00034809, 0.00033497, 0.00035017, 0.00035587, 0.00033599, 0.00029354, 0.00024199, 0.00019572, 0.00016336, 0.00014591, 0.00013891, 0.00013632, 0.00013386, 0.00013014, 0.00012590, 0.00012213, 0.00011868, 0.00011412, },
		{ -0.00012100, -0.00014015, -0.00015740, -0.00017443, -0.00019235, -0.00021061, -0.00022695, -0.00023865, -0.00024436, -0.00024541, -0.00024565, -0.00024936, -0.00025840, -0.00027005, -0.00027725, -0.00027163, -0.00024839, -0.00021039, -0.00016898, -0.00013993, -0.00013531, -0.00015439, -0.00017828, -0.00017216, -0.00009646, 0.00007567, 0.00034062, 0.00065385, 0.00093304, 0.00107646, 0.00099285, 0.00063373, 0.00001720, -0.00076639, -0.00157067, -0.00222825, -0.00259449, -0.00258671, -0.00220669, -0.00153923, -0.00072778, 0.00006505, 0.00069954, 0.00109076, 0.00122193, 0.00113731, 0.00091926, 0.00065883, 0.00042968, 0.00027282, 0.00019459, 0.00017576, 0.00018599, 0.00019757, 0.00019386, 0.00017094, 0.00013405, 0.00009170, 0.00005078, 0.00001439, -0.00001764, -0.00004678, -0.00007393, -0.00009890, },
		{ 0.00001414, 0.00000971, 0.00000727, 0.00000810, 0.00001279, 0.00002027, 0.00002780, 0.00003213, 0.00003139, 0.00002642, 0.00002064, 0.00001819, 0.00002112, 0.00002723, 0.00003017, 0.00002206, -0.00000202, -0.00004020, -0.00008284, -0.00011597, -0.00012831, -0.00011905, -0.00010253, -0.00010634, -0.00016195, -0.00029041, -0.00048837, -0.00072083, -0.00092530, -0.00102814, -0.00096882, -0.00072426, -0.00032376, 0.00015199, 0.00058975, 0.00087533, 0.00092838, 0.00072868, 0.00032467, -0.00017899, -0.00065135, -0.00097281, -0.00106854, -0.00092698, -0.00059866, -0.00017670, 0.00023337, 0.00054393, 0.00070655, 0.00071847, 0.00061462, 0.00044993, 0.00027938, 0.00014252, 0.00005668, 0.00001895, 0.00001409, 0.00002396, 0.00003466, 0.00003948, 0.00003787, 0.00003241, 0.00002584, 0.00001962, },
		{ 0.00003200, 0.00002596, 0.00001781, 0.00000693, -0.00000676, -0.00002194, -0.00003585, -0.00004532, -0.00004860, -0.00004666, -0.00004315, -0.00004257, -0.00004761, -0.00005693, -0.00006508, -0.00006493, -0.00005174, -0.00002683, 0.00000139, 0.00002035, 0.00001963, -0.00000213, -0.00003299, -0.00004946, -0.00002454, 0.00005989, 0.00020134, 0.00037102, 0.00051799, 0.00058339, 0.00052075, 0.00031524, -0.00000582, -0.00037630, -0.00070876, -0.00092103, -0.00096299, -0.00083397, -0.00058420, -0.00029917, -0.00007205, 0.00002656, -0.00002899, -0.00021239, -0.00045570, -0.00067296, -0.00078856, -0.00076049, -0.00059110, -0.00032274, -0.00002128, 0.00024597, 0.00042983, 0.00051022, 0.00049644, 0.00041818, 0.00031232, 0.00021056, 0.00013201, 0.00008205, 0.00005594, 0.00004467, 0.00003991, 0.00003643, },
		{ 0.00004752, 0.00003247, 0.00002573, 0.00002521, 0.00002924, 0.00003566, 0.00004134, 0.00004297, 0.00003867, 0.00002925, 0.00001826, 0.00001029, 0.00000830, 0.00001147, 0.00001485, 0.00001159, -0.00000322, -0.00002905, -0.00005879, -0.00008133, -0.00008716, -0.00007472, -0.00005423, -0.00004616, -0.00007407, -0.00015371, -0.00028279, -0.00043632, -0.00057060, -0.00063592, -0.00059411, -0.00043450, -0.00018201, 0.00010702, 0.00036061, 0.00051371, 0.00052972, 0.00041296, 0.00020722, -0.00001931, -0.00019758, -0.00028106, -0.00026064, -0.00016569, -0.00005128, 0.00002336, 0.00001701, -0.00007896, -0.00023834, -0.00041062, -0.00053949, -0.00058238, -0.00052382, -0.00037856, -0.00018430, 0.00001236, 0.00017089, 0.00026759, 0.00029908, 0.00027852, 0.00022758, 0.00016771, 0.00011411, 0.00007386, },
		{ 0.00023251, 0.00019577, 0.00015508, 0.00011556, 0.00007930, 0.00004749, 0.00002185, 0.00000440, -0.00000388, -0.00000449, -0.00000160, -0.00000051, -0.00000507, -0.00001531, -0.00002704, -0.00003375, -0.00003027, -0.00001633, 0.00000220, 0.00001538, 0.00001440, -0.00000266, -0.00002723, -0.00004146, -0.00002469, 0.00003686, 0.00014106, 0.00026584, 0.00037308, 0.00041998, 0.00037453, 0.00022912, 0.00000691, -0.00024259, -0.00045791, -0.00058525, -0.00059635, -0.00049801, -0.00032934, -0.00014782, -0.00000915, 0.00005162, 0.00003064, -0.00004606, -0.00013489, -0.00019283, -0.00019418, -0.00013961, -0.00005430, 0.00002348, 0.00005844, 0.00003243, -0.00004870, -0.00015775, -0.00025680, -0.00031121, -0.00030149, -0.00022873, -0.00011237, 0.00001811, 0.00013374, 0.00021462, 0.00025381, 0.00025598, },
		{ -0.00001105, 0.00005217, 0.00009309, 0.00011355, 0.00011916, 0.00011537, 0.00010520, 0.00008930, 0.00006787, 0.00004279, 0.00001836, 0.00000011, -0.00000787, -0.00000540, 0.00000307, 0.00001024, 0.00000966, -0.00000058, -0.00001626, -0.00002873, -0.00002960, -0.00001632, 0.00000426, 0.00001677, 0.00000296, -0.00004951, -0.00013892, -0.00024598, -0.00033733, -0.00037581, -0.00033416, -0.00020696, -0.00001563, 0.00019594, 0.00037452, 0.00047482, 0.00047469, 0.00038246, 0.00023362, 0.00007787, -0.00003843, -0.00008788, -0.00006991, -0.00000838, 0.00005984, 0.00010031, 0.00009420, 0.00004431, -0.00002815, -0.00009400, -0.00012909, -0.00012433, -0.00008876, -0.00004484, -0.00001831, -0.00002710, -0.00007395, -0.00014560, -0.00021840, -0.00026764, -0.00027673, -0.00024241, -0.00017457, -0.00009116, },
		{ -0.00018336, -0.00016180, -0.00011963, -0.00006859, -0.00002038, 0.00001778, 0.00004469, 0.00006366, 0.00007895, 0.00009248, 0.00010245, 0.00010472, 0.00009586, 0.00007627, 0.00005126, 0.00002931, 0.00001824, 0.00002114, 0.00003440, 0.00004915, 0.00005584, 0.00004988, 0.00003551, 0.00002545, 0.00003588, 0.00007835, 0.00015216, 0.00024093, 0.00031561, 0.00034368, 0.00030150, 0.00018478, 0.00001297, -0.00017509, -0.00033298, -0.00042206, -0.00042455, -0.00034939, -0.00022858, -0.00010530, -0.00001855, 0.00001008, -0.00001764, -0.00007962, -0.00014396, -0.00018225, -0.00018033, -0.00014240, -0.00008734, -0.00003931, -0.00001706, -0.00002639, -0.00005885, -0.00009674, -0.00012169, -0.00012305, -0.00010219, -0.00007129, -0.00004722, -0.00004368, -0.00006509, -0.00010471, -0.00014778, -0.00017784, },
		{ -0.00004144, -0.00006940, -0.00009773, -0.00011511, -0.00011458, -0.00009652, -0.00006801, -0.00003919, -0.00001838, -0.00000842, -0.00000597, -0.00000410, 0.00000335, 0.00001784, 0.00003477, 0.00004555, 0.00004205, 0.00002125, -0.00001237, -0.00004829, -0.00007478, -0.00008500, -0.00008125, -0.00007511, -0.00008270, -0.00011676, -0.00017908, -0.00025672, -0.00032452, -0.00035347, -0.00032196, -0.00022558, -0.00008110, 0.00007741, 0.00020928, 0.00028098, 0.00027767, 0.00020802, 0.00010059, -0.00000695, -0.00008144, -0.00010546, -0.00008200, -0.00003111, 0.00001964, 0.00004675, 0.00003963, 0.00000344, -0.00004479, -0.00008459, -0.00010106, -0.00009067, -0.00006163, -0.00002907, -0.00000767, -0.00000521, -0.00001972, -0.00004150, -0.00005850, -0.00006240, -0.00005258, -0.00003603, -0.00002362, -0.00002449, },
		{ -0.00005422, -0.00004689, -0.00004522, -0.00005519, -0.00007716, -0.00010497, -0.00012834, -0.00013751, -0.00012791, -0.00010258, -0.00007102, -0.00004489, -0.00003262, -0.00003534, -0.00004635, -0.00005431, -0.00004894, -0.00002640, 0.00000832, 0.00004326, 0.00006516, 0.00006618, 0.00004868, 0.00002533, 0.00001416, 0.00003032, 0.00007806, 0.00014671, 0.00021280, 0.00024812, 0.00023069, 0.00015433, 0.00003265, -0.00010429, -0.00021993, -0.00028384, -0.00028228, -0.00022263, -0.00012999, -0.00003739, 0.00002637, 0.00004657, 0.00002635, -0.00001653, -0.00005834, -0.00007946, -0.00007182, -0.00004086, -0.00000170, 0.00002836, 0.00003748, 0.00002384, -0.00000465, -0.00003485, -0.00005471, -0.00005849, -0.00004856, -0.00003329, -0.00002229, -0.00002163, -0.00003115, -0.00004523, -0.00005631, -0.00005923, },
		{ -0.00003152, -0.00003505, -0.00003336, -0.00002379, -0.00000857, 0.00000558, 0.00001045, 0.00000069, -0.00002285, -0.00005239, -0.00007602, -0.00008274, -0.00006776, -0.00003525, 0.00000275, 0.00003118, 0.00003857, 0.00002238, -0.00000944, -0.00004170, -0.00005879, -0.00005218, -0.00002521, 0.00000741, 0.00002539, 0.00001190, -0.00003801, -0.00011335, -0.00018964, -0.00023732, -0.00023300, -0.00016908, -0.00005778, 0.00007221, 0.00018542, 0.00025188, 0.00025734, 0.00020753, 0.00012493, 0.00003969, -0.00002166, -0.00004532, -0.00003355, -0.00000170, 0.00002962, 0.00004351, 0.00003325, 0.00000379, -0.00003181, -0.00005898, -0.00006820, -0.00005863, -0.00003751, -0.00001613, -0.00000426, -0.00000604, -0.00001879, -0.00003516, -0.00004720, -0.00005032, -0.00004507, -0.00003616, -0.00002938, -0.00002812, },
		{ -0.00002230, -0.00002156, -0.00002497, -0.00003364, -0.00004474, -0.00005254, -0.00005135, -0.00003895, -0.00001879, 0.00000061, 0.00000906, -0.00000045, -0.00002773, -0.00006419, -0.00009554, -0.00010750, -0.00009220, -0.00005237, -0.00000118, 0.00004268, 0.00006278, 0.00005243, 0.00001850, -0.00002045, -0.00004142, -0.00002673, 0.00002752, 0.00010811, 0.00018824, 0.00023679, 0.00023011, 0.00016147, 0.00004459, -0.00009045, -0.00020734, -0.00027595, -0.00028231, -0.00023248, -0.00014908, -0.00006216, 0.00000189, 0.00002902, 0.00002073, -0.00000888, -0.00004051, -0.00005806, -0.00005466, -0.00003399, -0.00000714, 0.00001338, 0.00001939, 0.00001016, -0.00000815, -0.00002615, -0.00003589, -0.00003431, -0.00002403, -0.00001134, -0.00000269, -0.00000154, -0.00000705, -0.00001522, -0.00002149, -0.00002349, },
		{ -0.00001737, -0.00001702, -0.00001337, -0.00000624, 0.00000153, 0.00000524, 0.00000101, -0.00001157, -0.00002822, -0.00004121, -0.00004275, -0.00002911, -0.00000336, 0.00002468, 0.00004200, 0.00003795, 0.00000959, -0.00003574, -0.00008231, -0.00011185, -0.00011133, -0.00007935, -0.00002817, 0.00001997, 0.00004119, 0.00001987, -0.00004387, -0.00013249, -0.00021610, -0.00026286, -0.00025076, -0.00017633, -0.00005671, 0.00007567, 0.00018409, 0.00024001, 0.00023233, 0.00017024, 0.00007875, -0.00001103, -0.00007298, -0.00009427, -0.00007794, -0.00003951, 0.00000070, 0.00002588, 0.00002856, 0.00001186, -0.00001368, -0.00003579, -0.00004600, -0.00004263, -0.00003038, -0.00001709, -0.00000964, -0.00001084, -0.00001875, -0.00002839, -0.00003468, -0.00003502, -0.00003025, -0.00002364, -0.00001868, -0.00001696, },
		{ 0.00000284, 0.00000082, -0.00000393, -0.00001098, -0.00001761, -0.00002008, -0.00001590, -0.00000576, 0.00000591, 0.00001246, 0.00000819, -0.00000843, -0.00003293, -0.00005584, -0.00006630, -0.00005708, -0.00002864, 0.00000976, 0.00004288, 0.00005558, 0.00003971, -0.00000119, -0.00005167, -0.00008925, -0.00009311, -0.00005295, 0.00002532, 0.00011988, 0.00019964, 0.00023498, 0.00020869, 0.00012293, -0.00000066, -0.00012789, -0.00022352, -0.00026304, -0.00024054, -0.00016970, -0.00007800, 0.00000372, 0.00005192, 0.00005769, 0.00002804, -0.00001867, -0.00006085, -0.00008207, -0.00007636, -0.00004881, -0.00001189, 0.00002059, 0.00003903, 0.00004109, 0.00003112, 0.00001707, 0.00000637, 0.00000284, 0.00000585, 0.00001171, 0.00001615, 0.00001672, 0.00001364, 0.00000907, 0.00000543, 0.00000367, },
		{ -0.00002180, -0.00001950, -0.00001529, -0.00001014, -0.00000645, -0.00000695, -0.00001294, -0.00002292, -0.00003257, -0.00003634, -0.00003022, -0.00001422, 0.00000661, 0.00002365, 0.00002822, 0.00001573, -0.00001129, -0.00004316, -0.00006636, -0.00006913, -0.00004721, -0.00000698, 0.00003558, 0.00006044, 0.00005142, 0.00000364, -0.00007281, -0.00015492, -0.00021397, -0.00022578, -0.00017987, -0.00008422, 0.00003668, 0.00014971, 0.00022423, 0.00024240, 0.00020472, 0.00012909, 0.00004364, -0.00002353, -0.00005405, -0.00004469, -0.00000688, 0.00003909, 0.00007250, 0.00007996, 0.00005954, 0.00002011, -0.00002335, -0.00005615, -0.00006943, -0.00006255, -0.00004207, -0.00001802, 0.00000052, 0.00000864, 0.00000654, -0.00000188, -0.00001160, -0.00001888, -0.00002252, -0.00002340, -0.00002314, -0.00002269, },
		{ -0.00001208, -0.00001677, -0.00002093, -0.00002361, -0.00002329, -0.00001863, -0.00000965, 0.00000146, 0.00001063, 0.00001348, 0.00000750, -0.00000609, -0.00002208, -0.00003290, -0.00003179, -0.00001628, 0.00000988, 0.00003732, 0.00005447, 0.00005251, 0.00002996, -0.00000547, -0.00003876, -0.00005296, -0.00003624, 0.00001241, 0.00008112, 0.00014811, 0.00018896, 0.00018547, 0.00013285, 0.00004236, -0.00006191, -0.00015104, -0.00020118, -0.00020161, -0.00015806, -0.00009002, -0.00002317, 0.00001995, 0.00002776, 0.00000279, -0.00004054, -0.00008239, -0.00010552, -0.00010161, -0.00007364, -0.00003369, 0.00000258, 0.00002223, 0.00001976, -0.00000163, -0.00003222, -0.00006019, -0.00007626, -0.00007668, -0.00006369, -0.00004352, -0.00002333, -0.00000841, -0.00000084, 0.00000029, -0.00000261, -0.00000719, },
		{ 0.00000199, 0.00000855, 0.00001087, 0.00000943, 0.00000458, -0.00000327, -0.00001293, -0.00002207, -0.00002752, -0.00002649, -0.00001818, -0.00000483, 0.00000837, 0.00001503, 0.00001029, -0.00000661, -0.00003110, -0.00005447, -0.00006704, -0.00006250, -0.00004130, -0.00001160, 0.00001306, 0.00001879, -0.00000277, -0.00005011, -0.00011107, -0.00016627, -0.00019569, -0.00018614, -0.00013653, -0.00005900, 0.00002493, 0.00009169, 0.00012388, 0.00011616, 0.00007667, 0.00002348, -0.00002250, -0.00004534, -0.00003959, -0.00001149, 0.00002434, 0.00005129, 0.00005747, 0.00004014, 0.00000620, -0.00003114, -0.00005802, -0.00006539, -0.00005226, -0.00002543, 0.00000367, 0.00002383, 0.00002821, 0.00001647, -0.00000586, -0.00003020, -0.00004843, -0.00005575, -0.00005170, -0.00003945, -0.00002379, -0.00000908, },
		{ -0.00005556, -0.00005176, -0.00004325, -0.00003279, -0.00002242, -0.00001339, -0.00000662, -0.00000316, -0.00000419, -0.00001040, -0.00002103, -0.00003330, -0.00004274, -0.00004469, -0.00003635, -0.00001856, 0.00000366, 0.00002249, 0.00003030, 0.00002302, 0.00000262, -0.00002269, -0.00004100, -0.00004117, -0.00001760, 0.00002666, 0.00007997, 0.00012556, 0.00014737, 0.00013597, 0.00009242, 0.00002824, -0.00003835, -0.00008884, -0.00011088, -0.00010236, -0.00007158, -0.00003354, -0.00000388, 0.00000714, -0.00000187, -0.00002355, -0.00004548, -0.00005573, -0.00004782, -0.00002332, 0.00000911, 0.00003750, 0.00005157, 0.00004683, 0.00002615, -0.00000174, -0.00002599, -0.00003804, -0.00003472, -0.00001914, 0.00000113, 0.00001733, 0.00002298, 0.00001610, -0.00000064, -0.00002148, -0.00004012, -0.00005203, },
		{ 0.00000410, -0.00000932, -0.00002237, -0.00003162, -0.00003554, -0.00003436, -0.00002933, -0.00002170, -0.00001235, -0.00000190, 0.00000867, 0.00001751, 0.00002202, 0.00001985, 0.00001019, -0.00000504, -0.00002118, -0.00003205, -0.00003235, -0.00002025, 0.00000109, 0.00002385, 0.00003798, 0.00003499, 0.00001167, -0.00002788, -0.00007285, -0.00010905, -0.00012378, -0.00011050, -0.00007140, -0.00001685, 0.00003815, 0.00007925, 0.00009765, 0.00009277, 0.00007179, 0.00004622, 0.00002707, 0.00002055, 0.00002604, 0.00003700, 0.00004438, 0.00004087, 0.00002433, -0.00000129, -0.00002758, -0.00004529, -0.00004825, -0.00003594, -0.00001362, 0.00000984, 0.00002564, 0.00002847, 0.00001842, 0.00000056, -0.00001740, -0.00002837, -0.00002870, -0.00001928, -0.00000479, 0.00000851, 0.00001538, 0.00001353, },
		{ 0.00000273, 0.00000947, 0.00001179, 0.00000813, -0.00000066, -0.00001198, -0.00002281, -0.00003098, -0.00003569, -0.00003737, -0.00003681, -0.00003435, -0.00002960, -0.00002190, -0.00001120, 0.00000102, 0.00001166, 0.00001689, 0.00001380, 0.00000220, -0.00001444, -0.00002954, -0.00003548, -0.00002653, -0.00000156, 0.00003478, 0.00007294, 0.00010141, 0.00011064, 0.00009648, 0.00006184, 0.00001577, -0.00002964, -0.00006338, -0.00007918, -0.00007716, -0.00006301, -0.00004521, -0.00003133, -0.00002511, -0.00002549, -0.00002772, -0.00002615, -0.00001722, -0.00000141, 0.00001678, 0.00003072, 0.00003462, 0.00002619, 0.00000781, -0.00001422, -0.00003217, -0.00003990, -0.00003522, -0.00002062, -0.00000203, 0.00001365, 0.00002136, 0.00001967, 0.00001100, 0.00000025, -0.00000752, -0.00000922, -0.00000487, },
		{ 0.00000409, 0.00000384, 0.00000664, 0.00001218, 0.00001830, 0.00002214, 0.00002154, 0.00001604, 0.00000706, -0.00000285, -0.00001118, -0.00001658, -0.00001926, -0.00002053, -0.00002179, -0.00002345, -0.00002455, -0.00002317, -0.00001766, -0.00000800, 0.00000342, 0.00001216, 0.00001320, 0.00000309, -0.00001817, -0.00004635, -0.00007396, -0.00009246, -0.00009522, -0.00008000, -0.00004979, -0.00001196, 0.00002429, 0.00005099, 0.00006394, 0.00006358, 0.00005426, 0.00004192, 0.00003158, 0.00002550, 0.00002275, 0.00002035, 0.00001528, 0.00000632, -0.00000498, -0.00001491, -0.00001926, -0.00001544, -0.00000395, 0.00001145, 0.00002504, 0.00003135, 0.00002736, 0.00001379, -0.00000515, -0.00002343, -0.00003545, -0.00003818, -0.00003200, -0.00002031, -0.00000781, 0.00000147, 0.00000570, 0.00000569, },
		{ -0.00001038, -0.00000570, -0.00000226, 0.00000032, 0.00000341, 0.00000829, 0.00001518, 0.00002275, 0.00002853, 0.00002994, 0.00002556, 0.00001588, 0.00000331, -0.00000879, -0.00001743, -0.00002120, -0.00002079, -0.00001848, -0.00001690, -0.00001755, -0.00001995, -0.00002157, -0.00001905, -0.00000974, 0.00000666, 0.00002748, 0.00004761, 0.00006119, 0.00006365, 0.00005346, 0.00003278, 0.00000670, -0.00001850, -0.00003745, -0.00004736, -0.00004849, -0.00004353, -0.00003602, -0.00002881, -0.00002300, -0.00001794, -0.00001210, -0.00000437, 0.00000497, 0.00001408, 0.00002024, 0.00002125, 0.00001661, 0.00000814, -0.00000051, -0.00000522, -0.00000308, 0.00000623, 0.00002020, 0.00003430, 0.00004366, 0.00004491, 0.00003744, 0.00002353, 0.00000738, -0.00000656, -0.00001515, -0.00001756, -0.00001513, },
		{ -0.00000873, -0.00001798, -0.00002412, -0.00002734, -0.00002850, -0.00002827, -0.00002658, -0.00002257, -0.00001515, -0.00000392, 0.00001011, 0.00002437, 0.00003532, 0.00003970, 0.00003581, 0.00002437, 0.00000846, -0.00000751, -0.00001935, -0.00002466, -0.00002370, -0.00001923, -0.00001530, -0.00001560, -0.00002185, -0.00003302, -0.00004571, -0.00005535, -0.00005795, -0.00005149, -0.00003662, -0.00001628, 0.00000543, 0.00002463, 0.00003882, 0.00004727, 0.00005069, 0.00005045, 0.00004773, 0.00004309, 0.00003655, 0.00002813, 0.00001843, 0.00000888, 0.00000148, -0.00000193, -0.00000058, 0.00000469, 0.00001153, 0.00001682, 0.00001790, 0.00001371, 0.00000529, -0.00000450, -0.00001202, -0.00001432, -0.00001030, -0.00000120, 0.00000987, 0.00001908, 0.00002330, 0.00002118, 0.00001347, 0.00000253, },
		{ 0.00001429, 0.00001580, 0.00001348, 0.00000829, 0.00000158, -0.00000560, -0.00001273, -0.00001979, -0.00002666, -0.00003267, -0.00003633, -0.00003562, -0.00002879, -0.00001534, 0.00000332, 0.00002376, 0.00004131, 0.00005148, 0.00005145, 0.00004123, 0.00002379, 0.00000432, -0.00001144, -0.00001902, -0.00001666, -0.00000574, 0.00000969, 0.00002435, 0.00003345, 0.00003401, 0.00002559, 0.00001009, -0.00000916, -0.00002845, -0.00004475, -0.00005612, -0.00006182, -0.00006197, -0.00005719, -0.00004837, -0.00003671, -0.00002377, -0.00001148, -0.00000190, 0.00000328, 0.00000330, -0.00000123, -0.00000838, -0.00001542, -0.00001973, -0.00001969, -0.00001534, -0.00000835, -0.00000138, 0.00000298, 0.00000321, -0.00000052, -0.00000645, -0.00001192, -0.00001442, -0.00001263, -0.00000684, 0.00000119, 0.00000899, },
		{ -0.00000313, -0.00000051, 0.00000344, 0.00000749, 0.00001057, 0.00001214, 0.00001218, 0.00001095, 0.00000858, 0.00000476, -0.00000114, -0.00000971, -0.00002080, -0.00003310, -0.00004405, -0.00005039, -0.00004916, -0.00003894, -0.00002066, 0.00000217, 0.00002421, 0.00003968, 0.00004406, 0.00003548, 0.00001535, -0.00001194, -0.00004029, -0.00006350, -0.00007690, -0.00007836, -0.00006855, -0.00005037, -0.00002797, -0.00000563, 0.00001314, 0.00002603, 0.00003206, 0.00003134, 0.00002489, 0.00001434, 0.00000183, -0.00001024, -0.00001953, -0.00002424, -0.00002352, -0.00001786, -0.00000899, 0.00000049, 0.00000789, 0.00001129, 0.00001011, 0.00000531, -0.00000102, -0.00000645, -0.00000910, -0.00000830, -0.00000475, -0.00000013, 0.00000361, 0.00000510, 0.00000404, 0.00000123, -0.00000181, -0.00000357, },
		{ -0.00000210, -0.00000303, -0.00000358, -0.00000333, -0.00000227, -0.00000086, 0.00000028, 0.00000076, 0.00000065, 0.00000054, 0.00000116, 0.00000288, 0.00000530, 0.00000708, 0.00000633, 0.00000136, -0.00000842, -0.00002189, -0.00003609, -0.00004686, -0.00005005, -0.00004284, -0.00002485, 0.00000144, 0.00003122, 0.00005854, 0.00007783, 0.00008535, 0.00008002, 0.00006345, 0.00003931, 0.00001221, -0.00001344, -0.00003423, -0.00004816, -0.00005462, -0.00005405, -0.00004760, -0.00003689, -0.00002388, -0.00001069, 0.00000052, 0.00000792, 0.00001044, 0.00000808, 0.00000204, -0.00000550, -0.00001204, -0.00001545, -0.00001467, -0.00001005, -0.00000312, 0.00000393, 0.00000905, 0.00001104, 0.00000986, 0.00000650, 0.00000249, -0.00000076, -0.00000246, -0.00000266, -0.00000202, -0.00000141, -0.00000142, },
		{ 0.00000180, 0.00000045, -0.00000140, -0.00000325, -0.00000431, -0.00000381, -0.00000143, 0.00000242, 0.00000663, 0.00000977, 0.00001070, 0.00000910, 0.00000578, 0.00000239, 0.00000079, 0.00000214, 0.00000625, 0.00001136, 0.00001462, 0.00001314, 0.00000511, -0.00000921, -0.00002730, -0.00004495, -0.00005747, -0.00006115, -0.00005445, -0.00003848, -0.00001661, 0.00000656, 0.00002659, 0.00004033, 0.00004656, 0.00004591, 0.00004020, 0.00003163, 0.00002200, 0.00001245, 0.00000351, -0.00000452, -0.00001121, -0.00001587, -0.00001771, -0.00001624, -0.00001170, -0.00000519, 0.00000148, 0.00000637, 0.00000802, 0.00000599, 0.00000102, -0.00000524, -0.00001087, -0.00001428, -0.00001477, -0.00001262, -0.00000885, -0.00000471, -0.00000125, 0.00000108, 0.00000231, 0.00000279, 0.00000284, 0.00000255, },
		{ -0.00000480, -0.00000060, 0.00000329, 0.00000573, 0.00000576, 0.00000300, -0.00000199, -0.00000769, -0.00001211, -0.00001346, -0.00001092, -0.00000510, 0.00000210, 0.00000809, 0.00001067, 0.00000892, 0.00000367, -0.00000271, -0.00000721, -0.00000747, -0.00000281, 0.00000530, 0.00001370, 0.00001865, 0.00001734, 0.00000908, -0.00000427, -0.00001880, -0.00002989, -0.00003383, -0.00002918, -0.00001721, -0.00000150, 0.00001336, 0.00002336, 0.00002628, 0.00002229, 0.00001355, 0.00000324, -0.00000569, -0.00001153, -0.00001412, -0.00001455, -0.00001438, -0.00001481, -0.00001610, -0.00001758, -0.00001802, -0.00001638, -0.00001230, -0.00000641, -0.00000006, 0.00000513, 0.00000794, 0.00000787, 0.00000527, 0.00000112, -0.00000346, -0.00000749, -0.00001040, -0.00001195, -0.00001212, -0.00001092, -0.00000842, },
		{ -0.00000751, -0.00001297, -0.00001613, -0.00001587, -0.00001178, -0.00000453, 0.00000412, 0.00001167, 0.00001571, 0.00001479, 0.00000910, 0.00000057, -0.00000772, -0.00001260, -0.00001208, -0.00000616, 0.00000304, 0.00001205, 0.00001734, 0.00001678, 0.00001052, 0.00000099, -0.00000806, -0.00001309, -0.00001229, -0.00000646, 0.00000124, 0.00000647, 0.00000552, -0.00000304, -0.00001755, -0.00003361, -0.00004554, -0.00004839, -0.00003976, -0.00002080, 0.00000417, 0.00002899, 0.00004767, 0.00005628, 0.00005394, 0.00004282, 0.00002711, 0.00001140, -0.00000080, -0.00000794, -0.00001050, -0.00001027, -0.00000938, -0.00000932, -0.00001045, -0.00001211, -0.00001311, -0.00001236, -0.00000941, -0.00000460, 0.00000112, 0.00000656, 0.00001064, 0.00001264, 0.00001226, 0.00000956, 0.00000488, -0.00000114, },
		{ 0.00002317, 0.00002400, 0.00002092, 0.00001382, 0.00000362, -0.00000771, -0.00001744, -0.00002284, -0.00002210, -0.00001506, -0.00000359, 0.00000888, 0.00001838, 0.00002179, 0.00001802, 0.00000858, -0.00000292, -0.00001203, -0.00001512, -0.00001083, -0.00000064, 0.00001166, 0.00002146, 0.00002517, 0.00002170, 0.00001297, 0.00000311, -0.00000319, -0.00000275, 0.00000466, 0.00001604, 0.00002612, 0.00002943, 0.00002249, 0.00000536, -0.00001805, -0.00004116, -0.00005691, -0.00006015, -0.00004938, -0.00002725, 0.00000040, 0.00002650, 0.00004495, 0.00005241, 0.00004896, 0.00003760, 0.00002287, 0.00000919, -0.00000045, -0.00000514, -0.00000576, -0.00000424, -0.00000255, -0.00000197, -0.00000271, -0.00000416, -0.00000525, -0.00000498, -0.00000275, 0.00000147, 0.00000719, 0.00001351, 0.00001927, },
		{ -0.00000720, -0.00000265, 0.00000431, 0.00001277, 0.00002107, 0.00002700, 0.00002836, 0.00002368, 0.00001296, -0.00000197, -0.00001766, -0.00002990, -0.00003502, -0.00003124, -0.00001948, -0.00000330, 0.00001219, 0.00002198, 0.00002304, 0.00001543, 0.00000236, -0.00001098, -0.00001936, -0.00001955, -0.00001158, 0.00000126, 0.00001367, 0.00002054, 0.00001896, 0.00000953, -0.00000383, -0.00001539, -0.00001992, -0.00001482, -0.00000136, 0.00001571, 0.00002981, 0.00003504, 0.00002839, 0.00001101, -0.00001214, -0.00003394, -0.00004757, -0.00004877, -0.00003721, -0.00001638, 0.00000772, 0.00002866, 0.00004157, 0.00004445, 0.00003831, 0.00002642, 0.00001290, 0.00000131, -0.00000627, -0.00000953, -0.00000951, -0.00000792, -0.00000636, -0.00000583, -0.00000651, -0.00000792, -0.00000916, -0.00000921, },
		},
		{ { -0.00000743, -0.00000947, -0.00001292, -0.00001205, -0.00000301, 0.00001234, 0.00002697, 0.00003321, 0.00002813, 0.00001558, 0.00000342, -0.00000222, -0.00000118, 0.00000113, -0.00000137, -0.00000985, -0.00001866, -0.00001907, -0.00000602, 0.00001683, 0.00003807, 0.00004522, 0.00003294, 0.00000706, -0.00001867, -0.00003111, -0.00002582, -0.00000945, 0.00000539, 0.00000885, 0.00000030, -0.00001203, -0.00001769, -0.00001203, 0.00000030, 0.00000885, 0.00000539, -0.00000945, -0.00002582, -0.00003111, -0.00001867, 0.00000706, 0.00003294, 0.00004522, 0.00003807, 0.00001683, -0.00000602, -0.00001907, -0.00001866, -0.00000985, -0.00000137, 0.00000113, -0.00000118, -0.00000222, 0.00000342, 0.00001558, 0.00002813, 0.00003321, 0.00002697, 0.00001234, -0.00000301, -0.00001205, -0.00001292, -0.00000947, },
		{ -0.00001093, 0.00000180, 0.00001892, 0.00003207, 0.00003419, 0.00002441, 0.00000869, -0.00000434, -0.00000931, -0.00000740, -0.00000466, -0.00000649, -0.00001281, -0.00001761, -0.00001360, 0.00000143, 0.00002152, 0.00003530, 0.00003334, 0.00001491, -0.00001082, -0.00003015, -0.00003365, -0.00002209, -0.00000565, 0.00000334, -0.00000119, -0.00001504, -0.00002724, -0.00002820, -0.00001680, -0.00000129, 0.00000682, 0.00000130, -0.00001379, -0.00002646, -0.00002504, -0.00000654, 0.00002062, 0.00004166, 0.00004477, 0.00002849, 0.00000236, -0.00001954, -0.00002741, -0.00002103, -0.00000842, 0.00000083, 0.00000231, -0.00000097, -0.00000207, 0.00000373, 0.00001488, 0.00002455, 0.00002576, 0.00001660, 0.00000170, -0.00001116, -0.00001655, -0.00001480, -0.00001093, -0.00001003, -0.00001286, -0.00001510, },
		{ 0.00003439, 0.00003088, 0.00001563, -0.00000312, -0.00001606, -0.00001885, -0.00001433, -0.00000946, -0.00000941, -0.00001341, -0.00001547, -0.00000966, 0.00000431, 0.00001956, 0.00002618, 0.00001822, -0.00000145, -0.00002232, -0.00003291, -0.00002848, -0.00001412, -0.00000108, 0.00000132, -0.00000757, -0.00001936, -0.00002287, -0.00001272, 0.00000606, 0.00002093, 0.00002045, 0.00000281, -0.00002211, -0.00003860, -0.00003538, -0.00001321, 0.00001525, 0.00003340, 0.00003107, 0.00001058, -0.00001537, -0.00003228, -0.00003278, -0.00002023, -0.00000498, 0.00000347, 0.00000271, -0.00000211, -0.00000336, 0.00000272, 0.00001301, 0.00002010, 0.00001801, 0.00000679, -0.00000761, -0.00001771, -0.00001988, -0.00001643, -0.00001303, -0.00001362, -0.00001688, -0.00001718, -0.00000926, 0.00000679, 0.00002442, },
		{ -0.00000907, -0.00001949, -0.00001752, -0.00000831, -0.00000028, 0.00000135, -0.00000198, -0.00000405, 0.00000049, 0.00001111, 0.00002110, 0.00002239, 0.00001181, -0.00000586, -0.00002075, -0.00002470, -0.00001707, -0.00000520, 0.00000106, -0.00000324, -0.00001426, -0.00002221, -0.00001900, -0.00000519, 0.00000927, 0.00001154, -0.00000452, -0.00003253, -0.00005609, -0.00005860, -0.00003418, 0.00000767, 0.00004699, 0.00006497, 0.00005484, 0.00002520, -0.00000604, -0.00002294, -0.00002041, -0.00000554, 0.00000896, 0.00001388, 0.00000899, 0.00000150, -0.00000052, 0.00000540, 0.00001452, 0.00001897, 0.00001394, 0.00000132, -0.00001190, -0.00001899, -0.00001827, -0.00001371, -0.00001108, -0.00001269, -0.00001518, -0.00001213, 0.00000036, 0.00001950, 0.00003643, 0.00004169, 0.00003160, 0.00001114, },
		{ 0.00000010, 0.00000993, 0.00001037, 0.00000424, -0.00000104, 0.00000021, 0.00000736, 0.00001409, 0.00001353, 0.00000396, -0.00000955, -0.00001876, -0.00001853, -0.00001068, -0.00000257, -0.00000118, -0.00000735, -0.00001483, -0.00001510, -0.00000466, 0.00001103, 0.00002049, 0.00001454, -0.00000579, -0.00002781, -0.00003428, -0.00001488, 0.00002554, 0.00006798, 0.00008956, 0.00007708, 0.00003512, -0.00001644, -0.00005393, -0.00006317, -0.00004627, -0.00001829, 0.00000347, 0.00000991, 0.00000400, -0.00000386, -0.00000453, 0.00000330, 0.00001324, 0.00001692, 0.00001071, -0.00000175, -0.00001272, -0.00001642, -0.00001306, -0.00000792, -0.00000625, -0.00000841, -0.00000920, -0.00000215, 0.00001450, 0.00003490, 0.00004836, 0.00004623, 0.00002810, 0.00000275, -0.00001709, -0.00002263, -0.00001412, },
		{ 0.00000309, -0.00000623, -0.00000709, -0.00000044, 0.00000689, 0.00000801, 0.00000135, -0.00000846, -0.00001463, -0.00001376, -0.00000836, -0.00000443, -0.00000603, -0.00001148, -0.00001446, -0.00000950, 0.00000248, 0.00001335, 0.00001310, -0.00000210, -0.00002523, -0.00004136, -0.00003717, -0.00001096, 0.00002411, 0.00004674, 0.00004047, 0.00000483, -0.00004297, -0.00007738, -0.00007938, -0.00004754, 0.00000158, 0.00004373, 0.00006121, 0.00005198, 0.00002855, 0.00000819, 0.00000161, 0.00000758, 0.00001612, 0.00001714, 0.00000777, -0.00000658, -0.00001732, -0.00001959, -0.00001542, -0.00001102, -0.00001079, -0.00001309, -0.00001138, 0.00000000, 0.00001996, 0.00003959, 0.00004721, 0.00003636, 0.00001097, -0.00001626, -0.00003145, -0.00002809, -0.00001077, 0.00000840, 0.00001799, 0.00001438, },
		{ -0.00000665, 0.00000334, 0.00000651, 0.00000105, -0.00000839, -0.00001526, -0.00001639, -0.00001382, -0.00001206, -0.00001339, -0.00001550, -0.00001363, -0.00000555, 0.00000501, 0.00000987, 0.00000258, -0.00001535, -0.00003335, -0.00003770, -0.00002114, 0.00001058, 0.00004061, 0.00005098, 0.00003463, 0.00000124, -0.00002774, -0.00003233, -0.00000705, 0.00003468, 0.00006825, 0.00007277, 0.00004381, -0.00000422, -0.00004720, -0.00006587, -0.00005648, -0.00003099, -0.00000776, 0.00000047, -0.00000670, -0.00001965, -0.00002734, -0.00002530, -0.00001740, -0.00001089, -0.00000929, -0.00000909, -0.00000289, 0.00001348, 0.00003571, 0.00005229, 0.00005187, 0.00003165, 0.00000067, -0.00002494, -0.00003182, -0.00001771, 0.00000727, 0.00002737, 0.00003126, 0.00001864, -0.00000070, -0.00001432, -0.00001545, },
		{ 0.00000271, -0.00000826, -0.00001659, -0.00001849, -0.00001564, -0.00001215, -0.00001014, -0.00000810, -0.00000319, 0.00000466, 0.00001098, 0.00000951, -0.00000223, -0.00001903, -0.00002980, -0.00002462, -0.00000261, 0.00002553, 0.00004314, 0.00003828, 0.00001243, -0.00001933, -0.00003757, -0.00003141, -0.00000642, 0.00001841, 0.00002306, 0.00000035, -0.00003826, -0.00006899, -0.00007054, -0.00003765, 0.00001532, 0.00006303, 0.00008429, 0.00007382, 0.00004340, 0.00001261, -0.00000392, -0.00000416, 0.00000336, 0.00000844, 0.00000780, 0.00000682, 0.00001365, 0.00003060, 0.00005014, 0.00005887, 0.00004711, 0.00001696, -0.00001729, -0.00003745, -0.00003291, -0.00000768, 0.00002204, 0.00003811, 0.00003164, 0.00000792, -0.00001801, -0.00003151, -0.00002735, -0.00001170, 0.00000341, 0.00000862, },
		{ -0.00001921, -0.00001363, -0.00000680, -0.00000185, 0.00000178, 0.00000597, 0.00001044, 0.00001173, 0.00000612, -0.00000589, -0.00001776, -0.00002022, -0.00000800, 0.00001493, 0.00003626, 0.00004268, 0.00002924, 0.00000374, -0.00001743, -0.00001998, -0.00000171, 0.00002542, 0.00004320, 0.00003995, 0.00001936, -0.00000123, -0.00000256, 0.00002253, 0.00006251, 0.00009299, 0.00009172, 0.00005291, -0.00000851, -0.00006560, -0.00009528, -0.00009102, -0.00006443, -0.00003537, -0.00001820, -0.00001359, -0.00001120, -0.00000027, 0.00002017, 0.00003934, 0.00004256, 0.00002299, -0.00001148, -0.00004179, -0.00004956, -0.00002941, 0.00000696, 0.00003783, 0.00004509, 0.00002533, -0.00000868, -0.00003673, -0.00004403, -0.00002958, -0.00000515, 0.00001356, 0.00001705, 0.00000666, -0.00000835, -0.00001831, },
		{ 0.00000244, 0.00000401, 0.00000435, 0.00000426, 0.00000179, -0.00000507, -0.00001482, -0.00002167, -0.00001925, -0.00000609, 0.00001148, 0.00002237, 0.00001809, -0.00000051, -0.00002258, -0.00003405, -0.00002731, -0.00000720, 0.00001176, 0.00001528, -0.00000028, -0.00002483, -0.00004145, -0.00003884, -0.00002053, -0.00000361, -0.00000721, -0.00003821, -0.00008428, -0.00011977, -0.00012132, -0.00008296, -0.00002062, 0.00003730, 0.00006715, 0.00006376, 0.00004168, 0.00002343, 0.00002342, 0.00003840, 0.00005114, 0.00004425, 0.00001408, -0.00002519, -0.00005043, -0.00004534, -0.00001210, 0.00002917, 0.00005292, 0.00004433, 0.00000845, -0.00003359, -0.00005821, -0.00005380, -0.00002661, 0.00000486, 0.00002245, 0.00001905, 0.00000105, -0.00001800, -0.00002704, -0.00002359, -0.00001307, -0.00000302, },
		{ -0.00001454, -0.00001749, -0.00002229, -0.00002821, -0.00003111, -0.00002691, -0.00001598, -0.00000446, -0.00000066, -0.00000862, -0.00002405, -0.00003629, -0.00003557, -0.00002048, 0.00000008, 0.00001261, 0.00000868, -0.00000851, -0.00002585, -0.00002965, -0.00001603, 0.00000557, 0.00001883, 0.00001296, -0.00000801, -0.00002643, -0.00002286, 0.00000907, 0.00005558, 0.00008940, 0.00008659, 0.00004294, -0.00002166, -0.00007439, -0.00008956, -0.00006447, -0.00002063, 0.00001076, 0.00000938, -0.00002121, -0.00005665, -0.00006885, -0.00004524, 0.00000241, 0.00004563, 0.00005777, 0.00003121, -0.00001806, -0.00006110, -0.00007426, -0.00005313, -0.00001355, 0.00001986, 0.00002944, 0.00001401, -0.00001295, -0.00003390, -0.00003854, -0.00002853, -0.00001368, -0.00000376, -0.00000234, -0.00000650, -0.00001133, },
		{ -0.00001433, -0.00001539, -0.00001212, -0.00000464, 0.00000239, 0.00000333, -0.00000328, -0.00001247, -0.00001584, -0.00000821, 0.00000760, 0.00002171, 0.00002400, 0.00001188, -0.00000706, -0.00002001, -0.00001835, -0.00000453, 0.00000916, 0.00000961, -0.00000677, -0.00003032, -0.00004476, -0.00003976, -0.00002038, -0.00000552, -0.00001490, -0.00005321, -0.00010282, -0.00013173, -0.00011317, -0.00004449, 0.00004769, 0.00012084, 0.00014094, 0.00010289, 0.00003364, -0.00002429, -0.00003875, -0.00000693, 0.00004443, 0.00007728, 0.00006681, 0.00001656, -0.00004502, -0.00008348, -0.00007988, -0.00004118, 0.00000611, 0.00003383, 0.00002917, 0.00000027, -0.00003179, -0.00004779, -0.00004178, -0.00002216, -0.00000350, 0.00000397, -0.00000018, -0.00000894, -0.00001475, -0.00001517, -0.00001308, -0.00001253, },
		{ -0.00001225, -0.00000833, -0.00000754, -0.00001101, -0.00001480, -0.00001322, -0.00000421, 0.00000776, 0.00001421, 0.00000918, -0.00000550, -0.00002057, -0.00002572, -0.00001758, -0.00000278, 0.00000651, 0.00000195, -0.00001404, -0.00002929, -0.00003103, -0.00001628, 0.00000450, 0.00001463, 0.00000478, -0.00001716, -0.00002838, -0.00000617, 0.00005370, 0.00012836, 0.00017692, 0.00016381, 0.00008223, -0.00003761, -0.00014250, -0.00018521, -0.00015129, -0.00006701, 0.00001648, 0.00005335, 0.00002887, -0.00003445, -0.00009321, -0.00011079, -0.00007864, -0.00001868, 0.00003302, 0.00004958, 0.00002832, -0.00001084, -0.00004073, -0.00004467, -0.00002497, 0.00000185, 0.00001819, 0.00001680, 0.00000338, -0.00001004, -0.00001472, -0.00001059, -0.00000428, -0.00000248, -0.00000657, -0.00001237, -0.00001473, },
		{ -0.00000870, -0.00000727, -0.00000296, 0.00000517, 0.00001298, 0.00001424, 0.00000620, -0.00000706, -0.00001685, -0.00001634, -0.00000631, 0.00000471, 0.00000666, -0.00000404, -0.00002107, -0.00003257, -0.00003048, -0.00001768, -0.00000665, -0.00000980, -0.00002838, -0.00004893, -0.00005180, -0.00002670, 0.00001558, 0.00004549, 0.00003219, -0.00003400, -0.00012942, -0.00020525, -0.00021384, -0.00013787, -0.00000412, 0.00012816, 0.00019876, 0.00017950, 0.00008906, -0.00001919, -0.00008855, -0.00009043, -0.00003721, 0.00002971, 0.00006752, 0.00005726, 0.00001165, -0.00003719, -0.00006033, -0.00004938, -0.00001845, 0.00000851, 0.00001471, 0.00000019, -0.00002126, -0.00003385, -0.00003096, -0.00001820, -0.00000725, -0.00000643, -0.00001512, -0.00002564, -0.00003010, -0.00002624, -0.00001809, -0.00001146, },
		{ 0.00001125, 0.00001283, 0.00000906, -0.00000119, -0.00001337, -0.00002030, -0.00001803, -0.00000974, -0.00000385, -0.00000752, -0.00002056, -0.00003486, -0.00004048, -0.00003384, -0.00002130, -0.00001453, -0.00002067, -0.00003485, -0.00004196, -0.00002778, 0.00000808, 0.00004777, 0.00006462, 0.00004184, -0.00001218, -0.00006391, -0.00007315, -0.00001956, 0.00007954, 0.00017373, 0.00020592, 0.00014605, 0.00001135, -0.00014028, -0.00024090, -0.00024937, -0.00017263, -0.00005908, 0.00003092, 0.00005905, 0.00002709, -0.00003001, -0.00006968, -0.00006814, -0.00003175, 0.00001198, 0.00003538, 0.00002848, 0.00000277, -0.00001981, -0.00002329, -0.00000774, 0.00001312, 0.00002333, 0.00001621, -0.00000211, -0.00001865, -0.00002339, -0.00001561, -0.00000258, 0.00000718, 0.00001019, 0.00000920, 0.00000900, },
		{ -0.00001125, -0.00001933, -0.00002212, -0.00001735, -0.00000914, -0.00000512, -0.00001011, -0.00002169, -0.00003162, -0.00003245, -0.00002393, -0.00001378, -0.00001141, -0.00001935, -0.00002943, -0.00002811, -0.00000775, 0.00002447, 0.00004813, 0.00004242, 0.00000228, -0.00005418, -0.00009411, -0.00009084, -0.00004385, 0.00001671, 0.00004629, 0.00001452, -0.00007202, -0.00016860, -0.00021589, -0.00017488, -0.00005196, 0.00010202, 0.00021900, 0.00025109, 0.00019572, 0.00009395, 0.00000401, -0.00003255, -0.00001164, 0.00003698, 0.00007301, 0.00007195, 0.00003794, -0.00000351, -0.00002556, -0.00001825, 0.00000691, 0.00002745, 0.00002656, 0.00000400, -0.00002559, -0.00004423, -0.00004289, -0.00002606, -0.00000649, 0.00000472, 0.00000501, -0.00000019, -0.00000376, -0.00000316, -0.00000160, -0.00000393, },
		{ -0.00001056, -0.00000336, -0.00000108, -0.00000722, -0.00001822, -0.00002612, -0.00002517, -0.00001687, -0.00000883, -0.00000804, -0.00001390, -0.00001741, -0.00000815, 0.00001576, 0.00004288, 0.00005413, 0.00003627, -0.00000618, -0.00005016, -0.00006709, -0.00004209, 0.00001388, 0.00006873, 0.00008939, 0.00006482, 0.00001579, -0.00001680, 0.00000106, 0.00007132, 0.00015800, 0.00020620, 0.00017533, 0.00006572, -0.00007848, -0.00019310, -0.00022989, -0.00018282, -0.00008887, -0.00000393, 0.00003054, 0.00000909, -0.00003987, -0.00007640, -0.00007572, -0.00004184, -0.00000110, 0.00001813, 0.00000465, -0.00002992, -0.00006122, -0.00006944, -0.00005155, -0.00002083, 0.00000426, 0.00001273, 0.00000660, -0.00000356, -0.00000812, -0.00000528, -0.00000055, -0.00000047, -0.00000637, -0.00001346, -0.00001546, },
		{ -0.00000968, -0.00002207, -0.00003058, -0.00003008, -0.00002290, -0.00001651, -0.00001661, -0.00002139, -0.00002255, -0.00001284, 0.00000582, 0.00002073, 0.00001625, -0.00001394, -0.00005912, -0.00009570, -0.00010173, -0.00007252, -0.00002590, 0.00000798, 0.00000625, -0.00003066, -0.00007863, -0.00010622, -0.00009729, -0.00006378, -0.00003867, -0.00005262, -0.00010975, -0.00017979, -0.00021384, -0.00017487, -0.00006364, 0.00007752, 0.00018671, 0.00021726, 0.00016395, 0.00006441, -0.00002459, -0.00006067, -0.00003830, 0.00001281, 0.00005027, 0.00004689, 0.00000578, -0.00004534, -0.00007486, -0.00006785, -0.00003343, 0.00000439, 0.00002413, 0.00001995, 0.00000231, -0.00001242, -0.00001443, -0.00000606, 0.00000256, 0.00000296, -0.00000497, -0.00001394, -0.00001626, -0.00001065, -0.00000317, -0.00000184, },
		{ -0.00005902, -0.00005088, -0.00004340, -0.00004243, -0.00004596, -0.00004623, -0.00003758, -0.00002338, -0.00001564, -0.00002641, -0.00005709, -0.00009422, -0.00011626, -0.00010774, -0.00007084, -0.00002557, 0.00000278, 0.00000069, -0.00002412, -0.00004794, -0.00004784, -0.00001826, 0.00002506, 0.00005699, 0.00006264, 0.00005000, 0.00004528, 0.00007285, 0.00013345, 0.00019703, 0.00021759, 0.00016213, 0.00003528, -0.00011756, -0.00023248, -0.00026220, -0.00020260, -0.00009417, 0.00000239, 0.00004105, 0.00001455, -0.00004629, -0.00009555, -0.00010149, -0.00006370, -0.00000862, 0.00003093, 0.00003721, 0.00001620, -0.00001061, -0.00002328, -0.00001660, -0.00000088, 0.00000841, 0.00000310, -0.00001239, -0.00002618, -0.00002925, -0.00002249, -0.00001529, -0.00001735, -0.00003065, -0.00004805, -0.00005925, },
		{ -0.00001843, -0.00002103, -0.00002170, -0.00001438, -0.00000259, 0.00000216, -0.00001044, -0.00003968, -0.00007168, -0.00008735, -0.00007558, -0.00004170, -0.00000472, 0.00001537, 0.00001132, -0.00000669, -0.00001899, -0.00001134, 0.00001428, 0.00004119, 0.00005071, 0.00003660, 0.00000972, -0.00001114, -0.00001561, -0.00001243, -0.00002475, -0.00007157, -0.00014806, -0.00022015, -0.00024025, -0.00017636, -0.00003635, 0.00012975, 0.00025322, 0.00028323, 0.00021469, 0.00009016, -0.00002514, -0.00007995, -0.00006291, -0.00000330, 0.00005185, 0.00006798, 0.00004143, -0.00000389, -0.00003676, -0.00004032, -0.00002082, 0.00000057, 0.00000511, -0.00001032, -0.00003308, -0.00004624, -0.00004202, -0.00002694, -0.00001572, -0.00001926, -0.00003631, -0.00005498, -0.00006233, -0.00005405, -0.00003699, -0.00002281, },
		{ -0.00003105, -0.00002253, -0.00002048, -0.00003600, -0.00006799, -0.00010190, -0.00011888, -0.00010861, -0.00007647, -0.00003958, -0.00001472, -0.00000704, -0.00000794, -0.00000335, 0.00001433, 0.00003982, 0.00005920, 0.00006123, 0.00004673, 0.00002823, 0.00002009, 0.00002673, 0.00003909, 0.00004298, 0.00003331, 0.00002258, 0.00003487, 0.00008715, 0.00017057, 0.00024647, 0.00026362, 0.00018854, 0.00003028, -0.00015755, -0.00030142, -0.00034605, -0.00028373, -0.00015703, -0.00003332, 0.00003299, 0.00002777, -0.00002145, -0.00006840, -0.00007906, -0.00005043, -0.00000701, 0.00002020, 0.00001588, -0.00001098, -0.00003668, -0.00004164, -0.00002438, -0.00000093, 0.00000852, -0.00000488, -0.00003255, -0.00005568, -0.00005949, -0.00004360, -0.00002118, -0.00000810, -0.00001103, -0.00002355, -0.00003282, },
		{ -0.00004588, -0.00008148, -0.00011981, -0.00013843, -0.00012505, -0.00008473, -0.00003520, 0.00000557, 0.00003001, 0.00004274, 0.00005320, 0.00006603, 0.00007701, 0.00007757, 0.00006336, 0.00003950, 0.00001758, 0.00000679, 0.00000654, 0.00000712, -0.00000167, -0.00002079, -0.00004000, -0.00004652, -0.00003836, -0.00003127, -0.00005122, -0.00011494, -0.00021137, -0.00029858, -0.00032243, -0.00024801, -0.00008505, 0.00011046, 0.00026199, 0.00031216, 0.00025287, 0.00012790, 0.00000625, -0.00005704, -0.00004890, 0.00000085, 0.00004407, 0.00004687, 0.00000881, -0.00004189, -0.00007153, -0.00006515, -0.00003520, -0.00001008, -0.00001245, -0.00004312, -0.00008172, -0.00010250, -0.00009292, -0.00006176, -0.00003139, -0.00002097, -0.00003315, -0.00005382, -0.00006438, -0.00005668, -0.00003938, -0.00003112, },
		{ -0.00017045, -0.00015188, -0.00009627, -0.00002708, 0.00003161, 0.00006741, 0.00008204, 0.00008437, 0.00008128, 0.00007350, 0.00005869, 0.00003736, 0.00001564, 0.00000180, -0.00000027, 0.00000529, 0.00000930, 0.00000526, -0.00000475, -0.00001145, -0.00000664, 0.00000841, 0.00002224, 0.00002242, 0.00000860, -0.00000177, 0.00001785, 0.00008299, 0.00017992, 0.00026379, 0.00027895, 0.00019158, 0.00001513, -0.00018938, -0.00034154, -0.00038332, -0.00030996, -0.00017120, -0.00004177, 0.00002081, 0.00000561, -0.00005210, -0.00009869, -0.00009804, -0.00005235, 0.00000415, 0.00003208, 0.00001477, -0.00003177, -0.00007211, -0.00007725, -0.00004432, 0.00000312, 0.00003311, 0.00002813, -0.00000377, -0.00003669, -0.00004691, -0.00003057, -0.00000625, -0.00000112, -0.00003086, -0.00008763, -0.00014414, },
		{ -0.00000783, 0.00007619, 0.00012120, 0.00013056, 0.00011840, 0.00009831, 0.00007726, 0.00005702, 0.00003887, 0.00002605, 0.00002169, 0.00002495, 0.00002992, 0.00002928, 0.00002011, 0.00000662, -0.00000303, -0.00000382, 0.00000120, 0.00000293, -0.00000587, -0.00002307, -0.00003746, -0.00003846, -0.00002835, -0.00002660, -0.00005939, -0.00013873, -0.00024487, -0.00032681, -0.00032553, -0.00020910, 0.00000177, 0.00023576, 0.00040338, 0.00044288, 0.00035310, 0.00019301, 0.00004760, -0.00001957, 0.00000123, 0.00006719, 0.00011531, 0.00010393, 0.00003685, -0.00004384, -0.00008925, -0.00007713, -0.00002506, 0.00002421, 0.00003342, -0.00000416, -0.00006211, -0.00010144, -0.00009899, -0.00006385, -0.00003076, -0.00003566, -0.00009039, -0.00017347, -0.00024241, -0.00025892, -0.00021030, -0.00011457, },
		{ 0.00019590, 0.00015594, 0.00011464, 0.00008340, 0.00006369, 0.00005304, 0.00004931, 0.00005088, 0.00005464, 0.00005577, 0.00005044, 0.00003928, 0.00002791, 0.00002308, 0.00002728, 0.00003647, 0.00004320, 0.00004303, 0.00003863, 0.00003744, 0.00004424, 0.00005514, 0.00005940, 0.00004983, 0.00003421, 0.00003716, 0.00008671, 0.00019129, 0.00032231, 0.00041770, 0.00040990, 0.00026544, 0.00001233, -0.00026447, -0.00046123, -0.00050809, -0.00040594, -0.00022514, -0.00006564, -0.00000187, -0.00004472, -0.00014253, -0.00021865, -0.00022077, -0.00015090, -0.00005871, -0.00000489, -0.00002008, -0.00008647, -0.00015283, -0.00017147, -0.00013133, -0.00006534, -0.00002786, -0.00005772, -0.00015147, -0.00026360, -0.00033271, -0.00031647, -0.00021389, -0.00006305, 0.00008182, 0.00017819, 0.00021190, },
		{ 0.00008228, 0.00007056, 0.00006892, 0.00006949, 0.00006831, 0.00006335, 0.00005315, 0.00003767, 0.00001978, 0.00000469, -0.00000331, -0.00000432, -0.00000329, -0.00000655, -0.00001695, -0.00003192, -0.00004629, -0.00005761, -0.00006875, -0.00008478, -0.00010602, -0.00012401, -0.00012600, -0.00010728, -0.00008256, -0.00008520, -0.00014940, -0.00028317, -0.00044954, -0.00057266, -0.00057193, -0.00040830, -0.00011575, 0.00020461, 0.00042949, 0.00047526, 0.00034185, 0.00011220, -0.00009362, -0.00018379, -0.00014148, -0.00002625, 0.00006876, 0.00007626, -0.00000626, -0.00012081, -0.00019140, -0.00017586, -0.00009355, -0.00001189, -0.00000176, -0.00009073, -0.00024374, -0.00038213, -0.00042830, -0.00034824, -0.00016765, 0.00004520, 0.00021665, 0.00030227, 0.00030063, 0.00024323, 0.00017117, 0.00011416, },
		{ 0.00004109, 0.00002923, 0.00001192, -0.00000774, -0.00002715, -0.00004339, -0.00005337, -0.00005563, -0.00005214, -0.00004778, -0.00004725, -0.00005180, -0.00005854, -0.00006320, -0.00006382, -0.00006169, -0.00005860, -0.00005289, -0.00003899, -0.00001258, 0.00002208, 0.00004978, 0.00005231, 0.00002430, -0.00001400, -0.00001939, 0.00005146, 0.00020915, 0.00040931, 0.00056031, 0.00056393, 0.00037075, 0.00001838, -0.00037405, -0.00065793, -0.00072876, -0.00058044, -0.00030818, -0.00005645, 0.00005910, 0.00001106, -0.00013284, -0.00025798, -0.00027574, -0.00017688, -0.00003383, 0.00004756, -0.00000417, -0.00018028, -0.00039466, -0.00053208, -0.00051160, -0.00032839, -0.00005225, 0.00021249, 0.00037968, 0.00041739, 0.00035094, 0.00023797, 0.00013410, 0.00006937, 0.00004445, 0.00004249, 0.00004490, },
		{ 0.00005049, 0.00004171, 0.00003845, 0.00004033, 0.00004633, 0.00005387, 0.00005916, 0.00005932, 0.00005436, 0.00004708, 0.00004088, 0.00003721, 0.00003527, 0.00003413, 0.00003466, 0.00003876, 0.00004555, 0.00004823, 0.00003606, 0.00000247, -0.00004577, -0.00008696, -0.00009587, -0.00006412, -0.00001607, -0.00000534, -0.00008693, -0.00027712, -0.00052566, -0.00072346, -0.00075014, -0.00054062, -0.00013317, 0.00033334, 0.00067874, 0.00076977, 0.00058931, 0.00024631, -0.00008214, -0.00024564, -0.00019951, -0.00002342, 0.00013018, 0.00012736, -0.00006921, -0.00037898, -0.00064914, -0.00073845, -0.00059128, -0.00026202, 0.00011896, 0.00041398, 0.00054153, 0.00050275, 0.00036337, 0.00020894, 0.00010156, 0.00005978, 0.00006568, 0.00008743, 0.00010063, 0.00009732, 0.00008242, 0.00006476, },
		{ 0.00005083, 0.00004629, 0.00004130, 0.00003450, 0.00002492, 0.00001330, 0.00000204, -0.00000635, -0.00001085, -0.00001199, -0.00001057, -0.00000658, -0.00000015, 0.00000635, 0.00000776, -0.00000072, -0.00001777, -0.00003248, -0.00002836, 0.00000515, 0.00006054, 0.00011021, 0.00012087, 0.00007906, 0.00001154, -0.00001773, 0.00005969, 0.00026794, 0.00055140, 0.00078069, 0.00080757, 0.00054541, 0.00003140, -0.00057023, -0.00103490, -0.00118540, -0.00098412, -0.00055783, -0.00013703, 0.00006030, -0.00006044, -0.00042426, -0.00082886, -0.00105081, -0.00095753, -0.00056691, -0.00002774, 0.00046182, 0.00074976, 0.00079014, 0.00064299, 0.00042449, 0.00024187, 0.00014869, 0.00013807, 0.00016726, 0.00019197, 0.00018960, 0.00016237, 0.00012580, 0.00009434, 0.00007384, 0.00006246, 0.00005580, },
		{ 0.00005809, 0.00005926, 0.00005970, 0.00006017, 0.00006181, 0.00006500, 0.00006907, 0.00007301, 0.00007602, 0.00007698, 0.00007365, 0.00006311, 0.00004454, 0.00002219, 0.00000529, 0.00000251, 0.00001362, 0.00002469, 0.00001331, -0.00003646, -0.00011825, -0.00019988, -0.00024016, -0.00022004, -0.00016853, -0.00016126, -0.00028245, -0.00056554, -0.00094688, -0.00126809, -0.00133907, -0.00103770, -0.00039439, 0.00039081, 0.00101947, 0.00122733, 0.00091381, 0.00020011, -0.00061786, -0.00120135, -0.00132010, -0.00094849, -0.00026174, 0.00045921, 0.00096067, 0.00111924, 0.00096988, 0.00065729, 0.00034756, 0.00015058, 0.00008849, 0.00011452, 0.00015962, 0.00017516, 0.00015059, 0.00010447, 0.00006262, 0.00004022, 0.00003692, 0.00004329, 0.00005017, 0.00005386, 0.00005533, 0.00005655, },
		{ 0.00017918, 0.00015609, 0.00013342, 0.00011150, 0.00009002, 0.00006849, 0.00004622, 0.00002225, -0.00000397, -0.00003089, -0.00005383, -0.00006663, -0.00006635, -0.00005794, -0.00005413, -0.00006800, -0.00010177, -0.00013998, -0.00015518, -0.00012613, -0.00005778, 0.00001316, 0.00003745, -0.00000897, -0.00009171, -0.00011690, 0.00002431, 0.00038114, 0.00087925, 0.00131310, 0.00141559, 0.00098355, -0.00000040, -0.00131738, -0.00257095, -0.00333834, -0.00334669, -0.00259021, -0.00133091, 0.00001658, 0.00105858, 0.00157126, 0.00155941, 0.00120957, 0.00077453, 0.00045385, 0.00032656, 0.00035578, 0.00044559, 0.00050866, 0.00050750, 0.00045603, 0.00039277, 0.00034952, 0.00033482, 0.00033707, 0.00033877, 0.00032942, 0.00030928, 0.00028480, 0.00026174, 0.00024151, 0.00022218, 0.00020155, },
		{ -0.00003249, -0.00004182, -0.00004876, -0.00005456, -0.00006020, -0.00006593, -0.00007059, -0.00007158, -0.00006669, -0.00005721, -0.00004971, -0.00005380, -0.00007557, -0.00011100, -0.00014550, -0.00016269, -0.00015860, -0.00015074, -0.00017207, -0.00024925, -0.00037779, -0.00051384, -0.00059676, -0.00059654, -0.00055902, -0.00061354, -0.00092212, -0.00158312, -0.00253608, -0.00352642, -0.00416700, -0.00408468, -0.00309088, -0.00129282, 0.00091583, 0.00298970, 0.00443451, 0.00498599, 0.00468573, 0.00382674, 0.00280476, 0.00195224, 0.00143065, 0.00121841, 0.00117981, 0.00116449, 0.00108280, 0.00092710, 0.00074423, 0.00058827, 0.00048549, 0.00042729, 0.00038644, 0.00033944, 0.00027949, 0.00021471, 0.00015755, 0.00011487, 0.00008511, 0.00006200, 0.00004022, 0.00001825, -0.00000236, -0.00001954, },
		{ -0.00030959, -0.00031172, -0.00031770, -0.00032629, -0.00033574, -0.00034513, -0.00035602, -0.00037266, -0.00039957, -0.00043698, -0.00047801, -0.00051114, -0.00052829, -0.00053331, -0.00054317, -0.00057748, -0.00064113, -0.00071256, -0.00075046, -0.00072075, -0.00062928, -0.00053505, -0.00052430, -0.00064841, -0.00085630, -0.00096671, -0.00071413, 0.00013550, 0.00164670, 0.00362566, 0.00563858, 0.00714663, 0.00770638, 0.00714663, 0.00563858, 0.00362566, 0.00164670, 0.00013550, -0.00071413, -0.00096671, -0.00085630, -0.00064841, -0.00052430, -0.00053505, -0.00062928, -0.00072075, -0.00075046, -0.00071256, -0.00064113, -0.00057748, -0.00054317, -0.00053331, -0.00052829, -0.00051114, -0.00047801, -0.00043698, -0.00039957, -0.00037266, -0.00035602, -0.00034513, -0.00033574, -0.00032629, -0.00031770, -0.00031172, },
		{ -0.00003249, -0.00001954, -0.00000236, 0.00001825, 0.00004022, 0.00006200, 0.00008511, 0.00011487, 0.00015755, 0.00021471, 0.00027949, 0.00033944, 0.00038644, 0.00042729, 0.00048549, 0.00058827, 0.00074423, 0.00092710, 0.00108280, 0.00116449, 0.00117981, 0.00121841, 0.00143065, 0.00195224, 0.00280476, 0.00382674, 0.00468573, 0.00498599, 0.00443451, 0.00298970, 0.00091583, -0.00129282, -0.00309088, -0.00408468, -0.00416700, -0.00352642, -0.00253608, -0.00158312, -0.00092212, -0.00061354, -0.00055902, -0.00059654, -0.00059676, -0.00051384, -0.00037779, -0.00024925, -0.00017207, -0.00015074, -0.00015860, -0.00016269, -0.00014550, -0.00011100, -0.00007557, -0.00005380, -0.00004971, -0.00005721, -0.00006669, -0.00007158, -0.00007059, -0.00006593, -0.00006020, -0.00005456, -0.00004876, -0.00004182, },
		{ 0.00017918, 0.00020155, 0.00022218, 0.00024151, 0.00026174, 0.00028480, 0.00030928, 0.00032942, 0.00033877, 0.00033707, 0.00033482, 0.00034952, 0.00039277, 0.00045603, 0.00050750, 0.00050866, 0.00044559, 0.00035578, 0.00032656, 0.00045385, 0.00077453, 0.00120957, 0.00155941, 0.00157126, 0.00105858, 0.00001658, -0.00133091, -0.00259021, -0.00334669, -0.00333834, -0.00257095, -0.00131738, -0.00000040, 0.00098355, 0.00141559, 0.00131310, 0.00087925, 0.00038114, 0.00002431, -0.00011690, -0.00009171, -0.00000897, 0.00003745, 0.00001316, -0.00005778, -0.00012613, -0.00015518, -0.00013998, -0.00010177, -0.00006800, -0.00005413, -0.00005794, -0.00006635, -0.00006663, -0.00005383, -0.00003089, -0.00000397, 0.00002225, 0.00004622, 0.00006849, 0.00009002, 0.00011150, 0.00013342, 0.00015609, },
		{ 0.00005809, 0.00005655, 0.00005533, 0.00005386, 0.00005017, 0.00004329, 0.00003692, 0.00004022, 0.00006262, 0.00010447, 0.00015059, 0.00017516, 0.00015962, 0.00011452, 0.00008849, 0.00015058, 0.00034756, 0.00065729, 0.00096988, 0.00111924, 0.00096067, 0.00045921, -0.00026174, -0.00094849, -0.00132010, -0.00120135, -0.00061786, 0.00020011, 0.00091381, 0.00122733, 0.00101947, 0.00039081, -0.00039439, -0.00103770, -0.00133907, -0.00126809, -0.00094688, -0.00056554, -0.00028245, -0.00016126, -0.00016853, -0.00022004, -0.00024016, -0.00019988, -0.00011825, -0.00003646, 0.00001331, 0.00002469, 0.00001362, 0.00000251, 0.00000529, 0.00002219, 0.00004454, 0.00006311, 0.00007365, 0.00007698, 0.00007602, 0.00007301, 0.00006907, 0.00006500, 0.00006181, 0.00006017, 0.00005970, 0.00005926, },
		{ 0.00005083, 0.00005580, 0.00006246, 0.00007384, 0.00009434, 0.00012580, 0.00016237, 0.00018960, 0.00019197, 0.00016726, 0.00013807, 0.00014869, 0.00024187, 0.00042449, 0.00064299, 0.00079014, 0.00074976, 0.00046182, -0.00002774, -0.00056691, -0.00095753, -0.00105081, -0.00082886, -0.00042426, -0.00006044, 0.00006030, -0.00013703, -0.00055783, -0.00098412, -0.00118540, -0.00103490, -0.00057023, 0.00003140, 0.00054541, 0.00080757, 0.00078069, 0.00055140, 0.00026794, 0.00005969, -0.00001773, 0.00001154, 0.00007906, 0.00012087, 0.00011021, 0.00006054, 0.00000515, -0.00002836, -0.00003248, -0.00001777, -0.00000072, 0.00000776, 0.00000635, -0.00000015, -0.00000658, -0.00001057, -0.00001199, -0.00001085, -0.00000635, 0.00000204, 0.00001330, 0.00002492, 0.00003450, 0.00004130, 0.00004629, },
		{ 0.00005049, 0.00006476, 0.00008242, 0.00009732, 0.00010063, 0.00008743, 0.00006568, 0.00005978, 0.00010156, 0.00020894, 0.00036337, 0.00050275, 0.00054153, 0.00041398, 0.00011896, -0.00026202, -0.00059128, -0.00073845, -0.00064914, -0.00037898, -0.00006921, 0.00012736, 0.00013018, -0.00002342, -0.00019951, -0.00024564, -0.00008214, 0.00024631, 0.00058931, 0.00076977, 0.00067874, 0.00033334, -0.00013317, -0.00054062, -0.00075014, -0.00072346, -0.00052566, -0.00027712, -0.00008693, -0.00000534, -0.00001607, -0.00006412, -0.00009587, -0.00008696, -0.00004577, 0.00000247, 0.00003606, 0.00004823, 0.00004555, 0.00003876, 0.00003466, 0.00003413, 0.00003527, 0.00003721, 0.00004088, 0.00004708, 0.00005436, 0.00005932, 0.00005916, 0.00005387, 0.00004633, 0.00004033, 0.00003845, 0.00004171, },
		{ 0.00004109, 0.00004490, 0.00004249, 0.00004445, 0.00006937, 0.00013410, 0.00023797, 0.00035094, 0.00041739, 0.00037968, 0.00021249, -0.00005225, -0.00032839, -0.00051160, -0.00053208, -0.00039466, -0.00018028, -0.00000417, 0.00004756, -0.00003383, -0.00017688, -0.00027574, -0.00025798, -0.00013284, 0.00001106, 0.00005910, -0.00005645, -0.00030818, -0.00058044, -0.00072876, -0.00065793, -0.00037405, 0.00001838, 0.00037075, 0.00056393, 0.00056031, 0.00040931, 0.00020915, 0.00005146, -0.00001939, -0.00001400, 0.00002430, 0.00005231, 0.00004978, 0.00002208, -0.00001258, -0.00003899, -0.00005289, -0.00005860, -0.00006169, -0.00006382, -0.00006320, -0.00005854, -0.00005180, -0.00004725, -0.00004778, -0.00005214, -0.00005563, -0.00005337, -0.00004339, -0.00002715, -0.00000774, 0.00001192, 0.00002923, },
		{ 0.00008228, 0.00011416, 0.00017117, 0.00024323, 0.00030063, 0.00030227, 0.00021665, 0.00004520, -0.00016765, -0.00034824, -0.00042830, -0.00038213, -0.00024374, -0.00009073, -0.00000176, -0.00001189, -0.00009355, -0.00017586, -0.00019140, -0.00012081, -0.00000626, 0.00007626, 0.00006876, -0.00002625, -0.00014148, -0.00018379, -0.00009362, 0.00011220, 0.00034185, 0.00047526, 0.00042949, 0.00020461, -0.00011575, -0.00040830, -0.00057193, -0.00057266, -0.00044954, -0.00028317, -0.00014940, -0.00008520, -0.00008256, -0.00010728, -0.00012600, -0.00012401, -0.00010602, -0.00008478, -0.00006875, -0.00005761, -0.00004629, -0.00003192, -0.00001695, -0.00000655, -0.00000329, -0.00000432, -0.00000331, 0.00000469, 0.00001978, 0.00003767, 0.00005315, 0.00006335, 0.00006831, 0.00006949, 0.00006892, 0.00007056, },
		{ 0.00019590, 0.00021190, 0.00017819, 0.00008182, -0.00006305, -0.00021389, -0.00031647, -0.00033271, -0.00026360, -0.00015147, -0.00005772, -0.00002786, -0.00006534, -0.00013133, -0.00017147, -0.00015283, -0.00008647, -0.00002008, -0.00000489, -0.00005871, -0.00015090, -0.00022077, -0.00021865, -0.00014253, -0.00004472, -0.00000187, -0.00006564, -0.00022514, -0.00040594, -0.00050809, -0.00046123, -0.00026447, 0.00001233, 0.00026544, 0.00040990, 0.00041770, 0.00032231, 0.00019129, 0.00008671, 0.00003716, 0.00003421, 0.00004983, 0.00005940, 0.00005514, 0.00004424, 0.00003744, 0.00003863, 0.00004303, 0.00004320, 0.00003647, 0.00002728, 0.00002308, 0.00002791, 0.00003928, 0.00005044, 0.00005577, 0.00005464, 0.00005088, 0.00004931, 0.00005304, 0.00006369, 0.00008340, 0.00011464, 0.00015594, },
		{ -0.00000783, -0.00011457, -0.00021030, -0.00025892, -0.00024241, -0.00017347, -0.00009039, -0.00003566, -0.00003076, -0.00006385, -0.00009899, -0.00010144, -0.00006211, -0.00000416, 0.00003342, 0.00002421, -0.00002506, -0.00007713, -0.00008925, -0.00004384, 0.00003685, 0.00010393, 0.00011531, 0.00006719, 0.00000123, -0.00001957, 0.00004760, 0.00019301, 0.00035310, 0.00044288, 0.00040338, 0.00023576, 0.00000177, -0.00020910, -0.00032553, -0.00032681, -0.00024487, -0.00013873, -0.00005939, -0.00002660, -0.00002835, -0.00003846, -0.00003746, -0.00002307, -0.00000587, 0.00000293, 0.00000120, -0.00000382, -0.00000303, 0.00000662, 0.00002011, 0.00002928, 0.00002992, 0.00002495, 0.00002169, 0.00002605, 0.00003887, 0.00005702, 0.00007726, 0.00009831, 0.00011840, 0.00013056, 0.00012120, 0.00007619, },
		{ -0.00017045, -0.00014414, -0.00008763, -0.00003086, -0.00000112, -0.00000625, -0.00003057, -0.00004691, -0.00003669, -0.00000377, 0.00002813, 0.00003311, 0.00000312, -0.00004432, -0.00007725, -0.00007211, -0.00003177, 0.00001477, 0.00003208, 0.00000415, -0.00005235, -0.00009804, -0.00009869, -0.00005210, 0.00000561, 0.00002081, -0.00004177, -0.00017120, -0.00030996, -0.00038332, -0.00034154, -0.00018938, 0.00001513, 0.00019158, 0.00027895, 0.00026379, 0.00017992, 0.00008299, 0.00001785, -0.00000177, 0.00000860, 0.00002242, 0.00002224, 0.00000841, -0.00000664, -0.00001145, -0.00000475, 0.00000526, 0.00000930, 0.00000529, -0.00000027, 0.00000180, 0.00001564, 0.00003736, 0.00005869, 0.00007350, 0.00008128, 0.00008437, 0.00008204, 0.00006741, 0.00003161, -0.00002708, -0.00009627, -0.00015188, },
		{ -0.00004588, -0.00003112, -0.00003938, -0.00005668, -0.00006438, -0.00005382, -0.00003315, -0.00002097, -0.00003139, -0.00006176, -0.00009292, -0.00010250, -0.00008172, -0.00004312, -0.00001245, -0.00001008, -0.00003520, -0.00006515, -0.00007153, -0.00004189, 0.00000881, 0.00004687, 0.00004407, 0.00000085, -0.00004890, -0.00005704, 0.00000625, 0.00012790, 0.00025287, 0.00031216, 0.00026199, 0.00011046, -0.00008505, -0.00024801, -0.00032243, -0.00029858, -0.00021137, -0.00011494, -0.00005122, -0.00003127, -0.00003836, -0.00004652, -0.00004000, -0.00002079, -0.00000167, 0.00000712, 0.00000654, 0.00000679, 0.00001758, 0.00003950, 0.00006336, 0.00007757, 0.00007701, 0.00006603, 0.00005320, 0.00004274, 0.00003001, 0.00000557, -0.00003520, -0.00008473, -0.00012505, -0.00013843, -0.00011981, -0.00008148, },
		{ -0.00003105, -0.00003282, -0.00002355, -0.00001103, -0.00000810, -0.00002118, -0.00004360, -0.00005949, -0.00005568, -0.00003255, -0.00000488, 0.00000852, -0.00000093, -0.00002438, -0.00004164, -0.00003668, -0.00001098, 0.00001588, 0.00002020, -0.00000701, -0.00005043, -0.00007906, -0.00006840, -0.00002145, 0.00002777, 0.00003299, -0.00003332, -0.00015703, -0.00028373, -0.00034605, -0.00030142, -0.00015755, 0.00003028, 0.00018854, 0.00026362, 0.00024647, 0.00017057, 0.00008715, 0.00003487, 0.00002258, 0.00003331, 0.00004298, 0.00003909, 0.00002673, 0.00002009, 0.00002823, 0.00004673, 0.00006123, 0.00005920, 0.00003982, 0.00001433, -0.00000335, -0.00000794, -0.00000704, -0.00001472, -0.00003958, -0.00007647, -0.00010861, -0.00011888, -0.00010190, -0.00006799, -0.00003600, -0.00002048, -0.00002253, },
		{ -0.00001843, -0.00002281, -0.00003699, -0.00005405, -0.00006233, -0.00005498, -0.00003631, -0.00001926, -0.00001572, -0.00002694, -0.00004202, -0.00004624, -0.00003308, -0.00001032, 0.00000511, 0.00000057, -0.00002082, -0.00004032, -0.00003676, -0.00000389, 0.00004143, 0.00006798, 0.00005185, -0.00000330, -0.00006291, -0.00007995, -0.00002514, 0.00009016, 0.00021469, 0.00028323, 0.00025322, 0.00012975, -0.00003635, -0.00017636, -0.00024025, -0.00022015, -0.00014806, -0.00007157, -0.00002475, -0.00001243, -0.00001561, -0.00001114, 0.00000972, 0.00003660, 0.00005071, 0.00004119, 0.00001428, -0.00001134, -0.00001899, -0.00000669, 0.00001132, 0.00001537, -0.00000472, -0.00004170, -0.00007558, -0.00008735, -0.00007168, -0.00003968, -0.00001044, 0.00000216, -0.00000259, -0.00001438, -0.00002170, -0.00002103, },
		{ -0.00005902, -0.00005925, -0.00004805, -0.00003065, -0.00001735, -0.00001529, -0.00002249, -0.00002925, -0.00002618, -0.00001239, 0.00000310, 0.00000841, -0.00000088, -0.00001660, -0.00002328, -0.00001061, 0.00001620, 0.00003721, 0.00003093, -0.00000862, -0.00006370, -0.00010149, -0.00009555, -0.00004629, 0.00001455, 0.00004105, 0.00000239, -0.00009417, -0.00020260, -0.00026220, -0.00023248, -0.00011756, 0.00003528, 0.00016213, 0.00021759, 0.00019703, 0.00013345, 0.00007285, 0.00004528, 0.00005000, 0.00006264, 0.00005699, 0.00002506, -0.00001826, -0.00004784, -0.00004794, -0.00002412, 0.00000069, 0.00000278, -0.00002557, -0.00007084, -0.00010774, -0.00011626, -0.00009422, -0.00005709, -0.00002641, -0.00001564, -0.00002338, -0.00003758, -0.00004623, -0.00004596, -0.00004243, -0.00004340, -0.00005088, },
		{ -0.00000968, -0.00000184, -0.00000317, -0.00001065, -0.00001626, -0.00001394, -0.00000497, 0.00000296, 0.00000256, -0.00000606, -0.00001443, -0.00001242, 0.00000231, 0.00001995, 0.00002413, 0.00000439, -0.00003343, -0.00006785, -0.00007486, -0.00004534, 0.00000578, 0.00004689, 0.00005027, 0.00001281, -0.00003830, -0.00006067, -0.00002459, 0.00006441, 0.00016395, 0.00021726, 0.00018671, 0.00007752, -0.00006364, -0.00017487, -0.00021384, -0.00017979, -0.00010975, -0.00005262, -0.00003867, -0.00006378, -0.00009729, -0.00010622, -0.00007863, -0.00003066, 0.00000625, 0.00000798, -0.00002590, -0.00007252, -0.00010173, -0.00009570, -0.00005912, -0.00001394, 0.00001625, 0.00002073, 0.00000582, -0.00001284, -0.00002255, -0.00002139, -0.00001661, -0.00001651, -0.00002290, -0.00003008, -0.00003058, -0.00002207, },
		{ -0.00001056, -0.00001546, -0.00001346, -0.00000637, -0.00000047, -0.00000055, -0.00000528, -0.00000812, -0.00000356, 0.00000660, 0.00001273, 0.00000426, -0.00002083, -0.00005155, -0.00006944, -0.00006122, -0.00002992, 0.00000465, 0.00001813, -0.00000110, -0.00004184, -0.00007572, -0.00007640, -0.00003987, 0.00000909, 0.00003054, -0.00000393, -0.00008887, -0.00018282, -0.00022989, -0.00019310, -0.00007848, 0.00006572, 0.00017533, 0.00020620, 0.00015800, 0.00007132, 0.00000106, -0.00001680, 0.00001579, 0.00006482, 0.00008939, 0.00006873, 0.00001388, -0.00004209, -0.00006709, -0.00005016, -0.00000618, 0.00003627, 0.00005413, 0.00004288, 0.00001576, -0.00000815, -0.00001741, -0.00001390, -0.00000804, -0.00000883, -0.00001687, -0.00002517, -0.00002612, -0.00001822, -0.00000722, -0.00000108, -0.00000336, },
		{ -0.00001125, -0.00000393, -0.00000160, -0.00000316, -0.00000376, -0.00000019, 0.00000501, 0.00000472, -0.00000649, -0.00002606, -0.00004289, -0.00004423, -0.00002559, 0.00000400, 0.00002656, 0.00002745, 0.00000691, -0.00001825, -0.00002556, -0.00000351, 0.00003794, 0.00007195, 0.00007301, 0.00003698, -0.00001164, -0.00003255, 0.00000401, 0.00009395, 0.00019572, 0.00025109, 0.00021900, 0.00010202, -0.00005196, -0.00017488, -0.00021589, -0.00016860, -0.00007202, 0.00001452, 0.00004629, 0.00001671, -0.00004385, -0.00009084, -0.00009411, -0.00005418, 0.00000228, 0.00004242, 0.00004813, 0.00002447, -0.00000775, -0.00002811, -0.00002943, -0.00001935, -0.00001141, -0.00001378, -0.00002393, -0.00003245, -0.00003162, -0.00002169, -0.00001011, -0.00000512, -0.00000914, -0.00001735, -0.00002212, -0.00001933, },
		{ 0.00001125, 0.00000900, 0.00000920, 0.00001019, 0.00000718, -0.00000258, -0.00001561, -0.00002339, -0.00001865, -0.00000211, 0.00001621, 0.00002333, 0.00001312, -0.00000774, -0.00002329, -0.00001981, 0.00000277, 0.00002848, 0.00003538, 0.00001198, -0.00003175, -0.00006814, -0.00006968, -0.00003001, 0.00002709, 0.00005905, 0.00003092, -0.00005908, -0.00017263, -0.00024937, -0.00024090, -0.00014028, 0.00001135, 0.00014605, 0.00020592, 0.00017373, 0.00007954, -0.00001956, -0.00007315, -0.00006391, -0.00001218, 0.00004184, 0.00006462, 0.00004777, 0.00000808, -0.00002778, -0.00004196, -0.00003485, -0.00002067, -0.00001453, -0.00002130, -0.00003384, -0.00004048, -0.00003486, -0.00002056, -0.00000752, -0.00000385, -0.00000974, -0.00001803, -0.00002030, -0.00001337, -0.00000119, 0.00000906, 0.00001283, },
		{ -0.00000870, -0.00001146, -0.00001809, -0.00002624, -0.00003010, -0.00002564, -0.00001512, -0.00000643, -0.00000725, -0.00001820, -0.00003096, -0.00003385, -0.00002126, 0.00000019, 0.00001471, 0.00000851, -0.00001845, -0.00004938, -0.00006033, -0.00003719, 0.00001165, 0.00005726, 0.00006752, 0.00002971, -0.00003721, -0.00009043, -0.00008855, -0.00001919, 0.00008906, 0.00017950, 0.00019876, 0.00012816, -0.00000412, -0.00013787, -0.00021384, -0.00020525, -0.00012942, -0.00003400, 0.00003219, 0.00004549, 0.00001558, -0.00002670, -0.00005180, -0.00004893, -0.00002838, -0.00000980, -0.00000665, -0.00001768, -0.00003048, -0.00003257, -0.00002107, -0.00000404, 0.00000666, 0.00000471, -0.00000631, -0.00001634, -0.00001685, -0.00000706, 0.00000620, 0.00001424, 0.00001298, 0.00000517, -0.00000296, -0.00000727, },
		{ -0.00001225, -0.00001473, -0.00001237, -0.00000657, -0.00000248, -0.00000428, -0.00001059, -0.00001472, -0.00001004, 0.00000338, 0.00001680, 0.00001819, 0.00000185, -0.00002497, -0.00004467, -0.00004073, -0.00001084, 0.00002832, 0.00004958, 0.00003302, -0.00001868, -0.00007864, -0.00011079, -0.00009321, -0.00003445, 0.00002887, 0.00005335, 0.00001648, -0.00006701, -0.00015129, -0.00018521, -0.00014250, -0.00003761, 0.00008223, 0.00016381, 0.00017692, 0.00012836, 0.00005370, -0.00000617, -0.00002838, -0.00001716, 0.00000478, 0.00001463, 0.00000450, -0.00001628, -0.00003103, -0.00002929, -0.00001404, 0.00000195, 0.00000651, -0.00000278, -0.00001758, -0.00002572, -0.00002057, -0.00000550, 0.00000918, 0.00001421, 0.00000776, -0.00000421, -0.00001322, -0.00001480, -0.00001101, -0.00000754, -0.00000833, },
		{ -0.00001433, -0.00001253, -0.00001308, -0.00001517, -0.00001475, -0.00000894, -0.00000018, 0.00000397, -0.00000350, -0.00002216, -0.00004178, -0.00004779, -0.00003179, 0.00000027, 0.00002917, 0.00003383, 0.00000611, -0.00004118, -0.00007988, -0.00008348, -0.00004502, 0.00001656, 0.00006681, 0.00007728, 0.00004443, -0.00000693, -0.00003875, -0.00002429, 0.00003364, 0.00010289, 0.00014094, 0.00012084, 0.00004769, -0.00004449, -0.00011317, -0.00013173, -0.00010282, -0.00005321, -0.00001490, -0.00000552, -0.00002038, -0.00003976, -0.00004476, -0.00003032, -0.00000677, 0.00000961, 0.00000916, -0.00000453, -0.00001835, -0.00002001, -0.00000706, 0.00001188, 0.00002400, 0.00002171, 0.00000760, -0.00000821, -0.00001584, -0.00001247, -0.00000328, 0.00000333, 0.00000239, -0.00000464, -0.00001212, -0.00001539, },
		{ -0.00001454, -0.00001133, -0.00000650, -0.00000234, -0.00000376, -0.00001368, -0.00002853, -0.00003854, -0.00003390, -0.00001295, 0.00001401, 0.00002944, 0.00001986, -0.00001355, -0.00005313, -0.00007426, -0.00006110, -0.00001806, 0.00003121, 0.00005777, 0.00004563, 0.00000241, -0.00004524, -0.00006885, -0.00005665, -0.00002121, 0.00000938, 0.00001076, -0.00002063, -0.00006447, -0.00008956, -0.00007439, -0.00002166, 0.00004294, 0.00008659, 0.00008940, 0.00005558, 0.00000907, -0.00002286, -0.00002643, -0.00000801, 0.00001296, 0.00001883, 0.00000557, -0.00001603, -0.00002965, -0.00002585, -0.00000851, 0.00000868, 0.00001261, 0.00000008, -0.00002048, -0.00003557, -0.00003629, -0.00002405, -0.00000862, -0.00000066, -0.00000446, -0.00001598, -0.00002691, -0.00003111, -0.00002821, -0.00002229, -0.00001749, },
		{ 0.00000244, -0.00000302, -0.00001307, -0.00002359, -0.00002704, -0.00001800, 0.00000105, 0.00001905, 0.00002245, 0.00000486, -0.00002661, -0.00005380, -0.00005821, -0.00003359, 0.00000845, 0.00004433, 0.00005292, 0.00002917, -0.00001210, -0.00004534, -0.00005043, -0.00002519, 0.00001408, 0.00004425, 0.00005114, 0.00003840, 0.00002342, 0.00002343, 0.00004168, 0.00006376, 0.00006715, 0.00003730, -0.00002062, -0.00008296, -0.00012132, -0.00011977, -0.00008428, -0.00003821, -0.00000721, -0.00000361, -0.00002053, -0.00003884, -0.00004145, -0.00002483, -0.00000028, 0.00001528, 0.00001176, -0.00000720, -0.00002731, -0.00003405, -0.00002258, -0.00000051, 0.00001809, 0.00002237, 0.00001148, -0.00000609, -0.00001925, -0.00002167, -0.00001482, -0.00000507, 0.00000179, 0.00000426, 0.00000435, 0.00000401, },
		{ -0.00001921, -0.00001831, -0.00000835, 0.00000666, 0.00001705, 0.00001356, -0.00000515, -0.00002958, -0.00004403, -0.00003673, -0.00000868, 0.00002533, 0.00004509, 0.00003783, 0.00000696, -0.00002941, -0.00004956, -0.00004179, -0.00001148, 0.00002299, 0.00004256, 0.00003934, 0.00002017, -0.00000027, -0.00001120, -0.00001359, -0.00001820, -0.00003537, -0.00006443, -0.00009102, -0.00009528, -0.00006560, -0.00000851, 0.00005291, 0.00009172, 0.00009299, 0.00006251, 0.00002253, -0.00000256, -0.00000123, 0.00001936, 0.00003995, 0.00004320, 0.00002542, -0.00000171, -0.00001998, -0.00001743, 0.00000374, 0.00002924, 0.00004268, 0.00003626, 0.00001493, -0.00000800, -0.00002022, -0.00001776, -0.00000589, 0.00000612, 0.00001173, 0.00001044, 0.00000597, 0.00000178, -0.00000185, -0.00000680, -0.00001363, },
		{ 0.00000271, 0.00000862, 0.00000341, -0.00001170, -0.00002735, -0.00003151, -0.00001801, 0.00000792, 0.00003164, 0.00003811, 0.00002204, -0.00000768, -0.00003291, -0.00003745, -0.00001729, 0.00001696, 0.00004711, 0.00005887, 0.00005014, 0.00003060, 0.00001365, 0.00000682, 0.00000780, 0.00000844, 0.00000336, -0.00000416, -0.00000392, 0.00001261, 0.00004340, 0.00007382, 0.00008429, 0.00006303, 0.00001532, -0.00003765, -0.00007054, -0.00006899, -0.00003826, 0.00000035, 0.00002306, 0.00001841, -0.00000642, -0.00003141, -0.00003757, -0.00001933, 0.00001243, 0.00003828, 0.00004314, 0.00002553, -0.00000261, -0.00002462, -0.00002980, -0.00001903, -0.00000223, 0.00000951, 0.00001098, 0.00000466, -0.00000319, -0.00000810, -0.00001014, -0.00001215, -0.00001564, -0.00001849, -0.00001659, -0.00000826, },
		{ -0.00000665, -0.00001545, -0.00001432, -0.00000070, 0.00001864, 0.00003126, 0.00002737, 0.00000727, -0.00001771, -0.00003182, -0.00002494, 0.00000067, 0.00003165, 0.00005187, 0.00005229, 0.00003571, 0.00001348, -0.00000289, -0.00000909, -0.00000929, -0.00001089, -0.00001740, -0.00002530, -0.00002734, -0.00001965, -0.00000670, 0.00000047, -0.00000776, -0.00003099, -0.00005648, -0.00006587, -0.00004720, -0.00000422, 0.00004381, 0.00007277, 0.00006825, 0.00003468, -0.00000705, -0.00003233, -0.00002774, 0.00000124, 0.00003463, 0.00005098, 0.00004061, 0.00001058, -0.00002114, -0.00003770, -0.00003335, -0.00001535, 0.00000258, 0.00000987, 0.00000501, -0.00000555, -0.00001363, -0.00001550, -0.00001339, -0.00001206, -0.00001382, -0.00001639, -0.00001526, -0.00000839, 0.00000105, 0.00000651, 0.00000334, },
		{ 0.00000309, 0.00001438, 0.00001799, 0.00000840, -0.00001077, -0.00002809, -0.00003145, -0.00001626, 0.00001097, 0.00003636, 0.00004721, 0.00003959, 0.00001996, 0.00000000, -0.00001138, -0.00001309, -0.00001079, -0.00001102, -0.00001542, -0.00001959, -0.00001732, -0.00000658, 0.00000777, 0.00001714, 0.00001612, 0.00000758, 0.00000161, 0.00000819, 0.00002855, 0.00005198, 0.00006121, 0.00004373, 0.00000158, -0.00004754, -0.00007938, -0.00007738, -0.00004297, 0.00000483, 0.00004047, 0.00004674, 0.00002411, -0.00001096, -0.00003717, -0.00004136, -0.00002523, -0.00000210, 0.00001310, 0.00001335, 0.00000248, -0.00000950, -0.00001446, -0.00001148, -0.00000603, -0.00000443, -0.00000836, -0.00001376, -0.00001463, -0.00000846, 0.00000135, 0.00000801, 0.00000689, -0.00000044, -0.00000709, -0.00000623, },
		{ 0.00000010, -0.00001412, -0.00002263, -0.00001709, 0.00000275, 0.00002810, 0.00004623, 0.00004836, 0.00003490, 0.00001450, -0.00000215, -0.00000920, -0.00000841, -0.00000625, -0.00000792, -0.00001306, -0.00001642, -0.00001272, -0.00000175, 0.00001071, 0.00001692, 0.00001324, 0.00000330, -0.00000453, -0.00000386, 0.00000400, 0.00000991, 0.00000347, -0.00001829, -0.00004627, -0.00006317, -0.00005393, -0.00001644, 0.00003512, 0.00007708, 0.00008956, 0.00006798, 0.00002554, -0.00001488, -0.00003428, -0.00002781, -0.00000579, 0.00001454, 0.00002049, 0.00001103, -0.00000466, -0.00001510, -0.00001483, -0.00000735, -0.00000118, -0.00000257, -0.00001068, -0.00001853, -0.00001876, -0.00000955, 0.00000396, 0.00001353, 0.00001409, 0.00000736, 0.00000021, -0.00000104, 0.00000424, 0.00001037, 0.00000993, },
		{ -0.00000907, 0.00001114, 0.00003160, 0.00004169, 0.00003643, 0.00001950, 0.00000036, -0.00001213, -0.00001518, -0.00001269, -0.00001108, -0.00001371, -0.00001827, -0.00001899, -0.00001190, 0.00000132, 0.00001394, 0.00001897, 0.00001452, 0.00000540, -0.00000052, 0.00000150, 0.00000899, 0.00001388, 0.00000896, -0.00000554, -0.00002041, -0.00002294, -0.00000604, 0.00002520, 0.00005484, 0.00006497, 0.00004699, 0.00000767, -0.00003418, -0.00005860, -0.00005609, -0.00003253, -0.00000452, 0.00001154, 0.00000927, -0.00000519, -0.00001900, -0.00002221, -0.00001426, -0.00000324, 0.00000106, -0.00000520, -0.00001707, -0.00002470, -0.00002075, -0.00000586, 0.00001181, 0.00002239, 0.00002110, 0.00001111, 0.00000049, -0.00000405, -0.00000198, 0.00000135, -0.00000028, -0.00000831, -0.00001752, -0.00001949, },
		{ 0.00003439, 0.00002442, 0.00000679, -0.00000926, -0.00001718, -0.00001688, -0.00001362, -0.00001303, -0.00001643, -0.00001988, -0.00001771, -0.00000761, 0.00000679, 0.00001801, 0.00002010, 0.00001301, 0.00000272, -0.00000336, -0.00000211, 0.00000271, 0.00000347, -0.00000498, -0.00002023, -0.00003278, -0.00003228, -0.00001537, 0.00001058, 0.00003107, 0.00003340, 0.00001525, -0.00001321, -0.00003538, -0.00003860, -0.00002211, 0.00000281, 0.00002045, 0.00002093, 0.00000606, -0.00001272, -0.00002287, -0.00001936, -0.00000757, 0.00000132, -0.00000108, -0.00001412, -0.00002848, -0.00003291, -0.00002232, -0.00000145, 0.00001822, 0.00002618, 0.00001956, 0.00000431, -0.00000966, -0.00001547, -0.00001341, -0.00000941, -0.00000946, -0.00001433, -0.00001885, -0.00001606, -0.00000312, 0.00001563, 0.00003088, },
		{ -0.00001093, -0.00001510, -0.00001286, -0.00001003, -0.00001093, -0.00001480, -0.00001655, -0.00001116, 0.00000170, 0.00001660, 0.00002576, 0.00002455, 0.00001488, 0.00000373, -0.00000207, -0.00000097, 0.00000231, 0.00000083, -0.00000842, -0.00002103, -0.00002741, -0.00001954, 0.00000236, 0.00002849, 0.00004477, 0.00004166, 0.00002062, -0.00000654, -0.00002504, -0.00002646, -0.00001379, 0.00000130, 0.00000682, -0.00000129, -0.00001680, -0.00002820, -0.00002724, -0.00001504, -0.00000119, 0.00000334, -0.00000565, -0.00002209, -0.00003365, -0.00003015, -0.00001082, 0.00001491, 0.00003334, 0.00003530, 0.00002152, 0.00000143, -0.00001360, -0.00001761, -0.00001281, -0.00000649, -0.00000466, -0.00000740, -0.00000931, -0.00000434, 0.00000869, 0.00002441, 0.00003419, 0.00003207, 0.00001892, 0.00000180, },
		},
		{ { -0.00000646, -0.00000135, 0.00000197, -0.00000527, -0.00001143, -0.00001645, -0.00003137, -0.00004186, -0.00002987, -0.00001618, -0.00002209, -0.00002510, -0.00000905, -0.00000103, -0.00001031, -0.00000585, 0.00001230, 0.00000933, -0.00000766, -0.00001163, -0.00002850, -0.00007224, -0.00007377, -0.00000643, 0.00002785, -0.00001744, -0.00002375, 0.00005555, 0.00006423, -0.00005854, -0.00010766, 0.00003138, 0.00013850, 0.00003138, -0.00010766, -0.00005854, 0.00006423, 0.00005555, -0.00002375, -0.00001744, 0.00002785, -0.00000643, -0.00007377, -0.00007224, -0.00002850, -0.00001163, -0.00000766, 0.00000933, 0.00001230, -0.00000585, -0.00001031, -0.00000103, -0.00000905, -0.00002510, -0.00002209, -0.00001618, -0.00002987, -0.00004186, -0.00003137, -0.00001645, -0.00001143, -0.00000527, 0.00000197, -0.00000135, },
		{ -0.00000174, -0.00001993, -0.00003717, -0.00004088, -0.00004042, -0.00003460, -0.00001427, 0.00000123, -0.00000939, -0.00002297, -0.00001269, 0.00000186, 0.00000036, -0.00000035, 0.00000232, -0.00001542, -0.00003973, -0.00003630, -0.00002763, -0.00004326, -0.00003599, 0.00002072, 0.00003804, -0.00002888, -0.00005749, 0.00002327, 0.00006656, -0.00002542, -0.00008419, 0.00002333, 0.00011736, 0.00001782, -0.00011302, -0.00005361, 0.00008772, 0.00008571, -0.00001675, -0.00004391, -0.00001218, -0.00003725, -0.00008517, -0.00005961, 0.00000486, 0.00002455, 0.00001226, 0.00000990, 0.00000166, -0.00001952, -0.00001902, -0.00000024, -0.00000638, -0.00003278, -0.00003867, -0.00002804, -0.00002861, -0.00002852, -0.00000949, 0.00000664, 0.00000136, -0.00000636, -0.00000336, -0.00000182, -0.00000328, 0.00000049, },
		{ -0.00005782, -0.00004641, -0.00002590, -0.00001042, 0.00000473, 0.00001087, -0.00000639, -0.00002535, -0.00001680, 0.00000172, 0.00000043, -0.00001150, -0.00002116, -0.00003832, -0.00004913, -0.00002561, 0.00000346, -0.00000581, -0.00001720, 0.00001144, 0.00001676, -0.00004700, -0.00007183, 0.00001956, 0.00008647, 0.00000246, -0.00008807, -0.00001686, 0.00008823, 0.00003498, -0.00007156, -0.00002831, 0.00008636, 0.00006724, -0.00004751, -0.00007917, -0.00003277, -0.00003003, -0.00005572, -0.00002257, 0.00003492, 0.00003532, 0.00000215, -0.00000793, -0.00001029, -0.00002390, -0.00001953, 0.00000079, -0.00000882, -0.00004274, -0.00004834, -0.00002407, -0.00001020, -0.00000708, 0.00000646, 0.00001474, 0.00000057, -0.00001217, -0.00000482, 0.00000318, -0.00000189, -0.00001032, -0.00002278, -0.00004473, },
		{ 0.00000281, 0.00001160, 0.00000980, 0.00000466, -0.00001270, -0.00002996, -0.00002275, -0.00000512, -0.00000831, -0.00002503, -0.00003332, -0.00003619, -0.00003010, 0.00000054, 0.00002638, 0.00000411, -0.00002909, -0.00001412, 0.00000313, -0.00003869, -0.00006130, 0.00001718, 0.00008098, -0.00000190, -0.00010578, -0.00004087, 0.00009743, 0.00008803, -0.00002662, -0.00004276, 0.00002790, 0.00001807, -0.00006209, -0.00006916, -0.00000976, 0.00000222, -0.00002261, -0.00000184, 0.00003538, 0.00001805, -0.00001988, -0.00001907, -0.00000717, -0.00001800, -0.00002041, -0.00000456, -0.00001360, -0.00004482, -0.00004268, -0.00000639, 0.00001167, 0.00000532, 0.00000703, 0.00001166, -0.00000105, -0.00001193, -0.00000285, 0.00000174, -0.00001642, -0.00003799, -0.00004949, -0.00005698, -0.00005235, -0.00002561, },
		{ -0.00000598, -0.00001223, -0.00001892, -0.00002972, -0.00002607, -0.00001206, -0.00001331, -0.00002559, -0.00002603, -0.00001797, -0.00000887, 0.00001067, 0.00002260, -0.00000639, -0.00004171, -0.00002218, 0.00001133, -0.00002003, -0.00006153, -0.00001395, 0.00004846, -0.00000635, -0.00009030, -0.00002173, 0.00012545, 0.00011854, -0.00003356, -0.00010525, -0.00004038, 0.00000735, -0.00001631, -0.00001834, 0.00001521, 0.00001103, -0.00001642, -0.00000249, 0.00002103, -0.00000324, -0.00003221, -0.00001397, 0.00000603, -0.00001377, -0.00002864, -0.00001217, -0.00001011, -0.00003241, -0.00002645, 0.00001045, 0.00002213, 0.00000360, -0.00000017, 0.00001067, 0.00000390, -0.00001115, -0.00001239, -0.00001859, -0.00004494, -0.00006351, -0.00005492, -0.00003706, -0.00001986, 0.00000344, 0.00001928, 0.00001069, },
		{ -0.00001081, -0.00001811, -0.00002390, -0.00001598, -0.00001225, -0.00001706, -0.00001372, -0.00000536, -0.00000311, 0.00000345, 0.00001028, -0.00000980, -0.00003627, -0.00001700, 0.00001462, -0.00002221, -0.00008376, -0.00005313, 0.00003014, 0.00002257, -0.00004220, 0.00000017, 0.00011438, 0.00009167, -0.00007321, -0.00014934, -0.00004974, 0.00005754, 0.00005379, 0.00000983, -0.00000463, -0.00001184, -0.00001606, 0.00000343, 0.00001158, -0.00001995, -0.00003516, 0.00000027, 0.00001965, -0.00001603, -0.00003962, -0.00001236, 0.00000620, -0.00001234, -0.00001414, 0.00001477, 0.00002067, -0.00000195, -0.00000321, 0.00001442, 0.00000496, -0.00002604, -0.00004154, -0.00004760, -0.00006084, -0.00005778, -0.00002482, 0.00000513, 0.00001042, 0.00000779, 0.00000332, -0.00001301, -0.00002676, -0.00001988, },
		{ -0.00002590, -0.00002428, -0.00001247, -0.00000903, -0.00000606, -0.00000175, -0.00000440, -0.00000299, 0.00000611, -0.00000387, -0.00002913, -0.00002402, -0.00000327, -0.00003727, -0.00009602, -0.00006380, 0.00003943, 0.00005995, -0.00001287, -0.00001979, 0.00005941, 0.00005201, -0.00007525, -0.00012473, -0.00000666, 0.00010822, 0.00007741, -0.00001701, -0.00004940, -0.00002160, 0.00001075, 0.00002093, -0.00000362, -0.00004125, -0.00003126, 0.00002314, 0.00003264, -0.00002351, -0.00004873, -0.00000279, 0.00002939, 0.00000346, -0.00001395, 0.00000930, 0.00001818, -0.00000278, -0.00000500, 0.00000883, -0.00001232, -0.00005725, -0.00007054, -0.00005273, -0.00003714, -0.00001705, 0.00001659, 0.00002924, 0.00000330, -0.00002504, -0.00003099, -0.00002949, -0.00002359, -0.00000309, 0.00001119, -0.00000420, },
		{ -0.00001126, 0.00000156, 0.00000172, 0.00000084, -0.00000397, -0.00000562, 0.00000176, -0.00000614, -0.00003539, -0.00004219, -0.00002090, -0.00003413, -0.00007816, -0.00005426, 0.00003838, 0.00005906, -0.00002512, -0.00005892, 0.00002008, 0.00005973, -0.00001941, -0.00006767, 0.00001710, 0.00009949, 0.00004668, -0.00005401, -0.00006137, 0.00000793, 0.00004845, 0.00001980, -0.00003795, -0.00005812, -0.00000693, 0.00005685, 0.00003775, -0.00004157, -0.00005555, 0.00001660, 0.00005451, 0.00000968, -0.00002416, 0.00000187, 0.00001898, -0.00000577, -0.00002014, -0.00001517, -0.00003882, -0.00007621, -0.00006504, -0.00001512, 0.00001457, 0.00001919, 0.00002179, 0.00000538, -0.00003794, -0.00006213, -0.00004054, -0.00000784, 0.00000971, 0.00002226, 0.00002248, -0.00000753, -0.00004055, -0.00003706, },
		{ 0.00000817, -0.00000139, -0.00000854, -0.00001396, -0.00001136, -0.00001716, -0.00004193, -0.00004777, -0.00001785, -0.00001224, -0.00005524, -0.00006336, -0.00000174, 0.00002319, -0.00003875, -0.00006001, 0.00003426, 0.00010440, 0.00003746, -0.00004670, -0.00001386, 0.00005211, 0.00002216, -0.00004617, -0.00003038, 0.00004160, 0.00005448, -0.00001081, -0.00006739, -0.00004271, 0.00004058, 0.00008123, 0.00001665, -0.00007016, -0.00004851, 0.00005226, 0.00007763, 0.00000058, -0.00004520, -0.00000677, 0.00001739, -0.00002084, -0.00004989, -0.00003945, -0.00004092, -0.00005281, -0.00002015, 0.00003590, 0.00004256, 0.00000411, -0.00002450, -0.00004108, -0.00006148, -0.00005396, -0.00000336, 0.00003879, 0.00003715, 0.00001933, 0.00000306, -0.00002661, -0.00005274, -0.00003794, 0.00000115, 0.00001732, },
		{ -0.00001932, -0.00002338, -0.00002212, -0.00002088, -0.00003383, -0.00003664, -0.00000837, -0.00000131, -0.00005595, -0.00009782, -0.00005599, -0.00000193, -0.00001363, -0.00001522, 0.00006740, 0.00012531, 0.00004509, -0.00006775, -0.00005585, 0.00002957, 0.00004258, -0.00000517, -0.00000500, 0.00003677, 0.00002804, -0.00003344, -0.00005576, 0.00000403, 0.00007726, 0.00006619, -0.00002823, -0.00008256, -0.00000777, 0.00009903, 0.00007901, -0.00003735, -0.00007757, -0.00001287, 0.00001521, -0.00004113, -0.00007269, -0.00003262, 0.00000149, 0.00000122, 0.00001886, 0.00004320, 0.00001050, -0.00005478, -0.00007286, -0.00004548, -0.00002300, 0.00000072, 0.00004160, 0.00005622, 0.00001935, -0.00001939, -0.00002595, -0.00002941, -0.00003866, -0.00002100, 0.00001537, 0.00002314, 0.00000079, -0.00001528, },
		{ -0.00002185, -0.00001669, -0.00002135, -0.00002863, -0.00001414, -0.00001216, -0.00006901, -0.00012445, -0.00008639, 0.00000101, 0.00003143, 0.00002789, 0.00006987, 0.00010089, 0.00002509, -0.00007727, -0.00005714, 0.00004680, 0.00007657, 0.00001665, -0.00001658, 0.00000776, 0.00001485, -0.00001354, -0.00001263, 0.00003756, 0.00006753, 0.00001987, -0.00005548, -0.00004782, 0.00005695, 0.00011927, 0.00003034, -0.00010202, -0.00009959, 0.00000355, 0.00002528, -0.00005164, -0.00006967, 0.00001156, 0.00006322, 0.00003202, -0.00000149, -0.00001079, -0.00004755, -0.00008793, -0.00005693, 0.00001782, 0.00004900, 0.00003744, 0.00003147, 0.00001646, -0.00002705, -0.00004925, -0.00002127, 0.00000324, -0.00000791, -0.00001270, 0.00000601, 0.00000721, -0.00001685, -0.00002787, -0.00002107, -0.00002055, },
		{ -0.00001841, -0.00003034, -0.00002860, -0.00002435, -0.00006747, -0.00012440, -0.00009738, -0.00000004, 0.00005161, 0.00003939, 0.00004782, 0.00007153, 0.00002777, -0.00004614, -0.00002806, 0.00006137, 0.00007985, 0.00000642, -0.00003655, 0.00000082, 0.00003866, 0.00002758, 0.00001303, 0.00002625, 0.00002860, -0.00000662, -0.00003030, 0.00001641, 0.00009122, 0.00007311, -0.00005912, -0.00014840, -0.00007034, 0.00005672, 0.00004450, -0.00005653, -0.00004935, 0.00006396, 0.00009474, -0.00000121, -0.00007739, -0.00006886, -0.00005095, -0.00004003, 0.00001917, 0.00008757, 0.00007228, -0.00000035, -0.00003589, -0.00003329, -0.00003832, -0.00002798, 0.00001817, 0.00003854, -0.00000198, -0.00003851, -0.00002611, -0.00000967, -0.00001907, -0.00002184, -0.00000959, -0.00001181, -0.00002292, -0.00001965, },
		{ -0.00003486, -0.00002623, -0.00005349, -0.00011691, -0.00011953, -0.00003242, 0.00004045, 0.00004283, 0.00004457, 0.00007373, 0.00005837, -0.00000372, -0.00001194, 0.00004591, 0.00005998, 0.00000091, -0.00002713, 0.00002346, 0.00006798, 0.00004423, 0.00000182, -0.00000224, 0.00001342, 0.00001521, 0.00001910, 0.00004557, 0.00005258, -0.00001846, -0.00012153, -0.00011822, 0.00001935, 0.00011652, 0.00003909, -0.00007634, -0.00003148, 0.00009493, 0.00007486, -0.00007838, -0.00014053, -0.00005189, 0.00003493, 0.00005084, 0.00006573, 0.00007680, 0.00001283, -0.00007844, -0.00008409, -0.00002261, 0.00001079, 0.00001680, 0.00003642, 0.00002965, -0.00003294, -0.00007537, -0.00004278, 0.00000189, 0.00000134, -0.00001026, -0.00000637, -0.00001475, -0.00003314, -0.00002829, -0.00001628, -0.00002723, },
		{ -0.00004773, -0.00012044, -0.00015824, -0.00008249, 0.00002701, 0.00006939, 0.00007537, 0.00009476, 0.00008180, 0.00001582, -0.00001747, 0.00002342, 0.00005405, 0.00002306, 0.00000218, 0.00003961, 0.00006821, 0.00003349, -0.00001091, -0.00000196, 0.00003553, 0.00004630, 0.00002807, 0.00000685, -0.00002270, -0.00007210, -0.00009224, -0.00001482, 0.00010972, 0.00012087, -0.00001676, -0.00010863, -0.00001291, 0.00010574, 0.00002854, -0.00013855, -0.00012685, 0.00005145, 0.00014436, 0.00007792, 0.00000486, -0.00001399, -0.00005534, -0.00010013, -0.00005301, 0.00004170, 0.00005984, 0.00001469, -0.00000322, -0.00000949, -0.00005273, -0.00007556, -0.00002449, 0.00002777, 0.00001210, -0.00002150, -0.00002147, -0.00002039, -0.00003391, -0.00002554, -0.00000547, -0.00001789, -0.00003918, -0.00003240, },
		{ -0.00019261, -0.00012661, 0.00002491, 0.00011389, 0.00011753, 0.00010608, 0.00008075, 0.00001918, -0.00001671, 0.00002306, 0.00006941, 0.00005163, 0.00001771, 0.00003026, 0.00005207, 0.00003225, 0.00000612, 0.00002194, 0.00004895, 0.00003271, -0.00001605, -0.00004759, -0.00004768, -0.00003226, 0.00000611, 0.00007767, 0.00012135, 0.00004869, -0.00009082, -0.00011304, 0.00002690, 0.00011219, -0.00000985, -0.00014408, -0.00004361, 0.00016426, 0.00017043, -0.00001760, -0.00012631, -0.00007651, -0.00002820, -0.00003164, 0.00000862, 0.00006998, 0.00003925, -0.00004398, -0.00005318, -0.00001175, -0.00001897, -0.00004352, -0.00001181, 0.00002278, -0.00000819, -0.00004505, -0.00002521, -0.00000055, -0.00001344, -0.00001844, -0.00000412, -0.00002122, -0.00005530, -0.00005035, -0.00005039, -0.00012472, },
		{ 0.00001991, 0.00015387, 0.00015361, 0.00010649, 0.00007153, 0.00003029, 0.00000592, 0.00003977, 0.00008221, 0.00006198, 0.00001519, 0.00001730, 0.00005084, 0.00005363, 0.00003154, 0.00002429, 0.00001944, -0.00001363, -0.00004993, -0.00004065, 0.00000997, 0.00005570, 0.00007652, 0.00008048, 0.00004930, -0.00003820, -0.00011289, -0.00005955, 0.00008254, 0.00010955, -0.00003400, -0.00011053, 0.00004288, 0.00019833, 0.00007757, -0.00017178, -0.00020002, -0.00001238, 0.00009305, 0.00003518, -0.00000856, 0.00001051, -0.00001990, -0.00007774, -0.00004087, 0.00004198, 0.00002973, -0.00004254, -0.00004846, -0.00001206, -0.00002110, -0.00003520, 0.00000603, 0.00003605, -0.00000193, -0.00003874, -0.00002995, -0.00003674, -0.00007034, -0.00006379, -0.00004684, -0.00011723, -0.00021385, -0.00016549, },
		{ 0.00018877, 0.00011245, 0.00007062, 0.00005333, 0.00003887, 0.00005430, 0.00008016, 0.00005996, 0.00001622, 0.00002030, 0.00006181, 0.00006865, 0.00002986, -0.00000555, -0.00001992, -0.00002799, -0.00001813, 0.00002943, 0.00008557, 0.00009707, 0.00005939, 0.00001331, -0.00002302, -0.00005737, -0.00005841, 0.00001860, 0.00010981, 0.00007739, -0.00005541, -0.00007459, 0.00007803, 0.00013962, -0.00006191, -0.00025881, -0.00013403, 0.00014502, 0.00018348, -0.00001647, -0.00012661, -0.00005893, -0.00000850, -0.00002566, 0.00001310, 0.00007565, 0.00001985, -0.00009581, -0.00009852, -0.00001234, 0.00002094, 0.00001011, 0.00003310, 0.00003849, -0.00003100, -0.00008576, -0.00005832, -0.00003423, -0.00006370, -0.00006687, -0.00004173, -0.00010582, -0.00022975, -0.00020870, 0.00000043, 0.00018224, },
		{ 0.00007951, 0.00007915, 0.00006996, 0.00006275, 0.00007189, 0.00005983, 0.00002700, 0.00002755, 0.00005977, 0.00005868, 0.00001137, -0.00002401, -0.00001492, 0.00001673, 0.00005267, 0.00008944, 0.00010203, 0.00006371, -0.00000017, -0.00003304, -0.00002290, -0.00000493, 0.00001226, 0.00005027, 0.00007579, 0.00001966, -0.00007161, -0.00005199, 0.00006272, 0.00004868, -0.00014589, -0.00021929, 0.00001873, 0.00025771, 0.00013180, -0.00017727, -0.00022686, -0.00001221, 0.00010908, 0.00004224, -0.00000116, 0.00002287, -0.00003221, -0.00012648, -0.00008327, 0.00005064, 0.00008311, 0.00002102, -0.00000189, -0.00000616, -0.00006863, -0.00010985, -0.00005340, -0.00000637, -0.00005048, -0.00008102, -0.00004970, -0.00009438, -0.00023785, -0.00025428, -0.00003411, 0.00020037, 0.00023025, 0.00013290, },
		{ 0.00009479, 0.00006759, 0.00006253, 0.00006037, 0.00003659, 0.00002598, 0.00004421, 0.00004625, 0.00001605, 0.00000279, 0.00003447, 0.00007688, 0.00009570, 0.00009088, 0.00006448, 0.00001690, -0.00002184, -0.00001426, 0.00002448, 0.00004035, 0.00002213, 0.00000969, 0.00000758, -0.00002504, -0.00006678, -0.00003702, 0.00003099, -0.00001048, -0.00013953, -0.00010890, 0.00013131, 0.00023196, -0.00002916, -0.00030563, -0.00018001, 0.00015636, 0.00021605, -0.00000852, -0.00012779, -0.00004736, -0.00000988, -0.00005947, -0.00001514, 0.00009568, 0.00007517, -0.00004491, -0.00007421, -0.00003063, -0.00005184, -0.00008820, -0.00003620, 0.00000660, -0.00006018, -0.00011138, -0.00005817, -0.00006801, -0.00023181, -0.00030572, -0.00009061, 0.00020605, 0.00028295, 0.00017396, 0.00009807, 0.00010161, },
		{ 0.00005436, 0.00006179, 0.00004606, 0.00002554, 0.00003702, 0.00005656, 0.00005501, 0.00005741, 0.00008460, 0.00010596, 0.00009334, 0.00006194, 0.00003313, 0.00000989, 0.00000117, 0.00001839, 0.00003784, 0.00001834, -0.00003030, -0.00005179, -0.00003399, -0.00002691, -0.00004181, -0.00002721, 0.00000697, -0.00002336, -0.00008642, -0.00002381, 0.00013589, 0.00011100, -0.00015965, -0.00028447, -0.00000564, 0.00029970, 0.00016768, -0.00019301, -0.00024510, 0.00001324, 0.00014149, 0.00003495, -0.00002345, 0.00003041, -0.00000404, -0.00011114, -0.00008865, 0.00001891, 0.00000760, -0.00007752, -0.00006341, -0.00001614, -0.00007367, -0.00012319, -0.00004604, -0.00003056, -0.00023015, -0.00037819, -0.00017034, 0.00021028, 0.00036036, 0.00024110, 0.00011996, 0.00010858, 0.00010523, 0.00006833, },
		{ 0.00006205, 0.00003408, 0.00004486, 0.00008648, 0.00010978, 0.00011202, 0.00011517, 0.00010834, 0.00007708, 0.00004215, 0.00002479, 0.00001621, 0.00000655, 0.00000097, -0.00000863, -0.00004063, -0.00007680, -0.00007089, -0.00002579, -0.00000251, -0.00002094, -0.00002758, -0.00000581, -0.00001304, -0.00004340, -0.00000405, 0.00007310, 0.00001002, -0.00016802, -0.00014845, 0.00014417, 0.00028094, -0.00002413, -0.00035243, -0.00018842, 0.00022960, 0.00029080, -0.00001394, -0.00018393, -0.00008368, -0.00001993, -0.00007134, -0.00002869, 0.00007758, 0.00002066, -0.00013030, -0.00012529, -0.00002239, -0.00004099, -0.00010395, -0.00004027, -0.00002047, -0.00025069, -0.00045847, -0.00024440, 0.00023352, 0.00046120, 0.00031484, 0.00012176, 0.00008736, 0.00010551, 0.00008059, 0.00006319, 0.00007395, },
		{ 0.00006773, 0.00013509, 0.00017659, 0.00016520, 0.00013278, 0.00009610, 0.00005185, 0.00001449, -0.00000257, -0.00001443, -0.00003474, -0.00005180, -0.00005858, -0.00006684, -0.00007144, -0.00004679, -0.00000277, 0.00000926, -0.00002029, -0.00003389, -0.00000890, -0.00000148, -0.00002835, -0.00002070, 0.00002057, -0.00001157, -0.00009277, -0.00002942, 0.00015790, 0.00013103, -0.00019000, -0.00032697, 0.00003685, 0.00042442, 0.00024245, -0.00025265, -0.00036506, -0.00005772, 0.00012528, 0.00002779, -0.00002382, 0.00003798, -0.00003233, -0.00018526, -0.00013751, 0.00003370, 0.00002036, -0.00011189, -0.00009126, -0.00004322, -0.00025433, -0.00049722, -0.00029250, 0.00025379, 0.00054096, 0.00036108, 0.00010026, 0.00006314, 0.00013072, 0.00012910, 0.00009356, 0.00008610, 0.00007180, 0.00004598, },
		{ 0.00025226, 0.00019963, 0.00011234, 0.00003908, -0.00001616, -0.00005213, -0.00006414, -0.00006727, -0.00007509, -0.00007547, -0.00005990, -0.00004547, -0.00004059, -0.00002692, -0.00000309, -0.00000125, -0.00002607, -0.00003155, -0.00000322, 0.00000440, -0.00002801, -0.00003383, 0.00000767, 0.00001310, -0.00002890, -0.00000254, 0.00007556, 0.00000193, -0.00019694, -0.00014689, 0.00023522, 0.00040251, -0.00001619, -0.00049183, -0.00034009, 0.00018222, 0.00030926, -0.00000891, -0.00018156, -0.00005812, -0.00002275, -0.00013715, -0.00008679, 0.00008256, 0.00002139, -0.00018930, -0.00016990, -0.00002642, -0.00018207, -0.00049236, -0.00036932, 0.00020132, 0.00056517, 0.00040211, 0.00011426, 0.00008953, 0.00019835, 0.00019295, 0.00010780, 0.00007165, 0.00007584, 0.00008485, 0.00013407, 0.00021795, },
		{ 0.00000514, -0.00008376, -0.00011923, -0.00012404, -0.00010665, -0.00008209, -0.00006615, -0.00005210, -0.00003235, -0.00002252, -0.00003091, -0.00003454, -0.00001932, -0.00000879, -0.00001715, -0.00001676, 0.00000355, 0.00000317, -0.00002967, -0.00003381, 0.00001234, 0.00002949, -0.00001445, -0.00002803, 0.00001138, -0.00001578, -0.00009135, 0.00000570, 0.00023968, 0.00018352, -0.00026616, -0.00049548, -0.00006790, 0.00044645, 0.00028498, -0.00027352, -0.00038253, -0.00000862, 0.00015992, -0.00003516, -0.00011332, 0.00001458, -0.00004254, -0.00024917, -0.00017169, 0.00007560, -0.00007552, -0.00053933, -0.00055576, 0.00006491, 0.00059727, 0.00052943, 0.00022060, 0.00015534, 0.00024626, 0.00020882, 0.00008576, 0.00006462, 0.00013839, 0.00020513, 0.00025710, 0.00030061, 0.00027234, 0.00014819, },
		{ -0.00017724, -0.00013084, -0.00008542, -0.00004831, -0.00003098, -0.00002749, -0.00002461, -0.00002813, -0.00004230, -0.00004479, -0.00002097, 0.00000319, 0.00000255, -0.00000468, 0.00000046, -0.00000485, -0.00003010, -0.00002948, 0.00001264, 0.00002710, -0.00001731, -0.00003645, 0.00000993, 0.00002519, -0.00001706, 0.00001349, 0.00009617, -0.00001843, -0.00029626, -0.00025882, 0.00022117, 0.00046601, -0.00001284, -0.00057430, -0.00036161, 0.00028857, 0.00038642, -0.00009984, -0.00033376, -0.00010599, -0.00000569, -0.00016563, -0.00009758, 0.00015091, -0.00006903, -0.00070206, -0.00080774, -0.00003954, 0.00074405, 0.00077083, 0.00035526, 0.00015562, 0.00020612, 0.00019470, 0.00012101, 0.00017444, 0.00031806, 0.00038082, 0.00033958, 0.00026137, 0.00014282, -0.00002381, -0.00016259, -0.00020568, },
		{ -0.00003891, -0.00002234, -0.00002223, -0.00003318, -0.00003885, -0.00003911, -0.00004139, -0.00003331, -0.00000507, 0.00001840, 0.00001140, -0.00000778, -0.00001151, -0.00001264, -0.00002660, -0.00002398, 0.00001059, 0.00002536, -0.00000724, -0.00002003, 0.00002193, 0.00003091, -0.00003373, -0.00006055, -0.00001567, -0.00004580, -0.00013688, -0.00001750, 0.00027527, 0.00021058, -0.00034514, -0.00061323, -0.00003667, 0.00062881, 0.00036432, -0.00044331, -0.00059704, -0.00002268, 0.00026596, -0.00002110, -0.00015470, 0.00003818, -0.00016888, -0.00084650, -0.00095689, -0.00002483, 0.00096046, 0.00098866, 0.00038747, 0.00005968, 0.00017424, 0.00031797, 0.00034982, 0.00042271, 0.00050642, 0.00042936, 0.00021291, 0.00001612, -0.00011386, -0.00021425, -0.00026218, -0.00022150, -0.00013603, -0.00007182, },
		{ -0.00003871, -0.00004205, -0.00003885, -0.00002781, -0.00002090, -0.00001412, 0.00000330, 0.00001607, 0.00000374, -0.00001712, -0.00001859, -0.00001146, -0.00001716, -0.00001613, 0.00001140, 0.00002836, 0.00000409, -0.00001570, 0.00000121, -0.00000850, -0.00007313, -0.00008834, -0.00000640, 0.00004196, 0.00000364, 0.00002976, 0.00010361, -0.00006844, -0.00042374, -0.00033943, 0.00033723, 0.00067903, -0.00000290, -0.00083391, -0.00056465, 0.00039599, 0.00058968, -0.00013273, -0.00051810, -0.00016977, -0.00011706, -0.00076138, -0.00098550, -0.00005898, 0.00102492, 0.00106435, 0.00038401, 0.00008312, 0.00038733, 0.00069595, 0.00069856, 0.00058022, 0.00043520, 0.00017090, -0.00014447, -0.00031274, -0.00030694, -0.00025245, -0.00020183, -0.00012978, -0.00005745, -0.00002981, -0.00003527, -0.00003879, },
		{ -0.00001661, -0.00000865, -0.00000744, -0.00001173, -0.00000658, 0.00000256, -0.00000334, -0.00001629, -0.00001362, -0.00000362, -0.00000792, -0.00001161, 0.00000784, 0.00002234, -0.00000227, -0.00003344, -0.00003165, -0.00003284, -0.00006153, -0.00004416, 0.00004719, 0.00008530, 0.00000110, -0.00006957, -0.00006367, -0.00013441, -0.00023854, -0.00002326, 0.00043295, 0.00037179, -0.00043018, -0.00088597, -0.00011939, 0.00087553, 0.00054874, -0.00067840, -0.00097492, -0.00007076, 0.00033021, -0.00054691, -0.00123362, -0.00044882, 0.00088587, 0.00123244, 0.00071275, 0.00050407, 0.00087400, 0.00109312, 0.00079335, 0.00032224, -0.00002523, -0.00029452, -0.00048177, -0.00045795, -0.00026592, -0.00010778, -0.00006225, -0.00005552, -0.00004624, -0.00005450, -0.00006882, -0.00005847, -0.00003467, -0.00002238, },
		{ -0.00001358, -0.00002101, -0.00001727, -0.00000273, 0.00000161, -0.00000605, -0.00000646, -0.00000121, -0.00000812, -0.00001679, -0.00000527, 0.00000598, -0.00001553, -0.00004387, -0.00003564, -0.00001469, -0.00001760, -0.00000577, 0.00004204, 0.00003502, -0.00006817, -0.00013033, -0.00006867, -0.00002165, -0.00004124, 0.00006070, 0.00022553, 0.00000842, -0.00053736, -0.00051324, 0.00040362, 0.00094169, -0.00002253, -0.00133602, -0.00099228, 0.00060005, 0.00093226, -0.00074222, -0.00209053, -0.00112779, 0.00101198, 0.00203591, 0.00164631, 0.00117701, 0.00117283, 0.00098984, 0.00031223, -0.00035034, -0.00059877, -0.00058457, -0.00049700, -0.00030786, -0.00006600, 0.00004514, -0.00001058, -0.00008851, -0.00010282, -0.00008899, -0.00007176, -0.00003735, -0.00000167, 0.00000612, -0.00000382, -0.00000913, },
		{ -0.00001607, 0.00000043, 0.00000829, -0.00000080, -0.00000866, -0.00000913, -0.00001656, -0.00002628, -0.00001813, -0.00000423, -0.00001361, -0.00002962, -0.00001539, 0.00000772, -0.00000348, -0.00002325, -0.00001368, -0.00002125, -0.00007512, -0.00007826, 0.00001635, 0.00006165, -0.00002191, -0.00006254, -0.00000320, -0.00008970, -0.00028688, -0.00007016, 0.00053572, 0.00046451, -0.00072817, -0.00149563, -0.00030721, 0.00147383, 0.00097031, -0.00180428, -0.00334735, -0.00135951, 0.00212949, 0.00369446, 0.00278413, 0.00136275, 0.00064002, 0.00014886, -0.00052685, -0.00094752, -0.00079890, -0.00043077, -0.00020300, -0.00007303, 0.00003839, 0.00003837, -0.00007946, -0.00015828, -0.00012470, -0.00006051, -0.00002656, -0.00000301, 0.00001565, 0.00000785, -0.00001407, -0.00002167, -0.00001950, -0.00002109, },
		{ 0.00000510, -0.00000447, -0.00001629, -0.00001838, -0.00002224, -0.00002873, -0.00001970, -0.00000046, 0.00000110, -0.00001077, -0.00000743, -0.00000183, -0.00002755, -0.00005718, -0.00004165, -0.00001102, -0.00001368, -0.00000602, 0.00003980, 0.00001809, -0.00010988, -0.00015731, -0.00002800, 0.00005511, -0.00001285, 0.00004415, 0.00020848, -0.00013184, -0.00096010, -0.00093383, 0.00061971, 0.00172211, -0.00003973, -0.00338848, -0.00402547, -0.00025786, 0.00437412, 0.00550895, 0.00308486, 0.00032688, -0.00079255, -0.00091678, -0.00100477, -0.00093400, -0.00047527, -0.00003034, 0.00004883, -0.00004650, -0.00006618, -0.00006676, -0.00012402, -0.00014056, -0.00006648, 0.00000052, 0.00000201, -0.00000999, -0.00000323, -0.00000503, -0.00001927, -0.00002048, -0.00001244, -0.00001393, -0.00001514, -0.00000319, },
		{ -0.00001808, -0.00002037, -0.00002025, -0.00002328, -0.00001459, 0.00000492, 0.00000817, -0.00000947, -0.00002161, -0.00002622, -0.00004408, -0.00005418, -0.00001810, 0.00002993, 0.00002644, -0.00000547, -0.00001731, -0.00005391, -0.00013053, -0.00010547, 0.00007178, 0.00015923, 0.00000092, -0.00015962, -0.00016987, -0.00032132, -0.00057629, -0.00015680, 0.00095066, 0.00089902, -0.00164235, -0.00435824, -0.00314825, 0.00221247, 0.00676630, 0.00616284, 0.00176209, -0.00177325, -0.00227014, -0.00126582, -0.00060615, -0.00036314, -0.00005332, 0.00014596, -0.00001019, -0.00022816, -0.00020264, -0.00007022, -0.00003758, -0.00004987, -0.00001606, 0.00000547, -0.00002668, -0.00004509, -0.00001646, 0.00000641, -0.00000235, -0.00001164, -0.00001124, -0.00001844, -0.00002537, -0.00001435, -0.00000066, -0.00000572, },
		{ -0.00001769, -0.00001938, -0.00001424, 0.00000084, 0.00000265, -0.00001720, -0.00003264, -0.00003112, -0.00002985, -0.00002458, 0.00000666, 0.00003459, 0.00000771, -0.00004449, -0.00005494, -0.00004576, -0.00005833, -0.00002361, 0.00008120, 0.00006935, -0.00015644, -0.00032426, -0.00021417, -0.00007345, -0.00008255, 0.00009040, 0.00034966, -0.00046139, -0.00266148, -0.00375327, -0.00079297, 0.00486750, 0.00780545, 0.00486750, -0.00079297, -0.00375327, -0.00266148, -0.00046139, 0.00034966, 0.00009040, -0.00008255, -0.00007345, -0.00021417, -0.00032426, -0.00015644, 0.00006935, 0.00008120, -0.00002361, -0.00005833, -0.00004576, -0.00005494, -0.00004449, 0.00000771, 0.00003459, 0.00000666, -0.00002458, -0.00002985, -0.00003112, -0.00003264, -0.00001720, 0.00000265, 0.00000084, -0.00001424, -0.00001938, },
		{ -0.00001808, -0.00000572, -0.00000066, -0.00001435, -0.00002537, -0.00001844, -0.00001124, -0.00001164, -0.00000235, 0.00000641, -0.00001646, -0.00004509, -0.00002668, 0.00000547, -0.00001606, -0.00004987, -0.00003758, -0.00007022, -0.00020264, -0.00022816, -0.00001019, 0.00014596, -0.00005332, -0.00036314, -0.00060615, -0.00126582, -0.00227014, -0.00177325, 0.00176209, 0.00616284, 0.00676630, 0.00221247, -0.00314825, -0.00435824, -0.00164235, 0.00089902, 0.00095066, -0.00015680, -0.00057629, -0.00032132, -0.00016987, -0.00015962, 0.00000092, 0.00015923, 0.00007178, -0.00010547, -0.00013053, -0.00005391, -0.00001731, -0.00000547, 0.00002644, 0.00002993, -0.00001810, -0.00005418, -0.00004408, -0.00002622, -0.00002161, -0.00000947, 0.00000817, 0.00000492, -0.00001459, -0.00002328, -0.00002025, -0.00002037, },
		{ 0.00000510, -0.00000319, -0.00001514, -0.00001393, -0.00001244, -0.00002048, -0.00001927, -0.00000503, -0.00000323, -0.00000999, 0.00000201, 0.00000052, -0.00006648, -0.00014056, -0.00012402, -0.00006676, -0.00006618, -0.00004650, 0.00004883, -0.00003034, -0.00047527, -0.00093400, -0.00100477, -0.00091678, -0.00079255, 0.00032688, 0.00308486, 0.00550895, 0.00437412, -0.00025786, -0.00402547, -0.00338848, -0.00003973, 0.00172211, 0.00061971, -0.00093383, -0.00096010, -0.00013184, 0.00020848, 0.00004415, -0.00001285, 0.00005511, -0.00002800, -0.00015731, -0.00010988, 0.00001809, 0.00003980, -0.00000602, -0.00001368, -0.00001102, -0.00004165, -0.00005718, -0.00002755, -0.00000183, -0.00000743, -0.00001077, 0.00000110, -0.00000046, -0.00001970, -0.00002873, -0.00002224, -0.00001838, -0.00001629, -0.00000447, },
		{ -0.00001607, -0.00002109, -0.00001950, -0.00002167, -0.00001407, 0.00000785, 0.00001565, -0.00000301, -0.00002656, -0.00006051, -0.00012470, -0.00015828, -0.00007946, 0.00003837, 0.00003839, -0.00007303, -0.00020300, -0.00043077, -0.00079890, -0.00094752, -0.00052685, 0.00014886, 0.00064002, 0.00136275, 0.00278413, 0.00369446, 0.00212949, -0.00135951, -0.00334735, -0.00180428, 0.00097031, 0.00147383, -0.00030721, -0.00149563, -0.00072817, 0.00046451, 0.00053572, -0.00007016, -0.00028688, -0.00008970, -0.00000320, -0.00006254, -0.00002191, 0.00006165, 0.00001635, -0.00007826, -0.00007512, -0.00002125, -0.00001368, -0.00002325, -0.00000348, 0.00000772, -0.00001539, -0.00002962, -0.00001361, -0.00000423, -0.00001813, -0.00002628, -0.00001656, -0.00000913, -0.00000866, -0.00000080, 0.00000829, 0.00000043, },
		{ -0.00001358, -0.00000913, -0.00000382, 0.00000612, -0.00000167, -0.00003735, -0.00007176, -0.00008899, -0.00010282, -0.00008851, -0.00001058, 0.00004514, -0.00006600, -0.00030786, -0.00049700, -0.00058457, -0.00059877, -0.00035034, 0.00031223, 0.00098984, 0.00117283, 0.00117701, 0.00164631, 0.00203591, 0.00101198, -0.00112779, -0.00209053, -0.00074222, 0.00093226, 0.00060005, -0.00099228, -0.00133602, -0.00002253, 0.00094169, 0.00040362, -0.00051324, -0.00053736, 0.00000842, 0.00022553, 0.00006070, -0.00004124, -0.00002165, -0.00006867, -0.00013033, -0.00006817, 0.00003502, 0.00004204, -0.00000577, -0.00001760, -0.00001469, -0.00003564, -0.00004387, -0.00001553, 0.00000598, -0.00000527, -0.00001679, -0.00000812, -0.00000121, -0.00000646, -0.00000605, 0.00000161, -0.00000273, -0.00001727, -0.00002101, },
		{ -0.00001661, -0.00002238, -0.00003467, -0.00005847, -0.00006882, -0.00005450, -0.00004624, -0.00005552, -0.00006225, -0.00010778, -0.00026592, -0.00045795, -0.00048177, -0.00029452, -0.00002523, 0.00032224, 0.00079335, 0.00109312, 0.00087400, 0.00050407, 0.00071275, 0.00123244, 0.00088587, -0.00044882, -0.00123362, -0.00054691, 0.00033021, -0.00007076, -0.00097492, -0.00067840, 0.00054874, 0.00087553, -0.00011939, -0.00088597, -0.00043018, 0.00037179, 0.00043295, -0.00002326, -0.00023854, -0.00013441, -0.00006367, -0.00006957, 0.00000110, 0.00008530, 0.00004719, -0.00004416, -0.00006153, -0.00003284, -0.00003165, -0.00003344, -0.00000227, 0.00002234, 0.00000784, -0.00001161, -0.00000792, -0.00000362, -0.00001362, -0.00001629, -0.00000334, 0.00000256, -0.00000658, -0.00001173, -0.00000744, -0.00000865, },
		{ -0.00003871, -0.00003879, -0.00003527, -0.00002981, -0.00005745, -0.00012978, -0.00020183, -0.00025245, -0.00030694, -0.00031274, -0.00014447, 0.00017090, 0.00043520, 0.00058022, 0.00069856, 0.00069595, 0.00038733, 0.00008312, 0.00038401, 0.00106435, 0.00102492, -0.00005898, -0.00098550, -0.00076138, -0.00011706, -0.00016977, -0.00051810, -0.00013273, 0.00058968, 0.00039599, -0.00056465, -0.00083391, -0.00000290, 0.00067903, 0.00033723, -0.00033943, -0.00042374, -0.00006844, 0.00010361, 0.00002976, 0.00000364, 0.00004196, -0.00000640, -0.00008834, -0.00007313, -0.00000850, 0.00000121, -0.00001570, 0.00000409, 0.00002836, 0.00001140, -0.00001613, -0.00001716, -0.00001146, -0.00001859, -0.00001712, 0.00000374, 0.00001607, 0.00000330, -0.00001412, -0.00002090, -0.00002781, -0.00003885, -0.00004205, },
		{ -0.00003891, -0.00007182, -0.00013603, -0.00022150, -0.00026218, -0.00021425, -0.00011386, 0.00001612, 0.00021291, 0.00042936, 0.00050642, 0.00042271, 0.00034982, 0.00031797, 0.00017424, 0.00005968, 0.00038747, 0.00098866, 0.00096046, -0.00002483, -0.00095689, -0.00084650, -0.00016888, 0.00003818, -0.00015470, -0.00002110, 0.00026596, -0.00002268, -0.00059704, -0.00044331, 0.00036432, 0.00062881, -0.00003667, -0.00061323, -0.00034514, 0.00021058, 0.00027527, -0.00001750, -0.00013688, -0.00004580, -0.00001567, -0.00006055, -0.00003373, 0.00003091, 0.00002193, -0.00002003, -0.00000724, 0.00002536, 0.00001059, -0.00002398, -0.00002660, -0.00001264, -0.00001151, -0.00000778, 0.00001140, 0.00001840, -0.00000507, -0.00003331, -0.00004139, -0.00003911, -0.00003885, -0.00003318, -0.00002223, -0.00002234, },
		{ -0.00017724, -0.00020568, -0.00016259, -0.00002381, 0.00014282, 0.00026137, 0.00033958, 0.00038082, 0.00031806, 0.00017444, 0.00012101, 0.00019470, 0.00020612, 0.00015562, 0.00035526, 0.00077083, 0.00074405, -0.00003954, -0.00080774, -0.00070206, -0.00006903, 0.00015091, -0.00009758, -0.00016563, -0.00000569, -0.00010599, -0.00033376, -0.00009984, 0.00038642, 0.00028857, -0.00036161, -0.00057430, -0.00001284, 0.00046601, 0.00022117, -0.00025882, -0.00029626, -0.00001843, 0.00009617, 0.00001349, -0.00001706, 0.00002519, 0.00000993, -0.00003645, -0.00001731, 0.00002710, 0.00001264, -0.00002948, -0.00003010, -0.00000485, 0.00000046, -0.00000468, 0.00000255, 0.00000319, -0.00002097, -0.00004479, -0.00004230, -0.00002813, -0.00002461, -0.00002749, -0.00003098, -0.00004831, -0.00008542, -0.00013084, },
		{ 0.00000514, 0.00014819, 0.00027234, 0.00030061, 0.00025710, 0.00020513, 0.00013839, 0.00006462, 0.00008576, 0.00020882, 0.00024626, 0.00015534, 0.00022060, 0.00052943, 0.00059727, 0.00006491, -0.00055576, -0.00053933, -0.00007552, 0.00007560, -0.00017169, -0.00024917, -0.00004254, 0.00001458, -0.00011332, -0.00003516, 0.00015992, -0.00000862, -0.00038253, -0.00027352, 0.00028498, 0.00044645, -0.00006790, -0.00049548, -0.00026616, 0.00018352, 0.00023968, 0.00000570, -0.00009135, -0.00001578, 0.00001138, -0.00002803, -0.00001445, 0.00002949, 0.00001234, -0.00003381, -0.00002967, 0.00000317, 0.00000355, -0.00001676, -0.00001715, -0.00000879, -0.00001932, -0.00003454, -0.00003091, -0.00002252, -0.00003235, -0.00005210, -0.00006615, -0.00008209, -0.00010665, -0.00012404, -0.00011923, -0.00008376, },
		{ 0.00025226, 0.00021795, 0.00013407, 0.00008485, 0.00007584, 0.00007165, 0.00010780, 0.00019295, 0.00019835, 0.00008953, 0.00011426, 0.00040211, 0.00056517, 0.00020132, -0.00036932, -0.00049236, -0.00018207, -0.00002642, -0.00016990, -0.00018930, 0.00002139, 0.00008256, -0.00008679, -0.00013715, -0.00002275, -0.00005812, -0.00018156, -0.00000891, 0.00030926, 0.00018222, -0.00034009, -0.00049183, -0.00001619, 0.00040251, 0.00023522, -0.00014689, -0.00019694, 0.00000193, 0.00007556, -0.00000254, -0.00002890, 0.00001310, 0.00000767, -0.00003383, -0.00002801, 0.00000440, -0.00000322, -0.00003155, -0.00002607, -0.00000125, -0.00000309, -0.00002692, -0.00004059, -0.00004547, -0.00005990, -0.00007547, -0.00007509, -0.00006727, -0.00006414, -0.00005213, -0.00001616, 0.00003908, 0.00011234, 0.00019963, },
		{ 0.00006773, 0.00004598, 0.00007180, 0.00008610, 0.00009356, 0.00012910, 0.00013072, 0.00006314, 0.00010026, 0.00036108, 0.00054096, 0.00025379, -0.00029250, -0.00049722, -0.00025433, -0.00004322, -0.00009126, -0.00011189, 0.00002036, 0.00003370, -0.00013751, -0.00018526, -0.00003233, 0.00003798, -0.00002382, 0.00002779, 0.00012528, -0.00005772, -0.00036506, -0.00025265, 0.00024245, 0.00042442, 0.00003685, -0.00032697, -0.00019000, 0.00013103, 0.00015790, -0.00002942, -0.00009277, -0.00001157, 0.00002057, -0.00002070, -0.00002835, -0.00000148, -0.00000890, -0.00003389, -0.00002029, 0.00000926, -0.00000277, -0.00004679, -0.00007144, -0.00006684, -0.00005858, -0.00005180, -0.00003474, -0.00001443, -0.00000257, 0.00001449, 0.00005185, 0.00009610, 0.00013278, 0.00016520, 0.00017659, 0.00013509, },
		{ 0.00006205, 0.00007395, 0.00006319, 0.00008059, 0.00010551, 0.00008736, 0.00012176, 0.00031484, 0.00046120, 0.00023352, -0.00024440, -0.00045847, -0.00025069, -0.00002047, -0.00004027, -0.00010395, -0.00004099, -0.00002239, -0.00012529, -0.00013030, 0.00002066, 0.00007758, -0.00002869, -0.00007134, -0.00001993, -0.00008368, -0.00018393, -0.00001394, 0.00029080, 0.00022960, -0.00018842, -0.00035243, -0.00002413, 0.00028094, 0.00014417, -0.00014845, -0.00016802, 0.00001002, 0.00007310, -0.00000405, -0.00004340, -0.00001304, -0.00000581, -0.00002758, -0.00002094, -0.00000251, -0.00002579, -0.00007089, -0.00007680, -0.00004063, -0.00000863, 0.00000097, 0.00000655, 0.00001621, 0.00002479, 0.00004215, 0.00007708, 0.00010834, 0.00011517, 0.00011202, 0.00010978, 0.00008648, 0.00004486, 0.00003408, },
		{ 0.00005436, 0.00006833, 0.00010523, 0.00010858, 0.00011996, 0.00024110, 0.00036036, 0.00021028, -0.00017034, -0.00037819, -0.00023015, -0.00003056, -0.00004604, -0.00012319, -0.00007367, -0.00001614, -0.00006341, -0.00007752, 0.00000760, 0.00001891, -0.00008865, -0.00011114, -0.00000404, 0.00003041, -0.00002345, 0.00003495, 0.00014149, 0.00001324, -0.00024510, -0.00019301, 0.00016768, 0.00029970, -0.00000564, -0.00028447, -0.00015965, 0.00011100, 0.00013589, -0.00002381, -0.00008642, -0.00002336, 0.00000697, -0.00002721, -0.00004181, -0.00002691, -0.00003399, -0.00005179, -0.00003030, 0.00001834, 0.00003784, 0.00001839, 0.00000117, 0.00000989, 0.00003313, 0.00006194, 0.00009334, 0.00010596, 0.00008460, 0.00005741, 0.00005501, 0.00005656, 0.00003702, 0.00002554, 0.00004606, 0.00006179, },
		{ 0.00009479, 0.00010161, 0.00009807, 0.00017396, 0.00028295, 0.00020605, -0.00009061, -0.00030572, -0.00023181, -0.00006801, -0.00005817, -0.00011138, -0.00006018, 0.00000660, -0.00003620, -0.00008820, -0.00005184, -0.00003063, -0.00007421, -0.00004491, 0.00007517, 0.00009568, -0.00001514, -0.00005947, -0.00000988, -0.00004736, -0.00012779, -0.00000852, 0.00021605, 0.00015636, -0.00018001, -0.00030563, -0.00002916, 0.00023196, 0.00013131, -0.00010890, -0.00013953, -0.00001048, 0.00003099, -0.00003702, -0.00006678, -0.00002504, 0.00000758, 0.00000969, 0.00002213, 0.00004035, 0.00002448, -0.00001426, -0.00002184, 0.00001690, 0.00006448, 0.00009088, 0.00009570, 0.00007688, 0.00003447, 0.00000279, 0.00001605, 0.00004625, 0.00004421, 0.00002598, 0.00003659, 0.00006037, 0.00006253, 0.00006759, },
		{ 0.00007951, 0.00013290, 0.00023025, 0.00020037, -0.00003411, -0.00025428, -0.00023785, -0.00009438, -0.00004970, -0.00008102, -0.00005048, -0.00000637, -0.00005340, -0.00010985, -0.00006863, -0.00000616, -0.00000189, 0.00002102, 0.00008311, 0.00005064, -0.00008327, -0.00012648, -0.00003221, 0.00002287, -0.00000116, 0.00004224, 0.00010908, -0.00001221, -0.00022686, -0.00017727, 0.00013180, 0.00025771, 0.00001873, -0.00021929, -0.00014589, 0.00004868, 0.00006272, -0.00005199, -0.00007161, 0.00001966, 0.00007579, 0.00005027, 0.00001226, -0.00000493, -0.00002290, -0.00003304, -0.00000017, 0.00006371, 0.00010203, 0.00008944, 0.00005267, 0.00001673, -0.00001492, -0.00002401, 0.00001137, 0.00005868, 0.00005977, 0.00002755, 0.00002700, 0.00005983, 0.00007189, 0.00006275, 0.00006996, 0.00007915, },
		{ 0.00018877, 0.00018224, 0.00000043, -0.00020870, -0.00022975, -0.00010582, -0.00004173, -0.00006687, -0.00006370, -0.00003423, -0.00005832, -0.00008576, -0.00003100, 0.00003849, 0.00003310, 0.00001011, 0.00002094, -0.00001234, -0.00009852, -0.00009581, 0.00001985, 0.00007565, 0.00001310, -0.00002566, -0.00000850, -0.00005893, -0.00012661, -0.00001647, 0.00018348, 0.00014502, -0.00013403, -0.00025881, -0.00006191, 0.00013962, 0.00007803, -0.00007459, -0.00005541, 0.00007739, 0.00010981, 0.00001860, -0.00005841, -0.00005737, -0.00002302, 0.00001331, 0.00005939, 0.00009707, 0.00008557, 0.00002943, -0.00001813, -0.00002799, -0.00001992, -0.00000555, 0.00002986, 0.00006865, 0.00006181, 0.00002030, 0.00001622, 0.00005996, 0.00008016, 0.00005430, 0.00003887, 0.00005333, 0.00007062, 0.00011245, },
		{ 0.00001991, -0.00016549, -0.00021385, -0.00011723, -0.00004684, -0.00006379, -0.00007034, -0.00003674, -0.00002995, -0.00003874, -0.00000193, 0.00003605, 0.00000603, -0.00003520, -0.00002110, -0.00001206, -0.00004846, -0.00004254, 0.00002973, 0.00004198, -0.00004087, -0.00007774, -0.00001990, 0.00001051, -0.00000856, 0.00003518, 0.00009305, -0.00001238, -0.00020002, -0.00017178, 0.00007757, 0.00019833, 0.00004288, -0.00011053, -0.00003400, 0.00010955, 0.00008254, -0.00005955, -0.00011289, -0.00003820, 0.00004930, 0.00008048, 0.00007652, 0.00005570, 0.00000997, -0.00004065, -0.00004993, -0.00001363, 0.00001944, 0.00002429, 0.00003154, 0.00005363, 0.00005084, 0.00001730, 0.00001519, 0.00006198, 0.00008221, 0.00003977, 0.00000592, 0.00003029, 0.00007153, 0.00010649, 0.00015361, 0.00015387, },
		{ -0.00019261, -0.00012472, -0.00005039, -0.00005035, -0.00005530, -0.00002122, -0.00000412, -0.00001844, -0.00001344, -0.00000055, -0.00002521, -0.00004505, -0.00000819, 0.00002278, -0.00001181, -0.00004352, -0.00001897, -0.00001175, -0.00005318, -0.00004398, 0.00003925, 0.00006998, 0.00000862, -0.00003164, -0.00002820, -0.00007651, -0.00012631, -0.00001760, 0.00017043, 0.00016426, -0.00004361, -0.00014408, -0.00000985, 0.00011219, 0.00002690, -0.00011304, -0.00009082, 0.00004869, 0.00012135, 0.00007767, 0.00000611, -0.00003226, -0.00004768, -0.00004759, -0.00001605, 0.00003271, 0.00004895, 0.00002194, 0.00000612, 0.00003225, 0.00005207, 0.00003026, 0.00001771, 0.00005163, 0.00006941, 0.00002306, -0.00001671, 0.00001918, 0.00008075, 0.00010608, 0.00011753, 0.00011389, 0.00002491, -0.00012661, },
		{ -0.00004773, -0.00003240, -0.00003918, -0.00001789, -0.00000547, -0.00002554, -0.00003391, -0.00002039, -0.00002147, -0.00002150, 0.00001210, 0.00002777, -0.00002449, -0.00007556, -0.00005273, -0.00000949, -0.00000322, 0.00001469, 0.00005984, 0.00004170, -0.00005301, -0.00010013, -0.00005534, -0.00001399, 0.00000486, 0.00007792, 0.00014436, 0.00005145, -0.00012685, -0.00013855, 0.00002854, 0.00010574, -0.00001291, -0.00010863, -0.00001676, 0.00012087, 0.00010972, -0.00001482, -0.00009224, -0.00007210, -0.00002270, 0.00000685, 0.00002807, 0.00004630, 0.00003553, -0.00000196, -0.00001091, 0.00003349, 0.00006821, 0.00003961, 0.00000218, 0.00002306, 0.00005405, 0.00002342, -0.00001747, 0.00001582, 0.00008180, 0.00009476, 0.00007537, 0.00006939, 0.00002701, -0.00008249, -0.00015824, -0.00012044, },
		{ -0.00003486, -0.00002723, -0.00001628, -0.00002829, -0.00003314, -0.00001475, -0.00000637, -0.00001026, 0.00000134, 0.00000189, -0.00004278, -0.00007537, -0.00003294, 0.00002965, 0.00003642, 0.00001680, 0.00001079, -0.00002261, -0.00008409, -0.00007844, 0.00001283, 0.00007680, 0.00006573, 0.00005084, 0.00003493, -0.00005189, -0.00014053, -0.00007838, 0.00007486, 0.00009493, -0.00003148, -0.00007634, 0.00003909, 0.00011652, 0.00001935, -0.00011822, -0.00012153, -0.00001846, 0.00005258, 0.00004557, 0.00001910, 0.00001521, 0.00001342, -0.00000224, 0.00000182, 0.00004423, 0.00006798, 0.00002346, -0.00002713, 0.00000091, 0.00005998, 0.00004591, -0.00001194, -0.00000372, 0.00005837, 0.00007373, 0.00004457, 0.00004283, 0.00004045, -0.00003242, -0.00011953, -0.00011691, -0.00005349, -0.00002623, },
		{ -0.00001841, -0.00001965, -0.00002292, -0.00001181, -0.00000959, -0.00002184, -0.00001907, -0.00000967, -0.00002611, -0.00003851, -0.00000198, 0.00003854, 0.00001817, -0.00002798, -0.00003832, -0.00003329, -0.00003589, -0.00000035, 0.00007228, 0.00008757, 0.00001917, -0.00004003, -0.00005095, -0.00006886, -0.00007739, -0.00000121, 0.00009474, 0.00006396, -0.00004935, -0.00005653, 0.00004450, 0.00005672, -0.00007034, -0.00014840, -0.00005912, 0.00007311, 0.00009122, 0.00001641, -0.00003030, -0.00000662, 0.00002860, 0.00002625, 0.00001303, 0.00002758, 0.00003866, 0.00000082, -0.00003655, 0.00000642, 0.00007985, 0.00006137, -0.00002806, -0.00004614, 0.00002777, 0.00007153, 0.00004782, 0.00003939, 0.00005161, -0.00000004, -0.00009738, -0.00012440, -0.00006747, -0.00002435, -0.00002860, -0.00003034, },
		{ -0.00002185, -0.00002055, -0.00002107, -0.00002787, -0.00001685, 0.00000721, 0.00000601, -0.00001270, -0.00000791, 0.00000324, -0.00002127, -0.00004925, -0.00002705, 0.00001646, 0.00003147, 0.00003744, 0.00004900, 0.00001782, -0.00005693, -0.00008793, -0.00004755, -0.00001079, -0.00000149, 0.00003202, 0.00006322, 0.00001156, -0.00006967, -0.00005164, 0.00002528, 0.00000355, -0.00009959, -0.00010202, 0.00003034, 0.00011927, 0.00005695, -0.00004782, -0.00005548, 0.00001987, 0.00006753, 0.00003756, -0.00001263, -0.00001354, 0.00001485, 0.00000776, -0.00001658, 0.00001665, 0.00007657, 0.00004680, -0.00005714, -0.00007727, 0.00002509, 0.00010089, 0.00006987, 0.00002789, 0.00003143, 0.00000101, -0.00008639, -0.00012445, -0.00006901, -0.00001216, -0.00001414, -0.00002863, -0.00002135, -0.00001669, },
		{ -0.00001932, -0.00001528, 0.00000079, 0.00002314, 0.00001537, -0.00002100, -0.00003866, -0.00002941, -0.00002595, -0.00001939, 0.00001935, 0.00005622, 0.00004160, 0.00000072, -0.00002300, -0.00004548, -0.00007286, -0.00005478, 0.00001050, 0.00004320, 0.00001886, 0.00000122, 0.00000149, -0.00003262, -0.00007269, -0.00004113, 0.00001521, -0.00001287, -0.00007757, -0.00003735, 0.00007901, 0.00009903, -0.00000777, -0.00008256, -0.00002823, 0.00006619, 0.00007726, 0.00000403, -0.00005576, -0.00003344, 0.00002804, 0.00003677, -0.00000500, -0.00000517, 0.00004258, 0.00002957, -0.00005585, -0.00006775, 0.00004509, 0.00012531, 0.00006740, -0.00001522, -0.00001363, -0.00000193, -0.00005599, -0.00009782, -0.00005595, -0.00000131, -0.00000837, -0.00003664, -0.00003383, -0.00002088, -0.00002212, -0.00002338, },
		{ 0.00000817, 0.00001732, 0.00000115, -0.00003794, -0.00005274, -0.00002661, 0.00000306, 0.00001933, 0.00003715, 0.00003879, -0.00000336, -0.00005396, -0.00006148, -0.00004108, -0.00002450, 0.00000411, 0.00004256, 0.00003590, -0.00002015, -0.00005281, -0.00004092, -0.00003945, -0.00004989, -0.00002084, 0.00001739, -0.00000677, -0.00004520, 0.00000058, 0.00007763, 0.00005226, -0.00004851, -0.00007016, 0.00001665, 0.00008123, 0.00004058, -0.00004271, -0.00006739, -0.00001081, 0.00005448, 0.00004160, -0.00003038, -0.00004617, 0.00002216, 0.00005211, -0.00001386, -0.00004670, 0.00003746, 0.00010440, 0.00003426, -0.00006001, -0.00003875, 0.00002319, -0.00000174, -0.00006336, -0.00005524, -0.00001224, -0.00001785, -0.00004777, -0.00004193, -0.00001716, -0.00001136, -0.00001396, -0.00000854, -0.00000139, },
		{ -0.00001126, -0.00003706, -0.00004055, -0.00000753, 0.00002248, 0.00002226, 0.00000971, -0.00000784, -0.00004054, -0.00006213, -0.00003794, 0.00000538, 0.00002179, 0.00001919, 0.00001457, -0.00001512, -0.00006504, -0.00007621, -0.00003882, -0.00001517, -0.00002014, -0.00000577, 0.00001898, 0.00000187, -0.00002416, 0.00000968, 0.00005451, 0.00001660, -0.00005555, -0.00004157, 0.00003775, 0.00005685, -0.00000693, -0.00005812, -0.00003795, 0.00001980, 0.00004845, 0.00000793, -0.00006137, -0.00005401, 0.00004668, 0.00009949, 0.00001710, -0.00006767, -0.00001941, 0.00005973, 0.00002008, -0.00005892, -0.00002512, 0.00005906, 0.00003838, -0.00005426, -0.00007816, -0.00003413, -0.00002090, -0.00004219, -0.00003539, -0.00000614, 0.00000176, -0.00000562, -0.00000397, 0.00000084, 0.00000172, 0.00000156, },
		{ -0.00002590, -0.00000420, 0.00001119, -0.00000309, -0.00002359, -0.00002949, -0.00003099, -0.00002504, 0.00000330, 0.00002924, 0.00001659, -0.00001705, -0.00003714, -0.00005273, -0.00007054, -0.00005725, -0.00001232, 0.00000883, -0.00000500, -0.00000278, 0.00001818, 0.00000930, -0.00001395, 0.00000346, 0.00002939, -0.00000279, -0.00004873, -0.00002351, 0.00003264, 0.00002314, -0.00003126, -0.00004125, -0.00000362, 0.00002093, 0.00001075, -0.00002160, -0.00004940, -0.00001701, 0.00007741, 0.00010822, -0.00000666, -0.00012473, -0.00007525, 0.00005201, 0.00005941, -0.00001979, -0.00001287, 0.00005995, 0.00003943, -0.00006380, -0.00009602, -0.00003727, -0.00000327, -0.00002402, -0.00002913, -0.00000387, 0.00000611, -0.00000299, -0.00000440, -0.00000175, -0.00000606, -0.00000903, -0.00001247, -0.00002428, },
		{ -0.00001081, -0.00001988, -0.00002676, -0.00001301, 0.00000332, 0.00000779, 0.00001042, 0.00000513, -0.00002482, -0.00005778, -0.00006084, -0.00004760, -0.00004154, -0.00002604, 0.00000496, 0.00001442, -0.00000321, -0.00000195, 0.00002067, 0.00001477, -0.00001414, -0.00001234, 0.00000620, -0.00001236, -0.00003962, -0.00001603, 0.00001965, 0.00000027, -0.00003516, -0.00001995, 0.00001158, 0.00000343, -0.00001606, -0.00001184, -0.00000463, 0.00000983, 0.00005379, 0.00005754, -0.00004974, -0.00014934, -0.00007321, 0.00009167, 0.00011438, 0.00000017, -0.00004220, 0.00002257, 0.00003014, -0.00005313, -0.00008376, -0.00002221, 0.00001462, -0.00001700, -0.00003627, -0.00000980, 0.00001028, 0.00000345, -0.00000311, -0.00000536, -0.00001372, -0.00001706, -0.00001225, -0.00001598, -0.00002390, -0.00001811, },
		{ -0.00000598, 0.00001069, 0.00001928, 0.00000344, -0.00001986, -0.00003706, -0.00005492, -0.00006351, -0.00004494, -0.00001859, -0.00001239, -0.00001115, 0.00000390, 0.00001067, -0.00000017, 0.00000360, 0.00002213, 0.00001045, -0.00002645, -0.00003241, -0.00001011, -0.00001217, -0.00002864, -0.00001377, 0.00000603, -0.00001397, -0.00003221, -0.00000324, 0.00002103, -0.00000249, -0.00001642, 0.00001103, 0.00001521, -0.00001834, -0.00001631, 0.00000735, -0.00004038, -0.00010525, -0.00003356, 0.00011854, 0.00012545, -0.00002173, -0.00009030, -0.00000635, 0.00004846, -0.00001395, -0.00006153, -0.00002003, 0.00001133, -0.00002218, -0.00004171, -0.00000639, 0.00002260, 0.00001067, -0.00000887, -0.00001797, -0.00002603, -0.00002559, -0.00001331, -0.00001206, -0.00002607, -0.00002972, -0.00001892, -0.00001223, },
		{ 0.00000281, -0.00002561, -0.00005235, -0.00005698, -0.00004949, -0.00003799, -0.00001642, 0.00000174, -0.00000285, -0.00001193, -0.00000105, 0.00001166, 0.00000703, 0.00000532, 0.00001167, -0.00000639, -0.00004268, -0.00004482, -0.00001360, -0.00000456, -0.00002041, -0.00001800, -0.00000717, -0.00001907, -0.00001988, 0.00001805, 0.00003538, -0.00000184, -0.00002261, 0.00000222, -0.00000976, -0.00006916, -0.00006209, 0.00001807, 0.00002790, -0.00004276, -0.00002662, 0.00008803, 0.00009743, -0.00004087, -0.00010578, -0.00000190, 0.00008098, 0.00001718, -0.00006130, -0.00003869, 0.00000313, -0.00001412, -0.00002909, 0.00000411, 0.00002638, 0.00000054, -0.00003010, -0.00003619, -0.00003332, -0.00002503, -0.00000831, -0.00000512, -0.00002275, -0.00002996, -0.00001270, 0.00000466, 0.00000980, 0.00001160, },
		{ -0.00005782, -0.00004473, -0.00002278, -0.00001032, -0.00000189, 0.00000318, -0.00000482, -0.00001217, 0.00000057, 0.00001474, 0.00000646, -0.00000708, -0.00001020, -0.00002407, -0.00004834, -0.00004274, -0.00000882, 0.00000079, -0.00001953, -0.00002390, -0.00001029, -0.00000793, 0.00000215, 0.00003532, 0.00003492, -0.00002257, -0.00005572, -0.00003003, -0.00003277, -0.00007917, -0.00004751, 0.00006724, 0.00008636, -0.00002831, -0.00007156, 0.00003498, 0.00008823, -0.00001686, -0.00008807, 0.00000246, 0.00008647, 0.00001956, -0.00007183, -0.00004700, 0.00001676, 0.00001144, -0.00001720, -0.00000581, 0.00000346, -0.00002561, -0.00004913, -0.00003832, -0.00002116, -0.00001150, 0.00000043, 0.00000172, -0.00001680, -0.00002535, -0.00000639, 0.00001087, 0.00000473, -0.00001042, -0.00002590, -0.00004641, },
		{ -0.00000174, 0.00000049, -0.00000328, -0.00000182, -0.00000336, -0.00000636, 0.00000136, 0.00000664, -0.00000949, -0.00002852, -0.00002861, -0.00002804, -0.00003867, -0.00003278, -0.00000638, -0.00000024, -0.00001902, -0.00001952, 0.00000166, 0.00000990, 0.00001226, 0.00002455, 0.00000486, -0.00005961, -0.00008517, -0.00003725, -0.00001218, -0.00004391, -0.00001675, 0.00008571, 0.00008772, -0.00005361, -0.00011302, 0.00001782, 0.00011736, 0.00002333, -0.00008419, -0.00002542, 0.00006656, 0.00002327, -0.00005749, -0.00002888, 0.00003804, 0.00002072, -0.00003599, -0.00004326, -0.00002763, -0.00003630, -0.00003973, -0.00001542, 0.00000232, -0.00000035, 0.00000036, 0.00000186, -0.00001269, -0.00002297, -0.00000939, 0.00000123, -0.00001427, -0.00003460, -0.00004042, -0.00004088, -0.00003717, -0.00001993, },
		},
		{ { -0.00002259, -0.00002273, -0.00001663, 0.00000074, -0.00002270, -0.00002975, -0.00002265, -0.00009550, -0.00007259, 0.00000448, 0.00005141, 0.00012231, 0.00002405, -0.00002232, 0.00006705, -0.00000597, 0.00000055, 0.00004204, -0.00002508, 0.00001237, 0.00000466, 0.00000466, 0.00006224, 0.00000427, 0.00001233, 0.00003282, -0.00002133, -0.00002615, -0.00003398, 0.00003251, 0.00010331, -0.00002326, -0.00013768, -0.00002326, 0.00010331, 0.00003251, -0.00003398, -0.00002615, -0.00002133, 0.00003282, 0.00001233, 0.00000427, 0.00006224, 0.00000466, 0.00000466, 0.00001237, -0.00002508, 0.00004204, 0.00000055, -0.00000597, 0.00006705, -0.00002232, 0.00002405, 0.00012231, 0.00005141, 0.00000448, -0.00007259, -0.00009550, -0.00002265, -0.00002975, -0.00002270, 0.00000074, -0.00001663, -0.00002273, },
		{ -0.00002255, -0.00000763, -0.00000328, -0.00003739, -0.00002438, -0.00003747, -0.00011677, -0.00005865, 0.00003414, 0.00009313, 0.00010383, -0.00001104, 0.00001734, 0.00006097, -0.00002516, 0.00003053, 0.00002720, -0.00003505, 0.00002264, 0.00002088, 0.00003193, 0.00002772, -0.00000223, 0.00006866, 0.00001812, -0.00008627, -0.00003192, 0.00005543, 0.00008281, -0.00000158, -0.00011944, -0.00006261, 0.00010307, 0.00006521, -0.00008084, -0.00001302, 0.00003530, -0.00002507, -0.00000070, 0.00000657, 0.00004719, 0.00004301, -0.00003600, 0.00002354, 0.00000622, -0.00002092, 0.00004542, -0.00001645, 0.00001840, 0.00004401, -0.00003607, 0.00007642, 0.00009989, 0.00000693, -0.00000288, -0.00007819, -0.00007760, -0.00001477, -0.00002657, -0.00000988, -0.00000632, -0.00002127, -0.00001972, -0.00002471, },
		{ -0.00000306, -0.00001618, -0.00004383, -0.00001866, -0.00006498, -0.00012888, -0.00002685, 0.00007019, 0.00011128, 0.00007058, -0.00001373, 0.00004907, 0.00003421, -0.00001646, 0.00005084, -0.00000834, -0.00002357, 0.00005552, 0.00002785, 0.00001431, 0.00001183, 0.00005531, 0.00007972, -0.00007345, -0.00009942, 0.00006651, 0.00010405, -0.00000888, -0.00010058, -0.00006933, 0.00006425, 0.00010180, -0.00006965, -0.00008534, 0.00009180, 0.00001268, -0.00007636, 0.00001675, 0.00003656, 0.00004705, -0.00000952, -0.00002894, 0.00005097, -0.00002498, -0.00000526, 0.00004890, -0.00003201, 0.00003269, 0.00001810, -0.00001550, 0.00010502, 0.00004590, -0.00000755, -0.00000010, -0.00009074, -0.00005780, -0.00000456, -0.00002135, -0.00000786, -0.00001495, -0.00001818, -0.00002047, -0.00002774, -0.00001663, },
		{ -0.00002940, -0.00004256, -0.00002194, -0.00009621, -0.00012211, 0.00001330, 0.00009792, 0.00011164, 0.00004333, 0.00000070, 0.00006102, 0.00001430, 0.00000641, 0.00003960, -0.00002988, 0.00002533, 0.00006910, -0.00000258, 0.00000847, 0.00005615, 0.00009064, -0.00000131, -0.00012981, 0.00000146, 0.00013796, 0.00001262, -0.00011304, -0.00005938, 0.00004408, 0.00008628, -0.00001327, -0.00012928, 0.00003700, 0.00011803, -0.00009513, -0.00005404, 0.00007164, 0.00002002, 0.00001450, -0.00002731, 0.00000959, 0.00003830, -0.00005185, 0.00003714, 0.00002709, -0.00004818, 0.00005346, 0.00000186, 0.00001059, 0.00009588, 0.00000139, 0.00000732, -0.00001514, -0.00010036, -0.00002528, -0.00000005, -0.00002361, -0.00001023, -0.00001606, -0.00001520, -0.00002675, -0.00002500, -0.00000990, -0.00000719, },
		{ -0.00004003, -0.00003325, -0.00012074, -0.00010146, 0.00005159, 0.00011793, 0.00010285, 0.00002443, 0.00002083, 0.00006294, 0.00000494, 0.00001810, 0.00001993, -0.00000815, 0.00006739, 0.00003329, -0.00001901, 0.00005769, 0.00009320, 0.00004011, -0.00008060, -0.00006895, 0.00010312, 0.00006803, -0.00011171, -0.00008032, 0.00005556, 0.00006372, 0.00000046, -0.00009174, -0.00005321, 0.00013319, 0.00001031, -0.00014196, 0.00005129, 0.00005582, -0.00002953, 0.00001952, -0.00001838, 0.00002462, 0.00000418, -0.00003595, 0.00006611, -0.00002936, -0.00002925, 0.00007873, -0.00002560, 0.00002915, 0.00007451, -0.00001613, 0.00001854, -0.00004622, -0.00008329, 0.00000590, -0.00001365, -0.00002483, -0.00000735, -0.00001532, -0.00001922, -0.00003003, -0.00001644, -0.00001075, -0.00001580, -0.00003773, },
		{ -0.00004620, -0.00013880, -0.00007518, 0.00008904, 0.00013241, 0.00008604, 0.00001618, 0.00004496, 0.00005559, -0.00000394, 0.00002557, 0.00002315, 0.00002723, 0.00006008, -0.00000411, 0.00002655, 0.00010599, 0.00005567, -0.00002923, -0.00006385, 0.00002422, 0.00007791, -0.00006233, -0.00010879, 0.00004545, 0.00008078, -0.00002228, -0.00005991, -0.00006060, 0.00004664, 0.00009989, -0.00010391, -0.00005693, 0.00010417, -0.00004020, -0.00001317, 0.00004623, -0.00003811, 0.00002857, -0.00000742, -0.00000519, 0.00004489, -0.00006909, 0.00003490, 0.00006395, -0.00006183, 0.00006174, 0.00005154, -0.00002590, 0.00001456, -0.00006326, -0.00004622, 0.00000937, -0.00003156, -0.00001205, -0.00000726, -0.00002052, -0.00002279, -0.00002483, -0.00001297, -0.00001958, -0.00002028, -0.00004388, -0.00003903, },
		{ -0.00015798, -0.00004226, 0.00012986, 0.00013724, 0.00006552, 0.00002328, 0.00006344, 0.00003523, -0.00000301, 0.00004698, 0.00003545, 0.00003367, 0.00003377, 0.00001527, 0.00008895, 0.00008262, -0.00000892, -0.00002586, -0.00000455, 0.00002171, -0.00002517, -0.00009075, 0.00000665, 0.00009740, -0.00001306, -0.00008204, -0.00002555, -0.00000070, 0.00006315, 0.00000769, -0.00010426, 0.00004982, 0.00002889, -0.00008642, 0.00007421, 0.00000562, -0.00005970, 0.00006536, -0.00002046, -0.00000083, 0.00000975, -0.00005071, 0.00008531, -0.00000579, -0.00005745, 0.00010437, 0.00000515, -0.00003028, 0.00001590, -0.00006077, -0.00002240, -0.00000881, -0.00002767, 0.00000152, -0.00002102, -0.00002148, -0.00001873, -0.00002159, -0.00002080, -0.00002446, -0.00002021, -0.00005390, -0.00003403, -0.00006010, },
		{ 0.00000827, 0.00017170, 0.00012816, 0.00005036, 0.00004021, 0.00006496, 0.00001649, 0.00002224, 0.00006798, 0.00002615, 0.00002582, 0.00004493, 0.00006704, 0.00009198, 0.00001762, -0.00001366, 0.00001782, -0.00001791, -0.00005463, -0.00005485, -0.00000341, 0.00007643, 0.00002064, -0.00009630, -0.00003814, 0.00002743, -0.00000110, 0.00002777, -0.00002919, -0.00003401, 0.00005231, -0.00007204, -0.00000723, 0.00010486, -0.00009456, 0.00000208, 0.00009021, -0.00007678, 0.00001156, 0.00000476, -0.00001406, 0.00007344, -0.00006967, 0.00000648, 0.00010453, -0.00005668, -0.00000110, 0.00002017, -0.00006459, -0.00001720, -0.00001610, -0.00000875, -0.00000776, -0.00003267, -0.00000757, -0.00002015, -0.00002911, -0.00002788, -0.00001870, -0.00002724, -0.00006327, -0.00001821, -0.00008748, -0.00018037, },
		{ 0.00020179, 0.00010588, 0.00004625, 0.00005541, 0.00005555, 0.00001966, 0.00005703, 0.00006066, 0.00000796, 0.00004933, 0.00008647, 0.00008072, 0.00004194, -0.00000611, 0.00002630, -0.00000099, -0.00009717, -0.00007119, 0.00002135, 0.00006021, 0.00002529, -0.00007102, -0.00007317, 0.00003670, 0.00001122, -0.00003368, 0.00002480, -0.00001186, -0.00000653, -0.00001948, -0.00006886, 0.00009577, 0.00000865, -0.00012009, 0.00012221, 0.00001058, -0.00011574, 0.00007767, -0.00001027, -0.00000689, 0.00003690, -0.00007504, 0.00007044, 0.00004219, -0.00007555, 0.00005385, -0.00000786, -0.00007331, -0.00000383, -0.00000927, -0.00000591, -0.00002872, -0.00001863, 0.00000164, -0.00003948, -0.00003301, -0.00002386, -0.00001480, -0.00004491, -0.00005495, -0.00000268, -0.00014536, -0.00018865, 0.00008653, },
		{ 0.00007976, 0.00005434, 0.00006420, 0.00005127, 0.00004117, 0.00007084, 0.00003385, 0.00002351, 0.00010356, 0.00010061, 0.00004319, 0.00000786, 0.00001694, 0.00001970, -0.00008751, -0.00012265, 0.00001379, 0.00008248, 0.00001555, -0.00005480, -0.00007146, 0.00000164, 0.00004310, -0.00004752, -0.00002165, 0.00005365, -0.00001680, -0.00003652, -0.00005007, 0.00001480, 0.00008991, -0.00010531, -0.00000673, 0.00014909, -0.00013600, -0.00003123, 0.00012643, -0.00008538, 0.00001224, 0.00002236, -0.00005559, 0.00008314, -0.00002239, -0.00002901, 0.00007185, -0.00006414, -0.00005001, 0.00001512, -0.00001600, -0.00001752, -0.00002826, 0.00000522, -0.00002001, -0.00005544, -0.00001946, -0.00002273, -0.00002156, -0.00005519, -0.00002617, -0.00001926, -0.00022725, -0.00015038, 0.00018241, 0.00020248, },
		{ 0.00007040, 0.00006815, 0.00005784, 0.00005882, 0.00005982, 0.00002872, 0.00008141, 0.00013845, 0.00006363, 0.00000312, 0.00001061, 0.00001576, -0.00005254, -0.00013585, -0.00004356, 0.00009363, 0.00003944, -0.00006271, -0.00005789, -0.00001497, 0.00002717, -0.00000958, -0.00006532, 0.00003426, 0.00004923, -0.00005897, -0.00004983, -0.00001304, 0.00005956, -0.00000203, -0.00009945, 0.00012068, 0.00001749, -0.00017591, 0.00013379, 0.00003791, -0.00014334, 0.00009339, -0.00001369, -0.00005113, 0.00007927, -0.00004016, 0.00002173, 0.00002749, -0.00008635, 0.00000158, 0.00000160, -0.00003050, -0.00001216, -0.00001203, 0.00000103, -0.00005441, -0.00003987, -0.00000989, -0.00003581, -0.00002101, -0.00004777, -0.00000550, -0.00008517, -0.00029186, -0.00005145, 0.00026129, 0.00017124, 0.00006732, },
		{ 0.00007100, 0.00006887, 0.00006084, 0.00005265, 0.00006840, 0.00014151, 0.00011846, 0.00000559, -0.00001255, 0.00000525, -0.00003552, -0.00010621, -0.00008611, 0.00005000, 0.00007333, -0.00004772, -0.00006941, -0.00000535, 0.00000498, -0.00000358, -0.00004303, -0.00002690, 0.00006707, -0.00000306, -0.00008070, -0.00001789, 0.00001890, 0.00003024, -0.00005532, -0.00000119, 0.00011312, -0.00012984, -0.00003543, 0.00018499, -0.00014425, -0.00005438, 0.00015159, -0.00011768, -0.00000883, 0.00009279, -0.00005054, 0.00002918, -0.00001871, -0.00005061, 0.00001931, -0.00003776, -0.00001153, 0.00000177, -0.00001721, -0.00002471, -0.00005991, -0.00001184, -0.00003021, -0.00003592, -0.00000214, -0.00004761, -0.00001673, -0.00017546, -0.00030041, 0.00008142, 0.00029552, 0.00013181, 0.00007704, 0.00008268, },
		{ 0.00007508, 0.00005823, 0.00007976, 0.00013317, 0.00015630, 0.00005090, -0.00004368, -0.00002383, -0.00002980, -0.00007765, -0.00008151, -0.00000572, 0.00006038, -0.00000565, -0.00007343, -0.00001398, 0.00001370, -0.00002538, -0.00002428, -0.00003074, 0.00001199, 0.00004892, -0.00004992, -0.00005549, 0.00000994, 0.00000693, -0.00000204, -0.00003088, 0.00005443, 0.00000372, -0.00012611, 0.00012757, 0.00003587, -0.00020967, 0.00013597, 0.00004939, -0.00019034, 0.00011866, 0.00005740, -0.00008610, 0.00003768, -0.00002318, -0.00001908, -0.00001552, -0.00003746, 0.00003067, -0.00002230, -0.00003567, -0.00003080, -0.00004281, -0.00001279, -0.00005353, 0.00000393, -0.00000216, -0.00007446, -0.00004163, -0.00025827, -0.00025385, 0.00021209, 0.00029257, 0.00010772, 0.00009460, 0.00008506, 0.00007913, },
		{ 0.00007893, 0.00014507, 0.00017378, 0.00010270, -0.00003724, -0.00007784, -0.00004124, -0.00005606, -0.00006221, -0.00002429, 0.00001529, 0.00000808, -0.00004327, -0.00003118, 0.00001765, -0.00002215, -0.00004481, -0.00000781, -0.00001403, 0.00001892, 0.00001730, -0.00005753, -0.00002343, 0.00000119, -0.00000737, 0.00000575, -0.00000293, 0.00002606, -0.00005665, -0.00000738, 0.00012527, -0.00014576, -0.00006070, 0.00020357, -0.00016063, -0.00008078, 0.00021651, -0.00008219, -0.00006943, 0.00007921, -0.00003716, -0.00002987, -0.00003351, 0.00001609, 0.00002314, -0.00007321, -0.00001719, -0.00002481, -0.00004440, -0.00002953, -0.00003296, 0.00004491, -0.00005991, -0.00009094, -0.00006057, -0.00033884, -0.00016333, 0.00033331, 0.00026881, 0.00009441, 0.00010603, 0.00008891, 0.00009131, 0.00007304, },
		{ 0.00020767, 0.00013693, -0.00000750, -0.00011222, -0.00008839, -0.00004298, -0.00004259, -0.00002120, -0.00000881, -0.00001800, -0.00002082, -0.00001872, 0.00000306, -0.00000779, -0.00005572, -0.00002195, 0.00000858, -0.00001997, 0.00001764, 0.00000291, -0.00005026, -0.00001770, -0.00001307, 0.00000837, 0.00001344, -0.00001698, -0.00000283, -0.00002181, 0.00005653, -0.00000827, -0.00015325, 0.00012698, 0.00004457, -0.00023533, 0.00015224, 0.00011441, -0.00020170, 0.00006315, 0.00006658, -0.00009178, -0.00002015, 0.00000094, 0.00004129, -0.00004742, -0.00007149, 0.00002612, -0.00005721, -0.00005034, -0.00001378, 0.00000980, 0.00002191, -0.00013559, -0.00005088, -0.00011277, -0.00042017, -0.00000706, 0.00043718, 0.00021439, 0.00008618, 0.00012256, 0.00009967, 0.00009486, 0.00008310, 0.00014747, },
		{ 0.00000926, -0.00011879, -0.00013659, -0.00006472, -0.00002018, -0.00001478, -0.00001099, -0.00003687, -0.00003653, 0.00000146, 0.00000699, -0.00001189, -0.00004429, -0.00004662, 0.00000995, -0.00000398, -0.00002595, 0.00002937, -0.00000941, -0.00005382, -0.00001939, -0.00000311, 0.00002405, -0.00001141, -0.00002747, 0.00001934, 0.00001201, 0.00000936, -0.00008637, -0.00001707, 0.00013300, -0.00015221, -0.00006370, 0.00024238, -0.00012939, -0.00011438, 0.00018052, -0.00006952, -0.00008735, 0.00004405, 0.00001073, -0.00000216, -0.00008947, -0.00000321, 0.00001090, -0.00010751, -0.00000208, 0.00001364, 0.00000797, -0.00004993, -0.00014942, 0.00000724, -0.00024781, -0.00043872, 0.00022418, 0.00047188, 0.00013072, 0.00011015, 0.00014784, 0.00009790, 0.00010162, 0.00014675, 0.00023769, 0.00018820, },
		{ -0.00015841, -0.00010211, -0.00002369, 0.00000136, -0.00001102, -0.00003420, -0.00005512, -0.00001152, 0.00002778, -0.00000863, -0.00004360, -0.00004789, -0.00001603, 0.00001546, -0.00002622, -0.00000960, 0.00003619, -0.00003648, -0.00005413, -0.00000558, 0.00001089, 0.00001022, -0.00003405, 0.00000349, 0.00003880, -0.00001938, -0.00004234, -0.00004068, 0.00006388, -0.00001010, -0.00015114, 0.00014803, 0.00006583, -0.00022969, 0.00012215, 0.00008465, -0.00018680, 0.00005277, 0.00004947, -0.00004718, -0.00002819, -0.00005717, 0.00003328, -0.00007069, -0.00007942, 0.00007284, -0.00002356, -0.00002718, -0.00009362, -0.00010323, -0.00000553, -0.00042213, -0.00031805, 0.00045385, 0.00039853, 0.00008065, 0.00017772, 0.00014369, 0.00009014, 0.00016771, 0.00025685, 0.00024873, 0.00004043, -0.00013898, },
		{ -0.00005247, 0.00000164, 0.00000029, -0.00003005, -0.00005161, -0.00003704, 0.00002194, 0.00001700, -0.00004384, -0.00005019, -0.00002731, 0.00000138, 0.00000167, -0.00003118, 0.00001400, 0.00001821, -0.00006054, -0.00003388, 0.00000814, 0.00000441, -0.00000897, -0.00001956, 0.00003566, 0.00000525, -0.00005802, -0.00001921, 0.00001670, 0.00001446, -0.00008842, 0.00000865, 0.00014270, -0.00015677, -0.00005170, 0.00020706, -0.00015938, -0.00007358, 0.00017176, -0.00009207, -0.00005146, 0.00001335, -0.00004185, -0.00000648, -0.00010402, 0.00003185, 0.00005578, -0.00010564, -0.00001311, -0.00010754, -0.00007041, -0.00009745, -0.00053234, -0.00008595, 0.00057968, 0.00028064, 0.00012149, 0.00022394, 0.00009968, 0.00015492, 0.00030606, 0.00029624, 0.00009477, -0.00015904, -0.00019383, -0.00011234, },
		{ 0.00000609, -0.00001482, -0.00004515, -0.00004181, -0.00000630, 0.00002575, -0.00001915, -0.00005944, -0.00003193, -0.00001140, 0.00000271, -0.00000922, -0.00002266, 0.00002152, -0.00001128, -0.00005999, -0.00000950, 0.00000175, -0.00000531, -0.00000351, 0.00000470, 0.00002283, -0.00004124, -0.00003952, 0.00002260, -0.00000362, -0.00004473, -0.00002313, 0.00009040, -0.00003720, -0.00014686, 0.00016960, 0.00000226, -0.00022781, 0.00019277, 0.00003773, -0.00021183, 0.00009421, -0.00000582, -0.00009261, -0.00001592, -0.00004141, 0.00008678, -0.00007031, -0.00010255, 0.00003650, -0.00016612, -0.00005151, -0.00020440, -0.00055718, 0.00016007, 0.00060913, 0.00022020, 0.00019711, 0.00019251, 0.00011863, 0.00033824, 0.00038137, 0.00013653, -0.00015238, -0.00025017, -0.00012436, -0.00005391, -0.00003386, },
		{ -0.00002974, -0.00004205, -0.00001836, 0.00000502, 0.00000154, -0.00004586, -0.00004727, -0.00001294, -0.00000762, 0.00000184, -0.00001508, -0.00001687, 0.00001568, -0.00002938, -0.00004442, -0.00000339, -0.00000931, 0.00000584, 0.00000798, -0.00000096, -0.00001012, -0.00004335, 0.00000263, 0.00001083, -0.00004488, -0.00001615, 0.00004888, 0.00000582, -0.00013033, 0.00005067, 0.00013753, -0.00023278, 0.00002041, 0.00025414, -0.00026276, -0.00005741, 0.00020674, -0.00018366, -0.00006174, 0.00005854, -0.00001205, 0.00001811, -0.00015219, 0.00001312, -0.00001133, -0.00024095, 0.00002369, -0.00034563, -0.00054417, 0.00041163, 0.00062266, 0.00020163, 0.00021466, 0.00017910, 0.00030411, 0.00049120, 0.00021058, -0.00016102, -0.00028763, -0.00017010, -0.00003368, -0.00005335, -0.00003336, 0.00000942, },
		{ -0.00002401, -0.00000543, -0.00000965, -0.00002363, -0.00004674, -0.00002525, -0.00000566, -0.00000693, 0.00000065, -0.00002433, -0.00001325, 0.00001087, -0.00003659, -0.00003438, -0.00000692, -0.00000126, 0.00002162, -0.00000652, -0.00002119, -0.00001499, -0.00001816, 0.00000753, -0.00002468, -0.00002915, 0.00003638, 0.00001447, -0.00009187, -0.00003450, 0.00015137, -0.00009379, -0.00017727, 0.00029044, -0.00004304, -0.00033337, 0.00027502, 0.00001418, -0.00030881, 0.00015917, 0.00004236, -0.00009006, -0.00003546, -0.00010441, 0.00008092, -0.00017720, -0.00017850, 0.00009460, -0.00061195, -0.00042818, 0.00072856, 0.00058934, 0.00014109, 0.00024417, 0.00033671, 0.00052533, 0.00035642, -0.00016753, -0.00034150, -0.00020712, -0.00004328, -0.00003023, -0.00008325, -0.00000956, 0.00001173, -0.00004125, },
		{ -0.00001730, -0.00003212, -0.00002800, -0.00003066, -0.00001058, -0.00000664, -0.00000573, -0.00000731, -0.00003345, -0.00000290, 0.00000602, -0.00004474, -0.00002771, 0.00000039, 0.00001531, 0.00001158, -0.00003420, -0.00001672, 0.00000046, -0.00001490, -0.00001121, -0.00002181, 0.00001115, 0.00002220, -0.00005822, -0.00006857, 0.00008524, 0.00004169, -0.00021039, 0.00010114, 0.00022283, -0.00036835, -0.00000181, 0.00034892, -0.00036804, -0.00007537, 0.00032411, -0.00020071, -0.00008721, 0.00001720, -0.00006737, 0.00001125, -0.00027428, 0.00003434, -0.00005162, -0.00091986, -0.00004799, 0.00101437, 0.00039977, 0.00012894, 0.00046030, 0.00057075, 0.00045978, -0.00008472, -0.00043896, -0.00024682, -0.00005227, -0.00002161, -0.00008565, -0.00006945, 0.00003005, -0.00000508, -0.00004199, -0.00000339, },
		{ -0.00003587, -0.00001144, -0.00001665, -0.00000778, -0.00000990, -0.00000757, -0.00002019, -0.00003219, 0.00001023, -0.00000791, -0.00004972, -0.00001112, 0.00001302, 0.00001305, -0.00001714, -0.00003805, 0.00000761, -0.00000351, -0.00002832, -0.00000465, 0.00000250, 0.00000600, -0.00002719, -0.00006251, 0.00001066, 0.00007697, -0.00009940, -0.00008726, 0.00024710, -0.00010871, -0.00031513, 0.00036860, -0.00002394, -0.00044719, 0.00038151, 0.00007227, -0.00040308, 0.00015667, -0.00002087, -0.00013267, -0.00003712, -0.00020217, 0.00015813, -0.00046033, -0.00097386, 0.00054570, 0.00102868, 0.00017132, 0.00040462, 0.00077750, 0.00051544, 0.00000947, -0.00047685, -0.00035968, -0.00003441, -0.00002506, -0.00008384, -0.00009436, -0.00001177, 0.00003926, -0.00003195, -0.00002397, -0.00000125, -0.00004765, },
		{ 0.00000527, -0.00001666, -0.00001106, -0.00001318, -0.00001349, -0.00002910, -0.00002055, 0.00001434, -0.00002589, -0.00003909, 0.00000887, 0.00000964, -0.00000658, -0.00003124, -0.00001644, 0.00001519, -0.00002522, -0.00001860, 0.00001802, -0.00000784, -0.00002755, -0.00003202, -0.00002205, 0.00002970, 0.00000297, -0.00010406, 0.00007618, 0.00011407, -0.00030398, 0.00004572, 0.00032551, -0.00043620, -0.00001207, 0.00049163, -0.00045946, -0.00014899, 0.00035354, -0.00031885, -0.00007214, 0.00003138, -0.00015396, 0.00004553, -0.00085764, -0.00065030, 0.00101769, 0.00079033, 0.00027799, 0.00091288, 0.00075192, 0.00000508, -0.00045699, -0.00047187, -0.00007751, 0.00002212, -0.00011455, -0.00009356, -0.00004252, 0.00002246, 0.00001462, -0.00003839, -0.00000543, -0.00002958, -0.00006647, -0.00000947, },
		{ -0.00002510, -0.00001291, -0.00001738, -0.00001993, -0.00002970, -0.00000768, 0.00000829, -0.00003361, -0.00001781, 0.00001201, -0.00000904, -0.00001721, -0.00002287, -0.00000052, 0.00000053, -0.00002886, 0.00001018, 0.00000895, -0.00004352, -0.00002418, -0.00000389, -0.00001882, 0.00000700, -0.00001585, -0.00003783, 0.00008934, -0.00007950, -0.00018616, 0.00027517, -0.00005680, -0.00038052, 0.00047051, -0.00000484, -0.00062215, 0.00042773, 0.00003812, -0.00053937, 0.00030248, -0.00006667, -0.00025095, -0.00009968, -0.00105283, -0.00020777, 0.00122347, 0.00073496, 0.00080879, 0.00109471, 0.00011486, -0.00056262, -0.00049759, -0.00017380, 0.00005519, -0.00010135, -0.00013977, -0.00002370, -0.00001505, 0.00001631, -0.00000095, -0.00002912, -0.00001551, -0.00006391, -0.00004604, 0.00002616, 0.00000069, },
		{ -0.00001370, -0.00002331, -0.00002256, -0.00002455, 0.00000111, 0.00000123, -0.00002932, -0.00000548, -0.00000367, -0.00001966, -0.00001007, -0.00001461, -0.00000146, -0.00000789, -0.00001051, 0.00001546, -0.00003019, -0.00004672, 0.00001247, -0.00001089, -0.00004200, 0.00002325, -0.00000182, -0.00003169, 0.00002034, -0.00011533, 0.00001583, 0.00017278, -0.00029329, 0.00005673, 0.00040922, -0.00058715, -0.00009324, 0.00059931, -0.00063758, -0.00013875, 0.00057536, -0.00057168, -0.00015101, -0.00015516, -0.00127290, 0.00020528, 0.00153992, 0.00111784, 0.00117228, 0.00047851, -0.00070190, -0.00062140, -0.00017407, 0.00000504, -0.00005153, -0.00017964, -0.00004060, 0.00001119, -0.00003833, 0.00002304, -0.00000245, -0.00004039, -0.00004231, -0.00006326, 0.00000086, 0.00003643, -0.00002184, -0.00002749, },
		{ -0.00002709, -0.00001962, -0.00001762, 0.00000751, -0.00000288, -0.00002561, -0.00000952, -0.00001775, -0.00001149, -0.00000327, -0.00001748, -0.00000023, -0.00000138, -0.00000340, -0.00001017, -0.00004912, -0.00001027, 0.00002094, -0.00005104, -0.00002173, 0.00005905, -0.00004235, -0.00005829, 0.00001489, -0.00005760, 0.00006446, -0.00001536, -0.00016541, 0.00031766, -0.00011296, -0.00056185, 0.00055966, -0.00001481, -0.00082849, 0.00070348, 0.00006262, -0.00097950, 0.00045454, -0.00059704, -0.00162628, 0.00105208, 0.00226607, 0.00139021, 0.00068253, -0.00056612, -0.00093406, -0.00017593, 0.00002707, -0.00009005, -0.00015228, -0.00009542, 0.00004088, -0.00004214, -0.00003520, 0.00005285, -0.00003501, -0.00006664, -0.00004012, -0.00002815, 0.00003112, 0.00001530, -0.00003900, -0.00002173, -0.00001935, },
		{ -0.00001413, -0.00001066, 0.00001260, -0.00000993, -0.00002997, -0.00001696, -0.00001590, 0.00000115, -0.00000872, -0.00001708, 0.00001077, -0.00000333, -0.00001875, -0.00002838, -0.00003083, 0.00001529, -0.00001143, -0.00006017, 0.00003475, 0.00003274, -0.00011044, -0.00001625, 0.00004629, -0.00005994, 0.00002815, -0.00003574, 0.00004006, 0.00014337, -0.00045233, 0.00001145, 0.00053822, -0.00077370, -0.00007477, 0.00095654, -0.00105789, -0.00039702, 0.00091812, -0.00191928, -0.00132155, 0.00279240, 0.00275349, 0.00071190, -0.00042914, -0.00108654, -0.00042945, 0.00017170, -0.00012803, -0.00018933, -0.00007831, 0.00000576, 0.00000412, -0.00008570, 0.00002821, 0.00003755, -0.00009756, -0.00005252, -0.00000399, -0.00000357, 0.00002809, -0.00001038, -0.00004100, -0.00001958, -0.00002763, -0.00002273, },
		{ -0.00000342, 0.00001240, -0.00002435, -0.00003485, -0.00001449, -0.00000534, 0.00000394, -0.00001587, -0.00000519, 0.00001257, -0.00002351, -0.00002863, -0.00001986, -0.00001072, 0.00000757, -0.00003584, -0.00002600, 0.00005189, -0.00004754, -0.00010571, 0.00006845, 0.00000396, -0.00008930, 0.00006865, 0.00001898, 0.00002821, -0.00013486, -0.00028260, 0.00040785, -0.00011433, -0.00074030, 0.00086905, 0.00000450, -0.00150297, 0.00103830, -0.00005018, -0.00332644, 0.00078605, 0.00462879, 0.00181162, -0.00076341, -0.00107539, -0.00069760, 0.00013341, 0.00002628, -0.00032271, -0.00006528, 0.00001958, -0.00001502, -0.00005600, -0.00004392, 0.00006697, -0.00004661, -0.00011306, 0.00001398, 0.00001260, -0.00001007, 0.00001807, -0.00002446, -0.00004100, -0.00002374, -0.00002758, -0.00001223, -0.00001124, },
		{ 0.00000209, -0.00004000, -0.00002929, -0.00000457, 0.00000114, 0.00000059, -0.00001457, 0.00000337, -0.00000554, -0.00003855, -0.00001564, -0.00000926, -0.00001080, -0.00000268, -0.00003161, -0.00000271, 0.00001326, -0.00009136, -0.00002679, 0.00008332, -0.00008319, -0.00002436, 0.00013159, -0.00004704, -0.00008867, -0.00015502, 0.00003747, 0.00023221, -0.00056879, 0.00005303, 0.00083942, -0.00127505, -0.00038457, 0.00166283, -0.00255508, -0.00277063, 0.00436299, 0.00461414, -0.00062709, -0.00167335, -0.00065233, -0.00007845, 0.00014700, -0.00029904, -0.00020937, 0.00010935, -0.00003847, -0.00006422, -0.00003416, 0.00000598, 0.00001856, -0.00011198, -0.00004241, 0.00006262, -0.00001909, -0.00001410, 0.00001811, -0.00004051, -0.00004438, -0.00002118, -0.00001773, -0.00000548, -0.00001103, 0.00000312, },
		{ -0.00004490, -0.00001360, 0.00000224, 0.00000267, -0.00000069, -0.00001271, -0.00000389, -0.00002530, -0.00002980, 0.00000032, -0.00001581, -0.00001028, 0.00000154, -0.00003271, -0.00001233, -0.00001964, -0.00006971, 0.00003061, 0.00002132, -0.00009122, 0.00009225, 0.00006156, -0.00016903, -0.00006083, -0.00002405, 0.00008183, -0.00012188, -0.00035247, 0.00060939, -0.00018888, -0.00137618, 0.00129890, 0.00003045, -0.00446576, 0.00083457, 0.00688679, 0.00175248, -0.00276989, -0.00100569, 0.00010156, -0.00002465, -0.00015639, -0.00031090, 0.00006393, 0.00006571, -0.00014493, -0.00000706, -0.00000334, -0.00002136, -0.00004218, -0.00008488, 0.00004207, 0.00003899, -0.00005384, 0.00000896, 0.00000631, -0.00006382, -0.00003494, -0.00000853, -0.00001098, -0.00000557, -0.00000943, 0.00000389, -0.00001621, },
		{ 0.00000142, 0.00000246, 0.00000498, -0.00000351, -0.00002051, -0.00001613, -0.00002675, -0.00001093, -0.00000136, -0.00002287, 0.00000499, -0.00000610, -0.00005454, -0.00001448, -0.00001715, -0.00004220, 0.00003169, -0.00001832, -0.00001806, 0.00010576, -0.00009842, -0.00014468, 0.00006313, -0.00003415, -0.00004710, -0.00016015, 0.00007027, 0.00031451, -0.00098834, -0.00012075, 0.00148304, -0.00291695, -0.00318733, 0.00560217, 0.00575811, -0.00226436, -0.00264409, 0.00046082, 0.00006670, -0.00035635, -0.00017772, -0.00006937, 0.00012645, -0.00011501, -0.00009490, 0.00007194, -0.00007054, -0.00005630, -0.00002618, -0.00002058, 0.00005960, -0.00001574, -0.00004546, 0.00002815, -0.00003101, -0.00006872, -0.00000758, -0.00000314, -0.00001274, -0.00000501, -0.00000842, -0.00000495, -0.00003110, -0.00003402, },
		{ 0.00000127, 0.00000699, -0.00001575, -0.00003153, -0.00001623, -0.00001541, -0.00000298, -0.00001119, -0.00001392, 0.00001216, -0.00003991, -0.00006156, 0.00001160, -0.00001904, -0.00003831, 0.00003481, -0.00000671, 0.00001539, 0.00000666, -0.00015506, 0.00000668, 0.00005114, -0.00015026, -0.00001749, -0.00000465, 0.00008295, -0.00028161, -0.00069256, 0.00094648, -0.00044894, -0.00428429, 0.00127986, 0.00782616, 0.00127986, -0.00428429, -0.00044894, 0.00094648, -0.00069256, -0.00028161, 0.00008295, -0.00000465, -0.00001749, -0.00015026, 0.00005114, 0.00000668, -0.00015506, 0.00000666, 0.00001539, -0.00000671, 0.00003481, -0.00003831, -0.00001904, 0.00001160, -0.00006156, -0.00003991, 0.00001216, -0.00001392, -0.00001119, -0.00000298, -0.00001541, -0.00001623, -0.00003153, -0.00001575, 0.00000699, },
		{ 0.00000142, -0.00003402, -0.00003110, -0.00000495, -0.00000842, -0.00000501, -0.00001274, -0.00000314, -0.00000758, -0.00006872, -0.00003101, 0.00002815, -0.00004546, -0.00001574, 0.00005960, -0.00002058, -0.00002618, -0.00005630, -0.00007054, 0.00007194, -0.00009490, -0.00011501, 0.00012645, -0.00006937, -0.00017772, -0.00035635, 0.00006670, 0.00046082, -0.00264409, -0.00226436, 0.00575811, 0.00560217, -0.00318733, -0.00291695, 0.00148304, -0.00012075, -0.00098834, 0.00031451, 0.00007027, -0.00016015, -0.00004710, -0.00003415, 0.00006313, -0.00014468, -0.00009842, 0.00010576, -0.00001806, -0.00001832, 0.00003169, -0.00004220, -0.00001715, -0.00001448, -0.00005454, -0.00000610, 0.00000499, -0.00002287, -0.00000136, -0.00001093, -0.00002675, -0.00001613, -0.00002051, -0.00000351, 0.00000498, 0.00000246, },
		{ -0.00004490, -0.00001621, 0.00000389, -0.00000943, -0.00000557, -0.00001098, -0.00000853, -0.00003494, -0.00006382, 0.00000631, 0.00000896, -0.00005384, 0.00003899, 0.00004207, -0.00008488, -0.00004218, -0.00002136, -0.00000334, -0.00000706, -0.00014493, 0.00006571, 0.00006393, -0.00031090, -0.00015639, -0.00002465, 0.00010156, -0.00100569, -0.00276989, 0.00175248, 0.00688679, 0.00083457, -0.00446576, 0.00003045, 0.00129890, -0.00137618, -0.00018888, 0.00060939, -0.00035247, -0.00012188, 0.00008183, -0.00002405, -0.00006083, -0.00016903, 0.00006156, 0.00009225, -0.00009122, 0.00002132, 0.00003061, -0.00006971, -0.00001964, -0.00001233, -0.00003271, 0.00000154, -0.00001028, -0.00001581, 0.00000032, -0.00002980, -0.00002530, -0.00000389, -0.00001271, -0.00000069, 0.00000267, 0.00000224, -0.00001360, },
		{ 0.00000209, 0.00000312, -0.00001103, -0.00000548, -0.00001773, -0.00002118, -0.00004438, -0.00004051, 0.00001811, -0.00001410, -0.00001909, 0.00006262, -0.00004241, -0.00011198, 0.00001856, 0.00000598, -0.00003416, -0.00006422, -0.00003847, 0.00010935, -0.00020937, -0.00029904, 0.00014700, -0.00007845, -0.00065233, -0.00167335, -0.00062709, 0.00461414, 0.00436299, -0.00277063, -0.00255508, 0.00166283, -0.00038457, -0.00127505, 0.00083942, 0.00005303, -0.00056879, 0.00023221, 0.00003747, -0.00015502, -0.00008867, -0.00004704, 0.00013159, -0.00002436, -0.00008319, 0.00008332, -0.00002679, -0.00009136, 0.00001326, -0.00000271, -0.00003161, -0.00000268, -0.00001080, -0.00000926, -0.00001564, -0.00003855, -0.00000554, 0.00000337, -0.00001457, 0.00000059, 0.00000114, -0.00000457, -0.00002929, -0.00004000, },
		{ -0.00000342, -0.00001124, -0.00001223, -0.00002758, -0.00002374, -0.00004100, -0.00002446, 0.00001807, -0.00001007, 0.00001260, 0.00001398, -0.00011306, -0.00004661, 0.00006697, -0.00004392, -0.00005600, -0.00001502, 0.00001958, -0.00006528, -0.00032271, 0.00002628, 0.00013341, -0.00069760, -0.00107539, -0.00076341, 0.00181162, 0.00462879, 0.00078605, -0.00332644, -0.00005018, 0.00103830, -0.00150297, 0.00000450, 0.00086905, -0.00074030, -0.00011433, 0.00040785, -0.00028260, -0.00013486, 0.00002821, 0.00001898, 0.00006865, -0.00008930, 0.00000396, 0.00006845, -0.00010571, -0.00004754, 0.00005189, -0.00002600, -0.00003584, 0.00000757, -0.00001072, -0.00001986, -0.00002863, -0.00002351, 0.00001257, -0.00000519, -0.00001587, 0.00000394, -0.00000534, -0.00001449, -0.00003485, -0.00002435, 0.00001240, },
		{ -0.00001413, -0.00002273, -0.00002763, -0.00001958, -0.00004100, -0.00001038, 0.00002809, -0.00000357, -0.00000399, -0.00005252, -0.00009756, 0.00003755, 0.00002821, -0.00008570, 0.00000412, 0.00000576, -0.00007831, -0.00018933, -0.00012803, 0.00017170, -0.00042945, -0.00108654, -0.00042914, 0.00071190, 0.00275349, 0.00279240, -0.00132155, -0.00191928, 0.00091812, -0.00039702, -0.00105789, 0.00095654, -0.00007477, -0.00077370, 0.00053822, 0.00001145, -0.00045233, 0.00014337, 0.00004006, -0.00003574, 0.00002815, -0.00005994, 0.00004629, -0.00001625, -0.00011044, 0.00003274, 0.00003475, -0.00006017, -0.00001143, 0.00001529, -0.00003083, -0.00002838, -0.00001875, -0.00000333, 0.00001077, -0.00001708, -0.00000872, 0.00000115, -0.00001590, -0.00001696, -0.00002997, -0.00000993, 0.00001260, -0.00001066, },
		{ -0.00002709, -0.00001935, -0.00002173, -0.00003900, 0.00001530, 0.00003112, -0.00002815, -0.00004012, -0.00006664, -0.00003501, 0.00005285, -0.00003520, -0.00004214, 0.00004088, -0.00009542, -0.00015228, -0.00009005, 0.00002707, -0.00017593, -0.00093406, -0.00056612, 0.00068253, 0.00139021, 0.00226607, 0.00105208, -0.00162628, -0.00059704, 0.00045454, -0.00097950, 0.00006262, 0.00070348, -0.00082849, -0.00001481, 0.00055966, -0.00056185, -0.00011296, 0.00031766, -0.00016541, -0.00001536, 0.00006446, -0.00005760, 0.00001489, -0.00005829, -0.00004235, 0.00005905, -0.00002173, -0.00005104, 0.00002094, -0.00001027, -0.00004912, -0.00001017, -0.00000340, -0.00000138, -0.00000023, -0.00001748, -0.00000327, -0.00001149, -0.00001775, -0.00000952, -0.00002561, -0.00000288, 0.00000751, -0.00001762, -0.00001962, },
		{ -0.00001370, -0.00002749, -0.00002184, 0.00003643, 0.00000086, -0.00006326, -0.00004231, -0.00004039, -0.00000245, 0.00002304, -0.00003833, 0.00001119, -0.00004060, -0.00017964, -0.00005153, 0.00000504, -0.00017407, -0.00062140, -0.00070190, 0.00047851, 0.00117228, 0.00111784, 0.00153992, 0.00020528, -0.00127290, -0.00015516, -0.00015101, -0.00057168, 0.00057536, -0.00013875, -0.00063758, 0.00059931, -0.00009324, -0.00058715, 0.00040922, 0.00005673, -0.00029329, 0.00017278, 0.00001583, -0.00011533, 0.00002034, -0.00003169, -0.00000182, 0.00002325, -0.00004200, -0.00001089, 0.00001247, -0.00004672, -0.00003019, 0.00001546, -0.00001051, -0.00000789, -0.00000146, -0.00001461, -0.00001007, -0.00001966, -0.00000367, -0.00000548, -0.00002932, 0.00000123, 0.00000111, -0.00002455, -0.00002256, -0.00002331, },
		{ -0.00002510, 0.00000069, 0.00002616, -0.00004604, -0.00006391, -0.00001551, -0.00002912, -0.00000095, 0.00001631, -0.00001505, -0.00002370, -0.00013977, -0.00010135, 0.00005519, -0.00017380, -0.00049759, -0.00056262, 0.00011486, 0.00109471, 0.00080879, 0.00073496, 0.00122347, -0.00020777, -0.00105283, -0.00009968, -0.00025095, -0.00006667, 0.00030248, -0.00053937, 0.00003812, 0.00042773, -0.00062215, -0.00000484, 0.00047051, -0.00038052, -0.00005680, 0.00027517, -0.00018616, -0.00007950, 0.00008934, -0.00003783, -0.00001585, 0.00000700, -0.00001882, -0.00000389, -0.00002418, -0.00004352, 0.00000895, 0.00001018, -0.00002886, 0.00000053, -0.00000052, -0.00002287, -0.00001721, -0.00000904, 0.00001201, -0.00001781, -0.00003361, 0.00000829, -0.00000768, -0.00002970, -0.00001993, -0.00001738, -0.00001291, },
		{ 0.00000527, -0.00000947, -0.00006647, -0.00002958, -0.00000543, -0.00003839, 0.00001462, 0.00002246, -0.00004252, -0.00009356, -0.00011455, 0.00002212, -0.00007751, -0.00047187, -0.00045699, 0.00000508, 0.00075192, 0.00091288, 0.00027799, 0.00079033, 0.00101769, -0.00065030, -0.00085764, 0.00004553, -0.00015396, 0.00003138, -0.00007214, -0.00031885, 0.00035354, -0.00014899, -0.00045946, 0.00049163, -0.00001207, -0.00043620, 0.00032551, 0.00004572, -0.00030398, 0.00011407, 0.00007618, -0.00010406, 0.00000297, 0.00002970, -0.00002205, -0.00003202, -0.00002755, -0.00000784, 0.00001802, -0.00001860, -0.00002522, 0.00001519, -0.00001644, -0.00003124, -0.00000658, 0.00000964, 0.00000887, -0.00003909, -0.00002589, 0.00001434, -0.00002055, -0.00002910, -0.00001349, -0.00001318, -0.00001106, -0.00001666, },
		{ -0.00003587, -0.00004765, -0.00000125, -0.00002397, -0.00003195, 0.00003926, -0.00001177, -0.00009436, -0.00008384, -0.00002506, -0.00003441, -0.00035968, -0.00047685, 0.00000947, 0.00051544, 0.00077750, 0.00040462, 0.00017132, 0.00102868, 0.00054570, -0.00097386, -0.00046033, 0.00015813, -0.00020217, -0.00003712, -0.00013267, -0.00002087, 0.00015667, -0.00040308, 0.00007227, 0.00038151, -0.00044719, -0.00002394, 0.00036860, -0.00031513, -0.00010871, 0.00024710, -0.00008726, -0.00009940, 0.00007697, 0.00001066, -0.00006251, -0.00002719, 0.00000600, 0.00000250, -0.00000465, -0.00002832, -0.00000351, 0.00000761, -0.00003805, -0.00001714, 0.00001305, 0.00001302, -0.00001112, -0.00004972, -0.00000791, 0.00001023, -0.00003219, -0.00002019, -0.00000757, -0.00000990, -0.00000778, -0.00001665, -0.00001144, },
		{ -0.00001730, -0.00000339, -0.00004199, -0.00000508, 0.00003005, -0.00006945, -0.00008565, -0.00002161, -0.00005227, -0.00024682, -0.00043896, -0.00008472, 0.00045978, 0.00057075, 0.00046030, 0.00012894, 0.00039977, 0.00101437, -0.00004799, -0.00091986, -0.00005162, 0.00003434, -0.00027428, 0.00001125, -0.00006737, 0.00001720, -0.00008721, -0.00020071, 0.00032411, -0.00007537, -0.00036804, 0.00034892, -0.00000181, -0.00036835, 0.00022283, 0.00010114, -0.00021039, 0.00004169, 0.00008524, -0.00006857, -0.00005822, 0.00002220, 0.00001115, -0.00002181, -0.00001121, -0.00001490, 0.00000046, -0.00001672, -0.00003420, 0.00001158, 0.00001531, 0.00000039, -0.00002771, -0.00004474, 0.00000602, -0.00000290, -0.00003345, -0.00000731, -0.00000573, -0.00000664, -0.00001058, -0.00003066, -0.00002800, -0.00003212, },
		{ -0.00002401, -0.00004125, 0.00001173, -0.00000956, -0.00008325, -0.00003023, -0.00004328, -0.00020712, -0.00034150, -0.00016753, 0.00035642, 0.00052533, 0.00033671, 0.00024417, 0.00014109, 0.00058934, 0.00072856, -0.00042818, -0.00061195, 0.00009460, -0.00017850, -0.00017720, 0.00008092, -0.00010441, -0.00003546, -0.00009006, 0.00004236, 0.00015917, -0.00030881, 0.00001418, 0.00027502, -0.00033337, -0.00004304, 0.00029044, -0.00017727, -0.00009379, 0.00015137, -0.00003450, -0.00009187, 0.00001447, 0.00003638, -0.00002915, -0.00002468, 0.00000753, -0.00001816, -0.00001499, -0.00002119, -0.00000652, 0.00002162, -0.00000126, -0.00000692, -0.00003438, -0.00003659, 0.00001087, -0.00001325, -0.00002433, 0.00000065, -0.00000693, -0.00000566, -0.00002525, -0.00004674, -0.00002363, -0.00000965, -0.00000543, },
		{ -0.00002974, 0.00000942, -0.00003336, -0.00005335, -0.00003368, -0.00017010, -0.00028763, -0.00016102, 0.00021058, 0.00049120, 0.00030411, 0.00017910, 0.00021466, 0.00020163, 0.00062266, 0.00041163, -0.00054417, -0.00034563, 0.00002369, -0.00024095, -0.00001133, 0.00001312, -0.00015219, 0.00001811, -0.00001205, 0.00005854, -0.00006174, -0.00018366, 0.00020674, -0.00005741, -0.00026276, 0.00025414, 0.00002041, -0.00023278, 0.00013753, 0.00005067, -0.00013033, 0.00000582, 0.00004888, -0.00001615, -0.00004488, 0.00001083, 0.00000263, -0.00004335, -0.00001012, -0.00000096, 0.00000798, 0.00000584, -0.00000931, -0.00000339, -0.00004442, -0.00002938, 0.00001568, -0.00001687, -0.00001508, 0.00000184, -0.00000762, -0.00001294, -0.00004727, -0.00004586, 0.00000154, 0.00000502, -0.00001836, -0.00004205, },
		{ 0.00000609, -0.00003386, -0.00005391, -0.00012436, -0.00025017, -0.00015238, 0.00013653, 0.00038137, 0.00033824, 0.00011863, 0.00019251, 0.00019711, 0.00022020, 0.00060913, 0.00016007, -0.00055718, -0.00020440, -0.00005151, -0.00016612, 0.00003650, -0.00010255, -0.00007031, 0.00008678, -0.00004141, -0.00001592, -0.00009261, -0.00000582, 0.00009421, -0.00021183, 0.00003773, 0.00019277, -0.00022781, 0.00000226, 0.00016960, -0.00014686, -0.00003720, 0.00009040, -0.00002313, -0.00004473, -0.00000362, 0.00002260, -0.00003952, -0.00004124, 0.00002283, 0.00000470, -0.00000351, -0.00000531, 0.00000175, -0.00000950, -0.00005999, -0.00001128, 0.00002152, -0.00002266, -0.00000922, 0.00000271, -0.00001140, -0.00003193, -0.00005944, -0.00001915, 0.00002575, -0.00000630, -0.00004181, -0.00004515, -0.00001482, },
		{ -0.00005247, -0.00011234, -0.00019383, -0.00015904, 0.00009477, 0.00029624, 0.00030606, 0.00015492, 0.00009968, 0.00022394, 0.00012149, 0.00028064, 0.00057968, -0.00008595, -0.00053234, -0.00009745, -0.00007041, -0.00010754, -0.00001311, -0.00010564, 0.00005578, 0.00003185, -0.00010402, -0.00000648, -0.00004185, 0.00001335, -0.00005146, -0.00009207, 0.00017176, -0.00007358, -0.00015938, 0.00020706, -0.00005170, -0.00015677, 0.00014270, 0.00000865, -0.00008842, 0.00001446, 0.00001670, -0.00001921, -0.00005802, 0.00000525, 0.00003566, -0.00001956, -0.00000897, 0.00000441, 0.00000814, -0.00003388, -0.00006054, 0.00001821, 0.00001400, -0.00003118, 0.00000167, 0.00000138, -0.00002731, -0.00005019, -0.00004384, 0.00001700, 0.00002194, -0.00003704, -0.00005161, -0.00003005, 0.00000029, 0.00000164, },
		{ -0.00015841, -0.00013898, 0.00004043, 0.00024873, 0.00025685, 0.00016771, 0.00009014, 0.00014369, 0.00017772, 0.00008065, 0.00039853, 0.00045385, -0.00031805, -0.00042213, -0.00000553, -0.00010323, -0.00009362, -0.00002718, -0.00002356, 0.00007284, -0.00007942, -0.00007069, 0.00003328, -0.00005717, -0.00002819, -0.00004718, 0.00004947, 0.00005277, -0.00018680, 0.00008465, 0.00012215, -0.00022969, 0.00006583, 0.00014803, -0.00015114, -0.00001010, 0.00006388, -0.00004068, -0.00004234, -0.00001938, 0.00003880, 0.00000349, -0.00003405, 0.00001022, 0.00001089, -0.00000558, -0.00005413, -0.00003648, 0.00003619, -0.00000960, -0.00002622, 0.00001546, -0.00001603, -0.00004789, -0.00004360, -0.00000863, 0.00002778, -0.00001152, -0.00005512, -0.00003420, -0.00001102, 0.00000136, -0.00002369, -0.00010211, },
		{ 0.00000926, 0.00018820, 0.00023769, 0.00014675, 0.00010162, 0.00009790, 0.00014784, 0.00011015, 0.00013072, 0.00047188, 0.00022418, -0.00043872, -0.00024781, 0.00000724, -0.00014942, -0.00004993, 0.00000797, 0.00001364, -0.00000208, -0.00010751, 0.00001090, -0.00000321, -0.00008947, -0.00000216, 0.00001073, 0.00004405, -0.00008735, -0.00006952, 0.00018052, -0.00011438, -0.00012939, 0.00024238, -0.00006370, -0.00015221, 0.00013300, -0.00001707, -0.00008637, 0.00000936, 0.00001201, 0.00001934, -0.00002747, -0.00001141, 0.00002405, -0.00000311, -0.00001939, -0.00005382, -0.00000941, 0.00002937, -0.00002595, -0.00000398, 0.00000995, -0.00004662, -0.00004429, -0.00001189, 0.00000699, 0.00000146, -0.00003653, -0.00003687, -0.00001099, -0.00001478, -0.00002018, -0.00006472, -0.00013659, -0.00011879, },
		{ 0.00020767, 0.00014747, 0.00008310, 0.00009486, 0.00009967, 0.00012256, 0.00008618, 0.00021439, 0.00043718, -0.00000706, -0.00042017, -0.00011277, -0.00005088, -0.00013559, 0.00002191, 0.00000980, -0.00001378, -0.00005034, -0.00005721, 0.00002612, -0.00007149, -0.00004742, 0.00004129, 0.00000094, -0.00002015, -0.00009178, 0.00006658, 0.00006315, -0.00020170, 0.00011441, 0.00015224, -0.00023533, 0.00004457, 0.00012698, -0.00015325, -0.00000827, 0.00005653, -0.00002181, -0.00000283, -0.00001698, 0.00001344, 0.00000837, -0.00001307, -0.00001770, -0.00005026, 0.00000291, 0.00001764, -0.00001997, 0.00000858, -0.00002195, -0.00005572, -0.00000779, 0.00000306, -0.00001872, -0.00002082, -0.00001800, -0.00000881, -0.00002120, -0.00004259, -0.00004298, -0.00008839, -0.00011222, -0.00000750, 0.00013693, },
		{ 0.00007893, 0.00007304, 0.00009131, 0.00008891, 0.00010603, 0.00009441, 0.00026881, 0.00033331, -0.00016333, -0.00033884, -0.00006057, -0.00009094, -0.00005991, 0.00004491, -0.00003296, -0.00002953, -0.00004440, -0.00002481, -0.00001719, -0.00007321, 0.00002314, 0.00001609, -0.00003351, -0.00002987, -0.00003716, 0.00007921, -0.00006943, -0.00008219, 0.00021651, -0.00008078, -0.00016063, 0.00020357, -0.00006070, -0.00014576, 0.00012527, -0.00000738, -0.00005665, 0.00002606, -0.00000293, 0.00000575, -0.00000737, 0.00000119, -0.00002343, -0.00005753, 0.00001730, 0.00001892, -0.00001403, -0.00000781, -0.00004481, -0.00002215, 0.00001765, -0.00003118, -0.00004327, 0.00000808, 0.00001529, -0.00002429, -0.00006221, -0.00005606, -0.00004124, -0.00007784, -0.00003724, 0.00010270, 0.00017378, 0.00014507, },
		{ 0.00007508, 0.00007913, 0.00008506, 0.00009460, 0.00010772, 0.00029257, 0.00021209, -0.00025385, -0.00025827, -0.00004163, -0.00007446, -0.00000216, 0.00000393, -0.00005353, -0.00001279, -0.00004281, -0.00003080, -0.00003567, -0.00002230, 0.00003067, -0.00003746, -0.00001552, -0.00001908, -0.00002318, 0.00003768, -0.00008610, 0.00005740, 0.00011866, -0.00019034, 0.00004939, 0.00013597, -0.00020967, 0.00003587, 0.00012757, -0.00012611, 0.00000372, 0.00005443, -0.00003088, -0.00000204, 0.00000693, 0.00000994, -0.00005549, -0.00004992, 0.00004892, 0.00001199, -0.00003074, -0.00002428, -0.00002538, 0.00001370, -0.00001398, -0.00007343, -0.00000565, 0.00006038, -0.00000572, -0.00008151, -0.00007765, -0.00002980, -0.00002383, -0.00004368, 0.00005090, 0.00015630, 0.00013317, 0.00007976, 0.00005823, },
		{ 0.00007100, 0.00008268, 0.00007704, 0.00013181, 0.00029552, 0.00008142, -0.00030041, -0.00017546, -0.00001673, -0.00004761, -0.00000214, -0.00003592, -0.00003021, -0.00001184, -0.00005991, -0.00002471, -0.00001721, 0.00000177, -0.00001153, -0.00003776, 0.00001931, -0.00005061, -0.00001871, 0.00002918, -0.00005054, 0.00009279, -0.00000883, -0.00011768, 0.00015159, -0.00005438, -0.00014425, 0.00018499, -0.00003543, -0.00012984, 0.00011312, -0.00000119, -0.00005532, 0.00003024, 0.00001890, -0.00001789, -0.00008070, -0.00000306, 0.00006707, -0.00002690, -0.00004303, -0.00000358, 0.00000498, -0.00000535, -0.00006941, -0.00004772, 0.00007333, 0.00005000, -0.00008611, -0.00010621, -0.00003552, 0.00000525, -0.00001255, 0.00000559, 0.00011846, 0.00014151, 0.00006840, 0.00005265, 0.00006084, 0.00006887, },
		{ 0.00007040, 0.00006732, 0.00017124, 0.00026129, -0.00005145, -0.00029186, -0.00008517, -0.00000550, -0.00004777, -0.00002101, -0.00003581, -0.00000989, -0.00003987, -0.00005441, 0.00000103, -0.00001203, -0.00001216, -0.00003050, 0.00000160, 0.00000158, -0.00008635, 0.00002749, 0.00002173, -0.00004016, 0.00007927, -0.00005113, -0.00001369, 0.00009339, -0.00014334, 0.00003791, 0.00013379, -0.00017591, 0.00001749, 0.00012068, -0.00009945, -0.00000203, 0.00005956, -0.00001304, -0.00004983, -0.00005897, 0.00004923, 0.00003426, -0.00006532, -0.00000958, 0.00002717, -0.00001497, -0.00005789, -0.00006271, 0.00003944, 0.00009363, -0.00004356, -0.00013585, -0.00005254, 0.00001576, 0.00001061, 0.00000312, 0.00006363, 0.00013845, 0.00008141, 0.00002872, 0.00005982, 0.00005882, 0.00005784, 0.00006815, },
		{ 0.00007976, 0.00020248, 0.00018241, -0.00015038, -0.00022725, -0.00001926, -0.00002617, -0.00005519, -0.00002156, -0.00002273, -0.00001946, -0.00005544, -0.00002001, 0.00000522, -0.00002826, -0.00001752, -0.00001600, 0.00001512, -0.00005001, -0.00006414, 0.00007185, -0.00002901, -0.00002239, 0.00008314, -0.00005559, 0.00002236, 0.00001224, -0.00008538, 0.00012643, -0.00003123, -0.00013600, 0.00014909, -0.00000673, -0.00010531, 0.00008991, 0.00001480, -0.00005007, -0.00003652, -0.00001680, 0.00005365, -0.00002165, -0.00004752, 0.00004310, 0.00000164, -0.00007146, -0.00005480, 0.00001555, 0.00008248, 0.00001379, -0.00012265, -0.00008751, 0.00001970, 0.00001694, 0.00000786, 0.00004319, 0.00010061, 0.00010356, 0.00002351, 0.00003385, 0.00007084, 0.00004117, 0.00005127, 0.00006420, 0.00005434, },
		{ 0.00020179, 0.00008653, -0.00018865, -0.00014536, -0.00000268, -0.00005495, -0.00004491, -0.00001480, -0.00002386, -0.00003301, -0.00003948, 0.00000164, -0.00001863, -0.00002872, -0.00000591, -0.00000927, -0.00000383, -0.00007331, -0.00000786, 0.00005385, -0.00007555, 0.00004219, 0.00007044, -0.00007504, 0.00003690, -0.00000689, -0.00001027, 0.00007767, -0.00011574, 0.00001058, 0.00012221, -0.00012009, 0.00000865, 0.00009577, -0.00006886, -0.00001948, -0.00000653, -0.00001186, 0.00002480, -0.00003368, 0.00001122, 0.00003670, -0.00007317, -0.00007102, 0.00002529, 0.00006021, 0.00002135, -0.00007119, -0.00009717, -0.00000099, 0.00002630, -0.00000611, 0.00004194, 0.00008072, 0.00008647, 0.00004933, 0.00000796, 0.00006066, 0.00005703, 0.00001966, 0.00005555, 0.00005541, 0.00004625, 0.00010588, },
		{ 0.00000827, -0.00018037, -0.00008748, -0.00001821, -0.00006327, -0.00002724, -0.00001870, -0.00002788, -0.00002911, -0.00002015, -0.00000757, -0.00003267, -0.00000776, -0.00000875, -0.00001610, -0.00001720, -0.00006459, 0.00002017, -0.00000110, -0.00005668, 0.00010453, 0.00000648, -0.00006967, 0.00007344, -0.00001406, 0.00000476, 0.00001156, -0.00007678, 0.00009021, 0.00000208, -0.00009456, 0.00010486, -0.00000723, -0.00007204, 0.00005231, -0.00003401, -0.00002919, 0.00002777, -0.00000110, 0.00002743, -0.00003814, -0.00009630, 0.00002064, 0.00007643, -0.00000341, -0.00005485, -0.00005463, -0.00001791, 0.00001782, -0.00001366, 0.00001762, 0.00009198, 0.00006704, 0.00004493, 0.00002582, 0.00002615, 0.00006798, 0.00002224, 0.00001649, 0.00006496, 0.00004021, 0.00005036, 0.00012816, 0.00017170, },
		{ -0.00015798, -0.00006010, -0.00003403, -0.00005390, -0.00002021, -0.00002446, -0.00002080, -0.00002159, -0.00001873, -0.00002148, -0.00002102, 0.00000152, -0.00002767, -0.00000881, -0.00002240, -0.00006077, 0.00001590, -0.00003028, 0.00000515, 0.00010437, -0.00005745, -0.00000579, 0.00008531, -0.00005071, 0.00000975, -0.00000083, -0.00002046, 0.00006536, -0.00005970, 0.00000562, 0.00007421, -0.00008642, 0.00002889, 0.00004982, -0.00010426, 0.00000769, 0.00006315, -0.00000070, -0.00002555, -0.00008204, -0.00001306, 0.00009740, 0.00000665, -0.00009075, -0.00002517, 0.00002171, -0.00000455, -0.00002586, -0.00000892, 0.00008262, 0.00008895, 0.00001527, 0.00003377, 0.00003367, 0.00003545, 0.00004698, -0.00000301, 0.00003523, 0.00006344, 0.00002328, 0.00006552, 0.00013724, 0.00012986, -0.00004226, },
		{ -0.00004620, -0.00003903, -0.00004388, -0.00002028, -0.00001958, -0.00001297, -0.00002483, -0.00002279, -0.00002052, -0.00000726, -0.00001205, -0.00003156, 0.00000937, -0.00004622, -0.00006326, 0.00001456, -0.00002590, 0.00005154, 0.00006174, -0.00006183, 0.00006395, 0.00003490, -0.00006909, 0.00004489, -0.00000519, -0.00000742, 0.00002857, -0.00003811, 0.00004623, -0.00001317, -0.00004020, 0.00010417, -0.00005693, -0.00010391, 0.00009989, 0.00004664, -0.00006060, -0.00005991, -0.00002228, 0.00008078, 0.00004545, -0.00010879, -0.00006233, 0.00007791, 0.00002422, -0.00006385, -0.00002923, 0.00005567, 0.00010599, 0.00002655, -0.00000411, 0.00006008, 0.00002723, 0.00002315, 0.00002557, -0.00000394, 0.00005559, 0.00004496, 0.00001618, 0.00008604, 0.00013241, 0.00008904, -0.00007518, -0.00013880, },
		{ -0.00004003, -0.00003773, -0.00001580, -0.00001075, -0.00001644, -0.00003003, -0.00001922, -0.00001532, -0.00000735, -0.00002483, -0.00001365, 0.00000590, -0.00008329, -0.00004622, 0.00001854, -0.00001613, 0.00007451, 0.00002915, -0.00002560, 0.00007873, -0.00002925, -0.00002936, 0.00006611, -0.00003595, 0.00000418, 0.00002462, -0.00001838, 0.00001952, -0.00002953, 0.00005582, 0.00005129, -0.00014196, 0.00001031, 0.00013319, -0.00005321, -0.00009174, 0.00000046, 0.00006372, 0.00005556, -0.00008032, -0.00011171, 0.00006803, 0.00010312, -0.00006895, -0.00008060, 0.00004011, 0.00009320, 0.00005769, -0.00001901, 0.00003329, 0.00006739, -0.00000815, 0.00001993, 0.00001810, 0.00000494, 0.00006294, 0.00002083, 0.00002443, 0.00010285, 0.00011793, 0.00005159, -0.00010146, -0.00012074, -0.00003325, },
		{ -0.00002940, -0.00000719, -0.00000990, -0.00002500, -0.00002675, -0.00001520, -0.00001606, -0.00001023, -0.00002361, -0.00000005, -0.00002528, -0.00010036, -0.00001514, 0.00000732, 0.00000139, 0.00009588, 0.00001059, 0.00000186, 0.00005346, -0.00004818, 0.00002709, 0.00003714, -0.00005185, 0.00003830, 0.00000959, -0.00002731, 0.00001450, 0.00002002, 0.00007164, -0.00005404, -0.00009513, 0.00011803, 0.00003700, -0.00012928, -0.00001327, 0.00008628, 0.00004408, -0.00005938, -0.00011304, 0.00001262, 0.00013796, 0.00000146, -0.00012981, -0.00000131, 0.00009064, 0.00005615, 0.00000847, -0.00000258, 0.00006910, 0.00002533, -0.00002988, 0.00003960, 0.00000641, 0.00001430, 0.00006102, 0.00000070, 0.00004333, 0.00011164, 0.00009792, 0.00001330, -0.00012211, -0.00009621, -0.00002194, -0.00004256, },
		{ -0.00000306, -0.00001663, -0.00002774, -0.00002047, -0.00001818, -0.00001495, -0.00000786, -0.00002135, -0.00000456, -0.00005780, -0.00009074, -0.00000010, -0.00000755, 0.00004590, 0.00010502, -0.00001550, 0.00001810, 0.00003269, -0.00003201, 0.00004890, -0.00000526, -0.00002498, 0.00005097, -0.00002894, -0.00000952, 0.00004705, 0.00003656, 0.00001675, -0.00007636, 0.00001268, 0.00009180, -0.00008534, -0.00006965, 0.00010180, 0.00006425, -0.00006933, -0.00010058, -0.00000888, 0.00010405, 0.00006651, -0.00009942, -0.00007345, 0.00007972, 0.00005531, 0.00001183, 0.00001431, 0.00002785, 0.00005552, -0.00002357, -0.00000834, 0.00005084, -0.00001646, 0.00003421, 0.00004907, -0.00001373, 0.00007058, 0.00011128, 0.00007019, -0.00002685, -0.00012888, -0.00006498, -0.00001866, -0.00004383, -0.00001618, },
		{ -0.00002255, -0.00002471, -0.00001972, -0.00002127, -0.00000632, -0.00000988, -0.00002657, -0.00001477, -0.00007760, -0.00007819, -0.00000288, 0.00000693, 0.00009989, 0.00007642, -0.00003607, 0.00004401, 0.00001840, -0.00001645, 0.00004542, -0.00002092, 0.00000622, 0.00002354, -0.00003600, 0.00004301, 0.00004719, 0.00000657, -0.00000070, -0.00002507, 0.00003530, -0.00001302, -0.00008084, 0.00006521, 0.00010307, -0.00006261, -0.00011944, -0.00000158, 0.00008281, 0.00005543, -0.00003192, -0.00008627, 0.00001812, 0.00006866, -0.00000223, 0.00002772, 0.00003193, 0.00002088, 0.00002264, -0.00003505, 0.00002720, 0.00003053, -0.00002516, 0.00006097, 0.00001734, -0.00001104, 0.00010383, 0.00009313, 0.00003414, -0.00005865, -0.00011677, -0.00003747, -0.00002438, -0.00003739, -0.00000328, -0.00000763, },
		},
		{ { 0.00038930, 0.00007774, -0.00011589, -0.00003351, 0.00002420, -0.00000282, 0.00006542, 0.00006420, 0.00008239, -0.00001387, -0.00004442, -0.00011592, -0.00000519, -0.00000060, -0.00003403, -0.00000735, 0.00001345, -0.00004970, 0.00003868, -0.00003782, -0.00000358, 0.00000238, -0.00000521, -0.00002870, -0.00001323, -0.00000530, 0.00007486, -0.00004268, -0.00002713, -0.00000387, 0.00001561, -0.00001578, 0.00002820, -0.00001578, 0.00001561, -0.00000387, -0.00002713, -0.00004268, 0.00007486, -0.00000530, -0.00001323, -0.00002870, -0.00000521, 0.00000238, -0.00000358, -0.00003782, 0.00003868, -0.00004970, 0.00001345, -0.00000735, -0.00003403, -0.00000060, -0.00000519, -0.00011592, -0.00004442, -0.00001387, 0.00008239, 0.00006420, 0.00006542, -0.00000282, 0.00002420, -0.00003351, -0.00011589, 0.00007774, },
		{ 0.00008662, -0.00009693, -0.00003430, 0.00003832, -0.00001042, 0.00009072, 0.00008187, 0.00006710, -0.00002664, -0.00007310, -0.00011050, 0.00002337, -0.00003074, -0.00002243, 0.00000569, -0.00001272, -0.00002847, 0.00003362, -0.00005162, 0.00001439, -0.00000822, -0.00000629, -0.00002862, -0.00001560, 0.00002106, 0.00005694, -0.00005957, -0.00001547, 0.00000409, 0.00000464, -0.00000430, 0.00002329, -0.00002021, 0.00002526, -0.00001779, -0.00003165, -0.00001987, 0.00008101, -0.00003278, -0.00000231, -0.00003460, 0.00000248, 0.00000536, -0.00002011, -0.00001462, 0.00002854, -0.00006018, 0.00004089, -0.00003470, -0.00002628, 0.00001906, -0.00004313, -0.00009932, -0.00002146, -0.00000856, 0.00009592, 0.00004125, 0.00004787, 0.00000440, 0.00000831, -0.00002847, -0.00013505, 0.00006549, 0.00039156, },
		{ -0.00007652, -0.00003142, 0.00004934, -0.00001526, 0.00012174, 0.00009405, 0.00004813, -0.00005017, -0.00009585, -0.00008786, 0.00003395, -0.00005833, 0.00000127, 0.00000191, -0.00002823, -0.00000612, 0.00001739, -0.00005284, 0.00002722, -0.00002139, -0.00000338, -0.00003262, -0.00000801, 0.00004021, 0.00003099, -0.00006714, -0.00000039, 0.00000445, -0.00000238, 0.00000899, 0.00001051, -0.00001501, 0.00002812, -0.00003325, -0.00002758, 0.00000427, 0.00007432, -0.00005518, 0.00001311, -0.00004529, 0.00001712, -0.00000274, -0.00002755, 0.00001014, 0.00000521, -0.00005361, 0.00005803, -0.00006666, 0.00000037, 0.00001756, -0.00007615, -0.00006490, -0.00001261, -0.00000380, 0.00010590, 0.00001490, 0.00003905, 0.00000793, -0.00000658, -0.00002024, -0.00015550, 0.00005103, 0.00039767, 0.00009177, },
		{ -0.00002617, 0.00005734, -0.00001516, 0.00015678, 0.00010005, 0.00002122, -0.00008159, -0.00010321, -0.00005897, 0.00002633, -0.00007354, 0.00002621, -0.00001414, -0.00002934, 0.00000863, -0.00000236, -0.00004331, 0.00003142, -0.00003269, 0.00000010, -0.00003704, 0.00000775, 0.00004777, 0.00000255, -0.00006427, 0.00001348, -0.00000135, -0.00000245, 0.00001836, -0.00000507, -0.00000143, 0.00002072, -0.00004459, -0.00001620, 0.00002528, 0.00005643, -0.00006784, 0.00002778, -0.00005738, 0.00003606, -0.00002194, -0.00002017, 0.00002640, -0.00002308, -0.00003045, 0.00005608, -0.00008972, 0.00003666, -0.00000806, -0.00008995, -0.00002471, -0.00001898, 0.00000655, 0.00010816, -0.00001052, 0.00003746, 0.00000586, -0.00001731, -0.00001158, -0.00017666, 0.00003509, 0.00040616, 0.00009421, -0.00005360, },
		{ 0.00006296, -0.00000847, 0.00019466, 0.00009685, -0.00001693, -0.00011258, -0.00009215, -0.00003574, 0.00000753, -0.00007328, 0.00004291, -0.00003377, -0.00001852, 0.00001125, -0.00001787, -0.00002829, 0.00002719, -0.00003969, 0.00000166, -0.00003782, 0.00002722, 0.00004234, -0.00002244, -0.00005252, 0.00002225, -0.00000960, 0.00000419, 0.00001990, -0.00001726, 0.00001576, 0.00000313, -0.00004686, -0.00000107, 0.00004001, 0.00003101, -0.00006865, 0.00003688, -0.00006632, 0.00005420, -0.00004804, 0.00000319, 0.00002565, -0.00004439, 0.00000111, 0.00003353, -0.00009212, 0.00006740, -0.00005049, -0.00007665, 0.00000616, -0.00003380, 0.00002390, 0.00009931, -0.00002948, 0.00003894, -0.00000103, -0.00002209, -0.00000574, -0.00019626, 0.00001685, 0.00041603, 0.00009578, -0.00002828, -0.00001948, },
		{ 0.00000693, 0.00023330, 0.00007896, -0.00006515, -0.00013295, -0.00006727, -0.00002594, -0.00001226, -0.00006127, 0.00004690, -0.00004837, -0.00000253, 0.00000250, -0.00002414, -0.00001396, 0.00001740, -0.00004242, 0.00000077, -0.00003219, 0.00004465, 0.00002579, -0.00003916, -0.00003556, 0.00002403, -0.00001596, 0.00001435, 0.00001301, -0.00002118, 0.00003001, -0.00002074, -0.00003769, 0.00001323, 0.00004723, 0.00000268, -0.00005827, 0.00003743, -0.00006809, 0.00006570, -0.00007355, 0.00003788, 0.00000464, -0.00004756, 0.00002751, -0.00000190, -0.00006959, 0.00007746, -0.00009402, -0.00003896, 0.00001624, -0.00004573, 0.00004365, 0.00007972, -0.00003787, 0.00003821, -0.00000920, -0.00002190, -0.00000462, -0.00021153, -0.00000607, 0.00042780, 0.00009793, -0.00000123, -0.00001168, 0.00006667, },
		{ 0.00026761, 0.00004075, -0.00011619, -0.00013561, -0.00003834, -0.00003044, -0.00002453, -0.00004529, 0.00003971, -0.00005292, 0.00001106, -0.00001272, -0.00002034, -0.00000498, 0.00000589, -0.00004250, -0.00000106, -0.00001990, 0.00005488, 0.00000267, -0.00004518, -0.00001809, 0.00001951, -0.00001726, 0.00002334, 0.00000059, -0.00001525, 0.00003538, -0.00004431, -0.00001820, 0.00002268, 0.00004756, -0.00002413, -0.00003962, 0.00002902, -0.00006070, 0.00006601, -0.00008996, 0.00007438, -0.00003253, -0.00002694, 0.00003485, -0.00003519, -0.00002855, 0.00005871, -0.00012042, 0.00000998, 0.00000293, -0.00004444, 0.00005717, 0.00005471, -0.00003552, 0.00003143, -0.00001409, -0.00002036, -0.00000739, -0.00022109, -0.00003636, 0.00044345, 0.00010083, 0.00002735, -0.00000318, 0.00006917, 0.00003400, },
		{ -0.00001963, -0.00015835, -0.00012016, -0.00001564, -0.00004387, -0.00002585, -0.00003306, 0.00002705, -0.00004740, 0.00001703, -0.00002800, -0.00000928, -0.00000326, -0.00000424, -0.00004184, -0.00000120, -0.00000335, 0.00005491, -0.00002124, -0.00004087, -0.00000431, 0.00001150, -0.00001270, 0.00002704, -0.00001224, -0.00000177, 0.00002888, -0.00006052, 0.00000726, 0.00002508, 0.00004280, -0.00004600, -0.00001685, 0.00001351, -0.00004480, 0.00005363, -0.00009068, 0.00010119, -0.00007509, 0.00001442, 0.00001476, -0.00004939, 0.00001506, 0.00001472, -0.00011668, 0.00005150, -0.00002582, -0.00002604, 0.00005630, 0.00003276, -0.00002688, 0.00001859, -0.00001277, -0.00002196, -0.00001064, -0.00022625, -0.00007520, 0.00046519, 0.00010350, 0.00005836, 0.00000455, 0.00007268, 0.00007537, 0.00028828, },
		{ -0.00017991, -0.00009353, -0.00000535, -0.00005804, -0.00001841, -0.00002886, 0.00001559, -0.00003586, 0.00001418, -0.00003815, 0.00000434, -0.00000799, -0.00001155, -0.00004141, 0.00000262, 0.00001325, 0.00004474, -0.00004058, -0.00002905, 0.00000326, 0.00000373, -0.00000408, 0.00002352, -0.00002020, 0.00001391, 0.00001164, -0.00006420, 0.00003254, 0.00002058, 0.00003515, -0.00006093, 0.00000577, -0.00000573, -0.00002334, 0.00003072, -0.00007345, 0.00010877, -0.00010886, 0.00006493, -0.00003128, -0.00003267, 0.00004116, -0.00003902, -0.00008144, 0.00006919, -0.00005486, 0.00000450, 0.00003800, 0.00002169, -0.00001923, 0.00000391, -0.00000611, -0.00002935, -0.00001027, -0.00023080, -0.00012098, 0.00049357, 0.00010522, 0.00009328, 0.00000875, 0.00008243, 0.00013044, 0.00028298, -0.00009682, },
		{ -0.00006703, -0.00000711, -0.00006624, -0.00000792, -0.00003228, 0.00000975, -0.00002385, 0.00000501, -0.00004089, 0.00001606, -0.00001680, -0.00001605, -0.00004077, 0.00001109, 0.00002550, 0.00002719, -0.00005184, -0.00001390, 0.00000438, -0.00000062, 0.00000507, 0.00001386, -0.00001997, 0.00002546, -0.00001135, -0.00005386, 0.00005164, 0.00001137, 0.00002640, -0.00006828, 0.00002470, -0.00002503, -0.00000045, 0.00000240, -0.00004132, 0.00009300, -0.00012117, 0.00010784, -0.00009112, 0.00001598, 0.00003357, -0.00008002, -0.00002699, 0.00005637, -0.00006861, 0.00003528, 0.00000673, 0.00002459, -0.00001905, -0.00000606, 0.00000094, -0.00004118, -0.00000426, -0.00023899, -0.00016965, 0.00052641, 0.00010736, 0.00013230, 0.00000681, 0.00010680, 0.00019214, 0.00024036, -0.00017725, -0.00017439, },
		{ -0.00001495, -0.00006625, -0.00000009, -0.00003944, 0.00001026, -0.00001572, -0.00000609, -0.00003702, 0.00002308, -0.00002711, -0.00001843, -0.00003851, 0.00002280, 0.00003029, 0.00000675, -0.00005409, 0.00000041, 0.00000083, -0.00000011, 0.00001100, 0.00000166, -0.00001153, 0.00002793, -0.00003299, -0.00003213, 0.00006062, 0.00000063, 0.00001768, -0.00006850, 0.00003766, -0.00004132, 0.00001999, -0.00002501, -0.00000165, 0.00005679, -0.00010550, 0.00012711, -0.00014537, 0.00008468, -0.00001259, -0.00008740, 0.00002441, 0.00001996, -0.00005858, 0.00005307, -0.00002702, 0.00003786, -0.00002825, -0.00000614, 0.00000148, -0.00005199, 0.00000519, -0.00025262, -0.00021698, 0.00055938, 0.00011424, 0.00017235, -0.00000075, 0.00015480, 0.00024523, 0.00015577, -0.00024209, -0.00014433, -0.00005130, },
		{ -0.00006076, 0.00000226, -0.00004566, 0.00001463, -0.00001299, -0.00001508, -0.00002931, 0.00002463, -0.00003707, -0.00001907, -0.00003334, 0.00003478, 0.00002692, -0.00001191, -0.00004872, 0.00001094, -0.00000446, 0.00000459, 0.00001127, -0.00000847, 0.00000187, 0.00001966, -0.00004646, -0.00000486, 0.00005861, -0.00000870, 0.00000951, -0.00006272, 0.00004364, -0.00005269, 0.00003548, -0.00004614, 0.00003633, 0.00000922, -0.00006407, 0.00011335, -0.00017405, 0.00015168, -0.00008727, -0.00005084, 0.00004884, -0.00002149, -0.00002813, 0.00005007, -0.00005155, 0.00005261, -0.00004261, 0.00000398, -0.00000945, -0.00005460, 0.00001166, -0.00026896, -0.00026157, 0.00058861, 0.00013188, 0.00020654, -0.00000725, 0.00023117, 0.00026827, 0.00003620, -0.00027339, -0.00010125, -0.00005138, -0.00002086, },
		{ 0.00000024, -0.00004809, 0.00001915, -0.00001438, -0.00001974, -0.00002092, 0.00002140, -0.00004569, -0.00001763, -0.00002503, 0.00004365, 0.00001706, -0.00002536, -0.00003869, 0.00001652, -0.00000866, 0.00001110, 0.00000574, -0.00001275, 0.00001502, 0.00000287, -0.00004751, 0.00002088, 0.00004777, -0.00001489, 0.00000219, -0.00005257, 0.00004278, -0.00005840, 0.00004525, -0.00005816, 0.00006465, -0.00003762, -0.00000727, 0.00006762, -0.00016395, 0.00019253, -0.00016745, 0.00002377, 0.00003064, -0.00004492, 0.00000799, 0.00002868, -0.00006002, 0.00005930, -0.00005351, 0.00001846, -0.00003134, -0.00004426, 0.00000726, -0.00028087, -0.00030697, 0.00061390, 0.00016475, 0.00022639, 0.00000118, 0.00033058, 0.00023955, -0.00009843, -0.00026131, -0.00006163, -0.00006446, -0.00001923, -0.00005512, },
		{ -0.00004692, 0.00002113, -0.00001719, -0.00002005, -0.00001416, 0.00001456, -0.00005237, -0.00001332, -0.00001494, 0.00004702, 0.00000398, -0.00003219, -0.00002725, 0.00001777, -0.00001000, 0.00001644, -0.00000344, -0.00000970, 0.00002272, -0.00001716, -0.00003579, 0.00003901, 0.00003228, -0.00001778, -0.00000387, -0.00003998, 0.00003607, -0.00005855, 0.00005003, -0.00006134, 0.00007906, -0.00007270, 0.00004989, 0.00000149, -0.00011400, 0.00018903, -0.00022445, 0.00011468, -0.00003009, -0.00003264, 0.00002945, 0.00000160, -0.00005475, 0.00005334, -0.00005255, 0.00002743, -0.00005713, -0.00002225, -0.00001297, -0.00027995, -0.00036113, 0.00064077, 0.00021182, 0.00022670, 0.00004265, 0.00043392, 0.00014587, -0.00021909, -0.00020914, -0.00004023, -0.00008141, -0.00000984, -0.00005388, -0.00000243, },
		{ 0.00002015, -0.00001903, -0.00001737, -0.00001021, 0.00000518, -0.00005624, -0.00000572, -0.00000556, 0.00004441, -0.00000882, -0.00003290, -0.00001695, 0.00001636, -0.00000829, 0.00001822, -0.00001277, -0.00000067, 0.00002177, -0.00003398, -0.00001482, 0.00004619, 0.00001674, -0.00001845, -0.00000797, -0.00002710, 0.00002517, -0.00005376, 0.00005116, -0.00005832, 0.00008002, -0.00008973, 0.00009302, -0.00006682, -0.00003634, 0.00013641, -0.00023410, 0.00019080, -0.00011583, 0.00001926, 0.00001939, -0.00001309, -0.00004735, 0.00003897, -0.00003723, 0.00002208, -0.00007574, 0.00000303, -0.00004719, -0.00026136, -0.00043285, 0.00067936, 0.00026447, 0.00021141, 0.00013235, 0.00050953, -0.00000939, -0.00029627, -0.00013349, -0.00004376, -0.00009135, 0.00000220, -0.00005794, -0.00000210, -0.00004472, },
		{ -0.00001896, -0.00001322, -0.00000948, -0.00000584, -0.00005604, 0.00000441, 0.00000052, 0.00003726, -0.00001877, -0.00002924, -0.00000910, 0.00001418, -0.00000470, 0.00001561, -0.00001875, 0.00001071, 0.00001209, -0.00004222, 0.00000942, 0.00004249, 0.00000465, -0.00001848, -0.00000934, -0.00001605, 0.00001221, -0.00004494, 0.00004981, -0.00005265, 0.00007184, -0.00008883, 0.00011332, -0.00011879, 0.00004750, 0.00004592, -0.00018595, 0.00022256, -0.00019674, 0.00009760, -0.00002716, -0.00000022, -0.00005393, 0.00002891, -0.00001438, 0.00000056, -0.00007769, 0.00001906, -0.00008582, -0.00022789, -0.00052627, 0.00073977, 0.00030833, 0.00019721, 0.00027317, 0.00052055, -0.00020258, -0.00031049, -0.00005909, -0.00006783, -0.00008728, 0.00000932, -0.00006392, 0.00000232, -0.00004439, 0.00001779, },
		{ -0.00000852, -0.00001230, -0.00001724, -0.00005057, 0.00001487, 0.00000190, 0.00002815, -0.00002486, -0.00002323, -0.00000383, 0.00001256, -0.00000108, 0.00000953, -0.00001928, 0.00001998, -0.00000327, -0.00003934, 0.00003050, 0.00003092, -0.00000244, -0.00001905, -0.00000748, -0.00000867, -0.00000040, -0.00003328, 0.00004653, -0.00004713, 0.00006047, -0.00007567, 0.00011017, -0.00014204, 0.00011397, -0.00005835, -0.00008799, 0.00019279, -0.00024098, 0.00017559, -0.00009925, 0.00004451, -0.00008727, 0.00003933, 0.00000057, -0.00002853, -0.00006149, 0.00001543, -0.00011492, -0.00019087, -0.00063636, 0.00082525, 0.00032987, 0.00021184, 0.00044774, 0.00043731, -0.00039320, -0.00026060, -0.00000963, -0.00009874, -0.00006983, 0.00000594, -0.00006607, 0.00000849, -0.00004706, 0.00001620, -0.00001750, },
		{ -0.00001916, -0.00002696, -0.00003958, 0.00002267, -0.00000097, 0.00001956, -0.00002751, -0.00001636, -0.00000056, 0.00001192, 0.00000090, 0.00000219, -0.00001434, 0.00002354, -0.00001932, -0.00002632, 0.00004354, 0.00001602, -0.00000500, -0.00002043, -0.00000264, -0.00000605, -0.00001033, -0.00002034, 0.00004132, -0.00004286, 0.00005092, -0.00005873, 0.00009075, -0.00013446, 0.00014566, -0.00014748, 0.00003527, 0.00010368, -0.00022659, 0.00022214, -0.00017244, 0.00011043, -0.00014912, 0.00008180, -0.00001064, -0.00004729, -0.00003735, -0.00000918, -0.00012324, -0.00016604, -0.00074841, 0.00092672, 0.00032611, 0.00028568, 0.00061646, 0.00025055, -0.00053469, -0.00016607, 0.00000159, -0.00011959, -0.00004731, -0.00000807, -0.00005971, 0.00001187, -0.00005116, 0.00001637, -0.00001563, -0.00000363, },
		{ -0.00003222, -0.00002465, 0.00002521, -0.00000594, 0.00001285, -0.00002783, -0.00000931, 0.00000136, 0.00001186, 0.00000045, -0.00000383, -0.00000595, 0.00001988, -0.00003094, -0.00000709, 0.00004653, 0.00000222, -0.00000484, -0.00002194, 0.00000407, -0.00000818, -0.00001576, -0.00000805, 0.00003418, -0.00003918, 0.00004543, -0.00004556, 0.00006655, -0.00010433, 0.00013723, -0.00019762, 0.00015067, -0.00002310, -0.00015010, 0.00021361, -0.00021720, 0.00017412, -0.00022659, 0.00015562, -0.00006036, -0.00003502, -0.00002581, -0.00004281, -0.00011001, -0.00016539, -0.00084315, 0.00102212, 0.00031308, 0.00043800, 0.00072431, -0.00001964, -0.00058938, -0.00006149, -0.00002268, -0.00011824, -0.00003179, -0.00002663, -0.00004432, 0.00000847, -0.00005323, 0.00001736, -0.00001365, 0.00000096, -0.00003031, },
		{ -0.00000920, 0.00002167, -0.00001016, 0.00000782, -0.00002669, -0.00000217, 0.00000223, 0.00001167, -0.00000213, -0.00000661, 0.00000275, 0.00001010, -0.00003457, 0.00001295, 0.00004049, -0.00000751, -0.00000407, -0.00002235, 0.00001085, -0.00001384, -0.00001583, 0.00000153, 0.00002571, -0.00003454, 0.00004307, -0.00003990, 0.00004856, -0.00006665, 0.00009684, -0.00019805, 0.00022757, -0.00015333, -0.00002869, 0.00014330, -0.00021015, 0.00020739, -0.00029451, 0.00024439, -0.00014692, 0.00002211, -0.00005045, -0.00006194, -0.00008962, -0.00018818, -0.00090698, 0.00108321, 0.00032860, 0.00066251, 0.00071634, -0.00032403, -0.00054256, 0.00001528, -0.00006454, -0.00009328, -0.00003300, -0.00004008, -0.00002451, -0.00000252, -0.00005001, 0.00001681, -0.00001062, 0.00000386, -0.00004489, -0.00003042, },
		{ 0.00001359, -0.00001150, 0.00000319, -0.00002418, 0.00000500, 0.00000194, 0.00001076, -0.00000571, -0.00000556, 0.00000867, -0.00000270, -0.00002930, 0.00002874, 0.00002869, -0.00001223, -0.00000419, -0.00002062, 0.00001575, -0.00002089, -0.00001103, 0.00000671, 0.00001733, -0.00002782, 0.00004086, -0.00004069, 0.00004300, -0.00003706, 0.00004257, -0.00015451, 0.00024808, -0.00025174, 0.00010521, 0.00002494, -0.00014373, 0.00018828, -0.00032385, 0.00031906, -0.00025104, 0.00012199, -0.00012705, -0.00003954, -0.00008969, -0.00021548, -0.00094332, 0.00108905, 0.00042534, 0.00091824, 0.00055757, -0.00059529, -0.00040973, 0.00003707, -0.00009829, -0.00005540, -0.00005287, -0.00004002, -0.00000828, -0.00001763, -0.00004071, 0.00001258, -0.00000493, 0.00000273, -0.00006022, -0.00002047, 0.00000239, },
		{ -0.00000964, -0.00000237, -0.00001963, 0.00001172, 0.00000023, 0.00000907, -0.00000885, -0.00000149, 0.00000986, -0.00001452, -0.00001690, 0.00003700, 0.00001525, -0.00001283, -0.00000555, -0.00001640, 0.00001731, -0.00002680, -0.00000308, 0.00000673, 0.00001103, -0.00001929, 0.00003576, -0.00004316, 0.00004922, -0.00002581, -0.00000468, -0.00008646, 0.00021271, -0.00029379, 0.00021485, -0.00011139, -0.00003018, 0.00011062, -0.00029302, 0.00034713, -0.00033940, 0.00024262, -0.00025336, 0.00004252, -0.00014172, -0.00021243, -0.00097977, 0.00104204, 0.00065405, 0.00113158, 0.00025072, -0.00076754, -0.00023352, -0.00000194, -0.00010148, -0.00002293, -0.00008368, -0.00002406, -0.00000300, -0.00003037, -0.00002794, 0.00000449, 0.00000428, -0.00000485, -0.00007193, -0.00000399, 0.00000648, 0.00000456, },
		{ -0.00000923, -0.00001219, 0.00001690, -0.00000290, 0.00000699, -0.00001045, 0.00000382, 0.00000615, -0.00002194, -0.00000105, 0.00003699, 0.00000373, -0.00001126, -0.00000750, -0.00001031, 0.00001510, -0.00002947, 0.00000542, 0.00000220, 0.00000862, -0.00001079, 0.00002659, -0.00004167, 0.00006068, -0.00003430, -0.00002895, -0.00001962, 0.00013930, -0.00027354, 0.00027169, -0.00022934, 0.00010178, -0.00001196, -0.00019758, 0.00030495, -0.00037485, 0.00034591, -0.00040441, 0.00018179, -0.00026680, -0.00013944, -0.00106617, 0.00097993, 0.00104103, 0.00121153, -0.00015581, -0.00079837, -0.00007024, -0.00008355, -0.00006506, -0.00001361, -0.00011076, 0.00000212, -0.00001139, -0.00003443, -0.00001683, -0.00000497, 0.00001577, -0.00001976, -0.00007523, 0.00001449, 0.00000194, -0.00000129, -0.00000631, },
		{ -0.00000170, 0.00001922, -0.00000689, 0.00000514, -0.00001012, 0.00000837, -0.00000102, -0.00002309, 0.00001404, 0.00003027, -0.00000380, -0.00000948, -0.00000883, -0.00000372, 0.00000985, -0.00002785, 0.00001181, -0.00000495, 0.00001088, -0.00000501, 0.00001485, -0.00003266, 0.00006831, -0.00005517, -0.00002478, 0.00002351, 0.00005599, -0.00020468, 0.00026462, -0.00029879, 0.00021666, -0.00014940, -0.00005432, 0.00018823, -0.00033020, 0.00038910, -0.00053651, 0.00034950, -0.00046161, 0.00003038, -0.00126121, 0.00097708, 0.00156717, 0.00107575, -0.00057444, -0.00068543, 0.00002952, -0.00017242, 0.00000184, -0.00003586, -0.00011897, 0.00002650, -0.00002936, -0.00002718, -0.00001238, -0.00001151, 0.00002581, -0.00003996, -0.00006697, 0.00002869, -0.00000897, -0.00000116, -0.00000446, -0.00001619, },
		{ 0.00001772, -0.00001066, 0.00000399, -0.00000819, 0.00001066, -0.00000925, -0.00001811, 0.00002491, 0.00001980, -0.00000702, -0.00000875, -0.00000847, 0.00000178, 0.00000321, -0.00002227, 0.00001416, -0.00001200, 0.00001696, -0.00000417, 0.00000430, -0.00001661, 0.00006483, -0.00007601, 0.00000161, 0.00003132, -0.00000915, -0.00011462, 0.00020282, -0.00030575, 0.00028449, -0.00026514, 0.00010273, 0.00001642, -0.00020000, 0.00034020, -0.00060000, 0.00049471, -0.00069119, 0.00029343, -0.00161013, 0.00113111, 0.00215757, 0.00068005, -0.00089966, -0.00047069, 0.00003955, -0.00023027, 0.00007506, -0.00008389, -0.00010016, 0.00003598, -0.00004713, -0.00001164, -0.00001655, -0.00001139, 0.00002935, -0.00006013, -0.00004762, 0.00003290, -0.00002104, 0.00000500, -0.00000677, -0.00002086, 0.00001074, },
		{ -0.00001301, 0.00000355, -0.00000542, 0.00001014, -0.00001605, -0.00000882, 0.00002974, 0.00000882, -0.00000698, -0.00000931, -0.00000595, 0.00000493, -0.00000287, -0.00001430, 0.00001192, -0.00001630, 0.00002449, -0.00000877, -0.00000087, 0.00000198, 0.00004808, -0.00008490, 0.00003560, 0.00000634, -0.00003778, -0.00003424, 0.00011130, -0.00025629, 0.00029100, -0.00032873, 0.00023290, -0.00017164, -0.00000575, 0.00019166, -0.00055671, 0.00055860, -0.00089308, 0.00060519, -0.00212019, 0.00153622, 0.00268802, 0.00004125, -0.00103839, -0.00022975, -0.00003159, -0.00023187, 0.00012543, -0.00013895, -0.00005786, 0.00002268, -0.00005349, 0.00000425, -0.00002657, -0.00000396, 0.00002246, -0.00007312, -0.00002198, 0.00002445, -0.00002803, 0.00001410, -0.00001404, -0.00002067, 0.00002279, 0.00001242, },
		{ 0.00000340, -0.00000262, 0.00000716, -0.00001965, 0.00000205, 0.00002853, -0.00000012, -0.00000538, -0.00001053, -0.00000161, 0.00000525, -0.00000683, -0.00000612, 0.00000604, -0.00001626, 0.00003035, -0.00001697, 0.00000194, 0.00001648, 0.00002201, -0.00007538, 0.00006061, -0.00003699, -0.00002668, 0.00001298, 0.00002117, -0.00017302, 0.00024191, -0.00032645, 0.00030268, -0.00032954, 0.00020865, -0.00003335, -0.00039599, 0.00049470, -0.00099364, 0.00088336, -0.00274329, 0.00224978, 0.00300979, -0.00075569, -0.00094014, -0.00004852, -0.00014269, -0.00017654, 0.00013172, -0.00017688, -0.00000700, -0.00001153, -0.00004120, 0.00001113, -0.00003561, 0.00000700, 0.00000496, -0.00007269, 0.00000182, 0.00000536, -0.00002529, 0.00002111, -0.00002449, -0.00001428, 0.00003153, 0.00000462, -0.00001307, },
		{ -0.00000029, 0.00000272, -0.00001938, 0.00001180, 0.00002275, -0.00000572, -0.00000384, -0.00001135, 0.00000361, 0.00000309, -0.00000794, 0.00000020, -0.00000137, -0.00001191, 0.00003189, -0.00002513, 0.00001225, 0.00002110, -0.00000496, -0.00004902, 0.00006477, -0.00007884, 0.00001271, 0.00001812, -0.00004146, -0.00008598, 0.00015968, -0.00026567, 0.00029724, -0.00041862, 0.00038995, -0.00028529, -0.00014276, 0.00028820, -0.00093239, 0.00102622, -0.00337387, 0.00326312, 0.00298761, -0.00157100, -0.00061646, 0.00000505, -0.00023455, -0.00008994, 0.00009114, -0.00017881, 0.00003193, -0.00005477, -0.00001114, 0.00000287, -0.00003583, 0.00001409, -0.00001827, -0.00005662, 0.00001517, -0.00001787, -0.00001194, 0.00002114, -0.00003410, -0.00000250, 0.00003458, -0.00000342, -0.00001085, 0.00000292, },
		{ -0.00000204, -0.00001566, 0.00001843, 0.00001458, -0.00000796, -0.00000332, -0.00001077, 0.00000852, -0.00000059, -0.00000648, 0.00000345, -0.00000787, -0.00000493, 0.00002802, -0.00002914, 0.00002631, 0.00001352, -0.00002359, -0.00001451, 0.00004536, -0.00010196, 0.00006008, -0.00001182, -0.00006372, -0.00002156, 0.00007448, -0.00017152, 0.00022528, -0.00042111, 0.00049283, -0.00050266, 0.00014662, -0.00003357, -0.00068624, 0.00094250, -0.00386635, 0.00448720, 0.00254005, -0.00224203, -0.00014224, -0.00009532, -0.00025213, -0.00001506, 0.00002250, -0.00014052, 0.00004187, -0.00008969, 0.00002714, -0.00001978, -0.00002259, 0.00000939, -0.00003815, -0.00002866, 0.00001253, -0.00003646, 0.00000833, 0.00001179, -0.00003818, 0.00001158, 0.00003119, -0.00000921, -0.00000724, 0.00000170, 0.00000144, },
		{ -0.00000963, 0.00002105, 0.00000625, -0.00000769, -0.00000404, -0.00000821, 0.00001205, -0.00000467, -0.00000347, 0.00000357, -0.00001157, 0.00000205, 0.00001978, -0.00002611, 0.00003840, -0.00000387, -0.00002768, 0.00001616, 0.00000929, -0.00009805, 0.00009469, -0.00005844, -0.00004868, 0.00000618, 0.00001281, -0.00007703, 0.00011537, -0.00034595, 0.00049510, -0.00063164, 0.00040017, -0.00040448, -0.00028454, 0.00058375, -0.00406926, 0.00576068, 0.00166969, -0.00262532, 0.00036294, -0.00032395, -0.00016611, 0.00000450, -0.00003973, -0.00007618, 0.00001764, -0.00010115, 0.00005905, -0.00004864, 0.00000243, -0.00001144, -0.00004465, 0.00000194, -0.00000577, -0.00004277, 0.00002843, -0.00000551, -0.00003336, 0.00002350, 0.00002270, -0.00001100, -0.00000378, -0.00000019, 0.00000272, -0.00000624, },
		{ 0.00001977, -0.00000059, -0.00000621, -0.00000551, -0.00000362, 0.00001348, -0.00000817, -0.00000016, 0.00000155, -0.00001179, 0.00000641, 0.00001005, -0.00001571, 0.00004307, -0.00002478, -0.00001669, 0.00003269, -0.00003021, -0.00007008, 0.00010314, -0.00010018, -0.00001182, -0.00000127, -0.00001149, -0.00001102, 0.00000523, -0.00022514, 0.00040509, -0.00064396, 0.00055285, -0.00073709, 0.00019204, -0.00003197, -0.00386749, 0.00688075, 0.00047218, -0.00263601, 0.00076808, -0.00061024, 0.00001438, -0.00005725, -0.00005868, -0.00001420, -0.00003022, -0.00008368, 0.00007151, -0.00007115, 0.00003071, -0.00004601, -0.00003123, 0.00002333, -0.00003292, -0.00003371, 0.00004062, -0.00002567, -0.00001928, 0.00002908, 0.00001223, -0.00000849, -0.00000199, -0.00000215, 0.00000365, -0.00000933, -0.00000268, },
		{ -0.00000503, -0.00000475, -0.00000682, 0.00000248, 0.00001251, -0.00001053, 0.00000244, -0.00000110, -0.00000922, 0.00000664, 0.00000244, -0.00000045, 0.00003733, -0.00004150, 0.00000410, 0.00003057, -0.00005924, -0.00003013, 0.00008349, -0.00012061, 0.00002655, -0.00002982, -0.00000070, 0.00001175, -0.00007198, -0.00010187, 0.00025825, -0.00054640, 0.00056805, -0.00094921, 0.00063023, -0.00081429, -0.00321943, 0.00764956, -0.00088016, -0.00227231, 0.00096854, -0.00086100, 0.00024281, -0.00019647, -0.00001050, 0.00001351, -0.00007821, -0.00004529, 0.00005954, -0.00007623, 0.00004997, -0.00008473, 0.00000149, 0.00002580, -0.00005771, -0.00001239, 0.00003980, -0.00004178, 0.00000101, 0.00002607, 0.00000360, -0.00000303, -0.00000267, -0.00000330, 0.00000414, -0.00001110, 0.00000393, 0.00001541, },
		{ -0.00000416, -0.00000695, 0.00000922, 0.00000921, -0.00001151, 0.00000386, -0.00000297, -0.00000554, 0.00000290, -0.00000007, 0.00001510, 0.00002190, -0.00004788, 0.00002586, 0.00001264, -0.00006875, 0.00000614, 0.00004462, -0.00011376, 0.00004963, -0.00006016, 0.00002984, -0.00000558, -0.00009881, -0.00001512, 0.00010427, -0.00037808, 0.00045010, -0.00099073, 0.00091584, -0.00161686, -0.00217692, 0.00792335, -0.00217692, -0.00161686, 0.00091584, -0.00099073, 0.00045010, -0.00037808, 0.00010427, -0.00001512, -0.00009881, -0.00000558, 0.00002984, -0.00006016, 0.00004963, -0.00011376, 0.00004462, 0.00000614, -0.00006875, 0.00001264, 0.00002586, -0.00004788, 0.00002190, 0.00001510, -0.00000007, 0.00000290, -0.00000554, -0.00000297, 0.00000386, -0.00001151, 0.00000921, 0.00000922, -0.00000695, },
		{ -0.00000503, 0.00001541, 0.00000393, -0.00001110, 0.00000414, -0.00000330, -0.00000267, -0.00000303, 0.00000360, 0.00002607, 0.00000101, -0.00004178, 0.00003980, -0.00001239, -0.00005771, 0.00002580, 0.00000149, -0.00008473, 0.00004997, -0.00007623, 0.00005954, -0.00004529, -0.00007821, 0.00001351, -0.00001050, -0.00019647, 0.00024281, -0.00086100, 0.00096854, -0.00227231, -0.00088016, 0.00764956, -0.00321943, -0.00081429, 0.00063023, -0.00094921, 0.00056805, -0.00054640, 0.00025825, -0.00010187, -0.00007198, 0.00001175, -0.00000070, -0.00002982, 0.00002655, -0.00012061, 0.00008349, -0.00003013, -0.00005924, 0.00003057, 0.00000410, -0.00004150, 0.00003733, -0.00000045, 0.00000244, 0.00000664, -0.00000922, -0.00000110, 0.00000244, -0.00001053, 0.00001251, 0.00000248, -0.00000682, -0.00000475, },
		{ 0.00001977, -0.00000268, -0.00000933, 0.00000365, -0.00000215, -0.00000199, -0.00000849, 0.00001223, 0.00002908, -0.00001928, -0.00002567, 0.00004062, -0.00003371, -0.00003292, 0.00002333, -0.00003123, -0.00004601, 0.00003071, -0.00007115, 0.00007151, -0.00008368, -0.00003022, -0.00001420, -0.00005868, -0.00005725, 0.00001438, -0.00061024, 0.00076808, -0.00263601, 0.00047218, 0.00688075, -0.00386749, -0.00003197, 0.00019204, -0.00073709, 0.00055285, -0.00064396, 0.00040509, -0.00022514, 0.00000523, -0.00001102, -0.00001149, -0.00000127, -0.00001182, -0.00010018, 0.00010314, -0.00007008, -0.00003021, 0.00003269, -0.00001669, -0.00002478, 0.00004307, -0.00001571, 0.00001005, 0.00000641, -0.00001179, 0.00000155, -0.00000016, -0.00000817, 0.00001348, -0.00000362, -0.00000551, -0.00000621, -0.00000059, },
		{ -0.00000963, -0.00000624, 0.00000272, -0.00000019, -0.00000378, -0.00001100, 0.00002270, 0.00002350, -0.00003336, -0.00000551, 0.00002843, -0.00004277, -0.00000577, 0.00000194, -0.00004465, -0.00001144, 0.00000243, -0.00004864, 0.00005905, -0.00010115, 0.00001764, -0.00007618, -0.00003973, 0.00000450, -0.00016611, -0.00032395, 0.00036294, -0.00262532, 0.00166969, 0.00576068, -0.00406926, 0.00058375, -0.00028454, -0.00040448, 0.00040017, -0.00063164, 0.00049510, -0.00034595, 0.00011537, -0.00007703, 0.00001281, 0.00000618, -0.00004868, -0.00005844, 0.00009469, -0.00009805, 0.00000929, 0.00001616, -0.00002768, -0.00000387, 0.00003840, -0.00002611, 0.00001978, 0.00000205, -0.00001157, 0.00000357, -0.00000347, -0.00000467, 0.00001205, -0.00000821, -0.00000404, -0.00000769, 0.00000625, 0.00002105, },
		{ -0.00000204, 0.00000144, 0.00000170, -0.00000724, -0.00000921, 0.00003119, 0.00001158, -0.00003818, 0.00001179, 0.00000833, -0.00003646, 0.00001253, -0.00002866, -0.00003815, 0.00000939, -0.00002259, -0.00001978, 0.00002714, -0.00008969, 0.00004187, -0.00014052, 0.00002250, -0.00001506, -0.00025213, -0.00009532, -0.00014224, -0.00224203, 0.00254005, 0.00448720, -0.00386635, 0.00094250, -0.00068624, -0.00003357, 0.00014662, -0.00050266, 0.00049283, -0.00042111, 0.00022528, -0.00017152, 0.00007448, -0.00002156, -0.00006372, -0.00001182, 0.00006008, -0.00010196, 0.00004536, -0.00001451, -0.00002359, 0.00001352, 0.00002631, -0.00002914, 0.00002802, -0.00000493, -0.00000787, 0.00000345, -0.00000648, -0.00000059, 0.00000852, -0.00001077, -0.00000332, -0.00000796, 0.00001458, 0.00001843, -0.00001566, },
		{ -0.00000029, 0.00000292, -0.00001085, -0.00000342, 0.00003458, -0.00000250, -0.00003410, 0.00002114, -0.00001194, -0.00001787, 0.00001517, -0.00005662, -0.00001827, 0.00001409, -0.00003583, 0.00000287, -0.00001114, -0.00005477, 0.00003193, -0.00017881, 0.00009114, -0.00008994, -0.00023455, 0.00000505, -0.00061646, -0.00157100, 0.00298761, 0.00326312, -0.00337387, 0.00102622, -0.00093239, 0.00028820, -0.00014276, -0.00028529, 0.00038995, -0.00041862, 0.00029724, -0.00026567, 0.00015968, -0.00008598, -0.00004146, 0.00001812, 0.00001271, -0.00007884, 0.00006477, -0.00004902, -0.00000496, 0.00002110, 0.00001225, -0.00002513, 0.00003189, -0.00001191, -0.00000137, 0.00000020, -0.00000794, 0.00000309, 0.00000361, -0.00001135, -0.00000384, -0.00000572, 0.00002275, 0.00001180, -0.00001938, 0.00000272, },
		{ 0.00000340, -0.00001307, 0.00000462, 0.00003153, -0.00001428, -0.00002449, 0.00002111, -0.00002529, 0.00000536, 0.00000182, -0.00007269, 0.00000496, 0.00000700, -0.00003561, 0.00001113, -0.00004120, -0.00001153, -0.00000700, -0.00017688, 0.00013172, -0.00017654, -0.00014269, -0.00004852, -0.00094014, -0.00075569, 0.00300979, 0.00224978, -0.00274329, 0.00088336, -0.00099364, 0.00049470, -0.00039599, -0.00003335, 0.00020865, -0.00032954, 0.00030268, -0.00032645, 0.00024191, -0.00017302, 0.00002117, 0.00001298, -0.00002668, -0.00003699, 0.00006061, -0.00007538, 0.00002201, 0.00001648, 0.00000194, -0.00001697, 0.00003035, -0.00001626, 0.00000604, -0.00000612, -0.00000683, 0.00000525, -0.00000161, -0.00001053, -0.00000538, -0.00000012, 0.00002853, 0.00000205, -0.00001965, 0.00000716, -0.00000262, },
		{ -0.00001301, 0.00001242, 0.00002279, -0.00002067, -0.00001404, 0.00001410, -0.00002803, 0.00002445, -0.00002198, -0.00007312, 0.00002246, -0.00000396, -0.00002657, 0.00000425, -0.00005349, 0.00002268, -0.00005786, -0.00013895, 0.00012543, -0.00023187, -0.00003159, -0.00022975, -0.00103839, 0.00004125, 0.00268802, 0.00153622, -0.00212019, 0.00060519, -0.00089308, 0.00055860, -0.00055671, 0.00019166, -0.00000575, -0.00017164, 0.00023290, -0.00032873, 0.00029100, -0.00025629, 0.00011130, -0.00003424, -0.00003778, 0.00000634, 0.00003560, -0.00008490, 0.00004808, 0.00000198, -0.00000087, -0.00000877, 0.00002449, -0.00001630, 0.00001192, -0.00001430, -0.00000287, 0.00000493, -0.00000595, -0.00000931, -0.00000698, 0.00000882, 0.00002974, -0.00000882, -0.00001605, 0.00001014, -0.00000542, 0.00000355, },
		{ 0.00001772, 0.00001074, -0.00002086, -0.00000677, 0.00000500, -0.00002104, 0.00003290, -0.00004762, -0.00006013, 0.00002935, -0.00001139, -0.00001655, -0.00001164, -0.00004713, 0.00003598, -0.00010016, -0.00008389, 0.00007506, -0.00023027, 0.00003955, -0.00047069, -0.00089966, 0.00068005, 0.00215757, 0.00113111, -0.00161013, 0.00029343, -0.00069119, 0.00049471, -0.00060000, 0.00034020, -0.00020000, 0.00001642, 0.00010273, -0.00026514, 0.00028449, -0.00030575, 0.00020282, -0.00011462, -0.00000915, 0.00003132, 0.00000161, -0.00007601, 0.00006483, -0.00001661, 0.00000430, -0.00000417, 0.00001696, -0.00001200, 0.00001416, -0.00002227, 0.00000321, 0.00000178, -0.00000847, -0.00000875, -0.00000702, 0.00001980, 0.00002491, -0.00001811, -0.00000925, 0.00001066, -0.00000819, 0.00000399, -0.00001066, },
		{ -0.00000170, -0.00001619, -0.00000446, -0.00000116, -0.00000897, 0.00002869, -0.00006697, -0.00003996, 0.00002581, -0.00001151, -0.00001238, -0.00002718, -0.00002936, 0.00002650, -0.00011897, -0.00003586, 0.00000184, -0.00017242, 0.00002952, -0.00068543, -0.00057444, 0.00107575, 0.00156717, 0.00097708, -0.00126121, 0.00003038, -0.00046161, 0.00034950, -0.00053651, 0.00038910, -0.00033020, 0.00018823, -0.00005432, -0.00014940, 0.00021666, -0.00029879, 0.00026462, -0.00020468, 0.00005599, 0.00002351, -0.00002478, -0.00005517, 0.00006831, -0.00003266, 0.00001485, -0.00000501, 0.00001088, -0.00000495, 0.00001181, -0.00002785, 0.00000985, -0.00000372, -0.00000883, -0.00000948, -0.00000380, 0.00003027, 0.00001404, -0.00002309, -0.00000102, 0.00000837, -0.00001012, 0.00000514, -0.00000689, 0.00001922, },
		{ -0.00000923, -0.00000631, -0.00000129, 0.00000194, 0.00001449, -0.00007523, -0.00001976, 0.00001577, -0.00000497, -0.00001683, -0.00003443, -0.00001139, 0.00000212, -0.00011076, -0.00001361, -0.00006506, -0.00008355, -0.00007024, -0.00079837, -0.00015581, 0.00121153, 0.00104103, 0.00097993, -0.00106617, -0.00013944, -0.00026680, 0.00018179, -0.00040441, 0.00034591, -0.00037485, 0.00030495, -0.00019758, -0.00001196, 0.00010178, -0.00022934, 0.00027169, -0.00027354, 0.00013930, -0.00001962, -0.00002895, -0.00003430, 0.00006068, -0.00004167, 0.00002659, -0.00001079, 0.00000862, 0.00000220, 0.00000542, -0.00002947, 0.00001510, -0.00001031, -0.00000750, -0.00001126, 0.00000373, 0.00003699, -0.00000105, -0.00002194, 0.00000615, 0.00000382, -0.00001045, 0.00000699, -0.00000290, 0.00001690, -0.00001219, },
		{ -0.00000964, 0.00000456, 0.00000648, -0.00000399, -0.00007193, -0.00000485, 0.00000428, 0.00000449, -0.00002794, -0.00003037, -0.00000300, -0.00002406, -0.00008368, -0.00002293, -0.00010148, -0.00000194, -0.00023352, -0.00076754, 0.00025072, 0.00113158, 0.00065405, 0.00104204, -0.00097977, -0.00021243, -0.00014172, 0.00004252, -0.00025336, 0.00024262, -0.00033940, 0.00034713, -0.00029302, 0.00011062, -0.00003018, -0.00011139, 0.00021485, -0.00029379, 0.00021271, -0.00008646, -0.00000468, -0.00002581, 0.00004922, -0.00004316, 0.00003576, -0.00001929, 0.00001103, 0.00000673, -0.00000308, -0.00002680, 0.00001731, -0.00001640, -0.00000555, -0.00001283, 0.00001525, 0.00003700, -0.00001690, -0.00001452, 0.00000986, -0.00000149, -0.00000885, 0.00000907, 0.00000023, 0.00001172, -0.00001963, -0.00000237, },
		{ 0.00001359, 0.00000239, -0.00002047, -0.00006022, 0.00000273, -0.00000493, 0.00001258, -0.00004071, -0.00001763, -0.00000828, -0.00004002, -0.00005287, -0.00005540, -0.00009829, 0.00003707, -0.00040973, -0.00059529, 0.00055757, 0.00091824, 0.00042534, 0.00108905, -0.00094332, -0.00021548, -0.00008969, -0.00003954, -0.00012705, 0.00012199, -0.00025104, 0.00031906, -0.00032385, 0.00018828, -0.00014373, 0.00002494, 0.00010521, -0.00025174, 0.00024808, -0.00015451, 0.00004257, -0.00003706, 0.00004300, -0.00004069, 0.00004086, -0.00002782, 0.00001733, 0.00000671, -0.00001103, -0.00002089, 0.00001575, -0.00002062, -0.00000419, -0.00001223, 0.00002869, 0.00002874, -0.00002930, -0.00000270, 0.00000867, -0.00000556, -0.00000571, 0.00001076, 0.00000194, 0.00000500, -0.00002418, 0.00000319, -0.00001150, },
		{ -0.00000920, -0.00003042, -0.00004489, 0.00000386, -0.00001062, 0.00001681, -0.00005001, -0.00000252, -0.00002451, -0.00004008, -0.00003300, -0.00009328, -0.00006454, 0.00001528, -0.00054256, -0.00032403, 0.00071634, 0.00066251, 0.00032860, 0.00108321, -0.00090698, -0.00018818, -0.00008962, -0.00006194, -0.00005045, 0.00002211, -0.00014692, 0.00024439, -0.00029451, 0.00020739, -0.00021015, 0.00014330, -0.00002869, -0.00015333, 0.00022757, -0.00019805, 0.00009684, -0.00006665, 0.00004856, -0.00003990, 0.00004307, -0.00003454, 0.00002571, 0.00000153, -0.00001583, -0.00001384, 0.00001085, -0.00002235, -0.00000407, -0.00000751, 0.00004049, 0.00001295, -0.00003457, 0.00001010, 0.00000275, -0.00000661, -0.00000213, 0.00001167, 0.00000223, -0.00000217, -0.00002669, 0.00000782, -0.00001016, 0.00002167, },
		{ -0.00003222, -0.00003031, 0.00000096, -0.00001365, 0.00001736, -0.00005323, 0.00000847, -0.00004432, -0.00002663, -0.00003179, -0.00011824, -0.00002268, -0.00006149, -0.00058938, -0.00001964, 0.00072431, 0.00043800, 0.00031308, 0.00102212, -0.00084315, -0.00016539, -0.00011001, -0.00004281, -0.00002581, -0.00003502, -0.00006036, 0.00015562, -0.00022659, 0.00017412, -0.00021720, 0.00021361, -0.00015010, -0.00002310, 0.00015067, -0.00019762, 0.00013723, -0.00010433, 0.00006655, -0.00004556, 0.00004543, -0.00003918, 0.00003418, -0.00000805, -0.00001576, -0.00000818, 0.00000407, -0.00002194, -0.00000484, 0.00000222, 0.00004653, -0.00000709, -0.00003094, 0.00001988, -0.00000595, -0.00000383, 0.00000045, 0.00001186, 0.00000136, -0.00000931, -0.00002783, 0.00001285, -0.00000594, 0.00002521, -0.00002465, },
		{ -0.00001916, -0.00000363, -0.00001563, 0.00001637, -0.00005116, 0.00001187, -0.00005971, -0.00000807, -0.00004731, -0.00011959, 0.00000159, -0.00016607, -0.00053469, 0.00025055, 0.00061646, 0.00028568, 0.00032611, 0.00092672, -0.00074841, -0.00016604, -0.00012324, -0.00000918, -0.00003735, -0.00004729, -0.00001064, 0.00008180, -0.00014912, 0.00011043, -0.00017244, 0.00022214, -0.00022659, 0.00010368, 0.00003527, -0.00014748, 0.00014566, -0.00013446, 0.00009075, -0.00005873, 0.00005092, -0.00004286, 0.00004132, -0.00002034, -0.00001033, -0.00000605, -0.00000264, -0.00002043, -0.00000500, 0.00001602, 0.00004354, -0.00002632, -0.00001932, 0.00002354, -0.00001434, 0.00000219, 0.00000090, 0.00001192, -0.00000056, -0.00001636, -0.00002751, 0.00001956, -0.00000097, 0.00002267, -0.00003958, -0.00002696, },
		{ -0.00000852, -0.00001750, 0.00001620, -0.00004706, 0.00000849, -0.00006607, 0.00000594, -0.00006983, -0.00009874, -0.00000963, -0.00026060, -0.00039320, 0.00043731, 0.00044774, 0.00021184, 0.00032987, 0.00082525, -0.00063636, -0.00019087, -0.00011492, 0.00001543, -0.00006149, -0.00002853, 0.00000057, 0.00003933, -0.00008727, 0.00004451, -0.00009925, 0.00017559, -0.00024098, 0.00019279, -0.00008799, -0.00005835, 0.00011397, -0.00014204, 0.00011017, -0.00007567, 0.00006047, -0.00004713, 0.00004653, -0.00003328, -0.00000040, -0.00000867, -0.00000748, -0.00001905, -0.00000244, 0.00003092, 0.00003050, -0.00003934, -0.00000327, 0.00001998, -0.00001928, 0.00000953, -0.00000108, 0.00001256, -0.00000383, -0.00002323, -0.00002486, 0.00002815, 0.00000190, 0.00001487, -0.00005057, -0.00001724, -0.00001230, },
		{ -0.00001896, 0.00001779, -0.00004439, 0.00000232, -0.00006392, 0.00000932, -0.00008728, -0.00006783, -0.00005909, -0.00031049, -0.00020258, 0.00052055, 0.00027317, 0.00019721, 0.00030833, 0.00073977, -0.00052627, -0.00022789, -0.00008582, 0.00001906, -0.00007769, 0.00000056, -0.00001438, 0.00002891, -0.00005393, -0.00000022, -0.00002716, 0.00009760, -0.00019674, 0.00022256, -0.00018595, 0.00004592, 0.00004750, -0.00011879, 0.00011332, -0.00008883, 0.00007184, -0.00005265, 0.00004981, -0.00004494, 0.00001221, -0.00001605, -0.00000934, -0.00001848, 0.00000465, 0.00004249, 0.00000942, -0.00004222, 0.00001209, 0.00001071, -0.00001875, 0.00001561, -0.00000470, 0.00001418, -0.00000910, -0.00002924, -0.00001877, 0.00003726, 0.00000052, 0.00000441, -0.00005604, -0.00000584, -0.00000948, -0.00001322, },
		{ 0.00002015, -0.00004472, -0.00000210, -0.00005794, 0.00000220, -0.00009135, -0.00004376, -0.00013349, -0.00029627, -0.00000939, 0.00050953, 0.00013235, 0.00021141, 0.00026447, 0.00067936, -0.00043285, -0.00026136, -0.00004719, 0.00000303, -0.00007574, 0.00002208, -0.00003723, 0.00003897, -0.00004735, -0.00001309, 0.00001939, 0.00001926, -0.00011583, 0.00019080, -0.00023410, 0.00013641, -0.00003634, -0.00006682, 0.00009302, -0.00008973, 0.00008002, -0.00005832, 0.00005116, -0.00005376, 0.00002517, -0.00002710, -0.00000797, -0.00001845, 0.00001674, 0.00004619, -0.00001482, -0.00003398, 0.00002177, -0.00000067, -0.00001277, 0.00001822, -0.00000829, 0.00001636, -0.00001695, -0.00003290, -0.00000882, 0.00004441, -0.00000556, -0.00000572, -0.00005624, 0.00000518, -0.00001021, -0.00001737, -0.00001903, },
		{ -0.00004692, -0.00000243, -0.00005388, -0.00000984, -0.00008141, -0.00004023, -0.00020914, -0.00021909, 0.00014587, 0.00043392, 0.00004265, 0.00022670, 0.00021182, 0.00064077, -0.00036113, -0.00027995, -0.00001297, -0.00002225, -0.00005713, 0.00002743, -0.00005255, 0.00005334, -0.00005475, 0.00000160, 0.00002945, -0.00003264, -0.00003009, 0.00011468, -0.00022445, 0.00018903, -0.00011400, 0.00000149, 0.00004989, -0.00007270, 0.00007906, -0.00006134, 0.00005003, -0.00005855, 0.00003607, -0.00003998, -0.00000387, -0.00001778, 0.00003228, 0.00003901, -0.00003579, -0.00001716, 0.00002272, -0.00000970, -0.00000344, 0.00001644, -0.00001000, 0.00001777, -0.00002725, -0.00003219, 0.00000398, 0.00004702, -0.00001494, -0.00001332, -0.00005237, 0.00001456, -0.00001416, -0.00002005, -0.00001719, 0.00002113, },
		{ 0.00000024, -0.00005512, -0.00001923, -0.00006446, -0.00006163, -0.00026131, -0.00009843, 0.00023955, 0.00033058, 0.00000118, 0.00022639, 0.00016475, 0.00061390, -0.00030697, -0.00028087, 0.00000726, -0.00004426, -0.00003134, 0.00001846, -0.00005351, 0.00005930, -0.00006002, 0.00002868, 0.00000799, -0.00004492, 0.00003064, 0.00002377, -0.00016745, 0.00019253, -0.00016395, 0.00006762, -0.00000727, -0.00003762, 0.00006465, -0.00005816, 0.00004525, -0.00005840, 0.00004278, -0.00005257, 0.00000219, -0.00001489, 0.00004777, 0.00002088, -0.00004751, 0.00000287, 0.00001502, -0.00001275, 0.00000574, 0.00001110, -0.00000866, 0.00001652, -0.00003869, -0.00002536, 0.00001706, 0.00004365, -0.00002503, -0.00001763, -0.00004569, 0.00002140, -0.00002092, -0.00001974, -0.00001438, 0.00001915, -0.00004809, },
		{ -0.00006076, -0.00002086, -0.00005138, -0.00010125, -0.00027339, 0.00003620, 0.00026827, 0.00023117, -0.00000725, 0.00020654, 0.00013188, 0.00058861, -0.00026157, -0.00026896, 0.00001166, -0.00005460, -0.00000945, 0.00000398, -0.00004261, 0.00005261, -0.00005155, 0.00005007, -0.00002813, -0.00002149, 0.00004884, -0.00005084, -0.00008727, 0.00015168, -0.00017405, 0.00011335, -0.00006407, 0.00000922, 0.00003633, -0.00004614, 0.00003548, -0.00005269, 0.00004364, -0.00006272, 0.00000951, -0.00000870, 0.00005861, -0.00000486, -0.00004646, 0.00001966, 0.00000187, -0.00000847, 0.00001127, 0.00000459, -0.00000446, 0.00001094, -0.00004872, -0.00001191, 0.00002692, 0.00003478, -0.00003334, -0.00001907, -0.00003707, 0.00002463, -0.00002931, -0.00001508, -0.00001299, 0.00001463, -0.00004566, 0.00000226, },
		{ -0.00001495, -0.00005130, -0.00014433, -0.00024209, 0.00015577, 0.00024523, 0.00015480, -0.00000075, 0.00017235, 0.00011424, 0.00055938, -0.00021698, -0.00025262, 0.00000519, -0.00005199, 0.00000148, -0.00000614, -0.00002825, 0.00003786, -0.00002702, 0.00005307, -0.00005858, 0.00001996, 0.00002441, -0.00008740, -0.00001259, 0.00008468, -0.00014537, 0.00012711, -0.00010550, 0.00005679, -0.00000165, -0.00002501, 0.00001999, -0.00004132, 0.00003766, -0.00006850, 0.00001768, 0.00000063, 0.00006062, -0.00003213, -0.00003299, 0.00002793, -0.00001153, 0.00000166, 0.00001100, -0.00000011, 0.00000083, 0.00000041, -0.00005409, 0.00000675, 0.00003029, 0.00002280, -0.00003851, -0.00001843, -0.00002711, 0.00002308, -0.00003702, -0.00000609, -0.00001572, 0.00001026, -0.00003944, -0.00000009, -0.00006625, },
		{ -0.00006703, -0.00017439, -0.00017725, 0.00024036, 0.00019214, 0.00010680, 0.00000681, 0.00013230, 0.00010736, 0.00052641, -0.00016965, -0.00023899, -0.00000426, -0.00004118, 0.00000094, -0.00000606, -0.00001905, 0.00002459, 0.00000673, 0.00003528, -0.00006861, 0.00005637, -0.00002699, -0.00008002, 0.00003357, 0.00001598, -0.00009112, 0.00010784, -0.00012117, 0.00009300, -0.00004132, 0.00000240, -0.00000045, -0.00002503, 0.00002470, -0.00006828, 0.00002640, 0.00001137, 0.00005164, -0.00005386, -0.00001135, 0.00002546, -0.00001997, 0.00001386, 0.00000507, -0.00000062, 0.00000438, -0.00001390, -0.00005184, 0.00002719, 0.00002550, 0.00001109, -0.00004077, -0.00001605, -0.00001680, 0.00001606, -0.00004089, 0.00000501, -0.00002385, 0.00000975, -0.00003228, -0.00000792, -0.00006624, -0.00000711, },
		{ -0.00017991, -0.00009682, 0.00028298, 0.00013044, 0.00008243, 0.00000875, 0.00009328, 0.00010522, 0.00049357, -0.00012098, -0.00023080, -0.00001027, -0.00002935, -0.00000611, 0.00000391, -0.00001923, 0.00002169, 0.00003800, 0.00000450, -0.00005486, 0.00006919, -0.00008144, -0.00003902, 0.00004116, -0.00003267, -0.00003128, 0.00006493, -0.00010886, 0.00010877, -0.00007345, 0.00003072, -0.00002334, -0.00000573, 0.00000577, -0.00006093, 0.00003515, 0.00002058, 0.00003254, -0.00006420, 0.00001164, 0.00001391, -0.00002020, 0.00002352, -0.00000408, 0.00000373, 0.00000326, -0.00002905, -0.00004058, 0.00004474, 0.00001325, 0.00000262, -0.00004141, -0.00001155, -0.00000799, 0.00000434, -0.00003815, 0.00001418, -0.00003586, 0.00001559, -0.00002886, -0.00001841, -0.00005804, -0.00000535, -0.00009353, },
		{ -0.00001963, 0.00028828, 0.00007537, 0.00007268, 0.00000455, 0.00005836, 0.00010350, 0.00046519, -0.00007520, -0.00022625, -0.00001064, -0.00002196, -0.00001277, 0.00001859, -0.00002688, 0.00003276, 0.00005630, -0.00002604, -0.00002582, 0.00005150, -0.00011668, 0.00001472, 0.00001506, -0.00004939, 0.00001476, 0.00001442, -0.00007509, 0.00010119, -0.00009068, 0.00005363, -0.00004480, 0.00001351, -0.00001685, -0.00004600, 0.00004280, 0.00002508, 0.00000726, -0.00006052, 0.00002888, -0.00000177, -0.00001224, 0.00002704, -0.00001270, 0.00001150, -0.00000431, -0.00004087, -0.00002124, 0.00005491, -0.00000335, -0.00000120, -0.00004184, -0.00000424, -0.00000326, -0.00000928, -0.00002800, 0.00001703, -0.00004740, 0.00002705, -0.00003306, -0.00002585, -0.00004387, -0.00001564, -0.00012016, -0.00015835, },
		{ 0.00026761, 0.00003400, 0.00006917, -0.00000318, 0.00002735, 0.00010083, 0.00044345, -0.00003636, -0.00022109, -0.00000739, -0.00002036, -0.00001409, 0.00003143, -0.00003552, 0.00005471, 0.00005717, -0.00004444, 0.00000293, 0.00000998, -0.00012042, 0.00005871, -0.00002855, -0.00003519, 0.00003485, -0.00002694, -0.00003253, 0.00007438, -0.00008996, 0.00006601, -0.00006070, 0.00002902, -0.00003962, -0.00002413, 0.00004756, 0.00002268, -0.00001820, -0.00004431, 0.00003538, -0.00001525, 0.00000059, 0.00002334, -0.00001726, 0.00001951, -0.00001809, -0.00004518, 0.00000267, 0.00005488, -0.00001990, -0.00000106, -0.00004250, 0.00000589, -0.00000498, -0.00002034, -0.00001272, 0.00001106, -0.00005292, 0.00003971, -0.00004529, -0.00002453, -0.00003044, -0.00003834, -0.00013561, -0.00011619, 0.00004075, },
		{ 0.00000693, 0.00006667, -0.00001168, -0.00000123, 0.00009793, 0.00042780, -0.00000607, -0.00021153, -0.00000462, -0.00002190, -0.00000920, 0.00003821, -0.00003787, 0.00007972, 0.00004365, -0.00004573, 0.00001624, -0.00003896, -0.00009402, 0.00007746, -0.00006959, -0.00000190, 0.00002751, -0.00004756, 0.00000464, 0.00003788, -0.00007355, 0.00006570, -0.00006809, 0.00003743, -0.00005827, 0.00000268, 0.00004723, 0.00001323, -0.00003769, -0.00002074, 0.00003001, -0.00002118, 0.00001301, 0.00001435, -0.00001596, 0.00002403, -0.00003556, -0.00003916, 0.00002579, 0.00004465, -0.00003219, 0.00000077, -0.00004242, 0.00001740, -0.00001396, -0.00002414, 0.00000250, -0.00000253, -0.00004837, 0.00004690, -0.00006127, -0.00001226, -0.00002594, -0.00006727, -0.00013295, -0.00006515, 0.00007896, 0.00023330, },
		{ 0.00006296, -0.00001948, -0.00002828, 0.00009578, 0.00041603, 0.00001685, -0.00019626, -0.00000574, -0.00002209, -0.00000103, 0.00003894, -0.00002948, 0.00009931, 0.00002390, -0.00003380, 0.00000616, -0.00007665, -0.00005049, 0.00006740, -0.00009212, 0.00003353, 0.00000111, -0.00004439, 0.00002565, 0.00000319, -0.00004804, 0.00005420, -0.00006632, 0.00003688, -0.00006865, 0.00003101, 0.00004001, -0.00000107, -0.00004686, 0.00000313, 0.00001576, -0.00001726, 0.00001990, 0.00000419, -0.00000960, 0.00002225, -0.00005252, -0.00002244, 0.00004234, 0.00002722, -0.00003782, 0.00000166, -0.00003969, 0.00002719, -0.00002829, -0.00001787, 0.00001125, -0.00001852, -0.00003377, 0.00004291, -0.00007328, 0.00000753, -0.00003574, -0.00009215, -0.00011258, -0.00001693, 0.00009685, 0.00019466, -0.00000847, },
		{ -0.00002617, -0.00005360, 0.00009421, 0.00040616, 0.00003509, -0.00017666, -0.00001158, -0.00001731, 0.00000586, 0.00003746, -0.00001052, 0.00010816, 0.00000655, -0.00001898, -0.00002471, -0.00008995, -0.00000806, 0.00003666, -0.00008972, 0.00005608, -0.00003045, -0.00002308, 0.00002640, -0.00002017, -0.00002194, 0.00003606, -0.00005738, 0.00002778, -0.00006784, 0.00005643, 0.00002528, -0.00001620, -0.00004459, 0.00002072, -0.00000143, -0.00000507, 0.00001836, -0.00000245, -0.00000135, 0.00001348, -0.00006427, 0.00000255, 0.00004777, 0.00000775, -0.00003704, 0.00000010, -0.00003269, 0.00003142, -0.00004331, -0.00000236, 0.00000863, -0.00002934, -0.00001414, 0.00002621, -0.00007354, 0.00002633, -0.00005897, -0.00010321, -0.00008159, 0.00002122, 0.00010005, 0.00015678, -0.00001516, 0.00005734, },
		{ -0.00007652, 0.00009177, 0.00039767, 0.00005103, -0.00015550, -0.00002024, -0.00000658, 0.00000793, 0.00003905, 0.00001490, 0.00010590, -0.00000380, -0.00001261, -0.00006490, -0.00007615, 0.00001756, 0.00000037, -0.00006666, 0.00005803, -0.00005361, 0.00000521, 0.00001014, -0.00002755, -0.00000274, 0.00001712, -0.00004529, 0.00001311, -0.00005518, 0.00007432, 0.00000427, -0.00002758, -0.00003325, 0.00002812, -0.00001501, 0.00001051, 0.00000899, -0.00000238, 0.00000445, -0.00000039, -0.00006714, 0.00003099, 0.00004021, -0.00000801, -0.00003262, -0.00000338, -0.00002139, 0.00002722, -0.00005284, 0.00001739, -0.00000612, -0.00002823, 0.00000191, 0.00000127, -0.00005833, 0.00003395, -0.00008786, -0.00009585, -0.00005017, 0.00004813, 0.00009405, 0.00012174, -0.00001526, 0.00004934, -0.00003142, },
		{ 0.00008662, 0.00039156, 0.00006549, -0.00013505, -0.00002847, 0.00000831, 0.00000440, 0.00004787, 0.00004125, 0.00009592, -0.00000856, -0.00002146, -0.00009932, -0.00004313, 0.00001906, -0.00002628, -0.00003470, 0.00004089, -0.00006018, 0.00002854, -0.00001462, -0.00002011, 0.00000536, 0.00000248, -0.00003460, -0.00000231, -0.00003278, 0.00008101, -0.00001987, -0.00003165, -0.00001779, 0.00002526, -0.00002021, 0.00002329, -0.00000430, 0.00000464, 0.00000409, -0.00001547, -0.00005957, 0.00005694, 0.00002106, -0.00001560, -0.00002862, -0.00000629, -0.00000822, 0.00001439, -0.00005162, 0.00003362, -0.00002847, -0.00001272, 0.00000569, -0.00002243, -0.00003074, 0.00002337, -0.00011050, -0.00007310, -0.00002664, 0.00006710, 0.00008187, 0.00009072, -0.00001042, 0.00003832, -0.00003430, -0.00009693, },
		},
		{ { -0.00031327, -0.00003332, 0.00018164, 0.00005877, 0.00004930, 0.00005126, 0.00004373, 0.00008168, 0.00019248, -0.00000842, -0.00016781, -0.00004510, -0.00003255, -0.00001774, -0.00005655, -0.00001050, -0.00001422, -0.00001207, 0.00001372, -0.00000802, -0.00001377, -0.00001357, -0.00001191, 0.00000182, -0.00002266, 0.00000209, -0.00000404, 0.00000383, -0.00000145, -0.00000842, 0.00003682, -0.00001179, -0.00001114, -0.00001179, 0.00003682, -0.00000842, -0.00000145, 0.00000383, -0.00000404, 0.00000209, -0.00002266, 0.00000182, -0.00001191, -0.00001357, -0.00001377, -0.00000802, 0.00001372, -0.00001207, -0.00001422, -0.00001050, -0.00005655, -0.00001774, -0.00003255, -0.00004510, -0.00016781, -0.00000842, 0.00019248, 0.00008168, 0.00004373, 0.00005126, 0.00004930, 0.00005877, 0.00018164, -0.00003332, },
		{ -0.00004265, 0.00015855, 0.00005993, 0.00004477, 0.00004248, 0.00003187, 0.00006997, 0.00015251, 0.00007560, -0.00014924, -0.00008070, -0.00002404, -0.00002034, -0.00004840, -0.00002288, -0.00000703, -0.00002015, 0.00001529, -0.00000658, -0.00000985, -0.00001462, -0.00001487, 0.00000340, -0.00001708, -0.00000889, 0.00000595, -0.00000549, 0.00000838, -0.00001884, 0.00003331, 0.00000094, -0.00001318, -0.00001596, 0.00002685, 0.00001179, -0.00001567, 0.00001512, -0.00001284, 0.00000800, -0.00002176, -0.00000363, -0.00000871, -0.00001177, -0.00001625, -0.00001078, 0.00001147, -0.00000474, -0.00001912, -0.00000265, -0.00005678, -0.00002212, -0.00003433, -0.00003034, -0.00014150, -0.00011517, 0.00020847, 0.00011356, 0.00004754, 0.00006553, 0.00005428, 0.00005723, 0.00020401, -0.00002037, -0.00031540, },
		{ 0.00013588, 0.00006037, 0.00003978, 0.00003922, 0.00001456, 0.00006832, 0.00011121, 0.00012061, -0.00009363, -0.00011903, -0.00002199, -0.00002101, -0.00004092, -0.00003140, -0.00000474, -0.00002294, 0.00001152, -0.00000228, -0.00000906, -0.00001136, -0.00001967, 0.00000373, -0.00001025, -0.00001721, 0.00000912, -0.00000634, 0.00000882, -0.00001687, 0.00001806, 0.00001806, -0.00001663, -0.00001372, 0.00000921, 0.00003225, -0.00002418, 0.00001873, -0.00001171, 0.00000279, -0.00001180, -0.00001368, -0.00000388, -0.00001232, -0.00001389, -0.00001741, 0.00001166, -0.00000242, -0.00001713, -0.00000398, -0.00004557, -0.00003630, -0.00002506, -0.00003860, -0.00008499, -0.00021090, 0.00017476, 0.00017057, 0.00004513, 0.00008319, 0.00006094, 0.00005600, 0.00022464, -0.00000319, -0.00032138, -0.00004959, },
		{ 0.00006084, 0.00003297, 0.00004110, -0.00000511, 0.00006889, 0.00007997, 0.00012996, -0.00002443, -0.00013764, -0.00003750, -0.00001369, -0.00003996, -0.00003090, -0.00001088, -0.00001817, 0.00000128, 0.00000585, -0.00001253, -0.00000388, -0.00002558, 0.00000364, -0.00000530, -0.00001911, 0.00000366, 0.00000031, 0.00000269, -0.00000767, -0.00000069, 0.00003121, -0.00001576, -0.00001089, -0.00000684, 0.00004233, -0.00001864, 0.00000915, 0.00000191, -0.00001232, 0.00000321, -0.00002421, 0.00000029, -0.00001441, -0.00000809, -0.00002556, 0.00001266, -0.00000398, -0.00001004, -0.00001146, -0.00002818, -0.00005336, -0.00001254, -0.00005590, -0.00002822, -0.00025951, 0.00007840, 0.00024263, 0.00004709, 0.00009809, 0.00007232, 0.00005454, 0.00024396, 0.00001742, -0.00032997, -0.00005539, 0.00011396, },
		{ 0.00002340, 0.00004710, -0.00002452, 0.00006771, 0.00006027, 0.00011781, 0.00003425, -0.00012426, -0.00006949, -0.00000190, -0.00004349, -0.00002371, -0.00002188, -0.00000996, -0.00001090, 0.00001366, -0.00001632, 0.00000316, -0.00002849, 0.00000118, -0.00000167, -0.00001627, -0.00000578, 0.00000763, -0.00000259, 0.00000037, -0.00001363, 0.00003333, -0.00000672, -0.00001179, -0.00001475, 0.00003708, 0.00000121, -0.00001002, 0.00002125, -0.00002797, 0.00001360, -0.00002765, -0.00000114, -0.00001399, -0.00000400, -0.00002942, 0.00000965, -0.00000458, -0.00000413, -0.00001707, -0.00001488, -0.00006164, -0.00000960, -0.00006360, 0.00000152, -0.00024301, -0.00006648, 0.00030003, 0.00007200, 0.00010140, 0.00009282, 0.00005085, 0.00026419, 0.00003939, -0.00033942, -0.00006103, 0.00009248, 0.00006249, },
		{ 0.00005464, -0.00004075, 0.00006275, 0.00004961, 0.00009772, 0.00007005, -0.00008395, -0.00010341, 0.00000285, -0.00004376, -0.00001733, -0.00002978, -0.00000594, -0.00001783, 0.00001492, -0.00001474, 0.00000410, -0.00002429, -0.00000569, 0.00000252, -0.00001277, -0.00001265, 0.00000875, -0.00000134, 0.00000200, -0.00001642, 0.00002355, 0.00000863, -0.00001566, -0.00001445, 0.00002014, 0.00002655, -0.00002735, 0.00003387, -0.00003213, 0.00001011, -0.00001862, -0.00001106, -0.00000817, -0.00000580, -0.00002496, -0.00000011, -0.00000101, -0.00000422, -0.00001475, -0.00001282, -0.00005429, -0.00002269, -0.00005304, -0.00000361, -0.00017406, -0.00021732, 0.00030167, 0.00013673, 0.00008820, 0.00012427, 0.00004396, 0.00028726, 0.00006117, -0.00034882, -0.00006677, 0.00007095, 0.00006568, 0.00001178, },
		{ -0.00005073, 0.00005255, 0.00004588, 0.00007667, 0.00008468, -0.00003442, -0.00012039, -0.00000975, -0.00003511, -0.00001743, -0.00002884, -0.00001093, -0.00001557, 0.00000688, -0.00000573, -0.00000252, -0.00001335, -0.00001543, 0.00000699, -0.00001076, -0.00001378, 0.00000209, 0.00000581, -0.00000134, -0.00001190, 0.00000782, 0.00002326, -0.00001742, -0.00001119, 0.00000100, 0.00004468, -0.00003070, 0.00002998, -0.00001863, -0.00000860, 0.00000043, -0.00002591, 0.00000058, -0.00001211, -0.00001443, -0.00001314, 0.00000419, -0.00000895, -0.00000579, -0.00002002, -0.00003562, -0.00004511, -0.00003237, -0.00002793, -0.00009099, -0.00031989, 0.00021605, 0.00024060, 0.00006645, 0.00016085, 0.00003759, 0.00031173, 0.00008427, -0.00035970, -0.00007143, 0.00004868, 0.00006975, 0.00000052, 0.00005993, },
		{ 0.00003707, 0.00004745, 0.00005643, 0.00008701, 0.00000635, -0.00011011, -0.00003900, -0.00002091, -0.00002220, -0.00002053, -0.00002251, -0.00000727, -0.00000638, 0.00000624, -0.00001170, -0.00000184, -0.00002228, 0.00000828, -0.00000872, -0.00001121, -0.00000772, 0.00001254, -0.00000340, -0.00000703, -0.00000521, 0.00002950, -0.00001202, -0.00001060, -0.00001149, 0.00004681, -0.00001552, 0.00000988, 0.00000729, -0.00003303, 0.00001843, -0.00003592, 0.00000474, -0.00001624, -0.00000574, -0.00002093, 0.00000391, -0.00001179, 0.00000214, -0.00002677, -0.00001857, -0.00006100, -0.00002019, -0.00004426, -0.00003626, -0.00033614, 0.00004460, 0.00035358, 0.00006055, 0.00018808, 0.00004089, 0.00033187, 0.00011404, -0.00037634, -0.00007241, 0.00002471, 0.00007369, -0.00000787, 0.00006009, -0.00005300, },
		{ 0.00005122, 0.00003828, 0.00008392, 0.00003079, -0.00007797, -0.00007099, -0.00001316, -0.00002336, -0.00001250, -0.00003244, -0.00000138, -0.00001617, 0.00001269, -0.00001508, 0.00000196, -0.00001986, 0.00000267, -0.00000381, -0.00000945, -0.00001389, 0.00001198, 0.00000114, -0.00000699, -0.00001017, 0.00002442, 0.00000084, -0.00001401, -0.00001454, 0.00003335, 0.00001171, -0.00001510, 0.00003102, -0.00004747, 0.00002183, -0.00003127, -0.00000273, -0.00001156, -0.00000657, -0.00001618, -0.00000730, -0.00000722, 0.00000174, -0.00002371, -0.00001504, -0.00005719, -0.00003061, -0.00003405, -0.00003136, -0.00026703, -0.00016825, 0.00042002, 0.00010233, 0.00019045, 0.00006330, 0.00034123, 0.00015672, -0.00040325, -0.00006752, -0.00000151, 0.00007669, -0.00001237, 0.00005513, -0.00004951, 0.00001919, },
		{ 0.00002546, 0.00007614, 0.00004352, -0.00004109, -0.00008666, -0.00002393, -0.00001365, -0.00001171, -0.00003327, -0.00000506, -0.00001566, 0.00000766, -0.00000744, -0.00000599, -0.00000671, -0.00000934, 0.00000404, -0.00001078, -0.00001302, 0.00000241, 0.00001175, -0.00001098, -0.00000826, 0.00001187, 0.00001574, -0.00001746, -0.00001204, 0.00001295, 0.00003710, -0.00002921, 0.00003749, -0.00004017, 0.00000425, -0.00001032, -0.00002139, 0.00000186, -0.00001752, 0.00000036, -0.00002688, 0.00000294, -0.00000688, -0.00000991, -0.00002625, -0.00003406, -0.00006098, -0.00000113, -0.00006445, -0.00015490, -0.00034884, 0.00038232, 0.00020971, 0.00016573, 0.00010514, 0.00033953, 0.00021388, -0.00044066, -0.00005804, -0.00002843, 0.00007766, -0.00001352, 0.00004785, -0.00004514, 0.00000409, 0.00005235, },
		{ 0.00006133, 0.00005398, -0.00001623, -0.00007489, -0.00005462, 0.00000501, -0.00001756, -0.00002475, -0.00001809, -0.00000563, -0.00000702, 0.00000837, -0.00002143, 0.00001087, -0.00002121, 0.00001036, -0.00001289, -0.00000762, -0.00001110, 0.00002175, -0.00001304, -0.00000552, -0.00000029, 0.00002490, -0.00001528, -0.00001043, -0.00000370, 0.00004795, -0.00002233, 0.00002142, -0.00001212, -0.00002723, 0.00001664, -0.00004102, 0.00001528, -0.00003028, 0.00001868, -0.00004371, 0.00000907, -0.00001525, 0.00000558, -0.00004133, -0.00000639, -0.00009214, 0.00003078, -0.00009958, -0.00006114, -0.00042890, 0.00021512, 0.00036415, 0.00013655, 0.00015126, 0.00033728, 0.00027866, -0.00048169, -0.00005054, -0.00005137, 0.00007432, -0.00001209, 0.00004166, -0.00004462, -0.00000391, 0.00004701, 0.00002110, },
		{ 0.00006681, -0.00000855, -0.00004101, -0.00009175, 0.00001992, -0.00002107, -0.00001459, -0.00003239, 0.00000539, -0.00002209, 0.00002248, -0.00003367, 0.00002175, -0.00002367, 0.00000889, -0.00001102, -0.00000368, -0.00002031, 0.00002303, -0.00000723, -0.00000735, -0.00000593, 0.00002369, -0.00000518, -0.00001306, -0.00001075, 0.00004033, 0.00000328, -0.00000830, 0.00002188, -0.00005442, 0.00003207, -0.00004703, 0.00001667, -0.00003317, 0.00002579, -0.00004505, 0.00000059, -0.00001323, 0.00001088, -0.00004556, 0.00000701, -0.00010172, 0.00003575, -0.00010203, -0.00003129, -0.00038273, -0.00004961, 0.00050352, 0.00014717, 0.00017584, 0.00035197, 0.00033861, -0.00051482, -0.00005463, -0.00006417, 0.00006379, -0.00000810, 0.00003798, -0.00004917, -0.00000492, 0.00003585, 0.00002436, 0.00004000, },
		{ -0.00000945, -0.00000403, -0.00011379, 0.00001543, -0.00001181, -0.00001223, -0.00003801, 0.00000734, -0.00002724, 0.00002477, -0.00003294, 0.00001747, -0.00001145, -0.00000285, -0.00000294, -0.00000528, -0.00002008, 0.00001246, 0.00000674, -0.00001367, -0.00000489, 0.00001402, 0.00000921, -0.00001762, -0.00000996, 0.00002071, 0.00003426, -0.00003353, 0.00004164, -0.00005883, 0.00002249, -0.00003102, 0.00000033, -0.00002058, 0.00001559, -0.00002640, -0.00002449, 0.00000150, 0.00000160, -0.00003239, -0.00000244, -0.00008094, 0.00000576, -0.00006270, -0.00006877, -0.00024434, -0.00032576, 0.00054230, 0.00024034, 0.00016029, 0.00039523, 0.00038561, -0.00053197, -0.00007640, -0.00006389, 0.00004521, -0.00000167, 0.00003604, -0.00005590, -0.00000389, 0.00002448, 0.00002932, 0.00001828, 0.00007779, },
		{ 0.00001606, -0.00010526, -0.00001475, 0.00001209, -0.00002017, -0.00003137, -0.00000336, -0.00001910, 0.00001264, -0.00001785, -0.00000112, 0.00001149, -0.00001965, 0.00000762, -0.00001074, -0.00001223, -0.00000521, 0.00002199, -0.00001847, -0.00000258, 0.00000287, 0.00002044, -0.00001835, -0.00000780, 0.00000075, 0.00005448, -0.00003807, 0.00003439, -0.00003421, -0.00001092, 0.00000104, -0.00002649, 0.00000087, -0.00000527, 0.00000319, -0.00005522, 0.00001914, -0.00001412, -0.00000963, -0.00002588, -0.00004234, -0.00004167, -0.00000475, -0.00013476, -0.00009133, -0.00050897, 0.00041475, 0.00042297, 0.00011519, 0.00045849, 0.00042646, -0.00053699, -0.00011183, -0.00005549, 0.00002287, 0.00000483, 0.00003490, -0.00006057, -0.00000715, 0.00001965, 0.00002868, 0.00000444, 0.00007835, -0.00000486, },
		{ -0.00006837, -0.00005994, 0.00003921, -0.00003025, -0.00001883, -0.00002039, -0.00000430, -0.00000630, 0.00000270, -0.00002320, 0.00003281, -0.00003042, 0.00001248, -0.00001372, -0.00000412, -0.00002011, 0.00002861, -0.00001447, -0.00000508, -0.00000294, 0.00002253, -0.00001126, -0.00000947, -0.00001063, 0.00005490, -0.00001791, 0.00000351, 0.00000783, -0.00005050, 0.00002994, -0.00004619, 0.00001539, -0.00002083, 0.00002574, -0.00007310, 0.00002331, -0.00002081, 0.00000579, -0.00004315, -0.00001167, -0.00007461, 0.00003282, -0.00017490, -0.00000236, -0.00052945, 0.00011973, 0.00064147, 0.00008916, 0.00050996, 0.00048466, -0.00054685, -0.00014633, -0.00005167, 0.00000596, 0.00000624, 0.00003543, -0.00006118, -0.00001755, 0.00002386, 0.00001951, 0.00000237, 0.00006414, 0.00001282, 0.00001201, },
		{ -0.00009652, 0.00005047, -0.00002892, -0.00001207, -0.00003189, 0.00000499, -0.00001924, 0.00001523, -0.00003466, 0.00003924, -0.00002524, 0.00000519, -0.00000876, -0.00000273, -0.00002406, 0.00002064, 0.00000057, -0.00001356, -0.00000183, 0.00001544, 0.00000197, -0.00001461, -0.00001234, 0.00003832, 0.00001634, -0.00003286, 0.00004372, -0.00007164, 0.00003460, -0.00004225, 0.00000919, -0.00001734, 0.00002663, -0.00006488, 0.00000357, -0.00000814, 0.00000126, -0.00003819, -0.00000930, -0.00006947, 0.00002362, -0.00015626, -0.00001249, -0.00039203, -0.00026125, 0.00078862, 0.00015126, 0.00051112, 0.00058744, -0.00058116, -0.00016352, -0.00006554, 0.00000319, -0.00000225, 0.00003953, -0.00005905, -0.00003251, 0.00003304, 0.00000602, 0.00000793, 0.00004028, 0.00003831, -0.00000600, -0.00002310, },
		{ 0.00003261, -0.00000857, -0.00001820, -0.00002980, 0.00000043, -0.00001758, 0.00001130, -0.00002792, 0.00002579, -0.00000323, -0.00001274, 0.00000322, -0.00000890, -0.00001643, 0.00000098, 0.00002066, -0.00002237, 0.00000145, 0.00000515, 0.00001459, -0.00001797, -0.00001004, 0.00001609, 0.00004648, -0.00005301, 0.00005230, -0.00005848, 0.00000687, -0.00001236, -0.00001777, 0.00000524, 0.00000610, -0.00003359, -0.00003425, 0.00001867, -0.00002013, -0.00001249, -0.00003441, -0.00002913, -0.00002566, -0.00008989, -0.00009414, -0.00017764, -0.00059051, 0.00074824, 0.00034823, 0.00044805, 0.00074340, -0.00064481, -0.00015943, -0.00009931, 0.00001429, -0.00001929, 0.00004632, -0.00005597, -0.00004689, 0.00003977, -0.00000363, 0.00001194, 0.00001836, 0.00005769, -0.00001830, 0.00000525, -0.00010255, },
		{ 0.00002365, -0.00003238, -0.00001700, -0.00001556, -0.00000376, -0.00000588, -0.00000781, 0.00000029, 0.00002491, -0.00003074, 0.00001409, -0.00001638, -0.00000455, -0.00001989, 0.00003443, -0.00002278, -0.00000003, -0.00000093, 0.00001962, -0.00001427, -0.00001027, -0.00000028, 0.00005825, -0.00004524, 0.00002856, -0.00001544, -0.00004042, 0.00002680, -0.00004870, 0.00003094, -0.00001907, 0.00000098, -0.00006866, 0.00003968, -0.00003963, 0.00001437, -0.00006437, 0.00001741, -0.00007773, -0.00002351, -0.00017646, -0.00000154, -0.00073768, 0.00046030, 0.00065796, 0.00035823, 0.00092627, -0.00071624, -0.00015157, -0.00013752, 0.00002537, -0.00003400, 0.00004900, -0.00005026, -0.00005821, 0.00003960, -0.00000442, 0.00000790, 0.00000757, 0.00005973, -0.00001066, 0.00000316, -0.00007334, -0.00000985, },
		{ -0.00003949, -0.00000597, -0.00003082, 0.00000993, -0.00002322, 0.00001114, -0.00002088, 0.00004251, -0.00003509, 0.00001385, -0.00001665, 0.00000170, -0.00003048, 0.00003248, -0.00000973, -0.00000928, 0.00000031, 0.00001517, -0.00000330, -0.00001467, -0.00000648, 0.00004952, -0.00001493, -0.00001271, 0.00003384, -0.00007911, 0.00004806, -0.00005965, 0.00003785, -0.00002638, 0.00001447, -0.00007567, 0.00003340, -0.00003607, 0.00001869, -0.00007074, 0.00003472, -0.00008984, -0.00000832, -0.00019200, 0.00005020, -0.00064837, -0.00002926, 0.00097055, 0.00032902, 0.00108197, -0.00075414, -0.00017290, -0.00015272, 0.00001399, -0.00002988, 0.00003683, -0.00003636, -0.00006935, 0.00003520, 0.00000061, -0.00000211, 0.00000732, 0.00004522, 0.00001497, -0.00002165, -0.00002671, -0.00005383, 0.00004731, },
		{ -0.00000912, -0.00003255, 0.00001037, -0.00002703, 0.00001467, -0.00002343, 0.00003766, -0.00001853, -0.00000136, -0.00000621, -0.00000260, -0.00002605, 0.00001405, 0.00001312, -0.00002189, 0.00000514, 0.00000604, 0.00000916, -0.00001884, -0.00000668, 0.00002998, 0.00001955, -0.00004596, 0.00006039, -0.00008272, 0.00003150, -0.00003651, 0.00001445, -0.00000433, -0.00000476, -0.00004616, -0.00000614, -0.00000320, -0.00000851, -0.00004111, 0.00000679, -0.00004444, -0.00006214, -0.00012141, -0.00003119, -0.00037821, -0.00056844, 0.00111991, 0.00045631, 0.00116366, -0.00072443, -0.00025008, -0.00012367, -0.00003578, 0.00000363, 0.00000324, -0.00000994, -0.00008538, 0.00003401, 0.00000234, -0.00000881, 0.00000790, 0.00002672, 0.00004188, -0.00004546, 0.00000716, -0.00007076, 0.00004161, -0.00002531, },
		{ -0.00001780, -0.00000600, -0.00001340, -0.00000089, -0.00000516, 0.00001181, 0.00001319, -0.00002449, 0.00000948, -0.00001362, -0.00001169, -0.00001168, 0.00003436, -0.00002827, 0.00000589, 0.00000013, 0.00001537, -0.00001657, -0.00000828, 0.00001303, 0.00003911, -0.00005105, 0.00004825, -0.00004425, -0.00002020, 0.00001233, -0.00002928, 0.00003712, -0.00004551, 0.00000497, -0.00006123, 0.00004212, -0.00005194, 0.00000959, -0.00004971, 0.00003625, -0.00015433, -0.00000686, -0.00017425, -0.00006844, -0.00095905, 0.00096134, 0.00077699, 0.00117898, -0.00063296, -0.00037862, -0.00005533, -0.00011757, 0.00005902, -0.00004500, 0.00002472, -0.00010662, 0.00004081, -0.00000647, -0.00000400, -0.00000089, 0.00001858, 0.00005140, -0.00004537, 0.00000544, -0.00004575, 0.00000146, 0.00001055, -0.00002770, },
		{ -0.00002956, 0.00000820, -0.00002545, 0.00002197, -0.00001995, 0.00004322, -0.00004061, 0.00001894, -0.00002156, 0.00000135, -0.00003101, 0.00004151, -0.00002040, -0.00000255, 0.00000215, 0.00001169, -0.00000584, -0.00001479, 0.00000620, 0.00003589, -0.00002580, 0.00000504, 0.00001733, -0.00008016, 0.00005794, -0.00006567, 0.00007114, -0.00007938, 0.00004607, -0.00009902, 0.00006867, -0.00008057, 0.00004765, -0.00009539, 0.00010333, -0.00022420, 0.00007566, -0.00027311, 0.00012813, -0.00105233, 0.00045989, 0.00121203, 0.00121758, -0.00053873, -0.00051467, 0.00001562, -0.00019920, 0.00010869, -0.00008619, 0.00005241, -0.00012461, 0.00005171, -0.00002366, 0.00001105, -0.00002001, 0.00002522, 0.00003729, -0.00001578, -0.00003148, 0.00001022, -0.00005500, 0.00005024, -0.00004825, 0.00000298, },
		{ 0.00002146, -0.00004206, 0.00003975, -0.00003819, 0.00005359, -0.00003642, 0.00001299, -0.00001836, 0.00000375, -0.00003451, 0.00002923, 0.00000150, -0.00001780, 0.00000964, 0.00000175, 0.00000851, -0.00002269, 0.00000718, 0.00001672, 0.00001346, -0.00004261, 0.00006758, -0.00011231, 0.00006821, -0.00006721, 0.00007220, -0.00008007, 0.00004946, -0.00009307, 0.00005254, -0.00006991, 0.00004486, -0.00009605, 0.00011472, -0.00022208, 0.00006816, -0.00025534, 0.00013095, -0.00083038, -0.00026106, 0.00156187, 0.00142982, -0.00052798, -0.00059759, 0.00003991, -0.00023949, 0.00011935, -0.00009510, 0.00005494, -0.00012660, 0.00005597, -0.00003806, 0.00002488, -0.00003922, 0.00003758, 0.00001041, 0.00002754, -0.00007944, 0.00006506, -0.00009493, 0.00006709, -0.00005112, 0.00001311, -0.00004400, },
		{ -0.00003946, 0.00003650, -0.00003203, 0.00003724, -0.00001022, -0.00000734, -0.00000455, -0.00000514, -0.00002295, 0.00000335, 0.00002763, -0.00003035, 0.00001453, -0.00000608, 0.00001789, -0.00002473, 0.00000782, -0.00000371, 0.00004363, -0.00006622, 0.00007805, -0.00009473, 0.00003039, -0.00002755, 0.00003617, -0.00004354, 0.00001200, -0.00004382, -0.00000261, -0.00002295, 0.00000163, -0.00004946, 0.00006660, -0.00014596, -0.00002685, -0.00013021, -0.00002265, -0.00042578, -0.00094781, 0.00157372, 0.00195141, -0.00065209, -0.00059737, -0.00000652, -0.00021936, 0.00007754, -0.00006335, 0.00002693, -0.00010788, 0.00004640, -0.00003905, 0.00002506, -0.00004574, 0.00004147, -0.00001045, 0.00005947, -0.00010703, 0.00008572, -0.00009166, 0.00004430, -0.00002602, 0.00000326, -0.00003899, 0.00001559, },
		{ 0.00001591, -0.00000704, 0.00000385, 0.00002485, -0.00002936, 0.00000991, -0.00001693, -0.00000651, -0.00002238, 0.00004391, -0.00002985, 0.00000961, -0.00000541, 0.00001686, -0.00001704, 0.00000274, -0.00001465, 0.00005053, -0.00005403, 0.00004520, -0.00003597, -0.00003653, 0.00002956, -0.00001246, 0.00000542, -0.00003911, 0.00001703, -0.00006222, 0.00002626, -0.00004811, 0.00000826, 0.00000064, -0.00004815, -0.00014208, 0.00001370, -0.00019874, -0.00006196, -0.00131943, 0.00106103, 0.00280393, -0.00085998, -0.00055985, -0.00008926, -0.00016785, 0.00001001, -0.00001530, -0.00001136, -0.00008316, 0.00003079, -0.00002895, 0.00001124, -0.00003758, 0.00003180, -0.00001492, 0.00006512, -0.00009811, 0.00006116, -0.00004467, -0.00000971, 0.00001626, -0.00001868, -0.00001962, -0.00000539, -0.00002114, },
		{ 0.00001778, -0.00002513, 0.00004748, -0.00003583, 0.00001219, -0.00002045, 0.00000247, -0.00003482, 0.00004080, -0.00001294, -0.00000489, 0.00000245, 0.00000832, -0.00000409, -0.00000478, -0.00001704, 0.00003778, -0.00001858, -0.00000616, 0.00002735, -0.00008975, 0.00006151, -0.00003337, 0.00002688, -0.00006149, 0.00004353, -0.00007926, 0.00003148, -0.00005625, 0.00002346, -0.00002016, -0.00000530, -0.00018439, 0.00006174, -0.00024487, 0.00005878, -0.00120067, 0.00001903, 0.00382591, -0.00097039, -0.00061438, -0.00011444, -0.00016058, -0.00001926, -0.00000456, -0.00001675, -0.00008580, 0.00003374, -0.00002720, 0.00000087, -0.00003049, 0.00002107, -0.00001174, 0.00005247, -0.00006460, 0.00001213, 0.00001589, -0.00006145, 0.00004617, -0.00002990, -0.00000475, -0.00002380, -0.00000422, -0.00000468, },
		{ -0.00002972, 0.00004121, -0.00001643, -0.00000387, -0.00001029, -0.00000205, -0.00002953, 0.00001976, 0.00001301, -0.00001976, 0.00000836, 0.00000235, 0.00000308, -0.00000454, -0.00002165, 0.00002281, 0.00001200, -0.00003798, 0.00005183, -0.00008787, 0.00003339, 0.00000319, -0.00000759, -0.00002540, 0.00000486, -0.00002445, -0.00003483, 0.00000703, -0.00003913, 0.00004700, -0.00006907, -0.00009422, -0.00005381, -0.00008137, -0.00013855, -0.00061719, -0.00132845, 0.00468490, -0.00071781, -0.00092295, 0.00002677, -0.00028178, 0.00005817, -0.00008616, 0.00005410, -0.00015000, 0.00008329, -0.00005998, 0.00002049, -0.00005044, 0.00003336, -0.00002406, 0.00004776, -0.00003948, -0.00002046, 0.00004548, -0.00007068, 0.00003288, -0.00000769, -0.00001413, -0.00002023, -0.00000782, -0.00000601, 0.00002244, },
		{ 0.00000557, 0.00002399, -0.00003188, 0.00000849, -0.00001590, -0.00001316, -0.00000740, 0.00003350, -0.00002229, 0.00000192, 0.00000893, -0.00000515, 0.00001199, -0.00003882, 0.00002282, 0.00001338, -0.00002352, 0.00001496, -0.00001851, -0.00005661, 0.00009313, -0.00009254, 0.00006360, -0.00009169, 0.00009085, -0.00015854, 0.00012868, -0.00016856, 0.00019241, -0.00022857, 0.00011160, -0.00030554, 0.00024708, -0.00055935, 0.00020078, -0.00258775, 0.00498516, 0.00013182, -0.00158395, 0.00038359, -0.00056486, 0.00026642, -0.00027635, 0.00021178, -0.00028377, 0.00018786, -0.00013932, 0.00008563, -0.00011416, 0.00008564, -0.00007088, 0.00007533, -0.00005344, -0.00000260, 0.00001370, -0.00001709, -0.00003332, 0.00005218, -0.00005245, 0.00001169, -0.00003882, 0.00001909, -0.00000027, -0.00000444, },
		{ 0.00006643, -0.00005544, 0.00002335, -0.00002810, 0.00000168, -0.00002590, 0.00003583, -0.00000540, -0.00001984, 0.00002929, -0.00002838, 0.00004278, -0.00006788, 0.00004106, -0.00001801, 0.00003441, -0.00006853, 0.00008974, -0.00017016, 0.00019491, -0.00018648, 0.00016313, -0.00020013, 0.00021538, -0.00028524, 0.00025366, -0.00030732, 0.00035251, -0.00040937, 0.00034125, -0.00057780, 0.00059008, -0.00098565, 0.00091448, -0.00334423, 0.00443368, 0.00164302, -0.00253691, 0.00088171, -0.00094620, 0.00054987, -0.00052489, 0.00041195, -0.00045014, 0.00031889, -0.00024546, 0.00018256, -0.00021059, 0.00016906, -0.00014737, 0.00013592, -0.00011056, 0.00006765, -0.00007283, 0.00008201, -0.00012874, 0.00012684, -0.00010205, 0.00005821, -0.00008489, 0.00005906, -0.00003888, 0.00003728, -0.00004223, },
		{ -0.00005796, 0.00002288, -0.00002906, 0.00000513, -0.00002732, 0.00001768, 0.00002585, -0.00004782, 0.00005272, -0.00005387, 0.00007320, -0.00009447, 0.00006315, -0.00006185, 0.00010365, -0.00015281, 0.00018018, -0.00024690, 0.00025027, -0.00023356, 0.00021644, -0.00025973, 0.00028392, -0.00034755, 0.00031294, -0.00038068, 0.00044228, -0.00051151, 0.00047466, -0.00072612, 0.00077054, -0.00118692, 0.00122801, -0.00333588, 0.00298887, 0.00362198, -0.00352725, 0.00131555, -0.00126742, 0.00077917, -0.00072268, 0.00056295, -0.00057326, 0.00041427, -0.00032740, 0.00026644, -0.00029780, 0.00024496, -0.00021946, 0.00019950, -0.00018060, 0.00015433, -0.00016942, 0.00017619, -0.00020457, 0.00017540, -0.00013096, 0.00009202, -0.00012038, 0.00008954, -0.00006961, 0.00007086, -0.00007663, 0.00008801, },
		{ 0.00000524, -0.00001766, -0.00000341, -0.00001465, -0.00001079, 0.00005548, -0.00006471, 0.00006195, -0.00006326, 0.00008367, -0.00009873, 0.00006653, -0.00008693, 0.00014171, -0.00018673, 0.00019944, -0.00023843, 0.00021818, -0.00019676, 0.00018589, -0.00023000, 0.00025391, -0.00030293, 0.00026277, -0.00033890, 0.00040278, -0.00046466, 0.00042889, -0.00065529, 0.00067952, -0.00103643, 0.00102014, -0.00256850, 0.00091500, 0.00563490, -0.00416783, 0.00141804, -0.00133840, 0.00080605, -0.00074937, 0.00056656, -0.00057253, 0.00040647, -0.00032613, 0.00028159, -0.00032165, 0.00026211, -0.00023956, 0.00022006, -0.00021465, 0.00020254, -0.00021671, 0.00020838, -0.00021316, 0.00016229, -0.00011240, 0.00008922, -0.00012193, 0.00008827, -0.00007117, 0.00007538, -0.00007863, 0.00007553, -0.00003354, },
		{ -0.00000215, -0.00001574, 0.00000044, -0.00003296, 0.00006562, -0.00005548, 0.00004448, -0.00004459, 0.00006241, -0.00006905, 0.00003668, -0.00007177, 0.00012161, -0.00014402, 0.00012885, -0.00013781, 0.00010120, -0.00008215, 0.00007671, -0.00011508, 0.00013044, -0.00015906, 0.00011124, -0.00018673, 0.00023517, -0.00026790, 0.00020401, -0.00037142, 0.00033412, -0.00057175, 0.00039149, -0.00131492, -0.00129365, 0.00714295, -0.00408551, 0.00098601, -0.00103702, 0.00054389, -0.00053985, 0.00037339, -0.00040890, 0.00026201, -0.00020969, 0.00019422, -0.00024723, 0.00018704, -0.00017567, 0.00016460, -0.00017569, 0.00017166, -0.00017570, 0.00014852, -0.00013719, 0.00008071, -0.00004373, 0.00004557, -0.00008357, 0.00004946, -0.00003827, 0.00004642, -0.00004679, 0.00003357, 0.00000831, -0.00001968, },
		{ -0.00002140, 0.00000596, -0.00003618, 0.00004764, -0.00001799, 0.00000233, -0.00000161, 0.00001470, -0.00001196, -0.00002123, -0.00001904, 0.00004846, -0.00003913, -0.00000344, 0.00001382, -0.00005278, 0.00006206, -0.00006446, 0.00003775, -0.00003566, 0.00002878, -0.00008427, 0.00001778, 0.00000118, 0.00000890, -0.00011635, 0.00002101, -0.00013240, 0.00002987, -0.00039371, 0.00000206, -0.00309171, 0.00770270, -0.00309171, 0.00000206, -0.00039371, 0.00002987, -0.00013240, 0.00002101, -0.00011635, 0.00000890, 0.00000118, 0.00001778, -0.00008427, 0.00002878, -0.00003566, 0.00003775, -0.00006446, 0.00006206, -0.00005278, 0.00001382, -0.00000344, -0.00003913, 0.00004846, -0.00001904, -0.00002123, -0.00001196, 0.00001470, -0.00000161, 0.00000233, -0.00001799, 0.00004764, -0.00003618, 0.00000596, },
		{ -0.00000215, -0.00001968, 0.00000831, 0.00003357, -0.00004679, 0.00004642, -0.00003827, 0.00004946, -0.00008357, 0.00004557, -0.00004373, 0.00008071, -0.00013719, 0.00014852, -0.00017570, 0.00017166, -0.00017569, 0.00016460, -0.00017567, 0.00018704, -0.00024723, 0.00019422, -0.00020969, 0.00026201, -0.00040890, 0.00037339, -0.00053985, 0.00054389, -0.00103702, 0.00098601, -0.00408551, 0.00714295, -0.00129365, -0.00131492, 0.00039149, -0.00057175, 0.00033412, -0.00037142, 0.00020401, -0.00026790, 0.00023517, -0.00018673, 0.00011124, -0.00015906, 0.00013044, -0.00011508, 0.00007671, -0.00008215, 0.00010120, -0.00013781, 0.00012885, -0.00014402, 0.00012161, -0.00007177, 0.00003668, -0.00006905, 0.00006241, -0.00004459, 0.00004448, -0.00005548, 0.00006562, -0.00003296, 0.00000044, -0.00001574, },
		{ 0.00000524, -0.00003354, 0.00007553, -0.00007863, 0.00007538, -0.00007117, 0.00008827, -0.00012193, 0.00008922, -0.00011240, 0.00016229, -0.00021316, 0.00020838, -0.00021671, 0.00020254, -0.00021465, 0.00022006, -0.00023956, 0.00026211, -0.00032165, 0.00028159, -0.00032613, 0.00040647, -0.00057253, 0.00056656, -0.00074937, 0.00080605, -0.00133840, 0.00141804, -0.00416783, 0.00563490, 0.00091500, -0.00256850, 0.00102014, -0.00103643, 0.00067952, -0.00065529, 0.00042889, -0.00046466, 0.00040278, -0.00033890, 0.00026277, -0.00030293, 0.00025391, -0.00023000, 0.00018589, -0.00019676, 0.00021818, -0.00023843, 0.00019944, -0.00018673, 0.00014171, -0.00008693, 0.00006653, -0.00009873, 0.00008367, -0.00006326, 0.00006195, -0.00006471, 0.00005548, -0.00001079, -0.00001465, -0.00000341, -0.00001766, },
		{ -0.00005796, 0.00008801, -0.00007663, 0.00007086, -0.00006961, 0.00008954, -0.00012038, 0.00009202, -0.00013096, 0.00017540, -0.00020457, 0.00017619, -0.00016942, 0.00015433, -0.00018060, 0.00019950, -0.00021946, 0.00024496, -0.00029780, 0.00026644, -0.00032740, 0.00041427, -0.00057326, 0.00056295, -0.00072268, 0.00077917, -0.00126742, 0.00131555, -0.00352725, 0.00362198, 0.00298887, -0.00333588, 0.00122801, -0.00118692, 0.00077054, -0.00072612, 0.00047466, -0.00051151, 0.00044228, -0.00038068, 0.00031294, -0.00034755, 0.00028392, -0.00025973, 0.00021644, -0.00023356, 0.00025027, -0.00024690, 0.00018018, -0.00015281, 0.00010365, -0.00006185, 0.00006315, -0.00009447, 0.00007320, -0.00005387, 0.00005272, -0.00004782, 0.00002585, 0.00001768, -0.00002732, 0.00000513, -0.00002906, 0.00002288, },
		{ 0.00006643, -0.00004223, 0.00003728, -0.00003888, 0.00005906, -0.00008489, 0.00005821, -0.00010205, 0.00012684, -0.00012874, 0.00008201, -0.00007283, 0.00006765, -0.00011056, 0.00013592, -0.00014737, 0.00016906, -0.00021059, 0.00018256, -0.00024546, 0.00031889, -0.00045014, 0.00041195, -0.00052489, 0.00054987, -0.00094620, 0.00088171, -0.00253691, 0.00164302, 0.00443368, -0.00334423, 0.00091448, -0.00098565, 0.00059008, -0.00057780, 0.00034125, -0.00040937, 0.00035251, -0.00030732, 0.00025366, -0.00028524, 0.00021538, -0.00020013, 0.00016313, -0.00018648, 0.00019491, -0.00017016, 0.00008974, -0.00006853, 0.00003441, -0.00001801, 0.00004106, -0.00006788, 0.00004278, -0.00002838, 0.00002929, -0.00001984, -0.00000540, 0.00003583, -0.00002590, 0.00000168, -0.00002810, 0.00002335, -0.00005544, },
		{ 0.00000557, -0.00000444, -0.00000027, 0.00001909, -0.00003882, 0.00001169, -0.00005245, 0.00005218, -0.00003332, -0.00001709, 0.00001370, -0.00000260, -0.00005344, 0.00007533, -0.00007088, 0.00008564, -0.00011416, 0.00008563, -0.00013932, 0.00018786, -0.00028377, 0.00021178, -0.00027635, 0.00026642, -0.00056486, 0.00038359, -0.00158395, 0.00013182, 0.00498516, -0.00258775, 0.00020078, -0.00055935, 0.00024708, -0.00030554, 0.00011160, -0.00022857, 0.00019241, -0.00016856, 0.00012868, -0.00015854, 0.00009085, -0.00009169, 0.00006360, -0.00009254, 0.00009313, -0.00005661, -0.00001851, 0.00001496, -0.00002352, 0.00001338, 0.00002282, -0.00003882, 0.00001199, -0.00000515, 0.00000893, 0.00000192, -0.00002229, 0.00003350, -0.00000740, -0.00001316, -0.00001590, 0.00000849, -0.00003188, 0.00002399, },
		{ -0.00002972, 0.00002244, -0.00000601, -0.00000782, -0.00002023, -0.00001413, -0.00000769, 0.00003288, -0.00007068, 0.00004548, -0.00002046, -0.00003948, 0.00004776, -0.00002406, 0.00003336, -0.00005044, 0.00002049, -0.00005998, 0.00008329, -0.00015000, 0.00005410, -0.00008616, 0.00005817, -0.00028178, 0.00002677, -0.00092295, -0.00071781, 0.00468490, -0.00132845, -0.00061719, -0.00013855, -0.00008137, -0.00005381, -0.00009422, -0.00006907, 0.00004700, -0.00003913, 0.00000703, -0.00003483, -0.00002445, 0.00000486, -0.00002540, -0.00000759, 0.00000319, 0.00003339, -0.00008787, 0.00005183, -0.00003798, 0.00001200, 0.00002281, -0.00002165, -0.00000454, 0.00000308, 0.00000235, 0.00000836, -0.00001976, 0.00001301, 0.00001976, -0.00002953, -0.00000205, -0.00001029, -0.00000387, -0.00001643, 0.00004121, },
		{ 0.00001778, -0.00000468, -0.00000422, -0.00002380, -0.00000475, -0.00002990, 0.00004617, -0.00006145, 0.00001589, 0.00001213, -0.00006460, 0.00005247, -0.00001174, 0.00002107, -0.00003049, 0.00000087, -0.00002720, 0.00003374, -0.00008580, -0.00001675, -0.00000456, -0.00001926, -0.00016058, -0.00011444, -0.00061438, -0.00097039, 0.00382591, 0.00001903, -0.00120067, 0.00005878, -0.00024487, 0.00006174, -0.00018439, -0.00000530, -0.00002016, 0.00002346, -0.00005625, 0.00003148, -0.00007926, 0.00004353, -0.00006149, 0.00002688, -0.00003337, 0.00006151, -0.00008975, 0.00002735, -0.00000616, -0.00001858, 0.00003778, -0.00001704, -0.00000478, -0.00000409, 0.00000832, 0.00000245, -0.00000489, -0.00001294, 0.00004080, -0.00003482, 0.00000247, -0.00002045, 0.00001219, -0.00003583, 0.00004748, -0.00002513, },
		{ 0.00001591, -0.00002114, -0.00000539, -0.00001962, -0.00001868, 0.00001626, -0.00000971, -0.00004467, 0.00006116, -0.00009811, 0.00006512, -0.00001492, 0.00003180, -0.00003758, 0.00001124, -0.00002895, 0.00003079, -0.00008316, -0.00001136, -0.00001530, 0.00001001, -0.00016785, -0.00008926, -0.00055985, -0.00085998, 0.00280393, 0.00106103, -0.00131943, -0.00006196, -0.00019874, 0.00001370, -0.00014208, -0.00004815, 0.00000064, 0.00000826, -0.00004811, 0.00002626, -0.00006222, 0.00001703, -0.00003911, 0.00000542, -0.00001246, 0.00002956, -0.00003653, -0.00003597, 0.00004520, -0.00005403, 0.00005053, -0.00001465, 0.00000274, -0.00001704, 0.00001686, -0.00000541, 0.00000961, -0.00002985, 0.00004391, -0.00002238, -0.00000651, -0.00001693, 0.00000991, -0.00002936, 0.00002485, 0.00000385, -0.00000704, },
		{ -0.00003946, 0.00001559, -0.00003899, 0.00000326, -0.00002602, 0.00004430, -0.00009166, 0.00008572, -0.00010703, 0.00005947, -0.00001045, 0.00004147, -0.00004574, 0.00002506, -0.00003905, 0.00004640, -0.00010788, 0.00002693, -0.00006335, 0.00007754, -0.00021936, -0.00000652, -0.00059737, -0.00065209, 0.00195141, 0.00157372, -0.00094781, -0.00042578, -0.00002265, -0.00013021, -0.00002685, -0.00014596, 0.00006660, -0.00004946, 0.00000163, -0.00002295, -0.00000261, -0.00004382, 0.00001200, -0.00004354, 0.00003617, -0.00002755, 0.00003039, -0.00009473, 0.00007805, -0.00006622, 0.00004363, -0.00000371, 0.00000782, -0.00002473, 0.00001789, -0.00000608, 0.00001453, -0.00003035, 0.00002763, 0.00000335, -0.00002295, -0.00000514, -0.00000455, -0.00000734, -0.00001022, 0.00003724, -0.00003203, 0.00003650, },
		{ 0.00002146, -0.00004400, 0.00001311, -0.00005112, 0.00006709, -0.00009493, 0.00006506, -0.00007944, 0.00002754, 0.00001041, 0.00003758, -0.00003922, 0.00002488, -0.00003806, 0.00005597, -0.00012660, 0.00005494, -0.00009510, 0.00011935, -0.00023949, 0.00003991, -0.00059759, -0.00052798, 0.00142982, 0.00156187, -0.00026106, -0.00083038, 0.00013095, -0.00025534, 0.00006816, -0.00022208, 0.00011472, -0.00009605, 0.00004486, -0.00006991, 0.00005254, -0.00009307, 0.00004946, -0.00008007, 0.00007220, -0.00006721, 0.00006821, -0.00011231, 0.00006758, -0.00004261, 0.00001346, 0.00001672, 0.00000718, -0.00002269, 0.00000851, 0.00000175, 0.00000964, -0.00001780, 0.00000150, 0.00002923, -0.00003451, 0.00000375, -0.00001836, 0.00001299, -0.00003642, 0.00005359, -0.00003819, 0.00003975, -0.00004206, },
		{ -0.00002956, 0.00000298, -0.00004825, 0.00005024, -0.00005500, 0.00001022, -0.00003148, -0.00001578, 0.00003729, 0.00002522, -0.00002001, 0.00001105, -0.00002366, 0.00005171, -0.00012461, 0.00005241, -0.00008619, 0.00010869, -0.00019920, 0.00001562, -0.00051467, -0.00053873, 0.00121758, 0.00121203, 0.00045989, -0.00105233, 0.00012813, -0.00027311, 0.00007566, -0.00022420, 0.00010333, -0.00009539, 0.00004765, -0.00008057, 0.00006867, -0.00009902, 0.00004607, -0.00007938, 0.00007114, -0.00006567, 0.00005794, -0.00008016, 0.00001733, 0.00000504, -0.00002580, 0.00003589, 0.00000620, -0.00001479, -0.00000584, 0.00001169, 0.00000215, -0.00000255, -0.00002040, 0.00004151, -0.00003101, 0.00000135, -0.00002156, 0.00001894, -0.00004061, 0.00004322, -0.00001995, 0.00002197, -0.00002545, 0.00000820, },
		{ -0.00001780, -0.00002770, 0.00001055, 0.00000146, -0.00004575, 0.00000544, -0.00004537, 0.00005140, 0.00001858, -0.00000089, -0.00000400, -0.00000647, 0.00004081, -0.00010662, 0.00002472, -0.00004500, 0.00005902, -0.00011757, -0.00005533, -0.00037862, -0.00063296, 0.00117898, 0.00077699, 0.00096134, -0.00095905, -0.00006844, -0.00017425, -0.00000686, -0.00015433, 0.00003625, -0.00004971, 0.00000959, -0.00005194, 0.00004212, -0.00006123, 0.00000497, -0.00004551, 0.00003712, -0.00002928, 0.00001233, -0.00002020, -0.00004425, 0.00004825, -0.00005105, 0.00003911, 0.00001303, -0.00000828, -0.00001657, 0.00001537, 0.00000013, 0.00000589, -0.00002827, 0.00003436, -0.00001168, -0.00001169, -0.00001362, 0.00000948, -0.00002449, 0.00001319, 0.00001181, -0.00000516, -0.00000089, -0.00001340, -0.00000600, },
		{ -0.00000912, -0.00002531, 0.00004161, -0.00007076, 0.00000716, -0.00004546, 0.00004188, 0.00002672, 0.00000790, -0.00000881, 0.00000234, 0.00003401, -0.00008538, -0.00000994, 0.00000324, 0.00000363, -0.00003578, -0.00012367, -0.00025008, -0.00072443, 0.00116366, 0.00045631, 0.00111991, -0.00056844, -0.00037821, -0.00003119, -0.00012141, -0.00006214, -0.00004444, 0.00000679, -0.00004111, -0.00000851, -0.00000320, -0.00000614, -0.00004616, -0.00000476, -0.00000433, 0.00001445, -0.00003651, 0.00003150, -0.00008272, 0.00006039, -0.00004596, 0.00001955, 0.00002998, -0.00000668, -0.00001884, 0.00000916, 0.00000604, 0.00000514, -0.00002189, 0.00001312, 0.00001405, -0.00002605, -0.00000260, -0.00000621, -0.00000136, -0.00001853, 0.00003766, -0.00002343, 0.00001467, -0.00002703, 0.00001037, -0.00003255, },
		{ -0.00003949, 0.00004731, -0.00005383, -0.00002671, -0.00002165, 0.00001497, 0.00004522, 0.00000732, -0.00000211, 0.00000061, 0.00003520, -0.00006935, -0.00003636, 0.00003683, -0.00002988, 0.00001399, -0.00015272, -0.00017290, -0.00075414, 0.00108197, 0.00032902, 0.00097055, -0.00002926, -0.00064837, 0.00005020, -0.00019200, -0.00000832, -0.00008984, 0.00003472, -0.00007074, 0.00001869, -0.00003607, 0.00003340, -0.00007567, 0.00001447, -0.00002638, 0.00003785, -0.00005965, 0.00004806, -0.00007911, 0.00003384, -0.00001271, -0.00001493, 0.00004952, -0.00000648, -0.00001467, -0.00000330, 0.00001517, 0.00000031, -0.00000928, -0.00000973, 0.00003248, -0.00003048, 0.00000170, -0.00001665, 0.00001385, -0.00003509, 0.00004251, -0.00002088, 0.00001114, -0.00002322, 0.00000993, -0.00003082, -0.00000597, },
		{ 0.00002365, -0.00000985, -0.00007334, 0.00000316, -0.00001066, 0.00005973, 0.00000757, 0.00000790, -0.00000442, 0.00003960, -0.00005821, -0.00005026, 0.00004900, -0.00003400, 0.00002537, -0.00013752, -0.00015157, -0.00071624, 0.00092627, 0.00035823, 0.00065796, 0.00046030, -0.00073768, -0.00000154, -0.00017646, -0.00002351, -0.00007773, 0.00001741, -0.00006437, 0.00001437, -0.00003963, 0.00003968, -0.00006866, 0.00000098, -0.00001907, 0.00003094, -0.00004870, 0.00002680, -0.00004042, -0.00001544, 0.00002856, -0.00004524, 0.00005825, -0.00000028, -0.00001027, -0.00001427, 0.00001962, -0.00000093, -0.00000003, -0.00002278, 0.00003443, -0.00001989, -0.00000455, -0.00001638, 0.00001409, -0.00003074, 0.00002491, 0.00000029, -0.00000781, -0.00000588, -0.00000376, -0.00001556, -0.00001700, -0.00003238, },
		{ 0.00003261, -0.00010255, 0.00000525, -0.00001830, 0.00005769, 0.00001836, 0.00001194, -0.00000363, 0.00003977, -0.00004689, -0.00005597, 0.00004632, -0.00001929, 0.00001429, -0.00009931, -0.00015943, -0.00064481, 0.00074340, 0.00044805, 0.00034823, 0.00074824, -0.00059051, -0.00017764, -0.00009414, -0.00008989, -0.00002566, -0.00002913, -0.00003441, -0.00001249, -0.00002013, 0.00001867, -0.00003425, -0.00003359, 0.00000610, 0.00000524, -0.00001777, -0.00001236, 0.00000687, -0.00005848, 0.00005230, -0.00005301, 0.00004648, 0.00001609, -0.00001004, -0.00001797, 0.00001459, 0.00000515, 0.00000145, -0.00002237, 0.00002066, 0.00000098, -0.00001643, -0.00000890, 0.00000322, -0.00001274, -0.00000323, 0.00002579, -0.00002792, 0.00001130, -0.00001758, 0.00000043, -0.00002980, -0.00001820, -0.00000857, },
		{ -0.00009652, -0.00002310, -0.00000600, 0.00003831, 0.00004028, 0.00000793, 0.00000602, 0.00003304, -0.00003251, -0.00005905, 0.00003953, -0.00000225, 0.00000319, -0.00006554, -0.00016352, -0.00058116, 0.00058744, 0.00051112, 0.00015126, 0.00078862, -0.00026125, -0.00039203, -0.00001249, -0.00015626, 0.00002362, -0.00006947, -0.00000930, -0.00003819, 0.00000126, -0.00000814, 0.00000357, -0.00006488, 0.00002663, -0.00001734, 0.00000919, -0.00004225, 0.00003460, -0.00007164, 0.00004372, -0.00003286, 0.00001634, 0.00003832, -0.00001234, -0.00001461, 0.00000197, 0.00001544, -0.00000183, -0.00001356, 0.00000057, 0.00002064, -0.00002406, -0.00000273, -0.00000876, 0.00000519, -0.00002524, 0.00003924, -0.00003466, 0.00001523, -0.00001924, 0.00000499, -0.00003189, -0.00001207, -0.00002892, 0.00005047, },
		{ -0.00006837, 0.00001201, 0.00001282, 0.00006414, 0.00000237, 0.00001951, 0.00002386, -0.00001755, -0.00006118, 0.00003543, 0.00000624, 0.00000596, -0.00005167, -0.00014633, -0.00054685, 0.00048466, 0.00050996, 0.00008916, 0.00064147, 0.00011973, -0.00052945, -0.00000236, -0.00017490, 0.00003282, -0.00007461, -0.00001167, -0.00004315, 0.00000579, -0.00002081, 0.00002331, -0.00007310, 0.00002574, -0.00002083, 0.00001539, -0.00004619, 0.00002994, -0.00005050, 0.00000783, 0.00000351, -0.00001791, 0.00005490, -0.00001063, -0.00000947, -0.00001126, 0.00002253, -0.00000294, -0.00000508, -0.00001447, 0.00002861, -0.00002011, -0.00000412, -0.00001372, 0.00001248, -0.00003042, 0.00003281, -0.00002320, 0.00000270, -0.00000630, -0.00000430, -0.00002039, -0.00001883, -0.00003025, 0.00003921, -0.00005994, },
		{ 0.00001606, -0.00000486, 0.00007835, 0.00000444, 0.00002868, 0.00001965, -0.00000715, -0.00006057, 0.00003490, 0.00000483, 0.00002287, -0.00005549, -0.00011183, -0.00053699, 0.00042646, 0.00045849, 0.00011519, 0.00042297, 0.00041475, -0.00050897, -0.00009133, -0.00013476, -0.00000475, -0.00004167, -0.00004234, -0.00002588, -0.00000963, -0.00001412, 0.00001914, -0.00005522, 0.00000319, -0.00000527, 0.00000087, -0.00002649, 0.00000104, -0.00001092, -0.00003421, 0.00003439, -0.00003807, 0.00005448, 0.00000075, -0.00000780, -0.00001835, 0.00002044, 0.00000287, -0.00000258, -0.00001847, 0.00002199, -0.00000521, -0.00001223, -0.00001074, 0.00000762, -0.00001965, 0.00001149, -0.00000112, -0.00001785, 0.00001264, -0.00001910, -0.00000336, -0.00003137, -0.00002017, 0.00001209, -0.00001475, -0.00010526, },
		{ -0.00000945, 0.00007779, 0.00001828, 0.00002932, 0.00002448, -0.00000389, -0.00005590, 0.00003604, -0.00000167, 0.00004521, -0.00006389, -0.00007640, -0.00053197, 0.00038561, 0.00039523, 0.00016029, 0.00024034, 0.00054230, -0.00032576, -0.00024434, -0.00006877, -0.00006270, 0.00000576, -0.00008094, -0.00000244, -0.00003239, 0.00000160, 0.00000150, -0.00002449, -0.00002640, 0.00001559, -0.00002058, 0.00000033, -0.00003102, 0.00002249, -0.00005883, 0.00004164, -0.00003353, 0.00003426, 0.00002071, -0.00000996, -0.00001762, 0.00000921, 0.00001402, -0.00000489, -0.00001367, 0.00000674, 0.00001246, -0.00002008, -0.00000528, -0.00000294, -0.00000285, -0.00001145, 0.00001747, -0.00003294, 0.00002477, -0.00002724, 0.00000734, -0.00003801, -0.00001223, -0.00001181, 0.00001543, -0.00011379, -0.00000403, },
		{ 0.00006681, 0.00004000, 0.00002436, 0.00003585, -0.00000492, -0.00004917, 0.00003798, -0.00000810, 0.00006379, -0.00006417, -0.00005463, -0.00051482, 0.00033861, 0.00035197, 0.00017584, 0.00014717, 0.00050352, -0.00004961, -0.00038273, -0.00003129, -0.00010203, 0.00003575, -0.00010172, 0.00000701, -0.00004556, 0.00001088, -0.00001323, 0.00000059, -0.00004505, 0.00002579, -0.00003317, 0.00001667, -0.00004703, 0.00003207, -0.00005442, 0.00002188, -0.00000830, 0.00000328, 0.00004033, -0.00001075, -0.00001306, -0.00000518, 0.00002369, -0.00000593, -0.00000735, -0.00000723, 0.00002303, -0.00002031, -0.00000368, -0.00001102, 0.00000889, -0.00002367, 0.00002175, -0.00003367, 0.00002248, -0.00002209, 0.00000539, -0.00003239, -0.00001459, -0.00002107, 0.00001992, -0.00009175, -0.00004101, -0.00000855, },
		{ 0.00006133, 0.00002110, 0.00004701, -0.00000391, -0.00004462, 0.00004166, -0.00001209, 0.00007432, -0.00005137, -0.00005054, -0.00048169, 0.00027866, 0.00033728, 0.00015126, 0.00013655, 0.00036415, 0.00021512, -0.00042890, -0.00006114, -0.00009958, 0.00003078, -0.00009214, -0.00000639, -0.00004133, 0.00000558, -0.00001525, 0.00000907, -0.00004371, 0.00001868, -0.00003028, 0.00001528, -0.00004102, 0.00001664, -0.00002723, -0.00001212, 0.00002142, -0.00002233, 0.00004795, -0.00000370, -0.00001043, -0.00001528, 0.00002490, -0.00000029, -0.00000552, -0.00001304, 0.00002175, -0.00001110, -0.00000762, -0.00001289, 0.00001036, -0.00002121, 0.00001087, -0.00002143, 0.00000837, -0.00000702, -0.00000563, -0.00001809, -0.00002475, -0.00001756, 0.00000501, -0.00005462, -0.00007489, -0.00001623, 0.00005398, },
		{ 0.00002546, 0.00005235, 0.00000409, -0.00004514, 0.00004785, -0.00001352, 0.00007766, -0.00002843, -0.00005804, -0.00044066, 0.00021388, 0.00033953, 0.00010514, 0.00016573, 0.00020971, 0.00038232, -0.00034884, -0.00015490, -0.00006445, -0.00000113, -0.00006098, -0.00003406, -0.00002625, -0.00000991, -0.00000688, 0.00000294, -0.00002688, 0.00000036, -0.00001752, 0.00000186, -0.00002139, -0.00001032, 0.00000425, -0.00004017, 0.00003749, -0.00002921, 0.00003710, 0.00001295, -0.00001204, -0.00001746, 0.00001574, 0.00001187, -0.00000826, -0.00001098, 0.00001175, 0.00000241, -0.00001302, -0.00001078, 0.00000404, -0.00000934, -0.00000671, -0.00000599, -0.00000744, 0.00000766, -0.00001566, -0.00000506, -0.00003327, -0.00001171, -0.00001365, -0.00002393, -0.00008666, -0.00004109, 0.00004352, 0.00007614, },
		{ 0.00005122, 0.00001919, -0.00004951, 0.00005513, -0.00001237, 0.00007669, -0.00000151, -0.00006752, -0.00040325, 0.00015672, 0.00034123, 0.00006330, 0.00019045, 0.00010233, 0.00042002, -0.00016825, -0.00026703, -0.00003136, -0.00003405, -0.00003061, -0.00005719, -0.00001504, -0.00002371, 0.00000174, -0.00000722, -0.00000730, -0.00001618, -0.00000657, -0.00001156, -0.00000273, -0.00003127, 0.00002183, -0.00004747, 0.00003102, -0.00001510, 0.00001171, 0.00003335, -0.00001454, -0.00001401, 0.00000084, 0.00002442, -0.00001017, -0.00000699, 0.00000114, 0.00001198, -0.00001389, -0.00000945, -0.00000381, 0.00000267, -0.00001986, 0.00000196, -0.00001508, 0.00001269, -0.00001617, -0.00000138, -0.00003244, -0.00001250, -0.00002336, -0.00001316, -0.00007099, -0.00007797, 0.00003079, 0.00008392, 0.00003828, },
		{ 0.00003707, -0.00005300, 0.00006009, -0.00000787, 0.00007369, 0.00002471, -0.00007241, -0.00037634, 0.00011404, 0.00033187, 0.00004089, 0.00018808, 0.00006055, 0.00035358, 0.00004460, -0.00033614, -0.00003626, -0.00004426, -0.00002019, -0.00006100, -0.00001857, -0.00002677, 0.00000214, -0.00001179, 0.00000391, -0.00002093, -0.00000574, -0.00001624, 0.00000474, -0.00003592, 0.00001843, -0.00003303, 0.00000729, 0.00000988, -0.00001552, 0.00004681, -0.00001149, -0.00001060, -0.00001202, 0.00002950, -0.00000521, -0.00000703, -0.00000340, 0.00001254, -0.00000772, -0.00001121, -0.00000872, 0.00000828, -0.00002228, -0.00000184, -0.00001170, 0.00000624, -0.00000638, -0.00000727, -0.00002251, -0.00002053, -0.00002220, -0.00002091, -0.00003900, -0.00011011, 0.00000635, 0.00008701, 0.00005643, 0.00004745, },
		{ -0.00005073, 0.00005993, 0.00000052, 0.00006975, 0.00004868, -0.00007143, -0.00035970, 0.00008427, 0.00031173, 0.00003759, 0.00016085, 0.00006645, 0.00024060, 0.00021605, -0.00031989, -0.00009099, -0.00002793, -0.00003237, -0.00004511, -0.00003562, -0.00002002, -0.00000579, -0.00000895, 0.00000419, -0.00001314, -0.00001443, -0.00001211, 0.00000058, -0.00002591, 0.00000043, -0.00000860, -0.00001863, 0.00002998, -0.00003070, 0.00004468, 0.00000100, -0.00001119, -0.00001742, 0.00002326, 0.00000782, -0.00001190, -0.00000134, 0.00000581, 0.00000209, -0.00001378, -0.00001076, 0.00000699, -0.00001543, -0.00001335, -0.00000252, -0.00000573, 0.00000688, -0.00001557, -0.00001093, -0.00002884, -0.00001743, -0.00003511, -0.00000975, -0.00012039, -0.00003442, 0.00008468, 0.00007667, 0.00004588, 0.00005255, },
		{ 0.00005464, 0.00001178, 0.00006568, 0.00007095, -0.00006677, -0.00034882, 0.00006117, 0.00028726, 0.00004396, 0.00012427, 0.00008820, 0.00013673, 0.00030167, -0.00021732, -0.00017406, -0.00000361, -0.00005304, -0.00002269, -0.00005429, -0.00001282, -0.00001475, -0.00000422, -0.00000101, -0.00000011, -0.00002496, -0.00000580, -0.00000817, -0.00001106, -0.00001862, 0.00001011, -0.00003213, 0.00003387, -0.00002735, 0.00002655, 0.00002014, -0.00001445, -0.00001566, 0.00000863, 0.00002355, -0.00001642, 0.00000200, -0.00000134, 0.00000875, -0.00001265, -0.00001277, 0.00000252, -0.00000569, -0.00002429, 0.00000410, -0.00001474, 0.00001492, -0.00001783, -0.00000594, -0.00002978, -0.00001733, -0.00004376, 0.00000285, -0.00010341, -0.00008395, 0.00007005, 0.00009772, 0.00004961, 0.00006275, -0.00004075, },
		{ 0.00002340, 0.00006249, 0.00009248, -0.00006103, -0.00033942, 0.00003939, 0.00026419, 0.00005085, 0.00009282, 0.00010140, 0.00007200, 0.00030003, -0.00006648, -0.00024301, 0.00000152, -0.00006360, -0.00000960, -0.00006164, -0.00001488, -0.00001707, -0.00000413, -0.00000458, 0.00000965, -0.00002942, -0.00000400, -0.00001399, -0.00000114, -0.00002765, 0.00001360, -0.00002797, 0.00002125, -0.00001002, 0.00000121, 0.00003708, -0.00001475, -0.00001179, -0.00000672, 0.00003333, -0.00001363, 0.00000037, -0.00000259, 0.00000763, -0.00000578, -0.00001627, -0.00000167, 0.00000118, -0.00002849, 0.00000316, -0.00001632, 0.00001366, -0.00001090, -0.00000996, -0.00002188, -0.00002371, -0.00004349, -0.00000190, -0.00006949, -0.00012426, 0.00003425, 0.00011781, 0.00006027, 0.00006771, -0.00002452, 0.00004710, },
		{ 0.00006084, 0.00011396, -0.00005539, -0.00032997, 0.00001742, 0.00024396, 0.00005454, 0.00007232, 0.00009809, 0.00004709, 0.00024263, 0.00007840, -0.00025951, -0.00002822, -0.00005590, -0.00001254, -0.00005336, -0.00002818, -0.00001146, -0.00001004, -0.00000398, 0.00001266, -0.00002556, -0.00000809, -0.00001441, 0.00000029, -0.00002421, 0.00000321, -0.00001232, 0.00000191, 0.00000915, -0.00001864, 0.00004233, -0.00000684, -0.00001089, -0.00001576, 0.00003121, -0.00000069, -0.00000767, 0.00000269, 0.00000031, 0.00000366, -0.00001911, -0.00000530, 0.00000364, -0.00002558, -0.00000388, -0.00001253, 0.00000585, 0.00000128, -0.00001817, -0.00001088, -0.00003090, -0.00003996, -0.00001369, -0.00003750, -0.00013764, -0.00002443, 0.00012996, 0.00007997, 0.00006889, -0.00000511, 0.00004110, 0.00003297, },
		{ 0.00013588, -0.00004959, -0.00032138, -0.00000319, 0.00022464, 0.00005600, 0.00006094, 0.00008319, 0.00004513, 0.00017057, 0.00017476, -0.00021090, -0.00008499, -0.00003860, -0.00002506, -0.00003630, -0.00004557, -0.00000398, -0.00001713, -0.00000242, 0.00001166, -0.00001741, -0.00001389, -0.00001232, -0.00000388, -0.00001368, -0.00001180, 0.00000279, -0.00001171, 0.00001873, -0.00002418, 0.00003225, 0.00000921, -0.00001372, -0.00001663, 0.00001806, 0.00001806, -0.00001687, 0.00000882, -0.00000634, 0.00000912, -0.00001721, -0.00001025, 0.00000373, -0.00001967, -0.00001136, -0.00000906, -0.00000228, 0.00001152, -0.00002294, -0.00000474, -0.00003140, -0.00004092, -0.00002101, -0.00002199, -0.00011903, -0.00009363, 0.00012061, 0.00011121, 0.00006832, 0.00001456, 0.00003922, 0.00003978, 0.00006037, },
		{ -0.00004265, -0.00031540, -0.00002037, 0.00020401, 0.00005723, 0.00005428, 0.00006553, 0.00004754, 0.00011356, 0.00020847, -0.00011517, -0.00014150, -0.00003034, -0.00003433, -0.00002212, -0.00005678, -0.00000265, -0.00001912, -0.00000474, 0.00001147, -0.00001078, -0.00001625, -0.00001177, -0.00000871, -0.00000363, -0.00002176, 0.00000800, -0.00001284, 0.00001512, -0.00001567, 0.00001179, 0.00002685, -0.00001596, -0.00001318, 0.00000094, 0.00003331, -0.00001884, 0.00000838, -0.00000549, 0.00000595, -0.00000889, -0.00001708, 0.00000340, -0.00001487, -0.00001462, -0.00000985, -0.00000658, 0.00001529, -0.00002015, -0.00000703, -0.00002288, -0.00004840, -0.00002034, -0.00002404, -0.00008070, -0.00014924, 0.00007560, 0.00015251, 0.00006997, 0.00003187, 0.00004248, 0.00004477, 0.00005993, 0.00015855, },
		},
		{ { -0.00001773, -0.00001812, 0.00000506, -0.00001612, -0.00001362, -0.00001665, -0.00003874, -0.00003895, -0.00017727, 0.00000512, 0.00025223, 0.00006771, 0.00006203, 0.00005434, 0.00009478, 0.00007951, 0.00018877, 0.00001991, -0.00019261, -0.00004771, -0.00003484, -0.00001839, -0.00002182, -0.00001929, 0.00000821, -0.00001122, -0.00002586, -0.00001077, -0.00000594, 0.00000285, -0.00005778, -0.00000169, -0.00000642, -0.00000169, -0.00005778, 0.00000285, -0.00000594, -0.00001077, -0.00002586, -0.00001122, 0.00000821, -0.00001929, -0.00002182, -0.00001839, -0.00003484, -0.00004771, -0.00019261, 0.00001991, 0.00018877, 0.00007951, 0.00009478, 0.00005434, 0.00006203, 0.00006771, 0.00025223, 0.00000512, -0.00017727, -0.00003895, -0.00003874, -0.00001665, -0.00001362, -0.00001612, 0.00000506, -0.00001812, },
		{ -0.00002041, -0.00000451, 0.00000039, -0.00002105, -0.00000869, -0.00004209, -0.00002237, -0.00013088, -0.00008379, 0.00019960, 0.00013507, 0.00003406, 0.00006178, 0.00006758, 0.00007915, 0.00011245, 0.00015387, -0.00012661, -0.00012043, -0.00002622, -0.00003032, -0.00001666, -0.00002335, -0.00000136, 0.00000160, -0.00002424, -0.00001807, -0.00001219, 0.00001164, -0.00004636, -0.00001989, -0.00000130, 0.00000054, -0.00004469, -0.00002557, 0.00001073, -0.00001983, -0.00000416, -0.00003703, 0.00001735, -0.00001525, -0.00002053, -0.00001963, -0.00002721, -0.00003238, -0.00012471, -0.00016549, 0.00018224, 0.00013290, 0.00010161, 0.00006832, 0.00007394, 0.00004596, 0.00021793, 0.00014817, -0.00020571, -0.00007185, -0.00003882, -0.00002241, -0.00000917, -0.00002113, -0.00000323, -0.00000576, -0.00001943, },
		{ -0.00001634, 0.00000825, -0.00001731, -0.00000748, -0.00003889, -0.00002227, -0.00008546, -0.00011926, 0.00011232, 0.00017656, 0.00004484, 0.00004604, 0.00006252, 0.00006995, 0.00007062, 0.00015361, 0.00002492, -0.00015824, -0.00005347, -0.00002858, -0.00002133, -0.00002209, -0.00000851, 0.00000175, -0.00001244, -0.00002386, -0.00001888, 0.00000984, -0.00002586, -0.00003712, 0.00000201, -0.00000323, -0.00002274, -0.00005231, 0.00001932, -0.00002672, 0.00001123, -0.00004051, 0.00000119, 0.00000082, -0.00002104, -0.00002289, -0.00001626, -0.00003916, -0.00005038, -0.00021384, 0.00000043, 0.00023025, 0.00009807, 0.00010522, 0.00006318, 0.00007179, 0.00013405, 0.00027232, -0.00016262, -0.00013606, -0.00003530, -0.00003471, -0.00000385, -0.00001954, -0.00001518, -0.00000070, -0.00001428, -0.00002030, },
		{ -0.00000085, -0.00000278, -0.00001177, -0.00002785, -0.00003322, -0.00004834, -0.00012407, 0.00003905, 0.00016517, 0.00008646, 0.00002552, 0.00006036, 0.00006273, 0.00005333, 0.00010649, 0.00011389, -0.00008249, -0.00011690, -0.00002434, -0.00002862, -0.00002086, -0.00001394, 0.00000087, -0.00000899, -0.00001594, -0.00002969, 0.00000470, -0.00001038, -0.00004084, -0.00000523, -0.00000177, -0.00001028, -0.00005693, 0.00000348, -0.00001297, -0.00000305, -0.00000749, -0.00003790, 0.00002317, -0.00002784, -0.00001179, -0.00002827, -0.00001786, -0.00005033, -0.00011721, -0.00020870, 0.00020037, 0.00017396, 0.00010858, 0.00008058, 0.00008609, 0.00008483, 0.00030059, -0.00002383, -0.00022153, -0.00002985, -0.00005850, 0.00000609, -0.00002171, -0.00001397, -0.00001440, 0.00000079, -0.00002332, -0.00001842, },
		{ 0.00000157, -0.00000662, -0.00002094, -0.00003889, -0.00003102, -0.00010668, -0.00001620, 0.00013275, 0.00010976, 0.00003700, 0.00003657, 0.00007187, 0.00003886, 0.00007152, 0.00011753, 0.00002702, -0.00011953, -0.00006746, -0.00001412, -0.00003381, -0.00001133, -0.00000394, -0.00000604, -0.00001222, -0.00002604, -0.00001266, 0.00000477, -0.00004038, -0.00001138, -0.00000331, -0.00000185, -0.00004944, -0.00001982, 0.00000337, -0.00002355, 0.00002252, -0.00005270, 0.00001541, -0.00001682, -0.00000956, -0.00003312, -0.00000545, -0.00005528, -0.00004682, -0.00022974, -0.00003411, 0.00028295, 0.00011996, 0.00010550, 0.00009355, 0.00007582, 0.00025709, 0.00014280, -0.00026221, -0.00005748, -0.00006885, -0.00000171, -0.00001411, -0.00001248, -0.00002541, 0.00000260, -0.00001463, -0.00002228, -0.00000870, },
		{ 0.00000252, -0.00001417, -0.00003915, -0.00002753, -0.00008212, -0.00005217, 0.00009607, 0.00011200, 0.00005654, 0.00002596, 0.00005981, 0.00005429, 0.00003028, 0.00010607, 0.00006939, -0.00003242, -0.00012439, -0.00001215, -0.00003663, -0.00001714, -0.00000560, -0.00000173, -0.00001703, -0.00001203, -0.00002993, 0.00001091, -0.00003456, -0.00001641, -0.00000632, 0.00000323, -0.00003795, -0.00003702, 0.00000783, -0.00002944, 0.00002230, -0.00002657, -0.00002096, 0.00000724, -0.00002181, -0.00001473, -0.00002551, -0.00002120, -0.00006377, -0.00010580, -0.00025427, 0.00020605, 0.00024110, 0.00008736, 0.00012910, 0.00007164, 0.00020512, 0.00026135, -0.00021427, -0.00012981, -0.00005453, -0.00003739, 0.00000782, -0.00002051, -0.00001848, -0.00001724, 0.00000488, -0.00002877, -0.00000918, -0.00000609, },
		{ 0.00000326, -0.00004143, -0.00002465, -0.00006619, -0.00006418, 0.00005181, 0.00011514, 0.00005498, 0.00004419, 0.00002698, 0.00008015, 0.00000591, 0.00008074, 0.00007536, 0.00004045, -0.00009737, -0.00006900, -0.00000835, -0.00004191, 0.00000179, -0.00000437, -0.00001369, -0.00001328, -0.00002272, -0.00000635, -0.00001423, -0.00003133, 0.00000140, -0.00000478, -0.00001638, -0.00005487, 0.00001046, -0.00003095, 0.00000975, 0.00000310, -0.00003862, 0.00000604, -0.00001904, -0.00000634, -0.00003389, -0.00000409, -0.00007032, -0.00004171, -0.00023784, -0.00009061, 0.00036037, 0.00012176, 0.00013072, 0.00010779, 0.00013838, 0.00033956, -0.00011388, -0.00020185, -0.00004626, -0.00007179, 0.00001561, -0.00001931, -0.00001128, -0.00003268, 0.00000813, -0.00001975, -0.00001660, -0.00000650, -0.00000338, },
		{ -0.00003335, -0.00002817, -0.00005214, -0.00006731, 0.00001446, 0.00010830, 0.00005738, 0.00004622, 0.00002753, 0.00005994, 0.00003975, 0.00001917, 0.00009476, 0.00004283, -0.00000003, -0.00012445, -0.00000130, -0.00004775, -0.00000612, -0.00000297, -0.00000533, -0.00002556, -0.00000509, -0.00002531, 0.00000127, -0.00004182, 0.00000668, -0.00001213, 0.00000178, -0.00006347, 0.00000517, -0.00002500, -0.00000780, 0.00001937, -0.00002937, -0.00001267, -0.00000963, -0.00001023, -0.00002036, -0.00001842, -0.00003672, -0.00006686, -0.00009436, -0.00030571, 0.00021028, 0.00031484, 0.00006314, 0.00019294, 0.00006461, 0.00038081, 0.00001611, -0.00025247, -0.00005555, -0.00008902, -0.00000304, -0.00000506, -0.00001167, -0.00003116, -0.00000951, -0.00000050, -0.00002632, -0.00000125, -0.00001634, 0.00001603, },
		{ -0.00004234, -0.00003239, -0.00007513, -0.00000261, 0.00007705, 0.00008457, 0.00001603, 0.00005974, 0.00001620, 0.00008219, -0.00001672, 0.00008179, 0.00004456, 0.00005161, -0.00008639, -0.00005595, -0.00001783, -0.00003538, 0.00000613, -0.00000309, -0.00002601, -0.00000828, -0.00001677, -0.00000935, -0.00002983, -0.00000946, 0.00000061, -0.00000281, -0.00004490, -0.00002477, 0.00000334, -0.00004050, 0.00003719, -0.00002591, -0.00000787, -0.00002608, 0.00000138, -0.00002144, -0.00001341, -0.00002992, -0.00006368, -0.00004968, -0.00023180, -0.00017033, 0.00046120, 0.00010026, 0.00019835, 0.00008575, 0.00031805, 0.00021289, -0.00030696, -0.00006228, -0.00010284, -0.00002659, -0.00000326, -0.00000238, -0.00002989, -0.00002164, 0.00000106, -0.00001817, -0.00000816, -0.00001366, 0.00000370, -0.00000511, },
		{ -0.00002256, -0.00007550, -0.00001447, 0.00004212, 0.00010593, 0.00000276, 0.00005866, 0.00002028, 0.00006196, 0.00002305, 0.00001581, 0.00007372, 0.00003939, 0.00000101, -0.00009781, -0.00001223, -0.00004218, -0.00000385, 0.00000347, -0.00001794, -0.00002500, 0.00000175, -0.00002294, -0.00001615, -0.00002848, 0.00001478, -0.00001189, -0.00001855, -0.00005774, 0.00002928, -0.00006209, 0.00003883, -0.00001935, 0.00000328, -0.00003848, 0.00000193, -0.00002147, -0.00000052, -0.00003871, -0.00003421, -0.00008101, -0.00006799, -0.00037818, 0.00023352, 0.00036108, 0.00008953, 0.00020882, 0.00017443, 0.00042935, -0.00031275, -0.00010780, -0.00008854, -0.00006054, -0.00001002, 0.00000638, -0.00002461, -0.00002625, -0.00001081, -0.00000427, -0.00001683, -0.00000366, -0.00001716, 0.00001836, -0.00004483, },
		{ -0.00005993, -0.00003477, 0.00002476, 0.00009331, 0.00003444, 0.00001135, 0.00006178, 0.00001517, 0.00006939, -0.00001748, 0.00005836, 0.00004781, 0.00003143, -0.00005598, -0.00005523, -0.00002089, -0.00002911, 0.00001029, -0.00000885, -0.00003329, 0.00000046, -0.00001266, -0.00002206, -0.00002858, 0.00000649, -0.00000102, -0.00001235, -0.00006080, 0.00001663, -0.00003790, -0.00000332, 0.00001939, -0.00002123, -0.00000194, -0.00004275, 0.00001213, -0.00002518, -0.00000190, -0.00005829, -0.00005046, -0.00005816, -0.00023013, -0.00024439, 0.00054096, 0.00011426, 0.00024625, 0.00012101, 0.00050641, -0.00014449, -0.00026593, -0.00001060, -0.00012473, 0.00000198, -0.00001649, 0.00000662, -0.00004412, -0.00000746, -0.00001365, -0.00000531, -0.00000796, -0.00001863, 0.00001136, -0.00002101, -0.00003094, },
		{ -0.00005184, 0.00001618, 0.00006191, 0.00007685, -0.00002404, 0.00006862, 0.00001728, 0.00005161, 0.00002341, -0.00000373, 0.00007152, 0.00002789, -0.00000193, -0.00006336, -0.00003412, -0.00002401, -0.00000979, 0.00001069, -0.00003617, -0.00001147, 0.00000189, -0.00002507, -0.00002801, -0.00000705, 0.00001169, -0.00001111, -0.00004756, -0.00001701, 0.00000542, -0.00005392, 0.00005626, -0.00004921, 0.00003857, -0.00007534, 0.00002781, -0.00004502, 0.00003607, -0.00008573, -0.00000635, -0.00011137, -0.00003054, -0.00045846, 0.00025380, 0.00040211, 0.00015534, 0.00019470, 0.00042270, 0.00017089, -0.00045797, 0.00004512, -0.00015830, 0.00000049, -0.00004512, 0.00003456, -0.00005421, -0.00000186, -0.00002966, 0.00000594, -0.00001165, -0.00001150, -0.00000782, 0.00000315, -0.00003458, -0.00004551, },
		{ 0.00000651, 0.00003310, 0.00009567, -0.00001495, 0.00002984, 0.00005082, 0.00001769, 0.00005403, -0.00001195, 0.00002776, 0.00006986, -0.00001363, -0.00000174, -0.00007816, -0.00000326, -0.00003626, 0.00002262, -0.00003008, -0.00002114, 0.00000039, -0.00000903, -0.00003864, -0.00001017, 0.00000707, 0.00000394, -0.00004151, -0.00003711, 0.00002183, -0.00006144, 0.00004163, -0.00002701, 0.00001821, -0.00003290, -0.00002446, -0.00000816, 0.00000605, -0.00003098, -0.00005338, -0.00006016, -0.00004602, -0.00025068, -0.00029249, 0.00056517, 0.00022060, 0.00020612, 0.00034982, 0.00043519, -0.00048178, -0.00006602, -0.00007948, -0.00006651, -0.00002670, 0.00000768, -0.00001813, -0.00002758, -0.00001542, -0.00001556, 0.00000780, -0.00001720, -0.00001154, 0.00000252, -0.00001935, -0.00004063, -0.00005861, },
		{ 0.00000986, 0.00009085, 0.00001671, -0.00000558, 0.00005361, 0.00003024, 0.00002304, 0.00004590, -0.00004615, 0.00010088, -0.00001522, 0.00002319, -0.00005426, -0.00003726, -0.00001699, -0.00000638, 0.00000055, -0.00003830, -0.00000033, -0.00000101, -0.00003275, -0.00002404, 0.00000535, 0.00001070, -0.00002601, -0.00005269, 0.00001923, -0.00004104, 0.00000076, 0.00001649, -0.00002794, 0.00002968, -0.00007552, 0.00002281, -0.00003517, 0.00003851, -0.00010983, 0.00000663, -0.00012317, -0.00002045, -0.00049720, 0.00020133, 0.00052943, 0.00015562, 0.00031797, 0.00058021, -0.00029453, -0.00030788, 0.00003835, -0.00014058, 0.00000545, -0.00004451, 0.00002990, -0.00005721, 0.00000768, -0.00004390, 0.00002231, -0.00001617, -0.00001268, -0.00000472, -0.00000882, -0.00002695, -0.00006687, 0.00000094, },
		{ 0.00006445, 0.00005264, -0.00001995, 0.00003151, 0.00005205, 0.00000216, 0.00005996, -0.00002807, 0.00002508, 0.00006739, -0.00003875, 0.00003838, -0.00009602, 0.00001463, -0.00004170, 0.00002639, -0.00004911, 0.00000234, -0.00001028, -0.00000636, -0.00004831, 0.00001170, -0.00000014, 0.00000499, -0.00007050, 0.00001461, -0.00002446, -0.00002297, 0.00003150, -0.00003829, 0.00003645, -0.00005270, -0.00001178, -0.00002107, 0.00003313, -0.00006860, -0.00003618, -0.00007365, -0.00004025, -0.00025431, -0.00036931, 0.00059728, 0.00035526, 0.00017424, 0.00069855, -0.00002524, -0.00049701, 0.00003838, -0.00012404, -0.00001608, -0.00005496, 0.00002641, -0.00004167, -0.00000351, -0.00003567, -0.00000230, 0.00001136, -0.00002664, 0.00000042, -0.00001718, -0.00000313, -0.00007148, -0.00000867, 0.00000113, },
		{ 0.00008941, -0.00002801, 0.00002426, 0.00003223, 0.00003959, 0.00000089, 0.00006135, -0.00007728, 0.00012530, -0.00006002, 0.00005906, -0.00006379, -0.00002221, -0.00002217, 0.00000412, -0.00002560, -0.00001540, -0.00000583, -0.00000022, -0.00004272, -0.00000637, 0.00000363, 0.00001445, -0.00005721, -0.00001509, 0.00000415, -0.00004545, 0.00003748, -0.00003326, 0.00001684, -0.00000946, -0.00004349, -0.00001203, 0.00001014, -0.00000614, -0.00008817, -0.00001612, -0.00010394, -0.00004321, -0.00049234, 0.00006492, 0.00077084, 0.00005969, 0.00069595, 0.00032224, -0.00058458, -0.00007304, -0.00006677, -0.00004989, -0.00004578, -0.00000550, -0.00001105, -0.00002327, -0.00001472, -0.00003347, 0.00002833, -0.00002401, -0.00000488, -0.00001680, -0.00000128, -0.00004682, -0.00004066, 0.00001836, 0.00001687, },
		{ -0.00001815, 0.00001942, 0.00000610, 0.00006819, -0.00002715, 0.00007983, -0.00005715, 0.00004507, 0.00003425, -0.00002512, 0.00003943, -0.00008376, 0.00001134, -0.00002908, 0.00000347, -0.00003972, 0.00001231, -0.00001901, -0.00000879, -0.00004266, 0.00002215, -0.00000319, -0.00001229, -0.00006501, 0.00004259, -0.00007283, 0.00004903, -0.00003585, 0.00001082, -0.00000319, -0.00001894, -0.00004843, 0.00002097, -0.00000186, -0.00005181, -0.00006339, -0.00004097, -0.00009124, -0.00018205, -0.00055575, 0.00074406, 0.00038748, 0.00038733, 0.00079335, -0.00059877, -0.00020301, -0.00006619, -0.00003759, -0.00005834, -0.00001733, -0.00001370, -0.00001370, -0.00001763, -0.00003168, 0.00000406, 0.00001056, -0.00003014, 0.00000352, -0.00002610, -0.00000281, -0.00007684, 0.00003780, -0.00002187, 0.00010200, },
		{ -0.00001366, 0.00002192, 0.00003347, 0.00002344, 0.00000640, 0.00004679, -0.00006777, 0.00010439, -0.00005892, 0.00005995, -0.00005314, -0.00002003, -0.00001411, -0.00000580, -0.00003629, 0.00000934, -0.00001950, 0.00000081, -0.00004480, 0.00001047, -0.00000193, 0.00000886, -0.00007618, 0.00003593, -0.00005475, 0.00001785, -0.00000032, -0.00002257, 0.00001472, -0.00001172, -0.00004251, -0.00001231, 0.00002105, -0.00003061, -0.00007750, -0.00002237, -0.00011187, -0.00002641, -0.00053932, -0.00003953, 0.00098866, 0.00008313, 0.00109312, -0.00035035, -0.00043077, -0.00004651, -0.00007023, -0.00002362, -0.00005393, -0.00000604, -0.00002127, -0.00000579, -0.00003287, -0.00001573, 0.00002533, -0.00002951, 0.00000314, -0.00003158, 0.00000922, -0.00007092, 0.00001831, -0.00001429, 0.00006368, 0.00002941, },
		{ 0.00004893, -0.00001093, 0.00006796, -0.00003657, 0.00007655, -0.00005586, 0.00003745, 0.00002008, -0.00001288, 0.00003013, -0.00006153, 0.00000314, -0.00001719, -0.00002762, -0.00000765, 0.00000167, -0.00001951, -0.00001358, -0.00002643, 0.00002069, -0.00000497, -0.00003879, -0.00002012, 0.00001053, -0.00005690, 0.00007231, -0.00008406, 0.00005987, -0.00005315, 0.00002976, -0.00009849, 0.00008313, -0.00007419, 0.00000762, -0.00012527, 0.00002038, -0.00016988, -0.00007551, -0.00080773, 0.00096047, 0.00038401, 0.00087401, 0.00031223, -0.00079891, 0.00004883, -0.00020265, 0.00008118, -0.00013054, 0.00003978, -0.00007514, 0.00004202, -0.00006155, 0.00000119, -0.00000727, 0.00001262, -0.00002970, -0.00000325, -0.00002032, -0.00002582, -0.00003033, 0.00002445, -0.00000020, 0.00008554, -0.00004996, },
		{ -0.00000198, 0.00004421, 0.00000080, 0.00001663, 0.00002956, -0.00004671, 0.00005972, -0.00001980, 0.00002256, -0.00001395, -0.00003869, 0.00001144, -0.00004325, -0.00001162, 0.00000991, -0.00002389, -0.00000454, -0.00003239, 0.00001479, -0.00000276, -0.00001515, -0.00005279, 0.00004323, -0.00008790, 0.00008760, -0.00007842, 0.00004173, -0.00004395, 0.00004201, -0.00009578, 0.00005066, -0.00004488, 0.00001893, -0.00013028, 0.00003372, -0.00018928, 0.00007562, -0.00070205, -0.00002482, 0.00106436, 0.00050408, 0.00098984, -0.00094752, -0.00003034, -0.00022817, 0.00006934, -0.00010548, 0.00001808, -0.00007827, 0.00003500, -0.00004418, -0.00000852, -0.00002005, 0.00002708, -0.00003383, 0.00000438, -0.00003392, -0.00000254, -0.00005182, 0.00004032, -0.00003307, 0.00009705, -0.00004068, 0.00003269, },
		{ 0.00000180, 0.00003864, -0.00001659, 0.00004256, -0.00001388, -0.00001942, 0.00005940, -0.00004221, 0.00004846, -0.00006130, 0.00001676, -0.00003598, -0.00002849, 0.00001227, -0.00001028, -0.00002040, -0.00001009, -0.00001413, 0.00001820, -0.00002011, -0.00004090, 0.00001888, -0.00004752, 0.00001919, 0.00001286, -0.00005298, 0.00003928, -0.00004084, 0.00001987, -0.00008324, 0.00007519, -0.00008863, 0.00002068, -0.00013749, 0.00002141, -0.00017167, -0.00006902, -0.00095687, 0.00102493, 0.00071276, 0.00117283, -0.00052685, -0.00047527, -0.00001019, -0.00015645, 0.00007178, -0.00010990, 0.00001634, -0.00006819, 0.00004717, -0.00007315, 0.00002191, -0.00001733, 0.00001232, -0.00002804, -0.00000893, -0.00002097, -0.00003402, 0.00002210, -0.00002292, 0.00005936, 0.00000994, -0.00001607, 0.00003550, },
		{ 0.00002756, 0.00000774, -0.00000519, 0.00005210, -0.00006768, 0.00005200, 0.00000016, -0.00000635, 0.00001718, -0.00004701, 0.00002072, -0.00007223, 0.00002455, -0.00000792, -0.00001799, -0.00001216, -0.00001233, 0.00000932, -0.00000575, -0.00003943, 0.00000124, -0.00001077, -0.00004001, 0.00007682, -0.00010010, 0.00007001, -0.00007771, 0.00007568, -0.00012646, 0.00009570, -0.00011111, 0.00007760, -0.00018524, 0.00008258, -0.00024916, 0.00015093, -0.00084649, -0.00005897, 0.00123245, 0.00117701, 0.00014887, -0.00093400, 0.00014596, -0.00032427, 0.00015923, -0.00015732, 0.00006164, -0.00013034, 0.00008529, -0.00008836, 0.00003089, -0.00003647, 0.00002947, -0.00003385, -0.00000150, -0.00002760, -0.00002694, 0.00000966, -0.00000495, 0.00001329, 0.00005568, -0.00004762, 0.00004627, -0.00000226, },
		{ 0.00001483, -0.00000502, 0.00002214, 0.00001709, -0.00007526, 0.00011437, -0.00009031, 0.00008097, -0.00007183, 0.00003803, -0.00007377, 0.00000486, 0.00000216, -0.00000716, -0.00002863, 0.00000622, -0.00001394, 0.00001899, -0.00004987, 0.00000151, -0.00000147, -0.00005093, 0.00006575, -0.00005532, 0.00000864, -0.00001988, 0.00001312, -0.00003219, -0.00001512, -0.00000402, -0.00002867, -0.00003231, -0.00008677, -0.00004252, -0.00009756, -0.00016887, -0.00098549, 0.00088588, 0.00164632, 0.00064003, -0.00100477, -0.00005332, -0.00021417, 0.00000092, -0.00002800, -0.00002192, -0.00006868, 0.00000108, -0.00000641, -0.00003374, 0.00000991, -0.00001447, 0.00000765, -0.00002837, -0.00000583, -0.00004183, 0.00000756, 0.00001224, -0.00002305, 0.00007650, -0.00004770, 0.00002805, 0.00001340, 0.00001301, },
		{ 0.00003675, -0.00004618, 0.00009947, -0.00012474, 0.00009166, -0.00002174, -0.00000191, 0.00001956, -0.00002889, -0.00000643, -0.00005961, 0.00003532, -0.00001907, -0.00001376, -0.00001235, 0.00000347, 0.00000188, -0.00002083, -0.00003261, 0.00003204, -0.00006885, 0.00005086, -0.00001397, -0.00003162, 0.00001053, -0.00002564, 0.00002289, -0.00005945, 0.00003043, -0.00007132, 0.00003800, -0.00013713, 0.00001460, -0.00016561, 0.00003820, -0.00076136, -0.00044881, 0.00203592, 0.00136276, -0.00091678, -0.00036314, -0.00007345, -0.00015962, 0.00005510, -0.00006255, -0.00002165, -0.00006958, 0.00004194, -0.00006057, 0.00002518, -0.00002805, 0.00001309, -0.00002072, -0.00001306, -0.00002723, -0.00002506, 0.00005025, -0.00005739, 0.00008046, -0.00003228, 0.00000683, 0.00001519, 0.00002623, -0.00001356, },
		{ -0.00003040, 0.00004667, -0.00000667, -0.00007323, 0.00012543, -0.00010579, 0.00008646, -0.00005750, 0.00002784, -0.00008517, 0.00003492, -0.00001988, 0.00000603, -0.00003961, 0.00002940, -0.00002414, 0.00001740, -0.00007268, 0.00006324, -0.00007737, 0.00003495, 0.00000488, -0.00002818, -0.00000854, -0.00000849, -0.00000114, -0.00000986, -0.00002343, -0.00001991, -0.00002380, -0.00002273, -0.00011330, -0.00000568, -0.00015468, -0.00011704, -0.00123361, 0.00101199, 0.00278414, -0.00079254, -0.00060614, -0.00008255, -0.00016987, -0.00001285, -0.00000321, -0.00004124, -0.00006368, 0.00000363, -0.00001568, -0.00001708, 0.00001137, -0.00002892, 0.00002055, -0.00004342, 0.00000695, -0.00006680, 0.00007577, -0.00005843, 0.00004928, 0.00000609, -0.00002272, 0.00001908, 0.00002859, -0.00001265, 0.00002802, },
		{ -0.00005403, 0.00010821, -0.00014936, 0.00011853, -0.00004088, 0.00000246, 0.00002327, -0.00001744, -0.00003726, -0.00002257, 0.00001805, -0.00001396, -0.00001603, -0.00000278, 0.00000969, -0.00000676, -0.00004112, 0.00001157, -0.00000120, -0.00005187, 0.00007794, -0.00007649, 0.00003519, -0.00005891, 0.00004226, -0.00004734, 0.00003497, -0.00008367, 0.00002781, -0.00005811, -0.00003514, -0.00010597, -0.00002109, -0.00016976, -0.00054690, -0.00112778, 0.00369447, 0.00032689, -0.00126582, 0.00009040, -0.00032132, 0.00004416, -0.00008970, 0.00006070, -0.00013441, 0.00002975, -0.00004580, 0.00001348, -0.00001579, -0.00000256, -0.00001158, -0.00000406, -0.00002337, -0.00003704, 0.00001964, 0.00001858, -0.00003822, 0.00007766, -0.00007212, 0.00004555, -0.00000664, 0.00003754, -0.00003346, 0.00004159, },
		{ 0.00007740, -0.00004975, -0.00003357, 0.00009742, -0.00008808, 0.00006656, -0.00002376, -0.00001219, -0.00005572, 0.00003538, -0.00003221, 0.00001965, -0.00004873, 0.00005451, -0.00004520, 0.00001522, -0.00006966, 0.00009475, -0.00014051, 0.00014437, -0.00012630, 0.00009307, -0.00012660, 0.00010909, -0.00012777, 0.00014150, -0.00018391, 0.00012530, -0.00018154, 0.00015993, -0.00033374, 0.00026597, -0.00051809, 0.00033023, -0.00209052, 0.00212950, 0.00308487, -0.00227013, 0.00034966, -0.00057628, 0.00020849, -0.00028688, 0.00022553, -0.00023854, 0.00010361, -0.00013689, 0.00009616, -0.00009136, 0.00007555, -0.00009278, 0.00007309, -0.00008643, 0.00003098, -0.00007162, 0.00010979, -0.00011290, 0.00012134, -0.00009225, 0.00005256, -0.00003032, 0.00006751, -0.00005577, 0.00005446, -0.00006139, },
		{ 0.00005752, -0.00010526, 0.00008802, -0.00001687, -0.00002543, 0.00005554, -0.00004392, -0.00003004, -0.00000184, -0.00000324, 0.00000027, -0.00002351, 0.00001660, 0.00000058, -0.00001286, -0.00005164, 0.00006396, -0.00007837, 0.00005146, -0.00001759, -0.00001237, -0.00001646, -0.00001220, -0.00000850, 0.00001326, -0.00001393, -0.00005770, -0.00000889, -0.00000861, -0.00009982, -0.00002267, -0.00013272, -0.00007075, -0.00074221, -0.00135950, 0.00550896, -0.00177324, -0.00046139, -0.00015679, -0.00013183, -0.00007016, 0.00000843, -0.00002326, -0.00006844, -0.00001750, -0.00001843, 0.00000570, 0.00000193, -0.00002943, 0.00001001, -0.00002382, -0.00001049, -0.00005200, 0.00007738, -0.00005956, 0.00004868, -0.00001483, -0.00001847, 0.00001639, 0.00001985, 0.00000401, -0.00001083, 0.00000791, -0.00001702, },
		{ -0.00004039, -0.00002663, 0.00008821, -0.00008420, 0.00006422, -0.00001676, -0.00003278, -0.00002261, 0.00002102, -0.00003517, 0.00003264, -0.00005555, 0.00007763, -0.00007757, 0.00002528, -0.00004935, 0.00007486, -0.00012685, 0.00017044, -0.00020001, 0.00018349, -0.00022684, 0.00021606, -0.00024508, 0.00029081, -0.00036504, 0.00030927, -0.00038251, 0.00038643, -0.00059703, 0.00058970, -0.00097491, 0.00093227, -0.00334734, 0.00437413, 0.00176210, -0.00266147, 0.00095066, -0.00096010, 0.00053573, -0.00053735, 0.00043295, -0.00042374, 0.00027527, -0.00029626, 0.00023968, -0.00019694, 0.00015789, -0.00016803, 0.00013588, -0.00013954, 0.00006271, -0.00005542, 0.00008253, -0.00009084, 0.00010970, -0.00012154, 0.00009120, -0.00005549, 0.00007725, -0.00006740, 0.00004843, -0.00004942, 0.00005377, },
		{ -0.00004277, 0.00003497, 0.00002332, -0.00005855, 0.00008570, -0.00007918, 0.00000222, -0.00000250, -0.00001995, 0.00002314, -0.00004157, 0.00005226, -0.00003735, 0.00000355, -0.00005653, 0.00009493, -0.00013855, 0.00016426, -0.00017177, 0.00014503, -0.00017726, 0.00015637, -0.00019300, 0.00022961, -0.00025264, 0.00018224, -0.00027351, 0.00028859, -0.00044330, 0.00039600, -0.00067838, 0.00060006, -0.00180427, -0.00025785, 0.00616285, -0.00375326, 0.00089903, -0.00093382, 0.00046452, -0.00051323, 0.00037179, -0.00033942, 0.00021058, -0.00025882, 0.00018352, -0.00014689, 0.00013103, -0.00014845, 0.00011099, -0.00010891, 0.00004867, -0.00007460, 0.00010954, -0.00011305, 0.00012085, -0.00011824, 0.00007310, -0.00004784, 0.00006618, -0.00004273, 0.00001979, -0.00002161, 0.00000982, 0.00000734, },
		{ -0.00007157, 0.00011735, -0.00010767, 0.00008771, -0.00004752, -0.00000977, -0.00001643, 0.00001157, -0.00003127, 0.00003775, -0.00004851, 0.00007901, -0.00009959, 0.00004450, -0.00003148, 0.00002855, -0.00004360, 0.00007757, -0.00013402, 0.00013181, -0.00018001, 0.00016769, -0.00018841, 0.00024246, -0.00034008, 0.00028499, -0.00036160, 0.00036433, -0.00056464, 0.00054875, -0.00099226, 0.00097032, -0.00402546, 0.00676631, -0.00079296, -0.00164234, 0.00061972, -0.00072817, 0.00040363, -0.00043017, 0.00033724, -0.00034514, 0.00022117, -0.00026616, 0.00023522, -0.00018999, 0.00014417, -0.00015965, 0.00013130, -0.00014590, 0.00007802, -0.00003400, 0.00002689, -0.00001677, 0.00001934, -0.00005913, 0.00005694, -0.00002824, 0.00004057, -0.00003797, 0.00001074, -0.00000464, -0.00001632, 0.00002789, },
		{ 0.00001781, 0.00003137, -0.00005362, 0.00006723, -0.00006917, 0.00001103, 0.00000342, -0.00004126, 0.00005684, -0.00007016, 0.00009902, -0.00010202, 0.00005672, -0.00007635, 0.00010574, -0.00014408, 0.00019833, -0.00025880, 0.00025772, -0.00030562, 0.00029970, -0.00035242, 0.00042442, -0.00049182, 0.00044646, -0.00057429, 0.00062882, -0.00083390, 0.00087554, -0.00133601, 0.00147384, -0.00338847, 0.00221249, 0.00486751, -0.00435823, 0.00172212, -0.00149562, 0.00094170, -0.00088596, 0.00067904, -0.00061322, 0.00046602, -0.00049547, 0.00040251, -0.00032697, 0.00028094, -0.00028447, 0.00023196, -0.00021929, 0.00013962, -0.00011053, 0.00011219, -0.00010863, 0.00011651, -0.00014841, 0.00011926, -0.00008256, 0.00008122, -0.00005813, 0.00002092, -0.00001185, -0.00001835, 0.00001806, -0.00002832, },
		{ 0.00013848, -0.00011303, 0.00008635, -0.00006210, 0.00001519, -0.00001607, -0.00000363, -0.00000693, 0.00001665, -0.00000777, 0.00003034, -0.00007034, 0.00003908, -0.00001292, -0.00000985, 0.00004288, -0.00006191, 0.00001873, -0.00002915, -0.00000564, -0.00002413, 0.00003685, -0.00001618, -0.00006790, -0.00001283, -0.00003667, -0.00000289, -0.00011938, -0.00002252, -0.00030720, -0.00003972, -0.00314824, 0.00780546, -0.00314824, -0.00003972, -0.00030720, -0.00002252, -0.00011938, -0.00000289, -0.00003667, -0.00001283, -0.00006790, -0.00001618, 0.00003685, -0.00002413, -0.00000564, -0.00002915, 0.00001873, -0.00006191, 0.00004288, -0.00000985, -0.00001292, 0.00003908, -0.00007034, 0.00003034, -0.00000777, 0.00001665, -0.00000693, -0.00000363, -0.00001607, 0.00001519, -0.00006210, 0.00008635, -0.00011303, },
		{ 0.00001781, -0.00002832, 0.00001806, -0.00001835, -0.00001185, 0.00002092, -0.00005813, 0.00008122, -0.00008256, 0.00011926, -0.00014841, 0.00011651, -0.00010863, 0.00011219, -0.00011053, 0.00013962, -0.00021929, 0.00023196, -0.00028447, 0.00028094, -0.00032697, 0.00040251, -0.00049547, 0.00046602, -0.00061322, 0.00067904, -0.00088596, 0.00094170, -0.00149562, 0.00172212, -0.00435823, 0.00486751, 0.00221249, -0.00338847, 0.00147384, -0.00133601, 0.00087554, -0.00083390, 0.00062882, -0.00057429, 0.00044646, -0.00049182, 0.00042442, -0.00035242, 0.00029970, -0.00030562, 0.00025772, -0.00025880, 0.00019833, -0.00014408, 0.00010574, -0.00007635, 0.00005672, -0.00010202, 0.00009902, -0.00007016, 0.00005684, -0.00004126, 0.00000342, 0.00001103, -0.00006917, 0.00006723, -0.00005362, 0.00003137, },
		{ -0.00007157, 0.00002789, -0.00001632, -0.00000464, 0.00001074, -0.00003797, 0.00004057, -0.00002824, 0.00005694, -0.00005913, 0.00001934, -0.00001677, 0.00002689, -0.00003400, 0.00007802, -0.00014590, 0.00013130, -0.00015965, 0.00014417, -0.00018999, 0.00023522, -0.00026616, 0.00022117, -0.00034514, 0.00033724, -0.00043017, 0.00040363, -0.00072817, 0.00061972, -0.00164234, -0.00079296, 0.00676631, -0.00402546, 0.00097032, -0.00099226, 0.00054875, -0.00056464, 0.00036433, -0.00036160, 0.00028499, -0.00034008, 0.00024246, -0.00018841, 0.00016769, -0.00018001, 0.00013181, -0.00013402, 0.00007757, -0.00004360, 0.00002855, -0.00003148, 0.00004450, -0.00009959, 0.00007901, -0.00004851, 0.00003775, -0.00003127, 0.00001157, -0.00001643, -0.00000977, -0.00004752, 0.00008771, -0.00010767, 0.00011735, },
		{ -0.00004277, 0.00000734, 0.00000982, -0.00002161, 0.00001979, -0.00004273, 0.00006618, -0.00004784, 0.00007310, -0.00011824, 0.00012085, -0.00011305, 0.00010954, -0.00007460, 0.00004867, -0.00010891, 0.00011099, -0.00014845, 0.00013103, -0.00014689, 0.00018352, -0.00025882, 0.00021058, -0.00033942, 0.00037179, -0.00051323, 0.00046452, -0.00093382, 0.00089903, -0.00375326, 0.00616285, -0.00025785, -0.00180427, 0.00060006, -0.00067838, 0.00039600, -0.00044330, 0.00028859, -0.00027351, 0.00018224, -0.00025264, 0.00022961, -0.00019300, 0.00015637, -0.00017726, 0.00014503, -0.00017177, 0.00016426, -0.00013855, 0.00009493, -0.00005653, 0.00000355, -0.00003735, 0.00005226, -0.00004157, 0.00002314, -0.00001995, -0.00000250, 0.00000222, -0.00007918, 0.00008570, -0.00005855, 0.00002332, 0.00003497, },
		{ -0.00004039, 0.00005377, -0.00004942, 0.00004843, -0.00006740, 0.00007725, -0.00005549, 0.00009120, -0.00012154, 0.00010970, -0.00009084, 0.00008253, -0.00005542, 0.00006271, -0.00013954, 0.00013588, -0.00016803, 0.00015789, -0.00019694, 0.00023968, -0.00029626, 0.00027527, -0.00042374, 0.00043295, -0.00053735, 0.00053573, -0.00096010, 0.00095066, -0.00266147, 0.00176210, 0.00437413, -0.00334734, 0.00093227, -0.00097491, 0.00058970, -0.00059703, 0.00038643, -0.00038251, 0.00030927, -0.00036504, 0.00029081, -0.00024508, 0.00021606, -0.00022684, 0.00018349, -0.00020001, 0.00017044, -0.00012685, 0.00007486, -0.00004935, 0.00002528, -0.00007757, 0.00007763, -0.00005555, 0.00003264, -0.00003517, 0.00002102, -0.00002261, -0.00003278, -0.00001676, 0.00006422, -0.00008420, 0.00008821, -0.00002663, },
		{ 0.00005752, -0.00001702, 0.00000791, -0.00001083, 0.00000401, 0.00001985, 0.00001639, -0.00001847, -0.00001483, 0.00004868, -0.00005956, 0.00007738, -0.00005200, -0.00001049, -0.00002382, 0.00001001, -0.00002943, 0.00000193, 0.00000570, -0.00001843, -0.00001750, -0.00006844, -0.00002326, 0.00000843, -0.00007016, -0.00013183, -0.00015679, -0.00046139, -0.00177324, 0.00550896, -0.00135950, -0.00074221, -0.00007075, -0.00013272, -0.00002267, -0.00009982, -0.00000861, -0.00000889, -0.00005770, -0.00001393, 0.00001326, -0.00000850, -0.00001220, -0.00001646, -0.00001237, -0.00001759, 0.00005146, -0.00007837, 0.00006396, -0.00005164, -0.00001286, 0.00000058, 0.00001660, -0.00002351, 0.00000027, -0.00000324, -0.00000184, -0.00003004, -0.00004392, 0.00005554, -0.00002543, -0.00001687, 0.00008802, -0.00010526, },
		{ 0.00007740, -0.00006139, 0.00005446, -0.00005577, 0.00006751, -0.00003032, 0.00005256, -0.00009225, 0.00012134, -0.00011290, 0.00010979, -0.00007162, 0.00003098, -0.00008643, 0.00007309, -0.00009278, 0.00007555, -0.00009136, 0.00009616, -0.00013689, 0.00010361, -0.00023854, 0.00022553, -0.00028688, 0.00020849, -0.00057628, 0.00034966, -0.00227013, 0.00308487, 0.00212950, -0.00209052, 0.00033023, -0.00051809, 0.00026597, -0.00033374, 0.00015993, -0.00018154, 0.00012530, -0.00018391, 0.00014150, -0.00012777, 0.00010909, -0.00012660, 0.00009307, -0.00012630, 0.00014437, -0.00014051, 0.00009475, -0.00006966, 0.00001522, -0.00004520, 0.00005451, -0.00004873, 0.00001965, -0.00003221, 0.00003538, -0.00005572, -0.00001219, -0.00002376, 0.00006656, -0.00008808, 0.00009742, -0.00003357, -0.00004975, },
		{ -0.00005403, 0.00004159, -0.00003346, 0.00003754, -0.00000664, 0.00004555, -0.00007212, 0.00007766, -0.00003822, 0.00001858, 0.00001964, -0.00003704, -0.00002337, -0.00000406, -0.00001158, -0.00000256, -0.00001579, 0.00001348, -0.00004580, 0.00002975, -0.00013441, 0.00006070, -0.00008970, 0.00004416, -0.00032132, 0.00009040, -0.00126582, 0.00032689, 0.00369447, -0.00112778, -0.00054690, -0.00016976, -0.00002109, -0.00010597, -0.00003514, -0.00005811, 0.00002781, -0.00008367, 0.00003497, -0.00004734, 0.00004226, -0.00005891, 0.00003519, -0.00007649, 0.00007794, -0.00005187, -0.00000120, 0.00001157, -0.00004112, -0.00000676, 0.00000969, -0.00000278, -0.00001603, -0.00001396, 0.00001805, -0.00002257, -0.00003726, -0.00001744, 0.00002327, 0.00000246, -0.00004088, 0.00011853, -0.00014936, 0.00010821, },
		{ -0.00003040, 0.00002802, -0.00001265, 0.00002859, 0.00001908, -0.00002272, 0.00000609, 0.00004928, -0.00005843, 0.00007577, -0.00006680, 0.00000695, -0.00004342, 0.00002055, -0.00002892, 0.00001137, -0.00001708, -0.00001568, 0.00000363, -0.00006368, -0.00004124, -0.00000321, -0.00001285, -0.00016987, -0.00008255, -0.00060614, -0.00079254, 0.00278414, 0.00101199, -0.00123361, -0.00011704, -0.00015468, -0.00000568, -0.00011330, -0.00002273, -0.00002380, -0.00001991, -0.00002343, -0.00000986, -0.00000114, -0.00000849, -0.00000854, -0.00002818, 0.00000488, 0.00003495, -0.00007737, 0.00006324, -0.00007268, 0.00001740, -0.00002414, 0.00002940, -0.00003961, 0.00000603, -0.00001988, 0.00003492, -0.00008517, 0.00002784, -0.00005750, 0.00008646, -0.00010579, 0.00012543, -0.00007323, -0.00000667, 0.00004667, },
		{ 0.00003675, -0.00001356, 0.00002623, 0.00001519, 0.00000683, -0.00003228, 0.00008046, -0.00005739, 0.00005025, -0.00002506, -0.00002723, -0.00001306, -0.00002072, 0.00001309, -0.00002805, 0.00002518, -0.00006057, 0.00004194, -0.00006958, -0.00002165, -0.00006255, 0.00005510, -0.00015962, -0.00007345, -0.00036314, -0.00091678, 0.00136276, 0.00203592, -0.00044881, -0.00076136, 0.00003820, -0.00016561, 0.00001460, -0.00013713, 0.00003800, -0.00007132, 0.00003043, -0.00005945, 0.00002289, -0.00002564, 0.00001053, -0.00003162, -0.00001397, 0.00005086, -0.00006885, 0.00003204, -0.00003261, -0.00002083, 0.00000188, 0.00000347, -0.00001235, -0.00001376, -0.00001907, 0.00003532, -0.00005961, -0.00000643, -0.00002889, 0.00001956, -0.00000191, -0.00002174, 0.00009166, -0.00012474, 0.00009947, -0.00004618, },
		{ 0.00001483, 0.00001301, 0.00001340, 0.00002805, -0.00004770, 0.00007650, -0.00002305, 0.00001224, 0.00000756, -0.00004183, -0.00000583, -0.00002837, 0.00000765, -0.00001447, 0.00000991, -0.00003374, -0.00000641, 0.00000108, -0.00006868, -0.00002192, -0.00002800, 0.00000092, -0.00021417, -0.00005332, -0.00100477, 0.00064003, 0.00164632, 0.00088588, -0.00098549, -0.00016887, -0.00009756, -0.00004252, -0.00008677, -0.00003231, -0.00002867, -0.00000402, -0.00001512, -0.00003219, 0.00001312, -0.00001988, 0.00000864, -0.00005532, 0.00006575, -0.00005093, -0.00000147, 0.00000151, -0.00004987, 0.00001899, -0.00001394, 0.00000622, -0.00002863, -0.00000716, 0.00000216, 0.00000486, -0.00007377, 0.00003803, -0.00007183, 0.00008097, -0.00009031, 0.00011437, -0.00007526, 0.00001709, 0.00002214, -0.00000502, },
		{ 0.00002756, -0.00000226, 0.00004627, -0.00004762, 0.00005568, 0.00001329, -0.00000495, 0.00000966, -0.00002694, -0.00002760, -0.00000150, -0.00003385, 0.00002947, -0.00003647, 0.00003089, -0.00008836, 0.00008529, -0.00013034, 0.00006164, -0.00015732, 0.00015923, -0.00032427, 0.00014596, -0.00093400, 0.00014887, 0.00117701, 0.00123245, -0.00005897, -0.00084649, 0.00015093, -0.00024916, 0.00008258, -0.00018524, 0.00007760, -0.00011111, 0.00009570, -0.00012646, 0.00007568, -0.00007771, 0.00007001, -0.00010010, 0.00007682, -0.00004001, -0.00001077, 0.00000124, -0.00003943, -0.00000575, 0.00000932, -0.00001233, -0.00001216, -0.00001799, -0.00000792, 0.00002455, -0.00007223, 0.00002072, -0.00004701, 0.00001718, -0.00000635, 0.00000016, 0.00005200, -0.00006768, 0.00005210, -0.00000519, 0.00000774, },
		{ 0.00000180, 0.00003550, -0.00001607, 0.00000994, 0.00005936, -0.00002292, 0.00002210, -0.00003402, -0.00002097, -0.00000893, -0.00002804, 0.00001232, -0.00001733, 0.00002191, -0.00007315, 0.00004717, -0.00006819, 0.00001634, -0.00010990, 0.00007178, -0.00015645, -0.00001019, -0.00047527, -0.00052685, 0.00117283, 0.00071276, 0.00102493, -0.00095687, -0.00006902, -0.00017167, 0.00002141, -0.00013749, 0.00002068, -0.00008863, 0.00007519, -0.00008324, 0.00001987, -0.00004084, 0.00003928, -0.00005298, 0.00001286, 0.00001919, -0.00004752, 0.00001888, -0.00004090, -0.00002011, 0.00001820, -0.00001413, -0.00001009, -0.00002040, -0.00001028, 0.00001227, -0.00002849, -0.00003598, 0.00001676, -0.00006130, 0.00004846, -0.00004221, 0.00005940, -0.00001942, -0.00001388, 0.00004256, -0.00001659, 0.00003864, },
		{ -0.00000198, 0.00003269, -0.00004068, 0.00009705, -0.00003307, 0.00004032, -0.00005182, -0.00000254, -0.00003392, 0.00000438, -0.00003383, 0.00002708, -0.00002005, -0.00000852, -0.00004418, 0.00003500, -0.00007827, 0.00001808, -0.00010548, 0.00006934, -0.00022817, -0.00003034, -0.00094752, 0.00098984, 0.00050408, 0.00106436, -0.00002482, -0.00070205, 0.00007562, -0.00018928, 0.00003372, -0.00013028, 0.00001893, -0.00004488, 0.00005066, -0.00009578, 0.00004201, -0.00004395, 0.00004173, -0.00007842, 0.00008760, -0.00008790, 0.00004323, -0.00005279, -0.00001515, -0.00000276, 0.00001479, -0.00003239, -0.00000454, -0.00002389, 0.00000991, -0.00001162, -0.00004325, 0.00001144, -0.00003869, -0.00001395, 0.00002256, -0.00001980, 0.00005972, -0.00004671, 0.00002956, 0.00001663, 0.00000080, 0.00004421, },
		{ 0.00004893, -0.00004996, 0.00008554, -0.00000020, 0.00002445, -0.00003033, -0.00002582, -0.00002032, -0.00000325, -0.00002970, 0.00001262, -0.00000727, 0.00000119, -0.00006155, 0.00004202, -0.00007514, 0.00003978, -0.00013054, 0.00008118, -0.00020265, 0.00004883, -0.00079891, 0.00031223, 0.00087401, 0.00038401, 0.00096047, -0.00080773, -0.00007551, -0.00016988, 0.00002038, -0.00012527, 0.00000762, -0.00007419, 0.00008313, -0.00009849, 0.00002976, -0.00005315, 0.00005987, -0.00008406, 0.00007231, -0.00005690, 0.00001053, -0.00002012, -0.00003879, -0.00000497, 0.00002069, -0.00002643, -0.00001358, -0.00001951, 0.00000167, -0.00000765, -0.00002762, -0.00001719, 0.00000314, -0.00006153, 0.00003013, -0.00001288, 0.00002008, 0.00003745, -0.00005586, 0.00007655, -0.00003657, 0.00006796, -0.00001093, },
		{ -0.00001366, 0.00002941, 0.00006368, -0.00001429, 0.00001831, -0.00007092, 0.00000922, -0.00003158, 0.00000314, -0.00002951, 0.00002533, -0.00001573, -0.00003287, -0.00000579, -0.00002127, -0.00000604, -0.00005393, -0.00002362, -0.00007023, -0.00004651, -0.00043077, -0.00035035, 0.00109312, 0.00008313, 0.00098866, -0.00003953, -0.00053932, -0.00002641, -0.00011187, -0.00002237, -0.00007750, -0.00003061, 0.00002105, -0.00001231, -0.00004251, -0.00001172, 0.00001472, -0.00002257, -0.00000032, 0.00001785, -0.00005475, 0.00003593, -0.00007618, 0.00000886, -0.00000193, 0.00001047, -0.00004480, 0.00000081, -0.00001950, 0.00000934, -0.00003629, -0.00000580, -0.00001411, -0.00002003, -0.00005314, 0.00005995, -0.00005892, 0.00010439, -0.00006777, 0.00004679, 0.00000640, 0.00002344, 0.00003347, 0.00002192, },
		{ -0.00001815, 0.00010200, -0.00002187, 0.00003780, -0.00007684, -0.00000281, -0.00002610, 0.00000352, -0.00003014, 0.00001056, 0.00000406, -0.00003168, -0.00001763, -0.00001370, -0.00001370, -0.00001733, -0.00005834, -0.00003759, -0.00006619, -0.00020301, -0.00059877, 0.00079335, 0.00038733, 0.00038748, 0.00074406, -0.00055575, -0.00018205, -0.00009124, -0.00004097, -0.00006339, -0.00005181, -0.00000186, 0.00002097, -0.00004843, -0.00001894, -0.00000319, 0.00001082, -0.00003585, 0.00004903, -0.00007283, 0.00004259, -0.00006501, -0.00001229, -0.00000319, 0.00002215, -0.00004266, -0.00000879, -0.00001901, 0.00001231, -0.00003972, 0.00000347, -0.00002908, 0.00001134, -0.00008376, 0.00003943, -0.00002512, 0.00003425, 0.00004507, -0.00005715, 0.00007983, -0.00002715, 0.00006819, 0.00000610, 0.00001942, },
		{ 0.00008941, 0.00001687, 0.00001836, -0.00004066, -0.00004682, -0.00000128, -0.00001680, -0.00000488, -0.00002401, 0.00002833, -0.00003347, -0.00001472, -0.00002327, -0.00001105, -0.00000550, -0.00004578, -0.00004989, -0.00006677, -0.00007304, -0.00058458, 0.00032224, 0.00069595, 0.00005969, 0.00077084, 0.00006492, -0.00049234, -0.00004321, -0.00010394, -0.00001612, -0.00008817, -0.00000614, 0.00001014, -0.00001203, -0.00004349, -0.00000946, 0.00001684, -0.00003326, 0.00003748, -0.00004545, 0.00000415, -0.00001509, -0.00005721, 0.00001445, 0.00000363, -0.00000637, -0.00004272, -0.00000022, -0.00000583, -0.00001540, -0.00002560, 0.00000412, -0.00002217, -0.00002221, -0.00006379, 0.00005906, -0.00006002, 0.00012530, -0.00007728, 0.00006135, 0.00000089, 0.00003959, 0.00003223, 0.00002426, -0.00002801, },
		{ 0.00006445, 0.00000113, -0.00000867, -0.00007148, -0.00000313, -0.00001718, 0.00000042, -0.00002664, 0.00001136, -0.00000230, -0.00003567, -0.00000351, -0.00004167, 0.00002641, -0.00005496, -0.00001608, -0.00012404, 0.00003838, -0.00049701, -0.00002524, 0.00069855, 0.00017424, 0.00035526, 0.00059728, -0.00036931, -0.00025431, -0.00004025, -0.00007365, -0.00003618, -0.00006860, 0.00003313, -0.00002107, -0.00001178, -0.00005270, 0.00003645, -0.00003829, 0.00003150, -0.00002297, -0.00002446, 0.00001461, -0.00007050, 0.00000499, -0.00000014, 0.00001170, -0.00004831, -0.00000636, -0.00001028, 0.00000234, -0.00004911, 0.00002639, -0.00004170, 0.00001463, -0.00009602, 0.00003838, -0.00003875, 0.00006739, 0.00002508, -0.00002807, 0.00005996, 0.00000216, 0.00005205, 0.00003151, -0.00001995, 0.00005264, },
		{ 0.00000986, 0.00000094, -0.00006687, -0.00002695, -0.00000882, -0.00000472, -0.00001268, -0.00001617, 0.00002231, -0.00004390, 0.00000768, -0.00005721, 0.00002990, -0.00004451, 0.00000545, -0.00014058, 0.00003835, -0.00030788, -0.00029453, 0.00058021, 0.00031797, 0.00015562, 0.00052943, 0.00020133, -0.00049720, -0.00002045, -0.00012317, 0.00000663, -0.00010983, 0.00003851, -0.00003517, 0.00002281, -0.00007552, 0.00002968, -0.00002794, 0.00001649, 0.00000076, -0.00004104, 0.00001923, -0.00005269, -0.00002601, 0.00001070, 0.00000535, -0.00002404, -0.00003275, -0.00000101, -0.00000033, -0.00003830, 0.00000055, -0.00000638, -0.00001699, -0.00003726, -0.00005426, 0.00002319, -0.00001522, 0.00010088, -0.00004615, 0.00004590, 0.00002304, 0.00003024, 0.00005361, -0.00000558, 0.00001671, 0.00009085, },
		{ 0.00000651, -0.00005861, -0.00004063, -0.00001935, 0.00000252, -0.00001154, -0.00001720, 0.00000780, -0.00001556, -0.00001542, -0.00002758, -0.00001813, 0.00000768, -0.00002670, -0.00006651, -0.00007948, -0.00006602, -0.00048178, 0.00043519, 0.00034982, 0.00020612, 0.00022060, 0.00056517, -0.00029249, -0.00025068, -0.00004602, -0.00006016, -0.00005338, -0.00003098, 0.00000605, -0.00000816, -0.00002446, -0.00003290, 0.00001821, -0.00002701, 0.00004163, -0.00006144, 0.00002183, -0.00003711, -0.00004151, 0.00000394, 0.00000707, -0.00001017, -0.00003864, -0.00000903, 0.00000039, -0.00002114, -0.00003008, 0.00002262, -0.00003626, -0.00000326, -0.00007816, -0.00000174, -0.00001363, 0.00006986, 0.00002776, -0.00001195, 0.00005403, 0.00001769, 0.00005082, 0.00002984, -0.00001495, 0.00009567, 0.00003310, },
		{ -0.00005184, -0.00004551, -0.00003458, 0.00000315, -0.00000782, -0.00001150, -0.00001165, 0.00000594, -0.00002966, -0.00000186, -0.00005421, 0.00003456, -0.00004512, 0.00000049, -0.00015830, 0.00004512, -0.00045797, 0.00017089, 0.00042270, 0.00019470, 0.00015534, 0.00040211, 0.00025380, -0.00045846, -0.00003054, -0.00011137, -0.00000635, -0.00008573, 0.00003607, -0.00004502, 0.00002781, -0.00007534, 0.00003857, -0.00004921, 0.00005626, -0.00005392, 0.00000542, -0.00001701, -0.00004756, -0.00001111, 0.00001169, -0.00000705, -0.00002801, -0.00002507, 0.00000189, -0.00001147, -0.00003617, 0.00001069, -0.00000979, -0.00002401, -0.00003412, -0.00006336, -0.00000193, 0.00002789, 0.00007152, -0.00000373, 0.00002341, 0.00005161, 0.00001728, 0.00006862, -0.00002404, 0.00007685, 0.00006191, 0.00001618, },
		{ -0.00005993, -0.00003094, -0.00002101, 0.00001136, -0.00001863, -0.00000796, -0.00000531, -0.00001365, -0.00000746, -0.00004412, 0.00000662, -0.00001649, 0.00000198, -0.00012473, -0.00001060, -0.00026593, -0.00014449, 0.00050641, 0.00012101, 0.00024625, 0.00011426, 0.00054096, -0.00024439, -0.00023013, -0.00005816, -0.00005046, -0.00005829, -0.00000190, -0.00002518, 0.00001213, -0.00004275, -0.00000194, -0.00002123, 0.00001939, -0.00000332, -0.00003790, 0.00001663, -0.00006080, -0.00001235, -0.00000102, 0.00000649, -0.00002858, -0.00002206, -0.00001266, 0.00000046, -0.00003329, -0.00000885, 0.00001029, -0.00002911, -0.00002089, -0.00005523, -0.00005598, 0.00003143, 0.00004781, 0.00005836, -0.00001748, 0.00006939, 0.00001517, 0.00006178, 0.00001135, 0.00003444, 0.00009331, 0.00002476, -0.00003477, },
		{ -0.00002256, -0.00004483, 0.00001836, -0.00001716, -0.00000366, -0.00001683, -0.00000427, -0.00001081, -0.00002625, -0.00002461, 0.00000638, -0.00001002, -0.00006054, -0.00008854, -0.00010780, -0.00031275, 0.00042935, 0.00017443, 0.00020882, 0.00008953, 0.00036108, 0.00023352, -0.00037818, -0.00006799, -0.00008101, -0.00003421, -0.00003871, -0.00000052, -0.00002147, 0.00000193, -0.00003848, 0.00000328, -0.00001935, 0.00003883, -0.00006209, 0.00002928, -0.00005774, -0.00001855, -0.00001189, 0.00001478, -0.00002848, -0.00001615, -0.00002294, 0.00000175, -0.00002500, -0.00001794, 0.00000347, -0.00000385, -0.00004218, -0.00001223, -0.00009781, 0.00000101, 0.00003939, 0.00007372, 0.00001581, 0.00002305, 0.00006196, 0.00002028, 0.00005866, 0.00000276, 0.00010593, 0.00004212, -0.00001447, -0.00007550, },
		{ -0.00004234, -0.00000511, 0.00000370, -0.00001366, -0.00000816, -0.00001817, 0.00000106, -0.00002164, -0.00002989, -0.00000238, -0.00000326, -0.00002659, -0.00010284, -0.00006228, -0.00030696, 0.00021289, 0.00031805, 0.00008575, 0.00019835, 0.00010026, 0.00046120, -0.00017033, -0.00023180, -0.00004968, -0.00006368, -0.00002992, -0.00001341, -0.00002144, 0.00000138, -0.00002608, -0.00000787, -0.00002591, 0.00003719, -0.00004050, 0.00000334, -0.00002477, -0.00004490, -0.00000281, 0.00000061, -0.00000946, -0.00002983, -0.00000935, -0.00001677, -0.00000828, -0.00002601, -0.00000309, 0.00000613, -0.00003538, -0.00001783, -0.00005595, -0.00008639, 0.00005161, 0.00004456, 0.00008179, -0.00001672, 0.00008219, 0.00001620, 0.00005974, 0.00001603, 0.00008457, 0.00007705, -0.00000261, -0.00007513, -0.00003239, },
		{ -0.00003335, 0.00001603, -0.00001634, -0.00000125, -0.00002632, -0.00000050, -0.00000951, -0.00003116, -0.00001167, -0.00000506, -0.00000304, -0.00008902, -0.00005555, -0.00025247, 0.00001611, 0.00038081, 0.00006461, 0.00019294, 0.00006314, 0.00031484, 0.00021028, -0.00030571, -0.00009436, -0.00006686, -0.00003672, -0.00001842, -0.00002036, -0.00001023, -0.00000963, -0.00001267, -0.00002937, 0.00001937, -0.00000780, -0.00002500, 0.00000517, -0.00006347, 0.00000178, -0.00001213, 0.00000668, -0.00004182, 0.00000127, -0.00002531, -0.00000509, -0.00002556, -0.00000533, -0.00000297, -0.00000612, -0.00004775, -0.00000130, -0.00012445, -0.00000003, 0.00004283, 0.00009476, 0.00001917, 0.00003975, 0.00005994, 0.00002753, 0.00004622, 0.00005738, 0.00010830, 0.00001446, -0.00006731, -0.00005214, -0.00002817, },
		{ 0.00000326, -0.00000338, -0.00000650, -0.00001660, -0.00001975, 0.00000813, -0.00003268, -0.00001128, -0.00001931, 0.00001561, -0.00007179, -0.00004626, -0.00020185, -0.00011388, 0.00033956, 0.00013838, 0.00010779, 0.00013072, 0.00012176, 0.00036037, -0.00009061, -0.00023784, -0.00004171, -0.00007032, -0.00000409, -0.00003389, -0.00000634, -0.00001904, 0.00000604, -0.00003862, 0.00000310, 0.00000975, -0.00003095, 0.00001046, -0.00005487, -0.00001638, -0.00000478, 0.00000140, -0.00003133, -0.00001423, -0.00000635, -0.00002272, -0.00001328, -0.00001369, -0.00000437, 0.00000179, -0.00004191, -0.00000835, -0.00006900, -0.00009737, 0.00004045, 0.00007536, 0.00008074, 0.00000591, 0.00008015, 0.00002698, 0.00004419, 0.00005498, 0.00011514, 0.00005181, -0.00006418, -0.00006619, -0.00002465, -0.00004143, },
		{ 0.00000252, -0.00000609, -0.00000918, -0.00002877, 0.00000488, -0.00001724, -0.00001848, -0.00002051, 0.00000782, -0.00003739, -0.00005453, -0.00012981, -0.00021427, 0.00026135, 0.00020512, 0.00007164, 0.00012910, 0.00008736, 0.00024110, 0.00020605, -0.00025427, -0.00010580, -0.00006377, -0.00002120, -0.00002551, -0.00001473, -0.00002181, 0.00000724, -0.00002096, -0.00002657, 0.00002230, -0.00002944, 0.00000783, -0.00003702, -0.00003795, 0.00000323, -0.00000632, -0.00001641, -0.00003456, 0.00001091, -0.00002993, -0.00001203, -0.00001703, -0.00000173, -0.00000560, -0.00001714, -0.00003663, -0.00001215, -0.00012439, -0.00003242, 0.00006939, 0.00010607, 0.00003028, 0.00005429, 0.00005981, 0.00002596, 0.00005654, 0.00011200, 0.00009607, -0.00005217, -0.00008212, -0.00002753, -0.00003915, -0.00001417, },
		{ 0.00000157, -0.00000870, -0.00002228, -0.00001463, 0.00000260, -0.00002541, -0.00001248, -0.00001411, -0.00000171, -0.00006885, -0.00005748, -0.00026221, 0.00014280, 0.00025709, 0.00007582, 0.00009355, 0.00010550, 0.00011996, 0.00028295, -0.00003411, -0.00022974, -0.00004682, -0.00005528, -0.00000545, -0.00003312, -0.00000956, -0.00001682, 0.00001541, -0.00005270, 0.00002252, -0.00002355, 0.00000337, -0.00001982, -0.00004944, -0.00000185, -0.00000331, -0.00001138, -0.00004038, 0.00000477, -0.00001266, -0.00002604, -0.00001222, -0.00000604, -0.00000394, -0.00001133, -0.00003381, -0.00001412, -0.00006746, -0.00011953, 0.00002702, 0.00011753, 0.00007152, 0.00003886, 0.00007187, 0.00003657, 0.00003700, 0.00010976, 0.00013275, -0.00001620, -0.00010668, -0.00003102, -0.00003889, -0.00002094, -0.00000662, },
		{ -0.00000085, -0.00001842, -0.00002332, 0.00000079, -0.00001440, -0.00001397, -0.00002171, 0.00000609, -0.00005850, -0.00002985, -0.00022153, -0.00002383, 0.00030059, 0.00008483, 0.00008609, 0.00008058, 0.00010858, 0.00017396, 0.00020037, -0.00020870, -0.00011721, -0.00005033, -0.00001786, -0.00002827, -0.00001179, -0.00002784, 0.00002317, -0.00003790, -0.00000749, -0.00000305, -0.00001297, 0.00000348, -0.00005693, -0.00001028, -0.00000177, -0.00000523, -0.00004084, -0.00001038, 0.00000470, -0.00002969, -0.00001594, -0.00000899, 0.00000087, -0.00001394, -0.00002086, -0.00002862, -0.00002434, -0.00011690, -0.00008249, 0.00011389, 0.00010649, 0.00005333, 0.00006273, 0.00006036, 0.00002552, 0.00008646, 0.00016517, 0.00003905, -0.00012407, -0.00004834, -0.00003322, -0.00002785, -0.00001177, -0.00000278, },
		{ -0.00001634, -0.00002030, -0.00001428, -0.00000070, -0.00001518, -0.00001954, -0.00000385, -0.00003471, -0.00003530, -0.00013606, -0.00016262, 0.00027232, 0.00013405, 0.00007179, 0.00006318, 0.00010522, 0.00009807, 0.00023025, 0.00000043, -0.00021384, -0.00005038, -0.00003916, -0.00001626, -0.00002289, -0.00002104, 0.00000082, 0.00000119, -0.00004051, 0.00001123, -0.00002672, 0.00001932, -0.00005231, -0.00002274, -0.00000323, 0.00000201, -0.00003712, -0.00002586, 0.00000984, -0.00001888, -0.00002386, -0.00001244, 0.00000175, -0.00000851, -0.00002209, -0.00002133, -0.00002858, -0.00005347, -0.00015824, 0.00002492, 0.00015361, 0.00007062, 0.00006995, 0.00006252, 0.00004604, 0.00004484, 0.00017656, 0.00011232, -0.00011926, -0.00008546, -0.00002227, -0.00003889, -0.00000748, -0.00001731, 0.00000825, },
		{ -0.00002041, -0.00001943, -0.00000576, -0.00000323, -0.00002113, -0.00000917, -0.00002241, -0.00003882, -0.00007185, -0.00020571, 0.00014817, 0.00021793, 0.00004596, 0.00007394, 0.00006832, 0.00010161, 0.00013290, 0.00018224, -0.00016549, -0.00012471, -0.00003238, -0.00002721, -0.00001963, -0.00002053, -0.00001525, 0.00001735, -0.00003703, -0.00000416, -0.00001983, 0.00001073, -0.00002557, -0.00004469, 0.00000054, -0.00000130, -0.00001989, -0.00004636, 0.00001164, -0.00001219, -0.00001807, -0.00002424, 0.00000160, -0.00000136, -0.00002335, -0.00001666, -0.00003032, -0.00002622, -0.00012043, -0.00012661, 0.00015387, 0.00011245, 0.00007915, 0.00006758, 0.00006178, 0.00003406, 0.00013507, 0.00019960, -0.00008379, -0.00013088, -0.00002237, -0.00004209, -0.00000869, -0.00002105, 0.00000039, -0.00000451, },
		},
		{ { 0.00000139, 0.00000153, -0.00004479, 0.00000220, -0.00000333, -0.00001406, -0.00002704, -0.00001367, -0.00002509, 0.00000526, -0.00003590, -0.00001735, -0.00002408, -0.00002982, 0.00000600, -0.00005257, -0.00015851, 0.00000916, 0.00020757, 0.00007884, 0.00007501, 0.00007093, 0.00007035, 0.00007973, 0.00020178, 0.00000828, -0.00015795, -0.00004615, -0.00003997, -0.00002932, -0.00000298, -0.00002246, -0.00002250, -0.00002246, -0.00000298, -0.00002932, -0.00003997, -0.00004615, -0.00015795, 0.00000828, 0.00020178, 0.00007973, 0.00007035, 0.00007093, 0.00007501, 0.00007884, 0.00020757, 0.00000916, -0.00015851, -0.00005257, 0.00000600, -0.00002982, -0.00002408, -0.00001735, -0.00003590, 0.00000526, -0.00002509, -0.00001367, -0.00002704, -0.00001406, -0.00000333, 0.00000220, -0.00004479, 0.00000153, },
		{ 0.00000257, -0.00001348, -0.00003989, 0.00001251, -0.00001057, -0.00001955, -0.00002326, -0.00001288, -0.00001665, -0.00001145, -0.00003216, -0.00000548, -0.00004212, -0.00001491, 0.00000154, -0.00010222, -0.00011889, 0.00013683, 0.00014497, 0.00005814, 0.00006879, 0.00006809, 0.00005429, 0.00010585, 0.00017169, -0.00004225, -0.00013877, -0.00003320, -0.00004250, -0.00001610, -0.00000755, -0.00002264, -0.00002462, -0.00001654, -0.00000710, -0.00003766, -0.00003897, -0.00006006, -0.00018034, 0.00008653, 0.00020247, 0.00006729, 0.00008263, 0.00007906, 0.00007296, 0.00014738, 0.00018810, -0.00013908, -0.00011244, -0.00003396, 0.00000932, -0.00004133, -0.00000346, -0.00004770, -0.00000950, 0.00000068, -0.00002748, -0.00001932, -0.00002268, -0.00001116, 0.00000321, -0.00001610, -0.00003390, 0.00000711, },
		{ 0.00000236, -0.00002918, -0.00002424, 0.00001271, -0.00001753, -0.00002249, -0.00001733, -0.00001103, -0.00001664, -0.00002801, -0.00000968, -0.00001841, -0.00004521, 0.00000021, -0.00002378, -0.00013669, -0.00000761, 0.00017367, 0.00007966, 0.00006075, 0.00005776, 0.00006413, 0.00004620, 0.00012813, 0.00012985, -0.00007517, -0.00012072, -0.00002189, -0.00004377, -0.00000320, -0.00001654, -0.00001964, -0.00002765, -0.00000982, -0.00001571, -0.00004380, -0.00003397, -0.00008744, -0.00018862, 0.00018241, 0.00017123, 0.00007701, 0.00008501, 0.00009125, 0.00008302, 0.00023760, 0.00004034, -0.00019393, -0.00005402, -0.00003347, 0.00001163, -0.00004207, -0.00000132, -0.00006653, 0.00002613, -0.00002185, -0.00002172, -0.00002760, -0.00001218, -0.00001096, 0.00000398, -0.00003100, -0.00001564, 0.00000510, },
		{ -0.00000445, -0.00003473, -0.00000982, 0.00000761, -0.00002446, -0.00001986, -0.00001313, -0.00000775, -0.00003065, -0.00002364, 0.00000499, -0.00004186, -0.00003012, 0.00000128, -0.00006482, -0.00011232, 0.00010260, 0.00013306, 0.00005255, 0.00005872, 0.00005119, 0.00005534, 0.00005031, 0.00013721, 0.00008903, -0.00010145, -0.00009619, -0.00001862, -0.00003733, 0.00000081, -0.00002119, -0.00002038, -0.00002491, -0.00001066, -0.00002020, -0.00005382, -0.00001815, -0.00014531, -0.00015036, 0.00026130, 0.00013179, 0.00009457, 0.00008886, 0.00009480, 0.00014667, 0.00024864, -0.00015914, -0.00012446, -0.00005346, -0.00000966, -0.00000517, -0.00002406, -0.00002965, -0.00004609, 0.00003639, -0.00003901, -0.00001957, -0.00002754, -0.00000543, -0.00000936, -0.00000486, -0.00003143, -0.00000340, 0.00000279, },
		{ -0.00001437, -0.00002986, -0.00000277, 0.00000121, -0.00002962, -0.00001342, -0.00000985, -0.00001054, -0.00004673, 0.00000153, -0.00000633, -0.00005166, -0.00001109, -0.00002026, -0.00008848, -0.00003734, 0.00015620, 0.00006829, 0.00005972, 0.00004108, 0.00005547, 0.00004015, 0.00006547, 0.00013238, 0.00005158, -0.00012210, -0.00006495, -0.00002434, -0.00002264, -0.00000625, -0.00001810, -0.00002666, -0.00001635, -0.00001949, -0.00002013, -0.00006320, -0.00000262, -0.00022720, -0.00005142, 0.00029553, 0.00010771, 0.00010600, 0.00009963, 0.00010155, 0.00025677, 0.00009468, -0.00025027, -0.00003378, -0.00008335, 0.00002995, -0.00003205, -0.00000551, -0.00006397, 0.00000081, 0.00001527, -0.00004101, -0.00002373, -0.00001770, -0.00000552, -0.00000835, -0.00001614, -0.00002041, -0.00000058, 0.00000126, },
		{ -0.00001684, -0.00002549, 0.00000135, -0.00000758, -0.00002901, -0.00000749, -0.00000659, -0.00002522, -0.00004585, 0.00002574, -0.00003707, -0.00003425, -0.00001485, -0.00004306, -0.00007793, 0.00005080, 0.00014140, 0.00002862, 0.00007074, 0.00001957, 0.00006488, 0.00002322, 0.00008599, 0.00011790, 0.00001329, -0.00012887, -0.00003744, -0.00002970, -0.00000982, -0.00001487, -0.00001512, -0.00002994, -0.00001288, -0.00002437, -0.00002716, -0.00005487, -0.00001920, -0.00029181, 0.00008145, 0.00029258, 0.00009440, 0.00012253, 0.00009785, 0.00016765, 0.00029616, -0.00015247, -0.00017020, -0.00003033, -0.00006955, 0.00003915, -0.00003849, -0.00001559, -0.00006332, 0.00003107, -0.00001041, -0.00004101, -0.00002117, -0.00001095, -0.00000496, -0.00001533, -0.00001605, -0.00001260, 0.00000070, -0.00000523, },
		{ -0.00000940, -0.00002920, 0.00000840, -0.00002044, -0.00002010, -0.00000566, -0.00000560, -0.00004724, -0.00001914, 0.00002192, -0.00005516, -0.00001105, -0.00004266, -0.00004132, -0.00004377, 0.00011836, 0.00008130, 0.00003375, 0.00005693, 0.00001640, 0.00006336, 0.00001612, 0.00010281, 0.00009789, -0.00002686, -0.00011676, -0.00002263, -0.00002652, -0.00000780, -0.00001599, -0.00001913, -0.00002474, -0.00002071, -0.00001861, -0.00004483, -0.00002610, -0.00008511, -0.00030037, 0.00021212, 0.00026882, 0.00008617, 0.00014781, 0.00009009, 0.00030600, 0.00013645, -0.00028772, -0.00004338, -0.00008575, -0.00001187, 0.00001452, -0.00002921, -0.00004239, -0.00002822, 0.00002804, -0.00002450, -0.00004439, -0.00000852, -0.00001271, -0.00000293, -0.00002668, -0.00000380, -0.00001447, 0.00000405, -0.00001578, },
		{ -0.00000536, -0.00003349, 0.00001445, -0.00003209, -0.00000723, -0.00000686, -0.00001289, -0.00005940, 0.00001701, -0.00001154, -0.00003690, -0.00002126, -0.00005613, -0.00002391, 0.00000550, 0.00013835, 0.00002341, 0.00006055, 0.00002214, 0.00003513, 0.00004488, 0.00002436, 0.00011159, 0.00007016, -0.00005866, -0.00009549, -0.00001474, -0.00002131, -0.00001017, -0.00001525, -0.00002271, -0.00002150, -0.00002779, -0.00001471, -0.00005510, -0.00000542, -0.00017540, -0.00025381, 0.00033334, 0.00021440, 0.00011014, 0.00014366, 0.00015487, 0.00038130, -0.00016110, -0.00020721, -0.00002170, -0.00009446, 0.00002236, -0.00000105, -0.00004048, -0.00004021, -0.00000364, 0.00001802, -0.00004054, -0.00003495, -0.00000313, -0.00001115, -0.00001088, -0.00002523, 0.00000346, -0.00001577, 0.00000126, -0.00001764, },
		{ -0.00001769, -0.00002577, 0.00001034, -0.00003335, 0.00000073, -0.00000755, -0.00003188, -0.00004381, 0.00002779, -0.00003654, -0.00000884, -0.00006226, -0.00002986, -0.00001263, 0.00006354, 0.00010346, 0.00000786, 0.00006787, -0.00000311, 0.00005550, 0.00002075, 0.00004326, 0.00011123, 0.00003411, -0.00007260, -0.00007759, -0.00000454, -0.00002356, -0.00000729, -0.00002044, -0.00001865, -0.00002902, -0.00002377, -0.00002147, -0.00004769, -0.00001666, -0.00025821, -0.00016329, 0.00043720, 0.00013073, 0.00017771, 0.00009965, 0.00033820, 0.00021051, -0.00034158, -0.00005236, -0.00008394, -0.00004262, 0.00001620, -0.00000255, -0.00006674, -0.00000407, -0.00001013, 0.00001806, -0.00006385, -0.00000759, -0.00001391, -0.00000133, -0.00002975, -0.00000547, -0.00000510, -0.00000862, -0.00001138, -0.00000355, },
		{ -0.00003897, -0.00000779, -0.00000279, -0.00002422, 0.00000192, -0.00001133, -0.00005013, -0.00000860, 0.00000147, -0.00001801, -0.00002432, -0.00007770, 0.00000519, 0.00000304, 0.00010052, 0.00004923, 0.00002604, 0.00004687, -0.00000404, 0.00006285, 0.00000062, 0.00007051, 0.00009309, 0.00000445, -0.00007820, -0.00005780, -0.00000002, -0.00002478, -0.00000720, -0.00002141, -0.00002007, -0.00003292, -0.00002264, -0.00002092, -0.00004752, -0.00004156, -0.00033878, -0.00000701, 0.00047191, 0.00008066, 0.00022393, 0.00011860, 0.00049116, -0.00016760, -0.00024690, -0.00002515, -0.00009366, -0.00001516, 0.00002294, -0.00003511, -0.00005261, 0.00001252, -0.00001417, 0.00000625, -0.00006875, 0.00001214, -0.00002286, 0.00000035, -0.00003850, 0.00001264, -0.00001699, -0.00000317, -0.00001955, 0.00001213, },
		{ -0.00004960, 0.00000614, -0.00001314, -0.00001498, 0.00000279, -0.00002724, -0.00004355, 0.00000703, -0.00002081, 0.00001528, -0.00008154, -0.00003557, 0.00001054, 0.00004310, 0.00008637, 0.00002572, 0.00003535, 0.00002547, 0.00000484, 0.00006093, -0.00001381, 0.00010376, 0.00005136, -0.00000291, -0.00009075, -0.00002527, -0.00001362, -0.00001201, -0.00002096, -0.00000749, -0.00003940, -0.00001937, -0.00003573, -0.00000205, -0.00007437, -0.00006050, -0.00042011, 0.00022423, 0.00039856, 0.00012150, 0.00019250, 0.00030408, 0.00035637, -0.00043902, -0.00003449, -0.00011464, -0.00002380, -0.00003844, 0.00005274, -0.00009766, 0.00001389, -0.00001917, 0.00000889, -0.00003106, -0.00003995, 0.00000498, -0.00001580, -0.00001561, -0.00002345, 0.00001084, -0.00001740, -0.00000997, -0.00000893, 0.00000898, },
		{ -0.00004462, 0.00001099, -0.00001675, -0.00000912, 0.00000147, -0.00004782, -0.00001184, -0.00001868, 0.00000809, -0.00000573, -0.00010624, 0.00001571, 0.00000779, 0.00008064, 0.00004483, 0.00003357, 0.00002305, 0.00001799, 0.00001420, 0.00004898, -0.00001112, 0.00012225, 0.00000688, -0.00000013, -0.00010037, 0.00000591, -0.00003153, 0.00000156, -0.00003261, 0.00000172, -0.00005536, -0.00000980, -0.00003583, -0.00000207, -0.00009086, -0.00011270, -0.00043866, 0.00045390, 0.00028067, 0.00019712, 0.00017909, 0.00052530, -0.00008477, -0.00035975, 0.00002204, -0.00013986, 0.00001109, -0.00003531, 0.00003744, -0.00011316, 0.00006253, -0.00005392, 0.00002808, -0.00006161, -0.00000614, -0.00001029, -0.00000925, -0.00002860, -0.00000328, -0.00000016, -0.00001452, -0.00001711, 0.00000975, -0.00001101, },
		{ -0.00003647, 0.00001579, -0.00002255, 0.00000177, -0.00001595, -0.00004422, 0.00000311, -0.00004324, 0.00006039, -0.00008612, -0.00005257, 0.00001689, 0.00004187, 0.00006696, 0.00003368, 0.00002713, 0.00001982, 0.00000631, 0.00003411, 0.00001725, 0.00002397, 0.00009982, -0.00000760, -0.00001517, -0.00008330, 0.00000938, -0.00002765, -0.00000772, -0.00001857, -0.00001993, -0.00003978, -0.00003012, 0.00000402, -0.00005982, -0.00005080, -0.00024774, -0.00031799, 0.00057972, 0.00022022, 0.00021467, 0.00033669, 0.00045975, -0.00047690, -0.00007758, -0.00010143, -0.00004069, -0.00004224, 0.00002810, -0.00004671, -0.00004251, 0.00003889, -0.00004554, 0.00001153, -0.00005459, 0.00000151, -0.00001081, -0.00001985, -0.00001872, -0.00000133, -0.00000139, -0.00002278, -0.00000648, 0.00001314, -0.00002760, },
		{ -0.00002926, 0.00002164, -0.00003107, 0.00001556, -0.00004653, -0.00000772, -0.00003113, -0.00000562, 0.00005001, -0.00013586, 0.00001967, -0.00000616, 0.00009191, 0.00001519, 0.00005999, -0.00000825, 0.00003950, -0.00001656, 0.00006087, -0.00002242, 0.00007634, 0.00004583, 0.00000727, -0.00004625, -0.00004623, -0.00000880, -0.00000872, -0.00002867, 0.00000528, -0.00005433, -0.00001176, -0.00005344, 0.00004500, -0.00013551, 0.00000733, -0.00042205, -0.00008589, 0.00060917, 0.00020166, 0.00024417, 0.00057074, 0.00000944, -0.00047192, 0.00005512, -0.00017972, 0.00004079, -0.00008580, 0.00006687, -0.00011209, 0.00004197, -0.00001583, -0.00001913, -0.00001455, -0.00003276, -0.00000271, -0.00001073, -0.00002837, -0.00000337, -0.00000783, -0.00000045, -0.00003115, 0.00001316, 0.00000050, -0.00003426, },
		{ -0.00001116, 0.00001412, -0.00002611, 0.00001005, -0.00005563, 0.00001772, -0.00007338, 0.00007336, -0.00004355, -0.00008752, 0.00002627, 0.00001757, 0.00008889, -0.00000419, 0.00006730, -0.00002998, 0.00005074, -0.00002526, 0.00006695, -0.00003617, 0.00010494, 0.00000132, 0.00001849, -0.00006329, -0.00002241, -0.00001609, -0.00000588, -0.00002821, 0.00000109, -0.00005983, -0.00001271, -0.00003287, 0.00002200, -0.00014933, -0.00000545, -0.00053227, 0.00016013, 0.00062271, 0.00014111, 0.00046031, 0.00051543, -0.00045702, -0.00017385, -0.00005160, -0.00009550, 0.00000403, -0.00004402, 0.00001846, -0.00008498, 0.00005950, -0.00003840, -0.00001723, -0.00001240, -0.00003166, 0.00000754, -0.00003084, -0.00001016, -0.00001047, 0.00000058, -0.00001637, -0.00001705, 0.00001541, -0.00000681, -0.00004430, },
		{ 0.00001833, -0.00000948, -0.00000387, -0.00002185, -0.00002206, -0.00001391, -0.00004767, 0.00009367, -0.00012264, -0.00000101, -0.00001369, 0.00008257, 0.00002648, 0.00003321, 0.00002524, -0.00000844, 0.00003043, -0.00000607, 0.00004391, -0.00001559, 0.00009580, -0.00001619, 0.00001452, -0.00006080, -0.00001721, -0.00000926, -0.00001749, -0.00001198, -0.00002465, -0.00004274, -0.00002945, 0.00000989, -0.00004984, -0.00010315, -0.00009737, -0.00055711, 0.00041169, 0.00058939, 0.00012897, 0.00077751, 0.00000507, -0.00049762, 0.00000499, -0.00015235, 0.00000568, -0.00005610, 0.00000588, -0.00004229, -0.00002068, 0.00003471, -0.00004229, -0.00001972, -0.00000278, -0.00003589, 0.00001525, -0.00004914, 0.00001547, -0.00002883, 0.00001524, -0.00003798, 0.00001166, -0.00000116, -0.00000328, -0.00005987, },
		{ 0.00003630, -0.00002584, 0.00000869, -0.00004471, 0.00001379, -0.00006933, 0.00003949, 0.00001382, -0.00009716, 0.00001781, -0.00000895, 0.00010593, -0.00001908, 0.00006901, -0.00002367, 0.00002710, 0.00000045, 0.00001830, 0.00001800, 0.00001050, 0.00007443, -0.00002597, 0.00001585, -0.00006462, -0.00000384, -0.00001599, -0.00001214, -0.00001717, -0.00003074, -0.00004433, -0.00001370, 0.00000806, -0.00009353, -0.00007032, -0.00020432, -0.00054409, 0.00072862, 0.00039982, 0.00040465, 0.00075193, -0.00056263, -0.00017410, -0.00009010, -0.00007837, -0.00001510, -0.00003425, -0.00002146, -0.00002629, -0.00000681, 0.00003159, -0.00006980, 0.00001318, -0.00002607, -0.00001148, -0.00001030, -0.00003020, 0.00001019, -0.00002519, 0.00000766, -0.00003413, 0.00002171, -0.00000921, -0.00000939, -0.00006042, },
		{ 0.00002949, -0.00001985, -0.00000770, -0.00002528, -0.00000526, -0.00006264, 0.00008254, -0.00007116, -0.00001790, -0.00002587, 0.00005563, 0.00005764, -0.00000265, 0.00005544, -0.00003515, 0.00004194, -0.00001655, 0.00003258, 0.00000176, 0.00002906, 0.00005146, -0.00003035, 0.00002012, -0.00007334, 0.00001511, -0.00003049, 0.00000180, -0.00003563, -0.00002475, -0.00005027, 0.00001373, -0.00002709, -0.00010745, -0.00005142, -0.00034555, -0.00042811, 0.00101443, 0.00017137, 0.00091291, 0.00011487, -0.00062141, 0.00002704, -0.00018938, 0.00001952, -0.00006430, -0.00000343, -0.00005640, 0.00001529, -0.00001843, 0.00003051, -0.00009145, 0.00005181, -0.00006024, 0.00002089, -0.00004675, 0.00000894, -0.00001859, -0.00000348, -0.00001667, -0.00000645, 0.00000593, 0.00000185, -0.00003377, -0.00003636, },
		{ 0.00001776, -0.00001392, -0.00002417, 0.00000508, -0.00005781, 0.00001562, 0.00002140, -0.00005459, -0.00000454, -0.00002924, 0.00009317, 0.00000842, 0.00002778, 0.00002255, -0.00002517, 0.00004531, -0.00003211, 0.00005335, -0.00002570, 0.00006164, 0.00000507, -0.00000117, -0.00000791, -0.00005004, 0.00000159, -0.00001152, -0.00002228, -0.00001715, -0.00005715, -0.00000201, -0.00002347, -0.00001302, -0.00016603, 0.00002378, -0.00061187, -0.00004792, 0.00102874, 0.00027803, 0.00109474, -0.00070189, -0.00017594, -0.00012806, -0.00006533, -0.00003853, -0.00000714, -0.00007064, 0.00000656, -0.00001817, 0.00002122, -0.00002689, -0.00004763, 0.00003466, -0.00005111, 0.00001242, -0.00004355, 0.00001801, -0.00002831, 0.00000049, -0.00002114, 0.00000805, -0.00000522, 0.00000825, -0.00005401, -0.00000930, },
		{ 0.00001904, -0.00003063, -0.00000347, -0.00001487, -0.00005471, 0.00006028, -0.00005479, 0.00002174, -0.00006384, 0.00004010, 0.00005612, 0.00001426, 0.00002081, 0.00001228, -0.00002102, 0.00004880, -0.00004829, 0.00007862, -0.00006193, 0.00010428, -0.00005676, 0.00005379, -0.00006419, 0.00000155, -0.00003777, 0.00003068, -0.00007318, 0.00002616, -0.00010745, 0.00007291, -0.00010556, 0.00003659, -0.00024086, 0.00009468, -0.00091978, 0.00054578, 0.00079039, 0.00080883, 0.00047854, -0.00093405, 0.00017169, -0.00032274, 0.00010931, -0.00014500, 0.00007186, -0.00015515, 0.00010566, -0.00009132, 0.00008321, -0.00010581, 0.00003265, -0.00002181, -0.00001096, -0.00002423, -0.00000787, -0.00000466, -0.00001489, -0.00001496, -0.00000091, -0.00000344, 0.00000450, -0.00000548, -0.00005371, 0.00000303, },
		{ 0.00001211, -0.00004291, 0.00002728, -0.00007136, 0.00002538, -0.00000334, -0.00002511, 0.00002425, -0.00008059, 0.00009063, 0.00001180, 0.00003187, 0.00000459, 0.00000614, -0.00000536, 0.00002699, -0.00002936, 0.00006385, -0.00005755, 0.00010444, -0.00007563, 0.00007179, -0.00008640, 0.00001928, -0.00003747, 0.00002315, -0.00007146, 0.00001094, -0.00007936, 0.00005586, -0.00010247, -0.00001124, -0.00017841, -0.00005153, -0.00097378, 0.00101776, 0.00073502, 0.00117233, -0.00056610, -0.00042944, 0.00002627, -0.00020940, 0.00006566, -0.00009496, 0.00000660, -0.00009851, 0.00009216, -0.00008329, 0.00006834, -0.00011054, 0.00005896, -0.00004208, -0.00000396, -0.00002760, 0.00000247, -0.00001122, -0.00001815, -0.00001009, 0.00000475, -0.00000890, 0.00001098, -0.00001929, -0.00005014, 0.00001742, },
		{ -0.00002678, -0.00000946, 0.00000175, -0.00007091, 0.00007652, -0.00009068, 0.00007796, -0.00006892, -0.00000130, 0.00005529, 0.00002769, 0.00000460, 0.00002347, -0.00002506, 0.00003705, -0.00002946, 0.00003480, -0.00000589, 0.00000638, 0.00004210, -0.00002909, 0.00002742, -0.00005065, -0.00001555, 0.00001608, -0.00004741, -0.00000319, -0.00007064, 0.00003191, -0.00007024, 0.00001320, -0.00017711, 0.00003443, -0.00046024, -0.00065021, 0.00122354, 0.00111790, 0.00068257, -0.00108651, 0.00013342, -0.00029905, 0.00006390, -0.00011506, 0.00005107, -0.00014476, 0.00006147, -0.00002446, 0.00000386, -0.00001635, -0.00004245, 0.00002315, -0.00001891, -0.00003209, 0.00000595, -0.00002184, 0.00000752, -0.00004334, 0.00002286, -0.00001951, 0.00001030, -0.00000302, -0.00001759, -0.00005742, 0.00004904, },
		{ -0.00006520, 0.00004322, -0.00007305, 0.00002075, 0.00000673, -0.00006226, 0.00010317, -0.00012978, 0.00007973, -0.00000224, 0.00006220, -0.00003606, 0.00005090, -0.00005193, 0.00006601, -0.00006919, 0.00008520, -0.00006977, 0.00007034, -0.00002248, 0.00002165, -0.00001877, -0.00001913, -0.00003354, 0.00004128, -0.00008946, 0.00003331, -0.00010398, 0.00008684, -0.00015212, 0.00008100, -0.00027419, 0.00015822, -0.00085755, -0.00020768, 0.00153999, 0.00139027, -0.00042909, -0.00069757, 0.00014701, -0.00031091, 0.00012642, -0.00015030, 0.00006306, -0.00016911, 0.00013150, -0.00008940, 0.00004619, -0.00005840, -0.00000192, 0.00000690, -0.00002214, -0.00002726, 0.00001110, -0.00002471, 0.00000262, -0.00004123, 0.00003569, -0.00003400, 0.00002412, -0.00001298, -0.00002333, -0.00004981, 0.00006719, },
		{ -0.00004740, 0.00003682, -0.00009619, 0.00009750, -0.00010870, 0.00006810, 0.00000151, -0.00007342, 0.00006867, 0.00000426, 0.00004297, -0.00002899, 0.00003823, -0.00003603, 0.00004480, -0.00005081, 0.00007334, -0.00007514, 0.00008304, -0.00004025, 0.00002910, -0.00002324, -0.00002992, 0.00000091, -0.00000217, -0.00005716, -0.00000646, -0.00004136, 0.00001817, -0.00010434, 0.00001133, -0.00020209, 0.00004562, -0.00105274, 0.00020536, 0.00226614, 0.00071196, -0.00107535, -0.00007843, -0.00015638, -0.00006938, -0.00001752, -0.00003419, -0.00006090, -0.00004712, 0.00006856, -0.00006004, 0.00001478, -0.00003180, -0.00001595, 0.00002960, -0.00006260, 0.00002213, -0.00002920, 0.00001079, -0.00003953, 0.00000526, 0.00000353, -0.00001136, 0.00000844, 0.00000128, -0.00005539, -0.00000294, 0.00003437, },
		{ 0.00001134, -0.00003802, -0.00001295, 0.00004555, -0.00011162, 0.00013803, -0.00009937, 0.00001815, 0.00001234, 0.00004718, -0.00000955, 0.00000954, 0.00000411, -0.00000527, 0.00000965, -0.00001416, 0.00003679, -0.00005569, 0.00007917, -0.00005064, 0.00003760, -0.00003722, -0.00002020, 0.00001070, -0.00002820, -0.00004184, -0.00001589, -0.00001200, -0.00003540, -0.00006730, -0.00003704, -0.00015387, -0.00009959, -0.00127282, 0.00105217, 0.00275356, -0.00076335, -0.00065228, -0.00002463, -0.00017772, -0.00000466, -0.00004713, -0.00002409, -0.00008874, 0.00001890, 0.00002806, -0.00005770, 0.00002024, -0.00003794, 0.00000287, 0.00001057, -0.00005830, 0.00003631, -0.00004493, 0.00002257, -0.00005803, 0.00003881, -0.00002744, 0.00001350, -0.00000729, 0.00001003, -0.00008060, 0.00004934, -0.00002154, },
		{ 0.00002755, -0.00008193, 0.00008089, -0.00008022, 0.00001271, 0.00006659, -0.00008622, 0.00003285, 0.00000659, 0.00004704, -0.00002734, 0.00002457, -0.00000749, -0.00000092, 0.00000467, -0.00000699, 0.00002226, -0.00005124, 0.00009269, -0.00008619, 0.00007913, -0.00009185, 0.00004400, -0.00004721, 0.00001334, -0.00009260, 0.00005857, -0.00009002, 0.00001726, -0.00013259, 0.00003146, -0.00025086, -0.00015507, -0.00162619, 0.00279248, 0.00181170, -0.00167329, 0.00010161, -0.00035632, 0.00008295, -0.00016016, 0.00008180, -0.00015507, 0.00002814, -0.00003582, 0.00006437, -0.00011543, 0.00008924, -0.00010416, 0.00007687, -0.00006867, 0.00001438, -0.00001622, -0.00000367, -0.00001924, -0.00001939, 0.00001935, -0.00001695, 0.00000580, 0.00000700, -0.00001780, -0.00005887, 0.00005376, -0.00003356, },
		{ -0.00002543, -0.00002216, 0.00005567, -0.00011294, 0.00010414, -0.00003185, -0.00002128, -0.00000067, 0.00003657, 0.00001449, -0.00001841, 0.00002852, -0.00002052, 0.00001148, -0.00001036, 0.00001214, -0.00001379, -0.00000893, 0.00005730, -0.00006952, 0.00006650, -0.00008741, 0.00004942, -0.00005149, -0.00000583, -0.00006173, 0.00004238, -0.00008717, -0.00002081, -0.00007207, -0.00006658, -0.00015092, -0.00059695, -0.00132146, 0.00462887, -0.00062701, -0.00100563, 0.00006674, -0.00028159, 0.00007027, -0.00012189, 0.00003744, -0.00013491, 0.00004000, -0.00001544, 0.00001574, -0.00007960, 0.00007608, -0.00009951, 0.00008513, -0.00009197, 0.00004879, -0.00004479, 0.00001665, -0.00004237, 0.00001200, -0.00000282, -0.00000290, -0.00000199, 0.00001897, -0.00004974, -0.00001670, 0.00002491, -0.00000098, },
		{ -0.00005979, 0.00006384, -0.00005927, -0.00000878, 0.00005552, -0.00002607, -0.00002502, 0.00001679, 0.00002003, 0.00001951, -0.00003814, 0.00006531, -0.00007685, 0.00007758, -0.00008547, 0.00009328, -0.00011778, 0.00011856, -0.00008229, 0.00006306, -0.00006960, 0.00005271, -0.00009212, 0.00009418, -0.00018367, 0.00015917, -0.00020068, 0.00015672, -0.00031879, 0.00030255, -0.00057159, 0.00045462, -0.00191919, 0.00078614, 0.00461422, -0.00276982, 0.00046088, -0.00069251, 0.00031454, -0.00035246, 0.00023220, -0.00028263, 0.00014332, -0.00016547, 0.00017270, -0.00018625, 0.00011397, -0.00008736, 0.00004159, -0.00003460, 0.00000573, -0.00002322, 0.00001439, -0.00004073, 0.00000933, -0.00002182, 0.00002607, -0.00003084, 0.00003029, -0.00001297, -0.00003643, -0.00001176, 0.00002788, -0.00000059, },
		{ 0.00000058, 0.00004419, -0.00010047, 0.00008291, -0.00003389, 0.00003537, -0.00007631, 0.00007167, -0.00002952, 0.00004622, -0.00005973, 0.00009016, -0.00011581, 0.00012634, -0.00014344, 0.00015149, -0.00019044, 0.00021640, -0.00020180, 0.00018043, -0.00018688, 0.00017169, -0.00021187, 0.00020671, -0.00030882, 0.00032412, -0.00040305, 0.00035359, -0.00053931, 0.00057543, -0.00097941, 0.00091820, -0.00332635, 0.00436308, 0.00175256, -0.00264402, 0.00094654, -0.00098829, 0.00060942, -0.00056878, 0.00040784, -0.00045236, 0.00031762, -0.00029335, 0.00027509, -0.00030407, 0.00024700, -0.00021050, 0.00015126, -0.00013043, 0.00009031, -0.00008850, 0.00006382, -0.00008642, 0.00005650, -0.00005666, 0.00005444, -0.00005529, 0.00005961, -0.00005000, -0.00000645, -0.00002909, 0.00006326, -0.00006049, },
		{ 0.00008640, -0.00006922, -0.00000147, 0.00003262, -0.00001293, 0.00001276, -0.00005399, 0.00005585, -0.00001316, 0.00000561, 0.00000205, 0.00001053, -0.00003130, 0.00003783, -0.00005448, 0.00004929, -0.00008088, 0.00011431, -0.00011448, 0.00008456, -0.00007366, 0.00003766, -0.00005745, 0.00001415, -0.00007538, 0.00007227, -0.00014897, 0.00003817, -0.00013869, 0.00006270, -0.00039694, -0.00005009, -0.00277054, 0.00688688, -0.00226428, -0.00044887, -0.00012069, -0.00018883, 0.00005305, -0.00011432, 0.00001144, -0.00011299, 0.00005669, -0.00005687, 0.00004564, -0.00010880, 0.00010104, -0.00009389, 0.00005056, -0.00003730, 0.00000856, -0.00001018, -0.00001713, -0.00000832, -0.00000741, 0.00000371, -0.00000118, -0.00000200, 0.00001485, -0.00001941, -0.00003392, 0.00000779, 0.00004675, -0.00009162, },
		{ 0.00006437, -0.00011933, 0.00010342, -0.00008074, 0.00009189, -0.00009506, 0.00005134, -0.00004017, 0.00007422, -0.00009457, 0.00012218, -0.00013605, 0.00013373, -0.00014433, 0.00013588, -0.00016073, 0.00015213, -0.00012950, 0.00012205, -0.00015947, 0.00019269, -0.00026283, 0.00027497, -0.00036807, 0.00038150, -0.00045945, 0.00042776, -0.00063754, 0.00070354, -0.00105782, 0.00103839, -0.00255499, 0.00083466, 0.00575820, -0.00428421, 0.00148311, -0.00137612, 0.00083946, -0.00074027, 0.00053823, -0.00056186, 0.00040919, -0.00038056, 0.00032545, -0.00031521, 0.00022274, -0.00017737, 0.00013742, -0.00014697, 0.00014260, -0.00015124, 0.00013291, -0.00015332, 0.00012521, -0.00012615, 0.00011311, -0.00009944, 0.00008994, -0.00006881, 0.00005238, -0.00010417, 0.00009999, -0.00005310, -0.00001315, },
		{ -0.00006249, -0.00002314, 0.00006533, -0.00008523, 0.00011812, -0.00014189, 0.00010422, -0.00008639, 0.00010487, -0.00012010, 0.00014906, -0.00017596, 0.00018493, -0.00020976, 0.00020347, -0.00023543, 0.00024228, -0.00022980, 0.00020696, -0.00022790, 0.00025406, -0.00033343, 0.00034887, -0.00044722, 0.00049162, -0.00062214, 0.00059934, -0.00082845, 0.00095660, -0.00150289, 0.00166291, -0.00446567, 0.00560226, 0.00127995, -0.00291687, 0.00129897, -0.00127499, 0.00086910, -0.00077367, 0.00055967, -0.00058716, 0.00047048, -0.00043625, 0.00036853, -0.00036843, 0.00029035, -0.00023288, 0.00016950, -0.00015688, 0.00014793, -0.00015230, 0.00012689, -0.00014583, 0.00012752, -0.00012988, 0.00012067, -0.00010530, 0.00009580, -0.00007198, 0.00004989, -0.00010382, 0.00013329, -0.00012917, 0.00010192, },
		{ -0.00013756, 0.00010318, -0.00006954, 0.00003710, 0.00001040, -0.00005686, 0.00002894, -0.00000720, 0.00000866, -0.00000674, 0.00001746, -0.00003548, 0.00003580, -0.00006078, 0.00004448, -0.00006380, 0.00006572, -0.00005181, 0.00000216, 0.00002032, -0.00004312, -0.00000187, -0.00002399, -0.00001210, -0.00000485, -0.00009323, -0.00001478, -0.00007472, 0.00000456, -0.00038450, 0.00003053, -0.00318724, 0.00782624, -0.00318724, 0.00003053, -0.00038450, 0.00000456, -0.00007472, -0.00001478, -0.00009323, -0.00000485, -0.00001210, -0.00002399, -0.00000187, -0.00004312, 0.00002032, 0.00000216, -0.00005181, 0.00006572, -0.00006380, 0.00004448, -0.00006078, 0.00003580, -0.00003548, 0.00001746, -0.00000674, 0.00000866, -0.00000720, 0.00002894, -0.00005686, 0.00001040, 0.00003710, -0.00006954, 0.00010318, },
		{ -0.00006249, 0.00010192, -0.00012917, 0.00013329, -0.00010382, 0.00004989, -0.00007198, 0.00009580, -0.00010530, 0.00012067, -0.00012988, 0.00012752, -0.00014583, 0.00012689, -0.00015230, 0.00014793, -0.00015688, 0.00016950, -0.00023288, 0.00029035, -0.00036843, 0.00036853, -0.00043625, 0.00047048, -0.00058716, 0.00055967, -0.00077367, 0.00086910, -0.00127499, 0.00129897, -0.00291687, 0.00127995, 0.00560226, -0.00446567, 0.00166291, -0.00150289, 0.00095660, -0.00082845, 0.00059934, -0.00062214, 0.00049162, -0.00044722, 0.00034887, -0.00033343, 0.00025406, -0.00022790, 0.00020696, -0.00022980, 0.00024228, -0.00023543, 0.00020347, -0.00020976, 0.00018493, -0.00017596, 0.00014906, -0.00012010, 0.00010487, -0.00008639, 0.00010422, -0.00014189, 0.00011812, -0.00008523, 0.00006533, -0.00002314, },
		{ 0.00006437, -0.00001315, -0.00005310, 0.00009999, -0.00010417, 0.00005238, -0.00006881, 0.00008994, -0.00009944, 0.00011311, -0.00012615, 0.00012521, -0.00015332, 0.00013291, -0.00015124, 0.00014260, -0.00014697, 0.00013742, -0.00017737, 0.00022274, -0.00031521, 0.00032545, -0.00038056, 0.00040919, -0.00056186, 0.00053823, -0.00074027, 0.00083946, -0.00137612, 0.00148311, -0.00428421, 0.00575820, 0.00083466, -0.00255499, 0.00103839, -0.00105782, 0.00070354, -0.00063754, 0.00042776, -0.00045945, 0.00038150, -0.00036807, 0.00027497, -0.00026283, 0.00019269, -0.00015947, 0.00012205, -0.00012950, 0.00015213, -0.00016073, 0.00013588, -0.00014433, 0.00013373, -0.00013605, 0.00012218, -0.00009457, 0.00007422, -0.00004017, 0.00005134, -0.00009506, 0.00009189, -0.00008074, 0.00010342, -0.00011933, },
		{ 0.00008640, -0.00009162, 0.00004675, 0.00000779, -0.00003392, -0.00001941, 0.00001485, -0.00000200, -0.00000118, 0.00000371, -0.00000741, -0.00000832, -0.00001713, -0.00001018, 0.00000856, -0.00003730, 0.00005056, -0.00009389, 0.00010104, -0.00010880, 0.00004564, -0.00005687, 0.00005669, -0.00011299, 0.00001144, -0.00011432, 0.00005305, -0.00018883, -0.00012069, -0.00044887, -0.00226428, 0.00688688, -0.00277054, -0.00005009, -0.00039694, 0.00006270, -0.00013869, 0.00003817, -0.00014897, 0.00007227, -0.00007538, 0.00001415, -0.00005745, 0.00003766, -0.00007366, 0.00008456, -0.00011448, 0.00011431, -0.00008088, 0.00004929, -0.00005448, 0.00003783, -0.00003130, 0.00001053, 0.00000205, 0.00000561, -0.00001316, 0.00005585, -0.00005399, 0.00001276, -0.00001293, 0.00003262, -0.00000147, -0.00006922, },
		{ 0.00000058, -0.00006049, 0.00006326, -0.00002909, -0.00000645, -0.00005000, 0.00005961, -0.00005529, 0.00005444, -0.00005666, 0.00005650, -0.00008642, 0.00006382, -0.00008850, 0.00009031, -0.00013043, 0.00015126, -0.00021050, 0.00024700, -0.00030407, 0.00027509, -0.00029335, 0.00031762, -0.00045236, 0.00040784, -0.00056878, 0.00060942, -0.00098829, 0.00094654, -0.00264402, 0.00175256, 0.00436308, -0.00332635, 0.00091820, -0.00097941, 0.00057543, -0.00053931, 0.00035359, -0.00040305, 0.00032412, -0.00030882, 0.00020671, -0.00021187, 0.00017169, -0.00018688, 0.00018043, -0.00020180, 0.00021640, -0.00019044, 0.00015149, -0.00014344, 0.00012634, -0.00011581, 0.00009016, -0.00005973, 0.00004622, -0.00002952, 0.00007167, -0.00007631, 0.00003537, -0.00003389, 0.00008291, -0.00010047, 0.00004419, },
		{ -0.00005979, -0.00000059, 0.00002788, -0.00001176, -0.00003643, -0.00001297, 0.00003029, -0.00003084, 0.00002607, -0.00002182, 0.00000933, -0.00004073, 0.00001439, -0.00002322, 0.00000573, -0.00003460, 0.00004159, -0.00008736, 0.00011397, -0.00018625, 0.00017270, -0.00016547, 0.00014332, -0.00028263, 0.00023220, -0.00035246, 0.00031454, -0.00069251, 0.00046088, -0.00276982, 0.00461422, 0.00078614, -0.00191919, 0.00045462, -0.00057159, 0.00030255, -0.00031879, 0.00015672, -0.00020068, 0.00015917, -0.00018367, 0.00009418, -0.00009212, 0.00005271, -0.00006960, 0.00006306, -0.00008229, 0.00011856, -0.00011778, 0.00009328, -0.00008547, 0.00007758, -0.00007685, 0.00006531, -0.00003814, 0.00001951, 0.00002003, 0.00001679, -0.00002502, -0.00002607, 0.00005552, -0.00000878, -0.00005927, 0.00006384, },
		{ -0.00002543, -0.00000098, 0.00002491, -0.00001670, -0.00004974, 0.00001897, -0.00000199, -0.00000290, -0.00000282, 0.00001200, -0.00004237, 0.00001665, -0.00004479, 0.00004879, -0.00009197, 0.00008513, -0.00009951, 0.00007608, -0.00007960, 0.00001574, -0.00001544, 0.00004000, -0.00013491, 0.00003744, -0.00012189, 0.00007027, -0.00028159, 0.00006674, -0.00100563, -0.00062701, 0.00462887, -0.00132146, -0.00059695, -0.00015092, -0.00006658, -0.00007207, -0.00002081, -0.00008717, 0.00004238, -0.00006173, -0.00000583, -0.00005149, 0.00004942, -0.00008741, 0.00006650, -0.00006952, 0.00005730, -0.00000893, -0.00001379, 0.00001214, -0.00001036, 0.00001148, -0.00002052, 0.00002852, -0.00001841, 0.00001449, 0.00003657, -0.00000067, -0.00002128, -0.00003185, 0.00010414, -0.00011294, 0.00005567, -0.00002216, },
		{ 0.00002755, -0.00003356, 0.00005376, -0.00005887, -0.00001780, 0.00000700, 0.00000580, -0.00001695, 0.00001935, -0.00001939, -0.00001924, -0.00000367, -0.00001622, 0.00001438, -0.00006867, 0.00007687, -0.00010416, 0.00008924, -0.00011543, 0.00006437, -0.00003582, 0.00002814, -0.00015507, 0.00008180, -0.00016016, 0.00008295, -0.00035632, 0.00010161, -0.00167329, 0.00181170, 0.00279248, -0.00162619, -0.00015507, -0.00025086, 0.00003146, -0.00013259, 0.00001726, -0.00009002, 0.00005857, -0.00009260, 0.00001334, -0.00004721, 0.00004400, -0.00009185, 0.00007913, -0.00008619, 0.00009269, -0.00005124, 0.00002226, -0.00000699, 0.00000467, -0.00000092, -0.00000749, 0.00002457, -0.00002734, 0.00004704, 0.00000659, 0.00003285, -0.00008622, 0.00006659, 0.00001271, -0.00008022, 0.00008089, -0.00008193, },
		{ 0.00001134, -0.00002154, 0.00004934, -0.00008060, 0.00001003, -0.00000729, 0.00001350, -0.00002744, 0.00003881, -0.00005803, 0.00002257, -0.00004493, 0.00003631, -0.00005830, 0.00001057, 0.00000287, -0.00003794, 0.00002024, -0.00005770, 0.00002806, 0.00001890, -0.00008874, -0.00002409, -0.00004713, -0.00000466, -0.00017772, -0.00002463, -0.00065228, -0.00076335, 0.00275356, 0.00105217, -0.00127282, -0.00009959, -0.00015387, -0.00003704, -0.00006730, -0.00003540, -0.00001200, -0.00001589, -0.00004184, -0.00002820, 0.00001070, -0.00002020, -0.00003722, 0.00003760, -0.00005064, 0.00007917, -0.00005569, 0.00003679, -0.00001416, 0.00000965, -0.00000527, 0.00000411, 0.00000954, -0.00000955, 0.00004718, 0.00001234, 0.00001815, -0.00009937, 0.00013803, -0.00011162, 0.00004555, -0.00001295, -0.00003802, },
		{ -0.00004740, 0.00003437, -0.00000294, -0.00005539, 0.00000128, 0.00000844, -0.00001136, 0.00000353, 0.00000526, -0.00003953, 0.00001079, -0.00002920, 0.00002213, -0.00006260, 0.00002960, -0.00001595, -0.00003180, 0.00001478, -0.00006004, 0.00006856, -0.00004712, -0.00006090, -0.00003419, -0.00001752, -0.00006938, -0.00015638, -0.00007843, -0.00107535, 0.00071196, 0.00226614, 0.00020536, -0.00105274, 0.00004562, -0.00020209, 0.00001133, -0.00010434, 0.00001817, -0.00004136, -0.00000646, -0.00005716, -0.00000217, 0.00000091, -0.00002992, -0.00002324, 0.00002910, -0.00004025, 0.00008304, -0.00007514, 0.00007334, -0.00005081, 0.00004480, -0.00003603, 0.00003823, -0.00002899, 0.00004297, 0.00000426, 0.00006867, -0.00007342, 0.00000151, 0.00006810, -0.00010870, 0.00009750, -0.00009619, 0.00003682, },
		{ -0.00006520, 0.00006719, -0.00004981, -0.00002333, -0.00001298, 0.00002412, -0.00003400, 0.00003569, -0.00004123, 0.00000262, -0.00002471, 0.00001110, -0.00002726, -0.00002214, 0.00000690, -0.00000192, -0.00005840, 0.00004619, -0.00008940, 0.00013150, -0.00016911, 0.00006306, -0.00015030, 0.00012642, -0.00031091, 0.00014701, -0.00069757, -0.00042909, 0.00139027, 0.00153999, -0.00020768, -0.00085755, 0.00015822, -0.00027419, 0.00008100, -0.00015212, 0.00008684, -0.00010398, 0.00003331, -0.00008946, 0.00004128, -0.00003354, -0.00001913, -0.00001877, 0.00002165, -0.00002248, 0.00007034, -0.00006977, 0.00008520, -0.00006919, 0.00006601, -0.00005193, 0.00005090, -0.00003606, 0.00006220, -0.00000224, 0.00007973, -0.00012978, 0.00010317, -0.00006226, 0.00000673, 0.00002075, -0.00007305, 0.00004322, },
		{ -0.00002678, 0.00004904, -0.00005742, -0.00001759, -0.00000302, 0.00001030, -0.00001951, 0.00002286, -0.00004334, 0.00000752, -0.00002184, 0.00000595, -0.00003209, -0.00001891, 0.00002315, -0.00004245, -0.00001635, 0.00000386, -0.00002446, 0.00006147, -0.00014476, 0.00005107, -0.00011506, 0.00006390, -0.00029905, 0.00013342, -0.00108651, 0.00068257, 0.00111790, 0.00122354, -0.00065021, -0.00046024, 0.00003443, -0.00017711, 0.00001320, -0.00007024, 0.00003191, -0.00007064, -0.00000319, -0.00004741, 0.00001608, -0.00001555, -0.00005065, 0.00002742, -0.00002909, 0.00004210, 0.00000638, -0.00000589, 0.00003480, -0.00002946, 0.00003705, -0.00002506, 0.00002347, 0.00000460, 0.00002769, 0.00005529, -0.00000130, -0.00006892, 0.00007796, -0.00009068, 0.00007652, -0.00007091, 0.00000175, -0.00000946, },
		{ 0.00001211, 0.00001742, -0.00005014, -0.00001929, 0.00001098, -0.00000890, 0.00000475, -0.00001009, -0.00001815, -0.00001122, 0.00000247, -0.00002760, -0.00000396, -0.00004208, 0.00005896, -0.00011054, 0.00006834, -0.00008329, 0.00009216, -0.00009851, 0.00000660, -0.00009496, 0.00006566, -0.00020940, 0.00002627, -0.00042944, -0.00056610, 0.00117233, 0.00073502, 0.00101776, -0.00097378, -0.00005153, -0.00017841, -0.00001124, -0.00010247, 0.00005586, -0.00007936, 0.00001094, -0.00007146, 0.00002315, -0.00003747, 0.00001928, -0.00008640, 0.00007179, -0.00007563, 0.00010444, -0.00005755, 0.00006385, -0.00002936, 0.00002699, -0.00000536, 0.00000614, 0.00000459, 0.00003187, 0.00001180, 0.00009063, -0.00008059, 0.00002425, -0.00002511, -0.00000334, 0.00002538, -0.00007136, 0.00002728, -0.00004291, },
		{ 0.00001904, 0.00000303, -0.00005371, -0.00000548, 0.00000450, -0.00000344, -0.00000091, -0.00001496, -0.00001489, -0.00000466, -0.00000787, -0.00002423, -0.00001096, -0.00002181, 0.00003265, -0.00010581, 0.00008321, -0.00009132, 0.00010566, -0.00015515, 0.00007186, -0.00014500, 0.00010931, -0.00032274, 0.00017169, -0.00093405, 0.00047854, 0.00080883, 0.00079039, 0.00054578, -0.00091978, 0.00009468, -0.00024086, 0.00003659, -0.00010556, 0.00007291, -0.00010745, 0.00002616, -0.00007318, 0.00003068, -0.00003777, 0.00000155, -0.00006419, 0.00005379, -0.00005676, 0.00010428, -0.00006193, 0.00007862, -0.00004829, 0.00004880, -0.00002102, 0.00001228, 0.00002081, 0.00001426, 0.00005612, 0.00004010, -0.00006384, 0.00002174, -0.00005479, 0.00006028, -0.00005471, -0.00001487, -0.00000347, -0.00003063, },
		{ 0.00001776, -0.00000930, -0.00005401, 0.00000825, -0.00000522, 0.00000805, -0.00002114, 0.00000049, -0.00002831, 0.00001801, -0.00004355, 0.00001242, -0.00005111, 0.00003466, -0.00004763, -0.00002689, 0.00002122, -0.00001817, 0.00000656, -0.00007064, -0.00000714, -0.00003853, -0.00006533, -0.00012806, -0.00017594, -0.00070189, 0.00109474, 0.00027803, 0.00102874, -0.00004792, -0.00061187, 0.00002378, -0.00016603, -0.00001302, -0.00002347, -0.00000201, -0.00005715, -0.00001715, -0.00002228, -0.00001152, 0.00000159, -0.00005004, -0.00000791, -0.00000117, 0.00000507, 0.00006164, -0.00002570, 0.00005335, -0.00003211, 0.00004531, -0.00002517, 0.00002255, 0.00002778, 0.00000842, 0.00009317, -0.00002924, -0.00000454, -0.00005459, 0.00002140, 0.00001562, -0.00005781, 0.00000508, -0.00002417, -0.00001392, },
		{ 0.00002949, -0.00003636, -0.00003377, 0.00000185, 0.00000593, -0.00000645, -0.00001667, -0.00000348, -0.00001859, 0.00000894, -0.00004675, 0.00002089, -0.00006024, 0.00005181, -0.00009145, 0.00003051, -0.00001843, 0.00001529, -0.00005640, -0.00000343, -0.00006430, 0.00001952, -0.00018938, 0.00002704, -0.00062141, 0.00011487, 0.00091291, 0.00017137, 0.00101443, -0.00042811, -0.00034555, -0.00005142, -0.00010745, -0.00002709, 0.00001373, -0.00005027, -0.00002475, -0.00003563, 0.00000180, -0.00003049, 0.00001511, -0.00007334, 0.00002012, -0.00003035, 0.00005146, 0.00002906, 0.00000176, 0.00003258, -0.00001655, 0.00004194, -0.00003515, 0.00005544, -0.00000265, 0.00005764, 0.00005563, -0.00002587, -0.00001790, -0.00007116, 0.00008254, -0.00006264, -0.00000526, -0.00002528, -0.00000770, -0.00001985, },
		{ 0.00003630, -0.00006042, -0.00000939, -0.00000921, 0.00002171, -0.00003413, 0.00000766, -0.00002519, 0.00001019, -0.00003020, -0.00001030, -0.00001148, -0.00002607, 0.00001318, -0.00006980, 0.00003159, -0.00000681, -0.00002629, -0.00002146, -0.00003425, -0.00001510, -0.00007837, -0.00009010, -0.00017410, -0.00056263, 0.00075193, 0.00040465, 0.00039982, 0.00072862, -0.00054409, -0.00020432, -0.00007032, -0.00009353, 0.00000806, -0.00001370, -0.00004433, -0.00003074, -0.00001717, -0.00001214, -0.00001599, -0.00000384, -0.00006462, 0.00001585, -0.00002597, 0.00007443, 0.00001050, 0.00001800, 0.00001830, 0.00000045, 0.00002710, -0.00002367, 0.00006901, -0.00001908, 0.00010593, -0.00000895, 0.00001781, -0.00009716, 0.00001382, 0.00003949, -0.00006933, 0.00001379, -0.00004471, 0.00000869, -0.00002584, },
		{ 0.00001833, -0.00005987, -0.00000328, -0.00000116, 0.00001166, -0.00003798, 0.00001524, -0.00002883, 0.00001547, -0.00004914, 0.00001525, -0.00003589, -0.00000278, -0.00001972, -0.00004229, 0.00003471, -0.00002068, -0.00004229, 0.00000588, -0.00005610, 0.00000568, -0.00015235, 0.00000499, -0.00049762, 0.00000507, 0.00077751, 0.00012897, 0.00058939, 0.00041169, -0.00055711, -0.00009737, -0.00010315, -0.00004984, 0.00000989, -0.00002945, -0.00004274, -0.00002465, -0.00001198, -0.00001749, -0.00000926, -0.00001721, -0.00006080, 0.00001452, -0.00001619, 0.00009580, -0.00001559, 0.00004391, -0.00000607, 0.00003043, -0.00000844, 0.00002524, 0.00003321, 0.00002648, 0.00008257, -0.00001369, -0.00000101, -0.00012264, 0.00009367, -0.00004767, -0.00001391, -0.00002206, -0.00002185, -0.00000387, -0.00000948, },
		{ -0.00001116, -0.00004430, -0.00000681, 0.00001541, -0.00001705, -0.00001637, 0.00000058, -0.00001047, -0.00001016, -0.00003084, 0.00000754, -0.00003166, -0.00001240, -0.00001723, -0.00003840, 0.00005950, -0.00008498, 0.00001846, -0.00004402, 0.00000403, -0.00009550, -0.00005160, -0.00017385, -0.00045702, 0.00051543, 0.00046031, 0.00014111, 0.00062271, 0.00016013, -0.00053227, -0.00000545, -0.00014933, 0.00002200, -0.00003287, -0.00001271, -0.00005983, 0.00000109, -0.00002821, -0.00000588, -0.00001609, -0.00002241, -0.00006329, 0.00001849, 0.00000132, 0.00010494, -0.00003617, 0.00006695, -0.00002526, 0.00005074, -0.00002998, 0.00006730, -0.00000419, 0.00008889, 0.00001757, 0.00002627, -0.00008752, -0.00004355, 0.00007336, -0.00007338, 0.00001772, -0.00005563, 0.00001005, -0.00002611, 0.00001412, },
		{ -0.00002926, -0.00003426, 0.00000050, 0.00001316, -0.00003115, -0.00000045, -0.00000783, -0.00000337, -0.00002837, -0.00001073, -0.00000271, -0.00003276, -0.00001455, -0.00001913, -0.00001583, 0.00004197, -0.00011209, 0.00006687, -0.00008580, 0.00004079, -0.00017972, 0.00005512, -0.00047192, 0.00000944, 0.00057074, 0.00024417, 0.00020166, 0.00060917, -0.00008589, -0.00042205, 0.00000733, -0.00013551, 0.00004500, -0.00005344, -0.00001176, -0.00005433, 0.00000528, -0.00002867, -0.00000872, -0.00000880, -0.00004623, -0.00004625, 0.00000727, 0.00004583, 0.00007634, -0.00002242, 0.00006087, -0.00001656, 0.00003950, -0.00000825, 0.00005999, 0.00001519, 0.00009191, -0.00000616, 0.00001967, -0.00013586, 0.00005001, -0.00000562, -0.00003113, -0.00000772, -0.00004653, 0.00001556, -0.00003107, 0.00002164, },
		{ -0.00003647, -0.00002760, 0.00001314, -0.00000648, -0.00002278, -0.00000139, -0.00000133, -0.00001872, -0.00001985, -0.00001081, 0.00000151, -0.00005459, 0.00001153, -0.00004554, 0.00003889, -0.00004251, -0.00004671, 0.00002810, -0.00004224, -0.00004069, -0.00010143, -0.00007758, -0.00047690, 0.00045975, 0.00033669, 0.00021467, 0.00022022, 0.00057972, -0.00031799, -0.00024774, -0.00005080, -0.00005982, 0.00000402, -0.00003012, -0.00003978, -0.00001993, -0.00001857, -0.00000772, -0.00002765, 0.00000938, -0.00008330, -0.00001517, -0.00000760, 0.00009982, 0.00002397, 0.00001725, 0.00003411, 0.00000631, 0.00001982, 0.00002713, 0.00003368, 0.00006696, 0.00004187, 0.00001689, -0.00005257, -0.00008612, 0.00006039, -0.00004324, 0.00000311, -0.00004422, -0.00001595, 0.00000177, -0.00002255, 0.00001579, },
		{ -0.00004462, -0.00001101, 0.00000975, -0.00001711, -0.00001452, -0.00000016, -0.00000328, -0.00002860, -0.00000925, -0.00001029, -0.00000614, -0.00006161, 0.00002808, -0.00005392, 0.00006253, -0.00011316, 0.00003744, -0.00003531, 0.00001109, -0.00013986, 0.00002204, -0.00035975, -0.00008477, 0.00052530, 0.00017909, 0.00019712, 0.00028067, 0.00045390, -0.00043866, -0.00011270, -0.00009086, -0.00000207, -0.00003583, -0.00000980, -0.00005536, 0.00000172, -0.00003261, 0.00000156, -0.00003153, 0.00000591, -0.00010037, -0.00000013, 0.00000688, 0.00012225, -0.00001112, 0.00004898, 0.00001420, 0.00001799, 0.00002305, 0.00003357, 0.00004483, 0.00008064, 0.00000779, 0.00001571, -0.00010624, -0.00000573, 0.00000809, -0.00001868, -0.00001184, -0.00004782, 0.00000147, -0.00000912, -0.00001675, 0.00001099, },
		{ -0.00004960, 0.00000898, -0.00000893, -0.00000997, -0.00001740, 0.00001084, -0.00002345, -0.00001561, -0.00001580, 0.00000498, -0.00003995, -0.00003106, 0.00000889, -0.00001917, 0.00001389, -0.00009766, 0.00005274, -0.00003844, -0.00002380, -0.00011464, -0.00003449, -0.00043902, 0.00035637, 0.00030408, 0.00019250, 0.00012150, 0.00039856, 0.00022423, -0.00042011, -0.00006050, -0.00007437, -0.00000205, -0.00003573, -0.00001937, -0.00003940, -0.00000749, -0.00002096, -0.00001201, -0.00001362, -0.00002527, -0.00009075, -0.00000291, 0.00005136, 0.00010376, -0.00001381, 0.00006093, 0.00000484, 0.00002547, 0.00003535, 0.00002572, 0.00008637, 0.00004310, 0.00001054, -0.00003557, -0.00008154, 0.00001528, -0.00002081, 0.00000703, -0.00004355, -0.00002724, 0.00000279, -0.00001498, -0.00001314, 0.00000614, },
		{ -0.00003897, 0.00001213, -0.00001955, -0.00000317, -0.00001699, 0.00001264, -0.00003850, 0.00000035, -0.00002286, 0.00001214, -0.00006875, 0.00000625, -0.00001417, 0.00001252, -0.00005261, -0.00003511, 0.00002294, -0.00001516, -0.00009366, -0.00002515, -0.00024690, -0.00016760, 0.00049116, 0.00011860, 0.00022393, 0.00008066, 0.00047191, -0.00000701, -0.00033878, -0.00004156, -0.00004752, -0.00002092, -0.00002264, -0.00003292, -0.00002007, -0.00002141, -0.00000720, -0.00002478, -0.00000002, -0.00005780, -0.00007820, 0.00000445, 0.00009309, 0.00007051, 0.00000062, 0.00006285, -0.00000404, 0.00004687, 0.00002604, 0.00004923, 0.00010052, 0.00000304, 0.00000519, -0.00007770, -0.00002432, -0.00001801, 0.00000147, -0.00000860, -0.00005013, -0.00001133, 0.00000192, -0.00002422, -0.00000279, -0.00000779, },
		{ -0.00001769, -0.00000355, -0.00001138, -0.00000862, -0.00000510, -0.00000547, -0.00002975, -0.00000133, -0.00001391, -0.00000759, -0.00006385, 0.00001806, -0.00001013, -0.00000407, -0.00006674, -0.00000255, 0.00001620, -0.00004262, -0.00008394, -0.00005236, -0.00034158, 0.00021051, 0.00033820, 0.00009965, 0.00017771, 0.00013073, 0.00043720, -0.00016329, -0.00025821, -0.00001666, -0.00004769, -0.00002147, -0.00002377, -0.00002902, -0.00001865, -0.00002044, -0.00000729, -0.00002356, -0.00000454, -0.00007759, -0.00007260, 0.00003411, 0.00011123, 0.00004326, 0.00002075, 0.00005550, -0.00000311, 0.00006787, 0.00000786, 0.00010346, 0.00006354, -0.00001263, -0.00002986, -0.00006226, -0.00000884, -0.00003654, 0.00002779, -0.00004381, -0.00003188, -0.00000755, 0.00000073, -0.00003335, 0.00001034, -0.00002577, },
		{ -0.00000536, -0.00001764, 0.00000126, -0.00001577, 0.00000346, -0.00002523, -0.00001088, -0.00001115, -0.00000313, -0.00003495, -0.00004054, 0.00001802, -0.00000364, -0.00004021, -0.00004048, -0.00000105, 0.00002236, -0.00009446, -0.00002170, -0.00020721, -0.00016110, 0.00038130, 0.00015487, 0.00014366, 0.00011014, 0.00021440, 0.00033334, -0.00025381, -0.00017540, -0.00000542, -0.00005510, -0.00001471, -0.00002779, -0.00002150, -0.00002271, -0.00001525, -0.00001017, -0.00002131, -0.00001474, -0.00009549, -0.00005866, 0.00007016, 0.00011159, 0.00002436, 0.00004488, 0.00003513, 0.00002214, 0.00006055, 0.00002341, 0.00013835, 0.00000550, -0.00002391, -0.00005613, -0.00002126, -0.00003690, -0.00001154, 0.00001701, -0.00005940, -0.00001289, -0.00000686, -0.00000723, -0.00003209, 0.00001445, -0.00003349, },
		{ -0.00000940, -0.00001578, 0.00000405, -0.00001447, -0.00000380, -0.00002668, -0.00000293, -0.00001271, -0.00000852, -0.00004439, -0.00002450, 0.00002804, -0.00002822, -0.00004239, -0.00002921, 0.00001452, -0.00001187, -0.00008575, -0.00004338, -0.00028772, 0.00013645, 0.00030600, 0.00009009, 0.00014781, 0.00008617, 0.00026882, 0.00021212, -0.00030037, -0.00008511, -0.00002610, -0.00004483, -0.00001861, -0.00002071, -0.00002474, -0.00001913, -0.00001599, -0.00000780, -0.00002652, -0.00002263, -0.00011676, -0.00002686, 0.00009789, 0.00010281, 0.00001612, 0.00006336, 0.00001640, 0.00005693, 0.00003375, 0.00008130, 0.00011836, -0.00004377, -0.00004132, -0.00004266, -0.00001105, -0.00005516, 0.00002192, -0.00001914, -0.00004724, -0.00000560, -0.00000566, -0.00002010, -0.00002044, 0.00000840, -0.00002920, },
		{ -0.00001684, -0.00000523, 0.00000070, -0.00001260, -0.00001605, -0.00001533, -0.00000496, -0.00001095, -0.00002117, -0.00004101, -0.00001041, 0.00003107, -0.00006332, -0.00001559, -0.00003849, 0.00003915, -0.00006955, -0.00003033, -0.00017020, -0.00015247, 0.00029616, 0.00016765, 0.00009785, 0.00012253, 0.00009440, 0.00029258, 0.00008145, -0.00029181, -0.00001920, -0.00005487, -0.00002716, -0.00002437, -0.00001288, -0.00002994, -0.00001512, -0.00001487, -0.00000982, -0.00002970, -0.00003744, -0.00012887, 0.00001329, 0.00011790, 0.00008599, 0.00002322, 0.00006488, 0.00001957, 0.00007074, 0.00002862, 0.00014140, 0.00005080, -0.00007793, -0.00004306, -0.00001485, -0.00003425, -0.00003707, 0.00002574, -0.00004585, -0.00002522, -0.00000659, -0.00000749, -0.00002901, -0.00000758, 0.00000135, -0.00002549, },
		{ -0.00001437, 0.00000126, -0.00000058, -0.00002041, -0.00001614, -0.00000835, -0.00000552, -0.00001770, -0.00002373, -0.00004101, 0.00001527, 0.00000081, -0.00006397, -0.00000551, -0.00003205, 0.00002995, -0.00008335, -0.00003378, -0.00025027, 0.00009468, 0.00025677, 0.00010155, 0.00009963, 0.00010600, 0.00010771, 0.00029553, -0.00005142, -0.00022720, -0.00000262, -0.00006320, -0.00002013, -0.00001949, -0.00001635, -0.00002666, -0.00001810, -0.00000625, -0.00002264, -0.00002434, -0.00006495, -0.00012210, 0.00005158, 0.00013238, 0.00006547, 0.00004015, 0.00005547, 0.00004108, 0.00005972, 0.00006829, 0.00015620, -0.00003734, -0.00008848, -0.00002026, -0.00001109, -0.00005166, -0.00000633, 0.00000153, -0.00004673, -0.00001054, -0.00000985, -0.00001342, -0.00002962, 0.00000121, -0.00000277, -0.00002986, },
		{ -0.00000445, 0.00000279, -0.00000340, -0.00003143, -0.00000486, -0.00000936, -0.00000543, -0.00002754, -0.00001957, -0.00003901, 0.00003639, -0.00004609, -0.00002965, -0.00002406, -0.00000517, -0.00000966, -0.00005346, -0.00012446, -0.00015914, 0.00024864, 0.00014667, 0.00009480, 0.00008886, 0.00009457, 0.00013179, 0.00026130, -0.00015036, -0.00014531, -0.00001815, -0.00005382, -0.00002020, -0.00001066, -0.00002491, -0.00002038, -0.00002119, 0.00000081, -0.00003733, -0.00001862, -0.00009619, -0.00010145, 0.00008903, 0.00013721, 0.00005031, 0.00005534, 0.00005119, 0.00005872, 0.00005255, 0.00013306, 0.00010260, -0.00011232, -0.00006482, 0.00000128, -0.00003012, -0.00004186, 0.00000499, -0.00002364, -0.00003065, -0.00000775, -0.00001313, -0.00001986, -0.00002446, 0.00000761, -0.00000982, -0.00003473, },
		{ 0.00000236, 0.00000510, -0.00001564, -0.00003100, 0.00000398, -0.00001096, -0.00001218, -0.00002760, -0.00002172, -0.00002185, 0.00002613, -0.00006653, -0.00000132, -0.00004207, 0.00001163, -0.00003347, -0.00005402, -0.00019393, 0.00004034, 0.00023760, 0.00008302, 0.00009125, 0.00008501, 0.00007701, 0.00017123, 0.00018241, -0.00018862, -0.00008744, -0.00003397, -0.00004380, -0.00001571, -0.00000982, -0.00002765, -0.00001964, -0.00001654, -0.00000320, -0.00004377, -0.00002189, -0.00012072, -0.00007517, 0.00012985, 0.00012813, 0.00004620, 0.00006413, 0.00005776, 0.00006075, 0.00007966, 0.00017367, -0.00000761, -0.00013669, -0.00002378, 0.00000021, -0.00004521, -0.00001841, -0.00000968, -0.00002801, -0.00001664, -0.00001103, -0.00001733, -0.00002249, -0.00001753, 0.00001271, -0.00002424, -0.00002918, },
		{ 0.00000257, 0.00000711, -0.00003390, -0.00001610, 0.00000321, -0.00001116, -0.00002268, -0.00001932, -0.00002748, 0.00000068, -0.00000950, -0.00004770, -0.00000346, -0.00004133, 0.00000932, -0.00003396, -0.00011244, -0.00013908, 0.00018810, 0.00014738, 0.00007296, 0.00007906, 0.00008263, 0.00006729, 0.00020247, 0.00008653, -0.00018034, -0.00006006, -0.00003897, -0.00003766, -0.00000710, -0.00001654, -0.00002462, -0.00002264, -0.00000755, -0.00001610, -0.00004250, -0.00003320, -0.00013877, -0.00004225, 0.00017169, 0.00010585, 0.00005429, 0.00006809, 0.00006879, 0.00005814, 0.00014497, 0.00013683, -0.00011889, -0.00010222, 0.00000154, -0.00001491, -0.00004212, -0.00000548, -0.00003216, -0.00001145, -0.00001665, -0.00001288, -0.00002326, -0.00001955, -0.00001057, 0.00001251, -0.00003989, -0.00001348, },
		},
		{ { -0.00000794, -0.00000720, 0.00002317, -0.00000751, -0.00000480, 0.00000180, -0.00000210, -0.00000313, 0.00001429, -0.00000873, -0.00001038, 0.00000409, 0.00000273, 0.00000410, -0.00005556, 0.00000199, -0.00001208, -0.00002180, 0.00000284, -0.00001737, -0.00002230, -0.00003152, -0.00005422, -0.00004144, -0.00018336, -0.00001105, 0.00023251, 0.00004752, 0.00003200, 0.00001414, -0.00012100, 0.00010662, 0.00038548, 0.00010662, -0.00012100, 0.00001414, 0.00003200, 0.00004752, 0.00023251, -0.00001105, -0.00018336, -0.00004144, -0.00005422, -0.00003152, -0.00002230, -0.00001737, 0.00000284, -0.00002180, -0.00001208, 0.00000199, -0.00005556, 0.00000410, 0.00000273, 0.00000409, -0.00001038, -0.00000873, 0.00001429, -0.00000313, -0.00000210, 0.00000180, -0.00000480, -0.00000751, 0.00002317, -0.00000720, },
		{ -0.00000853, -0.00000265, 0.00002400, -0.00001297, -0.00000060, 0.00000045, -0.00000303, -0.00000051, 0.00001580, -0.00001798, -0.00000570, 0.00000384, 0.00000947, -0.00000932, -0.00005176, 0.00000855, -0.00001677, -0.00001950, 0.00000082, -0.00001702, -0.00002156, -0.00003505, -0.00004689, -0.00006940, -0.00016180, 0.00005217, 0.00019577, 0.00003247, 0.00002596, 0.00000971, -0.00014015, 0.00009519, 0.00038775, 0.00011412, -0.00009890, 0.00001962, 0.00003643, 0.00007386, 0.00025598, -0.00009116, -0.00017784, -0.00002449, -0.00005923, -0.00002812, -0.00002349, -0.00001696, 0.00000367, -0.00002269, -0.00000719, -0.00000908, -0.00005203, 0.00001353, -0.00000487, 0.00000569, -0.00001513, 0.00000253, 0.00000899, -0.00000357, -0.00000142, 0.00000255, -0.00000842, -0.00000114, 0.00001927, -0.00000921, },
		{ -0.00000999, 0.00000431, 0.00002092, -0.00001613, 0.00000329, -0.00000140, -0.00000358, 0.00000344, 0.00001348, -0.00002412, -0.00000226, 0.00000664, 0.00001179, -0.00002237, -0.00004325, 0.00001087, -0.00002093, -0.00001529, -0.00000393, -0.00001337, -0.00002497, -0.00003336, -0.00004522, -0.00009773, -0.00011963, 0.00009309, 0.00015508, 0.00002573, 0.00001781, 0.00000727, -0.00015740, 0.00008021, 0.00039386, 0.00011868, -0.00007393, 0.00002584, 0.00003991, 0.00011411, 0.00025381, -0.00017457, -0.00014778, -0.00002362, -0.00005631, -0.00002938, -0.00002149, -0.00001868, 0.00000543, -0.00002314, -0.00000261, -0.00002379, -0.00004012, 0.00001538, -0.00000922, 0.00000570, -0.00001756, 0.00001347, 0.00000119, -0.00000181, -0.00000141, 0.00000284, -0.00001092, 0.00000488, 0.00001351, -0.00000916, },
		{ -0.00001147, 0.00001277, 0.00001382, -0.00001587, 0.00000573, -0.00000325, -0.00000333, 0.00000749, 0.00000829, -0.00002734, 0.00000032, 0.00001218, 0.00000813, -0.00003162, -0.00003279, 0.00000943, -0.00002361, -0.00001014, -0.00001098, -0.00000624, -0.00003364, -0.00002379, -0.00005519, -0.00011511, -0.00006859, 0.00011355, 0.00011556, 0.00002521, 0.00000693, 0.00000810, -0.00017443, 0.00006287, 0.00040235, 0.00012213, -0.00004678, 0.00003241, 0.00004467, 0.00016771, 0.00021462, -0.00024241, -0.00010471, -0.00003603, -0.00004523, -0.00003616, -0.00001522, -0.00002364, 0.00000907, -0.00002340, 0.00000029, -0.00003945, -0.00002148, 0.00000851, -0.00000752, 0.00000147, -0.00001515, 0.00002118, -0.00000684, 0.00000123, -0.00000202, 0.00000279, -0.00001212, 0.00000956, 0.00000719, -0.00000792, },
		{ -0.00001174, 0.00002107, 0.00000362, -0.00001178, 0.00000576, -0.00000431, -0.00000227, 0.00001057, 0.00000158, -0.00002850, 0.00000341, 0.00001830, -0.00000066, -0.00003554, -0.00002242, 0.00000458, -0.00002329, -0.00000645, -0.00001761, 0.00000153, -0.00004474, -0.00000857, -0.00007716, -0.00011458, -0.00002038, 0.00011916, 0.00007930, 0.00002924, -0.00000676, 0.00001279, -0.00019235, 0.00004392, 0.00041222, 0.00012590, -0.00001764, 0.00003787, 0.00005594, 0.00022758, 0.00013374, -0.00027673, -0.00006509, -0.00005258, -0.00003115, -0.00004507, -0.00000705, -0.00003025, 0.00001364, -0.00002252, -0.00000084, -0.00005170, -0.00000064, -0.00000479, 0.00000025, -0.00000781, -0.00000656, 0.00002330, -0.00001263, 0.00000404, -0.00000266, 0.00000231, -0.00001195, 0.00001226, 0.00000147, -0.00000651, },
		{ -0.00000951, 0.00002700, -0.00000771, -0.00000453, 0.00000300, -0.00000381, -0.00000086, 0.00001214, -0.00000560, -0.00002827, 0.00000829, 0.00002214, -0.00001198, -0.00003436, -0.00001339, -0.00000327, -0.00001863, -0.00000695, -0.00002008, 0.00000524, -0.00005254, 0.00000558, -0.00010497, -0.00009652, 0.00001778, 0.00011537, 0.00004749, 0.00003566, -0.00002194, 0.00002027, -0.00021061, 0.00002258, 0.00042398, 0.00013014, 0.00001439, 0.00003948, 0.00008205, 0.00027852, 0.00001811, -0.00026764, -0.00004368, -0.00006240, -0.00002163, -0.00005032, -0.00000154, -0.00003502, 0.00001672, -0.00001888, -0.00000841, -0.00005575, 0.00001610, -0.00001928, 0.00001100, -0.00002031, 0.00000738, 0.00001908, -0.00001442, 0.00000510, -0.00000246, 0.00000108, -0.00001040, 0.00001264, -0.00000275, -0.00000583, },
		{ -0.00000390, 0.00002836, -0.00001744, 0.00000412, -0.00000199, -0.00000143, 0.00000028, 0.00001218, -0.00001273, -0.00002658, 0.00001518, 0.00002154, -0.00002281, -0.00002933, -0.00000662, -0.00001293, -0.00000965, -0.00001294, -0.00001590, 0.00000101, -0.00005135, 0.00001045, -0.00012834, -0.00006801, 0.00004469, 0.00010520, 0.00002185, 0.00004134, -0.00003585, 0.00002780, -0.00022695, -0.00000350, 0.00043963, 0.00013386, 0.00005078, 0.00003466, 0.00013201, 0.00029908, -0.00011237, -0.00021840, -0.00004722, -0.00005850, -0.00002229, -0.00004720, -0.00000269, -0.00003468, 0.00001615, -0.00001160, -0.00002333, -0.00004843, 0.00002298, -0.00002870, 0.00001967, -0.00003200, 0.00002353, 0.00000987, -0.00001192, 0.00000361, -0.00000076, -0.00000125, -0.00000749, 0.00001064, -0.00000498, -0.00000636, },
		{ 0.00000504, 0.00002368, -0.00002284, 0.00001167, -0.00000769, 0.00000242, 0.00000076, 0.00001095, -0.00001979, -0.00002257, 0.00002275, 0.00001604, -0.00003098, -0.00002170, -0.00000316, -0.00002207, 0.00000146, -0.00002292, -0.00000576, -0.00001157, -0.00003895, 0.00000069, -0.00013751, -0.00003919, 0.00006366, 0.00008930, 0.00000440, 0.00004297, -0.00004532, 0.00003213, -0.00023865, -0.00003699, 0.00046137, 0.00013632, 0.00009170, 0.00002396, 0.00021056, 0.00026759, -0.00022873, -0.00014560, -0.00007129, -0.00004150, -0.00003329, -0.00003516, -0.00001134, -0.00002839, 0.00001171, -0.00000188, -0.00004352, -0.00003020, 0.00001733, -0.00002837, 0.00002136, -0.00003818, 0.00003744, -0.00000120, -0.00000645, -0.00000013, 0.00000249, -0.00000471, -0.00000346, 0.00000656, -0.00000525, -0.00000792, },
		{ 0.00001601, 0.00001296, -0.00002210, 0.00001571, -0.00001211, 0.00000663, 0.00000065, 0.00000858, -0.00002666, -0.00001515, 0.00002853, 0.00000706, -0.00003569, -0.00001235, -0.00000419, -0.00002752, 0.00001063, -0.00003257, 0.00000591, -0.00002822, -0.00001879, -0.00002285, -0.00012791, -0.00001838, 0.00007895, 0.00006787, -0.00000388, 0.00003867, -0.00004860, 0.00003139, -0.00024436, -0.00007903, 0.00048976, 0.00013891, 0.00013405, 0.00001409, 0.00031232, 0.00017089, -0.00030149, -0.00007395, -0.00010219, -0.00001972, -0.00004856, -0.00001879, -0.00002403, -0.00001875, 0.00000585, 0.00000654, -0.00006369, -0.00000586, 0.00000113, -0.00001740, 0.00001365, -0.00003545, 0.00004491, -0.00001030, -0.00000052, -0.00000475, 0.00000650, -0.00000885, 0.00000112, 0.00000112, -0.00000416, -0.00000951, },
		{ 0.00002648, -0.00000197, -0.00001506, 0.00001479, -0.00001346, 0.00000977, 0.00000054, 0.00000476, -0.00003267, -0.00000392, 0.00002994, -0.00000285, -0.00003737, -0.00000190, -0.00001040, -0.00002649, 0.00001348, -0.00003634, 0.00001246, -0.00004121, 0.00000061, -0.00005239, -0.00010258, -0.00000842, 0.00009248, 0.00004279, -0.00000449, 0.00002925, -0.00004666, 0.00002642, -0.00024541, -0.00012797, 0.00052259, 0.00014591, 0.00017094, 0.00001895, 0.00041818, 0.00001236, -0.00031121, -0.00002710, -0.00012305, -0.00000521, -0.00005849, -0.00000604, -0.00003431, -0.00001084, 0.00000284, 0.00000864, -0.00007668, 0.00001647, -0.00001914, 0.00000056, -0.00000203, -0.00002343, 0.00004366, -0.00001432, 0.00000321, -0.00000830, 0.00000986, -0.00001262, 0.00000527, -0.00000460, -0.00000271, -0.00000953, },
		{ 0.00003321, -0.00001766, -0.00000359, 0.00000910, -0.00001092, 0.00001070, 0.00000116, -0.00000114, -0.00003633, 0.00001011, 0.00002556, -0.00001118, -0.00003681, 0.00000867, -0.00002103, -0.00001818, 0.00000750, -0.00003022, 0.00000819, -0.00004275, 0.00000906, -0.00007602, -0.00007102, -0.00000597, 0.00010245, 0.00001836, -0.00000160, 0.00001826, -0.00004315, 0.00002064, -0.00024565, -0.00017976, 0.00055557, 0.00016336, 0.00019386, 0.00005668, 0.00049644, -0.00018430, -0.00025680, -0.00001831, -0.00012169, -0.00000767, -0.00005471, -0.00000426, -0.00003589, -0.00000964, 0.00000637, 0.00000052, -0.00007626, 0.00002821, -0.00003472, 0.00001842, -0.00002062, -0.00000515, 0.00003430, -0.00001202, 0.00000298, -0.00000910, 0.00001104, -0.00001477, 0.00000787, -0.00000941, -0.00000197, -0.00000627, },
		{ 0.00003321, -0.00002990, 0.00000888, 0.00000057, -0.00000510, 0.00000910, 0.00000288, -0.00000971, -0.00003562, 0.00002437, 0.00001588, -0.00001658, -0.00003435, 0.00001751, -0.00003330, -0.00000483, -0.00000609, -0.00001422, -0.00000843, -0.00002911, -0.00000045, -0.00008274, -0.00004489, -0.00000410, 0.00010472, 0.00000011, -0.00000051, 0.00001029, -0.00004257, 0.00001819, -0.00024936, -0.00023012, 0.00058480, 0.00019572, 0.00019757, 0.00014252, 0.00051022, -0.00037856, -0.00015775, -0.00004484, -0.00009674, -0.00002907, -0.00003485, -0.00001613, -0.00002615, -0.00001709, 0.00001707, -0.00001802, -0.00006019, 0.00002383, -0.00003804, 0.00002847, -0.00003522, 0.00001379, 0.00002020, -0.00000450, -0.00000138, -0.00000645, 0.00000905, -0.00001428, 0.00000794, -0.00001236, -0.00000255, 0.00000131, },
		{ 0.00002495, -0.00003502, 0.00001838, -0.00000772, 0.00000210, 0.00000578, 0.00000530, -0.00002080, -0.00002879, 0.00003532, 0.00000331, -0.00001926, -0.00002960, 0.00002202, -0.00004274, 0.00000837, -0.00002208, 0.00000661, -0.00003293, -0.00000336, -0.00002773, -0.00006776, -0.00003262, 0.00000335, 0.00009586, -0.00000787, -0.00000507, 0.00000830, -0.00004761, 0.00002112, -0.00025840, -0.00027762, 0.00061009, 0.00024199, 0.00018599, 0.00027938, 0.00042983, -0.00052382, -0.00004870, -0.00008876, -0.00005885, -0.00006163, -0.00000465, -0.00003751, -0.00000815, -0.00003038, 0.00003112, -0.00004207, -0.00003222, 0.00000367, -0.00002599, 0.00002564, -0.00003990, 0.00002736, 0.00000623, 0.00000529, -0.00000835, -0.00000102, 0.00000393, -0.00001087, 0.00000513, -0.00001311, -0.00000424, 0.00001290, },
		{ 0.00000917, -0.00003124, 0.00002179, -0.00001260, 0.00000809, 0.00000239, 0.00000708, -0.00003310, -0.00001534, 0.00003970, -0.00000879, -0.00002053, -0.00002190, 0.00001985, -0.00004469, 0.00001503, -0.00003290, 0.00002365, -0.00005584, 0.00002468, -0.00006419, -0.00003525, -0.00003534, 0.00001784, 0.00007627, -0.00000540, -0.00001531, 0.00001147, -0.00005693, 0.00002723, -0.00027005, -0.00032578, 0.00063695, 0.00029354, 0.00017576, 0.00044993, 0.00024597, -0.00058238, 0.00003243, -0.00012433, -0.00002639, -0.00009067, 0.00002384, -0.00005863, 0.00001016, -0.00004263, 0.00004109, -0.00006255, -0.00000163, -0.00002543, -0.00000174, 0.00000984, -0.00003217, 0.00003135, -0.00000308, 0.00001371, -0.00001534, 0.00000531, -0.00000312, -0.00000524, -0.00000006, -0.00001211, -0.00000576, 0.00002642, },
		{ -0.00001087, -0.00001948, 0.00001802, -0.00001208, 0.00001067, 0.00000079, 0.00000633, -0.00004405, 0.00000332, 0.00003581, -0.00001743, -0.00002179, -0.00001120, 0.00001019, -0.00003635, 0.00001029, -0.00003179, 0.00002822, -0.00006630, 0.00004200, -0.00009554, 0.00000275, -0.00004635, 0.00003477, 0.00005126, 0.00000307, -0.00002704, 0.00001485, -0.00006508, 0.00003017, -0.00027725, -0.00038254, 0.00067554, 0.00033599, 0.00019459, 0.00061462, -0.00002128, -0.00053949, 0.00005844, -0.00012909, -0.00001706, -0.00010106, 0.00003748, -0.00006820, 0.00001939, -0.00004600, 0.00003903, -0.00006943, 0.00001976, -0.00005226, 0.00002615, -0.00001362, -0.00001422, 0.00002504, -0.00000522, 0.00001790, -0.00001969, 0.00001011, -0.00001005, 0.00000102, -0.00000641, -0.00001045, -0.00000514, 0.00003831, },
		{ -0.00003011, -0.00000330, 0.00000858, -0.00000616, 0.00000892, 0.00000214, 0.00000136, -0.00005039, 0.00002376, 0.00002437, -0.00002120, -0.00002345, 0.00000102, -0.00000504, -0.00001856, -0.00000661, -0.00001628, 0.00001573, -0.00005708, 0.00003795, -0.00010750, 0.00003118, -0.00005431, 0.00004555, 0.00002931, 0.00001024, -0.00003375, 0.00001159, -0.00006493, 0.00002206, -0.00027163, -0.00045665, 0.00073595, 0.00035587, 0.00027282, 0.00071847, -0.00032274, -0.00041062, 0.00002348, -0.00009400, -0.00003931, -0.00008459, 0.00002836, -0.00005898, 0.00001338, -0.00003579, 0.00002059, -0.00005615, 0.00002223, -0.00006539, 0.00004683, -0.00003594, 0.00000781, 0.00001145, -0.00000051, 0.00001682, -0.00001973, 0.00001129, -0.00001467, 0.00000599, -0.00001230, -0.00000932, -0.00000045, 0.00004445, },
		{ -0.00004313, 0.00001219, -0.00000292, 0.00000304, 0.00000367, 0.00000625, -0.00000842, -0.00004916, 0.00004131, 0.00000846, -0.00002079, -0.00002455, 0.00001166, -0.00002118, 0.00000366, -0.00003110, 0.00000988, -0.00001129, -0.00002864, 0.00000959, -0.00009220, 0.00003857, -0.00004894, 0.00004205, 0.00001824, 0.00000966, -0.00003027, -0.00000322, -0.00005174, -0.00000202, -0.00024839, -0.00055225, 0.00082144, 0.00035017, 0.00042968, 0.00070655, -0.00059110, -0.00023834, -0.00005430, -0.00002815, -0.00008734, -0.00004479, -0.00000170, -0.00003181, -0.00000714, -0.00001368, -0.00001189, -0.00002335, 0.00000258, -0.00005802, 0.00005157, -0.00004825, 0.00002619, -0.00000395, 0.00000814, 0.00001153, -0.00001542, 0.00000789, -0.00001545, 0.00000802, -0.00001638, -0.00000938, 0.00000919, 0.00004157, },
		{ -0.00004600, 0.00002198, -0.00001203, 0.00001205, -0.00000271, 0.00001136, -0.00002189, -0.00003894, 0.00005148, -0.00000751, -0.00001848, -0.00002317, 0.00001689, -0.00003205, 0.00002249, -0.00005447, 0.00003732, -0.00004316, 0.00000976, -0.00003574, -0.00005237, 0.00002238, -0.00002640, 0.00002125, 0.00002114, -0.00000058, -0.00001633, -0.00002905, -0.00002683, -0.00004020, -0.00021039, -0.00066428, 0.00092291, 0.00033497, 0.00065883, 0.00054393, -0.00076049, -0.00007896, -0.00013961, 0.00004431, -0.00014240, 0.00000344, -0.00004086, 0.00000379, -0.00003399, 0.00001186, -0.00004881, 0.00002011, -0.00003369, -0.00003114, 0.00003750, -0.00004529, 0.00003462, -0.00001544, 0.00001661, 0.00000469, -0.00000838, 0.00000049, -0.00001204, 0.00000637, -0.00001802, -0.00001027, 0.00002287, 0.00002866, },
		{ -0.00003776, 0.00002304, -0.00001512, 0.00001734, -0.00000721, 0.00001462, -0.00003609, -0.00002066, 0.00005145, -0.00001935, -0.00001690, -0.00001766, 0.00001380, -0.00003235, 0.00003030, -0.00006704, 0.00005447, -0.00006636, 0.00004288, -0.00008231, -0.00000118, -0.00000944, 0.00000832, -0.00001237, 0.00003440, -0.00001626, 0.00000220, -0.00005879, 0.00000139, -0.00008284, -0.00016898, -0.00077800, 0.00101830, 0.00034809, 0.00091926, 0.00023337, -0.00078856, 0.00001701, -0.00019418, 0.00009420, -0.00018033, 0.00003963, -0.00007182, 0.00003325, -0.00005466, 0.00002856, -0.00007636, 0.00005954, -0.00007364, 0.00000620, 0.00000911, -0.00002758, 0.00003072, -0.00001926, 0.00002125, -0.00000058, -0.00000123, -0.00000899, -0.00000550, 0.00000148, -0.00001758, -0.00001050, 0.00003760, 0.00000772, },
		{ -0.00002094, 0.00001543, -0.00001083, 0.00001678, -0.00000747, 0.00001314, -0.00004686, 0.00000217, 0.00004123, -0.00002466, -0.00001755, -0.00000800, 0.00000220, -0.00002025, 0.00002302, -0.00006250, 0.00005251, -0.00006913, 0.00005558, -0.00011185, 0.00004268, -0.00004170, 0.00004326, -0.00004829, 0.00004915, -0.00002873, 0.00001538, -0.00008133, 0.00002035, -0.00011597, -0.00013993, -0.00087414, 0.00107939, 0.00044224, 0.00113731, -0.00017670, -0.00067296, 0.00002336, -0.00019283, 0.00010031, -0.00018225, 0.00004675, -0.00007946, 0.00004351, -0.00005806, 0.00002588, -0.00008207, 0.00007996, -0.00010161, 0.00004014, -0.00002332, -0.00000129, 0.00001678, -0.00001491, 0.00002024, -0.00000193, 0.00000330, -0.00001786, 0.00000204, -0.00000519, -0.00001610, -0.00000794, 0.00004896, -0.00001638, },
		{ -0.00000091, 0.00000236, -0.00000064, 0.00001052, -0.00000281, 0.00000511, -0.00005005, 0.00002421, 0.00002379, -0.00002370, -0.00001995, 0.00000342, -0.00001444, 0.00000109, 0.00000262, -0.00004130, 0.00002996, -0.00004721, 0.00003971, -0.00011133, 0.00006278, -0.00005879, 0.00006516, -0.00007478, 0.00005584, -0.00002960, 0.00001440, -0.00008716, 0.00001963, -0.00012831, -0.00013531, -0.00093909, 0.00108523, 0.00066819, 0.00122193, -0.00059866, -0.00045570, -0.00005128, -0.00013489, 0.00005984, -0.00014396, 0.00001964, -0.00005834, 0.00002962, -0.00004051, 0.00000070, -0.00006085, 0.00007250, -0.00010552, 0.00005747, -0.00004782, 0.00002433, -0.00000141, -0.00000498, 0.00001408, 0.00000148, 0.00000328, -0.00002352, 0.00000808, -0.00001170, -0.00001481, -0.00000080, 0.00005241, -0.00003721, },
		{ 0.00001588, -0.00001098, 0.00001166, 0.00000099, 0.00000530, -0.00000921, -0.00004284, 0.00003968, 0.00000432, -0.00001923, -0.00002157, 0.00001216, -0.00002954, 0.00002385, -0.00002269, -0.00001160, -0.00000547, -0.00000698, -0.00000119, -0.00007935, 0.00005243, -0.00005218, 0.00006618, -0.00008500, 0.00004988, -0.00001632, -0.00000266, -0.00007472, -0.00000213, -0.00011905, -0.00015439, -0.00097623, 0.00103822, 0.00105226, 0.00109076, -0.00092698, -0.00021239, -0.00016569, -0.00004606, -0.00000838, -0.00007962, -0.00003111, -0.00001653, -0.00000170, -0.00000888, -0.00003951, -0.00001867, 0.00003909, -0.00008239, 0.00005129, -0.00005573, 0.00004087, -0.00001722, 0.00000632, 0.00000497, 0.00000888, -0.00000190, -0.00002424, 0.00001044, -0.00001624, -0.00001438, 0.00001140, 0.00004495, -0.00004877, },
		{ 0.00002414, -0.00001936, 0.00002146, -0.00000806, 0.00001370, -0.00002730, -0.00002485, 0.00004406, -0.00001144, -0.00001530, -0.00001905, 0.00001320, -0.00003548, 0.00003798, -0.00004100, 0.00001306, -0.00003876, 0.00003558, -0.00005167, -0.00002817, 0.00001850, -0.00002521, 0.00004868, -0.00008125, 0.00003551, 0.00000426, -0.00002723, -0.00005423, -0.00003299, -0.00010253, -0.00017828, -0.00101319, 0.00097611, 0.00157537, 0.00069954, -0.00106854, -0.00002899, -0.00026064, 0.00003064, -0.00006991, -0.00001764, -0.00008200, 0.00002635, -0.00003355, 0.00002073, -0.00007794, 0.00002804, -0.00000688, -0.00004054, 0.00002434, -0.00004548, 0.00004438, -0.00002615, 0.00001528, -0.00000437, 0.00001843, -0.00001148, -0.00001953, 0.00000792, -0.00001771, -0.00001455, 0.00002711, 0.00002650, -0.00004757, },
		{ 0.00002167, -0.00001955, 0.00002517, -0.00001309, 0.00001865, -0.00004495, 0.00000144, 0.00003548, -0.00001902, -0.00001560, -0.00000974, 0.00000309, -0.00002653, 0.00003499, -0.00004117, 0.00001879, -0.00005296, 0.00006044, -0.00008925, 0.00001997, -0.00002045, 0.00000741, 0.00002533, -0.00007511, 0.00002545, 0.00001677, -0.00004146, -0.00004616, -0.00004946, -0.00010634, -0.00017216, -0.00109978, 0.00097326, 0.00216266, 0.00006505, -0.00097281, 0.00002656, -0.00028106, 0.00005162, -0.00008788, 0.00001008, -0.00010546, 0.00004657, -0.00004532, 0.00002902, -0.00009427, 0.00005769, -0.00004469, 0.00000279, -0.00001149, -0.00002355, 0.00003700, -0.00002772, 0.00002035, -0.00001210, 0.00002813, -0.00002377, -0.00001024, 0.00000052, -0.00001587, -0.00001412, 0.00004282, 0.00000040, -0.00003394, },
		{ 0.00001013, -0.00001158, 0.00002170, -0.00001229, 0.00001734, -0.00005747, 0.00003122, 0.00001535, -0.00001666, -0.00002185, 0.00000666, -0.00001817, -0.00000156, 0.00001167, -0.00001760, -0.00000277, -0.00003624, 0.00005142, -0.00009311, 0.00004119, -0.00004142, 0.00002539, 0.00001416, -0.00008270, 0.00003588, 0.00000296, -0.00002469, -0.00007407, -0.00002454, -0.00016195, -0.00009646, -0.00129470, 0.00112730, 0.00268993, -0.00072778, -0.00065135, -0.00007205, -0.00019758, -0.00000915, -0.00003843, -0.00001855, -0.00008144, 0.00002637, -0.00002166, 0.00000189, -0.00007298, 0.00005192, -0.00005405, 0.00002776, -0.00003959, -0.00000187, 0.00002604, -0.00002549, 0.00002275, -0.00001794, 0.00003655, -0.00003671, 0.00000183, -0.00001069, -0.00001121, -0.00001153, 0.00005394, -0.00002725, -0.00001214, },
		{ -0.00000556, 0.00000126, 0.00001297, -0.00000646, 0.00000908, -0.00006115, 0.00005854, -0.00001194, -0.00000574, -0.00003302, 0.00002748, -0.00004635, 0.00003478, -0.00002788, 0.00002666, -0.00005011, 0.00001241, 0.00000364, -0.00005295, 0.00001987, -0.00002673, 0.00001190, 0.00003032, -0.00011676, 0.00007835, -0.00004951, 0.00003686, -0.00015371, 0.00005989, -0.00029041, 0.00007567, -0.00164317, 0.00153240, 0.00300850, -0.00153923, -0.00017899, -0.00029917, -0.00001931, -0.00014782, 0.00007787, -0.00010530, -0.00000695, -0.00003739, 0.00003969, -0.00006216, -0.00001103, 0.00000372, -0.00002353, 0.00001995, -0.00004534, 0.00000714, 0.00002055, -0.00002511, 0.00002550, -0.00002300, 0.00004309, -0.00004837, 0.00001434, -0.00002388, -0.00000452, -0.00000569, 0.00005628, -0.00004938, 0.00001101, },
		{ -0.00001904, 0.00001367, 0.00000311, 0.00000124, -0.00000427, -0.00005445, 0.00007783, -0.00004029, 0.00000969, -0.00004571, 0.00004761, -0.00007396, 0.00007294, -0.00007285, 0.00007997, -0.00011107, 0.00008112, -0.00007281, 0.00002532, -0.00004387, 0.00002752, -0.00003801, 0.00007806, -0.00017908, 0.00015216, -0.00013892, 0.00014106, -0.00028279, 0.00020134, -0.00048837, 0.00034062, -0.00215249, 0.00224596, 0.00298312, -0.00220669, 0.00032467, -0.00058420, 0.00020722, -0.00032934, 0.00023362, -0.00022858, 0.00010059, -0.00012999, 0.00012493, -0.00014908, 0.00007875, -0.00007800, 0.00004364, -0.00002317, -0.00002250, -0.00000388, 0.00002707, -0.00003133, 0.00003158, -0.00002881, 0.00004773, -0.00005719, 0.00002489, -0.00003689, 0.00000351, 0.00000324, 0.00004767, -0.00006015, 0.00002839, },
		{ -0.00002497, 0.00002054, -0.00000319, 0.00000647, -0.00001880, -0.00003848, 0.00008535, -0.00006350, 0.00002435, -0.00005535, 0.00006119, -0.00009246, 0.00010141, -0.00010905, 0.00012556, -0.00016627, 0.00014811, -0.00015492, 0.00011988, -0.00013249, 0.00010811, -0.00011335, 0.00014671, -0.00025672, 0.00024093, -0.00024598, 0.00026584, -0.00043632, 0.00037102, -0.00072083, 0.00065385, -0.00277454, 0.00325930, 0.00253240, -0.00258671, 0.00072868, -0.00083397, 0.00041296, -0.00049801, 0.00038246, -0.00034939, 0.00020802, -0.00022263, 0.00020753, -0.00023248, 0.00017024, -0.00016970, 0.00012909, -0.00009002, 0.00002348, -0.00003354, 0.00004622, -0.00004521, 0.00004192, -0.00003602, 0.00005045, -0.00006197, 0.00003134, -0.00004760, 0.00001245, 0.00001355, 0.00002899, -0.00005691, 0.00003504, },
		{ -0.00002104, 0.00001896, -0.00000275, 0.00000552, -0.00002989, -0.00001661, 0.00008002, -0.00007690, 0.00003345, -0.00005795, 0.00006365, -0.00009522, 0.00011064, -0.00012378, 0.00014737, -0.00019569, 0.00018896, -0.00021397, 0.00019964, -0.00021610, 0.00018824, -0.00018964, 0.00021280, -0.00032452, 0.00031561, -0.00033733, 0.00037308, -0.00057060, 0.00051799, -0.00092530, 0.00093304, -0.00340378, 0.00448338, 0.00165894, -0.00259449, 0.00092838, -0.00096299, 0.00052972, -0.00059635, 0.00047469, -0.00042455, 0.00027767, -0.00028228, 0.00025734, -0.00028231, 0.00023233, -0.00024054, 0.00020472, -0.00015806, 0.00007667, -0.00007158, 0.00007179, -0.00006301, 0.00005426, -0.00004353, 0.00005069, -0.00006182, 0.00003206, -0.00005405, 0.00002200, 0.00002229, 0.00000417, -0.00004116, 0.00002981, },
		{ -0.00000886, 0.00000953, 0.00000466, -0.00000304, -0.00003383, 0.00000656, 0.00006345, -0.00007836, 0.00003401, -0.00005149, 0.00005346, -0.00008000, 0.00009648, -0.00011050, 0.00013597, -0.00018614, 0.00018547, -0.00022578, 0.00023498, -0.00026286, 0.00023679, -0.00023732, 0.00024812, -0.00035347, 0.00034368, -0.00037581, 0.00041998, -0.00063592, 0.00058339, -0.00102814, 0.00107646, -0.00389464, 0.00575687, 0.00045842, -0.00222825, 0.00087533, -0.00092103, 0.00051371, -0.00058525, 0.00047482, -0.00042206, 0.00028098, -0.00028384, 0.00025188, -0.00027595, 0.00024001, -0.00026304, 0.00024240, -0.00020161, 0.00011616, -0.00010236, 0.00009277, -0.00007716, 0.00006358, -0.00004849, 0.00004727, -0.00005612, 0.00002603, -0.00005462, 0.00003163, 0.00002628, -0.00002080, -0.00001805, 0.00001571, },
		{ 0.00000672, -0.00000383, 0.00001604, -0.00001755, -0.00002918, 0.00002659, 0.00003931, -0.00006855, 0.00002559, -0.00003662, 0.00003278, -0.00004979, 0.00006184, -0.00007140, 0.00009242, -0.00013653, 0.00013285, -0.00017987, 0.00020869, -0.00025076, 0.00023011, -0.00023300, 0.00023069, -0.00032196, 0.00030150, -0.00033416, 0.00037453, -0.00059411, 0.00052075, -0.00096882, 0.00099285, -0.00409567, 0.00687694, -0.00089680, -0.00157067, 0.00058975, -0.00070876, 0.00036061, -0.00045791, 0.00037452, -0.00033298, 0.00020928, -0.00021993, 0.00018542, -0.00020734, 0.00018409, -0.00022352, 0.00022423, -0.00020118, 0.00012388, -0.00011088, 0.00009765, -0.00007918, 0.00006394, -0.00004736, 0.00003882, -0.00004475, 0.00001314, -0.00004816, 0.00004020, 0.00002336, -0.00003976, 0.00000536, -0.00000136, },
		{ 0.00001950, -0.00001539, 0.00002612, -0.00003361, -0.00001721, 0.00004033, 0.00001221, -0.00005037, 0.00001009, -0.00001628, 0.00000670, -0.00001196, 0.00001577, -0.00001685, 0.00002824, -0.00005900, 0.00004236, -0.00008422, 0.00012293, -0.00017633, 0.00016147, -0.00016908, 0.00015433, -0.00022558, 0.00018478, -0.00020696, 0.00022912, -0.00043450, 0.00031524, -0.00072426, 0.00063373, -0.00389177, 0.00764575, -0.00219629, -0.00076639, 0.00015199, -0.00037630, 0.00010702, -0.00024259, 0.00019594, -0.00017509, 0.00007741, -0.00010429, 0.00007221, -0.00009045, 0.00007567, -0.00012789, 0.00014971, -0.00015104, 0.00009169, -0.00008884, 0.00007925, -0.00006338, 0.00005099, -0.00003745, 0.00002463, -0.00002845, -0.00000563, -0.00003423, 0.00004591, 0.00001336, -0.00004839, 0.00002249, -0.00001482, },
		{ 0.00002441, -0.00001992, 0.00002943, -0.00004554, -0.00000150, 0.00004656, -0.00001344, -0.00002797, -0.00000916, 0.00000543, -0.00001850, 0.00002429, -0.00002964, 0.00003815, -0.00003835, 0.00002493, -0.00006191, 0.00003668, -0.00000066, -0.00005671, 0.00004459, -0.00005778, 0.00003265, -0.00008110, 0.00001297, -0.00001563, 0.00000691, -0.00018201, -0.00000582, -0.00032376, 0.00001720, -0.00324136, 0.00791953, -0.00324136, 0.00001720, -0.00032376, -0.00000582, -0.00018201, 0.00000691, -0.00001563, 0.00001297, -0.00008110, 0.00003265, -0.00005778, 0.00004459, -0.00005671, -0.00000066, 0.00003668, -0.00006191, 0.00002493, -0.00003835, 0.00003815, -0.00002964, 0.00002429, -0.00001850, 0.00000543, -0.00000916, -0.00002797, -0.00001344, 0.00004656, -0.00000150, -0.00004554, 0.00002943, -0.00001992, },
		{ 0.00001950, -0.00001482, 0.00002249, -0.00004839, 0.00001336, 0.00004591, -0.00003423, -0.00000563, -0.00002845, 0.00002463, -0.00003745, 0.00005099, -0.00006338, 0.00007925, -0.00008884, 0.00009169, -0.00015104, 0.00014971, -0.00012789, 0.00007567, -0.00009045, 0.00007221, -0.00010429, 0.00007741, -0.00017509, 0.00019594, -0.00024259, 0.00010702, -0.00037630, 0.00015199, -0.00076639, -0.00219629, 0.00764575, -0.00389177, 0.00063373, -0.00072426, 0.00031524, -0.00043450, 0.00022912, -0.00020696, 0.00018478, -0.00022558, 0.00015433, -0.00016908, 0.00016147, -0.00017633, 0.00012293, -0.00008422, 0.00004236, -0.00005900, 0.00002824, -0.00001685, 0.00001577, -0.00001196, 0.00000670, -0.00001628, 0.00001009, -0.00005037, 0.00001221, 0.00004033, -0.00001721, -0.00003361, 0.00002612, -0.00001539, },
		{ 0.00000672, -0.00000136, 0.00000536, -0.00003976, 0.00002336, 0.00004020, -0.00004816, 0.00001314, -0.00004475, 0.00003882, -0.00004736, 0.00006394, -0.00007918, 0.00009765, -0.00011088, 0.00012388, -0.00020118, 0.00022423, -0.00022352, 0.00018409, -0.00020734, 0.00018542, -0.00021993, 0.00020928, -0.00033298, 0.00037452, -0.00045791, 0.00036061, -0.00070876, 0.00058975, -0.00157067, -0.00089680, 0.00687694, -0.00409567, 0.00099285, -0.00096882, 0.00052075, -0.00059411, 0.00037453, -0.00033416, 0.00030150, -0.00032196, 0.00023069, -0.00023300, 0.00023011, -0.00025076, 0.00020869, -0.00017987, 0.00013285, -0.00013653, 0.00009242, -0.00007140, 0.00006184, -0.00004979, 0.00003278, -0.00003662, 0.00002559, -0.00006855, 0.00003931, 0.00002659, -0.00002918, -0.00001755, 0.00001604, -0.00000383, },
		{ -0.00000886, 0.00001571, -0.00001805, -0.00002080, 0.00002628, 0.00003163, -0.00005462, 0.00002603, -0.00005612, 0.00004727, -0.00004849, 0.00006358, -0.00007716, 0.00009277, -0.00010236, 0.00011616, -0.00020161, 0.00024240, -0.00026304, 0.00024001, -0.00027595, 0.00025188, -0.00028384, 0.00028098, -0.00042206, 0.00047482, -0.00058525, 0.00051371, -0.00092103, 0.00087533, -0.00222825, 0.00045842, 0.00575687, -0.00389464, 0.00107646, -0.00102814, 0.00058339, -0.00063592, 0.00041998, -0.00037581, 0.00034368, -0.00035347, 0.00024812, -0.00023732, 0.00023679, -0.00026286, 0.00023498, -0.00022578, 0.00018547, -0.00018614, 0.00013597, -0.00011050, 0.00009648, -0.00008000, 0.00005346, -0.00005149, 0.00003401, -0.00007836, 0.00006345, 0.00000656, -0.00003383, -0.00000304, 0.00000466, 0.00000953, },
		{ -0.00002104, 0.00002981, -0.00004116, 0.00000417, 0.00002229, 0.00002200, -0.00005405, 0.00003206, -0.00006182, 0.00005069, -0.00004353, 0.00005426, -0.00006301, 0.00007179, -0.00007158, 0.00007667, -0.00015806, 0.00020472, -0.00024054, 0.00023233, -0.00028231, 0.00025734, -0.00028228, 0.00027767, -0.00042455, 0.00047469, -0.00059635, 0.00052972, -0.00096299, 0.00092838, -0.00259449, 0.00165894, 0.00448338, -0.00340378, 0.00093304, -0.00092530, 0.00051799, -0.00057060, 0.00037308, -0.00033733, 0.00031561, -0.00032452, 0.00021280, -0.00018964, 0.00018824, -0.00021610, 0.00019964, -0.00021397, 0.00018896, -0.00019569, 0.00014737, -0.00012378, 0.00011064, -0.00009522, 0.00006365, -0.00005795, 0.00003345, -0.00007690, 0.00008002, -0.00001661, -0.00002989, 0.00000552, -0.00000275, 0.00001896, },
		{ -0.00002497, 0.00003504, -0.00005691, 0.00002899, 0.00001355, 0.00001245, -0.00004760, 0.00003134, -0.00006197, 0.00005045, -0.00003602, 0.00004192, -0.00004521, 0.00004622, -0.00003354, 0.00002348, -0.00009002, 0.00012909, -0.00016970, 0.00017024, -0.00023248, 0.00020753, -0.00022263, 0.00020802, -0.00034939, 0.00038246, -0.00049801, 0.00041296, -0.00083397, 0.00072868, -0.00258671, 0.00253240, 0.00325930, -0.00277454, 0.00065385, -0.00072083, 0.00037102, -0.00043632, 0.00026584, -0.00024598, 0.00024093, -0.00025672, 0.00014671, -0.00011335, 0.00010811, -0.00013249, 0.00011988, -0.00015492, 0.00014811, -0.00016627, 0.00012556, -0.00010905, 0.00010141, -0.00009246, 0.00006119, -0.00005535, 0.00002435, -0.00006350, 0.00008535, -0.00003848, -0.00001880, 0.00000647, -0.00000319, 0.00002054, },
		{ -0.00001904, 0.00002839, -0.00006015, 0.00004767, 0.00000324, 0.00000351, -0.00003689, 0.00002489, -0.00005719, 0.00004773, -0.00002881, 0.00003158, -0.00003133, 0.00002707, -0.00000388, -0.00002250, -0.00002317, 0.00004364, -0.00007800, 0.00007875, -0.00014908, 0.00012493, -0.00012999, 0.00010059, -0.00022858, 0.00023362, -0.00032934, 0.00020722, -0.00058420, 0.00032467, -0.00220669, 0.00298312, 0.00224596, -0.00215249, 0.00034062, -0.00048837, 0.00020134, -0.00028279, 0.00014106, -0.00013892, 0.00015216, -0.00017908, 0.00007806, -0.00003801, 0.00002752, -0.00004387, 0.00002532, -0.00007281, 0.00008112, -0.00011107, 0.00007997, -0.00007285, 0.00007294, -0.00007396, 0.00004761, -0.00004571, 0.00000969, -0.00004029, 0.00007783, -0.00005445, -0.00000427, 0.00000124, 0.00000311, 0.00001367, },
		{ -0.00000556, 0.00001101, -0.00004938, 0.00005628, -0.00000569, -0.00000452, -0.00002388, 0.00001434, -0.00004837, 0.00004309, -0.00002300, 0.00002550, -0.00002511, 0.00002055, 0.00000714, -0.00004534, 0.00001995, -0.00002353, 0.00000372, -0.00001103, -0.00006216, 0.00003969, -0.00003739, -0.00000695, -0.00010530, 0.00007787, -0.00014782, -0.00001931, -0.00029917, -0.00017899, -0.00153923, 0.00300850, 0.00153240, -0.00164317, 0.00007567, -0.00029041, 0.00005989, -0.00015371, 0.00003686, -0.00004951, 0.00007835, -0.00011676, 0.00003032, 0.00001190, -0.00002673, 0.00001987, -0.00005295, 0.00000364, 0.00001241, -0.00005011, 0.00002666, -0.00002788, 0.00003478, -0.00004635, 0.00002748, -0.00003302, -0.00000574, -0.00001194, 0.00005854, -0.00006115, 0.00000908, -0.00000646, 0.00001297, 0.00000126, },
		{ 0.00001013, -0.00001214, -0.00002725, 0.00005394, -0.00001153, -0.00001121, -0.00001069, 0.00000183, -0.00003671, 0.00003655, -0.00001794, 0.00002275, -0.00002549, 0.00002604, -0.00000187, -0.00003959, 0.00002776, -0.00005405, 0.00005192, -0.00007298, 0.00000189, -0.00002166, 0.00002637, -0.00008144, -0.00001855, -0.00003843, -0.00000915, -0.00019758, -0.00007205, -0.00065135, -0.00072778, 0.00268993, 0.00112730, -0.00129470, -0.00009646, -0.00016195, -0.00002454, -0.00007407, -0.00002469, 0.00000296, 0.00003588, -0.00008270, 0.00001416, 0.00002539, -0.00004142, 0.00004119, -0.00009311, 0.00005142, -0.00003624, -0.00000277, -0.00001760, 0.00001167, -0.00000156, -0.00001817, 0.00000666, -0.00002185, -0.00001666, 0.00001535, 0.00003122, -0.00005747, 0.00001734, -0.00001229, 0.00002170, -0.00001158, },
		{ 0.00002167, -0.00003394, 0.00000040, 0.00004282, -0.00001412, -0.00001587, 0.00000052, -0.00001024, -0.00002377, 0.00002813, -0.00001210, 0.00002035, -0.00002772, 0.00003700, -0.00002355, -0.00001149, 0.00000279, -0.00004469, 0.00005769, -0.00009427, 0.00002902, -0.00004532, 0.00004657, -0.00010546, 0.00001008, -0.00008788, 0.00005162, -0.00028106, 0.00002656, -0.00097281, 0.00006505, 0.00216266, 0.00097326, -0.00109978, -0.00017216, -0.00010634, -0.00004946, -0.00004616, -0.00004146, 0.00001677, 0.00002545, -0.00007511, 0.00002533, 0.00000741, -0.00002045, 0.00001997, -0.00008925, 0.00006044, -0.00005296, 0.00001879, -0.00004117, 0.00003499, -0.00002653, 0.00000309, -0.00000974, -0.00001560, -0.00001902, 0.00003548, 0.00000144, -0.00004495, 0.00001865, -0.00001309, 0.00002517, -0.00001955, },
		{ 0.00002414, -0.00004757, 0.00002650, 0.00002711, -0.00001455, -0.00001771, 0.00000792, -0.00001953, -0.00001148, 0.00001843, -0.00000437, 0.00001528, -0.00002615, 0.00004438, -0.00004548, 0.00002434, -0.00004054, -0.00000688, 0.00002804, -0.00007794, 0.00002073, -0.00003355, 0.00002635, -0.00008200, -0.00001764, -0.00006991, 0.00003064, -0.00026064, -0.00002899, -0.00106854, 0.00069954, 0.00157537, 0.00097611, -0.00101319, -0.00017828, -0.00010253, -0.00003299, -0.00005423, -0.00002723, 0.00000426, 0.00003551, -0.00008125, 0.00004868, -0.00002521, 0.00001850, -0.00002817, -0.00005167, 0.00003558, -0.00003876, 0.00001306, -0.00004100, 0.00003798, -0.00003548, 0.00001320, -0.00001905, -0.00001530, -0.00001144, 0.00004406, -0.00002485, -0.00002730, 0.00001370, -0.00000806, 0.00002146, -0.00001936, },
		{ 0.00001588, -0.00004877, 0.00004495, 0.00001140, -0.00001438, -0.00001624, 0.00001044, -0.00002424, -0.00000190, 0.00000888, 0.00000497, 0.00000632, -0.00001722, 0.00004087, -0.00005573, 0.00005129, -0.00008239, 0.00003909, -0.00001867, -0.00003951, -0.00000888, -0.00000170, -0.00001653, -0.00003111, -0.00007962, -0.00000838, -0.00004606, -0.00016569, -0.00021239, -0.00092698, 0.00109076, 0.00105226, 0.00103822, -0.00097623, -0.00015439, -0.00011905, -0.00000213, -0.00007472, -0.00000266, -0.00001632, 0.00004988, -0.00008500, 0.00006618, -0.00005218, 0.00005243, -0.00007935, -0.00000119, -0.00000698, -0.00000547, -0.00001160, -0.00002269, 0.00002385, -0.00002954, 0.00001216, -0.00002157, -0.00001923, 0.00000432, 0.00003968, -0.00004284, -0.00000921, 0.00000530, 0.00000099, 0.00001166, -0.00001098, },
		{ -0.00000091, -0.00003721, 0.00005241, -0.00000080, -0.00001481, -0.00001170, 0.00000808, -0.00002352, 0.00000328, 0.00000148, 0.00001408, -0.00000498, -0.00000141, 0.00002433, -0.00004782, 0.00005747, -0.00010552, 0.00007250, -0.00006085, 0.00000070, -0.00004051, 0.00002962, -0.00005834, 0.00001964, -0.00014396, 0.00005984, -0.00013489, -0.00005128, -0.00045570, -0.00059866, 0.00122193, 0.00066819, 0.00108523, -0.00093909, -0.00013531, -0.00012831, 0.00001963, -0.00008716, 0.00001440, -0.00002960, 0.00005584, -0.00007478, 0.00006516, -0.00005879, 0.00006278, -0.00011133, 0.00003971, -0.00004721, 0.00002996, -0.00004130, 0.00000262, 0.00000109, -0.00001444, 0.00000342, -0.00001995, -0.00002370, 0.00002379, 0.00002421, -0.00005005, 0.00000511, -0.00000281, 0.00001052, -0.00000064, 0.00000236, },
		{ -0.00002094, -0.00001638, 0.00004896, -0.00000794, -0.00001610, -0.00000519, 0.00000204, -0.00001786, 0.00000330, -0.00000193, 0.00002024, -0.00001491, 0.00001678, -0.00000129, -0.00002332, 0.00004014, -0.00010161, 0.00007996, -0.00008207, 0.00002588, -0.00005806, 0.00004351, -0.00007946, 0.00004675, -0.00018225, 0.00010031, -0.00019283, 0.00002336, -0.00067296, -0.00017670, 0.00113731, 0.00044224, 0.00107939, -0.00087414, -0.00013993, -0.00011597, 0.00002035, -0.00008133, 0.00001538, -0.00002873, 0.00004915, -0.00004829, 0.00004326, -0.00004170, 0.00004268, -0.00011185, 0.00005558, -0.00006913, 0.00005251, -0.00006250, 0.00002302, -0.00002025, 0.00000220, -0.00000800, -0.00001755, -0.00002466, 0.00004123, 0.00000217, -0.00004686, 0.00001314, -0.00000747, 0.00001678, -0.00001083, 0.00001543, },
		{ -0.00003776, 0.00000772, 0.00003760, -0.00001050, -0.00001758, 0.00000148, -0.00000550, -0.00000899, -0.00000123, -0.00000058, 0.00002125, -0.00001926, 0.00003072, -0.00002758, 0.00000911, 0.00000620, -0.00007364, 0.00005954, -0.00007636, 0.00002856, -0.00005466, 0.00003325, -0.00007182, 0.00003963, -0.00018033, 0.00009420, -0.00019418, 0.00001701, -0.00078856, 0.00023337, 0.00091926, 0.00034809, 0.00101830, -0.00077800, -0.00016898, -0.00008284, 0.00000139, -0.00005879, 0.00000220, -0.00001626, 0.00003440, -0.00001237, 0.00000832, -0.00000944, -0.00000118, -0.00008231, 0.00004288, -0.00006636, 0.00005447, -0.00006704, 0.00003030, -0.00003235, 0.00001380, -0.00001766, -0.00001690, -0.00001935, 0.00005145, -0.00002066, -0.00003609, 0.00001462, -0.00000721, 0.00001734, -0.00001512, 0.00002304, },
		{ -0.00004600, 0.00002866, 0.00002287, -0.00001027, -0.00001802, 0.00000637, -0.00001204, 0.00000049, -0.00000838, 0.00000469, 0.00001661, -0.00001544, 0.00003462, -0.00004529, 0.00003750, -0.00003114, -0.00003369, 0.00002011, -0.00004881, 0.00001186, -0.00003399, 0.00000379, -0.00004086, 0.00000344, -0.00014240, 0.00004431, -0.00013961, -0.00007896, -0.00076049, 0.00054393, 0.00065883, 0.00033497, 0.00092291, -0.00066428, -0.00021039, -0.00004020, -0.00002683, -0.00002905, -0.00001633, -0.00000058, 0.00002114, 0.00002125, -0.00002640, 0.00002238, -0.00005237, -0.00003574, 0.00000976, -0.00004316, 0.00003732, -0.00005447, 0.00002249, -0.00003205, 0.00001689, -0.00002317, -0.00001848, -0.00000751, 0.00005148, -0.00003894, -0.00002189, 0.00001136, -0.00000271, 0.00001205, -0.00001203, 0.00002198, },
		{ -0.00004313, 0.00004157, 0.00000919, -0.00000938, -0.00001638, 0.00000802, -0.00001545, 0.00000789, -0.00001542, 0.00001153, 0.00000814, -0.00000395, 0.00002619, -0.00004825, 0.00005157, -0.00005802, 0.00000258, -0.00002335, -0.00001189, -0.00001368, -0.00000714, -0.00003181, -0.00000170, -0.00004479, -0.00008734, -0.00002815, -0.00005430, -0.00023834, -0.00059110, 0.00070655, 0.00042968, 0.00035017, 0.00082144, -0.00055225, -0.00024839, -0.00000202, -0.00005174, -0.00000322, -0.00003027, 0.00000966, 0.00001824, 0.00004205, -0.00004894, 0.00003857, -0.00009220, 0.00000959, -0.00002864, -0.00001129, 0.00000988, -0.00003110, 0.00000366, -0.00002118, 0.00001166, -0.00002455, -0.00002079, 0.00000846, 0.00004131, -0.00004916, -0.00000842, 0.00000625, 0.00000367, 0.00000304, -0.00000292, 0.00001219, },
		{ -0.00003011, 0.00004445, -0.00000045, -0.00000932, -0.00001230, 0.00000599, -0.00001467, 0.00001129, -0.00001973, 0.00001682, -0.00000051, 0.00001145, 0.00000781, -0.00003594, 0.00004683, -0.00006539, 0.00002223, -0.00005615, 0.00002059, -0.00003579, 0.00001338, -0.00005898, 0.00002836, -0.00008459, -0.00003931, -0.00009400, 0.00002348, -0.00041062, -0.00032274, 0.00071847, 0.00027282, 0.00035587, 0.00073595, -0.00045665, -0.00027163, 0.00002206, -0.00006493, 0.00001159, -0.00003375, 0.00001024, 0.00002931, 0.00004555, -0.00005431, 0.00003118, -0.00010750, 0.00003795, -0.00005708, 0.00001573, -0.00001628, -0.00000661, -0.00001856, -0.00000504, 0.00000102, -0.00002345, -0.00002120, 0.00002437, 0.00002376, -0.00005039, 0.00000136, 0.00000214, 0.00000892, -0.00000616, 0.00000858, -0.00000330, },
		{ -0.00001087, 0.00003831, -0.00000514, -0.00001045, -0.00000641, 0.00000102, -0.00001005, 0.00001011, -0.00001969, 0.00001790, -0.00000522, 0.00002504, -0.00001422, -0.00001362, 0.00002615, -0.00005226, 0.00001976, -0.00006943, 0.00003903, -0.00004600, 0.00001939, -0.00006820, 0.00003748, -0.00010106, -0.00001706, -0.00012909, 0.00005844, -0.00053949, -0.00002128, 0.00061462, 0.00019459, 0.00033599, 0.00067554, -0.00038254, -0.00027725, 0.00003017, -0.00006508, 0.00001485, -0.00002704, 0.00000307, 0.00005126, 0.00003477, -0.00004635, 0.00000275, -0.00009554, 0.00004200, -0.00006630, 0.00002822, -0.00003179, 0.00001029, -0.00003635, 0.00001019, -0.00001120, -0.00002179, -0.00001743, 0.00003581, 0.00000332, -0.00004405, 0.00000633, 0.00000079, 0.00001067, -0.00001208, 0.00001802, -0.00001948, },
		{ 0.00000917, 0.00002642, -0.00000576, -0.00001211, -0.00000006, -0.00000524, -0.00000312, 0.00000531, -0.00001534, 0.00001371, -0.00000308, 0.00003135, -0.00003217, 0.00000984, -0.00000174, -0.00002543, -0.00000163, -0.00006255, 0.00004109, -0.00004263, 0.00001016, -0.00005863, 0.00002384, -0.00009067, -0.00002639, -0.00012433, 0.00003243, -0.00058238, 0.00024597, 0.00044993, 0.00017576, 0.00029354, 0.00063695, -0.00032578, -0.00027005, 0.00002723, -0.00005693, 0.00001147, -0.00001531, -0.00000540, 0.00007627, 0.00001784, -0.00003534, -0.00003525, -0.00006419, 0.00002468, -0.00005584, 0.00002365, -0.00003290, 0.00001503, -0.00004469, 0.00001985, -0.00002190, -0.00002053, -0.00000879, 0.00003970, -0.00001534, -0.00003310, 0.00000708, 0.00000239, 0.00000809, -0.00001260, 0.00002179, -0.00003124, },
		{ 0.00002495, 0.00001290, -0.00000424, -0.00001311, 0.00000513, -0.00001087, 0.00000393, -0.00000102, -0.00000835, 0.00000529, 0.00000623, 0.00002736, -0.00003990, 0.00002564, -0.00002599, 0.00000367, -0.00003222, -0.00004207, 0.00003112, -0.00003038, -0.00000815, -0.00003751, -0.00000465, -0.00006163, -0.00005885, -0.00008876, -0.00004870, -0.00052382, 0.00042983, 0.00027938, 0.00018599, 0.00024199, 0.00061009, -0.00027762, -0.00025840, 0.00002112, -0.00004761, 0.00000830, -0.00000507, -0.00000787, 0.00009586, 0.00000335, -0.00003262, -0.00006776, -0.00002773, -0.00000336, -0.00003293, 0.00000661, -0.00002208, 0.00000837, -0.00004274, 0.00002202, -0.00002960, -0.00001926, 0.00000331, 0.00003532, -0.00002879, -0.00002080, 0.00000530, 0.00000578, 0.00000210, -0.00000772, 0.00001838, -0.00003502, },
		{ 0.00003321, 0.00000131, -0.00000255, -0.00001236, 0.00000794, -0.00001428, 0.00000905, -0.00000645, -0.00000138, -0.00000450, 0.00002020, 0.00001379, -0.00003522, 0.00002847, -0.00003804, 0.00002383, -0.00006019, -0.00001802, 0.00001707, -0.00001709, -0.00002615, -0.00001613, -0.00003485, -0.00002907, -0.00009674, -0.00004484, -0.00015775, -0.00037856, 0.00051022, 0.00014252, 0.00019757, 0.00019572, 0.00058480, -0.00023012, -0.00024936, 0.00001819, -0.00004257, 0.00001029, -0.00000051, 0.00000011, 0.00010472, -0.00000410, -0.00004489, -0.00008274, -0.00000045, -0.00002911, -0.00000843, -0.00001422, -0.00000609, -0.00000483, -0.00003330, 0.00001751, -0.00003435, -0.00001658, 0.00001588, 0.00002437, -0.00003562, -0.00000971, 0.00000288, 0.00000910, -0.00000510, 0.00000057, 0.00000888, -0.00002990, },
		{ 0.00003321, -0.00000627, -0.00000197, -0.00000941, 0.00000787, -0.00001477, 0.00001104, -0.00000910, 0.00000298, -0.00001202, 0.00003430, -0.00000515, -0.00002062, 0.00001842, -0.00003472, 0.00002821, -0.00007626, 0.00000052, 0.00000637, -0.00000964, -0.00003589, -0.00000426, -0.00005471, -0.00000767, -0.00012169, -0.00001831, -0.00025680, -0.00018430, 0.00049644, 0.00005668, 0.00019386, 0.00016336, 0.00055557, -0.00017976, -0.00024565, 0.00002064, -0.00004315, 0.00001826, -0.00000160, 0.00001836, 0.00010245, -0.00000597, -0.00007102, -0.00007602, 0.00000906, -0.00004275, 0.00000819, -0.00003022, 0.00000750, -0.00001818, -0.00002103, 0.00000867, -0.00003681, -0.00001118, 0.00002556, 0.00001011, -0.00003633, -0.00000114, 0.00000116, 0.00001070, -0.00001092, 0.00000910, -0.00000359, -0.00001766, },
		{ 0.00002648, -0.00000953, -0.00000271, -0.00000460, 0.00000527, -0.00001262, 0.00000986, -0.00000830, 0.00000321, -0.00001432, 0.00004366, -0.00002343, -0.00000203, 0.00000056, -0.00001914, 0.00001647, -0.00007668, 0.00000864, 0.00000284, -0.00001084, -0.00003431, -0.00000604, -0.00005849, -0.00000521, -0.00012305, -0.00002710, -0.00031121, 0.00001236, 0.00041818, 0.00001895, 0.00017094, 0.00014591, 0.00052259, -0.00012797, -0.00024541, 0.00002642, -0.00004666, 0.00002925, -0.00000449, 0.00004279, 0.00009248, -0.00000842, -0.00010258, -0.00005239, 0.00000061, -0.00004121, 0.00001246, -0.00003634, 0.00001348, -0.00002649, -0.00001040, -0.00000190, -0.00003737, -0.00000285, 0.00002994, -0.00000392, -0.00003267, 0.00000476, 0.00000054, 0.00000977, -0.00001346, 0.00001479, -0.00001506, -0.00000197, },
		{ 0.00001601, -0.00000951, -0.00000416, 0.00000112, 0.00000112, -0.00000885, 0.00000650, -0.00000475, -0.00000052, -0.00001030, 0.00004491, -0.00003545, 0.00001365, -0.00001740, 0.00000113, -0.00000586, -0.00006369, 0.00000654, 0.00000585, -0.00001875, -0.00002403, -0.00001879, -0.00004856, -0.00001972, -0.00010219, -0.00007395, -0.00030149, 0.00017089, 0.00031232, 0.00001409, 0.00013405, 0.00013891, 0.00048976, -0.00007903, -0.00024436, 0.00003139, -0.00004860, 0.00003867, -0.00000388, 0.00006787, 0.00007895, -0.00001838, -0.00012791, -0.00002285, -0.00001879, -0.00002822, 0.00000591, -0.00003257, 0.00001063, -0.00002752, -0.00000419, -0.00001235, -0.00003569, 0.00000706, 0.00002853, -0.00001515, -0.00002666, 0.00000858, 0.00000065, 0.00000663, -0.00001211, 0.00001571, -0.00002210, 0.00001296, },
		{ 0.00000504, -0.00000792, -0.00000525, 0.00000656, -0.00000346, -0.00000471, 0.00000249, -0.00000013, -0.00000645, -0.00000120, 0.00003744, -0.00003818, 0.00002136, -0.00002837, 0.00001733, -0.00003020, -0.00004352, -0.00000188, 0.00001171, -0.00002839, -0.00001134, -0.00003516, -0.00003329, -0.00004150, -0.00007129, -0.00014560, -0.00022873, 0.00026759, 0.00021056, 0.00002396, 0.00009170, 0.00013632, 0.00046137, -0.00003699, -0.00023865, 0.00003213, -0.00004532, 0.00004297, 0.00000440, 0.00008930, 0.00006366, -0.00003919, -0.00013751, 0.00000069, -0.00003895, -0.00001157, -0.00000576, -0.00002292, 0.00000146, -0.00002207, -0.00000316, -0.00002170, -0.00003098, 0.00001604, 0.00002275, -0.00002257, -0.00001979, 0.00001095, 0.00000076, 0.00000242, -0.00000769, 0.00001167, -0.00002284, 0.00002368, },
		{ -0.00000390, -0.00000636, -0.00000498, 0.00001064, -0.00000749, -0.00000125, -0.00000076, 0.00000361, -0.00001192, 0.00000987, 0.00002353, -0.00003200, 0.00001967, -0.00002870, 0.00002298, -0.00004843, -0.00002333, -0.00001160, 0.00001615, -0.00003468, -0.00000269, -0.00004720, -0.00002229, -0.00005850, -0.00004722, -0.00021840, -0.00011237, 0.00029908, 0.00013201, 0.00003466, 0.00005078, 0.00013386, 0.00043963, -0.00000350, -0.00022695, 0.00002780, -0.00003585, 0.00004134, 0.00002185, 0.00010520, 0.00004469, -0.00006801, -0.00012834, 0.00001045, -0.00005135, 0.00000101, -0.00001590, -0.00001294, -0.00000965, -0.00001293, -0.00000662, -0.00002933, -0.00002281, 0.00002154, 0.00001518, -0.00002658, -0.00001273, 0.00001218, 0.00000028, -0.00000143, -0.00000199, 0.00000412, -0.00001744, 0.00002836, },
		{ -0.00000951, -0.00000583, -0.00000275, 0.00001264, -0.00001040, 0.00000108, -0.00000246, 0.00000510, -0.00001442, 0.00001908, 0.00000738, -0.00002031, 0.00001100, -0.00001928, 0.00001610, -0.00005575, -0.00000841, -0.00001888, 0.00001672, -0.00003502, -0.00000154, -0.00005032, -0.00002163, -0.00006240, -0.00004368, -0.00026764, 0.00001811, 0.00027852, 0.00008205, 0.00003948, 0.00001439, 0.00013014, 0.00042398, 0.00002258, -0.00021061, 0.00002027, -0.00002194, 0.00003566, 0.00004749, 0.00011537, 0.00001778, -0.00009652, -0.00010497, 0.00000558, -0.00005254, 0.00000524, -0.00002008, -0.00000695, -0.00001863, -0.00000327, -0.00001339, -0.00003436, -0.00001198, 0.00002214, 0.00000829, -0.00002827, -0.00000560, 0.00001214, -0.00000086, -0.00000381, 0.00000300, -0.00000453, -0.00000771, 0.00002700, },
		{ -0.00001174, -0.00000651, 0.00000147, 0.00001226, -0.00001195, 0.00000231, -0.00000266, 0.00000404, -0.00001263, 0.00002330, -0.00000656, -0.00000781, 0.00000025, -0.00000479, -0.00000064, -0.00005170, -0.00000084, -0.00002252, 0.00001364, -0.00003025, -0.00000705, -0.00004507, -0.00003115, -0.00005258, -0.00006509, -0.00027673, 0.00013374, 0.00022758, 0.00005594, 0.00003787, -0.00001764, 0.00012590, 0.00041222, 0.00004392, -0.00019235, 0.00001279, -0.00000676, 0.00002924, 0.00007930, 0.00011916, -0.00002038, -0.00011458, -0.00007716, -0.00000857, -0.00004474, 0.00000153, -0.00001761, -0.00000645, -0.00002329, 0.00000458, -0.00002242, -0.00003554, -0.00000066, 0.00001830, 0.00000341, -0.00002850, 0.00000158, 0.00001057, -0.00000227, -0.00000431, 0.00000576, -0.00001178, 0.00000362, 0.00002107, },
		{ -0.00001147, -0.00000792, 0.00000719, 0.00000956, -0.00001212, 0.00000279, -0.00000202, 0.00000123, -0.00000684, 0.00002118, -0.00001515, 0.00000147, -0.00000752, 0.00000851, -0.00002148, -0.00003945, 0.00000029, -0.00002340, 0.00000907, -0.00002364, -0.00001522, -0.00003616, -0.00004523, -0.00003603, -0.00010471, -0.00024241, 0.00021462, 0.00016771, 0.00004467, 0.00003241, -0.00004678, 0.00012213, 0.00040235, 0.00006287, -0.00017443, 0.00000810, 0.00000693, 0.00002521, 0.00011556, 0.00011355, -0.00006859, -0.00011511, -0.00005519, -0.00002379, -0.00003364, -0.00000624, -0.00001098, -0.00001014, -0.00002361, 0.00000943, -0.00003279, -0.00003162, 0.00000813, 0.00001218, 0.00000032, -0.00002734, 0.00000829, 0.00000749, -0.00000333, -0.00000325, 0.00000573, -0.00001587, 0.00001382, 0.00001277, },
		{ -0.00000999, -0.00000916, 0.00001351, 0.00000488, -0.00001092, 0.00000284, -0.00000141, -0.00000181, 0.00000119, 0.00001347, -0.00001756, 0.00000570, -0.00000922, 0.00001538, -0.00004012, -0.00002379, -0.00000261, -0.00002314, 0.00000543, -0.00001868, -0.00002149, -0.00002938, -0.00005631, -0.00002362, -0.00014778, -0.00017457, 0.00025381, 0.00011411, 0.00003991, 0.00002584, -0.00007393, 0.00011868, 0.00039386, 0.00008021, -0.00015740, 0.00000727, 0.00001781, 0.00002573, 0.00015508, 0.00009309, -0.00011963, -0.00009773, -0.00004522, -0.00003336, -0.00002497, -0.00001337, -0.00000393, -0.00001529, -0.00002093, 0.00001087, -0.00004325, -0.00002237, 0.00001179, 0.00000664, -0.00000226, -0.00002412, 0.00001348, 0.00000344, -0.00000358, -0.00000140, 0.00000329, -0.00001613, 0.00002092, 0.00000431, },
		{ -0.00000853, -0.00000921, 0.00001927, -0.00000114, -0.00000842, 0.00000255, -0.00000142, -0.00000357, 0.00000899, 0.00000253, -0.00001513, 0.00000569, -0.00000487, 0.00001353, -0.00005203, -0.00000908, -0.00000719, -0.00002269, 0.00000367, -0.00001696, -0.00002349, -0.00002812, -0.00005923, -0.00002449, -0.00017784, -0.00009116, 0.00025598, 0.00007386, 0.00003643, 0.00001962, -0.00009890, 0.00011412, 0.00038775, 0.00009519, -0.00014015, 0.00000971, 0.00002596, 0.00003247, 0.00019577, 0.00005217, -0.00016180, -0.00006940, -0.00004689, -0.00003505, -0.00002156, -0.00001702, 0.00000082, -0.00001950, -0.00001677, 0.00000855, -0.00005176, -0.00000932, 0.00000947, 0.00000384, -0.00000570, -0.00001798, 0.00001580, -0.00000051, -0.00000303, 0.00000045, -0.00000060, -0.00001297, 0.00002400, -0.00000265, },
		},
		{ { -0.00000743, -0.00001093, 0.00003439, -0.00000907, 0.00000010, 0.00000309, -0.00000665, 0.00000271, -0.00001921, 0.00000244, -0.00001454, -0.00001433, -0.00001225, -0.00000870, 0.00001125, -0.00001125, -0.00001056, -0.00000968, -0.00005902, -0.00001843, -0.00003105, -0.00004588, -0.00017045, -0.00000783, 0.00019590, 0.00008228, 0.00004109, 0.00005049, 0.00005083, 0.00005809, 0.00017918, -0.00003249, -0.00030959, -0.00003249, 0.00017918, 0.00005809, 0.00005083, 0.00005049, 0.00004109, 0.00008228, 0.00019590, -0.00000783, -0.00017045, -0.00004588, -0.00003105, -0.00001843, -0.00005902, -0.00000968, -0.00001056, -0.00001125, 0.00001125, -0.00000870, -0.00001225, -0.00001433, -0.00001454, 0.00000244, -0.00001921, 0.00000271, -0.00000665, 0.00000309, 0.00000010, -0.00000907, 0.00003439, -0.00001093, },
		{ -0.00000947, -0.00001510, 0.00002442, 0.00001114, -0.00001412, 0.00001438, -0.00001545, 0.00000862, -0.00001831, -0.00000302, -0.00001133, -0.00001253, -0.00001473, -0.00001146, 0.00000900, -0.00000393, -0.00001546, -0.00000184, -0.00005925, -0.00002281, -0.00003282, -0.00003112, -0.00014414, -0.00011457, 0.00021190, 0.00011416, 0.00004490, 0.00006476, 0.00005580, 0.00005655, 0.00020155, -0.00001954, -0.00031172, -0.00004182, 0.00015609, 0.00005926, 0.00004629, 0.00004171, 0.00002923, 0.00007056, 0.00015594, 0.00007619, -0.00015188, -0.00008148, -0.00002253, -0.00002103, -0.00005088, -0.00002207, -0.00000336, -0.00001933, 0.00001283, -0.00000727, -0.00000833, -0.00001539, -0.00001749, 0.00000401, -0.00001363, -0.00000826, 0.00000334, -0.00000623, 0.00000993, -0.00001949, 0.00003088, 0.00000180, },
		{ -0.00001292, -0.00001286, 0.00000679, 0.00003160, -0.00002263, 0.00001799, -0.00001432, 0.00000341, -0.00000835, -0.00001307, -0.00000650, -0.00001308, -0.00001237, -0.00001809, 0.00000920, -0.00000160, -0.00001346, -0.00000317, -0.00004805, -0.00003699, -0.00002355, -0.00003938, -0.00008763, -0.00021030, 0.00017819, 0.00017117, 0.00004249, 0.00008242, 0.00006246, 0.00005533, 0.00022218, -0.00000236, -0.00031770, -0.00004876, 0.00013342, 0.00005970, 0.00004130, 0.00003845, 0.00001192, 0.00006892, 0.00011464, 0.00012120, -0.00009627, -0.00011981, -0.00002048, -0.00002170, -0.00004340, -0.00003058, -0.00000108, -0.00002212, 0.00000906, -0.00000296, -0.00000754, -0.00001212, -0.00002229, 0.00000435, -0.00000680, -0.00001659, 0.00000651, -0.00000709, 0.00001037, -0.00001752, 0.00001563, 0.00001892, },
		{ -0.00001205, -0.00001003, -0.00000926, 0.00004169, -0.00001709, 0.00000840, -0.00000070, -0.00001170, 0.00000666, -0.00002359, -0.00000234, -0.00001517, -0.00000657, -0.00002624, 0.00001019, -0.00000316, -0.00000637, -0.00001065, -0.00003065, -0.00005405, -0.00001103, -0.00005668, -0.00003086, -0.00025892, 0.00008182, 0.00024323, 0.00004445, 0.00009732, 0.00007384, 0.00005386, 0.00024151, 0.00001825, -0.00032629, -0.00005456, 0.00011150, 0.00006017, 0.00003450, 0.00004033, -0.00000774, 0.00006949, 0.00008340, 0.00013056, -0.00002708, -0.00013843, -0.00003600, -0.00001438, -0.00004243, -0.00003008, -0.00000722, -0.00001735, -0.00000119, 0.00000517, -0.00001101, -0.00000464, -0.00002821, 0.00000426, -0.00000185, -0.00001849, 0.00000105, -0.00000044, 0.00000424, -0.00000831, -0.00000312, 0.00003207, },
		{ -0.00000301, -0.00001093, -0.00001718, 0.00003643, 0.00000275, -0.00001077, 0.00001864, -0.00002735, 0.00001705, -0.00002704, -0.00000376, -0.00001475, -0.00000248, -0.00003010, 0.00000718, -0.00000376, -0.00000047, -0.00001626, -0.00001735, -0.00006233, -0.00000810, -0.00006438, -0.00000112, -0.00024241, -0.00006305, 0.00030063, 0.00006937, 0.00010063, 0.00009434, 0.00005017, 0.00026174, 0.00004022, -0.00033574, -0.00006020, 0.00009002, 0.00006181, 0.00002492, 0.00004633, -0.00002715, 0.00006831, 0.00006369, 0.00011840, 0.00003161, -0.00012505, -0.00006799, -0.00000259, -0.00004596, -0.00002290, -0.00001822, -0.00000914, -0.00001337, 0.00001298, -0.00001480, 0.00000239, -0.00003111, 0.00000179, 0.00000178, -0.00001564, -0.00000839, 0.00000689, -0.00000104, -0.00000028, -0.00001606, 0.00003419, },
		{ 0.00001234, -0.00001480, -0.00001688, 0.00001950, 0.00002810, -0.00002809, 0.00003126, -0.00003151, 0.00001356, -0.00001800, -0.00001368, -0.00000894, -0.00000428, -0.00002564, -0.00000258, -0.00000019, -0.00000055, -0.00001394, -0.00001529, -0.00005498, -0.00002118, -0.00005382, -0.00000625, -0.00017347, -0.00021389, 0.00030227, 0.00013410, 0.00008743, 0.00012580, 0.00004329, 0.00028480, 0.00006200, -0.00034513, -0.00006593, 0.00006849, 0.00006500, 0.00001330, 0.00005387, -0.00004339, 0.00006335, 0.00005304, 0.00009831, 0.00006741, -0.00008473, -0.00010190, 0.00000216, -0.00004623, -0.00001651, -0.00002612, -0.00000512, -0.00002030, 0.00001424, -0.00001322, 0.00000333, -0.00002691, -0.00000507, 0.00000597, -0.00001215, -0.00001526, 0.00000801, 0.00000021, 0.00000135, -0.00001885, 0.00002441, },
		{ 0.00002697, -0.00001655, -0.00001362, 0.00000036, 0.00004623, -0.00003145, 0.00002737, -0.00001801, -0.00000515, 0.00000105, -0.00002853, -0.00000018, -0.00001059, -0.00001512, -0.00001561, 0.00000501, -0.00000528, -0.00000497, -0.00002249, -0.00003631, -0.00004360, -0.00003315, -0.00003057, -0.00009039, -0.00031647, 0.00021665, 0.00023797, 0.00006568, 0.00016237, 0.00003692, 0.00030928, 0.00008511, -0.00035602, -0.00007059, 0.00004622, 0.00006907, 0.00000204, 0.00005916, -0.00005337, 0.00005315, 0.00004931, 0.00007726, 0.00008204, -0.00003520, -0.00011888, -0.00001044, -0.00003758, -0.00001661, -0.00002517, -0.00001011, -0.00001803, 0.00000620, -0.00000421, -0.00000328, -0.00001598, -0.00001482, 0.00001044, -0.00001014, -0.00001639, 0.00000135, 0.00000736, -0.00000198, -0.00001433, 0.00000869, },
		{ 0.00003321, -0.00001116, -0.00001303, -0.00001213, 0.00004836, -0.00001626, 0.00000727, 0.00000792, -0.00002958, 0.00001905, -0.00003854, 0.00000397, -0.00001472, -0.00000643, -0.00002339, 0.00000472, -0.00000812, 0.00000296, -0.00002925, -0.00001926, -0.00005949, -0.00002097, -0.00004691, -0.00003566, -0.00033271, 0.00004520, 0.00035094, 0.00005978, 0.00018960, 0.00004022, 0.00032942, 0.00011487, -0.00037266, -0.00007158, 0.00002225, 0.00007301, -0.00000635, 0.00005932, -0.00005563, 0.00003767, 0.00005088, 0.00005702, 0.00008437, 0.00000557, -0.00010861, -0.00003968, -0.00002338, -0.00002139, -0.00001687, -0.00002169, -0.00000974, -0.00000706, 0.00000776, -0.00001247, -0.00000446, -0.00002167, 0.00001173, -0.00000810, -0.00001382, -0.00000846, 0.00001409, -0.00000405, -0.00000946, -0.00000434, },
		{ 0.00002813, 0.00000170, -0.00001643, -0.00001518, 0.00003490, 0.00001097, -0.00001771, 0.00003164, -0.00004403, 0.00002245, -0.00003390, -0.00000350, -0.00001004, -0.00000725, -0.00001865, -0.00000649, -0.00000356, 0.00000256, -0.00002618, -0.00001572, -0.00005568, -0.00003139, -0.00003669, -0.00003076, -0.00026360, -0.00016765, 0.00041739, 0.00010156, 0.00019197, 0.00006262, 0.00033877, 0.00015755, -0.00039957, -0.00006669, -0.00000397, 0.00007602, -0.00001085, 0.00005436, -0.00005214, 0.00001978, 0.00005464, 0.00003887, 0.00008128, 0.00003001, -0.00007647, -0.00007168, -0.00001564, -0.00002255, -0.00000883, -0.00003162, -0.00000385, -0.00001685, 0.00001421, -0.00001584, -0.00000066, -0.00001925, 0.00000612, -0.00000319, -0.00001206, -0.00001463, 0.00001353, 0.00000049, -0.00000941, -0.00000931, },
		{ 0.00001558, 0.00001660, -0.00001988, -0.00001269, 0.00001450, 0.00003636, -0.00003182, 0.00003811, -0.00003673, 0.00000486, -0.00001295, -0.00002216, 0.00000338, -0.00001820, -0.00000211, -0.00002606, 0.00000660, -0.00000606, -0.00001239, -0.00002694, -0.00003255, -0.00006176, -0.00000377, -0.00006385, -0.00015147, -0.00034824, 0.00037968, 0.00020894, 0.00016726, 0.00010447, 0.00033707, 0.00021471, -0.00043698, -0.00005721, -0.00003089, 0.00007698, -0.00001199, 0.00004708, -0.00004778, 0.00000469, 0.00005577, 0.00002605, 0.00007350, 0.00004274, -0.00003958, -0.00008735, -0.00002641, -0.00001284, -0.00000804, -0.00003245, -0.00000752, -0.00001634, 0.00000918, -0.00000821, -0.00000862, -0.00000609, -0.00000589, 0.00000466, -0.00001339, -0.00001376, 0.00000396, 0.00001111, -0.00001341, -0.00000740, },
		{ 0.00000342, 0.00002576, -0.00001771, -0.00001108, -0.00000215, 0.00004721, -0.00002494, 0.00002204, -0.00000868, -0.00002661, 0.00001401, -0.00004178, 0.00001680, -0.00003096, 0.00001621, -0.00004289, 0.00001273, -0.00001443, 0.00000310, -0.00004202, -0.00000488, -0.00009292, 0.00002813, -0.00009899, -0.00005772, -0.00042830, 0.00021249, 0.00036337, 0.00013807, 0.00015059, 0.00033482, 0.00027949, -0.00047801, -0.00004971, -0.00005383, 0.00007365, -0.00001057, 0.00004088, -0.00004725, -0.00000331, 0.00005044, 0.00002169, 0.00005869, 0.00005320, -0.00001472, -0.00007558, -0.00005709, 0.00000582, -0.00001390, -0.00002393, -0.00002056, -0.00000631, -0.00000550, 0.00000760, -0.00002405, 0.00001148, -0.00001776, 0.00001098, -0.00001550, -0.00000836, -0.00000955, 0.00002110, -0.00001547, -0.00000466, },
		{ -0.00000222, 0.00002455, -0.00000761, -0.00001371, -0.00000920, 0.00003959, 0.00000067, -0.00000768, 0.00002533, -0.00005380, 0.00002944, -0.00004779, 0.00001819, -0.00003385, 0.00002333, -0.00004423, 0.00000426, -0.00001242, 0.00000841, -0.00004624, 0.00000852, -0.00010250, 0.00003311, -0.00010144, -0.00002786, -0.00038213, -0.00005225, 0.00050275, 0.00014869, 0.00017516, 0.00034952, 0.00033944, -0.00051114, -0.00005380, -0.00006663, 0.00006311, -0.00000658, 0.00003721, -0.00005180, -0.00000432, 0.00003928, 0.00002495, 0.00003736, 0.00006603, -0.00000704, -0.00004170, -0.00009422, 0.00002073, -0.00001741, -0.00001378, -0.00003486, 0.00000471, -0.00002057, 0.00002171, -0.00003629, 0.00002237, -0.00002022, 0.00000951, -0.00001363, -0.00000443, -0.00001876, 0.00002239, -0.00000966, -0.00000649, },
		{ -0.00000118, 0.00001488, 0.00000679, -0.00001827, -0.00000841, 0.00001996, 0.00003165, -0.00003291, 0.00004509, -0.00005821, 0.00001986, -0.00003179, 0.00000185, -0.00002126, 0.00001312, -0.00002559, -0.00002083, 0.00000231, -0.00000088, -0.00003308, -0.00000093, -0.00008172, 0.00000312, -0.00006211, -0.00006534, -0.00024374, -0.00032839, 0.00054153, 0.00024187, 0.00015962, 0.00039277, 0.00038644, -0.00052829, -0.00007557, -0.00006635, 0.00004454, -0.00000015, 0.00003527, -0.00005854, -0.00000329, 0.00002791, 0.00002992, 0.00001564, 0.00007701, -0.00000794, -0.00000472, -0.00011626, 0.00001625, -0.00000815, -0.00001141, -0.00004048, 0.00000666, -0.00002572, 0.00002400, -0.00003557, 0.00001809, -0.00000800, -0.00000223, -0.00000555, -0.00000603, -0.00001853, 0.00001181, 0.00000431, -0.00001281, },
		{ 0.00000113, 0.00000373, 0.00001801, -0.00001899, -0.00000625, 0.00000000, 0.00005187, -0.00003745, 0.00003783, -0.00003359, -0.00001355, 0.00000027, -0.00002497, 0.00000019, -0.00000774, 0.00000400, -0.00005155, 0.00001995, -0.00001660, -0.00001032, -0.00002438, -0.00004312, -0.00004432, -0.00000416, -0.00013133, -0.00009073, -0.00051160, 0.00041398, 0.00042449, 0.00011452, 0.00045603, 0.00042729, -0.00053331, -0.00011100, -0.00005794, 0.00002219, 0.00000635, 0.00003413, -0.00006320, -0.00000655, 0.00002308, 0.00002928, 0.00000180, 0.00007757, -0.00000335, 0.00001537, -0.00010774, -0.00001394, 0.00001576, -0.00001935, -0.00003384, -0.00000404, -0.00001758, 0.00001188, -0.00002048, -0.00000051, 0.00001493, -0.00001903, 0.00000501, -0.00001148, -0.00001068, -0.00000586, 0.00001956, -0.00001761, },
		{ -0.00000137, -0.00000207, 0.00002010, -0.00001190, -0.00000792, -0.00001138, 0.00005229, -0.00001729, 0.00000696, 0.00000845, -0.00005313, 0.00002917, -0.00004467, 0.00001471, -0.00002329, 0.00002656, -0.00006944, 0.00002413, -0.00002328, 0.00000511, -0.00004164, -0.00001245, -0.00007725, 0.00003342, -0.00017147, -0.00000176, -0.00053208, 0.00011896, 0.00064299, 0.00008849, 0.00050750, 0.00048549, -0.00054317, -0.00014550, -0.00005413, 0.00000529, 0.00000776, 0.00003466, -0.00006382, -0.00001695, 0.00002728, 0.00002011, -0.00000027, 0.00006336, 0.00001433, 0.00001132, -0.00007084, -0.00005912, 0.00004288, -0.00002943, -0.00002130, -0.00002107, -0.00000278, -0.00000706, 0.00000008, -0.00002258, 0.00003626, -0.00002980, 0.00000987, -0.00001446, -0.00000257, -0.00002075, 0.00002618, -0.00001360, },
		{ -0.00000985, -0.00000097, 0.00001301, 0.00000132, -0.00001306, -0.00001309, 0.00003571, 0.00001696, -0.00002941, 0.00004433, -0.00007426, 0.00003383, -0.00004073, 0.00000851, -0.00001981, 0.00002745, -0.00006122, 0.00000439, -0.00001061, 0.00000057, -0.00003668, -0.00001008, -0.00007211, 0.00002421, -0.00015283, -0.00001189, -0.00039466, -0.00026202, 0.00079014, 0.00015058, 0.00050866, 0.00058827, -0.00057748, -0.00016269, -0.00006800, 0.00000251, -0.00000072, 0.00003876, -0.00006169, -0.00003192, 0.00003647, 0.00000662, 0.00000529, 0.00003950, 0.00003982, -0.00000669, -0.00002557, -0.00009570, 0.00005413, -0.00002811, -0.00001453, -0.00003257, 0.00000651, -0.00002001, 0.00001261, -0.00003405, 0.00004268, -0.00002462, 0.00000258, -0.00000950, -0.00000118, -0.00002470, 0.00001822, 0.00000143, },
		{ -0.00001866, 0.00000231, 0.00000272, 0.00001394, -0.00001642, -0.00001079, 0.00001348, 0.00004711, -0.00004956, 0.00005292, -0.00006110, 0.00000611, -0.00001084, -0.00001845, 0.00000277, 0.00000691, -0.00002992, -0.00003343, 0.00001620, -0.00002082, -0.00001098, -0.00003520, -0.00003177, -0.00002506, -0.00008647, -0.00009355, -0.00018028, -0.00059128, 0.00074976, 0.00034756, 0.00044559, 0.00074423, -0.00064113, -0.00015860, -0.00010177, 0.00001362, -0.00001777, 0.00004555, -0.00005860, -0.00004629, 0.00004320, -0.00000303, 0.00000930, 0.00001758, 0.00005920, -0.00001899, 0.00000278, -0.00010173, 0.00003627, -0.00000775, -0.00002067, -0.00003048, 0.00000195, -0.00001835, 0.00000868, -0.00002731, 0.00002924, -0.00000261, -0.00001535, 0.00000248, -0.00000735, -0.00001707, -0.00000145, 0.00002152, },
		{ -0.00001907, 0.00000083, -0.00000336, 0.00001897, -0.00001272, -0.00001102, -0.00000289, 0.00005887, -0.00004179, 0.00002917, -0.00001806, -0.00004118, 0.00002832, -0.00004938, 0.00002848, -0.00001825, 0.00000465, -0.00006785, 0.00003721, -0.00004032, 0.00001588, -0.00006515, 0.00001477, -0.00007713, -0.00002008, -0.00017586, -0.00000417, -0.00073845, 0.00046182, 0.00065729, 0.00035578, 0.00092710, -0.00071256, -0.00015074, -0.00013998, 0.00002469, -0.00003248, 0.00004823, -0.00005289, -0.00005761, 0.00004303, -0.00000382, 0.00000526, 0.00000679, 0.00006123, -0.00001134, 0.00000069, -0.00007252, -0.00000618, 0.00002447, -0.00003485, -0.00001768, -0.00001404, -0.00000453, -0.00000851, -0.00000720, 0.00000374, 0.00002553, -0.00003335, 0.00001335, -0.00001483, -0.00000520, -0.00002232, 0.00003530, },
		{ -0.00000602, -0.00000842, -0.00000211, 0.00001452, -0.00000175, -0.00001542, -0.00000909, 0.00005014, -0.00001148, -0.00001210, 0.00003121, -0.00007988, 0.00004958, -0.00006033, 0.00003538, -0.00002556, 0.00001813, -0.00007486, 0.00003093, -0.00003676, 0.00002020, -0.00007153, 0.00003208, -0.00008925, -0.00000489, -0.00019140, 0.00004756, -0.00064914, -0.00002774, 0.00096988, 0.00032656, 0.00108280, -0.00075046, -0.00017207, -0.00015518, 0.00001331, -0.00002836, 0.00003606, -0.00003899, -0.00006875, 0.00003863, 0.00000120, -0.00000475, 0.00000654, 0.00004673, 0.00001428, -0.00002412, -0.00002590, -0.00005016, 0.00004813, -0.00004196, -0.00000665, -0.00002929, 0.00000916, -0.00002585, 0.00001176, -0.00001743, 0.00004314, -0.00003770, 0.00001310, -0.00001510, 0.00000106, -0.00003291, 0.00003334, },
		{ 0.00001683, -0.00002103, 0.00000271, 0.00000540, 0.00001071, -0.00001959, -0.00000929, 0.00003060, 0.00002299, -0.00004534, 0.00005777, -0.00008348, 0.00003302, -0.00003719, 0.00001198, -0.00000351, -0.00000110, -0.00004534, -0.00000862, -0.00000389, -0.00000701, -0.00004189, 0.00000415, -0.00004384, -0.00005871, -0.00012081, -0.00003383, -0.00037898, -0.00056691, 0.00111924, 0.00045385, 0.00116449, -0.00072075, -0.00024925, -0.00012613, -0.00003646, 0.00000515, 0.00000247, -0.00001258, -0.00008478, 0.00003744, 0.00000293, -0.00001145, 0.00000712, 0.00002823, 0.00004119, -0.00004794, 0.00000798, -0.00006709, 0.00004242, -0.00002778, -0.00000980, -0.00003103, 0.00000961, -0.00002965, 0.00001528, -0.00001998, 0.00003828, -0.00002114, -0.00000210, -0.00000466, -0.00000324, -0.00002848, 0.00001491, },
		{ 0.00003807, -0.00002741, 0.00000347, -0.00000052, 0.00001692, -0.00001732, -0.00001089, 0.00001365, 0.00004256, -0.00005043, 0.00004563, -0.00004502, -0.00001868, 0.00001165, -0.00003175, 0.00003794, -0.00004184, 0.00000578, -0.00006370, 0.00004143, -0.00005043, 0.00000881, -0.00005235, 0.00003685, -0.00015090, -0.00000626, -0.00017688, -0.00006921, -0.00095753, 0.00096067, 0.00077453, 0.00117981, -0.00062928, -0.00037779, -0.00005778, -0.00011825, 0.00006054, -0.00004577, 0.00002208, -0.00010602, 0.00004424, -0.00000587, -0.00000664, -0.00000167, 0.00002009, 0.00005071, -0.00004784, 0.00000625, -0.00004209, 0.00000228, 0.00000808, -0.00002838, -0.00001628, -0.00000677, -0.00001603, -0.00000028, -0.00000171, 0.00001243, 0.00001058, -0.00002523, 0.00001103, -0.00001426, -0.00001412, -0.00001082, },
		{ 0.00004522, -0.00001954, -0.00000498, 0.00000150, 0.00001324, -0.00000658, -0.00001740, 0.00000682, 0.00003934, -0.00002519, 0.00000241, 0.00001656, -0.00007864, 0.00005726, -0.00006814, 0.00007195, -0.00007572, 0.00004689, -0.00010149, 0.00006798, -0.00007906, 0.00004687, -0.00009804, 0.00010393, -0.00022077, 0.00007626, -0.00027574, 0.00012736, -0.00105081, 0.00045921, 0.00120957, 0.00121841, -0.00053505, -0.00051384, 0.00001316, -0.00019988, 0.00011021, -0.00008696, 0.00004978, -0.00012401, 0.00005514, -0.00002307, 0.00000841, -0.00002079, 0.00002673, 0.00003660, -0.00001826, -0.00003066, 0.00001388, -0.00005418, 0.00004777, -0.00004893, 0.00000450, -0.00003032, 0.00000557, -0.00002483, 0.00002542, -0.00001933, 0.00004061, -0.00004136, 0.00002049, -0.00002221, -0.00000108, -0.00003015, },
		{ 0.00003294, 0.00000236, -0.00002023, 0.00000899, 0.00000330, 0.00000777, -0.00002530, 0.00000780, 0.00002017, 0.00001408, -0.00004524, 0.00006681, -0.00011079, 0.00006752, -0.00006968, 0.00007301, -0.00007640, 0.00005027, -0.00009555, 0.00005185, -0.00006840, 0.00004407, -0.00009869, 0.00011531, -0.00021865, 0.00006876, -0.00025798, 0.00013018, -0.00082886, -0.00026174, 0.00155941, 0.00143065, -0.00052430, -0.00059676, 0.00003745, -0.00024016, 0.00012087, -0.00009587, 0.00005231, -0.00012600, 0.00005940, -0.00003746, 0.00002224, -0.00004000, 0.00003909, 0.00000972, 0.00002506, -0.00007863, 0.00006873, -0.00009411, 0.00006462, -0.00005180, 0.00001463, -0.00004476, 0.00001883, -0.00004145, 0.00004320, -0.00003757, 0.00005098, -0.00003717, 0.00001454, -0.00001900, 0.00000132, -0.00003365, },
		{ 0.00000706, 0.00002849, -0.00003278, 0.00001388, -0.00000453, 0.00001714, -0.00002734, 0.00000844, -0.00000027, 0.00004425, -0.00006885, 0.00007728, -0.00009321, 0.00002971, -0.00003001, 0.00003698, -0.00003987, 0.00001281, -0.00004629, -0.00000330, -0.00002145, 0.00000085, -0.00005210, 0.00006719, -0.00014253, -0.00002625, -0.00013284, -0.00002342, -0.00042426, -0.00094849, 0.00157126, 0.00195224, -0.00064841, -0.00059654, -0.00000897, -0.00022004, 0.00007906, -0.00006412, 0.00002430, -0.00010728, 0.00004983, -0.00003846, 0.00002242, -0.00004652, 0.00004298, -0.00001114, 0.00005699, -0.00010622, 0.00008939, -0.00009084, 0.00004184, -0.00002670, 0.00000478, -0.00003976, 0.00001296, -0.00003884, 0.00003995, -0.00003141, 0.00003463, -0.00001096, -0.00000579, -0.00000519, -0.00000757, -0.00002209, },
		{ -0.00001867, 0.00004477, -0.00003228, 0.00000896, -0.00000386, 0.00001612, -0.00001965, 0.00000336, -0.00001120, 0.00005114, -0.00005665, 0.00004443, -0.00003445, -0.00003721, 0.00002709, -0.00001164, 0.00000909, -0.00003830, 0.00001455, -0.00006291, 0.00002777, -0.00004890, 0.00000561, 0.00000123, -0.00004472, -0.00014148, 0.00001106, -0.00019951, -0.00006044, -0.00132010, 0.00105858, 0.00280476, -0.00085630, -0.00055902, -0.00009171, -0.00016853, 0.00001154, -0.00001607, -0.00001400, -0.00008256, 0.00003421, -0.00002835, 0.00000860, -0.00003836, 0.00003331, -0.00001561, 0.00006264, -0.00009729, 0.00006482, -0.00004385, -0.00001218, 0.00001558, -0.00001716, -0.00002038, -0.00000801, -0.00002053, 0.00001936, -0.00000642, 0.00000124, 0.00002411, -0.00002781, 0.00000927, -0.00001936, -0.00000565, },
		{ -0.00003111, 0.00004166, -0.00001537, -0.00000554, 0.00000400, 0.00000758, -0.00000670, -0.00000416, -0.00001359, 0.00003840, -0.00002121, -0.00000693, 0.00002887, -0.00009043, 0.00005905, -0.00003255, 0.00003054, -0.00006067, 0.00004105, -0.00007995, 0.00003299, -0.00005704, 0.00002081, -0.00001957, -0.00000187, -0.00018379, 0.00005910, -0.00024564, 0.00006030, -0.00120135, 0.00001658, 0.00382674, -0.00096671, -0.00061354, -0.00011690, -0.00016126, -0.00001773, -0.00000534, -0.00001939, -0.00008520, 0.00003716, -0.00002660, -0.00000177, -0.00003127, 0.00002258, -0.00001243, 0.00005000, -0.00006378, 0.00001579, 0.00001671, -0.00006391, 0.00004549, -0.00002838, -0.00000552, -0.00002643, -0.00000361, -0.00000123, 0.00001841, -0.00002774, 0.00004674, -0.00003428, 0.00001154, -0.00002287, 0.00000334, },
		{ -0.00002582, 0.00002062, 0.00001058, -0.00002041, 0.00000991, 0.00000161, 0.00000047, -0.00000392, -0.00001820, 0.00002342, 0.00000938, -0.00003875, 0.00005335, -0.00008855, 0.00003092, 0.00000401, -0.00000393, -0.00002459, 0.00000239, -0.00002514, -0.00003332, 0.00000625, -0.00004177, 0.00004760, -0.00006564, -0.00009362, -0.00005645, -0.00008214, -0.00013703, -0.00061786, -0.00133091, 0.00468573, -0.00071413, -0.00092212, 0.00002431, -0.00028245, 0.00005969, -0.00008693, 0.00005146, -0.00014940, 0.00008671, -0.00005939, 0.00001785, -0.00005122, 0.00003487, -0.00002475, 0.00004528, -0.00003867, -0.00001680, 0.00004629, -0.00007315, 0.00003219, -0.00000617, -0.00001490, -0.00002286, -0.00000721, -0.00000256, 0.00002306, -0.00003233, 0.00004047, -0.00001488, -0.00000452, -0.00001272, -0.00000119, },
		{ -0.00000945, -0.00000654, 0.00003107, -0.00002294, 0.00000347, 0.00000819, -0.00000776, 0.00001261, -0.00003537, 0.00002343, 0.00001076, -0.00002429, 0.00001648, -0.00001919, -0.00005908, 0.00009395, -0.00008887, 0.00006441, -0.00009417, 0.00009016, -0.00015703, 0.00012790, -0.00017120, 0.00019301, -0.00022514, 0.00011220, -0.00030818, 0.00024631, -0.00055783, 0.00020011, -0.00259021, 0.00498599, 0.00013550, -0.00158312, 0.00038114, -0.00056554, 0.00026794, -0.00027712, 0.00020915, -0.00028317, 0.00019129, -0.00013873, 0.00008299, -0.00011494, 0.00008715, -0.00007157, 0.00007285, -0.00005262, 0.00000106, 0.00001452, -0.00001956, -0.00003400, 0.00005370, -0.00005321, 0.00000907, -0.00003821, 0.00002253, 0.00000035, -0.00000705, 0.00000483, 0.00002554, -0.00003253, 0.00000606, -0.00001504, },
		{ 0.00000539, -0.00002504, 0.00003340, -0.00000604, -0.00001829, 0.00002855, -0.00003099, 0.00004340, -0.00006443, 0.00004168, -0.00002063, 0.00003364, -0.00006701, 0.00008906, -0.00017263, 0.00019572, -0.00018282, 0.00016395, -0.00020260, 0.00021469, -0.00028373, 0.00025287, -0.00030996, 0.00035310, -0.00040594, 0.00034185, -0.00058044, 0.00058931, -0.00098412, 0.00091381, -0.00334669, 0.00443451, 0.00164670, -0.00253608, 0.00087925, -0.00094688, 0.00055140, -0.00052566, 0.00040931, -0.00044954, 0.00032231, -0.00024487, 0.00017992, -0.00021137, 0.00017057, -0.00014806, 0.00013345, -0.00010975, 0.00007132, -0.00007202, 0.00007954, -0.00012942, 0.00012836, -0.00010282, 0.00005558, -0.00008428, 0.00006251, -0.00003826, 0.00003468, -0.00004297, 0.00006798, -0.00005609, 0.00002093, -0.00002724, },
		{ 0.00000885, -0.00002646, 0.00001525, 0.00002520, -0.00004627, 0.00005198, -0.00005648, 0.00007382, -0.00009102, 0.00006376, -0.00006447, 0.00010289, -0.00015129, 0.00017950, -0.00024937, 0.00025109, -0.00022989, 0.00021726, -0.00026220, 0.00028323, -0.00034605, 0.00031216, -0.00038332, 0.00044288, -0.00050809, 0.00047526, -0.00072876, 0.00076977, -0.00118540, 0.00122733, -0.00333834, 0.00298970, 0.00362566, -0.00352642, 0.00131310, -0.00126809, 0.00078069, -0.00072346, 0.00056031, -0.00057266, 0.00041770, -0.00032681, 0.00026379, -0.00029858, 0.00024647, -0.00022015, 0.00019703, -0.00017979, 0.00015800, -0.00016860, 0.00017373, -0.00020525, 0.00017692, -0.00013173, 0.00008940, -0.00011977, 0.00009299, -0.00006899, 0.00006825, -0.00007738, 0.00008956, -0.00005860, 0.00002045, -0.00002820, },
		{ 0.00000030, -0.00001379, -0.00001321, 0.00005484, -0.00006317, 0.00006121, -0.00006587, 0.00008429, -0.00009528, 0.00006715, -0.00008956, 0.00014094, -0.00018521, 0.00019876, -0.00024090, 0.00021900, -0.00019310, 0.00018671, -0.00023248, 0.00025322, -0.00030142, 0.00026199, -0.00034154, 0.00040338, -0.00046123, 0.00042949, -0.00065793, 0.00067874, -0.00103490, 0.00101947, -0.00257095, 0.00091583, 0.00563858, -0.00416700, 0.00141559, -0.00133907, 0.00080757, -0.00075014, 0.00056393, -0.00057193, 0.00040990, -0.00032553, 0.00027895, -0.00032243, 0.00026362, -0.00024025, 0.00021759, -0.00021384, 0.00020620, -0.00021589, 0.00020592, -0.00021384, 0.00016381, -0.00011317, 0.00008659, -0.00012132, 0.00009172, -0.00007054, 0.00007277, -0.00007938, 0.00007708, -0.00003418, 0.00000281, -0.00001680, },
		{ -0.00001203, 0.00000130, -0.00003538, 0.00006497, -0.00005393, 0.00004373, -0.00004720, 0.00006303, -0.00006560, 0.00003730, -0.00007439, 0.00012084, -0.00014250, 0.00012816, -0.00014028, 0.00010202, -0.00007848, 0.00007752, -0.00011756, 0.00012975, -0.00015755, 0.00011046, -0.00018938, 0.00023576, -0.00026447, 0.00020461, -0.00037405, 0.00033334, -0.00057023, 0.00039081, -0.00131738, -0.00129282, 0.00714663, -0.00408468, 0.00098355, -0.00103770, 0.00054541, -0.00054062, 0.00037075, -0.00040830, 0.00026544, -0.00020910, 0.00019158, -0.00024801, 0.00018854, -0.00017636, 0.00016213, -0.00017487, 0.00017533, -0.00017488, 0.00014605, -0.00013787, 0.00008223, -0.00004449, 0.00004294, -0.00008296, 0.00005291, -0.00003765, 0.00004381, -0.00004754, 0.00003512, 0.00000767, -0.00002211, -0.00000129, },
		{ -0.00001769, 0.00000682, -0.00003860, 0.00004699, -0.00001644, 0.00000158, -0.00000422, 0.00001532, -0.00000851, -0.00002062, -0.00002166, 0.00004769, -0.00003761, -0.00000412, 0.00001135, -0.00005196, 0.00006572, -0.00006364, 0.00003528, -0.00003635, 0.00003028, -0.00008505, 0.00001513, 0.00000177, 0.00001233, -0.00011575, 0.00001838, -0.00013317, 0.00003140, -0.00039439, -0.00000040, -0.00309088, 0.00770638, -0.00309088, -0.00000040, -0.00039439, 0.00003140, -0.00013317, 0.00001838, -0.00011575, 0.00001233, 0.00000177, 0.00001513, -0.00008505, 0.00003028, -0.00003635, 0.00003528, -0.00006364, 0.00006572, -0.00005196, 0.00001135, -0.00000412, -0.00003761, 0.00004769, -0.00002166, -0.00002062, -0.00000851, 0.00001532, -0.00000422, 0.00000158, -0.00001644, 0.00004699, -0.00003860, 0.00000682, },
		{ -0.00001203, -0.00000129, -0.00002211, 0.00000767, 0.00003512, -0.00004754, 0.00004381, -0.00003765, 0.00005291, -0.00008296, 0.00004294, -0.00004449, 0.00008223, -0.00013787, 0.00014605, -0.00017488, 0.00017533, -0.00017487, 0.00016213, -0.00017636, 0.00018854, -0.00024801, 0.00019158, -0.00020910, 0.00026544, -0.00040830, 0.00037075, -0.00054062, 0.00054541, -0.00103770, 0.00098355, -0.00408468, 0.00714663, -0.00129282, -0.00131738, 0.00039081, -0.00057023, 0.00033334, -0.00037405, 0.00020461, -0.00026447, 0.00023576, -0.00018938, 0.00011046, -0.00015755, 0.00012975, -0.00011756, 0.00007752, -0.00007848, 0.00010202, -0.00014028, 0.00012816, -0.00014250, 0.00012084, -0.00007439, 0.00003730, -0.00006560, 0.00006303, -0.00004720, 0.00004373, -0.00005393, 0.00006497, -0.00003538, 0.00000130, },
		{ 0.00000030, -0.00001680, 0.00000281, -0.00003418, 0.00007708, -0.00007938, 0.00007277, -0.00007054, 0.00009172, -0.00012132, 0.00008659, -0.00011317, 0.00016381, -0.00021384, 0.00020592, -0.00021589, 0.00020620, -0.00021384, 0.00021759, -0.00024025, 0.00026362, -0.00032243, 0.00027895, -0.00032553, 0.00040990, -0.00057193, 0.00056393, -0.00075014, 0.00080757, -0.00133907, 0.00141559, -0.00416700, 0.00563858, 0.00091583, -0.00257095, 0.00101947, -0.00103490, 0.00067874, -0.00065793, 0.00042949, -0.00046123, 0.00040338, -0.00034154, 0.00026199, -0.00030142, 0.00025322, -0.00023248, 0.00018671, -0.00019310, 0.00021900, -0.00024090, 0.00019876, -0.00018521, 0.00014094, -0.00008956, 0.00006715, -0.00009528, 0.00008429, -0.00006587, 0.00006121, -0.00006317, 0.00005484, -0.00001321, -0.00001379, },
		{ 0.00000885, -0.00002820, 0.00002045, -0.00005860, 0.00008956, -0.00007738, 0.00006825, -0.00006899, 0.00009299, -0.00011977, 0.00008940, -0.00013173, 0.00017692, -0.00020525, 0.00017373, -0.00016860, 0.00015800, -0.00017979, 0.00019703, -0.00022015, 0.00024647, -0.00029858, 0.00026379, -0.00032681, 0.00041770, -0.00057266, 0.00056031, -0.00072346, 0.00078069, -0.00126809, 0.00131310, -0.00352642, 0.00362566, 0.00298970, -0.00333834, 0.00122733, -0.00118540, 0.00076977, -0.00072876, 0.00047526, -0.00050809, 0.00044288, -0.00038332, 0.00031216, -0.00034605, 0.00028323, -0.00026220, 0.00021726, -0.00022989, 0.00025109, -0.00024937, 0.00017950, -0.00015129, 0.00010289, -0.00006447, 0.00006376, -0.00009102, 0.00007382, -0.00005648, 0.00005198, -0.00004627, 0.00002520, 0.00001525, -0.00002646, },
		{ 0.00000539, -0.00002724, 0.00002093, -0.00005609, 0.00006798, -0.00004297, 0.00003468, -0.00003826, 0.00006251, -0.00008428, 0.00005558, -0.00010282, 0.00012836, -0.00012942, 0.00007954, -0.00007202, 0.00007132, -0.00010975, 0.00013345, -0.00014806, 0.00017057, -0.00021137, 0.00017992, -0.00024487, 0.00032231, -0.00044954, 0.00040931, -0.00052566, 0.00055140, -0.00094688, 0.00087925, -0.00253608, 0.00164670, 0.00443451, -0.00334669, 0.00091381, -0.00098412, 0.00058931, -0.00058044, 0.00034185, -0.00040594, 0.00035310, -0.00030996, 0.00025287, -0.00028373, 0.00021469, -0.00020260, 0.00016395, -0.00018282, 0.00019572, -0.00017263, 0.00008906, -0.00006701, 0.00003364, -0.00002063, 0.00004168, -0.00006443, 0.00004340, -0.00003099, 0.00002855, -0.00001829, -0.00000604, 0.00003340, -0.00002504, },
		{ -0.00000945, -0.00001504, 0.00000606, -0.00003253, 0.00002554, 0.00000483, -0.00000705, 0.00000035, 0.00002253, -0.00003821, 0.00000907, -0.00005321, 0.00005370, -0.00003400, -0.00001956, 0.00001452, 0.00000106, -0.00005262, 0.00007285, -0.00007157, 0.00008715, -0.00011494, 0.00008299, -0.00013873, 0.00019129, -0.00028317, 0.00020915, -0.00027712, 0.00026794, -0.00056554, 0.00038114, -0.00158312, 0.00013550, 0.00498599, -0.00259021, 0.00020011, -0.00055783, 0.00024631, -0.00030818, 0.00011220, -0.00022514, 0.00019301, -0.00017120, 0.00012790, -0.00015703, 0.00009016, -0.00009417, 0.00006441, -0.00008887, 0.00009395, -0.00005908, -0.00001919, 0.00001648, -0.00002429, 0.00001076, 0.00002343, -0.00003537, 0.00001261, -0.00000776, 0.00000819, 0.00000347, -0.00002294, 0.00003107, -0.00000654, },
		{ -0.00002582, -0.00000119, -0.00001272, -0.00000452, -0.00001488, 0.00004047, -0.00003233, 0.00002306, -0.00000256, -0.00000721, -0.00002286, -0.00001490, -0.00000617, 0.00003219, -0.00007315, 0.00004629, -0.00001680, -0.00003867, 0.00004528, -0.00002475, 0.00003487, -0.00005122, 0.00001785, -0.00005939, 0.00008671, -0.00014940, 0.00005146, -0.00008693, 0.00005969, -0.00028245, 0.00002431, -0.00092212, -0.00071413, 0.00468573, -0.00133091, -0.00061786, -0.00013703, -0.00008214, -0.00005645, -0.00009362, -0.00006564, 0.00004760, -0.00004177, 0.00000625, -0.00003332, -0.00002514, 0.00000239, -0.00002459, -0.00000393, 0.00000401, 0.00003092, -0.00008855, 0.00005335, -0.00003875, 0.00000938, 0.00002342, -0.00001820, -0.00000392, 0.00000047, 0.00000161, 0.00000991, -0.00002041, 0.00001058, 0.00002062, },
		{ -0.00003111, 0.00000334, -0.00002287, 0.00001154, -0.00003428, 0.00004674, -0.00002774, 0.00001841, -0.00000123, -0.00000361, -0.00002643, -0.00000552, -0.00002838, 0.00004549, -0.00006391, 0.00001671, 0.00001579, -0.00006378, 0.00005000, -0.00001243, 0.00002258, -0.00003127, -0.00000177, -0.00002660, 0.00003716, -0.00008520, -0.00001939, -0.00000534, -0.00001773, -0.00016126, -0.00011690, -0.00061354, -0.00096671, 0.00382674, 0.00001658, -0.00120135, 0.00006030, -0.00024564, 0.00005910, -0.00018379, -0.00000187, -0.00001957, 0.00002081, -0.00005704, 0.00003299, -0.00007995, 0.00004105, -0.00006067, 0.00003054, -0.00003255, 0.00005905, -0.00009043, 0.00002887, -0.00000693, -0.00002121, 0.00003840, -0.00001359, -0.00000416, -0.00000670, 0.00000758, 0.00000400, -0.00000554, -0.00001537, 0.00004166, },
		{ -0.00001867, -0.00000565, -0.00001936, 0.00000927, -0.00002781, 0.00002411, 0.00000124, -0.00000642, 0.00001936, -0.00002053, -0.00000801, -0.00002038, -0.00001716, 0.00001558, -0.00001218, -0.00004385, 0.00006482, -0.00009729, 0.00006264, -0.00001561, 0.00003331, -0.00003836, 0.00000860, -0.00002835, 0.00003421, -0.00008256, -0.00001400, -0.00001607, 0.00001154, -0.00016853, -0.00009171, -0.00055902, -0.00085630, 0.00280476, 0.00105858, -0.00132010, -0.00006044, -0.00019951, 0.00001106, -0.00014148, -0.00004472, 0.00000123, 0.00000561, -0.00004890, 0.00002777, -0.00006291, 0.00001455, -0.00003830, 0.00000909, -0.00001164, 0.00002709, -0.00003721, -0.00003445, 0.00004443, -0.00005665, 0.00005114, -0.00001120, 0.00000336, -0.00001965, 0.00001612, -0.00000386, 0.00000896, -0.00003228, 0.00004477, },
		{ 0.00000706, -0.00002209, -0.00000757, -0.00000519, -0.00000579, -0.00001096, 0.00003463, -0.00003141, 0.00003995, -0.00003884, 0.00001296, -0.00003976, 0.00000478, -0.00002670, 0.00004184, -0.00009084, 0.00008939, -0.00010622, 0.00005699, -0.00001114, 0.00004298, -0.00004652, 0.00002242, -0.00003846, 0.00004983, -0.00010728, 0.00002430, -0.00006412, 0.00007906, -0.00022004, -0.00000897, -0.00059654, -0.00064841, 0.00195224, 0.00157126, -0.00094849, -0.00042426, -0.00002342, -0.00013284, -0.00002625, -0.00014253, 0.00006719, -0.00005210, 0.00000085, -0.00002145, -0.00000330, -0.00004629, 0.00001281, -0.00003987, 0.00003698, -0.00003001, 0.00002971, -0.00009321, 0.00007728, -0.00006885, 0.00004425, -0.00000027, 0.00000844, -0.00002734, 0.00001714, -0.00000453, 0.00001388, -0.00003278, 0.00002849, },
		{ 0.00003294, -0.00003365, 0.00000132, -0.00001900, 0.00001454, -0.00003717, 0.00005098, -0.00003757, 0.00004320, -0.00004145, 0.00001883, -0.00004476, 0.00001463, -0.00005180, 0.00006462, -0.00009411, 0.00006873, -0.00007863, 0.00002506, 0.00000972, 0.00003909, -0.00004000, 0.00002224, -0.00003746, 0.00005940, -0.00012600, 0.00005231, -0.00009587, 0.00012087, -0.00024016, 0.00003745, -0.00059676, -0.00052430, 0.00143065, 0.00155941, -0.00026174, -0.00082886, 0.00013018, -0.00025798, 0.00006876, -0.00021865, 0.00011531, -0.00009869, 0.00004407, -0.00006840, 0.00005185, -0.00009555, 0.00005027, -0.00007640, 0.00007301, -0.00006968, 0.00006752, -0.00011079, 0.00006681, -0.00004524, 0.00001408, 0.00002017, 0.00000780, -0.00002530, 0.00000777, 0.00000330, 0.00000899, -0.00002023, 0.00000236, },
		{ 0.00004522, -0.00003015, -0.00000108, -0.00002221, 0.00002049, -0.00004136, 0.00004061, -0.00001933, 0.00002542, -0.00002483, 0.00000557, -0.00003032, 0.00000450, -0.00004893, 0.00004777, -0.00005418, 0.00001388, -0.00003066, -0.00001826, 0.00003660, 0.00002673, -0.00002079, 0.00000841, -0.00002307, 0.00005514, -0.00012401, 0.00004978, -0.00008696, 0.00011021, -0.00019988, 0.00001316, -0.00051384, -0.00053505, 0.00121841, 0.00120957, 0.00045921, -0.00105081, 0.00012736, -0.00027574, 0.00007626, -0.00022077, 0.00010393, -0.00009804, 0.00004687, -0.00007906, 0.00006798, -0.00010149, 0.00004689, -0.00007572, 0.00007195, -0.00006814, 0.00005726, -0.00007864, 0.00001656, 0.00000241, -0.00002519, 0.00003934, 0.00000682, -0.00001740, -0.00000658, 0.00001324, 0.00000150, -0.00000498, -0.00001954, },
		{ 0.00003807, -0.00001082, -0.00001412, -0.00001426, 0.00001103, -0.00002523, 0.00001058, 0.00001243, -0.00000171, -0.00000028, -0.00001603, -0.00000677, -0.00001628, -0.00002838, 0.00000808, 0.00000228, -0.00004209, 0.00000625, -0.00004784, 0.00005071, 0.00002009, -0.00000167, -0.00000664, -0.00000587, 0.00004424, -0.00010602, 0.00002208, -0.00004577, 0.00006054, -0.00011825, -0.00005778, -0.00037779, -0.00062928, 0.00117981, 0.00077453, 0.00096067, -0.00095753, -0.00006921, -0.00017688, -0.00000626, -0.00015090, 0.00003685, -0.00005235, 0.00000881, -0.00005043, 0.00004143, -0.00006370, 0.00000578, -0.00004184, 0.00003794, -0.00003175, 0.00001165, -0.00001868, -0.00004502, 0.00004563, -0.00005043, 0.00004256, 0.00001365, -0.00001089, -0.00001732, 0.00001692, -0.00000052, 0.00000347, -0.00002741, },
		{ 0.00001683, 0.00001491, -0.00002848, -0.00000324, -0.00000466, -0.00000210, -0.00002114, 0.00003828, -0.00001998, 0.00001528, -0.00002965, 0.00000961, -0.00003103, -0.00000980, -0.00002778, 0.00004242, -0.00006709, 0.00000798, -0.00004794, 0.00004119, 0.00002823, 0.00000712, -0.00001145, 0.00000293, 0.00003744, -0.00008478, -0.00001258, 0.00000247, 0.00000515, -0.00003646, -0.00012613, -0.00024925, -0.00072075, 0.00116449, 0.00045385, 0.00111924, -0.00056691, -0.00037898, -0.00003383, -0.00012081, -0.00005871, -0.00004384, 0.00000415, -0.00004189, -0.00000701, -0.00000389, -0.00000862, -0.00004534, -0.00000110, -0.00000351, 0.00001198, -0.00003719, 0.00003302, -0.00008348, 0.00005777, -0.00004534, 0.00002299, 0.00003060, -0.00000929, -0.00001959, 0.00001071, 0.00000540, 0.00000271, -0.00002103, },
		{ -0.00000602, 0.00003334, -0.00003291, 0.00000106, -0.00001510, 0.00001310, -0.00003770, 0.00004314, -0.00001743, 0.00001176, -0.00002585, 0.00000916, -0.00002929, -0.00000665, -0.00004196, 0.00004813, -0.00005016, -0.00002590, -0.00002412, 0.00001428, 0.00004673, 0.00000654, -0.00000475, 0.00000120, 0.00003863, -0.00006875, -0.00003899, 0.00003606, -0.00002836, 0.00001331, -0.00015518, -0.00017207, -0.00075046, 0.00108280, 0.00032656, 0.00096988, -0.00002774, -0.00064914, 0.00004756, -0.00019140, -0.00000489, -0.00008925, 0.00003208, -0.00007153, 0.00002020, -0.00003676, 0.00003093, -0.00007486, 0.00001813, -0.00002556, 0.00003538, -0.00006033, 0.00004958, -0.00007988, 0.00003121, -0.00001210, -0.00001148, 0.00005014, -0.00000909, -0.00001542, -0.00000175, 0.00001452, -0.00000211, -0.00000842, },
		{ -0.00001907, 0.00003530, -0.00002232, -0.00000520, -0.00001483, 0.00001335, -0.00003335, 0.00002553, 0.00000374, -0.00000720, -0.00000851, -0.00000453, -0.00001404, -0.00001768, -0.00003485, 0.00002447, -0.00000618, -0.00007252, 0.00000069, -0.00001134, 0.00006123, 0.00000679, 0.00000526, -0.00000382, 0.00004303, -0.00005761, -0.00005289, 0.00004823, -0.00003248, 0.00002469, -0.00013998, -0.00015074, -0.00071256, 0.00092710, 0.00035578, 0.00065729, 0.00046182, -0.00073845, -0.00000417, -0.00017586, -0.00002008, -0.00007713, 0.00001477, -0.00006515, 0.00001588, -0.00004032, 0.00003721, -0.00006785, 0.00000465, -0.00001825, 0.00002848, -0.00004938, 0.00002832, -0.00004118, -0.00001806, 0.00002917, -0.00004179, 0.00005887, -0.00000289, -0.00001102, -0.00001272, 0.00001897, -0.00000336, 0.00000083, },
		{ -0.00001866, 0.00002152, -0.00000145, -0.00001707, -0.00000735, 0.00000248, -0.00001535, -0.00000261, 0.00002924, -0.00002731, 0.00000868, -0.00001835, 0.00000195, -0.00003048, -0.00002067, -0.00000775, 0.00003627, -0.00010173, 0.00000278, -0.00001899, 0.00005920, 0.00001758, 0.00000930, -0.00000303, 0.00004320, -0.00004629, -0.00005860, 0.00004555, -0.00001777, 0.00001362, -0.00010177, -0.00015860, -0.00064113, 0.00074423, 0.00044559, 0.00034756, 0.00074976, -0.00059128, -0.00018028, -0.00009355, -0.00008647, -0.00002506, -0.00003177, -0.00003520, -0.00001098, -0.00002082, 0.00001620, -0.00003343, -0.00002992, 0.00000691, 0.00000277, -0.00001845, -0.00001084, 0.00000611, -0.00006110, 0.00005292, -0.00004956, 0.00004711, 0.00001348, -0.00001079, -0.00001642, 0.00001394, 0.00000272, 0.00000231, },
		{ -0.00000985, 0.00000143, 0.00001822, -0.00002470, -0.00000118, -0.00000950, 0.00000258, -0.00002462, 0.00004268, -0.00003405, 0.00001261, -0.00002001, 0.00000651, -0.00003257, -0.00001453, -0.00002811, 0.00005413, -0.00009570, -0.00002557, -0.00000669, 0.00003982, 0.00003950, 0.00000529, 0.00000662, 0.00003647, -0.00003192, -0.00006169, 0.00003876, -0.00000072, 0.00000251, -0.00006800, -0.00016269, -0.00057748, 0.00058827, 0.00050866, 0.00015058, 0.00079014, -0.00026202, -0.00039466, -0.00001189, -0.00015283, 0.00002421, -0.00007211, -0.00001008, -0.00003668, 0.00000057, -0.00001061, 0.00000439, -0.00006122, 0.00002745, -0.00001981, 0.00000851, -0.00004073, 0.00003383, -0.00007426, 0.00004433, -0.00002941, 0.00001696, 0.00003571, -0.00001309, -0.00001306, 0.00000132, 0.00001301, -0.00000097, },
		{ -0.00000137, -0.00001360, 0.00002618, -0.00002075, -0.00000257, -0.00001446, 0.00000987, -0.00002980, 0.00003626, -0.00002258, 0.00000008, -0.00000706, -0.00000278, -0.00002107, -0.00002130, -0.00002943, 0.00004288, -0.00005912, -0.00007084, 0.00001132, 0.00001433, 0.00006336, -0.00000027, 0.00002011, 0.00002728, -0.00001695, -0.00006382, 0.00003466, 0.00000776, 0.00000529, -0.00005413, -0.00014550, -0.00054317, 0.00048549, 0.00050750, 0.00008849, 0.00064299, 0.00011896, -0.00053208, -0.00000176, -0.00017147, 0.00003342, -0.00007725, -0.00001245, -0.00004164, 0.00000511, -0.00002328, 0.00002413, -0.00006944, 0.00002656, -0.00002329, 0.00001471, -0.00004467, 0.00002917, -0.00005313, 0.00000845, 0.00000696, -0.00001729, 0.00005229, -0.00001138, -0.00000792, -0.00001190, 0.00002010, -0.00000207, },
		{ 0.00000113, -0.00001761, 0.00001956, -0.00000586, -0.00001068, -0.00001148, 0.00000501, -0.00001903, 0.00001493, -0.00000051, -0.00002048, 0.00001188, -0.00001758, -0.00000404, -0.00003384, -0.00001935, 0.00001576, -0.00001394, -0.00010774, 0.00001537, -0.00000335, 0.00007757, 0.00000180, 0.00002928, 0.00002308, -0.00000655, -0.00006320, 0.00003413, 0.00000635, 0.00002219, -0.00005794, -0.00011100, -0.00053331, 0.00042729, 0.00045603, 0.00011452, 0.00042449, 0.00041398, -0.00051160, -0.00009073, -0.00013133, -0.00000416, -0.00004432, -0.00004312, -0.00002438, -0.00001032, -0.00001660, 0.00001995, -0.00005155, 0.00000400, -0.00000774, 0.00000019, -0.00002497, 0.00000027, -0.00001355, -0.00003359, 0.00003783, -0.00003745, 0.00005187, 0.00000000, -0.00000625, -0.00001899, 0.00001801, 0.00000373, },
		{ -0.00000118, -0.00001281, 0.00000431, 0.00001181, -0.00001853, -0.00000603, -0.00000555, -0.00000223, -0.00000800, 0.00001809, -0.00003557, 0.00002400, -0.00002572, 0.00000666, -0.00004048, -0.00001141, -0.00000815, 0.00001625, -0.00011626, -0.00000472, -0.00000794, 0.00007701, 0.00001564, 0.00002992, 0.00002791, -0.00000329, -0.00005854, 0.00003527, -0.00000015, 0.00004454, -0.00006635, -0.00007557, -0.00052829, 0.00038644, 0.00039277, 0.00015962, 0.00024187, 0.00054153, -0.00032839, -0.00024374, -0.00006534, -0.00006211, 0.00000312, -0.00008172, -0.00000093, -0.00003308, -0.00000088, 0.00000231, -0.00002083, -0.00002559, 0.00001312, -0.00002126, 0.00000185, -0.00003179, 0.00001986, -0.00005821, 0.00004509, -0.00003291, 0.00003165, 0.00001996, -0.00000841, -0.00001827, 0.00000679, 0.00001488, },
		{ -0.00000222, -0.00000649, -0.00000966, 0.00002239, -0.00001876, -0.00000443, -0.00001363, 0.00000951, -0.00002022, 0.00002237, -0.00003629, 0.00002171, -0.00002057, 0.00000471, -0.00003486, -0.00001378, -0.00001741, 0.00002073, -0.00009422, -0.00004170, -0.00000704, 0.00006603, 0.00003736, 0.00002495, 0.00003928, -0.00000432, -0.00005180, 0.00003721, -0.00000658, 0.00006311, -0.00006663, -0.00005380, -0.00051114, 0.00033944, 0.00034952, 0.00017516, 0.00014869, 0.00050275, -0.00005225, -0.00038213, -0.00002786, -0.00010144, 0.00003311, -0.00010250, 0.00000852, -0.00004624, 0.00000841, -0.00001242, 0.00000426, -0.00004423, 0.00002333, -0.00003385, 0.00001819, -0.00004779, 0.00002944, -0.00005380, 0.00002533, -0.00000768, 0.00000067, 0.00003959, -0.00000920, -0.00001371, -0.00000761, 0.00002455, },
		{ 0.00000342, -0.00000466, -0.00001547, 0.00002110, -0.00000955, -0.00000836, -0.00001550, 0.00001098, -0.00001776, 0.00001148, -0.00002405, 0.00000760, -0.00000550, -0.00000631, -0.00002056, -0.00002393, -0.00001390, 0.00000582, -0.00005709, -0.00007558, -0.00001472, 0.00005320, 0.00005869, 0.00002169, 0.00005044, -0.00000331, -0.00004725, 0.00004088, -0.00001057, 0.00007365, -0.00005383, -0.00004971, -0.00047801, 0.00027949, 0.00033482, 0.00015059, 0.00013807, 0.00036337, 0.00021249, -0.00042830, -0.00005772, -0.00009899, 0.00002813, -0.00009292, -0.00000488, -0.00004202, 0.00000310, -0.00001443, 0.00001273, -0.00004289, 0.00001621, -0.00003096, 0.00001680, -0.00004178, 0.00001401, -0.00002661, -0.00000868, 0.00002204, -0.00002494, 0.00004721, -0.00000215, -0.00001108, -0.00001771, 0.00002576, },
		{ 0.00001558, -0.00000740, -0.00001341, 0.00001111, 0.00000396, -0.00001376, -0.00001339, 0.00000466, -0.00000589, -0.00000609, -0.00000862, -0.00000821, 0.00000918, -0.00001634, -0.00000752, -0.00003245, -0.00000804, -0.00001284, -0.00002641, -0.00008735, -0.00003958, 0.00004274, 0.00007350, 0.00002605, 0.00005577, 0.00000469, -0.00004778, 0.00004708, -0.00001199, 0.00007698, -0.00003089, -0.00005721, -0.00043698, 0.00021471, 0.00033707, 0.00010447, 0.00016726, 0.00020894, 0.00037968, -0.00034824, -0.00015147, -0.00006385, -0.00000377, -0.00006176, -0.00003255, -0.00002694, -0.00001239, -0.00000606, 0.00000660, -0.00002606, -0.00000211, -0.00001820, 0.00000338, -0.00002216, -0.00001295, 0.00000486, -0.00003673, 0.00003811, -0.00003182, 0.00003636, 0.00001450, -0.00001269, -0.00001988, 0.00001660, },
		{ 0.00002813, -0.00000931, -0.00000941, 0.00000049, 0.00001353, -0.00001463, -0.00001206, -0.00000319, 0.00000612, -0.00001925, -0.00000066, -0.00001584, 0.00001421, -0.00001685, -0.00000385, -0.00003162, -0.00000883, -0.00002255, -0.00001564, -0.00007168, -0.00007647, 0.00003001, 0.00008128, 0.00003887, 0.00005464, 0.00001978, -0.00005214, 0.00005436, -0.00001085, 0.00007602, -0.00000397, -0.00006669, -0.00039957, 0.00015755, 0.00033877, 0.00006262, 0.00019197, 0.00010156, 0.00041739, -0.00016765, -0.00026360, -0.00003076, -0.00003669, -0.00003139, -0.00005568, -0.00001572, -0.00002618, 0.00000256, -0.00000356, -0.00000649, -0.00001865, -0.00000725, -0.00001004, -0.00000350, -0.00003390, 0.00002245, -0.00004403, 0.00003164, -0.00001771, 0.00001097, 0.00003490, -0.00001518, -0.00001643, 0.00000170, },
		{ 0.00003321, -0.00000434, -0.00000946, -0.00000405, 0.00001409, -0.00000846, -0.00001382, -0.00000810, 0.00001173, -0.00002167, -0.00000446, -0.00001247, 0.00000776, -0.00000706, -0.00000974, -0.00002169, -0.00001687, -0.00002139, -0.00002338, -0.00003968, -0.00010861, 0.00000557, 0.00008437, 0.00005702, 0.00005088, 0.00003767, -0.00005563, 0.00005932, -0.00000635, 0.00007301, 0.00002225, -0.00007158, -0.00037266, 0.00011487, 0.00032942, 0.00004022, 0.00018960, 0.00005978, 0.00035094, 0.00004520, -0.00033271, -0.00003566, -0.00004691, -0.00002097, -0.00005949, -0.00001926, -0.00002925, 0.00000296, -0.00000812, 0.00000472, -0.00002339, -0.00000643, -0.00001472, 0.00000397, -0.00003854, 0.00001905, -0.00002958, 0.00000792, 0.00000727, -0.00001626, 0.00004836, -0.00001213, -0.00001303, -0.00001116, },
		{ 0.00002697, 0.00000869, -0.00001433, -0.00000198, 0.00000736, 0.00000135, -0.00001639, -0.00001014, 0.00001044, -0.00001482, -0.00001598, -0.00000328, -0.00000421, 0.00000620, -0.00001803, -0.00001011, -0.00002517, -0.00001661, -0.00003758, -0.00001044, -0.00011888, -0.00003520, 0.00008204, 0.00007726, 0.00004931, 0.00005315, -0.00005337, 0.00005916, 0.00000204, 0.00006907, 0.00004622, -0.00007059, -0.00035602, 0.00008511, 0.00030928, 0.00003692, 0.00016237, 0.00006568, 0.00023797, 0.00021665, -0.00031647, -0.00009039, -0.00003057, -0.00003315, -0.00004360, -0.00003631, -0.00002249, -0.00000497, -0.00000528, 0.00000501, -0.00001561, -0.00001512, -0.00001059, -0.00000018, -0.00002853, 0.00000105, -0.00000515, -0.00001801, 0.00002737, -0.00003145, 0.00004623, 0.00000036, -0.00001362, -0.00001655, },
		{ 0.00001234, 0.00002441, -0.00001885, 0.00000135, 0.00000021, 0.00000801, -0.00001526, -0.00001215, 0.00000597, -0.00000507, -0.00002691, 0.00000333, -0.00001322, 0.00001424, -0.00002030, -0.00000512, -0.00002612, -0.00001651, -0.00004623, 0.00000216, -0.00010190, -0.00008473, 0.00006741, 0.00009831, 0.00005304, 0.00006335, -0.00004339, 0.00005387, 0.00001330, 0.00006500, 0.00006849, -0.00006593, -0.00034513, 0.00006200, 0.00028480, 0.00004329, 0.00012580, 0.00008743, 0.00013410, 0.00030227, -0.00021389, -0.00017347, -0.00000625, -0.00005382, -0.00002118, -0.00005498, -0.00001529, -0.00001394, -0.00000055, -0.00000019, -0.00000258, -0.00002564, -0.00000428, -0.00000894, -0.00001368, -0.00001800, 0.00001356, -0.00003151, 0.00003126, -0.00002809, 0.00002810, 0.00001950, -0.00001688, -0.00001480, },
		{ -0.00000301, 0.00003419, -0.00001606, -0.00000028, -0.00000104, 0.00000689, -0.00000839, -0.00001564, 0.00000178, 0.00000179, -0.00003111, 0.00000239, -0.00001480, 0.00001298, -0.00001337, -0.00000914, -0.00001822, -0.00002290, -0.00004596, -0.00000259, -0.00006799, -0.00012505, 0.00003161, 0.00011840, 0.00006369, 0.00006831, -0.00002715, 0.00004633, 0.00002492, 0.00006181, 0.00009002, -0.00006020, -0.00033574, 0.00004022, 0.00026174, 0.00005017, 0.00009434, 0.00010063, 0.00006937, 0.00030063, -0.00006305, -0.00024241, -0.00000112, -0.00006438, -0.00000810, -0.00006233, -0.00001735, -0.00001626, -0.00000047, -0.00000376, 0.00000718, -0.00003010, -0.00000248, -0.00001475, -0.00000376, -0.00002704, 0.00001705, -0.00002735, 0.00001864, -0.00001077, 0.00000275, 0.00003643, -0.00001718, -0.00001093, },
		{ -0.00001205, 0.00003207, -0.00000312, -0.00000831, 0.00000424, -0.00000044, 0.00000105, -0.00001849, -0.00000185, 0.00000426, -0.00002821, -0.00000464, -0.00001101, 0.00000517, -0.00000119, -0.00001735, -0.00000722, -0.00003008, -0.00004243, -0.00001438, -0.00003600, -0.00013843, -0.00002708, 0.00013056, 0.00008340, 0.00006949, -0.00000774, 0.00004033, 0.00003450, 0.00006017, 0.00011150, -0.00005456, -0.00032629, 0.00001825, 0.00024151, 0.00005386, 0.00007384, 0.00009732, 0.00004445, 0.00024323, 0.00008182, -0.00025892, -0.00003086, -0.00005668, -0.00001103, -0.00005405, -0.00003065, -0.00001065, -0.00000637, -0.00000316, 0.00001019, -0.00002624, -0.00000657, -0.00001517, -0.00000234, -0.00002359, 0.00000666, -0.00001170, -0.00000070, 0.00000840, -0.00001709, 0.00004169, -0.00000926, -0.00001003, },
		{ -0.00001292, 0.00001892, 0.00001563, -0.00001752, 0.00001037, -0.00000709, 0.00000651, -0.00001659, -0.00000680, 0.00000435, -0.00002229, -0.00001212, -0.00000754, -0.00000296, 0.00000906, -0.00002212, -0.00000108, -0.00003058, -0.00004340, -0.00002170, -0.00002048, -0.00011981, -0.00009627, 0.00012120, 0.00011464, 0.00006892, 0.00001192, 0.00003845, 0.00004130, 0.00005970, 0.00013342, -0.00004876, -0.00031770, -0.00000236, 0.00022218, 0.00005533, 0.00006246, 0.00008242, 0.00004249, 0.00017117, 0.00017819, -0.00021030, -0.00008763, -0.00003938, -0.00002355, -0.00003699, -0.00004805, -0.00000317, -0.00001346, -0.00000160, 0.00000920, -0.00001809, -0.00001237, -0.00001308, -0.00000650, -0.00001307, -0.00000835, 0.00000341, -0.00001432, 0.00001799, -0.00002263, 0.00003160, 0.00000679, -0.00001286, },
		{ -0.00000947, 0.00000180, 0.00003088, -0.00001949, 0.00000993, -0.00000623, 0.00000334, -0.00000826, -0.00001363, 0.00000401, -0.00001749, -0.00001539, -0.00000833, -0.00000727, 0.00001283, -0.00001933, -0.00000336, -0.00002207, -0.00005088, -0.00002103, -0.00002253, -0.00008148, -0.00015188, 0.00007619, 0.00015594, 0.00007056, 0.00002923, 0.00004171, 0.00004629, 0.00005926, 0.00015609, -0.00004182, -0.00031172, -0.00001954, 0.00020155, 0.00005655, 0.00005580, 0.00006476, 0.00004490, 0.00011416, 0.00021190, -0.00011457, -0.00014414, -0.00003112, -0.00003282, -0.00002281, -0.00005925, -0.00000184, -0.00001546, -0.00000393, 0.00000900, -0.00001146, -0.00001473, -0.00001253, -0.00001133, -0.00000302, -0.00001831, 0.00000862, -0.00001545, 0.00001438, -0.00001412, 0.00001114, 0.00002442, -0.00001510, },
		},
		{ { -0.00000644, -0.00000171, -0.00005779, 0.00000283, -0.00000596, -0.00001079, -0.00002587, -0.00001124, 0.00000819, -0.00001930, -0.00002183, -0.00001840, -0.00003485, -0.00004772, -0.00019261, 0.00001991, 0.00018877, 0.00007951, 0.00009478, 0.00005435, 0.00006204, 0.00006772, 0.00025224, 0.00000513, -0.00017726, -0.00003893, -0.00003873, -0.00001664, -0.00001361, -0.00001610, 0.00000508, -0.00001810, -0.00001772, -0.00001810, 0.00000508, -0.00001610, -0.00001361, -0.00001664, -0.00003873, -0.00003893, -0.00017726, 0.00000513, 0.00025224, 0.00006772, 0.00006204, 0.00005435, 0.00009478, 0.00007951, 0.00018877, 0.00001991, -0.00019261, -0.00004772, -0.00003485, -0.00001840, -0.00002183, -0.00001930, 0.00000819, -0.00001124, -0.00002587, -0.00001079, -0.00000596, 0.00000283, -0.00005779, -0.00000171, },
		{ -0.00000132, 0.00000052, -0.00004470, -0.00002558, 0.00001072, -0.00001985, -0.00000417, -0.00003704, 0.00001734, -0.00001526, -0.00002054, -0.00001964, -0.00002721, -0.00003239, -0.00012471, -0.00016549, 0.00018224, 0.00013290, 0.00010161, 0.00006833, 0.00007394, 0.00004596, 0.00021794, 0.00014818, -0.00020570, -0.00007184, -0.00003881, -0.00002240, -0.00000915, -0.00002112, -0.00000321, -0.00000575, -0.00001941, -0.00002039, -0.00000450, 0.00000041, -0.00002104, -0.00000867, -0.00004208, -0.00002236, -0.00013087, -0.00008378, 0.00019961, 0.00013508, 0.00003407, 0.00006178, 0.00006759, 0.00007915, 0.00011245, 0.00015387, -0.00012661, -0.00012043, -0.00002623, -0.00003033, -0.00001667, -0.00002336, -0.00000137, 0.00000158, -0.00002426, -0.00001809, -0.00001220, 0.00001162, -0.00004638, -0.00001991, },
		{ 0.00000200, -0.00000325, -0.00002276, -0.00005232, 0.00001931, -0.00002673, 0.00001121, -0.00004052, 0.00000118, 0.00000081, -0.00002105, -0.00002290, -0.00001627, -0.00003916, -0.00005038, -0.00021384, 0.00000043, 0.00023025, 0.00009807, 0.00010523, 0.00006319, 0.00007179, 0.00013405, 0.00027233, -0.00016261, -0.00013605, -0.00003529, -0.00003469, -0.00000384, -0.00001953, -0.00001516, -0.00000069, -0.00001427, -0.00002028, -0.00001632, 0.00000826, -0.00001730, -0.00000746, -0.00003887, -0.00002226, -0.00008544, -0.00011925, 0.00011233, 0.00017657, 0.00004484, 0.00004605, 0.00006252, 0.00006996, 0.00007062, 0.00015361, 0.00002491, -0.00015824, -0.00005348, -0.00002859, -0.00002134, -0.00002210, -0.00000852, 0.00000174, -0.00001245, -0.00002388, -0.00001890, 0.00000983, -0.00002588, -0.00003714, },
		{ -0.00000525, -0.00000179, -0.00001030, -0.00005695, 0.00000346, -0.00001298, -0.00000307, -0.00000750, -0.00003791, 0.00002316, -0.00002785, -0.00001179, -0.00002827, -0.00001787, -0.00005033, -0.00011722, -0.00020870, 0.00020037, 0.00017396, 0.00010858, 0.00008059, 0.00008610, 0.00008484, 0.00030060, -0.00002382, -0.00022151, -0.00002983, -0.00005849, 0.00000610, -0.00002170, -0.00001395, -0.00001438, 0.00000081, -0.00002330, -0.00001841, -0.00000083, -0.00000276, -0.00001175, -0.00002783, -0.00003321, -0.00004833, -0.00012406, 0.00003906, 0.00016518, 0.00008647, 0.00002553, 0.00006036, 0.00006274, 0.00005333, 0.00010648, 0.00011388, -0.00008249, -0.00011690, -0.00002435, -0.00002863, -0.00002087, -0.00001395, 0.00000085, -0.00000901, -0.00001596, -0.00002970, 0.00000468, -0.00001040, -0.00004085, },
		{ -0.00001140, -0.00000333, -0.00000187, -0.00004946, -0.00001983, 0.00000335, -0.00002356, 0.00002250, -0.00005271, 0.00001539, -0.00001683, -0.00000957, -0.00003312, -0.00000545, -0.00005528, -0.00004683, -0.00022974, -0.00003411, 0.00028295, 0.00011996, 0.00010551, 0.00009356, 0.00007583, 0.00025710, 0.00014281, -0.00026220, -0.00005746, -0.00006884, -0.00000169, -0.00001409, -0.00001247, -0.00002539, 0.00000262, -0.00001461, -0.00002226, -0.00000869, 0.00000158, -0.00000661, -0.00002093, -0.00003888, -0.00003100, -0.00010667, -0.00001619, 0.00013276, 0.00010976, 0.00003700, 0.00003657, 0.00007187, 0.00003886, 0.00007152, 0.00011753, 0.00002701, -0.00011953, -0.00006747, -0.00001413, -0.00003382, -0.00001135, -0.00000396, -0.00000605, -0.00001223, -0.00002605, -0.00001268, 0.00000475, -0.00004040, },
		{ -0.00001643, -0.00000633, 0.00000321, -0.00003796, -0.00003703, 0.00000782, -0.00002946, 0.00002228, -0.00002658, -0.00002097, 0.00000723, -0.00002182, -0.00001473, -0.00002552, -0.00002120, -0.00006377, -0.00010580, -0.00025427, 0.00020605, 0.00024111, 0.00008737, 0.00012910, 0.00007165, 0.00020513, 0.00026136, -0.00021426, -0.00012979, -0.00005452, -0.00003737, 0.00000783, -0.00002050, -0.00001846, -0.00001723, 0.00000490, -0.00002875, -0.00000916, -0.00000608, 0.00000253, -0.00001415, -0.00003914, -0.00002752, -0.00008211, -0.00005216, 0.00009608, 0.00011200, 0.00005654, 0.00002596, 0.00005981, 0.00005429, 0.00003028, 0.00010607, 0.00006939, -0.00003242, -0.00012440, -0.00001216, -0.00003664, -0.00001715, -0.00000561, -0.00000174, -0.00001704, -0.00001204, -0.00002994, 0.00001089, -0.00003458, },
		{ -0.00003135, 0.00000138, -0.00000479, -0.00001639, -0.00005489, 0.00001045, -0.00003096, 0.00000974, 0.00000309, -0.00003863, 0.00000603, -0.00001904, -0.00000634, -0.00003389, -0.00000410, -0.00007032, -0.00004171, -0.00023784, -0.00009060, 0.00036037, 0.00012177, 0.00013072, 0.00010780, 0.00013839, 0.00033957, -0.00011386, -0.00020184, -0.00004625, -0.00007178, 0.00001563, -0.00001929, -0.00001126, -0.00003266, 0.00000815, -0.00001973, -0.00001658, -0.00000648, -0.00000337, 0.00000327, -0.00004142, -0.00002464, -0.00006617, -0.00006417, 0.00005182, 0.00011515, 0.00005499, 0.00004419, 0.00002698, 0.00008015, 0.00000591, 0.00008074, 0.00007536, 0.00004045, -0.00009738, -0.00006901, -0.00000836, -0.00004192, 0.00000177, -0.00000438, -0.00001370, -0.00001330, -0.00002274, -0.00000637, -0.00001425, },
		{ -0.00004184, 0.00000666, -0.00001214, 0.00000177, -0.00006348, 0.00000516, -0.00002501, -0.00000781, 0.00001936, -0.00002938, -0.00001268, -0.00000964, -0.00001024, -0.00002037, -0.00001842, -0.00003672, -0.00006686, -0.00009436, -0.00030571, 0.00021029, 0.00031485, 0.00006315, 0.00019295, 0.00006462, 0.00038082, 0.00001612, -0.00025245, -0.00005553, -0.00008900, -0.00000302, -0.00000505, -0.00001166, -0.00003115, -0.00000949, -0.00000049, -0.00002631, -0.00000124, -0.00001632, 0.00001604, -0.00003333, -0.00002816, -0.00005213, -0.00006730, 0.00001447, 0.00010831, 0.00005739, 0.00004623, 0.00002753, 0.00005994, 0.00003975, 0.00001917, 0.00009475, 0.00004283, -0.00000004, -0.00012446, -0.00000131, -0.00004777, -0.00000613, -0.00000298, -0.00000535, -0.00002557, -0.00000511, -0.00002533, 0.00000125, },
		{ -0.00002985, -0.00000947, 0.00000059, -0.00000282, -0.00004492, -0.00002479, 0.00000332, -0.00004051, 0.00003718, -0.00002592, -0.00000788, -0.00002609, 0.00000137, -0.00002145, -0.00001341, -0.00002992, -0.00006368, -0.00004968, -0.00023179, -0.00017032, 0.00046121, 0.00010027, 0.00019836, 0.00008576, 0.00031806, 0.00021291, -0.00030695, -0.00006226, -0.00010283, -0.00002657, -0.00000324, -0.00000237, -0.00002987, -0.00002163, 0.00000107, -0.00001815, -0.00000815, -0.00001364, 0.00000371, -0.00000510, -0.00004232, -0.00003238, -0.00007512, -0.00000260, 0.00007705, 0.00008457, 0.00001603, 0.00005975, 0.00001620, 0.00008219, -0.00001673, 0.00008179, 0.00004456, 0.00005160, -0.00008640, -0.00005596, -0.00001785, -0.00003539, 0.00000611, -0.00000311, -0.00002602, -0.00000829, -0.00001679, -0.00000937, },
		{ -0.00001617, -0.00002850, 0.00001477, -0.00001191, -0.00001856, -0.00005775, 0.00002927, -0.00006210, 0.00003882, -0.00001936, 0.00000327, -0.00003849, 0.00000192, -0.00002147, -0.00000053, -0.00003871, -0.00003421, -0.00008100, -0.00006799, -0.00037818, 0.00023353, 0.00036109, 0.00008954, 0.00020883, 0.00017444, 0.00042936, -0.00031274, -0.00010779, -0.00008852, -0.00006052, -0.00001001, 0.00000640, -0.00002459, -0.00002624, -0.00001079, -0.00000425, -0.00001682, -0.00000365, -0.00001714, 0.00001837, -0.00004481, -0.00002255, -0.00007550, -0.00001446, 0.00004212, 0.00010593, 0.00000277, 0.00005866, 0.00002028, 0.00006196, 0.00002304, 0.00001580, 0.00007371, 0.00003938, 0.00000100, -0.00009782, -0.00001225, -0.00004219, -0.00000387, 0.00000345, -0.00001796, -0.00002502, 0.00000173, -0.00002295, },
		{ -0.00002208, -0.00002860, 0.00000648, -0.00000103, -0.00001237, -0.00006081, 0.00001661, -0.00003791, -0.00000334, 0.00001938, -0.00002124, -0.00000195, -0.00004275, 0.00001212, -0.00002518, -0.00000191, -0.00005829, -0.00005045, -0.00005815, -0.00023013, -0.00024438, 0.00054097, 0.00011427, 0.00024627, 0.00012102, 0.00050642, -0.00014447, -0.00026592, -0.00001059, -0.00012471, 0.00000200, -0.00001647, 0.00000664, -0.00004410, -0.00000745, -0.00001363, -0.00000529, -0.00000794, -0.00001862, 0.00001138, -0.00002100, -0.00003093, -0.00005992, -0.00003477, 0.00002477, 0.00009331, 0.00003444, 0.00001135, 0.00006178, 0.00001517, 0.00006939, -0.00001749, 0.00005835, 0.00004781, 0.00003142, -0.00005600, -0.00005524, -0.00002090, -0.00002913, 0.00001028, -0.00000887, -0.00003331, 0.00000044, -0.00001268, },
		{ -0.00002509, -0.00002802, -0.00000706, 0.00001168, -0.00001112, -0.00004758, -0.00001703, 0.00000540, -0.00005393, 0.00005625, -0.00004922, 0.00003857, -0.00007535, 0.00002780, -0.00004502, 0.00003607, -0.00008573, -0.00000635, -0.00011136, -0.00003054, -0.00045845, 0.00025381, 0.00040212, 0.00015535, 0.00019471, 0.00042271, 0.00017090, -0.00045795, 0.00004514, -0.00015828, 0.00000051, -0.00004510, 0.00003457, -0.00005419, -0.00000185, -0.00002964, 0.00000596, -0.00001163, -0.00001148, -0.00000781, 0.00000316, -0.00003457, -0.00004550, -0.00005183, 0.00001618, 0.00006191, 0.00007685, -0.00002404, 0.00006862, 0.00001728, 0.00005161, 0.00002340, -0.00000373, 0.00007151, 0.00002788, -0.00000194, -0.00006337, -0.00003413, -0.00002403, -0.00000980, 0.00001067, -0.00003619, -0.00001149, 0.00000187, },
		{ -0.00000904, -0.00003866, -0.00001019, 0.00000705, 0.00000392, -0.00004152, -0.00003712, 0.00002181, -0.00006146, 0.00004162, -0.00002702, 0.00001820, -0.00003291, -0.00002447, -0.00000816, 0.00000605, -0.00003098, -0.00005338, -0.00006016, -0.00004602, -0.00025067, -0.00029248, 0.00056518, 0.00022061, 0.00020613, 0.00034983, 0.00043520, -0.00048176, -0.00006600, -0.00007946, -0.00006649, -0.00002668, 0.00000770, -0.00001812, -0.00002756, -0.00001540, -0.00001555, 0.00000781, -0.00001719, -0.00001153, 0.00000253, -0.00001934, -0.00004062, -0.00005860, 0.00000652, 0.00003311, 0.00009567, -0.00001494, 0.00002984, 0.00005081, 0.00001768, 0.00005403, -0.00001196, 0.00002776, 0.00006985, -0.00001364, -0.00000175, -0.00007817, -0.00000328, -0.00003628, 0.00002260, -0.00003010, -0.00002116, 0.00000037, },
		{ -0.00000102, -0.00003277, -0.00002405, 0.00000533, 0.00001069, -0.00002602, -0.00005270, 0.00001921, -0.00004105, 0.00000075, 0.00001648, -0.00002795, 0.00002968, -0.00007553, 0.00002281, -0.00003518, 0.00003851, -0.00010982, 0.00000663, -0.00012316, -0.00002045, -0.00049720, 0.00020134, 0.00052944, 0.00015563, 0.00031798, 0.00058022, -0.00029452, -0.00030786, 0.00003837, -0.00014057, 0.00000547, -0.00004449, 0.00002992, -0.00005720, 0.00000770, -0.00004388, 0.00002232, -0.00001615, -0.00001266, -0.00000471, -0.00000881, -0.00002694, -0.00006686, 0.00000095, 0.00000987, 0.00009086, 0.00001671, -0.00000558, 0.00005361, 0.00003023, 0.00002304, 0.00004589, -0.00004616, 0.00010087, -0.00001523, 0.00002318, -0.00005427, -0.00003728, -0.00001701, -0.00000639, 0.00000054, -0.00003831, -0.00000035, },
		{ -0.00001030, -0.00000637, -0.00004833, 0.00001168, -0.00000016, 0.00000498, -0.00007052, 0.00001460, -0.00002448, -0.00002298, 0.00003150, -0.00003829, 0.00003645, -0.00005271, -0.00001178, -0.00002107, 0.00003313, -0.00006860, -0.00003618, -0.00007365, -0.00004025, -0.00025430, -0.00036930, 0.00059729, 0.00035527, 0.00017425, 0.00069857, -0.00002522, -0.00049699, 0.00003839, -0.00012402, -0.00001607, -0.00005495, 0.00002643, -0.00004166, -0.00000349, -0.00003565, -0.00000229, 0.00001138, -0.00002663, 0.00000044, -0.00001717, -0.00000312, -0.00007147, -0.00000866, 0.00000114, 0.00006445, 0.00005264, -0.00001995, 0.00003151, 0.00005205, 0.00000216, 0.00005996, -0.00002808, 0.00002507, 0.00006738, -0.00003876, 0.00003837, -0.00009603, 0.00001462, -0.00004171, 0.00002637, -0.00004913, 0.00000233, },
		{ -0.00000584, -0.00000024, -0.00004273, -0.00000638, 0.00000361, 0.00001444, -0.00005723, -0.00001510, 0.00000414, -0.00004546, 0.00003747, -0.00003327, 0.00001683, -0.00000946, -0.00004350, -0.00001203, 0.00001014, -0.00000614, -0.00008817, -0.00001611, -0.00010393, -0.00004320, -0.00049233, 0.00006493, 0.00077085, 0.00005970, 0.00069597, 0.00032226, -0.00058456, -0.00007302, -0.00006675, -0.00004987, -0.00004577, -0.00000548, -0.00001103, -0.00002326, -0.00001471, -0.00003346, 0.00002834, -0.00002400, -0.00000487, -0.00001678, -0.00000127, -0.00004681, -0.00004066, 0.00001837, 0.00001688, 0.00008941, -0.00002801, 0.00002426, 0.00003223, 0.00003959, 0.00000088, 0.00006134, -0.00007729, 0.00012529, -0.00006003, 0.00005905, -0.00006381, -0.00002222, -0.00002218, 0.00000411, -0.00002561, -0.00001542, },
		{ 0.00001230, -0.00001902, -0.00000881, -0.00004268, 0.00002214, -0.00000320, -0.00001231, -0.00006503, 0.00004258, -0.00007284, 0.00004902, -0.00003586, 0.00001082, -0.00000319, -0.00001894, -0.00004843, 0.00002097, -0.00000186, -0.00005181, -0.00006338, -0.00004096, -0.00009123, -0.00018204, -0.00055573, 0.00074407, 0.00038749, 0.00038735, 0.00079336, -0.00059876, -0.00020299, -0.00006618, -0.00003758, -0.00005833, -0.00001731, -0.00001368, -0.00001368, -0.00001761, -0.00003166, 0.00000408, 0.00001058, -0.00003012, 0.00000353, -0.00002609, -0.00000280, -0.00007683, 0.00003781, -0.00002186, 0.00010200, -0.00001815, 0.00001941, 0.00000609, 0.00006819, -0.00002716, 0.00007983, -0.00005716, 0.00004506, 0.00003424, -0.00002514, 0.00003941, -0.00008377, 0.00001132, -0.00002910, 0.00000345, -0.00003973, },
		{ 0.00000932, -0.00001952, 0.00000079, -0.00004482, 0.00001045, -0.00000194, 0.00000885, -0.00007619, 0.00003592, -0.00005476, 0.00001784, -0.00000033, -0.00002258, 0.00001472, -0.00001173, -0.00004251, -0.00001231, 0.00002105, -0.00003061, -0.00007749, -0.00002236, -0.00011186, -0.00002640, -0.00053930, -0.00003952, 0.00098868, 0.00008314, 0.00109313, -0.00035033, -0.00043076, -0.00004649, -0.00007021, -0.00002361, -0.00005391, -0.00000602, -0.00002126, -0.00000578, -0.00003285, -0.00001572, 0.00002534, -0.00002950, 0.00000315, -0.00003157, 0.00000923, -0.00007091, 0.00001832, -0.00001429, 0.00006368, 0.00002941, -0.00001366, 0.00002192, 0.00003347, 0.00002343, 0.00000639, 0.00004678, -0.00006778, 0.00010438, -0.00005894, 0.00005994, -0.00005315, -0.00002004, -0.00001413, -0.00000582, -0.00003631, },
		{ -0.00000767, 0.00000165, -0.00001953, -0.00001360, -0.00002645, 0.00002068, -0.00000499, -0.00003880, -0.00002013, 0.00001052, -0.00005691, 0.00007230, -0.00008407, 0.00005986, -0.00005315, 0.00002976, -0.00009849, 0.00008313, -0.00007418, 0.00000763, -0.00012527, 0.00002038, -0.00016988, -0.00007550, -0.00080772, 0.00096048, 0.00038403, 0.00087402, 0.00031225, -0.00079889, 0.00004884, -0.00020263, 0.00008120, -0.00013053, 0.00003980, -0.00007512, 0.00004203, -0.00006154, 0.00000120, -0.00000726, 0.00001263, -0.00002969, -0.00000324, -0.00002031, -0.00002582, -0.00003033, 0.00002445, -0.00000019, 0.00008554, -0.00004996, 0.00004893, -0.00001094, 0.00006796, -0.00003657, 0.00007654, -0.00005587, 0.00003744, 0.00002006, -0.00001289, 0.00003012, -0.00006154, 0.00000312, -0.00001721, -0.00002764, },
		{ -0.00001164, 0.00000990, -0.00002390, -0.00000456, -0.00003240, 0.00001477, -0.00000277, -0.00001516, -0.00005280, 0.00004322, -0.00008791, 0.00008759, -0.00007842, 0.00004172, -0.00004395, 0.00004201, -0.00009578, 0.00005066, -0.00004488, 0.00001894, -0.00013027, 0.00003373, -0.00018927, 0.00007563, -0.00070204, -0.00002480, 0.00106438, 0.00050409, 0.00098986, -0.00094751, -0.00003033, -0.00022815, 0.00006936, -0.00010546, 0.00001809, -0.00007826, 0.00003501, -0.00004416, -0.00000851, -0.00002004, 0.00002709, -0.00003382, 0.00000439, -0.00003391, -0.00000254, -0.00005181, 0.00004032, -0.00003307, 0.00009704, -0.00004068, 0.00003269, -0.00000199, 0.00004421, 0.00000080, 0.00001662, 0.00002955, -0.00004672, 0.00005971, -0.00001981, 0.00002255, -0.00001397, -0.00003870, 0.00001143, -0.00004327, },
		{ -0.00002851, 0.00001226, -0.00001029, -0.00002041, -0.00001011, -0.00001414, 0.00001819, -0.00002013, -0.00004091, 0.00001887, -0.00004753, 0.00001919, 0.00001285, -0.00005298, 0.00003927, -0.00004084, 0.00001987, -0.00008324, 0.00007520, -0.00008862, 0.00002069, -0.00013748, 0.00002142, -0.00017166, -0.00006901, -0.00095686, 0.00102494, 0.00071277, 0.00117285, -0.00052683, -0.00047525, -0.00001017, -0.00015643, 0.00007179, -0.00010988, 0.00001635, -0.00006817, 0.00004718, -0.00007313, 0.00002192, -0.00001732, 0.00001233, -0.00002803, -0.00000892, -0.00002096, -0.00003401, 0.00002210, -0.00002292, 0.00005936, 0.00000994, -0.00001608, 0.00003550, 0.00000179, 0.00003863, -0.00001660, 0.00004255, -0.00001389, -0.00001943, 0.00005938, -0.00004223, 0.00004844, -0.00006131, 0.00001674, -0.00003600, },
		{ -0.00007225, 0.00002453, -0.00000794, -0.00001800, -0.00001217, -0.00001234, 0.00000931, -0.00000577, -0.00003944, 0.00000123, -0.00001078, -0.00004002, 0.00007682, -0.00010011, 0.00007000, -0.00007772, 0.00007568, -0.00012645, 0.00009571, -0.00011111, 0.00007761, -0.00018523, 0.00008259, -0.00024914, 0.00015094, -0.00084647, -0.00005896, 0.00123247, 0.00117703, 0.00014888, -0.00093399, 0.00014598, -0.00032425, 0.00015924, -0.00015730, 0.00006165, -0.00013032, 0.00008530, -0.00008834, 0.00003090, -0.00003646, 0.00002948, -0.00003384, -0.00000149, -0.00002759, -0.00002693, 0.00000966, -0.00000495, 0.00001329, 0.00005568, -0.00004762, 0.00004627, -0.00000227, 0.00002756, 0.00000773, -0.00000520, 0.00005209, -0.00006770, 0.00005199, 0.00000014, -0.00000637, 0.00001716, -0.00004702, 0.00002071, },
		{ -0.00007379, 0.00000485, 0.00000214, -0.00000718, -0.00002865, 0.00000620, -0.00001395, 0.00001898, -0.00004989, 0.00000150, -0.00000148, -0.00005094, 0.00006575, -0.00005532, 0.00000864, -0.00001988, 0.00001312, -0.00003219, -0.00001512, -0.00000401, -0.00002866, -0.00003231, -0.00008676, -0.00004251, -0.00009755, -0.00016886, -0.00098547, 0.00088590, 0.00164634, 0.00064004, -0.00100475, -0.00005331, -0.00021415, 0.00000094, -0.00002799, -0.00002190, -0.00006866, 0.00000110, -0.00000640, -0.00003373, 0.00000992, -0.00001446, 0.00000766, -0.00002836, -0.00000582, -0.00004183, 0.00000756, 0.00001224, -0.00002305, 0.00007650, -0.00004771, 0.00002804, 0.00001339, 0.00001300, 0.00001482, -0.00000503, 0.00002213, 0.00001708, -0.00007528, 0.00011436, -0.00009033, 0.00008096, -0.00007185, 0.00003802, },
		{ -0.00000645, -0.00005962, 0.00003531, -0.00001908, -0.00001378, -0.00001236, 0.00000345, 0.00000187, -0.00002084, -0.00003262, 0.00003203, -0.00006885, 0.00005086, -0.00001397, -0.00003163, 0.00001053, -0.00002564, 0.00002290, -0.00005945, 0.00003043, -0.00007132, 0.00003801, -0.00013712, 0.00001461, -0.00016560, 0.00003821, -0.00076135, -0.00044879, 0.00203594, 0.00136278, -0.00091676, -0.00036312, -0.00007344, -0.00015960, 0.00005512, -0.00006253, -0.00002164, -0.00006957, 0.00004196, -0.00006055, 0.00002519, -0.00002804, 0.00001310, -0.00002071, -0.00001305, -0.00002723, -0.00002506, 0.00005025, -0.00005739, 0.00008046, -0.00003228, 0.00000683, 0.00001518, 0.00002622, -0.00001356, 0.00003674, -0.00004619, 0.00009946, -0.00012475, 0.00009164, -0.00002175, -0.00000192, 0.00001954, -0.00002890, },
		{ 0.00002783, -0.00008519, 0.00003490, -0.00001989, 0.00000602, -0.00003963, 0.00002939, -0.00002416, 0.00001739, -0.00007269, 0.00006323, -0.00007738, 0.00003494, 0.00000487, -0.00002818, -0.00000854, -0.00000849, -0.00000114, -0.00000986, -0.00002342, -0.00001990, -0.00002379, -0.00002272, -0.00011329, -0.00000567, -0.00015467, -0.00011703, -0.00123360, 0.00101201, 0.00278416, -0.00079253, -0.00060613, -0.00008253, -0.00016986, -0.00001283, -0.00000319, -0.00004123, -0.00006366, 0.00000365, -0.00001567, -0.00001706, 0.00001138, -0.00002891, 0.00002056, -0.00004341, 0.00000696, -0.00006680, 0.00007577, -0.00005843, 0.00004928, 0.00000608, -0.00002272, 0.00001907, 0.00002858, -0.00001266, 0.00002801, -0.00003041, 0.00004666, -0.00000669, -0.00007324, 0.00012542, -0.00010581, 0.00008645, -0.00005751, },
		{ -0.00001746, -0.00003727, -0.00002259, 0.00001804, -0.00001398, -0.00001604, -0.00000280, 0.00000968, -0.00000678, -0.00004113, 0.00001156, -0.00000121, -0.00005188, 0.00007793, -0.00007650, 0.00003519, -0.00005891, 0.00004226, -0.00004733, 0.00003497, -0.00008366, 0.00002782, -0.00005810, -0.00003513, -0.00010596, -0.00002107, -0.00016974, -0.00054688, -0.00112777, 0.00369449, 0.00032691, -0.00126580, 0.00009042, -0.00032130, 0.00004417, -0.00008968, 0.00006072, -0.00013440, 0.00002976, -0.00004579, 0.00001349, -0.00001578, -0.00000255, -0.00001157, -0.00000406, -0.00002337, -0.00003704, 0.00001965, 0.00001858, -0.00003822, 0.00007765, -0.00007212, 0.00004554, -0.00000664, 0.00003753, -0.00003347, 0.00004158, -0.00005404, 0.00010819, -0.00014937, 0.00011851, -0.00004089, 0.00000244, 0.00002325, },
		{ -0.00002377, -0.00001220, -0.00005574, 0.00003536, -0.00003222, 0.00001964, -0.00004874, 0.00005450, -0.00004521, 0.00001521, -0.00006967, 0.00009475, -0.00014052, 0.00014437, -0.00012630, 0.00009306, -0.00012660, 0.00010909, -0.00012777, 0.00014151, -0.00018390, 0.00012530, -0.00018153, 0.00015995, -0.00033373, 0.00026598, -0.00051807, 0.00033024, -0.00209051, 0.00212952, 0.00308489, -0.00227011, 0.00034968, -0.00057626, 0.00020850, -0.00028687, 0.00022555, -0.00023853, 0.00010362, -0.00013688, 0.00009617, -0.00009135, 0.00007556, -0.00009277, 0.00007309, -0.00008643, 0.00003098, -0.00007162, 0.00010979, -0.00011290, 0.00012133, -0.00009226, 0.00005256, -0.00003033, 0.00006750, -0.00005578, 0.00005445, -0.00006140, 0.00007738, -0.00004977, -0.00003359, 0.00009741, -0.00008809, 0.00006654, },
		{ 0.00005553, -0.00004393, -0.00003005, -0.00000186, -0.00000326, 0.00000025, -0.00002352, 0.00001659, 0.00000057, -0.00001287, -0.00005165, 0.00006396, -0.00007837, 0.00005145, -0.00001759, -0.00001237, -0.00001646, -0.00001220, -0.00000850, 0.00001326, -0.00001392, -0.00005769, -0.00000888, -0.00000860, -0.00009981, -0.00002266, -0.00013270, -0.00007074, -0.00074219, -0.00135949, 0.00550898, -0.00177323, -0.00046137, -0.00015678, -0.00013181, -0.00007014, 0.00000844, -0.00002324, -0.00006842, -0.00001749, -0.00001842, 0.00000571, 0.00000194, -0.00002942, 0.00001001, -0.00002382, -0.00001049, -0.00005200, 0.00007738, -0.00005957, 0.00004868, -0.00001484, -0.00001848, 0.00001638, 0.00001985, 0.00000400, -0.00001084, 0.00000790, -0.00001704, 0.00005751, -0.00010527, 0.00008800, -0.00001689, -0.00002544, },
		{ 0.00006421, -0.00001678, -0.00003279, -0.00002263, 0.00002101, -0.00003518, 0.00003262, -0.00005557, 0.00007762, -0.00007758, 0.00002527, -0.00004936, 0.00007486, -0.00012685, 0.00017044, -0.00020001, 0.00018349, -0.00022684, 0.00021606, -0.00024508, 0.00029082, -0.00036504, 0.00030928, -0.00038250, 0.00038644, -0.00059701, 0.00058971, -0.00097489, 0.00093229, -0.00334733, 0.00437414, 0.00176211, -0.00266145, 0.00095068, -0.00096008, 0.00053574, -0.00053734, 0.00043297, -0.00042372, 0.00027528, -0.00029624, 0.00023969, -0.00019694, 0.00015790, -0.00016802, 0.00013588, -0.00013953, 0.00006271, -0.00005542, 0.00008253, -0.00009084, 0.00010970, -0.00012155, 0.00009120, -0.00005550, 0.00007724, -0.00006741, 0.00004842, -0.00004943, 0.00005376, -0.00004041, -0.00002664, 0.00008820, -0.00008422, },
		{ -0.00005856, 0.00008569, -0.00007919, 0.00000220, -0.00000252, -0.00001997, 0.00002312, -0.00004159, 0.00005225, -0.00003736, 0.00000354, -0.00005654, 0.00009493, -0.00013855, 0.00016426, -0.00017177, 0.00014503, -0.00017726, 0.00015638, -0.00019299, 0.00022962, -0.00025264, 0.00018225, -0.00027350, 0.00028860, -0.00044329, 0.00039602, -0.00067837, 0.00060008, -0.00180425, -0.00025783, 0.00616287, -0.00375325, 0.00089905, -0.00093381, 0.00046453, -0.00051322, 0.00037181, -0.00033941, 0.00021060, -0.00025880, 0.00018353, -0.00014688, 0.00013104, -0.00014845, 0.00011100, -0.00010890, 0.00004867, -0.00007460, 0.00010954, -0.00011305, 0.00012085, -0.00011824, 0.00007309, -0.00004785, 0.00006617, -0.00004274, 0.00001978, -0.00002162, 0.00000980, 0.00000732, -0.00004279, 0.00003495, 0.00002330, },
		{ -0.00010769, 0.00008770, -0.00004754, -0.00000979, -0.00001644, 0.00001156, -0.00003128, 0.00003773, -0.00004853, 0.00007900, -0.00009960, 0.00004449, -0.00003148, 0.00002854, -0.00004361, 0.00007757, -0.00013402, 0.00013181, -0.00018000, 0.00016769, -0.00018840, 0.00024247, -0.00034007, 0.00028500, -0.00036159, 0.00036434, -0.00056462, 0.00054876, -0.00099225, 0.00097034, -0.00402544, 0.00676633, -0.00079294, -0.00164233, 0.00061973, -0.00072815, 0.00040364, -0.00043016, 0.00033725, -0.00034512, 0.00022118, -0.00026615, 0.00023523, -0.00018999, 0.00014418, -0.00015965, 0.00013131, -0.00014590, 0.00007802, -0.00003401, 0.00002689, -0.00001677, 0.00001934, -0.00005913, 0.00005693, -0.00002825, 0.00004056, -0.00003798, 0.00001072, -0.00000466, -0.00001634, 0.00002787, -0.00007159, 0.00011733, },
		{ 0.00003135, -0.00005364, 0.00006721, -0.00006918, 0.00001101, 0.00000341, -0.00004127, 0.00005683, -0.00007017, 0.00009901, -0.00010203, 0.00005671, -0.00007635, 0.00010573, -0.00014408, 0.00019833, -0.00025880, 0.00025772, -0.00030562, 0.00029971, -0.00035241, 0.00042443, -0.00049181, 0.00044647, -0.00057427, 0.00062883, -0.00083388, 0.00087555, -0.00133600, 0.00147385, -0.00338845, 0.00221250, 0.00486752, -0.00435821, 0.00172213, -0.00149561, 0.00094172, -0.00088595, 0.00067905, -0.00061321, 0.00046603, -0.00049546, 0.00040252, -0.00032696, 0.00028095, -0.00028446, 0.00023196, -0.00021929, 0.00013962, -0.00011053, 0.00011219, -0.00010864, 0.00011651, -0.00014841, 0.00011925, -0.00008258, 0.00008121, -0.00005814, 0.00002091, -0.00001186, -0.00001837, 0.00001805, -0.00002834, 0.00001779, },
		{ 0.00013847, -0.00011305, 0.00008634, -0.00006212, 0.00001518, -0.00001608, -0.00000365, -0.00000695, 0.00001663, -0.00000779, 0.00003033, -0.00007035, 0.00003908, -0.00001292, -0.00000985, 0.00004288, -0.00006191, 0.00001873, -0.00002915, -0.00000564, -0.00002412, 0.00003686, -0.00001617, -0.00006789, -0.00001282, -0.00003665, -0.00000288, -0.00011937, -0.00002250, -0.00030718, -0.00003970, -0.00314822, 0.00780548, -0.00314822, -0.00003970, -0.00030718, -0.00002250, -0.00011937, -0.00000288, -0.00003665, -0.00001282, -0.00006789, -0.00001617, 0.00003686, -0.00002412, -0.00000564, -0.00002915, 0.00001873, -0.00006191, 0.00004288, -0.00000985, -0.00001292, 0.00003908, -0.00007035, 0.00003033, -0.00000779, 0.00001663, -0.00000695, -0.00000365, -0.00001608, 0.00001518, -0.00006212, 0.00008634, -0.00011305, },
		{ 0.00003135, 0.00001779, -0.00002834, 0.00001805, -0.00001837, -0.00001186, 0.00002091, -0.00005814, 0.00008121, -0.00008258, 0.00011925, -0.00014841, 0.00011651, -0.00010864, 0.00011219, -0.00011053, 0.00013962, -0.00021929, 0.00023196, -0.00028446, 0.00028095, -0.00032696, 0.00040252, -0.00049546, 0.00046603, -0.00061321, 0.00067905, -0.00088595, 0.00094172, -0.00149561, 0.00172213, -0.00435821, 0.00486752, 0.00221250, -0.00338845, 0.00147385, -0.00133600, 0.00087555, -0.00083388, 0.00062883, -0.00057427, 0.00044647, -0.00049181, 0.00042443, -0.00035241, 0.00029971, -0.00030562, 0.00025772, -0.00025880, 0.00019833, -0.00014408, 0.00010573, -0.00007635, 0.00005671, -0.00010203, 0.00009901, -0.00007017, 0.00005683, -0.00004127, 0.00000341, 0.00001101, -0.00006918, 0.00006721, -0.00005364, },
		{ -0.00010769, 0.00011733, -0.00007159, 0.00002787, -0.00001634, -0.00000466, 0.00001072, -0.00003798, 0.00004056, -0.00002825, 0.00005693, -0.00005913, 0.00001934, -0.00001677, 0.00002689, -0.00003401, 0.00007802, -0.00014590, 0.00013131, -0.00015965, 0.00014418, -0.00018999, 0.00023523, -0.00026615, 0.00022118, -0.00034512, 0.00033725, -0.00043016, 0.00040364, -0.00072815, 0.00061973, -0.00164233, -0.00079294, 0.00676633, -0.00402544, 0.00097034, -0.00099225, 0.00054876, -0.00056462, 0.00036434, -0.00036159, 0.00028500, -0.00034007, 0.00024247, -0.00018840, 0.00016769, -0.00018000, 0.00013181, -0.00013402, 0.00007757, -0.00004361, 0.00002854, -0.00003148, 0.00004449, -0.00009960, 0.00007900, -0.00004853, 0.00003773, -0.00003128, 0.00001156, -0.00001644, -0.00000979, -0.00004754, 0.00008770, },
		{ -0.00005856, 0.00002330, 0.00003495, -0.00004279, 0.00000732, 0.00000980, -0.00002162, 0.00001978, -0.00004274, 0.00006617, -0.00004785, 0.00007309, -0.00011824, 0.00012085, -0.00011305, 0.00010954, -0.00007460, 0.00004867, -0.00010890, 0.00011100, -0.00014845, 0.00013104, -0.00014688, 0.00018353, -0.00025880, 0.00021060, -0.00033941, 0.00037181, -0.00051322, 0.00046453, -0.00093381, 0.00089905, -0.00375325, 0.00616287, -0.00025783, -0.00180425, 0.00060008, -0.00067837, 0.00039602, -0.00044329, 0.00028860, -0.00027350, 0.00018225, -0.00025264, 0.00022962, -0.00019299, 0.00015638, -0.00017726, 0.00014503, -0.00017177, 0.00016426, -0.00013855, 0.00009493, -0.00005654, 0.00000354, -0.00003736, 0.00005225, -0.00004159, 0.00002312, -0.00001997, -0.00000252, 0.00000220, -0.00007919, 0.00008569, },
		{ 0.00006421, -0.00008422, 0.00008820, -0.00002664, -0.00004041, 0.00005376, -0.00004943, 0.00004842, -0.00006741, 0.00007724, -0.00005550, 0.00009120, -0.00012155, 0.00010970, -0.00009084, 0.00008253, -0.00005542, 0.00006271, -0.00013953, 0.00013588, -0.00016802, 0.00015790, -0.00019694, 0.00023969, -0.00029624, 0.00027528, -0.00042372, 0.00043297, -0.00053734, 0.00053574, -0.00096008, 0.00095068, -0.00266145, 0.00176211, 0.00437414, -0.00334733, 0.00093229, -0.00097489, 0.00058971, -0.00059701, 0.00038644, -0.00038250, 0.00030928, -0.00036504, 0.00029082, -0.00024508, 0.00021606, -0.00022684, 0.00018349, -0.00020001, 0.00017044, -0.00012685, 0.00007486, -0.00004936, 0.00002527, -0.00007758, 0.00007762, -0.00005557, 0.00003262, -0.00003518, 0.00002101, -0.00002263, -0.00003279, -0.00001678, },
		{ 0.00005553, -0.00002544, -0.00001689, 0.00008800, -0.00010527, 0.00005751, -0.00001704, 0.00000790, -0.00001084, 0.00000400, 0.00001985, 0.00001638, -0.00001848, -0.00001484, 0.00004868, -0.00005957, 0.00007738, -0.00005200, -0.00001049, -0.00002382, 0.00001001, -0.00002942, 0.00000194, 0.00000571, -0.00001842, -0.00001749, -0.00006842, -0.00002324, 0.00000844, -0.00007014, -0.00013181, -0.00015678, -0.00046137, -0.00177323, 0.00550898, -0.00135949, -0.00074219, -0.00007074, -0.00013270, -0.00002266, -0.00009981, -0.00000860, -0.00000888, -0.00005769, -0.00001392, 0.00001326, -0.00000850, -0.00001220, -0.00001646, -0.00001237, -0.00001759, 0.00005145, -0.00007837, 0.00006396, -0.00005165, -0.00001287, 0.00000057, 0.00001659, -0.00002352, 0.00000025, -0.00000326, -0.00000186, -0.00003005, -0.00004393, },
		{ -0.00002377, 0.00006654, -0.00008809, 0.00009741, -0.00003359, -0.00004977, 0.00007738, -0.00006140, 0.00005445, -0.00005578, 0.00006750, -0.00003033, 0.00005256, -0.00009226, 0.00012133, -0.00011290, 0.00010979, -0.00007162, 0.00003098, -0.00008643, 0.00007309, -0.00009277, 0.00007556, -0.00009135, 0.00009617, -0.00013688, 0.00010362, -0.00023853, 0.00022555, -0.00028687, 0.00020850, -0.00057626, 0.00034968, -0.00227011, 0.00308489, 0.00212952, -0.00209051, 0.00033024, -0.00051807, 0.00026598, -0.00033373, 0.00015995, -0.00018153, 0.00012530, -0.00018390, 0.00014151, -0.00012777, 0.00010909, -0.00012660, 0.00009306, -0.00012630, 0.00014437, -0.00014052, 0.00009475, -0.00006967, 0.00001521, -0.00004521, 0.00005450, -0.00004874, 0.00001964, -0.00003222, 0.00003536, -0.00005574, -0.00001220, },
		{ -0.00001746, 0.00002325, 0.00000244, -0.00004089, 0.00011851, -0.00014937, 0.00010819, -0.00005404, 0.00004158, -0.00003347, 0.00003753, -0.00000664, 0.00004554, -0.00007212, 0.00007765, -0.00003822, 0.00001858, 0.00001965, -0.00003704, -0.00002337, -0.00000406, -0.00001157, -0.00000255, -0.00001578, 0.00001349, -0.00004579, 0.00002976, -0.00013440, 0.00006072, -0.00008968, 0.00004417, -0.00032130, 0.00009042, -0.00126580, 0.00032691, 0.00369449, -0.00112777, -0.00054688, -0.00016974, -0.00002107, -0.00010596, -0.00003513, -0.00005810, 0.00002782, -0.00008366, 0.00003497, -0.00004733, 0.00004226, -0.00005891, 0.00003519, -0.00007650, 0.00007793, -0.00005188, -0.00000121, 0.00001156, -0.00004113, -0.00000678, 0.00000968, -0.00000280, -0.00001604, -0.00001398, 0.00001804, -0.00002259, -0.00003727, },
		{ 0.00002783, -0.00005751, 0.00008645, -0.00010581, 0.00012542, -0.00007324, -0.00000669, 0.00004666, -0.00003041, 0.00002801, -0.00001266, 0.00002858, 0.00001907, -0.00002272, 0.00000608, 0.00004928, -0.00005843, 0.00007577, -0.00006680, 0.00000696, -0.00004341, 0.00002056, -0.00002891, 0.00001138, -0.00001706, -0.00001567, 0.00000365, -0.00006366, -0.00004123, -0.00000319, -0.00001283, -0.00016986, -0.00008253, -0.00060613, -0.00079253, 0.00278416, 0.00101201, -0.00123360, -0.00011703, -0.00015467, -0.00000567, -0.00011329, -0.00002272, -0.00002379, -0.00001990, -0.00002342, -0.00000986, -0.00000114, -0.00000849, -0.00000854, -0.00002818, 0.00000487, 0.00003494, -0.00007738, 0.00006323, -0.00007269, 0.00001739, -0.00002416, 0.00002939, -0.00003963, 0.00000602, -0.00001989, 0.00003490, -0.00008519, },
		{ -0.00000645, -0.00002890, 0.00001954, -0.00000192, -0.00002175, 0.00009164, -0.00012475, 0.00009946, -0.00004619, 0.00003674, -0.00001356, 0.00002622, 0.00001518, 0.00000683, -0.00003228, 0.00008046, -0.00005739, 0.00005025, -0.00002506, -0.00002723, -0.00001305, -0.00002071, 0.00001310, -0.00002804, 0.00002519, -0.00006055, 0.00004196, -0.00006957, -0.00002164, -0.00006253, 0.00005512, -0.00015960, -0.00007344, -0.00036312, -0.00091676, 0.00136278, 0.00203594, -0.00044879, -0.00076135, 0.00003821, -0.00016560, 0.00001461, -0.00013712, 0.00003801, -0.00007132, 0.00003043, -0.00005945, 0.00002290, -0.00002564, 0.00001053, -0.00003163, -0.00001397, 0.00005086, -0.00006885, 0.00003203, -0.00003262, -0.00002084, 0.00000187, 0.00000345, -0.00001236, -0.00001378, -0.00001908, 0.00003531, -0.00005962, },
		{ -0.00007379, 0.00003802, -0.00007185, 0.00008096, -0.00009033, 0.00011436, -0.00007528, 0.00001708, 0.00002213, -0.00000503, 0.00001482, 0.00001300, 0.00001339, 0.00002804, -0.00004771, 0.00007650, -0.00002305, 0.00001224, 0.00000756, -0.00004183, -0.00000582, -0.00002836, 0.00000766, -0.00001446, 0.00000992, -0.00003373, -0.00000640, 0.00000110, -0.00006866, -0.00002190, -0.00002799, 0.00000094, -0.00021415, -0.00005331, -0.00100475, 0.00064004, 0.00164634, 0.00088590, -0.00098547, -0.00016886, -0.00009755, -0.00004251, -0.00008676, -0.00003231, -0.00002866, -0.00000401, -0.00001512, -0.00003219, 0.00001312, -0.00001988, 0.00000864, -0.00005532, 0.00006575, -0.00005094, -0.00000148, 0.00000150, -0.00004989, 0.00001898, -0.00001395, 0.00000620, -0.00002865, -0.00000718, 0.00000214, 0.00000485, },
		{ -0.00007225, 0.00002071, -0.00004702, 0.00001716, -0.00000637, 0.00000014, 0.00005199, -0.00006770, 0.00005209, -0.00000520, 0.00000773, 0.00002756, -0.00000227, 0.00004627, -0.00004762, 0.00005568, 0.00001329, -0.00000495, 0.00000966, -0.00002693, -0.00002759, -0.00000149, -0.00003384, 0.00002948, -0.00003646, 0.00003090, -0.00008834, 0.00008530, -0.00013032, 0.00006165, -0.00015730, 0.00015924, -0.00032425, 0.00014598, -0.00093399, 0.00014888, 0.00117703, 0.00123247, -0.00005896, -0.00084647, 0.00015094, -0.00024914, 0.00008259, -0.00018523, 0.00007761, -0.00011111, 0.00009571, -0.00012645, 0.00007568, -0.00007772, 0.00007000, -0.00010011, 0.00007682, -0.00004002, -0.00001078, 0.00000123, -0.00003944, -0.00000577, 0.00000931, -0.00001234, -0.00001217, -0.00001800, -0.00000794, 0.00002453, },
		{ -0.00002851, -0.00003600, 0.00001674, -0.00006131, 0.00004844, -0.00004223, 0.00005938, -0.00001943, -0.00001389, 0.00004255, -0.00001660, 0.00003863, 0.00000179, 0.00003550, -0.00001608, 0.00000994, 0.00005936, -0.00002292, 0.00002210, -0.00003401, -0.00002096, -0.00000892, -0.00002803, 0.00001233, -0.00001732, 0.00002192, -0.00007313, 0.00004718, -0.00006817, 0.00001635, -0.00010988, 0.00007179, -0.00015643, -0.00001017, -0.00047525, -0.00052683, 0.00117285, 0.00071277, 0.00102494, -0.00095686, -0.00006901, -0.00017166, 0.00002142, -0.00013748, 0.00002069, -0.00008862, 0.00007520, -0.00008324, 0.00001987, -0.00004084, 0.00003927, -0.00005298, 0.00001285, 0.00001919, -0.00004753, 0.00001887, -0.00004091, -0.00002013, 0.00001819, -0.00001414, -0.00001011, -0.00002041, -0.00001029, 0.00001226, },
		{ -0.00001164, -0.00004327, 0.00001143, -0.00003870, -0.00001397, 0.00002255, -0.00001981, 0.00005971, -0.00004672, 0.00002955, 0.00001662, 0.00000080, 0.00004421, -0.00000199, 0.00003269, -0.00004068, 0.00009704, -0.00003307, 0.00004032, -0.00005181, -0.00000254, -0.00003391, 0.00000439, -0.00003382, 0.00002709, -0.00002004, -0.00000851, -0.00004416, 0.00003501, -0.00007826, 0.00001809, -0.00010546, 0.00006936, -0.00022815, -0.00003033, -0.00094751, 0.00098986, 0.00050409, 0.00106438, -0.00002480, -0.00070204, 0.00007563, -0.00018927, 0.00003373, -0.00013027, 0.00001894, -0.00004488, 0.00005066, -0.00009578, 0.00004201, -0.00004395, 0.00004172, -0.00007842, 0.00008759, -0.00008791, 0.00004322, -0.00005280, -0.00001516, -0.00000277, 0.00001477, -0.00003240, -0.00000456, -0.00002390, 0.00000990, },
		{ -0.00000767, -0.00002764, -0.00001721, 0.00000312, -0.00006154, 0.00003012, -0.00001289, 0.00002006, 0.00003744, -0.00005587, 0.00007654, -0.00003657, 0.00006796, -0.00001094, 0.00004893, -0.00004996, 0.00008554, -0.00000019, 0.00002445, -0.00003033, -0.00002582, -0.00002031, -0.00000324, -0.00002969, 0.00001263, -0.00000726, 0.00000120, -0.00006154, 0.00004203, -0.00007512, 0.00003980, -0.00013053, 0.00008120, -0.00020263, 0.00004884, -0.00079889, 0.00031225, 0.00087402, 0.00038403, 0.00096048, -0.00080772, -0.00007550, -0.00016988, 0.00002038, -0.00012527, 0.00000763, -0.00007418, 0.00008313, -0.00009849, 0.00002976, -0.00005315, 0.00005986, -0.00008407, 0.00007230, -0.00005691, 0.00001052, -0.00002013, -0.00003880, -0.00000499, 0.00002068, -0.00002645, -0.00001360, -0.00001953, 0.00000165, },
		{ 0.00000932, -0.00003631, -0.00000582, -0.00001413, -0.00002004, -0.00005315, 0.00005994, -0.00005894, 0.00010438, -0.00006778, 0.00004678, 0.00000639, 0.00002343, 0.00003347, 0.00002192, -0.00001366, 0.00002941, 0.00006368, -0.00001429, 0.00001832, -0.00007091, 0.00000923, -0.00003157, 0.00000315, -0.00002950, 0.00002534, -0.00001572, -0.00003285, -0.00000578, -0.00002126, -0.00000602, -0.00005391, -0.00002361, -0.00007021, -0.00004649, -0.00043076, -0.00035033, 0.00109313, 0.00008314, 0.00098868, -0.00003952, -0.00053930, -0.00002640, -0.00011186, -0.00002236, -0.00007749, -0.00003061, 0.00002105, -0.00001231, -0.00004251, -0.00001173, 0.00001472, -0.00002258, -0.00000033, 0.00001784, -0.00005476, 0.00003592, -0.00007619, 0.00000885, -0.00000194, 0.00001045, -0.00004482, 0.00000079, -0.00001952, },
		{ 0.00001230, -0.00003973, 0.00000345, -0.00002910, 0.00001132, -0.00008377, 0.00003941, -0.00002514, 0.00003424, 0.00004506, -0.00005716, 0.00007983, -0.00002716, 0.00006819, 0.00000609, 0.00001941, -0.00001815, 0.00010200, -0.00002186, 0.00003781, -0.00007683, -0.00000280, -0.00002609, 0.00000353, -0.00003012, 0.00001058, 0.00000408, -0.00003166, -0.00001761, -0.00001368, -0.00001368, -0.00001731, -0.00005833, -0.00003758, -0.00006618, -0.00020299, -0.00059876, 0.00079336, 0.00038735, 0.00038749, 0.00074407, -0.00055573, -0.00018204, -0.00009123, -0.00004096, -0.00006338, -0.00005181, -0.00000186, 0.00002097, -0.00004843, -0.00001894, -0.00000319, 0.00001082, -0.00003586, 0.00004902, -0.00007284, 0.00004258, -0.00006503, -0.00001231, -0.00000320, 0.00002214, -0.00004268, -0.00000881, -0.00001902, },
		{ -0.00000584, -0.00001542, -0.00002561, 0.00000411, -0.00002218, -0.00002222, -0.00006381, 0.00005905, -0.00006003, 0.00012529, -0.00007729, 0.00006134, 0.00000088, 0.00003959, 0.00003223, 0.00002426, -0.00002801, 0.00008941, 0.00001688, 0.00001837, -0.00004066, -0.00004681, -0.00000127, -0.00001678, -0.00000487, -0.00002400, 0.00002834, -0.00003346, -0.00001471, -0.00002326, -0.00001103, -0.00000548, -0.00004577, -0.00004987, -0.00006675, -0.00007302, -0.00058456, 0.00032226, 0.00069597, 0.00005970, 0.00077085, 0.00006493, -0.00049233, -0.00004320, -0.00010393, -0.00001611, -0.00008817, -0.00000614, 0.00001014, -0.00001203, -0.00004350, -0.00000946, 0.00001683, -0.00003327, 0.00003747, -0.00004546, 0.00000414, -0.00001510, -0.00005723, 0.00001444, 0.00000361, -0.00000638, -0.00004273, -0.00000024, },
		{ -0.00001030, 0.00000233, -0.00004913, 0.00002637, -0.00004171, 0.00001462, -0.00009603, 0.00003837, -0.00003876, 0.00006738, 0.00002507, -0.00002808, 0.00005996, 0.00000216, 0.00005205, 0.00003151, -0.00001995, 0.00005264, 0.00006445, 0.00000114, -0.00000866, -0.00007147, -0.00000312, -0.00001717, 0.00000044, -0.00002663, 0.00001138, -0.00000229, -0.00003565, -0.00000349, -0.00004166, 0.00002643, -0.00005495, -0.00001607, -0.00012402, 0.00003839, -0.00049699, -0.00002522, 0.00069857, 0.00017425, 0.00035527, 0.00059729, -0.00036930, -0.00025430, -0.00004025, -0.00007365, -0.00003618, -0.00006860, 0.00003313, -0.00002107, -0.00001178, -0.00005271, 0.00003645, -0.00003829, 0.00003150, -0.00002298, -0.00002448, 0.00001460, -0.00007052, 0.00000498, -0.00000016, 0.00001168, -0.00004833, -0.00000637, },
		{ -0.00000102, -0.00000035, -0.00003831, 0.00000054, -0.00000639, -0.00001701, -0.00003728, -0.00005427, 0.00002318, -0.00001523, 0.00010087, -0.00004616, 0.00004589, 0.00002304, 0.00003023, 0.00005361, -0.00000558, 0.00001671, 0.00009086, 0.00000987, 0.00000095, -0.00006686, -0.00002694, -0.00000881, -0.00000471, -0.00001266, -0.00001615, 0.00002232, -0.00004388, 0.00000770, -0.00005720, 0.00002992, -0.00004449, 0.00000547, -0.00014057, 0.00003837, -0.00030786, -0.00029452, 0.00058022, 0.00031798, 0.00015563, 0.00052944, 0.00020134, -0.00049720, -0.00002045, -0.00012316, 0.00000663, -0.00010982, 0.00003851, -0.00003518, 0.00002281, -0.00007553, 0.00002968, -0.00002795, 0.00001648, 0.00000075, -0.00004105, 0.00001921, -0.00005270, -0.00002602, 0.00001069, 0.00000533, -0.00002405, -0.00003277, },
		{ -0.00000904, 0.00000037, -0.00002116, -0.00003010, 0.00002260, -0.00003628, -0.00000328, -0.00007817, -0.00000175, -0.00001364, 0.00006985, 0.00002776, -0.00001196, 0.00005403, 0.00001768, 0.00005081, 0.00002984, -0.00001494, 0.00009567, 0.00003311, 0.00000652, -0.00005860, -0.00004062, -0.00001934, 0.00000253, -0.00001153, -0.00001719, 0.00000781, -0.00001555, -0.00001540, -0.00002756, -0.00001812, 0.00000770, -0.00002668, -0.00006649, -0.00007946, -0.00006600, -0.00048176, 0.00043520, 0.00034983, 0.00020613, 0.00022061, 0.00056518, -0.00029248, -0.00025067, -0.00004602, -0.00006016, -0.00005338, -0.00003098, 0.00000605, -0.00000816, -0.00002447, -0.00003291, 0.00001820, -0.00002702, 0.00004162, -0.00006146, 0.00002181, -0.00003712, -0.00004152, 0.00000392, 0.00000705, -0.00001019, -0.00003866, },
		{ -0.00002509, 0.00000187, -0.00001149, -0.00003619, 0.00001067, -0.00000980, -0.00002403, -0.00003413, -0.00006337, -0.00000194, 0.00002788, 0.00007151, -0.00000373, 0.00002340, 0.00005161, 0.00001728, 0.00006862, -0.00002404, 0.00007685, 0.00006191, 0.00001618, -0.00005183, -0.00004550, -0.00003457, 0.00000316, -0.00000781, -0.00001148, -0.00001163, 0.00000596, -0.00002964, -0.00000185, -0.00005419, 0.00003457, -0.00004510, 0.00000051, -0.00015828, 0.00004514, -0.00045795, 0.00017090, 0.00042271, 0.00019471, 0.00015535, 0.00040212, 0.00025381, -0.00045845, -0.00003054, -0.00011136, -0.00000635, -0.00008573, 0.00003607, -0.00004502, 0.00002780, -0.00007535, 0.00003857, -0.00004922, 0.00005625, -0.00005393, 0.00000540, -0.00001703, -0.00004758, -0.00001112, 0.00001168, -0.00000706, -0.00002802, },
		{ -0.00002208, -0.00001268, 0.00000044, -0.00003331, -0.00000887, 0.00001028, -0.00002913, -0.00002090, -0.00005524, -0.00005600, 0.00003142, 0.00004781, 0.00005835, -0.00001749, 0.00006939, 0.00001517, 0.00006178, 0.00001135, 0.00003444, 0.00009331, 0.00002477, -0.00003477, -0.00005992, -0.00003093, -0.00002100, 0.00001138, -0.00001862, -0.00000794, -0.00000529, -0.00001363, -0.00000745, -0.00004410, 0.00000664, -0.00001647, 0.00000200, -0.00012471, -0.00001059, -0.00026592, -0.00014447, 0.00050642, 0.00012102, 0.00024627, 0.00011427, 0.00054097, -0.00024438, -0.00023013, -0.00005815, -0.00005045, -0.00005829, -0.00000191, -0.00002518, 0.00001212, -0.00004275, -0.00000195, -0.00002124, 0.00001938, -0.00000334, -0.00003791, 0.00001661, -0.00006081, -0.00001237, -0.00000103, 0.00000648, -0.00002860, },
		{ -0.00001617, -0.00002295, 0.00000173, -0.00002502, -0.00001796, 0.00000345, -0.00000387, -0.00004219, -0.00001225, -0.00009782, 0.00000100, 0.00003938, 0.00007371, 0.00001580, 0.00002304, 0.00006196, 0.00002028, 0.00005866, 0.00000277, 0.00010593, 0.00004212, -0.00001446, -0.00007550, -0.00002255, -0.00004481, 0.00001837, -0.00001714, -0.00000365, -0.00001682, -0.00000425, -0.00001079, -0.00002624, -0.00002459, 0.00000640, -0.00001001, -0.00006052, -0.00008852, -0.00010779, -0.00031274, 0.00042936, 0.00017444, 0.00020883, 0.00008954, 0.00036109, 0.00023353, -0.00037818, -0.00006799, -0.00008100, -0.00003421, -0.00003871, -0.00000053, -0.00002147, 0.00000192, -0.00003849, 0.00000327, -0.00001936, 0.00003882, -0.00006210, 0.00002927, -0.00005775, -0.00001856, -0.00001191, 0.00001477, -0.00002850, },
		{ -0.00002985, -0.00000937, -0.00001679, -0.00000829, -0.00002602, -0.00000311, 0.00000611, -0.00003539, -0.00001785, -0.00005596, -0.00008640, 0.00005160, 0.00004456, 0.00008179, -0.00001673, 0.00008219, 0.00001620, 0.00005975, 0.00001603, 0.00008457, 0.00007705, -0.00000260, -0.00007512, -0.00003238, -0.00004232, -0.00000510, 0.00000371, -0.00001364, -0.00000815, -0.00001815, 0.00000107, -0.00002163, -0.00002987, -0.00000237, -0.00000324, -0.00002657, -0.00010283, -0.00006226, -0.00030695, 0.00021291, 0.00031806, 0.00008576, 0.00019836, 0.00010027, 0.00046121, -0.00017032, -0.00023179, -0.00004968, -0.00006368, -0.00002992, -0.00001341, -0.00002145, 0.00000137, -0.00002609, -0.00000788, -0.00002592, 0.00003718, -0.00004051, 0.00000332, -0.00002479, -0.00004492, -0.00000282, 0.00000059, -0.00000947, },
		{ -0.00004184, 0.00000125, -0.00002533, -0.00000511, -0.00002557, -0.00000535, -0.00000298, -0.00000613, -0.00004777, -0.00000131, -0.00012446, -0.00000004, 0.00004283, 0.00009475, 0.00001917, 0.00003975, 0.00005994, 0.00002753, 0.00004623, 0.00005739, 0.00010831, 0.00001447, -0.00006730, -0.00005213, -0.00002816, -0.00003333, 0.00001604, -0.00001632, -0.00000124, -0.00002631, -0.00000049, -0.00000949, -0.00003115, -0.00001166, -0.00000505, -0.00000302, -0.00008900, -0.00005553, -0.00025245, 0.00001612, 0.00038082, 0.00006462, 0.00019295, 0.00006315, 0.00031485, 0.00021029, -0.00030571, -0.00009436, -0.00006686, -0.00003672, -0.00001842, -0.00002037, -0.00001024, -0.00000964, -0.00001268, -0.00002938, 0.00001936, -0.00000781, -0.00002501, 0.00000516, -0.00006348, 0.00000177, -0.00001214, 0.00000666, },
		{ -0.00003135, -0.00001425, -0.00000637, -0.00002274, -0.00001330, -0.00001370, -0.00000438, 0.00000177, -0.00004192, -0.00000836, -0.00006901, -0.00009738, 0.00004045, 0.00007536, 0.00008074, 0.00000591, 0.00008015, 0.00002698, 0.00004419, 0.00005499, 0.00011515, 0.00005182, -0.00006417, -0.00006617, -0.00002464, -0.00004142, 0.00000327, -0.00000337, -0.00000648, -0.00001658, -0.00001973, 0.00000815, -0.00003266, -0.00001126, -0.00001929, 0.00001563, -0.00007178, -0.00004625, -0.00020184, -0.00011386, 0.00033957, 0.00013839, 0.00010780, 0.00013072, 0.00012177, 0.00036037, -0.00009060, -0.00023784, -0.00004171, -0.00007032, -0.00000410, -0.00003389, -0.00000634, -0.00001904, 0.00000603, -0.00003863, 0.00000309, 0.00000974, -0.00003096, 0.00001045, -0.00005489, -0.00001639, -0.00000479, 0.00000138, },
		{ -0.00001643, -0.00003458, 0.00001089, -0.00002994, -0.00001204, -0.00001704, -0.00000174, -0.00000561, -0.00001715, -0.00003664, -0.00001216, -0.00012440, -0.00003242, 0.00006939, 0.00010607, 0.00003028, 0.00005429, 0.00005981, 0.00002596, 0.00005654, 0.00011200, 0.00009608, -0.00005216, -0.00008211, -0.00002752, -0.00003914, -0.00001415, 0.00000253, -0.00000608, -0.00000916, -0.00002875, 0.00000490, -0.00001723, -0.00001846, -0.00002050, 0.00000783, -0.00003737, -0.00005452, -0.00012979, -0.00021426, 0.00026136, 0.00020513, 0.00007165, 0.00012910, 0.00008737, 0.00024111, 0.00020605, -0.00025427, -0.00010580, -0.00006377, -0.00002120, -0.00002552, -0.00001473, -0.00002182, 0.00000723, -0.00002097, -0.00002658, 0.00002228, -0.00002946, 0.00000782, -0.00003703, -0.00003796, 0.00000321, -0.00000633, },
		{ -0.00001140, -0.00004040, 0.00000475, -0.00001268, -0.00002605, -0.00001223, -0.00000605, -0.00000396, -0.00001135, -0.00003382, -0.00001413, -0.00006747, -0.00011953, 0.00002701, 0.00011753, 0.00007152, 0.00003886, 0.00007187, 0.00003657, 0.00003700, 0.00010976, 0.00013276, -0.00001619, -0.00010667, -0.00003100, -0.00003888, -0.00002093, -0.00000661, 0.00000158, -0.00000869, -0.00002226, -0.00001461, 0.00000262, -0.00002539, -0.00001247, -0.00001409, -0.00000169, -0.00006884, -0.00005746, -0.00026220, 0.00014281, 0.00025710, 0.00007583, 0.00009356, 0.00010551, 0.00011996, 0.00028295, -0.00003411, -0.00022974, -0.00004683, -0.00005528, -0.00000545, -0.00003312, -0.00000957, -0.00001683, 0.00001539, -0.00005271, 0.00002250, -0.00002356, 0.00000335, -0.00001983, -0.00004946, -0.00000187, -0.00000333, },
		{ -0.00000525, -0.00004085, -0.00001040, 0.00000468, -0.00002970, -0.00001596, -0.00000901, 0.00000085, -0.00001395, -0.00002087, -0.00002863, -0.00002435, -0.00011690, -0.00008249, 0.00011388, 0.00010648, 0.00005333, 0.00006274, 0.00006036, 0.00002553, 0.00008647, 0.00016518, 0.00003906, -0.00012406, -0.00004833, -0.00003321, -0.00002783, -0.00001175, -0.00000276, -0.00000083, -0.00001841, -0.00002330, 0.00000081, -0.00001438, -0.00001395, -0.00002170, 0.00000610, -0.00005849, -0.00002983, -0.00022151, -0.00002382, 0.00030060, 0.00008484, 0.00008610, 0.00008059, 0.00010858, 0.00017396, 0.00020037, -0.00020870, -0.00011722, -0.00005033, -0.00001787, -0.00002827, -0.00001179, -0.00002785, 0.00002316, -0.00003791, -0.00000750, -0.00000307, -0.00001298, 0.00000346, -0.00005695, -0.00001030, -0.00000179, },
		{ 0.00000200, -0.00003714, -0.00002588, 0.00000983, -0.00001890, -0.00002388, -0.00001245, 0.00000174, -0.00000852, -0.00002210, -0.00002134, -0.00002859, -0.00005348, -0.00015824, 0.00002491, 0.00015361, 0.00007062, 0.00006996, 0.00006252, 0.00004605, 0.00004484, 0.00017657, 0.00011233, -0.00011925, -0.00008544, -0.00002226, -0.00003887, -0.00000746, -0.00001730, 0.00000826, -0.00001632, -0.00002028, -0.00001427, -0.00000069, -0.00001516, -0.00001953, -0.00000384, -0.00003469, -0.00003529, -0.00013605, -0.00016261, 0.00027233, 0.00013405, 0.00007179, 0.00006319, 0.00010523, 0.00009807, 0.00023025, 0.00000043, -0.00021384, -0.00005038, -0.00003916, -0.00001627, -0.00002290, -0.00002105, 0.00000081, 0.00000118, -0.00004052, 0.00001121, -0.00002673, 0.00001931, -0.00005232, -0.00002276, -0.00000325, },
		{ -0.00000132, -0.00001991, -0.00004638, 0.00001162, -0.00001220, -0.00001809, -0.00002426, 0.00000158, -0.00000137, -0.00002336, -0.00001667, -0.00003033, -0.00002623, -0.00012043, -0.00012661, 0.00015387, 0.00011245, 0.00007915, 0.00006759, 0.00006178, 0.00003407, 0.00013508, 0.00019961, -0.00008378, -0.00013087, -0.00002236, -0.00004208, -0.00000867, -0.00002104, 0.00000041, -0.00000450, -0.00002039, -0.00001941, -0.00000575, -0.00000321, -0.00002112, -0.00000915, -0.00002240, -0.00003881, -0.00007184, -0.00020570, 0.00014818, 0.00021794, 0.00004596, 0.00007394, 0.00006833, 0.00010161, 0.00013290, 0.00018224, -0.00016549, -0.00012471, -0.00003239, -0.00002721, -0.00001964, -0.00002054, -0.00001526, 0.00001734, -0.00003704, -0.00000417, -0.00001985, 0.00001072, -0.00002558, -0.00004470, 0.00000052, },
		},
		{ { -0.00002250, -0.00002246, -0.00000298, -0.00002932, -0.00003997, -0.00004615, -0.00015795, 0.00000828, 0.00020178, 0.00007973, 0.00007035, 0.00007093, 0.00007501, 0.00007884, 0.00020757, 0.00000916, -0.00015851, -0.00005257, 0.00000600, -0.00002982, -0.00002408, -0.00001735, -0.00003590, 0.00000526, -0.00002509, -0.00001367, -0.00002704, -0.00001406, -0.00000333, 0.00000220, -0.00004479, 0.00000153, 0.00000139, 0.00000153, -0.00004479, 0.00000220, -0.00000333, -0.00001406, -0.00002704, -0.00001367, -0.00002509, 0.00000526, -0.00003590, -0.00001735, -0.00002408, -0.00002982, 0.00000600, -0.00005257, -0.00015851, 0.00000916, 0.00020757, 0.00007884, 0.00007501, 0.00007093, 0.00007035, 0.00007973, 0.00020178, 0.00000828, -0.00015795, -0.00004615, -0.00003997, -0.00002932, -0.00000298, -0.00002246, },
		{ -0.00002264, -0.00002462, -0.00001654, -0.00000710, -0.00003766, -0.00003897, -0.00006006, -0.00018034, 0.00008653, 0.00020247, 0.00006729, 0.00008263, 0.00007906, 0.00007296, 0.00014738, 0.00018810, -0.00013908, -0.00011244, -0.00003396, 0.00000932, -0.00004133, -0.00000346, -0.00004770, -0.00000950, 0.00000068, -0.00002748, -0.00001932, -0.00002268, -0.00001116, 0.00000321, -0.00001610, -0.00003390, 0.00000711, 0.00000257, -0.00001348, -0.00003989, 0.00001251, -0.00001057, -0.00001955, -0.00002326, -0.00001288, -0.00001665, -0.00001145, -0.00003216, -0.00000548, -0.00004212, -0.00001491, 0.00000154, -0.00010222, -0.00011889, 0.00013683, 0.00014497, 0.00005814, 0.00006879, 0.00006809, 0.00005429, 0.00010585, 0.00017169, -0.00004225, -0.00013877, -0.00003320, -0.00004250, -0.00001610, -0.00000755, },
		{ -0.00001654, -0.00001964, -0.00002765, -0.00000982, -0.00001571, -0.00004380, -0.00003397, -0.00008744, -0.00018862, 0.00018241, 0.00017123, 0.00007701, 0.00008501, 0.00009125, 0.00008302, 0.00023760, 0.00004034, -0.00019393, -0.00005402, -0.00003347, 0.00001163, -0.00004207, -0.00000132, -0.00006653, 0.00002613, -0.00002185, -0.00002172, -0.00002760, -0.00001218, -0.00001096, 0.00000398, -0.00003100, -0.00001564, 0.00000510, 0.00000236, -0.00002918, -0.00002424, 0.00001271, -0.00001753, -0.00002249, -0.00001733, -0.00001103, -0.00001664, -0.00002801, -0.00000968, -0.00001841, -0.00004521, 0.00000021, -0.00002378, -0.00013669, -0.00000761, 0.00017367, 0.00007966, 0.00006075, 0.00005776, 0.00006413, 0.00004620, 0.00012813, 0.00012985, -0.00007517, -0.00012072, -0.00002189, -0.00004377, -0.00000320, },
		{ 0.00000081, -0.00002119, -0.00002038, -0.00002491, -0.00001066, -0.00002020, -0.00005382, -0.00001815, -0.00014531, -0.00015036, 0.00026130, 0.00013179, 0.00009457, 0.00008886, 0.00009480, 0.00014667, 0.00024864, -0.00015914, -0.00012446, -0.00005346, -0.00000966, -0.00000517, -0.00002406, -0.00002965, -0.00004609, 0.00003639, -0.00003901, -0.00001957, -0.00002754, -0.00000543, -0.00000936, -0.00000486, -0.00003143, -0.00000340, 0.00000279, -0.00000445, -0.00003473, -0.00000982, 0.00000761, -0.00002446, -0.00001986, -0.00001313, -0.00000775, -0.00003065, -0.00002364, 0.00000499, -0.00004186, -0.00003012, 0.00000128, -0.00006482, -0.00011232, 0.00010260, 0.00013306, 0.00005255, 0.00005872, 0.00005119, 0.00005534, 0.00005031, 0.00013721, 0.00008903, -0.00010145, -0.00009619, -0.00001862, -0.00003733, },
		{ -0.00002264, -0.00000625, -0.00001810, -0.00002666, -0.00001635, -0.00001949, -0.00002013, -0.00006320, -0.00000262, -0.00022720, -0.00005142, 0.00029553, 0.00010771, 0.00010600, 0.00009963, 0.00010155, 0.00025677, 0.00009468, -0.00025027, -0.00003378, -0.00008335, 0.00002995, -0.00003205, -0.00000551, -0.00006397, 0.00000081, 0.00001527, -0.00004101, -0.00002373, -0.00001770, -0.00000552, -0.00000835, -0.00001614, -0.00002041, -0.00000058, 0.00000126, -0.00001437, -0.00002986, -0.00000277, 0.00000121, -0.00002962, -0.00001342, -0.00000985, -0.00001054, -0.00004673, 0.00000153, -0.00000633, -0.00005166, -0.00001109, -0.00002026, -0.00008848, -0.00003734, 0.00015620, 0.00006829, 0.00005972, 0.00004108, 0.00005547, 0.00004015, 0.00006547, 0.00013238, 0.00005158, -0.00012210, -0.00006495, -0.00002434, },
		{ -0.00002970, -0.00000982, -0.00001487, -0.00001512, -0.00002994, -0.00001288, -0.00002437, -0.00002716, -0.00005487, -0.00001920, -0.00029181, 0.00008145, 0.00029258, 0.00009440, 0.00012253, 0.00009785, 0.00016765, 0.00029616, -0.00015247, -0.00017020, -0.00003033, -0.00006955, 0.00003915, -0.00003849, -0.00001559, -0.00006332, 0.00003107, -0.00001041, -0.00004101, -0.00002117, -0.00001095, -0.00000496, -0.00001533, -0.00001605, -0.00001260, 0.00000070, -0.00000523, -0.00001684, -0.00002549, 0.00000135, -0.00000758, -0.00002901, -0.00000749, -0.00000659, -0.00002522, -0.00004585, 0.00002574, -0.00003707, -0.00003425, -0.00001485, -0.00004306, -0.00007793, 0.00005080, 0.00014140, 0.00002862, 0.00007074, 0.00001957, 0.00006488, 0.00002322, 0.00008599, 0.00011790, 0.00001329, -0.00012887, -0.00003744, },
		{ -0.00002263, -0.00002652, -0.00000780, -0.00001599, -0.00001913, -0.00002474, -0.00002071, -0.00001861, -0.00004483, -0.00002610, -0.00008511, -0.00030037, 0.00021212, 0.00026882, 0.00008617, 0.00014781, 0.00009009, 0.00030600, 0.00013645, -0.00028772, -0.00004338, -0.00008575, -0.00001187, 0.00001452, -0.00002921, -0.00004239, -0.00002822, 0.00002804, -0.00002450, -0.00004439, -0.00000852, -0.00001271, -0.00000293, -0.00002668, -0.00000380, -0.00001447, 0.00000405, -0.00001578, -0.00000940, -0.00002920, 0.00000840, -0.00002044, -0.00002010, -0.00000566, -0.00000560, -0.00004724, -0.00001914, 0.00002192, -0.00005516, -0.00001105, -0.00004266, -0.00004132, -0.00004377, 0.00011836, 0.00008130, 0.00003375, 0.00005693, 0.00001640, 0.00006336, 0.00001612, 0.00010281, 0.00009789, -0.00002686, -0.00011676, },
		{ -0.00009549, -0.00001474, -0.00002131, -0.00001017, -0.00001525, -0.00002271, -0.00002150, -0.00002779, -0.00001471, -0.00005510, -0.00000542, -0.00017540, -0.00025381, 0.00033334, 0.00021440, 0.00011014, 0.00014366, 0.00015487, 0.00038130, -0.00016110, -0.00020721, -0.00002170, -0.00009446, 0.00002236, -0.00000105, -0.00004048, -0.00004021, -0.00000364, 0.00001802, -0.00004054, -0.00003495, -0.00000313, -0.00001115, -0.00001088, -0.00002523, 0.00000346, -0.00001577, 0.00000126, -0.00001764, -0.00000536, -0.00003349, 0.00001445, -0.00003209, -0.00000723, -0.00000686, -0.00001289, -0.00005940, 0.00001701, -0.00001154, -0.00003690, -0.00002126, -0.00005613, -0.00002391, 0.00000550, 0.00013835, 0.00002341, 0.00006055, 0.00002214, 0.00003513, 0.00004488, 0.00002436, 0.00011159, 0.00007016, -0.00005866, },
		{ -0.00007260, -0.00007759, -0.00000454, -0.00002356, -0.00000729, -0.00002044, -0.00001865, -0.00002902, -0.00002377, -0.00002147, -0.00004769, -0.00001666, -0.00025821, -0.00016329, 0.00043720, 0.00013073, 0.00017771, 0.00009965, 0.00033820, 0.00021051, -0.00034158, -0.00005236, -0.00008394, -0.00004262, 0.00001620, -0.00000255, -0.00006674, -0.00000407, -0.00001013, 0.00001806, -0.00006385, -0.00000759, -0.00001391, -0.00000133, -0.00002975, -0.00000547, -0.00000510, -0.00000862, -0.00001138, -0.00000355, -0.00001769, -0.00002577, 0.00001034, -0.00003335, 0.00000073, -0.00000755, -0.00003188, -0.00004381, 0.00002779, -0.00003654, -0.00000884, -0.00006226, -0.00002986, -0.00001263, 0.00006354, 0.00010346, 0.00000786, 0.00006787, -0.00000311, 0.00005550, 0.00002075, 0.00004326, 0.00011123, 0.00003411, },
		{ 0.00000445, -0.00007820, -0.00005780, -0.00000002, -0.00002478, -0.00000720, -0.00002141, -0.00002007, -0.00003292, -0.00002264, -0.00002092, -0.00004752, -0.00004156, -0.00033878, -0.00000701, 0.00047191, 0.00008066, 0.00022393, 0.00011860, 0.00049116, -0.00016760, -0.00024690, -0.00002515, -0.00009366, -0.00001516, 0.00002294, -0.00003511, -0.00005261, 0.00001252, -0.00001417, 0.00000625, -0.00006875, 0.00001214, -0.00002286, 0.00000035, -0.00003850, 0.00001264, -0.00001699, -0.00000317, -0.00001955, 0.00001213, -0.00003897, -0.00000779, -0.00000279, -0.00002422, 0.00000192, -0.00001133, -0.00005013, -0.00000860, 0.00000147, -0.00001801, -0.00002432, -0.00007770, 0.00000519, 0.00000304, 0.00010052, 0.00004923, 0.00002604, 0.00004687, -0.00000404, 0.00006285, 0.00000062, 0.00007051, 0.00009309, },
		{ 0.00005136, -0.00000291, -0.00009075, -0.00002527, -0.00001362, -0.00001201, -0.00002096, -0.00000749, -0.00003940, -0.00001937, -0.00003573, -0.00000205, -0.00007437, -0.00006050, -0.00042011, 0.00022423, 0.00039856, 0.00012150, 0.00019250, 0.00030408, 0.00035637, -0.00043902, -0.00003449, -0.00011464, -0.00002380, -0.00003844, 0.00005274, -0.00009766, 0.00001389, -0.00001917, 0.00000889, -0.00003106, -0.00003995, 0.00000498, -0.00001580, -0.00001561, -0.00002345, 0.00001084, -0.00001740, -0.00000997, -0.00000893, 0.00000898, -0.00004960, 0.00000614, -0.00001314, -0.00001498, 0.00000279, -0.00002724, -0.00004355, 0.00000703, -0.00002081, 0.00001528, -0.00008154, -0.00003557, 0.00001054, 0.00004310, 0.00008637, 0.00002572, 0.00003535, 0.00002547, 0.00000484, 0.00006093, -0.00001381, 0.00010376, },
		{ 0.00012225, 0.00000688, -0.00000013, -0.00010037, 0.00000591, -0.00003153, 0.00000156, -0.00003261, 0.00000172, -0.00005536, -0.00000980, -0.00003583, -0.00000207, -0.00009086, -0.00011270, -0.00043866, 0.00045390, 0.00028067, 0.00019712, 0.00017909, 0.00052530, -0.00008477, -0.00035975, 0.00002204, -0.00013986, 0.00001109, -0.00003531, 0.00003744, -0.00011316, 0.00006253, -0.00005392, 0.00002808, -0.00006161, -0.00000614, -0.00001029, -0.00000925, -0.00002860, -0.00000328, -0.00000016, -0.00001452, -0.00001711, 0.00000975, -0.00001101, -0.00004462, 0.00001099, -0.00001675, -0.00000912, 0.00000147, -0.00004782, -0.00001184, -0.00001868, 0.00000809, -0.00000573, -0.00010624, 0.00001571, 0.00000779, 0.00008064, 0.00004483, 0.00003357, 0.00002305, 0.00001799, 0.00001420, 0.00004898, -0.00001112, },
		{ 0.00002397, 0.00009982, -0.00000760, -0.00001517, -0.00008330, 0.00000938, -0.00002765, -0.00000772, -0.00001857, -0.00001993, -0.00003978, -0.00003012, 0.00000402, -0.00005982, -0.00005080, -0.00024774, -0.00031799, 0.00057972, 0.00022022, 0.00021467, 0.00033669, 0.00045975, -0.00047690, -0.00007758, -0.00010143, -0.00004069, -0.00004224, 0.00002810, -0.00004671, -0.00004251, 0.00003889, -0.00004554, 0.00001153, -0.00005459, 0.00000151, -0.00001081, -0.00001985, -0.00001872, -0.00000133, -0.00000139, -0.00002278, -0.00000648, 0.00001314, -0.00002760, -0.00003647, 0.00001579, -0.00002255, 0.00000177, -0.00001595, -0.00004422, 0.00000311, -0.00004324, 0.00006039, -0.00008612, -0.00005257, 0.00001689, 0.00004187, 0.00006696, 0.00003368, 0.00002713, 0.00001982, 0.00000631, 0.00003411, 0.00001725, },
		{ -0.00002242, 0.00007634, 0.00004583, 0.00000727, -0.00004625, -0.00004623, -0.00000880, -0.00000872, -0.00002867, 0.00000528, -0.00005433, -0.00001176, -0.00005344, 0.00004500, -0.00013551, 0.00000733, -0.00042205, -0.00008589, 0.00060917, 0.00020166, 0.00024417, 0.00057074, 0.00000944, -0.00047192, 0.00005512, -0.00017972, 0.00004079, -0.00008580, 0.00006687, -0.00011209, 0.00004197, -0.00001583, -0.00001913, -0.00001455, -0.00003276, -0.00000271, -0.00001073, -0.00002837, -0.00000337, -0.00000783, -0.00000045, -0.00003115, 0.00001316, 0.00000050, -0.00003426, -0.00002926, 0.00002164, -0.00003107, 0.00001556, -0.00004653, -0.00000772, -0.00003113, -0.00000562, 0.00005001, -0.00013586, 0.00001967, -0.00000616, 0.00009191, 0.00001519, 0.00005999, -0.00000825, 0.00003950, -0.00001656, 0.00006087, },
		{ 0.00006695, -0.00003617, 0.00010494, 0.00000132, 0.00001849, -0.00006329, -0.00002241, -0.00001609, -0.00000588, -0.00002821, 0.00000109, -0.00005983, -0.00001271, -0.00003287, 0.00002200, -0.00014933, -0.00000545, -0.00053227, 0.00016013, 0.00062271, 0.00014111, 0.00046031, 0.00051543, -0.00045702, -0.00017385, -0.00005160, -0.00009550, 0.00000403, -0.00004402, 0.00001846, -0.00008498, 0.00005950, -0.00003840, -0.00001723, -0.00001240, -0.00003166, 0.00000754, -0.00003084, -0.00001016, -0.00001047, 0.00000058, -0.00001637, -0.00001705, 0.00001541, -0.00000681, -0.00004430, -0.00001116, 0.00001412, -0.00002611, 0.00001005, -0.00005563, 0.00001772, -0.00007338, 0.00007336, -0.00004355, -0.00008752, 0.00002627, 0.00001757, 0.00008889, -0.00000419, 0.00006730, -0.00002998, 0.00005074, -0.00002526, },
		{ -0.00000607, 0.00004391, -0.00001559, 0.00009580, -0.00001619, 0.00001452, -0.00006080, -0.00001721, -0.00000926, -0.00001749, -0.00001198, -0.00002465, -0.00004274, -0.00002945, 0.00000989, -0.00004984, -0.00010315, -0.00009737, -0.00055711, 0.00041169, 0.00058939, 0.00012897, 0.00077751, 0.00000507, -0.00049762, 0.00000499, -0.00015235, 0.00000568, -0.00005610, 0.00000588, -0.00004229, -0.00002068, 0.00003471, -0.00004229, -0.00001972, -0.00000278, -0.00003589, 0.00001525, -0.00004914, 0.00001547, -0.00002883, 0.00001524, -0.00003798, 0.00001166, -0.00000116, -0.00000328, -0.00005987, 0.00001833, -0.00000948, -0.00000387, -0.00002185, -0.00002206, -0.00001391, -0.00004767, 0.00009367, -0.00012264, -0.00000101, -0.00001369, 0.00008257, 0.00002648, 0.00003321, 0.00002524, -0.00000844, 0.00003043, },
		{ 0.00000045, 0.00001830, 0.00001800, 0.00001050, 0.00007443, -0.00002597, 0.00001585, -0.00006462, -0.00000384, -0.00001599, -0.00001214, -0.00001717, -0.00003074, -0.00004433, -0.00001370, 0.00000806, -0.00009353, -0.00007032, -0.00020432, -0.00054409, 0.00072862, 0.00039982, 0.00040465, 0.00075193, -0.00056263, -0.00017410, -0.00009010, -0.00007837, -0.00001510, -0.00003425, -0.00002146, -0.00002629, -0.00000681, 0.00003159, -0.00006980, 0.00001318, -0.00002607, -0.00001148, -0.00001030, -0.00003020, 0.00001019, -0.00002519, 0.00000766, -0.00003413, 0.00002171, -0.00000921, -0.00000939, -0.00006042, 0.00003630, -0.00002584, 0.00000869, -0.00004471, 0.00001379, -0.00006933, 0.00003949, 0.00001382, -0.00009716, 0.00001781, -0.00000895, 0.00010593, -0.00001908, 0.00006901, -0.00002367, 0.00002710, },
		{ 0.00004194, -0.00001655, 0.00003258, 0.00000176, 0.00002906, 0.00005146, -0.00003035, 0.00002012, -0.00007334, 0.00001511, -0.00003049, 0.00000180, -0.00003563, -0.00002475, -0.00005027, 0.00001373, -0.00002709, -0.00010745, -0.00005142, -0.00034555, -0.00042811, 0.00101443, 0.00017137, 0.00091291, 0.00011487, -0.00062141, 0.00002704, -0.00018938, 0.00001952, -0.00006430, -0.00000343, -0.00005640, 0.00001529, -0.00001843, 0.00003051, -0.00009145, 0.00005181, -0.00006024, 0.00002089, -0.00004675, 0.00000894, -0.00001859, -0.00000348, -0.00001667, -0.00000645, 0.00000593, 0.00000185, -0.00003377, -0.00003636, 0.00002949, -0.00001985, -0.00000770, -0.00002528, -0.00000526, -0.00006264, 0.00008254, -0.00007116, -0.00001790, -0.00002587, 0.00005563, 0.00005764, -0.00000265, 0.00005544, -0.00003515, },
		{ -0.00002517, 0.00004531, -0.00003211, 0.00005335, -0.00002570, 0.00006164, 0.00000507, -0.00000117, -0.00000791, -0.00005004, 0.00000159, -0.00001152, -0.00002228, -0.00001715, -0.00005715, -0.00000201, -0.00002347, -0.00001302, -0.00016603, 0.00002378, -0.00061187, -0.00004792, 0.00102874, 0.00027803, 0.00109474, -0.00070189, -0.00017594, -0.00012806, -0.00006533, -0.00003853, -0.00000714, -0.00007064, 0.00000656, -0.00001817, 0.00002122, -0.00002689, -0.00004763, 0.00003466, -0.00005111, 0.00001242, -0.00004355, 0.00001801, -0.00002831, 0.00000049, -0.00002114, 0.00000805, -0.00000522, 0.00000825, -0.00005401, -0.00000930, 0.00001776, -0.00001392, -0.00002417, 0.00000508, -0.00005781, 0.00001562, 0.00002140, -0.00005459, -0.00000454, -0.00002924, 0.00009317, 0.00000842, 0.00002778, 0.00002255, },
		{ 0.00001228, -0.00002102, 0.00004880, -0.00004829, 0.00007862, -0.00006193, 0.00010428, -0.00005676, 0.00005379, -0.00006419, 0.00000155, -0.00003777, 0.00003068, -0.00007318, 0.00002616, -0.00010745, 0.00007291, -0.00010556, 0.00003659, -0.00024086, 0.00009468, -0.00091978, 0.00054578, 0.00079039, 0.00080883, 0.00047854, -0.00093405, 0.00017169, -0.00032274, 0.00010931, -0.00014500, 0.00007186, -0.00015515, 0.00010566, -0.00009132, 0.00008321, -0.00010581, 0.00003265, -0.00002181, -0.00001096, -0.00002423, -0.00000787, -0.00000466, -0.00001489, -0.00001496, -0.00000091, -0.00000344, 0.00000450, -0.00000548, -0.00005371, 0.00000303, 0.00001904, -0.00003063, -0.00000347, -0.00001487, -0.00005471, 0.00006028, -0.00005479, 0.00002174, -0.00006384, 0.00004010, 0.00005612, 0.00001426, 0.00002081, },
		{ 0.00000459, 0.00000614, -0.00000536, 0.00002699, -0.00002936, 0.00006385, -0.00005755, 0.00010444, -0.00007563, 0.00007179, -0.00008640, 0.00001928, -0.00003747, 0.00002315, -0.00007146, 0.00001094, -0.00007936, 0.00005586, -0.00010247, -0.00001124, -0.00017841, -0.00005153, -0.00097378, 0.00101776, 0.00073502, 0.00117233, -0.00056610, -0.00042944, 0.00002627, -0.00020940, 0.00006566, -0.00009496, 0.00000660, -0.00009851, 0.00009216, -0.00008329, 0.00006834, -0.00011054, 0.00005896, -0.00004208, -0.00000396, -0.00002760, 0.00000247, -0.00001122, -0.00001815, -0.00001009, 0.00000475, -0.00000890, 0.00001098, -0.00001929, -0.00005014, 0.00001742, 0.00001211, -0.00004291, 0.00002728, -0.00007136, 0.00002538, -0.00000334, -0.00002511, 0.00002425, -0.00008059, 0.00009063, 0.00001180, 0.00003187, },
		{ 0.00000460, 0.00002347, -0.00002506, 0.00003705, -0.00002946, 0.00003480, -0.00000589, 0.00000638, 0.00004210, -0.00002909, 0.00002742, -0.00005065, -0.00001555, 0.00001608, -0.00004741, -0.00000319, -0.00007064, 0.00003191, -0.00007024, 0.00001320, -0.00017711, 0.00003443, -0.00046024, -0.00065021, 0.00122354, 0.00111790, 0.00068257, -0.00108651, 0.00013342, -0.00029905, 0.00006390, -0.00011506, 0.00005107, -0.00014476, 0.00006147, -0.00002446, 0.00000386, -0.00001635, -0.00004245, 0.00002315, -0.00001891, -0.00003209, 0.00000595, -0.00002184, 0.00000752, -0.00004334, 0.00002286, -0.00001951, 0.00001030, -0.00000302, -0.00001759, -0.00005742, 0.00004904, -0.00002678, -0.00000946, 0.00000175, -0.00007091, 0.00007652, -0.00009068, 0.00007796, -0.00006892, -0.00000130, 0.00005529, 0.00002769, },
		{ 0.00006220, -0.00003606, 0.00005090, -0.00005193, 0.00006601, -0.00006919, 0.00008520, -0.00006977, 0.00007034, -0.00002248, 0.00002165, -0.00001877, -0.00001913, -0.00003354, 0.00004128, -0.00008946, 0.00003331, -0.00010398, 0.00008684, -0.00015212, 0.00008100, -0.00027419, 0.00015822, -0.00085755, -0.00020768, 0.00153999, 0.00139027, -0.00042909, -0.00069757, 0.00014701, -0.00031091, 0.00012642, -0.00015030, 0.00006306, -0.00016911, 0.00013150, -0.00008940, 0.00004619, -0.00005840, -0.00000192, 0.00000690, -0.00002214, -0.00002726, 0.00001110, -0.00002471, 0.00000262, -0.00004123, 0.00003569, -0.00003400, 0.00002412, -0.00001298, -0.00002333, -0.00004981, 0.00006719, -0.00006520, 0.00004322, -0.00007305, 0.00002075, 0.00000673, -0.00006226, 0.00010317, -0.00012978, 0.00007973, -0.00000224, },
		{ 0.00000426, 0.00004297, -0.00002899, 0.00003823, -0.00003603, 0.00004480, -0.00005081, 0.00007334, -0.00007514, 0.00008304, -0.00004025, 0.00002910, -0.00002324, -0.00002992, 0.00000091, -0.00000217, -0.00005716, -0.00000646, -0.00004136, 0.00001817, -0.00010434, 0.00001133, -0.00020209, 0.00004562, -0.00105274, 0.00020536, 0.00226614, 0.00071196, -0.00107535, -0.00007843, -0.00015638, -0.00006938, -0.00001752, -0.00003419, -0.00006090, -0.00004712, 0.00006856, -0.00006004, 0.00001478, -0.00003180, -0.00001595, 0.00002960, -0.00006260, 0.00002213, -0.00002920, 0.00001079, -0.00003953, 0.00000526, 0.00000353, -0.00001136, 0.00000844, 0.00000128, -0.00005539, -0.00000294, 0.00003437, -0.00004740, 0.00003682, -0.00009619, 0.00009750, -0.00010870, 0.00006810, 0.00000151, -0.00007342, 0.00006867, },
		{ 0.00001234, 0.00004718, -0.00000955, 0.00000954, 0.00000411, -0.00000527, 0.00000965, -0.00001416, 0.00003679, -0.00005569, 0.00007917, -0.00005064, 0.00003760, -0.00003722, -0.00002020, 0.00001070, -0.00002820, -0.00004184, -0.00001589, -0.00001200, -0.00003540, -0.00006730, -0.00003704, -0.00015387, -0.00009959, -0.00127282, 0.00105217, 0.00275356, -0.00076335, -0.00065228, -0.00002463, -0.00017772, -0.00000466, -0.00004713, -0.00002409, -0.00008874, 0.00001890, 0.00002806, -0.00005770, 0.00002024, -0.00003794, 0.00000287, 0.00001057, -0.00005830, 0.00003631, -0.00004493, 0.00002257, -0.00005803, 0.00003881, -0.00002744, 0.00001350, -0.00000729, 0.00001003, -0.00008060, 0.00004934, -0.00002154, 0.00001134, -0.00003802, -0.00001295, 0.00004555, -0.00011162, 0.00013803, -0.00009937, 0.00001815, },
		{ 0.00003285, 0.00000659, 0.00004704, -0.00002734, 0.00002457, -0.00000749, -0.00000092, 0.00000467, -0.00000699, 0.00002226, -0.00005124, 0.00009269, -0.00008619, 0.00007913, -0.00009185, 0.00004400, -0.00004721, 0.00001334, -0.00009260, 0.00005857, -0.00009002, 0.00001726, -0.00013259, 0.00003146, -0.00025086, -0.00015507, -0.00162619, 0.00279248, 0.00181170, -0.00167329, 0.00010161, -0.00035632, 0.00008295, -0.00016016, 0.00008180, -0.00015507, 0.00002814, -0.00003582, 0.00006437, -0.00011543, 0.00008924, -0.00010416, 0.00007687, -0.00006867, 0.00001438, -0.00001622, -0.00000367, -0.00001924, -0.00001939, 0.00001935, -0.00001695, 0.00000580, 0.00000700, -0.00001780, -0.00005887, 0.00005376, -0.00003356, 0.00002755, -0.00008193, 0.00008089, -0.00008022, 0.00001271, 0.00006659, -0.00008622, },
		{ -0.00002128, -0.00000067, 0.00003657, 0.00001449, -0.00001841, 0.00002852, -0.00002052, 0.00001148, -0.00001036, 0.00001214, -0.00001379, -0.00000893, 0.00005730, -0.00006952, 0.00006650, -0.00008741, 0.00004942, -0.00005149, -0.00000583, -0.00006173, 0.00004238, -0.00008717, -0.00002081, -0.00007207, -0.00006658, -0.00015092, -0.00059695, -0.00132146, 0.00462887, -0.00062701, -0.00100563, 0.00006674, -0.00028159, 0.00007027, -0.00012189, 0.00003744, -0.00013491, 0.00004000, -0.00001544, 0.00001574, -0.00007960, 0.00007608, -0.00009951, 0.00008513, -0.00009197, 0.00004879, -0.00004479, 0.00001665, -0.00004237, 0.00001200, -0.00000282, -0.00000290, -0.00000199, 0.00001897, -0.00004974, -0.00001670, 0.00002491, -0.00000098, -0.00002543, -0.00002216, 0.00005567, -0.00011294, 0.00010414, -0.00003185, },
		{ -0.00002607, -0.00002502, 0.00001679, 0.00002003, 0.00001951, -0.00003814, 0.00006531, -0.00007685, 0.00007758, -0.00008547, 0.00009328, -0.00011778, 0.00011856, -0.00008229, 0.00006306, -0.00006960, 0.00005271, -0.00009212, 0.00009418, -0.00018367, 0.00015917, -0.00020068, 0.00015672, -0.00031879, 0.00030255, -0.00057159, 0.00045462, -0.00191919, 0.00078614, 0.00461422, -0.00276982, 0.00046088, -0.00069251, 0.00031454, -0.00035246, 0.00023220, -0.00028263, 0.00014332, -0.00016547, 0.00017270, -0.00018625, 0.00011397, -0.00008736, 0.00004159, -0.00003460, 0.00000573, -0.00002322, 0.00001439, -0.00004073, 0.00000933, -0.00002182, 0.00002607, -0.00003084, 0.00003029, -0.00001297, -0.00003643, -0.00001176, 0.00002788, -0.00000059, -0.00005979, 0.00006384, -0.00005927, -0.00000878, 0.00005552, },
		{ -0.00003389, 0.00003537, -0.00007631, 0.00007167, -0.00002952, 0.00004622, -0.00005973, 0.00009016, -0.00011581, 0.00012634, -0.00014344, 0.00015149, -0.00019044, 0.00021640, -0.00020180, 0.00018043, -0.00018688, 0.00017169, -0.00021187, 0.00020671, -0.00030882, 0.00032412, -0.00040305, 0.00035359, -0.00053931, 0.00057543, -0.00097941, 0.00091820, -0.00332635, 0.00436308, 0.00175256, -0.00264402, 0.00094654, -0.00098829, 0.00060942, -0.00056878, 0.00040784, -0.00045236, 0.00031762, -0.00029335, 0.00027509, -0.00030407, 0.00024700, -0.00021050, 0.00015126, -0.00013043, 0.00009031, -0.00008850, 0.00006382, -0.00008642, 0.00005650, -0.00005666, 0.00005444, -0.00005529, 0.00005961, -0.00005000, -0.00000645, -0.00002909, 0.00006326, -0.00006049, 0.00000058, 0.00004419, -0.00010047, 0.00008291, },
		{ 0.00003262, -0.00001293, 0.00001276, -0.00005399, 0.00005585, -0.00001316, 0.00000561, 0.00000205, 0.00001053, -0.00003130, 0.00003783, -0.00005448, 0.00004929, -0.00008088, 0.00011431, -0.00011448, 0.00008456, -0.00007366, 0.00003766, -0.00005745, 0.00001415, -0.00007538, 0.00007227, -0.00014897, 0.00003817, -0.00013869, 0.00006270, -0.00039694, -0.00005009, -0.00277054, 0.00688688, -0.00226428, -0.00044887, -0.00012069, -0.00018883, 0.00005305, -0.00011432, 0.00001144, -0.00011299, 0.00005669, -0.00005687, 0.00004564, -0.00010880, 0.00010104, -0.00009389, 0.00005056, -0.00003730, 0.00000856, -0.00001018, -0.00001713, -0.00000832, -0.00000741, 0.00000371, -0.00000118, -0.00000200, 0.00001485, -0.00001941, -0.00003392, 0.00000779, 0.00004675, -0.00009162, 0.00008640, -0.00006922, -0.00000147, },
		{ 0.00010342, -0.00008074, 0.00009189, -0.00009506, 0.00005134, -0.00004017, 0.00007422, -0.00009457, 0.00012218, -0.00013605, 0.00013373, -0.00014433, 0.00013588, -0.00016073, 0.00015213, -0.00012950, 0.00012205, -0.00015947, 0.00019269, -0.00026283, 0.00027497, -0.00036807, 0.00038150, -0.00045945, 0.00042776, -0.00063754, 0.00070354, -0.00105782, 0.00103839, -0.00255499, 0.00083466, 0.00575820, -0.00428421, 0.00148311, -0.00137612, 0.00083946, -0.00074027, 0.00053823, -0.00056186, 0.00040919, -0.00038056, 0.00032545, -0.00031521, 0.00022274, -0.00017737, 0.00013742, -0.00014697, 0.00014260, -0.00015124, 0.00013291, -0.00015332, 0.00012521, -0.00012615, 0.00011311, -0.00009944, 0.00008994, -0.00006881, 0.00005238, -0.00010417, 0.00009999, -0.00005310, -0.00001315, 0.00006437, -0.00011933, },
		{ -0.00002314, 0.00006533, -0.00008523, 0.00011812, -0.00014189, 0.00010422, -0.00008639, 0.00010487, -0.00012010, 0.00014906, -0.00017596, 0.00018493, -0.00020976, 0.00020347, -0.00023543, 0.00024228, -0.00022980, 0.00020696, -0.00022790, 0.00025406, -0.00033343, 0.00034887, -0.00044722, 0.00049162, -0.00062214, 0.00059934, -0.00082845, 0.00095660, -0.00150289, 0.00166291, -0.00446567, 0.00560226, 0.00127995, -0.00291687, 0.00129897, -0.00127499, 0.00086910, -0.00077367, 0.00055967, -0.00058716, 0.00047048, -0.00043625, 0.00036853, -0.00036843, 0.00029035, -0.00023288, 0.00016950, -0.00015688, 0.00014793, -0.00015230, 0.00012689, -0.00014583, 0.00012752, -0.00012988, 0.00012067, -0.00010530, 0.00009580, -0.00007198, 0.00004989, -0.00010382, 0.00013329, -0.00012917, 0.00010192, -0.00006249, },
		{ -0.00013756, 0.00010318, -0.00006954, 0.00003710, 0.00001040, -0.00005686, 0.00002894, -0.00000720, 0.00000866, -0.00000674, 0.00001746, -0.00003548, 0.00003580, -0.00006078, 0.00004448, -0.00006380, 0.00006572, -0.00005181, 0.00000216, 0.00002032, -0.00004312, -0.00000187, -0.00002399, -0.00001210, -0.00000485, -0.00009323, -0.00001478, -0.00007472, 0.00000456, -0.00038450, 0.00003053, -0.00318724, 0.00782624, -0.00318724, 0.00003053, -0.00038450, 0.00000456, -0.00007472, -0.00001478, -0.00009323, -0.00000485, -0.00001210, -0.00002399, -0.00000187, -0.00004312, 0.00002032, 0.00000216, -0.00005181, 0.00006572, -0.00006380, 0.00004448, -0.00006078, 0.00003580, -0.00003548, 0.00001746, -0.00000674, 0.00000866, -0.00000720, 0.00002894, -0.00005686, 0.00001040, 0.00003710, -0.00006954, 0.00010318, },
		{ -0.00002314, -0.00006249, 0.00010192, -0.00012917, 0.00013329, -0.00010382, 0.00004989, -0.00007198, 0.00009580, -0.00010530, 0.00012067, -0.00012988, 0.00012752, -0.00014583, 0.00012689, -0.00015230, 0.00014793, -0.00015688, 0.00016950, -0.00023288, 0.00029035, -0.00036843, 0.00036853, -0.00043625, 0.00047048, -0.00058716, 0.00055967, -0.00077367, 0.00086910, -0.00127499, 0.00129897, -0.00291687, 0.00127995, 0.00560226, -0.00446567, 0.00166291, -0.00150289, 0.00095660, -0.00082845, 0.00059934, -0.00062214, 0.00049162, -0.00044722, 0.00034887, -0.00033343, 0.00025406, -0.00022790, 0.00020696, -0.00022980, 0.00024228, -0.00023543, 0.00020347, -0.00020976, 0.00018493, -0.00017596, 0.00014906, -0.00012010, 0.00010487, -0.00008639, 0.00010422, -0.00014189, 0.00011812, -0.00008523, 0.00006533, },
		{ 0.00010342, -0.00011933, 0.00006437, -0.00001315, -0.00005310, 0.00009999, -0.00010417, 0.00005238, -0.00006881, 0.00008994, -0.00009944, 0.00011311, -0.00012615, 0.00012521, -0.00015332, 0.00013291, -0.00015124, 0.00014260, -0.00014697, 0.00013742, -0.00017737, 0.00022274, -0.00031521, 0.00032545, -0.00038056, 0.00040919, -0.00056186, 0.00053823, -0.00074027, 0.00083946, -0.00137612, 0.00148311, -0.00428421, 0.00575820, 0.00083466, -0.00255499, 0.00103839, -0.00105782, 0.00070354, -0.00063754, 0.00042776, -0.00045945, 0.00038150, -0.00036807, 0.00027497, -0.00026283, 0.00019269, -0.00015947, 0.00012205, -0.00012950, 0.00015213, -0.00016073, 0.00013588, -0.00014433, 0.00013373, -0.00013605, 0.00012218, -0.00009457, 0.00007422, -0.00004017, 0.00005134, -0.00009506, 0.00009189, -0.00008074, },
		{ 0.00003262, -0.00000147, -0.00006922, 0.00008640, -0.00009162, 0.00004675, 0.00000779, -0.00003392, -0.00001941, 0.00001485, -0.00000200, -0.00000118, 0.00000371, -0.00000741, -0.00000832, -0.00001713, -0.00001018, 0.00000856, -0.00003730, 0.00005056, -0.00009389, 0.00010104, -0.00010880, 0.00004564, -0.00005687, 0.00005669, -0.00011299, 0.00001144, -0.00011432, 0.00005305, -0.00018883, -0.00012069, -0.00044887, -0.00226428, 0.00688688, -0.00277054, -0.00005009, -0.00039694, 0.00006270, -0.00013869, 0.00003817, -0.00014897, 0.00007227, -0.00007538, 0.00001415, -0.00005745, 0.00003766, -0.00007366, 0.00008456, -0.00011448, 0.00011431, -0.00008088, 0.00004929, -0.00005448, 0.00003783, -0.00003130, 0.00001053, 0.00000205, 0.00000561, -0.00001316, 0.00005585, -0.00005399, 0.00001276, -0.00001293, },
		{ -0.00003389, 0.00008291, -0.00010047, 0.00004419, 0.00000058, -0.00006049, 0.00006326, -0.00002909, -0.00000645, -0.00005000, 0.00005961, -0.00005529, 0.00005444, -0.00005666, 0.00005650, -0.00008642, 0.00006382, -0.00008850, 0.00009031, -0.00013043, 0.00015126, -0.00021050, 0.00024700, -0.00030407, 0.00027509, -0.00029335, 0.00031762, -0.00045236, 0.00040784, -0.00056878, 0.00060942, -0.00098829, 0.00094654, -0.00264402, 0.00175256, 0.00436308, -0.00332635, 0.00091820, -0.00097941, 0.00057543, -0.00053931, 0.00035359, -0.00040305, 0.00032412, -0.00030882, 0.00020671, -0.00021187, 0.00017169, -0.00018688, 0.00018043, -0.00020180, 0.00021640, -0.00019044, 0.00015149, -0.00014344, 0.00012634, -0.00011581, 0.00009016, -0.00005973, 0.00004622, -0.00002952, 0.00007167, -0.00007631, 0.00003537, },
		{ -0.00002607, 0.00005552, -0.00000878, -0.00005927, 0.00006384, -0.00005979, -0.00000059, 0.00002788, -0.00001176, -0.00003643, -0.00001297, 0.00003029, -0.00003084, 0.00002607, -0.00002182, 0.00000933, -0.00004073, 0.00001439, -0.00002322, 0.00000573, -0.00003460, 0.00004159, -0.00008736, 0.00011397, -0.00018625, 0.00017270, -0.00016547, 0.00014332, -0.00028263, 0.00023220, -0.00035246, 0.00031454, -0.00069251, 0.00046088, -0.00276982, 0.00461422, 0.00078614, -0.00191919, 0.00045462, -0.00057159, 0.00030255, -0.00031879, 0.00015672, -0.00020068, 0.00015917, -0.00018367, 0.00009418, -0.00009212, 0.00005271, -0.00006960, 0.00006306, -0.00008229, 0.00011856, -0.00011778, 0.00009328, -0.00008547, 0.00007758, -0.00007685, 0.00006531, -0.00003814, 0.00001951, 0.00002003, 0.00001679, -0.00002502, },
		{ -0.00002128, -0.00003185, 0.00010414, -0.00011294, 0.00005567, -0.00002216, -0.00002543, -0.00000098, 0.00002491, -0.00001670, -0.00004974, 0.00001897, -0.00000199, -0.00000290, -0.00000282, 0.00001200, -0.00004237, 0.00001665, -0.00004479, 0.00004879, -0.00009197, 0.00008513, -0.00009951, 0.00007608, -0.00007960, 0.00001574, -0.00001544, 0.00004000, -0.00013491, 0.00003744, -0.00012189, 0.00007027, -0.00028159, 0.00006674, -0.00100563, -0.00062701, 0.00462887, -0.00132146, -0.00059695, -0.00015092, -0.00006658, -0.00007207, -0.00002081, -0.00008717, 0.00004238, -0.00006173, -0.00000583, -0.00005149, 0.00004942, -0.00008741, 0.00006650, -0.00006952, 0.00005730, -0.00000893, -0.00001379, 0.00001214, -0.00001036, 0.00001148, -0.00002052, 0.00002852, -0.00001841, 0.00001449, 0.00003657, -0.00000067, },
		{ 0.00003285, -0.00008622, 0.00006659, 0.00001271, -0.00008022, 0.00008089, -0.00008193, 0.00002755, -0.00003356, 0.00005376, -0.00005887, -0.00001780, 0.00000700, 0.00000580, -0.00001695, 0.00001935, -0.00001939, -0.00001924, -0.00000367, -0.00001622, 0.00001438, -0.00006867, 0.00007687, -0.00010416, 0.00008924, -0.00011543, 0.00006437, -0.00003582, 0.00002814, -0.00015507, 0.00008180, -0.00016016, 0.00008295, -0.00035632, 0.00010161, -0.00167329, 0.00181170, 0.00279248, -0.00162619, -0.00015507, -0.00025086, 0.00003146, -0.00013259, 0.00001726, -0.00009002, 0.00005857, -0.00009260, 0.00001334, -0.00004721, 0.00004400, -0.00009185, 0.00007913, -0.00008619, 0.00009269, -0.00005124, 0.00002226, -0.00000699, 0.00000467, -0.00000092, -0.00000749, 0.00002457, -0.00002734, 0.00004704, 0.00000659, },
		{ 0.00001234, 0.00001815, -0.00009937, 0.00013803, -0.00011162, 0.00004555, -0.00001295, -0.00003802, 0.00001134, -0.00002154, 0.00004934, -0.00008060, 0.00001003, -0.00000729, 0.00001350, -0.00002744, 0.00003881, -0.00005803, 0.00002257, -0.00004493, 0.00003631, -0.00005830, 0.00001057, 0.00000287, -0.00003794, 0.00002024, -0.00005770, 0.00002806, 0.00001890, -0.00008874, -0.00002409, -0.00004713, -0.00000466, -0.00017772, -0.00002463, -0.00065228, -0.00076335, 0.00275356, 0.00105217, -0.00127282, -0.00009959, -0.00015387, -0.00003704, -0.00006730, -0.00003540, -0.00001200, -0.00001589, -0.00004184, -0.00002820, 0.00001070, -0.00002020, -0.00003722, 0.00003760, -0.00005064, 0.00007917, -0.00005569, 0.00003679, -0.00001416, 0.00000965, -0.00000527, 0.00000411, 0.00000954, -0.00000955, 0.00004718, },
		{ 0.00000426, 0.00006867, -0.00007342, 0.00000151, 0.00006810, -0.00010870, 0.00009750, -0.00009619, 0.00003682, -0.00004740, 0.00003437, -0.00000294, -0.00005539, 0.00000128, 0.00000844, -0.00001136, 0.00000353, 0.00000526, -0.00003953, 0.00001079, -0.00002920, 0.00002213, -0.00006260, 0.00002960, -0.00001595, -0.00003180, 0.00001478, -0.00006004, 0.00006856, -0.00004712, -0.00006090, -0.00003419, -0.00001752, -0.00006938, -0.00015638, -0.00007843, -0.00107535, 0.00071196, 0.00226614, 0.00020536, -0.00105274, 0.00004562, -0.00020209, 0.00001133, -0.00010434, 0.00001817, -0.00004136, -0.00000646, -0.00005716, -0.00000217, 0.00000091, -0.00002992, -0.00002324, 0.00002910, -0.00004025, 0.00008304, -0.00007514, 0.00007334, -0.00005081, 0.00004480, -0.00003603, 0.00003823, -0.00002899, 0.00004297, },
		{ 0.00006220, -0.00000224, 0.00007973, -0.00012978, 0.00010317, -0.00006226, 0.00000673, 0.00002075, -0.00007305, 0.00004322, -0.00006520, 0.00006719, -0.00004981, -0.00002333, -0.00001298, 0.00002412, -0.00003400, 0.00003569, -0.00004123, 0.00000262, -0.00002471, 0.00001110, -0.00002726, -0.00002214, 0.00000690, -0.00000192, -0.00005840, 0.00004619, -0.00008940, 0.00013150, -0.00016911, 0.00006306, -0.00015030, 0.00012642, -0.00031091, 0.00014701, -0.00069757, -0.00042909, 0.00139027, 0.00153999, -0.00020768, -0.00085755, 0.00015822, -0.00027419, 0.00008100, -0.00015212, 0.00008684, -0.00010398, 0.00003331, -0.00008946, 0.00004128, -0.00003354, -0.00001913, -0.00001877, 0.00002165, -0.00002248, 0.00007034, -0.00006977, 0.00008520, -0.00006919, 0.00006601, -0.00005193, 0.00005090, -0.00003606, },
		{ 0.00000460, 0.00002769, 0.00005529, -0.00000130, -0.00006892, 0.00007796, -0.00009068, 0.00007652, -0.00007091, 0.00000175, -0.00000946, -0.00002678, 0.00004904, -0.00005742, -0.00001759, -0.00000302, 0.00001030, -0.00001951, 0.00002286, -0.00004334, 0.00000752, -0.00002184, 0.00000595, -0.00003209, -0.00001891, 0.00002315, -0.00004245, -0.00001635, 0.00000386, -0.00002446, 0.00006147, -0.00014476, 0.00005107, -0.00011506, 0.00006390, -0.00029905, 0.00013342, -0.00108651, 0.00068257, 0.00111790, 0.00122354, -0.00065021, -0.00046024, 0.00003443, -0.00017711, 0.00001320, -0.00007024, 0.00003191, -0.00007064, -0.00000319, -0.00004741, 0.00001608, -0.00001555, -0.00005065, 0.00002742, -0.00002909, 0.00004210, 0.00000638, -0.00000589, 0.00003480, -0.00002946, 0.00003705, -0.00002506, 0.00002347, },
		{ 0.00000459, 0.00003187, 0.00001180, 0.00009063, -0.00008059, 0.00002425, -0.00002511, -0.00000334, 0.00002538, -0.00007136, 0.00002728, -0.00004291, 0.00001211, 0.00001742, -0.00005014, -0.00001929, 0.00001098, -0.00000890, 0.00000475, -0.00001009, -0.00001815, -0.00001122, 0.00000247, -0.00002760, -0.00000396, -0.00004208, 0.00005896, -0.00011054, 0.00006834, -0.00008329, 0.00009216, -0.00009851, 0.00000660, -0.00009496, 0.00006566, -0.00020940, 0.00002627, -0.00042944, -0.00056610, 0.00117233, 0.00073502, 0.00101776, -0.00097378, -0.00005153, -0.00017841, -0.00001124, -0.00010247, 0.00005586, -0.00007936, 0.00001094, -0.00007146, 0.00002315, -0.00003747, 0.00001928, -0.00008640, 0.00007179, -0.00007563, 0.00010444, -0.00005755, 0.00006385, -0.00002936, 0.00002699, -0.00000536, 0.00000614, },
		{ 0.00001228, 0.00002081, 0.00001426, 0.00005612, 0.00004010, -0.00006384, 0.00002174, -0.00005479, 0.00006028, -0.00005471, -0.00001487, -0.00000347, -0.00003063, 0.00001904, 0.00000303, -0.00005371, -0.00000548, 0.00000450, -0.00000344, -0.00000091, -0.00001496, -0.00001489, -0.00000466, -0.00000787, -0.00002423, -0.00001096, -0.00002181, 0.00003265, -0.00010581, 0.00008321, -0.00009132, 0.00010566, -0.00015515, 0.00007186, -0.00014500, 0.00010931, -0.00032274, 0.00017169, -0.00093405, 0.00047854, 0.00080883, 0.00079039, 0.00054578, -0.00091978, 0.00009468, -0.00024086, 0.00003659, -0.00010556, 0.00007291, -0.00010745, 0.00002616, -0.00007318, 0.00003068, -0.00003777, 0.00000155, -0.00006419, 0.00005379, -0.00005676, 0.00010428, -0.00006193, 0.00007862, -0.00004829, 0.00004880, -0.00002102, },
		{ -0.00002517, 0.00002255, 0.00002778, 0.00000842, 0.00009317, -0.00002924, -0.00000454, -0.00005459, 0.00002140, 0.00001562, -0.00005781, 0.00000508, -0.00002417, -0.00001392, 0.00001776, -0.00000930, -0.00005401, 0.00000825, -0.00000522, 0.00000805, -0.00002114, 0.00000049, -0.00002831, 0.00001801, -0.00004355, 0.00001242, -0.00005111, 0.00003466, -0.00004763, -0.00002689, 0.00002122, -0.00001817, 0.00000656, -0.00007064, -0.00000714, -0.00003853, -0.00006533, -0.00012806, -0.00017594, -0.00070189, 0.00109474, 0.00027803, 0.00102874, -0.00004792, -0.00061187, 0.00002378, -0.00016603, -0.00001302, -0.00002347, -0.00000201, -0.00005715, -0.00001715, -0.00002228, -0.00001152, 0.00000159, -0.00005004, -0.00000791, -0.00000117, 0.00000507, 0.00006164, -0.00002570, 0.00005335, -0.00003211, 0.00004531, },
		{ 0.00004194, -0.00003515, 0.00005544, -0.00000265, 0.00005764, 0.00005563, -0.00002587, -0.00001790, -0.00007116, 0.00008254, -0.00006264, -0.00000526, -0.00002528, -0.00000770, -0.00001985, 0.00002949, -0.00003636, -0.00003377, 0.00000185, 0.00000593, -0.00000645, -0.00001667, -0.00000348, -0.00001859, 0.00000894, -0.00004675, 0.00002089, -0.00006024, 0.00005181, -0.00009145, 0.00003051, -0.00001843, 0.00001529, -0.00005640, -0.00000343, -0.00006430, 0.00001952, -0.00018938, 0.00002704, -0.00062141, 0.00011487, 0.00091291, 0.00017137, 0.00101443, -0.00042811, -0.00034555, -0.00005142, -0.00010745, -0.00002709, 0.00001373, -0.00005027, -0.00002475, -0.00003563, 0.00000180, -0.00003049, 0.00001511, -0.00007334, 0.00002012, -0.00003035, 0.00005146, 0.00002906, 0.00000176, 0.00003258, -0.00001655, },
		{ 0.00000045, 0.00002710, -0.00002367, 0.00006901, -0.00001908, 0.00010593, -0.00000895, 0.00001781, -0.00009716, 0.00001382, 0.00003949, -0.00006933, 0.00001379, -0.00004471, 0.00000869, -0.00002584, 0.00003630, -0.00006042, -0.00000939, -0.00000921, 0.00002171, -0.00003413, 0.00000766, -0.00002519, 0.00001019, -0.00003020, -0.00001030, -0.00001148, -0.00002607, 0.00001318, -0.00006980, 0.00003159, -0.00000681, -0.00002629, -0.00002146, -0.00003425, -0.00001510, -0.00007837, -0.00009010, -0.00017410, -0.00056263, 0.00075193, 0.00040465, 0.00039982, 0.00072862, -0.00054409, -0.00020432, -0.00007032, -0.00009353, 0.00000806, -0.00001370, -0.00004433, -0.00003074, -0.00001717, -0.00001214, -0.00001599, -0.00000384, -0.00006462, 0.00001585, -0.00002597, 0.00007443, 0.00001050, 0.00001800, 0.00001830, },
		{ -0.00000607, 0.00003043, -0.00000844, 0.00002524, 0.00003321, 0.00002648, 0.00008257, -0.00001369, -0.00000101, -0.00012264, 0.00009367, -0.00004767, -0.00001391, -0.00002206, -0.00002185, -0.00000387, -0.00000948, 0.00001833, -0.00005987, -0.00000328, -0.00000116, 0.00001166, -0.00003798, 0.00001524, -0.00002883, 0.00001547, -0.00004914, 0.00001525, -0.00003589, -0.00000278, -0.00001972, -0.00004229, 0.00003471, -0.00002068, -0.00004229, 0.00000588, -0.00005610, 0.00000568, -0.00015235, 0.00000499, -0.00049762, 0.00000507, 0.00077751, 0.00012897, 0.00058939, 0.00041169, -0.00055711, -0.00009737, -0.00010315, -0.00004984, 0.00000989, -0.00002945, -0.00004274, -0.00002465, -0.00001198, -0.00001749, -0.00000926, -0.00001721, -0.00006080, 0.00001452, -0.00001619, 0.00009580, -0.00001559, 0.00004391, },
		{ 0.00006695, -0.00002526, 0.00005074, -0.00002998, 0.00006730, -0.00000419, 0.00008889, 0.00001757, 0.00002627, -0.00008752, -0.00004355, 0.00007336, -0.00007338, 0.00001772, -0.00005563, 0.00001005, -0.00002611, 0.00001412, -0.00001116, -0.00004430, -0.00000681, 0.00001541, -0.00001705, -0.00001637, 0.00000058, -0.00001047, -0.00001016, -0.00003084, 0.00000754, -0.00003166, -0.00001240, -0.00001723, -0.00003840, 0.00005950, -0.00008498, 0.00001846, -0.00004402, 0.00000403, -0.00009550, -0.00005160, -0.00017385, -0.00045702, 0.00051543, 0.00046031, 0.00014111, 0.00062271, 0.00016013, -0.00053227, -0.00000545, -0.00014933, 0.00002200, -0.00003287, -0.00001271, -0.00005983, 0.00000109, -0.00002821, -0.00000588, -0.00001609, -0.00002241, -0.00006329, 0.00001849, 0.00000132, 0.00010494, -0.00003617, },
		{ -0.00002242, 0.00006087, -0.00001656, 0.00003950, -0.00000825, 0.00005999, 0.00001519, 0.00009191, -0.00000616, 0.00001967, -0.00013586, 0.00005001, -0.00000562, -0.00003113, -0.00000772, -0.00004653, 0.00001556, -0.00003107, 0.00002164, -0.00002926, -0.00003426, 0.00000050, 0.00001316, -0.00003115, -0.00000045, -0.00000783, -0.00000337, -0.00002837, -0.00001073, -0.00000271, -0.00003276, -0.00001455, -0.00001913, -0.00001583, 0.00004197, -0.00011209, 0.00006687, -0.00008580, 0.00004079, -0.00017972, 0.00005512, -0.00047192, 0.00000944, 0.00057074, 0.00024417, 0.00020166, 0.00060917, -0.00008589, -0.00042205, 0.00000733, -0.00013551, 0.00004500, -0.00005344, -0.00001176, -0.00005433, 0.00000528, -0.00002867, -0.00000872, -0.00000880, -0.00004623, -0.00004625, 0.00000727, 0.00004583, 0.00007634, },
		{ 0.00002397, 0.00001725, 0.00003411, 0.00000631, 0.00001982, 0.00002713, 0.00003368, 0.00006696, 0.00004187, 0.00001689, -0.00005257, -0.00008612, 0.00006039, -0.00004324, 0.00000311, -0.00004422, -0.00001595, 0.00000177, -0.00002255, 0.00001579, -0.00003647, -0.00002760, 0.00001314, -0.00000648, -0.00002278, -0.00000139, -0.00000133, -0.00001872, -0.00001985, -0.00001081, 0.00000151, -0.00005459, 0.00001153, -0.00004554, 0.00003889, -0.00004251, -0.00004671, 0.00002810, -0.00004224, -0.00004069, -0.00010143, -0.00007758, -0.00047690, 0.00045975, 0.00033669, 0.00021467, 0.00022022, 0.00057972, -0.00031799, -0.00024774, -0.00005080, -0.00005982, 0.00000402, -0.00003012, -0.00003978, -0.00001993, -0.00001857, -0.00000772, -0.00002765, 0.00000938, -0.00008330, -0.00001517, -0.00000760, 0.00009982, },
		{ 0.00012225, -0.00001112, 0.00004898, 0.00001420, 0.00001799, 0.00002305, 0.00003357, 0.00004483, 0.00008064, 0.00000779, 0.00001571, -0.00010624, -0.00000573, 0.00000809, -0.00001868, -0.00001184, -0.00004782, 0.00000147, -0.00000912, -0.00001675, 0.00001099, -0.00004462, -0.00001101, 0.00000975, -0.00001711, -0.00001452, -0.00000016, -0.00000328, -0.00002860, -0.00000925, -0.00001029, -0.00000614, -0.00006161, 0.00002808, -0.00005392, 0.00006253, -0.00011316, 0.00003744, -0.00003531, 0.00001109, -0.00013986, 0.00002204, -0.00035975, -0.00008477, 0.00052530, 0.00017909, 0.00019712, 0.00028067, 0.00045390, -0.00043866, -0.00011270, -0.00009086, -0.00000207, -0.00003583, -0.00000980, -0.00005536, 0.00000172, -0.00003261, 0.00000156, -0.00003153, 0.00000591, -0.00010037, -0.00000013, 0.00000688, },
		{ 0.00005136, 0.00010376, -0.00001381, 0.00006093, 0.00000484, 0.00002547, 0.00003535, 0.00002572, 0.00008637, 0.00004310, 0.00001054, -0.00003557, -0.00008154, 0.00001528, -0.00002081, 0.00000703, -0.00004355, -0.00002724, 0.00000279, -0.00001498, -0.00001314, 0.00000614, -0.00004960, 0.00000898, -0.00000893, -0.00000997, -0.00001740, 0.00001084, -0.00002345, -0.00001561, -0.00001580, 0.00000498, -0.00003995, -0.00003106, 0.00000889, -0.00001917, 0.00001389, -0.00009766, 0.00005274, -0.00003844, -0.00002380, -0.00011464, -0.00003449, -0.00043902, 0.00035637, 0.00030408, 0.00019250, 0.00012150, 0.00039856, 0.00022423, -0.00042011, -0.00006050, -0.00007437, -0.00000205, -0.00003573, -0.00001937, -0.00003940, -0.00000749, -0.00002096, -0.00001201, -0.00001362, -0.00002527, -0.00009075, -0.00000291, },
		{ 0.00000445, 0.00009309, 0.00007051, 0.00000062, 0.00006285, -0.00000404, 0.00004687, 0.00002604, 0.00004923, 0.00010052, 0.00000304, 0.00000519, -0.00007770, -0.00002432, -0.00001801, 0.00000147, -0.00000860, -0.00005013, -0.00001133, 0.00000192, -0.00002422, -0.00000279, -0.00000779, -0.00003897, 0.00001213, -0.00001955, -0.00000317, -0.00001699, 0.00001264, -0.00003850, 0.00000035, -0.00002286, 0.00001214, -0.00006875, 0.00000625, -0.00001417, 0.00001252, -0.00005261, -0.00003511, 0.00002294, -0.00001516, -0.00009366, -0.00002515, -0.00024690, -0.00016760, 0.00049116, 0.00011860, 0.00022393, 0.00008066, 0.00047191, -0.00000701, -0.00033878, -0.00004156, -0.00004752, -0.00002092, -0.00002264, -0.00003292, -0.00002007, -0.00002141, -0.00000720, -0.00002478, -0.00000002, -0.00005780, -0.00007820, },
		{ -0.00007260, 0.00003411, 0.00011123, 0.00004326, 0.00002075, 0.00005550, -0.00000311, 0.00006787, 0.00000786, 0.00010346, 0.00006354, -0.00001263, -0.00002986, -0.00006226, -0.00000884, -0.00003654, 0.00002779, -0.00004381, -0.00003188, -0.00000755, 0.00000073, -0.00003335, 0.00001034, -0.00002577, -0.00001769, -0.00000355, -0.00001138, -0.00000862, -0.00000510, -0.00000547, -0.00002975, -0.00000133, -0.00001391, -0.00000759, -0.00006385, 0.00001806, -0.00001013, -0.00000407, -0.00006674, -0.00000255, 0.00001620, -0.00004262, -0.00008394, -0.00005236, -0.00034158, 0.00021051, 0.00033820, 0.00009965, 0.00017771, 0.00013073, 0.00043720, -0.00016329, -0.00025821, -0.00001666, -0.00004769, -0.00002147, -0.00002377, -0.00002902, -0.00001865, -0.00002044, -0.00000729, -0.00002356, -0.00000454, -0.00007759, },
		{ -0.00009549, -0.00005866, 0.00007016, 0.00011159, 0.00002436, 0.00004488, 0.00003513, 0.00002214, 0.00006055, 0.00002341, 0.00013835, 0.00000550, -0.00002391, -0.00005613, -0.00002126, -0.00003690, -0.00001154, 0.00001701, -0.00005940, -0.00001289, -0.00000686, -0.00000723, -0.00003209, 0.00001445, -0.00003349, -0.00000536, -0.00001764, 0.00000126, -0.00001577, 0.00000346, -0.00002523, -0.00001088, -0.00001115, -0.00000313, -0.00003495, -0.00004054, 0.00001802, -0.00000364, -0.00004021, -0.00004048, -0.00000105, 0.00002236, -0.00009446, -0.00002170, -0.00020721, -0.00016110, 0.00038130, 0.00015487, 0.00014366, 0.00011014, 0.00021440, 0.00033334, -0.00025381, -0.00017540, -0.00000542, -0.00005510, -0.00001471, -0.00002779, -0.00002150, -0.00002271, -0.00001525, -0.00001017, -0.00002131, -0.00001474, },
		{ -0.00002263, -0.00011676, -0.00002686, 0.00009789, 0.00010281, 0.00001612, 0.00006336, 0.00001640, 0.00005693, 0.00003375, 0.00008130, 0.00011836, -0.00004377, -0.00004132, -0.00004266, -0.00001105, -0.00005516, 0.00002192, -0.00001914, -0.00004724, -0.00000560, -0.00000566, -0.00002010, -0.00002044, 0.00000840, -0.00002920, -0.00000940, -0.00001578, 0.00000405, -0.00001447, -0.00000380, -0.00002668, -0.00000293, -0.00001271, -0.00000852, -0.00004439, -0.00002450, 0.00002804, -0.00002822, -0.00004239, -0.00002921, 0.00001452, -0.00001187, -0.00008575, -0.00004338, -0.00028772, 0.00013645, 0.00030600, 0.00009009, 0.00014781, 0.00008617, 0.00026882, 0.00021212, -0.00030037, -0.00008511, -0.00002610, -0.00004483, -0.00001861, -0.00002071, -0.00002474, -0.00001913, -0.00001599, -0.00000780, -0.00002652, },
		{ -0.00002970, -0.00003744, -0.00012887, 0.00001329, 0.00011790, 0.00008599, 0.00002322, 0.00006488, 0.00001957, 0.00007074, 0.00002862, 0.00014140, 0.00005080, -0.00007793, -0.00004306, -0.00001485, -0.00003425, -0.00003707, 0.00002574, -0.00004585, -0.00002522, -0.00000659, -0.00000749, -0.00002901, -0.00000758, 0.00000135, -0.00002549, -0.00001684, -0.00000523, 0.00000070, -0.00001260, -0.00001605, -0.00001533, -0.00000496, -0.00001095, -0.00002117, -0.00004101, -0.00001041, 0.00003107, -0.00006332, -0.00001559, -0.00003849, 0.00003915, -0.00006955, -0.00003033, -0.00017020, -0.00015247, 0.00029616, 0.00016765, 0.00009785, 0.00012253, 0.00009440, 0.00029258, 0.00008145, -0.00029181, -0.00001920, -0.00005487, -0.00002716, -0.00002437, -0.00001288, -0.00002994, -0.00001512, -0.00001487, -0.00000982, },
		{ -0.00002264, -0.00002434, -0.00006495, -0.00012210, 0.00005158, 0.00013238, 0.00006547, 0.00004015, 0.00005547, 0.00004108, 0.00005972, 0.00006829, 0.00015620, -0.00003734, -0.00008848, -0.00002026, -0.00001109, -0.00005166, -0.00000633, 0.00000153, -0.00004673, -0.00001054, -0.00000985, -0.00001342, -0.00002962, 0.00000121, -0.00000277, -0.00002986, -0.00001437, 0.00000126, -0.00000058, -0.00002041, -0.00001614, -0.00000835, -0.00000552, -0.00001770, -0.00002373, -0.00004101, 0.00001527, 0.00000081, -0.00006397, -0.00000551, -0.00003205, 0.00002995, -0.00008335, -0.00003378, -0.00025027, 0.00009468, 0.00025677, 0.00010155, 0.00009963, 0.00010600, 0.00010771, 0.00029553, -0.00005142, -0.00022720, -0.00000262, -0.00006320, -0.00002013, -0.00001949, -0.00001635, -0.00002666, -0.00001810, -0.00000625, },
		{ 0.00000081, -0.00003733, -0.00001862, -0.00009619, -0.00010145, 0.00008903, 0.00013721, 0.00005031, 0.00005534, 0.00005119, 0.00005872, 0.00005255, 0.00013306, 0.00010260, -0.00011232, -0.00006482, 0.00000128, -0.00003012, -0.00004186, 0.00000499, -0.00002364, -0.00003065, -0.00000775, -0.00001313, -0.00001986, -0.00002446, 0.00000761, -0.00000982, -0.00003473, -0.00000445, 0.00000279, -0.00000340, -0.00003143, -0.00000486, -0.00000936, -0.00000543, -0.00002754, -0.00001957, -0.00003901, 0.00003639, -0.00004609, -0.00002965, -0.00002406, -0.00000517, -0.00000966, -0.00005346, -0.00012446, -0.00015914, 0.00024864, 0.00014667, 0.00009480, 0.00008886, 0.00009457, 0.00013179, 0.00026130, -0.00015036, -0.00014531, -0.00001815, -0.00005382, -0.00002020, -0.00001066, -0.00002491, -0.00002038, -0.00002119, },
		{ -0.00001654, -0.00000320, -0.00004377, -0.00002189, -0.00012072, -0.00007517, 0.00012985, 0.00012813, 0.00004620, 0.00006413, 0.00005776, 0.00006075, 0.00007966, 0.00017367, -0.00000761, -0.00013669, -0.00002378, 0.00000021, -0.00004521, -0.00001841, -0.00000968, -0.00002801, -0.00001664, -0.00001103, -0.00001733, -0.00002249, -0.00001753, 0.00001271, -0.00002424, -0.00002918, 0.00000236, 0.00000510, -0.00001564, -0.00003100, 0.00000398, -0.00001096, -0.00001218, -0.00002760, -0.00002172, -0.00002185, 0.00002613, -0.00006653, -0.00000132, -0.00004207, 0.00001163, -0.00003347, -0.00005402, -0.00019393, 0.00004034, 0.00023760, 0.00008302, 0.00009125, 0.00008501, 0.00007701, 0.00017123, 0.00018241, -0.00018862, -0.00008744, -0.00003397, -0.00004380, -0.00001571, -0.00000982, -0.00002765, -0.00001964, },
		{ -0.00002264, -0.00000755, -0.00001610, -0.00004250, -0.00003320, -0.00013877, -0.00004225, 0.00017169, 0.00010585, 0.00005429, 0.00006809, 0.00006879, 0.00005814, 0.00014497, 0.00013683, -0.00011889, -0.00010222, 0.00000154, -0.00001491, -0.00004212, -0.00000548, -0.00003216, -0.00001145, -0.00001665, -0.00001288, -0.00002326, -0.00001955, -0.00001057, 0.00001251, -0.00003989, -0.00001348, 0.00000257, 0.00000711, -0.00003390, -0.00001610, 0.00000321, -0.00001116, -0.00002268, -0.00001932, -0.00002748, 0.00000068, -0.00000950, -0.00004770, -0.00000346, -0.00004133, 0.00000932, -0.00003396, -0.00011244, -0.00013908, 0.00018810, 0.00014738, 0.00007296, 0.00007906, 0.00008263, 0.00006729, 0.00020247, 0.00008653, -0.00018034, -0.00006006, -0.00003897, -0.00003766, -0.00000710, -0.00001654, -0.00002462, },
		},
		{ { 0.00035497, 0.00011617, -0.00010470, 0.00000490, 0.00003082, 0.00003813, 0.00023966, 0.00001024, -0.00021387, -0.00003711, -0.00004279, -0.00002942, -0.00003172, -0.00001564, -0.00000780, 0.00001084, -0.00004259, 0.00001065, -0.00006016, 0.00002209, -0.00001835, 0.00002171, -0.00003706, 0.00002824, -0.00001622, 0.00001689, -0.00002452, 0.00003093, -0.00003415, 0.00002147, -0.00000839, 0.00002455, -0.00003846, 0.00002455, -0.00000839, 0.00002147, -0.00003415, 0.00003093, -0.00002452, 0.00001689, -0.00001622, 0.00002824, -0.00003706, 0.00002171, -0.00001835, 0.00002209, -0.00006016, 0.00001065, -0.00004259, 0.00001084, -0.00000780, -0.00001564, -0.00003172, -0.00002942, -0.00004279, -0.00003711, -0.00021387, 0.00001024, 0.00023966, 0.00003813, 0.00003082, 0.00000490, -0.00010470, 0.00011617, },
		{ 0.00010729, 0.00035723, 0.00012132, -0.00008179, 0.00001159, 0.00003453, 0.00006520, 0.00026034, -0.00006695, -0.00020835, -0.00002138, -0.00004870, -0.00002378, -0.00003439, -0.00001344, -0.00001016, 0.00001230, -0.00003770, -0.00000051, -0.00005836, 0.00003368, -0.00002738, 0.00002505, -0.00004403, 0.00004068, -0.00002152, 0.00001682, -0.00002499, 0.00003267, -0.00003834, 0.00002842, -0.00001274, 0.00002263, -0.00003904, 0.00002899, -0.00000710, 0.00001544, -0.00002939, 0.00002860, -0.00002437, 0.00001923, -0.00001472, 0.00001775, -0.00003021, 0.00001985, -0.00001028, 0.00000664, -0.00005490, 0.00001751, -0.00004729, 0.00001077, -0.00000682, -0.00001680, -0.00002970, -0.00003492, -0.00003496, -0.00006361, -0.00019232, 0.00007063, 0.00020535, 0.00002273, 0.00002521, -0.00000035, -0.00012511, },
		{ -0.00014407, 0.00009504, 0.00036334, 0.00012376, -0.00005646, 0.00001939, 0.00003703, 0.00010657, 0.00025504, -0.00014738, -0.00017829, -0.00002146, -0.00004708, -0.00002254, -0.00003406, -0.00001310, -0.00001175, 0.00001417, -0.00003312, -0.00001509, -0.00004841, 0.00003777, -0.00003324, 0.00002689, -0.00004869, 0.00005273, -0.00002932, 0.00001905, -0.00002622, 0.00003394, -0.00004143, 0.00003503, -0.00001893, 0.00002276, -0.00004050, 0.00003583, -0.00000972, 0.00001172, -0.00002495, 0.00002580, -0.00002393, 0.00002301, -0.00001704, 0.00001031, -0.00002464, 0.00002119, -0.00000673, -0.00000829, -0.00004518, 0.00002032, -0.00005145, 0.00001261, -0.00000880, -0.00001438, -0.00003205, -0.00003490, -0.00003320, -0.00009024, -0.00015015, 0.00010883, 0.00016672, 0.00001604, 0.00001720, -0.00000323, },
		{ -0.00000244, -0.00016323, 0.00008058, 0.00037183, 0.00012533, -0.00002941, 0.00002789, 0.00004054, 0.00016164, 0.00021242, -0.00021222, -0.00013522, -0.00003454, -0.00003769, -0.00002660, -0.00002964, -0.00001577, -0.00001158, 0.00001615, -0.00003022, -0.00003042, -0.00003193, 0.00003323, -0.00003311, 0.00002458, -0.00004851, 0.00006147, -0.00003736, 0.00002265, -0.00002811, 0.00003488, -0.00004321, 0.00004031, -0.00002567, 0.00002407, -0.00004199, 0.00004416, -0.00001636, 0.00001145, -0.00002197, 0.00002303, -0.00002278, 0.00002700, -0.00002222, 0.00000576, -0.00002001, 0.00002541, -0.00000928, -0.00001928, -0.00003378, 0.00001959, -0.00005413, 0.00001542, -0.00001332, -0.00000816, -0.00003987, -0.00002670, -0.00004349, -0.00010569, -0.00009911, 0.00012672, 0.00012884, 0.00001593, 0.00000618, },
		{ -0.00000794, 0.00000260, -0.00018368, 0.00006464, 0.00038171, 0.00012748, -0.00000084, 0.00003563, 0.00005030, 0.00022334, 0.00012783, -0.00024352, -0.00009559, -0.00005149, -0.00002565, -0.00003258, -0.00002347, -0.00001988, -0.00001059, 0.00001918, -0.00003135, -0.00004213, -0.00001346, 0.00002232, -0.00002695, 0.00001731, -0.00004211, 0.00006452, -0.00004315, 0.00002611, -0.00003007, 0.00003537, -0.00004362, 0.00004361, -0.00003178, 0.00002554, -0.00004225, 0.00005233, -0.00002610, 0.00001505, -0.00002145, 0.00002110, -0.00002093, 0.00003013, -0.00002894, 0.00000324, -0.00001499, 0.00003038, -0.00001708, -0.00002475, -0.00002276, 0.00001564, -0.00005381, 0.00001681, -0.00001769, -0.00000100, -0.00005037, -0.00001253, -0.00006618, -0.00010303, -0.00005090, 0.00012993, 0.00009379, 0.00002078, },
		{ 0.00002838, -0.00002383, 0.00001083, -0.00020484, 0.00004640, 0.00039347, 0.00013038, 0.00003018, 0.00003982, 0.00007466, 0.00027643, 0.00000826, -0.00023144, -0.00007419, -0.00006143, -0.00001853, -0.00003474, -0.00002009, -0.00002195, -0.00001114, 0.00002487, -0.00003892, -0.00004543, 0.00000075, 0.00001025, -0.00001785, 0.00000687, -0.00003033, 0.00006113, -0.00004494, 0.00002791, -0.00003122, 0.00003508, -0.00004264, 0.00004458, -0.00003636, 0.00002626, -0.00004002, 0.00005811, -0.00003697, 0.00002185, -0.00002374, 0.00002077, -0.00001885, 0.00003184, -0.00003611, 0.00000210, -0.00000827, 0.00003326, -0.00002754, -0.00002495, -0.00001338, 0.00000889, -0.00004915, 0.00001408, -0.00001820, 0.00000242, -0.00005781, 0.00000090, -0.00009512, -0.00008266, -0.00001274, 0.00012392, 0.00006277, },
		{ 0.00003748, 0.00003560, -0.00003872, 0.00001949, -0.00022444, 0.00002348, 0.00040912, 0.00013305, 0.00006510, 0.00003789, 0.00012266, 0.00029947, -0.00012637, -0.00017927, -0.00007772, -0.00005736, -0.00002191, -0.00002838, -0.00002349, -0.00001873, -0.00001537, 0.00003408, -0.00005384, -0.00003717, 0.00000495, 0.00000326, -0.00001084, -0.00000274, -0.00001625, 0.00005264, -0.00004243, 0.00002722, -0.00003089, 0.00003367, -0.00004028, 0.00004315, -0.00003892, 0.00002574, -0.00003442, 0.00005932, -0.00004626, 0.00003008, -0.00002829, 0.00002238, -0.00001715, 0.00003213, -0.00004325, 0.00000242, 0.00000031, 0.00003188, -0.00003767, -0.00002108, -0.00000655, 0.00000049, -0.00004017, 0.00000595, -0.00001236, -0.00000178, -0.00005649, 0.00000540, -0.00012000, -0.00005167, 0.00001417, 0.00011174, },
		{ 0.00009407, 0.00001993, 0.00003913, -0.00004944, 0.00002533, -0.00023971, -0.00000681, 0.00043086, 0.00013477, 0.00010412, 0.00003032, 0.00019903, 0.00027075, -0.00024703, -0.00010362, -0.00010180, -0.00003991, -0.00003592, -0.00001298, -0.00003448, -0.00000943, -0.00002346, 0.00004557, -0.00007403, -0.00001782, -0.00000350, 0.00000600, -0.00001081, -0.00000682, -0.00000433, 0.00004217, -0.00003697, 0.00002436, -0.00002902, 0.00003107, -0.00003678, 0.00003962, -0.00003950, 0.00002420, -0.00002548, 0.00005449, -0.00005124, 0.00003725, -0.00003360, 0.00002552, -0.00001625, 0.00003126, -0.00005030, 0.00000508, 0.00000943, 0.00002579, -0.00004527, -0.00001440, -0.00000331, -0.00000722, -0.00002906, -0.00000606, -0.00000089, -0.00001401, -0.00004422, -0.00000440, -0.00013106, -0.00002025, 0.00003314, },
		{ 0.00004843, 0.00007112, 0.00001111, 0.00003706, -0.00005423, 0.00002645, -0.00024927, -0.00004565, 0.00045924, 0.00013691, 0.00014417, 0.00002383, 0.00029844, 0.00017707, -0.00032421, -0.00002922, -0.00013270, -0.00001739, -0.00005448, 0.00000682, -0.00004959, 0.00000334, -0.00003292, 0.00005561, -0.00009420, 0.00000782, -0.00002259, 0.00001935, -0.00002016, -0.00000201, 0.00000125, 0.00003354, -0.00003104, 0.00002068, -0.00002637, 0.00002776, -0.00003270, 0.00003471, -0.00003868, 0.00002260, -0.00001450, 0.00004362, -0.00005009, 0.00004097, -0.00003767, 0.00002909, -0.00001606, 0.00002933, -0.00005717, 0.00001117, 0.00001658, 0.00001642, -0.00004958, -0.00000578, -0.00000486, -0.00001109, -0.00001989, -0.00001760, 0.00001176, -0.00002999, -0.00002443, -0.00002763, -0.00012370, 0.00000328, },
		{ 0.00001605, 0.00006196, 0.00004477, 0.00000951, 0.00003017, -0.00005404, 0.00002369, -0.00025443, -0.00009143, 0.00049208, 0.00014379, 0.00017836, 0.00003226, 0.00040179, 0.00002181, -0.00033844, 0.00002024, -0.00015355, -0.00000187, -0.00006792, 0.00002306, -0.00006233, 0.00001445, -0.00003946, 0.00005915, -0.00010719, 0.00003162, -0.00004582, 0.00003963, -0.00003746, 0.00001208, -0.00000176, 0.00002987, -0.00002730, 0.00001810, -0.00002435, 0.00002474, -0.00002902, 0.00002949, -0.00003746, 0.00002256, -0.00000403, 0.00002853, -0.00004267, 0.00003977, -0.00003872, 0.00003168, -0.00001600, 0.00002606, -0.00006318, 0.00002112, 0.00001919, 0.00000633, -0.00005102, 0.00000418, -0.00001188, -0.00000835, -0.00001704, -0.00002312, 0.00001895, -0.00004200, -0.00000563, -0.00005652, -0.00010095, },
		{ -0.00007227, 0.00002136, 0.00007193, 0.00001935, 0.00001100, 0.00002200, -0.00005249, 0.00002043, -0.00025898, -0.00014010, 0.00052506, 0.00016143, 0.00019820, 0.00007372, 0.00047739, -0.00017138, -0.00028855, 0.00003146, -0.00015220, -0.00000308, -0.00006787, 0.00002834, -0.00006639, 0.00001890, -0.00003934, 0.00005226, -0.00010676, 0.00004497, -0.00006440, 0.00005972, -0.00005762, 0.00003237, -0.00001273, 0.00003238, -0.00002753, 0.00001832, -0.00002447, 0.00002329, -0.00002686, 0.00002515, -0.00003690, 0.00002578, 0.00000269, 0.00001268, -0.00003085, 0.00003386, -0.00003594, 0.00003213, -0.00001536, 0.00002079, -0.00006684, 0.00003392, 0.00001581, -0.00000198, -0.00005037, 0.00001449, -0.00002360, 0.00000180, -0.00002301, -0.00001857, 0.00001496, -0.00004226, 0.00000198, -0.00007916, },
		{ -0.00008458, -0.00004931, 0.00002612, 0.00007420, 0.00000040, 0.00001026, 0.00001711, -0.00005409, 0.00002080, -0.00026717, -0.00018743, 0.00055428, 0.00019430, 0.00019852, 0.00016342, 0.00048841, -0.00036201, -0.00019401, 0.00000719, -0.00012725, -0.00002295, -0.00005191, 0.00001995, -0.00005914, 0.00001473, -0.00003190, 0.00003476, -0.00009070, 0.00004233, -0.00007072, 0.00007191, -0.00007373, 0.00005326, -0.00002828, 0.00003999, -0.00003189, 0.00002202, -0.00002772, 0.00002440, -0.00002718, 0.00002264, -0.00003763, 0.00003331, 0.00000270, 0.00000028, -0.00001806, 0.00002516, -0.00002992, 0.00003014, -0.00001374, 0.00001294, -0.00006613, 0.00004700, 0.00000695, -0.00000716, -0.00004800, 0.00002330, -0.00003722, 0.00001708, -0.00003661, -0.00000397, -0.00000174, -0.00002704, -0.00000859, },
		{ -0.00003715, -0.00006799, -0.00004046, 0.00003647, 0.00006534, -0.00000800, 0.00000349, 0.00001843, -0.00006148, 0.00002682, -0.00028081, -0.00023202, 0.00057957, 0.00024137, 0.00018323, 0.00030424, 0.00040517, -0.00050349, -0.00008943, -0.00003467, -0.00008936, -0.00005376, -0.00002573, 0.00000198, -0.00004360, 0.00000471, -0.00002095, 0.00001153, -0.00006273, 0.00002403, -0.00006166, 0.00007109, -0.00007983, 0.00006868, -0.00004354, 0.00004972, -0.00003887, 0.00002852, -0.00003404, 0.00002835, -0.00003033, 0.00002227, -0.00003941, 0.00004483, -0.00000556, -0.00000499, -0.00000827, 0.00001676, -0.00002258, 0.00002652, -0.00001156, 0.00000264, -0.00005931, 0.00005686, -0.00000503, -0.00000941, -0.00004350, 0.00002803, -0.00004828, 0.00003230, -0.00005260, 0.00001565, -0.00002668, 0.00000054, },
		{ 0.00003069, -0.00007509, -0.00003356, -0.00004683, 0.00005382, 0.00004576, -0.00000565, -0.00000935, 0.00002509, -0.00007332, 0.00003626, -0.00029714, -0.00027742, 0.00060644, 0.00029403, 0.00016903, 0.00047881, 0.00021841, -0.00055818, -0.00001267, -0.00006842, -0.00005690, -0.00008084, -0.00000135, -0.00001579, -0.00002771, -0.00000433, -0.00001386, -0.00000835, -0.00003214, -0.00000312, -0.00004034, 0.00005715, -0.00007344, 0.00007442, -0.00005394, 0.00005796, -0.00004585, 0.00003593, -0.00004222, 0.00003444, -0.00003583, 0.00002360, -0.00004099, 0.00005827, -0.00002134, -0.00000136, -0.00000462, 0.00001182, -0.00001650, 0.00002293, -0.00001015, -0.00000880, -0.00004585, 0.00006023, -0.00001674, -0.00001006, -0.00003619, 0.00002631, -0.00005209, 0.00004105, -0.00006341, 0.00003171, -0.00005038, },
		{ -0.00006198, 0.00005034, -0.00010811, 0.00000663, -0.00006165, 0.00007357, 0.00002074, 0.00000300, -0.00002403, 0.00003214, -0.00008413, 0.00004273, -0.00030905, -0.00033158, 0.00064503, 0.00033788, 0.00018366, 0.00064753, -0.00005177, -0.00051136, 0.00000912, -0.00007160, -0.00004757, -0.00008905, 0.00000813, -0.00002215, -0.00002083, -0.00000455, -0.00001856, -0.00001485, -0.00001075, -0.00002790, -0.00001530, 0.00003539, -0.00005673, 0.00006975, -0.00005698, 0.00006183, -0.00005020, 0.00004181, -0.00005019, 0.00004108, -0.00004242, 0.00002555, -0.00004038, 0.00007007, -0.00004139, 0.00001026, -0.00000817, 0.00001235, -0.00001389, 0.00002123, -0.00001140, -0.00001883, -0.00002719, 0.00005544, -0.00002521, -0.00001049, -0.00002606, 0.00001733, -0.00004581, 0.00003843, -0.00006231, 0.00003549, },
		{ 0.00002245, -0.00005423, 0.00004884, -0.00012191, 0.00003750, -0.00007358, 0.00008709, -0.00000121, 0.00001064, -0.00003401, 0.00003268, -0.00008674, 0.00003834, -0.00030814, -0.00040330, 0.00070544, 0.00035943, 0.00025749, 0.00075539, -0.00035617, -0.00037853, -0.00002989, -0.00003519, -0.00006983, -0.00007024, -0.00000517, -0.00000986, -0.00002909, 0.00000869, -0.00003939, -0.00000143, -0.00000828, -0.00003895, 0.00000262, 0.00001460, -0.00003581, 0.00005766, -0.00005297, 0.00006029, -0.00005024, 0.00004406, -0.00005577, 0.00004632, -0.00004851, 0.00002691, -0.00003566, 0.00007611, -0.00006062, 0.00002632, -0.00001744, 0.00001833, -0.00001568, 0.00002257, -0.00001698, -0.00002419, -0.00000676, 0.00004319, -0.00002903, -0.00001115, -0.00001455, 0.00000300, -0.00003030, 0.00002369, -0.00004679, },
		{ -0.00002063, -0.00000490, -0.00002759, 0.00002323, -0.00010861, 0.00004758, -0.00007228, 0.00008622, -0.00001228, 0.00001082, -0.00003409, 0.00002175, -0.00007639, 0.00001811, -0.00028954, -0.00049672, 0.00079092, 0.00035566, 0.00040982, 0.00074741, -0.00062743, -0.00020231, -0.00011149, 0.00003171, -0.00011785, -0.00002794, -0.00003938, 0.00002020, -0.00005174, 0.00003371, -0.00007396, 0.00003129, -0.00002793, -0.00002944, 0.00000474, 0.00000361, -0.00001843, 0.00004360, -0.00004480, 0.00005443, -0.00004594, 0.00004171, -0.00005739, 0.00004854, -0.00005273, 0.00002703, -0.00002594, 0.00007312, -0.00007364, 0.00004167, -0.00002882, 0.00002766, -0.00002101, 0.00002677, -0.00002751, -0.00002195, 0.00001080, 0.00002659, -0.00002889, -0.00001106, -0.00000476, -0.00001201, -0.00001055, 0.00000134, },
		{ -0.00001990, 0.00000681, -0.00003686, 0.00000871, -0.00001920, -0.00007092, 0.00003427, -0.00005389, 0.00006792, -0.00000937, 0.00000162, -0.00002396, -0.00000014, -0.00005439, -0.00001612, -0.00025607, -0.00060681, 0.00089239, 0.00034263, 0.00063432, 0.00058864, -0.00079968, -0.00003904, -0.00020037, 0.00010493, -0.00017291, 0.00002293, -0.00008262, 0.00005848, -0.00008060, 0.00006199, -0.00011268, 0.00007441, -0.00006420, -0.00000040, -0.00001179, 0.00000771, -0.00001085, 0.00003329, -0.00003660, 0.00004690, -0.00003889, 0.00003532, -0.00005472, 0.00004698, -0.00005446, 0.00002626, -0.00001214, 0.00006008, -0.00007652, 0.00005135, -0.00003785, 0.00003684, -0.00002753, 0.00003206, -0.00004182, -0.00001068, 0.00002096, 0.00001004, -0.00002706, -0.00000834, -0.00000053, -0.00002155, 0.00000567, },
		{ 0.00001073, -0.00003038, 0.00002396, -0.00005992, 0.00003944, -0.00006274, -0.00002198, 0.00000552, -0.00002335, 0.00003665, 0.00000389, -0.00001274, -0.00000948, -0.00002593, -0.00002911, -0.00005475, -0.00021906, -0.00071885, 0.00098779, 0.00035815, 0.00089006, 0.00028179, -0.00083051, 0.00006072, -0.00025821, 0.00015530, -0.00021084, 0.00006185, -0.00011754, 0.00009039, -0.00010311, 0.00008125, -0.00014171, 0.00011329, -0.00010416, 0.00003909, -0.00004246, 0.00002633, -0.00001547, 0.00003048, -0.00003201, 0.00004082, -0.00003174, 0.00002681, -0.00004881, 0.00004208, -0.00005405, 0.00002610, 0.00000277, 0.00003900, -0.00006828, 0.00005230, -0.00004091, 0.00004237, -0.00003222, 0.00003560, -0.00005697, 0.00000866, 0.00002094, -0.00000225, -0.00002617, -0.00000133, -0.00000473, -0.00002033, },
		{ -0.00000653, 0.00000059, -0.00002380, 0.00002201, -0.00006232, 0.00004951, -0.00008913, 0.00001954, -0.00002352, 0.00000743, 0.00000290, 0.00001865, -0.00002363, -0.00000052, -0.00004454, -0.00001308, -0.00008385, -0.00019422, -0.00081360, 0.00104888, 0.00045489, 0.00110340, -0.00012473, -0.00071756, 0.00007075, -0.00025981, 0.00016159, -0.00021277, 0.00007180, -0.00012900, 0.00010285, -0.00010819, 0.00008091, -0.00014856, 0.00013294, -0.00013213, 0.00007516, -0.00007697, 0.00005330, -0.00002997, 0.00003565, -0.00003284, 0.00003856, -0.00002721, 0.00001886, -0.00004176, 0.00003530, -0.00005254, 0.00002866, 0.00001434, 0.00001476, -0.00005146, 0.00004460, -0.00003663, 0.00004210, -0.00003273, 0.00003449, -0.00006879, 0.00003257, 0.00001071, -0.00000787, -0.00002773, 0.00000996, -0.00001756, },
		{ -0.00003553, 0.00001667, -0.00002273, -0.00000064, -0.00000055, -0.00003979, 0.00003076, -0.00008540, 0.00003722, -0.00003727, 0.00002522, -0.00002163, 0.00002533, -0.00002268, -0.00000586, -0.00004649, -0.00001671, -0.00009216, -0.00019358, -0.00087743, 0.00105472, 0.00068360, 0.00118334, -0.00054337, -0.00050283, -0.00000039, -0.00020448, 0.00012101, -0.00017448, 0.00004756, -0.00011153, 0.00009088, -0.00009213, 0.00005783, -0.00012813, 0.00012449, -0.00013604, 0.00009457, -0.00010333, 0.00007937, -0.00004856, 0.00004621, -0.00003862, 0.00004097, -0.00002724, 0.00001406, -0.00003610, 0.00002859, -0.00005116, 0.00003575, 0.00001804, -0.00000621, -0.00003143, 0.00003145, -0.00002649, 0.00003618, -0.00002838, 0.00002692, -0.00007310, 0.00005569, -0.00000672, -0.00000710, -0.00003121, 0.00002314, },
		{ 0.00003373, -0.00005206, 0.00004143, -0.00005102, 0.00003091, -0.00003598, 0.00000125, -0.00001324, -0.00005016, 0.00002441, -0.00002724, 0.00002222, -0.00003010, 0.00001937, -0.00000735, -0.00002739, -0.00003029, -0.00004132, -0.00007894, -0.00021636, -0.00091377, 0.00100771, 0.00107058, 0.00104757, -0.00086859, -0.00026189, -0.00011148, -0.00011788, 0.00005238, -0.00011014, -0.00000030, -0.00007314, 0.00006118, -0.00006178, 0.00001947, -0.00008639, 0.00008988, -0.00011291, 0.00009041, -0.00011285, 0.00009613, -0.00006461, 0.00005794, -0.00004712, 0.00004727, -0.00003241, 0.00001414, -0.00003397, 0.00002375, -0.00005059, 0.00004784, 0.00001086, -0.00001793, -0.00001464, 0.00001805, -0.00001429, 0.00002703, -0.00002061, 0.00001314, -0.00006708, 0.00007223, -0.00002619, -0.00000268, -0.00003412, },
		{ -0.00003305, 0.00003671, -0.00005950, 0.00005769, -0.00007234, 0.00005731, -0.00006926, 0.00004486, -0.00006697, 0.00000429, -0.00001200, 0.00000321, 0.00000083, -0.00002483, 0.00000500, 0.00001549, -0.00005647, -0.00000615, -0.00007495, -0.00005855, -0.00024366, -0.00095022, 0.00094560, 0.00159672, 0.00065187, -0.00100732, -0.00008066, -0.00020335, -0.00004301, -0.00000985, -0.00004816, -0.00004829, -0.00003343, 0.00003064, -0.00003324, -0.00001740, -0.00003977, 0.00004251, -0.00007106, 0.00006540, -0.00010396, 0.00009960, -0.00007362, 0.00006712, -0.00005566, 0.00005565, -0.00004199, 0.00001956, -0.00003659, 0.00002189, -0.00005056, 0.00006339, -0.00000727, -0.00001688, -0.00000637, 0.00000961, -0.00000464, 0.00001840, -0.00001260, -0.00000433, -0.00005036, 0.00007767, -0.00004195, 0.00000133, },
		{ 0.00000124, -0.00002535, 0.00002861, -0.00005212, 0.00005694, -0.00007550, 0.00006465, -0.00008347, 0.00007096, -0.00010797, 0.00005569, -0.00005345, 0.00003933, -0.00002625, -0.00001743, -0.00000506, 0.00003044, -0.00007524, 0.00000539, -0.00009408, -0.00005861, -0.00024062, -0.00103662, 0.00094275, 0.00218712, 0.00001307, -0.00090907, -0.00002709, -0.00022093, -0.00002344, -0.00002881, -0.00002045, -0.00006889, -0.00001610, 0.00001985, -0.00002579, -0.00003244, -0.00000984, 0.00000312, -0.00002774, 0.00003141, -0.00008311, 0.00009197, -0.00007511, 0.00007221, -0.00006238, 0.00006412, -0.00005429, 0.00002949, -0.00004397, 0.00002325, -0.00004989, 0.00007887, -0.00003301, -0.00000341, -0.00000884, 0.00000939, -0.00000112, 0.00001384, -0.00000808, -0.00002129, -0.00002537, 0.00007010, -0.00004954, },
		{ -0.00004717, -0.00000465, -0.00001071, 0.00000942, -0.00002876, 0.00003593, -0.00005489, 0.00004456, -0.00006675, 0.00006337, -0.00011535, 0.00008012, -0.00007688, 0.00006079, -0.00004095, -0.00002401, 0.00000537, 0.00001923, -0.00006297, -0.00001925, -0.00007168, -0.00011064, -0.00016762, -0.00123166, 0.00109678, 0.00271757, -0.00078388, -0.00058539, -0.00012746, -0.00013491, -0.00008520, 0.00001937, -0.00004908, -0.00004206, -0.00003887, 0.00004415, -0.00005352, -0.00001017, -0.00001497, -0.00000798, -0.00000276, 0.00000503, -0.00006225, 0.00008051, -0.00007264, 0.00007442, -0.00006701, 0.00007126, -0.00006723, 0.00004211, -0.00005502, 0.00002736, -0.00004700, 0.00008971, -0.00006029, 0.00001823, -0.00002039, 0.00001734, -0.00000481, 0.00001514, -0.00000986, -0.00003305, 0.00000307, 0.00005096, },
		{ 0.00002459, -0.00003626, -0.00001535, 0.00000822, -0.00001668, 0.00000593, -0.00000124, -0.00001352, -0.00000148, -0.00001810, 0.00001721, -0.00007879, 0.00006193, -0.00006461, 0.00005073, -0.00002807, -0.00005734, 0.00004784, -0.00003049, -0.00000584, -0.00009585, 0.00001038, -0.00023572, 0.00000220, -0.00158057, 0.00150189, 0.00303934, -0.00159918, -0.00011116, -0.00035609, 0.00004558, -0.00022441, 0.00013415, -0.00013583, 0.00003515, -0.00010488, 0.00010582, -0.00011794, 0.00005245, -0.00006219, 0.00002065, -0.00001057, 0.00000085, -0.00005376, 0.00007429, -0.00007186, 0.00007678, -0.00007070, 0.00007647, -0.00007889, 0.00005507, -0.00006791, 0.00003342, -0.00004081, 0.00009173, -0.00008201, 0.00004123, -0.00003607, 0.00003017, -0.00001381, 0.00002150, -0.00001863, -0.00003591, 0.00002902, },
		{ 0.00004693, -0.00000288, -0.00002083, -0.00002744, 0.00002637, -0.00004219, 0.00004243, -0.00004380, 0.00003698, -0.00006131, 0.00005061, -0.00005747, -0.00000418, 0.00000120, -0.00001270, 0.00000418, 0.00001666, -0.00011920, 0.00012165, -0.00011705, 0.00009405, -0.00022216, 0.00014965, -0.00043054, 0.00026524, -0.00209064, 0.00221545, 0.00301716, -0.00227022, 0.00039401, -0.00064238, 0.00027402, -0.00040603, 0.00028813, -0.00025910, 0.00014530, -0.00019937, 0.00019102, -0.00020498, 0.00014258, -0.00014258, 0.00008580, -0.00005370, 0.00002512, -0.00006500, 0.00007986, -0.00007751, 0.00008228, -0.00007497, 0.00007976, -0.00008770, 0.00006597, -0.00008049, 0.00004074, -0.00003148, 0.00008275, -0.00009235, 0.00005846, -0.00004955, 0.00004260, -0.00002397, 0.00002976, -0.00003251, -0.00002834, },
		{ -0.00001146, 0.00005309, -0.00002528, -0.00000616, -0.00003637, 0.00003787, -0.00005861, 0.00006924, -0.00007757, 0.00007989, -0.00011557, 0.00011760, -0.00013765, 0.00008672, -0.00008455, 0.00006564, -0.00006791, 0.00008258, -0.00019667, 0.00021042, -0.00022117, 0.00021467, -0.00037320, 0.00031736, -0.00066012, 0.00057701, -0.00271373, 0.00322878, 0.00256961, -0.00265351, 0.00079916, -0.00089314, 0.00048131, -0.00057435, 0.00043496, -0.00037992, 0.00025520, -0.00029353, 0.00027325, -0.00028825, 0.00023410, -0.00023263, 0.00016910, -0.00012054, 0.00007237, -0.00009460, 0.00009784, -0.00009068, 0.00009183, -0.00008048, 0.00008110, -0.00009249, 0.00007268, -0.00009064, 0.00004891, -0.00002074, 0.00006364, -0.00008867, 0.00006496, -0.00005548, 0.00004948, -0.00003061, 0.00003556, -0.00004758, },
		{ -0.00005924, 0.00001136, 0.00004640, -0.00003795, 0.00000293, -0.00003814, 0.00003818, -0.00005932, 0.00007682, -0.00008988, 0.00009917, -0.00014424, 0.00015845, -0.00019465, 0.00016285, -0.00016545, 0.00014362, -0.00014110, 0.00014627, -0.00026459, 0.00028509, -0.00030952, 0.00031796, -0.00050531, 0.00046257, -0.00086200, 0.00085517, -0.00334432, 0.00445287, 0.00169924, -0.00266419, 0.00099962, -0.00102287, 0.00059926, -0.00067191, 0.00052497, -0.00045507, 0.00032716, -0.00035430, 0.00032235, -0.00033771, 0.00029590, -0.00030150, 0.00024251, -0.00018858, 0.00012665, -0.00013228, 0.00012204, -0.00010763, 0.00010320, -0.00008616, 0.00007997, -0.00009234, 0.00007354, -0.00009642, 0.00005763, -0.00001153, 0.00003836, -0.00007247, 0.00005959, -0.00005156, 0.00004794, -0.00003054, 0.00003520, },
		{ 0.00002723, -0.00006376, 0.00003550, 0.00002851, -0.00003876, 0.00000349, -0.00003076, 0.00002579, -0.00004210, 0.00006105, -0.00007421, 0.00008541, -0.00013414, 0.00015495, -0.00020430, 0.00019461, -0.00020970, 0.00019018, -0.00018585, 0.00017954, -0.00029393, 0.00031317, -0.00034498, 0.00036116, -0.00056880, 0.00052646, -0.00096256, 0.00099803, -0.00383679, 0.00572635, 0.00050173, -0.00230049, 0.00094692, -0.00098135, 0.00058405, -0.00065959, 0.00052270, -0.00045258, 0.00033260, -0.00035658, 0.00031583, -0.00033075, 0.00030298, -0.00032174, 0.00027788, -0.00023213, 0.00016704, -0.00016241, 0.00014147, -0.00012078, 0.00011137, -0.00008918, 0.00007519, -0.00008664, 0.00006756, -0.00009621, 0.00006637, -0.00000704, 0.00001290, -0.00004890, 0.00004534, -0.00003937, 0.00003857, -0.00002351, },
		{ -0.00001255, 0.00001332, -0.00005969, 0.00005651, 0.00000309, -0.00002838, -0.00000493, -0.00001486, 0.00000288, -0.00000996, 0.00002484, -0.00003277, 0.00003968, -0.00008420, 0.00010233, -0.00015615, 0.00016484, -0.00019531, 0.00018165, -0.00017881, 0.00016043, -0.00026309, 0.00027098, -0.00030032, 0.00031226, -0.00052550, 0.00046256, -0.00090131, 0.00091432, -0.00403971, 0.00684642, -0.00085061, -0.00164504, 0.00066130, -0.00076923, 0.00043138, -0.00053061, 0.00041983, -0.00036350, 0.00026282, -0.00029299, 0.00024800, -0.00026130, 0.00024614, -0.00027969, 0.00025737, -0.00023170, 0.00017546, -0.00016999, 0.00014461, -0.00012168, 0.00011041, -0.00008600, 0.00006542, -0.00007526, 0.00005460, -0.00008885, 0.00007403, -0.00000943, -0.00000659, -0.00002503, 0.00002815, -0.00002379, 0.00002528, },
		{ 0.00001380, -0.00000290, -0.00000215, -0.00004831, 0.00007124, -0.00002525, -0.00000972, -0.00002043, 0.00000660, -0.00002543, 0.00002971, -0.00002273, 0.00002402, -0.00002645, -0.00000654, 0.00001184, -0.00005819, 0.00007573, -0.00011882, 0.00011134, -0.00011238, 0.00008277, -0.00016766, 0.00015427, -0.00017013, 0.00016371, -0.00036479, 0.00025606, -0.00065516, 0.00055556, -0.00383793, 0.00761523, -0.00214737, -0.00084247, 0.00022312, -0.00043662, 0.00017783, -0.00031323, 0.00023853, -0.00020561, 0.00013265, -0.00017726, 0.00013312, -0.00014335, 0.00013649, -0.00018129, 0.00018047, -0.00018156, 0.00014377, -0.00014673, 0.00012432, -0.00010465, 0.00009600, -0.00007397, 0.00004993, -0.00005897, 0.00003566, -0.00007392, 0.00007878, -0.00001888, -0.00001576, -0.00000744, 0.00001457, -0.00001101, },
		{ -0.00000610, 0.00000937, -0.00000003, -0.00001349, -0.00003318, 0.00007846, -0.00005205, 0.00001305, -0.00003967, 0.00002949, -0.00005284, 0.00006769, -0.00006957, 0.00008118, -0.00009477, 0.00007730, -0.00009243, 0.00006507, -0.00005105, 0.00000260, -0.00000703, 0.00000116, -0.00003981, -0.00002439, -0.00001755, 0.00002412, -0.00006129, -0.00011156, -0.00006571, -0.00025346, -0.00006015, -0.00318987, 0.00788901, -0.00318987, -0.00006015, -0.00025346, -0.00006571, -0.00011156, -0.00006129, 0.00002412, -0.00001755, -0.00002439, -0.00003981, 0.00000116, -0.00000703, 0.00000260, -0.00005105, 0.00006507, -0.00009243, 0.00007730, -0.00009477, 0.00008118, -0.00006957, 0.00006769, -0.00005284, 0.00002949, -0.00003967, 0.00001305, -0.00005205, 0.00007846, -0.00003318, -0.00001349, -0.00000003, 0.00000937, },
		{ 0.00001380, -0.00001101, 0.00001457, -0.00000744, -0.00001576, -0.00001888, 0.00007878, -0.00007392, 0.00003566, -0.00005897, 0.00004993, -0.00007397, 0.00009600, -0.00010465, 0.00012432, -0.00014673, 0.00014377, -0.00018156, 0.00018047, -0.00018129, 0.00013649, -0.00014335, 0.00013312, -0.00017726, 0.00013265, -0.00020561, 0.00023853, -0.00031323, 0.00017783, -0.00043662, 0.00022312, -0.00084247, -0.00214737, 0.00761523, -0.00383793, 0.00055556, -0.00065516, 0.00025606, -0.00036479, 0.00016371, -0.00017013, 0.00015427, -0.00016766, 0.00008277, -0.00011238, 0.00011134, -0.00011882, 0.00007573, -0.00005819, 0.00001184, -0.00000654, -0.00002645, 0.00002402, -0.00002273, 0.00002971, -0.00002543, 0.00000660, -0.00002043, -0.00000972, -0.00002525, 0.00007124, -0.00004831, -0.00000215, -0.00000290, },
		{ -0.00001255, 0.00002528, -0.00002379, 0.00002815, -0.00002503, -0.00000659, -0.00000943, 0.00007403, -0.00008885, 0.00005460, -0.00007526, 0.00006542, -0.00008600, 0.00011041, -0.00012168, 0.00014461, -0.00016999, 0.00017546, -0.00023170, 0.00025737, -0.00027969, 0.00024614, -0.00026130, 0.00024800, -0.00029299, 0.00026282, -0.00036350, 0.00041983, -0.00053061, 0.00043138, -0.00076923, 0.00066130, -0.00164504, -0.00085061, 0.00684642, -0.00403971, 0.00091432, -0.00090131, 0.00046256, -0.00052550, 0.00031226, -0.00030032, 0.00027098, -0.00026309, 0.00016043, -0.00017881, 0.00018165, -0.00019531, 0.00016484, -0.00015615, 0.00010233, -0.00008420, 0.00003968, -0.00003277, 0.00002484, -0.00000996, 0.00000288, -0.00001486, -0.00000493, -0.00002838, 0.00000309, 0.00005651, -0.00005969, 0.00001332, },
		{ 0.00002723, -0.00002351, 0.00003857, -0.00003937, 0.00004534, -0.00004890, 0.00001290, -0.00000704, 0.00006637, -0.00009621, 0.00006756, -0.00008664, 0.00007519, -0.00008918, 0.00011137, -0.00012078, 0.00014147, -0.00016241, 0.00016704, -0.00023213, 0.00027788, -0.00032174, 0.00030298, -0.00033075, 0.00031583, -0.00035658, 0.00033260, -0.00045258, 0.00052270, -0.00065959, 0.00058405, -0.00098135, 0.00094692, -0.00230049, 0.00050173, 0.00572635, -0.00383679, 0.00099803, -0.00096256, 0.00052646, -0.00056880, 0.00036116, -0.00034498, 0.00031317, -0.00029393, 0.00017954, -0.00018585, 0.00019018, -0.00020970, 0.00019461, -0.00020430, 0.00015495, -0.00013414, 0.00008541, -0.00007421, 0.00006105, -0.00004210, 0.00002579, -0.00003076, 0.00000349, -0.00003876, 0.00002851, 0.00003550, -0.00006376, },
		{ -0.00005924, 0.00003520, -0.00003054, 0.00004794, -0.00005156, 0.00005959, -0.00007247, 0.00003836, -0.00001153, 0.00005763, -0.00009642, 0.00007354, -0.00009234, 0.00007997, -0.00008616, 0.00010320, -0.00010763, 0.00012204, -0.00013228, 0.00012665, -0.00018858, 0.00024251, -0.00030150, 0.00029590, -0.00033771, 0.00032235, -0.00035430, 0.00032716, -0.00045507, 0.00052497, -0.00067191, 0.00059926, -0.00102287, 0.00099962, -0.00266419, 0.00169924, 0.00445287, -0.00334432, 0.00085517, -0.00086200, 0.00046257, -0.00050531, 0.00031796, -0.00030952, 0.00028509, -0.00026459, 0.00014627, -0.00014110, 0.00014362, -0.00016545, 0.00016285, -0.00019465, 0.00015845, -0.00014424, 0.00009917, -0.00008988, 0.00007682, -0.00005932, 0.00003818, -0.00003814, 0.00000293, -0.00003795, 0.00004640, 0.00001136, },
		{ -0.00001146, -0.00004758, 0.00003556, -0.00003061, 0.00004948, -0.00005548, 0.00006496, -0.00008867, 0.00006364, -0.00002074, 0.00004891, -0.00009064, 0.00007268, -0.00009249, 0.00008110, -0.00008048, 0.00009183, -0.00009068, 0.00009784, -0.00009460, 0.00007237, -0.00012054, 0.00016910, -0.00023263, 0.00023410, -0.00028825, 0.00027325, -0.00029353, 0.00025520, -0.00037992, 0.00043496, -0.00057435, 0.00048131, -0.00089314, 0.00079916, -0.00265351, 0.00256961, 0.00322878, -0.00271373, 0.00057701, -0.00066012, 0.00031736, -0.00037320, 0.00021467, -0.00022117, 0.00021042, -0.00019667, 0.00008258, -0.00006791, 0.00006564, -0.00008455, 0.00008672, -0.00013765, 0.00011760, -0.00011557, 0.00007989, -0.00007757, 0.00006924, -0.00005861, 0.00003787, -0.00003637, -0.00000616, -0.00002528, 0.00005309, },
		{ 0.00004693, -0.00002834, -0.00003251, 0.00002976, -0.00002397, 0.00004260, -0.00004955, 0.00005846, -0.00009235, 0.00008275, -0.00003148, 0.00004074, -0.00008049, 0.00006597, -0.00008770, 0.00007976, -0.00007497, 0.00008228, -0.00007751, 0.00007986, -0.00006500, 0.00002512, -0.00005370, 0.00008580, -0.00014258, 0.00014258, -0.00020498, 0.00019102, -0.00019937, 0.00014530, -0.00025910, 0.00028813, -0.00040603, 0.00027402, -0.00064238, 0.00039401, -0.00227022, 0.00301716, 0.00221545, -0.00209064, 0.00026524, -0.00043054, 0.00014965, -0.00022216, 0.00009405, -0.00011705, 0.00012165, -0.00011920, 0.00001666, 0.00000418, -0.00001270, 0.00000120, -0.00000418, -0.00005747, 0.00005061, -0.00006131, 0.00003698, -0.00004380, 0.00004243, -0.00004219, 0.00002637, -0.00002744, -0.00002083, -0.00000288, },
		{ 0.00002459, 0.00002902, -0.00003591, -0.00001863, 0.00002150, -0.00001381, 0.00003017, -0.00003607, 0.00004123, -0.00008201, 0.00009173, -0.00004081, 0.00003342, -0.00006791, 0.00005507, -0.00007889, 0.00007647, -0.00007070, 0.00007678, -0.00007186, 0.00007429, -0.00005376, 0.00000085, -0.00001057, 0.00002065, -0.00006219, 0.00005245, -0.00011794, 0.00010582, -0.00010488, 0.00003515, -0.00013583, 0.00013415, -0.00022441, 0.00004558, -0.00035609, -0.00011116, -0.00159918, 0.00303934, 0.00150189, -0.00158057, 0.00000220, -0.00023572, 0.00001038, -0.00009585, -0.00000584, -0.00003049, 0.00004784, -0.00005734, -0.00002807, 0.00005073, -0.00006461, 0.00006193, -0.00007879, 0.00001721, -0.00001810, -0.00000148, -0.00001352, -0.00000124, 0.00000593, -0.00001668, 0.00000822, -0.00001535, -0.00003626, },
		{ -0.00004717, 0.00005096, 0.00000307, -0.00003305, -0.00000986, 0.00001514, -0.00000481, 0.00001734, -0.00002039, 0.00001823, -0.00006029, 0.00008971, -0.00004700, 0.00002736, -0.00005502, 0.00004211, -0.00006723, 0.00007126, -0.00006701, 0.00007442, -0.00007264, 0.00008051, -0.00006225, 0.00000503, -0.00000276, -0.00000798, -0.00001497, -0.00001017, -0.00005352, 0.00004415, -0.00003887, -0.00004206, -0.00004908, 0.00001937, -0.00008520, -0.00013491, -0.00012746, -0.00058539, -0.00078388, 0.00271757, 0.00109678, -0.00123166, -0.00016762, -0.00011064, -0.00007168, -0.00001925, -0.00006297, 0.00001923, 0.00000537, -0.00002401, -0.00004095, 0.00006079, -0.00007688, 0.00008012, -0.00011535, 0.00006337, -0.00006675, 0.00004456, -0.00005489, 0.00003593, -0.00002876, 0.00000942, -0.00001071, -0.00000465, },
		{ 0.00000124, -0.00004954, 0.00007010, -0.00002537, -0.00002129, -0.00000808, 0.00001384, -0.00000112, 0.00000939, -0.00000884, -0.00000341, -0.00003301, 0.00007887, -0.00004989, 0.00002325, -0.00004397, 0.00002949, -0.00005429, 0.00006412, -0.00006238, 0.00007221, -0.00007511, 0.00009197, -0.00008311, 0.00003141, -0.00002774, 0.00000312, -0.00000984, -0.00003244, -0.00002579, 0.00001985, -0.00001610, -0.00006889, -0.00002045, -0.00002881, -0.00002344, -0.00022093, -0.00002709, -0.00090907, 0.00001307, 0.00218712, 0.00094275, -0.00103662, -0.00024062, -0.00005861, -0.00009408, 0.00000539, -0.00007524, 0.00003044, -0.00000506, -0.00001743, -0.00002625, 0.00003933, -0.00005345, 0.00005569, -0.00010797, 0.00007096, -0.00008347, 0.00006465, -0.00007550, 0.00005694, -0.00005212, 0.00002861, -0.00002535, },
		{ -0.00003305, 0.00000133, -0.00004195, 0.00007767, -0.00005036, -0.00000433, -0.00001260, 0.00001840, -0.00000464, 0.00000961, -0.00000637, -0.00001688, -0.00000727, 0.00006339, -0.00005056, 0.00002189, -0.00003659, 0.00001956, -0.00004199, 0.00005565, -0.00005566, 0.00006712, -0.00007362, 0.00009960, -0.00010396, 0.00006540, -0.00007106, 0.00004251, -0.00003977, -0.00001740, -0.00003324, 0.00003064, -0.00003343, -0.00004829, -0.00004816, -0.00000985, -0.00004301, -0.00020335, -0.00008066, -0.00100732, 0.00065187, 0.00159672, 0.00094560, -0.00095022, -0.00024366, -0.00005855, -0.00007495, -0.00000615, -0.00005647, 0.00001549, 0.00000500, -0.00002483, 0.00000083, 0.00000321, -0.00001200, 0.00000429, -0.00006697, 0.00004486, -0.00006926, 0.00005731, -0.00007234, 0.00005769, -0.00005950, 0.00003671, },
		{ 0.00003373, -0.00003412, -0.00000268, -0.00002619, 0.00007223, -0.00006708, 0.00001314, -0.00002061, 0.00002703, -0.00001429, 0.00001805, -0.00001464, -0.00001793, 0.00001086, 0.00004784, -0.00005059, 0.00002375, -0.00003397, 0.00001414, -0.00003241, 0.00004727, -0.00004712, 0.00005794, -0.00006461, 0.00009613, -0.00011285, 0.00009041, -0.00011291, 0.00008988, -0.00008639, 0.00001947, -0.00006178, 0.00006118, -0.00007314, -0.00000030, -0.00011014, 0.00005238, -0.00011788, -0.00011148, -0.00026189, -0.00086859, 0.00104757, 0.00107058, 0.00100771, -0.00091377, -0.00021636, -0.00007894, -0.00004132, -0.00003029, -0.00002739, -0.00000735, 0.00001937, -0.00003010, 0.00002222, -0.00002724, 0.00002441, -0.00005016, -0.00001324, 0.00000125, -0.00003598, 0.00003091, -0.00005102, 0.00004143, -0.00005206, },
		{ -0.00003553, 0.00002314, -0.00003121, -0.00000710, -0.00000672, 0.00005569, -0.00007310, 0.00002692, -0.00002838, 0.00003618, -0.00002649, 0.00003145, -0.00003143, -0.00000621, 0.00001804, 0.00003575, -0.00005116, 0.00002859, -0.00003610, 0.00001406, -0.00002724, 0.00004097, -0.00003862, 0.00004621, -0.00004856, 0.00007937, -0.00010333, 0.00009457, -0.00013604, 0.00012449, -0.00012813, 0.00005783, -0.00009213, 0.00009088, -0.00011153, 0.00004756, -0.00017448, 0.00012101, -0.00020448, -0.00000039, -0.00050283, -0.00054337, 0.00118334, 0.00068360, 0.00105472, -0.00087743, -0.00019358, -0.00009216, -0.00001671, -0.00004649, -0.00000586, -0.00002268, 0.00002533, -0.00002163, 0.00002522, -0.00003727, 0.00003722, -0.00008540, 0.00003076, -0.00003979, -0.00000055, -0.00000064, -0.00002273, 0.00001667, },
		{ -0.00000653, -0.00001756, 0.00000996, -0.00002773, -0.00000787, 0.00001071, 0.00003257, -0.00006879, 0.00003449, -0.00003273, 0.00004210, -0.00003663, 0.00004460, -0.00005146, 0.00001476, 0.00001434, 0.00002866, -0.00005254, 0.00003530, -0.00004176, 0.00001886, -0.00002721, 0.00003856, -0.00003284, 0.00003565, -0.00002997, 0.00005330, -0.00007697, 0.00007516, -0.00013213, 0.00013294, -0.00014856, 0.00008091, -0.00010819, 0.00010285, -0.00012900, 0.00007180, -0.00021277, 0.00016159, -0.00025981, 0.00007075, -0.00071756, -0.00012473, 0.00110340, 0.00045489, 0.00104888, -0.00081360, -0.00019422, -0.00008385, -0.00001308, -0.00004454, -0.00000052, -0.00002363, 0.00001865, 0.00000290, 0.00000743, -0.00002352, 0.00001954, -0.00008913, 0.00004951, -0.00006232, 0.00002201, -0.00002380, 0.00000059, },
		{ 0.00001073, -0.00002033, -0.00000473, -0.00000133, -0.00002617, -0.00000225, 0.00002094, 0.00000866, -0.00005697, 0.00003560, -0.00003222, 0.00004237, -0.00004091, 0.00005230, -0.00006828, 0.00003900, 0.00000277, 0.00002610, -0.00005405, 0.00004208, -0.00004881, 0.00002681, -0.00003174, 0.00004082, -0.00003201, 0.00003048, -0.00001547, 0.00002633, -0.00004246, 0.00003909, -0.00010416, 0.00011329, -0.00014171, 0.00008125, -0.00010311, 0.00009039, -0.00011754, 0.00006185, -0.00021084, 0.00015530, -0.00025821, 0.00006072, -0.00083051, 0.00028179, 0.00089006, 0.00035815, 0.00098779, -0.00071885, -0.00021906, -0.00005475, -0.00002911, -0.00002593, -0.00000948, -0.00001274, 0.00000389, 0.00003665, -0.00002335, 0.00000552, -0.00002198, -0.00006274, 0.00003944, -0.00005992, 0.00002396, -0.00003038, },
		{ -0.00001990, 0.00000567, -0.00002155, -0.00000053, -0.00000834, -0.00002706, 0.00001004, 0.00002096, -0.00001068, -0.00004182, 0.00003206, -0.00002753, 0.00003684, -0.00003785, 0.00005135, -0.00007652, 0.00006008, -0.00001214, 0.00002626, -0.00005446, 0.00004698, -0.00005472, 0.00003532, -0.00003889, 0.00004690, -0.00003660, 0.00003329, -0.00001085, 0.00000771, -0.00001179, -0.00000040, -0.00006420, 0.00007441, -0.00011268, 0.00006199, -0.00008060, 0.00005848, -0.00008262, 0.00002293, -0.00017291, 0.00010493, -0.00020037, -0.00003904, -0.00079968, 0.00058864, 0.00063432, 0.00034263, 0.00089239, -0.00060681, -0.00025607, -0.00001612, -0.00005439, -0.00000014, -0.00002396, 0.00000162, -0.00000937, 0.00006792, -0.00005389, 0.00003427, -0.00007092, -0.00001920, 0.00000871, -0.00003686, 0.00000681, },
		{ -0.00002063, 0.00000134, -0.00001055, -0.00001201, -0.00000476, -0.00001106, -0.00002889, 0.00002659, 0.00001080, -0.00002195, -0.00002751, 0.00002677, -0.00002101, 0.00002766, -0.00002882, 0.00004167, -0.00007364, 0.00007312, -0.00002594, 0.00002703, -0.00005273, 0.00004854, -0.00005739, 0.00004171, -0.00004594, 0.00005443, -0.00004480, 0.00004360, -0.00001843, 0.00000361, 0.00000474, -0.00002944, -0.00002793, 0.00003129, -0.00007396, 0.00003371, -0.00005174, 0.00002020, -0.00003938, -0.00002794, -0.00011785, 0.00003171, -0.00011149, -0.00020231, -0.00062743, 0.00074741, 0.00040982, 0.00035566, 0.00079092, -0.00049672, -0.00028954, 0.00001811, -0.00007639, 0.00002175, -0.00003409, 0.00001082, -0.00001228, 0.00008622, -0.00007228, 0.00004758, -0.00010861, 0.00002323, -0.00002759, -0.00000490, },
		{ 0.00002245, -0.00004679, 0.00002369, -0.00003030, 0.00000300, -0.00001455, -0.00001115, -0.00002903, 0.00004319, -0.00000676, -0.00002419, -0.00001698, 0.00002257, -0.00001568, 0.00001833, -0.00001744, 0.00002632, -0.00006062, 0.00007611, -0.00003566, 0.00002691, -0.00004851, 0.00004632, -0.00005577, 0.00004406, -0.00005024, 0.00006029, -0.00005297, 0.00005766, -0.00003581, 0.00001460, 0.00000262, -0.00003895, -0.00000828, -0.00000143, -0.00003939, 0.00000869, -0.00002909, -0.00000986, -0.00000517, -0.00007024, -0.00006983, -0.00003519, -0.00002989, -0.00037853, -0.00035617, 0.00075539, 0.00025749, 0.00035943, 0.00070544, -0.00040330, -0.00030814, 0.00003834, -0.00008674, 0.00003268, -0.00003401, 0.00001064, -0.00000121, 0.00008709, -0.00007358, 0.00003750, -0.00012191, 0.00004884, -0.00005423, },
		{ -0.00006198, 0.00003549, -0.00006231, 0.00003843, -0.00004581, 0.00001733, -0.00002606, -0.00001049, -0.00002521, 0.00005544, -0.00002719, -0.00001883, -0.00001140, 0.00002123, -0.00001389, 0.00001235, -0.00000817, 0.00001026, -0.00004139, 0.00007007, -0.00004038, 0.00002555, -0.00004242, 0.00004108, -0.00005019, 0.00004181, -0.00005020, 0.00006183, -0.00005698, 0.00006975, -0.00005673, 0.00003539, -0.00001530, -0.00002790, -0.00001075, -0.00001485, -0.00001856, -0.00000455, -0.00002083, -0.00002215, 0.00000813, -0.00008905, -0.00004757, -0.00007160, 0.00000912, -0.00051136, -0.00005177, 0.00064753, 0.00018366, 0.00033788, 0.00064503, -0.00033158, -0.00030905, 0.00004273, -0.00008413, 0.00003214, -0.00002403, 0.00000300, 0.00002074, 0.00007357, -0.00006165, 0.00000663, -0.00010811, 0.00005034, },
		{ 0.00003069, -0.00005038, 0.00003171, -0.00006341, 0.00004105, -0.00005209, 0.00002631, -0.00003619, -0.00001006, -0.00001674, 0.00006023, -0.00004585, -0.00000880, -0.00001015, 0.00002293, -0.00001650, 0.00001182, -0.00000462, -0.00000136, -0.00002134, 0.00005827, -0.00004099, 0.00002360, -0.00003583, 0.00003444, -0.00004222, 0.00003593, -0.00004585, 0.00005796, -0.00005394, 0.00007442, -0.00007344, 0.00005715, -0.00004034, -0.00000312, -0.00003214, -0.00000835, -0.00001386, -0.00000433, -0.00002771, -0.00001579, -0.00000135, -0.00008084, -0.00005690, -0.00006842, -0.00001267, -0.00055818, 0.00021841, 0.00047881, 0.00016903, 0.00029403, 0.00060644, -0.00027742, -0.00029714, 0.00003626, -0.00007332, 0.00002509, -0.00000935, -0.00000565, 0.00004576, 0.00005382, -0.00004683, -0.00003356, -0.00007509, },
		{ -0.00003715, 0.00000054, -0.00002668, 0.00001565, -0.00005260, 0.00003230, -0.00004828, 0.00002803, -0.00004350, -0.00000941, -0.00000503, 0.00005686, -0.00005931, 0.00000264, -0.00001156, 0.00002652, -0.00002258, 0.00001676, -0.00000827, -0.00000499, -0.00000556, 0.00004483, -0.00003941, 0.00002227, -0.00003033, 0.00002835, -0.00003404, 0.00002852, -0.00003887, 0.00004972, -0.00004354, 0.00006868, -0.00007983, 0.00007109, -0.00006166, 0.00002403, -0.00006273, 0.00001153, -0.00002095, 0.00000471, -0.00004360, 0.00000198, -0.00002573, -0.00005376, -0.00008936, -0.00003467, -0.00008943, -0.00050349, 0.00040517, 0.00030424, 0.00018323, 0.00024137, 0.00057957, -0.00023202, -0.00028081, 0.00002682, -0.00006148, 0.00001843, 0.00000349, -0.00000800, 0.00006534, 0.00003647, -0.00004046, -0.00006799, },
		{ -0.00008458, -0.00000859, -0.00002704, -0.00000174, -0.00000397, -0.00003661, 0.00001708, -0.00003722, 0.00002330, -0.00004800, -0.00000716, 0.00000695, 0.00004700, -0.00006613, 0.00001294, -0.00001374, 0.00003014, -0.00002992, 0.00002516, -0.00001806, 0.00000028, 0.00000270, 0.00003331, -0.00003763, 0.00002264, -0.00002718, 0.00002440, -0.00002772, 0.00002202, -0.00003189, 0.00003999, -0.00002828, 0.00005326, -0.00007373, 0.00007191, -0.00007072, 0.00004233, -0.00009070, 0.00003476, -0.00003190, 0.00001473, -0.00005914, 0.00001995, -0.00005191, -0.00002295, -0.00012725, 0.00000719, -0.00019401, -0.00036201, 0.00048841, 0.00016342, 0.00019852, 0.00019430, 0.00055428, -0.00018743, -0.00026717, 0.00002080, -0.00005409, 0.00001711, 0.00001026, 0.00000040, 0.00007420, 0.00002612, -0.00004931, },
		{ -0.00007227, -0.00007916, 0.00000198, -0.00004226, 0.00001496, -0.00001857, -0.00002301, 0.00000180, -0.00002360, 0.00001449, -0.00005037, -0.00000198, 0.00001581, 0.00003392, -0.00006684, 0.00002079, -0.00001536, 0.00003213, -0.00003594, 0.00003386, -0.00003085, 0.00001268, 0.00000269, 0.00002578, -0.00003690, 0.00002515, -0.00002686, 0.00002329, -0.00002447, 0.00001832, -0.00002753, 0.00003238, -0.00001273, 0.00003237, -0.00005762, 0.00005972, -0.00006440, 0.00004497, -0.00010676, 0.00005226, -0.00003934, 0.00001890, -0.00006639, 0.00002834, -0.00006787, -0.00000308, -0.00015220, 0.00003146, -0.00028855, -0.00017138, 0.00047739, 0.00007372, 0.00019820, 0.00016143, 0.00052506, -0.00014010, -0.00025898, 0.00002043, -0.00005249, 0.00002200, 0.00001100, 0.00001935, 0.00007193, 0.00002136, },
		{ 0.00001605, -0.00010095, -0.00005652, -0.00000563, -0.00004200, 0.00001895, -0.00002312, -0.00001704, -0.00000835, -0.00001188, 0.00000418, -0.00005102, 0.00000633, 0.00001919, 0.00002112, -0.00006318, 0.00002606, -0.00001600, 0.00003168, -0.00003872, 0.00003977, -0.00004267, 0.00002853, -0.00000403, 0.00002256, -0.00003746, 0.00002949, -0.00002902, 0.00002474, -0.00002435, 0.00001810, -0.00002730, 0.00002987, -0.00000176, 0.00001208, -0.00003746, 0.00003963, -0.00004582, 0.00003162, -0.00010719, 0.00005915, -0.00003946, 0.00001445, -0.00006233, 0.00002306, -0.00006792, -0.00000187, -0.00015355, 0.00002024, -0.00033844, 0.00002181, 0.00040179, 0.00003226, 0.00017836, 0.00014379, 0.00049208, -0.00009143, -0.00025443, 0.00002369, -0.00005404, 0.00003017, 0.00000951, 0.00004477, 0.00006196, },
		{ 0.00004843, 0.00000328, -0.00012370, -0.00002763, -0.00002443, -0.00002999, 0.00001176, -0.00001760, -0.00001989, -0.00001109, -0.00000486, -0.00000578, -0.00004958, 0.00001642, 0.00001658, 0.00001117, -0.00005717, 0.00002933, -0.00001606, 0.00002909, -0.00003767, 0.00004097, -0.00005009, 0.00004362, -0.00001450, 0.00002260, -0.00003868, 0.00003471, -0.00003270, 0.00002776, -0.00002637, 0.00002068, -0.00003104, 0.00003354, 0.00000125, -0.00000201, -0.00002016, 0.00001935, -0.00002259, 0.00000782, -0.00009420, 0.00005561, -0.00003292, 0.00000334, -0.00004959, 0.00000682, -0.00005448, -0.00001739, -0.00013270, -0.00002922, -0.00032421, 0.00017707, 0.00029844, 0.00002383, 0.00014417, 0.00013691, 0.00045924, -0.00004565, -0.00024927, 0.00002645, -0.00005423, 0.00003706, 0.00001111, 0.00007112, },
		{ 0.00009407, 0.00003314, -0.00002025, -0.00013106, -0.00000440, -0.00004422, -0.00001401, -0.00000089, -0.00000606, -0.00002906, -0.00000722, -0.00000331, -0.00001440, -0.00004527, 0.00002579, 0.00000943, 0.00000508, -0.00005030, 0.00003126, -0.00001625, 0.00002552, -0.00003360, 0.00003725, -0.00005124, 0.00005449, -0.00002548, 0.00002420, -0.00003950, 0.00003962, -0.00003678, 0.00003107, -0.00002902, 0.00002436, -0.00003697, 0.00004217, -0.00000433, -0.00000682, -0.00001081, 0.00000600, -0.00000350, -0.00001782, -0.00007403, 0.00004557, -0.00002346, -0.00000943, -0.00003448, -0.00001298, -0.00003592, -0.00003991, -0.00010180, -0.00010362, -0.00024703, 0.00027075, 0.00019903, 0.00003032, 0.00010412, 0.00013477, 0.00043086, -0.00000681, -0.00023971, 0.00002533, -0.00004944, 0.00003913, 0.00001993, },
		{ 0.00003748, 0.00011174, 0.00001417, -0.00005167, -0.00012000, 0.00000540, -0.00005649, -0.00000178, -0.00001236, 0.00000595, -0.00004017, 0.00000049, -0.00000655, -0.00002108, -0.00003767, 0.00003188, 0.00000031, 0.00000242, -0.00004325, 0.00003213, -0.00001715, 0.00002238, -0.00002829, 0.00003008, -0.00004626, 0.00005932, -0.00003442, 0.00002574, -0.00003892, 0.00004315, -0.00004028, 0.00003367, -0.00003089, 0.00002722, -0.00004243, 0.00005264, -0.00001625, -0.00000274, -0.00001084, 0.00000326, 0.00000495, -0.00003717, -0.00005384, 0.00003408, -0.00001537, -0.00001873, -0.00002349, -0.00002838, -0.00002191, -0.00005736, -0.00007772, -0.00017927, -0.00012637, 0.00029947, 0.00012266, 0.00003789, 0.00006510, 0.00013305, 0.00040912, 0.00002348, -0.00022444, 0.00001949, -0.00003872, 0.00003560, },
		{ 0.00002838, 0.00006277, 0.00012392, -0.00001274, -0.00008266, -0.00009512, 0.00000090, -0.00005781, 0.00000242, -0.00001820, 0.00001408, -0.00004915, 0.00000889, -0.00001338, -0.00002495, -0.00002754, 0.00003326, -0.00000827, 0.00000210, -0.00003611, 0.00003184, -0.00001885, 0.00002077, -0.00002374, 0.00002185, -0.00003697, 0.00005811, -0.00004002, 0.00002626, -0.00003636, 0.00004458, -0.00004264, 0.00003508, -0.00003122, 0.00002791, -0.00004494, 0.00006113, -0.00003033, 0.00000687, -0.00001785, 0.00001025, 0.00000075, -0.00004543, -0.00003892, 0.00002487, -0.00001114, -0.00002195, -0.00002009, -0.00003474, -0.00001853, -0.00006143, -0.00007419, -0.00023144, 0.00000826, 0.00027643, 0.00007466, 0.00003982, 0.00003018, 0.00013038, 0.00039347, 0.00004640, -0.00020484, 0.00001083, -0.00002383, },
		{ -0.00000794, 0.00002078, 0.00009379, 0.00012993, -0.00005090, -0.00010303, -0.00006618, -0.00001253, -0.00005037, -0.00000100, -0.00001769, 0.00001681, -0.00005381, 0.00001564, -0.00002276, -0.00002475, -0.00001708, 0.00003038, -0.00001499, 0.00000324, -0.00002894, 0.00003013, -0.00002093, 0.00002110, -0.00002145, 0.00001505, -0.00002610, 0.00005233, -0.00004225, 0.00002554, -0.00003178, 0.00004361, -0.00004362, 0.00003537, -0.00003007, 0.00002611, -0.00004315, 0.00006452, -0.00004211, 0.00001731, -0.00002695, 0.00002232, -0.00001346, -0.00004213, -0.00003135, 0.00001918, -0.00001059, -0.00001988, -0.00002347, -0.00003258, -0.00002565, -0.00005149, -0.00009559, -0.00024352, 0.00012783, 0.00022334, 0.00005030, 0.00003563, -0.00000084, 0.00012748, 0.00038171, 0.00006464, -0.00018368, 0.00000260, },
		{ -0.00000244, 0.00000618, 0.00001593, 0.00012884, 0.00012672, -0.00009911, -0.00010569, -0.00004349, -0.00002670, -0.00003987, -0.00000816, -0.00001332, 0.00001542, -0.00005413, 0.00001959, -0.00003378, -0.00001928, -0.00000928, 0.00002541, -0.00002001, 0.00000576, -0.00002222, 0.00002700, -0.00002278, 0.00002303, -0.00002197, 0.00001145, -0.00001636, 0.00004416, -0.00004199, 0.00002407, -0.00002567, 0.00004031, -0.00004321, 0.00003488, -0.00002811, 0.00002265, -0.00003736, 0.00006147, -0.00004851, 0.00002458, -0.00003311, 0.00003323, -0.00003193, -0.00003042, -0.00003022, 0.00001615, -0.00001158, -0.00001577, -0.00002964, -0.00002660, -0.00003769, -0.00003454, -0.00013522, -0.00021222, 0.00021242, 0.00016164, 0.00004054, 0.00002789, -0.00002941, 0.00012533, 0.00037183, 0.00008058, -0.00016323, },
		{ -0.00014407, -0.00000323, 0.00001720, 0.00001604, 0.00016672, 0.00010883, -0.00015015, -0.00009024, -0.00003320, -0.00003490, -0.00003205, -0.00001438, -0.00000880, 0.00001261, -0.00005145, 0.00002032, -0.00004518, -0.00000829, -0.00000673, 0.00002119, -0.00002464, 0.00001031, -0.00001704, 0.00002301, -0.00002393, 0.00002580, -0.00002495, 0.00001172, -0.00000972, 0.00003583, -0.00004050, 0.00002276, -0.00001893, 0.00003503, -0.00004143, 0.00003394, -0.00002622, 0.00001905, -0.00002932, 0.00005273, -0.00004869, 0.00002689, -0.00003324, 0.00003777, -0.00004841, -0.00001509, -0.00003312, 0.00001417, -0.00001175, -0.00001310, -0.00003406, -0.00002254, -0.00004708, -0.00002146, -0.00017829, -0.00014738, 0.00025504, 0.00010657, 0.00003703, 0.00001939, -0.00005646, 0.00012376, 0.00036334, 0.00009504, },
		{ 0.00010729, -0.00012511, -0.00000035, 0.00002521, 0.00002273, 0.00020535, 0.00007063, -0.00019232, -0.00006361, -0.00003496, -0.00003492, -0.00002970, -0.00001680, -0.00000682, 0.00001077, -0.00004729, 0.00001751, -0.00005490, 0.00000664, -0.00001028, 0.00001985, -0.00003021, 0.00001775, -0.00001472, 0.00001923, -0.00002437, 0.00002860, -0.00002939, 0.00001544, -0.00000710, 0.00002899, -0.00003904, 0.00002263, -0.00001274, 0.00002842, -0.00003834, 0.00003267, -0.00002499, 0.00001682, -0.00002152, 0.00004068, -0.00004403, 0.00002505, -0.00002738, 0.00003368, -0.00005836, -0.00000051, -0.00003770, 0.00001230, -0.00001016, -0.00001344, -0.00003439, -0.00002378, -0.00004870, -0.00002138, -0.00020835, -0.00006695, 0.00026034, 0.00006520, 0.00003453, 0.00001159, -0.00008179, 0.00012132, 0.00035723, },
		},
	}
	;

	float f32_4[16][32][32] = {
		{ { -0.00081354, -0.00008218, 0.00031092, 0.00021543, 0.00006194, 0.00037482, -0.00004131, -0.00009310, 0.00019781, -0.00012537, -0.00030332, 0.00011771, 0.00021859, -0.00029858, 0.00007687, -0.00002241, -0.00005683, -0.00002241, 0.00007687, -0.00029858, 0.00021859, 0.00011771, -0.00030332, -0.00012537, 0.00019781, -0.00009310, -0.00004131, 0.00037482, 0.00006194, 0.00021543, 0.00031092, -0.00008218, },
		{ -0.00002376, -0.00082643, -0.00011480, 0.00017490, 0.00024383, -0.00001075, 0.00028498, 0.00004094, -0.00016563, 0.00015980, 0.00007933, -0.00029480, -0.00016151, 0.00043300, -0.00023368, -0.00005112, 0.00003134, -0.00002862, -0.00010006, 0.00012677, -0.00020099, -0.00003221, 0.00020726, -0.00012905, -0.00027871, 0.00008391, -0.00000526, -0.00008740, 0.00040738, 0.00018291, 0.00020291, 0.00043077, },
		{ 0.00054810, 0.00005793, -0.00087618, -0.00011417, 0.00003605, 0.00024386, -0.00000907, 0.00016008, 0.00008990, -0.00014490, 0.00003845, 0.00014897, -0.00005141, -0.00047219, 0.00044855, -0.00001430, -0.00015959, 0.00001398, 0.00002618, -0.00013308, 0.00007125, -0.00004043, -0.00016614, 0.00010507, 0.00005809, -0.00026479, -0.00010381, -0.00000731, -0.00009779, 0.00041061, 0.00032006, 0.00022331, },
		{ 0.00027684, 0.00066894, 0.00018287, -0.00097365, -0.00010287, -0.00006414, 0.00018603, 0.00003855, 0.00007060, 0.00005445, -0.00007283, -0.00000038, -0.00001586, 0.00029378, -0.00059948, 0.00021968, 0.00023496, -0.00015690, -0.00005367, 0.00004693, -0.00009529, -0.00003356, 0.00006349, -0.00014189, -0.00006198, 0.00011574, -0.00011888, -0.00022332, -0.00011849, -0.00014354, 0.00039720, 0.00047193, },
		{ 0.00063900, 0.00038586, 0.00078315, 0.00036589, -0.00109526, -0.00012927, -0.00010388, 0.00009930, 0.00006224, 0.00005798, -0.00001552, -0.00008641, 0.00015691, -0.00032430, 0.00050123, -0.00043012, -0.00013468, 0.00035746, -0.00003469, -0.00009685, 0.00000195, -0.00002609, -0.00010412, 0.00006090, -0.00003512, -0.00015498, 0.00003036, 0.00002449, -0.00019926, -0.00024790, -0.00027969, 0.00031667, },
		{ 0.00009375, 0.00075994, 0.00059286, 0.00090624, 0.00058431, -0.00119493, -0.00021393, -0.00012321, 0.00004903, 0.00003066, 0.00006355, 0.00001100, -0.00028545, 0.00044914, -0.00054255, 0.00039937, -0.00004631, -0.00040677, 0.00027948, 0.00012007, -0.00005184, -0.00007592, 0.00000700, -0.00010187, -0.00000683, 0.00004379, -0.00012703, -0.00009859, 0.00005682, -0.00008289, -0.00028023, -0.00046339, },
		{ -0.00056846, -0.00027172, 0.00070076, 0.00088827, 0.00111948, 0.00080467, -0.00126751, -0.00030862, -0.00019288, 0.00004973, 0.00001747, -0.00004430, 0.00022294, -0.00060915, 0.00065977, -0.00046633, 0.00002068, 0.00032353, -0.00044540, 0.00006487, 0.00019116, 0.00007409, -0.00011260, -0.00002825, -0.00005912, -0.00006153, 0.00004130, -0.00003580, -0.00016076, -0.00001949, 0.00000581, -0.00018734, },
		{ -0.00005750, -0.00049510, -0.00064589, 0.00034649, 0.00113094, 0.00152804, 0.00107182, -0.00139098, -0.00034039, -0.00031465, -0.00000148, 0.00016402, -0.00034314, 0.00054793, -0.00084306, 0.00057678, -0.00008530, -0.00041119, 0.00046422, -0.00025996, -0.00013490, 0.00012436, 0.00020139, -0.00005652, -0.00009717, -0.00003982, -0.00006241, -0.00001337, 0.00002908, -0.00012584, -0.00011355, -0.00000758, },
		{ -0.00008991, -0.00001278, -0.00027522, -0.00082488, -0.00026226, 0.00106874, 0.00212534, 0.00157117, -0.00166232, -0.00033305, -0.00034980, -0.00027538, 0.00053902, -0.00073627, 0.00076000, -0.00077095, 0.00016904, 0.00039441, -0.00063788, 0.00032246, -0.00000388, -0.00019512, -0.00003096, 0.00023901, 0.00007265, -0.00012748, -0.00007501, -0.00003354, -0.00005709, 0.00002787, -0.00004908, -0.00014158, },
		{ -0.00010071, -0.00014443, -0.00009163, -0.00006681, -0.00069200, -0.00087855, 0.00050077, 0.00266625, 0.00255331, -0.00203340, -0.00049387, -0.00008235, -0.00084741, 0.00100911, -0.00096624, 0.00063545, -0.00035587, -0.00036766, 0.00069260, -0.00054741, 0.00003244, 0.00014628, -0.00010549, -0.00016313, 0.00015727, 0.00019416, -0.00007364, -0.00013333, -0.00002680, -0.00006017, -0.00001011, -0.00000139, },
		{ 0.00000003, -0.00005548, -0.00011300, -0.00020909, -0.00002739, -0.00033799, -0.00116993, -0.00051354, 0.00269788, 0.00410961, -0.00217783, -0.00113727, 0.00058258, -0.00154698, 0.00125820, -0.00078690, 0.00014332, 0.00020244, -0.00073959, 0.00065342, -0.00024205, -0.00019590, 0.00011851, 0.00003565, -0.00018691, 0.00001147, 0.00022841, 0.00004084, -0.00015215, -0.00006237, -0.00004745, -0.00003527, },
		{ -0.00003629, -0.00000538, -0.00005950, -0.00003514, -0.00023615, -0.00017209, -0.00002300, -0.00095452, -0.00157703, 0.00180346, 0.00594120, -0.00155434, -0.00243340, 0.00142803, -0.00197707, 0.00098593, -0.00017458, -0.00049112, 0.00059336, -0.00076227, 0.00034687, 0.00005291, -0.00023320, -0.00001798, 0.00010975, -0.00010574, -0.00010302, 0.00015905, 0.00014107, -0.00009871, -0.00010637, -0.00005948, },
		{ -0.00009952, -0.00004745, 0.00002034, -0.00009563, -0.00000585, -0.00012686, -0.00034787, 0.00001511, -0.00037488, -0.00212484, -0.00006455, 0.00733359, 0.00028820, -0.00413984, 0.00193078, -0.00174707, 0.00015626, 0.00059825, -0.00092007, 0.00061372, -0.00048409, 0.00000029, 0.00016046, -0.00011014, -0.00012697, 0.00007217, 0.00000126, -0.00012528, 0.00004264, 0.00016278, -0.00000350, -0.00010616, },
		{ -0.00004699, -0.00012389, -0.00010035, 0.00006096, -0.00009640, -0.00007390, 0.00002773, -0.00036040, -0.00023855, 0.00015077, -0.00177287, -0.00240568, 0.00744463, 0.00332231, -0.00554996, 0.00153315, -0.00077992, -0.00090520, 0.00111177, -0.00091442, 0.00030871, -0.00013356, -0.00017741, 0.00006752, 0.00003233, -0.00012116, -0.00002379, 0.00005157, -0.00007047, -0.00004087, 0.00010419, 0.00006543, },
		{ 0.00006348, 0.00002734, -0.00008733, -0.00017456, 0.00005774, -0.00002368, -0.00018222, 0.00007718, -0.00015304, -0.00053781, 0.00019981, -0.00061264, -0.00432842, 0.00578577, 0.00685060, -0.00577837, 0.00003875, 0.00054233, -0.00167392, 0.00107852, -0.00052749, -0.00007476, 0.00006057, -0.00013675, -0.00009600, 0.00006858, -0.00002333, -0.00008535, 0.00002622, -0.00000585, -0.00005279, 0.00002269, },
		{ -0.00001901, 0.00000906, 0.00005264, 0.00000208, -0.00020429, -0.00002289, 0.00007581, -0.00021573, -0.00004253, 0.00011974, -0.00055190, -0.00034434, 0.00077805, -0.00500139, 0.00259622, 0.00970957, -0.00426228, -0.00214528, 0.00154319, -0.00176018, 0.00054701, -0.00005129, -0.00028099, 0.00002914, 0.00000916, -0.00018112, -0.00001786, 0.00006944, -0.00006432, -0.00002436, 0.00001372, -0.00001982, },
		{ 0.00000072, -0.00000778, -0.00003549, 0.00000675, 0.00007738, -0.00014190, -0.00013580, 0.00010356, -0.00011869, -0.00022864, 0.00020166, -0.00013121, -0.00117791, 0.00165230, -0.00413042, -0.00116676, 0.01081045, -0.00116676, -0.00413042, 0.00165230, -0.00117791, -0.00013121, 0.00020166, -0.00022864, -0.00011869, 0.00010356, -0.00013580, -0.00014190, 0.00007738, 0.00000675, -0.00003549, -0.00000778, },
		{ -0.00001901, -0.00001982, 0.00001372, -0.00002436, -0.00006432, 0.00006944, -0.00001786, -0.00018112, 0.00000916, 0.00002914, -0.00028099, -0.00005129, 0.00054701, -0.00176018, 0.00154319, -0.00214528, -0.00426228, 0.00970957, 0.00259622, -0.00500139, 0.00077805, -0.00034434, -0.00055190, 0.00011974, -0.00004253, -0.00021573, 0.00007581, -0.00002289, -0.00020429, 0.00000208, 0.00005264, 0.00000906, },
		{ 0.00006348, 0.00002269, -0.00005279, -0.00000585, 0.00002622, -0.00008535, -0.00002333, 0.00006858, -0.00009600, -0.00013675, 0.00006057, -0.00007476, -0.00052749, 0.00107852, -0.00167392, 0.00054233, 0.00003875, -0.00577837, 0.00685060, 0.00578577, -0.00432842, -0.00061264, 0.00019981, -0.00053781, -0.00015304, 0.00007718, -0.00018222, -0.00002368, 0.00005774, -0.00017456, -0.00008733, 0.00002734, },
		{ -0.00004699, 0.00006543, 0.00010419, -0.00004087, -0.00007047, 0.00005157, -0.00002379, -0.00012116, 0.00003233, 0.00006752, -0.00017741, -0.00013356, 0.00030871, -0.00091442, 0.00111177, -0.00090520, -0.00077992, 0.00153315, -0.00554996, 0.00332231, 0.00744463, -0.00240568, -0.00177287, 0.00015077, -0.00023855, -0.00036040, 0.00002773, -0.00007390, -0.00009640, 0.00006096, -0.00010035, -0.00012389, },
		{ -0.00009952, -0.00010616, -0.00000350, 0.00016278, 0.00004264, -0.00012528, 0.00000126, 0.00007217, -0.00012697, -0.00011014, 0.00016046, 0.00000029, -0.00048409, 0.00061372, -0.00092007, 0.00059825, 0.00015626, -0.00174707, 0.00193078, -0.00413984, 0.00028820, 0.00733359, -0.00006455, -0.00212484, -0.00037488, 0.00001511, -0.00034787, -0.00012686, -0.00000585, -0.00009563, 0.00002034, -0.00004745, },
		{ -0.00003629, -0.00005948, -0.00010637, -0.00009871, 0.00014107, 0.00015905, -0.00010302, -0.00010574, 0.00010975, -0.00001798, -0.00023320, 0.00005291, 0.00034687, -0.00076227, 0.00059336, -0.00049112, -0.00017458, 0.00098593, -0.00197707, 0.00142803, -0.00243340, -0.00155434, 0.00594120, 0.00180346, -0.00157703, -0.00095452, -0.00002300, -0.00017209, -0.00023615, -0.00003514, -0.00005950, -0.00000538, },
		{ 0.00000003, -0.00003527, -0.00004745, -0.00006237, -0.00015215, 0.00004084, 0.00022841, 0.00001147, -0.00018691, 0.00003565, 0.00011851, -0.00019590, -0.00024205, 0.00065342, -0.00073959, 0.00020244, 0.00014332, -0.00078690, 0.00125820, -0.00154698, 0.00058258, -0.00113727, -0.00217783, 0.00410961, 0.00269788, -0.00051354, -0.00116993, -0.00033799, -0.00002739, -0.00020909, -0.00011300, -0.00005548, },
		{ -0.00010071, -0.00000139, -0.00001011, -0.00006017, -0.00002680, -0.00013333, -0.00007364, 0.00019416, 0.00015727, -0.00016313, -0.00010549, 0.00014628, 0.00003244, -0.00054741, 0.00069260, -0.00036766, -0.00035587, 0.00063545, -0.00096624, 0.00100911, -0.00084741, -0.00008235, -0.00049387, -0.00203340, 0.00255331, 0.00266625, 0.00050077, -0.00087855, -0.00069200, -0.00006681, -0.00009163, -0.00014443, },
		{ -0.00008991, -0.00014158, -0.00004908, 0.00002787, -0.00005709, -0.00003354, -0.00007501, -0.00012748, 0.00007265, 0.00023901, -0.00003096, -0.00019512, -0.00000388, 0.00032246, -0.00063788, 0.00039441, 0.00016904, -0.00077095, 0.00076000, -0.00073627, 0.00053902, -0.00027538, -0.00034980, -0.00033305, -0.00166232, 0.00157117, 0.00212534, 0.00106874, -0.00026226, -0.00082488, -0.00027522, -0.00001278, },
		{ -0.00005750, -0.00000758, -0.00011355, -0.00012584, 0.00002908, -0.00001337, -0.00006241, -0.00003982, -0.00009717, -0.00005652, 0.00020139, 0.00012436, -0.00013490, -0.00025996, 0.00046422, -0.00041119, -0.00008530, 0.00057678, -0.00084306, 0.00054793, -0.00034314, 0.00016402, -0.00000148, -0.00031465, -0.00034039, -0.00139098, 0.00107182, 0.00152804, 0.00113094, 0.00034649, -0.00064589, -0.00049510, },
		{ -0.00056846, -0.00018734, 0.00000581, -0.00001949, -0.00016076, -0.00003580, 0.00004130, -0.00006153, -0.00005912, -0.00002825, -0.00011260, 0.00007409, 0.00019116, 0.00006487, -0.00044540, 0.00032353, 0.00002068, -0.00046633, 0.00065977, -0.00060915, 0.00022294, -0.00004430, 0.00001747, 0.00004973, -0.00019288, -0.00030862, -0.00126751, 0.00080467, 0.00111948, 0.00088827, 0.00070076, -0.00027172, },
		{ 0.00009375, -0.00046339, -0.00028023, -0.00008289, 0.00005682, -0.00009859, -0.00012703, 0.00004379, -0.00000683, -0.00010187, 0.00000700, -0.00007592, -0.00005184, 0.00012007, 0.00027948, -0.00040677, -0.00004631, 0.00039937, -0.00054255, 0.00044914, -0.00028545, 0.00001100, 0.00006355, 0.00003066, 0.00004903, -0.00012321, -0.00021393, -0.00119493, 0.00058431, 0.00090624, 0.00059286, 0.00075994, },
		{ 0.00063900, 0.00031667, -0.00027969, -0.00024790, -0.00019926, 0.00002449, 0.00003036, -0.00015498, -0.00003512, 0.00006090, -0.00010412, -0.00002609, 0.00000195, -0.00009685, -0.00003469, 0.00035746, -0.00013468, -0.00043012, 0.00050123, -0.00032430, 0.00015691, -0.00008641, -0.00001552, 0.00005798, 0.00006224, 0.00009930, -0.00010388, -0.00012927, -0.00109526, 0.00036589, 0.00078315, 0.00038586, },
		{ 0.00027684, 0.00047193, 0.00039720, -0.00014354, -0.00011849, -0.00022332, -0.00011888, 0.00011574, -0.00006198, -0.00014189, 0.00006349, -0.00003356, -0.00009529, 0.00004693, -0.00005367, -0.00015690, 0.00023496, 0.00021968, -0.00059948, 0.00029378, -0.00001586, -0.00000038, -0.00007283, 0.00005445, 0.00007060, 0.00003855, 0.00018603, -0.00006414, -0.00010287, -0.00097365, 0.00018287, 0.00066894, },
		{ 0.00054810, 0.00022331, 0.00032006, 0.00041061, -0.00009779, -0.00000731, -0.00010381, -0.00026479, 0.00005809, 0.00010507, -0.00016614, -0.00004043, 0.00007125, -0.00013308, 0.00002618, 0.00001398, -0.00015959, -0.00001430, 0.00044855, -0.00047219, -0.00005141, 0.00014897, 0.00003845, -0.00014490, 0.00008990, 0.00016008, -0.00000907, 0.00024386, 0.00003605, -0.00011417, -0.00087618, 0.00005793, },
		{ -0.00002376, 0.00043077, 0.00020291, 0.00018291, 0.00040738, -0.00008740, -0.00000526, 0.00008391, -0.00027871, -0.00012905, 0.00020726, -0.00003221, -0.00020099, 0.00012677, -0.00010006, -0.00002862, 0.00003134, -0.00005112, -0.00023368, 0.00043300, -0.00016151, -0.00029480, 0.00007933, 0.00015980, -0.00016563, 0.00004094, 0.00028498, -0.00001075, 0.00024383, 0.00017490, -0.00011480, -0.00082643, },
		},
		{ { -0.00007265, -0.00008128, -0.00009401, -0.00000824, 0.00001683, -0.00011477, -0.00002396, 0.00000704, -0.00020500, 0.00010532, 0.00042439, -0.00003107, -0.00021104, -0.00001114, -0.00017695, 0.00004783, 0.00043950, 0.00004783, -0.00017695, -0.00001114, -0.00021104, -0.00003107, 0.00042439, 0.00010532, -0.00020500, 0.00000704, -0.00002396, -0.00011477, 0.00001683, -0.00000824, -0.00009401, -0.00008128, },
		{ -0.00008822, -0.00009826, -0.00007550, -0.00007181, -0.00006091, -0.00001452, -0.00011791, -0.00006277, 0.00017870, -0.00007219, -0.00022779, 0.00024053, 0.00018949, -0.00017266, 0.00007757, -0.00000408, -0.00034759, 0.00007652, 0.00035411, 0.00009959, 0.00006233, -0.00015984, -0.00032855, 0.00021473, 0.00043061, 0.00002371, -0.00008638, -0.00010606, -0.00021173, -0.00003562, 0.00008577, -0.00002328, },
		{ 0.00006698, -0.00000023, -0.00007152, -0.00012148, -0.00008294, -0.00005599, -0.00000441, -0.00008340, -0.00020676, 0.00010299, 0.00021788, -0.00020432, -0.00007370, 0.00022852, -0.00014482, -0.00009221, 0.00030228, -0.00008580, -0.00033081, 0.00004585, 0.00019343, 0.00026975, 0.00018087, -0.00030178, -0.00019384, 0.00035637, 0.00032111, 0.00007105, -0.00008909, -0.00032138, -0.00023103, 0.00004570, },
		{ -0.00012748, 0.00001468, 0.00003855, 0.00005344, -0.00007005, -0.00012746, -0.00010086, -0.00002827, 0.00003854, -0.00017965, -0.00017772, 0.00023624, 0.00009110, -0.00020600, 0.00010757, 0.00001289, -0.00033859, 0.00009262, 0.00032451, -0.00009481, -0.00020212, -0.00008184, 0.00012377, 0.00042516, 0.00013972, -0.00025604, 0.00004962, 0.00035224, 0.00027476, 0.00008072, -0.00028578, -0.00040836, },
		{ -0.00048416, -0.00026828, -0.00007733, -0.00008968, 0.00005511, 0.00007251, -0.00005382, -0.00010716, -0.00014819, 0.00001119, 0.00005800, -0.00023719, -0.00008099, 0.00023336, -0.00003994, -0.00007244, 0.00012472, -0.00023532, -0.00030258, 0.00016910, 0.00021794, 0.00002098, -0.00011030, -0.00018035, 0.00018278, 0.00042756, 0.00008083, -0.00002766, 0.00024005, 0.00035809, 0.00026636, -0.00014621, },
		{ 0.00006859, -0.00054310, -0.00044107, -0.00006104, -0.00015406, -0.00013707, 0.00005325, 0.00003778, 0.00003981, -0.00010505, -0.00023126, 0.00007619, 0.00007217, -0.00024453, -0.00002135, 0.00014710, -0.00004855, 0.00002833, -0.00002755, -0.00030947, -0.00016974, 0.00006505, 0.00012885, 0.00014448, -0.00008835, -0.00011881, 0.00026580, 0.00031507, 0.00012742, 0.00021853, 0.00040687, 0.00046736, },
		{ 0.00067665, 0.00049558, -0.00042605, -0.00077883, -0.00017698, -0.00001627, -0.00021972, -0.00009789, -0.00007654, 0.00003988, 0.00021249, -0.00019412, -0.00030730, 0.00018622, 0.00008045, -0.00025369, -0.00002834, 0.00010821, 0.00001674, -0.00003414, -0.00017368, -0.00018736, -0.00011406, -0.00010649, 0.00008566, 0.00014599, -0.00004774, 0.00007891, 0.00030561, 0.00025183, 0.00023331, 0.00040269, },
		{ 0.00030862, 0.00064441, 0.00096082, 0.00014451, -0.00094179, -0.00061395, -0.00000394, -0.00013882, -0.00009075, -0.00011953, -0.00032366, 0.00019807, 0.00035407, -0.00039022, -0.00029494, 0.00032386, 0.00004280, -0.00029915, -0.00002900, 0.00015908, 0.00001182, -0.00020100, -0.00016780, -0.00002335, -0.00016031, -0.00021570, -0.00000028, 0.00003006, 0.00005675, 0.00028803, 0.00038287, 0.00031899, },
		{ 0.00060062, 0.00046081, 0.00045835, 0.00102513, 0.00085443, -0.00057237, -0.00099141, -0.00022542, -0.00019349, -0.00016649, 0.00020696, -0.00036775, -0.00050647, 0.00050571, 0.00032090, -0.00059752, -0.00014746, 0.00041418, -0.00007086, -0.00033615, 0.00004470, 0.00020605, -0.00011556, -0.00030378, -0.00004388, 0.00004615, -0.00020603, -0.00025713, -0.00017688, -0.00011718, 0.00013869, 0.00045839, },
		{ 0.00009629, 0.00064018, 0.00090868, 0.00070436, 0.00094843, 0.00125279, -0.00000515, -0.00114590, -0.00037626, -0.00009591, -0.00064135, 0.00014189, 0.00040683, -0.00080138, -0.00042905, 0.00078407, 0.00008260, -0.00066668, 0.00007859, 0.00037538, -0.00021762, -0.00028758, 0.00015322, 0.00014270, -0.00026295, -0.00025847, 0.00005806, 0.00002851, -0.00018569, -0.00029534, -0.00037466, -0.00027864, },
		{ -0.00051266, -0.00051647, -0.00004956, 0.00080417, 0.00118494, 0.00135299, 0.00181230, 0.00071567, -0.00143094, -0.00094141, 0.00051615, -0.00053314, -0.00092252, 0.00070106, 0.00026052, -0.00115090, -0.00009249, 0.00082360, -0.00022103, -0.00052835, 0.00024605, 0.00020474, -0.00029642, -0.00014766, 0.00018503, -0.00001568, -0.00029958, -0.00013490, 0.00006601, 0.00000094, -0.00014105, -0.00032989, },
		{ -0.00012315, -0.00029473, -0.00061044, -0.00081993, -0.00023647, 0.00074021, 0.00154875, 0.00277693, 0.00252246, -0.00096240, -0.00262371, 0.00017519, 0.00098977, -0.00131849, -0.00070753, 0.00113039, -0.00018681, -0.00115318, 0.00027615, 0.00058016, -0.00039064, -0.00026802, 0.00024975, 0.00000304, -0.00024407, -0.00000958, 0.00008808, -0.00015156, -0.00021243, -0.00004098, 0.00002473, -0.00002258, },
		{ -0.00001216, -0.00008739, -0.00010058, -0.00019105, -0.00071169, -0.00104307, -0.00058614, 0.00029128, 0.00248584, 0.00488212, 0.00222596, -0.00337135, -0.00275251, 0.00160088, 0.00066148, -0.00187457, -0.00010314, 0.00110781, -0.00061285, -0.00080249, 0.00042315, 0.00021947, -0.00034309, -0.00005550, 0.00011223, -0.00010652, -0.00010745, 0.00002757, -0.00006689, -0.00018480, -0.00009540, 0.00000438, },
		{ -0.00001894, 0.00004671, -0.00006424, -0.00018181, -0.00003415, -0.00012861, -0.00071266, -0.00110613, -0.00141639, -0.00026124, 0.00446076, 0.00683103, 0.00048356, -0.00540469, -0.00168323, 0.00244419, -0.00019190, -0.00179311, 0.00044438, 0.00067409, -0.00071628, -0.00035193, 0.00029168, -0.00003906, -0.00017299, 0.00000629, -0.00004014, -0.00009016, -0.00000094, -0.00003448, -0.00016881, -0.00015137, },
		{ -0.00018782, -0.00012956, 0.00003338, 0.00008230, -0.00015262, -0.00021309, -0.00001436, -0.00015550, -0.00038939, -0.00125053, -0.00286648, -0.00017073, 0.00727857, 0.00747530, -0.00216704, -0.00603590, -0.00013829, 0.00225521, -0.00090659, -0.00118553, 0.00057308, 0.00017381, -0.00048147, -0.00007165, 0.00004149, -0.00011298, -0.00004123, -0.00004057, -0.00010038, -0.00002602, 0.00001075, -0.00010294, },
		{ -0.00002939, -0.00010856, -0.00018145, -0.00012654, 0.00008147, 0.00004528, -0.00018956, -0.00015598, -0.00021866, -0.00008091, 0.00029575, -0.00194825, -0.00447506, 0.00121161, 0.00988037, 0.00625034, -0.00444990, -0.00510774, 0.00089364, 0.00134129, -0.00103232, -0.00052346, 0.00030928, -0.00008143, -0.00014650, -0.00002641, -0.00013463, -0.00008798, -0.00001700, -0.00007948, -0.00007478, -0.00001050, },
		{ -0.00004963, -0.00005296, -0.00004060, -0.00008110, -0.00018389, -0.00008870, 0.00005481, -0.00006208, -0.00003237, -0.00017999, -0.00066131, 0.00040523, 0.00093916, -0.00340185, -0.00526560, 0.00372041, 0.01094712, 0.00372041, -0.00526560, -0.00340185, 0.00093916, 0.00040523, -0.00066131, -0.00017999, -0.00003237, -0.00006208, 0.00005481, -0.00008870, -0.00018389, -0.00008110, -0.00004060, -0.00005296, },
		{ -0.00002939, -0.00001050, -0.00007478, -0.00007948, -0.00001700, -0.00008798, -0.00013463, -0.00002641, -0.00014650, -0.00008143, 0.00030928, -0.00052346, -0.00103232, 0.00134129, 0.00089364, -0.00510774, -0.00444990, 0.00625034, 0.00988037, 0.00121161, -0.00447506, -0.00194825, 0.00029575, -0.00008091, -0.00021866, -0.00015598, -0.00018956, 0.00004528, 0.00008147, -0.00012654, -0.00018145, -0.00010856, },
		{ -0.00018782, -0.00010294, 0.00001075, -0.00002602, -0.00010038, -0.00004057, -0.00004123, -0.00011298, 0.00004149, -0.00007165, -0.00048147, 0.00017381, 0.00057308, -0.00118553, -0.00090659, 0.00225521, -0.00013829, -0.00603590, -0.00216704, 0.00747530, 0.00727857, -0.00017073, -0.00286648, -0.00125053, -0.00038939, -0.00015550, -0.00001436, -0.00021309, -0.00015262, 0.00008230, 0.00003338, -0.00012956, },
		{ -0.00001894, -0.00015137, -0.00016881, -0.00003448, -0.00000094, -0.00009016, -0.00004014, 0.00000629, -0.00017299, -0.00003906, 0.00029168, -0.00035193, -0.00071628, 0.00067409, 0.00044438, -0.00179311, -0.00019190, 0.00244419, -0.00168323, -0.00540469, 0.00048356, 0.00683103, 0.00446076, -0.00026124, -0.00141639, -0.00110613, -0.00071266, -0.00012861, -0.00003415, -0.00018181, -0.00006424, 0.00004671, },
		{ -0.00001216, 0.00000438, -0.00009540, -0.00018480, -0.00006689, 0.00002757, -0.00010745, -0.00010652, 0.00011223, -0.00005550, -0.00034309, 0.00021947, 0.00042315, -0.00080249, -0.00061285, 0.00110781, -0.00010314, -0.00187457, 0.00066148, 0.00160088, -0.00275251, -0.00337135, 0.00222596, 0.00488212, 0.00248584, 0.00029128, -0.00058614, -0.00104307, -0.00071169, -0.00019105, -0.00010058, -0.00008739, },
		{ -0.00012315, -0.00002258, 0.00002473, -0.00004098, -0.00021243, -0.00015156, 0.00008808, -0.00000958, -0.00024407, 0.00000304, 0.00024975, -0.00026802, -0.00039064, 0.00058016, 0.00027615, -0.00115318, -0.00018681, 0.00113039, -0.00070753, -0.00131849, 0.00098977, 0.00017519, -0.00262371, -0.00096240, 0.00252246, 0.00277693, 0.00154875, 0.00074021, -0.00023647, -0.00081993, -0.00061044, -0.00029473, },
		{ -0.00051266, -0.00032989, -0.00014105, 0.00000094, 0.00006601, -0.00013490, -0.00029958, -0.00001568, 0.00018503, -0.00014766, -0.00029642, 0.00020474, 0.00024605, -0.00052835, -0.00022103, 0.00082360, -0.00009249, -0.00115090, 0.00026052, 0.00070106, -0.00092252, -0.00053314, 0.00051615, -0.00094141, -0.00143094, 0.00071567, 0.00181230, 0.00135299, 0.00118494, 0.00080417, -0.00004956, -0.00051647, },
		{ 0.00009629, -0.00027864, -0.00037466, -0.00029534, -0.00018569, 0.00002851, 0.00005806, -0.00025847, -0.00026295, 0.00014270, 0.00015322, -0.00028758, -0.00021762, 0.00037538, 0.00007859, -0.00066668, 0.00008260, 0.00078407, -0.00042905, -0.00080138, 0.00040683, 0.00014189, -0.00064135, -0.00009591, -0.00037626, -0.00114590, -0.00000515, 0.00125279, 0.00094843, 0.00070436, 0.00090868, 0.00064018, },
		{ 0.00060062, 0.00045839, 0.00013869, -0.00011718, -0.00017688, -0.00025713, -0.00020603, 0.00004615, -0.00004388, -0.00030378, -0.00011556, 0.00020605, 0.00004470, -0.00033615, -0.00007086, 0.00041418, -0.00014746, -0.00059752, 0.00032090, 0.00050571, -0.00050647, -0.00036775, 0.00020696, -0.00016649, -0.00019349, -0.00022542, -0.00099141, -0.00057237, 0.00085443, 0.00102513, 0.00045835, 0.00046081, },
		{ 0.00030862, 0.00031899, 0.00038287, 0.00028803, 0.00005675, 0.00003006, -0.00000028, -0.00021570, -0.00016031, -0.00002335, -0.00016780, -0.00020100, 0.00001182, 0.00015908, -0.00002900, -0.00029915, 0.00004280, 0.00032386, -0.00029494, -0.00039022, 0.00035407, 0.00019807, -0.00032366, -0.00011953, -0.00009075, -0.00013882, -0.00000394, -0.00061395, -0.00094179, 0.00014451, 0.00096082, 0.00064441, },
		{ 0.00067665, 0.00040269, 0.00023331, 0.00025183, 0.00030561, 0.00007891, -0.00004774, 0.00014599, 0.00008566, -0.00010649, -0.00011406, -0.00018736, -0.00017368, -0.00003414, 0.00001674, 0.00010821, -0.00002834, -0.00025369, 0.00008045, 0.00018622, -0.00030730, -0.00019412, 0.00021249, 0.00003988, -0.00007654, -0.00009789, -0.00021972, -0.00001627, -0.00017698, -0.00077883, -0.00042605, 0.00049558, },
		{ 0.00006859, 0.00046736, 0.00040687, 0.00021853, 0.00012742, 0.00031507, 0.00026580, -0.00011881, -0.00008835, 0.00014448, 0.00012885, 0.00006505, -0.00016974, -0.00030947, -0.00002755, 0.00002833, -0.00004855, 0.00014710, -0.00002135, -0.00024453, 0.00007217, 0.00007619, -0.00023126, -0.00010505, 0.00003981, 0.00003778, 0.00005325, -0.00013707, -0.00015406, -0.00006104, -0.00044107, -0.00054310, },
		{ -0.00048416, -0.00014621, 0.00026636, 0.00035809, 0.00024005, -0.00002766, 0.00008083, 0.00042756, 0.00018278, -0.00018035, -0.00011030, 0.00002098, 0.00021794, 0.00016910, -0.00030258, -0.00023532, 0.00012472, -0.00007244, -0.00003994, 0.00023336, -0.00008099, -0.00023719, 0.00005800, 0.00001119, -0.00014819, -0.00010716, -0.00005382, 0.00007251, 0.00005511, -0.00008968, -0.00007733, -0.00026828, },
		{ -0.00012748, -0.00040836, -0.00028578, 0.00008072, 0.00027476, 0.00035224, 0.00004962, -0.00025604, 0.00013972, 0.00042516, 0.00012377, -0.00008184, -0.00020212, -0.00009481, 0.00032451, 0.00009262, -0.00033859, 0.00001289, 0.00010757, -0.00020600, 0.00009110, 0.00023624, -0.00017772, -0.00017965, 0.00003854, -0.00002827, -0.00010086, -0.00012746, -0.00007005, 0.00005344, 0.00003855, 0.00001468, },
		{ 0.00006698, 0.00004570, -0.00023103, -0.00032138, -0.00008909, 0.00007105, 0.00032111, 0.00035637, -0.00019384, -0.00030178, 0.00018087, 0.00026975, 0.00019343, 0.00004585, -0.00033081, -0.00008580, 0.00030228, -0.00009221, -0.00014482, 0.00022852, -0.00007370, -0.00020432, 0.00021788, 0.00010299, -0.00020676, -0.00008340, -0.00000441, -0.00005599, -0.00008294, -0.00012148, -0.00007152, -0.00000023, },
		{ -0.00008822, -0.00002328, 0.00008577, -0.00003562, -0.00021173, -0.00010606, -0.00008638, 0.00002371, 0.00043061, 0.00021473, -0.00032855, -0.00015984, 0.00006233, 0.00009959, 0.00035411, 0.00007652, -0.00034759, -0.00000408, 0.00007757, -0.00017266, 0.00018949, 0.00024053, -0.00022779, -0.00007219, 0.00017870, -0.00006277, -0.00011791, -0.00001452, -0.00006091, -0.00007181, -0.00007550, -0.00009826, },
		},
		{ { -0.00004532, -0.00004693, -0.00005682, -0.00008985, -0.00014834, -0.00018322, -0.00012043, 0.00003034, 0.00013307, 0.00007047, -0.00007655, -0.00009323, 0.00008360, 0.00023683, 0.00012432, -0.00018335, -0.00035030, -0.00018335, 0.00012432, 0.00023683, 0.00008360, -0.00009323, -0.00007655, 0.00007047, 0.00013307, 0.00003034, -0.00012043, -0.00018322, -0.00014834, -0.00008985, -0.00005682, -0.00004693, },
		{ -0.00003855, -0.00001445, -0.00002335, -0.00003633, -0.00002780, -0.00003009, -0.00009632, -0.00019331, -0.00020181, -0.00007989, 0.00003473, -0.00001419, -0.00015797, -0.00014897, 0.00010140, 0.00034499, 0.00027476, -0.00007103, -0.00032521, -0.00023310, 0.00004963, 0.00017786, 0.00004412, -0.00013656, -0.00013864, 0.00001060, 0.00011953, 0.00008395, -0.00003158, -0.00011599, -0.00012687, -0.00008742, },
		{ -0.00013085, -0.00011119, -0.00004980, -0.00001086, -0.00003399, -0.00006860, -0.00003936, 0.00002468, 0.00001319, -0.00008428, -0.00012161, -0.00000764, 0.00010558, -0.00000641, -0.00028976, -0.00040110, -0.00013387, 0.00026763, 0.00037022, 0.00008239, -0.00024052, -0.00024881, -0.00000578, 0.00015299, 0.00006236, -0.00011765, -0.00016403, -0.00005929, 0.00005059, 0.00006180, -0.00000738, -0.00008868, },
		{ 0.00000398, -0.00005615, -0.00011782, -0.00011717, -0.00004031, 0.00002086, -0.00002594, -0.00012516, -0.00013515, -0.00004921, -0.00002596, -0.00013075, -0.00017648, 0.00002572, 0.00032359, 0.00033848, -0.00003188, -0.00041090, -0.00038137, -0.00001236, 0.00023587, 0.00010100, -0.00018619, -0.00025381, -0.00006371, 0.00010786, 0.00007015, -0.00007897, -0.00014804, -0.00009489, -0.00001218, 0.00002402, },
		{ -0.00006697, -0.00004197, -0.00002662, -0.00005789, -0.00013237, -0.00015015, -0.00004077, 0.00008951, 0.00007110, -0.00006905, -0.00011882, -0.00000897, 0.00003727, -0.00017018, -0.00043124, -0.00035376, 0.00009319, 0.00045899, 0.00035813, -0.00003825, -0.00024337, -0.00009027, 0.00010678, 0.00003476, -0.00020584, -0.00029615, -0.00014859, 0.00002620, 0.00004504, -0.00004320, -0.00010084, -0.00009355, },
		{ -0.00003741, -0.00002836, -0.00005489, -0.00006954, -0.00004204, -0.00005305, -0.00015909, -0.00023608, -0.00012710, 0.00007772, 0.00010993, -0.00007642, -0.00017781, 0.00003408, 0.00031871, 0.00024008, -0.00021168, -0.00054246, -0.00038008, 0.00005005, 0.00025222, 0.00010727, -0.00005716, 0.00000412, 0.00013186, 0.00006275, -0.00016500, -0.00029256, -0.00021748, -0.00008740, -0.00004187, -0.00004912, },
		{ -0.00017724, -0.00014789, -0.00004165, 0.00002998, -0.00000870, -0.00005438, -0.00000716, 0.00002322, -0.00011108, -0.00028487, -0.00021176, 0.00009581, 0.00023853, -0.00003528, -0.00040467, -0.00034680, 0.00014656, 0.00051934, 0.00036839, -0.00006117, -0.00025885, -0.00011811, 0.00003387, -0.00002119, -0.00012584, -0.00007635, 0.00004345, 0.00002559, -0.00012546, -0.00022454, -0.00020256, -0.00016753, },
		{ -0.00012844, -0.00016031, -0.00029025, -0.00031653, -0.00015917, -0.00001065, -0.00003568, -0.00009742, 0.00000436, 0.00014427, 0.00001726, -0.00032564, -0.00041550, -0.00000590, 0.00047905, 0.00042727, -0.00014326, -0.00057543, -0.00040744, 0.00009067, 0.00035832, 0.00026554, 0.00013279, 0.00017131, 0.00023140, 0.00013311, -0.00003945, -0.00011580, -0.00011509, -0.00015801, -0.00022367, -0.00020124, },
		{ -0.00040641, -0.00025110, -0.00001101, 0.00000275, -0.00022919, -0.00037701, -0.00024576, -0.00007668, -0.00015609, -0.00031827, -0.00017340, 0.00022210, 0.00031995, -0.00014477, -0.00066057, -0.00053677, 0.00016904, 0.00068893, 0.00051010, -0.00004567, -0.00033906, -0.00021019, 0.00000376, 0.00006365, 0.00009135, 0.00023924, 0.00040800, 0.00042376, 0.00029146, 0.00011422, -0.00009007, -0.00031210, },
		{ 0.00008568, -0.00041259, -0.00070530, -0.00051540, -0.00005825, 0.00014342, -0.00006317, -0.00025357, -0.00007261, 0.00018743, 0.00000319, -0.00052024, -0.00066902, -0.00011318, 0.00052956, 0.00044463, -0.00029790, -0.00082143, -0.00055888, 0.00011340, 0.00047692, 0.00035566, 0.00012768, 0.00007575, 0.00009857, 0.00004584, 0.00000328, 0.00011564, 0.00033867, 0.00051690, 0.00056568, 0.00044182, },
		{ 0.00078882, 0.00085746, 0.00048961, -0.00028077, -0.00092716, -0.00090540, -0.00036041, -0.00002346, -0.00025395, -0.00051613, -0.00014084, 0.00061801, 0.00079399, 0.00001904, -0.00087041, -0.00084494, 0.00001744, 0.00067538, 0.00045547, -0.00022673, -0.00056570, -0.00036303, -0.00003848, 0.00008532, 0.00010498, 0.00018866, 0.00027428, 0.00024057, 0.00014801, 0.00015045, 0.00029720, 0.00053844, },
		{ 0.00026726, 0.00044884, 0.00082152, 0.00114559, 0.00094287, 0.00006210, -0.00087410, -0.00100208, -0.00031498, 0.00017237, -0.00032361, -0.00122734, -0.00123453, -0.00002742, 0.00119288, 0.00113902, -0.00000921, -0.00091481, -0.00073756, 0.00004074, 0.00047065, 0.00029955, -0.00001091, -0.00008985, -0.00002937, -0.00001958, -0.00001802, 0.00008875, 0.00024513, 0.00031679, 0.00028762, 0.00024299, },
		{ 0.00060645, 0.00063536, 0.00058733, 0.00061746, 0.00092228, 0.00131951, 0.00114677, 0.00000899, -0.00138382, -0.00166041, -0.00039925, 0.00107043, 0.00101712, -0.00061973, -0.00205572, -0.00172908, -0.00007161, 0.00115244, 0.00094353, -0.00004220, -0.00061617, -0.00045145, -0.00008283, 0.00004193, 0.00000803, 0.00003783, 0.00011132, 0.00011512, 0.00009232, 0.00016104, 0.00032346, 0.00049194, },
		{ 0.00005347, 0.00030892, 0.00066549, 0.00102167, 0.00125218, 0.00147089, 0.00198018, 0.00264460, 0.00251984, 0.00072475, -0.00210424, -0.00379614, -0.00272112, 0.00022001, 0.00227747, 0.00170168, -0.00051307, -0.00197393, -0.00154896, -0.00020301, 0.00056198, 0.00037768, -0.00007570, -0.00023563, -0.00018609, -0.00019522, -0.00026010, -0.00024799, -0.00017458, -0.00014050, -0.00014153, -0.00009258, },
		{ -0.00043758, -0.00054896, -0.00064862, -0.00063288, -0.00036994, 0.00009931, 0.00060239, 0.00126010, 0.00257640, 0.00458469, 0.00595398, 0.00472132, 0.00053067, -0.00417160, -0.00597385, -0.00380059, 0.00001304, 0.00217339, 0.00159927, -0.00013565, -0.00106490, -0.00079788, -0.00022303, -0.00003061, -0.00010814, -0.00011887, -0.00005178, -0.00006801, -0.00017112, -0.00025166, -0.00028994, -0.00034526, },
		{ -0.00011967, -0.00014663, -0.00019743, -0.00033376, -0.00063134, -0.00100826, -0.00124815, -0.00132530, -0.00144522, -0.00141443, -0.00021280, 0.00303322, 0.00729971, 0.00953126, 0.00720198, 0.00118830, -0.00450569, -0.00630428, -0.00415495, -0.00093857, 0.00073503, 0.00061417, -0.00000106, -0.00019664, -0.00009872, -0.00010568, -0.00022118, -0.00024172, -0.00015153, -0.00008356, -0.00008385, -0.00010296, },
		{ -0.00013451, -0.00013756, -0.00014836, -0.00013264, -0.00003982, 0.00005474, -0.00002959, -0.00029950, -0.00054308, -0.00080545, -0.00163283, -0.00323405, -0.00434076, -0.00271878, 0.00237184, 0.00839537, 0.01110393, 0.00839537, 0.00237184, -0.00271878, -0.00434076, -0.00323405, -0.00163283, -0.00080545, -0.00054308, -0.00029950, -0.00002959, 0.00005474, -0.00003982, -0.00013264, -0.00014836, -0.00013756, },
		{ -0.00011967, -0.00010296, -0.00008385, -0.00008356, -0.00015153, -0.00024172, -0.00022118, -0.00010568, -0.00009872, -0.00019664, -0.00000106, 0.00061417, 0.00073503, -0.00093857, -0.00415495, -0.00630428, -0.00450569, 0.00118830, 0.00720198, 0.00953126, 0.00729971, 0.00303322, -0.00021280, -0.00141443, -0.00144522, -0.00132530, -0.00124815, -0.00100826, -0.00063134, -0.00033376, -0.00019743, -0.00014663, },
		{ -0.00043758, -0.00034526, -0.00028994, -0.00025166, -0.00017112, -0.00006801, -0.00005178, -0.00011887, -0.00010814, -0.00003061, -0.00022303, -0.00079788, -0.00106490, -0.00013565, 0.00159927, 0.00217339, 0.00001304, -0.00380059, -0.00597385, -0.00417160, 0.00053067, 0.00472132, 0.00595398, 0.00458469, 0.00257640, 0.00126010, 0.00060239, 0.00009931, -0.00036994, -0.00063288, -0.00064862, -0.00054896, },
		{ 0.00005347, -0.00009258, -0.00014153, -0.00014050, -0.00017458, -0.00024799, -0.00026010, -0.00019522, -0.00018609, -0.00023563, -0.00007570, 0.00037768, 0.00056198, -0.00020301, -0.00154896, -0.00197393, -0.00051307, 0.00170168, 0.00227747, 0.00022001, -0.00272112, -0.00379614, -0.00210424, 0.00072475, 0.00251984, 0.00264460, 0.00198018, 0.00147089, 0.00125218, 0.00102167, 0.00066549, 0.00030892, },
		{ 0.00060645, 0.00049194, 0.00032346, 0.00016104, 0.00009232, 0.00011512, 0.00011132, 0.00003783, 0.00000803, 0.00004193, -0.00008283, -0.00045145, -0.00061617, -0.00004220, 0.00094353, 0.00115244, -0.00007161, -0.00172908, -0.00205572, -0.00061973, 0.00101712, 0.00107043, -0.00039925, -0.00166041, -0.00138382, 0.00000899, 0.00114677, 0.00131951, 0.00092228, 0.00061746, 0.00058733, 0.00063536, },
		{ 0.00026726, 0.00024299, 0.00028762, 0.00031679, 0.00024513, 0.00008875, -0.00001802, -0.00001958, -0.00002937, -0.00008985, -0.00001091, 0.00029955, 0.00047065, 0.00004074, -0.00073756, -0.00091481, -0.00000921, 0.00113902, 0.00119288, -0.00002742, -0.00123453, -0.00122734, -0.00032361, 0.00017237, -0.00031498, -0.00100208, -0.00087410, 0.00006210, 0.00094287, 0.00114559, 0.00082152, 0.00044884, },
		{ 0.00078882, 0.00053844, 0.00029720, 0.00015045, 0.00014801, 0.00024057, 0.00027428, 0.00018866, 0.00010498, 0.00008532, -0.00003848, -0.00036303, -0.00056570, -0.00022673, 0.00045547, 0.00067538, 0.00001744, -0.00084494, -0.00087041, 0.00001904, 0.00079399, 0.00061801, -0.00014084, -0.00051613, -0.00025395, -0.00002346, -0.00036041, -0.00090540, -0.00092716, -0.00028077, 0.00048961, 0.00085746, },
		{ 0.00008568, 0.00044182, 0.00056568, 0.00051690, 0.00033867, 0.00011564, 0.00000328, 0.00004584, 0.00009857, 0.00007575, 0.00012768, 0.00035566, 0.00047692, 0.00011340, -0.00055888, -0.00082143, -0.00029790, 0.00044463, 0.00052956, -0.00011318, -0.00066902, -0.00052024, 0.00000319, 0.00018743, -0.00007261, -0.00025357, -0.00006317, 0.00014342, -0.00005825, -0.00051540, -0.00070530, -0.00041259, },
		{ -0.00040641, -0.00031210, -0.00009007, 0.00011422, 0.00029146, 0.00042376, 0.00040800, 0.00023924, 0.00009135, 0.00006365, 0.00000376, -0.00021019, -0.00033906, -0.00004567, 0.00051010, 0.00068893, 0.00016904, -0.00053677, -0.00066057, -0.00014477, 0.00031995, 0.00022210, -0.00017340, -0.00031827, -0.00015609, -0.00007668, -0.00024576, -0.00037701, -0.00022919, 0.00000275, -0.00001101, -0.00025110, },
		{ -0.00012844, -0.00020124, -0.00022367, -0.00015801, -0.00011509, -0.00011580, -0.00003945, 0.00013311, 0.00023140, 0.00017131, 0.00013279, 0.00026554, 0.00035832, 0.00009067, -0.00040744, -0.00057543, -0.00014326, 0.00042727, 0.00047905, -0.00000590, -0.00041550, -0.00032564, 0.00001726, 0.00014427, 0.00000436, -0.00009742, -0.00003568, -0.00001065, -0.00015917, -0.00031653, -0.00029025, -0.00016031, },
		{ -0.00017724, -0.00016753, -0.00020256, -0.00022454, -0.00012546, 0.00002559, 0.00004345, -0.00007635, -0.00012584, -0.00002119, 0.00003387, -0.00011811, -0.00025885, -0.00006117, 0.00036839, 0.00051934, 0.00014656, -0.00034680, -0.00040467, -0.00003528, 0.00023853, 0.00009581, -0.00021176, -0.00028487, -0.00011108, 0.00002322, -0.00000716, -0.00005438, -0.00000870, 0.00002998, -0.00004165, -0.00014789, },
		{ -0.00003741, -0.00004912, -0.00004187, -0.00008740, -0.00021748, -0.00029256, -0.00016500, 0.00006275, 0.00013186, 0.00000412, -0.00005716, 0.00010727, 0.00025222, 0.00005005, -0.00038008, -0.00054246, -0.00021168, 0.00024008, 0.00031871, 0.00003408, -0.00017781, -0.00007642, 0.00010993, 0.00007772, -0.00012710, -0.00023608, -0.00015909, -0.00005305, -0.00004204, -0.00006954, -0.00005489, -0.00002836, },
		{ -0.00006697, -0.00009355, -0.00010084, -0.00004320, 0.00004504, 0.00002620, -0.00014859, -0.00029615, -0.00020584, 0.00003476, 0.00010678, -0.00009027, -0.00024337, -0.00003825, 0.00035813, 0.00045899, 0.00009319, -0.00035376, -0.00043124, -0.00017018, 0.00003727, -0.00000897, -0.00011882, -0.00006905, 0.00007110, 0.00008951, -0.00004077, -0.00015015, -0.00013237, -0.00005789, -0.00002662, -0.00004197, },
		{ 0.00000398, 0.00002402, -0.00001218, -0.00009489, -0.00014804, -0.00007897, 0.00007015, 0.00010786, -0.00006371, -0.00025381, -0.00018619, 0.00010100, 0.00023587, -0.00001236, -0.00038137, -0.00041090, -0.00003188, 0.00033848, 0.00032359, 0.00002572, -0.00017648, -0.00013075, -0.00002596, -0.00004921, -0.00013515, -0.00012516, -0.00002594, 0.00002086, -0.00004031, -0.00011717, -0.00011782, -0.00005615, },
		{ -0.00013085, -0.00008868, -0.00000738, 0.00006180, 0.00005059, -0.00005929, -0.00016403, -0.00011765, 0.00006236, 0.00015299, -0.00000578, -0.00024881, -0.00024052, 0.00008239, 0.00037022, 0.00026763, -0.00013387, -0.00040110, -0.00028976, -0.00000641, 0.00010558, -0.00000764, -0.00012161, -0.00008428, 0.00001319, 0.00002468, -0.00003936, -0.00006860, -0.00003399, -0.00001086, -0.00004980, -0.00011119, },
		{ -0.00003855, -0.00008742, -0.00012687, -0.00011599, -0.00003158, 0.00008395, 0.00011953, 0.00001060, -0.00013864, -0.00013656, 0.00004412, 0.00017786, 0.00004963, -0.00023310, -0.00032521, -0.00007103, 0.00027476, 0.00034499, 0.00010140, -0.00014897, -0.00015797, -0.00001419, 0.00003473, -0.00007989, -0.00020181, -0.00019331, -0.00009632, -0.00003009, -0.00002780, -0.00003633, -0.00002335, -0.00001445, },
		},
		{ { -0.00001223, -0.00000885, -0.00000064, 0.00000737, 0.00000902, -0.00000048, -0.00002241, -0.00005360, -0.00008704, -0.00011390, -0.00012638, -0.00012031, -0.00009677, -0.00006210, -0.00002608, 0.00000084, 0.00001079, 0.00000084, -0.00002608, -0.00006210, -0.00009677, -0.00012031, -0.00012638, -0.00011390, -0.00008704, -0.00005360, -0.00002241, -0.00000048, 0.00000902, 0.00000737, -0.00000064, -0.00000885, },
		{ -0.00000667, -0.00000944, -0.00001662, -0.00002225, -0.00002073, -0.00000905, 0.00001178, 0.00003682, 0.00005856, 0.00006943, 0.00006430, 0.00004245, 0.00000803, -0.00003098, -0.00006506, -0.00008607, -0.00008965, -0.00007659, -0.00005243, -0.00002574, -0.00000533, 0.00000248, -0.00000429, -0.00002283, -0.00004663, -0.00006780, -0.00007964, -0.00007885, -0.00006637, -0.00004685, -0.00002682, -0.00001232, },
		{ -0.00000750, 0.00001498, 0.00003579, 0.00004864, 0.00004949, 0.00003791, 0.00001727, -0.00000627, -0.00002565, -0.00003519, -0.00003229, -0.00001823, 0.00000229, 0.00002277, 0.00003700, 0.00004103, 0.00003434, 0.00001986, 0.00000286, -0.00001097, -0.00001743, -0.00001523, -0.00000632, 0.00000492, 0.00001332, 0.00001485, 0.00000814, -0.00000493, -0.00001985, -0.00003095, -0.00003345, -0.00002518, },
		{ -0.00003071, -0.00004917, -0.00005753, -0.00005314, -0.00003686, -0.00001278, 0.00001312, 0.00003467, 0.00004732, 0.00004929, 0.00004183, 0.00002851, 0.00001385, 0.00000173, -0.00000579, -0.00000894, -0.00000985, -0.00001144, -0.00001604, -0.00002433, -0.00003484, -0.00004439, -0.00004917, -0.00004626, -0.00003482, -0.00001674, 0.00000368, 0.00002089, 0.00002979, 0.00002737, 0.00001378, -0.00000753, },
		{ 0.00002243, 0.00002536, 0.00001563, -0.00000345, -0.00002580, -0.00004474, -0.00005524, -0.00005552, -0.00004748, -0.00003574, -0.00002582, -0.00002198, -0.00002557, -0.00003439, -0.00004351, -0.00004697, -0.00004007, -0.00002130, 0.00000682, 0.00003819, 0.00006485, 0.00007948, 0.00007768, 0.00005958, 0.00002983, -0.00000361, -0.00003192, -0.00004804, -0.00004877, -0.00003563, -0.00001423, 0.00000762, },
		{ -0.00002936, -0.00002515, -0.00001006, 0.00000775, 0.00001967, 0.00002022, 0.00000926, -0.00000772, -0.00002155, -0.00002282, -0.00000564, 0.00002968, 0.00007595, 0.00012085, 0.00015060, 0.00015431, 0.00012775, 0.00007507, 0.00000799, -0.00005740, -0.00010525, -0.00012461, -0.00011247, -0.00007451, -0.00002309, 0.00002667, 0.00006147, 0.00007373, 0.00006339, 0.00003729, 0.00000637, -0.00001833, },
		{ 0.00000528, -0.00000699, -0.00002601, -0.00003927, -0.00003718, -0.00001730, 0.00001417, 0.00004434, 0.00005824, 0.00004468, 0.00000108, -0.00006437, -0.00013428, -0.00018691, -0.00020291, -0.00017176, -0.00009573, 0.00001000, 0.00012119, 0.00021128, 0.00025900, 0.00025431, 0.00020111, 0.00011559, 0.00002116, -0.00005874, -0.00010731, -0.00011869, -0.00009858, -0.00006115, -0.00002342, 0.00000078, },
		{ -0.00000261, 0.00003610, 0.00007167, 0.00008662, 0.00007135, 0.00002837, -0.00002798, -0.00007638, -0.00009634, -0.00007609, -0.00001793, 0.00006118, 0.00013411, 0.00017190, 0.00015320, 0.00007187, -0.00005998, -0.00021465, -0.00035635, -0.00045143, -0.00047810, -0.00043250, -0.00032935, -0.00019710, -0.00006901, 0.00002670, 0.00007467, 0.00007569, 0.00004470, 0.00000429, -0.00002381, -0.00002634, },
		{ -0.00006114, -0.00011799, -0.00015771, -0.00016144, -0.00012451, -0.00005923, 0.00000865, 0.00004866, 0.00003668, -0.00003551, -0.00015574, -0.00029355, -0.00040874, -0.00046358, -0.00043488, -0.00032189, -0.00014746, 0.00004809, 0.00021853, 0.00032511, 0.00034750, 0.00028896, 0.00017424, 0.00004093, -0.00007265, -0.00013915, -0.00014964, -0.00011468, -0.00005881, -0.00001071, 0.00000742, -0.00001227, },
		{ -0.00011500, -0.00006087, -0.00003696, -0.00005715, -0.00011570, -0.00018752, -0.00023594, -0.00022547, -0.00013525, 0.00003147, 0.00024542, 0.00045797, 0.00061456, 0.00067083, 0.00060639, 0.00043161, 0.00018538, -0.00007579, -0.00029378, -0.00042516, -0.00045259, -0.00038806, -0.00026716, -0.00013639, -0.00003811, 0.00000221, -0.00001766, -0.00007865, -0.00014926, -0.00019823, -0.00020628, -0.00017274, },
		{ -0.00012612, -0.00013864, -0.00010684, -0.00002562, 0.00008680, 0.00019307, 0.00024863, 0.00021766, 0.00008734, -0.00012450, -0.00037155, -0.00059085, -0.00072076, -0.00071965, -0.00057924, -0.00032827, -0.00002511, 0.00025852, 0.00045770, 0.00053217, 0.00047644, 0.00031925, 0.00011288, -0.00008386, -0.00022351, -0.00028279, -0.00026671, -0.00020310, -0.00012988, -0.00008006, -0.00006967, -0.00009292, },
		{ -0.00011706, -0.00011930, -0.00017910, -0.00029269, -0.00042988, -0.00054255, -0.00058077, -0.00051107, -0.00033061, -0.00007191, 0.00020387, 0.00042365, 0.00052390, 0.00047034, 0.00026928, -0.00003304, -0.00036311, -0.00064110, -0.00080339, -0.00081994, -0.00070113, -0.00049227, -0.00025795, -0.00006149, 0.00005376, 0.00007529, 0.00002159, -0.00006805, -0.00014993, -0.00019255, -0.00018759, -0.00015144, },
		{ -0.00051875, -0.00050242, -0.00040211, -0.00023894, -0.00006219, 0.00006494, 0.00008604, -0.00002820, -0.00026720, -0.00057984, -0.00088683, -0.00110277, -0.00116105, -0.00103404, -0.00074189, -0.00034741, 0.00006165, 0.00039734, 0.00059546, 0.00063168, 0.00052496, 0.00032781, 0.00010731, -0.00007688, -0.00018989, -0.00022927, -0.00022071, -0.00020417, -0.00021591, -0.00027290, -0.00036565, -0.00046170, },
		{ -0.00002385, -0.00027062, -0.00058088, -0.00090121, -0.00115727, -0.00127431, -0.00120112, -0.00092940, -0.00050175, -0.00000483, 0.00045126, 0.00076161, 0.00085487, 0.00071277, 0.00037495, -0.00007241, -0.00052064, -0.00086799, -0.00104562, -0.00103266, -0.00085668, -0.00058021, -0.00027874, -0.00001762, 0.00016463, 0.00026183, 0.00029363, 0.00029066, 0.00027747, 0.00026014, 0.00022331, 0.00013782, },
		{ 0.00017150, 0.00040966, 0.00061961, 0.00073790, 0.00069741, 0.00044970, -0.00001357, -0.00064808, -0.00135860, -0.00201640, -0.00248853, -0.00267111, -0.00251713, -0.00205048, -0.00136159, -0.00058580, 0.00012966, 0.00066078, 0.00093420, 0.00093980, 0.00072583, 0.00037946, -0.00000051, -0.00033004, -0.00055758, -0.00066906, -0.00068031, -0.00062162, -0.00052113, -0.00039371, -0.00023936, -0.00005111, },
		{ 0.00041158, 0.00050522, 0.00062387, 0.00080328, 0.00107155, 0.00143098, 0.00184421, 0.00223123, 0.00248074, 0.00247467, 0.00211996, 0.00137862, 0.00028677, -0.00104413, -0.00244541, -0.00372496, -0.00470785, -0.00527319, -0.00537749, -0.00505870, -0.00442076, -0.00360392, -0.00274991, -0.00197165, -0.00133523, -0.00085745, -0.00051708, -0.00027395, -0.00008832, 0.00006631, 0.00020036, 0.00031463, },
		{ 0.00121347, 0.00123441, 0.00129219, 0.00137621, 0.00148154, 0.00162181, 0.00183695, 0.00219064, 0.00275566, 0.00359007, 0.00471092, 0.00607419, 0.00756878, 0.00902842, 0.01026056, 0.01108550, 0.01137583, 0.01108550, 0.01026056, 0.00902842, 0.00756878, 0.00607419, 0.00471092, 0.00359007, 0.00275566, 0.00219064, 0.00183695, 0.00162181, 0.00148154, 0.00137621, 0.00129219, 0.00123441, },
		{ 0.00041158, 0.00031463, 0.00020036, 0.00006631, -0.00008832, -0.00027395, -0.00051708, -0.00085745, -0.00133523, -0.00197165, -0.00274991, -0.00360392, -0.00442076, -0.00505870, -0.00537749, -0.00527319, -0.00470785, -0.00372496, -0.00244541, -0.00104413, 0.00028677, 0.00137862, 0.00211996, 0.00247467, 0.00248074, 0.00223123, 0.00184421, 0.00143098, 0.00107155, 0.00080328, 0.00062387, 0.00050522, },
		{ 0.00017150, -0.00005111, -0.00023936, -0.00039371, -0.00052113, -0.00062162, -0.00068031, -0.00066906, -0.00055758, -0.00033004, -0.00000051, 0.00037946, 0.00072583, 0.00093980, 0.00093420, 0.00066078, 0.00012966, -0.00058580, -0.00136159, -0.00205048, -0.00251713, -0.00267111, -0.00248853, -0.00201640, -0.00135860, -0.00064808, -0.00001357, 0.00044970, 0.00069741, 0.00073790, 0.00061961, 0.00040966, },
		{ -0.00002385, 0.00013782, 0.00022331, 0.00026014, 0.00027747, 0.00029066, 0.00029363, 0.00026183, 0.00016463, -0.00001762, -0.00027874, -0.00058021, -0.00085668, -0.00103266, -0.00104562, -0.00086799, -0.00052064, -0.00007241, 0.00037495, 0.00071277, 0.00085487, 0.00076161, 0.00045126, -0.00000483, -0.00050175, -0.00092940, -0.00120112, -0.00127431, -0.00115727, -0.00090121, -0.00058088, -0.00027062, },
		{ -0.00051875, -0.00046170, -0.00036565, -0.00027290, -0.00021591, -0.00020417, -0.00022071, -0.00022927, -0.00018989, -0.00007688, 0.00010731, 0.00032781, 0.00052496, 0.00063168, 0.00059546, 0.00039734, 0.00006165, -0.00034741, -0.00074189, -0.00103404, -0.00116105, -0.00110277, -0.00088683, -0.00057984, -0.00026720, -0.00002820, 0.00008604, 0.00006494, -0.00006219, -0.00023894, -0.00040211, -0.00050242, },
		{ -0.00011706, -0.00015144, -0.00018759, -0.00019255, -0.00014993, -0.00006805, 0.00002159, 0.00007529, 0.00005376, -0.00006149, -0.00025795, -0.00049227, -0.00070113, -0.00081994, -0.00080339, -0.00064110, -0.00036311, -0.00003304, 0.00026928, 0.00047034, 0.00052390, 0.00042365, 0.00020387, -0.00007191, -0.00033061, -0.00051107, -0.00058077, -0.00054255, -0.00042988, -0.00029269, -0.00017910, -0.00011930, },
		{ -0.00012612, -0.00009292, -0.00006967, -0.00008006, -0.00012988, -0.00020310, -0.00026671, -0.00028279, -0.00022351, -0.00008386, 0.00011288, 0.00031925, 0.00047644, 0.00053217, 0.00045770, 0.00025852, -0.00002511, -0.00032827, -0.00057924, -0.00071965, -0.00072076, -0.00059085, -0.00037155, -0.00012450, 0.00008734, 0.00021766, 0.00024863, 0.00019307, 0.00008680, -0.00002562, -0.00010684, -0.00013864, },
		{ -0.00011500, -0.00017274, -0.00020628, -0.00019823, -0.00014926, -0.00007865, -0.00001766, 0.00000221, -0.00003811, -0.00013639, -0.00026716, -0.00038806, -0.00045259, -0.00042516, -0.00029378, -0.00007579, 0.00018538, 0.00043161, 0.00060639, 0.00067083, 0.00061456, 0.00045797, 0.00024542, 0.00003147, -0.00013525, -0.00022547, -0.00023594, -0.00018752, -0.00011570, -0.00005715, -0.00003696, -0.00006087, },
		{ -0.00006114, -0.00001227, 0.00000742, -0.00001071, -0.00005881, -0.00011468, -0.00014964, -0.00013915, -0.00007265, 0.00004093, 0.00017424, 0.00028896, 0.00034750, 0.00032511, 0.00021853, 0.00004809, -0.00014746, -0.00032189, -0.00043488, -0.00046358, -0.00040874, -0.00029355, -0.00015574, -0.00003551, 0.00003668, 0.00004866, 0.00000865, -0.00005923, -0.00012451, -0.00016144, -0.00015771, -0.00011799, },
		{ -0.00000261, -0.00002634, -0.00002381, 0.00000429, 0.00004470, 0.00007569, 0.00007467, 0.00002670, -0.00006901, -0.00019710, -0.00032935, -0.00043250, -0.00047810, -0.00045143, -0.00035635, -0.00021465, -0.00005998, 0.00007187, 0.00015320, 0.00017190, 0.00013411, 0.00006118, -0.00001793, -0.00007609, -0.00009634, -0.00007638, -0.00002798, 0.00002837, 0.00007135, 0.00008662, 0.00007167, 0.00003610, },
		{ 0.00000528, 0.00000078, -0.00002342, -0.00006115, -0.00009858, -0.00011869, -0.00010731, -0.00005874, 0.00002116, 0.00011559, 0.00020111, 0.00025431, 0.00025900, 0.00021128, 0.00012119, 0.00001000, -0.00009573, -0.00017176, -0.00020291, -0.00018691, -0.00013428, -0.00006437, 0.00000108, 0.00004468, 0.00005824, 0.00004434, 0.00001417, -0.00001730, -0.00003718, -0.00003927, -0.00002601, -0.00000699, },
		{ -0.00002936, -0.00001833, 0.00000637, 0.00003729, 0.00006339, 0.00007373, 0.00006147, 0.00002667, -0.00002309, -0.00007451, -0.00011247, -0.00012461, -0.00010525, -0.00005740, 0.00000799, 0.00007507, 0.00012775, 0.00015431, 0.00015060, 0.00012085, 0.00007595, 0.00002968, -0.00000564, -0.00002282, -0.00002155, -0.00000772, 0.00000926, 0.00002022, 0.00001967, 0.00000775, -0.00001006, -0.00002515, },
		{ 0.00002243, 0.00000762, -0.00001423, -0.00003563, -0.00004877, -0.00004804, -0.00003192, -0.00000361, 0.00002983, 0.00005958, 0.00007768, 0.00007948, 0.00006485, 0.00003819, 0.00000682, -0.00002130, -0.00004007, -0.00004697, -0.00004351, -0.00003439, -0.00002557, -0.00002198, -0.00002582, -0.00003574, -0.00004748, -0.00005552, -0.00005524, -0.00004474, -0.00002580, -0.00000345, 0.00001563, 0.00002536, },
		{ -0.00003071, -0.00000753, 0.00001378, 0.00002737, 0.00002979, 0.00002089, 0.00000368, -0.00001674, -0.00003482, -0.00004626, -0.00004917, -0.00004439, -0.00003484, -0.00002433, -0.00001604, -0.00001144, -0.00000985, -0.00000894, -0.00000579, 0.00000173, 0.00001385, 0.00002851, 0.00004183, 0.00004929, 0.00004732, 0.00003467, 0.00001312, -0.00001278, -0.00003686, -0.00005314, -0.00005753, -0.00004917, },
		{ -0.00000750, -0.00002518, -0.00003345, -0.00003095, -0.00001985, -0.00000493, 0.00000814, 0.00001485, 0.00001332, 0.00000492, -0.00000632, -0.00001523, -0.00001743, -0.00001097, 0.00000286, 0.00001986, 0.00003434, 0.00004103, 0.00003700, 0.00002277, 0.00000229, -0.00001823, -0.00003229, -0.00003519, -0.00002565, -0.00000627, 0.00001727, 0.00003791, 0.00004949, 0.00004864, 0.00003579, 0.00001498, },
		{ -0.00000667, -0.00001232, -0.00002682, -0.00004685, -0.00006637, -0.00007885, -0.00007964, -0.00006780, -0.00004663, -0.00002283, -0.00000429, 0.00000248, -0.00000533, -0.00002574, -0.00005243, -0.00007659, -0.00008965, -0.00008607, -0.00006506, -0.00003098, 0.00000803, 0.00004245, 0.00006430, 0.00006943, 0.00005856, 0.00003682, 0.00001178, -0.00000905, -0.00002073, -0.00002225, -0.00001662, -0.00000944, },
		},
		{ { 0.00000072, -0.00001982, -0.00005279, -0.00004087, 0.00004264, 0.00015905, 0.00022841, 0.00019416, 0.00007265, -0.00005652, -0.00011260, -0.00007592, 0.00000195, 0.00004693, 0.00002618, -0.00002862, -0.00005683, -0.00002862, 0.00002618, 0.00004693, 0.00000195, -0.00007592, -0.00011260, -0.00005652, 0.00007265, 0.00019416, 0.00022841, 0.00015905, 0.00004264, -0.00004087, -0.00005279, -0.00001982, },
		{ -0.00001901, 0.00002269, 0.00010419, 0.00016278, 0.00014107, 0.00004084, -0.00007364, -0.00012748, -0.00009717, -0.00002825, 0.00000700, -0.00002609, -0.00009529, -0.00013308, -0.00010006, -0.00002241, 0.00003134, 0.00001398, -0.00005367, -0.00009685, -0.00005184, 0.00007409, 0.00020139, 0.00023901, 0.00015727, 0.00001147, -0.00010302, -0.00012528, -0.00007047, -0.00000585, 0.00001372, -0.00000778, },
		{ 0.00006348, 0.00006543, -0.00000350, -0.00009871, -0.00015215, -0.00013333, -0.00007501, -0.00003982, -0.00005912, -0.00010187, -0.00010412, -0.00003356, 0.00007125, 0.00012677, 0.00007687, -0.00005112, -0.00015959, -0.00015690, -0.00003469, 0.00012007, 0.00019116, 0.00012436, -0.00003096, -0.00016313, -0.00018691, -0.00010574, 0.00000126, 0.00005157, 0.00002622, -0.00002436, -0.00003549, 0.00000906, },
		{ -0.00004699, -0.00010616, -0.00010637, -0.00006237, -0.00002680, -0.00003354, -0.00006241, -0.00006153, -0.00000683, 0.00006090, 0.00006349, -0.00004043, -0.00020099, -0.00029858, -0.00023368, -0.00001430, 0.00023496, 0.00035746, 0.00027948, 0.00006487, -0.00013490, -0.00019512, -0.00010549, 0.00003565, 0.00010975, 0.00007217, -0.00002379, -0.00008535, -0.00006432, 0.00000675, 0.00005264, 0.00002734, },
		{ -0.00009952, -0.00005948, -0.00004745, -0.00006017, -0.00005709, -0.00001337, 0.00004130, 0.00004379, -0.00003512, -0.00014189, -0.00016614, -0.00003221, 0.00021859, 0.00043300, 0.00044855, 0.00021968, -0.00013468, -0.00040677, -0.00044540, -0.00025996, -0.00000388, 0.00014628, 0.00011851, -0.00001798, -0.00012697, -0.00012116, -0.00002333, 0.00006944, 0.00007738, 0.00000208, -0.00008733, -0.00012389, },
		{ -0.00003629, -0.00003527, -0.00001011, 0.00002787, 0.00002908, -0.00003580, -0.00012703, -0.00015498, -0.00006198, 0.00010507, 0.00020726, 0.00011771, -0.00016151, -0.00047219, -0.00059948, -0.00043012, -0.00004631, 0.00032353, 0.00046422, 0.00032246, 0.00003244, -0.00019590, -0.00023320, -0.00011014, 0.00003233, 0.00006858, -0.00001786, -0.00014190, -0.00020429, -0.00017456, -0.00010035, -0.00004745, },
		{ 0.00000003, -0.00000139, -0.00004908, -0.00012584, -0.00016076, -0.00009859, 0.00003036, 0.00011574, 0.00005809, -0.00012905, -0.00030332, -0.00029480, -0.00005141, 0.00029378, 0.00050123, 0.00039937, 0.00002068, -0.00041119, -0.00063788, -0.00054741, -0.00024205, 0.00005291, 0.00016046, 0.00006752, -0.00009600, -0.00018112, -0.00013580, -0.00002289, 0.00005774, 0.00006096, 0.00002034, -0.00000538, },
		{ -0.00010071, -0.00014158, -0.00011355, -0.00001949, 0.00005682, 0.00002449, -0.00011888, -0.00026479, -0.00027871, -0.00012537, 0.00007933, 0.00014897, -0.00001586, -0.00032430, -0.00054255, -0.00046633, -0.00008530, 0.00039441, 0.00069260, 0.00065342, 0.00034687, 0.00000029, -0.00017741, -0.00013675, 0.00000916, 0.00010356, 0.00007581, -0.00002368, -0.00009640, -0.00009563, -0.00005950, -0.00005548, },
		{ -0.00008991, -0.00000758, 0.00000581, -0.00008289, -0.00019926, -0.00022332, -0.00010381, 0.00008391, 0.00019781, 0.00015980, 0.00003845, -0.00000038, 0.00015691, 0.00044914, 0.00065977, 0.00057678, 0.00016904, -0.00036766, -0.00073959, -0.00076227, -0.00048409, -0.00013356, 0.00006057, 0.00002914, -0.00011869, -0.00021573, -0.00018222, -0.00007390, -0.00000585, -0.00003514, -0.00011300, -0.00014443, },
		{ -0.00005750, -0.00018734, -0.00028023, -0.00024790, -0.00011849, -0.00000731, -0.00000526, -0.00009310, -0.00016563, -0.00014490, -0.00007283, -0.00008641, -0.00028545, -0.00060915, -0.00084306, -0.00077095, -0.00035587, 0.00020244, 0.00059336, 0.00061372, 0.00030871, -0.00007476, -0.00028099, -0.00022864, -0.00004253, 0.00007718, 0.00002773, -0.00012686, -0.00023615, -0.00020909, -0.00009163, -0.00001278, },
		{ -0.00056846, -0.00046339, -0.00027969, -0.00014354, -0.00009779, -0.00008740, -0.00004131, 0.00004094, 0.00008990, 0.00005445, -0.00001552, 0.00001100, 0.00022294, 0.00054793, 0.00076000, 0.00063545, 0.00014332, -0.00049112, -0.00092007, -0.00091442, -0.00052749, -0.00005129, 0.00020166, 0.00011974, -0.00015304, -0.00036040, -0.00034787, -0.00017209, -0.00002739, -0.00006681, -0.00027522, -0.00049510, },
		{ 0.00009375, 0.00031667, 0.00039720, 0.00041061, 0.00040738, 0.00037482, 0.00028498, 0.00016008, 0.00007060, 0.00005798, 0.00006355, -0.00004430, -0.00034314, -0.00073627, -0.00096624, -0.00078690, -0.00017458, 0.00059825, 0.00111177, 0.00107852, 0.00054701, -0.00013121, -0.00055190, -0.00053781, -0.00023855, 0.00001511, -0.00002300, -0.00033799, -0.00069200, -0.00082488, -0.00064589, -0.00027172, },
		{ 0.00063900, 0.00047193, 0.00032006, 0.00018291, 0.00006194, -0.00001075, -0.00000907, 0.00003855, 0.00006224, 0.00003066, 0.00001747, 0.00016402, 0.00053902, 0.00100911, 0.00125820, 0.00098593, 0.00015626, -0.00090520, -0.00167392, -0.00176018, -0.00117791, -0.00034434, 0.00019981, 0.00015077, -0.00037488, -0.00095452, -0.00116993, -0.00087855, -0.00026226, 0.00034649, 0.00070076, 0.00075994, },
		{ 0.00027684, 0.00022331, 0.00020291, 0.00021543, 0.00024383, 0.00024386, 0.00018603, 0.00009930, 0.00004903, 0.00004973, -0.00000148, -0.00027538, -0.00084741, -0.00154698, -0.00197707, -0.00174707, -0.00077992, 0.00054233, 0.00154319, 0.00165230, 0.00077805, -0.00061264, -0.00177287, -0.00212484, -0.00157703, -0.00051354, 0.00050077, 0.00106874, 0.00113094, 0.00088827, 0.00059286, 0.00038586, },
		{ 0.00054810, 0.00043077, 0.00031092, 0.00017490, 0.00003605, -0.00006414, -0.00010388, -0.00012321, -0.00019288, -0.00031465, -0.00034980, -0.00008235, 0.00058258, 0.00142803, 0.00193078, 0.00153315, 0.00003875, -0.00214528, -0.00413042, -0.00500139, -0.00432842, -0.00240568, -0.00006455, 0.00180346, 0.00269788, 0.00266625, 0.00212534, 0.00152804, 0.00111948, 0.00090624, 0.00078315, 0.00066894, },
		{ -0.00002376, -0.00008218, -0.00011480, -0.00011417, -0.00010287, -0.00012927, -0.00021393, -0.00030862, -0.00034039, -0.00033305, -0.00049387, -0.00113727, -0.00243340, -0.00413984, -0.00554996, -0.00577837, -0.00426228, -0.00116676, 0.00259622, 0.00578577, 0.00744463, 0.00733359, 0.00594120, 0.00410961, 0.00255331, 0.00157117, 0.00107182, 0.00080467, 0.00058431, 0.00036589, 0.00018287, 0.00005793, },
		{ -0.00081354, -0.00082643, -0.00087618, -0.00097365, -0.00109526, -0.00119493, -0.00126751, -0.00139098, -0.00166232, -0.00203340, -0.00217783, -0.00155434, 0.00028820, 0.00332231, 0.00685060, 0.00970957, 0.01081045, 0.00970957, 0.00685060, 0.00332231, 0.00028820, -0.00155434, -0.00217783, -0.00203340, -0.00166232, -0.00139098, -0.00126751, -0.00119493, -0.00109526, -0.00097365, -0.00087618, -0.00082643, },
		{ -0.00002376, 0.00005793, 0.00018287, 0.00036589, 0.00058431, 0.00080467, 0.00107182, 0.00157117, 0.00255331, 0.00410961, 0.00594120, 0.00733359, 0.00744463, 0.00578577, 0.00259622, -0.00116676, -0.00426228, -0.00577837, -0.00554996, -0.00413984, -0.00243340, -0.00113727, -0.00049387, -0.00033305, -0.00034039, -0.00030862, -0.00021393, -0.00012927, -0.00010287, -0.00011417, -0.00011480, -0.00008218, },
		{ 0.00054810, 0.00066894, 0.00078315, 0.00090624, 0.00111948, 0.00152804, 0.00212534, 0.00266625, 0.00269788, 0.00180346, -0.00006455, -0.00240568, -0.00432842, -0.00500139, -0.00413042, -0.00214528, 0.00003875, 0.00153315, 0.00193078, 0.00142803, 0.00058258, -0.00008235, -0.00034980, -0.00031465, -0.00019288, -0.00012321, -0.00010388, -0.00006414, 0.00003605, 0.00017490, 0.00031092, 0.00043077, },
		{ 0.00027684, 0.00038586, 0.00059286, 0.00088827, 0.00113094, 0.00106874, 0.00050077, -0.00051354, -0.00157703, -0.00212484, -0.00177287, -0.00061264, 0.00077805, 0.00165230, 0.00154319, 0.00054233, -0.00077992, -0.00174707, -0.00197707, -0.00154698, -0.00084741, -0.00027538, -0.00000148, 0.00004973, 0.00004903, 0.00009930, 0.00018603, 0.00024386, 0.00024383, 0.00021543, 0.00020291, 0.00022331, },
		{ 0.00063900, 0.00075994, 0.00070076, 0.00034649, -0.00026226, -0.00087855, -0.00116993, -0.00095452, -0.00037488, 0.00015077, 0.00019981, -0.00034434, -0.00117791, -0.00176018, -0.00167392, -0.00090520, 0.00015626, 0.00098593, 0.00125820, 0.00100911, 0.00053902, 0.00016402, 0.00001747, 0.00003066, 0.00006224, 0.00003855, -0.00000907, -0.00001075, 0.00006194, 0.00018291, 0.00032006, 0.00047193, },
		{ 0.00009375, -0.00027172, -0.00064589, -0.00082488, -0.00069200, -0.00033799, -0.00002300, 0.00001511, -0.00023855, -0.00053781, -0.00055190, -0.00013121, 0.00054701, 0.00107852, 0.00111177, 0.00059825, -0.00017458, -0.00078690, -0.00096624, -0.00073627, -0.00034314, -0.00004430, 0.00006355, 0.00005798, 0.00007060, 0.00016008, 0.00028498, 0.00037482, 0.00040738, 0.00041061, 0.00039720, 0.00031667, },
		{ -0.00056846, -0.00049510, -0.00027522, -0.00006681, -0.00002739, -0.00017209, -0.00034787, -0.00036040, -0.00015304, 0.00011974, 0.00020166, -0.00005129, -0.00052749, -0.00091442, -0.00092007, -0.00049112, 0.00014332, 0.00063545, 0.00076000, 0.00054793, 0.00022294, 0.00001100, -0.00001552, 0.00005445, 0.00008990, 0.00004094, -0.00004131, -0.00008740, -0.00009779, -0.00014354, -0.00027969, -0.00046339, },
		{ -0.00005750, -0.00001278, -0.00009163, -0.00020909, -0.00023615, -0.00012686, 0.00002773, 0.00007718, -0.00004253, -0.00022864, -0.00028099, -0.00007476, 0.00030871, 0.00061372, 0.00059336, 0.00020244, -0.00035587, -0.00077095, -0.00084306, -0.00060915, -0.00028545, -0.00008641, -0.00007283, -0.00014490, -0.00016563, -0.00009310, -0.00000526, -0.00000731, -0.00011849, -0.00024790, -0.00028023, -0.00018734, },
		{ -0.00008991, -0.00014443, -0.00011300, -0.00003514, -0.00000585, -0.00007390, -0.00018222, -0.00021573, -0.00011869, 0.00002914, 0.00006057, -0.00013356, -0.00048409, -0.00076227, -0.00073959, -0.00036766, 0.00016904, 0.00057678, 0.00065977, 0.00044914, 0.00015691, -0.00000038, 0.00003845, 0.00015980, 0.00019781, 0.00008391, -0.00010381, -0.00022332, -0.00019926, -0.00008289, 0.00000581, -0.00000758, },
		{ -0.00010071, -0.00005548, -0.00005950, -0.00009563, -0.00009640, -0.00002368, 0.00007581, 0.00010356, 0.00000916, -0.00013675, -0.00017741, 0.00000029, 0.00034687, 0.00065342, 0.00069260, 0.00039441, -0.00008530, -0.00046633, -0.00054255, -0.00032430, -0.00001586, 0.00014897, 0.00007933, -0.00012537, -0.00027871, -0.00026479, -0.00011888, 0.00002449, 0.00005682, -0.00001949, -0.00011355, -0.00014158, },
		{ 0.00000003, -0.00000538, 0.00002034, 0.00006096, 0.00005774, -0.00002289, -0.00013580, -0.00018112, -0.00009600, 0.00006752, 0.00016046, 0.00005291, -0.00024205, -0.00054741, -0.00063788, -0.00041119, 0.00002068, 0.00039937, 0.00050123, 0.00029378, -0.00005141, -0.00029480, -0.00030332, -0.00012905, 0.00005809, 0.00011574, 0.00003036, -0.00009859, -0.00016076, -0.00012584, -0.00004908, -0.00000139, },
		{ -0.00003629, -0.00004745, -0.00010035, -0.00017456, -0.00020429, -0.00014190, -0.00001786, 0.00006858, 0.00003233, -0.00011014, -0.00023320, -0.00019590, 0.00003244, 0.00032246, 0.00046422, 0.00032353, -0.00004631, -0.00043012, -0.00059948, -0.00047219, -0.00016151, 0.00011771, 0.00020726, 0.00010507, -0.00006198, -0.00015498, -0.00012703, -0.00003580, 0.00002908, 0.00002787, -0.00001011, -0.00003527, },
		{ -0.00009952, -0.00012389, -0.00008733, 0.00000208, 0.00007738, 0.00006944, -0.00002333, -0.00012116, -0.00012697, -0.00001798, 0.00011851, 0.00014628, -0.00000388, -0.00025996, -0.00044540, -0.00040677, -0.00013468, 0.00021968, 0.00044855, 0.00043300, 0.00021859, -0.00003221, -0.00016614, -0.00014189, -0.00003512, 0.00004379, 0.00004130, -0.00001337, -0.00005709, -0.00006017, -0.00004745, -0.00005948, },
		{ -0.00004699, 0.00002734, 0.00005264, 0.00000675, -0.00006432, -0.00008535, -0.00002379, 0.00007217, 0.00010975, 0.00003565, -0.00010549, -0.00019512, -0.00013490, 0.00006487, 0.00027948, 0.00035746, 0.00023496, -0.00001430, -0.00023368, -0.00029858, -0.00020099, -0.00004043, 0.00006349, 0.00006090, -0.00000683, -0.00006153, -0.00006241, -0.00003354, -0.00002680, -0.00006237, -0.00010637, -0.00010616, },
		{ 0.00006348, 0.00000906, -0.00003549, -0.00002436, 0.00002622, 0.00005157, 0.00000126, -0.00010574, -0.00018691, -0.00016313, -0.00003096, 0.00012436, 0.00019116, 0.00012007, -0.00003469, -0.00015690, -0.00015959, -0.00005112, 0.00007687, 0.00012677, 0.00007125, -0.00003356, -0.00010412, -0.00010187, -0.00005912, -0.00003982, -0.00007501, -0.00013333, -0.00015215, -0.00009871, -0.00000350, 0.00006543, },
		{ -0.00001901, -0.00000778, 0.00001372, -0.00000585, -0.00007047, -0.00012528, -0.00010302, 0.00001147, 0.00015727, 0.00023901, 0.00020139, 0.00007409, -0.00005184, -0.00009685, -0.00005367, 0.00001398, 0.00003134, -0.00002241, -0.00010006, -0.00013308, -0.00009529, -0.00002609, 0.00000700, -0.00002825, -0.00009717, -0.00012748, -0.00007364, 0.00004084, 0.00014107, 0.00016278, 0.00010419, 0.00002269, },
		},
		{ { -0.00005022, -0.00001108, 0.00001021, -0.00003497, -0.00006731, -0.00015189, -0.00029981, -0.00025858, -0.00004388, -0.00002323, -0.00011383, 0.00006538, 0.00021836, -0.00009431, -0.00033026, 0.00007710, 0.00044009, 0.00007710, -0.00033026, -0.00009431, 0.00021836, 0.00006538, -0.00011383, -0.00002323, -0.00004388, -0.00025858, -0.00029981, -0.00015189, -0.00006731, -0.00003497, 0.00001021, -0.00001108, },
		{ -0.00002997, -0.00010349, -0.00016930, -0.00018522, -0.00021275, -0.00013513, 0.00005795, 0.00004615, -0.00016020, -0.00010626, 0.00012918, 0.00002139, -0.00020162, 0.00004639, 0.00035469, 0.00004843, -0.00034701, -0.00008526, 0.00032500, 0.00016952, -0.00016942, -0.00018714, -0.00016769, -0.00030378, -0.00026306, -0.00001590, 0.00008775, 0.00002716, -0.00000144, -0.00002657, -0.00007536, -0.00005356, },
		{ -0.00018836, -0.00015186, -0.00009582, -0.00004131, 0.00006579, 0.00002839, -0.00020603, -0.00021558, 0.00008589, 0.00014481, -0.00010988, -0.00008135, 0.00019398, 0.00010017, -0.00017636, -0.00000349, 0.00030283, 0.00009312, -0.00030216, -0.00030914, -0.00017346, -0.00020089, -0.00011556, 0.00014259, 0.00018481, -0.00000991, -0.00010787, -0.00009065, -0.00010092, -0.00008006, -0.00004119, -0.00010914, },
		{ -0.00001943, 0.00000396, 0.00002441, 0.00000072, -0.00018580, -0.00025713, -0.00000017, 0.00014622, -0.00008802, -0.00017993, 0.00012426, 0.00027029, 0.00006291, -0.00001055, 0.00007815, -0.00009166, -0.00033809, -0.00023491, -0.00002722, -0.00003391, 0.00001193, 0.00020605, 0.00015310, -0.00014789, -0.00024440, -0.00010694, -0.00004063, -0.00004112, -0.00001758, -0.00008169, -0.00018203, -0.00013011, },
		{ -0.00001258, -0.00002291, -0.00014128, -0.00029546, -0.00017688, 0.00003017, -0.00004751, -0.00011848, 0.00018320, 0.00042565, 0.00018142, -0.00015926, -0.00021045, -0.00017208, -0.00014427, 0.00001338, 0.00012514, 0.00002866, 0.00001697, 0.00015919, 0.00004470, -0.00028770, -0.00029665, 0.00000271, 0.00011181, 0.00000579, -0.00004178, -0.00008856, -0.00018448, -0.00012712, 0.00003283, 0.00004622, },
		{ -0.00012348, -0.00033012, -0.00037478, -0.00011718, 0.00005686, 0.00007914, 0.00026613, 0.00042798, 0.00014022, -0.00030123, -0.00032796, -0.00003048, 0.00019007, 0.00022907, 0.00010806, -0.00007202, -0.00004822, 0.00010844, -0.00002888, -0.00033615, -0.00021773, 0.00020451, 0.00024943, -0.00005592, -0.00017348, -0.00011352, -0.00013521, -0.00008930, 0.00008089, 0.00008175, -0.00006474, -0.00008781, },
		{ -0.00051289, -0.00027875, 0.00013869, 0.00028815, 0.00030584, 0.00031540, 0.00008125, -0.00025555, -0.00019330, 0.00021531, 0.00042499, 0.00024111, -0.00007315, -0.00020551, -0.00003952, 0.00014743, -0.00002812, -0.00029904, -0.00007086, 0.00037526, 0.00024582, -0.00026835, -0.00034351, -0.00003955, 0.00004095, -0.00002699, 0.00005422, 0.00004469, -0.00015317, -0.00018230, -0.00010100, -0.00029506, },
		{ 0.00009617, 0.00045839, 0.00038299, 0.00025205, 0.00012775, -0.00002724, 0.00005011, 0.00035692, 0.00043119, 0.00010592, -0.00022721, -0.00020377, 0.00009160, 0.00023378, -0.00002102, -0.00025346, 0.00004292, 0.00041418, 0.00007847, -0.00052858, -0.00039097, 0.00021905, 0.00029118, -0.00007220, -0.00014708, -0.00006267, -0.00019014, -0.00021364, -0.00003464, -0.00019147, -0.00061077, -0.00051670, },
		{ 0.00060062, 0.00031910, 0.00023353, 0.00021886, 0.00024047, 0.00035273, 0.00032166, 0.00002430, -0.00020441, -0.00007160, 0.00021842, 0.00023674, -0.00008057, -0.00024420, 0.00008067, 0.00032397, -0.00014746, -0.00066680, -0.00022126, 0.00057983, 0.00042273, -0.00035243, -0.00048202, -0.00008201, -0.00003296, -0.00015656, -0.00001491, -0.00012910, -0.00071210, -0.00082026, -0.00004979, 0.00064007, },
		{ 0.00030874, 0.00040291, 0.00040719, 0.00035850, 0.00027525, 0.00007160, -0.00008580, 0.00000763, 0.00017928, 0.00010354, -0.00017722, -0.00023677, 0.00007249, 0.00018644, -0.00029482, -0.00059752, 0.00008248, 0.00082337, 0.00027582, -0.00080291, -0.00071677, 0.00017326, 0.00030870, -0.00018058, -0.00021924, -0.00015604, -0.00071315, -0.00104348, -0.00023680, 0.00080394, 0.00090857, 0.00046081, },
		{ 0.00067688, 0.00046769, 0.00026678, 0.00008121, -0.00008854, -0.00010548, -0.00002337, -0.00006219, -0.00020621, -0.00017916, 0.00005842, 0.00007652, -0.00030707, -0.00039010, 0.00032090, 0.00078395, -0.00009272, -0.00115351, -0.00061327, 0.00067359, 0.00057254, -0.00052405, -0.00066190, -0.00008149, -0.00038994, -0.00110662, -0.00058655, 0.00073988, 0.00118472, 0.00070425, 0.00045835, 0.00064452, },
		{ 0.00006892, -0.00014579, -0.00028529, -0.00032083, -0.00021115, -0.00011418, -0.00011733, -0.00008285, 0.00003903, 0.00001161, -0.00023093, -0.00019389, 0.00035419, 0.00050571, -0.00042917, -0.00115113, -0.00018714, 0.00110739, 0.00044389, -0.00118608, -0.00103290, 0.00040464, 0.00029517, -0.00125107, -0.00141688, 0.00029086, 0.00154842, 0.00135276, 0.00094832, 0.00102513, 0.00096094, 0.00049581, },
		{ -0.00048374, -0.00040787, -0.00023048, -0.00003504, 0.00001742, -0.00001394, -0.00000386, -0.00002778, -0.00014777, -0.00010472, 0.00021272, 0.00019818, -0.00050647, -0.00080150, 0.00026029, 0.00113006, -0.00010356, -0.00179360, -0.00090714, 0.00134071, 0.00093857, -0.00194883, -0.00286702, -0.00026174, 0.00248542, 0.00277660, 0.00181207, 0.00125267, 0.00085443, 0.00014462, -0.00042583, -0.00054277, },
		{ -0.00012699, 0.00004624, 0.00008636, -0.00000765, -0.00006033, -0.00005544, -0.00010036, -0.00010674, 0.00004014, 0.00004010, -0.00032355, -0.00036775, 0.00040672, 0.00070083, -0.00070785, -0.00187499, -0.00019240, 0.00225467, 0.00089306, -0.00340245, -0.00447564, -0.00017127, 0.00446027, 0.00488170, 0.00252213, 0.00071544, -0.00000526, -0.00057237, -0.00094167, -0.00077860, -0.00044075, -0.00026786, },
		{ 0.00006753, -0.00002270, -0.00009341, -0.00007123, -0.00008239, -0.00012697, -0.00005340, 0.00003811, -0.00007631, -0.00011942, 0.00020696, 0.00014178, -0.00092274, -0.00131882, 0.00066107, 0.00244370, -0.00013884, -0.00510832, -0.00526619, 0.00121103, 0.00727802, 0.00683054, 0.00222554, -0.00096273, -0.00143117, -0.00114601, -0.00099141, -0.00061383, -0.00017675, -0.00006071, -0.00007691, 0.00001517, },
		{ -0.00008764, -0.00008069, -0.00007492, -0.00012093, -0.00006956, 0.00007293, 0.00005358, -0.00009767, -0.00009063, -0.00016649, -0.00064146, -0.00053337, 0.00098944, 0.00160046, -0.00168372, -0.00603645, -0.00445049, 0.00371981, 0.00987979, 0.00747475, 0.00048307, -0.00337176, -0.00262403, -0.00094164, -0.00037637, -0.00022542, -0.00000383, -0.00001605, -0.00015373, -0.00008926, 0.00003905, 0.00000032, },
		{ -0.00007206, -0.00009768, -0.00007097, 0.00005394, 0.00005553, -0.00013674, -0.00021949, -0.00013870, -0.00019349, -0.00009603, 0.00051592, 0.00017486, -0.00275293, -0.00540518, -0.00216759, 0.00624975, 0.01094653, 0.00624975, -0.00216759, -0.00540518, -0.00275293, 0.00017486, 0.00051592, -0.00009603, -0.00019349, -0.00013870, -0.00021949, -0.00013674, 0.00005553, 0.00005394, -0.00007097, -0.00009768, },
		{ -0.00008764, 0.00000032, 0.00003905, -0.00008926, -0.00015373, -0.00001605, -0.00000383, -0.00022542, -0.00037637, -0.00094164, -0.00262403, -0.00337176, 0.00048307, 0.00747475, 0.00987979, 0.00371981, -0.00445049, -0.00603645, -0.00168372, 0.00160046, 0.00098944, -0.00053337, -0.00064146, -0.00016649, -0.00009063, -0.00009767, 0.00005358, 0.00007293, -0.00006956, -0.00012093, -0.00007492, -0.00008069, },
		{ 0.00006753, 0.00001517, -0.00007691, -0.00006071, -0.00017675, -0.00061383, -0.00099141, -0.00114601, -0.00143117, -0.00096273, 0.00222554, 0.00683054, 0.00727802, 0.00121103, -0.00526619, -0.00510832, -0.00013884, 0.00244370, 0.00066107, -0.00131882, -0.00092274, 0.00014178, 0.00020696, -0.00011942, -0.00007631, 0.00003811, -0.00005340, -0.00012697, -0.00008239, -0.00007123, -0.00009341, -0.00002270, },
		{ -0.00012699, -0.00026786, -0.00044075, -0.00077860, -0.00094167, -0.00057237, -0.00000526, 0.00071544, 0.00252213, 0.00488170, 0.00446027, -0.00017127, -0.00447564, -0.00340245, 0.00089306, 0.00225467, -0.00019240, -0.00187499, -0.00070785, 0.00070083, 0.00040672, -0.00036775, -0.00032355, 0.00004010, 0.00004014, -0.00010674, -0.00010036, -0.00005544, -0.00006033, -0.00000765, 0.00008636, 0.00004624, },
		{ -0.00048374, -0.00054277, -0.00042583, 0.00014462, 0.00085443, 0.00125267, 0.00181207, 0.00277660, 0.00248542, -0.00026174, -0.00286702, -0.00194883, 0.00093857, 0.00134071, -0.00090714, -0.00179360, -0.00010356, 0.00113006, 0.00026029, -0.00080150, -0.00050647, 0.00019818, 0.00021272, -0.00010472, -0.00014777, -0.00002778, -0.00000386, -0.00001394, 0.00001742, -0.00003504, -0.00023048, -0.00040787, },
		{ 0.00006892, 0.00049581, 0.00096094, 0.00102513, 0.00094832, 0.00135276, 0.00154842, 0.00029086, -0.00141688, -0.00125107, 0.00029517, 0.00040464, -0.00103290, -0.00118608, 0.00044389, 0.00110739, -0.00018714, -0.00115113, -0.00042917, 0.00050571, 0.00035419, -0.00019389, -0.00023093, 0.00001161, 0.00003903, -0.00008285, -0.00011733, -0.00011418, -0.00021115, -0.00032083, -0.00028529, -0.00014579, },
		{ 0.00067688, 0.00064452, 0.00045835, 0.00070425, 0.00118472, 0.00073988, -0.00058655, -0.00110662, -0.00038994, -0.00008149, -0.00066190, -0.00052405, 0.00057254, 0.00067359, -0.00061327, -0.00115351, -0.00009272, 0.00078395, 0.00032090, -0.00039010, -0.00030707, 0.00007652, 0.00005842, -0.00017916, -0.00020621, -0.00006219, -0.00002337, -0.00010548, -0.00008854, 0.00008121, 0.00026678, 0.00046769, },
		{ 0.00030874, 0.00046081, 0.00090857, 0.00080394, -0.00023680, -0.00104348, -0.00071315, -0.00015604, -0.00021924, -0.00018058, 0.00030870, 0.00017326, -0.00071677, -0.00080291, 0.00027582, 0.00082337, 0.00008248, -0.00059752, -0.00029482, 0.00018644, 0.00007249, -0.00023677, -0.00017722, 0.00010354, 0.00017928, 0.00000763, -0.00008580, 0.00007160, 0.00027525, 0.00035850, 0.00040719, 0.00040291, },
		{ 0.00060062, 0.00064007, -0.00004979, -0.00082026, -0.00071210, -0.00012910, -0.00001491, -0.00015656, -0.00003296, -0.00008201, -0.00048202, -0.00035243, 0.00042273, 0.00057983, -0.00022126, -0.00066680, -0.00014746, 0.00032397, 0.00008067, -0.00024420, -0.00008057, 0.00023674, 0.00021842, -0.00007160, -0.00020441, 0.00002430, 0.00032166, 0.00035273, 0.00024047, 0.00021886, 0.00023353, 0.00031910, },
		{ 0.00009617, -0.00051670, -0.00061077, -0.00019147, -0.00003464, -0.00021364, -0.00019014, -0.00006267, -0.00014708, -0.00007220, 0.00029118, 0.00021905, -0.00039097, -0.00052858, 0.00007847, 0.00041418, 0.00004292, -0.00025346, -0.00002102, 0.00023378, 0.00009160, -0.00020377, -0.00022721, 0.00010592, 0.00043119, 0.00035692, 0.00005011, -0.00002724, 0.00012775, 0.00025205, 0.00038299, 0.00045839, },
		{ -0.00051289, -0.00029506, -0.00010100, -0.00018230, -0.00015317, 0.00004469, 0.00005422, -0.00002699, 0.00004095, -0.00003955, -0.00034351, -0.00026835, 0.00024582, 0.00037526, -0.00007086, -0.00029904, -0.00002812, 0.00014743, -0.00003952, -0.00020551, -0.00007315, 0.00024111, 0.00042499, 0.00021531, -0.00019330, -0.00025555, 0.00008125, 0.00031540, 0.00030584, 0.00028815, 0.00013869, -0.00027875, },
		{ -0.00012348, -0.00008781, -0.00006474, 0.00008175, 0.00008089, -0.00008930, -0.00013521, -0.00011352, -0.00017348, -0.00005592, 0.00024943, 0.00020451, -0.00021773, -0.00033615, -0.00002888, 0.00010844, -0.00004822, -0.00007202, 0.00010806, 0.00022907, 0.00019007, -0.00003048, -0.00032796, -0.00030123, 0.00014022, 0.00042798, 0.00026613, 0.00007914, 0.00005686, -0.00011718, -0.00037478, -0.00033012, },
		{ -0.00001258, 0.00004622, 0.00003283, -0.00012712, -0.00018448, -0.00008856, -0.00004178, 0.00000579, 0.00011181, 0.00000271, -0.00029665, -0.00028770, 0.00004470, 0.00015919, 0.00001697, 0.00002866, 0.00012514, 0.00001338, -0.00014427, -0.00017208, -0.00021045, -0.00015926, 0.00018142, 0.00042565, 0.00018320, -0.00011848, -0.00004751, 0.00003017, -0.00017688, -0.00029546, -0.00014128, -0.00002291, },
		{ -0.00001943, -0.00013011, -0.00018203, -0.00008169, -0.00001758, -0.00004112, -0.00004063, -0.00010694, -0.00024440, -0.00014789, 0.00015310, 0.00020605, 0.00001193, -0.00003391, -0.00002722, -0.00023491, -0.00033809, -0.00009166, 0.00007815, -0.00001055, 0.00006291, 0.00027029, 0.00012426, -0.00017993, -0.00008802, 0.00014622, -0.00000017, -0.00025713, -0.00018580, 0.00000072, 0.00002441, 0.00000396, },
		{ -0.00018836, -0.00010914, -0.00004119, -0.00008006, -0.00010092, -0.00009065, -0.00010787, -0.00000991, 0.00018481, 0.00014259, -0.00011556, -0.00020089, -0.00017346, -0.00030914, -0.00030216, 0.00009312, 0.00030283, -0.00000349, -0.00017636, 0.00010017, 0.00019398, -0.00008135, -0.00010988, 0.00014481, 0.00008589, -0.00021558, -0.00020603, 0.00002839, 0.00006579, -0.00004131, -0.00009582, -0.00015186, },
		{ -0.00002997, -0.00005356, -0.00007536, -0.00002657, -0.00000144, 0.00002716, 0.00008775, -0.00001590, -0.00026306, -0.00030378, -0.00016769, -0.00018714, -0.00016942, 0.00016952, 0.00032500, -0.00008526, -0.00034701, 0.00004843, 0.00035469, 0.00004639, -0.00020162, 0.00002139, 0.00012918, -0.00010626, -0.00016020, 0.00004615, 0.00005795, -0.00013513, -0.00021275, -0.00018522, -0.00016930, -0.00010349, },
		},
		{ { -0.00013623, -0.00010453, -0.00029108, -0.00014101, 0.00009256, 0.00008973, 0.00027586, 0.00004780, 0.00009341, 0.00017315, 0.00003519, 0.00010786, -0.00024361, -0.00001343, 0.00036845, -0.00007326, -0.00035270, -0.00007326, 0.00036845, -0.00001343, -0.00024361, 0.00010786, 0.00003519, 0.00017315, 0.00009341, 0.00004780, 0.00027586, 0.00008973, 0.00009256, -0.00014101, -0.00029108, -0.00010453, },
		{ -0.00012123, -0.00034640, -0.00014204, 0.00016127, 0.00024611, 0.00024215, 0.00000524, 0.00024130, 0.00023323, -0.00001986, -0.00005656, -0.00009051, 0.00023481, 0.00008062, -0.00032744, -0.00018575, 0.00027253, 0.00026586, -0.00038244, -0.00003849, 0.00025282, -0.00011679, 0.00013462, 0.00006571, 0.00010054, 0.00019024, -0.00001705, 0.00011536, -0.00017509, -0.00025280, -0.00008541, -0.00013928, },
		{ -0.00043872, -0.00009309, 0.00032370, 0.00031776, 0.00014959, 0.00011761, 0.00041006, 0.00013494, -0.00012451, 0.00000472, 0.00010654, 0.00009993, -0.00024228, -0.00023533, 0.00012193, 0.00034276, -0.00013564, -0.00041197, 0.00035789, 0.00005065, -0.00025753, 0.00026737, 0.00000581, 0.00007772, 0.00010657, -0.00001861, 0.00011156, -0.00024850, -0.00017226, -0.00008513, -0.00015008, -0.00014820, },
		{ 0.00005297, 0.00049218, 0.00028860, 0.00015203, 0.00034063, 0.00042582, -0.00003762, -0.00007503, 0.00013246, 0.00003452, -0.00018726, -0.00025058, 0.00004740, 0.00023443, 0.00009917, -0.00040287, -0.00003295, 0.00045875, -0.00037948, -0.00005985, 0.00036015, -0.00020814, 0.00012965, 0.00008690, -0.00002839, 0.00003807, -0.00026060, -0.00006915, -0.00015310, -0.00013436, -0.00019899, -0.00055010, },
		{ 0.00060669, 0.00024396, 0.00029879, 0.00051887, 0.00029352, -0.00011397, 0.00004478, 0.00006334, -0.00020608, -0.00025488, -0.00000755, 0.00017563, 0.00008120, -0.00015120, -0.00029153, 0.00033741, 0.00009295, -0.00054186, 0.00036972, 0.00009250, -0.00033700, 0.00035763, -0.00003690, -0.00008887, 0.00000827, -0.00019572, -0.00005292, -0.00024329, -0.00004153, -0.00033533, -0.00064976, 0.00030841, },
		{ 0.00026823, 0.00054003, 0.00056764, 0.00011628, -0.00011326, 0.00002692, -0.00016440, -0.00029638, -0.00006478, 0.00015122, 0.00004189, -0.00009562, -0.00016021, -0.00000817, 0.00032252, -0.00035400, -0.00021108, 0.00052067, -0.00040561, -0.00004361, 0.00047889, -0.00036145, -0.00000994, 0.00004217, -0.00018659, -0.00012001, -0.00022275, 0.00005302, -0.00063291, -0.00063402, 0.00066498, 0.00063560, },
		{ 0.00079041, 0.00044379, -0.00008801, -0.00015618, -0.00012413, -0.00029196, -0.00014883, 0.00010679, 0.00006059, -0.00013879, -0.00007894, -0.00001642, 0.00010382, 0.00002466, -0.00043148, 0.00024068, 0.00014788, -0.00057360, 0.00051216, 0.00011537, -0.00056412, 0.00030052, -0.00008259, -0.00023613, -0.00010928, -0.00010725, -0.00003131, -0.00100983, -0.00037108, 0.00102116, 0.00058757, 0.00044982, },
		{ 0.00008764, -0.00031004, -0.00022184, -0.00022322, -0.00021689, 0.00002596, 0.00006908, -0.00011942, -0.00014087, 0.00006808, 0.00003250, -0.00000940, -0.00017755, -0.00017042, 0.00031931, -0.00034547, -0.00014143, 0.00069098, -0.00055691, -0.00022515, 0.00047163, -0.00045121, -0.00007620, -0.00003175, -0.00010029, -0.00030122, -0.00124972, 0.00009817, 0.00125167, 0.00061770, 0.00082250, 0.00085904, },
		{ -0.00040435, -0.00019940, -0.00020124, -0.00008680, 0.00004480, -0.00008003, -0.00016580, 0.00000837, 0.00013068, -0.00008212, -0.00012338, -0.00013182, 0.00003703, 0.00003468, -0.00040334, 0.00042910, 0.00017109, -0.00081947, 0.00045706, 0.00004171, -0.00061593, 0.00037717, -0.00022417, -0.00019821, -0.00054479, -0.00132687, 0.00060125, 0.00147039, 0.00092252, 0.00114656, 0.00049120, -0.00041062, },
		{ -0.00012661, -0.00016621, -0.00004128, -0.00004344, -0.00014911, -0.00006106, 0.00011730, 0.00002795, -0.00020404, -0.00008605, -0.00002703, -0.00000921, -0.00017721, -0.00003396, 0.00048088, -0.00053472, -0.00029594, 0.00067697, -0.00073658, -0.00004196, 0.00056147, -0.00079902, -0.00000263, -0.00080716, -0.00144679, 0.00125896, 0.00197967, 0.00131975, 0.00094385, -0.00027919, -0.00070333, -0.00024905, },
		{ -0.00017592, -0.00004852, -0.00010108, -0.00009596, 0.00004882, 0.00008172, -0.00012283, -0.00019554, 0.00001142, -0.00005028, -0.00011906, -0.00007583, 0.00023985, -0.00000407, -0.00065851, 0.00044660, 0.00001902, -0.00091384, 0.00094377, -0.00020352, -0.00106604, 0.00061260, -0.00163455, -0.00141600, 0.00257526, 0.00264410, 0.00114701, 0.00006307, -0.00092558, -0.00051344, -0.00000896, -0.00015847, },
		{ -0.00003682, -0.00009379, -0.00001324, 0.00006003, -0.00003381, -0.00018562, -0.00009855, 0.00002292, -0.00013622, -0.00006928, 0.00011052, 0.00009713, -0.00041367, -0.00014272, 0.00053153, -0.00084336, -0.00000824, 0.00115268, -0.00154947, -0.00013679, 0.00073347, -0.00323577, -0.00021437, 0.00458355, 0.00251934, 0.00000923, -0.00087313, -0.00090381, -0.00005629, 0.00000480, -0.00028842, -0.00014657, },
		{ -0.00006721, 0.00002296, -0.00000915, -0.00011822, -0.00015073, -0.00003232, -0.00004113, -0.00012623, 0.00007086, 0.00007832, -0.00021044, -0.00032381, 0.00032200, -0.00011121, -0.00086883, 0.00114000, -0.00007137, -0.00197443, 0.00159813, -0.00094014, -0.00434248, 0.00303166, 0.00595284, 0.00072425, -0.00138358, -0.00100111, -0.00035882, 0.00014538, -0.00022714, -0.00031470, -0.00004032, -0.00002776, },
		{ 0.00000291, -0.00009045, -0.00012910, -0.00009224, -0.00003003, -0.00007037, -0.00002701, 0.00008927, -0.00012650, -0.00028355, 0.00001910, 0.00022416, -0.00066706, 0.00002063, 0.00119385, -0.00172884, -0.00051358, 0.00217225, -0.00415651, -0.00272050, 0.00729814, 0.00472018, -0.00210474, -0.00166017, -0.00031401, -0.00002188, -0.00006120, -0.00037496, -0.00015733, 0.00003131, -0.00005430, -0.00004221, },
		{ -0.00013262, -0.00008965, -0.00005921, -0.00003856, -0.00003575, 0.00001979, -0.00004101, -0.00023548, -0.00010976, 0.00014610, -0.00017135, -0.00051827, 0.00079557, -0.00002645, -0.00205548, 0.00170118, 0.00001190, -0.00630585, 0.00237012, 0.00952969, 0.00052953, -0.00379664, -0.00039901, 0.00017334, -0.00025237, -0.00025160, -0.00024371, -0.00000882, -0.00000738, -0.00006894, -0.00002685, -0.00005722, },
		{ -0.00004078, -0.00004933, -0.00002558, -0.00001262, -0.00004137, -0.00015039, -0.00015849, 0.00002455, 0.00000619, -0.00031621, 0.00000516, 0.00061959, -0.00123355, -0.00061949, 0.00227696, -0.00380173, -0.00450726, 0.00839365, 0.00720042, -0.00417274, -0.00272162, 0.00107067, -0.00032263, -0.00051454, -0.00007065, -0.00007463, -0.00003384, -0.00005305, -0.00004144, -0.00005813, -0.00011888, -0.00011296, },
		{ -0.00004771, -0.00001668, -0.00005156, -0.00011823, -0.00013261, -0.00005245, -0.00000583, -0.00009559, -0.00015403, 0.00018939, -0.00013926, -0.00122636, 0.00101736, 0.00021950, -0.00597499, 0.00118674, 0.01110221, 0.00118674, -0.00597499, 0.00021950, 0.00101736, -0.00122636, -0.00013926, 0.00018939, -0.00015403, -0.00009559, -0.00000583, -0.00005245, -0.00013261, -0.00011823, -0.00005156, -0.00001668, },
		{ -0.00004078, -0.00011296, -0.00011888, -0.00005813, -0.00004144, -0.00005305, -0.00003384, -0.00007463, -0.00007065, -0.00051454, -0.00032263, 0.00107067, -0.00272162, -0.00417274, 0.00720042, 0.00839365, -0.00450726, -0.00380173, 0.00227696, -0.00061949, -0.00123355, 0.00061959, 0.00000516, -0.00031621, 0.00000619, 0.00002455, -0.00015849, -0.00015039, -0.00004137, -0.00001262, -0.00002558, -0.00004933, },
		{ -0.00013262, -0.00005722, -0.00002685, -0.00006894, -0.00000738, -0.00000882, -0.00024371, -0.00025160, -0.00025237, 0.00017334, -0.00039901, -0.00379664, 0.00052953, 0.00952969, 0.00237012, -0.00630585, 0.00001190, 0.00170118, -0.00205548, -0.00002645, 0.00079557, -0.00051827, -0.00017135, 0.00014610, -0.00010976, -0.00023548, -0.00004101, 0.00001979, -0.00003575, -0.00003856, -0.00005921, -0.00008965, },
		{ 0.00000291, -0.00004221, -0.00005430, 0.00003131, -0.00015733, -0.00037496, -0.00006120, -0.00002188, -0.00031401, -0.00166017, -0.00210474, 0.00472018, 0.00729814, -0.00272050, -0.00415651, 0.00217225, -0.00051358, -0.00172884, 0.00119385, 0.00002063, -0.00066706, 0.00022416, 0.00001910, -0.00028355, -0.00012650, 0.00008927, -0.00002701, -0.00007037, -0.00003003, -0.00009224, -0.00012910, -0.00009045, },
		{ -0.00006721, -0.00002776, -0.00004032, -0.00031470, -0.00022714, 0.00014538, -0.00035882, -0.00100111, -0.00138358, 0.00072425, 0.00595284, 0.00303166, -0.00434248, -0.00094014, 0.00159813, -0.00197443, -0.00007137, 0.00114000, -0.00086883, -0.00011121, 0.00032200, -0.00032381, -0.00021044, 0.00007832, 0.00007086, -0.00012623, -0.00004113, -0.00003232, -0.00015073, -0.00011822, -0.00000915, 0.00002296, },
		{ -0.00003682, -0.00014657, -0.00028842, 0.00000480, -0.00005629, -0.00090381, -0.00087313, 0.00000923, 0.00251934, 0.00458355, -0.00021437, -0.00323577, 0.00073347, -0.00013679, -0.00154947, 0.00115268, -0.00000824, -0.00084336, 0.00053153, -0.00014272, -0.00041367, 0.00009713, 0.00011052, -0.00006928, -0.00013622, 0.00002292, -0.00009855, -0.00018562, -0.00003381, 0.00006003, -0.00001324, -0.00009379, },
		{ -0.00017592, -0.00015847, -0.00000896, -0.00051344, -0.00092558, 0.00006307, 0.00114701, 0.00264410, 0.00257526, -0.00141600, -0.00163455, 0.00061260, -0.00106604, -0.00020352, 0.00094377, -0.00091384, 0.00001902, 0.00044660, -0.00065851, -0.00000407, 0.00023985, -0.00007583, -0.00011906, -0.00005028, 0.00001142, -0.00019554, -0.00012283, 0.00008172, 0.00004882, -0.00009596, -0.00010108, -0.00004852, },
		{ -0.00012661, -0.00024905, -0.00070333, -0.00027919, 0.00094385, 0.00131975, 0.00197967, 0.00125896, -0.00144679, -0.00080716, -0.00000263, -0.00079902, 0.00056147, -0.00004196, -0.00073658, 0.00067697, -0.00029594, -0.00053472, 0.00048088, -0.00003396, -0.00017721, -0.00000921, -0.00002703, -0.00008605, -0.00020404, 0.00002795, 0.00011730, -0.00006106, -0.00014911, -0.00004344, -0.00004128, -0.00016621, },
		{ -0.00040435, -0.00041062, 0.00049120, 0.00114656, 0.00092252, 0.00147039, 0.00060125, -0.00132687, -0.00054479, -0.00019821, -0.00022417, 0.00037717, -0.00061593, 0.00004171, 0.00045706, -0.00081947, 0.00017109, 0.00042910, -0.00040334, 0.00003468, 0.00003703, -0.00013182, -0.00012338, -0.00008212, 0.00013068, 0.00000837, -0.00016580, -0.00008003, 0.00004480, -0.00008680, -0.00020124, -0.00019940, },
		{ 0.00008764, 0.00085904, 0.00082250, 0.00061770, 0.00125167, 0.00009817, -0.00124972, -0.00030122, -0.00010029, -0.00003175, -0.00007620, -0.00045121, 0.00047163, -0.00022515, -0.00055691, 0.00069098, -0.00014143, -0.00034547, 0.00031931, -0.00017042, -0.00017755, -0.00000940, 0.00003250, 0.00006808, -0.00014087, -0.00011942, 0.00006908, 0.00002596, -0.00021689, -0.00022322, -0.00022184, -0.00031004, },
		{ 0.00079041, 0.00044982, 0.00058757, 0.00102116, -0.00037108, -0.00100983, -0.00003131, -0.00010725, -0.00010928, -0.00023613, -0.00008259, 0.00030052, -0.00056412, 0.00011537, 0.00051216, -0.00057360, 0.00014788, 0.00024068, -0.00043148, 0.00002466, 0.00010382, -0.00001642, -0.00007894, -0.00013879, 0.00006059, 0.00010679, -0.00014883, -0.00029196, -0.00012413, -0.00015618, -0.00008801, 0.00044379, },
		{ 0.00026823, 0.00063560, 0.00066498, -0.00063402, -0.00063291, 0.00005302, -0.00022275, -0.00012001, -0.00018659, 0.00004217, -0.00000994, -0.00036145, 0.00047889, -0.00004361, -0.00040561, 0.00052067, -0.00021108, -0.00035400, 0.00032252, -0.00000817, -0.00016021, -0.00009562, 0.00004189, 0.00015122, -0.00006478, -0.00029638, -0.00016440, 0.00002692, -0.00011326, 0.00011628, 0.00056764, 0.00054003, },
		{ 0.00060669, 0.00030841, -0.00064976, -0.00033533, -0.00004153, -0.00024329, -0.00005292, -0.00019572, 0.00000827, -0.00008887, -0.00003690, 0.00035763, -0.00033700, 0.00009250, 0.00036972, -0.00054186, 0.00009295, 0.00033741, -0.00029153, -0.00015120, 0.00008120, 0.00017563, -0.00000755, -0.00025488, -0.00020608, 0.00006334, 0.00004478, -0.00011397, 0.00029352, 0.00051887, 0.00029879, 0.00024396, },
		{ 0.00005297, -0.00055010, -0.00019899, -0.00013436, -0.00015310, -0.00006915, -0.00026060, 0.00003807, -0.00002839, 0.00008690, 0.00012965, -0.00020814, 0.00036015, -0.00005985, -0.00037948, 0.00045875, -0.00003295, -0.00040287, 0.00009917, 0.00023443, 0.00004740, -0.00025058, -0.00018726, 0.00003452, 0.00013246, -0.00007503, -0.00003762, 0.00042582, 0.00034063, 0.00015203, 0.00028860, 0.00049218, },
		{ -0.00043872, -0.00014820, -0.00015008, -0.00008513, -0.00017226, -0.00024850, 0.00011156, -0.00001861, 0.00010657, 0.00007772, 0.00000581, 0.00026737, -0.00025753, 0.00005065, 0.00035789, -0.00041197, -0.00013564, 0.00034276, 0.00012193, -0.00023533, -0.00024228, 0.00009993, 0.00010654, 0.00000472, -0.00012451, 0.00013494, 0.00041006, 0.00011761, 0.00014959, 0.00031776, 0.00032370, -0.00009309, },
		{ -0.00012123, -0.00013928, -0.00008541, -0.00025280, -0.00017509, 0.00011536, -0.00001705, 0.00019024, 0.00010054, 0.00006571, 0.00013462, -0.00011679, 0.00025282, -0.00003849, -0.00038244, 0.00026586, 0.00027253, -0.00018575, -0.00032744, 0.00008062, 0.00023481, -0.00009051, -0.00005656, -0.00001986, 0.00023323, 0.00024130, 0.00000524, 0.00024215, 0.00024611, 0.00016127, -0.00014204, -0.00034640, },
		},
		{ { 0.00123669, 0.00022541, -0.00001892, 0.00003640, -0.00005969, -0.00024421, -0.00006401, -0.00013513, -0.00005525, -0.00016828, 0.00022471, -0.00019235, 0.00009853, -0.00004030, 0.00001647, -0.00008576, 0.00003333, -0.00008576, 0.00001647, -0.00004030, 0.00009853, -0.00019235, 0.00022471, -0.00016828, -0.00005525, -0.00013513, -0.00006401, -0.00024421, -0.00005969, 0.00003640, -0.00001892, 0.00022541, },
		{ 0.00029361, 0.00019073, -0.00000433, -0.00010760, -0.00034559, 0.00000709, -0.00014584, -0.00012175, -0.00006296, 0.00016696, -0.00019687, 0.00012457, -0.00006365, 0.00001116, -0.00006316, 0.00002338, -0.00009668, 0.00002394, -0.00001869, 0.00005947, -0.00015394, 0.00024861, -0.00027881, 0.00005833, -0.00017929, -0.00009598, -0.00012849, -0.00006269, 0.00006612, -0.00020360, 0.00014358, 0.00125763, },
		{ 0.00042497, -0.00008507, -0.00019729, -0.00040038, 0.00007910, -0.00019266, -0.00013224, 0.00000980, 0.00009768, -0.00017252, 0.00013279, -0.00008505, 0.00001188, -0.00003737, -0.00000354, -0.00009044, 0.00002824, -0.00000079, 0.00001521, -0.00008536, 0.00022358, -0.00036212, 0.00019164, -0.00027590, -0.00006041, -0.00004314, -0.00009906, 0.00009972, -0.00036913, 0.00004443, 0.00131541, 0.00036331, },
		{ -0.00023355, -0.00029640, -0.00039979, 0.00011906, -0.00024467, -0.00009728, 0.00003553, 0.00003937, -0.00013115, 0.00012290, -0.00010025, 0.00001964, -0.00001716, -0.00003955, -0.00006637, 0.00002449, 0.00001356, -0.00002580, 0.00000162, 0.00014687, -0.00039054, 0.00030636, -0.00039959, 0.00004864, -0.00002865, -0.00013179, 0.00013033, -0.00051403, -0.00007421, 0.00139943, 0.00046373, 0.00066456, },
		{ -0.00036253, -0.00036004, 0.00011131, -0.00027132, -0.00004141, 0.00001587, 0.00000798, -0.00008751, 0.00009926, -0.00010592, 0.00003228, -0.00000887, -0.00007423, -0.00002893, 0.00001018, 0.00002621, -0.00005696, 0.00009033, 0.00002963, -0.00034999, 0.00036490, -0.00050827, 0.00020908, -0.00010490, -0.00011999, 0.00013235, -0.00062173, -0.00022412, 0.00150475, 0.00063133, 0.00086568, -0.00045916, },
		{ -0.00031425, 0.00006234, -0.00025416, 0.00000669, -0.00003345, 0.00000872, -0.00005463, 0.00006958, -0.00010090, 0.00004528, -0.00001448, -0.00009776, 0.00001359, -0.00001380, 0.00003962, -0.00007528, 0.00016386, -0.00010581, -0.00024487, 0.00034251, -0.00055588, 0.00037482, -0.00026089, -0.00003689, 0.00007383, -0.00066222, -0.00043318, 0.00164503, 0.00089467, 0.00096522, -0.00074112, -0.00036093, },
		{ -0.00000316, -0.00019352, 0.00002482, -0.00008912, 0.00003542, -0.00004003, 0.00004254, -0.00008682, 0.00005339, -0.00003122, -0.00010384, 0.00005156, -0.00004312, 0.00005552, -0.00008183, 0.00020967, -0.00023198, -0.00009737, 0.00023593, -0.00050751, 0.00048864, -0.00045481, 0.00011618, -0.00006636, -0.00060320, -0.00074243, 0.00186017, 0.00125623, 0.00089676, -0.00102711, -0.00028046, -0.00029768, },
		{ -0.00010782, 0.00000513, -0.00012884, 0.00007366, -0.00004387, 0.00002515, -0.00006784, 0.00005261, -0.00005267, -0.00009136, 0.00007685, -0.00007122, 0.00007383, -0.00008093, 0.00022288, -0.00032385, 0.00005866, 0.00006549, -0.00035197, 0.00049989, -0.00062490, 0.00030557, -0.00028364, -0.00042683, -0.00119324, 0.00221386, 0.00167858, 0.00061297, -0.00124408, -0.00014146, -0.00033200, -0.00005333, },
		{ -0.00004374, -0.00013889, 0.00010639, -0.00005954, 0.00002065, -0.00004945, 0.00004165, -0.00007100, -0.00006449, 0.00008517, -0.00009131, 0.00009225, -0.00007821, 0.00020707, -0.00036562, 0.00018737, -0.00013006, -0.00010751, 0.00038156, -0.00070806, 0.00047279, -0.00054119, -0.00014522, -0.00180789, 0.00277888, 0.00208134, 0.00010687, -0.00131881, 0.00000772, -0.00041443, -0.00006309, -0.00002595, },
		{ -0.00011832, 0.00011996, -0.00007653, 0.00002769, -0.00003679, 0.00002253, -0.00007965, -0.00003106, 0.00007713, -0.00009847, 0.00010675, -0.00007839, 0.00017259, -0.00035463, 0.00026117, -0.00030449, 0.00018141, 0.00014083, -0.00066035, 0.00055195, -0.00077534, 0.00019192, -0.00257041, 0.00361329, 0.00235263, -0.00057556, -0.00120170, 0.00010492, -0.00051596, -0.00002572, 0.00002444, -0.00010059, },
		{ 0.00010863, -0.00008430, 0.00004088, -0.00003302, 0.00000004, -0.00007542, 0.00000013, 0.00005775, -0.00009125, 0.00011277, -0.00008349, 0.00013283, -0.00030119, 0.00026825, -0.00041748, 0.00045560, -0.00018045, -0.00047259, 0.00049155, -0.00091225, 0.00050220, -0.00341531, 0.00473413, 0.00237351, -0.00133725, -0.00088614, 0.00009491, -0.00058987, 0.00004222, 0.00002841, -0.00014031, -0.00007824, },
		{ -0.00007608, 0.00005272, -0.00003823, -0.00001990, -0.00005940, 0.00002206, 0.00003451, -0.00007215, 0.00010679, -0.00009214, 0.00009984, -0.00022469, 0.00021520, -0.00044618, 0.00065748, -0.00051592, -0.00017579, 0.00027359, -0.00089085, 0.00068821, -0.00423001, 0.00609741, 0.00204992, -0.00204752, -0.00041643, -0.00005044, -0.00058768, 0.00010511, -0.00000919, -0.00014404, -0.00003686, 0.00007592, },
		{ 0.00005610, -0.00004928, -0.00003215, -0.00003636, 0.00003156, 0.00001484, -0.00004685, 0.00008773, -0.00010012, 0.00008074, -0.00014712, 0.00012394, -0.00039134, 0.00074712, -0.00079260, 0.00016571, -0.00007683, -0.00068493, 0.00066386, -0.00487289, 0.00759199, 0.00134266, -0.00257138, 0.00011918, -0.00031937, -0.00047751, 0.00012004, -0.00006410, -0.00010711, -0.00001187, 0.00003264, -0.00005103, },
		{ -0.00006072, -0.00003417, -0.00001297, 0.00002992, 0.00000365, -0.00002248, 0.00005760, -0.00010206, 0.00007591, -0.00008633, 0.00002415, -0.00027615, 0.00071318, -0.00095115, 0.00047162, -0.00049497, -0.00031642, 0.00038160, -0.00520348, 0.00905163, 0.00028653, -0.00280297, 0.00060908, -0.00065958, -0.00025806, 0.00005276, -0.00010029, -0.00004183, -0.00001327, -0.00000777, -0.00001430, 0.00004664, },
		{ -0.00002692, 0.00000460, 0.00002191, 0.00000191, -0.00000534, 0.00002128, -0.00009356, 0.00007968, -0.00005119, -0.00005625, -0.00013834, 0.00057519, -0.00096213, 0.00066831, -0.00089251, 0.00014501, -0.00014809, -0.00511741, 0.01028378, -0.00100837, -0.00269341, 0.00094707, -0.00099074, 0.00003666, -0.00010815, -0.00008274, 0.00002605, -0.00003906, -0.00003467, 0.00002510, 0.00002402, -0.00006643, },
		{ 0.00001290, 0.00001369, 0.00000665, 0.00000099, -0.00001466, -0.00007305, 0.00008305, -0.00003993, -0.00009944, -0.00001811, 0.00037680, -0.00083343, 0.00070970, -0.00118160, 0.00059711, -0.00085191, -0.00457602, 0.01110872, -0.00237477, -0.00226486, 0.00106074, -0.00122651, 0.00034410, -0.00034369, 0.00000913, 0.00006606, -0.00007557, -0.00004451, 0.00005775, -0.00000795, -0.00006149, -0.00001462, },
		{ 0.00001031, 0.00001228, -0.00000335, -0.00004377, -0.00004269, 0.00007735, -0.00004233, -0.00010226, 0.00005408, 0.00017201, -0.00060664, 0.00058995, -0.00129954, 0.00093103, -0.00160630, -0.00362188, 0.01139905, -0.00362188, -0.00160630, 0.00093103, -0.00129954, 0.00058995, -0.00060664, 0.00017201, 0.00005408, -0.00010226, -0.00004233, 0.00007735, -0.00004269, -0.00004377, -0.00000335, 0.00001228, },
		{ 0.00001290, -0.00001462, -0.00006149, -0.00000795, 0.00005775, -0.00004451, -0.00007557, 0.00006606, 0.00000913, -0.00034369, 0.00034410, -0.00122651, 0.00106074, -0.00226486, -0.00237477, 0.01110872, -0.00457602, -0.00085191, 0.00059711, -0.00118160, 0.00070970, -0.00083343, 0.00037680, -0.00001811, -0.00009944, -0.00003993, 0.00008305, -0.00007305, -0.00001466, 0.00000099, 0.00000665, 0.00001369, },
		{ -0.00002692, -0.00006643, 0.00002402, 0.00002510, -0.00003467, -0.00003906, 0.00002605, -0.00008274, -0.00010815, 0.00003666, -0.00099074, 0.00094707, -0.00269341, -0.00100837, 0.01028378, -0.00511741, -0.00014809, 0.00014501, -0.00089251, 0.00066831, -0.00096213, 0.00057519, -0.00013834, -0.00005625, -0.00005119, 0.00007968, -0.00009356, 0.00002128, -0.00000534, 0.00000191, 0.00002191, 0.00000460, },
		{ -0.00006072, 0.00004664, -0.00001430, -0.00000777, -0.00001327, -0.00004183, -0.00010029, 0.00005276, -0.00025806, -0.00065958, 0.00060908, -0.00280297, 0.00028653, 0.00905163, -0.00520348, 0.00038160, -0.00031642, -0.00049497, 0.00047162, -0.00095115, 0.00071318, -0.00027615, 0.00002415, -0.00008633, 0.00007591, -0.00010206, 0.00005760, -0.00002248, 0.00000365, 0.00002992, -0.00001297, -0.00003417, },
		{ 0.00005610, -0.00005103, 0.00003264, -0.00001187, -0.00010711, -0.00006410, 0.00012004, -0.00047751, -0.00031937, 0.00011918, -0.00257138, 0.00134266, 0.00759199, -0.00487289, 0.00066386, -0.00068493, -0.00007683, 0.00016571, -0.00079260, 0.00074712, -0.00039134, 0.00012394, -0.00014712, 0.00008074, -0.00010012, 0.00008773, -0.00004685, 0.00001484, 0.00003156, -0.00003636, -0.00003215, -0.00004928, },
		{ -0.00007608, 0.00007592, -0.00003686, -0.00014404, -0.00000919, 0.00010511, -0.00058768, -0.00005044, -0.00041643, -0.00204752, 0.00204992, 0.00609741, -0.00423001, 0.00068821, -0.00089085, 0.00027359, -0.00017579, -0.00051592, 0.00065748, -0.00044618, 0.00021520, -0.00022469, 0.00009984, -0.00009214, 0.00010679, -0.00007215, 0.00003451, 0.00002206, -0.00005940, -0.00001990, -0.00003823, 0.00005272, },
		{ 0.00010863, -0.00007824, -0.00014031, 0.00002841, 0.00004222, -0.00058987, 0.00009491, -0.00088614, -0.00133725, 0.00237351, 0.00473413, -0.00341531, 0.00050220, -0.00091225, 0.00049155, -0.00047259, -0.00018045, 0.00045560, -0.00041748, 0.00026825, -0.00030119, 0.00013283, -0.00008349, 0.00011277, -0.00009125, 0.00005775, 0.00000013, -0.00007542, 0.00000004, -0.00003302, 0.00004088, -0.00008430, },
		{ -0.00011832, -0.00010059, 0.00002444, -0.00002572, -0.00051596, 0.00010492, -0.00120170, -0.00057556, 0.00235263, 0.00361329, -0.00257041, 0.00019192, -0.00077534, 0.00055195, -0.00066035, 0.00014083, 0.00018141, -0.00030449, 0.00026117, -0.00035463, 0.00017259, -0.00007839, 0.00010675, -0.00009847, 0.00007713, -0.00003106, -0.00007965, 0.00002253, -0.00003679, 0.00002769, -0.00007653, 0.00011996, },
		{ -0.00004374, -0.00002595, -0.00006309, -0.00041443, 0.00000772, -0.00131881, 0.00010687, 0.00208134, 0.00277888, -0.00180789, -0.00014522, -0.00054119, 0.00047279, -0.00070806, 0.00038156, -0.00010751, -0.00013006, 0.00018737, -0.00036562, 0.00020707, -0.00007821, 0.00009225, -0.00009131, 0.00008517, -0.00006449, -0.00007100, 0.00004165, -0.00004945, 0.00002065, -0.00005954, 0.00010639, -0.00013889, },
		{ -0.00010782, -0.00005333, -0.00033200, -0.00014146, -0.00124408, 0.00061297, 0.00167858, 0.00221386, -0.00119324, -0.00042683, -0.00028364, 0.00030557, -0.00062490, 0.00049989, -0.00035197, 0.00006549, 0.00005866, -0.00032385, 0.00022288, -0.00008093, 0.00007383, -0.00007122, 0.00007685, -0.00009136, -0.00005267, 0.00005261, -0.00006784, 0.00002515, -0.00004387, 0.00007366, -0.00012884, 0.00000513, },
		{ -0.00000316, -0.00029768, -0.00028046, -0.00102711, 0.00089676, 0.00125623, 0.00186017, -0.00074243, -0.00060320, -0.00006636, 0.00011618, -0.00045481, 0.00048864, -0.00050751, 0.00023593, -0.00009737, -0.00023198, 0.00020967, -0.00008183, 0.00005552, -0.00004312, 0.00005156, -0.00010384, -0.00003122, 0.00005339, -0.00008682, 0.00004254, -0.00004003, 0.00003542, -0.00008912, 0.00002482, -0.00019352, },
		{ -0.00031425, -0.00036093, -0.00074112, 0.00096522, 0.00089467, 0.00164503, -0.00043318, -0.00066222, 0.00007383, -0.00003689, -0.00026089, 0.00037482, -0.00055588, 0.00034251, -0.00024487, -0.00010581, 0.00016386, -0.00007528, 0.00003962, -0.00001380, 0.00001359, -0.00009776, -0.00001448, 0.00004528, -0.00010090, 0.00006958, -0.00005463, 0.00000872, -0.00003345, 0.00000669, -0.00025416, 0.00006234, },
		{ -0.00036253, -0.00045916, 0.00086568, 0.00063133, 0.00150475, -0.00022412, -0.00062173, 0.00013235, -0.00011999, -0.00010490, 0.00020908, -0.00050827, 0.00036490, -0.00034999, 0.00002963, 0.00009033, -0.00005696, 0.00002621, 0.00001018, -0.00002893, -0.00007423, -0.00000887, 0.00003228, -0.00010592, 0.00009926, -0.00008751, 0.00000798, 0.00001587, -0.00004141, -0.00027132, 0.00011131, -0.00036004, },
		{ -0.00023355, 0.00066456, 0.00046373, 0.00139943, -0.00007421, -0.00051403, 0.00013033, -0.00013179, -0.00002865, 0.00004864, -0.00039959, 0.00030636, -0.00039054, 0.00014687, 0.00000162, -0.00002580, 0.00001356, 0.00002449, -0.00006637, -0.00003955, -0.00001716, 0.00001964, -0.00010025, 0.00012290, -0.00013115, 0.00003937, 0.00003553, -0.00009728, -0.00024467, 0.00011906, -0.00039979, -0.00029640, },
		{ 0.00042497, 0.00036331, 0.00131541, 0.00004443, -0.00036913, 0.00009972, -0.00009906, -0.00004314, -0.00006041, -0.00027590, 0.00019164, -0.00036212, 0.00022358, -0.00008536, 0.00001521, -0.00000079, 0.00002824, -0.00009044, -0.00000354, -0.00003737, 0.00001188, -0.00008505, 0.00013279, -0.00017252, 0.00009768, 0.00000980, -0.00013224, -0.00019266, 0.00007910, -0.00040038, -0.00019729, -0.00008507, },
		{ 0.00029361, 0.00125763, 0.00014358, -0.00020360, 0.00006612, -0.00006269, -0.00012849, -0.00009598, -0.00017929, 0.00005833, -0.00027881, 0.00024861, -0.00015394, 0.00005947, -0.00001869, 0.00002394, -0.00009668, 0.00002338, -0.00006316, 0.00001116, -0.00006365, 0.00012457, -0.00019687, 0.00016696, -0.00006296, -0.00012175, -0.00014584, 0.00000709, -0.00034559, -0.00010760, -0.00000433, 0.00019073, },
		},
		{ { -0.00083402, -0.00002813, 0.00056232, 0.00028092, 0.00063061, 0.00009826, -0.00055346, -0.00006087, -0.00010937, -0.00010421, 0.00001477, -0.00003215, -0.00010839, -0.00004347, 0.00007707, -0.00002404, -0.00002044, -0.00002404, 0.00007707, -0.00004347, -0.00010839, -0.00003215, 0.00001477, -0.00010421, -0.00010937, -0.00006087, -0.00055346, 0.00009826, 0.00063061, 0.00028092, 0.00056232, -0.00002813, },
		{ -0.00008655, 0.00044499, 0.00022740, 0.00046354, 0.00032119, -0.00044839, -0.00019071, -0.00002703, -0.00014509, 0.00001335, -0.00003112, -0.00006835, -0.00010264, 0.00007902, 0.00001766, -0.00004098, -0.00001281, 0.00002265, 0.00003087, -0.00013276, -0.00004330, 0.00000936, -0.00005899, -0.00016388, -0.00001615, -0.00048010, -0.00026720, 0.00075155, 0.00038995, 0.00068315, 0.00005356, -0.00084692, },
		{ 0.00032513, 0.00020700, 0.00031167, 0.00040172, -0.00026469, -0.00028360, -0.00001364, -0.00011705, -0.00003434, -0.00000596, -0.00005632, -0.00010285, 0.00001009, 0.00009915, -0.00007395, 0.00000869, -0.00002190, 0.00005617, -0.00009620, -0.00009620, 0.00003508, -0.00006300, -0.00013246, -0.00009500, -0.00026022, -0.00064137, 0.00069237, 0.00059695, 0.00079737, 0.00017850, -0.00089667, -0.00011917, },
		{ 0.00021951, 0.00017452, 0.00041513, -0.00012854, -0.00025127, -0.00010235, -0.00002299, -0.00011110, 0.00003201, -0.00006904, -0.00005885, -0.00008512, 0.00015774, -0.00006203, -0.00001088, -0.00001077, 0.00001027, -0.00000679, -0.00017042, 0.00007570, -0.00009913, -0.00005459, -0.00021246, -0.00005181, -0.00082036, 0.00033810, 0.00089235, 0.00092045, 0.00036152, -0.00099413, -0.00011854, 0.00018912, },
		{ 0.00005354, 0.00041190, -0.00008279, -0.00012186, -0.00021871, 0.00005332, -0.00014602, 0.00003322, -0.00006596, -0.00002327, -0.00013856, 0.00013604, 0.00002148, -0.00007551, 0.00003981, -0.00006080, 0.00006850, -0.00020014, 0.00007248, -0.00009990, -0.00002531, -0.00023952, -0.00001239, -0.00068748, -0.00027066, 0.00113502, 0.00113369, 0.00057994, -0.00111575, -0.00010724, 0.00005026, 0.00024791, },
		{ 0.00037934, -0.00007240, -0.00001068, -0.00024277, 0.00002098, -0.00008385, -0.00003166, -0.00002224, -0.00003001, -0.00011974, 0.00003581, 0.00013789, -0.00013032, 0.00006516, -0.00008182, 0.00006057, -0.00013775, -0.00000815, -0.00002718, -0.00009335, -0.00013023, -0.00015709, -0.00033348, -0.00088694, 0.00107283, 0.00154225, 0.00080030, -0.00121542, -0.00013364, -0.00004993, 0.00024794, -0.00001914, },
		{ -0.00002631, -0.00000863, -0.00012327, -0.00012238, 0.00004510, -0.00012288, 0.00003243, -0.00005888, -0.00006142, -0.00007868, 0.00020724, -0.00010805, 0.00001485, -0.00002027, -0.00003220, -0.00001372, -0.00012106, 0.00007231, -0.00020168, 0.00002436, -0.00033287, -0.00001848, -0.00117832, 0.00050486, 0.00213955, 0.00106745, -0.00128800, -0.00021830, -0.00008966, 0.00019012, -0.00001746, 0.00028950, },
		{ -0.00009647, 0.00006446, -0.00026829, 0.00013048, -0.00015083, 0.00003492, -0.00005801, -0.00002623, -0.00013251, 0.00017299, 0.00000643, -0.00009215, 0.00007569, -0.00013003, 0.00007273, -0.00016637, 0.00010006, -0.00023519, 0.00007381, -0.00034540, 0.00001963, -0.00096291, -0.00050945, 0.00268046, 0.00156680, -0.00141146, -0.00031299, -0.00010900, 0.00010338, 0.00003016, 0.00016459, 0.00005594, },
		{ 0.00017835, -0.00028221, 0.00007283, -0.00005784, -0.00004399, -0.00000330, -0.00004553, -0.00010221, 0.00005149, 0.00015224, -0.00017332, 0.00011328, -0.00013584, 0.00003647, -0.00008126, 0.00000566, -0.00013814, -0.00004590, -0.00013804, -0.00023403, -0.00038328, -0.00157294, 0.00271209, 0.00254894, -0.00168280, -0.00034476, -0.00017867, 0.00005312, 0.00005384, 0.00007512, 0.00010490, -0.00016900, },
		{ -0.00012887, -0.00011431, 0.00010922, -0.00015076, 0.00006443, -0.00008828, -0.00003328, -0.00007768, 0.00023397, -0.00014954, 0.00003917, -0.00002685, -0.00010600, 0.00008226, -0.00014026, 0.00000968, -0.00023201, 0.00013474, -0.00053329, 0.00014238, -0.00212075, 0.00181768, 0.00410524, -0.00205388, -0.00033742, -0.00030044, 0.00005382, 0.00002227, 0.00006250, 0.00006945, -0.00014827, 0.00014035, },
		{ -0.00028858, 0.00021141, -0.00017501, 0.00006702, -0.00009053, 0.00000196, -0.00013377, 0.00019635, -0.00001737, -0.00010197, 0.00010964, -0.00022905, 0.00017520, -0.00018091, 0.00004111, -0.00028436, 0.00021666, -0.00054738, 0.00019142, -0.00176879, -0.00005034, 0.00593683, -0.00219832, -0.00049824, -0.00033559, 0.00000261, 0.00000908, 0.00006807, -0.00000052, -0.00007620, 0.00001899, 0.00007583, },
		{ 0.00012185, -0.00004108, -0.00003690, -0.00001998, -0.00003113, -0.00009708, 0.00006906, 0.00013795, -0.00019159, 0.00013741, -0.00019176, 0.00006765, -0.00000321, -0.00015302, -0.00007813, -0.00003629, -0.00012669, -0.00035274, -0.00060855, -0.00239146, 0.00732922, -0.00157483, -0.00114164, -0.00006814, -0.00027129, 0.00015563, -0.00003978, 0.00002600, -0.00008978, -0.00001984, 0.00014547, -0.00028006, },
		{ 0.00020972, -0.00019747, 0.00008484, -0.00010033, -0.00001921, -0.00005687, 0.00020475, -0.00013138, -0.00001275, 0.00003658, -0.00022731, 0.00034336, -0.00050354, 0.00030534, -0.00051249, 0.00055153, -0.00118631, 0.00078214, -0.00431420, 0.00744026, 0.00026771, -0.00243777, 0.00059680, -0.00084332, 0.00053063, -0.00033862, 0.00023794, -0.00028882, 0.00013745, -0.00001937, -0.00003667, -0.00015737, },
		{ -0.00029506, 0.00014036, -0.00013811, 0.00002576, -0.00010188, 0.00013366, 0.00006840, -0.00026883, 0.00032661, -0.00053267, 0.00064992, -0.00078172, 0.00061035, -0.00089942, 0.00108303, -0.00176857, 0.00165638, -0.00498717, 0.00578140, 0.00330183, -0.00414421, 0.00144224, -0.00154289, 0.00100072, -0.00073175, 0.00056293, -0.00061252, 0.00042968, -0.00032780, 0.00030852, -0.00046805, 0.00042413, },
		{ 0.00009046, -0.00010509, 0.00000501, -0.00005871, -0.00002110, 0.00028301, -0.00045427, 0.00046836, -0.00062314, 0.00068909, -0.00075905, 0.00058999, -0.00090507, 0.00111629, -0.00168231, 0.00154727, -0.00411620, 0.00259185, 0.00683011, -0.00555433, 0.00194500, -0.00197298, 0.00124981, -0.00096172, 0.00077500, -0.00084643, 0.00064031, -0.00054605, 0.00051597, -0.00059534, 0.00043968, -0.00023015, },
		{ -0.00002744, -0.00004979, 0.00000894, -0.00014331, 0.00036098, -0.00041565, 0.00032767, -0.00039645, 0.00039090, -0.00038711, 0.00019907, -0.00047612, 0.00060277, -0.00091359, 0.00054642, -0.00213106, -0.00117113, 0.00968908, -0.00578274, 0.00154736, -0.00174298, 0.00097754, -0.00078238, 0.00065045, -0.00077432, 0.00055733, -0.00046983, 0.00041411, -0.00042597, 0.00021081, -0.00001078, -0.00003753, },
		{ -0.00007799, 0.00002631, -0.00014600, 0.00023849, -0.00014355, -0.00004217, 0.00003542, -0.00008880, 0.00014958, -0.00035924, 0.00015832, -0.00017006, 0.00014786, -0.00077583, 0.00005297, -0.00426665, 0.01078997, -0.00426665, 0.00005297, -0.00077583, 0.00014786, -0.00017006, 0.00015832, -0.00035924, 0.00014958, -0.00008880, 0.00003542, -0.00004217, -0.00014355, 0.00023849, -0.00014600, 0.00002631, },
		{ -0.00002744, -0.00003753, -0.00001078, 0.00021081, -0.00042597, 0.00041411, -0.00046983, 0.00055733, -0.00077432, 0.00065045, -0.00078238, 0.00097754, -0.00174298, 0.00154736, -0.00578274, 0.00968908, -0.00117113, -0.00213106, 0.00054642, -0.00091359, 0.00060277, -0.00047612, 0.00019907, -0.00038711, 0.00039090, -0.00039645, 0.00032767, -0.00041565, 0.00036098, -0.00014331, 0.00000894, -0.00004979, },
		{ 0.00009046, -0.00023015, 0.00043968, -0.00059534, 0.00051597, -0.00054605, 0.00064031, -0.00084643, 0.00077500, -0.00096172, 0.00124981, -0.00197298, 0.00194500, -0.00555433, 0.00683011, 0.00259185, -0.00411620, 0.00154727, -0.00168231, 0.00111629, -0.00090507, 0.00058999, -0.00075905, 0.00068909, -0.00062314, 0.00046836, -0.00045427, 0.00028301, -0.00002110, -0.00005871, 0.00000501, -0.00010509, },
		{ -0.00029506, 0.00042413, -0.00046805, 0.00030852, -0.00032780, 0.00042968, -0.00061252, 0.00056293, -0.00073175, 0.00100072, -0.00154289, 0.00144224, -0.00414421, 0.00330183, 0.00578140, -0.00498717, 0.00165638, -0.00176857, 0.00108303, -0.00089942, 0.00061035, -0.00078172, 0.00064992, -0.00053267, 0.00032661, -0.00026883, 0.00006840, 0.00013366, -0.00010188, 0.00002576, -0.00013811, 0.00014036, },
		{ 0.00020972, -0.00015737, -0.00003667, -0.00001937, 0.00013745, -0.00028882, 0.00023794, -0.00033862, 0.00053063, -0.00084332, 0.00059680, -0.00243777, 0.00026771, 0.00744026, -0.00431420, 0.00078214, -0.00118631, 0.00055153, -0.00051249, 0.00030534, -0.00050354, 0.00034336, -0.00022731, 0.00003658, -0.00001275, -0.00013138, 0.00020475, -0.00005687, -0.00001921, -0.00010033, 0.00008484, -0.00019747, },
		{ 0.00012185, -0.00028006, 0.00014547, -0.00001984, -0.00008978, 0.00002600, -0.00003978, 0.00015563, -0.00027129, -0.00006814, -0.00114164, -0.00157483, 0.00732922, -0.00239146, -0.00060855, -0.00035274, -0.00012669, -0.00003629, -0.00007813, -0.00015302, -0.00000321, 0.00006765, -0.00019176, 0.00013741, -0.00019159, 0.00013795, 0.00006906, -0.00009708, -0.00003113, -0.00001998, -0.00003690, -0.00004108, },
		{ -0.00028858, 0.00007583, 0.00001899, -0.00007620, -0.00000052, 0.00006807, 0.00000908, 0.00000261, -0.00033559, -0.00049824, -0.00219832, 0.00593683, -0.00005034, -0.00176879, 0.00019142, -0.00054738, 0.00021666, -0.00028436, 0.00004111, -0.00018091, 0.00017520, -0.00022905, 0.00010964, -0.00010197, -0.00001737, 0.00019635, -0.00013377, 0.00000196, -0.00009053, 0.00006702, -0.00017501, 0.00021141, },
		{ -0.00012887, 0.00014035, -0.00014827, 0.00006945, 0.00006250, 0.00002227, 0.00005382, -0.00030044, -0.00033742, -0.00205388, 0.00410524, 0.00181768, -0.00212075, 0.00014238, -0.00053329, 0.00013474, -0.00023201, 0.00000968, -0.00014026, 0.00008226, -0.00010600, -0.00002685, 0.00003917, -0.00014954, 0.00023397, -0.00007768, -0.00003328, -0.00008828, 0.00006443, -0.00015076, 0.00010922, -0.00011431, },
		{ 0.00017835, -0.00016900, 0.00010490, 0.00007512, 0.00005384, 0.00005312, -0.00017867, -0.00034476, -0.00168280, 0.00254894, 0.00271209, -0.00157294, -0.00038328, -0.00023403, -0.00013804, -0.00004590, -0.00013814, 0.00000566, -0.00008126, 0.00003647, -0.00013584, 0.00011328, -0.00017332, 0.00015224, 0.00005149, -0.00010221, -0.00004553, -0.00000330, -0.00004399, -0.00005784, 0.00007283, -0.00028221, },
		{ -0.00009647, 0.00005594, 0.00016459, 0.00003016, 0.00010338, -0.00010900, -0.00031299, -0.00141146, 0.00156680, 0.00268046, -0.00050945, -0.00096291, 0.00001963, -0.00034540, 0.00007381, -0.00023519, 0.00010006, -0.00016637, 0.00007273, -0.00013003, 0.00007569, -0.00009215, 0.00000643, 0.00017299, -0.00013251, -0.00002623, -0.00005801, 0.00003492, -0.00015083, 0.00013048, -0.00026829, 0.00006446, },
		{ -0.00002631, 0.00028950, -0.00001746, 0.00019012, -0.00008966, -0.00021830, -0.00128800, 0.00106745, 0.00213955, 0.00050486, -0.00117832, -0.00001848, -0.00033287, 0.00002436, -0.00020168, 0.00007231, -0.00012106, -0.00001372, -0.00003220, -0.00002027, 0.00001485, -0.00010805, 0.00020724, -0.00007868, -0.00006142, -0.00005888, 0.00003243, -0.00012288, 0.00004510, -0.00012238, -0.00012327, -0.00000863, },
		{ 0.00037934, -0.00001914, 0.00024794, -0.00004993, -0.00013364, -0.00121542, 0.00080030, 0.00154225, 0.00107283, -0.00088694, -0.00033348, -0.00015709, -0.00013023, -0.00009335, -0.00002718, -0.00000815, -0.00013775, 0.00006057, -0.00008182, 0.00006516, -0.00013032, 0.00013789, 0.00003581, -0.00011974, -0.00003001, -0.00002224, -0.00003166, -0.00008385, 0.00002098, -0.00024277, -0.00001068, -0.00007240, },
		{ 0.00005354, 0.00024791, 0.00005026, -0.00010724, -0.00111575, 0.00057994, 0.00113369, 0.00113502, -0.00027066, -0.00068748, -0.00001239, -0.00023952, -0.00002531, -0.00009990, 0.00007248, -0.00020014, 0.00006850, -0.00006080, 0.00003981, -0.00007551, 0.00002148, 0.00013604, -0.00013856, -0.00002327, -0.00006596, 0.00003322, -0.00014602, 0.00005332, -0.00021871, -0.00012186, -0.00008279, 0.00041190, },
		{ 0.00021951, 0.00018912, -0.00011854, -0.00099413, 0.00036152, 0.00092045, 0.00089235, 0.00033810, -0.00082036, -0.00005181, -0.00021246, -0.00005459, -0.00009913, 0.00007570, -0.00017042, -0.00000679, 0.00001027, -0.00001077, -0.00001088, -0.00006203, 0.00015774, -0.00008512, -0.00005885, -0.00006904, 0.00003201, -0.00011110, -0.00002299, -0.00010235, -0.00025127, -0.00012854, 0.00041513, 0.00017452, },
		{ 0.00032513, -0.00011917, -0.00089667, 0.00017850, 0.00079737, 0.00059695, 0.00069237, -0.00064137, -0.00026022, -0.00009500, -0.00013246, -0.00006300, 0.00003508, -0.00009620, -0.00009620, 0.00005617, -0.00002190, 0.00000869, -0.00007395, 0.00009915, 0.00001009, -0.00010285, -0.00005632, -0.00000596, -0.00003434, -0.00011705, -0.00001364, -0.00028360, -0.00026469, 0.00040172, 0.00031167, 0.00020700, },
		{ -0.00008655, -0.00084692, 0.00005356, 0.00068315, 0.00038995, 0.00075155, -0.00026720, -0.00048010, -0.00001615, -0.00016388, -0.00005899, 0.00000936, -0.00004330, -0.00013276, 0.00003087, 0.00002265, -0.00001281, -0.00004098, 0.00001766, 0.00007902, -0.00010264, -0.00006835, -0.00003112, 0.00001335, -0.00014509, -0.00002703, -0.00019071, -0.00044839, 0.00032119, 0.00046354, 0.00022740, 0.00044499, },
		},
		{ { -0.00007307, -0.00008864, 0.00006659, -0.00012783, -0.00048446, 0.00006835, 0.00067649, 0.00030854, 0.00060062, 0.00009637, -0.00051250, -0.00012291, -0.00001186, -0.00001859, -0.00018743, -0.00002898, -0.00004921, -0.00002898, -0.00018743, -0.00001859, -0.00001186, -0.00012291, -0.00051250, 0.00009637, 0.00060062, 0.00030854, 0.00067649, 0.00006835, -0.00048446, -0.00012783, 0.00006659, -0.00008864, },
		{ -0.00008170, -0.00002367, 0.00004535, -0.00040866, -0.00014644, 0.00046720, 0.00040260, 0.00031899, 0.00045848, -0.00027847, -0.00032965, -0.00002228, 0.00000473, -0.00015098, -0.00010253, -0.00001008, -0.00005255, -0.00010817, -0.00012921, 0.00004701, -0.00008716, -0.00029457, -0.00051639, 0.00064018, 0.00046072, 0.00064424, 0.00049535, -0.00054340, -0.00026863, 0.00001429, -0.00000064, -0.00009869, },
		{ -0.00009440, 0.00008542, -0.00023132, -0.00028601, 0.00026620, 0.00040678, 0.00023331, 0.00038296, 0.00013885, -0.00037443, -0.00014075, 0.00002509, -0.00009501, -0.00016839, 0.00001118, -0.00007437, -0.00004021, -0.00018110, 0.00003368, -0.00006401, -0.00010042, -0.00061036, -0.00004956, 0.00090860, 0.00045819, 0.00096058, -0.00042635, -0.00044143, -0.00007772, 0.00003814, -0.00007194, -0.00007592, },
		{ -0.00000859, -0.00003592, -0.00032161, 0.00008056, 0.00035800, 0.00021853, 0.00025191, 0.00028820, -0.00011694, -0.00029504, 0.00000130, -0.00004059, -0.00018439, -0.00003406, -0.00002561, -0.00007909, -0.00008075, -0.00012624, 0.00008254, -0.00018165, -0.00019097, -0.00081993, 0.00080408, 0.00070420, 0.00102490, 0.00014421, -0.00077918, -0.00006143, -0.00009009, 0.00005302, -0.00012189, -0.00007220, },
		{ 0.00001653, -0.00021196, -0.00008925, 0.00027467, 0.00024005, 0.00012751, 0.00030577, 0.00005698, -0.00017658, -0.00018534, 0.00006640, -0.00021201, -0.00006647, -0.00000053, -0.00009998, -0.00001665, -0.00018359, 0.00008171, -0.00015246, -0.00003407, -0.00071169, -0.00023655, 0.00118478, 0.00094820, 0.00085413, -0.00094214, -0.00017737, -0.00015448, 0.00005469, -0.00007046, -0.00008333, -0.00006126, },
		{ -0.00011501, -0.00010623, 0.00007097, 0.00035224, -0.00002758, 0.00031523, 0.00007915, 0.00003036, -0.00025678, 0.00002890, -0.00013449, -0.00015114, 0.00002799, -0.00008977, -0.00004022, -0.00008768, -0.00008847, 0.00004544, -0.00021301, -0.00012861, -0.00104315, 0.00074004, 0.00135275, 0.00125249, -0.00057272, -0.00061434, -0.00001669, -0.00013749, 0.00007209, -0.00012785, -0.00005634, -0.00001482, },
		{ -0.00002413, -0.00008647, 0.00032111, 0.00004970, 0.00008100, 0.00026603, -0.00004744, 0.00000007, -0.00020564, 0.00005848, -0.00029916, 0.00008849, -0.00010706, -0.00003978, -0.00004093, -0.00013439, 0.00005498, -0.00018948, -0.00001436, -0.00071274, -0.00058630, 0.00154851, 0.00181200, -0.00000550, -0.00099180, -0.00000436, -0.00022014, 0.00005283, -0.00005421, -0.00010121, -0.00000471, -0.00011815, },
		{ 0.00000696, 0.00002371, 0.00035646, -0.00025588, 0.00042780, -0.00011851, 0.00014634, -0.00021531, 0.00004656, -0.00025804, -0.00001526, -0.00000919, -0.00010617, 0.00000658, -0.00011274, -0.00002625, -0.00006199, -0.00015598, -0.00015558, -0.00110629, 0.00029105, 0.00277663, 0.00071532, -0.00114629, -0.00022583, -0.00013924, -0.00009831, 0.00003739, -0.00010752, -0.00002857, -0.00008363, -0.00006293, },
		{ -0.00020500, 0.00043069, -0.00019368, 0.00013996, 0.00018308, -0.00008799, 0.00008605, -0.00015990, -0.00004345, -0.00026253, 0.00018542, -0.00024372, 0.00011252, -0.00017275, 0.00004166, -0.00014641, -0.00003237, -0.00021874, -0.00038955, -0.00141662, 0.00248554, 0.00252210, -0.00143133, -0.00037667, -0.00019392, -0.00009116, -0.00007693, 0.00003946, -0.00014849, 0.00003831, -0.00020692, 0.00017862, },
		{ 0.00010541, 0.00021489, -0.00030155, 0.00042546, -0.00018000, 0.00014487, -0.00010607, -0.00002292, -0.00030336, 0.00014310, -0.00014731, 0.00000334, -0.00005526, -0.00003889, -0.00007157, -0.00008143, -0.00018007, -0.00008107, -0.00125076, -0.00026154, 0.00488177, -0.00096280, -0.00094183, -0.00009633, -0.00016691, -0.00011992, 0.00003952, -0.00010535, 0.00001096, -0.00017981, 0.00010291, -0.00007219, },
		{ 0.00042456, -0.00032831, 0.00018117, 0.00012412, -0.00010990, 0.00012927, -0.00011364, -0.00016739, -0.00011517, 0.00015357, -0.00029612, 0.00024999, -0.00034293, 0.00029176, -0.00048147, 0.00030919, -0.00066147, 0.00029552, -0.00286678, 0.00446041, 0.00222557, -0.00262412, 0.00051573, -0.00064176, 0.00020657, -0.00032402, 0.00021219, -0.00023149, 0.00005784, -0.00017780, 0.00021788, -0.00022771, },
		{ -0.00003084, -0.00015954, 0.00027010, -0.00008145, 0.00002139, 0.00006548, -0.00018695, -0.00020061, 0.00020641, -0.00028728, 0.00020498, -0.00026786, 0.00021955, -0.00035193, 0.00017373, -0.00052363, 0.00040500, -0.00194855, -0.00017108, 0.00683064, -0.00337176, 0.00017476, -0.00053356, 0.00014150, -0.00036810, 0.00019777, -0.00019436, 0.00007603, -0.00023727, 0.00023624, -0.00020423, 0.00024069, },
		{ -0.00021074, 0.00006268, 0.00019382, -0.00020170, 0.00021837, -0.00016933, -0.00017329, 0.00001217, 0.00004500, -0.00021738, 0.00024621, -0.00039055, 0.00042315, -0.00071636, 0.00057292, -0.00103255, 0.00093886, -0.00447541, 0.00727818, 0.00048315, -0.00275293, 0.00098936, -0.00092291, 0.00040648, -0.00050676, 0.00035384, -0.00030746, 0.00007208, -0.00008099, 0.00009119, -0.00007353, 0.00018973, },
		{ -0.00001079, 0.00009998, 0.00004626, -0.00009438, 0.00016951, -0.00030908, -0.00003379, 0.00015938, -0.00033592, 0.00037554, -0.00052827, 0.00058016, -0.00080257, 0.00067392, -0.00118577, 0.00134099, -0.00340221, 0.00121122, 0.00747488, -0.00540511, 0.00160046, -0.00131889, 0.00070071, -0.00080168, 0.00050547, -0.00039038, 0.00018613, -0.00024453, 0.00023344, -0.00020584, 0.00022876, -0.00017236, },
		{ -0.00017656, 0.00035452, -0.00033038, 0.00032492, -0.00030219, -0.00002719, 0.00001704, -0.00002876, -0.00007070, 0.00007867, -0.00022103, 0.00027606, -0.00061301, 0.00044415, -0.00090689, 0.00089329, -0.00526599, 0.00987995, -0.00216746, -0.00168365, 0.00066109, -0.00070788, 0.00026022, -0.00042929, 0.00032074, -0.00029502, 0.00008045, -0.00002127, -0.00003978, 0.00010780, -0.00014452, 0.00007792, },
		{ 0.00004825, 0.00007695, -0.00008539, 0.00009302, -0.00023497, 0.00002863, 0.00010845, -0.00029899, 0.00041426, -0.00066668, 0.00082352, -0.00115334, 0.00110758, -0.00179341, 0.00225486, -0.00510813, 0.00371999, 0.00624991, -0.00603632, 0.00244380, -0.00187492, 0.00113009, -0.00115114, 0.00078391, -0.00059760, 0.00032386, -0.00025361, 0.00014726, -0.00007220, 0.00001319, -0.00009186, -0.00000368, },
		{ 0.00043992, -0.00034718, 0.00030267, -0.00033824, 0.00012502, -0.00004832, -0.00002818, 0.00004288, -0.00014746, 0.00008251, -0.00009265, -0.00018704, -0.00010344, -0.00019226, -0.00013868, -0.00445032, 0.01094670, -0.00445032, -0.00013868, -0.00019226, -0.00010344, -0.00018704, -0.00009265, 0.00008251, -0.00014746, 0.00004288, -0.00002818, -0.00004832, 0.00012502, -0.00033824, 0.00030267, -0.00034718, },
		{ 0.00004825, -0.00000368, -0.00009186, 0.00001319, -0.00007220, 0.00014726, -0.00025361, 0.00032386, -0.00059760, 0.00078391, -0.00115114, 0.00113009, -0.00187492, 0.00244380, -0.00603632, 0.00624991, 0.00371999, -0.00510813, 0.00225486, -0.00179341, 0.00110758, -0.00115334, 0.00082352, -0.00066668, 0.00041426, -0.00029899, 0.00010845, 0.00002863, -0.00023497, 0.00009302, -0.00008539, 0.00007695, },
		{ -0.00017656, 0.00007792, -0.00014452, 0.00010780, -0.00003978, -0.00002127, 0.00008045, -0.00029502, 0.00032074, -0.00042929, 0.00026022, -0.00070788, 0.00066109, -0.00168365, -0.00216746, 0.00987995, -0.00526599, 0.00089329, -0.00090689, 0.00044415, -0.00061301, 0.00027606, -0.00022103, 0.00007867, -0.00007070, -0.00002876, 0.00001704, -0.00002719, -0.00030219, 0.00032492, -0.00033038, 0.00035452, },
		{ -0.00001079, -0.00017236, 0.00022876, -0.00020584, 0.00023344, -0.00024453, 0.00018613, -0.00039038, 0.00050547, -0.00080168, 0.00070071, -0.00131889, 0.00160046, -0.00540511, 0.00747488, 0.00121122, -0.00340221, 0.00134099, -0.00118577, 0.00067392, -0.00080257, 0.00058016, -0.00052827, 0.00037554, -0.00033592, 0.00015938, -0.00003379, -0.00030908, 0.00016951, -0.00009438, 0.00004626, 0.00009998, },
		{ -0.00021074, 0.00018973, -0.00007353, 0.00009119, -0.00008099, 0.00007208, -0.00030746, 0.00035384, -0.00050676, 0.00040648, -0.00092291, 0.00098936, -0.00275293, 0.00048315, 0.00727818, -0.00447541, 0.00093886, -0.00103255, 0.00057292, -0.00071636, 0.00042315, -0.00039055, 0.00024621, -0.00021738, 0.00004500, 0.00001217, -0.00017329, -0.00016933, 0.00021837, -0.00020170, 0.00019382, 0.00006268, },
		{ -0.00003084, 0.00024069, -0.00020423, 0.00023624, -0.00023727, 0.00007603, -0.00019436, 0.00019777, -0.00036810, 0.00014150, -0.00053356, 0.00017476, -0.00337176, 0.00683064, -0.00017108, -0.00194855, 0.00040500, -0.00052363, 0.00017373, -0.00035193, 0.00021955, -0.00026786, 0.00020498, -0.00028728, 0.00020641, -0.00020061, -0.00018695, 0.00006548, 0.00002139, -0.00008145, 0.00027010, -0.00015954, },
		{ 0.00042456, -0.00022771, 0.00021788, -0.00017780, 0.00005784, -0.00023149, 0.00021219, -0.00032402, 0.00020657, -0.00064176, 0.00051573, -0.00262412, 0.00222557, 0.00446041, -0.00286678, 0.00029552, -0.00066147, 0.00030919, -0.00048147, 0.00029176, -0.00034293, 0.00024999, -0.00029612, 0.00015357, -0.00011517, -0.00016739, -0.00011364, 0.00012927, -0.00010990, 0.00012412, 0.00018117, -0.00032831, },
		{ 0.00010541, -0.00007219, 0.00010291, -0.00017981, 0.00001096, -0.00010535, 0.00003952, -0.00011992, -0.00016691, -0.00009633, -0.00094183, -0.00096280, 0.00488177, -0.00026154, -0.00125076, -0.00008107, -0.00018007, -0.00008143, -0.00007157, -0.00003889, -0.00005526, 0.00000334, -0.00014731, 0.00014310, -0.00030336, -0.00002292, -0.00010607, 0.00014487, -0.00018000, 0.00042546, -0.00030155, 0.00021489, },
		{ -0.00020500, 0.00017862, -0.00020692, 0.00003831, -0.00014849, 0.00003946, -0.00007693, -0.00009116, -0.00019392, -0.00037667, -0.00143133, 0.00252210, 0.00248554, -0.00141662, -0.00038955, -0.00021874, -0.00003237, -0.00014641, 0.00004166, -0.00017275, 0.00011252, -0.00024372, 0.00018542, -0.00026253, -0.00004345, -0.00015990, 0.00008605, -0.00008799, 0.00018308, 0.00013996, -0.00019368, 0.00043069, },
		{ 0.00000696, -0.00006293, -0.00008363, -0.00002857, -0.00010752, 0.00003739, -0.00009831, -0.00013924, -0.00022583, -0.00114629, 0.00071532, 0.00277663, 0.00029105, -0.00110629, -0.00015558, -0.00015598, -0.00006199, -0.00002625, -0.00011274, 0.00000658, -0.00010617, -0.00000919, -0.00001526, -0.00025804, 0.00004656, -0.00021531, 0.00014634, -0.00011851, 0.00042780, -0.00025588, 0.00035646, 0.00002371, },
		{ -0.00002413, -0.00011815, -0.00000471, -0.00010121, -0.00005421, 0.00005283, -0.00022014, -0.00000436, -0.00099180, -0.00000550, 0.00181200, 0.00154851, -0.00058630, -0.00071274, -0.00001436, -0.00018948, 0.00005498, -0.00013439, -0.00004093, -0.00003978, -0.00010706, 0.00008849, -0.00029916, 0.00005848, -0.00020564, 0.00000007, -0.00004744, 0.00026603, 0.00008100, 0.00004970, 0.00032111, -0.00008647, },
		{ -0.00011501, -0.00001482, -0.00005634, -0.00012785, 0.00007209, -0.00013749, -0.00001669, -0.00061434, -0.00057272, 0.00125249, 0.00135275, 0.00074004, -0.00104315, -0.00012861, -0.00021301, 0.00004544, -0.00008847, -0.00008768, -0.00004022, -0.00008977, 0.00002799, -0.00015114, -0.00013449, 0.00002890, -0.00025678, 0.00003036, 0.00007915, 0.00031523, -0.00002758, 0.00035224, 0.00007097, -0.00010623, },
		{ 0.00001653, -0.00006126, -0.00008333, -0.00007046, 0.00005469, -0.00015448, -0.00017737, -0.00094214, 0.00085413, 0.00094820, 0.00118478, -0.00023655, -0.00071169, -0.00003407, -0.00015246, 0.00008171, -0.00018359, -0.00001665, -0.00009998, -0.00000053, -0.00006647, -0.00021201, 0.00006640, -0.00018534, -0.00017658, 0.00005698, 0.00030577, 0.00012751, 0.00024005, 0.00027467, -0.00008925, -0.00021196, },
		{ -0.00000859, -0.00007220, -0.00012189, 0.00005302, -0.00009009, -0.00006143, -0.00077918, 0.00014421, 0.00102490, 0.00070420, 0.00080408, -0.00081993, -0.00019097, -0.00018165, 0.00008254, -0.00012624, -0.00008075, -0.00007909, -0.00002561, -0.00003406, -0.00018439, -0.00004059, 0.00000130, -0.00029504, -0.00011694, 0.00028820, 0.00025191, 0.00021853, 0.00035800, 0.00008056, -0.00032161, -0.00003592, },
		{ -0.00009440, -0.00007592, -0.00007194, 0.00003814, -0.00007772, -0.00044143, -0.00042635, 0.00096058, 0.00045819, 0.00090860, -0.00004956, -0.00061036, -0.00010042, -0.00006401, 0.00003368, -0.00018110, -0.00004021, -0.00007437, 0.00001118, -0.00016839, -0.00009501, 0.00002509, -0.00014075, -0.00037443, 0.00013885, 0.00038296, 0.00023331, 0.00040678, 0.00026620, -0.00028601, -0.00023132, 0.00008542, },
		{ -0.00008170, -0.00009869, -0.00000064, 0.00001429, -0.00026863, -0.00054340, 0.00049535, 0.00064424, 0.00046072, 0.00064018, -0.00051639, -0.00029457, -0.00008716, 0.00004701, -0.00012921, -0.00010817, -0.00005255, -0.00001008, -0.00010253, -0.00015098, 0.00000473, -0.00002228, -0.00032965, -0.00027847, 0.00045848, 0.00031899, 0.00040260, 0.00046720, -0.00014644, -0.00040866, 0.00004535, -0.00002367, },
		},
		{ { -0.00004532, -0.00003855, -0.00013085, 0.00000398, -0.00006697, -0.00003741, -0.00017724, -0.00012844, -0.00040641, 0.00008568, 0.00078882, 0.00026726, 0.00060645, 0.00005347, -0.00043758, -0.00011967, -0.00013451, -0.00011967, -0.00043758, 0.00005347, 0.00060645, 0.00026726, 0.00078882, 0.00008568, -0.00040641, -0.00012844, -0.00017724, -0.00003741, -0.00006697, 0.00000398, -0.00013085, -0.00003855, },
		{ -0.00004693, -0.00008742, -0.00008868, 0.00002402, -0.00009355, -0.00004912, -0.00016753, -0.00020124, -0.00031210, 0.00044182, 0.00053844, 0.00024299, 0.00049194, -0.00009258, -0.00034526, -0.00010296, -0.00013756, -0.00014663, -0.00054896, 0.00030892, 0.00063536, 0.00044884, 0.00085746, -0.00041259, -0.00025110, -0.00016031, -0.00014789, -0.00002836, -0.00004197, -0.00005615, -0.00011119, -0.00001445, },
		{ -0.00005682, -0.00012687, -0.00000738, -0.00001218, -0.00010084, -0.00004187, -0.00020256, -0.00022367, -0.00009007, 0.00056568, 0.00029720, 0.00028762, 0.00032346, -0.00014153, -0.00028994, -0.00008385, -0.00014836, -0.00019743, -0.00064862, 0.00066549, 0.00058733, 0.00082152, 0.00048961, -0.00070530, -0.00001101, -0.00029025, -0.00004165, -0.00005489, -0.00002662, -0.00011782, -0.00004980, -0.00002335, },
		{ -0.00008985, -0.00011599, 0.00006180, -0.00009489, -0.00004320, -0.00008740, -0.00022454, -0.00015801, 0.00011422, 0.00051690, 0.00015045, 0.00031679, 0.00016104, -0.00014050, -0.00025166, -0.00008356, -0.00013264, -0.00033376, -0.00063288, 0.00102167, 0.00061746, 0.00114559, -0.00028077, -0.00051540, 0.00000275, -0.00031653, 0.00002998, -0.00006954, -0.00005789, -0.00011717, -0.00001086, -0.00003633, },
		{ -0.00014834, -0.00003158, 0.00005059, -0.00014804, 0.00004504, -0.00021748, -0.00012546, -0.00011509, 0.00029146, 0.00033867, 0.00014801, 0.00024513, 0.00009232, -0.00017458, -0.00017112, -0.00015153, -0.00003982, -0.00063134, -0.00036994, 0.00125218, 0.00092228, 0.00094287, -0.00092716, -0.00005825, -0.00022919, -0.00015917, -0.00000870, -0.00004204, -0.00013237, -0.00004031, -0.00003399, -0.00002780, },
		{ -0.00018322, 0.00008395, -0.00005929, -0.00007897, 0.00002620, -0.00029256, 0.00002559, -0.00011580, 0.00042376, 0.00011564, 0.00024057, 0.00008875, 0.00011512, -0.00024799, -0.00006801, -0.00024172, 0.00005474, -0.00100826, 0.00009931, 0.00147089, 0.00131951, 0.00006210, -0.00090540, 0.00014342, -0.00037701, -0.00001065, -0.00005438, -0.00005305, -0.00015015, 0.00002086, -0.00006860, -0.00003009, },
		{ -0.00012043, 0.00011953, -0.00016403, 0.00007015, -0.00014859, -0.00016500, 0.00004345, -0.00003945, 0.00040800, 0.00000328, 0.00027428, -0.00001802, 0.00011132, -0.00026010, -0.00005178, -0.00022118, -0.00002959, -0.00124815, 0.00060239, 0.00198018, 0.00114677, -0.00087410, -0.00036041, -0.00006317, -0.00024576, -0.00003568, -0.00000716, -0.00015909, -0.00004077, -0.00002594, -0.00003936, -0.00009632, },
		{ 0.00003034, 0.00001060, -0.00011765, 0.00010786, -0.00029615, 0.00006275, -0.00007635, 0.00013311, 0.00023924, 0.00004584, 0.00018866, -0.00001958, 0.00003783, -0.00019522, -0.00011887, -0.00010568, -0.00029950, -0.00132530, 0.00126010, 0.00264460, 0.00000899, -0.00100208, -0.00002346, -0.00025357, -0.00007668, -0.00009742, 0.00002322, -0.00023608, 0.00008951, -0.00012516, 0.00002468, -0.00019331, },
		{ 0.00013307, -0.00013864, 0.00006236, -0.00006371, -0.00020584, 0.00013186, -0.00012584, 0.00023140, 0.00009135, 0.00009857, 0.00010498, -0.00002937, 0.00000803, -0.00018609, -0.00010814, -0.00009872, -0.00054308, -0.00144522, 0.00257640, 0.00251984, -0.00138382, -0.00031498, -0.00025395, -0.00007261, -0.00015609, 0.00000436, -0.00011108, -0.00012710, 0.00007110, -0.00013515, 0.00001319, -0.00020181, },
		{ 0.00007047, -0.00013656, 0.00015299, -0.00025381, 0.00003476, 0.00000412, -0.00002119, 0.00017131, 0.00006365, 0.00007575, 0.00008532, -0.00008985, 0.00004193, -0.00023563, -0.00003061, -0.00019664, -0.00080545, -0.00141443, 0.00458469, 0.00072475, -0.00166041, 0.00017237, -0.00051613, 0.00018743, -0.00031827, 0.00014427, -0.00028487, 0.00007772, -0.00006905, -0.00004921, -0.00008428, -0.00007989, },
		{ -0.00007655, 0.00004412, -0.00000578, -0.00018619, 0.00010678, -0.00005716, 0.00003387, 0.00013279, 0.00000376, 0.00012768, -0.00003848, -0.00001091, -0.00008283, -0.00007570, -0.00022303, -0.00000106, -0.00163283, -0.00021280, 0.00595398, -0.00210424, -0.00039925, -0.00032361, -0.00014084, 0.00000319, -0.00017340, 0.00001726, -0.00021176, 0.00010993, -0.00011882, -0.00002596, -0.00012161, 0.00003473, },
		{ -0.00009323, 0.00017786, -0.00024881, 0.00010100, -0.00009027, 0.00010727, -0.00011811, 0.00026554, -0.00021019, 0.00035566, -0.00036303, 0.00029955, -0.00045145, 0.00037768, -0.00079788, 0.00061417, -0.00323405, 0.00303322, 0.00472132, -0.00379614, 0.00107043, -0.00122734, 0.00061801, -0.00052024, 0.00022210, -0.00032564, 0.00009581, -0.00007642, -0.00000897, -0.00013075, -0.00000764, -0.00001419, },
		{ 0.00008360, 0.00004963, -0.00024052, 0.00023587, -0.00024337, 0.00025222, -0.00025885, 0.00035832, -0.00033906, 0.00047692, -0.00056570, 0.00047065, -0.00061617, 0.00056198, -0.00106490, 0.00073503, -0.00434076, 0.00729971, 0.00053067, -0.00272112, 0.00101712, -0.00123453, 0.00079399, -0.00066902, 0.00031995, -0.00041550, 0.00023853, -0.00017781, 0.00003727, -0.00017648, 0.00010558, -0.00015797, },
		{ 0.00023683, -0.00023310, 0.00008239, -0.00001236, -0.00003825, 0.00005005, -0.00006117, 0.00009067, -0.00004567, 0.00011340, -0.00022673, 0.00004074, -0.00004220, -0.00020301, -0.00013565, -0.00093857, -0.00271878, 0.00953126, -0.00417160, 0.00022001, -0.00061973, -0.00002742, 0.00001904, -0.00011318, -0.00014477, -0.00000590, -0.00003528, 0.00003408, -0.00017018, 0.00002572, -0.00000641, -0.00014897, },
		{ 0.00012432, -0.00032521, 0.00037022, -0.00038137, 0.00035813, -0.00038008, 0.00036839, -0.00040744, 0.00051010, -0.00055888, 0.00045547, -0.00073756, 0.00094353, -0.00154896, 0.00159927, -0.00415495, 0.00237184, 0.00720198, -0.00597385, 0.00227747, -0.00205572, 0.00119288, -0.00087041, 0.00052956, -0.00066057, 0.00047905, -0.00040467, 0.00031871, -0.00043124, 0.00032359, -0.00028976, 0.00010140, },
		{ -0.00018335, -0.00007103, 0.00026763, -0.00041090, 0.00045899, -0.00054246, 0.00051934, -0.00057543, 0.00068893, -0.00082143, 0.00067538, -0.00091481, 0.00115244, -0.00197393, 0.00217339, -0.00630428, 0.00839537, 0.00118830, -0.00380059, 0.00170168, -0.00172908, 0.00113902, -0.00084494, 0.00044463, -0.00053677, 0.00042727, -0.00034680, 0.00024008, -0.00035376, 0.00033848, -0.00040110, 0.00034499, },
		{ -0.00035030, 0.00027476, -0.00013387, -0.00003188, 0.00009319, -0.00021168, 0.00014656, -0.00014326, 0.00016904, -0.00029790, 0.00001744, -0.00000921, -0.00007161, -0.00051307, 0.00001304, -0.00450569, 0.01110393, -0.00450569, 0.00001304, -0.00051307, -0.00007161, -0.00000921, 0.00001744, -0.00029790, 0.00016904, -0.00014326, 0.00014656, -0.00021168, 0.00009319, -0.00003188, -0.00013387, 0.00027476, },
		{ -0.00018335, 0.00034499, -0.00040110, 0.00033848, -0.00035376, 0.00024008, -0.00034680, 0.00042727, -0.00053677, 0.00044463, -0.00084494, 0.00113902, -0.00172908, 0.00170168, -0.00380059, 0.00118830, 0.00839537, -0.00630428, 0.00217339, -0.00197393, 0.00115244, -0.00091481, 0.00067538, -0.00082143, 0.00068893, -0.00057543, 0.00051934, -0.00054246, 0.00045899, -0.00041090, 0.00026763, -0.00007103, },
		{ 0.00012432, 0.00010140, -0.00028976, 0.00032359, -0.00043124, 0.00031871, -0.00040467, 0.00047905, -0.00066057, 0.00052956, -0.00087041, 0.00119288, -0.00205572, 0.00227747, -0.00597385, 0.00720198, 0.00237184, -0.00415495, 0.00159927, -0.00154896, 0.00094353, -0.00073756, 0.00045547, -0.00055888, 0.00051010, -0.00040744, 0.00036839, -0.00038008, 0.00035813, -0.00038137, 0.00037022, -0.00032521, },
		{ 0.00023683, -0.00014897, -0.00000641, 0.00002572, -0.00017018, 0.00003408, -0.00003528, -0.00000590, -0.00014477, -0.00011318, 0.00001904, -0.00002742, -0.00061973, 0.00022001, -0.00417160, 0.00953126, -0.00271878, -0.00093857, -0.00013565, -0.00020301, -0.00004220, 0.00004074, -0.00022673, 0.00011340, -0.00004567, 0.00009067, -0.00006117, 0.00005005, -0.00003825, -0.00001236, 0.00008239, -0.00023310, },
		{ 0.00008360, -0.00015797, 0.00010558, -0.00017648, 0.00003727, -0.00017781, 0.00023853, -0.00041550, 0.00031995, -0.00066902, 0.00079399, -0.00123453, 0.00101712, -0.00272112, 0.00053067, 0.00729971, -0.00434076, 0.00073503, -0.00106490, 0.00056198, -0.00061617, 0.00047065, -0.00056570, 0.00047692, -0.00033906, 0.00035832, -0.00025885, 0.00025222, -0.00024337, 0.00023587, -0.00024052, 0.00004963, },
		{ -0.00009323, -0.00001419, -0.00000764, -0.00013075, -0.00000897, -0.00007642, 0.00009581, -0.00032564, 0.00022210, -0.00052024, 0.00061801, -0.00122734, 0.00107043, -0.00379614, 0.00472132, 0.00303322, -0.00323405, 0.00061417, -0.00079788, 0.00037768, -0.00045145, 0.00029955, -0.00036303, 0.00035566, -0.00021019, 0.00026554, -0.00011811, 0.00010727, -0.00009027, 0.00010100, -0.00024881, 0.00017786, },
		{ -0.00007655, 0.00003473, -0.00012161, -0.00002596, -0.00011882, 0.00010993, -0.00021176, 0.00001726, -0.00017340, 0.00000319, -0.00014084, -0.00032361, -0.00039925, -0.00210424, 0.00595398, -0.00021280, -0.00163283, -0.00000106, -0.00022303, -0.00007570, -0.00008283, -0.00001091, -0.00003848, 0.00012768, 0.00000376, 0.00013279, 0.00003387, -0.00005716, 0.00010678, -0.00018619, -0.00000578, 0.00004412, },
		{ 0.00007047, -0.00007989, -0.00008428, -0.00004921, -0.00006905, 0.00007772, -0.00028487, 0.00014427, -0.00031827, 0.00018743, -0.00051613, 0.00017237, -0.00166041, 0.00072475, 0.00458469, -0.00141443, -0.00080545, -0.00019664, -0.00003061, -0.00023563, 0.00004193, -0.00008985, 0.00008532, 0.00007575, 0.00006365, 0.00017131, -0.00002119, 0.00000412, 0.00003476, -0.00025381, 0.00015299, -0.00013656, },
		{ 0.00013307, -0.00020181, 0.00001319, -0.00013515, 0.00007110, -0.00012710, -0.00011108, 0.00000436, -0.00015609, -0.00007261, -0.00025395, -0.00031498, -0.00138382, 0.00251984, 0.00257640, -0.00144522, -0.00054308, -0.00009872, -0.00010814, -0.00018609, 0.00000803, -0.00002937, 0.00010498, 0.00009857, 0.00009135, 0.00023140, -0.00012584, 0.00013186, -0.00020584, -0.00006371, 0.00006236, -0.00013864, },
		{ 0.00003034, -0.00019331, 0.00002468, -0.00012516, 0.00008951, -0.00023608, 0.00002322, -0.00009742, -0.00007668, -0.00025357, -0.00002346, -0.00100208, 0.00000899, 0.00264460, 0.00126010, -0.00132530, -0.00029950, -0.00010568, -0.00011887, -0.00019522, 0.00003783, -0.00001958, 0.00018866, 0.00004584, 0.00023924, 0.00013311, -0.00007635, 0.00006275, -0.00029615, 0.00010786, -0.00011765, 0.00001060, },
		{ -0.00012043, -0.00009632, -0.00003936, -0.00002594, -0.00004077, -0.00015909, -0.00000716, -0.00003568, -0.00024576, -0.00006317, -0.00036041, -0.00087410, 0.00114677, 0.00198018, 0.00060239, -0.00124815, -0.00002959, -0.00022118, -0.00005178, -0.00026010, 0.00011132, -0.00001802, 0.00027428, 0.00000328, 0.00040800, -0.00003945, 0.00004345, -0.00016500, -0.00014859, 0.00007015, -0.00016403, 0.00011953, },
		{ -0.00018322, -0.00003009, -0.00006860, 0.00002086, -0.00015015, -0.00005305, -0.00005438, -0.00001065, -0.00037701, 0.00014342, -0.00090540, 0.00006210, 0.00131951, 0.00147089, 0.00009931, -0.00100826, 0.00005474, -0.00024172, -0.00006801, -0.00024799, 0.00011512, 0.00008875, 0.00024057, 0.00011564, 0.00042376, -0.00011580, 0.00002559, -0.00029256, 0.00002620, -0.00007897, -0.00005929, 0.00008395, },
		{ -0.00014834, -0.00002780, -0.00003399, -0.00004031, -0.00013237, -0.00004204, -0.00000870, -0.00015917, -0.00022919, -0.00005825, -0.00092716, 0.00094287, 0.00092228, 0.00125218, -0.00036994, -0.00063134, -0.00003982, -0.00015153, -0.00017112, -0.00017458, 0.00009232, 0.00024513, 0.00014801, 0.00033867, 0.00029146, -0.00011509, -0.00012546, -0.00021748, 0.00004504, -0.00014804, 0.00005059, -0.00003158, },
		{ -0.00008985, -0.00003633, -0.00001086, -0.00011717, -0.00005789, -0.00006954, 0.00002998, -0.00031653, 0.00000275, -0.00051540, -0.00028077, 0.00114559, 0.00061746, 0.00102167, -0.00063288, -0.00033376, -0.00013264, -0.00008356, -0.00025166, -0.00014050, 0.00016104, 0.00031679, 0.00015045, 0.00051690, 0.00011422, -0.00015801, -0.00022454, -0.00008740, -0.00004320, -0.00009489, 0.00006180, -0.00011599, },
		{ -0.00005682, -0.00002335, -0.00004980, -0.00011782, -0.00002662, -0.00005489, -0.00004165, -0.00029025, -0.00001101, -0.00070530, 0.00048961, 0.00082152, 0.00058733, 0.00066549, -0.00064862, -0.00019743, -0.00014836, -0.00008385, -0.00028994, -0.00014153, 0.00032346, 0.00028762, 0.00029720, 0.00056568, -0.00009007, -0.00022367, -0.00020256, -0.00004187, -0.00010084, -0.00001218, -0.00000738, -0.00012687, },
		{ -0.00004693, -0.00001445, -0.00011119, -0.00005615, -0.00004197, -0.00002836, -0.00014789, -0.00016031, -0.00025110, -0.00041259, 0.00085746, 0.00044884, 0.00063536, 0.00030892, -0.00054896, -0.00014663, -0.00013756, -0.00010296, -0.00034526, -0.00009258, 0.00049194, 0.00024299, 0.00053844, 0.00044182, -0.00031210, -0.00020124, -0.00016753, -0.00004912, -0.00009355, 0.00002402, -0.00008868, -0.00008742, },
		},
		{ { -0.00001223, -0.00000667, -0.00000750, -0.00003071, 0.00002243, -0.00002936, 0.00000528, -0.00000261, -0.00006114, -0.00011500, -0.00012612, -0.00011706, -0.00051875, -0.00002385, 0.00017150, 0.00041158, 0.00121347, 0.00041158, 0.00017150, -0.00002385, -0.00051875, -0.00011706, -0.00012612, -0.00011500, -0.00006114, -0.00000261, 0.00000528, -0.00002936, 0.00002243, -0.00003071, -0.00000750, -0.00000667, },
		{ -0.00000885, -0.00001232, -0.00002518, -0.00000753, 0.00000762, -0.00001833, 0.00000078, -0.00002634, -0.00001227, -0.00017274, -0.00009292, -0.00015144, -0.00046170, 0.00013782, -0.00005111, 0.00031463, 0.00123441, 0.00050522, 0.00040966, -0.00027062, -0.00050242, -0.00011930, -0.00013864, -0.00006087, -0.00011799, 0.00003610, -0.00000699, -0.00002515, 0.00002536, -0.00004917, 0.00001498, -0.00000944, },
		{ -0.00000064, -0.00002682, -0.00003345, 0.00001378, -0.00001423, 0.00000637, -0.00002342, -0.00002381, 0.00000742, -0.00020628, -0.00006967, -0.00018759, -0.00036565, 0.00022331, -0.00023936, 0.00020036, 0.00129219, 0.00062387, 0.00061961, -0.00058088, -0.00040211, -0.00017910, -0.00010684, -0.00003696, -0.00015771, 0.00007167, -0.00002601, -0.00001006, 0.00001563, -0.00005753, 0.00003579, -0.00001662, },
		{ 0.00000737, -0.00004685, -0.00003095, 0.00002737, -0.00003563, 0.00003729, -0.00006115, 0.00000429, -0.00001071, -0.00019823, -0.00008006, -0.00019255, -0.00027290, 0.00026014, -0.00039371, 0.00006631, 0.00137621, 0.00080328, 0.00073790, -0.00090121, -0.00023894, -0.00029269, -0.00002562, -0.00005715, -0.00016144, 0.00008662, -0.00003927, 0.00000775, -0.00000345, -0.00005314, 0.00004864, -0.00002225, },
		{ 0.00000902, -0.00006637, -0.00001985, 0.00002979, -0.00004877, 0.00006339, -0.00009858, 0.00004470, -0.00005881, -0.00014926, -0.00012988, -0.00014993, -0.00021591, 0.00027747, -0.00052113, -0.00008832, 0.00148154, 0.00107155, 0.00069741, -0.00115727, -0.00006219, -0.00042988, 0.00008680, -0.00011570, -0.00012451, 0.00007135, -0.00003718, 0.00001967, -0.00002580, -0.00003686, 0.00004949, -0.00002073, },
		{ -0.00000048, -0.00007885, -0.00000493, 0.00002089, -0.00004804, 0.00007373, -0.00011869, 0.00007569, -0.00011468, -0.00007865, -0.00020310, -0.00006805, -0.00020417, 0.00029066, -0.00062162, -0.00027395, 0.00162181, 0.00143098, 0.00044970, -0.00127431, 0.00006494, -0.00054255, 0.00019307, -0.00018752, -0.00005923, 0.00002837, -0.00001730, 0.00002022, -0.00004474, -0.00001278, 0.00003791, -0.00000905, },
		{ -0.00002241, -0.00007964, 0.00000814, 0.00000368, -0.00003192, 0.00006147, -0.00010731, 0.00007467, -0.00014964, -0.00001766, -0.00026671, 0.00002159, -0.00022071, 0.00029363, -0.00068031, -0.00051708, 0.00183695, 0.00184421, -0.00001357, -0.00120112, 0.00008604, -0.00058077, 0.00024863, -0.00023594, 0.00000865, -0.00002798, 0.00001417, 0.00000926, -0.00005524, 0.00001312, 0.00001727, 0.00001178, },
		{ -0.00005360, -0.00006780, 0.00001485, -0.00001674, -0.00000361, 0.00002667, -0.00005874, 0.00002670, -0.00013915, 0.00000221, -0.00028279, 0.00007529, -0.00022927, 0.00026183, -0.00066906, -0.00085745, 0.00219064, 0.00223123, -0.00064808, -0.00092940, -0.00002820, -0.00051107, 0.00021766, -0.00022547, 0.00004866, -0.00007638, 0.00004434, -0.00000772, -0.00005552, 0.00003467, -0.00000627, 0.00003682, },
		{ -0.00008704, -0.00004663, 0.00001332, -0.00003482, 0.00002983, -0.00002309, 0.00002116, -0.00006901, -0.00007265, -0.00003811, -0.00022351, 0.00005376, -0.00018989, 0.00016463, -0.00055758, -0.00133523, 0.00275566, 0.00248074, -0.00135860, -0.00050175, -0.00026720, -0.00033061, 0.00008734, -0.00013525, 0.00003668, -0.00009634, 0.00005824, -0.00002155, -0.00004748, 0.00004732, -0.00002565, 0.00005856, },
		{ -0.00011390, -0.00002283, 0.00000492, -0.00004626, 0.00005958, -0.00007451, 0.00011559, -0.00019710, 0.00004093, -0.00013639, -0.00008386, -0.00006149, -0.00007688, -0.00001762, -0.00033004, -0.00197165, 0.00359007, 0.00247467, -0.00201640, -0.00000483, -0.00057984, -0.00007191, -0.00012450, 0.00003147, -0.00003551, -0.00007609, 0.00004468, -0.00002282, -0.00003574, 0.00004929, -0.00003519, 0.00006943, },
		{ -0.00012638, -0.00000429, -0.00000632, -0.00004917, 0.00007768, -0.00011247, 0.00020111, -0.00032935, 0.00017424, -0.00026716, 0.00011288, -0.00025795, 0.00010731, -0.00027874, -0.00000051, -0.00274991, 0.00471092, 0.00211996, -0.00248853, 0.00045126, -0.00088683, 0.00020387, -0.00037155, 0.00024542, -0.00015574, -0.00001793, 0.00000108, -0.00000564, -0.00002582, 0.00004183, -0.00003229, 0.00006430, },
		{ -0.00012031, 0.00000248, -0.00001523, -0.00004439, 0.00007948, -0.00012461, 0.00025431, -0.00043250, 0.00028896, -0.00038806, 0.00031925, -0.00049227, 0.00032781, -0.00058021, 0.00037946, -0.00360392, 0.00607419, 0.00137862, -0.00267111, 0.00076161, -0.00110277, 0.00042365, -0.00059085, 0.00045797, -0.00029355, 0.00006118, -0.00006437, 0.00002968, -0.00002198, 0.00002851, -0.00001823, 0.00004245, },
		{ -0.00009677, -0.00000533, -0.00001743, -0.00003484, 0.00006485, -0.00010525, 0.00025900, -0.00047810, 0.00034750, -0.00045259, 0.00047644, -0.00070113, 0.00052496, -0.00085668, 0.00072583, -0.00442076, 0.00756878, 0.00028677, -0.00251713, 0.00085487, -0.00116105, 0.00052390, -0.00072076, 0.00061456, -0.00040874, 0.00013411, -0.00013428, 0.00007595, -0.00002557, 0.00001385, 0.00000229, 0.00000803, },
		{ -0.00006210, -0.00002574, -0.00001097, -0.00002433, 0.00003819, -0.00005740, 0.00021128, -0.00045143, 0.00032511, -0.00042516, 0.00053217, -0.00081994, 0.00063168, -0.00103266, 0.00093980, -0.00505870, 0.00902842, -0.00104413, -0.00205048, 0.00071277, -0.00103404, 0.00047034, -0.00071965, 0.00067083, -0.00046358, 0.00017190, -0.00018691, 0.00012085, -0.00003439, 0.00000173, 0.00002277, -0.00003098, },
		{ -0.00002608, -0.00005243, 0.00000286, -0.00001604, 0.00000682, 0.00000799, 0.00012119, -0.00035635, 0.00021853, -0.00029378, 0.00045770, -0.00080339, 0.00059546, -0.00104562, 0.00093420, -0.00537749, 0.01026056, -0.00244541, -0.00136159, 0.00037495, -0.00074189, 0.00026928, -0.00057924, 0.00060639, -0.00043488, 0.00015320, -0.00020291, 0.00015060, -0.00004351, -0.00000579, 0.00003700, -0.00006506, },
		{ 0.00000084, -0.00007659, 0.00001986, -0.00001144, -0.00002130, 0.00007507, 0.00001000, -0.00021465, 0.00004809, -0.00007579, 0.00025852, -0.00064110, 0.00039734, -0.00086799, 0.00066078, -0.00527319, 0.01108550, -0.00372496, -0.00058580, -0.00007241, -0.00034741, -0.00003304, -0.00032827, 0.00043161, -0.00032189, 0.00007187, -0.00017176, 0.00015431, -0.00004697, -0.00000894, 0.00004103, -0.00008607, },
		{ 0.00001079, -0.00008965, 0.00003434, -0.00000985, -0.00004007, 0.00012775, -0.00009573, -0.00005998, -0.00014746, 0.00018538, -0.00002511, -0.00036311, 0.00006165, -0.00052064, 0.00012966, -0.00470785, 0.01137583, -0.00470785, 0.00012966, -0.00052064, 0.00006165, -0.00036311, -0.00002511, 0.00018538, -0.00014746, -0.00005998, -0.00009573, 0.00012775, -0.00004007, -0.00000985, 0.00003434, -0.00008965, },
		{ 0.00000084, -0.00008607, 0.00004103, -0.00000894, -0.00004697, 0.00015431, -0.00017176, 0.00007187, -0.00032189, 0.00043161, -0.00032827, -0.00003304, -0.00034741, -0.00007241, -0.00058580, -0.00372496, 0.01108550, -0.00527319, 0.00066078, -0.00086799, 0.00039734, -0.00064110, 0.00025852, -0.00007579, 0.00004809, -0.00021465, 0.00001000, 0.00007507, -0.00002130, -0.00001144, 0.00001986, -0.00007659, },
		{ -0.00002608, -0.00006506, 0.00003700, -0.00000579, -0.00004351, 0.00015060, -0.00020291, 0.00015320, -0.00043488, 0.00060639, -0.00057924, 0.00026928, -0.00074189, 0.00037495, -0.00136159, -0.00244541, 0.01026056, -0.00537749, 0.00093420, -0.00104562, 0.00059546, -0.00080339, 0.00045770, -0.00029378, 0.00021853, -0.00035635, 0.00012119, 0.00000799, 0.00000682, -0.00001604, 0.00000286, -0.00005243, },
		{ -0.00006210, -0.00003098, 0.00002277, 0.00000173, -0.00003439, 0.00012085, -0.00018691, 0.00017190, -0.00046358, 0.00067083, -0.00071965, 0.00047034, -0.00103404, 0.00071277, -0.00205048, -0.00104413, 0.00902842, -0.00505870, 0.00093980, -0.00103266, 0.00063168, -0.00081994, 0.00053217, -0.00042516, 0.00032511, -0.00045143, 0.00021128, -0.00005740, 0.00003819, -0.00002433, -0.00001097, -0.00002574, },
		{ -0.00009677, 0.00000803, 0.00000229, 0.00001385, -0.00002557, 0.00007595, -0.00013428, 0.00013411, -0.00040874, 0.00061456, -0.00072076, 0.00052390, -0.00116105, 0.00085487, -0.00251713, 0.00028677, 0.00756878, -0.00442076, 0.00072583, -0.00085668, 0.00052496, -0.00070113, 0.00047644, -0.00045259, 0.00034750, -0.00047810, 0.00025900, -0.00010525, 0.00006485, -0.00003484, -0.00001743, -0.00000533, },
		{ -0.00012031, 0.00004245, -0.00001823, 0.00002851, -0.00002198, 0.00002968, -0.00006437, 0.00006118, -0.00029355, 0.00045797, -0.00059085, 0.00042365, -0.00110277, 0.00076161, -0.00267111, 0.00137862, 0.00607419, -0.00360392, 0.00037946, -0.00058021, 0.00032781, -0.00049227, 0.00031925, -0.00038806, 0.00028896, -0.00043250, 0.00025431, -0.00012461, 0.00007948, -0.00004439, -0.00001523, 0.00000248, },
		{ -0.00012638, 0.00006430, -0.00003229, 0.00004183, -0.00002582, -0.00000564, 0.00000108, -0.00001793, -0.00015574, 0.00024542, -0.00037155, 0.00020387, -0.00088683, 0.00045126, -0.00248853, 0.00211996, 0.00471092, -0.00274991, -0.00000051, -0.00027874, 0.00010731, -0.00025795, 0.00011288, -0.00026716, 0.00017424, -0.00032935, 0.00020111, -0.00011247, 0.00007768, -0.00004917, -0.00000632, -0.00000429, },
		{ -0.00011390, 0.00006943, -0.00003519, 0.00004929, -0.00003574, -0.00002282, 0.00004468, -0.00007609, -0.00003551, 0.00003147, -0.00012450, -0.00007191, -0.00057984, -0.00000483, -0.00201640, 0.00247467, 0.00359007, -0.00197165, -0.00033004, -0.00001762, -0.00007688, -0.00006149, -0.00008386, -0.00013639, 0.00004093, -0.00019710, 0.00011559, -0.00007451, 0.00005958, -0.00004626, 0.00000492, -0.00002283, },
		{ -0.00008704, 0.00005856, -0.00002565, 0.00004732, -0.00004748, -0.00002155, 0.00005824, -0.00009634, 0.00003668, -0.00013525, 0.00008734, -0.00033061, -0.00026720, -0.00050175, -0.00135860, 0.00248074, 0.00275566, -0.00133523, -0.00055758, 0.00016463, -0.00018989, 0.00005376, -0.00022351, -0.00003811, -0.00007265, -0.00006901, 0.00002116, -0.00002309, 0.00002983, -0.00003482, 0.00001332, -0.00004663, },
		{ -0.00005360, 0.00003682, -0.00000627, 0.00003467, -0.00005552, -0.00000772, 0.00004434, -0.00007638, 0.00004866, -0.00022547, 0.00021766, -0.00051107, -0.00002820, -0.00092940, -0.00064808, 0.00223123, 0.00219064, -0.00085745, -0.00066906, 0.00026183, -0.00022927, 0.00007529, -0.00028279, 0.00000221, -0.00013915, 0.00002670, -0.00005874, 0.00002667, -0.00000361, -0.00001674, 0.00001485, -0.00006780, },
		{ -0.00002241, 0.00001178, 0.00001727, 0.00001312, -0.00005524, 0.00000926, 0.00001417, -0.00002798, 0.00000865, -0.00023594, 0.00024863, -0.00058077, 0.00008604, -0.00120112, -0.00001357, 0.00184421, 0.00183695, -0.00051708, -0.00068031, 0.00029363, -0.00022071, 0.00002159, -0.00026671, -0.00001766, -0.00014964, 0.00007467, -0.00010731, 0.00006147, -0.00003192, 0.00000368, 0.00000814, -0.00007964, },
		{ -0.00000048, -0.00000905, 0.00003791, -0.00001278, -0.00004474, 0.00002022, -0.00001730, 0.00002837, -0.00005923, -0.00018752, 0.00019307, -0.00054255, 0.00006494, -0.00127431, 0.00044970, 0.00143098, 0.00162181, -0.00027395, -0.00062162, 0.00029066, -0.00020417, -0.00006805, -0.00020310, -0.00007865, -0.00011468, 0.00007569, -0.00011869, 0.00007373, -0.00004804, 0.00002089, -0.00000493, -0.00007885, },
		{ 0.00000902, -0.00002073, 0.00004949, -0.00003686, -0.00002580, 0.00001967, -0.00003718, 0.00007135, -0.00012451, -0.00011570, 0.00008680, -0.00042988, -0.00006219, -0.00115727, 0.00069741, 0.00107155, 0.00148154, -0.00008832, -0.00052113, 0.00027747, -0.00021591, -0.00014993, -0.00012988, -0.00014926, -0.00005881, 0.00004470, -0.00009858, 0.00006339, -0.00004877, 0.00002979, -0.00001985, -0.00006637, },
		{ 0.00000737, -0.00002225, 0.00004864, -0.00005314, -0.00000345, 0.00000775, -0.00003927, 0.00008662, -0.00016144, -0.00005715, -0.00002562, -0.00029269, -0.00023894, -0.00090121, 0.00073790, 0.00080328, 0.00137621, 0.00006631, -0.00039371, 0.00026014, -0.00027290, -0.00019255, -0.00008006, -0.00019823, -0.00001071, 0.00000429, -0.00006115, 0.00003729, -0.00003563, 0.00002737, -0.00003095, -0.00004685, },
		{ -0.00000064, -0.00001662, 0.00003579, -0.00005753, 0.00001563, -0.00001006, -0.00002601, 0.00007167, -0.00015771, -0.00003696, -0.00010684, -0.00017910, -0.00040211, -0.00058088, 0.00061961, 0.00062387, 0.00129219, 0.00020036, -0.00023936, 0.00022331, -0.00036565, -0.00018759, -0.00006967, -0.00020628, 0.00000742, -0.00002381, -0.00002342, 0.00000637, -0.00001423, 0.00001378, -0.00003345, -0.00002682, },
		{ -0.00000885, -0.00000944, 0.00001498, -0.00004917, 0.00002536, -0.00002515, -0.00000699, 0.00003610, -0.00011799, -0.00006087, -0.00013864, -0.00011930, -0.00050242, -0.00027062, 0.00040966, 0.00050522, 0.00123441, 0.00031463, -0.00005111, 0.00013782, -0.00046170, -0.00015144, -0.00009292, -0.00017274, -0.00001227, -0.00002634, 0.00000078, -0.00001833, 0.00000762, -0.00000753, -0.00002518, -0.00001232, },
		},
		{ { 0.00000072, -0.00001901, 0.00006348, -0.00004699, -0.00009952, -0.00003629, 0.00000003, -0.00010071, -0.00008991, -0.00005750, -0.00056846, 0.00009375, 0.00063900, 0.00027684, 0.00054810, -0.00002376, -0.00081354, -0.00002376, 0.00054810, 0.00027684, 0.00063900, 0.00009375, -0.00056846, -0.00005750, -0.00008991, -0.00010071, 0.00000003, -0.00003629, -0.00009952, -0.00004699, 0.00006348, -0.00001901, },
		{ -0.00001982, -0.00000778, 0.00000906, 0.00002734, -0.00012389, -0.00004745, -0.00000538, -0.00005548, -0.00014443, -0.00001278, -0.00049510, -0.00027172, 0.00075994, 0.00038586, 0.00066894, 0.00005793, -0.00082643, -0.00008218, 0.00043077, 0.00022331, 0.00047193, 0.00031667, -0.00046339, -0.00018734, -0.00000758, -0.00014158, -0.00000139, -0.00003527, -0.00005948, -0.00010616, 0.00006543, 0.00002269, },
		{ -0.00005279, 0.00001372, -0.00003549, 0.00005264, -0.00008733, -0.00010035, 0.00002034, -0.00005950, -0.00011300, -0.00009163, -0.00027522, -0.00064589, 0.00070076, 0.00059286, 0.00078315, 0.00018287, -0.00087618, -0.00011480, 0.00031092, 0.00020291, 0.00032006, 0.00039720, -0.00027969, -0.00028023, 0.00000581, -0.00011355, -0.00004908, -0.00001011, -0.00004745, -0.00010637, -0.00000350, 0.00010419, },
		{ -0.00004087, -0.00000585, -0.00002436, 0.00000675, 0.00000208, -0.00017456, 0.00006096, -0.00009563, -0.00003514, -0.00020909, -0.00006681, -0.00082488, 0.00034649, 0.00088827, 0.00090624, 0.00036589, -0.00097365, -0.00011417, 0.00017490, 0.00021543, 0.00018291, 0.00041061, -0.00014354, -0.00024790, -0.00008289, -0.00001949, -0.00012584, 0.00002787, -0.00006017, -0.00006237, -0.00009871, 0.00016278, },
		{ 0.00004264, -0.00007047, 0.00002622, -0.00006432, 0.00007738, -0.00020429, 0.00005774, -0.00009640, -0.00000585, -0.00023615, -0.00002739, -0.00069200, -0.00026226, 0.00113094, 0.00111948, 0.00058431, -0.00109526, -0.00010287, 0.00003605, 0.00024383, 0.00006194, 0.00040738, -0.00009779, -0.00011849, -0.00019926, 0.00005682, -0.00016076, 0.00002908, -0.00005709, -0.00002680, -0.00015215, 0.00014107, },
		{ 0.00015905, -0.00012528, 0.00005157, -0.00008535, 0.00006944, -0.00014190, -0.00002289, -0.00002368, -0.00007390, -0.00012686, -0.00017209, -0.00033799, -0.00087855, 0.00106874, 0.00152804, 0.00080467, -0.00119493, -0.00012927, -0.00006414, 0.00024386, -0.00001075, 0.00037482, -0.00008740, -0.00000731, -0.00022332, 0.00002449, -0.00009859, -0.00003580, -0.00001337, -0.00003354, -0.00013333, 0.00004084, },
		{ 0.00022841, -0.00010302, 0.00000126, -0.00002379, -0.00002333, -0.00001786, -0.00013580, 0.00007581, -0.00018222, 0.00002773, -0.00034787, -0.00002300, -0.00116993, 0.00050077, 0.00212534, 0.00107182, -0.00126751, -0.00021393, -0.00010388, 0.00018603, -0.00000907, 0.00028498, -0.00004131, -0.00000526, -0.00010381, -0.00011888, 0.00003036, -0.00012703, 0.00004130, -0.00006241, -0.00007501, -0.00007364, },
		{ 0.00019416, 0.00001147, -0.00010574, 0.00007217, -0.00012116, 0.00006858, -0.00018112, 0.00010356, -0.00021573, 0.00007718, -0.00036040, 0.00001511, -0.00095452, -0.00051354, 0.00266625, 0.00157117, -0.00139098, -0.00030862, -0.00012321, 0.00009930, 0.00003855, 0.00016008, 0.00004094, -0.00009310, 0.00008391, -0.00026479, 0.00011574, -0.00015498, 0.00004379, -0.00006153, -0.00003982, -0.00012748, },
		{ 0.00007265, 0.00015727, -0.00018691, 0.00010975, -0.00012697, 0.00003233, -0.00009600, 0.00000916, -0.00011869, -0.00004253, -0.00015304, -0.00023855, -0.00037488, -0.00157703, 0.00269788, 0.00255331, -0.00166232, -0.00034039, -0.00019288, 0.00004903, 0.00006224, 0.00007060, 0.00008990, -0.00016563, 0.00019781, -0.00027871, 0.00005809, -0.00006198, -0.00003512, -0.00000683, -0.00005912, -0.00009717, },
		{ -0.00005652, 0.00023901, -0.00016313, 0.00003565, -0.00001798, -0.00011014, 0.00006752, -0.00013675, 0.00002914, -0.00022864, 0.00011974, -0.00053781, 0.00015077, -0.00212484, 0.00180346, 0.00410961, -0.00203340, -0.00033305, -0.00031465, 0.00004973, 0.00003066, 0.00005798, 0.00005445, -0.00014490, 0.00015980, -0.00012537, -0.00012905, 0.00010507, -0.00014189, 0.00006090, -0.00010187, -0.00002825, },
		{ -0.00011260, 0.00020139, -0.00003096, -0.00010549, 0.00011851, -0.00023320, 0.00016046, -0.00017741, 0.00006057, -0.00028099, 0.00020166, -0.00055190, 0.00019981, -0.00177287, -0.00006455, 0.00594120, -0.00217783, -0.00049387, -0.00034980, -0.00000148, 0.00001747, 0.00006355, -0.00001552, -0.00007283, 0.00003845, 0.00007933, -0.00030332, 0.00020726, -0.00016614, 0.00006349, -0.00010412, 0.00000700, },
		{ -0.00007592, 0.00007409, 0.00012436, -0.00019512, 0.00014628, -0.00019590, 0.00005291, 0.00000029, -0.00013356, -0.00007476, -0.00005129, -0.00013121, -0.00034434, -0.00061264, -0.00240568, 0.00733359, -0.00155434, -0.00113727, -0.00008235, -0.00027538, 0.00016402, -0.00004430, 0.00001100, -0.00008641, -0.00000038, 0.00014897, -0.00029480, 0.00011771, -0.00003221, -0.00004043, -0.00003356, -0.00002609, },
		{ 0.00000195, -0.00005184, 0.00019116, -0.00013490, -0.00000388, 0.00003244, -0.00024205, 0.00034687, -0.00048409, 0.00030871, -0.00052749, 0.00054701, -0.00117791, 0.00077805, -0.00432842, 0.00744463, 0.00028820, -0.00243340, 0.00058258, -0.00084741, 0.00053902, -0.00034314, 0.00022294, -0.00028545, 0.00015691, -0.00001586, -0.00005141, -0.00016151, 0.00021859, -0.00020099, 0.00007125, -0.00009529, },
		{ 0.00004693, -0.00009685, 0.00012007, 0.00006487, -0.00025996, 0.00032246, -0.00054741, 0.00065342, -0.00076227, 0.00061372, -0.00091442, 0.00107852, -0.00176018, 0.00165230, -0.00500139, 0.00578577, 0.00332231, -0.00413984, 0.00142803, -0.00154698, 0.00100911, -0.00073627, 0.00054793, -0.00060915, 0.00044914, -0.00032430, 0.00029378, -0.00047219, 0.00043300, -0.00029858, 0.00012677, -0.00013308, },
		{ 0.00002618, -0.00005367, -0.00003469, 0.00027948, -0.00044540, 0.00046422, -0.00063788, 0.00069260, -0.00073959, 0.00059336, -0.00092007, 0.00111177, -0.00167392, 0.00154319, -0.00413042, 0.00259622, 0.00685060, -0.00554996, 0.00193078, -0.00197707, 0.00125820, -0.00096624, 0.00076000, -0.00084306, 0.00065977, -0.00054255, 0.00050123, -0.00059948, 0.00044855, -0.00023368, 0.00007687, -0.00010006, },
		{ -0.00002862, 0.00001398, -0.00015690, 0.00035746, -0.00040677, 0.00032353, -0.00041119, 0.00039441, -0.00036766, 0.00020244, -0.00049112, 0.00059825, -0.00090520, 0.00054233, -0.00214528, -0.00116676, 0.00970957, -0.00577837, 0.00153315, -0.00174707, 0.00098593, -0.00078690, 0.00063545, -0.00077095, 0.00057678, -0.00046633, 0.00039937, -0.00043012, 0.00021968, -0.00001430, -0.00005112, -0.00002241, },
		{ -0.00005683, 0.00003134, -0.00015959, 0.00023496, -0.00013468, -0.00004631, 0.00002068, -0.00008530, 0.00016904, -0.00035587, 0.00014332, -0.00017458, 0.00015626, -0.00077992, 0.00003875, -0.00426228, 0.01081045, -0.00426228, 0.00003875, -0.00077992, 0.00015626, -0.00017458, 0.00014332, -0.00035587, 0.00016904, -0.00008530, 0.00002068, -0.00004631, -0.00013468, 0.00023496, -0.00015959, 0.00003134, },
		{ -0.00002862, -0.00002241, -0.00005112, -0.00001430, 0.00021968, -0.00043012, 0.00039937, -0.00046633, 0.00057678, -0.00077095, 0.00063545, -0.00078690, 0.00098593, -0.00174707, 0.00153315, -0.00577837, 0.00970957, -0.00116676, -0.00214528, 0.00054233, -0.00090520, 0.00059825, -0.00049112, 0.00020244, -0.00036766, 0.00039441, -0.00041119, 0.00032353, -0.00040677, 0.00035746, -0.00015690, 0.00001398, },
		{ 0.00002618, -0.00010006, 0.00007687, -0.00023368, 0.00044855, -0.00059948, 0.00050123, -0.00054255, 0.00065977, -0.00084306, 0.00076000, -0.00096624, 0.00125820, -0.00197707, 0.00193078, -0.00554996, 0.00685060, 0.00259622, -0.00413042, 0.00154319, -0.00167392, 0.00111177, -0.00092007, 0.00059336, -0.00073959, 0.00069260, -0.00063788, 0.00046422, -0.00044540, 0.00027948, -0.00003469, -0.00005367, },
		{ 0.00004693, -0.00013308, 0.00012677, -0.00029858, 0.00043300, -0.00047219, 0.00029378, -0.00032430, 0.00044914, -0.00060915, 0.00054793, -0.00073627, 0.00100911, -0.00154698, 0.00142803, -0.00413984, 0.00332231, 0.00578577, -0.00500139, 0.00165230, -0.00176018, 0.00107852, -0.00091442, 0.00061372, -0.00076227, 0.00065342, -0.00054741, 0.00032246, -0.00025996, 0.00006487, 0.00012007, -0.00009685, },
		{ 0.00000195, -0.00009529, 0.00007125, -0.00020099, 0.00021859, -0.00016151, -0.00005141, -0.00001586, 0.00015691, -0.00028545, 0.00022294, -0.00034314, 0.00053902, -0.00084741, 0.00058258, -0.00243340, 0.00028820, 0.00744463, -0.00432842, 0.00077805, -0.00117791, 0.00054701, -0.00052749, 0.00030871, -0.00048409, 0.00034687, -0.00024205, 0.00003244, -0.00000388, -0.00013490, 0.00019116, -0.00005184, },
		{ -0.00007592, -0.00002609, -0.00003356, -0.00004043, -0.00003221, 0.00011771, -0.00029480, 0.00014897, -0.00000038, -0.00008641, 0.00001100, -0.00004430, 0.00016402, -0.00027538, -0.00008235, -0.00113727, -0.00155434, 0.00733359, -0.00240568, -0.00061264, -0.00034434, -0.00013121, -0.00005129, -0.00007476, -0.00013356, 0.00000029, 0.00005291, -0.00019590, 0.00014628, -0.00019512, 0.00012436, 0.00007409, },
		{ -0.00011260, 0.00000700, -0.00010412, 0.00006349, -0.00016614, 0.00020726, -0.00030332, 0.00007933, 0.00003845, -0.00007283, -0.00001552, 0.00006355, 0.00001747, -0.00000148, -0.00034980, -0.00049387, -0.00217783, 0.00594120, -0.00006455, -0.00177287, 0.00019981, -0.00055190, 0.00020166, -0.00028099, 0.00006057, -0.00017741, 0.00016046, -0.00023320, 0.00011851, -0.00010549, -0.00003096, 0.00020139, },
		{ -0.00005652, -0.00002825, -0.00010187, 0.00006090, -0.00014189, 0.00010507, -0.00012905, -0.00012537, 0.00015980, -0.00014490, 0.00005445, 0.00005798, 0.00003066, 0.00004973, -0.00031465, -0.00033305, -0.00203340, 0.00410961, 0.00180346, -0.00212484, 0.00015077, -0.00053781, 0.00011974, -0.00022864, 0.00002914, -0.00013675, 0.00006752, -0.00011014, -0.00001798, 0.00003565, -0.00016313, 0.00023901, },
		{ 0.00007265, -0.00009717, -0.00005912, -0.00000683, -0.00003512, -0.00006198, 0.00005809, -0.00027871, 0.00019781, -0.00016563, 0.00008990, 0.00007060, 0.00006224, 0.00004903, -0.00019288, -0.00034039, -0.00166232, 0.00255331, 0.00269788, -0.00157703, -0.00037488, -0.00023855, -0.00015304, -0.00004253, -0.00011869, 0.00000916, -0.00009600, 0.00003233, -0.00012697, 0.00010975, -0.00018691, 0.00015727, },
		{ 0.00019416, -0.00012748, -0.00003982, -0.00006153, 0.00004379, -0.00015498, 0.00011574, -0.00026479, 0.00008391, -0.00009310, 0.00004094, 0.00016008, 0.00003855, 0.00009930, -0.00012321, -0.00030862, -0.00139098, 0.00157117, 0.00266625, -0.00051354, -0.00095452, 0.00001511, -0.00036040, 0.00007718, -0.00021573, 0.00010356, -0.00018112, 0.00006858, -0.00012116, 0.00007217, -0.00010574, 0.00001147, },
		{ 0.00022841, -0.00007364, -0.00007501, -0.00006241, 0.00004130, -0.00012703, 0.00003036, -0.00011888, -0.00010381, -0.00000526, -0.00004131, 0.00028498, -0.00000907, 0.00018603, -0.00010388, -0.00021393, -0.00126751, 0.00107182, 0.00212534, 0.00050077, -0.00116993, -0.00002300, -0.00034787, 0.00002773, -0.00018222, 0.00007581, -0.00013580, -0.00001786, -0.00002333, -0.00002379, 0.00000126, -0.00010302, },
		{ 0.00015905, 0.00004084, -0.00013333, -0.00003354, -0.00001337, -0.00003580, -0.00009859, 0.00002449, -0.00022332, -0.00000731, -0.00008740, 0.00037482, -0.00001075, 0.00024386, -0.00006414, -0.00012927, -0.00119493, 0.00080467, 0.00152804, 0.00106874, -0.00087855, -0.00033799, -0.00017209, -0.00012686, -0.00007390, -0.00002368, -0.00002289, -0.00014190, 0.00006944, -0.00008535, 0.00005157, -0.00012528, },
		{ 0.00004264, 0.00014107, -0.00015215, -0.00002680, -0.00005709, 0.00002908, -0.00016076, 0.00005682, -0.00019926, -0.00011849, -0.00009779, 0.00040738, 0.00006194, 0.00024383, 0.00003605, -0.00010287, -0.00109526, 0.00058431, 0.00111948, 0.00113094, -0.00026226, -0.00069200, -0.00002739, -0.00023615, -0.00000585, -0.00009640, 0.00005774, -0.00020429, 0.00007738, -0.00006432, 0.00002622, -0.00007047, },
		{ -0.00004087, 0.00016278, -0.00009871, -0.00006237, -0.00006017, 0.00002787, -0.00012584, -0.00001949, -0.00008289, -0.00024790, -0.00014354, 0.00041061, 0.00018291, 0.00021543, 0.00017490, -0.00011417, -0.00097365, 0.00036589, 0.00090624, 0.00088827, 0.00034649, -0.00082488, -0.00006681, -0.00020909, -0.00003514, -0.00009563, 0.00006096, -0.00017456, 0.00000208, 0.00000675, -0.00002436, -0.00000585, },
		{ -0.00005279, 0.00010419, -0.00000350, -0.00010637, -0.00004745, -0.00001011, -0.00004908, -0.00011355, 0.00000581, -0.00028023, -0.00027969, 0.00039720, 0.00032006, 0.00020291, 0.00031092, -0.00011480, -0.00087618, 0.00018287, 0.00078315, 0.00059286, 0.00070076, -0.00064589, -0.00027522, -0.00009163, -0.00011300, -0.00005950, 0.00002034, -0.00010035, -0.00008733, 0.00005264, -0.00003549, 0.00001372, },
		{ -0.00001982, 0.00002269, 0.00006543, -0.00010616, -0.00005948, -0.00003527, -0.00000139, -0.00014158, -0.00000758, -0.00018734, -0.00046339, 0.00031667, 0.00047193, 0.00022331, 0.00043077, -0.00008218, -0.00082643, 0.00005793, 0.00066894, 0.00038586, 0.00075994, -0.00027172, -0.00049510, -0.00001278, -0.00014443, -0.00005548, -0.00000538, -0.00004745, -0.00012389, 0.00002734, 0.00000906, -0.00000778, },
		},
		{ { -0.00004963, -0.00002939, -0.00018782, -0.00001894, -0.00001216, -0.00012315, -0.00051266, 0.00009629, 0.00060062, 0.00030862, 0.00067665, 0.00006859, -0.00048416, -0.00012748, 0.00006698, -0.00008822, -0.00007265, -0.00008822, 0.00006698, -0.00012748, -0.00048416, 0.00006859, 0.00067665, 0.00030862, 0.00060062, 0.00009629, -0.00051266, -0.00012315, -0.00001216, -0.00001894, -0.00018782, -0.00002939, },
		{ -0.00001050, -0.00005296, -0.00010856, -0.00012956, 0.00004671, -0.00008739, -0.00029473, -0.00051647, 0.00064018, 0.00046081, 0.00064441, 0.00049558, -0.00054310, -0.00026828, 0.00001468, -0.00000023, -0.00009826, -0.00008128, -0.00002328, 0.00004570, -0.00040836, -0.00014621, 0.00046736, 0.00040269, 0.00031899, 0.00045839, -0.00027864, -0.00032989, -0.00002258, 0.00000438, -0.00015137, -0.00010294, },
		{ 0.00001075, -0.00007478, -0.00004060, -0.00018145, 0.00003338, -0.00006424, -0.00010058, -0.00061044, -0.00004956, 0.00090868, 0.00045835, 0.00096082, -0.00042605, -0.00044107, -0.00007733, 0.00003855, -0.00007152, -0.00007550, -0.00009401, 0.00008577, -0.00023103, -0.00028578, 0.00026636, 0.00040687, 0.00023331, 0.00038287, 0.00013869, -0.00037466, -0.00014105, 0.00002473, -0.00009540, -0.00016881, },
		{ -0.00003448, -0.00002602, -0.00007948, -0.00008110, -0.00012654, 0.00008230, -0.00018181, -0.00019105, -0.00081993, 0.00080417, 0.00070436, 0.00102513, 0.00014451, -0.00077883, -0.00006104, -0.00008968, 0.00005344, -0.00012148, -0.00007181, -0.00000824, -0.00003562, -0.00032138, 0.00008072, 0.00035809, 0.00021853, 0.00025183, 0.00028803, -0.00011718, -0.00029534, 0.00000094, -0.00004098, -0.00018480, },
		{ -0.00006689, -0.00000094, -0.00010038, -0.00001700, -0.00018389, 0.00008147, -0.00015262, -0.00003415, -0.00071169, -0.00023647, 0.00118494, 0.00094843, 0.00085443, -0.00094179, -0.00017698, -0.00015406, 0.00005511, -0.00007005, -0.00008294, -0.00006091, 0.00001683, -0.00021173, -0.00008909, 0.00027476, 0.00024005, 0.00012742, 0.00030561, 0.00005675, -0.00017688, -0.00018569, 0.00006601, -0.00021243, },
		{ -0.00015156, 0.00002757, -0.00009016, -0.00004057, -0.00008798, -0.00008870, 0.00004528, -0.00021309, -0.00012861, -0.00104307, 0.00074021, 0.00135299, 0.00125279, -0.00057237, -0.00061395, -0.00001627, -0.00013707, 0.00007251, -0.00012746, -0.00005599, -0.00001452, -0.00011477, -0.00010606, 0.00007105, 0.00035224, -0.00002766, 0.00031507, 0.00007891, 0.00003006, -0.00025713, 0.00002851, -0.00013490, },
		{ -0.00029958, 0.00008808, -0.00010745, -0.00004014, -0.00004123, -0.00013463, 0.00005481, -0.00018956, -0.00001436, -0.00071266, -0.00058614, 0.00154875, 0.00181230, -0.00000515, -0.00099141, -0.00000394, -0.00021972, 0.00005325, -0.00005382, -0.00010086, -0.00000441, -0.00011791, -0.00002396, -0.00008638, 0.00032111, 0.00004962, 0.00008083, 0.00026580, -0.00004774, -0.00000028, -0.00020603, 0.00005806, },
		{ -0.00025847, -0.00001568, -0.00000958, -0.00010652, 0.00000629, -0.00011298, -0.00002641, -0.00006208, -0.00015598, -0.00015550, -0.00110613, 0.00029128, 0.00277693, 0.00071567, -0.00114590, -0.00022542, -0.00013882, -0.00009789, 0.00003778, -0.00010716, -0.00002827, -0.00008340, -0.00006277, 0.00000704, 0.00002371, 0.00035637, -0.00025604, 0.00042756, -0.00011881, 0.00014599, -0.00021570, 0.00004615, },
		{ -0.00004388, -0.00026295, 0.00018503, -0.00024407, 0.00011223, -0.00017299, 0.00004149, -0.00014650, -0.00003237, -0.00021866, -0.00038939, -0.00141639, 0.00248584, 0.00252246, -0.00143094, -0.00037626, -0.00019349, -0.00009075, -0.00007654, 0.00003981, -0.00014819, 0.00003854, -0.00020676, 0.00017870, -0.00020500, 0.00043061, -0.00019384, 0.00013972, 0.00018278, -0.00008835, 0.00008566, -0.00016031, },
		{ -0.00002335, -0.00030378, 0.00014270, -0.00014766, 0.00000304, -0.00005550, -0.00003906, -0.00007165, -0.00008143, -0.00017999, -0.00008091, -0.00125053, -0.00026124, 0.00488212, -0.00096240, -0.00094141, -0.00009591, -0.00016649, -0.00011953, 0.00003988, -0.00010505, 0.00001119, -0.00017965, 0.00010299, -0.00007219, 0.00010532, 0.00021473, -0.00030178, 0.00042516, -0.00018035, 0.00014448, -0.00010649, },
		{ -0.00011406, -0.00016780, -0.00011556, 0.00015322, -0.00029642, 0.00024975, -0.00034309, 0.00029168, -0.00048147, 0.00030928, -0.00066131, 0.00029575, -0.00286648, 0.00446076, 0.00222596, -0.00262371, 0.00051615, -0.00064135, 0.00020696, -0.00032366, 0.00021249, -0.00023126, 0.00005800, -0.00017772, 0.00021788, -0.00022779, 0.00042439, -0.00032855, 0.00018087, 0.00012377, -0.00011030, 0.00012885, },
		{ 0.00006505, -0.00018736, -0.00020100, 0.00020605, -0.00028758, 0.00020474, -0.00026802, 0.00021947, -0.00035193, 0.00017381, -0.00052346, 0.00040523, -0.00194825, -0.00017073, 0.00683103, -0.00337135, 0.00017519, -0.00053314, 0.00014189, -0.00036775, 0.00019807, -0.00019412, 0.00007619, -0.00023719, 0.00023624, -0.00020432, 0.00024053, -0.00003107, -0.00015984, 0.00026975, -0.00008184, 0.00002098, },
		{ 0.00021794, -0.00016974, -0.00017368, 0.00001182, 0.00004470, -0.00021762, 0.00024605, -0.00039064, 0.00042315, -0.00071628, 0.00057308, -0.00103232, 0.00093916, -0.00447506, 0.00727857, 0.00048356, -0.00275251, 0.00098977, -0.00092252, 0.00040683, -0.00050647, 0.00035407, -0.00030730, 0.00007217, -0.00008099, 0.00009110, -0.00007370, 0.00018949, -0.00021104, 0.00006233, 0.00019343, -0.00020212, },
		{ -0.00009481, 0.00016910, -0.00030947, -0.00003414, 0.00015908, -0.00033615, 0.00037538, -0.00052835, 0.00058016, -0.00080249, 0.00067409, -0.00118553, 0.00134129, -0.00340185, 0.00121161, 0.00747530, -0.00540469, 0.00160088, -0.00131849, 0.00070106, -0.00080138, 0.00050571, -0.00039022, 0.00018622, -0.00024453, 0.00023336, -0.00020600, 0.00022852, -0.00017266, -0.00001114, 0.00009959, 0.00004585, },
		{ -0.00033081, 0.00032451, -0.00030258, -0.00002755, 0.00001674, -0.00002900, -0.00007086, 0.00007859, -0.00022103, 0.00027615, -0.00061285, 0.00044438, -0.00090659, 0.00089364, -0.00526560, 0.00988037, -0.00216704, -0.00168323, 0.00066148, -0.00070753, 0.00026052, -0.00042905, 0.00032090, -0.00029494, 0.00008045, -0.00002135, -0.00003994, 0.00010757, -0.00014482, 0.00007757, -0.00017695, 0.00035411, },
		{ 0.00007652, -0.00008580, 0.00009262, -0.00023532, 0.00002833, 0.00010821, -0.00029915, 0.00041418, -0.00066668, 0.00082360, -0.00115318, 0.00110781, -0.00179311, 0.00225521, -0.00510774, 0.00372041, 0.00625034, -0.00603590, 0.00244419, -0.00187457, 0.00113039, -0.00115090, 0.00078407, -0.00059752, 0.00032386, -0.00025369, 0.00014710, -0.00007244, 0.00001289, -0.00009221, -0.00000408, 0.00004783, },
		{ 0.00043950, -0.00034759, 0.00030228, -0.00033859, 0.00012472, -0.00004855, -0.00002834, 0.00004280, -0.00014746, 0.00008260, -0.00009249, -0.00018681, -0.00010314, -0.00019190, -0.00013829, -0.00444990, 0.01094712, -0.00444990, -0.00013829, -0.00019190, -0.00010314, -0.00018681, -0.00009249, 0.00008260, -0.00014746, 0.00004280, -0.00002834, -0.00004855, 0.00012472, -0.00033859, 0.00030228, -0.00034759, },
		{ 0.00007652, 0.00004783, -0.00000408, -0.00009221, 0.00001289, -0.00007244, 0.00014710, -0.00025369, 0.00032386, -0.00059752, 0.00078407, -0.00115090, 0.00113039, -0.00187457, 0.00244419, -0.00603590, 0.00625034, 0.00372041, -0.00510774, 0.00225521, -0.00179311, 0.00110781, -0.00115318, 0.00082360, -0.00066668, 0.00041418, -0.00029915, 0.00010821, 0.00002833, -0.00023532, 0.00009262, -0.00008580, },
		{ -0.00033081, 0.00035411, -0.00017695, 0.00007757, -0.00014482, 0.00010757, -0.00003994, -0.00002135, 0.00008045, -0.00029494, 0.00032090, -0.00042905, 0.00026052, -0.00070753, 0.00066148, -0.00168323, -0.00216704, 0.00988037, -0.00526560, 0.00089364, -0.00090659, 0.00044438, -0.00061285, 0.00027615, -0.00022103, 0.00007859, -0.00007086, -0.00002900, 0.00001674, -0.00002755, -0.00030258, 0.00032451, },
		{ -0.00009481, 0.00004585, 0.00009959, -0.00001114, -0.00017266, 0.00022852, -0.00020600, 0.00023336, -0.00024453, 0.00018622, -0.00039022, 0.00050571, -0.00080138, 0.00070106, -0.00131849, 0.00160088, -0.00540469, 0.00747530, 0.00121161, -0.00340185, 0.00134129, -0.00118553, 0.00067409, -0.00080249, 0.00058016, -0.00052835, 0.00037538, -0.00033615, 0.00015908, -0.00003414, -0.00030947, 0.00016910, },
		{ 0.00021794, -0.00020212, 0.00019343, 0.00006233, -0.00021104, 0.00018949, -0.00007370, 0.00009110, -0.00008099, 0.00007217, -0.00030730, 0.00035407, -0.00050647, 0.00040683, -0.00092252, 0.00098977, -0.00275251, 0.00048356, 0.00727857, -0.00447506, 0.00093916, -0.00103232, 0.00057308, -0.00071628, 0.00042315, -0.00039064, 0.00024605, -0.00021762, 0.00004470, 0.00001182, -0.00017368, -0.00016974, },
		{ 0.00006505, 0.00002098, -0.00008184, 0.00026975, -0.00015984, -0.00003107, 0.00024053, -0.00020432, 0.00023624, -0.00023719, 0.00007619, -0.00019412, 0.00019807, -0.00036775, 0.00014189, -0.00053314, 0.00017519, -0.00337135, 0.00683103, -0.00017073, -0.00194825, 0.00040523, -0.00052346, 0.00017381, -0.00035193, 0.00021947, -0.00026802, 0.00020474, -0.00028758, 0.00020605, -0.00020100, -0.00018736, },
		{ -0.00011406, 0.00012885, -0.00011030, 0.00012377, 0.00018087, -0.00032855, 0.00042439, -0.00022779, 0.00021788, -0.00017772, 0.00005800, -0.00023126, 0.00021249, -0.00032366, 0.00020696, -0.00064135, 0.00051615, -0.00262371, 0.00222596, 0.00446076, -0.00286648, 0.00029575, -0.00066131, 0.00030928, -0.00048147, 0.00029168, -0.00034309, 0.00024975, -0.00029642, 0.00015322, -0.00011556, -0.00016780, },
		{ -0.00002335, -0.00010649, 0.00014448, -0.00018035, 0.00042516, -0.00030178, 0.00021473, 0.00010532, -0.00007219, 0.00010299, -0.00017965, 0.00001119, -0.00010505, 0.00003988, -0.00011953, -0.00016649, -0.00009591, -0.00094141, -0.00096240, 0.00488212, -0.00026124, -0.00125053, -0.00008091, -0.00017999, -0.00008143, -0.00007165, -0.00003906, -0.00005550, 0.00000304, -0.00014766, 0.00014270, -0.00030378, },
		{ -0.00004388, -0.00016031, 0.00008566, -0.00008835, 0.00018278, 0.00013972, -0.00019384, 0.00043061, -0.00020500, 0.00017870, -0.00020676, 0.00003854, -0.00014819, 0.00003981, -0.00007654, -0.00009075, -0.00019349, -0.00037626, -0.00143094, 0.00252246, 0.00248584, -0.00141639, -0.00038939, -0.00021866, -0.00003237, -0.00014650, 0.00004149, -0.00017299, 0.00011223, -0.00024407, 0.00018503, -0.00026295, },
		{ -0.00025847, 0.00004615, -0.00021570, 0.00014599, -0.00011881, 0.00042756, -0.00025604, 0.00035637, 0.00002371, 0.00000704, -0.00006277, -0.00008340, -0.00002827, -0.00010716, 0.00003778, -0.00009789, -0.00013882, -0.00022542, -0.00114590, 0.00071567, 0.00277693, 0.00029128, -0.00110613, -0.00015550, -0.00015598, -0.00006208, -0.00002641, -0.00011298, 0.00000629, -0.00010652, -0.00000958, -0.00001568, },
		{ -0.00029958, 0.00005806, -0.00020603, -0.00000028, -0.00004774, 0.00026580, 0.00008083, 0.00004962, 0.00032111, -0.00008638, -0.00002396, -0.00011791, -0.00000441, -0.00010086, -0.00005382, 0.00005325, -0.00021972, -0.00000394, -0.00099141, -0.00000515, 0.00181230, 0.00154875, -0.00058614, -0.00071266, -0.00001436, -0.00018956, 0.00005481, -0.00013463, -0.00004123, -0.00004014, -0.00010745, 0.00008808, },
		{ -0.00015156, -0.00013490, 0.00002851, -0.00025713, 0.00003006, 0.00007891, 0.00031507, -0.00002766, 0.00035224, 0.00007105, -0.00010606, -0.00011477, -0.00001452, -0.00005599, -0.00012746, 0.00007251, -0.00013707, -0.00001627, -0.00061395, -0.00057237, 0.00125279, 0.00135299, 0.00074021, -0.00104307, -0.00012861, -0.00021309, 0.00004528, -0.00008870, -0.00008798, -0.00004057, -0.00009016, 0.00002757, },
		{ -0.00006689, -0.00021243, 0.00006601, -0.00018569, -0.00017688, 0.00005675, 0.00030561, 0.00012742, 0.00024005, 0.00027476, -0.00008909, -0.00021173, 0.00001683, -0.00006091, -0.00008294, -0.00007005, 0.00005511, -0.00015406, -0.00017698, -0.00094179, 0.00085443, 0.00094843, 0.00118494, -0.00023647, -0.00071169, -0.00003415, -0.00015262, 0.00008147, -0.00018389, -0.00001700, -0.00010038, -0.00000094, },
		{ -0.00003448, -0.00018480, -0.00004098, 0.00000094, -0.00029534, -0.00011718, 0.00028803, 0.00025183, 0.00021853, 0.00035809, 0.00008072, -0.00032138, -0.00003562, -0.00000824, -0.00007181, -0.00012148, 0.00005344, -0.00008968, -0.00006104, -0.00077883, 0.00014451, 0.00102513, 0.00070436, 0.00080417, -0.00081993, -0.00019105, -0.00018181, 0.00008230, -0.00012654, -0.00008110, -0.00007948, -0.00002602, },
		{ 0.00001075, -0.00016881, -0.00009540, 0.00002473, -0.00014105, -0.00037466, 0.00013869, 0.00038287, 0.00023331, 0.00040687, 0.00026636, -0.00028578, -0.00023103, 0.00008577, -0.00009401, -0.00007550, -0.00007152, 0.00003855, -0.00007733, -0.00044107, -0.00042605, 0.00096082, 0.00045835, 0.00090868, -0.00004956, -0.00061044, -0.00010058, -0.00006424, 0.00003338, -0.00018145, -0.00004060, -0.00007478, },
		{ -0.00001050, -0.00010294, -0.00015137, 0.00000438, -0.00002258, -0.00032989, -0.00027864, 0.00045839, 0.00031899, 0.00040269, 0.00046736, -0.00014621, -0.00040836, 0.00004570, -0.00002328, -0.00008128, -0.00009826, -0.00000023, 0.00001468, -0.00026828, -0.00054310, 0.00049558, 0.00064441, 0.00046081, 0.00064018, -0.00051647, -0.00029473, -0.00008739, 0.00004671, -0.00012956, -0.00010856, -0.00005296, },
		},
		{ { -0.00013451, -0.00011967, -0.00043758, 0.00005347, 0.00060645, 0.00026726, 0.00078882, 0.00008568, -0.00040641, -0.00012844, -0.00017724, -0.00003741, -0.00006697, 0.00000398, -0.00013085, -0.00003855, -0.00004532, -0.00003855, -0.00013085, 0.00000398, -0.00006697, -0.00003741, -0.00017724, -0.00012844, -0.00040641, 0.00008568, 0.00078882, 0.00026726, 0.00060645, 0.00005347, -0.00043758, -0.00011967, },
		{ -0.00010296, -0.00013756, -0.00014663, -0.00054896, 0.00030892, 0.00063536, 0.00044884, 0.00085746, -0.00041259, -0.00025110, -0.00016031, -0.00014789, -0.00002836, -0.00004197, -0.00005615, -0.00011119, -0.00001445, -0.00004693, -0.00008742, -0.00008868, 0.00002402, -0.00009355, -0.00004912, -0.00016753, -0.00020124, -0.00031210, 0.00044182, 0.00053844, 0.00024299, 0.00049194, -0.00009258, -0.00034526, },
		{ -0.00028994, -0.00008385, -0.00014836, -0.00019743, -0.00064862, 0.00066549, 0.00058733, 0.00082152, 0.00048961, -0.00070530, -0.00001101, -0.00029025, -0.00004165, -0.00005489, -0.00002662, -0.00011782, -0.00004980, -0.00002335, -0.00005682, -0.00012687, -0.00000738, -0.00001218, -0.00010084, -0.00004187, -0.00020256, -0.00022367, -0.00009007, 0.00056568, 0.00029720, 0.00028762, 0.00032346, -0.00014153, },
		{ -0.00014050, -0.00025166, -0.00008356, -0.00013264, -0.00033376, -0.00063288, 0.00102167, 0.00061746, 0.00114559, -0.00028077, -0.00051540, 0.00000275, -0.00031653, 0.00002998, -0.00006954, -0.00005789, -0.00011717, -0.00001086, -0.00003633, -0.00008985, -0.00011599, 0.00006180, -0.00009489, -0.00004320, -0.00008740, -0.00022454, -0.00015801, 0.00011422, 0.00051690, 0.00015045, 0.00031679, 0.00016104, },
		{ 0.00009232, -0.00017458, -0.00017112, -0.00015153, -0.00003982, -0.00063134, -0.00036994, 0.00125218, 0.00092228, 0.00094287, -0.00092716, -0.00005825, -0.00022919, -0.00015917, -0.00000870, -0.00004204, -0.00013237, -0.00004031, -0.00003399, -0.00002780, -0.00014834, -0.00003158, 0.00005059, -0.00014804, 0.00004504, -0.00021748, -0.00012546, -0.00011509, 0.00029146, 0.00033867, 0.00014801, 0.00024513, },
		{ 0.00008875, 0.00011512, -0.00024799, -0.00006801, -0.00024172, 0.00005474, -0.00100826, 0.00009931, 0.00147089, 0.00131951, 0.00006210, -0.00090540, 0.00014342, -0.00037701, -0.00001065, -0.00005438, -0.00005305, -0.00015015, 0.00002086, -0.00006860, -0.00003009, -0.00018322, 0.00008395, -0.00005929, -0.00007897, 0.00002620, -0.00029256, 0.00002559, -0.00011580, 0.00042376, 0.00011564, 0.00024057, },
		{ 0.00027428, -0.00001802, 0.00011132, -0.00026010, -0.00005178, -0.00022118, -0.00002959, -0.00124815, 0.00060239, 0.00198018, 0.00114677, -0.00087410, -0.00036041, -0.00006317, -0.00024576, -0.00003568, -0.00000716, -0.00015909, -0.00004077, -0.00002594, -0.00003936, -0.00009632, -0.00012043, 0.00011953, -0.00016403, 0.00007015, -0.00014859, -0.00016500, 0.00004345, -0.00003945, 0.00040800, 0.00000328, },
		{ 0.00004584, 0.00018866, -0.00001958, 0.00003783, -0.00019522, -0.00011887, -0.00010568, -0.00029950, -0.00132530, 0.00126010, 0.00264460, 0.00000899, -0.00100208, -0.00002346, -0.00025357, -0.00007668, -0.00009742, 0.00002322, -0.00023608, 0.00008951, -0.00012516, 0.00002468, -0.00019331, 0.00003034, 0.00001060, -0.00011765, 0.00010786, -0.00029615, 0.00006275, -0.00007635, 0.00013311, 0.00023924, },
		{ 0.00009135, 0.00009857, 0.00010498, -0.00002937, 0.00000803, -0.00018609, -0.00010814, -0.00009872, -0.00054308, -0.00144522, 0.00257640, 0.00251984, -0.00138382, -0.00031498, -0.00025395, -0.00007261, -0.00015609, 0.00000436, -0.00011108, -0.00012710, 0.00007110, -0.00013515, 0.00001319, -0.00020181, 0.00013307, -0.00013864, 0.00006236, -0.00006371, -0.00020584, 0.00013186, -0.00012584, 0.00023140, },
		{ 0.00017131, 0.00006365, 0.00007575, 0.00008532, -0.00008985, 0.00004193, -0.00023563, -0.00003061, -0.00019664, -0.00080545, -0.00141443, 0.00458469, 0.00072475, -0.00166041, 0.00017237, -0.00051613, 0.00018743, -0.00031827, 0.00014427, -0.00028487, 0.00007772, -0.00006905, -0.00004921, -0.00008428, -0.00007989, 0.00007047, -0.00013656, 0.00015299, -0.00025381, 0.00003476, 0.00000412, -0.00002119, },
		{ 0.00003387, 0.00013279, 0.00000376, 0.00012768, -0.00003848, -0.00001091, -0.00008283, -0.00007570, -0.00022303, -0.00000106, -0.00163283, -0.00021280, 0.00595398, -0.00210424, -0.00039925, -0.00032361, -0.00014084, 0.00000319, -0.00017340, 0.00001726, -0.00021176, 0.00010993, -0.00011882, -0.00002596, -0.00012161, 0.00003473, -0.00007655, 0.00004412, -0.00000578, -0.00018619, 0.00010678, -0.00005716, },
		{ 0.00010727, -0.00011811, 0.00026554, -0.00021019, 0.00035566, -0.00036303, 0.00029955, -0.00045145, 0.00037768, -0.00079788, 0.00061417, -0.00323405, 0.00303322, 0.00472132, -0.00379614, 0.00107043, -0.00122734, 0.00061801, -0.00052024, 0.00022210, -0.00032564, 0.00009581, -0.00007642, -0.00000897, -0.00013075, -0.00000764, -0.00001419, -0.00009323, 0.00017786, -0.00024881, 0.00010100, -0.00009027, },
		{ -0.00024337, 0.00025222, -0.00025885, 0.00035832, -0.00033906, 0.00047692, -0.00056570, 0.00047065, -0.00061617, 0.00056198, -0.00106490, 0.00073503, -0.00434076, 0.00729971, 0.00053067, -0.00272112, 0.00101712, -0.00123453, 0.00079399, -0.00066902, 0.00031995, -0.00041550, 0.00023853, -0.00017781, 0.00003727, -0.00017648, 0.00010558, -0.00015797, 0.00008360, 0.00004963, -0.00024052, 0.00023587, },
		{ -0.00001236, -0.00003825, 0.00005005, -0.00006117, 0.00009067, -0.00004567, 0.00011340, -0.00022673, 0.00004074, -0.00004220, -0.00020301, -0.00013565, -0.00093857, -0.00271878, 0.00953126, -0.00417160, 0.00022001, -0.00061973, -0.00002742, 0.00001904, -0.00011318, -0.00014477, -0.00000590, -0.00003528, 0.00003408, -0.00017018, 0.00002572, -0.00000641, -0.00014897, 0.00023683, -0.00023310, 0.00008239, },
		{ 0.00037022, -0.00038137, 0.00035813, -0.00038008, 0.00036839, -0.00040744, 0.00051010, -0.00055888, 0.00045547, -0.00073756, 0.00094353, -0.00154896, 0.00159927, -0.00415495, 0.00237184, 0.00720198, -0.00597385, 0.00227747, -0.00205572, 0.00119288, -0.00087041, 0.00052956, -0.00066057, 0.00047905, -0.00040467, 0.00031871, -0.00043124, 0.00032359, -0.00028976, 0.00010140, 0.00012432, -0.00032521, },
		{ -0.00007103, 0.00026763, -0.00041090, 0.00045899, -0.00054246, 0.00051934, -0.00057543, 0.00068893, -0.00082143, 0.00067538, -0.00091481, 0.00115244, -0.00197393, 0.00217339, -0.00630428, 0.00839537, 0.00118830, -0.00380059, 0.00170168, -0.00172908, 0.00113902, -0.00084494, 0.00044463, -0.00053677, 0.00042727, -0.00034680, 0.00024008, -0.00035376, 0.00033848, -0.00040110, 0.00034499, -0.00018335, },
		{ -0.00035030, 0.00027476, -0.00013387, -0.00003188, 0.00009319, -0.00021168, 0.00014656, -0.00014326, 0.00016904, -0.00029790, 0.00001744, -0.00000921, -0.00007161, -0.00051307, 0.00001304, -0.00450569, 0.01110393, -0.00450569, 0.00001304, -0.00051307, -0.00007161, -0.00000921, 0.00001744, -0.00029790, 0.00016904, -0.00014326, 0.00014656, -0.00021168, 0.00009319, -0.00003188, -0.00013387, 0.00027476, },
		{ -0.00007103, -0.00018335, 0.00034499, -0.00040110, 0.00033848, -0.00035376, 0.00024008, -0.00034680, 0.00042727, -0.00053677, 0.00044463, -0.00084494, 0.00113902, -0.00172908, 0.00170168, -0.00380059, 0.00118830, 0.00839537, -0.00630428, 0.00217339, -0.00197393, 0.00115244, -0.00091481, 0.00067538, -0.00082143, 0.00068893, -0.00057543, 0.00051934, -0.00054246, 0.00045899, -0.00041090, 0.00026763, },
		{ 0.00037022, -0.00032521, 0.00012432, 0.00010140, -0.00028976, 0.00032359, -0.00043124, 0.00031871, -0.00040467, 0.00047905, -0.00066057, 0.00052956, -0.00087041, 0.00119288, -0.00205572, 0.00227747, -0.00597385, 0.00720198, 0.00237184, -0.00415495, 0.00159927, -0.00154896, 0.00094353, -0.00073756, 0.00045547, -0.00055888, 0.00051010, -0.00040744, 0.00036839, -0.00038008, 0.00035813, -0.00038137, },
		{ -0.00001236, 0.00008239, -0.00023310, 0.00023683, -0.00014897, -0.00000641, 0.00002572, -0.00017018, 0.00003408, -0.00003528, -0.00000590, -0.00014477, -0.00011318, 0.00001904, -0.00002742, -0.00061973, 0.00022001, -0.00417160, 0.00953126, -0.00271878, -0.00093857, -0.00013565, -0.00020301, -0.00004220, 0.00004074, -0.00022673, 0.00011340, -0.00004567, 0.00009067, -0.00006117, 0.00005005, -0.00003825, },
		{ -0.00024337, 0.00023587, -0.00024052, 0.00004963, 0.00008360, -0.00015797, 0.00010558, -0.00017648, 0.00003727, -0.00017781, 0.00023853, -0.00041550, 0.00031995, -0.00066902, 0.00079399, -0.00123453, 0.00101712, -0.00272112, 0.00053067, 0.00729971, -0.00434076, 0.00073503, -0.00106490, 0.00056198, -0.00061617, 0.00047065, -0.00056570, 0.00047692, -0.00033906, 0.00035832, -0.00025885, 0.00025222, },
		{ 0.00010727, -0.00009027, 0.00010100, -0.00024881, 0.00017786, -0.00009323, -0.00001419, -0.00000764, -0.00013075, -0.00000897, -0.00007642, 0.00009581, -0.00032564, 0.00022210, -0.00052024, 0.00061801, -0.00122734, 0.00107043, -0.00379614, 0.00472132, 0.00303322, -0.00323405, 0.00061417, -0.00079788, 0.00037768, -0.00045145, 0.00029955, -0.00036303, 0.00035566, -0.00021019, 0.00026554, -0.00011811, },
		{ 0.00003387, -0.00005716, 0.00010678, -0.00018619, -0.00000578, 0.00004412, -0.00007655, 0.00003473, -0.00012161, -0.00002596, -0.00011882, 0.00010993, -0.00021176, 0.00001726, -0.00017340, 0.00000319, -0.00014084, -0.00032361, -0.00039925, -0.00210424, 0.00595398, -0.00021280, -0.00163283, -0.00000106, -0.00022303, -0.00007570, -0.00008283, -0.00001091, -0.00003848, 0.00012768, 0.00000376, 0.00013279, },
		{ 0.00017131, -0.00002119, 0.00000412, 0.00003476, -0.00025381, 0.00015299, -0.00013656, 0.00007047, -0.00007989, -0.00008428, -0.00004921, -0.00006905, 0.00007772, -0.00028487, 0.00014427, -0.00031827, 0.00018743, -0.00051613, 0.00017237, -0.00166041, 0.00072475, 0.00458469, -0.00141443, -0.00080545, -0.00019664, -0.00003061, -0.00023563, 0.00004193, -0.00008985, 0.00008532, 0.00007575, 0.00006365, },
		{ 0.00009135, 0.00023140, -0.00012584, 0.00013186, -0.00020584, -0.00006371, 0.00006236, -0.00013864, 0.00013307, -0.00020181, 0.00001319, -0.00013515, 0.00007110, -0.00012710, -0.00011108, 0.00000436, -0.00015609, -0.00007261, -0.00025395, -0.00031498, -0.00138382, 0.00251984, 0.00257640, -0.00144522, -0.00054308, -0.00009872, -0.00010814, -0.00018609, 0.00000803, -0.00002937, 0.00010498, 0.00009857, },
		{ 0.00004584, 0.00023924, 0.00013311, -0.00007635, 0.00006275, -0.00029615, 0.00010786, -0.00011765, 0.00001060, 0.00003034, -0.00019331, 0.00002468, -0.00012516, 0.00008951, -0.00023608, 0.00002322, -0.00009742, -0.00007668, -0.00025357, -0.00002346, -0.00100208, 0.00000899, 0.00264460, 0.00126010, -0.00132530, -0.00029950, -0.00010568, -0.00011887, -0.00019522, 0.00003783, -0.00001958, 0.00018866, },
		{ 0.00027428, 0.00000328, 0.00040800, -0.00003945, 0.00004345, -0.00016500, -0.00014859, 0.00007015, -0.00016403, 0.00011953, -0.00012043, -0.00009632, -0.00003936, -0.00002594, -0.00004077, -0.00015909, -0.00000716, -0.00003568, -0.00024576, -0.00006317, -0.00036041, -0.00087410, 0.00114677, 0.00198018, 0.00060239, -0.00124815, -0.00002959, -0.00022118, -0.00005178, -0.00026010, 0.00011132, -0.00001802, },
		{ 0.00008875, 0.00024057, 0.00011564, 0.00042376, -0.00011580, 0.00002559, -0.00029256, 0.00002620, -0.00007897, -0.00005929, 0.00008395, -0.00018322, -0.00003009, -0.00006860, 0.00002086, -0.00015015, -0.00005305, -0.00005438, -0.00001065, -0.00037701, 0.00014342, -0.00090540, 0.00006210, 0.00131951, 0.00147089, 0.00009931, -0.00100826, 0.00005474, -0.00024172, -0.00006801, -0.00024799, 0.00011512, },
		{ 0.00009232, 0.00024513, 0.00014801, 0.00033867, 0.00029146, -0.00011509, -0.00012546, -0.00021748, 0.00004504, -0.00014804, 0.00005059, -0.00003158, -0.00014834, -0.00002780, -0.00003399, -0.00004031, -0.00013237, -0.00004204, -0.00000870, -0.00015917, -0.00022919, -0.00005825, -0.00092716, 0.00094287, 0.00092228, 0.00125218, -0.00036994, -0.00063134, -0.00003982, -0.00015153, -0.00017112, -0.00017458, },
		{ -0.00014050, 0.00016104, 0.00031679, 0.00015045, 0.00051690, 0.00011422, -0.00015801, -0.00022454, -0.00008740, -0.00004320, -0.00009489, 0.00006180, -0.00011599, -0.00008985, -0.00003633, -0.00001086, -0.00011717, -0.00005789, -0.00006954, 0.00002998, -0.00031653, 0.00000275, -0.00051540, -0.00028077, 0.00114559, 0.00061746, 0.00102167, -0.00063288, -0.00033376, -0.00013264, -0.00008356, -0.00025166, },
		{ -0.00028994, -0.00014153, 0.00032346, 0.00028762, 0.00029720, 0.00056568, -0.00009007, -0.00022367, -0.00020256, -0.00004187, -0.00010084, -0.00001218, -0.00000738, -0.00012687, -0.00005682, -0.00002335, -0.00004980, -0.00011782, -0.00002662, -0.00005489, -0.00004165, -0.00029025, -0.00001101, -0.00070530, 0.00048961, 0.00082152, 0.00058733, 0.00066549, -0.00064862, -0.00019743, -0.00014836, -0.00008385, },
		{ -0.00010296, -0.00034526, -0.00009258, 0.00049194, 0.00024299, 0.00053844, 0.00044182, -0.00031210, -0.00020124, -0.00016753, -0.00004912, -0.00009355, 0.00002402, -0.00008868, -0.00008742, -0.00004693, -0.00001445, -0.00011119, -0.00005615, -0.00004197, -0.00002836, -0.00014789, -0.00016031, -0.00025110, -0.00041259, 0.00085746, 0.00044884, 0.00063536, 0.00030892, -0.00054896, -0.00014663, -0.00013756, },
		},
		{ { 0.00104084, 0.00045932, 0.00026447, -0.00005817, -0.00054403, -0.00013668, -0.00015960, 0.00006321, -0.00023377, 0.00005285, -0.00004755, 0.00010186, -0.00012492, 0.00011521, -0.00018680, 0.00017926, -0.00018486, 0.00017926, -0.00018680, 0.00011521, -0.00012492, 0.00010186, -0.00004755, 0.00005285, -0.00023377, 0.00006321, -0.00015960, -0.00013668, -0.00054403, -0.00005817, 0.00026447, 0.00045932, },
		{ 0.00039111, 0.00106177, 0.00052901, 0.00050406, -0.00028378, -0.00054244, -0.00012011, -0.00020977, 0.00014508, -0.00029062, 0.00009292, -0.00008026, 0.00012691, -0.00013439, 0.00010951, -0.00017450, 0.00017864, -0.00018148, 0.00017096, -0.00019404, 0.00012666, -0.00012831, 0.00009364, -0.00003622, 0.00003227, -0.00018490, -0.00002249, -0.00009410, -0.00018248, -0.00047791, 0.00009030, 0.00003023, },
		{ -0.00017942, 0.00030929, 0.00111956, 0.00062943, 0.00070518, -0.00056575, -0.00046197, -0.00015444, -0.00021952, 0.00019547, -0.00033034, 0.00013430, -0.00012354, 0.00016364, -0.00015701, 0.00011445, -0.00016322, 0.00017301, -0.00017327, 0.00015339, -0.00019202, 0.00013771, -0.00014015, 0.00010141, -0.00004979, 0.00004232, -0.00016521, -0.00008313, -0.00004513, -0.00022223, -0.00037879, 0.00017104, },
		{ 0.00021178, -0.00036410, 0.00021013, 0.00120357, 0.00079704, 0.00080472, -0.00085173, -0.00032296, -0.00023686, -0.00018216, 0.00019944, -0.00033407, 0.00015929, -0.00016395, 0.00020304, -0.00018897, 0.00013216, -0.00015889, 0.00016828, -0.00016526, 0.00013001, -0.00017977, 0.00014291, -0.00015334, 0.00011840, -0.00008252, 0.00008204, -0.00018335, -0.00010029, -0.00003738, -0.00022282, -0.00028911, },
		{ -0.00024119, 0.00024150, -0.00052963, 0.00009149, 0.00130890, 0.00106037, 0.00073627, -0.00106870, -0.00017379, -0.00033840, -0.00011421, 0.00016184, -0.00029714, 0.00015789, -0.00019085, 0.00023569, -0.00022372, 0.00016127, -0.00016521, 0.00017001, -0.00016361, 0.00010697, -0.00015983, 0.00013915, -0.00016037, 0.00013407, -0.00012076, 0.00013772, -0.00023144, -0.00007364, -0.00007734, -0.00016802, },
		{ -0.00006665, -0.00024419, 0.00027509, -0.00067453, -0.00005842, 0.00144917, 0.00142193, 0.00045247, -0.00114343, -0.00007658, -0.00041230, -0.00005133, 0.00010693, -0.00023186, 0.00013210, -0.00020069, 0.00025529, -0.00025408, 0.00019722, -0.00018235, 0.00018120, -0.00017311, 0.00009094, -0.00013734, 0.00012649, -0.00015588, 0.00013791, -0.00014746, 0.00018703, -0.00028732, -0.00002163, -0.00014935, },
		{ -0.00022045, 0.00004907, -0.00028056, 0.00030571, -0.00078223, -0.00026748, 0.00166432, 0.00184428, -0.00005363, -0.00102633, -0.00008659, -0.00041012, -0.00003640, 0.00007074, -0.00016398, 0.00009559, -0.00019851, 0.00026099, -0.00027459, 0.00023353, -0.00020673, 0.00020087, -0.00019504, 0.00008671, -0.00011823, 0.00010809, -0.00013849, 0.00012331, -0.00014820, 0.00020669, -0.00032228, 0.00002519, },
		{ 0.00003590, -0.00025242, 0.00013442, -0.00031330, 0.00030772, -0.00082271, -0.00057673, 0.00201801, 0.00224704, -0.00073606, -0.00071077, -0.00023195, -0.00029995, -0.00010368, 0.00008829, -0.00012397, 0.00006891, -0.00019611, 0.00025762, -0.00028309, 0.00026366, -0.00023202, 0.00022411, -0.00022623, 0.00009537, -0.00010726, 0.00008911, -0.00011144, 0.00009043, -0.00011681, 0.00018096, -0.00031178, },
		{ -0.00024528, -0.00000826, -0.00021685, 0.00014891, -0.00030149, 0.00024921, -0.00076370, -0.00102754, 0.00258303, 0.00251833, -0.00149775, -0.00024106, -0.00050087, -0.00008049, -0.00026459, 0.00018016, -0.00013595, 0.00007172, -0.00020737, 0.00025385, -0.00028115, 0.00028273, -0.00025112, 0.00024350, -0.00025967, 0.00011370, -0.00010648, 0.00007504, -0.00008176, 0.00004678, -0.00005850, 0.00010820, },
		{ 0.00000288, -0.00013170, -0.00010487, -0.00010780, 0.00007266, -0.00021840, 0.00010901, -0.00058733, -0.00164219, 0.00341744, 0.00253922, -0.00220802, 0.00029455, -0.00084109, 0.00021422, -0.00050012, 0.00034304, -0.00020815, 0.00011492, -0.00024251, 0.00025868, -0.00027316, 0.00028870, -0.00025834, 0.00025153, -0.00028654, 0.00013514, -0.00011459, 0.00007001, -0.00005813, 0.00000542, 0.00001078, },
		{ 0.00006853, -0.00010765, 0.00000161, -0.00022856, 0.00005264, -0.00008333, -0.00006533, -0.00010827, -0.00030572, -0.00240471, 0.00453828, 0.00221562, -0.00273188, 0.00078445, -0.00117225, 0.00052166, -0.00076308, 0.00054783, -0.00032838, 0.00019531, -0.00030330, 0.00027778, -0.00026452, 0.00028268, -0.00025118, 0.00024321, -0.00029901, 0.00015189, -0.00012759, 0.00007568, -0.00004824, -0.00001893, },
		{ -0.00001441, 0.00009243, -0.00019096, 0.00011633, -0.00033724, 0.00021838, -0.00027725, 0.00012406, -0.00036582, 0.00003143, -0.00324960, 0.00590156, 0.00150836, -0.00296347, 0.00112244, -0.00140801, 0.00076751, -0.00098987, 0.00074622, -0.00046619, 0.00029511, -0.00038087, 0.00031077, -0.00025941, 0.00026819, -0.00023109, 0.00021793, -0.00029294, 0.00015750, -0.00014024, 0.00009089, -0.00005645, },
		{ -0.00008250, 0.00002400, 0.00006740, -0.00021938, 0.00017486, -0.00038485, 0.00033220, -0.00044733, 0.00029129, -0.00059997, 0.00034171, -0.00406431, 0.00739614, 0.00045223, -0.00285391, 0.00123611, -0.00148104, 0.00088727, -0.00111857, 0.00088421, -0.00058138, 0.00038637, -0.00045737, 0.00035053, -0.00025924, 0.00024976, -0.00020299, 0.00017996, -0.00026941, 0.00014920, -0.00014800, 0.00011229, },
		{ 0.00013563, -0.00012156, 0.00009258, -0.00000931, -0.00017883, 0.00015248, -0.00033648, 0.00034345, -0.00053050, 0.00037044, -0.00073687, 0.00052772, -0.00470719, 0.00885578, -0.00084267, -0.00242536, 0.00110640, -0.00136310, 0.00084587, -0.00110759, 0.00091815, -0.00063622, 0.00043942, -0.00051081, 0.00038501, -0.00026196, 0.00023145, -0.00017367, 0.00013743, -0.00023473, 0.00012900, -0.00014871, },
		{ -0.00014340, 0.00015724, -0.00016581, 0.00017956, -0.00012655, -0.00007371, 0.00004590, -0.00018094, 0.00022512, -0.00048279, 0.00031005, -0.00071548, 0.00050336, -0.00503778, 0.01008793, -0.00220907, -0.00176680, 0.00077249, -0.00107402, 0.00064918, -0.00094904, 0.00082851, -0.00060751, 0.00043233, -0.00052180, 0.00040082, -0.00026286, 0.00021555, -0.00014969, 0.00009999, -0.00019872, 0.00010320, },
		{ 0.00008060, -0.00013593, 0.00017515, -0.00020682, 0.00026827, -0.00026199, 0.00007380, -0.00012454, 0.00006352, -0.00001561, -0.00029502, 0.00009209, -0.00050956, 0.00022110, -0.00495171, 0.01091287, -0.00345618, -0.00101241, 0.00032039, -0.00067648, 0.00034327, -0.00067236, 0.00062663, -0.00049452, 0.00035854, -0.00048003, 0.00038761, -0.00025630, 0.00020214, -0.00013538, 0.00007592, -0.00017179, },
		{ -0.00016184, 0.00006968, -0.00013163, 0.00018949, -0.00023799, 0.00034180, -0.00038816, 0.00022982, -0.00032009, 0.00035244, -0.00033689, 0.00000177, -0.00025833, -0.00014105, -0.00030859, -0.00441032, 0.01120319, -0.00441032, -0.00030859, -0.00014105, -0.00025833, 0.00000177, -0.00033689, 0.00035244, -0.00032009, 0.00022982, -0.00038816, 0.00034180, -0.00023799, 0.00018949, -0.00013163, 0.00006968, },
		{ 0.00008060, -0.00017179, 0.00007592, -0.00013538, 0.00020214, -0.00025630, 0.00038761, -0.00048003, 0.00035854, -0.00049452, 0.00062663, -0.00067236, 0.00034327, -0.00067648, 0.00032039, -0.00101241, -0.00345618, 0.01091287, -0.00495171, 0.00022110, -0.00050956, 0.00009209, -0.00029502, -0.00001561, 0.00006352, -0.00012454, 0.00007380, -0.00026199, 0.00026827, -0.00020682, 0.00017515, -0.00013593, },
		{ -0.00014340, 0.00010320, -0.00019872, 0.00009999, -0.00014969, 0.00021555, -0.00026286, 0.00040082, -0.00052180, 0.00043233, -0.00060751, 0.00082851, -0.00094904, 0.00064918, -0.00107402, 0.00077249, -0.00176680, -0.00220907, 0.01008793, -0.00503778, 0.00050336, -0.00071548, 0.00031005, -0.00048279, 0.00022512, -0.00018094, 0.00004590, -0.00007371, -0.00012655, 0.00017956, -0.00016581, 0.00015724, },
		{ 0.00013563, -0.00014871, 0.00012900, -0.00023473, 0.00013743, -0.00017367, 0.00023145, -0.00026196, 0.00038501, -0.00051081, 0.00043942, -0.00063622, 0.00091815, -0.00110759, 0.00084587, -0.00136310, 0.00110640, -0.00242536, -0.00084267, 0.00885578, -0.00470719, 0.00052772, -0.00073687, 0.00037044, -0.00053050, 0.00034345, -0.00033648, 0.00015248, -0.00017883, -0.00000931, 0.00009258, -0.00012156, },
		{ -0.00008250, 0.00011229, -0.00014800, 0.00014920, -0.00026941, 0.00017996, -0.00020299, 0.00024976, -0.00025924, 0.00035053, -0.00045737, 0.00038637, -0.00058138, 0.00088421, -0.00111857, 0.00088727, -0.00148104, 0.00123611, -0.00285391, 0.00045223, 0.00739614, -0.00406431, 0.00034171, -0.00059997, 0.00029129, -0.00044733, 0.00033220, -0.00038485, 0.00017486, -0.00021938, 0.00006740, 0.00002400, },
		{ -0.00001441, -0.00005645, 0.00009089, -0.00014024, 0.00015750, -0.00029294, 0.00021793, -0.00023109, 0.00026819, -0.00025941, 0.00031077, -0.00038087, 0.00029511, -0.00046619, 0.00074622, -0.00098987, 0.00076751, -0.00140801, 0.00112244, -0.00296347, 0.00150836, 0.00590156, -0.00324960, 0.00003143, -0.00036582, 0.00012406, -0.00027725, 0.00021838, -0.00033724, 0.00011633, -0.00019096, 0.00009243, },
		{ 0.00006853, -0.00001893, -0.00004824, 0.00007568, -0.00012759, 0.00015189, -0.00029901, 0.00024321, -0.00025118, 0.00028268, -0.00026452, 0.00027778, -0.00030330, 0.00019531, -0.00032838, 0.00054783, -0.00076308, 0.00052166, -0.00117225, 0.00078445, -0.00273188, 0.00221562, 0.00453828, -0.00240471, -0.00030572, -0.00010827, -0.00006533, -0.00008333, 0.00005264, -0.00022856, 0.00000161, -0.00010765, },
		{ 0.00000288, 0.00001078, 0.00000542, -0.00005813, 0.00007001, -0.00011459, 0.00013514, -0.00028654, 0.00025153, -0.00025834, 0.00028870, -0.00027316, 0.00025868, -0.00024251, 0.00011492, -0.00020815, 0.00034304, -0.00050012, 0.00021422, -0.00084109, 0.00029455, -0.00220802, 0.00253922, 0.00341744, -0.00164219, -0.00058733, 0.00010901, -0.00021840, 0.00007266, -0.00010780, -0.00010487, -0.00013170, },
		{ -0.00024528, 0.00010820, -0.00005850, 0.00004678, -0.00008176, 0.00007504, -0.00010648, 0.00011370, -0.00025967, 0.00024350, -0.00025112, 0.00028273, -0.00028115, 0.00025385, -0.00020737, 0.00007172, -0.00013595, 0.00018016, -0.00026459, -0.00008049, -0.00050087, -0.00024106, -0.00149775, 0.00251833, 0.00258303, -0.00102754, -0.00076370, 0.00024921, -0.00030149, 0.00014891, -0.00021685, -0.00000826, },
		{ 0.00003590, -0.00031178, 0.00018096, -0.00011681, 0.00009043, -0.00011144, 0.00008911, -0.00010726, 0.00009537, -0.00022623, 0.00022411, -0.00023202, 0.00026366, -0.00028309, 0.00025762, -0.00019611, 0.00006891, -0.00012397, 0.00008829, -0.00010368, -0.00029995, -0.00023195, -0.00071077, -0.00073606, 0.00224704, 0.00201801, -0.00057673, -0.00082271, 0.00030772, -0.00031330, 0.00013442, -0.00025242, },
		{ -0.00022045, 0.00002519, -0.00032228, 0.00020669, -0.00014820, 0.00012331, -0.00013849, 0.00010809, -0.00011823, 0.00008671, -0.00019504, 0.00020087, -0.00020673, 0.00023353, -0.00027459, 0.00026099, -0.00019851, 0.00009559, -0.00016398, 0.00007074, -0.00003640, -0.00041012, -0.00008659, -0.00102633, -0.00005363, 0.00184428, 0.00166432, -0.00026748, -0.00078223, 0.00030571, -0.00028056, 0.00004907, },
		{ -0.00006665, -0.00014935, -0.00002163, -0.00028732, 0.00018703, -0.00014746, 0.00013791, -0.00015588, 0.00012649, -0.00013734, 0.00009094, -0.00017311, 0.00018120, -0.00018235, 0.00019722, -0.00025408, 0.00025529, -0.00020069, 0.00013210, -0.00023186, 0.00010693, -0.00005133, -0.00041230, -0.00007658, -0.00114343, 0.00045247, 0.00142193, 0.00144917, -0.00005842, -0.00067453, 0.00027509, -0.00024419, },
		{ -0.00024119, -0.00016802, -0.00007734, -0.00007364, -0.00023144, 0.00013772, -0.00012076, 0.00013407, -0.00016037, 0.00013915, -0.00015983, 0.00010697, -0.00016361, 0.00017001, -0.00016521, 0.00016127, -0.00022372, 0.00023569, -0.00019085, 0.00015789, -0.00029714, 0.00016184, -0.00011421, -0.00033840, -0.00017379, -0.00106870, 0.00073627, 0.00106037, 0.00130890, 0.00009149, -0.00052963, 0.00024150, },
		{ 0.00021178, -0.00028911, -0.00022282, -0.00003738, -0.00010029, -0.00018335, 0.00008204, -0.00008252, 0.00011840, -0.00015334, 0.00014291, -0.00017977, 0.00013001, -0.00016526, 0.00016828, -0.00015889, 0.00013216, -0.00018897, 0.00020304, -0.00016395, 0.00015929, -0.00033407, 0.00019944, -0.00018216, -0.00023686, -0.00032296, -0.00085173, 0.00080472, 0.00079704, 0.00120357, 0.00021013, -0.00036410, },
		{ -0.00017942, 0.00017104, -0.00037879, -0.00022223, -0.00004513, -0.00008313, -0.00016521, 0.00004232, -0.00004979, 0.00010141, -0.00014015, 0.00013771, -0.00019202, 0.00015339, -0.00017327, 0.00017301, -0.00016322, 0.00011445, -0.00015701, 0.00016364, -0.00012354, 0.00013430, -0.00033034, 0.00019547, -0.00021952, -0.00015444, -0.00046197, -0.00056575, 0.00070518, 0.00062943, 0.00111956, 0.00030929, },
		{ 0.00039111, 0.00003023, 0.00009030, -0.00047791, -0.00018248, -0.00009410, -0.00002249, -0.00018490, 0.00003227, -0.00003622, 0.00009364, -0.00012831, 0.00012666, -0.00019404, 0.00017096, -0.00018148, 0.00017864, -0.00017450, 0.00010951, -0.00013439, 0.00012691, -0.00008026, 0.00009292, -0.00029062, 0.00014508, -0.00020977, -0.00012011, -0.00054244, -0.00028378, 0.00050406, 0.00052901, 0.00106177, },
		},
	}
	;

	float f128_5[32][128][128] = {
		{ { -0.00011173, -0.00001466, 0.00005435, 0.00002146, 0.00001633, 0.00001645, -0.00000177, 0.00001876, 0.00002035, 0.00004356, 0.00002513, -0.00001884, -0.00004351, -0.00001056, -0.00001025, -0.00001345, 0.00000468, -0.00000803, -0.00000545, 0.00000016, -0.00000856, 0.00001059, -0.00001329, -0.00000229, -0.00000326, -0.00000204, 0.00000840, -0.00000682, -0.00000824, -0.00000507, 0.00000964, 0.00000442, -0.00000862, 0.00000014, -0.00000233, 0.00000567, 0.00000026, -0.00000539, 0.00000160, -0.00000325, 0.00000517, 0.00000209, 0.00000042, -0.00000061, -0.00000258, -0.00000554, 0.00000587, -0.00000325, 0.00000200, 0.00000112, 0.00000206, -0.00000837, -0.00000346, 0.00000901, -0.00000249, -0.00000149, 0.00001413, -0.00000927, -0.00000246, -0.00001117, 0.00001156, 0.00000157, -0.00000205, 0.00000052, -0.00000476, 0.00000052, -0.00000205, 0.00000157, 0.00001156, -0.00001117, -0.00000246, -0.00000927, 0.00001413, -0.00000149, -0.00000249, 0.00000901, -0.00000346, -0.00000837, 0.00000206, 0.00000112, 0.00000200, -0.00000325, 0.00000587, -0.00000554, -0.00000258, -0.00000061, 0.00000042, 0.00000209, 0.00000517, -0.00000325, 0.00000160, -0.00000539, 0.00000026, 0.00000567, -0.00000233, 0.00000014, -0.00000862, 0.00000442, 0.00000964, -0.00000507, -0.00000824, -0.00000682, 0.00000840, -0.00000204, -0.00000326, -0.00000229, -0.00001329, 0.00001059, -0.00000856, 0.00000016, -0.00000545, -0.00000803, 0.00000468, -0.00001345, -0.00001025, -0.00001056, -0.00004351, -0.00001884, 0.00002513, 0.00004356, 0.00002035, 0.00001876, -0.00000177, 0.00001645, 0.00001633, 0.00002146, 0.00005435, -0.00001466, },
		{ -0.00001278, -0.00011181, -0.00001646, 0.00005061, 0.00002123, 0.00001502, 0.00001685, -0.00000458, 0.00001940, 0.00001637, 0.00003977, 0.00002913, -0.00001069, -0.00004177, -0.00001693, -0.00000750, -0.00001370, 0.00000291, -0.00000509, -0.00000843, 0.00000264, -0.00001066, 0.00001050, -0.00001099, -0.00000438, -0.00000070, -0.00000542, 0.00000959, -0.00000485, -0.00000849, -0.00000628, 0.00000641, 0.00000883, -0.00000867, -0.00000117, -0.00000202, 0.00000403, 0.00000282, -0.00000643, 0.00000123, -0.00000299, 0.00000471, 0.00000228, 0.00000069, -0.00000089, -0.00000117, -0.00000695, 0.00000521, -0.00000196, 0.00000132, 0.00000113, 0.00000241, -0.00000680, -0.00000548, 0.00000752, 0.00000131, -0.00000608, 0.00001681, -0.00000861, -0.00000189, -0.00001201, 0.00000892, 0.00000457, -0.00000310, 0.00000125, -0.00000446, -0.00000069, -0.00000068, -0.00000122, 0.00001312, -0.00000927, -0.00000354, -0.00000913, 0.00001066, 0.00000286, -0.00000503, 0.00000888, -0.00000050, -0.00000990, 0.00000158, 0.00000103, 0.00000268, -0.00000421, 0.00000586, -0.00000367, -0.00000388, -0.00000060, 0.00000036, 0.00000175, 0.00000562, -0.00000331, 0.00000152, -0.00000385, -0.00000225, 0.00000666, -0.00000196, 0.00000064, -0.00000746, -0.00000017, 0.00001167, -0.00000294, -0.00000827, -0.00000798, 0.00000616, 0.00000155, -0.00000530, -0.00000096, -0.00001444, 0.00000932, -0.00000560, -0.00000248, -0.00000286, -0.00001001, 0.00000499, -0.00001161, -0.00001395, -0.00000512, -0.00004262, -0.00002737, 0.00001962, 0.00004655, 0.00002526, 0.00001829, 0.00000074, 0.00001648, 0.00001738, 0.00002179, 0.00005814, },
		{ 0.00006207, -0.00001083, -0.00011206, -0.00001811, 0.00004685, 0.00002119, 0.00001341, 0.00001766, -0.00000755, 0.00001996, 0.00001341, 0.00003535, 0.00003194, -0.00000351, -0.00003769, -0.00002325, -0.00000647, -0.00001211, -0.00000012, -0.00000165, -0.00001122, 0.00000439, -0.00001148, 0.00000893, -0.00000775, -0.00000686, 0.00000191, -0.00000811, 0.00000946, -0.00000208, -0.00000900, -0.00000663, 0.00000244, 0.00001244, -0.00000731, -0.00000326, -0.00000119, 0.00000201, 0.00000512, -0.00000680, 0.00000044, -0.00000263, 0.00000432, 0.00000228, 0.00000113, -0.00000130, 0.00000014, -0.00000776, 0.00000394, -0.00000056, 0.00000087, 0.00000087, 0.00000285, -0.00000557, -0.00000625, 0.00000450, 0.00000585, -0.00001028, 0.00001825, -0.00000698, -0.00000195, -0.00001182, 0.00000547, 0.00000738, -0.00000355, 0.00000139, -0.00000360, -0.00000220, 0.00000071, -0.00000345, 0.00001343, -0.00000646, -0.00000492, -0.00000846, 0.00000689, 0.00000641, -0.00000605, 0.00000730, 0.00000298, -0.00001103, 0.00000078, 0.00000099, 0.00000323, -0.00000474, 0.00000522, -0.00000156, -0.00000487, -0.00000093, 0.00000048, 0.00000132, 0.00000595, -0.00000309, 0.00000102, -0.00000204, -0.00000440, 0.00000683, -0.00000093, 0.00000051, -0.00000560, -0.00000435, 0.00001223, 0.00000001, -0.00000844, -0.00000847, 0.00000325, 0.00000483, -0.00000644, -0.00000058, -0.00001445, 0.00000699, -0.00000233, -0.00000472, -0.00000111, -0.00001079, 0.00000392, -0.00000869, -0.00001773, -0.00000135, -0.00003926, -0.00003547, 0.00001236, 0.00004851, 0.00003098, 0.00001821, 0.00000290, 0.00001693, 0.00001830, 0.00002211, },
		{ 0.00002234, 0.00006616, -0.00000875, -0.00011254, -0.00001955, 0.00004295, 0.00002142, 0.00001148, 0.00001878, -0.00001048, 0.00002019, 0.00001158, 0.00003040, 0.00003384, 0.00000239, -0.00003188, -0.00002848, -0.00000771, -0.00000880, -0.00000388, 0.00000164, -0.00001322, 0.00000495, -0.00001082, 0.00000607, -0.00000399, -0.00000924, 0.00000406, -0.00000967, 0.00000790, 0.00000133, -0.00000961, -0.00000637, -0.00000171, 0.00001472, -0.00000441, -0.00000596, -0.00000009, -0.00000004, 0.00000688, -0.00000646, -0.00000069, -0.00000229, 0.00000407, 0.00000209, 0.00000166, -0.00000173, 0.00000120, -0.00000794, 0.00000224, 0.00000066, 0.00000088, 0.00000018, 0.00000356, -0.00000491, -0.00000570, 0.00000029, 0.00001049, -0.00001347, 0.00001814, -0.00000434, -0.00000258, -0.00001076, 0.00000164, 0.00000960, -0.00000324, 0.00000097, -0.00000236, -0.00000374, 0.00000181, -0.00000487, 0.00001251, -0.00000304, -0.00000627, -0.00000761, 0.00000330, 0.00000876, -0.00000551, 0.00000462, 0.00000642, -0.00001141, -0.00000045, 0.00000106, 0.00000355, -0.00000482, 0.00000409, 0.00000054, -0.00000538, -0.00000162, 0.00000073, 0.00000090, 0.00000609, -0.00000259, 0.00000021, -0.00000019, -0.00000593, 0.00000611, 0.00000061, 0.00000002, -0.00000350, -0.00000765, 0.00001125, 0.00000351, -0.00000853, -0.00000850, 0.00000010, 0.00000737, -0.00000649, -0.00000111, -0.00001355, 0.00000406, 0.00000071, -0.00000612, -0.00000041, -0.00001040, 0.00000184, -0.00000537, -0.00002082, 0.00000037, -0.00003395, -0.00004226, 0.00000329, 0.00004908, 0.00003742, 0.00001871, 0.00000468, 0.00001771, 0.00001918, },
		{ 0.00002012, 0.00002242, 0.00007039, -0.00000647, -0.00011335, -0.00002072, 0.00003886, 0.00002191, 0.00000932, 0.00002003, -0.00001313, 0.00001986, 0.00001088, 0.00002508, 0.00003499, 0.00000704, -0.00002522, -0.00003172, -0.00001137, -0.00000427, -0.00000762, 0.00000408, -0.00001394, 0.00000409, -0.00000878, 0.00000231, -0.00000029, -0.00001099, 0.00000532, -0.00000986, 0.00000504, 0.00000505, -0.00001006, -0.00000585, -0.00000548, 0.00001527, -0.00000009, -0.00000889, 0.00000091, -0.00000180, 0.00000789, -0.00000542, -0.00000197, -0.00000210, 0.00000401, 0.00000176, 0.00000219, -0.00000204, 0.00000194, -0.00000759, 0.00000041, 0.00000140, 0.00000157, -0.00000105, 0.00000466, -0.00000495, -0.00000396, -0.00000457, 0.00001454, -0.00001512, 0.00001636, -0.00000087, -0.00000360, -0.00000911, -0.00000217, 0.00001092, -0.00000213, 0.00000012, -0.00000097, -0.00000505, 0.00000238, -0.00000537, 0.00001055, 0.00000060, -0.00000726, -0.00000686, 0.00000025, 0.00000973, -0.00000363, 0.00000137, 0.00000932, -0.00001081, -0.00000211, 0.00000122, 0.00000365, -0.00000453, 0.00000268, 0.00000239, -0.00000531, -0.00000261, 0.00000099, 0.00000060, 0.00000599, -0.00000183, -0.00000076, 0.00000146, -0.00000671, 0.00000460, 0.00000241, -0.00000052, -0.00000158, -0.00000979, 0.00000889, 0.00000715, -0.00000829, -0.00000830, -0.00000294, 0.00000888, -0.00000545, -0.00000237, -0.00001209, 0.00000104, 0.00000303, -0.00000642, -0.00000076, -0.00000912, -0.00000075, -0.00000230, -0.00002266, 0.00000005, -0.00002753, -0.00004689, -0.00000737, 0.00004780, 0.00004446, 0.00001993, 0.00000612, 0.00001875, },
		{ 0.00001995, 0.00002117, 0.00002238, 0.00007470, -0.00000389, -0.00011455, -0.00002156, 0.00003459, 0.00002261, 0.00000707, 0.00002120, -0.00001529, 0.00001880, 0.00001124, 0.00001962, 0.00003542, 0.00001071, -0.00001862, -0.00003235, -0.00001716, 0.00000067, -0.00001052, 0.00000508, -0.00001309, 0.00000183, -0.00000570, -0.00000172, 0.00000274, -0.00001165, 0.00000542, -0.00000869, 0.00000120, 0.00000861, -0.00000998, -0.00000543, -0.00000839, 0.00001393, 0.00000528, -0.00001158, 0.00000148, -0.00000301, 0.00000807, -0.00000381, -0.00000324, -0.00000215, 0.00000413, 0.00000134, 0.00000265, -0.00000221, 0.00000237, -0.00000690, -0.00000121, 0.00000135, 0.00000306, -0.00000284, 0.00000613, -0.00000561, -0.00000136, -0.00000943, 0.00001733, -0.00001491, 0.00001303, 0.00000313, -0.00000473, -0.00000719, -0.00000555, 0.00001116, -0.00000033, -0.00000091, 0.00000032, -0.00000588, 0.00000228, -0.00000499, 0.00000786, 0.00000403, -0.00000760, -0.00000645, -0.00000205, 0.00000934, -0.00000078, -0.00000187, 0.00001123, -0.00000916, -0.00000406, 0.00000137, 0.00000359, -0.00000399, 0.00000120, 0.00000380, -0.00000463, -0.00000378, 0.00000114, 0.00000048, 0.00000563, -0.00000090, -0.00000172, 0.00000271, -0.00000668, 0.00000250, 0.00000414, -0.00000078, -0.00000014, -0.00001068, 0.00000550, 0.00001046, -0.00000750, -0.00000804, -0.00000560, 0.00000926, -0.00000348, -0.00000400, -0.00001049, -0.00000163, 0.00000431, -0.00000561, -0.00000190, -0.00000737, -0.00000331, -0.00000002, -0.00002300, -0.00000195, -0.00002096, -0.00004879, -0.00001906, 0.00004413, 0.00005190, 0.00002207, 0.00000727, },
		{ 0.00000817, 0.00002129, 0.00002237, 0.00002224, 0.00007901, -0.00000091, -0.00011619, -0.00002211, 0.00003017, 0.00002340, 0.00000490, 0.00002208, -0.00001678, 0.00001698, 0.00001245, 0.00001433, 0.00003501, 0.00001381, -0.00001290, -0.00003017, -0.00002436, 0.00000501, -0.00001181, 0.00000428, -0.00001070, -0.00000147, -0.00000217, -0.00000535, 0.00000453, -0.00001095, 0.00000432, -0.00000638, -0.00000308, 0.00001147, -0.00000906, -0.00000544, -0.00001013, 0.00001081, 0.00001116, -0.00001351, 0.00000129, -0.00000353, 0.00000746, -0.00000185, -0.00000427, -0.00000253, 0.00000439, 0.00000091, 0.00000299, -0.00000224, 0.00000263, -0.00000614, -0.00000228, 0.00000030, 0.00000536, -0.00000508, 0.00000786, -0.00000669, 0.00000163, -0.00001362, 0.00001836, -0.00001278, 0.00000847, 0.00000720, -0.00000564, -0.00000534, -0.00000820, 0.00001030, 0.00000192, -0.00000184, 0.00000125, -0.00000610, 0.00000150, -0.00000392, 0.00000486, 0.00000681, -0.00000708, -0.00000648, -0.00000352, 0.00000779, 0.00000252, -0.00000455, 0.00001190, -0.00000656, -0.00000608, 0.00000138, 0.00000349, -0.00000334, -0.00000015, 0.00000466, -0.00000343, -0.00000494, 0.00000105, 0.00000060, 0.00000506, 0.00000008, -0.00000251, 0.00000346, -0.00000591, 0.00000011, 0.00000546, -0.00000054, 0.00000070, -0.00001045, 0.00000157, 0.00001301, -0.00000601, -0.00000782, -0.00000771, 0.00000856, -0.00000092, -0.00000564, -0.00000912, -0.00000361, 0.00000445, -0.00000388, -0.00000345, -0.00000561, -0.00000535, 0.00000117, -0.00002193, -0.00000498, -0.00001521, -0.00004775, -0.00003093, 0.00003756, 0.00005940, 0.00002532, },
		{ 0.00002992, 0.00000887, 0.00002271, 0.00002371, 0.00002204, 0.00008327, 0.00000250, -0.00011824, -0.00002244, 0.00002573, 0.00002413, 0.00000299, 0.00002251, -0.00001755, 0.00001449, 0.00001421, 0.00000959, 0.00003358, 0.00001676, -0.00000860, -0.00002546, -0.00003194, 0.00000774, -0.00001094, 0.00000165, -0.00000712, -0.00000525, 0.00000110, -0.00000791, 0.00000470, -0.00000885, 0.00000224, -0.00000338, -0.00000720, 0.00001313, -0.00000706, -0.00000604, -0.00001059, 0.00000627, 0.00001686, -0.00001419, 0.00000019, -0.00000335, 0.00000623, 0.00000019, -0.00000490, -0.00000324, 0.00000477, 0.00000050, 0.00000324, -0.00000223, 0.00000289, -0.00000562, -0.00000252, -0.00000180, 0.00000832, -0.00000753, 0.00000957, -0.00000785, 0.00000445, -0.00001655, 0.00001736, -0.00000893, 0.00000322, 0.00001084, -0.00000603, -0.00000380, -0.00000994, 0.00000845, 0.00000430, -0.00000239, 0.00000166, -0.00000567, 0.00000018, -0.00000244, 0.00000196, 0.00000864, -0.00000565, -0.00000689, -0.00000427, 0.00000543, 0.00000573, -0.00000626, 0.00001124, -0.00000327, -0.00000787, 0.00000110, 0.00000341, -0.00000272, -0.00000125, 0.00000493, -0.00000186, -0.00000592, 0.00000063, 0.00000093, 0.00000437, 0.00000099, -0.00000301, 0.00000367, -0.00000459, -0.00000223, 0.00000612, 0.00000030, 0.00000094, -0.00000936, -0.00000239, 0.00001447, -0.00000384, -0.00000760, -0.00000926, 0.00000695, 0.00000185, -0.00000690, -0.00000826, -0.00000475, 0.00000356, -0.00000159, -0.00000496, -0.00000427, -0.00000658, 0.00000121, -0.00001979, -0.00000828, -0.00001100, -0.00004397, -0.00004191, 0.00002776, 0.00006644, },
		{ 0.00007226, 0.00003611, 0.00000944, 0.00002420, 0.00002520, 0.00002181, 0.00008744, 0.00000636, -0.00012064, -0.00002266, 0.00002139, 0.00002464, 0.00000147, 0.00002242, -0.00001765, 0.00001154, 0.00001610, 0.00000577, 0.00003097, 0.00001987, -0.00000592, -0.00001891, -0.00003868, 0.00000802, -0.00000775, -0.00000245, -0.00000298, -0.00000875, 0.00000344, -0.00000893, 0.00000314, -0.00000560, -0.00000037, -0.00000029, -0.00001050, 0.00001319, -0.00000392, -0.00000724, -0.00000992, 0.00000087, 0.00002167, -0.00001327, -0.00000183, -0.00000261, 0.00000464, 0.00000203, -0.00000496, -0.00000430, 0.00000521, 0.00000009, 0.00000347, -0.00000234, 0.00000339, -0.00000560, -0.00000177, -0.00000487, 0.00001165, -0.00000984, 0.00001094, -0.00000873, 0.00000662, -0.00001779, 0.00001434, -0.00000380, -0.00000210, 0.00001359, -0.00000567, -0.00000274, -0.00001075, 0.00000587, 0.00000645, -0.00000234, 0.00000149, -0.00000468, -0.00000147, -0.00000091, -0.00000046, 0.00000933, -0.00000341, -0.00000754, -0.00000450, 0.00000268, 0.00000837, -0.00000674, 0.00000940, 0.00000030, -0.00000912, 0.00000042, 0.00000343, -0.00000219, -0.00000201, 0.00000467, -0.00000011, -0.00000653, -0.00000015, 0.00000142, 0.00000365, 0.00000170, -0.00000315, 0.00000339, -0.00000294, -0.00000421, 0.00000593, 0.00000170, 0.00000076, -0.00000780, -0.00000590, 0.00001466, -0.00000114, -0.00000730, -0.00001030, 0.00000472, 0.00000445, -0.00000751, -0.00000805, -0.00000507, 0.00000189, 0.00000082, -0.00000603, -0.00000362, -0.00000693, 0.00000030, -0.00001713, -0.00001114, -0.00000874, -0.00003809, -0.00005085, 0.00001471, },
		{ -0.00000125, 0.00007592, 0.00004408, 0.00001004, 0.00002566, 0.00002689, 0.00002154, 0.00009155, 0.00001061, -0.00012329, -0.00002290, 0.00001727, 0.00002485, 0.00000038, 0.00002185, -0.00001725, 0.00000844, 0.00001771, 0.00000317, 0.00002714, 0.00002324, -0.00000467, -0.00001153, -0.00004345, 0.00000539, -0.00000252, -0.00000732, 0.00000089, -0.00001126, 0.00000430, -0.00000819, 0.00000007, -0.00000170, -0.00000291, 0.00000226, -0.00001245, 0.00001149, 0.00000022, -0.00000883, -0.00000846, -0.00000472, 0.00002495, -0.00001060, -0.00000458, -0.00000158, 0.00000299, 0.00000340, -0.00000439, -0.00000564, 0.00000568, -0.00000034, 0.00000381, -0.00000274, 0.00000432, -0.00000625, -0.00000003, -0.00000862, 0.00001490, -0.00001160, 0.00001165, -0.00000896, 0.00000772, -0.00001716, 0.00000957, 0.00000195, -0.00000681, 0.00001507, -0.00000448, -0.00000216, -0.00001074, 0.00000290, 0.00000804, -0.00000157, 0.00000081, -0.00000332, -0.00000316, 0.00000037, -0.00000212, 0.00000886, -0.00000061, -0.00000816, -0.00000446, -0.00000007, 0.00001008, -0.00000595, 0.00000669, 0.00000372, -0.00000958, -0.00000070, 0.00000353, -0.00000180, -0.00000246, 0.00000402, 0.00000159, -0.00000666, -0.00000126, 0.00000196, 0.00000303, 0.00000213, -0.00000293, 0.00000276, -0.00000125, -0.00000557, 0.00000488, 0.00000348, 0.00000042, -0.00000614, -0.00000863, 0.00001362, 0.00000180, -0.00000676, -0.00001094, 0.00000214, 0.00000655, -0.00000733, -0.00000845, -0.00000474, -0.00000019, 0.00000293, -0.00000635, -0.00000375, -0.00000651, -0.00000118, -0.00001451, -0.00001300, -0.00000847, -0.00003104, -0.00005672, },
		{ -0.00005882, -0.00001930, 0.00007639, 0.00005384, 0.00001093, 0.00002698, 0.00002887, 0.00002118, 0.00009565, 0.00001520, -0.00012608, -0.00002327, 0.00001344, 0.00002471, -0.00000031, 0.00002094, -0.00001662, 0.00000553, 0.00001869, 0.00000197, 0.00002221, 0.00002672, -0.00000435, -0.00000442, -0.00004535, -0.00000014, 0.00000403, -0.00001202, 0.00000368, -0.00001217, 0.00000341, -0.00000581, -0.00000403, 0.00000215, -0.00000472, 0.00000372, -0.00001273, 0.00000810, 0.00000499, -0.00001045, -0.00000667, -0.00000980, 0.00002624, -0.00000629, -0.00000768, -0.00000063, 0.00000163, 0.00000408, -0.00000314, -0.00000722, 0.00000618, -0.00000091, 0.00000442, -0.00000360, 0.00000581, -0.00000760, 0.00000250, -0.00001263, 0.00001759, -0.00001241, 0.00001138, -0.00000829, 0.00000756, -0.00001475, 0.00000358, 0.00000758, -0.00001033, 0.00001508, -0.00000254, -0.00000195, -0.00001011, -0.00000009, 0.00000880, -0.00000009, -0.00000022, -0.00000185, -0.00000461, 0.00000113, -0.00000289, 0.00000739, 0.00000240, -0.00000847, -0.00000441, -0.00000250, 0.00001068, -0.00000405, 0.00000356, 0.00000656, -0.00000911, -0.00000220, 0.00000368, -0.00000152, -0.00000264, 0.00000312, 0.00000306, -0.00000623, -0.00000262, 0.00000242, 0.00000261, 0.00000225, -0.00000241, 0.00000194, 0.00000025, -0.00000616, 0.00000307, 0.00000537, 0.00000022, -0.00000470, -0.00001037, 0.00001153, 0.00000465, -0.00000581, -0.00001128, -0.00000052, 0.00000793, -0.00000637, -0.00000932, -0.00000404, -0.00000231, 0.00000441, -0.00000581, -0.00000454, -0.00000561, -0.00000279, -0.00001241, -0.00001359, -0.00000982, -0.00002394, },
		{ -0.00001786, -0.00005685, -0.00003821, 0.00007266, 0.00006516, 0.00001252, 0.00002797, 0.00003126, 0.00002066, 0.00009979, 0.00002006, -0.00012892, -0.00002383, 0.00000987, 0.00002429, -0.00000073, 0.00001989, -0.00001603, 0.00000313, 0.00001880, 0.00000213, 0.00001650, 0.00002996, -0.00000427, 0.00000138, -0.00004394, -0.00000803, 0.00001086, -0.00001554, 0.00000468, -0.00001120, 0.00000086, -0.00000224, -0.00000841, 0.00000521, -0.00000525, 0.00000375, -0.00001128, 0.00000339, 0.00000984, -0.00001164, -0.00000504, -0.00001381, 0.00002536, -0.00000071, -0.00001061, -0.00000013, 0.00000083, 0.00000395, -0.00000126, -0.00000896, 0.00000674, -0.00000171, 0.00000543, -0.00000502, 0.00000786, -0.00000952, 0.00000547, -0.00001635, 0.00001922, -0.00001196, 0.00000999, -0.00000661, 0.00000612, -0.00001088, -0.00000294, 0.00001237, -0.00001224, 0.00001362, -0.00000007, -0.00000191, -0.00000913, -0.00000276, 0.00000862, 0.00000196, -0.00000134, -0.00000053, -0.00000558, 0.00000123, -0.00000281, 0.00000520, 0.00000524, -0.00000823, -0.00000453, -0.00000438, 0.00001016, -0.00000137, 0.00000049, 0.00000845, -0.00000767, -0.00000393, 0.00000375, -0.00000129, -0.00000264, 0.00000214, 0.00000414, -0.00000527, -0.00000406, 0.00000267, 0.00000245, 0.00000208, -0.00000169, 0.00000114, 0.00000135, -0.00000594, 0.00000073, 0.00000704, 0.00000041, -0.00000369, -0.00001112, 0.00000875, 0.00000702, -0.00000436, -0.00001140, -0.00000307, 0.00000850, -0.00000477, -0.00001040, -0.00000324, -0.00000414, 0.00000506, -0.00000448, -0.00000575, -0.00000460, -0.00000414, -0.00001115, -0.00001297, -0.00001215, },
		{ -0.00001468, -0.00001369, -0.00005108, -0.00005641, 0.00006392, 0.00007746, 0.00001538, 0.00002841, 0.00003419, 0.00001991, 0.00010402, 0.00002518, -0.00013181, -0.00002454, 0.00000650, 0.00002372, -0.00000107, 0.00001893, -0.00001573, 0.00000142, 0.00001799, 0.00000341, 0.00001051, 0.00003244, -0.00000370, 0.00000510, -0.00003929, -0.00001728, 0.00001673, -0.00001700, 0.00000351, -0.00000839, -0.00000291, 0.00000179, -0.00001226, 0.00000679, -0.00000412, 0.00000224, -0.00000834, -0.00000207, 0.00001412, -0.00001195, -0.00000401, -0.00001634, 0.00002239, 0.00000554, -0.00001280, -0.00000046, 0.00000080, 0.00000295, 0.00000116, -0.00001080, 0.00000742, -0.00000288, 0.00000693, -0.00000699, 0.00001034, -0.00001172, 0.00000842, -0.00001924, 0.00001936, -0.00001006, 0.00000746, -0.00000400, 0.00000362, -0.00000607, -0.00000922, 0.00001570, -0.00001234, 0.00001092, 0.00000258, -0.00000179, -0.00000807, -0.00000487, 0.00000748, 0.00000432, -0.00000226, 0.00000039, -0.00000593, 0.00000066, -0.00000205, 0.00000268, 0.00000751, -0.00000727, -0.00000490, -0.00000566, 0.00000868, 0.00000167, -0.00000206, 0.00000921, -0.00000540, -0.00000566, 0.00000362, -0.00000104, -0.00000256, 0.00000123, 0.00000476, -0.00000386, -0.00000541, 0.00000259, 0.00000257, 0.00000168, -0.00000093, 0.00000051, 0.00000193, -0.00000500, -0.00000182, 0.00000816, 0.00000116, -0.00000316, -0.00001105, 0.00000569, 0.00000861, -0.00000240, -0.00001128, -0.00000541, 0.00000829, -0.00000274, -0.00001142, -0.00000262, -0.00000547, 0.00000486, -0.00000258, -0.00000703, -0.00000382, -0.00000494, -0.00001080, -0.00001144, },
		{ -0.00000951, -0.00001664, -0.00001190, -0.00004230, -0.00007223, 0.00004980, 0.00008976, 0.00002018, 0.00002809, 0.00003774, 0.00001895, 0.00010831, 0.00003059, -0.00013480, -0.00002530, 0.00000318, 0.00002316, -0.00000149, 0.00001825, -0.00001589, 0.00000045, 0.00001640, 0.00000541, 0.00000484, 0.00003363, -0.00000207, 0.00000642, -0.00003203, -0.00002656, 0.00002047, -0.00001582, 0.00000019, -0.00000421, -0.00000719, 0.00000545, -0.00001478, 0.00000643, -0.00000125, -0.00000057, -0.00000443, -0.00000753, 0.00001718, -0.00001100, -0.00000385, -0.00001730, 0.00001775, 0.00001169, -0.00001367, -0.00000185, 0.00000162, 0.00000113, 0.00000399, -0.00001267, 0.00000827, -0.00000446, 0.00000890, -0.00000937, 0.00001298, -0.00001381, 0.00001082, -0.00002079, 0.00001777, -0.00000675, 0.00000397, -0.00000070, 0.00000042, -0.00000094, -0.00001456, 0.00001716, -0.00001070, 0.00000738, 0.00000499, -0.00000135, -0.00000713, -0.00000626, 0.00000554, 0.00000668, -0.00000272, 0.00000076, -0.00000562, -0.00000047, -0.00000088, 0.00000022, 0.00000893, -0.00000555, -0.00000549, -0.00000639, 0.00000652, 0.00000462, -0.00000374, 0.00000879, -0.00000256, -0.00000712, 0.00000319, -0.00000073, -0.00000247, 0.00000047, 0.00000490, -0.00000216, -0.00000651, 0.00000214, 0.00000291, 0.00000117, -0.00000026, 0.00000019, 0.00000198, -0.00000354, -0.00000424, 0.00000851, 0.00000250, -0.00000303, -0.00001043, 0.00000278, 0.00000921, -0.00000004, -0.00001086, -0.00000747, 0.00000740, -0.00000057, -0.00001215, -0.00000233, -0.00000623, 0.00000394, -0.00000042, -0.00000804, -0.00000351, -0.00000509, -0.00001121, },
		{ -0.00001204, -0.00000774, -0.00001746, -0.00001253, -0.00003172, -0.00008410, 0.00003045, 0.00010069, 0.00002759, 0.00002692, 0.00004190, 0.00001787, 0.00011252, 0.00003645, -0.00013803, -0.00002596, -0.00000027, 0.00002276, -0.00000215, 0.00001794, -0.00001652, 0.00000010, 0.00001432, 0.00000758, 0.00000009, 0.00003314, 0.00000092, 0.00000547, -0.00002321, -0.00003446, 0.00002119, -0.00001189, -0.00000480, 0.00000055, -0.00001107, 0.00000789, -0.00001540, 0.00000397, 0.00000311, -0.00000418, -0.00000024, -0.00001221, 0.00001850, -0.00000864, -0.00000460, -0.00001682, 0.00001207, 0.00001694, -0.00001278, -0.00000440, 0.00000326, -0.00000137, 0.00000706, -0.00001453, 0.00000932, -0.00000647, 0.00001122, -0.00001190, 0.00001540, -0.00001531, 0.00001218, -0.00002064, 0.00001441, -0.00000224, -0.00000014, 0.00000291, -0.00000299, 0.00000386, -0.00001838, 0.00001660, -0.00000764, 0.00000354, 0.00000679, -0.00000043, -0.00000643, -0.00000694, 0.00000306, 0.00000868, -0.00000251, 0.00000054, -0.00000475, -0.00000194, 0.00000038, -0.00000183, 0.00000933, -0.00000317, -0.00000616, -0.00000672, 0.00000401, 0.00000708, -0.00000435, 0.00000735, 0.00000048, -0.00000806, 0.00000237, -0.00000034, -0.00000241, -0.00000009, 0.00000462, -0.00000034, -0.00000720, 0.00000130, 0.00000340, 0.00000067, 0.00000020, 0.00000023, 0.00000157, -0.00000180, -0.00000620, 0.00000795, 0.00000432, -0.00000308, -0.00000959, 0.00000036, 0.00000881, 0.00000250, -0.00001006, -0.00000927, 0.00000599, 0.00000150, -0.00001242, -0.00000244, -0.00000646, 0.00000253, 0.00000166, -0.00000853, -0.00000377, -0.00000469, },
		{ -0.00000397, -0.00001288, -0.00000660, -0.00001684, -0.00001518, -0.00002081, -0.00009077, 0.00000672, 0.00010862, 0.00003814, 0.00002502, 0.00004650, 0.00001690, 0.00011646, 0.00004296, -0.00014166, -0.00002637, -0.00000396, 0.00002259, -0.00000307, 0.00001799, -0.00001751, 0.00000012, 0.00001217, 0.00000940, -0.00000327, 0.00003078, 0.00000524, 0.00000283, -0.00001417, -0.00003970, 0.00001844, -0.00000563, -0.00001058, 0.00000492, -0.00001368, 0.00000847, -0.00001384, -0.00000038, 0.00000839, -0.00000789, 0.00000347, -0.00001544, 0.00001778, -0.00000494, -0.00000605, -0.00001532, 0.00000617, 0.00002050, -0.00000984, -0.00000803, 0.00000555, -0.00000434, 0.00001021, -0.00001630, 0.00001059, -0.00000881, 0.00001365, -0.00001420, 0.00001716, -0.00001580, 0.00001211, -0.00001866, 0.00000946, 0.00000304, -0.00000440, 0.00000639, -0.00000612, 0.00000777, -0.00002035, 0.00001414, -0.00000366, -0.00000005, 0.00000765, 0.00000103, -0.00000597, -0.00000701, 0.00000035, 0.00001002, -0.00000154, -0.00000019, -0.00000352, -0.00000346, 0.00000141, -0.00000319, 0.00000870, -0.00000035, -0.00000670, -0.00000685, 0.00000148, 0.00000874, -0.00000386, 0.00000516, 0.00000335, -0.00000827, 0.00000117, 0.00000010, -0.00000239, -0.00000045, 0.00000401, 0.00000141, -0.00000739, 0.00000012, 0.00000390, 0.00000031, 0.00000036, 0.00000062, 0.00000084, -0.00000006, -0.00000748, 0.00000649, 0.00000640, -0.00000305, -0.00000883, -0.00000135, 0.00000751, 0.00000493, -0.00000876, -0.00001081, 0.00000420, 0.00000327, -0.00001217, -0.00000289, -0.00000632, 0.00000089, 0.00000338, -0.00000838, -0.00000452, },
		{ -0.00000554, -0.00000327, -0.00001336, -0.00000641, -0.00001485, -0.00001907, -0.00001107, -0.00009159, -0.00001990, 0.00011177, 0.00005205, 0.00002278, 0.00005117, 0.00001637, 0.00011985, 0.00005032, -0.00014585, -0.00002643, -0.00000795, 0.00002264, -0.00000419, 0.00001825, -0.00001863, 0.00000016, 0.00001038, 0.00001042, -0.00000499, 0.00002667, 0.00001047, -0.00000057, -0.00000625, -0.00004141, 0.00001240, 0.00000205, -0.00001603, 0.00000789, -0.00001430, 0.00000687, -0.00001025, -0.00000605, 0.00001378, -0.00001091, 0.00000599, -0.00001675, 0.00001503, -0.00000027, -0.00000777, -0.00001340, 0.00000087, 0.00002176, -0.00000484, -0.00001248, 0.00000823, -0.00000755, 0.00001326, -0.00001793, 0.00001199, -0.00001126, 0.00001583, -0.00001585, 0.00001783, -0.00001491, 0.00001044, -0.00001495, 0.00000336, 0.00000851, -0.00000828, 0.00000928, -0.00000852, 0.00001040, -0.00002037, 0.00001018, 0.00000060, -0.00000284, 0.00000739, 0.00000295, -0.00000565, -0.00000665, -0.00000227, 0.00001049, 0.00000014, -0.00000124, -0.00000218, -0.00000476, 0.00000197, -0.00000376, 0.00000718, 0.00000258, -0.00000688, -0.00000697, -0.00000081, 0.00000946, -0.00000243, 0.00000262, 0.00000568, -0.00000765, -0.00000033, 0.00000052, -0.00000237, -0.00000065, 0.00000319, 0.00000294, -0.00000704, -0.00000126, 0.00000427, 0.00000017, 0.00000022, 0.00000128, -0.00000002, 0.00000142, -0.00000794, 0.00000429, 0.00000843, -0.00000267, -0.00000837, -0.00000229, 0.00000558, 0.00000693, -0.00000692, -0.00001206, 0.00000218, 0.00000460, -0.00001143, -0.00000355, -0.00000597, -0.00000071, 0.00000455, -0.00000761, },
		{ -0.00000641, -0.00000655, -0.00000290, -0.00001322, -0.00000721, -0.00001186, -0.00002320, -0.00000373, -0.00008661, -0.00004736, 0.00010856, 0.00006907, 0.00002093, 0.00005543, 0.00001670, 0.00012241, 0.00005872, -0.00015067, -0.00002612, -0.00001218, 0.00002276, -0.00000531, 0.00001848, -0.00001960, -0.00000011, 0.00000928, 0.00001037, -0.00000506, 0.00002121, 0.00001591, -0.00000369, -0.00000059, -0.00003925, 0.00000380, 0.00000986, -0.00001994, 0.00000863, -0.00001256, 0.00000317, -0.00000515, -0.00001219, 0.00001841, -0.00001252, 0.00000684, -0.00001601, 0.00001058, 0.00000477, -0.00000918, -0.00001174, -0.00000305, 0.00002037, 0.00000198, -0.00001732, 0.00001098, -0.00001074, 0.00001602, -0.00001931, 0.00001338, -0.00001354, 0.00001737, -0.00001644, 0.00001708, -0.00001248, 0.00000720, -0.00000984, -0.00000329, 0.00001354, -0.00001129, 0.00001121, -0.00000989, 0.00001156, -0.00001862, 0.00000527, 0.00000448, -0.00000446, 0.00000602, 0.00000514, -0.00000531, -0.00000609, -0.00000449, 0.00001000, 0.00000235, -0.00000235, -0.00000101, -0.00000561, 0.00000193, -0.00000357, 0.00000504, 0.00000528, -0.00000649, -0.00000720, -0.00000269, 0.00000924, -0.00000034, 0.00000014, 0.00000718, -0.00000622, -0.00000197, 0.00000082, -0.00000231, -0.00000074, 0.00000230, 0.00000412, -0.00000618, -0.00000270, 0.00000440, 0.00000030, -0.00000018, 0.00000208, -0.00000082, 0.00000246, -0.00000756, 0.00000163, 0.00001005, -0.00000175, -0.00000829, -0.00000256, 0.00000338, 0.00000822, -0.00000455, -0.00001296, 0.00000001, 0.00000546, -0.00001032, -0.00000427, -0.00000560, -0.00000211, 0.00000512, },
		{ 0.00000518, -0.00000502, -0.00000723, -0.00000306, -0.00001239, -0.00000880, -0.00000849, -0.00002659, 0.00000036, -0.00007661, -0.00007325, 0.00009777, 0.00008832, 0.00002048, 0.00005867, 0.00001828, 0.00012396, 0.00006822, -0.00015609, -0.00002558, -0.00001644, 0.00002273, -0.00000620, 0.00001843, -0.00002013, -0.00000097, 0.00000908, 0.00000922, -0.00000379, 0.00001508, 0.00002074, -0.00000556, 0.00000209, -0.00003355, -0.00000606, 0.00001636, -0.00002127, 0.00000671, -0.00000855, -0.00000211, 0.00000063, -0.00001782, 0.00002142, -0.00001222, 0.00000585, -0.00001342, 0.00000507, 0.00000938, -0.00000964, -0.00001095, -0.00000502, 0.00001630, 0.00001009, -0.00002202, 0.00001346, -0.00001366, 0.00001833, -0.00002034, 0.00001457, -0.00001529, 0.00001785, -0.00001563, 0.00001474, -0.00000855, 0.00000268, -0.00000388, -0.00000978, 0.00001752, -0.00001303, 0.00001197, -0.00001011, 0.00001130, -0.00001550, 0.00000010, 0.00000741, -0.00000469, 0.00000372, 0.00000730, -0.00000476, -0.00000553, -0.00000612, 0.00000859, 0.00000482, -0.00000322, -0.00000022, -0.00000586, 0.00000126, -0.00000277, 0.00000263, 0.00000740, -0.00000543, -0.00000759, -0.00000411, 0.00000821, 0.00000206, -0.00000190, 0.00000770, -0.00000412, -0.00000356, 0.00000089, -0.00000217, -0.00000077, 0.00000141, 0.00000490, -0.00000491, -0.00000403, 0.00000418, 0.00000070, -0.00000076, 0.00000288, -0.00000140, 0.00000296, -0.00000647, -0.00000117, 0.00001098, -0.00000018, -0.00000855, -0.00000239, 0.00000127, 0.00000861, -0.00000179, -0.00001340, -0.00000225, 0.00000585, -0.00000897, -0.00000488, -0.00000533, -0.00000321, },
		{ -0.00000405, 0.00000489, -0.00000373, -0.00000738, -0.00000382, -0.00001101, -0.00001077, -0.00000543, -0.00002843, 0.00000096, -0.00006306, -0.00009513, 0.00007889, 0.00010830, 0.00002266, 0.00006037, 0.00002137, 0.00012446, 0.00007874, -0.00016190, -0.00002507, -0.00002045, 0.00002224, -0.00000659, 0.00001788, -0.00002005, -0.00000253, 0.00000975, 0.00000720, -0.00000168, 0.00000905, 0.00002416, -0.00000546, 0.00000164, -0.00002524, -0.00001560, 0.00002017, -0.00001939, 0.00000216, -0.00000278, -0.00000812, 0.00000607, -0.00002199, 0.00002223, -0.00000984, 0.00000321, -0.00000955, -0.00000063, 0.00001274, -0.00000855, -0.00001152, -0.00000476, 0.00000985, 0.00001880, -0.00002600, 0.00001534, -0.00001610, 0.00002002, -0.00002084, 0.00001528, -0.00001614, 0.00001694, -0.00001324, 0.00001086, -0.00000339, -0.00000265, 0.00000224, -0.00001541, 0.00001996, -0.00001329, 0.00001152, -0.00000924, 0.00000985, -0.00001153, -0.00000467, 0.00000897, -0.00000356, 0.00000082, 0.00000910, -0.00000381, -0.00000513, -0.00000707, 0.00000646, 0.00000723, -0.00000360, 0.00000003, -0.00000551, 0.00000010, -0.00000162, 0.00000032, 0.00000869, -0.00000369, -0.00000809, -0.00000511, 0.00000660, 0.00000439, -0.00000321, 0.00000724, -0.00000159, -0.00000485, 0.00000064, -0.00000193, -0.00000079, 0.00000061, 0.00000526, -0.00000337, -0.00000510, 0.00000358, 0.00000128, -0.00000140, 0.00000353, -0.00000164, 0.00000293, -0.00000489, -0.00000374, 0.00001100, 0.00000197, -0.00000897, -0.00000207, -0.00000041, 0.00000808, 0.00000113, -0.00001326, -0.00000458, 0.00000582, -0.00000755, -0.00000528, -0.00000519, },
		{ -0.00000516, -0.00000471, 0.00000446, -0.00000278, -0.00000690, -0.00000506, -0.00000939, -0.00001261, -0.00000333, -0.00002826, -0.00000155, -0.00004788, -0.00011083, 0.00005231, 0.00012688, 0.00002872, 0.00006017, 0.00002602, 0.00012409, 0.00008999, -0.00016776, -0.00002493, -0.00002387, 0.00002102, -0.00000631, 0.00001671, -0.00001932, -0.00000473, 0.00001108, 0.00000470, 0.00000062, 0.00000388, 0.00002557, -0.00000312, -0.00000149, -0.00001574, -0.00002316, 0.00002029, -0.00001418, -0.00000446, 0.00000381, -0.00001379, 0.00001017, -0.00002398, 0.00002057, -0.00000561, -0.00000054, -0.00000520, -0.00000554, 0.00001409, -0.00000553, -0.00001369, -0.00000231, 0.00000161, 0.00002725, -0.00002871, 0.00001640, -0.00001787, 0.00002094, -0.00002065, 0.00001525, -0.00001577, 0.00001444, -0.00000930, 0.00000569, 0.00000253, -0.00000814, 0.00000781, -0.00001957, 0.00002056, -0.00001208, 0.00001002, -0.00000751, 0.00000762, -0.00000733, -0.00000847, 0.00000896, -0.00000130, -0.00000222, 0.00001021, -0.00000237, -0.00000494, -0.00000737, 0.00000386, 0.00000921, -0.00000329, -0.00000028, -0.00000465, -0.00000135, -0.00000042, -0.00000159, 0.00000903, -0.00000138, -0.00000856, -0.00000581, 0.00000467, 0.00000633, -0.00000363, 0.00000594, 0.00000106, -0.00000564, 0.00000004, -0.00000160, -0.00000079, -0.00000008, 0.00000526, -0.00000172, -0.00000577, 0.00000261, 0.00000196, -0.00000197, 0.00000393, -0.00000151, 0.00000246, -0.00000307, -0.00000579, 0.00001004, 0.00000451, -0.00000927, -0.00000188, -0.00000146, 0.00000673, 0.00000391, -0.00001242, -0.00000691, 0.00000540, -0.00000615, -0.00000544, },
		{ -0.00000539, -0.00000514, -0.00000532, 0.00000406, -0.00000230, -0.00000586, -0.00000651, -0.00000792, -0.00001385, -0.00000258, -0.00002606, -0.00000627, -0.00003319, -0.00011882, 0.00001938, 0.00014152, 0.00003968, 0.00005810, 0.00003196, 0.00012330, 0.00010153, -0.00017326, -0.00002554, -0.00002642, 0.00001890, -0.00000529, 0.00001498, -0.00001808, -0.00000735, 0.00001272, 0.00000223, 0.00000248, 0.00000016, 0.00002473, 0.00000124, -0.00000635, -0.00000666, -0.00002733, 0.00001630, -0.00000617, -0.00001212, 0.00001004, -0.00001804, 0.00001211, -0.00002343, 0.00001663, -0.00000014, -0.00000457, -0.00000135, -0.00000876, 0.00001289, -0.00000046, -0.00001737, 0.00000198, -0.00000759, 0.00003457, -0.00002973, 0.00001646, -0.00001886, 0.00002095, -0.00001960, 0.00001422, -0.00001395, 0.00001032, -0.00000404, -0.00000030, 0.00000859, -0.00001313, 0.00001220, -0.00002185, 0.00001927, -0.00000962, 0.00000778, -0.00000530, 0.00000505, -0.00000346, -0.00001091, 0.00000743, 0.00000168, -0.00000494, 0.00001040, -0.00000043, -0.00000494, -0.00000716, 0.00000112, 0.00001050, -0.00000223, -0.00000105, -0.00000347, -0.00000283, 0.00000054, -0.00000285, 0.00000842, 0.00000127, -0.00000882, -0.00000637, 0.00000269, 0.00000763, -0.00000317, 0.00000405, 0.00000350, -0.00000579, -0.00000088, -0.00000124, -0.00000076, -0.00000065, 0.00000495, -0.00000009, -0.00000597, 0.00000136, 0.00000259, -0.00000235, 0.00000401, -0.00000104, 0.00000170, -0.00000131, -0.00000708, 0.00000818, 0.00000714, -0.00000918, -0.00000206, -0.00000181, 0.00000481, 0.00000621, -0.00001082, -0.00000915, 0.00000461, -0.00000485, },
		{ -0.00000366, -0.00000522, -0.00000502, -0.00000598, 0.00000382, -0.00000230, -0.00000446, -0.00000781, -0.00000700, -0.00001414, -0.00000332, -0.00002225, -0.00001192, -0.00002097, -0.00011847, -0.00001757, 0.00014956, 0.00005606, 0.00005468, 0.00003857, 0.00012273, 0.00011281, -0.00017795, -0.00002722, -0.00002793, 0.00001585, -0.00000367, 0.00001289, -0.00001660, -0.00001002, 0.00001422, 0.00000026, 0.00000344, -0.00000182, 0.00002178, 0.00000696, -0.00001165, 0.00000039, -0.00002718, 0.00000843, 0.00000354, -0.00001946, 0.00001473, -0.00002001, 0.00001147, -0.00002043, 0.00001104, 0.00000565, -0.00000790, 0.00000109, -0.00000957, 0.00000891, 0.00000647, -0.00002222, 0.00000747, -0.00001676, 0.00003995, -0.00002878, 0.00001551, -0.00001899, 0.00001994, -0.00001755, 0.00001204, -0.00001063, 0.00000479, 0.00000208, -0.00000649, 0.00001410, -0.00001697, 0.00001492, -0.00002207, 0.00001630, -0.00000631, 0.00000523, -0.00000299, 0.00000260, -0.00000035, -0.00001182, 0.00000464, 0.00000487, -0.00000691, 0.00000956, 0.00000189, -0.00000499, -0.00000663, -0.00000144, 0.00001089, -0.00000049, -0.00000211, -0.00000221, -0.00000409, 0.00000105, -0.00000339, 0.00000703, 0.00000395, -0.00000867, -0.00000691, 0.00000085, 0.00000818, -0.00000196, 0.00000191, 0.00000543, -0.00000523, -0.00000204, -0.00000092, -0.00000068, -0.00000113, 0.00000443, 0.00000137, -0.00000569, -0.00000005, 0.00000304, -0.00000247, 0.00000379, -0.00000032, 0.00000085, 0.00000018, -0.00000752, 0.00000562, 0.00000952, -0.00000846, -0.00000269, -0.00000156, 0.00000268, 0.00000773, -0.00000850, -0.00001115, 0.00000344, },
		{ 0.00000185, -0.00000256, -0.00000501, -0.00000471, -0.00000675, 0.00000375, -0.00000265, -0.00000303, -0.00000861, -0.00000690, -0.00001337, -0.00000529, -0.00001761, -0.00001707, -0.00001274, -0.00011014, -0.00005550, 0.00014857, 0.00007764, 0.00005098, 0.00004497, 0.00012313, 0.00012334, -0.00018152, -0.00003010, -0.00002842, 0.00001208, -0.00000175, 0.00001078, -0.00001524, -0.00001239, 0.00001521, -0.00000085, 0.00000326, -0.00000213, 0.00001728, 0.00001303, -0.00001600, 0.00000416, -0.00002249, -0.00000234, 0.00001344, -0.00002508, 0.00001689, -0.00001921, 0.00000831, -0.00001556, 0.00000476, 0.00001067, -0.00000957, 0.00000136, -0.00000758, 0.00000230, 0.00001472, -0.00002761, 0.00001336, -0.00002494, 0.00004277, -0.00002577, 0.00001361, -0.00001822, 0.00001785, -0.00001442, 0.00000867, -0.00000595, -0.00000175, 0.00000844, -0.00001219, 0.00001844, -0.00001918, 0.00001573, -0.00002033, 0.00001207, -0.00000269, 0.00000283, -0.00000097, 0.00000064, 0.00000172, -0.00001126, 0.00000105, 0.00000772, -0.00000781, 0.00000775, 0.00000437, -0.00000494, -0.00000600, -0.00000355, 0.00001033, 0.00000176, -0.00000317, -0.00000113, -0.00000490, 0.00000099, -0.00000323, 0.00000510, 0.00000634, -0.00000797, -0.00000751, -0.00000071, 0.00000799, -0.00000023, -0.00000014, 0.00000665, -0.00000401, -0.00000329, -0.00000074, -0.00000050, -0.00000153, 0.00000377, 0.00000257, -0.00000497, -0.00000148, 0.00000320, -0.00000230, 0.00000332, 0.00000050, 0.00000009, 0.00000121, -0.00000714, 0.00000268, 0.00001128, -0.00000695, -0.00000375, -0.00000094, 0.00000072, 0.00000829, -0.00000561, -0.00001271, },
		{ -0.00001360, -0.00000017, -0.00000154, -0.00000483, -0.00000420, -0.00000757, 0.00000377, -0.00000314, -0.00000188, -0.00000862, -0.00000771, -0.00001172, -0.00000799, -0.00001310, -0.00002046, -0.00000932, -0.00009524, -0.00009095, 0.00013679, 0.00010328, 0.00004858, 0.00005011, 0.00012524, 0.00013275, -0.00018388, -0.00003410, -0.00002814, 0.00000794, 0.00000005, 0.00000906, -0.00001437, -0.00001413, 0.00001544, -0.00000099, 0.00000200, -0.00000116, 0.00001208, 0.00001832, -0.00001820, 0.00000397, -0.00001378, -0.00001448, 0.00002181, -0.00002774, 0.00001593, -0.00001568, 0.00000319, -0.00000981, -0.00000100, 0.00001381, -0.00000880, -0.00000098, -0.00000280, -0.00000637, 0.00002349, -0.00003274, 0.00001877, -0.00003124, 0.00004264, -0.00002085, 0.00001094, -0.00001657, 0.00001465, -0.00001024, 0.00000420, -0.00000022, -0.00000871, 0.00001435, -0.00001673, 0.00002109, -0.00001951, 0.00001466, -0.00001702, 0.00000718, 0.00000067, 0.00000099, 0.00000047, -0.00000062, 0.00000266, -0.00000950, -0.00000282, 0.00000976, -0.00000751, 0.00000515, 0.00000671, -0.00000457, -0.00000545, -0.00000505, 0.00000888, 0.00000423, -0.00000397, -0.00000043, -0.00000516, 0.00000036, -0.00000255, 0.00000296, 0.00000816, -0.00000663, -0.00000818, -0.00000197, 0.00000717, 0.00000171, -0.00000180, 0.00000704, -0.00000227, -0.00000443, -0.00000078, -0.00000022, -0.00000185, 0.00000303, 0.00000347, -0.00000390, -0.00000277, 0.00000300, -0.00000189, 0.00000271, 0.00000129, -0.00000043, 0.00000172, -0.00000609, -0.00000028, 0.00001214, -0.00000468, -0.00000503, -0.00000027, -0.00000074, 0.00000783, -0.00000243, },
		{ 0.00000068, -0.00001360, -0.00000261, -0.00000062, -0.00000472, -0.00000352, -0.00000836, 0.00000372, -0.00000355, -0.00000129, -0.00000774, -0.00000929, -0.00000957, -0.00001071, -0.00000968, -0.00002118, -0.00001072, -0.00007604, -0.00012046, 0.00011346, 0.00013082, 0.00004940, 0.00005302, 0.00012954, 0.00014099, -0.00018520, -0.00003889, -0.00002755, 0.00000394, 0.00000126, 0.00000810, -0.00001422, -0.00001511, 0.00001489, -0.00000029, 0.00000002, 0.00000041, 0.00000714, 0.00002179, -0.00001745, -0.00000017, -0.00000229, -0.00002611, 0.00002707, -0.00002667, 0.00001186, -0.00001001, -0.00000288, -0.00000442, -0.00000503, 0.00001419, -0.00000515, -0.00000594, 0.00000431, -0.00001624, 0.00003179, -0.00003680, 0.00002289, -0.00003501, 0.00003947, -0.00001436, 0.00000773, -0.00001410, 0.00001040, -0.00000514, -0.00000108, 0.00000605, -0.00001541, 0.00001911, -0.00001959, 0.00002178, -0.00001796, 0.00001201, -0.00001267, 0.00000229, 0.00000325, -0.00000002, 0.00000118, -0.00000110, 0.00000259, -0.00000693, -0.00000640, 0.00001064, -0.00000605, 0.00000211, 0.00000861, -0.00000376, -0.00000510, -0.00000587, 0.00000674, 0.00000658, -0.00000426, -0.00000024, -0.00000487, -0.00000070, -0.00000157, 0.00000094, 0.00000918, -0.00000469, -0.00000884, -0.00000297, 0.00000589, 0.00000358, -0.00000286, 0.00000665, -0.00000024, -0.00000529, -0.00000110, 0.00000016, -0.00000212, 0.00000227, 0.00000404, -0.00000258, -0.00000378, 0.00000242, -0.00000132, 0.00000209, 0.00000190, -0.00000062, 0.00000173, -0.00000462, -0.00000289, 0.00001194, -0.00000177, -0.00000626, 0.00000015, -0.00000147, 0.00000647, },
		{ 0.00000449, 0.00000329, -0.00001257, -0.00000539, 0.00000012, -0.00000463, -0.00000279, -0.00000894, 0.00000346, -0.00000365, -0.00000139, -0.00000606, -0.00001129, -0.00000750, -0.00001272, -0.00000810, -0.00001888, -0.00001609, -0.00005534, -0.00014110, 0.00007913, 0.00015723, 0.00005543, 0.00005302, 0.00013616, 0.00014839, -0.00018599, -0.00004389, -0.00002723, 0.00000061, 0.00000148, 0.00000809, -0.00001484, -0.00001542, 0.00001378, 0.00000086, -0.00000212, 0.00000182, 0.00000339, 0.00002271, -0.00001352, -0.00000751, 0.00001023, -0.00003529, 0.00002806, -0.00002173, 0.00000532, -0.00000329, -0.00000859, -0.00000068, -0.00000631, 0.00001129, 0.00000135, -0.00001310, 0.00001294, -0.00002619, 0.00003860, -0.00003901, 0.00002509, -0.00003587, 0.00003347, -0.00000679, 0.00000426, -0.00001089, 0.00000525, 0.00000062, -0.00000677, 0.00001228, -0.00002117, 0.00002216, -0.00002045, 0.00002050, -0.00001481, 0.00000828, -0.00000796, -0.00000196, 0.00000468, -0.00000006, 0.00000118, -0.00000089, 0.00000176, -0.00000404, -0.00000924, 0.00001023, -0.00000367, -0.00000097, 0.00000977, -0.00000242, -0.00000500, -0.00000607, 0.00000419, 0.00000848, -0.00000387, -0.00000057, -0.00000411, -0.00000201, -0.00000058, -0.00000068, 0.00000933, -0.00000230, -0.00000935, -0.00000380, 0.00000438, 0.00000513, -0.00000322, 0.00000561, 0.00000183, -0.00000570, -0.00000170, 0.00000060, -0.00000229, 0.00000151, 0.00000431, -0.00000115, -0.00000443, 0.00000151, -0.00000070, 0.00000157, 0.00000225, -0.00000046, 0.00000135, -0.00000298, -0.00000487, 0.00001066, 0.00000146, -0.00000710, 0.00000003, -0.00000139, },
		{ -0.00000062, 0.00000229, 0.00000503, -0.00001049, -0.00000829, 0.00000056, -0.00000446, -0.00000217, -0.00000916, 0.00000285, -0.00000337, -0.00000215, -0.00000385, -0.00001322, -0.00000608, -0.00001342, -0.00000871, -0.00001383, -0.00002395, -0.00003608, -0.00015085, 0.00003572, 0.00017888, 0.00006834, 0.00005000, 0.00014470, 0.00015567, -0.00018704, -0.00004842, -0.00002775, -0.00000164, 0.00000050, 0.00000902, -0.00001605, -0.00001538, 0.00001253, 0.00000193, -0.00000378, 0.00000235, 0.00000147, 0.00002080, -0.00000681, -0.00001673, 0.00002181, -0.00004042, 0.00002428, -0.00001352, -0.00000248, 0.00000306, -0.00001257, 0.00000029, -0.00000417, 0.00000507, 0.00001023, -0.00002161, 0.00002199, -0.00003502, 0.00004299, -0.00003881, 0.00002498, -0.00003377, 0.00002512, 0.00000125, 0.00000082, -0.00000707, -0.00000057, 0.00000664, -0.00001234, 0.00001782, -0.00002534, 0.00002310, -0.00001926, 0.00001749, -0.00001054, 0.00000410, -0.00000356, -0.00000506, 0.00000477, 0.00000080, 0.00000061, -0.00000020, 0.00000050, -0.00000129, -0.00001100, 0.00000857, -0.00000072, -0.00000365, 0.00001001, -0.00000058, -0.00000508, -0.00000580, 0.00000156, 0.00000967, -0.00000275, -0.00000132, -0.00000309, -0.00000332, 0.00000016, -0.00000172, 0.00000864, 0.00000031, -0.00000954, -0.00000456, 0.00000281, 0.00000618, -0.00000289, 0.00000414, 0.00000367, -0.00000559, -0.00000253, 0.00000104, -0.00000235, 0.00000076, 0.00000432, 0.00000027, -0.00000464, 0.00000035, -0.00000016, 0.00000126, 0.00000230, -0.00000000, 0.00000076, -0.00000147, -0.00000602, 0.00000845, 0.00000465, -0.00000723, -0.00000075, },
		{ -0.00000217, 0.00000058, 0.00000033, 0.00000562, -0.00000747, -0.00001105, 0.00000051, -0.00000408, -0.00000181, -0.00000891, 0.00000186, -0.00000272, -0.00000340, -0.00000156, -0.00001452, -0.00000582, -0.00001251, -0.00001143, -0.00000688, -0.00003237, -0.00002091, -0.00014900, -0.00001350, 0.00019191, 0.00008911, 0.00004462, 0.00015422, 0.00016393, -0.00018924, -0.00005182, -0.00002955, -0.00000259, -0.00000160, 0.00001059, -0.00001740, -0.00001549, 0.00001169, 0.00000235, -0.00000441, 0.00000149, 0.00000166, 0.00001633, 0.00000172, -0.00002616, 0.00003060, -0.00004051, 0.00001607, -0.00000333, -0.00000992, 0.00000752, -0.00001360, -0.00000225, 0.00000151, -0.00000399, 0.00002055, -0.00003036, 0.00003023, -0.00004162, 0.00004426, -0.00003594, 0.00002249, -0.00002896, 0.00001510, 0.00000912, -0.00000234, -0.00000279, -0.00000674, 0.00001247, -0.00001725, 0.00002207, -0.00002750, 0.00002182, -0.00001624, 0.00001324, -0.00000580, 0.00000015, -0.00000005, -0.00000670, 0.00000354, 0.00000231, -0.00000024, 0.00000070, -0.00000083, 0.00000095, -0.00001154, 0.00000589, 0.00000232, -0.00000557, 0.00000925, 0.00000163, -0.00000520, -0.00000526, -0.00000087, 0.00000998, -0.00000099, -0.00000228, -0.00000202, -0.00000439, 0.00000049, -0.00000210, 0.00000728, 0.00000287, -0.00000926, -0.00000534, 0.00000132, 0.00000666, -0.00000201, 0.00000251, 0.00000506, -0.00000493, -0.00000351, 0.00000140, -0.00000227, 0.00000003, 0.00000411, 0.00000159, -0.00000443, -0.00000093, 0.00000016, 0.00000121, 0.00000209, 0.00000063, 0.00000015, -0.00000029, -0.00000628, 0.00000558, 0.00000739, -0.00000642, },
		{ -0.00000461, -0.00000406, 0.00000183, -0.00000098, 0.00000494, -0.00000382, -0.00001330, -0.00000018, -0.00000341, -0.00000180, -0.00000819, 0.00000059, -0.00000188, -0.00000479, 0.00000036, -0.00001475, -0.00000698, -0.00001003, -0.00001570, 0.00000071, -0.00003938, -0.00001177, -0.00013624, -0.00006426, 0.00019282, 0.00011761, 0.00003839, 0.00016331, 0.00017444, -0.00019347, -0.00005360, -0.00003280, -0.00000240, -0.00000444, 0.00001224, -0.00001829, -0.00001634, 0.00001177, 0.00000166, -0.00000367, -0.00000090, 0.00000377, 0.00001003, 0.00001072, -0.00003408, 0.00003522, -0.00003538, 0.00000459, 0.00000701, -0.00001524, 0.00000873, -0.00001088, -0.00000848, 0.00001028, -0.00001490, 0.00003108, -0.00003810, 0.00003651, -0.00004510, 0.00004207, -0.00003043, 0.00001784, -0.00002198, 0.00000425, 0.00001621, -0.00000501, 0.00000176, -0.00001286, 0.00001759, -0.00002095, 0.00002458, -0.00002743, 0.00001849, -0.00001187, 0.00000840, -0.00000125, -0.00000299, 0.00000218, -0.00000680, 0.00000126, 0.00000412, -0.00000107, 0.00000153, -0.00000193, 0.00000243, -0.00001089, 0.00000258, 0.00000500, -0.00000648, 0.00000757, 0.00000398, -0.00000520, -0.00000466, -0.00000283, 0.00000937, 0.00000121, -0.00000321, -0.00000115, -0.00000505, 0.00000030, -0.00000189, 0.00000548, 0.00000507, -0.00000841, -0.00000615, -0.00000002, 0.00000659, -0.00000076, 0.00000099, 0.00000587, -0.00000380, -0.00000448, 0.00000158, -0.00000203, -0.00000066, 0.00000372, 0.00000273, -0.00000383, -0.00000219, 0.00000018, 0.00000142, 0.00000169, 0.00000130, -0.00000031, 0.00000041, -0.00000574, 0.00000242, 0.00000930, },
		{ 0.00001011, -0.00000188, -0.00000608, 0.00000273, -0.00000133, 0.00000308, 0.00000003, -0.00001467, -0.00000166, -0.00000240, -0.00000217, -0.00000707, -0.00000077, -0.00000110, -0.00000597, 0.00000148, -0.00001365, -0.00000951, -0.00000640, -0.00002062, 0.00000754, -0.00004331, -0.00000955, -0.00011465, -0.00011173, 0.00017901, 0.00015241, 0.00003361, 0.00017030, 0.00018840, -0.00020034, -0.00005367, -0.00003727, -0.00000153, -0.00000735, 0.00001323, -0.00001806, -0.00001844, 0.00001316, -0.00000034, -0.00000155, -0.00000458, 0.00000720, 0.00000301, 0.00001876, -0.00003905, 0.00003499, -0.00002580, -0.00000827, 0.00001539, -0.00001682, 0.00000580, -0.00000418, -0.00001797, 0.00002114, -0.00002630, 0.00004042, -0.00004363, 0.00003987, -0.00004494, 0.00003643, -0.00002268, 0.00001152, -0.00001354, -0.00000658, 0.00002200, -0.00000705, 0.00000637, -0.00001850, 0.00002152, -0.00002304, 0.00002510, -0.00002519, 0.00001355, -0.00000681, 0.00000367, 0.00000248, -0.00000490, 0.00000297, -0.00000552, -0.00000168, 0.00000580, -0.00000159, 0.00000206, -0.00000261, 0.00000306, -0.00000928, -0.00000090, 0.00000691, -0.00000627, 0.00000517, 0.00000618, -0.00000489, -0.00000417, -0.00000418, 0.00000792, 0.00000358, -0.00000384, -0.00000064, -0.00000522, -0.00000037, -0.00000125, 0.00000356, 0.00000668, -0.00000696, -0.00000695, -0.00000121, 0.00000608, 0.00000063, -0.00000020, 0.00000606, -0.00000234, -0.00000530, 0.00000155, -0.00000165, -0.00000129, 0.00000317, 0.00000365, -0.00000292, -0.00000330, -0.00000015, 0.00000182, 0.00000121, 0.00000187, -0.00000049, 0.00000060, -0.00000458, -0.00000064, },
		{ -0.00000322, 0.00000971, 0.00000151, -0.00000782, 0.00000291, -0.00000058, 0.00000034, 0.00000356, -0.00001485, -0.00000394, -0.00000112, -0.00000279, -0.00000579, -0.00000194, -0.00000070, -0.00000660, 0.00000152, -0.00001120, -0.00001304, -0.00000235, -0.00002511, 0.00001228, -0.00004307, -0.00001404, -0.00008746, -0.00015119, 0.00014930, 0.00019060, 0.00003315, 0.00017358, 0.00020673, -0.00021001, -0.00005240, -0.00004233, -0.00000078, -0.00000948, 0.00001280, -0.00001616, -0.00002211, 0.00001593, -0.00000353, 0.00000159, -0.00000899, 0.00001107, -0.00000352, 0.00002456, -0.00004021, 0.00003015, -0.00001336, -0.00002020, 0.00001982, -0.00001352, -0.00000154, 0.00000603, -0.00002971, 0.00003270, -0.00003666, 0.00004723, -0.00004597, 0.00003974, -0.00004104, 0.00002779, -0.00001336, 0.00000422, -0.00000444, -0.00001658, 0.00002611, -0.00000839, 0.00001081, -0.00002323, 0.00002384, -0.00002325, 0.00002362, -0.00002113, 0.00000763, -0.00000181, -0.00000028, 0.00000494, -0.00000538, 0.00000240, -0.00000324, -0.00000476, 0.00000696, -0.00000157, 0.00000219, -0.00000278, 0.00000292, -0.00000703, -0.00000409, 0.00000778, -0.00000501, 0.00000239, 0.00000794, -0.00000413, -0.00000390, -0.00000488, 0.00000584, 0.00000580, -0.00000395, -0.00000060, -0.00000489, -0.00000141, -0.00000041, 0.00000180, 0.00000753, -0.00000499, -0.00000763, -0.00000228, 0.00000525, 0.00000193, -0.00000094, 0.00000568, -0.00000075, -0.00000584, 0.00000127, -0.00000116, -0.00000181, 0.00000250, 0.00000433, -0.00000180, -0.00000415, -0.00000081, 0.00000232, 0.00000079, 0.00000223, -0.00000034, 0.00000033, -0.00000307, },
		{ -0.00000148, -0.00000506, 0.00000815, 0.00000516, -0.00000886, 0.00000212, 0.00000117, -0.00000283, 0.00000624, -0.00001364, -0.00000691, 0.00000025, -0.00000347, -0.00000463, -0.00000262, -0.00000091, -0.00000647, 0.00000045, -0.00000771, -0.00001694, 0.00000125, -0.00002814, 0.00001403, -0.00003838, -0.00002388, -0.00005857, -0.00017859, 0.00010431, 0.00022801, 0.00004001, 0.00017195, 0.00022974, -0.00022200, -0.00005065, -0.00004697, -0.00000113, -0.00000998, 0.00001035, -0.00001229, -0.00002733, 0.00001984, -0.00000747, 0.00000513, -0.00001330, 0.00001440, -0.00000851, 0.00002729, -0.00003746, 0.00002180, -0.00000028, -0.00002883, 0.00001873, -0.00000494, -0.00001281, 0.00001867, -0.00004222, 0.00004333, -0.00004450, 0.00005043, -0.00004453, 0.00003595, -0.00003372, 0.00001691, -0.00000331, -0.00000330, 0.00000449, -0.00002507, 0.00002833, -0.00000904, 0.00001485, -0.00002664, 0.00002430, -0.00002157, 0.00002038, -0.00001580, 0.00000150, 0.00000242, -0.00000296, 0.00000590, -0.00000448, 0.00000078, -0.00000045, -0.00000748, 0.00000729, -0.00000092, 0.00000190, -0.00000249, 0.00000219, -0.00000454, -0.00000662, 0.00000750, -0.00000294, -0.00000039, 0.00000899, -0.00000286, -0.00000387, -0.00000501, 0.00000340, 0.00000755, -0.00000341, -0.00000102, -0.00000418, -0.00000262, 0.00000040, 0.00000044, 0.00000757, -0.00000266, -0.00000806, -0.00000331, 0.00000424, 0.00000299, -0.00000118, 0.00000487, 0.00000077, -0.00000599, 0.00000077, -0.00000061, -0.00000219, 0.00000173, 0.00000477, -0.00000055, -0.00000466, -0.00000176, 0.00000280, 0.00000053, 0.00000230, 0.00000014, -0.00000025, },
		{ -0.00000098, -0.00000010, -0.00000601, 0.00000565, 0.00000861, -0.00000885, 0.00000031, 0.00000365, -0.00000586, 0.00000760, -0.00001103, -0.00001027, 0.00000141, -0.00000393, -0.00000388, -0.00000256, -0.00000188, -0.00000556, -0.00000154, -0.00000371, -0.00002044, 0.00000354, -0.00002894, 0.00001241, -0.00002982, -0.00003688, -0.00003201, -0.00019122, 0.00004661, 0.00025952, 0.00005685, 0.00016505, 0.00025695, -0.00023525, -0.00004974, -0.00005002, -0.00000357, -0.00000815, 0.00000555, -0.00000654, -0.00003373, 0.00002429, -0.00001141, 0.00000824, -0.00001667, 0.00001635, -0.00001124, 0.00002676, -0.00003149, 0.00001175, 0.00001095, -0.00003211, 0.00001133, 0.00000850, -0.00002678, 0.00003213, -0.00005375, 0.00005140, -0.00004857, 0.00004937, -0.00003921, 0.00002884, -0.00002374, 0.00000481, 0.00000652, -0.00001029, 0.00001257, -0.00003156, 0.00002861, -0.00000905, 0.00001827, -0.00002846, 0.00002280, -0.00001818, 0.00001587, -0.00000992, -0.00000406, 0.00000529, -0.00000408, 0.00000537, -0.00000248, -0.00000144, 0.00000232, -0.00000940, 0.00000662, 0.00000030, 0.00000130, -0.00000188, 0.00000112, -0.00000218, -0.00000820, 0.00000615, -0.00000039, -0.00000278, 0.00000917, -0.00000112, -0.00000401, -0.00000470, 0.00000092, 0.00000859, -0.00000220, -0.00000177, -0.00000327, -0.00000379, 0.00000095, -0.00000038, 0.00000687, -0.00000022, -0.00000808, -0.00000436, 0.00000318, 0.00000371, -0.00000096, 0.00000383, 0.00000203, -0.00000573, 0.00000011, -0.00000009, -0.00000238, 0.00000088, 0.00000496, 0.00000073, -0.00000481, -0.00000288, 0.00000312, 0.00000051, 0.00000209, 0.00000087, },
		{ 0.00000171, -0.00000165, 0.00000090, -0.00000605, 0.00000255, 0.00001138, -0.00000759, -0.00000236, 0.00000640, -0.00000817, 0.00000738, -0.00000724, -0.00001360, 0.00000202, -0.00000389, -0.00000379, -0.00000163, -0.00000357, -0.00000406, -0.00000402, 0.00000013, -0.00002275, 0.00000388, -0.00002719, 0.00000772, -0.00001871, -0.00005033, -0.00001133, -0.00018808, -0.00001940, 0.00027966, 0.00008536, 0.00015362, 0.00028707, -0.00024809, -0.00005119, -0.00005031, -0.00000882, -0.00000369, -0.00000146, 0.00000056, -0.00004054, 0.00002843, -0.00001452, 0.00001013, -0.00001844, 0.00001644, -0.00001157, 0.00002349, -0.00002374, 0.00000227, 0.00001798, -0.00002868, -0.00000223, 0.00002546, -0.00004163, 0.00004451, -0.00006258, 0.00005557, -0.00004810, 0.00004390, -0.00003040, 0.00001913, -0.00001208, -0.00000741, 0.00001524, -0.00001612, 0.00001923, -0.00003578, 0.00002710, -0.00000855, 0.00002089, -0.00002850, 0.00001949, -0.00001351, 0.00001070, -0.00000424, -0.00000839, 0.00000645, -0.00000362, 0.00000363, 0.00000016, -0.00000373, 0.00000457, -0.00001024, 0.00000499, 0.00000191, 0.00000058, -0.00000113, -0.00000000, -0.00000026, -0.00000875, 0.00000394, 0.00000221, -0.00000447, 0.00000844, 0.00000095, -0.00000420, -0.00000416, -0.00000133, 0.00000879, -0.00000041, -0.00000267, -0.00000238, -0.00000471, 0.00000110, -0.00000064, 0.00000562, 0.00000207, -0.00000757, -0.00000544, 0.00000212, 0.00000405, -0.00000040, 0.00000276, 0.00000290, -0.00000508, -0.00000059, 0.00000032, -0.00000236, -0.00000001, 0.00000491, 0.00000198, -0.00000460, -0.00000407, 0.00000318, 0.00000075, 0.00000165, },
		{ 0.00000107, 0.00000253, -0.00000212, 0.00000139, -0.00000530, -0.00000073, 0.00001310, -0.00000506, -0.00000556, 0.00000888, -0.00000928, 0.00000552, -0.00000270, -0.00001638, 0.00000173, -0.00000314, -0.00000448, 0.00000011, -0.00000575, -0.00000235, -0.00000643, 0.00000310, -0.00002328, 0.00000198, -0.00002308, 0.00000089, -0.00000688, -0.00006153, 0.00000094, -0.00017018, -0.00008792, 0.00028338, 0.00012575, 0.00013968, 0.00031807, -0.00025858, -0.00005651, -0.00004699, -0.00001721, 0.00000322, -0.00001004, 0.00000811, -0.00004677, 0.00003134, -0.00001601, 0.00001025, -0.00001831, 0.00001469, -0.00000998, 0.00001863, -0.00001609, -0.00000436, 0.00001902, -0.00001821, -0.00002076, 0.00004390, -0.00005518, 0.00005392, -0.00006726, 0.00005498, -0.00004287, 0.00003443, -0.00001891, 0.00000785, 0.00000007, -0.00001864, 0.00002210, -0.00002033, 0.00002415, -0.00003767, 0.00002405, -0.00000765, 0.00002258, -0.00002675, 0.00001468, -0.00000815, 0.00000559, 0.00000055, -0.00001102, 0.00000579, -0.00000184, 0.00000111, 0.00000293, -0.00000559, 0.00000594, -0.00000991, 0.00000257, 0.00000362, -0.00000003, -0.00000045, -0.00000096, 0.00000103, -0.00000830, 0.00000121, 0.00000448, -0.00000524, 0.00000688, 0.00000314, -0.00000425, -0.00000359, -0.00000313, 0.00000812, 0.00000176, -0.00000347, -0.00000169, -0.00000523, 0.00000080, -0.00000040, 0.00000405, 0.00000395, -0.00000647, -0.00000653, 0.00000108, 0.00000407, 0.00000035, 0.00000184, 0.00000331, -0.00000416, -0.00000122, 0.00000056, -0.00000214, -0.00000088, 0.00000462, 0.00000314, -0.00000407, -0.00000519, 0.00000291, 0.00000123, },
		{ 0.00000187, 0.00000048, 0.00000319, -0.00000228, 0.00000138, -0.00000399, -0.00000377, 0.00001351, -0.00000145, -0.00000880, 0.00001055, -0.00000893, 0.00000227, 0.00000199, -0.00001807, 0.00000031, -0.00000162, -0.00000590, 0.00000242, -0.00000803, -0.00000090, -0.00000820, 0.00000464, -0.00002173, -0.00000198, -0.00001732, -0.00000674, 0.00000367, -0.00006824, 0.00000363, -0.00014036, -0.00015233, 0.00026676, 0.00017633, 0.00012644, 0.00034748, -0.00026480, -0.00006685, -0.00003973, -0.00002843, 0.00001182, -0.00001914, 0.00001494, -0.00005136, 0.00003220, -0.00001539, 0.00000842, -0.00001648, 0.00001168, -0.00000749, 0.00001380, -0.00001057, -0.00000623, 0.00001319, -0.00000146, -0.00004213, 0.00006128, -0.00006526, 0.00005877, -0.00006685, 0.00004934, -0.00003328, 0.00002189, -0.00000590, -0.00000381, 0.00001156, -0.00002795, 0.00002661, -0.00002272, 0.00002722, -0.00003737, 0.00001982, -0.00000654, 0.00002327, -0.00002338, 0.00000890, -0.00000275, 0.00000121, 0.00000392, -0.00001173, 0.00000352, 0.00000082, -0.00000161, 0.00000528, -0.00000664, 0.00000626, -0.00000851, -0.00000030, 0.00000511, -0.00000032, 0.00000002, -0.00000159, 0.00000163, -0.00000707, -0.00000163, 0.00000606, -0.00000502, 0.00000469, 0.00000517, -0.00000401, -0.00000314, -0.00000434, 0.00000667, 0.00000403, -0.00000393, -0.00000137, -0.00000531, 0.00000008, 0.00000017, 0.00000244, 0.00000521, -0.00000482, -0.00000752, 0.00000004, 0.00000384, 0.00000111, 0.00000119, 0.00000331, -0.00000311, -0.00000165, 0.00000059, -0.00000175, -0.00000168, 0.00000410, 0.00000418, -0.00000329, -0.00000613, 0.00000230, },
		{ 0.00000138, 0.00000255, -0.00000000, 0.00000360, -0.00000210, 0.00000095, -0.00000242, -0.00000620, 0.00001255, 0.00000285, -0.00001153, 0.00001101, -0.00000705, -0.00000192, 0.00000612, -0.00001825, -0.00000233, 0.00000058, -0.00000781, 0.00000491, -0.00000991, -0.00000021, -0.00000887, 0.00000445, -0.00001821, -0.00000741, -0.00001102, -0.00001362, 0.00001117, -0.00006906, -0.00000282, -0.00010302, -0.00020607, 0.00022784, 0.00023337, 0.00011805, 0.00037282, -0.00026527, -0.00008266, -0.00002893, -0.00004157, 0.00002087, -0.00002743, 0.00001984, -0.00005345, 0.00003057, -0.00001264, 0.00000502, -0.00001370, 0.00000848, -0.00000553, 0.00001073, -0.00000895, -0.00000217, 0.00000069, 0.00001972, -0.00006358, 0.00007498, -0.00007001, 0.00005804, -0.00006105, 0.00003906, -0.00002029, 0.00000756, 0.00000729, -0.00001465, 0.00002138, -0.00003468, 0.00002856, -0.00002330, 0.00002851, -0.00003519, 0.00001485, -0.00000535, 0.00002298, -0.00001869, 0.00000275, 0.00000200, -0.00000190, 0.00000555, -0.00001061, 0.00000007, 0.00000374, -0.00000399, 0.00000681, -0.00000667, 0.00000555, -0.00000632, -0.00000321, 0.00000606, -0.00000017, 0.00000020, -0.00000184, 0.00000160, -0.00000532, -0.00000421, 0.00000674, -0.00000389, 0.00000219, 0.00000678, -0.00000335, -0.00000290, -0.00000494, 0.00000467, 0.00000610, -0.00000388, -0.00000147, -0.00000497, -0.00000091, 0.00000086, 0.00000103, 0.00000575, -0.00000272, -0.00000828, -0.00000106, 0.00000346, 0.00000176, 0.00000087, 0.00000297, -0.00000209, -0.00000182, 0.00000043, -0.00000124, -0.00000233, 0.00000334, 0.00000506, -0.00000232, -0.00000681, },
		{ -0.00000718, 0.00000022, 0.00000314, -0.00000027, 0.00000372, -0.00000161, 0.00000027, -0.00000086, -0.00000776, 0.00001034, 0.00000735, -0.00001324, 0.00001004, -0.00000391, -0.00000637, 0.00000903, -0.00001664, -0.00000604, 0.00000315, -0.00000982, 0.00000709, -0.00001092, -0.00000066, -0.00000817, 0.00000257, -0.00001321, -0.00001334, -0.00000543, -0.00001831, 0.00001438, -0.00006371, -0.00001651, -0.00006349, -0.00024350, 0.00016705, 0.00029122, 0.00011899, 0.00039218, -0.00025936, -0.00010352, -0.00001577, -0.00005511, 0.00002885, -0.00003353, 0.00002182, -0.00005258, 0.00002652, -0.00000829, 0.00000095, -0.00001115, 0.00000647, -0.00000562, 0.00001099, -0.00001241, 0.00000799, -0.00001709, 0.00004259, -0.00008206, 0.00008267, -0.00006819, 0.00005144, -0.00005031, 0.00002514, -0.00000526, -0.00000708, 0.00001937, -0.00002366, 0.00002880, -0.00003848, 0.00002801, -0.00002230, 0.00002826, -0.00003154, 0.00000957, -0.00000425, 0.00002181, -0.00001312, -0.00000310, 0.00000553, -0.00000339, 0.00000542, -0.00000797, -0.00000397, 0.00000633, -0.00000555, 0.00000727, -0.00000566, 0.00000403, -0.00000374, -0.00000575, 0.00000626, 0.00000049, 0.00000010, -0.00000174, 0.00000109, -0.00000340, -0.00000621, 0.00000645, -0.00000207, -0.00000029, 0.00000775, -0.00000223, -0.00000290, -0.00000501, 0.00000237, 0.00000768, -0.00000320, -0.00000194, -0.00000434, -0.00000200, 0.00000145, 0.00000002, 0.00000557, -0.00000039, -0.00000865, -0.00000227, 0.00000301, 0.00000222, 0.00000085, 0.00000244, -0.00000126, -0.00000168, 0.00000013, -0.00000071, -0.00000276, 0.00000236, 0.00000576, -0.00000122, },
		{ -0.00000005, -0.00000723, -0.00000107, 0.00000352, -0.00000027, 0.00000358, -0.00000093, -0.00000048, 0.00000042, -0.00000836, 0.00000717, 0.00001148, -0.00001354, 0.00000764, 0.00000004, -0.00001034, 0.00001018, -0.00001325, -0.00001044, 0.00000562, -0.00001144, 0.00000847, -0.00001069, -0.00000244, -0.00000617, -0.00000061, -0.00000757, -0.00001864, -0.00000177, -0.00001978, 0.00001292, -0.00005308, -0.00003435, -0.00002728, -0.00026072, 0.00008754, 0.00034281, 0.00013336, 0.00040480, -0.00024757, -0.00012804, -0.00000212, -0.00006724, 0.00003420, -0.00003634, 0.00002040, -0.00004891, 0.00002076, -0.00000343, -0.00000245, -0.00001029, 0.00000709, -0.00000907, 0.00001559, -0.00002123, 0.00002333, -0.00003781, 0.00006397, -0.00009469, 0.00008267, -0.00005946, 0.00003952, -0.00003572, 0.00000903, 0.00001022, -0.00002058, 0.00002923, -0.00003010, 0.00003341, -0.00003934, 0.00002532, -0.00002011, 0.00002683, -0.00002688, 0.00000440, -0.00000336, 0.00001992, -0.00000716, -0.00000802, 0.00000743, -0.00000318, 0.00000378, -0.00000438, -0.00000790, 0.00000803, -0.00000600, 0.00000662, -0.00000380, 0.00000204, -0.00000119, -0.00000757, 0.00000559, 0.00000156, -0.00000018, -0.00000140, 0.00000032, -0.00000159, -0.00000742, 0.00000527, 0.00000013, -0.00000241, 0.00000794, -0.00000069, -0.00000305, -0.00000470, 0.00000006, 0.00000857, -0.00000190, -0.00000268, -0.00000358, -0.00000300, 0.00000176, -0.00000050, 0.00000477, 0.00000193, -0.00000851, -0.00000360, 0.00000251, 0.00000248, 0.00000106, 0.00000188, -0.00000071, -0.00000126, -0.00000022, -0.00000024, -0.00000292, 0.00000121, 0.00000624, },
		{ 0.00000647, 0.00000115, -0.00000698, -0.00000237, 0.00000359, -0.00000000, 0.00000323, -0.00000019, -0.00000112, 0.00000125, -0.00000802, 0.00000347, 0.00001472, -0.00001225, 0.00000413, 0.00000413, -0.00001313, 0.00000928, -0.00000838, -0.00001496, 0.00000746, -0.00001222, 0.00000869, -0.00000904, -0.00000543, -0.00000322, -0.00000440, -0.00000227, -0.00002223, -0.00000091, -0.00001766, 0.00000727, -0.00003906, -0.00005263, 0.00000073, -0.00025613, -0.00000498, 0.00038038, 0.00016409, 0.00041150, -0.00023170, -0.00015395, 0.00000976, -0.00007612, 0.00003568, -0.00003533, 0.00001576, -0.00004329, 0.00001457, 0.00000041, -0.00000362, -0.00001256, 0.00001152, -0.00001668, 0.00002475, -0.00003472, 0.00004189, -0.00005842, 0.00008059, -0.00009922, 0.00007431, -0.00004440, 0.00002355, -0.00001889, -0.00000755, 0.00002455, -0.00003167, 0.00003607, -0.00003361, 0.00003516, -0.00003758, 0.00002101, -0.00001720, 0.00002460, -0.00002168, -0.00000029, -0.00000280, 0.00001751, -0.00000135, -0.00001152, 0.00000751, -0.00000142, 0.00000106, -0.00000049, -0.00001109, 0.00000845, -0.00000526, 0.00000502, -0.00000144, 0.00000001, 0.00000092, -0.00000845, 0.00000410, 0.00000288, -0.00000048, -0.00000096, -0.00000051, -0.00000016, -0.00000778, 0.00000340, 0.00000235, -0.00000389, 0.00000732, 0.00000113, -0.00000324, -0.00000419, -0.00000200, 0.00000865, -0.00000010, -0.00000347, -0.00000289, -0.00000373, 0.00000171, -0.00000054, 0.00000354, 0.00000399, -0.00000776, -0.00000501, 0.00000197, 0.00000256, 0.00000138, 0.00000143, -0.00000050, -0.00000065, -0.00000052, 0.00000009, -0.00000280, -0.00000005, },
		{ -0.00000133, 0.00000639, 0.00000235, -0.00000648, -0.00000356, 0.00000333, 0.00000047, 0.00000281, 0.00000048, -0.00000152, 0.00000157, -0.00000692, -0.00000029, 0.00001666, -0.00000941, -0.00000002, 0.00000765, -0.00001420, 0.00000636, -0.00000262, -0.00001888, 0.00000812, -0.00001180, 0.00000757, -0.00000610, -0.00000927, 0.00000007, -0.00000798, 0.00000170, -0.00002330, -0.00000319, -0.00001227, -0.00000123, -0.00002424, -0.00006761, 0.00001700, -0.00023083, -0.00010273, 0.00039650, 0.00021208, 0.00041503, -0.00021472, -0.00017846, 0.00001758, -0.00008031, 0.00003274, -0.00003075, 0.00000892, -0.00003724, 0.00000973, 0.00000152, -0.00000106, -0.00001908, 0.00002036, -0.00002847, 0.00003780, -0.00005121, 0.00006099, -0.00007567, 0.00008971, -0.00009438, 0.00005799, -0.00002451, 0.00000540, -0.00000167, -0.00002291, 0.00003637, -0.00003948, 0.00003952, -0.00003419, 0.00003434, -0.00003373, 0.00001575, -0.00001407, 0.00002195, -0.00001637, -0.00000424, -0.00000263, 0.00001483, 0.00000383, -0.00001327, 0.00000585, 0.00000149, -0.00000213, 0.00000305, -0.00001305, 0.00000744, -0.00000349, 0.00000281, 0.00000101, -0.00000167, 0.00000229, -0.00000832, 0.00000199, 0.00000419, -0.00000063, -0.00000056, -0.00000121, 0.00000077, -0.00000735, 0.00000114, 0.00000426, -0.00000457, 0.00000598, 0.00000305, -0.00000330, -0.00000367, -0.00000362, 0.00000794, 0.00000199, -0.00000411, -0.00000244, -0.00000410, 0.00000126, -0.00000021, 0.00000212, 0.00000555, -0.00000640, -0.00000640, 0.00000133, 0.00000252, 0.00000168, 0.00000119, -0.00000060, 0.00000001, -0.00000064, 0.00000024, -0.00000242, },
		{ -0.00000182, -0.00000252, 0.00000599, 0.00000352, -0.00000577, -0.00000456, 0.00000272, 0.00000106, 0.00000244, 0.00000095, -0.00000159, 0.00000141, -0.00000533, -0.00000368, 0.00001708, -0.00000532, -0.00000413, 0.00000995, -0.00001327, 0.00000178, 0.00000325, -0.00002148, 0.00000722, -0.00001005, 0.00000522, -0.00000223, -0.00001334, 0.00000296, -0.00001048, 0.00000359, -0.00002148, -0.00000834, -0.00000466, -0.00001064, -0.00001139, -0.00007619, 0.00001987, -0.00018846, -0.00019669, 0.00038515, 0.00027561, 0.00042002, -0.00020040, -0.00019871, 0.00001954, -0.00007918, 0.00002572, -0.00002378, 0.00000162, -0.00003280, 0.00000816, -0.00000172, 0.00000632, -0.00003034, 0.00003346, -0.00004362, 0.00005315, -0.00006829, 0.00007759, -0.00008657, 0.00008943, -0.00008012, 0.00003522, -0.00000198, -0.00001279, 0.00001408, -0.00003558, 0.00004470, -0.00004357, 0.00003965, -0.00003221, 0.00003148, -0.00002848, 0.00001019, -0.00001115, 0.00001921, -0.00001131, -0.00000722, -0.00000289, 0.00001211, 0.00000800, -0.00001316, 0.00000275, 0.00000499, -0.00000515, 0.00000567, -0.00001348, 0.00000510, -0.00000102, 0.00000044, 0.00000312, -0.00000268, 0.00000278, -0.00000728, -0.00000047, 0.00000522, -0.00000049, -0.00000030, -0.00000167, 0.00000114, -0.00000630, -0.00000118, 0.00000558, -0.00000440, 0.00000413, 0.00000481, -0.00000310, -0.00000326, -0.00000469, 0.00000653, 0.00000412, -0.00000437, -0.00000233, -0.00000409, 0.00000050, 0.00000034, 0.00000077, 0.00000647, -0.00000451, -0.00000763, 0.00000053, 0.00000246, 0.00000187, 0.00000119, -0.00000091, 0.00000059, -0.00000053, 0.00000019, },
		{ -0.00000001, -0.00000110, -0.00000353, 0.00000524, 0.00000462, -0.00000489, -0.00000530, 0.00000183, 0.00000162, 0.00000223, 0.00000115, -0.00000134, 0.00000089, -0.00000355, -0.00000633, 0.00001597, -0.00000047, -0.00000753, 0.00001055, -0.00001041, -0.00000377, 0.00000833, -0.00002219, 0.00000460, -0.00000709, 0.00000201, 0.00000196, -0.00001690, 0.00000469, -0.00001127, 0.00000303, -0.00001699, -0.00001547, 0.00000359, -0.00001878, -0.00000297, -0.00007643, 0.00000979, -0.00013487, -0.00027766, 0.00034278, 0.00034997, 0.00043267, -0.00019268, -0.00021244, 0.00001479, -0.00007319, 0.00001605, -0.00001642, -0.00000384, -0.00003214, 0.00001163, -0.00001043, 0.00001895, -0.00004603, 0.00004978, -0.00006046, 0.00006856, -0.00008317, 0.00008875, -0.00008886, 0.00007911, -0.00005767, 0.00000845, 0.00002057, -0.00002887, 0.00002680, -0.00004452, 0.00004907, -0.00004400, 0.00003690, -0.00002829, 0.00002726, -0.00002255, 0.00000494, -0.00000878, 0.00001662, -0.00000675, -0.00000914, -0.00000357, 0.00000953, 0.00001090, -0.00001130, -0.00000128, 0.00000844, -0.00000743, 0.00000698, -0.00001236, 0.00000175, 0.00000165, -0.00000163, 0.00000455, -0.00000283, 0.00000242, -0.00000559, -0.00000291, 0.00000572, 0.00000002, -0.00000022, -0.00000184, 0.00000102, -0.00000488, -0.00000325, 0.00000614, -0.00000349, 0.00000203, 0.00000619, -0.00000254, -0.00000305, -0.00000521, 0.00000464, 0.00000599, -0.00000411, -0.00000258, -0.00000374, -0.00000045, 0.00000090, -0.00000031, 0.00000669, -0.00000226, -0.00000853, -0.00000050, 0.00000241, 0.00000191, 0.00000140, -0.00000130, 0.00000093, -0.00000014, },
		{ 0.00000049, -0.00000028, -0.00000037, -0.00000425, 0.00000417, 0.00000560, -0.00000388, -0.00000578, 0.00000075, 0.00000204, 0.00000227, 0.00000108, -0.00000084, 0.00000018, -0.00000187, -0.00000800, 0.00001351, 0.00000451, -0.00000962, 0.00000925, -0.00000602, -0.00000944, 0.00001179, -0.00002064, 0.00000040, -0.00000335, -0.00000147, 0.00000575, -0.00001924, 0.00000473, -0.00001003, 0.00000014, -0.00001055, -0.00002321, 0.00001070, -0.00002365, -0.00000069, -0.00006789, -0.00001071, -0.00007724, -0.00033752, 0.00026909, 0.00042759, 0.00046004, -0.00019492, -0.00021855, 0.00000386, -0.00006409, 0.00000611, -0.00001131, -0.00000497, -0.00003729, 0.00002136, -0.00002497, 0.00003634, -0.00006490, 0.00006748, -0.00007669, 0.00008140, -0.00009312, 0.00009214, -0.00008142, 0.00005944, -0.00002937, -0.00001932, 0.00004053, -0.00004108, 0.00003543, -0.00004927, 0.00004952, -0.00004124, 0.00003201, -0.00002320, 0.00002240, -0.00001661, 0.00000048, -0.00000716, 0.00001432, -0.00000289, -0.00000998, -0.00000463, 0.00000725, 0.00001245, -0.00000802, -0.00000563, 0.00001123, -0.00000853, 0.00000686, -0.00000990, -0.00000213, 0.00000402, -0.00000303, 0.00000509, -0.00000213, 0.00000139, -0.00000357, -0.00000498, 0.00000552, 0.00000092, -0.00000030, -0.00000177, 0.00000054, -0.00000336, -0.00000482, 0.00000592, -0.00000201, -0.00000002, 0.00000700, -0.00000158, -0.00000304, -0.00000526, 0.00000252, 0.00000737, -0.00000326, -0.00000314, -0.00000320, -0.00000140, 0.00000129, -0.00000100, 0.00000627, 0.00000013, -0.00000892, -0.00000178, 0.00000239, 0.00000180, 0.00000171, -0.00000161, 0.00000096, },
		{ 0.00000063, 0.00000127, -0.00000050, 0.00000028, -0.00000462, 0.00000281, 0.00000639, -0.00000275, -0.00000598, -0.00000041, 0.00000221, 0.00000258, 0.00000077, -0.00000020, -0.00000054, -0.00000053, -0.00000864, 0.00001010, 0.00000899, -0.00000996, 0.00000617, -0.00000078, -0.00001426, 0.00001298, -0.00001684, -0.00000491, 0.00000052, -0.00000451, 0.00000841, -0.00001983, 0.00000280, -0.00000686, -0.00000449, -0.00000334, -0.00002997, 0.00001495, -0.00002380, -0.00000514, -0.00005174, -0.00003751, -0.00002318, -0.00037031, 0.00016753, 0.00049848, 0.00050907, -0.00020897, -0.00021770, -0.00001115, -0.00005479, -0.00000099, -0.00001132, 0.00000050, -0.00004960, 0.00003770, -0.00004472, 0.00005705, -0.00008492, 0.00008410, -0.00008969, 0.00008911, -0.00009590, 0.00008641, -0.00006441, 0.00003241, 0.00000160, -0.00004492, 0.00005563, -0.00004821, 0.00003955, -0.00004992, 0.00004660, -0.00003612, 0.00002586, -0.00001773, 0.00001755, -0.00001120, -0.00000289, -0.00000636, 0.00001237, 0.00000018, -0.00000980, -0.00000597, 0.00000534, 0.00001269, -0.00000378, -0.00000966, 0.00001288, -0.00000826, 0.00000539, -0.00000652, -0.00000596, 0.00000562, -0.00000349, 0.00000470, -0.00000073, -0.00000001, -0.00000159, -0.00000640, 0.00000458, 0.00000209, -0.00000043, -0.00000154, -0.00000012, -0.00000197, -0.00000575, 0.00000498, -0.00000026, -0.00000175, 0.00000716, -0.00000027, -0.00000317, -0.00000498, 0.00000043, 0.00000807, -0.00000186, -0.00000385, -0.00000263, -0.00000220, 0.00000139, -0.00000124, 0.00000533, 0.00000238, -0.00000868, -0.00000329, 0.00000236, 0.00000161, 0.00000200, -0.00000172, },
		{ -0.00000154, 0.00000001, 0.00000205, -0.00000055, 0.00000075, -0.00000462, 0.00000126, 0.00000691, -0.00000152, -0.00000595, -0.00000154, 0.00000204, 0.00000315, 0.00000033, 0.00000043, -0.00000109, 0.00000033, -0.00000833, 0.00000619, 0.00001238, -0.00000839, 0.00000174, 0.00000444, -0.00001739, 0.00001160, -0.00001117, -0.00001058, 0.00000376, -0.00000640, 0.00000943, -0.00001842, -0.00000093, -0.00000231, -0.00000985, 0.00000321, -0.00003424, 0.00001511, -0.00001865, -0.00001566, -0.00003054, -0.00006551, 0.00002039, -0.00037312, 0.00004529, 0.00055122, 0.00058542, -0.00023441, -0.00021258, -0.00002665, -0.00004913, -0.00000185, -0.00001903, 0.00001414, -0.00006946, 0.00005988, -0.00006799, 0.00007877, -0.00010340, 0.00009691, -0.00009695, 0.00008962, -0.00009016, 0.00007143, -0.00003934, 0.00000100, 0.00003168, -0.00006546, 0.00006432, -0.00004985, 0.00003938, -0.00004709, 0.00004118, -0.00002962, 0.00001936, -0.00001257, 0.00001323, -0.00000669, -0.00000503, -0.00000632, 0.00001075, 0.00000245, -0.00000873, -0.00000745, 0.00000379, 0.00001183, 0.00000087, -0.00001277, 0.00001307, -0.00000664, 0.00000290, -0.00000276, -0.00000918, 0.00000610, -0.00000295, 0.00000354, 0.00000108, -0.00000145, 0.00000001, -0.00000700, 0.00000296, 0.00000336, -0.00000049, -0.00000126, -0.00000081, -0.00000090, -0.00000600, 0.00000354, 0.00000148, -0.00000293, 0.00000664, 0.00000125, -0.00000331, -0.00000453, -0.00000140, 0.00000802, -0.00000005, -0.00000453, -0.00000219, -0.00000271, 0.00000114, -0.00000109, 0.00000411, 0.00000425, -0.00000776, -0.00000493, 0.00000224, 0.00000143, 0.00000216, },
		{ 0.00000209, -0.00000106, -0.00000080, 0.00000266, -0.00000036, 0.00000103, -0.00000429, -0.00000035, 0.00000709, -0.00000020, -0.00000572, -0.00000253, 0.00000152, 0.00000389, -0.00000013, 0.00000092, -0.00000136, 0.00000067, -0.00000728, 0.00000231, 0.00001427, -0.00000505, -0.00000336, 0.00000874, -0.00001822, 0.00000774, -0.00000436, -0.00001569, 0.00000560, -0.00000664, 0.00000855, -0.00001517, -0.00000593, 0.00000272, -0.00001473, 0.00000776, -0.00003487, 0.00001066, -0.00000864, -0.00003045, -0.00000783, -0.00008949, 0.00004826, -0.00034668, -0.00008727, 0.00057425, 0.00069224, -0.00026796, -0.00020798, -0.00003784, -0.00005126, 0.00000658, -0.00003621, 0.00003637, -0.00009599, 0.00008598, -0.00009218, 0.00009857, -0.00011746, 0.00010334, -0.00009649, 0.00008170, -0.00007570, 0.00004837, -0.00000887, -0.00003116, 0.00005746, -0.00007876, 0.00006593, -0.00004640, 0.00003575, -0.00004176, 0.00003432, -0.00002274, 0.00001330, -0.00000825, 0.00000977, -0.00000327, -0.00000598, -0.00000687, 0.00000940, 0.00000395, -0.00000694, -0.00000893, 0.00000253, 0.00001017, 0.00000535, -0.00001454, 0.00001175, -0.00000393, -0.00000011, 0.00000078, -0.00001132, 0.00000533, -0.00000155, 0.00000188, 0.00000295, -0.00000257, 0.00000099, -0.00000674, 0.00000085, 0.00000450, -0.00000035, -0.00000101, -0.00000139, -0.00000025, -0.00000566, 0.00000182, 0.00000294, -0.00000345, 0.00000556, 0.00000282, -0.00000335, -0.00000407, -0.00000282, 0.00000726, 0.00000196, -0.00000496, -0.00000202, -0.00000288, 0.00000056, -0.00000068, 0.00000283, 0.00000555, -0.00000620, -0.00000655, 0.00000191, 0.00000136, },
		{ 0.00000147, 0.00000176, -0.00000035, -0.00000162, 0.00000296, 0.00000014, 0.00000112, -0.00000370, -0.00000189, 0.00000686, 0.00000118, -0.00000533, -0.00000331, 0.00000068, 0.00000469, -0.00000046, 0.00000118, -0.00000131, 0.00000057, -0.00000580, -0.00000108, 0.00001448, -0.00000033, -0.00000831, 0.00001133, -0.00001648, 0.00000191, 0.00000259, -0.00001932, 0.00000548, -0.00000497, 0.00000591, -0.00001059, -0.00001132, 0.00000714, -0.00001792, 0.00000923, -0.00003134, 0.00000194, 0.00000481, -0.00004681, 0.00001252, -0.00010500, 0.00005770, -0.00029536, -0.00021757, 0.00055737, 0.00082914, -0.00030322, -0.00021038, -0.00003940, -0.00006504, 0.00002642, -0.00006336, 0.00006622, -0.00012703, 0.00011302, -0.00011402, 0.00011330, -0.00012438, 0.00010139, -0.00008718, 0.00006527, -0.00005356, 0.00001954, 0.00002361, -0.00006038, 0.00007609, -0.00008365, 0.00006082, -0.00003897, 0.00002989, -0.00003510, 0.00002710, -0.00001634, 0.00000827, -0.00000506, 0.00000729, -0.00000094, -0.00000589, -0.00000781, 0.00000824, 0.00000476, -0.00000466, -0.00001023, 0.00000147, 0.00000803, 0.00000916, -0.00001473, 0.00000908, -0.00000057, -0.00000311, 0.00000360, -0.00001212, 0.00000337, 0.00000042, 0.00000009, 0.00000452, -0.00000313, 0.00000125, -0.00000574, -0.00000147, 0.00000528, 0.00000008, -0.00000085, -0.00000179, -0.00000002, -0.00000488, 0.00000011, 0.00000391, -0.00000331, 0.00000409, 0.00000424, -0.00000316, -0.00000372, -0.00000374, 0.00000590, 0.00000391, -0.00000498, -0.00000219, -0.00000273, -0.00000023, -0.00000018, 0.00000171, 0.00000618, -0.00000414, -0.00000793, 0.00000126, },
		{ 0.00000021, 0.00000178, 0.00000121, 0.00000046, -0.00000228, 0.00000283, 0.00000091, 0.00000110, -0.00000297, -0.00000322, 0.00000618, 0.00000258, -0.00000479, -0.00000386, -0.00000042, 0.00000541, -0.00000058, 0.00000115, -0.00000095, 0.00000015, -0.00000419, -0.00000363, 0.00001304, 0.00000515, -0.00001228, 0.00001169, -0.00001234, -0.00000501, 0.00000857, -0.00002070, 0.00000312, -0.00000156, 0.00000201, -0.00000555, -0.00001606, 0.00000988, -0.00001849, 0.00000711, -0.00002389, -0.00000986, 0.00001954, -0.00006170, 0.00002697, -0.00010910, 0.00004888, -0.00022675, -0.00033222, 0.00049323, 0.00099149, -0.00033092, -0.00022728, -0.00002631, -0.00009319, 0.00005827, -0.00009934, 0.00010122, -0.00015922, 0.00013732, -0.00013002, 0.00012006, -0.00012210, 0.00009000, -0.00006908, 0.00004140, -0.00002587, -0.00001192, 0.00005440, -0.00008335, 0.00008573, -0.00008009, 0.00005024, -0.00002917, 0.00002320, -0.00002826, 0.00002042, -0.00001106, 0.00000458, -0.00000308, 0.00000574, 0.00000043, -0.00000500, -0.00000890, 0.00000717, 0.00000498, -0.00000212, -0.00001116, 0.00000048, 0.00000576, 0.00001195, -0.00001331, 0.00000545, 0.00000290, -0.00000557, 0.00000532, -0.00001151, 0.00000050, 0.00000255, -0.00000144, 0.00000553, -0.00000300, 0.00000083, -0.00000421, -0.00000369, 0.00000552, 0.00000084, -0.00000079, -0.00000198, -0.00000015, -0.00000387, -0.00000136, 0.00000430, -0.00000259, 0.00000246, 0.00000532, -0.00000268, -0.00000354, -0.00000418, 0.00000415, 0.00000555, -0.00000447, -0.00000266, -0.00000234, -0.00000111, 0.00000023, 0.00000092, 0.00000613, -0.00000180, -0.00000886, },
		{ -0.00000917, -0.00000125, 0.00000224, 0.00000053, 0.00000120, -0.00000259, 0.00000223, 0.00000187, 0.00000108, -0.00000222, -0.00000421, 0.00000506, 0.00000393, -0.00000411, -0.00000416, -0.00000167, 0.00000593, -0.00000040, 0.00000085, -0.00000039, -0.00000040, -0.00000274, -0.00000516, 0.00001024, 0.00001066, -0.00001458, 0.00000964, -0.00000637, -0.00001192, 0.00001255, -0.00001935, -0.00000134, 0.00000305, -0.00000231, -0.00000107, -0.00001911, 0.00001011, -0.00001598, 0.00000153, -0.00001352, -0.00002286, 0.00003301, -0.00007223, 0.00003299, -0.00010088, 0.00002486, -0.00015072, -0.00041863, 0.00037868, 0.00117016, -0.00033961, -0.00026611, 0.00000520, -0.00013665, 0.00010082, -0.00014130, 0.00013759, -0.00018842, 0.00015498, -0.00013700, 0.00011667, -0.00010959, 0.00006932, -0.00004340, 0.00001228, 0.00000438, -0.00004249, 0.00008006, -0.00009769, 0.00008579, -0.00006923, 0.00003616, -0.00001886, 0.00001703, -0.00002217, 0.00001494, -0.00000724, 0.00000228, -0.00000218, 0.00000495, 0.00000102, -0.00000360, -0.00000992, 0.00000613, 0.00000473, 0.00000045, -0.00001159, -0.00000056, 0.00000364, 0.00001349, -0.00001051, 0.00000135, 0.00000592, -0.00000707, 0.00000577, -0.00000965, -0.00000285, 0.00000440, -0.00000239, 0.00000580, -0.00000217, -0.00000011, -0.00000245, -0.00000553, 0.00000510, 0.00000188, -0.00000077, -0.00000202, -0.00000052, -0.00000283, -0.00000243, 0.00000408, -0.00000147, 0.00000091, 0.00000593, -0.00000189, -0.00000351, -0.00000422, 0.00000224, 0.00000668, -0.00000340, -0.00000336, -0.00000186, -0.00000189, 0.00000040, 0.00000054, 0.00000551, 0.00000056, },
		{ 0.00000266, -0.00000873, -0.00000305, 0.00000274, -0.00000012, 0.00000170, -0.00000245, 0.00000121, 0.00000284, 0.00000120, -0.00000156, -0.00000479, 0.00000355, 0.00000512, -0.00000329, -0.00000425, -0.00000294, 0.00000613, 0.00000007, 0.00000033, 0.00000026, -0.00000090, -0.00000165, -0.00000565, 0.00000653, 0.00001545, -0.00001475, 0.00000540, 0.00000051, -0.00001765, 0.00001378, -0.00001528, -0.00000733, 0.00000802, -0.00000605, 0.00000186, -0.00001969, 0.00000744, -0.00001050, -0.00000667, -0.00000184, -0.00003485, 0.00004280, -0.00007624, 0.00002952, -0.00008165, -0.00000883, -0.00007819, -0.00046659, 0.00021563, 0.00135186, -0.00031682, -0.00033303, 0.00005672, -0.00019402, 0.00015071, -0.00018483, 0.00017059, -0.00021018, 0.00016240, -0.00013259, 0.00010208, -0.00008708, 0.00004073, -0.00001249, -0.00001910, 0.00003382, -0.00006877, 0.00009788, -0.00010217, 0.00007693, -0.00005310, 0.00002093, -0.00000982, 0.00001241, -0.00001741, 0.00001095, -0.00000490, 0.00000121, -0.00000209, 0.00000468, 0.00000110, -0.00000196, -0.00001069, 0.00000506, 0.00000412, 0.00000282, -0.00001141, -0.00000173, 0.00000186, 0.00001377, -0.00000670, -0.00000261, 0.00000803, -0.00000739, 0.00000500, -0.00000690, -0.00000618, 0.00000555, -0.00000256, 0.00000532, -0.00000081, -0.00000132, -0.00000077, -0.00000674, 0.00000401, 0.00000308, -0.00000071, -0.00000196, -0.00000101, -0.00000194, -0.00000301, 0.00000338, -0.00000019, -0.00000035, 0.00000602, -0.00000084, -0.00000359, -0.00000398, 0.00000040, 0.00000716, -0.00000186, -0.00000410, -0.00000146, -0.00000245, 0.00000025, 0.00000056, 0.00000451, },
		{ 0.00000336, 0.00000426, -0.00000752, -0.00000505, 0.00000310, -0.00000058, 0.00000182, -0.00000183, -0.00000010, 0.00000365, 0.00000157, -0.00000108, -0.00000493, 0.00000178, 0.00000605, -0.00000230, -0.00000416, -0.00000414, 0.00000597, 0.00000081, -0.00000029, 0.00000087, -0.00000121, -0.00000102, -0.00000524, 0.00000245, 0.00001893, -0.00001266, -0.00000045, 0.00000723, -0.00002118, 0.00001191, -0.00000893, -0.00001391, 0.00001229, -0.00000824, 0.00000245, -0.00001741, 0.00000202, -0.00000281, -0.00001611, 0.00000917, -0.00004368, 0.00004703, -0.00007274, 0.00001727, -0.00005482, -0.00004505, -0.00001966, -0.00046970, 0.00001145, 0.00152015, -0.00025056, -0.00043168, 0.00012693, -0.00026137, 0.00020268, -0.00022444, 0.00019519, -0.00022045, 0.00015692, -0.00011567, 0.00007661, -0.00005614, 0.00000675, 0.00002054, -0.00004936, 0.00005915, -0.00008797, 0.00010623, -0.00009693, 0.00006100, -0.00003437, 0.00000691, -0.00000346, 0.00000993, -0.00001419, 0.00000845, -0.00000386, 0.00000105, -0.00000250, 0.00000467, 0.00000087, -0.00000029, -0.00001109, 0.00000395, 0.00000329, 0.00000480, -0.00001055, -0.00000304, 0.00000053, 0.00001291, -0.00000238, -0.00000590, 0.00000891, -0.00000649, 0.00000326, -0.00000373, -0.00000898, 0.00000572, -0.00000191, 0.00000423, 0.00000086, -0.00000250, 0.00000056, -0.00000722, 0.00000236, 0.00000426, -0.00000050, -0.00000187, -0.00000150, -0.00000132, -0.00000311, 0.00000234, 0.00000103, -0.00000116, 0.00000560, 0.00000034, -0.00000367, -0.00000362, -0.00000117, 0.00000694, -0.00000000, -0.00000470, -0.00000126, -0.00000271, -0.00000022, 0.00000086, },
		{ 0.00000131, 0.00000230, 0.00000520, -0.00000564, -0.00000701, 0.00000315, -0.00000072, 0.00000150, -0.00000081, -0.00000149, 0.00000409, 0.00000227, -0.00000081, -0.00000468, -0.00000013, 0.00000660, -0.00000115, -0.00000393, -0.00000516, 0.00000543, 0.00000171, -0.00000088, 0.00000131, -0.00000125, -0.00000086, -0.00000419, -0.00000147, 0.00002068, -0.00000853, -0.00000703, 0.00001267, -0.00002182, 0.00000709, -0.00000122, -0.00001994, 0.00001483, -0.00000809, 0.00000032, -0.00001244, -0.00000544, 0.00000586, -0.00002507, 0.00001760, -0.00004765, 0.00004481, -0.00006215, -0.00000139, -0.00002531, -0.00007607, 0.00001623, -0.00042631, -0.00022139, 0.00165685, -0.00013102, -0.00056201, 0.00021122, -0.00033235, 0.00025011, -0.00025421, 0.00020670, -0.00021616, 0.00013726, -0.00008664, 0.00004205, -0.00001952, -0.00002927, 0.00005215, -0.00007521, 0.00007763, -0.00009831, 0.00010479, -0.00008335, 0.00004067, -0.00001590, -0.00000390, -0.00000062, 0.00000966, -0.00001232, 0.00000716, -0.00000375, 0.00000141, -0.00000307, 0.00000471, 0.00000053, 0.00000121, -0.00001106, 0.00000281, 0.00000233, 0.00000627, -0.00000905, -0.00000447, -0.00000037, 0.00001117, 0.00000192, -0.00000809, 0.00000845, -0.00000458, 0.00000095, -0.00000065, -0.00001086, 0.00000480, -0.00000057, 0.00000277, 0.00000253, -0.00000338, 0.00000134, -0.00000695, 0.00000034, 0.00000521, -0.00000005, -0.00000181, -0.00000189, -0.00000101, -0.00000281, 0.00000118, 0.00000199, -0.00000147, 0.00000478, 0.00000152, -0.00000365, -0.00000325, -0.00000235, 0.00000610, 0.00000195, -0.00000498, -0.00000138, -0.00000264, -0.00000094, },
		{ -0.00000179, 0.00000170, 0.00000151, 0.00000542, -0.00000327, -0.00000868, 0.00000272, -0.00000047, 0.00000075, 0.00000047, -0.00000270, 0.00000398, 0.00000333, -0.00000070, -0.00000412, -0.00000198, 0.00000668, 0.00000015, -0.00000359, -0.00000595, 0.00000456, 0.00000265, -0.00000134, 0.00000151, -0.00000101, -0.00000106, -0.00000280, -0.00000477, 0.00002056, -0.00000288, -0.00001335, 0.00001593, -0.00001933, 0.00000000, 0.00000662, -0.00002419, 0.00001481, -0.00000523, -0.00000440, -0.00000545, -0.00001375, 0.00001397, -0.00003185, 0.00002189, -0.00004591, 0.00003646, -0.00004632, -0.00002271, 0.00000117, -0.00009484, 0.00002387, -0.00033996, -0.00046627, 0.00174398, 0.00004779, -0.00071961, 0.00030172, -0.00039879, 0.00028579, -0.00026861, 0.00020158, -0.00019580, 0.00010384, -0.00004750, 0.00000143, 0.00001915, -0.00006360, 0.00007896, -0.00009392, 0.00008748, -0.00009921, 0.00009452, -0.00006387, 0.00001903, -0.00000032, -0.00001014, -0.00000144, 0.00001115, -0.00001132, 0.00000666, -0.00000414, 0.00000193, -0.00000357, 0.00000465, 0.00000019, 0.00000245, -0.00001059, 0.00000166, 0.00000134, 0.00000712, -0.00000697, -0.00000590, -0.00000094, 0.00000889, 0.00000571, -0.00000890, 0.00000676, -0.00000200, -0.00000145, 0.00000186, -0.00001155, 0.00000286, 0.00000120, 0.00000126, 0.00000391, -0.00000376, 0.00000152, -0.00000606, -0.00000179, 0.00000574, 0.00000069, -0.00000179, -0.00000216, -0.00000100, -0.00000227, 0.00000010, 0.00000257, -0.00000129, 0.00000372, 0.00000254, -0.00000345, -0.00000298, -0.00000310, 0.00000477, 0.00000376, -0.00000478, -0.00000181, -0.00000231, },
		{ -0.00000185, -0.00000258, 0.00000189, 0.00000113, 0.00000499, -0.00000069, -0.00000980, 0.00000173, 0.00000018, -0.00000031, 0.00000177, -0.00000351, 0.00000324, 0.00000467, -0.00000067, -0.00000338, -0.00000360, 0.00000622, 0.00000156, -0.00000314, -0.00000649, 0.00000343, 0.00000351, -0.00000157, 0.00000144, -0.00000053, -0.00000148, -0.00000138, -0.00000712, 0.00001869, 0.00000353, -0.00001844, 0.00001642, -0.00001399, -0.00000824, 0.00001321, -0.00002568, 0.00001181, 0.00000022, -0.00001106, 0.00000241, -0.00002146, 0.00001997, -0.00003505, 0.00002119, -0.00003864, 0.00002349, -0.00002825, -0.00004220, 0.00001908, -0.00009604, 0.00000155, -0.00021916, -0.00070397, 0.00176580, 0.00028667, -0.00089540, 0.00038788, -0.00045158, 0.00030287, -0.00026339, 0.00017802, -0.00015976, 0.00005886, -0.00000166, -0.00004125, 0.00005588, -0.00009267, 0.00009815, -0.00010370, 0.00008808, -0.00009138, 0.00007748, -0.00004150, -0.00000087, 0.00001040, -0.00001130, -0.00000537, 0.00001358, -0.00001058, 0.00000650, -0.00000466, 0.00000233, -0.00000383, 0.00000442, -0.00000009, 0.00000335, -0.00000973, 0.00000056, 0.00000038, 0.00000734, -0.00000448, -0.00000720, -0.00000130, 0.00000643, 0.00000860, -0.00000829, 0.00000412, 0.00000078, -0.00000348, 0.00000347, -0.00001099, 0.00000014, 0.00000303, 0.00000002, 0.00000478, -0.00000355, 0.00000115, -0.00000476, -0.00000376, 0.00000570, 0.00000169, -0.00000178, -0.00000230, -0.00000121, -0.00000165, -0.00000076, 0.00000272, -0.00000075, 0.00000260, 0.00000327, -0.00000303, -0.00000283, -0.00000344, 0.00000314, 0.00000520, -0.00000405, -0.00000251, },
		{ -0.00000333, -0.00000142, -0.00000319, 0.00000176, 0.00000117, 0.00000406, 0.00000179, -0.00001015, 0.00000015, 0.00000111, -0.00000149, 0.00000284, -0.00000372, 0.00000188, 0.00000613, -0.00000056, -0.00000262, -0.00000486, 0.00000523, 0.00000298, -0.00000257, -0.00000680, 0.00000215, 0.00000415, -0.00000152, 0.00000114, 0.00000010, -0.00000195, -0.00000020, -0.00000839, 0.00001544, 0.00000987, -0.00002151, 0.00001401, -0.00000657, -0.00001627, 0.00001731, -0.00002378, 0.00000591, 0.00000760, -0.00001856, 0.00000974, -0.00002710, 0.00002263, -0.00003392, 0.00001562, -0.00002713, 0.00000849, -0.00001165, -0.00005535, 0.00002397, -0.00007697, -0.00004833, -0.00007647, -0.00091465, 0.00171075, 0.00058037, -0.00107601, 0.00045746, -0.00048184, 0.00029593, -0.00023632, 0.00013641, -0.00011046, 0.00000612, 0.00004645, -0.00008163, 0.00008682, -0.00011352, 0.00010791, -0.00010398, 0.00008006, -0.00007662, 0.00005650, -0.00001946, -0.00001646, 0.00001513, -0.00000772, -0.00001130, 0.00001595, -0.00000950, 0.00000631, -0.00000504, 0.00000243, -0.00000381, 0.00000404, -0.00000027, 0.00000392, -0.00000854, -0.00000046, -0.00000048, 0.00000697, -0.00000176, -0.00000820, -0.00000163, 0.00000410, 0.00001037, -0.00000639, 0.00000098, 0.00000328, -0.00000476, 0.00000401, -0.00000932, -0.00000297, 0.00000453, -0.00000070, 0.00000503, -0.00000278, 0.00000038, -0.00000331, -0.00000532, 0.00000505, 0.00000288, -0.00000171, -0.00000236, -0.00000155, -0.00000108, -0.00000129, 0.00000248, 0.00000002, 0.00000161, 0.00000363, -0.00000240, -0.00000279, -0.00000345, 0.00000143, 0.00000610, -0.00000281, },
		{ -0.00000116, -0.00000410, -0.00000116, -0.00000349, 0.00000128, 0.00000155, 0.00000286, 0.00000388, -0.00000964, -0.00000190, 0.00000213, -0.00000255, 0.00000345, -0.00000322, 0.00000000, 0.00000751, -0.00000025, -0.00000197, -0.00000564, 0.00000375, 0.00000431, -0.00000185, -0.00000689, 0.00000084, 0.00000450, -0.00000119, 0.00000069, 0.00000075, -0.00000231, 0.00000060, -0.00000863, 0.00001133, 0.00001533, -0.00002211, 0.00000905, 0.00000175, -0.00002266, 0.00001796, -0.00001841, -0.00000221, 0.00001579, -0.00002553, 0.00001514, -0.00002951, 0.00002127, -0.00002853, 0.00000622, -0.00001360, -0.00000533, -0.00000032, -0.00005842, 0.00001327, -0.00003802, -0.00011944, 0.00007286, -0.00107994, 0.00157317, 0.00091734, -0.00124473, 0.00049797, -0.00048222, 0.00026190, -0.00018767, 0.00007941, -0.00005210, -0.00004944, 0.00009195, -0.00011550, 0.00010884, -0.00012426, 0.00010766, -0.00009543, 0.00006516, -0.00005754, 0.00003468, -0.00000061, -0.00002602, 0.00001379, -0.00000056, -0.00001780, 0.00001725, -0.00000764, 0.00000589, -0.00000515, 0.00000222, -0.00000354, 0.00000354, -0.00000037, 0.00000416, -0.00000712, -0.00000133, -0.00000123, 0.00000608, 0.00000097, -0.00000874, -0.00000206, 0.00000212, 0.00001094, -0.00000352, -0.00000217, 0.00000508, -0.00000508, 0.00000348, -0.00000683, -0.00000602, 0.00000537, -0.00000074, 0.00000467, -0.00000160, -0.00000056, -0.00000198, -0.00000629, 0.00000382, 0.00000411, -0.00000151, -0.00000239, -0.00000193, -0.00000070, -0.00000148, 0.00000196, 0.00000082, 0.00000088, 0.00000362, -0.00000162, -0.00000281, -0.00000326, -0.00000015, 0.00000637, },
		{ 0.00000602, 0.00000070, -0.00000461, -0.00000121, -0.00000344, 0.00000049, 0.00000214, 0.00000166, 0.00000534, -0.00000826, -0.00000422, 0.00000299, -0.00000326, 0.00000345, -0.00000205, -0.00000216, 0.00000855, 0.00000042, -0.00000150, -0.00000591, 0.00000189, 0.00000541, -0.00000096, -0.00000682, -0.00000039, 0.00000449, -0.00000064, 0.00000017, 0.00000130, -0.00000247, 0.00000096, -0.00000803, 0.00000694, 0.00001929, -0.00002018, 0.00000232, 0.00000965, -0.00002619, 0.00001477, -0.00001012, -0.00001141, 0.00002343, -0.00003059, 0.00001746, -0.00002804, 0.00001598, -0.00001982, -0.00000506, -0.00000088, -0.00001465, 0.00000252, -0.00004910, -0.00001322, 0.00001728, -0.00020242, 0.00021265, -0.00118507, 0.00135427, 0.00128028, -0.00138307, 0.00049823, -0.00044806, 0.00020070, -0.00012046, 0.00001180, 0.00000974, -0.00010231, 0.00013014, -0.00013945, 0.00011991, -0.00012431, 0.00009812, -0.00007982, 0.00004592, -0.00003712, 0.00001497, 0.00001287, -0.00002887, 0.00000724, 0.00000852, -0.00002336, 0.00001672, -0.00000483, 0.00000525, -0.00000501, 0.00000175, -0.00000312, 0.00000301, -0.00000041, 0.00000411, -0.00000555, -0.00000202, -0.00000185, 0.00000479, 0.00000348, -0.00000872, -0.00000268, 0.00000064, 0.00001043, -0.00000013, -0.00000484, 0.00000589, -0.00000438, 0.00000210, -0.00000395, -0.00000858, 0.00000531, -0.00000012, 0.00000384, -0.00000021, -0.00000143, -0.00000098, -0.00000661, 0.00000213, 0.00000520, -0.00000107, -0.00000243, -0.00000225, -0.00000054, -0.00000138, 0.00000132, 0.00000150, 0.00000049, 0.00000327, -0.00000080, -0.00000281, -0.00000297, -0.00000147, },
		{ -0.00000244, 0.00000514, 0.00000254, -0.00000470, -0.00000161, -0.00000309, -0.00000048, 0.00000274, 0.00000070, 0.00000604, -0.00000616, -0.00000653, 0.00000346, -0.00000345, 0.00000278, -0.00000034, -0.00000431, 0.00000901, 0.00000151, -0.00000125, -0.00000572, -0.00000016, 0.00000614, 0.00000013, -0.00000662, -0.00000146, 0.00000412, 0.00000005, -0.00000029, 0.00000169, -0.00000237, 0.00000091, -0.00000686, 0.00000282, 0.00002138, -0.00001604, -0.00000510, 0.00001578, -0.00002608, 0.00000795, -0.00000002, -0.00002023, 0.00002909, -0.00003263, 0.00001607, -0.00002279, 0.00000766, -0.00000951, -0.00001576, 0.00000807, -0.00001659, -0.00000516, -0.00002698, -0.00005318, 0.00008272, -0.00028606, 0.00032773, -0.00122049, 0.00106256, 0.00164737, -0.00147269, 0.00045004, -0.00037838, 0.00011557, -0.00004023, -0.00006005, 0.00006888, -0.00014722, 0.00015719, -0.00015129, 0.00011946, -0.00011445, 0.00008118, -0.00005977, 0.00002528, -0.00001824, -0.00000027, 0.00001981, -0.00002541, -0.00000288, 0.00001758, -0.00002667, 0.00001396, -0.00000121, 0.00000458, -0.00000475, 0.00000115, -0.00000266, 0.00000249, -0.00000038, 0.00000381, -0.00000393, -0.00000248, -0.00000235, 0.00000326, 0.00000558, -0.00000807, -0.00000350, -0.00000035, 0.00000907, 0.00000333, -0.00000663, 0.00000559, -0.00000281, 0.00000019, -0.00000111, -0.00001029, 0.00000429, 0.00000105, 0.00000276, 0.00000113, -0.00000201, -0.00000044, -0.00000630, 0.00000020, 0.00000595, -0.00000035, -0.00000249, -0.00000249, -0.00000061, -0.00000110, 0.00000069, 0.00000194, 0.00000045, 0.00000271, -0.00000004, -0.00000272, -0.00000270, },
		{ -0.00000249, -0.00000303, 0.00000388, 0.00000415, -0.00000427, -0.00000233, -0.00000254, -0.00000147, 0.00000315, 0.00000013, 0.00000597, -0.00000357, -0.00000848, 0.00000330, -0.00000303, 0.00000151, 0.00000166, -0.00000612, 0.00000871, 0.00000302, -0.00000117, -0.00000515, -0.00000221, 0.00000638, 0.00000140, -0.00000632, -0.00000231, 0.00000343, 0.00000076, -0.00000060, 0.00000185, -0.00000203, 0.00000057, -0.00000545, -0.00000060, 0.00002153, -0.00001039, -0.00001200, 0.00001909, -0.00002208, -0.00000164, 0.00001038, -0.00002711, 0.00003161, -0.00003100, 0.00001102, -0.00001460, -0.00000209, 0.00000016, -0.00002322, 0.00001065, -0.00000929, -0.00002379, 0.00000624, -0.00010203, 0.00015024, -0.00035883, 0.00040573, -0.00118300, 0.00071328, 0.00199416, -0.00149744, 0.00034953, -0.00027636, 0.00001293, 0.00004554, -0.00012898, 0.00011927, -0.00017977, 0.00017063, -0.00015040, 0.00010835, -0.00009670, 0.00005957, -0.00003829, 0.00000610, -0.00000329, -0.00000957, 0.00002012, -0.00001697, -0.00001445, 0.00002479, -0.00002687, 0.00000906, 0.00000278, 0.00000417, -0.00000452, 0.00000053, -0.00000224, 0.00000203, -0.00000029, 0.00000330, -0.00000238, -0.00000267, -0.00000274, 0.00000161, 0.00000712, -0.00000681, -0.00000447, -0.00000091, 0.00000716, 0.00000640, -0.00000729, 0.00000424, -0.00000065, -0.00000185, 0.00000128, -0.00001092, 0.00000237, 0.00000251, 0.00000170, 0.00000219, -0.00000215, -0.00000039, -0.00000550, -0.00000173, 0.00000622, 0.00000065, -0.00000254, -0.00000264, -0.00000085, -0.00000075, 0.00000022, 0.00000209, 0.00000069, 0.00000208, 0.00000054, -0.00000250, },
		{ -0.00000213, -0.00000235, -0.00000330, 0.00000243, 0.00000532, -0.00000329, -0.00000326, -0.00000196, -0.00000229, 0.00000323, 0.00000005, 0.00000526, -0.00000084, -0.00000979, 0.00000240, -0.00000202, -0.00000018, 0.00000363, -0.00000730, 0.00000756, 0.00000484, -0.00000116, -0.00000434, -0.00000407, 0.00000603, 0.00000283, -0.00000592, -0.00000292, 0.00000251, 0.00000137, -0.00000070, 0.00000179, -0.00000152, 0.00000007, -0.00000404, -0.00000308, 0.00001998, -0.00000411, -0.00001727, 0.00001895, -0.00001463, -0.00001261, 0.00001937, -0.00003072, 0.00003030, -0.00002575, 0.00000313, -0.00000500, -0.00001117, 0.00000682, -0.00002512, 0.00000511, 0.00000773, -0.00005203, 0.00004694, -0.00015353, 0.00021106, -0.00041037, 0.00043846, -0.00107619, 0.00032720, 0.00229591, -0.00144546, 0.00019816, -0.00014937, -0.00009824, 0.00012828, -0.00018795, 0.00015585, -0.00019709, 0.00016971, -0.00013775, 0.00008878, -0.00007396, 0.00003645, -0.00001832, -0.00000921, 0.00000617, -0.00001251, 0.00001472, -0.00000555, -0.00002529, 0.00002871, -0.00002369, 0.00000255, 0.00000651, 0.00000437, -0.00000446, -0.00000000, -0.00000192, 0.00000161, -0.00000014, 0.00000264, -0.00000096, -0.00000258, -0.00000306, -0.00000000, 0.00000801, -0.00000503, -0.00000545, -0.00000121, 0.00000505, 0.00000873, -0.00000677, 0.00000208, 0.00000171, -0.00000362, 0.00000290, -0.00001044, -0.00000019, 0.00000393, 0.00000091, 0.00000281, -0.00000182, -0.00000076, -0.00000441, -0.00000342, 0.00000590, 0.00000187, -0.00000254, -0.00000274, -0.00000119, -0.00000048, -0.00000002, 0.00000198, 0.00000109, 0.00000150, 0.00000087, },
		{ 0.00000094, -0.00000167, -0.00000227, -0.00000334, 0.00000098, 0.00000593, -0.00000183, -0.00000423, -0.00000152, -0.00000280, 0.00000289, 0.00000042, 0.00000411, 0.00000170, -0.00001022, 0.00000075, -0.00000059, -0.00000200, 0.00000523, -0.00000762, 0.00000561, 0.00000676, -0.00000104, -0.00000347, -0.00000555, 0.00000509, 0.00000432, -0.00000540, -0.00000331, 0.00000144, 0.00000178, -0.00000055, 0.00000156, -0.00000091, -0.00000044, -0.00000284, -0.00000457, 0.00001720, 0.00000184, -0.00002006, 0.00001531, -0.00000474, -0.00002327, 0.00002537, -0.00003019, 0.00002517, -0.00001766, -0.00000615, 0.00000403, -0.00001736, 0.00000835, -0.00001988, -0.00000905, 0.00003338, -0.00008683, 0.00009003, -0.00020064, 0.00025688, -0.00043287, 0.00042289, -0.00091002, -0.00007133, 0.00252998, -0.00131076, 0.00000307, -0.00000835, -0.00020742, 0.00019945, -0.00023100, 0.00017528, -0.00019822, 0.00015547, -0.00011577, 0.00006390, -0.00004961, 0.00001493, -0.00000234, -0.00001908, 0.00000959, -0.00000968, 0.00000530, 0.00000655, -0.00003343, 0.00002853, -0.00001746, -0.00000467, 0.00000925, 0.00000544, -0.00000462, -0.00000044, -0.00000167, 0.00000120, 0.00000010, 0.00000189, 0.00000023, -0.00000220, -0.00000329, -0.00000150, 0.00000825, -0.00000289, -0.00000631, -0.00000140, 0.00000303, 0.00001009, -0.00000516, -0.00000052, 0.00000387, -0.00000481, 0.00000361, -0.00000895, -0.00000306, 0.00000498, 0.00000061, 0.00000292, -0.00000105, -0.00000139, -0.00000327, -0.00000467, 0.00000500, 0.00000318, -0.00000238, -0.00000282, -0.00000155, -0.00000037, -0.00000001, 0.00000169, 0.00000151, 0.00000111, },
		{ 0.00000096, 0.00000079, -0.00000118, -0.00000219, -0.00000323, -0.00000030, 0.00000594, -0.00000005, -0.00000502, -0.00000138, -0.00000292, 0.00000216, 0.00000110, 0.00000281, 0.00000371, -0.00000966, -0.00000150, 0.00000101, -0.00000363, 0.00000618, -0.00000697, 0.00000303, 0.00000853, -0.00000067, -0.00000268, -0.00000654, 0.00000358, 0.00000576, -0.00000473, -0.00000350, 0.00000034, 0.00000190, -0.00000016, 0.00000120, -0.00000029, -0.00000089, -0.00000193, -0.00000520, 0.00001379, 0.00000659, -0.00001991, 0.00000868, 0.00000612, -0.00003187, 0.00002721, -0.00002536, 0.00001695, -0.00000816, -0.00001491, 0.00001036, -0.00001870, 0.00000334, -0.00000701, -0.00003097, 0.00006504, -0.00012384, 0.00012965, -0.00023652, 0.00028109, -0.00042208, 0.00036132, -0.00069971, -0.00045672, 0.00267826, -0.00109438, -0.00022327, 0.00013326, -0.00030383, 0.00025165, -0.00025406, 0.00017631, -0.00018422, 0.00013059, -0.00008797, 0.00003737, -0.00002692, -0.00000242, 0.00000801, -0.00002292, 0.00000740, -0.00000249, -0.00000596, 0.00001712, -0.00003752, 0.00002417, -0.00000911, -0.00001152, 0.00001040, 0.00000748, -0.00000494, -0.00000084, -0.00000146, 0.00000077, 0.00000042, 0.00000113, 0.00000112, -0.00000156, -0.00000346, -0.00000279, 0.00000787, -0.00000060, -0.00000686, -0.00000164, 0.00000134, 0.00001044, -0.00000275, -0.00000316, 0.00000547, -0.00000521, 0.00000341, -0.00000672, -0.00000584, 0.00000539, 0.00000086, 0.00000256, -0.00000000, -0.00000209, -0.00000229, -0.00000535, 0.00000359, 0.00000442, -0.00000199, -0.00000293, -0.00000185, -0.00000046, 0.00000019, 0.00000133, 0.00000182, },
		{ 0.00000194, 0.00000106, 0.00000050, -0.00000074, -0.00000206, -0.00000307, -0.00000130, 0.00000540, 0.00000183, -0.00000544, -0.00000163, -0.00000266, 0.00000114, 0.00000189, 0.00000165, 0.00000498, -0.00000814, -0.00000411, 0.00000245, -0.00000478, 0.00000628, -0.00000542, 0.00000012, 0.00000985, 0.00000009, -0.00000205, -0.00000697, 0.00000163, 0.00000700, -0.00000387, -0.00000354, -0.00000069, 0.00000169, 0.00000042, 0.00000080, 0.00000028, -0.00000122, -0.00000129, -0.00000523, 0.00001038, 0.00000954, -0.00001690, 0.00000009, 0.00001629, -0.00003693, 0.00002437, -0.00001682, 0.00000709, 0.00000088, -0.00002111, 0.00001211, -0.00001385, -0.00000866, 0.00001274, -0.00005838, 0.00009894, -0.00015799, 0.00016003, -0.00025553, 0.00027964, -0.00037780, 0.00026104, -0.00046405, -0.00080464, 0.00272904, -0.00080464, -0.00046405, 0.00026104, -0.00037780, 0.00027964, -0.00025553, 0.00016003, -0.00015799, 0.00009894, -0.00005838, 0.00001274, -0.00000866, -0.00001385, 0.00001211, -0.00002111, 0.00000088, 0.00000709, -0.00001682, 0.00002437, -0.00003693, 0.00001629, 0.00000009, -0.00001690, 0.00000954, 0.00001038, -0.00000523, -0.00000129, -0.00000122, 0.00000028, 0.00000080, 0.00000042, 0.00000169, -0.00000069, -0.00000354, -0.00000387, 0.00000700, 0.00000163, -0.00000697, -0.00000205, 0.00000009, 0.00000985, 0.00000012, -0.00000542, 0.00000628, -0.00000478, 0.00000245, -0.00000411, -0.00000814, 0.00000498, 0.00000165, 0.00000189, 0.00000114, -0.00000266, -0.00000163, -0.00000544, 0.00000183, 0.00000540, -0.00000130, -0.00000307, -0.00000206, -0.00000074, 0.00000050, 0.00000106, },
		{ 0.00000096, 0.00000182, 0.00000133, 0.00000019, -0.00000046, -0.00000185, -0.00000293, -0.00000199, 0.00000442, 0.00000359, -0.00000535, -0.00000229, -0.00000209, -0.00000000, 0.00000256, 0.00000086, 0.00000539, -0.00000584, -0.00000672, 0.00000341, -0.00000521, 0.00000547, -0.00000316, -0.00000275, 0.00001044, 0.00000134, -0.00000164, -0.00000686, -0.00000060, 0.00000787, -0.00000279, -0.00000346, -0.00000156, 0.00000112, 0.00000113, 0.00000042, 0.00000077, -0.00000146, -0.00000084, -0.00000494, 0.00000748, 0.00001040, -0.00001152, -0.00000911, 0.00002417, -0.00003752, 0.00001712, -0.00000596, -0.00000249, 0.00000740, -0.00002292, 0.00000801, -0.00000242, -0.00002692, 0.00003737, -0.00008797, 0.00013059, -0.00018422, 0.00017631, -0.00025406, 0.00025165, -0.00030383, 0.00013326, -0.00022327, -0.00109438, 0.00267826, -0.00045672, -0.00069971, 0.00036132, -0.00042208, 0.00028109, -0.00023652, 0.00012965, -0.00012384, 0.00006504, -0.00003097, -0.00000701, 0.00000334, -0.00001870, 0.00001036, -0.00001491, -0.00000816, 0.00001695, -0.00002536, 0.00002721, -0.00003187, 0.00000612, 0.00000868, -0.00001991, 0.00000659, 0.00001379, -0.00000520, -0.00000193, -0.00000089, -0.00000029, 0.00000120, -0.00000016, 0.00000190, 0.00000034, -0.00000350, -0.00000473, 0.00000576, 0.00000358, -0.00000654, -0.00000268, -0.00000067, 0.00000853, 0.00000303, -0.00000697, 0.00000618, -0.00000363, 0.00000101, -0.00000150, -0.00000966, 0.00000371, 0.00000281, 0.00000110, 0.00000216, -0.00000292, -0.00000138, -0.00000502, -0.00000005, 0.00000594, -0.00000030, -0.00000323, -0.00000219, -0.00000118, 0.00000079, },
		{ 0.00000094, 0.00000111, 0.00000151, 0.00000169, -0.00000001, -0.00000037, -0.00000155, -0.00000282, -0.00000238, 0.00000318, 0.00000500, -0.00000467, -0.00000327, -0.00000139, -0.00000105, 0.00000292, 0.00000061, 0.00000498, -0.00000306, -0.00000895, 0.00000361, -0.00000481, 0.00000387, -0.00000052, -0.00000516, 0.00001009, 0.00000303, -0.00000140, -0.00000631, -0.00000289, 0.00000825, -0.00000150, -0.00000329, -0.00000220, 0.00000023, 0.00000189, 0.00000010, 0.00000120, -0.00000167, -0.00000044, -0.00000462, 0.00000544, 0.00000925, -0.00000467, -0.00001746, 0.00002853, -0.00003343, 0.00000655, 0.00000530, -0.00000968, 0.00000959, -0.00001908, -0.00000234, 0.00001493, -0.00004961, 0.00006390, -0.00011577, 0.00015547, -0.00019822, 0.00017528, -0.00023100, 0.00019945, -0.00020742, -0.00000835, 0.00000307, -0.00131076, 0.00252998, -0.00007133, -0.00091002, 0.00042289, -0.00043287, 0.00025688, -0.00020064, 0.00009003, -0.00008683, 0.00003338, -0.00000905, -0.00001988, 0.00000835, -0.00001736, 0.00000403, -0.00000615, -0.00001766, 0.00002517, -0.00003019, 0.00002537, -0.00002327, -0.00000474, 0.00001531, -0.00002006, 0.00000184, 0.00001720, -0.00000457, -0.00000284, -0.00000044, -0.00000091, 0.00000156, -0.00000055, 0.00000178, 0.00000144, -0.00000331, -0.00000540, 0.00000432, 0.00000509, -0.00000555, -0.00000347, -0.00000104, 0.00000676, 0.00000561, -0.00000762, 0.00000523, -0.00000200, -0.00000059, 0.00000075, -0.00001022, 0.00000170, 0.00000411, 0.00000042, 0.00000289, -0.00000280, -0.00000152, -0.00000423, -0.00000183, 0.00000593, 0.00000098, -0.00000334, -0.00000227, -0.00000167, },
		{ -0.00000213, 0.00000087, 0.00000150, 0.00000109, 0.00000198, -0.00000002, -0.00000048, -0.00000119, -0.00000274, -0.00000254, 0.00000187, 0.00000590, -0.00000342, -0.00000441, -0.00000076, -0.00000182, 0.00000281, 0.00000091, 0.00000393, -0.00000019, -0.00001044, 0.00000290, -0.00000362, 0.00000171, 0.00000208, -0.00000677, 0.00000873, 0.00000505, -0.00000121, -0.00000545, -0.00000503, 0.00000801, -0.00000000, -0.00000306, -0.00000258, -0.00000096, 0.00000264, -0.00000014, 0.00000161, -0.00000192, -0.00000000, -0.00000446, 0.00000437, 0.00000651, 0.00000255, -0.00002369, 0.00002871, -0.00002529, -0.00000555, 0.00001472, -0.00001251, 0.00000617, -0.00000921, -0.00001832, 0.00003645, -0.00007396, 0.00008878, -0.00013775, 0.00016971, -0.00019709, 0.00015585, -0.00018795, 0.00012828, -0.00009824, -0.00014937, 0.00019816, -0.00144546, 0.00229591, 0.00032720, -0.00107619, 0.00043846, -0.00041037, 0.00021106, -0.00015353, 0.00004694, -0.00005203, 0.00000773, 0.00000511, -0.00002512, 0.00000682, -0.00001117, -0.00000500, 0.00000313, -0.00002575, 0.00003030, -0.00003072, 0.00001937, -0.00001261, -0.00001463, 0.00001895, -0.00001727, -0.00000411, 0.00001998, -0.00000308, -0.00000404, 0.00000007, -0.00000152, 0.00000179, -0.00000070, 0.00000137, 0.00000251, -0.00000292, -0.00000592, 0.00000283, 0.00000603, -0.00000407, -0.00000434, -0.00000116, 0.00000484, 0.00000756, -0.00000730, 0.00000363, -0.00000018, -0.00000202, 0.00000240, -0.00000979, -0.00000084, 0.00000526, 0.00000005, 0.00000323, -0.00000229, -0.00000196, -0.00000326, -0.00000329, 0.00000532, 0.00000243, -0.00000330, -0.00000235, },
		{ -0.00000249, -0.00000250, 0.00000054, 0.00000208, 0.00000069, 0.00000209, 0.00000022, -0.00000075, -0.00000085, -0.00000264, -0.00000254, 0.00000065, 0.00000622, -0.00000173, -0.00000550, -0.00000039, -0.00000215, 0.00000219, 0.00000170, 0.00000251, 0.00000237, -0.00001092, 0.00000128, -0.00000185, -0.00000065, 0.00000424, -0.00000729, 0.00000640, 0.00000716, -0.00000091, -0.00000447, -0.00000681, 0.00000712, 0.00000161, -0.00000274, -0.00000267, -0.00000238, 0.00000330, -0.00000029, 0.00000203, -0.00000224, 0.00000053, -0.00000452, 0.00000417, 0.00000278, 0.00000906, -0.00002687, 0.00002479, -0.00001445, -0.00001697, 0.00002012, -0.00000957, -0.00000329, 0.00000610, -0.00003829, 0.00005957, -0.00009670, 0.00010835, -0.00015040, 0.00017063, -0.00017977, 0.00011927, -0.00012898, 0.00004554, 0.00001293, -0.00027636, 0.00034953, -0.00149744, 0.00199416, 0.00071328, -0.00118300, 0.00040573, -0.00035883, 0.00015024, -0.00010203, 0.00000624, -0.00002379, -0.00000929, 0.00001065, -0.00002322, 0.00000016, -0.00000209, -0.00001460, 0.00001102, -0.00003100, 0.00003161, -0.00002711, 0.00001038, -0.00000164, -0.00002208, 0.00001909, -0.00001200, -0.00001039, 0.00002153, -0.00000060, -0.00000545, 0.00000057, -0.00000203, 0.00000185, -0.00000060, 0.00000076, 0.00000343, -0.00000231, -0.00000632, 0.00000140, 0.00000638, -0.00000221, -0.00000515, -0.00000117, 0.00000302, 0.00000871, -0.00000612, 0.00000166, 0.00000151, -0.00000303, 0.00000330, -0.00000848, -0.00000357, 0.00000597, 0.00000013, 0.00000315, -0.00000147, -0.00000254, -0.00000233, -0.00000427, 0.00000415, 0.00000388, -0.00000303, },
		{ -0.00000244, -0.00000270, -0.00000272, -0.00000004, 0.00000271, 0.00000045, 0.00000194, 0.00000069, -0.00000110, -0.00000061, -0.00000249, -0.00000249, -0.00000035, 0.00000595, 0.00000020, -0.00000630, -0.00000044, -0.00000201, 0.00000113, 0.00000276, 0.00000105, 0.00000429, -0.00001029, -0.00000111, 0.00000019, -0.00000281, 0.00000559, -0.00000663, 0.00000333, 0.00000907, -0.00000035, -0.00000350, -0.00000807, 0.00000558, 0.00000326, -0.00000235, -0.00000248, -0.00000393, 0.00000381, -0.00000038, 0.00000249, -0.00000266, 0.00000115, -0.00000475, 0.00000458, -0.00000121, 0.00001396, -0.00002667, 0.00001758, -0.00000288, -0.00002541, 0.00001981, -0.00000027, -0.00001824, 0.00002528, -0.00005977, 0.00008118, -0.00011445, 0.00011946, -0.00015129, 0.00015719, -0.00014722, 0.00006888, -0.00006005, -0.00004023, 0.00011557, -0.00037838, 0.00045004, -0.00147269, 0.00164737, 0.00106256, -0.00122049, 0.00032773, -0.00028606, 0.00008272, -0.00005318, -0.00002698, -0.00000516, -0.00001659, 0.00000807, -0.00001576, -0.00000951, 0.00000766, -0.00002279, 0.00001607, -0.00003263, 0.00002909, -0.00002023, -0.00000002, 0.00000795, -0.00002608, 0.00001578, -0.00000510, -0.00001604, 0.00002138, 0.00000282, -0.00000686, 0.00000091, -0.00000237, 0.00000169, -0.00000029, 0.00000005, 0.00000412, -0.00000146, -0.00000662, 0.00000013, 0.00000614, -0.00000016, -0.00000572, -0.00000125, 0.00000151, 0.00000901, -0.00000431, -0.00000034, 0.00000278, -0.00000345, 0.00000346, -0.00000653, -0.00000616, 0.00000604, 0.00000070, 0.00000274, -0.00000048, -0.00000309, -0.00000161, -0.00000470, 0.00000254, 0.00000514, },
		{ 0.00000602, -0.00000147, -0.00000297, -0.00000281, -0.00000080, 0.00000327, 0.00000049, 0.00000150, 0.00000132, -0.00000138, -0.00000054, -0.00000225, -0.00000243, -0.00000107, 0.00000520, 0.00000213, -0.00000661, -0.00000098, -0.00000143, -0.00000021, 0.00000384, -0.00000012, 0.00000531, -0.00000858, -0.00000395, 0.00000210, -0.00000438, 0.00000589, -0.00000484, -0.00000013, 0.00001043, 0.00000064, -0.00000268, -0.00000872, 0.00000348, 0.00000479, -0.00000185, -0.00000202, -0.00000555, 0.00000411, -0.00000041, 0.00000301, -0.00000312, 0.00000175, -0.00000501, 0.00000525, -0.00000483, 0.00001672, -0.00002336, 0.00000852, 0.00000724, -0.00002887, 0.00001287, 0.00001497, -0.00003712, 0.00004592, -0.00007982, 0.00009812, -0.00012431, 0.00011991, -0.00013945, 0.00013014, -0.00010231, 0.00000974, 0.00001180, -0.00012046, 0.00020070, -0.00044806, 0.00049823, -0.00138307, 0.00128028, 0.00135427, -0.00118507, 0.00021265, -0.00020242, 0.00001728, -0.00001322, -0.00004910, 0.00000252, -0.00001465, -0.00000088, -0.00000506, -0.00001982, 0.00001598, -0.00002804, 0.00001746, -0.00003059, 0.00002343, -0.00001141, -0.00001012, 0.00001477, -0.00002619, 0.00000965, 0.00000232, -0.00002018, 0.00001929, 0.00000694, -0.00000803, 0.00000096, -0.00000247, 0.00000130, 0.00000017, -0.00000064, 0.00000449, -0.00000039, -0.00000682, -0.00000096, 0.00000541, 0.00000189, -0.00000591, -0.00000150, 0.00000042, 0.00000855, -0.00000216, -0.00000205, 0.00000345, -0.00000326, 0.00000299, -0.00000422, -0.00000826, 0.00000534, 0.00000166, 0.00000214, 0.00000049, -0.00000344, -0.00000121, -0.00000461, 0.00000070, },
		{ -0.00000116, 0.00000637, -0.00000015, -0.00000326, -0.00000281, -0.00000162, 0.00000362, 0.00000088, 0.00000082, 0.00000196, -0.00000148, -0.00000070, -0.00000193, -0.00000239, -0.00000151, 0.00000411, 0.00000382, -0.00000629, -0.00000198, -0.00000056, -0.00000160, 0.00000467, -0.00000074, 0.00000537, -0.00000602, -0.00000683, 0.00000348, -0.00000508, 0.00000508, -0.00000217, -0.00000352, 0.00001094, 0.00000212, -0.00000206, -0.00000874, 0.00000097, 0.00000608, -0.00000123, -0.00000133, -0.00000712, 0.00000416, -0.00000037, 0.00000354, -0.00000354, 0.00000222, -0.00000515, 0.00000589, -0.00000764, 0.00001725, -0.00001780, -0.00000056, 0.00001379, -0.00002602, -0.00000061, 0.00003468, -0.00005754, 0.00006516, -0.00009543, 0.00010766, -0.00012426, 0.00010884, -0.00011550, 0.00009195, -0.00004944, -0.00005210, 0.00007941, -0.00018767, 0.00026190, -0.00048222, 0.00049797, -0.00124473, 0.00091734, 0.00157317, -0.00107994, 0.00007286, -0.00011944, -0.00003802, 0.00001327, -0.00005842, -0.00000032, -0.00000533, -0.00001360, 0.00000622, -0.00002853, 0.00002127, -0.00002951, 0.00001514, -0.00002553, 0.00001579, -0.00000221, -0.00001841, 0.00001796, -0.00002266, 0.00000175, 0.00000905, -0.00002211, 0.00001533, 0.00001133, -0.00000863, 0.00000060, -0.00000231, 0.00000075, 0.00000069, -0.00000119, 0.00000450, 0.00000084, -0.00000689, -0.00000185, 0.00000431, 0.00000375, -0.00000564, -0.00000197, -0.00000025, 0.00000751, 0.00000000, -0.00000322, 0.00000345, -0.00000255, 0.00000213, -0.00000190, -0.00000964, 0.00000388, 0.00000286, 0.00000155, 0.00000128, -0.00000349, -0.00000116, -0.00000410, },
		{ -0.00000333, -0.00000281, 0.00000610, 0.00000143, -0.00000345, -0.00000279, -0.00000240, 0.00000363, 0.00000161, 0.00000002, 0.00000248, -0.00000129, -0.00000108, -0.00000155, -0.00000236, -0.00000171, 0.00000288, 0.00000505, -0.00000532, -0.00000331, 0.00000038, -0.00000278, 0.00000503, -0.00000070, 0.00000453, -0.00000297, -0.00000932, 0.00000401, -0.00000476, 0.00000328, 0.00000098, -0.00000639, 0.00001037, 0.00000410, -0.00000163, -0.00000820, -0.00000176, 0.00000697, -0.00000048, -0.00000046, -0.00000854, 0.00000392, -0.00000027, 0.00000404, -0.00000381, 0.00000243, -0.00000504, 0.00000631, -0.00000950, 0.00001595, -0.00001130, -0.00000772, 0.00001513, -0.00001646, -0.00001946, 0.00005650, -0.00007662, 0.00008006, -0.00010398, 0.00010791, -0.00011352, 0.00008682, -0.00008163, 0.00004645, 0.00000612, -0.00011046, 0.00013641, -0.00023632, 0.00029593, -0.00048184, 0.00045746, -0.00107601, 0.00058037, 0.00171075, -0.00091465, -0.00007647, -0.00004833, -0.00007697, 0.00002397, -0.00005535, -0.00001165, 0.00000849, -0.00002713, 0.00001562, -0.00003392, 0.00002263, -0.00002710, 0.00000974, -0.00001856, 0.00000760, 0.00000591, -0.00002378, 0.00001731, -0.00001627, -0.00000657, 0.00001401, -0.00002151, 0.00000987, 0.00001544, -0.00000839, -0.00000020, -0.00000195, 0.00000010, 0.00000114, -0.00000152, 0.00000415, 0.00000215, -0.00000680, -0.00000257, 0.00000298, 0.00000523, -0.00000486, -0.00000262, -0.00000056, 0.00000613, 0.00000188, -0.00000372, 0.00000284, -0.00000149, 0.00000111, 0.00000015, -0.00001015, 0.00000179, 0.00000406, 0.00000117, 0.00000176, -0.00000319, -0.00000142, },
		{ -0.00000185, -0.00000251, -0.00000405, 0.00000520, 0.00000314, -0.00000344, -0.00000283, -0.00000303, 0.00000327, 0.00000260, -0.00000075, 0.00000272, -0.00000076, -0.00000165, -0.00000121, -0.00000230, -0.00000178, 0.00000169, 0.00000570, -0.00000376, -0.00000476, 0.00000115, -0.00000355, 0.00000478, 0.00000002, 0.00000303, 0.00000014, -0.00001099, 0.00000347, -0.00000348, 0.00000078, 0.00000412, -0.00000829, 0.00000860, 0.00000643, -0.00000130, -0.00000720, -0.00000448, 0.00000734, 0.00000038, 0.00000056, -0.00000973, 0.00000335, -0.00000009, 0.00000442, -0.00000383, 0.00000233, -0.00000466, 0.00000650, -0.00001058, 0.00001358, -0.00000537, -0.00001130, 0.00001040, -0.00000087, -0.00004150, 0.00007748, -0.00009138, 0.00008808, -0.00010370, 0.00009815, -0.00009267, 0.00005588, -0.00004125, -0.00000166, 0.00005886, -0.00015976, 0.00017802, -0.00026339, 0.00030287, -0.00045158, 0.00038788, -0.00089540, 0.00028667, 0.00176580, -0.00070397, -0.00021916, 0.00000155, -0.00009604, 0.00001908, -0.00004220, -0.00002825, 0.00002349, -0.00003864, 0.00002119, -0.00003505, 0.00001997, -0.00002146, 0.00000241, -0.00001106, 0.00000022, 0.00001181, -0.00002568, 0.00001321, -0.00000824, -0.00001399, 0.00001642, -0.00001844, 0.00000353, 0.00001869, -0.00000712, -0.00000138, -0.00000148, -0.00000053, 0.00000144, -0.00000157, 0.00000351, 0.00000343, -0.00000649, -0.00000314, 0.00000156, 0.00000622, -0.00000360, -0.00000338, -0.00000067, 0.00000467, 0.00000324, -0.00000351, 0.00000177, -0.00000031, 0.00000018, 0.00000173, -0.00000980, -0.00000069, 0.00000499, 0.00000113, 0.00000189, -0.00000258, },
		{ -0.00000179, -0.00000231, -0.00000181, -0.00000478, 0.00000376, 0.00000477, -0.00000310, -0.00000298, -0.00000345, 0.00000254, 0.00000372, -0.00000129, 0.00000257, 0.00000010, -0.00000227, -0.00000100, -0.00000216, -0.00000179, 0.00000069, 0.00000574, -0.00000179, -0.00000606, 0.00000152, -0.00000376, 0.00000391, 0.00000126, 0.00000120, 0.00000286, -0.00001155, 0.00000186, -0.00000145, -0.00000200, 0.00000676, -0.00000890, 0.00000571, 0.00000889, -0.00000094, -0.00000590, -0.00000697, 0.00000712, 0.00000134, 0.00000166, -0.00001059, 0.00000245, 0.00000019, 0.00000465, -0.00000357, 0.00000193, -0.00000414, 0.00000666, -0.00001132, 0.00001115, -0.00000144, -0.00001014, -0.00000032, 0.00001903, -0.00006387, 0.00009452, -0.00009921, 0.00008748, -0.00009392, 0.00007896, -0.00006360, 0.00001915, 0.00000143, -0.00004750, 0.00010384, -0.00019580, 0.00020158, -0.00026861, 0.00028579, -0.00039879, 0.00030172, -0.00071961, 0.00004779, 0.00174398, -0.00046627, -0.00033996, 0.00002387, -0.00009484, 0.00000117, -0.00002271, -0.00004632, 0.00003646, -0.00004591, 0.00002189, -0.00003185, 0.00001397, -0.00001375, -0.00000545, -0.00000440, -0.00000523, 0.00001481, -0.00002419, 0.00000662, 0.00000000, -0.00001933, 0.00001593, -0.00001335, -0.00000288, 0.00002056, -0.00000477, -0.00000280, -0.00000106, -0.00000101, 0.00000151, -0.00000134, 0.00000265, 0.00000456, -0.00000595, -0.00000359, 0.00000015, 0.00000668, -0.00000198, -0.00000412, -0.00000070, 0.00000333, 0.00000398, -0.00000270, 0.00000047, 0.00000075, -0.00000047, 0.00000272, -0.00000868, -0.00000327, 0.00000542, 0.00000151, 0.00000170, },
		{ 0.00000131, -0.00000094, -0.00000264, -0.00000138, -0.00000498, 0.00000195, 0.00000610, -0.00000235, -0.00000325, -0.00000365, 0.00000152, 0.00000478, -0.00000147, 0.00000199, 0.00000118, -0.00000281, -0.00000101, -0.00000189, -0.00000181, -0.00000005, 0.00000521, 0.00000034, -0.00000695, 0.00000134, -0.00000338, 0.00000253, 0.00000277, -0.00000057, 0.00000480, -0.00001086, -0.00000065, 0.00000095, -0.00000458, 0.00000845, -0.00000809, 0.00000192, 0.00001117, -0.00000037, -0.00000447, -0.00000905, 0.00000627, 0.00000233, 0.00000281, -0.00001106, 0.00000121, 0.00000053, 0.00000471, -0.00000307, 0.00000141, -0.00000375, 0.00000716, -0.00001232, 0.00000966, -0.00000062, -0.00000390, -0.00001590, 0.00004067, -0.00008335, 0.00010479, -0.00009831, 0.00007763, -0.00007521, 0.00005215, -0.00002927, -0.00001952, 0.00004205, -0.00008664, 0.00013726, -0.00021616, 0.00020670, -0.00025421, 0.00025011, -0.00033235, 0.00021122, -0.00056201, -0.00013102, 0.00165685, -0.00022139, -0.00042631, 0.00001623, -0.00007607, -0.00002531, -0.00000139, -0.00006215, 0.00004481, -0.00004765, 0.00001760, -0.00002507, 0.00000586, -0.00000544, -0.00001244, 0.00000032, -0.00000809, 0.00001483, -0.00001994, -0.00000122, 0.00000709, -0.00002182, 0.00001267, -0.00000703, -0.00000853, 0.00002068, -0.00000147, -0.00000419, -0.00000086, -0.00000125, 0.00000131, -0.00000088, 0.00000171, 0.00000543, -0.00000516, -0.00000393, -0.00000115, 0.00000660, -0.00000013, -0.00000468, -0.00000081, 0.00000227, 0.00000409, -0.00000149, -0.00000081, 0.00000150, -0.00000072, 0.00000315, -0.00000701, -0.00000564, 0.00000520, 0.00000230, },
		{ 0.00000336, 0.00000086, -0.00000022, -0.00000271, -0.00000126, -0.00000470, -0.00000000, 0.00000694, -0.00000117, -0.00000362, -0.00000367, 0.00000034, 0.00000560, -0.00000116, 0.00000103, 0.00000234, -0.00000311, -0.00000132, -0.00000150, -0.00000187, -0.00000050, 0.00000426, 0.00000236, -0.00000722, 0.00000056, -0.00000250, 0.00000086, 0.00000423, -0.00000191, 0.00000572, -0.00000898, -0.00000373, 0.00000326, -0.00000649, 0.00000891, -0.00000590, -0.00000238, 0.00001291, 0.00000053, -0.00000304, -0.00001055, 0.00000480, 0.00000329, 0.00000395, -0.00001109, -0.00000029, 0.00000087, 0.00000467, -0.00000250, 0.00000105, -0.00000386, 0.00000845, -0.00001419, 0.00000993, -0.00000346, 0.00000691, -0.00003437, 0.00006100, -0.00009693, 0.00010623, -0.00008797, 0.00005915, -0.00004936, 0.00002054, 0.00000675, -0.00005614, 0.00007661, -0.00011567, 0.00015692, -0.00022045, 0.00019519, -0.00022444, 0.00020268, -0.00026137, 0.00012693, -0.00043168, -0.00025056, 0.00152015, 0.00001145, -0.00046970, -0.00001966, -0.00004505, -0.00005482, 0.00001727, -0.00007274, 0.00004703, -0.00004368, 0.00000917, -0.00001611, -0.00000281, 0.00000202, -0.00001741, 0.00000245, -0.00000824, 0.00001229, -0.00001391, -0.00000893, 0.00001191, -0.00002118, 0.00000723, -0.00000045, -0.00001266, 0.00001893, 0.00000245, -0.00000524, -0.00000102, -0.00000121, 0.00000087, -0.00000029, 0.00000081, 0.00000597, -0.00000414, -0.00000416, -0.00000230, 0.00000605, 0.00000178, -0.00000493, -0.00000108, 0.00000157, 0.00000365, -0.00000010, -0.00000183, 0.00000182, -0.00000058, 0.00000310, -0.00000505, -0.00000752, 0.00000426, },
		{ 0.00000266, 0.00000451, 0.00000056, 0.00000025, -0.00000245, -0.00000146, -0.00000410, -0.00000186, 0.00000716, 0.00000040, -0.00000398, -0.00000359, -0.00000084, 0.00000602, -0.00000035, -0.00000019, 0.00000338, -0.00000301, -0.00000194, -0.00000101, -0.00000196, -0.00000071, 0.00000308, 0.00000401, -0.00000674, -0.00000077, -0.00000132, -0.00000081, 0.00000532, -0.00000256, 0.00000555, -0.00000618, -0.00000690, 0.00000500, -0.00000739, 0.00000803, -0.00000261, -0.00000670, 0.00001377, 0.00000186, -0.00000173, -0.00001141, 0.00000282, 0.00000412, 0.00000506, -0.00001069, -0.00000196, 0.00000110, 0.00000468, -0.00000209, 0.00000121, -0.00000490, 0.00001095, -0.00001741, 0.00001241, -0.00000982, 0.00002093, -0.00005310, 0.00007693, -0.00010217, 0.00009788, -0.00006877, 0.00003382, -0.00001910, -0.00001249, 0.00004073, -0.00008708, 0.00010208, -0.00013259, 0.00016240, -0.00021018, 0.00017059, -0.00018483, 0.00015071, -0.00019402, 0.00005672, -0.00033303, -0.00031682, 0.00135186, 0.00021563, -0.00046659, -0.00007819, -0.00000883, -0.00008165, 0.00002952, -0.00007624, 0.00004280, -0.00003485, -0.00000184, -0.00000667, -0.00001050, 0.00000744, -0.00001969, 0.00000186, -0.00000605, 0.00000802, -0.00000733, -0.00001528, 0.00001378, -0.00001765, 0.00000051, 0.00000540, -0.00001475, 0.00001545, 0.00000653, -0.00000565, -0.00000165, -0.00000090, 0.00000026, 0.00000033, 0.00000007, 0.00000613, -0.00000294, -0.00000425, -0.00000329, 0.00000512, 0.00000355, -0.00000479, -0.00000156, 0.00000120, 0.00000284, 0.00000121, -0.00000245, 0.00000170, -0.00000012, 0.00000274, -0.00000305, -0.00000873, },
		{ -0.00000917, 0.00000056, 0.00000551, 0.00000054, 0.00000040, -0.00000189, -0.00000186, -0.00000336, -0.00000340, 0.00000668, 0.00000224, -0.00000422, -0.00000351, -0.00000189, 0.00000593, 0.00000091, -0.00000147, 0.00000408, -0.00000243, -0.00000283, -0.00000052, -0.00000202, -0.00000077, 0.00000188, 0.00000510, -0.00000553, -0.00000245, -0.00000011, -0.00000217, 0.00000580, -0.00000239, 0.00000440, -0.00000285, -0.00000965, 0.00000577, -0.00000707, 0.00000592, 0.00000135, -0.00001051, 0.00001349, 0.00000364, -0.00000056, -0.00001159, 0.00000045, 0.00000473, 0.00000613, -0.00000992, -0.00000360, 0.00000102, 0.00000495, -0.00000218, 0.00000228, -0.00000724, 0.00001494, -0.00002217, 0.00001703, -0.00001886, 0.00003616, -0.00006923, 0.00008579, -0.00009769, 0.00008006, -0.00004249, 0.00000438, 0.00001228, -0.00004340, 0.00006932, -0.00010959, 0.00011667, -0.00013700, 0.00015498, -0.00018842, 0.00013759, -0.00014130, 0.00010082, -0.00013665, 0.00000520, -0.00026611, -0.00033961, 0.00117016, 0.00037868, -0.00041863, -0.00015072, 0.00002486, -0.00010088, 0.00003299, -0.00007223, 0.00003301, -0.00002286, -0.00001352, 0.00000153, -0.00001598, 0.00001011, -0.00001911, -0.00000107, -0.00000231, 0.00000305, -0.00000134, -0.00001935, 0.00001255, -0.00001192, -0.00000637, 0.00000964, -0.00001458, 0.00001066, 0.00001024, -0.00000516, -0.00000274, -0.00000040, -0.00000039, 0.00000085, -0.00000040, 0.00000593, -0.00000167, -0.00000416, -0.00000411, 0.00000393, 0.00000506, -0.00000421, -0.00000222, 0.00000108, 0.00000187, 0.00000223, -0.00000259, 0.00000120, 0.00000053, 0.00000224, -0.00000125, },
		{ 0.00000021, -0.00000886, -0.00000180, 0.00000613, 0.00000092, 0.00000023, -0.00000111, -0.00000234, -0.00000266, -0.00000447, 0.00000555, 0.00000415, -0.00000418, -0.00000354, -0.00000268, 0.00000532, 0.00000246, -0.00000259, 0.00000430, -0.00000136, -0.00000387, -0.00000015, -0.00000198, -0.00000079, 0.00000084, 0.00000552, -0.00000369, -0.00000421, 0.00000083, -0.00000300, 0.00000553, -0.00000144, 0.00000255, 0.00000050, -0.00001151, 0.00000532, -0.00000557, 0.00000290, 0.00000545, -0.00001331, 0.00001195, 0.00000576, 0.00000048, -0.00001116, -0.00000212, 0.00000498, 0.00000717, -0.00000890, -0.00000500, 0.00000043, 0.00000574, -0.00000308, 0.00000458, -0.00001106, 0.00002042, -0.00002826, 0.00002320, -0.00002917, 0.00005024, -0.00008009, 0.00008573, -0.00008335, 0.00005440, -0.00001192, -0.00002587, 0.00004140, -0.00006908, 0.00009000, -0.00012210, 0.00012006, -0.00013002, 0.00013732, -0.00015922, 0.00010122, -0.00009934, 0.00005827, -0.00009319, -0.00002631, -0.00022728, -0.00033092, 0.00099149, 0.00049323, -0.00033222, -0.00022675, 0.00004888, -0.00010910, 0.00002697, -0.00006170, 0.00001954, -0.00000986, -0.00002389, 0.00000711, -0.00001849, 0.00000988, -0.00001606, -0.00000555, 0.00000201, -0.00000156, 0.00000312, -0.00002070, 0.00000857, -0.00000501, -0.00001234, 0.00001169, -0.00001228, 0.00000515, 0.00001304, -0.00000363, -0.00000419, 0.00000015, -0.00000095, 0.00000115, -0.00000058, 0.00000541, -0.00000042, -0.00000386, -0.00000479, 0.00000258, 0.00000618, -0.00000322, -0.00000297, 0.00000110, 0.00000091, 0.00000283, -0.00000228, 0.00000046, 0.00000121, 0.00000178, },
		{ 0.00000147, 0.00000126, -0.00000793, -0.00000414, 0.00000618, 0.00000171, -0.00000018, -0.00000023, -0.00000273, -0.00000219, -0.00000498, 0.00000391, 0.00000590, -0.00000374, -0.00000372, -0.00000316, 0.00000424, 0.00000409, -0.00000331, 0.00000391, 0.00000011, -0.00000488, -0.00000002, -0.00000179, -0.00000085, 0.00000008, 0.00000528, -0.00000147, -0.00000574, 0.00000125, -0.00000313, 0.00000452, 0.00000009, 0.00000042, 0.00000337, -0.00001212, 0.00000360, -0.00000311, -0.00000057, 0.00000908, -0.00001473, 0.00000916, 0.00000803, 0.00000147, -0.00001023, -0.00000466, 0.00000476, 0.00000824, -0.00000781, -0.00000589, -0.00000094, 0.00000729, -0.00000506, 0.00000827, -0.00001634, 0.00002710, -0.00003510, 0.00002989, -0.00003897, 0.00006082, -0.00008365, 0.00007609, -0.00006038, 0.00002361, 0.00001954, -0.00005356, 0.00006527, -0.00008718, 0.00010139, -0.00012438, 0.00011330, -0.00011402, 0.00011302, -0.00012703, 0.00006622, -0.00006336, 0.00002642, -0.00006504, -0.00003940, -0.00021038, -0.00030322, 0.00082914, 0.00055737, -0.00021757, -0.00029536, 0.00005770, -0.00010500, 0.00001252, -0.00004681, 0.00000481, 0.00000194, -0.00003134, 0.00000923, -0.00001792, 0.00000714, -0.00001132, -0.00001059, 0.00000591, -0.00000497, 0.00000548, -0.00001932, 0.00000259, 0.00000191, -0.00001648, 0.00001133, -0.00000831, -0.00000033, 0.00001448, -0.00000108, -0.00000580, 0.00000057, -0.00000131, 0.00000118, -0.00000046, 0.00000469, 0.00000068, -0.00000331, -0.00000533, 0.00000118, 0.00000686, -0.00000189, -0.00000370, 0.00000112, 0.00000014, 0.00000296, -0.00000162, -0.00000035, 0.00000176, },
		{ 0.00000209, 0.00000136, 0.00000191, -0.00000655, -0.00000620, 0.00000555, 0.00000283, -0.00000068, 0.00000056, -0.00000288, -0.00000202, -0.00000496, 0.00000196, 0.00000726, -0.00000282, -0.00000407, -0.00000335, 0.00000282, 0.00000556, -0.00000345, 0.00000294, 0.00000182, -0.00000566, -0.00000025, -0.00000139, -0.00000101, -0.00000035, 0.00000450, 0.00000085, -0.00000674, 0.00000099, -0.00000257, 0.00000295, 0.00000188, -0.00000155, 0.00000533, -0.00001132, 0.00000078, -0.00000011, -0.00000393, 0.00001175, -0.00001454, 0.00000535, 0.00001017, 0.00000253, -0.00000893, -0.00000694, 0.00000395, 0.00000940, -0.00000687, -0.00000598, -0.00000327, 0.00000977, -0.00000825, 0.00001330, -0.00002274, 0.00003432, -0.00004176, 0.00003575, -0.00004640, 0.00006593, -0.00007876, 0.00005746, -0.00003116, -0.00000887, 0.00004837, -0.00007570, 0.00008170, -0.00009649, 0.00010334, -0.00011746, 0.00009857, -0.00009218, 0.00008598, -0.00009599, 0.00003637, -0.00003621, 0.00000658, -0.00005126, -0.00003784, -0.00020798, -0.00026796, 0.00069224, 0.00057425, -0.00008727, -0.00034668, 0.00004826, -0.00008949, -0.00000783, -0.00003045, -0.00000864, 0.00001066, -0.00003487, 0.00000776, -0.00001473, 0.00000272, -0.00000593, -0.00001517, 0.00000855, -0.00000664, 0.00000560, -0.00001569, -0.00000436, 0.00000774, -0.00001822, 0.00000874, -0.00000336, -0.00000505, 0.00001427, 0.00000231, -0.00000728, 0.00000067, -0.00000136, 0.00000092, -0.00000013, 0.00000389, 0.00000152, -0.00000253, -0.00000572, -0.00000020, 0.00000709, -0.00000035, -0.00000429, 0.00000103, -0.00000036, 0.00000266, -0.00000080, -0.00000106, },
		{ -0.00000154, 0.00000216, 0.00000143, 0.00000224, -0.00000493, -0.00000776, 0.00000425, 0.00000411, -0.00000109, 0.00000114, -0.00000271, -0.00000219, -0.00000453, -0.00000005, 0.00000802, -0.00000140, -0.00000453, -0.00000331, 0.00000125, 0.00000664, -0.00000293, 0.00000148, 0.00000354, -0.00000600, -0.00000090, -0.00000081, -0.00000126, -0.00000049, 0.00000336, 0.00000296, -0.00000700, 0.00000001, -0.00000145, 0.00000108, 0.00000354, -0.00000295, 0.00000610, -0.00000918, -0.00000276, 0.00000290, -0.00000664, 0.00001307, -0.00001277, 0.00000087, 0.00001183, 0.00000379, -0.00000745, -0.00000873, 0.00000245, 0.00001075, -0.00000632, -0.00000503, -0.00000669, 0.00001323, -0.00001257, 0.00001936, -0.00002962, 0.00004118, -0.00004709, 0.00003938, -0.00004985, 0.00006432, -0.00006546, 0.00003168, 0.00000100, -0.00003934, 0.00007143, -0.00009016, 0.00008962, -0.00009695, 0.00009691, -0.00010340, 0.00007877, -0.00006799, 0.00005988, -0.00006946, 0.00001414, -0.00001903, -0.00000185, -0.00004913, -0.00002665, -0.00021258, -0.00023441, 0.00058542, 0.00055122, 0.00004529, -0.00037312, 0.00002039, -0.00006551, -0.00003054, -0.00001566, -0.00001865, 0.00001511, -0.00003424, 0.00000321, -0.00000985, -0.00000231, -0.00000093, -0.00001842, 0.00000943, -0.00000640, 0.00000376, -0.00001058, -0.00001117, 0.00001160, -0.00001739, 0.00000444, 0.00000174, -0.00000839, 0.00001238, 0.00000619, -0.00000833, 0.00000033, -0.00000109, 0.00000043, 0.00000033, 0.00000315, 0.00000204, -0.00000154, -0.00000595, -0.00000152, 0.00000691, 0.00000126, -0.00000462, 0.00000075, -0.00000055, 0.00000205, 0.00000001, },
		{ 0.00000063, -0.00000172, 0.00000200, 0.00000161, 0.00000236, -0.00000329, -0.00000868, 0.00000238, 0.00000533, -0.00000124, 0.00000139, -0.00000220, -0.00000263, -0.00000385, -0.00000186, 0.00000807, 0.00000043, -0.00000498, -0.00000317, -0.00000027, 0.00000716, -0.00000175, -0.00000026, 0.00000498, -0.00000575, -0.00000197, -0.00000012, -0.00000154, -0.00000043, 0.00000209, 0.00000458, -0.00000640, -0.00000159, -0.00000001, -0.00000073, 0.00000470, -0.00000349, 0.00000562, -0.00000596, -0.00000652, 0.00000539, -0.00000826, 0.00001288, -0.00000966, -0.00000378, 0.00001269, 0.00000534, -0.00000597, -0.00000980, 0.00000018, 0.00001237, -0.00000636, -0.00000289, -0.00001120, 0.00001755, -0.00001773, 0.00002586, -0.00003612, 0.00004660, -0.00004992, 0.00003955, -0.00004821, 0.00005563, -0.00004492, 0.00000160, 0.00003241, -0.00006441, 0.00008641, -0.00009590, 0.00008911, -0.00008969, 0.00008410, -0.00008492, 0.00005705, -0.00004472, 0.00003770, -0.00004960, 0.00000050, -0.00001132, -0.00000099, -0.00005479, -0.00001115, -0.00021770, -0.00020897, 0.00050907, 0.00049848, 0.00016753, -0.00037031, -0.00002318, -0.00003751, -0.00005174, -0.00000514, -0.00002380, 0.00001495, -0.00002997, -0.00000334, -0.00000449, -0.00000686, 0.00000280, -0.00001983, 0.00000841, -0.00000451, 0.00000052, -0.00000491, -0.00001684, 0.00001298, -0.00001426, -0.00000078, 0.00000617, -0.00000996, 0.00000899, 0.00001010, -0.00000864, -0.00000053, -0.00000054, -0.00000020, 0.00000077, 0.00000258, 0.00000221, -0.00000041, -0.00000598, -0.00000275, 0.00000639, 0.00000281, -0.00000462, 0.00000028, -0.00000050, 0.00000127, },
		{ 0.00000049, 0.00000096, -0.00000161, 0.00000171, 0.00000180, 0.00000239, -0.00000178, -0.00000892, 0.00000013, 0.00000627, -0.00000100, 0.00000129, -0.00000140, -0.00000320, -0.00000314, -0.00000326, 0.00000737, 0.00000252, -0.00000526, -0.00000304, -0.00000158, 0.00000700, -0.00000002, -0.00000201, 0.00000592, -0.00000482, -0.00000336, 0.00000054, -0.00000177, -0.00000030, 0.00000092, 0.00000552, -0.00000498, -0.00000357, 0.00000139, -0.00000213, 0.00000509, -0.00000303, 0.00000402, -0.00000213, -0.00000990, 0.00000686, -0.00000853, 0.00001123, -0.00000563, -0.00000802, 0.00001245, 0.00000725, -0.00000463, -0.00000998, -0.00000289, 0.00001432, -0.00000716, 0.00000048, -0.00001661, 0.00002240, -0.00002320, 0.00003201, -0.00004124, 0.00004952, -0.00004927, 0.00003543, -0.00004108, 0.00004053, -0.00001932, -0.00002937, 0.00005944, -0.00008142, 0.00009214, -0.00009312, 0.00008140, -0.00007669, 0.00006748, -0.00006490, 0.00003634, -0.00002497, 0.00002136, -0.00003729, -0.00000497, -0.00001131, 0.00000611, -0.00006409, 0.00000386, -0.00021855, -0.00019492, 0.00046004, 0.00042759, 0.00026909, -0.00033752, -0.00007724, -0.00001071, -0.00006789, -0.00000069, -0.00002365, 0.00001070, -0.00002321, -0.00001055, 0.00000014, -0.00001003, 0.00000473, -0.00001924, 0.00000575, -0.00000147, -0.00000335, 0.00000040, -0.00002064, 0.00001179, -0.00000944, -0.00000602, 0.00000925, -0.00000962, 0.00000451, 0.00001351, -0.00000800, -0.00000187, 0.00000018, -0.00000084, 0.00000108, 0.00000227, 0.00000204, 0.00000075, -0.00000578, -0.00000388, 0.00000560, 0.00000417, -0.00000425, -0.00000037, -0.00000028, },
		{ -0.00000001, -0.00000014, 0.00000093, -0.00000130, 0.00000140, 0.00000191, 0.00000241, -0.00000050, -0.00000853, -0.00000226, 0.00000669, -0.00000031, 0.00000090, -0.00000045, -0.00000374, -0.00000258, -0.00000411, 0.00000599, 0.00000464, -0.00000521, -0.00000305, -0.00000254, 0.00000619, 0.00000203, -0.00000349, 0.00000614, -0.00000325, -0.00000488, 0.00000102, -0.00000184, -0.00000022, 0.00000002, 0.00000572, -0.00000291, -0.00000559, 0.00000242, -0.00000283, 0.00000455, -0.00000163, 0.00000165, 0.00000175, -0.00001236, 0.00000698, -0.00000743, 0.00000844, -0.00000128, -0.00001130, 0.00001090, 0.00000953, -0.00000357, -0.00000914, -0.00000675, 0.00001662, -0.00000878, 0.00000494, -0.00002255, 0.00002726, -0.00002829, 0.00003690, -0.00004400, 0.00004907, -0.00004452, 0.00002680, -0.00002887, 0.00002057, 0.00000845, -0.00005767, 0.00007911, -0.00008886, 0.00008875, -0.00008317, 0.00006856, -0.00006046, 0.00004978, -0.00004603, 0.00001895, -0.00001043, 0.00001163, -0.00003214, -0.00000384, -0.00001642, 0.00001605, -0.00007319, 0.00001479, -0.00021244, -0.00019268, 0.00043267, 0.00034997, 0.00034278, -0.00027766, -0.00013487, 0.00000979, -0.00007643, -0.00000297, -0.00001878, 0.00000359, -0.00001547, -0.00001699, 0.00000303, -0.00001127, 0.00000469, -0.00001690, 0.00000196, 0.00000201, -0.00000709, 0.00000460, -0.00002219, 0.00000833, -0.00000377, -0.00001041, 0.00001055, -0.00000753, -0.00000047, 0.00001597, -0.00000633, -0.00000355, 0.00000089, -0.00000134, 0.00000115, 0.00000223, 0.00000162, 0.00000183, -0.00000530, -0.00000489, 0.00000462, 0.00000524, -0.00000353, -0.00000110, },
		{ -0.00000182, 0.00000019, -0.00000053, 0.00000059, -0.00000091, 0.00000119, 0.00000187, 0.00000246, 0.00000053, -0.00000763, -0.00000451, 0.00000647, 0.00000077, 0.00000034, 0.00000050, -0.00000409, -0.00000233, -0.00000437, 0.00000412, 0.00000653, -0.00000469, -0.00000326, -0.00000310, 0.00000481, 0.00000413, -0.00000440, 0.00000558, -0.00000118, -0.00000630, 0.00000114, -0.00000167, -0.00000030, -0.00000049, 0.00000522, -0.00000047, -0.00000728, 0.00000278, -0.00000268, 0.00000312, 0.00000044, -0.00000102, 0.00000510, -0.00001348, 0.00000567, -0.00000515, 0.00000499, 0.00000275, -0.00001316, 0.00000800, 0.00001211, -0.00000289, -0.00000722, -0.00001131, 0.00001921, -0.00001115, 0.00001019, -0.00002848, 0.00003148, -0.00003221, 0.00003965, -0.00004357, 0.00004470, -0.00003558, 0.00001408, -0.00001279, -0.00000198, 0.00003522, -0.00008012, 0.00008943, -0.00008657, 0.00007759, -0.00006829, 0.00005315, -0.00004362, 0.00003346, -0.00003034, 0.00000632, -0.00000172, 0.00000816, -0.00003280, 0.00000162, -0.00002378, 0.00002572, -0.00007918, 0.00001954, -0.00019871, -0.00020040, 0.00042002, 0.00027561, 0.00038515, -0.00019669, -0.00018846, 0.00001987, -0.00007619, -0.00001139, -0.00001064, -0.00000466, -0.00000834, -0.00002148, 0.00000359, -0.00001048, 0.00000296, -0.00001334, -0.00000223, 0.00000522, -0.00001005, 0.00000722, -0.00002148, 0.00000325, 0.00000178, -0.00001327, 0.00000995, -0.00000413, -0.00000532, 0.00001708, -0.00000368, -0.00000533, 0.00000141, -0.00000159, 0.00000095, 0.00000244, 0.00000106, 0.00000272, -0.00000456, -0.00000577, 0.00000352, 0.00000599, -0.00000252, },
		{ -0.00000133, -0.00000242, 0.00000024, -0.00000064, 0.00000001, -0.00000060, 0.00000119, 0.00000168, 0.00000252, 0.00000133, -0.00000640, -0.00000640, 0.00000555, 0.00000212, -0.00000021, 0.00000126, -0.00000410, -0.00000244, -0.00000411, 0.00000199, 0.00000794, -0.00000362, -0.00000367, -0.00000330, 0.00000305, 0.00000598, -0.00000457, 0.00000426, 0.00000114, -0.00000735, 0.00000077, -0.00000121, -0.00000056, -0.00000063, 0.00000419, 0.00000199, -0.00000832, 0.00000229, -0.00000167, 0.00000101, 0.00000281, -0.00000349, 0.00000744, -0.00001305, 0.00000305, -0.00000213, 0.00000149, 0.00000585, -0.00001327, 0.00000383, 0.00001483, -0.00000263, -0.00000424, -0.00001637, 0.00002195, -0.00001407, 0.00001575, -0.00003373, 0.00003434, -0.00003419, 0.00003952, -0.00003948, 0.00003637, -0.00002291, -0.00000167, 0.00000540, -0.00002451, 0.00005799, -0.00009438, 0.00008971, -0.00007567, 0.00006099, -0.00005121, 0.00003780, -0.00002847, 0.00002036, -0.00001908, -0.00000106, 0.00000152, 0.00000973, -0.00003724, 0.00000892, -0.00003075, 0.00003274, -0.00008031, 0.00001758, -0.00017846, -0.00021472, 0.00041503, 0.00021208, 0.00039650, -0.00010273, -0.00023083, 0.00001700, -0.00006761, -0.00002424, -0.00000123, -0.00001227, -0.00000319, -0.00002330, 0.00000170, -0.00000798, 0.00000007, -0.00000927, -0.00000610, 0.00000757, -0.00001180, 0.00000812, -0.00001888, -0.00000262, 0.00000636, -0.00001420, 0.00000765, -0.00000002, -0.00000941, 0.00001666, -0.00000029, -0.00000692, 0.00000157, -0.00000152, 0.00000048, 0.00000281, 0.00000047, 0.00000333, -0.00000356, -0.00000648, 0.00000235, 0.00000639, },
		{ 0.00000647, -0.00000005, -0.00000280, 0.00000009, -0.00000052, -0.00000065, -0.00000050, 0.00000143, 0.00000138, 0.00000256, 0.00000197, -0.00000501, -0.00000776, 0.00000399, 0.00000354, -0.00000054, 0.00000171, -0.00000373, -0.00000289, -0.00000347, -0.00000010, 0.00000865, -0.00000200, -0.00000419, -0.00000324, 0.00000113, 0.00000732, -0.00000389, 0.00000235, 0.00000340, -0.00000778, -0.00000016, -0.00000051, -0.00000096, -0.00000048, 0.00000288, 0.00000410, -0.00000845, 0.00000092, 0.00000001, -0.00000144, 0.00000502, -0.00000526, 0.00000845, -0.00001109, -0.00000049, 0.00000106, -0.00000142, 0.00000751, -0.00001152, -0.00000135, 0.00001751, -0.00000280, -0.00000029, -0.00002168, 0.00002460, -0.00001720, 0.00002101, -0.00003758, 0.00003516, -0.00003361, 0.00003607, -0.00003167, 0.00002455, -0.00000755, -0.00001889, 0.00002355, -0.00004440, 0.00007431, -0.00009922, 0.00008059, -0.00005842, 0.00004189, -0.00003472, 0.00002475, -0.00001668, 0.00001152, -0.00001256, -0.00000362, 0.00000041, 0.00001457, -0.00004329, 0.00001576, -0.00003533, 0.00003568, -0.00007612, 0.00000976, -0.00015395, -0.00023170, 0.00041150, 0.00016409, 0.00038038, -0.00000498, -0.00025613, 0.00000073, -0.00005263, -0.00003906, 0.00000727, -0.00001766, -0.00000091, -0.00002223, -0.00000227, -0.00000440, -0.00000322, -0.00000543, -0.00000904, 0.00000869, -0.00001222, 0.00000746, -0.00001496, -0.00000838, 0.00000928, -0.00001313, 0.00000413, 0.00000413, -0.00001225, 0.00001472, 0.00000347, -0.00000802, 0.00000125, -0.00000112, -0.00000019, 0.00000323, -0.00000000, 0.00000359, -0.00000237, -0.00000698, 0.00000115, },
		{ -0.00000005, 0.00000624, 0.00000121, -0.00000292, -0.00000024, -0.00000022, -0.00000126, -0.00000071, 0.00000188, 0.00000106, 0.00000248, 0.00000251, -0.00000360, -0.00000851, 0.00000193, 0.00000477, -0.00000050, 0.00000176, -0.00000300, -0.00000358, -0.00000268, -0.00000190, 0.00000857, 0.00000006, -0.00000470, -0.00000305, -0.00000069, 0.00000794, -0.00000241, 0.00000013, 0.00000527, -0.00000742, -0.00000159, 0.00000032, -0.00000140, -0.00000018, 0.00000156, 0.00000559, -0.00000757, -0.00000119, 0.00000204, -0.00000380, 0.00000662, -0.00000600, 0.00000803, -0.00000790, -0.00000438, 0.00000378, -0.00000318, 0.00000743, -0.00000802, -0.00000716, 0.00001992, -0.00000336, 0.00000440, -0.00002688, 0.00002683, -0.00002011, 0.00002532, -0.00003934, 0.00003341, -0.00003010, 0.00002923, -0.00002058, 0.00001022, 0.00000903, -0.00003572, 0.00003952, -0.00005946, 0.00008267, -0.00009469, 0.00006397, -0.00003781, 0.00002333, -0.00002123, 0.00001559, -0.00000907, 0.00000709, -0.00001029, -0.00000245, -0.00000343, 0.00002076, -0.00004891, 0.00002040, -0.00003634, 0.00003420, -0.00006724, -0.00000212, -0.00012804, -0.00024757, 0.00040480, 0.00013336, 0.00034281, 0.00008754, -0.00026072, -0.00002728, -0.00003435, -0.00005308, 0.00001292, -0.00001978, -0.00000177, -0.00001864, -0.00000757, -0.00000061, -0.00000617, -0.00000244, -0.00001069, 0.00000847, -0.00001144, 0.00000562, -0.00001044, -0.00001325, 0.00001018, -0.00001034, 0.00000004, 0.00000764, -0.00001354, 0.00001148, 0.00000717, -0.00000836, 0.00000042, -0.00000048, -0.00000093, 0.00000358, -0.00000027, 0.00000352, -0.00000107, -0.00000723, },
		{ -0.00000718, -0.00000122, 0.00000576, 0.00000236, -0.00000276, -0.00000071, 0.00000013, -0.00000168, -0.00000126, 0.00000244, 0.00000085, 0.00000222, 0.00000301, -0.00000227, -0.00000865, -0.00000039, 0.00000557, 0.00000002, 0.00000145, -0.00000200, -0.00000434, -0.00000194, -0.00000320, 0.00000768, 0.00000237, -0.00000501, -0.00000290, -0.00000223, 0.00000775, -0.00000029, -0.00000207, 0.00000645, -0.00000621, -0.00000340, 0.00000109, -0.00000174, 0.00000010, 0.00000049, 0.00000626, -0.00000575, -0.00000374, 0.00000403, -0.00000566, 0.00000727, -0.00000555, 0.00000633, -0.00000397, -0.00000797, 0.00000542, -0.00000339, 0.00000553, -0.00000310, -0.00001312, 0.00002181, -0.00000425, 0.00000957, -0.00003154, 0.00002826, -0.00002230, 0.00002801, -0.00003848, 0.00002880, -0.00002366, 0.00001937, -0.00000708, -0.00000526, 0.00002514, -0.00005031, 0.00005144, -0.00006819, 0.00008267, -0.00008206, 0.00004259, -0.00001709, 0.00000799, -0.00001241, 0.00001099, -0.00000562, 0.00000647, -0.00001115, 0.00000095, -0.00000829, 0.00002652, -0.00005258, 0.00002182, -0.00003353, 0.00002885, -0.00005511, -0.00001577, -0.00010352, -0.00025936, 0.00039218, 0.00011899, 0.00029122, 0.00016705, -0.00024350, -0.00006349, -0.00001651, -0.00006371, 0.00001438, -0.00001831, -0.00000543, -0.00001334, -0.00001321, 0.00000257, -0.00000817, -0.00000066, -0.00001092, 0.00000709, -0.00000982, 0.00000315, -0.00000604, -0.00001664, 0.00000903, -0.00000637, -0.00000391, 0.00001004, -0.00001324, 0.00000735, 0.00001034, -0.00000776, -0.00000086, 0.00000027, -0.00000161, 0.00000372, -0.00000027, 0.00000314, 0.00000022, },
		{ 0.00000138, -0.00000681, -0.00000232, 0.00000506, 0.00000334, -0.00000233, -0.00000124, 0.00000043, -0.00000182, -0.00000209, 0.00000297, 0.00000087, 0.00000176, 0.00000346, -0.00000106, -0.00000828, -0.00000272, 0.00000575, 0.00000103, 0.00000086, -0.00000091, -0.00000497, -0.00000147, -0.00000388, 0.00000610, 0.00000467, -0.00000494, -0.00000290, -0.00000335, 0.00000678, 0.00000219, -0.00000389, 0.00000674, -0.00000421, -0.00000532, 0.00000160, -0.00000184, 0.00000020, -0.00000017, 0.00000606, -0.00000321, -0.00000632, 0.00000555, -0.00000667, 0.00000681, -0.00000399, 0.00000374, 0.00000007, -0.00001061, 0.00000555, -0.00000190, 0.00000200, 0.00000275, -0.00001869, 0.00002298, -0.00000535, 0.00001485, -0.00003519, 0.00002851, -0.00002330, 0.00002856, -0.00003468, 0.00002138, -0.00001465, 0.00000729, 0.00000756, -0.00002029, 0.00003906, -0.00006105, 0.00005804, -0.00007001, 0.00007498, -0.00006358, 0.00001972, 0.00000069, -0.00000217, -0.00000895, 0.00001073, -0.00000553, 0.00000848, -0.00001370, 0.00000502, -0.00001264, 0.00003057, -0.00005345, 0.00001984, -0.00002743, 0.00002087, -0.00004157, -0.00002893, -0.00008266, -0.00026527, 0.00037282, 0.00011805, 0.00023337, 0.00022784, -0.00020607, -0.00010302, -0.00000282, -0.00006906, 0.00001117, -0.00001362, -0.00001102, -0.00000741, -0.00001821, 0.00000445, -0.00000887, -0.00000021, -0.00000991, 0.00000491, -0.00000781, 0.00000058, -0.00000233, -0.00001825, 0.00000612, -0.00000192, -0.00000705, 0.00001101, -0.00001153, 0.00000285, 0.00001255, -0.00000620, -0.00000242, 0.00000095, -0.00000210, 0.00000360, -0.00000000, 0.00000255, },
		{ 0.00000187, 0.00000230, -0.00000613, -0.00000329, 0.00000418, 0.00000410, -0.00000168, -0.00000175, 0.00000059, -0.00000165, -0.00000311, 0.00000331, 0.00000119, 0.00000111, 0.00000384, 0.00000004, -0.00000752, -0.00000482, 0.00000521, 0.00000244, 0.00000017, 0.00000008, -0.00000531, -0.00000137, -0.00000393, 0.00000403, 0.00000667, -0.00000434, -0.00000314, -0.00000401, 0.00000517, 0.00000469, -0.00000502, 0.00000606, -0.00000163, -0.00000707, 0.00000163, -0.00000159, 0.00000002, -0.00000032, 0.00000511, -0.00000030, -0.00000851, 0.00000626, -0.00000664, 0.00000528, -0.00000161, 0.00000082, 0.00000352, -0.00001173, 0.00000392, 0.00000121, -0.00000275, 0.00000890, -0.00002338, 0.00002327, -0.00000654, 0.00001982, -0.00003737, 0.00002722, -0.00002272, 0.00002661, -0.00002795, 0.00001156, -0.00000381, -0.00000590, 0.00002189, -0.00003328, 0.00004934, -0.00006685, 0.00005877, -0.00006526, 0.00006128, -0.00004213, -0.00000146, 0.00001319, -0.00000623, -0.00001057, 0.00001380, -0.00000749, 0.00001168, -0.00001648, 0.00000842, -0.00001539, 0.00003220, -0.00005136, 0.00001494, -0.00001914, 0.00001182, -0.00002843, -0.00003973, -0.00006685, -0.00026480, 0.00034748, 0.00012644, 0.00017633, 0.00026676, -0.00015233, -0.00014036, 0.00000363, -0.00006824, 0.00000367, -0.00000674, -0.00001732, -0.00000198, -0.00002173, 0.00000464, -0.00000820, -0.00000090, -0.00000803, 0.00000242, -0.00000590, -0.00000162, 0.00000031, -0.00001807, 0.00000199, 0.00000227, -0.00000893, 0.00001055, -0.00000880, -0.00000145, 0.00001351, -0.00000377, -0.00000399, 0.00000138, -0.00000228, 0.00000319, 0.00000048, },
		{ 0.00000107, 0.00000123, 0.00000291, -0.00000519, -0.00000407, 0.00000314, 0.00000462, -0.00000088, -0.00000214, 0.00000056, -0.00000122, -0.00000416, 0.00000331, 0.00000184, 0.00000035, 0.00000407, 0.00000108, -0.00000653, -0.00000647, 0.00000395, 0.00000405, -0.00000040, 0.00000080, -0.00000523, -0.00000169, -0.00000347, 0.00000176, 0.00000812, -0.00000313, -0.00000359, -0.00000425, 0.00000314, 0.00000688, -0.00000524, 0.00000448, 0.00000121, -0.00000830, 0.00000103, -0.00000096, -0.00000045, -0.00000003, 0.00000362, 0.00000257, -0.00000991, 0.00000594, -0.00000559, 0.00000293, 0.00000111, -0.00000184, 0.00000579, -0.00001102, 0.00000055, 0.00000559, -0.00000815, 0.00001468, -0.00002675, 0.00002258, -0.00000765, 0.00002405, -0.00003767, 0.00002415, -0.00002033, 0.00002210, -0.00001864, 0.00000007, 0.00000785, -0.00001891, 0.00003443, -0.00004287, 0.00005498, -0.00006726, 0.00005392, -0.00005518, 0.00004390, -0.00002076, -0.00001821, 0.00001902, -0.00000436, -0.00001609, 0.00001863, -0.00000998, 0.00001469, -0.00001831, 0.00001025, -0.00001601, 0.00003134, -0.00004677, 0.00000811, -0.00001004, 0.00000322, -0.00001721, -0.00004699, -0.00005651, -0.00025858, 0.00031807, 0.00013968, 0.00012575, 0.00028338, -0.00008792, -0.00017018, 0.00000094, -0.00006153, -0.00000688, 0.00000089, -0.00002308, 0.00000198, -0.00002328, 0.00000310, -0.00000643, -0.00000235, -0.00000575, 0.00000011, -0.00000448, -0.00000314, 0.00000173, -0.00001638, -0.00000270, 0.00000552, -0.00000928, 0.00000888, -0.00000556, -0.00000506, 0.00001310, -0.00000073, -0.00000530, 0.00000139, -0.00000212, 0.00000253, },
		{ 0.00000171, 0.00000165, 0.00000075, 0.00000318, -0.00000407, -0.00000460, 0.00000198, 0.00000491, -0.00000001, -0.00000236, 0.00000032, -0.00000059, -0.00000508, 0.00000290, 0.00000276, -0.00000040, 0.00000405, 0.00000212, -0.00000544, -0.00000757, 0.00000207, 0.00000562, -0.00000064, 0.00000110, -0.00000471, -0.00000238, -0.00000267, -0.00000041, 0.00000879, -0.00000133, -0.00000416, -0.00000420, 0.00000095, 0.00000844, -0.00000447, 0.00000221, 0.00000394, -0.00000875, -0.00000026, -0.00000000, -0.00000113, 0.00000058, 0.00000191, 0.00000499, -0.00001024, 0.00000457, -0.00000373, 0.00000016, 0.00000363, -0.00000362, 0.00000645, -0.00000839, -0.00000424, 0.00001070, -0.00001351, 0.00001949, -0.00002850, 0.00002089, -0.00000855, 0.00002710, -0.00003578, 0.00001923, -0.00001612, 0.00001524, -0.00000741, -0.00001208, 0.00001913, -0.00003040, 0.00004390, -0.00004810, 0.00005557, -0.00006258, 0.00004451, -0.00004163, 0.00002546, -0.00000223, -0.00002868, 0.00001798, 0.00000227, -0.00002374, 0.00002349, -0.00001157, 0.00001644, -0.00001844, 0.00001013, -0.00001452, 0.00002843, -0.00004054, 0.00000056, -0.00000146, -0.00000369, -0.00000882, -0.00005031, -0.00005119, -0.00024809, 0.00028707, 0.00015362, 0.00008536, 0.00027966, -0.00001940, -0.00018808, -0.00001133, -0.00005033, -0.00001871, 0.00000772, -0.00002719, 0.00000388, -0.00002275, 0.00000013, -0.00000402, -0.00000406, -0.00000357, -0.00000163, -0.00000379, -0.00000389, 0.00000202, -0.00001360, -0.00000724, 0.00000738, -0.00000817, 0.00000640, -0.00000236, -0.00000759, 0.00001138, 0.00000255, -0.00000605, 0.00000090, -0.00000165, },
		{ -0.00000098, 0.00000087, 0.00000209, 0.00000051, 0.00000312, -0.00000288, -0.00000481, 0.00000073, 0.00000496, 0.00000088, -0.00000238, -0.00000009, 0.00000011, -0.00000573, 0.00000203, 0.00000383, -0.00000096, 0.00000371, 0.00000318, -0.00000436, -0.00000808, -0.00000022, 0.00000687, -0.00000038, 0.00000095, -0.00000379, -0.00000327, -0.00000177, -0.00000220, 0.00000859, 0.00000092, -0.00000470, -0.00000401, -0.00000112, 0.00000917, -0.00000278, -0.00000039, 0.00000615, -0.00000820, -0.00000218, 0.00000112, -0.00000188, 0.00000130, 0.00000030, 0.00000662, -0.00000940, 0.00000232, -0.00000144, -0.00000248, 0.00000537, -0.00000408, 0.00000529, -0.00000406, -0.00000992, 0.00001587, -0.00001818, 0.00002280, -0.00002846, 0.00001827, -0.00000905, 0.00002861, -0.00003156, 0.00001257, -0.00001029, 0.00000652, 0.00000481, -0.00002374, 0.00002884, -0.00003921, 0.00004937, -0.00004857, 0.00005140, -0.00005375, 0.00003213, -0.00002678, 0.00000850, 0.00001133, -0.00003211, 0.00001095, 0.00001175, -0.00003149, 0.00002676, -0.00001124, 0.00001635, -0.00001667, 0.00000824, -0.00001141, 0.00002429, -0.00003373, -0.00000654, 0.00000555, -0.00000815, -0.00000357, -0.00005002, -0.00004974, -0.00023525, 0.00025695, 0.00016505, 0.00005685, 0.00025952, 0.00004661, -0.00019122, -0.00003201, -0.00003688, -0.00002982, 0.00001241, -0.00002894, 0.00000354, -0.00002044, -0.00000371, -0.00000154, -0.00000556, -0.00000188, -0.00000256, -0.00000388, -0.00000393, 0.00000141, -0.00001027, -0.00001103, 0.00000760, -0.00000586, 0.00000365, 0.00000031, -0.00000885, 0.00000861, 0.00000565, -0.00000601, -0.00000010, },
		{ -0.00000148, -0.00000025, 0.00000014, 0.00000230, 0.00000053, 0.00000280, -0.00000176, -0.00000466, -0.00000055, 0.00000477, 0.00000173, -0.00000219, -0.00000061, 0.00000077, -0.00000599, 0.00000077, 0.00000487, -0.00000118, 0.00000299, 0.00000424, -0.00000331, -0.00000806, -0.00000266, 0.00000757, 0.00000044, 0.00000040, -0.00000262, -0.00000418, -0.00000102, -0.00000341, 0.00000755, 0.00000340, -0.00000501, -0.00000387, -0.00000286, 0.00000899, -0.00000039, -0.00000294, 0.00000750, -0.00000662, -0.00000454, 0.00000219, -0.00000249, 0.00000190, -0.00000092, 0.00000729, -0.00000748, -0.00000045, 0.00000078, -0.00000448, 0.00000590, -0.00000296, 0.00000242, 0.00000150, -0.00001580, 0.00002038, -0.00002157, 0.00002430, -0.00002664, 0.00001485, -0.00000904, 0.00002833, -0.00002507, 0.00000449, -0.00000330, -0.00000331, 0.00001691, -0.00003372, 0.00003595, -0.00004453, 0.00005043, -0.00004450, 0.00004333, -0.00004222, 0.00001867, -0.00001281, -0.00000494, 0.00001873, -0.00002883, -0.00000028, 0.00002180, -0.00003746, 0.00002729, -0.00000851, 0.00001440, -0.00001330, 0.00000513, -0.00000747, 0.00001984, -0.00002733, -0.00001229, 0.00001035, -0.00000998, -0.00000113, -0.00004697, -0.00005065, -0.00022200, 0.00022974, 0.00017195, 0.00004001, 0.00022801, 0.00010431, -0.00017859, -0.00005857, -0.00002388, -0.00003838, 0.00001403, -0.00002814, 0.00000125, -0.00001694, -0.00000771, 0.00000045, -0.00000647, -0.00000091, -0.00000262, -0.00000463, -0.00000347, 0.00000025, -0.00000691, -0.00001364, 0.00000624, -0.00000283, 0.00000117, 0.00000212, -0.00000886, 0.00000516, 0.00000815, -0.00000506, },
		{ -0.00000322, -0.00000307, 0.00000033, -0.00000034, 0.00000223, 0.00000079, 0.00000232, -0.00000081, -0.00000415, -0.00000180, 0.00000433, 0.00000250, -0.00000181, -0.00000116, 0.00000127, -0.00000584, -0.00000075, 0.00000568, -0.00000094, 0.00000193, 0.00000525, -0.00000228, -0.00000763, -0.00000499, 0.00000753, 0.00000180, -0.00000041, -0.00000141, -0.00000489, -0.00000060, -0.00000395, 0.00000580, 0.00000584, -0.00000488, -0.00000390, -0.00000413, 0.00000794, 0.00000239, -0.00000501, 0.00000778, -0.00000409, -0.00000703, 0.00000292, -0.00000278, 0.00000219, -0.00000157, 0.00000696, -0.00000476, -0.00000324, 0.00000240, -0.00000538, 0.00000494, -0.00000028, -0.00000181, 0.00000763, -0.00002113, 0.00002362, -0.00002325, 0.00002384, -0.00002323, 0.00001081, -0.00000839, 0.00002611, -0.00001658, -0.00000444, 0.00000422, -0.00001336, 0.00002779, -0.00004104, 0.00003974, -0.00004597, 0.00004723, -0.00003666, 0.00003270, -0.00002971, 0.00000603, -0.00000154, -0.00001352, 0.00001982, -0.00002020, -0.00001336, 0.00003015, -0.00004021, 0.00002456, -0.00000352, 0.00001107, -0.00000899, 0.00000159, -0.00000353, 0.00001593, -0.00002211, -0.00001616, 0.00001280, -0.00000948, -0.00000078, -0.00004233, -0.00005240, -0.00021001, 0.00020673, 0.00017358, 0.00003315, 0.00019060, 0.00014930, -0.00015119, -0.00008746, -0.00001404, -0.00004307, 0.00001228, -0.00002511, -0.00000235, -0.00001304, -0.00001120, 0.00000152, -0.00000660, -0.00000070, -0.00000194, -0.00000579, -0.00000279, -0.00000112, -0.00000394, -0.00001485, 0.00000356, 0.00000034, -0.00000058, 0.00000291, -0.00000782, 0.00000151, 0.00000971, },
		{ 0.00001011, -0.00000064, -0.00000458, 0.00000060, -0.00000049, 0.00000187, 0.00000121, 0.00000182, -0.00000015, -0.00000330, -0.00000292, 0.00000365, 0.00000317, -0.00000129, -0.00000165, 0.00000155, -0.00000530, -0.00000234, 0.00000606, -0.00000020, 0.00000063, 0.00000608, -0.00000121, -0.00000695, -0.00000696, 0.00000668, 0.00000356, -0.00000125, -0.00000037, -0.00000522, -0.00000064, -0.00000384, 0.00000358, 0.00000792, -0.00000418, -0.00000417, -0.00000489, 0.00000618, 0.00000517, -0.00000627, 0.00000691, -0.00000090, -0.00000928, 0.00000306, -0.00000261, 0.00000206, -0.00000159, 0.00000580, -0.00000168, -0.00000552, 0.00000297, -0.00000490, 0.00000248, 0.00000367, -0.00000681, 0.00001355, -0.00002519, 0.00002510, -0.00002304, 0.00002152, -0.00001850, 0.00000637, -0.00000705, 0.00002200, -0.00000658, -0.00001354, 0.00001152, -0.00002268, 0.00003643, -0.00004494, 0.00003987, -0.00004363, 0.00004042, -0.00002630, 0.00002114, -0.00001797, -0.00000418, 0.00000580, -0.00001682, 0.00001539, -0.00000827, -0.00002580, 0.00003499, -0.00003905, 0.00001876, 0.00000301, 0.00000720, -0.00000458, -0.00000155, -0.00000034, 0.00001316, -0.00001844, -0.00001806, 0.00001323, -0.00000735, -0.00000153, -0.00003727, -0.00005367, -0.00020034, 0.00018840, 0.00017030, 0.00003361, 0.00015241, 0.00017901, -0.00011173, -0.00011465, -0.00000955, -0.00004331, 0.00000754, -0.00002062, -0.00000640, -0.00000951, -0.00001365, 0.00000148, -0.00000597, -0.00000110, -0.00000077, -0.00000707, -0.00000217, -0.00000240, -0.00000166, -0.00001467, 0.00000003, 0.00000308, -0.00000133, 0.00000273, -0.00000608, -0.00000188, },
		{ -0.00000461, 0.00000930, 0.00000242, -0.00000574, 0.00000041, -0.00000031, 0.00000130, 0.00000169, 0.00000142, 0.00000018, -0.00000219, -0.00000383, 0.00000273, 0.00000372, -0.00000066, -0.00000203, 0.00000158, -0.00000448, -0.00000380, 0.00000587, 0.00000099, -0.00000076, 0.00000659, -0.00000002, -0.00000615, -0.00000841, 0.00000507, 0.00000548, -0.00000189, 0.00000030, -0.00000505, -0.00000115, -0.00000321, 0.00000121, 0.00000937, -0.00000283, -0.00000466, -0.00000520, 0.00000398, 0.00000757, -0.00000648, 0.00000500, 0.00000258, -0.00001089, 0.00000243, -0.00000193, 0.00000153, -0.00000107, 0.00000412, 0.00000126, -0.00000680, 0.00000218, -0.00000299, -0.00000125, 0.00000840, -0.00001187, 0.00001849, -0.00002743, 0.00002458, -0.00002095, 0.00001759, -0.00001286, 0.00000176, -0.00000501, 0.00001621, 0.00000425, -0.00002198, 0.00001784, -0.00003043, 0.00004207, -0.00004510, 0.00003651, -0.00003810, 0.00003108, -0.00001490, 0.00001028, -0.00000848, -0.00001088, 0.00000873, -0.00001524, 0.00000701, 0.00000459, -0.00003538, 0.00003522, -0.00003408, 0.00001072, 0.00001003, 0.00000377, -0.00000090, -0.00000367, 0.00000166, 0.00001177, -0.00001634, -0.00001829, 0.00001224, -0.00000444, -0.00000240, -0.00003280, -0.00005360, -0.00019347, 0.00017444, 0.00016331, 0.00003839, 0.00011761, 0.00019282, -0.00006426, -0.00013624, -0.00001177, -0.00003938, 0.00000071, -0.00001570, -0.00001003, -0.00000698, -0.00001475, 0.00000036, -0.00000479, -0.00000188, 0.00000059, -0.00000819, -0.00000180, -0.00000341, -0.00000018, -0.00001330, -0.00000382, 0.00000494, -0.00000098, 0.00000183, -0.00000406, },
		{ -0.00000217, -0.00000642, 0.00000739, 0.00000558, -0.00000628, -0.00000029, 0.00000015, 0.00000063, 0.00000209, 0.00000121, 0.00000016, -0.00000093, -0.00000443, 0.00000159, 0.00000411, 0.00000003, -0.00000227, 0.00000140, -0.00000351, -0.00000493, 0.00000506, 0.00000251, -0.00000201, 0.00000666, 0.00000132, -0.00000534, -0.00000926, 0.00000287, 0.00000728, -0.00000210, 0.00000049, -0.00000439, -0.00000202, -0.00000228, -0.00000099, 0.00000998, -0.00000087, -0.00000526, -0.00000520, 0.00000163, 0.00000925, -0.00000557, 0.00000232, 0.00000589, -0.00001154, 0.00000095, -0.00000083, 0.00000070, -0.00000024, 0.00000231, 0.00000354, -0.00000670, -0.00000005, 0.00000015, -0.00000580, 0.00001324, -0.00001624, 0.00002182, -0.00002750, 0.00002207, -0.00001725, 0.00001247, -0.00000674, -0.00000279, -0.00000234, 0.00000912, 0.00001510, -0.00002896, 0.00002249, -0.00003594, 0.00004426, -0.00004162, 0.00003023, -0.00003036, 0.00002055, -0.00000399, 0.00000151, -0.00000225, -0.00001360, 0.00000752, -0.00000992, -0.00000333, 0.00001607, -0.00004051, 0.00003060, -0.00002616, 0.00000172, 0.00001633, 0.00000166, 0.00000149, -0.00000441, 0.00000235, 0.00001169, -0.00001549, -0.00001740, 0.00001059, -0.00000160, -0.00000259, -0.00002955, -0.00005182, -0.00018924, 0.00016393, 0.00015422, 0.00004462, 0.00008911, 0.00019191, -0.00001350, -0.00014900, -0.00002091, -0.00003237, -0.00000688, -0.00001143, -0.00001251, -0.00000582, -0.00001452, -0.00000156, -0.00000340, -0.00000272, 0.00000186, -0.00000891, -0.00000181, -0.00000408, 0.00000051, -0.00001105, -0.00000747, 0.00000562, 0.00000033, 0.00000058, },
		{ -0.00000062, -0.00000075, -0.00000723, 0.00000465, 0.00000845, -0.00000602, -0.00000147, 0.00000076, -0.00000000, 0.00000230, 0.00000126, -0.00000016, 0.00000035, -0.00000464, 0.00000027, 0.00000432, 0.00000076, -0.00000235, 0.00000104, -0.00000253, -0.00000559, 0.00000367, 0.00000414, -0.00000289, 0.00000618, 0.00000281, -0.00000456, -0.00000954, 0.00000031, 0.00000864, -0.00000172, 0.00000016, -0.00000332, -0.00000309, -0.00000132, -0.00000275, 0.00000967, 0.00000156, -0.00000580, -0.00000508, -0.00000058, 0.00001001, -0.00000365, -0.00000072, 0.00000857, -0.00001100, -0.00000129, 0.00000050, -0.00000020, 0.00000061, 0.00000080, 0.00000477, -0.00000506, -0.00000356, 0.00000410, -0.00001054, 0.00001749, -0.00001926, 0.00002310, -0.00002534, 0.00001782, -0.00001234, 0.00000664, -0.00000057, -0.00000707, 0.00000082, 0.00000125, 0.00002512, -0.00003377, 0.00002498, -0.00003881, 0.00004299, -0.00003502, 0.00002199, -0.00002161, 0.00001023, 0.00000507, -0.00000417, 0.00000029, -0.00001257, 0.00000306, -0.00000248, -0.00001352, 0.00002428, -0.00004042, 0.00002181, -0.00001673, -0.00000681, 0.00002080, 0.00000147, 0.00000235, -0.00000378, 0.00000193, 0.00001253, -0.00001538, -0.00001605, 0.00000902, 0.00000050, -0.00000164, -0.00002775, -0.00004842, -0.00018704, 0.00015567, 0.00014470, 0.00005000, 0.00006834, 0.00017888, 0.00003572, -0.00015085, -0.00003608, -0.00002395, -0.00001383, -0.00000871, -0.00001342, -0.00000608, -0.00001322, -0.00000385, -0.00000215, -0.00000337, 0.00000285, -0.00000916, -0.00000217, -0.00000446, 0.00000056, -0.00000829, -0.00001049, 0.00000503, 0.00000229, },
		{ 0.00000449, -0.00000139, 0.00000003, -0.00000710, 0.00000146, 0.00001066, -0.00000487, -0.00000298, 0.00000135, -0.00000046, 0.00000225, 0.00000157, -0.00000070, 0.00000151, -0.00000443, -0.00000115, 0.00000431, 0.00000151, -0.00000229, 0.00000060, -0.00000170, -0.00000570, 0.00000183, 0.00000561, -0.00000322, 0.00000513, 0.00000438, -0.00000380, -0.00000935, -0.00000230, 0.00000933, -0.00000068, -0.00000058, -0.00000201, -0.00000411, -0.00000057, -0.00000387, 0.00000848, 0.00000419, -0.00000607, -0.00000500, -0.00000242, 0.00000977, -0.00000097, -0.00000367, 0.00001023, -0.00000924, -0.00000404, 0.00000176, -0.00000089, 0.00000118, -0.00000006, 0.00000468, -0.00000196, -0.00000796, 0.00000828, -0.00001481, 0.00002050, -0.00002045, 0.00002216, -0.00002117, 0.00001228, -0.00000677, 0.00000062, 0.00000525, -0.00001089, 0.00000426, -0.00000679, 0.00003347, -0.00003587, 0.00002509, -0.00003901, 0.00003860, -0.00002619, 0.00001294, -0.00001310, 0.00000135, 0.00001129, -0.00000631, -0.00000068, -0.00000859, -0.00000329, 0.00000532, -0.00002173, 0.00002806, -0.00003529, 0.00001023, -0.00000751, -0.00001352, 0.00002271, 0.00000339, 0.00000182, -0.00000212, 0.00000086, 0.00001378, -0.00001542, -0.00001484, 0.00000809, 0.00000148, 0.00000061, -0.00002723, -0.00004389, -0.00018599, 0.00014839, 0.00013616, 0.00005302, 0.00005543, 0.00015723, 0.00007913, -0.00014110, -0.00005534, -0.00001609, -0.00001888, -0.00000810, -0.00001272, -0.00000750, -0.00001129, -0.00000606, -0.00000139, -0.00000365, 0.00000346, -0.00000894, -0.00000279, -0.00000463, 0.00000012, -0.00000539, -0.00001257, 0.00000329, },
		{ 0.00000068, 0.00000647, -0.00000147, 0.00000015, -0.00000626, -0.00000177, 0.00001194, -0.00000289, -0.00000462, 0.00000173, -0.00000062, 0.00000190, 0.00000209, -0.00000132, 0.00000242, -0.00000378, -0.00000258, 0.00000404, 0.00000227, -0.00000212, 0.00000016, -0.00000110, -0.00000529, -0.00000024, 0.00000665, -0.00000286, 0.00000358, 0.00000589, -0.00000297, -0.00000884, -0.00000469, 0.00000918, 0.00000094, -0.00000157, -0.00000070, -0.00000487, -0.00000024, -0.00000426, 0.00000658, 0.00000674, -0.00000587, -0.00000510, -0.00000376, 0.00000861, 0.00000211, -0.00000605, 0.00001064, -0.00000640, -0.00000693, 0.00000259, -0.00000110, 0.00000118, -0.00000002, 0.00000325, 0.00000229, -0.00001267, 0.00001201, -0.00001796, 0.00002178, -0.00001959, 0.00001911, -0.00001541, 0.00000605, -0.00000108, -0.00000514, 0.00001040, -0.00001410, 0.00000773, -0.00001436, 0.00003947, -0.00003501, 0.00002289, -0.00003680, 0.00003179, -0.00001624, 0.00000431, -0.00000594, -0.00000515, 0.00001419, -0.00000503, -0.00000442, -0.00000288, -0.00001001, 0.00001186, -0.00002667, 0.00002707, -0.00002611, -0.00000229, -0.00000017, -0.00001745, 0.00002179, 0.00000714, 0.00000041, 0.00000002, -0.00000029, 0.00001489, -0.00001511, -0.00001422, 0.00000810, 0.00000126, 0.00000394, -0.00002755, -0.00003889, -0.00018520, 0.00014099, 0.00012954, 0.00005302, 0.00004940, 0.00013082, 0.00011346, -0.00012046, -0.00007604, -0.00001072, -0.00002118, -0.00000968, -0.00001071, -0.00000957, -0.00000929, -0.00000774, -0.00000129, -0.00000355, 0.00000372, -0.00000836, -0.00000352, -0.00000472, -0.00000062, -0.00000261, -0.00001360, },
		{ -0.00001360, -0.00000243, 0.00000783, -0.00000074, -0.00000027, -0.00000503, -0.00000468, 0.00001214, -0.00000028, -0.00000609, 0.00000172, -0.00000043, 0.00000129, 0.00000271, -0.00000189, 0.00000300, -0.00000277, -0.00000390, 0.00000347, 0.00000303, -0.00000185, -0.00000022, -0.00000078, -0.00000443, -0.00000227, 0.00000704, -0.00000180, 0.00000171, 0.00000717, -0.00000197, -0.00000818, -0.00000663, 0.00000816, 0.00000296, -0.00000255, 0.00000036, -0.00000516, -0.00000043, -0.00000397, 0.00000423, 0.00000888, -0.00000505, -0.00000545, -0.00000457, 0.00000671, 0.00000515, -0.00000751, 0.00000976, -0.00000282, -0.00000950, 0.00000266, -0.00000062, 0.00000047, 0.00000099, 0.00000067, 0.00000718, -0.00001702, 0.00001466, -0.00001951, 0.00002109, -0.00001673, 0.00001435, -0.00000871, -0.00000022, 0.00000420, -0.00001024, 0.00001465, -0.00001657, 0.00001094, -0.00002085, 0.00004264, -0.00003124, 0.00001877, -0.00003274, 0.00002349, -0.00000637, -0.00000280, -0.00000098, -0.00000880, 0.00001381, -0.00000100, -0.00000981, 0.00000319, -0.00001568, 0.00001593, -0.00002774, 0.00002181, -0.00001448, -0.00001378, 0.00000397, -0.00001820, 0.00001832, 0.00001208, -0.00000116, 0.00000200, -0.00000099, 0.00001544, -0.00001413, -0.00001437, 0.00000906, 0.00000005, 0.00000794, -0.00002814, -0.00003410, -0.00018388, 0.00013275, 0.00012524, 0.00005011, 0.00004858, 0.00010328, 0.00013679, -0.00009095, -0.00009524, -0.00000932, -0.00002046, -0.00001310, -0.00000799, -0.00001172, -0.00000771, -0.00000862, -0.00000188, -0.00000314, 0.00000377, -0.00000757, -0.00000420, -0.00000483, -0.00000154, -0.00000017, },
		{ 0.00000185, -0.00001271, -0.00000561, 0.00000829, 0.00000072, -0.00000094, -0.00000375, -0.00000695, 0.00001128, 0.00000268, -0.00000714, 0.00000121, 0.00000009, 0.00000050, 0.00000332, -0.00000230, 0.00000320, -0.00000148, -0.00000497, 0.00000257, 0.00000377, -0.00000153, -0.00000050, -0.00000074, -0.00000329, -0.00000401, 0.00000665, -0.00000014, -0.00000023, 0.00000799, -0.00000071, -0.00000751, -0.00000797, 0.00000634, 0.00000510, -0.00000323, 0.00000099, -0.00000490, -0.00000113, -0.00000317, 0.00000176, 0.00001033, -0.00000355, -0.00000600, -0.00000494, 0.00000437, 0.00000775, -0.00000781, 0.00000772, 0.00000105, -0.00001126, 0.00000172, 0.00000064, -0.00000097, 0.00000283, -0.00000269, 0.00001207, -0.00002033, 0.00001573, -0.00001918, 0.00001844, -0.00001219, 0.00000844, -0.00000175, -0.00000595, 0.00000867, -0.00001442, 0.00001785, -0.00001822, 0.00001361, -0.00002577, 0.00004277, -0.00002494, 0.00001336, -0.00002761, 0.00001472, 0.00000230, -0.00000758, 0.00000136, -0.00000957, 0.00001067, 0.00000476, -0.00001556, 0.00000831, -0.00001921, 0.00001689, -0.00002508, 0.00001344, -0.00000234, -0.00002249, 0.00000416, -0.00001600, 0.00001303, 0.00001728, -0.00000213, 0.00000326, -0.00000085, 0.00001521, -0.00001239, -0.00001524, 0.00001078, -0.00000175, 0.00001208, -0.00002842, -0.00003010, -0.00018152, 0.00012334, 0.00012313, 0.00004497, 0.00005098, 0.00007764, 0.00014857, -0.00005550, -0.00011014, -0.00001274, -0.00001707, -0.00001761, -0.00000529, -0.00001337, -0.00000690, -0.00000861, -0.00000303, -0.00000265, 0.00000375, -0.00000675, -0.00000471, -0.00000501, -0.00000256, },
		{ -0.00000366, 0.00000344, -0.00001115, -0.00000850, 0.00000773, 0.00000268, -0.00000156, -0.00000269, -0.00000846, 0.00000952, 0.00000562, -0.00000752, 0.00000018, 0.00000085, -0.00000032, 0.00000379, -0.00000247, 0.00000304, -0.00000005, -0.00000569, 0.00000137, 0.00000443, -0.00000113, -0.00000068, -0.00000092, -0.00000204, -0.00000523, 0.00000543, 0.00000191, -0.00000196, 0.00000818, 0.00000085, -0.00000691, -0.00000867, 0.00000395, 0.00000703, -0.00000339, 0.00000105, -0.00000409, -0.00000221, -0.00000211, -0.00000049, 0.00001089, -0.00000144, -0.00000663, -0.00000499, 0.00000189, 0.00000956, -0.00000691, 0.00000487, 0.00000464, -0.00001182, -0.00000035, 0.00000260, -0.00000299, 0.00000523, -0.00000631, 0.00001630, -0.00002207, 0.00001492, -0.00001697, 0.00001410, -0.00000649, 0.00000208, 0.00000479, -0.00001063, 0.00001204, -0.00001755, 0.00001994, -0.00001899, 0.00001551, -0.00002878, 0.00003995, -0.00001676, 0.00000747, -0.00002222, 0.00000647, 0.00000891, -0.00000957, 0.00000109, -0.00000790, 0.00000565, 0.00001104, -0.00002043, 0.00001147, -0.00002001, 0.00001473, -0.00001946, 0.00000354, 0.00000843, -0.00002718, 0.00000039, -0.00001165, 0.00000696, 0.00002178, -0.00000182, 0.00000344, 0.00000026, 0.00001422, -0.00001002, -0.00001660, 0.00001289, -0.00000367, 0.00001585, -0.00002793, -0.00002722, -0.00017795, 0.00011281, 0.00012273, 0.00003857, 0.00005468, 0.00005606, 0.00014956, -0.00001757, -0.00011847, -0.00002097, -0.00001192, -0.00002225, -0.00000332, -0.00001414, -0.00000700, -0.00000781, -0.00000446, -0.00000230, 0.00000382, -0.00000598, -0.00000502, -0.00000522, },
		{ -0.00000539, -0.00000485, 0.00000461, -0.00000915, -0.00001082, 0.00000621, 0.00000481, -0.00000181, -0.00000206, -0.00000918, 0.00000714, 0.00000818, -0.00000708, -0.00000131, 0.00000170, -0.00000104, 0.00000401, -0.00000235, 0.00000259, 0.00000136, -0.00000597, -0.00000009, 0.00000495, -0.00000065, -0.00000076, -0.00000124, -0.00000088, -0.00000579, 0.00000350, 0.00000405, -0.00000317, 0.00000763, 0.00000269, -0.00000637, -0.00000882, 0.00000127, 0.00000842, -0.00000285, 0.00000054, -0.00000283, -0.00000347, -0.00000105, -0.00000223, 0.00001050, 0.00000112, -0.00000716, -0.00000494, -0.00000043, 0.00001040, -0.00000494, 0.00000168, 0.00000743, -0.00001091, -0.00000346, 0.00000505, -0.00000530, 0.00000778, -0.00000962, 0.00001927, -0.00002185, 0.00001220, -0.00001313, 0.00000859, -0.00000030, -0.00000404, 0.00001032, -0.00001395, 0.00001422, -0.00001960, 0.00002095, -0.00001886, 0.00001646, -0.00002973, 0.00003457, -0.00000759, 0.00000198, -0.00001737, -0.00000046, 0.00001289, -0.00000876, -0.00000135, -0.00000457, -0.00000014, 0.00001663, -0.00002343, 0.00001211, -0.00001804, 0.00001004, -0.00001212, -0.00000617, 0.00001630, -0.00002733, -0.00000666, -0.00000635, 0.00000124, 0.00002473, 0.00000016, 0.00000248, 0.00000223, 0.00001272, -0.00000735, -0.00001808, 0.00001498, -0.00000529, 0.00001890, -0.00002642, -0.00002554, -0.00017326, 0.00010153, 0.00012330, 0.00003196, 0.00005810, 0.00003968, 0.00014152, 0.00001938, -0.00011882, -0.00003319, -0.00000627, -0.00002606, -0.00000258, -0.00001385, -0.00000792, -0.00000651, -0.00000586, -0.00000230, 0.00000406, -0.00000532, -0.00000514, },
		{ -0.00000516, -0.00000544, -0.00000615, 0.00000540, -0.00000691, -0.00001242, 0.00000391, 0.00000673, -0.00000146, -0.00000188, -0.00000927, 0.00000451, 0.00001004, -0.00000579, -0.00000307, 0.00000246, -0.00000151, 0.00000393, -0.00000197, 0.00000196, 0.00000261, -0.00000577, -0.00000172, 0.00000526, -0.00000008, -0.00000079, -0.00000160, 0.00000004, -0.00000564, 0.00000106, 0.00000594, -0.00000363, 0.00000633, 0.00000467, -0.00000581, -0.00000856, -0.00000138, 0.00000903, -0.00000159, -0.00000042, -0.00000135, -0.00000465, -0.00000028, -0.00000329, 0.00000921, 0.00000386, -0.00000737, -0.00000494, -0.00000237, 0.00001021, -0.00000222, -0.00000130, 0.00000896, -0.00000847, -0.00000733, 0.00000762, -0.00000751, 0.00001002, -0.00001208, 0.00002056, -0.00001957, 0.00000781, -0.00000814, 0.00000253, 0.00000569, -0.00000930, 0.00001444, -0.00001577, 0.00001525, -0.00002065, 0.00002094, -0.00001787, 0.00001640, -0.00002871, 0.00002725, 0.00000161, -0.00000231, -0.00001369, -0.00000553, 0.00001409, -0.00000554, -0.00000520, -0.00000054, -0.00000561, 0.00002057, -0.00002398, 0.00001017, -0.00001379, 0.00000381, -0.00000446, -0.00001418, 0.00002029, -0.00002316, -0.00001574, -0.00000149, -0.00000312, 0.00002557, 0.00000388, 0.00000062, 0.00000470, 0.00001108, -0.00000473, -0.00001932, 0.00001671, -0.00000631, 0.00002102, -0.00002387, -0.00002493, -0.00016776, 0.00008999, 0.00012409, 0.00002602, 0.00006017, 0.00002872, 0.00012688, 0.00005231, -0.00011083, -0.00004788, -0.00000155, -0.00002826, -0.00000333, -0.00001261, -0.00000939, -0.00000506, -0.00000690, -0.00000278, 0.00000446, -0.00000471, },
		{ -0.00000405, -0.00000519, -0.00000528, -0.00000755, 0.00000582, -0.00000458, -0.00001326, 0.00000113, 0.00000808, -0.00000041, -0.00000207, -0.00000897, 0.00000197, 0.00001100, -0.00000374, -0.00000489, 0.00000293, -0.00000164, 0.00000353, -0.00000140, 0.00000128, 0.00000358, -0.00000510, -0.00000337, 0.00000526, 0.00000061, -0.00000079, -0.00000193, 0.00000064, -0.00000485, -0.00000159, 0.00000724, -0.00000321, 0.00000439, 0.00000660, -0.00000511, -0.00000809, -0.00000369, 0.00000869, 0.00000032, -0.00000162, 0.00000010, -0.00000551, 0.00000003, -0.00000360, 0.00000723, 0.00000646, -0.00000707, -0.00000513, -0.00000381, 0.00000910, 0.00000082, -0.00000356, 0.00000897, -0.00000467, -0.00001153, 0.00000985, -0.00000924, 0.00001152, -0.00001329, 0.00001996, -0.00001541, 0.00000224, -0.00000265, -0.00000339, 0.00001086, -0.00001324, 0.00001694, -0.00001614, 0.00001528, -0.00002084, 0.00002002, -0.00001610, 0.00001534, -0.00002600, 0.00001880, 0.00000985, -0.00000476, -0.00001152, -0.00000855, 0.00001274, -0.00000063, -0.00000955, 0.00000321, -0.00000984, 0.00002223, -0.00002199, 0.00000607, -0.00000812, -0.00000278, 0.00000216, -0.00001939, 0.00002017, -0.00001560, -0.00002524, 0.00000164, -0.00000546, 0.00002416, 0.00000905, -0.00000168, 0.00000720, 0.00000975, -0.00000253, -0.00002005, 0.00001788, -0.00000659, 0.00002224, -0.00002045, -0.00002507, -0.00016190, 0.00007874, 0.00012446, 0.00002137, 0.00006037, 0.00002266, 0.00010830, 0.00007889, -0.00009513, -0.00006306, 0.00000096, -0.00002843, -0.00000543, -0.00001077, -0.00001101, -0.00000382, -0.00000738, -0.00000373, 0.00000489, },
		{ 0.00000518, -0.00000321, -0.00000533, -0.00000488, -0.00000897, 0.00000585, -0.00000225, -0.00001340, -0.00000179, 0.00000861, 0.00000127, -0.00000239, -0.00000855, -0.00000018, 0.00001098, -0.00000117, -0.00000647, 0.00000296, -0.00000140, 0.00000288, -0.00000076, 0.00000070, 0.00000418, -0.00000403, -0.00000491, 0.00000490, 0.00000141, -0.00000077, -0.00000217, 0.00000089, -0.00000356, -0.00000412, 0.00000770, -0.00000190, 0.00000206, 0.00000821, -0.00000411, -0.00000759, -0.00000543, 0.00000740, 0.00000263, -0.00000277, 0.00000126, -0.00000586, -0.00000022, -0.00000322, 0.00000482, 0.00000859, -0.00000612, -0.00000553, -0.00000476, 0.00000730, 0.00000372, -0.00000469, 0.00000741, 0.00000010, -0.00001550, 0.00001130, -0.00001011, 0.00001197, -0.00001303, 0.00001752, -0.00000978, -0.00000388, 0.00000268, -0.00000855, 0.00001474, -0.00001563, 0.00001785, -0.00001529, 0.00001457, -0.00002034, 0.00001833, -0.00001366, 0.00001346, -0.00002202, 0.00001009, 0.00001630, -0.00000502, -0.00001095, -0.00000964, 0.00000938, 0.00000507, -0.00001342, 0.00000585, -0.00001222, 0.00002142, -0.00001782, 0.00000063, -0.00000211, -0.00000855, 0.00000671, -0.00002127, 0.00001636, -0.00000606, -0.00003355, 0.00000209, -0.00000556, 0.00002074, 0.00001508, -0.00000379, 0.00000922, 0.00000908, -0.00000097, -0.00002013, 0.00001843, -0.00000620, 0.00002273, -0.00001644, -0.00002558, -0.00015609, 0.00006822, 0.00012396, 0.00001828, 0.00005867, 0.00002048, 0.00008832, 0.00009777, -0.00007325, -0.00007661, 0.00000036, -0.00002659, -0.00000849, -0.00000880, -0.00001239, -0.00000306, -0.00000723, -0.00000502, },
		{ -0.00000641, 0.00000512, -0.00000211, -0.00000560, -0.00000427, -0.00001032, 0.00000546, 0.00000001, -0.00001296, -0.00000455, 0.00000822, 0.00000338, -0.00000256, -0.00000829, -0.00000175, 0.00001005, 0.00000163, -0.00000756, 0.00000246, -0.00000082, 0.00000208, -0.00000018, 0.00000030, 0.00000440, -0.00000270, -0.00000618, 0.00000412, 0.00000230, -0.00000074, -0.00000231, 0.00000082, -0.00000197, -0.00000622, 0.00000718, 0.00000014, -0.00000034, 0.00000924, -0.00000269, -0.00000720, -0.00000649, 0.00000528, 0.00000504, -0.00000357, 0.00000193, -0.00000561, -0.00000101, -0.00000235, 0.00000235, 0.00001000, -0.00000449, -0.00000609, -0.00000531, 0.00000514, 0.00000602, -0.00000446, 0.00000448, 0.00000527, -0.00001862, 0.00001156, -0.00000989, 0.00001121, -0.00001129, 0.00001354, -0.00000329, -0.00000984, 0.00000720, -0.00001248, 0.00001708, -0.00001644, 0.00001737, -0.00001354, 0.00001338, -0.00001931, 0.00001602, -0.00001074, 0.00001098, -0.00001732, 0.00000198, 0.00002037, -0.00000305, -0.00001174, -0.00000918, 0.00000477, 0.00001058, -0.00001601, 0.00000684, -0.00001252, 0.00001841, -0.00001219, -0.00000515, 0.00000317, -0.00001256, 0.00000863, -0.00001994, 0.00000986, 0.00000380, -0.00003925, -0.00000059, -0.00000369, 0.00001591, 0.00002121, -0.00000506, 0.00001037, 0.00000928, -0.00000011, -0.00001960, 0.00001848, -0.00000531, 0.00002276, -0.00001218, -0.00002612, -0.00015067, 0.00005872, 0.00012241, 0.00001670, 0.00005543, 0.00002093, 0.00006907, 0.00010856, -0.00004736, -0.00008661, -0.00000373, -0.00002320, -0.00001186, -0.00000721, -0.00001322, -0.00000290, -0.00000655, },
		{ -0.00000554, -0.00000761, 0.00000455, -0.00000071, -0.00000597, -0.00000355, -0.00001143, 0.00000460, 0.00000218, -0.00001206, -0.00000692, 0.00000693, 0.00000558, -0.00000229, -0.00000837, -0.00000267, 0.00000843, 0.00000429, -0.00000794, 0.00000142, -0.00000002, 0.00000128, 0.00000022, 0.00000017, 0.00000427, -0.00000126, -0.00000704, 0.00000294, 0.00000319, -0.00000065, -0.00000237, 0.00000052, -0.00000033, -0.00000765, 0.00000568, 0.00000262, -0.00000243, 0.00000946, -0.00000081, -0.00000697, -0.00000688, 0.00000258, 0.00000718, -0.00000376, 0.00000197, -0.00000476, -0.00000218, -0.00000124, 0.00000014, 0.00001049, -0.00000227, -0.00000665, -0.00000565, 0.00000295, 0.00000739, -0.00000284, 0.00000060, 0.00001018, -0.00002037, 0.00001040, -0.00000852, 0.00000928, -0.00000828, 0.00000851, 0.00000336, -0.00001495, 0.00001044, -0.00001491, 0.00001783, -0.00001585, 0.00001583, -0.00001126, 0.00001199, -0.00001793, 0.00001326, -0.00000755, 0.00000823, -0.00001248, -0.00000484, 0.00002176, 0.00000087, -0.00001340, -0.00000777, -0.00000027, 0.00001503, -0.00001675, 0.00000599, -0.00001091, 0.00001378, -0.00000605, -0.00001025, 0.00000687, -0.00001430, 0.00000789, -0.00001603, 0.00000205, 0.00001240, -0.00004141, -0.00000625, -0.00000057, 0.00001047, 0.00002667, -0.00000499, 0.00001042, 0.00001038, 0.00000016, -0.00001863, 0.00001825, -0.00000419, 0.00002264, -0.00000795, -0.00002643, -0.00014585, 0.00005032, 0.00011985, 0.00001637, 0.00005117, 0.00002278, 0.00005205, 0.00011177, -0.00001990, -0.00009159, -0.00001107, -0.00001907, -0.00001485, -0.00000641, -0.00001336, -0.00000327, },
		{ -0.00000397, -0.00000452, -0.00000838, 0.00000338, 0.00000089, -0.00000632, -0.00000289, -0.00001217, 0.00000327, 0.00000420, -0.00001081, -0.00000876, 0.00000493, 0.00000751, -0.00000135, -0.00000883, -0.00000305, 0.00000640, 0.00000649, -0.00000748, -0.00000006, 0.00000084, 0.00000062, 0.00000036, 0.00000031, 0.00000390, 0.00000012, -0.00000739, 0.00000141, 0.00000401, -0.00000045, -0.00000239, 0.00000010, 0.00000117, -0.00000827, 0.00000335, 0.00000516, -0.00000386, 0.00000874, 0.00000148, -0.00000685, -0.00000670, -0.00000035, 0.00000870, -0.00000319, 0.00000141, -0.00000346, -0.00000352, -0.00000019, -0.00000154, 0.00001002, 0.00000035, -0.00000701, -0.00000597, 0.00000103, 0.00000765, -0.00000005, -0.00000366, 0.00001414, -0.00002035, 0.00000777, -0.00000612, 0.00000639, -0.00000440, 0.00000304, 0.00000946, -0.00001866, 0.00001211, -0.00001580, 0.00001716, -0.00001420, 0.00001365, -0.00000881, 0.00001059, -0.00001630, 0.00001021, -0.00000434, 0.00000555, -0.00000803, -0.00000984, 0.00002050, 0.00000617, -0.00001532, -0.00000605, -0.00000494, 0.00001778, -0.00001544, 0.00000347, -0.00000789, 0.00000839, -0.00000038, -0.00001384, 0.00000847, -0.00001368, 0.00000492, -0.00001058, -0.00000563, 0.00001844, -0.00003970, -0.00001417, 0.00000283, 0.00000524, 0.00003078, -0.00000327, 0.00000940, 0.00001217, 0.00000012, -0.00001751, 0.00001799, -0.00000307, 0.00002259, -0.00000396, -0.00002637, -0.00014166, 0.00004296, 0.00011646, 0.00001690, 0.00004650, 0.00002502, 0.00003814, 0.00010862, 0.00000672, -0.00009077, -0.00002081, -0.00001518, -0.00001684, -0.00000660, -0.00001288, },
		{ -0.00001204, -0.00000469, -0.00000377, -0.00000853, 0.00000166, 0.00000253, -0.00000646, -0.00000244, -0.00001242, 0.00000150, 0.00000599, -0.00000927, -0.00001006, 0.00000250, 0.00000881, 0.00000036, -0.00000959, -0.00000308, 0.00000432, 0.00000795, -0.00000620, -0.00000180, 0.00000157, 0.00000023, 0.00000020, 0.00000067, 0.00000340, 0.00000130, -0.00000720, -0.00000034, 0.00000462, -0.00000009, -0.00000241, -0.00000034, 0.00000237, -0.00000806, 0.00000048, 0.00000735, -0.00000435, 0.00000708, 0.00000401, -0.00000672, -0.00000616, -0.00000317, 0.00000933, -0.00000183, 0.00000038, -0.00000194, -0.00000475, 0.00000054, -0.00000251, 0.00000868, 0.00000306, -0.00000694, -0.00000643, -0.00000043, 0.00000679, 0.00000354, -0.00000764, 0.00001660, -0.00001838, 0.00000386, -0.00000299, 0.00000291, -0.00000014, -0.00000224, 0.00001441, -0.00002064, 0.00001218, -0.00001531, 0.00001540, -0.00001190, 0.00001122, -0.00000647, 0.00000932, -0.00001453, 0.00000706, -0.00000137, 0.00000326, -0.00000440, -0.00001278, 0.00001694, 0.00001207, -0.00001682, -0.00000460, -0.00000864, 0.00001850, -0.00001221, -0.00000024, -0.00000418, 0.00000311, 0.00000397, -0.00001540, 0.00000789, -0.00001107, 0.00000055, -0.00000480, -0.00001189, 0.00002119, -0.00003446, -0.00002321, 0.00000547, 0.00000092, 0.00003314, 0.00000009, 0.00000758, 0.00001432, 0.00000010, -0.00001652, 0.00001794, -0.00000215, 0.00002276, -0.00000027, -0.00002596, -0.00013803, 0.00003645, 0.00011252, 0.00001787, 0.00004190, 0.00002692, 0.00002759, 0.00010069, 0.00003045, -0.00008410, -0.00003172, -0.00001253, -0.00001746, -0.00000774, },
		{ -0.00000951, -0.00001121, -0.00000509, -0.00000351, -0.00000804, -0.00000042, 0.00000394, -0.00000623, -0.00000233, -0.00001215, -0.00000057, 0.00000740, -0.00000747, -0.00001086, -0.00000004, 0.00000921, 0.00000278, -0.00001043, -0.00000303, 0.00000250, 0.00000851, -0.00000424, -0.00000354, 0.00000198, 0.00000019, -0.00000026, 0.00000117, 0.00000291, 0.00000214, -0.00000651, -0.00000216, 0.00000490, 0.00000047, -0.00000247, -0.00000073, 0.00000319, -0.00000712, -0.00000256, 0.00000879, -0.00000374, 0.00000462, 0.00000652, -0.00000639, -0.00000549, -0.00000555, 0.00000893, 0.00000022, -0.00000088, -0.00000047, -0.00000562, 0.00000076, -0.00000272, 0.00000668, 0.00000554, -0.00000626, -0.00000713, -0.00000135, 0.00000499, 0.00000738, -0.00001070, 0.00001716, -0.00001456, -0.00000094, 0.00000042, -0.00000070, 0.00000397, -0.00000675, 0.00001777, -0.00002079, 0.00001082, -0.00001381, 0.00001298, -0.00000937, 0.00000890, -0.00000446, 0.00000827, -0.00001267, 0.00000399, 0.00000113, 0.00000162, -0.00000185, -0.00001367, 0.00001169, 0.00001775, -0.00001730, -0.00000385, -0.00001100, 0.00001718, -0.00000753, -0.00000443, -0.00000057, -0.00000125, 0.00000643, -0.00001478, 0.00000545, -0.00000719, -0.00000421, 0.00000019, -0.00001582, 0.00002047, -0.00002656, -0.00003203, 0.00000642, -0.00000207, 0.00003363, 0.00000484, 0.00000541, 0.00001640, 0.00000045, -0.00001589, 0.00001825, -0.00000149, 0.00002316, 0.00000318, -0.00002530, -0.00013480, 0.00003059, 0.00010831, 0.00001895, 0.00003774, 0.00002809, 0.00002018, 0.00008976, 0.00004980, -0.00007223, -0.00004230, -0.00001190, -0.00001664, },
		{ -0.00001468, -0.00001144, -0.00001080, -0.00000494, -0.00000382, -0.00000703, -0.00000258, 0.00000486, -0.00000547, -0.00000262, -0.00001142, -0.00000274, 0.00000829, -0.00000541, -0.00001128, -0.00000240, 0.00000861, 0.00000569, -0.00001105, -0.00000316, 0.00000116, 0.00000816, -0.00000182, -0.00000500, 0.00000193, 0.00000051, -0.00000093, 0.00000168, 0.00000257, 0.00000259, -0.00000541, -0.00000386, 0.00000476, 0.00000123, -0.00000256, -0.00000104, 0.00000362, -0.00000566, -0.00000540, 0.00000921, -0.00000206, 0.00000167, 0.00000868, -0.00000566, -0.00000490, -0.00000727, 0.00000751, 0.00000268, -0.00000205, 0.00000066, -0.00000593, 0.00000039, -0.00000226, 0.00000432, 0.00000748, -0.00000487, -0.00000807, -0.00000179, 0.00000258, 0.00001092, -0.00001234, 0.00001570, -0.00000922, -0.00000607, 0.00000362, -0.00000400, 0.00000746, -0.00001006, 0.00001936, -0.00001924, 0.00000842, -0.00001172, 0.00001034, -0.00000699, 0.00000693, -0.00000288, 0.00000742, -0.00001080, 0.00000116, 0.00000295, 0.00000080, -0.00000046, -0.00001280, 0.00000554, 0.00002239, -0.00001634, -0.00000401, -0.00001195, 0.00001412, -0.00000207, -0.00000834, 0.00000224, -0.00000412, 0.00000679, -0.00001226, 0.00000179, -0.00000291, -0.00000839, 0.00000351, -0.00001700, 0.00001673, -0.00001728, -0.00003929, 0.00000510, -0.00000370, 0.00003244, 0.00001051, 0.00000341, 0.00001799, 0.00000142, -0.00001573, 0.00001893, -0.00000107, 0.00002372, 0.00000650, -0.00002454, -0.00013181, 0.00002518, 0.00010402, 0.00001991, 0.00003419, 0.00002841, 0.00001538, 0.00007746, 0.00006392, -0.00005641, -0.00005108, -0.00001369, },
		{ -0.00001786, -0.00001215, -0.00001297, -0.00001115, -0.00000414, -0.00000460, -0.00000575, -0.00000448, 0.00000506, -0.00000414, -0.00000324, -0.00001040, -0.00000477, 0.00000850, -0.00000307, -0.00001140, -0.00000436, 0.00000702, 0.00000875, -0.00001112, -0.00000369, 0.00000041, 0.00000704, 0.00000073, -0.00000594, 0.00000135, 0.00000114, -0.00000169, 0.00000208, 0.00000245, 0.00000267, -0.00000406, -0.00000527, 0.00000414, 0.00000214, -0.00000264, -0.00000129, 0.00000375, -0.00000393, -0.00000767, 0.00000845, 0.00000049, -0.00000137, 0.00001016, -0.00000438, -0.00000453, -0.00000823, 0.00000524, 0.00000520, -0.00000281, 0.00000123, -0.00000558, -0.00000053, -0.00000134, 0.00000196, 0.00000862, -0.00000276, -0.00000913, -0.00000191, -0.00000007, 0.00001362, -0.00001224, 0.00001237, -0.00000294, -0.00001088, 0.00000612, -0.00000661, 0.00000999, -0.00001196, 0.00001922, -0.00001635, 0.00000547, -0.00000952, 0.00000786, -0.00000502, 0.00000543, -0.00000171, 0.00000674, -0.00000896, -0.00000126, 0.00000395, 0.00000083, -0.00000013, -0.00001061, -0.00000071, 0.00002536, -0.00001381, -0.00000504, -0.00001164, 0.00000984, 0.00000339, -0.00001128, 0.00000375, -0.00000525, 0.00000521, -0.00000841, -0.00000224, 0.00000086, -0.00001120, 0.00000468, -0.00001554, 0.00001086, -0.00000803, -0.00004394, 0.00000138, -0.00000427, 0.00002996, 0.00001650, 0.00000213, 0.00001880, 0.00000313, -0.00001603, 0.00001989, -0.00000073, 0.00002429, 0.00000987, -0.00002383, -0.00012892, 0.00002006, 0.00009979, 0.00002066, 0.00003126, 0.00002797, 0.00001252, 0.00006516, 0.00007266, -0.00003821, -0.00005685, },
		{ -0.00005882, -0.00002394, -0.00000982, -0.00001359, -0.00001241, -0.00000279, -0.00000561, -0.00000454, -0.00000581, 0.00000441, -0.00000231, -0.00000404, -0.00000932, -0.00000637, 0.00000793, -0.00000052, -0.00001128, -0.00000581, 0.00000465, 0.00001153, -0.00001037, -0.00000470, 0.00000022, 0.00000537, 0.00000307, -0.00000616, 0.00000025, 0.00000194, -0.00000241, 0.00000225, 0.00000261, 0.00000242, -0.00000262, -0.00000623, 0.00000306, 0.00000312, -0.00000264, -0.00000152, 0.00000368, -0.00000220, -0.00000911, 0.00000656, 0.00000356, -0.00000405, 0.00001068, -0.00000250, -0.00000441, -0.00000847, 0.00000240, 0.00000739, -0.00000289, 0.00000113, -0.00000461, -0.00000185, -0.00000022, -0.00000009, 0.00000880, -0.00000009, -0.00001011, -0.00000195, -0.00000254, 0.00001508, -0.00001033, 0.00000758, 0.00000358, -0.00001475, 0.00000756, -0.00000829, 0.00001138, -0.00001241, 0.00001759, -0.00001263, 0.00000250, -0.00000760, 0.00000581, -0.00000360, 0.00000442, -0.00000091, 0.00000618, -0.00000722, -0.00000314, 0.00000408, 0.00000163, -0.00000063, -0.00000768, -0.00000629, 0.00002624, -0.00000980, -0.00000667, -0.00001045, 0.00000499, 0.00000810, -0.00001273, 0.00000372, -0.00000472, 0.00000215, -0.00000403, -0.00000581, 0.00000341, -0.00001217, 0.00000368, -0.00001202, 0.00000403, -0.00000014, -0.00004535, -0.00000442, -0.00000435, 0.00002672, 0.00002221, 0.00000197, 0.00001869, 0.00000553, -0.00001662, 0.00002094, -0.00000031, 0.00002471, 0.00001344, -0.00002327, -0.00012608, 0.00001520, 0.00009565, 0.00002118, 0.00002887, 0.00002698, 0.00001093, 0.00005384, 0.00007639, -0.00001930, },
		{ -0.00000125, -0.00005672, -0.00003104, -0.00000847, -0.00001300, -0.00001451, -0.00000118, -0.00000651, -0.00000375, -0.00000635, 0.00000293, -0.00000019, -0.00000474, -0.00000845, -0.00000733, 0.00000655, 0.00000214, -0.00001094, -0.00000676, 0.00000180, 0.00001362, -0.00000863, -0.00000614, 0.00000042, 0.00000348, 0.00000488, -0.00000557, -0.00000125, 0.00000276, -0.00000293, 0.00000213, 0.00000303, 0.00000196, -0.00000126, -0.00000666, 0.00000159, 0.00000402, -0.00000246, -0.00000180, 0.00000353, -0.00000070, -0.00000958, 0.00000372, 0.00000669, -0.00000595, 0.00001008, -0.00000007, -0.00000446, -0.00000816, -0.00000061, 0.00000886, -0.00000212, 0.00000037, -0.00000316, -0.00000332, 0.00000081, -0.00000157, 0.00000804, 0.00000290, -0.00001074, -0.00000216, -0.00000448, 0.00001507, -0.00000681, 0.00000195, 0.00000957, -0.00001716, 0.00000772, -0.00000896, 0.00001165, -0.00001160, 0.00001490, -0.00000862, -0.00000003, -0.00000625, 0.00000432, -0.00000274, 0.00000381, -0.00000034, 0.00000568, -0.00000564, -0.00000439, 0.00000340, 0.00000299, -0.00000158, -0.00000458, -0.00001060, 0.00002495, -0.00000472, -0.00000846, -0.00000883, 0.00000022, 0.00001149, -0.00001245, 0.00000226, -0.00000291, -0.00000170, 0.00000007, -0.00000819, 0.00000430, -0.00001126, 0.00000089, -0.00000732, -0.00000252, 0.00000539, -0.00004345, -0.00001153, -0.00000467, 0.00002324, 0.00002714, 0.00000317, 0.00001771, 0.00000844, -0.00001725, 0.00002185, 0.00000038, 0.00002485, 0.00001727, -0.00002290, -0.00012329, 0.00001061, 0.00009155, 0.00002154, 0.00002689, 0.00002566, 0.00001004, 0.00004408, 0.00007592, },
		{ 0.00007226, 0.00001471, -0.00005085, -0.00003809, -0.00000874, -0.00001114, -0.00001713, 0.00000030, -0.00000693, -0.00000362, -0.00000603, 0.00000082, 0.00000189, -0.00000507, -0.00000805, -0.00000751, 0.00000445, 0.00000472, -0.00001030, -0.00000730, -0.00000114, 0.00001466, -0.00000590, -0.00000780, 0.00000076, 0.00000170, 0.00000593, -0.00000421, -0.00000294, 0.00000339, -0.00000315, 0.00000170, 0.00000365, 0.00000142, -0.00000015, -0.00000653, -0.00000011, 0.00000467, -0.00000201, -0.00000219, 0.00000343, 0.00000042, -0.00000912, 0.00000030, 0.00000940, -0.00000674, 0.00000837, 0.00000268, -0.00000450, -0.00000754, -0.00000341, 0.00000933, -0.00000046, -0.00000091, -0.00000147, -0.00000468, 0.00000149, -0.00000234, 0.00000645, 0.00000587, -0.00001075, -0.00000274, -0.00000567, 0.00001359, -0.00000210, -0.00000380, 0.00001434, -0.00001779, 0.00000662, -0.00000873, 0.00001094, -0.00000984, 0.00001165, -0.00000487, -0.00000177, -0.00000560, 0.00000339, -0.00000234, 0.00000347, 0.00000009, 0.00000521, -0.00000430, -0.00000496, 0.00000203, 0.00000464, -0.00000261, -0.00000183, -0.00001327, 0.00002167, 0.00000087, -0.00000992, -0.00000724, -0.00000392, 0.00001319, -0.00001050, -0.00000029, -0.00000037, -0.00000560, 0.00000314, -0.00000893, 0.00000344, -0.00000875, -0.00000298, -0.00000245, -0.00000775, 0.00000802, -0.00003868, -0.00001891, -0.00000592, 0.00001987, 0.00003097, 0.00000577, 0.00001610, 0.00001154, -0.00001765, 0.00002242, 0.00000147, 0.00002464, 0.00002139, -0.00002266, -0.00012064, 0.00000636, 0.00008744, 0.00002181, 0.00002520, 0.00002420, 0.00000944, 0.00003611, },
		{ 0.00002992, 0.00006644, 0.00002776, -0.00004191, -0.00004397, -0.00001100, -0.00000828, -0.00001979, 0.00000121, -0.00000658, -0.00000427, -0.00000496, -0.00000159, 0.00000356, -0.00000475, -0.00000826, -0.00000690, 0.00000185, 0.00000695, -0.00000926, -0.00000760, -0.00000384, 0.00001447, -0.00000239, -0.00000936, 0.00000094, 0.00000030, 0.00000612, -0.00000223, -0.00000459, 0.00000367, -0.00000301, 0.00000099, 0.00000437, 0.00000093, 0.00000063, -0.00000592, -0.00000186, 0.00000493, -0.00000125, -0.00000272, 0.00000341, 0.00000110, -0.00000787, -0.00000327, 0.00001124, -0.00000626, 0.00000573, 0.00000543, -0.00000427, -0.00000689, -0.00000565, 0.00000864, 0.00000196, -0.00000244, 0.00000018, -0.00000567, 0.00000166, -0.00000239, 0.00000430, 0.00000845, -0.00000994, -0.00000380, -0.00000603, 0.00001084, 0.00000322, -0.00000893, 0.00001736, -0.00001655, 0.00000445, -0.00000785, 0.00000957, -0.00000753, 0.00000832, -0.00000180, -0.00000252, -0.00000562, 0.00000289, -0.00000223, 0.00000324, 0.00000050, 0.00000477, -0.00000324, -0.00000490, 0.00000019, 0.00000623, -0.00000335, 0.00000019, -0.00001419, 0.00001686, 0.00000627, -0.00001059, -0.00000604, -0.00000706, 0.00001313, -0.00000720, -0.00000338, 0.00000224, -0.00000885, 0.00000470, -0.00000791, 0.00000110, -0.00000525, -0.00000712, 0.00000165, -0.00001094, 0.00000774, -0.00003194, -0.00002546, -0.00000860, 0.00001676, 0.00003358, 0.00000959, 0.00001421, 0.00001449, -0.00001755, 0.00002251, 0.00000299, 0.00002413, 0.00002573, -0.00002244, -0.00011824, 0.00000250, 0.00008327, 0.00002204, 0.00002371, 0.00002271, 0.00000887, },
		{ 0.00000817, 0.00002532, 0.00005940, 0.00003756, -0.00003093, -0.00004775, -0.00001521, -0.00000498, -0.00002193, 0.00000117, -0.00000535, -0.00000561, -0.00000345, -0.00000388, 0.00000445, -0.00000361, -0.00000912, -0.00000564, -0.00000092, 0.00000856, -0.00000771, -0.00000782, -0.00000601, 0.00001301, 0.00000157, -0.00001045, 0.00000070, -0.00000054, 0.00000546, 0.00000011, -0.00000591, 0.00000346, -0.00000251, 0.00000008, 0.00000506, 0.00000060, 0.00000105, -0.00000494, -0.00000343, 0.00000466, -0.00000015, -0.00000334, 0.00000349, 0.00000138, -0.00000608, -0.00000656, 0.00001190, -0.00000455, 0.00000252, 0.00000779, -0.00000352, -0.00000648, -0.00000708, 0.00000681, 0.00000486, -0.00000392, 0.00000150, -0.00000610, 0.00000125, -0.00000184, 0.00000192, 0.00001030, -0.00000820, -0.00000534, -0.00000564, 0.00000720, 0.00000847, -0.00001278, 0.00001836, -0.00001362, 0.00000163, -0.00000669, 0.00000786, -0.00000508, 0.00000536, 0.00000030, -0.00000228, -0.00000614, 0.00000263, -0.00000224, 0.00000299, 0.00000091, 0.00000439, -0.00000253, -0.00000427, -0.00000185, 0.00000746, -0.00000353, 0.00000129, -0.00001351, 0.00001116, 0.00001081, -0.00001013, -0.00000544, -0.00000906, 0.00001147, -0.00000308, -0.00000638, 0.00000432, -0.00001095, 0.00000453, -0.00000535, -0.00000217, -0.00000147, -0.00001070, 0.00000428, -0.00001181, 0.00000501, -0.00002436, -0.00003017, -0.00001290, 0.00001381, 0.00003501, 0.00001433, 0.00001245, 0.00001698, -0.00001678, 0.00002208, 0.00000490, 0.00002340, 0.00003017, -0.00002211, -0.00011619, -0.00000091, 0.00007901, 0.00002224, 0.00002237, 0.00002129, },
		{ 0.00001995, 0.00000727, 0.00002207, 0.00005190, 0.00004413, -0.00001906, -0.00004879, -0.00002096, -0.00000195, -0.00002300, -0.00000002, -0.00000331, -0.00000737, -0.00000190, -0.00000561, 0.00000431, -0.00000163, -0.00001049, -0.00000400, -0.00000348, 0.00000926, -0.00000560, -0.00000804, -0.00000750, 0.00001046, 0.00000550, -0.00001068, -0.00000014, -0.00000078, 0.00000414, 0.00000250, -0.00000668, 0.00000271, -0.00000172, -0.00000090, 0.00000563, 0.00000048, 0.00000114, -0.00000378, -0.00000463, 0.00000380, 0.00000120, -0.00000399, 0.00000359, 0.00000137, -0.00000406, -0.00000916, 0.00001123, -0.00000187, -0.00000078, 0.00000934, -0.00000205, -0.00000645, -0.00000760, 0.00000403, 0.00000786, -0.00000499, 0.00000228, -0.00000588, 0.00000032, -0.00000091, -0.00000033, 0.00001116, -0.00000555, -0.00000719, -0.00000473, 0.00000313, 0.00001303, -0.00001491, 0.00001733, -0.00000943, -0.00000136, -0.00000561, 0.00000613, -0.00000284, 0.00000306, 0.00000135, -0.00000121, -0.00000690, 0.00000237, -0.00000221, 0.00000265, 0.00000134, 0.00000413, -0.00000215, -0.00000324, -0.00000381, 0.00000807, -0.00000301, 0.00000148, -0.00001158, 0.00000528, 0.00001393, -0.00000839, -0.00000543, -0.00000998, 0.00000861, 0.00000120, -0.00000869, 0.00000542, -0.00001165, 0.00000274, -0.00000172, -0.00000570, 0.00000183, -0.00001309, 0.00000508, -0.00001052, 0.00000067, -0.00001716, -0.00003235, -0.00001862, 0.00001071, 0.00003542, 0.00001962, 0.00001124, 0.00001880, -0.00001529, 0.00002120, 0.00000707, 0.00002261, 0.00003459, -0.00002156, -0.00011455, -0.00000389, 0.00007470, 0.00002238, 0.00002117, },
		{ 0.00002012, 0.00001875, 0.00000612, 0.00001993, 0.00004446, 0.00004780, -0.00000737, -0.00004689, -0.00002753, 0.00000005, -0.00002266, -0.00000230, -0.00000075, -0.00000912, -0.00000076, -0.00000642, 0.00000303, 0.00000104, -0.00001209, -0.00000237, -0.00000545, 0.00000888, -0.00000294, -0.00000830, -0.00000829, 0.00000715, 0.00000889, -0.00000979, -0.00000158, -0.00000052, 0.00000241, 0.00000460, -0.00000671, 0.00000146, -0.00000076, -0.00000183, 0.00000599, 0.00000060, 0.00000099, -0.00000261, -0.00000531, 0.00000239, 0.00000268, -0.00000453, 0.00000365, 0.00000122, -0.00000211, -0.00001081, 0.00000932, 0.00000137, -0.00000363, 0.00000973, 0.00000025, -0.00000686, -0.00000726, 0.00000060, 0.00001055, -0.00000537, 0.00000238, -0.00000505, -0.00000097, 0.00000012, -0.00000213, 0.00001092, -0.00000217, -0.00000911, -0.00000360, -0.00000087, 0.00001636, -0.00001512, 0.00001454, -0.00000457, -0.00000396, -0.00000495, 0.00000466, -0.00000105, 0.00000157, 0.00000140, 0.00000041, -0.00000759, 0.00000194, -0.00000204, 0.00000219, 0.00000176, 0.00000401, -0.00000210, -0.00000197, -0.00000542, 0.00000789, -0.00000180, 0.00000091, -0.00000889, -0.00000009, 0.00001527, -0.00000548, -0.00000585, -0.00001006, 0.00000505, 0.00000504, -0.00000986, 0.00000532, -0.00001099, -0.00000029, 0.00000231, -0.00000878, 0.00000409, -0.00001394, 0.00000408, -0.00000762, -0.00000427, -0.00001137, -0.00003172, -0.00002522, 0.00000704, 0.00003499, 0.00002508, 0.00001088, 0.00001986, -0.00001313, 0.00002003, 0.00000932, 0.00002191, 0.00003886, -0.00002072, -0.00011335, -0.00000647, 0.00007039, 0.00002242, },
		{ 0.00002234, 0.00001918, 0.00001771, 0.00000468, 0.00001871, 0.00003742, 0.00004908, 0.00000329, -0.00004226, -0.00003395, 0.00000037, -0.00002082, -0.00000537, 0.00000184, -0.00001040, -0.00000041, -0.00000612, 0.00000071, 0.00000406, -0.00001355, -0.00000111, -0.00000649, 0.00000737, 0.00000010, -0.00000850, -0.00000853, 0.00000351, 0.00001125, -0.00000765, -0.00000350, 0.00000002, 0.00000061, 0.00000611, -0.00000593, -0.00000019, 0.00000021, -0.00000259, 0.00000609, 0.00000090, 0.00000073, -0.00000162, -0.00000538, 0.00000054, 0.00000409, -0.00000482, 0.00000355, 0.00000106, -0.00000045, -0.00001141, 0.00000642, 0.00000462, -0.00000551, 0.00000876, 0.00000330, -0.00000761, -0.00000627, -0.00000304, 0.00001251, -0.00000487, 0.00000181, -0.00000374, -0.00000236, 0.00000097, -0.00000324, 0.00000960, 0.00000164, -0.00001076, -0.00000258, -0.00000434, 0.00001814, -0.00001347, 0.00001049, 0.00000029, -0.00000570, -0.00000491, 0.00000356, 0.00000018, 0.00000088, 0.00000066, 0.00000224, -0.00000794, 0.00000120, -0.00000173, 0.00000166, 0.00000209, 0.00000407, -0.00000229, -0.00000069, -0.00000646, 0.00000688, -0.00000004, -0.00000009, -0.00000596, -0.00000441, 0.00001472, -0.00000171, -0.00000637, -0.00000961, 0.00000133, 0.00000790, -0.00000967, 0.00000406, -0.00000924, -0.00000399, 0.00000607, -0.00001082, 0.00000495, -0.00001322, 0.00000164, -0.00000388, -0.00000880, -0.00000771, -0.00002848, -0.00003188, 0.00000239, 0.00003384, 0.00003040, 0.00001158, 0.00002019, -0.00001048, 0.00001878, 0.00001148, 0.00002142, 0.00004295, -0.00001955, -0.00011254, -0.00000875, 0.00006616, },
		{ 0.00006207, 0.00002211, 0.00001830, 0.00001693, 0.00000290, 0.00001821, 0.00003098, 0.00004851, 0.00001236, -0.00003547, -0.00003926, -0.00000135, -0.00001773, -0.00000869, 0.00000392, -0.00001079, -0.00000111, -0.00000472, -0.00000233, 0.00000699, -0.00001445, -0.00000058, -0.00000644, 0.00000483, 0.00000325, -0.00000847, -0.00000844, 0.00000001, 0.00001223, -0.00000435, -0.00000560, 0.00000051, -0.00000093, 0.00000683, -0.00000440, -0.00000204, 0.00000102, -0.00000309, 0.00000595, 0.00000132, 0.00000048, -0.00000093, -0.00000487, -0.00000156, 0.00000522, -0.00000474, 0.00000323, 0.00000099, 0.00000078, -0.00001103, 0.00000298, 0.00000730, -0.00000605, 0.00000641, 0.00000689, -0.00000846, -0.00000492, -0.00000646, 0.00001343, -0.00000345, 0.00000071, -0.00000220, -0.00000360, 0.00000139, -0.00000355, 0.00000738, 0.00000547, -0.00001182, -0.00000195, -0.00000698, 0.00001825, -0.00001028, 0.00000585, 0.00000450, -0.00000625, -0.00000557, 0.00000285, 0.00000087, 0.00000087, -0.00000056, 0.00000394, -0.00000776, 0.00000014, -0.00000130, 0.00000113, 0.00000228, 0.00000432, -0.00000263, 0.00000044, -0.00000680, 0.00000512, 0.00000201, -0.00000119, -0.00000326, -0.00000731, 0.00001244, 0.00000244, -0.00000663, -0.00000900, -0.00000208, 0.00000946, -0.00000811, 0.00000191, -0.00000686, -0.00000775, 0.00000893, -0.00001148, 0.00000439, -0.00001122, -0.00000165, -0.00000012, -0.00001211, -0.00000647, -0.00002325, -0.00003769, -0.00000351, 0.00003194, 0.00003535, 0.00001341, 0.00001996, -0.00000755, 0.00001766, 0.00001341, 0.00002119, 0.00004685, -0.00001811, -0.00011206, -0.00001083, },
		{ -0.00001278, 0.00005814, 0.00002179, 0.00001738, 0.00001648, 0.00000074, 0.00001829, 0.00002526, 0.00004655, 0.00001962, -0.00002737, -0.00004262, -0.00000512, -0.00001395, -0.00001161, 0.00000499, -0.00001001, -0.00000286, -0.00000248, -0.00000560, 0.00000932, -0.00001444, -0.00000096, -0.00000530, 0.00000155, 0.00000616, -0.00000798, -0.00000827, -0.00000294, 0.00001167, -0.00000017, -0.00000746, 0.00000064, -0.00000196, 0.00000666, -0.00000225, -0.00000385, 0.00000152, -0.00000331, 0.00000562, 0.00000175, 0.00000036, -0.00000060, -0.00000388, -0.00000367, 0.00000586, -0.00000421, 0.00000268, 0.00000103, 0.00000158, -0.00000990, -0.00000050, 0.00000888, -0.00000503, 0.00000286, 0.00001066, -0.00000913, -0.00000354, -0.00000927, 0.00001312, -0.00000122, -0.00000068, -0.00000069, -0.00000446, 0.00000125, -0.00000310, 0.00000457, 0.00000892, -0.00001201, -0.00000189, -0.00000861, 0.00001681, -0.00000608, 0.00000131, 0.00000752, -0.00000548, -0.00000680, 0.00000241, 0.00000113, 0.00000132, -0.00000196, 0.00000521, -0.00000695, -0.00000117, -0.00000089, 0.00000069, 0.00000228, 0.00000471, -0.00000299, 0.00000123, -0.00000643, 0.00000282, 0.00000403, -0.00000202, -0.00000117, -0.00000867, 0.00000883, 0.00000641, -0.00000628, -0.00000849, -0.00000485, 0.00000959, -0.00000542, -0.00000070, -0.00000438, -0.00001099, 0.00001050, -0.00001066, 0.00000264, -0.00000843, -0.00000509, 0.00000291, -0.00001370, -0.00000750, -0.00001693, -0.00004177, -0.00001069, 0.00002913, 0.00003977, 0.00001637, 0.00001940, -0.00000458, 0.00001685, 0.00001502, 0.00002123, 0.00005061, -0.00001646, -0.00011181, },
		},
		{ { -0.00000449, -0.00000624, -0.00000097, 0.00000166, -0.00000876, -0.00000710, -0.00001049, -0.00000868, -0.00001425, -0.00004168, -0.00002026, 0.00002523, 0.00004641, 0.00001444, 0.00000097, 0.00002348, 0.00000206, 0.00001185, 0.00001754, 0.00000673, 0.00001765, 0.00002020, 0.00001862, 0.00000186, 0.00000185, -0.00002623, -0.00002471, 0.00001668, -0.00000054, -0.00000116, -0.00001237, -0.00001483, 0.00001007, -0.00000309, -0.00000475, 0.00000152, -0.00000076, -0.00000006, 0.00000269, -0.00000300, -0.00000503, 0.00000147, -0.00000366, -0.00000378, 0.00000090, 0.00000124, -0.00000570, -0.00000792, 0.00000313, 0.00000926, -0.00000153, 0.00000356, 0.00000001, -0.00001691, 0.00001298, -0.00000044, -0.00001221, 0.00000827, -0.00001027, 0.00001273, 0.00000757, -0.00001637, 0.00000148, 0.00000269, -0.00000002, 0.00000269, 0.00000148, -0.00001637, 0.00000757, 0.00001273, -0.00001027, 0.00000827, -0.00001221, -0.00000044, 0.00001298, -0.00001691, 0.00000001, 0.00000356, -0.00000153, 0.00000926, 0.00000313, -0.00000792, -0.00000570, 0.00000124, 0.00000090, -0.00000378, -0.00000366, 0.00000147, -0.00000503, -0.00000300, 0.00000269, -0.00000006, -0.00000076, 0.00000152, -0.00000475, -0.00000309, 0.00001007, -0.00001483, -0.00001237, -0.00000116, -0.00000054, 0.00001668, -0.00002471, -0.00002623, 0.00000185, 0.00000186, 0.00001862, 0.00002020, 0.00001765, 0.00000673, 0.00001754, 0.00001185, 0.00000206, 0.00002348, 0.00000097, 0.00001444, 0.00004641, 0.00002523, -0.00002026, -0.00004168, -0.00001425, -0.00000868, -0.00001049, -0.00000710, -0.00000876, 0.00000166, -0.00000097, -0.00000624, },
		{ -0.00000579, -0.00000491, -0.00000566, -0.00000357, 0.00000227, -0.00000579, -0.00000748, -0.00001035, -0.00000916, -0.00001050, -0.00003859, -0.00002431, 0.00001639, 0.00004440, 0.00002310, -0.00000452, 0.00002125, 0.00000693, 0.00000591, 0.00001828, 0.00000782, 0.00001295, 0.00001680, 0.00002322, 0.00000273, 0.00000581, -0.00001347, -0.00003686, 0.00001165, 0.00000803, -0.00000449, -0.00000892, -0.00001682, 0.00000660, 0.00000136, -0.00000680, -0.00000089, 0.00000282, -0.00000108, 0.00000095, 0.00000111, -0.00000728, 0.00000058, -0.00000109, -0.00000501, 0.00000085, 0.00000107, -0.00000428, -0.00000703, -0.00000221, 0.00000905, 0.00000315, 0.00000076, 0.00000530, -0.00001870, 0.00000527, 0.00000913, -0.00001675, 0.00000922, -0.00000568, 0.00000089, 0.00001447, -0.00001132, -0.00000488, 0.00000391, 0.00000158, -0.00000087, 0.00000637, -0.00001472, -0.00000463, 0.00002142, -0.00000748, 0.00000240, -0.00000574, -0.00000825, 0.00001501, -0.00001068, -0.00000560, 0.00000530, -0.00000355, 0.00000583, 0.00000822, -0.00000592, -0.00000855, 0.00000147, 0.00000111, -0.00000285, -0.00000474, 0.00000001, -0.00000178, -0.00000532, 0.00000115, 0.00000309, -0.00000326, 0.00000034, -0.00000028, -0.00000695, 0.00001057, -0.00000937, -0.00001788, 0.00000208, -0.00000464, 0.00001364, -0.00001026, -0.00003136, -0.00000727, 0.00000055, 0.00001482, 0.00002028, 0.00002320, 0.00000847, 0.00001411, 0.00001829, -0.00000007, 0.00002164, 0.00000849, 0.00000792, 0.00004470, 0.00003469, -0.00001446, -0.00004456, -0.00001837, -0.00000845, -0.00001122, -0.00000595, -0.00001075, -0.00000088, 0.00000169, },
		{ 0.00000317, -0.00000393, -0.00000578, -0.00000494, -0.00000516, 0.00000114, -0.00000297, -0.00000664, -0.00001032, -0.00001030, -0.00000709, -0.00003513, -0.00002743, 0.00000919, 0.00003909, 0.00003177, -0.00000612, 0.00001538, 0.00001231, 0.00000234, 0.00001636, 0.00000961, 0.00001110, 0.00001069, 0.00002619, 0.00000653, 0.00000342, 0.00000187, -0.00004074, -0.00000147, 0.00001677, -0.00000413, -0.00000881, -0.00001633, 0.00000255, 0.00000392, -0.00000557, -0.00000501, 0.00000438, 0.00000153, -0.00000295, 0.00000439, -0.00000655, -0.00000276, 0.00000181, -0.00000534, -0.00000006, 0.00000185, -0.00000436, -0.00000511, -0.00000510, 0.00000471, 0.00000855, -0.00000059, 0.00000768, -0.00001505, -0.00000561, 0.00001621, -0.00001604, 0.00000439, 0.00000307, -0.00000829, 0.00001344, -0.00000236, -0.00000865, 0.00000125, 0.00000509, -0.00000360, 0.00000680, -0.00000719, -0.00001673, 0.00002229, 0.00000213, -0.00000450, -0.00000136, -0.00001169, 0.00001187, -0.00000312, -0.00000881, 0.00000474, -0.00000304, 0.00000103, 0.00001049, -0.00000090, -0.00001125, 0.00000039, 0.00000221, -0.00000269, -0.00000445, -0.00000239, 0.00000026, -0.00000469, -0.00000251, 0.00000556, -0.00000252, -0.00000390, 0.00000393, -0.00000771, 0.00000733, -0.00000186, -0.00002238, 0.00000190, -0.00000321, 0.00000635, 0.00000045, -0.00002789, -0.00001716, -0.00000418, 0.00001212, 0.00001831, 0.00002717, 0.00001379, 0.00000961, 0.00002287, 0.00000167, 0.00001692, 0.00001562, 0.00000488, 0.00003995, 0.00004335, -0.00000634, -0.00004692, -0.00002321, -0.00000813, -0.00001252, -0.00000493, -0.00001103, -0.00000468, },
		{ -0.00000843, 0.00000261, -0.00000113, -0.00000622, -0.00000484, -0.00000554, -0.00000084, -0.00000131, -0.00000483, -0.00000964, -0.00001216, -0.00000443, -0.00003091, -0.00003023, 0.00000387, 0.00003176, 0.00003833, -0.00000328, 0.00000764, 0.00001602, 0.00000161, 0.00001333, 0.00000996, 0.00001234, 0.00000448, 0.00002464, 0.00001452, -0.00000248, 0.00001329, -0.00003378, -0.00001821, 0.00002029, 0.00000151, -0.00001043, -0.00001601, 0.00000016, 0.00000386, -0.00000256, -0.00000779, 0.00000196, 0.00000683, -0.00000612, 0.00000436, -0.00000253, -0.00000689, 0.00000314, -0.00000377, -0.00000205, 0.00000322, -0.00000462, -0.00000439, -0.00000434, -0.00000186, 0.00001150, 0.00000154, 0.00000650, -0.00000776, -0.00001535, 0.00001723, -0.00000955, -0.00000370, 0.00001100, -0.00001081, 0.00000585, 0.00000562, -0.00000716, -0.00000407, 0.00000750, -0.00000288, 0.00000263, 0.00000231, -0.00002336, 0.00001431, 0.00001433, -0.00000770, -0.00000132, -0.00001111, 0.00000640, 0.00000245, -0.00000849, 0.00000260, -0.00000164, -0.00000277, 0.00000903, 0.00000535, -0.00001169, -0.00000270, 0.00000398, -0.00000276, -0.00000380, -0.00000411, -0.00000001, -0.00000193, -0.00000585, 0.00000515, 0.00000139, -0.00000862, 0.00000502, -0.00000457, 0.00000192, 0.00000480, -0.00002315, -0.00000260, 0.00000141, -0.00000014, 0.00000428, -0.00001903, -0.00002285, -0.00001268, 0.00000868, 0.00001607, 0.00002816, 0.00002166, 0.00000653, 0.00002403, 0.00000674, 0.00001155, 0.00002040, 0.00000540, 0.00003361, 0.00004990, 0.00000410, -0.00004804, -0.00002913, -0.00000774, -0.00001386, -0.00000489, -0.00000965, },
		{ -0.00000743, -0.00001090, 0.00000003, 0.00000146, -0.00000538, -0.00000550, -0.00000521, -0.00000261, -0.00000112, -0.00000287, -0.00000775, -0.00001428, -0.00000317, -0.00002569, -0.00003286, -0.00000003, 0.00002401, 0.00004145, 0.00000319, 0.00000032, 0.00001691, 0.00000269, 0.00001111, 0.00000813, 0.00001480, 0.00000145, 0.00001735, 0.00002463, -0.00000674, 0.00001627, -0.00001852, -0.00003168, 0.00001531, 0.00001039, -0.00001023, -0.00001810, -0.00000025, 0.00000270, -0.00000051, -0.00000705, -0.00000389, 0.00001171, -0.00000558, 0.00000041, 0.00000290, -0.00000924, 0.00000154, -0.00000053, -0.00000413, 0.00000387, -0.00000348, -0.00000595, -0.00000114, -0.00000721, 0.00000947, 0.00000708, 0.00000354, -0.00000044, -0.00002005, 0.00001139, 0.00000001, -0.00001072, 0.00001411, -0.00000635, -0.00000379, 0.00000833, -0.00000083, -0.00000864, 0.00000614, 0.00000154, -0.00000343, 0.00000905, -0.00002192, 0.00000066, 0.00002328, -0.00000441, -0.00000472, -0.00000919, 0.00000181, 0.00000443, -0.00000578, 0.00000083, -0.00000098, -0.00000446, 0.00000492, 0.00001018, -0.00000866, -0.00000708, 0.00000520, -0.00000209, -0.00000374, -0.00000440, -0.00000210, 0.00000095, -0.00000675, 0.00000160, 0.00000635, -0.00001083, 0.00000184, 0.00000089, -0.00000278, 0.00000808, -0.00001949, -0.00000959, 0.00000545, -0.00000243, 0.00000217, -0.00001008, -0.00002173, -0.00002222, 0.00000214, 0.00001422, 0.00002639, 0.00002988, 0.00000691, 0.00002183, 0.00001329, 0.00000776, 0.00002204, 0.00000845, 0.00002729, 0.00005359, 0.00001622, -0.00004702, -0.00003621, -0.00000772, -0.00001459, -0.00000614, },
		{ -0.00000835, -0.00000547, -0.00001147, -0.00000360, 0.00000254, -0.00000311, -0.00000634, -0.00000500, -0.00000353, -0.00000197, -0.00000172, -0.00000469, -0.00001573, -0.00000379, -0.00001970, -0.00003493, -0.00000332, 0.00001715, 0.00004086, 0.00001155, -0.00000482, 0.00001531, 0.00000382, 0.00001065, 0.00000524, 0.00001562, 0.00000355, 0.00000609, 0.00003242, -0.00000493, 0.00001081, -0.00000141, -0.00003608, 0.00000301, 0.00001757, -0.00000513, -0.00002244, -0.00000075, 0.00000302, -0.00000133, -0.00000304, -0.00001019, 0.00001271, -0.00000031, -0.00000552, 0.00000662, -0.00000788, -0.00000263, 0.00000271, -0.00000465, 0.00000263, -0.00000046, -0.00000887, 0.00000164, -0.00000823, 0.00000237, 0.00001339, 0.00000206, 0.00000366, -0.00001841, 0.00000128, 0.00000811, -0.00001310, 0.00001144, 0.00000197, -0.00001040, 0.00000439, 0.00000690, -0.00000893, 0.00000059, 0.00000727, -0.00000753, 0.00001005, -0.00001380, -0.00001292, 0.00002462, 0.00000453, -0.00000812, -0.00000890, -0.00000031, 0.00000364, -0.00000312, 0.00000103, -0.00000141, -0.00000463, 0.00000038, 0.00001161, -0.00000272, -0.00001083, 0.00000457, -0.00000019, -0.00000441, -0.00000365, -0.00000453, 0.00000206, -0.00000470, -0.00000338, 0.00000951, -0.00000893, -0.00000438, 0.00000565, -0.00000428, 0.00000740, -0.00001309, -0.00001590, 0.00000603, -0.00000029, -0.00000215, -0.00000536, -0.00001490, -0.00002848, -0.00000831, 0.00001169, 0.00002304, 0.00003632, 0.00001141, 0.00001778, 0.00001911, 0.00000679, 0.00002108, 0.00001249, 0.00002218, 0.00005439, 0.00002905, -0.00004318, -0.00004404, -0.00000875, -0.00001436, },
		{ -0.00001328, -0.00001080, -0.00000459, -0.00001048, -0.00000684, 0.00000143, -0.00000014, -0.00000638, -0.00000546, -0.00000368, -0.00000299, -0.00000197, -0.00000120, -0.00001554, -0.00000624, -0.00001368, -0.00003582, -0.00000670, 0.00001186, 0.00003718, 0.00002000, -0.00000708, 0.00001249, 0.00000378, 0.00001132, 0.00000356, 0.00001271, 0.00001002, -0.00000492, 0.00003358, 0.00000390, 0.00000136, 0.00001051, -0.00003001, -0.00001118, 0.00001807, 0.00000493, -0.00002601, -0.00000434, 0.00000637, -0.00000457, 0.00000159, -0.00001333, 0.00000822, 0.00000782, -0.00000982, 0.00000611, -0.00000299, -0.00000716, 0.00000367, -0.00000255, -0.00000040, 0.00000323, -0.00001078, 0.00000119, -0.00000418, -0.00000693, 0.00001647, 0.00000463, 0.00000360, -0.00001229, -0.00000855, 0.00001100, -0.00001008, 0.00000528, 0.00000961, -0.00001095, -0.00000403, 0.00001149, -0.00000352, -0.00000671, 0.00001070, -0.00000707, 0.00000565, -0.00000340, -0.00002109, 0.00001770, 0.00001501, -0.00000775, -0.00001142, -0.00000081, 0.00000243, -0.00000259, 0.00000318, -0.00000181, -0.00000475, -0.00000274, 0.00000941, 0.00000397, -0.00001193, 0.00000168, 0.00000241, -0.00000511, -0.00000290, -0.00000588, 0.00000078, -0.00000095, -0.00000728, 0.00000909, -0.00000359, -0.00001070, 0.00000697, -0.00000204, 0.00000436, -0.00000677, -0.00001903, 0.00000274, 0.00000384, -0.00000470, -0.00000605, -0.00000645, -0.00002814, -0.00002080, 0.00000642, 0.00001904, 0.00003983, 0.00001923, 0.00001411, 0.00002266, 0.00000861, 0.00001884, 0.00001623, 0.00001873, 0.00005272, 0.00004162, -0.00003618, -0.00005193, -0.00001135, },
		{ -0.00001574, -0.00001182, -0.00001283, -0.00000490, -0.00000895, -0.00000851, -0.00000155, 0.00000218, -0.00000495, -0.00000644, -0.00000371, -0.00000337, -0.00000343, 0.00000156, -0.00001316, -0.00000982, -0.00000862, -0.00003506, -0.00001049, 0.00000819, 0.00003147, 0.00002721, -0.00000678, 0.00000975, 0.00000263, 0.00001126, 0.00000501, 0.00000620, 0.00001760, -0.00001112, 0.00002659, 0.00001658, -0.00000577, 0.00001342, -0.00001732, -0.00002040, 0.00001015, 0.00001616, -0.00002453, -0.00001259, 0.00001167, -0.00000744, 0.00000346, -0.00001135, -0.00000031, 0.00001485, -0.00000936, 0.00000127, 0.00000287, -0.00000907, 0.00000105, 0.00000156, -0.00000348, 0.00000536, -0.00000950, -0.00000327, 0.00000271, -0.00001388, 0.00001329, 0.00001124, 0.00000130, -0.00000554, -0.00001401, 0.00000767, -0.00000400, -0.00000057, 0.00001320, -0.00000576, -0.00001246, 0.00000979, 0.00000577, -0.00001173, 0.00000927, -0.00000228, -0.00000094, 0.00000439, -0.00002153, 0.00000581, 0.00002191, -0.00000194, -0.00001517, -0.00000209, 0.00000289, -0.00000446, 0.00000554, -0.00000049, -0.00000578, -0.00000402, 0.00000508, 0.00000883, -0.00000950, -0.00000252, 0.00000448, -0.00000493, -0.00000295, -0.00000569, -0.00000212, 0.00000239, -0.00000828, 0.00000529, 0.00000271, -0.00001420, 0.00000400, 0.00000229, 0.00000159, -0.00000271, -0.00001851, -0.00000260, 0.00000683, -0.00000354, -0.00001005, -0.00000097, -0.00002096, -0.00003129, -0.00000289, 0.00001408, 0.00004027, 0.00002874, 0.00001269, 0.00002360, 0.00001224, 0.00001667, 0.00001901, 0.00001688, 0.00004913, 0.00005310, -0.00002600, -0.00005913, },
		{ -0.00006494, -0.00002189, -0.00001051, -0.00001421, -0.00000588, -0.00000801, -0.00000833, -0.00000510, 0.00000257, -0.00000217, -0.00000712, -0.00000431, -0.00000292, -0.00000525, 0.00000253, -0.00000885, -0.00001328, -0.00000536, -0.00003258, -0.00001462, 0.00000581, 0.00002476, 0.00003250, -0.00000465, 0.00000765, 0.00000162, 0.00000864, 0.00000975, -0.00000144, 0.00002219, -0.00001007, 0.00001373, 0.00002751, -0.00000602, 0.00000859, -0.00000499, -0.00002027, -0.00000301, 0.00002259, -0.00001543, -0.00002380, 0.00001537, -0.00000664, 0.00000082, -0.00000530, -0.00000896, 0.00001698, -0.00000353, -0.00000524, 0.00000613, -0.00000658, -0.00000426, 0.00000527, -0.00000423, 0.00000421, -0.00000468, -0.00000985, 0.00000826, -0.00001469, 0.00000380, 0.00001890, 0.00000035, -0.00000169, -0.00001359, 0.00000039, 0.00000121, -0.00000309, 0.00001216, 0.00000203, -0.00001665, 0.00000185, 0.00001464, -0.00001110, 0.00000319, 0.00000394, -0.00000571, 0.00000678, -0.00001582, -0.00000571, 0.00002190, 0.00000768, -0.00001683, -0.00000613, 0.00000521, -0.00000699, 0.00000579, 0.00000331, -0.00000734, -0.00000442, 0.00000079, 0.00001024, -0.00000439, -0.00000613, 0.00000485, -0.00000352, -0.00000380, -0.00000449, -0.00000510, 0.00000365, -0.00000622, 0.00000005, 0.00000721, -0.00001353, -0.00000187, 0.00000596, 0.00000108, -0.00000131, -0.00001580, -0.00000741, 0.00000681, 0.00000054, -0.00001370, -0.00000102, -0.00001002, -0.00003576, -0.00001531, 0.00000682, 0.00003777, 0.00003827, 0.00001457, 0.00002258, 0.00001638, 0.00001546, 0.00002076, 0.00001642, 0.00004409, 0.00006278, -0.00001272, },
		{ 0.00000350, -0.00006865, -0.00002970, -0.00000962, -0.00001520, -0.00000679, -0.00000818, -0.00000703, -0.00000764, 0.00000061, 0.00000092, -0.00000656, -0.00000569, -0.00000216, -0.00000636, 0.00000133, -0.00000367, -0.00001533, -0.00000424, -0.00002878, -0.00001874, 0.00000429, 0.00001788, 0.00003564, -0.00000128, 0.00000580, 0.00000234, 0.00000295, 0.00001579, -0.00000671, 0.00002110, -0.00000271, -0.00000002, 0.00003187, 0.00000092, 0.00000142, 0.00000084, -0.00001163, -0.00001459, 0.00001960, -0.00000025, -0.00003322, 0.00001353, -0.00000075, -0.00000509, 0.00000125, -0.00001369, 0.00001292, 0.00000504, -0.00000930, 0.00000444, -0.00000054, -0.00000923, 0.00000568, -0.00000121, 0.00000003, 0.00000165, -0.00001491, 0.00000872, -0.00000850, -0.00000858, 0.00002310, 0.00000364, -0.00000187, -0.00000897, -0.00000674, 0.00000236, -0.00000180, 0.00000858, 0.00000855, -0.00001475, -0.00000911, 0.00001858, -0.00000419, -0.00000446, 0.00000783, -0.00000625, 0.00000431, -0.00000815, -0.00001239, 0.00001523, 0.00001690, -0.00001350, -0.00001278, 0.00000753, -0.00000768, 0.00000262, 0.00000863, -0.00000784, -0.00000533, -0.00000198, 0.00000835, 0.00000124, -0.00000742, 0.00000331, -0.00000141, -0.00000480, -0.00000328, -0.00000686, 0.00000244, -0.00000251, -0.00000419, 0.00000834, -0.00000951, -0.00000794, 0.00000678, 0.00000301, -0.00000123, -0.00001300, -0.00001005, 0.00000410, 0.00000500, -0.00001411, -0.00000593, -0.00000008, -0.00003241, -0.00002756, -0.00000384, 0.00003194, 0.00004646, 0.00001995, 0.00002081, 0.00002001, 0.00001551, 0.00002168, 0.00001721, 0.00003806, 0.00006987, },
		{ 0.00007347, 0.00002221, -0.00006933, -0.00003923, -0.00000917, -0.00001620, -0.00000725, -0.00000913, -0.00000592, -0.00000814, -0.00000299, 0.00000283, -0.00000433, -0.00000724, -0.00000202, -0.00000604, -0.00000155, 0.00000088, -0.00001512, -0.00000493, -0.00002431, -0.00002249, 0.00000324, 0.00001145, 0.00003653, 0.00000319, 0.00000334, 0.00000558, -0.00000444, 0.00001999, -0.00000700, 0.00001450, 0.00000716, -0.00000962, 0.00002821, 0.00001116, -0.00000199, -0.00000159, -0.00000033, -0.00001779, 0.00000712, 0.00001534, -0.00003546, 0.00000462, 0.00000832, -0.00001057, 0.00000443, -0.00001272, 0.00000482, 0.00001186, -0.00000790, -0.00000159, 0.00000562, -0.00001031, 0.00000134, 0.00000469, -0.00000474, 0.00000610, -0.00001519, 0.00000300, 0.00000183, -0.00001847, 0.00002027, 0.00001137, -0.00000428, -0.00000375, -0.00001008, -0.00000117, 0.00000120, 0.00000554, 0.00001138, -0.00000802, -0.00001842, 0.00001520, 0.00000631, -0.00000945, 0.00000713, -0.00000302, 0.00000006, -0.00000270, -0.00001283, 0.00000550, 0.00002133, -0.00000478, -0.00001946, 0.00000714, -0.00000504, -0.00000317, 0.00001307, -0.00000566, -0.00000734, -0.00000316, 0.00000471, 0.00000515, -0.00000587, 0.00000084, 0.00000037, -0.00000520, -0.00000277, -0.00000704, -0.00000037, 0.00000084, -0.00000588, 0.00000643, -0.00000437, -0.00001177, 0.00000430, 0.00000590, -0.00000058, -0.00001140, -0.00001060, 0.00000052, 0.00000744, -0.00001073, -0.00001259, 0.00000493, -0.00002280, -0.00003545, -0.00001733, 0.00002176, 0.00005194, 0.00002855, 0.00001959, 0.00002252, 0.00001678, 0.00002197, 0.00001912, 0.00003178, },
		{ 0.00002631, 0.00007288, 0.00004235, -0.00006572, -0.00005051, -0.00000924, -0.00001727, -0.00000755, -0.00000995, -0.00000613, -0.00000670, -0.00000665, 0.00000247, -0.00000102, -0.00000782, -0.00000318, -0.00000444, -0.00000482, 0.00000344, -0.00001260, -0.00000657, -0.00001988, -0.00002561, 0.00000229, 0.00000604, 0.00003503, 0.00000884, -0.00000027, 0.00001073, -0.00001083, 0.00001964, -0.00000204, 0.00000518, 0.00001508, -0.00001253, 0.00001904, 0.00001912, 0.00000153, -0.00000873, 0.00000606, -0.00000995, -0.00000944, 0.00002410, -0.00002792, -0.00000888, 0.00001612, -0.00001182, 0.00000264, -0.00000770, -0.00000295, 0.00001326, -0.00000122, -0.00000830, 0.00000769, -0.00000567, -0.00000642, 0.00001030, -0.00000667, 0.00000596, -0.00000991, -0.00000655, 0.00001113, -0.00002124, 0.00001006, 0.00002065, -0.00000544, -0.00000125, -0.00000842, -0.00000726, 0.00000273, 0.00000509, 0.00001066, 0.00000020, -0.00002241, 0.00000555, 0.00001570, -0.00000874, 0.00000244, 0.00000113, -0.00000256, -0.00000143, -0.00000905, -0.00000251, 0.00001892, 0.00000660, -0.00002249, 0.00000245, 0.00000026, -0.00000905, 0.00001417, -0.00000045, -0.00000965, -0.00000383, 0.00000124, 0.00000618, -0.00000243, -0.00000102, 0.00000101, -0.00000478, -0.00000299, -0.00000616, -0.00000335, 0.00000238, -0.00000500, 0.00000319, -0.00000038, -0.00001246, -0.00000014, 0.00000766, 0.00000166, -0.00001077, -0.00001045, -0.00000208, 0.00000709, -0.00000529, -0.00001743, 0.00000384, -0.00001109, -0.00003616, -0.00003088, 0.00000655, 0.00005280, 0.00003960, 0.00002025, 0.00002364, 0.00001908, 0.00002182, 0.00002184, },
		{ 0.00002468, 0.00002278, 0.00006808, 0.00006206, -0.00005650, -0.00006318, -0.00001037, -0.00001803, -0.00000840, -0.00000977, -0.00000788, -0.00000443, -0.00000880, -0.00000018, 0.00000193, -0.00000643, -0.00000550, -0.00000255, -0.00000705, 0.00000342, -0.00000861, -0.00000809, -0.00001593, -0.00002798, 0.00000109, 0.00000213, 0.00003109, 0.00001561, -0.00000475, 0.00001593, -0.00001349, 0.00001393, 0.00000596, -0.00000304, 0.00001820, -0.00000966, 0.00000907, 0.00002094, 0.00001036, -0.00001386, 0.00000263, 0.00000541, -0.00002162, 0.00002139, -0.00001296, -0.00002131, 0.00001822, -0.00000755, -0.00000236, -0.00000263, -0.00000643, 0.00000873, 0.00000709, -0.00001092, 0.00000334, 0.00000321, -0.00001365, 0.00001173, -0.00000346, 0.00000101, -0.00000141, -0.00001528, 0.00001456, -0.00001550, -0.00000416, 0.00002686, -0.00000243, -0.00000256, -0.00000337, -0.00001234, 0.00000058, 0.00000701, 0.00000852, 0.00000652, -0.00002009, -0.00000654, 0.00001966, -0.00000230, -0.00000308, 0.00000301, -0.00000193, -0.00000326, -0.00000492, -0.00000549, 0.00001113, 0.00001606, -0.00001937, -0.00000566, 0.00000557, -0.00001218, 0.00001092, 0.00000637, -0.00001055, -0.00000514, -0.00000101, 0.00000471, 0.00000101, -0.00000109, 0.00000058, -0.00000399, -0.00000345, -0.00000505, -0.00000540, 0.00000188, -0.00000282, 0.00000056, 0.00000139, -0.00001082, -0.00000439, 0.00000700, 0.00000500, -0.00000978, -0.00001095, -0.00000303, 0.00000497, -0.00000033, -0.00001845, -0.00000155, -0.00000187, -0.00002980, -0.00004043, -0.00001275, 0.00004674, 0.00005151, 0.00002421, 0.00002344, 0.00002204, 0.00002163, },
		{ 0.00002206, 0.00002686, 0.00002185, 0.00006007, 0.00007899, -0.00004095, -0.00007604, -0.00001356, -0.00001779, -0.00001044, -0.00000842, -0.00001045, -0.00000284, -0.00000872, -0.00000392, 0.00000290, -0.00000301, -0.00000786, -0.00000170, -0.00000735, 0.00000124, -0.00000448, -0.00000870, -0.00001257, -0.00002954, -0.00000073, 0.00000008, 0.00002491, 0.00002296, -0.00000895, 0.00001892, -0.00001094, 0.00000434, 0.00001353, -0.00000735, 0.00001650, -0.00000434, 0.00000240, 0.00001663, 0.00001923, -0.00001129, -0.00000974, 0.00002010, -0.00002239, 0.00000822, 0.00000275, -0.00002694, 0.00001327, -0.00000011, -0.00000658, -0.00000114, -0.00000470, 0.00000151, 0.00001184, -0.00000651, -0.00000602, 0.00001179, -0.00001578, 0.00000676, 0.00000433, -0.00000599, 0.00000599, -0.00001873, 0.00001020, -0.00000393, -0.00001681, 0.00002625, 0.00000517, -0.00000606, 0.00000184, -0.00001354, -0.00000517, 0.00000916, 0.00000747, 0.00000932, -0.00001337, -0.00001645, 0.00001650, 0.00000680, -0.00000564, 0.00000117, 0.00000089, -0.00000524, -0.00000369, -0.00000334, 0.00000213, 0.00001964, -0.00001045, -0.00001398, 0.00000785, -0.00001117, 0.00000453, 0.00001218, -0.00000862, -0.00000734, -0.00000220, 0.00000216, 0.00000280, 0.00000068, 0.00000003, -0.00000353, -0.00000371, -0.00000426, -0.00000623, 0.00000010, -0.00000092, -0.00000043, 0.00000137, -0.00000859, -0.00000689, 0.00000419, 0.00000787, -0.00000722, -0.00001242, -0.00000312, 0.00000289, 0.00000240, -0.00001600, -0.00000784, 0.00000214, -0.00001949, -0.00004267, -0.00003283, 0.00003200, 0.00006130, 0.00003262, 0.00002266, 0.00002488, },
		{ 0.00002658, 0.00002372, 0.00002801, 0.00002328, 0.00005070, 0.00009102, -0.00001947, -0.00008702, -0.00001999, -0.00001619, -0.00001364, -0.00000653, -0.00001264, -0.00000290, -0.00000693, -0.00000694, 0.00000111, 0.00000126, -0.00000864, -0.00000276, -0.00000591, -0.00000180, -0.00000154, -0.00000814, -0.00000955, -0.00003022, -0.00000352, -0.00000002, 0.00001707, 0.00002983, -0.00001131, 0.00001804, -0.00000362, -0.00000609, 0.00001769, -0.00000720, 0.00001221, -0.00000000, 0.00000024, 0.00000981, 0.00002257, 0.00000016, -0.00002461, 0.00002554, -0.00001032, -0.00000862, 0.00001169, -0.00002356, 0.00000420, 0.00000588, -0.00000656, -0.00000396, -0.00000054, -0.00000327, 0.00000949, 0.00000380, -0.00001545, 0.00001474, -0.00001026, -0.00000347, 0.00001317, -0.00001072, 0.00000835, -0.00001511, 0.00000005, 0.00000805, -0.00002283, 0.00001802, 0.00001486, -0.00000864, 0.00000432, -0.00001016, -0.00001223, 0.00000893, 0.00000869, 0.00000930, -0.00000558, -0.00002107, 0.00000784, 0.00001410, -0.00000311, -0.00000310, 0.00000312, -0.00000473, -0.00000604, 0.00000110, -0.00000357, 0.00001616, 0.00000091, -0.00001866, 0.00000558, -0.00000694, -0.00000228, 0.00001467, -0.00000384, -0.00000943, -0.00000329, -0.00000006, 0.00000239, 0.00000317, 0.00000049, -0.00000372, -0.00000376, -0.00000380, -0.00000622, -0.00000184, -0.00000027, 0.00000000, 0.00000092, -0.00000721, -0.00000746, 0.00000070, 0.00000872, -0.00000296, -0.00001390, -0.00000379, 0.00000213, 0.00000291, -0.00001212, -0.00001212, 0.00000112, -0.00000968, -0.00003696, -0.00004876, 0.00000897, 0.00006475, 0.00004556, 0.00002302, },
		{ 0.00002708, 0.00002644, 0.00002653, 0.00002854, 0.00002606, 0.00004196, 0.00009708, 0.00000615, -0.00009368, -0.00003026, -0.00001359, -0.00001729, -0.00000513, -0.00001356, -0.00000452, -0.00000487, -0.00000786, -0.00000281, 0.00000433, -0.00000664, -0.00000546, -0.00000389, -0.00000419, -0.00000044, -0.00000687, -0.00000656, -0.00002980, -0.00000750, 0.00000154, 0.00000856, 0.00003490, -0.00001049, 0.00001288, 0.00000644, -0.00001417, 0.00001720, -0.00000417, 0.00000806, 0.00000184, 0.00000071, 0.00000518, 0.00001795, 0.00001620, -0.00003347, 0.00001762, 0.00000918, -0.00002001, 0.00000985, -0.00001424, -0.00000328, 0.00000659, -0.00000204, -0.00000816, 0.00000131, -0.00000188, 0.00000065, 0.00001490, -0.00001893, 0.00000905, 0.00000148, -0.00001452, 0.00001823, -0.00000969, 0.00000441, -0.00000626, -0.00001103, 0.00001497, -0.00002017, 0.00000494, 0.00002254, -0.00000767, 0.00000316, -0.00000386, -0.00001749, 0.00000498, 0.00001135, 0.00000856, 0.00000043, -0.00002009, -0.00000242, 0.00001611, 0.00000366, -0.00000653, 0.00000225, -0.00000125, -0.00000999, 0.00000394, -0.00000356, 0.00000795, 0.00001001, -0.00001741, -0.00000022, -0.00000216, -0.00000680, 0.00001313, 0.00000219, -0.00000990, -0.00000501, -0.00000138, 0.00000044, 0.00000479, 0.00000252, -0.00000412, -0.00000399, -0.00000344, -0.00000592, -0.00000319, -0.00000088, 0.00000076, 0.00000125, -0.00000698, -0.00000714, -0.00000178, 0.00000709, 0.00000181, -0.00001390, -0.00000589, 0.00000258, 0.00000254, -0.00000892, -0.00001339, -0.00000248, -0.00000385, -0.00002591, -0.00005596, -0.00001881, 0.00005759, 0.00006087, },
		{ 0.00007375, 0.00003720, 0.00002509, 0.00002944, 0.00002946, 0.00002909, 0.00003512, 0.00009748, 0.00003329, -0.00009410, -0.00004393, -0.00001113, -0.00002039, -0.00000498, -0.00001318, -0.00000675, -0.00000393, -0.00000659, -0.00000692, 0.00000431, -0.00000201, -0.00000833, -0.00000284, -0.00000502, -0.00000091, -0.00000574, -0.00000349, -0.00002795, -0.00001265, 0.00000414, 0.00000060, 0.00003696, -0.00000585, 0.00000435, 0.00001643, -0.00001774, 0.00001282, -0.00000063, 0.00000552, 0.00000210, 0.00000053, 0.00000554, 0.00000760, 0.00002953, -0.00003049, -0.00000089, 0.00002649, -0.00001948, -0.00000051, -0.00000588, -0.00000435, 0.00000200, 0.00000342, -0.00000907, -0.00000254, 0.00000540, -0.00000978, 0.00002019, -0.00001283, -0.00000383, 0.00001427, -0.00002076, 0.00001604, -0.00000232, -0.00000368, 0.00000334, -0.00001788, 0.00001405, -0.00001074, -0.00000802, 0.00002474, -0.00000264, -0.00000031, 0.00000253, -0.00001874, -0.00000198, 0.00001329, 0.00000901, 0.00000372, -0.00001561, -0.00001028, 0.00001214, 0.00001125, -0.00000598, -0.00000207, 0.00000330, -0.00001239, 0.00000261, 0.00000126, -0.00000049, 0.00001326, -0.00001089, -0.00000645, 0.00000040, -0.00000791, 0.00000886, 0.00000713, -0.00000781, -0.00000716, -0.00000225, -0.00000177, 0.00000450, 0.00000562, -0.00000378, -0.00000481, -0.00000303, -0.00000563, -0.00000386, -0.00000201, 0.00000076, 0.00000268, -0.00000709, -0.00000734, -0.00000249, 0.00000393, 0.00000533, -0.00001154, -0.00000909, 0.00000302, 0.00000284, -0.00000731, -0.00001266, -0.00000573, -0.00000290, -0.00001418, -0.00005255, -0.00004499, 0.00003769, },
		{ 0.00000698, 0.00007765, 0.00005368, 0.00002502, 0.00003081, 0.00003155, 0.00003194, 0.00003040, 0.00009351, 0.00005946, -0.00008751, -0.00005955, -0.00001017, -0.00002233, -0.00000607, -0.00001222, -0.00000848, -0.00000453, -0.00000444, -0.00000900, 0.00000063, 0.00000354, -0.00000938, -0.00000371, -0.00000441, -0.00000192, -0.00000548, -0.00000066, -0.00002428, -0.00001866, 0.00000684, -0.00000560, 0.00003517, 0.00000236, -0.00000565, 0.00002374, -0.00001623, 0.00000647, 0.00000175, 0.00000415, 0.00000300, -0.00000253, 0.00001007, -0.00000294, 0.00003389, -0.00001580, -0.00002162, 0.00003295, -0.00000713, -0.00001185, -0.00000502, 0.00000136, -0.00000347, 0.00000457, -0.00000357, -0.00001148, 0.00001390, -0.00001507, 0.00001539, 0.00000156, -0.00001814, 0.00002162, -0.00001838, 0.00000661, 0.00000846, -0.00001145, 0.00000898, -0.00001767, 0.00000635, 0.00000071, -0.00001612, 0.00002048, 0.00000458, -0.00000363, 0.00000667, -0.00001576, -0.00000958, 0.00001247, 0.00001108, 0.00000529, -0.00001053, -0.00001363, 0.00000447, 0.00001561, -0.00000044, -0.00000764, 0.00000585, -0.00001104, -0.00000264, 0.00000739, -0.00000491, 0.00001007, -0.00000246, -0.00000976, -0.00000040, -0.00000658, 0.00000428, 0.00000921, -0.00000363, -0.00000859, -0.00000346, -0.00000328, 0.00000231, 0.00000848, -0.00000187, -0.00000609, -0.00000281, -0.00000530, -0.00000417, -0.00000286, -0.00000039, 0.00000452, -0.00000633, -0.00000877, -0.00000192, 0.00000092, 0.00000632, -0.00000722, -0.00001199, 0.00000211, 0.00000442, -0.00000668, -0.00001175, -0.00000698, -0.00000508, -0.00000614, -0.00004059, -0.00006259, },
		{ -0.00006699, -0.00002804, 0.00006685, 0.00007317, 0.00003003, 0.00002972, 0.00003441, 0.00003521, 0.00002731, 0.00008652, 0.00008308, -0.00007438, -0.00007527, -0.00001150, -0.00002324, -0.00000784, -0.00001146, -0.00000938, -0.00000585, -0.00000334, -0.00000783, -0.00000538, 0.00000738, -0.00000724, -0.00000617, -0.00000350, -0.00000233, -0.00000621, 0.00000113, -0.00001859, -0.00002477, 0.00000854, -0.00000901, 0.00002933, 0.00001299, -0.00001484, 0.00002659, -0.00001043, 0.00000013, 0.00000270, 0.00000235, 0.00000613, -0.00000801, 0.00001501, -0.00000814, 0.00002754, 0.00000462, -0.00003521, 0.00002562, 0.00001023, -0.00001561, -0.00001339, 0.00000871, -0.00000343, -0.00000179, 0.00000706, -0.00002059, 0.00001709, -0.00001061, 0.00000138, 0.00001811, -0.00002652, 0.00001912, -0.00000762, -0.00000631, 0.00001770, -0.00001453, 0.00000819, -0.00001111, -0.00000404, 0.00000907, -0.00001713, 0.00001175, 0.00001089, -0.00000481, 0.00000788, -0.00001012, -0.00001522, 0.00000833, 0.00001350, 0.00000694, -0.00000682, -0.00001295, -0.00000312, 0.00001435, 0.00000814, -0.00001090, 0.00000431, -0.00000620, -0.00000915, 0.00001092, -0.00000352, 0.00000307, 0.00000371, -0.00000850, -0.00000342, -0.00000507, 0.00000142, 0.00000832, 0.00000092, -0.00000808, -0.00000530, -0.00000400, -0.00000082, 0.00000973, 0.00000167, -0.00000713, -0.00000317, -0.00000478, -0.00000448, -0.00000306, -0.00000229, 0.00000562, -0.00000399, -0.00001106, -0.00000144, -0.00000052, 0.00000479, -0.00000243, -0.00001311, -0.00000062, 0.00000659, -0.00000564, -0.00001191, -0.00000653, -0.00000749, -0.00000370, -0.00002530, },
		{ -0.00001265, -0.00005827, -0.00005811, 0.00003959, 0.00008860, 0.00004320, 0.00002740, 0.00003628, 0.00003987, 0.00002567, 0.00007713, 0.00010354, -0.00005563, -0.00008980, -0.00001509, -0.00002375, -0.00000978, -0.00001102, -0.00001012, -0.00000642, -0.00000459, -0.00000405, -0.00001097, 0.00000728, -0.00000218, -0.00000864, -0.00000364, -0.00000167, -0.00000721, 0.00000097, -0.00001126, -0.00002980, 0.00000814, -0.00000898, 0.00001995, 0.00002421, -0.00002096, 0.00002424, -0.00000187, -0.00000492, 0.00000309, -0.00000122, 0.00001108, -0.00001313, 0.00001616, -0.00000562, 0.00001402, 0.00002235, -0.00003627, 0.00000880, 0.00002288, -0.00000759, -0.00002600, 0.00000968, 0.00000611, -0.00001377, 0.00001705, -0.00002322, 0.00001056, 0.00000301, -0.00001568, 0.00002867, -0.00002387, 0.00000705, 0.00000703, -0.00001697, 0.00002099, -0.00001098, 0.00000183, -0.00000186, -0.00001224, 0.00001129, -0.00001214, 0.00000232, 0.00001376, -0.00000352, 0.00000709, -0.00000407, -0.00001743, 0.00000208, 0.00001432, 0.00000974, -0.00000455, -0.00001053, -0.00000747, 0.00000799, 0.00001585, -0.00000912, -0.00000096, -0.00000037, -0.00001356, 0.00000968, 0.00000225, -0.00000346, 0.00000486, -0.00000374, -0.00000597, -0.00000536, 0.00000073, 0.00000595, 0.00000383, -0.00000527, -0.00000716, -0.00000457, -0.00000368, 0.00000872, 0.00000599, -0.00000694, -0.00000433, -0.00000412, -0.00000488, -0.00000275, -0.00000409, 0.00000517, -0.00000032, -0.00001298, -0.00000231, -0.00000002, 0.00000195, 0.00000109, -0.00001179, -0.00000445, 0.00000802, -0.00000322, -0.00001299, -0.00000593, -0.00000801, -0.00000566, },
		{ -0.00000880, -0.00000644, -0.00004137, -0.00007480, 0.00000014, 0.00009123, 0.00006425, 0.00002774, 0.00003507, 0.00004593, 0.00002640, 0.00006523, 0.00012044, -0.00003173, -0.00010265, -0.00002045, -0.00002419, -0.00001215, -0.00001016, -0.00001184, -0.00000528, -0.00000786, 0.00000004, -0.00001338, 0.00000277, 0.00000381, -0.00000911, -0.00000545, -0.00000055, -0.00000724, -0.00000154, -0.00000343, -0.00003231, 0.00000483, -0.00000558, 0.00000840, 0.00003372, -0.00002226, 0.00001702, 0.00000761, -0.00000816, 0.00000391, -0.00000646, 0.00001555, -0.00001463, 0.00001150, 0.00000275, -0.00000014, 0.00003111, -0.00002601, -0.00000853, 0.00002386, 0.00000897, -0.00003341, -0.00000157, 0.00002309, -0.00002434, 0.00001912, -0.00001525, -0.00000483, 0.00001972, -0.00002725, 0.00002732, -0.00001025, -0.00000947, 0.00001894, -0.00002057, 0.00001677, -0.00000220, -0.00000654, 0.00000559, -0.00001503, 0.00000755, -0.00000464, -0.00000427, 0.00001261, -0.00000107, 0.00000589, 0.00000084, -0.00001641, -0.00000407, 0.00001226, 0.00001324, -0.00000237, -0.00000872, -0.00000763, -0.00000039, 0.00001892, -0.00000211, -0.00000719, 0.00000328, -0.00001389, 0.00000431, 0.00000885, -0.00000595, 0.00000111, 0.00000140, -0.00000560, -0.00000777, 0.00000093, 0.00000420, 0.00000406, -0.00000115, -0.00000778, -0.00000563, -0.00000565, 0.00000584, 0.00000966, -0.00000493, -0.00000592, -0.00000365, -0.00000523, -0.00000236, -0.00000507, 0.00000311, 0.00000360, -0.00001328, -0.00000495, 0.00000159, -0.00000056, 0.00000231, -0.00000860, -0.00000797, 0.00000767, 0.00000046, -0.00001381, -0.00000661, -0.00000647, },
		{ -0.00000441, -0.00001010, -0.00000671, -0.00002389, -0.00007426, -0.00004171, 0.00007459, 0.00008776, 0.00003582, 0.00003038, 0.00005150, 0.00003129, 0.00005104, 0.00013255, -0.00000231, -0.00011368, -0.00002761, -0.00002390, -0.00001598, -0.00000795, -0.00001509, -0.00000288, -0.00001126, 0.00000177, -0.00001137, -0.00000440, 0.00000785, -0.00000646, -0.00000812, -0.00000037, -0.00000539, -0.00000579, 0.00000308, -0.00003105, -0.00000146, 0.00000025, -0.00000323, 0.00003923, -0.00001790, 0.00000626, 0.00001620, -0.00000951, 0.00000552, -0.00001174, 0.00001674, -0.00001088, 0.00000227, 0.00001233, -0.00000925, 0.00002976, -0.00001094, -0.00001813, 0.00001326, 0.00002469, -0.00002727, -0.00002404, 0.00003912, -0.00002480, 0.00000908, 0.00000181, -0.00002267, 0.00003089, -0.00002679, 0.00001364, 0.00000879, -0.00002293, 0.00002272, -0.00001570, 0.00000706, 0.00000796, -0.00001276, 0.00000821, -0.00001215, 0.00000070, 0.00000148, -0.00000640, 0.00000894, 0.00000063, 0.00000521, 0.00000436, -0.00001359, -0.00000820, 0.00000763, 0.00001572, 0.00000117, -0.00000847, -0.00000508, -0.00000696, 0.00001576, 0.00000748, -0.00001076, 0.00000289, -0.00001065, -0.00000243, 0.00001275, -0.00000337, -0.00000476, 0.00000357, -0.00000174, -0.00001065, -0.00000001, 0.00000434, 0.00000220, 0.00000239, -0.00000623, -0.00000715, -0.00000689, 0.00000214, 0.00001153, -0.00000130, -0.00000716, -0.00000374, -0.00000540, -0.00000226, -0.00000498, 0.00000011, 0.00000649, -0.00001138, -0.00000871, 0.00000293, -0.00000154, 0.00000139, -0.00000490, -0.00000997, 0.00000548, 0.00000434, -0.00001317, -0.00000877, },
		{ -0.00001134, -0.00000370, -0.00000855, -0.00001019, -0.00001247, -0.00005915, -0.00007404, 0.00003833, 0.00010391, 0.00005482, 0.00002519, 0.00005315, 0.00004153, 0.00003636, 0.00013738, 0.00003311, -0.00012181, -0.00003789, -0.00002144, -0.00002222, -0.00000429, -0.00001900, -0.00000109, -0.00001242, -0.00000032, -0.00000601, -0.00001104, 0.00000780, -0.00000138, -0.00000974, -0.00000235, -0.00000185, -0.00001007, 0.00000634, -0.00002562, -0.00000978, 0.00000673, -0.00001254, 0.00003909, -0.00000835, -0.00000581, 0.00002219, -0.00000896, 0.00000738, -0.00001490, 0.00001308, -0.00000291, -0.00000790, 0.00001859, -0.00001125, 0.00002214, 0.00000137, -0.00001719, -0.00000167, 0.00002928, -0.00000596, -0.00004886, 0.00004326, -0.00001011, -0.00001074, 0.00002063, -0.00003412, 0.00002991, -0.00001345, -0.00000676, 0.00002485, -0.00002715, 0.00001691, -0.00000494, -0.00000372, 0.00001533, -0.00001411, 0.00000564, -0.00000592, -0.00000557, 0.00000387, -0.00000490, 0.00000529, 0.00000043, 0.00000478, 0.00000720, -0.00001046, -0.00000975, 0.00000218, 0.00001544, 0.00000632, -0.00000892, -0.00000242, -0.00000918, 0.00000782, 0.00001549, -0.00000928, -0.00000088, -0.00000627, -0.00000749, 0.00001233, 0.00000250, -0.00000901, 0.00000118, 0.00000385, -0.00001152, -0.00000329, 0.00000600, 0.00000010, 0.00000363, -0.00000265, -0.00000824, -0.00000803, -0.00000136, 0.00001121, 0.00000297, -0.00000721, -0.00000442, -0.00000541, -0.00000256, -0.00000408, -0.00000285, 0.00000752, -0.00000766, -0.00001231, 0.00000287, -0.00000081, -0.00000064, -0.00000201, -0.00001012, 0.00000227, 0.00000721, -0.00001074, },
		{ -0.00000728, -0.00001285, -0.00000510, -0.00000551, -0.00001259, -0.00000972, -0.00003748, -0.00008779, -0.00001005, 0.00010209, 0.00008269, 0.00002572, 0.00004797, 0.00005590, 0.00002509, 0.00013225, 0.00007353, -0.00012406, -0.00005367, -0.00001563, -0.00003052, -0.00000071, -0.00002149, -0.00000205, -0.00001009, -0.00000552, -0.00000021, -0.00001424, 0.00000358, 0.00000370, -0.00000839, -0.00000645, 0.00000180, -0.00001218, 0.00000525, -0.00001698, -0.00001813, 0.00001162, -0.00001753, 0.00003294, 0.00000450, -0.00001650, 0.00002433, -0.00000663, 0.00000864, -0.00001444, 0.00000510, 0.00000629, -0.00001541, 0.00001964, -0.00000808, 0.00001385, 0.00000693, -0.00000974, -0.00001085, 0.00001797, 0.00002301, -0.00006295, 0.00002829, 0.00001702, -0.00003152, 0.00003161, -0.00003262, 0.00001591, 0.00000732, -0.00002490, 0.00003079, -0.00002037, 0.00000457, 0.00000661, -0.00001116, 0.00001746, -0.00001046, -0.00000008, 0.00000032, -0.00000868, 0.00000255, -0.00000218, 0.00000363, -0.00000125, 0.00000342, 0.00001006, -0.00000760, -0.00000953, -0.00000196, 0.00001185, 0.00001175, -0.00000826, -0.00000171, -0.00000700, -0.00000136, 0.00001850, -0.00000305, -0.00000532, -0.00000350, -0.00000936, 0.00000847, 0.00000840, -0.00000907, -0.00000470, 0.00000803, -0.00000867, -0.00000833, 0.00000738, -0.00000034, 0.00000215, 0.00000157, -0.00000773, -0.00000945, -0.00000427, 0.00000919, 0.00000664, -0.00000571, -0.00000532, -0.00000548, -0.00000318, -0.00000285, -0.00000499, 0.00000660, -0.00000316, -0.00001459, 0.00000109, 0.00000094, -0.00000257, -0.00000056, -0.00000893, -0.00000088, 0.00000853, },
		{ 0.00000863, -0.00000394, -0.00001255, -0.00000769, -0.00000350, -0.00001130, -0.00001344, -0.00001880, -0.00008085, -0.00005724, 0.00007611, 0.00011070, 0.00003875, 0.00003652, 0.00006989, 0.00002203, 0.00011621, 0.00011518, -0.00011580, -0.00007685, -0.00000726, -0.00003848, 0.00000008, -0.00002041, -0.00000671, -0.00000508, -0.00001123, 0.00000300, -0.00001301, -0.00000259, 0.00000591, -0.00000359, -0.00001089, 0.00000322, -0.00001062, 0.00000034, -0.00000746, -0.00002410, 0.00001313, -0.00001740, 0.00002207, 0.00001769, -0.00002334, 0.00002222, -0.00000299, 0.00000867, -0.00001016, -0.00000500, 0.00001354, -0.00001856, 0.00001656, -0.00000345, 0.00000847, 0.00000725, -0.00000369, -0.00000765, -0.00000482, 0.00004593, -0.00005599, -0.00000413, 0.00004601, -0.00004199, 0.00002809, -0.00001780, -0.00000550, 0.00002638, -0.00003286, 0.00002426, -0.00000599, -0.00000843, 0.00001408, -0.00001293, 0.00001458, -0.00000385, -0.00000596, 0.00000400, -0.00000811, -0.00000059, -0.00000061, 0.00000428, -0.00000273, 0.00000013, 0.00001277, -0.00000455, -0.00000900, -0.00000354, 0.00000611, 0.00001528, -0.00000516, -0.00000324, -0.00000251, -0.00000801, 0.00001549, 0.00000514, -0.00000733, -0.00000361, -0.00000867, 0.00000365, 0.00001157, -0.00000488, -0.00001100, 0.00000811, -0.00000246, -0.00001298, 0.00000652, 0.00000162, -0.00000092, 0.00000449, -0.00000504, -0.00001080, -0.00000682, 0.00000633, 0.00000887, -0.00000303, -0.00000580, -0.00000573, -0.00000405, -0.00000174, -0.00000592, 0.00000428, 0.00000094, -0.00001497, -0.00000193, 0.00000275, -0.00000367, -0.00000037, -0.00000725, -0.00000337, },
		{ -0.00000528, 0.00000822, -0.00000140, -0.00001099, -0.00000962, -0.00000427, -0.00000677, -0.00001841, -0.00000972, -0.00005905, -0.00008924, 0.00002847, 0.00012559, 0.00006714, 0.00002448, 0.00007711, 0.00003017, 0.00009204, 0.00015152, -0.00009257, -0.00010678, 0.00000009, -0.00004229, -0.00000435, -0.00001518, -0.00001391, 0.00000003, -0.00001460, 0.00000223, -0.00000887, -0.00000744, 0.00000375, 0.00000289, -0.00001279, 0.00000071, -0.00000568, -0.00000610, -0.00000001, -0.00002587, 0.00001073, -0.00001299, 0.00000909, 0.00002819, -0.00002502, 0.00001664, 0.00000114, 0.00000730, -0.00000298, -0.00001462, 0.00001707, -0.00001799, 0.00001183, 0.00000024, 0.00000552, 0.00000757, -0.00000535, 0.00000661, -0.00002703, 0.00004989, -0.00002686, -0.00004262, 0.00006298, -0.00003452, 0.00001042, 0.00000401, -0.00002502, 0.00003551, -0.00002758, 0.00000862, 0.00000932, -0.00001652, 0.00001511, -0.00000966, 0.00000896, 0.00000292, -0.00000967, 0.00000435, -0.00000521, -0.00000324, -0.00000123, 0.00000594, -0.00000221, -0.00000492, 0.00001411, -0.00000045, -0.00000907, -0.00000285, 0.00000051, 0.00001523, 0.00000015, -0.00000557, 0.00000150, -0.00001006, 0.00000829, 0.00001153, -0.00000528, -0.00000566, -0.00000750, 0.00000015, 0.00001120, 0.00000138, -0.00001442, 0.00000348, 0.00000469, -0.00001469, 0.00000264, 0.00000502, -0.00000352, 0.00000474, -0.00000077, -0.00001108, -0.00000941, 0.00000328, 0.00000951, -0.00000003, -0.00000537, -0.00000599, -0.00000513, -0.00000108, -0.00000570, 0.00000136, 0.00000385, -0.00001361, -0.00000535, 0.00000393, -0.00000386, -0.00000089, -0.00000562, },
		{ -0.00000407, -0.00000702, 0.00000768, 0.00000059, -0.00000946, -0.00000947, -0.00000754, -0.00000194, -0.00001969, -0.00001108, -0.00003339, -0.00009751, -0.00002902, 0.00011505, 0.00010590, 0.00002158, 0.00007285, 0.00004807, 0.00006639, 0.00017543, -0.00005280, -0.00013888, 0.00000074, -0.00003863, -0.00001434, -0.00000758, -0.00002079, 0.00000247, -0.00001424, -0.00000146, -0.00000488, -0.00000859, -0.00000172, 0.00000759, -0.00000981, -0.00000542, 0.00000035, -0.00001093, 0.00000313, -0.00002320, 0.00000548, -0.00000637, -0.00000307, 0.00003397, -0.00002182, 0.00000927, 0.00000491, 0.00000451, 0.00000580, -0.00002186, 0.00001685, -0.00001540, 0.00000695, 0.00000343, 0.00000167, 0.00001250, -0.00001490, 0.00002285, -0.00003524, 0.00003030, 0.00001420, -0.00007074, 0.00005785, -0.00000983, -0.00001361, 0.00002293, -0.00003442, 0.00003136, -0.00001209, -0.00000899, 0.00001913, -0.00001711, 0.00001074, -0.00000406, 0.00000334, 0.00000783, -0.00001044, 0.00000226, -0.00000200, -0.00000400, -0.00000340, 0.00000661, 0.00000102, -0.00001018, 0.00001261, 0.00000495, -0.00000954, -0.00000137, -0.00000279, 0.00001147, 0.00000589, -0.00000668, 0.00000321, -0.00000789, 0.00000034, 0.00001346, 0.00000001, -0.00000728, -0.00000772, -0.00000128, 0.00000855, 0.00000672, -0.00001328, -0.00000401, 0.00000957, -0.00001217, -0.00000322, 0.00000781, -0.00000392, 0.00000238, 0.00000347, -0.00000937, -0.00001198, 0.00000014, 0.00000901, 0.00000247, -0.00000407, -0.00000588, -0.00000634, -0.00000110, -0.00000462, -0.00000140, 0.00000527, -0.00001110, -0.00000848, 0.00000428, -0.00000343, -0.00000171, },
		{ -0.00000280, -0.00000234, -0.00000872, 0.00000671, 0.00000286, -0.00000901, -0.00000736, -0.00001115, -0.00000008, -0.00001552, -0.00001826, -0.00001469, -0.00008288, -0.00007913, 0.00007429, 0.00014164, 0.00003706, 0.00005794, 0.00006919, 0.00004748, 0.00018248, 0.00000019, -0.00016528, -0.00001064, -0.00002699, -0.00002726, -0.00000112, -0.00002448, 0.00000099, -0.00001113, -0.00000538, -0.00000360, -0.00000622, -0.00000714, 0.00000722, -0.00000192, -0.00001245, 0.00000409, -0.00001168, 0.00000164, -0.00001774, -0.00000044, 0.00000013, -0.00001232, 0.00003473, -0.00001542, 0.00000185, 0.00000815, 0.00000012, 0.00001497, -0.00002572, 0.00001365, -0.00001173, 0.00000126, 0.00000851, -0.00000585, 0.00002200, -0.00002572, 0.00002925, -0.00002243, -0.00000556, 0.00004965, -0.00007499, 0.00003020, 0.00002252, -0.00003257, 0.00003067, -0.00003064, 0.00001677, 0.00000645, -0.00002132, 0.00002024, -0.00001146, 0.00000423, 0.00000095, -0.00000052, 0.00001034, -0.00000889, -0.00000069, 0.00000004, -0.00000294, -0.00000545, 0.00000494, 0.00000603, -0.00001337, 0.00000759, 0.00001070, -0.00000922, -0.00000074, -0.00000293, 0.00000566, 0.00000972, -0.00000540, 0.00000263, -0.00000364, -0.00000521, 0.00001066, 0.00000582, -0.00000626, -0.00000966, -0.00000167, 0.00000580, 0.00000911, -0.00000831, -0.00001108, 0.00000991, -0.00000624, -0.00000867, 0.00000810, -0.00000179, -0.00000115, 0.00000589, -0.00000557, -0.00001381, -0.00000332, 0.00000786, 0.00000409, -0.00000238, -0.00000507, -0.00000739, -0.00000193, -0.00000312, -0.00000348, 0.00000532, -0.00000805, -0.00001095, 0.00000394, -0.00000260, },
		{ -0.00000132, -0.00000433, -0.00000043, -0.00000989, 0.00000455, 0.00000599, -0.00000951, -0.00000487, -0.00001255, -0.00000251, -0.00000803, -0.00002446, -0.00000856, -0.00005510, -0.00010644, 0.00001022, 0.00015667, 0.00007364, 0.00004056, 0.00008439, 0.00004116, 0.00017359, 0.00005853, -0.00017759, -0.00003609, -0.00001084, -0.00003840, 0.00000094, -0.00002397, -0.00000324, -0.00000786, -0.00000710, -0.00000528, -0.00000311, -0.00000870, 0.00000085, 0.00000788, -0.00001628, 0.00000294, -0.00000809, -0.00000258, -0.00001222, -0.00000498, 0.00000516, -0.00001818, 0.00003163, -0.00000778, -0.00000482, 0.00001155, -0.00000604, 0.00002318, -0.00002545, 0.00000778, -0.00000634, -0.00000714, 0.00001698, -0.00001600, 0.00003062, -0.00002833, 0.00001847, 0.00000735, -0.00004122, 0.00006298, -0.00005214, -0.00000957, 0.00004849, -0.00003731, 0.00002483, -0.00001691, -0.00000121, 0.00002053, -0.00002446, 0.00001379, -0.00000346, -0.00000104, 0.00000377, -0.00000247, 0.00001104, -0.00000618, -0.00000326, 0.00000057, -0.00000119, -0.00000591, 0.00000106, 0.00001072, -0.00001279, -0.00000009, 0.00001493, -0.00000695, -0.00000170, -0.00000068, 0.00000029, 0.00001014, -0.00000218, 0.00000120, 0.00000027, -0.00000696, 0.00000514, 0.00000918, -0.00000207, -0.00001187, -0.00000278, 0.00000451, 0.00000850, -0.00000201, -0.00001479, 0.00000565, 0.00000051, -0.00001134, 0.00000537, 0.00000169, -0.00000383, 0.00000558, -0.00000071, -0.00001389, -0.00000716, 0.00000621, 0.00000499, -0.00000095, -0.00000356, -0.00000780, -0.00000354, -0.00000171, -0.00000458, 0.00000433, -0.00000498, -0.00001269, 0.00000303, },
		{ 0.00000147, 0.00000057, -0.00000622, 0.00000112, -0.00000973, 0.00000079, 0.00000959, -0.00000973, -0.00000391, -0.00001059, -0.00000778, -0.00000149, -0.00002478, -0.00001333, -0.00002800, -0.00010432, -0.00005957, 0.00013672, 0.00012317, 0.00003378, 0.00008700, 0.00004735, 0.00015512, 0.00011272, -0.00017053, -0.00007260, 0.00000343, -0.00004344, -0.00000228, -0.00002083, -0.00000758, -0.00000665, -0.00000636, -0.00000762, -0.00000289, -0.00000466, -0.00000898, 0.00001466, -0.00001370, -0.00000310, -0.00000243, -0.00000635, -0.00000898, -0.00000748, 0.00000885, -0.00002135, 0.00002622, 0.00000007, -0.00001128, 0.00001619, -0.00001335, 0.00002839, -0.00001996, -0.00000124, 0.00000201, -0.00001888, 0.00002684, -0.00002342, 0.00003063, -0.00001657, -0.00000735, 0.00004021, -0.00005896, 0.00004732, -0.00001166, -0.00004478, 0.00005659, -0.00002584, 0.00000989, -0.00000051, -0.00001525, 0.00002586, -0.00001905, 0.00000402, 0.00000285, -0.00000338, 0.00000466, -0.00000338, 0.00001081, -0.00000313, -0.00000501, 0.00000017, -0.00000001, -0.00000437, -0.00000359, 0.00001292, -0.00000825, -0.00000818, 0.00001579, -0.00000251, -0.00000365, 0.00000218, -0.00000261, 0.00000736, 0.00000122, 0.00000057, 0.00000258, -0.00000566, -0.00000015, 0.00000856, 0.00000372, -0.00001216, -0.00000570, 0.00000465, 0.00000649, 0.00000290, -0.00001404, -0.00000110, 0.00000514, -0.00001025, 0.00000080, 0.00000453, -0.00000431, 0.00000302, 0.00000348, -0.00001162, -0.00001085, 0.00000382, 0.00000552, -0.00000012, -0.00000185, -0.00000714, -0.00000559, -0.00000093, -0.00000465, 0.00000272, -0.00000220, -0.00001368, },
		{ -0.00001384, -0.00000086, 0.00000299, -0.00000790, 0.00000150, -0.00000774, -0.00000409, 0.00001235, -0.00000818, -0.00000541, -0.00000626, -0.00001262, 0.00000058, -0.00001888, -0.00002193, -0.00001270, -0.00007808, -0.00011327, 0.00007889, 0.00016721, 0.00004912, 0.00007772, 0.00005964, 0.00013576, 0.00015583, -0.00014435, -0.00011299, 0.00000947, -0.00004108, -0.00000857, -0.00001810, -0.00000994, -0.00000755, -0.00000538, -0.00000728, -0.00000752, 0.00000312, -0.00001710, 0.00001412, -0.00000452, -0.00001100, 0.00000160, -0.00000713, -0.00000847, -0.00000903, 0.00001228, -0.00002266, 0.00001911, 0.00000854, -0.00001875, 0.00002217, -0.00001951, 0.00002771, -0.00000813, -0.00001360, 0.00001308, -0.00003134, 0.00003243, -0.00002130, 0.00001739, 0.00000766, -0.00003677, 0.00005928, -0.00004896, 0.00000948, 0.00002840, -0.00006067, 0.00004389, -0.00000432, -0.00000570, 0.00001149, -0.00002129, 0.00002269, -0.00000917, -0.00000443, 0.00000535, -0.00000336, 0.00000497, -0.00000409, 0.00000997, 0.00000007, -0.00000613, -0.00000039, 0.00000001, -0.00000160, -0.00000708, 0.00001162, -0.00000131, -0.00001402, 0.00001257, 0.00000298, -0.00000520, 0.00000396, -0.00000248, 0.00000311, 0.00000285, 0.00000137, 0.00000362, -0.00000319, -0.00000301, 0.00000464, 0.00000837, -0.00000912, -0.00000991, 0.00000482, 0.00000504, 0.00000493, -0.00000993, -0.00000727, 0.00000586, -0.00000638, -0.00000344, 0.00000518, -0.00000259, -0.00000019, 0.00000552, -0.00000739, -0.00001333, 0.00000047, 0.00000577, 0.00000028, -0.00000060, -0.00000535, -0.00000741, -0.00000121, -0.00000391, 0.00000096, 0.00000004, },
		{ 0.00000157, -0.00001310, -0.00000393, 0.00000547, -0.00000852, 0.00000012, -0.00000426, -0.00000873, 0.00001283, -0.00000427, -0.00000856, -0.00000203, -0.00001420, -0.00000230, -0.00001062, -0.00002658, -0.00001256, -0.00004243, -0.00013421, -0.00000467, 0.00018348, 0.00008895, 0.00006613, 0.00006881, 0.00012162, 0.00018645, -0.00010425, -0.00014920, 0.00000418, -0.00003396, -0.00001404, -0.00001788, -0.00001067, -0.00000822, -0.00000726, -0.00000253, -0.00001545, 0.00000978, -0.00001843, 0.00000526, 0.00000763, -0.00001606, 0.00000119, -0.00000466, -0.00000894, -0.00001160, 0.00001604, -0.00002173, 0.00000959, 0.00001867, -0.00002747, 0.00002737, -0.00002064, 0.00001849, 0.00000972, -0.00002751, 0.00002362, -0.00003874, 0.00002732, -0.00000616, -0.00000648, 0.00003426, -0.00005480, 0.00005363, -0.00001469, -0.00003297, 0.00005010, -0.00005178, 0.00001748, 0.00001646, -0.00001439, 0.00001584, -0.00001980, 0.00001471, 0.00000043, -0.00000884, 0.00000457, -0.00000282, 0.00000577, -0.00000462, 0.00000827, 0.00000351, -0.00000684, -0.00000074, -0.00000083, 0.00000101, -0.00000811, 0.00000738, 0.00000551, -0.00001582, 0.00000625, 0.00000757, -0.00000512, 0.00000406, -0.00000025, -0.00000038, 0.00000187, 0.00000277, 0.00000463, -0.00000118, -0.00000307, -0.00000024, 0.00000959, -0.00000319, -0.00001348, 0.00000339, 0.00000520, 0.00000446, -0.00000490, -0.00001046, 0.00000296, -0.00000200, -0.00000551, 0.00000347, 0.00000002, -0.00000230, 0.00000499, -0.00000258, -0.00001362, -0.00000357, 0.00000543, 0.00000068, -0.00000024, -0.00000291, -0.00000820, -0.00000259, -0.00000288, -0.00000048, },
		{ -0.00000131, 0.00000239, -0.00001155, -0.00000738, 0.00000719, -0.00000745, -0.00000278, -0.00000048, -0.00001161, 0.00001042, 0.00000113, -0.00001137, -0.00000015, -0.00001188, -0.00000743, -0.00000516, -0.00002345, -0.00002230, -0.00001429, -0.00011845, -0.00008986, 0.00015603, 0.00014205, 0.00006683, 0.00006909, 0.00011255, 0.00020837, -0.00005692, -0.00017608, -0.00001081, -0.00002705, -0.00001602, -0.00001934, -0.00001263, -0.00000557, -0.00001342, 0.00000500, -0.00002205, 0.00001011, -0.00001112, -0.00000846, 0.00001687, -0.00001463, -0.00000362, -0.00000132, -0.00000766, -0.00001624, 0.00001899, -0.00001682, -0.00000344, 0.00003047, -0.00003526, 0.00002756, -0.00001286, 0.00000041, 0.00003026, -0.00003829, 0.00002798, -0.00003506, 0.00000869, 0.00001856, -0.00003115, 0.00004986, -0.00005101, 0.00002431, 0.00002845, -0.00006022, 0.00004450, -0.00002467, -0.00000933, 0.00002708, -0.00001365, 0.00001412, -0.00001434, 0.00000627, 0.00000677, -0.00000932, 0.00000266, -0.00000335, 0.00000701, -0.00000417, 0.00000533, 0.00000704, -0.00000700, -0.00000108, -0.00000173, 0.00000233, -0.00000661, 0.00000200, 0.00000995, -0.00001346, -0.00000096, 0.00000953, -0.00000325, 0.00000310, 0.00000241, -0.00000162, -0.00000101, 0.00000312, 0.00000654, 0.00000003, -0.00000157, -0.00000356, 0.00000688, 0.00000345, -0.00001432, -0.00000021, 0.00000649, 0.00000312, -0.00000122, -0.00001009, -0.00000159, 0.00000060, -0.00000491, 0.00000058, 0.00000185, -0.00000243, 0.00000278, 0.00000109, -0.00001147, -0.00000736, 0.00000398, 0.00000136, -0.00000066, -0.00000071, -0.00000750, -0.00000460, -0.00000224, },
		{ -0.00000251, -0.00000159, 0.00000272, -0.00000947, -0.00001063, 0.00000751, -0.00000469, -0.00000623, 0.00000219, -0.00001196, 0.00000587, 0.00000616, -0.00001207, -0.00000120, -0.00000756, -0.00001088, -0.00000541, -0.00001457, -0.00003194, -0.00000463, -0.00007680, -0.00015001, 0.00008420, 0.00018568, 0.00009114, 0.00006339, 0.00010249, 0.00022656, -0.00000618, -0.00019333, -0.00003078, -0.00002421, -0.00001549, -0.00001870, -0.00001915, 0.00000140, -0.00002178, 0.00001055, -0.00002220, 0.00000191, 0.00000190, -0.00002053, 0.00001819, -0.00000662, -0.00000954, -0.00000062, -0.00000313, -0.00002160, 0.00001819, -0.00000606, -0.00001962, 0.00004128, -0.00003753, 0.00001820, 0.00000503, -0.00002266, 0.00004677, -0.00003964, 0.00002104, -0.00001810, -0.00001947, 0.00004268, -0.00004406, 0.00004518, -0.00002562, -0.00001581, 0.00006075, -0.00005997, 0.00001600, 0.00000562, -0.00002420, 0.00002455, -0.00000684, 0.00001074, -0.00000864, -0.00000007, 0.00000958, -0.00000782, 0.00000163, -0.00000508, 0.00000749, -0.00000194, 0.00000139, 0.00000997, -0.00000619, -0.00000178, -0.00000206, 0.00000207, -0.00000361, -0.00000251, 0.00001101, -0.00000845, -0.00000664, 0.00000832, -0.00000060, 0.00000227, 0.00000421, -0.00000058, -0.00000392, 0.00000114, 0.00000906, 0.00000130, -0.00000022, -0.00000408, 0.00000173, 0.00000806, -0.00001154, -0.00000502, 0.00000745, 0.00000254, 0.00000016, -0.00000733, -0.00000534, 0.00000037, -0.00000270, -0.00000175, 0.00000188, -0.00000098, 0.00000051, 0.00000255, -0.00000773, -0.00000967, 0.00000127, 0.00000207, -0.00000126, 0.00000049, -0.00000545, -0.00000634, },
		{ -0.00000693, -0.00000369, -0.00000177, 0.00000292, -0.00000718, -0.00001319, 0.00000624, -0.00000097, -0.00000895, 0.00000295, -0.00001020, 0.00000084, 0.00000918, -0.00001027, -0.00000377, -0.00000418, -0.00001026, -0.00001011, -0.00000606, -0.00003297, -0.00001363, -0.00002960, -0.00016735, -0.00001453, 0.00019467, 0.00013846, 0.00006381, 0.00008443, 0.00024200, 0.00004924, -0.00020382, -0.00005187, -0.00002510, -0.00001715, -0.00001179, -0.00003090, 0.00000978, -0.00002717, 0.00000906, -0.00001372, -0.00001217, 0.00001409, -0.00002485, 0.00001062, 0.00000396, -0.00001185, -0.00000456, 0.00000323, -0.00002403, 0.00001062, 0.00001033, -0.00003552, 0.00004592, -0.00002926, -0.00000244, 0.00002924, -0.00004265, 0.00005124, -0.00002694, 0.00000197, 0.00000778, -0.00004681, 0.00005384, -0.00003686, 0.00002070, 0.00001015, -0.00004853, 0.00006805, -0.00003331, -0.00002020, 0.00002355, -0.00002222, 0.00001352, -0.00000044, 0.00000943, -0.00000425, -0.00000418, 0.00001037, -0.00000640, 0.00000206, -0.00000672, 0.00000590, 0.00000203, -0.00000249, 0.00001136, -0.00000407, -0.00000296, -0.00000170, 0.00000075, -0.00000058, -0.00000488, 0.00000922, -0.00000303, -0.00000932, 0.00000490, 0.00000125, 0.00000238, 0.00000493, 0.00000158, -0.00000508, -0.00000310, 0.00001072, 0.00000381, 0.00000010, -0.00000230, -0.00000328, 0.00000886, -0.00000609, -0.00000905, 0.00000674, 0.00000334, -0.00000021, -0.00000421, -0.00000673, -0.00000205, -0.00000071, -0.00000236, 0.00000034, 0.00000072, -0.00000047, 0.00000189, -0.00000398, -0.00000966, -0.00000211, 0.00000213, -0.00000146, 0.00000046, -0.00000287, },
		{ -0.00000081, -0.00000602, -0.00000518, -0.00000245, 0.00000316, -0.00000482, -0.00001488, 0.00000370, 0.00000270, -0.00001013, 0.00000203, -0.00000763, -0.00000315, 0.00000981, -0.00000723, -0.00000562, -0.00000361, -0.00000628, -0.00001483, -0.00000383, -0.00002331, -0.00003166, 0.00000291, -0.00014096, -0.00011023, 0.00015361, 0.00019255, 0.00008564, 0.00005755, 0.00024880, 0.00011374, -0.00020912, -0.00007468, -0.00002496, -0.00002638, 0.00000218, -0.00004387, 0.00001349, -0.00002436, -0.00000140, 0.00000056, -0.00002557, 0.00001884, -0.00001936, -0.00000186, 0.00001102, -0.00000765, -0.00001149, 0.00000713, -0.00001960, -0.00000420, 0.00002841, -0.00004492, 0.00003886, -0.00000843, -0.00003065, 0.00005096, -0.00004988, 0.00003831, -0.00000089, -0.00002365, 0.00003242, -0.00006112, 0.00004403, -0.00001064, -0.00001281, 0.00003966, -0.00005960, 0.00004850, 0.00000632, -0.00004609, 0.00002107, -0.00000817, 0.00000304, 0.00000044, 0.00001056, -0.00000010, -0.00000733, 0.00001066, -0.00000576, 0.00000291, -0.00000661, 0.00000189, 0.00000653, -0.00000486, 0.00001064, -0.00000088, -0.00000429, -0.00000106, -0.00000073, 0.00000140, -0.00000502, 0.00000608, 0.00000095, -0.00000906, 0.00000113, 0.00000119, 0.00000320, 0.00000530, 0.00000342, -0.00000374, -0.00000808, 0.00000979, 0.00000788, -0.00000014, 0.00000010, -0.00000601, 0.00000600, -0.00000027, -0.00001060, 0.00000416, 0.00000490, -0.00000098, -0.00000229, -0.00000570, -0.00000491, -0.00000042, -0.00000133, -0.00000158, 0.00000136, 0.00000019, 0.00000031, -0.00000174, -0.00000751, -0.00000493, 0.00000096, -0.00000113, -0.00000035, },
		{ -0.00000112, 0.00000006, -0.00000405, -0.00000599, -0.00000396, 0.00000322, -0.00000227, -0.00001577, 0.00000045, 0.00000557, -0.00000980, 0.00000039, -0.00000545, -0.00000564, 0.00000907, -0.00000480, -0.00000543, -0.00000532, -0.00000222, -0.00001499, -0.00000961, -0.00000813, -0.00004536, 0.00000995, -0.00008696, -0.00017281, 0.00006633, 0.00022606, 0.00013714, 0.00003199, 0.00023654, 0.00018928, -0.00020507, -0.00010484, -0.00001797, -0.00004468, 0.00001823, -0.00005040, 0.00000673, -0.00001181, -0.00001749, 0.00001385, -0.00003148, 0.00001348, -0.00000755, -0.00001234, 0.00001021, 0.00000156, -0.00001626, 0.00000391, -0.00000725, -0.00002233, 0.00004094, -0.00004140, 0.00001791, 0.00002120, -0.00005710, 0.00005937, -0.00003779, 0.00000917, 0.00003091, -0.00004515, 0.00004436, -0.00005460, 0.00001460, 0.00002364, -0.00003955, 0.00004947, -0.00004576, 0.00001292, 0.00004035, -0.00004994, 0.00000138, 0.00000673, 0.00000078, -0.00000444, 0.00001114, 0.00000570, -0.00001025, 0.00001074, -0.00000515, 0.00000258, -0.00000397, -0.00000343, 0.00000969, -0.00000473, 0.00000806, 0.00000260, -0.00000517, -0.00000063, -0.00000173, 0.00000201, -0.00000375, 0.00000318, 0.00000281, -0.00000718, -0.00000128, -0.00000070, 0.00000362, 0.00000625, 0.00000421, -0.00000062, -0.00001163, 0.00000547, 0.00001241, 0.00000045, 0.00000155, -0.00000585, 0.00000129, 0.00000357, -0.00000924, 0.00000075, 0.00000606, -0.00000107, -0.00000198, -0.00000348, -0.00000652, -0.00000205, 0.00000019, -0.00000259, 0.00000051, 0.00000173, -0.00000067, -0.00000159, -0.00000448, -0.00000596, -0.00000138, -0.00000082, },
		{ -0.00000135, -0.00000135, -0.00000017, -0.00000205, -0.00000542, -0.00000595, 0.00000251, 0.00000061, -0.00001589, -0.00000316, 0.00000749, -0.00000861, -0.00000098, -0.00000399, -0.00000741, 0.00000851, -0.00000399, -0.00000387, -0.00000678, -0.00000153, -0.00000911, -0.00001974, 0.00000366, -0.00004535, -0.00000553, -0.00003061, -0.00018524, -0.00004313, 0.00021250, 0.00021050, 0.00002703, 0.00019778, 0.00026950, -0.00018101, -0.00014899, -0.00000275, -0.00006657, 0.00002672, -0.00004329, -0.00001176, 0.00000619, -0.00003179, 0.00001912, -0.00002700, 0.00000111, 0.00000351, -0.00001519, 0.00000214, 0.00000987, -0.00001308, -0.00000791, 0.00000907, -0.00003599, 0.00004046, -0.00002235, -0.00001289, 0.00004991, -0.00007014, 0.00004683, -0.00000727, -0.00002709, 0.00005607, -0.00005156, 0.00003684, -0.00002854, -0.00002344, 0.00005020, -0.00004674, 0.00003623, -0.00001590, -0.00002135, 0.00005458, -0.00003176, -0.00002340, 0.00001148, 0.00000830, -0.00001031, 0.00000726, 0.00001355, -0.00001203, 0.00000964, -0.00000332, 0.00000023, 0.00000040, -0.00000802, 0.00001006, -0.00000222, 0.00000473, 0.00000538, -0.00000518, -0.00000064, -0.00000203, 0.00000161, -0.00000215, 0.00000140, 0.00000307, -0.00000531, -0.00000166, -0.00000319, 0.00000243, 0.00000801, 0.00000429, 0.00000272, -0.00001216, -0.00000144, 0.00001537, 0.00000303, 0.00000167, -0.00000377, -0.00000291, 0.00000431, -0.00000609, -0.00000195, 0.00000605, -0.00000028, -0.00000265, -0.00000154, -0.00000617, -0.00000462, 0.00000081, -0.00000218, -0.00000126, 0.00000286, -0.00000015, -0.00000291, -0.00000221, -0.00000472, -0.00000397, },
		{ -0.00000542, -0.00000318, -0.00000122, -0.00000072, -0.00000101, -0.00000350, -0.00000746, 0.00000047, 0.00000363, -0.00001502, -0.00000709, 0.00000871, -0.00000722, -0.00000192, -0.00000250, -0.00000965, 0.00000885, -0.00000403, -0.00000313, -0.00000545, -0.00000532, 0.00000011, -0.00002743, 0.00000511, -0.00003118, -0.00002994, 0.00000572, -0.00015095, -0.00013945, 0.00014046, 0.00027992, 0.00006179, 0.00013661, 0.00033817, -0.00012454, -0.00020821, 0.00001372, -0.00008069, 0.00001830, -0.00002095, -0.00003639, 0.00002100, -0.00003688, 0.00001353, -0.00001521, -0.00001118, 0.00000755, -0.00000998, -0.00000716, 0.00001011, 0.00000023, -0.00002405, 0.00002106, -0.00003699, 0.00002368, 0.00000797, -0.00004320, 0.00006546, -0.00006127, 0.00001374, 0.00003189, -0.00005625, 0.00006280, -0.00003749, 0.00001199, 0.00000647, -0.00005429, 0.00005630, -0.00003135, 0.00000810, 0.00001420, -0.00003959, 0.00004574, -0.00000235, -0.00003923, 0.00000166, 0.00002007, -0.00001059, -0.00000249, 0.00002109, -0.00001061, 0.00000647, 0.00000012, -0.00000354, 0.00000462, -0.00001002, 0.00000743, 0.00000143, 0.00000200, 0.00000688, -0.00000436, -0.00000099, -0.00000184, 0.00000082, -0.00000096, 0.00000071, 0.00000283, -0.00000450, -0.00000065, -0.00000458, -0.00000083, 0.00000980, 0.00000472, 0.00000485, -0.00000950, -0.00000875, 0.00001488, 0.00000753, 0.00000139, -0.00000146, -0.00000505, 0.00000249, -0.00000306, -0.00000295, 0.00000508, 0.00000079, -0.00000335, -0.00000074, -0.00000435, -0.00000667, -0.00000019, -0.00000090, -0.00000279, 0.00000267, 0.00000168, -0.00000426, -0.00000189, -0.00000188, },
		{ 0.00000097, -0.00000464, -0.00000584, -0.00000160, -0.00000079, -0.00000123, -0.00000117, -0.00000741, -0.00000288, 0.00000603, -0.00001268, -0.00001138, 0.00000935, -0.00000567, -0.00000307, 0.00000004, -0.00001278, 0.00000923, -0.00000287, -0.00000527, -0.00000079, -0.00001133, 0.00000740, -0.00002707, -0.00000452, -0.00001086, -0.00004736, 0.00001260, -0.00009163, -0.00019137, 0.00002247, 0.00030995, 0.00014232, 0.00007270, 0.00037461, -0.00002992, -0.00027273, 0.00001693, -0.00007543, -0.00000989, 0.00000982, -0.00005654, 0.00002431, -0.00003007, 0.00000051, -0.00000339, -0.00001682, 0.00000342, -0.00000225, -0.00000963, -0.00000119, 0.00001925, -0.00003525, 0.00001992, -0.00002164, -0.00000517, 0.00003850, -0.00006011, 0.00005891, -0.00003040, -0.00002986, 0.00006409, -0.00006504, 0.00004605, -0.00000681, -0.00001945, 0.00003539, -0.00006508, 0.00003883, -0.00000156, -0.00001994, 0.00003038, -0.00003673, 0.00002198, 0.00002320, -0.00003782, -0.00001817, 0.00002684, -0.00000145, -0.00001519, 0.00002419, -0.00000460, 0.00000163, 0.00000424, -0.00000707, 0.00000680, -0.00000884, 0.00000298, 0.00000455, 0.00000071, 0.00000721, -0.00000315, -0.00000142, -0.00000141, 0.00000006, -0.00000035, 0.00000051, 0.00000299, -0.00000473, 0.00000035, -0.00000380, -0.00000532, 0.00001021, 0.00000635, 0.00000541, -0.00000490, -0.00001401, 0.00001042, 0.00001243, 0.00000217, -0.00000006, -0.00000508, -0.00000018, -0.00000162, -0.00000249, 0.00000409, 0.00000149, -0.00000361, -0.00000099, -0.00000221, -0.00000717, -0.00000241, 0.00000011, -0.00000323, 0.00000118, 0.00000377, -0.00000433, -0.00000354, },
		{ -0.00000607, 0.00000222, -0.00000167, -0.00000802, -0.00000334, -0.00000018, -0.00000209, 0.00000027, -0.00000531, -0.00000661, 0.00000672, -0.00000861, -0.00001569, 0.00000902, -0.00000338, -0.00000535, 0.00000394, -0.00001573, 0.00000764, 0.00000131, -0.00001041, 0.00000494, -0.00001534, 0.00000830, -0.00001781, -0.00001940, 0.00000447, -0.00004835, -0.00000357, -0.00003618, -0.00018598, -0.00010708, 0.00027141, 0.00025210, 0.00003639, 0.00036467, 0.00009457, -0.00032236, -0.00000855, -0.00004659, -0.00005048, 0.00003587, -0.00006206, 0.00001347, -0.00001562, -0.00001212, 0.00000173, -0.00001446, -0.00000364, -0.00000036, -0.00000063, -0.00002010, 0.00003362, -0.00003166, 0.00000191, 0.00000555, -0.00003451, 0.00005558, -0.00005438, 0.00003007, 0.00001249, -0.00006755, 0.00007468, -0.00004786, 0.00001104, 0.00002820, -0.00004302, 0.00004607, -0.00005216, 0.00000578, 0.00002765, -0.00003420, 0.00002719, -0.00001882, -0.00000262, 0.00003369, -0.00002068, -0.00003730, 0.00002136, 0.00001507, -0.00002463, 0.00001960, 0.00000503, -0.00000284, 0.00000738, -0.00000881, 0.00000617, -0.00000536, -0.00000142, 0.00000594, 0.00000079, 0.00000706, -0.00000200, -0.00000181, -0.00000088, -0.00000050, -0.00000009, 0.00000020, 0.00000377, -0.00000514, 0.00000007, -0.00000098, -0.00000932, 0.00000808, 0.00000910, 0.00000511, -0.00000024, -0.00001571, 0.00000318, 0.00001554, 0.00000489, 0.00000053, -0.00000410, -0.00000207, -0.00000201, -0.00000178, 0.00000396, 0.00000175, -0.00000360, -0.00000169, -0.00000072, -0.00000612, -0.00000469, -0.00000009, -0.00000259, -0.00000076, 0.00000495, -0.00000267, },
		{ 0.00000012, -0.00000792, 0.00000110, 0.00000229, -0.00000822, -0.00000655, 0.00000037, -0.00000244, -0.00000003, -0.00000181, -0.00000918, 0.00000491, -0.00000328, -0.00001906, 0.00000697, 0.00000029, -0.00000895, 0.00000810, -0.00001637, 0.00000236, 0.00000843, -0.00001614, 0.00000793, -0.00001391, 0.00000214, -0.00000406, -0.00003083, 0.00000740, -0.00003440, -0.00002606, -0.00000654, -0.00013467, -0.00020529, 0.00015781, 0.00035272, 0.00005498, 0.00031173, 0.00022580, -0.00033397, -0.00006987, -0.00000222, -0.00008794, 0.00004421, -0.00004965, -0.00000612, -0.00000243, -0.00001733, -0.00000112, -0.00000930, -0.00000525, -0.00000954, 0.00001667, -0.00003569, 0.00003191, -0.00000907, -0.00002774, 0.00003229, -0.00005007, 0.00004955, -0.00002602, -0.00001083, 0.00005066, -0.00008360, 0.00005717, -0.00001041, -0.00002820, 0.00005238, -0.00004782, 0.00003526, -0.00002280, -0.00002785, 0.00004243, -0.00002928, 0.00000978, 0.00000152, -0.00001602, 0.00002717, 0.00000243, -0.00004506, 0.00000304, 0.00003157, -0.00002480, 0.00000755, 0.00001468, -0.00000438, 0.00000844, -0.00000831, 0.00000337, -0.00000132, -0.00000426, 0.00000550, 0.00000138, 0.00000707, -0.00000100, -0.00000233, -0.00000021, -0.00000089, 0.00000009, -0.00000043, 0.00000469, -0.00000473, -0.00000188, 0.00000259, -0.00001108, 0.00000335, 0.00001183, 0.00000521, 0.00000304, -0.00001391, -0.00000447, 0.00001522, 0.00000906, 0.00000140, -0.00000331, -0.00000259, -0.00000319, -0.00000209, 0.00000470, 0.00000222, -0.00000377, -0.00000234, -0.00000013, -0.00000430, -0.00000591, -0.00000159, -0.00000166, -0.00000218, 0.00000468, },
		{ 0.00000329, 0.00000285, -0.00000803, -0.00000185, 0.00000538, -0.00000570, -0.00001020, -0.00000037, -0.00000152, -0.00000184, 0.00000151, -0.00000919, 0.00000082, 0.00000193, -0.00002023, 0.00000261, 0.00000526, -0.00001285, 0.00001042, -0.00001284, -0.00000649, 0.00001598, -0.00001873, 0.00000530, -0.00000686, -0.00000748, 0.00000728, -0.00003260, -0.00000161, -0.00001578, -0.00003826, -0.00000784, -0.00006720, -0.00023975, -0.00000681, 0.00039699, 0.00013631, 0.00024040, 0.00033414, -0.00029296, -0.00015918, 0.00003901, -0.00010605, 0.00002970, -0.00002568, -0.00002373, 0.00000176, -0.00001416, -0.00000625, -0.00000993, 0.00000400, -0.00002700, 0.00003128, -0.00003527, 0.00000878, 0.00002660, -0.00005531, 0.00004360, -0.00004204, 0.00002066, 0.00001411, -0.00004674, 0.00006801, -0.00007068, 0.00001714, 0.00003205, -0.00005506, 0.00005510, -0.00003203, 0.00000998, 0.00000888, -0.00004782, 0.00003693, -0.00001015, -0.00000998, 0.00001261, -0.00001430, 0.00001088, 0.00001942, -0.00003696, -0.00002129, 0.00003916, -0.00001376, -0.00000751, 0.00001981, -0.00000173, 0.00000782, -0.00000631, -0.00000021, 0.00000173, -0.00000512, 0.00000409, 0.00000158, 0.00000742, 0.00000021, -0.00000320, 0.00000050, -0.00000107, 0.00000029, -0.00000117, 0.00000507, -0.00000301, -0.00000481, 0.00000520, -0.00000976, -0.00000275, 0.00001298, 0.00000654, 0.00000466, -0.00001000, -0.00001020, 0.00001132, 0.00001295, 0.00000372, -0.00000308, -0.00000234, -0.00000376, -0.00000384, 0.00000536, 0.00000367, -0.00000421, -0.00000300, -0.00000008, -0.00000267, -0.00000568, -0.00000361, -0.00000141, -0.00000262, },
		{ -0.00000243, 0.00000163, 0.00000457, -0.00000638, -0.00000520, 0.00000618, -0.00000108, -0.00001248, -0.00000328, 0.00000039, -0.00000386, 0.00000299, -0.00000634, -0.00000411, 0.00000527, -0.00001838, -0.00000371, 0.00001034, -0.00001513, 0.00000898, -0.00000501, -0.00001640, 0.00002019, -0.00001541, -0.00000275, 0.00000240, -0.00001498, 0.00001111, -0.00002458, -0.00001509, -0.00000429, -0.00003313, -0.00002783, -0.00001677, -0.00020376, -0.00017409, 0.00034989, 0.00025838, 0.00018864, 0.00039787, -0.00020236, -0.00025471, 0.00005635, -0.00009715, -0.00000117, -0.00000323, -0.00003008, -0.00000372, -0.00000925, -0.00000458, -0.00002174, 0.00002124, -0.00004158, 0.00002974, -0.00001221, -0.00002990, 0.00005986, -0.00006436, 0.00002963, -0.00001126, -0.00001942, 0.00004825, -0.00006147, 0.00005701, -0.00003400, -0.00002951, 0.00006146, -0.00005820, 0.00003568, -0.00000380, -0.00001647, 0.00002917, -0.00004746, 0.00001575, 0.00001140, -0.00002073, 0.00000993, -0.00000260, -0.00000386, 0.00002270, -0.00001698, -0.00004095, 0.00003255, 0.00000460, -0.00001885, 0.00001756, 0.00000389, 0.00000724, -0.00000399, -0.00000337, 0.00000318, -0.00000463, 0.00000283, 0.00000103, 0.00000760, 0.00000211, -0.00000441, 0.00000098, -0.00000088, 0.00000041, -0.00000163, 0.00000449, -0.00000029, -0.00000744, 0.00000558, -0.00000588, -0.00000824, 0.00001143, 0.00000892, 0.00000538, -0.00000575, -0.00001279, 0.00000526, 0.00001462, 0.00000767, -0.00000264, -0.00000245, -0.00000300, -0.00000619, 0.00000465, 0.00000626, -0.00000429, -0.00000408, -0.00000014, -0.00000168, -0.00000445, -0.00000507, -0.00000227, },
		{ -0.00000386, -0.00000239, 0.00000038, 0.00000507, -0.00000381, -0.00000749, 0.00000444, 0.00000386, -0.00001186, -0.00000800, 0.00000192, -0.00000451, 0.00000186, -0.00000178, -0.00000774, 0.00000549, -0.00001386, -0.00001054, 0.00001366, -0.00001410, 0.00000341, 0.00000496, -0.00002358, 0.00001816, -0.00000634, -0.00001278, 0.00000887, -0.00001632, 0.00000717, -0.00001235, -0.00002374, -0.00000579, -0.00001622, -0.00004567, -0.00000372, -0.00011987, -0.00029022, 0.00020764, 0.00037270, 0.00018978, 0.00041644, -0.00008315, -0.00033016, 0.00003803, -0.00006733, -0.00003330, 0.00000591, -0.00002427, -0.00001084, -0.00001274, 0.00000934, -0.00004131, 0.00003473, -0.00003924, 0.00000461, 0.00002798, -0.00006730, 0.00007205, -0.00004443, -0.00000775, 0.00002945, -0.00005208, 0.00005990, -0.00004784, 0.00002295, 0.00001053, -0.00006373, 0.00006545, -0.00003722, 0.00000376, 0.00002308, -0.00003157, 0.00003143, -0.00003029, -0.00000940, 0.00002367, -0.00001771, -0.00000237, 0.00000906, -0.00000843, 0.00001290, 0.00000481, -0.00004720, 0.00001336, 0.00002212, -0.00002107, 0.00000885, 0.00000907, 0.00000846, -0.00000183, -0.00000580, 0.00000338, -0.00000381, 0.00000229, 0.00000020, 0.00000689, 0.00000481, -0.00000547, 0.00000075, -0.00000022, 0.00000037, -0.00000161, 0.00000306, 0.00000255, -0.00000860, 0.00000360, -0.00000103, -0.00001134, 0.00000729, 0.00001122, 0.00000628, -0.00000236, -0.00001250, -0.00000074, 0.00001306, 0.00001209, -0.00000072, -0.00000351, -0.00000141, -0.00000766, 0.00000189, 0.00000910, -0.00000308, -0.00000576, -0.00000032, -0.00000117, -0.00000305, -0.00000535, },
		{ -0.00000456, -0.00000530, -0.00000305, -0.00000045, 0.00000483, -0.00000138, -0.00000806, 0.00000123, 0.00000713, -0.00000800, -0.00001287, 0.00000139, -0.00000306, -0.00000112, 0.00000234, -0.00000836, 0.00000276, -0.00000829, -0.00001588, 0.00001372, -0.00000941, -0.00000465, 0.00001356, -0.00002486, 0.00000978, 0.00000511, -0.00001988, 0.00000914, -0.00001161, -0.00000016, -0.00000324, -0.00002251, -0.00001678, -0.00000104, -0.00004475, -0.00002631, -0.00002928, -0.00031925, 0.00000445, 0.00042241, 0.00025263, 0.00041367, 0.00003645, -0.00036823, -0.00001172, -0.00003359, -0.00005222, -0.00000038, -0.00001578, -0.00000762, -0.00003043, 0.00003120, -0.00005582, 0.00002971, -0.00001208, -0.00003873, 0.00006763, -0.00008261, 0.00005114, 0.00000178, -0.00005375, 0.00006020, -0.00006041, 0.00004226, -0.00001193, -0.00001801, 0.00004358, -0.00007220, 0.00004328, -0.00000279, -0.00002560, 0.00003645, -0.00002982, 0.00001857, -0.00000710, -0.00002634, 0.00002135, -0.00000472, -0.00001499, 0.00001217, -0.00000147, -0.00000218, 0.00001804, -0.00003796, -0.00001046, 0.00003056, -0.00001326, -0.00000179, 0.00001023, 0.00001173, 0.00000083, -0.00000782, 0.00000296, -0.00000328, 0.00000223, -0.00000004, 0.00000489, 0.00000780, -0.00000562, -0.00000048, 0.00000080, 0.00000026, -0.00000117, 0.00000130, 0.00000461, -0.00000786, 0.00000016, 0.00000295, -0.00001133, 0.00000181, 0.00001205, 0.00000794, 0.00000005, -0.00001064, -0.00000488, 0.00000878, 0.00001503, 0.00000337, -0.00000494, -0.00000030, -0.00000713, -0.00000230, 0.00001065, -0.00000004, -0.00000753, -0.00000106, -0.00000073, -0.00000206, },
		{ -0.00000144, -0.00000336, -0.00000591, -0.00000428, -0.00000136, 0.00000445, 0.00000037, -0.00000729, -0.00000186, 0.00000759, -0.00000217, -0.00001562, -0.00000207, -0.00000024, -0.00000407, 0.00000402, -0.00000575, -0.00000126, -0.00000384, -0.00001826, 0.00001031, -0.00000245, -0.00001227, 0.00001768, -0.00001962, -0.00000196, 0.00001415, -0.00002072, 0.00000358, -0.00000482, -0.00000507, -0.00000142, -0.00001359, -0.00002718, 0.00000026, -0.00002236, -0.00006347, 0.00002786, -0.00025946, -0.00019813, 0.00036823, 0.00035085, 0.00042671, 0.00013795, -0.00036992, -0.00007520, -0.00001335, -0.00005412, -0.00001181, -0.00001913, 0.00001280, -0.00005691, 0.00004642, -0.00004931, -0.00000204, 0.00003423, -0.00008172, 0.00008422, -0.00006189, -0.00000020, 0.00005720, -0.00008637, 0.00006356, -0.00003811, 0.00000249, 0.00002917, -0.00004645, 0.00005165, -0.00005366, 0.00000606, 0.00002874, -0.00003958, 0.00003179, -0.00001490, 0.00000030, 0.00001040, -0.00002875, 0.00000762, 0.00000910, -0.00001956, 0.00000461, 0.00001116, -0.00001265, 0.00001765, -0.00001881, -0.00002873, 0.00002609, 0.00000016, -0.00000856, 0.00000593, 0.00001516, 0.00000527, -0.00000962, 0.00000212, -0.00000293, 0.00000185, 0.00000094, 0.00000201, 0.00001006, -0.00000425, -0.00000257, 0.00000174, 0.00000033, -0.00000057, -0.00000024, 0.00000537, -0.00000561, -0.00000336, 0.00000486, -0.00000880, -0.00000318, 0.00001064, 0.00000997, 0.00000216, -0.00000856, -0.00000657, 0.00000358, 0.00001490, 0.00000899, -0.00000524, -0.00000084, -0.00000472, -0.00000626, 0.00000965, 0.00000431, -0.00000829, -0.00000276, -0.00000029, },
		{ -0.00000028, -0.00000084, -0.00000233, -0.00000564, -0.00000537, -0.00000282, 0.00000404, 0.00000164, -0.00000615, -0.00000362, 0.00000563, 0.00000335, -0.00001469, -0.00000768, 0.00000203, -0.00000510, 0.00000265, -0.00000140, -0.00000416, -0.00000202, -0.00001767, 0.00000480, 0.00000438, -0.00001680, 0.00001625, -0.00001009, -0.00001253, 0.00001729, -0.00001552, -0.00000392, -0.00000096, -0.00000446, -0.00000527, -0.00000463, -0.00002733, -0.00001502, 0.00000932, -0.00008763, 0.00003131, -0.00014417, -0.00033600, 0.00020992, 0.00042956, 0.00048411, 0.00022639, -0.00035261, -0.00013318, -0.00001236, -0.00004986, -0.00001150, -0.00004312, 0.00004410, -0.00007509, 0.00003720, -0.00001319, -0.00005412, 0.00007982, -0.00010017, 0.00006211, -0.00000722, -0.00006313, 0.00009685, -0.00008685, 0.00003389, 0.00000636, -0.00004106, 0.00005543, -0.00004917, 0.00003353, -0.00001917, -0.00002894, 0.00004353, -0.00003388, 0.00001397, 0.00000332, -0.00001282, 0.00001563, -0.00001862, -0.00000850, 0.00001507, -0.00001354, -0.00000882, 0.00002061, -0.00001229, 0.00000621, 0.00000027, -0.00003434, 0.00001139, 0.00001171, -0.00000771, -0.00000181, 0.00001565, 0.00001203, -0.00001054, 0.00000048, -0.00000222, 0.00000055, 0.00000295, -0.00000068, 0.00001064, -0.00000133, -0.00000484, 0.00000203, 0.00000076, -0.00000000, -0.00000120, 0.00000493, -0.00000275, -0.00000579, 0.00000455, -0.00000517, -0.00000625, 0.00000731, 0.00001135, 0.00000467, -0.00000689, -0.00000652, -0.00000051, 0.00001142, 0.00001423, -0.00000295, -0.00000302, -0.00000183, -0.00000822, 0.00000606, 0.00000841, -0.00000703, -0.00000519, },
		{ -0.00000735, -0.00000133, -0.00000011, -0.00000146, -0.00000507, -0.00000565, -0.00000472, 0.00000312, 0.00000296, -0.00000536, -0.00000397, 0.00000277, 0.00000650, -0.00001019, -0.00001322, 0.00000172, -0.00000356, -0.00000060, 0.00000220, -0.00000421, -0.00000270, -0.00001557, -0.00000073, 0.00000899, -0.00001726, 0.00001075, -0.00000013, -0.00001834, 0.00001432, -0.00000785, -0.00000856, -0.00000251, -0.00000005, -0.00000877, -0.00000325, -0.00001461, -0.00003817, 0.00003117, -0.00008051, -0.00001094, -0.00002599, -0.00037013, -0.00000879, 0.00042879, 0.00058380, 0.00033085, -0.00033571, -0.00017989, -0.00001923, -0.00005871, 0.00001156, -0.00008126, 0.00006612, -0.00006426, -0.00000579, 0.00004501, -0.00010477, 0.00009867, -0.00007701, 0.00000283, 0.00006138, -0.00010972, 0.00009956, -0.00004979, -0.00001854, 0.00005299, -0.00006732, 0.00005367, -0.00002574, 0.00000057, 0.00001374, -0.00004695, 0.00003689, -0.00001422, -0.00000611, 0.00001479, -0.00001515, 0.00000971, -0.00000398, -0.00001784, 0.00001001, -0.00000121, -0.00001992, 0.00002052, -0.00000178, -0.00000810, 0.00001044, -0.00002686, -0.00000578, 0.00001498, 0.00000002, -0.00000836, 0.00001087, 0.00001984, -0.00000902, -0.00000243, -0.00000091, -0.00000154, 0.00000503, -0.00000203, 0.00000926, 0.00000241, -0.00000638, 0.00000131, 0.00000147, 0.00000052, -0.00000156, 0.00000373, -0.00000020, -0.00000664, 0.00000277, -0.00000199, -0.00000697, 0.00000330, 0.00001106, 0.00000763, -0.00000532, -0.00000602, -0.00000229, 0.00000593, 0.00001686, 0.00000220, -0.00000547, -0.00000026, -0.00000746, 0.00000124, 0.00001050, -0.00000365, },
		{ 0.00000069, -0.00000789, -0.00000353, 0.00000029, -0.00000033, -0.00000476, -0.00000509, -0.00000625, 0.00000110, 0.00000451, -0.00000486, -0.00000376, 0.00000071, 0.00000657, -0.00000395, -0.00001613, -0.00000187, -0.00000061, -0.00000344, 0.00000309, -0.00000146, -0.00000412, -0.00001396, -0.00000474, 0.00001081, -0.00001466, 0.00000405, 0.00000698, -0.00001867, 0.00000826, -0.00000230, -0.00000793, -0.00000756, 0.00000281, -0.00000603, -0.00001155, 0.00000452, -0.00005438, 0.00002899, -0.00004297, -0.00006898, 0.00004783, -0.00030355, -0.00021604, 0.00031412, 0.00068413, 0.00048801, -0.00032060, -0.00023074, -0.00001212, -0.00009384, 0.00005068, -0.00011032, 0.00005401, -0.00001353, -0.00007356, 0.00010085, -0.00012552, 0.00007222, -0.00001343, -0.00007244, 0.00011336, -0.00011588, 0.00005917, 0.00001255, -0.00007132, 0.00007947, -0.00006293, 0.00002427, 0.00001143, -0.00002937, 0.00002996, -0.00004266, 0.00001489, 0.00000741, -0.00001796, 0.00001477, -0.00000824, -0.00000055, 0.00000637, -0.00001622, -0.00000238, 0.00000945, -0.00002237, 0.00001067, 0.00001236, -0.00001649, 0.00000850, -0.00001239, -0.00001685, 0.00000827, 0.00000962, -0.00000909, 0.00000122, 0.00002566, -0.00000368, -0.00000635, 0.00000053, -0.00000363, 0.00000605, -0.00000148, 0.00000649, 0.00000582, -0.00000650, -0.00000030, 0.00000213, 0.00000110, -0.00000145, 0.00000231, 0.00000153, -0.00000617, 0.00000062, -0.00000012, -0.00000597, 0.00000006, 0.00000886, 0.00001032, -0.00000314, -0.00000603, -0.00000193, 0.00000065, 0.00001556, 0.00000886, -0.00000616, -0.00000102, -0.00000481, -0.00000271, 0.00000968, },
		{ 0.00000660, 0.00000408, -0.00000614, -0.00000605, -0.00000050, 0.00000141, -0.00000471, -0.00000445, -0.00000645, -0.00000207, 0.00000571, -0.00000398, -0.00000402, 0.00000025, 0.00000452, 0.00000140, -0.00001499, -0.00000753, 0.00000141, -0.00000394, 0.00000104, 0.00000221, -0.00000400, -0.00001402, -0.00000724, 0.00001086, -0.00001110, -0.00000139, 0.00001031, -0.00001571, 0.00000323, -0.00000133, -0.00000374, -0.00001097, -0.00000085, 0.00000419, -0.00002429, 0.00001838, -0.00005198, 0.00000238, 0.00000632, -0.00010822, 0.00005660, -0.00017882, -0.00034272, 0.00009909, 0.00071636, 0.00071458, -0.00027743, -0.00031519, 0.00002379, -0.00014766, 0.00007896, -0.00010060, -0.00000604, 0.00006672, -0.00013792, 0.00012255, -0.00009616, 0.00000170, 0.00006797, -0.00013066, 0.00012182, -0.00007400, -0.00001069, 0.00007470, -0.00010019, 0.00007218, -0.00002892, -0.00001906, 0.00004348, -0.00004112, 0.00002405, -0.00002207, -0.00000963, 0.00001958, -0.00001703, 0.00000576, 0.00000128, -0.00000750, 0.00000780, -0.00000612, -0.00001430, 0.00001191, -0.00001542, -0.00000347, 0.00002202, -0.00001450, -0.00000191, 0.00000018, -0.00001710, -0.00000445, 0.00001451, -0.00000232, -0.00000974, 0.00002608, 0.00000543, -0.00000982, 0.00000107, -0.00000479, 0.00000538, 0.00000059, 0.00000352, 0.00000795, -0.00000515, -0.00000226, 0.00000232, 0.00000178, -0.00000101, 0.00000104, 0.00000236, -0.00000501, -0.00000106, 0.00000042, -0.00000442, -0.00000151, 0.00000548, 0.00001163, 0.00000009, -0.00000657, -0.00000067, -0.00000245, 0.00001071, 0.00001456, -0.00000364, -0.00000355, -0.00000220, -0.00000421, },
		{ -0.00000318, 0.00000319, 0.00000499, -0.00000274, -0.00000733, -0.00000304, 0.00000333, -0.00000415, -0.00000470, -0.00000502, -0.00000545, 0.00000554, -0.00000212, -0.00000510, 0.00000090, 0.00000226, 0.00000395, -0.00001041, -0.00001264, 0.00000046, -0.00000186, -0.00000225, 0.00000422, -0.00000107, -0.00001512, -0.00000966, 0.00001071, -0.00000808, -0.00000493, 0.00001131, -0.00001278, 0.00000161, -0.00000363, -0.00000077, -0.00000803, -0.00001165, 0.00001688, -0.00003199, 0.00001753, -0.00002989, -0.00003497, 0.00004233, -0.00010806, 0.00001327, -0.00005670, -0.00035381, -0.00015377, 0.00061594, 0.00098284, -0.00014997, -0.00045467, 0.00007726, -0.00018763, 0.00006067, -0.00003446, -0.00010302, 0.00014358, -0.00016235, 0.00008766, -0.00001817, -0.00008838, 0.00013141, -0.00014218, 0.00007754, 0.00000160, -0.00008158, 0.00010951, -0.00009088, 0.00003308, 0.00001966, -0.00005608, 0.00005462, -0.00002958, 0.00000221, 0.00000162, -0.00002426, 0.00001742, -0.00000657, -0.00000487, 0.00000641, -0.00000726, 0.00000177, 0.00000552, -0.00001874, 0.00000488, -0.00000401, -0.00001437, 0.00002234, -0.00000398, -0.00001303, 0.00000414, -0.00000802, -0.00001575, 0.00001042, 0.00000927, -0.00001679, 0.00001943, 0.00001611, -0.00001068, -0.00000007, -0.00000469, 0.00000329, 0.00000307, 0.00000149, 0.00000848, -0.00000287, -0.00000386, 0.00000184, 0.00000244, -0.00000030, 0.00000007, 0.00000252, -0.00000375, -0.00000202, 0.00000023, -0.00000334, -0.00000148, 0.00000223, 0.00001085, 0.00000399, -0.00000688, 0.00000005, -0.00000261, 0.00000424, 0.00001696, 0.00000191, -0.00000593, -0.00000146, },
		{ -0.00000309, -0.00000116, 0.00000152, 0.00000334, 0.00000042, -0.00000608, -0.00000681, 0.00000421, -0.00000217, -0.00000620, -0.00000277, -0.00000753, 0.00000329, 0.00000052, -0.00000636, 0.00000137, 0.00000142, 0.00000351, -0.00000472, -0.00001460, -0.00000376, 0.00000108, -0.00000422, 0.00000296, 0.00000387, -0.00001514, -0.00001357, 0.00001109, -0.00000554, -0.00000780, 0.00001233, -0.00001182, 0.00000248, -0.00000498, -0.00000347, 0.00000185, -0.00002461, 0.00002363, -0.00002713, 0.00000101, 0.00000141, -0.00006327, 0.00004823, -0.00007166, -0.00004596, 0.00001332, -0.00026459, -0.00035656, 0.00035939, 0.00121496, 0.00011832, -0.00063566, 0.00010364, -0.00016897, -0.00002685, 0.00007679, -0.00019883, 0.00017387, -0.00012092, -0.00000186, 0.00008123, -0.00015893, 0.00014440, -0.00009618, -0.00000471, 0.00007985, -0.00012295, 0.00010111, -0.00004631, -0.00002110, 0.00006120, -0.00007001, 0.00003978, -0.00000169, -0.00002193, 0.00001544, -0.00002341, 0.00000446, 0.00000510, -0.00000990, 0.00000415, -0.00000146, -0.00000608, 0.00001192, -0.00001347, -0.00000719, 0.00000466, -0.00001712, 0.00001416, 0.00000859, -0.00001771, -0.00000158, 0.00000383, -0.00001892, -0.00000200, 0.00001967, -0.00001590, 0.00000724, 0.00002454, -0.00000733, -0.00000264, -0.00000385, 0.00000069, 0.00000484, 0.00000088, 0.00000787, -0.00000049, -0.00000467, 0.00000084, 0.00000282, 0.00000065, -0.00000054, 0.00000229, -0.00000267, -0.00000244, -0.00000000, -0.00000308, -0.00000069, 0.00000030, 0.00000816, 0.00000747, -0.00000606, -0.00000052, -0.00000057, -0.00000132, 0.00001505, 0.00000863, -0.00000601, },
		{ -0.00000271, -0.00000584, -0.00000031, 0.00000245, 0.00000076, 0.00000128, -0.00000229, -0.00001009, 0.00000266, 0.00000143, -0.00000816, -0.00000128, -0.00000713, -0.00000064, 0.00000278, -0.00000673, 0.00000064, 0.00000231, 0.00000158, -0.00000062, -0.00001248, -0.00000942, 0.00000221, -0.00000318, -0.00000103, 0.00000826, -0.00001192, -0.00001925, 0.00001099, -0.00000201, -0.00001163, 0.00001451, -0.00001199, 0.00000229, -0.00000132, -0.00001257, 0.00001401, -0.00003171, 0.00001802, -0.00000946, -0.00002262, 0.00002603, -0.00006827, 0.00002468, -0.00002196, -0.00008330, 0.00001568, -0.00013263, -0.00043505, -0.00001240, 0.00130473, 0.00054618, -0.00079421, 0.00003836, -0.00006131, -0.00017184, 0.00018887, -0.00024199, 0.00012603, -0.00001569, -0.00011510, 0.00015869, -0.00017425, 0.00009463, -0.00000795, -0.00009138, 0.00012687, -0.00011667, 0.00005282, 0.00001462, -0.00006702, 0.00007804, -0.00005548, 0.00000656, 0.00002794, -0.00003497, 0.00001334, -0.00001045, -0.00001017, 0.00001003, -0.00000665, -0.00000308, 0.00000473, -0.00001004, 0.00001026, -0.00000185, -0.00001712, 0.00000550, -0.00001206, 0.00000294, 0.00001636, -0.00001335, -0.00001236, 0.00001078, -0.00001177, -0.00001722, 0.00002271, -0.00000667, -0.00000598, 0.00002714, 0.00000007, -0.00000516, -0.00000327, -0.00000152, 0.00000533, 0.00000135, 0.00000695, 0.00000141, -0.00000471, -0.00000024, 0.00000271, 0.00000168, -0.00000072, 0.00000182, -0.00000182, -0.00000269, 0.00000007, -0.00000332, -0.00000017, 0.00000011, 0.00000463, 0.00000936, -0.00000374, -0.00000208, 0.00000208, -0.00000412, 0.00000971, 0.00001383, },
		{ 0.00001537, 0.00000319, -0.00000742, -0.00000198, 0.00000499, -0.00000033, -0.00000107, 0.00000232, -0.00001072, -0.00000173, 0.00000553, -0.00000897, -0.00000194, -0.00000434, -0.00000463, 0.00000323, -0.00000562, -0.00000124, 0.00000392, 0.00000028, 0.00000040, -0.00000757, -0.00001350, -0.00000030, 0.00000043, -0.00000526, 0.00000932, -0.00000492, -0.00002490, 0.00000824, 0.00000390, -0.00001676, 0.00001663, -0.00001026, -0.00000245, 0.00000812, -0.00002384, 0.00002085, -0.00002686, 0.00000031, 0.00001318, -0.00004027, 0.00003214, -0.00004938, -0.00001156, 0.00001327, -0.00008022, -0.00002704, -0.00002889, -0.00036475, -0.00039912, 0.00116159, 0.00108320, -0.00082569, -0.00016977, 0.00012313, -0.00032135, 0.00023954, -0.00019317, 0.00000087, 0.00011770, -0.00020275, 0.00017451, -0.00012057, -0.00000177, 0.00008639, -0.00014520, 0.00012193, -0.00006581, -0.00001434, 0.00006613, -0.00008603, 0.00006467, -0.00002037, -0.00002939, 0.00004495, -0.00003059, -0.00000147, 0.00000517, -0.00001754, 0.00000524, 0.00000180, -0.00000964, 0.00000640, -0.00000790, 0.00000281, 0.00000958, -0.00001931, -0.00000154, -0.00000412, -0.00000495, 0.00001605, -0.00000291, -0.00002101, 0.00000835, 0.00000216, -0.00002770, 0.00001550, 0.00000697, -0.00001472, 0.00002250, 0.00000915, -0.00000565, -0.00000356, -0.00000304, 0.00000474, 0.00000214, 0.00000634, 0.00000279, -0.00000438, -0.00000106, 0.00000218, 0.00000252, -0.00000044, 0.00000125, -0.00000111, -0.00000298, 0.00000035, -0.00000344, -0.00000052, 0.00000118, 0.00000163, 0.00000907, -0.00000048, -0.00000365, 0.00000380, -0.00000373, 0.00000319, },
		{ -0.00000203, 0.00001280, 0.00000936, -0.00000588, -0.00000574, 0.00000677, 0.00000183, -0.00000550, 0.00000500, -0.00000746, -0.00000775, 0.00000806, -0.00000721, -0.00000485, -0.00000072, -0.00000668, 0.00000126, -0.00000361, -0.00000322, 0.00000472, 0.00000089, -0.00000112, -0.00000269, -0.00001361, -0.00000607, 0.00000408, -0.00000679, 0.00000591, 0.00000394, -0.00002740, 0.00000118, 0.00001184, -0.00002129, 0.00001577, -0.00000377, -0.00001234, 0.00001959, -0.00003025, 0.00001649, -0.00001023, -0.00002193, 0.00002918, -0.00004224, 0.00001889, -0.00001929, -0.00003853, 0.00001833, -0.00004608, -0.00007025, -0.00000208, -0.00018554, -0.00067607, 0.00075881, 0.00160482, -0.00062281, -0.00051905, 0.00031720, -0.00039726, 0.00017865, -0.00005043, -0.00016100, 0.00022178, -0.00021950, 0.00011334, -0.00001537, -0.00010570, 0.00014608, -0.00014278, 0.00006753, 0.00000681, -0.00007173, 0.00008793, -0.00007270, 0.00002942, 0.00001906, -0.00005263, 0.00004239, -0.00001212, -0.00001932, 0.00001382, -0.00001392, -0.00000587, 0.00000898, -0.00001063, 0.00000221, -0.00000185, -0.00000519, 0.00001521, -0.00001282, -0.00001177, 0.00000057, -0.00000610, 0.00000947, 0.00000746, -0.00002209, -0.00000238, 0.00001561, -0.00002798, 0.00000025, 0.00001855, -0.00001536, 0.00001254, 0.00001628, -0.00000289, -0.00000434, -0.00000425, 0.00000368, 0.00000260, 0.00000611, 0.00000403, -0.00000407, -0.00000156, 0.00000153, 0.00000292, 0.00000025, 0.00000068, -0.00000051, -0.00000327, 0.00000048, -0.00000300, -0.00000168, 0.00000248, 0.00000017, 0.00000697, 0.00000260, -0.00000423, 0.00000396, -0.00000117, },
		{ 0.00000193, -0.00000442, 0.00000746, 0.00001313, -0.00000091, -0.00000941, 0.00000554, 0.00000700, -0.00000918, 0.00000352, -0.00000110, -0.00001273, 0.00000714, -0.00000282, -0.00000858, 0.00000151, -0.00000563, -0.00000226, -0.00000225, -0.00000407, 0.00000394, 0.00000311, -0.00000309, -0.00000049, -0.00000949, -0.00001243, 0.00000469, -0.00000424, -0.00000044, 0.00001101, -0.00002422, -0.00000969, 0.00001920, -0.00002207, 0.00000949, 0.00000769, -0.00002393, 0.00002663, -0.00002612, 0.00000079, 0.00001111, -0.00003776, 0.00003019, -0.00002822, -0.00000348, 0.00000435, -0.00004269, -0.00000169, -0.00001002, -0.00007466, -0.00005464, 0.00001423, -0.00074622, 0.00016193, 0.00194344, -0.00012949, -0.00093020, 0.00041434, -0.00033233, -0.00000039, 0.00014034, -0.00029195, 0.00024283, -0.00014760, -0.00000526, 0.00009842, -0.00017253, 0.00014497, -0.00008585, -0.00001189, 0.00006964, -0.00009682, 0.00007438, -0.00003636, -0.00001077, 0.00004670, -0.00005528, 0.00002337, 0.00001033, -0.00003013, 0.00001105, -0.00000237, -0.00001610, 0.00000944, -0.00000517, -0.00000494, 0.00000350, -0.00000911, 0.00001339, -0.00000145, -0.00001889, -0.00000154, -0.00000183, 0.00000182, 0.00001226, -0.00001491, -0.00001574, 0.00002166, -0.00001760, -0.00001664, 0.00002230, -0.00000815, 0.00000171, 0.00001845, 0.00000256, -0.00000430, -0.00000567, 0.00000253, 0.00000261, 0.00000585, 0.00000556, -0.00000379, -0.00000202, 0.00000110, 0.00000280, 0.00000114, 0.00000029, -0.00000008, -0.00000338, 0.00000018, -0.00000198, -0.00000302, 0.00000306, 0.00000042, 0.00000419, 0.00000442, -0.00000355, 0.00000294, },
		{ 0.00000164, 0.00000424, -0.00000396, 0.00000174, 0.00001296, 0.00000584, -0.00001032, 0.00000065, 0.00001283, -0.00000909, -0.00000232, 0.00000556, -0.00001396, 0.00000244, 0.00000254, -0.00001086, 0.00000084, -0.00000208, -0.00000520, -0.00000297, -0.00000353, 0.00000211, 0.00000538, -0.00000333, -0.00000180, -0.00000338, -0.00001592, 0.00000070, 0.00000108, -0.00000626, 0.00001305, -0.00001533, -0.00002120, 0.00002222, -0.00001675, -0.00000204, 0.00002075, -0.00003137, 0.00002391, -0.00001098, -0.00001986, 0.00002681, -0.00003926, 0.00001630, -0.00000730, -0.00002052, 0.00001021, -0.00002684, -0.00002523, -0.00000135, -0.00002970, -0.00014279, 0.00013795, -0.00058557, -0.00048004, 0.00194866, 0.00061339, -0.00125467, 0.00031176, -0.00011088, -0.00024532, 0.00029865, -0.00032676, 0.00015856, -0.00000929, -0.00013303, 0.00017152, -0.00017321, 0.00008408, -0.00000036, -0.00008186, 0.00009752, -0.00008325, 0.00003596, 0.00000479, -0.00003940, 0.00005419, -0.00003981, -0.00000181, 0.00002599, -0.00002890, -0.00000057, 0.00000968, -0.00001909, 0.00000216, 0.00000314, -0.00001016, 0.00000443, -0.00000767, 0.00000671, 0.00000889, -0.00001870, -0.00000939, 0.00000310, -0.00000181, 0.00000987, -0.00000346, -0.00002484, 0.00001737, -0.00000131, -0.00002786, 0.00001624, 0.00000279, -0.00000515, 0.00001500, 0.00000834, -0.00000206, -0.00000728, 0.00000119, 0.00000252, 0.00000504, 0.00000736, -0.00000316, -0.00000282, 0.00000103, 0.00000237, 0.00000190, 0.00000023, 0.00000013, -0.00000320, -0.00000055, -0.00000079, -0.00000387, 0.00000252, 0.00000171, 0.00000195, 0.00000457, -0.00000208, },
		{ -0.00000076, 0.00000080, 0.00000533, -0.00000179, -0.00000224, 0.00000915, 0.00001151, -0.00000690, -0.00000625, 0.00001598, -0.00000381, -0.00001016, 0.00000897, -0.00001029, -0.00000423, 0.00000623, -0.00001009, -0.00000248, 0.00000179, -0.00000546, -0.00000593, -0.00000271, 0.00000070, 0.00000610, -0.00000105, -0.00000517, 0.00000134, -0.00001433, -0.00000672, 0.00000578, -0.00000815, 0.00000922, -0.00000374, -0.00002887, 0.00001825, -0.00000582, -0.00001535, 0.00002994, -0.00002974, 0.00001070, 0.00000957, -0.00003577, 0.00002916, -0.00002640, -0.00000387, 0.00000787, -0.00002296, 0.00000017, -0.00000756, -0.00002896, -0.00003005, 0.00004007, -0.00020120, 0.00012658, -0.00025934, -0.00098238, 0.00155322, 0.00145557, -0.00132060, -0.00003508, 0.00020729, -0.00045789, 0.00034356, -0.00023577, -0.00000663, 0.00013987, -0.00021437, 0.00017161, -0.00010829, -0.00000892, 0.00007622, -0.00011434, 0.00008298, -0.00004246, -0.00000735, 0.00003345, -0.00004822, 0.00004364, -0.00001613, -0.00002188, 0.00002918, -0.00001783, -0.00001382, 0.00001526, -0.00001309, -0.00000890, 0.00000869, -0.00000994, 0.00000059, -0.00000327, -0.00000011, 0.00001375, -0.00001135, -0.00001828, 0.00000375, 0.00000039, 0.00000308, 0.00000624, -0.00002546, 0.00000513, 0.00001345, -0.00002890, 0.00000318, 0.00001154, -0.00000541, 0.00000815, 0.00001144, 0.00000268, -0.00000816, -0.00000084, 0.00000270, 0.00000363, 0.00000892, -0.00000165, -0.00000405, 0.00000108, 0.00000202, 0.00000228, 0.00000050, 0.00000016, -0.00000275, -0.00000146, 0.00000009, -0.00000387, 0.00000114, 0.00000305, 0.00000102, 0.00000350, },
		{ 0.00000213, -0.00000025, 0.00000050, 0.00000559, 0.00000068, -0.00000357, 0.00000367, 0.00001357, 0.00000025, -0.00001191, 0.00001402, 0.00000534, -0.00001603, 0.00000681, -0.00000316, -0.00000966, 0.00000622, -0.00000646, -0.00000644, 0.00000349, -0.00000237, -0.00000949, -0.00000351, 0.00000083, 0.00000492, 0.00000265, -0.00000786, 0.00000211, -0.00000824, -0.00001402, 0.00000634, -0.00000498, 0.00000192, 0.00000584, -0.00002934, 0.00000776, 0.00000704, -0.00002523, 0.00003082, -0.00001825, -0.00000824, 0.00002666, -0.00003942, 0.00001767, -0.00000708, -0.00001857, 0.00000939, -0.00001275, -0.00001239, -0.00000363, -0.00000385, -0.00007912, 0.00008918, -0.00017935, -0.00001243, 0.00010093, -0.00119384, 0.00081491, 0.00217129, -0.00100261, -0.00057206, 0.00049633, -0.00053350, 0.00023406, -0.00004451, -0.00018829, 0.00023509, -0.00021261, 0.00009914, -0.00000765, -0.00009332, 0.00011283, -0.00010038, 0.00003819, 0.00000358, -0.00003684, 0.00004144, -0.00003977, 0.00002459, 0.00000467, -0.00003040, 0.00002143, -0.00000399, -0.00002154, 0.00001189, -0.00000173, -0.00001741, 0.00000756, -0.00000451, -0.00000502, 0.00000015, -0.00000346, 0.00001234, -0.00000080, -0.00002290, -0.00000186, 0.00000612, -0.00000303, 0.00000973, -0.00001812, -0.00000893, 0.00002052, -0.00002021, -0.00001075, 0.00001353, -0.00000007, 0.00000195, 0.00001000, 0.00000849, -0.00000689, -0.00000374, 0.00000303, 0.00000206, 0.00000952, 0.00000093, -0.00000531, 0.00000075, 0.00000204, 0.00000228, 0.00000095, 0.00000012, -0.00000218, -0.00000224, 0.00000035, -0.00000312, -0.00000041, 0.00000367, 0.00000134, },
		{ 0.00000228, 0.00000136, -0.00000062, 0.00000028, 0.00000566, 0.00000256, -0.00000275, -0.00000104, 0.00001118, 0.00000845, -0.00001320, 0.00000700, 0.00001471, -0.00001643, -0.00000047, 0.00000387, -0.00001097, 0.00000253, -0.00000212, -0.00000854, 0.00000186, 0.00000263, -0.00001100, -0.00000717, 0.00000220, 0.00000308, 0.00000561, -0.00000765, -0.00000120, -0.00000077, -0.00001728, 0.00000148, 0.00000114, -0.00000447, 0.00000954, -0.00002256, -0.00000536, 0.00001659, -0.00002768, 0.00002270, -0.00000100, -0.00002501, 0.00003307, -0.00002975, -0.00000062, 0.00000763, -0.00002038, -0.00000059, -0.00000160, -0.00001211, -0.00002269, 0.00003754, -0.00011361, 0.00007837, -0.00007053, -0.00020798, 0.00035120, -0.00105831, -0.00008806, 0.00253257, -0.00028683, -0.00114375, 0.00060936, -0.00041189, -0.00000450, 0.00017196, -0.00030927, 0.00023278, -0.00012818, -0.00001272, 0.00008415, -0.00013493, 0.00009894, -0.00005265, -0.00001281, 0.00003423, -0.00004396, 0.00003213, -0.00002391, 0.00000754, 0.00001615, -0.00002813, 0.00000945, 0.00000537, -0.00002087, 0.00000253, 0.00000853, -0.00001863, -0.00000002, 0.00000235, -0.00000828, -0.00000009, -0.00000279, 0.00000738, 0.00000783, -0.00002047, -0.00001164, 0.00001067, -0.00000438, 0.00000654, -0.00000730, -0.00001848, 0.00001798, -0.00000672, -0.00001936, 0.00000793, 0.00000680, 0.00000000, 0.00000459, 0.00001271, -0.00000255, -0.00000688, 0.00000289, 0.00000103, 0.00000873, 0.00000418, -0.00000588, -0.00000030, 0.00000241, 0.00000213, 0.00000134, 0.00000013, -0.00000164, -0.00000266, 0.00000000, -0.00000209, -0.00000151, 0.00000334, },
		{ 0.00000240, 0.00000306, 0.00000149, -0.00000131, -0.00000047, 0.00000582, 0.00000384, -0.00000108, -0.00000330, 0.00000555, 0.00001426, -0.00000886, -0.00000235, 0.00002015, -0.00001032, -0.00000955, 0.00000692, -0.00000749, -0.00000248, 0.00000011, -0.00000745, -0.00000207, 0.00000662, -0.00000844, -0.00001302, 0.00000291, 0.00000246, 0.00000623, -0.00000431, -0.00000625, 0.00000415, -0.00001461, -0.00000680, 0.00000601, -0.00000617, 0.00000672, -0.00001212, -0.00001567, 0.00001906, -0.00002231, 0.00000920, 0.00001519, -0.00003303, 0.00002725, -0.00001272, -0.00001549, 0.00001053, -0.00001113, -0.00001122, -0.00000275, 0.00000705, -0.00005464, 0.00006732, -0.00010147, -0.00000065, 0.00008208, -0.00035932, 0.00039595, -0.00064055, -0.00091152, 0.00239480, 0.00069999, -0.00153713, 0.00044317, -0.00011456, -0.00028358, 0.00032002, -0.00031555, 0.00013218, 0.00000102, -0.00011426, 0.00012942, -0.00012164, 0.00004767, 0.00000281, -0.00004713, 0.00004010, -0.00003291, 0.00001650, -0.00001101, -0.00000183, 0.00001876, -0.00002091, 0.00000032, 0.00000709, -0.00001437, -0.00000674, 0.00001246, -0.00001217, -0.00000979, 0.00000584, -0.00000682, -0.00000361, -0.00000045, 0.00000283, 0.00001130, -0.00001214, -0.00002084, 0.00001002, -0.00000026, -0.00000006, 0.00000140, -0.00002025, 0.00000866, 0.00000499, -0.00001958, -0.00000203, 0.00001039, 0.00000341, -0.00000185, 0.00001276, 0.00000431, -0.00000877, 0.00000154, 0.00000091, 0.00000676, 0.00000718, -0.00000514, -0.00000201, 0.00000275, 0.00000216, 0.00000155, 0.00000021, -0.00000122, -0.00000267, -0.00000069, -0.00000124, -0.00000190, },
		{ -0.00000171, 0.00000139, 0.00000325, 0.00000223, -0.00000156, -0.00000191, 0.00000578, 0.00000502, 0.00000028, -0.00000293, -0.00000069, 0.00001525, -0.00000033, -0.00000991, 0.00001919, 0.00000012, -0.00001586, 0.00000410, -0.00000145, -0.00000535, -0.00000134, -0.00000418, -0.00000559, 0.00000704, -0.00000194, -0.00001841, 0.00000062, 0.00000406, 0.00000468, 0.00000031, -0.00000974, 0.00000417, -0.00000738, -0.00001418, 0.00000588, -0.00000243, 0.00000041, -0.00000339, -0.00001926, 0.00001437, -0.00001269, -0.00000378, 0.00002461, -0.00003056, 0.00001396, 0.00000252, -0.00001989, 0.00000283, -0.00000078, -0.00001031, -0.00002111, 0.00003602, -0.00007639, 0.00005896, -0.00003427, -0.00011694, 0.00020602, -0.00038477, 0.00022855, -0.00010308, -0.00143034, 0.00175909, 0.00171550, -0.00155907, -0.00000573, 0.00025177, -0.00048612, 0.00033050, -0.00020304, -0.00002329, 0.00011774, -0.00016328, 0.00011462, -0.00006721, -0.00001294, 0.00004076, -0.00005406, 0.00002590, -0.00001607, 0.00000548, -0.00000594, -0.00000465, 0.00001770, -0.00001507, -0.00000267, 0.00000335, -0.00000773, -0.00001080, 0.00000904, -0.00000205, -0.00001618, 0.00000326, -0.00000167, -0.00000754, 0.00000041, 0.00000126, 0.00000969, -0.00000198, -0.00002496, 0.00000330, 0.00000645, -0.00000534, 0.00000443, -0.00001531, -0.00000183, 0.00001019, -0.00001276, -0.00001096, 0.00000783, 0.00001005, -0.00000548, 0.00000781, 0.00001146, -0.00000775, -0.00000114, 0.00000138, 0.00000440, 0.00000900, -0.00000297, -0.00000381, 0.00000259, 0.00000250, 0.00000166, 0.00000028, -0.00000095, -0.00000237, -0.00000142, -0.00000086, },
		{ -0.00000095, -0.00000130, 0.00000069, 0.00000289, 0.00000297, -0.00000097, -0.00000363, 0.00000495, 0.00000650, 0.00000098, -0.00000106, -0.00000499, 0.00001119, 0.00000890, -0.00001224, 0.00001244, 0.00001043, -0.00001601, -0.00000322, 0.00000312, -0.00000372, -0.00000551, -0.00000157, -0.00000604, 0.00000338, 0.00000586, -0.00001999, -0.00000570, 0.00000690, 0.00000285, 0.00000358, -0.00000955, -0.00000011, 0.00000045, -0.00001652, -0.00000006, 0.00000388, -0.00000442, -0.00000038, -0.00001602, 0.00000633, -0.00000418, -0.00001163, 0.00002582, -0.00002136, 0.00000101, 0.00000930, -0.00001435, -0.00000657, -0.00000073, 0.00000673, -0.00004813, 0.00005464, -0.00006574, 0.00000487, 0.00006439, -0.00021382, 0.00023638, -0.00026160, -0.00006779, 0.00035745, -0.00151991, 0.00078277, 0.00247698, -0.00111732, -0.00062153, 0.00053207, -0.00051899, 0.00019279, -0.00002107, -0.00016628, 0.00017290, -0.00014471, 0.00005482, -0.00000203, -0.00005522, 0.00004893, -0.00003845, 0.00000555, -0.00000563, 0.00000368, -0.00000638, -0.00000635, 0.00001827, -0.00001334, -0.00000133, -0.00000048, -0.00000588, -0.00000856, 0.00000175, 0.00000597, -0.00001569, -0.00000426, 0.00000347, -0.00000860, -0.00000181, 0.00000244, 0.00000581, 0.00000560, -0.00002226, -0.00000682, 0.00001133, -0.00000605, 0.00000197, -0.00000789, -0.00000842, 0.00000814, -0.00000361, -0.00001434, -0.00000017, 0.00001566, -0.00000368, -0.00000046, 0.00001581, -0.00000315, -0.00000423, 0.00000159, 0.00000252, 0.00000922, 0.00000004, -0.00000493, 0.00000170, 0.00000299, 0.00000187, 0.00000032, -0.00000084, -0.00000188, -0.00000193, },
		{ -0.00000211, -0.00000136, -0.00000096, 0.00000039, 0.00000231, 0.00000325, 0.00000031, -0.00000484, 0.00000296, 0.00000814, 0.00000149, 0.00000078, -0.00000604, 0.00000403, 0.00001504, -0.00000842, 0.00000333, 0.00001627, -0.00000973, -0.00001092, 0.00000264, 0.00000194, -0.00000910, -0.00000253, -0.00000281, -0.00000220, 0.00001120, -0.00001581, -0.00001454, 0.00000832, 0.00000298, 0.00000404, -0.00000606, -0.00000556, 0.00000470, -0.00001243, -0.00000893, 0.00000793, -0.00000400, -0.00000334, -0.00000977, 0.00000024, -0.00000049, -0.00001371, 0.00002200, -0.00001179, -0.00000582, 0.00000760, -0.00000647, -0.00000718, -0.00001554, 0.00003213, -0.00006536, 0.00004366, -0.00001645, -0.00007584, 0.00014850, -0.00023757, 0.00014995, -0.00003731, -0.00035872, 0.00058715, -0.00119982, -0.00026967, 0.00275934, -0.00026967, -0.00119982, 0.00058715, -0.00035872, -0.00003731, 0.00014995, -0.00023757, 0.00014850, -0.00007584, -0.00001645, 0.00004366, -0.00006536, 0.00003213, -0.00001554, -0.00000718, -0.00000647, 0.00000760, -0.00000582, -0.00001179, 0.00002200, -0.00001371, -0.00000049, 0.00000024, -0.00000977, -0.00000334, -0.00000400, 0.00000793, -0.00000893, -0.00001243, 0.00000470, -0.00000556, -0.00000606, 0.00000404, 0.00000298, 0.00000832, -0.00001454, -0.00001581, 0.00001120, -0.00000220, -0.00000281, -0.00000253, -0.00000910, 0.00000194, 0.00000264, -0.00001092, -0.00000973, 0.00001627, 0.00000333, -0.00000842, 0.00001504, 0.00000403, -0.00000604, 0.00000078, 0.00000149, 0.00000814, 0.00000296, -0.00000484, 0.00000031, 0.00000325, 0.00000231, 0.00000039, -0.00000096, -0.00000136, },
		{ -0.00000095, -0.00000193, -0.00000188, -0.00000084, 0.00000032, 0.00000187, 0.00000299, 0.00000170, -0.00000493, 0.00000004, 0.00000922, 0.00000252, 0.00000159, -0.00000423, -0.00000315, 0.00001581, -0.00000046, -0.00000368, 0.00001566, -0.00000017, -0.00001434, -0.00000361, 0.00000814, -0.00000842, -0.00000789, 0.00000197, -0.00000605, 0.00001133, -0.00000682, -0.00002226, 0.00000560, 0.00000581, 0.00000244, -0.00000181, -0.00000860, 0.00000347, -0.00000426, -0.00001569, 0.00000597, 0.00000175, -0.00000856, -0.00000588, -0.00000048, -0.00000133, -0.00001334, 0.00001827, -0.00000635, -0.00000638, 0.00000368, -0.00000563, 0.00000555, -0.00003845, 0.00004893, -0.00005522, -0.00000203, 0.00005482, -0.00014471, 0.00017290, -0.00016628, -0.00002107, 0.00019279, -0.00051899, 0.00053207, -0.00062153, -0.00111732, 0.00247698, 0.00078277, -0.00151991, 0.00035745, -0.00006779, -0.00026160, 0.00023638, -0.00021382, 0.00006439, 0.00000487, -0.00006574, 0.00005464, -0.00004813, 0.00000673, -0.00000073, -0.00000657, -0.00001435, 0.00000930, 0.00000101, -0.00002136, 0.00002582, -0.00001163, -0.00000418, 0.00000633, -0.00001602, -0.00000038, -0.00000442, 0.00000388, -0.00000006, -0.00001652, 0.00000045, -0.00000011, -0.00000955, 0.00000358, 0.00000285, 0.00000690, -0.00000570, -0.00001999, 0.00000586, 0.00000338, -0.00000604, -0.00000157, -0.00000551, -0.00000372, 0.00000312, -0.00000322, -0.00001601, 0.00001043, 0.00001244, -0.00001224, 0.00000890, 0.00001119, -0.00000499, -0.00000106, 0.00000098, 0.00000650, 0.00000495, -0.00000363, -0.00000097, 0.00000297, 0.00000289, 0.00000069, -0.00000130, },
		{ -0.00000171, -0.00000086, -0.00000142, -0.00000237, -0.00000095, 0.00000028, 0.00000166, 0.00000250, 0.00000259, -0.00000381, -0.00000297, 0.00000900, 0.00000440, 0.00000138, -0.00000114, -0.00000775, 0.00001146, 0.00000781, -0.00000548, 0.00001005, 0.00000783, -0.00001096, -0.00001276, 0.00001019, -0.00000183, -0.00001531, 0.00000443, -0.00000534, 0.00000645, 0.00000330, -0.00002496, -0.00000198, 0.00000969, 0.00000126, 0.00000041, -0.00000754, -0.00000167, 0.00000326, -0.00001618, -0.00000205, 0.00000904, -0.00001080, -0.00000773, 0.00000335, -0.00000267, -0.00001507, 0.00001770, -0.00000465, -0.00000594, 0.00000548, -0.00001607, 0.00002590, -0.00005406, 0.00004076, -0.00001294, -0.00006721, 0.00011462, -0.00016328, 0.00011774, -0.00002329, -0.00020304, 0.00033050, -0.00048612, 0.00025177, -0.00000573, -0.00155907, 0.00171550, 0.00175909, -0.00143034, -0.00010308, 0.00022855, -0.00038477, 0.00020602, -0.00011694, -0.00003427, 0.00005896, -0.00007639, 0.00003602, -0.00002111, -0.00001031, -0.00000078, 0.00000283, -0.00001989, 0.00000252, 0.00001396, -0.00003056, 0.00002461, -0.00000378, -0.00001269, 0.00001437, -0.00001926, -0.00000339, 0.00000041, -0.00000243, 0.00000588, -0.00001418, -0.00000738, 0.00000417, -0.00000974, 0.00000031, 0.00000468, 0.00000406, 0.00000062, -0.00001841, -0.00000194, 0.00000704, -0.00000559, -0.00000418, -0.00000134, -0.00000535, -0.00000145, 0.00000410, -0.00001586, 0.00000012, 0.00001919, -0.00000991, -0.00000033, 0.00001525, -0.00000069, -0.00000293, 0.00000028, 0.00000502, 0.00000578, -0.00000191, -0.00000156, 0.00000223, 0.00000325, 0.00000139, },
		{ 0.00000240, -0.00000190, -0.00000124, -0.00000069, -0.00000267, -0.00000122, 0.00000021, 0.00000155, 0.00000216, 0.00000275, -0.00000201, -0.00000514, 0.00000718, 0.00000676, 0.00000091, 0.00000154, -0.00000877, 0.00000431, 0.00001276, -0.00000185, 0.00000341, 0.00001039, -0.00000203, -0.00001958, 0.00000499, 0.00000866, -0.00002025, 0.00000140, -0.00000006, -0.00000026, 0.00001002, -0.00002084, -0.00001214, 0.00001130, 0.00000283, -0.00000045, -0.00000361, -0.00000682, 0.00000584, -0.00000979, -0.00001217, 0.00001246, -0.00000674, -0.00001437, 0.00000709, 0.00000032, -0.00002091, 0.00001876, -0.00000183, -0.00001101, 0.00001650, -0.00003291, 0.00004010, -0.00004713, 0.00000281, 0.00004767, -0.00012164, 0.00012942, -0.00011426, 0.00000102, 0.00013218, -0.00031555, 0.00032002, -0.00028358, -0.00011456, 0.00044317, -0.00153713, 0.00069999, 0.00239480, -0.00091152, -0.00064055, 0.00039595, -0.00035932, 0.00008208, -0.00000065, -0.00010147, 0.00006732, -0.00005464, 0.00000705, -0.00000275, -0.00001122, -0.00001113, 0.00001053, -0.00001549, -0.00001272, 0.00002725, -0.00003303, 0.00001519, 0.00000920, -0.00002231, 0.00001906, -0.00001567, -0.00001212, 0.00000672, -0.00000617, 0.00000601, -0.00000680, -0.00001461, 0.00000415, -0.00000625, -0.00000431, 0.00000623, 0.00000246, 0.00000291, -0.00001302, -0.00000844, 0.00000662, -0.00000207, -0.00000745, 0.00000011, -0.00000248, -0.00000749, 0.00000692, -0.00000955, -0.00001032, 0.00002015, -0.00000235, -0.00000886, 0.00001426, 0.00000555, -0.00000330, -0.00000108, 0.00000384, 0.00000582, -0.00000047, -0.00000131, 0.00000149, 0.00000306, },
		{ 0.00000228, 0.00000334, -0.00000151, -0.00000209, 0.00000000, -0.00000266, -0.00000164, 0.00000013, 0.00000134, 0.00000213, 0.00000241, -0.00000030, -0.00000588, 0.00000418, 0.00000873, 0.00000103, 0.00000289, -0.00000688, -0.00000255, 0.00001271, 0.00000459, 0.00000000, 0.00000680, 0.00000793, -0.00001936, -0.00000672, 0.00001798, -0.00001848, -0.00000730, 0.00000654, -0.00000438, 0.00001067, -0.00001164, -0.00002047, 0.00000783, 0.00000738, -0.00000279, -0.00000009, -0.00000828, 0.00000235, -0.00000002, -0.00001863, 0.00000853, 0.00000253, -0.00002087, 0.00000537, 0.00000945, -0.00002813, 0.00001615, 0.00000754, -0.00002391, 0.00003213, -0.00004396, 0.00003423, -0.00001281, -0.00005265, 0.00009894, -0.00013493, 0.00008415, -0.00001272, -0.00012818, 0.00023278, -0.00030927, 0.00017196, -0.00000450, -0.00041189, 0.00060936, -0.00114375, -0.00028683, 0.00253257, -0.00008806, -0.00105831, 0.00035120, -0.00020798, -0.00007053, 0.00007837, -0.00011361, 0.00003754, -0.00002269, -0.00001211, -0.00000160, -0.00000059, -0.00002038, 0.00000763, -0.00000062, -0.00002975, 0.00003307, -0.00002501, -0.00000100, 0.00002270, -0.00002768, 0.00001659, -0.00000536, -0.00002256, 0.00000954, -0.00000447, 0.00000114, 0.00000148, -0.00001728, -0.00000077, -0.00000120, -0.00000765, 0.00000561, 0.00000308, 0.00000220, -0.00000717, -0.00001100, 0.00000263, 0.00000186, -0.00000854, -0.00000212, 0.00000253, -0.00001097, 0.00000387, -0.00000047, -0.00001643, 0.00001471, 0.00000700, -0.00001320, 0.00000845, 0.00001118, -0.00000104, -0.00000275, 0.00000256, 0.00000566, 0.00000028, -0.00000062, 0.00000136, },
		{ 0.00000213, 0.00000134, 0.00000367, -0.00000041, -0.00000312, 0.00000035, -0.00000224, -0.00000218, 0.00000012, 0.00000095, 0.00000228, 0.00000204, 0.00000075, -0.00000531, 0.00000093, 0.00000952, 0.00000206, 0.00000303, -0.00000374, -0.00000689, 0.00000849, 0.00001000, 0.00000195, -0.00000007, 0.00001353, -0.00001075, -0.00002021, 0.00002052, -0.00000893, -0.00001812, 0.00000973, -0.00000303, 0.00000612, -0.00000186, -0.00002290, -0.00000080, 0.00001234, -0.00000346, 0.00000015, -0.00000502, -0.00000451, 0.00000756, -0.00001741, -0.00000173, 0.00001189, -0.00002154, -0.00000399, 0.00002143, -0.00003040, 0.00000467, 0.00002459, -0.00003977, 0.00004144, -0.00003684, 0.00000358, 0.00003819, -0.00010038, 0.00011283, -0.00009332, -0.00000765, 0.00009914, -0.00021261, 0.00023509, -0.00018829, -0.00004451, 0.00023406, -0.00053350, 0.00049633, -0.00057206, -0.00100261, 0.00217129, 0.00081491, -0.00119384, 0.00010093, -0.00001243, -0.00017935, 0.00008918, -0.00007912, -0.00000385, -0.00000363, -0.00001239, -0.00001275, 0.00000939, -0.00001857, -0.00000708, 0.00001767, -0.00003942, 0.00002666, -0.00000824, -0.00001825, 0.00003082, -0.00002523, 0.00000704, 0.00000776, -0.00002934, 0.00000584, 0.00000192, -0.00000498, 0.00000634, -0.00001402, -0.00000824, 0.00000211, -0.00000786, 0.00000265, 0.00000492, 0.00000083, -0.00000351, -0.00000949, -0.00000237, 0.00000349, -0.00000644, -0.00000646, 0.00000622, -0.00000966, -0.00000316, 0.00000681, -0.00001603, 0.00000534, 0.00001402, -0.00001191, 0.00000025, 0.00001357, 0.00000367, -0.00000357, 0.00000068, 0.00000559, 0.00000050, -0.00000025, },
		{ -0.00000076, 0.00000350, 0.00000102, 0.00000305, 0.00000114, -0.00000387, 0.00000009, -0.00000146, -0.00000275, 0.00000016, 0.00000050, 0.00000228, 0.00000202, 0.00000108, -0.00000405, -0.00000165, 0.00000892, 0.00000363, 0.00000270, -0.00000084, -0.00000816, 0.00000268, 0.00001144, 0.00000815, -0.00000541, 0.00001154, 0.00000318, -0.00002890, 0.00001345, 0.00000513, -0.00002546, 0.00000624, 0.00000308, 0.00000039, 0.00000375, -0.00001828, -0.00001135, 0.00001375, -0.00000011, -0.00000327, 0.00000059, -0.00000994, 0.00000869, -0.00000890, -0.00001309, 0.00001526, -0.00001382, -0.00001783, 0.00002918, -0.00002188, -0.00001613, 0.00004364, -0.00004822, 0.00003345, -0.00000735, -0.00004246, 0.00008298, -0.00011434, 0.00007622, -0.00000892, -0.00010829, 0.00017161, -0.00021437, 0.00013987, -0.00000663, -0.00023577, 0.00034356, -0.00045789, 0.00020729, -0.00003508, -0.00132060, 0.00145557, 0.00155322, -0.00098238, -0.00025934, 0.00012658, -0.00020120, 0.00004007, -0.00003005, -0.00002896, -0.00000756, 0.00000017, -0.00002296, 0.00000787, -0.00000387, -0.00002640, 0.00002916, -0.00003577, 0.00000957, 0.00001070, -0.00002974, 0.00002994, -0.00001535, -0.00000582, 0.00001825, -0.00002887, -0.00000374, 0.00000922, -0.00000815, 0.00000578, -0.00000672, -0.00001433, 0.00000134, -0.00000517, -0.00000105, 0.00000610, 0.00000070, -0.00000271, -0.00000593, -0.00000546, 0.00000179, -0.00000248, -0.00001009, 0.00000623, -0.00000423, -0.00001029, 0.00000897, -0.00001016, -0.00000381, 0.00001598, -0.00000625, -0.00000690, 0.00001151, 0.00000915, -0.00000224, -0.00000179, 0.00000533, 0.00000080, },
		{ 0.00000164, -0.00000208, 0.00000457, 0.00000195, 0.00000171, 0.00000252, -0.00000387, -0.00000079, -0.00000055, -0.00000320, 0.00000013, 0.00000023, 0.00000190, 0.00000237, 0.00000103, -0.00000282, -0.00000316, 0.00000736, 0.00000504, 0.00000252, 0.00000119, -0.00000728, -0.00000206, 0.00000834, 0.00001500, -0.00000515, 0.00000279, 0.00001624, -0.00002786, -0.00000131, 0.00001737, -0.00002484, -0.00000346, 0.00000987, -0.00000181, 0.00000310, -0.00000939, -0.00001870, 0.00000889, 0.00000671, -0.00000767, 0.00000443, -0.00001016, 0.00000314, 0.00000216, -0.00001909, 0.00000968, -0.00000057, -0.00002890, 0.00002599, -0.00000181, -0.00003981, 0.00005419, -0.00003940, 0.00000479, 0.00003596, -0.00008325, 0.00009752, -0.00008186, -0.00000036, 0.00008408, -0.00017321, 0.00017152, -0.00013303, -0.00000929, 0.00015856, -0.00032676, 0.00029865, -0.00024532, -0.00011088, 0.00031176, -0.00125467, 0.00061339, 0.00194866, -0.00048004, -0.00058557, 0.00013795, -0.00014279, -0.00002970, -0.00000135, -0.00002523, -0.00002684, 0.00001021, -0.00002052, -0.00000730, 0.00001630, -0.00003926, 0.00002681, -0.00001986, -0.00001098, 0.00002391, -0.00003137, 0.00002075, -0.00000204, -0.00001675, 0.00002222, -0.00002120, -0.00001533, 0.00001305, -0.00000626, 0.00000108, 0.00000070, -0.00001592, -0.00000338, -0.00000180, -0.00000333, 0.00000538, 0.00000211, -0.00000353, -0.00000297, -0.00000520, -0.00000208, 0.00000084, -0.00001086, 0.00000254, 0.00000244, -0.00001396, 0.00000556, -0.00000232, -0.00000909, 0.00001283, 0.00000065, -0.00001032, 0.00000584, 0.00001296, 0.00000174, -0.00000396, 0.00000424, },
		{ 0.00000193, 0.00000294, -0.00000355, 0.00000442, 0.00000419, 0.00000042, 0.00000306, -0.00000302, -0.00000198, 0.00000018, -0.00000338, -0.00000008, 0.00000029, 0.00000114, 0.00000280, 0.00000110, -0.00000202, -0.00000379, 0.00000556, 0.00000585, 0.00000261, 0.00000253, -0.00000567, -0.00000430, 0.00000256, 0.00001845, 0.00000171, -0.00000815, 0.00002230, -0.00001664, -0.00001760, 0.00002166, -0.00001574, -0.00001491, 0.00001226, 0.00000182, -0.00000183, -0.00000154, -0.00001889, -0.00000145, 0.00001339, -0.00000911, 0.00000350, -0.00000494, -0.00000517, 0.00000944, -0.00001610, -0.00000237, 0.00001105, -0.00003013, 0.00001033, 0.00002337, -0.00005528, 0.00004670, -0.00001077, -0.00003636, 0.00007438, -0.00009682, 0.00006964, -0.00001189, -0.00008585, 0.00014497, -0.00017253, 0.00009842, -0.00000526, -0.00014760, 0.00024283, -0.00029195, 0.00014034, -0.00000039, -0.00033233, 0.00041434, -0.00093020, -0.00012949, 0.00194344, 0.00016193, -0.00074622, 0.00001423, -0.00005464, -0.00007466, -0.00001002, -0.00000169, -0.00004269, 0.00000435, -0.00000348, -0.00002822, 0.00003019, -0.00003776, 0.00001111, 0.00000079, -0.00002612, 0.00002663, -0.00002393, 0.00000769, 0.00000949, -0.00002207, 0.00001920, -0.00000969, -0.00002422, 0.00001101, -0.00000044, -0.00000424, 0.00000469, -0.00001243, -0.00000949, -0.00000049, -0.00000309, 0.00000311, 0.00000394, -0.00000407, -0.00000225, -0.00000226, -0.00000563, 0.00000151, -0.00000858, -0.00000282, 0.00000714, -0.00001273, -0.00000110, 0.00000352, -0.00000918, 0.00000700, 0.00000554, -0.00000941, -0.00000091, 0.00001313, 0.00000746, -0.00000442, },
		{ -0.00000203, -0.00000117, 0.00000396, -0.00000423, 0.00000260, 0.00000697, 0.00000017, 0.00000248, -0.00000168, -0.00000300, 0.00000048, -0.00000327, -0.00000051, 0.00000068, 0.00000025, 0.00000292, 0.00000153, -0.00000156, -0.00000407, 0.00000403, 0.00000611, 0.00000260, 0.00000368, -0.00000425, -0.00000434, -0.00000289, 0.00001628, 0.00001254, -0.00001536, 0.00001855, 0.00000025, -0.00002798, 0.00001561, -0.00000238, -0.00002209, 0.00000746, 0.00000947, -0.00000610, 0.00000057, -0.00001177, -0.00001282, 0.00001521, -0.00000519, -0.00000185, 0.00000221, -0.00001063, 0.00000898, -0.00000587, -0.00001392, 0.00001382, -0.00001932, -0.00001212, 0.00004239, -0.00005263, 0.00001906, 0.00002942, -0.00007270, 0.00008793, -0.00007173, 0.00000681, 0.00006753, -0.00014278, 0.00014608, -0.00010570, -0.00001537, 0.00011334, -0.00021950, 0.00022178, -0.00016100, -0.00005043, 0.00017865, -0.00039726, 0.00031720, -0.00051905, -0.00062281, 0.00160482, 0.00075881, -0.00067607, -0.00018554, -0.00000208, -0.00007025, -0.00004608, 0.00001833, -0.00003853, -0.00001929, 0.00001889, -0.00004224, 0.00002918, -0.00002193, -0.00001023, 0.00001649, -0.00003025, 0.00001959, -0.00001234, -0.00000377, 0.00001577, -0.00002129, 0.00001184, 0.00000118, -0.00002740, 0.00000394, 0.00000591, -0.00000679, 0.00000408, -0.00000607, -0.00001361, -0.00000269, -0.00000112, 0.00000089, 0.00000472, -0.00000322, -0.00000361, 0.00000126, -0.00000668, -0.00000072, -0.00000485, -0.00000721, 0.00000806, -0.00000775, -0.00000746, 0.00000500, -0.00000550, 0.00000183, 0.00000677, -0.00000574, -0.00000588, 0.00000936, 0.00001280, },
		{ 0.00001537, 0.00000319, -0.00000373, 0.00000380, -0.00000365, -0.00000048, 0.00000907, 0.00000163, 0.00000118, -0.00000052, -0.00000344, 0.00000035, -0.00000298, -0.00000111, 0.00000125, -0.00000044, 0.00000252, 0.00000218, -0.00000106, -0.00000438, 0.00000279, 0.00000634, 0.00000214, 0.00000474, -0.00000304, -0.00000356, -0.00000565, 0.00000915, 0.00002250, -0.00001472, 0.00000697, 0.00001550, -0.00002770, 0.00000216, 0.00000835, -0.00002101, -0.00000291, 0.00001605, -0.00000495, -0.00000412, -0.00000154, -0.00001931, 0.00000958, 0.00000281, -0.00000790, 0.00000640, -0.00000964, 0.00000180, 0.00000524, -0.00001754, 0.00000517, -0.00000147, -0.00003059, 0.00004495, -0.00002939, -0.00002037, 0.00006467, -0.00008603, 0.00006613, -0.00001434, -0.00006581, 0.00012193, -0.00014520, 0.00008639, -0.00000177, -0.00012057, 0.00017451, -0.00020275, 0.00011770, 0.00000087, -0.00019317, 0.00023954, -0.00032135, 0.00012313, -0.00016977, -0.00082569, 0.00108320, 0.00116159, -0.00039912, -0.00036475, -0.00002889, -0.00002704, -0.00008022, 0.00001327, -0.00001156, -0.00004938, 0.00003214, -0.00004027, 0.00001318, 0.00000031, -0.00002686, 0.00002085, -0.00002384, 0.00000812, -0.00000245, -0.00001026, 0.00001663, -0.00001676, 0.00000390, 0.00000824, -0.00002490, -0.00000492, 0.00000932, -0.00000526, 0.00000043, -0.00000030, -0.00001350, -0.00000757, 0.00000040, 0.00000028, 0.00000392, -0.00000124, -0.00000562, 0.00000323, -0.00000463, -0.00000434, -0.00000194, -0.00000897, 0.00000553, -0.00000173, -0.00001072, 0.00000232, -0.00000107, -0.00000033, 0.00000499, -0.00000198, -0.00000742, 0.00000319, },
		{ -0.00000271, 0.00001383, 0.00000971, -0.00000412, 0.00000208, -0.00000208, -0.00000374, 0.00000936, 0.00000463, 0.00000011, -0.00000017, -0.00000332, 0.00000007, -0.00000269, -0.00000182, 0.00000182, -0.00000072, 0.00000168, 0.00000271, -0.00000024, -0.00000471, 0.00000141, 0.00000695, 0.00000135, 0.00000533, -0.00000152, -0.00000327, -0.00000516, 0.00000007, 0.00002714, -0.00000598, -0.00000667, 0.00002271, -0.00001722, -0.00001177, 0.00001078, -0.00001236, -0.00001335, 0.00001636, 0.00000294, -0.00001206, 0.00000550, -0.00001712, -0.00000185, 0.00001026, -0.00001004, 0.00000473, -0.00000308, -0.00000665, 0.00001003, -0.00001017, -0.00001045, 0.00001334, -0.00003497, 0.00002794, 0.00000656, -0.00005548, 0.00007804, -0.00006702, 0.00001462, 0.00005282, -0.00011667, 0.00012687, -0.00009138, -0.00000795, 0.00009463, -0.00017425, 0.00015869, -0.00011510, -0.00001569, 0.00012603, -0.00024199, 0.00018887, -0.00017184, -0.00006131, 0.00003836, -0.00079421, 0.00054618, 0.00130473, -0.00001240, -0.00043505, -0.00013263, 0.00001568, -0.00008330, -0.00002196, 0.00002468, -0.00006827, 0.00002603, -0.00002262, -0.00000946, 0.00001802, -0.00003171, 0.00001401, -0.00001257, -0.00000132, 0.00000229, -0.00001199, 0.00001451, -0.00001163, -0.00000201, 0.00001099, -0.00001925, -0.00001192, 0.00000826, -0.00000103, -0.00000318, 0.00000221, -0.00000942, -0.00001248, -0.00000062, 0.00000158, 0.00000231, 0.00000064, -0.00000673, 0.00000278, -0.00000064, -0.00000713, -0.00000128, -0.00000816, 0.00000143, 0.00000266, -0.00001009, -0.00000229, 0.00000128, 0.00000076, 0.00000245, -0.00000031, -0.00000584, },
		{ -0.00000309, -0.00000601, 0.00000863, 0.00001505, -0.00000132, -0.00000057, -0.00000052, -0.00000606, 0.00000747, 0.00000816, 0.00000030, -0.00000069, -0.00000308, -0.00000000, -0.00000244, -0.00000267, 0.00000229, -0.00000054, 0.00000065, 0.00000282, 0.00000084, -0.00000467, -0.00000049, 0.00000787, 0.00000088, 0.00000484, 0.00000069, -0.00000385, -0.00000264, -0.00000733, 0.00002454, 0.00000724, -0.00001590, 0.00001967, -0.00000200, -0.00001892, 0.00000383, -0.00000158, -0.00001771, 0.00000859, 0.00001416, -0.00001712, 0.00000466, -0.00000719, -0.00001347, 0.00001192, -0.00000608, -0.00000146, 0.00000415, -0.00000990, 0.00000510, 0.00000446, -0.00002341, 0.00001544, -0.00002193, -0.00000169, 0.00003978, -0.00007001, 0.00006120, -0.00002110, -0.00004631, 0.00010111, -0.00012295, 0.00007985, -0.00000471, -0.00009618, 0.00014440, -0.00015893, 0.00008123, -0.00000186, -0.00012092, 0.00017387, -0.00019883, 0.00007679, -0.00002685, -0.00016897, 0.00010364, -0.00063566, 0.00011832, 0.00121496, 0.00035939, -0.00035656, -0.00026459, 0.00001332, -0.00004596, -0.00007166, 0.00004823, -0.00006327, 0.00000141, 0.00000101, -0.00002713, 0.00002363, -0.00002461, 0.00000185, -0.00000347, -0.00000498, 0.00000248, -0.00001182, 0.00001233, -0.00000780, -0.00000554, 0.00001109, -0.00001357, -0.00001514, 0.00000387, 0.00000296, -0.00000422, 0.00000108, -0.00000376, -0.00001460, -0.00000472, 0.00000351, 0.00000142, 0.00000137, -0.00000636, 0.00000052, 0.00000329, -0.00000753, -0.00000277, -0.00000620, -0.00000217, 0.00000421, -0.00000681, -0.00000608, 0.00000042, 0.00000334, 0.00000152, -0.00000116, },
		{ -0.00000318, -0.00000146, -0.00000593, 0.00000191, 0.00001696, 0.00000424, -0.00000261, 0.00000005, -0.00000688, 0.00000399, 0.00001085, 0.00000223, -0.00000148, -0.00000334, 0.00000023, -0.00000202, -0.00000375, 0.00000252, 0.00000007, -0.00000030, 0.00000244, 0.00000184, -0.00000386, -0.00000287, 0.00000848, 0.00000149, 0.00000307, 0.00000329, -0.00000469, -0.00000007, -0.00001068, 0.00001611, 0.00001943, -0.00001679, 0.00000927, 0.00001042, -0.00001575, -0.00000802, 0.00000414, -0.00001303, -0.00000398, 0.00002234, -0.00001437, -0.00000401, 0.00000488, -0.00001874, 0.00000552, 0.00000177, -0.00000726, 0.00000641, -0.00000487, -0.00000657, 0.00001742, -0.00002426, 0.00000162, 0.00000221, -0.00002958, 0.00005462, -0.00005608, 0.00001966, 0.00003308, -0.00009088, 0.00010951, -0.00008158, 0.00000160, 0.00007754, -0.00014218, 0.00013141, -0.00008838, -0.00001817, 0.00008766, -0.00016235, 0.00014358, -0.00010302, -0.00003446, 0.00006067, -0.00018763, 0.00007726, -0.00045467, -0.00014997, 0.00098284, 0.00061594, -0.00015377, -0.00035381, -0.00005670, 0.00001327, -0.00010806, 0.00004233, -0.00003497, -0.00002989, 0.00001753, -0.00003199, 0.00001688, -0.00001165, -0.00000803, -0.00000077, -0.00000363, 0.00000161, -0.00001278, 0.00001131, -0.00000493, -0.00000808, 0.00001071, -0.00000966, -0.00001512, -0.00000107, 0.00000422, -0.00000225, -0.00000186, 0.00000046, -0.00001264, -0.00001041, 0.00000395, 0.00000226, 0.00000090, -0.00000510, -0.00000212, 0.00000554, -0.00000545, -0.00000502, -0.00000470, -0.00000415, 0.00000333, -0.00000304, -0.00000733, -0.00000274, 0.00000499, 0.00000319, },
		{ 0.00000660, -0.00000421, -0.00000220, -0.00000355, -0.00000364, 0.00001456, 0.00001071, -0.00000245, -0.00000067, -0.00000657, 0.00000009, 0.00001163, 0.00000548, -0.00000151, -0.00000442, 0.00000042, -0.00000106, -0.00000501, 0.00000236, 0.00000104, -0.00000101, 0.00000178, 0.00000232, -0.00000226, -0.00000515, 0.00000795, 0.00000352, 0.00000059, 0.00000538, -0.00000479, 0.00000107, -0.00000982, 0.00000543, 0.00002608, -0.00000974, -0.00000232, 0.00001451, -0.00000445, -0.00001710, 0.00000018, -0.00000191, -0.00001450, 0.00002202, -0.00000347, -0.00001542, 0.00001191, -0.00001430, -0.00000612, 0.00000780, -0.00000750, 0.00000128, 0.00000576, -0.00001703, 0.00001958, -0.00000963, -0.00002207, 0.00002405, -0.00004112, 0.00004348, -0.00001906, -0.00002892, 0.00007218, -0.00010019, 0.00007470, -0.00001069, -0.00007400, 0.00012182, -0.00013066, 0.00006797, 0.00000170, -0.00009616, 0.00012255, -0.00013792, 0.00006672, -0.00000604, -0.00010060, 0.00007896, -0.00014766, 0.00002379, -0.00031519, -0.00027743, 0.00071458, 0.00071636, 0.00009909, -0.00034272, -0.00017882, 0.00005660, -0.00010822, 0.00000632, 0.00000238, -0.00005198, 0.00001838, -0.00002429, 0.00000419, -0.00000085, -0.00001097, -0.00000374, -0.00000133, 0.00000323, -0.00001571, 0.00001031, -0.00000139, -0.00001110, 0.00001086, -0.00000724, -0.00001402, -0.00000400, 0.00000221, 0.00000104, -0.00000394, 0.00000141, -0.00000753, -0.00001499, 0.00000140, 0.00000452, 0.00000025, -0.00000402, -0.00000398, 0.00000571, -0.00000207, -0.00000645, -0.00000445, -0.00000471, 0.00000141, -0.00000050, -0.00000605, -0.00000614, 0.00000408, },
		{ 0.00000069, 0.00000968, -0.00000271, -0.00000481, -0.00000102, -0.00000616, 0.00000886, 0.00001556, 0.00000065, -0.00000193, -0.00000603, -0.00000314, 0.00001032, 0.00000886, 0.00000006, -0.00000597, -0.00000012, 0.00000062, -0.00000617, 0.00000153, 0.00000231, -0.00000145, 0.00000110, 0.00000213, -0.00000030, -0.00000650, 0.00000582, 0.00000649, -0.00000148, 0.00000605, -0.00000363, 0.00000053, -0.00000635, -0.00000368, 0.00002566, 0.00000122, -0.00000909, 0.00000962, 0.00000827, -0.00001685, -0.00001239, 0.00000850, -0.00001649, 0.00001236, 0.00001067, -0.00002237, 0.00000945, -0.00000238, -0.00001622, 0.00000637, -0.00000055, -0.00000824, 0.00001477, -0.00001796, 0.00000741, 0.00001489, -0.00004266, 0.00002996, -0.00002937, 0.00001143, 0.00002427, -0.00006293, 0.00007947, -0.00007132, 0.00001255, 0.00005917, -0.00011588, 0.00011336, -0.00007244, -0.00001343, 0.00007222, -0.00012552, 0.00010085, -0.00007356, -0.00001353, 0.00005401, -0.00011032, 0.00005068, -0.00009384, -0.00001212, -0.00023074, -0.00032060, 0.00048801, 0.00068413, 0.00031412, -0.00021604, -0.00030355, 0.00004783, -0.00006898, -0.00004297, 0.00002899, -0.00005438, 0.00000452, -0.00001155, -0.00000603, 0.00000281, -0.00000756, -0.00000793, -0.00000230, 0.00000826, -0.00001867, 0.00000698, 0.00000405, -0.00001466, 0.00001081, -0.00000474, -0.00001396, -0.00000412, -0.00000146, 0.00000309, -0.00000344, -0.00000061, -0.00000187, -0.00001613, -0.00000395, 0.00000657, 0.00000071, -0.00000376, -0.00000486, 0.00000451, 0.00000110, -0.00000625, -0.00000509, -0.00000476, -0.00000033, 0.00000029, -0.00000353, -0.00000789, },
		{ -0.00000735, -0.00000365, 0.00001050, 0.00000124, -0.00000746, -0.00000026, -0.00000547, 0.00000220, 0.00001686, 0.00000593, -0.00000229, -0.00000602, -0.00000532, 0.00000763, 0.00001106, 0.00000330, -0.00000697, -0.00000199, 0.00000277, -0.00000664, -0.00000020, 0.00000373, -0.00000156, 0.00000052, 0.00000147, 0.00000131, -0.00000638, 0.00000241, 0.00000926, -0.00000203, 0.00000503, -0.00000154, -0.00000091, -0.00000243, -0.00000902, 0.00001984, 0.00001087, -0.00000836, 0.00000002, 0.00001498, -0.00000578, -0.00002686, 0.00001044, -0.00000810, -0.00000178, 0.00002052, -0.00001992, -0.00000121, 0.00001001, -0.00001784, -0.00000398, 0.00000971, -0.00001515, 0.00001479, -0.00000611, -0.00001422, 0.00003689, -0.00004695, 0.00001374, 0.00000057, -0.00002574, 0.00005367, -0.00006732, 0.00005299, -0.00001854, -0.00004979, 0.00009956, -0.00010972, 0.00006138, 0.00000283, -0.00007701, 0.00009867, -0.00010477, 0.00004501, -0.00000579, -0.00006426, 0.00006612, -0.00008126, 0.00001156, -0.00005871, -0.00001923, -0.00017989, -0.00033571, 0.00033085, 0.00058380, 0.00042879, -0.00000879, -0.00037013, -0.00002599, -0.00001094, -0.00008051, 0.00003117, -0.00003817, -0.00001461, -0.00000325, -0.00000877, -0.00000005, -0.00000251, -0.00000856, -0.00000785, 0.00001432, -0.00001834, -0.00000013, 0.00001075, -0.00001726, 0.00000899, -0.00000073, -0.00001557, -0.00000270, -0.00000421, 0.00000220, -0.00000060, -0.00000356, 0.00000172, -0.00001322, -0.00001019, 0.00000650, 0.00000277, -0.00000397, -0.00000536, 0.00000296, 0.00000312, -0.00000472, -0.00000565, -0.00000507, -0.00000146, -0.00000011, -0.00000133, },
		{ -0.00000028, -0.00000519, -0.00000703, 0.00000841, 0.00000606, -0.00000822, -0.00000183, -0.00000302, -0.00000295, 0.00001423, 0.00001142, -0.00000051, -0.00000652, -0.00000689, 0.00000467, 0.00001135, 0.00000731, -0.00000625, -0.00000517, 0.00000455, -0.00000579, -0.00000275, 0.00000493, -0.00000120, -0.00000000, 0.00000076, 0.00000203, -0.00000484, -0.00000133, 0.00001064, -0.00000068, 0.00000295, 0.00000055, -0.00000222, 0.00000048, -0.00001054, 0.00001203, 0.00001565, -0.00000181, -0.00000771, 0.00001171, 0.00001139, -0.00003434, 0.00000027, 0.00000621, -0.00001229, 0.00002061, -0.00000882, -0.00001354, 0.00001507, -0.00000850, -0.00001862, 0.00001563, -0.00001282, 0.00000332, 0.00001397, -0.00003388, 0.00004353, -0.00002894, -0.00001917, 0.00003353, -0.00004917, 0.00005543, -0.00004106, 0.00000636, 0.00003389, -0.00008685, 0.00009685, -0.00006313, -0.00000722, 0.00006211, -0.00010017, 0.00007982, -0.00005412, -0.00001319, 0.00003720, -0.00007509, 0.00004410, -0.00004312, -0.00001150, -0.00004986, -0.00001236, -0.00013318, -0.00035261, 0.00022639, 0.00048411, 0.00042956, 0.00020992, -0.00033600, -0.00014417, 0.00003131, -0.00008763, 0.00000932, -0.00001502, -0.00002733, -0.00000463, -0.00000527, -0.00000446, -0.00000096, -0.00000392, -0.00001552, 0.00001729, -0.00001253, -0.00001009, 0.00001625, -0.00001680, 0.00000438, 0.00000480, -0.00001767, -0.00000202, -0.00000416, -0.00000140, 0.00000265, -0.00000510, 0.00000203, -0.00000768, -0.00001469, 0.00000335, 0.00000563, -0.00000362, -0.00000615, 0.00000164, 0.00000404, -0.00000282, -0.00000537, -0.00000564, -0.00000233, -0.00000084, },
		{ -0.00000144, -0.00000029, -0.00000276, -0.00000829, 0.00000431, 0.00000965, -0.00000626, -0.00000472, -0.00000084, -0.00000524, 0.00000899, 0.00001490, 0.00000358, -0.00000657, -0.00000856, 0.00000216, 0.00000997, 0.00001064, -0.00000318, -0.00000880, 0.00000486, -0.00000336, -0.00000561, 0.00000537, -0.00000024, -0.00000057, 0.00000033, 0.00000174, -0.00000257, -0.00000425, 0.00001006, 0.00000201, 0.00000094, 0.00000185, -0.00000293, 0.00000212, -0.00000962, 0.00000527, 0.00001516, 0.00000593, -0.00000856, 0.00000016, 0.00002609, -0.00002873, -0.00001881, 0.00001765, -0.00001265, 0.00001116, 0.00000461, -0.00001956, 0.00000910, 0.00000762, -0.00002875, 0.00001040, 0.00000030, -0.00001490, 0.00003179, -0.00003958, 0.00002874, 0.00000606, -0.00005366, 0.00005165, -0.00004645, 0.00002917, 0.00000249, -0.00003811, 0.00006356, -0.00008637, 0.00005720, -0.00000020, -0.00006189, 0.00008422, -0.00008172, 0.00003423, -0.00000204, -0.00004931, 0.00004642, -0.00005691, 0.00001280, -0.00001913, -0.00001181, -0.00005412, -0.00001335, -0.00007520, -0.00036992, 0.00013795, 0.00042671, 0.00035085, 0.00036823, -0.00019813, -0.00025946, 0.00002786, -0.00006347, -0.00002236, 0.00000026, -0.00002718, -0.00001359, -0.00000142, -0.00000507, -0.00000482, 0.00000358, -0.00002072, 0.00001415, -0.00000196, -0.00001962, 0.00001768, -0.00001227, -0.00000245, 0.00001031, -0.00001826, -0.00000384, -0.00000126, -0.00000575, 0.00000402, -0.00000407, -0.00000024, -0.00000207, -0.00001562, -0.00000217, 0.00000759, -0.00000186, -0.00000729, 0.00000037, 0.00000445, -0.00000136, -0.00000428, -0.00000591, -0.00000336, },
		{ -0.00000456, -0.00000206, -0.00000073, -0.00000106, -0.00000753, -0.00000004, 0.00001065, -0.00000230, -0.00000713, -0.00000030, -0.00000494, 0.00000337, 0.00001503, 0.00000878, -0.00000488, -0.00001064, 0.00000005, 0.00000794, 0.00001205, 0.00000181, -0.00001133, 0.00000295, 0.00000016, -0.00000786, 0.00000461, 0.00000130, -0.00000117, 0.00000026, 0.00000080, -0.00000048, -0.00000562, 0.00000780, 0.00000489, -0.00000004, 0.00000223, -0.00000328, 0.00000296, -0.00000782, 0.00000083, 0.00001173, 0.00001023, -0.00000179, -0.00001326, 0.00003056, -0.00001046, -0.00003796, 0.00001804, -0.00000218, -0.00000147, 0.00001217, -0.00001499, -0.00000472, 0.00002135, -0.00002634, -0.00000710, 0.00001857, -0.00002982, 0.00003645, -0.00002560, -0.00000279, 0.00004328, -0.00007220, 0.00004358, -0.00001801, -0.00001193, 0.00004226, -0.00006041, 0.00006020, -0.00005375, 0.00000178, 0.00005114, -0.00008261, 0.00006763, -0.00003873, -0.00001208, 0.00002971, -0.00005582, 0.00003120, -0.00003043, -0.00000762, -0.00001578, -0.00000038, -0.00005222, -0.00003359, -0.00001172, -0.00036823, 0.00003645, 0.00041367, 0.00025263, 0.00042241, 0.00000445, -0.00031925, -0.00002928, -0.00002631, -0.00004475, -0.00000104, -0.00001678, -0.00002251, -0.00000324, -0.00000016, -0.00001161, 0.00000914, -0.00001988, 0.00000511, 0.00000978, -0.00002486, 0.00001356, -0.00000465, -0.00000941, 0.00001372, -0.00001588, -0.00000829, 0.00000276, -0.00000836, 0.00000234, -0.00000112, -0.00000306, 0.00000139, -0.00001287, -0.00000800, 0.00000713, 0.00000123, -0.00000806, -0.00000138, 0.00000483, -0.00000045, -0.00000305, -0.00000530, },
		{ -0.00000386, -0.00000535, -0.00000305, -0.00000117, -0.00000032, -0.00000576, -0.00000308, 0.00000910, 0.00000189, -0.00000766, -0.00000141, -0.00000351, -0.00000072, 0.00001209, 0.00001306, -0.00000074, -0.00001250, -0.00000236, 0.00000628, 0.00001122, 0.00000729, -0.00001134, -0.00000103, 0.00000360, -0.00000860, 0.00000255, 0.00000306, -0.00000161, 0.00000037, -0.00000022, 0.00000075, -0.00000547, 0.00000481, 0.00000689, 0.00000020, 0.00000229, -0.00000381, 0.00000338, -0.00000580, -0.00000183, 0.00000846, 0.00000907, 0.00000885, -0.00002107, 0.00002212, 0.00001336, -0.00004720, 0.00000481, 0.00001290, -0.00000843, 0.00000906, -0.00000237, -0.00001771, 0.00002367, -0.00000940, -0.00003029, 0.00003143, -0.00003157, 0.00002308, 0.00000376, -0.00003722, 0.00006545, -0.00006373, 0.00001053, 0.00002295, -0.00004784, 0.00005990, -0.00005208, 0.00002945, -0.00000775, -0.00004443, 0.00007205, -0.00006730, 0.00002798, 0.00000461, -0.00003924, 0.00003473, -0.00004131, 0.00000934, -0.00001274, -0.00001084, -0.00002427, 0.00000591, -0.00003330, -0.00006733, 0.00003803, -0.00033016, -0.00008315, 0.00041644, 0.00018978, 0.00037270, 0.00020764, -0.00029022, -0.00011987, -0.00000372, -0.00004567, -0.00001622, -0.00000579, -0.00002374, -0.00001235, 0.00000717, -0.00001632, 0.00000887, -0.00001278, -0.00000634, 0.00001816, -0.00002358, 0.00000496, 0.00000341, -0.00001410, 0.00001366, -0.00001054, -0.00001386, 0.00000549, -0.00000774, -0.00000178, 0.00000186, -0.00000451, 0.00000192, -0.00000800, -0.00001186, 0.00000386, 0.00000444, -0.00000749, -0.00000381, 0.00000507, 0.00000038, -0.00000239, },
		{ -0.00000243, -0.00000227, -0.00000507, -0.00000445, -0.00000168, -0.00000014, -0.00000408, -0.00000429, 0.00000626, 0.00000465, -0.00000619, -0.00000300, -0.00000245, -0.00000264, 0.00000767, 0.00001462, 0.00000526, -0.00001279, -0.00000575, 0.00000538, 0.00000892, 0.00001143, -0.00000824, -0.00000588, 0.00000558, -0.00000744, -0.00000029, 0.00000449, -0.00000163, 0.00000041, -0.00000088, 0.00000098, -0.00000441, 0.00000211, 0.00000760, 0.00000103, 0.00000283, -0.00000463, 0.00000318, -0.00000337, -0.00000399, 0.00000724, 0.00000389, 0.00001756, -0.00001885, 0.00000460, 0.00003255, -0.00004095, -0.00001698, 0.00002270, -0.00000386, -0.00000260, 0.00000993, -0.00002073, 0.00001140, 0.00001575, -0.00004746, 0.00002917, -0.00001647, -0.00000380, 0.00003568, -0.00005820, 0.00006146, -0.00002951, -0.00003400, 0.00005701, -0.00006147, 0.00004825, -0.00001942, -0.00001126, 0.00002963, -0.00006436, 0.00005986, -0.00002990, -0.00001221, 0.00002974, -0.00004158, 0.00002124, -0.00002174, -0.00000458, -0.00000925, -0.00000372, -0.00003008, -0.00000323, -0.00000117, -0.00009715, 0.00005635, -0.00025471, -0.00020236, 0.00039787, 0.00018864, 0.00025838, 0.00034989, -0.00017409, -0.00020376, -0.00001677, -0.00002783, -0.00003313, -0.00000429, -0.00001509, -0.00002458, 0.00001111, -0.00001498, 0.00000240, -0.00000275, -0.00001541, 0.00002019, -0.00001640, -0.00000501, 0.00000898, -0.00001513, 0.00001034, -0.00000371, -0.00001838, 0.00000527, -0.00000411, -0.00000634, 0.00000299, -0.00000386, 0.00000039, -0.00000328, -0.00001248, -0.00000108, 0.00000618, -0.00000520, -0.00000638, 0.00000457, 0.00000163, },
		{ 0.00000329, -0.00000262, -0.00000141, -0.00000361, -0.00000568, -0.00000267, -0.00000008, -0.00000300, -0.00000421, 0.00000367, 0.00000536, -0.00000384, -0.00000376, -0.00000234, -0.00000308, 0.00000372, 0.00001295, 0.00001132, -0.00001020, -0.00001000, 0.00000466, 0.00000654, 0.00001298, -0.00000275, -0.00000976, 0.00000520, -0.00000481, -0.00000301, 0.00000507, -0.00000117, 0.00000029, -0.00000107, 0.00000050, -0.00000320, 0.00000021, 0.00000742, 0.00000158, 0.00000409, -0.00000512, 0.00000173, -0.00000021, -0.00000631, 0.00000782, -0.00000173, 0.00001981, -0.00000751, -0.00001376, 0.00003916, -0.00002129, -0.00003696, 0.00001942, 0.00001088, -0.00001430, 0.00001261, -0.00000998, -0.00001015, 0.00003693, -0.00004782, 0.00000888, 0.00000998, -0.00003203, 0.00005510, -0.00005506, 0.00003205, 0.00001714, -0.00007068, 0.00006801, -0.00004674, 0.00001411, 0.00002066, -0.00004204, 0.00004360, -0.00005531, 0.00002660, 0.00000878, -0.00003527, 0.00003128, -0.00002700, 0.00000400, -0.00000993, -0.00000625, -0.00001416, 0.00000176, -0.00002373, -0.00002568, 0.00002970, -0.00010605, 0.00003901, -0.00015918, -0.00029296, 0.00033414, 0.00024040, 0.00013631, 0.00039699, -0.00000681, -0.00023975, -0.00006720, -0.00000784, -0.00003826, -0.00001578, -0.00000161, -0.00003260, 0.00000728, -0.00000748, -0.00000686, 0.00000530, -0.00001873, 0.00001598, -0.00000649, -0.00001284, 0.00001042, -0.00001285, 0.00000526, 0.00000261, -0.00002023, 0.00000193, 0.00000082, -0.00000919, 0.00000151, -0.00000184, -0.00000152, -0.00000037, -0.00001020, -0.00000570, 0.00000538, -0.00000185, -0.00000803, 0.00000285, },
		{ 0.00000012, 0.00000468, -0.00000218, -0.00000166, -0.00000159, -0.00000591, -0.00000430, -0.00000013, -0.00000234, -0.00000377, 0.00000222, 0.00000470, -0.00000209, -0.00000319, -0.00000259, -0.00000331, 0.00000140, 0.00000906, 0.00001522, -0.00000447, -0.00001391, 0.00000304, 0.00000521, 0.00001183, 0.00000335, -0.00001108, 0.00000259, -0.00000188, -0.00000473, 0.00000469, -0.00000043, 0.00000009, -0.00000089, -0.00000021, -0.00000233, -0.00000100, 0.00000707, 0.00000138, 0.00000550, -0.00000426, -0.00000132, 0.00000337, -0.00000831, 0.00000844, -0.00000438, 0.00001468, 0.00000755, -0.00002480, 0.00003157, 0.00000304, -0.00004506, 0.00000243, 0.00002717, -0.00001602, 0.00000152, 0.00000978, -0.00002928, 0.00004243, -0.00002785, -0.00002280, 0.00003526, -0.00004782, 0.00005238, -0.00002820, -0.00001041, 0.00005717, -0.00008360, 0.00005066, -0.00001083, -0.00002602, 0.00004955, -0.00005007, 0.00003229, -0.00002774, -0.00000907, 0.00003191, -0.00003569, 0.00001667, -0.00000954, -0.00000525, -0.00000930, -0.00000112, -0.00001733, -0.00000243, -0.00000612, -0.00004965, 0.00004421, -0.00008794, -0.00000222, -0.00006987, -0.00033397, 0.00022580, 0.00031173, 0.00005498, 0.00035272, 0.00015781, -0.00020529, -0.00013467, -0.00000654, -0.00002606, -0.00003440, 0.00000740, -0.00003083, -0.00000406, 0.00000214, -0.00001391, 0.00000793, -0.00001614, 0.00000843, 0.00000236, -0.00001637, 0.00000810, -0.00000895, 0.00000029, 0.00000697, -0.00001906, -0.00000328, 0.00000491, -0.00000918, -0.00000181, -0.00000003, -0.00000244, 0.00000037, -0.00000655, -0.00000822, 0.00000229, 0.00000110, -0.00000792, },
		{ -0.00000607, -0.00000267, 0.00000495, -0.00000076, -0.00000259, -0.00000009, -0.00000469, -0.00000612, -0.00000072, -0.00000169, -0.00000360, 0.00000175, 0.00000396, -0.00000178, -0.00000201, -0.00000207, -0.00000410, 0.00000053, 0.00000489, 0.00001554, 0.00000318, -0.00001571, -0.00000024, 0.00000511, 0.00000910, 0.00000808, -0.00000932, -0.00000098, 0.00000007, -0.00000514, 0.00000377, 0.00000020, -0.00000009, -0.00000050, -0.00000088, -0.00000181, -0.00000200, 0.00000706, 0.00000079, 0.00000594, -0.00000142, -0.00000536, 0.00000617, -0.00000881, 0.00000738, -0.00000284, 0.00000503, 0.00001960, -0.00002463, 0.00001507, 0.00002136, -0.00003730, -0.00002068, 0.00003369, -0.00000262, -0.00001882, 0.00002719, -0.00003420, 0.00002765, 0.00000578, -0.00005216, 0.00004607, -0.00004302, 0.00002820, 0.00001104, -0.00004786, 0.00007468, -0.00006755, 0.00001249, 0.00003007, -0.00005438, 0.00005558, -0.00003451, 0.00000555, 0.00000191, -0.00003166, 0.00003362, -0.00002010, -0.00000063, -0.00000036, -0.00000364, -0.00001446, 0.00000173, -0.00001212, -0.00001562, 0.00001347, -0.00006206, 0.00003587, -0.00005048, -0.00004659, -0.00000855, -0.00032236, 0.00009457, 0.00036467, 0.00003639, 0.00025210, 0.00027141, -0.00010708, -0.00018598, -0.00003618, -0.00000357, -0.00004835, 0.00000447, -0.00001940, -0.00001781, 0.00000830, -0.00001534, 0.00000494, -0.00001041, 0.00000131, 0.00000764, -0.00001573, 0.00000394, -0.00000535, -0.00000338, 0.00000902, -0.00001569, -0.00000861, 0.00000672, -0.00000661, -0.00000531, 0.00000027, -0.00000209, -0.00000018, -0.00000334, -0.00000802, -0.00000167, 0.00000222, },
		{ 0.00000097, -0.00000354, -0.00000433, 0.00000377, 0.00000118, -0.00000323, 0.00000011, -0.00000241, -0.00000717, -0.00000221, -0.00000099, -0.00000361, 0.00000149, 0.00000409, -0.00000249, -0.00000162, -0.00000018, -0.00000508, -0.00000006, 0.00000217, 0.00001243, 0.00001042, -0.00001401, -0.00000490, 0.00000541, 0.00000635, 0.00001021, -0.00000532, -0.00000380, 0.00000035, -0.00000473, 0.00000299, 0.00000051, -0.00000035, 0.00000006, -0.00000141, -0.00000142, -0.00000315, 0.00000721, 0.00000071, 0.00000455, 0.00000298, -0.00000884, 0.00000680, -0.00000707, 0.00000424, 0.00000163, -0.00000460, 0.00002419, -0.00001519, -0.00000145, 0.00002684, -0.00001817, -0.00003782, 0.00002320, 0.00002198, -0.00003673, 0.00003038, -0.00001994, -0.00000156, 0.00003883, -0.00006508, 0.00003539, -0.00001945, -0.00000681, 0.00004605, -0.00006504, 0.00006409, -0.00002986, -0.00003040, 0.00005891, -0.00006011, 0.00003850, -0.00000517, -0.00002164, 0.00001992, -0.00003525, 0.00001925, -0.00000119, -0.00000963, -0.00000225, 0.00000342, -0.00001682, -0.00000339, 0.00000051, -0.00003007, 0.00002431, -0.00005654, 0.00000982, -0.00000989, -0.00007543, 0.00001693, -0.00027273, -0.00002992, 0.00037461, 0.00007270, 0.00014232, 0.00030995, 0.00002247, -0.00019137, -0.00009163, 0.00001260, -0.00004736, -0.00001086, -0.00000452, -0.00002707, 0.00000740, -0.00001133, -0.00000079, -0.00000527, -0.00000287, 0.00000923, -0.00001278, 0.00000004, -0.00000307, -0.00000567, 0.00000935, -0.00001138, -0.00001268, 0.00000603, -0.00000288, -0.00000741, -0.00000117, -0.00000123, -0.00000079, -0.00000160, -0.00000584, -0.00000464, },
		{ -0.00000542, -0.00000188, -0.00000189, -0.00000426, 0.00000168, 0.00000267, -0.00000279, -0.00000090, -0.00000019, -0.00000667, -0.00000435, -0.00000074, -0.00000335, 0.00000079, 0.00000508, -0.00000295, -0.00000306, 0.00000249, -0.00000505, -0.00000146, 0.00000139, 0.00000753, 0.00001488, -0.00000875, -0.00000950, 0.00000485, 0.00000472, 0.00000980, -0.00000083, -0.00000458, -0.00000065, -0.00000450, 0.00000283, 0.00000071, -0.00000096, 0.00000082, -0.00000184, -0.00000099, -0.00000436, 0.00000688, 0.00000200, 0.00000143, 0.00000743, -0.00001002, 0.00000462, -0.00000354, 0.00000012, 0.00000647, -0.00001061, 0.00002109, -0.00000249, -0.00001059, 0.00002007, 0.00000166, -0.00003923, -0.00000235, 0.00004574, -0.00003959, 0.00001420, 0.00000810, -0.00003135, 0.00005630, -0.00005429, 0.00000647, 0.00001199, -0.00003749, 0.00006280, -0.00005625, 0.00003189, 0.00001374, -0.00006127, 0.00006546, -0.00004320, 0.00000797, 0.00002368, -0.00003699, 0.00002106, -0.00002405, 0.00000023, 0.00001011, -0.00000716, -0.00000998, 0.00000755, -0.00001118, -0.00001521, 0.00001353, -0.00003688, 0.00002100, -0.00003639, -0.00002095, 0.00001830, -0.00008069, 0.00001372, -0.00020821, -0.00012454, 0.00033817, 0.00013661, 0.00006179, 0.00027992, 0.00014046, -0.00013945, -0.00015095, 0.00000572, -0.00002994, -0.00003118, 0.00000511, -0.00002743, 0.00000011, -0.00000532, -0.00000545, -0.00000313, -0.00000403, 0.00000885, -0.00000965, -0.00000250, -0.00000192, -0.00000722, 0.00000871, -0.00000709, -0.00001502, 0.00000363, 0.00000047, -0.00000746, -0.00000350, -0.00000101, -0.00000072, -0.00000122, -0.00000318, },
		{ -0.00000135, -0.00000397, -0.00000472, -0.00000221, -0.00000291, -0.00000015, 0.00000286, -0.00000126, -0.00000218, 0.00000081, -0.00000462, -0.00000617, -0.00000154, -0.00000265, -0.00000028, 0.00000605, -0.00000195, -0.00000609, 0.00000431, -0.00000291, -0.00000377, 0.00000167, 0.00000303, 0.00001537, -0.00000144, -0.00001216, 0.00000272, 0.00000429, 0.00000801, 0.00000243, -0.00000319, -0.00000166, -0.00000531, 0.00000307, 0.00000140, -0.00000215, 0.00000161, -0.00000203, -0.00000064, -0.00000518, 0.00000538, 0.00000473, -0.00000222, 0.00001006, -0.00000802, 0.00000040, 0.00000023, -0.00000332, 0.00000964, -0.00001203, 0.00001355, 0.00000726, -0.00001031, 0.00000830, 0.00001148, -0.00002340, -0.00003176, 0.00005458, -0.00002135, -0.00001590, 0.00003623, -0.00004674, 0.00005020, -0.00002344, -0.00002854, 0.00003684, -0.00005156, 0.00005607, -0.00002709, -0.00000727, 0.00004683, -0.00007014, 0.00004991, -0.00001289, -0.00002235, 0.00004046, -0.00003599, 0.00000907, -0.00000791, -0.00001308, 0.00000987, 0.00000214, -0.00001519, 0.00000351, 0.00000111, -0.00002700, 0.00001912, -0.00003179, 0.00000619, -0.00001176, -0.00004329, 0.00002672, -0.00006657, -0.00000275, -0.00014899, -0.00018101, 0.00026950, 0.00019778, 0.00002703, 0.00021050, 0.00021250, -0.00004313, -0.00018524, -0.00003061, -0.00000553, -0.00004535, 0.00000366, -0.00001974, -0.00000911, -0.00000153, -0.00000678, -0.00000387, -0.00000399, 0.00000851, -0.00000741, -0.00000399, -0.00000098, -0.00000861, 0.00000749, -0.00000316, -0.00001589, 0.00000061, 0.00000251, -0.00000595, -0.00000542, -0.00000205, -0.00000017, -0.00000135, },
		{ -0.00000112, -0.00000082, -0.00000138, -0.00000596, -0.00000448, -0.00000159, -0.00000067, 0.00000173, 0.00000051, -0.00000259, 0.00000019, -0.00000205, -0.00000652, -0.00000348, -0.00000198, -0.00000107, 0.00000606, 0.00000075, -0.00000924, 0.00000357, 0.00000129, -0.00000585, 0.00000155, 0.00000045, 0.00001241, 0.00000547, -0.00001163, -0.00000062, 0.00000421, 0.00000625, 0.00000362, -0.00000070, -0.00000128, -0.00000718, 0.00000281, 0.00000318, -0.00000375, 0.00000201, -0.00000173, -0.00000063, -0.00000517, 0.00000260, 0.00000806, -0.00000473, 0.00000969, -0.00000343, -0.00000397, 0.00000258, -0.00000515, 0.00001074, -0.00001025, 0.00000570, 0.00001114, -0.00000444, 0.00000078, 0.00000673, 0.00000138, -0.00004994, 0.00004035, 0.00001292, -0.00004576, 0.00004947, -0.00003955, 0.00002364, 0.00001460, -0.00005460, 0.00004436, -0.00004515, 0.00003091, 0.00000917, -0.00003779, 0.00005937, -0.00005710, 0.00002120, 0.00001791, -0.00004140, 0.00004094, -0.00002233, -0.00000725, 0.00000391, -0.00001626, 0.00000156, 0.00001021, -0.00001234, -0.00000755, 0.00001348, -0.00003148, 0.00001385, -0.00001749, -0.00001181, 0.00000673, -0.00005040, 0.00001823, -0.00004468, -0.00001797, -0.00010484, -0.00020507, 0.00018928, 0.00023654, 0.00003199, 0.00013714, 0.00022606, 0.00006633, -0.00017281, -0.00008696, 0.00000995, -0.00004536, -0.00000813, -0.00000961, -0.00001499, -0.00000222, -0.00000532, -0.00000543, -0.00000480, 0.00000907, -0.00000564, -0.00000545, 0.00000039, -0.00000980, 0.00000557, 0.00000045, -0.00001577, -0.00000227, 0.00000322, -0.00000396, -0.00000599, -0.00000405, 0.00000006, },
		{ -0.00000081, -0.00000035, -0.00000113, 0.00000096, -0.00000493, -0.00000751, -0.00000174, 0.00000031, 0.00000019, 0.00000136, -0.00000158, -0.00000133, -0.00000042, -0.00000491, -0.00000570, -0.00000229, -0.00000098, 0.00000490, 0.00000416, -0.00001060, -0.00000027, 0.00000600, -0.00000601, 0.00000010, -0.00000014, 0.00000788, 0.00000979, -0.00000808, -0.00000374, 0.00000342, 0.00000530, 0.00000320, 0.00000119, 0.00000113, -0.00000906, 0.00000095, 0.00000608, -0.00000502, 0.00000140, -0.00000073, -0.00000106, -0.00000429, -0.00000088, 0.00001064, -0.00000486, 0.00000653, 0.00000189, -0.00000661, 0.00000291, -0.00000576, 0.00001066, -0.00000733, -0.00000010, 0.00001056, 0.00000044, 0.00000304, -0.00000817, 0.00002107, -0.00004609, 0.00000632, 0.00004850, -0.00005960, 0.00003966, -0.00001281, -0.00001064, 0.00004403, -0.00006112, 0.00003242, -0.00002365, -0.00000089, 0.00003831, -0.00004988, 0.00005096, -0.00003065, -0.00000843, 0.00003886, -0.00004492, 0.00002841, -0.00000420, -0.00001960, 0.00000713, -0.00001149, -0.00000765, 0.00001102, -0.00000186, -0.00001936, 0.00001884, -0.00002557, 0.00000056, -0.00000140, -0.00002436, 0.00001349, -0.00004387, 0.00000218, -0.00002638, -0.00002496, -0.00007468, -0.00020912, 0.00011374, 0.00024880, 0.00005755, 0.00008564, 0.00019255, 0.00015361, -0.00011023, -0.00014096, 0.00000291, -0.00003166, -0.00002331, -0.00000383, -0.00001483, -0.00000628, -0.00000361, -0.00000562, -0.00000723, 0.00000981, -0.00000315, -0.00000763, 0.00000203, -0.00001013, 0.00000270, 0.00000370, -0.00001488, -0.00000482, 0.00000316, -0.00000245, -0.00000518, -0.00000602, },
		{ -0.00000693, -0.00000287, 0.00000046, -0.00000146, 0.00000213, -0.00000211, -0.00000966, -0.00000398, 0.00000189, -0.00000047, 0.00000072, 0.00000034, -0.00000236, -0.00000071, -0.00000205, -0.00000673, -0.00000421, -0.00000021, 0.00000334, 0.00000674, -0.00000905, -0.00000609, 0.00000886, -0.00000328, -0.00000230, 0.00000010, 0.00000381, 0.00001072, -0.00000310, -0.00000508, 0.00000158, 0.00000493, 0.00000238, 0.00000125, 0.00000490, -0.00000932, -0.00000303, 0.00000922, -0.00000488, -0.00000058, 0.00000075, -0.00000170, -0.00000296, -0.00000407, 0.00001136, -0.00000249, 0.00000203, 0.00000590, -0.00000672, 0.00000206, -0.00000640, 0.00001037, -0.00000418, -0.00000425, 0.00000943, -0.00000044, 0.00001352, -0.00002222, 0.00002355, -0.00002020, -0.00003331, 0.00006805, -0.00004853, 0.00001015, 0.00002070, -0.00003686, 0.00005384, -0.00004681, 0.00000778, 0.00000197, -0.00002694, 0.00005124, -0.00004265, 0.00002924, -0.00000244, -0.00002926, 0.00004592, -0.00003552, 0.00001033, 0.00001062, -0.00002403, 0.00000323, -0.00000456, -0.00001185, 0.00000396, 0.00001062, -0.00002485, 0.00001409, -0.00001217, -0.00001372, 0.00000906, -0.00002717, 0.00000978, -0.00003090, -0.00001179, -0.00001715, -0.00002510, -0.00005187, -0.00020382, 0.00004924, 0.00024200, 0.00008443, 0.00006381, 0.00013846, 0.00019467, -0.00001453, -0.00016735, -0.00002960, -0.00001363, -0.00003297, -0.00000606, -0.00001011, -0.00001026, -0.00000418, -0.00000377, -0.00001027, 0.00000918, 0.00000084, -0.00001020, 0.00000295, -0.00000895, -0.00000097, 0.00000624, -0.00001319, -0.00000718, 0.00000292, -0.00000177, -0.00000369, },
		{ -0.00000251, -0.00000634, -0.00000545, 0.00000049, -0.00000126, 0.00000207, 0.00000127, -0.00000967, -0.00000773, 0.00000255, 0.00000051, -0.00000098, 0.00000188, -0.00000175, -0.00000270, 0.00000037, -0.00000534, -0.00000733, 0.00000016, 0.00000254, 0.00000745, -0.00000502, -0.00001154, 0.00000806, 0.00000173, -0.00000408, -0.00000022, 0.00000130, 0.00000906, 0.00000114, -0.00000392, -0.00000058, 0.00000421, 0.00000227, -0.00000060, 0.00000832, -0.00000664, -0.00000845, 0.00001101, -0.00000251, -0.00000361, 0.00000207, -0.00000206, -0.00000178, -0.00000619, 0.00000997, 0.00000139, -0.00000194, 0.00000749, -0.00000508, 0.00000163, -0.00000782, 0.00000958, -0.00000007, -0.00000864, 0.00001074, -0.00000684, 0.00002455, -0.00002420, 0.00000562, 0.00001600, -0.00005997, 0.00006075, -0.00001581, -0.00002562, 0.00004518, -0.00004406, 0.00004268, -0.00001947, -0.00001810, 0.00002104, -0.00003964, 0.00004677, -0.00002266, 0.00000503, 0.00001820, -0.00003753, 0.00004128, -0.00001962, -0.00000606, 0.00001819, -0.00002160, -0.00000313, -0.00000062, -0.00000954, -0.00000662, 0.00001819, -0.00002053, 0.00000190, 0.00000191, -0.00002220, 0.00001055, -0.00002178, 0.00000140, -0.00001915, -0.00001870, -0.00001549, -0.00002421, -0.00003078, -0.00019333, -0.00000618, 0.00022656, 0.00010249, 0.00006339, 0.00009114, 0.00018568, 0.00008420, -0.00015001, -0.00007680, -0.00000463, -0.00003194, -0.00001457, -0.00000541, -0.00001088, -0.00000756, -0.00000120, -0.00001207, 0.00000616, 0.00000587, -0.00001196, 0.00000219, -0.00000623, -0.00000469, 0.00000751, -0.00001063, -0.00000947, 0.00000272, -0.00000159, },
		{ -0.00000131, -0.00000224, -0.00000460, -0.00000750, -0.00000071, -0.00000066, 0.00000136, 0.00000398, -0.00000736, -0.00001147, 0.00000109, 0.00000278, -0.00000243, 0.00000185, 0.00000058, -0.00000491, 0.00000060, -0.00000159, -0.00001009, -0.00000122, 0.00000312, 0.00000649, -0.00000021, -0.00001432, 0.00000345, 0.00000688, -0.00000356, -0.00000157, 0.00000003, 0.00000654, 0.00000312, -0.00000101, -0.00000162, 0.00000241, 0.00000310, -0.00000325, 0.00000953, -0.00000096, -0.00001346, 0.00000995, 0.00000200, -0.00000661, 0.00000233, -0.00000173, -0.00000108, -0.00000700, 0.00000704, 0.00000533, -0.00000417, 0.00000701, -0.00000335, 0.00000266, -0.00000932, 0.00000677, 0.00000627, -0.00001434, 0.00001412, -0.00001365, 0.00002708, -0.00000933, -0.00002467, 0.00004450, -0.00006022, 0.00002845, 0.00002431, -0.00005101, 0.00004986, -0.00003115, 0.00001856, 0.00000869, -0.00003506, 0.00002798, -0.00003829, 0.00003026, 0.00000041, -0.00001286, 0.00002756, -0.00003526, 0.00003047, -0.00000344, -0.00001682, 0.00001899, -0.00001624, -0.00000766, -0.00000132, -0.00000362, -0.00001463, 0.00001687, -0.00000846, -0.00001112, 0.00001011, -0.00002205, 0.00000500, -0.00001342, -0.00000557, -0.00001263, -0.00001934, -0.00001602, -0.00002705, -0.00001081, -0.00017608, -0.00005692, 0.00020837, 0.00011255, 0.00006909, 0.00006683, 0.00014205, 0.00015603, -0.00008986, -0.00011845, -0.00001429, -0.00002230, -0.00002345, -0.00000516, -0.00000743, -0.00001188, -0.00000015, -0.00001137, 0.00000113, 0.00001042, -0.00001161, -0.00000048, -0.00000278, -0.00000745, 0.00000719, -0.00000738, -0.00001155, 0.00000239, },
		{ 0.00000157, -0.00000048, -0.00000288, -0.00000259, -0.00000820, -0.00000291, -0.00000024, 0.00000068, 0.00000543, -0.00000357, -0.00001362, -0.00000258, 0.00000499, -0.00000230, 0.00000002, 0.00000347, -0.00000551, -0.00000200, 0.00000296, -0.00001046, -0.00000490, 0.00000446, 0.00000520, 0.00000339, -0.00001348, -0.00000319, 0.00000959, -0.00000024, -0.00000307, -0.00000118, 0.00000463, 0.00000277, 0.00000187, -0.00000038, -0.00000025, 0.00000406, -0.00000512, 0.00000757, 0.00000625, -0.00001582, 0.00000551, 0.00000738, -0.00000811, 0.00000101, -0.00000083, -0.00000074, -0.00000684, 0.00000351, 0.00000827, -0.00000462, 0.00000577, -0.00000282, 0.00000457, -0.00000884, 0.00000043, 0.00001471, -0.00001980, 0.00001584, -0.00001439, 0.00001646, 0.00001748, -0.00005178, 0.00005010, -0.00003297, -0.00001469, 0.00005363, -0.00005480, 0.00003426, -0.00000648, -0.00000616, 0.00002732, -0.00003874, 0.00002362, -0.00002751, 0.00000972, 0.00001849, -0.00002064, 0.00002737, -0.00002747, 0.00001867, 0.00000959, -0.00002173, 0.00001604, -0.00001160, -0.00000894, -0.00000466, 0.00000119, -0.00001606, 0.00000763, 0.00000526, -0.00001843, 0.00000978, -0.00001545, -0.00000253, -0.00000726, -0.00000822, -0.00001067, -0.00001788, -0.00001404, -0.00003396, 0.00000418, -0.00014920, -0.00010425, 0.00018645, 0.00012162, 0.00006881, 0.00006613, 0.00008895, 0.00018348, -0.00000467, -0.00013421, -0.00004243, -0.00001256, -0.00002658, -0.00001062, -0.00000230, -0.00001420, -0.00000203, -0.00000856, -0.00000427, 0.00001283, -0.00000873, -0.00000426, 0.00000012, -0.00000852, 0.00000547, -0.00000393, -0.00001310, },
		{ -0.00001384, 0.00000004, 0.00000096, -0.00000391, -0.00000121, -0.00000741, -0.00000535, -0.00000060, 0.00000028, 0.00000577, 0.00000047, -0.00001333, -0.00000739, 0.00000552, -0.00000019, -0.00000259, 0.00000518, -0.00000344, -0.00000638, 0.00000586, -0.00000727, -0.00000993, 0.00000493, 0.00000504, 0.00000482, -0.00000991, -0.00000912, 0.00000837, 0.00000464, -0.00000301, -0.00000319, 0.00000362, 0.00000137, 0.00000285, 0.00000311, -0.00000248, 0.00000396, -0.00000520, 0.00000298, 0.00001257, -0.00001402, -0.00000131, 0.00001162, -0.00000708, -0.00000160, 0.00000001, -0.00000039, -0.00000613, 0.00000007, 0.00000997, -0.00000409, 0.00000497, -0.00000336, 0.00000535, -0.00000443, -0.00000917, 0.00002269, -0.00002129, 0.00001149, -0.00000570, -0.00000432, 0.00004389, -0.00006067, 0.00002840, 0.00000948, -0.00004896, 0.00005928, -0.00003677, 0.00000766, 0.00001739, -0.00002130, 0.00003243, -0.00003134, 0.00001308, -0.00001360, -0.00000813, 0.00002771, -0.00001951, 0.00002217, -0.00001875, 0.00000854, 0.00001911, -0.00002266, 0.00001228, -0.00000903, -0.00000847, -0.00000713, 0.00000160, -0.00001100, -0.00000452, 0.00001412, -0.00001710, 0.00000312, -0.00000752, -0.00000728, -0.00000538, -0.00000755, -0.00000994, -0.00001810, -0.00000857, -0.00004108, 0.00000947, -0.00011299, -0.00014435, 0.00015583, 0.00013576, 0.00005964, 0.00007772, 0.00004912, 0.00016721, 0.00007889, -0.00011327, -0.00007808, -0.00001270, -0.00002193, -0.00001888, 0.00000058, -0.00001262, -0.00000626, -0.00000541, -0.00000818, 0.00001235, -0.00000409, -0.00000774, 0.00000150, -0.00000790, 0.00000299, -0.00000086, },
		{ 0.00000147, -0.00001368, -0.00000220, 0.00000272, -0.00000465, -0.00000093, -0.00000559, -0.00000714, -0.00000185, -0.00000012, 0.00000552, 0.00000382, -0.00001085, -0.00001162, 0.00000348, 0.00000302, -0.00000431, 0.00000453, 0.00000080, -0.00001025, 0.00000514, -0.00000110, -0.00001404, 0.00000290, 0.00000649, 0.00000465, -0.00000570, -0.00001216, 0.00000372, 0.00000856, -0.00000015, -0.00000566, 0.00000258, 0.00000057, 0.00000122, 0.00000736, -0.00000261, 0.00000218, -0.00000365, -0.00000251, 0.00001579, -0.00000818, -0.00000825, 0.00001292, -0.00000359, -0.00000437, -0.00000001, 0.00000017, -0.00000501, -0.00000313, 0.00001081, -0.00000338, 0.00000466, -0.00000338, 0.00000285, 0.00000402, -0.00001905, 0.00002586, -0.00001525, -0.00000051, 0.00000989, -0.00002584, 0.00005659, -0.00004478, -0.00001166, 0.00004732, -0.00005896, 0.00004021, -0.00000735, -0.00001657, 0.00003063, -0.00002342, 0.00002684, -0.00001888, 0.00000201, -0.00000124, -0.00001996, 0.00002839, -0.00001335, 0.00001619, -0.00001128, 0.00000007, 0.00002622, -0.00002135, 0.00000885, -0.00000748, -0.00000898, -0.00000635, -0.00000243, -0.00000310, -0.00001370, 0.00001466, -0.00000898, -0.00000466, -0.00000289, -0.00000762, -0.00000636, -0.00000665, -0.00000758, -0.00002083, -0.00000228, -0.00004344, 0.00000343, -0.00007260, -0.00017053, 0.00011272, 0.00015512, 0.00004735, 0.00008700, 0.00003378, 0.00012317, 0.00013672, -0.00005957, -0.00010432, -0.00002800, -0.00001333, -0.00002478, -0.00000149, -0.00000778, -0.00001059, -0.00000391, -0.00000973, 0.00000959, 0.00000079, -0.00000973, 0.00000112, -0.00000622, 0.00000057, },
		{ -0.00000132, 0.00000303, -0.00001269, -0.00000498, 0.00000433, -0.00000458, -0.00000171, -0.00000354, -0.00000780, -0.00000356, -0.00000095, 0.00000499, 0.00000621, -0.00000716, -0.00001389, -0.00000071, 0.00000558, -0.00000383, 0.00000169, 0.00000537, -0.00001134, 0.00000051, 0.00000565, -0.00001479, -0.00000201, 0.00000850, 0.00000451, -0.00000278, -0.00001187, -0.00000207, 0.00000918, 0.00000514, -0.00000696, 0.00000027, 0.00000120, -0.00000218, 0.00001014, 0.00000029, -0.00000068, -0.00000170, -0.00000695, 0.00001493, -0.00000009, -0.00001279, 0.00001072, 0.00000106, -0.00000591, -0.00000119, 0.00000057, -0.00000326, -0.00000618, 0.00001104, -0.00000247, 0.00000377, -0.00000104, -0.00000346, 0.00001379, -0.00002446, 0.00002053, -0.00000121, -0.00001691, 0.00002483, -0.00003731, 0.00004849, -0.00000957, -0.00005214, 0.00006298, -0.00004122, 0.00000735, 0.00001847, -0.00002833, 0.00003062, -0.00001600, 0.00001698, -0.00000714, -0.00000634, 0.00000778, -0.00002545, 0.00002318, -0.00000604, 0.00001155, -0.00000482, -0.00000778, 0.00003163, -0.00001818, 0.00000516, -0.00000498, -0.00001222, -0.00000258, -0.00000809, 0.00000294, -0.00001628, 0.00000788, 0.00000085, -0.00000870, -0.00000311, -0.00000528, -0.00000710, -0.00000786, -0.00000324, -0.00002397, 0.00000094, -0.00003840, -0.00001084, -0.00003609, -0.00017759, 0.00005853, 0.00017359, 0.00004116, 0.00008439, 0.00004056, 0.00007364, 0.00015667, 0.00001022, -0.00010644, -0.00005510, -0.00000856, -0.00002446, -0.00000803, -0.00000251, -0.00001255, -0.00000487, -0.00000951, 0.00000599, 0.00000455, -0.00000989, -0.00000043, -0.00000433, },
		{ -0.00000280, -0.00000260, 0.00000394, -0.00001095, -0.00000805, 0.00000532, -0.00000348, -0.00000312, -0.00000193, -0.00000739, -0.00000507, -0.00000238, 0.00000409, 0.00000786, -0.00000332, -0.00001381, -0.00000557, 0.00000589, -0.00000115, -0.00000179, 0.00000810, -0.00000867, -0.00000624, 0.00000991, -0.00001108, -0.00000831, 0.00000911, 0.00000580, -0.00000167, -0.00000966, -0.00000626, 0.00000582, 0.00001066, -0.00000521, -0.00000364, 0.00000263, -0.00000540, 0.00000972, 0.00000566, -0.00000293, -0.00000074, -0.00000922, 0.00001070, 0.00000759, -0.00001337, 0.00000603, 0.00000494, -0.00000545, -0.00000294, 0.00000004, -0.00000069, -0.00000889, 0.00001034, -0.00000052, 0.00000095, 0.00000423, -0.00001146, 0.00002024, -0.00002132, 0.00000645, 0.00001677, -0.00003064, 0.00003067, -0.00003257, 0.00002252, 0.00003020, -0.00007499, 0.00004965, -0.00000556, -0.00002243, 0.00002925, -0.00002572, 0.00002200, -0.00000585, 0.00000851, 0.00000126, -0.00001173, 0.00001365, -0.00002572, 0.00001497, 0.00000012, 0.00000815, 0.00000185, -0.00001542, 0.00003473, -0.00001232, 0.00000013, -0.00000044, -0.00001774, 0.00000164, -0.00001168, 0.00000409, -0.00001245, -0.00000192, 0.00000722, -0.00000714, -0.00000622, -0.00000360, -0.00000538, -0.00001113, 0.00000099, -0.00002448, -0.00000112, -0.00002726, -0.00002699, -0.00001064, -0.00016528, 0.00000019, 0.00018248, 0.00004748, 0.00006919, 0.00005794, 0.00003706, 0.00014164, 0.00007429, -0.00007913, -0.00008288, -0.00001469, -0.00001826, -0.00001552, -0.00000008, -0.00001115, -0.00000736, -0.00000901, 0.00000286, 0.00000671, -0.00000872, -0.00000234, },
		{ -0.00000407, -0.00000171, -0.00000343, 0.00000428, -0.00000848, -0.00001110, 0.00000527, -0.00000140, -0.00000462, -0.00000110, -0.00000634, -0.00000588, -0.00000407, 0.00000247, 0.00000901, 0.00000014, -0.00001198, -0.00000937, 0.00000347, 0.00000238, -0.00000392, 0.00000781, -0.00000322, -0.00001217, 0.00000957, -0.00000401, -0.00001328, 0.00000672, 0.00000855, -0.00000128, -0.00000772, -0.00000728, 0.00000001, 0.00001346, 0.00000034, -0.00000789, 0.00000321, -0.00000668, 0.00000589, 0.00001147, -0.00000279, -0.00000137, -0.00000954, 0.00000495, 0.00001261, -0.00001018, 0.00000102, 0.00000661, -0.00000340, -0.00000400, -0.00000200, 0.00000226, -0.00001044, 0.00000783, 0.00000334, -0.00000406, 0.00001074, -0.00001711, 0.00001913, -0.00000899, -0.00001209, 0.00003136, -0.00003442, 0.00002293, -0.00001361, -0.00000983, 0.00005785, -0.00007074, 0.00001420, 0.00003030, -0.00003524, 0.00002285, -0.00001490, 0.00001250, 0.00000167, 0.00000343, 0.00000695, -0.00001540, 0.00001685, -0.00002186, 0.00000580, 0.00000451, 0.00000491, 0.00000927, -0.00002182, 0.00003397, -0.00000307, -0.00000637, 0.00000548, -0.00002320, 0.00000313, -0.00001093, 0.00000035, -0.00000542, -0.00000981, 0.00000759, -0.00000172, -0.00000859, -0.00000488, -0.00000146, -0.00001424, 0.00000247, -0.00002079, -0.00000758, -0.00001434, -0.00003863, 0.00000074, -0.00013888, -0.00005280, 0.00017543, 0.00006639, 0.00004807, 0.00007285, 0.00002158, 0.00010590, 0.00011505, -0.00002902, -0.00009751, -0.00003339, -0.00001108, -0.00001969, -0.00000194, -0.00000754, -0.00000947, -0.00000946, 0.00000059, 0.00000768, -0.00000702, },
		{ -0.00000528, -0.00000562, -0.00000089, -0.00000386, 0.00000393, -0.00000535, -0.00001361, 0.00000385, 0.00000136, -0.00000570, -0.00000108, -0.00000513, -0.00000599, -0.00000537, -0.00000003, 0.00000951, 0.00000328, -0.00000941, -0.00001108, -0.00000077, 0.00000474, -0.00000352, 0.00000502, 0.00000264, -0.00001469, 0.00000469, 0.00000348, -0.00001442, 0.00000138, 0.00001120, 0.00000015, -0.00000750, -0.00000566, -0.00000528, 0.00001153, 0.00000829, -0.00001006, 0.00000150, -0.00000557, 0.00000015, 0.00001523, 0.00000051, -0.00000285, -0.00000907, -0.00000045, 0.00001411, -0.00000492, -0.00000221, 0.00000594, -0.00000123, -0.00000324, -0.00000521, 0.00000435, -0.00000967, 0.00000292, 0.00000896, -0.00000966, 0.00001511, -0.00001652, 0.00000932, 0.00000862, -0.00002758, 0.00003551, -0.00002502, 0.00000401, 0.00001042, -0.00003452, 0.00006298, -0.00004262, -0.00002686, 0.00004989, -0.00002703, 0.00000661, -0.00000535, 0.00000757, 0.00000552, 0.00000024, 0.00001183, -0.00001799, 0.00001707, -0.00001462, -0.00000298, 0.00000730, 0.00000114, 0.00001664, -0.00002502, 0.00002819, 0.00000909, -0.00001299, 0.00001073, -0.00002587, -0.00000001, -0.00000610, -0.00000568, 0.00000071, -0.00001279, 0.00000289, 0.00000375, -0.00000744, -0.00000887, 0.00000223, -0.00001460, 0.00000003, -0.00001391, -0.00001518, -0.00000435, -0.00004229, 0.00000009, -0.00010678, -0.00009257, 0.00015152, 0.00009204, 0.00003017, 0.00007711, 0.00002448, 0.00006714, 0.00012559, 0.00002847, -0.00008924, -0.00005905, -0.00000972, -0.00001841, -0.00000677, -0.00000427, -0.00000962, -0.00001099, -0.00000140, 0.00000822, },
		{ 0.00000863, -0.00000337, -0.00000725, -0.00000037, -0.00000367, 0.00000275, -0.00000193, -0.00001497, 0.00000094, 0.00000428, -0.00000592, -0.00000174, -0.00000405, -0.00000573, -0.00000580, -0.00000303, 0.00000887, 0.00000633, -0.00000682, -0.00001080, -0.00000504, 0.00000449, -0.00000092, 0.00000162, 0.00000652, -0.00001298, -0.00000246, 0.00000811, -0.00001100, -0.00000488, 0.00001157, 0.00000365, -0.00000867, -0.00000361, -0.00000733, 0.00000514, 0.00001549, -0.00000801, -0.00000251, -0.00000324, -0.00000516, 0.00001528, 0.00000611, -0.00000354, -0.00000900, -0.00000455, 0.00001277, 0.00000013, -0.00000273, 0.00000428, -0.00000061, -0.00000059, -0.00000811, 0.00000400, -0.00000596, -0.00000385, 0.00001458, -0.00001293, 0.00001408, -0.00000843, -0.00000599, 0.00002426, -0.00003286, 0.00002638, -0.00000550, -0.00001780, 0.00002809, -0.00004199, 0.00004601, -0.00000413, -0.00005599, 0.00004593, -0.00000482, -0.00000765, -0.00000369, 0.00000725, 0.00000847, -0.00000345, 0.00001656, -0.00001856, 0.00001354, -0.00000500, -0.00001016, 0.00000867, -0.00000299, 0.00002222, -0.00002334, 0.00001769, 0.00002207, -0.00001740, 0.00001313, -0.00002410, -0.00000746, 0.00000034, -0.00001062, 0.00000322, -0.00001089, -0.00000359, 0.00000591, -0.00000259, -0.00001301, 0.00000300, -0.00001123, -0.00000508, -0.00000671, -0.00002041, 0.00000008, -0.00003848, -0.00000726, -0.00007685, -0.00011580, 0.00011518, 0.00011621, 0.00002203, 0.00006989, 0.00003652, 0.00003875, 0.00011070, 0.00007611, -0.00005724, -0.00008085, -0.00001880, -0.00001344, -0.00001130, -0.00000350, -0.00000769, -0.00001255, -0.00000394, },
		{ -0.00000728, 0.00000853, -0.00000088, -0.00000893, -0.00000056, -0.00000257, 0.00000094, 0.00000109, -0.00001459, -0.00000316, 0.00000660, -0.00000499, -0.00000285, -0.00000318, -0.00000548, -0.00000532, -0.00000571, 0.00000664, 0.00000919, -0.00000427, -0.00000945, -0.00000773, 0.00000157, 0.00000215, -0.00000034, 0.00000738, -0.00000833, -0.00000867, 0.00000803, -0.00000470, -0.00000907, 0.00000840, 0.00000847, -0.00000936, -0.00000350, -0.00000532, -0.00000305, 0.00001850, -0.00000136, -0.00000700, -0.00000171, -0.00000826, 0.00001175, 0.00001185, -0.00000196, -0.00000953, -0.00000760, 0.00001006, 0.00000342, -0.00000125, 0.00000363, -0.00000218, 0.00000255, -0.00000868, 0.00000032, -0.00000008, -0.00001046, 0.00001746, -0.00001116, 0.00000661, 0.00000457, -0.00002037, 0.00003079, -0.00002490, 0.00000732, 0.00001591, -0.00003262, 0.00003161, -0.00003152, 0.00001702, 0.00002829, -0.00006295, 0.00002301, 0.00001797, -0.00001085, -0.00000974, 0.00000693, 0.00001385, -0.00000808, 0.00001964, -0.00001541, 0.00000629, 0.00000510, -0.00001444, 0.00000864, -0.00000663, 0.00002433, -0.00001650, 0.00000450, 0.00003294, -0.00001753, 0.00001162, -0.00001813, -0.00001698, 0.00000525, -0.00001218, 0.00000180, -0.00000645, -0.00000839, 0.00000370, 0.00000358, -0.00001424, -0.00000021, -0.00000552, -0.00001009, -0.00000205, -0.00002149, -0.00000071, -0.00003052, -0.00001563, -0.00005367, -0.00012406, 0.00007353, 0.00013225, 0.00002509, 0.00005590, 0.00004797, 0.00002572, 0.00008269, 0.00010209, -0.00001005, -0.00008779, -0.00003748, -0.00000972, -0.00001259, -0.00000551, -0.00000510, -0.00001285, },
		{ -0.00001134, -0.00001074, 0.00000721, 0.00000227, -0.00001012, -0.00000201, -0.00000064, -0.00000081, 0.00000287, -0.00001231, -0.00000766, 0.00000752, -0.00000285, -0.00000408, -0.00000256, -0.00000541, -0.00000442, -0.00000721, 0.00000297, 0.00001121, -0.00000136, -0.00000803, -0.00000824, -0.00000265, 0.00000363, 0.00000010, 0.00000600, -0.00000329, -0.00001152, 0.00000385, 0.00000118, -0.00000901, 0.00000250, 0.00001233, -0.00000749, -0.00000627, -0.00000088, -0.00000928, 0.00001549, 0.00000782, -0.00000918, -0.00000242, -0.00000892, 0.00000632, 0.00001544, 0.00000218, -0.00000975, -0.00001046, 0.00000720, 0.00000478, 0.00000043, 0.00000529, -0.00000490, 0.00000387, -0.00000557, -0.00000592, 0.00000564, -0.00001411, 0.00001533, -0.00000372, -0.00000494, 0.00001691, -0.00002715, 0.00002485, -0.00000676, -0.00001345, 0.00002991, -0.00003412, 0.00002063, -0.00001074, -0.00001011, 0.00004326, -0.00004886, -0.00000596, 0.00002928, -0.00000167, -0.00001719, 0.00000137, 0.00002214, -0.00001125, 0.00001859, -0.00000790, -0.00000291, 0.00001308, -0.00001490, 0.00000738, -0.00000896, 0.00002219, -0.00000581, -0.00000835, 0.00003909, -0.00001254, 0.00000673, -0.00000978, -0.00002562, 0.00000634, -0.00001007, -0.00000185, -0.00000235, -0.00000974, -0.00000138, 0.00000780, -0.00001104, -0.00000601, -0.00000032, -0.00001242, -0.00000109, -0.00001900, -0.00000429, -0.00002222, -0.00002144, -0.00003789, -0.00012181, 0.00003311, 0.00013738, 0.00003636, 0.00004153, 0.00005315, 0.00002519, 0.00005482, 0.00010391, 0.00003833, -0.00007404, -0.00005915, -0.00001247, -0.00001019, -0.00000855, -0.00000370, },
		{ -0.00000441, -0.00000877, -0.00001317, 0.00000434, 0.00000548, -0.00000997, -0.00000490, 0.00000139, -0.00000154, 0.00000293, -0.00000871, -0.00001138, 0.00000649, 0.00000011, -0.00000498, -0.00000226, -0.00000540, -0.00000374, -0.00000716, -0.00000130, 0.00001153, 0.00000214, -0.00000689, -0.00000715, -0.00000623, 0.00000239, 0.00000220, 0.00000434, -0.00000001, -0.00001065, -0.00000174, 0.00000357, -0.00000476, -0.00000337, 0.00001275, -0.00000243, -0.00001065, 0.00000289, -0.00001076, 0.00000748, 0.00001576, -0.00000696, -0.00000508, -0.00000847, 0.00000117, 0.00001572, 0.00000763, -0.00000820, -0.00001359, 0.00000436, 0.00000521, 0.00000063, 0.00000894, -0.00000640, 0.00000148, 0.00000070, -0.00001215, 0.00000821, -0.00001276, 0.00000796, 0.00000706, -0.00001570, 0.00002272, -0.00002293, 0.00000879, 0.00001364, -0.00002679, 0.00003089, -0.00002267, 0.00000181, 0.00000908, -0.00002480, 0.00003912, -0.00002404, -0.00002727, 0.00002469, 0.00001326, -0.00001813, -0.00001094, 0.00002976, -0.00000925, 0.00001233, 0.00000227, -0.00001088, 0.00001674, -0.00001174, 0.00000552, -0.00000951, 0.00001620, 0.00000626, -0.00001790, 0.00003923, -0.00000323, 0.00000025, -0.00000146, -0.00003105, 0.00000308, -0.00000579, -0.00000539, -0.00000037, -0.00000812, -0.00000646, 0.00000785, -0.00000440, -0.00001137, 0.00000177, -0.00001126, -0.00000288, -0.00001509, -0.00000795, -0.00001598, -0.00002390, -0.00002761, -0.00011368, -0.00000231, 0.00013255, 0.00005104, 0.00003129, 0.00005150, 0.00003038, 0.00003582, 0.00008776, 0.00007459, -0.00004171, -0.00007426, -0.00002389, -0.00000671, -0.00001010, },
		{ -0.00000880, -0.00000647, -0.00000661, -0.00001381, 0.00000046, 0.00000767, -0.00000797, -0.00000860, 0.00000231, -0.00000056, 0.00000159, -0.00000495, -0.00001328, 0.00000360, 0.00000311, -0.00000507, -0.00000236, -0.00000523, -0.00000365, -0.00000592, -0.00000493, 0.00000966, 0.00000584, -0.00000565, -0.00000563, -0.00000778, -0.00000115, 0.00000406, 0.00000420, 0.00000093, -0.00000777, -0.00000560, 0.00000140, 0.00000111, -0.00000595, 0.00000885, 0.00000431, -0.00001389, 0.00000328, -0.00000719, -0.00000211, 0.00001892, -0.00000039, -0.00000763, -0.00000872, -0.00000237, 0.00001324, 0.00001226, -0.00000407, -0.00001641, 0.00000084, 0.00000589, -0.00000107, 0.00001261, -0.00000427, -0.00000464, 0.00000755, -0.00001503, 0.00000559, -0.00000654, -0.00000220, 0.00001677, -0.00002057, 0.00001894, -0.00000947, -0.00001025, 0.00002732, -0.00002725, 0.00001972, -0.00000483, -0.00001525, 0.00001912, -0.00002434, 0.00002309, -0.00000157, -0.00003341, 0.00000897, 0.00002386, -0.00000853, -0.00002601, 0.00003111, -0.00000014, 0.00000275, 0.00001150, -0.00001463, 0.00001555, -0.00000646, 0.00000391, -0.00000816, 0.00000761, 0.00001702, -0.00002226, 0.00003372, 0.00000840, -0.00000558, 0.00000483, -0.00003231, -0.00000343, -0.00000154, -0.00000724, -0.00000055, -0.00000545, -0.00000911, 0.00000381, 0.00000277, -0.00001338, 0.00000004, -0.00000786, -0.00000528, -0.00001184, -0.00001016, -0.00001215, -0.00002419, -0.00002045, -0.00010265, -0.00003173, 0.00012044, 0.00006523, 0.00002640, 0.00004593, 0.00003507, 0.00002774, 0.00006425, 0.00009123, 0.00000014, -0.00007480, -0.00004137, -0.00000644, },
		{ -0.00001265, -0.00000566, -0.00000801, -0.00000593, -0.00001299, -0.00000322, 0.00000802, -0.00000445, -0.00001179, 0.00000109, 0.00000195, -0.00000002, -0.00000231, -0.00001298, -0.00000032, 0.00000517, -0.00000409, -0.00000275, -0.00000488, -0.00000412, -0.00000433, -0.00000694, 0.00000599, 0.00000872, -0.00000368, -0.00000457, -0.00000716, -0.00000527, 0.00000383, 0.00000595, 0.00000073, -0.00000536, -0.00000597, -0.00000374, 0.00000486, -0.00000346, 0.00000225, 0.00000968, -0.00001356, -0.00000037, -0.00000096, -0.00000912, 0.00001585, 0.00000799, -0.00000747, -0.00001053, -0.00000455, 0.00000974, 0.00001432, 0.00000208, -0.00001743, -0.00000407, 0.00000709, -0.00000352, 0.00001376, 0.00000232, -0.00001214, 0.00001129, -0.00001224, -0.00000186, 0.00000183, -0.00001098, 0.00002099, -0.00001697, 0.00000703, 0.00000705, -0.00002387, 0.00002867, -0.00001568, 0.00000301, 0.00001056, -0.00002322, 0.00001705, -0.00001377, 0.00000611, 0.00000968, -0.00002600, -0.00000759, 0.00002288, 0.00000880, -0.00003627, 0.00002235, 0.00001402, -0.00000562, 0.00001616, -0.00001313, 0.00001108, -0.00000122, 0.00000309, -0.00000492, -0.00000187, 0.00002424, -0.00002096, 0.00002421, 0.00001995, -0.00000898, 0.00000814, -0.00002980, -0.00001126, 0.00000097, -0.00000721, -0.00000167, -0.00000364, -0.00000864, -0.00000218, 0.00000728, -0.00001097, -0.00000405, -0.00000459, -0.00000642, -0.00001012, -0.00001102, -0.00000978, -0.00002375, -0.00001509, -0.00008980, -0.00005563, 0.00010354, 0.00007713, 0.00002567, 0.00003987, 0.00003628, 0.00002740, 0.00004320, 0.00008860, 0.00003959, -0.00005811, -0.00005827, },
		{ -0.00006699, -0.00002530, -0.00000370, -0.00000749, -0.00000653, -0.00001191, -0.00000564, 0.00000659, -0.00000062, -0.00001311, -0.00000243, 0.00000479, -0.00000052, -0.00000144, -0.00001106, -0.00000399, 0.00000562, -0.00000229, -0.00000306, -0.00000448, -0.00000478, -0.00000317, -0.00000713, 0.00000167, 0.00000973, -0.00000082, -0.00000400, -0.00000530, -0.00000808, 0.00000092, 0.00000832, 0.00000142, -0.00000507, -0.00000342, -0.00000850, 0.00000371, 0.00000307, -0.00000352, 0.00001092, -0.00000915, -0.00000620, 0.00000431, -0.00001090, 0.00000814, 0.00001435, -0.00000312, -0.00001295, -0.00000682, 0.00000694, 0.00001350, 0.00000833, -0.00001522, -0.00001012, 0.00000788, -0.00000481, 0.00001089, 0.00001175, -0.00001713, 0.00000907, -0.00000404, -0.00001111, 0.00000819, -0.00001453, 0.00001770, -0.00000631, -0.00000762, 0.00001912, -0.00002652, 0.00001811, 0.00000138, -0.00001061, 0.00001709, -0.00002059, 0.00000706, -0.00000179, -0.00000343, 0.00000871, -0.00001339, -0.00001561, 0.00001023, 0.00002562, -0.00003521, 0.00000462, 0.00002754, -0.00000814, 0.00001501, -0.00000801, 0.00000613, 0.00000235, 0.00000270, 0.00000013, -0.00001043, 0.00002659, -0.00001484, 0.00001299, 0.00002933, -0.00000901, 0.00000854, -0.00002477, -0.00001859, 0.00000113, -0.00000621, -0.00000233, -0.00000350, -0.00000617, -0.00000724, 0.00000738, -0.00000538, -0.00000783, -0.00000334, -0.00000585, -0.00000938, -0.00001146, -0.00000784, -0.00002324, -0.00001150, -0.00007527, -0.00007438, 0.00008308, 0.00008652, 0.00002731, 0.00003521, 0.00003441, 0.00002972, 0.00003003, 0.00007317, 0.00006685, -0.00002804, },
		{ 0.00000698, -0.00006259, -0.00004059, -0.00000614, -0.00000508, -0.00000698, -0.00001175, -0.00000668, 0.00000442, 0.00000211, -0.00001199, -0.00000722, 0.00000632, 0.00000092, -0.00000192, -0.00000877, -0.00000633, 0.00000452, -0.00000039, -0.00000286, -0.00000417, -0.00000530, -0.00000281, -0.00000609, -0.00000187, 0.00000848, 0.00000231, -0.00000328, -0.00000346, -0.00000859, -0.00000363, 0.00000921, 0.00000428, -0.00000658, -0.00000040, -0.00000976, -0.00000246, 0.00001007, -0.00000491, 0.00000739, -0.00000264, -0.00001104, 0.00000585, -0.00000764, -0.00000044, 0.00001561, 0.00000447, -0.00001363, -0.00001053, 0.00000529, 0.00001108, 0.00001247, -0.00000958, -0.00001576, 0.00000667, -0.00000363, 0.00000458, 0.00002048, -0.00001612, 0.00000071, 0.00000635, -0.00001767, 0.00000898, -0.00001145, 0.00000846, 0.00000661, -0.00001838, 0.00002162, -0.00001814, 0.00000156, 0.00001539, -0.00001507, 0.00001390, -0.00001148, -0.00000357, 0.00000457, -0.00000347, 0.00000136, -0.00000502, -0.00001185, -0.00000713, 0.00003295, -0.00002162, -0.00001580, 0.00003389, -0.00000294, 0.00001007, -0.00000253, 0.00000300, 0.00000415, 0.00000175, 0.00000647, -0.00001623, 0.00002374, -0.00000565, 0.00000236, 0.00003517, -0.00000560, 0.00000684, -0.00001866, -0.00002428, -0.00000066, -0.00000548, -0.00000192, -0.00000441, -0.00000371, -0.00000938, 0.00000354, 0.00000063, -0.00000900, -0.00000444, -0.00000453, -0.00000848, -0.00001222, -0.00000607, -0.00002233, -0.00001017, -0.00005955, -0.00008751, 0.00005946, 0.00009351, 0.00003040, 0.00003194, 0.00003155, 0.00003081, 0.00002502, 0.00005368, 0.00007765, },
		{ 0.00007375, 0.00003769, -0.00004499, -0.00005255, -0.00001418, -0.00000290, -0.00000573, -0.00001266, -0.00000731, 0.00000284, 0.00000302, -0.00000909, -0.00001154, 0.00000533, 0.00000393, -0.00000249, -0.00000734, -0.00000709, 0.00000268, 0.00000076, -0.00000201, -0.00000386, -0.00000563, -0.00000303, -0.00000481, -0.00000378, 0.00000562, 0.00000450, -0.00000177, -0.00000225, -0.00000716, -0.00000781, 0.00000713, 0.00000886, -0.00000791, 0.00000040, -0.00000645, -0.00001089, 0.00001326, -0.00000049, 0.00000126, 0.00000261, -0.00001239, 0.00000330, -0.00000207, -0.00000598, 0.00001125, 0.00001214, -0.00001028, -0.00001561, 0.00000372, 0.00000901, 0.00001329, -0.00000198, -0.00001874, 0.00000253, -0.00000031, -0.00000264, 0.00002474, -0.00000802, -0.00001074, 0.00001405, -0.00001788, 0.00000334, -0.00000368, -0.00000232, 0.00001604, -0.00002076, 0.00001427, -0.00000383, -0.00001283, 0.00002019, -0.00000978, 0.00000540, -0.00000254, -0.00000907, 0.00000342, 0.00000200, -0.00000435, -0.00000588, -0.00000051, -0.00001948, 0.00002649, -0.00000089, -0.00003049, 0.00002953, 0.00000760, 0.00000554, 0.00000053, 0.00000210, 0.00000552, -0.00000063, 0.00001282, -0.00001774, 0.00001643, 0.00000435, -0.00000585, 0.00003696, 0.00000060, 0.00000414, -0.00001265, -0.00002795, -0.00000349, -0.00000574, -0.00000091, -0.00000502, -0.00000284, -0.00000833, -0.00000201, 0.00000431, -0.00000692, -0.00000659, -0.00000393, -0.00000675, -0.00001318, -0.00000498, -0.00002039, -0.00001113, -0.00004393, -0.00009410, 0.00003329, 0.00009748, 0.00003512, 0.00002909, 0.00002946, 0.00002944, 0.00002509, 0.00003720, },
		{ 0.00002708, 0.00006087, 0.00005759, -0.00001881, -0.00005596, -0.00002591, -0.00000385, -0.00000248, -0.00001339, -0.00000892, 0.00000254, 0.00000258, -0.00000589, -0.00001390, 0.00000181, 0.00000709, -0.00000178, -0.00000714, -0.00000698, 0.00000125, 0.00000076, -0.00000088, -0.00000319, -0.00000592, -0.00000344, -0.00000399, -0.00000412, 0.00000252, 0.00000479, 0.00000044, -0.00000138, -0.00000501, -0.00000990, 0.00000219, 0.00001313, -0.00000680, -0.00000216, -0.00000022, -0.00001741, 0.00001001, 0.00000795, -0.00000356, 0.00000394, -0.00000999, -0.00000125, 0.00000225, -0.00000653, 0.00000366, 0.00001611, -0.00000242, -0.00002009, 0.00000043, 0.00000856, 0.00001135, 0.00000498, -0.00001749, -0.00000386, 0.00000316, -0.00000767, 0.00002254, 0.00000494, -0.00002017, 0.00001497, -0.00001103, -0.00000626, 0.00000441, -0.00000969, 0.00001823, -0.00001452, 0.00000148, 0.00000905, -0.00001893, 0.00001490, 0.00000065, -0.00000188, 0.00000131, -0.00000816, -0.00000204, 0.00000659, -0.00000328, -0.00001424, 0.00000985, -0.00002001, 0.00000918, 0.00001762, -0.00003347, 0.00001620, 0.00001795, 0.00000518, 0.00000071, 0.00000184, 0.00000806, -0.00000417, 0.00001720, -0.00001417, 0.00000644, 0.00001288, -0.00001049, 0.00003490, 0.00000856, 0.00000154, -0.00000750, -0.00002980, -0.00000656, -0.00000687, -0.00000044, -0.00000419, -0.00000389, -0.00000546, -0.00000664, 0.00000433, -0.00000281, -0.00000786, -0.00000487, -0.00000452, -0.00001356, -0.00000513, -0.00001729, -0.00001359, -0.00003026, -0.00009368, 0.00000615, 0.00009708, 0.00004196, 0.00002606, 0.00002854, 0.00002653, 0.00002644, },
		{ 0.00002658, 0.00002302, 0.00004556, 0.00006475, 0.00000897, -0.00004876, -0.00003696, -0.00000968, 0.00000112, -0.00001212, -0.00001212, 0.00000291, 0.00000213, -0.00000379, -0.00001390, -0.00000296, 0.00000872, 0.00000070, -0.00000746, -0.00000721, 0.00000092, 0.00000000, -0.00000027, -0.00000184, -0.00000622, -0.00000380, -0.00000376, -0.00000372, 0.00000049, 0.00000317, 0.00000239, -0.00000006, -0.00000329, -0.00000943, -0.00000384, 0.00001467, -0.00000228, -0.00000694, 0.00000558, -0.00001866, 0.00000091, 0.00001616, -0.00000357, 0.00000110, -0.00000604, -0.00000473, 0.00000312, -0.00000310, -0.00000311, 0.00001410, 0.00000784, -0.00002107, -0.00000558, 0.00000930, 0.00000869, 0.00000893, -0.00001223, -0.00001016, 0.00000432, -0.00000864, 0.00001486, 0.00001802, -0.00002283, 0.00000805, 0.00000005, -0.00001511, 0.00000835, -0.00001072, 0.00001317, -0.00000347, -0.00001026, 0.00001474, -0.00001545, 0.00000380, 0.00000949, -0.00000327, -0.00000054, -0.00000396, -0.00000656, 0.00000588, 0.00000420, -0.00002356, 0.00001169, -0.00000862, -0.00001032, 0.00002554, -0.00002461, 0.00000016, 0.00002257, 0.00000981, 0.00000024, -0.00000000, 0.00001221, -0.00000720, 0.00001769, -0.00000609, -0.00000362, 0.00001804, -0.00001131, 0.00002983, 0.00001707, -0.00000002, -0.00000352, -0.00003022, -0.00000955, -0.00000814, -0.00000154, -0.00000180, -0.00000591, -0.00000276, -0.00000864, 0.00000126, 0.00000111, -0.00000694, -0.00000693, -0.00000290, -0.00001264, -0.00000653, -0.00001364, -0.00001619, -0.00001999, -0.00008702, -0.00001947, 0.00009102, 0.00005070, 0.00002328, 0.00002801, 0.00002372, },
		{ 0.00002206, 0.00002488, 0.00002266, 0.00003262, 0.00006130, 0.00003200, -0.00003283, -0.00004267, -0.00001949, 0.00000214, -0.00000784, -0.00001600, 0.00000240, 0.00000289, -0.00000312, -0.00001242, -0.00000722, 0.00000787, 0.00000419, -0.00000689, -0.00000859, 0.00000137, -0.00000043, -0.00000092, 0.00000010, -0.00000623, -0.00000426, -0.00000371, -0.00000353, 0.00000003, 0.00000068, 0.00000280, 0.00000216, -0.00000220, -0.00000734, -0.00000862, 0.00001218, 0.00000453, -0.00001117, 0.00000785, -0.00001398, -0.00001045, 0.00001964, 0.00000213, -0.00000334, -0.00000369, -0.00000524, 0.00000089, 0.00000117, -0.00000564, 0.00000680, 0.00001650, -0.00001645, -0.00001337, 0.00000932, 0.00000747, 0.00000916, -0.00000517, -0.00001354, 0.00000184, -0.00000606, 0.00000517, 0.00002625, -0.00001681, -0.00000393, 0.00001020, -0.00001873, 0.00000599, -0.00000599, 0.00000433, 0.00000676, -0.00001578, 0.00001179, -0.00000602, -0.00000651, 0.00001184, 0.00000151, -0.00000470, -0.00000114, -0.00000658, -0.00000011, 0.00001327, -0.00002694, 0.00000275, 0.00000822, -0.00002239, 0.00002010, -0.00000974, -0.00001129, 0.00001923, 0.00001663, 0.00000240, -0.00000434, 0.00001650, -0.00000735, 0.00001353, 0.00000434, -0.00001094, 0.00001892, -0.00000895, 0.00002296, 0.00002491, 0.00000008, -0.00000073, -0.00002954, -0.00001257, -0.00000870, -0.00000448, 0.00000124, -0.00000735, -0.00000170, -0.00000786, -0.00000301, 0.00000290, -0.00000392, -0.00000872, -0.00000284, -0.00001045, -0.00000842, -0.00001044, -0.00001779, -0.00001356, -0.00007604, -0.00004095, 0.00007899, 0.00006007, 0.00002185, 0.00002686, },
		{ 0.00002468, 0.00002163, 0.00002204, 0.00002344, 0.00002421, 0.00005151, 0.00004674, -0.00001275, -0.00004043, -0.00002980, -0.00000187, -0.00000155, -0.00001845, -0.00000033, 0.00000497, -0.00000303, -0.00001095, -0.00000978, 0.00000500, 0.00000700, -0.00000439, -0.00001082, 0.00000139, 0.00000056, -0.00000282, 0.00000188, -0.00000540, -0.00000505, -0.00000345, -0.00000399, 0.00000058, -0.00000109, 0.00000101, 0.00000471, -0.00000101, -0.00000514, -0.00001055, 0.00000637, 0.00001092, -0.00001218, 0.00000557, -0.00000566, -0.00001937, 0.00001606, 0.00001113, -0.00000549, -0.00000492, -0.00000326, -0.00000193, 0.00000301, -0.00000308, -0.00000230, 0.00001966, -0.00000654, -0.00002009, 0.00000652, 0.00000852, 0.00000701, 0.00000058, -0.00001234, -0.00000337, -0.00000256, -0.00000243, 0.00002686, -0.00000416, -0.00001550, 0.00001456, -0.00001528, -0.00000141, 0.00000101, -0.00000346, 0.00001173, -0.00001365, 0.00000321, 0.00000334, -0.00001092, 0.00000709, 0.00000873, -0.00000643, -0.00000263, -0.00000236, -0.00000755, 0.00001822, -0.00002131, -0.00001296, 0.00002139, -0.00002162, 0.00000541, 0.00000263, -0.00001386, 0.00001036, 0.00002094, 0.00000907, -0.00000966, 0.00001820, -0.00000304, 0.00000596, 0.00001393, -0.00001349, 0.00001593, -0.00000475, 0.00001561, 0.00003109, 0.00000213, 0.00000109, -0.00002798, -0.00001593, -0.00000809, -0.00000861, 0.00000342, -0.00000705, -0.00000255, -0.00000550, -0.00000643, 0.00000193, -0.00000018, -0.00000880, -0.00000443, -0.00000788, -0.00000977, -0.00000840, -0.00001803, -0.00001037, -0.00006318, -0.00005650, 0.00006206, 0.00006808, 0.00002278, },
		{ 0.00002631, 0.00002184, 0.00002182, 0.00001908, 0.00002364, 0.00002025, 0.00003960, 0.00005280, 0.00000655, -0.00003088, -0.00003616, -0.00001109, 0.00000384, -0.00001743, -0.00000529, 0.00000709, -0.00000208, -0.00001045, -0.00001077, 0.00000166, 0.00000766, -0.00000014, -0.00001246, -0.00000038, 0.00000319, -0.00000500, 0.00000238, -0.00000335, -0.00000616, -0.00000299, -0.00000478, 0.00000101, -0.00000102, -0.00000243, 0.00000618, 0.00000124, -0.00000383, -0.00000965, -0.00000045, 0.00001417, -0.00000905, 0.00000026, 0.00000245, -0.00002249, 0.00000660, 0.00001892, -0.00000251, -0.00000905, -0.00000143, -0.00000256, 0.00000113, 0.00000244, -0.00000874, 0.00001570, 0.00000555, -0.00002241, 0.00000020, 0.00001066, 0.00000509, 0.00000273, -0.00000726, -0.00000842, -0.00000125, -0.00000544, 0.00002065, 0.00001006, -0.00002124, 0.00001113, -0.00000655, -0.00000991, 0.00000596, -0.00000667, 0.00001030, -0.00000642, -0.00000567, 0.00000769, -0.00000830, -0.00000122, 0.00001326, -0.00000295, -0.00000770, 0.00000264, -0.00001182, 0.00001612, -0.00000888, -0.00002792, 0.00002410, -0.00000944, -0.00000995, 0.00000606, -0.00000873, 0.00000153, 0.00001912, 0.00001904, -0.00001253, 0.00001508, 0.00000518, -0.00000204, 0.00001964, -0.00001083, 0.00001073, -0.00000027, 0.00000884, 0.00003503, 0.00000604, 0.00000229, -0.00002561, -0.00001988, -0.00000657, -0.00001260, 0.00000344, -0.00000482, -0.00000444, -0.00000318, -0.00000782, -0.00000102, 0.00000247, -0.00000665, -0.00000670, -0.00000613, -0.00000995, -0.00000755, -0.00001727, -0.00000924, -0.00005051, -0.00006572, 0.00004235, 0.00007288, },
		{ 0.00007347, 0.00003178, 0.00001912, 0.00002197, 0.00001678, 0.00002252, 0.00001959, 0.00002855, 0.00005194, 0.00002176, -0.00001733, -0.00003545, -0.00002280, 0.00000493, -0.00001259, -0.00001073, 0.00000744, 0.00000052, -0.00001060, -0.00001140, -0.00000058, 0.00000590, 0.00000430, -0.00001177, -0.00000437, 0.00000643, -0.00000588, 0.00000084, -0.00000037, -0.00000704, -0.00000277, -0.00000520, 0.00000037, 0.00000084, -0.00000587, 0.00000515, 0.00000471, -0.00000316, -0.00000734, -0.00000566, 0.00001307, -0.00000317, -0.00000504, 0.00000714, -0.00001946, -0.00000478, 0.00002133, 0.00000550, -0.00001283, -0.00000270, 0.00000006, -0.00000302, 0.00000713, -0.00000945, 0.00000631, 0.00001520, -0.00001842, -0.00000802, 0.00001138, 0.00000554, 0.00000120, -0.00000117, -0.00001008, -0.00000375, -0.00000428, 0.00001137, 0.00002027, -0.00001847, 0.00000183, 0.00000300, -0.00001519, 0.00000610, -0.00000474, 0.00000469, 0.00000134, -0.00001031, 0.00000562, -0.00000159, -0.00000790, 0.00001186, 0.00000482, -0.00001272, 0.00000443, -0.00001057, 0.00000832, 0.00000462, -0.00003546, 0.00001534, 0.00000712, -0.00001779, -0.00000033, -0.00000159, -0.00000199, 0.00001116, 0.00002821, -0.00000962, 0.00000716, 0.00001450, -0.00000700, 0.00001999, -0.00000444, 0.00000558, 0.00000334, 0.00000319, 0.00003653, 0.00001145, 0.00000324, -0.00002249, -0.00002431, -0.00000493, -0.00001512, 0.00000088, -0.00000155, -0.00000604, -0.00000202, -0.00000724, -0.00000433, 0.00000283, -0.00000299, -0.00000814, -0.00000592, -0.00000913, -0.00000725, -0.00001620, -0.00000917, -0.00003923, -0.00006933, 0.00002221, },
		{ 0.00000350, 0.00006987, 0.00003806, 0.00001721, 0.00002168, 0.00001551, 0.00002001, 0.00002081, 0.00001995, 0.00004646, 0.00003194, -0.00000384, -0.00002756, -0.00003241, -0.00000008, -0.00000593, -0.00001411, 0.00000500, 0.00000410, -0.00001005, -0.00001300, -0.00000123, 0.00000301, 0.00000678, -0.00000794, -0.00000951, 0.00000834, -0.00000419, -0.00000251, 0.00000244, -0.00000686, -0.00000328, -0.00000480, -0.00000141, 0.00000331, -0.00000742, 0.00000124, 0.00000835, -0.00000198, -0.00000533, -0.00000784, 0.00000863, 0.00000262, -0.00000768, 0.00000753, -0.00001278, -0.00001350, 0.00001690, 0.00001523, -0.00001239, -0.00000815, 0.00000431, -0.00000625, 0.00000783, -0.00000446, -0.00000419, 0.00001858, -0.00000911, -0.00001475, 0.00000855, 0.00000858, -0.00000180, 0.00000236, -0.00000674, -0.00000897, -0.00000187, 0.00000364, 0.00002310, -0.00000858, -0.00000850, 0.00000872, -0.00001491, 0.00000165, 0.00000003, -0.00000121, 0.00000568, -0.00000923, -0.00000054, 0.00000444, -0.00000930, 0.00000504, 0.00001292, -0.00001369, 0.00000125, -0.00000509, -0.00000075, 0.00001353, -0.00003322, -0.00000025, 0.00001960, -0.00001459, -0.00001163, 0.00000084, 0.00000142, 0.00000092, 0.00003187, -0.00000002, -0.00000271, 0.00002110, -0.00000671, 0.00001579, 0.00000295, 0.00000234, 0.00000580, -0.00000128, 0.00003564, 0.00001788, 0.00000429, -0.00001874, -0.00002878, -0.00000424, -0.00001533, -0.00000367, 0.00000133, -0.00000636, -0.00000216, -0.00000569, -0.00000656, 0.00000092, 0.00000061, -0.00000764, -0.00000703, -0.00000818, -0.00000679, -0.00001520, -0.00000962, -0.00002970, -0.00006865, },
		{ -0.00006494, -0.00001272, 0.00006278, 0.00004409, 0.00001642, 0.00002076, 0.00001546, 0.00001638, 0.00002258, 0.00001457, 0.00003827, 0.00003777, 0.00000682, -0.00001531, -0.00003576, -0.00001002, -0.00000102, -0.00001370, 0.00000054, 0.00000681, -0.00000741, -0.00001580, -0.00000131, 0.00000108, 0.00000596, -0.00000187, -0.00001353, 0.00000721, 0.00000005, -0.00000622, 0.00000365, -0.00000510, -0.00000449, -0.00000380, -0.00000352, 0.00000485, -0.00000613, -0.00000439, 0.00001024, 0.00000079, -0.00000442, -0.00000734, 0.00000331, 0.00000579, -0.00000699, 0.00000521, -0.00000613, -0.00001683, 0.00000768, 0.00002190, -0.00000571, -0.00001582, 0.00000678, -0.00000571, 0.00000394, 0.00000319, -0.00001110, 0.00001464, 0.00000185, -0.00001665, 0.00000203, 0.00001216, -0.00000309, 0.00000121, 0.00000039, -0.00001359, -0.00000169, 0.00000035, 0.00001890, 0.00000380, -0.00001469, 0.00000826, -0.00000985, -0.00000468, 0.00000421, -0.00000423, 0.00000527, -0.00000426, -0.00000658, 0.00000613, -0.00000524, -0.00000353, 0.00001698, -0.00000896, -0.00000530, 0.00000082, -0.00000664, 0.00001537, -0.00002380, -0.00001543, 0.00002259, -0.00000301, -0.00002027, -0.00000499, 0.00000859, -0.00000602, 0.00002751, 0.00001373, -0.00001007, 0.00002219, -0.00000144, 0.00000975, 0.00000864, 0.00000162, 0.00000765, -0.00000465, 0.00003250, 0.00002476, 0.00000581, -0.00001462, -0.00003258, -0.00000536, -0.00001328, -0.00000885, 0.00000253, -0.00000525, -0.00000292, -0.00000431, -0.00000712, -0.00000217, 0.00000257, -0.00000510, -0.00000833, -0.00000801, -0.00000588, -0.00001421, -0.00001051, -0.00002189, },
		{ -0.00001574, -0.00005913, -0.00002600, 0.00005310, 0.00004913, 0.00001688, 0.00001901, 0.00001667, 0.00001224, 0.00002360, 0.00001269, 0.00002874, 0.00004027, 0.00001408, -0.00000289, -0.00003129, -0.00002096, -0.00000097, -0.00001005, -0.00000354, 0.00000683, -0.00000260, -0.00001851, -0.00000271, 0.00000159, 0.00000229, 0.00000400, -0.00001420, 0.00000271, 0.00000529, -0.00000828, 0.00000239, -0.00000212, -0.00000569, -0.00000295, -0.00000493, 0.00000448, -0.00000252, -0.00000950, 0.00000883, 0.00000508, -0.00000402, -0.00000578, -0.00000049, 0.00000554, -0.00000446, 0.00000289, -0.00000209, -0.00001517, -0.00000194, 0.00002191, 0.00000581, -0.00002153, 0.00000439, -0.00000094, -0.00000228, 0.00000927, -0.00001173, 0.00000577, 0.00000979, -0.00001246, -0.00000576, 0.00001320, -0.00000057, -0.00000400, 0.00000767, -0.00001401, -0.00000554, 0.00000130, 0.00001124, 0.00001329, -0.00001388, 0.00000271, -0.00000327, -0.00000950, 0.00000536, -0.00000348, 0.00000156, 0.00000105, -0.00000907, 0.00000287, 0.00000127, -0.00000936, 0.00001485, -0.00000031, -0.00001135, 0.00000346, -0.00000744, 0.00001167, -0.00001259, -0.00002453, 0.00001616, 0.00001015, -0.00002040, -0.00001732, 0.00001342, -0.00000577, 0.00001658, 0.00002659, -0.00001112, 0.00001760, 0.00000620, 0.00000501, 0.00001126, 0.00000263, 0.00000975, -0.00000678, 0.00002721, 0.00003147, 0.00000819, -0.00001049, -0.00003506, -0.00000862, -0.00000982, -0.00001316, 0.00000156, -0.00000343, -0.00000337, -0.00000371, -0.00000644, -0.00000495, 0.00000218, -0.00000155, -0.00000851, -0.00000895, -0.00000490, -0.00001283, -0.00001182, },
		{ -0.00001328, -0.00001135, -0.00005193, -0.00003618, 0.00004162, 0.00005272, 0.00001873, 0.00001623, 0.00001884, 0.00000861, 0.00002266, 0.00001411, 0.00001923, 0.00003983, 0.00001904, 0.00000642, -0.00002080, -0.00002814, -0.00000645, -0.00000605, -0.00000470, 0.00000384, 0.00000274, -0.00001903, -0.00000677, 0.00000436, -0.00000204, 0.00000697, -0.00001070, -0.00000359, 0.00000909, -0.00000728, -0.00000095, 0.00000078, -0.00000588, -0.00000290, -0.00000511, 0.00000241, 0.00000168, -0.00001193, 0.00000397, 0.00000941, -0.00000274, -0.00000475, -0.00000181, 0.00000318, -0.00000259, 0.00000243, -0.00000081, -0.00001142, -0.00000775, 0.00001501, 0.00001770, -0.00002109, -0.00000340, 0.00000565, -0.00000707, 0.00001070, -0.00000671, -0.00000352, 0.00001149, -0.00000403, -0.00001095, 0.00000961, 0.00000528, -0.00001008, 0.00001100, -0.00000855, -0.00001229, 0.00000360, 0.00000463, 0.00001647, -0.00000693, -0.00000418, 0.00000119, -0.00001078, 0.00000323, -0.00000040, -0.00000255, 0.00000367, -0.00000716, -0.00000299, 0.00000611, -0.00000982, 0.00000782, 0.00000822, -0.00001333, 0.00000159, -0.00000457, 0.00000637, -0.00000434, -0.00002601, 0.00000493, 0.00001807, -0.00001118, -0.00003001, 0.00001051, 0.00000136, 0.00000390, 0.00003358, -0.00000492, 0.00001002, 0.00001271, 0.00000356, 0.00001132, 0.00000378, 0.00001249, -0.00000708, 0.00002000, 0.00003718, 0.00001186, -0.00000670, -0.00003582, -0.00001368, -0.00000624, -0.00001554, -0.00000120, -0.00000197, -0.00000299, -0.00000368, -0.00000546, -0.00000638, -0.00000014, 0.00000143, -0.00000684, -0.00001048, -0.00000459, -0.00001080, },
		{ -0.00000835, -0.00001436, -0.00000875, -0.00004404, -0.00004318, 0.00002905, 0.00005439, 0.00002218, 0.00001249, 0.00002108, 0.00000679, 0.00001911, 0.00001778, 0.00001141, 0.00003632, 0.00002304, 0.00001169, -0.00000831, -0.00002848, -0.00001490, -0.00000536, -0.00000215, -0.00000029, 0.00000603, -0.00001590, -0.00001309, 0.00000740, -0.00000428, 0.00000565, -0.00000438, -0.00000893, 0.00000951, -0.00000338, -0.00000470, 0.00000206, -0.00000453, -0.00000365, -0.00000441, -0.00000019, 0.00000457, -0.00001083, -0.00000272, 0.00001161, 0.00000038, -0.00000463, -0.00000141, 0.00000103, -0.00000312, 0.00000364, -0.00000031, -0.00000890, -0.00000812, 0.00000453, 0.00002462, -0.00001292, -0.00001380, 0.00001005, -0.00000753, 0.00000727, 0.00000059, -0.00000893, 0.00000690, 0.00000439, -0.00001040, 0.00000197, 0.00001144, -0.00001310, 0.00000811, 0.00000128, -0.00001841, 0.00000366, 0.00000206, 0.00001339, 0.00000237, -0.00000823, 0.00000164, -0.00000887, -0.00000046, 0.00000263, -0.00000465, 0.00000271, -0.00000263, -0.00000788, 0.00000662, -0.00000552, -0.00000031, 0.00001271, -0.00001019, -0.00000304, -0.00000133, 0.00000302, -0.00000075, -0.00002244, -0.00000513, 0.00001757, 0.00000301, -0.00003608, -0.00000141, 0.00001081, -0.00000493, 0.00003242, 0.00000609, 0.00000355, 0.00001562, 0.00000524, 0.00001065, 0.00000382, 0.00001531, -0.00000482, 0.00001155, 0.00004086, 0.00001715, -0.00000332, -0.00003493, -0.00001970, -0.00000379, -0.00001573, -0.00000469, -0.00000172, -0.00000197, -0.00000353, -0.00000500, -0.00000634, -0.00000311, 0.00000254, -0.00000360, -0.00001147, -0.00000547, },
		{ -0.00000743, -0.00000614, -0.00001459, -0.00000772, -0.00003621, -0.00004702, 0.00001622, 0.00005359, 0.00002729, 0.00000845, 0.00002204, 0.00000776, 0.00001329, 0.00002183, 0.00000691, 0.00002988, 0.00002639, 0.00001422, 0.00000214, -0.00002222, -0.00002173, -0.00001008, 0.00000217, -0.00000243, 0.00000545, -0.00000959, -0.00001949, 0.00000808, -0.00000278, 0.00000089, 0.00000184, -0.00001083, 0.00000635, 0.00000160, -0.00000675, 0.00000095, -0.00000210, -0.00000440, -0.00000374, -0.00000209, 0.00000520, -0.00000708, -0.00000866, 0.00001018, 0.00000492, -0.00000446, -0.00000098, 0.00000083, -0.00000578, 0.00000443, 0.00000181, -0.00000919, -0.00000472, -0.00000441, 0.00002328, 0.00000066, -0.00002192, 0.00000905, -0.00000343, 0.00000154, 0.00000614, -0.00000864, -0.00000083, 0.00000833, -0.00000379, -0.00000635, 0.00001411, -0.00001072, 0.00000001, 0.00001139, -0.00002005, -0.00000044, 0.00000354, 0.00000708, 0.00000947, -0.00000721, -0.00000114, -0.00000595, -0.00000348, 0.00000387, -0.00000413, -0.00000053, 0.00000154, -0.00000924, 0.00000290, 0.00000041, -0.00000558, 0.00001171, -0.00000389, -0.00000705, -0.00000051, 0.00000270, -0.00000025, -0.00001810, -0.00001023, 0.00001039, 0.00001531, -0.00003168, -0.00001852, 0.00001627, -0.00000674, 0.00002463, 0.00001735, 0.00000145, 0.00001480, 0.00000813, 0.00001111, 0.00000269, 0.00001691, 0.00000032, 0.00000319, 0.00004145, 0.00002401, -0.00000003, -0.00003286, -0.00002569, -0.00000317, -0.00001428, -0.00000775, -0.00000287, -0.00000112, -0.00000261, -0.00000521, -0.00000550, -0.00000538, 0.00000146, 0.00000003, -0.00001090, },
		{ -0.00000843, -0.00000965, -0.00000489, -0.00001386, -0.00000774, -0.00002913, -0.00004804, 0.00000410, 0.00004990, 0.00003361, 0.00000540, 0.00002040, 0.00001155, 0.00000674, 0.00002403, 0.00000653, 0.00002166, 0.00002816, 0.00001607, 0.00000868, -0.00001268, -0.00002285, -0.00001903, 0.00000428, -0.00000014, 0.00000141, -0.00000260, -0.00002315, 0.00000480, 0.00000192, -0.00000457, 0.00000502, -0.00000862, 0.00000139, 0.00000515, -0.00000585, -0.00000193, -0.00000001, -0.00000411, -0.00000380, -0.00000276, 0.00000398, -0.00000270, -0.00001169, 0.00000535, 0.00000903, -0.00000277, -0.00000164, 0.00000260, -0.00000849, 0.00000245, 0.00000640, -0.00001111, -0.00000132, -0.00000770, 0.00001433, 0.00001431, -0.00002336, 0.00000231, 0.00000263, -0.00000288, 0.00000750, -0.00000407, -0.00000716, 0.00000562, 0.00000585, -0.00001081, 0.00001100, -0.00000370, -0.00000955, 0.00001723, -0.00001535, -0.00000776, 0.00000650, 0.00000154, 0.00001150, -0.00000186, -0.00000434, -0.00000439, -0.00000462, 0.00000322, -0.00000205, -0.00000377, 0.00000314, -0.00000689, -0.00000253, 0.00000436, -0.00000612, 0.00000683, 0.00000196, -0.00000779, -0.00000256, 0.00000386, 0.00000016, -0.00001601, -0.00001043, 0.00000151, 0.00002029, -0.00001821, -0.00003378, 0.00001329, -0.00000248, 0.00001452, 0.00002464, 0.00000448, 0.00001234, 0.00000996, 0.00001333, 0.00000161, 0.00001602, 0.00000764, -0.00000328, 0.00003833, 0.00003176, 0.00000387, -0.00003023, -0.00003091, -0.00000443, -0.00001216, -0.00000964, -0.00000483, -0.00000131, -0.00000084, -0.00000554, -0.00000484, -0.00000622, -0.00000113, 0.00000261, },
		{ 0.00000317, -0.00000468, -0.00001103, -0.00000493, -0.00001252, -0.00000813, -0.00002321, -0.00004692, -0.00000634, 0.00004335, 0.00003995, 0.00000488, 0.00001562, 0.00001692, 0.00000167, 0.00002287, 0.00000961, 0.00001379, 0.00002717, 0.00001831, 0.00001212, -0.00000418, -0.00001716, -0.00002789, 0.00000045, 0.00000635, -0.00000321, 0.00000190, -0.00002238, -0.00000186, 0.00000733, -0.00000771, 0.00000393, -0.00000390, -0.00000252, 0.00000556, -0.00000251, -0.00000469, 0.00000026, -0.00000239, -0.00000445, -0.00000269, 0.00000221, 0.00000039, -0.00001125, -0.00000090, 0.00001049, 0.00000103, -0.00000304, 0.00000474, -0.00000881, -0.00000312, 0.00001187, -0.00001169, -0.00000136, -0.00000450, 0.00000213, 0.00002229, -0.00001673, -0.00000719, 0.00000680, -0.00000360, 0.00000509, 0.00000125, -0.00000865, -0.00000236, 0.00001344, -0.00000829, 0.00000307, 0.00000439, -0.00001604, 0.00001621, -0.00000561, -0.00001505, 0.00000768, -0.00000059, 0.00000855, 0.00000471, -0.00000510, -0.00000511, -0.00000436, 0.00000185, -0.00000006, -0.00000534, 0.00000181, -0.00000276, -0.00000655, 0.00000439, -0.00000295, 0.00000153, 0.00000438, -0.00000501, -0.00000557, 0.00000392, 0.00000255, -0.00001633, -0.00000881, -0.00000413, 0.00001677, -0.00000147, -0.00004074, 0.00000187, 0.00000342, 0.00000653, 0.00002619, 0.00001069, 0.00001110, 0.00000961, 0.00001636, 0.00000234, 0.00001231, 0.00001538, -0.00000612, 0.00003177, 0.00003909, 0.00000919, -0.00002743, -0.00003513, -0.00000709, -0.00001030, -0.00001032, -0.00000664, -0.00000297, 0.00000114, -0.00000516, -0.00000494, -0.00000578, -0.00000393, },
		{ -0.00000579, 0.00000169, -0.00000088, -0.00001075, -0.00000595, -0.00001122, -0.00000845, -0.00001837, -0.00004456, -0.00001446, 0.00003469, 0.00004470, 0.00000792, 0.00000849, 0.00002164, -0.00000007, 0.00001829, 0.00001411, 0.00000847, 0.00002320, 0.00002028, 0.00001482, 0.00000055, -0.00000727, -0.00003136, -0.00001026, 0.00001364, -0.00000464, 0.00000208, -0.00001788, -0.00000937, 0.00001057, -0.00000695, -0.00000028, 0.00000034, -0.00000326, 0.00000309, 0.00000115, -0.00000532, -0.00000178, 0.00000001, -0.00000474, -0.00000285, 0.00000111, 0.00000147, -0.00000855, -0.00000592, 0.00000822, 0.00000583, -0.00000355, 0.00000530, -0.00000560, -0.00001068, 0.00001501, -0.00000825, -0.00000574, 0.00000240, -0.00000748, 0.00002142, -0.00000463, -0.00001472, 0.00000637, -0.00000087, 0.00000158, 0.00000391, -0.00000488, -0.00001132, 0.00001447, 0.00000089, -0.00000568, 0.00000922, -0.00001675, 0.00000913, 0.00000527, -0.00001870, 0.00000530, 0.00000076, 0.00000315, 0.00000905, -0.00000221, -0.00000703, -0.00000428, 0.00000107, 0.00000085, -0.00000501, -0.00000109, 0.00000058, -0.00000728, 0.00000111, 0.00000095, -0.00000108, 0.00000282, -0.00000089, -0.00000680, 0.00000136, 0.00000660, -0.00001682, -0.00000892, -0.00000449, 0.00000803, 0.00001165, -0.00003686, -0.00001347, 0.00000581, 0.00000273, 0.00002322, 0.00001680, 0.00001295, 0.00000782, 0.00001828, 0.00000591, 0.00000693, 0.00002125, -0.00000452, 0.00002310, 0.00004440, 0.00001639, -0.00002431, -0.00003859, -0.00001050, -0.00000916, -0.00001035, -0.00000748, -0.00000579, 0.00000227, -0.00000357, -0.00000566, -0.00000491, },
		},
		{ { 0.00000245, 0.00000385, -0.00000350, -0.00001548, -0.00000991, 0.00000304, 0.00000358, -0.00000263, -0.00000405, -0.00000129, -0.00000153, -0.00000639, -0.00000402, 0.00000810, 0.00000529, -0.00000831, -0.00000339, -0.00000196, -0.00001065, -0.00000275, -0.00001052, -0.00002719, -0.00000914, 0.00000027, 0.00000442, 0.00003112, 0.00001763, -0.00000718, 0.00001095, 0.00001259, 0.00000616, 0.00001413, -0.00000036, -0.00000149, 0.00001402, 0.00000168, -0.00000928, 0.00000066, 0.00002001, 0.00002032, -0.00001227, -0.00000736, 0.00002108, -0.00000744, -0.00002644, 0.00000516, 0.00001503, -0.00000229, -0.00000875, -0.00000058, -0.00000261, -0.00000769, 0.00001537, 0.00001101, -0.00002224, -0.00000396, 0.00000898, -0.00000751, 0.00000940, 0.00000497, -0.00001513, 0.00000421, 0.00000491, -0.00000418, 0.00000233, -0.00000418, 0.00000491, 0.00000421, -0.00001513, 0.00000497, 0.00000940, -0.00000751, 0.00000898, -0.00000396, -0.00002224, 0.00001101, 0.00001537, -0.00000769, -0.00000261, -0.00000058, -0.00000875, -0.00000229, 0.00001503, 0.00000516, -0.00002644, -0.00000744, 0.00002108, -0.00000736, -0.00001227, 0.00002032, 0.00002001, 0.00000066, -0.00000928, 0.00000168, 0.00001402, -0.00000149, -0.00000036, 0.00001413, 0.00000616, 0.00001259, 0.00001095, -0.00000718, 0.00001763, 0.00003112, 0.00000442, 0.00000027, -0.00000914, -0.00002719, -0.00001052, -0.00000275, -0.00001065, -0.00000196, -0.00000339, -0.00000831, 0.00000529, 0.00000810, -0.00000402, -0.00000639, -0.00000153, -0.00000129, -0.00000405, -0.00000263, 0.00000358, 0.00000304, -0.00000991, -0.00001548, -0.00000350, 0.00000385, },
		{ 0.00000196, 0.00000337, 0.00000290, 0.00000173, -0.00000937, -0.00001411, -0.00000432, 0.00000393, 0.00000219, -0.00000355, -0.00000421, -0.00000019, -0.00000298, -0.00000828, 0.00000209, 0.00001056, -0.00000309, -0.00000698, -0.00000031, -0.00000921, -0.00000579, -0.00000298, -0.00002482, -0.00001696, 0.00000258, -0.00000184, 0.00001983, 0.00002866, -0.00000330, 0.00000147, 0.00001531, 0.00000471, 0.00001243, 0.00000872, -0.00000513, 0.00000636, 0.00000790, -0.00000242, -0.00000645, 0.00000301, 0.00002704, 0.00000874, -0.00002164, 0.00000874, 0.00001821, -0.00002051, -0.00001755, 0.00001320, 0.00001261, -0.00000966, -0.00000984, 0.00000518, -0.00000638, -0.00000098, 0.00002157, -0.00000745, -0.00002025, 0.00000794, 0.00000109, -0.00000011, 0.00000873, -0.00000935, -0.00000261, 0.00000677, -0.00000344, 0.00000244, -0.00000320, -0.00000101, 0.00001165, -0.00001295, -0.00000834, 0.00001783, -0.00000377, -0.00000282, 0.00000952, -0.00001780, -0.00000895, 0.00001912, 0.00000481, -0.00000893, -0.00000347, -0.00000081, -0.00000590, 0.00000136, 0.00001705, -0.00000764, -0.00002744, 0.00000689, 0.00001217, -0.00001333, 0.00000093, 0.00002309, 0.00001704, -0.00000563, -0.00001021, 0.00001335, 0.00000961, -0.00000542, 0.00000934, 0.00001000, 0.00000871, 0.00001670, -0.00000226, 0.00000435, 0.00003338, 0.00001706, -0.00000045, -0.00000497, -0.00002455, -0.00001852, -0.00000274, -0.00000973, -0.00000559, -0.00000023, -0.00000861, -0.00000205, 0.00000848, 0.00000221, -0.00000550, -0.00000474, -0.00000124, -0.00000139, -0.00000448, -0.00000135, 0.00000561, -0.00000116, -0.00001571, -0.00001165, },
		{ -0.00001714, -0.00000463, 0.00000401, 0.00000265, 0.00000287, -0.00000286, -0.00001179, -0.00001025, -0.00000152, 0.00000433, 0.00000127, -0.00000544, -0.00000304, 0.00000076, -0.00000690, -0.00000580, 0.00000950, 0.00000484, -0.00000750, -0.00000148, -0.00000587, -0.00000959, 0.00000056, -0.00001686, -0.00002465, 0.00000078, -0.00000037, 0.00000630, 0.00003090, 0.00000777, -0.00000624, 0.00001349, 0.00000706, 0.00000707, 0.00001456, -0.00000064, -0.00000052, 0.00000596, 0.00000299, -0.00000072, -0.00000978, 0.00001224, 0.00002939, -0.00001120, -0.00001584, 0.00002232, 0.00000536, -0.00002552, -0.00000714, 0.00001975, 0.00000538, -0.00001915, -0.00000091, 0.00000644, -0.00001075, 0.00001205, 0.00001338, -0.00002107, -0.00000682, 0.00000850, -0.00000012, 0.00000528, -0.00000419, -0.00000546, 0.00000710, -0.00000273, 0.00000089, 0.00000149, -0.00000767, 0.00001189, -0.00000058, -0.00002093, 0.00001269, 0.00001112, -0.00001159, 0.00000716, -0.00000198, -0.00001888, 0.00000650, 0.00001592, -0.00000249, -0.00001125, -0.00000154, 0.00000230, -0.00000757, 0.00000775, 0.00001399, -0.00002142, -0.00001843, 0.00001270, 0.00000175, -0.00000969, 0.00000935, 0.00002631, 0.00001146, -0.00001519, 0.00000069, 0.00001892, -0.00000038, 0.00000113, 0.00001169, 0.00000725, 0.00001673, 0.00000727, -0.00000348, 0.00002574, 0.00002940, 0.00000500, -0.00000363, -0.00002016, -0.00002428, -0.00000597, -0.00000752, -0.00000928, 0.00000020, -0.00000545, -0.00000710, 0.00000420, 0.00000574, -0.00000115, -0.00000546, -0.00000412, -0.00000020, -0.00000208, -0.00000515, 0.00000173, 0.00000549, -0.00000866, },
		{ 0.00000070, -0.00001527, -0.00001264, 0.00000093, 0.00000354, 0.00000267, 0.00000054, -0.00000662, -0.00001053, -0.00000781, 0.00000016, 0.00000556, -0.00000130, -0.00000672, 0.00000002, -0.00000160, -0.00000951, 0.00000227, 0.00001016, -0.00000325, -0.00000381, -0.00000263, -0.00001137, -0.00000117, -0.00000680, -0.00002711, -0.00000676, 0.00000429, -0.00000234, 0.00002371, 0.00001990, -0.00000768, 0.00000712, 0.00001106, 0.00000253, 0.00001433, 0.00000692, -0.00000199, 0.00000209, 0.00000100, 0.00000663, -0.00000603, -0.00000979, 0.00002771, 0.00001609, -0.00002437, 0.00000133, 0.00002392, -0.00000793, -0.00002488, 0.00000506, 0.00002290, -0.00001085, -0.00001836, 0.00001177, -0.00000205, -0.00000578, 0.00001954, -0.00000425, -0.00001946, 0.00000331, 0.00000595, 0.00000409, -0.00000291, -0.00000779, 0.00000769, 0.00000033, -0.00000403, 0.00000646, -0.00000810, 0.00000272, 0.00001282, -0.00002036, -0.00000514, 0.00002201, -0.00000535, -0.00000517, 0.00000738, -0.00001252, -0.00000841, 0.00001316, 0.00000995, -0.00000956, -0.00001180, 0.00000645, -0.00000069, -0.00000774, 0.00001597, 0.00000222, -0.00002767, -0.00000725, 0.00001069, -0.00000265, -0.00000526, 0.00001778, 0.00002896, -0.00000303, -0.00001454, 0.00001570, 0.00001189, -0.00000254, 0.00000840, 0.00000809, 0.00001345, 0.00001532, -0.00000271, 0.00001360, 0.00003482, 0.00001595, -0.00000115, -0.00001622, -0.00002739, -0.00001138, -0.00000543, -0.00001145, -0.00000231, -0.00000177, -0.00000826, -0.00000121, 0.00000503, 0.00000277, -0.00000241, -0.00000611, -0.00000261, 0.00000068, -0.00000413, -0.00000400, 0.00000520, },
		{ -0.00000037, 0.00000545, -0.00000731, -0.00001630, -0.00000575, 0.00000279, 0.00000312, 0.00000178, -0.00000291, -0.00000694, -0.00000926, -0.00000715, 0.00000342, 0.00000548, -0.00000540, -0.00000484, 0.00000159, -0.00000660, -0.00000563, 0.00000880, 0.00000326, -0.00000423, -0.00000094, -0.00001012, -0.00000611, 0.00000041, -0.00002244, -0.00001655, 0.00000625, -0.00000358, 0.00001164, 0.00002717, -0.00000218, -0.00000110, 0.00001372, 0.00000162, 0.00000927, 0.00001285, 0.00000068, 0.00000132, -0.00000338, 0.00000457, 0.00000758, -0.00001745, 0.00000442, 0.00003339, -0.00000610, -0.00002086, 0.00001600, 0.00001713, -0.00001837, -0.00001980, 0.00002024, 0.00001338, -0.00002599, -0.00000263, 0.00001383, -0.00001048, 0.00000653, 0.00001293, -0.00001583, -0.00000979, 0.00000648, 0.00000826, -0.00000055, -0.00001205, 0.00000513, 0.00000779, -0.00000836, 0.00000523, -0.00000063, -0.00000796, 0.00001564, -0.00000612, -0.00002121, 0.00001674, 0.00001057, -0.00001138, 0.00000302, -0.00000114, -0.00001191, 0.00000119, 0.00001376, 0.00000291, -0.00001739, -0.00000327, 0.00001114, -0.00000917, 0.00000000, 0.00001752, -0.00001225, -0.00002547, -0.00000043, 0.00000847, -0.00000500, -0.00000005, 0.00003004, 0.00002050, -0.00001626, -0.00000000, 0.00001990, 0.00000317, 0.00000331, 0.00000820, 0.00001007, 0.00001863, 0.00000417, 0.00000389, 0.00003157, 0.00002753, 0.00000535, -0.00001220, -0.00002849, -0.00001804, -0.00000465, -0.00001135, -0.00000655, 0.00000012, -0.00000649, -0.00000499, 0.00000187, 0.00000352, 0.00000168, -0.00000432, -0.00000602, -0.00000010, -0.00000030, -0.00000589, },
		{ -0.00000294, -0.00000516, 0.00000330, 0.00000052, -0.00001321, -0.00001171, -0.00000126, 0.00000290, 0.00000296, -0.00000107, -0.00000402, -0.00000567, -0.00001062, -0.00000497, 0.00000748, 0.00000191, -0.00000766, -0.00000105, -0.00000123, -0.00000818, 0.00000228, 0.00000696, -0.00000167, -0.00000020, -0.00000677, -0.00001088, 0.00000202, -0.00001321, -0.00002347, 0.00000283, 0.00000036, 0.00000026, 0.00002693, 0.00000801, -0.00000791, 0.00001266, 0.00000509, 0.00000242, 0.00001470, 0.00000499, 0.00000272, -0.00000294, -0.00000516, 0.00001429, -0.00000503, -0.00001825, 0.00002398, 0.00002092, -0.00002000, -0.00000796, 0.00002261, 0.00000758, -0.00002732, -0.00000563, 0.00002908, -0.00000896, -0.00002414, 0.00001463, 0.00000343, -0.00000848, 0.00001339, -0.00000024, -0.00001584, -0.00000250, 0.00001059, 0.00000696, -0.00001378, -0.00000422, 0.00001510, -0.00000503, -0.00000306, 0.00000729, -0.00001015, 0.00000650, 0.00000953, -0.00002270, -0.00000084, 0.00002058, -0.00000417, -0.00000594, 0.00000218, -0.00000564, -0.00000603, 0.00000557, 0.00001278, -0.00000860, -0.00001749, 0.00001030, 0.00000461, -0.00001254, 0.00001070, 0.00000973, -0.00002210, -0.00002073, 0.00000467, 0.00000650, -0.00000842, 0.00001310, 0.00003632, 0.00000161, -0.00001426, 0.00001445, 0.00001340, 0.00000278, 0.00000642, 0.00000739, 0.00001755, 0.00001238, 0.00000044, 0.00002276, 0.00003483, 0.00001543, -0.00000631, -0.00002747, -0.00002522, -0.00000641, -0.00000918, -0.00001058, -0.00000079, -0.00000370, -0.00000640, -0.00000129, 0.00000160, 0.00000342, 0.00000007, -0.00000645, -0.00000392, 0.00000148, },
		{ -0.00000092, 0.00000079, -0.00000510, -0.00000189, 0.00000310, -0.00000664, -0.00001302, -0.00000635, 0.00000008, 0.00000353, 0.00000104, -0.00000335, -0.00000194, -0.00000709, -0.00001240, 0.00000065, 0.00000862, -0.00000329, -0.00000605, 0.00000016, -0.00000507, -0.00000337, 0.00000540, 0.00000135, 0.00000091, -0.00000236, -0.00001303, -0.00000149, -0.00000383, -0.00002458, -0.00000490, 0.00000563, -0.00000656, 0.00001931, 0.00001943, -0.00001005, 0.00000631, 0.00001145, -0.00000220, 0.00001137, 0.00000969, 0.00000359, 0.00000250, -0.00001166, 0.00000621, 0.00001285, -0.00002053, -0.00000200, 0.00003050, 0.00000212, -0.00002058, 0.00000379, 0.00002499, -0.00000624, -0.00002939, 0.00001568, 0.00001947, -0.00002728, -0.00000529, 0.00001810, -0.00000690, -0.00000006, 0.00000982, -0.00000820, -0.00001171, 0.00000410, 0.00001470, -0.00000671, -0.00001631, 0.00001345, 0.00000682, -0.00001021, 0.00000688, -0.00000298, -0.00000472, 0.00001462, -0.00001083, -0.00001615, 0.00001618, 0.00000848, -0.00000748, -0.00000255, -0.00000129, -0.00000323, -0.00000335, 0.00000942, 0.00000658, -0.00001960, -0.00000510, 0.00001579, -0.00000697, -0.00000634, 0.00001558, -0.00000174, -0.00002758, -0.00001471, 0.00001035, -0.00000031, -0.00000527, 0.00003083, 0.00002654, -0.00001174, -0.00000035, 0.00001744, 0.00000850, 0.00000573, 0.00000509, 0.00001356, 0.00001829, 0.00000312, 0.00001312, 0.00003576, 0.00002633, 0.00000230, -0.00002355, -0.00003149, -0.00001159, -0.00000637, -0.00001259, -0.00000402, -0.00000167, -0.00000604, -0.00000324, -0.00000083, 0.00000201, 0.00000348, -0.00000297, -0.00000675, },
		{ -0.00000559, -0.00000492, 0.00000107, -0.00000158, -0.00000469, 0.00000077, -0.00000185, -0.00001021, -0.00000881, -0.00000415, 0.00000127, 0.00000351, -0.00000243, -0.00000179, -0.00000067, -0.00001195, -0.00001025, 0.00000682, 0.00000542, -0.00000611, -0.00000352, -0.00000198, -0.00000407, 0.00000157, 0.00000161, 0.00000171, 0.00000259, -0.00001132, -0.00000773, 0.00000201, -0.00001978, -0.00001414, 0.00000873, -0.00000676, 0.00000645, 0.00002740, -0.00000451, -0.00000441, 0.00001610, -0.00000061, 0.00000382, 0.00001295, 0.00000419, 0.00000724, -0.00000812, -0.00000863, 0.00001829, -0.00000403, -0.00002043, 0.00001597, 0.00002173, -0.00000912, -0.00001639, 0.00001457, 0.00002235, -0.00002360, -0.00001645, 0.00002895, -0.00000436, -0.00002562, 0.00001342, 0.00000894, -0.00000880, 0.00000472, 0.00000335, -0.00001043, -0.00000708, 0.00001354, 0.00000712, -0.00002066, 0.00000013, 0.00001789, -0.00000713, -0.00000142, 0.00000467, -0.00000827, 0.00000887, 0.00000267, -0.00001944, 0.00000314, 0.00001461, -0.00000070, -0.00000556, -0.00000430, 0.00000118, -0.00000240, -0.00000156, 0.00001218, -0.00000663, -0.00001910, 0.00001020, 0.00000877, -0.00001246, 0.00000308, 0.00001343, -0.00001315, -0.00003028, -0.00000440, 0.00001074, -0.00000824, 0.00001043, 0.00003831, 0.00000814, -0.00000959, 0.00001043, 0.00001435, 0.00000911, 0.00000458, 0.00000804, 0.00001988, 0.00000962, 0.00000643, 0.00003093, 0.00003516, 0.00001310, -0.00001639, -0.00003493, -0.00001971, -0.00000528, -0.00001190, -0.00000801, -0.00000142, -0.00000491, -0.00000414, -0.00000224, -0.00000077, 0.00000360, 0.00000161, },
		{ 0.00000387, -0.00000137, -0.00000639, -0.00000221, 0.00000095, -0.00000342, -0.00000227, -0.00000088, -0.00000704, -0.00000801, -0.00000620, -0.00000354, 0.00000340, 0.00000068, -0.00000350, 0.00000226, -0.00000435, -0.00001577, -0.00000373, 0.00000942, 0.00000097, -0.00000620, -0.00000350, -0.00000227, 0.00000033, -0.00000026, -0.00000010, 0.00000646, -0.00000521, -0.00001337, 0.00000220, -0.00001096, -0.00002156, 0.00000674, -0.00000030, -0.00000679, 0.00002659, 0.00000876, -0.00001399, 0.00001320, 0.00000748, -0.00000331, 0.00001182, 0.00000629, 0.00000761, 0.00000182, -0.00001676, 0.00000761, 0.00001342, -0.00001837, -0.00000595, 0.00002081, 0.00001144, -0.00001454, -0.00001107, 0.00002686, 0.00000842, -0.00003318, 0.00000655, 0.00002190, -0.00002263, -0.00000791, 0.00001872, -0.00000127, -0.00000627, 0.00000493, -0.00000052, -0.00001146, 0.00000238, 0.00001672, -0.00001155, -0.00001585, 0.00001712, 0.00000525, -0.00000800, 0.00000498, -0.00000437, 0.00000092, 0.00000894, -0.00001328, -0.00000759, 0.00001157, 0.00000578, -0.00000166, -0.00000804, -0.00000197, 0.00000490, -0.00000582, 0.00000379, 0.00000725, -0.00001755, -0.00000637, 0.00001572, -0.00000206, -0.00000996, 0.00000962, 0.00000709, -0.00002500, -0.00002637, 0.00000637, 0.00000227, -0.00000547, 0.00002906, 0.00002956, -0.00000254, -0.00000014, 0.00001304, 0.00001422, 0.00000821, 0.00000364, 0.00001643, 0.00001656, 0.00000484, 0.00002254, 0.00003965, 0.00002501, -0.00000635, -0.00003431, -0.00002870, -0.00000770, -0.00000958, -0.00001101, -0.00000292, -0.00000390, -0.00000456, -0.00000261, -0.00000279, 0.00000109, },
		{ -0.00000174, 0.00000283, 0.00000238, -0.00000384, -0.00000544, -0.00000013, -0.00000069, -0.00000286, -0.00000166, -0.00000593, -0.00000696, -0.00000465, -0.00000692, -0.00000109, 0.00000361, -0.00000293, 0.00000023, 0.00000319, -0.00001141, -0.00001440, 0.00000309, 0.00000843, -0.00000199, -0.00000689, -0.00000348, 0.00000192, -0.00000004, -0.00000443, 0.00000588, 0.00000379, -0.00001438, -0.00000329, -0.00000177, -0.00002357, -0.00000158, 0.00000917, -0.00001339, 0.00001474, 0.00002370, -0.00001449, 0.00000123, 0.00001613, -0.00000398, 0.00000562, 0.00000911, 0.00000537, 0.00000996, -0.00001267, -0.00000840, 0.00001748, -0.00000344, -0.00001609, 0.00000507, 0.00001794, 0.00000581, -0.00002140, 0.00000165, 0.00003287, -0.00001397, -0.00002500, 0.00002206, 0.00000173, -0.00002231, 0.00000919, 0.00001309, -0.00000724, -0.00000412, 0.00000533, -0.00000447, -0.00000843, 0.00001355, 0.00000414, -0.00002161, 0.00000380, 0.00001591, -0.00000518, -0.00000015, -0.00000022, -0.00000294, 0.00000876, -0.00000558, -0.00001183, 0.00000566, 0.00000643, 0.00000369, -0.00000456, -0.00000993, 0.00000636, 0.00000163, -0.00000658, 0.00000800, -0.00000433, -0.00001649, 0.00000684, 0.00001093, -0.00000839, -0.00000444, 0.00001293, -0.00000391, -0.00003378, -0.00001435, 0.00000829, -0.00000577, 0.00001009, 0.00003595, 0.00001597, -0.00000184, 0.00000548, 0.00001482, 0.00001489, 0.00000414, 0.00000969, 0.00002008, 0.00000849, 0.00001400, 0.00003830, 0.00003632, 0.00000629, -0.00002952, -0.00003625, -0.00001370, -0.00000768, -0.00001225, -0.00000532, -0.00000345, -0.00000491, -0.00000272, -0.00000333, },
		{ -0.00000305, -0.00000327, 0.00000023, 0.00000306, 0.00000039, -0.00000530, -0.00000337, 0.00000013, -0.00000149, -0.00000149, -0.00000564, -0.00000823, -0.00000253, -0.00000538, -0.00000695, 0.00000224, 0.00000044, -0.00000253, 0.00000500, -0.00000182, -0.00001630, -0.00000902, 0.00000730, 0.00000660, -0.00000510, -0.00000823, 0.00000128, 0.00000420, -0.00000669, -0.00000082, 0.00001057, -0.00000801, -0.00001096, 0.00000291, -0.00001829, -0.00001345, 0.00001459, -0.00000871, -0.00000305, 0.00003024, -0.00000250, -0.00001312, 0.00001688, 0.00000323, -0.00000142, 0.00000959, 0.00000381, 0.00001192, -0.00000114, -0.00001757, 0.00000842, 0.00000907, -0.00001175, -0.00000689, 0.00000627, 0.00001936, -0.00000376, -0.00002475, 0.00002206, 0.00002259, -0.00003003, -0.00000536, 0.00001983, -0.00001338, -0.00000920, 0.00001616, 0.00000433, -0.00001070, -0.00000049, 0.00000502, -0.00000832, 0.00000185, 0.00001317, -0.00001369, -0.00001109, 0.00001542, 0.00000411, -0.00000289, 0.00000023, -0.00000398, 0.00000681, 0.00000032, -0.00001274, 0.00000169, 0.00000477, 0.00000329, 0.00000381, -0.00001186, -0.00000265, 0.00001062, -0.00000570, -0.00000212, 0.00000473, -0.00001215, -0.00000663, 0.00001178, 0.00000369, -0.00001037, 0.00000202, 0.00001183, -0.00001903, -0.00003268, -0.00000275, 0.00000127, -0.00000193, 0.00002580, 0.00003066, 0.00000923, 0.00000099, 0.00000859, 0.00001879, 0.00001079, 0.00000451, 0.00001786, 0.00001466, 0.00000906, 0.00003144, 0.00004426, 0.00002127, -0.00002053, -0.00004098, -0.00002201, -0.00000767, -0.00001220, -0.00000782, -0.00000351, -0.00000525, -0.00000319, },
		{ -0.00000407, -0.00000288, -0.00000364, -0.00000167, 0.00000110, 0.00000275, -0.00000183, -0.00000528, -0.00000166, -0.00000061, -0.00000020, -0.00000371, -0.00001056, -0.00000403, -0.00000055, -0.00000864, -0.00000360, 0.00000247, -0.00000218, 0.00000233, 0.00000500, -0.00000911, -0.00001650, -0.00000323, 0.00001008, 0.00000377, -0.00000995, -0.00000497, 0.00000793, -0.00000220, -0.00000946, 0.00000910, 0.00000341, -0.00001435, 0.00000025, -0.00000837, -0.00002227, 0.00001038, 0.00000429, -0.00001599, 0.00002231, 0.00001568, -0.00001913, 0.00000680, 0.00001245, -0.00000365, 0.00000673, 0.00000376, 0.00000862, 0.00001002, -0.00001583, -0.00000462, 0.00001215, -0.00000354, -0.00000670, -0.00000636, 0.00001004, 0.00002123, -0.00002129, -0.00001321, 0.00003678, -0.00000030, -0.00002946, 0.00000987, 0.00000744, -0.00001489, 0.00000404, 0.00001467, -0.00000481, -0.00001080, 0.00000496, 0.00000094, -0.00000569, 0.00001008, -0.00000126, -0.00001657, 0.00000569, 0.00001050, 0.00000002, 0.00000013, -0.00000533, 0.00000396, 0.00000591, -0.00001253, -0.00000236, 0.00000619, -0.00000130, 0.00000723, -0.00000341, -0.00001294, 0.00000891, 0.00000520, -0.00000879, 0.00000121, -0.00000236, -0.00001074, 0.00000187, 0.00001078, -0.00000230, -0.00000961, 0.00000887, 0.00000287, -0.00003085, -0.00002292, -0.00000043, -0.00000309, 0.00001125, 0.00003158, 0.00002426, 0.00000719, 0.00000172, 0.00001486, 0.00001901, 0.00000585, 0.00001176, 0.00001877, 0.00000957, 0.00002211, 0.00004584, 0.00003690, -0.00000665, -0.00004200, -0.00003147, -0.00000997, -0.00001164, -0.00001024, -0.00000395, -0.00000537, },
		{ -0.00000508, -0.00000505, -0.00000309, -0.00000388, -0.00000223, -0.00000101, 0.00000171, 0.00000195, -0.00000384, -0.00000396, -0.00000153, 0.00000029, -0.00000008, -0.00001020, -0.00000942, 0.00000190, -0.00000423, -0.00000889, -0.00000021, -0.00000009, 0.00000002, 0.00000608, -0.00000009, -0.00001394, -0.00001421, 0.00000256, 0.00001197, -0.00000277, -0.00001211, 0.00000451, 0.00000690, -0.00001181, -0.00000114, 0.00001179, -0.00000891, -0.00000677, -0.00000042, -0.00002264, -0.00000212, 0.00001599, -0.00001582, 0.00000417, 0.00002788, -0.00001154, -0.00000782, 0.00001566, 0.00000011, 0.00000297, 0.00000460, 0.00000287, 0.00001559, -0.00000585, -0.00001424, 0.00000849, 0.00000100, -0.00000266, -0.00000488, -0.00000996, 0.00002364, 0.00000955, -0.00003388, 0.00001166, 0.00003331, -0.00002077, -0.00001733, 0.00001356, -0.00000220, -0.00000796, 0.00001163, 0.00000778, -0.00001243, -0.00000518, 0.00000715, -0.00000240, 0.00000234, 0.00000478, -0.00001258, -0.00000373, 0.00000951, 0.00000395, 0.00000304, -0.00000545, -0.00000179, 0.00001066, -0.00000850, -0.00000943, 0.00000895, -0.00000186, 0.00000140, 0.00000709, -0.00001288, -0.00000288, 0.00001297, -0.00000343, -0.00000734, -0.00000008, -0.00000520, -0.00000570, 0.00000581, 0.00000826, -0.00000788, -0.00000476, 0.00001143, -0.00001162, -0.00003266, -0.00001466, -0.00000376, 0.00000247, 0.00002229, 0.00003145, 0.00002114, 0.00000343, 0.00000567, 0.00002124, 0.00001357, 0.00000746, 0.00001791, 0.00001366, 0.00001485, 0.00004034, 0.00004933, 0.00001229, -0.00003756, -0.00004128, -0.00001483, -0.00001085, -0.00001263, -0.00000499, },
		{ -0.00000715, -0.00000452, -0.00000583, -0.00000337, -0.00000447, -0.00000263, -0.00000134, -0.00000110, 0.00000290, -0.00000013, -0.00000431, -0.00000326, -0.00000121, 0.00000259, -0.00000550, -0.00001423, -0.00000183, 0.00000204, -0.00000857, -0.00000563, -0.00000036, -0.00000021, 0.00000465, 0.00000467, -0.00000565, -0.00001634, -0.00001065, 0.00000963, 0.00000945, -0.00001103, -0.00000565, 0.00001174, -0.00000405, -0.00001213, 0.00000959, 0.00000225, -0.00001062, 0.00000083, -0.00001558, -0.00001515, 0.00001786, -0.00000396, -0.00001250, 0.00002620, 0.00000457, -0.00001708, 0.00001030, 0.00000578, 0.00000057, 0.00000650, -0.00000216, 0.00001318, 0.00000755, -0.00001714, 0.00000148, 0.00000390, -0.00000443, 0.00000509, -0.00001578, -0.00000055, 0.00003353, -0.00001583, -0.00002665, 0.00003349, 0.00001426, -0.00002898, -0.00000461, 0.00001086, -0.00000498, 0.00000031, 0.00001300, -0.00000266, -0.00001378, 0.00000228, 0.00000513, 0.00000045, 0.00000363, -0.00000748, -0.00000737, 0.00000549, 0.00000387, 0.00000636, -0.00000056, -0.00000855, 0.00000989, 0.00000107, -0.00001601, 0.00000605, 0.00000498, -0.00000647, 0.00000852, -0.00000240, -0.00001251, 0.00000876, 0.00000758, -0.00000850, -0.00000642, -0.00000118, -0.00000374, -0.00000278, 0.00000852, 0.00000361, -0.00001110, 0.00000265, 0.00000559, -0.00002294, -0.00002831, -0.00001305, -0.00000269, 0.00001222, 0.00002782, 0.00003229, 0.00001591, 0.00000070, 0.00001460, 0.00002120, 0.00000951, 0.00001388, 0.00001718, 0.00001243, 0.00003058, 0.00005435, 0.00003386, -0.00002518, -0.00004968, -0.00002306, -0.00001006, -0.00001456, },
		{ -0.00001506, -0.00001055, -0.00000421, -0.00000633, -0.00000360, -0.00000472, -0.00000419, -0.00000042, -0.00000264, 0.00000047, 0.00000283, -0.00000211, -0.00000389, -0.00000368, 0.00000184, 0.00000098, -0.00001368, -0.00000962, 0.00000393, -0.00000314, -0.00000852, -0.00000382, -0.00000111, 0.00000390, 0.00000581, 0.00000071, -0.00000965, -0.00001810, -0.00000365, 0.00001499, 0.00000057, -0.00001301, 0.00000611, 0.00000746, -0.00001378, -0.00000179, 0.00000957, -0.00000661, -0.00000245, -0.00000735, -0.00002164, 0.00000934, 0.00001034, -0.00001834, 0.00001300, 0.00001886, -0.00001618, 0.00000046, 0.00000940, -0.00000108, 0.00000878, -0.00000227, 0.00000334, 0.00001795, -0.00001108, -0.00000811, 0.00000788, -0.00000864, 0.00000634, -0.00000058, -0.00002439, 0.00002301, 0.00002255, -0.00003649, -0.00000176, 0.00003838, -0.00000790, -0.00002639, 0.00000437, 0.00000778, -0.00000359, 0.00000666, 0.00000766, -0.00001131, -0.00000822, 0.00000591, 0.00000520, 0.00000319, -0.00000611, -0.00000875, 0.00000389, 0.00000186, 0.00000461, 0.00000766, -0.00000963, 0.00000145, 0.00001093, -0.00001479, -0.00000446, 0.00001238, -0.00000628, 0.00000011, 0.00000729, -0.00001085, -0.00000237, 0.00001208, -0.00000021, -0.00001110, -0.00000563, 0.00000117, -0.00000371, -0.00000014, 0.00001019, -0.00000347, -0.00000896, 0.00000648, -0.00000479, -0.00002724, -0.00002605, -0.00001218, 0.00000426, 0.00001935, 0.00003342, 0.00003138, 0.00000738, 0.00000512, 0.00002160, 0.00001649, 0.00001162, 0.00001757, 0.00001400, 0.00002137, 0.00005049, 0.00005278, -0.00000384, -0.00005293, -0.00003521, -0.00001057, },
		{ -0.00001484, -0.00001362, -0.00001452, -0.00000482, -0.00000648, -0.00000434, -0.00000374, -0.00000650, -0.00000059, -0.00000142, -0.00000291, 0.00000266, 0.00000099, -0.00000252, -0.00000489, -0.00000195, 0.00000470, -0.00000732, -0.00001554, -0.00000065, 0.00000213, -0.00000649, -0.00000675, -0.00000458, 0.00000244, 0.00000700, 0.00000298, -0.00000201, -0.00001436, -0.00001704, 0.00000656, 0.00001309, -0.00000859, -0.00000501, 0.00001163, -0.00000467, -0.00001192, 0.00000639, 0.00000133, -0.00000391, -0.00000244, -0.00002080, -0.00000351, 0.00001844, -0.00001264, -0.00000304, 0.00002451, -0.00000748, -0.00000883, 0.00001065, -0.00000249, 0.00000762, 0.00000501, -0.00000772, 0.00001773, 0.00000358, -0.00001662, 0.00000814, -0.00000494, -0.00000608, 0.00001625, -0.00002117, -0.00001178, 0.00004002, -0.00000633, -0.00003568, 0.00002423, 0.00002610, -0.00002382, -0.00001774, 0.00001059, 0.00000535, -0.00000026, 0.00000828, -0.00000169, -0.00001277, -0.00000182, 0.00000787, 0.00000800, -0.00000429, -0.00001267, 0.00000290, 0.00000379, -0.00000172, 0.00001132, -0.00000154, -0.00000854, 0.00001262, -0.00000433, -0.00001513, 0.00001112, 0.00000279, -0.00000814, 0.00000701, -0.00000161, -0.00000890, 0.00000668, 0.00000827, -0.00000673, -0.00001287, -0.00000104, 0.00000251, -0.00000556, 0.00000481, 0.00000764, -0.00000886, -0.00000394, 0.00000414, -0.00001288, -0.00002958, -0.00002600, -0.00000653, 0.00001100, 0.00002643, 0.00003968, 0.00002357, 0.00000290, 0.00001438, 0.00002195, 0.00001409, 0.00001611, 0.00001659, 0.00001617, 0.00004026, 0.00006335, 0.00002372, -0.00004625, -0.00004971, },
		{ -0.00006186, -0.00002523, -0.00001117, -0.00001772, -0.00000701, -0.00000593, -0.00000618, -0.00000206, -0.00000746, -0.00000341, 0.00000099, -0.00000408, -0.00000033, 0.00000261, -0.00000003, -0.00000348, -0.00000580, 0.00000319, 0.00000071, -0.00001527, -0.00000825, 0.00000286, -0.00000239, -0.00000591, -0.00000790, -0.00000246, 0.00000804, 0.00000497, 0.00000089, -0.00000509, -0.00001961, -0.00000979, 0.00001380, 0.00000465, -0.00000951, 0.00000515, 0.00000562, -0.00001175, -0.00000339, 0.00000452, -0.00000107, 0.00000041, -0.00001594, -0.00001479, 0.00001779, -0.00000071, -0.00001475, 0.00002100, 0.00000461, -0.00001561, 0.00000970, -0.00000051, -0.00000038, 0.00001478, -0.00000950, 0.00000446, 0.00001870, -0.00001541, -0.00000116, 0.00000692, -0.00001746, 0.00001289, 0.00000496, -0.00003415, 0.00001755, 0.00003240, -0.00003232, -0.00001473, 0.00003679, 0.00000514, -0.00003016, -0.00000577, 0.00001393, 0.00000362, 0.00000265, 0.00000378, -0.00000811, -0.00000918, 0.00000254, 0.00001265, 0.00000357, -0.00001628, -0.00000357, 0.00000963, -0.00000466, 0.00000534, 0.00000957, -0.00001041, 0.00000448, 0.00000741, -0.00001676, 0.00000062, 0.00001163, -0.00000746, -0.00000079, 0.00000468, -0.00000648, -0.00000102, 0.00000889, 0.00000247, -0.00001296, -0.00001029, 0.00000526, -0.00000107, -0.00000448, 0.00000907, 0.00000152, -0.00000995, -0.00000104, -0.00000021, -0.00001880, -0.00003370, -0.00002243, 0.00000066, 0.00001688, 0.00003725, 0.00003951, 0.00001278, 0.00000689, 0.00002087, 0.00001922, 0.00001591, 0.00001816, 0.00001548, 0.00002865, 0.00006271, 0.00005103, -0.00002668, },
		{ 0.00000433, -0.00006491, -0.00004190, -0.00001045, -0.00001856, -0.00001127, -0.00000466, -0.00000844, -0.00000175, -0.00000550, -0.00000758, 0.00000140, -0.00000209, -0.00000354, 0.00000163, 0.00000167, -0.00000022, -0.00000655, -0.00000215, 0.00000536, -0.00000924, -0.00001374, -0.00000072, -0.00000032, -0.00000292, -0.00000691, -0.00000901, 0.00000484, 0.00000890, 0.00000137, 0.00000068, -0.00001146, -0.00002037, 0.00000077, 0.00001325, -0.00000193, -0.00000292, 0.00000718, -0.00000363, -0.00000925, -0.00000010, 0.00000130, 0.00000397, -0.00000877, -0.00002207, 0.00000978, 0.00001202, -0.00001895, 0.00001018, 0.00001718, -0.00001818, 0.00000342, 0.00000800, -0.00001080, 0.00001562, 0.00000296, -0.00001220, 0.00002091, -0.00000043, -0.00001470, 0.00001361, -0.00001209, -0.00000830, 0.00002423, -0.00002162, -0.00002148, 0.00003951, 0.00000472, -0.00003831, 0.00001165, 0.00003210, -0.00001524, -0.00002631, 0.00000633, 0.00001402, 0.00000284, 0.00000162, -0.00000230, -0.00000875, -0.00000584, 0.00000924, 0.00001435, -0.00001159, -0.00001537, 0.00001104, 0.00000169, -0.00000482, 0.00001294, -0.00000187, -0.00000526, 0.00001112, -0.00000865, -0.00001048, 0.00001188, 0.00000060, -0.00000669, 0.00000319, -0.00000104, -0.00000329, 0.00000392, 0.00000712, -0.00000472, -0.00001659, -0.00000187, 0.00000656, -0.00000586, 0.00000059, 0.00000901, -0.00000414, -0.00000922, -0.00000025, -0.00000391, -0.00002679, -0.00003574, -0.00001521, 0.00000589, 0.00002742, 0.00004657, 0.00003038, 0.00000767, 0.00001475, 0.00002220, 0.00001824, 0.00001884, 0.00001765, 0.00002034, 0.00005245, 0.00007033, },
		{ 0.00007610, 0.00004021, -0.00005279, -0.00006132, -0.00001534, -0.00001615, -0.00001708, -0.00000399, -0.00000911, -0.00000442, -0.00000181, -0.00000978, -0.00000174, 0.00000095, -0.00000446, -0.00000085, 0.00000128, 0.00000246, -0.00000323, -0.00000720, 0.00000418, -0.00000139, -0.00001444, -0.00000546, -0.00000099, -0.00000202, -0.00000214, -0.00001129, -0.00000373, 0.00001088, 0.00000452, 0.00000101, -0.00000202, -0.00001806, -0.00001451, 0.00000757, 0.00000902, -0.00000154, 0.00000169, 0.00000185, -0.00000641, -0.00000592, -0.00000196, 0.00000619, 0.00000113, -0.00002343, -0.00000374, 0.00002115, -0.00001387, -0.00000584, 0.00002608, -0.00001186, -0.00001063, 0.00001798, -0.00001168, 0.00000137, 0.00001986, -0.00001546, 0.00000518, 0.00001803, -0.00001770, 0.00000465, 0.00000601, -0.00002358, 0.00001551, 0.00001043, -0.00003766, 0.00000866, 0.00003692, -0.00002280, -0.00002391, 0.00002892, 0.00001497, -0.00002757, -0.00001473, 0.00001492, 0.00001234, 0.00000054, -0.00000272, -0.00000419, -0.00000811, -0.00000108, 0.00001789, 0.00000293, -0.00002274, 0.00000119, 0.00001222, -0.00000789, 0.00000524, 0.00000854, -0.00000714, 0.00000615, 0.00000126, -0.00001420, 0.00000438, 0.00000773, -0.00000565, -0.00000158, 0.00000047, -0.00000110, 0.00000106, 0.00000475, 0.00000340, -0.00001289, -0.00001289, 0.00000633, 0.00000161, -0.00000650, 0.00000559, 0.00000569, -0.00000824, -0.00000849, 0.00000120, -0.00001026, -0.00003558, -0.00003239, -0.00000930, 0.00001362, 0.00004247, 0.00004643, 0.00001935, 0.00001029, 0.00002048, 0.00002139, 0.00001969, 0.00002053, 0.00001772, 0.00003791, },
		{ 0.00002580, 0.00006787, 0.00007096, -0.00002358, -0.00007595, -0.00002909, -0.00001168, -0.00002208, -0.00000640, -0.00000684, -0.00000888, 0.00000016, -0.00000800, -0.00000664, 0.00000196, -0.00000294, -0.00000271, -0.00000047, 0.00000227, 0.00000199, -0.00000800, -0.00000142, 0.00000396, -0.00001085, -0.00000908, -0.00000207, -0.00000432, 0.00000088, -0.00000638, -0.00001197, 0.00000588, 0.00000977, 0.00000129, 0.00000159, -0.00000848, -0.00001950, -0.00000725, 0.00000960, 0.00000819, 0.00000045, -0.00000044, -0.00000144, -0.00000518, -0.00000802, 0.00000192, 0.00001084, -0.00001574, -0.00001896, 0.00002119, 0.00000087, -0.00002127, 0.00002372, 0.00000603, -0.00002576, 0.00001652, 0.00000356, -0.00001779, 0.00002246, 0.00000127, -0.00001526, 0.00002177, -0.00000316, -0.00001209, 0.00001673, -0.00001593, -0.00001073, 0.00002752, -0.00001757, -0.00002777, 0.00003247, 0.00001442, -0.00003398, -0.00000112, 0.00003055, -0.00000521, -0.00002845, -0.00000095, 0.00001944, 0.00000875, -0.00000518, -0.00000436, -0.00000350, -0.00000817, 0.00000944, 0.00001726, -0.00001624, -0.00001496, 0.00001483, 0.00000036, -0.00000509, 0.00001063, -0.00000082, 0.00000014, 0.00000591, -0.00001072, -0.00000387, 0.00000889, -0.00000087, -0.00000344, -0.00000165, -0.00000061, 0.00000274, 0.00000167, 0.00000437, -0.00000345, -0.00001697, -0.00000328, 0.00000787, -0.00000411, -0.00000314, 0.00000795, 0.00000111, -0.00001198, -0.00000562, 0.00000139, -0.00002103, -0.00003990, -0.00002747, -0.00000402, 0.00002887, 0.00005337, 0.00003743, 0.00001379, 0.00001623, 0.00002279, 0.00002114, 0.00002248, 0.00002014, },
		{ 0.00002440, 0.00002087, 0.00005114, 0.00008724, 0.00001786, -0.00007632, -0.00005138, -0.00000960, -0.00002279, -0.00001318, -0.00000332, -0.00001162, -0.00000199, -0.00000373, -0.00000992, -0.00000005, -0.00000092, -0.00000315, -0.00000150, -0.00000067, 0.00000503, -0.00000347, -0.00000720, 0.00000430, -0.00000504, -0.00001109, -0.00000224, -0.00000631, -0.00000198, 0.00000107, -0.00001266, -0.00000438, 0.00001094, 0.00000524, 0.00000154, -0.00000148, -0.00001351, -0.00001708, -0.00000284, 0.00001299, 0.00000939, -0.00000295, -0.00000299, 0.00000018, -0.00000842, -0.00000822, 0.00001257, 0.00000032, -0.00002729, 0.00000816, 0.00001920, -0.00002543, 0.00000565, 0.00002724, -0.00002692, -0.00000333, 0.00002394, -0.00002150, 0.00000302, 0.00002330, -0.00001896, 0.00000472, 0.00001638, -0.00001748, 0.00000800, 0.00000550, -0.00002596, 0.00001392, 0.00001460, -0.00003481, -0.00000144, 0.00003506, -0.00001043, -0.00002697, 0.00001641, 0.00001928, -0.00001929, -0.00002064, 0.00001107, 0.00002051, 0.00000061, -0.00000983, -0.00000107, -0.00000549, -0.00000317, 0.00001935, 0.00000121, -0.00002375, 0.00000407, 0.00001155, -0.00000710, 0.00000378, 0.00000548, -0.00000073, 0.00000587, -0.00000544, -0.00000841, 0.00000597, 0.00000250, -0.00000211, -0.00000238, -0.00000376, 0.00000385, 0.00000354, 0.00000087, 0.00000215, -0.00001131, -0.00001307, 0.00000478, 0.00000362, -0.00000643, 0.00000123, 0.00000798, -0.00000510, -0.00001348, -0.00000047, -0.00000411, -0.00003153, -0.00003989, -0.00002436, 0.00000752, 0.00004742, 0.00005384, 0.00002693, 0.00001455, 0.00002148, 0.00002287, 0.00002302, },
		{ 0.00002309, 0.00002692, 0.00002319, 0.00003488, 0.00008539, 0.00006022, -0.00005576, -0.00007576, -0.00001628, -0.00001772, -0.00002192, -0.00000270, -0.00001012, -0.00000702, -0.00000053, -0.00000989, -0.00000354, 0.00000013, -0.00000331, -0.00000082, -0.00000336, 0.00000307, 0.00000332, -0.00000860, 0.00000002, 0.00000036, -0.00001125, -0.00000292, -0.00000451, -0.00000815, 0.00000326, -0.00000553, -0.00001187, 0.00000474, 0.00000920, 0.00000350, 0.00000050, -0.00000594, -0.00001479, -0.00001622, 0.00000308, 0.00001938, 0.00000483, -0.00000871, 0.00000041, -0.00000084, -0.00001471, 0.00000194, 0.00001501, -0.00002019, -0.00001278, 0.00002784, -0.00001121, -0.00001961, 0.00003410, -0.00000588, -0.00002945, 0.00002733, -0.00000101, -0.00002262, 0.00002653, 0.00000063, -0.00001688, 0.00002012, -0.00000390, -0.00000948, 0.00001714, -0.00001597, -0.00001289, 0.00002587, -0.00001049, -0.00002950, 0.00002077, 0.00002027, -0.00002367, -0.00001124, 0.00002169, 0.00000407, -0.00002406, -0.00000920, 0.00002080, 0.00001548, -0.00001021, -0.00000721, 0.00000106, -0.00000771, 0.00000870, 0.00001565, -0.00001702, -0.00001182, 0.00001369, 0.00000046, -0.00000325, 0.00000531, 0.00000170, 0.00000580, -0.00000112, -0.00001017, 0.00000244, 0.00000383, -0.00000133, -0.00000004, -0.00000597, -0.00000021, 0.00000774, 0.00000061, 0.00000066, -0.00000333, -0.00001451, -0.00000469, 0.00000678, -0.00000156, -0.00000559, 0.00000536, 0.00000493, -0.00001182, -0.00001004, 0.00000192, -0.00001412, -0.00003795, -0.00004041, -0.00001853, 0.00002754, 0.00006023, 0.00004611, 0.00002035, 0.00001901, 0.00002411, },
		{ 0.00002392, 0.00002414, 0.00002659, 0.00002829, 0.00002624, 0.00006992, 0.00009058, -0.00001544, -0.00009071, -0.00003551, -0.00001045, -0.00002730, -0.00000791, -0.00000578, -0.00001131, -0.00000059, -0.00000762, -0.00000682, 0.00000056, -0.00000407, -0.00000032, -0.00000257, -0.00000240, 0.00000688, -0.00000405, -0.00000543, 0.00000276, -0.00000904, -0.00000566, -0.00000002, -0.00001110, -0.00000223, 0.00000197, -0.00001098, -0.00000406, 0.00000758, 0.00000729, 0.00000263, -0.00000324, -0.00000699, -0.00001764, -0.00001488, 0.00001550, 0.00002055, -0.00000634, -0.00000805, 0.00000533, -0.00000921, -0.00001316, 0.00001530, -0.00000023, -0.00002620, 0.00001706, 0.00001354, -0.00003246, 0.00001601, 0.00002483, -0.00003745, 0.00000375, 0.00002694, -0.00002794, 0.00000358, 0.00002473, -0.00001957, 0.00000285, 0.00001427, -0.00001507, 0.00000851, 0.00000551, -0.00002459, 0.00000900, 0.00001652, -0.00002689, -0.00001036, 0.00002720, 0.00000190, -0.00002335, 0.00000183, 0.00001734, -0.00000752, -0.00002231, 0.00000409, 0.00002586, 0.00000246, -0.00001525, 0.00000013, -0.00000201, -0.00000276, 0.00001648, -0.00000118, -0.00001933, 0.00000444, 0.00000851, -0.00000317, 0.00000106, 0.00000190, 0.00000671, 0.00000384, -0.00001053, -0.00000166, 0.00000535, -0.00000213, 0.00000150, -0.00000314, -0.00000662, 0.00000729, 0.00000589, -0.00000198, -0.00000083, -0.00000896, -0.00001126, 0.00000183, 0.00000437, -0.00000515, -0.00000242, 0.00000789, -0.00000193, -0.00001472, -0.00000424, -0.00000168, -0.00002340, -0.00004316, -0.00004135, -0.00000352, 0.00004966, 0.00006310, 0.00003580, 0.00001923, },
		{ 0.00002752, 0.00002269, 0.00002601, 0.00002560, 0.00003130, 0.00002616, 0.00005072, 0.00010133, 0.00003452, -0.00008522, -0.00006395, -0.00000874, -0.00002542, -0.00001748, -0.00000273, -0.00001238, -0.00000332, -0.00000478, -0.00000973, 0.00000079, -0.00000393, -0.00000269, 0.00000113, -0.00000609, 0.00000391, 0.00000313, -0.00000747, 0.00000127, -0.00000501, -0.00000946, 0.00000271, -0.00000745, -0.00001010, 0.00000277, -0.00000466, -0.00000804, 0.00000130, 0.00000724, 0.00000754, -0.00000188, -0.00000387, -0.00000821, -0.00002416, -0.00000438, 0.00002810, 0.00001009, -0.00001461, -0.00000010, 0.00000305, -0.00001766, -0.00000040, 0.00001601, -0.00001942, -0.00000613, 0.00002845, -0.00002024, -0.00001548, 0.00003890, -0.00001533, -0.00002952, 0.00003361, -0.00000420, -0.00002496, 0.00002705, 0.00000121, -0.00001719, 0.00001647, -0.00000278, -0.00000737, 0.00001525, -0.00001279, -0.00001427, 0.00002016, -0.00000246, -0.00002651, 0.00000701, 0.00002093, -0.00000970, -0.00001693, 0.00000769, 0.00001063, -0.00001476, -0.00001592, 0.00001816, 0.00002066, -0.00001200, -0.00000993, 0.00000355, -0.00000435, 0.00000720, 0.00000999, -0.00001394, -0.00000704, 0.00000875, 0.00000184, -0.00000063, -0.00000097, 0.00000505, 0.00000988, -0.00000747, -0.00000752, 0.00000654, -0.00000125, -0.00000097, 0.00000216, -0.00000853, 0.00000028, 0.00001079, 0.00000069, -0.00000313, -0.00000420, -0.00001055, -0.00000548, 0.00000424, 0.00000037, -0.00000638, 0.00000234, 0.00000612, -0.00000923, -0.00001223, -0.00000160, -0.00000818, -0.00003123, -0.00005035, -0.00003622, 0.00002009, 0.00006643, 0.00005750, },
		{ 0.00007418, 0.00004696, 0.00002460, 0.00002669, 0.00002638, 0.00003125, 0.00003024, 0.00003635, 0.00009382, 0.00007993, -0.00005535, -0.00009108, -0.00001952, -0.00001761, -0.00002654, -0.00000413, -0.00001060, -0.00000723, -0.00000184, -0.00001214, -0.00000091, -0.00000059, -0.00000721, 0.00000290, -0.00000392, -0.00000302, 0.00000688, -0.00000425, -0.00000207, -0.00000113, -0.00001172, 0.00000075, -0.00000003, -0.00001341, -0.00000287, -0.00000033, -0.00000609, -0.00000277, 0.00000149, 0.00000996, 0.00000401, -0.00000571, -0.00000045, -0.00001694, -0.00002477, 0.00001525, 0.00002883, -0.00000571, -0.00001287, 0.00000673, -0.00000652, -0.00001658, 0.00001263, 0.00000164, -0.00002058, 0.00002000, 0.00000697, -0.00003375, 0.00002196, 0.00002021, -0.00004167, 0.00000847, 0.00002709, -0.00003002, 0.00000267, 0.00002410, -0.00001767, 0.00000004, 0.00001209, -0.00001130, 0.00000674, 0.00000590, -0.00002030, 0.00000239, 0.00001576, -0.00001603, -0.00001624, 0.00001499, 0.00001170, -0.00001429, -0.00001190, 0.00001011, 0.00000459, -0.00001945, -0.00000354, 0.00002665, 0.00000524, -0.00001723, -0.00000083, 0.00000150, -0.00000002, 0.00001010, -0.00000354, -0.00001120, 0.00000271, 0.00000448, 0.00000189, -0.00000211, -0.00000072, 0.00001275, 0.00000101, -0.00001251, 0.00000375, 0.00000318, -0.00000483, 0.00000377, -0.00000374, -0.00000734, 0.00000887, 0.00000751, -0.00000359, -0.00000438, -0.00000669, -0.00000835, -0.00000101, 0.00000356, -0.00000366, -0.00000433, 0.00000567, 0.00000022, -0.00001239, -0.00000847, -0.00000291, -0.00001440, -0.00004132, -0.00005618, -0.00002057, 0.00004644, },
		{ 0.00000439, 0.00006986, 0.00007185, 0.00003685, 0.00002641, 0.00002804, 0.00003076, 0.00003416, 0.00002959, 0.00007551, 0.00010972, -0.00000631, -0.00010425, -0.00004387, -0.00000991, -0.00003064, -0.00001026, -0.00000750, -0.00001171, 0.00000056, -0.00001183, -0.00000686, 0.00000467, -0.00000876, -0.00000127, 0.00000199, -0.00000724, 0.00000409, 0.00000125, -0.00000389, 0.00000111, -0.00001098, -0.00000496, 0.00000579, -0.00001001, -0.00000925, -0.00000116, -0.00000384, -0.00000121, -0.00000351, 0.00000433, 0.00001165, -0.00000395, -0.00000192, -0.00000280, -0.00002864, -0.00001057, 0.00003066, 0.00001636, -0.00001572, -0.00000472, 0.00000649, -0.00001561, -0.00000659, 0.00001471, -0.00001236, -0.00000191, 0.00002371, -0.00002344, -0.00001117, 0.00003740, -0.00002000, -0.00002689, 0.00003449, -0.00000525, -0.00002489, 0.00002423, 0.00000253, -0.00001634, 0.00001184, -0.00000056, -0.00000588, 0.00001195, -0.00000771, -0.00001443, 0.00001202, 0.00000453, -0.00001968, -0.00000606, 0.00001654, 0.00000483, -0.00001775, -0.00000802, 0.00001314, -0.00000328, -0.00001890, 0.00001236, 0.00002229, -0.00001007, -0.00001148, 0.00000356, 0.00000095, 0.00000517, 0.00000207, -0.00000820, -0.00000198, 0.00000229, 0.00000364, 0.00000150, -0.00000590, 0.00000812, 0.00001128, -0.00001101, -0.00000390, 0.00000751, -0.00000470, -0.00000025, 0.00000253, -0.00000881, 0.00000106, 0.00001160, 0.00000127, -0.00000609, -0.00000550, -0.00000664, -0.00000523, 0.00000132, 0.00000070, -0.00000568, -0.00000023, 0.00000465, -0.00000540, -0.00001187, -0.00000702, -0.00000506, -0.00002236, -0.00005406, -0.00005462, },
		{ -0.00006516, -0.00004272, 0.00003538, 0.00008502, 0.00006191, 0.00003158, 0.00002811, 0.00003109, 0.00003729, 0.00002922, 0.00005419, 0.00011920, 0.00005093, -0.00009429, -0.00007634, -0.00000941, -0.00002810, -0.00001966, -0.00000432, -0.00001563, -0.00000041, -0.00000626, -0.00001528, 0.00000583, -0.00000319, -0.00000886, 0.00000470, -0.00000480, -0.00000205, 0.00000408, -0.00000310, 0.00000205, -0.00000741, -0.00001152, 0.00000661, -0.00000261, -0.00001213, -0.00000414, -0.00000534, 0.00000093, -0.00000088, -0.00000525, 0.00001166, 0.00000514, -0.00000639, 0.00000369, -0.00001593, -0.00002868, 0.00001091, 0.00003180, 0.00000096, -0.00001741, 0.00000343, 0.00000064, -0.00001922, 0.00000464, 0.00000753, -0.00001247, 0.00001602, 0.00000237, -0.00002997, 0.00002169, 0.00001513, -0.00003948, 0.00000980, 0.00002487, -0.00002794, 0.00000056, 0.00002157, -0.00001420, -0.00000280, 0.00001001, -0.00000726, 0.00000381, 0.00000636, -0.00001435, -0.00000411, 0.00001251, -0.00000467, -0.00001824, 0.00000102, 0.00001731, -0.00000252, -0.00002215, -0.00000018, 0.00001390, -0.00001272, -0.00000938, 0.00002225, 0.00000780, -0.00001505, -0.00000322, 0.00000428, 0.00000389, 0.00000207, -0.00000520, -0.00000201, 0.00000008, 0.00000078, 0.00000588, -0.00000459, -0.00000153, 0.00001549, -0.00000143, -0.00001103, 0.00000631, 0.00000044, -0.00000523, 0.00000404, -0.00000405, -0.00000587, 0.00000890, 0.00000783, -0.00000416, -0.00000721, -0.00000519, -0.00000531, -0.00000265, 0.00000138, -0.00000274, -0.00000440, 0.00000220, 0.00000093, -0.00000799, -0.00001139, -0.00000659, -0.00000792, -0.00003470, },
		{ -0.00001431, -0.00005008, -0.00007060, -0.00001938, 0.00006718, 0.00008819, 0.00005076, 0.00002930, 0.00002995, 0.00003999, 0.00003436, 0.00003584, 0.00010854, 0.00010372, -0.00005757, -0.00010716, -0.00002178, -0.00001969, -0.00003016, -0.00000328, -0.00001567, -0.00000755, 0.00000243, -0.00001916, -0.00000196, 0.00000631, -0.00001199, 0.00000003, 0.00000100, -0.00000550, 0.00000271, -0.00000181, 0.00000215, -0.00000146, -0.00001595, 0.00000172, 0.00000520, -0.00001116, -0.00000604, -0.00000774, -0.00000294, 0.00000644, -0.00000776, 0.00000119, 0.00001331, -0.00000366, 0.00000027, -0.00000166, -0.00002744, -0.00001381, 0.00002448, 0.00002377, -0.00001114, -0.00001397, 0.00001002, -0.00000912, -0.00001696, 0.00001298, -0.00000016, -0.00000217, 0.00001595, -0.00002014, -0.00000881, 0.00003066, -0.00001919, -0.00002274, 0.00003051, -0.00000429, -0.00002273, 0.00001899, 0.00000375, -0.00001449, 0.00000734, 0.00000169, -0.00000486, 0.00000825, -0.00000227, -0.00001321, 0.00000331, 0.00000910, -0.00001067, -0.00001621, 0.00000858, 0.00001676, -0.00001451, -0.00002163, 0.00001128, 0.00000741, -0.00001700, 0.00000483, 0.00001941, -0.00000506, -0.00001102, 0.00000167, 0.00000642, 0.00000287, -0.00000566, -0.00000177, 0.00000221, -0.00000327, 0.00000475, 0.00000237, -0.00000783, 0.00000980, 0.00000995, -0.00001074, -0.00000054, 0.00000558, -0.00000580, 0.00000035, 0.00000123, -0.00000709, 0.00000227, 0.00001047, 0.00000150, -0.00000755, -0.00000660, -0.00000395, -0.00000395, -0.00000114, -0.00000066, -0.00000419, -0.00000200, 0.00000158, -0.00000185, -0.00000950, -0.00001163, -0.00000575, },
		{ -0.00001096, -0.00000629, -0.00002494, -0.00006617, -0.00006679, 0.00001508, 0.00009173, 0.00008190, 0.00004153, 0.00002712, 0.00003930, 0.00004413, 0.00002664, 0.00008164, 0.00013888, 0.00000358, -0.00012337, -0.00005031, -0.00000895, -0.00003729, -0.00000878, -0.00000915, -0.00001830, 0.00000628, -0.00001272, -0.00001524, 0.00001039, -0.00000591, -0.00000805, 0.00000332, -0.00000416, 0.00000026, -0.00000168, 0.00000011, 0.00000556, -0.00001506, -0.00000787, 0.00001028, -0.00000678, -0.00000761, -0.00000628, -0.00001112, 0.00000783, 0.00000106, -0.00000975, 0.00001038, 0.00000551, -0.00000286, 0.00000314, -0.00001520, -0.00002404, 0.00000170, 0.00002854, 0.00001399, -0.00002187, -0.00000365, 0.00001177, -0.00002227, -0.00000659, 0.00001753, -0.00000480, 0.00000650, 0.00000164, -0.00002325, 0.00001543, 0.00001104, -0.00003228, 0.00000793, 0.00002095, -0.00002280, -0.00000192, 0.00001741, -0.00001026, -0.00000472, 0.00000797, -0.00000385, 0.00000092, 0.00000653, -0.00000807, -0.00000904, 0.00000754, 0.00000502, -0.00001660, -0.00001179, 0.00001819, 0.00000864, -0.00002716, -0.00001075, 0.00001819, -0.00000360, -0.00001181, 0.00001334, 0.00000907, -0.00000917, -0.00000563, 0.00000573, 0.00000734, -0.00000527, -0.00000578, 0.00000560, -0.00000201, -0.00000179, 0.00000724, -0.00000540, -0.00000054, 0.00001435, -0.00000251, -0.00000691, 0.00000575, -0.00000197, -0.00000384, 0.00000253, -0.00000404, -0.00000277, 0.00000806, 0.00000658, -0.00000412, -0.00000842, -0.00000471, -0.00000289, -0.00000282, -0.00000151, -0.00000268, -0.00000346, -0.00000129, 0.00000041, -0.00000358, -0.00001168, },
		{ -0.00000615, -0.00001334, -0.00001001, -0.00000940, -0.00003941, -0.00008087, -0.00004851, 0.00005455, 0.00010423, 0.00007144, 0.00003144, 0.00003202, 0.00005284, 0.00003221, 0.00004882, 0.00014455, 0.00007798, -0.00010995, -0.00009273, -0.00000485, -0.00003419, -0.00002331, 0.00000015, -0.00002377, -0.00000132, 0.00000059, -0.00002331, 0.00000297, 0.00000472, -0.00001135, -0.00000015, -0.00000178, -0.00000046, -0.00000087, -0.00000492, 0.00000950, -0.00000649, -0.00001826, 0.00000882, 0.00000131, -0.00001011, -0.00000226, -0.00001477, -0.00000182, 0.00001182, -0.00000902, -0.00000149, 0.00001006, 0.00000087, 0.00000278, -0.00000670, -0.00001920, -0.00001354, 0.00000920, 0.00003219, -0.00000191, -0.00002661, 0.00001429, -0.00000053, -0.00003016, 0.00001284, 0.00001405, -0.00000727, 0.00000822, -0.00001188, -0.00000951, 0.00002090, -0.00001384, -0.00001805, 0.00002323, -0.00000219, -0.00001894, 0.00001263, 0.00000401, -0.00001178, 0.00000380, 0.00000306, -0.00000403, 0.00000497, 0.00000223, -0.00001077, -0.00000429, 0.00001070, -0.00000155, -0.00002204, -0.00000071, 0.00002356, -0.00000884, -0.00003009, 0.00000573, 0.00001490, -0.00001051, -0.00000280, 0.00001231, 0.00000144, -0.00000810, -0.00000092, 0.00001014, 0.00000069, -0.00001099, 0.00000340, 0.00000492, -0.00000625, 0.00000431, 0.00000197, -0.00000637, 0.00000949, 0.00000678, -0.00000710, 0.00000169, 0.00000191, -0.00000488, 0.00000052, -0.00000103, -0.00000414, 0.00000396, 0.00000803, 0.00000081, -0.00000737, -0.00000708, -0.00000284, -0.00000212, -0.00000281, -0.00000295, -0.00000292, -0.00000309, -0.00000152, 0.00000011, },
		{ -0.00000077, -0.00000094, -0.00000904, -0.00001437, -0.00000985, -0.00001473, -0.00005925, -0.00008782, -0.00001430, 0.00008959, 0.00010682, 0.00005590, 0.00002447, 0.00005060, 0.00005041, 0.00002714, 0.00011765, 0.00014338, -0.00005635, -0.00013510, -0.00002099, -0.00001798, -0.00004075, 0.00000160, -0.00001688, -0.00001681, 0.00000905, -0.00001861, -0.00001133, 0.00000999, -0.00000685, -0.00000480, -0.00000197, -0.00000128, 0.00000296, -0.00000927, 0.00000518, 0.00000702, -0.00002188, -0.00000165, 0.00001056, -0.00001036, -0.00000180, -0.00000925, -0.00001478, 0.00001166, 0.00000257, -0.00000919, 0.00000432, 0.00000494, 0.00000622, -0.00000336, -0.00001470, -0.00001055, -0.00001212, 0.00001993, 0.00003145, -0.00002629, -0.00001196, 0.00002471, -0.00002666, -0.00001753, 0.00002997, -0.00000027, -0.00000531, 0.00000502, -0.00001627, 0.00000521, 0.00000864, -0.00002233, 0.00000408, 0.00001597, -0.00001616, -0.00000387, 0.00001204, -0.00000681, -0.00000508, 0.00000588, -0.00000156, -0.00000105, 0.00000619, -0.00000254, -0.00001162, 0.00000199, 0.00001150, -0.00001241, -0.00002096, 0.00001477, 0.00001622, -0.00002642, -0.00001903, 0.00001678, 0.00000596, -0.00001046, 0.00000182, 0.00000835, -0.00000105, -0.00000651, 0.00000557, 0.00000888, -0.00000995, -0.00000529, 0.00000989, -0.00000257, -0.00000306, 0.00000566, -0.00000422, 0.00000149, 0.00001014, -0.00000191, -0.00000172, 0.00000297, -0.00000348, -0.00000168, 0.00000004, -0.00000372, 0.00000107, 0.00000699, 0.00000393, -0.00000385, -0.00000789, -0.00000493, -0.00000139, -0.00000213, -0.00000416, -0.00000330, -0.00000281, -0.00000370, },
		{ -0.00000391, -0.00000328, -0.00000011, -0.00000294, -0.00001149, -0.00001624, -0.00000871, -0.00002546, -0.00008276, -0.00007829, 0.00002904, 0.00011646, 0.00009936, 0.00003474, 0.00003515, 0.00006635, 0.00003058, 0.00007267, 0.00017362, 0.00003115, -0.00015145, -0.00006529, 0.00000130, -0.00004692, -0.00001254, -0.00000205, -0.00002750, 0.00000392, -0.00000526, -0.00002130, 0.00000531, 0.00000019, -0.00000576, -0.00000283, -0.00000540, 0.00000685, -0.00000636, -0.00000674, 0.00001585, -0.00001268, -0.00001676, 0.00001361, -0.00000386, -0.00000671, -0.00000055, -0.00001912, -0.00000056, 0.00001192, -0.00000510, -0.00000329, 0.00000062, 0.00001048, 0.00000523, -0.00001783, -0.00000418, -0.00001048, -0.00001286, 0.00004081, 0.00000932, -0.00003966, 0.00001954, 0.00000686, -0.00004331, 0.00001552, 0.00002614, -0.00001568, 0.00000317, -0.00000173, -0.00001306, 0.00001071, -0.00000606, -0.00001340, 0.00001461, -0.00000007, -0.00001407, 0.00000637, 0.00000284, -0.00000846, 0.00000159, 0.00000314, -0.00000312, 0.00000256, 0.00000507, -0.00000756, -0.00000962, 0.00000962, 0.00000575, -0.00002315, -0.00000899, 0.00002398, -0.00000268, -0.00003187, -0.00000191, 0.00001794, -0.00000112, -0.00000905, 0.00000247, 0.00000751, -0.00000286, -0.00000286, 0.00001081, -0.00000105, -0.00001270, 0.00000607, 0.00000607, -0.00000624, 0.00000219, 0.00000083, -0.00000246, 0.00000737, 0.00000316, -0.00000168, 0.00000254, -0.00000194, -0.00000293, 0.00000019, -0.00000329, -0.00000087, 0.00000587, 0.00000499, -0.00000079, -0.00000600, -0.00000679, -0.00000278, -0.00000059, -0.00000377, -0.00000493, -0.00000276, },
		{ -0.00000415, -0.00000354, -0.00000449, -0.00000197, -0.00000076, -0.00000435, -0.00001577, -0.00001675, -0.00000726, -0.00004653, -0.00010044, -0.00005074, 0.00007453, 0.00013402, 0.00007590, 0.00002352, 0.00006256, 0.00005487, 0.00003805, 0.00015687, 0.00012357, -0.00011758, -0.00012606, 0.00000171, -0.00003220, -0.00003610, 0.00000623, -0.00002381, -0.00001107, 0.00000489, -0.00002076, -0.00000442, 0.00000341, -0.00000453, -0.00000025, -0.00001023, 0.00000361, 0.00000469, -0.00001628, 0.00001066, 0.00000543, -0.00002494, 0.00000473, 0.00000717, -0.00001063, 0.00000227, -0.00001255, -0.00001352, 0.00000966, 0.00000343, -0.00000196, -0.00000721, 0.00000362, 0.00002032, -0.00001540, -0.00001372, 0.00000939, -0.00002893, 0.00000835, 0.00004943, -0.00003021, -0.00001822, 0.00003607, -0.00003243, -0.00002449, 0.00004168, -0.00000037, -0.00001549, 0.00001110, -0.00001129, -0.00000593, 0.00000776, -0.00001201, -0.00000011, 0.00001048, -0.00000954, -0.00000466, 0.00000602, -0.00000437, -0.00000388, 0.00000373, -0.00000042, -0.00000176, 0.00000533, 0.00000156, -0.00001179, -0.00000297, 0.00001410, -0.00000717, -0.00002496, 0.00000823, 0.00001842, -0.00002079, -0.00002332, 0.00001070, 0.00001396, -0.00000616, -0.00000979, 0.00000547, 0.00000738, -0.00000483, 0.00000399, 0.00000785, -0.00001107, -0.00000400, 0.00001053, -0.00000147, -0.00000337, 0.00000209, -0.00000153, 0.00000355, 0.00000470, -0.00000007, 0.00000287, -0.00000044, -0.00000400, 0.00000013, -0.00000244, -0.00000307, 0.00000475, 0.00000599, 0.00000057, -0.00000359, -0.00000626, -0.00000513, -0.00000072, -0.00000170, -0.00000565, },
		{ -0.00000437, -0.00000552, -0.00000411, -0.00000490, -0.00000322, -0.00000206, -0.00000068, -0.00000713, -0.00002189, -0.00001191, -0.00001300, -0.00007495, -0.00010324, -0.00000977, 0.00012045, 0.00013220, 0.00004334, 0.00004064, 0.00007568, 0.00003518, 0.00010963, 0.00018291, -0.00003250, -0.00017075, -0.00003383, -0.00000526, -0.00005118, -0.00000270, -0.00000972, -0.00002358, 0.00000461, -0.00001365, -0.00001211, 0.00000308, -0.00000573, 0.00000379, -0.00000794, -0.00000838, 0.00001389, -0.00001220, -0.00000622, 0.00001822, -0.00001743, -0.00001137, 0.00001321, -0.00000718, -0.00000159, -0.00000340, -0.00001773, 0.00000052, 0.00000491, 0.00000505, -0.00000420, -0.00001442, 0.00002444, 0.00000692, -0.00003262, 0.00001553, -0.00000627, -0.00003712, 0.00004807, 0.00001916, -0.00004917, 0.00002522, 0.00001000, -0.00005277, 0.00002088, 0.00003238, -0.00002453, 0.00000216, 0.00000687, -0.00001809, 0.00000224, 0.00000166, -0.00000903, 0.00000647, 0.00000114, -0.00000877, 0.00000109, 0.00000033, -0.00000494, 0.00000062, 0.00000208, -0.00000205, 0.00000114, 0.00000615, -0.00000420, -0.00001214, 0.00000667, 0.00000992, -0.00002006, -0.00001456, 0.00001828, 0.00000224, -0.00002727, -0.00000955, 0.00001659, 0.00000859, -0.00001296, -0.00000793, 0.00001145, 0.00000391, -0.00000318, 0.00000822, -0.00000218, -0.00001086, 0.00000615, 0.00000599, -0.00000411, -0.00000094, -0.00000030, 0.00000213, 0.00000431, 0.00000036, 0.00000345, 0.00000232, -0.00000466, -0.00000108, -0.00000058, -0.00000476, 0.00000210, 0.00000750, 0.00000195, -0.00000273, -0.00000433, -0.00000587, -0.00000274, -0.00000022, },
		{ -0.00000062, -0.00000251, -0.00000553, -0.00000477, -0.00000595, -0.00000398, -0.00000272, -0.00000241, 0.00000043, -0.00001487, -0.00002367, -0.00000532, -0.00003150, -0.00010068, -0.00009011, 0.00004462, 0.00015772, 0.00010224, 0.00002990, 0.00007107, 0.00005758, 0.00006715, 0.00018960, 0.00007396, -0.00016589, -0.00009842, 0.00000926, -0.00004465, -0.00002415, 0.00000144, -0.00002574, -0.00000299, -0.00000687, -0.00001691, 0.00000325, -0.00000811, 0.00000120, 0.00000381, -0.00001871, 0.00000891, 0.00000429, -0.00001954, 0.00001394, 0.00000076, -0.00002148, 0.00000797, 0.00000096, -0.00000531, 0.00000145, -0.00001526, -0.00000564, 0.00000028, 0.00000375, 0.00001163, -0.00002328, 0.00000208, 0.00003690, -0.00003073, -0.00001128, 0.00003124, -0.00004581, -0.00000320, 0.00006306, -0.00003259, -0.00001961, 0.00004231, -0.00003848, -0.00002487, 0.00004908, -0.00000486, -0.00002077, 0.00001739, -0.00000942, -0.00001503, 0.00000763, -0.00000324, -0.00000322, 0.00000500, -0.00000408, -0.00000419, 0.00000011, -0.00000304, -0.00000166, 0.00000166, -0.00000013, -0.00000141, 0.00000412, 0.00000396, -0.00001002, -0.00000639, 0.00001297, -0.00000223, -0.00002359, 0.00000015, 0.00001502, -0.00001220, -0.00002305, 0.00000222, 0.00001879, -0.00000064, -0.00001905, 0.00000096, 0.00001426, -0.00000052, 0.00000161, 0.00000454, -0.00000904, -0.00000224, 0.00000846, 0.00000056, -0.00000325, -0.00000180, 0.00000158, 0.00000478, 0.00000005, 0.00000212, 0.00000566, -0.00000297, -0.00000371, 0.00000087, -0.00000421, -0.00000197, 0.00000753, 0.00000491, -0.00000250, -0.00000347, -0.00000461, -0.00000447, },
		{ -0.00000455, -0.00000179, -0.00000149, -0.00000498, -0.00000436, -0.00000659, -0.00000634, -0.00000196, -0.00000439, 0.00000019, -0.00000256, -0.00002421, -0.00001760, -0.00000650, -0.00005821, -0.00011856, -0.00005859, 0.00011135, 0.00016459, 0.00006304, 0.00004972, 0.00007604, 0.00005349, 0.00015772, 0.00016024, -0.00010300, -0.00016094, -0.00000931, -0.00002198, -0.00004284, 0.00000098, -0.00002025, -0.00001153, -0.00000159, -0.00002172, 0.00000242, -0.00000396, -0.00001074, 0.00001418, -0.00001382, -0.00000933, 0.00001720, -0.00001571, -0.00000317, 0.00001388, -0.00001754, -0.00000321, 0.00000562, -0.00000545, 0.00000384, -0.00001211, -0.00000870, 0.00000084, -0.00000907, 0.00002066, -0.00000450, -0.00003252, 0.00004025, 0.00000710, -0.00004697, 0.00003623, -0.00000276, -0.00005794, 0.00005188, 0.00002466, -0.00005155, 0.00003002, 0.00000730, -0.00005537, 0.00002730, 0.00003074, -0.00003060, 0.00000489, 0.00001127, -0.00002267, -0.00000332, 0.00000729, -0.00000500, 0.00000007, 0.00000105, -0.00000378, -0.00000276, -0.00000287, -0.00000176, 0.00000046, 0.00000047, -0.00000089, 0.00000052, 0.00000581, -0.00000127, -0.00001194, 0.00000286, 0.00001054, -0.00001399, -0.00001671, 0.00000809, 0.00000478, -0.00001819, -0.00001514, 0.00001203, 0.00001598, -0.00001426, -0.00001666, 0.00001240, 0.00001092, -0.00000154, 0.00000343, -0.00000281, -0.00000668, 0.00000461, 0.00000510, -0.00000143, -0.00000389, -0.00000086, 0.00000560, 0.00000151, -0.00000092, 0.00000664, 0.00000160, -0.00000550, -0.00000008, -0.00000167, -0.00000499, 0.00000452, 0.00000816, -0.00000059, -0.00000410, -0.00000336, },
		{ -0.00000351, -0.00000390, -0.00000245, -0.00000078, -0.00000490, -0.00000409, -0.00000487, -0.00000936, -0.00000330, -0.00000283, -0.00000424, 0.00000302, -0.00001206, -0.00002716, -0.00001031, -0.00001799, -0.00008759, -0.00012650, 0.00000102, 0.00016942, 0.00013483, 0.00004998, 0.00007186, 0.00006103, 0.00011925, 0.00020343, -0.00000424, -0.00018819, -0.00005979, -0.00000225, -0.00004842, -0.00000876, -0.00001301, -0.00002001, 0.00000408, -0.00002371, -0.00000709, 0.00000882, -0.00002081, 0.00000850, 0.00000564, -0.00002351, 0.00001181, 0.00000136, -0.00001609, 0.00001198, -0.00000587, -0.00000988, 0.00000496, -0.00000621, 0.00000643, -0.00000633, -0.00001624, 0.00000861, -0.00001602, 0.00000379, 0.00002918, -0.00004120, 0.00000087, 0.00005200, -0.00004562, -0.00000666, 0.00004958, -0.00005915, -0.00001209, 0.00006805, -0.00003211, -0.00001487, 0.00004103, -0.00004285, -0.00001822, 0.00004934, -0.00001190, -0.00001964, 0.00002133, -0.00001046, -0.00002007, 0.00000732, 0.00000289, -0.00000450, 0.00000007, -0.00000038, -0.00000287, -0.00000491, -0.00000261, 0.00000066, -0.00000009, -0.00000022, -0.00000047, 0.00000286, 0.00000480, -0.00000710, -0.00000787, 0.00000899, 0.00000149, -0.00001792, -0.00000788, 0.00000726, -0.00000314, -0.00001875, -0.00000587, 0.00001963, 0.00000411, -0.00002432, -0.00000413, 0.00001823, 0.00000548, -0.00000054, 0.00000002, -0.00000535, -0.00000028, 0.00000533, 0.00000227, -0.00000308, -0.00000440, 0.00000392, 0.00000488, -0.00000241, 0.00000368, 0.00000623, -0.00000379, -0.00000288, 0.00000031, -0.00000505, -0.00000024, 0.00000900, 0.00000337, -0.00000426, },
		{ -0.00000235, -0.00000424, -0.00000373, -0.00000308, 0.00000038, -0.00000415, -0.00000583, -0.00000207, -0.00000910, -0.00000823, -0.00000094, -0.00000533, -0.00000031, -0.00000028, -0.00002132, -0.00002384, -0.00000850, -0.00003574, -0.00012134, -0.00011032, 0.00008474, 0.00018892, 0.00009944, 0.00006325, 0.00006561, 0.00009228, 0.00020928, 0.00009856, -0.00016667, -0.00012249, -0.00000079, -0.00004255, -0.00002238, -0.00000474, -0.00002949, 0.00000557, -0.00001429, -0.00002628, 0.00001761, -0.00001357, -0.00001279, 0.00002085, -0.00001740, -0.00000757, 0.00001349, -0.00001382, 0.00000155, 0.00000139, -0.00001028, 0.00000555, -0.00000939, 0.00000338, 0.00000782, -0.00002610, 0.00000782, -0.00000008, -0.00002771, 0.00003901, -0.00000450, -0.00004970, 0.00005419, 0.00000547, -0.00005630, 0.00005351, -0.00000125, -0.00007030, 0.00005023, 0.00002432, -0.00004483, 0.00003233, -0.00000070, -0.00004952, 0.00003225, 0.00002155, -0.00003150, 0.00000967, 0.00001046, -0.00002506, -0.00000581, 0.00000979, -0.00000140, -0.00000400, -0.00000015, 0.00000017, -0.00000512, -0.00000585, 0.00000053, 0.00000058, -0.00000097, 0.00000017, 0.00000044, 0.00000461, 0.00000084, -0.00000961, -0.00000091, 0.00000815, -0.00000654, -0.00001577, -0.00000403, 0.00000453, -0.00000743, -0.00001719, 0.00000606, 0.00001906, -0.00001341, -0.00002197, 0.00001047, 0.00001657, 0.00000175, -0.00000169, -0.00000329, -0.00000190, 0.00000293, 0.00000376, 0.00000027, -0.00000553, -0.00000069, 0.00000687, -0.00000010, -0.00000072, 0.00000718, 0.00000092, -0.00000450, -0.00000006, -0.00000302, -0.00000392, 0.00000650, 0.00000727, },
		{ 0.00000898, 0.00000111, -0.00000416, -0.00000371, -0.00000442, 0.00000071, -0.00000098, -0.00000826, -0.00000233, -0.00000438, -0.00001230, -0.00000390, -0.00000234, -0.00000446, 0.00000167, -0.00000816, -0.00002503, -0.00002144, -0.00000916, -0.00006407, -0.00015291, -0.00005200, 0.00015970, 0.00017229, 0.00008855, 0.00006315, 0.00007050, 0.00019237, 0.00018648, -0.00010156, -0.00017705, -0.00002115, -0.00003019, -0.00003969, 0.00000449, -0.00003316, -0.00000714, 0.00000672, -0.00004014, 0.00000574, 0.00001000, -0.00002932, 0.00001375, 0.00000434, -0.00002080, 0.00000906, -0.00000346, -0.00000433, 0.00000238, -0.00001243, 0.00000941, -0.00000601, -0.00001206, 0.00002341, -0.00002110, -0.00001458, 0.00002690, -0.00003657, 0.00000378, 0.00004674, -0.00005707, 0.00000026, 0.00006147, -0.00005584, 0.00000010, 0.00005979, -0.00006563, -0.00001692, 0.00006212, -0.00002772, -0.00000483, 0.00003192, -0.00004338, -0.00000647, 0.00004146, -0.00001875, -0.00001276, 0.00002116, -0.00001316, -0.00002046, 0.00000620, 0.00000612, -0.00000399, -0.00000377, 0.00000147, -0.00000140, -0.00000836, -0.00000272, 0.00000222, -0.00000125, -0.00000028, 0.00000055, 0.00000178, 0.00000442, -0.00000379, -0.00000754, 0.00000343, 0.00000365, -0.00000993, -0.00001450, -0.00000267, 0.00000402, -0.00001180, -0.00001113, 0.00001660, 0.00000644, -0.00002507, -0.00000847, 0.00001872, 0.00001154, -0.00000138, -0.00000429, -0.00000191, 0.00000177, 0.00000275, 0.00000257, -0.00000303, -0.00000486, 0.00000474, 0.00000407, -0.00000241, 0.00000407, 0.00000498, -0.00000293, -0.00000171, -0.00000123, -0.00000512, 0.00000224, },
		{ -0.00000180, 0.00000813, 0.00000466, -0.00000317, -0.00000302, -0.00000535, -0.00000166, 0.00000319, -0.00000720, -0.00000698, -0.00000000, -0.00001019, -0.00001041, -0.00000172, -0.00000353, -0.00000192, -0.00000184, -0.00001313, -0.00002874, -0.00001845, -0.00001291, -0.00011118, -0.00015518, 0.00003777, 0.00019712, 0.00015525, 0.00008197, 0.00004830, 0.00015429, 0.00025108, -0.00000004, -0.00020961, -0.00006418, -0.00001184, -0.00005909, 0.00000451, -0.00001913, -0.00003336, 0.00002151, -0.00002976, -0.00002510, 0.00002726, -0.00002135, -0.00001088, 0.00001928, -0.00001502, -0.00000356, 0.00000086, -0.00000416, 0.00000622, -0.00001855, 0.00000543, 0.00001238, -0.00002981, 0.00001918, 0.00000889, -0.00004285, 0.00002929, -0.00000240, -0.00004707, 0.00005600, -0.00000436, -0.00006179, 0.00006231, 0.00000237, -0.00005709, 0.00006340, -0.00000197, -0.00007135, 0.00004201, 0.00001853, -0.00002962, 0.00003089, -0.00001135, -0.00003600, 0.00003321, 0.00000744, -0.00002651, 0.00001413, 0.00000518, -0.00002431, -0.00000592, 0.00000920, 0.00000155, -0.00000578, -0.00000183, 0.00000255, -0.00000618, -0.00000781, 0.00000154, 0.00000055, -0.00000170, 0.00000100, 0.00000061, 0.00000310, 0.00000201, -0.00000606, -0.00000405, 0.00000395, 0.00000064, -0.00001278, -0.00001533, 0.00000176, 0.00000216, -0.00001517, 0.00000053, 0.00001710, -0.00001135, -0.00002304, 0.00000660, 0.00001948, 0.00000597, -0.00000515, -0.00000404, 0.00000190, 0.00000239, 0.00000218, 0.00000021, -0.00000535, -0.00000000, 0.00000598, -0.00000024, 0.00000032, 0.00000552, 0.00000048, -0.00000227, -0.00000060, -0.00000456, },
		{ -0.00000318, -0.00000475, 0.00000530, 0.00000758, -0.00000170, -0.00000238, -0.00000401, -0.00000569, 0.00000405, -0.00000113, -0.00001127, -0.00000168, -0.00000344, -0.00001323, -0.00000693, -0.00000151, -0.00000205, -0.00000392, -0.00000415, -0.00001745, -0.00003550, -0.00000692, -0.00003644, -0.00016264, -0.00010793, 0.00012104, 0.00021128, 0.00014636, 0.00004855, 0.00009317, 0.00027527, 0.00013147, -0.00020115, -0.00013412, 0.00000652, -0.00006552, -0.00001990, 0.00001045, -0.00005159, 0.00000852, 0.00000283, -0.00004924, 0.00001616, 0.00000601, -0.00002841, 0.00001262, 0.00000030, -0.00000826, -0.00000185, -0.00000861, 0.00001514, -0.00001549, -0.00001661, 0.00003370, -0.00002431, -0.00001462, 0.00004253, -0.00004045, -0.00001002, 0.00004540, -0.00005703, 0.00000364, 0.00005999, -0.00006682, 0.00000006, 0.00006234, -0.00005797, 0.00000808, 0.00005948, -0.00006326, -0.00001779, 0.00004590, -0.00001929, 0.00000782, 0.00001714, -0.00003862, 0.00000671, 0.00002670, -0.00002278, -0.00000264, 0.00001655, -0.00001584, -0.00001701, 0.00000414, 0.00000682, -0.00000233, -0.00000611, 0.00000175, -0.00000046, -0.00000986, -0.00000304, 0.00000249, -0.00000172, -0.00000008, 0.00000130, 0.00000102, 0.00000330, -0.00000071, -0.00000599, -0.00000228, 0.00000440, -0.00000189, -0.00001880, -0.00001233, 0.00000766, -0.00000394, -0.00001219, 0.00001072, 0.00000548, -0.00002193, -0.00001091, 0.00001590, 0.00001599, -0.00000011, -0.00000720, -0.00000019, 0.00000385, 0.00000161, 0.00000111, -0.00000304, -0.00000338, 0.00000409, 0.00000294, -0.00000077, 0.00000334, 0.00000279, -0.00000115, -0.00000028, },
		{ 0.00000033, -0.00000110, -0.00000606, 0.00000089, 0.00000930, 0.00000084, -0.00000311, -0.00000106, -0.00000734, -0.00000046, 0.00000542, -0.00000920, -0.00000833, 0.00000033, -0.00000873, -0.00001255, -0.00000512, 0.00000118, -0.00000414, -0.00000700, -0.00000176, -0.00003083, -0.00003274, 0.00000373, -0.00009009, -0.00018319, -0.00002945, 0.00018275, 0.00022734, 0.00010722, 0.00003538, 0.00023496, 0.00026648, -0.00012236, -0.00022188, -0.00000159, -0.00003857, -0.00006500, 0.00002832, -0.00003696, -0.00002957, 0.00002874, -0.00004159, -0.00001792, 0.00002550, -0.00002200, -0.00000564, 0.00000661, -0.00000381, 0.00000123, -0.00002024, 0.00001379, 0.00000883, -0.00004298, 0.00002820, 0.00001521, -0.00005179, 0.00003940, 0.00000639, -0.00005850, 0.00005130, -0.00000568, -0.00006273, 0.00006720, -0.00000393, -0.00006567, 0.00006170, -0.00000122, -0.00004815, 0.00006365, -0.00000410, -0.00006129, 0.00002765, 0.00000933, -0.00000916, 0.00002542, -0.00002085, -0.00001779, 0.00002867, -0.00000763, -0.00001683, 0.00001614, -0.00000249, -0.00002051, -0.00000469, 0.00000643, 0.00000352, -0.00000574, -0.00000329, 0.00000311, -0.00000628, -0.00000832, 0.00000121, 0.00000018, -0.00000157, 0.00000152, 0.00000086, 0.00000165, 0.00000242, -0.00000227, -0.00000624, -0.00000059, 0.00000618, -0.00000908, -0.00002354, -0.00000276, 0.00000846, -0.00000963, -0.00000318, 0.00001076, -0.00000921, -0.00002004, 0.00000208, 0.00001887, 0.00001030, -0.00000563, -0.00000524, 0.00000386, 0.00000344, 0.00000056, -0.00000138, -0.00000370, 0.00000080, 0.00000396, 0.00000063, 0.00000134, 0.00000287, 0.00000019, },
		{ 0.00000061, 0.00000055, 0.00000129, -0.00000485, -0.00000409, 0.00000834, 0.00000515, -0.00000439, 0.00000010, -0.00000428, -0.00000660, 0.00000613, -0.00000116, -0.00001245, -0.00000331, -0.00000235, -0.00001129, -0.00001243, -0.00000072, 0.00000256, -0.00001214, -0.00000104, -0.00000696, -0.00004752, -0.00001167, -0.00001267, -0.00015106, -0.00015944, 0.00005092, 0.00024562, 0.00021858, 0.00003928, 0.00013601, 0.00034726, 0.00003922, -0.00028191, -0.00007105, 0.00001570, -0.00009511, 0.00000394, 0.00000639, -0.00005964, 0.00001973, -0.00000729, -0.00004488, 0.00001783, -0.00000196, -0.00001481, 0.00000157, -0.00000493, 0.00001387, -0.00002145, -0.00001217, 0.00003936, -0.00003977, -0.00001666, 0.00005848, -0.00004677, -0.00001294, 0.00005958, -0.00006005, -0.00000322, 0.00005981, -0.00007249, 0.00000298, 0.00006551, -0.00006752, -0.00000028, 0.00005385, -0.00005054, 0.00001566, 0.00004927, -0.00005215, -0.00001603, 0.00002291, -0.00000794, 0.00001942, 0.00000075, -0.00002847, 0.00001715, 0.00000825, -0.00002245, 0.00000735, 0.00000858, -0.00001705, -0.00001139, 0.00000150, 0.00000572, -0.00000043, -0.00000677, 0.00000097, -0.00000042, -0.00000947, -0.00000337, 0.00000144, -0.00000158, 0.00000040, 0.00000170, 0.00000057, 0.00000193, 0.00000173, -0.00000402, -0.00000689, 0.00000412, 0.00000425, -0.00002041, -0.00001972, 0.00000725, 0.00000313, -0.00000906, 0.00000353, 0.00000142, -0.00001631, -0.00001088, 0.00001060, 0.00001777, 0.00000337, -0.00000809, -0.00000062, 0.00000575, 0.00000189, -0.00000151, -0.00000301, -0.00000108, 0.00000278, 0.00000208, 0.00000102, 0.00000203, },
		{ 0.00000165, 0.00000059, -0.00000015, 0.00000249, -0.00000114, -0.00000708, 0.00000360, 0.00000958, -0.00000309, -0.00000247, 0.00000054, -0.00000830, 0.00000032, 0.00000572, -0.00000886, -0.00000971, -0.00000197, -0.00000456, -0.00001421, -0.00001067, 0.00000733, -0.00000629, -0.00001367, 0.00000882, -0.00002923, -0.00004590, 0.00000822, -0.00006186, -0.00018719, -0.00011247, 0.00014278, 0.00030481, 0.00014214, 0.00004478, 0.00032073, 0.00023926, -0.00024524, -0.00020083, 0.00004668, -0.00007000, -0.00005735, 0.00003854, -0.00004972, -0.00001871, 0.00002090, -0.00004259, -0.00000679, 0.00000898, -0.00001122, 0.00000093, -0.00001709, 0.00001697, 0.00000487, -0.00004674, 0.00003804, 0.00000790, -0.00006722, 0.00005418, 0.00001318, -0.00007053, 0.00005955, 0.00000091, -0.00006904, 0.00006870, -0.00000898, -0.00007104, 0.00006925, -0.00000285, -0.00006030, 0.00005159, -0.00000398, -0.00003105, 0.00005446, -0.00000613, -0.00004329, 0.00000928, -0.00000039, 0.00001157, 0.00001694, -0.00002549, 0.00000076, 0.00001869, -0.00001964, -0.00000510, 0.00001453, -0.00000996, -0.00001463, -0.00000309, 0.00000274, 0.00000423, -0.00000454, -0.00000400, 0.00000208, -0.00000575, -0.00000745, -0.00000019, -0.00000053, -0.00000077, 0.00000179, 0.00000107, 0.00000047, 0.00000237, 0.00000089, -0.00000739, -0.00000418, 0.00000919, -0.00000582, -0.00002743, -0.00000796, 0.00001056, -0.00000204, -0.00000447, 0.00000177, -0.00000779, -0.00001412, -0.00000200, 0.00001492, 0.00001402, -0.00000301, -0.00000665, 0.00000399, 0.00000545, -0.00000084, -0.00000348, -0.00000172, 0.00000141, 0.00000227, 0.00000152, },
		{ 0.00000178, 0.00000162, 0.00000086, -0.00000083, 0.00000145, 0.00000270, -0.00000567, -0.00000300, 0.00001038, 0.00000221, -0.00000601, 0.00000183, -0.00000423, -0.00000595, 0.00000561, -0.00000106, -0.00001089, -0.00000752, -0.00000147, -0.00000681, -0.00001904, -0.00000049, 0.00000751, -0.00001969, 0.00000022, 0.00000233, -0.00005209, -0.00002247, 0.00000011, -0.00011755, -0.00019980, -0.00004573, 0.00026568, 0.00029498, 0.00005438, 0.00020306, 0.00038124, -0.00007962, -0.00032347, -0.00000734, 0.00000488, -0.00010558, 0.00002105, -0.00000671, -0.00005031, 0.00001855, -0.00002079, -0.00002391, 0.00000574, -0.00001113, 0.00001135, -0.00002053, -0.00000919, 0.00004243, -0.00004764, -0.00001265, 0.00006357, -0.00006624, -0.00001208, 0.00007858, -0.00006841, -0.00000751, 0.00007048, -0.00007484, 0.00000432, 0.00006674, -0.00007892, 0.00000129, 0.00006181, -0.00005841, -0.00000137, 0.00003784, -0.00003471, 0.00002112, 0.00003252, -0.00003466, -0.00001351, -0.00000147, 0.00000400, 0.00002662, -0.00001251, -0.00001440, 0.00002157, -0.00000975, -0.00001775, 0.00001409, -0.00000070, -0.00001603, -0.00000545, -0.00000110, 0.00000358, 0.00000093, -0.00000593, -0.00000033, -0.00000122, -0.00000768, -0.00000360, -0.00000049, -0.00000111, 0.00000104, 0.00000185, 0.00000032, 0.00000071, 0.00000333, -0.00000232, -0.00000959, 0.00000322, 0.00000733, -0.00001949, -0.00002372, 0.00000337, 0.00000815, -0.00000300, -0.00000333, -0.00000464, -0.00000987, -0.00000868, 0.00000400, 0.00001682, 0.00000833, -0.00000701, -0.00000244, 0.00000704, 0.00000279, -0.00000394, -0.00000297, 0.00000063, 0.00000196, },
		{ 0.00000174, 0.00000204, 0.00000150, 0.00000099, -0.00000045, -0.00000060, 0.00000366, -0.00000075, -0.00000691, 0.00000576, 0.00000838, -0.00000572, -0.00000171, 0.00000073, -0.00000694, 0.00000048, 0.00000331, -0.00000568, -0.00001083, -0.00000670, 0.00000113, -0.00001489, -0.00001663, 0.00001208, -0.00000754, -0.00001868, 0.00001390, -0.00002370, -0.00005253, -0.00000282, -0.00003132, -0.00016548, -0.00020160, 0.00007815, 0.00036904, 0.00019274, 0.00009782, 0.00038980, 0.00015822, -0.00034008, -0.00015253, 0.00006403, -0.00009110, -0.00003806, 0.00002710, -0.00004827, -0.00000425, -0.00000485, -0.00002736, 0.00000460, -0.00002175, 0.00001249, 0.00000618, -0.00004845, 0.00004516, 0.00000574, -0.00007364, 0.00006421, 0.00000312, -0.00008607, 0.00007785, 0.00000699, -0.00008170, 0.00007446, -0.00000590, -0.00007151, 0.00007713, -0.00001107, -0.00007021, 0.00006093, -0.00000088, -0.00004710, 0.00003379, -0.00000480, -0.00000977, 0.00003850, -0.00000660, -0.00002244, -0.00000955, -0.00000796, 0.00002728, 0.00000754, -0.00002303, 0.00001527, 0.00000508, -0.00002583, 0.00000546, 0.00000943, -0.00001507, -0.00000812, -0.00000179, -0.00000069, 0.00000362, -0.00000287, -0.00000380, 0.00000004, -0.00000488, -0.00000571, -0.00000210, -0.00000144, 0.00000025, 0.00000196, 0.00000121, -0.00000035, 0.00000212, 0.00000284, -0.00000757, -0.00000609, 0.00000942, -0.00000368, -0.00002692, -0.00001277, 0.00000876, 0.00000568, -0.00000358, -0.00000760, -0.00000732, -0.00000697, -0.00000506, 0.00000883, 0.00001665, 0.00000160, -0.00000756, 0.00000301, 0.00000705, -0.00000163, -0.00000479, -0.00000072, },
		{ -0.00000304, 0.00000075, 0.00000250, 0.00000196, 0.00000040, 0.00000024, -0.00000125, 0.00000144, 0.00000346, -0.00000527, -0.00000112, 0.00001036, -0.00000053, -0.00000579, 0.00000148, -0.00000353, -0.00000314, 0.00000196, -0.00000086, -0.00000674, -0.00001334, -0.00000131, -0.00000134, -0.00002417, -0.00000152, 0.00001013, -0.00002263, -0.00000120, 0.00000638, -0.00004539, -0.00003728, -0.00000191, -0.00006692, -0.00022386, -0.00015608, 0.00025854, 0.00036212, 0.00011820, 0.00029277, 0.00034954, -0.00020573, -0.00031160, 0.00003462, -0.00002097, -0.00008752, 0.00002035, -0.00002443, -0.00002082, -0.00000074, -0.00003319, 0.00001234, -0.00002170, -0.00001814, 0.00004620, -0.00005146, -0.00001067, 0.00007202, -0.00007580, -0.00000963, 0.00008252, -0.00008922, -0.00000396, 0.00008962, -0.00008375, -0.00000012, 0.00007251, -0.00008051, 0.00001069, 0.00006464, -0.00007446, -0.00000180, 0.00004977, -0.00004132, -0.00000353, 0.00001824, -0.00001401, 0.00002323, 0.00001424, -0.00001480, -0.00001180, -0.00002160, 0.00001363, 0.00002748, -0.00001880, 0.00000092, 0.00001871, -0.00002345, -0.00001019, 0.00001572, -0.00000898, -0.00001285, -0.00000069, -0.00000308, 0.00000108, 0.00000131, -0.00000407, -0.00000163, -0.00000248, -0.00000527, -0.00000371, -0.00000263, -0.00000065, 0.00000164, 0.00000191, 0.00000019, -0.00000009, 0.00000400, -0.00000127, -0.00001010, 0.00000199, 0.00000719, -0.00001663, -0.00002420, -0.00000253, 0.00001049, 0.00000397, -0.00000857, -0.00001091, -0.00000404, -0.00000541, -0.00000256, 0.00001404, 0.00001332, -0.00000449, -0.00000410, 0.00000712, 0.00000328, -0.00000471, },
		{ -0.00000143, -0.00000442, -0.00000154, 0.00000191, 0.00000349, 0.00000021, -0.00000048, -0.00000019, -0.00000089, 0.00000363, -0.00000032, -0.00000484, 0.00000685, 0.00000536, -0.00000587, -0.00000142, -0.00000098, -0.00000273, -0.00000009, -0.00000156, -0.00000045, -0.00001133, -0.00001246, 0.00000595, -0.00001395, -0.00002059, 0.00001210, -0.00000700, -0.00001907, 0.00000986, -0.00001467, -0.00004937, -0.00002838, -0.00000359, -0.00012310, -0.00028224, -0.00000679, 0.00039646, 0.00026799, 0.00021372, 0.00041601, 0.00002111, -0.00037965, -0.00009047, 0.00004059, -0.00008751, -0.00001328, -0.00000872, -0.00002624, 0.00000857, -0.00004641, 0.00000548, 0.00001218, -0.00006194, 0.00004670, 0.00000790, -0.00008153, 0.00007579, 0.00000329, -0.00009527, 0.00008656, -0.00000438, -0.00009531, 0.00009353, -0.00000127, -0.00008270, 0.00007990, -0.00001177, -0.00006497, 0.00007434, -0.00001098, -0.00006076, 0.00004374, 0.00000187, -0.00002944, 0.00001237, -0.00000322, 0.00001054, 0.00002017, -0.00000468, -0.00000410, -0.00002491, -0.00001183, 0.00003409, -0.00000028, -0.00001346, 0.00002275, -0.00000923, -0.00002547, 0.00001212, 0.00000207, -0.00001663, -0.00000234, -0.00000107, -0.00000316, 0.00000192, -0.00000133, -0.00000288, -0.00000218, -0.00000395, -0.00000379, -0.00000389, -0.00000234, 0.00000103, 0.00000210, 0.00000130, -0.00000078, 0.00000186, 0.00000339, -0.00000688, -0.00000627, 0.00000724, -0.00000280, -0.00002294, -0.00001630, 0.00000431, 0.00001166, -0.00000148, -0.00001515, -0.00000738, -0.00000054, -0.00000717, 0.00000249, 0.00001788, 0.00000640, -0.00000697, 0.00000181, 0.00000677, },
		{ 0.00000549, 0.00000264, -0.00000299, -0.00000335, -0.00000082, 0.00000442, 0.00000227, -0.00000223, 0.00000005, -0.00000072, 0.00000098, 0.00000297, -0.00000350, 0.00000179, 0.00000759, -0.00000242, -0.00000358, -0.00000148, -0.00000197, 0.00000147, -0.00000411, -0.00000075, -0.00000130, -0.00001861, -0.00000243, 0.00000286, -0.00002481, -0.00000448, 0.00000813, -0.00001792, -0.00000573, 0.00000290, -0.00002692, -0.00005314, -0.00001773, -0.00000965, -0.00022830, -0.00026117, 0.00020938, 0.00040702, 0.00025505, 0.00039682, 0.00023142, -0.00031511, -0.00024230, 0.00003591, -0.00005206, -0.00003863, -0.00000556, -0.00004039, 0.00002605, -0.00003921, -0.00003993, 0.00005805, -0.00006149, -0.00002034, 0.00008300, -0.00008580, -0.00000865, 0.00009638, -0.00009984, -0.00000401, 0.00009191, -0.00010311, 0.00000535, 0.00008925, -0.00008853, 0.00000673, 0.00006517, -0.00007495, 0.00001419, 0.00005422, -0.00005999, -0.00000613, 0.00003225, -0.00002029, -0.00000646, -0.00000015, 0.00000659, 0.00002171, -0.00000058, 0.00000292, -0.00001155, -0.00003330, 0.00001835, 0.00002209, -0.00001645, 0.00001438, 0.00000971, -0.00003046, -0.00000220, 0.00001207, -0.00001451, -0.00000828, 0.00000216, -0.00000412, -0.00000130, 0.00000069, -0.00000189, -0.00000253, -0.00000363, -0.00000302, -0.00000373, -0.00000436, -0.00000049, 0.00000201, 0.00000199, 0.00000014, -0.00000038, 0.00000378, -0.00000093, -0.00000871, 0.00000055, 0.00000453, -0.00001267, -0.00002181, -0.00000866, 0.00001016, 0.00000966, -0.00001159, -0.00001549, 0.00000029, -0.00000267, -0.00000779, 0.00001079, 0.00001661, -0.00000117, -0.00000399, },
		{ -0.00000408, 0.00000109, 0.00000387, -0.00000009, -0.00000228, -0.00000398, 0.00000208, 0.00000581, -0.00000192, -0.00000232, 0.00000065, -0.00000053, 0.00000246, -0.00000032, -0.00000056, 0.00000653, 0.00000072, -0.00000323, -0.00000199, -0.00000416, 0.00000361, -0.00000128, -0.00000754, 0.00000487, -0.00001071, -0.00001716, 0.00000709, -0.00001084, -0.00002045, 0.00000493, -0.00000402, -0.00001351, -0.00000417, -0.00000346, -0.00003671, -0.00006232, 0.00001621, -0.00007106, -0.00033664, -0.00010610, 0.00036382, 0.00037862, 0.00040345, 0.00037509, -0.00016336, -0.00034571, -0.00003185, -0.00002373, -0.00005050, 0.00000880, -0.00006496, 0.00001731, 0.00001472, -0.00009694, 0.00005202, 0.00001307, -0.00010379, 0.00008526, 0.00000521, -0.00010938, 0.00010223, -0.00000172, -0.00010691, 0.00009952, -0.00001232, -0.00009291, 0.00009710, -0.00000881, -0.00007341, 0.00007393, -0.00001465, -0.00005076, 0.00006095, -0.00000838, -0.00004524, 0.00002146, 0.00000473, -0.00001162, -0.00000751, 0.00000030, 0.00002517, 0.00000422, -0.00000057, 0.00000777, -0.00003358, -0.00001196, 0.00003073, -0.00000446, 0.00000093, 0.00002240, -0.00002101, -0.00001987, 0.00001337, -0.00000571, -0.00001469, 0.00000178, -0.00000088, -0.00000446, -0.00000042, -0.00000032, -0.00000164, -0.00000384, -0.00000314, -0.00000230, -0.00000507, -0.00000306, 0.00000128, 0.00000222, 0.00000136, -0.00000080, 0.00000159, 0.00000273, -0.00000551, -0.00000520, 0.00000344, -0.00000290, -0.00001702, -0.00001800, -0.00000111, 0.00001457, 0.00000054, -0.00001931, -0.00000708, 0.00000354, -0.00000873, -0.00000214, 0.00001748, 0.00000960, },
		{ 0.00001685, 0.00000243, -0.00000126, 0.00000207, 0.00000082, 0.00000092, -0.00000378, -0.00000323, 0.00000678, 0.00000231, -0.00000503, -0.00000029, 0.00000046, 0.00000078, 0.00000089, -0.00000045, 0.00000557, 0.00000224, -0.00000302, -0.00000005, -0.00000637, 0.00000035, 0.00000654, -0.00001061, -0.00000115, 0.00000341, -0.00002067, -0.00000537, 0.00000376, -0.00001878, -0.00000981, 0.00000073, -0.00000497, -0.00001303, -0.00000755, -0.00000062, -0.00006632, -0.00004159, 0.00003942, -0.00021128, -0.00034016, 0.00010439, 0.00042132, 0.00048217, 0.00050083, 0.00002001, -0.00038514, -0.00012356, -0.00001293, -0.00007713, 0.00003510, -0.00005722, -0.00004863, 0.00008228, -0.00009204, -0.00003815, 0.00010430, -0.00010637, -0.00001686, 0.00011389, -0.00011615, -0.00000386, 0.00011021, -0.00011393, 0.00000218, 0.00008971, -0.00010414, 0.00001295, 0.00007784, -0.00008105, 0.00001075, 0.00005034, -0.00005893, 0.00001395, 0.00003818, -0.00003893, -0.00001094, 0.00001323, -0.00000043, -0.00000925, -0.00001312, 0.00002219, 0.00001735, -0.00000852, 0.00001485, -0.00001234, -0.00003504, 0.00001682, 0.00001243, -0.00000650, 0.00002343, -0.00000236, -0.00003035, 0.00000368, 0.00000450, -0.00001643, -0.00000338, 0.00000306, -0.00000426, -0.00000326, -0.00000064, -0.00000006, -0.00000280, -0.00000426, -0.00000151, -0.00000366, -0.00000528, -0.00000072, 0.00000203, 0.00000208, 0.00000021, -0.00000024, 0.00000282, -0.00000104, -0.00000612, -0.00000103, 0.00000054, -0.00000850, -0.00001766, -0.00001344, 0.00000770, 0.00001234, -0.00001245, -0.00001681, 0.00000280, -0.00000190, -0.00001074, 0.00000830, },
		{ -0.00000373, 0.00001531, 0.00001027, 0.00000125, 0.00000102, -0.00000122, 0.00000208, 0.00000087, -0.00000678, 0.00000220, 0.00000749, -0.00000382, -0.00000405, 0.00000106, 0.00000093, 0.00000048, -0.00000096, 0.00000552, 0.00000475, -0.00000461, 0.00000132, -0.00000345, -0.00000751, 0.00001012, -0.00000261, -0.00001262, 0.00000767, -0.00000844, -0.00001883, 0.00000243, -0.00000531, -0.00001547, -0.00000959, 0.00000130, -0.00000306, -0.00002521, 0.00000756, -0.00001847, -0.00009401, 0.00002324, -0.00002353, -0.00034726, -0.00021421, 0.00025482, 0.00053424, 0.00066449, 0.00025194, -0.00037627, -0.00023483, 0.00000742, -0.00011894, 0.00001953, 0.00002088, -0.00013173, 0.00007663, 0.00001130, -0.00014649, 0.00010366, 0.00000670, -0.00013847, 0.00011970, -0.00000100, -0.00012599, 0.00011805, -0.00000742, -0.00010614, 0.00009948, -0.00001806, -0.00007965, 0.00008728, -0.00001316, -0.00005624, 0.00005765, -0.00001339, -0.00003204, 0.00004022, -0.00000383, -0.00002734, -0.00000082, 0.00000663, 0.00000246, -0.00002110, 0.00000456, 0.00003128, -0.00000570, 0.00000464, 0.00001170, -0.00003410, -0.00000962, 0.00001890, -0.00000413, 0.00001631, 0.00001580, -0.00002773, -0.00001179, 0.00000930, -0.00001210, -0.00001023, 0.00000395, -0.00000105, -0.00000480, -0.00000283, -0.00000003, -0.00000045, -0.00000445, -0.00000255, -0.00000154, -0.00000543, -0.00000350, 0.00000095, 0.00000223, 0.00000141, -0.00000044, 0.00000125, 0.00000132, -0.00000373, -0.00000362, -0.00000094, -0.00000351, -0.00001077, -0.00001774, -0.00000595, 0.00001387, 0.00000126, -0.00001938, -0.00000555, 0.00000426, -0.00001011, },
		{ -0.00000377, -0.00001092, 0.00000702, 0.00001362, 0.00000608, 0.00000370, -0.00000249, -0.00000159, 0.00000556, -0.00000391, -0.00000506, 0.00000796, 0.00000188, -0.00000659, -0.00000149, 0.00000187, 0.00000133, -0.00000251, 0.00000346, 0.00000944, -0.00000446, -0.00000238, 0.00000337, -0.00001135, 0.00000287, 0.00001019, -0.00001488, -0.00000262, 0.00000546, -0.00001739, -0.00001038, 0.00000046, -0.00000601, -0.00001721, -0.00001077, 0.00001340, -0.00002161, -0.00002155, 0.00002548, -0.00007241, -0.00006481, 0.00006407, -0.00017200, -0.00037582, -0.00006526, 0.00040449, 0.00080265, 0.00058728, -0.00027326, -0.00040091, 0.00002608, -0.00010842, -0.00008731, 0.00012080, -0.00012730, -0.00004579, 0.00013671, -0.00014932, -0.00002786, 0.00014141, -0.00014651, -0.00000896, 0.00013216, -0.00013524, 0.00000175, 0.00011060, -0.00011422, 0.00000661, 0.00007670, -0.00009153, 0.00001644, 0.00005833, -0.00006279, 0.00001062, 0.00003165, -0.00003597, 0.00001024, 0.00002027, -0.00001625, -0.00001496, -0.00000349, 0.00001362, -0.00001070, -0.00001826, 0.00002948, 0.00001167, -0.00000866, 0.00001929, -0.00001301, -0.00002824, 0.00000947, 0.00000162, 0.00000765, 0.00002675, -0.00001358, -0.00002458, 0.00000568, -0.00000474, -0.00001496, 0.00000089, 0.00000254, -0.00000382, -0.00000466, -0.00000225, 0.00000097, -0.00000242, -0.00000414, -0.00000092, -0.00000352, -0.00000535, -0.00000125, 0.00000169, 0.00000209, 0.00000045, 0.00000016, 0.00000140, -0.00000129, -0.00000322, -0.00000268, -0.00000352, -0.00000489, -0.00001289, -0.00001584, 0.00000393, 0.00001113, -0.00001168, -0.00001411, 0.00000359, },
		{ -0.00000240, 0.00000162, -0.00001128, -0.00000219, 0.00001137, 0.00000869, 0.00000822, 0.00000134, -0.00000689, 0.00000422, 0.00000381, -0.00000803, 0.00000202, 0.00000702, -0.00000427, -0.00000532, 0.00000027, 0.00000347, -0.00000184, -0.00000192, 0.00001209, 0.00000195, -0.00000915, 0.00000586, -0.00000527, -0.00000935, 0.00001417, -0.00000329, -0.00001489, 0.00000577, -0.00000422, -0.00001648, -0.00000828, 0.00000217, -0.00000673, -0.00002859, 0.00000774, 0.00000890, -0.00004657, 0.00001490, -0.00000090, -0.00011069, 0.00001314, 0.00000625, -0.00030117, -0.00032755, 0.00006636, 0.00073944, 0.00097853, 0.00004678, -0.00059196, -0.00006094, 0.00001757, -0.00022737, 0.00011212, 0.00001920, -0.00019574, 0.00014036, -0.00000126, -0.00018474, 0.00014914, -0.00000488, -0.00015862, 0.00014166, -0.00000844, -0.00012794, 0.00011895, -0.00001155, -0.00009350, 0.00008594, -0.00001952, -0.00005882, 0.00006614, -0.00001297, -0.00003528, 0.00003485, -0.00000820, -0.00001286, 0.00001714, 0.00000135, -0.00001082, -0.00001830, 0.00000645, 0.00001044, -0.00002554, 0.00000808, 0.00002859, -0.00000816, 0.00000939, 0.00000895, -0.00002720, -0.00000675, 0.00000263, 0.00000009, 0.00002859, 0.00000617, -0.00002826, -0.00000435, 0.00000132, -0.00001589, -0.00000472, 0.00000442, -0.00000137, -0.00000459, -0.00000483, -0.00000043, 0.00000039, -0.00000396, -0.00000216, -0.00000148, -0.00000507, -0.00000363, 0.00000022, 0.00000200, 0.00000144, 0.00000020, 0.00000079, -0.00000030, -0.00000189, -0.00000226, -0.00000494, -0.00000420, -0.00000543, -0.00001572, -0.00000916, 0.00000982, -0.00000006, -0.00001561, },
		{ -0.00000998, -0.00000776, 0.00000298, -0.00000777, -0.00000850, 0.00000644, 0.00000778, 0.00000970, 0.00000936, -0.00000688, -0.00000365, 0.00000883, -0.00000303, -0.00000498, 0.00000641, 0.00000099, -0.00000595, -0.00000382, 0.00000309, 0.00000259, -0.00000621, 0.00000750, 0.00001223, -0.00001040, -0.00000109, 0.00000508, -0.00001394, 0.00000461, 0.00001077, -0.00001459, -0.00000563, 0.00000382, -0.00000832, -0.00001703, -0.00000630, 0.00001082, -0.00002527, -0.00002311, 0.00003114, -0.00002860, -0.00003392, 0.00004103, -0.00006677, -0.00007868, 0.00005084, -0.00011046, -0.00035980, -0.00029912, 0.00036485, 0.00120065, 0.00063996, -0.00061847, -0.00035999, 0.00018160, -0.00023409, -0.00007910, 0.00020195, -0.00020194, -0.00003244, 0.00017807, -0.00019849, -0.00001206, 0.00016167, -0.00017231, 0.00000038, 0.00013400, -0.00013806, 0.00000579, 0.00009782, -0.00010018, 0.00000750, 0.00005620, -0.00006792, 0.00001474, 0.00003531, -0.00003801, 0.00000654, 0.00001338, -0.00001123, 0.00000443, 0.00000421, 0.00000306, -0.00001688, -0.00001522, 0.00001908, -0.00000978, -0.00001559, 0.00002767, 0.00000633, -0.00000268, 0.00001689, -0.00001223, -0.00001654, -0.00000158, -0.00000721, 0.00002143, 0.00002459, -0.00002045, -0.00001583, 0.00000321, -0.00001332, -0.00001113, 0.00000388, 0.00000129, -0.00000321, -0.00000552, -0.00000375, 0.00000113, -0.00000157, -0.00000338, -0.00000105, -0.00000332, -0.00000480, -0.00000185, 0.00000104, 0.00000188, 0.00000085, 0.00000062, -0.00000012, -0.00000142, -0.00000080, -0.00000426, -0.00000670, -0.00000227, -0.00000841, -0.00001559, -0.00000033, 0.00000615, },
		{ 0.00000326, -0.00000341, -0.00000899, 0.00000196, -0.00000340, -0.00001170, 0.00000126, 0.00000594, 0.00000619, 0.00001515, 0.00000163, -0.00001094, 0.00000519, 0.00000545, -0.00000606, 0.00000113, 0.00000358, -0.00000265, -0.00000591, -0.00000180, 0.00000638, -0.00000381, -0.00000242, 0.00001712, -0.00000100, -0.00001137, 0.00000800, -0.00000646, -0.00000837, 0.00001346, -0.00000252, -0.00001336, -0.00000155, 0.00000167, -0.00001063, -0.00002276, 0.00001009, 0.00000298, -0.00004633, 0.00001270, 0.00001778, -0.00005940, 0.00001426, 0.00000966, -0.00010096, -0.00001736, 0.00001014, -0.00019454, -0.00044477, -0.00018765, 0.00098989, 0.00131297, -0.00022187, -0.00078342, 0.00014670, -0.00000211, -0.00032814, 0.00020912, 0.00000487, -0.00024391, 0.00019443, -0.00001919, -0.00020381, 0.00017772, -0.00001771, -0.00016028, 0.00014532, -0.00001428, -0.00011532, 0.00010417, -0.00001241, -0.00007225, 0.00006181, -0.00001602, -0.00003524, 0.00003853, -0.00000857, -0.00001515, 0.00001097, -0.00000055, 0.00000305, -0.00000312, 0.00000554, 0.00000156, -0.00002781, 0.00000351, 0.00001194, -0.00002069, 0.00000970, 0.00001929, -0.00000419, 0.00001236, 0.00000271, -0.00001553, -0.00000509, -0.00001309, 0.00000632, 0.00003467, -0.00000281, -0.00002306, 0.00000005, -0.00000839, -0.00001675, 0.00000044, 0.00000377, -0.00000174, -0.00000427, -0.00000615, -0.00000125, 0.00000080, -0.00000270, -0.00000189, -0.00000182, -0.00000431, -0.00000351, -0.00000063, 0.00000146, 0.00000141, 0.00000095, 0.00000025, -0.00000163, -0.00000038, -0.00000157, -0.00000786, -0.00000466, -0.00000167, -0.00001243, -0.00001035, },
		{ -0.00001307, -0.00000447, -0.00000160, -0.00000784, 0.00000080, 0.00000133, -0.00001222, -0.00000435, 0.00000532, 0.00000121, 0.00001314, 0.00001362, -0.00000946, -0.00000460, 0.00000893, -0.00000047, -0.00000240, 0.00000146, -0.00000020, -0.00000309, -0.00000683, 0.00000369, 0.00000394, -0.00000858, 0.00001068, 0.00001271, -0.00001320, -0.00000020, 0.00000393, -0.00001208, 0.00000453, 0.00000759, -0.00000879, -0.00001015, -0.00000116, 0.00000466, -0.00002393, -0.00001428, 0.00002647, -0.00003135, -0.00003219, 0.00004036, -0.00003007, -0.00003966, 0.00003634, -0.00004534, -0.00007757, -0.00000676, -0.00002796, -0.00029598, -0.00058422, 0.00031794, 0.00164058, 0.00064863, -0.00097161, -0.00027394, 0.00031720, -0.00036426, -0.00005555, 0.00026280, -0.00026445, -0.00001294, 0.00019806, -0.00022789, 0.00000422, 0.00016133, -0.00017717, 0.00000760, 0.00011951, -0.00012349, 0.00000643, 0.00007513, -0.00007483, 0.00000434, 0.00003303, -0.00003866, 0.00000849, 0.00001364, -0.00001250, 0.00000008, -0.00000075, 0.00000994, -0.00000154, -0.00000733, 0.00001539, -0.00001586, -0.00002101, 0.00001581, -0.00000620, -0.00000729, 0.00001870, 0.00000252, 0.00000602, 0.00001030, -0.00000918, -0.00000445, -0.00001312, -0.00001190, 0.00003058, 0.00001850, -0.00002107, -0.00000694, -0.00000296, -0.00001952, -0.00000630, 0.00000543, -0.00000006, -0.00000281, -0.00000595, -0.00000482, 0.00000064, -0.00000057, -0.00000225, -0.00000147, -0.00000308, -0.00000398, -0.00000231, 0.00000022, 0.00000141, 0.00000129, 0.00000100, -0.00000137, -0.00000138, 0.00000060, -0.00000559, -0.00000853, -0.00000078, -0.00000470, },
		{ 0.00000042, -0.00000845, -0.00000982, -0.00000466, -0.00000801, -0.00000101, 0.00000634, -0.00000891, -0.00001059, 0.00000441, -0.00000031, 0.00000454, 0.00001999, 0.00000164, -0.00001149, 0.00000400, 0.00000532, -0.00000027, -0.00000075, -0.00000249, 0.00000062, -0.00000569, -0.00000443, 0.00000832, -0.00000467, -0.00000159, 0.00001851, -0.00000293, -0.00001030, 0.00000583, -0.00000594, -0.00000423, 0.00000709, 0.00000009, -0.00000842, -0.00001281, 0.00000804, -0.00000318, -0.00003641, 0.00001582, 0.00001016, -0.00005560, 0.00001250, 0.00001950, -0.00005530, 0.00000145, 0.00000923, -0.00005854, -0.00006935, -0.00001220, -0.00003572, -0.00056555, -0.00047642, 0.00126988, 0.00162637, -0.00051872, -0.00092801, 0.00034963, -0.00004314, -0.00040232, 0.00028934, -0.00001660, -0.00026679, 0.00022604, -0.00003665, -0.00020035, 0.00018227, -0.00002733, -0.00014389, 0.00012925, -0.00001630, -0.00009149, 0.00007681, -0.00000953, -0.00004730, 0.00003265, -0.00000852, -0.00001381, 0.00001071, -0.00000171, 0.00000040, -0.00000858, 0.00000731, 0.00001328, -0.00001667, 0.00000742, 0.00000857, -0.00002862, -0.00000199, 0.00000845, -0.00000909, 0.00000904, 0.00000715, 0.00000318, 0.00001288, -0.00000314, -0.00000277, -0.00000552, -0.00002392, 0.00001211, 0.00003333, -0.00000823, -0.00001394, 0.00000036, -0.00001758, -0.00001518, 0.00000425, 0.00000264, -0.00000211, -0.00000417, -0.00000674, -0.00000214, 0.00000080, -0.00000122, -0.00000157, -0.00000219, -0.00000349, -0.00000321, -0.00000135, 0.00000065, 0.00000126, 0.00000159, -0.00000020, -0.00000238, 0.00000048, -0.00000166, -0.00000942, -0.00000480, },
		{ -0.00000898, -0.00000028, -0.00000192, -0.00000913, -0.00000823, -0.00001051, -0.00000545, 0.00000929, -0.00000090, -0.00001477, 0.00000014, 0.00000233, -0.00000383, 0.00001599, 0.00001476, -0.00000885, -0.00000420, 0.00000520, 0.00000395, 0.00000219, -0.00000613, -0.00000161, 0.00000093, -0.00000926, 0.00000322, 0.00000446, -0.00000791, 0.00001191, 0.00001023, -0.00001115, -0.00000080, -0.00000016, -0.00000509, 0.00000200, 0.00000131, 0.00000133, -0.00001536, -0.00000777, 0.00001748, -0.00002868, -0.00002051, 0.00003441, -0.00003189, -0.00003453, 0.00003371, -0.00002186, -0.00003593, 0.00000568, -0.00000596, -0.00006110, -0.00010521, 0.00007182, -0.00020056, -0.00089427, 0.00028232, 0.00207138, 0.00061241, -0.00128286, -0.00016586, 0.00041194, -0.00047056, -0.00002502, 0.00029223, -0.00029996, 0.00000615, 0.00019346, -0.00023012, 0.00001570, 0.00014144, -0.00015954, 0.00000995, 0.00009244, -0.00009472, 0.00000311, 0.00004768, -0.00004397, -0.00000182, 0.00001208, -0.00001023, -0.00000020, -0.00000285, 0.00000814, -0.00000637, -0.00000862, 0.00002357, -0.00000583, -0.00001325, 0.00001942, -0.00001192, -0.00002153, 0.00000620, -0.00000060, 0.00000318, 0.00000638, 0.00000061, 0.00001361, 0.00000306, -0.00000390, 0.00000421, -0.00002197, -0.00001190, 0.00003254, 0.00001071, -0.00001558, -0.00000015, -0.00001109, -0.00002264, -0.00000183, 0.00000572, -0.00000111, -0.00000285, -0.00000614, -0.00000530, -0.00000013, 0.00000022, -0.00000107, -0.00000172, -0.00000281, -0.00000322, -0.00000250, -0.00000062, 0.00000070, 0.00000157, 0.00000121, -0.00000203, -0.00000124, 0.00000081, -0.00000650, },
		{ -0.00000237, -0.00000968, -0.00000464, 0.00000117, -0.00000431, -0.00000833, -0.00001277, -0.00001248, 0.00000660, 0.00000917, -0.00001266, -0.00000732, 0.00000494, -0.00000633, 0.00000503, 0.00002103, 0.00000151, -0.00000788, 0.00000037, 0.00000379, 0.00000832, -0.00000340, -0.00000974, 0.00000405, -0.00000453, -0.00000644, 0.00000778, -0.00000370, 0.00000114, 0.00001489, -0.00000279, -0.00000606, -0.00000110, -0.00000306, 0.00000211, -0.00000375, 0.00000407, -0.00000175, -0.00002439, 0.00001278, 0.00000324, -0.00004337, 0.00001603, 0.00001186, -0.00004988, 0.00000327, 0.00001211, -0.00002922, -0.00002631, -0.00000225, 0.00001496, -0.00014058, -0.00005285, 0.00015431, -0.00068840, -0.00075697, 0.00155239, 0.00187647, -0.00079782, -0.00101185, 0.00050892, -0.00009071, -0.00043734, 0.00033439, -0.00003734, -0.00026086, 0.00022740, -0.00004777, -0.00017617, 0.00016144, -0.00003042, -0.00011383, 0.00009689, -0.00001374, -0.00006198, 0.00004302, -0.00000392, -0.00002375, 0.00000503, 0.00000064, 0.00000176, -0.00001152, 0.00000504, 0.00000935, -0.00001997, 0.00001316, 0.00001712, -0.00002187, 0.00000648, 0.00001052, -0.00002247, -0.00000887, 0.00000254, 0.00000498, 0.00000664, -0.00000388, 0.00001019, 0.00001119, -0.00000554, 0.00000759, -0.00000781, -0.00002750, 0.00001533, 0.00002553, -0.00000886, -0.00000341, -0.00000286, -0.00002452, -0.00001225, 0.00000626, 0.00000157, -0.00000250, -0.00000447, -0.00000673, -0.00000271, 0.00000050, -0.00000004, -0.00000107, -0.00000227, -0.00000286, -0.00000285, -0.00000183, -0.00000028, 0.00000090, 0.00000191, -0.00000040, -0.00000250, 0.00000056, },
		{ -0.00000115, 0.00000002, -0.00000694, -0.00000838, -0.00000045, 0.00000011, -0.00000475, -0.00001168, -0.00001872, -0.00000282, 0.00001525, -0.00000314, -0.00001329, 0.00000337, -0.00000320, -0.00000556, 0.00001759, 0.00001222, -0.00000462, -0.00000282, -0.00000137, 0.00000961, 0.00000737, -0.00001424, -0.00000339, 0.00000448, -0.00000950, 0.00000132, 0.00000343, -0.00000317, 0.00001041, 0.00000505, -0.00000406, -0.00000348, -0.00000607, 0.00000569, -0.00000270, -0.00000619, 0.00001246, -0.00001942, -0.00001359, 0.00002482, -0.00002824, -0.00002244, 0.00002883, -0.00002337, -0.00002872, 0.00000705, -0.00000094, -0.00002329, -0.00005023, 0.00005031, -0.00003445, -0.00023027, 0.00018161, -0.00009857, -0.00119202, 0.00027421, 0.00243518, 0.00054074, -0.00150668, -0.00006450, 0.00045379, -0.00052929, 0.00000182, 0.00028629, -0.00029972, 0.00001843, 0.00016672, -0.00020428, 0.00001869, 0.00010730, -0.00012335, 0.00000651, 0.00005894, -0.00005833, -0.00000310, 0.00002104, -0.00001444, -0.00000876, -0.00000304, 0.00001155, -0.00000847, -0.00001226, 0.00002011, -0.00001050, -0.00000999, 0.00002809, -0.00000733, -0.00001402, 0.00001624, -0.00000609, -0.00001851, -0.00000573, 0.00000554, 0.00001232, -0.00000491, 0.00000014, 0.00001722, -0.00000166, 0.00000271, 0.00000749, -0.00002593, -0.00000823, 0.00002713, 0.00000342, -0.00000600, 0.00000350, -0.00001899, -0.00002290, 0.00000129, 0.00000516, -0.00000167, -0.00000340, -0.00000628, -0.00000511, -0.00000079, 0.00000050, -0.00000009, -0.00000151, -0.00000251, -0.00000269, -0.00000247, -0.00000133, -0.00000012, 0.00000149, 0.00000128, -0.00000201, },
		{ -0.00000026, -0.00000212, -0.00000007, -0.00000341, -0.00000897, -0.00000429, 0.00000113, -0.00000044, -0.00000683, -0.00002003, -0.00001534, 0.00001216, 0.00000918, -0.00001260, -0.00000194, 0.00000135, -0.00001097, 0.00000782, 0.00001772, 0.00000173, -0.00000105, -0.00000492, 0.00000191, 0.00001719, -0.00000606, -0.00001610, 0.00000554, -0.00000226, -0.00000673, 0.00000377, -0.00000052, 0.00000552, 0.00000641, -0.00000015, 0.00000008, -0.00001127, 0.00000190, 0.00000861, -0.00001578, 0.00000790, 0.00000426, -0.00003140, 0.00001309, 0.00000627, -0.00003842, 0.00000750, 0.00000694, -0.00002573, -0.00001774, -0.00000317, 0.00001828, -0.00006896, -0.00002065, 0.00010920, -0.00022444, -0.00009801, 0.00034465, -0.00080829, -0.00098259, 0.00180250, 0.00202903, -0.00101210, -0.00103194, 0.00059413, -0.00012849, -0.00042931, 0.00033378, -0.00005002, -0.00022905, 0.00019805, -0.00004876, -0.00013748, 0.00012003, -0.00002601, -0.00007705, 0.00005554, -0.00000769, -0.00003287, 0.00001016, 0.00000227, -0.00000557, -0.00001534, 0.00000862, 0.00000986, -0.00002437, 0.00000925, 0.00001171, -0.00002190, 0.00001560, 0.00001557, -0.00001950, 0.00000322, 0.00000876, -0.00001274, -0.00001529, -0.00000314, 0.00001695, 0.00000368, -0.00001087, 0.00001372, 0.00000822, -0.00000322, 0.00001338, -0.00001083, -0.00002399, 0.00001484, 0.00001390, -0.00000528, 0.00000595, -0.00000793, -0.00002823, -0.00000918, 0.00000651, 0.00000086, -0.00000291, -0.00000523, -0.00000627, -0.00000271, -0.00000003, 0.00000054, -0.00000036, -0.00000194, -0.00000245, -0.00000254, -0.00000208, -0.00000112, 0.00000032, 0.00000180, },
		{ 0.00000097, 0.00000121, -0.00000141, -0.00000116, -0.00000139, -0.00000697, -0.00000716, -0.00000094, 0.00000164, -0.00000089, -0.00001499, -0.00002451, -0.00000001, 0.00001671, -0.00000462, -0.00000680, 0.00000406, -0.00001076, -0.00000353, 0.00001720, 0.00000702, 0.00000238, -0.00000123, -0.00000832, 0.00001505, 0.00001131, -0.00001999, -0.00000592, 0.00000591, -0.00000637, -0.00000210, 0.00000076, 0.00000534, 0.00000609, -0.00000153, 0.00000631, -0.00000789, -0.00001171, 0.00001693, -0.00000810, -0.00001245, 0.00001997, -0.00001983, -0.00001639, 0.00002183, -0.00001974, -0.00001901, 0.00000676, -0.00000347, -0.00001615, -0.00003866, 0.00003314, -0.00000572, -0.00012342, 0.00011822, -0.00000667, -0.00035930, 0.00028997, -0.00001747, -0.00143511, 0.00030243, 0.00267866, 0.00045130, -0.00161111, 0.00000336, 0.00044119, -0.00052750, 0.00001581, 0.00024852, -0.00026359, 0.00001998, 0.00012488, -0.00015622, 0.00001243, 0.00006669, -0.00007693, -0.00000149, 0.00002575, -0.00002240, -0.00001000, -0.00000034, 0.00000781, -0.00001391, -0.00001077, 0.00002316, -0.00001380, -0.00001484, 0.00002233, -0.00001090, -0.00000635, 0.00002447, -0.00000602, -0.00001117, 0.00000868, -0.00000007, -0.00001396, -0.00001576, 0.00001040, 0.00001764, -0.00001172, 0.00000018, 0.00001574, -0.00000215, 0.00000925, 0.00000573, -0.00002444, -0.00000292, 0.00001644, -0.00000183, 0.00000455, 0.00000413, -0.00002463, -0.00002111, 0.00000251, 0.00000414, -0.00000175, -0.00000440, -0.00000649, -0.00000428, -0.00000115, 0.00000020, 0.00000059, -0.00000083, -0.00000212, -0.00000237, -0.00000236, -0.00000184, -0.00000092, },
		{ -0.00000175, -0.00000046, 0.00000127, 0.00000014, -0.00000145, -0.00000110, -0.00000451, -0.00000773, -0.00000384, 0.00000081, 0.00000283, -0.00000621, -0.00002551, -0.00001549, 0.00001429, 0.00000599, -0.00000766, 0.00000466, -0.00000619, -0.00001331, 0.00001164, 0.00001097, 0.00000286, 0.00000665, -0.00001004, 0.00000088, 0.00002313, -0.00000754, -0.00001964, 0.00000331, 0.00000108, -0.00000479, -0.00000348, 0.00000463, 0.00001050, -0.00000499, 0.00000442, 0.00000601, -0.00002134, 0.00000743, 0.00001287, -0.00002441, 0.00000919, 0.00000743, -0.00002924, 0.00000630, 0.00000480, -0.00001884, -0.00001166, -0.00000337, 0.00001556, -0.00005081, -0.00002220, 0.00008032, -0.00012257, -0.00003910, 0.00020688, -0.00030995, -0.00013234, 0.00050038, -0.00091531, -0.00111301, 0.00198356, 0.00206489, -0.00112376, -0.00099520, 0.00059020, -0.00014395, -0.00038307, 0.00028773, -0.00005046, -0.00017940, 0.00014493, -0.00003921, -0.00009286, 0.00006757, -0.00001578, -0.00004098, 0.00001422, -0.00000056, -0.00000916, -0.00001523, 0.00000664, 0.00000533, -0.00002518, 0.00001306, 0.00001102, -0.00002699, 0.00000955, 0.00000921, -0.00001579, 0.00001440, 0.00001063, -0.00001224, -0.00000106, 0.00000510, -0.00000318, -0.00001932, -0.00000672, 0.00002341, 0.00000148, -0.00001269, 0.00001232, 0.00000514, 0.00000305, 0.00001427, -0.00001308, -0.00001573, 0.00001086, 0.00000181, 0.00000049, 0.00001220, -0.00001273, -0.00002845, -0.00000698, 0.00000525, 0.00000064, -0.00000337, -0.00000638, -0.00000550, -0.00000210, -0.00000071, 0.00000051, 0.00000048, -0.00000128, -0.00000215, -0.00000232, -0.00000220, },
		{ -0.00000230, -0.00000214, -0.00000160, 0.00000011, 0.00000100, -0.00000054, -0.00000122, -0.00000295, -0.00000676, -0.00000573, -0.00000146, 0.00000294, 0.00000170, -0.00001824, -0.00002663, 0.00000259, 0.00001288, -0.00000443, 0.00000395, 0.00000174, -0.00001865, 0.00000108, 0.00001436, 0.00000136, 0.00001005, -0.00000014, -0.00001260, 0.00001831, 0.00001251, -0.00002086, -0.00000948, 0.00000434, 0.00000019, -0.00000663, -0.00000284, 0.00001565, -0.00000014, -0.00000751, 0.00001652, -0.00001250, -0.00001481, 0.00002415, -0.00001172, -0.00001534, 0.00001934, -0.00001401, -0.00001504, 0.00000582, -0.00000156, -0.00001114, -0.00003047, 0.00002766, -0.00000072, -0.00009678, 0.00007731, 0.00001071, -0.00020000, 0.00018835, 0.00001243, -0.00046613, 0.00038192, 0.00002022, -0.00158410, 0.00036545, 0.00276433, 0.00036545, -0.00158410, 0.00002022, 0.00038192, -0.00046613, 0.00001243, 0.00018835, -0.00020000, 0.00001071, 0.00007731, -0.00009678, -0.00000072, 0.00002766, -0.00003047, -0.00001114, -0.00000156, 0.00000582, -0.00001504, -0.00001401, 0.00001934, -0.00001534, -0.00001172, 0.00002415, -0.00001481, -0.00001250, 0.00001652, -0.00000751, -0.00000014, 0.00001565, -0.00000284, -0.00000663, 0.00000019, 0.00000434, -0.00000948, -0.00002086, 0.00001251, 0.00001831, -0.00001260, -0.00000014, 0.00001005, 0.00000136, 0.00001436, 0.00000108, -0.00001865, 0.00000174, 0.00000395, -0.00000443, 0.00001288, 0.00000259, -0.00002663, -0.00001824, 0.00000170, 0.00000294, -0.00000146, -0.00000573, -0.00000676, -0.00000295, -0.00000122, -0.00000054, 0.00000100, 0.00000011, -0.00000160, -0.00000214, },
		{ -0.00000175, -0.00000220, -0.00000232, -0.00000215, -0.00000128, 0.00000048, 0.00000051, -0.00000071, -0.00000210, -0.00000550, -0.00000638, -0.00000337, 0.00000064, 0.00000525, -0.00000698, -0.00002845, -0.00001273, 0.00001220, 0.00000049, 0.00000181, 0.00001086, -0.00001573, -0.00001308, 0.00001427, 0.00000305, 0.00000514, 0.00001232, -0.00001269, 0.00000148, 0.00002341, -0.00000672, -0.00001932, -0.00000318, 0.00000510, -0.00000106, -0.00001224, 0.00001063, 0.00001440, -0.00001579, 0.00000921, 0.00000955, -0.00002699, 0.00001102, 0.00001306, -0.00002518, 0.00000533, 0.00000664, -0.00001523, -0.00000916, -0.00000056, 0.00001422, -0.00004098, -0.00001578, 0.00006757, -0.00009286, -0.00003921, 0.00014493, -0.00017940, -0.00005046, 0.00028773, -0.00038307, -0.00014395, 0.00059020, -0.00099520, -0.00112376, 0.00206489, 0.00198356, -0.00111301, -0.00091531, 0.00050038, -0.00013234, -0.00030995, 0.00020688, -0.00003910, -0.00012257, 0.00008032, -0.00002220, -0.00005081, 0.00001556, -0.00000337, -0.00001166, -0.00001884, 0.00000480, 0.00000630, -0.00002924, 0.00000743, 0.00000919, -0.00002441, 0.00001287, 0.00000743, -0.00002134, 0.00000601, 0.00000442, -0.00000499, 0.00001050, 0.00000463, -0.00000348, -0.00000479, 0.00000108, 0.00000331, -0.00001964, -0.00000754, 0.00002313, 0.00000088, -0.00001004, 0.00000665, 0.00000286, 0.00001097, 0.00001164, -0.00001331, -0.00000619, 0.00000466, -0.00000766, 0.00000599, 0.00001429, -0.00001549, -0.00002551, -0.00000621, 0.00000283, 0.00000081, -0.00000384, -0.00000773, -0.00000451, -0.00000110, -0.00000145, 0.00000014, 0.00000127, -0.00000046, },
		{ 0.00000097, -0.00000092, -0.00000184, -0.00000236, -0.00000237, -0.00000212, -0.00000083, 0.00000059, 0.00000020, -0.00000115, -0.00000428, -0.00000649, -0.00000440, -0.00000175, 0.00000414, 0.00000251, -0.00002111, -0.00002463, 0.00000413, 0.00000455, -0.00000183, 0.00001644, -0.00000292, -0.00002444, 0.00000573, 0.00000925, -0.00000215, 0.00001574, 0.00000018, -0.00001172, 0.00001764, 0.00001040, -0.00001576, -0.00001396, -0.00000007, 0.00000868, -0.00001117, -0.00000602, 0.00002447, -0.00000635, -0.00001090, 0.00002233, -0.00001484, -0.00001380, 0.00002316, -0.00001077, -0.00001391, 0.00000781, -0.00000034, -0.00001000, -0.00002240, 0.00002575, -0.00000149, -0.00007693, 0.00006669, 0.00001243, -0.00015622, 0.00012488, 0.00001998, -0.00026359, 0.00024852, 0.00001581, -0.00052750, 0.00044119, 0.00000336, -0.00161111, 0.00045130, 0.00267866, 0.00030243, -0.00143511, -0.00001747, 0.00028997, -0.00035930, -0.00000667, 0.00011822, -0.00012342, -0.00000572, 0.00003314, -0.00003866, -0.00001615, -0.00000347, 0.00000676, -0.00001901, -0.00001974, 0.00002183, -0.00001639, -0.00001983, 0.00001997, -0.00001245, -0.00000810, 0.00001693, -0.00001171, -0.00000789, 0.00000631, -0.00000153, 0.00000609, 0.00000534, 0.00000076, -0.00000210, -0.00000637, 0.00000591, -0.00000592, -0.00001999, 0.00001131, 0.00001505, -0.00000832, -0.00000123, 0.00000238, 0.00000702, 0.00001720, -0.00000353, -0.00001076, 0.00000406, -0.00000680, -0.00000462, 0.00001671, -0.00000001, -0.00002451, -0.00001499, -0.00000089, 0.00000164, -0.00000094, -0.00000716, -0.00000697, -0.00000139, -0.00000116, -0.00000141, 0.00000121, },
		{ -0.00000026, 0.00000180, 0.00000032, -0.00000112, -0.00000208, -0.00000254, -0.00000245, -0.00000194, -0.00000036, 0.00000054, -0.00000003, -0.00000271, -0.00000627, -0.00000523, -0.00000291, 0.00000086, 0.00000651, -0.00000918, -0.00002823, -0.00000793, 0.00000595, -0.00000528, 0.00001390, 0.00001484, -0.00002399, -0.00001083, 0.00001338, -0.00000322, 0.00000822, 0.00001372, -0.00001087, 0.00000368, 0.00001695, -0.00000314, -0.00001529, -0.00001274, 0.00000876, 0.00000322, -0.00001950, 0.00001557, 0.00001560, -0.00002190, 0.00001171, 0.00000925, -0.00002437, 0.00000986, 0.00000862, -0.00001534, -0.00000557, 0.00000227, 0.00001016, -0.00003287, -0.00000769, 0.00005554, -0.00007705, -0.00002601, 0.00012003, -0.00013748, -0.00004876, 0.00019805, -0.00022905, -0.00005002, 0.00033378, -0.00042931, -0.00012849, 0.00059413, -0.00103194, -0.00101210, 0.00202903, 0.00180250, -0.00098259, -0.00080829, 0.00034465, -0.00009801, -0.00022444, 0.00010920, -0.00002065, -0.00006896, 0.00001828, -0.00000317, -0.00001774, -0.00002573, 0.00000694, 0.00000750, -0.00003842, 0.00000627, 0.00001309, -0.00003140, 0.00000426, 0.00000790, -0.00001578, 0.00000861, 0.00000190, -0.00001127, 0.00000008, -0.00000015, 0.00000641, 0.00000552, -0.00000052, 0.00000377, -0.00000673, -0.00000226, 0.00000554, -0.00001610, -0.00000606, 0.00001719, 0.00000191, -0.00000492, -0.00000105, 0.00000173, 0.00001772, 0.00000782, -0.00001097, 0.00000135, -0.00000194, -0.00001260, 0.00000918, 0.00001216, -0.00001534, -0.00002003, -0.00000683, -0.00000044, 0.00000113, -0.00000429, -0.00000897, -0.00000341, -0.00000007, -0.00000212, },
		{ -0.00000115, -0.00000201, 0.00000128, 0.00000149, -0.00000012, -0.00000133, -0.00000247, -0.00000269, -0.00000251, -0.00000151, -0.00000009, 0.00000050, -0.00000079, -0.00000511, -0.00000628, -0.00000340, -0.00000167, 0.00000516, 0.00000129, -0.00002290, -0.00001899, 0.00000350, -0.00000600, 0.00000342, 0.00002713, -0.00000823, -0.00002593, 0.00000749, 0.00000271, -0.00000166, 0.00001722, 0.00000014, -0.00000491, 0.00001232, 0.00000554, -0.00000573, -0.00001851, -0.00000609, 0.00001624, -0.00001402, -0.00000733, 0.00002809, -0.00000999, -0.00001050, 0.00002011, -0.00001226, -0.00000847, 0.00001155, -0.00000304, -0.00000876, -0.00001444, 0.00002104, -0.00000310, -0.00005833, 0.00005894, 0.00000651, -0.00012335, 0.00010730, 0.00001869, -0.00020428, 0.00016672, 0.00001843, -0.00029972, 0.00028629, 0.00000182, -0.00052929, 0.00045379, -0.00006450, -0.00150668, 0.00054074, 0.00243518, 0.00027421, -0.00119202, -0.00009857, 0.00018161, -0.00023027, -0.00003445, 0.00005031, -0.00005023, -0.00002329, -0.00000094, 0.00000705, -0.00002872, -0.00002337, 0.00002883, -0.00002244, -0.00002824, 0.00002482, -0.00001359, -0.00001942, 0.00001246, -0.00000619, -0.00000270, 0.00000569, -0.00000607, -0.00000348, -0.00000406, 0.00000505, 0.00001041, -0.00000317, 0.00000343, 0.00000132, -0.00000950, 0.00000448, -0.00000339, -0.00001424, 0.00000737, 0.00000961, -0.00000137, -0.00000282, -0.00000462, 0.00001222, 0.00001759, -0.00000556, -0.00000320, 0.00000337, -0.00001329, -0.00000314, 0.00001525, -0.00000282, -0.00001872, -0.00001168, -0.00000475, 0.00000011, -0.00000045, -0.00000838, -0.00000694, 0.00000002, },
		{ -0.00000237, 0.00000056, -0.00000250, -0.00000040, 0.00000191, 0.00000090, -0.00000028, -0.00000183, -0.00000285, -0.00000286, -0.00000227, -0.00000107, -0.00000004, 0.00000050, -0.00000271, -0.00000673, -0.00000447, -0.00000250, 0.00000157, 0.00000626, -0.00001225, -0.00002452, -0.00000286, -0.00000341, -0.00000886, 0.00002553, 0.00001533, -0.00002750, -0.00000781, 0.00000759, -0.00000554, 0.00001119, 0.00001019, -0.00000388, 0.00000664, 0.00000498, 0.00000254, -0.00000887, -0.00002247, 0.00001052, 0.00000648, -0.00002187, 0.00001712, 0.00001316, -0.00001997, 0.00000935, 0.00000504, -0.00001152, 0.00000176, 0.00000064, 0.00000503, -0.00002375, -0.00000392, 0.00004302, -0.00006198, -0.00001374, 0.00009689, -0.00011383, -0.00003042, 0.00016144, -0.00017617, -0.00004777, 0.00022740, -0.00026086, -0.00003734, 0.00033439, -0.00043734, -0.00009071, 0.00050892, -0.00101185, -0.00079782, 0.00187647, 0.00155239, -0.00075697, -0.00068840, 0.00015431, -0.00005285, -0.00014058, 0.00001496, -0.00000225, -0.00002631, -0.00002922, 0.00001211, 0.00000327, -0.00004988, 0.00001186, 0.00001603, -0.00004337, 0.00000324, 0.00001278, -0.00002439, -0.00000175, 0.00000407, -0.00000375, 0.00000211, -0.00000306, -0.00000110, -0.00000606, -0.00000279, 0.00001489, 0.00000114, -0.00000370, 0.00000778, -0.00000644, -0.00000453, 0.00000405, -0.00000974, -0.00000340, 0.00000832, 0.00000379, 0.00000037, -0.00000788, 0.00000151, 0.00002103, 0.00000503, -0.00000633, 0.00000494, -0.00000732, -0.00001266, 0.00000917, 0.00000660, -0.00001248, -0.00001277, -0.00000833, -0.00000431, 0.00000117, -0.00000464, -0.00000968, },
		{ -0.00000898, -0.00000650, 0.00000081, -0.00000124, -0.00000203, 0.00000121, 0.00000157, 0.00000070, -0.00000062, -0.00000250, -0.00000322, -0.00000281, -0.00000172, -0.00000107, 0.00000022, -0.00000013, -0.00000530, -0.00000614, -0.00000285, -0.00000111, 0.00000572, -0.00000183, -0.00002264, -0.00001109, -0.00000015, -0.00001558, 0.00001071, 0.00003254, -0.00001190, -0.00002197, 0.00000421, -0.00000390, 0.00000306, 0.00001361, 0.00000061, 0.00000638, 0.00000318, -0.00000060, 0.00000620, -0.00002153, -0.00001192, 0.00001942, -0.00001325, -0.00000583, 0.00002357, -0.00000862, -0.00000637, 0.00000814, -0.00000285, -0.00000020, -0.00001023, 0.00001208, -0.00000182, -0.00004397, 0.00004768, 0.00000311, -0.00009472, 0.00009244, 0.00000995, -0.00015954, 0.00014144, 0.00001570, -0.00023012, 0.00019346, 0.00000615, -0.00029996, 0.00029223, -0.00002502, -0.00047056, 0.00041194, -0.00016586, -0.00128286, 0.00061241, 0.00207138, 0.00028232, -0.00089427, -0.00020056, 0.00007182, -0.00010521, -0.00006110, -0.00000596, 0.00000568, -0.00003593, -0.00002186, 0.00003371, -0.00003453, -0.00003189, 0.00003441, -0.00002051, -0.00002868, 0.00001748, -0.00000777, -0.00001536, 0.00000133, 0.00000131, 0.00000200, -0.00000509, -0.00000016, -0.00000080, -0.00001115, 0.00001023, 0.00001191, -0.00000791, 0.00000446, 0.00000322, -0.00000926, 0.00000093, -0.00000161, -0.00000613, 0.00000219, 0.00000395, 0.00000520, -0.00000420, -0.00000885, 0.00001476, 0.00001599, -0.00000383, 0.00000233, 0.00000014, -0.00001477, -0.00000090, 0.00000929, -0.00000545, -0.00001051, -0.00000823, -0.00000913, -0.00000192, -0.00000028, },
		{ 0.00000042, -0.00000480, -0.00000942, -0.00000166, 0.00000048, -0.00000238, -0.00000020, 0.00000159, 0.00000126, 0.00000065, -0.00000135, -0.00000321, -0.00000349, -0.00000219, -0.00000157, -0.00000122, 0.00000080, -0.00000214, -0.00000674, -0.00000417, -0.00000211, 0.00000264, 0.00000425, -0.00001518, -0.00001758, 0.00000036, -0.00001394, -0.00000823, 0.00003333, 0.00001211, -0.00002392, -0.00000552, -0.00000277, -0.00000314, 0.00001288, 0.00000318, 0.00000715, 0.00000904, -0.00000909, 0.00000845, -0.00000199, -0.00002862, 0.00000857, 0.00000742, -0.00001667, 0.00001328, 0.00000731, -0.00000858, 0.00000040, -0.00000171, 0.00001071, -0.00001381, -0.00000852, 0.00003265, -0.00004730, -0.00000953, 0.00007681, -0.00009149, -0.00001630, 0.00012925, -0.00014389, -0.00002733, 0.00018227, -0.00020035, -0.00003665, 0.00022604, -0.00026679, -0.00001660, 0.00028934, -0.00040232, -0.00004314, 0.00034963, -0.00092801, -0.00051872, 0.00162637, 0.00126988, -0.00047642, -0.00056555, -0.00003572, -0.00001220, -0.00006935, -0.00005854, 0.00000923, 0.00000145, -0.00005530, 0.00001950, 0.00001250, -0.00005560, 0.00001016, 0.00001582, -0.00003641, -0.00000318, 0.00000804, -0.00001281, -0.00000842, 0.00000009, 0.00000709, -0.00000423, -0.00000594, 0.00000583, -0.00001030, -0.00000293, 0.00001851, -0.00000159, -0.00000467, 0.00000832, -0.00000443, -0.00000569, 0.00000062, -0.00000249, -0.00000075, -0.00000027, 0.00000532, 0.00000400, -0.00001149, 0.00000164, 0.00001999, 0.00000454, -0.00000031, 0.00000441, -0.00001059, -0.00000891, 0.00000634, -0.00000101, -0.00000801, -0.00000466, -0.00000982, -0.00000845, },
		{ -0.00001307, -0.00000470, -0.00000078, -0.00000853, -0.00000559, 0.00000060, -0.00000138, -0.00000137, 0.00000100, 0.00000129, 0.00000141, 0.00000022, -0.00000231, -0.00000398, -0.00000308, -0.00000147, -0.00000225, -0.00000057, 0.00000064, -0.00000482, -0.00000595, -0.00000281, -0.00000006, 0.00000543, -0.00000630, -0.00001952, -0.00000296, -0.00000694, -0.00002107, 0.00001850, 0.00003058, -0.00001190, -0.00001312, -0.00000445, -0.00000918, 0.00001030, 0.00000602, 0.00000252, 0.00001870, -0.00000729, -0.00000620, 0.00001581, -0.00002101, -0.00001586, 0.00001539, -0.00000733, -0.00000154, 0.00000994, -0.00000075, 0.00000008, -0.00001250, 0.00001364, 0.00000849, -0.00003866, 0.00003303, 0.00000434, -0.00007483, 0.00007513, 0.00000643, -0.00012349, 0.00011951, 0.00000760, -0.00017717, 0.00016133, 0.00000422, -0.00022789, 0.00019806, -0.00001294, -0.00026445, 0.00026280, -0.00005555, -0.00036426, 0.00031720, -0.00027394, -0.00097161, 0.00064863, 0.00164058, 0.00031794, -0.00058422, -0.00029598, -0.00002796, -0.00000676, -0.00007757, -0.00004534, 0.00003634, -0.00003966, -0.00003007, 0.00004036, -0.00003219, -0.00003135, 0.00002647, -0.00001428, -0.00002393, 0.00000466, -0.00000116, -0.00001015, -0.00000879, 0.00000759, 0.00000453, -0.00001208, 0.00000393, -0.00000020, -0.00001320, 0.00001271, 0.00001068, -0.00000858, 0.00000394, 0.00000369, -0.00000683, -0.00000309, -0.00000020, 0.00000146, -0.00000240, -0.00000047, 0.00000893, -0.00000460, -0.00000946, 0.00001362, 0.00001314, 0.00000121, 0.00000532, -0.00000435, -0.00001222, 0.00000133, 0.00000080, -0.00000784, -0.00000160, -0.00000447, },
		{ 0.00000326, -0.00001035, -0.00001243, -0.00000167, -0.00000466, -0.00000786, -0.00000157, -0.00000038, -0.00000163, 0.00000025, 0.00000095, 0.00000141, 0.00000146, -0.00000063, -0.00000351, -0.00000431, -0.00000182, -0.00000189, -0.00000270, 0.00000080, -0.00000125, -0.00000615, -0.00000427, -0.00000174, 0.00000377, 0.00000044, -0.00001675, -0.00000839, 0.00000005, -0.00002306, -0.00000281, 0.00003467, 0.00000632, -0.00001309, -0.00000509, -0.00001553, 0.00000271, 0.00001236, -0.00000419, 0.00001929, 0.00000970, -0.00002069, 0.00001194, 0.00000351, -0.00002781, 0.00000156, 0.00000554, -0.00000312, 0.00000305, -0.00000055, 0.00001097, -0.00001515, -0.00000857, 0.00003853, -0.00003524, -0.00001602, 0.00006181, -0.00007225, -0.00001241, 0.00010417, -0.00011532, -0.00001428, 0.00014532, -0.00016028, -0.00001771, 0.00017772, -0.00020381, -0.00001919, 0.00019443, -0.00024391, 0.00000487, 0.00020912, -0.00032814, -0.00000211, 0.00014670, -0.00078342, -0.00022187, 0.00131297, 0.00098989, -0.00018765, -0.00044477, -0.00019454, 0.00001014, -0.00001736, -0.00010096, 0.00000966, 0.00001426, -0.00005940, 0.00001778, 0.00001270, -0.00004633, 0.00000298, 0.00001009, -0.00002276, -0.00001063, 0.00000167, -0.00000155, -0.00001336, -0.00000252, 0.00001346, -0.00000837, -0.00000646, 0.00000800, -0.00001137, -0.00000100, 0.00001712, -0.00000242, -0.00000381, 0.00000638, -0.00000180, -0.00000591, -0.00000265, 0.00000358, 0.00000113, -0.00000606, 0.00000545, 0.00000519, -0.00001094, 0.00000163, 0.00001515, 0.00000619, 0.00000594, 0.00000126, -0.00001170, -0.00000340, 0.00000196, -0.00000899, -0.00000341, },
		{ -0.00000998, 0.00000615, -0.00000033, -0.00001559, -0.00000841, -0.00000227, -0.00000670, -0.00000426, -0.00000080, -0.00000142, -0.00000012, 0.00000062, 0.00000085, 0.00000188, 0.00000104, -0.00000185, -0.00000480, -0.00000332, -0.00000105, -0.00000338, -0.00000157, 0.00000113, -0.00000375, -0.00000552, -0.00000321, 0.00000129, 0.00000388, -0.00001113, -0.00001332, 0.00000321, -0.00001583, -0.00002045, 0.00002459, 0.00002143, -0.00000721, -0.00000158, -0.00001654, -0.00001223, 0.00001689, -0.00000268, 0.00000633, 0.00002767, -0.00001559, -0.00000978, 0.00001908, -0.00001522, -0.00001688, 0.00000306, 0.00000421, 0.00000443, -0.00001123, 0.00001338, 0.00000654, -0.00003801, 0.00003531, 0.00001474, -0.00006792, 0.00005620, 0.00000750, -0.00010018, 0.00009782, 0.00000579, -0.00013806, 0.00013400, 0.00000038, -0.00017231, 0.00016167, -0.00001206, -0.00019849, 0.00017807, -0.00003244, -0.00020194, 0.00020195, -0.00007910, -0.00023409, 0.00018160, -0.00035999, -0.00061847, 0.00063996, 0.00120065, 0.00036485, -0.00029912, -0.00035980, -0.00011046, 0.00005084, -0.00007868, -0.00006677, 0.00004103, -0.00003392, -0.00002860, 0.00003114, -0.00002311, -0.00002527, 0.00001082, -0.00000630, -0.00001703, -0.00000832, 0.00000382, -0.00000563, -0.00001459, 0.00001077, 0.00000461, -0.00001394, 0.00000508, -0.00000109, -0.00001040, 0.00001223, 0.00000750, -0.00000621, 0.00000259, 0.00000309, -0.00000382, -0.00000595, 0.00000099, 0.00000641, -0.00000498, -0.00000303, 0.00000883, -0.00000365, -0.00000688, 0.00000936, 0.00000970, 0.00000778, 0.00000644, -0.00000850, -0.00000777, 0.00000298, -0.00000776, },
		{ -0.00000240, -0.00001561, -0.00000006, 0.00000982, -0.00000916, -0.00001572, -0.00000543, -0.00000420, -0.00000494, -0.00000226, -0.00000189, -0.00000030, 0.00000079, 0.00000020, 0.00000144, 0.00000200, 0.00000022, -0.00000363, -0.00000507, -0.00000148, -0.00000216, -0.00000396, 0.00000039, -0.00000043, -0.00000483, -0.00000459, -0.00000137, 0.00000442, -0.00000472, -0.00001589, 0.00000132, -0.00000435, -0.00002826, 0.00000617, 0.00002859, 0.00000009, 0.00000263, -0.00000675, -0.00002720, 0.00000895, 0.00000939, -0.00000816, 0.00002859, 0.00000808, -0.00002554, 0.00001044, 0.00000645, -0.00001830, -0.00001082, 0.00000135, 0.00001714, -0.00001286, -0.00000820, 0.00003485, -0.00003528, -0.00001297, 0.00006614, -0.00005882, -0.00001952, 0.00008594, -0.00009350, -0.00001155, 0.00011895, -0.00012794, -0.00000844, 0.00014166, -0.00015862, -0.00000488, 0.00014914, -0.00018474, -0.00000126, 0.00014036, -0.00019574, 0.00001920, 0.00011212, -0.00022737, 0.00001757, -0.00006094, -0.00059196, 0.00004678, 0.00097853, 0.00073944, 0.00006636, -0.00032755, -0.00030117, 0.00000625, 0.00001314, -0.00011069, -0.00000090, 0.00001490, -0.00004657, 0.00000890, 0.00000774, -0.00002859, -0.00000673, 0.00000217, -0.00000828, -0.00001648, -0.00000422, 0.00000577, -0.00001489, -0.00000329, 0.00001417, -0.00000935, -0.00000527, 0.00000586, -0.00000915, 0.00000195, 0.00001209, -0.00000192, -0.00000184, 0.00000347, 0.00000027, -0.00000532, -0.00000427, 0.00000702, 0.00000202, -0.00000803, 0.00000381, 0.00000422, -0.00000689, 0.00000134, 0.00000822, 0.00000869, 0.00001137, -0.00000219, -0.00001128, 0.00000162, },
		{ -0.00000377, 0.00000359, -0.00001411, -0.00001168, 0.00001113, 0.00000393, -0.00001584, -0.00001289, -0.00000489, -0.00000352, -0.00000268, -0.00000322, -0.00000129, 0.00000140, 0.00000016, 0.00000045, 0.00000209, 0.00000169, -0.00000125, -0.00000535, -0.00000352, -0.00000092, -0.00000414, -0.00000242, 0.00000097, -0.00000225, -0.00000466, -0.00000382, 0.00000254, 0.00000089, -0.00001496, -0.00000474, 0.00000568, -0.00002458, -0.00001358, 0.00002675, 0.00000765, 0.00000162, 0.00000947, -0.00002824, -0.00001301, 0.00001929, -0.00000866, 0.00001167, 0.00002948, -0.00001826, -0.00001070, 0.00001362, -0.00000349, -0.00001496, -0.00001625, 0.00002027, 0.00001024, -0.00003597, 0.00003165, 0.00001062, -0.00006279, 0.00005833, 0.00001644, -0.00009153, 0.00007670, 0.00000661, -0.00011422, 0.00011060, 0.00000175, -0.00013524, 0.00013216, -0.00000896, -0.00014651, 0.00014141, -0.00002786, -0.00014932, 0.00013671, -0.00004579, -0.00012730, 0.00012080, -0.00008731, -0.00010842, 0.00002608, -0.00040091, -0.00027326, 0.00058728, 0.00080265, 0.00040449, -0.00006526, -0.00037582, -0.00017200, 0.00006407, -0.00006481, -0.00007241, 0.00002548, -0.00002155, -0.00002161, 0.00001340, -0.00001077, -0.00001721, -0.00000601, 0.00000046, -0.00001038, -0.00001739, 0.00000546, -0.00000262, -0.00001488, 0.00001019, 0.00000287, -0.00001135, 0.00000337, -0.00000238, -0.00000446, 0.00000944, 0.00000346, -0.00000251, 0.00000133, 0.00000187, -0.00000149, -0.00000659, 0.00000188, 0.00000796, -0.00000506, -0.00000391, 0.00000556, -0.00000159, -0.00000249, 0.00000370, 0.00000608, 0.00001362, 0.00000702, -0.00001092, },
		{ -0.00000373, -0.00001011, 0.00000426, -0.00000555, -0.00001938, 0.00000126, 0.00001387, -0.00000595, -0.00001774, -0.00001077, -0.00000351, -0.00000094, -0.00000362, -0.00000373, 0.00000132, 0.00000125, -0.00000044, 0.00000141, 0.00000223, 0.00000095, -0.00000350, -0.00000543, -0.00000154, -0.00000255, -0.00000445, -0.00000045, -0.00000003, -0.00000283, -0.00000480, -0.00000105, 0.00000395, -0.00001023, -0.00001210, 0.00000930, -0.00001179, -0.00002773, 0.00001580, 0.00001631, -0.00000413, 0.00001890, -0.00000962, -0.00003410, 0.00001170, 0.00000464, -0.00000570, 0.00003128, 0.00000456, -0.00002110, 0.00000246, 0.00000663, -0.00000082, -0.00002734, -0.00000383, 0.00004022, -0.00003204, -0.00001339, 0.00005765, -0.00005624, -0.00001316, 0.00008728, -0.00007965, -0.00001806, 0.00009948, -0.00010614, -0.00000742, 0.00011805, -0.00012599, -0.00000100, 0.00011970, -0.00013847, 0.00000670, 0.00010366, -0.00014649, 0.00001130, 0.00007663, -0.00013173, 0.00002088, 0.00001953, -0.00011894, 0.00000742, -0.00023483, -0.00037627, 0.00025194, 0.00066449, 0.00053424, 0.00025482, -0.00021421, -0.00034726, -0.00002353, 0.00002324, -0.00009401, -0.00001847, 0.00000756, -0.00002521, -0.00000306, 0.00000130, -0.00000959, -0.00001547, -0.00000531, 0.00000243, -0.00001883, -0.00000844, 0.00000767, -0.00001262, -0.00000261, 0.00001012, -0.00000751, -0.00000345, 0.00000132, -0.00000461, 0.00000475, 0.00000552, -0.00000096, 0.00000048, 0.00000093, 0.00000106, -0.00000405, -0.00000382, 0.00000749, 0.00000220, -0.00000678, 0.00000087, 0.00000208, -0.00000122, 0.00000102, 0.00000125, 0.00001027, 0.00001531, },
		{ 0.00001685, 0.00000830, -0.00001074, -0.00000190, 0.00000280, -0.00001681, -0.00001245, 0.00001234, 0.00000770, -0.00001344, -0.00001766, -0.00000850, 0.00000054, -0.00000103, -0.00000612, -0.00000104, 0.00000282, -0.00000024, 0.00000021, 0.00000208, 0.00000203, -0.00000072, -0.00000528, -0.00000366, -0.00000151, -0.00000426, -0.00000280, -0.00000006, -0.00000064, -0.00000326, -0.00000426, 0.00000306, -0.00000338, -0.00001643, 0.00000450, 0.00000368, -0.00003035, -0.00000236, 0.00002343, -0.00000650, 0.00001243, 0.00001682, -0.00003504, -0.00001234, 0.00001485, -0.00000852, 0.00001735, 0.00002219, -0.00001312, -0.00000925, -0.00000043, 0.00001323, -0.00001094, -0.00003893, 0.00003818, 0.00001395, -0.00005893, 0.00005034, 0.00001075, -0.00008105, 0.00007784, 0.00001295, -0.00010414, 0.00008971, 0.00000218, -0.00011393, 0.00011021, -0.00000386, -0.00011615, 0.00011389, -0.00001686, -0.00010637, 0.00010430, -0.00003815, -0.00009204, 0.00008228, -0.00004863, -0.00005722, 0.00003510, -0.00007713, -0.00001293, -0.00012356, -0.00038514, 0.00002001, 0.00050083, 0.00048217, 0.00042132, 0.00010439, -0.00034016, -0.00021128, 0.00003942, -0.00004159, -0.00006632, -0.00000062, -0.00000755, -0.00001303, -0.00000497, 0.00000073, -0.00000981, -0.00001878, 0.00000376, -0.00000537, -0.00002067, 0.00000341, -0.00000115, -0.00001061, 0.00000654, 0.00000035, -0.00000637, -0.00000005, -0.00000302, 0.00000224, 0.00000557, -0.00000045, 0.00000089, 0.00000078, 0.00000046, -0.00000029, -0.00000503, 0.00000231, 0.00000678, -0.00000323, -0.00000378, 0.00000092, 0.00000082, 0.00000207, -0.00000126, 0.00000243, },
		{ -0.00000408, 0.00000960, 0.00001748, -0.00000214, -0.00000873, 0.00000354, -0.00000708, -0.00001931, 0.00000054, 0.00001457, -0.00000111, -0.00001800, -0.00001702, -0.00000290, 0.00000344, -0.00000520, -0.00000551, 0.00000273, 0.00000159, -0.00000080, 0.00000136, 0.00000222, 0.00000128, -0.00000306, -0.00000507, -0.00000230, -0.00000314, -0.00000384, -0.00000164, -0.00000032, -0.00000042, -0.00000446, -0.00000088, 0.00000178, -0.00001469, -0.00000571, 0.00001337, -0.00001987, -0.00002101, 0.00002240, 0.00000093, -0.00000446, 0.00003073, -0.00001196, -0.00003358, 0.00000777, -0.00000057, 0.00000422, 0.00002517, 0.00000030, -0.00000751, -0.00001162, 0.00000473, 0.00002146, -0.00004524, -0.00000838, 0.00006095, -0.00005076, -0.00001465, 0.00007393, -0.00007341, -0.00000881, 0.00009710, -0.00009291, -0.00001232, 0.00009952, -0.00010691, -0.00000172, 0.00010223, -0.00010938, 0.00000521, 0.00008526, -0.00010379, 0.00001307, 0.00005202, -0.00009694, 0.00001472, 0.00001731, -0.00006496, 0.00000880, -0.00005050, -0.00002373, -0.00003185, -0.00034571, -0.00016336, 0.00037509, 0.00040345, 0.00037862, 0.00036382, -0.00010610, -0.00033664, -0.00007106, 0.00001621, -0.00006232, -0.00003671, -0.00000346, -0.00000417, -0.00001351, -0.00000402, 0.00000493, -0.00002045, -0.00001084, 0.00000709, -0.00001716, -0.00001071, 0.00000487, -0.00000754, -0.00000128, 0.00000361, -0.00000416, -0.00000199, -0.00000323, 0.00000072, 0.00000653, -0.00000056, -0.00000032, 0.00000246, -0.00000053, 0.00000065, -0.00000232, -0.00000192, 0.00000581, 0.00000208, -0.00000398, -0.00000228, -0.00000009, 0.00000387, 0.00000109, },
		{ 0.00000549, -0.00000399, -0.00000117, 0.00001661, 0.00001079, -0.00000779, -0.00000267, 0.00000029, -0.00001549, -0.00001159, 0.00000966, 0.00001016, -0.00000866, -0.00002181, -0.00001267, 0.00000453, 0.00000055, -0.00000871, -0.00000093, 0.00000378, -0.00000038, 0.00000014, 0.00000199, 0.00000201, -0.00000049, -0.00000436, -0.00000373, -0.00000302, -0.00000363, -0.00000253, -0.00000189, 0.00000069, -0.00000130, -0.00000412, 0.00000216, -0.00000828, -0.00001451, 0.00001207, -0.00000220, -0.00003046, 0.00000971, 0.00001438, -0.00001645, 0.00002209, 0.00001835, -0.00003330, -0.00001155, 0.00000292, -0.00000058, 0.00002171, 0.00000659, -0.00000015, -0.00000646, -0.00002029, 0.00003225, -0.00000613, -0.00005999, 0.00005422, 0.00001419, -0.00007495, 0.00006517, 0.00000673, -0.00008853, 0.00008925, 0.00000535, -0.00010311, 0.00009191, -0.00000401, -0.00009984, 0.00009638, -0.00000865, -0.00008580, 0.00008300, -0.00002034, -0.00006149, 0.00005805, -0.00003993, -0.00003921, 0.00002605, -0.00004039, -0.00000556, -0.00003863, -0.00005206, 0.00003591, -0.00024230, -0.00031511, 0.00023142, 0.00039682, 0.00025505, 0.00040702, 0.00020938, -0.00026117, -0.00022830, -0.00000965, -0.00001773, -0.00005314, -0.00002692, 0.00000290, -0.00000573, -0.00001792, 0.00000813, -0.00000448, -0.00002481, 0.00000286, -0.00000243, -0.00001861, -0.00000130, -0.00000075, -0.00000411, 0.00000147, -0.00000197, -0.00000148, -0.00000358, -0.00000242, 0.00000759, 0.00000179, -0.00000350, 0.00000297, 0.00000098, -0.00000072, 0.00000005, -0.00000223, 0.00000227, 0.00000442, -0.00000082, -0.00000335, -0.00000299, 0.00000264, },
		{ -0.00000143, 0.00000677, 0.00000181, -0.00000697, 0.00000640, 0.00001788, 0.00000249, -0.00000717, -0.00000054, -0.00000738, -0.00001515, -0.00000148, 0.00001166, 0.00000431, -0.00001630, -0.00002294, -0.00000280, 0.00000724, -0.00000627, -0.00000688, 0.00000339, 0.00000186, -0.00000078, 0.00000130, 0.00000210, 0.00000103, -0.00000234, -0.00000389, -0.00000379, -0.00000395, -0.00000218, -0.00000288, -0.00000133, 0.00000192, -0.00000316, -0.00000107, -0.00000234, -0.00001663, 0.00000207, 0.00001212, -0.00002547, -0.00000923, 0.00002275, -0.00001346, -0.00000028, 0.00003409, -0.00001183, -0.00002491, -0.00000410, -0.00000468, 0.00002017, 0.00001054, -0.00000322, 0.00001237, -0.00002944, 0.00000187, 0.00004374, -0.00006076, -0.00001098, 0.00007434, -0.00006497, -0.00001177, 0.00007990, -0.00008270, -0.00000127, 0.00009353, -0.00009531, -0.00000438, 0.00008656, -0.00009527, 0.00000329, 0.00007579, -0.00008153, 0.00000790, 0.00004670, -0.00006194, 0.00001218, 0.00000548, -0.00004641, 0.00000857, -0.00002624, -0.00000872, -0.00001328, -0.00008751, 0.00004059, -0.00009047, -0.00037965, 0.00002111, 0.00041601, 0.00021372, 0.00026799, 0.00039646, -0.00000679, -0.00028224, -0.00012310, -0.00000359, -0.00002838, -0.00004937, -0.00001467, 0.00000986, -0.00001907, -0.00000700, 0.00001210, -0.00002059, -0.00001395, 0.00000595, -0.00001246, -0.00001133, -0.00000045, -0.00000156, -0.00000009, -0.00000273, -0.00000098, -0.00000142, -0.00000587, 0.00000536, 0.00000685, -0.00000484, -0.00000032, 0.00000363, -0.00000089, -0.00000019, -0.00000048, 0.00000021, 0.00000349, 0.00000191, -0.00000154, -0.00000442, },
		{ -0.00000304, -0.00000471, 0.00000328, 0.00000712, -0.00000410, -0.00000449, 0.00001332, 0.00001404, -0.00000256, -0.00000541, -0.00000404, -0.00001091, -0.00000857, 0.00000397, 0.00001049, -0.00000253, -0.00002420, -0.00001663, 0.00000719, 0.00000199, -0.00001010, -0.00000127, 0.00000400, -0.00000009, 0.00000019, 0.00000191, 0.00000164, -0.00000065, -0.00000263, -0.00000371, -0.00000527, -0.00000248, -0.00000163, -0.00000407, 0.00000131, 0.00000108, -0.00000308, -0.00000069, -0.00001285, -0.00000898, 0.00001572, -0.00001019, -0.00002345, 0.00001871, 0.00000092, -0.00001880, 0.00002748, 0.00001363, -0.00002160, -0.00001180, -0.00001480, 0.00001424, 0.00002323, -0.00001401, 0.00001824, -0.00000353, -0.00004132, 0.00004977, -0.00000180, -0.00007446, 0.00006464, 0.00001069, -0.00008051, 0.00007251, -0.00000012, -0.00008375, 0.00008962, -0.00000396, -0.00008922, 0.00008252, -0.00000963, -0.00007580, 0.00007202, -0.00001067, -0.00005146, 0.00004620, -0.00001814, -0.00002170, 0.00001234, -0.00003319, -0.00000074, -0.00002082, -0.00002443, 0.00002035, -0.00008752, -0.00002097, 0.00003462, -0.00031160, -0.00020573, 0.00034954, 0.00029277, 0.00011820, 0.00036212, 0.00025854, -0.00015608, -0.00022386, -0.00006692, -0.00000191, -0.00003728, -0.00004539, 0.00000638, -0.00000120, -0.00002263, 0.00001013, -0.00000152, -0.00002417, -0.00000134, -0.00000131, -0.00001334, -0.00000674, -0.00000086, 0.00000196, -0.00000314, -0.00000353, 0.00000148, -0.00000579, -0.00000053, 0.00001036, -0.00000112, -0.00000527, 0.00000346, 0.00000144, -0.00000125, 0.00000024, 0.00000040, 0.00000196, 0.00000250, 0.00000075, },
		{ 0.00000174, -0.00000072, -0.00000479, -0.00000163, 0.00000705, 0.00000301, -0.00000756, 0.00000160, 0.00001665, 0.00000883, -0.00000506, -0.00000697, -0.00000732, -0.00000760, -0.00000358, 0.00000568, 0.00000876, -0.00001277, -0.00002692, -0.00000368, 0.00000942, -0.00000609, -0.00000757, 0.00000284, 0.00000212, -0.00000035, 0.00000121, 0.00000196, 0.00000025, -0.00000144, -0.00000210, -0.00000571, -0.00000488, 0.00000004, -0.00000380, -0.00000287, 0.00000362, -0.00000069, -0.00000179, -0.00000812, -0.00001507, 0.00000943, 0.00000546, -0.00002583, 0.00000508, 0.00001527, -0.00002303, 0.00000754, 0.00002728, -0.00000796, -0.00000955, -0.00002244, -0.00000660, 0.00003850, -0.00000977, -0.00000480, 0.00003379, -0.00004710, -0.00000088, 0.00006093, -0.00007021, -0.00001107, 0.00007713, -0.00007151, -0.00000590, 0.00007446, -0.00008170, 0.00000699, 0.00007785, -0.00008607, 0.00000312, 0.00006421, -0.00007364, 0.00000574, 0.00004516, -0.00004845, 0.00000618, 0.00001249, -0.00002175, 0.00000460, -0.00002736, -0.00000485, -0.00000425, -0.00004827, 0.00002710, -0.00003806, -0.00009110, 0.00006403, -0.00015253, -0.00034008, 0.00015822, 0.00038980, 0.00009782, 0.00019274, 0.00036904, 0.00007815, -0.00020160, -0.00016548, -0.00003132, -0.00000282, -0.00005253, -0.00002370, 0.00001390, -0.00001868, -0.00000754, 0.00001208, -0.00001663, -0.00001489, 0.00000113, -0.00000670, -0.00001083, -0.00000568, 0.00000331, 0.00000048, -0.00000694, 0.00000073, -0.00000171, -0.00000572, 0.00000838, 0.00000576, -0.00000691, -0.00000075, 0.00000366, -0.00000060, -0.00000045, 0.00000099, 0.00000150, 0.00000204, },
		{ 0.00000178, 0.00000196, 0.00000063, -0.00000297, -0.00000394, 0.00000279, 0.00000704, -0.00000244, -0.00000701, 0.00000833, 0.00001682, 0.00000400, -0.00000868, -0.00000987, -0.00000464, -0.00000333, -0.00000300, 0.00000815, 0.00000337, -0.00002372, -0.00001949, 0.00000733, 0.00000322, -0.00000959, -0.00000232, 0.00000333, 0.00000071, 0.00000032, 0.00000185, 0.00000104, -0.00000111, -0.00000049, -0.00000360, -0.00000768, -0.00000122, -0.00000033, -0.00000593, 0.00000093, 0.00000358, -0.00000110, -0.00000545, -0.00001603, -0.00000070, 0.00001409, -0.00001775, -0.00000975, 0.00002157, -0.00001440, -0.00001251, 0.00002662, 0.00000400, -0.00000147, -0.00001351, -0.00003466, 0.00003252, 0.00002112, -0.00003471, 0.00003784, -0.00000137, -0.00005841, 0.00006181, 0.00000129, -0.00007892, 0.00006674, 0.00000432, -0.00007484, 0.00007048, -0.00000751, -0.00006841, 0.00007858, -0.00001208, -0.00006624, 0.00006357, -0.00001265, -0.00004764, 0.00004243, -0.00000919, -0.00002053, 0.00001135, -0.00001113, 0.00000574, -0.00002391, -0.00002079, 0.00001855, -0.00005031, -0.00000671, 0.00002105, -0.00010558, 0.00000488, -0.00000734, -0.00032347, -0.00007962, 0.00038124, 0.00020306, 0.00005438, 0.00029498, 0.00026568, -0.00004573, -0.00019980, -0.00011755, 0.00000011, -0.00002247, -0.00005209, 0.00000233, 0.00000022, -0.00001969, 0.00000751, -0.00000049, -0.00001904, -0.00000681, -0.00000147, -0.00000752, -0.00001089, -0.00000106, 0.00000561, -0.00000595, -0.00000423, 0.00000183, -0.00000601, 0.00000221, 0.00001038, -0.00000300, -0.00000567, 0.00000270, 0.00000145, -0.00000083, 0.00000086, 0.00000162, },
		{ 0.00000165, 0.00000152, 0.00000227, 0.00000141, -0.00000172, -0.00000348, -0.00000084, 0.00000545, 0.00000399, -0.00000665, -0.00000301, 0.00001402, 0.00001492, -0.00000200, -0.00001412, -0.00000779, 0.00000177, -0.00000447, -0.00000204, 0.00001056, -0.00000796, -0.00002743, -0.00000582, 0.00000919, -0.00000418, -0.00000739, 0.00000089, 0.00000237, 0.00000047, 0.00000107, 0.00000179, -0.00000077, -0.00000053, -0.00000019, -0.00000745, -0.00000575, 0.00000208, -0.00000400, -0.00000454, 0.00000423, 0.00000274, -0.00000309, -0.00001463, -0.00000996, 0.00001453, -0.00000510, -0.00001964, 0.00001869, 0.00000076, -0.00002549, 0.00001694, 0.00001157, -0.00000039, 0.00000928, -0.00004329, -0.00000613, 0.00005446, -0.00003105, -0.00000398, 0.00005159, -0.00006030, -0.00000285, 0.00006925, -0.00007104, -0.00000898, 0.00006870, -0.00006904, 0.00000091, 0.00005955, -0.00007053, 0.00001318, 0.00005418, -0.00006722, 0.00000790, 0.00003804, -0.00004674, 0.00000487, 0.00001697, -0.00001709, 0.00000093, -0.00001122, 0.00000898, -0.00000679, -0.00004259, 0.00002090, -0.00001871, -0.00004972, 0.00003854, -0.00005735, -0.00007000, 0.00004668, -0.00020083, -0.00024524, 0.00023926, 0.00032073, 0.00004478, 0.00014214, 0.00030481, 0.00014278, -0.00011247, -0.00018719, -0.00006186, 0.00000822, -0.00004590, -0.00002923, 0.00000882, -0.00001367, -0.00000629, 0.00000733, -0.00001067, -0.00001421, -0.00000456, -0.00000197, -0.00000971, -0.00000886, 0.00000572, 0.00000032, -0.00000830, 0.00000054, -0.00000247, -0.00000309, 0.00000958, 0.00000360, -0.00000708, -0.00000114, 0.00000249, -0.00000015, 0.00000059, },
		{ 0.00000061, 0.00000203, 0.00000102, 0.00000208, 0.00000278, -0.00000108, -0.00000301, -0.00000151, 0.00000189, 0.00000575, -0.00000062, -0.00000809, 0.00000337, 0.00001777, 0.00001060, -0.00001088, -0.00001631, 0.00000142, 0.00000353, -0.00000906, 0.00000313, 0.00000725, -0.00001972, -0.00002041, 0.00000425, 0.00000412, -0.00000689, -0.00000402, 0.00000173, 0.00000193, 0.00000057, 0.00000170, 0.00000040, -0.00000158, 0.00000144, -0.00000337, -0.00000947, -0.00000042, 0.00000097, -0.00000677, -0.00000043, 0.00000572, 0.00000150, -0.00001139, -0.00001705, 0.00000858, 0.00000735, -0.00002245, 0.00000825, 0.00001715, -0.00002847, 0.00000075, 0.00001942, -0.00000794, 0.00002291, -0.00001603, -0.00005215, 0.00004927, 0.00001566, -0.00005054, 0.00005385, -0.00000028, -0.00006752, 0.00006551, 0.00000298, -0.00007249, 0.00005981, -0.00000322, -0.00006005, 0.00005958, -0.00001294, -0.00004677, 0.00005848, -0.00001666, -0.00003977, 0.00003936, -0.00001217, -0.00002145, 0.00001387, -0.00000493, 0.00000157, -0.00001481, -0.00000196, 0.00001783, -0.00004488, -0.00000729, 0.00001973, -0.00005964, 0.00000639, 0.00000394, -0.00009511, 0.00001570, -0.00007105, -0.00028191, 0.00003922, 0.00034726, 0.00013601, 0.00003928, 0.00021858, 0.00024562, 0.00005092, -0.00015944, -0.00015106, -0.00001267, -0.00001167, -0.00004752, -0.00000696, -0.00000104, -0.00001214, 0.00000256, -0.00000072, -0.00001243, -0.00001129, -0.00000235, -0.00000331, -0.00001245, -0.00000116, 0.00000613, -0.00000660, -0.00000428, 0.00000010, -0.00000439, 0.00000515, 0.00000834, -0.00000409, -0.00000485, 0.00000129, 0.00000055, },
		{ 0.00000033, 0.00000019, 0.00000287, 0.00000134, 0.00000063, 0.00000396, 0.00000080, -0.00000370, -0.00000138, 0.00000056, 0.00000344, 0.00000386, -0.00000524, -0.00000563, 0.00001030, 0.00001887, 0.00000208, -0.00002004, -0.00000921, 0.00001076, -0.00000318, -0.00000963, 0.00000846, -0.00000276, -0.00002354, -0.00000908, 0.00000618, -0.00000059, -0.00000624, -0.00000227, 0.00000242, 0.00000165, 0.00000086, 0.00000152, -0.00000157, 0.00000018, 0.00000121, -0.00000832, -0.00000628, 0.00000311, -0.00000329, -0.00000574, 0.00000352, 0.00000643, -0.00000469, -0.00002051, -0.00000249, 0.00001614, -0.00001683, -0.00000763, 0.00002867, -0.00001779, -0.00002085, 0.00002542, -0.00000916, 0.00000933, 0.00002765, -0.00006129, -0.00000410, 0.00006365, -0.00004815, -0.00000122, 0.00006170, -0.00006567, -0.00000393, 0.00006720, -0.00006273, -0.00000568, 0.00005130, -0.00005850, 0.00000639, 0.00003940, -0.00005179, 0.00001521, 0.00002820, -0.00004298, 0.00000883, 0.00001379, -0.00002024, 0.00000123, -0.00000381, 0.00000661, -0.00000564, -0.00002200, 0.00002550, -0.00001792, -0.00004159, 0.00002874, -0.00002957, -0.00003696, 0.00002832, -0.00006500, -0.00003857, -0.00000159, -0.00022188, -0.00012236, 0.00026648, 0.00023496, 0.00003538, 0.00010722, 0.00022734, 0.00018275, -0.00002945, -0.00018319, -0.00009009, 0.00000373, -0.00003274, -0.00003083, -0.00000176, -0.00000700, -0.00000414, 0.00000118, -0.00000512, -0.00001255, -0.00000873, 0.00000033, -0.00000833, -0.00000920, 0.00000542, -0.00000046, -0.00000734, -0.00000106, -0.00000311, 0.00000084, 0.00000930, 0.00000089, -0.00000606, -0.00000110, },
		{ -0.00000318, -0.00000028, -0.00000115, 0.00000279, 0.00000334, -0.00000077, 0.00000294, 0.00000409, -0.00000338, -0.00000304, 0.00000111, 0.00000161, 0.00000385, -0.00000019, -0.00000720, -0.00000011, 0.00001599, 0.00001590, -0.00001091, -0.00002193, 0.00000548, 0.00001072, -0.00001219, -0.00000394, 0.00000766, -0.00001233, -0.00001880, -0.00000189, 0.00000440, -0.00000228, -0.00000599, -0.00000071, 0.00000330, 0.00000102, 0.00000130, -0.00000008, -0.00000172, 0.00000249, -0.00000304, -0.00000986, -0.00000046, 0.00000175, -0.00000611, -0.00000233, 0.00000682, 0.00000414, -0.00001701, -0.00001584, 0.00001655, -0.00000264, -0.00002278, 0.00002670, 0.00000671, -0.00003862, 0.00001714, 0.00000782, -0.00001929, 0.00004590, -0.00001779, -0.00006326, 0.00005948, 0.00000808, -0.00005797, 0.00006234, 0.00000006, -0.00006682, 0.00005999, 0.00000364, -0.00005703, 0.00004540, -0.00001002, -0.00004045, 0.00004253, -0.00001462, -0.00002431, 0.00003370, -0.00001661, -0.00001549, 0.00001514, -0.00000861, -0.00000185, -0.00000826, 0.00000030, 0.00001262, -0.00002841, 0.00000601, 0.00001616, -0.00004924, 0.00000283, 0.00000852, -0.00005159, 0.00001045, -0.00001990, -0.00006552, 0.00000652, -0.00013412, -0.00020115, 0.00013147, 0.00027527, 0.00009317, 0.00004855, 0.00014636, 0.00021128, 0.00012104, -0.00010793, -0.00016264, -0.00003644, -0.00000692, -0.00003550, -0.00001745, -0.00000415, -0.00000392, -0.00000205, -0.00000151, -0.00000693, -0.00001323, -0.00000344, -0.00000168, -0.00001127, -0.00000113, 0.00000405, -0.00000569, -0.00000401, -0.00000238, -0.00000170, 0.00000758, 0.00000530, -0.00000475, },
		{ -0.00000180, -0.00000456, -0.00000060, -0.00000227, 0.00000048, 0.00000552, 0.00000032, -0.00000024, 0.00000598, -0.00000000, -0.00000535, 0.00000021, 0.00000218, 0.00000239, 0.00000190, -0.00000404, -0.00000515, 0.00000597, 0.00001948, 0.00000660, -0.00002304, -0.00001135, 0.00001710, 0.00000053, -0.00001517, 0.00000216, 0.00000176, -0.00001533, -0.00001278, 0.00000064, 0.00000395, -0.00000405, -0.00000606, 0.00000201, 0.00000310, 0.00000061, 0.00000100, -0.00000170, 0.00000055, 0.00000154, -0.00000781, -0.00000618, 0.00000255, -0.00000183, -0.00000578, 0.00000155, 0.00000920, -0.00000592, -0.00002431, 0.00000518, 0.00001413, -0.00002651, 0.00000744, 0.00003321, -0.00003600, -0.00001135, 0.00003089, -0.00002962, 0.00001853, 0.00004201, -0.00007135, -0.00000197, 0.00006340, -0.00005709, 0.00000237, 0.00006231, -0.00006179, -0.00000436, 0.00005600, -0.00004707, -0.00000240, 0.00002929, -0.00004285, 0.00000889, 0.00001918, -0.00002981, 0.00001238, 0.00000543, -0.00001855, 0.00000622, -0.00000416, 0.00000086, -0.00000356, -0.00001502, 0.00001928, -0.00001088, -0.00002135, 0.00002726, -0.00002510, -0.00002976, 0.00002151, -0.00003336, -0.00001913, 0.00000451, -0.00005909, -0.00001184, -0.00006418, -0.00020961, -0.00000004, 0.00025108, 0.00015429, 0.00004830, 0.00008197, 0.00015525, 0.00019712, 0.00003777, -0.00015518, -0.00011118, -0.00001291, -0.00001845, -0.00002874, -0.00001313, -0.00000184, -0.00000192, -0.00000353, -0.00000172, -0.00001041, -0.00001019, -0.00000000, -0.00000698, -0.00000720, 0.00000319, -0.00000166, -0.00000535, -0.00000302, -0.00000317, 0.00000466, 0.00000813, },
		{ 0.00000898, 0.00000224, -0.00000512, -0.00000123, -0.00000171, -0.00000293, 0.00000498, 0.00000407, -0.00000241, 0.00000407, 0.00000474, -0.00000486, -0.00000303, 0.00000257, 0.00000275, 0.00000177, -0.00000191, -0.00000429, -0.00000138, 0.00001154, 0.00001872, -0.00000847, -0.00002507, 0.00000644, 0.00001660, -0.00001113, -0.00001180, 0.00000402, -0.00000267, -0.00001450, -0.00000993, 0.00000365, 0.00000343, -0.00000754, -0.00000379, 0.00000442, 0.00000178, 0.00000055, -0.00000028, -0.00000125, 0.00000222, -0.00000272, -0.00000836, -0.00000140, 0.00000147, -0.00000377, -0.00000399, 0.00000612, 0.00000620, -0.00002046, -0.00001316, 0.00002116, -0.00001276, -0.00001875, 0.00004146, -0.00000647, -0.00004338, 0.00003192, -0.00000483, -0.00002772, 0.00006212, -0.00001692, -0.00006563, 0.00005979, 0.00000010, -0.00005584, 0.00006147, 0.00000026, -0.00005707, 0.00004674, 0.00000378, -0.00003657, 0.00002690, -0.00001458, -0.00002110, 0.00002341, -0.00001206, -0.00000601, 0.00000941, -0.00001243, 0.00000238, -0.00000433, -0.00000346, 0.00000906, -0.00002080, 0.00000434, 0.00001375, -0.00002932, 0.00001000, 0.00000574, -0.00004014, 0.00000672, -0.00000714, -0.00003316, 0.00000449, -0.00003969, -0.00003019, -0.00002115, -0.00017705, -0.00010156, 0.00018648, 0.00019237, 0.00007050, 0.00006315, 0.00008855, 0.00017229, 0.00015970, -0.00005200, -0.00015291, -0.00006407, -0.00000916, -0.00002144, -0.00002503, -0.00000816, 0.00000167, -0.00000446, -0.00000234, -0.00000390, -0.00001230, -0.00000438, -0.00000233, -0.00000826, -0.00000098, 0.00000071, -0.00000442, -0.00000371, -0.00000416, 0.00000111, },
		{ -0.00000235, 0.00000727, 0.00000650, -0.00000392, -0.00000302, -0.00000006, -0.00000450, 0.00000092, 0.00000718, -0.00000072, -0.00000010, 0.00000687, -0.00000069, -0.00000553, 0.00000027, 0.00000376, 0.00000293, -0.00000190, -0.00000329, -0.00000169, 0.00000175, 0.00001657, 0.00001047, -0.00002197, -0.00001341, 0.00001906, 0.00000606, -0.00001719, -0.00000743, 0.00000453, -0.00000403, -0.00001577, -0.00000654, 0.00000815, -0.00000091, -0.00000961, 0.00000084, 0.00000461, 0.00000044, 0.00000017, -0.00000097, 0.00000058, 0.00000053, -0.00000585, -0.00000512, 0.00000017, -0.00000015, -0.00000400, -0.00000140, 0.00000979, -0.00000581, -0.00002506, 0.00001046, 0.00000967, -0.00003150, 0.00002155, 0.00003225, -0.00004952, -0.00000070, 0.00003233, -0.00004483, 0.00002432, 0.00005023, -0.00007030, -0.00000125, 0.00005351, -0.00005630, 0.00000547, 0.00005419, -0.00004970, -0.00000450, 0.00003901, -0.00002771, -0.00000008, 0.00000782, -0.00002610, 0.00000782, 0.00000338, -0.00000939, 0.00000555, -0.00001028, 0.00000139, 0.00000155, -0.00001382, 0.00001349, -0.00000757, -0.00001740, 0.00002085, -0.00001279, -0.00001357, 0.00001761, -0.00002628, -0.00001429, 0.00000557, -0.00002949, -0.00000474, -0.00002238, -0.00004255, -0.00000079, -0.00012249, -0.00016667, 0.00009856, 0.00020928, 0.00009228, 0.00006561, 0.00006325, 0.00009944, 0.00018892, 0.00008474, -0.00011032, -0.00012134, -0.00003574, -0.00000850, -0.00002384, -0.00002132, -0.00000028, -0.00000031, -0.00000533, -0.00000094, -0.00000823, -0.00000910, -0.00000207, -0.00000583, -0.00000415, 0.00000038, -0.00000308, -0.00000373, -0.00000424, },
		{ -0.00000351, -0.00000426, 0.00000337, 0.00000900, -0.00000024, -0.00000505, 0.00000031, -0.00000288, -0.00000379, 0.00000623, 0.00000368, -0.00000241, 0.00000488, 0.00000392, -0.00000440, -0.00000308, 0.00000227, 0.00000533, -0.00000028, -0.00000535, 0.00000002, -0.00000054, 0.00000548, 0.00001823, -0.00000413, -0.00002432, 0.00000411, 0.00001963, -0.00000587, -0.00001875, -0.00000314, 0.00000726, -0.00000788, -0.00001792, 0.00000149, 0.00000899, -0.00000787, -0.00000710, 0.00000480, 0.00000286, -0.00000047, -0.00000022, -0.00000009, 0.00000066, -0.00000261, -0.00000491, -0.00000287, -0.00000038, 0.00000007, -0.00000450, 0.00000289, 0.00000732, -0.00002007, -0.00001046, 0.00002133, -0.00001964, -0.00001190, 0.00004934, -0.00001822, -0.00004285, 0.00004103, -0.00001487, -0.00003211, 0.00006805, -0.00001209, -0.00005915, 0.00004958, -0.00000666, -0.00004562, 0.00005200, 0.00000087, -0.00004120, 0.00002918, 0.00000379, -0.00001602, 0.00000861, -0.00001624, -0.00000633, 0.00000643, -0.00000621, 0.00000496, -0.00000988, -0.00000587, 0.00001198, -0.00001609, 0.00000136, 0.00001181, -0.00002351, 0.00000564, 0.00000850, -0.00002081, 0.00000882, -0.00000709, -0.00002371, 0.00000408, -0.00002001, -0.00001301, -0.00000876, -0.00004842, -0.00000225, -0.00005979, -0.00018819, -0.00000424, 0.00020343, 0.00011925, 0.00006103, 0.00007186, 0.00004998, 0.00013483, 0.00016942, 0.00000102, -0.00012650, -0.00008759, -0.00001799, -0.00001031, -0.00002716, -0.00001206, 0.00000302, -0.00000424, -0.00000283, -0.00000330, -0.00000936, -0.00000487, -0.00000409, -0.00000490, -0.00000078, -0.00000245, -0.00000390, },
		{ -0.00000455, -0.00000336, -0.00000410, -0.00000059, 0.00000816, 0.00000452, -0.00000499, -0.00000167, -0.00000008, -0.00000550, 0.00000160, 0.00000664, -0.00000092, 0.00000151, 0.00000560, -0.00000086, -0.00000389, -0.00000143, 0.00000510, 0.00000461, -0.00000668, -0.00000281, 0.00000343, -0.00000154, 0.00001092, 0.00001240, -0.00001666, -0.00001426, 0.00001598, 0.00001203, -0.00001514, -0.00001819, 0.00000478, 0.00000809, -0.00001671, -0.00001399, 0.00001054, 0.00000286, -0.00001194, -0.00000127, 0.00000581, 0.00000052, -0.00000089, 0.00000047, 0.00000046, -0.00000176, -0.00000287, -0.00000276, -0.00000378, 0.00000105, 0.00000007, -0.00000500, 0.00000729, -0.00000332, -0.00002267, 0.00001127, 0.00000489, -0.00003060, 0.00003074, 0.00002730, -0.00005537, 0.00000730, 0.00003002, -0.00005155, 0.00002466, 0.00005188, -0.00005794, -0.00000276, 0.00003623, -0.00004697, 0.00000710, 0.00004025, -0.00003252, -0.00000450, 0.00002066, -0.00000907, 0.00000084, -0.00000870, -0.00001211, 0.00000384, -0.00000545, 0.00000562, -0.00000321, -0.00001754, 0.00001388, -0.00000317, -0.00001571, 0.00001720, -0.00000933, -0.00001382, 0.00001418, -0.00001074, -0.00000396, 0.00000242, -0.00002172, -0.00000159, -0.00001153, -0.00002025, 0.00000098, -0.00004284, -0.00002198, -0.00000931, -0.00016094, -0.00010300, 0.00016024, 0.00015772, 0.00005349, 0.00007604, 0.00004972, 0.00006304, 0.00016459, 0.00011135, -0.00005859, -0.00011856, -0.00005821, -0.00000650, -0.00001760, -0.00002421, -0.00000256, 0.00000019, -0.00000439, -0.00000196, -0.00000634, -0.00000659, -0.00000436, -0.00000498, -0.00000149, -0.00000179, },
		{ -0.00000062, -0.00000447, -0.00000461, -0.00000347, -0.00000250, 0.00000491, 0.00000753, -0.00000197, -0.00000421, 0.00000087, -0.00000371, -0.00000297, 0.00000566, 0.00000212, 0.00000005, 0.00000478, 0.00000158, -0.00000180, -0.00000325, 0.00000056, 0.00000846, -0.00000224, -0.00000904, 0.00000454, 0.00000161, -0.00000052, 0.00001426, 0.00000096, -0.00001905, -0.00000064, 0.00001879, 0.00000222, -0.00002305, -0.00001220, 0.00001502, 0.00000015, -0.00002359, -0.00000223, 0.00001297, -0.00000639, -0.00001002, 0.00000396, 0.00000412, -0.00000141, -0.00000013, 0.00000166, -0.00000166, -0.00000304, 0.00000011, -0.00000419, -0.00000408, 0.00000500, -0.00000322, -0.00000324, 0.00000763, -0.00001503, -0.00000942, 0.00001739, -0.00002077, -0.00000486, 0.00004908, -0.00002487, -0.00003848, 0.00004231, -0.00001961, -0.00003259, 0.00006306, -0.00000320, -0.00004581, 0.00003124, -0.00001128, -0.00003073, 0.00003690, 0.00000208, -0.00002328, 0.00001163, 0.00000375, 0.00000028, -0.00000564, -0.00001526, 0.00000145, -0.00000531, 0.00000096, 0.00000797, -0.00002148, 0.00000076, 0.00001394, -0.00001954, 0.00000429, 0.00000891, -0.00001871, 0.00000381, 0.00000120, -0.00000811, 0.00000325, -0.00001691, -0.00000687, -0.00000299, -0.00002574, 0.00000144, -0.00002415, -0.00004465, 0.00000926, -0.00009842, -0.00016589, 0.00007396, 0.00018960, 0.00006715, 0.00005758, 0.00007107, 0.00002990, 0.00010224, 0.00015772, 0.00004462, -0.00009011, -0.00010068, -0.00003150, -0.00000532, -0.00002367, -0.00001487, 0.00000043, -0.00000241, -0.00000272, -0.00000398, -0.00000595, -0.00000477, -0.00000553, -0.00000251, },
		{ -0.00000437, -0.00000022, -0.00000274, -0.00000587, -0.00000433, -0.00000273, 0.00000195, 0.00000750, 0.00000210, -0.00000476, -0.00000058, -0.00000108, -0.00000466, 0.00000232, 0.00000345, 0.00000036, 0.00000431, 0.00000213, -0.00000030, -0.00000094, -0.00000411, 0.00000599, 0.00000615, -0.00001086, -0.00000218, 0.00000822, -0.00000318, 0.00000391, 0.00001145, -0.00000793, -0.00001296, 0.00000859, 0.00001659, -0.00000955, -0.00002727, 0.00000224, 0.00001828, -0.00001456, -0.00002006, 0.00000992, 0.00000667, -0.00001214, -0.00000420, 0.00000615, 0.00000114, -0.00000205, 0.00000208, 0.00000062, -0.00000494, 0.00000033, 0.00000109, -0.00000877, 0.00000114, 0.00000647, -0.00000903, 0.00000166, 0.00000224, -0.00001809, 0.00000687, 0.00000216, -0.00002453, 0.00003238, 0.00002088, -0.00005277, 0.00001000, 0.00002522, -0.00004917, 0.00001916, 0.00004807, -0.00003712, -0.00000627, 0.00001553, -0.00003262, 0.00000692, 0.00002444, -0.00001442, -0.00000420, 0.00000505, 0.00000491, 0.00000052, -0.00001773, -0.00000340, -0.00000159, -0.00000718, 0.00001321, -0.00001137, -0.00001743, 0.00001822, -0.00000622, -0.00001220, 0.00001389, -0.00000838, -0.00000794, 0.00000379, -0.00000573, 0.00000308, -0.00001211, -0.00001365, 0.00000461, -0.00002358, -0.00000972, -0.00000270, -0.00005118, -0.00000526, -0.00003383, -0.00017075, -0.00003250, 0.00018291, 0.00010963, 0.00003518, 0.00007568, 0.00004064, 0.00004334, 0.00013220, 0.00012045, -0.00000977, -0.00010324, -0.00007495, -0.00001300, -0.00001191, -0.00002189, -0.00000713, -0.00000068, -0.00000206, -0.00000322, -0.00000490, -0.00000411, -0.00000552, },
		{ -0.00000415, -0.00000565, -0.00000170, -0.00000072, -0.00000513, -0.00000626, -0.00000359, 0.00000057, 0.00000599, 0.00000475, -0.00000307, -0.00000244, 0.00000013, -0.00000400, -0.00000044, 0.00000287, -0.00000007, 0.00000470, 0.00000355, -0.00000153, 0.00000209, -0.00000337, -0.00000147, 0.00001053, -0.00000400, -0.00001107, 0.00000785, 0.00000399, -0.00000483, 0.00000738, 0.00000547, -0.00000979, -0.00000616, 0.00001396, 0.00001070, -0.00002332, -0.00002079, 0.00001842, 0.00000823, -0.00002496, -0.00000717, 0.00001410, -0.00000297, -0.00001179, 0.00000156, 0.00000533, -0.00000176, -0.00000042, 0.00000373, -0.00000388, -0.00000437, 0.00000602, -0.00000466, -0.00000954, 0.00001048, -0.00000011, -0.00001201, 0.00000776, -0.00000593, -0.00001129, 0.00001110, -0.00001549, -0.00000037, 0.00004168, -0.00002449, -0.00003243, 0.00003607, -0.00001822, -0.00003021, 0.00004943, 0.00000835, -0.00002893, 0.00000939, -0.00001372, -0.00001540, 0.00002032, 0.00000362, -0.00000721, -0.00000196, 0.00000343, 0.00000966, -0.00001352, -0.00001255, 0.00000227, -0.00001063, 0.00000717, 0.00000473, -0.00002494, 0.00000543, 0.00001066, -0.00001628, 0.00000469, 0.00000361, -0.00001023, -0.00000025, -0.00000453, 0.00000341, -0.00000442, -0.00002076, 0.00000489, -0.00001107, -0.00002381, 0.00000623, -0.00003610, -0.00003220, 0.00000171, -0.00012606, -0.00011758, 0.00012357, 0.00015687, 0.00003805, 0.00005487, 0.00006256, 0.00002352, 0.00007590, 0.00013402, 0.00007453, -0.00005074, -0.00010044, -0.00004653, -0.00000726, -0.00001675, -0.00001577, -0.00000435, -0.00000076, -0.00000197, -0.00000449, -0.00000354, },
		{ -0.00000391, -0.00000276, -0.00000493, -0.00000377, -0.00000059, -0.00000278, -0.00000679, -0.00000600, -0.00000079, 0.00000499, 0.00000587, -0.00000087, -0.00000329, 0.00000019, -0.00000293, -0.00000194, 0.00000254, -0.00000168, 0.00000316, 0.00000737, -0.00000246, 0.00000083, 0.00000219, -0.00000624, 0.00000607, 0.00000607, -0.00001270, -0.00000105, 0.00001081, -0.00000286, -0.00000286, 0.00000751, 0.00000247, -0.00000905, -0.00000112, 0.00001794, -0.00000191, -0.00003187, -0.00000268, 0.00002398, -0.00000899, -0.00002315, 0.00000575, 0.00000962, -0.00000962, -0.00000756, 0.00000507, 0.00000256, -0.00000312, 0.00000314, 0.00000159, -0.00000846, 0.00000284, 0.00000637, -0.00001407, -0.00000007, 0.00001461, -0.00001340, -0.00000606, 0.00001071, -0.00001306, -0.00000173, 0.00000317, -0.00001568, 0.00002614, 0.00001552, -0.00004331, 0.00000686, 0.00001954, -0.00003966, 0.00000932, 0.00004081, -0.00001286, -0.00001048, -0.00000418, -0.00001783, 0.00000523, 0.00001048, 0.00000062, -0.00000329, -0.00000510, 0.00001192, -0.00000056, -0.00001912, -0.00000055, -0.00000671, -0.00000386, 0.00001361, -0.00001676, -0.00001268, 0.00001585, -0.00000674, -0.00000636, 0.00000685, -0.00000540, -0.00000283, -0.00000576, 0.00000019, 0.00000531, -0.00002130, -0.00000526, 0.00000392, -0.00002750, -0.00000205, -0.00001254, -0.00004692, 0.00000130, -0.00006529, -0.00015145, 0.00003115, 0.00017362, 0.00007267, 0.00003058, 0.00006635, 0.00003515, 0.00003474, 0.00009936, 0.00011646, 0.00002904, -0.00007829, -0.00008276, -0.00002546, -0.00000871, -0.00001624, -0.00001149, -0.00000294, -0.00000011, -0.00000328, },
		{ -0.00000077, -0.00000370, -0.00000281, -0.00000330, -0.00000416, -0.00000213, -0.00000139, -0.00000493, -0.00000789, -0.00000385, 0.00000393, 0.00000699, 0.00000107, -0.00000372, 0.00000004, -0.00000168, -0.00000348, 0.00000297, -0.00000172, -0.00000191, 0.00001014, 0.00000149, -0.00000422, 0.00000566, -0.00000306, -0.00000257, 0.00000989, -0.00000529, -0.00000995, 0.00000888, 0.00000557, -0.00000651, -0.00000105, 0.00000835, 0.00000182, -0.00001046, 0.00000596, 0.00001678, -0.00001903, -0.00002642, 0.00001622, 0.00001477, -0.00002096, -0.00001241, 0.00001150, 0.00000199, -0.00001162, -0.00000254, 0.00000619, -0.00000105, -0.00000156, 0.00000588, -0.00000508, -0.00000681, 0.00001204, -0.00000387, -0.00001616, 0.00001597, 0.00000408, -0.00002233, 0.00000864, 0.00000521, -0.00001627, 0.00000502, -0.00000531, -0.00000027, 0.00002997, -0.00001753, -0.00002666, 0.00002471, -0.00001196, -0.00002629, 0.00003145, 0.00001993, -0.00001212, -0.00001055, -0.00001470, -0.00000336, 0.00000622, 0.00000494, 0.00000432, -0.00000919, 0.00000257, 0.00001166, -0.00001478, -0.00000925, -0.00000180, -0.00001036, 0.00001056, -0.00000165, -0.00002188, 0.00000702, 0.00000518, -0.00000927, 0.00000296, -0.00000128, -0.00000197, -0.00000480, -0.00000685, 0.00000999, -0.00001133, -0.00001861, 0.00000905, -0.00001681, -0.00001688, 0.00000160, -0.00004075, -0.00001798, -0.00002099, -0.00013510, -0.00005635, 0.00014338, 0.00011765, 0.00002714, 0.00005041, 0.00005060, 0.00002447, 0.00005590, 0.00010682, 0.00008959, -0.00001430, -0.00008782, -0.00005925, -0.00001473, -0.00000985, -0.00001437, -0.00000904, -0.00000094, },
		{ -0.00000615, 0.00000011, -0.00000152, -0.00000309, -0.00000292, -0.00000295, -0.00000281, -0.00000212, -0.00000284, -0.00000708, -0.00000737, 0.00000081, 0.00000803, 0.00000396, -0.00000414, -0.00000103, 0.00000052, -0.00000488, 0.00000191, 0.00000169, -0.00000710, 0.00000678, 0.00000949, -0.00000637, 0.00000197, 0.00000431, -0.00000625, 0.00000492, 0.00000340, -0.00001099, 0.00000069, 0.00001014, -0.00000092, -0.00000810, 0.00000144, 0.00001231, -0.00000280, -0.00001051, 0.00001490, 0.00000573, -0.00003009, -0.00000884, 0.00002356, -0.00000071, -0.00002204, -0.00000155, 0.00001070, -0.00000429, -0.00001077, 0.00000223, 0.00000497, -0.00000403, 0.00000306, 0.00000380, -0.00001178, 0.00000401, 0.00001263, -0.00001894, -0.00000219, 0.00002323, -0.00001805, -0.00001384, 0.00002090, -0.00000951, -0.00001188, 0.00000822, -0.00000727, 0.00001405, 0.00001284, -0.00003016, -0.00000053, 0.00001429, -0.00002661, -0.00000191, 0.00003219, 0.00000920, -0.00001354, -0.00001920, -0.00000670, 0.00000278, 0.00000087, 0.00001006, -0.00000149, -0.00000902, 0.00001182, -0.00000182, -0.00001477, -0.00000226, -0.00001011, 0.00000131, 0.00000882, -0.00001826, -0.00000649, 0.00000950, -0.00000492, -0.00000087, -0.00000046, -0.00000178, -0.00000015, -0.00001135, 0.00000472, 0.00000297, -0.00002331, 0.00000059, -0.00000132, -0.00002377, 0.00000015, -0.00002331, -0.00003419, -0.00000485, -0.00009273, -0.00010995, 0.00007798, 0.00014455, 0.00004882, 0.00003221, 0.00005284, 0.00003202, 0.00003144, 0.00007144, 0.00010423, 0.00005455, -0.00004851, -0.00008087, -0.00003941, -0.00000940, -0.00001001, -0.00001334, },
		{ -0.00001096, -0.00001168, -0.00000358, 0.00000041, -0.00000129, -0.00000346, -0.00000268, -0.00000151, -0.00000282, -0.00000289, -0.00000471, -0.00000842, -0.00000412, 0.00000658, 0.00000806, -0.00000277, -0.00000404, 0.00000253, -0.00000384, -0.00000197, 0.00000575, -0.00000691, -0.00000251, 0.00001435, -0.00000054, -0.00000540, 0.00000724, -0.00000179, -0.00000201, 0.00000560, -0.00000578, -0.00000527, 0.00000734, 0.00000573, -0.00000563, -0.00000917, 0.00000907, 0.00001334, -0.00001181, -0.00000360, 0.00001819, -0.00001075, -0.00002716, 0.00000864, 0.00001819, -0.00001179, -0.00001660, 0.00000502, 0.00000754, -0.00000904, -0.00000807, 0.00000653, 0.00000092, -0.00000385, 0.00000797, -0.00000472, -0.00001026, 0.00001741, -0.00000192, -0.00002280, 0.00002095, 0.00000793, -0.00003228, 0.00001104, 0.00001543, -0.00002325, 0.00000164, 0.00000650, -0.00000480, 0.00001753, -0.00000659, -0.00002227, 0.00001177, -0.00000365, -0.00002187, 0.00001399, 0.00002854, 0.00000170, -0.00002404, -0.00001520, 0.00000314, -0.00000286, 0.00000551, 0.00001038, -0.00000975, 0.00000106, 0.00000783, -0.00001112, -0.00000628, -0.00000761, -0.00000678, 0.00001028, -0.00000787, -0.00001506, 0.00000556, 0.00000011, -0.00000168, 0.00000026, -0.00000416, 0.00000332, -0.00000805, -0.00000591, 0.00001039, -0.00001524, -0.00001272, 0.00000628, -0.00001830, -0.00000915, -0.00000878, -0.00003729, -0.00000895, -0.00005031, -0.00012337, 0.00000358, 0.00013888, 0.00008164, 0.00002664, 0.00004413, 0.00003930, 0.00002712, 0.00004153, 0.00008190, 0.00009173, 0.00001508, -0.00006679, -0.00006617, -0.00002494, -0.00000629, },
		{ -0.00001431, -0.00000575, -0.00001163, -0.00000950, -0.00000185, 0.00000158, -0.00000200, -0.00000419, -0.00000066, -0.00000114, -0.00000395, -0.00000395, -0.00000660, -0.00000755, 0.00000150, 0.00001047, 0.00000227, -0.00000709, 0.00000123, 0.00000035, -0.00000580, 0.00000558, -0.00000054, -0.00001074, 0.00000995, 0.00000980, -0.00000783, 0.00000237, 0.00000475, -0.00000327, 0.00000221, -0.00000177, -0.00000566, 0.00000287, 0.00000642, 0.00000167, -0.00001102, -0.00000506, 0.00001941, 0.00000483, -0.00001700, 0.00000741, 0.00001128, -0.00002163, -0.00001451, 0.00001676, 0.00000858, -0.00001621, -0.00001067, 0.00000910, 0.00000331, -0.00001321, -0.00000227, 0.00000825, -0.00000486, 0.00000169, 0.00000734, -0.00001449, 0.00000375, 0.00001899, -0.00002273, -0.00000429, 0.00003051, -0.00002274, -0.00001919, 0.00003066, -0.00000881, -0.00002014, 0.00001595, -0.00000217, -0.00000016, 0.00001298, -0.00001696, -0.00000912, 0.00001002, -0.00001397, -0.00001114, 0.00002377, 0.00002448, -0.00001381, -0.00002744, -0.00000166, 0.00000027, -0.00000366, 0.00001331, 0.00000119, -0.00000776, 0.00000644, -0.00000294, -0.00000774, -0.00000604, -0.00001116, 0.00000520, 0.00000172, -0.00001595, -0.00000146, 0.00000215, -0.00000181, 0.00000271, -0.00000550, 0.00000100, 0.00000003, -0.00001199, 0.00000631, -0.00000196, -0.00001916, 0.00000243, -0.00000755, -0.00001567, -0.00000328, -0.00003016, -0.00001969, -0.00002178, -0.00010716, -0.00005757, 0.00010372, 0.00010854, 0.00003584, 0.00003436, 0.00003999, 0.00002995, 0.00002930, 0.00005076, 0.00008819, 0.00006718, -0.00001938, -0.00007060, -0.00005008, },
		{ -0.00006516, -0.00003470, -0.00000792, -0.00000659, -0.00001139, -0.00000799, 0.00000093, 0.00000220, -0.00000440, -0.00000274, 0.00000138, -0.00000265, -0.00000531, -0.00000519, -0.00000721, -0.00000416, 0.00000783, 0.00000890, -0.00000587, -0.00000405, 0.00000404, -0.00000523, 0.00000044, 0.00000631, -0.00001103, -0.00000143, 0.00001549, -0.00000153, -0.00000459, 0.00000588, 0.00000078, 0.00000008, -0.00000201, -0.00000520, 0.00000207, 0.00000389, 0.00000428, -0.00000322, -0.00001505, 0.00000780, 0.00002225, -0.00000938, -0.00001272, 0.00001390, -0.00000018, -0.00002215, -0.00000252, 0.00001731, 0.00000102, -0.00001824, -0.00000467, 0.00001251, -0.00000411, -0.00001435, 0.00000636, 0.00000381, -0.00000726, 0.00001001, -0.00000280, -0.00001420, 0.00002157, 0.00000056, -0.00002794, 0.00002487, 0.00000980, -0.00003948, 0.00001513, 0.00002169, -0.00002997, 0.00000237, 0.00001602, -0.00001247, 0.00000753, 0.00000464, -0.00001922, 0.00000064, 0.00000343, -0.00001741, 0.00000096, 0.00003180, 0.00001091, -0.00002868, -0.00001593, 0.00000369, -0.00000639, 0.00000514, 0.00001166, -0.00000525, -0.00000088, 0.00000093, -0.00000534, -0.00000414, -0.00001213, -0.00000261, 0.00000661, -0.00001152, -0.00000741, 0.00000205, -0.00000310, 0.00000408, -0.00000205, -0.00000480, 0.00000470, -0.00000886, -0.00000319, 0.00000583, -0.00001528, -0.00000626, -0.00000041, -0.00001563, -0.00000432, -0.00001966, -0.00002810, -0.00000941, -0.00007634, -0.00009429, 0.00005093, 0.00011920, 0.00005419, 0.00002922, 0.00003729, 0.00003109, 0.00002811, 0.00003158, 0.00006191, 0.00008502, 0.00003538, -0.00004272, },
		{ 0.00000439, -0.00005462, -0.00005406, -0.00002236, -0.00000506, -0.00000702, -0.00001187, -0.00000540, 0.00000465, -0.00000023, -0.00000568, 0.00000070, 0.00000132, -0.00000523, -0.00000664, -0.00000550, -0.00000609, 0.00000127, 0.00001160, 0.00000106, -0.00000881, 0.00000253, -0.00000025, -0.00000470, 0.00000751, -0.00000390, -0.00001101, 0.00001128, 0.00000812, -0.00000590, 0.00000150, 0.00000364, 0.00000229, -0.00000198, -0.00000820, 0.00000207, 0.00000517, 0.00000095, 0.00000356, -0.00001148, -0.00001007, 0.00002229, 0.00001236, -0.00001890, -0.00000328, 0.00001314, -0.00000802, -0.00001775, 0.00000483, 0.00001654, -0.00000606, -0.00001968, 0.00000453, 0.00001202, -0.00001443, -0.00000771, 0.00001195, -0.00000588, -0.00000056, 0.00001184, -0.00001634, 0.00000253, 0.00002423, -0.00002489, -0.00000525, 0.00003449, -0.00002689, -0.00002000, 0.00003740, -0.00001117, -0.00002344, 0.00002371, -0.00000191, -0.00001236, 0.00001471, -0.00000659, -0.00001561, 0.00000649, -0.00000472, -0.00001572, 0.00001636, 0.00003066, -0.00001057, -0.00002864, -0.00000280, -0.00000192, -0.00000395, 0.00001165, 0.00000433, -0.00000351, -0.00000121, -0.00000384, -0.00000116, -0.00000925, -0.00001001, 0.00000579, -0.00000496, -0.00001098, 0.00000111, -0.00000389, 0.00000125, 0.00000409, -0.00000724, 0.00000199, -0.00000127, -0.00000876, 0.00000467, -0.00000686, -0.00001183, 0.00000056, -0.00001171, -0.00000750, -0.00001026, -0.00003064, -0.00000991, -0.00004387, -0.00010425, -0.00000631, 0.00010972, 0.00007551, 0.00002959, 0.00003416, 0.00003076, 0.00002804, 0.00002641, 0.00003685, 0.00007185, 0.00006986, },
		{ 0.00007418, 0.00004644, -0.00002057, -0.00005618, -0.00004132, -0.00001440, -0.00000291, -0.00000847, -0.00001239, 0.00000022, 0.00000567, -0.00000433, -0.00000366, 0.00000356, -0.00000101, -0.00000835, -0.00000669, -0.00000438, -0.00000359, 0.00000751, 0.00000887, -0.00000734, -0.00000374, 0.00000377, -0.00000483, 0.00000318, 0.00000375, -0.00001251, 0.00000101, 0.00001275, -0.00000072, -0.00000211, 0.00000189, 0.00000448, 0.00000271, -0.00001120, -0.00000354, 0.00001010, -0.00000002, 0.00000150, -0.00000083, -0.00001723, 0.00000524, 0.00002665, -0.00000354, -0.00001945, 0.00000459, 0.00001011, -0.00001190, -0.00001429, 0.00001170, 0.00001499, -0.00001624, -0.00001603, 0.00001576, 0.00000239, -0.00002030, 0.00000590, 0.00000674, -0.00001130, 0.00001209, 0.00000004, -0.00001767, 0.00002410, 0.00000267, -0.00003002, 0.00002709, 0.00000847, -0.00004167, 0.00002021, 0.00002196, -0.00003375, 0.00000697, 0.00002000, -0.00002058, 0.00000164, 0.00001263, -0.00001658, -0.00000652, 0.00000673, -0.00001287, -0.00000571, 0.00002883, 0.00001525, -0.00002477, -0.00001694, -0.00000045, -0.00000571, 0.00000401, 0.00000996, 0.00000149, -0.00000277, -0.00000609, -0.00000033, -0.00000287, -0.00001341, -0.00000003, 0.00000075, -0.00001172, -0.00000113, -0.00000207, -0.00000425, 0.00000688, -0.00000302, -0.00000392, 0.00000290, -0.00000721, -0.00000059, -0.00000091, -0.00001214, -0.00000184, -0.00000723, -0.00001060, -0.00000413, -0.00002654, -0.00001761, -0.00001952, -0.00009108, -0.00005535, 0.00007993, 0.00009382, 0.00003635, 0.00003024, 0.00003125, 0.00002638, 0.00002669, 0.00002460, 0.00004696, },
		{ 0.00002752, 0.00005750, 0.00006643, 0.00002009, -0.00003622, -0.00005035, -0.00003123, -0.00000818, -0.00000160, -0.00001223, -0.00000923, 0.00000612, 0.00000234, -0.00000638, 0.00000037, 0.00000424, -0.00000548, -0.00001055, -0.00000420, -0.00000313, 0.00000069, 0.00001079, 0.00000028, -0.00000853, 0.00000216, -0.00000097, -0.00000125, 0.00000654, -0.00000752, -0.00000747, 0.00000988, 0.00000505, -0.00000097, -0.00000063, 0.00000184, 0.00000875, -0.00000704, -0.00001394, 0.00000999, 0.00000720, -0.00000435, 0.00000355, -0.00000993, -0.00001200, 0.00002066, 0.00001816, -0.00001592, -0.00001476, 0.00001063, 0.00000769, -0.00001693, -0.00000970, 0.00002093, 0.00000701, -0.00002651, -0.00000246, 0.00002016, -0.00001427, -0.00001279, 0.00001525, -0.00000737, -0.00000278, 0.00001647, -0.00001719, 0.00000121, 0.00002705, -0.00002496, -0.00000420, 0.00003361, -0.00002952, -0.00001533, 0.00003890, -0.00001548, -0.00002024, 0.00002845, -0.00000613, -0.00001942, 0.00001601, -0.00000040, -0.00001766, 0.00000305, -0.00000010, -0.00001461, 0.00001009, 0.00002810, -0.00000438, -0.00002416, -0.00000821, -0.00000387, -0.00000188, 0.00000754, 0.00000724, 0.00000130, -0.00000804, -0.00000466, 0.00000277, -0.00001010, -0.00000745, 0.00000271, -0.00000946, -0.00000501, 0.00000127, -0.00000747, 0.00000313, 0.00000391, -0.00000609, 0.00000113, -0.00000269, -0.00000393, 0.00000079, -0.00000973, -0.00000478, -0.00000332, -0.00001238, -0.00000273, -0.00001748, -0.00002542, -0.00000874, -0.00006395, -0.00008522, 0.00003452, 0.00010133, 0.00005072, 0.00002616, 0.00003130, 0.00002560, 0.00002601, 0.00002269, },
		{ 0.00002392, 0.00001923, 0.00003580, 0.00006310, 0.00004966, -0.00000352, -0.00004135, -0.00004316, -0.00002340, -0.00000168, -0.00000424, -0.00001472, -0.00000193, 0.00000789, -0.00000242, -0.00000515, 0.00000437, 0.00000183, -0.00001126, -0.00000896, -0.00000083, -0.00000198, 0.00000589, 0.00000729, -0.00000662, -0.00000314, 0.00000150, -0.00000213, 0.00000535, -0.00000166, -0.00001053, 0.00000384, 0.00000671, 0.00000190, 0.00000106, -0.00000317, 0.00000851, 0.00000444, -0.00001933, -0.00000118, 0.00001648, -0.00000276, -0.00000201, 0.00000013, -0.00001525, 0.00000246, 0.00002586, 0.00000409, -0.00002231, -0.00000752, 0.00001734, 0.00000183, -0.00002335, 0.00000190, 0.00002720, -0.00001036, -0.00002689, 0.00001652, 0.00000900, -0.00002459, 0.00000551, 0.00000851, -0.00001507, 0.00001427, 0.00000285, -0.00001957, 0.00002473, 0.00000358, -0.00002794, 0.00002694, 0.00000375, -0.00003745, 0.00002483, 0.00001601, -0.00003246, 0.00001354, 0.00001706, -0.00002620, -0.00000023, 0.00001530, -0.00001316, -0.00000921, 0.00000533, -0.00000805, -0.00000634, 0.00002055, 0.00001550, -0.00001488, -0.00001764, -0.00000699, -0.00000324, 0.00000263, 0.00000729, 0.00000758, -0.00000406, -0.00001098, 0.00000197, -0.00000223, -0.00001110, -0.00000002, -0.00000566, -0.00000904, 0.00000276, -0.00000543, -0.00000405, 0.00000688, -0.00000240, -0.00000257, -0.00000032, -0.00000407, 0.00000056, -0.00000682, -0.00000762, -0.00000059, -0.00001131, -0.00000578, -0.00000791, -0.00002730, -0.00001045, -0.00003551, -0.00009071, -0.00001544, 0.00009058, 0.00006992, 0.00002624, 0.00002829, 0.00002659, 0.00002414, },
		{ 0.00002309, 0.00002411, 0.00001901, 0.00002035, 0.00004611, 0.00006023, 0.00002754, -0.00001853, -0.00004041, -0.00003795, -0.00001412, 0.00000192, -0.00001004, -0.00001182, 0.00000493, 0.00000536, -0.00000559, -0.00000156, 0.00000678, -0.00000469, -0.00001451, -0.00000333, 0.00000066, 0.00000061, 0.00000774, -0.00000021, -0.00000597, -0.00000004, -0.00000133, 0.00000383, 0.00000244, -0.00001017, -0.00000112, 0.00000580, 0.00000170, 0.00000531, -0.00000325, 0.00000046, 0.00001369, -0.00001182, -0.00001702, 0.00001565, 0.00000870, -0.00000771, 0.00000106, -0.00000721, -0.00001021, 0.00001548, 0.00002080, -0.00000920, -0.00002406, 0.00000407, 0.00002169, -0.00001124, -0.00002367, 0.00002027, 0.00002077, -0.00002950, -0.00001049, 0.00002587, -0.00001289, -0.00001597, 0.00001714, -0.00000948, -0.00000390, 0.00002012, -0.00001688, 0.00000063, 0.00002653, -0.00002262, -0.00000101, 0.00002733, -0.00002945, -0.00000588, 0.00003410, -0.00001961, -0.00001121, 0.00002784, -0.00001278, -0.00002019, 0.00001501, 0.00000194, -0.00001471, -0.00000084, 0.00000041, -0.00000871, 0.00000483, 0.00001938, 0.00000308, -0.00001622, -0.00001479, -0.00000594, 0.00000050, 0.00000350, 0.00000920, 0.00000474, -0.00001187, -0.00000553, 0.00000326, -0.00000815, -0.00000451, -0.00000292, -0.00001125, 0.00000036, 0.00000002, -0.00000860, 0.00000332, 0.00000307, -0.00000336, -0.00000082, -0.00000331, 0.00000013, -0.00000354, -0.00000989, -0.00000053, -0.00000702, -0.00001012, -0.00000270, -0.00002192, -0.00001772, -0.00001628, -0.00007576, -0.00005576, 0.00006022, 0.00008539, 0.00003488, 0.00002319, 0.00002692, },
		{ 0.00002440, 0.00002302, 0.00002287, 0.00002148, 0.00001455, 0.00002693, 0.00005384, 0.00004742, 0.00000752, -0.00002436, -0.00003989, -0.00003153, -0.00000411, -0.00000047, -0.00001348, -0.00000510, 0.00000798, 0.00000123, -0.00000643, 0.00000362, 0.00000478, -0.00001307, -0.00001131, 0.00000215, 0.00000087, 0.00000354, 0.00000385, -0.00000376, -0.00000238, -0.00000211, 0.00000250, 0.00000597, -0.00000841, -0.00000544, 0.00000587, -0.00000073, 0.00000548, 0.00000378, -0.00000710, 0.00001155, 0.00000407, -0.00002375, 0.00000121, 0.00001935, -0.00000317, -0.00000549, -0.00000107, -0.00000983, 0.00000061, 0.00002051, 0.00001107, -0.00002064, -0.00001929, 0.00001928, 0.00001641, -0.00002697, -0.00001043, 0.00003506, -0.00000144, -0.00003481, 0.00001460, 0.00001392, -0.00002596, 0.00000550, 0.00000800, -0.00001748, 0.00001638, 0.00000472, -0.00001896, 0.00002330, 0.00000302, -0.00002150, 0.00002394, -0.00000333, -0.00002692, 0.00002724, 0.00000565, -0.00002543, 0.00001920, 0.00000816, -0.00002729, 0.00000032, 0.00001257, -0.00000822, -0.00000842, 0.00000018, -0.00000299, -0.00000295, 0.00000939, 0.00001299, -0.00000284, -0.00001708, -0.00001351, -0.00000148, 0.00000154, 0.00000524, 0.00001094, -0.00000438, -0.00001266, 0.00000107, -0.00000198, -0.00000631, -0.00000224, -0.00001109, -0.00000504, 0.00000430, -0.00000720, -0.00000347, 0.00000503, -0.00000067, -0.00000150, -0.00000315, -0.00000092, -0.00000005, -0.00000992, -0.00000373, -0.00000199, -0.00001162, -0.00000332, -0.00001318, -0.00002279, -0.00000960, -0.00005138, -0.00007632, 0.00001786, 0.00008724, 0.00005114, 0.00002087, },
		{ 0.00002580, 0.00002014, 0.00002248, 0.00002114, 0.00002279, 0.00001623, 0.00001379, 0.00003743, 0.00005337, 0.00002887, -0.00000402, -0.00002747, -0.00003990, -0.00002103, 0.00000139, -0.00000562, -0.00001198, 0.00000111, 0.00000795, -0.00000314, -0.00000411, 0.00000787, -0.00000328, -0.00001697, -0.00000345, 0.00000437, 0.00000167, 0.00000274, -0.00000061, -0.00000165, -0.00000344, -0.00000087, 0.00000889, -0.00000387, -0.00001072, 0.00000591, 0.00000014, -0.00000082, 0.00001063, -0.00000509, 0.00000036, 0.00001483, -0.00001496, -0.00001624, 0.00001726, 0.00000944, -0.00000817, -0.00000350, -0.00000436, -0.00000518, 0.00000875, 0.00001944, -0.00000095, -0.00002845, -0.00000521, 0.00003055, -0.00000112, -0.00003398, 0.00001442, 0.00003247, -0.00002777, -0.00001757, 0.00002752, -0.00001073, -0.00001593, 0.00001673, -0.00001209, -0.00000316, 0.00002177, -0.00001526, 0.00000127, 0.00002246, -0.00001779, 0.00000356, 0.00001652, -0.00002576, 0.00000603, 0.00002372, -0.00002127, 0.00000087, 0.00002119, -0.00001896, -0.00001574, 0.00001084, 0.00000192, -0.00000802, -0.00000518, -0.00000144, -0.00000044, 0.00000045, 0.00000819, 0.00000960, -0.00000725, -0.00001950, -0.00000848, 0.00000159, 0.00000129, 0.00000977, 0.00000588, -0.00001197, -0.00000638, 0.00000088, -0.00000432, -0.00000207, -0.00000908, -0.00001085, 0.00000396, -0.00000142, -0.00000800, 0.00000199, 0.00000227, -0.00000047, -0.00000271, -0.00000294, 0.00000196, -0.00000664, -0.00000800, 0.00000016, -0.00000888, -0.00000684, -0.00000640, -0.00002208, -0.00001168, -0.00002909, -0.00007595, -0.00002358, 0.00007096, 0.00006787, },
		{ 0.00007610, 0.00003791, 0.00001772, 0.00002053, 0.00001969, 0.00002139, 0.00002048, 0.00001029, 0.00001935, 0.00004643, 0.00004247, 0.00001362, -0.00000930, -0.00003239, -0.00003558, -0.00001026, 0.00000120, -0.00000849, -0.00000824, 0.00000569, 0.00000559, -0.00000650, 0.00000161, 0.00000633, -0.00001289, -0.00001289, 0.00000340, 0.00000475, 0.00000106, -0.00000110, 0.00000047, -0.00000158, -0.00000565, 0.00000773, 0.00000438, -0.00001420, 0.00000126, 0.00000615, -0.00000714, 0.00000854, 0.00000524, -0.00000789, 0.00001222, 0.00000119, -0.00002274, 0.00000293, 0.00001789, -0.00000108, -0.00000811, -0.00000419, -0.00000272, 0.00000054, 0.00001234, 0.00001492, -0.00001473, -0.00002757, 0.00001497, 0.00002892, -0.00002391, -0.00002280, 0.00003692, 0.00000866, -0.00003766, 0.00001043, 0.00001551, -0.00002358, 0.00000601, 0.00000465, -0.00001770, 0.00001803, 0.00000518, -0.00001546, 0.00001986, 0.00000137, -0.00001168, 0.00001798, -0.00001063, -0.00001186, 0.00002608, -0.00000584, -0.00001387, 0.00002115, -0.00000374, -0.00002343, 0.00000113, 0.00000619, -0.00000196, -0.00000592, -0.00000641, 0.00000185, 0.00000169, -0.00000154, 0.00000902, 0.00000757, -0.00001451, -0.00001806, -0.00000202, 0.00000101, 0.00000452, 0.00001088, -0.00000373, -0.00001129, -0.00000214, -0.00000202, -0.00000099, -0.00000546, -0.00001444, -0.00000139, 0.00000418, -0.00000720, -0.00000323, 0.00000246, 0.00000128, -0.00000085, -0.00000446, 0.00000095, -0.00000174, -0.00000978, -0.00000181, -0.00000442, -0.00000911, -0.00000399, -0.00001708, -0.00001615, -0.00001534, -0.00006132, -0.00005279, 0.00004021, },
		{ 0.00000433, 0.00007033, 0.00005245, 0.00002034, 0.00001765, 0.00001884, 0.00001824, 0.00002220, 0.00001475, 0.00000767, 0.00003038, 0.00004657, 0.00002742, 0.00000589, -0.00001521, -0.00003574, -0.00002679, -0.00000391, -0.00000025, -0.00000922, -0.00000414, 0.00000901, 0.00000059, -0.00000586, 0.00000656, -0.00000187, -0.00001659, -0.00000472, 0.00000712, 0.00000392, -0.00000329, -0.00000104, 0.00000319, -0.00000669, 0.00000060, 0.00001188, -0.00001048, -0.00000865, 0.00001112, -0.00000526, -0.00000187, 0.00001294, -0.00000482, 0.00000169, 0.00001104, -0.00001537, -0.00001159, 0.00001435, 0.00000924, -0.00000584, -0.00000875, -0.00000230, 0.00000162, 0.00000284, 0.00001402, 0.00000633, -0.00002631, -0.00001524, 0.00003210, 0.00001165, -0.00003831, 0.00000472, 0.00003951, -0.00002148, -0.00002162, 0.00002423, -0.00000830, -0.00001209, 0.00001361, -0.00001470, -0.00000043, 0.00002091, -0.00001220, 0.00000296, 0.00001562, -0.00001080, 0.00000800, 0.00000342, -0.00001818, 0.00001718, 0.00001018, -0.00001895, 0.00001202, 0.00000978, -0.00002207, -0.00000877, 0.00000397, 0.00000130, -0.00000010, -0.00000925, -0.00000363, 0.00000718, -0.00000292, -0.00000193, 0.00001325, 0.00000077, -0.00002037, -0.00001146, 0.00000068, 0.00000137, 0.00000890, 0.00000484, -0.00000901, -0.00000691, -0.00000292, -0.00000032, -0.00000072, -0.00001374, -0.00000924, 0.00000536, -0.00000215, -0.00000655, -0.00000022, 0.00000167, 0.00000163, -0.00000354, -0.00000209, 0.00000140, -0.00000758, -0.00000550, -0.00000175, -0.00000844, -0.00000466, -0.00001127, -0.00001856, -0.00001045, -0.00004190, -0.00006491, },
		{ -0.00006186, -0.00002668, 0.00005103, 0.00006271, 0.00002865, 0.00001548, 0.00001816, 0.00001591, 0.00001922, 0.00002087, 0.00000689, 0.00001278, 0.00003951, 0.00003725, 0.00001688, 0.00000066, -0.00002243, -0.00003370, -0.00001880, -0.00000021, -0.00000104, -0.00000995, 0.00000152, 0.00000907, -0.00000448, -0.00000107, 0.00000526, -0.00001029, -0.00001296, 0.00000247, 0.00000889, -0.00000102, -0.00000648, 0.00000468, -0.00000079, -0.00000746, 0.00001163, 0.00000062, -0.00001676, 0.00000741, 0.00000448, -0.00001041, 0.00000957, 0.00000534, -0.00000466, 0.00000963, -0.00000357, -0.00001628, 0.00000357, 0.00001265, 0.00000254, -0.00000918, -0.00000811, 0.00000378, 0.00000265, 0.00000362, 0.00001393, -0.00000577, -0.00003016, 0.00000514, 0.00003679, -0.00001473, -0.00003232, 0.00003240, 0.00001755, -0.00003415, 0.00000496, 0.00001289, -0.00001746, 0.00000692, -0.00000116, -0.00001541, 0.00001870, 0.00000446, -0.00000950, 0.00001478, -0.00000038, -0.00000051, 0.00000970, -0.00001561, 0.00000461, 0.00002100, -0.00001475, -0.00000071, 0.00001779, -0.00001479, -0.00001594, 0.00000041, -0.00000107, 0.00000452, -0.00000339, -0.00001175, 0.00000562, 0.00000515, -0.00000951, 0.00000465, 0.00001380, -0.00000979, -0.00001961, -0.00000509, 0.00000089, 0.00000497, 0.00000804, -0.00000246, -0.00000790, -0.00000591, -0.00000239, 0.00000286, -0.00000825, -0.00001527, 0.00000071, 0.00000319, -0.00000580, -0.00000348, -0.00000003, 0.00000261, -0.00000033, -0.00000408, 0.00000099, -0.00000341, -0.00000746, -0.00000206, -0.00000618, -0.00000593, -0.00000701, -0.00001772, -0.00001117, -0.00002523, },
		{ -0.00001484, -0.00004971, -0.00004625, 0.00002372, 0.00006335, 0.00004026, 0.00001617, 0.00001659, 0.00001611, 0.00001409, 0.00002195, 0.00001438, 0.00000290, 0.00002357, 0.00003968, 0.00002643, 0.00001100, -0.00000653, -0.00002600, -0.00002958, -0.00001288, 0.00000414, -0.00000394, -0.00000886, 0.00000764, 0.00000481, -0.00000556, 0.00000251, -0.00000104, -0.00001287, -0.00000673, 0.00000827, 0.00000668, -0.00000890, -0.00000161, 0.00000701, -0.00000814, 0.00000279, 0.00001112, -0.00001513, -0.00000433, 0.00001262, -0.00000854, -0.00000154, 0.00001132, -0.00000172, 0.00000379, 0.00000290, -0.00001267, -0.00000429, 0.00000800, 0.00000787, -0.00000182, -0.00001277, -0.00000169, 0.00000828, -0.00000026, 0.00000535, 0.00001059, -0.00001774, -0.00002382, 0.00002610, 0.00002423, -0.00003568, -0.00000633, 0.00004002, -0.00001178, -0.00002117, 0.00001625, -0.00000608, -0.00000494, 0.00000814, -0.00001662, 0.00000358, 0.00001773, -0.00000772, 0.00000501, 0.00000762, -0.00000249, 0.00001065, -0.00000883, -0.00000748, 0.00002451, -0.00000304, -0.00001264, 0.00001844, -0.00000351, -0.00002080, -0.00000244, -0.00000391, 0.00000133, 0.00000639, -0.00001192, -0.00000467, 0.00001163, -0.00000501, -0.00000859, 0.00001309, 0.00000656, -0.00001704, -0.00001436, -0.00000201, 0.00000298, 0.00000700, 0.00000244, -0.00000458, -0.00000675, -0.00000649, 0.00000213, -0.00000065, -0.00001554, -0.00000732, 0.00000470, -0.00000195, -0.00000489, -0.00000252, 0.00000099, 0.00000266, -0.00000291, -0.00000142, -0.00000059, -0.00000650, -0.00000374, -0.00000434, -0.00000648, -0.00000482, -0.00001452, -0.00001362, },
		{ -0.00001506, -0.00001057, -0.00003521, -0.00005293, -0.00000384, 0.00005278, 0.00005049, 0.00002137, 0.00001400, 0.00001757, 0.00001162, 0.00001649, 0.00002160, 0.00000512, 0.00000738, 0.00003138, 0.00003342, 0.00001935, 0.00000426, -0.00001218, -0.00002605, -0.00002724, -0.00000479, 0.00000648, -0.00000896, -0.00000347, 0.00001019, -0.00000014, -0.00000371, 0.00000117, -0.00000563, -0.00001110, -0.00000021, 0.00001208, -0.00000237, -0.00001085, 0.00000729, 0.00000011, -0.00000628, 0.00001238, -0.00000446, -0.00001479, 0.00001093, 0.00000145, -0.00000963, 0.00000766, 0.00000461, 0.00000186, 0.00000389, -0.00000875, -0.00000611, 0.00000319, 0.00000520, 0.00000591, -0.00000822, -0.00001131, 0.00000766, 0.00000666, -0.00000359, 0.00000778, 0.00000437, -0.00002639, -0.00000790, 0.00003838, -0.00000176, -0.00003649, 0.00002255, 0.00002301, -0.00002439, -0.00000058, 0.00000634, -0.00000864, 0.00000788, -0.00000811, -0.00001108, 0.00001795, 0.00000334, -0.00000227, 0.00000878, -0.00000108, 0.00000940, 0.00000046, -0.00001618, 0.00001886, 0.00001300, -0.00001834, 0.00001034, 0.00000934, -0.00002164, -0.00000735, -0.00000245, -0.00000661, 0.00000957, -0.00000179, -0.00001378, 0.00000746, 0.00000611, -0.00001301, 0.00000057, 0.00001499, -0.00000365, -0.00001810, -0.00000965, 0.00000071, 0.00000581, 0.00000390, -0.00000111, -0.00000382, -0.00000852, -0.00000314, 0.00000393, -0.00000962, -0.00001368, 0.00000098, 0.00000184, -0.00000368, -0.00000389, -0.00000211, 0.00000283, 0.00000047, -0.00000264, -0.00000042, -0.00000419, -0.00000472, -0.00000360, -0.00000633, -0.00000421, -0.00001055, },
		{ -0.00000715, -0.00001456, -0.00001006, -0.00002306, -0.00004968, -0.00002518, 0.00003386, 0.00005435, 0.00003058, 0.00001243, 0.00001718, 0.00001388, 0.00000951, 0.00002120, 0.00001460, 0.00000070, 0.00001591, 0.00003229, 0.00002782, 0.00001222, -0.00000269, -0.00001305, -0.00002831, -0.00002294, 0.00000559, 0.00000265, -0.00001110, 0.00000361, 0.00000852, -0.00000278, -0.00000374, -0.00000118, -0.00000642, -0.00000850, 0.00000758, 0.00000876, -0.00001251, -0.00000240, 0.00000852, -0.00000647, 0.00000498, 0.00000605, -0.00001601, 0.00000107, 0.00000989, -0.00000855, -0.00000056, 0.00000636, 0.00000387, 0.00000549, -0.00000737, -0.00000748, 0.00000363, 0.00000045, 0.00000513, 0.00000228, -0.00001378, -0.00000266, 0.00001300, 0.00000031, -0.00000498, 0.00001086, -0.00000461, -0.00002898, 0.00001426, 0.00003349, -0.00002665, -0.00001583, 0.00003353, -0.00000055, -0.00001578, 0.00000509, -0.00000443, 0.00000390, 0.00000148, -0.00001714, 0.00000755, 0.00001318, -0.00000216, 0.00000650, 0.00000057, 0.00000578, 0.00001030, -0.00001708, 0.00000457, 0.00002620, -0.00001250, -0.00000396, 0.00001786, -0.00001515, -0.00001558, 0.00000083, -0.00001062, 0.00000225, 0.00000959, -0.00001213, -0.00000405, 0.00001174, -0.00000565, -0.00001103, 0.00000945, 0.00000963, -0.00001065, -0.00001634, -0.00000565, 0.00000467, 0.00000465, -0.00000021, -0.00000036, -0.00000563, -0.00000857, 0.00000204, -0.00000183, -0.00001423, -0.00000550, 0.00000259, -0.00000121, -0.00000326, -0.00000431, -0.00000013, 0.00000290, -0.00000110, -0.00000134, -0.00000263, -0.00000447, -0.00000337, -0.00000583, -0.00000452, },
		{ -0.00000508, -0.00000499, -0.00001263, -0.00001085, -0.00001483, -0.00004128, -0.00003756, 0.00001229, 0.00004933, 0.00004034, 0.00001485, 0.00001366, 0.00001791, 0.00000746, 0.00001357, 0.00002124, 0.00000567, 0.00000343, 0.00002114, 0.00003145, 0.00002229, 0.00000247, -0.00000376, -0.00001466, -0.00003266, -0.00001162, 0.00001143, -0.00000476, -0.00000788, 0.00000826, 0.00000581, -0.00000570, -0.00000520, -0.00000008, -0.00000734, -0.00000343, 0.00001297, -0.00000288, -0.00001288, 0.00000709, 0.00000140, -0.00000186, 0.00000895, -0.00000943, -0.00000850, 0.00001066, -0.00000179, -0.00000545, 0.00000304, 0.00000395, 0.00000951, -0.00000373, -0.00001258, 0.00000478, 0.00000234, -0.00000240, 0.00000715, -0.00000518, -0.00001243, 0.00000778, 0.00001163, -0.00000796, -0.00000220, 0.00001356, -0.00001733, -0.00002077, 0.00003331, 0.00001166, -0.00003388, 0.00000955, 0.00002364, -0.00000996, -0.00000488, -0.00000266, 0.00000100, 0.00000849, -0.00001424, -0.00000585, 0.00001559, 0.00000287, 0.00000460, 0.00000297, 0.00000011, 0.00001566, -0.00000782, -0.00001154, 0.00002788, 0.00000417, -0.00001582, 0.00001599, -0.00000212, -0.00002264, -0.00000042, -0.00000677, -0.00000891, 0.00001179, -0.00000114, -0.00001181, 0.00000690, 0.00000451, -0.00001211, -0.00000277, 0.00001197, 0.00000256, -0.00001421, -0.00001394, -0.00000009, 0.00000608, 0.00000002, -0.00000009, -0.00000021, -0.00000889, -0.00000423, 0.00000190, -0.00000942, -0.00001020, -0.00000008, 0.00000029, -0.00000153, -0.00000396, -0.00000384, 0.00000195, 0.00000171, -0.00000101, -0.00000223, -0.00000388, -0.00000309, -0.00000505, },
		{ -0.00000407, -0.00000537, -0.00000395, -0.00001024, -0.00001164, -0.00000997, -0.00003147, -0.00004200, -0.00000665, 0.00003690, 0.00004584, 0.00002211, 0.00000957, 0.00001877, 0.00001176, 0.00000585, 0.00001901, 0.00001486, 0.00000172, 0.00000719, 0.00002426, 0.00003158, 0.00001125, -0.00000309, -0.00000043, -0.00002292, -0.00003085, 0.00000287, 0.00000887, -0.00000961, -0.00000230, 0.00001078, 0.00000187, -0.00001074, -0.00000236, 0.00000121, -0.00000879, 0.00000520, 0.00000891, -0.00001294, -0.00000341, 0.00000723, -0.00000130, 0.00000619, -0.00000236, -0.00001253, 0.00000591, 0.00000396, -0.00000533, 0.00000013, 0.00000002, 0.00001050, 0.00000569, -0.00001657, -0.00000126, 0.00001008, -0.00000569, 0.00000094, 0.00000496, -0.00001080, -0.00000481, 0.00001467, 0.00000404, -0.00001489, 0.00000744, 0.00000987, -0.00002946, -0.00000030, 0.00003678, -0.00001321, -0.00002129, 0.00002123, 0.00001004, -0.00000636, -0.00000670, -0.00000354, 0.00001215, -0.00000462, -0.00001583, 0.00001002, 0.00000862, 0.00000376, 0.00000673, -0.00000365, 0.00001245, 0.00000680, -0.00001913, 0.00001568, 0.00002231, -0.00001599, 0.00000429, 0.00001038, -0.00002227, -0.00000837, 0.00000025, -0.00001435, 0.00000341, 0.00000910, -0.00000946, -0.00000220, 0.00000793, -0.00000497, -0.00000995, 0.00000377, 0.00001008, -0.00000323, -0.00001650, -0.00000911, 0.00000500, 0.00000233, -0.00000218, 0.00000247, -0.00000360, -0.00000864, -0.00000055, -0.00000403, -0.00001056, -0.00000371, -0.00000020, -0.00000061, -0.00000166, -0.00000528, -0.00000183, 0.00000275, 0.00000110, -0.00000167, -0.00000364, -0.00000288, },
		{ -0.00000305, -0.00000319, -0.00000525, -0.00000351, -0.00000782, -0.00001220, -0.00000767, -0.00002201, -0.00004098, -0.00002053, 0.00002127, 0.00004426, 0.00003144, 0.00000906, 0.00001466, 0.00001786, 0.00000451, 0.00001079, 0.00001879, 0.00000859, 0.00000099, 0.00000923, 0.00003066, 0.00002580, -0.00000193, 0.00000127, -0.00000275, -0.00003268, -0.00001903, 0.00001183, 0.00000202, -0.00001037, 0.00000369, 0.00001178, -0.00000663, -0.00001215, 0.00000473, -0.00000212, -0.00000570, 0.00001062, -0.00000265, -0.00001186, 0.00000381, 0.00000329, 0.00000477, 0.00000169, -0.00001274, 0.00000032, 0.00000681, -0.00000398, 0.00000023, -0.00000289, 0.00000411, 0.00001542, -0.00001109, -0.00001369, 0.00001317, 0.00000185, -0.00000832, 0.00000502, -0.00000049, -0.00001070, 0.00000433, 0.00001616, -0.00000920, -0.00001338, 0.00001983, -0.00000536, -0.00003003, 0.00002259, 0.00002206, -0.00002475, -0.00000376, 0.00001936, 0.00000627, -0.00000689, -0.00001175, 0.00000907, 0.00000842, -0.00001757, -0.00000114, 0.00001192, 0.00000381, 0.00000959, -0.00000142, 0.00000323, 0.00001688, -0.00001312, -0.00000250, 0.00003024, -0.00000305, -0.00000871, 0.00001459, -0.00001345, -0.00001829, 0.00000291, -0.00001096, -0.00000801, 0.00001057, -0.00000082, -0.00000669, 0.00000420, 0.00000128, -0.00000823, -0.00000510, 0.00000660, 0.00000730, -0.00000902, -0.00001630, -0.00000182, 0.00000500, -0.00000253, 0.00000044, 0.00000224, -0.00000695, -0.00000538, -0.00000253, -0.00000823, -0.00000564, -0.00000149, -0.00000149, 0.00000013, -0.00000337, -0.00000530, 0.00000039, 0.00000306, 0.00000023, -0.00000327, },
		{ -0.00000174, -0.00000333, -0.00000272, -0.00000491, -0.00000345, -0.00000532, -0.00001225, -0.00000768, -0.00001370, -0.00003625, -0.00002952, 0.00000629, 0.00003632, 0.00003830, 0.00001400, 0.00000849, 0.00002008, 0.00000969, 0.00000414, 0.00001489, 0.00001482, 0.00000548, -0.00000184, 0.00001597, 0.00003595, 0.00001009, -0.00000577, 0.00000829, -0.00001435, -0.00003378, -0.00000391, 0.00001293, -0.00000444, -0.00000839, 0.00001093, 0.00000684, -0.00001649, -0.00000433, 0.00000800, -0.00000658, 0.00000163, 0.00000636, -0.00000993, -0.00000456, 0.00000369, 0.00000643, 0.00000566, -0.00001183, -0.00000558, 0.00000876, -0.00000294, -0.00000022, -0.00000015, -0.00000518, 0.00001591, 0.00000380, -0.00002161, 0.00000414, 0.00001355, -0.00000843, -0.00000447, 0.00000533, -0.00000412, -0.00000724, 0.00001309, 0.00000919, -0.00002231, 0.00000173, 0.00002206, -0.00002500, -0.00001397, 0.00003287, 0.00000165, -0.00002140, 0.00000581, 0.00001794, 0.00000507, -0.00001609, -0.00000344, 0.00001748, -0.00000840, -0.00001267, 0.00000996, 0.00000537, 0.00000911, 0.00000562, -0.00000398, 0.00001613, 0.00000123, -0.00001449, 0.00002370, 0.00001474, -0.00001339, 0.00000917, -0.00000158, -0.00002357, -0.00000177, -0.00000329, -0.00001438, 0.00000379, 0.00000588, -0.00000443, -0.00000004, 0.00000192, -0.00000348, -0.00000689, -0.00000199, 0.00000843, 0.00000309, -0.00001440, -0.00001141, 0.00000319, 0.00000023, -0.00000293, 0.00000361, -0.00000109, -0.00000692, -0.00000465, -0.00000696, -0.00000593, -0.00000166, -0.00000286, -0.00000069, -0.00000013, -0.00000544, -0.00000384, 0.00000238, 0.00000283, },
		{ 0.00000387, 0.00000109, -0.00000279, -0.00000261, -0.00000456, -0.00000390, -0.00000292, -0.00001101, -0.00000958, -0.00000770, -0.00002870, -0.00003431, -0.00000635, 0.00002501, 0.00003965, 0.00002254, 0.00000484, 0.00001656, 0.00001643, 0.00000364, 0.00000821, 0.00001422, 0.00001304, -0.00000014, -0.00000254, 0.00002956, 0.00002906, -0.00000547, 0.00000227, 0.00000637, -0.00002637, -0.00002500, 0.00000709, 0.00000962, -0.00000996, -0.00000206, 0.00001572, -0.00000637, -0.00001755, 0.00000725, 0.00000379, -0.00000582, 0.00000490, -0.00000197, -0.00000804, -0.00000166, 0.00000578, 0.00001157, -0.00000759, -0.00001328, 0.00000894, 0.00000092, -0.00000437, 0.00000498, -0.00000800, 0.00000525, 0.00001712, -0.00001585, -0.00001155, 0.00001672, 0.00000238, -0.00001146, -0.00000052, 0.00000493, -0.00000627, -0.00000127, 0.00001872, -0.00000791, -0.00002263, 0.00002190, 0.00000655, -0.00003318, 0.00000842, 0.00002686, -0.00001107, -0.00001454, 0.00001144, 0.00002081, -0.00000595, -0.00001837, 0.00001342, 0.00000761, -0.00001676, 0.00000182, 0.00000761, 0.00000629, 0.00001182, -0.00000331, 0.00000748, 0.00001320, -0.00001399, 0.00000876, 0.00002659, -0.00000679, -0.00000030, 0.00000674, -0.00002156, -0.00001096, 0.00000220, -0.00001337, -0.00000521, 0.00000646, -0.00000010, -0.00000026, 0.00000033, -0.00000227, -0.00000350, -0.00000620, 0.00000097, 0.00000942, -0.00000373, -0.00001577, -0.00000435, 0.00000226, -0.00000350, 0.00000068, 0.00000340, -0.00000354, -0.00000620, -0.00000801, -0.00000704, -0.00000088, -0.00000227, -0.00000342, 0.00000095, -0.00000221, -0.00000639, -0.00000137, },
		{ -0.00000559, 0.00000161, 0.00000360, -0.00000077, -0.00000224, -0.00000414, -0.00000491, -0.00000142, -0.00000801, -0.00001190, -0.00000528, -0.00001971, -0.00003493, -0.00001639, 0.00001310, 0.00003516, 0.00003093, 0.00000643, 0.00000962, 0.00001988, 0.00000804, 0.00000458, 0.00000911, 0.00001435, 0.00001043, -0.00000959, 0.00000814, 0.00003831, 0.00001043, -0.00000824, 0.00001074, -0.00000440, -0.00003028, -0.00001315, 0.00001343, 0.00000308, -0.00001246, 0.00000877, 0.00001020, -0.00001910, -0.00000663, 0.00001218, -0.00000156, -0.00000240, 0.00000118, -0.00000430, -0.00000556, -0.00000070, 0.00001461, 0.00000314, -0.00001944, 0.00000267, 0.00000887, -0.00000827, 0.00000467, -0.00000142, -0.00000713, 0.00001789, 0.00000013, -0.00002066, 0.00000712, 0.00001354, -0.00000708, -0.00001043, 0.00000335, 0.00000472, -0.00000880, 0.00000894, 0.00001342, -0.00002562, -0.00000436, 0.00002895, -0.00001645, -0.00002360, 0.00002235, 0.00001457, -0.00001639, -0.00000912, 0.00002173, 0.00001597, -0.00002043, -0.00000403, 0.00001829, -0.00000863, -0.00000812, 0.00000724, 0.00000419, 0.00001295, 0.00000382, -0.00000061, 0.00001610, -0.00000441, -0.00000451, 0.00002740, 0.00000645, -0.00000676, 0.00000873, -0.00001414, -0.00001978, 0.00000201, -0.00000773, -0.00001132, 0.00000259, 0.00000171, 0.00000161, 0.00000157, -0.00000407, -0.00000198, -0.00000352, -0.00000611, 0.00000542, 0.00000682, -0.00001025, -0.00001195, -0.00000067, -0.00000179, -0.00000243, 0.00000351, 0.00000127, -0.00000415, -0.00000881, -0.00001021, -0.00000185, 0.00000077, -0.00000469, -0.00000158, 0.00000107, -0.00000492, },
		{ -0.00000092, -0.00000675, -0.00000297, 0.00000348, 0.00000201, -0.00000083, -0.00000324, -0.00000604, -0.00000167, -0.00000402, -0.00001259, -0.00000637, -0.00001159, -0.00003149, -0.00002355, 0.00000230, 0.00002633, 0.00003576, 0.00001312, 0.00000312, 0.00001829, 0.00001356, 0.00000509, 0.00000573, 0.00000850, 0.00001744, -0.00000035, -0.00001174, 0.00002654, 0.00003083, -0.00000527, -0.00000031, 0.00001035, -0.00001471, -0.00002758, -0.00000174, 0.00001558, -0.00000634, -0.00000697, 0.00001579, -0.00000510, -0.00001960, 0.00000658, 0.00000942, -0.00000335, -0.00000323, -0.00000129, -0.00000255, -0.00000748, 0.00000848, 0.00001618, -0.00001615, -0.00001083, 0.00001462, -0.00000472, -0.00000298, 0.00000688, -0.00001021, 0.00000682, 0.00001345, -0.00001631, -0.00000671, 0.00001470, 0.00000410, -0.00001171, -0.00000820, 0.00000982, -0.00000006, -0.00000690, 0.00001810, -0.00000529, -0.00002728, 0.00001947, 0.00001568, -0.00002939, -0.00000624, 0.00002499, 0.00000379, -0.00002058, 0.00000212, 0.00003050, -0.00000200, -0.00002053, 0.00001285, 0.00000621, -0.00001166, 0.00000250, 0.00000359, 0.00000969, 0.00001137, -0.00000220, 0.00001145, 0.00000631, -0.00001005, 0.00001943, 0.00001931, -0.00000656, 0.00000563, -0.00000490, -0.00002458, -0.00000383, -0.00000149, -0.00001303, -0.00000236, 0.00000091, 0.00000135, 0.00000540, -0.00000337, -0.00000507, 0.00000016, -0.00000605, -0.00000329, 0.00000862, 0.00000065, -0.00001240, -0.00000709, -0.00000194, -0.00000335, 0.00000104, 0.00000353, 0.00000008, -0.00000635, -0.00001302, -0.00000664, 0.00000310, -0.00000189, -0.00000510, 0.00000079, },
		{ -0.00000294, 0.00000148, -0.00000392, -0.00000645, 0.00000007, 0.00000342, 0.00000160, -0.00000129, -0.00000640, -0.00000370, -0.00000079, -0.00001058, -0.00000918, -0.00000641, -0.00002522, -0.00002747, -0.00000631, 0.00001543, 0.00003483, 0.00002276, 0.00000044, 0.00001238, 0.00001755, 0.00000739, 0.00000642, 0.00000278, 0.00001340, 0.00001445, -0.00001426, 0.00000161, 0.00003632, 0.00001310, -0.00000842, 0.00000650, 0.00000467, -0.00002073, -0.00002210, 0.00000973, 0.00001070, -0.00001254, 0.00000461, 0.00001030, -0.00001749, -0.00000860, 0.00001278, 0.00000557, -0.00000603, -0.00000564, 0.00000218, -0.00000594, -0.00000417, 0.00002058, -0.00000084, -0.00002270, 0.00000953, 0.00000650, -0.00001015, 0.00000729, -0.00000306, -0.00000503, 0.00001510, -0.00000422, -0.00001378, 0.00000696, 0.00001059, -0.00000250, -0.00001584, -0.00000024, 0.00001339, -0.00000848, 0.00000343, 0.00001463, -0.00002414, -0.00000896, 0.00002908, -0.00000563, -0.00002732, 0.00000758, 0.00002261, -0.00000796, -0.00002000, 0.00002092, 0.00002398, -0.00001825, -0.00000503, 0.00001429, -0.00000516, -0.00000294, 0.00000272, 0.00000499, 0.00001470, 0.00000242, 0.00000509, 0.00001266, -0.00000791, 0.00000801, 0.00002693, 0.00000026, 0.00000036, 0.00000283, -0.00002347, -0.00001321, 0.00000202, -0.00001088, -0.00000677, -0.00000020, -0.00000167, 0.00000696, 0.00000228, -0.00000818, -0.00000123, -0.00000105, -0.00000766, 0.00000191, 0.00000748, -0.00000497, -0.00001062, -0.00000567, -0.00000402, -0.00000107, 0.00000296, 0.00000290, -0.00000126, -0.00001171, -0.00001321, 0.00000052, 0.00000330, -0.00000516, },
		{ -0.00000037, -0.00000589, -0.00000030, -0.00000010, -0.00000602, -0.00000432, 0.00000168, 0.00000352, 0.00000187, -0.00000499, -0.00000649, 0.00000012, -0.00000655, -0.00001135, -0.00000465, -0.00001804, -0.00002849, -0.00001220, 0.00000535, 0.00002753, 0.00003157, 0.00000389, 0.00000417, 0.00001863, 0.00001007, 0.00000820, 0.00000331, 0.00000317, 0.00001990, -0.00000000, -0.00001626, 0.00002050, 0.00003004, -0.00000005, -0.00000500, 0.00000847, -0.00000043, -0.00002547, -0.00001225, 0.00001752, 0.00000000, -0.00000917, 0.00001114, -0.00000327, -0.00001739, 0.00000291, 0.00001376, 0.00000119, -0.00001191, -0.00000114, 0.00000302, -0.00001138, 0.00001057, 0.00001674, -0.00002121, -0.00000612, 0.00001564, -0.00000796, -0.00000063, 0.00000523, -0.00000836, 0.00000779, 0.00000513, -0.00001205, -0.00000055, 0.00000826, 0.00000648, -0.00000979, -0.00001583, 0.00001293, 0.00000653, -0.00001048, 0.00001383, -0.00000263, -0.00002599, 0.00001338, 0.00002024, -0.00001980, -0.00001837, 0.00001713, 0.00001600, -0.00002086, -0.00000610, 0.00003339, 0.00000442, -0.00001745, 0.00000758, 0.00000457, -0.00000338, 0.00000132, 0.00000068, 0.00001285, 0.00000927, 0.00000162, 0.00001372, -0.00000110, -0.00000218, 0.00002717, 0.00001164, -0.00000358, 0.00000625, -0.00001655, -0.00002244, 0.00000041, -0.00000611, -0.00001012, -0.00000094, -0.00000423, 0.00000326, 0.00000880, -0.00000563, -0.00000660, 0.00000159, -0.00000484, -0.00000540, 0.00000548, 0.00000342, -0.00000715, -0.00000926, -0.00000694, -0.00000291, 0.00000178, 0.00000312, 0.00000279, -0.00000575, -0.00001630, -0.00000731, 0.00000545, },
		{ 0.00000070, 0.00000520, -0.00000400, -0.00000413, 0.00000068, -0.00000261, -0.00000611, -0.00000241, 0.00000277, 0.00000503, -0.00000121, -0.00000826, -0.00000177, -0.00000231, -0.00001145, -0.00000543, -0.00001138, -0.00002739, -0.00001622, -0.00000115, 0.00001595, 0.00003482, 0.00001360, -0.00000271, 0.00001532, 0.00001345, 0.00000809, 0.00000840, -0.00000254, 0.00001189, 0.00001570, -0.00001454, -0.00000303, 0.00002896, 0.00001778, -0.00000526, -0.00000265, 0.00001069, -0.00000725, -0.00002767, 0.00000222, 0.00001597, -0.00000774, -0.00000069, 0.00000645, -0.00001180, -0.00000956, 0.00000995, 0.00001316, -0.00000841, -0.00001252, 0.00000738, -0.00000517, -0.00000535, 0.00002201, -0.00000514, -0.00002036, 0.00001282, 0.00000272, -0.00000810, 0.00000646, -0.00000403, 0.00000033, 0.00000769, -0.00000779, -0.00000291, 0.00000409, 0.00000595, 0.00000331, -0.00001946, -0.00000425, 0.00001954, -0.00000578, -0.00000205, 0.00001177, -0.00001836, -0.00001085, 0.00002290, 0.00000506, -0.00002488, -0.00000793, 0.00002392, 0.00000133, -0.00002437, 0.00001609, 0.00002771, -0.00000979, -0.00000603, 0.00000663, 0.00000100, 0.00000209, -0.00000199, 0.00000692, 0.00001433, 0.00000253, 0.00001106, 0.00000712, -0.00000768, 0.00001990, 0.00002371, -0.00000234, 0.00000429, -0.00000676, -0.00002711, -0.00000680, -0.00000117, -0.00001137, -0.00000263, -0.00000381, -0.00000325, 0.00001016, 0.00000227, -0.00000951, -0.00000160, 0.00000002, -0.00000672, -0.00000130, 0.00000556, 0.00000016, -0.00000781, -0.00001053, -0.00000662, 0.00000054, 0.00000267, 0.00000354, 0.00000093, -0.00001264, -0.00001527, },
		{ -0.00001714, -0.00000866, 0.00000549, 0.00000173, -0.00000515, -0.00000208, -0.00000020, -0.00000412, -0.00000546, -0.00000115, 0.00000574, 0.00000420, -0.00000710, -0.00000545, 0.00000020, -0.00000928, -0.00000752, -0.00000597, -0.00002428, -0.00002016, -0.00000363, 0.00000500, 0.00002940, 0.00002574, -0.00000348, 0.00000727, 0.00001673, 0.00000725, 0.00001169, 0.00000113, -0.00000038, 0.00001892, 0.00000069, -0.00001519, 0.00001146, 0.00002631, 0.00000935, -0.00000969, 0.00000175, 0.00001270, -0.00001843, -0.00002142, 0.00001399, 0.00000775, -0.00000757, 0.00000230, -0.00000154, -0.00001125, -0.00000249, 0.00001592, 0.00000650, -0.00001888, -0.00000198, 0.00000716, -0.00001159, 0.00001112, 0.00001269, -0.00002093, -0.00000058, 0.00001189, -0.00000767, 0.00000149, 0.00000089, -0.00000273, 0.00000710, -0.00000546, -0.00000419, 0.00000528, -0.00000012, 0.00000850, -0.00000682, -0.00002107, 0.00001338, 0.00001205, -0.00001075, 0.00000644, -0.00000091, -0.00001915, 0.00000538, 0.00001975, -0.00000714, -0.00002552, 0.00000536, 0.00002232, -0.00001584, -0.00001120, 0.00002939, 0.00001224, -0.00000978, -0.00000072, 0.00000299, 0.00000596, -0.00000052, -0.00000064, 0.00001456, 0.00000707, 0.00000706, 0.00001349, -0.00000624, 0.00000777, 0.00003090, 0.00000630, -0.00000037, 0.00000078, -0.00002465, -0.00001686, 0.00000056, -0.00000959, -0.00000587, -0.00000148, -0.00000750, 0.00000484, 0.00000950, -0.00000580, -0.00000690, 0.00000076, -0.00000304, -0.00000544, 0.00000127, 0.00000433, -0.00000152, -0.00001025, -0.00001179, -0.00000286, 0.00000287, 0.00000265, 0.00000401, -0.00000463, },
		{ 0.00000196, -0.00001165, -0.00001571, -0.00000116, 0.00000561, -0.00000135, -0.00000448, -0.00000139, -0.00000124, -0.00000474, -0.00000550, 0.00000221, 0.00000848, -0.00000205, -0.00000861, -0.00000023, -0.00000559, -0.00000973, -0.00000274, -0.00001852, -0.00002455, -0.00000497, -0.00000045, 0.00001706, 0.00003338, 0.00000435, -0.00000226, 0.00001670, 0.00000871, 0.00001000, 0.00000934, -0.00000542, 0.00000961, 0.00001335, -0.00001021, -0.00000563, 0.00001704, 0.00002309, 0.00000093, -0.00001333, 0.00001217, 0.00000689, -0.00002744, -0.00000764, 0.00001705, 0.00000136, -0.00000590, -0.00000081, -0.00000347, -0.00000893, 0.00000481, 0.00001912, -0.00000895, -0.00001780, 0.00000952, -0.00000282, -0.00000377, 0.00001783, -0.00000834, -0.00001295, 0.00001165, -0.00000101, -0.00000320, 0.00000244, -0.00000344, 0.00000677, -0.00000261, -0.00000935, 0.00000873, -0.00000011, 0.00000109, 0.00000794, -0.00002025, -0.00000745, 0.00002157, -0.00000098, -0.00000638, 0.00000518, -0.00000984, -0.00000966, 0.00001261, 0.00001320, -0.00001755, -0.00002051, 0.00001821, 0.00000874, -0.00002164, 0.00000874, 0.00002704, 0.00000301, -0.00000645, -0.00000242, 0.00000790, 0.00000636, -0.00000513, 0.00000872, 0.00001243, 0.00000471, 0.00001531, 0.00000147, -0.00000330, 0.00002866, 0.00001983, -0.00000184, 0.00000258, -0.00001696, -0.00002482, -0.00000298, -0.00000579, -0.00000921, -0.00000031, -0.00000698, -0.00000309, 0.00001056, 0.00000209, -0.00000828, -0.00000298, -0.00000019, -0.00000421, -0.00000355, 0.00000219, 0.00000393, -0.00000432, -0.00001411, -0.00000937, 0.00000173, 0.00000290, 0.00000337, },
		},
		{ { -0.00000478, -0.00000561, -0.00000180, 0.00000864, 0.00001115, 0.00000029, -0.00000665, -0.00000145, 0.00000140, -0.00000256, -0.00000091, 0.00000449, 0.00000142, -0.00000507, -0.00000443, -0.00000092, -0.00000069, -0.00000213, -0.00000340, -0.00000246, 0.00000157, 0.00000333, 0.00000073, -0.00000089, -0.00000038, -0.00000006, 0.00000064, 0.00000003, -0.00000263, -0.00000385, -0.00000547, -0.00000673, 0.00000052, 0.00000541, -0.00000636, -0.00000964, 0.00000970, 0.00000889, -0.00001851, -0.00001501, 0.00001167, 0.00000185, -0.00001405, 0.00001299, 0.00002606, -0.00000701, -0.00001480, 0.00001175, 0.00000728, -0.00000717, 0.00001218, 0.00001480, -0.00001584, -0.00001145, 0.00002043, 0.00001189, -0.00001540, -0.00000959, 0.00001012, 0.00001709, 0.00000811, -0.00002091, -0.00003242, 0.00001102, 0.00004628, 0.00001102, -0.00003242, -0.00002091, 0.00000811, 0.00001709, 0.00001012, -0.00000959, -0.00001540, 0.00001189, 0.00002043, -0.00001145, -0.00001584, 0.00001480, 0.00001218, -0.00000717, 0.00000728, 0.00001175, -0.00001480, -0.00000701, 0.00002606, 0.00001299, -0.00001405, 0.00000185, 0.00001167, -0.00001501, -0.00001851, 0.00000889, 0.00000970, -0.00000964, -0.00000636, 0.00000541, 0.00000052, -0.00000673, -0.00000547, -0.00000385, -0.00000263, 0.00000003, 0.00000064, -0.00000006, -0.00000038, -0.00000089, 0.00000073, 0.00000333, 0.00000157, -0.00000246, -0.00000340, -0.00000213, -0.00000069, -0.00000092, -0.00000443, -0.00000507, 0.00000142, 0.00000449, -0.00000091, -0.00000256, 0.00000140, -0.00000145, -0.00000665, 0.00000029, 0.00001115, 0.00000864, -0.00000180, -0.00000561, },
		{ -0.00000386, -0.00000565, -0.00000486, -0.00000509, 0.00000107, 0.00001071, 0.00000862, -0.00000297, -0.00000554, 0.00000043, -0.00000003, -0.00000313, 0.00000131, 0.00000417, -0.00000177, -0.00000535, -0.00000192, 0.00000001, -0.00000137, -0.00000242, -0.00000317, -0.00000209, 0.00000134, 0.00000258, 0.00000113, 0.00000000, -0.00000103, -0.00000077, 0.00000126, 0.00000013, -0.00000348, -0.00000480, -0.00000598, -0.00000441, 0.00000339, 0.00000209, -0.00000955, -0.00000279, 0.00001295, -0.00000238, -0.00002162, -0.00000047, 0.00001465, -0.00001040, -0.00001059, 0.00002535, 0.00001830, -0.00001843, -0.00000789, 0.00001546, -0.00000143, -0.00000589, 0.00001902, 0.00000916, -0.00001943, -0.00000425, 0.00001941, 0.00000484, -0.00001171, -0.00000685, 0.00000226, 0.00001578, 0.00001544, -0.00002057, -0.00003602, 0.00001096, 0.00004513, 0.00001208, -0.00002720, -0.00002374, 0.00000112, 0.00002147, 0.00001557, -0.00001428, -0.00001503, 0.00001858, 0.00001676, -0.00001584, -0.00000860, 0.00001588, 0.00000436, -0.00000327, 0.00001357, 0.00000378, -0.00001570, 0.00000729, 0.00002604, 0.00000004, -0.00000996, 0.00001125, 0.00000130, -0.00002513, -0.00000930, 0.00001587, 0.00000274, -0.00001281, -0.00000150, 0.00000605, -0.00000298, -0.00000768, -0.00000466, -0.00000303, -0.00000189, -0.00000034, 0.00000027, 0.00000080, -0.00000023, -0.00000178, 0.00000077, 0.00000391, 0.00000151, -0.00000270, -0.00000341, -0.00000168, -0.00000041, -0.00000260, -0.00000636, -0.00000343, 0.00000406, 0.00000349, -0.00000250, -0.00000120, 0.00000186, -0.00000353, -0.00000623, 0.00000376, 0.00001202, 0.00000585, },
		{ 0.00001133, 0.00000294, -0.00000504, -0.00000548, -0.00000506, -0.00000382, 0.00000447, 0.00001142, 0.00000470, -0.00000518, -0.00000313, 0.00000142, -0.00000201, -0.00000251, 0.00000345, 0.00000233, -0.00000468, -0.00000419, 0.00000054, 0.00000011, -0.00000215, -0.00000251, -0.00000274, -0.00000168, 0.00000076, 0.00000189, 0.00000196, 0.00000060, -0.00000205, -0.00000114, 0.00000197, -0.00000004, -0.00000443, -0.00000585, -0.00000562, -0.00000100, 0.00000435, -0.00000298, -0.00000922, 0.00000564, 0.00001046, -0.00001440, -0.00001684, 0.00001372, 0.00000924, -0.00002044, -0.00000020, 0.00003251, 0.00000509, -0.00002346, 0.00000173, 0.00001336, -0.00000907, -0.00000014, 0.00002197, 0.00000134, -0.00001785, 0.00000267, 0.00001382, 0.00000012, -0.00000552, -0.00000757, -0.00000573, 0.00001759, 0.00002109, -0.00002170, -0.00003712, 0.00001044, 0.00004220, 0.00001514, -0.00002200, -0.00002891, -0.00000353, 0.00002742, 0.00001729, -0.00001849, -0.00001076, 0.00002219, 0.00001020, -0.00001542, -0.00000070, 0.00001196, -0.00000115, 0.00000427, 0.00001486, -0.00000503, -0.00000945, 0.00001989, 0.00001877, -0.00000897, -0.00000077, 0.00001393, -0.00001278, -0.00002789, 0.00000226, 0.00001688, -0.00000484, -0.00001184, 0.00000287, 0.00000418, -0.00000590, -0.00000750, -0.00000392, -0.00000225, -0.00000132, -0.00000087, 0.00000025, 0.00000157, -0.00000054, -0.00000243, 0.00000110, 0.00000418, 0.00000130, -0.00000272, -0.00000319, -0.00000121, -0.00000077, -0.00000471, -0.00000719, -0.00000096, 0.00000554, 0.00000174, -0.00000304, 0.00000037, 0.00000129, -0.00000515, -0.00000450, 0.00000669, },
		{ -0.00000195, 0.00000856, 0.00000945, 0.00000035, -0.00000545, -0.00000533, -0.00000517, -0.00000160, 0.00000787, 0.00001027, 0.00000006, -0.00000567, -0.00000007, 0.00000111, -0.00000382, -0.00000071, 0.00000469, -0.00000073, -0.00000650, -0.00000189, 0.00000236, -0.00000044, -0.00000281, -0.00000242, -0.00000212, -0.00000139, -0.00000014, 0.00000149, 0.00000306, 0.00000072, -0.00000323, -0.00000112, 0.00000260, -0.00000049, -0.00000555, -0.00000673, -0.00000397, 0.00000257, 0.00000261, -0.00000794, -0.00000469, 0.00001260, 0.00000219, -0.00002290, -0.00000519, 0.00002271, -0.00000240, -0.00002398, 0.00001373, 0.00003163, -0.00000925, -0.00002100, 0.00000995, 0.00000636, -0.00001269, 0.00000754, 0.00001977, -0.00000544, -0.00001160, 0.00000646, 0.00000555, -0.00000071, 0.00000071, -0.00001168, -0.00001164, 0.00002124, 0.00002388, -0.00002265, -0.00003595, 0.00000830, 0.00003882, 0.00002020, -0.00001847, -0.00003502, -0.00000475, 0.00003269, 0.00001550, -0.00001985, -0.00000440, 0.00002114, 0.00000371, -0.00001008, 0.00000470, 0.00000468, -0.00000205, 0.00001237, 0.00001067, -0.00001073, 0.00000244, 0.00002690, 0.00000738, -0.00001138, 0.00000912, 0.00000867, -0.00002566, -0.00002311, 0.00001197, 0.00001257, -0.00001017, -0.00000783, 0.00000515, 0.00000084, -0.00000756, -0.00000677, -0.00000335, -0.00000143, -0.00000103, -0.00000140, 0.00000057, 0.00000201, -0.00000114, -0.00000265, 0.00000151, 0.00000408, 0.00000110, -0.00000251, -0.00000279, -0.00000093, -0.00000185, -0.00000672, -0.00000671, 0.00000165, 0.00000565, -0.00000006, -0.00000253, 0.00000156, 0.00000001, -0.00000584, },
		{ -0.00000153, -0.00000543, 0.00000083, 0.00000916, 0.00000682, -0.00000160, -0.00000526, -0.00000535, -0.00000491, 0.00000158, 0.00001052, 0.00000710, -0.00000429, -0.00000421, 0.00000272, -0.00000043, -0.00000473, 0.00000178, 0.00000439, -0.00000425, -0.00000672, 0.00000088, 0.00000322, -0.00000130, -0.00000324, -0.00000218, -0.00000138, -0.00000132, -0.00000122, 0.00000152, 0.00000419, 0.00000028, -0.00000429, -0.00000073, 0.00000301, -0.00000123, -0.00000679, -0.00000697, -0.00000100, 0.00000504, -0.00000163, -0.00001064, 0.00000308, 0.00001516, -0.00000958, -0.00002456, 0.00000973, 0.00002342, -0.00001575, -0.00001921, 0.00002624, 0.00002270, -0.00002007, -0.00001264, 0.00001355, -0.00000274, -0.00001112, 0.00001385, 0.00001324, -0.00000840, -0.00000300, 0.00000566, -0.00000260, 0.00000217, 0.00000462, -0.00001764, -0.00001421, 0.00002473, 0.00002388, -0.00002195, -0.00003367, 0.00000427, 0.00003645, 0.00002617, -0.00001748, -0.00004006, -0.00000286, 0.00003515, 0.00001191, -0.00001714, 0.00000132, 0.00001573, 0.00000019, -0.00000170, 0.00000557, -0.00000291, 0.00000194, 0.00001764, 0.00000285, -0.00001041, 0.00001634, 0.00002661, -0.00000371, -0.00000739, 0.00001508, -0.00000271, -0.00003307, -0.00001325, 0.00001678, 0.00000537, -0.00001164, -0.00000281, 0.00000477, -0.00000264, -0.00000791, -0.00000598, -0.00000286, -0.00000063, -0.00000105, -0.00000171, 0.00000107, 0.00000197, -0.00000176, -0.00000245, 0.00000178, 0.00000369, 0.00000101, -0.00000207, -0.00000228, -0.00000105, -0.00000353, -0.00000810, -0.00000508, 0.00000367, 0.00000459, -0.00000130, -0.00000127, 0.00000205, },
		{ 0.00000027, 0.00000176, -0.00000284, -0.00000408, 0.00000331, 0.00000851, 0.00000389, -0.00000278, -0.00000474, -0.00000553, -0.00000390, 0.00000544, 0.00001160, 0.00000225, -0.00000728, -0.00000113, 0.00000437, -0.00000267, -0.00000424, 0.00000409, 0.00000240, -0.00000726, -0.00000532, 0.00000334, 0.00000315, -0.00000209, -0.00000345, -0.00000183, -0.00000056, -0.00000158, -0.00000228, 0.00000203, 0.00000504, -0.00000064, -0.00000501, -0.00000010, 0.00000308, -0.00000228, -0.00000795, -0.00000609, 0.00000272, 0.00000528, -0.00000711, -0.00000954, 0.00001166, 0.00001180, -0.00002103, -0.00001830, 0.00002295, 0.00001570, -0.00002564, -0.00000747, 0.00003279, 0.00000859, -0.00002412, -0.00000197, 0.00001134, -0.00001046, -0.00000538, 0.00001617, 0.00000496, -0.00000640, 0.00000486, 0.00000073, -0.00000809, 0.00000701, 0.00000498, -0.00002317, -0.00001363, 0.00002622, 0.00002225, -0.00001900, -0.00003172, -0.00000090, 0.00003598, 0.00003139, -0.00001868, -0.00004219, 0.00000051, 0.00003381, 0.00000897, -0.00001087, 0.00000387, 0.00000800, 0.00000131, 0.00000666, 0.00000190, -0.00000765, 0.00000901, 0.00001791, -0.00000511, -0.00000343, 0.00002795, 0.00002003, -0.00001065, 0.00000013, 0.00001414, -0.00001610, -0.00003293, -0.00000230, 0.00001572, -0.00000163, -0.00000929, 0.00000122, 0.00000228, -0.00000525, -0.00000741, -0.00000539, -0.00000228, -0.00000000, -0.00000132, -0.00000167, 0.00000151, 0.00000149, -0.00000215, -0.00000197, 0.00000176, 0.00000316, 0.00000111, -0.00000145, -0.00000180, -0.00000175, -0.00000549, -0.00000844, -0.00000278, 0.00000461, 0.00000287, -0.00000163, },
		{ 0.00000109, -0.00000101, 0.00000166, 0.00000087, -0.00000360, -0.00000207, 0.00000510, 0.00000679, 0.00000108, -0.00000315, -0.00000414, -0.00000573, -0.00000186, 0.00000936, 0.00001047, -0.00000339, -0.00000808, 0.00000269, 0.00000439, -0.00000471, -0.00000241, 0.00000528, -0.00000088, -0.00000887, -0.00000279, 0.00000489, 0.00000248, -0.00000260, -0.00000356, -0.00000136, 0.00000022, -0.00000217, -0.00000307, 0.00000289, 0.00000536, -0.00000184, -0.00000521, 0.00000059, 0.00000268, -0.00000358, -0.00000862, -0.00000386, 0.00000613, 0.00000286, -0.00001184, -0.00000444, 0.00001798, 0.00000313, -0.00002809, -0.00000590, 0.00002996, 0.00000246, -0.00002817, 0.00000727, 0.00003097, -0.00000615, -0.00002074, 0.00000697, 0.00000453, -0.00001412, 0.00000185, 0.00001362, -0.00000190, -0.00000031, 0.00000936, -0.00000626, -0.00000971, 0.00001136, 0.00000218, -0.00002625, -0.00001133, 0.00002481, 0.00002063, -0.00001429, -0.00003114, -0.00000580, 0.00003723, 0.00003432, -0.00002062, -0.00004063, 0.00000314, 0.00002923, 0.00000882, -0.00000305, 0.00000208, 0.00000094, 0.00000671, 0.00001203, -0.00000438, -0.00000765, 0.00001602, 0.00001313, -0.00000953, 0.00000834, 0.00003395, 0.00001040, -0.00001168, 0.00000687, 0.00000613, -0.00002679, -0.00002597, 0.00000586, 0.00001003, -0.00000592, -0.00000459, 0.00000299, -0.00000114, -0.00000655, -0.00000664, -0.00000496, -0.00000157, 0.00000029, -0.00000163, -0.00000128, 0.00000167, 0.00000074, -0.00000216, -0.00000143, 0.00000143, 0.00000264, 0.00000139, -0.00000071, -0.00000150, -0.00000304, -0.00000728, -0.00000762, -0.00000044, 0.00000432, },
		{ 0.00000133, 0.00000299, -0.00000022, 0.00000033, 0.00000258, -0.00000034, -0.00000369, 0.00000021, 0.00000599, 0.00000429, -0.00000129, -0.00000283, -0.00000370, -0.00000559, 0.00000130, 0.00001242, 0.00000696, -0.00000855, -0.00000648, 0.00000608, 0.00000288, -0.00000564, 0.00000017, 0.00000467, -0.00000451, -0.00000859, 0.00000003, 0.00000526, 0.00000166, -0.00000278, -0.00000365, -0.00000075, 0.00000083, -0.00000302, -0.00000340, 0.00000386, 0.00000501, -0.00000298, -0.00000484, 0.00000108, 0.00000173, -0.00000492, -0.00000832, -0.00000057, 0.00000802, -0.00000163, -0.00001386, 0.00000324, 0.00001952, -0.00000816, -0.00002803, 0.00000860, 0.00002840, -0.00001141, -0.00002222, 0.00001993, 0.00002133, -0.00001699, -0.00001190, 0.00001124, -0.00000395, -0.00001294, 0.00000746, 0.00000738, -0.00000493, 0.00000741, 0.00000944, -0.00001255, -0.00000794, 0.00001313, -0.00000212, -0.00002596, -0.00000927, 0.00002086, 0.00002031, -0.00000910, -0.00003202, -0.00000904, 0.00003904, 0.00003436, -0.00002133, -0.00003599, 0.00000317, 0.00002330, 0.00001231, 0.00000361, -0.00000338, -0.00000281, 0.00001427, 0.00001284, -0.00001018, -0.00000303, 0.00001986, 0.00000543, -0.00000802, 0.00002131, 0.00003330, 0.00000166, -0.00000776, 0.00000893, -0.00000633, -0.00003127, -0.00001521, 0.00000880, 0.00000243, -0.00000644, 0.00000042, 0.00000237, -0.00000425, -0.00000676, -0.00000601, -0.00000447, -0.00000085, 0.00000018, -0.00000174, -0.00000068, 0.00000140, -0.00000001, -0.00000182, -0.00000103, 0.00000089, 0.00000225, 0.00000180, 0.00000008, -0.00000155, -0.00000480, -0.00000844, -0.00000582, },
		{ -0.00000865, -0.00000348, 0.00000210, 0.00000106, -0.00000073, 0.00000205, 0.00000292, -0.00000157, -0.00000313, 0.00000242, 0.00000590, 0.00000133, -0.00000294, -0.00000207, -0.00000349, -0.00000468, 0.00000527, 0.00001362, 0.00000153, -0.00001193, -0.00000297, 0.00000799, 0.00000055, -0.00000492, 0.00000244, 0.00000213, -0.00000734, -0.00000652, 0.00000224, 0.00000462, 0.00000113, -0.00000277, -0.00000375, 0.00000001, 0.00000112, -0.00000392, -0.00000315, 0.00000460, 0.00000401, -0.00000374, -0.00000402, 0.00000108, 0.00000028, -0.00000586, -0.00000674, 0.00000295, 0.00000757, -0.00000666, -0.00001210, 0.00001088, 0.00001537, -0.00001817, -0.00002058, 0.00002033, 0.00001898, -0.00002092, -0.00000981, 0.00002626, 0.00000712, -0.00002110, -0.00000127, 0.00001005, -0.00001084, -0.00000819, 0.00000921, 0.00000005, -0.00000341, 0.00001382, 0.00000589, -0.00001596, -0.00000466, 0.00001147, -0.00000572, -0.00002280, -0.00000907, 0.00001569, 0.00002164, -0.00000484, -0.00003349, -0.00000995, 0.00003983, 0.00003209, -0.00001925, -0.00003006, -0.00000005, 0.00001843, 0.00001861, 0.00000689, -0.00001027, -0.00000208, 0.00002096, 0.00000950, -0.00001266, 0.00000411, 0.00001882, -0.00000176, -0.00000046, 0.00003161, 0.00002740, -0.00000318, -0.00000206, 0.00000445, -0.00001894, -0.00002846, -0.00000467, 0.00000634, -0.00000416, -0.00000374, 0.00000399, 0.00000014, -0.00000628, -0.00000641, -0.00000561, -0.00000378, -0.00000034, -0.00000022, -0.00000152, -0.00000013, 0.00000077, -0.00000053, -0.00000129, -0.00000089, 0.00000032, 0.00000200, 0.00000229, 0.00000083, -0.00000207, -0.00000673, },
		{ -0.00000306, -0.00000843, -0.00000782, -0.00000118, 0.00000176, -0.00000090, -0.00000037, 0.00000376, 0.00000272, -0.00000261, -0.00000205, 0.00000430, 0.00000484, -0.00000170, -0.00000375, -0.00000118, -0.00000337, -0.00000261, 0.00000931, 0.00001222, -0.00000471, -0.00001261, 0.00000132, 0.00000786, -0.00000149, -0.00000265, 0.00000332, -0.00000173, -0.00000841, -0.00000337, 0.00000321, 0.00000347, 0.00000108, -0.00000277, -0.00000377, 0.00000084, 0.00000099, -0.00000461, -0.00000240, 0.00000475, 0.00000256, -0.00000379, -0.00000300, 0.00000031, -0.00000144, -0.00000586, -0.00000397, 0.00000556, 0.00000481, -0.00001025, -0.00000693, 0.00001562, 0.00000678, -0.00002326, -0.00000814, 0.00002536, 0.00000521, -0.00002279, 0.00000473, 0.00002443, -0.00000706, -0.00001827, 0.00000733, 0.00000485, -0.00001388, -0.00000251, 0.00000668, -0.00000542, 0.00000148, 0.00001671, 0.00000089, -0.00001567, -0.00000219, 0.00000706, -0.00000688, -0.00001834, -0.00001130, 0.00001089, 0.00002398, -0.00000236, -0.00003419, -0.00000889, 0.00003832, 0.00002910, -0.00001397, -0.00002501, -0.00000566, 0.00001646, 0.00002558, 0.00000604, -0.00001575, 0.00000241, 0.00002421, 0.00000419, -0.00001038, 0.00001070, 0.00001331, -0.00000516, 0.00001106, 0.00003664, 0.00001936, -0.00000332, 0.00000152, -0.00000566, -0.00002745, -0.00001997, 0.00000224, 0.00000025, -0.00000774, 0.00000050, 0.00000528, -0.00000246, -0.00000708, -0.00000603, -0.00000527, -0.00000290, -0.00000024, -0.00000068, -0.00000098, 0.00000015, -0.00000002, -0.00000072, -0.00000079, -0.00000097, -0.00000015, 0.00000185, 0.00000279, 0.00000141, },
		{ 0.00000330, 0.00000170, -0.00000442, -0.00000951, -0.00000615, 0.00000054, 0.00000049, -0.00000243, 0.00000071, 0.00000518, 0.00000211, -0.00000329, -0.00000060, 0.00000558, 0.00000292, -0.00000441, -0.00000376, -0.00000047, -0.00000296, 0.00000075, 0.00001233, 0.00000805, -0.00001021, -0.00001039, 0.00000495, 0.00000589, -0.00000218, 0.00000040, 0.00000209, -0.00000579, -0.00000727, -0.00000016, 0.00000278, 0.00000241, 0.00000142, -0.00000293, -0.00000357, 0.00000157, 0.00000045, -0.00000479, -0.00000138, 0.00000405, 0.00000109, -0.00000297, -0.00000224, -0.00000137, -0.00000290, -0.00000449, -0.00000071, 0.00000623, 0.00000078, -0.00001083, -0.00000017, 0.00001552, -0.00000323, -0.00002148, 0.00000501, 0.00002223, -0.00000797, -0.00001670, 0.00001647, 0.00001555, -0.00001703, -0.00001076, 0.00001141, -0.00000150, -0.00001268, 0.00000130, 0.00000130, -0.00000711, 0.00000722, 0.00001548, -0.00000303, -0.00001253, -0.00000230, 0.00000172, -0.00000504, -0.00001446, -0.00001536, 0.00000756, 0.00002603, -0.00000158, -0.00003296, -0.00000713, 0.00003423, 0.00002727, -0.00000637, -0.00002245, -0.00001167, 0.00001789, 0.00003070, 0.00000207, -0.00001758, 0.00000835, 0.00002292, -0.00000011, -0.00000394, 0.00001400, 0.00000566, -0.00000293, 0.00002312, 0.00003595, 0.00001248, -0.00000039, 0.00000005, -0.00001816, -0.00002931, -0.00000923, 0.00000389, -0.00000668, -0.00000783, 0.00000435, 0.00000453, -0.00000441, -0.00000704, -0.00000588, -0.00000475, -0.00000205, -0.00000055, -0.00000094, -0.00000031, 0.00000008, -0.00000072, -0.00000066, -0.00000050, -0.00000112, -0.00000047, 0.00000171, },
		{ -0.00000068, 0.00000156, 0.00000377, 0.00000159, -0.00000589, -0.00000967, -0.00000404, 0.00000129, -0.00000129, -0.00000320, 0.00000221, 0.00000611, 0.00000123, -0.00000355, 0.00000104, 0.00000608, 0.00000037, -0.00000643, -0.00000321, -0.00000008, -0.00000175, 0.00000504, 0.00001319, 0.00000174, -0.00001346, -0.00000598, 0.00000670, 0.00000299, -0.00000092, 0.00000300, -0.00000127, -0.00000872, -0.00000422, 0.00000207, 0.00000131, 0.00000190, 0.00000186, -0.00000321, -0.00000306, 0.00000194, -0.00000032, -0.00000421, -0.00000054, 0.00000241, 0.00000013, -0.00000136, -0.00000227, -0.00000381, -0.00000339, -0.00000170, 0.00000195, 0.00000460, -0.00000276, -0.00000793, 0.00000546, 0.00001041, -0.00001091, -0.00001341, 0.00001435, 0.00001247, -0.00001602, -0.00000526, 0.00002171, 0.00000314, -0.00002055, -0.00000215, 0.00001062, -0.00000608, -0.00000874, 0.00000152, -0.00000435, -0.00000478, 0.00001116, 0.00001126, -0.00000407, -0.00000850, -0.00000540, -0.00000248, -0.00000097, -0.00001253, -0.00001989, 0.00000586, 0.00002655, -0.00000154, -0.00002942, -0.00000618, 0.00002843, 0.00002791, 0.00000165, -0.00002271, -0.00001587, 0.00002165, 0.00003216, -0.00000272, -0.00001508, 0.00001290, 0.00001798, -0.00000092, 0.00000438, 0.00001281, -0.00000087, 0.00000461, 0.00003244, 0.00003118, 0.00000874, 0.00000240, -0.00000711, -0.00002878, -0.00002452, -0.00000008, 0.00000084, -0.00001187, -0.00000541, 0.00000645, 0.00000271, -0.00000523, -0.00000672, -0.00000588, -0.00000397, -0.00000149, -0.00000111, -0.00000090, 0.00000023, -0.00000024, -0.00000113, -0.00000055, -0.00000046, -0.00000119, },
		{ -0.00000054, -0.00000110, -0.00000090, 0.00000140, 0.00000411, 0.00000103, -0.00000710, -0.00000886, -0.00000206, 0.00000099, -0.00000306, -0.00000311, 0.00000382, 0.00000648, 0.00000026, -0.00000330, 0.00000265, 0.00000561, -0.00000244, -0.00000750, -0.00000246, 0.00000009, 0.00000074, 0.00000942, 0.00001104, -0.00000536, -0.00001345, -0.00000080, 0.00000597, 0.00000051, 0.00000209, 0.00000390, -0.00000598, -0.00000948, -0.00000019, 0.00000271, -0.00000049, 0.00000204, 0.00000207, -0.00000339, -0.00000228, 0.00000169, -0.00000095, -0.00000285, -0.00000039, 0.00000005, 0.00000021, 0.00000065, -0.00000361, -0.00000651, -0.00000222, 0.00000196, 0.00000279, 0.00000127, -0.00000398, -0.00000258, 0.00000749, 0.00000219, -0.00001324, -0.00000202, 0.00001679, -0.00000002, -0.00001653, 0.00000707, 0.00001933, -0.00000840, -0.00001791, 0.00000414, 0.00000662, -0.00000717, -0.00000459, -0.00000179, -0.00000771, 0.00000006, 0.00001158, 0.00000628, -0.00000190, -0.00000575, -0.00001052, -0.00000412, 0.00000366, -0.00001286, -0.00002341, 0.00000501, 0.00002495, -0.00000094, -0.00002421, -0.00000706, 0.00002260, 0.00003118, 0.00000809, -0.00002474, -0.00001676, 0.00002557, 0.00002967, -0.00000570, -0.00000939, 0.00001391, 0.00001173, 0.00000266, 0.00001166, 0.00000793, -0.00000349, 0.00001525, 0.00003714, 0.00002506, 0.00000800, 0.00000185, -0.00001801, -0.00003403, -0.00001547, 0.00000477, -0.00000475, -0.00001392, -0.00000221, 0.00000645, 0.00000094, -0.00000506, -0.00000649, -0.00000577, -0.00000306, -0.00000138, -0.00000165, -0.00000063, 0.00000046, -0.00000058, -0.00000124, -0.00000059, },
		{ -0.00000120, -0.00000086, -0.00000057, -0.00000088, -0.00000118, 0.00000129, 0.00000420, 0.00000009, -0.00000766, -0.00000724, -0.00000071, -0.00000017, -0.00000436, -0.00000226, 0.00000523, 0.00000628, -0.00000062, -0.00000251, 0.00000395, 0.00000413, -0.00000504, -0.00000760, -0.00000191, 0.00000048, 0.00000466, 0.00001263, 0.00000578, -0.00001144, -0.00001008, 0.00000344, 0.00000312, -0.00000034, 0.00000589, 0.00000236, -0.00001065, -0.00000775, 0.00000355, 0.00000177, -0.00000188, 0.00000256, 0.00000190, -0.00000314, -0.00000153, 0.00000072, -0.00000097, -0.00000095, -0.00000143, -0.00000249, 0.00000168, 0.00000236, -0.00000649, -0.00000857, 0.00000100, 0.00000534, 0.00000107, -0.00000222, -0.00000180, 0.00000298, 0.00000468, -0.00000587, -0.00000921, 0.00000838, 0.00001185, -0.00001045, -0.00001002, 0.00001571, 0.00001104, -0.00001544, -0.00001159, 0.00000621, 0.00000213, -0.00000485, -0.00000260, -0.00000694, -0.00000740, 0.00000492, 0.00000841, 0.00000285, 0.00000234, -0.00000563, -0.00001579, -0.00000299, 0.00000712, -0.00001466, -0.00002510, 0.00000377, 0.00002159, 0.00000115, -0.00001872, -0.00000982, 0.00001850, 0.00003601, 0.00001165, -0.00002659, -0.00001426, 0.00002743, 0.00002455, -0.00000513, -0.00000289, 0.00001087, 0.00000699, 0.00000965, 0.00001560, 0.00000174, -0.00000091, 0.00002598, 0.00003727, 0.00001999, 0.00000815, -0.00000373, -0.00002899, -0.00003251, -0.00000581, 0.00000460, -0.00001018, -0.00001296, 0.00000010, 0.00000491, -0.00000005, -0.00000439, -0.00000641, -0.00000535, -0.00000231, -0.00000165, -0.00000194, -0.00000037, 0.00000037, -0.00000071, },
		{ 0.00000013, -0.00000053, -0.00000123, -0.00000130, -0.00000043, -0.00000069, -0.00000152, 0.00000126, 0.00000389, -0.00000099, -0.00000731, -0.00000525, -0.00000034, -0.00000179, -0.00000487, -0.00000090, 0.00000617, 0.00000555, -0.00000121, -0.00000123, 0.00000458, 0.00000179, -0.00000693, -0.00000695, -0.00000178, 0.00000165, 0.00000962, 0.00001339, -0.00000181, -0.00001478, -0.00000431, 0.00000537, -0.00000072, 0.00000106, 0.00000901, -0.00000149, -0.00001379, -0.00000403, 0.00000587, -0.00000014, -0.00000240, 0.00000295, 0.00000149, -0.00000223, -0.00000124, -0.00000080, 0.00000007, 0.00000094, -0.00000389, -0.00000443, 0.00000456, 0.00000288, -0.00001067, -0.00000890, 0.00000600, 0.00000699, -0.00000303, -0.00000388, 0.00000352, 0.00000602, -0.00000225, -0.00001029, -0.00000027, 0.00001382, 0.00000187, -0.00001499, 0.00000036, 0.00001775, 0.00000050, -0.00001643, -0.00000500, 0.00000420, -0.00000040, -0.00000080, -0.00000389, -0.00001145, -0.00000371, 0.00000750, 0.00000309, 0.00000231, 0.00000663, -0.00000821, -0.00001931, -0.00000008, 0.00000834, -0.00001657, -0.00002504, 0.00000111, 0.00001756, 0.00000482, -0.00001437, -0.00001343, 0.00001719, 0.00004058, 0.00001233, -0.00002635, -0.00000969, 0.00002590, 0.00001913, -0.00000085, 0.00000181, 0.00000501, 0.00000578, 0.00001771, 0.00001547, -0.00000282, 0.00000622, 0.00003422, 0.00003446, 0.00001683, 0.00000632, -0.00001366, -0.00003617, -0.00002534, 0.00000111, 0.00000064, -0.00001341, -0.00001029, 0.00000066, 0.00000285, -0.00000012, -0.00000370, -0.00000625, -0.00000467, -0.00000199, -0.00000206, -0.00000194, -0.00000035, },
		{ -0.00000180, -0.00000068, 0.00000000, -0.00000014, -0.00000149, -0.00000169, -0.00000012, -0.00000068, -0.00000179, 0.00000123, 0.00000305, -0.00000186, -0.00000595, -0.00000342, -0.00000103, -0.00000336, -0.00000447, 0.00000065, 0.00000647, 0.00000442, -0.00000126, 0.00000036, 0.00000422, -0.00000107, -0.00000776, -0.00000598, -0.00000197, 0.00000415, 0.00001465, 0.00001080, -0.00001011, -0.00001434, 0.00000216, 0.00000441, -0.00000395, 0.00000438, 0.00001007, -0.00000663, -0.00001431, 0.00000043, 0.00000622, -0.00000204, -0.00000210, 0.00000272, 0.00000130, -0.00000071, -0.00000190, -0.00000239, 0.00000238, 0.00000206, -0.00000763, -0.00000493, 0.00000834, 0.00000140, -0.00001536, -0.00000658, 0.00001173, 0.00000564, -0.00000825, -0.00000206, 0.00001024, 0.00000452, -0.00001077, -0.00000881, 0.00001009, 0.00001218, -0.00000895, -0.00001222, 0.00001012, 0.00001292, -0.00000817, -0.00001223, -0.00000104, -0.00000002, 0.00000020, 0.00000260, -0.00000794, -0.00001320, 0.00000158, 0.00000659, -0.00000215, 0.00000457, 0.00000905, -0.00001225, -0.00002002, 0.00000293, 0.00000731, -0.00001732, -0.00002416, -0.00000327, 0.00001409, 0.00000926, -0.00001202, -0.00001627, 0.00001857, 0.00004318, 0.00001136, -0.00002298, -0.00000515, 0.00002114, 0.00001573, 0.00000570, 0.00000309, -0.00000124, 0.00000852, 0.00002418, 0.00001234, -0.00000349, 0.00001582, 0.00003879, 0.00003078, 0.00001458, 0.00000031, -0.00002509, -0.00003710, -0.00001557, 0.00000345, -0.00000469, -0.00001373, -0.00000757, -0.00000040, 0.00000119, 0.00000034, -0.00000317, -0.00000574, -0.00000398, -0.00000215, -0.00000231, },
		{ -0.00000262, -0.00000225, -0.00000180, -0.00000124, 0.00000015, 0.00000023, -0.00000193, -0.00000182, 0.00000016, -0.00000090, -0.00000184, 0.00000105, 0.00000173, -0.00000210, -0.00000380, -0.00000227, -0.00000255, -0.00000440, -0.00000327, 0.00000208, 0.00000602, 0.00000313, -0.00000058, 0.00000193, 0.00000265, -0.00000387, -0.00000740, -0.00000516, -0.00000207, 0.00000815, 0.00001836, 0.00000482, -0.00001705, -0.00001015, 0.00000734, 0.00000108, -0.00000520, 0.00000842, 0.00000833, -0.00001139, -0.00001203, 0.00000410, 0.00000484, -0.00000302, -0.00000153, 0.00000169, 0.00000190, 0.00000104, -0.00000382, -0.00000333, 0.00000579, 0.00000163, -0.00001202, -0.00000334, 0.00001206, -0.00000249, -0.00001923, -0.00000124, 0.00001650, 0.00000083, -0.00001251, 0.00000378, 0.00001564, -0.00000195, -0.00001738, -0.00000141, 0.00001764, 0.00000403, -0.00001614, -0.00000357, 0.00001524, 0.00000350, -0.00001200, -0.00000554, -0.00000099, -0.00000393, 0.00000338, 0.00000344, -0.00001286, -0.00001137, 0.00000607, 0.00000260, -0.00000531, 0.00000822, 0.00000862, -0.00001589, -0.00001805, 0.00000452, 0.00000497, -0.00001636, -0.00002364, -0.00000878, 0.00001195, 0.00001310, -0.00001152, -0.00001683, 0.00002148, 0.00004294, 0.00001058, -0.00001681, -0.00000256, 0.00001471, 0.00001559, 0.00001216, 0.00000089, -0.00000538, 0.00001398, 0.00002728, 0.00000850, 0.00000046, 0.00002543, 0.00004010, 0.00002748, 0.00001094, -0.00001010, -0.00003420, -0.00003170, -0.00000679, 0.00000141, -0.00000901, -0.00001184, -0.00000602, -0.00000226, 0.00000028, 0.00000085, -0.00000262, -0.00000480, -0.00000362, },
		{ -0.00000362, -0.00000378, -0.00000308, -0.00000196, -0.00000213, -0.00000178, 0.00000055, 0.00000032, -0.00000234, -0.00000164, 0.00000016, -0.00000123, -0.00000159, 0.00000045, 0.00000016, -0.00000141, -0.00000129, -0.00000215, -0.00000446, -0.00000462, -0.00000149, 0.00000308, 0.00000481, 0.00000199, 0.00000090, 0.00000298, -0.00000005, -0.00000600, -0.00000608, -0.00000479, -0.00000141, 0.00001323, 0.00001933, -0.00000358, -0.00002075, -0.00000342, 0.00000973, -0.00000318, -0.00000385, 0.00001145, 0.00000408, -0.00001406, -0.00000775, 0.00000573, 0.00000269, -0.00000263, -0.00000152, 0.00000016, 0.00000377, 0.00000230, -0.00000693, -0.00000288, 0.00000969, -0.00000089, -0.00001603, 0.00000052, 0.00001445, -0.00000856, -0.00002077, 0.00000656, 0.00001848, -0.00000674, -0.00001355, 0.00001263, 0.00001702, -0.00001183, -0.00001891, 0.00000959, 0.00001903, -0.00000751, -0.00001668, 0.00000730, 0.00001367, -0.00000671, -0.00001020, 0.00000044, -0.00000420, -0.00000550, 0.00000729, 0.00000115, -0.00001632, -0.00000671, 0.00000781, -0.00000281, -0.00000547, 0.00001129, 0.00000567, -0.00001749, -0.00001460, 0.00000393, 0.00000267, -0.00001402, -0.00002433, -0.00001429, 0.00001107, 0.00001505, -0.00001179, -0.00001442, 0.00002422, 0.00004023, 0.00001161, -0.00000934, -0.00000282, 0.00000885, 0.00001847, 0.00001630, -0.00000330, -0.00000590, 0.00001990, 0.00002671, 0.00000642, 0.00000821, 0.00003329, 0.00003946, 0.00002426, 0.00000368, -0.00002287, -0.00003783, -0.00002218, -0.00000167, -0.00000315, -0.00001094, -0.00000924, -0.00000596, -0.00000399, -0.00000009, 0.00000109, -0.00000173, },
		{ 0.00000117, -0.00000030, -0.00000261, -0.00000439, -0.00000326, -0.00000172, -0.00000276, -0.00000201, 0.00000094, 0.00000007, -0.00000240, -0.00000129, -0.00000030, -0.00000140, -0.00000113, -0.00000073, -0.00000126, 0.00000029, 0.00000097, -0.00000306, -0.00000623, -0.00000394, 0.00000054, 0.00000340, 0.00000294, 0.00000139, 0.00000303, 0.00000303, -0.00000353, -0.00000696, -0.00000427, -0.00000483, 0.00000059, 0.00001832, 0.00001665, -0.00001253, -0.00002014, 0.00000380, 0.00000879, -0.00000653, -0.00000040, 0.00001193, -0.00000134, -0.00001352, -0.00000310, 0.00000485, 0.00000110, -0.00000111, -0.00000280, -0.00000112, 0.00000695, 0.00000223, -0.00001066, -0.00000052, 0.00001306, -0.00000551, -0.00001847, 0.00000614, 0.00001434, -0.00001577, -0.00001876, 0.00001532, 0.00001627, -0.00001516, -0.00000985, 0.00002203, 0.00001276, -0.00002192, -0.00001389, 0.00002023, 0.00001314, -0.00001801, -0.00001020, 0.00001596, 0.00000597, -0.00001392, -0.00000421, 0.00000333, -0.00000866, -0.00000412, 0.00000978, -0.00000332, -0.00001665, -0.00000121, 0.00000611, -0.00000746, -0.00000313, 0.00001214, 0.00000161, -0.00001635, -0.00001129, 0.00000142, 0.00000148, -0.00001130, -0.00002632, -0.00001868, 0.00001057, 0.00001456, -0.00001138, -0.00000946, 0.00002542, 0.00003642, 0.00001507, -0.00000257, -0.00000541, 0.00000547, 0.00002275, 0.00001702, -0.00000716, -0.00000279, 0.00002405, 0.00002373, 0.00000773, 0.00001802, 0.00003888, 0.00003798, 0.00001939, -0.00000803, -0.00003441, -0.00003482, -0.00001206, -0.00000101, -0.00000777, -0.00001052, -0.00000732, -0.00000684, -0.00000501, -0.00000042, },
		{ -0.00000538, -0.00000113, 0.00000147, 0.00000156, -0.00000210, -0.00000512, -0.00000313, -0.00000181, -0.00000342, -0.00000187, 0.00000099, -0.00000032, -0.00000190, -0.00000113, -0.00000120, -0.00000116, -0.00000073, -0.00000251, -0.00000203, 0.00000277, 0.00000244, -0.00000472, -0.00000740, -0.00000246, 0.00000247, 0.00000289, 0.00000068, 0.00000165, 0.00000543, 0.00000174, -0.00000714, -0.00000656, -0.00000253, -0.00000485, 0.00000412, 0.00002189, 0.00001033, -0.00001970, -0.00001540, 0.00000928, 0.00000524, -0.00000741, 0.00000364, 0.00000910, -0.00000593, -0.00000971, 0.00000007, 0.00000197, 0.00000122, 0.00000061, -0.00000565, -0.00000118, 0.00001097, 0.00000021, -0.00001407, 0.00000381, 0.00001480, -0.00001165, -0.00001832, 0.00001238, 0.00001100, -0.00002244, -0.00001276, 0.00002287, 0.00000959, -0.00002177, -0.00000135, 0.00002882, 0.00000319, -0.00002853, -0.00000318, 0.00002639, 0.00000156, -0.00002338, 0.00000101, 0.00001898, -0.00000499, -0.00001586, 0.00000310, 0.00000244, -0.00001199, -0.00000074, 0.00000936, -0.00000795, -0.00001354, 0.00000288, 0.00000176, -0.00000967, 0.00000016, 0.00001019, -0.00000176, -0.00001290, -0.00000944, -0.00000196, 0.00000172, -0.00000931, -0.00002894, -0.00002136, 0.00000920, 0.00001200, -0.00000911, -0.00000322, 0.00002458, 0.00003327, 0.00002035, 0.00000195, -0.00000871, 0.00000534, 0.00002629, 0.00001471, -0.00000864, 0.00000253, 0.00002519, 0.00002035, 0.00001274, 0.00002816, 0.00004270, 0.00003561, 0.00001071, -0.00002278, -0.00004108, -0.00002651, -0.00000460, -0.00000369, -0.00001057, -0.00000892, -0.00000677, -0.00000771, },
		{ -0.00000730, -0.00000787, -0.00000564, -0.00000219, 0.00000234, 0.00000342, -0.00000215, -0.00000556, -0.00000294, -0.00000234, -0.00000374, -0.00000155, 0.00000047, -0.00000046, -0.00000089, -0.00000152, -0.00000225, -0.00000038, -0.00000077, -0.00000459, -0.00000180, 0.00000548, 0.00000282, -0.00000662, -0.00000769, -0.00000049, 0.00000401, 0.00000150, -0.00000157, 0.00000294, 0.00000747, -0.00000094, -0.00001013, -0.00000499, -0.00000127, -0.00000415, 0.00000877, 0.00002243, 0.00000152, -0.00002304, -0.00000803, 0.00001132, 0.00000084, -0.00000518, 0.00000627, 0.00000341, -0.00000769, -0.00000380, 0.00000030, -0.00000152, 0.00000358, 0.00000131, -0.00000965, 0.00000078, 0.00001483, -0.00000392, -0.00001604, 0.00000950, 0.00001399, -0.00001811, -0.00001513, 0.00001767, 0.00000459, -0.00002664, -0.00000345, 0.00002694, -0.00000040, -0.00002408, 0.00001027, 0.00003025, -0.00000936, -0.00002880, 0.00001008, 0.00002535, -0.00001195, -0.00002144, 0.00001288, 0.00001515, -0.00001532, -0.00001244, 0.00000875, -0.00000115, -0.00001262, 0.00000274, 0.00000597, -0.00001064, -0.00000813, 0.00000399, -0.00000341, -0.00000897, 0.00000251, 0.00000609, -0.00000307, -0.00000846, -0.00000955, -0.00000494, 0.00000291, -0.00000870, -0.00003110, -0.00002261, 0.00000597, 0.00000849, -0.00000466, 0.00000272, 0.00002234, 0.00003208, 0.00002595, 0.00000373, -0.00001076, 0.00000783, 0.00002736, 0.00001099, -0.00000675, 0.00000805, 0.00002344, 0.00001849, 0.00002064, 0.00003770, 0.00004543, 0.00003087, -0.00000299, -0.00003722, -0.00004076, -0.00001611, -0.00000156, -0.00000739, -0.00001099, -0.00000763, },
		{ -0.00000980, -0.00000761, -0.00000801, -0.00000723, -0.00000632, -0.00000321, 0.00000383, 0.00000479, -0.00000248, -0.00000554, -0.00000308, -0.00000310, -0.00000347, -0.00000148, -0.00000057, 0.00000007, 0.00000027, -0.00000265, -0.00000302, 0.00000074, -0.00000152, -0.00000646, -0.00000049, 0.00000774, 0.00000216, -0.00000815, -0.00000710, 0.00000166, 0.00000491, -0.00000068, -0.00000331, 0.00000515, 0.00000851, -0.00000463, -0.00001184, -0.00000279, -0.00000053, -0.00000209, 0.00001337, 0.00001896, -0.00000765, -0.00002144, -0.00000041, 0.00000944, -0.00000222, -0.00000050, 0.00000568, -0.00000347, -0.00000538, 0.00000219, -0.00000288, -0.00000386, 0.00000774, -0.00000014, -0.00001374, 0.00000506, 0.00001731, -0.00000966, -0.00001563, 0.00001544, 0.00001030, -0.00002341, -0.00000923, 0.00002044, -0.00000379, -0.00002685, 0.00000732, 0.00002599, -0.00001130, -0.00002062, 0.00002197, 0.00002509, -0.00002118, -0.00002190, 0.00002160, 0.00001689, -0.00002283, -0.00001273, 0.00002112, 0.00000582, -0.00002164, -0.00000558, 0.00001085, -0.00000531, -0.00001041, 0.00000435, 0.00000087, -0.00001012, -0.00000248, 0.00000188, -0.00000737, -0.00000614, 0.00000258, 0.00000135, -0.00000191, -0.00000457, -0.00001117, -0.00000666, 0.00000406, -0.00000937, -0.00003188, -0.00002333, 0.00000065, 0.00000536, 0.00000129, 0.00000723, 0.00002010, 0.00003315, 0.00003016, 0.00000355, -0.00001018, 0.00001130, 0.00002544, 0.00000794, -0.00000191, 0.00001203, 0.00002004, 0.00001919, 0.00003023, 0.00004662, 0.00004691, 0.00002158, -0.00002084, -0.00004730, -0.00003384, -0.00000731, -0.00000259, -0.00000987, },
		{ -0.00000568, -0.00000997, -0.00000852, -0.00000886, -0.00000798, -0.00000631, -0.00000759, -0.00000361, 0.00000554, 0.00000537, -0.00000259, -0.00000521, -0.00000386, -0.00000365, -0.00000266, -0.00000209, -0.00000169, 0.00000143, 0.00000102, -0.00000435, -0.00000312, 0.00000176, -0.00000301, -0.00000751, 0.00000156, 0.00000896, 0.00000086, -0.00000886, -0.00000584, 0.00000364, 0.00000497, -0.00000338, -0.00000410, 0.00000784, 0.00000804, -0.00000860, -0.00001198, -0.00000062, 0.00000009, 0.00000161, 0.00001633, 0.00001152, -0.00001462, -0.00001522, 0.00000488, 0.00000455, -0.00000214, 0.00000484, 0.00000100, -0.00000914, 0.00000097, 0.00000595, -0.00000869, -0.00000351, 0.00001246, -0.00000423, -0.00001653, 0.00001121, 0.00001735, -0.00001590, -0.00001248, 0.00002023, 0.00000415, -0.00002618, -0.00000181, 0.00001968, -0.00001225, -0.00002246, 0.00001703, 0.00001983, -0.00002003, -0.00001169, 0.00003024, 0.00001423, -0.00002842, -0.00000948, 0.00002743, 0.00000342, -0.00002741, -0.00000027, 0.00002289, -0.00000572, -0.00002227, 0.00000174, 0.00000922, -0.00000789, -0.00000657, 0.00000306, -0.00000389, -0.00000654, 0.00000134, -0.00000236, -0.00000880, -0.00000285, 0.00000015, -0.00000234, 0.00000105, -0.00000231, -0.00001322, -0.00000701, 0.00000425, -0.00001059, -0.00003100, -0.00002463, -0.00000618, 0.00000351, 0.00000743, 0.00001011, 0.00001932, 0.00003573, 0.00003191, 0.00000296, -0.00000675, 0.00001391, 0.00002129, 0.00000711, 0.00000447, 0.00001375, 0.00001662, 0.00002248, 0.00004069, 0.00005521, 0.00004563, 0.00000608, -0.00003969, -0.00004993, -0.00002311, -0.00000277, },
		{ -0.00001257, -0.00000303, -0.00000830, -0.00000806, -0.00000842, -0.00001048, -0.00000691, -0.00000590, -0.00000907, -0.00000309, 0.00000683, 0.00000532, -0.00000205, -0.00000502, -0.00000522, -0.00000352, -0.00000175, -0.00000356, -0.00000231, 0.00000338, 0.00000090, -0.00000609, -0.00000245, 0.00000213, -0.00000483, -0.00000729, 0.00000367, 0.00000885, -0.00000040, -0.00000857, -0.00000427, 0.00000520, 0.00000414, -0.00000615, -0.00000366, 0.00001029, 0.00000593, -0.00001192, -0.00001067, 0.00000094, 0.00000129, 0.00000657, 0.00001600, 0.00000138, -0.00001712, -0.00000609, 0.00000595, -0.00000128, 0.00000182, 0.00000845, -0.00000720, -0.00001126, 0.00000995, 0.00000572, -0.00001537, 0.00000024, 0.00001600, -0.00001055, -0.00001675, 0.00001809, 0.00001441, -0.00002124, -0.00000693, 0.00002258, -0.00000331, -0.00002563, 0.00000543, 0.00001530, -0.00001864, -0.00001417, 0.00002316, 0.00000985, -0.00002390, 0.00000060, 0.00003235, 0.00000058, -0.00002854, 0.00000477, 0.00002544, -0.00001083, -0.00002427, 0.00001163, 0.00001772, -0.00001558, -0.00001774, 0.00000680, 0.00000524, -0.00000770, -0.00000304, -0.00000084, -0.00000647, -0.00000129, 0.00000216, -0.00000686, -0.00000759, -0.00000079, -0.00000391, -0.00000390, 0.00000450, -0.00000192, -0.00001463, -0.00000656, 0.00000308, -0.00001138, -0.00002889, -0.00002722, -0.00001334, 0.00000301, 0.00001238, 0.00001207, 0.00002092, 0.00003845, 0.00003120, 0.00000352, -0.00000142, 0.00001451, 0.00001658, 0.00000897, 0.00001068, 0.00001358, 0.00001445, 0.00002786, 0.00005192, 0.00006312, 0.00003906, -0.00001535, -0.00005503, -0.00004443, },
		{ -0.00006266, -0.00003293, -0.00000584, -0.00000654, -0.00000866, -0.00000563, -0.00000984, -0.00001130, -0.00000530, -0.00000645, -0.00000999, -0.00000187, 0.00000713, 0.00000519, -0.00000083, -0.00000550, -0.00000673, -0.00000250, -0.00000138, -0.00000559, -0.00000192, 0.00000529, -0.00000022, -0.00000721, -0.00000136, 0.00000144, -0.00000629, -0.00000577, 0.00000503, 0.00000762, -0.00000098, -0.00000742, -0.00000272, 0.00000613, 0.00000245, -0.00000842, -0.00000204, 0.00001166, 0.00000251, -0.00001376, -0.00000851, 0.00000169, 0.00000384, 0.00001169, 0.00001131, -0.00000916, -0.00001394, 0.00000320, 0.00000222, -0.00000549, 0.00000895, 0.00000819, -0.00001705, -0.00000832, 0.00001910, 0.00000091, -0.00002065, 0.00000694, 0.00001680, -0.00001788, -0.00001372, 0.00002414, 0.00000877, -0.00002436, -0.00000007, 0.00002181, -0.00001048, -0.00002183, 0.00001071, 0.00000832, -0.00002118, -0.00000390, 0.00002403, -0.00000123, -0.00002151, 0.00001279, 0.00002735, -0.00001181, -0.00002126, 0.00001641, 0.00001613, -0.00002135, -0.00001471, 0.00001892, 0.00000760, -0.00002078, -0.00001033, 0.00000812, 0.00000109, -0.00000495, -0.00000148, -0.00000587, -0.00000600, 0.00000361, 0.00000011, -0.00000984, -0.00000474, -0.00000098, -0.00000809, -0.00000320, 0.00000707, -0.00000284, -0.00001478, -0.00000623, 0.00000082, -0.00001105, -0.00002649, -0.00003113, -0.00001968, 0.00000311, 0.00001534, 0.00001431, 0.00002483, 0.00004004, 0.00002902, 0.00000605, 0.00000424, 0.00001306, 0.00001301, 0.00001286, 0.00001558, 0.00001255, 0.00001405, 0.00003494, 0.00006415, 0.00006862, 0.00002484, -0.00003986, },
		{ 0.00000230, -0.00006249, -0.00006050, -0.00001966, -0.00000457, -0.00001052, -0.00000655, -0.00000435, -0.00001199, -0.00001064, -0.00000413, -0.00000778, -0.00000968, -0.00000064, 0.00000631, 0.00000563, 0.00000066, -0.00000688, -0.00000771, -0.00000088, -0.00000205, -0.00000749, -0.00000045, 0.00000638, -0.00000195, -0.00000722, -0.00000054, -0.00000026, -0.00000660, -0.00000344, 0.00000497, 0.00000588, -0.00000045, -0.00000577, -0.00000146, 0.00000628, 0.00000014, -0.00000959, 0.00000032, 0.00001124, -0.00000139, -0.00001358, -0.00000633, 0.00000189, 0.00000816, 0.00001531, 0.00000221, -0.00001731, -0.00000536, 0.00000975, -0.00000532, -0.00000592, 0.00001725, 0.00000296, -0.00002580, -0.00000026, 0.00002570, -0.00000767, -0.00002253, 0.00001511, 0.00001397, -0.00002446, -0.00000765, 0.00002786, 0.00000145, -0.00002444, 0.00000662, 0.00001799, -0.00001578, -0.00001577, 0.00001276, 0.00000060, -0.00001914, 0.00000567, 0.00001948, -0.00001017, -0.00001339, 0.00002125, 0.00001656, -0.00001915, -0.00000876, 0.00002175, 0.00000251, -0.00002495, -0.00000225, 0.00001935, -0.00000390, -0.00002021, -0.00000302, 0.00000592, -0.00000133, -0.00000101, -0.00000260, -0.00001009, -0.00000295, 0.00000639, -0.00000354, -0.00001032, -0.00000187, -0.00000335, -0.00001093, -0.00000103, 0.00000796, -0.00000406, -0.00001375, -0.00000673, -0.00000175, -0.00000955, -0.00002476, -0.00003564, -0.00002456, 0.00000264, 0.00001636, 0.00001790, 0.00003016, 0.00003997, 0.00002681, 0.00001031, 0.00000877, 0.00001058, 0.00001161, 0.00001753, 0.00001887, 0.00001171, 0.00001541, 0.00004396, 0.00007710, 0.00006862, },
		{ 0.00008901, 0.00005968, -0.00002639, -0.00007775, -0.00004964, -0.00000924, -0.00000793, -0.00001241, -0.00000332, -0.00000511, -0.00001341, -0.00000874, -0.00000432, -0.00000904, -0.00000800, -0.00000030, 0.00000483, 0.00000703, 0.00000171, -0.00000891, -0.00000761, 0.00000060, -0.00000380, -0.00000843, 0.00000159, 0.00000607, -0.00000348, -0.00000608, -0.00000076, -0.00000241, -0.00000524, -0.00000120, 0.00000331, 0.00000445, 0.00000120, -0.00000404, -0.00000062, 0.00000551, -0.00000237, -0.00000923, 0.00000258, 0.00000870, -0.00000463, -0.00001141, -0.00000496, 0.00000222, 0.00001407, 0.00001572, -0.00001009, -0.00002057, 0.00000677, 0.00001133, -0.00001436, -0.00000155, 0.00002404, -0.00000680, -0.00003071, 0.00001143, 0.00002755, -0.00001795, -0.00001995, 0.00002266, 0.00000768, -0.00002856, 0.00000035, 0.00002829, -0.00000602, -0.00002150, 0.00001164, 0.00001207, -0.00001814, -0.00000909, 0.00001129, -0.00000568, -0.00001313, 0.00001197, 0.00001102, -0.00001423, -0.00000194, 0.00002336, 0.00000324, -0.00001920, 0.00000498, 0.00001923, -0.00001100, -0.00002090, 0.00000875, 0.00001310, -0.00001303, -0.00001486, 0.00000173, 0.00000179, -0.00000115, 0.00000230, -0.00000581, -0.00001198, 0.00000117, 0.00000625, -0.00000701, -0.00000851, -0.00000043, -0.00000690, -0.00001163, 0.00000135, 0.00000716, -0.00000469, -0.00001220, -0.00000824, -0.00000379, -0.00000741, -0.00002421, -0.00003972, -0.00002809, 0.00000060, 0.00001617, 0.00002324, 0.00003571, 0.00003866, 0.00002573, 0.00001519, 0.00001154, 0.00000854, 0.00001243, 0.00002182, 0.00002099, 0.00001157, 0.00001860, 0.00005558, },
		{ 0.00002433, 0.00007008, 0.00009630, 0.00003964, -0.00005612, -0.00008171, -0.00003414, -0.00000480, -0.00001310, -0.00001115, -0.00000108, -0.00000743, -0.00001285, -0.00000672, -0.00000611, -0.00000916, -0.00000555, -0.00000138, 0.00000355, 0.00000919, 0.00000167, -0.00001081, -0.00000636, 0.00000107, -0.00000607, -0.00000783, 0.00000324, 0.00000437, -0.00000387, -0.00000439, -0.00000251, -0.00000405, -0.00000229, -0.00000004, 0.00000046, 0.00000406, 0.00000360, -0.00000254, -0.00000029, 0.00000375, -0.00000444, -0.00000725, 0.00000370, 0.00000430, -0.00000606, -0.00000782, -0.00000493, 0.00000353, 0.00002069, 0.00001167, -0.00002329, -0.00001749, 0.00001952, 0.00000712, -0.00002204, 0.00000705, 0.00002679, -0.00001918, -0.00002990, 0.00002406, 0.00002377, -0.00002731, -0.00001316, 0.00002753, -0.00000085, -0.00002906, 0.00000861, 0.00002530, -0.00001207, -0.00001634, 0.00001397, 0.00000552, -0.00001734, -0.00000357, 0.00000709, -0.00000872, -0.00000501, 0.00001337, 0.00000144, -0.00001218, 0.00000928, 0.00001847, -0.00000851, -0.00001214, 0.00001556, 0.00000993, -0.00002012, -0.00001110, 0.00001461, 0.00000261, -0.00001724, -0.00000723, 0.00000283, -0.00000219, 0.00000117, 0.00000360, -0.00000961, -0.00001106, 0.00000450, 0.00000365, -0.00000881, -0.00000554, -0.00000111, -0.00001020, -0.00001033, 0.00000281, 0.00000533, -0.00000436, -0.00001092, -0.00001037, -0.00000482, -0.00000551, -0.00002467, -0.00004249, -0.00003101, -0.00000339, 0.00001577, 0.00002988, 0.00004062, 0.00003721, 0.00002613, 0.00001935, 0.00001293, 0.00000813, 0.00001474, 0.00002518, 0.00002254, 0.00001224, },
		{ 0.00002378, 0.00001387, 0.00003395, 0.00008617, 0.00009428, 0.00000932, -0.00008013, -0.00007372, -0.00001960, -0.00000675, -0.00001668, -0.00000764, -0.00000148, -0.00000971, -0.00001002, -0.00000600, -0.00000880, -0.00000746, -0.00000342, -0.00000375, 0.00000328, 0.00001135, 0.00000043, -0.00001167, -0.00000454, -0.00000003, -0.00000781, -0.00000578, 0.00000345, 0.00000196, -0.00000256, -0.00000314, -0.00000565, -0.00000418, 0.00000154, -0.00000066, -0.00000268, 0.00000503, 0.00000619, -0.00000135, -0.00000058, 0.00000105, -0.00000532, -0.00000399, 0.00000267, -0.00000111, -0.00000480, -0.00000385, -0.00000629, 0.00000650, 0.00002652, 0.00000286, -0.00003451, -0.00000815, 0.00002968, -0.00000204, -0.00002580, 0.00001791, 0.00002405, -0.00003128, -0.00002304, 0.00003458, 0.00001500, -0.00003330, -0.00000362, 0.00002840, -0.00000974, -0.00002578, 0.00001535, 0.00001965, -0.00001554, -0.00001032, 0.00001337, -0.00000008, -0.00001412, -0.00000052, 0.00000186, -0.00000775, 0.00000265, 0.00000977, -0.00000606, -0.00000473, 0.00001669, 0.00000821, -0.00001508, -0.00000052, 0.00001972, -0.00000281, -0.00002222, 0.00000068, 0.00001367, -0.00000845, -0.00001591, -0.00000017, 0.00000079, -0.00000439, 0.00000426, 0.00000252, -0.00001228, -0.00000803, 0.00000563, -0.00000007, -0.00000837, -0.00000290, -0.00000357, -0.00001206, -0.00000790, 0.00000283, 0.00000337, -0.00000331, -0.00001044, -0.00001238, -0.00000492, -0.00000456, -0.00002546, -0.00004364, -0.00003423, -0.00000900, 0.00001581, 0.00003684, 0.00004478, 0.00003680, 0.00002746, 0.00002190, 0.00001398, 0.00000977, 0.00001767, 0.00002773, },
		{ 0.00002073, 0.00002978, 0.00002450, 0.00001741, 0.00004877, 0.00010028, 0.00007884, -0.00002677, -0.00009226, -0.00005700, -0.00001094, -0.00001266, -0.00001640, -0.00000425, -0.00000474, -0.00001001, -0.00000588, -0.00000749, -0.00001101, -0.00000412, -0.00000269, -0.00000652, 0.00000431, 0.00001249, -0.00000143, -0.00001086, -0.00000326, -0.00000256, -0.00000792, -0.00000313, 0.00000160, 0.00000003, 0.00000034, -0.00000338, -0.00000933, -0.00000224, 0.00000503, -0.00000310, -0.00000505, 0.00000708, 0.00000845, -0.00000033, -0.00000172, -0.00000228, -0.00000431, -0.00000032, -0.00000119, -0.00000617, -0.00000056, -0.00000069, -0.00000857, 0.00001135, 0.00002989, -0.00000978, -0.00004108, 0.00000578, 0.00003468, -0.00001390, -0.00002422, 0.00002824, 0.00001589, -0.00004010, -0.00001142, 0.00004041, 0.00000329, -0.00003445, 0.00000641, 0.00002505, -0.00001698, -0.00001954, 0.00001924, 0.00001274, -0.00001607, -0.00000493, 0.00001045, -0.00000357, -0.00000985, -0.00000036, -0.00000243, -0.00000334, 0.00000743, 0.00000271, -0.00000896, 0.00000557, 0.00001796, -0.00000406, -0.00001467, 0.00001164, 0.00001645, -0.00001460, -0.00001720, 0.00001024, 0.00000666, -0.00001662, -0.00001037, 0.00000417, -0.00000278, -0.00000422, 0.00000651, -0.00000019, -0.00001267, -0.00000441, 0.00000418, -0.00000332, -0.00000611, -0.00000174, -0.00000676, -0.00001200, -0.00000548, 0.00000158, 0.00000196, -0.00000217, -0.00001076, -0.00001363, -0.00000466, -0.00000481, -0.00002576, -0.00004350, -0.00003835, -0.00001543, 0.00001625, 0.00004317, 0.00004882, 0.00003811, 0.00002870, 0.00002280, 0.00001574, 0.00001312, },
		{ 0.00001875, 0.00001754, 0.00002401, 0.00003133, 0.00002453, 0.00002482, 0.00006870, 0.00010679, 0.00004856, -0.00006109, -0.00008931, -0.00003765, -0.00001046, -0.00001849, -0.00001220, -0.00000353, -0.00000942, -0.00000720, -0.00000225, -0.00001100, -0.00001140, -0.00000025, -0.00000382, -0.00000847, 0.00000609, 0.00001188, -0.00000276, -0.00000850, -0.00000365, -0.00000559, -0.00000580, -0.00000124, -0.00000211, -0.00000021, 0.00000388, -0.00000564, -0.00001223, 0.00000152, 0.00000694, -0.00000660, -0.00000587, 0.00000947, 0.00001019, 0.00000069, -0.00000405, -0.00000564, -0.00000098, 0.00000250, -0.00000791, -0.00000930, 0.00000623, 0.00000060, -0.00001094, 0.00001767, 0.00002940, -0.00002425, -0.00004125, 0.00002143, 0.00003327, -0.00002546, -0.00001746, 0.00003527, 0.00000397, -0.00004344, 0.00000234, 0.00004028, -0.00000860, -0.00003061, 0.00001462, 0.00001843, -0.00002106, -0.00001192, 0.00001975, 0.00000615, -0.00001408, -0.00000132, 0.00000639, -0.00000456, -0.00000609, -0.00000246, -0.00000416, 0.00000278, 0.00000791, -0.00000508, -0.00000627, 0.00001525, 0.00001277, -0.00001435, -0.00000784, 0.00002025, 0.00000729, -0.00002160, -0.00000738, 0.00001440, -0.00000370, -0.00001978, -0.00000316, 0.00000502, -0.00000597, -0.00000227, 0.00000686, -0.00000310, -0.00001065, -0.00000180, 0.00000080, -0.00000496, -0.00000322, -0.00000242, -0.00000942, -0.00001035, -0.00000395, -0.00000028, 0.00000128, -0.00000157, -0.00001144, -0.00001386, -0.00000474, -0.00000594, -0.00002505, -0.00004279, -0.00004336, -0.00002193, 0.00001646, 0.00004839, 0.00005370, 0.00004098, 0.00002901, 0.00002275, },
		{ 0.00002820, 0.00002272, 0.00002280, 0.00002266, 0.00002787, 0.00003177, 0.00002452, 0.00003862, 0.00009099, 0.00009946, 0.00000627, -0.00008517, -0.00007255, -0.00002245, -0.00001694, -0.00002050, -0.00000626, -0.00000683, -0.00001304, -0.00000164, -0.00000092, -0.00001516, -0.00000941, 0.00000261, -0.00000625, -0.00000865, 0.00000741, 0.00000959, -0.00000233, -0.00000559, -0.00000630, -0.00000763, -0.00000178, -0.00000145, -0.00000657, 0.00000183, 0.00000665, -0.00000957, -0.00001308, 0.00000599, 0.00000655, -0.00000980, -0.00000502, 0.00001129, 0.00001170, 0.00000171, -0.00000790, -0.00000813, 0.00000462, 0.00000303, -0.00001676, -0.00000908, 0.00001447, -0.00000059, -0.00001244, 0.00002444, 0.00002444, -0.00003791, -0.00003470, 0.00003554, 0.00002589, -0.00003391, -0.00000714, 0.00003718, -0.00000899, -0.00004053, 0.00001513, 0.00003448, -0.00001800, -0.00002299, 0.00001932, 0.00001031, -0.00002143, -0.00000470, 0.00001727, 0.00000120, -0.00001064, 0.00000005, 0.00000257, -0.00000352, -0.00000401, -0.00000540, -0.00000271, 0.00000830, 0.00000417, -0.00001065, 0.00000107, 0.00002100, 0.00000296, -0.00001946, 0.00000269, 0.00002256, -0.00000429, -0.00002182, 0.00000344, 0.00001208, -0.00001384, -0.00001777, 0.00000308, 0.00000308, -0.00000742, 0.00000015, 0.00000524, -0.00000480, -0.00000710, -0.00000121, -0.00000315, -0.00000471, -0.00000099, -0.00000444, -0.00001064, -0.00000798, -0.00000360, -0.00000199, 0.00000102, -0.00000179, -0.00001188, -0.00001332, -0.00000558, -0.00000731, -0.00002337, -0.00004216, -0.00004872, -0.00002821, 0.00001560, 0.00005276, 0.00006009, 0.00004468, },
		{ 0.00006798, 0.00004839, 0.00002687, 0.00002326, 0.00002730, 0.00002872, 0.00003229, 0.00003024, 0.00002660, 0.00006045, 0.00010983, 0.00007391, -0.00004072, -0.00009255, -0.00004768, -0.00001627, -0.00002625, -0.00001700, -0.00000194, -0.00001344, -0.00001334, 0.00000482, -0.00000261, -0.00001801, -0.00000569, 0.00000318, -0.00000858, -0.00000698, 0.00000692, 0.00000664, 0.00000048, -0.00000367, -0.00001079, -0.00000732, 0.00000282, -0.00000436, -0.00001013, 0.00000576, 0.00000731, -0.00001387, -0.00001132, 0.00000948, 0.00000410, -0.00001131, -0.00000324, 0.00001189, 0.00001361, 0.00000236, -0.00001334, -0.00000864, 0.00001176, 0.00000009, -0.00002626, -0.00000465, 0.00002258, -0.00000430, -0.00001236, 0.00003031, 0.00001538, -0.00004812, -0.00002265, 0.00004528, 0.00001442, -0.00003741, 0.00000415, 0.00003357, -0.00002000, -0.00003230, 0.00002421, 0.00002474, -0.00002307, -0.00001370, 0.00001990, 0.00000268, -0.00001851, 0.00000062, 0.00001288, -0.00000150, -0.00000703, -0.00000051, 0.00000002, -0.00000153, -0.00000400, -0.00000752, 0.00000136, 0.00001113, -0.00000220, -0.00001183, 0.00001049, 0.00002087, -0.00000810, -0.00001803, 0.00001301, 0.00001813, -0.00001417, -0.00001571, 0.00001136, 0.00000448, -0.00002067, -0.00001212, 0.00000655, -0.00000008, -0.00000683, 0.00000172, 0.00000247, -0.00000451, -0.00000343, -0.00000270, -0.00000634, -0.00000322, -0.00000024, -0.00000676, -0.00001022, -0.00000584, -0.00000414, -0.00000312, 0.00000066, -0.00000270, -0.00001169, -0.00001254, -0.00000717, -0.00000836, -0.00002117, -0.00004185, -0.00005377, -0.00003459, 0.00001314, 0.00005696, },
		{ 0.00000916, 0.00006161, 0.00007666, 0.00005186, 0.00002602, 0.00002424, 0.00003185, 0.00003632, 0.00003642, 0.00002658, 0.00003425, 0.00008938, 0.00011729, 0.00003000, -0.00008222, -0.00008130, -0.00002289, -0.00002015, -0.00003320, -0.00000926, -0.00000207, -0.00002076, -0.00000951, 0.00000961, -0.00000650, -0.00001798, -0.00000198, 0.00000127, -0.00000918, -0.00000447, 0.00000386, 0.00000467, 0.00000525, -0.00000416, -0.00001568, -0.00000410, 0.00000614, -0.00000943, -0.00001127, 0.00001014, 0.00000530, -0.00001670, -0.00000749, 0.00001040, 0.00000089, -0.00001023, -0.00000191, 0.00001126, 0.00001660, 0.00000189, -0.00002001, -0.00000620, 0.00001910, -0.00000695, -0.00003443, 0.00000402, 0.00002887, -0.00000996, -0.00001037, 0.00003395, 0.00000360, -0.00005295, -0.00000751, 0.00004900, 0.00000161, -0.00003557, 0.00001370, 0.00002558, -0.00002671, -0.00002095, 0.00002798, 0.00001365, -0.00002330, -0.00000504, 0.00001690, -0.00000279, -0.00001356, 0.00000325, 0.00000798, -0.00000203, -0.00000431, -0.00000215, -0.00000084, 0.00000018, -0.00000553, -0.00000751, 0.00000644, 0.00001019, -0.00000865, -0.00000799, 0.00001868, 0.00001494, -0.00001674, -0.00001099, 0.00001949, 0.00000883, -0.00001907, -0.00000588, 0.00001382, -0.00000557, -0.00002255, -0.00000531, 0.00000678, -0.00000279, -0.00000483, 0.00000173, -0.00000017, -0.00000240, -0.00000092, -0.00000546, -0.00000798, -0.00000157, -0.00000105, -0.00000835, -0.00000860, -0.00000455, -0.00000500, -0.00000367, -0.00000020, -0.00000384, -0.00001086, -0.00001201, -0.00000906, -0.00000889, -0.00001907, -0.00004162, -0.00005811, -0.00004169, },
		{ -0.00006187, -0.00004999, 0.00000431, 0.00006675, 0.00008521, 0.00005561, 0.00002638, 0.00002497, 0.00003678, 0.00004582, 0.00003868, 0.00002236, 0.00005108, 0.00012061, 0.00010571, -0.00002675, -0.00010821, -0.00005516, -0.00000597, -0.00003092, -0.00003394, -0.00000090, -0.00000739, -0.00002545, -0.00000279, 0.00001070, -0.00001041, -0.00001477, -0.00000034, -0.00000205, -0.00000703, -0.00000288, -0.00000138, 0.00000518, 0.00001047, -0.00000759, -0.00001900, 0.00000131, 0.00000656, -0.00001497, -0.00000941, 0.00001296, 0.00000135, -0.00001644, -0.00000331, 0.00000797, -0.00000105, -0.00000674, -0.00000251, 0.00001022, 0.00002098, -0.00000064, -0.00002695, -0.00000025, 0.00002482, -0.00001779, -0.00003920, 0.00001600, 0.00003191, -0.00001659, -0.00000670, 0.00003446, -0.00000884, -0.00005164, 0.00000773, 0.00004654, -0.00000974, -0.00002942, 0.00001950, 0.00001546, -0.00002806, -0.00000932, 0.00002636, 0.00000382, -0.00001949, 0.00000118, 0.00001174, -0.00000528, -0.00000816, 0.00000324, 0.00000379, -0.00000112, -0.00000298, -0.00000383, -0.00000034, 0.00000071, -0.00000739, -0.00000501, 0.00001051, 0.00000580, -0.00001255, -0.00000033, 0.00002275, 0.00000524, -0.00002023, -0.00000117, 0.00002001, -0.00000188, -0.00001766, 0.00000398, 0.00001030, -0.00001472, -0.00001970, 0.00000029, 0.00000458, -0.00000401, -0.00000256, 0.00000028, -0.00000162, 0.00000060, -0.00000022, -0.00000828, -0.00000809, -0.00000077, -0.00000281, -0.00000863, -0.00000657, -0.00000419, -0.00000566, -0.00000395, -0.00000161, -0.00000479, -0.00000974, -0.00001190, -0.00001076, -0.00000914, -0.00001741, -0.00004092, },
		{ -0.00001606, -0.00003939, -0.00006563, -0.00005944, -0.00000068, 0.00007163, 0.00009309, 0.00006073, 0.00002793, 0.00002487, 0.00004310, 0.00005651, 0.00003768, 0.00002132, 0.00007877, 0.00014557, 0.00007084, -0.00008576, -0.00011232, -0.00002261, -0.00000133, -0.00004256, -0.00002778, 0.00000374, -0.00001601, -0.00002505, 0.00000392, 0.00000771, -0.00001178, -0.00000970, -0.00000211, -0.00000473, -0.00000239, -0.00000391, -0.00000722, 0.00000879, 0.00001406, -0.00001314, -0.00001906, 0.00000696, 0.00000353, -0.00001870, -0.00000537, 0.00001246, -0.00000261, -0.00001245, -0.00000117, 0.00000278, 0.00000027, -0.00000225, -0.00000604, 0.00001018, 0.00002638, -0.00000605, -0.00003269, 0.00000898, 0.00002706, -0.00003111, -0.00003889, 0.00002955, 0.00003085, -0.00002297, -0.00000202, 0.00003165, -0.00001973, -0.00004476, 0.00002032, 0.00003917, -0.00001758, -0.00002099, 0.00002076, 0.00000582, -0.00002448, -0.00000002, 0.00002062, -0.00000284, -0.00001337, 0.00000407, 0.00000622, -0.00000488, -0.00000374, 0.00000133, 0.00000104, 0.00000027, -0.00000293, -0.00000474, 0.00000071, -0.00000017, -0.00000826, -0.00000068, 0.00001187, -0.00000038, -0.00001218, 0.00000859, 0.00002123, -0.00000500, -0.00001773, 0.00000784, 0.00001462, -0.00001030, -0.00001089, 0.00001046, 0.00000234, -0.00002030, -0.00001378, 0.00000322, 0.00000139, -0.00000357, -0.00000105, -0.00000181, -0.00000145, 0.00000328, -0.00000110, -0.00001014, -0.00000737, -0.00000134, -0.00000462, -0.00000761, -0.00000484, -0.00000444, -0.00000591, -0.00000435, -0.00000330, -0.00000539, -0.00000876, -0.00001197, -0.00001196, -0.00000961, },
		{ -0.00001280, -0.00001060, -0.00001457, -0.00003715, -0.00006996, -0.00006931, -0.00000557, 0.00007511, 0.00010063, 0.00006824, 0.00002987, 0.00002415, 0.00005185, 0.00006640, 0.00003359, 0.00002867, 0.00011500, 0.00015384, 0.00001447, -0.00013402, -0.00009449, 0.00000600, -0.00000829, -0.00004888, -0.00001757, 0.00000179, -0.00002403, -0.00001933, 0.00000729, 0.00000232, -0.00000891, -0.00000528, -0.00000712, -0.00000472, 0.00000307, -0.00000828, -0.00001142, 0.00001461, 0.00001427, -0.00001868, -0.00001542, 0.00001034, -0.00000193, -0.00001870, -0.00000133, 0.00000809, -0.00000395, -0.00000569, -0.00000328, -0.00000312, 0.00000603, 0.00000098, -0.00001238, 0.00001268, 0.00003162, -0.00001455, -0.00003556, 0.00002034, 0.00002441, -0.00004476, -0.00003272, 0.00004243, 0.00002563, -0.00002804, 0.00000275, 0.00002606, -0.00002722, -0.00003400, 0.00002837, 0.00002905, -0.00002101, -0.00001261, 0.00001802, -0.00000118, -0.00001763, 0.00000531, 0.00001290, -0.00000555, -0.00000696, 0.00000385, 0.00000188, -0.00000248, -0.00000115, -0.00000140, -0.00000020, 0.00000131, -0.00000359, -0.00000460, 0.00000140, -0.00000193, -0.00000721, 0.00000397, 0.00000986, -0.00000597, -0.00000734, 0.00001571, 0.00001457, -0.00001247, -0.00001049, 0.00001291, 0.00000546, -0.00001378, -0.00000152, 0.00001160, -0.00000719, -0.00002117, -0.00000714, 0.00000328, -0.00000137, -0.00000210, -0.00000075, -0.00000352, -0.00000003, 0.00000472, -0.00000275, -0.00001061, -0.00000674, -0.00000309, -0.00000572, -0.00000580, -0.00000375, -0.00000486, -0.00000591, -0.00000502, -0.00000488, -0.00000582, -0.00000817, -0.00001181, },
		{ -0.00000786, -0.00001114, -0.00001370, -0.00001204, -0.00001251, -0.00003486, -0.00007494, -0.00007859, -0.00001094, 0.00007623, 0.00010895, 0.00007835, 0.00003121, 0.00002412, 0.00006309, 0.00007283, 0.00002931, 0.00004906, 0.00015233, 0.00013644, -0.00005458, -0.00016012, -0.00006144, 0.00002213, -0.00002139, -0.00004624, -0.00000856, -0.00000661, -0.00002725, -0.00001064, 0.00000519, -0.00000235, -0.00000197, -0.00000405, -0.00001353, -0.00000102, 0.00000693, -0.00001512, -0.00001211, 0.00002042, 0.00001069, -0.00002155, -0.00000935, 0.00000939, -0.00000729, -0.00001436, -0.00000016, 0.00000100, -0.00000041, 0.00000134, -0.00001066, -0.00000675, 0.00001584, 0.00000052, -0.00002016, 0.00001872, 0.00003493, -0.00002547, -0.00003414, 0.00003188, 0.00001637, -0.00005611, -0.00002105, 0.00005236, 0.00001698, -0.00003103, 0.00000673, 0.00001886, -0.00003031, -0.00002170, 0.00003121, 0.00001860, -0.00002043, -0.00000615, 0.00001287, -0.00000437, -0.00000982, 0.00000630, 0.00000549, -0.00000475, -0.00000195, 0.00000153, -0.00000040, 0.00000065, -0.00000054, -0.00000393, -0.00000034, 0.00000157, -0.00000422, -0.00000367, 0.00000112, -0.00000355, -0.00000415, 0.00000721, 0.00000507, -0.00000872, 0.00000055, 0.00001860, 0.00000493, -0.00001497, -0.00000128, 0.00001254, -0.00000419, -0.00001161, 0.00000704, 0.00000741, -0.00001526, -0.00001784, -0.00000183, 0.00000127, -0.00000288, -0.00000051, -0.00000142, -0.00000416, 0.00000177, 0.00000466, -0.00000418, -0.00000994, -0.00000690, -0.00000539, -0.00000583, -0.00000381, -0.00000321, -0.00000518, -0.00000594, -0.00000581, -0.00000613, -0.00000644, },
		{ -0.00000715, -0.00000747, -0.00000746, -0.00001006, -0.00001503, -0.00001339, -0.00000990, -0.00003334, -0.00007989, -0.00008661, -0.00001809, 0.00007496, 0.00011939, 0.00009019, 0.00003169, 0.00002668, 0.00007514, 0.00007398, 0.00003057, 0.00008388, 0.00017897, 0.00008973, -0.00012198, -0.00015831, -0.00002471, 0.00002227, -0.00003266, -0.00003527, -0.00000594, -0.00001810, -0.00002324, -0.00000308, -0.00000211, -0.00000311, 0.00000681, -0.00000735, -0.00001851, 0.00000555, 0.00000709, -0.00002206, -0.00000882, 0.00002338, 0.00000474, -0.00001966, -0.00000367, 0.00000363, -0.00000938, -0.00000706, -0.00000424, -0.00000599, 0.00000887, 0.00000520, -0.00002243, -0.00000514, 0.00002761, -0.00000530, -0.00002701, 0.00002822, 0.00003451, -0.00003721, -0.00002780, 0.00004120, 0.00000370, -0.00006267, -0.00000544, 0.00005752, 0.00000624, -0.00003166, 0.00000930, 0.00001151, -0.00002897, -0.00001018, 0.00002940, 0.00000979, -0.00001711, -0.00000259, 0.00000729, -0.00000383, -0.00000321, 0.00000381, 0.00000012, -0.00000172, 0.00000073, -0.00000144, -0.00000052, 0.00000332, -0.00000145, -0.00000563, -0.00000006, 0.00000110, -0.00000429, -0.00000258, -0.00000015, -0.00000398, 0.00000013, 0.00000774, -0.00000077, -0.00000737, 0.00000895, 0.00001626, -0.00000461, -0.00001213, 0.00000662, 0.00000723, -0.00001111, -0.00000515, 0.00001189, -0.00000034, -0.00001969, -0.00001201, 0.00000101, -0.00000153, -0.00000310, 0.00000049, -0.00000237, -0.00000360, 0.00000303, 0.00000349, -0.00000472, -0.00000882, -0.00000803, -0.00000752, -0.00000514, -0.00000210, -0.00000295, -0.00000543, -0.00000618, -0.00000642, },
		{ -0.00000654, -0.00000662, -0.00000823, -0.00000884, -0.00000665, -0.00000905, -0.00001676, -0.00001399, -0.00000743, -0.00003304, -0.00008356, -0.00009361, -0.00002839, 0.00007234, 0.00013262, 0.00010220, 0.00003242, 0.00003327, 0.00008466, 0.00007069, 0.00004447, 0.00012881, 0.00018181, 0.00001813, -0.00017178, -0.00013096, 0.00000317, 0.00000931, -0.00003505, -0.00002078, -0.00001237, -0.00002733, -0.00001264, -0.00000069, -0.00001183, 0.00000169, 0.00001402, -0.00001438, -0.00001954, 0.00001247, 0.00000293, -0.00002606, -0.00000304, 0.00002134, -0.00000052, -0.00001265, -0.00000186, -0.00000532, -0.00000558, 0.00000008, -0.00001438, -0.00000916, 0.00002262, 0.00000275, -0.00003580, 0.00000347, 0.00003799, -0.00001659, -0.00003024, 0.00003978, 0.00002917, -0.00004749, -0.00001702, 0.00004603, -0.00001167, -0.00006273, 0.00001163, 0.00005696, -0.00000491, -0.00003009, 0.00001023, 0.00000540, -0.00002408, -0.00000125, 0.00002434, 0.00000374, -0.00001276, -0.00000183, 0.00000301, -0.00000069, 0.00000073, -0.00000053, -0.00000240, 0.00000192, 0.00000109, -0.00000381, 0.00000091, 0.00000477, -0.00000309, -0.00000634, 0.00000007, 0.00000031, -0.00000366, -0.00000200, -0.00000184, -0.00000264, 0.00000419, 0.00000528, -0.00000549, -0.00000212, 0.00001511, 0.00000947, -0.00001118, -0.00000540, 0.00001068, -0.00000077, -0.00001333, 0.00000286, 0.00001168, -0.00000888, -0.00001974, -0.00000580, 0.00000131, -0.00000396, -0.00000251, 0.00000067, -0.00000292, -0.00000234, 0.00000324, 0.00000198, -0.00000428, -0.00000795, -0.00000978, -0.00000904, -0.00000412, -0.00000075, -0.00000272, -0.00000580, },
		{ -0.00000251, -0.00000647, -0.00000673, -0.00000646, -0.00000966, -0.00001014, -0.00000548, -0.00000866, -0.00001831, -0.00001360, -0.00000618, -0.00003356, -0.00008481, -0.00010082, -0.00004244, 0.00007011, 0.00014799, 0.00011315, 0.00003591, 0.00004347, 0.00008788, 0.00006762, 0.00007654, 0.00017308, 0.00015081, -0.00006554, -0.00019146, -0.00008832, 0.00001325, -0.00000868, -0.00002548, -0.00000969, -0.00002648, -0.00002937, 0.00000077, -0.00000547, -0.00001973, 0.00001114, 0.00001653, -0.00002228, -0.00001575, 0.00001638, -0.00000397, -0.00002476, 0.00000197, 0.00001396, -0.00000143, -0.00000244, -0.00000665, -0.00001390, 0.00000489, 0.00000302, -0.00002906, -0.00000526, 0.00003751, -0.00000765, -0.00004664, 0.00001886, 0.00004333, -0.00003164, -0.00002779, 0.00005095, 0.00001885, -0.00005399, -0.00000351, 0.00004496, -0.00002691, -0.00005584, 0.00002728, 0.00005080, -0.00001482, -0.00002687, 0.00000975, 0.00000149, -0.00001716, 0.00000415, 0.00001782, 0.00000062, -0.00000884, -0.00000293, 0.00000103, 0.00000335, 0.00000160, -0.00000492, -0.00000228, 0.00000477, -0.00000015, -0.00000478, 0.00000294, 0.00000480, -0.00000468, -0.00000632, -0.00000024, -0.00000031, -0.00000262, -0.00000227, -0.00000307, 0.00000030, 0.00000658, 0.00000068, -0.00000721, 0.00000543, 0.00001703, 0.00000046, -0.00001308, 0.00000254, 0.00000995, -0.00000848, -0.00001069, 0.00000936, 0.00000687, -0.00001556, -0.00001612, -0.00000096, -0.00000017, -0.00000541, -0.00000176, 0.00000030, -0.00000273, -0.00000109, 0.00000246, 0.00000083, -0.00000328, -0.00000769, -0.00001159, -0.00000991, -0.00000316, 0.00000039, },
		{ -0.00000241, 0.00000151, -0.00000251, -0.00000735, -0.00000649, -0.00000631, -0.00001140, -0.00001089, -0.00000445, -0.00000909, -0.00001878, -0.00001269, -0.00000709, -0.00003363, -0.00008330, -0.00011004, -0.00005948, 0.00006972, 0.00016350, 0.00012321, 0.00004504, 0.00005441, 0.00008286, 0.00007286, 0.00012733, 0.00020121, 0.00008355, -0.00014275, -0.00017645, -0.00004537, 0.00000346, -0.00002169, -0.00000638, -0.00000796, -0.00004310, -0.00002205, 0.00001151, -0.00001623, -0.00002196, 0.00002180, 0.00001317, -0.00002717, -0.00000872, 0.00001464, -0.00001015, -0.00001778, 0.00000234, 0.00000340, 0.00000482, 0.00000701, -0.00001872, -0.00001769, 0.00002025, -0.00000167, -0.00004454, 0.00000720, 0.00004894, -0.00002524, -0.00005077, 0.00003849, 0.00004091, -0.00004722, -0.00001901, 0.00005887, 0.00000489, -0.00005502, 0.00001016, 0.00003785, -0.00003898, -0.00004299, 0.00003885, 0.00004020, -0.00002221, -0.00002272, 0.00000839, 0.00000014, -0.00000985, 0.00000602, 0.00001145, -0.00000020, -0.00000625, -0.00000459, 0.00000138, 0.00000667, -0.00000003, -0.00000796, -0.00000045, 0.00000605, -0.00000192, -0.00000423, 0.00000464, 0.00000372, -0.00000574, -0.00000601, -0.00000093, -0.00000046, -0.00000169, -0.00000324, -0.00000304, 0.00000395, 0.00000639, -0.00000434, -0.00000510, 0.00001283, 0.00001417, -0.00000794, -0.00001034, 0.00000886, 0.00000528, -0.00001325, -0.00000480, 0.00001206, -0.00000066, -0.00001867, -0.00001050, 0.00000163, -0.00000239, -0.00000584, -0.00000126, -0.00000014, -0.00000197, -0.00000041, 0.00000120, 0.00000032, -0.00000230, -0.00000795, -0.00001298, -0.00001043, },
		{ -0.00001383, -0.00001090, -0.00000177, 0.00000257, -0.00000294, -0.00000811, -0.00000582, -0.00000658, -0.00001299, -0.00001093, -0.00000436, -0.00000989, -0.00001748, -0.00001247, -0.00001027, -0.00003163, -0.00007998, -0.00012264, -0.00007749, 0.00007131, 0.00017676, 0.00013460, 0.00006143, 0.00006128, 0.00007160, 0.00009549, 0.00018974, 0.00019724, -0.00001211, -0.00019448, -0.00013229, -0.00001673, -0.00002059, -0.00002199, 0.00001517, -0.00001772, -0.00005529, -0.00000730, 0.00001485, -0.00002869, -0.00001707, 0.00002902, 0.00000564, -0.00002590, -0.00000223, 0.00000673, -0.00001140, -0.00000739, -0.00000487, -0.00000610, 0.00001875, 0.00001094, -0.00003596, -0.00001300, 0.00003639, -0.00001543, -0.00005589, 0.00002704, 0.00005249, -0.00004671, -0.00004536, 0.00005806, 0.00002995, -0.00005947, -0.00000507, 0.00006118, -0.00001024, -0.00005013, 0.00002119, 0.00002604, -0.00004539, -0.00002649, 0.00004449, 0.00002706, -0.00002640, -0.00001836, 0.00000685, 0.00000110, -0.00000359, 0.00000507, 0.00000633, 0.00000028, -0.00000511, -0.00000564, 0.00000332, 0.00000819, -0.00000296, -0.00000900, 0.00000186, 0.00000571, -0.00000324, -0.00000261, 0.00000543, 0.00000210, -0.00000613, -0.00000579, -0.00000169, -0.00000012, -0.00000132, -0.00000432, -0.00000146, 0.00000701, 0.00000365, -0.00000780, 0.00000030, 0.00001766, 0.00000755, -0.00001331, -0.00000451, 0.00001162, -0.00000116, -0.00001374, 0.00000175, 0.00001017, -0.00000836, -0.00001784, -0.00000473, 0.00000199, -0.00000443, -0.00000556, -0.00000105, -0.00000033, -0.00000110, -0.00000043, 0.00000005, 0.00000028, -0.00000176, -0.00000831, },
		{ -0.00000168, -0.00000837, -0.00001433, -0.00001145, -0.00000105, 0.00000332, -0.00000375, -0.00000832, -0.00000512, -0.00000741, -0.00001378, -0.00001067, -0.00000576, -0.00001003, -0.00001445, -0.00001435, -0.00001467, -0.00002645, -0.00007692, -0.00013864, -0.00009405, 0.00007312, 0.00018633, 0.00015106, 0.00008393, 0.00005940, 0.00006116, 0.00014168, 0.00024885, 0.00015036, -0.00011850, -0.00020738, -0.00007344, -0.00001141, -0.00004778, -0.00000751, 0.00003038, -0.00003588, -0.00005740, 0.00000926, 0.00000909, -0.00003715, -0.00000705, 0.00002898, -0.00000179, -0.00001771, -0.00000093, -0.00000495, -0.00000451, 0.00000188, -0.00002014, -0.00000950, 0.00003792, 0.00000546, -0.00005376, 0.00000159, 0.00004795, -0.00003681, -0.00005850, 0.00005039, 0.00004549, -0.00006696, -0.00003007, 0.00007270, 0.00001204, -0.00006509, 0.00001126, 0.00005679, -0.00002359, -0.00004030, 0.00002732, 0.00001208, -0.00004484, -0.00000941, 0.00004358, 0.00001357, -0.00002733, -0.00001431, 0.00000575, 0.00000363, 0.00000070, 0.00000243, 0.00000288, 0.00000109, -0.00000496, -0.00000542, 0.00000569, 0.00000761, -0.00000584, -0.00000816, 0.00000360, 0.00000429, -0.00000352, -0.00000062, 0.00000514, 0.00000051, -0.00000601, -0.00000582, -0.00000221, 0.00000039, -0.00000167, -0.00000477, 0.00000133, 0.00000828, -0.00000062, -0.00000824, 0.00000730, 0.00001837, -0.00000063, -0.00001442, 0.00000197, 0.00001033, -0.00000676, -0.00001027, 0.00000636, 0.00000448, -0.00001400, -0.00001390, -0.00000023, 0.00000074, -0.00000582, -0.00000490, -0.00000095, -0.00000026, -0.00000062, -0.00000085, -0.00000064, 0.00000025, },
		{ -0.00000092, -0.00000016, -0.00000174, -0.00000799, -0.00001476, -0.00001192, -0.00000028, 0.00000341, -0.00000455, -0.00000785, -0.00000498, -0.00000844, -0.00001325, -0.00001102, -0.00000859, -0.00000834, -0.00001077, -0.00001917, -0.00001836, -0.00001834, -0.00007643, -0.00015623, -0.00010764, 0.00007200, 0.00019302, 0.00017627, 0.00010811, 0.00004714, 0.00006249, 0.00021041, 0.00028478, 0.00006005, -0.00021223, -0.00017869, -0.00001870, -0.00002949, -0.00006592, 0.00001701, 0.00003262, -0.00005502, -0.00004783, 0.00002047, -0.00000321, -0.00003697, 0.00000301, 0.00002067, -0.00000374, -0.00000514, -0.00000842, -0.00001550, 0.00001108, 0.00000465, -0.00004079, -0.00000278, 0.00005725, -0.00001075, -0.00006630, 0.00002435, 0.00005009, -0.00006148, -0.00004981, 0.00007165, 0.00002809, -0.00008060, -0.00000736, 0.00007847, -0.00000916, -0.00006239, 0.00002638, 0.00004631, -0.00003254, -0.00002782, 0.00002753, -0.00000087, -0.00003762, 0.00000511, 0.00003683, 0.00000177, -0.00002553, -0.00001086, 0.00000550, 0.00000678, 0.00000273, -0.00000077, 0.00000098, 0.00000159, -0.00000506, -0.00000391, 0.00000737, 0.00000539, -0.00000765, -0.00000615, 0.00000419, 0.00000258, -0.00000270, 0.00000102, 0.00000400, -0.00000069, -0.00000569, -0.00000602, -0.00000232, 0.00000067, -0.00000247, -0.00000402, 0.00000440, 0.00000717, -0.00000475, -0.00000522, 0.00001359, 0.00001479, -0.00000785, -0.00001158, 0.00000674, 0.00000605, -0.00000948, -0.00000456, 0.00000732, -0.00000311, -0.00001631, -0.00000835, 0.00000231, -0.00000130, -0.00000646, -0.00000403, -0.00000076, -0.00000027, -0.00000069, -0.00000117, },
		{ -0.00000109, -0.00000104, -0.00000119, -0.00000101, -0.00000156, -0.00000737, -0.00001524, -0.00001202, 0.00000023, 0.00000274, -0.00000476, -0.00000698, -0.00000584, -0.00000888, -0.00001152, -0.00001298, -0.00001186, -0.00000423, -0.00000820, -0.00002649, -0.00001933, -0.00000922, -0.00007979, -0.00017234, -0.00011884, 0.00006477, 0.00020028, 0.00021165, 0.00012756, 0.00002847, 0.00008692, 0.00029084, 0.00027840, -0.00006139, -0.00027093, -0.00011800, 0.00001512, -0.00006194, -0.00006675, 0.00004169, 0.00002066, -0.00006630, -0.00003025, 0.00002071, -0.00001579, -0.00002691, 0.00000676, 0.00000687, 0.00000399, 0.00000644, -0.00002541, -0.00001908, 0.00003249, -0.00000330, -0.00006135, 0.00001527, 0.00007048, -0.00003560, -0.00006843, 0.00005052, 0.00004027, -0.00008345, -0.00003039, 0.00008516, 0.00000356, -0.00008360, 0.00001799, 0.00007360, -0.00002894, -0.00005187, 0.00003688, 0.00003193, -0.00003563, -0.00001557, 0.00002235, -0.00000994, -0.00002555, 0.00001461, 0.00002601, -0.00000691, -0.00002188, -0.00000801, 0.00000619, 0.00000958, 0.00000274, -0.00000364, 0.00000014, 0.00000160, -0.00000474, -0.00000165, 0.00000766, 0.00000243, -0.00000794, -0.00000389, 0.00000362, 0.00000131, -0.00000113, 0.00000188, 0.00000245, -0.00000139, -0.00000540, -0.00000616, -0.00000212, 0.00000041, -0.00000317, -0.00000198, 0.00000657, 0.00000401, -0.00000703, 0.00000048, 0.00001713, 0.00000812, -0.00001214, -0.00000638, 0.00000841, 0.00000082, -0.00000859, 0.00000105, 0.00000436, -0.00001036, -0.00001520, -0.00000279, 0.00000290, -0.00000345, -0.00000642, -0.00000294, -0.00000051, -0.00000073, },
		{ -0.00000051, -0.00000179, -0.00000135, -0.00000051, -0.00000184, -0.00000199, -0.00000098, -0.00000688, -0.00001556, -0.00001165, -0.00000005, 0.00000160, -0.00000393, -0.00000643, -0.00000764, -0.00000786, -0.00000944, -0.00001701, -0.00001405, 0.00000176, -0.00000843, -0.00003447, -0.00001666, -0.00000205, -0.00008628, -0.00018391, -0.00013058, 0.00005007, 0.00021309, 0.00025471, 0.00013686, 0.00001373, 0.00014101, 0.00036278, 0.00021830, -0.00019035, -0.00028061, -0.00004480, 0.00001921, -0.00009401, -0.00004964, 0.00005553, -0.00000018, -0.00006319, -0.00001269, 0.00000868, -0.00002097, -0.00001029, -0.00000079, -0.00000631, 0.00002235, 0.00001055, -0.00004883, -0.00001106, 0.00005388, -0.00002323, -0.00007518, 0.00004222, 0.00007222, -0.00006380, -0.00005758, 0.00007363, 0.00001932, -0.00009682, -0.00000404, 0.00008700, -0.00002257, -0.00007463, 0.00004029, 0.00005909, -0.00004296, -0.00003610, 0.00004063, 0.00001678, -0.00003297, -0.00000622, 0.00001375, -0.00001326, -0.00001151, 0.00001787, 0.00001351, -0.00001186, -0.00001732, -0.00000558, 0.00000760, 0.00001133, 0.00000135, -0.00000568, -0.00000016, 0.00000130, -0.00000367, 0.00000057, 0.00000645, -0.00000028, -0.00000686, -0.00000216, 0.00000233, 0.00000089, 0.00000063, 0.00000185, 0.00000092, -0.00000171, -0.00000523, -0.00000604, -0.00000189, -0.00000039, -0.00000316, 0.00000085, 0.00000695, -0.00000008, -0.00000637, 0.00000712, 0.00001680, 0.00000044, -0.00001271, -0.00000093, 0.00000691, -0.00000324, -0.00000476, 0.00000445, -0.00000151, -0.00001546, -0.00001153, 0.00000165, 0.00000200, -0.00000528, -0.00000575, -0.00000157, },
		{ -0.00000442, -0.00000009, -0.00000114, -0.00000320, -0.00000108, 0.00000001, -0.00000299, -0.00000262, -0.00000022, -0.00000671, -0.00001527, -0.00001109, -0.00000151, 0.00000071, -0.00000209, -0.00000699, -0.00000963, -0.00000498, -0.00000835, -0.00002256, -0.00001378, 0.00000784, -0.00001209, -0.00004038, -0.00001134, 0.00000038, -0.00009299, -0.00018959, -0.00014719, 0.00002973, 0.00023564, 0.00029895, 0.00013526, 0.00001760, 0.00022152, 0.00040090, 0.00010650, -0.00029764, -0.00024093, 0.00001757, -0.00000387, -0.00011090, -0.00002229, 0.00005110, -0.00001958, -0.00004475, -0.00000459, -0.00001142, -0.00001281, 0.00000569, -0.00002085, -0.00001143, 0.00004803, 0.00000202, -0.00007233, 0.00000994, 0.00006814, -0.00005242, -0.00007665, 0.00007228, 0.00005991, -0.00008837, -0.00003482, 0.00008749, -0.00000854, -0.00009774, 0.00002320, 0.00007635, -0.00004406, -0.00005561, 0.00005455, 0.00003846, -0.00004846, -0.00001893, 0.00003733, 0.00000395, -0.00002618, -0.00000142, 0.00000456, -0.00001054, 0.00000132, 0.00001513, 0.00000169, -0.00001327, -0.00001268, -0.00000331, 0.00000929, 0.00001167, -0.00000070, -0.00000676, -0.00000031, 0.00000107, -0.00000192, 0.00000202, 0.00000415, -0.00000203, -0.00000497, -0.00000133, 0.00000091, 0.00000133, 0.00000204, 0.00000115, -0.00000031, -0.00000186, -0.00000508, -0.00000562, -0.00000195, -0.00000142, -0.00000207, 0.00000356, 0.00000526, -0.00000355, -0.00000269, 0.00001263, 0.00001274, -0.00000607, -0.00001006, 0.00000294, 0.00000339, -0.00000471, 0.00000031, 0.00000456, -0.00000848, -0.00001756, -0.00000662, 0.00000440, 0.00000020, -0.00000655, },
		{ -0.00000198, -0.00000706, -0.00000244, 0.00000109, -0.00000257, -0.00000440, -0.00000028, -0.00000000, -0.00000431, -0.00000262, 0.00000023, -0.00000672, -0.00001401, -0.00001096, -0.00000409, 0.00000084, 0.00000010, -0.00000904, -0.00001064, -0.00000077, -0.00000940, -0.00002806, -0.00001068, 0.00001162, -0.00001806, -0.00004191, -0.00000630, -0.00000294, -0.00009580, -0.00019092, -0.00017237, 0.00000875, 0.00026861, 0.00033597, 0.00012961, 0.00005418, 0.00031272, 0.00038223, -0.00003917, -0.00035690, -0.00016642, 0.00004973, -0.00004087, -0.00010377, 0.00000187, 0.00002822, -0.00002609, -0.00001676, -0.00001297, -0.00003082, 0.00001007, 0.00001235, -0.00004967, -0.00000254, 0.00007402, -0.00002075, -0.00008815, 0.00004103, 0.00006928, -0.00008468, -0.00006323, 0.00009796, 0.00003495, -0.00010272, -0.00000485, 0.00008818, -0.00003684, -0.00008573, 0.00004484, 0.00005587, -0.00005591, -0.00003115, 0.00005801, 0.00001670, -0.00004508, -0.00000432, 0.00002851, -0.00000438, -0.00001780, -0.00000132, -0.00000240, -0.00000307, 0.00001030, 0.00000786, -0.00000760, -0.00001196, -0.00000843, -0.00000102, 0.00001070, 0.00001063, -0.00000273, -0.00000704, -0.00000048, 0.00000120, 0.00000009, 0.00000230, 0.00000146, -0.00000254, -0.00000298, -0.00000138, -0.00000009, 0.00000231, 0.00000278, 0.00000009, -0.00000118, -0.00000196, -0.00000480, -0.00000502, -0.00000246, -0.00000216, -0.00000000, 0.00000518, 0.00000203, -0.00000500, 0.00000303, 0.00001534, 0.00000625, -0.00000988, -0.00000567, 0.00000430, -0.00000041, -0.00000337, 0.00000460, 0.00000147, -0.00001468, -0.00001674, -0.00000174, 0.00000542, },
		{ 0.00000217, 0.00000494, -0.00000412, -0.00000657, -0.00000000, 0.00000143, -0.00000456, -0.00000485, 0.00000060, -0.00000081, -0.00000519, -0.00000214, -0.00000009, -0.00000635, -0.00001185, -0.00001201, -0.00000715, 0.00000241, 0.00000158, -0.00001215, -0.00000966, 0.00000339, -0.00001286, -0.00003142, -0.00000592, 0.00001119, -0.00002355, -0.00003847, -0.00000540, -0.00001035, -0.00009123, -0.00019225, -0.00020694, -0.00000656, 0.00030738, 0.00035956, 0.00013475, 0.00013056, 0.00038767, 0.00029443, -0.00018784, -0.00035325, -0.00008274, 0.00004304, -0.00007247, -0.00007362, 0.00000886, -0.00000512, -0.00001177, 0.00000976, -0.00003915, -0.00003917, 0.00004317, 0.00000275, -0.00007925, 0.00002225, 0.00009169, -0.00005466, -0.00008973, 0.00007555, 0.00005463, -0.00011204, -0.00003660, 0.00011234, 0.00000254, -0.00010286, 0.00002541, 0.00007543, -0.00005885, -0.00006392, 0.00005599, 0.00003086, -0.00005583, -0.00000718, 0.00005088, -0.00000128, -0.00003482, 0.00000482, 0.00001700, -0.00000754, -0.00001044, -0.00000461, -0.00000516, 0.00000668, 0.00001395, -0.00000165, -0.00001339, -0.00000901, -0.00000475, 0.00000131, 0.00001134, 0.00000854, -0.00000427, -0.00000676, -0.00000064, 0.00000183, 0.00000181, 0.00000144, -0.00000087, -0.00000199, -0.00000144, -0.00000195, -0.00000040, 0.00000339, 0.00000279, -0.00000097, -0.00000175, -0.00000206, -0.00000430, -0.00000452, -0.00000332, -0.00000216, 0.00000246, 0.00000507, -0.00000159, -0.00000370, 0.00000910, 0.00001451, -0.00000078, -0.00001054, -0.00000130, 0.00000338, -0.00000287, -0.00000011, 0.00000659, -0.00000367, -0.00001874, -0.00001379, },
		{ -0.00002011, -0.00000981, 0.00000462, 0.00000333, -0.00000575, -0.00000494, 0.00000233, 0.00000057, -0.00000644, -0.00000438, 0.00000085, -0.00000216, -0.00000509, -0.00000176, -0.00000121, -0.00000494, -0.00000939, -0.00001459, -0.00000963, 0.00000507, 0.00000142, -0.00001512, -0.00000652, 0.00000570, -0.00001760, -0.00003113, -0.00000206, 0.00000628, -0.00002516, -0.00003195, -0.00001167, -0.00001767, -0.00007847, -0.00019908, -0.00024753, -0.00001171, 0.00034249, 0.00037034, 0.00017039, 0.00024080, 0.00041412, 0.00014243, -0.00030303, -0.00028901, -0.00001892, 0.00000305, -0.00008040, -0.00003167, -0.00001021, -0.00003501, 0.00002410, 0.00002228, -0.00007744, -0.00002838, 0.00007710, -0.00002529, -0.00009974, 0.00005956, 0.00009372, -0.00009244, -0.00007414, 0.00010495, 0.00002606, -0.00012725, -0.00000244, 0.00011139, -0.00002985, -0.00008871, 0.00004888, 0.00005275, -0.00006967, -0.00003806, 0.00005479, 0.00000770, -0.00004489, 0.00001084, 0.00003613, -0.00001203, -0.00002128, 0.00000743, 0.00000591, -0.00000644, -0.00000592, -0.00000901, -0.00000314, 0.00001583, 0.00001220, -0.00001103, -0.00001564, -0.00000543, -0.00000156, 0.00000351, 0.00001090, 0.00000591, -0.00000508, -0.00000619, -0.00000065, 0.00000280, 0.00000281, -0.00000019, -0.00000237, -0.00000083, -0.00000062, -0.00000256, -0.00000006, 0.00000415, 0.00000223, -0.00000185, -0.00000214, -0.00000207, -0.00000359, -0.00000433, -0.00000421, -0.00000124, 0.00000448, 0.00000324, -0.00000421, 0.00000015, 0.00001372, 0.00001053, -0.00000654, -0.00000867, 0.00000170, 0.00000129, -0.00000317, 0.00000348, 0.00000568, -0.00000928, },
		{ 0.00000232, -0.00001387, -0.00001901, -0.00000577, 0.00000545, 0.00000100, -0.00000637, -0.00000225, 0.00000380, -0.00000137, -0.00000739, -0.00000335, -0.00000008, -0.00000337, -0.00000392, -0.00000222, -0.00000261, -0.00000214, -0.00000762, -0.00001837, -0.00001060, 0.00000774, -0.00000059, -0.00001637, -0.00000221, 0.00000487, -0.00002135, -0.00002714, -0.00000208, -0.00000116, -0.00002040, -0.00002639, -0.00002560, -0.00001975, -0.00006046, -0.00021542, -0.00028705, -0.00000739, 0.00036278, 0.00037902, 0.00025462, 0.00036322, 0.00036371, -0.00004959, -0.00035303, -0.00018438, 0.00000194, -0.00005199, -0.00005440, 0.00000426, -0.00005494, -0.00004648, 0.00007385, 0.00001144, -0.00011647, 0.00000394, 0.00010043, -0.00006787, -0.00010282, 0.00010129, 0.00007681, -0.00012471, -0.00004335, 0.00012144, -0.00001037, -0.00012614, 0.00003127, 0.00009522, -0.00005472, -0.00006413, 0.00006055, 0.00002626, -0.00006773, -0.00001474, 0.00004298, -0.00000819, -0.00002710, 0.00001955, 0.00001845, -0.00001449, -0.00000844, 0.00000436, -0.00000235, -0.00000304, -0.00000478, -0.00001209, 0.00000272, 0.00002189, 0.00000632, -0.00001838, -0.00001503, -0.00000194, 0.00000129, 0.00000524, 0.00000936, 0.00000331, -0.00000515, -0.00000548, -0.00000042, 0.00000381, 0.00000286, -0.00000202, -0.00000288, 0.00000039, -0.00000044, -0.00000284, 0.00000067, 0.00000436, 0.00000138, -0.00000247, -0.00000242, -0.00000190, -0.00000284, -0.00000450, -0.00000468, 0.00000036, 0.00000530, 0.00000034, -0.00000473, 0.00000544, 0.00001556, 0.00000469, -0.00000990, -0.00000556, 0.00000286, -0.00000057, -0.00000149, 0.00000584, },
		{ 0.00000117, 0.00000600, -0.00000239, -0.00001650, -0.00001609, -0.00000242, 0.00000465, -0.00000153, -0.00000551, 0.00000095, 0.00000370, -0.00000371, -0.00000690, -0.00000264, -0.00000221, -0.00000358, -0.00000223, -0.00000393, -0.00000339, 0.00000172, -0.00000743, -0.00002229, -0.00000978, 0.00000901, -0.00000358, -0.00001481, 0.00000124, 0.00000098, -0.00002156, -0.00002136, -0.00000797, -0.00000739, -0.00000924, -0.00002648, -0.00004422, -0.00001295, -0.00004332, -0.00024105, -0.00031748, -0.00000134, 0.00036075, 0.00040642, 0.00039591, 0.00046259, 0.00022204, -0.00024206, -0.00032087, -0.00007240, -0.00002947, -0.00009708, 0.00000336, 0.00001605, -0.00011488, -0.00002903, 0.00012330, -0.00002509, -0.00014273, 0.00005300, 0.00010360, -0.00011567, -0.00008472, 0.00013698, 0.00004311, -0.00014293, -0.00000390, 0.00012058, -0.00004615, -0.00010902, 0.00005666, 0.00006815, -0.00006692, -0.00003570, 0.00005898, 0.00000278, -0.00005514, 0.00000072, 0.00002515, -0.00001391, -0.00000798, 0.00001846, 0.00000272, -0.00000999, 0.00000055, -0.00000214, -0.00000668, 0.00000046, -0.00000618, -0.00001212, 0.00001036, 0.00002346, -0.00000164, -0.00002264, -0.00001261, 0.00000113, 0.00000387, 0.00000616, 0.00000694, 0.00000120, -0.00000461, -0.00000473, 0.00000006, 0.00000451, 0.00000206, -0.00000354, -0.00000256, 0.00000128, -0.00000062, -0.00000260, 0.00000143, 0.00000400, 0.00000047, -0.00000287, -0.00000260, -0.00000152, -0.00000227, -0.00000489, -0.00000443, 0.00000205, 0.00000452, -0.00000249, -0.00000275, 0.00001059, 0.00001425, -0.00000134, -0.00001066, -0.00000255, 0.00000257, -0.00000116, },
		{ 0.00000179, -0.00000017, 0.00000347, 0.00000384, -0.00000712, -0.00001690, -0.00001216, -0.00000026, 0.00000240, -0.00000354, -0.00000295, 0.00000372, 0.00000180, -0.00000540, -0.00000509, -0.00000315, -0.00000491, -0.00000231, -0.00000096, -0.00000666, -0.00000269, 0.00000561, -0.00000910, -0.00002495, -0.00000789, 0.00000784, -0.00000581, -0.00001055, 0.00000162, -0.00000422, -0.00001665, -0.00001719, -0.00001945, -0.00000831, 0.00000526, -0.00003532, -0.00006179, 0.00000259, -0.00003394, -0.00027025, -0.00033401, -0.00000729, 0.00033830, 0.00047913, 0.00058606, 0.00049757, -0.00000388, -0.00038935, -0.00021073, 0.00001056, -0.00010514, -0.00010901, 0.00007833, -0.00000795, -0.00017234, 0.00001899, 0.00015633, -0.00008110, -0.00014521, 0.00010784, 0.00008239, -0.00015679, -0.00004783, 0.00015714, -0.00000021, -0.00014222, 0.00003505, 0.00010271, -0.00007297, -0.00008059, 0.00006839, 0.00003726, -0.00006508, -0.00001055, 0.00004657, -0.00001238, -0.00003685, 0.00000584, 0.00000714, -0.00000984, 0.00000718, 0.00000990, -0.00000744, -0.00000161, 0.00000420, -0.00000929, -0.00000743, 0.00000240, -0.00000840, -0.00000857, 0.00001730, 0.00002047, -0.00000950, -0.00002372, -0.00000939, 0.00000376, 0.00000609, 0.00000600, 0.00000409, -0.00000013, -0.00000369, -0.00000395, 0.00000067, 0.00000463, 0.00000074, -0.00000439, -0.00000178, 0.00000168, -0.00000081, -0.00000191, 0.00000191, 0.00000320, -0.00000032, -0.00000311, -0.00000260, -0.00000100, -0.00000201, -0.00000518, -0.00000342, 0.00000310, 0.00000238, -0.00000406, 0.00000129, 0.00001408, 0.00001031, -0.00000614, -0.00000945, -0.00000052, },
		{ -0.00000731, 0.00000040, 0.00000149, 0.00000188, 0.00000434, 0.00000000, -0.00001080, -0.00001528, -0.00000795, 0.00000037, -0.00000094, -0.00000424, 0.00000100, 0.00000498, -0.00000140, -0.00000546, -0.00000278, -0.00000535, -0.00000708, 0.00000021, -0.00000102, -0.00000945, -0.00000028, 0.00000818, -0.00001192, -0.00002529, -0.00000651, 0.00000428, -0.00000534, -0.00000518, -0.00000238, -0.00000771, -0.00000710, -0.00001814, -0.00003353, -0.00000147, 0.00001774, -0.00005244, -0.00007204, 0.00002194, -0.00003650, -0.00029296, -0.00033911, -0.00004066, 0.00031054, 0.00062139, 0.00079723, 0.00043201, -0.00028368, -0.00045207, -0.00004826, 0.00003581, -0.00020184, -0.00007449, 0.00014797, -0.00006821, -0.00020815, 0.00008911, 0.00016032, -0.00014334, -0.00011936, 0.00015465, 0.00003971, -0.00018045, -0.00000020, 0.00015631, -0.00004308, -0.00012290, 0.00006455, 0.00007286, -0.00008530, -0.00004839, 0.00006524, 0.00001007, -0.00005188, 0.00000588, 0.00002852, -0.00001695, -0.00001884, 0.00000155, -0.00000585, 0.00000080, 0.00001493, -0.00000203, -0.00001064, 0.00000703, 0.00000287, -0.00001481, -0.00000601, 0.00000219, -0.00000950, -0.00000225, 0.00002143, 0.00001407, -0.00001552, -0.00002221, -0.00000610, 0.00000604, 0.00000769, 0.00000475, 0.00000134, -0.00000062, -0.00000262, -0.00000316, 0.00000121, 0.00000409, -0.00000070, -0.00000450, -0.00000093, 0.00000164, -0.00000074, -0.00000100, 0.00000195, 0.00000217, -0.00000094, -0.00000322, -0.00000239, -0.00000052, -0.00000206, -0.00000501, -0.00000199, 0.00000295, -0.00000039, -0.00000359, 0.00000623, 0.00001495, 0.00000498, -0.00000890, },
		{ -0.00000036, -0.00000962, -0.00000515, 0.00000067, 0.00000216, 0.00000400, 0.00000333, -0.00000449, -0.00001288, -0.00001213, -0.00000410, -0.00000076, -0.00000467, -0.00000295, 0.00000545, 0.00000399, -0.00000477, -0.00000353, -0.00000121, -0.00000893, -0.00000771, 0.00000302, -0.00000273, -0.00001103, 0.00000316, 0.00000841, -0.00001435, -0.00002325, -0.00000741, -0.00000028, -0.00000104, -0.00000136, -0.00001023, -0.00000650, 0.00000427, -0.00002614, -0.00004533, 0.00001228, 0.00002261, -0.00007312, -0.00007128, 0.00003659, -0.00004927, -0.00029853, -0.00034459, -0.00011171, 0.00030544, 0.00084529, 0.00098489, 0.00024680, -0.00056852, -0.00040870, 0.00012575, -0.00000877, -0.00028831, 0.00000490, 0.00018914, -0.00015307, -0.00020814, 0.00016489, 0.00013060, -0.00019556, -0.00006892, 0.00018114, -0.00001513, -0.00018051, 0.00004690, 0.00013484, -0.00007566, -0.00009038, 0.00007865, 0.00003902, -0.00008195, -0.00002034, 0.00005019, -0.00000778, -0.00003286, 0.00001137, 0.00001098, -0.00001214, -0.00000613, -0.00000847, -0.00001092, 0.00001320, 0.00001453, -0.00001291, -0.00000778, 0.00001294, -0.00000169, -0.00001751, -0.00000414, 0.00000042, -0.00000817, 0.00000506, 0.00002164, 0.00000611, -0.00001880, -0.00001909, -0.00000307, 0.00000803, 0.00000835, 0.00000264, -0.00000080, -0.00000038, -0.00000158, -0.00000237, 0.00000148, 0.00000299, -0.00000188, -0.00000402, -0.00000028, 0.00000140, -0.00000033, -0.00000015, 0.00000155, 0.00000110, -0.00000142, -0.00000318, -0.00000196, -0.00000022, -0.00000224, -0.00000422, -0.00000069, 0.00000145, -0.00000273, -0.00000096, 0.00001067, 0.00001307, },
		{ 0.00001336, 0.00000908, -0.00000463, -0.00000886, -0.00000342, 0.00000100, 0.00000365, 0.00000520, 0.00000072, -0.00000865, -0.00001326, -0.00000799, -0.00000116, -0.00000363, -0.00000780, 0.00000049, 0.00000903, 0.00000075, -0.00000691, -0.00000012, -0.00000143, -0.00001278, -0.00000647, 0.00000479, -0.00000548, -0.00001042, 0.00000614, 0.00000623, -0.00001467, -0.00002005, -0.00001163, -0.00000350, 0.00000657, -0.00000170, -0.00001936, 0.00000088, 0.00001309, -0.00004007, -0.00005006, 0.00002832, 0.00001687, -0.00008958, -0.00006103, 0.00003767, -0.00006332, -0.00028105, -0.00036991, -0.00021859, 0.00035840, 0.00114247, 0.00109682, -0.00005157, -0.00080216, -0.00026277, 0.00026636, -0.00011475, -0.00033508, 0.00011274, 0.00018594, -0.00024204, -0.00016813, 0.00022700, 0.00007235, -0.00022366, -0.00000496, 0.00018050, -0.00006925, -0.00015744, 0.00008250, 0.00009871, -0.00009137, -0.00005328, 0.00007602, 0.00000956, -0.00006617, -0.00000228, 0.00002919, -0.00001408, -0.00001442, 0.00000731, -0.00000105, -0.00000188, -0.00000127, -0.00001922, -0.00000810, 0.00002269, 0.00000782, -0.00001943, -0.00000142, 0.00001461, -0.00000717, -0.00001748, -0.00000316, -0.00000159, -0.00000411, 0.00001128, 0.00001802, -0.00000148, -0.00001928, -0.00001525, -0.00000034, 0.00000958, 0.00000785, 0.00000016, -0.00000204, 0.00000030, -0.00000069, -0.00000165, 0.00000136, 0.00000156, -0.00000256, -0.00000319, 0.00000005, 0.00000120, 0.00000035, 0.00000044, 0.00000084, 0.00000012, -0.00000177, -0.00000293, -0.00000143, -0.00000021, -0.00000227, -0.00000289, -0.00000011, -0.00000102, -0.00000370, 0.00000321, },
		{ -0.00000277, 0.00000781, 0.00001368, 0.00000408, -0.00000731, -0.00000734, -0.00000203, 0.00000191, 0.00000524, 0.00000491, -0.00000275, -0.00001184, -0.00001211, -0.00000336, 0.00000029, -0.00000784, -0.00000922, 0.00000552, 0.00001044, -0.00000381, -0.00000677, 0.00000346, -0.00000374, -0.00001545, -0.00000396, 0.00000447, -0.00000783, -0.00000755, 0.00000696, 0.00000278, -0.00001178, -0.00001779, -0.00001877, -0.00000309, 0.00001516, -0.00000740, -0.00002594, 0.00001320, 0.00001528, -0.00005558, -0.00004542, 0.00003960, 0.00000229, -0.00009403, -0.00004889, 0.00002044, -0.00006438, -0.00024431, -0.00043675, -0.00034276, 0.00050241, 0.00148067, 0.00108637, -0.00042880, -0.00093519, -0.00004313, 0.00033717, -0.00025560, -0.00032385, 0.00022279, 0.00013510, -0.00031216, -0.00009555, 0.00025934, -0.00000093, -0.00022022, 0.00005753, 0.00015352, -0.00011022, -0.00011800, 0.00009929, 0.00005745, -0.00008872, -0.00002065, 0.00006010, -0.00000938, -0.00004425, 0.00000361, 0.00000897, -0.00001039, -0.00000168, -0.00000218, -0.00000520, 0.00000882, -0.00000374, -0.00002626, -0.00000004, 0.00002625, -0.00000168, -0.00002025, 0.00000529, 0.00001218, -0.00001159, -0.00001573, -0.00000347, -0.00000238, 0.00000182, 0.00001473, 0.00001171, -0.00000727, -0.00001758, -0.00001130, 0.00000214, 0.00001041, 0.00000621, -0.00000213, -0.00000231, 0.00000112, 0.00000000, -0.00000105, 0.00000082, 0.00000010, -0.00000272, -0.00000231, 0.00000014, 0.00000119, 0.00000111, 0.00000065, -0.00000002, -0.00000075, -0.00000199, -0.00000246, -0.00000093, -0.00000043, -0.00000191, -0.00000139, -0.00000062, -0.00000369, },
		{ -0.00000217, -0.00000563, -0.00000008, 0.00001159, 0.00001170, -0.00000084, -0.00000850, -0.00000558, -0.00000052, 0.00000337, 0.00000606, 0.00000317, -0.00000624, -0.00001379, -0.00000967, 0.00000117, -0.00000019, -0.00001246, -0.00000814, 0.00001080, 0.00000896, -0.00000806, -0.00000423, 0.00000561, -0.00000739, -0.00001583, -0.00000165, 0.00000197, -0.00000815, -0.00000357, 0.00000455, 0.00000012, -0.00000596, -0.00001858, -0.00002677, 0.00000198, 0.00002136, -0.00001727, -0.00002651, 0.00002645, 0.00000913, -0.00006636, -0.00003347, 0.00003934, -0.00001409, -0.00008249, -0.00004675, -0.00001193, -0.00003759, -0.00020409, -0.00056067, -0.00044945, 0.00075650, 0.00180712, 0.00092614, -0.00082839, -0.00093826, 0.00020311, 0.00032047, -0.00039474, -0.00025310, 0.00030723, 0.00004721, -0.00034525, -0.00000725, 0.00025407, -0.00007161, -0.00018671, 0.00010441, 0.00010814, -0.00012984, -0.00007291, 0.00009564, 0.00002104, -0.00007138, 0.00000085, 0.00003756, -0.00001557, -0.00002327, -0.00000092, -0.00000516, -0.00000102, 0.00000305, -0.00001203, -0.00000220, 0.00001570, -0.00001057, -0.00002711, 0.00000921, 0.00002323, -0.00001017, -0.00001611, 0.00000975, 0.00000705, -0.00001389, -0.00001355, -0.00000462, -0.00000108, 0.00000806, 0.00001469, 0.00000440, -0.00001056, -0.00001452, -0.00000750, 0.00000433, 0.00001020, 0.00000376, -0.00000369, -0.00000179, 0.00000181, 0.00000050, -0.00000066, -0.00000006, -0.00000114, -0.00000244, -0.00000155, 0.00000014, 0.00000142, 0.00000175, 0.00000052, -0.00000089, -0.00000150, -0.00000205, -0.00000180, -0.00000060, -0.00000071, -0.00000109, -0.00000020, },
		{ 0.00000003, 0.00000022, -0.00000428, -0.00000611, 0.00000366, 0.00001363, 0.00000808, -0.00000489, -0.00000861, -0.00000368, 0.00000148, 0.00000486, 0.00000552, 0.00000055, -0.00000914, -0.00001451, -0.00000622, 0.00000488, -0.00000269, -0.00001617, -0.00000448, 0.00001458, 0.00000492, -0.00001030, -0.00000029, 0.00000523, -0.00001079, -0.00001379, -0.00000120, -0.00000165, -0.00000541, -0.00000046, -0.00000084, 0.00000031, 0.00000098, -0.00002339, -0.00003265, 0.00001060, 0.00002215, -0.00002768, -0.00001981, 0.00003503, -0.00000318, -0.00006667, -0.00002081, 0.00002446, -0.00002169, -0.00005785, -0.00006657, -0.00004704, 0.00002590, -0.00018624, -0.00074267, -0.00049398, 0.00111602, 0.00205876, 0.00061791, -0.00118403, -0.00081040, 0.00042299, 0.00022180, -0.00049648, -0.00013836, 0.00034527, -0.00005659, -0.00033348, 0.00007581, 0.00021385, -0.00012345, -0.00013273, 0.00012655, 0.00005669, -0.00012641, -0.00003333, 0.00007566, -0.00000315, -0.00004657, 0.00000875, 0.00001585, -0.00001089, -0.00000872, -0.00001128, -0.00001086, 0.00000878, 0.00000058, -0.00001806, 0.00000484, 0.00001667, -0.00001766, -0.00002189, 0.00001575, 0.00001530, -0.00001485, -0.00000924, 0.00001064, 0.00000115, -0.00001405, -0.00001188, -0.00000555, 0.00000230, 0.00001288, 0.00001146, -0.00000231, -0.00001146, -0.00001084, -0.00000388, 0.00000600, 0.00000876, 0.00000105, -0.00000423, -0.00000085, 0.00000219, 0.00000084, -0.00000053, -0.00000114, -0.00000200, -0.00000191, -0.00000097, 0.00000020, 0.00000179, 0.00000210, 0.00000014, -0.00000169, -0.00000213, -0.00000192, -0.00000104, -0.00000048, -0.00000083, },
		{ -0.00000051, -0.00000063, 0.00000110, -0.00000030, -0.00000620, -0.00000490, 0.00000739, 0.00001356, 0.00000375, -0.00000780, -0.00000805, -0.00000138, 0.00000402, 0.00000558, 0.00000355, -0.00000222, -0.00001119, -0.00001405, -0.00000207, 0.00000698, -0.00000672, -0.00001761, 0.00000081, 0.00001538, -0.00000023, -0.00000951, 0.00000325, 0.00000228, -0.00001214, -0.00001044, -0.00000370, -0.00000446, 0.00000013, -0.00000024, -0.00000750, 0.00000437, 0.00000631, -0.00003130, -0.00003375, 0.00001953, 0.00001656, -0.00003375, -0.00000805, 0.00003408, -0.00001554, -0.00005426, -0.00001685, -0.00000192, -0.00001033, -0.00003076, -0.00011492, -0.00006551, 0.00012295, -0.00022005, -0.00096381, -0.00043299, 0.00154833, 0.00217719, 0.00019541, -0.00143544, -0.00057953, 0.00057164, 0.00006761, -0.00053637, -0.00000701, 0.00032932, -0.00015136, -0.00028144, 0.00013508, 0.00015058, -0.00014645, -0.00007264, 0.00012237, 0.00001182, -0.00010465, -0.00000743, 0.00004733, -0.00001235, -0.00002237, 0.00000497, 0.00000077, -0.00000016, -0.00000301, -0.00002199, -0.00000902, 0.00001477, -0.00000589, -0.00001834, 0.00001189, 0.00001222, -0.00002143, -0.00001295, 0.00001716, 0.00000553, -0.00001468, -0.00000233, 0.00000808, -0.00000389, -0.00001281, -0.00001099, -0.00000528, 0.00000687, 0.00001507, 0.00000619, -0.00000728, -0.00001053, -0.00000699, -0.00000044, 0.00000679, 0.00000620, -0.00000128, -0.00000376, 0.00000011, 0.00000224, 0.00000107, -0.00000067, -0.00000221, -0.00000245, -0.00000131, -0.00000053, 0.00000042, 0.00000216, 0.00000212, -0.00000033, -0.00000238, -0.00000264, -0.00000157, -0.00000032, },
		{ -0.00000104, 0.00000027, -0.00000056, -0.00000014, 0.00000170, -0.00000157, -0.00000716, -0.00000228, 0.00001011, 0.00001158, -0.00000047, -0.00000966, -0.00000695, 0.00000161, 0.00000661, 0.00000488, 0.00000064, -0.00000455, -0.00001241, -0.00001236, 0.00000216, 0.00000679, -0.00001108, -0.00001589, 0.00000597, 0.00001260, -0.00000441, -0.00000590, 0.00000449, -0.00000201, -0.00001035, -0.00000774, -0.00000892, -0.00000449, 0.00000680, -0.00000394, -0.00001279, 0.00001155, 0.00000757, -0.00003942, -0.00002922, 0.00002446, 0.00000663, -0.00003136, 0.00000322, 0.00002200, -0.00001969, -0.00003242, -0.00003023, -0.00003011, 0.00002523, -0.00001724, -0.00018850, -0.00004682, 0.00023559, -0.00032867, -0.00118529, -0.00023766, 0.00199621, 0.00212398, -0.00028145, -0.00154318, -0.00029518, 0.00062362, -0.00010316, -0.00050766, 0.00011041, 0.00026667, -0.00021593, -0.00020419, 0.00015975, 0.00008128, -0.00013937, -0.00002090, 0.00009753, -0.00001731, -0.00007361, 0.00000171, 0.00001964, -0.00000859, -0.00000509, -0.00000541, -0.00000517, 0.00001087, -0.00000511, -0.00002866, -0.00000295, 0.00001506, -0.00001234, -0.00001352, 0.00001562, 0.00000492, -0.00002007, -0.00000373, 0.00001316, -0.00000287, -0.00001043, 0.00000242, 0.00000325, -0.00000718, -0.00001110, -0.00001045, -0.00000330, 0.00001123, 0.00001427, 0.00000037, -0.00001012, -0.00000847, -0.00000310, 0.00000265, 0.00000635, 0.00000296, -0.00000268, -0.00000256, 0.00000077, 0.00000204, 0.00000119, -0.00000104, -0.00000312, -0.00000256, -0.00000076, -0.00000013, 0.00000078, 0.00000235, 0.00000187, -0.00000079, -0.00000297, -0.00000298, },
		{ -0.00000348, -0.00000311, -0.00000042, 0.00000068, -0.00000047, 0.00000047, 0.00000156, -0.00000307, -0.00000668, 0.00000098, 0.00001114, 0.00000833, -0.00000407, -0.00001070, -0.00000513, 0.00000534, 0.00000845, 0.00000261, -0.00000244, -0.00000624, -0.00001281, -0.00000945, 0.00000560, 0.00000417, -0.00001413, -0.00001114, 0.00000889, 0.00000698, -0.00000568, -0.00000099, 0.00000235, -0.00000552, -0.00000562, -0.00000765, -0.00001523, -0.00000076, 0.00001202, -0.00001086, -0.00001431, 0.00001925, 0.00000393, -0.00004391, -0.00002094, 0.00002182, -0.00000264, -0.00001938, 0.00000674, 0.00000208, -0.00000836, -0.00001007, -0.00006488, -0.00004611, 0.00008161, -0.00003358, -0.00027248, 0.00002298, 0.00033307, -0.00051928, -0.00135530, 0.00009508, 0.00238883, 0.00189204, -0.00073972, -0.00149799, -0.00001535, 0.00057834, -0.00025067, -0.00042199, 0.00018929, 0.00017637, -0.00023899, -0.00012198, 0.00014941, 0.00002252, -0.00010928, 0.00001229, 0.00006253, -0.00002741, -0.00004335, -0.00000358, -0.00000042, 0.00000266, 0.00000248, -0.00001636, -0.00000304, 0.00001767, -0.00001155, -0.00002934, 0.00000325, 0.00001035, -0.00001556, -0.00000621, 0.00001457, -0.00000185, -0.00001397, 0.00000259, 0.00000539, -0.00000762, -0.00000409, 0.00000394, -0.00000224, -0.00000865, -0.00000955, -0.00000951, 0.00000008, 0.00001404, 0.00001102, -0.00000472, -0.00001103, -0.00000577, 0.00000081, 0.00000501, 0.00000457, -0.00000028, -0.00000294, -0.00000110, 0.00000097, 0.00000173, 0.00000119, -0.00000157, -0.00000376, -0.00000243, -0.00000029, 0.00000033, 0.00000119, 0.00000229, 0.00000145, -0.00000120, },
		{ 0.00000098, -0.00000155, -0.00000387, -0.00000300, 0.00000019, 0.00000095, -0.00000017, 0.00000090, 0.00000064, -0.00000412, -0.00000470, 0.00000395, 0.00001033, 0.00000453, -0.00000692, -0.00001099, -0.00000229, 0.00000934, 0.00000874, -0.00000080, -0.00000506, -0.00000742, -0.00001225, -0.00000552, 0.00000718, -0.00000025, -0.00001425, -0.00000468, 0.00000796, 0.00000050, -0.00000311, 0.00000291, -0.00000283, -0.00000620, 0.00000037, -0.00001120, -0.00002014, 0.00000602, 0.00001345, -0.00001853, -0.00001116, 0.00002395, -0.00000297, -0.00004162, -0.00001336, 0.00001075, -0.00000491, -0.00000102, -0.00000365, -0.00001759, 0.00002130, 0.00000054, -0.00011708, -0.00003626, 0.00014587, -0.00008999, -0.00034306, 0.00014422, 0.00037913, -0.00077656, -0.00142117, 0.00053732, 0.00265771, 0.00150926, -0.00111031, -0.00132214, 0.00020828, 0.00045811, -0.00034513, -0.00030449, 0.00021763, 0.00008267, -0.00022148, -0.00005383, 0.00011308, -0.00001443, -0.00006847, 0.00002358, 0.00002879, -0.00002124, -0.00002169, -0.00001719, -0.00000968, 0.00001484, 0.00000134, -0.00002312, 0.00000355, 0.00001826, -0.00001808, -0.00002480, 0.00000636, 0.00000326, -0.00001425, 0.00000036, 0.00000947, -0.00000531, -0.00000536, 0.00000423, -0.00000342, -0.00000805, 0.00000200, 0.00000240, -0.00000705, -0.00000881, -0.00000828, -0.00000755, 0.00000394, 0.00001454, 0.00000641, -0.00000833, -0.00001052, -0.00000269, 0.00000461, 0.00000618, 0.00000170, -0.00000281, -0.00000215, 0.00000014, 0.00000075, 0.00000142, 0.00000104, -0.00000215, -0.00000407, -0.00000217, 0.00000012, 0.00000087, 0.00000151, 0.00000196, },
		{ 0.00000163, 0.00000145, 0.00000053, -0.00000187, -0.00000409, -0.00000266, 0.00000070, 0.00000117, 0.00000030, 0.00000083, -0.00000077, -0.00000413, -0.00000166, 0.00000583, 0.00000797, 0.00000077, -0.00000907, -0.00001032, 0.00000168, 0.00001275, 0.00000713, -0.00000448, -0.00000693, -0.00000828, -0.00001042, -0.00000120, 0.00000604, -0.00000499, -0.00001056, 0.00000125, 0.00000280, -0.00000430, 0.00000264, 0.00000368, -0.00000924, -0.00000304, 0.00000512, -0.00001769, -0.00002137, 0.00001346, 0.00001019, -0.00002356, -0.00000474, 0.00002273, -0.00000918, -0.00003181, -0.00001209, -0.00000578, 0.00000506, 0.00001642, -0.00003002, -0.00002629, 0.00006537, -0.00001204, -0.00017503, 0.00000736, 0.00019819, -0.00018523, -0.00037396, 0.00030000, 0.00034257, -0.00106232, -0.00134386, 0.00103307, 0.00275336, 0.00103307, -0.00134386, -0.00106232, 0.00034257, 0.00030000, -0.00037396, -0.00018523, 0.00019819, 0.00000736, -0.00017503, -0.00001204, 0.00006537, -0.00002629, -0.00003002, 0.00001642, 0.00000506, -0.00000578, -0.00001209, -0.00003181, -0.00000918, 0.00002273, -0.00000474, -0.00002356, 0.00001019, 0.00001346, -0.00002137, -0.00001769, 0.00000512, -0.00000304, -0.00000924, 0.00000368, 0.00000264, -0.00000430, 0.00000280, 0.00000125, -0.00001056, -0.00000499, 0.00000604, -0.00000120, -0.00001042, -0.00000828, -0.00000693, -0.00000448, 0.00000713, 0.00001275, 0.00000168, -0.00001032, -0.00000907, 0.00000077, 0.00000797, 0.00000583, -0.00000166, -0.00000413, -0.00000077, 0.00000083, 0.00000030, 0.00000117, 0.00000070, -0.00000266, -0.00000409, -0.00000187, 0.00000053, 0.00000145, },
		{ 0.00000098, 0.00000196, 0.00000151, 0.00000087, 0.00000012, -0.00000217, -0.00000407, -0.00000215, 0.00000104, 0.00000142, 0.00000075, 0.00000014, -0.00000215, -0.00000281, 0.00000170, 0.00000618, 0.00000461, -0.00000269, -0.00001052, -0.00000833, 0.00000641, 0.00001454, 0.00000394, -0.00000755, -0.00000828, -0.00000881, -0.00000705, 0.00000240, 0.00000200, -0.00000805, -0.00000342, 0.00000423, -0.00000536, -0.00000531, 0.00000947, 0.00000036, -0.00001425, 0.00000326, 0.00000636, -0.00002480, -0.00001808, 0.00001826, 0.00000355, -0.00002312, 0.00000134, 0.00001484, -0.00000968, -0.00001719, -0.00002169, -0.00002124, 0.00002879, 0.00002358, -0.00006847, -0.00001443, 0.00011308, -0.00005383, -0.00022148, 0.00008267, 0.00021763, -0.00030449, -0.00034513, 0.00045811, 0.00020828, -0.00132214, -0.00111031, 0.00150926, 0.00265771, 0.00053732, -0.00142117, -0.00077656, 0.00037913, 0.00014422, -0.00034306, -0.00008999, 0.00014587, -0.00003626, -0.00011708, 0.00000054, 0.00002130, -0.00001759, -0.00000365, -0.00000102, -0.00000491, 0.00001075, -0.00001336, -0.00004162, -0.00000297, 0.00002395, -0.00001116, -0.00001853, 0.00001345, 0.00000602, -0.00002014, -0.00001120, 0.00000037, -0.00000620, -0.00000283, 0.00000291, -0.00000311, 0.00000050, 0.00000796, -0.00000468, -0.00001425, -0.00000025, 0.00000718, -0.00000552, -0.00001225, -0.00000742, -0.00000506, -0.00000080, 0.00000874, 0.00000934, -0.00000229, -0.00001099, -0.00000692, 0.00000453, 0.00001033, 0.00000395, -0.00000470, -0.00000412, 0.00000064, 0.00000090, -0.00000017, 0.00000095, 0.00000019, -0.00000300, -0.00000387, -0.00000155, },
		{ -0.00000348, -0.00000120, 0.00000145, 0.00000229, 0.00000119, 0.00000033, -0.00000029, -0.00000243, -0.00000376, -0.00000157, 0.00000119, 0.00000173, 0.00000097, -0.00000110, -0.00000294, -0.00000028, 0.00000457, 0.00000501, 0.00000081, -0.00000577, -0.00001103, -0.00000472, 0.00001102, 0.00001404, 0.00000008, -0.00000951, -0.00000955, -0.00000865, -0.00000224, 0.00000394, -0.00000409, -0.00000762, 0.00000539, 0.00000259, -0.00001397, -0.00000185, 0.00001457, -0.00000621, -0.00001556, 0.00001035, 0.00000325, -0.00002934, -0.00001155, 0.00001767, -0.00000304, -0.00001636, 0.00000248, 0.00000266, -0.00000042, -0.00000358, -0.00004335, -0.00002741, 0.00006253, 0.00001229, -0.00010928, 0.00002252, 0.00014941, -0.00012198, -0.00023899, 0.00017637, 0.00018929, -0.00042199, -0.00025067, 0.00057834, -0.00001535, -0.00149799, -0.00073972, 0.00189204, 0.00238883, 0.00009508, -0.00135530, -0.00051928, 0.00033307, 0.00002298, -0.00027248, -0.00003358, 0.00008161, -0.00004611, -0.00006488, -0.00001007, -0.00000836, 0.00000208, 0.00000674, -0.00001938, -0.00000264, 0.00002182, -0.00002094, -0.00004391, 0.00000393, 0.00001925, -0.00001431, -0.00001086, 0.00001202, -0.00000076, -0.00001523, -0.00000765, -0.00000562, -0.00000552, 0.00000235, -0.00000099, -0.00000568, 0.00000698, 0.00000889, -0.00001114, -0.00001413, 0.00000417, 0.00000560, -0.00000945, -0.00001281, -0.00000624, -0.00000244, 0.00000261, 0.00000845, 0.00000534, -0.00000513, -0.00001070, -0.00000407, 0.00000833, 0.00001114, 0.00000098, -0.00000668, -0.00000307, 0.00000156, 0.00000047, -0.00000047, 0.00000068, -0.00000042, -0.00000311, },
		{ -0.00000104, -0.00000298, -0.00000297, -0.00000079, 0.00000187, 0.00000235, 0.00000078, -0.00000013, -0.00000076, -0.00000256, -0.00000312, -0.00000104, 0.00000119, 0.00000204, 0.00000077, -0.00000256, -0.00000268, 0.00000296, 0.00000635, 0.00000265, -0.00000310, -0.00000847, -0.00001012, 0.00000037, 0.00001427, 0.00001123, -0.00000330, -0.00001045, -0.00001110, -0.00000718, 0.00000325, 0.00000242, -0.00001043, -0.00000287, 0.00001316, -0.00000373, -0.00002007, 0.00000492, 0.00001562, -0.00001352, -0.00001234, 0.00001506, -0.00000295, -0.00002866, -0.00000511, 0.00001087, -0.00000517, -0.00000541, -0.00000509, -0.00000859, 0.00001964, 0.00000171, -0.00007361, -0.00001731, 0.00009753, -0.00002090, -0.00013937, 0.00008128, 0.00015975, -0.00020419, -0.00021593, 0.00026667, 0.00011041, -0.00050766, -0.00010316, 0.00062362, -0.00029518, -0.00154318, -0.00028145, 0.00212398, 0.00199621, -0.00023766, -0.00118529, -0.00032867, 0.00023559, -0.00004682, -0.00018850, -0.00001724, 0.00002523, -0.00003011, -0.00003023, -0.00003242, -0.00001969, 0.00002200, 0.00000322, -0.00003136, 0.00000663, 0.00002446, -0.00002922, -0.00003942, 0.00000757, 0.00001155, -0.00001279, -0.00000394, 0.00000680, -0.00000449, -0.00000892, -0.00000774, -0.00001035, -0.00000201, 0.00000449, -0.00000590, -0.00000441, 0.00001260, 0.00000597, -0.00001589, -0.00001108, 0.00000679, 0.00000216, -0.00001236, -0.00001241, -0.00000455, 0.00000064, 0.00000488, 0.00000661, 0.00000161, -0.00000695, -0.00000966, -0.00000047, 0.00001158, 0.00001011, -0.00000228, -0.00000716, -0.00000157, 0.00000170, -0.00000014, -0.00000056, 0.00000027, },
		{ -0.00000051, -0.00000032, -0.00000157, -0.00000264, -0.00000238, -0.00000033, 0.00000212, 0.00000216, 0.00000042, -0.00000053, -0.00000131, -0.00000245, -0.00000221, -0.00000067, 0.00000107, 0.00000224, 0.00000011, -0.00000376, -0.00000128, 0.00000620, 0.00000679, -0.00000044, -0.00000699, -0.00001053, -0.00000728, 0.00000619, 0.00001507, 0.00000687, -0.00000528, -0.00001099, -0.00001281, -0.00000389, 0.00000808, -0.00000233, -0.00001468, 0.00000553, 0.00001716, -0.00001295, -0.00002143, 0.00001222, 0.00001189, -0.00001834, -0.00000589, 0.00001477, -0.00000902, -0.00002199, -0.00000301, -0.00000016, 0.00000077, 0.00000497, -0.00002237, -0.00001235, 0.00004733, -0.00000743, -0.00010465, 0.00001182, 0.00012237, -0.00007264, -0.00014645, 0.00015058, 0.00013508, -0.00028144, -0.00015136, 0.00032932, -0.00000701, -0.00053637, 0.00006761, 0.00057164, -0.00057953, -0.00143544, 0.00019541, 0.00217719, 0.00154833, -0.00043299, -0.00096381, -0.00022005, 0.00012295, -0.00006551, -0.00011492, -0.00003076, -0.00001033, -0.00000192, -0.00001685, -0.00005426, -0.00001554, 0.00003408, -0.00000805, -0.00003375, 0.00001656, 0.00001953, -0.00003375, -0.00003130, 0.00000631, 0.00000437, -0.00000750, -0.00000024, 0.00000013, -0.00000446, -0.00000370, -0.00001044, -0.00001214, 0.00000228, 0.00000325, -0.00000951, -0.00000023, 0.00001538, 0.00000081, -0.00001761, -0.00000672, 0.00000698, -0.00000207, -0.00001405, -0.00001119, -0.00000222, 0.00000355, 0.00000558, 0.00000402, -0.00000138, -0.00000805, -0.00000780, 0.00000375, 0.00001356, 0.00000739, -0.00000490, -0.00000620, -0.00000030, 0.00000110, -0.00000063, },
		{ 0.00000003, -0.00000083, -0.00000048, -0.00000104, -0.00000192, -0.00000213, -0.00000169, 0.00000014, 0.00000210, 0.00000179, 0.00000020, -0.00000097, -0.00000191, -0.00000200, -0.00000114, -0.00000053, 0.00000084, 0.00000219, -0.00000085, -0.00000423, 0.00000105, 0.00000876, 0.00000600, -0.00000388, -0.00001084, -0.00001146, -0.00000231, 0.00001146, 0.00001288, 0.00000230, -0.00000555, -0.00001188, -0.00001405, 0.00000115, 0.00001064, -0.00000924, -0.00001485, 0.00001530, 0.00001575, -0.00002189, -0.00001766, 0.00001667, 0.00000484, -0.00001806, 0.00000058, 0.00000878, -0.00001086, -0.00001128, -0.00000872, -0.00001089, 0.00001585, 0.00000875, -0.00004657, -0.00000315, 0.00007566, -0.00003333, -0.00012641, 0.00005669, 0.00012655, -0.00013273, -0.00012345, 0.00021385, 0.00007581, -0.00033348, -0.00005659, 0.00034527, -0.00013836, -0.00049648, 0.00022180, 0.00042299, -0.00081040, -0.00118403, 0.00061791, 0.00205876, 0.00111602, -0.00049398, -0.00074267, -0.00018624, 0.00002590, -0.00004704, -0.00006657, -0.00005785, -0.00002169, 0.00002446, -0.00002081, -0.00006667, -0.00000318, 0.00003503, -0.00001981, -0.00002768, 0.00002215, 0.00001060, -0.00003265, -0.00002339, 0.00000098, 0.00000031, -0.00000084, -0.00000046, -0.00000541, -0.00000165, -0.00000120, -0.00001379, -0.00001079, 0.00000523, -0.00000029, -0.00001030, 0.00000492, 0.00001458, -0.00000448, -0.00001617, -0.00000269, 0.00000488, -0.00000622, -0.00001451, -0.00000914, 0.00000055, 0.00000552, 0.00000486, 0.00000148, -0.00000368, -0.00000861, -0.00000489, 0.00000808, 0.00001363, 0.00000366, -0.00000611, -0.00000428, 0.00000022, },
		{ -0.00000217, -0.00000020, -0.00000109, -0.00000071, -0.00000060, -0.00000180, -0.00000205, -0.00000150, -0.00000089, 0.00000052, 0.00000175, 0.00000142, 0.00000014, -0.00000155, -0.00000244, -0.00000114, -0.00000006, -0.00000066, 0.00000050, 0.00000181, -0.00000179, -0.00000369, 0.00000376, 0.00001020, 0.00000433, -0.00000750, -0.00001452, -0.00001056, 0.00000440, 0.00001469, 0.00000806, -0.00000108, -0.00000462, -0.00001355, -0.00001389, 0.00000705, 0.00000975, -0.00001611, -0.00001017, 0.00002323, 0.00000921, -0.00002711, -0.00001057, 0.00001570, -0.00000220, -0.00001203, 0.00000305, -0.00000102, -0.00000516, -0.00000092, -0.00002327, -0.00001557, 0.00003756, 0.00000085, -0.00007138, 0.00002104, 0.00009564, -0.00007291, -0.00012984, 0.00010814, 0.00010441, -0.00018671, -0.00007161, 0.00025407, -0.00000725, -0.00034525, 0.00004721, 0.00030723, -0.00025310, -0.00039474, 0.00032047, 0.00020311, -0.00093826, -0.00082839, 0.00092614, 0.00180712, 0.00075650, -0.00044945, -0.00056067, -0.00020409, -0.00003759, -0.00001193, -0.00004675, -0.00008249, -0.00001409, 0.00003934, -0.00003347, -0.00006636, 0.00000913, 0.00002645, -0.00002651, -0.00001727, 0.00002136, 0.00000198, -0.00002677, -0.00001858, -0.00000596, 0.00000012, 0.00000455, -0.00000357, -0.00000815, 0.00000197, -0.00000165, -0.00001583, -0.00000739, 0.00000561, -0.00000423, -0.00000806, 0.00000896, 0.00001080, -0.00000814, -0.00001246, -0.00000019, 0.00000117, -0.00000967, -0.00001379, -0.00000624, 0.00000317, 0.00000606, 0.00000337, -0.00000052, -0.00000558, -0.00000850, -0.00000084, 0.00001170, 0.00001159, -0.00000008, -0.00000563, },
		{ -0.00000277, -0.00000369, -0.00000062, -0.00000139, -0.00000191, -0.00000043, -0.00000093, -0.00000246, -0.00000199, -0.00000075, -0.00000002, 0.00000065, 0.00000111, 0.00000119, 0.00000014, -0.00000231, -0.00000272, 0.00000010, 0.00000082, -0.00000105, 0.00000000, 0.00000112, -0.00000231, -0.00000213, 0.00000621, 0.00001041, 0.00000214, -0.00001130, -0.00001758, -0.00000727, 0.00001171, 0.00001473, 0.00000182, -0.00000238, -0.00000347, -0.00001573, -0.00001159, 0.00001218, 0.00000529, -0.00002025, -0.00000168, 0.00002625, -0.00000004, -0.00002626, -0.00000374, 0.00000882, -0.00000520, -0.00000218, -0.00000168, -0.00001039, 0.00000897, 0.00000361, -0.00004425, -0.00000938, 0.00006010, -0.00002065, -0.00008872, 0.00005745, 0.00009929, -0.00011800, -0.00011022, 0.00015352, 0.00005753, -0.00022022, -0.00000093, 0.00025934, -0.00009555, -0.00031216, 0.00013510, 0.00022279, -0.00032385, -0.00025560, 0.00033717, -0.00004313, -0.00093519, -0.00042880, 0.00108637, 0.00148067, 0.00050241, -0.00034276, -0.00043675, -0.00024431, -0.00006438, 0.00002044, -0.00004889, -0.00009403, 0.00000229, 0.00003960, -0.00004542, -0.00005558, 0.00001528, 0.00001320, -0.00002594, -0.00000740, 0.00001516, -0.00000309, -0.00001877, -0.00001779, -0.00001178, 0.00000278, 0.00000696, -0.00000755, -0.00000783, 0.00000447, -0.00000396, -0.00001545, -0.00000374, 0.00000346, -0.00000677, -0.00000381, 0.00001044, 0.00000552, -0.00000922, -0.00000784, 0.00000029, -0.00000336, -0.00001211, -0.00001184, -0.00000275, 0.00000491, 0.00000524, 0.00000191, -0.00000203, -0.00000734, -0.00000731, 0.00000408, 0.00001368, 0.00000781, },
		{ 0.00001336, 0.00000321, -0.00000370, -0.00000102, -0.00000011, -0.00000289, -0.00000227, -0.00000021, -0.00000143, -0.00000293, -0.00000177, 0.00000012, 0.00000084, 0.00000044, 0.00000035, 0.00000120, 0.00000005, -0.00000319, -0.00000256, 0.00000156, 0.00000136, -0.00000165, -0.00000069, 0.00000030, -0.00000204, 0.00000016, 0.00000785, 0.00000958, -0.00000034, -0.00001525, -0.00001928, -0.00000148, 0.00001802, 0.00001128, -0.00000411, -0.00000159, -0.00000316, -0.00001748, -0.00000717, 0.00001461, -0.00000142, -0.00001943, 0.00000782, 0.00002269, -0.00000810, -0.00001922, -0.00000127, -0.00000188, -0.00000105, 0.00000731, -0.00001442, -0.00001408, 0.00002919, -0.00000228, -0.00006617, 0.00000956, 0.00007602, -0.00005328, -0.00009137, 0.00009871, 0.00008250, -0.00015744, -0.00006925, 0.00018050, -0.00000496, -0.00022366, 0.00007235, 0.00022700, -0.00016813, -0.00024204, 0.00018594, 0.00011274, -0.00033508, -0.00011475, 0.00026636, -0.00026277, -0.00080216, -0.00005157, 0.00109682, 0.00114247, 0.00035840, -0.00021859, -0.00036991, -0.00028105, -0.00006332, 0.00003767, -0.00006103, -0.00008958, 0.00001687, 0.00002832, -0.00005006, -0.00004007, 0.00001309, 0.00000088, -0.00001936, -0.00000170, 0.00000657, -0.00000350, -0.00001163, -0.00002005, -0.00001467, 0.00000623, 0.00000614, -0.00001042, -0.00000548, 0.00000479, -0.00000647, -0.00001278, -0.00000143, -0.00000012, -0.00000691, 0.00000075, 0.00000903, 0.00000049, -0.00000780, -0.00000363, -0.00000116, -0.00000799, -0.00001326, -0.00000865, 0.00000072, 0.00000520, 0.00000365, 0.00000100, -0.00000342, -0.00000886, -0.00000463, 0.00000908, },
		{ -0.00000036, 0.00001307, 0.00001067, -0.00000096, -0.00000273, 0.00000145, -0.00000069, -0.00000422, -0.00000224, -0.00000022, -0.00000196, -0.00000318, -0.00000142, 0.00000110, 0.00000155, -0.00000015, -0.00000033, 0.00000140, -0.00000028, -0.00000402, -0.00000188, 0.00000299, 0.00000148, -0.00000237, -0.00000158, -0.00000038, -0.00000080, 0.00000264, 0.00000835, 0.00000803, -0.00000307, -0.00001909, -0.00001880, 0.00000611, 0.00002164, 0.00000506, -0.00000817, 0.00000042, -0.00000414, -0.00001751, -0.00000169, 0.00001294, -0.00000778, -0.00001291, 0.00001453, 0.00001320, -0.00001092, -0.00000847, -0.00000613, -0.00001214, 0.00001098, 0.00001137, -0.00003286, -0.00000778, 0.00005019, -0.00002034, -0.00008195, 0.00003902, 0.00007865, -0.00009038, -0.00007566, 0.00013484, 0.00004690, -0.00018051, -0.00001513, 0.00018114, -0.00006892, -0.00019556, 0.00013060, 0.00016489, -0.00020814, -0.00015307, 0.00018914, 0.00000490, -0.00028831, -0.00000877, 0.00012575, -0.00040870, -0.00056852, 0.00024680, 0.00098489, 0.00084529, 0.00030544, -0.00011171, -0.00034459, -0.00029853, -0.00004927, 0.00003659, -0.00007128, -0.00007312, 0.00002261, 0.00001228, -0.00004533, -0.00002614, 0.00000427, -0.00000650, -0.00001023, -0.00000136, -0.00000104, -0.00000028, -0.00000741, -0.00002325, -0.00001435, 0.00000841, 0.00000316, -0.00001103, -0.00000273, 0.00000302, -0.00000771, -0.00000893, -0.00000121, -0.00000353, -0.00000477, 0.00000399, 0.00000545, -0.00000295, -0.00000467, -0.00000076, -0.00000410, -0.00001213, -0.00001288, -0.00000449, 0.00000333, 0.00000400, 0.00000216, 0.00000067, -0.00000515, -0.00000962, },
		{ -0.00000731, -0.00000890, 0.00000498, 0.00001495, 0.00000623, -0.00000359, -0.00000039, 0.00000295, -0.00000199, -0.00000501, -0.00000206, -0.00000052, -0.00000239, -0.00000322, -0.00000094, 0.00000217, 0.00000195, -0.00000100, -0.00000074, 0.00000164, -0.00000093, -0.00000450, -0.00000070, 0.00000409, 0.00000121, -0.00000316, -0.00000262, -0.00000062, 0.00000134, 0.00000475, 0.00000769, 0.00000604, -0.00000610, -0.00002221, -0.00001552, 0.00001407, 0.00002143, -0.00000225, -0.00000950, 0.00000219, -0.00000601, -0.00001481, 0.00000287, 0.00000703, -0.00001064, -0.00000203, 0.00001493, 0.00000080, -0.00000585, 0.00000155, -0.00001884, -0.00001695, 0.00002852, 0.00000588, -0.00005188, 0.00001007, 0.00006524, -0.00004839, -0.00008530, 0.00007286, 0.00006455, -0.00012290, -0.00004308, 0.00015631, -0.00000020, -0.00018045, 0.00003971, 0.00015465, -0.00011936, -0.00014334, 0.00016032, 0.00008911, -0.00020815, -0.00006821, 0.00014797, -0.00007449, -0.00020184, 0.00003581, -0.00004826, -0.00045207, -0.00028368, 0.00043201, 0.00079723, 0.00062139, 0.00031054, -0.00004066, -0.00033911, -0.00029296, -0.00003650, 0.00002194, -0.00007204, -0.00005244, 0.00001774, -0.00000147, -0.00003353, -0.00001814, -0.00000710, -0.00000771, -0.00000238, -0.00000518, -0.00000534, 0.00000428, -0.00000651, -0.00002529, -0.00001192, 0.00000818, -0.00000028, -0.00000945, -0.00000102, 0.00000021, -0.00000708, -0.00000535, -0.00000278, -0.00000546, -0.00000140, 0.00000498, 0.00000100, -0.00000424, -0.00000094, 0.00000037, -0.00000795, -0.00001528, -0.00001080, 0.00000000, 0.00000434, 0.00000188, 0.00000149, 0.00000040, },
		{ 0.00000179, -0.00000052, -0.00000945, -0.00000614, 0.00001031, 0.00001408, 0.00000129, -0.00000406, 0.00000238, 0.00000310, -0.00000342, -0.00000518, -0.00000201, -0.00000100, -0.00000260, -0.00000311, -0.00000032, 0.00000320, 0.00000191, -0.00000191, -0.00000081, 0.00000168, -0.00000178, -0.00000439, 0.00000074, 0.00000463, 0.00000067, -0.00000395, -0.00000369, -0.00000013, 0.00000409, 0.00000600, 0.00000609, 0.00000376, -0.00000939, -0.00002372, -0.00000950, 0.00002047, 0.00001730, -0.00000857, -0.00000840, 0.00000240, -0.00000743, -0.00000929, 0.00000420, -0.00000161, -0.00000744, 0.00000990, 0.00000718, -0.00000984, 0.00000714, 0.00000584, -0.00003685, -0.00001238, 0.00004657, -0.00001055, -0.00006508, 0.00003726, 0.00006839, -0.00008059, -0.00007297, 0.00010271, 0.00003505, -0.00014222, -0.00000021, 0.00015714, -0.00004783, -0.00015679, 0.00008239, 0.00010784, -0.00014521, -0.00008110, 0.00015633, 0.00001899, -0.00017234, -0.00000795, 0.00007833, -0.00010901, -0.00010514, 0.00001056, -0.00021073, -0.00038935, -0.00000388, 0.00049757, 0.00058606, 0.00047913, 0.00033830, -0.00000729, -0.00033401, -0.00027025, -0.00003394, 0.00000259, -0.00006179, -0.00003532, 0.00000526, -0.00000831, -0.00001945, -0.00001719, -0.00001665, -0.00000422, 0.00000162, -0.00001055, -0.00000581, 0.00000784, -0.00000789, -0.00002495, -0.00000910, 0.00000561, -0.00000269, -0.00000666, -0.00000096, -0.00000231, -0.00000491, -0.00000315, -0.00000509, -0.00000540, 0.00000180, 0.00000372, -0.00000295, -0.00000354, 0.00000240, -0.00000026, -0.00001216, -0.00001690, -0.00000712, 0.00000384, 0.00000347, -0.00000017, },
		{ 0.00000117, -0.00000116, 0.00000257, -0.00000255, -0.00001066, -0.00000134, 0.00001425, 0.00001059, -0.00000275, -0.00000249, 0.00000452, 0.00000205, -0.00000443, -0.00000489, -0.00000227, -0.00000152, -0.00000260, -0.00000287, 0.00000047, 0.00000400, 0.00000143, -0.00000260, -0.00000062, 0.00000128, -0.00000256, -0.00000354, 0.00000206, 0.00000451, 0.00000006, -0.00000473, -0.00000461, 0.00000120, 0.00000694, 0.00000616, 0.00000387, 0.00000113, -0.00001261, -0.00002264, -0.00000164, 0.00002346, 0.00001036, -0.00001212, -0.00000618, 0.00000046, -0.00000668, -0.00000214, 0.00000055, -0.00000999, 0.00000272, 0.00001846, -0.00000798, -0.00001391, 0.00002515, 0.00000072, -0.00005514, 0.00000278, 0.00005898, -0.00003570, -0.00006692, 0.00006815, 0.00005666, -0.00010902, -0.00004615, 0.00012058, -0.00000390, -0.00014293, 0.00004311, 0.00013698, -0.00008472, -0.00011567, 0.00010360, 0.00005300, -0.00014273, -0.00002509, 0.00012330, -0.00002903, -0.00011488, 0.00001605, 0.00000336, -0.00009708, -0.00002947, -0.00007240, -0.00032087, -0.00024206, 0.00022204, 0.00046259, 0.00039591, 0.00040642, 0.00036075, -0.00000134, -0.00031748, -0.00024105, -0.00004332, -0.00001295, -0.00004422, -0.00002648, -0.00000924, -0.00000739, -0.00000797, -0.00002136, -0.00002156, 0.00000098, 0.00000124, -0.00001481, -0.00000358, 0.00000901, -0.00000978, -0.00002229, -0.00000743, 0.00000172, -0.00000339, -0.00000393, -0.00000223, -0.00000358, -0.00000221, -0.00000264, -0.00000690, -0.00000371, 0.00000370, 0.00000095, -0.00000551, -0.00000153, 0.00000465, -0.00000242, -0.00001609, -0.00001650, -0.00000239, 0.00000600, },
		{ 0.00000232, 0.00000584, -0.00000149, -0.00000057, 0.00000286, -0.00000556, -0.00000990, 0.00000469, 0.00001556, 0.00000544, -0.00000473, 0.00000034, 0.00000530, 0.00000036, -0.00000468, -0.00000450, -0.00000284, -0.00000190, -0.00000242, -0.00000247, 0.00000138, 0.00000436, 0.00000067, -0.00000284, -0.00000044, 0.00000039, -0.00000288, -0.00000202, 0.00000286, 0.00000381, -0.00000042, -0.00000548, -0.00000515, 0.00000331, 0.00000936, 0.00000524, 0.00000129, -0.00000194, -0.00001503, -0.00001838, 0.00000632, 0.00002189, 0.00000272, -0.00001209, -0.00000478, -0.00000304, -0.00000235, 0.00000436, -0.00000844, -0.00001449, 0.00001845, 0.00001955, -0.00002710, -0.00000819, 0.00004298, -0.00001474, -0.00006773, 0.00002626, 0.00006055, -0.00006413, -0.00005472, 0.00009522, 0.00003127, -0.00012614, -0.00001037, 0.00012144, -0.00004335, -0.00012471, 0.00007681, 0.00010129, -0.00010282, -0.00006787, 0.00010043, 0.00000394, -0.00011647, 0.00001144, 0.00007385, -0.00004648, -0.00005494, 0.00000426, -0.00005440, -0.00005199, 0.00000194, -0.00018438, -0.00035303, -0.00004959, 0.00036371, 0.00036322, 0.00025462, 0.00037902, 0.00036278, -0.00000739, -0.00028705, -0.00021542, -0.00006046, -0.00001975, -0.00002560, -0.00002639, -0.00002040, -0.00000116, -0.00000208, -0.00002714, -0.00002135, 0.00000487, -0.00000221, -0.00001637, -0.00000059, 0.00000774, -0.00001060, -0.00001837, -0.00000762, -0.00000214, -0.00000261, -0.00000222, -0.00000392, -0.00000337, -0.00000008, -0.00000335, -0.00000739, -0.00000137, 0.00000380, -0.00000225, -0.00000637, 0.00000100, 0.00000545, -0.00000577, -0.00001901, -0.00001387, },
		{ -0.00002011, -0.00000928, 0.00000568, 0.00000348, -0.00000317, 0.00000129, 0.00000170, -0.00000867, -0.00000654, 0.00001053, 0.00001372, 0.00000015, -0.00000421, 0.00000324, 0.00000448, -0.00000124, -0.00000421, -0.00000433, -0.00000359, -0.00000207, -0.00000214, -0.00000185, 0.00000223, 0.00000415, -0.00000006, -0.00000256, -0.00000062, -0.00000083, -0.00000237, -0.00000019, 0.00000281, 0.00000280, -0.00000065, -0.00000619, -0.00000508, 0.00000591, 0.00001090, 0.00000351, -0.00000156, -0.00000543, -0.00001564, -0.00001103, 0.00001220, 0.00001583, -0.00000314, -0.00000901, -0.00000592, -0.00000644, 0.00000591, 0.00000743, -0.00002128, -0.00001203, 0.00003613, 0.00001084, -0.00004489, 0.00000770, 0.00005479, -0.00003806, -0.00006967, 0.00005275, 0.00004888, -0.00008871, -0.00002985, 0.00011139, -0.00000244, -0.00012725, 0.00002606, 0.00010495, -0.00007414, -0.00009244, 0.00009372, 0.00005956, -0.00009974, -0.00002529, 0.00007710, -0.00002838, -0.00007744, 0.00002228, 0.00002410, -0.00003501, -0.00001021, -0.00003167, -0.00008040, 0.00000305, -0.00001892, -0.00028901, -0.00030303, 0.00014243, 0.00041412, 0.00024080, 0.00017039, 0.00037034, 0.00034249, -0.00001171, -0.00024753, -0.00019908, -0.00007847, -0.00001767, -0.00001167, -0.00003195, -0.00002516, 0.00000628, -0.00000206, -0.00003113, -0.00001760, 0.00000570, -0.00000652, -0.00001512, 0.00000142, 0.00000507, -0.00000963, -0.00001459, -0.00000939, -0.00000494, -0.00000121, -0.00000176, -0.00000509, -0.00000216, 0.00000085, -0.00000438, -0.00000644, 0.00000057, 0.00000233, -0.00000494, -0.00000575, 0.00000333, 0.00000462, -0.00000981, },
		{ 0.00000217, -0.00001379, -0.00001874, -0.00000367, 0.00000659, -0.00000011, -0.00000287, 0.00000338, -0.00000130, -0.00001054, -0.00000078, 0.00001451, 0.00000910, -0.00000370, -0.00000159, 0.00000507, 0.00000246, -0.00000216, -0.00000332, -0.00000452, -0.00000430, -0.00000206, -0.00000175, -0.00000097, 0.00000279, 0.00000339, -0.00000040, -0.00000195, -0.00000144, -0.00000199, -0.00000087, 0.00000144, 0.00000181, 0.00000183, -0.00000064, -0.00000676, -0.00000427, 0.00000854, 0.00001134, 0.00000131, -0.00000475, -0.00000901, -0.00001339, -0.00000165, 0.00001395, 0.00000668, -0.00000516, -0.00000461, -0.00001044, -0.00000754, 0.00001700, 0.00000482, -0.00003482, -0.00000128, 0.00005088, -0.00000718, -0.00005583, 0.00003086, 0.00005599, -0.00006392, -0.00005885, 0.00007543, 0.00002541, -0.00010286, 0.00000254, 0.00011234, -0.00003660, -0.00011204, 0.00005463, 0.00007555, -0.00008973, -0.00005466, 0.00009169, 0.00002225, -0.00007925, 0.00000275, 0.00004317, -0.00003917, -0.00003915, 0.00000976, -0.00001177, -0.00000512, 0.00000886, -0.00007362, -0.00007247, 0.00004304, -0.00008274, -0.00035325, -0.00018784, 0.00029443, 0.00038767, 0.00013056, 0.00013475, 0.00035956, 0.00030738, -0.00000656, -0.00020694, -0.00019225, -0.00009123, -0.00001035, -0.00000540, -0.00003847, -0.00002355, 0.00001119, -0.00000592, -0.00003142, -0.00001286, 0.00000339, -0.00000966, -0.00001215, 0.00000158, 0.00000241, -0.00000715, -0.00001201, -0.00001185, -0.00000635, -0.00000009, -0.00000214, -0.00000519, -0.00000081, 0.00000060, -0.00000485, -0.00000456, 0.00000143, -0.00000000, -0.00000657, -0.00000412, 0.00000494, },
		{ -0.00000198, 0.00000542, -0.00000174, -0.00001674, -0.00001468, 0.00000147, 0.00000460, -0.00000337, -0.00000041, 0.00000430, -0.00000567, -0.00000988, 0.00000625, 0.00001534, 0.00000303, -0.00000500, 0.00000203, 0.00000518, -0.00000000, -0.00000216, -0.00000246, -0.00000502, -0.00000480, -0.00000196, -0.00000118, 0.00000009, 0.00000278, 0.00000231, -0.00000009, -0.00000138, -0.00000298, -0.00000254, 0.00000146, 0.00000230, 0.00000009, 0.00000120, -0.00000048, -0.00000704, -0.00000273, 0.00001063, 0.00001070, -0.00000102, -0.00000843, -0.00001196, -0.00000760, 0.00000786, 0.00001030, -0.00000307, -0.00000240, -0.00000132, -0.00001780, -0.00000438, 0.00002851, -0.00000432, -0.00004508, 0.00001670, 0.00005801, -0.00003115, -0.00005591, 0.00005587, 0.00004484, -0.00008573, -0.00003684, 0.00008818, -0.00000485, -0.00010272, 0.00003495, 0.00009796, -0.00006323, -0.00008468, 0.00006928, 0.00004103, -0.00008815, -0.00002075, 0.00007402, -0.00000254, -0.00004967, 0.00001235, 0.00001007, -0.00003082, -0.00001297, -0.00001676, -0.00002609, 0.00002822, 0.00000187, -0.00010377, -0.00004087, 0.00004973, -0.00016642, -0.00035690, -0.00003917, 0.00038223, 0.00031272, 0.00005418, 0.00012961, 0.00033597, 0.00026861, 0.00000875, -0.00017237, -0.00019092, -0.00009580, -0.00000294, -0.00000630, -0.00004191, -0.00001806, 0.00001162, -0.00001068, -0.00002806, -0.00000940, -0.00000077, -0.00001064, -0.00000904, 0.00000010, 0.00000084, -0.00000409, -0.00001096, -0.00001401, -0.00000672, 0.00000023, -0.00000262, -0.00000431, -0.00000000, -0.00000028, -0.00000440, -0.00000257, 0.00000109, -0.00000244, -0.00000706, },
		{ -0.00000442, -0.00000655, 0.00000020, 0.00000440, -0.00000662, -0.00001756, -0.00000848, 0.00000456, 0.00000031, -0.00000471, 0.00000339, 0.00000294, -0.00001006, -0.00000607, 0.00001274, 0.00001263, -0.00000269, -0.00000355, 0.00000526, 0.00000356, -0.00000207, -0.00000142, -0.00000195, -0.00000562, -0.00000508, -0.00000186, -0.00000031, 0.00000115, 0.00000204, 0.00000133, 0.00000091, -0.00000133, -0.00000497, -0.00000203, 0.00000415, 0.00000202, -0.00000192, 0.00000107, -0.00000031, -0.00000676, -0.00000070, 0.00001167, 0.00000929, -0.00000331, -0.00001268, -0.00001327, 0.00000169, 0.00001513, 0.00000132, -0.00001054, 0.00000456, -0.00000142, -0.00002618, 0.00000395, 0.00003733, -0.00001893, -0.00004846, 0.00003846, 0.00005455, -0.00005561, -0.00004406, 0.00007635, 0.00002320, -0.00009774, -0.00000854, 0.00008749, -0.00003482, -0.00008837, 0.00005991, 0.00007228, -0.00007665, -0.00005242, 0.00006814, 0.00000994, -0.00007233, 0.00000202, 0.00004803, -0.00001143, -0.00002085, 0.00000569, -0.00001281, -0.00001142, -0.00000459, -0.00004475, -0.00001958, 0.00005110, -0.00002229, -0.00011090, -0.00000387, 0.00001757, -0.00024093, -0.00029764, 0.00010650, 0.00040090, 0.00022152, 0.00001760, 0.00013526, 0.00029895, 0.00023564, 0.00002973, -0.00014719, -0.00018959, -0.00009299, 0.00000038, -0.00001134, -0.00004038, -0.00001209, 0.00000784, -0.00001378, -0.00002256, -0.00000835, -0.00000498, -0.00000963, -0.00000699, -0.00000209, 0.00000071, -0.00000151, -0.00001109, -0.00001527, -0.00000671, -0.00000022, -0.00000262, -0.00000299, 0.00000001, -0.00000108, -0.00000320, -0.00000114, -0.00000009, },
		{ -0.00000051, -0.00000157, -0.00000575, -0.00000528, 0.00000200, 0.00000165, -0.00001153, -0.00001546, -0.00000151, 0.00000445, -0.00000476, -0.00000324, 0.00000691, -0.00000093, -0.00001271, 0.00000044, 0.00001680, 0.00000712, -0.00000637, -0.00000008, 0.00000695, 0.00000085, -0.00000316, -0.00000039, -0.00000189, -0.00000604, -0.00000523, -0.00000171, 0.00000092, 0.00000185, 0.00000063, 0.00000089, 0.00000233, -0.00000216, -0.00000686, -0.00000028, 0.00000645, 0.00000057, -0.00000367, 0.00000130, -0.00000016, -0.00000568, 0.00000135, 0.00001133, 0.00000760, -0.00000558, -0.00001732, -0.00001186, 0.00001351, 0.00001787, -0.00001151, -0.00001326, 0.00001375, -0.00000622, -0.00003297, 0.00001678, 0.00004063, -0.00003610, -0.00004296, 0.00005909, 0.00004029, -0.00007463, -0.00002257, 0.00008700, -0.00000404, -0.00009682, 0.00001932, 0.00007363, -0.00005758, -0.00006380, 0.00007222, 0.00004222, -0.00007518, -0.00002323, 0.00005388, -0.00001106, -0.00004883, 0.00001055, 0.00002235, -0.00000631, -0.00000079, -0.00001029, -0.00002097, 0.00000868, -0.00001269, -0.00006319, -0.00000018, 0.00005553, -0.00004964, -0.00009401, 0.00001921, -0.00004480, -0.00028061, -0.00019035, 0.00021830, 0.00036278, 0.00014101, 0.00001373, 0.00013686, 0.00025471, 0.00021309, 0.00005007, -0.00013058, -0.00018391, -0.00008628, -0.00000205, -0.00001666, -0.00003447, -0.00000843, 0.00000176, -0.00001405, -0.00001701, -0.00000944, -0.00000786, -0.00000764, -0.00000643, -0.00000393, 0.00000160, -0.00000005, -0.00001165, -0.00001556, -0.00000688, -0.00000098, -0.00000199, -0.00000184, -0.00000051, -0.00000135, -0.00000179, },
		{ -0.00000109, -0.00000073, -0.00000051, -0.00000294, -0.00000642, -0.00000345, 0.00000290, -0.00000279, -0.00001520, -0.00001036, 0.00000436, 0.00000105, -0.00000859, 0.00000082, 0.00000841, -0.00000638, -0.00001214, 0.00000812, 0.00001713, 0.00000048, -0.00000703, 0.00000401, 0.00000657, -0.00000198, -0.00000317, 0.00000041, -0.00000212, -0.00000616, -0.00000540, -0.00000139, 0.00000245, 0.00000188, -0.00000113, 0.00000131, 0.00000362, -0.00000389, -0.00000794, 0.00000243, 0.00000766, -0.00000165, -0.00000474, 0.00000160, 0.00000014, -0.00000364, 0.00000274, 0.00000958, 0.00000619, -0.00000801, -0.00002188, -0.00000691, 0.00002601, 0.00001461, -0.00002555, -0.00000994, 0.00002235, -0.00001557, -0.00003563, 0.00003193, 0.00003688, -0.00005187, -0.00002894, 0.00007360, 0.00001799, -0.00008360, 0.00000356, 0.00008516, -0.00003039, -0.00008345, 0.00004027, 0.00005052, -0.00006843, -0.00003560, 0.00007048, 0.00001527, -0.00006135, -0.00000330, 0.00003249, -0.00001908, -0.00002541, 0.00000644, 0.00000399, 0.00000687, 0.00000676, -0.00002691, -0.00001579, 0.00002071, -0.00003025, -0.00006630, 0.00002066, 0.00004169, -0.00006675, -0.00006194, 0.00001512, -0.00011800, -0.00027093, -0.00006139, 0.00027840, 0.00029084, 0.00008692, 0.00002847, 0.00012756, 0.00021165, 0.00020028, 0.00006477, -0.00011884, -0.00017234, -0.00007979, -0.00000922, -0.00001933, -0.00002649, -0.00000820, -0.00000423, -0.00001186, -0.00001298, -0.00001152, -0.00000888, -0.00000584, -0.00000698, -0.00000476, 0.00000274, 0.00000023, -0.00001202, -0.00001524, -0.00000737, -0.00000156, -0.00000101, -0.00000119, -0.00000104, },
		{ -0.00000092, -0.00000117, -0.00000069, -0.00000027, -0.00000076, -0.00000403, -0.00000646, -0.00000130, 0.00000231, -0.00000835, -0.00001631, -0.00000311, 0.00000732, -0.00000456, -0.00000948, 0.00000605, 0.00000674, -0.00001158, -0.00000785, 0.00001479, 0.00001359, -0.00000522, -0.00000475, 0.00000717, 0.00000440, -0.00000402, -0.00000247, 0.00000067, -0.00000232, -0.00000602, -0.00000569, -0.00000069, 0.00000400, 0.00000102, -0.00000270, 0.00000258, 0.00000419, -0.00000615, -0.00000765, 0.00000539, 0.00000737, -0.00000391, -0.00000506, 0.00000159, 0.00000098, -0.00000077, 0.00000273, 0.00000678, 0.00000550, -0.00001086, -0.00002553, 0.00000177, 0.00003683, 0.00000511, -0.00003762, -0.00000087, 0.00002753, -0.00002782, -0.00003254, 0.00004631, 0.00002638, -0.00006239, -0.00000916, 0.00007847, -0.00000736, -0.00008060, 0.00002809, 0.00007165, -0.00004981, -0.00006148, 0.00005009, 0.00002435, -0.00006630, -0.00001075, 0.00005725, -0.00000278, -0.00004079, 0.00000465, 0.00001108, -0.00001550, -0.00000842, -0.00000514, -0.00000374, 0.00002067, 0.00000301, -0.00003697, -0.00000321, 0.00002047, -0.00004783, -0.00005502, 0.00003262, 0.00001701, -0.00006592, -0.00002949, -0.00001870, -0.00017869, -0.00021223, 0.00006005, 0.00028478, 0.00021041, 0.00006249, 0.00004714, 0.00010811, 0.00017627, 0.00019302, 0.00007200, -0.00010764, -0.00015623, -0.00007643, -0.00001834, -0.00001836, -0.00001917, -0.00001077, -0.00000834, -0.00000859, -0.00001102, -0.00001325, -0.00000844, -0.00000498, -0.00000785, -0.00000455, 0.00000341, -0.00000028, -0.00001192, -0.00001476, -0.00000799, -0.00000174, -0.00000016, },
		{ -0.00000168, 0.00000025, -0.00000064, -0.00000085, -0.00000062, -0.00000026, -0.00000095, -0.00000490, -0.00000582, 0.00000074, -0.00000023, -0.00001390, -0.00001400, 0.00000448, 0.00000636, -0.00001027, -0.00000676, 0.00001033, 0.00000197, -0.00001442, -0.00000063, 0.00001837, 0.00000730, -0.00000824, -0.00000062, 0.00000828, 0.00000133, -0.00000477, -0.00000167, 0.00000039, -0.00000221, -0.00000582, -0.00000601, 0.00000051, 0.00000514, -0.00000062, -0.00000352, 0.00000429, 0.00000360, -0.00000816, -0.00000584, 0.00000761, 0.00000569, -0.00000542, -0.00000496, 0.00000109, 0.00000288, 0.00000243, 0.00000070, 0.00000363, 0.00000575, -0.00001431, -0.00002733, 0.00001357, 0.00004358, -0.00000941, -0.00004484, 0.00001208, 0.00002732, -0.00004030, -0.00002359, 0.00005679, 0.00001126, -0.00006509, 0.00001204, 0.00007270, -0.00003007, -0.00006696, 0.00004549, 0.00005039, -0.00005850, -0.00003681, 0.00004795, 0.00000159, -0.00005376, 0.00000546, 0.00003792, -0.00000950, -0.00002014, 0.00000188, -0.00000451, -0.00000495, -0.00000093, -0.00001771, -0.00000179, 0.00002898, -0.00000705, -0.00003715, 0.00000909, 0.00000926, -0.00005740, -0.00003588, 0.00003038, -0.00000751, -0.00004778, -0.00001141, -0.00007344, -0.00020738, -0.00011850, 0.00015036, 0.00024885, 0.00014168, 0.00006116, 0.00005940, 0.00008393, 0.00015106, 0.00018633, 0.00007312, -0.00009405, -0.00013864, -0.00007692, -0.00002645, -0.00001467, -0.00001435, -0.00001445, -0.00001003, -0.00000576, -0.00001067, -0.00001378, -0.00000741, -0.00000512, -0.00000832, -0.00000375, 0.00000332, -0.00000105, -0.00001145, -0.00001433, -0.00000837, },
		{ -0.00001383, -0.00000831, -0.00000176, 0.00000028, 0.00000005, -0.00000043, -0.00000110, -0.00000033, -0.00000105, -0.00000556, -0.00000443, 0.00000199, -0.00000473, -0.00001784, -0.00000836, 0.00001017, 0.00000175, -0.00001374, -0.00000116, 0.00001162, -0.00000451, -0.00001331, 0.00000755, 0.00001766, 0.00000030, -0.00000780, 0.00000365, 0.00000701, -0.00000146, -0.00000432, -0.00000132, -0.00000012, -0.00000169, -0.00000579, -0.00000613, 0.00000210, 0.00000543, -0.00000261, -0.00000324, 0.00000571, 0.00000186, -0.00000900, -0.00000296, 0.00000819, 0.00000332, -0.00000564, -0.00000511, 0.00000028, 0.00000633, 0.00000507, -0.00000359, 0.00000110, 0.00000685, -0.00001836, -0.00002640, 0.00002706, 0.00004449, -0.00002649, -0.00004539, 0.00002604, 0.00002119, -0.00005013, -0.00001024, 0.00006118, -0.00000507, -0.00005947, 0.00002995, 0.00005806, -0.00004536, -0.00004671, 0.00005249, 0.00002704, -0.00005589, -0.00001543, 0.00003639, -0.00001300, -0.00003596, 0.00001094, 0.00001875, -0.00000610, -0.00000487, -0.00000739, -0.00001140, 0.00000673, -0.00000223, -0.00002590, 0.00000564, 0.00002902, -0.00001707, -0.00002869, 0.00001485, -0.00000730, -0.00005529, -0.00001772, 0.00001517, -0.00002199, -0.00002059, -0.00001673, -0.00013229, -0.00019448, -0.00001211, 0.00019724, 0.00018974, 0.00009549, 0.00007160, 0.00006128, 0.00006143, 0.00013460, 0.00017676, 0.00007131, -0.00007749, -0.00012264, -0.00007998, -0.00003163, -0.00001027, -0.00001247, -0.00001748, -0.00000989, -0.00000436, -0.00001093, -0.00001299, -0.00000658, -0.00000582, -0.00000811, -0.00000294, 0.00000257, -0.00000177, -0.00001090, },
		{ -0.00000241, -0.00001043, -0.00001298, -0.00000795, -0.00000230, 0.00000032, 0.00000120, -0.00000041, -0.00000197, -0.00000014, -0.00000126, -0.00000584, -0.00000239, 0.00000163, -0.00001050, -0.00001867, -0.00000066, 0.00001206, -0.00000480, -0.00001325, 0.00000528, 0.00000886, -0.00001034, -0.00000794, 0.00001417, 0.00001283, -0.00000510, -0.00000434, 0.00000639, 0.00000395, -0.00000304, -0.00000324, -0.00000169, -0.00000046, -0.00000093, -0.00000601, -0.00000574, 0.00000372, 0.00000464, -0.00000423, -0.00000192, 0.00000605, -0.00000045, -0.00000796, -0.00000003, 0.00000667, 0.00000138, -0.00000459, -0.00000625, -0.00000020, 0.00001145, 0.00000602, -0.00000985, 0.00000014, 0.00000839, -0.00002272, -0.00002221, 0.00004020, 0.00003885, -0.00004299, -0.00003898, 0.00003785, 0.00001016, -0.00005502, 0.00000489, 0.00005887, -0.00001901, -0.00004722, 0.00004091, 0.00003849, -0.00005077, -0.00002524, 0.00004894, 0.00000720, -0.00004454, -0.00000167, 0.00002025, -0.00001769, -0.00001872, 0.00000701, 0.00000482, 0.00000340, 0.00000234, -0.00001778, -0.00001015, 0.00001464, -0.00000872, -0.00002717, 0.00001317, 0.00002180, -0.00002196, -0.00001623, 0.00001151, -0.00002205, -0.00004310, -0.00000796, -0.00000638, -0.00002169, 0.00000346, -0.00004537, -0.00017645, -0.00014275, 0.00008355, 0.00020121, 0.00012733, 0.00007286, 0.00008286, 0.00005441, 0.00004504, 0.00012321, 0.00016350, 0.00006972, -0.00005948, -0.00011004, -0.00008330, -0.00003363, -0.00000709, -0.00001269, -0.00001878, -0.00000909, -0.00000445, -0.00001089, -0.00001140, -0.00000631, -0.00000649, -0.00000735, -0.00000251, 0.00000151, },
		{ -0.00000251, 0.00000039, -0.00000316, -0.00000991, -0.00001159, -0.00000769, -0.00000328, 0.00000083, 0.00000246, -0.00000109, -0.00000273, 0.00000030, -0.00000176, -0.00000541, -0.00000017, -0.00000096, -0.00001612, -0.00001556, 0.00000687, 0.00000936, -0.00001069, -0.00000848, 0.00000995, 0.00000254, -0.00001308, 0.00000046, 0.00001703, 0.00000543, -0.00000721, 0.00000068, 0.00000658, 0.00000030, -0.00000307, -0.00000227, -0.00000262, -0.00000031, -0.00000024, -0.00000632, -0.00000468, 0.00000480, 0.00000294, -0.00000478, -0.00000015, 0.00000477, -0.00000228, -0.00000492, 0.00000160, 0.00000335, 0.00000103, -0.00000293, -0.00000884, 0.00000062, 0.00001782, 0.00000415, -0.00001716, 0.00000149, 0.00000975, -0.00002687, -0.00001482, 0.00005080, 0.00002728, -0.00005584, -0.00002691, 0.00004496, -0.00000351, -0.00005399, 0.00001885, 0.00005095, -0.00002779, -0.00003164, 0.00004333, 0.00001886, -0.00004664, -0.00000765, 0.00003751, -0.00000526, -0.00002906, 0.00000302, 0.00000489, -0.00001390, -0.00000665, -0.00000244, -0.00000143, 0.00001396, 0.00000197, -0.00002476, -0.00000397, 0.00001638, -0.00001575, -0.00002228, 0.00001653, 0.00001114, -0.00001973, -0.00000547, 0.00000077, -0.00002937, -0.00002648, -0.00000969, -0.00002548, -0.00000868, 0.00001325, -0.00008832, -0.00019146, -0.00006554, 0.00015081, 0.00017308, 0.00007654, 0.00006762, 0.00008788, 0.00004347, 0.00003591, 0.00011315, 0.00014799, 0.00007011, -0.00004244, -0.00010082, -0.00008481, -0.00003356, -0.00000618, -0.00001360, -0.00001831, -0.00000866, -0.00000548, -0.00001014, -0.00000966, -0.00000646, -0.00000673, -0.00000647, },
		{ -0.00000654, -0.00000580, -0.00000272, -0.00000075, -0.00000412, -0.00000904, -0.00000978, -0.00000795, -0.00000428, 0.00000198, 0.00000324, -0.00000234, -0.00000292, 0.00000067, -0.00000251, -0.00000396, 0.00000131, -0.00000580, -0.00001974, -0.00000888, 0.00001168, 0.00000286, -0.00001333, -0.00000077, 0.00001068, -0.00000540, -0.00001118, 0.00000947, 0.00001511, -0.00000212, -0.00000549, 0.00000528, 0.00000419, -0.00000264, -0.00000184, -0.00000200, -0.00000366, 0.00000031, 0.00000007, -0.00000634, -0.00000309, 0.00000477, 0.00000091, -0.00000381, 0.00000109, 0.00000192, -0.00000240, -0.00000053, 0.00000073, -0.00000069, 0.00000301, -0.00000183, -0.00001276, 0.00000374, 0.00002434, -0.00000125, -0.00002408, 0.00000540, 0.00001023, -0.00003009, -0.00000491, 0.00005696, 0.00001163, -0.00006273, -0.00001167, 0.00004603, -0.00001702, -0.00004749, 0.00002917, 0.00003978, -0.00003024, -0.00001659, 0.00003799, 0.00000347, -0.00003580, 0.00000275, 0.00002262, -0.00000916, -0.00001438, 0.00000008, -0.00000558, -0.00000532, -0.00000186, -0.00001265, -0.00000052, 0.00002134, -0.00000304, -0.00002606, 0.00000293, 0.00001247, -0.00001954, -0.00001438, 0.00001402, 0.00000169, -0.00001183, -0.00000069, -0.00001264, -0.00002733, -0.00001237, -0.00002078, -0.00003505, 0.00000931, 0.00000317, -0.00013096, -0.00017178, 0.00001813, 0.00018181, 0.00012881, 0.00004447, 0.00007069, 0.00008466, 0.00003327, 0.00003242, 0.00010220, 0.00013262, 0.00007234, -0.00002839, -0.00009361, -0.00008356, -0.00003304, -0.00000743, -0.00001399, -0.00001676, -0.00000905, -0.00000665, -0.00000884, -0.00000823, -0.00000662, },
		{ -0.00000715, -0.00000642, -0.00000618, -0.00000543, -0.00000295, -0.00000210, -0.00000514, -0.00000752, -0.00000803, -0.00000882, -0.00000472, 0.00000349, 0.00000303, -0.00000360, -0.00000237, 0.00000049, -0.00000310, -0.00000153, 0.00000101, -0.00001201, -0.00001969, -0.00000034, 0.00001189, -0.00000515, -0.00001111, 0.00000723, 0.00000662, -0.00001213, -0.00000461, 0.00001626, 0.00000895, -0.00000737, -0.00000077, 0.00000774, 0.00000013, -0.00000398, -0.00000015, -0.00000258, -0.00000429, 0.00000110, -0.00000006, -0.00000563, -0.00000145, 0.00000332, -0.00000052, -0.00000144, 0.00000073, -0.00000172, 0.00000012, 0.00000381, -0.00000321, -0.00000383, 0.00000729, -0.00000259, -0.00001711, 0.00000979, 0.00002940, -0.00001018, -0.00002897, 0.00001151, 0.00000930, -0.00003166, 0.00000624, 0.00005752, -0.00000544, -0.00006267, 0.00000370, 0.00004120, -0.00002780, -0.00003721, 0.00003451, 0.00002822, -0.00002701, -0.00000530, 0.00002761, -0.00000514, -0.00002243, 0.00000520, 0.00000887, -0.00000599, -0.00000424, -0.00000706, -0.00000938, 0.00000363, -0.00000367, -0.00001966, 0.00000474, 0.00002338, -0.00000882, -0.00002206, 0.00000709, 0.00000555, -0.00001851, -0.00000735, 0.00000681, -0.00000311, -0.00000211, -0.00000308, -0.00002324, -0.00001810, -0.00000594, -0.00003527, -0.00003266, 0.00002227, -0.00002471, -0.00015831, -0.00012198, 0.00008973, 0.00017897, 0.00008388, 0.00003057, 0.00007398, 0.00007514, 0.00002668, 0.00003169, 0.00009019, 0.00011939, 0.00007496, -0.00001809, -0.00008661, -0.00007989, -0.00003334, -0.00000990, -0.00001339, -0.00001503, -0.00001006, -0.00000746, -0.00000747, },
		{ -0.00000786, -0.00000644, -0.00000613, -0.00000581, -0.00000594, -0.00000518, -0.00000321, -0.00000381, -0.00000583, -0.00000539, -0.00000690, -0.00000994, -0.00000418, 0.00000466, 0.00000177, -0.00000416, -0.00000142, -0.00000051, -0.00000288, 0.00000127, -0.00000183, -0.00001784, -0.00001526, 0.00000741, 0.00000704, -0.00001161, -0.00000419, 0.00001254, -0.00000128, -0.00001497, 0.00000493, 0.00001860, 0.00000055, -0.00000872, 0.00000507, 0.00000721, -0.00000415, -0.00000355, 0.00000112, -0.00000367, -0.00000422, 0.00000157, -0.00000034, -0.00000393, -0.00000054, 0.00000065, -0.00000040, 0.00000153, -0.00000195, -0.00000475, 0.00000549, 0.00000630, -0.00000982, -0.00000437, 0.00001287, -0.00000615, -0.00002043, 0.00001860, 0.00003121, -0.00002170, -0.00003031, 0.00001886, 0.00000673, -0.00003103, 0.00001698, 0.00005236, -0.00002105, -0.00005611, 0.00001637, 0.00003188, -0.00003414, -0.00002547, 0.00003493, 0.00001872, -0.00002016, 0.00000052, 0.00001584, -0.00000675, -0.00001066, 0.00000134, -0.00000041, 0.00000100, -0.00000016, -0.00001436, -0.00000729, 0.00000939, -0.00000935, -0.00002155, 0.00001069, 0.00002042, -0.00001211, -0.00001512, 0.00000693, -0.00000102, -0.00001353, -0.00000405, -0.00000197, -0.00000235, 0.00000519, -0.00001064, -0.00002725, -0.00000661, -0.00000856, -0.00004624, -0.00002139, 0.00002213, -0.00006144, -0.00016012, -0.00005458, 0.00013644, 0.00015233, 0.00004906, 0.00002931, 0.00007283, 0.00006309, 0.00002412, 0.00003121, 0.00007835, 0.00010895, 0.00007623, -0.00001094, -0.00007859, -0.00007494, -0.00003486, -0.00001251, -0.00001204, -0.00001370, -0.00001114, },
		{ -0.00001280, -0.00001181, -0.00000817, -0.00000582, -0.00000488, -0.00000502, -0.00000591, -0.00000486, -0.00000375, -0.00000580, -0.00000572, -0.00000309, -0.00000674, -0.00001061, -0.00000275, 0.00000472, -0.00000003, -0.00000352, -0.00000075, -0.00000210, -0.00000137, 0.00000328, -0.00000714, -0.00002117, -0.00000719, 0.00001160, -0.00000152, -0.00001378, 0.00000546, 0.00001291, -0.00001049, -0.00001247, 0.00001457, 0.00001571, -0.00000734, -0.00000597, 0.00000986, 0.00000397, -0.00000721, -0.00000193, 0.00000140, -0.00000460, -0.00000359, 0.00000131, -0.00000020, -0.00000140, -0.00000115, -0.00000248, 0.00000188, 0.00000385, -0.00000696, -0.00000555, 0.00001290, 0.00000531, -0.00001763, -0.00000118, 0.00001802, -0.00001261, -0.00002101, 0.00002905, 0.00002837, -0.00003400, -0.00002722, 0.00002606, 0.00000275, -0.00002804, 0.00002563, 0.00004243, -0.00003272, -0.00004476, 0.00002441, 0.00002034, -0.00003556, -0.00001455, 0.00003162, 0.00001268, -0.00001238, 0.00000098, 0.00000603, -0.00000312, -0.00000328, -0.00000569, -0.00000395, 0.00000809, -0.00000133, -0.00001870, -0.00000193, 0.00001034, -0.00001542, -0.00001868, 0.00001427, 0.00001461, -0.00001142, -0.00000828, 0.00000307, -0.00000472, -0.00000712, -0.00000528, -0.00000891, 0.00000232, 0.00000729, -0.00001933, -0.00002403, 0.00000179, -0.00001757, -0.00004888, -0.00000829, 0.00000600, -0.00009449, -0.00013402, 0.00001447, 0.00015384, 0.00011500, 0.00002867, 0.00003359, 0.00006640, 0.00005185, 0.00002415, 0.00002987, 0.00006824, 0.00010063, 0.00007511, -0.00000557, -0.00006931, -0.00006996, -0.00003715, -0.00001457, -0.00001060, },
		{ -0.00001606, -0.00000961, -0.00001196, -0.00001197, -0.00000876, -0.00000539, -0.00000330, -0.00000435, -0.00000591, -0.00000444, -0.00000484, -0.00000761, -0.00000462, -0.00000134, -0.00000737, -0.00001014, -0.00000110, 0.00000328, -0.00000145, -0.00000181, -0.00000105, -0.00000357, 0.00000139, 0.00000322, -0.00001378, -0.00002030, 0.00000234, 0.00001046, -0.00001089, -0.00001030, 0.00001462, 0.00000784, -0.00001773, -0.00000500, 0.00002123, 0.00000859, -0.00001218, -0.00000038, 0.00001187, -0.00000068, -0.00000826, -0.00000017, 0.00000071, -0.00000474, -0.00000293, 0.00000027, 0.00000104, 0.00000133, -0.00000374, -0.00000488, 0.00000622, 0.00000407, -0.00001337, -0.00000284, 0.00002062, -0.00000002, -0.00002448, 0.00000582, 0.00002076, -0.00002099, -0.00001758, 0.00003917, 0.00002032, -0.00004476, -0.00001973, 0.00003165, -0.00000202, -0.00002297, 0.00003085, 0.00002955, -0.00003889, -0.00003111, 0.00002706, 0.00000898, -0.00003269, -0.00000605, 0.00002638, 0.00001018, -0.00000604, -0.00000225, 0.00000027, 0.00000278, -0.00000117, -0.00001245, -0.00000261, 0.00001246, -0.00000537, -0.00001870, 0.00000353, 0.00000696, -0.00001906, -0.00001314, 0.00001406, 0.00000879, -0.00000722, -0.00000391, -0.00000239, -0.00000473, -0.00000211, -0.00000970, -0.00001178, 0.00000771, 0.00000392, -0.00002505, -0.00001601, 0.00000374, -0.00002778, -0.00004256, -0.00000133, -0.00002261, -0.00011232, -0.00008576, 0.00007084, 0.00014557, 0.00007877, 0.00002132, 0.00003768, 0.00005651, 0.00004310, 0.00002487, 0.00002793, 0.00006073, 0.00009309, 0.00007163, -0.00000068, -0.00005944, -0.00006563, -0.00003939, },
		{ -0.00006187, -0.00004092, -0.00001741, -0.00000914, -0.00001076, -0.00001190, -0.00000974, -0.00000479, -0.00000161, -0.00000395, -0.00000566, -0.00000419, -0.00000657, -0.00000863, -0.00000281, -0.00000077, -0.00000809, -0.00000828, -0.00000022, 0.00000060, -0.00000162, 0.00000028, -0.00000256, -0.00000401, 0.00000458, 0.00000029, -0.00001970, -0.00001472, 0.00001030, 0.00000398, -0.00001766, -0.00000188, 0.00002001, -0.00000117, -0.00002023, 0.00000524, 0.00002275, -0.00000033, -0.00001255, 0.00000580, 0.00001051, -0.00000501, -0.00000739, 0.00000071, -0.00000034, -0.00000383, -0.00000298, -0.00000112, 0.00000379, 0.00000324, -0.00000816, -0.00000528, 0.00001174, 0.00000118, -0.00001949, 0.00000382, 0.00002636, -0.00000932, -0.00002806, 0.00001546, 0.00001950, -0.00002942, -0.00000974, 0.00004654, 0.00000773, -0.00005164, -0.00000884, 0.00003446, -0.00000670, -0.00001659, 0.00003191, 0.00001600, -0.00003920, -0.00001779, 0.00002482, -0.00000025, -0.00002695, -0.00000064, 0.00002098, 0.00001022, -0.00000251, -0.00000674, -0.00000105, 0.00000797, -0.00000331, -0.00001644, 0.00000135, 0.00001296, -0.00000941, -0.00001497, 0.00000656, 0.00000131, -0.00001900, -0.00000759, 0.00001047, 0.00000518, -0.00000138, -0.00000288, -0.00000703, -0.00000205, -0.00000034, -0.00001477, -0.00001041, 0.00001070, -0.00000279, -0.00002545, -0.00000739, -0.00000090, -0.00003394, -0.00003092, -0.00000597, -0.00005516, -0.00010821, -0.00002675, 0.00010571, 0.00012061, 0.00005108, 0.00002236, 0.00003868, 0.00004582, 0.00003678, 0.00002497, 0.00002638, 0.00005561, 0.00008521, 0.00006675, 0.00000431, -0.00004999, },
		{ 0.00000916, -0.00004169, -0.00005811, -0.00004162, -0.00001907, -0.00000889, -0.00000906, -0.00001201, -0.00001086, -0.00000384, -0.00000020, -0.00000367, -0.00000500, -0.00000455, -0.00000860, -0.00000835, -0.00000105, -0.00000157, -0.00000798, -0.00000546, -0.00000092, -0.00000240, -0.00000017, 0.00000173, -0.00000483, -0.00000279, 0.00000678, -0.00000531, -0.00002255, -0.00000557, 0.00001382, -0.00000588, -0.00001907, 0.00000883, 0.00001949, -0.00001099, -0.00001674, 0.00001494, 0.00001868, -0.00000799, -0.00000865, 0.00001019, 0.00000644, -0.00000751, -0.00000553, 0.00000018, -0.00000084, -0.00000215, -0.00000431, -0.00000203, 0.00000798, 0.00000325, -0.00001356, -0.00000279, 0.00001690, -0.00000504, -0.00002330, 0.00001365, 0.00002798, -0.00002095, -0.00002671, 0.00002558, 0.00001370, -0.00003557, 0.00000161, 0.00004900, -0.00000751, -0.00005295, 0.00000360, 0.00003395, -0.00001037, -0.00000996, 0.00002887, 0.00000402, -0.00003443, -0.00000695, 0.00001910, -0.00000620, -0.00002001, 0.00000189, 0.00001660, 0.00001126, -0.00000191, -0.00001023, 0.00000089, 0.00001040, -0.00000749, -0.00001670, 0.00000530, 0.00001014, -0.00001127, -0.00000943, 0.00000614, -0.00000410, -0.00001568, -0.00000416, 0.00000525, 0.00000467, 0.00000386, -0.00000447, -0.00000918, 0.00000127, -0.00000198, -0.00001798, -0.00000650, 0.00000961, -0.00000951, -0.00002076, -0.00000207, -0.00000926, -0.00003320, -0.00002015, -0.00002289, -0.00008130, -0.00008222, 0.00003000, 0.00011729, 0.00008938, 0.00003425, 0.00002658, 0.00003642, 0.00003632, 0.00003185, 0.00002424, 0.00002602, 0.00005186, 0.00007666, 0.00006161, },
		{ 0.00006798, 0.00005696, 0.00001314, -0.00003459, -0.00005377, -0.00004185, -0.00002117, -0.00000836, -0.00000717, -0.00001254, -0.00001169, -0.00000270, 0.00000066, -0.00000312, -0.00000414, -0.00000584, -0.00001022, -0.00000676, -0.00000024, -0.00000322, -0.00000634, -0.00000270, -0.00000343, -0.00000451, 0.00000247, 0.00000172, -0.00000683, -0.00000008, 0.00000655, -0.00001212, -0.00002067, 0.00000448, 0.00001136, -0.00001571, -0.00001417, 0.00001813, 0.00001301, -0.00001803, -0.00000810, 0.00002087, 0.00001049, -0.00001183, -0.00000220, 0.00001113, 0.00000136, -0.00000752, -0.00000400, -0.00000153, 0.00000002, -0.00000051, -0.00000703, -0.00000150, 0.00001288, 0.00000062, -0.00001851, 0.00000268, 0.00001990, -0.00001370, -0.00002307, 0.00002474, 0.00002421, -0.00003230, -0.00002000, 0.00003357, 0.00000415, -0.00003741, 0.00001442, 0.00004528, -0.00002265, -0.00004812, 0.00001538, 0.00003031, -0.00001236, -0.00000430, 0.00002258, -0.00000465, -0.00002626, 0.00000009, 0.00001176, -0.00000864, -0.00001334, 0.00000236, 0.00001361, 0.00001189, -0.00000324, -0.00001131, 0.00000410, 0.00000948, -0.00001132, -0.00001387, 0.00000731, 0.00000576, -0.00001013, -0.00000436, 0.00000282, -0.00000732, -0.00001079, -0.00000367, 0.00000048, 0.00000664, 0.00000692, -0.00000698, -0.00000858, 0.00000318, -0.00000569, -0.00001801, -0.00000261, 0.00000482, -0.00001334, -0.00001344, -0.00000194, -0.00001700, -0.00002625, -0.00001627, -0.00004768, -0.00009255, -0.00004072, 0.00007391, 0.00010983, 0.00006045, 0.00002660, 0.00003024, 0.00003229, 0.00002872, 0.00002730, 0.00002326, 0.00002687, 0.00004839, },
		{ 0.00002820, 0.00004468, 0.00006009, 0.00005276, 0.00001560, -0.00002821, -0.00004872, -0.00004216, -0.00002337, -0.00000731, -0.00000558, -0.00001332, -0.00001188, -0.00000179, 0.00000102, -0.00000199, -0.00000360, -0.00000798, -0.00001064, -0.00000444, -0.00000099, -0.00000471, -0.00000315, -0.00000121, -0.00000710, -0.00000480, 0.00000524, 0.00000015, -0.00000742, 0.00000308, 0.00000308, -0.00001777, -0.00001384, 0.00001208, 0.00000344, -0.00002182, -0.00000429, 0.00002256, 0.00000269, -0.00001946, 0.00000296, 0.00002100, 0.00000107, -0.00001065, 0.00000417, 0.00000830, -0.00000271, -0.00000540, -0.00000401, -0.00000352, 0.00000257, 0.00000005, -0.00001064, 0.00000120, 0.00001727, -0.00000470, -0.00002143, 0.00001031, 0.00001932, -0.00002299, -0.00001800, 0.00003448, 0.00001513, -0.00004053, -0.00000899, 0.00003718, -0.00000714, -0.00003391, 0.00002589, 0.00003554, -0.00003470, -0.00003791, 0.00002444, 0.00002444, -0.00001244, -0.00000059, 0.00001447, -0.00000908, -0.00001676, 0.00000303, 0.00000462, -0.00000813, -0.00000790, 0.00000171, 0.00001170, 0.00001129, -0.00000502, -0.00000980, 0.00000655, 0.00000599, -0.00001308, -0.00000957, 0.00000665, 0.00000183, -0.00000657, -0.00000145, -0.00000178, -0.00000763, -0.00000630, -0.00000559, -0.00000233, 0.00000959, 0.00000741, -0.00000865, -0.00000625, 0.00000261, -0.00000941, -0.00001516, -0.00000092, -0.00000164, -0.00001304, -0.00000683, -0.00000626, -0.00002050, -0.00001694, -0.00002245, -0.00007255, -0.00008517, 0.00000627, 0.00009946, 0.00009099, 0.00003862, 0.00002452, 0.00003177, 0.00002787, 0.00002266, 0.00002280, 0.00002272, },
		{ 0.00001875, 0.00002275, 0.00002901, 0.00004098, 0.00005370, 0.00004839, 0.00001646, -0.00002193, -0.00004336, -0.00004279, -0.00002505, -0.00000594, -0.00000474, -0.00001386, -0.00001144, -0.00000157, 0.00000128, -0.00000028, -0.00000395, -0.00001035, -0.00000942, -0.00000242, -0.00000322, -0.00000496, 0.00000080, -0.00000180, -0.00001065, -0.00000310, 0.00000686, -0.00000227, -0.00000597, 0.00000502, -0.00000316, -0.00001978, -0.00000370, 0.00001440, -0.00000738, -0.00002160, 0.00000729, 0.00002025, -0.00000784, -0.00001435, 0.00001277, 0.00001525, -0.00000627, -0.00000508, 0.00000791, 0.00000278, -0.00000416, -0.00000246, -0.00000609, -0.00000456, 0.00000639, -0.00000132, -0.00001408, 0.00000615, 0.00001975, -0.00001192, -0.00002106, 0.00001843, 0.00001462, -0.00003061, -0.00000860, 0.00004028, 0.00000234, -0.00004344, 0.00000397, 0.00003527, -0.00001746, -0.00002546, 0.00003327, 0.00002143, -0.00004125, -0.00002425, 0.00002940, 0.00001767, -0.00001094, 0.00000060, 0.00000623, -0.00000930, -0.00000791, 0.00000250, -0.00000098, -0.00000564, -0.00000405, 0.00000069, 0.00001019, 0.00000947, -0.00000587, -0.00000660, 0.00000694, 0.00000152, -0.00001223, -0.00000564, 0.00000388, -0.00000021, -0.00000211, -0.00000124, -0.00000580, -0.00000559, -0.00000365, -0.00000850, -0.00000276, 0.00001188, 0.00000609, -0.00000847, -0.00000382, -0.00000025, -0.00001140, -0.00001100, -0.00000225, -0.00000720, -0.00000942, -0.00000353, -0.00001220, -0.00001849, -0.00001046, -0.00003765, -0.00008931, -0.00006109, 0.00004856, 0.00010679, 0.00006870, 0.00002482, 0.00002453, 0.00003133, 0.00002401, 0.00001754, },
		{ 0.00002073, 0.00001312, 0.00001574, 0.00002280, 0.00002870, 0.00003811, 0.00004882, 0.00004317, 0.00001625, -0.00001543, -0.00003835, -0.00004350, -0.00002576, -0.00000481, -0.00000466, -0.00001363, -0.00001076, -0.00000217, 0.00000196, 0.00000158, -0.00000548, -0.00001200, -0.00000676, -0.00000174, -0.00000611, -0.00000332, 0.00000418, -0.00000441, -0.00001267, -0.00000019, 0.00000651, -0.00000422, -0.00000278, 0.00000417, -0.00001037, -0.00001662, 0.00000666, 0.00001024, -0.00001720, -0.00001460, 0.00001645, 0.00001164, -0.00001467, -0.00000406, 0.00001796, 0.00000557, -0.00000896, 0.00000271, 0.00000743, -0.00000334, -0.00000243, -0.00000036, -0.00000985, -0.00000357, 0.00001045, -0.00000493, -0.00001607, 0.00001274, 0.00001924, -0.00001954, -0.00001698, 0.00002505, 0.00000641, -0.00003445, 0.00000329, 0.00004041, -0.00001142, -0.00004010, 0.00001589, 0.00002824, -0.00002422, -0.00001390, 0.00003468, 0.00000578, -0.00004108, -0.00000978, 0.00002989, 0.00001135, -0.00000857, -0.00000069, -0.00000056, -0.00000617, -0.00000119, -0.00000032, -0.00000431, -0.00000228, -0.00000172, -0.00000033, 0.00000845, 0.00000708, -0.00000505, -0.00000310, 0.00000503, -0.00000224, -0.00000933, -0.00000338, 0.00000034, 0.00000003, 0.00000160, -0.00000313, -0.00000792, -0.00000256, -0.00000326, -0.00001086, -0.00000143, 0.00001249, 0.00000431, -0.00000652, -0.00000269, -0.00000412, -0.00001101, -0.00000749, -0.00000588, -0.00001001, -0.00000474, -0.00000425, -0.00001640, -0.00001266, -0.00001094, -0.00005700, -0.00009226, -0.00002677, 0.00007884, 0.00010028, 0.00004877, 0.00001741, 0.00002450, 0.00002978, },
		{ 0.00002378, 0.00002773, 0.00001767, 0.00000977, 0.00001398, 0.00002190, 0.00002746, 0.00003680, 0.00004478, 0.00003684, 0.00001581, -0.00000900, -0.00003423, -0.00004364, -0.00002546, -0.00000456, -0.00000492, -0.00001238, -0.00001044, -0.00000331, 0.00000337, 0.00000283, -0.00000790, -0.00001206, -0.00000357, -0.00000290, -0.00000837, -0.00000007, 0.00000563, -0.00000803, -0.00001228, 0.00000252, 0.00000426, -0.00000439, 0.00000079, -0.00000017, -0.00001591, -0.00000845, 0.00001367, 0.00000068, -0.00002222, -0.00000281, 0.00001972, -0.00000052, -0.00001508, 0.00000821, 0.00001669, -0.00000473, -0.00000606, 0.00000977, 0.00000265, -0.00000775, 0.00000186, -0.00000052, -0.00001412, -0.00000008, 0.00001337, -0.00001032, -0.00001554, 0.00001965, 0.00001535, -0.00002578, -0.00000974, 0.00002840, -0.00000362, -0.00003330, 0.00001500, 0.00003458, -0.00002304, -0.00003128, 0.00002405, 0.00001791, -0.00002580, -0.00000204, 0.00002968, -0.00000815, -0.00003451, 0.00000286, 0.00002652, 0.00000650, -0.00000629, -0.00000385, -0.00000480, -0.00000111, 0.00000267, -0.00000399, -0.00000532, 0.00000105, -0.00000058, -0.00000135, 0.00000619, 0.00000503, -0.00000268, -0.00000066, 0.00000154, -0.00000418, -0.00000565, -0.00000314, -0.00000256, 0.00000196, 0.00000345, -0.00000578, -0.00000781, -0.00000003, -0.00000454, -0.00001167, 0.00000043, 0.00001135, 0.00000328, -0.00000375, -0.00000342, -0.00000746, -0.00000880, -0.00000600, -0.00001002, -0.00000971, -0.00000148, -0.00000764, -0.00001668, -0.00000675, -0.00001960, -0.00007372, -0.00008013, 0.00000932, 0.00009428, 0.00008617, 0.00003395, 0.00001387, },
		{ 0.00002433, 0.00001224, 0.00002254, 0.00002518, 0.00001474, 0.00000813, 0.00001293, 0.00001935, 0.00002613, 0.00003721, 0.00004062, 0.00002988, 0.00001577, -0.00000339, -0.00003101, -0.00004249, -0.00002467, -0.00000551, -0.00000482, -0.00001037, -0.00001092, -0.00000436, 0.00000533, 0.00000281, -0.00001033, -0.00001020, -0.00000111, -0.00000554, -0.00000881, 0.00000365, 0.00000450, -0.00001106, -0.00000961, 0.00000360, 0.00000117, -0.00000219, 0.00000283, -0.00000723, -0.00001724, 0.00000261, 0.00001461, -0.00001110, -0.00002012, 0.00000993, 0.00001556, -0.00001214, -0.00000851, 0.00001847, 0.00000928, -0.00001218, 0.00000144, 0.00001337, -0.00000501, -0.00000872, 0.00000709, -0.00000357, -0.00001734, 0.00000552, 0.00001397, -0.00001634, -0.00001207, 0.00002530, 0.00000861, -0.00002906, -0.00000085, 0.00002753, -0.00001316, -0.00002731, 0.00002377, 0.00002406, -0.00002990, -0.00001918, 0.00002679, 0.00000705, -0.00002204, 0.00000712, 0.00001952, -0.00001749, -0.00002329, 0.00001167, 0.00002069, 0.00000353, -0.00000493, -0.00000782, -0.00000606, 0.00000430, 0.00000370, -0.00000725, -0.00000444, 0.00000375, -0.00000029, -0.00000254, 0.00000360, 0.00000406, 0.00000046, -0.00000004, -0.00000229, -0.00000405, -0.00000251, -0.00000439, -0.00000387, 0.00000437, 0.00000324, -0.00000783, -0.00000607, 0.00000107, -0.00000636, -0.00001081, 0.00000167, 0.00000919, 0.00000355, -0.00000138, -0.00000555, -0.00000916, -0.00000611, -0.00000672, -0.00001285, -0.00000743, -0.00000108, -0.00001115, -0.00001310, -0.00000480, -0.00003414, -0.00008171, -0.00005612, 0.00003964, 0.00009630, 0.00007008, },
		{ 0.00008901, 0.00005558, 0.00001860, 0.00001157, 0.00002099, 0.00002182, 0.00001243, 0.00000854, 0.00001154, 0.00001519, 0.00002573, 0.00003866, 0.00003571, 0.00002324, 0.00001617, 0.00000060, -0.00002809, -0.00003972, -0.00002421, -0.00000741, -0.00000379, -0.00000824, -0.00001220, -0.00000469, 0.00000716, 0.00000135, -0.00001163, -0.00000690, -0.00000043, -0.00000851, -0.00000701, 0.00000625, 0.00000117, -0.00001198, -0.00000581, 0.00000230, -0.00000115, 0.00000179, 0.00000173, -0.00001486, -0.00001303, 0.00001310, 0.00000875, -0.00002090, -0.00001100, 0.00001923, 0.00000498, -0.00001920, 0.00000324, 0.00002336, -0.00000194, -0.00001423, 0.00001102, 0.00001197, -0.00001313, -0.00000568, 0.00001129, -0.00000909, -0.00001814, 0.00001207, 0.00001164, -0.00002150, -0.00000602, 0.00002829, 0.00000035, -0.00002856, 0.00000768, 0.00002266, -0.00001995, -0.00001795, 0.00002755, 0.00001143, -0.00003071, -0.00000680, 0.00002404, -0.00000155, -0.00001436, 0.00001133, 0.00000677, -0.00002057, -0.00001009, 0.00001572, 0.00001407, 0.00000222, -0.00000496, -0.00001141, -0.00000463, 0.00000870, 0.00000258, -0.00000923, -0.00000237, 0.00000551, -0.00000062, -0.00000404, 0.00000120, 0.00000445, 0.00000331, -0.00000120, -0.00000524, -0.00000241, -0.00000076, -0.00000608, -0.00000348, 0.00000607, 0.00000159, -0.00000843, -0.00000380, 0.00000060, -0.00000761, -0.00000891, 0.00000171, 0.00000703, 0.00000483, -0.00000030, -0.00000800, -0.00000904, -0.00000432, -0.00000874, -0.00001341, -0.00000511, -0.00000332, -0.00001241, -0.00000793, -0.00000924, -0.00004964, -0.00007775, -0.00002639, 0.00005968, },
		{ 0.00000230, 0.00006862, 0.00007710, 0.00004396, 0.00001541, 0.00001171, 0.00001887, 0.00001753, 0.00001161, 0.00001058, 0.00000877, 0.00001031, 0.00002681, 0.00003997, 0.00003016, 0.00001790, 0.00001636, 0.00000264, -0.00002456, -0.00003564, -0.00002476, -0.00000955, -0.00000175, -0.00000673, -0.00001375, -0.00000406, 0.00000796, -0.00000103, -0.00001093, -0.00000335, -0.00000187, -0.00001032, -0.00000354, 0.00000639, -0.00000295, -0.00001009, -0.00000260, -0.00000101, -0.00000133, 0.00000592, -0.00000302, -0.00002021, -0.00000390, 0.00001935, -0.00000225, -0.00002495, 0.00000251, 0.00002175, -0.00000876, -0.00001915, 0.00001656, 0.00002125, -0.00001339, -0.00001017, 0.00001948, 0.00000567, -0.00001914, 0.00000060, 0.00001276, -0.00001577, -0.00001578, 0.00001799, 0.00000662, -0.00002444, 0.00000145, 0.00002786, -0.00000765, -0.00002446, 0.00001397, 0.00001511, -0.00002253, -0.00000767, 0.00002570, -0.00000026, -0.00002580, 0.00000296, 0.00001725, -0.00000592, -0.00000532, 0.00000975, -0.00000536, -0.00001731, 0.00000221, 0.00001531, 0.00000816, 0.00000189, -0.00000633, -0.00001358, -0.00000139, 0.00001124, 0.00000032, -0.00000959, 0.00000014, 0.00000628, -0.00000146, -0.00000577, -0.00000045, 0.00000588, 0.00000497, -0.00000344, -0.00000660, -0.00000026, -0.00000054, -0.00000722, -0.00000195, 0.00000638, -0.00000045, -0.00000749, -0.00000205, -0.00000088, -0.00000771, -0.00000688, 0.00000066, 0.00000563, 0.00000631, -0.00000064, -0.00000968, -0.00000778, -0.00000413, -0.00001064, -0.00001199, -0.00000435, -0.00000655, -0.00001052, -0.00000457, -0.00001966, -0.00006050, -0.00006249, },
		{ -0.00006266, -0.00003986, 0.00002484, 0.00006862, 0.00006415, 0.00003494, 0.00001405, 0.00001255, 0.00001558, 0.00001286, 0.00001301, 0.00001306, 0.00000424, 0.00000605, 0.00002902, 0.00004004, 0.00002483, 0.00001431, 0.00001534, 0.00000311, -0.00001968, -0.00003113, -0.00002649, -0.00001105, 0.00000082, -0.00000623, -0.00001478, -0.00000284, 0.00000707, -0.00000320, -0.00000809, -0.00000098, -0.00000474, -0.00000984, 0.00000011, 0.00000361, -0.00000600, -0.00000587, -0.00000148, -0.00000495, 0.00000109, 0.00000812, -0.00001033, -0.00002078, 0.00000760, 0.00001892, -0.00001471, -0.00002135, 0.00001613, 0.00001641, -0.00002126, -0.00001181, 0.00002735, 0.00001279, -0.00002151, -0.00000123, 0.00002403, -0.00000390, -0.00002118, 0.00000832, 0.00001071, -0.00002183, -0.00001048, 0.00002181, -0.00000007, -0.00002436, 0.00000877, 0.00002414, -0.00001372, -0.00001788, 0.00001680, 0.00000694, -0.00002065, 0.00000091, 0.00001910, -0.00000832, -0.00001705, 0.00000819, 0.00000895, -0.00000549, 0.00000222, 0.00000320, -0.00001394, -0.00000916, 0.00001131, 0.00001169, 0.00000384, 0.00000169, -0.00000851, -0.00001376, 0.00000251, 0.00001166, -0.00000204, -0.00000842, 0.00000245, 0.00000613, -0.00000272, -0.00000742, -0.00000098, 0.00000762, 0.00000503, -0.00000577, -0.00000629, 0.00000144, -0.00000136, -0.00000721, -0.00000022, 0.00000529, -0.00000192, -0.00000559, -0.00000138, -0.00000250, -0.00000673, -0.00000550, -0.00000083, 0.00000519, 0.00000713, -0.00000187, -0.00000999, -0.00000645, -0.00000530, -0.00001130, -0.00000984, -0.00000563, -0.00000866, -0.00000654, -0.00000584, -0.00003293, },
		{ -0.00001257, -0.00004443, -0.00005503, -0.00001535, 0.00003906, 0.00006312, 0.00005192, 0.00002786, 0.00001445, 0.00001358, 0.00001068, 0.00000897, 0.00001658, 0.00001451, -0.00000142, 0.00000352, 0.00003120, 0.00003845, 0.00002092, 0.00001207, 0.00001238, 0.00000301, -0.00001334, -0.00002722, -0.00002889, -0.00001138, 0.00000308, -0.00000656, -0.00001463, -0.00000192, 0.00000450, -0.00000390, -0.00000391, -0.00000079, -0.00000759, -0.00000686, 0.00000216, -0.00000129, -0.00000647, -0.00000084, -0.00000304, -0.00000770, 0.00000524, 0.00000680, -0.00001774, -0.00001558, 0.00001772, 0.00001163, -0.00002427, -0.00001083, 0.00002544, 0.00000477, -0.00002854, 0.00000058, 0.00003235, 0.00000060, -0.00002390, 0.00000985, 0.00002316, -0.00001417, -0.00001864, 0.00001530, 0.00000543, -0.00002563, -0.00000331, 0.00002258, -0.00000693, -0.00002124, 0.00001441, 0.00001809, -0.00001675, -0.00001055, 0.00001600, 0.00000024, -0.00001537, 0.00000572, 0.00000995, -0.00001126, -0.00000720, 0.00000845, 0.00000182, -0.00000128, 0.00000595, -0.00000609, -0.00001712, 0.00000138, 0.00001600, 0.00000657, 0.00000129, 0.00000094, -0.00001067, -0.00001192, 0.00000593, 0.00001029, -0.00000366, -0.00000615, 0.00000414, 0.00000520, -0.00000427, -0.00000857, -0.00000040, 0.00000885, 0.00000367, -0.00000729, -0.00000483, 0.00000213, -0.00000245, -0.00000609, 0.00000090, 0.00000338, -0.00000231, -0.00000356, -0.00000175, -0.00000352, -0.00000522, -0.00000502, -0.00000205, 0.00000532, 0.00000683, -0.00000309, -0.00000907, -0.00000590, -0.00000691, -0.00001048, -0.00000842, -0.00000806, -0.00000830, -0.00000303, },
		{ -0.00000568, -0.00000277, -0.00002311, -0.00004993, -0.00003969, 0.00000608, 0.00004563, 0.00005521, 0.00004069, 0.00002248, 0.00001662, 0.00001375, 0.00000447, 0.00000711, 0.00002129, 0.00001391, -0.00000675, 0.00000296, 0.00003191, 0.00003573, 0.00001932, 0.00001011, 0.00000743, 0.00000351, -0.00000618, -0.00002463, -0.00003100, -0.00001059, 0.00000425, -0.00000701, -0.00001322, -0.00000231, 0.00000105, -0.00000234, 0.00000015, -0.00000285, -0.00000880, -0.00000236, 0.00000134, -0.00000654, -0.00000389, 0.00000306, -0.00000657, -0.00000789, 0.00000922, 0.00000174, -0.00002227, -0.00000572, 0.00002289, -0.00000027, -0.00002741, 0.00000342, 0.00002743, -0.00000948, -0.00002842, 0.00001423, 0.00003024, -0.00001169, -0.00002003, 0.00001983, 0.00001703, -0.00002246, -0.00001225, 0.00001968, -0.00000181, -0.00002618, 0.00000415, 0.00002023, -0.00001248, -0.00001590, 0.00001735, 0.00001121, -0.00001653, -0.00000423, 0.00001246, -0.00000351, -0.00000869, 0.00000595, 0.00000097, -0.00000914, 0.00000100, 0.00000484, -0.00000214, 0.00000455, 0.00000488, -0.00001522, -0.00001462, 0.00001152, 0.00001633, 0.00000161, 0.00000009, -0.00000062, -0.00001198, -0.00000860, 0.00000804, 0.00000784, -0.00000410, -0.00000338, 0.00000497, 0.00000364, -0.00000584, -0.00000886, 0.00000086, 0.00000896, 0.00000156, -0.00000751, -0.00000301, 0.00000176, -0.00000312, -0.00000435, 0.00000102, 0.00000143, -0.00000169, -0.00000209, -0.00000266, -0.00000365, -0.00000386, -0.00000521, -0.00000259, 0.00000537, 0.00000554, -0.00000361, -0.00000759, -0.00000631, -0.00000798, -0.00000886, -0.00000852, -0.00000997, },
		{ -0.00000980, -0.00000987, -0.00000259, -0.00000731, -0.00003384, -0.00004730, -0.00002084, 0.00002158, 0.00004691, 0.00004662, 0.00003023, 0.00001919, 0.00002004, 0.00001203, -0.00000191, 0.00000794, 0.00002544, 0.00001130, -0.00001018, 0.00000355, 0.00003016, 0.00003315, 0.00002010, 0.00000723, 0.00000129, 0.00000536, 0.00000065, -0.00002333, -0.00003188, -0.00000937, 0.00000406, -0.00000666, -0.00001117, -0.00000457, -0.00000191, 0.00000135, 0.00000258, -0.00000614, -0.00000737, 0.00000188, -0.00000248, -0.00001012, 0.00000087, 0.00000435, -0.00001041, -0.00000531, 0.00001085, -0.00000558, -0.00002164, 0.00000582, 0.00002112, -0.00001273, -0.00002283, 0.00001689, 0.00002160, -0.00002190, -0.00002118, 0.00002509, 0.00002197, -0.00002062, -0.00001130, 0.00002599, 0.00000732, -0.00002685, -0.00000379, 0.00002044, -0.00000923, -0.00002341, 0.00001030, 0.00001544, -0.00001563, -0.00000966, 0.00001731, 0.00000506, -0.00001374, -0.00000014, 0.00000774, -0.00000386, -0.00000288, 0.00000219, -0.00000538, -0.00000347, 0.00000568, -0.00000050, -0.00000222, 0.00000944, -0.00000041, -0.00002144, -0.00000765, 0.00001896, 0.00001337, -0.00000209, -0.00000053, -0.00000279, -0.00001184, -0.00000463, 0.00000851, 0.00000515, -0.00000331, -0.00000068, 0.00000491, 0.00000166, -0.00000710, -0.00000815, 0.00000216, 0.00000774, -0.00000049, -0.00000646, -0.00000152, 0.00000074, -0.00000302, -0.00000265, 0.00000027, 0.00000007, -0.00000057, -0.00000148, -0.00000347, -0.00000310, -0.00000308, -0.00000554, -0.00000248, 0.00000479, 0.00000383, -0.00000321, -0.00000632, -0.00000723, -0.00000801, -0.00000761, },
		{ -0.00000730, -0.00000763, -0.00001099, -0.00000739, -0.00000156, -0.00001611, -0.00004076, -0.00003722, -0.00000299, 0.00003087, 0.00004543, 0.00003770, 0.00002064, 0.00001849, 0.00002344, 0.00000805, -0.00000675, 0.00001099, 0.00002736, 0.00000783, -0.00001076, 0.00000373, 0.00002595, 0.00003208, 0.00002234, 0.00000272, -0.00000466, 0.00000849, 0.00000597, -0.00002261, -0.00003110, -0.00000870, 0.00000291, -0.00000494, -0.00000955, -0.00000846, -0.00000307, 0.00000609, 0.00000251, -0.00000897, -0.00000341, 0.00000399, -0.00000813, -0.00001064, 0.00000597, 0.00000274, -0.00001262, -0.00000115, 0.00000875, -0.00001244, -0.00001532, 0.00001515, 0.00001288, -0.00002144, -0.00001195, 0.00002535, 0.00001008, -0.00002880, -0.00000936, 0.00003025, 0.00001027, -0.00002408, -0.00000040, 0.00002694, -0.00000345, -0.00002664, 0.00000459, 0.00001767, -0.00001513, -0.00001811, 0.00001399, 0.00000950, -0.00001604, -0.00000392, 0.00001483, 0.00000078, -0.00000965, 0.00000131, 0.00000358, -0.00000152, 0.00000030, -0.00000380, -0.00000769, 0.00000341, 0.00000627, -0.00000518, 0.00000084, 0.00001132, -0.00000803, -0.00002304, 0.00000152, 0.00002243, 0.00000877, -0.00000415, -0.00000127, -0.00000499, -0.00001013, -0.00000094, 0.00000747, 0.00000294, -0.00000157, 0.00000150, 0.00000401, -0.00000049, -0.00000769, -0.00000662, 0.00000282, 0.00000548, -0.00000180, -0.00000459, -0.00000077, -0.00000038, -0.00000225, -0.00000152, -0.00000089, -0.00000046, 0.00000047, -0.00000155, -0.00000374, -0.00000234, -0.00000294, -0.00000556, -0.00000215, 0.00000342, 0.00000234, -0.00000219, -0.00000564, -0.00000787, },
		{ -0.00000538, -0.00000771, -0.00000677, -0.00000892, -0.00001057, -0.00000369, -0.00000460, -0.00002651, -0.00004108, -0.00002278, 0.00001071, 0.00003561, 0.00004270, 0.00002816, 0.00001274, 0.00002035, 0.00002519, 0.00000253, -0.00000864, 0.00001471, 0.00002629, 0.00000534, -0.00000871, 0.00000195, 0.00002035, 0.00003327, 0.00002458, -0.00000322, -0.00000911, 0.00001200, 0.00000920, -0.00002136, -0.00002894, -0.00000931, 0.00000172, -0.00000196, -0.00000944, -0.00001290, -0.00000176, 0.00001019, 0.00000016, -0.00000967, 0.00000176, 0.00000288, -0.00001354, -0.00000795, 0.00000936, -0.00000074, -0.00001199, 0.00000244, 0.00000310, -0.00001586, -0.00000499, 0.00001898, 0.00000101, -0.00002338, 0.00000156, 0.00002639, -0.00000318, -0.00002853, 0.00000319, 0.00002882, -0.00000135, -0.00002177, 0.00000959, 0.00002287, -0.00001276, -0.00002244, 0.00001100, 0.00001238, -0.00001832, -0.00001165, 0.00001480, 0.00000381, -0.00001407, 0.00000021, 0.00001097, -0.00000118, -0.00000565, 0.00000061, 0.00000122, 0.00000197, 0.00000007, -0.00000971, -0.00000593, 0.00000910, 0.00000364, -0.00000741, 0.00000524, 0.00000928, -0.00001540, -0.00001970, 0.00001033, 0.00002189, 0.00000412, -0.00000485, -0.00000253, -0.00000656, -0.00000714, 0.00000174, 0.00000543, 0.00000165, 0.00000068, 0.00000289, 0.00000247, -0.00000246, -0.00000740, -0.00000472, 0.00000244, 0.00000277, -0.00000203, -0.00000251, -0.00000073, -0.00000116, -0.00000120, -0.00000113, -0.00000190, -0.00000032, 0.00000099, -0.00000187, -0.00000342, -0.00000181, -0.00000313, -0.00000512, -0.00000210, 0.00000156, 0.00000147, -0.00000113, },
		{ 0.00000117, -0.00000042, -0.00000501, -0.00000684, -0.00000732, -0.00001052, -0.00000777, -0.00000101, -0.00001206, -0.00003482, -0.00003441, -0.00000803, 0.00001939, 0.00003798, 0.00003888, 0.00001802, 0.00000773, 0.00002373, 0.00002405, -0.00000279, -0.00000716, 0.00001702, 0.00002275, 0.00000547, -0.00000541, -0.00000257, 0.00001507, 0.00003642, 0.00002542, -0.00000946, -0.00001138, 0.00001456, 0.00001057, -0.00001868, -0.00002632, -0.00001130, 0.00000148, 0.00000142, -0.00001129, -0.00001635, 0.00000161, 0.00001214, -0.00000313, -0.00000746, 0.00000611, -0.00000121, -0.00001665, -0.00000332, 0.00000978, -0.00000412, -0.00000866, 0.00000333, -0.00000421, -0.00001392, 0.00000597, 0.00001596, -0.00001020, -0.00001801, 0.00001314, 0.00002023, -0.00001389, -0.00002192, 0.00001276, 0.00002203, -0.00000985, -0.00001516, 0.00001627, 0.00001532, -0.00001876, -0.00001577, 0.00001434, 0.00000614, -0.00001847, -0.00000551, 0.00001306, -0.00000052, -0.00001066, 0.00000223, 0.00000695, -0.00000112, -0.00000280, -0.00000111, 0.00000110, 0.00000485, -0.00000310, -0.00001352, -0.00000134, 0.00001193, -0.00000040, -0.00000653, 0.00000879, 0.00000380, -0.00002014, -0.00001253, 0.00001665, 0.00001832, 0.00000059, -0.00000483, -0.00000427, -0.00000696, -0.00000353, 0.00000303, 0.00000303, 0.00000139, 0.00000294, 0.00000340, 0.00000054, -0.00000394, -0.00000623, -0.00000306, 0.00000097, 0.00000029, -0.00000126, -0.00000073, -0.00000113, -0.00000140, -0.00000030, -0.00000129, -0.00000240, 0.00000007, 0.00000094, -0.00000201, -0.00000276, -0.00000172, -0.00000326, -0.00000439, -0.00000261, -0.00000030, },
		{ -0.00000362, -0.00000173, 0.00000109, -0.00000009, -0.00000399, -0.00000596, -0.00000924, -0.00001094, -0.00000315, -0.00000167, -0.00002218, -0.00003783, -0.00002287, 0.00000368, 0.00002426, 0.00003946, 0.00003329, 0.00000821, 0.00000642, 0.00002671, 0.00001990, -0.00000590, -0.00000330, 0.00001630, 0.00001847, 0.00000885, -0.00000282, -0.00000934, 0.00001161, 0.00004023, 0.00002422, -0.00001442, -0.00001179, 0.00001505, 0.00001107, -0.00001429, -0.00002433, -0.00001402, 0.00000267, 0.00000393, -0.00001460, -0.00001749, 0.00000567, 0.00001129, -0.00000547, -0.00000281, 0.00000781, -0.00000671, -0.00001632, 0.00000115, 0.00000729, -0.00000550, -0.00000420, 0.00000044, -0.00001020, -0.00000671, 0.00001367, 0.00000730, -0.00001668, -0.00000751, 0.00001903, 0.00000959, -0.00001891, -0.00001183, 0.00001702, 0.00001263, -0.00001355, -0.00000674, 0.00001848, 0.00000656, -0.00002077, -0.00000856, 0.00001445, 0.00000052, -0.00001603, -0.00000089, 0.00000969, -0.00000288, -0.00000693, 0.00000230, 0.00000377, 0.00000016, -0.00000152, -0.00000263, 0.00000269, 0.00000573, -0.00000775, -0.00001406, 0.00000408, 0.00001145, -0.00000385, -0.00000318, 0.00000973, -0.00000342, -0.00002075, -0.00000358, 0.00001933, 0.00001323, -0.00000141, -0.00000479, -0.00000608, -0.00000600, -0.00000005, 0.00000298, 0.00000090, 0.00000199, 0.00000481, 0.00000308, -0.00000149, -0.00000462, -0.00000446, -0.00000215, -0.00000129, -0.00000141, 0.00000016, 0.00000045, -0.00000159, -0.00000123, 0.00000016, -0.00000164, -0.00000234, 0.00000032, 0.00000055, -0.00000178, -0.00000213, -0.00000196, -0.00000308, -0.00000378, },
		{ -0.00000262, -0.00000362, -0.00000480, -0.00000262, 0.00000085, 0.00000028, -0.00000226, -0.00000602, -0.00001184, -0.00000901, 0.00000141, -0.00000679, -0.00003170, -0.00003420, -0.00001010, 0.00001094, 0.00002748, 0.00004010, 0.00002543, 0.00000046, 0.00000850, 0.00002728, 0.00001398, -0.00000538, 0.00000089, 0.00001216, 0.00001559, 0.00001471, -0.00000256, -0.00001681, 0.00001058, 0.00004294, 0.00002148, -0.00001683, -0.00001152, 0.00001310, 0.00001195, -0.00000878, -0.00002364, -0.00001636, 0.00000497, 0.00000452, -0.00001805, -0.00001589, 0.00000862, 0.00000822, -0.00000531, 0.00000260, 0.00000607, -0.00001137, -0.00001286, 0.00000344, 0.00000338, -0.00000393, -0.00000099, -0.00000554, -0.00001200, 0.00000350, 0.00001524, -0.00000357, -0.00001614, 0.00000403, 0.00001764, -0.00000141, -0.00001738, -0.00000195, 0.00001564, 0.00000378, -0.00001251, 0.00000083, 0.00001650, -0.00000124, -0.00001923, -0.00000249, 0.00001206, -0.00000334, -0.00001202, 0.00000163, 0.00000579, -0.00000333, -0.00000382, 0.00000104, 0.00000190, 0.00000169, -0.00000153, -0.00000302, 0.00000484, 0.00000410, -0.00001203, -0.00001139, 0.00000833, 0.00000842, -0.00000520, 0.00000108, 0.00000734, -0.00001015, -0.00001705, 0.00000482, 0.00001836, 0.00000815, -0.00000207, -0.00000516, -0.00000740, -0.00000387, 0.00000265, 0.00000193, -0.00000058, 0.00000313, 0.00000602, 0.00000208, -0.00000327, -0.00000440, -0.00000255, -0.00000227, -0.00000380, -0.00000210, 0.00000173, 0.00000105, -0.00000184, -0.00000090, 0.00000016, -0.00000182, -0.00000193, 0.00000023, 0.00000015, -0.00000124, -0.00000180, -0.00000225, },
		{ -0.00000180, -0.00000231, -0.00000215, -0.00000398, -0.00000574, -0.00000317, 0.00000034, 0.00000119, -0.00000040, -0.00000757, -0.00001373, -0.00000469, 0.00000345, -0.00001557, -0.00003710, -0.00002509, 0.00000031, 0.00001458, 0.00003078, 0.00003879, 0.00001582, -0.00000349, 0.00001234, 0.00002418, 0.00000852, -0.00000124, 0.00000309, 0.00000570, 0.00001573, 0.00002114, -0.00000515, -0.00002298, 0.00001136, 0.00004318, 0.00001857, -0.00001627, -0.00001202, 0.00000926, 0.00001409, -0.00000327, -0.00002416, -0.00001732, 0.00000731, 0.00000293, -0.00002002, -0.00001225, 0.00000905, 0.00000457, -0.00000215, 0.00000659, 0.00000158, -0.00001320, -0.00000794, 0.00000260, 0.00000020, -0.00000002, -0.00000104, -0.00001223, -0.00000817, 0.00001292, 0.00001012, -0.00001222, -0.00000895, 0.00001218, 0.00001009, -0.00000881, -0.00001077, 0.00000452, 0.00001024, -0.00000206, -0.00000825, 0.00000564, 0.00001173, -0.00000658, -0.00001536, 0.00000140, 0.00000834, -0.00000493, -0.00000763, 0.00000206, 0.00000238, -0.00000239, -0.00000190, -0.00000071, 0.00000130, 0.00000272, -0.00000210, -0.00000204, 0.00000622, 0.00000043, -0.00001431, -0.00000663, 0.00001007, 0.00000438, -0.00000395, 0.00000441, 0.00000216, -0.00001434, -0.00001011, 0.00001080, 0.00001465, 0.00000415, -0.00000197, -0.00000598, -0.00000776, -0.00000107, 0.00000422, 0.00000036, -0.00000126, 0.00000442, 0.00000647, 0.00000065, -0.00000447, -0.00000336, -0.00000103, -0.00000342, -0.00000595, -0.00000186, 0.00000305, 0.00000123, -0.00000179, -0.00000068, -0.00000012, -0.00000169, -0.00000149, -0.00000014, 0.00000000, -0.00000068, },
		{ 0.00000013, -0.00000035, -0.00000194, -0.00000206, -0.00000199, -0.00000467, -0.00000625, -0.00000370, -0.00000012, 0.00000285, 0.00000066, -0.00001029, -0.00001341, 0.00000064, 0.00000111, -0.00002534, -0.00003617, -0.00001366, 0.00000632, 0.00001683, 0.00003446, 0.00003422, 0.00000622, -0.00000282, 0.00001547, 0.00001771, 0.00000578, 0.00000501, 0.00000181, -0.00000085, 0.00001913, 0.00002590, -0.00000969, -0.00002635, 0.00001233, 0.00004058, 0.00001719, -0.00001343, -0.00001437, 0.00000482, 0.00001756, 0.00000111, -0.00002504, -0.00001657, 0.00000834, -0.00000008, -0.00001931, -0.00000821, 0.00000663, 0.00000231, 0.00000309, 0.00000750, -0.00000371, -0.00001145, -0.00000389, -0.00000080, -0.00000040, 0.00000420, -0.00000500, -0.00001643, 0.00000050, 0.00001775, 0.00000036, -0.00001499, 0.00000187, 0.00001382, -0.00000027, -0.00001029, -0.00000225, 0.00000602, 0.00000352, -0.00000388, -0.00000303, 0.00000699, 0.00000600, -0.00000890, -0.00001067, 0.00000288, 0.00000456, -0.00000443, -0.00000389, 0.00000094, 0.00000007, -0.00000080, -0.00000124, -0.00000223, 0.00000149, 0.00000295, -0.00000240, -0.00000014, 0.00000587, -0.00000403, -0.00001379, -0.00000149, 0.00000901, 0.00000106, -0.00000072, 0.00000537, -0.00000431, -0.00001478, -0.00000181, 0.00001339, 0.00000962, 0.00000165, -0.00000178, -0.00000695, -0.00000693, 0.00000179, 0.00000458, -0.00000123, -0.00000121, 0.00000555, 0.00000617, -0.00000090, -0.00000487, -0.00000179, -0.00000034, -0.00000525, -0.00000731, -0.00000099, 0.00000389, 0.00000126, -0.00000152, -0.00000069, -0.00000043, -0.00000130, -0.00000123, -0.00000053, },
		{ -0.00000120, -0.00000071, 0.00000037, -0.00000037, -0.00000194, -0.00000165, -0.00000231, -0.00000535, -0.00000641, -0.00000439, -0.00000005, 0.00000491, 0.00000010, -0.00001296, -0.00001018, 0.00000460, -0.00000581, -0.00003251, -0.00002899, -0.00000373, 0.00000815, 0.00001999, 0.00003727, 0.00002598, -0.00000091, 0.00000174, 0.00001560, 0.00000965, 0.00000699, 0.00001087, -0.00000289, -0.00000513, 0.00002455, 0.00002743, -0.00001426, -0.00002659, 0.00001165, 0.00003601, 0.00001850, -0.00000982, -0.00001872, 0.00000115, 0.00002159, 0.00000377, -0.00002510, -0.00001466, 0.00000712, -0.00000299, -0.00001579, -0.00000563, 0.00000234, 0.00000285, 0.00000841, 0.00000492, -0.00000740, -0.00000694, -0.00000260, -0.00000485, 0.00000213, 0.00000621, -0.00001159, -0.00001544, 0.00001104, 0.00001571, -0.00001002, -0.00001045, 0.00001185, 0.00000838, -0.00000921, -0.00000587, 0.00000468, 0.00000298, -0.00000180, -0.00000222, 0.00000107, 0.00000534, 0.00000100, -0.00000857, -0.00000649, 0.00000236, 0.00000168, -0.00000249, -0.00000143, -0.00000095, -0.00000097, 0.00000072, -0.00000153, -0.00000314, 0.00000190, 0.00000256, -0.00000188, 0.00000177, 0.00000355, -0.00000775, -0.00001065, 0.00000236, 0.00000589, -0.00000034, 0.00000312, 0.00000344, -0.00001008, -0.00001144, 0.00000578, 0.00001263, 0.00000466, 0.00000048, -0.00000191, -0.00000760, -0.00000504, 0.00000413, 0.00000395, -0.00000251, -0.00000062, 0.00000628, 0.00000523, -0.00000226, -0.00000436, -0.00000017, -0.00000071, -0.00000724, -0.00000766, 0.00000009, 0.00000420, 0.00000129, -0.00000118, -0.00000088, -0.00000057, -0.00000086, },
		{ -0.00000054, -0.00000059, -0.00000124, -0.00000058, 0.00000046, -0.00000063, -0.00000165, -0.00000138, -0.00000306, -0.00000577, -0.00000649, -0.00000506, 0.00000094, 0.00000645, -0.00000221, -0.00001392, -0.00000475, 0.00000477, -0.00001547, -0.00003403, -0.00001801, 0.00000185, 0.00000800, 0.00002506, 0.00003714, 0.00001525, -0.00000349, 0.00000793, 0.00001166, 0.00000266, 0.00001173, 0.00001391, -0.00000939, -0.00000570, 0.00002967, 0.00002557, -0.00001676, -0.00002474, 0.00000809, 0.00003118, 0.00002260, -0.00000706, -0.00002421, -0.00000094, 0.00002495, 0.00000501, -0.00002341, -0.00001286, 0.00000366, -0.00000412, -0.00001052, -0.00000575, -0.00000190, 0.00000628, 0.00001158, 0.00000006, -0.00000771, -0.00000179, -0.00000459, -0.00000717, 0.00000662, 0.00000414, -0.00001791, -0.00000840, 0.00001933, 0.00000707, -0.00001653, -0.00000002, 0.00001679, -0.00000202, -0.00001324, 0.00000219, 0.00000749, -0.00000258, -0.00000398, 0.00000127, 0.00000279, 0.00000196, -0.00000222, -0.00000651, -0.00000361, 0.00000065, 0.00000021, 0.00000005, -0.00000039, -0.00000285, -0.00000095, 0.00000169, -0.00000228, -0.00000339, 0.00000207, 0.00000204, -0.00000049, 0.00000271, -0.00000019, -0.00000948, -0.00000598, 0.00000390, 0.00000209, 0.00000051, 0.00000597, -0.00000080, -0.00001345, -0.00000536, 0.00001104, 0.00000942, 0.00000074, 0.00000009, -0.00000246, -0.00000750, -0.00000244, 0.00000561, 0.00000265, -0.00000330, 0.00000026, 0.00000648, 0.00000382, -0.00000311, -0.00000306, 0.00000099, -0.00000206, -0.00000886, -0.00000710, 0.00000103, 0.00000411, 0.00000140, -0.00000090, -0.00000110, },
		{ -0.00000068, -0.00000119, -0.00000046, -0.00000055, -0.00000113, -0.00000024, 0.00000023, -0.00000090, -0.00000111, -0.00000149, -0.00000397, -0.00000588, -0.00000672, -0.00000523, 0.00000271, 0.00000645, -0.00000541, -0.00001187, 0.00000084, -0.00000008, -0.00002452, -0.00002878, -0.00000711, 0.00000240, 0.00000874, 0.00003118, 0.00003244, 0.00000461, -0.00000087, 0.00001281, 0.00000438, -0.00000092, 0.00001798, 0.00001290, -0.00001508, -0.00000272, 0.00003216, 0.00002165, -0.00001587, -0.00002271, 0.00000165, 0.00002791, 0.00002843, -0.00000618, -0.00002942, -0.00000154, 0.00002655, 0.00000586, -0.00001989, -0.00001253, -0.00000097, -0.00000248, -0.00000540, -0.00000850, -0.00000407, 0.00001126, 0.00001116, -0.00000478, -0.00000435, 0.00000152, -0.00000874, -0.00000608, 0.00001062, -0.00000215, -0.00002055, 0.00000314, 0.00002171, -0.00000526, -0.00001602, 0.00001247, 0.00001435, -0.00001341, -0.00001091, 0.00001041, 0.00000546, -0.00000793, -0.00000276, 0.00000460, 0.00000195, -0.00000170, -0.00000339, -0.00000381, -0.00000227, -0.00000136, 0.00000013, 0.00000241, -0.00000054, -0.00000421, -0.00000032, 0.00000194, -0.00000306, -0.00000321, 0.00000186, 0.00000190, 0.00000131, 0.00000207, -0.00000422, -0.00000872, -0.00000127, 0.00000300, -0.00000092, 0.00000299, 0.00000670, -0.00000598, -0.00001346, 0.00000174, 0.00001319, 0.00000504, -0.00000175, -0.00000008, -0.00000321, -0.00000643, 0.00000037, 0.00000608, 0.00000104, -0.00000355, 0.00000123, 0.00000611, 0.00000221, -0.00000320, -0.00000129, 0.00000129, -0.00000404, -0.00000967, -0.00000589, 0.00000159, 0.00000377, 0.00000156, },
		{ 0.00000330, 0.00000171, -0.00000047, -0.00000112, -0.00000050, -0.00000066, -0.00000072, 0.00000008, -0.00000031, -0.00000094, -0.00000055, -0.00000205, -0.00000475, -0.00000588, -0.00000704, -0.00000441, 0.00000453, 0.00000435, -0.00000783, -0.00000668, 0.00000389, -0.00000923, -0.00002931, -0.00001816, 0.00000005, -0.00000039, 0.00001248, 0.00003595, 0.00002312, -0.00000293, 0.00000566, 0.00001400, -0.00000394, -0.00000011, 0.00002292, 0.00000835, -0.00001758, 0.00000207, 0.00003070, 0.00001789, -0.00001167, -0.00002245, -0.00000637, 0.00002727, 0.00003423, -0.00000713, -0.00003296, -0.00000158, 0.00002603, 0.00000756, -0.00001536, -0.00001446, -0.00000504, 0.00000172, -0.00000230, -0.00001253, -0.00000303, 0.00001548, 0.00000722, -0.00000711, 0.00000130, 0.00000130, -0.00001268, -0.00000150, 0.00001141, -0.00001076, -0.00001703, 0.00001555, 0.00001647, -0.00001670, -0.00000797, 0.00002223, 0.00000501, -0.00002148, -0.00000323, 0.00001552, -0.00000017, -0.00001083, 0.00000078, 0.00000623, -0.00000071, -0.00000449, -0.00000290, -0.00000137, -0.00000224, -0.00000297, 0.00000109, 0.00000405, -0.00000138, -0.00000479, 0.00000045, 0.00000157, -0.00000357, -0.00000293, 0.00000142, 0.00000241, 0.00000278, -0.00000016, -0.00000727, -0.00000579, 0.00000209, 0.00000040, -0.00000218, 0.00000589, 0.00000495, -0.00001039, -0.00001021, 0.00000805, 0.00001233, 0.00000075, -0.00000296, -0.00000047, -0.00000376, -0.00000441, 0.00000292, 0.00000558, -0.00000060, -0.00000329, 0.00000211, 0.00000518, 0.00000071, -0.00000243, 0.00000049, 0.00000054, -0.00000615, -0.00000951, -0.00000442, 0.00000170, },
		{ -0.00000306, 0.00000141, 0.00000279, 0.00000185, -0.00000015, -0.00000097, -0.00000079, -0.00000072, -0.00000002, 0.00000015, -0.00000098, -0.00000068, -0.00000024, -0.00000290, -0.00000527, -0.00000603, -0.00000708, -0.00000246, 0.00000528, 0.00000050, -0.00000774, 0.00000025, 0.00000224, -0.00001997, -0.00002745, -0.00000566, 0.00000152, -0.00000332, 0.00001936, 0.00003664, 0.00001106, -0.00000516, 0.00001331, 0.00001070, -0.00001038, 0.00000419, 0.00002421, 0.00000241, -0.00001575, 0.00000604, 0.00002558, 0.00001646, -0.00000566, -0.00002501, -0.00001397, 0.00002910, 0.00003832, -0.00000889, -0.00003419, -0.00000236, 0.00002398, 0.00001089, -0.00001130, -0.00001834, -0.00000688, 0.00000706, -0.00000219, -0.00001567, 0.00000089, 0.00001671, 0.00000148, -0.00000542, 0.00000668, -0.00000251, -0.00001388, 0.00000485, 0.00000733, -0.00001827, -0.00000706, 0.00002443, 0.00000473, -0.00002279, 0.00000521, 0.00002536, -0.00000814, -0.00002326, 0.00000678, 0.00001562, -0.00000693, -0.00001025, 0.00000481, 0.00000556, -0.00000397, -0.00000586, -0.00000144, 0.00000031, -0.00000300, -0.00000379, 0.00000256, 0.00000475, -0.00000240, -0.00000461, 0.00000099, 0.00000084, -0.00000377, -0.00000277, 0.00000108, 0.00000347, 0.00000321, -0.00000337, -0.00000841, -0.00000173, 0.00000332, -0.00000265, -0.00000149, 0.00000786, 0.00000132, -0.00001261, -0.00000471, 0.00001222, 0.00000931, -0.00000261, -0.00000337, -0.00000118, -0.00000375, -0.00000170, 0.00000484, 0.00000430, -0.00000205, -0.00000261, 0.00000272, 0.00000376, -0.00000037, -0.00000090, 0.00000176, -0.00000118, -0.00000782, -0.00000843, },
		{ -0.00000865, -0.00000673, -0.00000207, 0.00000083, 0.00000229, 0.00000200, 0.00000032, -0.00000089, -0.00000129, -0.00000053, 0.00000077, -0.00000013, -0.00000152, -0.00000022, -0.00000034, -0.00000378, -0.00000561, -0.00000641, -0.00000628, 0.00000014, 0.00000399, -0.00000374, -0.00000416, 0.00000634, -0.00000467, -0.00002846, -0.00001894, 0.00000445, -0.00000206, -0.00000318, 0.00002740, 0.00003161, -0.00000046, -0.00000176, 0.00001882, 0.00000411, -0.00001266, 0.00000950, 0.00002096, -0.00000208, -0.00001027, 0.00000689, 0.00001861, 0.00001843, -0.00000005, -0.00003006, -0.00001925, 0.00003209, 0.00003983, -0.00000995, -0.00003349, -0.00000484, 0.00002164, 0.00001569, -0.00000907, -0.00002280, -0.00000572, 0.00001147, -0.00000466, -0.00001596, 0.00000589, 0.00001382, -0.00000341, 0.00000005, 0.00000921, -0.00000819, -0.00001084, 0.00001005, -0.00000127, -0.00002110, 0.00000712, 0.00002626, -0.00000981, -0.00002092, 0.00001898, 0.00002033, -0.00002058, -0.00001817, 0.00001537, 0.00001088, -0.00001210, -0.00000666, 0.00000757, 0.00000295, -0.00000674, -0.00000586, 0.00000028, 0.00000108, -0.00000402, -0.00000374, 0.00000401, 0.00000460, -0.00000315, -0.00000392, 0.00000112, 0.00000001, -0.00000375, -0.00000277, 0.00000113, 0.00000462, 0.00000224, -0.00000652, -0.00000734, 0.00000213, 0.00000244, -0.00000492, 0.00000055, 0.00000799, -0.00000297, -0.00001193, 0.00000153, 0.00001362, 0.00000527, -0.00000468, -0.00000349, -0.00000207, -0.00000294, 0.00000133, 0.00000590, 0.00000242, -0.00000313, -0.00000157, 0.00000292, 0.00000205, -0.00000073, 0.00000106, 0.00000210, -0.00000348, },
		{ 0.00000133, -0.00000582, -0.00000844, -0.00000480, -0.00000155, 0.00000008, 0.00000180, 0.00000225, 0.00000089, -0.00000103, -0.00000182, -0.00000001, 0.00000140, -0.00000068, -0.00000174, 0.00000018, -0.00000085, -0.00000447, -0.00000601, -0.00000676, -0.00000425, 0.00000237, 0.00000042, -0.00000644, 0.00000243, 0.00000880, -0.00001521, -0.00003127, -0.00000633, 0.00000893, -0.00000776, 0.00000166, 0.00003330, 0.00002131, -0.00000802, 0.00000543, 0.00001986, -0.00000303, -0.00001018, 0.00001284, 0.00001427, -0.00000281, -0.00000338, 0.00000361, 0.00001231, 0.00002330, 0.00000317, -0.00003599, -0.00002133, 0.00003436, 0.00003904, -0.00000904, -0.00003202, -0.00000910, 0.00002031, 0.00002086, -0.00000927, -0.00002596, -0.00000212, 0.00001313, -0.00000794, -0.00001255, 0.00000944, 0.00000741, -0.00000493, 0.00000738, 0.00000746, -0.00001294, -0.00000395, 0.00001124, -0.00001190, -0.00001699, 0.00002133, 0.00001993, -0.00002222, -0.00001141, 0.00002840, 0.00000860, -0.00002803, -0.00000816, 0.00001952, 0.00000324, -0.00001386, -0.00000163, 0.00000802, -0.00000057, -0.00000832, -0.00000492, 0.00000173, 0.00000108, -0.00000484, -0.00000298, 0.00000501, 0.00000386, -0.00000340, -0.00000302, 0.00000083, -0.00000075, -0.00000365, -0.00000278, 0.00000166, 0.00000526, 0.00000003, -0.00000859, -0.00000451, 0.00000467, 0.00000017, -0.00000564, 0.00000288, 0.00000608, -0.00000648, -0.00000855, 0.00000696, 0.00001242, 0.00000130, -0.00000559, -0.00000370, -0.00000283, -0.00000129, 0.00000429, 0.00000599, 0.00000021, -0.00000369, -0.00000034, 0.00000258, 0.00000033, -0.00000022, 0.00000299, },
		{ 0.00000109, 0.00000432, -0.00000044, -0.00000762, -0.00000728, -0.00000304, -0.00000150, -0.00000071, 0.00000139, 0.00000264, 0.00000143, -0.00000143, -0.00000216, 0.00000074, 0.00000167, -0.00000128, -0.00000163, 0.00000029, -0.00000157, -0.00000496, -0.00000664, -0.00000655, -0.00000114, 0.00000299, -0.00000459, -0.00000592, 0.00001003, 0.00000586, -0.00002597, -0.00002679, 0.00000613, 0.00000687, -0.00001168, 0.00001040, 0.00003395, 0.00000834, -0.00000953, 0.00001313, 0.00001602, -0.00000765, -0.00000438, 0.00001203, 0.00000671, 0.00000094, 0.00000208, -0.00000305, 0.00000882, 0.00002923, 0.00000314, -0.00004063, -0.00002062, 0.00003432, 0.00003723, -0.00000580, -0.00003114, -0.00001429, 0.00002063, 0.00002481, -0.00001133, -0.00002625, 0.00000218, 0.00001136, -0.00000971, -0.00000626, 0.00000936, -0.00000031, -0.00000190, 0.00001362, 0.00000185, -0.00001412, 0.00000453, 0.00000697, -0.00002074, -0.00000615, 0.00003097, 0.00000727, -0.00002817, 0.00000246, 0.00002996, -0.00000590, -0.00002809, 0.00000313, 0.00001798, -0.00000444, -0.00001184, 0.00000286, 0.00000613, -0.00000386, -0.00000862, -0.00000358, 0.00000268, 0.00000059, -0.00000521, -0.00000184, 0.00000536, 0.00000289, -0.00000307, -0.00000217, 0.00000022, -0.00000136, -0.00000356, -0.00000260, 0.00000248, 0.00000489, -0.00000279, -0.00000887, -0.00000088, 0.00000528, -0.00000241, -0.00000471, 0.00000439, 0.00000269, -0.00000808, -0.00000339, 0.00001047, 0.00000936, -0.00000186, -0.00000573, -0.00000414, -0.00000315, 0.00000108, 0.00000679, 0.00000510, -0.00000207, -0.00000360, 0.00000087, 0.00000166, -0.00000101, },
		{ 0.00000027, -0.00000163, 0.00000287, 0.00000461, -0.00000278, -0.00000844, -0.00000549, -0.00000175, -0.00000180, -0.00000145, 0.00000111, 0.00000316, 0.00000176, -0.00000197, -0.00000215, 0.00000149, 0.00000151, -0.00000167, -0.00000132, -0.00000000, -0.00000228, -0.00000539, -0.00000741, -0.00000525, 0.00000228, 0.00000122, -0.00000929, -0.00000163, 0.00001572, -0.00000230, -0.00003293, -0.00001610, 0.00001414, 0.00000013, -0.00001065, 0.00002003, 0.00002795, -0.00000343, -0.00000511, 0.00001791, 0.00000901, -0.00000765, 0.00000190, 0.00000666, 0.00000131, 0.00000800, 0.00000387, -0.00001087, 0.00000897, 0.00003381, 0.00000051, -0.00004219, -0.00001868, 0.00003139, 0.00003598, -0.00000090, -0.00003172, -0.00001900, 0.00002225, 0.00002622, -0.00001363, -0.00002317, 0.00000498, 0.00000701, -0.00000809, 0.00000073, 0.00000486, -0.00000640, 0.00000496, 0.00001617, -0.00000538, -0.00001046, 0.00001134, -0.00000197, -0.00002412, 0.00000859, 0.00003279, -0.00000747, -0.00002564, 0.00001570, 0.00002295, -0.00001830, -0.00002103, 0.00001180, 0.00001166, -0.00000954, -0.00000711, 0.00000528, 0.00000272, -0.00000609, -0.00000795, -0.00000228, 0.00000308, -0.00000010, -0.00000501, -0.00000064, 0.00000504, 0.00000203, -0.00000228, -0.00000158, -0.00000056, -0.00000183, -0.00000345, -0.00000209, 0.00000315, 0.00000334, -0.00000532, -0.00000726, 0.00000240, 0.00000409, -0.00000424, -0.00000267, 0.00000437, -0.00000113, -0.00000728, 0.00000225, 0.00001160, 0.00000544, -0.00000390, -0.00000553, -0.00000474, -0.00000278, 0.00000389, 0.00000851, 0.00000331, -0.00000408, -0.00000284, 0.00000176, },
		{ -0.00000153, 0.00000205, -0.00000127, -0.00000130, 0.00000459, 0.00000367, -0.00000508, -0.00000810, -0.00000353, -0.00000105, -0.00000228, -0.00000207, 0.00000101, 0.00000369, 0.00000178, -0.00000245, -0.00000176, 0.00000197, 0.00000107, -0.00000171, -0.00000105, -0.00000063, -0.00000286, -0.00000598, -0.00000791, -0.00000264, 0.00000477, -0.00000281, -0.00001164, 0.00000537, 0.00001678, -0.00001325, -0.00003307, -0.00000271, 0.00001508, -0.00000739, -0.00000371, 0.00002661, 0.00001634, -0.00001041, 0.00000285, 0.00001764, 0.00000194, -0.00000291, 0.00000557, -0.00000170, 0.00000019, 0.00001573, 0.00000132, -0.00001714, 0.00001191, 0.00003515, -0.00000286, -0.00004006, -0.00001748, 0.00002617, 0.00003645, 0.00000427, -0.00003367, -0.00002195, 0.00002388, 0.00002473, -0.00001421, -0.00001764, 0.00000462, 0.00000217, -0.00000260, 0.00000566, -0.00000300, -0.00000840, 0.00001324, 0.00001385, -0.00001112, -0.00000274, 0.00001355, -0.00001264, -0.00002007, 0.00002270, 0.00002624, -0.00001921, -0.00001575, 0.00002342, 0.00000973, -0.00002456, -0.00000958, 0.00001516, 0.00000308, -0.00001064, -0.00000163, 0.00000504, -0.00000100, -0.00000697, -0.00000679, -0.00000123, 0.00000301, -0.00000073, -0.00000429, 0.00000028, 0.00000419, 0.00000152, -0.00000122, -0.00000132, -0.00000138, -0.00000218, -0.00000324, -0.00000130, 0.00000322, 0.00000088, -0.00000672, -0.00000425, 0.00000439, 0.00000178, -0.00000473, -0.00000043, 0.00000272, -0.00000421, -0.00000429, 0.00000710, 0.00001052, 0.00000158, -0.00000491, -0.00000535, -0.00000526, -0.00000160, 0.00000682, 0.00000916, 0.00000083, -0.00000543, },
		{ -0.00000195, -0.00000584, 0.00000001, 0.00000156, -0.00000253, -0.00000006, 0.00000565, 0.00000165, -0.00000671, -0.00000672, -0.00000185, -0.00000093, -0.00000279, -0.00000251, 0.00000110, 0.00000408, 0.00000151, -0.00000265, -0.00000114, 0.00000201, 0.00000057, -0.00000140, -0.00000103, -0.00000143, -0.00000335, -0.00000677, -0.00000756, 0.00000084, 0.00000515, -0.00000783, -0.00001017, 0.00001257, 0.00001197, -0.00002311, -0.00002566, 0.00000867, 0.00000912, -0.00001138, 0.00000738, 0.00002690, 0.00000244, -0.00001073, 0.00001067, 0.00001237, -0.00000205, 0.00000468, 0.00000470, -0.00001008, 0.00000371, 0.00002114, -0.00000440, -0.00001985, 0.00001550, 0.00003269, -0.00000475, -0.00003502, -0.00001847, 0.00002020, 0.00003882, 0.00000830, -0.00003595, -0.00002265, 0.00002388, 0.00002124, -0.00001164, -0.00001168, 0.00000071, -0.00000071, 0.00000555, 0.00000646, -0.00001160, -0.00000544, 0.00001977, 0.00000754, -0.00001269, 0.00000636, 0.00000995, -0.00002100, -0.00000925, 0.00003163, 0.00001373, -0.00002398, -0.00000240, 0.00002271, -0.00000519, -0.00002290, 0.00000219, 0.00001260, -0.00000469, -0.00000794, 0.00000261, 0.00000257, -0.00000397, -0.00000673, -0.00000555, -0.00000049, 0.00000260, -0.00000112, -0.00000323, 0.00000072, 0.00000306, 0.00000149, -0.00000014, -0.00000139, -0.00000212, -0.00000242, -0.00000281, -0.00000044, 0.00000236, -0.00000189, -0.00000650, -0.00000073, 0.00000469, -0.00000071, -0.00000382, 0.00000111, -0.00000007, -0.00000567, 0.00000006, 0.00001027, 0.00000787, -0.00000160, -0.00000517, -0.00000533, -0.00000545, 0.00000035, 0.00000945, 0.00000856, },
		{ 0.00001133, 0.00000669, -0.00000450, -0.00000515, 0.00000129, 0.00000037, -0.00000304, 0.00000174, 0.00000554, -0.00000096, -0.00000719, -0.00000471, -0.00000077, -0.00000121, -0.00000319, -0.00000272, 0.00000130, 0.00000418, 0.00000110, -0.00000243, -0.00000054, 0.00000157, 0.00000025, -0.00000087, -0.00000132, -0.00000225, -0.00000392, -0.00000750, -0.00000590, 0.00000418, 0.00000287, -0.00001184, -0.00000484, 0.00001688, 0.00000226, -0.00002789, -0.00001278, 0.00001393, -0.00000077, -0.00000897, 0.00001877, 0.00001989, -0.00000945, -0.00000503, 0.00001486, 0.00000427, -0.00000115, 0.00001196, -0.00000070, -0.00001542, 0.00001020, 0.00002219, -0.00001076, -0.00001849, 0.00001729, 0.00002742, -0.00000353, -0.00002891, -0.00002200, 0.00001514, 0.00004220, 0.00001044, -0.00003712, -0.00002170, 0.00002109, 0.00001759, -0.00000573, -0.00000757, -0.00000552, 0.00000012, 0.00001382, 0.00000267, -0.00001785, 0.00000134, 0.00002197, -0.00000014, -0.00000907, 0.00001336, 0.00000173, -0.00002346, 0.00000509, 0.00003251, -0.00000020, -0.00002044, 0.00000924, 0.00001372, -0.00001684, -0.00001440, 0.00001046, 0.00000564, -0.00000922, -0.00000298, 0.00000435, -0.00000100, -0.00000562, -0.00000585, -0.00000443, -0.00000004, 0.00000197, -0.00000114, -0.00000205, 0.00000060, 0.00000196, 0.00000189, 0.00000076, -0.00000168, -0.00000274, -0.00000251, -0.00000215, 0.00000011, 0.00000054, -0.00000419, -0.00000468, 0.00000233, 0.00000345, -0.00000251, -0.00000201, 0.00000142, -0.00000313, -0.00000518, 0.00000470, 0.00001142, 0.00000447, -0.00000382, -0.00000506, -0.00000548, -0.00000504, 0.00000294, },
		{ -0.00000386, 0.00000585, 0.00001202, 0.00000376, -0.00000623, -0.00000353, 0.00000186, -0.00000120, -0.00000250, 0.00000349, 0.00000406, -0.00000343, -0.00000636, -0.00000260, -0.00000041, -0.00000168, -0.00000341, -0.00000270, 0.00000151, 0.00000391, 0.00000077, -0.00000178, -0.00000023, 0.00000080, 0.00000027, -0.00000034, -0.00000189, -0.00000303, -0.00000466, -0.00000768, -0.00000298, 0.00000605, -0.00000150, -0.00001281, 0.00000274, 0.00001587, -0.00000930, -0.00002513, 0.00000130, 0.00001125, -0.00000996, 0.00000004, 0.00002604, 0.00000729, -0.00001570, 0.00000378, 0.00001357, -0.00000327, 0.00000436, 0.00001588, -0.00000860, -0.00001584, 0.00001676, 0.00001858, -0.00001503, -0.00001428, 0.00001557, 0.00002147, 0.00000112, -0.00002374, -0.00002720, 0.00001208, 0.00004513, 0.00001096, -0.00003602, -0.00002057, 0.00001544, 0.00001578, 0.00000226, -0.00000685, -0.00001171, 0.00000484, 0.00001941, -0.00000425, -0.00001943, 0.00000916, 0.00001902, -0.00000589, -0.00000143, 0.00001546, -0.00000789, -0.00001843, 0.00001830, 0.00002535, -0.00001059, -0.00001040, 0.00001465, -0.00000047, -0.00002162, -0.00000238, 0.00001295, -0.00000279, -0.00000955, 0.00000209, 0.00000339, -0.00000441, -0.00000598, -0.00000480, -0.00000348, 0.00000013, 0.00000126, -0.00000077, -0.00000103, 0.00000000, 0.00000113, 0.00000258, 0.00000134, -0.00000209, -0.00000317, -0.00000242, -0.00000137, 0.00000001, -0.00000192, -0.00000535, -0.00000177, 0.00000417, 0.00000131, -0.00000313, -0.00000003, 0.00000043, -0.00000554, -0.00000297, 0.00000862, 0.00001071, 0.00000107, -0.00000509, -0.00000486, -0.00000565, },
		},
		{ { -0.00000105, -0.00000198, -0.00000382, -0.00000380, 0.00000081, 0.00000877, 0.00001365, 0.00000945, -0.00000188, -0.00001079, -0.00000963, -0.00000035, 0.00000813, 0.00000864, 0.00000108, -0.00000834, -0.00001140, -0.00000410, 0.00000783, 0.00001205, 0.00000249, -0.00001215, -0.00001742, -0.00000977, 0.00000091, 0.00000491, 0.00000254, -0.00000065, -0.00000167, -0.00000044, 0.00000188, 0.00000175, -0.00000393, -0.00001074, -0.00000930, 0.00000067, 0.00000658, -0.00000023, -0.00001123, -0.00001295, -0.00000584, -0.00000015, 0.00000125, 0.00000333, 0.00000558, 0.00000283, -0.00000225, -0.00000215, 0.00000022, -0.00000432, -0.00001133, -0.00000636, 0.00000758, 0.00000916, -0.00000571, -0.00001255, 0.00000118, 0.00001025, -0.00000705, -0.00002503, -0.00000658, 0.00003044, 0.00003061, -0.00001316, -0.00004110, -0.00001316, 0.00003061, 0.00003044, -0.00000658, -0.00002503, -0.00000705, 0.00001025, 0.00000118, -0.00001255, -0.00000571, 0.00000916, 0.00000758, -0.00000636, -0.00001133, -0.00000432, 0.00000022, -0.00000215, -0.00000225, 0.00000283, 0.00000558, 0.00000333, 0.00000125, -0.00000015, -0.00000584, -0.00001295, -0.00001123, -0.00000023, 0.00000658, 0.00000067, -0.00000930, -0.00001074, -0.00000393, 0.00000175, 0.00000188, -0.00000044, -0.00000167, -0.00000065, 0.00000254, 0.00000491, 0.00000091, -0.00000977, -0.00001742, -0.00001215, 0.00000249, 0.00001205, 0.00000783, -0.00000410, -0.00001140, -0.00000834, 0.00000108, 0.00000864, 0.00000813, -0.00000035, -0.00000963, -0.00001079, -0.00000188, 0.00000945, 0.00001365, 0.00000877, 0.00000081, -0.00000380, -0.00000382, -0.00000198, },
		{ -0.00000290, -0.00000323, -0.00000184, -0.00000104, -0.00000261, -0.00000525, -0.00000435, 0.00000292, 0.00001230, 0.00001491, 0.00000686, -0.00000566, -0.00001231, -0.00000844, 0.00000187, 0.00001023, 0.00000979, 0.00000003, -0.00001112, -0.00001253, -0.00000169, 0.00001023, 0.00000992, -0.00000233, -0.00001360, -0.00001418, -0.00000669, 0.00000087, 0.00000427, 0.00000329, -0.00000043, -0.00000273, 0.00000006, 0.00000428, 0.00000160, -0.00000780, -0.00001251, -0.00000582, 0.00000329, 0.00000305, -0.00000443, -0.00000888, -0.00000799, -0.00000650, -0.00000396, 0.00000301, 0.00000923, 0.00000632, -0.00000174, -0.00000350, -0.00000002, -0.00000278, -0.00001025, -0.00000758, 0.00000533, 0.00000866, -0.00000350, -0.00000924, 0.00000347, 0.00001050, -0.00000882, -0.00002796, -0.00000883, 0.00003071, 0.00003316, -0.00001093, -0.00004138, -0.00001529, 0.00002901, 0.00003041, -0.00000592, -0.00002426, -0.00000553, 0.00001239, 0.00000179, -0.00001521, -0.00000995, 0.00000720, 0.00000937, -0.00000351, -0.00001081, -0.00000616, -0.00000116, -0.00000141, -0.00000101, 0.00000183, 0.00000206, 0.00000032, 0.00000233, 0.00000537, 0.00000008, -0.00001190, -0.00001633, -0.00000685, 0.00000479, 0.00000529, -0.00000314, -0.00000908, -0.00000776, -0.00000310, 0.00000051, 0.00000152, -0.00000028, -0.00000263, -0.00000104, 0.00000448, 0.00000612, -0.00000233, -0.00001470, -0.00001780, -0.00000737, 0.00000662, 0.00001165, 0.00000526, -0.00000508, -0.00001045, -0.00000711, 0.00000178, 0.00000841, 0.00000640, -0.00000274, -0.00001011, -0.00000821, 0.00000128, 0.00000988, 0.00001132, 0.00000654, 0.00000056, },
		{ 0.00000979, 0.00000608, 0.00000093, -0.00000278, -0.00000332, -0.00000165, -0.00000128, -0.00000415, -0.00000671, -0.00000308, 0.00000685, 0.00001516, 0.00001378, 0.00000313, -0.00000863, -0.00001310, -0.00000729, 0.00000483, 0.00001373, 0.00001054, -0.00000332, -0.00001468, -0.00001186, 0.00000140, 0.00001031, 0.00000653, -0.00000434, -0.00001215, -0.00001252, -0.00000668, 0.00000160, 0.00000641, 0.00000357, -0.00000264, -0.00000300, 0.00000315, 0.00000539, -0.00000217, -0.00001086, -0.00001016, -0.00000294, 0.00000100, -0.00000017, -0.00000203, -0.00000471, -0.00000976, -0.00001091, -0.00000172, 0.00001000, 0.00001051, 0.00000137, -0.00000339, -0.00000114, -0.00000282, -0.00000923, -0.00000704, 0.00000459, 0.00000735, -0.00000402, -0.00000777, 0.00000698, 0.00001405, -0.00000843, -0.00003177, -0.00001358, 0.00002964, 0.00003637, -0.00000765, -0.00004174, -0.00001806, 0.00002761, 0.00003129, -0.00000528, -0.00002521, -0.00000589, 0.00001475, 0.00000497, -0.00001528, -0.00001403, 0.00000298, 0.00000891, -0.00000071, -0.00000847, -0.00000663, -0.00000318, -0.00000233, -0.00000001, 0.00000288, 0.00000080, -0.00000322, -0.00000029, 0.00000720, 0.00000611, -0.00000658, -0.00001682, -0.00001289, -0.00000093, 0.00000536, 0.00000225, -0.00000394, -0.00000753, -0.00000720, -0.00000349, 0.00000093, 0.00000157, -0.00000203, -0.00000359, 0.00000138, 0.00000712, 0.00000398, -0.00000791, -0.00001720, -0.00001449, -0.00000244, 0.00000844, 0.00001036, 0.00000343, -0.00000607, -0.00001039, -0.00000589, 0.00000331, 0.00000830, 0.00000416, -0.00000477, -0.00000948, -0.00000570, 0.00000274, 0.00000904, },
		{ -0.00000428, 0.00000298, 0.00000860, 0.00000986, 0.00000641, 0.00000054, -0.00000353, -0.00000332, -0.00000141, -0.00000243, -0.00000613, -0.00000654, 0.00000037, 0.00001054, 0.00001562, 0.00001124, -0.00000001, -0.00001111, -0.00001414, -0.00000514, 0.00000995, 0.00001722, 0.00000841, -0.00000833, -0.00001651, -0.00000983, 0.00000265, 0.00000897, 0.00000561, -0.00000362, -0.00001239, -0.00001390, -0.00000552, 0.00000552, 0.00000788, 0.00000078, -0.00000425, -0.00000016, 0.00000534, 0.00000253, -0.00000558, -0.00000963, -0.00000799, -0.00000498, -0.00000122, 0.00000328, 0.00000232, -0.00000754, -0.00001558, -0.00000898, 0.00000628, 0.00001231, 0.00000566, -0.00000057, -0.00000103, -0.00000403, -0.00000989, -0.00000634, 0.00000593, 0.00000752, -0.00000583, -0.00000943, 0.00000896, 0.00001965, -0.00000434, -0.00003400, -0.00002010, 0.00002587, 0.00003883, -0.00000317, -0.00004116, -0.00002131, 0.00002550, 0.00003262, -0.00000366, -0.00002646, -0.00000820, 0.00001548, 0.00000886, -0.00001260, -0.00001588, -0.00000163, 0.00000601, 0.00000019, -0.00000571, -0.00000517, -0.00000418, -0.00000425, -0.00000058, 0.00000415, 0.00000184, -0.00000484, -0.00000408, 0.00000513, 0.00000906, -0.00000024, -0.00001279, -0.00001507, -0.00000714, 0.00000118, 0.00000384, 0.00000144, -0.00000358, -0.00000795, -0.00000755, -0.00000231, 0.00000160, -0.00000037, -0.00000386, -0.00000166, 0.00000488, 0.00000666, -0.00000097, -0.00001180, -0.00001600, -0.00001020, 0.00000076, 0.00000900, 0.00000908, 0.00000136, -0.00000766, -0.00000999, -0.00000360, 0.00000507, 0.00000746, 0.00000173, -0.00000599, -0.00000852, },
		{ -0.00000048, -0.00000672, -0.00000804, -0.00000384, 0.00000339, 0.00000965, 0.00001086, 0.00000593, -0.00000102, -0.00000421, -0.00000293, -0.00000196, -0.00000436, -0.00000675, -0.00000398, 0.00000394, 0.00001192, 0.00001464, 0.00000927, -0.00000294, -0.00001440, -0.00001429, -0.00000010, 0.00001586, 0.00001773, 0.00000382, -0.00001209, -0.00001677, -0.00000925, 0.00000265, 0.00001017, 0.00000731, -0.00000501, -0.00001602, -0.00001380, -0.00000042, 0.00000894, 0.00000566, -0.00000187, -0.00000243, 0.00000223, 0.00000346, -0.00000028, -0.00000467, -0.00000843, -0.00001088, -0.00000703, 0.00000357, 0.00000891, -0.00000079, -0.00001495, -0.00001478, -0.00000063, 0.00000956, 0.00000799, 0.00000394, 0.00000205, -0.00000424, -0.00001223, -0.00000751, 0.00000789, 0.00001032, -0.00000637, -0.00001327, 0.00000743, 0.00002450, 0.00000287, -0.00003251, -0.00002612, 0.00001950, 0.00003892, 0.00000142, -0.00003912, -0.00002409, 0.00002253, 0.00003346, -0.00000121, -0.00002682, -0.00001123, 0.00001411, 0.00001152, -0.00000868, -0.00001494, -0.00000450, 0.00000220, -0.00000139, -0.00000441, -0.00000286, -0.00000329, -0.00000556, -0.00000246, 0.00000405, 0.00000359, -0.00000402, -0.00000645, 0.00000124, 0.00000809, 0.00000379, -0.00000692, -0.00001274, -0.00001040, -0.00000438, 0.00000119, 0.00000390, 0.00000145, -0.00000501, -0.00000904, -0.00000624, -0.00000082, 0.00000023, -0.00000268, -0.00000296, 0.00000176, 0.00000582, 0.00000320, -0.00000503, -0.00001236, -0.00001327, -0.00000693, 0.00000272, 0.00000909, 0.00000722, -0.00000133, -0.00000879, -0.00000822, -0.00000064, 0.00000614, 0.00000584, },
		{ 0.00000202, 0.00000605, 0.00000373, -0.00000236, -0.00000754, -0.00000829, -0.00000332, 0.00000530, 0.00001180, 0.00001097, 0.00000392, -0.00000259, -0.00000416, -0.00000311, -0.00000354, -0.00000521, -0.00000488, -0.00000109, 0.00000524, 0.00001196, 0.00001449, 0.00000741, -0.00000738, -0.00001734, -0.00001114, 0.00000667, 0.00001915, 0.00001531, -0.00000010, -0.00001449, -0.00001846, -0.00000972, 0.00000565, 0.00001429, 0.00000649, -0.00001056, -0.00001821, -0.00000896, 0.00000472, 0.00000830, 0.00000300, -0.00000112, -0.00000079, 0.00000060, 0.00000161, 0.00000139, -0.00000384, -0.00001275, -0.00001393, -0.00000143, 0.00001106, 0.00000657, -0.00000927, -0.00001557, -0.00000679, 0.00000301, 0.00000561, 0.00000698, 0.00000761, -0.00000119, -0.00001419, -0.00001118, 0.00000806, 0.00001471, -0.00000377, -0.00001659, 0.00000263, 0.00002598, 0.00001066, -0.00002706, -0.00002912, 0.00001233, 0.00003609, 0.00000448, -0.00003624, -0.00002563, 0.00001942, 0.00003338, 0.00000122, -0.00002610, -0.00001371, 0.00001155, 0.00001212, -0.00000549, -0.00001235, -0.00000469, -0.00000038, -0.00000428, -0.00000549, -0.00000156, -0.00000121, -0.00000514, -0.00000418, 0.00000245, 0.00000429, -0.00000201, -0.00000638, -0.00000182, 0.00000476, 0.00000409, -0.00000252, -0.00000794, -0.00000942, -0.00000793, -0.00000352, 0.00000237, 0.00000443, -0.00000038, -0.00000708, -0.00000837, -0.00000434, -0.00000122, -0.00000187, -0.00000271, -0.00000035, 0.00000330, 0.00000393, -0.00000002, -0.00000636, -0.00001120, -0.00001096, -0.00000465, 0.00000405, 0.00000837, 0.00000450, -0.00000379, -0.00000840, -0.00000517, },
		{ -0.00000492, -0.00000621, -0.00000177, 0.00000347, 0.00000486, 0.00000161, -0.00000416, -0.00000891, -0.00000847, -0.00000117, 0.00000861, 0.00001307, 0.00000916, 0.00000168, -0.00000319, -0.00000448, -0.00000451, -0.00000423, -0.00000344, -0.00000257, -0.00000063, 0.00000520, 0.00001328, 0.00001471, 0.00000350, -0.00001229, -0.00001703, -0.00000552, 0.00001145, 0.00001956, 0.00001330, -0.00000305, -0.00001881, -0.00002135, -0.00000657, 0.00001239, 0.00001577, 0.00000055, -0.00001516, -0.00001509, -0.00000334, 0.00000580, 0.00000680, 0.00000356, -0.00000056, -0.00000323, -0.00000081, 0.00000471, 0.00000295, -0.00000930, -0.00001762, -0.00000856, 0.00000770, 0.00001014, -0.00000214, -0.00001086, -0.00000812, -0.00000354, -0.00000143, 0.00000514, 0.00001257, 0.00000542, -0.00001294, -0.00001566, 0.00000498, 0.00001806, 0.00000165, -0.00001685, -0.00000302, 0.00002341, 0.00001599, -0.00001961, -0.00002805, 0.00000695, 0.00003155, 0.00000498, -0.00003402, -0.00002614, 0.00001703, 0.00003282, 0.00000302, -0.00002500, -0.00001518, 0.00000912, 0.00001128, -0.00000415, -0.00001004, -0.00000300, -0.00000049, -0.00000642, -0.00000819, -0.00000246, 0.00000034, -0.00000328, -0.00000442, 0.00000055, 0.00000342, -0.00000062, -0.00000473, -0.00000261, 0.00000163, 0.00000168, -0.00000145, -0.00000381, -0.00000559, -0.00000771, -0.00000710, -0.00000166, 0.00000385, 0.00000304, -0.00000305, -0.00000748, -0.00000678, -0.00000395, -0.00000257, -0.00000226, -0.00000110, 0.00000096, 0.00000240, 0.00000193, -0.00000106, -0.00000603, -0.00001016, -0.00000944, -0.00000292, 0.00000464, 0.00000660, 0.00000154, },
		{ 0.00000425, 0.00000409, -0.00000057, -0.00000398, -0.00000295, 0.00000067, 0.00000343, 0.00000327, -0.00000046, -0.00000652, -0.00001030, -0.00000685, 0.00000282, 0.00001114, 0.00001204, 0.00000685, 0.00000059, -0.00000399, -0.00000619, -0.00000514, -0.00000207, -0.00000109, -0.00000299, -0.00000169, 0.00000691, 0.00001554, 0.00001266, -0.00000171, -0.00001438, -0.00001395, -0.00000138, 0.00001374, 0.00002062, 0.00001208, -0.00000873, -0.00002504, -0.00002024, 0.00000144, 0.00001690, 0.00001137, -0.00000527, -0.00001433, -0.00001058, -0.00000150, 0.00000599, 0.00000845, 0.00000362, -0.00000452, -0.00000526, 0.00000359, 0.00000793, -0.00000277, -0.00001612, -0.00001334, 0.00000150, 0.00000812, 0.00000167, -0.00000394, -0.00000328, -0.00000537, -0.00000961, -0.00000252, 0.00001291, 0.00001297, -0.00000717, -0.00001778, -0.00000040, 0.00001808, 0.00000745, -0.00001340, -0.00000650, 0.00001858, 0.00001729, -0.00001324, -0.00002411, 0.00000502, 0.00002765, 0.00000352, -0.00003371, -0.00002682, 0.00001551, 0.00003265, 0.00000438, -0.00002422, -0.00001614, 0.00000748, 0.00001026, -0.00000432, -0.00000934, -0.00000122, 0.00000148, -0.00000629, -0.00001058, -0.00000521, -0.00000007, -0.00000148, -0.00000310, -0.00000024, 0.00000188, -0.00000069, -0.00000321, -0.00000159, 0.00000043, -0.00000110, -0.00000324, -0.00000252, -0.00000173, -0.00000446, -0.00000747, -0.00000534, 0.00000055, 0.00000346, 0.00000053, -0.00000437, -0.00000674, -0.00000615, -0.00000445, -0.00000273, -0.00000132, -0.00000043, 0.00000028, 0.00000133, 0.00000177, -0.00000057, -0.00000579, -0.00000992, -0.00000835, -0.00000161, },
		{ -0.00001019, -0.00000726, -0.00000090, 0.00000278, 0.00000181, -0.00000065, -0.00000136, -0.00000025, 0.00000147, 0.00000273, 0.00000165, -0.00000327, -0.00000915, -0.00000976, -0.00000288, 0.00000626, 0.00001108, 0.00001023, 0.00000581, -0.00000059, -0.00000644, -0.00000727, -0.00000259, 0.00000094, -0.00000181, -0.00000537, -0.00000076, 0.00001020, 0.00001565, 0.00000878, -0.00000459, -0.00001382, -0.00001224, 0.00000067, 0.00001741, 0.00002284, 0.00000738, -0.00001755, -0.00002746, -0.00001319, 0.00000813, 0.00001524, 0.00000631, -0.00000591, -0.00001191, -0.00000962, -0.00000022, 0.00000964, 0.00000925, -0.00000172, -0.00000812, -0.00000035, 0.00000876, 0.00000282, -0.00001105, -0.00001315, -0.00000291, 0.00000271, -0.00000044, -0.00000000, 0.00000490, -0.00000009, -0.00001362, -0.00001309, 0.00000645, 0.00001700, 0.00000137, -0.00001519, -0.00000502, 0.00001469, 0.00001077, -0.00000808, -0.00000622, 0.00001457, 0.00001551, -0.00001017, -0.00002007, 0.00000611, 0.00002639, 0.00000208, -0.00003523, -0.00002901, 0.00001392, 0.00003320, 0.00000609, -0.00002375, -0.00001732, 0.00000633, 0.00000989, -0.00000486, -0.00001025, -0.00000088, 0.00000389, -0.00000390, -0.00001099, -0.00000815, -0.00000247, -0.00000128, -0.00000143, 0.00000051, 0.00000105, -0.00000169, -0.00000294, -0.00000027, 0.00000115, -0.00000234, -0.00000589, -0.00000411, -0.00000017, -0.00000058, -0.00000479, -0.00000646, -0.00000307, 0.00000116, 0.00000176, -0.00000112, -0.00000456, -0.00000645, -0.00000614, -0.00000411, -0.00000195, -0.00000120, -0.00000143, -0.00000052, 0.00000201, 0.00000317, -0.00000022, -0.00000662, },
		{ 0.00000448, -0.00000141, -0.00000835, -0.00001013, -0.00000618, -0.00000130, 0.00000094, 0.00000126, 0.00000128, 0.00000107, 0.00000086, 0.00000160, 0.00000200, -0.00000098, -0.00000679, -0.00000985, -0.00000635, 0.00000090, 0.00000690, 0.00001000, 0.00000985, 0.00000478, -0.00000377, -0.00000853, -0.00000483, 0.00000126, 0.00000069, -0.00000501, -0.00000574, 0.00000229, 0.00001165, 0.00001376, 0.00000694, -0.00000512, -0.00001502, -0.00001263, 0.00000498, 0.00002322, 0.00002132, -0.00000183, -0.00002335, -0.00002331, -0.00000646, 0.00000884, 0.00001254, 0.00000610, -0.00000523, -0.00001321, -0.00000874, 0.00000539, 0.00001236, 0.00000342, -0.00000723, -0.00000358, 0.00000646, 0.00000467, -0.00000643, -0.00000914, -0.00000228, -0.00000125, -0.00000684, -0.00000266, 0.00001081, 0.00001032, -0.00000973, -0.00002096, -0.00000501, 0.00001413, 0.00000835, -0.00000841, -0.00000577, 0.00001034, 0.00001046, -0.00000386, -0.00000312, 0.00001366, 0.00001338, -0.00001034, -0.00001850, 0.00000792, 0.00002797, 0.00000276, -0.00003700, -0.00003302, 0.00001092, 0.00003383, 0.00000867, -0.00002290, -0.00001890, 0.00000491, 0.00001008, -0.00000476, -0.00001174, -0.00000229, 0.00000511, -0.00000063, -0.00000903, -0.00000950, -0.00000553, -0.00000309, -0.00000095, 0.00000192, 0.00000162, -0.00000233, -0.00000372, 0.00000006, 0.00000251, -0.00000165, -0.00000741, -0.00000696, -0.00000141, 0.00000158, -0.00000101, -0.00000470, -0.00000483, -0.00000196, 0.00000048, 0.00000043, -0.00000189, -0.00000488, -0.00000643, -0.00000543, -0.00000311, -0.00000198, -0.00000265, -0.00000259, 0.00000048, 0.00000452, },
		{ -0.00000196, 0.00000378, 0.00000727, 0.00000384, -0.00000412, -0.00000970, -0.00000948, -0.00000598, -0.00000248, 0.00000051, 0.00000286, 0.00000325, 0.00000198, 0.00000145, 0.00000187, 0.00000004, -0.00000491, -0.00000857, -0.00000720, -0.00000270, 0.00000173, 0.00000630, 0.00001055, 0.00000965, 0.00000139, -0.00000700, -0.00000685, -0.00000056, 0.00000184, -0.00000252, -0.00000636, -0.00000371, 0.00000372, 0.00001115, 0.00001404, 0.00000738, -0.00000810, -0.00001892, -0.00001018, 0.00001277, 0.00002570, 0.00001423, -0.00000915, -0.00002256, -0.00001883, -0.00000527, 0.00000873, 0.00001473, 0.00000666, -0.00000909, -0.00001436, -0.00000222, 0.00001051, 0.00000710, -0.00000368, -0.00000349, 0.00000449, 0.00000334, -0.00000554, -0.00000542, 0.00000280, 0.00000036, -0.00001246, -0.00001122, 0.00000959, 0.00002014, 0.00000139, -0.00002109, -0.00001615, 0.00000460, 0.00000944, -0.00000107, -0.00000186, 0.00000842, 0.00000801, -0.00000274, 0.00000011, 0.00001586, 0.00001357, -0.00001156, -0.00002009, 0.00000777, 0.00003063, 0.00000630, -0.00003691, -0.00003765, 0.00000588, 0.00003322, 0.00001174, -0.00002099, -0.00002030, 0.00000287, 0.00001010, -0.00000392, -0.00001271, -0.00000462, 0.00000451, 0.00000184, -0.00000577, -0.00000856, -0.00000748, -0.00000589, -0.00000234, 0.00000252, 0.00000303, -0.00000176, -0.00000447, -0.00000072, 0.00000317, -0.00000001, -0.00000706, -0.00000913, -0.00000423, 0.00000118, 0.00000167, -0.00000151, -0.00000406, -0.00000387, -0.00000187, -0.00000016, -0.00000038, -0.00000266, -0.00000518, -0.00000572, -0.00000417, -0.00000289, -0.00000362, -0.00000450, },
		{ -0.00000429, -0.00000609, -0.00000475, 0.00000149, 0.00000792, 0.00000787, 0.00000095, -0.00000669, -0.00001001, -0.00000936, -0.00000672, -0.00000248, 0.00000223, 0.00000450, 0.00000366, 0.00000261, 0.00000259, 0.00000082, -0.00000389, -0.00000747, -0.00000662, -0.00000387, -0.00000211, 0.00000111, 0.00000758, 0.00001156, 0.00000676, -0.00000258, -0.00000638, -0.00000284, 0.00000054, -0.00000086, -0.00000400, -0.00000519, -0.00000334, 0.00000317, 0.00001327, 0.00001696, 0.00000473, -0.00001487, -0.00001999, -0.00000297, 0.00001814, 0.00002214, 0.00000806, -0.00001029, -0.00002142, -0.00001971, -0.00000428, 0.00001428, 0.00001735, 0.00000129, -0.00001317, -0.00000836, 0.00000494, 0.00000643, -0.00000109, -0.00000070, 0.00000584, 0.00000234, -0.00000831, -0.00000598, 0.00000808, 0.00000783, -0.00001208, -0.00002066, 0.00000052, 0.00002327, 0.00001456, -0.00001238, -0.00002096, -0.00000716, 0.00000331, 0.00000214, 0.00000418, 0.00001075, 0.00000660, -0.00000422, 0.00000101, 0.00001903, 0.00001688, -0.00001104, -0.00002323, 0.00000442, 0.00003173, 0.00001145, -0.00003379, -0.00004086, -0.00000042, 0.00003045, 0.00001409, -0.00001811, -0.00002073, 0.00000063, 0.00000942, -0.00000298, -0.00001283, -0.00000674, 0.00000267, 0.00000263, -0.00000283, -0.00000604, -0.00000735, -0.00000803, -0.00000495, 0.00000143, 0.00000399, -0.00000022, -0.00000421, -0.00000176, 0.00000273, 0.00000127, -0.00000551, -0.00000963, -0.00000685, -0.00000098, 0.00000218, 0.00000112, -0.00000179, -0.00000373, -0.00000353, -0.00000187, -0.00000058, -0.00000119, -0.00000330, -0.00000483, -0.00000443, -0.00000348, },
		{ -0.00000388, -0.00000335, -0.00000428, -0.00000702, -0.00000743, -0.00000183, 0.00000655, 0.00001003, 0.00000558, -0.00000226, -0.00000805, -0.00001057, -0.00001018, -0.00000631, -0.00000025, 0.00000387, 0.00000432, 0.00000396, 0.00000438, 0.00000257, -0.00000281, -0.00000699, -0.00000608, -0.00000355, -0.00000366, -0.00000320, 0.00000262, 0.00000969, 0.00000966, 0.00000278, -0.00000272, -0.00000307, -0.00000193, -0.00000183, -0.00000177, -0.00000279, -0.00000598, -0.00000528, 0.00000573, 0.00001885, 0.00001659, -0.00000291, -0.00001939, -0.00001550, 0.00000282, 0.00001760, 0.00001914, 0.00000773, -0.00001166, -0.00002624, -0.00002021, 0.00000392, 0.00002056, 0.00001233, -0.00000550, -0.00000873, 0.00000008, 0.00000204, -0.00000273, 0.00000117, 0.00001038, 0.00000531, -0.00001108, -0.00001153, 0.00000857, 0.00001702, -0.00000404, -0.00002467, -0.00001208, 0.00001716, 0.00002283, 0.00000118, -0.00001654, -0.00001486, -0.00000702, -0.00000143, 0.00000785, 0.00001601, 0.00000870, -0.00000569, -0.00000081, 0.00002056, 0.00002182, -0.00000737, -0.00002525, -0.00000119, 0.00002954, 0.00001572, -0.00002833, -0.00004114, -0.00000614, 0.00002585, 0.00001459, -0.00001524, -0.00001998, -0.00000101, 0.00000824, -0.00000257, -0.00001258, -0.00000811, 0.00000068, 0.00000205, -0.00000129, -0.00000343, -0.00000553, -0.00000838, -0.00000730, -0.00000101, 0.00000350, 0.00000122, -0.00000291, -0.00000208, 0.00000177, 0.00000158, -0.00000393, -0.00000880, -0.00000814, -0.00000334, 0.00000096, 0.00000203, 0.00000031, -0.00000218, -0.00000361, -0.00000324, -0.00000179, -0.00000100, -0.00000190, -0.00000347, },
		{ -0.00000137, -0.00000249, -0.00000318, -0.00000270, -0.00000339, -0.00000681, -0.00000929, -0.00000548, 0.00000348, 0.00000995, 0.00000883, 0.00000251, -0.00000429, -0.00000957, -0.00001208, -0.00000970, -0.00000340, 0.00000177, 0.00000320, 0.00000400, 0.00000605, 0.00000539, -0.00000050, -0.00000621, -0.00000605, -0.00000319, -0.00000366, -0.00000537, -0.00000192, 0.00000536, 0.00000876, 0.00000623, 0.00000261, 0.00000011, -0.00000292, -0.00000472, -0.00000215, 0.00000041, -0.00000391, -0.00000978, -0.00000389, 0.00001282, 0.00002147, 0.00001029, -0.00000915, -0.00001814, -0.00001184, 0.00000296, 0.00001793, 0.00002227, 0.00000639, -0.00002053, -0.00003073, -0.00001146, 0.00001415, 0.00001762, 0.00000386, -0.00000286, 0.00000048, -0.00000187, -0.00000854, -0.00000176, 0.00001438, 0.00001288, -0.00000925, -0.00001853, 0.00000227, 0.00002196, 0.00000831, -0.00001980, -0.00002126, 0.00000470, 0.00002181, 0.00001257, -0.00000516, -0.00001414, -0.00001563, -0.00001028, 0.00000571, 0.00002035, 0.00001407, -0.00000440, -0.00000340, 0.00001936, 0.00002587, -0.00000137, -0.00002425, -0.00000660, 0.00002437, 0.00001705, -0.00002271, -0.00003860, -0.00000953, 0.00002094, 0.00001314, -0.00001361, -0.00001875, -0.00000160, 0.00000732, -0.00000270, -0.00001264, -0.00000902, -0.00000077, 0.00000107, -0.00000109, -0.00000192, -0.00000333, -0.00000703, -0.00000817, -0.00000356, 0.00000161, 0.00000153, -0.00000142, -0.00000145, 0.00000119, 0.00000125, -0.00000305, -0.00000763, -0.00000815, -0.00000475, -0.00000070, 0.00000139, 0.00000112, -0.00000059, -0.00000252, -0.00000347, -0.00000287, -0.00000160, },
		{ -0.00000320, -0.00000208, -0.00000127, -0.00000209, -0.00000298, -0.00000227, -0.00000206, -0.00000533, -0.00000958, -0.00000843, -0.00000056, 0.00000760, 0.00000990, 0.00000656, 0.00000052, -0.00000662, -0.00001224, -0.00001210, -0.00000622, -0.00000066, 0.00000088, 0.00000211, 0.00000605, 0.00000796, 0.00000309, -0.00000393, -0.00000562, -0.00000332, -0.00000356, -0.00000592, -0.00000447, 0.00000101, 0.00000494, 0.00000591, 0.00000670, 0.00000573, -0.00000047, -0.00000695, -0.00000520, 0.00000130, 0.00000011, -0.00000894, -0.00001042, 0.00000296, 0.00001758, 0.00001768, 0.00000475, -0.00000968, -0.00001779, -0.00001402, 0.00000469, 0.00002574, 0.00002415, -0.00000472, -0.00003027, -0.00002396, 0.00000167, 0.00001417, 0.00000859, 0.00000503, 0.00000728, -0.00000025, -0.00001436, -0.00001026, 0.00001301, 0.00002145, -0.00000089, -0.00002139, -0.00000805, 0.00001889, 0.00001843, -0.00000789, -0.00002175, -0.00000727, 0.00001245, 0.00001594, 0.00000682, -0.00000542, -0.00001727, -0.00001935, -0.00000216, 0.00001992, 0.00001966, 0.00000028, -0.00000405, 0.00001667, 0.00002742, 0.00000472, -0.00002036, -0.00000961, 0.00001826, 0.00001510, -0.00001918, -0.00003497, -0.00001017, 0.00001753, 0.00001085, -0.00001366, -0.00001821, -0.00000154, 0.00000724, -0.00000274, -0.00001319, -0.00001014, -0.00000179, 0.00000049, -0.00000140, -0.00000170, -0.00000196, -0.00000503, -0.00000738, -0.00000500, -0.00000070, 0.00000053, -0.00000070, -0.00000045, 0.00000135, 0.00000102, -0.00000278, -0.00000687, -0.00000765, -0.00000506, -0.00000173, 0.00000023, 0.00000064, 0.00000004, -0.00000133, -0.00000285, },
		{ -0.00000083, -0.00000220, -0.00000304, -0.00000205, -0.00000086, -0.00000170, -0.00000325, -0.00000267, -0.00000128, -0.00000328, -0.00000811, -0.00000953, -0.00000421, 0.00000357, 0.00000826, 0.00000870, 0.00000539, -0.00000203, -0.00001056, -0.00001336, -0.00000843, -0.00000259, -0.00000139, -0.00000096, 0.00000373, 0.00000853, 0.00000651, -0.00000003, -0.00000342, -0.00000304, -0.00000404, -0.00000633, -0.00000528, -0.00000146, 0.00000060, 0.00000215, 0.00000715, 0.00001096, 0.00000519, -0.00000594, -0.00000858, -0.00000105, 0.00000244, -0.00000495, -0.00001126, -0.00000498, 0.00000806, 0.00001574, 0.00001453, 0.00000511, -0.00001169, -0.00002442, -0.00001418, 0.00001619, 0.00003318, 0.00001378, -0.00001923, -0.00002742, -0.00000967, 0.00000454, 0.00000511, 0.00000852, 0.00001661, 0.00000889, -0.00001435, -0.00002045, 0.00000415, 0.00002524, 0.00001119, -0.00001642, -0.00001627, 0.00000908, 0.00002083, 0.00000478, -0.00001341, -0.00001218, 0.00000066, 0.00001023, 0.00001276, 0.00000613, -0.00001070, -0.00002304, -0.00001177, 0.00001368, 0.00002159, 0.00000615, -0.00000145, 0.00001498, 0.00002700, 0.00000896, -0.00001542, -0.00000971, 0.00001354, 0.00001140, -0.00001860, -0.00003246, -0.00000929, 0.00001641, 0.00000942, -0.00001468, -0.00001907, -0.00000188, 0.00000787, -0.00000201, -0.00001374, -0.00001181, -0.00000302, 0.00000035, -0.00000139, -0.00000208, -0.00000181, -0.00000360, -0.00000575, -0.00000493, -0.00000228, -0.00000106, -0.00000106, 0.00000008, 0.00000189, 0.00000129, -0.00000256, -0.00000658, -0.00000734, -0.00000481, -0.00000185, -0.00000052, -0.00000037, -0.00000035, },
		{ -0.00000112, -0.00000115, -0.00000104, -0.00000213, -0.00000314, -0.00000194, 0.00000002, -0.00000063, -0.00000323, -0.00000368, -0.00000173, -0.00000192, -0.00000573, -0.00000835, -0.00000591, -0.00000059, 0.00000419, 0.00000787, 0.00000891, 0.00000349, -0.00000691, -0.00001320, -0.00001012, -0.00000404, -0.00000280, -0.00000369, 0.00000005, 0.00000632, 0.00000770, 0.00000397, 0.00000087, -0.00000078, -0.00000427, -0.00000754, -0.00000585, -0.00000196, -0.00000201, -0.00000279, 0.00000365, 0.00001292, 0.00001148, -0.00000115, -0.00000923, -0.00000477, 0.00000118, -0.00000178, -0.00000772, -0.00000712, -0.00000113, 0.00000622, 0.00001492, 0.00001830, 0.00000354, -0.00002288, -0.00002967, -0.00000184, 0.00002976, 0.00002651, -0.00000365, -0.00002101, -0.00001348, -0.00000454, -0.00000468, 0.00000328, 0.00002159, 0.00002222, -0.00000514, -0.00002565, -0.00000926, 0.00002030, 0.00002073, -0.00000484, -0.00001696, -0.00000170, 0.00001473, 0.00001151, -0.00000136, -0.00000779, -0.00000640, -0.00000016, 0.00000984, 0.00001374, 0.00000032, -0.00001882, -0.00001762, 0.00000448, 0.00001797, 0.00000957, 0.00000305, 0.00001605, 0.00002673, 0.00001122, -0.00001148, -0.00000824, 0.00001122, 0.00000818, -0.00002001, -0.00003234, -0.00000893, 0.00001685, 0.00000992, -0.00001516, -0.00002094, -0.00000348, 0.00000832, -0.00000042, -0.00001357, -0.00001368, -0.00000490, 0.00000008, -0.00000086, -0.00000224, -0.00000239, -0.00000326, -0.00000441, -0.00000396, -0.00000263, -0.00000221, -0.00000199, -0.00000022, 0.00000212, 0.00000185, -0.00000201, -0.00000639, -0.00000737, -0.00000459, -0.00000142, -0.00000055, },
		{ -0.00000088, -0.00000010, -0.00000139, -0.00000177, -0.00000142, -0.00000247, -0.00000387, -0.00000248, 0.00000074, 0.00000111, -0.00000209, -0.00000425, -0.00000311, -0.00000222, -0.00000406, -0.00000577, -0.00000480, -0.00000285, -0.00000075, 0.00000387, 0.00000947, 0.00000846, -0.00000152, -0.00001103, -0.00001100, -0.00000545, -0.00000360, -0.00000512, -0.00000316, 0.00000236, 0.00000560, 0.00000565, 0.00000572, 0.00000412, -0.00000244, -0.00000888, -0.00000746, -0.00000194, -0.00000232, -0.00000630, -0.00000182, 0.00001058, 0.00001513, 0.00000534, -0.00000571, -0.00000648, -0.00000253, -0.00000240, -0.00000382, -0.00000347, -0.00000475, -0.00000522, 0.00000575, 0.00002306, 0.00002058, -0.00001007, -0.00003493, -0.00002010, 0.00001643, 0.00002865, 0.00000853, -0.00000966, -0.00000824, -0.00000643, -0.00001414, -0.00000947, 0.00001674, 0.00003235, 0.00001108, -0.00002070, -0.00002024, 0.00000780, 0.00002199, 0.00000750, -0.00000919, -0.00000680, 0.00000470, 0.00000941, 0.00000721, 0.00000243, -0.00000474, -0.00000815, 0.00000078, 0.00001326, 0.00000939, -0.00000898, -0.00001623, -0.00000256, 0.00001055, 0.00000807, 0.00000620, 0.00001929, 0.00002859, 0.00001315, -0.00000918, -0.00000717, 0.00001048, 0.00000677, -0.00002139, -0.00003411, -0.00001057, 0.00001706, 0.00001203, -0.00001387, -0.00002256, -0.00000628, 0.00000767, 0.00000136, -0.00001236, -0.00001503, -0.00000727, -0.00000083, -0.00000025, -0.00000184, -0.00000291, -0.00000371, -0.00000396, -0.00000307, -0.00000215, -0.00000240, -0.00000264, -0.00000100, 0.00000170, 0.00000218, -0.00000126, -0.00000601, -0.00000752, -0.00000458, },
		{ -0.00000762, -0.00000465, -0.00000034, 0.00000055, -0.00000135, -0.00000212, -0.00000154, -0.00000264, -0.00000487, -0.00000392, 0.00000039, 0.00000262, 0.00000008, -0.00000331, -0.00000406, -0.00000389, -0.00000445, -0.00000376, -0.00000164, -0.00000183, -0.00000413, -0.00000195, 0.00000624, 0.00001085, 0.00000428, -0.00000655, -0.00001014, -0.00000672, -0.00000465, -0.00000561, -0.00000460, -0.00000117, 0.00000112, 0.00000346, 0.00000840, 0.00001035, 0.00000257, -0.00000839, -0.00000974, -0.00000301, -0.00000163, -0.00000725, -0.00000638, 0.00000543, 0.00001428, 0.00001000, 0.00000062, -0.00000377, -0.00000526, -0.00000675, -0.00000358, 0.00000234, -0.00000117, -0.00001231, -0.00000838, 0.00001703, 0.00003171, 0.00000816, -0.00002844, -0.00003115, 0.00000016, 0.00002124, 0.00001210, -0.00000083, 0.00000169, 0.00000096, -0.00001596, -0.00002304, 0.00000202, 0.00003222, 0.00002693, -0.00000582, -0.00002185, -0.00000586, 0.00001384, 0.00001366, 0.00000252, -0.00000314, -0.00000227, 0.00000128, 0.00000773, 0.00001162, 0.00000402, -0.00000854, -0.00000801, 0.00000573, 0.00001148, 0.00000076, -0.00000849, -0.00000337, 0.00000380, 0.00000234, 0.00000522, 0.00002203, 0.00003265, 0.00001678, -0.00000737, -0.00000751, 0.00000952, 0.00000686, -0.00002120, -0.00003606, -0.00001418, 0.00001542, 0.00001425, -0.00001076, -0.00002263, -0.00000929, 0.00000566, 0.00000240, -0.00001055, -0.00001538, -0.00000946, -0.00000239, -0.00000013, -0.00000121, -0.00000296, -0.00000423, -0.00000423, -0.00000287, -0.00000167, -0.00000189, -0.00000252, -0.00000157, 0.00000089, 0.00000201, -0.00000069, -0.00000553, },
		{ -0.00000045, -0.00000526, -0.00000783, -0.00000478, 0.00000029, 0.00000144, -0.00000114, -0.00000243, -0.00000149, -0.00000232, -0.00000540, -0.00000568, -0.00000126, 0.00000274, 0.00000212, -0.00000086, -0.00000314, -0.00000531, -0.00000681, -0.00000415, 0.00000138, 0.00000208, -0.00000391, -0.00000688, 0.00000018, 0.00000912, 0.00000819, -0.00000067, -0.00000664, -0.00000675, -0.00000625, -0.00000646, -0.00000457, -0.00000244, -0.00000321, -0.00000197, 0.00000674, 0.00001492, 0.00000983, -0.00000440, -0.00001071, -0.00000531, -0.00000189, -0.00000665, -0.00000824, 0.00000054, 0.00000974, 0.00001004, 0.00000617, 0.00000308, -0.00000357, -0.00001179, -0.00000846, 0.00000551, 0.00000715, -0.00001148, -0.00002085, 0.00000294, 0.00003259, 0.00002425, -0.00001407, -0.00003193, -0.00001191, 0.00001000, 0.00000732, 0.00000015, 0.00000886, 0.00001359, -0.00000701, -0.00002908, -0.00001630, 0.00001979, 0.00003386, 0.00001255, -0.00001188, -0.00001268, 0.00000111, 0.00001024, 0.00001093, 0.00000667, -0.00000134, -0.00000632, 0.00000105, 0.00001406, 0.00001387, -0.00000119, -0.00001057, -0.00000340, 0.00000609, 0.00000488, 0.00000076, 0.00000239, 0.00000212, -0.00000388, -0.00000009, 0.00002133, 0.00003683, 0.00002270, -0.00000412, -0.00000860, 0.00000708, 0.00000687, -0.00001934, -0.00003651, -0.00001834, 0.00001157, 0.00001492, -0.00000697, -0.00002075, -0.00001125, 0.00000295, 0.00000220, -0.00000899, -0.00001491, -0.00001091, -0.00000407, -0.00000063, -0.00000089, -0.00000274, -0.00000440, -0.00000462, -0.00000330, -0.00000177, -0.00000136, -0.00000180, -0.00000151, 0.00000024, 0.00000157, },
		{ 0.00000008, 0.00000133, -0.00000027, -0.00000521, -0.00000840, -0.00000525, 0.00000089, 0.00000268, -0.00000058, -0.00000278, -0.00000167, -0.00000177, -0.00000510, -0.00000675, -0.00000341, 0.00000109, 0.00000249, 0.00000172, -0.00000010, -0.00000456, -0.00000935, -0.00000726, 0.00000181, 0.00000657, 0.00000007, -0.00000820, -0.00000595, 0.00000359, 0.00000795, 0.00000421, -0.00000080, -0.00000403, -0.00000734, -0.00000852, -0.00000470, -0.00000145, -0.00000495, -0.00000794, 0.00000083, 0.00001500, 0.00001633, 0.00000276, -0.00000817, -0.00000700, -0.00000388, -0.00000672, -0.00000787, -0.00000151, 0.00000458, 0.00000528, 0.00000725, 0.00001087, 0.00000367, -0.00001333, -0.00001648, 0.00000289, 0.00001530, -0.00000335, -0.00002650, -0.00001329, 0.00002369, 0.00003265, 0.00000168, -0.00002453, -0.00001613, 0.00000144, -0.00000075, -0.00000676, 0.00000747, 0.00002353, 0.00000925, -0.00002273, -0.00002886, 0.00000007, 0.00002741, 0.00002507, 0.00000514, -0.00000805, -0.00000807, -0.00000010, 0.00001087, 0.00001599, 0.00000693, -0.00000748, -0.00000727, 0.00000879, 0.00001851, 0.00000909, -0.00000522, -0.00000790, -0.00000249, 0.00000160, 0.00000590, 0.00001095, 0.00000689, -0.00000616, -0.00000681, 0.00001632, 0.00003828, 0.00002944, 0.00000158, -0.00000870, 0.00000334, 0.00000535, -0.00001705, -0.00003496, -0.00002135, 0.00000654, 0.00001333, -0.00000408, -0.00001768, -0.00001150, 0.00000067, 0.00000110, -0.00000830, -0.00001432, -0.00001158, -0.00000526, -0.00000135, -0.00000110, -0.00000274, -0.00000431, -0.00000471, -0.00000386, -0.00000244, -0.00000133, -0.00000106, -0.00000098, },
		{ -0.00000073, -0.00000049, 0.00000018, 0.00000147, 0.00000028, -0.00000498, -0.00000927, -0.00000639, 0.00000102, 0.00000413, 0.00000063, -0.00000280, -0.00000215, -0.00000162, -0.00000441, -0.00000662, -0.00000472, -0.00000138, 0.00000055, 0.00000246, 0.00000358, -0.00000098, -0.00000966, -0.00001132, -0.00000123, 0.00000863, 0.00000560, -0.00000502, -0.00000879, -0.00000313, 0.00000306, 0.00000521, 0.00000516, 0.00000187, -0.00000592, -0.00001106, -0.00000655, -0.00000003, -0.00000352, -0.00001134, -0.00000667, 0.00000991, 0.00001856, 0.00001025, -0.00000164, -0.00000539, -0.00000617, -0.00000903, -0.00000760, -0.00000031, 0.00000223, -0.00000166, 0.00000225, 0.00001522, 0.00001439, -0.00000860, -0.00002342, -0.00000533, 0.00001879, 0.00000815, -0.00002374, -0.00002581, 0.00000928, 0.00003193, 0.00001319, -0.00001403, -0.00001353, -0.00000084, -0.00000586, -0.00001652, -0.00000296, 0.00002387, 0.00002444, -0.00000609, -0.00002874, -0.00001733, 0.00001039, 0.00002502, 0.00002013, 0.00000593, -0.00000734, -0.00001004, 0.00000274, 0.00001857, 0.00001707, -0.00000085, -0.00001084, 0.00000006, 0.00001545, 0.00001590, 0.00000458, -0.00000479, -0.00000813, -0.00000610, 0.00000407, 0.00001686, 0.00001541, -0.00000234, -0.00001073, 0.00000905, 0.00003556, 0.00003436, 0.00000885, -0.00000652, -0.00000026, 0.00000205, -0.00001585, -0.00003224, -0.00002237, 0.00000190, 0.00001007, -0.00000304, -0.00001476, -0.00001041, -0.00000040, 0.00000001, -0.00000839, -0.00001424, -0.00001199, -0.00000580, -0.00000173, -0.00000156, -0.00000323, -0.00000440, -0.00000454, -0.00000415, -0.00000325, -0.00000180, },
		{ -0.00000387, -0.00000239, -0.00000073, -0.00000034, 0.00000005, 0.00000166, 0.00000132, -0.00000408, -0.00000986, -0.00000806, 0.00000022, 0.00000514, 0.00000233, -0.00000197, -0.00000242, -0.00000208, -0.00000423, -0.00000584, -0.00000440, -0.00000296, -0.00000275, 0.00000026, 0.00000548, 0.00000404, -0.00000645, -0.00001339, -0.00000616, 0.00000664, 0.00000919, 0.00000081, -0.00000655, -0.00000717, -0.00000428, 0.00000086, 0.00000787, 0.00000924, -0.00000069, -0.00001180, -0.00000998, -0.00000045, -0.00000070, -0.00001079, -0.00001203, 0.00000199, 0.00001488, 0.00001400, 0.00000653, 0.00000100, -0.00000574, -0.00001297, -0.00000992, 0.00000208, 0.00000436, -0.00000673, -0.00000732, 0.00001291, 0.00002425, 0.00000216, -0.00002521, -0.00001599, 0.00001573, 0.00001817, -0.00001459, -0.00003131, -0.00000534, 0.00002439, 0.00001808, -0.00000532, -0.00000825, 0.00000229, -0.00000447, -0.00002244, -0.00001726, 0.00001299, 0.00003012, 0.00001263, -0.00001585, -0.00002364, -0.00000834, 0.00001255, 0.00002479, 0.00002142, 0.00000372, -0.00001224, -0.00000759, 0.00001278, 0.00002244, 0.00000958, -0.00000675, -0.00000581, 0.00000745, 0.00001551, 0.00001274, 0.00000389, -0.00000689, -0.00001252, -0.00000303, 0.00001648, 0.00002269, 0.00000600, -0.00000909, 0.00000334, 0.00002990, 0.00003569, 0.00001547, -0.00000235, -0.00000232, -0.00000203, -0.00001637, -0.00002977, -0.00002178, -0.00000127, 0.00000646, -0.00000367, -0.00001305, -0.00000906, -0.00000034, -0.00000032, -0.00000863, -0.00001480, -0.00001273, -0.00000601, -0.00000149, -0.00000174, -0.00000403, -0.00000496, -0.00000448, -0.00000419, },
		{ -0.00000482, -0.00000431, -0.00000433, -0.00000282, -0.00000069, -0.00000037, -0.00000055, 0.00000134, 0.00000233, -0.00000244, -0.00000952, -0.00000954, -0.00000144, 0.00000503, 0.00000375, -0.00000032, -0.00000170, -0.00000253, -0.00000504, -0.00000564, -0.00000300, -0.00000243, -0.00000531, -0.00000408, 0.00000388, 0.00000768, -0.00000081, -0.00001146, -0.00000970, 0.00000170, 0.00000833, 0.00000544, -0.00000048, -0.00000574, -0.00000991, -0.00000739, 0.00000483, 0.00001447, 0.00000742, -0.00000855, -0.00001270, -0.00000337, 0.00000060, -0.00000773, -0.00001288, -0.00000453, 0.00000696, 0.00001121, 0.00001196, 0.00001077, -0.00000001, -0.00001565, -0.00001549, 0.00000249, 0.00000995, -0.00000672, -0.00001753, 0.00000362, 0.00002877, 0.00001573, -0.00002002, -0.00002473, 0.00000733, 0.00002325, -0.00000304, -0.00002968, -0.00001619, 0.00001398, 0.00001725, -0.00000070, -0.00000437, 0.00000686, 0.00000241, -0.00002040, -0.00002782, -0.00000427, 0.00002286, 0.00002370, 0.00000272, -0.00001621, -0.00001893, -0.00000517, 0.00001669, 0.00002945, 0.00001909, -0.00000411, -0.00001256, 0.00000276, 0.00001978, 0.00001736, 0.00000264, -0.00000467, 0.00000038, 0.00000916, 0.00001464, 0.00001266, 0.00000060, -0.00001303, -0.00001037, 0.00001029, 0.00002460, 0.00001450, -0.00000246, 0.00000228, 0.00002456, 0.00003390, 0.00001950, 0.00000209, -0.00000269, -0.00000558, -0.00001805, -0.00002836, -0.00002073, -0.00000300, 0.00000362, -0.00000496, -0.00001264, -0.00000841, 0.00000007, 0.00000025, -0.00000833, -0.00001558, -0.00001405, -0.00000641, -0.00000079, -0.00000133, -0.00000469, -0.00000589, },
		{ -0.00000487, -0.00000681, -0.00000547, -0.00000475, -0.00000496, -0.00000320, -0.00000038, -0.00000013, -0.00000132, 0.00000027, 0.00000261, -0.00000070, -0.00000813, -0.00000993, -0.00000307, 0.00000375, 0.00000390, 0.00000126, 0.00000023, -0.00000189, -0.00000620, -0.00000684, -0.00000204, -0.00000015, -0.00000538, -0.00000815, -0.00000074, 0.00000751, 0.00000411, -0.00000613, -0.00000901, -0.00000265, 0.00000345, 0.00000544, 0.00000546, 0.00000082, -0.00001002, -0.00001534, -0.00000354, 0.00001408, 0.00001490, -0.00000122, -0.00001175, -0.00000688, -0.00000131, -0.00000555, -0.00000993, -0.00000610, -0.00000074, 0.00000239, 0.00001054, 0.00001972, 0.00001114, -0.00001358, -0.00002236, -0.00000131, 0.00001584, -0.00000090, -0.00002402, -0.00000996, 0.00002553, 0.00002747, -0.00000900, -0.00002820, -0.00000312, 0.00002248, 0.00000712, -0.00002315, -0.00002176, 0.00000411, 0.00001323, 0.00000032, -0.00000366, 0.00000891, 0.00001034, -0.00001122, -0.00002922, -0.00001953, 0.00000626, 0.00002153, 0.00001670, -0.00000018, -0.00001638, -0.00001816, 0.00000077, 0.00002545, 0.00002997, 0.00001039, -0.00000791, -0.00000433, 0.00001143, 0.00001804, 0.00001149, 0.00000270, -0.00000082, 0.00000196, 0.00001009, 0.00001642, 0.00000975, -0.00000702, -0.00001303, 0.00000244, 0.00002053, 0.00001883, 0.00000558, 0.00000616, 0.00002269, 0.00003159, 0.00002086, 0.00000494, -0.00000257, -0.00000819, -0.00001984, -0.00002785, -0.00002006, -0.00000409, 0.00000175, -0.00000591, -0.00001283, -0.00000876, -0.00000002, 0.00000118, -0.00000727, -0.00001601, -0.00001572, -0.00000730, -0.00000005, -0.00000036, },
		{ 0.00000046, 0.00000091, -0.00000455, -0.00000743, -0.00000613, -0.00000541, -0.00000595, -0.00000391, -0.00000000, 0.00000062, -0.00000166, -0.00000113, 0.00000184, 0.00000019, -0.00000645, -0.00000899, -0.00000353, 0.00000227, 0.00000252, 0.00000149, 0.00000247, 0.00000042, -0.00000634, -0.00000898, -0.00000283, 0.00000210, -0.00000296, -0.00000958, -0.00000583, 0.00000345, 0.00000518, -0.00000061, -0.00000390, -0.00000273, -0.00000195, 0.00000027, 0.00000693, 0.00000873, -0.00000379, -0.00001828, -0.00001339, 0.00000724, 0.00001743, 0.00000714, -0.00000582, -0.00000740, -0.00000525, -0.00000716, -0.00000678, -0.00000222, -0.00000343, -0.00000836, 0.00000107, 0.00002272, 0.00002453, -0.00000481, -0.00002686, -0.00000912, 0.00001859, 0.00000874, -0.00002400, -0.00002341, 0.00001506, 0.00003347, 0.00000443, -0.00002535, -0.00001202, 0.00001721, 0.00001361, -0.00001459, -0.00002251, -0.00000342, 0.00000833, -0.00000041, -0.00000523, 0.00000695, 0.00001470, 0.00000023, -0.00002131, -0.00002577, -0.00001115, 0.00000781, 0.00001859, 0.00001453, -0.00000347, -0.00001965, -0.00001370, 0.00001200, 0.00003028, 0.00002327, 0.00000448, -0.00000311, 0.00000365, 0.00001217, 0.00001462, 0.00001145, 0.00000459, -0.00000078, 0.00000316, 0.00001369, 0.00001530, 0.00000200, -0.00000929, -0.00000216, 0.00001360, 0.00001750, 0.00001081, 0.00001228, 0.00002524, 0.00003174, 0.00002148, 0.00000577, -0.00000359, -0.00001066, -0.00002125, -0.00002746, -0.00001983, -0.00000531, 0.00000035, -0.00000615, -0.00001278, -0.00000963, -0.00000094, 0.00000181, -0.00000574, -0.00001585, -0.00001736, -0.00000863, },
		{ -0.00001880, -0.00001021, 0.00000069, 0.00000227, -0.00000386, -0.00000771, -0.00000657, -0.00000599, -0.00000716, -0.00000513, -0.00000002, 0.00000160, -0.00000126, -0.00000206, 0.00000054, -0.00000023, -0.00000571, -0.00000754, -0.00000229, 0.00000206, 0.00000056, -0.00000028, 0.00000342, 0.00000362, -0.00000443, -0.00001043, -0.00000522, 0.00000246, 0.00000006, -0.00000783, -0.00000836, -0.00000195, 0.00000151, 0.00000117, 0.00000253, 0.00000264, -0.00000337, -0.00000690, 0.00000220, 0.00001289, 0.00000558, -0.00001402, -0.00001928, -0.00000296, 0.00001276, 0.00001163, 0.00000294, -0.00000222, -0.00000738, -0.00001263, -0.00000763, 0.00000393, 0.00000121, -0.00001525, -0.00001362, 0.00001652, 0.00003486, 0.00000941, -0.00002561, -0.00001834, 0.00001630, 0.00001866, -0.00001733, -0.00003253, 0.00000042, 0.00003194, 0.00001621, -0.00001749, -0.00001696, 0.00000992, 0.00001592, -0.00000636, -0.00001979, -0.00000817, 0.00000367, -0.00000145, -0.00000700, 0.00000240, 0.00001354, 0.00000825, -0.00000890, -0.00002119, -0.00002098, -0.00000964, 0.00000788, 0.00001915, 0.00001101, -0.00001008, -0.00001888, -0.00000292, 0.00002035, 0.00002757, 0.00001788, 0.00000650, 0.00000246, 0.00000487, 0.00001102, 0.00001610, 0.00001291, 0.00000295, -0.00000084, 0.00000736, 0.00001474, 0.00000894, -0.00000153, -0.00000094, 0.00000846, 0.00001260, 0.00001112, 0.00001677, 0.00003032, 0.00003570, 0.00002406, 0.00000597, -0.00000635, -0.00001435, -0.00002280, -0.00002669, -0.00001949, -0.00000682, -0.00000122, -0.00000601, -0.00001209, -0.00001038, -0.00000246, 0.00000181, -0.00000422, -0.00001525, },
		{ -0.00000291, -0.00001444, -0.00002011, -0.00001201, 0.00000067, 0.00000372, -0.00000286, -0.00000773, -0.00000680, -0.00000631, -0.00000825, -0.00000667, -0.00000069, 0.00000221, -0.00000044, -0.00000206, -0.00000027, -0.00000141, -0.00000660, -0.00000709, -0.00000016, 0.00000400, -0.00000019, -0.00000322, 0.00000197, 0.00000586, -0.00000095, -0.00000964, -0.00000745, 0.00000063, 0.00000127, -0.00000475, -0.00000706, -0.00000502, -0.00000458, -0.00000264, 0.00000568, 0.00001082, 0.00000156, -0.00001094, -0.00000629, 0.00001019, 0.00001245, -0.00000481, -0.00001755, -0.00001063, 0.00000281, 0.00000842, 0.00000943, 0.00000797, -0.00000353, -0.00001850, -0.00001440, 0.00000718, 0.00001136, -0.00001389, -0.00002768, 0.00000173, 0.00003723, 0.00002494, -0.00001750, -0.00002529, 0.00000949, 0.00002553, -0.00000619, -0.00003492, -0.00001423, 0.00002367, 0.00002316, -0.00000740, -0.00001722, 0.00000304, 0.00001477, 0.00000007, -0.00001502, -0.00001022, -0.00000041, -0.00000257, -0.00000766, -0.00000192, 0.00000841, 0.00000980, 0.00000145, -0.00000981, -0.00001956, -0.00002160, -0.00000895, 0.00001109, 0.00001771, 0.00000337, -0.00001316, -0.00001110, 0.00000650, 0.00002168, 0.00002503, 0.00001927, 0.00000957, 0.00000221, 0.00000440, 0.00001435, 0.00001884, 0.00001079, 0.00000110, 0.00000239, 0.00000972, 0.00001058, 0.00000567, 0.00000482, 0.00000818, 0.00000824, 0.00000766, 0.00001713, 0.00003453, 0.00004226, 0.00003012, 0.00000788, -0.00000987, -0.00001994, -0.00002561, -0.00002591, -0.00001861, -0.00000826, -0.00000319, -0.00000596, -0.00001088, -0.00001063, -0.00000412, 0.00000127, },
		{ -0.00000569, 0.00000044, -0.00000173, -0.00001343, -0.00002136, -0.00001420, 0.00000021, 0.00000517, -0.00000148, -0.00000741, -0.00000688, -0.00000647, -0.00000905, -0.00000809, -0.00000175, 0.00000210, 0.00000001, -0.00000147, 0.00000011, -0.00000207, -0.00000864, -0.00000873, 0.00000105, 0.00000744, 0.00000161, -0.00000529, -0.00000135, 0.00000549, 0.00000209, -0.00000666, -0.00000743, -0.00000166, -0.00000031, -0.00000331, -0.00000352, -0.00000360, -0.00000887, -0.00001027, 0.00000251, 0.00001671, 0.00001102, -0.00000814, -0.00001300, 0.00000202, 0.00001227, 0.00000348, -0.00000907, -0.00001054, -0.00000655, -0.00000232, 0.00000821, 0.00001855, 0.00000770, -0.00001953, -0.00002503, 0.00000343, 0.00002186, -0.00000370, -0.00003506, -0.00001719, 0.00002961, 0.00003649, -0.00000440, -0.00002719, 0.00000056, 0.00002757, 0.00000611, -0.00003042, -0.00002513, 0.00001135, 0.00002392, 0.00000184, -0.00001365, -0.00000173, 0.00001147, 0.00000400, -0.00000943, -0.00000974, -0.00000359, -0.00000420, -0.00000748, -0.00000412, 0.00000274, 0.00000587, 0.00000535, 0.00000145, -0.00000955, -0.00002272, -0.00002255, -0.00000485, 0.00001234, 0.00001144, -0.00000189, -0.00000930, -0.00000330, 0.00001008, 0.00002281, 0.00002825, 0.00002144, 0.00000718, 0.00000046, 0.00000844, 0.00001915, 0.00001793, 0.00000791, 0.00000243, 0.00000458, 0.00000740, 0.00000866, 0.00001126, 0.00001248, 0.00000772, 0.00000386, 0.00001377, 0.00003529, 0.00004858, 0.00003901, 0.00001309, -0.00001219, -0.00002677, -0.00003037, -0.00002599, -0.00001742, -0.00000932, -0.00000535, -0.00000617, -0.00000943, -0.00001040, },
		{ -0.00000779, -0.00000975, -0.00000721, -0.00000070, -0.00000061, -0.00001201, -0.00002232, -0.00001676, -0.00000100, 0.00000625, 0.00000022, -0.00000652, -0.00000665, -0.00000660, -0.00000976, -0.00000922, -0.00000263, 0.00000155, -0.00000044, -0.00000130, 0.00000139, -0.00000116, -0.00001030, -0.00001215, -0.00000021, 0.00001035, 0.00000566, -0.00000460, -0.00000435, 0.00000259, 0.00000247, -0.00000351, -0.00000458, -0.00000178, -0.00000314, -0.00000529, -0.00000118, 0.00000128, -0.00000812, -0.00001714, -0.00000607, 0.00001601, 0.00001968, 0.00000084, -0.00001287, -0.00000604, 0.00000494, 0.00000480, 0.00000053, -0.00000162, -0.00000852, -0.00001546, -0.00000289, 0.00002293, 0.00002327, -0.00001189, -0.00003423, -0.00000773, 0.00002693, 0.00001166, -0.00003241, -0.00003387, 0.00001389, 0.00004011, 0.00000968, -0.00002341, -0.00000753, 0.00002492, 0.00001650, -0.00002083, -0.00003001, -0.00000150, 0.00001902, 0.00000778, -0.00000804, -0.00000361, 0.00000749, 0.00000526, -0.00000428, -0.00000702, -0.00000514, -0.00000639, -0.00000770, -0.00000435, -0.00000064, 0.00000013, 0.00000272, 0.00000699, 0.00000230, -0.00001406, -0.00002621, -0.00001972, -0.00000207, 0.00000855, 0.00000625, -0.00000102, -0.00000464, -0.00000002, 0.00001382, 0.00002922, 0.00003185, 0.00001794, 0.00000305, 0.00000335, 0.00001463, 0.00002059, 0.00001562, 0.00000762, 0.00000325, 0.00000288, 0.00000693, 0.00001464, 0.00001831, 0.00001156, 0.00000301, 0.00000941, 0.00003239, 0.00005213, 0.00004844, 0.00002136, -0.00001180, -0.00003328, -0.00003664, -0.00002750, -0.00001654, -0.00001020, -0.00000760, -0.00000659, },
		{ -0.00001023, -0.00000729, -0.00000586, -0.00000850, -0.00000859, -0.00000233, 0.00000019, -0.00001013, -0.00002254, -0.00001927, -0.00000304, 0.00000641, 0.00000172, -0.00000502, -0.00000569, -0.00000649, -0.00001068, -0.00001036, -0.00000305, 0.00000133, -0.00000150, -0.00000233, 0.00000236, 0.00000119, -0.00001010, -0.00001562, -0.00000398, 0.00001064, 0.00000978, -0.00000097, -0.00000454, -0.00000064, -0.00000036, -0.00000296, -0.00000075, 0.00000166, -0.00000411, -0.00000965, -0.00000280, 0.00000587, -0.00000253, -0.00001888, -0.00001539, 0.00000843, 0.00002198, 0.00001078, -0.00000498, -0.00000751, -0.00000436, -0.00000295, 0.00000376, 0.00001150, 0.00000050, -0.00002314, -0.00002067, 0.00001628, 0.00003656, 0.00000408, -0.00003631, -0.00002254, 0.00002326, 0.00002617, -0.00002060, -0.00004296, -0.00000505, 0.00003483, 0.00002049, -0.00001555, -0.00001263, 0.00001912, 0.00002298, -0.00000895, -0.00002840, -0.00001160, 0.00001050, 0.00000918, -0.00000257, -0.00000276, 0.00000422, 0.00000429, -0.00000078, -0.00000299, -0.00000436, -0.00000820, -0.00000909, -0.00000426, -0.00000130, -0.00000380, -0.00000301, 0.00000545, 0.00000956, -0.00000168, -0.00001937, -0.00002591, -0.00001738, -0.00000393, 0.00000493, 0.00000604, 0.00000065, -0.00000359, 0.00000415, 0.00002311, 0.00003605, 0.00002934, 0.00001185, 0.00000330, 0.00000925, 0.00001834, 0.00002026, 0.00001497, 0.00000684, 0.00000091, 0.00000305, 0.00001370, 0.00002212, 0.00001752, 0.00000617, 0.00000711, 0.00002789, 0.00005230, 0.00005611, 0.00003093, -0.00000867, -0.00003819, -0.00004325, -0.00003009, -0.00001633, -0.00001135, },
		{ -0.00001630, -0.00001288, -0.00001372, -0.00000883, -0.00000384, -0.00000635, -0.00000941, -0.00000439, 0.00000022, -0.00000825, -0.00002186, -0.00002098, -0.00000534, 0.00000539, 0.00000220, -0.00000350, -0.00000375, -0.00000550, -0.00001161, -0.00001202, -0.00000341, 0.00000192, -0.00000221, -0.00000439, 0.00000188, 0.00000362, -0.00000784, -0.00001719, -0.00000845, 0.00000779, 0.00001133, 0.00000345, -0.00000112, -0.00000113, -0.00000437, -0.00000642, 0.00000075, 0.00000723, -0.00000088, -0.00001293, -0.00000811, 0.00000638, 0.00000411, -0.00001454, -0.00001994, -0.00000159, 0.00001604, 0.00001481, 0.00000599, 0.00000057, -0.00000785, -0.00001582, -0.00000409, 0.00002045, 0.00001806, -0.00001899, -0.00003743, -0.00000129, 0.00004069, 0.00002344, -0.00002844, -0.00003497, 0.00001172, 0.00003442, -0.00000404, -0.00004229, -0.00002157, 0.00002288, 0.00002526, -0.00000649, -0.00001408, 0.00001217, 0.00002490, 0.00000230, -0.00002153, -0.00001668, 0.00000123, 0.00000622, 0.00000080, -0.00000022, 0.00000272, 0.00000217, 0.00000034, 0.00000080, -0.00000138, -0.00000824, -0.00001105, -0.00000537, -0.00000086, -0.00000463, -0.00000770, -0.00000036, 0.00000958, 0.00000745, -0.00000715, -0.00002135, -0.00002548, -0.00001891, -0.00000595, 0.00000526, 0.00000609, -0.00000107, -0.00000111, 0.00001461, 0.00003351, 0.00003659, 0.00002300, 0.00000921, 0.00000719, 0.00001397, 0.00002048, 0.00002076, 0.00001343, 0.00000329, 0.00000055, 0.00001010, 0.00002219, 0.00002244, 0.00001188, 0.00000826, 0.00002443, 0.00005054, 0.00006126, 0.00003990, -0.00000414, -0.00004147, -0.00004928, -0.00003282, },
		{ -0.00005490, -0.00003507, -0.00001549, -0.00001402, -0.00001804, -0.00001187, -0.00000253, -0.00000345, -0.00000906, -0.00000626, -0.00000061, -0.00000712, -0.00002083, -0.00002150, -0.00000684, 0.00000380, 0.00000110, -0.00000307, -0.00000135, -0.00000307, -0.00001175, -0.00001421, -0.00000449, 0.00000294, -0.00000177, -0.00000633, -0.00000014, 0.00000456, -0.00000494, -0.00001620, -0.00001115, 0.00000352, 0.00000912, 0.00000543, 0.00000401, 0.00000278, -0.00000597, -0.00001229, -0.00000244, 0.00001127, 0.00000586, -0.00001179, -0.00001357, 0.00000204, 0.00000698, -0.00000764, -0.00001730, -0.00000742, 0.00000563, 0.00000903, 0.00001201, 0.00001490, 0.00000008, -0.00002526, -0.00002144, 0.00001771, 0.00003622, -0.00000205, -0.00004469, -0.00002405, 0.00003224, 0.00003890, -0.00001244, -0.00003990, -0.00000330, 0.00003414, 0.00001164, -0.00003337, -0.00003158, 0.00000836, 0.00002363, 0.00000099, -0.00001253, 0.00000583, 0.00002281, 0.00001053, -0.00001185, -0.00001604, -0.00000584, 0.00000048, 0.00000087, 0.00000235, 0.00000323, 0.00000039, -0.00000066, 0.00000281, 0.00000251, -0.00000582, -0.00001195, -0.00000766, -0.00000136, -0.00000367, -0.00000920, -0.00000607, 0.00000430, 0.00000957, 0.00000337, -0.00001002, -0.00002302, -0.00002819, -0.00002045, -0.00000457, 0.00000524, 0.00000238, -0.00000138, 0.00000840, 0.00002750, 0.00003815, 0.00003216, 0.00001861, 0.00001022, 0.00001121, 0.00001789, 0.00002301, 0.00001982, 0.00000909, 0.00000162, 0.00000682, 0.00001932, 0.00002440, 0.00001754, 0.00001202, 0.00002346, 0.00004899, 0.00006487, 0.00004761, 0.00000037, -0.00004422, },
		{ 0.00000447, -0.00004753, -0.00006127, -0.00003737, -0.00001344, -0.00001361, -0.00002218, -0.00001646, -0.00000291, -0.00000078, -0.00000752, -0.00000702, -0.00000161, -0.00000715, -0.00002055, -0.00002137, -0.00000677, 0.00000300, -0.00000110, -0.00000471, 0.00000014, 0.00000052, -0.00001013, -0.00001602, -0.00000662, 0.00000333, -0.00000041, -0.00000698, -0.00000243, 0.00000357, -0.00000343, -0.00001401, -0.00001083, 0.00000082, 0.00000462, 0.00000294, 0.00000710, 0.00000969, -0.00000246, -0.00001660, -0.00000951, 0.00001034, 0.00001230, -0.00000588, -0.00001478, -0.00000384, 0.00000391, -0.00000408, -0.00001031, -0.00000465, -0.00000157, -0.00000426, 0.00000658, 0.00002690, 0.00001870, -0.00002289, -0.00004043, 0.00000121, 0.00004541, 0.00002207, -0.00003769, -0.00004322, 0.00001342, 0.00004445, 0.00000612, -0.00003569, -0.00001631, 0.00002672, 0.00002216, -0.00002014, -0.00003379, -0.00000459, 0.00001726, 0.00000516, -0.00000937, 0.00000126, 0.00001806, 0.00001441, -0.00000230, -0.00001068, -0.00000860, -0.00000553, -0.00000225, 0.00000321, 0.00000495, 0.00000013, -0.00000248, 0.00000241, 0.00000538, -0.00000169, -0.00001042, -0.00000959, -0.00000344, -0.00000321, -0.00000828, -0.00000881, -0.00000196, 0.00000561, 0.00000747, 0.00000116, -0.00001279, -0.00002742, -0.00003059, -0.00001867, -0.00000388, 0.00000081, -0.00000062, 0.00000561, 0.00002189, 0.00003593, 0.00003726, 0.00002796, 0.00001716, 0.00001225, 0.00001545, 0.00002229, 0.00002379, 0.00001585, 0.00000618, 0.00000612, 0.00001582, 0.00002355, 0.00002127, 0.00001644, 0.00002461, 0.00004888, 0.00006857, 0.00005486, },
		{ 0.00007324, 0.00006292, 0.00000908, -0.00005142, -0.00006941, -0.00004116, -0.00001095, -0.00001109, -0.00002462, -0.00002141, -0.00000524, 0.00000034, -0.00000579, -0.00000629, -0.00000149, -0.00000763, -0.00002182, -0.00002201, -0.00000552, 0.00000426, -0.00000277, -0.00000825, -0.00000076, 0.00000376, -0.00000665, -0.00001611, -0.00000896, 0.00000234, 0.00000058, -0.00000628, -0.00000335, 0.00000192, -0.00000428, -0.00001313, -0.00000860, 0.00000176, 0.00000130, -0.00000305, 0.00000511, 0.00001549, 0.00000553, -0.00001555, -0.00001634, 0.00000419, 0.00001393, 0.00000115, -0.00001003, -0.00000576, -0.00000239, -0.00000790, -0.00000583, 0.00000532, 0.00000133, -0.00001701, -0.00001071, 0.00002730, 0.00004032, -0.00000558, -0.00005067, -0.00002380, 0.00003955, 0.00004362, -0.00001811, -0.00005123, -0.00000894, 0.00003847, 0.00002073, -0.00002478, -0.00002350, 0.00001595, 0.00002600, -0.00000695, -0.00002950, -0.00001316, 0.00000883, 0.00000563, -0.00000609, -0.00000110, 0.00001243, 0.00001394, 0.00000446, -0.00000301, -0.00000647, -0.00000914, -0.00000704, 0.00000136, 0.00000627, 0.00000155, -0.00000347, 0.00000035, 0.00000586, 0.00000222, -0.00000659, -0.00000935, -0.00000584, -0.00000449, -0.00000729, -0.00000866, -0.00000573, -0.00000041, 0.00000504, 0.00000677, -0.00000129, -0.00001823, -0.00003135, -0.00002949, -0.00001744, -0.00000809, -0.00000404, 0.00000323, 0.00001783, 0.00003271, 0.00003907, 0.00003517, 0.00002552, 0.00001696, 0.00001537, 0.00002065, 0.00002510, 0.00002147, 0.00001260, 0.00000853, 0.00001373, 0.00002148, 0.00002286, 0.00001997, 0.00002647, 0.00005000, },
		{ 0.00002799, 0.00005133, 0.00007852, 0.00007229, 0.00001537, -0.00005471, -0.00007905, -0.00004757, -0.00000982, -0.00000741, -0.00002456, -0.00002485, -0.00000831, -0.00000071, -0.00000559, -0.00000505, 0.00000057, -0.00000697, -0.00002416, -0.00002477, -0.00000476, 0.00000759, -0.00000206, -0.00001197, -0.00000439, 0.00000466, -0.00000275, -0.00001382, -0.00000979, 0.00000057, -0.00000008, -0.00000567, -0.00000242, 0.00000176, -0.00000634, -0.00001523, -0.00000742, 0.00000568, 0.00000215, -0.00000880, -0.00000184, 0.00001604, 0.00001383, -0.00000865, -0.00001801, -0.00000321, 0.00000905, 0.00000360, -0.00000240, -0.00000010, -0.00000540, -0.00001779, -0.00001037, 0.00001530, 0.00001567, -0.00001964, -0.00003289, 0.00001173, 0.00005380, 0.00002187, -0.00004490, -0.00004699, 0.00001961, 0.00005370, 0.00000659, -0.00004557, -0.00002674, 0.00002427, 0.00002720, -0.00001199, -0.00002414, 0.00000585, 0.00002426, 0.00000316, -0.00002146, -0.00001646, 0.00000096, 0.00000314, -0.00000394, -0.00000154, 0.00000766, 0.00001042, 0.00000696, 0.00000390, -0.00000075, -0.00000865, -0.00001103, -0.00000275, 0.00000563, 0.00000353, -0.00000270, -0.00000163, 0.00000408, 0.00000402, -0.00000216, -0.00000643, -0.00000653, -0.00000670, -0.00000795, -0.00000781, -0.00000639, -0.00000464, -0.00000042, 0.00000570, 0.00000562, -0.00000654, -0.00002335, -0.00003162, -0.00002845, -0.00002118, -0.00001389, -0.00000314, 0.00001290, 0.00002923, 0.00003926, 0.00004028, 0.00003363, 0.00002399, 0.00001820, 0.00001989, 0.00002482, 0.00002521, 0.00001924, 0.00001325, 0.00001387, 0.00001973, 0.00002324, 0.00002232, },
		{ 0.00002332, 0.00002398, 0.00002909, 0.00005232, 0.00008357, 0.00008230, 0.00002353, -0.00005613, -0.00008867, -0.00005619, -0.00001146, -0.00000477, -0.00002301, -0.00002558, -0.00000982, -0.00000277, -0.00000803, -0.00000537, 0.00000383, -0.00000373, -0.00002570, -0.00002960, -0.00000644, 0.00001125, 0.00000162, -0.00001344, -0.00000913, 0.00000223, -0.00000089, -0.00001033, -0.00000781, 0.00000002, -0.00000222, -0.00000703, -0.00000105, 0.00000425, -0.00000697, -0.00001955, -0.00000986, 0.00000934, 0.00000733, -0.00001026, -0.00000994, 0.00001035, 0.00001716, 0.00000011, -0.00001252, -0.00000572, 0.00000108, -0.00000216, 0.00000130, 0.00001143, 0.00000101, -0.00002683, -0.00002532, 0.00001613, 0.00003541, -0.00000683, -0.00004865, -0.00001599, 0.00005053, 0.00004864, -0.00002326, -0.00005819, -0.00000671, 0.00004869, 0.00002703, -0.00002973, -0.00003462, 0.00000814, 0.00002537, -0.00000188, -0.00002029, -0.00000120, 0.00001937, 0.00000899, -0.00001254, -0.00001509, -0.00000446, -0.00000096, -0.00000361, -0.00000093, 0.00000485, 0.00000595, 0.00000541, 0.00000756, 0.00000583, -0.00000418, -0.00001196, -0.00000727, 0.00000255, 0.00000436, -0.00000061, -0.00000201, 0.00000158, 0.00000316, 0.00000066, -0.00000215, -0.00000442, -0.00000775, -0.00000995, -0.00000836, -0.00000583, -0.00000576, -0.00000483, 0.00000083, 0.00000594, 0.00000154, -0.00001188, -0.00002490, -0.00003133, -0.00003195, -0.00002738, -0.00001541, 0.00000338, 0.00002318, 0.00003772, 0.00004389, 0.00004130, 0.00003250, 0.00002366, 0.00002083, 0.00002402, 0.00002725, 0.00002524, 0.00001936, 0.00001614, 0.00001906, },
		{ 0.00002042, 0.00001981, 0.00002351, 0.00002524, 0.00003011, 0.00005331, 0.00008824, 0.00009206, 0.00003249, -0.00005569, -0.00009675, -0.00006511, -0.00001543, -0.00000500, -0.00002259, -0.00002453, -0.00000799, -0.00000314, -0.00001223, -0.00000905, 0.00000594, 0.00000186, -0.00002432, -0.00003459, -0.00001123, 0.00001265, 0.00000672, -0.00001128, -0.00001207, -0.00000228, -0.00000286, -0.00000846, -0.00000369, 0.00000276, -0.00000372, -0.00001090, -0.00000179, 0.00000821, -0.00000413, -0.00002307, -0.00001584, 0.00000917, 0.00001352, -0.00000619, -0.00001405, 0.00000197, 0.00001301, 0.00000425, -0.00000324, 0.00000069, -0.00000271, -0.00001420, -0.00000520, 0.00002156, 0.00001789, -0.00002619, -0.00004491, 0.00000244, 0.00004976, 0.00001870, -0.00004844, -0.00004514, 0.00002938, 0.00006323, 0.00000641, -0.00005266, -0.00002890, 0.00003195, 0.00003655, -0.00001107, -0.00003217, -0.00000405, 0.00001849, 0.00000338, -0.00001504, -0.00000484, 0.00001364, 0.00001085, -0.00000490, -0.00001060, -0.00000649, -0.00000503, -0.00000519, -0.00000051, 0.00000408, 0.00000259, 0.00000157, 0.00000705, 0.00001035, 0.00000231, -0.00000896, -0.00000980, -0.00000186, 0.00000290, 0.00000123, -0.00000055, 0.00000023, 0.00000080, 0.00000069, 0.00000117, -0.00000034, -0.00000607, -0.00001135, -0.00001052, -0.00000635, -0.00000524, -0.00000630, -0.00000374, 0.00000188, 0.00000333, -0.00000307, -0.00001403, -0.00002541, -0.00003485, -0.00003830, -0.00003072, -0.00001209, 0.00001122, 0.00003174, 0.00004485, 0.00004832, 0.00004253, 0.00003201, 0.00002407, 0.00002336, 0.00002772, 0.00003011, 0.00002626, },
		{ 0.00003299, 0.00003318, 0.00002678, 0.00002264, 0.00002420, 0.00002589, 0.00003080, 0.00005474, 0.00009342, 0.00010172, 0.00004107, -0.00005497, -0.00010341, -0.00007233, -0.00001930, -0.00000777, -0.00002580, -0.00002477, -0.00000344, 0.00000073, -0.00001519, -0.00001577, 0.00000436, 0.00000740, -0.00001950, -0.00003696, -0.00001739, 0.00001030, 0.00001015, -0.00000661, -0.00001091, -0.00000560, -0.00000808, -0.00001096, -0.00000040, 0.00000896, -0.00000169, -0.00001535, -0.00000612, 0.00001060, 0.00000163, -0.00002272, -0.00002218, 0.00000424, 0.00001608, 0.00000029, -0.00001148, -0.00000277, 0.00000426, -0.00000084, 0.00000220, 0.00001422, 0.00000448, -0.00002473, -0.00002282, 0.00002117, 0.00003911, -0.00001070, -0.00005860, -0.00002342, 0.00004905, 0.00004687, -0.00002986, -0.00006335, -0.00000225, 0.00005940, 0.00003246, -0.00003347, -0.00003897, 0.00001166, 0.00003418, 0.00000320, -0.00002339, -0.00000954, 0.00001097, 0.00000437, -0.00001068, -0.00000605, 0.00000847, 0.00000975, 0.00000019, -0.00000485, -0.00000512, -0.00000752, -0.00000796, -0.00000134, 0.00000433, 0.00000137, -0.00000218, 0.00000340, 0.00001100, 0.00000789, -0.00000313, -0.00000877, -0.00000541, -0.00000054, 0.00000122, 0.00000152, 0.00000095, -0.00000101, -0.00000139, 0.00000182, 0.00000345, -0.00000197, -0.00001022, -0.00001257, -0.00000863, -0.00000544, -0.00000581, -0.00000566, -0.00000259, 0.00000037, -0.00000016, -0.00000507, -0.00001491, -0.00002888, -0.00004111, -0.00004286, -0.00003027, -0.00000746, 0.00001800, 0.00003957, 0.00005239, 0.00005349, 0.00004406, 0.00003106, 0.00002392, 0.00002665, },
		{ 0.00002830, 0.00002513, 0.00003290, 0.00003837, 0.00003454, 0.00002849, 0.00002671, 0.00002609, 0.00003019, 0.00005601, 0.00010000, 0.00011288, 0.00004959, -0.00005573, -0.00011085, -0.00007788, -0.00002033, -0.00001017, -0.00003274, -0.00002972, 0.00000046, 0.00000893, -0.00001348, -0.00002249, -0.00000147, 0.00000960, -0.00001339, -0.00003524, -0.00002163, 0.00000541, 0.00000940, -0.00000289, -0.00000599, -0.00000435, -0.00001310, -0.00001821, -0.00000168, 0.00001583, 0.00000516, -0.00001676, -0.00001268, 0.00000874, 0.00000691, -0.00001810, -0.00002466, -0.00000235, 0.00001268, 0.00000336, -0.00000469, 0.00000091, -0.00000162, -0.00001389, -0.00000365, 0.00002675, 0.00002388, -0.00002398, -0.00004502, 0.00000502, 0.00005378, 0.00001703, -0.00005684, -0.00005168, 0.00003039, 0.00006526, 0.00000064, -0.00006296, -0.00003209, 0.00003933, 0.00004531, -0.00000958, -0.00003529, -0.00000367, 0.00002439, 0.00000950, -0.00001388, -0.00000931, 0.00000589, 0.00000332, -0.00000791, -0.00000600, 0.00000445, 0.00000684, 0.00000224, 0.00000031, -0.00000123, -0.00000734, -0.00001052, -0.00000369, 0.00000413, 0.00000192, -0.00000402, -0.00000109, 0.00000798, 0.00001017, 0.00000293, -0.00000443, -0.00000609, -0.00000420, -0.00000106, 0.00000241, 0.00000299, -0.00000083, -0.00000359, -0.00000010, 0.00000485, 0.00000251, -0.00000632, -0.00001239, -0.00001127, -0.00000747, -0.00000570, -0.00000545, -0.00000478, -0.00000357, -0.00000215, -0.00000165, -0.00000587, -0.00001805, -0.00003470, -0.00004618, -0.00004455, -0.00002895, -0.00000396, 0.00002418, 0.00004865, 0.00006203, 0.00005935, 0.00004383, },
		{ 0.00007380, 0.00006270, 0.00004007, 0.00002663, 0.00003040, 0.00003959, 0.00004134, 0.00003729, 0.00003304, 0.00002752, 0.00002780, 0.00005528, 0.00010726, 0.00012718, 0.00006034, -0.00005788, -0.00012176, -0.00008448, -0.00001799, -0.00000842, -0.00004023, -0.00004033, -0.00000071, 0.00001809, -0.00000597, -0.00002494, -0.00000880, 0.00000686, -0.00000987, -0.00003102, -0.00002137, 0.00000166, 0.00000484, -0.00000373, -0.00000087, 0.00000222, -0.00001351, -0.00002706, -0.00000925, 0.00001877, 0.00001468, -0.00001248, -0.00001737, 0.00000307, 0.00000776, -0.00001269, -0.00002152, -0.00000508, 0.00000588, -0.00000140, -0.00000069, 0.00001215, 0.00000302, -0.00002701, -0.00002270, 0.00002778, 0.00004871, -0.00000619, -0.00006014, -0.00002401, 0.00005204, 0.00004690, -0.00003679, -0.00006997, -0.00000023, 0.00006564, 0.00003094, -0.00004491, -0.00004916, 0.00001221, 0.00004206, 0.00000899, -0.00002265, -0.00000947, 0.00001392, 0.00000889, -0.00000774, -0.00000661, 0.00000386, 0.00000227, -0.00000620, -0.00000531, 0.00000179, 0.00000326, 0.00000145, 0.00000338, 0.00000356, -0.00000434, -0.00001133, -0.00000674, 0.00000246, 0.00000278, -0.00000364, -0.00000414, 0.00000323, 0.00000857, 0.00000655, 0.00000118, -0.00000337, -0.00000588, -0.00000442, 0.00000107, 0.00000460, 0.00000123, -0.00000405, -0.00000282, 0.00000337, 0.00000499, -0.00000140, -0.00000924, -0.00001204, -0.00001025, -0.00000736, -0.00000533, -0.00000482, -0.00000559, -0.00000550, -0.00000320, -0.00000219, -0.00000845, -0.00002305, -0.00003951, -0.00004890, -0.00004581, -0.00002942, -0.00000175, 0.00003207, 0.00006171, },
		{ 0.00000172, 0.00004576, 0.00007927, 0.00008337, 0.00006097, 0.00003644, 0.00002943, 0.00003635, 0.00004358, 0.00004636, 0.00004407, 0.00003329, 0.00002535, 0.00005106, 0.00011251, 0.00014424, 0.00007609, -0.00005856, -0.00013678, -0.00009603, -0.00001523, -0.00000094, -0.00004340, -0.00005357, -0.00000915, 0.00002289, 0.00000437, -0.00002076, -0.00001268, 0.00000129, -0.00001177, -0.00002861, -0.00001730, 0.00000282, 0.00000024, -0.00001000, -0.00000029, 0.00001095, -0.00000703, -0.00003216, -0.00002071, 0.00001449, 0.00002167, -0.00000360, -0.00001599, -0.00000211, 0.00000312, -0.00001194, -0.00001581, -0.00000006, 0.00000252, -0.00001296, -0.00000683, 0.00002349, 0.00002056, -0.00002943, -0.00004850, 0.00001067, 0.00006658, 0.00002583, -0.00005760, -0.00005518, 0.00003104, 0.00006613, -0.00000459, -0.00006994, -0.00003086, 0.00004822, 0.00004948, -0.00001790, -0.00004908, -0.00001072, 0.00002729, 0.00001620, -0.00000906, -0.00000675, 0.00000794, 0.00000549, -0.00000577, -0.00000450, 0.00000353, 0.00000200, -0.00000480, -0.00000411, 0.00000065, 0.00000012, -0.00000132, 0.00000353, 0.00000732, 0.00000051, -0.00000940, -0.00000889, -0.00000053, 0.00000243, -0.00000222, -0.00000477, -0.00000078, 0.00000447, 0.00000642, 0.00000533, 0.00000138, -0.00000443, -0.00000676, -0.00000196, 0.00000427, 0.00000358, -0.00000230, -0.00000421, 0.00000043, 0.00000446, 0.00000204, -0.00000451, -0.00000982, -0.00001154, -0.00001005, -0.00000689, -0.00000477, -0.00000550, -0.00000717, -0.00000632, -0.00000359, -0.00000424, -0.00001252, -0.00002703, -0.00004201, -0.00005126, -0.00004970, -0.00003271, },
		{ -0.00005673, -0.00005845, -0.00003636, 0.00001198, 0.00006669, 0.00009613, 0.00008714, 0.00005827, 0.00003692, 0.00003248, 0.00003921, 0.00005083, 0.00005742, 0.00004571, 0.00002693, 0.00004457, 0.00011297, 0.00016083, 0.00009736, -0.00005376, -0.00015297, -0.00011437, -0.00001703, 0.00000963, -0.00003926, -0.00006365, -0.00002239, 0.00001968, 0.00001149, -0.00001212, -0.00000933, -0.00000170, -0.00001803, -0.00003221, -0.00001382, 0.00000957, 0.00000058, -0.00001834, -0.00000675, 0.00001614, 0.00000407, -0.00002951, -0.00003012, 0.00000410, 0.00002107, 0.00000443, -0.00000800, -0.00000085, -0.00000298, -0.00001919, -0.00001455, 0.00001112, 0.00000957, -0.00002389, -0.00002533, 0.00002439, 0.00004504, -0.00001390, -0.00006844, -0.00002235, 0.00006582, 0.00006067, -0.00003415, -0.00007483, -0.00000266, 0.00006607, 0.00002744, -0.00005169, -0.00004975, 0.00002131, 0.00005098, 0.00000644, -0.00003554, -0.00002184, 0.00000988, 0.00001245, -0.00000120, -0.00000056, 0.00000763, 0.00000345, -0.00000601, -0.00000417, 0.00000320, 0.00000203, -0.00000357, -0.00000256, 0.00000103, -0.00000156, -0.00000463, 0.00000099, 0.00000849, 0.00000533, -0.00000499, -0.00000864, -0.00000349, 0.00000039, -0.00000149, -0.00000371, -0.00000249, 0.00000030, 0.00000329, 0.00000620, 0.00000565, -0.00000049, -0.00000647, -0.00000487, 0.00000186, 0.00000446, 0.00000047, -0.00000324, -0.00000171, 0.00000193, 0.00000254, -0.00000070, -0.00000558, -0.00000989, -0.00001157, -0.00000968, -0.00000639, -0.00000524, -0.00000673, -0.00000787, -0.00000674, -0.00000535, -0.00000763, -0.00001560, -0.00002842, -0.00004361, },
		{ -0.00001583, -0.00002782, -0.00004805, -0.00006900, -0.00007015, -0.00003349, 0.00003182, 0.00008919, 0.00010768, 0.00008952, 0.00005906, 0.00003634, 0.00003128, 0.00004665, 0.00006710, 0.00006325, 0.00003657, 0.00004051, 0.00010876, 0.00017274, 0.00012084, -0.00004164, -0.00016520, -0.00013701, -0.00002680, 0.00001743, -0.00002967, -0.00006601, -0.00003403, 0.00000988, 0.00001014, -0.00000552, 0.00000035, 0.00000328, -0.00002310, -0.00004226, -0.00001651, 0.00001793, 0.00000832, -0.00002276, -0.00001787, 0.00001343, 0.00001348, -0.00001994, -0.00003155, -0.00000620, 0.00001212, 0.00000454, 0.00000148, 0.00000998, -0.00000279, -0.00003192, -0.00002440, 0.00002068, 0.00002878, -0.00002385, -0.00005024, 0.00000761, 0.00006405, 0.00001800, -0.00007008, -0.00005983, 0.00004227, 0.00008355, 0.00000334, -0.00007359, -0.00003591, 0.00004695, 0.00004714, -0.00002370, -0.00005131, -0.00000340, 0.00003843, 0.00001957, -0.00001748, -0.00002035, -0.00000202, 0.00000322, -0.00000115, 0.00000356, 0.00001043, 0.00000452, -0.00000576, -0.00000478, 0.00000213, 0.00000161, -0.00000291, -0.00000117, 0.00000251, -0.00000126, -0.00000689, -0.00000293, 0.00000659, 0.00000809, 0.00000034, -0.00000553, -0.00000462, -0.00000239, -0.00000249, -0.00000269, -0.00000208, -0.00000198, -0.00000066, 0.00000387, 0.00000728, 0.00000381, -0.00000351, -0.00000581, -0.00000123, 0.00000314, 0.00000236, -0.00000067, -0.00000158, -0.00000040, 0.00000062, 0.00000036, -0.00000176, -0.00000591, -0.00001020, -0.00001156, -0.00000946, -0.00000673, -0.00000607, -0.00000724, -0.00000829, -0.00000837, -0.00000848, -0.00001035, },
		{ -0.00001139, -0.00001158, -0.00001028, -0.00001273, -0.00002888, -0.00006051, -0.00008717, -0.00007691, -0.00001989, 0.00005543, 0.00010789, 0.00011869, 0.00009538, 0.00005694, 0.00002909, 0.00003531, 0.00006686, 0.00007938, 0.00005427, 0.00004474, 0.00010471, 0.00017856, 0.00014103, -0.00002505, -0.00017006, -0.00015801, -0.00004292, 0.00001742, -0.00002105, -0.00006126, -0.00003773, 0.00000027, 0.00000019, -0.00000801, 0.00000983, 0.00001713, -0.00002002, -0.00005373, -0.00002782, 0.00002116, 0.00002052, -0.00001878, -0.00002700, 0.00000363, 0.00001479, -0.00000991, -0.00002371, -0.00000833, 0.00000057, -0.00000692, 0.00000334, 0.00002687, 0.00001053, -0.00004136, -0.00004635, 0.00001825, 0.00005374, -0.00000587, -0.00006919, -0.00002539, 0.00006500, 0.00005516, -0.00004833, -0.00008621, 0.00000249, 0.00008388, 0.00004071, -0.00005159, -0.00005569, 0.00001777, 0.00004871, 0.00000200, -0.00003841, -0.00001699, 0.00002060, 0.00001976, -0.00000385, -0.00001143, -0.00000491, -0.00000465, -0.00000628, 0.00000277, 0.00001244, 0.00000755, -0.00000361, -0.00000468, 0.00000092, 0.00000058, -0.00000320, -0.00000068, 0.00000410, 0.00000068, -0.00000702, -0.00000637, 0.00000252, 0.00000767, 0.00000440, -0.00000067, -0.00000290, -0.00000410, -0.00000480, -0.00000318, -0.00000112, -0.00000197, -0.00000324, 0.00000010, 0.00000576, 0.00000622, 0.00000051, -0.00000418, -0.00000314, 0.00000050, 0.00000220, 0.00000169, 0.00000057, -0.00000069, -0.00000162, -0.00000126, -0.00000048, -0.00000189, -0.00000620, -0.00001059, -0.00001187, -0.00000991, -0.00000728, -0.00000635, -0.00000750, -0.00000962, },
		{ -0.00000640, -0.00000816, -0.00001219, -0.00001521, -0.00001246, -0.00000572, -0.00000948, -0.00003815, -0.00008142, -0.00010278, -0.00007322, -0.00000133, 0.00007689, 0.00012817, 0.00013465, 0.00009682, 0.00004456, 0.00002632, 0.00005585, 0.00008561, 0.00007366, 0.00005938, 0.00010832, 0.00018262, 0.00015463, -0.00001061, -0.00016955, -0.00017219, -0.00005886, 0.00000988, -0.00002026, -0.00005608, -0.00003235, -0.00000093, -0.00001173, -0.00002186, 0.00001029, 0.00003398, -0.00000555, -0.00005832, -0.00004391, 0.00001452, 0.00002973, -0.00000761, -0.00002745, -0.00000643, 0.00000654, -0.00000817, -0.00001256, 0.00000217, -0.00000329, -0.00002606, -0.00001020, 0.00003957, 0.00003730, -0.00003658, -0.00007300, -0.00000272, 0.00007183, 0.00002826, -0.00006965, -0.00006359, 0.00004244, 0.00008184, -0.00000911, -0.00008972, -0.00003880, 0.00006100, 0.00006356, -0.00001851, -0.00005572, -0.00000857, 0.00003525, 0.00001611, -0.00001999, -0.00001736, 0.00000674, 0.00001174, 0.00000087, -0.00000245, -0.00000081, -0.00000691, -0.00001168, -0.00000199, 0.00001108, 0.00000988, -0.00000020, -0.00000298, 0.00000069, -0.00000032, -0.00000422, -0.00000148, 0.00000471, 0.00000313, -0.00000502, -0.00000776, -0.00000181, 0.00000442, 0.00000550, 0.00000383, 0.00000119, -0.00000324, -0.00000675, -0.00000529, -0.00000130, -0.00000092, -0.00000358, -0.00000287, 0.00000234, 0.00000573, 0.00000342, -0.00000099, -0.00000280, -0.00000175, 0.00000029, 0.00000230, 0.00000311, 0.00000129, -0.00000211, -0.00000369, -0.00000208, -0.00000021, -0.00000175, -0.00000669, -0.00001144, -0.00001270, -0.00001050, -0.00000753, },
		{ -0.00001359, -0.00001116, -0.00000772, -0.00000622, -0.00000955, -0.00001628, -0.00001785, -0.00000826, 0.00000094, -0.00001365, -0.00005744, -0.00010207, -0.00010926, -0.00006497, 0.00001577, 0.00010263, 0.00015669, 0.00014614, 0.00008409, 0.00003428, 0.00004282, 0.00007809, 0.00008423, 0.00007892, 0.00012428, 0.00019388, 0.00016491, -0.00000383, -0.00017117, -0.00018017, -0.00006738, 0.00000168, -0.00002895, -0.00005934, -0.00002426, 0.00000987, -0.00001558, -0.00004142, -0.00000327, 0.00004363, 0.00001596, -0.00005012, -0.00005566, -0.00000005, 0.00002867, 0.00000302, -0.00001795, -0.00000753, -0.00000502, -0.00001959, -0.00000913, 0.00002145, 0.00000947, -0.00004141, -0.00003948, 0.00003579, 0.00006869, -0.00001122, -0.00009083, -0.00003962, 0.00007055, 0.00006723, -0.00004620, -0.00009116, 0.00000199, 0.00008570, 0.00003288, -0.00006868, -0.00006598, 0.00002477, 0.00006448, 0.00001148, -0.00003935, -0.00002210, 0.00001633, 0.00001650, -0.00000561, -0.00000918, 0.00000178, 0.00000297, -0.00000203, 0.00000138, 0.00000510, -0.00000365, -0.00001345, -0.00000735, 0.00000658, 0.00000950, 0.00000252, -0.00000027, 0.00000196, -0.00000013, -0.00000503, -0.00000318, 0.00000374, 0.00000468, -0.00000197, -0.00000664, -0.00000447, 0.00000009, 0.00000341, 0.00000588, 0.00000575, 0.00000030, -0.00000658, -0.00000766, -0.00000322, -0.00000059, -0.00000249, -0.00000378, -0.00000086, 0.00000309, 0.00000388, 0.00000181, -0.00000069, -0.00000225, -0.00000190, 0.00000098, 0.00000431, 0.00000418, -0.00000019, -0.00000469, -0.00000494, -0.00000157, 0.00000063, -0.00000190, -0.00000780, -0.00001265, },
		{ -0.00000288, -0.00000907, -0.00001377, -0.00001502, -0.00001216, -0.00000708, -0.00000593, -0.00001288, -0.00002046, -0.00001548, -0.00000040, 0.00000140, -0.00002834, -0.00007786, -0.00011505, -0.00011301, -0.00005936, 0.00003910, 0.00014117, 0.00018287, 0.00014050, 0.00007008, 0.00004456, 0.00006382, 0.00007866, 0.00009133, 0.00014886, 0.00022003, 0.00018188, -0.00000334, -0.00018311, -0.00018991, -0.00006689, 0.00000231, -0.00004065, -0.00007542, -0.00002422, 0.00002756, -0.00000472, -0.00005527, -0.00002707, 0.00003805, 0.00003349, -0.00003122, -0.00005428, -0.00001318, 0.00001650, 0.00000305, -0.00000562, 0.00000531, -0.00000817, -0.00004022, -0.00002314, 0.00003727, 0.00003966, -0.00003915, -0.00007467, 0.00000854, 0.00008956, 0.00003108, -0.00008702, -0.00008021, 0.00004491, 0.00009499, -0.00000457, -0.00009553, -0.00004169, 0.00006441, 0.00006140, -0.00003252, -0.00007026, -0.00000963, 0.00004648, 0.00002711, -0.00001711, -0.00002059, 0.00000220, 0.00000820, -0.00000039, -0.00000008, 0.00000467, -0.00000094, -0.00000748, -0.00000068, 0.00000812, 0.00000178, -0.00001079, -0.00001007, 0.00000130, 0.00000635, 0.00000294, 0.00000187, 0.00000414, 0.00000154, -0.00000464, -0.00000472, 0.00000155, 0.00000445, 0.00000051, -0.00000390, -0.00000451, -0.00000319, -0.00000051, 0.00000468, 0.00000853, 0.00000504, -0.00000363, -0.00000843, -0.00000592, -0.00000196, -0.00000166, -0.00000303, -0.00000234, 0.00000014, 0.00000215, 0.00000271, 0.00000161, -0.00000099, -0.00000292, -0.00000118, 0.00000352, 0.00000617, 0.00000311, -0.00000321, -0.00000674, -0.00000461, -0.00000014, 0.00000106, },
		{ -0.00000295, 0.00000081, 0.00000043, -0.00000377, -0.00000996, -0.00001589, -0.00001743, -0.00001196, -0.00000531, -0.00000754, -0.00001738, -0.00002008, -0.00000851, 0.00000230, -0.00000854, -0.00004432, -0.00009114, -0.00012759, -0.00012260, -0.00004760, 0.00007969, 0.00018145, 0.00019098, 0.00013054, 0.00007685, 0.00006098, 0.00006109, 0.00008545, 0.00016917, 0.00026016, 0.00021658, 0.00000114, -0.00020700, -0.00021247, -0.00006485, 0.00001661, -0.00004373, -0.00009933, -0.00003996, 0.00004006, 0.00001802, -0.00005347, -0.00004906, 0.00001824, 0.00003578, -0.00001279, -0.00003859, -0.00001289, 0.00000192, -0.00001256, -0.00000334, 0.00002746, 0.00000734, -0.00005704, -0.00005570, 0.00003466, 0.00007692, -0.00001124, -0.00009891, -0.00003802, 0.00008585, 0.00007645, -0.00005702, -0.00010776, 0.00000131, 0.00009899, 0.00004041, -0.00007412, -0.00007203, 0.00002704, 0.00006678, 0.00000320, -0.00005379, -0.00002944, 0.00002075, 0.00002537, -0.00000056, -0.00000965, -0.00000168, -0.00000061, -0.00000317, 0.00000411, 0.00001036, 0.00000114, -0.00001047, -0.00000560, 0.00000656, 0.00000548, -0.00000583, -0.00000906, -0.00000208, 0.00000223, 0.00000093, 0.00000208, 0.00000581, 0.00000405, -0.00000267, -0.00000496, -0.00000070, 0.00000263, 0.00000122, -0.00000120, -0.00000246, -0.00000404, -0.00000409, 0.00000110, 0.00000821, 0.00000871, 0.00000109, -0.00000650, -0.00000760, -0.00000451, -0.00000236, -0.00000212, -0.00000209, -0.00000150, -0.00000025, 0.00000159, 0.00000257, 0.00000083, -0.00000234, -0.00000273, 0.00000147, 0.00000627, 0.00000600, 0.00000005, -0.00000613, -0.00000699, },
		{ -0.00000348, -0.00000709, -0.00000543, -0.00000180, 0.00000039, 0.00000012, -0.00000384, -0.00001217, -0.00001981, -0.00001863, -0.00000971, -0.00000508, -0.00001100, -0.00001776, -0.00001464, -0.00000634, -0.00000546, -0.00001904, -0.00005208, -0.00010541, -0.00015038, -0.00012869, -0.00001579, 0.00012677, 0.00020487, 0.00019347, 0.00014187, 0.00009041, 0.00005001, 0.00006132, 0.00017008, 0.00030164, 0.00027208, 0.00002385, -0.00023366, -0.00025355, -0.00007472, 0.00003899, -0.00002951, -0.00011811, -0.00006914, 0.00003607, 0.00004008, -0.00003566, -0.00005582, -0.00000381, 0.00002002, -0.00000959, -0.00001883, 0.00000681, 0.00000004, -0.00003790, -0.00002199, 0.00004463, 0.00004241, -0.00005412, -0.00009590, 0.00000468, 0.00010307, 0.00003822, -0.00009647, -0.00008847, 0.00005283, 0.00010657, -0.00000849, -0.00010960, -0.00004469, 0.00007692, 0.00007198, -0.00003591, -0.00007883, -0.00001069, 0.00005038, 0.00002477, -0.00002769, -0.00003043, 0.00000014, 0.00001208, 0.00000385, 0.00000139, 0.00000322, -0.00000367, -0.00000849, 0.00000225, 0.00001368, 0.00000632, -0.00000893, -0.00000933, 0.00000212, 0.00000552, -0.00000174, -0.00000565, -0.00000235, -0.00000062, -0.00000212, 0.00000019, 0.00000576, 0.00000606, 0.00000022, -0.00000347, -0.00000173, 0.00000036, 0.00000007, -0.00000001, 0.00000010, -0.00000250, -0.00000558, -0.00000285, 0.00000513, 0.00000962, 0.00000549, -0.00000238, -0.00000686, -0.00000661, -0.00000445, -0.00000244, -0.00000136, -0.00000150, -0.00000174, -0.00000039, 0.00000179, 0.00000183, -0.00000097, -0.00000298, -0.00000053, 0.00000482, 0.00000724, 0.00000337, },
		{ 0.00000684, 0.00000535, -0.00000040, -0.00000501, -0.00000573, -0.00000435, -0.00000237, 0.00000044, 0.00000118, -0.00000535, -0.00001699, -0.00002275, -0.00001710, -0.00000859, -0.00000705, -0.00001042, -0.00001261, -0.00001425, -0.00001553, -0.00001308, -0.00001878, -0.00006038, -0.00013404, -0.00017429, -0.00011465, 0.00002605, 0.00015921, 0.00022361, 0.00021975, 0.00016102, 0.00007111, 0.00003625, 0.00014635, 0.00032583, 0.00033761, 0.00007274, -0.00024649, -0.00030665, -0.00010673, 0.00005467, -0.00000041, -0.00011910, -0.00009843, 0.00001430, 0.00004645, -0.00001404, -0.00004267, -0.00001096, -0.00000313, -0.00002995, -0.00001319, 0.00003898, 0.00002350, -0.00005626, -0.00006184, 0.00003960, 0.00008503, -0.00002202, -0.00012433, -0.00004857, 0.00010091, 0.00009270, -0.00006180, -0.00012252, -0.00000087, 0.00010786, 0.00004143, -0.00008409, -0.00007588, 0.00003823, 0.00007984, 0.00000326, -0.00006304, -0.00003450, 0.00002314, 0.00002691, -0.00000562, -0.00001793, -0.00000748, -0.00000241, -0.00000258, 0.00000525, 0.00001056, 0.00000005, -0.00001079, -0.00000251, 0.00001250, 0.00001047, -0.00000429, -0.00000954, -0.00000206, 0.00000266, -0.00000049, -0.00000238, -0.00000033, -0.00000099, -0.00000431, -0.00000269, 0.00000378, 0.00000643, 0.00000273, -0.00000083, -0.00000095, -0.00000095, -0.00000190, -0.00000077, 0.00000153, 0.00000005, -0.00000459, -0.00000520, 0.00000101, 0.00000756, 0.00000763, 0.00000211, -0.00000371, -0.00000668, -0.00000655, -0.00000420, -0.00000157, -0.00000084, -0.00000175, -0.00000177, 0.00000009, 0.00000141, -0.00000002, -0.00000239, -0.00000164, 0.00000295, },
		{ -0.00000204, 0.00000153, 0.00000572, 0.00000576, 0.00000162, -0.00000226, -0.00000371, -0.00000473, -0.00000538, -0.00000246, 0.00000240, 0.00000086, -0.00001012, -0.00002115, -0.00002256, -0.00001655, -0.00001001, -0.00000463, -0.00000328, -0.00001285, -0.00002716, -0.00002485, -0.00000686, -0.00001657, -0.00008542, -0.00016886, -0.00018045, -0.00008971, 0.00005436, 0.00018920, 0.00027170, 0.00025739, 0.00014129, 0.00004002, 0.00011231, 0.00032139, 0.00039202, 0.00014042, -0.00023245, -0.00035406, -0.00015827, 0.00004896, 0.00002857, -0.00010057, -0.00011111, -0.00001241, 0.00003102, -0.00000765, -0.00002009, 0.00000770, -0.00001290, -0.00006722, -0.00003753, 0.00006381, 0.00007246, -0.00004792, -0.00010872, 0.00000268, 0.00011419, 0.00003469, -0.00012264, -0.00010761, 0.00006406, 0.00013019, -0.00000394, -0.00012503, -0.00005628, 0.00007890, 0.00007421, -0.00004189, -0.00008210, -0.00000083, 0.00006500, 0.00002888, -0.00003537, -0.00003842, -0.00000092, 0.00001435, 0.00000348, -0.00000270, -0.00000256, -0.00000944, -0.00001284, 0.00000044, 0.00001379, 0.00000675, -0.00000790, -0.00000571, 0.00000838, 0.00001118, 0.00000029, -0.00000669, -0.00000364, -0.00000066, -0.00000191, -0.00000116, 0.00000198, 0.00000066, -0.00000450, -0.00000492, 0.00000084, 0.00000494, 0.00000372, 0.00000167, 0.00000119, -0.00000050, -0.00000325, -0.00000268, 0.00000107, 0.00000191, -0.00000214, -0.00000512, -0.00000211, 0.00000389, 0.00000685, 0.00000497, 0.00000037, -0.00000436, -0.00000706, -0.00000631, -0.00000318, -0.00000087, -0.00000102, -0.00000188, -0.00000123, 0.00000016, -0.00000010, -0.00000192, },
		{ -0.00000078, -0.00000214, -0.00000249, 0.00000053, 0.00000471, 0.00000535, 0.00000216, -0.00000042, -0.00000082, -0.00000259, -0.00000623, -0.00000601, -0.00000015, 0.00000313, -0.00000322, -0.00001441, -0.00002180, -0.00002350, -0.00001926, -0.00000659, 0.00000611, -0.00000130, -0.00002890, -0.00004190, -0.00001791, 0.00000508, -0.00003113, -0.00012032, -0.00018913, -0.00017769, -0.00008039, 0.00008132, 0.00025650, 0.00033892, 0.00025202, 0.00009836, 0.00010053, 0.00029788, 0.00041751, 0.00020434, -0.00019449, -0.00037778, -0.00021154, 0.00001967, 0.00003815, -0.00007685, -0.00009995, -0.00002292, 0.00000455, -0.00002956, -0.00001093, 0.00004646, 0.00001000, -0.00009918, -0.00009343, 0.00005756, 0.00013015, -0.00000206, -0.00013872, -0.00006024, 0.00011037, 0.00009711, -0.00008399, -0.00014890, 0.00000180, 0.00013395, 0.00005662, -0.00009410, -0.00009291, 0.00003163, 0.00007920, -0.00000059, -0.00006505, -0.00002558, 0.00003838, 0.00003477, -0.00000997, -0.00002664, -0.00001210, -0.00000213, -0.00000086, 0.00000564, 0.00000785, -0.00000648, -0.00001915, -0.00000885, 0.00001015, 0.00001107, -0.00000188, -0.00000486, 0.00000475, 0.00000900, 0.00000243, -0.00000313, -0.00000247, -0.00000229, -0.00000421, -0.00000221, 0.00000285, 0.00000278, -0.00000289, -0.00000544, -0.00000163, 0.00000241, 0.00000291, 0.00000285, 0.00000345, 0.00000143, -0.00000300, -0.00000431, -0.00000079, 0.00000205, 0.00000013, -0.00000325, -0.00000308, 0.00000055, 0.00000405, 0.00000514, 0.00000342, -0.00000070, -0.00000528, -0.00000722, -0.00000538, -0.00000222, -0.00000079, -0.00000124, -0.00000149, -0.00000082, },
		{ -0.00000111, -0.00000089, -0.00000111, -0.00000266, -0.00000343, -0.00000065, 0.00000385, 0.00000489, 0.00000186, -0.00000002, 0.00000126, 0.00000067, -0.00000428, -0.00000741, -0.00000383, 0.00000098, -0.00000039, -0.00000680, -0.00001457, -0.00002384, -0.00002898, -0.00001709, 0.00000738, 0.00001296, -0.00001735, -0.00004959, -0.00003942, -0.00000164, 0.00000188, -0.00005678, -0.00014254, -0.00020379, -0.00019655, -0.00007514, 0.00015352, 0.00035494, 0.00036330, 0.00021215, 0.00014604, 0.00028916, 0.00041721, 0.00023986, -0.00015572, -0.00037511, -0.00024305, -0.00001612, 0.00001972, -0.00007237, -0.00007680, -0.00000192, -0.00000762, -0.00007463, -0.00003768, 0.00008122, 0.00006943, -0.00009782, -0.00016258, 0.00000644, 0.00016830, 0.00007403, -0.00012999, -0.00012759, 0.00006687, 0.00014007, -0.00001832, -0.00015447, -0.00006412, 0.00010141, 0.00009749, -0.00004248, -0.00009888, -0.00001448, 0.00005871, 0.00002428, -0.00003670, -0.00003012, 0.00001425, 0.00002500, 0.00000319, -0.00000979, -0.00000921, -0.00001226, -0.00001187, 0.00000351, 0.00001522, 0.00000258, -0.00001762, -0.00001625, 0.00000182, 0.00000983, 0.00000306, -0.00000081, 0.00000412, 0.00000646, 0.00000199, -0.00000105, -0.00000015, -0.00000165, -0.00000549, -0.00000426, 0.00000184, 0.00000384, -0.00000073, -0.00000434, -0.00000266, 0.00000014, 0.00000099, 0.00000233, 0.00000464, 0.00000373, -0.00000119, -0.00000456, -0.00000276, 0.00000064, 0.00000098, -0.00000106, -0.00000209, -0.00000105, 0.00000098, 0.00000309, 0.00000415, 0.00000246, -0.00000194, -0.00000603, -0.00000672, -0.00000430, -0.00000177, -0.00000100, },
		{ -0.00000352, -0.00000187, -0.00000115, -0.00000046, -0.00000055, -0.00000258, -0.00000444, -0.00000233, 0.00000258, 0.00000439, 0.00000146, -0.00000047, 0.00000186, 0.00000317, -0.00000097, -0.00000571, -0.00000536, -0.00000287, -0.00000280, -0.00000336, -0.00000524, -0.00001610, -0.00003211, -0.00003034, -0.00000227, 0.00001966, 0.00000053, -0.00004106, -0.00005349, -0.00002532, 0.00000123, -0.00001159, -0.00006858, -0.00016106, -0.00024871, -0.00022768, -0.00002024, 0.00027075, 0.00041568, 0.00034493, 0.00026176, 0.00034003, 0.00042456, 0.00024123, -0.00014880, -0.00036849, -0.00024188, -0.00002840, -0.00001267, -0.00010543, -0.00007092, 0.00004545, 0.00002066, -0.00011644, -0.00010633, 0.00007989, 0.00014624, -0.00004543, -0.00021225, -0.00008229, 0.00016073, 0.00015490, -0.00007566, -0.00017127, -0.00000458, 0.00014469, 0.00005118, -0.00012135, -0.00010970, 0.00004582, 0.00010534, 0.00000865, -0.00007645, -0.00004229, 0.00002611, 0.00002680, -0.00001224, -0.00001921, 0.00000252, 0.00001018, 0.00000258, 0.00000170, 0.00000144, -0.00001171, -0.00002068, -0.00000576, 0.00001466, 0.00001089, -0.00000990, -0.00001746, -0.00000619, 0.00000384, 0.00000371, 0.00000322, 0.00000644, 0.00000589, 0.00000071, -0.00000103, 0.00000140, 0.00000022, -0.00000497, -0.00000571, -0.00000020, 0.00000330, 0.00000067, -0.00000259, -0.00000221, -0.00000094, -0.00000086, 0.00000069, 0.00000425, 0.00000520, 0.00000124, -0.00000323, -0.00000363, -0.00000123, 0.00000029, 0.00000014, -0.00000036, -0.00000073, -0.00000082, 0.00000030, 0.00000261, 0.00000370, 0.00000137, -0.00000314, -0.00000618, -0.00000579, },
		{ -0.00000395, -0.00000564, -0.00000482, -0.00000350, -0.00000232, -0.00000060, 0.00000045, -0.00000142, -0.00000455, -0.00000398, 0.00000065, 0.00000327, 0.00000095, -0.00000098, 0.00000162, 0.00000411, 0.00000158, -0.00000235, -0.00000330, -0.00000456, -0.00000779, -0.00000620, 0.00000016, -0.00000441, -0.00002551, -0.00003776, -0.00001762, 0.00001335, 0.00001243, -0.00002189, -0.00004884, -0.00004498, -0.00002438, -0.00000499, -0.00000692, -0.00007478, -0.00021811, -0.00031882, -0.00021085, 0.00009145, 0.00035821, 0.00042984, 0.00041864, 0.00047900, 0.00049456, 0.00023823, -0.00019258, -0.00039958, -0.00022595, 0.00000643, -0.00002316, -0.00017052, -0.00011358, 0.00008909, 0.00009752, -0.00011805, -0.00019664, 0.00002064, 0.00020353, 0.00005187, -0.00021089, -0.00017963, 0.00009852, 0.00020751, 0.00001002, -0.00017078, -0.00007811, 0.00010853, 0.00009913, -0.00006346, -0.00012040, -0.00001030, 0.00008199, 0.00004035, -0.00004009, -0.00004512, -0.00000180, 0.00001259, -0.00000190, -0.00000418, 0.00000429, 0.00000061, -0.00000572, 0.00000313, 0.00001100, -0.00000317, -0.00002168, -0.00001522, 0.00000709, 0.00001322, -0.00000124, -0.00001259, -0.00000959, -0.00000309, -0.00000009, 0.00000420, 0.00000937, 0.00000777, 0.00000063, -0.00000197, 0.00000138, 0.00000178, -0.00000324, -0.00000572, -0.00000196, 0.00000173, 0.00000074, -0.00000133, -0.00000108, -0.00000073, -0.00000173, -0.00000101, 0.00000272, 0.00000523, 0.00000311, -0.00000108, -0.00000304, -0.00000239, -0.00000113, -0.00000005, 0.00000082, 0.00000059, -0.00000091, -0.00000172, -0.00000002, 0.00000276, 0.00000320, 0.00000008, },
		{ 0.00000306, 0.00000222, -0.00000118, -0.00000389, -0.00000464, -0.00000472, -0.00000428, -0.00000203, 0.00000073, 0.00000026, -0.00000326, -0.00000448, -0.00000129, 0.00000119, -0.00000042, -0.00000160, 0.00000148, 0.00000421, 0.00000231, 0.00000018, 0.00000093, -0.00000196, -0.00001080, -0.00001296, -0.00000182, 0.00000414, -0.00001264, -0.00003410, -0.00002870, -0.00000205, 0.00000965, -0.00000610, -0.00002843, -0.00004398, -0.00005152, -0.00003317, 0.00001259, 0.00000992, -0.00012408, -0.00031488, -0.00034877, -0.00013122, 0.00017934, 0.00039743, 0.00054491, 0.00069034, 0.00067597, 0.00029474, -0.00026785, -0.00050588, -0.00024364, 0.00008280, 0.00002833, -0.00023148, -0.00021579, 0.00008527, 0.00020083, -0.00005284, -0.00026689, -0.00009177, 0.00020406, 0.00016220, -0.00014606, -0.00024670, -0.00000243, 0.00020756, 0.00009663, -0.00012452, -0.00012599, 0.00004730, 0.00011047, -0.00000449, -0.00009774, -0.00004648, 0.00004259, 0.00004483, -0.00000825, -0.00002900, -0.00001367, -0.00000567, -0.00000637, 0.00000423, 0.00001295, 0.00000072, -0.00001306, -0.00000316, 0.00001360, 0.00000665, -0.00001507, -0.00001909, -0.00000239, 0.00000889, 0.00000332, -0.00000544, -0.00000744, -0.00000698, -0.00000562, 0.00000134, 0.00001025, 0.00001052, 0.00000250, -0.00000220, 0.00000043, 0.00000216, -0.00000150, -0.00000456, -0.00000260, 0.00000019, -0.00000016, -0.00000115, -0.00000028, 0.00000012, -0.00000147, -0.00000191, 0.00000100, 0.00000412, 0.00000375, 0.00000082, -0.00000151, -0.00000230, -0.00000218, -0.00000106, 0.00000087, 0.00000176, 0.00000016, -0.00000229, -0.00000236, 0.00000051, },
		{ -0.00000362, -0.00000188, 0.00000154, 0.00000271, 0.00000083, -0.00000152, -0.00000293, -0.00000448, -0.00000584, -0.00000440, -0.00000053, 0.00000113, -0.00000124, -0.00000319, -0.00000187, -0.00000105, -0.00000316, -0.00000329, 0.00000144, 0.00000469, 0.00000203, 0.00000041, 0.00000437, 0.00000370, -0.00000879, -0.00001837, -0.00000938, 0.00000449, -0.00000153, -0.00002162, -0.00002771, -0.00001545, -0.00000610, -0.00000620, -0.00000737, -0.00002112, -0.00005743, -0.00007158, -0.00001377, 0.00005213, -0.00001547, -0.00022508, -0.00038211, -0.00031650, -0.00007674, 0.00021313, 0.00055081, 0.00090313, 0.00097579, 0.00048371, -0.00030841, -0.00068959, -0.00035401, 0.00015456, 0.00015852, -0.00023353, -0.00035240, 0.00000071, 0.00028218, 0.00007585, -0.00027300, -0.00022230, 0.00013115, 0.00024066, -0.00003409, -0.00025376, -0.00010675, 0.00015284, 0.00015189, -0.00005144, -0.00013259, -0.00001311, 0.00008740, 0.00003411, -0.00005736, -0.00005389, 0.00000689, 0.00002798, 0.00000625, -0.00000782, -0.00000862, -0.00001604, -0.00001805, 0.00000173, 0.00001949, 0.00000779, -0.00001385, -0.00001078, 0.00000905, 0.00001184, -0.00000556, -0.00001595, -0.00000834, 0.00000136, 0.00000201, -0.00000061, -0.00000225, -0.00000617, -0.00000936, -0.00000357, 0.00000793, 0.00001191, 0.00000534, -0.00000088, -0.00000018, 0.00000164, -0.00000060, -0.00000316, -0.00000219, -0.00000054, -0.00000118, -0.00000180, -0.00000033, 0.00000073, -0.00000063, -0.00000181, -0.00000009, 0.00000269, 0.00000330, 0.00000176, 0.00000003, -0.00000128, -0.00000226, -0.00000199, 0.00000011, 0.00000213, 0.00000141, -0.00000171, },
		{ 0.00000223, -0.00000068, -0.00000389, -0.00000362, -0.00000037, 0.00000193, 0.00000148, 0.00000025, -0.00000061, -0.00000270, -0.00000589, -0.00000643, -0.00000298, 0.00000017, -0.00000004, -0.00000080, -0.00000015, -0.00000179, -0.00000650, -0.00000691, 0.00000031, 0.00000578, 0.00000237, -0.00000087, 0.00000494, 0.00000893, -0.00000254, -0.00001821, -0.00001669, -0.00000261, 0.00000083, -0.00000899, -0.00001499, -0.00001613, -0.00002319, -0.00002387, -0.00000240, 0.00000861, -0.00003487, -0.00009087, -0.00006239, 0.00003754, 0.00005473, -0.00009916, -0.00030466, -0.00039072, -0.00032050, -0.00009486, 0.00037091, 0.00100867, 0.00133559, 0.00084623, -0.00021880, -0.00089639, -0.00059160, 0.00014643, 0.00033858, -0.00013186, -0.00046466, -0.00016324, 0.00028821, 0.00022877, -0.00019313, -0.00031901, 0.00000118, 0.00025114, 0.00008494, -0.00019627, -0.00017598, 0.00006540, 0.00015767, 0.00001825, -0.00010181, -0.00004999, 0.00004702, 0.00004338, -0.00001982, -0.00003817, -0.00001098, 0.00000450, 0.00000185, 0.00000505, 0.00000500, -0.00001373, -0.00002679, -0.00000821, 0.00001830, 0.00001509, -0.00000823, -0.00001412, 0.00000168, 0.00001064, 0.00000142, -0.00000877, -0.00000833, -0.00000454, -0.00000313, -0.00000051, 0.00000204, -0.00000212, -0.00000932, -0.00000760, 0.00000346, 0.00001065, 0.00000735, 0.00000145, 0.00000038, 0.00000117, -0.00000047, -0.00000224, -0.00000139, -0.00000041, -0.00000163, -0.00000261, -0.00000109, 0.00000063, 0.00000004, -0.00000115, -0.00000032, 0.00000169, 0.00000245, 0.00000177, 0.00000088, -0.00000010, -0.00000153, -0.00000225, -0.00000077, 0.00000179, },
		{ -0.00000127, 0.00000125, 0.00000260, 0.00000029, -0.00000363, -0.00000468, -0.00000204, 0.00000066, 0.00000107, 0.00000087, 0.00000118, -0.00000031, -0.00000433, -0.00000686, -0.00000519, -0.00000233, -0.00000103, 0.00000083, 0.00000326, 0.00000036, -0.00000842, -0.00001187, -0.00000327, 0.00000610, 0.00000400, -0.00000154, 0.00000321, 0.00001083, 0.00000399, -0.00001250, -0.00001806, -0.00001090, -0.00000640, -0.00000544, 0.00000059, -0.00000161, -0.00002771, -0.00004789, -0.00001988, 0.00002412, 0.00000294, -0.00007587, -0.00009745, -0.00001402, 0.00005804, -0.00000281, -0.00016245, -0.00032485, -0.00045335, -0.00043502, 0.00000430, 0.00090398, 0.00163690, 0.00135435, 0.00008381, -0.00102071, -0.00093307, -0.00001190, 0.00049304, 0.00007702, -0.00048443, -0.00036121, 0.00018987, 0.00034473, -0.00004365, -0.00033804, -0.00014064, 0.00018575, 0.00016570, -0.00009844, -0.00018758, -0.00001908, 0.00011889, 0.00005852, -0.00005408, -0.00005487, 0.00001114, 0.00002956, 0.00000027, -0.00001431, -0.00000865, -0.00001105, -0.00001286, 0.00000382, 0.00001583, -0.00000252, -0.00002632, -0.00001789, 0.00001008, 0.00001697, -0.00000082, -0.00001176, -0.00000345, 0.00000561, 0.00000319, -0.00000221, -0.00000383, -0.00000586, -0.00000813, -0.00000427, 0.00000275, 0.00000201, -0.00000611, -0.00000868, -0.00000083, 0.00000727, 0.00000728, 0.00000338, 0.00000187, 0.00000151, -0.00000045, -0.00000190, -0.00000083, 0.00000013, -0.00000140, -0.00000301, -0.00000203, -0.00000011, 0.00000010, -0.00000060, -0.00000005, 0.00000137, 0.00000188, 0.00000139, 0.00000101, 0.00000064, -0.00000056, -0.00000186, },
		{ 0.00000018, -0.00000125, -0.00000139, 0.00000082, 0.00000282, 0.00000117, -0.00000314, -0.00000533, -0.00000340, -0.00000060, 0.00000023, 0.00000063, 0.00000194, 0.00000154, -0.00000213, -0.00000552, -0.00000565, -0.00000474, -0.00000409, -0.00000013, 0.00000622, 0.00000509, -0.00000694, -0.00001591, -0.00000915, 0.00000370, 0.00000553, -0.00000025, 0.00000183, 0.00000947, 0.00000678, -0.00000552, -0.00001231, -0.00001303, -0.00001692, -0.00001408, 0.00000742, 0.00001984, -0.00001338, -0.00006123, -0.00005039, 0.00001309, 0.00003104, -0.00003576, -0.00009375, -0.00006097, 0.00000632, 0.00001680, -0.00003322, -0.00015301, -0.00041384, -0.00068416, -0.00046968, 0.00054315, 0.00173995, 0.00189990, 0.00062098, -0.00094202, -0.00128603, -0.00034338, 0.00052627, 0.00033963, -0.00037009, -0.00051818, 0.00000334, 0.00036942, 0.00012389, -0.00026712, -0.00024084, 0.00007054, 0.00017992, -0.00000169, -0.00014591, -0.00006926, 0.00005991, 0.00006022, -0.00001448, -0.00003620, -0.00000555, 0.00000845, 0.00000057, 0.00000274, 0.00000552, -0.00001185, -0.00002569, -0.00000795, 0.00001652, 0.00000895, -0.00001736, -0.00002092, 0.00000029, 0.00001240, 0.00000323, -0.00000655, -0.00000399, 0.00000101, 0.00000084, 0.00000062, 0.00000145, -0.00000287, -0.00000987, -0.00000884, -0.00000012, 0.00000358, -0.00000212, -0.00000680, -0.00000295, 0.00000354, 0.00000524, 0.00000378, 0.00000325, 0.00000258, 0.00000008, -0.00000170, -0.00000065, 0.00000057, -0.00000079, -0.00000283, -0.00000264, -0.00000103, -0.00000039, -0.00000054, 0.00000015, 0.00000143, 0.00000180, 0.00000118, 0.00000081, 0.00000087, },
		{ 0.00000073, 0.00000090, 0.00000066, -0.00000072, -0.00000144, 0.00000043, 0.00000301, 0.00000215, -0.00000235, -0.00000565, -0.00000462, -0.00000192, -0.00000072, 0.00000014, 0.00000204, 0.00000236, -0.00000060, -0.00000351, -0.00000403, -0.00000521, -0.00000743, -0.00000391, 0.00000631, 0.00001007, -0.00000183, -0.00001634, -0.00001499, -0.00000197, 0.00000435, 0.00000184, 0.00000305, 0.00000794, 0.00000509, -0.00000263, -0.00000378, -0.00000601, -0.00002154, -0.00002924, -0.00000071, 0.00003430, 0.00001327, -0.00005280, -0.00007421, -0.00001863, 0.00002968, 0.00000145, -0.00005331, -0.00006555, -0.00005499, -0.00004308, 0.00001212, 0.00003966, -0.00021059, -0.00073944, -0.00091246, -0.00002892, 0.00154200, 0.00232110, 0.00132269, -0.00057949, -0.00151243, -0.00079523, 0.00036751, 0.00055827, -0.00013543, -0.00056330, -0.00021090, 0.00028436, 0.00024547, -0.00013411, -0.00026392, -0.00004671, 0.00013041, 0.00005620, -0.00007891, -0.00007345, 0.00001083, 0.00003411, 0.00000074, -0.00001249, -0.00000205, -0.00000480, -0.00001024, 0.00000593, 0.00001922, -0.00000087, -0.00002823, -0.00002109, 0.00000729, 0.00001322, -0.00000606, -0.00001607, -0.00000513, 0.00000497, 0.00000185, -0.00000286, -0.00000134, -0.00000039, -0.00000235, -0.00000035, 0.00000437, 0.00000173, -0.00000778, -0.00001127, -0.00000441, 0.00000198, 0.00000021, -0.00000368, -0.00000249, 0.00000123, 0.00000255, 0.00000254, 0.00000355, 0.00000366, 0.00000114, -0.00000117, -0.00000058, 0.00000076, -0.00000018, -0.00000229, -0.00000279, -0.00000173, -0.00000106, -0.00000093, -0.00000006, 0.00000141, 0.00000198, 0.00000131, },
		{ 0.00000203, 0.00000158, 0.00000086, 0.00000104, 0.00000115, -0.00000018, -0.00000153, -0.00000016, 0.00000295, 0.00000316, -0.00000109, -0.00000535, -0.00000559, -0.00000348, -0.00000204, -0.00000043, 0.00000217, 0.00000269, -0.00000026, -0.00000236, -0.00000152, -0.00000308, -0.00000872, -0.00000867, 0.00000273, 0.00001222, 0.00000457, -0.00001216, -0.00001736, -0.00000875, -0.00000104, 0.00000153, 0.00000632, 0.00000968, 0.00000223, -0.00000591, 0.00000069, 0.00000624, -0.00001486, -0.00003998, -0.00001964, 0.00003173, 0.00003642, -0.00002602, -0.00007446, -0.00004941, -0.00000115, 0.00000706, -0.00000533, -0.00001783, -0.00007315, -0.00013843, -0.00005375, 0.00015580, 0.00007631, -0.00056633, -0.00117829, -0.00068028, 0.00102907, 0.00246054, 0.00203494, 0.00005995, -0.00148086, -0.00124349, 0.00001310, 0.00063061, 0.00014627, -0.00046698, -0.00037321, 0.00011963, 0.00027583, 0.00000542, -0.00021015, -0.00012035, 0.00004856, 0.00005964, -0.00002240, -0.00004375, -0.00000863, 0.00000276, -0.00000836, -0.00000065, 0.00001087, -0.00000407, -0.00001981, -0.00000174, 0.00002386, 0.00001293, -0.00001992, -0.00002727, -0.00000551, 0.00000841, 0.00000069, -0.00000740, -0.00000379, -0.00000017, -0.00000293, -0.00000327, 0.00000107, 0.00000112, -0.00000361, -0.00000282, 0.00000410, 0.00000506, -0.00000363, -0.00001045, -0.00000750, -0.00000139, -0.00000008, -0.00000141, -0.00000060, 0.00000098, 0.00000069, 0.00000056, 0.00000254, 0.00000395, 0.00000223, -0.00000025, -0.00000029, 0.00000087, 0.00000030, -0.00000167, -0.00000260, -0.00000208, -0.00000159, -0.00000147, -0.00000063, 0.00000103, },
		{ -0.00000130, 0.00000036, 0.00000176, 0.00000165, 0.00000096, 0.00000125, 0.00000183, 0.00000067, -0.00000140, -0.00000094, 0.00000231, 0.00000371, 0.00000041, -0.00000416, -0.00000572, -0.00000500, -0.00000405, -0.00000160, 0.00000245, 0.00000343, -0.00000041, -0.00000278, -0.00000002, 0.00000043, -0.00000681, -0.00001155, -0.00000279, 0.00000996, 0.00000855, -0.00000540, -0.00001422, -0.00001254, -0.00000902, -0.00000390, 0.00000802, 0.00001521, 0.00000299, -0.00001228, -0.00000290, 0.00001550, 0.00000045, -0.00003785, -0.00003706, 0.00001363, 0.00004145, 0.00000211, -0.00004957, -0.00005549, -0.00003955, -0.00002848, 0.00001119, 0.00005632, -0.00001943, -0.00020173, -0.00019935, 0.00013239, 0.00032827, -0.00021831, -0.00117470, -0.00123582, 0.00029445, 0.00223040, 0.00256776, 0.00086614, -0.00112628, -0.00153390, -0.00045727, 0.00049872, 0.00036892, -0.00026097, -0.00042291, -0.00005831, 0.00021152, 0.00009570, -0.00011462, -0.00012947, -0.00002318, 0.00002131, -0.00000098, -0.00000702, 0.00000184, -0.00001330, -0.00002799, -0.00000654, 0.00001921, 0.00000604, -0.00002030, -0.00001148, 0.00001875, 0.00002115, -0.00000668, -0.00002374, -0.00001429, -0.00000150, -0.00000029, -0.00000089, 0.00000204, 0.00000001, -0.00000691, -0.00000696, 0.00000059, 0.00000307, -0.00000247, -0.00000433, 0.00000202, 0.00000586, 0.00000015, -0.00000739, -0.00000806, -0.00000439, -0.00000220, -0.00000108, 0.00000105, 0.00000210, 0.00000036, -0.00000098, 0.00000085, 0.00000324, 0.00000274, 0.00000068, 0.00000024, 0.00000110, 0.00000074, -0.00000105, -0.00000228, -0.00000216, -0.00000185, -0.00000188, },
		{ -0.00000192, -0.00000209, -0.00000186, -0.00000039, 0.00000125, 0.00000144, 0.00000075, 0.00000122, 0.00000251, 0.00000193, -0.00000064, -0.00000142, 0.00000121, 0.00000332, 0.00000138, -0.00000239, -0.00000450, -0.00000560, -0.00000645, -0.00000397, 0.00000213, 0.00000481, 0.00000018, -0.00000399, -0.00000059, 0.00000301, -0.00000281, -0.00001073, -0.00000690, 0.00000463, 0.00000748, -0.00000037, -0.00000698, -0.00001002, -0.00001503, -0.00001387, 0.00000402, 0.00002089, 0.00000951, -0.00001557, -0.00001233, 0.00001525, 0.00001498, -0.00002334, -0.00004081, -0.00000689, 0.00002567, 0.00001152, -0.00001637, -0.00002741, -0.00005399, -0.00008670, -0.00002787, 0.00010858, 0.00009246, -0.00017866, -0.00034946, -0.00002855, 0.00044064, 0.00018298, -0.00090272, -0.00154362, -0.00048803, 0.00165456, 0.00276537, 0.00165456, -0.00048803, -0.00154362, -0.00090272, 0.00018298, 0.00044064, -0.00002855, -0.00034946, -0.00017866, 0.00009246, 0.00010858, -0.00002787, -0.00008670, -0.00005399, -0.00002741, -0.00001637, 0.00001152, 0.00002567, -0.00000689, -0.00004081, -0.00002334, 0.00001498, 0.00001525, -0.00001233, -0.00001557, 0.00000951, 0.00002089, 0.00000402, -0.00001387, -0.00001503, -0.00001002, -0.00000698, -0.00000037, 0.00000748, 0.00000463, -0.00000690, -0.00001073, -0.00000281, 0.00000301, -0.00000059, -0.00000399, 0.00000018, 0.00000481, 0.00000213, -0.00000397, -0.00000645, -0.00000560, -0.00000450, -0.00000239, 0.00000138, 0.00000332, 0.00000121, -0.00000142, -0.00000064, 0.00000193, 0.00000251, 0.00000122, 0.00000075, 0.00000144, 0.00000125, -0.00000039, -0.00000186, -0.00000209, },
		{ -0.00000130, -0.00000188, -0.00000185, -0.00000216, -0.00000228, -0.00000105, 0.00000074, 0.00000110, 0.00000024, 0.00000068, 0.00000274, 0.00000324, 0.00000085, -0.00000098, 0.00000036, 0.00000210, 0.00000105, -0.00000108, -0.00000220, -0.00000439, -0.00000806, -0.00000739, 0.00000015, 0.00000586, 0.00000202, -0.00000433, -0.00000247, 0.00000307, 0.00000059, -0.00000696, -0.00000691, 0.00000001, 0.00000204, -0.00000089, -0.00000029, -0.00000150, -0.00001429, -0.00002374, -0.00000668, 0.00002115, 0.00001875, -0.00001148, -0.00002030, 0.00000604, 0.00001921, -0.00000654, -0.00002799, -0.00001330, 0.00000184, -0.00000702, -0.00000098, 0.00002131, -0.00002318, -0.00012947, -0.00011462, 0.00009570, 0.00021152, -0.00005831, -0.00042291, -0.00026097, 0.00036892, 0.00049872, -0.00045727, -0.00153390, -0.00112628, 0.00086614, 0.00256776, 0.00223040, 0.00029445, -0.00123582, -0.00117470, -0.00021831, 0.00032827, 0.00013239, -0.00019935, -0.00020173, -0.00001943, 0.00005632, 0.00001119, -0.00002848, -0.00003955, -0.00005549, -0.00004957, 0.00000211, 0.00004145, 0.00001363, -0.00003706, -0.00003785, 0.00000045, 0.00001550, -0.00000290, -0.00001228, 0.00000299, 0.00001521, 0.00000802, -0.00000390, -0.00000902, -0.00001254, -0.00001422, -0.00000540, 0.00000855, 0.00000996, -0.00000279, -0.00001155, -0.00000681, 0.00000043, -0.00000002, -0.00000278, -0.00000041, 0.00000343, 0.00000245, -0.00000160, -0.00000405, -0.00000500, -0.00000572, -0.00000416, 0.00000041, 0.00000371, 0.00000231, -0.00000094, -0.00000140, 0.00000067, 0.00000183, 0.00000125, 0.00000096, 0.00000165, 0.00000176, 0.00000036, },
		{ 0.00000203, 0.00000103, -0.00000063, -0.00000147, -0.00000159, -0.00000208, -0.00000260, -0.00000167, 0.00000030, 0.00000087, -0.00000029, -0.00000025, 0.00000223, 0.00000395, 0.00000254, 0.00000056, 0.00000069, 0.00000098, -0.00000060, -0.00000141, -0.00000008, -0.00000139, -0.00000750, -0.00001045, -0.00000363, 0.00000506, 0.00000410, -0.00000282, -0.00000361, 0.00000112, 0.00000107, -0.00000327, -0.00000293, -0.00000017, -0.00000379, -0.00000740, 0.00000069, 0.00000841, -0.00000551, -0.00002727, -0.00001992, 0.00001293, 0.00002386, -0.00000174, -0.00001981, -0.00000407, 0.00001087, -0.00000065, -0.00000836, 0.00000276, -0.00000863, -0.00004375, -0.00002240, 0.00005964, 0.00004856, -0.00012035, -0.00021015, 0.00000542, 0.00027583, 0.00011963, -0.00037321, -0.00046698, 0.00014627, 0.00063061, 0.00001310, -0.00124349, -0.00148086, 0.00005995, 0.00203494, 0.00246054, 0.00102907, -0.00068028, -0.00117829, -0.00056633, 0.00007631, 0.00015580, -0.00005375, -0.00013843, -0.00007315, -0.00001783, -0.00000533, 0.00000706, -0.00000115, -0.00004941, -0.00007446, -0.00002602, 0.00003642, 0.00003173, -0.00001964, -0.00003998, -0.00001486, 0.00000624, 0.00000069, -0.00000591, 0.00000223, 0.00000968, 0.00000632, 0.00000153, -0.00000104, -0.00000875, -0.00001736, -0.00001216, 0.00000457, 0.00001222, 0.00000273, -0.00000867, -0.00000872, -0.00000308, -0.00000152, -0.00000236, -0.00000026, 0.00000269, 0.00000217, -0.00000043, -0.00000204, -0.00000348, -0.00000559, -0.00000535, -0.00000109, 0.00000316, 0.00000295, -0.00000016, -0.00000153, -0.00000018, 0.00000115, 0.00000104, 0.00000086, 0.00000158, },
		{ 0.00000073, 0.00000131, 0.00000198, 0.00000141, -0.00000006, -0.00000093, -0.00000106, -0.00000173, -0.00000279, -0.00000229, -0.00000018, 0.00000076, -0.00000058, -0.00000117, 0.00000114, 0.00000366, 0.00000355, 0.00000254, 0.00000255, 0.00000123, -0.00000249, -0.00000368, 0.00000021, 0.00000198, -0.00000441, -0.00001127, -0.00000778, 0.00000173, 0.00000437, -0.00000035, -0.00000235, -0.00000039, -0.00000134, -0.00000286, 0.00000185, 0.00000497, -0.00000513, -0.00001607, -0.00000606, 0.00001322, 0.00000729, -0.00002109, -0.00002823, -0.00000087, 0.00001922, 0.00000593, -0.00001024, -0.00000480, -0.00000205, -0.00001249, 0.00000074, 0.00003411, 0.00001083, -0.00007345, -0.00007891, 0.00005620, 0.00013041, -0.00004671, -0.00026392, -0.00013411, 0.00024547, 0.00028436, -0.00021090, -0.00056330, -0.00013543, 0.00055827, 0.00036751, -0.00079523, -0.00151243, -0.00057949, 0.00132269, 0.00232110, 0.00154200, -0.00002892, -0.00091246, -0.00073944, -0.00021059, 0.00003966, 0.00001212, -0.00004308, -0.00005499, -0.00006555, -0.00005331, 0.00000145, 0.00002968, -0.00001863, -0.00007421, -0.00005280, 0.00001327, 0.00003430, -0.00000071, -0.00002924, -0.00002154, -0.00000601, -0.00000378, -0.00000263, 0.00000509, 0.00000794, 0.00000305, 0.00000184, 0.00000435, -0.00000197, -0.00001499, -0.00001634, -0.00000183, 0.00001007, 0.00000631, -0.00000391, -0.00000743, -0.00000521, -0.00000403, -0.00000351, -0.00000060, 0.00000236, 0.00000204, 0.00000014, -0.00000072, -0.00000192, -0.00000462, -0.00000565, -0.00000235, 0.00000215, 0.00000301, 0.00000043, -0.00000144, -0.00000072, 0.00000066, 0.00000090, },
		{ 0.00000018, 0.00000087, 0.00000081, 0.00000118, 0.00000180, 0.00000143, 0.00000015, -0.00000054, -0.00000039, -0.00000103, -0.00000264, -0.00000283, -0.00000079, 0.00000057, -0.00000065, -0.00000170, 0.00000008, 0.00000258, 0.00000325, 0.00000378, 0.00000524, 0.00000354, -0.00000295, -0.00000680, -0.00000212, 0.00000358, -0.00000012, -0.00000884, -0.00000987, -0.00000287, 0.00000145, 0.00000062, 0.00000084, 0.00000101, -0.00000399, -0.00000655, 0.00000323, 0.00001240, 0.00000029, -0.00002092, -0.00001736, 0.00000895, 0.00001652, -0.00000795, -0.00002569, -0.00001185, 0.00000552, 0.00000274, 0.00000057, 0.00000845, -0.00000555, -0.00003620, -0.00001448, 0.00006022, 0.00005991, -0.00006926, -0.00014591, -0.00000169, 0.00017992, 0.00007054, -0.00024084, -0.00026712, 0.00012389, 0.00036942, 0.00000334, -0.00051818, -0.00037009, 0.00033963, 0.00052627, -0.00034338, -0.00128603, -0.00094202, 0.00062098, 0.00189990, 0.00173995, 0.00054315, -0.00046968, -0.00068416, -0.00041384, -0.00015301, -0.00003322, 0.00001680, 0.00000632, -0.00006097, -0.00009375, -0.00003576, 0.00003104, 0.00001309, -0.00005039, -0.00006123, -0.00001338, 0.00001984, 0.00000742, -0.00001408, -0.00001692, -0.00001303, -0.00001231, -0.00000552, 0.00000678, 0.00000947, 0.00000183, -0.00000025, 0.00000553, 0.00000370, -0.00000915, -0.00001591, -0.00000694, 0.00000509, 0.00000622, -0.00000013, -0.00000409, -0.00000474, -0.00000565, -0.00000552, -0.00000213, 0.00000154, 0.00000194, 0.00000063, 0.00000023, -0.00000060, -0.00000340, -0.00000533, -0.00000314, 0.00000117, 0.00000282, 0.00000082, -0.00000139, -0.00000125, },
		{ -0.00000127, -0.00000186, -0.00000056, 0.00000064, 0.00000101, 0.00000139, 0.00000188, 0.00000137, -0.00000005, -0.00000060, 0.00000010, -0.00000011, -0.00000203, -0.00000301, -0.00000140, 0.00000013, -0.00000083, -0.00000190, -0.00000045, 0.00000151, 0.00000187, 0.00000338, 0.00000728, 0.00000727, -0.00000083, -0.00000868, -0.00000611, 0.00000201, 0.00000275, -0.00000427, -0.00000813, -0.00000586, -0.00000383, -0.00000221, 0.00000319, 0.00000561, -0.00000345, -0.00001176, -0.00000082, 0.00001697, 0.00001008, -0.00001789, -0.00002632, -0.00000252, 0.00001583, 0.00000382, -0.00001286, -0.00001105, -0.00000865, -0.00001431, 0.00000027, 0.00002956, 0.00001114, -0.00005487, -0.00005408, 0.00005852, 0.00011889, -0.00001908, -0.00018758, -0.00009844, 0.00016570, 0.00018575, -0.00014064, -0.00033804, -0.00004365, 0.00034473, 0.00018987, -0.00036121, -0.00048443, 0.00007702, 0.00049304, -0.00001190, -0.00093307, -0.00102071, 0.00008381, 0.00135435, 0.00163690, 0.00090398, 0.00000430, -0.00043502, -0.00045335, -0.00032485, -0.00016245, -0.00000281, 0.00005804, -0.00001402, -0.00009745, -0.00007587, 0.00000294, 0.00002412, -0.00001988, -0.00004789, -0.00002771, -0.00000161, 0.00000059, -0.00000544, -0.00000640, -0.00001090, -0.00001806, -0.00001250, 0.00000399, 0.00001083, 0.00000321, -0.00000154, 0.00000400, 0.00000610, -0.00000327, -0.00001187, -0.00000842, 0.00000036, 0.00000326, 0.00000083, -0.00000103, -0.00000233, -0.00000519, -0.00000686, -0.00000433, -0.00000031, 0.00000118, 0.00000087, 0.00000107, 0.00000066, -0.00000204, -0.00000468, -0.00000363, 0.00000029, 0.00000260, 0.00000125, },
		{ 0.00000223, 0.00000179, -0.00000077, -0.00000225, -0.00000153, -0.00000010, 0.00000088, 0.00000177, 0.00000245, 0.00000169, -0.00000032, -0.00000115, 0.00000004, 0.00000063, -0.00000109, -0.00000261, -0.00000163, -0.00000041, -0.00000139, -0.00000224, -0.00000047, 0.00000117, 0.00000038, 0.00000145, 0.00000735, 0.00001065, 0.00000346, -0.00000760, -0.00000932, -0.00000212, 0.00000204, -0.00000051, -0.00000313, -0.00000454, -0.00000833, -0.00000877, 0.00000142, 0.00001064, 0.00000168, -0.00001412, -0.00000823, 0.00001509, 0.00001830, -0.00000821, -0.00002679, -0.00001373, 0.00000500, 0.00000505, 0.00000185, 0.00000450, -0.00001098, -0.00003817, -0.00001982, 0.00004338, 0.00004702, -0.00004999, -0.00010181, 0.00001825, 0.00015767, 0.00006540, -0.00017598, -0.00019627, 0.00008494, 0.00025114, 0.00000118, -0.00031901, -0.00019313, 0.00022877, 0.00028821, -0.00016324, -0.00046466, -0.00013186, 0.00033858, 0.00014643, -0.00059160, -0.00089639, -0.00021880, 0.00084623, 0.00133559, 0.00100867, 0.00037091, -0.00009486, -0.00032050, -0.00039072, -0.00030466, -0.00009916, 0.00005473, 0.00003754, -0.00006239, -0.00009087, -0.00003487, 0.00000861, -0.00000240, -0.00002387, -0.00002319, -0.00001613, -0.00001499, -0.00000899, 0.00000083, -0.00000261, -0.00001669, -0.00001821, -0.00000254, 0.00000893, 0.00000494, -0.00000087, 0.00000237, 0.00000578, 0.00000031, -0.00000691, -0.00000650, -0.00000179, -0.00000015, -0.00000080, -0.00000004, 0.00000017, -0.00000298, -0.00000643, -0.00000589, -0.00000270, -0.00000061, 0.00000025, 0.00000148, 0.00000193, -0.00000037, -0.00000362, -0.00000389, -0.00000068, },
		{ -0.00000362, -0.00000171, 0.00000141, 0.00000213, 0.00000011, -0.00000199, -0.00000226, -0.00000128, 0.00000003, 0.00000176, 0.00000330, 0.00000269, -0.00000009, -0.00000181, -0.00000063, 0.00000073, -0.00000033, -0.00000180, -0.00000118, -0.00000054, -0.00000219, -0.00000316, -0.00000060, 0.00000164, -0.00000018, -0.00000088, 0.00000534, 0.00001191, 0.00000793, -0.00000357, -0.00000936, -0.00000617, -0.00000225, -0.00000061, 0.00000201, 0.00000136, -0.00000834, -0.00001595, -0.00000556, 0.00001184, 0.00000905, -0.00001078, -0.00001385, 0.00000779, 0.00001949, 0.00000173, -0.00001805, -0.00001604, -0.00000862, -0.00000782, 0.00000625, 0.00002798, 0.00000689, -0.00005389, -0.00005736, 0.00003411, 0.00008740, -0.00001311, -0.00013259, -0.00005144, 0.00015189, 0.00015284, -0.00010675, -0.00025376, -0.00003409, 0.00024066, 0.00013115, -0.00022230, -0.00027300, 0.00007585, 0.00028218, 0.00000071, -0.00035240, -0.00023353, 0.00015852, 0.00015456, -0.00035401, -0.00068959, -0.00030841, 0.00048371, 0.00097579, 0.00090313, 0.00055081, 0.00021313, -0.00007674, -0.00031650, -0.00038211, -0.00022508, -0.00001547, 0.00005213, -0.00001377, -0.00007158, -0.00005743, -0.00002112, -0.00000737, -0.00000620, -0.00000610, -0.00001545, -0.00002771, -0.00002162, -0.00000153, 0.00000449, -0.00000938, -0.00001837, -0.00000879, 0.00000370, 0.00000437, 0.00000041, 0.00000203, 0.00000469, 0.00000144, -0.00000329, -0.00000316, -0.00000105, -0.00000187, -0.00000319, -0.00000124, 0.00000113, -0.00000053, -0.00000440, -0.00000584, -0.00000448, -0.00000293, -0.00000152, 0.00000083, 0.00000271, 0.00000154, -0.00000188, },
		{ 0.00000306, 0.00000051, -0.00000236, -0.00000229, 0.00000016, 0.00000176, 0.00000087, -0.00000106, -0.00000218, -0.00000230, -0.00000151, 0.00000082, 0.00000375, 0.00000412, 0.00000100, -0.00000191, -0.00000147, 0.00000012, -0.00000028, -0.00000115, -0.00000016, 0.00000019, -0.00000260, -0.00000456, -0.00000150, 0.00000216, 0.00000043, -0.00000220, 0.00000250, 0.00001052, 0.00001025, 0.00000134, -0.00000562, -0.00000698, -0.00000744, -0.00000544, 0.00000332, 0.00000889, -0.00000239, -0.00001909, -0.00001507, 0.00000665, 0.00001360, -0.00000316, -0.00001306, 0.00000072, 0.00001295, 0.00000423, -0.00000637, -0.00000567, -0.00001367, -0.00002900, -0.00000825, 0.00004483, 0.00004259, -0.00004648, -0.00009774, -0.00000449, 0.00011047, 0.00004730, -0.00012599, -0.00012452, 0.00009663, 0.00020756, -0.00000243, -0.00024670, -0.00014606, 0.00016220, 0.00020406, -0.00009177, -0.00026689, -0.00005284, 0.00020083, 0.00008527, -0.00021579, -0.00023148, 0.00002833, 0.00008280, -0.00024364, -0.00050588, -0.00026785, 0.00029474, 0.00067597, 0.00069034, 0.00054491, 0.00039743, 0.00017934, -0.00013122, -0.00034877, -0.00031488, -0.00012408, 0.00000992, 0.00001259, -0.00003317, -0.00005152, -0.00004398, -0.00002843, -0.00000610, 0.00000965, -0.00000205, -0.00002870, -0.00003410, -0.00001264, 0.00000414, -0.00000182, -0.00001296, -0.00001080, -0.00000196, 0.00000093, 0.00000018, 0.00000231, 0.00000421, 0.00000148, -0.00000160, -0.00000042, 0.00000119, -0.00000129, -0.00000448, -0.00000326, 0.00000026, 0.00000073, -0.00000203, -0.00000428, -0.00000472, -0.00000464, -0.00000389, -0.00000118, 0.00000222, },
		{ -0.00000395, 0.00000008, 0.00000320, 0.00000276, -0.00000002, -0.00000172, -0.00000091, 0.00000059, 0.00000082, -0.00000005, -0.00000113, -0.00000239, -0.00000304, -0.00000108, 0.00000311, 0.00000523, 0.00000272, -0.00000101, -0.00000173, -0.00000073, -0.00000108, -0.00000133, 0.00000074, 0.00000173, -0.00000196, -0.00000572, -0.00000324, 0.00000178, 0.00000138, -0.00000197, 0.00000063, 0.00000777, 0.00000937, 0.00000420, -0.00000009, -0.00000309, -0.00000959, -0.00001259, -0.00000124, 0.00001322, 0.00000709, -0.00001522, -0.00002168, -0.00000317, 0.00001100, 0.00000313, -0.00000572, 0.00000061, 0.00000429, -0.00000418, -0.00000190, 0.00001259, -0.00000180, -0.00004512, -0.00004009, 0.00004035, 0.00008199, -0.00001030, -0.00012040, -0.00006346, 0.00009913, 0.00010853, -0.00007811, -0.00017078, 0.00001002, 0.00020751, 0.00009852, -0.00017963, -0.00021089, 0.00005187, 0.00020353, 0.00002064, -0.00019664, -0.00011805, 0.00009752, 0.00008909, -0.00011358, -0.00017052, -0.00002316, 0.00000643, -0.00022595, -0.00039958, -0.00019258, 0.00023823, 0.00049456, 0.00047900, 0.00041864, 0.00042984, 0.00035821, 0.00009145, -0.00021085, -0.00031882, -0.00021811, -0.00007478, -0.00000692, -0.00000499, -0.00002438, -0.00004498, -0.00004884, -0.00002189, 0.00001243, 0.00001335, -0.00001762, -0.00003776, -0.00002551, -0.00000441, 0.00000016, -0.00000620, -0.00000779, -0.00000456, -0.00000330, -0.00000235, 0.00000158, 0.00000411, 0.00000162, -0.00000098, 0.00000095, 0.00000327, 0.00000065, -0.00000398, -0.00000455, -0.00000142, 0.00000045, -0.00000060, -0.00000232, -0.00000350, -0.00000482, -0.00000564, },
		{ -0.00000352, -0.00000579, -0.00000618, -0.00000314, 0.00000137, 0.00000370, 0.00000261, 0.00000030, -0.00000082, -0.00000073, -0.00000036, 0.00000014, 0.00000029, -0.00000123, -0.00000363, -0.00000323, 0.00000124, 0.00000520, 0.00000425, 0.00000069, -0.00000086, -0.00000094, -0.00000221, -0.00000259, 0.00000067, 0.00000330, -0.00000020, -0.00000571, -0.00000497, 0.00000022, 0.00000140, -0.00000103, 0.00000071, 0.00000589, 0.00000644, 0.00000322, 0.00000371, 0.00000384, -0.00000619, -0.00001746, -0.00000990, 0.00001089, 0.00001466, -0.00000576, -0.00002068, -0.00001171, 0.00000144, 0.00000170, 0.00000258, 0.00001018, 0.00000252, -0.00001921, -0.00001224, 0.00002680, 0.00002611, -0.00004229, -0.00007645, 0.00000865, 0.00010534, 0.00004582, -0.00010970, -0.00012135, 0.00005118, 0.00014469, -0.00000458, -0.00017127, -0.00007566, 0.00015490, 0.00016073, -0.00008229, -0.00021225, -0.00004543, 0.00014624, 0.00007989, -0.00010633, -0.00011644, 0.00002066, 0.00004545, -0.00007092, -0.00010543, -0.00001267, -0.00002840, -0.00024188, -0.00036849, -0.00014880, 0.00024123, 0.00042456, 0.00034003, 0.00026176, 0.00034493, 0.00041568, 0.00027075, -0.00002024, -0.00022768, -0.00024871, -0.00016106, -0.00006858, -0.00001159, 0.00000123, -0.00002532, -0.00005349, -0.00004106, 0.00000053, 0.00001966, -0.00000227, -0.00003034, -0.00003211, -0.00001610, -0.00000524, -0.00000336, -0.00000280, -0.00000287, -0.00000536, -0.00000571, -0.00000097, 0.00000317, 0.00000186, -0.00000047, 0.00000146, 0.00000439, 0.00000258, -0.00000233, -0.00000444, -0.00000258, -0.00000055, -0.00000046, -0.00000115, -0.00000187, },
		{ -0.00000111, -0.00000100, -0.00000177, -0.00000430, -0.00000672, -0.00000603, -0.00000194, 0.00000246, 0.00000415, 0.00000309, 0.00000098, -0.00000105, -0.00000209, -0.00000106, 0.00000098, 0.00000064, -0.00000276, -0.00000456, -0.00000119, 0.00000373, 0.00000464, 0.00000233, 0.00000099, 0.00000014, -0.00000266, -0.00000434, -0.00000073, 0.00000384, 0.00000184, -0.00000426, -0.00000549, -0.00000165, -0.00000015, -0.00000105, 0.00000199, 0.00000646, 0.00000412, -0.00000081, 0.00000306, 0.00000983, 0.00000182, -0.00001625, -0.00001762, 0.00000258, 0.00001522, 0.00000351, -0.00001187, -0.00001226, -0.00000921, -0.00000979, 0.00000319, 0.00002500, 0.00001425, -0.00003012, -0.00003670, 0.00002428, 0.00005871, -0.00001448, -0.00009888, -0.00004248, 0.00009749, 0.00010141, -0.00006412, -0.00015447, -0.00001832, 0.00014007, 0.00006687, -0.00012759, -0.00012999, 0.00007403, 0.00016830, 0.00000644, -0.00016258, -0.00009782, 0.00006943, 0.00008122, -0.00003768, -0.00007463, -0.00000762, -0.00000192, -0.00007680, -0.00007237, 0.00001972, -0.00001612, -0.00024305, -0.00037511, -0.00015572, 0.00023986, 0.00041721, 0.00028916, 0.00014604, 0.00021215, 0.00036330, 0.00035494, 0.00015352, -0.00007514, -0.00019655, -0.00020379, -0.00014254, -0.00005678, 0.00000188, -0.00000164, -0.00003942, -0.00004959, -0.00001735, 0.00001296, 0.00000738, -0.00001709, -0.00002898, -0.00002384, -0.00001457, -0.00000680, -0.00000039, 0.00000098, -0.00000383, -0.00000741, -0.00000428, 0.00000067, 0.00000126, -0.00000002, 0.00000186, 0.00000489, 0.00000385, -0.00000065, -0.00000343, -0.00000266, -0.00000111, -0.00000089, },
		{ -0.00000078, -0.00000082, -0.00000149, -0.00000124, -0.00000079, -0.00000222, -0.00000538, -0.00000722, -0.00000528, -0.00000070, 0.00000342, 0.00000514, 0.00000405, 0.00000055, -0.00000308, -0.00000325, 0.00000013, 0.00000205, -0.00000079, -0.00000431, -0.00000300, 0.00000143, 0.00000345, 0.00000285, 0.00000291, 0.00000241, -0.00000163, -0.00000544, -0.00000289, 0.00000278, 0.00000285, -0.00000221, -0.00000421, -0.00000229, -0.00000247, -0.00000313, 0.00000243, 0.00000900, 0.00000475, -0.00000486, -0.00000188, 0.00001107, 0.00001015, -0.00000885, -0.00001915, -0.00000648, 0.00000785, 0.00000564, -0.00000086, -0.00000213, -0.00001210, -0.00002664, -0.00000997, 0.00003477, 0.00003838, -0.00002558, -0.00006505, -0.00000059, 0.00007920, 0.00003163, -0.00009291, -0.00009410, 0.00005662, 0.00013395, 0.00000180, -0.00014890, -0.00008399, 0.00009711, 0.00011037, -0.00006024, -0.00013872, -0.00000206, 0.00013015, 0.00005756, -0.00009343, -0.00009918, 0.00001000, 0.00004646, -0.00001093, -0.00002956, 0.00000455, -0.00002292, -0.00009995, -0.00007685, 0.00003815, 0.00001967, -0.00021154, -0.00037778, -0.00019449, 0.00020434, 0.00041751, 0.00029788, 0.00010053, 0.00009836, 0.00025202, 0.00033892, 0.00025650, 0.00008132, -0.00008039, -0.00017769, -0.00018913, -0.00012032, -0.00003113, 0.00000508, -0.00001791, -0.00004190, -0.00002890, -0.00000130, 0.00000611, -0.00000659, -0.00001926, -0.00002350, -0.00002180, -0.00001441, -0.00000322, 0.00000313, -0.00000015, -0.00000601, -0.00000623, -0.00000259, -0.00000082, -0.00000042, 0.00000216, 0.00000535, 0.00000471, 0.00000053, -0.00000249, -0.00000214, },
		{ -0.00000204, -0.00000192, -0.00000010, 0.00000016, -0.00000123, -0.00000188, -0.00000102, -0.00000087, -0.00000318, -0.00000631, -0.00000706, -0.00000436, 0.00000037, 0.00000497, 0.00000685, 0.00000389, -0.00000211, -0.00000512, -0.00000214, 0.00000191, 0.00000107, -0.00000268, -0.00000325, -0.00000050, 0.00000119, 0.00000167, 0.00000372, 0.00000494, 0.00000084, -0.00000492, -0.00000450, 0.00000066, 0.00000198, -0.00000116, -0.00000191, -0.00000066, -0.00000364, -0.00000669, 0.00000029, 0.00001118, 0.00000838, -0.00000571, -0.00000790, 0.00000675, 0.00001379, 0.00000044, -0.00001284, -0.00000944, -0.00000256, -0.00000270, 0.00000348, 0.00001435, -0.00000092, -0.00003842, -0.00003537, 0.00002888, 0.00006500, -0.00000083, -0.00008210, -0.00004189, 0.00007421, 0.00007890, -0.00005628, -0.00012503, -0.00000394, 0.00013019, 0.00006406, -0.00010761, -0.00012264, 0.00003469, 0.00011419, 0.00000268, -0.00010872, -0.00004792, 0.00007246, 0.00006381, -0.00003753, -0.00006722, -0.00001290, 0.00000770, -0.00002009, -0.00000765, 0.00003102, -0.00001241, -0.00011111, -0.00010057, 0.00002857, 0.00004896, -0.00015827, -0.00035406, -0.00023245, 0.00014042, 0.00039202, 0.00032139, 0.00011231, 0.00004002, 0.00014129, 0.00025739, 0.00027170, 0.00018920, 0.00005436, -0.00008971, -0.00018045, -0.00016886, -0.00008542, -0.00001657, -0.00000686, -0.00002485, -0.00002716, -0.00001285, -0.00000328, -0.00000463, -0.00001001, -0.00001655, -0.00002256, -0.00002115, -0.00001012, 0.00000086, 0.00000240, -0.00000246, -0.00000538, -0.00000473, -0.00000371, -0.00000226, 0.00000162, 0.00000576, 0.00000572, 0.00000153, },
		{ 0.00000684, 0.00000295, -0.00000164, -0.00000239, -0.00000002, 0.00000141, 0.00000009, -0.00000177, -0.00000175, -0.00000084, -0.00000157, -0.00000420, -0.00000655, -0.00000668, -0.00000371, 0.00000211, 0.00000763, 0.00000756, 0.00000101, -0.00000520, -0.00000459, 0.00000005, 0.00000153, -0.00000077, -0.00000190, -0.00000095, -0.00000095, -0.00000083, 0.00000273, 0.00000643, 0.00000378, -0.00000269, -0.00000431, -0.00000099, -0.00000033, -0.00000238, -0.00000049, 0.00000266, -0.00000206, -0.00000954, -0.00000429, 0.00001047, 0.00001250, -0.00000251, -0.00001079, 0.00000005, 0.00001056, 0.00000525, -0.00000258, -0.00000241, -0.00000748, -0.00001793, -0.00000562, 0.00002691, 0.00002314, -0.00003450, -0.00006304, 0.00000326, 0.00007984, 0.00003823, -0.00007588, -0.00008409, 0.00004143, 0.00010786, -0.00000087, -0.00012252, -0.00006180, 0.00009270, 0.00010091, -0.00004857, -0.00012433, -0.00002202, 0.00008503, 0.00003960, -0.00006184, -0.00005626, 0.00002350, 0.00003898, -0.00001319, -0.00002995, -0.00000313, -0.00001096, -0.00004267, -0.00001404, 0.00004645, 0.00001430, -0.00009843, -0.00011910, -0.00000041, 0.00005467, -0.00010673, -0.00030665, -0.00024649, 0.00007274, 0.00033761, 0.00032583, 0.00014635, 0.00003625, 0.00007111, 0.00016102, 0.00021975, 0.00022361, 0.00015921, 0.00002605, -0.00011465, -0.00017429, -0.00013404, -0.00006038, -0.00001878, -0.00001308, -0.00001553, -0.00001425, -0.00001261, -0.00001042, -0.00000705, -0.00000859, -0.00001710, -0.00002275, -0.00001699, -0.00000535, 0.00000118, 0.00000044, -0.00000237, -0.00000435, -0.00000573, -0.00000501, -0.00000040, 0.00000535, },
		{ -0.00000348, 0.00000337, 0.00000724, 0.00000482, -0.00000053, -0.00000298, -0.00000097, 0.00000183, 0.00000179, -0.00000039, -0.00000174, -0.00000150, -0.00000136, -0.00000244, -0.00000445, -0.00000661, -0.00000686, -0.00000238, 0.00000549, 0.00000962, 0.00000513, -0.00000285, -0.00000558, -0.00000250, 0.00000010, -0.00000001, 0.00000007, 0.00000036, -0.00000173, -0.00000347, 0.00000022, 0.00000606, 0.00000576, 0.00000019, -0.00000212, -0.00000062, -0.00000235, -0.00000565, -0.00000174, 0.00000552, 0.00000212, -0.00000933, -0.00000893, 0.00000632, 0.00001368, 0.00000225, -0.00000849, -0.00000367, 0.00000322, 0.00000139, 0.00000385, 0.00001208, 0.00000014, -0.00003043, -0.00002769, 0.00002477, 0.00005038, -0.00001069, -0.00007883, -0.00003591, 0.00007198, 0.00007692, -0.00004469, -0.00010960, -0.00000849, 0.00010657, 0.00005283, -0.00008847, -0.00009647, 0.00003822, 0.00010307, 0.00000468, -0.00009590, -0.00005412, 0.00004241, 0.00004463, -0.00002199, -0.00003790, 0.00000004, 0.00000681, -0.00001883, -0.00000959, 0.00002002, -0.00000381, -0.00005582, -0.00003566, 0.00004008, 0.00003607, -0.00006914, -0.00011811, -0.00002951, 0.00003899, -0.00007472, -0.00025355, -0.00023366, 0.00002385, 0.00027208, 0.00030164, 0.00017008, 0.00006132, 0.00005001, 0.00009041, 0.00014187, 0.00019347, 0.00020487, 0.00012677, -0.00001579, -0.00012869, -0.00015038, -0.00010541, -0.00005208, -0.00001904, -0.00000546, -0.00000634, -0.00001464, -0.00001776, -0.00001100, -0.00000508, -0.00000971, -0.00001863, -0.00001981, -0.00001217, -0.00000384, 0.00000012, 0.00000039, -0.00000180, -0.00000543, -0.00000709, },
		{ -0.00000295, -0.00000699, -0.00000613, 0.00000005, 0.00000600, 0.00000627, 0.00000147, -0.00000273, -0.00000234, 0.00000083, 0.00000257, 0.00000159, -0.00000025, -0.00000150, -0.00000209, -0.00000212, -0.00000236, -0.00000451, -0.00000760, -0.00000650, 0.00000109, 0.00000871, 0.00000821, 0.00000110, -0.00000409, -0.00000404, -0.00000246, -0.00000120, 0.00000122, 0.00000263, -0.00000070, -0.00000496, -0.00000267, 0.00000405, 0.00000581, 0.00000208, 0.00000093, 0.00000223, -0.00000208, -0.00000906, -0.00000583, 0.00000548, 0.00000656, -0.00000560, -0.00001047, 0.00000114, 0.00001036, 0.00000411, -0.00000317, -0.00000061, -0.00000168, -0.00000965, -0.00000056, 0.00002537, 0.00002075, -0.00002944, -0.00005379, 0.00000320, 0.00006678, 0.00002704, -0.00007203, -0.00007412, 0.00004041, 0.00009899, 0.00000131, -0.00010776, -0.00005702, 0.00007645, 0.00008585, -0.00003802, -0.00009891, -0.00001124, 0.00007692, 0.00003466, -0.00005570, -0.00005704, 0.00000734, 0.00002746, -0.00000334, -0.00001256, 0.00000192, -0.00001289, -0.00003859, -0.00001279, 0.00003578, 0.00001824, -0.00004906, -0.00005347, 0.00001802, 0.00004006, -0.00003996, -0.00009933, -0.00004373, 0.00001661, -0.00006485, -0.00021247, -0.00020700, 0.00000114, 0.00021658, 0.00026016, 0.00016917, 0.00008545, 0.00006109, 0.00006098, 0.00007685, 0.00013054, 0.00019098, 0.00018145, 0.00007969, -0.00004760, -0.00012260, -0.00012759, -0.00009114, -0.00004432, -0.00000854, 0.00000230, -0.00000851, -0.00002008, -0.00001738, -0.00000754, -0.00000531, -0.00001196, -0.00001743, -0.00001589, -0.00000996, -0.00000377, 0.00000043, 0.00000081, },
		{ -0.00000288, 0.00000106, -0.00000014, -0.00000461, -0.00000674, -0.00000321, 0.00000311, 0.00000617, 0.00000352, -0.00000118, -0.00000292, -0.00000099, 0.00000161, 0.00000271, 0.00000215, 0.00000014, -0.00000234, -0.00000303, -0.00000166, -0.00000196, -0.00000592, -0.00000843, -0.00000363, 0.00000504, 0.00000853, 0.00000468, -0.00000051, -0.00000319, -0.00000451, -0.00000390, 0.00000051, 0.00000445, 0.00000155, -0.00000472, -0.00000464, 0.00000154, 0.00000414, 0.00000187, 0.00000294, 0.00000635, 0.00000130, -0.00001007, -0.00001079, 0.00000178, 0.00000812, -0.00000068, -0.00000748, -0.00000094, 0.00000467, -0.00000008, -0.00000039, 0.00000820, 0.00000220, -0.00002059, -0.00001711, 0.00002711, 0.00004648, -0.00000963, -0.00007026, -0.00003252, 0.00006140, 0.00006441, -0.00004169, -0.00009553, -0.00000457, 0.00009499, 0.00004491, -0.00008021, -0.00008702, 0.00003108, 0.00008956, 0.00000854, -0.00007467, -0.00003915, 0.00003966, 0.00003727, -0.00002314, -0.00004022, -0.00000817, 0.00000531, -0.00000562, 0.00000305, 0.00001650, -0.00001318, -0.00005428, -0.00003122, 0.00003349, 0.00003805, -0.00002707, -0.00005527, -0.00000472, 0.00002756, -0.00002422, -0.00007542, -0.00004065, 0.00000231, -0.00006689, -0.00018991, -0.00018311, -0.00000334, 0.00018188, 0.00022003, 0.00014886, 0.00009133, 0.00007866, 0.00006382, 0.00004456, 0.00007008, 0.00014050, 0.00018287, 0.00014117, 0.00003910, -0.00005936, -0.00011301, -0.00011505, -0.00007786, -0.00002834, 0.00000140, -0.00000040, -0.00001548, -0.00002046, -0.00001288, -0.00000593, -0.00000708, -0.00001216, -0.00001502, -0.00001377, -0.00000907, },
		{ -0.00001359, -0.00001265, -0.00000780, -0.00000190, 0.00000063, -0.00000157, -0.00000494, -0.00000469, -0.00000019, 0.00000418, 0.00000431, 0.00000098, -0.00000190, -0.00000225, -0.00000069, 0.00000181, 0.00000388, 0.00000309, -0.00000086, -0.00000378, -0.00000249, -0.00000059, -0.00000322, -0.00000766, -0.00000658, 0.00000030, 0.00000575, 0.00000588, 0.00000341, 0.00000009, -0.00000447, -0.00000664, -0.00000197, 0.00000468, 0.00000374, -0.00000318, -0.00000503, -0.00000013, 0.00000196, -0.00000027, 0.00000252, 0.00000950, 0.00000658, -0.00000735, -0.00001345, -0.00000365, 0.00000510, 0.00000138, -0.00000203, 0.00000297, 0.00000178, -0.00000918, -0.00000561, 0.00001650, 0.00001633, -0.00002210, -0.00003935, 0.00001148, 0.00006448, 0.00002477, -0.00006598, -0.00006868, 0.00003288, 0.00008570, 0.00000199, -0.00009116, -0.00004620, 0.00006723, 0.00007055, -0.00003962, -0.00009083, -0.00001122, 0.00006869, 0.00003579, -0.00003948, -0.00004141, 0.00000947, 0.00002145, -0.00000913, -0.00001959, -0.00000502, -0.00000753, -0.00001795, 0.00000302, 0.00002867, -0.00000005, -0.00005566, -0.00005012, 0.00001596, 0.00004363, -0.00000327, -0.00004142, -0.00001558, 0.00000987, -0.00002426, -0.00005934, -0.00002895, 0.00000168, -0.00006738, -0.00018017, -0.00017117, -0.00000383, 0.00016491, 0.00019388, 0.00012428, 0.00007892, 0.00008423, 0.00007809, 0.00004282, 0.00003428, 0.00008409, 0.00014614, 0.00015669, 0.00010263, 0.00001577, -0.00006497, -0.00010926, -0.00010207, -0.00005744, -0.00001365, 0.00000094, -0.00000826, -0.00001785, -0.00001628, -0.00000955, -0.00000622, -0.00000772, -0.00001116, },
		{ -0.00000640, -0.00000753, -0.00001050, -0.00001270, -0.00001144, -0.00000669, -0.00000175, -0.00000021, -0.00000208, -0.00000369, -0.00000211, 0.00000129, 0.00000311, 0.00000230, 0.00000029, -0.00000175, -0.00000280, -0.00000099, 0.00000342, 0.00000573, 0.00000234, -0.00000287, -0.00000358, -0.00000092, -0.00000130, -0.00000529, -0.00000675, -0.00000324, 0.00000119, 0.00000383, 0.00000550, 0.00000442, -0.00000181, -0.00000776, -0.00000502, 0.00000313, 0.00000471, -0.00000148, -0.00000422, -0.00000032, 0.00000069, -0.00000298, -0.00000020, 0.00000988, 0.00001108, -0.00000199, -0.00001168, -0.00000691, -0.00000081, -0.00000245, 0.00000087, 0.00001174, 0.00000674, -0.00001736, -0.00001999, 0.00001611, 0.00003525, -0.00000857, -0.00005572, -0.00001851, 0.00006356, 0.00006100, -0.00003880, -0.00008972, -0.00000911, 0.00008184, 0.00004244, -0.00006359, -0.00006965, 0.00002826, 0.00007183, -0.00000272, -0.00007300, -0.00003658, 0.00003730, 0.00003957, -0.00001020, -0.00002606, -0.00000329, 0.00000217, -0.00001256, -0.00000817, 0.00000654, -0.00000643, -0.00002745, -0.00000761, 0.00002973, 0.00001452, -0.00004391, -0.00005832, -0.00000555, 0.00003398, 0.00001029, -0.00002186, -0.00001173, -0.00000093, -0.00003235, -0.00005608, -0.00002026, 0.00000988, -0.00005886, -0.00017219, -0.00016955, -0.00001061, 0.00015463, 0.00018262, 0.00010832, 0.00005938, 0.00007366, 0.00008561, 0.00005585, 0.00002632, 0.00004456, 0.00009682, 0.00013465, 0.00012817, 0.00007689, -0.00000133, -0.00007322, -0.00010278, -0.00008142, -0.00003815, -0.00000948, -0.00000572, -0.00001246, -0.00001521, -0.00001219, -0.00000816, },
		{ -0.00001139, -0.00000962, -0.00000750, -0.00000635, -0.00000728, -0.00000991, -0.00001187, -0.00001059, -0.00000620, -0.00000189, -0.00000048, -0.00000126, -0.00000162, -0.00000069, 0.00000057, 0.00000169, 0.00000220, 0.00000050, -0.00000314, -0.00000418, 0.00000051, 0.00000622, 0.00000576, 0.00000010, -0.00000324, -0.00000197, -0.00000112, -0.00000318, -0.00000480, -0.00000410, -0.00000290, -0.00000067, 0.00000440, 0.00000767, 0.00000252, -0.00000637, -0.00000702, 0.00000068, 0.00000410, -0.00000068, -0.00000320, 0.00000058, 0.00000092, -0.00000468, -0.00000361, 0.00000755, 0.00001244, 0.00000277, -0.00000628, -0.00000465, -0.00000491, -0.00001143, -0.00000385, 0.00001976, 0.00002060, -0.00001699, -0.00003841, 0.00000200, 0.00004871, 0.00001777, -0.00005569, -0.00005159, 0.00004071, 0.00008388, 0.00000249, -0.00008621, -0.00004833, 0.00005516, 0.00006500, -0.00002539, -0.00006919, -0.00000587, 0.00005374, 0.00001825, -0.00004635, -0.00004136, 0.00001053, 0.00002687, 0.00000334, -0.00000692, 0.00000057, -0.00000833, -0.00002371, -0.00000991, 0.00001479, 0.00000363, -0.00002700, -0.00001878, 0.00002052, 0.00002116, -0.00002782, -0.00005373, -0.00002002, 0.00001713, 0.00000983, -0.00000801, 0.00000019, 0.00000027, -0.00003773, -0.00006126, -0.00002105, 0.00001742, -0.00004292, -0.00015801, -0.00017006, -0.00002505, 0.00014103, 0.00017856, 0.00010471, 0.00004474, 0.00005427, 0.00007938, 0.00006686, 0.00003531, 0.00002909, 0.00005694, 0.00009538, 0.00011869, 0.00010789, 0.00005543, -0.00001989, -0.00007691, -0.00008717, -0.00006051, -0.00002888, -0.00001273, -0.00001028, -0.00001158, },
		{ -0.00001583, -0.00001035, -0.00000848, -0.00000837, -0.00000829, -0.00000724, -0.00000607, -0.00000673, -0.00000946, -0.00001156, -0.00001020, -0.00000591, -0.00000176, 0.00000036, 0.00000062, -0.00000040, -0.00000158, -0.00000067, 0.00000236, 0.00000314, -0.00000123, -0.00000581, -0.00000351, 0.00000381, 0.00000728, 0.00000387, -0.00000066, -0.00000198, -0.00000208, -0.00000269, -0.00000249, -0.00000239, -0.00000462, -0.00000553, 0.00000034, 0.00000809, 0.00000659, -0.00000293, -0.00000689, -0.00000126, 0.00000251, -0.00000117, -0.00000291, 0.00000161, 0.00000213, -0.00000478, -0.00000576, 0.00000452, 0.00001043, 0.00000356, -0.00000115, 0.00000322, -0.00000202, -0.00002035, -0.00001748, 0.00001957, 0.00003843, -0.00000340, -0.00005131, -0.00002370, 0.00004714, 0.00004695, -0.00003591, -0.00007359, 0.00000334, 0.00008355, 0.00004227, -0.00005983, -0.00007008, 0.00001800, 0.00006405, 0.00000761, -0.00005024, -0.00002385, 0.00002878, 0.00002068, -0.00002440, -0.00003192, -0.00000279, 0.00000998, 0.00000148, 0.00000454, 0.00001212, -0.00000620, -0.00003155, -0.00001994, 0.00001348, 0.00001343, -0.00001787, -0.00002276, 0.00000832, 0.00001793, -0.00001651, -0.00004226, -0.00002310, 0.00000328, 0.00000035, -0.00000552, 0.00001014, 0.00000988, -0.00003403, -0.00006601, -0.00002967, 0.00001743, -0.00002680, -0.00013701, -0.00016520, -0.00004164, 0.00012084, 0.00017274, 0.00010876, 0.00004051, 0.00003657, 0.00006325, 0.00006710, 0.00004665, 0.00003128, 0.00003634, 0.00005906, 0.00008952, 0.00010768, 0.00008919, 0.00003182, -0.00003349, -0.00007015, -0.00006900, -0.00004805, -0.00002782, },
		{ -0.00005673, -0.00004361, -0.00002842, -0.00001560, -0.00000763, -0.00000535, -0.00000674, -0.00000787, -0.00000673, -0.00000524, -0.00000639, -0.00000968, -0.00001157, -0.00000989, -0.00000558, -0.00000070, 0.00000254, 0.00000193, -0.00000171, -0.00000324, 0.00000047, 0.00000446, 0.00000186, -0.00000487, -0.00000647, -0.00000049, 0.00000565, 0.00000620, 0.00000329, 0.00000030, -0.00000249, -0.00000371, -0.00000149, 0.00000039, -0.00000349, -0.00000864, -0.00000499, 0.00000533, 0.00000849, 0.00000099, -0.00000463, -0.00000156, 0.00000103, -0.00000256, -0.00000357, 0.00000203, 0.00000320, -0.00000417, -0.00000601, 0.00000345, 0.00000763, -0.00000056, -0.00000120, 0.00001245, 0.00000988, -0.00002184, -0.00003554, 0.00000644, 0.00005098, 0.00002131, -0.00004975, -0.00005169, 0.00002744, 0.00006607, -0.00000266, -0.00007483, -0.00003415, 0.00006067, 0.00006582, -0.00002235, -0.00006844, -0.00001390, 0.00004504, 0.00002439, -0.00002533, -0.00002389, 0.00000957, 0.00001112, -0.00001455, -0.00001919, -0.00000298, -0.00000085, -0.00000800, 0.00000443, 0.00002107, 0.00000410, -0.00003012, -0.00002951, 0.00000407, 0.00001614, -0.00000675, -0.00001834, 0.00000058, 0.00000957, -0.00001382, -0.00003221, -0.00001803, -0.00000170, -0.00000933, -0.00001212, 0.00001149, 0.00001968, -0.00002239, -0.00006365, -0.00003926, 0.00000963, -0.00001703, -0.00011437, -0.00015297, -0.00005376, 0.00009736, 0.00016083, 0.00011297, 0.00004457, 0.00002693, 0.00004571, 0.00005742, 0.00005083, 0.00003921, 0.00003248, 0.00003692, 0.00005827, 0.00008714, 0.00009613, 0.00006669, 0.00001198, -0.00003636, -0.00005845, },
		{ 0.00000172, -0.00003271, -0.00004970, -0.00005126, -0.00004201, -0.00002703, -0.00001252, -0.00000424, -0.00000359, -0.00000632, -0.00000717, -0.00000550, -0.00000477, -0.00000689, -0.00001005, -0.00001154, -0.00000982, -0.00000451, 0.00000204, 0.00000446, 0.00000043, -0.00000421, -0.00000230, 0.00000358, 0.00000427, -0.00000196, -0.00000676, -0.00000443, 0.00000138, 0.00000533, 0.00000642, 0.00000447, -0.00000078, -0.00000477, -0.00000222, 0.00000243, -0.00000053, -0.00000889, -0.00000940, 0.00000051, 0.00000732, 0.00000353, -0.00000132, 0.00000012, 0.00000065, -0.00000411, -0.00000480, 0.00000200, 0.00000353, -0.00000450, -0.00000577, 0.00000549, 0.00000794, -0.00000675, -0.00000906, 0.00001620, 0.00002729, -0.00001072, -0.00004908, -0.00001790, 0.00004948, 0.00004822, -0.00003086, -0.00006994, -0.00000459, 0.00006613, 0.00003104, -0.00005518, -0.00005760, 0.00002583, 0.00006658, 0.00001067, -0.00004850, -0.00002943, 0.00002056, 0.00002349, -0.00000683, -0.00001296, 0.00000252, -0.00000006, -0.00001581, -0.00001194, 0.00000312, -0.00000211, -0.00001599, -0.00000360, 0.00002167, 0.00001449, -0.00002071, -0.00003216, -0.00000703, 0.00001095, -0.00000029, -0.00001000, 0.00000024, 0.00000282, -0.00001730, -0.00002861, -0.00001177, 0.00000129, -0.00001268, -0.00002076, 0.00000437, 0.00002289, -0.00000915, -0.00005357, -0.00004340, -0.00000094, -0.00001523, -0.00009603, -0.00013678, -0.00005856, 0.00007609, 0.00014424, 0.00011251, 0.00005106, 0.00002535, 0.00003329, 0.00004407, 0.00004636, 0.00004358, 0.00003635, 0.00002943, 0.00003644, 0.00006097, 0.00008337, 0.00007927, 0.00004576, },
		{ 0.00007380, 0.00006171, 0.00003207, -0.00000175, -0.00002942, -0.00004581, -0.00004890, -0.00003951, -0.00002305, -0.00000845, -0.00000219, -0.00000320, -0.00000550, -0.00000559, -0.00000482, -0.00000533, -0.00000736, -0.00001025, -0.00001204, -0.00000924, -0.00000140, 0.00000499, 0.00000337, -0.00000282, -0.00000405, 0.00000123, 0.00000460, 0.00000107, -0.00000442, -0.00000588, -0.00000337, 0.00000118, 0.00000655, 0.00000857, 0.00000323, -0.00000414, -0.00000364, 0.00000278, 0.00000246, -0.00000674, -0.00001133, -0.00000434, 0.00000356, 0.00000338, 0.00000145, 0.00000326, 0.00000179, -0.00000531, -0.00000620, 0.00000227, 0.00000386, -0.00000661, -0.00000774, 0.00000889, 0.00001392, -0.00000947, -0.00002265, 0.00000899, 0.00004206, 0.00001221, -0.00004916, -0.00004491, 0.00003094, 0.00006564, -0.00000023, -0.00006997, -0.00003679, 0.00004690, 0.00005204, -0.00002401, -0.00006014, -0.00000619, 0.00004871, 0.00002778, -0.00002270, -0.00002701, 0.00000302, 0.00001215, -0.00000069, -0.00000140, 0.00000588, -0.00000508, -0.00002152, -0.00001269, 0.00000776, 0.00000307, -0.00001737, -0.00001248, 0.00001468, 0.00001877, -0.00000925, -0.00002706, -0.00001351, 0.00000222, -0.00000087, -0.00000373, 0.00000484, 0.00000166, -0.00002137, -0.00003102, -0.00000987, 0.00000686, -0.00000880, -0.00002494, -0.00000597, 0.00001809, -0.00000071, -0.00004033, -0.00004023, -0.00000842, -0.00001799, -0.00008448, -0.00012176, -0.00005788, 0.00006034, 0.00012718, 0.00010726, 0.00005528, 0.00002780, 0.00002752, 0.00003304, 0.00003729, 0.00004134, 0.00003959, 0.00003040, 0.00002663, 0.00004007, 0.00006270, },
		{ 0.00002830, 0.00004383, 0.00005935, 0.00006203, 0.00004865, 0.00002418, -0.00000396, -0.00002895, -0.00004455, -0.00004618, -0.00003470, -0.00001805, -0.00000587, -0.00000165, -0.00000215, -0.00000357, -0.00000478, -0.00000545, -0.00000570, -0.00000747, -0.00001127, -0.00001239, -0.00000632, 0.00000251, 0.00000485, -0.00000010, -0.00000359, -0.00000083, 0.00000299, 0.00000241, -0.00000106, -0.00000420, -0.00000609, -0.00000443, 0.00000293, 0.00001017, 0.00000798, -0.00000109, -0.00000402, 0.00000192, 0.00000413, -0.00000369, -0.00001052, -0.00000734, -0.00000123, 0.00000031, 0.00000224, 0.00000684, 0.00000445, -0.00000600, -0.00000791, 0.00000332, 0.00000589, -0.00000931, -0.00001388, 0.00000950, 0.00002439, -0.00000367, -0.00003529, -0.00000958, 0.00004531, 0.00003933, -0.00003209, -0.00006296, 0.00000064, 0.00006526, 0.00003039, -0.00005168, -0.00005684, 0.00001703, 0.00005378, 0.00000502, -0.00004502, -0.00002398, 0.00002388, 0.00002675, -0.00000365, -0.00001389, -0.00000162, 0.00000091, -0.00000469, 0.00000336, 0.00001268, -0.00000235, -0.00002466, -0.00001810, 0.00000691, 0.00000874, -0.00001268, -0.00001676, 0.00000516, 0.00001583, -0.00000168, -0.00001821, -0.00001310, -0.00000435, -0.00000599, -0.00000289, 0.00000940, 0.00000541, -0.00002163, -0.00003524, -0.00001339, 0.00000960, -0.00000147, -0.00002249, -0.00001348, 0.00000893, 0.00000046, -0.00002972, -0.00003274, -0.00001017, -0.00002033, -0.00007788, -0.00011085, -0.00005573, 0.00004959, 0.00011288, 0.00010000, 0.00005601, 0.00003019, 0.00002609, 0.00002671, 0.00002849, 0.00003454, 0.00003837, 0.00003290, 0.00002513, },
		{ 0.00003299, 0.00002665, 0.00002392, 0.00003106, 0.00004406, 0.00005349, 0.00005239, 0.00003957, 0.00001800, -0.00000746, -0.00003027, -0.00004286, -0.00004111, -0.00002888, -0.00001491, -0.00000507, -0.00000016, 0.00000037, -0.00000259, -0.00000566, -0.00000581, -0.00000544, -0.00000863, -0.00001257, -0.00001022, -0.00000197, 0.00000345, 0.00000182, -0.00000139, -0.00000101, 0.00000095, 0.00000152, 0.00000122, -0.00000054, -0.00000541, -0.00000877, -0.00000313, 0.00000789, 0.00001100, 0.00000340, -0.00000218, 0.00000137, 0.00000433, -0.00000134, -0.00000796, -0.00000752, -0.00000512, -0.00000485, 0.00000019, 0.00000975, 0.00000847, -0.00000605, -0.00001068, 0.00000437, 0.00001097, -0.00000954, -0.00002339, 0.00000320, 0.00003418, 0.00001166, -0.00003897, -0.00003347, 0.00003246, 0.00005940, -0.00000225, -0.00006335, -0.00002986, 0.00004687, 0.00004905, -0.00002342, -0.00005860, -0.00001070, 0.00003911, 0.00002117, -0.00002282, -0.00002473, 0.00000448, 0.00001422, 0.00000220, -0.00000084, 0.00000426, -0.00000277, -0.00001148, 0.00000029, 0.00001608, 0.00000424, -0.00002218, -0.00002272, 0.00000163, 0.00001060, -0.00000612, -0.00001535, -0.00000169, 0.00000896, -0.00000040, -0.00001096, -0.00000808, -0.00000560, -0.00001091, -0.00000661, 0.00001015, 0.00001030, -0.00001739, -0.00003696, -0.00001950, 0.00000740, 0.00000436, -0.00001577, -0.00001519, 0.00000073, -0.00000344, -0.00002477, -0.00002580, -0.00000777, -0.00001930, -0.00007233, -0.00010341, -0.00005497, 0.00004107, 0.00010172, 0.00009342, 0.00005474, 0.00003080, 0.00002589, 0.00002420, 0.00002264, 0.00002678, 0.00003318, },
		{ 0.00002042, 0.00002626, 0.00003011, 0.00002772, 0.00002336, 0.00002407, 0.00003201, 0.00004253, 0.00004832, 0.00004485, 0.00003174, 0.00001122, -0.00001209, -0.00003072, -0.00003830, -0.00003485, -0.00002541, -0.00001403, -0.00000307, 0.00000333, 0.00000188, -0.00000374, -0.00000630, -0.00000524, -0.00000635, -0.00001052, -0.00001135, -0.00000607, -0.00000034, 0.00000117, 0.00000069, 0.00000080, 0.00000023, -0.00000055, 0.00000123, 0.00000290, -0.00000186, -0.00000980, -0.00000896, 0.00000231, 0.00001035, 0.00000705, 0.00000157, 0.00000259, 0.00000408, -0.00000051, -0.00000519, -0.00000503, -0.00000649, -0.00001060, -0.00000490, 0.00001085, 0.00001364, -0.00000484, -0.00001504, 0.00000338, 0.00001849, -0.00000405, -0.00003217, -0.00001107, 0.00003655, 0.00003195, -0.00002890, -0.00005266, 0.00000641, 0.00006323, 0.00002938, -0.00004514, -0.00004844, 0.00001870, 0.00004976, 0.00000244, -0.00004491, -0.00002619, 0.00001789, 0.00002156, -0.00000520, -0.00001420, -0.00000271, 0.00000069, -0.00000324, 0.00000425, 0.00001301, 0.00000197, -0.00001405, -0.00000619, 0.00001352, 0.00000917, -0.00001584, -0.00002307, -0.00000413, 0.00000821, -0.00000179, -0.00001090, -0.00000372, 0.00000276, -0.00000369, -0.00000846, -0.00000286, -0.00000228, -0.00001207, -0.00001128, 0.00000672, 0.00001265, -0.00001123, -0.00003459, -0.00002432, 0.00000186, 0.00000594, -0.00000905, -0.00001223, -0.00000314, -0.00000799, -0.00002453, -0.00002259, -0.00000500, -0.00001543, -0.00006511, -0.00009675, -0.00005569, 0.00003249, 0.00009206, 0.00008824, 0.00005331, 0.00003011, 0.00002524, 0.00002351, 0.00001981, },
		{ 0.00002332, 0.00001906, 0.00001614, 0.00001936, 0.00002524, 0.00002725, 0.00002402, 0.00002083, 0.00002366, 0.00003250, 0.00004130, 0.00004389, 0.00003772, 0.00002318, 0.00000338, -0.00001541, -0.00002738, -0.00003195, -0.00003133, -0.00002490, -0.00001188, 0.00000154, 0.00000594, 0.00000083, -0.00000483, -0.00000576, -0.00000583, -0.00000836, -0.00000995, -0.00000775, -0.00000442, -0.00000215, 0.00000066, 0.00000316, 0.00000158, -0.00000201, -0.00000061, 0.00000436, 0.00000255, -0.00000727, -0.00001196, -0.00000418, 0.00000583, 0.00000756, 0.00000541, 0.00000595, 0.00000485, -0.00000093, -0.00000361, -0.00000096, -0.00000446, -0.00001509, -0.00001254, 0.00000899, 0.00001937, -0.00000120, -0.00002029, -0.00000188, 0.00002537, 0.00000814, -0.00003462, -0.00002973, 0.00002703, 0.00004869, -0.00000671, -0.00005819, -0.00002326, 0.00004864, 0.00005053, -0.00001599, -0.00004865, -0.00000683, 0.00003541, 0.00001613, -0.00002532, -0.00002683, 0.00000101, 0.00001143, 0.00000130, -0.00000216, 0.00000108, -0.00000572, -0.00001252, 0.00000011, 0.00001716, 0.00001035, -0.00000994, -0.00001026, 0.00000733, 0.00000934, -0.00000986, -0.00001955, -0.00000697, 0.00000425, -0.00000105, -0.00000703, -0.00000222, 0.00000002, -0.00000781, -0.00001033, -0.00000089, 0.00000223, -0.00000913, -0.00001344, 0.00000162, 0.00001125, -0.00000644, -0.00002960, -0.00002570, -0.00000373, 0.00000383, -0.00000537, -0.00000803, -0.00000277, -0.00000982, -0.00002558, -0.00002301, -0.00000477, -0.00001146, -0.00005619, -0.00008867, -0.00005613, 0.00002353, 0.00008230, 0.00008357, 0.00005232, 0.00002909, 0.00002398, },
		{ 0.00002799, 0.00002232, 0.00002324, 0.00001973, 0.00001387, 0.00001325, 0.00001924, 0.00002521, 0.00002482, 0.00001989, 0.00001820, 0.00002399, 0.00003363, 0.00004028, 0.00003926, 0.00002923, 0.00001290, -0.00000314, -0.00001389, -0.00002118, -0.00002845, -0.00003162, -0.00002335, -0.00000654, 0.00000562, 0.00000570, -0.00000042, -0.00000464, -0.00000639, -0.00000781, -0.00000795, -0.00000670, -0.00000653, -0.00000643, -0.00000216, 0.00000402, 0.00000408, -0.00000163, -0.00000270, 0.00000353, 0.00000563, -0.00000275, -0.00001103, -0.00000865, -0.00000075, 0.00000390, 0.00000696, 0.00001042, 0.00000766, -0.00000154, -0.00000394, 0.00000314, 0.00000096, -0.00001646, -0.00002146, 0.00000316, 0.00002426, 0.00000585, -0.00002414, -0.00001199, 0.00002720, 0.00002427, -0.00002674, -0.00004557, 0.00000659, 0.00005370, 0.00001961, -0.00004699, -0.00004490, 0.00002187, 0.00005380, 0.00001173, -0.00003289, -0.00001964, 0.00001567, 0.00001530, -0.00001037, -0.00001779, -0.00000540, -0.00000010, -0.00000240, 0.00000360, 0.00000905, -0.00000321, -0.00001801, -0.00000865, 0.00001383, 0.00001604, -0.00000184, -0.00000880, 0.00000215, 0.00000568, -0.00000742, -0.00001523, -0.00000634, 0.00000176, -0.00000242, -0.00000567, -0.00000008, 0.00000057, -0.00000979, -0.00001382, -0.00000275, 0.00000466, -0.00000439, -0.00001197, -0.00000206, 0.00000759, -0.00000476, -0.00002477, -0.00002416, -0.00000697, 0.00000057, -0.00000505, -0.00000559, -0.00000071, -0.00000831, -0.00002485, -0.00002456, -0.00000741, -0.00000982, -0.00004757, -0.00007905, -0.00005471, 0.00001537, 0.00007229, 0.00007852, 0.00005133, },
		{ 0.00007324, 0.00005000, 0.00002647, 0.00001997, 0.00002286, 0.00002148, 0.00001373, 0.00000853, 0.00001260, 0.00002147, 0.00002510, 0.00002065, 0.00001537, 0.00001696, 0.00002552, 0.00003517, 0.00003907, 0.00003271, 0.00001783, 0.00000323, -0.00000404, -0.00000809, -0.00001744, -0.00002949, -0.00003135, -0.00001823, -0.00000129, 0.00000677, 0.00000504, -0.00000041, -0.00000573, -0.00000866, -0.00000729, -0.00000449, -0.00000584, -0.00000935, -0.00000659, 0.00000222, 0.00000586, 0.00000035, -0.00000347, 0.00000155, 0.00000627, 0.00000136, -0.00000704, -0.00000914, -0.00000647, -0.00000301, 0.00000446, 0.00001394, 0.00001243, -0.00000110, -0.00000609, 0.00000563, 0.00000883, -0.00001316, -0.00002950, -0.00000695, 0.00002600, 0.00001595, -0.00002350, -0.00002478, 0.00002073, 0.00003847, -0.00000894, -0.00005123, -0.00001811, 0.00004362, 0.00003955, -0.00002380, -0.00005067, -0.00000558, 0.00004032, 0.00002730, -0.00001071, -0.00001701, 0.00000133, 0.00000532, -0.00000583, -0.00000790, -0.00000239, -0.00000576, -0.00001003, 0.00000115, 0.00001393, 0.00000419, -0.00001634, -0.00001555, 0.00000553, 0.00001549, 0.00000511, -0.00000305, 0.00000130, 0.00000176, -0.00000860, -0.00001313, -0.00000428, 0.00000192, -0.00000335, -0.00000628, 0.00000058, 0.00000234, -0.00000896, -0.00001611, -0.00000665, 0.00000376, -0.00000076, -0.00000825, -0.00000277, 0.00000426, -0.00000552, -0.00002201, -0.00002182, -0.00000763, -0.00000149, -0.00000629, -0.00000579, 0.00000034, -0.00000524, -0.00002141, -0.00002462, -0.00001109, -0.00001095, -0.00004116, -0.00006941, -0.00005142, 0.00000908, 0.00006292, },
		{ 0.00000447, 0.00005486, 0.00006857, 0.00004888, 0.00002461, 0.00001644, 0.00002127, 0.00002355, 0.00001582, 0.00000612, 0.00000618, 0.00001585, 0.00002379, 0.00002229, 0.00001545, 0.00001225, 0.00001716, 0.00002796, 0.00003726, 0.00003593, 0.00002189, 0.00000561, -0.00000062, 0.00000081, -0.00000388, -0.00001867, -0.00003059, -0.00002742, -0.00001279, 0.00000116, 0.00000747, 0.00000561, -0.00000196, -0.00000881, -0.00000828, -0.00000321, -0.00000344, -0.00000959, -0.00001042, -0.00000169, 0.00000538, 0.00000241, -0.00000248, 0.00000013, 0.00000495, 0.00000321, -0.00000225, -0.00000553, -0.00000860, -0.00001068, -0.00000230, 0.00001441, 0.00001806, 0.00000126, -0.00000937, 0.00000516, 0.00001726, -0.00000459, -0.00003379, -0.00002014, 0.00002216, 0.00002672, -0.00001631, -0.00003569, 0.00000612, 0.00004445, 0.00001342, -0.00004322, -0.00003769, 0.00002207, 0.00004541, 0.00000121, -0.00004043, -0.00002289, 0.00001870, 0.00002690, 0.00000658, -0.00000426, -0.00000157, -0.00000465, -0.00001031, -0.00000408, 0.00000391, -0.00000384, -0.00001478, -0.00000588, 0.00001230, 0.00001034, -0.00000951, -0.00001660, -0.00000246, 0.00000969, 0.00000710, 0.00000294, 0.00000462, 0.00000082, -0.00001083, -0.00001401, -0.00000343, 0.00000357, -0.00000243, -0.00000698, -0.00000041, 0.00000333, -0.00000662, -0.00001602, -0.00001013, 0.00000052, 0.00000014, -0.00000471, -0.00000110, 0.00000300, -0.00000677, -0.00002137, -0.00002055, -0.00000715, -0.00000161, -0.00000702, -0.00000752, -0.00000078, -0.00000291, -0.00001646, -0.00002218, -0.00001361, -0.00001344, -0.00003737, -0.00006127, -0.00004753, },
		{ -0.00005490, -0.00004422, 0.00000037, 0.00004761, 0.00006487, 0.00004899, 0.00002346, 0.00001202, 0.00001754, 0.00002440, 0.00001932, 0.00000682, 0.00000162, 0.00000909, 0.00001982, 0.00002301, 0.00001789, 0.00001121, 0.00001022, 0.00001861, 0.00003216, 0.00003815, 0.00002750, 0.00000840, -0.00000138, 0.00000238, 0.00000524, -0.00000457, -0.00002045, -0.00002819, -0.00002302, -0.00001002, 0.00000337, 0.00000957, 0.00000430, -0.00000607, -0.00000920, -0.00000367, -0.00000136, -0.00000766, -0.00001195, -0.00000582, 0.00000251, 0.00000281, -0.00000066, 0.00000039, 0.00000323, 0.00000235, 0.00000087, 0.00000048, -0.00000584, -0.00001604, -0.00001185, 0.00001053, 0.00002281, 0.00000583, -0.00001253, 0.00000099, 0.00002363, 0.00000836, -0.00003158, -0.00003337, 0.00001164, 0.00003414, -0.00000330, -0.00003990, -0.00001244, 0.00003890, 0.00003224, -0.00002405, -0.00004469, -0.00000205, 0.00003622, 0.00001771, -0.00002144, -0.00002526, 0.00000008, 0.00001490, 0.00001201, 0.00000903, 0.00000563, -0.00000742, -0.00001730, -0.00000764, 0.00000698, 0.00000204, -0.00001357, -0.00001179, 0.00000586, 0.00001127, -0.00000244, -0.00001229, -0.00000597, 0.00000278, 0.00000401, 0.00000543, 0.00000912, 0.00000352, -0.00001115, -0.00001620, -0.00000494, 0.00000456, -0.00000014, -0.00000633, -0.00000177, 0.00000294, -0.00000449, -0.00001421, -0.00001175, -0.00000307, -0.00000135, -0.00000307, 0.00000110, 0.00000380, -0.00000684, -0.00002150, -0.00002083, -0.00000712, -0.00000061, -0.00000626, -0.00000906, -0.00000345, -0.00000253, -0.00001187, -0.00001804, -0.00001402, -0.00001549, -0.00003507, },
		{ -0.00001630, -0.00003282, -0.00004928, -0.00004147, -0.00000414, 0.00003990, 0.00006126, 0.00005054, 0.00002443, 0.00000826, 0.00001188, 0.00002244, 0.00002219, 0.00001010, 0.00000055, 0.00000329, 0.00001343, 0.00002076, 0.00002048, 0.00001397, 0.00000719, 0.00000921, 0.00002300, 0.00003659, 0.00003351, 0.00001461, -0.00000111, -0.00000107, 0.00000609, 0.00000526, -0.00000595, -0.00001891, -0.00002548, -0.00002135, -0.00000715, 0.00000745, 0.00000958, -0.00000036, -0.00000770, -0.00000463, -0.00000086, -0.00000537, -0.00001105, -0.00000824, -0.00000138, 0.00000080, 0.00000034, 0.00000217, 0.00000272, -0.00000022, 0.00000080, 0.00000622, 0.00000123, -0.00001668, -0.00002153, 0.00000230, 0.00002490, 0.00001217, -0.00001408, -0.00000649, 0.00002526, 0.00002288, -0.00002157, -0.00004229, -0.00000404, 0.00003442, 0.00001172, -0.00003497, -0.00002844, 0.00002344, 0.00004069, -0.00000129, -0.00003743, -0.00001899, 0.00001806, 0.00002045, -0.00000409, -0.00001582, -0.00000785, 0.00000057, 0.00000599, 0.00001481, 0.00001604, -0.00000159, -0.00001994, -0.00001454, 0.00000411, 0.00000638, -0.00000811, -0.00001293, -0.00000088, 0.00000723, 0.00000075, -0.00000642, -0.00000437, -0.00000113, -0.00000112, 0.00000345, 0.00001133, 0.00000779, -0.00000845, -0.00001719, -0.00000784, 0.00000362, 0.00000188, -0.00000439, -0.00000221, 0.00000192, -0.00000341, -0.00001202, -0.00001161, -0.00000550, -0.00000375, -0.00000350, 0.00000220, 0.00000539, -0.00000534, -0.00002098, -0.00002186, -0.00000825, 0.00000022, -0.00000439, -0.00000941, -0.00000635, -0.00000384, -0.00000883, -0.00001372, -0.00001288, },
		{ -0.00001023, -0.00001135, -0.00001633, -0.00003009, -0.00004325, -0.00003819, -0.00000867, 0.00003093, 0.00005611, 0.00005230, 0.00002789, 0.00000711, 0.00000617, 0.00001752, 0.00002212, 0.00001370, 0.00000305, 0.00000091, 0.00000684, 0.00001497, 0.00002026, 0.00001834, 0.00000925, 0.00000330, 0.00001185, 0.00002934, 0.00003605, 0.00002311, 0.00000415, -0.00000359, 0.00000065, 0.00000604, 0.00000493, -0.00000393, -0.00001738, -0.00002591, -0.00001937, -0.00000168, 0.00000956, 0.00000545, -0.00000301, -0.00000380, -0.00000130, -0.00000426, -0.00000909, -0.00000820, -0.00000436, -0.00000299, -0.00000078, 0.00000429, 0.00000422, -0.00000276, -0.00000257, 0.00000918, 0.00001050, -0.00001160, -0.00002840, -0.00000895, 0.00002298, 0.00001912, -0.00001263, -0.00001555, 0.00002049, 0.00003483, -0.00000505, -0.00004296, -0.00002060, 0.00002617, 0.00002326, -0.00002254, -0.00003631, 0.00000408, 0.00003656, 0.00001628, -0.00002067, -0.00002314, 0.00000050, 0.00001150, 0.00000376, -0.00000295, -0.00000436, -0.00000751, -0.00000498, 0.00001078, 0.00002198, 0.00000843, -0.00001539, -0.00001888, -0.00000253, 0.00000587, -0.00000280, -0.00000965, -0.00000411, 0.00000166, -0.00000075, -0.00000296, -0.00000036, -0.00000064, -0.00000454, -0.00000097, 0.00000978, 0.00001064, -0.00000398, -0.00001562, -0.00001010, 0.00000119, 0.00000236, -0.00000233, -0.00000150, 0.00000133, -0.00000305, -0.00001036, -0.00001068, -0.00000649, -0.00000569, -0.00000502, 0.00000172, 0.00000641, -0.00000304, -0.00001927, -0.00002254, -0.00001013, 0.00000019, -0.00000233, -0.00000859, -0.00000850, -0.00000586, -0.00000729, },
		{ -0.00000779, -0.00000659, -0.00000760, -0.00001020, -0.00001654, -0.00002750, -0.00003664, -0.00003328, -0.00001180, 0.00002136, 0.00004844, 0.00005213, 0.00003239, 0.00000941, 0.00000301, 0.00001156, 0.00001831, 0.00001464, 0.00000693, 0.00000288, 0.00000325, 0.00000762, 0.00001562, 0.00002059, 0.00001463, 0.00000335, 0.00000305, 0.00001794, 0.00003185, 0.00002922, 0.00001382, -0.00000002, -0.00000464, -0.00000102, 0.00000625, 0.00000855, -0.00000207, -0.00001972, -0.00002621, -0.00001406, 0.00000230, 0.00000699, 0.00000272, 0.00000013, -0.00000064, -0.00000435, -0.00000770, -0.00000639, -0.00000514, -0.00000702, -0.00000428, 0.00000526, 0.00000749, -0.00000361, -0.00000804, 0.00000778, 0.00001902, -0.00000150, -0.00003001, -0.00002083, 0.00001650, 0.00002492, -0.00000753, -0.00002341, 0.00000968, 0.00004011, 0.00001389, -0.00003387, -0.00003241, 0.00001166, 0.00002693, -0.00000773, -0.00003423, -0.00001189, 0.00002327, 0.00002293, -0.00000289, -0.00001546, -0.00000852, -0.00000162, 0.00000053, 0.00000480, 0.00000494, -0.00000604, -0.00001287, 0.00000084, 0.00001968, 0.00001601, -0.00000607, -0.00001714, -0.00000812, 0.00000128, -0.00000118, -0.00000529, -0.00000314, -0.00000178, -0.00000458, -0.00000351, 0.00000247, 0.00000259, -0.00000435, -0.00000460, 0.00000566, 0.00001035, -0.00000021, -0.00001215, -0.00001030, -0.00000116, 0.00000139, -0.00000130, -0.00000044, 0.00000155, -0.00000263, -0.00000922, -0.00000976, -0.00000660, -0.00000665, -0.00000652, 0.00000022, 0.00000625, -0.00000100, -0.00001676, -0.00002232, -0.00001201, -0.00000061, -0.00000070, -0.00000721, -0.00000975, },
		{ -0.00000569, -0.00001040, -0.00000943, -0.00000617, -0.00000535, -0.00000932, -0.00001742, -0.00002599, -0.00003037, -0.00002677, -0.00001219, 0.00001309, 0.00003901, 0.00004858, 0.00003529, 0.00001377, 0.00000386, 0.00000772, 0.00001248, 0.00001126, 0.00000866, 0.00000740, 0.00000458, 0.00000243, 0.00000791, 0.00001793, 0.00001915, 0.00000844, 0.00000046, 0.00000718, 0.00002144, 0.00002825, 0.00002281, 0.00001008, -0.00000330, -0.00000930, -0.00000189, 0.00001144, 0.00001234, -0.00000485, -0.00002255, -0.00002272, -0.00000955, 0.00000145, 0.00000535, 0.00000587, 0.00000274, -0.00000412, -0.00000748, -0.00000420, -0.00000359, -0.00000974, -0.00000943, 0.00000400, 0.00001147, -0.00000173, -0.00001365, 0.00000184, 0.00002392, 0.00001135, -0.00002513, -0.00003042, 0.00000611, 0.00002757, 0.00000056, -0.00002719, -0.00000440, 0.00003649, 0.00002961, -0.00001719, -0.00003506, -0.00000370, 0.00002186, 0.00000343, -0.00002503, -0.00001953, 0.00000770, 0.00001855, 0.00000821, -0.00000232, -0.00000655, -0.00001054, -0.00000907, 0.00000348, 0.00001227, 0.00000202, -0.00001300, -0.00000814, 0.00001102, 0.00001671, 0.00000251, -0.00001027, -0.00000887, -0.00000360, -0.00000352, -0.00000331, -0.00000031, -0.00000166, -0.00000743, -0.00000666, 0.00000209, 0.00000549, -0.00000135, -0.00000529, 0.00000161, 0.00000744, 0.00000105, -0.00000873, -0.00000864, -0.00000207, 0.00000011, -0.00000147, 0.00000001, 0.00000210, -0.00000175, -0.00000809, -0.00000905, -0.00000647, -0.00000688, -0.00000741, -0.00000148, 0.00000517, 0.00000021, -0.00001420, -0.00002136, -0.00001343, -0.00000173, 0.00000044, },
		{ -0.00000291, 0.00000127, -0.00000412, -0.00001063, -0.00001088, -0.00000596, -0.00000319, -0.00000826, -0.00001861, -0.00002591, -0.00002561, -0.00001994, -0.00000987, 0.00000788, 0.00003012, 0.00004226, 0.00003453, 0.00001713, 0.00000766, 0.00000824, 0.00000818, 0.00000482, 0.00000567, 0.00001058, 0.00000972, 0.00000239, 0.00000110, 0.00001079, 0.00001884, 0.00001435, 0.00000440, 0.00000221, 0.00000957, 0.00001927, 0.00002503, 0.00002168, 0.00000650, -0.00001110, -0.00001316, 0.00000337, 0.00001771, 0.00001109, -0.00000895, -0.00002160, -0.00001956, -0.00000981, 0.00000145, 0.00000980, 0.00000841, -0.00000192, -0.00000766, -0.00000257, -0.00000041, -0.00001022, -0.00001502, 0.00000007, 0.00001477, 0.00000304, -0.00001722, -0.00000740, 0.00002316, 0.00002367, -0.00001423, -0.00003492, -0.00000619, 0.00002553, 0.00000949, -0.00002529, -0.00001750, 0.00002494, 0.00003723, 0.00000173, -0.00002768, -0.00001389, 0.00001136, 0.00000718, -0.00001440, -0.00001850, -0.00000353, 0.00000797, 0.00000943, 0.00000842, 0.00000281, -0.00001063, -0.00001755, -0.00000481, 0.00001245, 0.00001019, -0.00000629, -0.00001094, 0.00000156, 0.00001082, 0.00000568, -0.00000264, -0.00000458, -0.00000502, -0.00000706, -0.00000475, 0.00000127, 0.00000063, -0.00000745, -0.00000964, -0.00000095, 0.00000586, 0.00000197, -0.00000322, -0.00000019, 0.00000400, -0.00000016, -0.00000709, -0.00000660, -0.00000141, -0.00000027, -0.00000206, -0.00000044, 0.00000221, -0.00000069, -0.00000667, -0.00000825, -0.00000631, -0.00000680, -0.00000773, -0.00000286, 0.00000372, 0.00000067, -0.00001201, -0.00002011, -0.00001444, },
		{ -0.00001880, -0.00001525, -0.00000422, 0.00000181, -0.00000246, -0.00001038, -0.00001209, -0.00000601, -0.00000122, -0.00000682, -0.00001949, -0.00002669, -0.00002280, -0.00001435, -0.00000635, 0.00000597, 0.00002406, 0.00003570, 0.00003032, 0.00001677, 0.00001112, 0.00001260, 0.00000846, -0.00000094, -0.00000153, 0.00000894, 0.00001474, 0.00000736, -0.00000084, 0.00000295, 0.00001291, 0.00001610, 0.00001102, 0.00000487, 0.00000246, 0.00000650, 0.00001788, 0.00002757, 0.00002035, -0.00000292, -0.00001888, -0.00001008, 0.00001101, 0.00001915, 0.00000788, -0.00000964, -0.00002098, -0.00002119, -0.00000890, 0.00000825, 0.00001354, 0.00000240, -0.00000700, -0.00000145, 0.00000367, -0.00000817, -0.00001979, -0.00000636, 0.00001592, 0.00000992, -0.00001696, -0.00001749, 0.00001621, 0.00003194, 0.00000042, -0.00003253, -0.00001733, 0.00001866, 0.00001630, -0.00001834, -0.00002561, 0.00000941, 0.00003486, 0.00001652, -0.00001362, -0.00001525, 0.00000121, 0.00000393, -0.00000763, -0.00001263, -0.00000738, -0.00000222, 0.00000294, 0.00001163, 0.00001276, -0.00000296, -0.00001928, -0.00001402, 0.00000558, 0.00001289, 0.00000220, -0.00000690, -0.00000337, 0.00000264, 0.00000253, 0.00000117, 0.00000151, -0.00000195, -0.00000836, -0.00000783, 0.00000006, 0.00000246, -0.00000522, -0.00001043, -0.00000443, 0.00000362, 0.00000342, -0.00000028, 0.00000056, 0.00000206, -0.00000229, -0.00000754, -0.00000571, -0.00000023, 0.00000054, -0.00000206, -0.00000126, 0.00000160, -0.00000002, -0.00000513, -0.00000716, -0.00000599, -0.00000657, -0.00000771, -0.00000386, 0.00000227, 0.00000069, -0.00001021, },
		{ 0.00000046, -0.00000863, -0.00001736, -0.00001585, -0.00000574, 0.00000181, -0.00000094, -0.00000963, -0.00001278, -0.00000615, 0.00000035, -0.00000531, -0.00001983, -0.00002746, -0.00002125, -0.00001066, -0.00000359, 0.00000577, 0.00002148, 0.00003174, 0.00002524, 0.00001228, 0.00001081, 0.00001750, 0.00001360, -0.00000216, -0.00000929, 0.00000200, 0.00001530, 0.00001369, 0.00000316, -0.00000078, 0.00000459, 0.00001145, 0.00001462, 0.00001217, 0.00000365, -0.00000311, 0.00000448, 0.00002327, 0.00003028, 0.00001200, -0.00001370, -0.00001965, -0.00000347, 0.00001453, 0.00001859, 0.00000781, -0.00001115, -0.00002577, -0.00002131, 0.00000023, 0.00001470, 0.00000695, -0.00000523, -0.00000041, 0.00000833, -0.00000342, -0.00002251, -0.00001459, 0.00001361, 0.00001721, -0.00001202, -0.00002535, 0.00000443, 0.00003347, 0.00001506, -0.00002341, -0.00002400, 0.00000874, 0.00001859, -0.00000912, -0.00002686, -0.00000481, 0.00002453, 0.00002272, 0.00000107, -0.00000836, -0.00000343, -0.00000222, -0.00000678, -0.00000716, -0.00000525, -0.00000740, -0.00000582, 0.00000714, 0.00001743, 0.00000724, -0.00001339, -0.00001828, -0.00000379, 0.00000873, 0.00000693, 0.00000027, -0.00000195, -0.00000273, -0.00000390, -0.00000061, 0.00000518, 0.00000345, -0.00000583, -0.00000958, -0.00000296, 0.00000210, -0.00000283, -0.00000898, -0.00000634, 0.00000042, 0.00000247, 0.00000149, 0.00000252, 0.00000227, -0.00000353, -0.00000899, -0.00000645, 0.00000019, 0.00000184, -0.00000113, -0.00000166, 0.00000062, -0.00000000, -0.00000391, -0.00000595, -0.00000541, -0.00000613, -0.00000743, -0.00000455, 0.00000091, },
		{ -0.00000487, -0.00000036, -0.00000005, -0.00000730, -0.00001572, -0.00001601, -0.00000727, 0.00000118, -0.00000002, -0.00000876, -0.00001283, -0.00000591, 0.00000175, -0.00000409, -0.00002006, -0.00002785, -0.00001984, -0.00000819, -0.00000257, 0.00000494, 0.00002086, 0.00003159, 0.00002269, 0.00000616, 0.00000558, 0.00001883, 0.00002053, 0.00000244, -0.00001303, -0.00000702, 0.00000975, 0.00001642, 0.00001009, 0.00000196, -0.00000082, 0.00000270, 0.00001149, 0.00001804, 0.00001143, -0.00000433, -0.00000791, 0.00001039, 0.00002997, 0.00002545, 0.00000077, -0.00001816, -0.00001638, -0.00000018, 0.00001670, 0.00002153, 0.00000626, -0.00001953, -0.00002922, -0.00001122, 0.00001034, 0.00000891, -0.00000366, 0.00000032, 0.00001323, 0.00000411, -0.00002176, -0.00002315, 0.00000712, 0.00002248, -0.00000312, -0.00002820, -0.00000900, 0.00002747, 0.00002553, -0.00000996, -0.00002402, -0.00000090, 0.00001584, -0.00000131, -0.00002236, -0.00001358, 0.00001114, 0.00001972, 0.00001054, 0.00000239, -0.00000074, -0.00000610, -0.00000993, -0.00000555, -0.00000131, -0.00000688, -0.00001175, -0.00000122, 0.00001490, 0.00001408, -0.00000354, -0.00001534, -0.00001002, 0.00000082, 0.00000546, 0.00000544, 0.00000345, -0.00000265, -0.00000901, -0.00000613, 0.00000411, 0.00000751, -0.00000074, -0.00000815, -0.00000538, -0.00000015, -0.00000204, -0.00000684, -0.00000620, -0.00000189, 0.00000023, 0.00000126, 0.00000390, 0.00000375, -0.00000307, -0.00000993, -0.00000813, -0.00000070, 0.00000261, 0.00000027, -0.00000132, -0.00000013, -0.00000038, -0.00000320, -0.00000496, -0.00000475, -0.00000547, -0.00000681, },
		{ -0.00000482, -0.00000589, -0.00000469, -0.00000133, -0.00000079, -0.00000641, -0.00001405, -0.00001558, -0.00000833, 0.00000025, 0.00000007, -0.00000841, -0.00001264, -0.00000496, 0.00000362, -0.00000300, -0.00002073, -0.00002836, -0.00001805, -0.00000558, -0.00000269, 0.00000209, 0.00001950, 0.00003390, 0.00002456, 0.00000228, -0.00000246, 0.00001450, 0.00002460, 0.00001029, -0.00001037, -0.00001303, 0.00000060, 0.00001266, 0.00001464, 0.00000916, 0.00000038, -0.00000467, 0.00000264, 0.00001736, 0.00001978, 0.00000276, -0.00001256, -0.00000411, 0.00001909, 0.00002945, 0.00001669, -0.00000517, -0.00001893, -0.00001621, 0.00000272, 0.00002370, 0.00002286, -0.00000427, -0.00002782, -0.00002040, 0.00000241, 0.00000686, -0.00000437, -0.00000070, 0.00001725, 0.00001398, -0.00001619, -0.00002968, -0.00000304, 0.00002325, 0.00000733, -0.00002473, -0.00002002, 0.00001573, 0.00002877, 0.00000362, -0.00001753, -0.00000672, 0.00000995, 0.00000249, -0.00001549, -0.00001565, -0.00000001, 0.00001077, 0.00001196, 0.00001121, 0.00000696, -0.00000453, -0.00001288, -0.00000773, 0.00000060, -0.00000337, -0.00001270, -0.00000855, 0.00000742, 0.00001447, 0.00000483, -0.00000739, -0.00000991, -0.00000574, -0.00000048, 0.00000544, 0.00000833, 0.00000170, -0.00000970, -0.00001146, -0.00000081, 0.00000768, 0.00000388, -0.00000408, -0.00000531, -0.00000243, -0.00000300, -0.00000564, -0.00000504, -0.00000253, -0.00000170, -0.00000032, 0.00000375, 0.00000503, -0.00000144, -0.00000954, -0.00000952, -0.00000244, 0.00000233, 0.00000134, -0.00000055, -0.00000037, -0.00000069, -0.00000282, -0.00000433, -0.00000431, },
		{ -0.00000387, -0.00000419, -0.00000448, -0.00000496, -0.00000403, -0.00000174, -0.00000149, -0.00000601, -0.00001273, -0.00001480, -0.00000863, -0.00000032, -0.00000034, -0.00000906, -0.00001305, -0.00000367, 0.00000646, -0.00000127, -0.00002178, -0.00002977, -0.00001637, -0.00000203, -0.00000232, -0.00000235, 0.00001547, 0.00003569, 0.00002990, 0.00000334, -0.00000909, 0.00000600, 0.00002269, 0.00001648, -0.00000303, -0.00001252, -0.00000689, 0.00000389, 0.00001274, 0.00001551, 0.00000745, -0.00000581, -0.00000675, 0.00000958, 0.00002244, 0.00001278, -0.00000759, -0.00001224, 0.00000372, 0.00002142, 0.00002479, 0.00001255, -0.00000834, -0.00002364, -0.00001585, 0.00001263, 0.00003012, 0.00001299, -0.00001726, -0.00002244, -0.00000447, 0.00000229, -0.00000825, -0.00000532, 0.00001808, 0.00002439, -0.00000534, -0.00003131, -0.00001459, 0.00001817, 0.00001573, -0.00001599, -0.00002521, 0.00000216, 0.00002425, 0.00001291, -0.00000732, -0.00000673, 0.00000436, 0.00000208, -0.00000992, -0.00001297, -0.00000574, 0.00000100, 0.00000653, 0.00001400, 0.00001488, 0.00000199, -0.00001203, -0.00001079, -0.00000070, -0.00000045, -0.00000998, -0.00001180, -0.00000069, 0.00000924, 0.00000787, 0.00000086, -0.00000428, -0.00000717, -0.00000655, 0.00000081, 0.00000919, 0.00000664, -0.00000616, -0.00001339, -0.00000645, 0.00000404, 0.00000548, 0.00000026, -0.00000275, -0.00000296, -0.00000440, -0.00000584, -0.00000423, -0.00000208, -0.00000242, -0.00000197, 0.00000233, 0.00000514, 0.00000022, -0.00000806, -0.00000986, -0.00000408, 0.00000132, 0.00000166, 0.00000005, -0.00000034, -0.00000073, -0.00000239, },
		{ -0.00000073, -0.00000180, -0.00000325, -0.00000415, -0.00000454, -0.00000440, -0.00000323, -0.00000156, -0.00000173, -0.00000580, -0.00001199, -0.00001424, -0.00000839, 0.00000001, -0.00000040, -0.00001041, -0.00001476, -0.00000304, 0.00001007, 0.00000190, -0.00002237, -0.00003224, -0.00001585, 0.00000205, -0.00000026, -0.00000652, 0.00000885, 0.00003436, 0.00003556, 0.00000905, -0.00001073, -0.00000234, 0.00001541, 0.00001686, 0.00000407, -0.00000610, -0.00000813, -0.00000479, 0.00000458, 0.00001590, 0.00001545, 0.00000006, -0.00001084, -0.00000085, 0.00001707, 0.00001857, 0.00000274, -0.00001004, -0.00000734, 0.00000593, 0.00002013, 0.00002502, 0.00001039, -0.00001733, -0.00002874, -0.00000609, 0.00002444, 0.00002387, -0.00000296, -0.00001652, -0.00000586, -0.00000084, -0.00001353, -0.00001403, 0.00001319, 0.00003193, 0.00000928, -0.00002581, -0.00002374, 0.00000815, 0.00001879, -0.00000533, -0.00002342, -0.00000860, 0.00001439, 0.00001522, 0.00000225, -0.00000166, 0.00000223, -0.00000031, -0.00000760, -0.00000903, -0.00000617, -0.00000539, -0.00000164, 0.00001025, 0.00001856, 0.00000991, -0.00000667, -0.00001134, -0.00000352, -0.00000003, -0.00000655, -0.00001106, -0.00000592, 0.00000187, 0.00000516, 0.00000521, 0.00000306, -0.00000313, -0.00000879, -0.00000502, 0.00000560, 0.00000863, -0.00000123, -0.00001132, -0.00000966, -0.00000098, 0.00000358, 0.00000246, 0.00000055, -0.00000138, -0.00000472, -0.00000662, -0.00000441, -0.00000162, -0.00000215, -0.00000280, 0.00000063, 0.00000413, 0.00000102, -0.00000639, -0.00000927, -0.00000498, 0.00000028, 0.00000147, 0.00000018, -0.00000049, },
		{ 0.00000008, -0.00000098, -0.00000106, -0.00000133, -0.00000244, -0.00000386, -0.00000471, -0.00000431, -0.00000274, -0.00000110, -0.00000135, -0.00000526, -0.00001158, -0.00001432, -0.00000830, 0.00000110, 0.00000067, -0.00001150, -0.00001768, -0.00000408, 0.00001333, 0.00000654, -0.00002135, -0.00003496, -0.00001705, 0.00000535, 0.00000334, -0.00000870, 0.00000158, 0.00002944, 0.00003828, 0.00001632, -0.00000681, -0.00000616, 0.00000689, 0.00001095, 0.00000590, 0.00000160, -0.00000249, -0.00000790, -0.00000522, 0.00000909, 0.00001851, 0.00000879, -0.00000727, -0.00000748, 0.00000693, 0.00001599, 0.00001087, -0.00000010, -0.00000807, -0.00000805, 0.00000514, 0.00002507, 0.00002741, 0.00000007, -0.00002886, -0.00002273, 0.00000925, 0.00002353, 0.00000747, -0.00000676, -0.00000075, 0.00000144, -0.00001613, -0.00002453, 0.00000168, 0.00003265, 0.00002369, -0.00001329, -0.00002650, -0.00000335, 0.00001530, 0.00000289, -0.00001648, -0.00001333, 0.00000367, 0.00001087, 0.00000725, 0.00000528, 0.00000458, -0.00000151, -0.00000787, -0.00000672, -0.00000388, -0.00000700, -0.00000817, 0.00000276, 0.00001633, 0.00001500, 0.00000083, -0.00000794, -0.00000495, -0.00000145, -0.00000470, -0.00000852, -0.00000734, -0.00000403, -0.00000080, 0.00000421, 0.00000795, 0.00000359, -0.00000595, -0.00000820, 0.00000007, 0.00000657, 0.00000181, -0.00000726, -0.00000935, -0.00000456, -0.00000010, 0.00000172, 0.00000249, 0.00000109, -0.00000341, -0.00000675, -0.00000510, -0.00000177, -0.00000167, -0.00000278, -0.00000058, 0.00000268, 0.00000089, -0.00000525, -0.00000840, -0.00000521, -0.00000027, 0.00000133, },
		{ -0.00000045, 0.00000157, 0.00000024, -0.00000151, -0.00000180, -0.00000136, -0.00000177, -0.00000330, -0.00000462, -0.00000440, -0.00000274, -0.00000089, -0.00000063, -0.00000407, -0.00001091, -0.00001491, -0.00000899, 0.00000220, 0.00000295, -0.00001125, -0.00002075, -0.00000697, 0.00001492, 0.00001157, -0.00001834, -0.00003651, -0.00001934, 0.00000687, 0.00000708, -0.00000860, -0.00000412, 0.00002270, 0.00003683, 0.00002133, -0.00000009, -0.00000388, 0.00000212, 0.00000239, 0.00000076, 0.00000488, 0.00000609, -0.00000340, -0.00001057, -0.00000119, 0.00001387, 0.00001406, 0.00000105, -0.00000632, -0.00000134, 0.00000667, 0.00001093, 0.00001024, 0.00000111, -0.00001268, -0.00001188, 0.00001255, 0.00003386, 0.00001979, -0.00001630, -0.00002908, -0.00000701, 0.00001359, 0.00000886, 0.00000015, 0.00000732, 0.00001000, -0.00001191, -0.00003193, -0.00001407, 0.00002425, 0.00003259, 0.00000294, -0.00002085, -0.00001148, 0.00000715, 0.00000551, -0.00000846, -0.00001179, -0.00000357, 0.00000308, 0.00000617, 0.00001004, 0.00000974, 0.00000054, -0.00000824, -0.00000665, -0.00000189, -0.00000531, -0.00001071, -0.00000440, 0.00000983, 0.00001492, 0.00000674, -0.00000197, -0.00000321, -0.00000244, -0.00000457, -0.00000646, -0.00000625, -0.00000675, -0.00000664, -0.00000067, 0.00000819, 0.00000912, 0.00000018, -0.00000688, -0.00000391, 0.00000208, 0.00000138, -0.00000415, -0.00000681, -0.00000531, -0.00000314, -0.00000086, 0.00000212, 0.00000274, -0.00000126, -0.00000568, -0.00000540, -0.00000232, -0.00000149, -0.00000243, -0.00000114, 0.00000144, 0.00000029, -0.00000478, -0.00000783, -0.00000526, },
		{ -0.00000762, -0.00000553, -0.00000069, 0.00000201, 0.00000089, -0.00000157, -0.00000252, -0.00000189, -0.00000167, -0.00000287, -0.00000423, -0.00000423, -0.00000296, -0.00000121, -0.00000013, -0.00000239, -0.00000946, -0.00001538, -0.00001055, 0.00000240, 0.00000566, -0.00000929, -0.00002263, -0.00001076, 0.00001425, 0.00001542, -0.00001418, -0.00003606, -0.00002120, 0.00000686, 0.00000952, -0.00000751, -0.00000737, 0.00001678, 0.00003265, 0.00002203, 0.00000522, 0.00000234, 0.00000380, -0.00000337, -0.00000849, 0.00000076, 0.00001148, 0.00000573, -0.00000801, -0.00000854, 0.00000402, 0.00001162, 0.00000773, 0.00000128, -0.00000227, -0.00000314, 0.00000252, 0.00001366, 0.00001384, -0.00000586, -0.00002185, -0.00000582, 0.00002693, 0.00003222, 0.00000202, -0.00002304, -0.00001596, 0.00000096, 0.00000169, -0.00000083, 0.00001210, 0.00002124, 0.00000016, -0.00003115, -0.00002844, 0.00000816, 0.00003171, 0.00001703, -0.00000838, -0.00001231, -0.00000117, 0.00000234, -0.00000358, -0.00000675, -0.00000526, -0.00000377, 0.00000062, 0.00001000, 0.00001428, 0.00000543, -0.00000638, -0.00000725, -0.00000163, -0.00000301, -0.00000974, -0.00000839, 0.00000257, 0.00001035, 0.00000840, 0.00000346, 0.00000112, -0.00000117, -0.00000460, -0.00000561, -0.00000465, -0.00000672, -0.00001014, -0.00000655, 0.00000428, 0.00001085, 0.00000624, -0.00000195, -0.00000413, -0.00000183, -0.00000164, -0.00000376, -0.00000445, -0.00000389, -0.00000406, -0.00000331, 0.00000008, 0.00000262, 0.00000039, -0.00000392, -0.00000487, -0.00000264, -0.00000154, -0.00000212, -0.00000135, 0.00000055, -0.00000034, -0.00000465, },
		{ -0.00000088, -0.00000458, -0.00000752, -0.00000601, -0.00000126, 0.00000218, 0.00000170, -0.00000100, -0.00000264, -0.00000240, -0.00000215, -0.00000307, -0.00000396, -0.00000371, -0.00000291, -0.00000184, -0.00000025, -0.00000083, -0.00000727, -0.00001503, -0.00001236, 0.00000136, 0.00000767, -0.00000628, -0.00002256, -0.00001387, 0.00001203, 0.00001706, -0.00001057, -0.00003411, -0.00002139, 0.00000677, 0.00001048, -0.00000717, -0.00000918, 0.00001315, 0.00002859, 0.00001929, 0.00000620, 0.00000807, 0.00001055, -0.00000256, -0.00001623, -0.00000898, 0.00000939, 0.00001326, 0.00000078, -0.00000815, -0.00000474, 0.00000243, 0.00000721, 0.00000941, 0.00000470, -0.00000680, -0.00000919, 0.00000750, 0.00002199, 0.00000780, -0.00002024, -0.00002070, 0.00001108, 0.00003235, 0.00001674, -0.00000947, -0.00001414, -0.00000643, -0.00000824, -0.00000966, 0.00000853, 0.00002865, 0.00001643, -0.00002010, -0.00003493, -0.00001007, 0.00002058, 0.00002306, 0.00000575, -0.00000522, -0.00000475, -0.00000347, -0.00000382, -0.00000240, -0.00000253, -0.00000648, -0.00000571, 0.00000534, 0.00001513, 0.00001058, -0.00000182, -0.00000630, -0.00000232, -0.00000194, -0.00000746, -0.00000888, -0.00000244, 0.00000412, 0.00000572, 0.00000565, 0.00000560, 0.00000236, -0.00000316, -0.00000512, -0.00000360, -0.00000545, -0.00001100, -0.00001103, -0.00000152, 0.00000846, 0.00000947, 0.00000387, -0.00000075, -0.00000285, -0.00000480, -0.00000577, -0.00000406, -0.00000222, -0.00000311, -0.00000425, -0.00000209, 0.00000111, 0.00000074, -0.00000248, -0.00000387, -0.00000247, -0.00000142, -0.00000177, -0.00000139, -0.00000010, },
		{ -0.00000112, -0.00000055, -0.00000142, -0.00000459, -0.00000737, -0.00000639, -0.00000201, 0.00000185, 0.00000212, -0.00000022, -0.00000199, -0.00000221, -0.00000263, -0.00000396, -0.00000441, -0.00000326, -0.00000239, -0.00000224, -0.00000086, 0.00000008, -0.00000490, -0.00001368, -0.00001357, -0.00000042, 0.00000832, -0.00000348, -0.00002094, -0.00001516, 0.00000992, 0.00001685, -0.00000893, -0.00003234, -0.00002001, 0.00000818, 0.00001122, -0.00000824, -0.00001148, 0.00001122, 0.00002673, 0.00001605, 0.00000305, 0.00000957, 0.00001797, 0.00000448, -0.00001762, -0.00001882, 0.00000032, 0.00001374, 0.00000984, -0.00000016, -0.00000640, -0.00000779, -0.00000136, 0.00001151, 0.00001473, -0.00000170, -0.00001696, -0.00000484, 0.00002073, 0.00002030, -0.00000926, -0.00002565, -0.00000514, 0.00002222, 0.00002159, 0.00000328, -0.00000468, -0.00000454, -0.00001348, -0.00002101, -0.00000365, 0.00002651, 0.00002976, -0.00000184, -0.00002967, -0.00002288, 0.00000354, 0.00001830, 0.00001492, 0.00000622, -0.00000113, -0.00000712, -0.00000772, -0.00000178, 0.00000118, -0.00000477, -0.00000923, -0.00000115, 0.00001148, 0.00001292, 0.00000365, -0.00000279, -0.00000201, -0.00000196, -0.00000585, -0.00000754, -0.00000427, -0.00000078, 0.00000087, 0.00000397, 0.00000770, 0.00000632, 0.00000005, -0.00000369, -0.00000280, -0.00000404, -0.00001012, -0.00001320, -0.00000691, 0.00000349, 0.00000891, 0.00000787, 0.00000419, -0.00000059, -0.00000591, -0.00000835, -0.00000573, -0.00000192, -0.00000173, -0.00000368, -0.00000323, -0.00000063, 0.00000002, -0.00000194, -0.00000314, -0.00000213, -0.00000104, -0.00000115, },
		{ -0.00000083, -0.00000035, -0.00000037, -0.00000052, -0.00000185, -0.00000481, -0.00000734, -0.00000658, -0.00000256, 0.00000129, 0.00000189, 0.00000008, -0.00000106, -0.00000106, -0.00000228, -0.00000493, -0.00000575, -0.00000360, -0.00000181, -0.00000208, -0.00000139, 0.00000035, -0.00000302, -0.00001181, -0.00001374, -0.00000201, 0.00000787, -0.00000188, -0.00001907, -0.00001468, 0.00000942, 0.00001641, -0.00000929, -0.00003246, -0.00001860, 0.00001140, 0.00001354, -0.00000971, -0.00001542, 0.00000896, 0.00002700, 0.00001498, -0.00000145, 0.00000615, 0.00002159, 0.00001368, -0.00001177, -0.00002304, -0.00001070, 0.00000613, 0.00001276, 0.00001023, 0.00000066, -0.00001218, -0.00001341, 0.00000478, 0.00002083, 0.00000908, -0.00001627, -0.00001642, 0.00001119, 0.00002524, 0.00000415, -0.00002045, -0.00001435, 0.00000889, 0.00001661, 0.00000852, 0.00000511, 0.00000454, -0.00000967, -0.00002742, -0.00001923, 0.00001378, 0.00003318, 0.00001619, -0.00001418, -0.00002442, -0.00001169, 0.00000511, 0.00001453, 0.00001574, 0.00000806, -0.00000498, -0.00001126, -0.00000495, 0.00000244, -0.00000105, -0.00000858, -0.00000594, 0.00000519, 0.00001096, 0.00000715, 0.00000215, 0.00000060, -0.00000146, -0.00000528, -0.00000633, -0.00000404, -0.00000304, -0.00000342, -0.00000003, 0.00000651, 0.00000853, 0.00000373, -0.00000096, -0.00000139, -0.00000259, -0.00000843, -0.00001336, -0.00001056, -0.00000203, 0.00000539, 0.00000870, 0.00000826, 0.00000357, -0.00000421, -0.00000953, -0.00000811, -0.00000328, -0.00000128, -0.00000267, -0.00000325, -0.00000170, -0.00000086, -0.00000205, -0.00000304, -0.00000220, },
		{ -0.00000320, -0.00000285, -0.00000133, 0.00000004, 0.00000064, 0.00000023, -0.00000173, -0.00000506, -0.00000765, -0.00000687, -0.00000278, 0.00000102, 0.00000135, -0.00000045, -0.00000070, 0.00000053, -0.00000070, -0.00000500, -0.00000738, -0.00000503, -0.00000196, -0.00000170, -0.00000140, 0.00000049, -0.00000179, -0.00001014, -0.00001319, -0.00000274, 0.00000724, -0.00000154, -0.00001821, -0.00001366, 0.00001085, 0.00001753, -0.00001017, -0.00003497, -0.00001918, 0.00001510, 0.00001826, -0.00000961, -0.00002036, 0.00000472, 0.00002742, 0.00001667, -0.00000405, 0.00000028, 0.00001966, 0.00001992, -0.00000216, -0.00001935, -0.00001727, -0.00000542, 0.00000682, 0.00001594, 0.00001245, -0.00000727, -0.00002175, -0.00000789, 0.00001843, 0.00001889, -0.00000805, -0.00002139, -0.00000089, 0.00002145, 0.00001301, -0.00001026, -0.00001436, -0.00000025, 0.00000728, 0.00000503, 0.00000859, 0.00001417, 0.00000167, -0.00002396, -0.00003027, -0.00000472, 0.00002415, 0.00002574, 0.00000469, -0.00001402, -0.00001779, -0.00000968, 0.00000475, 0.00001768, 0.00001758, 0.00000296, -0.00001042, -0.00000894, 0.00000011, 0.00000130, -0.00000520, -0.00000695, -0.00000047, 0.00000573, 0.00000670, 0.00000591, 0.00000494, 0.00000101, -0.00000447, -0.00000592, -0.00000356, -0.00000332, -0.00000562, -0.00000393, 0.00000309, 0.00000796, 0.00000605, 0.00000211, 0.00000088, -0.00000066, -0.00000622, -0.00001210, -0.00001224, -0.00000662, 0.00000052, 0.00000656, 0.00000990, 0.00000760, -0.00000056, -0.00000843, -0.00000958, -0.00000533, -0.00000206, -0.00000227, -0.00000298, -0.00000209, -0.00000127, -0.00000208, },
		{ -0.00000137, -0.00000160, -0.00000287, -0.00000347, -0.00000252, -0.00000059, 0.00000112, 0.00000139, -0.00000070, -0.00000475, -0.00000815, -0.00000763, -0.00000305, 0.00000125, 0.00000119, -0.00000145, -0.00000142, 0.00000153, 0.00000161, -0.00000356, -0.00000817, -0.00000703, -0.00000333, -0.00000192, -0.00000109, 0.00000107, -0.00000077, -0.00000902, -0.00001264, -0.00000270, 0.00000732, -0.00000160, -0.00001875, -0.00001361, 0.00001314, 0.00002094, -0.00000953, -0.00003860, -0.00002271, 0.00001705, 0.00002437, -0.00000660, -0.00002425, -0.00000137, 0.00002587, 0.00001936, -0.00000340, -0.00000440, 0.00001407, 0.00002035, 0.00000571, -0.00001028, -0.00001563, -0.00001414, -0.00000516, 0.00001257, 0.00002181, 0.00000470, -0.00002126, -0.00001980, 0.00000831, 0.00002196, 0.00000227, -0.00001853, -0.00000925, 0.00001288, 0.00001438, -0.00000176, -0.00000854, -0.00000187, 0.00000048, -0.00000286, 0.00000386, 0.00001762, 0.00001415, -0.00001146, -0.00003073, -0.00002053, 0.00000639, 0.00002227, 0.00001793, 0.00000296, -0.00001184, -0.00001814, -0.00000915, 0.00001029, 0.00002147, 0.00001282, -0.00000389, -0.00000978, -0.00000391, 0.00000041, -0.00000215, -0.00000472, -0.00000292, 0.00000011, 0.00000261, 0.00000623, 0.00000876, 0.00000536, -0.00000192, -0.00000537, -0.00000366, -0.00000319, -0.00000605, -0.00000621, -0.00000050, 0.00000539, 0.00000605, 0.00000400, 0.00000320, 0.00000177, -0.00000340, -0.00000970, -0.00001208, -0.00000957, -0.00000429, 0.00000251, 0.00000883, 0.00000995, 0.00000348, -0.00000548, -0.00000929, -0.00000681, -0.00000339, -0.00000270, -0.00000318, -0.00000249, },
		{ -0.00000388, -0.00000347, -0.00000190, -0.00000100, -0.00000179, -0.00000324, -0.00000361, -0.00000218, 0.00000031, 0.00000203, 0.00000096, -0.00000334, -0.00000814, -0.00000880, -0.00000393, 0.00000158, 0.00000177, -0.00000208, -0.00000291, 0.00000122, 0.00000350, -0.00000101, -0.00000730, -0.00000838, -0.00000553, -0.00000343, -0.00000129, 0.00000205, 0.00000068, -0.00000811, -0.00001258, -0.00000257, 0.00000824, -0.00000101, -0.00001998, -0.00001524, 0.00001459, 0.00002585, -0.00000614, -0.00004114, -0.00002833, 0.00001572, 0.00002954, -0.00000119, -0.00002525, -0.00000737, 0.00002182, 0.00002056, -0.00000081, -0.00000569, 0.00000870, 0.00001601, 0.00000785, -0.00000143, -0.00000702, -0.00001486, -0.00001654, 0.00000118, 0.00002283, 0.00001716, -0.00001208, -0.00002467, -0.00000404, 0.00001702, 0.00000857, -0.00001153, -0.00001108, 0.00000531, 0.00001038, 0.00000117, -0.00000273, 0.00000204, 0.00000008, -0.00000873, -0.00000550, 0.00001233, 0.00002056, 0.00000392, -0.00002021, -0.00002624, -0.00001166, 0.00000773, 0.00001914, 0.00001760, 0.00000282, -0.00001550, -0.00001939, -0.00000291, 0.00001659, 0.00001885, 0.00000573, -0.00000528, -0.00000598, -0.00000279, -0.00000177, -0.00000183, -0.00000193, -0.00000307, -0.00000272, 0.00000278, 0.00000966, 0.00000969, 0.00000262, -0.00000320, -0.00000366, -0.00000355, -0.00000608, -0.00000699, -0.00000281, 0.00000257, 0.00000438, 0.00000396, 0.00000432, 0.00000387, -0.00000025, -0.00000631, -0.00001018, -0.00001057, -0.00000805, -0.00000226, 0.00000558, 0.00001003, 0.00000655, -0.00000183, -0.00000743, -0.00000702, -0.00000428, -0.00000335, },
		{ -0.00000429, -0.00000348, -0.00000443, -0.00000483, -0.00000330, -0.00000119, -0.00000058, -0.00000187, -0.00000353, -0.00000373, -0.00000179, 0.00000112, 0.00000218, -0.00000098, -0.00000685, -0.00000963, -0.00000551, 0.00000127, 0.00000273, -0.00000176, -0.00000421, -0.00000022, 0.00000399, 0.00000143, -0.00000495, -0.00000803, -0.00000735, -0.00000604, -0.00000283, 0.00000263, 0.00000267, -0.00000674, -0.00001283, -0.00000298, 0.00000942, 0.00000063, -0.00002073, -0.00001811, 0.00001409, 0.00003045, -0.00000042, -0.00004086, -0.00003379, 0.00001145, 0.00003173, 0.00000442, -0.00002323, -0.00001104, 0.00001688, 0.00001903, 0.00000101, -0.00000422, 0.00000660, 0.00001075, 0.00000418, 0.00000214, 0.00000331, -0.00000716, -0.00002096, -0.00001238, 0.00001456, 0.00002327, 0.00000052, -0.00002066, -0.00001208, 0.00000783, 0.00000808, -0.00000598, -0.00000831, 0.00000234, 0.00000584, -0.00000070, -0.00000109, 0.00000643, 0.00000494, -0.00000836, -0.00001317, 0.00000129, 0.00001735, 0.00001428, -0.00000428, -0.00001971, -0.00002142, -0.00001029, 0.00000806, 0.00002214, 0.00001814, -0.00000297, -0.00001999, -0.00001487, 0.00000473, 0.00001696, 0.00001327, 0.00000317, -0.00000334, -0.00000519, -0.00000400, -0.00000086, 0.00000054, -0.00000284, -0.00000638, -0.00000258, 0.00000676, 0.00001156, 0.00000758, 0.00000111, -0.00000211, -0.00000387, -0.00000662, -0.00000747, -0.00000389, 0.00000082, 0.00000259, 0.00000261, 0.00000366, 0.00000450, 0.00000223, -0.00000248, -0.00000672, -0.00000936, -0.00001001, -0.00000669, 0.00000095, 0.00000787, 0.00000792, 0.00000149, -0.00000475, -0.00000609, },
		{ -0.00000196, -0.00000450, -0.00000362, -0.00000289, -0.00000417, -0.00000572, -0.00000518, -0.00000266, -0.00000038, -0.00000016, -0.00000187, -0.00000387, -0.00000406, -0.00000151, 0.00000167, 0.00000118, -0.00000423, -0.00000913, -0.00000706, -0.00000001, 0.00000317, -0.00000072, -0.00000447, -0.00000176, 0.00000303, 0.00000252, -0.00000234, -0.00000589, -0.00000748, -0.00000856, -0.00000577, 0.00000184, 0.00000451, -0.00000462, -0.00001271, -0.00000392, 0.00001010, 0.00000287, -0.00002030, -0.00002099, 0.00001174, 0.00003322, 0.00000588, -0.00003765, -0.00003691, 0.00000630, 0.00003063, 0.00000777, -0.00002009, -0.00001156, 0.00001357, 0.00001586, 0.00000011, -0.00000274, 0.00000801, 0.00000842, -0.00000186, -0.00000107, 0.00000944, 0.00000460, -0.00001615, -0.00002109, 0.00000139, 0.00002014, 0.00000959, -0.00001122, -0.00001246, 0.00000036, 0.00000280, -0.00000542, -0.00000554, 0.00000334, 0.00000449, -0.00000349, -0.00000368, 0.00000710, 0.00001051, -0.00000222, -0.00001436, -0.00000909, 0.00000666, 0.00001473, 0.00000873, -0.00000527, -0.00001883, -0.00002256, -0.00000915, 0.00001423, 0.00002570, 0.00001277, -0.00001018, -0.00001892, -0.00000810, 0.00000738, 0.00001404, 0.00001115, 0.00000372, -0.00000371, -0.00000636, -0.00000252, 0.00000184, -0.00000056, -0.00000685, -0.00000700, 0.00000139, 0.00000965, 0.00001055, 0.00000630, 0.00000173, -0.00000270, -0.00000720, -0.00000857, -0.00000491, 0.00000004, 0.00000187, 0.00000145, 0.00000198, 0.00000325, 0.00000286, 0.00000051, -0.00000248, -0.00000598, -0.00000948, -0.00000970, -0.00000412, 0.00000384, 0.00000727, 0.00000378, },
		{ 0.00000448, 0.00000452, 0.00000048, -0.00000259, -0.00000265, -0.00000198, -0.00000311, -0.00000543, -0.00000643, -0.00000488, -0.00000189, 0.00000043, 0.00000048, -0.00000196, -0.00000483, -0.00000470, -0.00000101, 0.00000158, -0.00000141, -0.00000696, -0.00000741, -0.00000165, 0.00000251, 0.00000006, -0.00000372, -0.00000233, 0.00000162, 0.00000192, -0.00000095, -0.00000309, -0.00000553, -0.00000950, -0.00000903, -0.00000063, 0.00000511, -0.00000229, -0.00001174, -0.00000476, 0.00001008, 0.00000491, -0.00001890, -0.00002290, 0.00000867, 0.00003383, 0.00001092, -0.00003302, -0.00003700, 0.00000276, 0.00002797, 0.00000792, -0.00001850, -0.00001034, 0.00001338, 0.00001366, -0.00000312, -0.00000386, 0.00001046, 0.00001034, -0.00000577, -0.00000841, 0.00000835, 0.00001413, -0.00000501, -0.00002096, -0.00000973, 0.00001032, 0.00001081, -0.00000266, -0.00000684, -0.00000125, -0.00000228, -0.00000914, -0.00000643, 0.00000467, 0.00000646, -0.00000358, -0.00000723, 0.00000342, 0.00001236, 0.00000539, -0.00000874, -0.00001321, -0.00000523, 0.00000610, 0.00001254, 0.00000884, -0.00000646, -0.00002331, -0.00002335, -0.00000183, 0.00002132, 0.00002322, 0.00000498, -0.00001263, -0.00001502, -0.00000512, 0.00000694, 0.00001376, 0.00001165, 0.00000229, -0.00000574, -0.00000501, 0.00000069, 0.00000126, -0.00000483, -0.00000853, -0.00000377, 0.00000478, 0.00000985, 0.00001000, 0.00000690, 0.00000090, -0.00000635, -0.00000985, -0.00000679, -0.00000098, 0.00000200, 0.00000160, 0.00000086, 0.00000107, 0.00000128, 0.00000126, 0.00000094, -0.00000130, -0.00000618, -0.00001013, -0.00000835, -0.00000141, },
		{ -0.00001019, -0.00000662, -0.00000022, 0.00000317, 0.00000201, -0.00000052, -0.00000143, -0.00000120, -0.00000195, -0.00000411, -0.00000614, -0.00000645, -0.00000456, -0.00000112, 0.00000176, 0.00000116, -0.00000307, -0.00000646, -0.00000479, -0.00000058, -0.00000017, -0.00000411, -0.00000589, -0.00000234, 0.00000115, -0.00000027, -0.00000294, -0.00000169, 0.00000105, 0.00000051, -0.00000143, -0.00000128, -0.00000247, -0.00000815, -0.00001099, -0.00000390, 0.00000389, -0.00000088, -0.00001025, -0.00000486, 0.00000989, 0.00000633, -0.00001732, -0.00002375, 0.00000609, 0.00003320, 0.00001392, -0.00002901, -0.00003523, 0.00000208, 0.00002639, 0.00000611, -0.00002007, -0.00001017, 0.00001551, 0.00001457, -0.00000622, -0.00000808, 0.00001077, 0.00001469, -0.00000502, -0.00001519, 0.00000137, 0.00001700, 0.00000645, -0.00001309, -0.00001362, -0.00000009, 0.00000490, -0.00000000, -0.00000044, 0.00000271, -0.00000291, -0.00001315, -0.00001105, 0.00000282, 0.00000876, -0.00000035, -0.00000812, -0.00000172, 0.00000925, 0.00000964, -0.00000022, -0.00000962, -0.00001191, -0.00000591, 0.00000631, 0.00001524, 0.00000813, -0.00001319, -0.00002746, -0.00001755, 0.00000738, 0.00002284, 0.00001741, 0.00000067, -0.00001224, -0.00001382, -0.00000459, 0.00000878, 0.00001565, 0.00001020, -0.00000076, -0.00000537, -0.00000181, 0.00000094, -0.00000259, -0.00000727, -0.00000644, -0.00000059, 0.00000581, 0.00001023, 0.00001108, 0.00000626, -0.00000288, -0.00000976, -0.00000915, -0.00000327, 0.00000165, 0.00000273, 0.00000147, -0.00000025, -0.00000136, -0.00000065, 0.00000181, 0.00000278, -0.00000090, -0.00000726, },
		{ 0.00000425, -0.00000161, -0.00000835, -0.00000992, -0.00000579, -0.00000057, 0.00000177, 0.00000133, 0.00000028, -0.00000043, -0.00000132, -0.00000273, -0.00000445, -0.00000615, -0.00000674, -0.00000437, 0.00000053, 0.00000346, 0.00000055, -0.00000534, -0.00000747, -0.00000446, -0.00000173, -0.00000252, -0.00000324, -0.00000110, 0.00000043, -0.00000159, -0.00000321, -0.00000069, 0.00000188, -0.00000024, -0.00000310, -0.00000148, -0.00000007, -0.00000521, -0.00001058, -0.00000629, 0.00000148, -0.00000122, -0.00000934, -0.00000432, 0.00001026, 0.00000748, -0.00001614, -0.00002422, 0.00000438, 0.00003265, 0.00001551, -0.00002682, -0.00003371, 0.00000352, 0.00002765, 0.00000502, -0.00002411, -0.00001324, 0.00001729, 0.00001858, -0.00000650, -0.00001340, 0.00000745, 0.00001808, -0.00000040, -0.00001778, -0.00000717, 0.00001297, 0.00001291, -0.00000252, -0.00000961, -0.00000537, -0.00000328, -0.00000394, 0.00000167, 0.00000812, 0.00000150, -0.00001334, -0.00001612, -0.00000277, 0.00000793, 0.00000359, -0.00000526, -0.00000452, 0.00000362, 0.00000845, 0.00000599, -0.00000150, -0.00001058, -0.00001433, -0.00000527, 0.00001137, 0.00001690, 0.00000144, -0.00002024, -0.00002504, -0.00000873, 0.00001208, 0.00002062, 0.00001374, -0.00000138, -0.00001395, -0.00001438, -0.00000171, 0.00001266, 0.00001554, 0.00000691, -0.00000169, -0.00000299, -0.00000109, -0.00000207, -0.00000514, -0.00000619, -0.00000399, 0.00000059, 0.00000685, 0.00001204, 0.00001114, 0.00000282, -0.00000685, -0.00001030, -0.00000652, -0.00000046, 0.00000327, 0.00000343, 0.00000067, -0.00000295, -0.00000398, -0.00000057, 0.00000409, },
		{ -0.00000492, 0.00000154, 0.00000660, 0.00000464, -0.00000292, -0.00000944, -0.00001016, -0.00000603, -0.00000106, 0.00000193, 0.00000240, 0.00000096, -0.00000110, -0.00000226, -0.00000257, -0.00000395, -0.00000678, -0.00000748, -0.00000305, 0.00000304, 0.00000385, -0.00000166, -0.00000710, -0.00000771, -0.00000559, -0.00000381, -0.00000145, 0.00000168, 0.00000163, -0.00000261, -0.00000473, -0.00000062, 0.00000342, 0.00000055, -0.00000442, -0.00000328, 0.00000034, -0.00000246, -0.00000819, -0.00000642, -0.00000049, -0.00000300, -0.00001004, -0.00000415, 0.00001128, 0.00000912, -0.00001518, -0.00002500, 0.00000302, 0.00003282, 0.00001703, -0.00002614, -0.00003402, 0.00000498, 0.00003155, 0.00000695, -0.00002805, -0.00001961, 0.00001599, 0.00002341, -0.00000302, -0.00001685, 0.00000165, 0.00001806, 0.00000498, -0.00001566, -0.00001294, 0.00000542, 0.00001257, 0.00000514, -0.00000143, -0.00000354, -0.00000812, -0.00001086, -0.00000214, 0.00001014, 0.00000770, -0.00000856, -0.00001762, -0.00000930, 0.00000295, 0.00000471, -0.00000081, -0.00000323, -0.00000056, 0.00000356, 0.00000680, 0.00000580, -0.00000334, -0.00001509, -0.00001516, 0.00000055, 0.00001577, 0.00001239, -0.00000657, -0.00002135, -0.00001881, -0.00000305, 0.00001330, 0.00001956, 0.00001145, -0.00000552, -0.00001703, -0.00001229, 0.00000350, 0.00001471, 0.00001328, 0.00000520, -0.00000063, -0.00000257, -0.00000344, -0.00000423, -0.00000451, -0.00000448, -0.00000319, 0.00000168, 0.00000916, 0.00001307, 0.00000861, -0.00000117, -0.00000847, -0.00000891, -0.00000416, 0.00000161, 0.00000486, 0.00000347, -0.00000177, -0.00000621, },
		{ 0.00000202, -0.00000517, -0.00000840, -0.00000379, 0.00000450, 0.00000837, 0.00000405, -0.00000465, -0.00001096, -0.00001120, -0.00000636, -0.00000002, 0.00000393, 0.00000330, -0.00000035, -0.00000271, -0.00000187, -0.00000122, -0.00000434, -0.00000837, -0.00000708, -0.00000038, 0.00000443, 0.00000237, -0.00000352, -0.00000793, -0.00000942, -0.00000794, -0.00000252, 0.00000409, 0.00000476, -0.00000182, -0.00000638, -0.00000201, 0.00000429, 0.00000245, -0.00000418, -0.00000514, -0.00000121, -0.00000156, -0.00000549, -0.00000428, -0.00000038, -0.00000469, -0.00001235, -0.00000549, 0.00001212, 0.00001155, -0.00001371, -0.00002610, 0.00000122, 0.00003338, 0.00001942, -0.00002563, -0.00003624, 0.00000448, 0.00003609, 0.00001233, -0.00002912, -0.00002706, 0.00001066, 0.00002598, 0.00000263, -0.00001659, -0.00000377, 0.00001471, 0.00000806, -0.00001118, -0.00001419, -0.00000119, 0.00000761, 0.00000698, 0.00000561, 0.00000301, -0.00000679, -0.00001557, -0.00000927, 0.00000657, 0.00001106, -0.00000143, -0.00001393, -0.00001275, -0.00000384, 0.00000139, 0.00000161, 0.00000060, -0.00000079, -0.00000112, 0.00000300, 0.00000830, 0.00000472, -0.00000896, -0.00001821, -0.00001056, 0.00000649, 0.00001429, 0.00000565, -0.00000972, -0.00001846, -0.00001449, -0.00000010, 0.00001531, 0.00001915, 0.00000667, -0.00001114, -0.00001734, -0.00000738, 0.00000741, 0.00001449, 0.00001196, 0.00000524, -0.00000109, -0.00000488, -0.00000521, -0.00000354, -0.00000311, -0.00000416, -0.00000259, 0.00000392, 0.00001097, 0.00001180, 0.00000530, -0.00000332, -0.00000829, -0.00000754, -0.00000236, 0.00000373, 0.00000605, },
		{ -0.00000048, 0.00000584, 0.00000614, -0.00000064, -0.00000822, -0.00000879, -0.00000133, 0.00000722, 0.00000909, 0.00000272, -0.00000693, -0.00001327, -0.00001236, -0.00000503, 0.00000320, 0.00000582, 0.00000176, -0.00000296, -0.00000268, 0.00000023, -0.00000082, -0.00000624, -0.00000904, -0.00000501, 0.00000145, 0.00000390, 0.00000119, -0.00000438, -0.00001040, -0.00001274, -0.00000692, 0.00000379, 0.00000809, 0.00000124, -0.00000645, -0.00000402, 0.00000359, 0.00000405, -0.00000246, -0.00000556, -0.00000329, -0.00000286, -0.00000441, -0.00000139, 0.00000220, -0.00000450, -0.00001494, -0.00000868, 0.00001152, 0.00001411, -0.00001123, -0.00002682, -0.00000121, 0.00003346, 0.00002253, -0.00002409, -0.00003912, 0.00000142, 0.00003892, 0.00001950, -0.00002612, -0.00003251, 0.00000287, 0.00002450, 0.00000743, -0.00001327, -0.00000637, 0.00001032, 0.00000789, -0.00000751, -0.00001223, -0.00000424, 0.00000205, 0.00000394, 0.00000799, 0.00000956, -0.00000063, -0.00001478, -0.00001495, -0.00000079, 0.00000891, 0.00000357, -0.00000703, -0.00001088, -0.00000843, -0.00000467, -0.00000028, 0.00000346, 0.00000223, -0.00000243, -0.00000187, 0.00000566, 0.00000894, -0.00000042, -0.00001380, -0.00001602, -0.00000501, 0.00000731, 0.00001017, 0.00000265, -0.00000925, -0.00001677, -0.00001209, 0.00000382, 0.00001773, 0.00001586, -0.00000010, -0.00001429, -0.00001440, -0.00000294, 0.00000927, 0.00001464, 0.00001192, 0.00000394, -0.00000398, -0.00000675, -0.00000436, -0.00000196, -0.00000293, -0.00000421, -0.00000102, 0.00000593, 0.00001086, 0.00000965, 0.00000339, -0.00000384, -0.00000804, -0.00000672, },
		{ -0.00000428, -0.00000852, -0.00000599, 0.00000173, 0.00000746, 0.00000507, -0.00000360, -0.00000999, -0.00000766, 0.00000136, 0.00000908, 0.00000900, 0.00000076, -0.00001020, -0.00001600, -0.00001180, -0.00000097, 0.00000666, 0.00000488, -0.00000166, -0.00000386, -0.00000037, 0.00000160, -0.00000231, -0.00000755, -0.00000795, -0.00000358, 0.00000144, 0.00000384, 0.00000118, -0.00000714, -0.00001507, -0.00001279, -0.00000024, 0.00000906, 0.00000513, -0.00000408, -0.00000484, 0.00000184, 0.00000415, -0.00000058, -0.00000425, -0.00000418, -0.00000517, -0.00000571, 0.00000019, 0.00000601, -0.00000163, -0.00001588, -0.00001260, 0.00000886, 0.00001548, -0.00000820, -0.00002646, -0.00000366, 0.00003262, 0.00002550, -0.00002131, -0.00004116, -0.00000317, 0.00003883, 0.00002587, -0.00002010, -0.00003400, -0.00000434, 0.00001965, 0.00000896, -0.00000943, -0.00000583, 0.00000752, 0.00000593, -0.00000634, -0.00000989, -0.00000403, -0.00000103, -0.00000057, 0.00000566, 0.00001231, 0.00000628, -0.00000898, -0.00001558, -0.00000754, 0.00000232, 0.00000328, -0.00000122, -0.00000498, -0.00000799, -0.00000963, -0.00000558, 0.00000253, 0.00000534, -0.00000016, -0.00000425, 0.00000078, 0.00000788, 0.00000552, -0.00000552, -0.00001390, -0.00001239, -0.00000362, 0.00000561, 0.00000897, 0.00000265, -0.00000983, -0.00001651, -0.00000833, 0.00000841, 0.00001722, 0.00000995, -0.00000514, -0.00001414, -0.00001111, -0.00000001, 0.00001124, 0.00001562, 0.00001054, 0.00000037, -0.00000654, -0.00000613, -0.00000243, -0.00000141, -0.00000332, -0.00000353, 0.00000054, 0.00000641, 0.00000986, 0.00000860, 0.00000298, },
		{ 0.00000979, 0.00000904, 0.00000274, -0.00000570, -0.00000948, -0.00000477, 0.00000416, 0.00000830, 0.00000331, -0.00000589, -0.00001039, -0.00000607, 0.00000343, 0.00001036, 0.00000844, -0.00000244, -0.00001449, -0.00001720, -0.00000791, 0.00000398, 0.00000712, 0.00000138, -0.00000359, -0.00000203, 0.00000157, 0.00000093, -0.00000349, -0.00000720, -0.00000753, -0.00000394, 0.00000225, 0.00000536, -0.00000093, -0.00001289, -0.00001682, -0.00000658, 0.00000611, 0.00000720, -0.00000029, -0.00000322, 0.00000080, 0.00000288, -0.00000001, -0.00000233, -0.00000318, -0.00000663, -0.00000847, -0.00000071, 0.00000891, 0.00000298, -0.00001403, -0.00001528, 0.00000497, 0.00001475, -0.00000589, -0.00002521, -0.00000528, 0.00003129, 0.00002761, -0.00001806, -0.00004174, -0.00000765, 0.00003637, 0.00002964, -0.00001358, -0.00003177, -0.00000843, 0.00001405, 0.00000698, -0.00000777, -0.00000402, 0.00000735, 0.00000459, -0.00000704, -0.00000923, -0.00000282, -0.00000114, -0.00000339, 0.00000137, 0.00001051, 0.00001000, -0.00000172, -0.00001091, -0.00000976, -0.00000471, -0.00000203, -0.00000017, 0.00000100, -0.00000294, -0.00001016, -0.00001086, -0.00000217, 0.00000539, 0.00000315, -0.00000300, -0.00000264, 0.00000357, 0.00000641, 0.00000160, -0.00000668, -0.00001252, -0.00001215, -0.00000434, 0.00000653, 0.00001031, 0.00000140, -0.00001186, -0.00001468, -0.00000332, 0.00001054, 0.00001373, 0.00000483, -0.00000729, -0.00001310, -0.00000863, 0.00000313, 0.00001378, 0.00001516, 0.00000685, -0.00000308, -0.00000671, -0.00000415, -0.00000128, -0.00000165, -0.00000332, -0.00000278, 0.00000093, 0.00000608, },
		{ -0.00000290, 0.00000056, 0.00000654, 0.00001132, 0.00000988, 0.00000128, -0.00000821, -0.00001011, -0.00000274, 0.00000640, 0.00000841, 0.00000178, -0.00000711, -0.00001045, -0.00000508, 0.00000526, 0.00001165, 0.00000662, -0.00000737, -0.00001780, -0.00001470, -0.00000233, 0.00000612, 0.00000448, -0.00000104, -0.00000263, -0.00000028, 0.00000152, 0.00000051, -0.00000310, -0.00000776, -0.00000908, -0.00000314, 0.00000529, 0.00000479, -0.00000685, -0.00001633, -0.00001190, 0.00000008, 0.00000537, 0.00000233, 0.00000032, 0.00000206, 0.00000183, -0.00000101, -0.00000141, -0.00000116, -0.00000616, -0.00001081, -0.00000351, 0.00000937, 0.00000720, -0.00000995, -0.00001521, 0.00000179, 0.00001239, -0.00000553, -0.00002426, -0.00000592, 0.00003041, 0.00002901, -0.00001529, -0.00004138, -0.00001093, 0.00003316, 0.00003071, -0.00000883, -0.00002796, -0.00000882, 0.00001050, 0.00000347, -0.00000924, -0.00000350, 0.00000866, 0.00000533, -0.00000758, -0.00001025, -0.00000278, -0.00000002, -0.00000350, -0.00000174, 0.00000632, 0.00000923, 0.00000301, -0.00000396, -0.00000650, -0.00000799, -0.00000888, -0.00000443, 0.00000305, 0.00000329, -0.00000582, -0.00001251, -0.00000780, 0.00000160, 0.00000428, 0.00000006, -0.00000273, -0.00000043, 0.00000329, 0.00000427, 0.00000087, -0.00000669, -0.00001418, -0.00001360, -0.00000233, 0.00000992, 0.00001023, -0.00000169, -0.00001253, -0.00001112, 0.00000003, 0.00000979, 0.00001023, 0.00000187, -0.00000844, -0.00001231, -0.00000566, 0.00000686, 0.00001491, 0.00001230, 0.00000292, -0.00000435, -0.00000525, -0.00000261, -0.00000104, -0.00000184, -0.00000323, },
		},
		{ { -0.00000137, -0.00000021, 0.00000148, 0.00000014, -0.00000555, -0.00001241, -0.00001462, -0.00000876, 0.00000225, 0.00001046, 0.00000929, -0.00000035, -0.00001007, -0.00001097, -0.00000187, 0.00000921, 0.00001211, 0.00000406, -0.00000749, -0.00001170, -0.00000462, 0.00000707, 0.00001236, 0.00000620, -0.00000625, -0.00001516, -0.00001530, -0.00000980, -0.00000527, -0.00000424, -0.00000322, 0.00000200, 0.00001015, 0.00001481, 0.00001114, 0.00000176, -0.00000498, -0.00000292, 0.00000572, 0.00001224, 0.00000964, -0.00000122, -0.00001308, -0.00001914, -0.00001827, -0.00001382, -0.00000873, -0.00000274, 0.00000535, 0.00001404, 0.00001872, 0.00001568, 0.00000615, -0.00000413, -0.00000941, -0.00000826, -0.00000361, 0.00000071, 0.00000302, 0.00000378, 0.00000349, 0.00000179, -0.00000151, -0.00000508, -0.00000665, -0.00000508, -0.00000151, 0.00000179, 0.00000349, 0.00000378, 0.00000302, 0.00000071, -0.00000361, -0.00000826, -0.00000941, -0.00000413, 0.00000615, 0.00001568, 0.00001872, 0.00001404, 0.00000535, -0.00000274, -0.00000873, -0.00001382, -0.00001827, -0.00001914, -0.00001308, -0.00000122, 0.00000964, 0.00001224, 0.00000572, -0.00000292, -0.00000498, 0.00000176, 0.00001114, 0.00001481, 0.00001015, 0.00000200, -0.00000322, -0.00000424, -0.00000527, -0.00000980, -0.00001530, -0.00001516, -0.00000625, 0.00000620, 0.00001236, 0.00000707, -0.00000462, -0.00001170, -0.00000749, 0.00000406, 0.00001211, 0.00000921, -0.00000187, -0.00001097, -0.00001007, -0.00000035, 0.00000929, 0.00001046, 0.00000225, -0.00000876, -0.00001462, -0.00001241, -0.00000555, 0.00000014, 0.00000148, -0.00000021, },
		{ 0.00000110, 0.00000158, -0.00000037, -0.00000130, 0.00000005, 0.00000134, -0.00000094, -0.00000745, -0.00001410, -0.00001478, -0.00000705, 0.00000453, 0.00001109, 0.00000702, -0.00000443, -0.00001306, -0.00001055, 0.00000166, 0.00001307, 0.00001351, 0.00000259, -0.00000966, -0.00001212, -0.00000283, 0.00000937, 0.00001342, 0.00000621, -0.00000547, -0.00001235, -0.00001158, -0.00000801, -0.00000740, -0.00000959, -0.00000896, -0.00000160, 0.00000914, 0.00001515, 0.00001169, 0.00000255, -0.00000333, -0.00000054, 0.00000772, 0.00001307, 0.00001005, 0.00000073, -0.00000833, -0.00001277, -0.00001357, -0.00001405, -0.00001495, -0.00001329, -0.00000619, 0.00000495, 0.00001470, 0.00001806, 0.00001447, 0.00000749, 0.00000114, -0.00000301, -0.00000545, -0.00000648, -0.00000535, -0.00000175, 0.00000270, 0.00000509, 0.00000374, -0.00000030, -0.00000432, -0.00000633, -0.00000611, -0.00000418, -0.00000066, 0.00000418, 0.00000841, 0.00000868, 0.00000305, -0.00000615, -0.00001294, -0.00001219, -0.00000395, 0.00000664, 0.00001377, 0.00001531, 0.00001293, 0.00000859, 0.00000212, -0.00000715, -0.00001696, -0.00002176, -0.00001720, -0.00000528, 0.00000599, 0.00000900, 0.00000344, -0.00000361, -0.00000446, 0.00000190, 0.00000946, 0.00001171, 0.00000795, 0.00000309, 0.00000136, 0.00000155, -0.00000107, -0.00000829, -0.00001545, -0.00001552, -0.00000673, 0.00000444, 0.00000871, 0.00000287, -0.00000685, -0.00001064, -0.00000452, 0.00000587, 0.00001070, 0.00000538, -0.00000521, -0.00001141, -0.00000769, 0.00000259, 0.00001050, 0.00000946, 0.00000043, -0.00000958, -0.00001370, -0.00001040, -0.00000367, },
		{ -0.00001207, -0.00000788, -0.00000153, 0.00000216, 0.00000173, -0.00000047, -0.00000113, 0.00000040, 0.00000121, -0.00000214, -0.00000946, -0.00001561, -0.00001442, -0.00000469, 0.00000715, 0.00001157, 0.00000436, -0.00000867, -0.00001586, -0.00000999, 0.00000466, 0.00001569, 0.00001366, 0.00000090, -0.00001071, -0.00001082, -0.00000009, 0.00001097, 0.00001240, 0.00000389, -0.00000608, -0.00000949, -0.00000645, -0.00000376, -0.00000646, -0.00001192, -0.00001279, -0.00000545, 0.00000552, 0.00001158, 0.00000888, 0.00000206, -0.00000072, 0.00000387, 0.00001119, 0.00001369, 0.00000853, -0.00000011, -0.00000592, -0.00000686, -0.00000640, -0.00000880, -0.00001402, -0.00001764, -0.00001533, -0.00000711, 0.00000309, 0.00001108, 0.00001525, 0.00001605, 0.00001381, 0.00000823, 0.00000000, -0.00000785, -0.00001139, -0.00000896, -0.00000271, 0.00000307, 0.00000538, 0.00000408, 0.00000079, -0.00000308, -0.00000662, -0.00000839, -0.00000634, 0.00000015, 0.00000820, 0.00001246, 0.00000921, -0.00000004, -0.00000930, -0.00001288, -0.00000967, -0.00000292, 0.00000361, 0.00000869, 0.00001273, 0.00001479, 0.00001196, 0.00000268, -0.00000972, -0.00001802, -0.00001696, -0.00000824, 0.00000056, 0.00000289, -0.00000091, -0.00000462, -0.00000284, 0.00000351, 0.00000863, 0.00000832, 0.00000425, 0.00000163, 0.00000295, 0.00000498, 0.00000256, -0.00000494, -0.00001197, -0.00001205, -0.00000481, 0.00000301, 0.00000386, -0.00000298, -0.00001046, -0.00001057, -0.00000227, 0.00000734, 0.00000968, 0.00000273, -0.00000714, -0.00001112, -0.00000583, 0.00000400, 0.00001003, 0.00000747, -0.00000142, -0.00000980, },
		{ 0.00000422, -0.00000400, -0.00001025, -0.00001033, -0.00000507, 0.00000100, 0.00000364, 0.00000222, -0.00000033, -0.00000080, 0.00000080, 0.00000102, -0.00000346, -0.00001146, -0.00001687, -0.00001360, -0.00000205, 0.00000959, 0.00001155, 0.00000146, -0.00001233, -0.00001742, -0.00000845, 0.00000741, 0.00001666, 0.00001172, -0.00000203, -0.00001147, -0.00000817, 0.00000396, 0.00001303, 0.00001090, 0.00000052, -0.00000800, -0.00000800, -0.00000242, 0.00000040, -0.00000381, -0.00001076, -0.00001261, -0.00000660, 0.00000195, 0.00000551, 0.00000240, -0.00000161, 0.00000026, 0.00000830, 0.00001577, 0.00001566, 0.00000769, -0.00000171, -0.00000587, -0.00000402, -0.00000110, -0.00000223, -0.00000789, -0.00001424, -0.00001723, -0.00001559, -0.00001033, -0.00000257, 0.00000694, 0.00001635, 0.00002183, 0.00001980, 0.00001030, -0.00000198, -0.00001073, -0.00001253, -0.00000864, -0.00000288, 0.00000175, 0.00000424, 0.00000443, 0.00000193, -0.00000303, -0.00000804, -0.00000922, -0.00000444, 0.00000400, 0.00001048, 0.00001058, 0.00000474, -0.00000262, -0.00000726, -0.00000835, -0.00000748, -0.00000550, -0.00000129, 0.00000583, 0.00001320, 0.00001565, 0.00001016, -0.00000043, -0.00000910, -0.00001081, -0.00000689, -0.00000323, -0.00000391, -0.00000711, -0.00000745, -0.00000228, 0.00000512, 0.00000877, 0.00000631, 0.00000136, -0.00000053, 0.00000202, 0.00000492, 0.00000340, -0.00000233, -0.00000690, -0.00000560, 0.00000019, 0.00000384, 0.00000025, -0.00000856, -0.00001478, -0.00001219, -0.00000226, 0.00000697, 0.00000817, 0.00000120, -0.00000708, -0.00000920, -0.00000359, 0.00000455, 0.00000814, },
		{ -0.00000029, 0.00000543, 0.00000602, 0.00000042, -0.00000735, -0.00001161, -0.00000945, -0.00000279, 0.00000349, 0.00000549, 0.00000332, 0.00000043, -0.00000003, 0.00000138, 0.00000072, -0.00000497, -0.00001341, -0.00001765, -0.00001218, 0.00000077, 0.00001145, 0.00001051, -0.00000198, -0.00001535, -0.00001730, -0.00000527, 0.00001063, 0.00001661, 0.00000799, -0.00000647, -0.00001291, -0.00000555, 0.00000823, 0.00001544, 0.00001002, -0.00000205, -0.00000958, -0.00000741, -0.00000053, 0.00000227, -0.00000198, -0.00000802, -0.00000885, -0.00000376, 0.00000117, 0.00000016, -0.00000539, -0.00000800, -0.00000206, 0.00000983, 0.00001884, 0.00001810, 0.00000883, -0.00000136, -0.00000570, -0.00000362, 0.00000030, 0.00000148, -0.00000119, -0.00000607, -0.00001146, -0.00001635, -0.00001901, -0.00001643, -0.00000666, 0.00000798, 0.00002084, 0.00002521, 0.00001936, 0.00000766, -0.00000319, -0.00000908, -0.00001005, -0.00000813, -0.00000473, -0.00000043, 0.00000342, 0.00000445, 0.00000133, -0.00000405, -0.00000738, -0.00000566, -0.00000004, 0.00000519, 0.00000668, 0.00000471, 0.00000195, -0.00000002, -0.00000210, -0.00000552, -0.00000892, -0.00000868, -0.00000283, 0.00000596, 0.00001187, 0.00001119, 0.00000568, 0.00000058, -0.00000092, -0.00000062, -0.00000263, -0.00000798, -0.00001262, -0.00001144, -0.00000388, 0.00000480, 0.00000834, 0.00000540, 0.00000039, -0.00000148, 0.00000050, 0.00000250, 0.00000118, -0.00000214, -0.00000286, 0.00000128, 0.00000677, 0.00000714, -0.00000042, -0.00001148, -0.00001775, -0.00001445, -0.00000459, 0.00000381, 0.00000483, -0.00000048, -0.00000585, -0.00000576, },
		{ -0.00000365, -0.00000523, -0.00000227, 0.00000353, 0.00000722, 0.00000499, -0.00000251, -0.00001051, -0.00001364, -0.00000988, -0.00000193, 0.00000494, 0.00000698, 0.00000474, 0.00000197, 0.00000160, 0.00000249, 0.00000049, -0.00000664, -0.00001528, -0.00001790, -0.00001006, 0.00000379, 0.00001263, 0.00000828, -0.00000618, -0.00001800, -0.00001583, -0.00000064, 0.00001455, 0.00001640, 0.00000371, -0.00001142, -0.00001492, -0.00000394, 0.00001108, 0.00001677, 0.00000933, -0.00000302, -0.00000920, -0.00000599, 0.00000023, 0.00000164, -0.00000255, -0.00000622, -0.00000411, 0.00000164, 0.00000376, -0.00000185, -0.00001080, -0.00001401, -0.00000659, 0.00000714, 0.00001750, 0.00001807, 0.00001051, 0.00000171, -0.00000288, -0.00000282, -0.00000080, 0.00000104, 0.00000217, 0.00000185, -0.00000187, -0.00000998, -0.00001958, -0.00002423, -0.00001881, -0.00000443, 0.00001164, 0.00002129, 0.00002138, 0.00001484, 0.00000679, 0.00000023, -0.00000465, -0.00000811, -0.00000915, -0.00000669, -0.00000184, 0.00000211, 0.00000248, -0.00000015, -0.00000263, -0.00000255, -0.00000058, 0.00000084, 0.00000065, 0.00000036, 0.00000179, 0.00000414, 0.00000426, 0.00000011, -0.00000625, -0.00000995, -0.00000791, -0.00000180, 0.00000372, 0.00000576, 0.00000559, 0.00000623, 0.00000805, 0.00000775, 0.00000217, -0.00000724, -0.00001442, -0.00001404, -0.00000653, 0.00000211, 0.00000590, 0.00000403, 0.00000049, -0.00000083, -0.00000006, -0.00000017, -0.00000240, -0.00000401, -0.00000116, 0.00000606, 0.00001226, 0.00001112, 0.00000154, -0.00001066, -0.00001743, -0.00001522, -0.00000757, -0.00000138, -0.00000067, },
		{ -0.00000889, -0.00000627, -0.00000691, -0.00000826, -0.00000653, -0.00000069, 0.00000613, 0.00000904, 0.00000528, -0.00000348, -0.00001201, -0.00001505, -0.00001090, -0.00000257, 0.00000458, 0.00000701, 0.00000551, 0.00000377, 0.00000407, 0.00000446, 0.00000075, -0.00000811, -0.00001689, -0.00001759, -0.00000736, 0.00000691, 0.00001317, 0.00000511, -0.00001090, -0.00002033, -0.00001358, 0.00000455, 0.00001854, 0.00001618, -0.00000006, -0.00001533, -0.00001628, -0.00000307, 0.00001173, 0.00001560, 0.00000738, -0.00000331, -0.00000682, -0.00000269, 0.00000167, 0.00000027, -0.00000478, -0.00000634, -0.00000101, 0.00000637, 0.00000745, -0.00000058, -0.00001154, -0.00001592, -0.00000979, 0.00000203, 0.00001110, 0.00001293, 0.00000950, 0.00000534, 0.00000265, 0.00000067, -0.00000121, -0.00000143, 0.00000159, 0.00000596, 0.00000640, -0.00000088, -0.00001351, -0.00002358, -0.00002398, -0.00001421, -0.00000043, 0.00001014, 0.00001455, 0.00001461, 0.00001302, 0.00001023, 0.00000520, -0.00000172, -0.00000768, -0.00000967, -0.00000745, -0.00000366, -0.00000099, 0.00000016, 0.00000113, 0.00000235, 0.00000244, 0.00000008, -0.00000347, -0.00000482, -0.00000184, 0.00000357, 0.00000678, 0.00000477, -0.00000092, -0.00000588, -0.00000713, -0.00000551, -0.00000366, -0.00000235, 0.00000031, 0.00000584, 0.00001202, 0.00001378, 0.00000800, -0.00000268, -0.00001140, -0.00001286, -0.00000771, -0.00000129, 0.00000174, 0.00000118, 0.00000002, 0.00000014, 0.00000025, -0.00000180, -0.00000526, -0.00000610, -0.00000110, 0.00000796, 0.00001469, 0.00001335, 0.00000405, -0.00000719, -0.00001358, -0.00001300, },
		{ -0.00000355, -0.00000808, -0.00000825, -0.00000736, -0.00000848, -0.00001129, -0.00001255, -0.00000932, -0.00000194, 0.00000577, 0.00000906, 0.00000557, -0.00000287, -0.00001120, -0.00001442, -0.00001091, -0.00000355, 0.00000273, 0.00000501, 0.00000452, 0.00000466, 0.00000654, 0.00000699, 0.00000173, -0.00000883, -0.00001765, -0.00001644, -0.00000419, 0.00000976, 0.00001290, 0.00000125, -0.00001554, -0.00002193, -0.00001072, 0.00000946, 0.00002149, 0.00001523, -0.00000318, -0.00001730, -0.00001566, -0.00000166, 0.00001097, 0.00001190, 0.00000323, -0.00000454, -0.00000414, 0.00000162, 0.00000438, 0.00000016, -0.00000623, -0.00000692, -0.00000001, 0.00000797, 0.00000882, 0.00000114, -0.00000845, -0.00001208, -0.00000812, -0.00000150, 0.00000237, 0.00000292, 0.00000348, 0.00000597, 0.00000809, 0.00000634, 0.00000113, -0.00000258, -0.00000037, 0.00000624, 0.00001034, 0.00000607, -0.00000544, -0.00001680, -0.00002095, -0.00001684, -0.00000904, -0.00000238, 0.00000216, 0.00000653, 0.00001159, 0.00001510, 0.00001394, 0.00000770, -0.00000043, -0.00000651, -0.00000891, -0.00000857, -0.00000675, -0.00000353, 0.00000104, 0.00000524, 0.00000607, 0.00000221, -0.00000385, -0.00000733, -0.00000539, 0.00000021, 0.00000488, 0.00000549, 0.00000280, -0.00000030, -0.00000227, -0.00000403, -0.00000657, -0.00000833, -0.00000607, 0.00000128, 0.00001014, 0.00001445, 0.00001095, 0.00000226, -0.00000542, -0.00000779, -0.00000565, -0.00000291, -0.00000211, -0.00000241, -0.00000174, -0.00000011, 0.00000008, -0.00000274, -0.00000649, -0.00000670, -0.00000093, 0.00000805, 0.00001401, 0.00001258, 0.00000492, },
		{ 0.00001182, 0.00000950, 0.00000344, -0.00000179, -0.00000359, -0.00000321, -0.00000387, -0.00000726, -0.00001182, -0.00001406, -0.00001154, -0.00000494, 0.00000227, 0.00000593, 0.00000386, -0.00000256, -0.00000912, -0.00001154, -0.00000861, -0.00000307, 0.00000095, 0.00000176, 0.00000158, 0.00000364, 0.00000765, 0.00000890, 0.00000287, -0.00000860, -0.00001690, -0.00001379, -0.00000033, 0.00001216, 0.00001152, -0.00000330, -0.00001972, -0.00002226, -0.00000701, 0.00001384, 0.00002275, 0.00001277, -0.00000634, -0.00001764, -0.00001279, 0.00000123, 0.00001031, 0.00000726, -0.00000231, -0.00000730, -0.00000308, 0.00000460, 0.00000685, 0.00000152, -0.00000496, -0.00000522, 0.00000094, 0.00000669, 0.00000594, -0.00000015, -0.00000503, -0.00000439, -0.00000076, -0.00000005, -0.00000418, -0.00000844, -0.00000666, 0.00000156, 0.00000974, 0.00001103, 0.00000529, -0.00000074, -0.00000056, 0.00000544, 0.00001052, 0.00000863, 0.00000025, -0.00000864, -0.00001280, -0.00001225, -0.00001070, -0.00001045, -0.00000974, -0.00000528, 0.00000329, 0.00001213, 0.00001646, 0.00001455, 0.00000846, 0.00000148, -0.00000445, -0.00000871, -0.00001042, -0.00000827, -0.00000241, 0.00000412, 0.00000696, 0.00000413, -0.00000191, -0.00000633, -0.00000615, -0.00000254, 0.00000107, 0.00000265, 0.00000286, 0.00000306, 0.00000279, 0.00000024, -0.00000489, -0.00000955, -0.00000948, -0.00000342, 0.00000500, 0.00001011, 0.00000911, 0.00000423, -0.00000004, -0.00000133, -0.00000097, -0.00000152, -0.00000343, -0.00000472, -0.00000381, -0.00000186, -0.00000166, -0.00000416, -0.00000677, -0.00000556, 0.00000052, 0.00000804, },
		{ -0.00000376, 0.00000290, 0.00000872, 0.00000984, 0.00000602, 0.00000071, -0.00000223, -0.00000181, -0.00000030, -0.00000070, -0.00000402, -0.00000857, -0.00001150, -0.00001093, -0.00000717, -0.00000249, 0.00000018, -0.00000085, -0.00000465, -0.00000804, -0.00000798, -0.00000430, -0.00000013, 0.00000106, -0.00000081, -0.00000200, 0.00000093, 0.00000652, 0.00000867, 0.00000276, -0.00000819, -0.00001459, -0.00000914, 0.00000478, 0.00001442, 0.00000913, -0.00000851, -0.00002339, -0.00002123, -0.00000234, 0.00001781, 0.00002244, 0.00000873, -0.00001001, -0.00001732, -0.00000877, 0.00000516, 0.00001054, 0.00000360, -0.00000699, -0.00001021, -0.00000393, 0.00000432, 0.00000643, 0.00000230, -0.00000140, 0.00000026, 0.00000469, 0.00000534, 0.00000005, -0.00000572, -0.00000485, 0.00000288, 0.00000946, 0.00000659, -0.00000523, -0.00001632, -0.00001693, -0.00000654, 0.00000600, 0.00001130, 0.00000800, 0.00000268, 0.00000206, 0.00000628, 0.00000979, 0.00000802, 0.00000217, -0.00000286, -0.00000423, -0.00000406, -0.00000631, -0.00001165, -0.00001612, -0.00001491, -0.00000717, 0.00000328, 0.00001140, 0.00001468, 0.00001361, 0.00000962, 0.00000357, -0.00000344, -0.00000892, -0.00000994, -0.00000585, 0.00000037, 0.00000414, 0.00000318, -0.00000072, -0.00000385, -0.00000419, -0.00000276, -0.00000158, -0.00000097, 0.00000039, 0.00000304, 0.00000507, 0.00000371, -0.00000138, -0.00000687, -0.00000849, -0.00000505, 0.00000048, 0.00000394, 0.00000385, 0.00000229, 0.00000197, 0.00000316, 0.00000366, 0.00000173, -0.00000174, -0.00000419, -0.00000441, -0.00000383, -0.00000460, -0.00000664, -0.00000721, },
		{ -0.00000716, -0.00000929, -0.00000803, -0.00000240, 0.00000497, 0.00000953, 0.00000863, 0.00000362, -0.00000151, -0.00000359, -0.00000251, -0.00000049, 0.00000018, -0.00000116, -0.00000363, -0.00000584, -0.00000683, -0.00000648, -0.00000564, -0.00000570, -0.00000741, -0.00000969, -0.00000990, -0.00000622, -0.00000005, 0.00000434, 0.00000361, -0.00000093, -0.00000400, -0.00000170, 0.00000378, 0.00000570, -0.00000002, -0.00000910, -0.00001181, -0.00000288, 0.00001134, 0.00001715, 0.00000645, -0.00001384, -0.00002642, -0.00001920, 0.00000280, 0.00002125, 0.00002103, 0.00000392, -0.00001374, -0.00001682, -0.00000502, 0.00000844, 0.00001077, 0.00000153, -0.00000880, -0.00001073, -0.00000479, 0.00000108, 0.00000160, -0.00000031, 0.00000164, 0.00000815, 0.00001222, 0.00000699, -0.00000522, -0.00001347, -0.00000876, 0.00000607, 0.00001781, 0.00001504, -0.00000107, -0.00001779, -0.00002267, -0.00001415, -0.00000124, 0.00000626, 0.00000637, 0.00000442, 0.00000557, 0.00000911, 0.00001040, 0.00000706, 0.00000198, -0.00000014, 0.00000168, 0.00000338, 0.00000030, -0.00000769, -0.00001574, -0.00001846, -0.00001439, -0.00000636, 0.00000180, 0.00000807, 0.00001200, 0.00001305, 0.00001028, 0.00000406, -0.00000277, -0.00000652, -0.00000573, -0.00000247, -0.00000024, -0.00000061, -0.00000218, -0.00000273, -0.00000189, -0.00000124, -0.00000192, -0.00000287, -0.00000195, 0.00000125, 0.00000435, 0.00000436, 0.00000089, -0.00000325, -0.00000481, -0.00000329, -0.00000107, -0.00000055, -0.00000150, -0.00000162, 0.00000057, 0.00000385, 0.00000551, 0.00000423, 0.00000129, -0.00000121, -0.00000272, -0.00000445, },
		{ 0.00000200, 0.00000012, -0.00000348, -0.00000802, -0.00001051, -0.00000819, -0.00000142, 0.00000594, 0.00000943, 0.00000752, 0.00000240, -0.00000230, -0.00000441, -0.00000414, -0.00000283, -0.00000144, -0.00000027, 0.00000043, 0.00000029, -0.00000083, -0.00000273, -0.00000532, -0.00000886, -0.00001305, -0.00001595, -0.00001460, -0.00000778, 0.00000160, 0.00000778, 0.00000705, 0.00000144, -0.00000287, -0.00000194, 0.00000155, 0.00000115, -0.00000555, -0.00001249, -0.00001026, 0.00000335, 0.00001839, 0.00002038, 0.00000442, -0.00001804, -0.00002804, -0.00001630, 0.00000757, 0.00002332, 0.00001845, -0.00000081, -0.00001632, -0.00001547, -0.00000201, 0.00000956, 0.00000912, -0.00000024, -0.00000769, -0.00000718, -0.00000289, -0.00000238, -0.00000640, -0.00000740, 0.00000102, 0.00001477, 0.00002136, 0.00001236, -0.00000652, -0.00001924, -0.00001416, 0.00000465, 0.00002083, 0.00002060, 0.00000467, -0.00001373, -0.00002168, -0.00001707, -0.00000767, -0.00000154, 0.00000028, 0.00000215, 0.00000679, 0.00001164, 0.00001209, 0.00000745, 0.00000222, 0.00000135, 0.00000489, 0.00000788, 0.00000546, -0.00000237, -0.00001104, -0.00001587, -0.00001562, -0.00001209, -0.00000721, -0.00000149, 0.00000478, 0.00000998, 0.00001168, 0.00000910, 0.00000428, 0.00000038, -0.00000118, -0.00000159, -0.00000269, -0.00000455, -0.00000541, -0.00000405, -0.00000157, -0.00000033, -0.00000126, -0.00000278, -0.00000249, 0.00000001, 0.00000257, 0.00000285, 0.00000087, -0.00000114, -0.00000134, -0.00000029, -0.00000014, -0.00000193, -0.00000419, -0.00000446, -0.00000195, 0.00000162, 0.00000387, 0.00000405, 0.00000313, },
		{ 0.00000055, 0.00000151, 0.00000256, 0.00000319, 0.00000176, -0.00000244, -0.00000745, -0.00000955, -0.00000648, 0.00000022, 0.00000629, 0.00000824, 0.00000583, 0.00000153, -0.00000212, -0.00000432, -0.00000541, -0.00000549, -0.00000407, -0.00000108, 0.00000231, 0.00000442, 0.00000433, 0.00000213, -0.00000198, -0.00000804, -0.00001520, -0.00002044, -0.00001976, -0.00001181, -0.00000038, 0.00000774, 0.00000850, 0.00000419, 0.00000080, 0.00000127, 0.00000204, -0.00000249, -0.00001196, -0.00001779, -0.00001088, 0.00000752, 0.00002376, 0.00002284, 0.00000318, -0.00001989, -0.00002684, -0.00001182, 0.00001164, 0.00002308, 0.00001392, -0.00000549, -0.00001693, -0.00001204, 0.00000122, 0.00000853, 0.00000434, -0.00000387, -0.00000575, -0.00000038, 0.00000311, -0.00000284, -0.00001390, -0.00001692, -0.00000416, 0.00001664, 0.00002791, 0.00001886, -0.00000339, -0.00001982, -0.00001688, 0.00000151, 0.00001857, 0.00002042, 0.00000755, -0.00000812, -0.00001563, -0.00001394, -0.00000958, -0.00000774, -0.00000718, -0.00000354, 0.00000403, 0.00001105, 0.00001248, 0.00000837, 0.00000384, 0.00000363, 0.00000730, 0.00001008, 0.00000794, 0.00000141, -0.00000566, -0.00001012, -0.00001185, -0.00001228, -0.00001177, -0.00000915, -0.00000384, 0.00000256, 0.00000724, 0.00000874, 0.00000796, 0.00000669, 0.00000533, 0.00000284, -0.00000130, -0.00000559, -0.00000755, -0.00000624, -0.00000335, -0.00000153, -0.00000169, -0.00000238, -0.00000177, 0.00000014, 0.00000158, 0.00000121, -0.00000023, -0.00000081, 0.00000022, 0.00000149, 0.00000105, -0.00000136, -0.00000389, -0.00000451, -0.00000304, -0.00000092, },
		{ -0.00000255, -0.00000200, -0.00000176, -0.00000179, -0.00000109, 0.00000066, 0.00000193, 0.00000061, -0.00000338, -0.00000723, -0.00000739, -0.00000296, 0.00000325, 0.00000719, 0.00000696, 0.00000387, 0.00000049, -0.00000202, -0.00000413, -0.00000623, -0.00000732, -0.00000593, -0.00000201, 0.00000261, 0.00000579, 0.00000668, 0.00000551, 0.00000205, -0.00000448, -0.00001339, -0.00002098, -0.00002224, -0.00001540, -0.00000442, 0.00000385, 0.00000588, 0.00000435, 0.00000433, 0.00000650, 0.00000561, -0.00000318, -0.00001634, -0.00002245, -0.00001259, 0.00000885, 0.00002529, 0.00002217, 0.00000126, -0.00001942, -0.00002176, -0.00000443, 0.00001585, 0.00002049, 0.00000663, -0.00001112, -0.00001611, -0.00000604, 0.00000613, 0.00000705, -0.00000278, -0.00001043, -0.00000577, 0.00000658, 0.00001173, 0.00000101, -0.00001720, -0.00002465, -0.00001144, 0.00001308, 0.00002846, 0.00002223, 0.00000129, -0.00001526, -0.00001422, 0.00000068, 0.00001432, 0.00001534, 0.00000543, -0.00000509, -0.00000876, -0.00000709, -0.00000636, -0.00000922, -0.00001190, -0.00000932, -0.00000142, 0.00000649, 0.00000936, 0.00000746, 0.00000542, 0.00000676, 0.00001029, 0.00001182, 0.00000893, 0.00000326, -0.00000174, -0.00000448, -0.00000623, -0.00000864, -0.00001132, -0.00001211, -0.00000964, -0.00000490, -0.00000006, 0.00000364, 0.00000650, 0.00000888, 0.00000980, 0.00000776, 0.00000275, -0.00000293, -0.00000637, -0.00000658, -0.00000509, -0.00000399, -0.00000380, -0.00000333, -0.00000166, 0.00000047, 0.00000143, 0.00000064, -0.00000065, -0.00000079, 0.00000037, 0.00000135, 0.00000080, -0.00000096, -0.00000239, },
		{ -0.00000059, -0.00000164, -0.00000161, -0.00000074, -0.00000034, -0.00000102, -0.00000185, -0.00000148, -0.00000020, -0.00000009, -0.00000276, -0.00000695, -0.00000897, -0.00000608, 0.00000063, 0.00000691, 0.00000903, 0.00000680, 0.00000292, -0.00000015, -0.00000232, -0.00000474, -0.00000741, -0.00000848, -0.00000635, -0.00000172, 0.00000286, 0.00000547, 0.00000629, 0.00000613, 0.00000434, -0.00000091, -0.00000959, -0.00001785, -0.00002041, -0.00001554, -0.00000728, -0.00000161, -0.00000031, 0.00000046, 0.00000478, 0.00001049, 0.00000999, -0.00000119, -0.00001686, -0.00002335, -0.00001241, 0.00000889, 0.00002263, 0.00001645, -0.00000377, -0.00001855, -0.00001344, 0.00000661, 0.00002179, 0.00001705, -0.00000293, -0.00001863, -0.00001556, 0.00000127, 0.00001307, 0.00000720, -0.00000964, -0.00001864, -0.00000897, 0.00001033, 0.00001925, 0.00000737, -0.00001504, -0.00002680, -0.00001646, 0.00000685, 0.00002285, 0.00001953, 0.00000335, -0.00000889, -0.00000698, 0.00000425, 0.00001199, 0.00000903, -0.00000020, -0.00000621, -0.00000487, -0.00000076, -0.00000074, -0.00000609, -0.00001155, -0.00001158, -0.00000613, -0.00000001, 0.00000281, 0.00000316, 0.00000465, 0.00000886, 0.00001324, 0.00001395, 0.00001012, 0.00000465, 0.00000096, -0.00000035, -0.00000143, -0.00000419, -0.00000808, -0.00001083, -0.00001099, -0.00000902, -0.00000615, -0.00000274, 0.00000159, 0.00000639, 0.00000976, 0.00000969, 0.00000612, 0.00000130, -0.00000225, -0.00000378, -0.00000440, -0.00000522, -0.00000593, -0.00000534, -0.00000318, -0.00000067, 0.00000062, 0.00000045, -0.00000005, 0.00000011, 0.00000064, 0.00000050, },
		{ 0.00000130, 0.00000159, 0.00000042, -0.00000142, -0.00000236, -0.00000172, -0.00000051, -0.00000013, -0.00000061, -0.00000060, 0.00000059, 0.00000146, -0.00000059, -0.00000600, -0.00001154, -0.00001245, -0.00000687, 0.00000214, 0.00000907, 0.00001058, 0.00000774, 0.00000384, 0.00000075, -0.00000220, -0.00000597, -0.00000935, -0.00000969, -0.00000592, -0.00000023, 0.00000388, 0.00000502, 0.00000474, 0.00000476, 0.00000420, 0.00000051, -0.00000666, -0.00001353, -0.00001530, -0.00001137, -0.00000638, -0.00000519, -0.00000700, -0.00000601, 0.00000136, 0.00001042, 0.00001149, 0.00000045, -0.00001455, -0.00001918, -0.00000731, 0.00001106, 0.00001814, 0.00000612, -0.00001354, -0.00002016, -0.00000465, 0.00002003, 0.00003030, 0.00001507, -0.00001256, -0.00002745, -0.00001680, 0.00000739, 0.00002046, 0.00000966, -0.00001353, -0.00002540, -0.00001394, 0.00000951, 0.00002215, 0.00001237, -0.00000953, -0.00002258, -0.00001580, 0.00000237, 0.00001435, 0.00001131, 0.00000034, -0.00000488, 0.00000111, 0.00001089, 0.00001325, 0.00000520, -0.00000563, -0.00000962, -0.00000489, 0.00000192, 0.00000338, -0.00000137, -0.00000712, -0.00000895, -0.00000704, -0.00000493, -0.00000447, -0.00000362, 0.00000059, 0.00000781, 0.00001401, 0.00001529, 0.00001169, 0.00000678, 0.00000385, 0.00000303, 0.00000204, -0.00000075, -0.00000466, -0.00000779, -0.00000919, -0.00000931, -0.00000869, -0.00000690, -0.00000322, 0.00000174, 0.00000592, 0.00000738, 0.00000608, 0.00000373, 0.00000192, 0.00000067, -0.00000100, -0.00000343, -0.00000562, -0.00000624, -0.00000514, -0.00000339, -0.00000201, -0.00000102, 0.00000012, },
		{ -0.00000556, -0.00000429, -0.00000177, 0.00000082, 0.00000188, 0.00000099, -0.00000059, -0.00000137, -0.00000120, -0.00000120, -0.00000190, -0.00000211, -0.00000019, 0.00000324, 0.00000478, 0.00000114, -0.00000697, -0.00001435, -0.00001534, -0.00000879, 0.00000093, 0.00000794, 0.00000973, 0.00000811, 0.00000585, 0.00000339, -0.00000074, -0.00000662, -0.00001122, -0.00001090, -0.00000538, 0.00000144, 0.00000514, 0.00000493, 0.00000353, 0.00000322, 0.00000284, -0.00000033, -0.00000621, -0.00001065, -0.00000972, -0.00000477, -0.00000182, -0.00000490, -0.00001069, -0.00001132, -0.00000349, 0.00000634, 0.00000769, -0.00000212, -0.00001316, -0.00001201, 0.00000319, 0.00001825, 0.00001594, -0.00000537, -0.00002690, -0.00002599, 0.00000096, 0.00003235, 0.00004003, 0.00001593, -0.00001932, -0.00003522, -0.00001967, 0.00001026, 0.00002552, 0.00001293, -0.00001354, -0.00002785, -0.00001753, 0.00000569, 0.00001937, 0.00001269, -0.00000468, -0.00001461, -0.00000929, 0.00000269, 0.00000750, 0.00000130, -0.00000668, -0.00000529, 0.00000608, 0.00001686, 0.00001650, 0.00000497, -0.00000774, -0.00001190, -0.00000668, 0.00000082, 0.00000371, 0.00000142, -0.00000191, -0.00000325, -0.00000371, -0.00000577, -0.00000902, -0.00000979, -0.00000517, 0.00000340, 0.00001102, 0.00001380, 0.00001205, 0.00000908, 0.00000746, 0.00000684, 0.00000528, 0.00000194, -0.00000203, -0.00000503, -0.00000666, -0.00000762, -0.00000820, -0.00000768, -0.00000532, -0.00000172, 0.00000140, 0.00000274, 0.00000267, 0.00000256, 0.00000307, 0.00000351, 0.00000272, 0.00000055, -0.00000203, -0.00000396, -0.00000505, -0.00000562, },
		{ 0.00000118, -0.00000146, -0.00000472, -0.00000722, -0.00000750, -0.00000545, -0.00000258, -0.00000057, 0.00000030, 0.00000091, 0.00000162, 0.00000139, -0.00000084, -0.00000402, -0.00000502, -0.00000167, 0.00000414, 0.00000715, 0.00000335, -0.00000571, -0.00001372, -0.00001495, -0.00000919, -0.00000128, 0.00000385, 0.00000543, 0.00000591, 0.00000675, 0.00000617, 0.00000164, -0.00000589, -0.00001154, -0.00001086, -0.00000441, 0.00000261, 0.00000554, 0.00000444, 0.00000281, 0.00000261, 0.00000192, -0.00000192, -0.00000761, -0.00001001, -0.00000591, 0.00000119, 0.00000370, -0.00000184, -0.00001012, -0.00001215, -0.00000559, 0.00000164, -0.00000023, -0.00001060, -0.00001669, -0.00000654, 0.00001579, 0.00003023, 0.00001910, -0.00001308, -0.00003963, -0.00003473, 0.00000126, 0.00003948, 0.00004745, 0.00001865, -0.00002147, -0.00003894, -0.00002205, 0.00000978, 0.00002615, 0.00001431, -0.00001118, -0.00002528, -0.00001705, 0.00000216, 0.00001292, 0.00000776, -0.00000362, -0.00000716, -0.00000005, 0.00000755, 0.00000509, -0.00000607, -0.00001392, -0.00000879, 0.00000638, 0.00001878, 0.00001828, 0.00000645, -0.00000618, -0.00001074, -0.00000712, -0.00000154, 0.00000098, 0.00000080, 0.00000094, 0.00000206, 0.00000151, -0.00000283, -0.00000906, -0.00001230, -0.00000946, -0.00000216, 0.00000494, 0.00000861, 0.00000933, 0.00000957, 0.00001049, 0.00001091, 0.00000914, 0.00000515, 0.00000067, -0.00000268, -0.00000469, -0.00000603, -0.00000685, -0.00000654, -0.00000479, -0.00000244, -0.00000096, -0.00000088, -0.00000133, -0.00000097, 0.00000057, 0.00000243, 0.00000354, 0.00000359, 0.00000279, },
		{ 0.00000020, 0.00000279, 0.00000520, 0.00000603, 0.00000414, -0.00000007, -0.00000462, -0.00000752, -0.00000827, -0.00000765, -0.00000623, -0.00000355, 0.00000062, 0.00000461, 0.00000541, 0.00000162, -0.00000420, -0.00000694, -0.00000371, 0.00000283, 0.00000647, 0.00000324, -0.00000458, -0.00001064, -0.00001057, -0.00000588, -0.00000158, -0.00000054, -0.00000088, 0.00000082, 0.00000463, 0.00000646, 0.00000279, -0.00000464, -0.00000980, -0.00000824, -0.00000170, 0.00000396, 0.00000491, 0.00000279, 0.00000167, 0.00000245, 0.00000181, -0.00000295, -0.00000906, -0.00001019, -0.00000377, 0.00000485, 0.00000710, 0.00000063, -0.00000745, -0.00000830, -0.00000210, 0.00000094, -0.00000771, -0.00002252, -0.00002619, -0.00000712, 0.00002481, 0.00004279, 0.00002708, -0.00001369, -0.00004643, -0.00004203, -0.00000254, 0.00003926, 0.00004871, 0.00002001, -0.00001988, -0.00003710, -0.00002129, 0.00000797, 0.00002243, 0.00001193, -0.00000919, -0.00001955, -0.00001213, 0.00000153, 0.00000637, 0.00000000, -0.00000664, -0.00000326, 0.00000767, 0.00001374, 0.00000659, -0.00000849, -0.00001753, -0.00001183, 0.00000373, 0.00001598, 0.00001607, 0.00000649, -0.00000329, -0.00000665, -0.00000465, -0.00000230, -0.00000204, -0.00000188, 0.00000064, 0.00000437, 0.00000531, 0.00000119, -0.00000564, -0.00001025, -0.00000980, -0.00000572, -0.00000145, 0.00000124, 0.00000339, 0.00000667, 0.00001075, 0.00001340, 0.00001278, 0.00000920, 0.00000457, 0.00000053, -0.00000256, -0.00000487, -0.00000611, -0.00000568, -0.00000375, -0.00000161, -0.00000074, -0.00000145, -0.00000279, -0.00000353, -0.00000315, -0.00000184, },
		{ -0.00000258, -0.00000452, -0.00000577, -0.00000511, -0.00000190, 0.00000286, 0.00000678, 0.00000780, 0.00000568, 0.00000176, -0.00000262, -0.00000700, -0.00001092, -0.00001268, -0.00001009, -0.00000310, 0.00000467, 0.00000801, 0.00000473, -0.00000183, -0.00000558, -0.00000352, 0.00000135, 0.00000314, -0.00000059, -0.00000624, -0.00000805, -0.00000432, 0.00000073, 0.00000167, -0.00000209, -0.00000580, -0.00000483, -0.00000014, 0.00000277, 0.00000018, -0.00000533, -0.00000750, -0.00000331, 0.00000358, 0.00000690, 0.00000458, 0.00000062, -0.00000023, 0.00000163, 0.00000130, -0.00000385, -0.00000969, -0.00000924, -0.00000115, 0.00000731, 0.00000754, -0.00000032, -0.00000666, -0.00000318, 0.00000593, 0.00000655, -0.00000962, -0.00003216, -0.00003812, -0.00001433, 0.00002569, 0.00004972, 0.00003524, -0.00000801, -0.00004422, -0.00004286, -0.00000618, 0.00003319, 0.00004221, 0.00001664, -0.00001768, -0.00003091, -0.00001622, 0.00000740, 0.00001661, 0.00000614, -0.00000946, -0.00001364, -0.00000491, 0.00000410, 0.00000244, -0.00000695, -0.00001130, -0.00000308, 0.00001118, 0.00001746, 0.00000894, -0.00000695, -0.00001608, -0.00001133, 0.00000152, 0.00001075, 0.00001011, 0.00000326, -0.00000205, -0.00000231, -0.00000036, -0.00000047, -0.00000284, -0.00000390, -0.00000112, 0.00000359, 0.00000582, 0.00000336, -0.00000169, -0.00000546, -0.00000622, -0.00000540, -0.00000508, -0.00000516, -0.00000360, 0.00000094, 0.00000719, 0.00001224, 0.00001398, 0.00001251, 0.00000923, 0.00000531, 0.00000117, -0.00000272, -0.00000535, -0.00000576, -0.00000409, -0.00000169, -0.00000006, 0.00000015, -0.00000085, },
		{ -0.00000137, 0.00000102, 0.00000226, 0.00000136, -0.00000181, -0.00000579, -0.00000810, -0.00000701, -0.00000293, 0.00000206, 0.00000597, 0.00000806, 0.00000810, 0.00000538, -0.00000083, -0.00000926, -0.00001577, -0.00001588, -0.00000886, 0.00000065, 0.00000621, 0.00000516, 0.00000086, -0.00000121, 0.00000062, 0.00000255, -0.00000002, -0.00000657, -0.00001115, -0.00000841, 0.00000027, 0.00000751, 0.00000713, 0.00000023, -0.00000635, -0.00000743, -0.00000463, -0.00000362, -0.00000678, -0.00001000, -0.00000737, 0.00000150, 0.00001012, 0.00001166, 0.00000615, 0.00000001, -0.00000141, 0.00000025, -0.00000085, -0.00000645, -0.00001079, -0.00000696, 0.00000378, 0.00001135, 0.00000733, -0.00000445, -0.00001029, -0.00000147, 0.00001381, 0.00001632, -0.00000403, -0.00003401, -0.00004599, -0.00002397, 0.00001829, 0.00004674, 0.00003742, -0.00000098, -0.00003447, -0.00003488, -0.00000503, 0.00002569, 0.00003005, 0.00000755, -0.00001807, -0.00002368, -0.00000831, 0.00000921, 0.00001142, -0.00000076, -0.00001176, -0.00000963, 0.00000161, 0.00000789, 0.00000144, -0.00001062, -0.00001437, -0.00000432, 0.00001071, 0.00001680, 0.00000897, -0.00000450, -0.00001122, -0.00000680, 0.00000222, 0.00000643, 0.00000313, -0.00000239, -0.00000377, -0.00000037, 0.00000307, 0.00000251, -0.00000108, -0.00000336, -0.00000183, 0.00000173, 0.00000368, 0.00000258, 0.00000025, -0.00000098, -0.00000106, -0.00000195, -0.00000475, -0.00000787, -0.00000842, -0.00000495, 0.00000117, 0.00000714, 0.00001106, 0.00001265, 0.00001240, 0.00001042, 0.00000667, 0.00000181, -0.00000266, -0.00000525, -0.00000547, -0.00000385, },
		{ -0.00000199, -0.00000549, -0.00000661, -0.00000491, -0.00000102, 0.00000309, 0.00000489, 0.00000309, -0.00000129, -0.00000566, -0.00000793, -0.00000764, -0.00000534, -0.00000135, 0.00000416, 0.00000959, 0.00001154, 0.00000708, -0.00000285, -0.00001275, -0.00001648, -0.00001249, -0.00000514, -0.00000029, 0.00000039, 0.00000031, 0.00000309, 0.00000733, 0.00000738, -0.00000014, -0.00001116, -0.00001671, -0.00001136, 0.00000109, 0.00001080, 0.00001115, 0.00000426, -0.00000245, -0.00000476, -0.00000537, -0.00000913, -0.00001557, -0.00001804, -0.00001091, 0.00000315, 0.00001458, 0.00001600, 0.00000930, 0.00000257, 0.00000074, 0.00000057, -0.00000386, -0.00001185, -0.00001475, -0.00000571, 0.00001026, 0.00001865, 0.00001003, -0.00000816, -0.00001704, -0.00000550, 0.00001624, 0.00002455, 0.00000604, -0.00002660, -0.00004436, -0.00002920, 0.00000749, 0.00003452, 0.00002995, 0.00000134, -0.00002264, -0.00002040, 0.00000261, 0.00002132, 0.00001700, -0.00000485, -0.00002193, -0.00001861, -0.00000076, 0.00001219, 0.00000814, -0.00000619, -0.00001417, -0.00000753, 0.00000555, 0.00001033, 0.00000167, -0.00001095, -0.00001403, -0.00000427, 0.00000864, 0.00001266, 0.00000557, -0.00000392, -0.00000616, -0.00000038, 0.00000576, 0.00000506, -0.00000177, -0.00000771, -0.00000721, -0.00000152, 0.00000361, 0.00000421, 0.00000152, -0.00000062, -0.00000028, 0.00000105, 0.00000116, 0.00000007, -0.00000030, 0.00000110, 0.00000273, 0.00000201, -0.00000179, -0.00000650, -0.00000903, -0.00000795, -0.00000423, 0.00000030, 0.00000458, 0.00000845, 0.00001159, 0.00001304, 0.00001186, 0.00000813, 0.00000299, },
		{ 0.00001360, 0.00001334, 0.00001006, 0.00000446, -0.00000193, -0.00000681, -0.00000805, -0.00000516, -0.00000001, 0.00000437, 0.00000588, 0.00000450, 0.00000142, -0.00000245, -0.00000666, -0.00000995, -0.00000984, -0.00000457, 0.00000417, 0.00001107, 0.00001108, 0.00000401, -0.00000498, -0.00001001, -0.00000980, -0.00000797, -0.00000802, -0.00000866, -0.00000534, 0.00000354, 0.00001260, 0.00001338, 0.00000299, -0.00001163, -0.00001904, -0.00001364, -0.00000068, 0.00000930, 0.00001052, 0.00000618, 0.00000281, 0.00000193, -0.00000125, -0.00001057, -0.00002197, -0.00002574, -0.00001635, 0.00000080, 0.00001388, 0.00001595, 0.00001084, 0.00000717, 0.00000751, 0.00000564, -0.00000449, -0.00001832, -0.00002243, -0.00000856, 0.00001478, 0.00002752, 0.00001722, -0.00000716, -0.00002226, -0.00001312, 0.00001120, 0.00002569, 0.00001425, -0.00001380, -0.00003250, -0.00002493, 0.00000039, 0.00001874, 0.00001442, -0.00000424, -0.00001491, -0.00000534, 0.00001405, 0.00002160, 0.00000747, -0.00001602, -0.00002710, -0.00001658, 0.00000368, 0.00001380, 0.00000595, -0.00000911, -0.00001476, -0.00000589, 0.00000699, 0.00001013, 0.00000110, -0.00000966, -0.00001076, -0.00000175, 0.00000740, 0.00000761, 0.00000012, -0.00000581, -0.00000334, 0.00000484, 0.00000986, 0.00000615, -0.00000320, -0.00001018, -0.00000985, -0.00000417, 0.00000130, 0.00000326, 0.00000275, 0.00000234, 0.00000264, 0.00000220, 0.00000018, -0.00000193, -0.00000182, 0.00000085, 0.00000371, 0.00000399, 0.00000109, -0.00000302, -0.00000595, -0.00000682, -0.00000626, -0.00000490, -0.00000254, 0.00000132, 0.00000628, 0.00001092, },
		{ -0.00000528, -0.00000122, 0.00000442, 0.00001010, 0.00001413, 0.00001490, 0.00001154, 0.00000487, -0.00000243, -0.00000719, -0.00000780, -0.00000504, -0.00000087, 0.00000323, 0.00000667, 0.00000853, 0.00000699, 0.00000099, -0.00000745, -0.00001337, -0.00001246, -0.00000522, 0.00000306, 0.00000687, 0.00000536, 0.00000221, 0.00000075, -0.00000011, -0.00000411, -0.00001174, -0.00001739, -0.00001397, -0.00000101, 0.00001271, 0.00001610, 0.00000618, -0.00000860, -0.00001610, -0.00001165, -0.00000139, 0.00000522, 0.00000522, 0.00000363, 0.00000553, 0.00000857, 0.00000510, -0.00000799, -0.00002328, -0.00002860, -0.00001909, -0.00000260, 0.00000824, 0.00000887, 0.00000635, 0.00000944, 0.00001658, 0.00001628, 0.00000085, -0.00002148, -0.00003116, -0.00001621, 0.00001357, 0.00003344, 0.00002646, -0.00000001, -0.00002122, -0.00001893, 0.00000161, 0.00001828, 0.00001512, -0.00000246, -0.00001519, -0.00001127, 0.00000175, 0.00000675, -0.00000297, -0.00001568, -0.00001420, 0.00000436, 0.00002365, 0.00002399, 0.00000274, -0.00002229, -0.00002991, -0.00001577, 0.00000461, 0.00001225, 0.00000303, -0.00001035, -0.00001309, -0.00000341, 0.00000728, 0.00000778, -0.00000110, -0.00000862, -0.00000651, 0.00000238, 0.00000785, 0.00000394, -0.00000485, -0.00000870, -0.00000302, 0.00000700, 0.00001196, 0.00000753, -0.00000211, -0.00000923, -0.00000983, -0.00000595, -0.00000191, 0.00000033, 0.00000180, 0.00000363, 0.00000495, 0.00000403, 0.00000092, -0.00000203, -0.00000235, -0.00000004, 0.00000261, 0.00000334, 0.00000195, -0.00000011, -0.00000169, -0.00000290, -0.00000444, -0.00000617, -0.00000686, },
		{ -0.00000075, -0.00000384, -0.00000682, -0.00000820, -0.00000710, -0.00000329, 0.00000272, 0.00000939, 0.00001425, 0.00001506, 0.00001144, 0.00000525, -0.00000085, -0.00000523, -0.00000754, -0.00000753, -0.00000426, 0.00000251, 0.00001038, 0.00001454, 0.00001128, 0.00000174, -0.00000832, -0.00001321, -0.00001193, -0.00000802, -0.00000490, -0.00000213, 0.00000280, 0.00000941, 0.00001225, 0.00000581, -0.00000829, -0.00002031, -0.00002021, -0.00000750, 0.00000738, 0.00001237, 0.00000503, -0.00000583, -0.00000961, -0.00000450, 0.00000199, 0.00000256, -0.00000137, -0.00000205, 0.00000433, 0.00001145, 0.00000900, -0.00000521, -0.00002094, -0.00002497, -0.00001483, -0.00000159, 0.00000220, -0.00000353, -0.00000623, 0.00000433, 0.00002222, 0.00002898, 0.00001266, -0.00001714, -0.00003554, -0.00002528, 0.00000603, 0.00003220, 0.00003253, 0.00001021, -0.00001285, -0.00001807, -0.00000651, 0.00000607, 0.00000778, 0.00000202, 0.00000029, 0.00000641, 0.00001100, 0.00000304, -0.00001508, -0.00002697, -0.00001841, 0.00000634, 0.00002644, 0.00002374, 0.00000036, -0.00002325, -0.00002798, -0.00001337, 0.00000394, 0.00000796, -0.00000159, -0.00001147, -0.00001030, -0.00000008, 0.00000752, 0.00000474, -0.00000419, -0.00000849, -0.00000304, 0.00000622, 0.00000911, 0.00000224, -0.00000754, -0.00001047, -0.00000370, 0.00000639, 0.00001101, 0.00000719, -0.00000071, -0.00000636, -0.00000730, -0.00000556, -0.00000386, -0.00000254, -0.00000033, 0.00000290, 0.00000525, 0.00000483, 0.00000207, -0.00000060, -0.00000120, 0.00000000, 0.00000124, 0.00000136, 0.00000086, 0.00000080, 0.00000122, 0.00000103, },
		{ 0.00000032, 0.00000197, 0.00000288, 0.00000203, -0.00000042, -0.00000365, -0.00000673, -0.00000864, -0.00000821, -0.00000463, 0.00000145, 0.00000784, 0.00001213, 0.00001330, 0.00001187, 0.00000850, 0.00000312, -0.00000397, -0.00001036, -0.00001189, -0.00000598, 0.00000511, 0.00001487, 0.00001743, 0.00001214, 0.00000333, -0.00000436, -0.00000994, -0.00001475, -0.00001817, -0.00001608, -0.00000540, 0.00001011, 0.00002017, 0.00001629, 0.00000051, -0.00001511, -0.00001892, -0.00001027, 0.00000054, 0.00000314, -0.00000256, -0.00000738, -0.00000393, 0.00000483, 0.00000902, 0.00000361, -0.00000520, -0.00000665, 0.00000182, 0.00001048, 0.00000753, -0.00000709, -0.00002000, -0.00001814, -0.00000349, 0.00000775, 0.00000224, -0.00001491, -0.00002368, -0.00000979, 0.00001869, 0.00003683, 0.00002652, -0.00000505, -0.00003105, -0.00002972, -0.00000381, 0.00002351, 0.00003109, 0.00001772, -0.00000088, -0.00000990, -0.00000825, -0.00000446, -0.00000343, -0.00000081, 0.00000877, 0.00002093, 0.00002276, 0.00000671, -0.00001821, -0.00003197, -0.00002214, 0.00000303, 0.00002135, 0.00001734, -0.00000369, -0.00002178, -0.00002207, -0.00000807, 0.00000411, 0.00000315, -0.00000696, -0.00001306, -0.00000784, 0.00000295, 0.00000772, 0.00000206, -0.00000689, -0.00000867, -0.00000092, 0.00000839, 0.00000967, 0.00000163, -0.00000783, -0.00001000, -0.00000367, 0.00000465, 0.00000788, 0.00000472, -0.00000063, -0.00000375, -0.00000400, -0.00000353, -0.00000373, -0.00000366, -0.00000183, 0.00000137, 0.00000379, 0.00000388, 0.00000227, 0.00000089, 0.00000075, 0.00000110, 0.00000080, -0.00000016, -0.00000059, },
		{ 0.00000100, -0.00000048, -0.00000093, 0.00000008, 0.00000166, 0.00000261, 0.00000250, 0.00000150, -0.00000031, -0.00000300, -0.00000610, -0.00000824, -0.00000800, -0.00000522, -0.00000101, 0.00000354, 0.00000823, 0.00001285, 0.00001561, 0.00001349, 0.00000516, -0.00000619, -0.00001393, -0.00001292, -0.00000388, 0.00000727, 0.00001461, 0.00001649, 0.00001471, 0.00001029, 0.00000170, -0.00001167, -0.00002497, -0.00002907, -0.00001843, 0.00000187, 0.00001832, 0.00002003, 0.00000836, -0.00000484, -0.00000909, -0.00000505, -0.00000214, -0.00000672, -0.00001421, -0.00001410, -0.00000257, 0.00001166, 0.00001576, 0.00000665, -0.00000511, -0.00000663, 0.00000266, 0.00000971, 0.00000244, -0.00001522, -0.00002536, -0.00001465, 0.00000955, 0.00002394, 0.00001205, -0.00001772, -0.00003810, -0.00002850, 0.00000474, 0.00003372, 0.00003476, 0.00000964, -0.00001783, -0.00002494, -0.00000949, 0.00001197, 0.00002224, 0.00001817, 0.00000852, 0.00000137, -0.00000329, -0.00000855, -0.00001207, -0.00000653, 0.00001017, 0.00002785, 0.00003064, 0.00001264, -0.00001394, -0.00002839, -0.00002058, -0.00000051, 0.00001167, 0.00000518, -0.00001179, -0.00002165, -0.00001555, -0.00000135, 0.00000591, -0.00000001, -0.00001098, -0.00001423, -0.00000613, 0.00000465, 0.00000714, -0.00000021, -0.00000850, -0.00000834, 0.00000031, 0.00000867, 0.00000866, 0.00000086, -0.00000685, -0.00000771, -0.00000229, 0.00000336, 0.00000437, 0.00000128, -0.00000192, -0.00000264, -0.00000176, -0.00000148, -0.00000235, -0.00000284, -0.00000166, 0.00000054, 0.00000199, 0.00000196, 0.00000137, 0.00000143, 0.00000207, 0.00000213, },
		{ 0.00000130, 0.00000220, 0.00000198, 0.00000072, -0.00000031, -0.00000008, 0.00000104, 0.00000187, 0.00000183, 0.00000141, 0.00000120, 0.00000093, -0.00000020, -0.00000233, -0.00000460, -0.00000617, -0.00000705, -0.00000744, -0.00000633, -0.00000174, 0.00000681, 0.00001597, 0.00001990, 0.00001491, 0.00000333, -0.00000783, -0.00001246, -0.00000979, -0.00000337, 0.00000359, 0.00001081, 0.00001835, 0.00002261, 0.00001741, 0.00000058, -0.00002065, -0.00003301, -0.00002800, -0.00001000, 0.00000719, 0.00001284, 0.00000834, 0.00000361, 0.00000488, 0.00000760, 0.00000250, -0.00001179, -0.00002443, -0.00002242, -0.00000506, 0.00001352, 0.00001787, 0.00000709, -0.00000432, -0.00000260, 0.00000935, 0.00001418, -0.00000045, -0.00002569, -0.00003728, -0.00001958, 0.00001624, 0.00003972, 0.00002888, -0.00000854, -0.00004132, -0.00004285, -0.00001422, 0.00001899, 0.00003132, 0.00001882, -0.00000173, -0.00001163, -0.00000654, 0.00000431, 0.00001097, 0.00001196, 0.00001116, 0.00000963, 0.00000407, -0.00000622, -0.00001398, -0.00000928, 0.00000885, 0.00002783, 0.00003143, 0.00001540, -0.00000742, -0.00001890, -0.00001286, -0.00000036, 0.00000249, -0.00000874, -0.00002268, -0.00002461, -0.00001190, 0.00000351, 0.00000772, -0.00000119, -0.00001218, -0.00001344, -0.00000433, 0.00000507, 0.00000537, -0.00000248, -0.00000914, -0.00000722, 0.00000124, 0.00000769, 0.00000628, -0.00000062, -0.00000580, -0.00000492, -0.00000020, 0.00000295, 0.00000183, -0.00000153, -0.00000350, -0.00000284, -0.00000125, -0.00000068, -0.00000113, -0.00000120, -0.00000026, 0.00000083, 0.00000102, 0.00000051, 0.00000044, },
		{ 0.00000083, 0.00000021, 0.00000058, 0.00000160, 0.00000188, 0.00000080, -0.00000062, -0.00000073, 0.00000079, 0.00000255, 0.00000298, 0.00000197, 0.00000070, 0.00000014, 0.00000010, -0.00000008, -0.00000045, -0.00000071, -0.00000137, -0.00000385, -0.00000843, -0.00001242, -0.00001133, -0.00000301, 0.00000918, 0.00001819, 0.00001892, 0.00001223, 0.00000342, -0.00000314, -0.00000704, -0.00000950, -0.00000919, -0.00000241, 0.00001137, 0.00002522, 0.00002815, 0.00001492, -0.00000745, -0.00002446, -0.00002649, -0.00001680, -0.00000671, -0.00000317, -0.00000229, 0.00000399, 0.00001578, 0.00002221, 0.00001250, -0.00001021, -0.00002861, -0.00002687, -0.00000690, 0.00001256, 0.00001458, 0.00000143, -0.00000810, 0.00000026, 0.00001910, 0.00002525, 0.00000438, -0.00003093, -0.00004985, -0.00003199, 0.00001150, 0.00004608, 0.00004408, 0.00000868, -0.00003054, -0.00004460, -0.00002856, -0.00000100, 0.00001632, 0.00001671, 0.00000896, 0.00000392, 0.00000393, 0.00000470, 0.00000386, 0.00000410, 0.00000774, 0.00001127, 0.00000828, -0.00000169, -0.00000985, -0.00000597, 0.00000992, 0.00002539, 0.00002682, 0.00001299, -0.00000346, -0.00000886, -0.00000240, 0.00000354, -0.00000283, -0.00001959, -0.00003242, -0.00002914, -0.00001216, 0.00000397, 0.00000690, -0.00000210, -0.00001104, -0.00001020, -0.00000145, 0.00000523, 0.00000306, -0.00000473, -0.00000924, -0.00000569, 0.00000199, 0.00000609, 0.00000329, -0.00000255, -0.00000524, -0.00000275, 0.00000147, 0.00000286, 0.00000048, -0.00000289, -0.00000425, -0.00000332, -0.00000187, -0.00000115, -0.00000080, 0.00000003, 0.00000110, 0.00000146, },
		{ 0.00000056, 0.00000156, 0.00000132, 0.00000048, 0.00000050, 0.00000178, 0.00000291, 0.00000220, -0.00000020, -0.00000217, -0.00000178, 0.00000071, 0.00000317, 0.00000385, 0.00000288, 0.00000146, 0.00000028, -0.00000084, -0.00000176, -0.00000145, 0.00000056, 0.00000250, 0.00000112, -0.00000491, -0.00001242, -0.00001548, -0.00001067, -0.00000042, 0.00000914, 0.00001375, 0.00001404, 0.00001286, 0.00001056, 0.00000448, -0.00000605, -0.00001557, -0.00001534, -0.00000211, 0.00001648, 0.00002699, 0.00002188, 0.00000636, -0.00000730, -0.00001205, -0.00001183, -0.00001456, -0.00002044, -0.00001991, -0.00000495, 0.00001841, 0.00003189, 0.00002186, -0.00000528, -0.00002672, -0.00002443, -0.00000311, 0.00001433, 0.00001026, -0.00000889, -0.00001901, -0.00000410, 0.00002475, 0.00003809, 0.00001725, -0.00002511, -0.00005421, -0.00004438, -0.00000248, 0.00003841, 0.00004824, 0.00002420, -0.00001103, -0.00003188, -0.00003025, -0.00001613, -0.00000309, 0.00000417, 0.00000892, 0.00001359, 0.00001552, 0.00001162, 0.00000442, 0.00000059, 0.00000318, 0.00000756, 0.00000669, 0.00000018, -0.00000400, 0.00000186, 0.00001543, 0.00002509, 0.00002149, 0.00000773, -0.00000335, -0.00000253, 0.00000574, 0.00000778, -0.00000466, -0.00002490, -0.00003717, -0.00003181, -0.00001442, 0.00000011, 0.00000210, -0.00000486, -0.00000963, -0.00000582, 0.00000237, 0.00000593, 0.00000132, -0.00000619, -0.00000869, -0.00000408, 0.00000236, 0.00000415, 0.00000030, -0.00000441, -0.00000510, -0.00000164, 0.00000201, 0.00000239, -0.00000020, -0.00000298, -0.00000394, -0.00000341, -0.00000265, -0.00000201, -0.00000095, },
		{ -0.00000318, -0.00000247, -0.00000077, 0.00000099, 0.00000136, 0.00000020, -0.00000073, 0.00000031, 0.00000287, 0.00000447, 0.00000324, 0.00000001, -0.00000241, -0.00000213, 0.00000013, 0.00000224, 0.00000303, 0.00000302, 0.00000288, 0.00000211, -0.00000008, -0.00000280, -0.00000355, -0.00000090, 0.00000299, 0.00000371, -0.00000099, -0.00000834, -0.00001284, -0.00001151, -0.00000645, -0.00000175, 0.00000161, 0.00000639, 0.00001414, 0.00002075, 0.00001879, 0.00000570, -0.00001119, -0.00001957, -0.00001318, 0.00000202, 0.00001378, 0.00001546, 0.00001128, 0.00000921, 0.00000989, 0.00000520, -0.00001076, -0.00003052, -0.00003657, -0.00001854, 0.00001306, 0.00003314, 0.00002489, -0.00000302, -0.00002356, -0.00001749, 0.00000722, 0.00002319, 0.00001167, -0.00001720, -0.00003346, -0.00001730, 0.00001967, 0.00004395, 0.00003154, -0.00000914, -0.00004477, -0.00004710, -0.00001675, 0.00002038, 0.00003761, 0.00002850, 0.00000660, -0.00001115, -0.00001838, -0.00001887, -0.00001680, -0.00001128, -0.00000030, 0.00001295, 0.00002084, 0.00001883, 0.00001053, 0.00000386, 0.00000262, 0.00000336, 0.00000131, -0.00000230, -0.00000068, 0.00000984, 0.00002331, 0.00002827, 0.00001938, 0.00000392, -0.00000471, -0.00000059, 0.00000836, 0.00000832, -0.00000609, -0.00002580, -0.00003593, -0.00003005, -0.00001557, -0.00000546, -0.00000548, -0.00001001, -0.00000984, -0.00000254, 0.00000552, 0.00000676, 0.00000065, -0.00000609, -0.00000697, -0.00000218, 0.00000244, 0.00000207, -0.00000236, -0.00000592, -0.00000525, -0.00000159, 0.00000131, 0.00000129, -0.00000067, -0.00000242, -0.00000309, -0.00000321, },
		{ -0.00000161, -0.00000223, -0.00000319, -0.00000370, -0.00000254, 0.00000001, 0.00000189, 0.00000130, -0.00000103, -0.00000230, -0.00000065, 0.00000272, 0.00000472, 0.00000368, 0.00000098, -0.00000067, -0.00000029, 0.00000077, 0.00000092, 0.00000053, 0.00000117, 0.00000302, 0.00000402, 0.00000215, -0.00000166, -0.00000396, -0.00000249, 0.00000090, 0.00000209, -0.00000075, -0.00000496, -0.00000663, -0.00000542, -0.00000497, -0.00000785, -0.00001081, -0.00000701, 0.00000600, 0.00002131, 0.00002729, 0.00001846, 0.00000128, -0.00001167, -0.00001356, -0.00000843, -0.00000440, -0.00000295, 0.00000239, 0.00001584, 0.00002953, 0.00002775, 0.00000361, -0.00002922, -0.00004496, -0.00002924, 0.00000517, 0.00002819, 0.00002050, -0.00000743, -0.00002517, -0.00001265, 0.00001908, 0.00003793, 0.00002255, -0.00001576, -0.00004281, -0.00003383, 0.00000336, 0.00003614, 0.00003735, 0.00000865, -0.00002394, -0.00003516, -0.00002115, 0.00000224, 0.00001724, 0.00001874, 0.00001297, 0.00000655, -0.00000018, -0.00000926, -0.00001821, -0.00002013, -0.00001108, 0.00000443, 0.00001676, 0.00001993, 0.00001607, 0.00001106, 0.00000744, 0.00000309, -0.00000309, -0.00000659, -0.00000082, 0.00001441, 0.00002956, 0.00003270, 0.00002104, 0.00000435, -0.00000416, -0.00000068, 0.00000587, 0.00000355, -0.00001007, -0.00002543, -0.00003088, -0.00002417, -0.00001384, -0.00000966, -0.00001279, -0.00001582, -0.00001177, -0.00000184, 0.00000615, 0.00000625, 0.00000031, -0.00000465, -0.00000397, 0.00000036, 0.00000280, 0.00000047, -0.00000424, -0.00000689, -0.00000568, -0.00000257, -0.00000049, -0.00000034, -0.00000104, },
		{ -0.00000330, -0.00000241, -0.00000133, -0.00000068, -0.00000153, -0.00000351, -0.00000445, -0.00000264, 0.00000085, 0.00000292, 0.00000167, -0.00000139, -0.00000292, -0.00000135, 0.00000152, 0.00000281, 0.00000191, 0.00000095, 0.00000184, 0.00000372, 0.00000387, 0.00000129, -0.00000170, -0.00000193, 0.00000093, 0.00000375, 0.00000350, 0.00000066, -0.00000155, -0.00000132, -0.00000048, -0.00000190, -0.00000509, -0.00000605, -0.00000220, 0.00000308, 0.00000274, -0.00000600, -0.00001670, -0.00001842, -0.00000641, 0.00001218, 0.00002434, 0.00002321, 0.00001341, 0.00000427, -0.00000092, -0.00000694, -0.00001729, -0.00002515, -0.00001765, 0.00000873, 0.00003836, 0.00004630, 0.00002162, -0.00001889, -0.00004292, -0.00003182, 0.00000123, 0.00002321, 0.00001322, -0.00001683, -0.00003367, -0.00001581, 0.00002431, 0.00005119, 0.00003963, -0.00000174, -0.00003886, -0.00004347, -0.00001670, 0.00001560, 0.00002829, 0.00001744, -0.00000163, -0.00001242, -0.00001134, -0.00000563, -0.00000179, 0.00000096, 0.00000620, 0.00001258, 0.00001357, 0.00000506, -0.00000850, -0.00001723, -0.00001494, -0.00000436, 0.00000670, 0.00001318, 0.00001566, 0.00001628, 0.00001422, 0.00000717, -0.00000297, -0.00000837, -0.00000166, 0.00001553, 0.00003176, 0.00003510, 0.00002400, 0.00000839, -0.00000035, -0.00000016, 0.00000104, -0.00000488, -0.00001696, -0.00002618, -0.00002540, -0.00001697, -0.00001015, -0.00001123, -0.00001721, -0.00001959, -0.00001364, -0.00000321, 0.00000374, 0.00000328, -0.00000111, -0.00000320, -0.00000063, 0.00000325, 0.00000381, 0.00000014, -0.00000458, -0.00000687, -0.00000622, -0.00000452, },
		{ -0.00000513, -0.00000618, -0.00000686, -0.00000678, -0.00000494, -0.00000185, 0.00000005, -0.00000104, -0.00000388, -0.00000511, -0.00000285, 0.00000110, 0.00000312, 0.00000173, -0.00000092, -0.00000172, -0.00000030, 0.00000081, -0.00000036, -0.00000231, -0.00000161, 0.00000282, 0.00000774, 0.00000860, 0.00000438, -0.00000119, -0.00000355, -0.00000190, 0.00000075, 0.00000163, 0.00000128, 0.00000184, 0.00000303, 0.00000158, -0.00000435, -0.00001102, -0.00001141, -0.00000288, 0.00000813, 0.00001093, 0.00000115, -0.00001380, -0.00002107, -0.00001455, -0.00000017, 0.00001116, 0.00001500, 0.00001596, 0.00001882, 0.00001939, 0.00000781, -0.00001739, -0.00004078, -0.00004015, -0.00000860, 0.00003377, 0.00005383, 0.00003491, -0.00000640, -0.00003423, -0.00002662, 0.00000401, 0.00002347, 0.00000985, -0.00002455, -0.00004477, -0.00002718, 0.00001751, 0.00005380, 0.00005351, 0.00001945, -0.00002011, -0.00003795, -0.00002896, -0.00000832, 0.00000676, 0.00001129, 0.00001079, 0.00001016, 0.00000777, 0.00000032, -0.00000955, -0.00001371, -0.00000698, 0.00000632, 0.00001564, 0.00001407, 0.00000430, -0.00000514, -0.00000835, -0.00000602, -0.00000182, 0.00000280, 0.00000880, 0.00001526, 0.00001761, 0.00001194, 0.00000117, -0.00000528, 0.00000040, 0.00001619, 0.00003095, 0.00003428, 0.00002551, 0.00001303, 0.00000495, 0.00000137, -0.00000373, -0.00001379, -0.00002451, -0.00002795, -0.00002152, -0.00001150, -0.00000709, -0.00001128, -0.00001823, -0.00001970, -0.00001341, -0.00000469, -0.00000031, -0.00000164, -0.00000419, -0.00000320, 0.00000128, 0.00000512, 0.00000493, 0.00000127, -0.00000277, },
		{ 0.00000464, 0.00000278, 0.00000072, -0.00000147, -0.00000471, -0.00000837, -0.00000988, -0.00000752, -0.00000296, -0.00000015, -0.00000122, -0.00000416, -0.00000507, -0.00000256, 0.00000087, 0.00000188, 0.00000032, -0.00000080, 0.00000064, 0.00000286, 0.00000208, -0.00000258, -0.00000708, -0.00000617, 0.00000087, 0.00000895, 0.00001196, 0.00000855, 0.00000269, -0.00000112, -0.00000221, -0.00000270, -0.00000335, -0.00000200, 0.00000280, 0.00000775, 0.00000660, -0.00000279, -0.00001404, -0.00001684, -0.00000733, 0.00000693, 0.00001340, 0.00000675, -0.00000613, -0.00001375, -0.00001191, -0.00000654, -0.00000452, -0.00000372, 0.00000463, 0.00002265, 0.00003730, 0.00002970, -0.00000414, -0.00004308, -0.00005542, -0.00002773, 0.00001988, 0.00004938, 0.00003865, 0.00000162, -0.00002550, -0.00001891, 0.00001052, 0.00002911, 0.00001416, -0.00002359, -0.00005008, -0.00003997, 0.00000068, 0.00004132, 0.00005441, 0.00003654, 0.00000589, -0.00001716, -0.00002527, -0.00002315, -0.00001658, -0.00000629, 0.00000806, 0.00002146, 0.00002493, 0.00001433, -0.00000379, -0.00001660, -0.00001596, -0.00000483, 0.00000698, 0.00001223, 0.00001114, 0.00000791, 0.00000490, 0.00000139, -0.00000265, -0.00000412, 0.00000021, 0.00000870, 0.00001469, 0.00001285, 0.00000545, 0.00000113, 0.00000656, 0.00001945, 0.00003050, 0.00003217, 0.00002510, 0.00001572, 0.00000850, 0.00000177, -0.00000833, -0.00002109, -0.00002993, -0.00002874, -0.00001890, -0.00000894, -0.00000664, -0.00001186, -0.00001741, -0.00001683, -0.00001060, -0.00000471, -0.00000376, -0.00000647, -0.00000781, -0.00000483, 0.00000058, 0.00000436, },
		{ -0.00000659, -0.00000219, 0.00000054, 0.00000166, 0.00000283, 0.00000401, 0.00000293, -0.00000180, -0.00000790, -0.00001094, -0.00000886, -0.00000429, -0.00000179, -0.00000293, -0.00000492, -0.00000426, -0.00000109, 0.00000105, -0.00000021, -0.00000275, -0.00000225, 0.00000232, 0.00000660, 0.00000512, -0.00000231, -0.00000958, -0.00000997, -0.00000280, 0.00000620, 0.00001088, 0.00001019, 0.00000748, 0.00000525, 0.00000237, -0.00000286, -0.00000818, -0.00000813, -0.00000056, 0.00000888, 0.00001070, 0.00000131, -0.00001198, -0.00001735, -0.00001023, 0.00000202, 0.00000804, 0.00000444, -0.00000153, -0.00000185, 0.00000167, -0.00000066, -0.00001291, -0.00002423, -0.00001724, 0.00001196, 0.00004367, 0.00004867, 0.00001611, -0.00003218, -0.00005795, -0.00004040, 0.00000431, 0.00003755, 0.00003373, 0.00000281, -0.00002133, -0.00001467, 0.00001441, 0.00003476, 0.00002356, -0.00001215, -0.00004298, -0.00004444, -0.00001696, 0.00001812, 0.00003907, 0.00003938, 0.00002643, 0.00000967, -0.00000677, -0.00002128, -0.00002917, -0.00002400, -0.00000520, 0.00001753, 0.00003011, 0.00002538, 0.00000879, -0.00000718, -0.00001369, -0.00001094, -0.00000439, 0.00000207, 0.00000821, 0.00001400, 0.00001672, 0.00001321, 0.00000461, -0.00000298, -0.00000383, 0.00000166, 0.00000741, 0.00000836, 0.00000604, 0.00000679, 0.00001446, 0.00002571, 0.00003307, 0.00003210, 0.00002488, 0.00001631, 0.00000825, -0.00000153, -0.00001452, -0.00002701, -0.00003217, -0.00002690, -0.00001598, -0.00000829, -0.00000869, -0.00001389, -0.00001656, -0.00001308, -0.00000674, -0.00000347, -0.00000540, -0.00000909, -0.00000984, },
		{ -0.00000598, -0.00000910, -0.00000891, -0.00000650, -0.00000490, -0.00000473, -0.00000355, 0.00000036, 0.00000484, 0.00000555, 0.00000092, -0.00000558, -0.00000878, -0.00000723, -0.00000440, -0.00000426, -0.00000651, -0.00000711, -0.00000362, 0.00000124, 0.00000232, -0.00000173, -0.00000615, -0.00000500, 0.00000201, 0.00000843, 0.00000759, -0.00000032, -0.00000834, -0.00000982, -0.00000473, 0.00000156, 0.00000492, 0.00000637, 0.00000897, 0.00001202, 0.00001071, 0.00000228, -0.00000857, -0.00001275, -0.00000617, 0.00000490, 0.00000943, 0.00000305, -0.00000735, -0.00001125, -0.00000621, -0.00000019, -0.00000134, -0.00000714, -0.00000655, 0.00000581, 0.00001960, 0.00001683, -0.00000788, -0.00003640, -0.00004049, -0.00000950, 0.00003521, 0.00005628, 0.00003400, -0.00001425, -0.00004861, -0.00004268, -0.00000647, 0.00002489, 0.00002505, -0.00000020, -0.00002153, -0.00001615, 0.00001086, 0.00003345, 0.00003047, 0.00000430, -0.00002391, -0.00003460, -0.00002459, -0.00000454, 0.00001382, 0.00002586, 0.00003151, 0.00002922, 0.00001628, -0.00000511, -0.00002466, -0.00003008, -0.00001756, 0.00000427, 0.00002125, 0.00002496, 0.00001756, 0.00000690, -0.00000142, -0.00000667, -0.00000945, -0.00000818, -0.00000075, 0.00001106, 0.00002070, 0.00002192, 0.00001464, 0.00000508, -0.00000030, -0.00000043, 0.00000103, 0.00000135, 0.00000249, 0.00000860, 0.00002027, 0.00003244, 0.00003849, 0.00003590, 0.00002736, 0.00001663, 0.00000468, -0.00000932, -0.00002372, -0.00003306, -0.00003235, -0.00002279, -0.00001209, -0.00000801, -0.00001146, -0.00001629, -0.00001597, -0.00001006, -0.00000399, -0.00000276, },
		{ -0.00000752, -0.00000315, -0.00000408, -0.00000727, -0.00000808, -0.00000582, -0.00000407, -0.00000563, -0.00000863, -0.00000849, -0.00000359, 0.00000234, 0.00000407, 0.00000070, -0.00000346, -0.00000405, -0.00000184, -0.00000152, -0.00000566, -0.00001077, -0.00001084, -0.00000443, 0.00000282, 0.00000390, -0.00000193, -0.00000793, -0.00000707, 0.00000035, 0.00000703, 0.00000665, 0.00000041, -0.00000491, -0.00000484, -0.00000175, -0.00000080, -0.00000269, -0.00000232, 0.00000449, 0.00001427, 0.00001796, 0.00001046, -0.00000289, -0.00001104, -0.00000859, -0.00000112, 0.00000169, -0.00000247, -0.00000616, -0.00000215, 0.00000593, 0.00000613, -0.00000718, -0.00002301, -0.00002202, 0.00000237, 0.00003226, 0.00003860, 0.00000953, -0.00003427, -0.00005515, -0.00003275, 0.00001576, 0.00004991, 0.00004226, 0.00000252, -0.00003314, -0.00003616, -0.00001027, 0.00001578, 0.00001764, -0.00000264, -0.00002195, -0.00002045, 0.00000036, 0.00002190, 0.00002751, 0.00001609, -0.00000104, -0.00001299, -0.00001698, -0.00001533, -0.00000883, 0.00000377, 0.00002035, 0.00003250, 0.00003089, 0.00001444, -0.00000700, -0.00002004, -0.00001848, -0.00000685, 0.00000552, 0.00001292, 0.00001560, 0.00001562, 0.00001265, 0.00000527, -0.00000462, -0.00001084, -0.00000777, 0.00000380, 0.00001656, 0.00002281, 0.00002070, 0.00001429, 0.00000834, 0.00000402, 0.00000005, -0.00000313, -0.00000191, 0.00000698, 0.00002195, 0.00003661, 0.00004430, 0.00004256, 0.00003302, 0.00001819, -0.00000033, -0.00001998, -0.00003527, -0.00003988, -0.00003223, -0.00001846, -0.00000850, -0.00000790, -0.00001348, -0.00001720, -0.00001431, },
		{ -0.00001503, -0.00001589, -0.00001032, -0.00000405, -0.00000307, -0.00000698, -0.00001000, -0.00000796, -0.00000309, -0.00000114, -0.00000451, -0.00000936, -0.00001013, -0.00000603, -0.00000177, -0.00000166, -0.00000433, -0.00000441, 0.00000045, 0.00000575, 0.00000464, -0.00000399, -0.00001311, -0.00001432, -0.00000672, 0.00000201, 0.00000368, -0.00000190, -0.00000732, -0.00000619, -0.00000014, 0.00000368, 0.00000142, -0.00000302, -0.00000298, 0.00000229, 0.00000609, 0.00000219, -0.00000687, -0.00001121, -0.00000419, 0.00000962, 0.00001850, 0.00001556, 0.00000545, -0.00000181, -0.00000253, -0.00000262, -0.00000836, -0.00001592, -0.00001377, 0.00000281, 0.00002154, 0.00002189, -0.00000307, -0.00003463, -0.00004220, -0.00001282, 0.00003254, 0.00005472, 0.00003184, -0.00001900, -0.00005565, -0.00004851, -0.00000637, 0.00003323, 0.00003896, 0.00001225, -0.00001820, -0.00002572, -0.00000910, 0.00001064, 0.00001343, -0.00000151, -0.00001831, -0.00002199, -0.00001140, 0.00000340, 0.00001299, 0.00001571, 0.00001446, 0.00001023, 0.00000156, -0.00001023, -0.00001806, -0.00001413, 0.00000185, 0.00002045, 0.00002975, 0.00002522, 0.00001241, 0.00000056, -0.00000542, -0.00000671, -0.00000587, -0.00000269, 0.00000442, 0.00001382, 0.00001965, 0.00001668, 0.00000616, -0.00000423, -0.00000680, -0.00000033, 0.00000980, 0.00001722, 0.00001970, 0.00001885, 0.00001635, 0.00001178, 0.00000479, -0.00000188, -0.00000290, 0.00000532, 0.00002096, 0.00003758, 0.00004838, 0.00004941, 0.00003974, 0.00002025, -0.00000575, -0.00003121, -0.00004670, -0.00004612, -0.00003221, -0.00001569, -0.00000732, -0.00000938, },
		{ -0.00001442, -0.00000947, -0.00001390, -0.00001769, -0.00001371, -0.00000462, 0.00000093, -0.00000194, -0.00000920, -0.00001265, -0.00000895, -0.00000270, -0.00000066, -0.00000405, -0.00000782, -0.00000735, -0.00000424, -0.00000409, -0.00000887, -0.00001319, -0.00000981, 0.00000144, 0.00001177, 0.00001166, 0.00000080, -0.00001122, -0.00001444, -0.00000831, -0.00000108, -0.00000011, -0.00000411, -0.00000594, -0.00000204, 0.00000282, 0.00000157, -0.00000576, -0.00001067, -0.00000551, 0.00000684, 0.00001465, 0.00000942, -0.00000460, -0.00001426, -0.00001098, 0.00000096, 0.00001026, 0.00001134, 0.00000912, 0.00001038, 0.00001275, 0.00000618, -0.00001303, -0.00003214, -0.00003067, -0.00000241, 0.00003247, 0.00004181, 0.00001200, -0.00003498, -0.00005799, -0.00003383, 0.00001972, 0.00005809, 0.00004937, 0.00000225, -0.00004313, -0.00005176, -0.00002294, 0.00001354, 0.00002739, 0.00001368, -0.00000802, -0.00001608, -0.00000660, 0.00000717, 0.00001092, 0.00000243, -0.00000974, -0.00001699, -0.00001756, -0.00001402, -0.00000748, 0.00000296, 0.00001554, 0.00002346, 0.00001971, 0.00000479, -0.00001170, -0.00001840, -0.00001154, 0.00000292, 0.00001568, 0.00002187, 0.00002258, 0.00002027, 0.00001489, 0.00000548, -0.00000549, -0.00001158, -0.00000764, 0.00000449, 0.00001644, 0.00002025, 0.00001498, 0.00000663, 0.00000198, 0.00000304, 0.00000709, 0.00001101, 0.00001416, 0.00001720, 0.00001921, 0.00001767, 0.00001159, 0.00000413, 0.00000119, 0.00000698, 0.00002086, 0.00003781, 0.00005115, 0.00005481, 0.00004463, 0.00002020, -0.00001266, -0.00004208, -0.00005577, -0.00004948, -0.00003094, },
		{ -0.00006225, -0.00004888, -0.00002696, -0.00001309, -0.00001367, -0.00002098, -0.00002225, -0.00001274, 0.00000027, 0.00000573, 0.00000036, -0.00000875, -0.00001241, -0.00000883, -0.00000413, -0.00000402, -0.00000704, -0.00000715, -0.00000231, 0.00000173, -0.00000226, -0.00001347, -0.00002183, -0.00001762, -0.00000213, 0.00001256, 0.00001510, 0.00000564, -0.00000543, -0.00000894, -0.00000578, -0.00000372, -0.00000680, -0.00001020, -0.00000674, 0.00000283, 0.00000881, 0.00000302, -0.00001074, -0.00001885, -0.00001168, 0.00000573, 0.00001812, 0.00001532, 0.00000197, -0.00000859, -0.00000901, -0.00000418, -0.00000258, -0.00000384, 0.00000123, 0.00001690, 0.00003156, 0.00002578, -0.00000588, -0.00004244, -0.00005109, -0.00001847, 0.00003201, 0.00005715, 0.00003264, -0.00002276, -0.00006207, -0.00005153, -0.00000049, 0.00004786, 0.00005530, 0.00002047, -0.00002406, -0.00004347, -0.00002962, -0.00000235, 0.00001332, 0.00000937, -0.00000319, -0.00000951, -0.00000521, 0.00000309, 0.00000743, 0.00000574, 0.00000050, -0.00000643, -0.00001488, -0.00002277, -0.00002437, -0.00001475, 0.00000374, 0.00002097, 0.00002640, 0.00001781, 0.00000261, -0.00000909, -0.00001227, -0.00000842, -0.00000097, 0.00000871, 0.00002006, 0.00002975, 0.00003171, 0.00002240, 0.00000573, -0.00000843, -0.00001176, -0.00000409, 0.00000737, 0.00001490, 0.00001625, 0.00001443, 0.00001292, 0.00001227, 0.00001118, 0.00000964, 0.00000965, 0.00001264, 0.00001694, 0.00001881, 0.00001614, 0.00001094, 0.00000821, 0.00001232, 0.00002422, 0.00004075, 0.00005534, 0.00005956, 0.00004647, 0.00001570, -0.00002314, -0.00005358, },
		{ 0.00000676, -0.00003796, -0.00006690, -0.00006752, -0.00004554, -0.00002068, -0.00001077, -0.00001769, -0.00002802, -0.00002722, -0.00001351, 0.00000191, 0.00000722, 0.00000150, -0.00000616, -0.00000780, -0.00000480, -0.00000439, -0.00000937, -0.00001377, -0.00000997, 0.00000092, 0.00000824, 0.00000254, -0.00001331, -0.00002558, -0.00002286, -0.00000740, 0.00000756, 0.00001137, 0.00000574, 0.00000047, 0.00000103, 0.00000270, -0.00000214, -0.00001272, -0.00001862, -0.00001129, 0.00000462, 0.00001408, 0.00000686, -0.00001096, -0.00002235, -0.00001647, 0.00000065, 0.00001336, 0.00001309, 0.00000582, 0.00000246, 0.00000433, 0.00000187, -0.00001106, -0.00002457, -0.00001955, 0.00001012, 0.00004412, 0.00004972, 0.00001368, -0.00003968, -0.00006574, -0.00003952, 0.00001909, 0.00006075, 0.00005003, -0.00000308, -0.00005280, -0.00005858, -0.00001935, 0.00002922, 0.00004859, 0.00002932, -0.00000640, -0.00002906, -0.00002703, -0.00001092, 0.00000180, 0.00000329, -0.00000189, -0.00000572, -0.00000509, -0.00000186, 0.00000182, 0.00000537, 0.00000748, 0.00000454, -0.00000596, -0.00002059, -0.00003029, -0.00002724, -0.00001202, 0.00000637, 0.00001790, 0.00001891, 0.00001278, 0.00000475, -0.00000252, -0.00000822, -0.00001003, -0.00000417, 0.00001028, 0.00002760, 0.00003788, 0.00003458, 0.00002017, 0.00000412, -0.00000483, -0.00000471, 0.00000041, 0.00000594, 0.00001058, 0.00001526, 0.00001975, 0.00002173, 0.00001939, 0.00001418, 0.00001009, 0.00001008, 0.00001340, 0.00001664, 0.00001708, 0.00001518, 0.00001433, 0.00001870, 0.00003056, 0.00004796, 0.00006344, 0.00006581, 0.00004635, },
		{ 0.00007525, 0.00007434, 0.00004577, -0.00000468, -0.00005529, -0.00008120, -0.00007240, -0.00004174, -0.00001470, -0.00000880, -0.00002070, -0.00003236, -0.00002943, -0.00001391, 0.00000041, 0.00000347, -0.00000166, -0.00000458, -0.00000072, 0.00000350, -0.00000074, -0.00001262, -0.00002080, -0.00001549, -0.00000024, 0.00001042, 0.00000558, -0.00001052, -0.00002280, -0.00002086, -0.00000883, 0.00000064, 0.00000098, -0.00000197, 0.00000123, 0.00001051, 0.00001489, 0.00000520, -0.00001339, -0.00002452, -0.00001723, 0.00000169, 0.00001389, 0.00000815, -0.00000835, -0.00001845, -0.00001365, -0.00000198, 0.00000315, -0.00000028, -0.00000086, 0.00001024, 0.00002442, 0.00002156, -0.00000648, -0.00004042, -0.00004688, -0.00001135, 0.00004200, 0.00006745, 0.00003900, -0.00002295, -0.00006706, -0.00005623, -0.00000081, 0.00005112, 0.00005693, 0.00001564, -0.00003471, -0.00005306, -0.00002982, 0.00001011, 0.00003363, 0.00002755, 0.00000424, -0.00001501, -0.00001924, -0.00001254, -0.00000480, -0.00000150, -0.00000188, -0.00000341, -0.00000506, -0.00000608, -0.00000445, 0.00000121, 0.00000823, 0.00001026, 0.00000277, -0.00001170, -0.00002454, -0.00002805, -0.00002142, -0.00001012, -0.00000009, 0.00000659, 0.00001061, 0.00001169, 0.00000794, -0.00000055, -0.00000870, -0.00000891, 0.00000229, 0.00002002, 0.00003416, 0.00003730, 0.00002992, 0.00001851, 0.00000922, 0.00000382, 0.00000108, 0.00000053, 0.00000359, 0.00001105, 0.00002037, 0.00002667, 0.00002658, 0.00002119, 0.00001493, 0.00001181, 0.00001260, 0.00001513, 0.00001674, 0.00001675, 0.00001745, 0.00002319, 0.00003736, 0.00005803, },
		{ 0.00002559, 0.00004317, 0.00006859, 0.00008801, 0.00008291, 0.00004404, -0.00001721, -0.00007211, -0.00009333, -0.00007528, -0.00003828, -0.00001167, -0.00001018, -0.00002452, -0.00003381, -0.00002679, -0.00001083, -0.00000126, -0.00000394, -0.00000992, -0.00000721, 0.00000442, 0.00001251, 0.00000612, -0.00001095, -0.00002308, -0.00001904, -0.00000401, 0.00000614, 0.00000181, -0.00001060, -0.00001684, -0.00001115, -0.00000220, -0.00000198, -0.00001068, -0.00001556, -0.00000576, 0.00001351, 0.00002441, 0.00001497, -0.00000739, -0.00002263, -0.00001856, -0.00000283, 0.00000653, 0.00000124, -0.00000914, -0.00001006, -0.00000051, 0.00000518, -0.00000446, -0.00002081, -0.00002110, 0.00000570, 0.00004094, 0.00004924, 0.00001377, -0.00004164, -0.00006925, -0.00004079, 0.00002349, 0.00007005, 0.00005935, 0.00000139, -0.00005373, -0.00006096, -0.00001876, 0.00003309, 0.00005142, 0.00002630, -0.00001550, -0.00003863, -0.00002960, -0.00000294, 0.00001715, 0.00001845, 0.00000650, -0.00000585, -0.00001103, -0.00000994, -0.00000674, -0.00000386, -0.00000192, -0.00000178, -0.00000426, -0.00000784, -0.00000860, -0.00000393, 0.00000381, 0.00000846, 0.00000554, -0.00000350, -0.00001293, -0.00001823, -0.00001933, -0.00001860, -0.00001668, -0.00001174, -0.00000306, 0.00000585, 0.00000921, 0.00000477, -0.00000287, -0.00000555, 0.00000128, 0.00001466, 0.00002709, 0.00003305, 0.00003247, 0.00002858, 0.00002351, 0.00001703, 0.00000907, 0.00000229, 0.00000098, 0.00000710, 0.00001771, 0.00002679, 0.00002971, 0.00002636, 0.00002042, 0.00001611, 0.00001533, 0.00001696, 0.00001842, 0.00001831, 0.00001885, },
		{ 0.00002297, 0.00002200, 0.00002138, 0.00002870, 0.00004972, 0.00007939, 0.00009927, 0.00008771, 0.00003776, -0.00003218, -0.00008715, -0.00010019, -0.00007286, -0.00003313, -0.00001155, -0.00001671, -0.00003195, -0.00003506, -0.00002067, -0.00000335, 0.00000012, -0.00001068, -0.00001996, -0.00001380, 0.00000421, 0.00001669, 0.00001132, -0.00000602, -0.00001814, -0.00001526, -0.00000479, -0.00000149, -0.00000969, -0.00001828, -0.00001418, 0.00000107, 0.00001131, 0.00000350, -0.00001640, -0.00002789, -0.00001717, 0.00000768, 0.00002446, 0.00001949, 0.00000093, -0.00001168, -0.00000927, -0.00000176, -0.00000392, -0.00001553, -0.00002002, -0.00000500, 0.00001893, 0.00002500, -0.00000066, -0.00003864, -0.00004999, -0.00001464, 0.00004343, 0.00007329, 0.00004364, -0.00002514, -0.00007616, -0.00006628, -0.00000481, 0.00005543, 0.00006571, 0.00002298, -0.00003152, -0.00005203, -0.00002736, 0.00001478, 0.00003739, 0.00002658, -0.00000180, -0.00002189, -0.00002135, -0.00000747, 0.00000478, 0.00000733, 0.00000264, -0.00000283, -0.00000568, -0.00000629, -0.00000572, -0.00000405, -0.00000178, -0.00000091, -0.00000305, -0.00000691, -0.00000885, -0.00000657, -0.00000179, 0.00000146, 0.00000102, -0.00000151, -0.00000347, -0.00000487, -0.00000846, -0.00001561, -0.00002314, -0.00002532, -0.00001951, -0.00000931, -0.00000158, 0.00000001, -0.00000188, -0.00000140, 0.00000457, 0.00001396, 0.00002245, 0.00002786, 0.00003119, 0.00003366, 0.00003400, 0.00002957, 0.00002010, 0.00000971, 0.00000440, 0.00000735, 0.00001633, 0.00002555, 0.00002995, 0.00002837, 0.00002358, 0.00001970, 0.00001924, 0.00002139, },
		{ 0.00002146, 0.00002208, 0.00002566, 0.00002802, 0.00002674, 0.00002599, 0.00003541, 0.00006075, 0.00009328, 0.00010958, 0.00008676, 0.00002335, -0.00005310, -0.00010248, -0.00010183, -0.00006332, -0.00002329, -0.00001124, -0.00002647, -0.00004339, -0.00003903, -0.00001530, 0.00000530, 0.00000448, -0.00001313, -0.00002616, -0.00001966, -0.00000017, 0.00001290, 0.00000889, -0.00000377, -0.00000907, -0.00000287, 0.00000261, -0.00000564, -0.00002377, -0.00003284, -0.00001931, 0.00000799, 0.00002446, 0.00001436, -0.00001258, -0.00003069, -0.00002427, -0.00000237, 0.00001300, 0.00001124, 0.00000260, 0.00000283, 0.00001169, 0.00001187, -0.00000874, -0.00003703, -0.00004292, -0.00001145, 0.00003512, 0.00005274, 0.00001815, -0.00004335, -0.00007624, -0.00004555, 0.00002767, 0.00008219, 0.00007087, 0.00000310, -0.00006407, -0.00007706, -0.00003120, 0.00002973, 0.00005572, 0.00003308, -0.00000998, -0.00003432, -0.00002459, 0.00000305, 0.00002180, 0.00001925, 0.00000378, -0.00000863, -0.00001023, -0.00000511, -0.00000111, -0.00000136, -0.00000349, -0.00000450, -0.00000416, -0.00000376, -0.00000361, -0.00000290, -0.00000161, -0.00000114, -0.00000253, -0.00000478, -0.00000581, -0.00000516, -0.00000463, -0.00000579, -0.00000735, -0.00000609, -0.00000101, 0.00000410, 0.00000338, -0.00000547, -0.00001820, -0.00002739, -0.00002873, -0.00002409, -0.00001845, -0.00001455, -0.00001091, -0.00000471, 0.00000424, 0.00001323, 0.00001991, 0.00002490, 0.00003055, 0.00003722, 0.00004168, 0.00003978, 0.00003084, 0.00001927, 0.00001161, 0.00001174, 0.00001833, 0.00002629, 0.00003056, 0.00002933, 0.00002485, },
		{ 0.00003297, 0.00003162, 0.00002698, 0.00002365, 0.00002480, 0.00002881, 0.00003084, 0.00002913, 0.00003022, 0.00004543, 0.00007829, 0.00011370, 0.00012228, 0.00008202, 0.00000079, -0.00008204, -0.00012164, -0.00010209, -0.00004919, -0.00000856, -0.00000747, -0.00003452, -0.00005485, -0.00004510, -0.00001404, 0.00000956, 0.00000724, -0.00001224, -0.00002508, -0.00001865, -0.00000291, 0.00000379, -0.00000314, -0.00001011, -0.00000320, 0.00001287, 0.00001781, -0.00000118, -0.00003202, -0.00004685, -0.00002993, 0.00000518, 0.00002812, 0.00002147, -0.00000342, -0.00002073, -0.00001792, -0.00000633, -0.00000392, -0.00001138, -0.00001080, 0.00001033, 0.00003764, 0.00003923, 0.00000043, -0.00005290, -0.00007210, -0.00003221, 0.00003816, 0.00007750, 0.00004713, -0.00003028, -0.00008812, -0.00007506, -0.00000129, 0.00007105, 0.00008348, 0.00003090, -0.00003855, -0.00006930, -0.00004514, 0.00000394, 0.00003478, 0.00002914, 0.00000282, -0.00001611, -0.00001422, 0.00000010, 0.00001054, 0.00000976, 0.00000295, -0.00000141, -0.00000094, 0.00000036, -0.00000140, -0.00000556, -0.00000851, -0.00000818, -0.00000564, -0.00000316, -0.00000175, -0.00000120, -0.00000135, -0.00000227, -0.00000338, -0.00000336, -0.00000191, -0.00000104, -0.00000351, -0.00000931, -0.00001420, -0.00001298, -0.00000477, 0.00000510, 0.00000907, 0.00000371, -0.00000782, -0.00001917, -0.00002633, -0.00002969, -0.00003121, -0.00003089, -0.00002669, -0.00001759, -0.00000576, 0.00000508, 0.00001335, 0.00002091, 0.00003042, 0.00004139, 0.00004948, 0.00004992, 0.00004193, 0.00002991, 0.00002049, 0.00001812, 0.00002244, 0.00002911, },
		{ 0.00002660, 0.00002747, 0.00003200, 0.00003527, 0.00003444, 0.00003099, 0.00002912, 0.00003099, 0.00003362, 0.00003210, 0.00002744, 0.00003068, 0.00005514, 0.00009991, 0.00014033, 0.00013955, 0.00007762, -0.00002534, -0.00011606, -0.00014515, -0.00010527, -0.00003675, 0.00000606, -0.00000115, -0.00003704, -0.00005994, -0.00004787, -0.00001540, 0.00000629, 0.00000237, -0.00001368, -0.00001914, -0.00000851, 0.00000217, -0.00000327, -0.00001982, -0.00002535, -0.00000655, 0.00002279, 0.00003311, 0.00000989, -0.00002924, -0.00005035, -0.00003651, -0.00000348, 0.00001815, 0.00001463, 0.00000010, -0.00000386, 0.00000459, 0.00000592, -0.00001413, -0.00004090, -0.00004083, 0.00000105, 0.00005633, 0.00007298, 0.00002561, -0.00005315, -0.00009601, -0.00006170, 0.00002396, 0.00008797, 0.00007479, -0.00000382, -0.00008004, -0.00009076, -0.00003146, 0.00004449, 0.00007617, 0.00004630, -0.00001128, -0.00004802, -0.00004270, -0.00001205, 0.00001273, 0.00001564, 0.00000435, -0.00000437, -0.00000292, 0.00000328, 0.00000568, 0.00000273, -0.00000040, 0.00000066, 0.00000429, 0.00000537, 0.00000127, -0.00000563, -0.00001072, -0.00001140, -0.00000842, -0.00000420, -0.00000086, 0.00000035, -0.00000089, -0.00000353, -0.00000510, -0.00000351, 0.00000046, 0.00000282, -0.00000033, -0.00000850, -0.00001613, -0.00001701, -0.00000987, 0.00000036, 0.00000675, 0.00000600, -0.00000015, -0.00000802, -0.00001590, -0.00002426, -0.00003294, -0.00003931, -0.00003994, -0.00003393, -0.00002393, -0.00001342, -0.00000320, 0.00000891, 0.00002478, 0.00004251, 0.00005656, 0.00006147, 0.00005601, 0.00004414, 0.00003255, },
		{ 0.00007196, 0.00006391, 0.00005170, 0.00004146, 0.00003637, 0.00003542, 0.00003533, 0.00003426, 0.00003389, 0.00003705, 0.00004263, 0.00004419, 0.00003638, 0.00002488, 0.00002783, 0.00006172, 0.00012031, 0.00016755, 0.00015813, 0.00007476, -0.00004938, -0.00014761, -0.00016753, -0.00011102, -0.00003113, 0.00001263, 0.00000118, -0.00003596, -0.00005521, -0.00004057, -0.00001280, -0.00000138, -0.00001194, -0.00002348, -0.00001569, 0.00000589, 0.00001564, -0.00000189, -0.00003171, -0.00004191, -0.00001776, 0.00002068, 0.00003757, 0.00001754, -0.00001909, -0.00003841, -0.00002782, -0.00000633, 0.00000031, -0.00000975, -0.00001341, 0.00000673, 0.00003567, 0.00003696, -0.00000606, -0.00006324, -0.00007911, -0.00002726, 0.00005644, 0.00009976, 0.00005962, -0.00003452, -0.00010339, -0.00008801, -0.00000256, 0.00007896, 0.00008911, 0.00002459, -0.00005578, -0.00008643, -0.00005028, 0.00001409, 0.00005303, 0.00004389, 0.00000634, -0.00002390, -0.00002815, -0.00001449, -0.00000172, 0.00000083, -0.00000188, -0.00000129, 0.00000381, 0.00000755, 0.00000542, -0.00000015, -0.00000288, 0.00000021, 0.00000581, 0.00000807, 0.00000426, -0.00000333, -0.00001009, -0.00001272, -0.00001077, -0.00000599, -0.00000113, 0.00000112, -0.00000041, -0.00000418, -0.00000648, -0.00000452, 0.00000060, 0.00000412, 0.00000180, -0.00000584, -0.00001354, -0.00001577, -0.00001156, -0.00000473, 0.00000008, 0.00000130, 0.00000032, -0.00000162, -0.00000558, -0.00001337, -0.00002454, -0.00003556, -0.00004242, -0.00004382, -0.00004140, -0.00003677, -0.00002891, -0.00001489, 0.00000659, 0.00003243, 0.00005583, 0.00007000, },
		{ 0.00000540, 0.00003799, 0.00006512, 0.00008107, 0.00008446, 0.00007823, 0.00006741, 0.00005604, 0.00004551, 0.00003607, 0.00002988, 0.00003130, 0.00004215, 0.00005613, 0.00006050, 0.00004775, 0.00002824, 0.00002824, 0.00006823, 0.00013703, 0.00018802, 0.00016928, 0.00006787, -0.00007063, -0.00016991, -0.00017940, -0.00011221, -0.00003090, 0.00000646, -0.00000884, -0.00003923, -0.00004463, -0.00002144, 0.00000040, -0.00000507, -0.00002982, -0.00004156, -0.00002095, 0.00001501, 0.00002985, 0.00000682, -0.00003210, -0.00004833, -0.00002670, 0.00000967, 0.00002512, 0.00000930, -0.00001409, -0.00001677, 0.00000056, 0.00000931, -0.00001119, -0.00004394, -0.00004741, -0.00000202, 0.00005994, 0.00007806, 0.00002348, -0.00006519, -0.00011035, -0.00006629, 0.00003458, 0.00010707, 0.00008889, -0.00000338, -0.00008985, -0.00009923, -0.00002974, 0.00005454, 0.00008407, 0.00004267, -0.00002637, -0.00006521, -0.00005121, -0.00000789, 0.00002483, 0.00002704, 0.00000898, -0.00000731, -0.00001080, -0.00000686, -0.00000534, -0.00000815, -0.00000926, -0.00000405, 0.00000444, 0.00000896, 0.00000615, -0.00000015, -0.00000328, -0.00000042, 0.00000528, 0.00000838, 0.00000600, -0.00000057, -0.00000765, -0.00001182, -0.00001152, -0.00000742, -0.00000222, 0.00000071, -0.00000045, -0.00000419, -0.00000664, -0.00000501, -0.00000046, 0.00000269, 0.00000106, -0.00000444, -0.00000954, -0.00001077, -0.00000853, -0.00000598, -0.00000532, -0.00000569, -0.00000489, -0.00000271, -0.00000181, -0.00000523, -0.00001334, -0.00002358, -0.00003312, -0.00004116, -0.00004832, -0.00005393, -0.00005453, -0.00004552, -0.00002468, },
		{ -0.00006784, -0.00007043, -0.00005949, -0.00003471, -0.00000013, 0.00003723, 0.00006988, 0.00009242, 0.00010288, 0.00010192, 0.00009083, 0.00007116, 0.00004711, 0.00002784, 0.00002424, 0.00003984, 0.00006367, 0.00007547, 0.00006373, 0.00004149, 0.00004086, 0.00008428, 0.00015527, 0.00019961, 0.00016452, 0.00004642, -0.00009639, -0.00018357, -0.00017462, -0.00009902, -0.00002712, -0.00000841, -0.00003386, -0.00005513, -0.00003840, 0.00000230, 0.00002373, 0.00000211, -0.00004044, -0.00005898, -0.00003306, 0.00001220, 0.00003301, 0.00001248, -0.00002388, -0.00003786, -0.00002006, 0.00000270, 0.00000115, -0.00002049, -0.00002851, -0.00000091, 0.00004065, 0.00004780, -0.00000167, -0.00007114, -0.00009281, -0.00003423, 0.00006246, 0.00011237, 0.00006568, -0.00004250, -0.00012016, -0.00010046, -0.00000167, 0.00009052, 0.00010042, 0.00002709, -0.00006079, -0.00009013, -0.00004551, 0.00002578, 0.00006285, 0.00004384, -0.00000393, -0.00003734, -0.00003658, -0.00001481, 0.00000300, 0.00000535, -0.00000061, -0.00000301, 0.00000054, 0.00000306, -0.00000112, -0.00000894, -0.00001237, -0.00000753, 0.00000132, 0.00000665, 0.00000520, 0.00000028, -0.00000241, -0.00000036, 0.00000422, 0.00000723, 0.00000614, 0.00000134, -0.00000476, -0.00000931, -0.00001027, -0.00000755, -0.00000332, -0.00000072, -0.00000147, -0.00000420, -0.00000572, -0.00000421, -0.00000115, 0.00000010, -0.00000204, -0.00000562, -0.00000721, -0.00000561, -0.00000314, -0.00000310, -0.00000619, -0.00000973, -0.00001053, -0.00000814, -0.00000513, -0.00000448, -0.00000710, -0.00001209, -0.00001891, -0.00002837, -0.00004131, -0.00005608, },
		{ -0.00001466, -0.00002631, -0.00004347, -0.00006238, -0.00007729, -0.00008234, -0.00007333, -0.00004929, -0.00001304, 0.00002967, 0.00007204, 0.00010770, 0.00013040, 0.00013394, 0.00011497, 0.00007879, 0.00004170, 0.00002385, 0.00003425, 0.00006093, 0.00007897, 0.00007393, 0.00005963, 0.00006892, 0.00011971, 0.00018647, 0.00020970, 0.00014361, 0.00000285, -0.00013788, -0.00019837, -0.00015713, -0.00006762, -0.00000967, -0.00002003, -0.00006582, -0.00008380, -0.00004525, 0.00001742, 0.00004634, 0.00001758, -0.00003647, -0.00006244, -0.00004009, 0.00000124, 0.00001839, 0.00000123, -0.00002048, -0.00001575, 0.00000871, 0.00001528, -0.00001804, -0.00006407, -0.00006813, -0.00000753, 0.00007402, 0.00009967, 0.00003405, -0.00007449, -0.00013106, -0.00008059, 0.00003800, 0.00012364, 0.00010325, -0.00000334, -0.00010333, -0.00011497, -0.00003718, 0.00005670, 0.00008904, 0.00004389, -0.00002861, -0.00006501, -0.00004395, 0.00000457, 0.00003555, 0.00003021, 0.00000462, -0.00001425, -0.00001542, -0.00000827, -0.00000624, -0.00001118, -0.00001427, -0.00000865, 0.00000206, 0.00000817, 0.00000458, -0.00000428, -0.00000979, -0.00000782, -0.00000165, 0.00000262, 0.00000226, -0.00000049, -0.00000178, -0.00000003, 0.00000329, 0.00000553, 0.00000504, 0.00000189, -0.00000254, -0.00000621, -0.00000743, -0.00000599, -0.00000355, -0.00000239, -0.00000331, -0.00000479, -0.00000463, -0.00000255, -0.00000083, -0.00000192, -0.00000551, -0.00000835, -0.00000736, -0.00000296, 0.00000105, 0.00000099, -0.00000338, -0.00000891, -0.00001222, -0.00001239, -0.00001088, -0.00000936, -0.00000833, -0.00000785, -0.00000914, },
		{ -0.00001321, -0.00001098, -0.00000875, -0.00000903, -0.00001426, -0.00002569, -0.00004283, -0.00006311, -0.00008179, -0.00009266, -0.00009001, -0.00007052, -0.00003395, 0.00001699, 0.00007578, 0.00013041, 0.00016442, 0.00016433, 0.00013068, 0.00008261, 0.00004762, 0.00004150, 0.00005589, 0.00006814, 0.00006722, 0.00006958, 0.00010412, 0.00017497, 0.00023940, 0.00023106, 0.00011744, -0.00005892, -0.00020026, -0.00022653, -0.00014078, -0.00002677, 0.00002259, -0.00001732, -0.00008894, -0.00011285, -0.00006295, 0.00001458, 0.00005213, 0.00002705, -0.00002272, -0.00004460, -0.00002648, -0.00000261, -0.00000773, -0.00003358, -0.00003827, 0.00000093, 0.00005193, 0.00005435, -0.00001407, -0.00010183, -0.00012365, -0.00004345, 0.00008143, 0.00014445, 0.00008622, -0.00004760, -0.00014384, -0.00012202, -0.00000471, 0.00010569, 0.00011953, 0.00003543, -0.00006709, -0.00010391, -0.00005725, 0.00001984, 0.00006019, 0.00004120, -0.00000593, -0.00003506, -0.00002806, -0.00000273, 0.00001330, 0.00001047, 0.00000038, -0.00000264, 0.00000216, 0.00000413, -0.00000404, -0.00001725, -0.00002366, -0.00001735, -0.00000393, 0.00000583, 0.00000626, 0.00000060, -0.00000414, -0.00000434, -0.00000183, -0.00000028, -0.00000101, -0.00000236, -0.00000221, -0.00000019, 0.00000233, 0.00000375, 0.00000328, 0.00000118, -0.00000156, -0.00000359, -0.00000400, -0.00000314, -0.00000249, -0.00000332, -0.00000512, -0.00000585, -0.00000413, -0.00000117, -0.00000011, -0.00000284, -0.00000772, -0.00001059, -0.00000861, -0.00000296, 0.00000213, 0.00000310, -0.00000027, -0.00000539, -0.00000965, -0.00001223, -0.00001362, -0.00001408, },
		{ -0.00000884, -0.00001158, -0.00001400, -0.00001553, -0.00001581, -0.00001499, -0.00001367, -0.00001298, -0.00001497, -0.00002246, -0.00003745, -0.00005914, -0.00008328, -0.00010325, -0.00011128, -0.00009925, -0.00006056, 0.00000492, 0.00008501, 0.00015546, 0.00019125, 0.00018324, 0.00014589, 0.00010632, 0.00008221, 0.00006982, 0.00005628, 0.00004440, 0.00006089, 0.00013062, 0.00023662, 0.00030939, 0.00027018, 0.00010083, -0.00012271, -0.00027409, -0.00027069, -0.00013964, 0.00000481, 0.00005591, -0.00000156, -0.00009104, -0.00012286, -0.00007496, -0.00000136, 0.00003164, 0.00001170, -0.00001735, -0.00001310, 0.00001501, 0.00001898, -0.00002616, -0.00008313, -0.00008452, -0.00000746, 0.00008893, 0.00011072, 0.00002138, -0.00011243, -0.00017315, -0.00009909, 0.00005463, 0.00016122, 0.00013369, 0.00000033, -0.00012375, -0.00013970, -0.00004697, 0.00006635, 0.00010770, 0.00005733, -0.00002735, -0.00007335, -0.00005543, -0.00000592, 0.00002666, 0.00002300, 0.00000099, -0.00001191, -0.00000726, 0.00000219, 0.00000237, -0.00000568, -0.00000960, -0.00000205, 0.00001025, 0.00001389, 0.00000321, -0.00001404, -0.00002469, -0.00002235, -0.00001156, -0.00000157, 0.00000235, 0.00000164, 0.00000039, 0.00000047, 0.00000068, -0.00000050, -0.00000267, -0.00000399, -0.00000327, -0.00000106, 0.00000109, 0.00000195, 0.00000127, -0.00000035, -0.00000180, -0.00000209, -0.00000111, -0.00000002, -0.00000049, -0.00000300, -0.00000588, -0.00000655, -0.00000412, -0.00000077, -0.00000003, -0.00000336, -0.00000847, -0.00001117, -0.00000913, -0.00000385, 0.00000090, 0.00000236, 0.00000054, -0.00000275, -0.00000597, },
		{ -0.00000090, -0.00000284, -0.00000456, -0.00000596, -0.00000737, -0.00000926, -0.00001216, -0.00001622, -0.00002045, -0.00002273, -0.00002138, -0.00001722, -0.00001402, -0.00001669, -0.00002889, -0.00005173, -0.00008318, -0.00011647, -0.00013839, -0.00013177, -0.00008479, -0.00000235, 0.00009132, 0.00016525, 0.00020138, 0.00020398, 0.00018903, 0.00016466, 0.00012551, 0.00007075, 0.00002650, 0.00004038, 0.00013938, 0.00028489, 0.00037587, 0.00031600, 0.00009806, -0.00017076, -0.00033713, -0.00031578, -0.00015398, 0.00001130, 0.00006702, 0.00000857, -0.00007636, -0.00010129, -0.00005844, -0.00000918, -0.00000685, -0.00003780, -0.00004244, 0.00000930, 0.00007384, 0.00007394, -0.00001492, -0.00012434, -0.00014891, -0.00004983, 0.00009679, 0.00016127, 0.00007893, -0.00008587, -0.00019332, -0.00015168, 0.00000271, 0.00014066, 0.00015494, 0.00004847, -0.00007916, -0.00012568, -0.00007021, 0.00002333, 0.00007445, 0.00005510, 0.00000014, -0.00003760, -0.00003629, -0.00001381, 0.00000134, 0.00000006, -0.00000528, -0.00000164, 0.00000846, 0.00001182, 0.00000188, -0.00001272, -0.00001729, -0.00000674, 0.00000928, 0.00001626, 0.00000850, -0.00000706, -0.00001880, -0.00002074, -0.00001547, -0.00000901, -0.00000467, -0.00000182, 0.00000098, 0.00000334, 0.00000361, 0.00000128, -0.00000206, -0.00000414, -0.00000389, -0.00000209, -0.00000034, 0.00000016, -0.00000073, -0.00000216, -0.00000281, -0.00000175, 0.00000056, 0.00000226, 0.00000148, -0.00000178, -0.00000522, -0.00000607, -0.00000385, -0.00000108, -0.00000097, -0.00000433, -0.00000866, -0.00001045, -0.00000832, -0.00000402, -0.00000057, 0.00000033, },
		{ -0.00000340, -0.00000105, -0.00000096, -0.00000255, -0.00000460, -0.00000604, -0.00000615, -0.00000501, -0.00000399, -0.00000545, -0.00001100, -0.00001957, -0.00002749, -0.00003076, -0.00002791, -0.00002066, -0.00001285, -0.00000963, -0.00001751, -0.00004265, -0.00008541, -0.00013402, -0.00016546, -0.00015720, -0.00010278, -0.00001735, 0.00007361, 0.00015230, 0.00021469, 0.00025845, 0.00026902, 0.00022670, 0.00013629, 0.00004863, 0.00003900, 0.00014458, 0.00031212, 0.00041429, 0.00034001, 0.00009126, -0.00020128, -0.00036913, -0.00033383, -0.00016279, -0.00000427, 0.00003928, -0.00001465, -0.00006932, -0.00005633, -0.00000071, 0.00001814, -0.00003659, -0.00010978, -0.00010815, -0.00000364, 0.00012284, 0.00015006, 0.00003575, -0.00013163, -0.00020572, -0.00011582, 0.00006455, 0.00018075, 0.00013445, -0.00003092, -0.00017289, -0.00017710, -0.00005174, 0.00009111, 0.00014012, 0.00007551, -0.00002998, -0.00008725, -0.00006612, -0.00000564, 0.00003580, 0.00003392, 0.00000793, -0.00001095, -0.00001190, -0.00000695, -0.00000935, -0.00001667, -0.00001618, -0.00000258, 0.00001393, 0.00001809, 0.00000586, -0.00001137, -0.00001829, -0.00001019, 0.00000409, 0.00001189, 0.00000820, -0.00000206, -0.00001080, -0.00001419, -0.00001377, -0.00001238, -0.00001060, -0.00000725, -0.00000220, 0.00000251, 0.00000443, 0.00000296, -0.00000020, -0.00000270, -0.00000332, -0.00000247, -0.00000146, -0.00000140, -0.00000249, -0.00000383, -0.00000399, -0.00000213, 0.00000097, 0.00000312, 0.00000250, -0.00000057, -0.00000364, -0.00000436, -0.00000277, -0.00000129, -0.00000232, -0.00000575, -0.00000902, -0.00000955, -0.00000701, },
		{ -0.00000926, -0.00000895, -0.00000625, -0.00000302, -0.00000101, -0.00000100, -0.00000285, -0.00000576, -0.00000824, -0.00000852, -0.00000593, -0.00000218, -0.00000082, -0.00000462, -0.00001328, -0.00002362, -0.00003197, -0.00003592, -0.00003397, -0.00002520, -0.00001142, -0.00000068, -0.00000625, -0.00003801, -0.00009176, -0.00014713, -0.00017899, -0.00017322, -0.00013198, -0.00006370, 0.00002877, 0.00014424, 0.00026465, 0.00034605, 0.00034225, 0.00025093, 0.00013784, 0.00010084, 0.00018831, 0.00034058, 0.00041878, 0.00031314, 0.00004384, -0.00023977, -0.00037203, -0.00030289, -0.00013201, -0.00001363, -0.00001978, -0.00008865, -0.00010324, -0.00001925, 0.00008618, 0.00009456, -0.00002538, -0.00017219, -0.00020084, -0.00006420, 0.00013167, 0.00021723, 0.00011361, -0.00009297, -0.00022688, -0.00017872, 0.00000179, 0.00015666, 0.00016169, 0.00002836, -0.00011981, -0.00016336, -0.00008462, 0.00003435, 0.00009654, 0.00007128, 0.00000341, -0.00004257, -0.00004055, -0.00001225, 0.00000775, 0.00000739, -0.00000026, -0.00000029, 0.00000569, 0.00000502, -0.00000759, -0.00002187, -0.00002294, -0.00000793, 0.00001054, 0.00001701, 0.00000788, -0.00000657, -0.00001344, -0.00000902, 0.00000003, 0.00000499, 0.00000319, -0.00000172, -0.00000535, -0.00000691, -0.00000837, -0.00001073, -0.00001241, -0.00001111, -0.00000660, -0.00000123, 0.00000221, 0.00000262, 0.00000101, -0.00000085, -0.00000180, -0.00000185, -0.00000177, -0.00000237, -0.00000375, -0.00000503, -0.00000486, -0.00000263, 0.00000062, 0.00000273, 0.00000227, -0.00000008, -0.00000211, -0.00000224, -0.00000115, -0.00000102, -0.00000322, -0.00000678, },
		{ -0.00000076, -0.00000352, -0.00000678, -0.00000858, -0.00000810, -0.00000599, -0.00000348, -0.00000157, -0.00000098, -0.00000232, -0.00000544, -0.00000876, -0.00000991, -0.00000777, -0.00000394, -0.00000167, -0.00000312, -0.00000794, -0.00001469, -0.00002296, -0.00003262, -0.00004076, -0.00004099, -0.00002865, -0.00000828, 0.00000532, -0.00000354, -0.00003892, -0.00008995, -0.00014039, -0.00017989, -0.00020302, -0.00019703, -0.00013745, -0.00000698, 0.00017349, 0.00033703, 0.00040921, 0.00037074, 0.00028571, 0.00025520, 0.00032402, 0.00042438, 0.00042103, 0.00023246, -0.00007767, -0.00033216, -0.00038086, -0.00022974, -0.00003742, 0.00002511, -0.00007331, -0.00020170, -0.00019783, -0.00003010, 0.00016449, 0.00020588, 0.00004352, -0.00018785, -0.00028494, -0.00015867, 0.00008551, 0.00024233, 0.00018752, -0.00001964, -0.00019859, -0.00020935, -0.00006498, 0.00009751, 0.00014715, 0.00006559, -0.00005731, -0.00011732, -0.00008354, -0.00000640, 0.00004427, 0.00004172, 0.00001095, -0.00001015, -0.00000911, -0.00000090, -0.00000212, -0.00001103, -0.00001346, -0.00000283, 0.00001089, 0.00001251, -0.00000108, -0.00001741, -0.00002152, -0.00001076, 0.00000393, 0.00000999, 0.00000495, -0.00000350, -0.00000692, -0.00000416, -0.00000037, -0.00000012, -0.00000290, -0.00000494, -0.00000422, -0.00000253, -0.00000294, -0.00000619, -0.00000995, -0.00001126, -0.00000922, -0.00000536, -0.00000192, -0.00000012, 0.00000019, -0.00000014, -0.00000046, -0.00000071, -0.00000126, -0.00000251, -0.00000427, -0.00000555, -0.00000517, -0.00000288, 0.00000005, 0.00000171, 0.00000126, -0.00000030, -0.00000117, -0.00000066, 0.00000006, },
		{ -0.00000059, 0.00000009, 0.00000015, -0.00000138, -0.00000397, -0.00000614, -0.00000684, -0.00000626, -0.00000518, -0.00000400, -0.00000279, -0.00000210, -0.00000283, -0.00000513, -0.00000756, -0.00000818, -0.00000682, -0.00000553, -0.00000631, -0.00000850, -0.00000955, -0.00000896, -0.00001058, -0.00001914, -0.00003363, -0.00004527, -0.00004419, -0.00002934, -0.00001091, -0.00000201, -0.00000848, -0.00002886, -0.00006378, -0.00012007, -0.00019779, -0.00027026, -0.00028407, -0.00019339, -0.00000564, 0.00020785, 0.00035960, 0.00041805, 0.00043428, 0.00048602, 0.00058272, 0.00062704, 0.00048893, 0.00014084, -0.00027286, -0.00051777, -0.00045306, -0.00016003, 0.00011028, 0.00014051, -0.00007032, -0.00030913, -0.00033935, -0.00011257, 0.00018788, 0.00031014, 0.00015588, -0.00013694, -0.00031997, -0.00025019, -0.00000295, 0.00020862, 0.00022320, 0.00005756, -0.00013083, -0.00019277, -0.00010653, 0.00002870, 0.00009791, 0.00006667, -0.00001095, -0.00006035, -0.00005382, -0.00001881, 0.00000458, 0.00000424, -0.00000282, 0.00000095, 0.00001254, 0.00001561, 0.00000263, -0.00001532, -0.00002099, -0.00000977, 0.00000617, 0.00001140, 0.00000268, -0.00000963, -0.00001378, -0.00000808, -0.00000040, 0.00000172, -0.00000155, -0.00000436, -0.00000283, 0.00000092, 0.00000191, -0.00000170, -0.00000672, -0.00000861, -0.00000610, -0.00000206, -0.00000027, -0.00000198, -0.00000546, -0.00000809, -0.00000861, -0.00000744, -0.00000564, -0.00000395, -0.00000256, -0.00000141, -0.00000052, -0.00000011, -0.00000055, -0.00000200, -0.00000394, -0.00000522, -0.00000481, -0.00000279, -0.00000049, 0.00000059, 0.00000018, -0.00000061, },
		{ -0.00000051, -0.00000061, -0.00000043, 0.00000026, 0.00000055, -0.00000061, -0.00000308, -0.00000535, -0.00000602, -0.00000511, -0.00000384, -0.00000317, -0.00000306, -0.00000317, -0.00000379, -0.00000535, -0.00000709, -0.00000709, -0.00000455, -0.00000187, -0.00000312, -0.00000948, -0.00001661, -0.00001800, -0.00001197, -0.00000487, -0.00000575, -0.00001726, -0.00003261, -0.00004223, -0.00004284, -0.00003812, -0.00003079, -0.00001800, 0.00000028, 0.00000757, -0.00002714, -0.00012242, -0.00025269, -0.00034808, -0.00033900, -0.00021205, -0.00002202, 0.00015991, 0.00031543, 0.00048998, 0.00071963, 0.00093971, 0.00098190, 0.00069813, 0.00012193, -0.00048754, -0.00079483, -0.00063639, -0.00016368, 0.00025646, 0.00032094, 0.00002391, -0.00035177, -0.00047821, -0.00025335, 0.00012858, 0.00035684, 0.00026815, -0.00003256, -0.00028395, -0.00029701, -0.00009721, 0.00012744, 0.00020285, 0.00010452, -0.00005298, -0.00013728, -0.00010725, -0.00002154, 0.00003718, 0.00003682, 0.00000519, -0.00001590, -0.00001401, -0.00000677, -0.00001140, -0.00002299, -0.00002343, -0.00000550, 0.00001723, 0.00002477, 0.00001159, -0.00000879, -0.00001861, -0.00001285, -0.00000128, 0.00000406, 0.00000074, -0.00000438, -0.00000485, -0.00000160, -0.00000025, -0.00000349, -0.00000781, -0.00000782, -0.00000259, 0.00000314, 0.00000397, -0.00000091, -0.00000726, -0.00001015, -0.00000821, -0.00000394, -0.00000077, -0.00000026, -0.00000177, -0.00000382, -0.00000551, -0.00000660, -0.00000706, -0.00000683, -0.00000583, -0.00000422, -0.00000239, -0.00000098, -0.00000062, -0.00000150, -0.00000304, -0.00000408, -0.00000378, -0.00000238, -0.00000101, },
		{ -0.00000215, -0.00000164, -0.00000151, -0.00000163, -0.00000121, 0.00000017, 0.00000160, 0.00000138, -0.00000122, -0.00000491, -0.00000733, -0.00000713, -0.00000497, -0.00000253, -0.00000088, -0.00000026, -0.00000104, -0.00000393, -0.00000843, -0.00001155, -0.00000963, -0.00000258, 0.00000408, 0.00000308, -0.00000741, -0.00002061, -0.00002654, -0.00002125, -0.00001050, -0.00000374, -0.00000512, -0.00001174, -0.00002050, -0.00003334, -0.00005176, -0.00006630, -0.00005757, -0.00001618, 0.00003532, 0.00004656, -0.00002263, -0.00016056, -0.00030156, -0.00037404, -0.00035543, -0.00027552, -0.00015897, 0.00003035, 0.00036484, 0.00084472, 0.00130454, 0.00145656, 0.00108692, 0.00026573, -0.00062254, -0.00110077, -0.00093126, -0.00029820, 0.00032072, 0.00050654, 0.00019941, -0.00028917, -0.00054995, -0.00040101, -0.00000163, 0.00031550, 0.00032822, 0.00008208, -0.00018850, -0.00027605, -0.00015587, 0.00003360, 0.00013599, 0.00010295, 0.00000232, -0.00006959, -0.00007335, -0.00003778, -0.00000935, -0.00000397, -0.00000531, 0.00000320, 0.00001581, 0.00001427, -0.00000685, -0.00003101, -0.00003619, -0.00001749, 0.00000867, 0.00002156, 0.00001505, -0.00000004, -0.00000955, -0.00000925, -0.00000515, -0.00000373, -0.00000475, -0.00000371, 0.00000081, 0.00000464, 0.00000281, -0.00000429, -0.00001070, -0.00001079, -0.00000477, 0.00000182, 0.00000355, -0.00000027, -0.00000591, -0.00000914, -0.00000857, -0.00000572, -0.00000281, -0.00000103, -0.00000043, -0.00000080, -0.00000200, -0.00000386, -0.00000590, -0.00000741, -0.00000785, -0.00000705, -0.00000533, -0.00000341, -0.00000213, -0.00000187, -0.00000225, -0.00000247, },
		{ -0.00000356, -0.00000230, -0.00000110, -0.00000033, -0.00000059, -0.00000175, -0.00000262, -0.00000181, 0.00000065, 0.00000283, 0.00000246, -0.00000094, -0.00000543, -0.00000839, -0.00000866, -0.00000685, -0.00000396, -0.00000051, 0.00000267, 0.00000329, -0.00000097, -0.00000917, -0.00001573, -0.00001435, -0.00000451, 0.00000628, 0.00000816, -0.00000192, -0.00001714, -0.00002716, -0.00002760, -0.00002216, -0.00001586, -0.00000874, 0.00000086, 0.00000725, -0.00000329, -0.00003788, -0.00008046, -0.00009583, -0.00005973, 0.00001067, 0.00006051, 0.00003984, -0.00005277, -0.00017198, -0.00027425, -0.00035924, -0.00044696, -0.00050231, -0.00039463, 0.00002595, 0.00076060, 0.00154717, 0.00194545, 0.00161879, 0.00061842, -0.00057368, -0.00132444, -0.00127122, -0.00056846, 0.00023732, 0.00061050, 0.00039476, -0.00012874, -0.00051046, -0.00048903, -0.00015167, 0.00019625, 0.00030442, 0.00015527, -0.00007508, -0.00019717, -0.00015564, -0.00003340, 0.00005437, 0.00006023, 0.00001730, -0.00001917, -0.00002776, -0.00002441, -0.00002774, -0.00003385, -0.00002591, -0.00000014, 0.00002523, 0.00002799, 0.00000531, -0.00002316, -0.00003461, -0.00002344, -0.00000304, 0.00000978, 0.00000973, 0.00000352, -0.00000078, -0.00000198, -0.00000374, -0.00000742, -0.00000937, -0.00000570, 0.00000207, 0.00000752, 0.00000559, -0.00000228, -0.00000958, -0.00001095, -0.00000666, -0.00000141, 0.00000049, -0.00000154, -0.00000493, -0.00000705, -0.00000712, -0.00000592, -0.00000432, -0.00000267, -0.00000110, -0.00000001, 0.00000005, -0.00000118, -0.00000339, -0.00000583, -0.00000770, -0.00000840, -0.00000785, -0.00000649, -0.00000494, },
		{ -0.00000870, -0.00000878, -0.00000796, -0.00000619, -0.00000355, -0.00000074, 0.00000091, 0.00000044, -0.00000153, -0.00000298, -0.00000213, 0.00000066, 0.00000302, 0.00000274, -0.00000027, -0.00000410, -0.00000698, -0.00000859, -0.00000914, -0.00000785, -0.00000349, 0.00000293, 0.00000693, 0.00000377, -0.00000618, -0.00001585, -0.00001702, -0.00000830, 0.00000274, 0.00000650, 0.00000034, -0.00000997, -0.00001774, -0.00002265, -0.00002829, -0.00003310, -0.00002767, -0.00000583, 0.00002130, 0.00002758, -0.00000472, -0.00006142, -0.00010115, -0.00009072, -0.00003843, 0.00001167, 0.00002284, -0.00000069, -0.00003241, -0.00007933, -0.00020290, -0.00045183, -0.00074765, -0.00084712, -0.00047465, 0.00043087, 0.00155278, 0.00230186, 0.00217778, 0.00114375, -0.00028607, -0.00135089, -0.00153849, -0.00090786, -0.00000298, 0.00056217, 0.00052075, 0.00006877, -0.00036496, -0.00047177, -0.00025543, 0.00004823, 0.00020506, 0.00015536, 0.00000698, -0.00009843, -0.00010430, -0.00005131, -0.00000636, 0.00000383, -0.00000245, -0.00000178, 0.00000457, -0.00000043, -0.00002166, -0.00004162, -0.00003890, -0.00001216, 0.00001737, 0.00002669, 0.00001206, -0.00001044, -0.00002237, -0.00001896, -0.00000890, -0.00000217, -0.00000058, 0.00000012, 0.00000240, 0.00000356, 0.00000015, -0.00000637, -0.00001012, -0.00000698, 0.00000082, 0.00000665, 0.00000588, -0.00000016, -0.00000618, -0.00000817, -0.00000644, -0.00000409, -0.00000352, -0.00000456, -0.00000566, -0.00000584, -0.00000533, -0.00000478, -0.00000430, -0.00000351, -0.00000221, -0.00000075, 0.00000016, -0.00000001, -0.00000136, -0.00000351, -0.00000584, -0.00000770, },
		{ -0.00000204, -0.00000376, -0.00000574, -0.00000769, -0.00000926, -0.00000980, -0.00000857, -0.00000549, -0.00000174, 0.00000071, 0.00000066, -0.00000108, -0.00000228, -0.00000145, 0.00000064, 0.00000188, 0.00000114, -0.00000067, -0.00000205, -0.00000315, -0.00000547, -0.00000921, -0.00001146, -0.00000854, -0.00000070, 0.00000637, 0.00000617, -0.00000184, -0.00001088, -0.00001330, -0.00000846, -0.00000300, -0.00000289, -0.00000652, -0.00000755, -0.00000489, -0.00000664, -0.00002070, -0.00004112, -0.00004818, -0.00002719, 0.00001154, 0.00003601, 0.00002133, -0.00002396, -0.00006442, -0.00007419, -0.00006361, -0.00006166, -0.00007012, -0.00004816, 0.00003522, 0.00012257, 0.00006200, -0.00027142, -0.00080060, -0.00118798, -0.00100993, -0.00008731, 0.00127885, 0.00239825, 0.00260554, 0.00172089, 0.00020679, -0.00111849, -0.00161224, -0.00119659, -0.00033531, 0.00034921, 0.00050491, 0.00021483, -0.00017184, -0.00035250, -0.00026757, -0.00006532, 0.00007555, 0.00008875, 0.00002772, -0.00002451, -0.00003518, -0.00002634, -0.00002677, -0.00003604, -0.00003394, -0.00001249, 0.00001119, 0.00001403, -0.00000691, -0.00003113, -0.00003605, -0.00001875, 0.00000393, 0.00001390, 0.00000803, -0.00000309, -0.00000873, -0.00000787, -0.00000635, -0.00000771, -0.00000922, -0.00000662, -0.00000044, 0.00000395, 0.00000241, -0.00000327, -0.00000722, -0.00000562, -0.00000023, 0.00000392, 0.00000371, 0.00000031, -0.00000291, -0.00000411, -0.00000414, -0.00000475, -0.00000634, -0.00000773, -0.00000773, -0.00000639, -0.00000475, -0.00000373, -0.00000334, -0.00000304, -0.00000242, -0.00000153, -0.00000074, -0.00000043, -0.00000086, },
		{ -0.00000125, -0.00000134, -0.00000163, -0.00000210, -0.00000283, -0.00000403, -0.00000586, -0.00000804, -0.00000959, -0.00000930, -0.00000687, -0.00000353, -0.00000122, -0.00000080, -0.00000123, -0.00000080, 0.00000070, 0.00000158, 0.00000031, -0.00000221, -0.00000323, -0.00000140, 0.00000094, -0.00000020, -0.00000572, -0.00001142, -0.00001174, -0.00000590, 0.00000077, 0.00000228, -0.00000137, -0.00000450, -0.00000307, -0.00000055, -0.00000403, -0.00001423, -0.00002158, -0.00001570, 0.00000064, 0.00001008, -0.00000266, -0.00003200, -0.00005275, -0.00004418, -0.00001275, 0.00001316, 0.00001355, -0.00000153, -0.00000769, -0.00000436, -0.00002651, -0.00010159, -0.00018965, -0.00018537, -0.00001594, 0.00023786, 0.00033289, 0.00003351, -0.00064126, -0.00131224, -0.00142704, -0.00065758, 0.00078341, 0.00218577, 0.00276597, 0.00218577, 0.00078341, -0.00065758, -0.00142704, -0.00131224, -0.00064126, 0.00003351, 0.00033289, 0.00023786, -0.00001594, -0.00018537, -0.00018965, -0.00010159, -0.00002651, -0.00000436, -0.00000769, -0.00000153, 0.00001355, 0.00001316, -0.00001275, -0.00004418, -0.00005275, -0.00003200, -0.00000266, 0.00001008, 0.00000064, -0.00001570, -0.00002158, -0.00001423, -0.00000403, -0.00000055, -0.00000307, -0.00000450, -0.00000137, 0.00000228, 0.00000077, -0.00000590, -0.00001174, -0.00001142, -0.00000572, -0.00000020, 0.00000094, -0.00000140, -0.00000323, -0.00000221, 0.00000031, 0.00000158, 0.00000070, -0.00000080, -0.00000123, -0.00000080, -0.00000122, -0.00000353, -0.00000687, -0.00000930, -0.00000959, -0.00000804, -0.00000586, -0.00000403, -0.00000283, -0.00000210, -0.00000163, -0.00000134, },
		{ -0.00000204, -0.00000086, -0.00000043, -0.00000074, -0.00000153, -0.00000242, -0.00000304, -0.00000334, -0.00000373, -0.00000475, -0.00000639, -0.00000773, -0.00000773, -0.00000634, -0.00000475, -0.00000414, -0.00000411, -0.00000291, 0.00000031, 0.00000371, 0.00000392, -0.00000023, -0.00000562, -0.00000722, -0.00000327, 0.00000241, 0.00000395, -0.00000044, -0.00000662, -0.00000922, -0.00000771, -0.00000635, -0.00000787, -0.00000873, -0.00000309, 0.00000803, 0.00001390, 0.00000393, -0.00001875, -0.00003605, -0.00003113, -0.00000691, 0.00001403, 0.00001119, -0.00001249, -0.00003394, -0.00003604, -0.00002677, -0.00002634, -0.00003518, -0.00002451, 0.00002772, 0.00008875, 0.00007555, -0.00006532, -0.00026757, -0.00035250, -0.00017184, 0.00021483, 0.00050491, 0.00034921, -0.00033531, -0.00119659, -0.00161224, -0.00111849, 0.00020679, 0.00172089, 0.00260554, 0.00239825, 0.00127885, -0.00008731, -0.00100993, -0.00118798, -0.00080060, -0.00027142, 0.00006200, 0.00012257, 0.00003522, -0.00004816, -0.00007012, -0.00006166, -0.00006361, -0.00007419, -0.00006442, -0.00002396, 0.00002133, 0.00003601, 0.00001154, -0.00002719, -0.00004818, -0.00004112, -0.00002070, -0.00000664, -0.00000489, -0.00000755, -0.00000652, -0.00000289, -0.00000300, -0.00000846, -0.00001330, -0.00001088, -0.00000184, 0.00000617, 0.00000637, -0.00000070, -0.00000854, -0.00001146, -0.00000921, -0.00000547, -0.00000315, -0.00000205, -0.00000067, 0.00000114, 0.00000188, 0.00000064, -0.00000145, -0.00000228, -0.00000108, 0.00000066, 0.00000071, -0.00000174, -0.00000549, -0.00000857, -0.00000980, -0.00000926, -0.00000769, -0.00000574, -0.00000376, },
		{ -0.00000870, -0.00000770, -0.00000584, -0.00000351, -0.00000136, -0.00000001, 0.00000016, -0.00000075, -0.00000221, -0.00000351, -0.00000430, -0.00000478, -0.00000533, -0.00000584, -0.00000566, -0.00000456, -0.00000352, -0.00000409, -0.00000644, -0.00000817, -0.00000618, -0.00000016, 0.00000588, 0.00000665, 0.00000082, -0.00000698, -0.00001012, -0.00000637, 0.00000015, 0.00000356, 0.00000240, 0.00000012, -0.00000058, -0.00000217, -0.00000890, -0.00001896, -0.00002237, -0.00001044, 0.00001206, 0.00002669, 0.00001737, -0.00001216, -0.00003890, -0.00004162, -0.00002166, -0.00000043, 0.00000457, -0.00000178, -0.00000245, 0.00000383, -0.00000636, -0.00005131, -0.00010430, -0.00009843, 0.00000698, 0.00015536, 0.00020506, 0.00004823, -0.00025543, -0.00047177, -0.00036496, 0.00006877, 0.00052075, 0.00056217, -0.00000298, -0.00090786, -0.00153849, -0.00135089, -0.00028607, 0.00114375, 0.00217778, 0.00230186, 0.00155278, 0.00043087, -0.00047465, -0.00084712, -0.00074765, -0.00045183, -0.00020290, -0.00007933, -0.00003241, -0.00000069, 0.00002284, 0.00001167, -0.00003843, -0.00009072, -0.00010115, -0.00006142, -0.00000472, 0.00002758, 0.00002130, -0.00000583, -0.00002767, -0.00003310, -0.00002829, -0.00002265, -0.00001774, -0.00000997, 0.00000034, 0.00000650, 0.00000274, -0.00000830, -0.00001702, -0.00001585, -0.00000618, 0.00000377, 0.00000693, 0.00000293, -0.00000349, -0.00000785, -0.00000914, -0.00000859, -0.00000698, -0.00000410, -0.00000027, 0.00000274, 0.00000302, 0.00000066, -0.00000213, -0.00000298, -0.00000153, 0.00000044, 0.00000091, -0.00000074, -0.00000355, -0.00000619, -0.00000796, -0.00000878, },
		{ -0.00000356, -0.00000494, -0.00000649, -0.00000785, -0.00000840, -0.00000770, -0.00000583, -0.00000339, -0.00000118, 0.00000005, -0.00000001, -0.00000110, -0.00000267, -0.00000432, -0.00000592, -0.00000712, -0.00000705, -0.00000493, -0.00000154, 0.00000049, -0.00000141, -0.00000666, -0.00001095, -0.00000958, -0.00000228, 0.00000559, 0.00000752, 0.00000207, -0.00000570, -0.00000937, -0.00000742, -0.00000374, -0.00000198, -0.00000078, 0.00000352, 0.00000973, 0.00000978, -0.00000304, -0.00002344, -0.00003461, -0.00002316, 0.00000531, 0.00002799, 0.00002523, -0.00000014, -0.00002591, -0.00003385, -0.00002774, -0.00002441, -0.00002776, -0.00001917, 0.00001730, 0.00006023, 0.00005437, -0.00003340, -0.00015564, -0.00019717, -0.00007508, 0.00015527, 0.00030442, 0.00019625, -0.00015167, -0.00048903, -0.00051046, -0.00012874, 0.00039476, 0.00061050, 0.00023732, -0.00056846, -0.00127122, -0.00132444, -0.00057368, 0.00061842, 0.00161879, 0.00194545, 0.00154717, 0.00076060, 0.00002595, -0.00039463, -0.00050231, -0.00044696, -0.00035924, -0.00027425, -0.00017198, -0.00005277, 0.00003984, 0.00006051, 0.00001067, -0.00005973, -0.00009583, -0.00008046, -0.00003788, -0.00000329, 0.00000725, 0.00000086, -0.00000874, -0.00001586, -0.00002216, -0.00002760, -0.00002716, -0.00001714, -0.00000192, 0.00000816, 0.00000628, -0.00000451, -0.00001435, -0.00001573, -0.00000917, -0.00000097, 0.00000329, 0.00000267, -0.00000051, -0.00000396, -0.00000685, -0.00000866, -0.00000839, -0.00000543, -0.00000094, 0.00000246, 0.00000283, 0.00000065, -0.00000181, -0.00000262, -0.00000175, -0.00000059, -0.00000033, -0.00000110, -0.00000230, },
		{ -0.00000215, -0.00000247, -0.00000225, -0.00000187, -0.00000213, -0.00000341, -0.00000533, -0.00000705, -0.00000785, -0.00000741, -0.00000590, -0.00000386, -0.00000200, -0.00000080, -0.00000043, -0.00000103, -0.00000281, -0.00000572, -0.00000857, -0.00000914, -0.00000591, -0.00000027, 0.00000355, 0.00000182, -0.00000477, -0.00001079, -0.00001070, -0.00000429, 0.00000281, 0.00000464, 0.00000081, -0.00000371, -0.00000475, -0.00000373, -0.00000515, -0.00000925, -0.00000955, -0.00000004, 0.00001505, 0.00002156, 0.00000867, -0.00001749, -0.00003619, -0.00003101, -0.00000685, 0.00001427, 0.00001581, 0.00000320, -0.00000531, -0.00000397, -0.00000935, -0.00003778, -0.00007335, -0.00006959, 0.00000232, 0.00010295, 0.00013599, 0.00003360, -0.00015587, -0.00027605, -0.00018850, 0.00008208, 0.00032822, 0.00031550, -0.00000163, -0.00040101, -0.00054995, -0.00028917, 0.00019941, 0.00050654, 0.00032072, -0.00029820, -0.00093126, -0.00110077, -0.00062254, 0.00026573, 0.00108692, 0.00145656, 0.00130454, 0.00084472, 0.00036484, 0.00003035, -0.00015897, -0.00027552, -0.00035543, -0.00037404, -0.00030156, -0.00016056, -0.00002263, 0.00004656, 0.00003532, -0.00001618, -0.00005757, -0.00006630, -0.00005176, -0.00003334, -0.00002050, -0.00001174, -0.00000512, -0.00000374, -0.00001050, -0.00002125, -0.00002654, -0.00002061, -0.00000741, 0.00000308, 0.00000408, -0.00000258, -0.00000963, -0.00001155, -0.00000843, -0.00000393, -0.00000104, -0.00000026, -0.00000088, -0.00000253, -0.00000497, -0.00000713, -0.00000733, -0.00000491, -0.00000122, 0.00000138, 0.00000160, 0.00000017, -0.00000121, -0.00000163, -0.00000151, -0.00000164, },
		{ -0.00000051, -0.00000101, -0.00000238, -0.00000378, -0.00000408, -0.00000304, -0.00000150, -0.00000062, -0.00000098, -0.00000239, -0.00000422, -0.00000583, -0.00000683, -0.00000706, -0.00000660, -0.00000551, -0.00000382, -0.00000177, -0.00000026, -0.00000077, -0.00000394, -0.00000821, -0.00001015, -0.00000726, -0.00000091, 0.00000397, 0.00000314, -0.00000259, -0.00000782, -0.00000781, -0.00000349, -0.00000025, -0.00000160, -0.00000485, -0.00000438, 0.00000074, 0.00000406, -0.00000128, -0.00001285, -0.00001861, -0.00000879, 0.00001159, 0.00002477, 0.00001723, -0.00000550, -0.00002343, -0.00002299, -0.00001140, -0.00000677, -0.00001401, -0.00001590, 0.00000519, 0.00003682, 0.00003718, -0.00002154, -0.00010725, -0.00013728, -0.00005298, 0.00010452, 0.00020285, 0.00012744, -0.00009721, -0.00029701, -0.00028395, -0.00003256, 0.00026815, 0.00035684, 0.00012858, -0.00025335, -0.00047821, -0.00035177, 0.00002391, 0.00032094, 0.00025646, -0.00016368, -0.00063639, -0.00079483, -0.00048754, 0.00012193, 0.00069813, 0.00098190, 0.00093971, 0.00071963, 0.00048998, 0.00031543, 0.00015991, -0.00002202, -0.00021205, -0.00033900, -0.00034808, -0.00025269, -0.00012242, -0.00002714, 0.00000757, 0.00000028, -0.00001800, -0.00003079, -0.00003812, -0.00004284, -0.00004223, -0.00003261, -0.00001726, -0.00000575, -0.00000487, -0.00001197, -0.00001800, -0.00001661, -0.00000948, -0.00000312, -0.00000187, -0.00000455, -0.00000709, -0.00000709, -0.00000535, -0.00000379, -0.00000317, -0.00000306, -0.00000317, -0.00000384, -0.00000511, -0.00000602, -0.00000535, -0.00000308, -0.00000061, 0.00000055, 0.00000026, -0.00000043, -0.00000061, },
		{ -0.00000059, -0.00000061, 0.00000018, 0.00000059, -0.00000049, -0.00000279, -0.00000481, -0.00000522, -0.00000394, -0.00000200, -0.00000055, -0.00000011, -0.00000052, -0.00000141, -0.00000256, -0.00000395, -0.00000564, -0.00000744, -0.00000861, -0.00000809, -0.00000546, -0.00000198, -0.00000027, -0.00000206, -0.00000610, -0.00000861, -0.00000672, -0.00000170, 0.00000191, 0.00000092, -0.00000283, -0.00000436, -0.00000155, 0.00000172, -0.00000040, -0.00000808, -0.00001378, -0.00000963, 0.00000268, 0.00001140, 0.00000617, -0.00000977, -0.00002099, -0.00001532, 0.00000263, 0.00001561, 0.00001254, 0.00000095, -0.00000282, 0.00000424, 0.00000458, -0.00001881, -0.00005382, -0.00006035, -0.00001095, 0.00006667, 0.00009791, 0.00002870, -0.00010653, -0.00019277, -0.00013083, 0.00005756, 0.00022320, 0.00020862, -0.00000295, -0.00025019, -0.00031997, -0.00013694, 0.00015588, 0.00031014, 0.00018788, -0.00011257, -0.00033935, -0.00030913, -0.00007032, 0.00014051, 0.00011028, -0.00016003, -0.00045306, -0.00051777, -0.00027286, 0.00014084, 0.00048893, 0.00062704, 0.00058272, 0.00048602, 0.00043428, 0.00041805, 0.00035960, 0.00020785, -0.00000564, -0.00019339, -0.00028407, -0.00027026, -0.00019779, -0.00012007, -0.00006378, -0.00002886, -0.00000848, -0.00000201, -0.00001091, -0.00002934, -0.00004419, -0.00004527, -0.00003363, -0.00001914, -0.00001058, -0.00000896, -0.00000955, -0.00000850, -0.00000631, -0.00000553, -0.00000682, -0.00000818, -0.00000756, -0.00000513, -0.00000283, -0.00000210, -0.00000279, -0.00000400, -0.00000518, -0.00000626, -0.00000684, -0.00000614, -0.00000397, -0.00000138, 0.00000015, 0.00000009, },
		{ -0.00000076, 0.00000006, -0.00000066, -0.00000117, -0.00000030, 0.00000126, 0.00000171, 0.00000005, -0.00000288, -0.00000517, -0.00000555, -0.00000427, -0.00000251, -0.00000126, -0.00000071, -0.00000046, -0.00000014, 0.00000019, -0.00000012, -0.00000192, -0.00000536, -0.00000922, -0.00001126, -0.00000995, -0.00000619, -0.00000294, -0.00000253, -0.00000422, -0.00000494, -0.00000290, -0.00000012, -0.00000037, -0.00000416, -0.00000692, -0.00000350, 0.00000495, 0.00000999, 0.00000393, -0.00001076, -0.00002152, -0.00001741, -0.00000108, 0.00001251, 0.00001089, -0.00000283, -0.00001346, -0.00001103, -0.00000212, -0.00000090, -0.00000911, -0.00001015, 0.00001095, 0.00004172, 0.00004427, -0.00000640, -0.00008354, -0.00011732, -0.00005731, 0.00006559, 0.00014715, 0.00009751, -0.00006498, -0.00020935, -0.00019859, -0.00001964, 0.00018752, 0.00024233, 0.00008551, -0.00015867, -0.00028494, -0.00018785, 0.00004352, 0.00020588, 0.00016449, -0.00003010, -0.00019783, -0.00020170, -0.00007331, 0.00002511, -0.00003742, -0.00022974, -0.00038086, -0.00033216, -0.00007767, 0.00023246, 0.00042103, 0.00042438, 0.00032402, 0.00025520, 0.00028571, 0.00037074, 0.00040921, 0.00033703, 0.00017349, -0.00000698, -0.00013745, -0.00019703, -0.00020302, -0.00017989, -0.00014039, -0.00008995, -0.00003892, -0.00000354, 0.00000532, -0.00000828, -0.00002865, -0.00004099, -0.00004076, -0.00003262, -0.00002296, -0.00001469, -0.00000794, -0.00000312, -0.00000167, -0.00000394, -0.00000777, -0.00000991, -0.00000876, -0.00000544, -0.00000232, -0.00000098, -0.00000157, -0.00000348, -0.00000599, -0.00000810, -0.00000858, -0.00000678, -0.00000352, },
		{ -0.00000926, -0.00000678, -0.00000322, -0.00000102, -0.00000115, -0.00000224, -0.00000211, -0.00000008, 0.00000227, 0.00000273, 0.00000062, -0.00000263, -0.00000486, -0.00000503, -0.00000375, -0.00000237, -0.00000177, -0.00000185, -0.00000180, -0.00000085, 0.00000101, 0.00000262, 0.00000221, -0.00000123, -0.00000660, -0.00001111, -0.00001241, -0.00001073, -0.00000837, -0.00000691, -0.00000535, -0.00000172, 0.00000319, 0.00000499, 0.00000003, -0.00000902, -0.00001344, -0.00000657, 0.00000788, 0.00001701, 0.00001054, -0.00000793, -0.00002294, -0.00002187, -0.00000759, 0.00000502, 0.00000569, -0.00000029, -0.00000026, 0.00000739, 0.00000775, -0.00001225, -0.00004055, -0.00004257, 0.00000341, 0.00007128, 0.00009654, 0.00003435, -0.00008462, -0.00016336, -0.00011981, 0.00002836, 0.00016169, 0.00015666, 0.00000179, -0.00017872, -0.00022688, -0.00009297, 0.00011361, 0.00021723, 0.00013167, -0.00006420, -0.00020084, -0.00017219, -0.00002538, 0.00009456, 0.00008618, -0.00001925, -0.00010324, -0.00008865, -0.00001978, -0.00001363, -0.00013201, -0.00030289, -0.00037203, -0.00023977, 0.00004384, 0.00031314, 0.00041878, 0.00034058, 0.00018831, 0.00010084, 0.00013784, 0.00025093, 0.00034225, 0.00034605, 0.00026465, 0.00014424, 0.00002877, -0.00006370, -0.00013198, -0.00017322, -0.00017899, -0.00014713, -0.00009176, -0.00003801, -0.00000625, -0.00000068, -0.00001142, -0.00002520, -0.00003397, -0.00003592, -0.00003197, -0.00002362, -0.00001328, -0.00000462, -0.00000082, -0.00000218, -0.00000593, -0.00000852, -0.00000824, -0.00000576, -0.00000285, -0.00000100, -0.00000101, -0.00000302, -0.00000625, -0.00000895, },
		{ -0.00000340, -0.00000701, -0.00000955, -0.00000902, -0.00000575, -0.00000232, -0.00000129, -0.00000277, -0.00000436, -0.00000364, -0.00000057, 0.00000250, 0.00000312, 0.00000097, -0.00000213, -0.00000399, -0.00000383, -0.00000249, -0.00000140, -0.00000146, -0.00000247, -0.00000332, -0.00000270, -0.00000020, 0.00000296, 0.00000443, 0.00000251, -0.00000220, -0.00000725, -0.00001060, -0.00001238, -0.00001377, -0.00001419, -0.00001080, -0.00000206, 0.00000820, 0.00001189, 0.00000409, -0.00001019, -0.00001829, -0.00001137, 0.00000586, 0.00001809, 0.00001393, -0.00000258, -0.00001618, -0.00001667, -0.00000935, -0.00000695, -0.00001190, -0.00001095, 0.00000793, 0.00003392, 0.00003580, -0.00000564, -0.00006612, -0.00008725, -0.00002998, 0.00007551, 0.00014012, 0.00009111, -0.00005174, -0.00017710, -0.00017289, -0.00003092, 0.00013445, 0.00018075, 0.00006455, -0.00011582, -0.00020572, -0.00013163, 0.00003575, 0.00015006, 0.00012284, -0.00000364, -0.00010815, -0.00010978, -0.00003659, 0.00001814, -0.00000071, -0.00005633, -0.00006932, -0.00001465, 0.00003928, -0.00000427, -0.00016279, -0.00033383, -0.00036913, -0.00020128, 0.00009126, 0.00034001, 0.00041429, 0.00031212, 0.00014458, 0.00003900, 0.00004863, 0.00013629, 0.00022670, 0.00026902, 0.00025845, 0.00021469, 0.00015230, 0.00007361, -0.00001735, -0.00010278, -0.00015720, -0.00016546, -0.00013402, -0.00008541, -0.00004265, -0.00001751, -0.00000963, -0.00001285, -0.00002066, -0.00002791, -0.00003076, -0.00002749, -0.00001957, -0.00001100, -0.00000545, -0.00000399, -0.00000501, -0.00000615, -0.00000604, -0.00000460, -0.00000255, -0.00000096, -0.00000105, },
		{ -0.00000090, 0.00000033, -0.00000057, -0.00000402, -0.00000832, -0.00001045, -0.00000866, -0.00000433, -0.00000097, -0.00000108, -0.00000385, -0.00000607, -0.00000522, -0.00000178, 0.00000148, 0.00000226, 0.00000056, -0.00000175, -0.00000281, -0.00000216, -0.00000073, 0.00000016, -0.00000034, -0.00000209, -0.00000389, -0.00000414, -0.00000206, 0.00000128, 0.00000361, 0.00000334, 0.00000098, -0.00000182, -0.00000467, -0.00000901, -0.00001547, -0.00002074, -0.00001880, -0.00000706, 0.00000850, 0.00001626, 0.00000928, -0.00000674, -0.00001729, -0.00001272, 0.00000188, 0.00001182, 0.00000846, -0.00000164, -0.00000528, 0.00000006, 0.00000134, -0.00001381, -0.00003629, -0.00003760, 0.00000014, 0.00005510, 0.00007445, 0.00002333, -0.00007021, -0.00012568, -0.00007916, 0.00004847, 0.00015494, 0.00014066, 0.00000271, -0.00015168, -0.00019332, -0.00008587, 0.00007893, 0.00016127, 0.00009679, -0.00004983, -0.00014891, -0.00012434, -0.00001492, 0.00007394, 0.00007384, 0.00000930, -0.00004244, -0.00003780, -0.00000685, -0.00000918, -0.00005844, -0.00010129, -0.00007636, 0.00000857, 0.00006702, 0.00001130, -0.00015398, -0.00031578, -0.00033713, -0.00017076, 0.00009806, 0.00031600, 0.00037587, 0.00028489, 0.00013938, 0.00004038, 0.00002650, 0.00007075, 0.00012551, 0.00016466, 0.00018903, 0.00020398, 0.00020138, 0.00016525, 0.00009132, -0.00000235, -0.00008479, -0.00013177, -0.00013839, -0.00011647, -0.00008318, -0.00005173, -0.00002889, -0.00001669, -0.00001402, -0.00001722, -0.00002138, -0.00002273, -0.00002045, -0.00001622, -0.00001216, -0.00000926, -0.00000737, -0.00000596, -0.00000456, -0.00000284, },
		{ -0.00000884, -0.00000597, -0.00000275, 0.00000054, 0.00000236, 0.00000090, -0.00000385, -0.00000913, -0.00001117, -0.00000847, -0.00000336, -0.00000003, -0.00000077, -0.00000412, -0.00000655, -0.00000588, -0.00000300, -0.00000049, -0.00000002, -0.00000111, -0.00000209, -0.00000180, -0.00000035, 0.00000127, 0.00000195, 0.00000109, -0.00000106, -0.00000327, -0.00000399, -0.00000267, -0.00000050, 0.00000068, 0.00000047, 0.00000039, 0.00000164, 0.00000235, -0.00000157, -0.00001156, -0.00002235, -0.00002469, -0.00001404, 0.00000321, 0.00001389, 0.00001025, -0.00000205, -0.00000960, -0.00000568, 0.00000237, 0.00000219, -0.00000726, -0.00001191, 0.00000099, 0.00002300, 0.00002666, -0.00000592, -0.00005543, -0.00007335, -0.00002735, 0.00005733, 0.00010770, 0.00006635, -0.00004697, -0.00013970, -0.00012375, 0.00000033, 0.00013369, 0.00016122, 0.00005463, -0.00009909, -0.00017315, -0.00011243, 0.00002138, 0.00011072, 0.00008893, -0.00000746, -0.00008452, -0.00008313, -0.00002616, 0.00001898, 0.00001501, -0.00001310, -0.00001735, 0.00001170, 0.00003164, -0.00000136, -0.00007496, -0.00012286, -0.00009104, -0.00000156, 0.00005591, 0.00000481, -0.00013964, -0.00027069, -0.00027409, -0.00012271, 0.00010083, 0.00027018, 0.00030939, 0.00023662, 0.00013062, 0.00006089, 0.00004440, 0.00005628, 0.00006982, 0.00008221, 0.00010632, 0.00014589, 0.00018324, 0.00019125, 0.00015546, 0.00008501, 0.00000492, -0.00006056, -0.00009925, -0.00011128, -0.00010325, -0.00008328, -0.00005914, -0.00003745, -0.00002246, -0.00001497, -0.00001298, -0.00001367, -0.00001499, -0.00001581, -0.00001553, -0.00001400, -0.00001158, },
		{ -0.00001321, -0.00001408, -0.00001362, -0.00001223, -0.00000965, -0.00000539, -0.00000027, 0.00000310, 0.00000213, -0.00000296, -0.00000861, -0.00001059, -0.00000772, -0.00000284, -0.00000011, -0.00000117, -0.00000413, -0.00000585, -0.00000512, -0.00000332, -0.00000249, -0.00000314, -0.00000400, -0.00000359, -0.00000156, 0.00000118, 0.00000328, 0.00000375, 0.00000233, -0.00000019, -0.00000221, -0.00000236, -0.00000101, -0.00000028, -0.00000183, -0.00000434, -0.00000414, 0.00000060, 0.00000626, 0.00000583, -0.00000393, -0.00001735, -0.00002366, -0.00001725, -0.00000404, 0.00000413, 0.00000216, -0.00000264, 0.00000038, 0.00001047, 0.00001330, -0.00000273, -0.00002806, -0.00003506, -0.00000593, 0.00004120, 0.00006019, 0.00001984, -0.00005725, -0.00010391, -0.00006709, 0.00003543, 0.00011953, 0.00010569, -0.00000471, -0.00012202, -0.00014384, -0.00004760, 0.00008622, 0.00014445, 0.00008143, -0.00004345, -0.00012365, -0.00010183, -0.00001407, 0.00005435, 0.00005193, 0.00000093, -0.00003827, -0.00003358, -0.00000773, -0.00000261, -0.00002648, -0.00004460, -0.00002272, 0.00002705, 0.00005213, 0.00001458, -0.00006295, -0.00011285, -0.00008894, -0.00001732, 0.00002259, -0.00002677, -0.00014078, -0.00022653, -0.00020026, -0.00005892, 0.00011744, 0.00023106, 0.00023940, 0.00017497, 0.00010412, 0.00006958, 0.00006722, 0.00006814, 0.00005589, 0.00004150, 0.00004762, 0.00008261, 0.00013068, 0.00016433, 0.00016442, 0.00013041, 0.00007578, 0.00001699, -0.00003395, -0.00007052, -0.00009001, -0.00009266, -0.00008179, -0.00006311, -0.00004283, -0.00002569, -0.00001426, -0.00000903, -0.00000875, -0.00001098, },
		{ -0.00001466, -0.00000914, -0.00000785, -0.00000833, -0.00000936, -0.00001088, -0.00001239, -0.00001222, -0.00000891, -0.00000338, 0.00000099, 0.00000105, -0.00000296, -0.00000736, -0.00000835, -0.00000551, -0.00000192, -0.00000083, -0.00000255, -0.00000463, -0.00000479, -0.00000331, -0.00000239, -0.00000355, -0.00000599, -0.00000743, -0.00000621, -0.00000254, 0.00000189, 0.00000504, 0.00000553, 0.00000329, -0.00000003, -0.00000178, -0.00000049, 0.00000226, 0.00000262, -0.00000165, -0.00000782, -0.00000979, -0.00000428, 0.00000458, 0.00000817, 0.00000206, -0.00000865, -0.00001427, -0.00001118, -0.00000624, -0.00000827, -0.00001542, -0.00001425, 0.00000462, 0.00003021, 0.00003555, 0.00000457, -0.00004395, -0.00006501, -0.00002861, 0.00004389, 0.00008904, 0.00005670, -0.00003718, -0.00011497, -0.00010333, -0.00000334, 0.00010325, 0.00012364, 0.00003800, -0.00008059, -0.00013106, -0.00007449, 0.00003405, 0.00009967, 0.00007402, -0.00000753, -0.00006813, -0.00006407, -0.00001804, 0.00001528, 0.00000871, -0.00001575, -0.00002048, 0.00000123, 0.00001839, 0.00000124, -0.00004009, -0.00006244, -0.00003647, 0.00001758, 0.00004634, 0.00001742, -0.00004525, -0.00008380, -0.00006582, -0.00002003, -0.00000967, -0.00006762, -0.00015713, -0.00019837, -0.00013788, 0.00000285, 0.00014361, 0.00020970, 0.00018647, 0.00011971, 0.00006892, 0.00005963, 0.00007393, 0.00007897, 0.00006093, 0.00003425, 0.00002385, 0.00004170, 0.00007879, 0.00011497, 0.00013394, 0.00013040, 0.00010770, 0.00007204, 0.00002967, -0.00001304, -0.00004929, -0.00007333, -0.00008234, -0.00007729, -0.00006238, -0.00004347, -0.00002631, },
		{ -0.00006784, -0.00005608, -0.00004131, -0.00002837, -0.00001891, -0.00001209, -0.00000710, -0.00000448, -0.00000513, -0.00000814, -0.00001053, -0.00000973, -0.00000619, -0.00000310, -0.00000314, -0.00000561, -0.00000721, -0.00000562, -0.00000204, 0.00000010, -0.00000115, -0.00000421, -0.00000572, -0.00000420, -0.00000147, -0.00000072, -0.00000332, -0.00000755, -0.00001027, -0.00000931, -0.00000476, 0.00000134, 0.00000614, 0.00000723, 0.00000422, -0.00000036, -0.00000241, 0.00000028, 0.00000520, 0.00000665, 0.00000132, -0.00000753, -0.00001237, -0.00000894, -0.00000112, 0.00000306, 0.00000054, -0.00000301, -0.00000061, 0.00000535, 0.00000300, -0.00001481, -0.00003658, -0.00003734, -0.00000393, 0.00004384, 0.00006285, 0.00002578, -0.00004551, -0.00009013, -0.00006079, 0.00002709, 0.00010042, 0.00009052, -0.00000167, -0.00010046, -0.00012016, -0.00004250, 0.00006568, 0.00011237, 0.00006246, -0.00003423, -0.00009281, -0.00007114, -0.00000167, 0.00004780, 0.00004065, -0.00000091, -0.00002851, -0.00002049, 0.00000115, 0.00000270, -0.00002006, -0.00003786, -0.00002388, 0.00001248, 0.00003301, 0.00001220, -0.00003306, -0.00005898, -0.00004044, 0.00000211, 0.00002373, 0.00000230, -0.00003840, -0.00005513, -0.00003386, -0.00000841, -0.00002712, -0.00009902, -0.00017462, -0.00018357, -0.00009639, 0.00004642, 0.00016452, 0.00019961, 0.00015527, 0.00008428, 0.00004086, 0.00004149, 0.00006373, 0.00007547, 0.00006367, 0.00003984, 0.00002424, 0.00002784, 0.00004711, 0.00007116, 0.00009083, 0.00010192, 0.00010288, 0.00009242, 0.00006988, 0.00003723, -0.00000013, -0.00003471, -0.00005949, -0.00007043, },
		{ 0.00000540, -0.00002468, -0.00004552, -0.00005453, -0.00005393, -0.00004832, -0.00004116, -0.00003312, -0.00002358, -0.00001334, -0.00000523, -0.00000181, -0.00000271, -0.00000489, -0.00000569, -0.00000532, -0.00000598, -0.00000853, -0.00001077, -0.00000954, -0.00000444, 0.00000106, 0.00000269, -0.00000046, -0.00000501, -0.00000664, -0.00000419, -0.00000045, 0.00000071, -0.00000222, -0.00000742, -0.00001152, -0.00001182, -0.00000765, -0.00000057, 0.00000600, 0.00000838, 0.00000528, -0.00000042, -0.00000328, -0.00000015, 0.00000615, 0.00000896, 0.00000444, -0.00000405, -0.00000926, -0.00000815, -0.00000534, -0.00000686, -0.00001080, -0.00000731, 0.00000898, 0.00002704, 0.00002483, -0.00000789, -0.00005121, -0.00006521, -0.00002637, 0.00004267, 0.00008407, 0.00005454, -0.00002974, -0.00009923, -0.00008985, -0.00000338, 0.00008889, 0.00010707, 0.00003458, -0.00006629, -0.00011035, -0.00006519, 0.00002348, 0.00007806, 0.00005994, -0.00000202, -0.00004741, -0.00004394, -0.00001119, 0.00000931, 0.00000056, -0.00001677, -0.00001409, 0.00000930, 0.00002512, 0.00000967, -0.00002670, -0.00004833, -0.00003210, 0.00000682, 0.00002985, 0.00001501, -0.00002095, -0.00004156, -0.00002982, -0.00000507, 0.00000040, -0.00002144, -0.00004463, -0.00003923, -0.00000884, 0.00000646, -0.00003090, -0.00011221, -0.00017940, -0.00016991, -0.00007063, 0.00006787, 0.00016928, 0.00018802, 0.00013703, 0.00006823, 0.00002824, 0.00002824, 0.00004775, 0.00006050, 0.00005613, 0.00004215, 0.00003130, 0.00002988, 0.00003607, 0.00004551, 0.00005604, 0.00006741, 0.00007823, 0.00008446, 0.00008107, 0.00006512, 0.00003799, },
		{ 0.00007196, 0.00007000, 0.00005583, 0.00003243, 0.00000659, -0.00001489, -0.00002891, -0.00003677, -0.00004140, -0.00004382, -0.00004242, -0.00003556, -0.00002454, -0.00001337, -0.00000558, -0.00000162, 0.00000032, 0.00000130, 0.00000008, -0.00000473, -0.00001156, -0.00001577, -0.00001354, -0.00000584, 0.00000180, 0.00000412, 0.00000060, -0.00000452, -0.00000648, -0.00000418, -0.00000041, 0.00000112, -0.00000113, -0.00000599, -0.00001077, -0.00001272, -0.00001009, -0.00000333, 0.00000426, 0.00000807, 0.00000581, 0.00000021, -0.00000288, -0.00000015, 0.00000542, 0.00000755, 0.00000381, -0.00000129, -0.00000188, 0.00000083, -0.00000172, -0.00001449, -0.00002815, -0.00002390, 0.00000634, 0.00004389, 0.00005303, 0.00001409, -0.00005028, -0.00008643, -0.00005578, 0.00002459, 0.00008911, 0.00007896, -0.00000256, -0.00008801, -0.00010339, -0.00003452, 0.00005962, 0.00009976, 0.00005644, -0.00002726, -0.00007911, -0.00006324, -0.00000606, 0.00003696, 0.00003567, 0.00000673, -0.00001341, -0.00000975, 0.00000031, -0.00000633, -0.00002782, -0.00003841, -0.00001909, 0.00001754, 0.00003757, 0.00002068, -0.00001776, -0.00004191, -0.00003171, -0.00000189, 0.00001564, 0.00000589, -0.00001569, -0.00002348, -0.00001194, -0.00000138, -0.00001280, -0.00004057, -0.00005521, -0.00003596, 0.00000118, 0.00001263, -0.00003113, -0.00011102, -0.00016753, -0.00014761, -0.00004938, 0.00007476, 0.00015813, 0.00016755, 0.00012031, 0.00006172, 0.00002783, 0.00002488, 0.00003638, 0.00004419, 0.00004263, 0.00003705, 0.00003389, 0.00003426, 0.00003533, 0.00003542, 0.00003637, 0.00004146, 0.00005170, 0.00006391, },
		{ 0.00002660, 0.00003255, 0.00004414, 0.00005601, 0.00006147, 0.00005656, 0.00004251, 0.00002478, 0.00000891, -0.00000320, -0.00001342, -0.00002393, -0.00003393, -0.00003994, -0.00003931, -0.00003294, -0.00002426, -0.00001590, -0.00000802, -0.00000015, 0.00000600, 0.00000675, 0.00000036, -0.00000987, -0.00001701, -0.00001613, -0.00000850, -0.00000033, 0.00000282, 0.00000046, -0.00000351, -0.00000510, -0.00000353, -0.00000089, 0.00000035, -0.00000086, -0.00000420, -0.00000842, -0.00001140, -0.00001072, -0.00000563, 0.00000127, 0.00000537, 0.00000429, 0.00000066, -0.00000040, 0.00000273, 0.00000568, 0.00000328, -0.00000292, -0.00000437, 0.00000435, 0.00001564, 0.00001273, -0.00001205, -0.00004270, -0.00004802, -0.00001128, 0.00004630, 0.00007617, 0.00004449, -0.00003146, -0.00009076, -0.00008004, -0.00000382, 0.00007479, 0.00008797, 0.00002396, -0.00006170, -0.00009601, -0.00005315, 0.00002561, 0.00007298, 0.00005633, 0.00000105, -0.00004083, -0.00004090, -0.00001413, 0.00000592, 0.00000459, -0.00000386, 0.00000010, 0.00001463, 0.00001815, -0.00000348, -0.00003651, -0.00005035, -0.00002924, 0.00000989, 0.00003311, 0.00002279, -0.00000655, -0.00002535, -0.00001982, -0.00000327, 0.00000217, -0.00000851, -0.00001914, -0.00001368, 0.00000237, 0.00000629, -0.00001540, -0.00004787, -0.00005994, -0.00003704, -0.00000115, 0.00000606, -0.00003675, -0.00010527, -0.00014515, -0.00011606, -0.00002534, 0.00007762, 0.00013955, 0.00014033, 0.00009991, 0.00005514, 0.00003068, 0.00002744, 0.00003210, 0.00003362, 0.00003099, 0.00002912, 0.00003099, 0.00003444, 0.00003527, 0.00003200, 0.00002747, },
		{ 0.00003297, 0.00002911, 0.00002244, 0.00001812, 0.00002049, 0.00002991, 0.00004193, 0.00004992, 0.00004948, 0.00004139, 0.00003042, 0.00002091, 0.00001335, 0.00000508, -0.00000576, -0.00001759, -0.00002669, -0.00003089, -0.00003121, -0.00002969, -0.00002633, -0.00001917, -0.00000782, 0.00000371, 0.00000907, 0.00000510, -0.00000477, -0.00001298, -0.00001420, -0.00000931, -0.00000351, -0.00000104, -0.00000191, -0.00000336, -0.00000338, -0.00000227, -0.00000135, -0.00000120, -0.00000175, -0.00000316, -0.00000564, -0.00000818, -0.00000851, -0.00000556, -0.00000140, 0.00000036, -0.00000094, -0.00000141, 0.00000295, 0.00000976, 0.00001054, 0.00000010, -0.00001422, -0.00001611, 0.00000282, 0.00002914, 0.00003478, 0.00000394, -0.00004514, -0.00006930, -0.00003855, 0.00003090, 0.00008348, 0.00007105, -0.00000129, -0.00007506, -0.00008812, -0.00003028, 0.00004713, 0.00007750, 0.00003816, -0.00003221, -0.00007210, -0.00005290, 0.00000043, 0.00003923, 0.00003764, 0.00001033, -0.00001080, -0.00001138, -0.00000392, -0.00000633, -0.00001792, -0.00002073, -0.00000342, 0.00002147, 0.00002812, 0.00000518, -0.00002993, -0.00004685, -0.00003202, -0.00000118, 0.00001781, 0.00001287, -0.00000320, -0.00001011, -0.00000314, 0.00000379, -0.00000291, -0.00001865, -0.00002508, -0.00001224, 0.00000724, 0.00000956, -0.00001404, -0.00004510, -0.00005485, -0.00003452, -0.00000747, -0.00000856, -0.00004919, -0.00010209, -0.00012164, -0.00008204, 0.00000079, 0.00008202, 0.00012228, 0.00011370, 0.00007829, 0.00004543, 0.00003022, 0.00002913, 0.00003084, 0.00002881, 0.00002480, 0.00002365, 0.00002698, 0.00003162, },
		{ 0.00002146, 0.00002485, 0.00002933, 0.00003056, 0.00002629, 0.00001833, 0.00001174, 0.00001161, 0.00001927, 0.00003084, 0.00003978, 0.00004168, 0.00003722, 0.00003055, 0.00002490, 0.00001991, 0.00001323, 0.00000424, -0.00000471, -0.00001091, -0.00001455, -0.00001845, -0.00002409, -0.00002873, -0.00002739, -0.00001820, -0.00000547, 0.00000338, 0.00000410, -0.00000101, -0.00000609, -0.00000735, -0.00000579, -0.00000463, -0.00000516, -0.00000581, -0.00000478, -0.00000253, -0.00000114, -0.00000161, -0.00000290, -0.00000361, -0.00000376, -0.00000416, -0.00000450, -0.00000349, -0.00000136, -0.00000111, -0.00000511, -0.00001023, -0.00000863, 0.00000378, 0.00001925, 0.00002180, 0.00000305, -0.00002459, -0.00003432, -0.00000998, 0.00003308, 0.00005572, 0.00002973, -0.00003120, -0.00007706, -0.00006407, 0.00000310, 0.00007087, 0.00008219, 0.00002767, -0.00004555, -0.00007624, -0.00004335, 0.00001815, 0.00005274, 0.00003512, -0.00001145, -0.00004292, -0.00003703, -0.00000874, 0.00001187, 0.00001169, 0.00000283, 0.00000260, 0.00001124, 0.00001300, -0.00000237, -0.00002427, -0.00003069, -0.00001258, 0.00001436, 0.00002446, 0.00000799, -0.00001931, -0.00003284, -0.00002377, -0.00000564, 0.00000261, -0.00000287, -0.00000907, -0.00000377, 0.00000889, 0.00001290, -0.00000017, -0.00001966, -0.00002616, -0.00001313, 0.00000448, 0.00000530, -0.00001530, -0.00003903, -0.00004339, -0.00002647, -0.00001124, -0.00002329, -0.00006332, -0.00010183, -0.00010248, -0.00005310, 0.00002335, 0.00008676, 0.00010958, 0.00009328, 0.00006075, 0.00003541, 0.00002599, 0.00002674, 0.00002802, 0.00002566, 0.00002208, },
		{ 0.00002297, 0.00002139, 0.00001924, 0.00001970, 0.00002358, 0.00002837, 0.00002995, 0.00002555, 0.00001633, 0.00000735, 0.00000440, 0.00000971, 0.00002010, 0.00002957, 0.00003400, 0.00003366, 0.00003119, 0.00002786, 0.00002245, 0.00001396, 0.00000457, -0.00000140, -0.00000188, 0.00000001, -0.00000158, -0.00000931, -0.00001951, -0.00002532, -0.00002314, -0.00001561, -0.00000846, -0.00000487, -0.00000347, -0.00000151, 0.00000102, 0.00000146, -0.00000179, -0.00000657, -0.00000885, -0.00000691, -0.00000305, -0.00000091, -0.00000178, -0.00000405, -0.00000572, -0.00000629, -0.00000568, -0.00000283, 0.00000264, 0.00000733, 0.00000478, -0.00000747, -0.00002135, -0.00002189, -0.00000180, 0.00002658, 0.00003739, 0.00001478, -0.00002736, -0.00005203, -0.00003152, 0.00002298, 0.00006571, 0.00005543, -0.00000481, -0.00006628, -0.00007616, -0.00002514, 0.00004364, 0.00007329, 0.00004343, -0.00001464, -0.00004999, -0.00003864, -0.00000066, 0.00002500, 0.00001893, -0.00000500, -0.00002002, -0.00001553, -0.00000392, -0.00000176, -0.00000927, -0.00001168, 0.00000093, 0.00001949, 0.00002446, 0.00000768, -0.00001717, -0.00002789, -0.00001640, 0.00000350, 0.00001131, 0.00000107, -0.00001418, -0.00001828, -0.00000969, -0.00000149, -0.00000479, -0.00001526, -0.00001814, -0.00000602, 0.00001132, 0.00001669, 0.00000421, -0.00001380, -0.00001996, -0.00001068, 0.00000012, -0.00000335, -0.00002067, -0.00003506, -0.00003195, -0.00001671, -0.00001155, -0.00003313, -0.00007286, -0.00010019, -0.00008715, -0.00003218, 0.00003776, 0.00008771, 0.00009927, 0.00007939, 0.00004972, 0.00002870, 0.00002138, 0.00002200, },
		{ 0.00002559, 0.00001885, 0.00001831, 0.00001842, 0.00001696, 0.00001533, 0.00001611, 0.00002042, 0.00002636, 0.00002971, 0.00002679, 0.00001771, 0.00000710, 0.00000098, 0.00000229, 0.00000907, 0.00001703, 0.00002351, 0.00002858, 0.00003247, 0.00003305, 0.00002709, 0.00001466, 0.00000128, -0.00000555, -0.00000287, 0.00000477, 0.00000921, 0.00000585, -0.00000306, -0.00001174, -0.00001668, -0.00001860, -0.00001933, -0.00001823, -0.00001293, -0.00000350, 0.00000554, 0.00000846, 0.00000381, -0.00000393, -0.00000860, -0.00000784, -0.00000426, -0.00000178, -0.00000192, -0.00000386, -0.00000674, -0.00000994, -0.00001103, -0.00000585, 0.00000650, 0.00001845, 0.00001715, -0.00000294, -0.00002960, -0.00003863, -0.00001550, 0.00002630, 0.00005142, 0.00003309, -0.00001876, -0.00006096, -0.00005373, 0.00000139, 0.00005935, 0.00007005, 0.00002349, -0.00004079, -0.00006925, -0.00004164, 0.00001377, 0.00004924, 0.00004094, 0.00000570, -0.00002110, -0.00002081, -0.00000446, 0.00000518, -0.00000051, -0.00001006, -0.00000914, 0.00000124, 0.00000653, -0.00000283, -0.00001856, -0.00002263, -0.00000739, 0.00001497, 0.00002441, 0.00001351, -0.00000576, -0.00001556, -0.00001068, -0.00000198, -0.00000220, -0.00001115, -0.00001684, -0.00001060, 0.00000181, 0.00000614, -0.00000401, -0.00001904, -0.00002308, -0.00001095, 0.00000612, 0.00001251, 0.00000442, -0.00000721, -0.00000992, -0.00000394, -0.00000126, -0.00001083, -0.00002679, -0.00003381, -0.00002452, -0.00001018, -0.00001167, -0.00003828, -0.00007528, -0.00009333, -0.00007211, -0.00001721, 0.00004404, 0.00008291, 0.00008801, 0.00006859, 0.00004317, },
		{ 0.00007525, 0.00005803, 0.00003736, 0.00002319, 0.00001745, 0.00001675, 0.00001674, 0.00001513, 0.00001260, 0.00001181, 0.00001493, 0.00002119, 0.00002658, 0.00002667, 0.00002037, 0.00001105, 0.00000359, 0.00000053, 0.00000108, 0.00000382, 0.00000922, 0.00001851, 0.00002992, 0.00003730, 0.00003416, 0.00002002, 0.00000229, -0.00000891, -0.00000870, -0.00000055, 0.00000794, 0.00001169, 0.00001061, 0.00000659, -0.00000009, -0.00001012, -0.00002142, -0.00002805, -0.00002454, -0.00001170, 0.00000277, 0.00001026, 0.00000823, 0.00000121, -0.00000445, -0.00000608, -0.00000506, -0.00000341, -0.00000188, -0.00000150, -0.00000480, -0.00001254, -0.00001924, -0.00001501, 0.00000424, 0.00002755, 0.00003363, 0.00001011, -0.00002982, -0.00005306, -0.00003471, 0.00001564, 0.00005693, 0.00005112, -0.00000081, -0.00005623, -0.00006706, -0.00002295, 0.00003900, 0.00006745, 0.00004200, -0.00001135, -0.00004688, -0.00004042, -0.00000648, 0.00002156, 0.00002442, 0.00001024, -0.00000086, -0.00000028, 0.00000315, -0.00000198, -0.00001365, -0.00001845, -0.00000835, 0.00000815, 0.00001389, 0.00000169, -0.00001723, -0.00002452, -0.00001339, 0.00000520, 0.00001489, 0.00001051, 0.00000123, -0.00000197, 0.00000098, 0.00000064, -0.00000883, -0.00002086, -0.00002280, -0.00001052, 0.00000558, 0.00001042, -0.00000024, -0.00001549, -0.00002080, -0.00001262, -0.00000074, 0.00000350, -0.00000072, -0.00000458, -0.00000166, 0.00000347, 0.00000041, -0.00001391, -0.00002943, -0.00003236, -0.00002070, -0.00000880, -0.00001470, -0.00004174, -0.00007240, -0.00008120, -0.00005529, -0.00000468, 0.00004577, 0.00007434, },
		{ 0.00000676, 0.00004635, 0.00006581, 0.00006344, 0.00004796, 0.00003056, 0.00001870, 0.00001433, 0.00001518, 0.00001708, 0.00001664, 0.00001340, 0.00001008, 0.00001009, 0.00001418, 0.00001939, 0.00002173, 0.00001975, 0.00001526, 0.00001058, 0.00000594, 0.00000041, -0.00000471, -0.00000483, 0.00000412, 0.00002017, 0.00003458, 0.00003788, 0.00002760, 0.00001028, -0.00000417, -0.00001003, -0.00000822, -0.00000252, 0.00000475, 0.00001278, 0.00001891, 0.00001790, 0.00000637, -0.00001202, -0.00002724, -0.00003029, -0.00002059, -0.00000596, 0.00000454, 0.00000748, 0.00000537, 0.00000182, -0.00000186, -0.00000509, -0.00000572, -0.00000189, 0.00000329, 0.00000180, -0.00001092, -0.00002703, -0.00002906, -0.00000640, 0.00002932, 0.00004859, 0.00002922, -0.00001935, -0.00005858, -0.00005280, -0.00000308, 0.00005003, 0.00006075, 0.00001909, -0.00003952, -0.00006574, -0.00003968, 0.00001368, 0.00004972, 0.00004412, 0.00001012, -0.00001955, -0.00002457, -0.00001106, 0.00000187, 0.00000433, 0.00000246, 0.00000582, 0.00001309, 0.00001336, 0.00000065, -0.00001647, -0.00002235, -0.00001096, 0.00000686, 0.00001408, 0.00000462, -0.00001129, -0.00001862, -0.00001272, -0.00000214, 0.00000270, 0.00000103, 0.00000047, 0.00000574, 0.00001137, 0.00000756, -0.00000740, -0.00002286, -0.00002558, -0.00001331, 0.00000254, 0.00000824, 0.00000092, -0.00000997, -0.00001377, -0.00000937, -0.00000439, -0.00000480, -0.00000780, -0.00000616, 0.00000150, 0.00000722, 0.00000191, -0.00001351, -0.00002722, -0.00002802, -0.00001769, -0.00001077, -0.00002068, -0.00004554, -0.00006752, -0.00006690, -0.00003796, },
		{ -0.00006225, -0.00005358, -0.00002314, 0.00001570, 0.00004647, 0.00005956, 0.00005534, 0.00004075, 0.00002422, 0.00001232, 0.00000821, 0.00001094, 0.00001614, 0.00001881, 0.00001694, 0.00001264, 0.00000965, 0.00000964, 0.00001118, 0.00001227, 0.00001292, 0.00001443, 0.00001625, 0.00001490, 0.00000737, -0.00000409, -0.00001176, -0.00000843, 0.00000573, 0.00002240, 0.00003171, 0.00002975, 0.00002006, 0.00000871, -0.00000097, -0.00000842, -0.00001227, -0.00000909, 0.00000261, 0.00001781, 0.00002640, 0.00002097, 0.00000374, -0.00001475, -0.00002437, -0.00002277, -0.00001488, -0.00000643, 0.00000050, 0.00000574, 0.00000743, 0.00000309, -0.00000521, -0.00000951, -0.00000319, 0.00000937, 0.00001332, -0.00000235, -0.00002962, -0.00004347, -0.00002406, 0.00002047, 0.00005530, 0.00004786, -0.00000049, -0.00005153, -0.00006207, -0.00002276, 0.00003264, 0.00005715, 0.00003201, -0.00001847, -0.00005109, -0.00004244, -0.00000588, 0.00002578, 0.00003156, 0.00001690, 0.00000123, -0.00000384, -0.00000258, -0.00000418, -0.00000901, -0.00000859, 0.00000197, 0.00001532, 0.00001812, 0.00000573, -0.00001168, -0.00001885, -0.00001074, 0.00000302, 0.00000881, 0.00000283, -0.00000674, -0.00001020, -0.00000680, -0.00000372, -0.00000578, -0.00000894, -0.00000543, 0.00000564, 0.00001510, 0.00001256, -0.00000213, -0.00001762, -0.00002183, -0.00001347, -0.00000226, 0.00000173, -0.00000231, -0.00000715, -0.00000704, -0.00000402, -0.00000413, -0.00000883, -0.00001241, -0.00000875, 0.00000036, 0.00000573, 0.00000027, -0.00001274, -0.00002225, -0.00002098, -0.00001367, -0.00001309, -0.00002696, -0.00004888, },
		{ -0.00001442, -0.00003094, -0.00004948, -0.00005577, -0.00004208, -0.00001266, 0.00002020, 0.00004463, 0.00005481, 0.00005115, 0.00003781, 0.00002086, 0.00000698, 0.00000119, 0.00000413, 0.00001159, 0.00001767, 0.00001921, 0.00001720, 0.00001416, 0.00001101, 0.00000709, 0.00000304, 0.00000198, 0.00000663, 0.00001498, 0.00002025, 0.00001644, 0.00000449, -0.00000764, -0.00001158, -0.00000549, 0.00000548, 0.00001489, 0.00002027, 0.00002258, 0.00002187, 0.00001568, 0.00000292, -0.00001154, -0.00001840, -0.00001170, 0.00000479, 0.00001971, 0.00002346, 0.00001554, 0.00000296, -0.00000748, -0.00001402, -0.00001756, -0.00001699, -0.00000974, 0.00000243, 0.00001092, 0.00000717, -0.00000660, -0.00001608, -0.00000802, 0.00001368, 0.00002739, 0.00001354, -0.00002294, -0.00005176, -0.00004313, 0.00000225, 0.00004937, 0.00005809, 0.00001972, -0.00003383, -0.00005799, -0.00003498, 0.00001200, 0.00004181, 0.00003247, -0.00000241, -0.00003067, -0.00003214, -0.00001303, 0.00000618, 0.00001275, 0.00001038, 0.00000912, 0.00001134, 0.00001026, 0.00000096, -0.00001098, -0.00001426, -0.00000460, 0.00000942, 0.00001465, 0.00000684, -0.00000551, -0.00001067, -0.00000576, 0.00000157, 0.00000282, -0.00000204, -0.00000594, -0.00000411, -0.00000011, -0.00000108, -0.00000831, -0.00001444, -0.00001122, 0.00000080, 0.00001166, 0.00001177, 0.00000144, -0.00000981, -0.00001319, -0.00000887, -0.00000409, -0.00000424, -0.00000735, -0.00000782, -0.00000405, -0.00000066, -0.00000270, -0.00000895, -0.00001265, -0.00000920, -0.00000194, 0.00000093, -0.00000462, -0.00001371, -0.00001769, -0.00001390, -0.00000947, },
		{ -0.00001503, -0.00000938, -0.00000732, -0.00001569, -0.00003221, -0.00004612, -0.00004670, -0.00003121, -0.00000575, 0.00002025, 0.00003974, 0.00004941, 0.00004838, 0.00003758, 0.00002096, 0.00000532, -0.00000290, -0.00000188, 0.00000479, 0.00001178, 0.00001635, 0.00001885, 0.00001970, 0.00001722, 0.00000980, -0.00000033, -0.00000680, -0.00000423, 0.00000616, 0.00001668, 0.00001965, 0.00001382, 0.00000442, -0.00000269, -0.00000587, -0.00000671, -0.00000542, 0.00000056, 0.00001241, 0.00002522, 0.00002975, 0.00002045, 0.00000185, -0.00001413, -0.00001806, -0.00001023, 0.00000156, 0.00001023, 0.00001446, 0.00001571, 0.00001299, 0.00000340, -0.00001140, -0.00002199, -0.00001831, -0.00000151, 0.00001343, 0.00001064, -0.00000910, -0.00002572, -0.00001820, 0.00001225, 0.00003896, 0.00003323, -0.00000637, -0.00004851, -0.00005565, -0.00001900, 0.00003184, 0.00005472, 0.00003254, -0.00001282, -0.00004220, -0.00003463, -0.00000307, 0.00002189, 0.00002154, 0.00000281, -0.00001377, -0.00001592, -0.00000836, -0.00000262, -0.00000253, -0.00000181, 0.00000545, 0.00001556, 0.00001850, 0.00000962, -0.00000419, -0.00001121, -0.00000687, 0.00000219, 0.00000609, 0.00000229, -0.00000298, -0.00000302, 0.00000142, 0.00000368, -0.00000014, -0.00000619, -0.00000732, -0.00000190, 0.00000368, 0.00000201, -0.00000672, -0.00001432, -0.00001311, -0.00000399, 0.00000464, 0.00000575, 0.00000045, -0.00000441, -0.00000433, -0.00000166, -0.00000177, -0.00000603, -0.00001013, -0.00000936, -0.00000451, -0.00000114, -0.00000309, -0.00000796, -0.00001000, -0.00000698, -0.00000307, -0.00000405, -0.00001032, -0.00001589, },
		{ -0.00000752, -0.00001431, -0.00001720, -0.00001348, -0.00000790, -0.00000850, -0.00001846, -0.00003223, -0.00003988, -0.00003527, -0.00001998, -0.00000033, 0.00001819, 0.00003302, 0.00004256, 0.00004430, 0.00003661, 0.00002195, 0.00000698, -0.00000191, -0.00000313, 0.00000005, 0.00000402, 0.00000834, 0.00001429, 0.00002070, 0.00002281, 0.00001656, 0.00000380, -0.00000777, -0.00001084, -0.00000462, 0.00000527, 0.00001265, 0.00001562, 0.00001560, 0.00001292, 0.00000552, -0.00000685, -0.00001848, -0.00002004, -0.00000700, 0.00001444, 0.00003089, 0.00003250, 0.00002035, 0.00000377, -0.00000883, -0.00001533, -0.00001698, -0.00001299, -0.00000104, 0.00001609, 0.00002751, 0.00002190, 0.00000036, -0.00002045, -0.00002195, -0.00000264, 0.00001764, 0.00001578, -0.00001027, -0.00003616, -0.00003314, 0.00000252, 0.00004226, 0.00004991, 0.00001576, -0.00003275, -0.00005515, -0.00003427, 0.00000953, 0.00003860, 0.00003226, 0.00000237, -0.00002202, -0.00002301, -0.00000718, 0.00000613, 0.00000593, -0.00000215, -0.00000616, -0.00000247, 0.00000169, -0.00000112, -0.00000859, -0.00001104, -0.00000289, 0.00001046, 0.00001796, 0.00001427, 0.00000449, -0.00000232, -0.00000269, -0.00000080, -0.00000175, -0.00000484, -0.00000491, 0.00000041, 0.00000665, 0.00000703, 0.00000035, -0.00000707, -0.00000793, -0.00000193, 0.00000390, 0.00000282, -0.00000443, -0.00001084, -0.00001077, -0.00000566, -0.00000152, -0.00000184, -0.00000405, -0.00000346, 0.00000070, 0.00000407, 0.00000234, -0.00000359, -0.00000849, -0.00000863, -0.00000563, -0.00000407, -0.00000582, -0.00000808, -0.00000727, -0.00000408, -0.00000315, },
		{ -0.00000598, -0.00000276, -0.00000399, -0.00001006, -0.00001597, -0.00001629, -0.00001146, -0.00000801, -0.00001209, -0.00002279, -0.00003235, -0.00003306, -0.00002372, -0.00000932, 0.00000468, 0.00001663, 0.00002736, 0.00003590, 0.00003849, 0.00003244, 0.00002027, 0.00000860, 0.00000249, 0.00000135, 0.00000103, -0.00000043, -0.00000030, 0.00000508, 0.00001464, 0.00002192, 0.00002070, 0.00001106, -0.00000075, -0.00000818, -0.00000945, -0.00000667, -0.00000142, 0.00000690, 0.00001756, 0.00002496, 0.00002125, 0.00000427, -0.00001756, -0.00003008, -0.00002466, -0.00000511, 0.00001628, 0.00002922, 0.00003151, 0.00002586, 0.00001382, -0.00000454, -0.00002459, -0.00003460, -0.00002391, 0.00000430, 0.00003047, 0.00003345, 0.00001086, -0.00001615, -0.00002153, -0.00000020, 0.00002505, 0.00002489, -0.00000647, -0.00004268, -0.00004861, -0.00001425, 0.00003400, 0.00005628, 0.00003521, -0.00000950, -0.00004049, -0.00003640, -0.00000788, 0.00001683, 0.00001960, 0.00000581, -0.00000655, -0.00000714, -0.00000134, -0.00000019, -0.00000621, -0.00001125, -0.00000735, 0.00000305, 0.00000943, 0.00000490, -0.00000617, -0.00001275, -0.00000857, 0.00000228, 0.00001071, 0.00001202, 0.00000897, 0.00000637, 0.00000492, 0.00000156, -0.00000473, -0.00000982, -0.00000834, -0.00000032, 0.00000759, 0.00000843, 0.00000201, -0.00000500, -0.00000615, -0.00000173, 0.00000232, 0.00000124, -0.00000362, -0.00000711, -0.00000651, -0.00000426, -0.00000440, -0.00000723, -0.00000878, -0.00000558, 0.00000092, 0.00000555, 0.00000484, 0.00000036, -0.00000355, -0.00000473, -0.00000490, -0.00000650, -0.00000891, -0.00000910, },
		{ -0.00000659, -0.00000984, -0.00000909, -0.00000540, -0.00000347, -0.00000674, -0.00001308, -0.00001656, -0.00001389, -0.00000869, -0.00000829, -0.00001598, -0.00002690, -0.00003217, -0.00002701, -0.00001452, -0.00000153, 0.00000825, 0.00001631, 0.00002488, 0.00003210, 0.00003307, 0.00002571, 0.00001446, 0.00000679, 0.00000604, 0.00000836, 0.00000741, 0.00000166, -0.00000383, -0.00000298, 0.00000461, 0.00001321, 0.00001672, 0.00001400, 0.00000821, 0.00000207, -0.00000439, -0.00001094, -0.00001369, -0.00000718, 0.00000879, 0.00002538, 0.00003011, 0.00001753, -0.00000520, -0.00002400, -0.00002917, -0.00002128, -0.00000677, 0.00000967, 0.00002643, 0.00003938, 0.00003907, 0.00001812, -0.00001696, -0.00004444, -0.00004298, -0.00001215, 0.00002356, 0.00003476, 0.00001441, -0.00001467, -0.00002133, 0.00000281, 0.00003373, 0.00003755, 0.00000431, -0.00004040, -0.00005795, -0.00003218, 0.00001611, 0.00004867, 0.00004367, 0.00001196, -0.00001724, -0.00002423, -0.00001291, -0.00000066, 0.00000167, -0.00000185, -0.00000153, 0.00000444, 0.00000804, 0.00000202, -0.00001023, -0.00001735, -0.00001198, 0.00000131, 0.00001070, 0.00000888, -0.00000056, -0.00000813, -0.00000818, -0.00000286, 0.00000237, 0.00000525, 0.00000748, 0.00001019, 0.00001088, 0.00000620, -0.00000280, -0.00000997, -0.00000958, -0.00000231, 0.00000512, 0.00000660, 0.00000232, -0.00000225, -0.00000275, -0.00000021, 0.00000105, -0.00000109, -0.00000426, -0.00000492, -0.00000293, -0.00000179, -0.00000429, -0.00000886, -0.00001094, -0.00000790, -0.00000180, 0.00000293, 0.00000401, 0.00000283, 0.00000166, 0.00000054, -0.00000219, },
		{ 0.00000464, 0.00000436, 0.00000058, -0.00000483, -0.00000781, -0.00000647, -0.00000376, -0.00000471, -0.00001060, -0.00001683, -0.00001741, -0.00001186, -0.00000664, -0.00000894, -0.00001890, -0.00002874, -0.00002993, -0.00002109, -0.00000833, 0.00000177, 0.00000850, 0.00001572, 0.00002510, 0.00003217, 0.00003050, 0.00001945, 0.00000656, 0.00000113, 0.00000545, 0.00001285, 0.00001469, 0.00000870, 0.00000021, -0.00000412, -0.00000265, 0.00000139, 0.00000490, 0.00000791, 0.00001114, 0.00001223, 0.00000698, -0.00000483, -0.00001596, -0.00001660, -0.00000379, 0.00001433, 0.00002493, 0.00002146, 0.00000806, -0.00000629, -0.00001658, -0.00002315, -0.00002527, -0.00001716, 0.00000589, 0.00003654, 0.00005441, 0.00004132, 0.00000068, -0.00003997, -0.00005008, -0.00002359, 0.00001416, 0.00002911, 0.00001052, -0.00001891, -0.00002550, 0.00000162, 0.00003865, 0.00004938, 0.00001988, -0.00002773, -0.00005542, -0.00004308, -0.00000414, 0.00002970, 0.00003730, 0.00002265, 0.00000463, -0.00000372, -0.00000452, -0.00000654, -0.00001191, -0.00001375, -0.00000613, 0.00000675, 0.00001340, 0.00000693, -0.00000733, -0.00001684, -0.00001404, -0.00000279, 0.00000660, 0.00000775, 0.00000280, -0.00000200, -0.00000335, -0.00000270, -0.00000221, -0.00000112, 0.00000269, 0.00000855, 0.00001196, 0.00000895, 0.00000087, -0.00000617, -0.00000708, -0.00000258, 0.00000208, 0.00000286, 0.00000064, -0.00000080, 0.00000032, 0.00000188, 0.00000087, -0.00000256, -0.00000507, -0.00000416, -0.00000122, -0.00000015, -0.00000296, -0.00000752, -0.00000988, -0.00000837, -0.00000471, -0.00000147, 0.00000072, 0.00000278, },
		{ -0.00000513, -0.00000277, 0.00000127, 0.00000493, 0.00000512, 0.00000128, -0.00000320, -0.00000419, -0.00000164, -0.00000031, -0.00000469, -0.00001341, -0.00001970, -0.00001823, -0.00001128, -0.00000709, -0.00001150, -0.00002152, -0.00002795, -0.00002451, -0.00001379, -0.00000373, 0.00000137, 0.00000495, 0.00001303, 0.00002551, 0.00003428, 0.00003095, 0.00001619, 0.00000040, -0.00000528, 0.00000117, 0.00001194, 0.00001761, 0.00001526, 0.00000880, 0.00000280, -0.00000182, -0.00000602, -0.00000835, -0.00000514, 0.00000430, 0.00001407, 0.00001564, 0.00000632, -0.00000698, -0.00001371, -0.00000955, 0.00000032, 0.00000777, 0.00001016, 0.00001079, 0.00001129, 0.00000676, -0.00000832, -0.00002896, -0.00003795, -0.00002011, 0.00001945, 0.00005351, 0.00005380, 0.00001751, -0.00002718, -0.00004477, -0.00002455, 0.00000985, 0.00002347, 0.00000401, -0.00002662, -0.00003423, -0.00000640, 0.00003491, 0.00005383, 0.00003377, -0.00000860, -0.00004015, -0.00004078, -0.00001739, 0.00000781, 0.00001939, 0.00001882, 0.00001596, 0.00001500, 0.00001116, -0.00000017, -0.00001455, -0.00002107, -0.00001380, 0.00000115, 0.00001093, 0.00000813, -0.00000288, -0.00001141, -0.00001102, -0.00000435, 0.00000158, 0.00000303, 0.00000184, 0.00000128, 0.00000163, 0.00000075, -0.00000190, -0.00000355, -0.00000119, 0.00000438, 0.00000860, 0.00000774, 0.00000282, -0.00000161, -0.00000231, -0.00000036, 0.00000081, -0.00000030, -0.00000172, -0.00000092, 0.00000173, 0.00000312, 0.00000110, -0.00000285, -0.00000511, -0.00000388, -0.00000104, 0.00000005, -0.00000185, -0.00000494, -0.00000678, -0.00000686, -0.00000618, },
		{ -0.00000330, -0.00000452, -0.00000622, -0.00000687, -0.00000458, 0.00000014, 0.00000381, 0.00000325, -0.00000063, -0.00000320, -0.00000111, 0.00000328, 0.00000374, -0.00000321, -0.00001364, -0.00001959, -0.00001721, -0.00001123, -0.00001015, -0.00001697, -0.00002540, -0.00002618, -0.00001696, -0.00000488, 0.00000104, -0.00000016, -0.00000035, 0.00000839, 0.00002400, 0.00003510, 0.00003176, 0.00001553, -0.00000166, -0.00000837, -0.00000297, 0.00000717, 0.00001422, 0.00001628, 0.00001566, 0.00001318, 0.00000670, -0.00000436, -0.00001494, -0.00001723, -0.00000850, 0.00000506, 0.00001357, 0.00001258, 0.00000620, 0.00000096, -0.00000179, -0.00000563, -0.00001134, -0.00001242, -0.00000163, 0.00001744, 0.00002829, 0.00001560, -0.00001670, -0.00004347, -0.00003886, -0.00000174, 0.00003963, 0.00005119, 0.00002431, -0.00001581, -0.00003367, -0.00001683, 0.00001322, 0.00002321, 0.00000123, -0.00003182, -0.00004292, -0.00001889, 0.00002162, 0.00004630, 0.00003836, 0.00000873, -0.00001765, -0.00002515, -0.00001729, -0.00000694, -0.00000092, 0.00000427, 0.00001341, 0.00002321, 0.00002434, 0.00001218, -0.00000641, -0.00001842, -0.00001670, -0.00000600, 0.00000274, 0.00000308, -0.00000220, -0.00000605, -0.00000509, -0.00000190, -0.00000048, -0.00000132, -0.00000155, 0.00000066, 0.00000350, 0.00000375, 0.00000093, -0.00000193, -0.00000170, 0.00000129, 0.00000387, 0.00000372, 0.00000184, 0.00000095, 0.00000191, 0.00000281, 0.00000152, -0.00000135, -0.00000292, -0.00000139, 0.00000167, 0.00000292, 0.00000085, -0.00000264, -0.00000445, -0.00000351, -0.00000153, -0.00000068, -0.00000133, -0.00000241, },
		{ -0.00000161, -0.00000104, -0.00000034, -0.00000049, -0.00000257, -0.00000568, -0.00000689, -0.00000424, 0.00000047, 0.00000280, 0.00000036, -0.00000397, -0.00000465, 0.00000031, 0.00000625, 0.00000615, -0.00000184, -0.00001177, -0.00001582, -0.00001279, -0.00000966, -0.00001384, -0.00002417, -0.00003088, -0.00002543, -0.00001007, 0.00000355, 0.00000587, -0.00000068, -0.00000416, 0.00000435, 0.00002104, 0.00003270, 0.00002956, 0.00001441, -0.00000082, -0.00000659, -0.00000309, 0.00000309, 0.00000744, 0.00001106, 0.00001607, 0.00001993, 0.00001676, 0.00000443, -0.00001108, -0.00002013, -0.00001821, -0.00000926, -0.00000018, 0.00000655, 0.00001297, 0.00001874, 0.00001724, 0.00000224, -0.00002115, -0.00003516, -0.00002394, 0.00000865, 0.00003735, 0.00003614, 0.00000336, -0.00003383, -0.00004281, -0.00001576, 0.00002255, 0.00003793, 0.00001908, -0.00001265, -0.00002517, -0.00000743, 0.00002050, 0.00002819, 0.00000517, -0.00002924, -0.00004496, -0.00002922, 0.00000361, 0.00002775, 0.00002953, 0.00001584, 0.00000239, -0.00000295, -0.00000440, -0.00000843, -0.00001356, -0.00001167, 0.00000128, 0.00001846, 0.00002729, 0.00002131, 0.00000600, -0.00000701, -0.00001081, -0.00000785, -0.00000497, -0.00000542, -0.00000663, -0.00000496, -0.00000075, 0.00000209, 0.00000090, -0.00000249, -0.00000396, -0.00000166, 0.00000215, 0.00000402, 0.00000302, 0.00000117, 0.00000053, 0.00000092, 0.00000077, -0.00000029, -0.00000067, 0.00000098, 0.00000368, 0.00000472, 0.00000272, -0.00000065, -0.00000230, -0.00000103, 0.00000130, 0.00000189, 0.00000001, -0.00000254, -0.00000370, -0.00000319, -0.00000223, },
		{ -0.00000318, -0.00000321, -0.00000309, -0.00000242, -0.00000067, 0.00000129, 0.00000131, -0.00000159, -0.00000525, -0.00000592, -0.00000236, 0.00000207, 0.00000244, -0.00000218, -0.00000697, -0.00000609, 0.00000065, 0.00000676, 0.00000552, -0.00000254, -0.00000984, -0.00001001, -0.00000548, -0.00000546, -0.00001557, -0.00003005, -0.00003593, -0.00002580, -0.00000609, 0.00000832, 0.00000836, -0.00000059, -0.00000471, 0.00000392, 0.00001938, 0.00002827, 0.00002331, 0.00000984, -0.00000068, -0.00000230, 0.00000131, 0.00000336, 0.00000262, 0.00000386, 0.00001053, 0.00001883, 0.00002084, 0.00001295, -0.00000030, -0.00001128, -0.00001680, -0.00001887, -0.00001838, -0.00001115, 0.00000660, 0.00002850, 0.00003761, 0.00002038, -0.00001675, -0.00004710, -0.00004477, -0.00000914, 0.00003154, 0.00004395, 0.00001967, -0.00001730, -0.00003346, -0.00001720, 0.00001167, 0.00002319, 0.00000722, -0.00001749, -0.00002356, -0.00000302, 0.00002489, 0.00003314, 0.00001306, -0.00001854, -0.00003657, -0.00003052, -0.00001076, 0.00000520, 0.00000989, 0.00000921, 0.00001128, 0.00001546, 0.00001378, 0.00000202, -0.00001318, -0.00001957, -0.00001119, 0.00000570, 0.00001879, 0.00002075, 0.00001414, 0.00000639, 0.00000161, -0.00000175, -0.00000645, -0.00001151, -0.00001284, -0.00000834, -0.00000099, 0.00000371, 0.00000299, -0.00000090, -0.00000355, -0.00000280, -0.00000008, 0.00000211, 0.00000288, 0.00000302, 0.00000303, 0.00000224, 0.00000013, -0.00000213, -0.00000241, 0.00000001, 0.00000324, 0.00000447, 0.00000287, 0.00000031, -0.00000073, 0.00000020, 0.00000136, 0.00000099, -0.00000077, -0.00000247, },
		{ 0.00000056, -0.00000095, -0.00000201, -0.00000265, -0.00000341, -0.00000394, -0.00000298, -0.00000020, 0.00000239, 0.00000201, -0.00000164, -0.00000510, -0.00000441, 0.00000030, 0.00000415, 0.00000236, -0.00000408, -0.00000869, -0.00000619, 0.00000132, 0.00000593, 0.00000237, -0.00000582, -0.00000963, -0.00000486, 0.00000210, 0.00000011, -0.00001442, -0.00003181, -0.00003717, -0.00002490, -0.00000466, 0.00000778, 0.00000574, -0.00000253, -0.00000335, 0.00000773, 0.00002149, 0.00002509, 0.00001543, 0.00000186, -0.00000400, 0.00000018, 0.00000669, 0.00000756, 0.00000318, 0.00000059, 0.00000442, 0.00001162, 0.00001552, 0.00001359, 0.00000892, 0.00000417, -0.00000309, -0.00001613, -0.00003025, -0.00003188, -0.00001103, 0.00002420, 0.00004824, 0.00003841, -0.00000248, -0.00004438, -0.00005421, -0.00002511, 0.00001725, 0.00003809, 0.00002475, -0.00000410, -0.00001901, -0.00000889, 0.00001026, 0.00001433, -0.00000311, -0.00002443, -0.00002672, -0.00000528, 0.00002186, 0.00003189, 0.00001841, -0.00000495, -0.00001991, -0.00002044, -0.00001456, -0.00001183, -0.00001205, -0.00000730, 0.00000636, 0.00002188, 0.00002699, 0.00001648, -0.00000211, -0.00001534, -0.00001557, -0.00000605, 0.00000448, 0.00001056, 0.00001286, 0.00001404, 0.00001375, 0.00000914, -0.00000042, -0.00001067, -0.00001548, -0.00001242, -0.00000491, 0.00000112, 0.00000250, 0.00000056, -0.00000145, -0.00000176, -0.00000084, 0.00000028, 0.00000146, 0.00000288, 0.00000385, 0.00000317, 0.00000071, -0.00000178, -0.00000217, -0.00000020, 0.00000220, 0.00000291, 0.00000178, 0.00000050, 0.00000048, 0.00000132, 0.00000156, },
		{ 0.00000083, 0.00000146, 0.00000110, 0.00000003, -0.00000080, -0.00000115, -0.00000187, -0.00000332, -0.00000425, -0.00000289, 0.00000048, 0.00000286, 0.00000147, -0.00000275, -0.00000524, -0.00000255, 0.00000329, 0.00000609, 0.00000199, -0.00000569, -0.00000924, -0.00000473, 0.00000306, 0.00000523, -0.00000145, -0.00001020, -0.00001104, -0.00000210, 0.00000690, 0.00000397, -0.00001216, -0.00002914, -0.00003242, -0.00001959, -0.00000283, 0.00000354, -0.00000240, -0.00000886, -0.00000346, 0.00001299, 0.00002682, 0.00002539, 0.00000992, -0.00000597, -0.00000985, -0.00000169, 0.00000828, 0.00001127, 0.00000774, 0.00000410, 0.00000386, 0.00000470, 0.00000393, 0.00000392, 0.00000896, 0.00001671, 0.00001632, -0.00000100, -0.00002856, -0.00004460, -0.00003054, 0.00000868, 0.00004408, 0.00004608, 0.00001150, -0.00003199, -0.00004985, -0.00003093, 0.00000438, 0.00002525, 0.00001910, 0.00000026, -0.00000810, 0.00000143, 0.00001458, 0.00001256, -0.00000690, -0.00002687, -0.00002861, -0.00001021, 0.00001250, 0.00002221, 0.00001578, 0.00000399, -0.00000229, -0.00000317, -0.00000671, -0.00001680, -0.00002649, -0.00002446, -0.00000745, 0.00001492, 0.00002815, 0.00002522, 0.00001137, -0.00000241, -0.00000919, -0.00000950, -0.00000704, -0.00000314, 0.00000342, 0.00001223, 0.00001892, 0.00001819, 0.00000918, -0.00000301, -0.00001133, -0.00001242, -0.00000843, -0.00000385, -0.00000137, -0.00000071, -0.00000045, -0.00000008, 0.00000010, 0.00000014, 0.00000070, 0.00000197, 0.00000298, 0.00000255, 0.00000079, -0.00000073, -0.00000062, 0.00000080, 0.00000188, 0.00000160, 0.00000058, 0.00000021, },
		{ 0.00000130, 0.00000044, 0.00000051, 0.00000102, 0.00000083, -0.00000026, -0.00000120, -0.00000113, -0.00000068, -0.00000125, -0.00000284, -0.00000350, -0.00000153, 0.00000183, 0.00000295, -0.00000020, -0.00000492, -0.00000580, -0.00000062, 0.00000628, 0.00000769, 0.00000124, -0.00000722, -0.00000914, -0.00000248, 0.00000537, 0.00000507, -0.00000433, -0.00001344, -0.00001218, -0.00000119, 0.00000772, 0.00000351, -0.00001190, -0.00002461, -0.00002268, -0.00000874, 0.00000249, -0.00000036, -0.00001286, -0.00001890, -0.00000742, 0.00001540, 0.00003143, 0.00002783, 0.00000885, -0.00000928, -0.00001398, -0.00000622, 0.00000407, 0.00000963, 0.00001116, 0.00001196, 0.00001097, 0.00000431, -0.00000654, -0.00001163, -0.00000173, 0.00001882, 0.00003132, 0.00001899, -0.00001422, -0.00004285, -0.00004132, -0.00000854, 0.00002888, 0.00003972, 0.00001624, -0.00001958, -0.00003728, -0.00002569, -0.00000045, 0.00001418, 0.00000935, -0.00000260, -0.00000432, 0.00000709, 0.00001787, 0.00001352, -0.00000506, -0.00002242, -0.00002443, -0.00001179, 0.00000250, 0.00000760, 0.00000488, 0.00000361, 0.00000834, 0.00001284, 0.00000719, -0.00001000, -0.00002800, -0.00003301, -0.00002065, 0.00000058, 0.00001741, 0.00002261, 0.00001835, 0.00001081, 0.00000359, -0.00000337, -0.00000979, -0.00001246, -0.00000783, 0.00000333, 0.00001491, 0.00001990, 0.00001597, 0.00000681, -0.00000174, -0.00000633, -0.00000744, -0.00000705, -0.00000617, -0.00000460, -0.00000233, -0.00000020, 0.00000093, 0.00000120, 0.00000141, 0.00000183, 0.00000187, 0.00000104, -0.00000008, -0.00000031, 0.00000072, 0.00000198, 0.00000220, },
		{ 0.00000100, 0.00000213, 0.00000207, 0.00000143, 0.00000137, 0.00000196, 0.00000199, 0.00000054, -0.00000166, -0.00000284, -0.00000235, -0.00000148, -0.00000176, -0.00000264, -0.00000192, 0.00000128, 0.00000437, 0.00000336, -0.00000229, -0.00000771, -0.00000685, 0.00000086, 0.00000866, 0.00000867, 0.00000031, -0.00000834, -0.00000850, -0.00000021, 0.00000714, 0.00000465, -0.00000613, -0.00001423, -0.00001098, -0.00000001, 0.00000591, -0.00000135, -0.00001555, -0.00002165, -0.00001179, 0.00000518, 0.00001167, -0.00000051, -0.00002058, -0.00002839, -0.00001394, 0.00001264, 0.00003064, 0.00002785, 0.00001017, -0.00000653, -0.00001207, -0.00000855, -0.00000329, 0.00000137, 0.00000852, 0.00001817, 0.00002224, 0.00001197, -0.00000949, -0.00002494, -0.00001783, 0.00000964, 0.00003476, 0.00003372, 0.00000474, -0.00002850, -0.00003810, -0.00001772, 0.00001205, 0.00002394, 0.00000955, -0.00001465, -0.00002536, -0.00001522, 0.00000244, 0.00000971, 0.00000266, -0.00000663, -0.00000511, 0.00000665, 0.00001576, 0.00001166, -0.00000257, -0.00001410, -0.00001421, -0.00000672, -0.00000214, -0.00000505, -0.00000909, -0.00000484, 0.00000836, 0.00002003, 0.00001832, 0.00000187, -0.00001843, -0.00002907, -0.00002497, -0.00001167, 0.00000170, 0.00001029, 0.00001471, 0.00001649, 0.00001461, 0.00000727, -0.00000388, -0.00001292, -0.00001393, -0.00000619, 0.00000516, 0.00001349, 0.00001561, 0.00001285, 0.00000823, 0.00000354, -0.00000101, -0.00000522, -0.00000800, -0.00000824, -0.00000610, -0.00000300, -0.00000031, 0.00000150, 0.00000250, 0.00000261, 0.00000166, 0.00000008, -0.00000093, -0.00000048, },
		{ 0.00000032, -0.00000059, -0.00000016, 0.00000080, 0.00000110, 0.00000075, 0.00000089, 0.00000227, 0.00000388, 0.00000379, 0.00000137, -0.00000183, -0.00000366, -0.00000373, -0.00000353, -0.00000400, -0.00000375, -0.00000063, 0.00000472, 0.00000788, 0.00000465, -0.00000367, -0.00001000, -0.00000783, 0.00000163, 0.00000967, 0.00000839, -0.00000092, -0.00000867, -0.00000689, 0.00000206, 0.00000772, 0.00000295, -0.00000784, -0.00001306, -0.00000696, 0.00000315, 0.00000411, -0.00000807, -0.00002207, -0.00002178, -0.00000369, 0.00001734, 0.00002135, 0.00000303, -0.00002214, -0.00003197, -0.00001821, 0.00000671, 0.00002276, 0.00002093, 0.00000877, -0.00000081, -0.00000343, -0.00000446, -0.00000825, -0.00000990, -0.00000088, 0.00001772, 0.00003109, 0.00002351, -0.00000381, -0.00002972, -0.00003105, -0.00000505, 0.00002652, 0.00003683, 0.00001869, -0.00000979, -0.00002368, -0.00001491, 0.00000224, 0.00000775, -0.00000349, -0.00001814, -0.00002000, -0.00000709, 0.00000753, 0.00001048, 0.00000182, -0.00000665, -0.00000520, 0.00000361, 0.00000902, 0.00000483, -0.00000393, -0.00000738, -0.00000256, 0.00000314, 0.00000054, -0.00001027, -0.00001892, -0.00001511, 0.00000051, 0.00001629, 0.00002017, 0.00001011, -0.00000540, -0.00001608, -0.00001817, -0.00001475, -0.00000994, -0.00000436, 0.00000333, 0.00001214, 0.00001743, 0.00001487, 0.00000511, -0.00000598, -0.00001189, -0.00001036, -0.00000397, 0.00000312, 0.00000850, 0.00001187, 0.00001330, 0.00001213, 0.00000784, 0.00000145, -0.00000463, -0.00000821, -0.00000864, -0.00000673, -0.00000365, -0.00000042, 0.00000203, 0.00000288, 0.00000197, },
		{ -0.00000075, 0.00000103, 0.00000122, 0.00000080, 0.00000086, 0.00000136, 0.00000124, 0.00000000, -0.00000120, -0.00000060, 0.00000207, 0.00000483, 0.00000525, 0.00000290, -0.00000033, -0.00000254, -0.00000386, -0.00000556, -0.00000730, -0.00000636, -0.00000071, 0.00000719, 0.00001101, 0.00000639, -0.00000370, -0.00001047, -0.00000754, 0.00000224, 0.00000911, 0.00000622, -0.00000304, -0.00000849, -0.00000419, 0.00000474, 0.00000752, -0.00000008, -0.00001030, -0.00001147, -0.00000159, 0.00000796, 0.00000394, -0.00001337, -0.00002798, -0.00002325, 0.00000036, 0.00002374, 0.00002644, 0.00000634, -0.00001841, -0.00002697, -0.00001508, 0.00000304, 0.00001100, 0.00000641, 0.00000029, 0.00000202, 0.00000778, 0.00000607, -0.00000651, -0.00001807, -0.00001285, 0.00001021, 0.00003253, 0.00003220, 0.00000603, -0.00002528, -0.00003554, -0.00001714, 0.00001266, 0.00002898, 0.00002222, 0.00000433, -0.00000623, -0.00000353, 0.00000220, -0.00000159, -0.00001483, -0.00002497, -0.00002094, -0.00000521, 0.00000900, 0.00001145, 0.00000433, -0.00000205, -0.00000137, 0.00000256, 0.00000199, -0.00000450, -0.00000961, -0.00000583, 0.00000503, 0.00001237, 0.00000738, -0.00000750, -0.00002021, -0.00002031, -0.00000829, 0.00000581, 0.00001225, 0.00000941, 0.00000280, -0.00000213, -0.00000490, -0.00000802, -0.00001193, -0.00001321, -0.00000832, 0.00000174, 0.00001128, 0.00001454, 0.00001038, 0.00000251, -0.00000426, -0.00000753, -0.00000754, -0.00000523, -0.00000085, 0.00000525, 0.00001144, 0.00001506, 0.00001425, 0.00000939, 0.00000272, -0.00000329, -0.00000710, -0.00000820, -0.00000682, -0.00000384, },
		{ -0.00000528, -0.00000686, -0.00000617, -0.00000444, -0.00000290, -0.00000169, -0.00000011, 0.00000195, 0.00000334, 0.00000261, -0.00000004, -0.00000235, -0.00000203, 0.00000092, 0.00000403, 0.00000495, 0.00000363, 0.00000180, 0.00000033, -0.00000191, -0.00000595, -0.00000983, -0.00000923, -0.00000211, 0.00000753, 0.00001196, 0.00000700, -0.00000302, -0.00000870, -0.00000485, 0.00000394, 0.00000785, 0.00000238, -0.00000651, -0.00000862, -0.00000110, 0.00000778, 0.00000728, -0.00000341, -0.00001309, -0.00001035, 0.00000303, 0.00001225, 0.00000461, -0.00001577, -0.00002991, -0.00002229, 0.00000274, 0.00002399, 0.00002365, 0.00000436, -0.00001420, -0.00001568, -0.00000297, 0.00000675, 0.00000175, -0.00001127, -0.00001519, -0.00000246, 0.00001512, 0.00001828, 0.00000161, -0.00001893, -0.00002122, -0.00000001, 0.00002646, 0.00003344, 0.00001357, -0.00001621, -0.00003116, -0.00002148, 0.00000085, 0.00001628, 0.00001658, 0.00000944, 0.00000635, 0.00000887, 0.00000824, -0.00000260, -0.00001909, -0.00002860, -0.00002328, -0.00000799, 0.00000510, 0.00000857, 0.00000553, 0.00000363, 0.00000522, 0.00000522, -0.00000139, -0.00001165, -0.00001610, -0.00000860, 0.00000618, 0.00001610, 0.00001271, -0.00000101, -0.00001397, -0.00001739, -0.00001174, -0.00000411, -0.00000011, 0.00000075, 0.00000221, 0.00000536, 0.00000687, 0.00000306, -0.00000522, -0.00001246, -0.00001337, -0.00000745, 0.00000099, 0.00000699, 0.00000853, 0.00000667, 0.00000323, -0.00000087, -0.00000504, -0.00000780, -0.00000719, -0.00000243, 0.00000487, 0.00001154, 0.00001490, 0.00001413, 0.00001010, 0.00000442, -0.00000122, },
		{ 0.00001360, 0.00001092, 0.00000628, 0.00000132, -0.00000254, -0.00000490, -0.00000626, -0.00000682, -0.00000595, -0.00000302, 0.00000109, 0.00000399, 0.00000371, 0.00000085, -0.00000182, -0.00000193, 0.00000018, 0.00000220, 0.00000264, 0.00000234, 0.00000275, 0.00000326, 0.00000130, -0.00000417, -0.00000985, -0.00001018, -0.00000320, 0.00000615, 0.00000986, 0.00000484, -0.00000334, -0.00000581, 0.00000012, 0.00000761, 0.00000740, -0.00000175, -0.00001076, -0.00000966, 0.00000110, 0.00001013, 0.00000699, -0.00000589, -0.00001476, -0.00000911, 0.00000595, 0.00001380, 0.00000368, -0.00001658, -0.00002710, -0.00001602, 0.00000747, 0.00002160, 0.00001405, -0.00000534, -0.00001491, -0.00000424, 0.00001442, 0.00001874, 0.00000039, -0.00002493, -0.00003250, -0.00001380, 0.00001425, 0.00002569, 0.00001120, -0.00001312, -0.00002226, -0.00000716, 0.00001722, 0.00002752, 0.00001478, -0.00000856, -0.00002243, -0.00001832, -0.00000449, 0.00000564, 0.00000751, 0.00000717, 0.00001084, 0.00001595, 0.00001388, 0.00000080, -0.00001635, -0.00002574, -0.00002197, -0.00001057, -0.00000125, 0.00000193, 0.00000281, 0.00000618, 0.00001052, 0.00000930, -0.00000068, -0.00001364, -0.00001904, -0.00001163, 0.00000299, 0.00001338, 0.00001260, 0.00000354, -0.00000534, -0.00000866, -0.00000802, -0.00000797, -0.00000980, -0.00001001, -0.00000498, 0.00000401, 0.00001108, 0.00001107, 0.00000417, -0.00000457, -0.00000984, -0.00000995, -0.00000666, -0.00000245, 0.00000142, 0.00000450, 0.00000588, 0.00000437, -0.00000001, -0.00000516, -0.00000805, -0.00000681, -0.00000193, 0.00000446, 0.00001006, 0.00001334, },
		{ -0.00000199, 0.00000299, 0.00000813, 0.00001186, 0.00001304, 0.00001159, 0.00000845, 0.00000458, 0.00000030, -0.00000423, -0.00000795, -0.00000903, -0.00000650, -0.00000179, 0.00000201, 0.00000273, 0.00000110, -0.00000030, 0.00000007, 0.00000116, 0.00000105, -0.00000028, -0.00000062, 0.00000152, 0.00000421, 0.00000361, -0.00000152, -0.00000721, -0.00000771, -0.00000177, 0.00000506, 0.00000576, -0.00000038, -0.00000616, -0.00000392, 0.00000557, 0.00001266, 0.00000864, -0.00000427, -0.00001403, -0.00001095, 0.00000167, 0.00001033, 0.00000555, -0.00000753, -0.00001417, -0.00000619, 0.00000814, 0.00001219, -0.00000076, -0.00001861, -0.00002193, -0.00000485, 0.00001700, 0.00002132, 0.00000261, -0.00002040, -0.00002264, 0.00000134, 0.00002995, 0.00003452, 0.00000749, -0.00002920, -0.00004436, -0.00002660, 0.00000604, 0.00002455, 0.00001624, -0.00000550, -0.00001704, -0.00000816, 0.00001003, 0.00001865, 0.00001026, -0.00000571, -0.00001475, -0.00001185, -0.00000386, 0.00000057, 0.00000074, 0.00000257, 0.00000930, 0.00001600, 0.00001458, 0.00000315, -0.00001091, -0.00001804, -0.00001557, -0.00000913, -0.00000537, -0.00000476, -0.00000245, 0.00000426, 0.00001115, 0.00001080, 0.00000109, -0.00001136, -0.00001671, -0.00001116, -0.00000014, 0.00000738, 0.00000733, 0.00000309, 0.00000031, 0.00000039, -0.00000029, -0.00000514, -0.00001249, -0.00001648, -0.00001275, -0.00000285, 0.00000708, 0.00001154, 0.00000959, 0.00000416, -0.00000135, -0.00000534, -0.00000764, -0.00000793, -0.00000566, -0.00000129, 0.00000309, 0.00000489, 0.00000309, -0.00000102, -0.00000491, -0.00000661, -0.00000549, },
		{ -0.00000137, -0.00000385, -0.00000547, -0.00000525, -0.00000266, 0.00000181, 0.00000667, 0.00001042, 0.00001240, 0.00001265, 0.00001106, 0.00000714, 0.00000117, -0.00000495, -0.00000842, -0.00000787, -0.00000475, -0.00000195, -0.00000106, -0.00000098, 0.00000025, 0.00000258, 0.00000368, 0.00000173, -0.00000183, -0.00000336, -0.00000108, 0.00000251, 0.00000307, -0.00000037, -0.00000377, -0.00000239, 0.00000313, 0.00000643, 0.00000222, -0.00000680, -0.00001122, -0.00000450, 0.00000897, 0.00001680, 0.00001071, -0.00000432, -0.00001437, -0.00001062, 0.00000144, 0.00000789, 0.00000161, -0.00000963, -0.00001176, -0.00000076, 0.00001142, 0.00000921, -0.00000831, -0.00002368, -0.00001807, 0.00000755, 0.00003005, 0.00002569, -0.00000503, -0.00003488, -0.00003447, -0.00000098, 0.00003742, 0.00004674, 0.00001829, -0.00002397, -0.00004599, -0.00003401, -0.00000403, 0.00001632, 0.00001381, -0.00000147, -0.00001029, -0.00000445, 0.00000733, 0.00001135, 0.00000378, -0.00000696, -0.00001079, -0.00000645, -0.00000085, 0.00000025, -0.00000141, 0.00000001, 0.00000615, 0.00001166, 0.00001012, 0.00000150, -0.00000737, -0.00001000, -0.00000678, -0.00000362, -0.00000463, -0.00000743, -0.00000635, 0.00000023, 0.00000713, 0.00000751, 0.00000027, -0.00000841, -0.00001115, -0.00000657, -0.00000002, 0.00000255, 0.00000062, -0.00000121, 0.00000086, 0.00000516, 0.00000621, 0.00000065, -0.00000886, -0.00001588, -0.00001577, -0.00000926, -0.00000083, 0.00000538, 0.00000810, 0.00000806, 0.00000597, 0.00000206, -0.00000293, -0.00000701, -0.00000810, -0.00000579, -0.00000181, 0.00000136, 0.00000226, 0.00000102, },
		{ -0.00000258, -0.00000085, 0.00000015, -0.00000006, -0.00000169, -0.00000409, -0.00000576, -0.00000535, -0.00000272, 0.00000117, 0.00000531, 0.00000923, 0.00001251, 0.00001398, 0.00001224, 0.00000719, 0.00000094, -0.00000360, -0.00000516, -0.00000508, -0.00000540, -0.00000622, -0.00000546, -0.00000169, 0.00000336, 0.00000582, 0.00000359, -0.00000112, -0.00000390, -0.00000284, -0.00000047, -0.00000036, -0.00000231, -0.00000205, 0.00000326, 0.00001011, 0.00001075, 0.00000152, -0.00001133, -0.00001608, -0.00000695, 0.00000894, 0.00001746, 0.00001118, -0.00000308, -0.00001130, -0.00000695, 0.00000244, 0.00000410, -0.00000491, -0.00001364, -0.00000946, 0.00000614, 0.00001661, 0.00000740, -0.00001622, -0.00003091, -0.00001768, 0.00001664, 0.00004221, 0.00003319, -0.00000618, -0.00004286, -0.00004422, -0.00000801, 0.00003524, 0.00004972, 0.00002569, -0.00001433, -0.00003812, -0.00003216, -0.00000962, 0.00000655, 0.00000593, -0.00000318, -0.00000666, -0.00000032, 0.00000754, 0.00000731, -0.00000115, -0.00000924, -0.00000969, -0.00000385, 0.00000130, 0.00000163, -0.00000023, 0.00000062, 0.00000458, 0.00000690, 0.00000358, -0.00000331, -0.00000750, -0.00000533, 0.00000018, 0.00000277, -0.00000014, -0.00000483, -0.00000580, -0.00000209, 0.00000167, 0.00000073, -0.00000432, -0.00000805, -0.00000624, -0.00000059, 0.00000314, 0.00000135, -0.00000352, -0.00000558, -0.00000183, 0.00000473, 0.00000801, 0.00000467, -0.00000310, -0.00001009, -0.00001268, -0.00001092, -0.00000700, -0.00000262, 0.00000176, 0.00000568, 0.00000780, 0.00000678, 0.00000286, -0.00000190, -0.00000511, -0.00000577, -0.00000452, },
		{ 0.00000020, -0.00000184, -0.00000315, -0.00000353, -0.00000279, -0.00000145, -0.00000074, -0.00000161, -0.00000375, -0.00000568, -0.00000611, -0.00000487, -0.00000256, 0.00000053, 0.00000457, 0.00000920, 0.00001278, 0.00001340, 0.00001075, 0.00000667, 0.00000339, 0.00000124, -0.00000145, -0.00000572, -0.00000980, -0.00001025, -0.00000564, 0.00000119, 0.00000531, 0.00000437, 0.00000064, -0.00000188, -0.00000204, -0.00000230, -0.00000465, -0.00000665, -0.00000329, 0.00000649, 0.00001607, 0.00001598, 0.00000373, -0.00001183, -0.00001753, -0.00000849, 0.00000659, 0.00001374, 0.00000767, -0.00000326, -0.00000664, 0.00000000, 0.00000637, 0.00000153, -0.00001213, -0.00001955, -0.00000919, 0.00001193, 0.00002243, 0.00000797, -0.00002129, -0.00003710, -0.00001988, 0.00002001, 0.00004871, 0.00003926, -0.00000254, -0.00004203, -0.00004643, -0.00001369, 0.00002708, 0.00004279, 0.00002481, -0.00000712, -0.00002619, -0.00002252, -0.00000771, 0.00000094, -0.00000210, -0.00000830, -0.00000745, 0.00000063, 0.00000710, 0.00000485, -0.00000377, -0.00001019, -0.00000906, -0.00000295, 0.00000181, 0.00000245, 0.00000167, 0.00000279, 0.00000491, 0.00000396, -0.00000170, -0.00000824, -0.00000980, -0.00000464, 0.00000279, 0.00000646, 0.00000463, 0.00000082, -0.00000088, -0.00000054, -0.00000158, -0.00000588, -0.00001057, -0.00001064, -0.00000458, 0.00000324, 0.00000647, 0.00000283, -0.00000371, -0.00000694, -0.00000420, 0.00000162, 0.00000541, 0.00000461, 0.00000062, -0.00000355, -0.00000623, -0.00000765, -0.00000827, -0.00000752, -0.00000462, -0.00000007, 0.00000414, 0.00000603, 0.00000520, 0.00000279, },
		{ 0.00000118, 0.00000279, 0.00000359, 0.00000354, 0.00000243, 0.00000057, -0.00000097, -0.00000133, -0.00000088, -0.00000096, -0.00000244, -0.00000479, -0.00000654, -0.00000685, -0.00000603, -0.00000469, -0.00000268, 0.00000067, 0.00000515, 0.00000914, 0.00001091, 0.00001049, 0.00000957, 0.00000933, 0.00000861, 0.00000494, -0.00000216, -0.00000946, -0.00001230, -0.00000906, -0.00000283, 0.00000151, 0.00000206, 0.00000094, 0.00000080, 0.00000098, -0.00000154, -0.00000712, -0.00001074, -0.00000618, 0.00000645, 0.00001828, 0.00001878, 0.00000638, -0.00000879, -0.00001392, -0.00000607, 0.00000509, 0.00000755, -0.00000005, -0.00000716, -0.00000362, 0.00000776, 0.00001292, 0.00000216, -0.00001705, -0.00002528, -0.00001118, 0.00001431, 0.00002615, 0.00000978, -0.00002205, -0.00003894, -0.00002147, 0.00001865, 0.00004745, 0.00003948, 0.00000126, -0.00003473, -0.00003963, -0.00001308, 0.00001910, 0.00003023, 0.00001579, -0.00000654, -0.00001669, -0.00001060, -0.00000023, 0.00000164, -0.00000559, -0.00001215, -0.00001012, -0.00000184, 0.00000370, 0.00000119, -0.00000591, -0.00001001, -0.00000761, -0.00000192, 0.00000192, 0.00000261, 0.00000281, 0.00000444, 0.00000554, 0.00000261, -0.00000441, -0.00001086, -0.00001154, -0.00000589, 0.00000164, 0.00000617, 0.00000675, 0.00000591, 0.00000543, 0.00000385, -0.00000128, -0.00000919, -0.00001495, -0.00001372, -0.00000571, 0.00000335, 0.00000715, 0.00000414, -0.00000167, -0.00000502, -0.00000402, -0.00000084, 0.00000139, 0.00000162, 0.00000091, 0.00000030, -0.00000057, -0.00000258, -0.00000545, -0.00000750, -0.00000722, -0.00000472, -0.00000146, },
		{ -0.00000556, -0.00000562, -0.00000505, -0.00000396, -0.00000203, 0.00000055, 0.00000272, 0.00000351, 0.00000307, 0.00000256, 0.00000267, 0.00000274, 0.00000140, -0.00000172, -0.00000532, -0.00000768, -0.00000820, -0.00000762, -0.00000666, -0.00000503, -0.00000203, 0.00000194, 0.00000528, 0.00000684, 0.00000746, 0.00000908, 0.00001205, 0.00001380, 0.00001102, 0.00000340, -0.00000517, -0.00000979, -0.00000902, -0.00000577, -0.00000371, -0.00000325, -0.00000191, 0.00000142, 0.00000371, 0.00000082, -0.00000668, -0.00001190, -0.00000774, 0.00000497, 0.00001650, 0.00001686, 0.00000608, -0.00000529, -0.00000668, 0.00000130, 0.00000750, 0.00000269, -0.00000929, -0.00001461, -0.00000468, 0.00001269, 0.00001937, 0.00000569, -0.00001753, -0.00002785, -0.00001354, 0.00001293, 0.00002552, 0.00001026, -0.00001967, -0.00003522, -0.00001932, 0.00001593, 0.00004003, 0.00003235, 0.00000096, -0.00002599, -0.00002690, -0.00000537, 0.00001594, 0.00001825, 0.00000319, -0.00001201, -0.00001316, -0.00000212, 0.00000769, 0.00000634, -0.00000349, -0.00001132, -0.00001069, -0.00000490, -0.00000182, -0.00000477, -0.00000972, -0.00001065, -0.00000621, -0.00000033, 0.00000284, 0.00000322, 0.00000353, 0.00000493, 0.00000514, 0.00000144, -0.00000538, -0.00001090, -0.00001122, -0.00000662, -0.00000074, 0.00000339, 0.00000585, 0.00000811, 0.00000973, 0.00000794, 0.00000093, -0.00000879, -0.00001534, -0.00001435, -0.00000697, 0.00000114, 0.00000478, 0.00000324, -0.00000019, -0.00000211, -0.00000190, -0.00000120, -0.00000120, -0.00000137, -0.00000059, 0.00000099, 0.00000188, 0.00000082, -0.00000177, -0.00000429, },
		{ 0.00000130, 0.00000012, -0.00000102, -0.00000201, -0.00000339, -0.00000514, -0.00000624, -0.00000562, -0.00000343, -0.00000100, 0.00000067, 0.00000192, 0.00000373, 0.00000608, 0.00000738, 0.00000592, 0.00000174, -0.00000322, -0.00000690, -0.00000869, -0.00000931, -0.00000919, -0.00000779, -0.00000466, -0.00000075, 0.00000204, 0.00000303, 0.00000385, 0.00000678, 0.00001169, 0.00001529, 0.00001401, 0.00000781, 0.00000059, -0.00000362, -0.00000447, -0.00000493, -0.00000704, -0.00000895, -0.00000712, -0.00000137, 0.00000338, 0.00000192, -0.00000489, -0.00000962, -0.00000563, 0.00000520, 0.00001325, 0.00001089, 0.00000111, -0.00000488, 0.00000034, 0.00001131, 0.00001435, 0.00000237, -0.00001580, -0.00002258, -0.00000953, 0.00001237, 0.00002215, 0.00000951, -0.00001394, -0.00002540, -0.00001353, 0.00000966, 0.00002046, 0.00000739, -0.00001680, -0.00002745, -0.00001256, 0.00001507, 0.00003030, 0.00002003, -0.00000465, -0.00002016, -0.00001354, 0.00000612, 0.00001814, 0.00001106, -0.00000731, -0.00001918, -0.00001455, 0.00000045, 0.00001149, 0.00001042, 0.00000136, -0.00000601, -0.00000700, -0.00000519, -0.00000638, -0.00001137, -0.00001530, -0.00001353, -0.00000666, 0.00000051, 0.00000420, 0.00000476, 0.00000474, 0.00000502, 0.00000388, -0.00000023, -0.00000592, -0.00000969, -0.00000935, -0.00000597, -0.00000220, 0.00000075, 0.00000384, 0.00000774, 0.00001058, 0.00000907, 0.00000214, -0.00000687, -0.00001245, -0.00001154, -0.00000600, -0.00000059, 0.00000146, 0.00000059, -0.00000060, -0.00000061, -0.00000013, -0.00000051, -0.00000172, -0.00000236, -0.00000142, 0.00000042, 0.00000159, },
		{ -0.00000059, 0.00000050, 0.00000064, 0.00000011, -0.00000005, 0.00000045, 0.00000062, -0.00000067, -0.00000318, -0.00000534, -0.00000593, -0.00000522, -0.00000440, -0.00000378, -0.00000225, 0.00000130, 0.00000612, 0.00000969, 0.00000976, 0.00000639, 0.00000159, -0.00000274, -0.00000615, -0.00000902, -0.00001099, -0.00001083, -0.00000808, -0.00000419, -0.00000143, -0.00000035, 0.00000096, 0.00000465, 0.00001012, 0.00001395, 0.00001324, 0.00000886, 0.00000465, 0.00000316, 0.00000281, -0.00000001, -0.00000613, -0.00001158, -0.00001155, -0.00000609, -0.00000074, -0.00000076, -0.00000487, -0.00000621, -0.00000020, 0.00000903, 0.00001199, 0.00000425, -0.00000698, -0.00000889, 0.00000335, 0.00001953, 0.00002285, 0.00000685, -0.00001646, -0.00002680, -0.00001504, 0.00000737, 0.00001925, 0.00001033, -0.00000897, -0.00001864, -0.00000964, 0.00000720, 0.00001307, 0.00000127, -0.00001556, -0.00001863, -0.00000293, 0.00001705, 0.00002179, 0.00000661, -0.00001344, -0.00001855, -0.00000377, 0.00001645, 0.00002263, 0.00000889, -0.00001241, -0.00002335, -0.00001686, -0.00000119, 0.00000999, 0.00001049, 0.00000478, 0.00000046, -0.00000031, -0.00000161, -0.00000728, -0.00001554, -0.00002041, -0.00001785, -0.00000959, -0.00000091, 0.00000434, 0.00000613, 0.00000629, 0.00000547, 0.00000286, -0.00000172, -0.00000635, -0.00000848, -0.00000741, -0.00000474, -0.00000232, -0.00000015, 0.00000292, 0.00000680, 0.00000903, 0.00000691, 0.00000063, -0.00000608, -0.00000897, -0.00000695, -0.00000276, -0.00000009, -0.00000020, -0.00000148, -0.00000185, -0.00000102, -0.00000034, -0.00000074, -0.00000161, -0.00000164, },
		{ -0.00000255, -0.00000239, -0.00000096, 0.00000080, 0.00000135, 0.00000037, -0.00000079, -0.00000065, 0.00000064, 0.00000143, 0.00000047, -0.00000166, -0.00000333, -0.00000380, -0.00000399, -0.00000509, -0.00000658, -0.00000637, -0.00000293, 0.00000275, 0.00000776, 0.00000980, 0.00000888, 0.00000650, 0.00000364, -0.00000006, -0.00000490, -0.00000964, -0.00001211, -0.00001132, -0.00000864, -0.00000623, -0.00000448, -0.00000174, 0.00000326, 0.00000893, 0.00001182, 0.00001029, 0.00000676, 0.00000542, 0.00000746, 0.00000936, 0.00000649, -0.00000142, -0.00000932, -0.00001190, -0.00000922, -0.00000636, -0.00000709, -0.00000876, -0.00000509, 0.00000543, 0.00001534, 0.00001432, 0.00000068, -0.00001422, -0.00001526, 0.00000129, 0.00002223, 0.00002846, 0.00001308, -0.00001144, -0.00002465, -0.00001720, 0.00000101, 0.00001173, 0.00000658, -0.00000577, -0.00001043, -0.00000278, 0.00000705, 0.00000613, -0.00000604, -0.00001611, -0.00001112, 0.00000663, 0.00002049, 0.00001585, -0.00000443, -0.00002176, -0.00001942, 0.00000126, 0.00002217, 0.00002529, 0.00000885, -0.00001259, -0.00002245, -0.00001634, -0.00000318, 0.00000561, 0.00000650, 0.00000433, 0.00000435, 0.00000588, 0.00000385, -0.00000442, -0.00001540, -0.00002224, -0.00002098, -0.00001339, -0.00000448, 0.00000205, 0.00000551, 0.00000668, 0.00000579, 0.00000261, -0.00000201, -0.00000593, -0.00000732, -0.00000623, -0.00000413, -0.00000202, 0.00000049, 0.00000387, 0.00000696, 0.00000719, 0.00000325, -0.00000296, -0.00000739, -0.00000723, -0.00000338, 0.00000061, 0.00000193, 0.00000066, -0.00000109, -0.00000179, -0.00000176, -0.00000200, },
		{ 0.00000055, -0.00000092, -0.00000304, -0.00000451, -0.00000389, -0.00000136, 0.00000105, 0.00000149, 0.00000022, -0.00000081, -0.00000023, 0.00000121, 0.00000158, 0.00000014, -0.00000177, -0.00000238, -0.00000169, -0.00000153, -0.00000335, -0.00000624, -0.00000755, -0.00000559, -0.00000130, 0.00000284, 0.00000533, 0.00000669, 0.00000796, 0.00000874, 0.00000724, 0.00000256, -0.00000384, -0.00000915, -0.00001177, -0.00001228, -0.00001185, -0.00001012, -0.00000566, 0.00000141, 0.00000794, 0.00001008, 0.00000730, 0.00000363, 0.00000384, 0.00000837, 0.00001248, 0.00001105, 0.00000403, -0.00000354, -0.00000718, -0.00000774, -0.00000958, -0.00001394, -0.00001563, -0.00000812, 0.00000755, 0.00002042, 0.00001857, 0.00000151, -0.00001688, -0.00001982, -0.00000339, 0.00001886, 0.00002791, 0.00001664, -0.00000416, -0.00001692, -0.00001390, -0.00000284, 0.00000311, -0.00000038, -0.00000575, -0.00000387, 0.00000434, 0.00000853, 0.00000122, -0.00001204, -0.00001693, -0.00000549, 0.00001392, 0.00002308, 0.00001164, -0.00001182, -0.00002684, -0.00001989, 0.00000318, 0.00002284, 0.00002376, 0.00000752, -0.00001088, -0.00001779, -0.00001196, -0.00000249, 0.00000204, 0.00000127, 0.00000080, 0.00000419, 0.00000850, 0.00000774, -0.00000038, -0.00001181, -0.00001976, -0.00002044, -0.00001520, -0.00000804, -0.00000198, 0.00000213, 0.00000433, 0.00000442, 0.00000231, -0.00000108, -0.00000407, -0.00000549, -0.00000541, -0.00000432, -0.00000212, 0.00000153, 0.00000583, 0.00000824, 0.00000629, 0.00000022, -0.00000648, -0.00000955, -0.00000745, -0.00000244, 0.00000176, 0.00000319, 0.00000256, 0.00000151, },
		{ 0.00000200, 0.00000313, 0.00000405, 0.00000387, 0.00000162, -0.00000195, -0.00000446, -0.00000419, -0.00000193, -0.00000014, -0.00000029, -0.00000134, -0.00000114, 0.00000087, 0.00000285, 0.00000257, 0.00000001, -0.00000249, -0.00000278, -0.00000126, -0.00000033, -0.00000157, -0.00000405, -0.00000541, -0.00000455, -0.00000269, -0.00000159, -0.00000118, 0.00000038, 0.00000428, 0.00000910, 0.00001168, 0.00000998, 0.00000478, -0.00000149, -0.00000721, -0.00001209, -0.00001562, -0.00001587, -0.00001104, -0.00000237, 0.00000546, 0.00000788, 0.00000489, 0.00000135, 0.00000222, 0.00000745, 0.00001209, 0.00001164, 0.00000679, 0.00000215, 0.00000028, -0.00000154, -0.00000767, -0.00001707, -0.00002168, -0.00001373, 0.00000467, 0.00002060, 0.00002083, 0.00000465, -0.00001416, -0.00001924, -0.00000652, 0.00001236, 0.00002136, 0.00001477, 0.00000102, -0.00000740, -0.00000640, -0.00000238, -0.00000289, -0.00000718, -0.00000769, -0.00000024, 0.00000912, 0.00000956, -0.00000201, -0.00001547, -0.00001632, -0.00000081, 0.00001845, 0.00002332, 0.00000757, -0.00001630, -0.00002804, -0.00001804, 0.00000442, 0.00002038, 0.00001839, 0.00000335, -0.00001026, -0.00001249, -0.00000555, 0.00000115, 0.00000155, -0.00000194, -0.00000287, 0.00000144, 0.00000705, 0.00000778, 0.00000160, -0.00000778, -0.00001460, -0.00001595, -0.00001305, -0.00000886, -0.00000532, -0.00000273, -0.00000083, 0.00000029, 0.00000043, -0.00000027, -0.00000144, -0.00000283, -0.00000414, -0.00000441, -0.00000230, 0.00000240, 0.00000752, 0.00000943, 0.00000594, -0.00000142, -0.00000819, -0.00001051, -0.00000802, -0.00000348, 0.00000012, },
		{ -0.00000716, -0.00000445, -0.00000272, -0.00000121, 0.00000129, 0.00000423, 0.00000551, 0.00000385, 0.00000057, -0.00000162, -0.00000150, -0.00000055, -0.00000107, -0.00000329, -0.00000481, -0.00000325, 0.00000089, 0.00000436, 0.00000435, 0.00000125, -0.00000195, -0.00000287, -0.00000192, -0.00000124, -0.00000189, -0.00000273, -0.00000218, -0.00000061, -0.00000024, -0.00000247, -0.00000573, -0.00000652, -0.00000277, 0.00000406, 0.00001028, 0.00001305, 0.00001200, 0.00000807, 0.00000180, -0.00000636, -0.00001439, -0.00001846, -0.00001574, -0.00000769, 0.00000030, 0.00000338, 0.00000168, -0.00000014, 0.00000198, 0.00000706, 0.00001040, 0.00000911, 0.00000557, 0.00000442, 0.00000637, 0.00000626, -0.00000124, -0.00001415, -0.00002267, -0.00001779, -0.00000107, 0.00001504, 0.00001781, 0.00000607, -0.00000876, -0.00001347, -0.00000522, 0.00000699, 0.00001222, 0.00000815, 0.00000164, -0.00000031, 0.00000160, 0.00000108, -0.00000479, -0.00001073, -0.00000880, 0.00000153, 0.00001077, 0.00000844, -0.00000502, -0.00001682, -0.00001374, 0.00000392, 0.00002103, 0.00002125, 0.00000280, -0.00001920, -0.00002642, -0.00001384, 0.00000645, 0.00001715, 0.00001134, -0.00000288, -0.00001181, -0.00000910, -0.00000002, 0.00000570, 0.00000378, -0.00000170, -0.00000400, -0.00000093, 0.00000361, 0.00000434, -0.00000005, -0.00000622, -0.00000990, -0.00000969, -0.00000741, -0.00000570, -0.00000564, -0.00000648, -0.00000683, -0.00000584, -0.00000363, -0.00000116, 0.00000018, -0.00000049, -0.00000251, -0.00000359, -0.00000151, 0.00000362, 0.00000863, 0.00000953, 0.00000497, -0.00000240, -0.00000803, -0.00000929, },
		{ -0.00000376, -0.00000721, -0.00000664, -0.00000460, -0.00000383, -0.00000441, -0.00000419, -0.00000174, 0.00000173, 0.00000366, 0.00000316, 0.00000197, 0.00000229, 0.00000385, 0.00000394, 0.00000048, -0.00000505, -0.00000849, -0.00000687, -0.00000138, 0.00000371, 0.00000507, 0.00000304, 0.00000039, -0.00000097, -0.00000158, -0.00000276, -0.00000419, -0.00000385, -0.00000072, 0.00000318, 0.00000414, 0.00000037, -0.00000585, -0.00000994, -0.00000892, -0.00000344, 0.00000357, 0.00000962, 0.00001361, 0.00001468, 0.00001140, 0.00000328, -0.00000717, -0.00001491, -0.00001612, -0.00001165, -0.00000631, -0.00000406, -0.00000423, -0.00000286, 0.00000217, 0.00000802, 0.00000979, 0.00000628, 0.00000206, 0.00000268, 0.00000800, 0.00001130, 0.00000600, -0.00000654, -0.00001693, -0.00001632, -0.00000523, 0.00000659, 0.00000946, 0.00000288, -0.00000485, -0.00000572, 0.00000005, 0.00000534, 0.00000469, 0.00000026, -0.00000140, 0.00000230, 0.00000643, 0.00000432, -0.00000393, -0.00001021, -0.00000699, 0.00000360, 0.00001054, 0.00000516, -0.00000877, -0.00001732, -0.00001001, 0.00000873, 0.00002244, 0.00001781, -0.00000234, -0.00002123, -0.00002339, -0.00000851, 0.00000913, 0.00001442, 0.00000478, -0.00000914, -0.00001459, -0.00000819, 0.00000276, 0.00000867, 0.00000652, 0.00000093, -0.00000200, -0.00000081, 0.00000106, -0.00000013, -0.00000430, -0.00000798, -0.00000804, -0.00000465, -0.00000085, 0.00000018, -0.00000249, -0.00000717, -0.00001093, -0.00001150, -0.00000857, -0.00000402, -0.00000070, -0.00000030, -0.00000181, -0.00000223, 0.00000071, 0.00000602, 0.00000984, 0.00000872, 0.00000290, },
		{ 0.00001182, 0.00000804, 0.00000052, -0.00000556, -0.00000677, -0.00000416, -0.00000166, -0.00000186, -0.00000381, -0.00000472, -0.00000343, -0.00000152, -0.00000097, -0.00000133, -0.00000004, 0.00000423, 0.00000911, 0.00001011, 0.00000500, -0.00000342, -0.00000948, -0.00000955, -0.00000489, 0.00000024, 0.00000279, 0.00000306, 0.00000286, 0.00000265, 0.00000107, -0.00000254, -0.00000615, -0.00000633, -0.00000191, 0.00000413, 0.00000696, 0.00000412, -0.00000241, -0.00000827, -0.00001042, -0.00000871, -0.00000445, 0.00000148, 0.00000846, 0.00001455, 0.00001646, 0.00001213, 0.00000329, -0.00000528, -0.00000974, -0.00001045, -0.00001070, -0.00001225, -0.00001280, -0.00000864, 0.00000025, 0.00000863, 0.00001052, 0.00000544, -0.00000056, -0.00000074, 0.00000529, 0.00001103, 0.00000974, 0.00000156, -0.00000666, -0.00000844, -0.00000418, -0.00000005, -0.00000076, -0.00000439, -0.00000503, -0.00000015, 0.00000594, 0.00000669, 0.00000094, -0.00000522, -0.00000496, 0.00000152, 0.00000685, 0.00000460, -0.00000308, -0.00000730, -0.00000231, 0.00000726, 0.00001031, 0.00000123, -0.00001279, -0.00001764, -0.00000634, 0.00001277, 0.00002275, 0.00001384, -0.00000701, -0.00002226, -0.00001972, -0.00000330, 0.00001152, 0.00001216, -0.00000033, -0.00001379, -0.00001690, -0.00000860, 0.00000287, 0.00000890, 0.00000765, 0.00000364, 0.00000158, 0.00000176, 0.00000095, -0.00000307, -0.00000861, -0.00001154, -0.00000912, -0.00000256, 0.00000386, 0.00000593, 0.00000227, -0.00000494, -0.00001154, -0.00001406, -0.00001182, -0.00000726, -0.00000387, -0.00000321, -0.00000359, -0.00000179, 0.00000344, 0.00000950, },
		{ -0.00000355, 0.00000492, 0.00001258, 0.00001401, 0.00000805, -0.00000093, -0.00000670, -0.00000649, -0.00000274, 0.00000008, -0.00000011, -0.00000174, -0.00000241, -0.00000211, -0.00000291, -0.00000565, -0.00000779, -0.00000542, 0.00000226, 0.00001095, 0.00001445, 0.00001014, 0.00000128, -0.00000607, -0.00000833, -0.00000657, -0.00000403, -0.00000227, -0.00000030, 0.00000280, 0.00000549, 0.00000488, 0.00000021, -0.00000539, -0.00000733, -0.00000385, 0.00000221, 0.00000607, 0.00000524, 0.00000104, -0.00000353, -0.00000675, -0.00000857, -0.00000891, -0.00000651, -0.00000043, 0.00000770, 0.00001394, 0.00001510, 0.00001159, 0.00000653, 0.00000216, -0.00000238, -0.00000904, -0.00001684, -0.00002095, -0.00001680, -0.00000544, 0.00000607, 0.00001034, 0.00000624, -0.00000037, -0.00000258, 0.00000113, 0.00000634, 0.00000809, 0.00000597, 0.00000348, 0.00000292, 0.00000237, -0.00000150, -0.00000812, -0.00001208, -0.00000845, 0.00000114, 0.00000882, 0.00000797, -0.00000001, -0.00000692, -0.00000623, 0.00000016, 0.00000438, 0.00000162, -0.00000414, -0.00000454, 0.00000323, 0.00001190, 0.00001097, -0.00000166, -0.00001566, -0.00001730, -0.00000318, 0.00001523, 0.00002149, 0.00000946, -0.00001072, -0.00002193, -0.00001554, 0.00000125, 0.00001290, 0.00000976, -0.00000419, -0.00001644, -0.00001765, -0.00000883, 0.00000173, 0.00000699, 0.00000654, 0.00000466, 0.00000452, 0.00000501, 0.00000273, -0.00000355, -0.00001091, -0.00001442, -0.00001120, -0.00000287, 0.00000557, 0.00000906, 0.00000577, -0.00000194, -0.00000932, -0.00001255, -0.00001129, -0.00000848, -0.00000736, -0.00000825, -0.00000808, },
		{ -0.00000889, -0.00001300, -0.00001358, -0.00000719, 0.00000405, 0.00001335, 0.00001469, 0.00000796, -0.00000110, -0.00000610, -0.00000526, -0.00000180, 0.00000025, 0.00000014, 0.00000002, 0.00000118, 0.00000174, -0.00000129, -0.00000771, -0.00001286, -0.00001140, -0.00000268, 0.00000800, 0.00001378, 0.00001202, 0.00000584, 0.00000031, -0.00000235, -0.00000366, -0.00000551, -0.00000713, -0.00000588, -0.00000092, 0.00000477, 0.00000678, 0.00000357, -0.00000184, -0.00000482, -0.00000347, 0.00000008, 0.00000244, 0.00000235, 0.00000113, 0.00000016, -0.00000099, -0.00000366, -0.00000745, -0.00000967, -0.00000768, -0.00000172, 0.00000520, 0.00001023, 0.00001302, 0.00001461, 0.00001455, 0.00001014, -0.00000043, -0.00001421, -0.00002398, -0.00002358, -0.00001351, -0.00000088, 0.00000640, 0.00000596, 0.00000159, -0.00000143, -0.00000121, 0.00000067, 0.00000265, 0.00000534, 0.00000950, 0.00001293, 0.00001110, 0.00000203, -0.00000979, -0.00001592, -0.00001154, -0.00000058, 0.00000745, 0.00000637, -0.00000101, -0.00000634, -0.00000478, 0.00000027, 0.00000167, -0.00000269, -0.00000682, -0.00000331, 0.00000738, 0.00001560, 0.00001173, -0.00000307, -0.00001628, -0.00001533, -0.00000006, 0.00001618, 0.00001854, 0.00000455, -0.00001358, -0.00002033, -0.00001090, 0.00000511, 0.00001317, 0.00000691, -0.00000736, -0.00001759, -0.00001689, -0.00000811, 0.00000075, 0.00000446, 0.00000407, 0.00000377, 0.00000551, 0.00000701, 0.00000458, -0.00000257, -0.00001090, -0.00001505, -0.00001201, -0.00000348, 0.00000528, 0.00000904, 0.00000613, -0.00000069, -0.00000653, -0.00000826, -0.00000691, -0.00000627, },
		{ -0.00000365, -0.00000067, -0.00000138, -0.00000757, -0.00001522, -0.00001743, -0.00001066, 0.00000154, 0.00001112, 0.00001226, 0.00000606, -0.00000116, -0.00000401, -0.00000240, -0.00000017, -0.00000006, -0.00000083, 0.00000049, 0.00000403, 0.00000590, 0.00000211, -0.00000653, -0.00001404, -0.00001442, -0.00000724, 0.00000217, 0.00000775, 0.00000805, 0.00000623, 0.00000559, 0.00000576, 0.00000372, -0.00000180, -0.00000791, -0.00000995, -0.00000625, 0.00000011, 0.00000426, 0.00000414, 0.00000179, 0.00000036, 0.00000065, 0.00000084, -0.00000058, -0.00000255, -0.00000263, -0.00000015, 0.00000248, 0.00000211, -0.00000184, -0.00000669, -0.00000915, -0.00000811, -0.00000465, 0.00000023, 0.00000679, 0.00001484, 0.00002138, 0.00002129, 0.00001164, -0.00000443, -0.00001881, -0.00002423, -0.00001958, -0.00000998, -0.00000187, 0.00000185, 0.00000217, 0.00000104, -0.00000080, -0.00000282, -0.00000288, 0.00000171, 0.00001051, 0.00001807, 0.00001750, 0.00000714, -0.00000659, -0.00001401, -0.00001080, -0.00000185, 0.00000376, 0.00000164, -0.00000411, -0.00000622, -0.00000255, 0.00000164, 0.00000023, -0.00000599, -0.00000920, -0.00000302, 0.00000933, 0.00001677, 0.00001108, -0.00000394, -0.00001492, -0.00001142, 0.00000371, 0.00001640, 0.00001455, -0.00000064, -0.00001583, -0.00001800, -0.00000618, 0.00000828, 0.00001263, 0.00000379, -0.00001006, -0.00001790, -0.00001528, -0.00000664, 0.00000049, 0.00000249, 0.00000160, 0.00000197, 0.00000474, 0.00000698, 0.00000494, -0.00000193, -0.00000988, -0.00001364, -0.00001051, -0.00000251, 0.00000499, 0.00000722, 0.00000353, -0.00000227, -0.00000523, },
		{ -0.00000029, -0.00000576, -0.00000585, -0.00000048, 0.00000483, 0.00000381, -0.00000459, -0.00001445, -0.00001775, -0.00001148, -0.00000042, 0.00000714, 0.00000677, 0.00000128, -0.00000286, -0.00000214, 0.00000118, 0.00000250, 0.00000050, -0.00000148, 0.00000039, 0.00000540, 0.00000834, 0.00000480, -0.00000388, -0.00001144, -0.00001262, -0.00000798, -0.00000263, -0.00000062, -0.00000092, 0.00000058, 0.00000568, 0.00001119, 0.00001187, 0.00000596, -0.00000283, -0.00000868, -0.00000892, -0.00000552, -0.00000210, -0.00000002, 0.00000195, 0.00000471, 0.00000668, 0.00000519, -0.00000004, -0.00000566, -0.00000738, -0.00000405, 0.00000133, 0.00000445, 0.00000342, -0.00000043, -0.00000473, -0.00000813, -0.00001005, -0.00000908, -0.00000319, 0.00000766, 0.00001936, 0.00002521, 0.00002084, 0.00000798, -0.00000666, -0.00001643, -0.00001901, -0.00001635, -0.00001146, -0.00000607, -0.00000119, 0.00000148, 0.00000030, -0.00000362, -0.00000570, -0.00000136, 0.00000883, 0.00001810, 0.00001884, 0.00000983, -0.00000206, -0.00000800, -0.00000539, 0.00000016, 0.00000117, -0.00000376, -0.00000885, -0.00000802, -0.00000198, 0.00000227, -0.00000053, -0.00000741, -0.00000958, -0.00000205, 0.00001002, 0.00001544, 0.00000823, -0.00000555, -0.00001291, -0.00000647, 0.00000799, 0.00001661, 0.00001063, -0.00000527, -0.00001730, -0.00001535, -0.00000198, 0.00001051, 0.00001145, 0.00000077, -0.00001218, -0.00001765, -0.00001341, -0.00000497, 0.00000072, 0.00000138, -0.00000003, 0.00000043, 0.00000332, 0.00000549, 0.00000349, -0.00000279, -0.00000945, -0.00001161, -0.00000735, 0.00000042, 0.00000602, 0.00000543, },
		{ 0.00000422, 0.00000814, 0.00000455, -0.00000359, -0.00000920, -0.00000708, 0.00000120, 0.00000817, 0.00000697, -0.00000226, -0.00001219, -0.00001478, -0.00000856, 0.00000025, 0.00000384, 0.00000019, -0.00000560, -0.00000690, -0.00000233, 0.00000340, 0.00000492, 0.00000202, -0.00000053, 0.00000136, 0.00000631, 0.00000877, 0.00000512, -0.00000228, -0.00000745, -0.00000711, -0.00000391, -0.00000323, -0.00000689, -0.00001081, -0.00000910, -0.00000043, 0.00001016, 0.00001565, 0.00001320, 0.00000583, -0.00000129, -0.00000550, -0.00000748, -0.00000835, -0.00000726, -0.00000262, 0.00000474, 0.00001058, 0.00001048, 0.00000400, -0.00000444, -0.00000922, -0.00000804, -0.00000303, 0.00000193, 0.00000443, 0.00000424, 0.00000175, -0.00000288, -0.00000864, -0.00001253, -0.00001073, -0.00000198, 0.00001030, 0.00001980, 0.00002183, 0.00001635, 0.00000694, -0.00000257, -0.00001033, -0.00001559, -0.00001723, -0.00001424, -0.00000789, -0.00000223, -0.00000110, -0.00000402, -0.00000587, -0.00000171, 0.00000769, 0.00001566, 0.00001577, 0.00000830, 0.00000026, -0.00000161, 0.00000240, 0.00000551, 0.00000195, -0.00000660, -0.00001261, -0.00001076, -0.00000381, 0.00000040, -0.00000242, -0.00000800, -0.00000800, 0.00000052, 0.00001090, 0.00001303, 0.00000396, -0.00000817, -0.00001147, -0.00000203, 0.00001172, 0.00001666, 0.00000741, -0.00000845, -0.00001742, -0.00001233, 0.00000146, 0.00001155, 0.00000959, -0.00000205, -0.00001360, -0.00001687, -0.00001146, -0.00000346, 0.00000102, 0.00000080, -0.00000080, -0.00000033, 0.00000222, 0.00000364, 0.00000100, -0.00000507, -0.00001033, -0.00001025, -0.00000400, },
		{ -0.00001207, -0.00000980, -0.00000142, 0.00000747, 0.00001003, 0.00000400, -0.00000583, -0.00001112, -0.00000714, 0.00000273, 0.00000968, 0.00000734, -0.00000227, -0.00001057, -0.00001046, -0.00000298, 0.00000386, 0.00000301, -0.00000481, -0.00001205, -0.00001197, -0.00000494, 0.00000256, 0.00000498, 0.00000295, 0.00000163, 0.00000425, 0.00000832, 0.00000863, 0.00000351, -0.00000284, -0.00000462, -0.00000091, 0.00000289, 0.00000056, -0.00000824, -0.00001696, -0.00001802, -0.00000972, 0.00000268, 0.00001196, 0.00001479, 0.00001273, 0.00000869, 0.00000361, -0.00000292, -0.00000967, -0.00001288, -0.00000930, -0.00000004, 0.00000921, 0.00001246, 0.00000820, 0.00000015, -0.00000634, -0.00000839, -0.00000662, -0.00000308, 0.00000079, 0.00000408, 0.00000538, 0.00000307, -0.00000271, -0.00000896, -0.00001139, -0.00000785, 0.00000000, 0.00000823, 0.00001381, 0.00001605, 0.00001525, 0.00001108, 0.00000309, -0.00000711, -0.00001533, -0.00001764, -0.00001402, -0.00000880, -0.00000640, -0.00000686, -0.00000592, -0.00000011, 0.00000853, 0.00001369, 0.00001119, 0.00000387, -0.00000072, 0.00000206, 0.00000888, 0.00001158, 0.00000552, -0.00000545, -0.00001279, -0.00001192, -0.00000646, -0.00000376, -0.00000645, -0.00000949, -0.00000608, 0.00000389, 0.00001240, 0.00001097, -0.00000009, -0.00001082, -0.00001071, 0.00000090, 0.00001366, 0.00001569, 0.00000466, -0.00000999, -0.00001586, -0.00000867, 0.00000436, 0.00001157, 0.00000715, -0.00000469, -0.00001442, -0.00001561, -0.00000946, -0.00000214, 0.00000121, 0.00000040, -0.00000113, -0.00000047, 0.00000173, 0.00000216, -0.00000153, -0.00000788, },
		{ 0.00000110, -0.00000367, -0.00001040, -0.00001370, -0.00000958, 0.00000043, 0.00000946, 0.00001050, 0.00000259, -0.00000769, -0.00001141, -0.00000521, 0.00000538, 0.00001070, 0.00000587, -0.00000452, -0.00001064, -0.00000685, 0.00000287, 0.00000871, 0.00000444, -0.00000673, -0.00001552, -0.00001545, -0.00000829, -0.00000107, 0.00000155, 0.00000136, 0.00000309, 0.00000795, 0.00001171, 0.00000946, 0.00000190, -0.00000446, -0.00000361, 0.00000344, 0.00000900, 0.00000599, -0.00000528, -0.00001720, -0.00002176, -0.00001696, -0.00000715, 0.00000212, 0.00000859, 0.00001293, 0.00001531, 0.00001377, 0.00000664, -0.00000395, -0.00001219, -0.00001294, -0.00000615, 0.00000305, 0.00000868, 0.00000841, 0.00000418, -0.00000066, -0.00000418, -0.00000611, -0.00000633, -0.00000432, -0.00000030, 0.00000374, 0.00000509, 0.00000270, -0.00000175, -0.00000535, -0.00000648, -0.00000545, -0.00000301, 0.00000114, 0.00000749, 0.00001447, 0.00001806, 0.00001470, 0.00000495, -0.00000619, -0.00001329, -0.00001495, -0.00001405, -0.00001357, -0.00001277, -0.00000833, 0.00000073, 0.00001005, 0.00001307, 0.00000772, -0.00000054, -0.00000333, 0.00000255, 0.00001169, 0.00001515, 0.00000914, -0.00000160, -0.00000896, -0.00000959, -0.00000740, -0.00000801, -0.00001158, -0.00001235, -0.00000547, 0.00000621, 0.00001342, 0.00000937, -0.00000283, -0.00001212, -0.00000966, 0.00000259, 0.00001351, 0.00001307, 0.00000166, -0.00001055, -0.00001306, -0.00000443, 0.00000702, 0.00001109, 0.00000453, -0.00000705, -0.00001478, -0.00001410, -0.00000745, -0.00000094, 0.00000134, 0.00000005, -0.00000130, -0.00000037, 0.00000158, },
		},
		{ { -0.00000248, -0.00000260, -0.00000293, -0.00000342, -0.00000400, -0.00000455, -0.00000497, -0.00000516, -0.00000506, -0.00000465, -0.00000394, -0.00000294, -0.00000166, -0.00000013, 0.00000158, 0.00000334, 0.00000491, 0.00000608, 0.00000670, 0.00000678, 0.00000654, 0.00000634, 0.00000654, 0.00000730, 0.00000840, 0.00000933, 0.00000938, 0.00000797, 0.00000496, 0.00000078, -0.00000364, -0.00000720, -0.00000905, -0.00000895, -0.00000737, -0.00000527, -0.00000359, -0.00000284, -0.00000278, -0.00000252, -0.00000101, 0.00000233, 0.00000714, 0.00001206, 0.00001514, 0.00001467, 0.00000999, 0.00000203, -0.00000691, -0.00001393, -0.00001660, -0.00001400, -0.00000713, 0.00000142, 0.00000850, 0.00001164, 0.00001003, 0.00000482, -0.00000147, -0.00000606, -0.00000712, -0.00000452, 0.00000015, 0.00000443, 0.00000615, 0.00000443, 0.00000015, -0.00000452, -0.00000712, -0.00000606, -0.00000147, 0.00000482, 0.00001003, 0.00001164, 0.00000850, 0.00000142, -0.00000713, -0.00001400, -0.00001660, -0.00001393, -0.00000691, 0.00000203, 0.00000999, 0.00001467, 0.00001514, 0.00001206, 0.00000714, 0.00000233, -0.00000101, -0.00000252, -0.00000278, -0.00000284, -0.00000359, -0.00000527, -0.00000737, -0.00000895, -0.00000905, -0.00000720, -0.00000364, 0.00000078, 0.00000496, 0.00000797, 0.00000938, 0.00000933, 0.00000840, 0.00000730, 0.00000654, 0.00000634, 0.00000654, 0.00000678, 0.00000670, 0.00000608, 0.00000491, 0.00000334, 0.00000158, -0.00000013, -0.00000166, -0.00000294, -0.00000394, -0.00000465, -0.00000506, -0.00000516, -0.00000497, -0.00000455, -0.00000400, -0.00000342, -0.00000293, -0.00000260, },
		{ -0.00000239, -0.00000236, -0.00000226, -0.00000213, -0.00000203, -0.00000203, -0.00000220, -0.00000256, -0.00000310, -0.00000375, -0.00000442, -0.00000504, -0.00000558, -0.00000605, -0.00000646, -0.00000677, -0.00000687, -0.00000659, -0.00000575, -0.00000432, -0.00000239, -0.00000023, 0.00000184, 0.00000361, 0.00000507, 0.00000643, 0.00000801, 0.00001001, 0.00001233, 0.00001449, 0.00001572, 0.00001525, 0.00001269, 0.00000824, 0.00000276, -0.00000251, -0.00000643, -0.00000836, -0.00000841, -0.00000735, -0.00000616, -0.00000554, -0.00000549, -0.00000532, -0.00000399, -0.00000074, 0.00000432, 0.00000999, 0.00001430, 0.00001527, 0.00001181, 0.00000436, -0.00000504, -0.00001341, -0.00001791, -0.00001699, -0.00001102, -0.00000223, 0.00000619, 0.00001131, 0.00001164, 0.00000769, 0.00000160, -0.00000379, -0.00000622, -0.00000495, -0.00000101, 0.00000336, 0.00000583, 0.00000508, 0.00000142, -0.00000335, -0.00000681, -0.00000705, -0.00000364, 0.00000213, 0.00000775, 0.00001062, 0.00000916, 0.00000356, -0.00000423, -0.00001130, -0.00001496, -0.00001377, -0.00000809, 0.00000009, 0.00000804, 0.00001334, 0.00001475, 0.00001252, 0.00000810, 0.00000338, -0.00000007, -0.00000164, -0.00000168, -0.00000117, -0.00000111, -0.00000204, -0.00000377, -0.00000553, -0.00000640, -0.00000574, -0.00000350, -0.00000028, 0.00000299, 0.00000537, 0.00000633, 0.00000583, 0.00000435, 0.00000260, 0.00000126, 0.00000076, 0.00000116, 0.00000222, 0.00000352, 0.00000467, 0.00000538, 0.00000551, 0.00000511, 0.00000427, 0.00000318, 0.00000198, 0.00000081, -0.00000022, -0.00000105, -0.00000167, -0.00000208, -0.00000231, },
		{ 0.00000150, 0.00000144, 0.00000135, 0.00000115, 0.00000082, 0.00000038, -0.00000010, -0.00000054, -0.00000089, -0.00000118, -0.00000144, -0.00000171, -0.00000199, -0.00000224, -0.00000242, -0.00000254, -0.00000267, -0.00000298, -0.00000361, -0.00000463, -0.00000594, -0.00000727, -0.00000828, -0.00000865, -0.00000829, -0.00000731, -0.00000598, -0.00000457, -0.00000317, -0.00000163, 0.00000042, 0.00000330, 0.00000703, 0.00001122, 0.00001501, 0.00001738, 0.00001756, 0.00001533, 0.00001121, 0.00000631, 0.00000189, -0.00000112, -0.00000249, -0.00000270, -0.00000263, -0.00000305, -0.00000412, -0.00000525, -0.00000537, -0.00000348, 0.00000065, 0.00000617, 0.00001123, 0.00001367, 0.00001193, 0.00000584, -0.00000310, -0.00001215, -0.00001830, -0.00001943, -0.00001522, -0.00000724, 0.00000160, 0.00000823, 0.00001066, 0.00000866, 0.00000383, -0.00000128, -0.00000428, -0.00000397, -0.00000088, 0.00000317, 0.00000592, 0.00000580, 0.00000271, -0.00000196, -0.00000597, -0.00000727, -0.00000503, -0.00000006, 0.00000551, 0.00000911, 0.00000891, 0.00000463, -0.00000231, -0.00000930, -0.00001364, -0.00001360, -0.00000908, -0.00000160, 0.00000632, 0.00001223, 0.00001457, 0.00001321, 0.00000925, 0.00000449, 0.00000056, -0.00000161, -0.00000207, -0.00000160, -0.00000118, -0.00000150, -0.00000260, -0.00000396, -0.00000472, -0.00000419, -0.00000217, 0.00000095, 0.00000434, 0.00000703, 0.00000827, 0.00000781, 0.00000591, 0.00000323, 0.00000054, -0.00000149, -0.00000249, -0.00000242, -0.00000152, -0.00000021, 0.00000112, 0.00000215, 0.00000273, 0.00000285, 0.00000264, 0.00000227, 0.00000190, 0.00000164, },
		{ -0.00000032, -0.00000072, -0.00000087, -0.00000061, 0.00000007, 0.00000100, 0.00000195, 0.00000266, 0.00000297, 0.00000285, 0.00000238, 0.00000169, 0.00000090, 0.00000008, -0.00000072, -0.00000147, -0.00000209, -0.00000248, -0.00000251, -0.00000215, -0.00000152, -0.00000087, -0.00000054, -0.00000076, -0.00000162, -0.00000294, -0.00000436, -0.00000553, -0.00000621, -0.00000644, -0.00000646, -0.00000658, -0.00000692, -0.00000730, -0.00000724, -0.00000617, -0.00000370, 0.00000006, 0.00000446, 0.00000846, 0.00001102, 0.00001154, 0.00001011, 0.00000752, 0.00000492, 0.00000329, 0.00000304, 0.00000384, 0.00000482, 0.00000505, 0.00000410, 0.00000231, 0.00000070, 0.00000044, 0.00000224, 0.00000580, 0.00000977, 0.00001216, 0.00001126, 0.00000639, -0.00000160, -0.00001047, -0.00001743, -0.00002020, -0.00001792, -0.00001155, -0.00000350, 0.00000337, 0.00000686, 0.00000633, 0.00000282, -0.00000146, -0.00000421, -0.00000405, -0.00000114, 0.00000302, 0.00000629, 0.00000698, 0.00000464, 0.00000030, -0.00000400, -0.00000618, -0.00000508, -0.00000108, 0.00000407, 0.00000797, 0.00000863, 0.00000536, -0.00000086, -0.00000779, -0.00001281, -0.00001395, -0.00001069, -0.00000409, 0.00000368, 0.00001019, 0.00001367, 0.00001358, 0.00001063, 0.00000635, 0.00000236, -0.00000027, -0.00000131, -0.00000129, -0.00000109, -0.00000144, -0.00000255, -0.00000404, -0.00000513, -0.00000505, -0.00000339, -0.00000033, 0.00000347, 0.00000703, 0.00000946, 0.00001019, 0.00000918, 0.00000684, 0.00000387, 0.00000104, -0.00000107, -0.00000216, -0.00000228, -0.00000172, -0.00000089, -0.00000018, 0.00000014, 0.00000005, },
		{ 0.00000066, 0.00000126, 0.00000129, 0.00000064, -0.00000055, -0.00000187, -0.00000287, -0.00000320, -0.00000273, -0.00000158, -0.00000005, 0.00000146, 0.00000266, 0.00000336, 0.00000356, 0.00000335, 0.00000283, 0.00000206, 0.00000107, -0.00000009, -0.00000131, -0.00000241, -0.00000313, -0.00000332, -0.00000297, -0.00000227, -0.00000155, -0.00000116, -0.00000126, -0.00000177, -0.00000240, -0.00000277, -0.00000265, -0.00000209, -0.00000145, -0.00000121, -0.00000174, -0.00000304, -0.00000469, -0.00000599, -0.00000624, -0.00000515, -0.00000298, -0.00000051, 0.00000127, 0.00000164, 0.00000045, -0.00000169, -0.00000370, -0.00000448, -0.00000339, -0.00000058, 0.00000312, 0.00000657, 0.00000887, 0.00000977, 0.00000979, 0.00000988, 0.00001085, 0.00001281, 0.00001500, 0.00001598, 0.00001431, 0.00000928, 0.00000144, -0.00000739, -0.00001478, -0.00001854, -0.00001766, -0.00001275, -0.00000584, 0.00000042, 0.00000386, 0.00000360, 0.00000038, -0.00000383, -0.00000680, -0.00000700, -0.00000428, 0.00000012, 0.00000416, 0.00000605, 0.00000499, 0.00000165, -0.00000224, -0.00000463, -0.00000418, -0.00000086, 0.00000396, 0.00000811, 0.00000952, 0.00000722, 0.00000172, -0.00000514, -0.00001089, -0.00001345, -0.00001189, -0.00000678, 0.00000014, 0.00000662, 0.00001082, 0.00001190, 0.00001019, 0.00000688, 0.00000345, 0.00000102, -0.00000000, 0.00000003, 0.00000033, 0.00000014, -0.00000092, -0.00000264, -0.00000434, -0.00000520, -0.00000463, -0.00000248, 0.00000083, 0.00000449, 0.00000757, 0.00000934, 0.00000946, 0.00000807, 0.00000571, 0.00000310, 0.00000092, -0.00000035, -0.00000063, -0.00000014, },
		{ 0.00000012, -0.00000125, -0.00000137, -0.00000041, 0.00000114, 0.00000259, 0.00000335, 0.00000308, 0.00000185, 0.00000006, -0.00000172, -0.00000293, -0.00000325, -0.00000266, -0.00000142, 0.00000007, 0.00000144, 0.00000246, 0.00000308, 0.00000336, 0.00000339, 0.00000320, 0.00000275, 0.00000200, 0.00000097, -0.00000020, -0.00000127, -0.00000203, -0.00000237, -0.00000240, -0.00000239, -0.00000265, -0.00000335, -0.00000437, -0.00000535, -0.00000578, -0.00000530, -0.00000384, -0.00000176, 0.00000029, 0.00000167, 0.00000202, 0.00000146, 0.00000050, -0.00000015, -0.00000002, 0.00000088, 0.00000199, 0.00000238, 0.00000125, -0.00000171, -0.00000606, -0.00001072, -0.00001431, -0.00001574, -0.00001457, -0.00001113, -0.00000633, -0.00000121, 0.00000352, 0.00000770, 0.00001157, 0.00001545, 0.00001926, 0.00002231, 0.00002347, 0.00002160, 0.00001623, 0.00000799, -0.00000137, -0.00000947, -0.00001415, -0.00001430, -0.00001035, -0.00000415, 0.00000175, 0.00000508, 0.00000476, 0.00000125, -0.00000369, -0.00000781, -0.00000938, -0.00000790, -0.00000428, -0.00000030, 0.00000218, 0.00000212, -0.00000021, -0.00000340, -0.00000557, -0.00000527, -0.00000216, 0.00000274, 0.00000751, 0.00001012, 0.00000932, 0.00000518, -0.00000091, -0.00000677, -0.00001032, -0.00001038, -0.00000706, -0.00000164, 0.00000397, 0.00000799, 0.00000942, 0.00000830, 0.00000555, 0.00000249, 0.00000024, -0.00000066, -0.00000036, 0.00000045, 0.00000098, 0.00000068, -0.00000048, -0.00000204, -0.00000327, -0.00000347, -0.00000229, 0.00000013, 0.00000320, 0.00000611, 0.00000804, 0.00000849, 0.00000739, 0.00000515, 0.00000246, },
		{ 0.00000975, 0.00000874, 0.00000620, 0.00000282, -0.00000050, -0.00000288, -0.00000378, -0.00000315, -0.00000141, 0.00000069, 0.00000235, 0.00000300, 0.00000249, 0.00000112, -0.00000055, -0.00000186, -0.00000238, -0.00000198, -0.00000087, 0.00000053, 0.00000179, 0.00000267, 0.00000310, 0.00000321, 0.00000316, 0.00000307, 0.00000292, 0.00000265, 0.00000220, 0.00000162, 0.00000105, 0.00000068, 0.00000057, 0.00000062, 0.00000052, -0.00000010, -0.00000147, -0.00000354, -0.00000588, -0.00000786, -0.00000885, -0.00000853, -0.00000701, -0.00000481, -0.00000264, -0.00000105, -0.00000020, 0.00000023, 0.00000082, 0.00000212, 0.00000425, 0.00000676, 0.00000874, 0.00000914, 0.00000724, 0.00000298, -0.00000296, -0.00000938, -0.00001494, -0.00001864, -0.00002003, -0.00001920, -0.00001656, -0.00001252, -0.00000732, -0.00000106, 0.00000603, 0.00001335, 0.00001979, 0.00002390, 0.00002440, 0.00002073, 0.00001351, 0.00000450, -0.00000383, -0.00000911, -0.00000999, -0.00000660, -0.00000055, 0.00000565, 0.00000962, 0.00000996, 0.00000678, 0.00000155, -0.00000358, -0.00000675, -0.00000710, -0.00000511, -0.00000227, -0.00000032, -0.00000042, -0.00000262, -0.00000581, -0.00000832, -0.00000865, -0.00000620, -0.00000163, 0.00000347, 0.00000718, 0.00000808, 0.00000586, 0.00000144, -0.00000342, -0.00000681, -0.00000746, -0.00000515, -0.00000077, 0.00000408, 0.00000778, 0.00000924, 0.00000830, 0.00000560, 0.00000230, -0.00000046, -0.00000197, -0.00000215, -0.00000149, -0.00000068, -0.00000036, -0.00000076, -0.00000166, -0.00000249, -0.00000261, -0.00000157, 0.00000067, 0.00000370, 0.00000676, 0.00000901, },
		{ -0.00000118, 0.00000217, 0.00000596, 0.00000926, 0.00001120, 0.00001122, 0.00000927, 0.00000588, 0.00000196, -0.00000147, -0.00000358, -0.00000403, -0.00000304, -0.00000129, 0.00000034, 0.00000116, 0.00000084, -0.00000044, -0.00000209, -0.00000340, -0.00000382, -0.00000315, -0.00000158, 0.00000040, 0.00000224, 0.00000354, 0.00000418, 0.00000425, 0.00000399, 0.00000363, 0.00000325, 0.00000287, 0.00000245, 0.00000202, 0.00000172, 0.00000172, 0.00000212, 0.00000280, 0.00000344, 0.00000358, 0.00000280, 0.00000100, -0.00000158, -0.00000433, -0.00000656, -0.00000771, -0.00000767, -0.00000674, -0.00000551, -0.00000461, -0.00000432, -0.00000450, -0.00000459, -0.00000393, -0.00000208, 0.00000091, 0.00000445, 0.00000760, 0.00000940, 0.00000923, 0.00000702, 0.00000318, -0.00000158, -0.00000648, -0.00001094, -0.00001454, -0.00001694, -0.00001779, -0.00001665, -0.00001318, -0.00000740, 0.00000004, 0.00000780, 0.00001408, 0.00001718, 0.00001610, 0.00001109, 0.00000365, -0.00000383, -0.00000887, -0.00000981, -0.00000643, -0.00000001, 0.00000714, 0.00001256, 0.00001458, 0.00001291, 0.00000863, 0.00000365, -0.00000013, -0.00000163, -0.00000095, 0.00000079, 0.00000202, 0.00000155, -0.00000090, -0.00000454, -0.00000792, -0.00000956, -0.00000867, -0.00000551, -0.00000131, 0.00000227, 0.00000382, 0.00000280, -0.00000028, -0.00000405, -0.00000687, -0.00000746, -0.00000541, -0.00000132, 0.00000354, 0.00000768, 0.00000995, 0.00000993, 0.00000797, 0.00000496, 0.00000196, -0.00000025, -0.00000137, -0.00000161, -0.00000148, -0.00000154, -0.00000209, -0.00000306, -0.00000399, -0.00000428, -0.00000341, },
		{ -0.00000179, -0.00000350, -0.00000481, -0.00000507, -0.00000387, -0.00000118, 0.00000251, 0.00000636, 0.00000937, 0.00001072, 0.00001006, 0.00000762, 0.00000417, 0.00000074, -0.00000169, -0.00000257, -0.00000195, -0.00000042, 0.00000112, 0.00000180, 0.00000119, -0.00000063, -0.00000306, -0.00000526, -0.00000647, -0.00000631, -0.00000484, -0.00000257, -0.00000013, 0.00000191, 0.00000324, 0.00000387, 0.00000400, 0.00000388, 0.00000368, 0.00000342, 0.00000309, 0.00000271, 0.00000238, 0.00000231, 0.00000265, 0.00000341, 0.00000435, 0.00000505, 0.00000503, 0.00000402, 0.00000210, -0.00000033, -0.00000259, -0.00000408, -0.00000446, -0.00000385, -0.00000272, -0.00000173, -0.00000142, -0.00000197, -0.00000318, -0.00000455, -0.00000550, -0.00000566, -0.00000494, -0.00000357, -0.00000193, -0.00000038, 0.00000083, 0.00000158, 0.00000179, 0.00000133, 0.00000008, -0.00000204, -0.00000483, -0.00000772, -0.00000981, -0.00001016, -0.00000815, -0.00000388, 0.00000164, 0.00000676, 0.00000965, 0.00000903, 0.00000473, -0.00000212, -0.00000936, -0.00001457, -0.00001590, -0.00001279, -0.00000617, 0.00000190, 0.00000898, 0.00001321, 0.00001388, 0.00001163, 0.00000804, 0.00000488, 0.00000337, 0.00000373, 0.00000516, 0.00000636, 0.00000617, 0.00000412, 0.00000071, -0.00000291, -0.00000539, -0.00000586, -0.00000430, -0.00000156, 0.00000100, 0.00000211, 0.00000112, -0.00000171, -0.00000537, -0.00000845, -0.00000972, -0.00000860, -0.00000533, -0.00000086, 0.00000354, 0.00000677, 0.00000821, 0.00000789, 0.00000634, 0.00000437, 0.00000268, 0.00000165, 0.00000123, 0.00000105, 0.00000065, -0.00000029, },
		{ 0.00000174, 0.00000237, 0.00000280, 0.00000262, 0.00000165, 0.00000006, -0.00000167, -0.00000285, -0.00000289, -0.00000151, 0.00000112, 0.00000436, 0.00000725, 0.00000890, 0.00000872, 0.00000672, 0.00000353, 0.00000016, -0.00000231, -0.00000316, -0.00000226, -0.00000016, 0.00000217, 0.00000368, 0.00000366, 0.00000202, -0.00000076, -0.00000378, -0.00000608, -0.00000704, -0.00000651, -0.00000486, -0.00000273, -0.00000076, 0.00000063, 0.00000134, 0.00000154, 0.00000153, 0.00000153, 0.00000162, 0.00000178, 0.00000194, 0.00000211, 0.00000239, 0.00000294, 0.00000382, 0.00000492, 0.00000594, 0.00000643, 0.00000605, 0.00000466, 0.00000247, -0.00000001, -0.00000214, -0.00000336, -0.00000341, -0.00000235, -0.00000055, 0.00000145, 0.00000316, 0.00000418, 0.00000430, 0.00000345, 0.00000167, -0.00000091, -0.00000401, -0.00000721, -0.00000993, -0.00001160, -0.00001181, -0.00001053, -0.00000815, -0.00000542, -0.00000317, -0.00000197, -0.00000183, -0.00000219, -0.00000210, -0.00000071, 0.00000228, 0.00000633, 0.00001011, 0.00001197, 0.00001060, 0.00000560, -0.00000215, -0.00001071, -0.00001770, -0.00002110, -0.00001998, -0.00001485, -0.00000741, 0.00000008, 0.00000564, 0.00000824, 0.00000806, 0.00000624, 0.00000430, 0.00000345, 0.00000410, 0.00000577, 0.00000744, 0.00000806, 0.00000710, 0.00000477, 0.00000194, -0.00000026, -0.00000100, -0.00000009, 0.00000190, 0.00000393, 0.00000488, 0.00000405, 0.00000146, -0.00000220, -0.00000580, -0.00000826, -0.00000886, -0.00000755, -0.00000486, -0.00000167, 0.00000110, 0.00000283, 0.00000336, 0.00000295, 0.00000215, 0.00000150, 0.00000135, },
		{ -0.00000029, -0.00000124, -0.00000203, -0.00000228, -0.00000186, -0.00000089, 0.00000025, 0.00000113, 0.00000140, 0.00000099, 0.00000017, -0.00000055, -0.00000056, 0.00000056, 0.00000283, 0.00000578, 0.00000858, 0.00001025, 0.00001007, 0.00000788, 0.00000417, -0.00000004, -0.00000352, -0.00000528, -0.00000493, -0.00000283, 0.00000008, 0.00000260, 0.00000375, 0.00000313, 0.00000101, -0.00000180, -0.00000429, -0.00000565, -0.00000556, -0.00000424, -0.00000231, -0.00000049, 0.00000067, 0.00000098, 0.00000059, -0.00000014, -0.00000083, -0.00000125, -0.00000133, -0.00000117, -0.00000084, -0.00000038, 0.00000030, 0.00000128, 0.00000261, 0.00000418, 0.00000569, 0.00000677, 0.00000708, 0.00000641, 0.00000479, 0.00000248, -0.00000010, -0.00000241, -0.00000395, -0.00000430, -0.00000322, -0.00000071, 0.00000288, 0.00000692, 0.00001049, 0.00001259, 0.00001243, 0.00000965, 0.00000452, -0.00000203, -0.00000868, -0.00001409, -0.00001731, -0.00001805, -0.00001677, -0.00001434, -0.00001169, -0.00000937, -0.00000732, -0.00000500, -0.00000173, 0.00000281, 0.00000821, 0.00001333, 0.00001658, 0.00001655, 0.00001263, 0.00000537, -0.00000360, -0.00001200, -0.00001772, -0.00001947, -0.00001725, -0.00001223, -0.00000631, -0.00000131, 0.00000158, 0.00000220, 0.00000130, 0.00000006, -0.00000044, 0.00000029, 0.00000202, 0.00000397, 0.00000527, 0.00000536, 0.00000427, 0.00000258, 0.00000117, 0.00000077, 0.00000166, 0.00000353, 0.00000562, 0.00000705, 0.00000714, 0.00000573, 0.00000317, 0.00000020, -0.00000234, -0.00000383, -0.00000404, -0.00000317, -0.00000176, -0.00000041, 0.00000037, 0.00000037, },
		{ 0.00000138, 0.00000215, 0.00000257, 0.00000238, 0.00000155, 0.00000025, -0.00000120, -0.00000243, -0.00000314, -0.00000325, -0.00000290, -0.00000239, -0.00000207, -0.00000212, -0.00000245, -0.00000269, -0.00000230, -0.00000080, 0.00000195, 0.00000560, 0.00000934, 0.00001213, 0.00001302, 0.00001158, 0.00000805, 0.00000339, -0.00000111, -0.00000421, -0.00000519, -0.00000410, -0.00000169, 0.00000084, 0.00000237, 0.00000224, 0.00000049, -0.00000219, -0.00000478, -0.00000633, -0.00000632, -0.00000481, -0.00000238, 0.00000012, 0.00000196, 0.00000272, 0.00000241, 0.00000140, 0.00000019, -0.00000080, -0.00000136, -0.00000153, -0.00000152, -0.00000158, -0.00000180, -0.00000214, -0.00000237, -0.00000218, -0.00000132, 0.00000029, 0.00000251, 0.00000492, 0.00000695, 0.00000795, 0.00000746, 0.00000535, 0.00000194, -0.00000192, -0.00000512, -0.00000652, -0.00000536, -0.00000156, 0.00000413, 0.00001029, 0.00001524, 0.00001749, 0.00001631, 0.00001188, 0.00000527, -0.00000196, -0.00000831, -0.00001280, -0.00001520, -0.00001593, -0.00001569, -0.00001506, -0.00001414, -0.00001254, -0.00000967, -0.00000517, 0.00000071, 0.00000699, 0.00001217, 0.00001473, 0.00001374, 0.00000930, 0.00000254, -0.00000466, -0.00001037, -0.00001319, -0.00001278, -0.00000981, -0.00000572, -0.00000204, 0.00000012, 0.00000044, -0.00000062, -0.00000214, -0.00000317, -0.00000320, -0.00000223, -0.00000077, 0.00000051, 0.00000110, 0.00000088, 0.00000017, -0.00000048, -0.00000053, 0.00000028, 0.00000185, 0.00000371, 0.00000528, 0.00000607, 0.00000586, 0.00000477, 0.00000321, 0.00000169, 0.00000064, 0.00000030, 0.00000064, },
		{ 0.00000162, 0.00000092, 0.00000056, 0.00000064, 0.00000116, 0.00000197, 0.00000285, 0.00000352, 0.00000376, 0.00000345, 0.00000266, 0.00000160, 0.00000054, -0.00000032, -0.00000090, -0.00000135, -0.00000190, -0.00000272, -0.00000381, -0.00000484, -0.00000526, -0.00000450, -0.00000223, 0.00000141, 0.00000573, 0.00000966, 0.00001206, 0.00001221, 0.00001001, 0.00000614, 0.00000182, -0.00000162, -0.00000322, -0.00000273, -0.00000069, 0.00000181, 0.00000355, 0.00000365, 0.00000191, -0.00000115, -0.00000449, -0.00000701, -0.00000790, -0.00000697, -0.00000465, -0.00000179, 0.00000066, 0.00000203, 0.00000213, 0.00000125, 0.00000002, -0.00000085, -0.00000088, -0.00000001, 0.00000140, 0.00000271, 0.00000317, 0.00000230, 0.00000002, -0.00000319, -0.00000644, -0.00000864, -0.00000888, -0.00000678, -0.00000264, 0.00000250, 0.00000720, 0.00001003, 0.00001009, 0.00000734, 0.00000271, -0.00000219, -0.00000560, -0.00000622, -0.00000366, 0.00000135, 0.00000727, 0.00001227, 0.00001485, 0.00001435, 0.00001112, 0.00000624, 0.00000110, -0.00000317, -0.00000604, -0.00000765, -0.00000855, -0.00000927, -0.00000998, -0.00001037, -0.00000983, -0.00000783, -0.00000429, 0.00000021, 0.00000452, 0.00000730, 0.00000760, 0.00000519, 0.00000075, -0.00000433, -0.00000849, -0.00001052, -0.00000997, -0.00000728, -0.00000353, -0.00000002, 0.00000223, 0.00000281, 0.00000191, 0.00000021, -0.00000154, -0.00000274, -0.00000321, -0.00000310, -0.00000276, -0.00000250, -0.00000246, -0.00000255, -0.00000252, -0.00000214, -0.00000130, -0.00000008, 0.00000128, 0.00000248, 0.00000324, 0.00000345, 0.00000313, 0.00000243, },
		{ 0.00000156, 0.00000214, 0.00000244, 0.00000241, 0.00000200, 0.00000124, 0.00000027, -0.00000070, -0.00000137, -0.00000153, -0.00000111, -0.00000024, 0.00000081, 0.00000173, 0.00000231, 0.00000250, 0.00000242, 0.00000226, 0.00000214, 0.00000200, 0.00000164, 0.00000083, -0.00000050, -0.00000214, -0.00000357, -0.00000414, -0.00000330, -0.00000092, 0.00000258, 0.00000628, 0.00000905, 0.00000994, 0.00000860, 0.00000542, 0.00000143, -0.00000204, -0.00000386, -0.00000353, -0.00000134, 0.00000174, 0.00000442, 0.00000558, 0.00000472, 0.00000207, -0.00000145, -0.00000464, -0.00000643, -0.00000628, -0.00000438, -0.00000154, 0.00000112, 0.00000255, 0.00000218, 0.00000013, -0.00000283, -0.00000551, -0.00000673, -0.00000571, -0.00000244, 0.00000230, 0.00000709, 0.00001032, 0.00001069, 0.00000774, 0.00000201, -0.00000500, -0.00001128, -0.00001496, -0.00001494, -0.00001127, -0.00000513, 0.00000150, 0.00000661, 0.00000882, 0.00000784, 0.00000457, 0.00000066, -0.00000211, -0.00000256, -0.00000047, 0.00000329, 0.00000725, 0.00000990, 0.00001026, 0.00000828, 0.00000471, 0.00000078, -0.00000238, -0.00000413, -0.00000447, -0.00000393, -0.00000320, -0.00000271, -0.00000248, -0.00000214, -0.00000123, 0.00000049, 0.00000277, 0.00000490, 0.00000594, 0.00000518, 0.00000241, -0.00000186, -0.00000656, -0.00001040, -0.00001234, -0.00001192, -0.00000938, -0.00000556, -0.00000151, 0.00000179, 0.00000376, 0.00000431, 0.00000369, 0.00000236, 0.00000076, -0.00000078, -0.00000213, -0.00000322, -0.00000403, -0.00000451, -0.00000459, -0.00000425, -0.00000352, -0.00000251, -0.00000137, -0.00000024, 0.00000076, },
		{ -0.00000392, -0.00000322, -0.00000259, -0.00000193, -0.00000107, 0.00000010, 0.00000150, 0.00000285, 0.00000375, 0.00000386, 0.00000304, 0.00000146, -0.00000042, -0.00000209, -0.00000313, -0.00000336, -0.00000294, -0.00000219, -0.00000145, -0.00000093, -0.00000061, -0.00000032, 0.00000011, 0.00000070, 0.00000129, 0.00000157, 0.00000132, 0.00000051, -0.00000051, -0.00000117, -0.00000086, 0.00000076, 0.00000350, 0.00000668, 0.00000926, 0.00001025, 0.00000910, 0.00000595, 0.00000166, -0.00000251, -0.00000528, -0.00000588, -0.00000428, -0.00000122, 0.00000204, 0.00000422, 0.00000444, 0.00000257, -0.00000071, -0.00000417, -0.00000646, -0.00000657, -0.00000423, -0.00000004, 0.00000469, 0.00000832, 0.00000945, 0.00000738, 0.00000241, -0.00000414, -0.00001031, -0.00001405, -0.00001397, -0.00000976, -0.00000246, 0.00000587, 0.00001271, 0.00001593, 0.00001448, 0.00000877, 0.00000056, -0.00000768, -0.00001357, -0.00001564, -0.00001377, -0.00000914, -0.00000371, 0.00000061, 0.00000265, 0.00000234, 0.00000065, -0.00000093, -0.00000110, 0.00000069, 0.00000398, 0.00000759, 0.00001007, 0.00001034, 0.00000811, 0.00000399, -0.00000082, -0.00000499, -0.00000759, -0.00000828, -0.00000736, -0.00000547, -0.00000322, -0.00000098, 0.00000121, 0.00000349, 0.00000593, 0.00000831, 0.00001012, 0.00001068, 0.00000948, 0.00000641, 0.00000187, -0.00000324, -0.00000784, -0.00001099, -0.00001214, -0.00001126, -0.00000880, -0.00000545, -0.00000194, 0.00000113, 0.00000338, 0.00000463, 0.00000486, 0.00000416, 0.00000269, 0.00000073, -0.00000138, -0.00000326, -0.00000459, -0.00000521, -0.00000516, -0.00000464, },
		{ -0.00000241, -0.00000323, -0.00000333, -0.00000311, -0.00000302, -0.00000336, -0.00000411, -0.00000490, -0.00000518, -0.00000452, -0.00000282, -0.00000042, 0.00000203, 0.00000379, 0.00000435, 0.00000363, 0.00000200, 0.00000012, -0.00000140, -0.00000217, -0.00000221, -0.00000184, -0.00000149, -0.00000148, -0.00000187, -0.00000244, -0.00000289, -0.00000299, -0.00000272, -0.00000228, -0.00000198, -0.00000198, -0.00000216, -0.00000212, -0.00000135, 0.00000052, 0.00000346, 0.00000695, 0.00001007, 0.00001182, 0.00001147, 0.00000892, 0.00000478, 0.00000018, -0.00000354, -0.00000533, -0.00000477, -0.00000226, 0.00000111, 0.00000394, 0.00000496, 0.00000350, -0.00000027, -0.00000530, -0.00000997, -0.00001258, -0.00001190, -0.00000760, -0.00000052, 0.00000755, 0.00001429, 0.00001758, 0.00001618, 0.00001022, 0.00000126, -0.00000814, -0.00001520, -0.00001778, -0.00001516, -0.00000825, 0.00000067, 0.00000874, 0.00001351, 0.00001372, 0.00000970, 0.00000315, -0.00000358, -0.00000838, -0.00001013, -0.00000901, -0.00000625, -0.00000349, -0.00000206, -0.00000237, -0.00000385, -0.00000526, -0.00000529, -0.00000325, 0.00000065, 0.00000533, 0.00000926, 0.00001110, 0.00001018, 0.00000673, 0.00000175, -0.00000344, -0.00000761, -0.00001005, -0.00001062, -0.00000964, -0.00000761, -0.00000496, -0.00000194, 0.00000132, 0.00000471, 0.00000795, 0.00001058, 0.00001208, 0.00001204, 0.00001032, 0.00000717, 0.00000317, -0.00000093, -0.00000446, -0.00000690, -0.00000803, -0.00000786, -0.00000657, -0.00000441, -0.00000173, 0.00000109, 0.00000358, 0.00000531, 0.00000592, 0.00000530, 0.00000365, 0.00000142, -0.00000078, },
		{ 0.00000655, 0.00000518, 0.00000305, 0.00000100, -0.00000025, -0.00000036, 0.00000046, 0.00000155, 0.00000209, 0.00000149, -0.00000031, -0.00000281, -0.00000512, -0.00000639, -0.00000615, -0.00000454, -0.00000223, -0.00000013, 0.00000098, 0.00000081, -0.00000037, -0.00000188, -0.00000294, -0.00000304, -0.00000213, -0.00000062, 0.00000085, 0.00000171, 0.00000168, 0.00000084, -0.00000043, -0.00000172, -0.00000277, -0.00000354, -0.00000417, -0.00000478, -0.00000533, -0.00000555, -0.00000500, -0.00000332, -0.00000045, 0.00000325, 0.00000699, 0.00000976, 0.00001071, 0.00000944, 0.00000620, 0.00000186, -0.00000232, -0.00000506, -0.00000547, -0.00000335, 0.00000067, 0.00000533, 0.00000900, 0.00001021, 0.00000810, 0.00000277, -0.00000464, -0.00001217, -0.00001757, -0.00001897, -0.00001553, -0.00000782, 0.00000224, 0.00001191, 0.00001841, 0.00001988, 0.00001599, 0.00000811, -0.00000118, -0.00000894, -0.00001288, -0.00001214, -0.00000752, -0.00000109, 0.00000463, 0.00000768, 0.00000728, 0.00000408, -0.00000027, -0.00000391, -0.00000551, -0.00000484, -0.00000271, -0.00000055, 0.00000032, -0.00000069, -0.00000320, -0.00000604, -0.00000776, -0.00000734, -0.00000459, -0.00000027, 0.00000423, 0.00000744, 0.00000832, 0.00000667, 0.00000304, -0.00000149, -0.00000575, -0.00000884, -0.00001035, -0.00001032, -0.00000905, -0.00000693, -0.00000431, -0.00000141, 0.00000153, 0.00000429, 0.00000658, 0.00000811, 0.00000870, 0.00000832, 0.00000713, 0.00000538, 0.00000334, 0.00000124, -0.00000075, -0.00000246, -0.00000371, -0.00000421, -0.00000375, -0.00000224, 0.00000014, 0.00000287, 0.00000525, 0.00000660, },
		{ 0.00000032, 0.00000322, 0.00000599, 0.00000765, 0.00000757, 0.00000576, 0.00000294, 0.00000025, -0.00000120, -0.00000080, 0.00000128, 0.00000416, 0.00000661, 0.00000760, 0.00000671, 0.00000432, 0.00000137, -0.00000102, -0.00000208, -0.00000172, -0.00000054, 0.00000048, 0.00000044, -0.00000107, -0.00000375, -0.00000675, -0.00000902, -0.00000974, -0.00000866, -0.00000611, -0.00000290, 0.00000008, 0.00000217, 0.00000312, 0.00000306, 0.00000234, 0.00000131, 0.00000016, -0.00000107, -0.00000241, -0.00000378, -0.00000495, -0.00000551, -0.00000505, -0.00000335, -0.00000056, 0.00000280, 0.00000591, 0.00000787, 0.00000798, 0.00000600, 0.00000221, -0.00000254, -0.00000705, -0.00001003, -0.00001043, -0.00000780, -0.00000250, 0.00000429, 0.00001079, 0.00001504, 0.00001551, 0.00001164, 0.00000413, -0.00000511, -0.00001349, -0.00001851, -0.00001864, -0.00001382, -0.00000558, 0.00000348, 0.00001058, 0.00001375, 0.00001246, 0.00000778, 0.00000187, -0.00000289, -0.00000486, -0.00000369, -0.00000040, 0.00000318, 0.00000526, 0.00000481, 0.00000203, -0.00000185, -0.00000515, -0.00000650, -0.00000547, -0.00000266, 0.00000055, 0.00000268, 0.00000279, 0.00000088, -0.00000213, -0.00000486, -0.00000607, -0.00000519, -0.00000252, 0.00000090, 0.00000374, 0.00000493, 0.00000397, 0.00000113, -0.00000277, -0.00000668, -0.00000968, -0.00001119, -0.00001108, -0.00000957, -0.00000708, -0.00000411, -0.00000114, 0.00000143, 0.00000331, 0.00000436, 0.00000461, 0.00000428, 0.00000369, 0.00000314, 0.00000279, 0.00000259, 0.00000231, 0.00000166, 0.00000052, -0.00000091, -0.00000216, -0.00000261, -0.00000179, },
		{ -0.00000094, -0.00000278, -0.00000385, -0.00000331, -0.00000089, 0.00000288, 0.00000682, 0.00000951, 0.00000989, 0.00000773, 0.00000378, -0.00000050, -0.00000351, -0.00000417, -0.00000232, 0.00000124, 0.00000507, 0.00000781, 0.00000869, 0.00000782, 0.00000610, 0.00000471, 0.00000456, 0.00000579, 0.00000777, 0.00000929, 0.00000917, 0.00000676, 0.00000225, -0.00000340, -0.00000880, -0.00001268, -0.00001428, -0.00001357, -0.00001116, -0.00000796, -0.00000484, -0.00000236, -0.00000069, 0.00000029, 0.00000081, 0.00000103, 0.00000099, 0.00000067, 0.00000004, -0.00000080, -0.00000162, -0.00000209, -0.00000193, -0.00000094, 0.00000084, 0.00000314, 0.00000547, 0.00000719, 0.00000764, 0.00000629, 0.00000293, -0.00000215, -0.00000809, -0.00001354, -0.00001694, -0.00001703, -0.00001328, -0.00000626, 0.00000242, 0.00001044, 0.00001552, 0.00001620, 0.00001237, 0.00000536, -0.00000246, -0.00000858, -0.00001123, -0.00000998, -0.00000585, -0.00000082, 0.00000305, 0.00000441, 0.00000318, 0.00000046, -0.00000203, -0.00000280, -0.00000125, 0.00000203, 0.00000560, 0.00000778, 0.00000746, 0.00000459, 0.00000019, -0.00000408, -0.00000668, -0.00000679, -0.00000465, -0.00000138, 0.00000155, 0.00000296, 0.00000247, 0.00000059, -0.00000160, -0.00000290, -0.00000263, -0.00000080, 0.00000182, 0.00000409, 0.00000494, 0.00000378, 0.00000069, -0.00000367, -0.00000830, -0.00001213, -0.00001438, -0.00001463, -0.00001293, -0.00000968, -0.00000557, -0.00000138, 0.00000209, 0.00000427, 0.00000495, 0.00000429, 0.00000284, 0.00000132, 0.00000037, 0.00000025, 0.00000080, 0.00000146, 0.00000157, 0.00000074, },
		{ 0.00000121, 0.00000265, 0.00000320, 0.00000213, -0.00000050, -0.00000385, -0.00000651, -0.00000712, -0.00000499, -0.00000049, 0.00000502, 0.00000964, 0.00001175, 0.00001059, 0.00000666, 0.00000148, -0.00000303, -0.00000532, -0.00000485, -0.00000220, 0.00000124, 0.00000392, 0.00000491, 0.00000421, 0.00000279, 0.00000201, 0.00000302, 0.00000613, 0.00001066, 0.00001517, 0.00001801, 0.00001803, 0.00001493, 0.00000943, 0.00000287, -0.00000325, -0.00000777, -0.00001022, -0.00001080, -0.00001015, -0.00000902, -0.00000798, -0.00000726, -0.00000680, -0.00000638, -0.00000573, -0.00000472, -0.00000335, -0.00000174, -0.00000014, 0.00000118, 0.00000195, 0.00000205, 0.00000155, 0.00000078, 0.00000028, 0.00000064, 0.00000219, 0.00000479, 0.00000770, 0.00000970, 0.00000950, 0.00000627, 0.00000009, -0.00000786, -0.00001556, -0.00002077, -0.00002184, -0.00001837, -0.00001137, -0.00000297, 0.00000436, 0.00000873, 0.00000944, 0.00000719, 0.00000364, 0.00000063, -0.00000061, 0.00000006, 0.00000175, 0.00000307, 0.00000288, 0.00000088, -0.00000222, -0.00000502, -0.00000612, -0.00000477, -0.00000132, 0.00000297, 0.00000639, 0.00000760, 0.00000615, 0.00000266, -0.00000146, -0.00000462, -0.00000577, -0.00000475, -0.00000231, 0.00000032, 0.00000198, 0.00000212, 0.00000098, -0.00000057, -0.00000149, -0.00000104, 0.00000089, 0.00000371, 0.00000637, 0.00000775, 0.00000708, 0.00000416, -0.00000057, -0.00000621, -0.00001162, -0.00001571, -0.00001770, -0.00001722, -0.00001443, -0.00000999, -0.00000488, -0.00000025, 0.00000295, 0.00000425, 0.00000374, 0.00000214, 0.00000044, -0.00000044, -0.00000010, },
		{ -0.00000170, -0.00000334, -0.00000363, -0.00000203, 0.00000111, 0.00000455, 0.00000675, 0.00000647, 0.00000346, -0.00000139, -0.00000628, -0.00000921, -0.00000884, -0.00000509, 0.00000078, 0.00000666, 0.00001041, 0.00001077, 0.00000785, 0.00000304, -0.00000162, -0.00000432, -0.00000424, -0.00000185, 0.00000139, 0.00000371, 0.00000392, 0.00000192, -0.00000126, -0.00000393, -0.00000448, -0.00000210, 0.00000284, 0.00000902, 0.00001461, 0.00001805, 0.00001857, 0.00001636, 0.00001241, 0.00000805, 0.00000439, 0.00000202, 0.00000091, 0.00000053, 0.00000017, -0.00000074, -0.00000253, -0.00000510, -0.00000807, -0.00001079, -0.00001255, -0.00001272, -0.00001094, -0.00000735, -0.00000259, 0.00000217, 0.00000562, 0.00000667, 0.00000495, 0.00000106, -0.00000352, -0.00000686, -0.00000732, -0.00000426, 0.00000162, 0.00000841, 0.00001362, 0.00001512, 0.00001198, 0.00000485, -0.00000424, -0.00001263, -0.00001807, -0.00001945, -0.00001712, -0.00001259, -0.00000779, -0.00000421, -0.00000236, -0.00000175, -0.00000133, -0.00000014, 0.00000209, 0.00000480, 0.00000687, 0.00000724, 0.00000548, 0.00000212, -0.00000159, -0.00000414, -0.00000452, -0.00000263, 0.00000065, 0.00000383, 0.00000551, 0.00000496, 0.00000243, -0.00000101, -0.00000397, -0.00000538, -0.00000488, -0.00000296, -0.00000066, 0.00000095, 0.00000126, 0.00000034, -0.00000110, -0.00000205, -0.00000169, 0.00000024, 0.00000339, 0.00000685, 0.00000951, 0.00001038, 0.00000892, 0.00000519, -0.00000023, -0.00000633, -0.00001191, -0.00001586, -0.00001739, -0.00001628, -0.00001293, -0.00000833, -0.00000378, -0.00000051, 0.00000076, 0.00000009, },
		{ -0.00000238, 0.00000002, 0.00000004, -0.00000230, -0.00000596, -0.00000922, -0.00001041, -0.00000857, -0.00000400, 0.00000182, 0.00000675, 0.00000886, 0.00000729, 0.00000267, -0.00000311, -0.00000762, -0.00000893, -0.00000643, -0.00000110, 0.00000487, 0.00000907, 0.00000985, 0.00000703, 0.00000194, -0.00000317, -0.00000616, -0.00000589, -0.00000268, 0.00000188, 0.00000567, 0.00000700, 0.00000533, 0.00000143, -0.00000296, -0.00000594, -0.00000625, -0.00000375, 0.00000063, 0.00000531, 0.00000877, 0.00001012, 0.00000936, 0.00000731, 0.00000514, 0.00000394, 0.00000431, 0.00000623, 0.00000912, 0.00001205, 0.00001398, 0.00001400, 0.00001157, 0.00000663, -0.00000020, -0.00000769, -0.00001413, -0.00001777, -0.00001735, -0.00001266, -0.00000480, 0.00000396, 0.00001085, 0.00001349, 0.00001082, 0.00000359, -0.00000582, -0.00001416, -0.00001848, -0.00001727, -0.00001092, -0.00000163, 0.00000749, 0.00001357, 0.00001504, 0.00001200, 0.00000606, -0.00000053, -0.00000579, -0.00000876, -0.00000962, -0.00000935, -0.00000905, -0.00000929, -0.00000989, -0.00001008, -0.00000906, -0.00000651, -0.00000287, 0.00000079, 0.00000327, 0.00000382, 0.00000251, 0.00000026, -0.00000165, -0.00000212, -0.00000078, 0.00000185, 0.00000460, 0.00000619, 0.00000586, 0.00000366, 0.00000041, -0.00000265, -0.00000443, -0.00000448, -0.00000309, -0.00000114, 0.00000034, 0.00000066, -0.00000027, -0.00000191, -0.00000336, -0.00000375, -0.00000259, 0.00000005, 0.00000358, 0.00000702, 0.00000938, 0.00000989, 0.00000818, 0.00000444, -0.00000065, -0.00000600, -0.00001039, -0.00001279, -0.00001265, -0.00001017, -0.00000626, },
		{ -0.00001084, -0.00001001, -0.00000654, -0.00000142, 0.00000363, 0.00000675, 0.00000660, 0.00000299, -0.00000295, -0.00000910, -0.00001311, -0.00001339, -0.00000978, -0.00000369, 0.00000245, 0.00000616, 0.00000598, 0.00000214, -0.00000353, -0.00000835, -0.00000996, -0.00000735, -0.00000133, 0.00000578, 0.00001117, 0.00001267, 0.00000974, 0.00000359, -0.00000328, -0.00000822, -0.00000952, -0.00000703, -0.00000214, 0.00000292, 0.00000609, 0.00000635, 0.00000404, 0.00000061, -0.00000217, -0.00000296, -0.00000145, 0.00000161, 0.00000478, 0.00000659, 0.00000616, 0.00000346, -0.00000069, -0.00000499, -0.00000808, -0.00000891, -0.00000690, -0.00000211, 0.00000486, 0.00001283, 0.00002020, 0.00002516, 0.00002612, 0.00002219, 0.00001370, 0.00000238, -0.00000891, -0.00001695, -0.00001921, -0.00001486, -0.00000517, 0.00000669, 0.00001670, 0.00002137, 0.00001898, 0.00001023, -0.00000204, -0.00001393, -0.00002189, -0.00002395, -0.00002028, -0.00001286, -0.00000458, 0.00000198, 0.00000543, 0.00000583, 0.00000435, 0.00000249, 0.00000126, 0.00000088, 0.00000078, 0.00000014, -0.00000152, -0.00000405, -0.00000667, -0.00000841, -0.00000862, -0.00000735, -0.00000531, -0.00000355, -0.00000288, -0.00000350, -0.00000485, -0.00000593, -0.00000576, -0.00000395, -0.00000086, 0.00000251, 0.00000495, 0.00000564, 0.00000445, 0.00000200, -0.00000062, -0.00000234, -0.00000259, -0.00000148, 0.00000030, 0.00000184, 0.00000238, 0.00000167, 0.00000000, -0.00000191, -0.00000326, -0.00000346, -0.00000230, -0.00000005, 0.00000266, 0.00000500, 0.00000616, 0.00000558, 0.00000314, -0.00000077, -0.00000522, -0.00000898, },
		{ 0.00000127, -0.00000325, -0.00000857, -0.00001330, -0.00001590, -0.00001518, -0.00001088, -0.00000388, 0.00000387, 0.00000995, 0.00001239, 0.00001041, 0.00000485, -0.00000209, -0.00000764, -0.00000965, -0.00000745, -0.00000220, 0.00000353, 0.00000692, 0.00000607, 0.00000095, -0.00000654, -0.00001334, -0.00001646, -0.00001433, -0.00000746, 0.00000176, 0.00000999, 0.00001437, 0.00001362, 0.00000849, 0.00000135, -0.00000490, -0.00000808, -0.00000751, -0.00000418, -0.00000011, 0.00000265, 0.00000296, 0.00000100, -0.00000188, -0.00000394, -0.00000381, -0.00000112, 0.00000338, 0.00000819, 0.00001166, 0.00001257, 0.00001046, 0.00000563, -0.00000105, -0.00000838, -0.00001507, -0.00001975, -0.00002116, -0.00001835, -0.00001107, -0.00000010, 0.00001253, 0.00002392, 0.00003094, 0.00003137, 0.00002477, 0.00001292, -0.00000055, -0.00001132, -0.00001585, -0.00001272, -0.00000316, 0.00000934, 0.00002033, 0.00002600, 0.00002455, 0.00001667, 0.00000516, -0.00000633, -0.00001469, -0.00001837, -0.00001769, -0.00001435, -0.00001050, -0.00000770, -0.00000646, -0.00000624, -0.00000599, -0.00000488, -0.00000275, -0.00000016, 0.00000198, 0.00000289, 0.00000240, 0.00000098, -0.00000056, -0.00000146, -0.00000152, -0.00000109, -0.00000091, -0.00000166, -0.00000351, -0.00000599, -0.00000820, -0.00000917, -0.00000841, -0.00000610, -0.00000308, -0.00000046, 0.00000086, 0.00000059, -0.00000090, -0.00000271, -0.00000390, -0.00000387, -0.00000256, -0.00000045, 0.00000169, 0.00000315, 0.00000356, 0.00000298, 0.00000184, 0.00000073, 0.00000019, 0.00000050, 0.00000163, 0.00000322, 0.00000464, 0.00000517, 0.00000415, },
		{ 0.00000639, 0.00000851, 0.00000978, 0.00000938, 0.00000665, 0.00000149, -0.00000543, -0.00001265, -0.00001823, -0.00002040, -0.00001824, -0.00001218, -0.00000404, 0.00000356, 0.00000813, 0.00000832, 0.00000454, -0.00000110, -0.00000566, -0.00000663, -0.00000302, 0.00000405, 0.00001174, 0.00001662, 0.00001617, 0.00000989, -0.00000037, -0.00001109, -0.00001851, -0.00002011, -0.00001565, -0.00000717, 0.00000189, 0.00000821, 0.00000986, 0.00000701, 0.00000169, -0.00000333, -0.00000576, -0.00000477, -0.00000123, 0.00000283, 0.00000528, 0.00000485, 0.00000167, -0.00000291, -0.00000696, -0.00000881, -0.00000774, -0.00000405, 0.00000119, 0.00000666, 0.00001124, 0.00001417, 0.00001497, 0.00001327, 0.00000879, 0.00000153, -0.00000786, -0.00001780, -0.00002591, -0.00002957, -0.00002691, -0.00001772, -0.00000387, 0.00001101, 0.00002267, 0.00002774, 0.00002501, 0.00001599, 0.00000436, -0.00000538, -0.00000961, -0.00000696, 0.00000131, 0.00001190, 0.00002085, 0.00002511, 0.00002358, 0.00001734, 0.00000892, 0.00000115, -0.00000403, -0.00000614, -0.00000607, -0.00000539, -0.00000547, -0.00000683, -0.00000906, -0.00001112, -0.00001202, -0.00001132, -0.00000931, -0.00000678, -0.00000456, -0.00000312, -0.00000236, -0.00000177, -0.00000085, 0.00000059, 0.00000220, 0.00000330, 0.00000325, 0.00000179, -0.00000072, -0.00000344, -0.00000542, -0.00000603, -0.00000523, -0.00000359, -0.00000202, -0.00000134, -0.00000191, -0.00000351, -0.00000546, -0.00000698, -0.00000747, -0.00000678, -0.00000518, -0.00000323, -0.00000148, -0.00000027, 0.00000033, 0.00000055, 0.00000074, 0.00000127, 0.00000240, 0.00000418, },
		{ -0.00000467, -0.00000421, -0.00000304, -0.00000093, 0.00000211, 0.00000577, 0.00000931, 0.00001167, 0.00001175, 0.00000887, 0.00000319, -0.00000413, -0.00001109, -0.00001556, -0.00001603, -0.00001240, -0.00000614, 0.00000015, 0.00000373, 0.00000294, -0.00000202, -0.00000894, -0.00001450, -0.00001554, -0.00001059, -0.00000055, 0.00001143, 0.00002119, 0.00002520, 0.00002201, 0.00001290, 0.00000132, -0.00000853, -0.00001341, -0.00001229, -0.00000669, 0.00000021, 0.00000497, 0.00000546, 0.00000168, -0.00000439, -0.00000983, -0.00001211, -0.00001021, -0.00000495, 0.00000150, 0.00000668, 0.00000890, 0.00000783, 0.00000442, 0.00000021, -0.00000339, -0.00000569, -0.00000672, -0.00000681, -0.00000616, -0.00000447, -0.00000123, 0.00000377, 0.00000992, 0.00001551, 0.00001821, 0.00001600, 0.00000823, -0.00000376, -0.00001677, -0.00002677, -0.00003043, -0.00002643, -0.00001615, -0.00000313, 0.00000824, 0.00001444, 0.00001417, 0.00000877, 0.00000149, -0.00000396, -0.00000493, -0.00000091, 0.00000645, 0.00001421, 0.00001946, 0.00002052, 0.00001754, 0.00001219, 0.00000678, 0.00000317, 0.00000206, 0.00000287, 0.00000420, 0.00000463, 0.00000338, 0.00000061, -0.00000283, -0.00000589, -0.00000786, -0.00000863, -0.00000862, -0.00000840, -0.00000835, -0.00000840, -0.00000816, -0.00000717, -0.00000530, -0.00000289, -0.00000060, 0.00000083, 0.00000100, 0.00000003, -0.00000147, -0.00000268, -0.00000301, -0.00000229, -0.00000093, 0.00000038, 0.00000097, 0.00000048, -0.00000098, -0.00000291, -0.00000471, -0.00000589, -0.00000629, -0.00000604, -0.00000547, -0.00000494, -0.00000464, -0.00000461, -0.00000470, },
		{ -0.00000168, -0.00000191, -0.00000250, -0.00000341, -0.00000453, -0.00000567, -0.00000648, -0.00000650, -0.00000533, -0.00000282, 0.00000064, 0.00000411, 0.00000629, 0.00000606, 0.00000301, -0.00000213, -0.00000759, -0.00001112, -0.00001095, -0.00000666, 0.00000038, 0.00000738, 0.00001114, 0.00000947, 0.00000222, -0.00000836, -0.00001838, -0.00002380, -0.00002213, -0.00001351, -0.00000081, 0.00001152, 0.00001921, 0.00001995, 0.00001427, 0.00000527, -0.00000281, -0.00000644, -0.00000429, 0.00000228, 0.00000994, 0.00001492, 0.00001469, 0.00000901, -0.00000009, -0.00000932, -0.00001561, -0.00001734, -0.00001482, -0.00000979, -0.00000445, -0.00000033, 0.00000221, 0.00000383, 0.00000540, 0.00000721, 0.00000857, 0.00000811, 0.00000465, -0.00000183, -0.00000977, -0.00001634, -0.00001858, -0.00001478, -0.00000549, 0.00000645, 0.00001682, 0.00002173, 0.00001922, 0.00001009, -0.00000239, -0.00001381, -0.00002040, -0.00002057, -0.00001530, -0.00000761, -0.00000105, 0.00000187, 0.00000067, -0.00000306, -0.00000659, -0.00000739, -0.00000436, 0.00000176, 0.00000880, 0.00001418, 0.00001612, 0.00001435, 0.00001005, 0.00000530, 0.00000201, 0.00000120, 0.00000264, 0.00000519, 0.00000743, 0.00000832, 0.00000762, 0.00000577, 0.00000357, 0.00000167, 0.00000028, -0.00000082, -0.00000202, -0.00000355, -0.00000526, -0.00000671, -0.00000740, -0.00000707, -0.00000590, -0.00000443, -0.00000333, -0.00000303, -0.00000356, -0.00000448, -0.00000519, -0.00000518, -0.00000430, -0.00000283, -0.00000131, -0.00000027, 0.00000001, -0.00000042, -0.00000121, -0.00000194, -0.00000234, -0.00000233, -0.00000205, -0.00000177, },
		{ -0.00000454, -0.00000391, -0.00000302, -0.00000204, -0.00000106, -0.00000012, 0.00000069, 0.00000122, 0.00000123, 0.00000056, -0.00000069, -0.00000208, -0.00000298, -0.00000283, -0.00000155, 0.00000023, 0.00000132, 0.00000044, -0.00000309, -0.00000876, -0.00001479, -0.00001866, -0.00001813, -0.00001234, -0.00000245, 0.00000846, 0.00001653, 0.00001858, 0.00001358, 0.00000324, -0.00000853, -0.00001713, -0.00001920, -0.00001405, -0.00000400, 0.00000659, 0.00001316, 0.00001296, 0.00000625, -0.00000384, -0.00001267, -0.00001613, -0.00001241, -0.00000269, 0.00000938, 0.00001940, 0.00002400, 0.00002218, 0.00001543, 0.00000676, -0.00000084, -0.00000571, -0.00000800, -0.00000915, -0.00001069, -0.00001306, -0.00001521, -0.00001503, -0.00001068, -0.00000186, 0.00000948, 0.00001961, 0.00002445, 0.00002136, 0.00001053, -0.00000482, -0.00001955, -0.00002858, -0.00002893, -0.00002087, -0.00000774, 0.00000549, 0.00001427, 0.00001623, 0.00001190, 0.00000417, -0.00000323, -0.00000742, -0.00000752, -0.00000475, -0.00000160, -0.00000038, -0.00000209, -0.00000597, -0.00000994, -0.00001174, -0.00001001, -0.00000501, 0.00000156, 0.00000735, 0.00001040, 0.00000997, 0.00000675, 0.00000246, -0.00000102, -0.00000245, -0.00000160, 0.00000079, 0.00000353, 0.00000560, 0.00000654, 0.00000650, 0.00000601, 0.00000553, 0.00000521, 0.00000488, 0.00000418, 0.00000290, 0.00000116, -0.00000063, -0.00000195, -0.00000251, -0.00000235, -0.00000186, -0.00000157, -0.00000185, -0.00000275, -0.00000395, -0.00000497, -0.00000543, -0.00000523, -0.00000458, -0.00000390, -0.00000354, -0.00000365, -0.00000409, -0.00000456, -0.00000477, },
		{ -0.00000207, -0.00000307, -0.00000408, -0.00000487, -0.00000542, -0.00000582, -0.00000618, -0.00000645, -0.00000643, -0.00000585, -0.00000457, -0.00000277, -0.00000087, 0.00000059, 0.00000133, 0.00000149, 0.00000166, 0.00000250, 0.00000428, 0.00000650, 0.00000788, 0.00000683, 0.00000227, -0.00000561, -0.00001503, -0.00002303, -0.00002656, -0.00002384, -0.00001530, -0.00000369, 0.00000683, 0.00001240, 0.00001109, 0.00000388, -0.00000563, -0.00001272, -0.00001359, -0.00000712, 0.00000447, 0.00001651, 0.00002379, 0.00002286, 0.00001353, -0.00000098, -0.00001540, -0.00002460, -0.00002569, -0.00001906, -0.00000794, 0.00000321, 0.00001081, 0.00001356, 0.00001269, 0.00001091, 0.00001064, 0.00001255, 0.00001506, 0.00001522, 0.00001041, 0.00000006, -0.00001349, -0.00002565, -0.00003137, -0.00002745, -0.00001419, 0.00000441, 0.00002200, 0.00003237, 0.00003186, 0.00002084, 0.00000344, -0.00001420, -0.00002635, -0.00002987, -0.00002513, -0.00001543, -0.00000529, 0.00000157, 0.00000362, 0.00000182, -0.00000134, -0.00000325, -0.00000255, 0.00000035, 0.00000367, 0.00000531, 0.00000407, 0.00000023, -0.00000455, -0.00000810, -0.00000879, -0.00000629, -0.00000171, 0.00000298, 0.00000584, 0.00000585, 0.00000324, -0.00000073, -0.00000438, -0.00000643, -0.00000639, -0.00000467, -0.00000217, 0.00000018, 0.00000188, 0.00000289, 0.00000353, 0.00000412, 0.00000475, 0.00000524, 0.00000528, 0.00000468, 0.00000352, 0.00000218, 0.00000109, 0.00000056, 0.00000057, 0.00000084, 0.00000095, 0.00000064, -0.00000010, -0.00000101, -0.00000171, -0.00000195, -0.00000174, -0.00000134, -0.00000113, -0.00000136, },
		{ -0.00000128, -0.00000050, 0.00000014, 0.00000046, 0.00000050, 0.00000049, 0.00000063, 0.00000091, 0.00000109, 0.00000075, -0.00000038, -0.00000226, -0.00000443, -0.00000627, -0.00000723, -0.00000719, -0.00000649, -0.00000573, -0.00000535, -0.00000528, -0.00000487, -0.00000319, 0.00000030, 0.00000525, 0.00001020, 0.00001304, 0.00001187, 0.00000608, -0.00000310, -0.00001279, -0.00001948, -0.00002059, -0.00001570, -0.00000698, 0.00000152, 0.00000566, 0.00000303, -0.00000570, -0.00001684, -0.00002512, -0.00002596, -0.00001756, -0.00000190, 0.00001591, 0.00002965, 0.00003457, 0.00002934, 0.00001653, 0.00000138, -0.00001058, -0.00001586, -0.00001440, -0.00000919, -0.00000455, -0.00000372, -0.00000716, -0.00001227, -0.00001474, -0.00001089, 0.00000015, 0.00001555, 0.00002967, 0.00003639, 0.00003192, 0.00001669, -0.00000448, -0.00002414, -0.00003520, -0.00003373, -0.00002056, -0.00000089, 0.00001794, 0.00002935, 0.00003001, 0.00002083, 0.00000618, -0.00000819, -0.00001765, -0.00002030, -0.00001728, -0.00001176, -0.00000717, -0.00000555, -0.00000691, -0.00000949, -0.00001103, -0.00001000, -0.00000636, -0.00000152, 0.00000248, 0.00000408, 0.00000298, 0.00000021, -0.00000244, -0.00000332, -0.00000176, 0.00000164, 0.00000536, 0.00000767, 0.00000749, 0.00000476, 0.00000043, -0.00000404, -0.00000736, -0.00000885, -0.00000860, -0.00000723, -0.00000547, -0.00000382, -0.00000241, -0.00000110, 0.00000026, 0.00000164, 0.00000283, 0.00000355, 0.00000361, 0.00000310, 0.00000233, 0.00000168, 0.00000135, 0.00000134, 0.00000136, 0.00000113, 0.00000048, -0.00000046, -0.00000138, -0.00000191, -0.00000185, },
		{ -0.00000002, -0.00000135, -0.00000241, -0.00000300, -0.00000323, -0.00000337, -0.00000365, -0.00000406, -0.00000427, -0.00000381, -0.00000236, -0.00000001, 0.00000270, 0.00000495, 0.00000604, 0.00000573, 0.00000436, 0.00000266, 0.00000133, 0.00000064, 0.00000023, -0.00000064, -0.00000257, -0.00000553, -0.00000863, -0.00001042, -0.00000963, -0.00000597, -0.00000051, 0.00000450, 0.00000663, 0.00000445, -0.00000154, -0.00000886, -0.00001397, -0.00001391, -0.00000781, 0.00000231, 0.00001216, 0.00001688, 0.00001327, 0.00000145, -0.00001485, -0.00002958, -0.00003689, -0.00003357, -0.00002058, -0.00000273, 0.00001341, 0.00002227, 0.00002173, 0.00001389, 0.00000398, -0.00000228, -0.00000144, 0.00000587, 0.00001524, 0.00002056, 0.00001707, 0.00000392, -0.00001503, -0.00003257, -0.00004118, -0.00003640, -0.00001891, 0.00000543, 0.00002790, 0.00004040, 0.00003874, 0.00002424, 0.00000307, -0.00001645, -0.00002721, -0.00002608, -0.00001480, 0.00000112, 0.00001501, 0.00002184, 0.00002003, 0.00001168, 0.00000111, -0.00000725, -0.00001078, -0.00000954, -0.00000577, -0.00000247, -0.00000187, -0.00000441, -0.00000878, -0.00001278, -0.00001456, -0.00001347, -0.00001030, -0.00000673, -0.00000435, -0.00000386, -0.00000478, -0.00000579, -0.00000545, -0.00000303, 0.00000111, 0.00000574, 0.00000928, 0.00001054, 0.00000919, 0.00000584, 0.00000164, -0.00000221, -0.00000493, -0.00000633, -0.00000670, -0.00000649, -0.00000602, -0.00000540, -0.00000457, -0.00000348, -0.00000223, -0.00000107, -0.00000024, 0.00000012, 0.00000015, 0.00000012, 0.00000031, 0.00000083, 0.00000151, 0.00000199, 0.00000193, 0.00000121, },
		{ -0.00000136, -0.00000008, 0.00000075, 0.00000094, 0.00000063, 0.00000016, -0.00000018, -0.00000037, -0.00000071, -0.00000162, -0.00000338, -0.00000582, -0.00000829, -0.00000991, -0.00000996, -0.00000827, -0.00000535, -0.00000219, 0.00000028, 0.00000162, 0.00000209, 0.00000251, 0.00000369, 0.00000591, 0.00000859, 0.00001052, 0.00001048, 0.00000799, 0.00000369, -0.00000077, -0.00000354, -0.00000358, -0.00000134, 0.00000122, 0.00000156, -0.00000213, -0.00000969, -0.00001866, -0.00002509, -0.00002526, -0.00001757, -0.00000366, 0.00001183, 0.00002303, 0.00002528, 0.00001738, 0.00000227, -0.00001396, -0.00002470, -0.00002569, -0.00001708, -0.00000339, 0.00000845, 0.00001234, 0.00000597, -0.00000789, -0.00002239, -0.00002964, -0.00002437, -0.00000663, 0.00001781, 0.00003963, 0.00004979, 0.00004342, 0.00002201, -0.00000700, -0.00003318, -0.00004728, -0.00004490, -0.00002807, -0.00000413, 0.00001746, 0.00002904, 0.00002763, 0.00001575, -0.00000019, -0.00001293, -0.00001745, -0.00001286, -0.00000230, 0.00000891, 0.00001577, 0.00001580, 0.00000980, 0.00000108, -0.00000633, -0.00000958, -0.00000808, -0.00000354, 0.00000122, 0.00000368, 0.00000277, -0.00000084, -0.00000532, -0.00000874, -0.00001005, -0.00000944, -0.00000807, -0.00000730, -0.00000788, -0.00000958, -0.00001134, -0.00001184, -0.00001026, -0.00000665, -0.00000195, 0.00000246, 0.00000539, 0.00000627, 0.00000530, 0.00000324, 0.00000097, -0.00000084, -0.00000196, -0.00000247, -0.00000262, -0.00000259, -0.00000241, -0.00000207, -0.00000160, -0.00000118, -0.00000104, -0.00000136, -0.00000211, -0.00000298, -0.00000354, -0.00000346, -0.00000264, },
		{ 0.00000008, -0.00000188, -0.00000316, -0.00000357, -0.00000331, -0.00000283, -0.00000255, -0.00000256, -0.00000263, -0.00000233, -0.00000139, 0.00000007, 0.00000144, 0.00000192, 0.00000091, -0.00000159, -0.00000489, -0.00000788, -0.00000956, -0.00000951, -0.00000818, -0.00000660, -0.00000585, -0.00000642, -0.00000785, -0.00000893, -0.00000833, -0.00000533, -0.00000037, 0.00000511, 0.00000929, 0.00001096, 0.00001021, 0.00000841, 0.00000754, 0.00000896, 0.00001247, 0.00001607, 0.00001675, 0.00001203, 0.00000146, -0.00001261, -0.00002563, -0.00003267, -0.00003067, -0.00002006, -0.00000499, 0.00000829, 0.00001412, 0.00001013, -0.00000144, -0.00001448, -0.00002168, -0.00001799, -0.00000328, 0.00001707, 0.00003417, 0.00003935, 0.00002818, 0.00000281, -0.00002833, -0.00005367, -0.00006314, -0.00005233, -0.00002449, 0.00001066, 0.00004071, 0.00005544, 0.00005075, 0.00002993, 0.00000203, -0.00002213, -0.00003430, -0.00003184, -0.00001819, -0.00000082, 0.00001235, 0.00001633, 0.00001084, -0.00000009, -0.00001044, -0.00001511, -0.00001211, -0.00000320, 0.00000728, 0.00001460, 0.00001583, 0.00001092, 0.00000249, -0.00000567, -0.00001036, -0.00001033, -0.00000653, -0.00000138, 0.00000255, 0.00000376, 0.00000229, -0.00000057, -0.00000315, -0.00000429, -0.00000392, -0.00000287, -0.00000236, -0.00000319, -0.00000538, -0.00000813, -0.00001026, -0.00001086, -0.00000970, -0.00000727, -0.00000456, -0.00000253, -0.00000171, -0.00000207, -0.00000314, -0.00000431, -0.00000510, -0.00000529, -0.00000490, -0.00000404, -0.00000278, -0.00000119, 0.00000064, 0.00000244, 0.00000380, 0.00000431, 0.00000373, 0.00000216, },
		{ 0.00000156, 0.00000430, 0.00000563, 0.00000553, 0.00000444, 0.00000307, 0.00000201, 0.00000150, 0.00000128, 0.00000088, -0.00000012, -0.00000168, -0.00000333, -0.00000432, -0.00000410, -0.00000269, -0.00000075, 0.00000068, 0.00000071, -0.00000087, -0.00000343, -0.00000576, -0.00000669, -0.00000573, -0.00000344, -0.00000122, -0.00000061, -0.00000246, -0.00000642, -0.00001098, -0.00001421, -0.00001468, -0.00001226, -0.00000820, -0.00000445, -0.00000257, -0.00000278, -0.00000368, -0.00000294, 0.00000134, 0.00000936, 0.00001902, 0.00002657, 0.00002828, 0.00002234, 0.00001013, -0.00000399, -0.00001441, -0.00001686, -0.00001064, 0.00000066, 0.00001048, 0.00001225, 0.00000263, -0.00001630, -0.00003732, -0.00005076, -0.00004866, -0.00002860, 0.00000455, 0.00003992, 0.00006462, 0.00006887, 0.00005008, 0.00001413, -0.00002666, -0.00005831, -0.00007048, -0.00006033, -0.00003328, -0.00000044, 0.00002602, 0.00003768, 0.00003280, 0.00001627, -0.00000308, -0.00001662, -0.00001950, -0.00001223, 0.00000016, 0.00001092, 0.00001482, 0.00001041, 0.00000032, -0.00001034, -0.00001650, -0.00001551, -0.00000814, 0.00000206, 0.00001055, 0.00001392, 0.00001130, 0.00000443, -0.00000339, -0.00000892, -0.00001041, -0.00000813, -0.00000394, -0.00000015, 0.00000159, 0.00000097, -0.00000111, -0.00000317, -0.00000403, -0.00000335, -0.00000170, -0.00000015, 0.00000043, -0.00000025, -0.00000180, -0.00000338, -0.00000424, -0.00000409, -0.00000321, -0.00000222, -0.00000181, -0.00000237, -0.00000387, -0.00000597, -0.00000821, -0.00001019, -0.00001165, -0.00001244, -0.00001237, -0.00001129, -0.00000907, -0.00000586, -0.00000208, },
		{ -0.00001615, -0.00001607, -0.00001426, -0.00001109, -0.00000738, -0.00000407, -0.00000185, -0.00000087, -0.00000069, -0.00000056, 0.00000015, 0.00000160, 0.00000333, 0.00000457, 0.00000468, 0.00000352, 0.00000166, 0.00000003, -0.00000055, 0.00000012, 0.00000143, 0.00000223, 0.00000147, -0.00000115, -0.00000489, -0.00000819, -0.00000944, -0.00000783, -0.00000386, 0.00000078, 0.00000395, 0.00000411, 0.00000113, -0.00000362, -0.00000793, -0.00000994, -0.00000913, -0.00000666, -0.00000477, -0.00000546, -0.00000929, -0.00001473, -0.00001877, -0.00001836, -0.00001218, -0.00000164, 0.00000940, 0.00001631, 0.00001600, 0.00000877, -0.00000136, -0.00000826, -0.00000654, 0.00000550, 0.00002432, 0.00004198, 0.00004924, 0.00003968, 0.00001312, -0.00002343, -0.00005785, -0.00007740, -0.00007390, -0.00004727, -0.00000601, 0.00003582, 0.00006394, 0.00006926, 0.00005126, 0.00001788, -0.00001793, -0.00004344, -0.00005110, -0.00004094, -0.00001985, 0.00000193, 0.00001537, 0.00001629, 0.00000654, -0.00000752, -0.00001832, -0.00002065, -0.00001376, -0.00000131, 0.00001072, 0.00001698, 0.00001513, 0.00000669, -0.00000397, -0.00001187, -0.00001372, -0.00000925, -0.00000107, 0.00000682, 0.00001091, 0.00000974, 0.00000428, -0.00000276, -0.00000833, -0.00001049, -0.00000906, -0.00000542, -0.00000169, 0.00000043, 0.00000032, -0.00000147, -0.00000369, -0.00000513, -0.00000521, -0.00000415, -0.00000266, -0.00000150, -0.00000104, -0.00000113, -0.00000127, -0.00000098, -0.00000007, 0.00000125, 0.00000250, 0.00000317, 0.00000295, 0.00000179, -0.00000016, -0.00000271, -0.00000568, -0.00000889, -0.00001203, -0.00001465, },
		{ 0.00000172, -0.00000078, -0.00000412, -0.00000774, -0.00001083, -0.00001258, -0.00001258, -0.00001107, -0.00000889, -0.00000709, -0.00000647, -0.00000716, -0.00000855, -0.00000962, -0.00000946, -0.00000777, -0.00000502, -0.00000223, -0.00000036, 0.00000022, -0.00000004, -0.00000015, 0.00000084, 0.00000321, 0.00000622, 0.00000842, 0.00000835, 0.00000543, 0.00000035, -0.00000500, -0.00000842, -0.00000848, -0.00000524, -0.00000037, 0.00000364, 0.00000474, 0.00000237, -0.00000216, -0.00000636, -0.00000783, -0.00000563, -0.00000095, 0.00000344, 0.00000456, 0.00000091, -0.00000649, -0.00001436, -0.00001874, -0.00001707, -0.00000986, -0.00000085, 0.00000454, 0.00000195, -0.00000935, -0.00002532, -0.00003826, -0.00003998, -0.00002564, 0.00000331, 0.00003881, 0.00006874, 0.00008152, 0.00007093, 0.00003910, -0.00000391, -0.00004366, -0.00006687, -0.00006651, -0.00004443, -0.00001039, 0.00002208, 0.00004116, 0.00004124, 0.00002470, 0.00000028, -0.00002111, -0.00003107, -0.00002705, -0.00001284, 0.00000368, 0.00001439, 0.00001459, 0.00000479, -0.00000994, -0.00002265, -0.00002777, -0.00002351, -0.00001236, 0.00000039, 0.00000932, 0.00001124, 0.00000651, -0.00000157, -0.00000850, -0.00001080, -0.00000752, -0.00000045, 0.00000699, 0.00001142, 0.00001105, 0.00000632, -0.00000055, -0.00000671, -0.00001003, -0.00000991, -0.00000726, -0.00000386, -0.00000141, -0.00000077, -0.00000178, -0.00000355, -0.00000505, -0.00000564, -0.00000530, -0.00000446, -0.00000363, -0.00000313, -0.00000290, -0.00000264, -0.00000204, -0.00000100, 0.00000035, 0.00000173, 0.00000288, 0.00000365, 0.00000403, 0.00000394, 0.00000325, },
		{ 0.00000056, 0.00000111, 0.00000192, 0.00000267, 0.00000281, 0.00000183, -0.00000035, -0.00000326, -0.00000594, -0.00000737, -0.00000696, -0.00000493, -0.00000226, -0.00000032, -0.00000021, -0.00000222, -0.00000566, -0.00000923, -0.00001162, -0.00001219, -0.00001124, -0.00000981, -0.00000900, -0.00000931, -0.00001027, -0.00001065, -0.00000918, -0.00000537, 0.00000002, 0.00000522, 0.00000817, 0.00000768, 0.00000410, -0.00000068, -0.00000408, -0.00000412, -0.00000048, 0.00000520, 0.00001004, 0.00001136, 0.00000810, 0.00000148, -0.00000548, -0.00000946, -0.00000859, -0.00000355, 0.00000266, 0.00000622, 0.00000455, -0.00000206, -0.00001025, -0.00001509, -0.00001263, -0.00000234, 0.00001191, 0.00002304, 0.00002374, 0.00001012, -0.00001577, -0.00004592, -0.00006903, -0.00007493, -0.00005904, -0.00002473, 0.00001752, 0.00005395, 0.00007279, 0.00006890, 0.00004564, 0.00001329, -0.00001532, -0.00003004, -0.00002738, -0.00001138, 0.00000865, 0.00002254, 0.00002379, 0.00001215, -0.00000665, -0.00002387, -0.00003186, -0.00002743, -0.00001308, 0.00000446, 0.00001745, 0.00002058, 0.00001319, -0.00000085, -0.00001515, -0.00002386, -0.00002415, -0.00001719, -0.00000719, 0.00000084, 0.00000344, 0.00000015, -0.00000655, -0.00001274, -0.00001497, -0.00001190, -0.00000470, 0.00000362, 0.00000978, 0.00001164, 0.00000906, 0.00000366, -0.00000209, -0.00000603, -0.00000723, -0.00000610, -0.00000394, -0.00000219, -0.00000174, -0.00000262, -0.00000422, -0.00000574, -0.00000657, -0.00000659, -0.00000602, -0.00000520, -0.00000438, -0.00000360, -0.00000281, -0.00000196, -0.00000112, -0.00000043, 0.00000000, 0.00000026, },
		{ -0.00000077, 0.00000015, 0.00000028, -0.00000013, -0.00000056, -0.00000049, 0.00000025, 0.00000132, 0.00000194, 0.00000134, -0.00000079, -0.00000397, -0.00000702, -0.00000857, -0.00000769, -0.00000444, 0.00000007, 0.00000411, 0.00000609, 0.00000533, 0.00000234, -0.00000148, -0.00000457, -0.00000602, -0.00000604, -0.00000575, -0.00000650, -0.00000897, -0.00001266, -0.00001604, -0.00001730, -0.00001540, -0.00001081, -0.00000538, -0.00000159, -0.00000120, -0.00000424, -0.00000883, -0.00001195, -0.00001097, -0.00000505, 0.00000419, 0.00001332, 0.00001865, 0.00001813, 0.00001243, 0.00000474, -0.00000081, -0.00000131, 0.00000323, 0.00000965, 0.00001311, 0.00000960, -0.00000157, -0.00001683, -0.00002940, -0.00003228, -0.00002169, 0.00000054, 0.00002697, 0.00004712, 0.00005172, 0.00003685, 0.00000599, -0.00003097, -0.00006139, -0.00007495, -0.00006768, -0.00004342, -0.00001205, 0.00001485, 0.00002873, 0.00002738, 0.00001530, 0.00000113, -0.00000664, -0.00000359, 0.00000878, 0.00002405, 0.00003408, 0.00003303, 0.00002005, -0.00000026, -0.00002005, -0.00003173, -0.00003131, -0.00001999, -0.00000338, 0.00001115, 0.00001782, 0.00001472, 0.00000434, -0.00000797, -0.00001667, -0.00001849, -0.00001371, -0.00000565, 0.00000118, 0.00000327, -0.00000029, -0.00000765, -0.00001524, -0.00001959, -0.00001886, -0.00001352, -0.00000594, 0.00000089, 0.00000470, 0.00000482, 0.00000221, -0.00000125, -0.00000375, -0.00000433, -0.00000315, -0.00000121, 0.00000032, 0.00000064, -0.00000037, -0.00000230, -0.00000446, -0.00000631, -0.00000752, -0.00000805, -0.00000794, -0.00000723, -0.00000594, -0.00000422, -0.00000236, },
		{ -0.00000822, -0.00000778, -0.00000630, -0.00000424, -0.00000231, -0.00000115, -0.00000105, -0.00000168, -0.00000229, -0.00000210, -0.00000078, 0.00000124, 0.00000293, 0.00000307, 0.00000101, -0.00000290, -0.00000730, -0.00001036, -0.00001062, -0.00000774, -0.00000273, 0.00000248, 0.00000591, 0.00000647, 0.00000445, 0.00000133, -0.00000108, -0.00000162, -0.00000041, 0.00000122, 0.00000148, -0.00000076, -0.00000525, -0.00001030, -0.00001366, -0.00001369, -0.00001045, -0.00000581, -0.00000267, -0.00000346, -0.00000871, -0.00001655, -0.00002336, -0.00002550, -0.00002115, -0.00001142, 0.00000000, 0.00000854, 0.00001102, 0.00000740, 0.00000104, -0.00000287, 0.00000000, 0.00001067, 0.00002567, 0.00003823, 0.00004114, 0.00003035, 0.00000731, -0.00002086, -0.00004389, -0.00005254, -0.00004260, -0.00001694, 0.00001538, 0.00004263, 0.00005509, 0.00004896, 0.00002758, -0.00000020, -0.00002408, -0.00003656, -0.00003588, -0.00002610, -0.00001458, -0.00000823, -0.00001007, -0.00001807, -0.00002637, -0.00002847, -0.00002065, -0.00000403, 0.00001580, 0.00003125, 0.00003622, 0.00002889, 0.00001254, -0.00000594, -0.00001916, -0.00002237, -0.00001542, -0.00000249, 0.00001004, 0.00001654, 0.00001457, 0.00000569, -0.00000558, -0.00001409, -0.00001638, -0.00001217, -0.00000416, 0.00000347, 0.00000703, 0.00000502, -0.00000146, -0.00000945, -0.00001567, -0.00001795, -0.00001611, -0.00001171, -0.00000711, -0.00000430, -0.00000402, -0.00000571, -0.00000791, -0.00000915, -0.00000861, -0.00000640, -0.00000331, -0.00000041, 0.00000151, 0.00000211, 0.00000150, 0.00000003, -0.00000194, -0.00000407, -0.00000606, -0.00000757, },
		{ -0.00000050, -0.00000118, -0.00000256, -0.00000410, -0.00000516, -0.00000517, -0.00000404, -0.00000225, -0.00000070, -0.00000027, -0.00000130, -0.00000334, -0.00000531, -0.00000597, -0.00000467, -0.00000174, 0.00000149, 0.00000330, 0.00000247, -0.00000103, -0.00000588, -0.00000995, -0.00001128, -0.00000906, -0.00000407, 0.00000161, 0.00000564, 0.00000653, 0.00000436, 0.00000074, -0.00000213, -0.00000262, -0.00000061, 0.00000245, 0.00000433, 0.00000330, -0.00000078, -0.00000620, -0.00001025, -0.00001061, -0.00000677, -0.00000056, 0.00000458, 0.00000528, 0.00000009, -0.00000953, -0.00001961, -0.00002551, -0.00002425, -0.00001624, -0.00000542, 0.00000245, 0.00000260, -0.00000610, -0.00002001, -0.00003198, -0.00003432, -0.00002255, 0.00000208, 0.00003241, 0.00005789, 0.00006879, 0.00006031, 0.00003476, 0.00000088, -0.00002968, -0.00004688, -0.00004620, -0.00003005, -0.00000623, 0.00001564, 0.00002829, 0.00002947, 0.00002217, 0.00001241, 0.00000590, 0.00000512, 0.00000838, 0.00001106, 0.00000838, -0.00000175, -0.00001734, -0.00003297, -0.00004234, -0.00004127, -0.00002987, -0.00001259, 0.00000366, 0.00001263, 0.00001146, 0.00000189, -0.00001071, -0.00001984, -0.00002092, -0.00001320, 0.00000002, 0.00001308, 0.00002061, 0.00001988, 0.00001182, 0.00000041, -0.00000938, -0.00001375, -0.00001168, -0.00000504, 0.00000256, 0.00000755, 0.00000798, 0.00000416, -0.00000177, -0.00000711, -0.00000989, -0.00000969, -0.00000755, -0.00000531, -0.00000456, -0.00000595, -0.00000899, -0.00001245, -0.00001500, -0.00001582, -0.00001478, -0.00001239, -0.00000936, -0.00000635, -0.00000378, -0.00000185, -0.00000073, },
		{ -0.00001559, -0.00001429, -0.00001246, -0.00001044, -0.00000876, -0.00000780, -0.00000759, -0.00000768, -0.00000725, -0.00000562, -0.00000264, 0.00000103, 0.00000412, 0.00000533, 0.00000402, 0.00000064, -0.00000335, -0.00000612, -0.00000640, -0.00000415, -0.00000070, 0.00000189, 0.00000182, -0.00000141, -0.00000664, -0.00001152, -0.00001362, -0.00001163, -0.00000613, 0.00000069, 0.00000598, 0.00000771, 0.00000563, 0.00000139, -0.00000235, -0.00000334, -0.00000104, 0.00000312, 0.00000642, 0.00000639, 0.00000230, -0.00000430, -0.00001019, -0.00001217, -0.00000884, -0.00000166, 0.00000555, 0.00000845, 0.00000446, -0.00000551, -0.00001708, -0.00002422, -0.00002219, -0.00001019, 0.00000757, 0.00002323, 0.00002852, 0.00001864, -0.00000512, -0.00003522, -0.00006051, -0.00007061, -0.00006023, -0.00003153, 0.00000660, 0.00004193, 0.00006365, 0.00006638, 0.00005181, 0.00002740, 0.00000287, -0.00001388, -0.00001956, -0.00001599, -0.00000822, -0.00000151, 0.00000135, 0.00000106, 0.00000080, 0.00000396, 0.00001177, 0.00002223, 0.00003090, 0.00003304, 0.00002619, 0.00001159, -0.00000608, -0.00002072, -0.00002760, -0.00002548, -0.00001715, -0.00000804, -0.00000363, -0.00000680, -0.00001652, -0.00002836, -0.00003658, -0.00003681, -0.00002805, -0.00001309, 0.00000278, 0.00001414, 0.00001771, 0.00001358, 0.00000484, -0.00000403, -0.00000922, -0.00000911, -0.00000468, 0.00000131, 0.00000584, 0.00000706, 0.00000494, 0.00000106, -0.00000245, -0.00000396, -0.00000307, -0.00000067, 0.00000169, 0.00000259, 0.00000137, -0.00000171, -0.00000574, -0.00000971, -0.00001287, -0.00001496, -0.00001601, -0.00001619, },
		{ 0.00000025, -0.00000231, -0.00000526, -0.00000840, -0.00001128, -0.00001342, -0.00001463, -0.00001515, -0.00001557, -0.00001643, -0.00001781, -0.00001910, -0.00001918, -0.00001700, -0.00001225, -0.00000571, 0.00000085, 0.00000542, 0.00000666, 0.00000464, 0.00000084, -0.00000251, -0.00000351, -0.00000165, 0.00000193, 0.00000492, 0.00000504, 0.00000132, -0.00000525, -0.00001205, -0.00001605, -0.00001528, -0.00000993, -0.00000228, 0.00000429, 0.00000700, 0.00000506, 0.00000005, -0.00000488, -0.00000670, -0.00000415, 0.00000153, 0.00000720, 0.00000952, 0.00000682, 0.00000027, -0.00000661, -0.00000967, -0.00000650, 0.00000190, 0.00001117, 0.00001543, 0.00001024, -0.00000473, -0.00002473, -0.00004144, -0.00004637, -0.00003477, -0.00000837, 0.00002464, 0.00005237, 0.00006388, 0.00005364, 0.00002398, -0.00001564, -0.00005223, -0.00007417, -0.00007550, -0.00005769, -0.00002848, 0.00000175, 0.00002412, 0.00003435, 0.00003350, 0.00002631, 0.00001832, 0.00001317, 0.00001131, 0.00001066, 0.00000856, 0.00000378, -0.00000246, -0.00000702, -0.00000666, -0.00000001, 0.00001131, 0.00002316, 0.00003070, 0.00003070, 0.00002313, 0.00001119, -0.00000012, -0.00000649, -0.00000630, -0.00000141, 0.00000380, 0.00000461, -0.00000168, -0.00001431, -0.00002932, -0.00004127, -0.00004567, -0.00004091, -0.00002896, -0.00001432, -0.00000210, 0.00000418, 0.00000385, -0.00000101, -0.00000680, -0.00001024, -0.00000972, -0.00000581, -0.00000067, 0.00000323, 0.00000432, 0.00000257, -0.00000070, -0.00000363, -0.00000476, -0.00000366, -0.00000095, 0.00000218, 0.00000459, 0.00000565, 0.00000539, 0.00000418, 0.00000241, },
		{ -0.00000220, -0.00000052, 0.00000152, 0.00000379, 0.00000576, 0.00000676, 0.00000637, 0.00000472, 0.00000240, 0.00000013, -0.00000172, -0.00000342, -0.00000573, -0.00000945, -0.00001467, -0.00002053, -0.00002529, -0.00002711, -0.00002495, -0.00001924, -0.00001187, -0.00000540, -0.00000189, -0.00000192, -0.00000425, -0.00000646, -0.00000625, -0.00000267, 0.00000319, 0.00000867, 0.00001082, 0.00000800, 0.00000087, -0.00000772, -0.00001399, -0.00001506, -0.00001042, -0.00000227, 0.00000550, 0.00000920, 0.00000721, 0.00000078, -0.00000654, -0.00001082, -0.00000985, -0.00000437, 0.00000221, 0.00000565, 0.00000329, -0.00000416, -0.00001256, -0.00001619, -0.00001074, 0.00000401, 0.00002324, 0.00003855, 0.00004153, 0.00002767, -0.00000092, -0.00003555, -0.00006392, -0.00007486, -0.00006296, -0.00003096, 0.00001101, 0.00004924, 0.00007158, 0.00007185, 0.00005176, 0.00001963, -0.00001331, -0.00003735, -0.00004757, -0.00004479, -0.00003402, -0.00002147, -0.00001170, -0.00000607, -0.00000310, -0.00000020, 0.00000437, 0.00001030, 0.00001554, 0.00001765, 0.00001550, 0.00001018, 0.00000468, 0.00000239, 0.00000529, 0.00001274, 0.00002168, 0.00002799, 0.00002856, 0.00002285, 0.00001324, 0.00000387, -0.00000136, -0.00000075, 0.00000434, 0.00001014, 0.00001223, 0.00000777, -0.00000312, -0.00001732, -0.00003009, -0.00003716, -0.00003656, -0.00002941, -0.00001920, -0.00001017, -0.00000540, -0.00000569, -0.00000944, -0.00001373, -0.00001583, -0.00001443, -0.00001005, -0.00000462, -0.00000033, 0.00000138, 0.00000038, -0.00000231, -0.00000518, -0.00000702, -0.00000739, -0.00000657, -0.00000517, -0.00000368, },
		{ -0.00000409, -0.00000527, -0.00000704, -0.00000921, -0.00001107, -0.00001173, -0.00001064, -0.00000795, -0.00000449, -0.00000134, 0.00000076, 0.00000183, 0.00000256, 0.00000382, 0.00000598, 0.00000844, 0.00000980, 0.00000850, 0.00000378, -0.00000372, -0.00001198, -0.00001839, -0.00002098, -0.00001939, -0.00001516, -0.00001096, -0.00000927, -0.00001106, -0.00001519, -0.00001888, -0.00001922, -0.00001470, -0.00000625, 0.00000301, 0.00000917, 0.00000947, 0.00000380, -0.00000507, -0.00001272, -0.00001516, -0.00001089, -0.00000166, 0.00000827, 0.00001427, 0.00001367, 0.00000719, -0.00000137, -0.00000716, -0.00000687, -0.00000077, 0.00000712, 0.00001095, 0.00000608, -0.00000803, -0.00002674, -0.00004172, -0.00004447, -0.00003041, -0.00000165, 0.00003303, 0.00006120, 0.00007161, 0.00005890, 0.00002606, -0.00001647, -0.00005462, -0.00007604, -0.00007462, -0.00005240, -0.00001824, 0.00001604, 0.00004026, 0.00004931, 0.00004420, 0.00003044, 0.00001499, 0.00000311, -0.00000332, -0.00000565, -0.00000677, -0.00000901, -0.00001285, -0.00001682, -0.00001869, -0.00001702, -0.00001214, -0.00000602, -0.00000118, 0.00000082, 0.00000030, -0.00000061, 0.00000084, 0.00000639, 0.00001564, 0.00002598, 0.00003368, 0.00003569, 0.00003114, 0.00002185, 0.00001153, 0.00000399, 0.00000143, 0.00000337, 0.00000705, 0.00000880, 0.00000596, -0.00000181, -0.00001235, -0.00002199, -0.00002730, -0.00002671, -0.00002111, -0.00001342, -0.00000713, -0.00000474, -0.00000672, -0.00001153, -0.00001651, -0.00001919, -0.00001840, -0.00001460, -0.00000941, -0.00000473, -0.00000181, -0.00000089, -0.00000135, -0.00000231, -0.00000323, },
		{ -0.00000715, -0.00000500, -0.00000233, 0.00000063, 0.00000298, 0.00000366, 0.00000211, -0.00000136, -0.00000557, -0.00000905, -0.00001081, -0.00001086, -0.00001008, -0.00000973, -0.00001055, -0.00001222, -0.00001343, -0.00001257, -0.00000874, -0.00000242, 0.00000450, 0.00000950, 0.00001069, 0.00000781, 0.00000248, -0.00000255, -0.00000482, -0.00000348, 0.00000016, 0.00000319, 0.00000265, -0.00000281, -0.00001199, -0.00002147, -0.00002714, -0.00002629, -0.00001905, -0.00000860, 0.00000031, 0.00000357, -0.00000015, -0.00000864, -0.00001716, -0.00002077, -0.00001685, -0.00000661, 0.00000539, 0.00001348, 0.00001383, 0.00000662, -0.00000372, -0.00001054, -0.00000838, 0.00000389, 0.00002187, 0.00003700, 0.00004030, 0.00002666, -0.00000216, -0.00003744, -0.00006652, -0.00007778, -0.00006552, -0.00003252, 0.00001070, 0.00004986, 0.00007237, 0.00007190, 0.00005046, 0.00001703, -0.00001636, -0.00003937, -0.00004691, -0.00004014, -0.00002491, -0.00000857, 0.00000335, 0.00000889, 0.00000957, 0.00000871, 0.00000919, 0.00001193, 0.00001571, 0.00001819, 0.00001748, 0.00001330, 0.00000705, 0.00000099, -0.00000310, -0.00000486, -0.00000539, -0.00000643, -0.00000908, -0.00001289, -0.00001588, -0.00001547, -0.00000993, 0.00000054, 0.00001350, 0.00002526, 0.00003235, 0.00003310, 0.00002824, 0.00002047, 0.00001308, 0.00000837, 0.00000666, 0.00000635, 0.00000501, 0.00000081, -0.00000635, -0.00001468, -0.00002130, -0.00002370, -0.00002096, -0.00001432, -0.00000658, -0.00000088, 0.00000069, -0.00000211, -0.00000772, -0.00001366, -0.00001774, -0.00001889, -0.00001743, -0.00001457, -0.00001160, -0.00000916, },
		{ -0.00001209, -0.00001303, -0.00001451, -0.00001629, -0.00001735, -0.00001647, -0.00001309, -0.00000776, -0.00000203, 0.00000231, 0.00000414, 0.00000358, 0.00000189, 0.00000074, 0.00000116, 0.00000289, 0.00000443, 0.00000385, -0.00000007, -0.00000697, -0.00001486, -0.00002098, -0.00002311, -0.00002075, -0.00001540, -0.00000981, -0.00000653, -0.00000651, -0.00000855, -0.00000985, -0.00000763, -0.00000081, 0.00000907, 0.00001829, 0.00002260, 0.00001942, 0.00000926, -0.00000418, -0.00001565, -0.00002076, -0.00001818, -0.00001033, -0.00000225, 0.00000096, -0.00000314, -0.00001294, -0.00002334, -0.00002831, -0.00002402, -0.00001122, 0.00000469, 0.00001597, 0.00001628, 0.00000413, -0.00001574, -0.00003393, -0.00004035, -0.00002895, -0.00000097, 0.00003478, 0.00006502, 0.00007734, 0.00006529, 0.00003132, -0.00001397, -0.00005577, -0.00008077, -0.00008200, -0.00006107, -0.00002690, 0.00000825, 0.00003362, 0.00004370, 0.00003936, 0.00002632, 0.00001189, 0.00000167, -0.00000237, -0.00000184, -0.00000019, -0.00000046, -0.00000363, -0.00000835, -0.00001202, -0.00001232, -0.00000857, -0.00000194, 0.00000526, 0.00001086, 0.00001388, 0.00001481, 0.00001495, 0.00001540, 0.00001625, 0.00001640, 0.00001432, 0.00000913, 0.00000143, -0.00000661, -0.00001204, -0.00001248, -0.00000730, 0.00000202, 0.00001240, 0.00002055, 0.00002427, 0.00002327, 0.00001899, 0.00001367, 0.00000904, 0.00000558, 0.00000244, -0.00000178, -0.00000791, -0.00001545, -0.00002253, -0.00002669, -0.00002611, -0.00002058, -0.00001174, -0.00000251, 0.00000420, 0.00000663, 0.00000479, 0.00000024, -0.00000486, -0.00000880, -0.00001103, },
		{ 0.00000272, 0.00000139, 0.00000139, 0.00000239, 0.00000297, 0.00000155, -0.00000253, -0.00000853, -0.00001451, -0.00001836, -0.00001889, -0.00001645, -0.00001275, -0.00000987, -0.00000908, -0.00001006, -0.00001100, -0.00000963, -0.00000456, 0.00000374, 0.00001287, 0.00001953, 0.00002108, 0.00001683, 0.00000848, -0.00000077, -0.00000775, -0.00001098, -0.00001129, -0.00001128, -0.00001369, -0.00001960, -0.00002749, -0.00003371, -0.00003422, -0.00002680, -0.00001253, 0.00000429, 0.00001789, 0.00002364, 0.00002032, 0.00001073, 0.00000039, -0.00000521, -0.00000349, 0.00000381, 0.00001138, 0.00001310, 0.00000542, -0.00001029, -0.00002779, -0.00003856, -0.00003590, -0.00001863, 0.00000754, 0.00003184, 0.00004282, 0.00003355, 0.00000515, -0.00003291, -0.00006617, -0.00008090, -0.00006970, -0.00003473, 0.00001297, 0.00005765, 0.00008489, 0.00008681, 0.00006465, 0.00002755, -0.00001151, -0.00004082, -0.00005398, -0.00005127, -0.00003820, -0.00002226, -0.00000947, -0.00000228, 0.00000047, 0.00000194, 0.00000500, 0.00001053, 0.00001712, 0.00002206, 0.00002296, 0.00001910, 0.00001177, 0.00000353, -0.00000304, -0.00000653, -0.00000708, -0.00000592, -0.00000439, -0.00000310, -0.00000160, 0.00000107, 0.00000557, 0.00001152, 0.00001739, 0.00002112, 0.00002121, 0.00001755, 0.00001171, 0.00000628, 0.00000380, 0.00000554, 0.00001096, 0.00001801, 0.00002410, 0.00002717, 0.00002651, 0.00002271, 0.00001706, 0.00001067, 0.00000392, -0.00000357, -0.00001230, -0.00002209, -0.00003167, -0.00003888, -0.00004151, -0.00003838, -0.00003007, -0.00001885, -0.00000785, 0.00000020, 0.00000406, 0.00000428, },
		{ -0.00001924, -0.00001219, -0.00000973, -0.00001034, -0.00001125, -0.00000994, -0.00000541, 0.00000128, 0.00000766, 0.00001112, 0.00001037, 0.00000608, 0.00000051, -0.00000371, -0.00000507, -0.00000401, -0.00000267, -0.00000371, -0.00000865, -0.00001679, -0.00002530, -0.00003038, -0.00002915, -0.00002115, -0.00000868, 0.00000415, 0.00001337, 0.00001698, 0.00001577, 0.00001269, 0.00001104, 0.00001250, 0.00001602, 0.00001824, 0.00001531, 0.00000521, -0.00001073, -0.00002792, -0.00004043, -0.00004367, -0.00003676, -0.00002310, -0.00000884, 0.00000006, 0.00000083, -0.00000464, -0.00001078, -0.00001124, -0.00000251, 0.00001368, 0.00003053, 0.00003909, 0.00003265, 0.00001059, -0.00002035, -0.00004810, -0.00006020, -0.00004931, -0.00001700, 0.00002606, 0.00006390, 0.00008141, 0.00007047, 0.00003341, -0.00001778, -0.00006593, -0.00009521, -0.00009692, -0.00007221, -0.00003091, 0.00001279, 0.00004590, 0.00006103, 0.00005803, 0.00004267, 0.00002317, 0.00000650, -0.00000398, -0.00000881, -0.00001090, -0.00001321, -0.00001695, -0.00002110, -0.00002331, -0.00002145, -0.00001495, -0.00000525, 0.00000497, 0.00001299, 0.00001729, 0.00001799, 0.00001647, 0.00001437, 0.00001271, 0.00001143, 0.00000978, 0.00000707, 0.00000334, -0.00000042, -0.00000269, -0.00000228, 0.00000091, 0.00000577, 0.00001038, 0.00001301, 0.00001299, 0.00001109, 0.00000918, 0.00000929, 0.00001267, 0.00001914, 0.00002725, 0.00003487, 0.00004010, 0.00004187, 0.00004000, 0.00003490, 0.00002697, 0.00001634, 0.00000308, -0.00001234, -0.00002850, -0.00004299, -0.00005297, -0.00005623, -0.00005221, -0.00004246, -0.00003023, },
		{ -0.00006293, -0.00005753, -0.00004534, -0.00003104, -0.00001957, -0.00001426, -0.00001548, -0.00002079, -0.00002610, -0.00002776, -0.00002418, -0.00001651, -0.00000787, -0.00000162, 0.00000044, -0.00000104, -0.00000331, -0.00000316, 0.00000119, 0.00000893, 0.00001678, 0.00002053, 0.00001721, 0.00000674, -0.00000780, -0.00002145, -0.00002961, -0.00003024, -0.00002463, -0.00001668, -0.00001069, -0.00000907, -0.00001107, -0.00001316, -0.00001108, -0.00000241, 0.00001176, 0.00002686, 0.00003685, 0.00003714, 0.00002703, 0.00001030, -0.00000644, -0.00001690, -0.00001827, -0.00001266, -0.00000607, -0.00000516, -0.00001348, -0.00002900, -0.00004445, -0.00005042, -0.00004015, -0.00001349, 0.00002188, 0.00005282, 0.00006602, 0.00005394, 0.00001872, -0.00002780, -0.00006829, -0.00008657, -0.00007421, -0.00003405, 0.00002066, 0.00007128, 0.00010085, 0.00010023, 0.00007102, 0.00002427, -0.00002429, -0.00006035, -0.00007577, -0.00007047, -0.00005098, -0.00002672, -0.00000589, 0.00000728, 0.00001300, 0.00001438, 0.00001499, 0.00001679, 0.00001939, 0.00002084, 0.00001908, 0.00001341, 0.00000498, -0.00000376, -0.00001027, -0.00001310, -0.00001241, -0.00000959, -0.00000636, -0.00000383, -0.00000202, -0.00000017, 0.00000260, 0.00000650, 0.00001085, 0.00001438, 0.00001592, 0.00001515, 0.00001278, 0.00001020, 0.00000880, 0.00000915, 0.00001075, 0.00001232, 0.00001253, 0.00001086, 0.00000789, 0.00000521, 0.00000467, 0.00000756, 0.00001409, 0.00002331, 0.00003353, 0.00004292, 0.00004997, 0.00005355, 0.00005280, 0.00004696, 0.00003547, 0.00001845, -0.00000278, -0.00002539, -0.00004543, -0.00005884, },
		{ 0.00000531, -0.00002425, -0.00005093, -0.00006887, -0.00007421, -0.00006660, -0.00004965, -0.00002966, -0.00001337, -0.00000541, -0.00000665, -0.00001415, -0.00002279, -0.00002777, -0.00002683, -0.00002115, -0.00001447, -0.00001096, -0.00001287, -0.00001917, -0.00002597, -0.00002845, -0.00002340, -0.00001102, 0.00000489, 0.00001863, 0.00002524, 0.00002292, 0.00001380, 0.00000285, -0.00000467, -0.00000594, -0.00000202, 0.00000280, 0.00000332, -0.00000356, -0.00001694, -0.00003193, -0.00004188, -0.00004170, -0.00003044, -0.00001208, 0.00000632, 0.00001798, 0.00001984, 0.00001407, 0.00000695, 0.00000541, 0.00001306, 0.00002762, 0.00004134, 0.00004444, 0.00003014, -0.00000122, -0.00004092, -0.00007468, -0.00008827, -0.00007389, -0.00003420, 0.00001781, 0.00006341, 0.00008535, 0.00007479, 0.00003496, -0.00001991, -0.00007007, -0.00009776, -0.00009377, -0.00006041, -0.00000995, 0.00004060, 0.00007603, 0.00008807, 0.00007732, 0.00005155, 0.00002159, -0.00000322, -0.00001811, -0.00002342, -0.00002293, -0.00002114, -0.00002083, -0.00002214, -0.00002313, -0.00002147, -0.00001593, -0.00000720, 0.00000255, 0.00001077, 0.00001578, 0.00001739, 0.00001665, 0.00001506, 0.00001366, 0.00001253, 0.00001100, 0.00000829, 0.00000422, -0.00000050, -0.00000452, -0.00000655, -0.00000600, -0.00000330, 0.00000039, 0.00000376, 0.00000609, 0.00000751, 0.00000882, 0.00001088, 0.00001400, 0.00001764, 0.00002060, 0.00002166, 0.00002022, 0.00001673, 0.00001257, 0.00000957, 0.00000930, 0.00001260, 0.00001943, 0.00002899, 0.00003998, 0.00005082, 0.00005960, 0.00006408, 0.00006195, 0.00005137, 0.00003191, },
		{ 0.00007698, 0.00007964, 0.00007133, 0.00005043, 0.00001851, -0.00001928, -0.00005522, -0.00008121, -0.00009151, -0.00008492, -0.00006538, -0.00004056, -0.00001905, -0.00000704, -0.00000607, -0.00001288, -0.00002121, -0.00002501, -0.00002131, -0.00001160, -0.00000091, 0.00000479, 0.00000171, -0.00000967, -0.00002475, -0.00003693, -0.00004079, -0.00003475, -0.00002183, -0.00000807, 0.00000053, 0.00000108, -0.00000471, -0.00001144, -0.00001292, -0.00000552, 0.00000957, 0.00002660, 0.00003784, 0.00003737, 0.00002413, 0.00000264, -0.00001905, -0.00003315, -0.00003593, -0.00002948, -0.00002047, -0.00001643, -0.00002139, -0.00003316, -0.00004379, -0.00004325, -0.00002477, 0.00001077, 0.00005372, 0.00008866, 0.00010056, 0.00008148, 0.00003483, -0.00002498, -0.00007753, -0.00010417, -0.00009539, -0.00005465, 0.00000288, 0.00005627, 0.00008680, 0.00008489, 0.00005312, 0.00000450, -0.00004322, -0.00007444, -0.00008114, -0.00006473, -0.00003412, -0.00000118, 0.00002415, 0.00003717, 0.00003889, 0.00003412, 0.00002836, 0.00002508, 0.00002459, 0.00002469, 0.00002249, 0.00001620, 0.00000620, -0.00000531, -0.00001540, -0.00002184, -0.00002389, -0.00002229, -0.00001852, -0.00001389, -0.00000897, -0.00000361, 0.00000253, 0.00000935, 0.00001604, 0.00002129, 0.00002385, 0.00002321, 0.00001987, 0.00001512, 0.00001043, 0.00000685, 0.00000458, 0.00000317, 0.00000198, 0.00000071, -0.00000031, -0.00000031, 0.00000150, 0.00000535, 0.00001062, 0.00001605, 0.00002021, 0.00002212, 0.00002164, 0.00001950, 0.00001703, 0.00001569, 0.00001675, 0.00002103, 0.00002890, 0.00004009, 0.00005351, 0.00006692, },
		{ 0.00002598, 0.00003671, 0.00005201, 0.00006987, 0.00008618, 0.00009534, 0.00009182, 0.00007233, 0.00003764, -0.00000660, -0.00005102, -0.00008521, -0.00010128, -0.00009683, -0.00007595, -0.00004782, -0.00002316, -0.00001007, -0.00001101, -0.00002212, -0.00003531, -0.00004218, -0.00003787, -0.00002325, -0.00000432, 0.00001083, 0.00001595, 0.00000964, -0.00000409, -0.00001794, -0.00002507, -0.00002256, -0.00001294, -0.00000282, 0.00000059, -0.00000660, -0.00002260, -0.00004047, -0.00005126, -0.00004838, -0.00003104, -0.00000485, 0.00002062, 0.00003628, 0.00003787, 0.00002786, 0.00001400, 0.00000513, 0.00000629, 0.00001565, 0.00002498, 0.00002368, 0.00000457, -0.00003143, -0.00007403, -0.00010694, -0.00011445, -0.00008861, -0.00003367, 0.00003440, 0.00009336, 0.00012312, 0.00011359, 0.00006868, 0.00000484, -0.00005532, -0.00009160, -0.00009359, -0.00006381, -0.00001578, 0.00003200, 0.00006337, 0.00007018, 0.00005417, 0.00002479, -0.00000564, -0.00002710, -0.00003527, -0.00003193, -0.00002274, -0.00001386, -0.00000896, -0.00000814, -0.00000861, -0.00000683, -0.00000059, 0.00000983, 0.00002190, 0.00003217, 0.00003784, 0.00003787, 0.00003305, 0.00002532, 0.00001669, 0.00000850, 0.00000121, -0.00000522, -0.00001075, -0.00001489, -0.00001679, -0.00001571, -0.00001151, -0.00000496, 0.00000251, 0.00000933, 0.00001442, 0.00001748, 0.00001895, 0.00001957, 0.00001983, 0.00001975, 0.00001890, 0.00001686, 0.00001362, 0.00000978, 0.00000642, 0.00000465, 0.00000511, 0.00000769, 0.00001156, 0.00001555, 0.00001862, 0.00002016, 0.00002017, 0.00001914, 0.00001796, 0.00001778, 0.00001999, },
		{ 0.00002457, 0.00002358, 0.00002233, 0.00002234, 0.00002577, 0.00003476, 0.00005024, 0.00007073, 0.00009174, 0.00010630, 0.00010689, 0.00008833, 0.00005037, -0.00000100, -0.00005454, -0.00009706, -0.00011789, -0.00011286, -0.00008609, -0.00004865, -0.00001444, 0.00000509, 0.00000522, -0.00001058, -0.00003267, -0.00004974, -0.00005391, -0.00004401, -0.00002561, -0.00000793, 0.00000092, -0.00000205, -0.00001328, -0.00002460, -0.00002759, -0.00001806, 0.00000142, 0.00002250, 0.00003490, 0.00003148, 0.00001197, -0.00001659, -0.00004296, -0.00005692, -0.00005409, -0.00003798, -0.00001810, -0.00000501, -0.00000477, -0.00001547, -0.00002785, -0.00002984, -0.00001307, 0.00002194, 0.00006438, 0.00009685, 0.00010253, 0.00007302, 0.00001309, -0.00005987, -0.00012187, -0.00015139, -0.00013780, -0.00008564, -0.00001294, 0.00005570, 0.00009837, 0.00010370, 0.00007423, 0.00002429, -0.00002640, -0.00006063, -0.00006958, -0.00005491, -0.00002635, 0.00000338, 0.00002406, 0.00003144, 0.00002762, 0.00001863, 0.00001083, 0.00000782, 0.00000929, 0.00001199, 0.00001197, 0.00000689, -0.00000272, -0.00001388, -0.00002259, -0.00002574, -0.00002231, -0.00001355, -0.00000209, 0.00000925, 0.00001855, 0.00002512, 0.00002920, 0.00003131, 0.00003171, 0.00003026, 0.00002675, 0.00002129, 0.00001459, 0.00000786, 0.00000240, -0.00000094, -0.00000204, -0.00000145, 0.00000004, 0.00000180, 0.00000369, 0.00000595, 0.00000890, 0.00001255, 0.00001643, 0.00001970, 0.00002153, 0.00002152, 0.00001994, 0.00001760, 0.00001551, 0.00001451, 0.00001491, 0.00001656, 0.00001893, 0.00002140, 0.00002342, 0.00002455, },
		{ 0.00002336, 0.00002520, 0.00002731, 0.00002921, 0.00003023, 0.00002988, 0.00002841, 0.00002735, 0.00002950, 0.00003810, 0.00005523, 0.00008003, 0.00010757, 0.00012922, 0.00013496, 0.00011694, 0.00007321, 0.00000976, -0.00006010, -0.00011945, -0.00015332, -0.00015407, -0.00012435, -0.00007632, -0.00002714, 0.00000747, 0.00001904, 0.00000897, -0.00001286, -0.00003316, -0.00004142, -0.00003452, -0.00001768, -0.00000142, 0.00000402, -0.00000595, -0.00002771, -0.00005099, -0.00006377, -0.00005819, -0.00003462, -0.00000173, 0.00002739, 0.00004140, 0.00003609, 0.00001640, -0.00000618, -0.00001935, -0.00001651, -0.00000037, 0.00001793, 0.00002417, 0.00000853, -0.00002859, -0.00007515, -0.00011166, -0.00011929, -0.00008856, -0.00002457, 0.00005372, 0.00012008, 0.00015105, 0.00013508, 0.00007719, -0.00000294, -0.00007852, -0.00012569, -0.00013203, -0.00010021, -0.00004563, 0.00001047, 0.00004945, 0.00006169, 0.00004869, 0.00002066, -0.00000899, -0.00002944, -0.00003616, -0.00003128, -0.00002108, -0.00001214, -0.00000815, -0.00000874, -0.00001050, -0.00000938, -0.00000317, 0.00000740, 0.00001899, 0.00002743, 0.00002964, 0.00002500, 0.00001535, 0.00000403, -0.00000563, -0.00001156, -0.00001334, -0.00001189, -0.00000854, -0.00000437, 0.00000023, 0.00000532, 0.00001099, 0.00001697, 0.00002257, 0.00002690, 0.00002936, 0.00002986, 0.00002885, 0.00002701, 0.00002488, 0.00002265, 0.00002017, 0.00001726, 0.00001395, 0.00001073, 0.00000832, 0.00000743, 0.00000834, 0.00001077, 0.00001395, 0.00001700, 0.00001923, 0.00002043, 0.00002081, 0.00002079, 0.00002081, 0.00002116, 0.00002200, },
		{ 0.00002482, 0.00002678, 0.00002851, 0.00003001, 0.00003147, 0.00003317, 0.00003512, 0.00003681, 0.00003723, 0.00003535, 0.00003109, 0.00002617, 0.00002437, 0.00003060, 0.00004880, 0.00007943, 0.00011757, 0.00015291, 0.00017214, 0.00016345, 0.00012139, 0.00005032, -0.00003558, -0.00011597, -0.00017096, -0.00018806, -0.00016675, -0.00011866, -0.00006310, -0.00001953, -0.00000015, -0.00000565, -0.00002588, -0.00004522, -0.00005005, -0.00003509, -0.00000572, 0.00002482, 0.00004191, 0.00003664, 0.00001024, -0.00002634, -0.00005759, -0.00007071, -0.00006169, -0.00003709, -0.00001070, 0.00000347, -0.00000154, -0.00002169, -0.00004347, -0.00005035, -0.00003127, 0.00001281, 0.00006764, 0.00011072, 0.00012064, 0.00008681, 0.00001518, -0.00007270, -0.00014732, -0.00018240, -0.00016506, -0.00010089, -0.00001170, 0.00007281, 0.00012622, 0.00013467, 0.00010083, 0.00004139, -0.00002055, -0.00006459, -0.00008004, -0.00006825, -0.00003999, -0.00000955, 0.00001145, 0.00001810, 0.00001261, 0.00000163, -0.00000778, -0.00001159, -0.00001007, -0.00000673, -0.00000576, -0.00000955, -0.00001741, -0.00002608, -0.00003148, -0.00003078, -0.00002370, -0.00001251, -0.00000082, 0.00000805, 0.00001234, 0.00001228, 0.00000955, 0.00000623, 0.00000379, 0.00000267, 0.00000245, 0.00000242, 0.00000215, 0.00000180, 0.00000191, 0.00000309, 0.00000562, 0.00000930, 0.00001366, 0.00001816, 0.00002250, 0.00002659, 0.00003041, 0.00003379, 0.00003629, 0.00003736, 0.00003657, 0.00003391, 0.00002989, 0.00002541, 0.00002141, 0.00001859, 0.00001720, 0.00001707, 0.00001785, 0.00001920, 0.00002088, 0.00002280, },
		{ 0.00002611, 0.00002333, 0.00002195, 0.00002220, 0.00002394, 0.00002677, 0.00003027, 0.00003423, 0.00003881, 0.00004429, 0.00005054, 0.00005650, 0.00006018, 0.00005929, 0.00005265, 0.00004164, 0.00003084, 0.00002723, 0.00003766, 0.00006542, 0.00010726, 0.00015259, 0.00018564, 0.00019052, 0.00015744, 0.00008770, -0.00000506, -0.00009826, -0.00016769, -0.00019611, -0.00017972, -0.00012975, -0.00006813, -0.00001894, 0.00000098, -0.00001160, -0.00004569, -0.00008153, -0.00009959, -0.00008925, -0.00005327, -0.00000614, 0.00003297, 0.00004918, 0.00003876, 0.00001059, -0.00001850, -0.00003226, -0.00002333, 0.00000263, 0.00002929, 0.00003750, 0.00001509, -0.00003589, -0.00009829, -0.00014601, -0.00015485, -0.00011353, -0.00002975, 0.00007131, 0.00015593, 0.00019457, 0.00017321, 0.00009876, -0.00000370, -0.00010052, -0.00016190, -0.00017229, -0.00013464, -0.00006755, 0.00000321, 0.00005482, 0.00007509, 0.00006510, 0.00003632, 0.00000422, -0.00001853, -0.00002643, -0.00002168, -0.00001136, -0.00000299, -0.00000094, -0.00000494, -0.00001122, -0.00001513, -0.00001391, -0.00000795, -0.00000039, 0.00000470, 0.00000448, -0.00000132, -0.00001050, -0.00001957, -0.00002548, -0.00002690, -0.00002453, -0.00002046, -0.00001694, -0.00001532, -0.00001558, -0.00001663, -0.00001706, -0.00001592, -0.00001313, -0.00000936, -0.00000559, -0.00000255, -0.00000048, 0.00000090, 0.00000208, 0.00000355, 0.00000554, 0.00000819, 0.00001155, 0.00001575, 0.00002092, 0.00002702, 0.00003369, 0.00004022, 0.00004571, 0.00004936, 0.00005072, 0.00004984, 0.00004717, 0.00004332, 0.00003888, 0.00003429, 0.00002991, },
		{ 0.00007279, 0.00007300, 0.00007059, 0.00006573, 0.00005897, 0.00005124, 0.00004358, 0.00003683, 0.00003140, 0.00002738, 0.00002482, 0.00002422, 0.00002659, 0.00003303, 0.00004393, 0.00005805, 0.00007221, 0.00008213, 0.00008421, 0.00007780, 0.00006660, 0.00005831, 0.00006198, 0.00008382, 0.00012300, 0.00016973, 0.00020684, 0.00021520, 0.00018119, 0.00010337, -0.00000471, -0.00011728, -0.00020442, -0.00024256, -0.00022341, -0.00015749, -0.00007058, 0.00000584, 0.00004648, 0.00004161, -0.00000003, -0.00005590, -0.00009990, -0.00011372, -0.00009436, -0.00005447, -0.00001551, 0.00000301, -0.00000672, -0.00003666, -0.00006661, -0.00007393, -0.00004507, 0.00001648, 0.00008991, 0.00014474, 0.00015367, 0.00010489, 0.00000854, -0.00010566, -0.00019906, -0.00023866, -0.00020972, -0.00012134, -0.00000289, 0.00010718, 0.00017541, 0.00018482, 0.00013948, 0.00006092, -0.00002149, -0.00008184, -0.00010634, -0.00009609, -0.00006386, -0.00002695, 0.00000032, 0.00001156, 0.00000884, -0.00000031, -0.00000772, -0.00000861, -0.00000320, 0.00000444, 0.00000923, 0.00000800, 0.00000099, -0.00000853, -0.00001624, -0.00001895, -0.00001608, -0.00000960, -0.00000287, 0.00000119, 0.00000133, -0.00000165, -0.00000570, -0.00000871, -0.00000966, -0.00000897, -0.00000813, -0.00000878, -0.00001184, -0.00001709, -0.00002331, -0.00002896, -0.00003276, -0.00003423, -0.00003361, -0.00003165, -0.00002910, -0.00002651, -0.00002406, -0.00002173, -0.00001939, -0.00001688, -0.00001400, -0.00001041, -0.00000570, 0.00000055, 0.00000853, 0.00001808, 0.00002863, 0.00003936, 0.00004944, 0.00005819, 0.00006517, 0.00007011, },
		{ 0.00000377, 0.00001859, 0.00003454, 0.00005105, 0.00006721, 0.00008180, 0.00009359, 0.00010167, 0.00010575, 0.00010613, 0.00010336, 0.00009787, 0.00008974, 0.00007886, 0.00006553, 0.00005111, 0.00003820, 0.00003010, 0.00002950, 0.00003696, 0.00005007, 0.00006398, 0.00007341, 0.00007563, 0.00007276, 0.00007212, 0.00008372, 0.00011549, 0.00016792, 0.00023080, 0.00028411, 0.00030367, 0.00027005, 0.00017746, 0.00003873, -0.00011630, -0.00024928, -0.00032608, -0.00032894, -0.00026326, -0.00015582, -0.00004493, 0.00003392, 0.00006106, 0.00003908, -0.00001000, -0.00005562, -0.00007279, -0.00005335, -0.00000939, 0.00003282, 0.00004554, 0.00001342, -0.00005783, -0.00014213, -0.00020308, -0.00020923, -0.00014819, -0.00003336, 0.00009981, 0.00020660, 0.00024995, 0.00021448, 0.00011218, -0.00002236, -0.00014513, -0.00021872, -0.00022501, -0.00016980, -0.00007833, 0.00001582, 0.00008356, 0.00010980, 0.00009634, 0.00005802, 0.00001463, -0.00001757, -0.00003127, -0.00002875, -0.00001846, -0.00000962, -0.00000776, -0.00001288, -0.00002062, -0.00002541, -0.00002367, -0.00001552, -0.00000437, 0.00000504, 0.00000915, 0.00000707, 0.00000070, -0.00000653, -0.00001151, -0.00001284, -0.00001116, -0.00000844, -0.00000669, -0.00000686, -0.00000847, -0.00001002, -0.00000997, -0.00000763, -0.00000355, 0.00000074, 0.00000345, 0.00000333, 0.00000013, -0.00000545, -0.00001214, -0.00001868, -0.00002426, -0.00002861, -0.00003188, -0.00003444, -0.00003660, -0.00003858, -0.00004048, -0.00004232, -0.00004406, -0.00004546, -0.00004612, -0.00004548, -0.00004295, -0.00003811, -0.00003083, -0.00002123, -0.00000960, },
		{ -0.00006593, -0.00006919, -0.00007086, -0.00007058, -0.00006770, -0.00006138, -0.00005088, -0.00003591, -0.00001686, 0.00000519, 0.00002882, 0.00005275, 0.00007614, 0.00009855, 0.00011948, 0.00013790, 0.00015194, 0.00015929, 0.00015821, 0.00014862, 0.00013280, 0.00011495, 0.00009967, 0.00008982, 0.00008498, 0.00008147, 0.00007438, 0.00006092, 0.00004369, 0.00003171, 0.00003808, 0.00007449, 0.00014435, 0.00023777, 0.00033107, 0.00039248, 0.00039252, 0.00031582, 0.00016927, -0.00001710, -0.00019821, -0.00032841, -0.00037738, -0.00034112, -0.00024338, -0.00012647, -0.00003485, 0.00000216, -0.00001866, -0.00007416, -0.00012576, -0.00013693, -0.00008990, 0.00000547, 0.00011469, 0.00019247, 0.00020162, 0.00012936, -0.00000598, -0.00016071, -0.00028185, -0.00032694, -0.00027990, -0.00015657, 0.00000176, 0.00014408, 0.00022812, 0.00023442, 0.00017093, 0.00006730, -0.00003806, -0.00011257, -0.00013988, -0.00012277, -0.00007853, -0.00002960, 0.00000591, 0.00002015, 0.00001612, 0.00000362, -0.00000678, -0.00000878, -0.00000242, 0.00000734, 0.00001408, 0.00001353, 0.00000554, -0.00000627, -0.00001675, -0.00002183, -0.00002033, -0.00001411, -0.00000674, -0.00000158, -0.00000022, -0.00000210, -0.00000520, -0.00000738, -0.00000757, -0.00000617, -0.00000462, -0.00000445, -0.00000634, -0.00000977, -0.00001327, -0.00001521, -0.00001458, -0.00001148, -0.00000695, -0.00000251, 0.00000052, 0.00000148, 0.00000044, -0.00000198, -0.00000500, -0.00000804, -0.00001083, -0.00001339, -0.00001592, -0.00001867, -0.00002189, -0.00002581, -0.00003056, -0.00003619, -0.00004252, -0.00004916, -0.00005560, -0.00006132, },
		{ -0.00001677, -0.00002137, -0.00002672, -0.00003277, -0.00003974, -0.00004800, -0.00005782, -0.00006903, -0.00008079, -0.00009167, -0.00010004, -0.00010454, -0.00010449, -0.00009983, -0.00009075, -0.00007720, -0.00005852, -0.00003371, -0.00000216, 0.00003543, 0.00007645, 0.00011675, 0.00015184, 0.00017862, 0.00019664, 0.00020817, 0.00021668, 0.00022440, 0.00023021, 0.00022921, 0.00021481, 0.00018283, 0.00013585, 0.00008540, 0.00005013, 0.00004954, 0.00009522, 0.00018317, 0.00029090, 0.00038200, 0.00041754, 0.00037066, 0.00023887, 0.00004848, -0.00015175, -0.00030628, -0.00037263, -0.00033757, -0.00022315, -0.00007936, 0.00003434, 0.00007176, 0.00001818, -0.00010344, -0.00024103, -0.00033362, -0.00033545, -0.00023444, -0.00005763, 0.00013839, 0.00028847, 0.00034291, 0.00028534, 0.00013767, -0.00004907, -0.00021414, -0.00030869, -0.00031150, -0.00023344, -0.00010993, 0.00001437, 0.00010231, 0.00013580, 0.00011865, 0.00007081, 0.00001773, -0.00002044, -0.00003527, -0.00003045, -0.00001711, -0.00000696, -0.00000670, -0.00001585, -0.00002843, -0.00003695, -0.00003646, -0.00002677, -0.00001200, 0.00000186, 0.00000997, 0.00001068, 0.00000573, -0.00000110, -0.00000616, -0.00000759, -0.00000586, -0.00000306, -0.00000144, -0.00000218, -0.00000487, -0.00000794, -0.00000968, -0.00000920, -0.00000687, -0.00000404, -0.00000227, -0.00000253, -0.00000476, -0.00000797, -0.00001080, -0.00001218, -0.00001180, -0.00001011, -0.00000802, -0.00000644, -0.00000592, -0.00000646, -0.00000771, -0.00000912, -0.00001024, -0.00001079, -0.00001075, -0.00001027, -0.00000962, -0.00000921, -0.00000945, -0.00001070, -0.00001315, },
		{ -0.00001513, -0.00001414, -0.00001348, -0.00001327, -0.00001327, -0.00001310, -0.00001246, -0.00001145, -0.00001073, -0.00001142, -0.00001470, -0.00002132, -0.00003125, -0.00004375, -0.00005772, -0.00007232, -0.00008729, -0.00010288, -0.00011919, -0.00013540, -0.00014920, -0.00015710, -0.00015542, -0.00014178, -0.00011621, -0.00008128, -0.00004089, 0.00000167, 0.00004546, 0.00009227, 0.00014508, 0.00020517, 0.00026916, 0.00032803, 0.00036921, 0.00038159, 0.00036166, 0.00031774, 0.00026958, 0.00024222, 0.00025557, 0.00031378, 0.00039908, 0.00047412, 0.00049343, 0.00042085, 0.00024638, -0.00000452, -0.00027439, -0.00049015, -0.00058766, -0.00053555, -0.00034908, -0.00008796, 0.00016222, 0.00031875, 0.00033045, 0.00019570, -0.00003717, -0.00028713, -0.00046921, -0.00052385, -0.00043633, -0.00023961, -0.00000017, 0.00020683, 0.00032356, 0.00032717, 0.00023340, 0.00008639, -0.00006022, -0.00016280, -0.00020081, -0.00017932, -0.00012192, -0.00005805, -0.00001082, 0.00001020, 0.00000916, -0.00000157, -0.00000915, -0.00000646, 0.00000564, 0.00002026, 0.00002909, 0.00002691, 0.00001393, -0.00000479, -0.00002215, -0.00003244, -0.00003363, -0.00002766, -0.00001880, -0.00001135, -0.00000764, -0.00000740, -0.00000856, -0.00000873, -0.00000665, -0.00000274, 0.00000131, 0.00000359, 0.00000308, 0.00000010, -0.00000394, -0.00000727, -0.00000874, -0.00000819, -0.00000643, -0.00000470, -0.00000396, -0.00000445, -0.00000574, -0.00000700, -0.00000750, -0.00000698, -0.00000575, -0.00000444, -0.00000375, -0.00000412, -0.00000561, -0.00000795, -0.00001067, -0.00001325, -0.00001527, -0.00001643, -0.00001666, -0.00001611, },
		{ -0.00001038, -0.00001245, -0.00001380, -0.00001450, -0.00001494, -0.00001566, -0.00001709, -0.00001923, -0.00002161, -0.00002336, -0.00002361, -0.00002197, -0.00001875, -0.00001493, -0.00001171, -0.00000993, -0.00000970, -0.00001047, -0.00001164, -0.00001326, -0.00001660, -0.00002392, -0.00003763, -0.00005892, -0.00008684, -0.00011812, -0.00014824, -0.00017317, -0.00019088, -0.00020185, -0.00020794, -0.00021014, -0.00020637, -0.00019076, -0.00015538, -0.00009402, -0.00000659, 0.00009848, 0.00020543, 0.00029663, 0.00036035, 0.00039753, 0.00042362, 0.00046351, 0.00054006, 0.00066040, 0.00080595, 0.00093168, 0.00097720, 0.00088733, 0.00063556, 0.00024106, -0.00022831, -0.00067061, -0.00097848, -0.00107302, -0.00093093, -0.00059479, -0.00016186, 0.00024540, 0.00051716, 0.00058962, 0.00046242, 0.00019462, -0.00011836, -0.00037804, -0.00051426, -0.00050405, -0.00037274, -0.00017874, 0.00001057, 0.00014243, 0.00019334, 0.00017090, 0.00010439, 0.00002940, -0.00002651, -0.00005189, -0.00005132, -0.00003882, -0.00002895, -0.00002973, -0.00004026, -0.00005302, -0.00005904, -0.00005290, -0.00003528, -0.00001217, 0.00000839, 0.00002011, 0.00002098, 0.00001346, 0.00000266, -0.00000638, -0.00001086, -0.00001099, -0.00000909, -0.00000790, -0.00000898, -0.00001203, -0.00001528, -0.00001672, -0.00001518, -0.00001102, -0.00000577, -0.00000137, 0.00000081, 0.00000050, -0.00000147, -0.00000380, -0.00000540, -0.00000585, -0.00000547, -0.00000498, -0.00000501, -0.00000575, -0.00000689, -0.00000781, -0.00000798, -0.00000715, -0.00000552, -0.00000360, -0.00000200, -0.00000122, -0.00000154, -0.00000293, -0.00000516, -0.00000781, },
		{ -0.00000289, -0.00000339, -0.00000475, -0.00000664, -0.00000853, -0.00000986, -0.00001021, -0.00000961, -0.00000857, -0.00000791, -0.00000841, -0.00001044, -0.00001370, -0.00001733, -0.00002036, -0.00002222, -0.00002311, -0.00002386, -0.00002541, -0.00002805, -0.00003098, -0.00003238, -0.00003029, -0.00002372, -0.00001358, -0.00000275, 0.00000486, 0.00000586, -0.00000108, -0.00001485, -0.00003269, -0.00005219, -0.00007314, -0.00009823, -0.00013176, -0.00017667, -0.00023124, -0.00028736, -0.00033185, -0.00035081, -0.00033554, -0.00028707, -0.00021658, -0.00014047, -0.00007146, -0.00000936, 0.00006336, 0.00017784, 0.00036615, 0.00064346, 0.00099198, 0.00135443, 0.00164218, 0.00175818, 0.00162833, 0.00123102, 0.00061337, -0.00011337, -0.00080000, -0.00129983, -0.00151076, -0.00140485, -0.00103543, -0.00051931, -0.00000020, 0.00039400, 0.00058599, 0.00056530, 0.00038201, 0.00012258, -0.00012238, -0.00028653, -0.00034351, -0.00030645, -0.00021435, -0.00011251, -0.00003510, 0.00000393, 0.00001013, 0.00000034, -0.00000806, -0.00000537, 0.00000779, 0.00002318, 0.00003093, 0.00002517, 0.00000682, -0.00001736, -0.00003846, -0.00004963, -0.00004887, -0.00003913, -0.00002620, -0.00001566, -0.00001049, -0.00001027, -0.00001221, -0.00001312, -0.00001117, -0.00000668, -0.00000166, 0.00000154, 0.00000149, -0.00000160, -0.00000611, -0.00000996, -0.00001163, -0.00001081, -0.00000832, -0.00000559, -0.00000382, -0.00000349, -0.00000426, -0.00000533, -0.00000595, -0.00000580, -0.00000512, -0.00000443, -0.00000425, -0.00000478, -0.00000584, -0.00000699, -0.00000773, -0.00000774, -0.00000700, -0.00000574, -0.00000436, -0.00000328, },
		{ -0.00000589, -0.00000513, -0.00000411, -0.00000317, -0.00000276, -0.00000328, -0.00000492, -0.00000749, -0.00001040, -0.00001284, -0.00001406, -0.00001377, -0.00001229, -0.00001043, -0.00000906, -0.00000869, -0.00000909, -0.00000948, -0.00000900, -0.00000738, -0.00000541, -0.00000479, -0.00000740, -0.00001420, -0.00002447, -0.00003577, -0.00004474, -0.00004867, -0.00004673, -0.00004049, -0.00003303, -0.00002726, -0.00002408, -0.00002176, -0.00001683, -0.00000661, 0.00000814, 0.00002182, 0.00002504, 0.00000832, -0.00003330, -0.00009679, -0.00017109, -0.00024128, -0.00029543, -0.00033112, -0.00035758, -0.00039141, -0.00044612, -0.00051942, -0.00058444, -0.00059035, -0.00047530, -0.00018912, 0.00028124, 0.00089368, 0.00154905, 0.00210941, 0.00243313, 0.00241670, 0.00202967, 0.00132995, 0.00045321, -0.00042177, -0.00112098, -0.00152138, -0.00158079, -0.00134261, -0.00091531, -0.00043498, -0.00002423, 0.00023919, 0.00033532, 0.00029497, 0.00017831, 0.00004864, -0.00004866, -0.00009599, -0.00010084, -0.00008460, -0.00006903, -0.00006633, -0.00007611, -0.00008891, -0.00009329, -0.00008246, -0.00005746, -0.00002598, 0.00000185, 0.00001828, 0.00002101, 0.00001328, 0.00000149, -0.00000824, -0.00001268, -0.00001231, -0.00001017, -0.00000966, -0.00001256, -0.00001826, -0.00002429, -0.00002784, -0.00002724, -0.00002270, -0.00001606, -0.00000976, -0.00000564, -0.00000424, -0.00000479, -0.00000587, -0.00000623, -0.00000538, -0.00000372, -0.00000215, -0.00000148, -0.00000203, -0.00000351, -0.00000523, -0.00000650, -0.00000696, -0.00000666, -0.00000597, -0.00000535, -0.00000509, -0.00000525, -0.00000568, -0.00000608, -0.00000621, },
		{ -0.00000581, -0.00000590, -0.00000612, -0.00000627, -0.00000612, -0.00000544, -0.00000418, -0.00000255, -0.00000106, -0.00000035, -0.00000095, -0.00000297, -0.00000600, -0.00000921, -0.00001175, -0.00001316, -0.00001365, -0.00001400, -0.00001504, -0.00001709, -0.00001951, -0.00002086, -0.00001955, -0.00001480, -0.00000732, 0.00000066, 0.00000619, 0.00000697, 0.00000248, -0.00000562, -0.00001432, -0.00002085, -0.00002436, -0.00002658, -0.00003095, -0.00004051, -0.00005545, -0.00007187, -0.00008271, -0.00008072, -0.00006240, -0.00003078, 0.00000469, 0.00003164, 0.00004046, 0.00002934, 0.00000580, -0.00001705, -0.00002863, -0.00002994, -0.00003856, -0.00008574, -0.00020426, -0.00041048, -0.00068737, -0.00097672, -0.00118641, -0.00121365, -0.00097840, -0.00045620, 0.00030187, 0.00117128, 0.00198021, 0.00255341, 0.00276051, 0.00255341, 0.00198021, 0.00117128, 0.00030187, -0.00045620, -0.00097840, -0.00121365, -0.00118641, -0.00097672, -0.00068737, -0.00041048, -0.00020426, -0.00008574, -0.00003856, -0.00002994, -0.00002863, -0.00001705, 0.00000580, 0.00002934, 0.00004046, 0.00003164, 0.00000469, -0.00003078, -0.00006240, -0.00008072, -0.00008271, -0.00007187, -0.00005545, -0.00004051, -0.00003095, -0.00002658, -0.00002436, -0.00002085, -0.00001432, -0.00000562, 0.00000248, 0.00000697, 0.00000619, 0.00000066, -0.00000732, -0.00001480, -0.00001955, -0.00002086, -0.00001951, -0.00001709, -0.00001504, -0.00001400, -0.00001365, -0.00001316, -0.00001175, -0.00000921, -0.00000600, -0.00000297, -0.00000095, -0.00000035, -0.00000106, -0.00000255, -0.00000418, -0.00000544, -0.00000612, -0.00000627, -0.00000612, -0.00000590, },
		{ -0.00000589, -0.00000621, -0.00000608, -0.00000568, -0.00000525, -0.00000509, -0.00000535, -0.00000597, -0.00000666, -0.00000696, -0.00000650, -0.00000523, -0.00000351, -0.00000203, -0.00000148, -0.00000215, -0.00000372, -0.00000538, -0.00000623, -0.00000587, -0.00000479, -0.00000424, -0.00000564, -0.00000976, -0.00001606, -0.00002270, -0.00002724, -0.00002784, -0.00002429, -0.00001826, -0.00001256, -0.00000966, -0.00001017, -0.00001231, -0.00001268, -0.00000824, 0.00000149, 0.00001328, 0.00002101, 0.00001828, 0.00000185, -0.00002598, -0.00005746, -0.00008246, -0.00009329, -0.00008891, -0.00007611, -0.00006633, -0.00006903, -0.00008460, -0.00010084, -0.00009599, -0.00004866, 0.00004864, 0.00017831, 0.00029497, 0.00033532, 0.00023919, -0.00002423, -0.00043498, -0.00091531, -0.00134261, -0.00158079, -0.00152138, -0.00112098, -0.00042177, 0.00045321, 0.00132995, 0.00202967, 0.00241670, 0.00243313, 0.00210941, 0.00154905, 0.00089368, 0.00028124, -0.00018912, -0.00047530, -0.00059035, -0.00058444, -0.00051942, -0.00044612, -0.00039141, -0.00035758, -0.00033112, -0.00029543, -0.00024128, -0.00017109, -0.00009679, -0.00003330, 0.00000832, 0.00002504, 0.00002182, 0.00000814, -0.00000661, -0.00001683, -0.00002176, -0.00002408, -0.00002726, -0.00003303, -0.00004049, -0.00004673, -0.00004867, -0.00004474, -0.00003577, -0.00002447, -0.00001420, -0.00000740, -0.00000479, -0.00000541, -0.00000738, -0.00000900, -0.00000948, -0.00000909, -0.00000869, -0.00000906, -0.00001043, -0.00001229, -0.00001377, -0.00001406, -0.00001284, -0.00001040, -0.00000749, -0.00000492, -0.00000328, -0.00000276, -0.00000317, -0.00000411, -0.00000513, },
		{ -0.00000289, -0.00000328, -0.00000436, -0.00000574, -0.00000700, -0.00000774, -0.00000773, -0.00000699, -0.00000584, -0.00000478, -0.00000425, -0.00000443, -0.00000512, -0.00000580, -0.00000595, -0.00000533, -0.00000426, -0.00000349, -0.00000382, -0.00000559, -0.00000832, -0.00001081, -0.00001163, -0.00000996, -0.00000611, -0.00000160, 0.00000149, 0.00000154, -0.00000166, -0.00000668, -0.00001117, -0.00001312, -0.00001221, -0.00001027, -0.00001049, -0.00001566, -0.00002620, -0.00003913, -0.00004887, -0.00004963, -0.00003846, -0.00001736, 0.00000682, 0.00002517, 0.00003093, 0.00002318, 0.00000779, -0.00000537, -0.00000806, 0.00000034, 0.00001013, 0.00000393, -0.00003510, -0.00011251, -0.00021435, -0.00030645, -0.00034351, -0.00028653, -0.00012238, 0.00012258, 0.00038201, 0.00056530, 0.00058599, 0.00039400, -0.00000020, -0.00051931, -0.00103543, -0.00140485, -0.00151076, -0.00129983, -0.00080000, -0.00011337, 0.00061337, 0.00123102, 0.00162833, 0.00175818, 0.00164218, 0.00135443, 0.00099198, 0.00064346, 0.00036615, 0.00017784, 0.00006336, -0.00000936, -0.00007146, -0.00014047, -0.00021658, -0.00028707, -0.00033554, -0.00035081, -0.00033185, -0.00028736, -0.00023124, -0.00017667, -0.00013176, -0.00009823, -0.00007314, -0.00005219, -0.00003269, -0.00001485, -0.00000108, 0.00000586, 0.00000486, -0.00000275, -0.00001358, -0.00002372, -0.00003029, -0.00003238, -0.00003098, -0.00002805, -0.00002541, -0.00002386, -0.00002311, -0.00002222, -0.00002036, -0.00001733, -0.00001370, -0.00001044, -0.00000841, -0.00000791, -0.00000857, -0.00000961, -0.00001021, -0.00000986, -0.00000853, -0.00000664, -0.00000475, -0.00000339, },
		{ -0.00001038, -0.00000781, -0.00000516, -0.00000293, -0.00000154, -0.00000122, -0.00000200, -0.00000360, -0.00000552, -0.00000715, -0.00000798, -0.00000781, -0.00000689, -0.00000575, -0.00000501, -0.00000498, -0.00000547, -0.00000585, -0.00000540, -0.00000380, -0.00000147, 0.00000050, 0.00000081, -0.00000137, -0.00000577, -0.00001102, -0.00001518, -0.00001672, -0.00001528, -0.00001203, -0.00000898, -0.00000790, -0.00000909, -0.00001099, -0.00001086, -0.00000638, 0.00000266, 0.00001346, 0.00002098, 0.00002011, 0.00000839, -0.00001217, -0.00003528, -0.00005290, -0.00005904, -0.00005302, -0.00004026, -0.00002973, -0.00002895, -0.00003882, -0.00005132, -0.00005189, -0.00002651, 0.00002940, 0.00010439, 0.00017090, 0.00019334, 0.00014243, 0.00001057, -0.00017874, -0.00037274, -0.00050405, -0.00051426, -0.00037804, -0.00011836, 0.00019462, 0.00046242, 0.00058962, 0.00051716, 0.00024540, -0.00016186, -0.00059479, -0.00093093, -0.00107302, -0.00097848, -0.00067061, -0.00022831, 0.00024106, 0.00063556, 0.00088733, 0.00097720, 0.00093168, 0.00080595, 0.00066040, 0.00054006, 0.00046351, 0.00042362, 0.00039753, 0.00036035, 0.00029663, 0.00020543, 0.00009848, -0.00000659, -0.00009402, -0.00015538, -0.00019076, -0.00020637, -0.00021014, -0.00020794, -0.00020185, -0.00019088, -0.00017317, -0.00014824, -0.00011812, -0.00008684, -0.00005892, -0.00003763, -0.00002392, -0.00001660, -0.00001326, -0.00001164, -0.00001047, -0.00000970, -0.00000993, -0.00001171, -0.00001493, -0.00001875, -0.00002197, -0.00002361, -0.00002336, -0.00002161, -0.00001923, -0.00001709, -0.00001566, -0.00001494, -0.00001450, -0.00001380, -0.00001245, },
		{ -0.00001513, -0.00001611, -0.00001666, -0.00001643, -0.00001527, -0.00001325, -0.00001067, -0.00000795, -0.00000561, -0.00000412, -0.00000375, -0.00000444, -0.00000575, -0.00000698, -0.00000750, -0.00000700, -0.00000574, -0.00000445, -0.00000396, -0.00000470, -0.00000643, -0.00000819, -0.00000874, -0.00000727, -0.00000394, 0.00000010, 0.00000308, 0.00000359, 0.00000131, -0.00000274, -0.00000665, -0.00000873, -0.00000856, -0.00000740, -0.00000764, -0.00001135, -0.00001880, -0.00002766, -0.00003363, -0.00003244, -0.00002215, -0.00000479, 0.00001393, 0.00002691, 0.00002909, 0.00002026, 0.00000564, -0.00000646, -0.00000915, -0.00000157, 0.00000916, 0.00001020, -0.00001082, -0.00005805, -0.00012192, -0.00017932, -0.00020081, -0.00016280, -0.00006022, 0.00008639, 0.00023340, 0.00032717, 0.00032356, 0.00020683, -0.00000017, -0.00023961, -0.00043633, -0.00052385, -0.00046921, -0.00028713, -0.00003717, 0.00019570, 0.00033045, 0.00031875, 0.00016222, -0.00008796, -0.00034908, -0.00053555, -0.00058766, -0.00049015, -0.00027439, -0.00000452, 0.00024638, 0.00042085, 0.00049343, 0.00047412, 0.00039908, 0.00031378, 0.00025557, 0.00024222, 0.00026958, 0.00031774, 0.00036166, 0.00038159, 0.00036921, 0.00032803, 0.00026916, 0.00020517, 0.00014508, 0.00009227, 0.00004546, 0.00000167, -0.00004089, -0.00008128, -0.00011621, -0.00014178, -0.00015542, -0.00015710, -0.00014920, -0.00013540, -0.00011919, -0.00010288, -0.00008729, -0.00007232, -0.00005772, -0.00004375, -0.00003125, -0.00002132, -0.00001470, -0.00001142, -0.00001073, -0.00001145, -0.00001246, -0.00001310, -0.00001327, -0.00001327, -0.00001348, -0.00001414, },
		{ -0.00001677, -0.00001315, -0.00001070, -0.00000945, -0.00000921, -0.00000962, -0.00001027, -0.00001075, -0.00001079, -0.00001024, -0.00000912, -0.00000771, -0.00000646, -0.00000592, -0.00000644, -0.00000802, -0.00001011, -0.00001180, -0.00001218, -0.00001080, -0.00000797, -0.00000476, -0.00000253, -0.00000227, -0.00000404, -0.00000687, -0.00000920, -0.00000968, -0.00000794, -0.00000487, -0.00000218, -0.00000144, -0.00000306, -0.00000586, -0.00000759, -0.00000616, -0.00000110, 0.00000573, 0.00001068, 0.00000997, 0.00000186, -0.00001200, -0.00002677, -0.00003646, -0.00003695, -0.00002843, -0.00001585, -0.00000670, -0.00000696, -0.00001711, -0.00003045, -0.00003527, -0.00002044, 0.00001773, 0.00007081, 0.00011865, 0.00013580, 0.00010231, 0.00001437, -0.00010993, -0.00023344, -0.00031150, -0.00030869, -0.00021414, -0.00004907, 0.00013767, 0.00028534, 0.00034291, 0.00028847, 0.00013839, -0.00005763, -0.00023444, -0.00033545, -0.00033362, -0.00024103, -0.00010344, 0.00001818, 0.00007176, 0.00003434, -0.00007936, -0.00022315, -0.00033757, -0.00037263, -0.00030628, -0.00015175, 0.00004848, 0.00023887, 0.00037066, 0.00041754, 0.00038200, 0.00029090, 0.00018317, 0.00009522, 0.00004954, 0.00005013, 0.00008540, 0.00013585, 0.00018283, 0.00021481, 0.00022921, 0.00023021, 0.00022440, 0.00021668, 0.00020817, 0.00019664, 0.00017862, 0.00015184, 0.00011675, 0.00007645, 0.00003543, -0.00000216, -0.00003371, -0.00005852, -0.00007720, -0.00009075, -0.00009983, -0.00010449, -0.00010454, -0.00010004, -0.00009167, -0.00008079, -0.00006903, -0.00005782, -0.00004800, -0.00003974, -0.00003277, -0.00002672, -0.00002137, },
		{ -0.00006593, -0.00006132, -0.00005560, -0.00004916, -0.00004252, -0.00003619, -0.00003056, -0.00002581, -0.00002189, -0.00001867, -0.00001592, -0.00001339, -0.00001083, -0.00000804, -0.00000500, -0.00000198, 0.00000044, 0.00000148, 0.00000052, -0.00000251, -0.00000695, -0.00001148, -0.00001458, -0.00001521, -0.00001327, -0.00000977, -0.00000634, -0.00000445, -0.00000462, -0.00000617, -0.00000757, -0.00000738, -0.00000520, -0.00000210, -0.00000022, -0.00000158, -0.00000674, -0.00001411, -0.00002033, -0.00002183, -0.00001675, -0.00000627, 0.00000554, 0.00001353, 0.00001408, 0.00000734, -0.00000242, -0.00000878, -0.00000678, 0.00000362, 0.00001612, 0.00002015, 0.00000591, -0.00002960, -0.00007853, -0.00012277, -0.00013988, -0.00011257, -0.00003806, 0.00006730, 0.00017093, 0.00023442, 0.00022812, 0.00014408, 0.00000176, -0.00015657, -0.00027990, -0.00032694, -0.00028185, -0.00016071, -0.00000598, 0.00012936, 0.00020162, 0.00019247, 0.00011469, 0.00000547, -0.00008990, -0.00013693, -0.00012576, -0.00007416, -0.00001866, 0.00000216, -0.00003485, -0.00012647, -0.00024338, -0.00034112, -0.00037738, -0.00032841, -0.00019821, -0.00001710, 0.00016927, 0.00031582, 0.00039252, 0.00039248, 0.00033107, 0.00023777, 0.00014435, 0.00007449, 0.00003808, 0.00003171, 0.00004369, 0.00006092, 0.00007438, 0.00008147, 0.00008498, 0.00008982, 0.00009967, 0.00011495, 0.00013280, 0.00014862, 0.00015821, 0.00015929, 0.00015194, 0.00013790, 0.00011948, 0.00009855, 0.00007614, 0.00005275, 0.00002882, 0.00000519, -0.00001686, -0.00003591, -0.00005088, -0.00006138, -0.00006770, -0.00007058, -0.00007086, -0.00006919, },
		{ 0.00000377, -0.00000960, -0.00002123, -0.00003083, -0.00003811, -0.00004295, -0.00004548, -0.00004612, -0.00004546, -0.00004406, -0.00004232, -0.00004048, -0.00003858, -0.00003660, -0.00003444, -0.00003188, -0.00002861, -0.00002426, -0.00001868, -0.00001214, -0.00000545, 0.00000013, 0.00000333, 0.00000345, 0.00000074, -0.00000355, -0.00000763, -0.00000997, -0.00001002, -0.00000847, -0.00000686, -0.00000669, -0.00000844, -0.00001116, -0.00001284, -0.00001151, -0.00000653, 0.00000070, 0.00000707, 0.00000915, 0.00000504, -0.00000437, -0.00001552, -0.00002367, -0.00002541, -0.00002062, -0.00001288, -0.00000776, -0.00000962, -0.00001846, -0.00002875, -0.00003127, -0.00001757, 0.00001463, 0.00005802, 0.00009634, 0.00010980, 0.00008356, 0.00001582, -0.00007833, -0.00016980, -0.00022501, -0.00021872, -0.00014513, -0.00002236, 0.00011218, 0.00021448, 0.00024995, 0.00020660, 0.00009981, -0.00003336, -0.00014819, -0.00020923, -0.00020308, -0.00014213, -0.00005783, 0.00001342, 0.00004554, 0.00003282, -0.00000939, -0.00005335, -0.00007279, -0.00005562, -0.00001000, 0.00003908, 0.00006106, 0.00003392, -0.00004493, -0.00015582, -0.00026326, -0.00032894, -0.00032608, -0.00024928, -0.00011630, 0.00003873, 0.00017746, 0.00027005, 0.00030367, 0.00028411, 0.00023080, 0.00016792, 0.00011549, 0.00008372, 0.00007212, 0.00007276, 0.00007563, 0.00007341, 0.00006398, 0.00005007, 0.00003696, 0.00002950, 0.00003010, 0.00003820, 0.00005111, 0.00006553, 0.00007886, 0.00008974, 0.00009787, 0.00010336, 0.00010613, 0.00010575, 0.00010167, 0.00009359, 0.00008180, 0.00006721, 0.00005105, 0.00003454, 0.00001859, },
		{ 0.00007279, 0.00007011, 0.00006517, 0.00005819, 0.00004944, 0.00003936, 0.00002863, 0.00001808, 0.00000853, 0.00000055, -0.00000570, -0.00001041, -0.00001400, -0.00001688, -0.00001939, -0.00002173, -0.00002406, -0.00002651, -0.00002910, -0.00003165, -0.00003361, -0.00003423, -0.00003276, -0.00002896, -0.00002331, -0.00001709, -0.00001184, -0.00000878, -0.00000813, -0.00000897, -0.00000966, -0.00000871, -0.00000570, -0.00000165, 0.00000133, 0.00000119, -0.00000287, -0.00000960, -0.00001608, -0.00001895, -0.00001624, -0.00000853, 0.00000099, 0.00000800, 0.00000923, 0.00000444, -0.00000320, -0.00000861, -0.00000772, -0.00000031, 0.00000884, 0.00001156, 0.00000032, -0.00002695, -0.00006386, -0.00009609, -0.00010634, -0.00008184, -0.00002149, 0.00006092, 0.00013948, 0.00018482, 0.00017541, 0.00010718, -0.00000289, -0.00012134, -0.00020972, -0.00023866, -0.00019906, -0.00010566, 0.00000854, 0.00010489, 0.00015367, 0.00014474, 0.00008991, 0.00001648, -0.00004507, -0.00007393, -0.00006661, -0.00003666, -0.00000672, 0.00000301, -0.00001551, -0.00005447, -0.00009436, -0.00011372, -0.00009990, -0.00005590, -0.00000003, 0.00004161, 0.00004648, 0.00000584, -0.00007058, -0.00015749, -0.00022341, -0.00024256, -0.00020442, -0.00011728, -0.00000471, 0.00010337, 0.00018119, 0.00021520, 0.00020684, 0.00016973, 0.00012300, 0.00008382, 0.00006198, 0.00005831, 0.00006660, 0.00007780, 0.00008421, 0.00008213, 0.00007221, 0.00005805, 0.00004393, 0.00003303, 0.00002659, 0.00002422, 0.00002482, 0.00002738, 0.00003140, 0.00003683, 0.00004358, 0.00005124, 0.00005897, 0.00006573, 0.00007059, 0.00007300, },
		{ 0.00002611, 0.00002991, 0.00003429, 0.00003888, 0.00004332, 0.00004717, 0.00004984, 0.00005072, 0.00004936, 0.00004571, 0.00004022, 0.00003369, 0.00002702, 0.00002092, 0.00001575, 0.00001155, 0.00000819, 0.00000554, 0.00000355, 0.00000208, 0.00000090, -0.00000048, -0.00000255, -0.00000559, -0.00000936, -0.00001313, -0.00001592, -0.00001706, -0.00001663, -0.00001558, -0.00001532, -0.00001694, -0.00002046, -0.00002453, -0.00002690, -0.00002548, -0.00001957, -0.00001050, -0.00000132, 0.00000448, 0.00000470, -0.00000039, -0.00000795, -0.00001391, -0.00001513, -0.00001122, -0.00000494, -0.00000094, -0.00000299, -0.00001136, -0.00002168, -0.00002643, -0.00001853, 0.00000422, 0.00003632, 0.00006510, 0.00007509, 0.00005482, 0.00000321, -0.00006755, -0.00013464, -0.00017229, -0.00016190, -0.00010052, -0.00000370, 0.00009876, 0.00017321, 0.00019457, 0.00015593, 0.00007131, -0.00002975, -0.00011353, -0.00015485, -0.00014601, -0.00009829, -0.00003589, 0.00001509, 0.00003750, 0.00002929, 0.00000263, -0.00002333, -0.00003226, -0.00001850, 0.00001059, 0.00003876, 0.00004918, 0.00003297, -0.00000614, -0.00005327, -0.00008925, -0.00009959, -0.00008153, -0.00004569, -0.00001160, 0.00000098, -0.00001894, -0.00006813, -0.00012975, -0.00017972, -0.00019611, -0.00016769, -0.00009826, -0.00000506, 0.00008770, 0.00015744, 0.00019052, 0.00018564, 0.00015259, 0.00010726, 0.00006542, 0.00003766, 0.00002723, 0.00003084, 0.00004164, 0.00005265, 0.00005929, 0.00006018, 0.00005650, 0.00005054, 0.00004429, 0.00003881, 0.00003423, 0.00003027, 0.00002677, 0.00002394, 0.00002220, 0.00002195, 0.00002333, },
		{ 0.00002482, 0.00002280, 0.00002088, 0.00001920, 0.00001785, 0.00001707, 0.00001720, 0.00001859, 0.00002141, 0.00002541, 0.00002989, 0.00003391, 0.00003657, 0.00003736, 0.00003629, 0.00003379, 0.00003041, 0.00002659, 0.00002250, 0.00001816, 0.00001366, 0.00000930, 0.00000562, 0.00000309, 0.00000191, 0.00000180, 0.00000215, 0.00000242, 0.00000245, 0.00000267, 0.00000379, 0.00000623, 0.00000955, 0.00001228, 0.00001234, 0.00000805, -0.00000082, -0.00001251, -0.00002370, -0.00003078, -0.00003148, -0.00002608, -0.00001741, -0.00000955, -0.00000576, -0.00000673, -0.00001007, -0.00001159, -0.00000778, 0.00000163, 0.00001261, 0.00001810, 0.00001145, -0.00000955, -0.00003999, -0.00006825, -0.00008004, -0.00006459, -0.00002055, 0.00004139, 0.00010083, 0.00013467, 0.00012622, 0.00007281, -0.00001170, -0.00010089, -0.00016506, -0.00018240, -0.00014732, -0.00007270, 0.00001518, 0.00008681, 0.00012064, 0.00011072, 0.00006764, 0.00001281, -0.00003127, -0.00005035, -0.00004347, -0.00002169, -0.00000154, 0.00000347, -0.00001070, -0.00003709, -0.00006169, -0.00007071, -0.00005759, -0.00002634, 0.00001024, 0.00003664, 0.00004191, 0.00002482, -0.00000572, -0.00003509, -0.00005005, -0.00004522, -0.00002588, -0.00000565, -0.00000015, -0.00001953, -0.00006310, -0.00011866, -0.00016675, -0.00018806, -0.00017096, -0.00011597, -0.00003558, 0.00005032, 0.00012139, 0.00016345, 0.00017214, 0.00015291, 0.00011757, 0.00007943, 0.00004880, 0.00003060, 0.00002437, 0.00002617, 0.00003109, 0.00003535, 0.00003723, 0.00003681, 0.00003512, 0.00003317, 0.00003147, 0.00003001, 0.00002851, 0.00002678, },
		{ 0.00002336, 0.00002200, 0.00002116, 0.00002081, 0.00002079, 0.00002081, 0.00002043, 0.00001923, 0.00001700, 0.00001395, 0.00001077, 0.00000834, 0.00000743, 0.00000832, 0.00001073, 0.00001395, 0.00001726, 0.00002017, 0.00002265, 0.00002488, 0.00002701, 0.00002885, 0.00002986, 0.00002936, 0.00002690, 0.00002257, 0.00001697, 0.00001099, 0.00000532, 0.00000023, -0.00000437, -0.00000854, -0.00001189, -0.00001334, -0.00001156, -0.00000563, 0.00000403, 0.00001535, 0.00002500, 0.00002964, 0.00002743, 0.00001899, 0.00000740, -0.00000317, -0.00000938, -0.00001050, -0.00000874, -0.00000815, -0.00001214, -0.00002108, -0.00003128, -0.00003616, -0.00002944, -0.00000899, 0.00002066, 0.00004869, 0.00006169, 0.00004945, 0.00001047, -0.00004563, -0.00010021, -0.00013203, -0.00012569, -0.00007852, -0.00000294, 0.00007719, 0.00013508, 0.00015105, 0.00012008, 0.00005372, -0.00002457, -0.00008856, -0.00011929, -0.00011166, -0.00007515, -0.00002859, 0.00000853, 0.00002417, 0.00001793, -0.00000037, -0.00001651, -0.00001935, -0.00000618, 0.00001640, 0.00003609, 0.00004140, 0.00002739, -0.00000173, -0.00003462, -0.00005819, -0.00006377, -0.00005099, -0.00002771, -0.00000595, 0.00000402, -0.00000142, -0.00001768, -0.00003452, -0.00004142, -0.00003316, -0.00001286, 0.00000897, 0.00001904, 0.00000747, -0.00002714, -0.00007632, -0.00012435, -0.00015407, -0.00015332, -0.00011945, -0.00006010, 0.00000976, 0.00007321, 0.00011694, 0.00013496, 0.00012922, 0.00010757, 0.00008003, 0.00005523, 0.00003810, 0.00002950, 0.00002735, 0.00002841, 0.00002988, 0.00003023, 0.00002921, 0.00002731, 0.00002520, },
		{ 0.00002457, 0.00002455, 0.00002342, 0.00002140, 0.00001893, 0.00001656, 0.00001491, 0.00001451, 0.00001551, 0.00001760, 0.00001994, 0.00002152, 0.00002153, 0.00001970, 0.00001643, 0.00001255, 0.00000890, 0.00000595, 0.00000369, 0.00000180, 0.00000004, -0.00000145, -0.00000204, -0.00000094, 0.00000240, 0.00000786, 0.00001459, 0.00002129, 0.00002675, 0.00003026, 0.00003171, 0.00003131, 0.00002920, 0.00002512, 0.00001855, 0.00000925, -0.00000209, -0.00001355, -0.00002231, -0.00002574, -0.00002259, -0.00001388, -0.00000272, 0.00000689, 0.00001197, 0.00001199, 0.00000929, 0.00000782, 0.00001083, 0.00001863, 0.00002762, 0.00003144, 0.00002406, 0.00000338, -0.00002635, -0.00005491, -0.00006958, -0.00006063, -0.00002640, 0.00002429, 0.00007423, 0.00010370, 0.00009837, 0.00005570, -0.00001294, -0.00008564, -0.00013780, -0.00015139, -0.00012187, -0.00005987, 0.00001309, 0.00007302, 0.00010253, 0.00009685, 0.00006438, 0.00002194, -0.00001307, -0.00002984, -0.00002785, -0.00001547, -0.00000477, -0.00000501, -0.00001810, -0.00003798, -0.00005409, -0.00005692, -0.00004296, -0.00001659, 0.00001197, 0.00003148, 0.00003490, 0.00002250, 0.00000142, -0.00001806, -0.00002759, -0.00002460, -0.00001328, -0.00000205, 0.00000092, -0.00000793, -0.00002561, -0.00004401, -0.00005391, -0.00004974, -0.00003267, -0.00001058, 0.00000522, 0.00000509, -0.00001444, -0.00004865, -0.00008609, -0.00011286, -0.00011789, -0.00009706, -0.00005454, -0.00000100, 0.00005037, 0.00008833, 0.00010689, 0.00010630, 0.00009174, 0.00007073, 0.00005024, 0.00003476, 0.00002577, 0.00002234, 0.00002233, 0.00002358, },
		{ 0.00002598, 0.00001999, 0.00001778, 0.00001796, 0.00001914, 0.00002017, 0.00002016, 0.00001862, 0.00001555, 0.00001156, 0.00000769, 0.00000511, 0.00000465, 0.00000642, 0.00000978, 0.00001362, 0.00001686, 0.00001890, 0.00001975, 0.00001983, 0.00001957, 0.00001895, 0.00001748, 0.00001442, 0.00000933, 0.00000251, -0.00000496, -0.00001151, -0.00001571, -0.00001679, -0.00001489, -0.00001075, -0.00000522, 0.00000121, 0.00000850, 0.00001669, 0.00002532, 0.00003305, 0.00003787, 0.00003784, 0.00003217, 0.00002190, 0.00000983, -0.00000059, -0.00000683, -0.00000861, -0.00000814, -0.00000896, -0.00001386, -0.00002274, -0.00003193, -0.00003527, -0.00002710, -0.00000564, 0.00002479, 0.00005417, 0.00007018, 0.00006337, 0.00003200, -0.00001578, -0.00006381, -0.00009359, -0.00009160, -0.00005532, 0.00000484, 0.00006868, 0.00011359, 0.00012312, 0.00009336, 0.00003440, -0.00003367, -0.00008861, -0.00011445, -0.00010694, -0.00007403, -0.00003143, 0.00000457, 0.00002368, 0.00002498, 0.00001565, 0.00000629, 0.00000513, 0.00001400, 0.00002786, 0.00003787, 0.00003628, 0.00002062, -0.00000485, -0.00003104, -0.00004838, -0.00005126, -0.00004047, -0.00002260, -0.00000660, 0.00000059, -0.00000282, -0.00001294, -0.00002256, -0.00002507, -0.00001794, -0.00000409, 0.00000964, 0.00001595, 0.00001083, -0.00000432, -0.00002325, -0.00003787, -0.00004218, -0.00003531, -0.00002212, -0.00001101, -0.00001007, -0.00002316, -0.00004782, -0.00007595, -0.00009683, -0.00010128, -0.00008521, -0.00005102, -0.00000660, 0.00003764, 0.00007233, 0.00009182, 0.00009534, 0.00008618, 0.00006987, 0.00005201, 0.00003671, },
		{ 0.00007698, 0.00006692, 0.00005351, 0.00004009, 0.00002890, 0.00002103, 0.00001675, 0.00001569, 0.00001703, 0.00001950, 0.00002164, 0.00002212, 0.00002021, 0.00001605, 0.00001062, 0.00000535, 0.00000150, -0.00000031, -0.00000031, 0.00000071, 0.00000198, 0.00000317, 0.00000458, 0.00000685, 0.00001043, 0.00001512, 0.00001987, 0.00002321, 0.00002385, 0.00002129, 0.00001604, 0.00000935, 0.00000253, -0.00000361, -0.00000897, -0.00001389, -0.00001852, -0.00002229, -0.00002389, -0.00002184, -0.00001540, -0.00000531, 0.00000620, 0.00001620, 0.00002249, 0.00002469, 0.00002459, 0.00002508, 0.00002836, 0.00003412, 0.00003889, 0.00003717, 0.00002415, -0.00000118, -0.00003412, -0.00006473, -0.00008114, -0.00007444, -0.00004322, 0.00000450, 0.00005312, 0.00008489, 0.00008680, 0.00005627, 0.00000288, -0.00005465, -0.00009539, -0.00010417, -0.00007753, -0.00002498, 0.00003483, 0.00008148, 0.00010056, 0.00008866, 0.00005372, 0.00001077, -0.00002477, -0.00004325, -0.00004379, -0.00003316, -0.00002139, -0.00001643, -0.00002047, -0.00002948, -0.00003593, -0.00003315, -0.00001905, 0.00000264, 0.00002413, 0.00003737, 0.00003784, 0.00002660, 0.00000957, -0.00000552, -0.00001292, -0.00001144, -0.00000471, 0.00000108, 0.00000053, -0.00000807, -0.00002183, -0.00003475, -0.00004079, -0.00003693, -0.00002475, -0.00000967, 0.00000171, 0.00000479, -0.00000091, -0.00001160, -0.00002131, -0.00002501, -0.00002121, -0.00001288, -0.00000607, -0.00000704, -0.00001905, -0.00004056, -0.00006538, -0.00008492, -0.00009151, -0.00008121, -0.00005522, -0.00001928, 0.00001851, 0.00005043, 0.00007133, 0.00007964, },
		{ 0.00000531, 0.00003191, 0.00005137, 0.00006195, 0.00006408, 0.00005960, 0.00005082, 0.00003998, 0.00002899, 0.00001943, 0.00001260, 0.00000930, 0.00000957, 0.00001257, 0.00001673, 0.00002022, 0.00002166, 0.00002060, 0.00001764, 0.00001400, 0.00001088, 0.00000882, 0.00000751, 0.00000609, 0.00000376, 0.00000039, -0.00000330, -0.00000600, -0.00000655, -0.00000452, -0.00000050, 0.00000422, 0.00000829, 0.00001100, 0.00001253, 0.00001366, 0.00001506, 0.00001665, 0.00001739, 0.00001578, 0.00001077, 0.00000255, -0.00000720, -0.00001593, -0.00002147, -0.00002313, -0.00002214, -0.00002083, -0.00002114, -0.00002293, -0.00002342, -0.00001811, -0.00000322, 0.00002159, 0.00005155, 0.00007732, 0.00008807, 0.00007603, 0.00004060, -0.00000995, -0.00006041, -0.00009377, -0.00009776, -0.00007007, -0.00001991, 0.00003496, 0.00007479, 0.00008535, 0.00006341, 0.00001781, -0.00003420, -0.00007389, -0.00008827, -0.00007468, -0.00004092, -0.00000122, 0.00003014, 0.00004444, 0.00004134, 0.00002762, 0.00001306, 0.00000541, 0.00000695, 0.00001407, 0.00001984, 0.00001798, 0.00000632, -0.00001208, -0.00003044, -0.00004170, -0.00004188, -0.00003193, -0.00001694, -0.00000356, 0.00000332, 0.00000280, -0.00000202, -0.00000594, -0.00000467, 0.00000285, 0.00001380, 0.00002292, 0.00002524, 0.00001863, 0.00000489, -0.00001102, -0.00002340, -0.00002845, -0.00002597, -0.00001917, -0.00001287, -0.00001096, -0.00001447, -0.00002115, -0.00002683, -0.00002777, -0.00002279, -0.00001415, -0.00000665, -0.00000541, -0.00001337, -0.00002966, -0.00004965, -0.00006660, -0.00007421, -0.00006887, -0.00005093, -0.00002425, },
		{ -0.00006293, -0.00005884, -0.00004543, -0.00002539, -0.00000278, 0.00001845, 0.00003547, 0.00004696, 0.00005280, 0.00005355, 0.00004997, 0.00004292, 0.00003353, 0.00002331, 0.00001409, 0.00000756, 0.00000467, 0.00000521, 0.00000789, 0.00001086, 0.00001253, 0.00001232, 0.00001075, 0.00000915, 0.00000880, 0.00001020, 0.00001278, 0.00001515, 0.00001592, 0.00001438, 0.00001085, 0.00000650, 0.00000260, -0.00000017, -0.00000202, -0.00000383, -0.00000636, -0.00000959, -0.00001241, -0.00001310, -0.00001027, -0.00000376, 0.00000498, 0.00001341, 0.00001908, 0.00002084, 0.00001939, 0.00001679, 0.00001499, 0.00001438, 0.00001300, 0.00000728, -0.00000589, -0.00002672, -0.00005098, -0.00007047, -0.00007577, -0.00006035, -0.00002429, 0.00002427, 0.00007102, 0.00010023, 0.00010085, 0.00007128, 0.00002066, -0.00003405, -0.00007421, -0.00008657, -0.00006829, -0.00002780, 0.00001872, 0.00005394, 0.00006602, 0.00005282, 0.00002188, -0.00001349, -0.00004015, -0.00005042, -0.00004445, -0.00002900, -0.00001348, -0.00000516, -0.00000607, -0.00001266, -0.00001827, -0.00001690, -0.00000644, 0.00001030, 0.00002703, 0.00003714, 0.00003685, 0.00002686, 0.00001176, -0.00000241, -0.00001108, -0.00001316, -0.00001107, -0.00000907, -0.00001069, -0.00001668, -0.00002463, -0.00003024, -0.00002961, -0.00002145, -0.00000780, 0.00000674, 0.00001721, 0.00002053, 0.00001678, 0.00000893, 0.00000119, -0.00000316, -0.00000331, -0.00000104, 0.00000044, -0.00000162, -0.00000787, -0.00001651, -0.00002418, -0.00002776, -0.00002610, -0.00002079, -0.00001548, -0.00001426, -0.00001957, -0.00003104, -0.00004534, -0.00005753, },
		{ -0.00001924, -0.00003023, -0.00004246, -0.00005221, -0.00005623, -0.00005297, -0.00004299, -0.00002850, -0.00001234, 0.00000308, 0.00001634, 0.00002697, 0.00003490, 0.00004000, 0.00004187, 0.00004010, 0.00003487, 0.00002725, 0.00001914, 0.00001267, 0.00000929, 0.00000918, 0.00001109, 0.00001299, 0.00001301, 0.00001038, 0.00000577, 0.00000091, -0.00000228, -0.00000269, -0.00000042, 0.00000334, 0.00000707, 0.00000978, 0.00001143, 0.00001271, 0.00001437, 0.00001647, 0.00001799, 0.00001729, 0.00001299, 0.00000497, -0.00000525, -0.00001495, -0.00002145, -0.00002331, -0.00002110, -0.00001695, -0.00001321, -0.00001090, -0.00000881, -0.00000398, 0.00000650, 0.00002317, 0.00004267, 0.00005803, 0.00006103, 0.00004590, 0.00001279, -0.00003091, -0.00007221, -0.00009692, -0.00009521, -0.00006593, -0.00001778, 0.00003341, 0.00007047, 0.00008141, 0.00006390, 0.00002606, -0.00001700, -0.00004931, -0.00006020, -0.00004810, -0.00002035, 0.00001059, 0.00003265, 0.00003909, 0.00003053, 0.00001368, -0.00000251, -0.00001124, -0.00001078, -0.00000464, 0.00000083, 0.00000006, -0.00000884, -0.00002310, -0.00003676, -0.00004367, -0.00004043, -0.00002792, -0.00001073, 0.00000521, 0.00001531, 0.00001824, 0.00001602, 0.00001250, 0.00001104, 0.00001269, 0.00001577, 0.00001698, 0.00001337, 0.00000415, -0.00000868, -0.00002115, -0.00002915, -0.00003038, -0.00002530, -0.00001679, -0.00000865, -0.00000371, -0.00000267, -0.00000401, -0.00000507, -0.00000371, 0.00000051, 0.00000608, 0.00001037, 0.00001112, 0.00000766, 0.00000128, -0.00000541, -0.00000994, -0.00001125, -0.00001034, -0.00000973, -0.00001219, },
		{ 0.00000272, 0.00000428, 0.00000406, 0.00000020, -0.00000785, -0.00001885, -0.00003007, -0.00003838, -0.00004151, -0.00003888, -0.00003167, -0.00002209, -0.00001230, -0.00000357, 0.00000392, 0.00001067, 0.00001706, 0.00002271, 0.00002651, 0.00002717, 0.00002410, 0.00001801, 0.00001096, 0.00000554, 0.00000380, 0.00000628, 0.00001171, 0.00001755, 0.00002121, 0.00002112, 0.00001739, 0.00001152, 0.00000557, 0.00000107, -0.00000160, -0.00000310, -0.00000439, -0.00000592, -0.00000708, -0.00000653, -0.00000304, 0.00000353, 0.00001177, 0.00001910, 0.00002296, 0.00002206, 0.00001712, 0.00001053, 0.00000500, 0.00000194, 0.00000047, -0.00000228, -0.00000947, -0.00002226, -0.00003820, -0.00005127, -0.00005398, -0.00004082, -0.00001151, 0.00002755, 0.00006465, 0.00008681, 0.00008489, 0.00005765, 0.00001297, -0.00003473, -0.00006970, -0.00008090, -0.00006617, -0.00003291, 0.00000515, 0.00003355, 0.00004282, 0.00003184, 0.00000754, -0.00001863, -0.00003590, -0.00003856, -0.00002779, -0.00001029, 0.00000542, 0.00001310, 0.00001138, 0.00000381, -0.00000349, -0.00000521, 0.00000039, 0.00001073, 0.00002032, 0.00002364, 0.00001789, 0.00000429, -0.00001253, -0.00002680, -0.00003422, -0.00003371, -0.00002749, -0.00001960, -0.00001369, -0.00001128, -0.00001129, -0.00001098, -0.00000775, -0.00000077, 0.00000848, 0.00001683, 0.00002108, 0.00001953, 0.00001287, 0.00000374, -0.00000456, -0.00000963, -0.00001100, -0.00001006, -0.00000908, -0.00000987, -0.00001275, -0.00001645, -0.00001889, -0.00001836, -0.00001451, -0.00000853, -0.00000253, 0.00000155, 0.00000297, 0.00000239, 0.00000139, 0.00000139, },
		{ -0.00001209, -0.00001103, -0.00000880, -0.00000486, 0.00000024, 0.00000479, 0.00000663, 0.00000420, -0.00000251, -0.00001174, -0.00002058, -0.00002611, -0.00002669, -0.00002253, -0.00001545, -0.00000791, -0.00000178, 0.00000244, 0.00000558, 0.00000904, 0.00001367, 0.00001899, 0.00002327, 0.00002427, 0.00002055, 0.00001240, 0.00000202, -0.00000730, -0.00001248, -0.00001204, -0.00000661, 0.00000143, 0.00000913, 0.00001432, 0.00001640, 0.00001625, 0.00001540, 0.00001495, 0.00001481, 0.00001388, 0.00001086, 0.00000526, -0.00000194, -0.00000857, -0.00001232, -0.00001202, -0.00000835, -0.00000363, -0.00000046, -0.00000019, -0.00000184, -0.00000237, 0.00000167, 0.00001189, 0.00002632, 0.00003936, 0.00004370, 0.00003362, 0.00000825, -0.00002690, -0.00006107, -0.00008200, -0.00008077, -0.00005577, -0.00001397, 0.00003132, 0.00006529, 0.00007734, 0.00006502, 0.00003478, -0.00000097, -0.00002895, -0.00004035, -0.00003393, -0.00001574, 0.00000413, 0.00001628, 0.00001597, 0.00000469, -0.00001122, -0.00002402, -0.00002831, -0.00002334, -0.00001294, -0.00000314, 0.00000096, -0.00000225, -0.00001033, -0.00001818, -0.00002076, -0.00001565, -0.00000418, 0.00000926, 0.00001942, 0.00002260, 0.00001829, 0.00000907, -0.00000081, -0.00000763, -0.00000985, -0.00000855, -0.00000651, -0.00000653, -0.00000981, -0.00001540, -0.00002075, -0.00002311, -0.00002098, -0.00001486, -0.00000697, -0.00000007, 0.00000385, 0.00000443, 0.00000289, 0.00000116, 0.00000074, 0.00000189, 0.00000358, 0.00000414, 0.00000231, -0.00000203, -0.00000776, -0.00001309, -0.00001647, -0.00001735, -0.00001629, -0.00001451, -0.00001303, },
		{ -0.00000715, -0.00000916, -0.00001160, -0.00001457, -0.00001743, -0.00001889, -0.00001774, -0.00001366, -0.00000772, -0.00000211, 0.00000069, -0.00000088, -0.00000658, -0.00001432, -0.00002096, -0.00002370, -0.00002130, -0.00001468, -0.00000635, 0.00000081, 0.00000501, 0.00000635, 0.00000666, 0.00000837, 0.00001308, 0.00002047, 0.00002824, 0.00003310, 0.00003235, 0.00002526, 0.00001350, 0.00000054, -0.00000993, -0.00001547, -0.00001588, -0.00001289, -0.00000908, -0.00000643, -0.00000539, -0.00000486, -0.00000310, 0.00000099, 0.00000705, 0.00001330, 0.00001748, 0.00001819, 0.00001571, 0.00001193, 0.00000919, 0.00000871, 0.00000957, 0.00000889, 0.00000335, -0.00000857, -0.00002491, -0.00004014, -0.00004691, -0.00003937, -0.00001636, 0.00001703, 0.00005046, 0.00007190, 0.00007237, 0.00004986, 0.00001070, -0.00003252, -0.00006552, -0.00007778, -0.00006652, -0.00003744, -0.00000216, 0.00002666, 0.00004030, 0.00003700, 0.00002187, 0.00000389, -0.00000838, -0.00001054, -0.00000372, 0.00000662, 0.00001383, 0.00001348, 0.00000539, -0.00000661, -0.00001685, -0.00002077, -0.00001716, -0.00000864, -0.00000015, 0.00000357, 0.00000031, -0.00000860, -0.00001905, -0.00002629, -0.00002714, -0.00002147, -0.00001199, -0.00000281, 0.00000265, 0.00000319, 0.00000016, -0.00000348, -0.00000482, -0.00000255, 0.00000248, 0.00000781, 0.00001069, 0.00000950, 0.00000450, -0.00000242, -0.00000874, -0.00001257, -0.00001343, -0.00001222, -0.00001055, -0.00000973, -0.00001008, -0.00001086, -0.00001081, -0.00000905, -0.00000557, -0.00000136, 0.00000211, 0.00000366, 0.00000298, 0.00000063, -0.00000233, -0.00000500, },
		{ -0.00000409, -0.00000323, -0.00000231, -0.00000135, -0.00000089, -0.00000181, -0.00000473, -0.00000941, -0.00001460, -0.00001840, -0.00001919, -0.00001651, -0.00001153, -0.00000672, -0.00000474, -0.00000713, -0.00001342, -0.00002111, -0.00002671, -0.00002730, -0.00002199, -0.00001235, -0.00000181, 0.00000596, 0.00000880, 0.00000705, 0.00000337, 0.00000143, 0.00000399, 0.00001153, 0.00002185, 0.00003114, 0.00003569, 0.00003368, 0.00002598, 0.00001564, 0.00000639, 0.00000084, -0.00000061, 0.00000030, 0.00000082, -0.00000118, -0.00000602, -0.00001214, -0.00001702, -0.00001869, -0.00001682, -0.00001285, -0.00000901, -0.00000677, -0.00000565, -0.00000332, 0.00000311, 0.00001499, 0.00003044, 0.00004420, 0.00004931, 0.00004026, 0.00001604, -0.00001824, -0.00005240, -0.00007462, -0.00007604, -0.00005462, -0.00001647, 0.00002606, 0.00005890, 0.00007161, 0.00006120, 0.00003303, -0.00000165, -0.00003041, -0.00004447, -0.00004172, -0.00002674, -0.00000803, 0.00000608, 0.00001095, 0.00000712, -0.00000077, -0.00000687, -0.00000716, -0.00000137, 0.00000719, 0.00001367, 0.00001427, 0.00000827, -0.00000166, -0.00001089, -0.00001516, -0.00001272, -0.00000507, 0.00000380, 0.00000947, 0.00000917, 0.00000301, -0.00000625, -0.00001470, -0.00001922, -0.00001888, -0.00001519, -0.00001106, -0.00000927, -0.00001096, -0.00001516, -0.00001939, -0.00002098, -0.00001839, -0.00001198, -0.00000372, 0.00000378, 0.00000850, 0.00000980, 0.00000844, 0.00000598, 0.00000382, 0.00000256, 0.00000183, 0.00000076, -0.00000134, -0.00000449, -0.00000795, -0.00001064, -0.00001173, -0.00001107, -0.00000921, -0.00000704, -0.00000527, },
		{ -0.00000220, -0.00000368, -0.00000517, -0.00000657, -0.00000739, -0.00000702, -0.00000518, -0.00000231, 0.00000038, 0.00000138, -0.00000033, -0.00000462, -0.00001005, -0.00001443, -0.00001583, -0.00001373, -0.00000944, -0.00000569, -0.00000540, -0.00001017, -0.00001920, -0.00002941, -0.00003656, -0.00003716, -0.00003009, -0.00001732, -0.00000312, 0.00000777, 0.00001223, 0.00001014, 0.00000434, -0.00000075, -0.00000136, 0.00000387, 0.00001324, 0.00002285, 0.00002856, 0.00002799, 0.00002168, 0.00001274, 0.00000529, 0.00000239, 0.00000468, 0.00001018, 0.00001550, 0.00001765, 0.00001554, 0.00001030, 0.00000437, -0.00000020, -0.00000310, -0.00000607, -0.00001170, -0.00002147, -0.00003402, -0.00004479, -0.00004757, -0.00003735, -0.00001331, 0.00001963, 0.00005176, 0.00007185, 0.00007158, 0.00004924, 0.00001101, -0.00003096, -0.00006296, -0.00007486, -0.00006392, -0.00003555, -0.00000092, 0.00002767, 0.00004153, 0.00003855, 0.00002324, 0.00000401, -0.00001074, -0.00001619, -0.00001256, -0.00000416, 0.00000329, 0.00000565, 0.00000221, -0.00000437, -0.00000985, -0.00001082, -0.00000654, 0.00000078, 0.00000721, 0.00000920, 0.00000550, -0.00000227, -0.00001042, -0.00001506, -0.00001399, -0.00000772, 0.00000087, 0.00000800, 0.00001082, 0.00000867, 0.00000319, -0.00000267, -0.00000625, -0.00000646, -0.00000425, -0.00000192, -0.00000189, -0.00000540, -0.00001187, -0.00001924, -0.00002495, -0.00002711, -0.00002529, -0.00002053, -0.00001467, -0.00000945, -0.00000573, -0.00000342, -0.00000172, 0.00000013, 0.00000240, 0.00000472, 0.00000637, 0.00000676, 0.00000576, 0.00000379, 0.00000152, -0.00000052, },
		{ 0.00000025, 0.00000241, 0.00000418, 0.00000539, 0.00000565, 0.00000459, 0.00000218, -0.00000095, -0.00000366, -0.00000476, -0.00000363, -0.00000070, 0.00000257, 0.00000432, 0.00000323, -0.00000067, -0.00000581, -0.00000972, -0.00001024, -0.00000680, -0.00000101, 0.00000385, 0.00000418, -0.00000210, -0.00001432, -0.00002896, -0.00004091, -0.00004567, -0.00004127, -0.00002932, -0.00001431, -0.00000168, 0.00000461, 0.00000380, -0.00000141, -0.00000630, -0.00000649, -0.00000012, 0.00001119, 0.00002313, 0.00003070, 0.00003070, 0.00002316, 0.00001131, -0.00000001, -0.00000666, -0.00000702, -0.00000246, 0.00000378, 0.00000856, 0.00001066, 0.00001131, 0.00001317, 0.00001832, 0.00002631, 0.00003350, 0.00003435, 0.00002412, 0.00000175, -0.00002848, -0.00005769, -0.00007550, -0.00007417, -0.00005223, -0.00001564, 0.00002398, 0.00005364, 0.00006388, 0.00005237, 0.00002464, -0.00000837, -0.00003477, -0.00004637, -0.00004144, -0.00002473, -0.00000473, 0.00001024, 0.00001543, 0.00001117, 0.00000190, -0.00000650, -0.00000967, -0.00000661, 0.00000027, 0.00000682, 0.00000952, 0.00000720, 0.00000153, -0.00000415, -0.00000670, -0.00000488, 0.00000005, 0.00000506, 0.00000700, 0.00000429, -0.00000228, -0.00000993, -0.00001528, -0.00001605, -0.00001205, -0.00000525, 0.00000132, 0.00000504, 0.00000492, 0.00000193, -0.00000165, -0.00000351, -0.00000251, 0.00000084, 0.00000464, 0.00000666, 0.00000542, 0.00000085, -0.00000571, -0.00001225, -0.00001700, -0.00001918, -0.00001910, -0.00001781, -0.00001643, -0.00001557, -0.00001515, -0.00001463, -0.00001342, -0.00001128, -0.00000840, -0.00000526, -0.00000231, },
		{ -0.00001559, -0.00001619, -0.00001601, -0.00001496, -0.00001287, -0.00000971, -0.00000574, -0.00000171, 0.00000137, 0.00000259, 0.00000169, -0.00000067, -0.00000307, -0.00000396, -0.00000245, 0.00000106, 0.00000494, 0.00000706, 0.00000584, 0.00000131, -0.00000468, -0.00000911, -0.00000922, -0.00000403, 0.00000484, 0.00001358, 0.00001771, 0.00001414, 0.00000278, -0.00001309, -0.00002805, -0.00003681, -0.00003658, -0.00002836, -0.00001652, -0.00000680, -0.00000363, -0.00000804, -0.00001715, -0.00002548, -0.00002760, -0.00002072, -0.00000608, 0.00001159, 0.00002619, 0.00003304, 0.00003090, 0.00002223, 0.00001177, 0.00000396, 0.00000080, 0.00000106, 0.00000135, -0.00000151, -0.00000822, -0.00001599, -0.00001956, -0.00001388, 0.00000287, 0.00002740, 0.00005181, 0.00006638, 0.00006365, 0.00004193, 0.00000660, -0.00003153, -0.00006023, -0.00007061, -0.00006051, -0.00003522, -0.00000512, 0.00001864, 0.00002852, 0.00002323, 0.00000757, -0.00001019, -0.00002219, -0.00002422, -0.00001708, -0.00000551, 0.00000446, 0.00000845, 0.00000555, -0.00000166, -0.00000884, -0.00001217, -0.00001019, -0.00000430, 0.00000230, 0.00000639, 0.00000642, 0.00000312, -0.00000104, -0.00000334, -0.00000235, 0.00000139, 0.00000563, 0.00000771, 0.00000598, 0.00000069, -0.00000613, -0.00001163, -0.00001362, -0.00001152, -0.00000664, -0.00000141, 0.00000182, 0.00000189, -0.00000070, -0.00000415, -0.00000640, -0.00000612, -0.00000335, 0.00000064, 0.00000402, 0.00000533, 0.00000412, 0.00000103, -0.00000264, -0.00000562, -0.00000725, -0.00000768, -0.00000759, -0.00000780, -0.00000876, -0.00001044, -0.00001246, -0.00001429, },
		{ -0.00000050, -0.00000073, -0.00000185, -0.00000378, -0.00000635, -0.00000936, -0.00001239, -0.00001478, -0.00001582, -0.00001500, -0.00001245, -0.00000899, -0.00000595, -0.00000456, -0.00000531, -0.00000755, -0.00000969, -0.00000989, -0.00000711, -0.00000177, 0.00000416, 0.00000798, 0.00000755, 0.00000256, -0.00000504, -0.00001168, -0.00001375, -0.00000938, 0.00000041, 0.00001182, 0.00001988, 0.00002061, 0.00001308, 0.00000002, -0.00001320, -0.00002092, -0.00001984, -0.00001071, 0.00000189, 0.00001146, 0.00001263, 0.00000366, -0.00001259, -0.00002987, -0.00004127, -0.00004234, -0.00003297, -0.00001734, -0.00000175, 0.00000838, 0.00001106, 0.00000838, 0.00000512, 0.00000590, 0.00001241, 0.00002217, 0.00002947, 0.00002829, 0.00001564, -0.00000623, -0.00003005, -0.00004620, -0.00004688, -0.00002968, 0.00000088, 0.00003476, 0.00006031, 0.00006879, 0.00005789, 0.00003241, 0.00000208, -0.00002255, -0.00003432, -0.00003198, -0.00002001, -0.00000610, 0.00000260, 0.00000245, -0.00000542, -0.00001624, -0.00002425, -0.00002551, -0.00001961, -0.00000953, 0.00000009, 0.00000528, 0.00000458, -0.00000056, -0.00000677, -0.00001061, -0.00001025, -0.00000620, -0.00000078, 0.00000330, 0.00000433, 0.00000245, -0.00000061, -0.00000262, -0.00000213, 0.00000074, 0.00000436, 0.00000653, 0.00000564, 0.00000161, -0.00000407, -0.00000906, -0.00001128, -0.00000995, -0.00000588, -0.00000103, 0.00000247, 0.00000330, 0.00000149, -0.00000174, -0.00000467, -0.00000597, -0.00000531, -0.00000334, -0.00000130, -0.00000027, -0.00000070, -0.00000225, -0.00000404, -0.00000517, -0.00000516, -0.00000410, -0.00000256, -0.00000118, },
		{ -0.00000822, -0.00000757, -0.00000606, -0.00000407, -0.00000194, 0.00000003, 0.00000150, 0.00000211, 0.00000151, -0.00000041, -0.00000331, -0.00000640, -0.00000861, -0.00000915, -0.00000791, -0.00000571, -0.00000402, -0.00000430, -0.00000711, -0.00001171, -0.00001611, -0.00001795, -0.00001567, -0.00000945, -0.00000146, 0.00000502, 0.00000703, 0.00000347, -0.00000416, -0.00001217, -0.00001638, -0.00001409, -0.00000558, 0.00000569, 0.00001457, 0.00001654, 0.00001004, -0.00000249, -0.00001542, -0.00002237, -0.00001916, -0.00000594, 0.00001254, 0.00002889, 0.00003622, 0.00003125, 0.00001580, -0.00000403, -0.00002065, -0.00002847, -0.00002637, -0.00001807, -0.00001007, -0.00000823, -0.00001458, -0.00002610, -0.00003588, -0.00003656, -0.00002408, -0.00000020, 0.00002758, 0.00004896, 0.00005509, 0.00004263, 0.00001538, -0.00001694, -0.00004260, -0.00005254, -0.00004389, -0.00002086, 0.00000731, 0.00003035, 0.00004114, 0.00003823, 0.00002567, 0.00001067, 0.00000000, -0.00000287, 0.00000104, 0.00000740, 0.00001102, 0.00000854, 0.00000000, -0.00001142, -0.00002115, -0.00002550, -0.00002336, -0.00001655, -0.00000871, -0.00000346, -0.00000267, -0.00000581, -0.00001045, -0.00001369, -0.00001366, -0.00001030, -0.00000525, -0.00000076, 0.00000148, 0.00000122, -0.00000041, -0.00000162, -0.00000108, 0.00000133, 0.00000445, 0.00000647, 0.00000591, 0.00000248, -0.00000273, -0.00000774, -0.00001062, -0.00001036, -0.00000730, -0.00000290, 0.00000101, 0.00000307, 0.00000293, 0.00000124, -0.00000078, -0.00000210, -0.00000229, -0.00000168, -0.00000105, -0.00000115, -0.00000231, -0.00000424, -0.00000630, -0.00000778, },
		{ -0.00000077, -0.00000236, -0.00000422, -0.00000594, -0.00000723, -0.00000794, -0.00000805, -0.00000752, -0.00000631, -0.00000446, -0.00000230, -0.00000037, 0.00000064, 0.00000032, -0.00000121, -0.00000315, -0.00000433, -0.00000375, -0.00000125, 0.00000221, 0.00000482, 0.00000470, 0.00000089, -0.00000594, -0.00001352, -0.00001886, -0.00001959, -0.00001524, -0.00000765, -0.00000029, 0.00000327, 0.00000118, -0.00000565, -0.00001371, -0.00001849, -0.00001667, -0.00000797, 0.00000434, 0.00001472, 0.00001782, 0.00001115, -0.00000338, -0.00001999, -0.00003131, -0.00003173, -0.00002005, -0.00000026, 0.00002005, 0.00003303, 0.00003408, 0.00002405, 0.00000878, -0.00000359, -0.00000664, 0.00000113, 0.00001530, 0.00002738, 0.00002873, 0.00001485, -0.00001205, -0.00004342, -0.00006768, -0.00007495, -0.00006139, -0.00003097, 0.00000599, 0.00003685, 0.00005172, 0.00004712, 0.00002697, 0.00000054, -0.00002169, -0.00003228, -0.00002940, -0.00001683, -0.00000157, 0.00000960, 0.00001311, 0.00000965, 0.00000323, -0.00000131, -0.00000081, 0.00000474, 0.00001243, 0.00001813, 0.00001865, 0.00001332, 0.00000419, -0.00000505, -0.00001097, -0.00001195, -0.00000883, -0.00000424, -0.00000120, -0.00000159, -0.00000538, -0.00001081, -0.00001540, -0.00001730, -0.00001604, -0.00001266, -0.00000897, -0.00000650, -0.00000575, -0.00000604, -0.00000602, -0.00000457, -0.00000148, 0.00000234, 0.00000533, 0.00000609, 0.00000411, 0.00000007, -0.00000444, -0.00000769, -0.00000857, -0.00000702, -0.00000397, -0.00000079, 0.00000134, 0.00000194, 0.00000132, 0.00000025, -0.00000049, -0.00000056, -0.00000013, 0.00000028, 0.00000015, },
		{ 0.00000056, 0.00000026, 0.00000000, -0.00000043, -0.00000112, -0.00000196, -0.00000281, -0.00000360, -0.00000438, -0.00000520, -0.00000602, -0.00000659, -0.00000657, -0.00000574, -0.00000422, -0.00000262, -0.00000174, -0.00000219, -0.00000394, -0.00000610, -0.00000723, -0.00000603, -0.00000209, 0.00000366, 0.00000906, 0.00001164, 0.00000978, 0.00000362, -0.00000470, -0.00001190, -0.00001497, -0.00001274, -0.00000655, 0.00000015, 0.00000344, 0.00000084, -0.00000719, -0.00001719, -0.00002415, -0.00002386, -0.00001515, -0.00000085, 0.00001319, 0.00002058, 0.00001745, 0.00000446, -0.00001308, -0.00002743, -0.00003186, -0.00002387, -0.00000665, 0.00001215, 0.00002379, 0.00002254, 0.00000865, -0.00001138, -0.00002738, -0.00003004, -0.00001532, 0.00001329, 0.00004564, 0.00006890, 0.00007279, 0.00005395, 0.00001752, -0.00002473, -0.00005904, -0.00007493, -0.00006903, -0.00004592, -0.00001577, 0.00001012, 0.00002374, 0.00002304, 0.00001191, -0.00000234, -0.00001263, -0.00001509, -0.00001025, -0.00000206, 0.00000455, 0.00000622, 0.00000266, -0.00000355, -0.00000859, -0.00000946, -0.00000548, 0.00000148, 0.00000810, 0.00001136, 0.00001004, 0.00000520, -0.00000048, -0.00000412, -0.00000408, -0.00000068, 0.00000410, 0.00000768, 0.00000817, 0.00000522, 0.00000002, -0.00000537, -0.00000918, -0.00001065, -0.00001027, -0.00000931, -0.00000900, -0.00000981, -0.00001124, -0.00001219, -0.00001162, -0.00000923, -0.00000566, -0.00000222, -0.00000021, -0.00000032, -0.00000226, -0.00000493, -0.00000696, -0.00000737, -0.00000594, -0.00000326, -0.00000035, 0.00000183, 0.00000281, 0.00000267, 0.00000192, 0.00000111, },
		{ 0.00000172, 0.00000325, 0.00000394, 0.00000403, 0.00000365, 0.00000288, 0.00000173, 0.00000035, -0.00000100, -0.00000204, -0.00000264, -0.00000290, -0.00000313, -0.00000363, -0.00000446, -0.00000530, -0.00000564, -0.00000505, -0.00000355, -0.00000178, -0.00000077, -0.00000141, -0.00000386, -0.00000726, -0.00000991, -0.00001003, -0.00000671, -0.00000055, 0.00000632, 0.00001105, 0.00001142, 0.00000699, -0.00000045, -0.00000752, -0.00001080, -0.00000850, -0.00000157, 0.00000651, 0.00001124, 0.00000932, 0.00000039, -0.00001236, -0.00002351, -0.00002777, -0.00002265, -0.00000994, 0.00000479, 0.00001459, 0.00001439, 0.00000368, -0.00001284, -0.00002705, -0.00003107, -0.00002111, 0.00000028, 0.00002470, 0.00004124, 0.00004116, 0.00002208, -0.00001039, -0.00004443, -0.00006651, -0.00006687, -0.00004366, -0.00000391, 0.00003910, 0.00007093, 0.00008152, 0.00006874, 0.00003881, 0.00000331, -0.00002564, -0.00003998, -0.00003826, -0.00002532, -0.00000935, 0.00000195, 0.00000454, -0.00000085, -0.00000986, -0.00001707, -0.00001874, -0.00001436, -0.00000649, 0.00000091, 0.00000456, 0.00000344, -0.00000095, -0.00000563, -0.00000783, -0.00000636, -0.00000216, 0.00000237, 0.00000474, 0.00000364, -0.00000037, -0.00000524, -0.00000848, -0.00000842, -0.00000500, 0.00000035, 0.00000543, 0.00000835, 0.00000842, 0.00000622, 0.00000321, 0.00000084, -0.00000015, -0.00000004, 0.00000022, -0.00000036, -0.00000223, -0.00000502, -0.00000777, -0.00000946, -0.00000962, -0.00000855, -0.00000716, -0.00000647, -0.00000709, -0.00000889, -0.00001107, -0.00001258, -0.00001258, -0.00001083, -0.00000774, -0.00000412, -0.00000078, },
		{ -0.00001615, -0.00001465, -0.00001203, -0.00000889, -0.00000568, -0.00000271, -0.00000016, 0.00000179, 0.00000295, 0.00000317, 0.00000250, 0.00000125, -0.00000007, -0.00000098, -0.00000127, -0.00000113, -0.00000104, -0.00000150, -0.00000266, -0.00000415, -0.00000521, -0.00000513, -0.00000369, -0.00000147, 0.00000032, 0.00000043, -0.00000169, -0.00000542, -0.00000906, -0.00001049, -0.00000833, -0.00000276, 0.00000428, 0.00000974, 0.00001091, 0.00000682, -0.00000107, -0.00000925, -0.00001372, -0.00001187, -0.00000397, 0.00000669, 0.00001513, 0.00001698, 0.00001072, -0.00000131, -0.00001376, -0.00002065, -0.00001832, -0.00000752, 0.00000654, 0.00001629, 0.00001537, 0.00000193, -0.00001985, -0.00004094, -0.00005110, -0.00004344, -0.00001793, 0.00001788, 0.00005126, 0.00006926, 0.00006394, 0.00003582, -0.00000601, -0.00004727, -0.00007390, -0.00007740, -0.00005785, -0.00002343, 0.00001312, 0.00003968, 0.00004924, 0.00004198, 0.00002432, 0.00000550, -0.00000654, -0.00000826, -0.00000136, 0.00000877, 0.00001600, 0.00001631, 0.00000940, -0.00000164, -0.00001218, -0.00001836, -0.00001877, -0.00001473, -0.00000929, -0.00000546, -0.00000477, -0.00000666, -0.00000913, -0.00000994, -0.00000793, -0.00000362, 0.00000113, 0.00000411, 0.00000395, 0.00000078, -0.00000386, -0.00000783, -0.00000944, -0.00000819, -0.00000489, -0.00000115, 0.00000147, 0.00000223, 0.00000143, 0.00000012, -0.00000055, 0.00000003, 0.00000166, 0.00000352, 0.00000468, 0.00000457, 0.00000333, 0.00000160, 0.00000015, -0.00000056, -0.00000069, -0.00000087, -0.00000185, -0.00000407, -0.00000738, -0.00001109, -0.00001426, -0.00001607, },
		{ 0.00000156, -0.00000208, -0.00000586, -0.00000907, -0.00001129, -0.00001237, -0.00001244, -0.00001165, -0.00001019, -0.00000821, -0.00000597, -0.00000387, -0.00000237, -0.00000181, -0.00000222, -0.00000321, -0.00000409, -0.00000424, -0.00000338, -0.00000180, -0.00000025, 0.00000043, -0.00000015, -0.00000170, -0.00000335, -0.00000403, -0.00000317, -0.00000111, 0.00000097, 0.00000159, -0.00000015, -0.00000394, -0.00000813, -0.00001041, -0.00000892, -0.00000339, 0.00000443, 0.00001130, 0.00001392, 0.00001055, 0.00000206, -0.00000814, -0.00001551, -0.00001650, -0.00001034, 0.00000032, 0.00001041, 0.00001482, 0.00001092, 0.00000016, -0.00001223, -0.00001950, -0.00001662, -0.00000308, 0.00001627, 0.00003280, 0.00003768, 0.00002602, -0.00000044, -0.00003328, -0.00006033, -0.00007048, -0.00005831, -0.00002666, 0.00001413, 0.00005008, 0.00006887, 0.00006462, 0.00003992, 0.00000455, -0.00002860, -0.00004866, -0.00005076, -0.00003732, -0.00001630, 0.00000263, 0.00001225, 0.00001048, 0.00000066, -0.00001064, -0.00001686, -0.00001441, -0.00000399, 0.00001013, 0.00002234, 0.00002828, 0.00002657, 0.00001902, 0.00000936, 0.00000134, -0.00000294, -0.00000368, -0.00000278, -0.00000257, -0.00000445, -0.00000820, -0.00001226, -0.00001468, -0.00001421, -0.00001098, -0.00000642, -0.00000246, -0.00000061, -0.00000122, -0.00000344, -0.00000573, -0.00000669, -0.00000576, -0.00000343, -0.00000087, 0.00000071, 0.00000068, -0.00000075, -0.00000269, -0.00000410, -0.00000432, -0.00000333, -0.00000168, -0.00000012, 0.00000088, 0.00000128, 0.00000150, 0.00000201, 0.00000307, 0.00000444, 0.00000553, 0.00000563, 0.00000430, },
		{ 0.00000008, 0.00000216, 0.00000373, 0.00000431, 0.00000380, 0.00000244, 0.00000064, -0.00000119, -0.00000278, -0.00000404, -0.00000490, -0.00000529, -0.00000510, -0.00000431, -0.00000314, -0.00000207, -0.00000171, -0.00000253, -0.00000456, -0.00000727, -0.00000970, -0.00001086, -0.00001026, -0.00000813, -0.00000538, -0.00000319, -0.00000236, -0.00000287, -0.00000392, -0.00000429, -0.00000315, -0.00000057, 0.00000229, 0.00000376, 0.00000255, -0.00000138, -0.00000653, -0.00001033, -0.00001036, -0.00000567, 0.00000249, 0.00001092, 0.00001583, 0.00001460, 0.00000728, -0.00000320, -0.00001211, -0.00001511, -0.00001044, -0.00000009, 0.00001084, 0.00001633, 0.00001235, -0.00000082, -0.00001819, -0.00003184, -0.00003430, -0.00002213, 0.00000203, 0.00002993, 0.00005075, 0.00005544, 0.00004071, 0.00001066, -0.00002449, -0.00005233, -0.00006314, -0.00005367, -0.00002833, 0.00000281, 0.00002818, 0.00003935, 0.00003417, 0.00001707, -0.00000328, -0.00001799, -0.00002168, -0.00001448, -0.00000144, 0.00001013, 0.00001412, 0.00000829, -0.00000499, -0.00002006, -0.00003067, -0.00003267, -0.00002563, -0.00001261, 0.00000146, 0.00001203, 0.00001675, 0.00001607, 0.00001247, 0.00000896, 0.00000754, 0.00000841, 0.00001021, 0.00001096, 0.00000929, 0.00000511, -0.00000037, -0.00000533, -0.00000833, -0.00000893, -0.00000785, -0.00000642, -0.00000585, -0.00000660, -0.00000818, -0.00000951, -0.00000956, -0.00000788, -0.00000489, -0.00000159, 0.00000091, 0.00000192, 0.00000144, 0.00000007, -0.00000139, -0.00000233, -0.00000263, -0.00000256, -0.00000255, -0.00000283, -0.00000331, -0.00000357, -0.00000316, -0.00000188, },
		{ -0.00000136, -0.00000264, -0.00000346, -0.00000354, -0.00000298, -0.00000211, -0.00000136, -0.00000104, -0.00000118, -0.00000160, -0.00000207, -0.00000241, -0.00000259, -0.00000262, -0.00000247, -0.00000196, -0.00000084, 0.00000097, 0.00000324, 0.00000530, 0.00000627, 0.00000539, 0.00000246, -0.00000195, -0.00000665, -0.00001026, -0.00001184, -0.00001134, -0.00000958, -0.00000788, -0.00000730, -0.00000807, -0.00000944, -0.00001005, -0.00000874, -0.00000532, -0.00000084, 0.00000277, 0.00000368, 0.00000122, -0.00000354, -0.00000808, -0.00000958, -0.00000633, 0.00000108, 0.00000980, 0.00001580, 0.00001577, 0.00000891, -0.00000230, -0.00001286, -0.00001745, -0.00001293, -0.00000019, 0.00001575, 0.00002763, 0.00002904, 0.00001746, -0.00000413, -0.00002807, -0.00004490, -0.00004728, -0.00003318, -0.00000700, 0.00002201, 0.00004342, 0.00004979, 0.00003963, 0.00001781, -0.00000663, -0.00002437, -0.00002964, -0.00002239, -0.00000789, 0.00000597, 0.00001234, 0.00000845, -0.00000339, -0.00001708, -0.00002569, -0.00002470, -0.00001396, 0.00000227, 0.00001738, 0.00002528, 0.00002303, 0.00001183, -0.00000366, -0.00001757, -0.00002526, -0.00002509, -0.00001866, -0.00000969, -0.00000213, 0.00000156, 0.00000122, -0.00000134, -0.00000358, -0.00000354, -0.00000077, 0.00000369, 0.00000799, 0.00001048, 0.00001052, 0.00000859, 0.00000591, 0.00000369, 0.00000251, 0.00000209, 0.00000162, 0.00000028, -0.00000219, -0.00000535, -0.00000827, -0.00000996, -0.00000991, -0.00000829, -0.00000582, -0.00000338, -0.00000162, -0.00000071, -0.00000037, -0.00000018, 0.00000016, 0.00000063, 0.00000094, 0.00000075, -0.00000008, },
		{ -0.00000002, 0.00000121, 0.00000193, 0.00000199, 0.00000151, 0.00000083, 0.00000031, 0.00000012, 0.00000015, 0.00000012, -0.00000024, -0.00000107, -0.00000223, -0.00000348, -0.00000457, -0.00000540, -0.00000602, -0.00000649, -0.00000670, -0.00000633, -0.00000493, -0.00000221, 0.00000164, 0.00000584, 0.00000919, 0.00001054, 0.00000928, 0.00000574, 0.00000111, -0.00000303, -0.00000545, -0.00000579, -0.00000478, -0.00000386, -0.00000435, -0.00000673, -0.00001030, -0.00001347, -0.00001456, -0.00001278, -0.00000878, -0.00000441, -0.00000187, -0.00000247, -0.00000577, -0.00000954, -0.00001078, -0.00000725, 0.00000111, 0.00001168, 0.00002003, 0.00002184, 0.00001501, 0.00000112, -0.00001480, -0.00002608, -0.00002721, -0.00001645, 0.00000307, 0.00002424, 0.00003874, 0.00004040, 0.00002790, 0.00000543, -0.00001891, -0.00003640, -0.00004118, -0.00003257, -0.00001503, 0.00000392, 0.00001707, 0.00002056, 0.00001524, 0.00000587, -0.00000144, -0.00000228, 0.00000398, 0.00001389, 0.00002173, 0.00002227, 0.00001341, -0.00000273, -0.00002058, -0.00003357, -0.00003689, -0.00002958, -0.00001485, 0.00000145, 0.00001327, 0.00001688, 0.00001216, 0.00000231, -0.00000781, -0.00001391, -0.00001397, -0.00000886, -0.00000154, 0.00000445, 0.00000663, 0.00000450, -0.00000051, -0.00000597, -0.00000963, -0.00001042, -0.00000863, -0.00000553, -0.00000257, -0.00000064, 0.00000023, 0.00000064, 0.00000133, 0.00000266, 0.00000436, 0.00000573, 0.00000604, 0.00000495, 0.00000270, -0.00000001, -0.00000236, -0.00000381, -0.00000427, -0.00000406, -0.00000365, -0.00000337, -0.00000323, -0.00000300, -0.00000241, -0.00000135, },
		{ -0.00000128, -0.00000185, -0.00000191, -0.00000138, -0.00000046, 0.00000048, 0.00000113, 0.00000136, 0.00000134, 0.00000135, 0.00000168, 0.00000233, 0.00000310, 0.00000361, 0.00000355, 0.00000283, 0.00000164, 0.00000026, -0.00000110, -0.00000241, -0.00000382, -0.00000547, -0.00000723, -0.00000860, -0.00000885, -0.00000736, -0.00000404, 0.00000043, 0.00000476, 0.00000749, 0.00000767, 0.00000536, 0.00000164, -0.00000176, -0.00000332, -0.00000244, 0.00000021, 0.00000298, 0.00000408, 0.00000248, -0.00000152, -0.00000636, -0.00001000, -0.00001103, -0.00000949, -0.00000691, -0.00000555, -0.00000717, -0.00001176, -0.00001728, -0.00002030, -0.00001765, -0.00000819, 0.00000618, 0.00002083, 0.00003001, 0.00002935, 0.00001794, -0.00000089, -0.00002056, -0.00003373, -0.00003520, -0.00002414, -0.00000448, 0.00001669, 0.00003192, 0.00003639, 0.00002967, 0.00001555, 0.00000015, -0.00001089, -0.00001474, -0.00001227, -0.00000716, -0.00000372, -0.00000455, -0.00000919, -0.00001440, -0.00001586, -0.00001058, 0.00000138, 0.00001653, 0.00002934, 0.00003457, 0.00002965, 0.00001591, -0.00000190, -0.00001756, -0.00002596, -0.00002512, -0.00001684, -0.00000570, 0.00000303, 0.00000566, 0.00000152, -0.00000698, -0.00001570, -0.00002059, -0.00001948, -0.00001279, -0.00000310, 0.00000608, 0.00001187, 0.00001304, 0.00001020, 0.00000525, 0.00000030, -0.00000319, -0.00000487, -0.00000528, -0.00000535, -0.00000573, -0.00000649, -0.00000719, -0.00000723, -0.00000627, -0.00000443, -0.00000226, -0.00000038, 0.00000075, 0.00000109, 0.00000091, 0.00000063, 0.00000049, 0.00000050, 0.00000046, 0.00000014, -0.00000050, },
		{ -0.00000207, -0.00000136, -0.00000113, -0.00000134, -0.00000174, -0.00000195, -0.00000171, -0.00000101, -0.00000010, 0.00000064, 0.00000095, 0.00000084, 0.00000057, 0.00000056, 0.00000109, 0.00000218, 0.00000352, 0.00000468, 0.00000528, 0.00000524, 0.00000475, 0.00000412, 0.00000353, 0.00000289, 0.00000188, 0.00000018, -0.00000217, -0.00000467, -0.00000639, -0.00000643, -0.00000438, -0.00000073, 0.00000324, 0.00000585, 0.00000584, 0.00000298, -0.00000171, -0.00000629, -0.00000879, -0.00000810, -0.00000455, 0.00000023, 0.00000407, 0.00000531, 0.00000367, 0.00000035, -0.00000255, -0.00000325, -0.00000134, 0.00000182, 0.00000362, 0.00000157, -0.00000529, -0.00001543, -0.00002513, -0.00002987, -0.00002635, -0.00001420, 0.00000344, 0.00002084, 0.00003186, 0.00003237, 0.00002200, 0.00000441, -0.00001419, -0.00002745, -0.00003137, -0.00002565, -0.00001349, 0.00000006, 0.00001041, 0.00001522, 0.00001506, 0.00001255, 0.00001064, 0.00001091, 0.00001269, 0.00001356, 0.00001081, 0.00000321, -0.00000794, -0.00001906, -0.00002569, -0.00002460, -0.00001540, -0.00000098, 0.00001353, 0.00002286, 0.00002379, 0.00001651, 0.00000447, -0.00000712, -0.00001359, -0.00001272, -0.00000563, 0.00000388, 0.00001109, 0.00001240, 0.00000683, -0.00000369, -0.00001530, -0.00002384, -0.00002656, -0.00002303, -0.00001503, -0.00000561, 0.00000227, 0.00000683, 0.00000788, 0.00000650, 0.00000428, 0.00000250, 0.00000166, 0.00000149, 0.00000133, 0.00000059, -0.00000087, -0.00000277, -0.00000457, -0.00000585, -0.00000643, -0.00000645, -0.00000618, -0.00000582, -0.00000542, -0.00000487, -0.00000408, -0.00000307, },
		{ -0.00000454, -0.00000477, -0.00000456, -0.00000409, -0.00000365, -0.00000354, -0.00000390, -0.00000458, -0.00000523, -0.00000543, -0.00000497, -0.00000395, -0.00000275, -0.00000185, -0.00000157, -0.00000186, -0.00000235, -0.00000251, -0.00000195, -0.00000063, 0.00000116, 0.00000290, 0.00000418, 0.00000488, 0.00000521, 0.00000553, 0.00000601, 0.00000650, 0.00000654, 0.00000560, 0.00000353, 0.00000079, -0.00000160, -0.00000245, -0.00000102, 0.00000246, 0.00000675, 0.00000997, 0.00001040, 0.00000735, 0.00000156, -0.00000501, -0.00001001, -0.00001174, -0.00000994, -0.00000597, -0.00000209, -0.00000038, -0.00000160, -0.00000475, -0.00000752, -0.00000742, -0.00000323, 0.00000417, 0.00001190, 0.00001623, 0.00001427, 0.00000549, -0.00000774, -0.00002087, -0.00002893, -0.00002858, -0.00001955, -0.00000482, 0.00001053, 0.00002136, 0.00002445, 0.00001961, 0.00000948, -0.00000186, -0.00001068, -0.00001503, -0.00001521, -0.00001306, -0.00001069, -0.00000915, -0.00000800, -0.00000571, -0.00000084, 0.00000676, 0.00001543, 0.00002218, 0.00002400, 0.00001940, 0.00000938, -0.00000269, -0.00001241, -0.00001613, -0.00001267, -0.00000384, 0.00000625, 0.00001296, 0.00001316, 0.00000659, -0.00000400, -0.00001405, -0.00001920, -0.00001713, -0.00000853, 0.00000324, 0.00001358, 0.00001858, 0.00001653, 0.00000846, -0.00000245, -0.00001234, -0.00001813, -0.00001866, -0.00001479, -0.00000876, -0.00000309, 0.00000044, 0.00000132, 0.00000023, -0.00000155, -0.00000283, -0.00000298, -0.00000208, -0.00000069, 0.00000056, 0.00000123, 0.00000122, 0.00000069, -0.00000012, -0.00000106, -0.00000204, -0.00000302, -0.00000391, },
		{ -0.00000168, -0.00000177, -0.00000205, -0.00000233, -0.00000234, -0.00000194, -0.00000121, -0.00000042, 0.00000001, -0.00000027, -0.00000131, -0.00000283, -0.00000430, -0.00000518, -0.00000519, -0.00000448, -0.00000356, -0.00000303, -0.00000333, -0.00000443, -0.00000590, -0.00000707, -0.00000740, -0.00000671, -0.00000526, -0.00000355, -0.00000202, -0.00000082, 0.00000028, 0.00000167, 0.00000357, 0.00000577, 0.00000762, 0.00000832, 0.00000743, 0.00000519, 0.00000264, 0.00000120, 0.00000201, 0.00000530, 0.00001005, 0.00001435, 0.00001612, 0.00001418, 0.00000880, 0.00000176, -0.00000436, -0.00000739, -0.00000659, -0.00000306, 0.00000067, 0.00000187, -0.00000105, -0.00000761, -0.00001530, -0.00002057, -0.00002040, -0.00001381, -0.00000239, 0.00001009, 0.00001922, 0.00002173, 0.00001682, 0.00000645, -0.00000549, -0.00001478, -0.00001858, -0.00001634, -0.00000977, -0.00000183, 0.00000465, 0.00000811, 0.00000857, 0.00000721, 0.00000540, 0.00000383, 0.00000221, -0.00000033, -0.00000445, -0.00000979, -0.00001482, -0.00001734, -0.00001561, -0.00000932, -0.00000009, 0.00000901, 0.00001469, 0.00001492, 0.00000994, 0.00000228, -0.00000429, -0.00000644, -0.00000281, 0.00000527, 0.00001427, 0.00001995, 0.00001921, 0.00001152, -0.00000081, -0.00001351, -0.00002213, -0.00002380, -0.00001838, -0.00000836, 0.00000222, 0.00000947, 0.00001114, 0.00000738, 0.00000038, -0.00000666, -0.00001095, -0.00001112, -0.00000759, -0.00000213, 0.00000301, 0.00000606, 0.00000629, 0.00000411, 0.00000064, -0.00000282, -0.00000533, -0.00000650, -0.00000648, -0.00000567, -0.00000453, -0.00000341, -0.00000250, -0.00000191, },
		{ -0.00000467, -0.00000470, -0.00000461, -0.00000464, -0.00000494, -0.00000547, -0.00000604, -0.00000629, -0.00000589, -0.00000471, -0.00000291, -0.00000098, 0.00000048, 0.00000097, 0.00000038, -0.00000093, -0.00000229, -0.00000301, -0.00000268, -0.00000147, 0.00000003, 0.00000100, 0.00000083, -0.00000060, -0.00000289, -0.00000530, -0.00000717, -0.00000816, -0.00000840, -0.00000835, -0.00000840, -0.00000862, -0.00000863, -0.00000786, -0.00000589, -0.00000283, 0.00000061, 0.00000338, 0.00000463, 0.00000420, 0.00000287, 0.00000206, 0.00000317, 0.00000678, 0.00001219, 0.00001754, 0.00002052, 0.00001946, 0.00001421, 0.00000645, -0.00000091, -0.00000493, -0.00000396, 0.00000149, 0.00000877, 0.00001417, 0.00001444, 0.00000824, -0.00000313, -0.00001615, -0.00002643, -0.00003043, -0.00002677, -0.00001677, -0.00000376, 0.00000823, 0.00001600, 0.00001821, 0.00001551, 0.00000992, 0.00000377, -0.00000123, -0.00000447, -0.00000616, -0.00000681, -0.00000672, -0.00000569, -0.00000339, 0.00000021, 0.00000442, 0.00000783, 0.00000890, 0.00000668, 0.00000150, -0.00000495, -0.00001021, -0.00001211, -0.00000983, -0.00000439, 0.00000168, 0.00000546, 0.00000497, 0.00000021, -0.00000669, -0.00001229, -0.00001341, -0.00000853, 0.00000132, 0.00001290, 0.00002201, 0.00002520, 0.00002119, 0.00001143, -0.00000055, -0.00001059, -0.00001554, -0.00001450, -0.00000894, -0.00000202, 0.00000294, 0.00000373, 0.00000015, -0.00000614, -0.00001240, -0.00001603, -0.00001556, -0.00001109, -0.00000413, 0.00000319, 0.00000887, 0.00001175, 0.00001167, 0.00000931, 0.00000577, 0.00000211, -0.00000093, -0.00000304, -0.00000421, },
		{ 0.00000639, 0.00000418, 0.00000240, 0.00000127, 0.00000074, 0.00000055, 0.00000033, -0.00000027, -0.00000148, -0.00000323, -0.00000518, -0.00000678, -0.00000747, -0.00000698, -0.00000546, -0.00000351, -0.00000191, -0.00000134, -0.00000202, -0.00000359, -0.00000523, -0.00000603, -0.00000542, -0.00000344, -0.00000072, 0.00000179, 0.00000325, 0.00000330, 0.00000220, 0.00000059, -0.00000085, -0.00000177, -0.00000236, -0.00000312, -0.00000456, -0.00000678, -0.00000931, -0.00001132, -0.00001202, -0.00001112, -0.00000906, -0.00000683, -0.00000547, -0.00000539, -0.00000607, -0.00000614, -0.00000403, 0.00000115, 0.00000892, 0.00001734, 0.00002358, 0.00002511, 0.00002085, 0.00001190, 0.00000131, -0.00000696, -0.00000961, -0.00000538, 0.00000436, 0.00001599, 0.00002501, 0.00002774, 0.00002267, 0.00001101, -0.00000387, -0.00001772, -0.00002691, -0.00002957, -0.00002591, -0.00001780, -0.00000786, 0.00000153, 0.00000879, 0.00001327, 0.00001497, 0.00001417, 0.00001124, 0.00000666, 0.00000119, -0.00000405, -0.00000774, -0.00000881, -0.00000696, -0.00000291, 0.00000167, 0.00000485, 0.00000528, 0.00000283, -0.00000123, -0.00000477, -0.00000576, -0.00000333, 0.00000169, 0.00000701, 0.00000986, 0.00000821, 0.00000189, -0.00000717, -0.00001565, -0.00002011, -0.00001851, -0.00001109, -0.00000037, 0.00000989, 0.00001617, 0.00001662, 0.00001174, 0.00000405, -0.00000302, -0.00000663, -0.00000566, -0.00000110, 0.00000454, 0.00000832, 0.00000813, 0.00000356, -0.00000404, -0.00001218, -0.00001824, -0.00002040, -0.00001823, -0.00001265, -0.00000543, 0.00000149, 0.00000665, 0.00000938, 0.00000978, 0.00000851, },
		{ 0.00000127, 0.00000415, 0.00000517, 0.00000464, 0.00000322, 0.00000163, 0.00000050, 0.00000019, 0.00000073, 0.00000184, 0.00000298, 0.00000356, 0.00000315, 0.00000169, -0.00000045, -0.00000256, -0.00000387, -0.00000390, -0.00000271, -0.00000090, 0.00000059, 0.00000086, -0.00000046, -0.00000308, -0.00000610, -0.00000841, -0.00000917, -0.00000820, -0.00000599, -0.00000351, -0.00000166, -0.00000091, -0.00000109, -0.00000152, -0.00000146, -0.00000056, 0.00000098, 0.00000240, 0.00000289, 0.00000198, -0.00000016, -0.00000275, -0.00000488, -0.00000599, -0.00000624, -0.00000646, -0.00000770, -0.00001050, -0.00001435, -0.00001769, -0.00001837, -0.00001469, -0.00000633, 0.00000516, 0.00001667, 0.00002455, 0.00002600, 0.00002033, 0.00000934, -0.00000316, -0.00001272, -0.00001585, -0.00001132, -0.00000055, 0.00001292, 0.00002477, 0.00003137, 0.00003094, 0.00002392, 0.00001253, -0.00000010, -0.00001107, -0.00001835, -0.00002116, -0.00001975, -0.00001507, -0.00000838, -0.00000105, 0.00000563, 0.00001046, 0.00001257, 0.00001166, 0.00000819, 0.00000338, -0.00000112, -0.00000381, -0.00000394, -0.00000188, 0.00000100, 0.00000296, 0.00000265, -0.00000011, -0.00000418, -0.00000751, -0.00000808, -0.00000490, 0.00000135, 0.00000849, 0.00001362, 0.00001437, 0.00000999, 0.00000176, -0.00000746, -0.00001433, -0.00001646, -0.00001334, -0.00000654, 0.00000095, 0.00000607, 0.00000692, 0.00000353, -0.00000220, -0.00000745, -0.00000965, -0.00000764, -0.00000209, 0.00000485, 0.00001041, 0.00001239, 0.00000995, 0.00000387, -0.00000388, -0.00001088, -0.00001518, -0.00001590, -0.00001330, -0.00000857, -0.00000325, },
		{ -0.00001084, -0.00000898, -0.00000522, -0.00000077, 0.00000314, 0.00000558, 0.00000616, 0.00000500, 0.00000266, -0.00000005, -0.00000230, -0.00000346, -0.00000326, -0.00000191, 0.00000000, 0.00000167, 0.00000238, 0.00000184, 0.00000030, -0.00000148, -0.00000259, -0.00000234, -0.00000062, 0.00000200, 0.00000445, 0.00000564, 0.00000495, 0.00000251, -0.00000086, -0.00000395, -0.00000576, -0.00000593, -0.00000485, -0.00000350, -0.00000288, -0.00000355, -0.00000531, -0.00000735, -0.00000862, -0.00000841, -0.00000667, -0.00000405, -0.00000152, 0.00000014, 0.00000078, 0.00000088, 0.00000126, 0.00000249, 0.00000435, 0.00000583, 0.00000543, 0.00000198, -0.00000458, -0.00001286, -0.00002028, -0.00002395, -0.00002189, -0.00001393, -0.00000204, 0.00001023, 0.00001898, 0.00002137, 0.00001670, 0.00000669, -0.00000517, -0.00001486, -0.00001921, -0.00001695, -0.00000891, 0.00000238, 0.00001370, 0.00002219, 0.00002612, 0.00002516, 0.00002020, 0.00001283, 0.00000486, -0.00000211, -0.00000690, -0.00000891, -0.00000808, -0.00000499, -0.00000069, 0.00000346, 0.00000616, 0.00000659, 0.00000478, 0.00000161, -0.00000145, -0.00000296, -0.00000217, 0.00000061, 0.00000404, 0.00000635, 0.00000609, 0.00000292, -0.00000214, -0.00000703, -0.00000952, -0.00000822, -0.00000328, 0.00000359, 0.00000974, 0.00001267, 0.00001117, 0.00000578, -0.00000133, -0.00000735, -0.00000996, -0.00000835, -0.00000353, 0.00000214, 0.00000598, 0.00000616, 0.00000245, -0.00000369, -0.00000978, -0.00001339, -0.00001311, -0.00000910, -0.00000295, 0.00000299, 0.00000660, 0.00000675, 0.00000363, -0.00000142, -0.00000654, -0.00001001, },
		{ -0.00000238, -0.00000626, -0.00001017, -0.00001265, -0.00001279, -0.00001039, -0.00000600, -0.00000065, 0.00000444, 0.00000818, 0.00000989, 0.00000938, 0.00000702, 0.00000358, 0.00000005, -0.00000259, -0.00000375, -0.00000336, -0.00000191, -0.00000027, 0.00000066, 0.00000034, -0.00000114, -0.00000309, -0.00000448, -0.00000443, -0.00000265, 0.00000041, 0.00000366, 0.00000586, 0.00000619, 0.00000460, 0.00000185, -0.00000078, -0.00000212, -0.00000165, 0.00000026, 0.00000251, 0.00000382, 0.00000327, 0.00000079, -0.00000287, -0.00000651, -0.00000906, -0.00001008, -0.00000989, -0.00000929, -0.00000905, -0.00000935, -0.00000962, -0.00000876, -0.00000579, -0.00000053, 0.00000606, 0.00001200, 0.00001504, 0.00001357, 0.00000749, -0.00000163, -0.00001092, -0.00001727, -0.00001848, -0.00001416, -0.00000582, 0.00000359, 0.00001082, 0.00001349, 0.00001085, 0.00000396, -0.00000480, -0.00001266, -0.00001735, -0.00001777, -0.00001413, -0.00000769, -0.00000020, 0.00000663, 0.00001157, 0.00001400, 0.00001398, 0.00001205, 0.00000912, 0.00000623, 0.00000431, 0.00000394, 0.00000514, 0.00000731, 0.00000936, 0.00001012, 0.00000877, 0.00000531, 0.00000063, -0.00000375, -0.00000625, -0.00000594, -0.00000296, 0.00000143, 0.00000533, 0.00000700, 0.00000567, 0.00000188, -0.00000268, -0.00000589, -0.00000616, -0.00000317, 0.00000194, 0.00000703, 0.00000985, 0.00000907, 0.00000487, -0.00000110, -0.00000643, -0.00000893, -0.00000762, -0.00000311, 0.00000267, 0.00000729, 0.00000886, 0.00000675, 0.00000182, -0.00000400, -0.00000857, -0.00001041, -0.00000922, -0.00000596, -0.00000230, 0.00000004, 0.00000002, },
		{ -0.00000170, 0.00000009, 0.00000076, -0.00000051, -0.00000378, -0.00000833, -0.00001293, -0.00001628, -0.00001739, -0.00001586, -0.00001191, -0.00000633, -0.00000023, 0.00000519, 0.00000892, 0.00001038, 0.00000951, 0.00000685, 0.00000339, 0.00000024, -0.00000169, -0.00000205, -0.00000110, 0.00000034, 0.00000126, 0.00000095, -0.00000066, -0.00000296, -0.00000488, -0.00000538, -0.00000397, -0.00000101, 0.00000243, 0.00000496, 0.00000551, 0.00000383, 0.00000065, -0.00000263, -0.00000452, -0.00000414, -0.00000159, 0.00000212, 0.00000548, 0.00000724, 0.00000687, 0.00000480, 0.00000209, -0.00000014, -0.00000133, -0.00000175, -0.00000236, -0.00000421, -0.00000779, -0.00001259, -0.00001712, -0.00001945, -0.00001807, -0.00001263, -0.00000424, 0.00000485, 0.00001198, 0.00001512, 0.00001362, 0.00000841, 0.00000162, -0.00000426, -0.00000732, -0.00000686, -0.00000352, 0.00000106, 0.00000495, 0.00000667, 0.00000562, 0.00000217, -0.00000259, -0.00000735, -0.00001094, -0.00001272, -0.00001255, -0.00001079, -0.00000807, -0.00000510, -0.00000253, -0.00000074, 0.00000017, 0.00000053, 0.00000091, 0.00000202, 0.00000439, 0.00000805, 0.00001241, 0.00001636, 0.00001857, 0.00001805, 0.00001461, 0.00000902, 0.00000284, -0.00000210, -0.00000448, -0.00000393, -0.00000126, 0.00000192, 0.00000392, 0.00000371, 0.00000139, -0.00000185, -0.00000424, -0.00000432, -0.00000162, 0.00000304, 0.00000785, 0.00001077, 0.00001041, 0.00000666, 0.00000078, -0.00000509, -0.00000884, -0.00000921, -0.00000628, -0.00000139, 0.00000346, 0.00000647, 0.00000675, 0.00000455, 0.00000111, -0.00000203, -0.00000363, -0.00000334, },
		{ 0.00000121, -0.00000010, -0.00000044, 0.00000044, 0.00000214, 0.00000374, 0.00000425, 0.00000295, -0.00000025, -0.00000488, -0.00000999, -0.00001443, -0.00001722, -0.00001770, -0.00001571, -0.00001162, -0.00000621, -0.00000057, 0.00000416, 0.00000708, 0.00000775, 0.00000637, 0.00000371, 0.00000089, -0.00000104, -0.00000149, -0.00000057, 0.00000098, 0.00000212, 0.00000198, 0.00000032, -0.00000231, -0.00000475, -0.00000577, -0.00000462, -0.00000146, 0.00000266, 0.00000615, 0.00000760, 0.00000639, 0.00000297, -0.00000132, -0.00000477, -0.00000612, -0.00000502, -0.00000222, 0.00000088, 0.00000288, 0.00000307, 0.00000175, 0.00000006, -0.00000061, 0.00000063, 0.00000364, 0.00000719, 0.00000944, 0.00000873, 0.00000436, -0.00000297, -0.00001137, -0.00001837, -0.00002184, -0.00002077, -0.00001556, -0.00000786, 0.00000009, 0.00000627, 0.00000950, 0.00000970, 0.00000770, 0.00000479, 0.00000219, 0.00000064, 0.00000028, 0.00000078, 0.00000155, 0.00000205, 0.00000195, 0.00000118, -0.00000014, -0.00000174, -0.00000335, -0.00000472, -0.00000573, -0.00000638, -0.00000680, -0.00000726, -0.00000798, -0.00000902, -0.00001015, -0.00001080, -0.00001022, -0.00000777, -0.00000325, 0.00000287, 0.00000943, 0.00001493, 0.00001803, 0.00001801, 0.00001517, 0.00001066, 0.00000613, 0.00000302, 0.00000201, 0.00000279, 0.00000421, 0.00000491, 0.00000392, 0.00000124, -0.00000220, -0.00000485, -0.00000532, -0.00000303, 0.00000148, 0.00000666, 0.00001059, 0.00001175, 0.00000964, 0.00000502, -0.00000049, -0.00000499, -0.00000712, -0.00000651, -0.00000385, -0.00000050, 0.00000213, 0.00000320, 0.00000265, },
		{ -0.00000094, 0.00000074, 0.00000157, 0.00000146, 0.00000080, 0.00000025, 0.00000037, 0.00000132, 0.00000284, 0.00000429, 0.00000495, 0.00000427, 0.00000209, -0.00000138, -0.00000557, -0.00000968, -0.00001293, -0.00001463, -0.00001438, -0.00001213, -0.00000830, -0.00000367, 0.00000069, 0.00000378, 0.00000494, 0.00000409, 0.00000182, -0.00000080, -0.00000263, -0.00000290, -0.00000160, 0.00000059, 0.00000247, 0.00000296, 0.00000155, -0.00000138, -0.00000465, -0.00000679, -0.00000668, -0.00000408, 0.00000019, 0.00000459, 0.00000746, 0.00000778, 0.00000560, 0.00000203, -0.00000125, -0.00000280, -0.00000203, 0.00000046, 0.00000318, 0.00000441, 0.00000305, -0.00000082, -0.00000585, -0.00000998, -0.00001123, -0.00000858, -0.00000246, 0.00000536, 0.00001237, 0.00001620, 0.00001552, 0.00001044, 0.00000242, -0.00000626, -0.00001328, -0.00001703, -0.00001694, -0.00001354, -0.00000809, -0.00000215, 0.00000293, 0.00000629, 0.00000764, 0.00000719, 0.00000547, 0.00000314, 0.00000084, -0.00000094, -0.00000193, -0.00000209, -0.00000162, -0.00000080, 0.00000004, 0.00000067, 0.00000099, 0.00000103, 0.00000081, 0.00000029, -0.00000069, -0.00000236, -0.00000484, -0.00000796, -0.00001116, -0.00001357, -0.00001428, -0.00001268, -0.00000880, -0.00000340, 0.00000225, 0.00000676, 0.00000917, 0.00000929, 0.00000777, 0.00000579, 0.00000456, 0.00000471, 0.00000610, 0.00000782, 0.00000869, 0.00000781, 0.00000507, 0.00000124, -0.00000232, -0.00000417, -0.00000351, -0.00000050, 0.00000378, 0.00000773, 0.00000989, 0.00000951, 0.00000682, 0.00000288, -0.00000089, -0.00000331, -0.00000385, -0.00000278, },
		{ 0.00000032, -0.00000179, -0.00000261, -0.00000216, -0.00000091, 0.00000052, 0.00000166, 0.00000231, 0.00000259, 0.00000279, 0.00000314, 0.00000369, 0.00000428, 0.00000461, 0.00000436, 0.00000331, 0.00000143, -0.00000114, -0.00000411, -0.00000708, -0.00000957, -0.00001108, -0.00001119, -0.00000968, -0.00000668, -0.00000277, 0.00000113, 0.00000397, 0.00000493, 0.00000374, 0.00000090, -0.00000252, -0.00000519, -0.00000607, -0.00000486, -0.00000213, 0.00000088, 0.00000279, 0.00000268, 0.00000055, -0.00000266, -0.00000547, -0.00000650, -0.00000515, -0.00000185, 0.00000203, 0.00000481, 0.00000526, 0.00000318, -0.00000040, -0.00000369, -0.00000486, -0.00000289, 0.00000187, 0.00000778, 0.00001246, 0.00001375, 0.00001058, 0.00000348, -0.00000558, -0.00001382, -0.00001864, -0.00001851, -0.00001349, -0.00000511, 0.00000413, 0.00001164, 0.00001551, 0.00001504, 0.00001079, 0.00000429, -0.00000250, -0.00000780, -0.00001043, -0.00001003, -0.00000705, -0.00000254, 0.00000221, 0.00000600, 0.00000798, 0.00000787, 0.00000591, 0.00000280, -0.00000056, -0.00000335, -0.00000505, -0.00000551, -0.00000495, -0.00000378, -0.00000241, -0.00000107, 0.00000016, 0.00000131, 0.00000234, 0.00000306, 0.00000312, 0.00000217, 0.00000008, -0.00000290, -0.00000611, -0.00000866, -0.00000974, -0.00000902, -0.00000675, -0.00000375, -0.00000107, 0.00000044, 0.00000048, -0.00000054, -0.00000172, -0.00000208, -0.00000102, 0.00000137, 0.00000432, 0.00000671, 0.00000760, 0.00000661, 0.00000416, 0.00000128, -0.00000080, -0.00000120, 0.00000025, 0.00000294, 0.00000576, 0.00000757, 0.00000765, 0.00000599, 0.00000322, },
		{ 0.00000655, 0.00000660, 0.00000525, 0.00000287, 0.00000014, -0.00000224, -0.00000375, -0.00000421, -0.00000371, -0.00000246, -0.00000075, 0.00000124, 0.00000334, 0.00000538, 0.00000713, 0.00000832, 0.00000870, 0.00000811, 0.00000658, 0.00000429, 0.00000153, -0.00000141, -0.00000431, -0.00000693, -0.00000905, -0.00001032, -0.00001035, -0.00000884, -0.00000575, -0.00000149, 0.00000304, 0.00000667, 0.00000832, 0.00000744, 0.00000423, -0.00000027, -0.00000459, -0.00000734, -0.00000776, -0.00000604, -0.00000320, -0.00000069, 0.00000032, -0.00000055, -0.00000271, -0.00000484, -0.00000551, -0.00000391, -0.00000027, 0.00000408, 0.00000728, 0.00000768, 0.00000463, -0.00000109, -0.00000752, -0.00001214, -0.00001288, -0.00000894, -0.00000118, 0.00000811, 0.00001599, 0.00001988, 0.00001841, 0.00001191, 0.00000224, -0.00000782, -0.00001553, -0.00001897, -0.00001757, -0.00001217, -0.00000464, 0.00000277, 0.00000810, 0.00001021, 0.00000900, 0.00000533, 0.00000067, -0.00000335, -0.00000547, -0.00000506, -0.00000232, 0.00000186, 0.00000620, 0.00000944, 0.00001071, 0.00000976, 0.00000699, 0.00000325, -0.00000045, -0.00000332, -0.00000500, -0.00000555, -0.00000533, -0.00000478, -0.00000417, -0.00000354, -0.00000277, -0.00000172, -0.00000043, 0.00000084, 0.00000168, 0.00000171, 0.00000085, -0.00000062, -0.00000213, -0.00000304, -0.00000294, -0.00000188, -0.00000037, 0.00000081, 0.00000098, -0.00000013, -0.00000223, -0.00000454, -0.00000615, -0.00000639, -0.00000512, -0.00000281, -0.00000031, 0.00000149, 0.00000209, 0.00000155, 0.00000046, -0.00000036, -0.00000025, 0.00000100, 0.00000305, 0.00000518, },
		{ -0.00000241, -0.00000078, 0.00000142, 0.00000365, 0.00000530, 0.00000592, 0.00000531, 0.00000358, 0.00000109, -0.00000173, -0.00000441, -0.00000657, -0.00000786, -0.00000803, -0.00000690, -0.00000446, -0.00000093, 0.00000317, 0.00000717, 0.00001032, 0.00001204, 0.00001208, 0.00001058, 0.00000795, 0.00000471, 0.00000132, -0.00000194, -0.00000496, -0.00000761, -0.00000964, -0.00001062, -0.00001005, -0.00000761, -0.00000344, 0.00000175, 0.00000673, 0.00001018, 0.00001110, 0.00000926, 0.00000533, 0.00000065, -0.00000325, -0.00000529, -0.00000526, -0.00000385, -0.00000237, -0.00000206, -0.00000349, -0.00000625, -0.00000901, -0.00001013, -0.00000838, -0.00000358, 0.00000315, 0.00000970, 0.00001372, 0.00001351, 0.00000874, 0.00000067, -0.00000825, -0.00001516, -0.00001778, -0.00001520, -0.00000814, 0.00000126, 0.00001022, 0.00001618, 0.00001758, 0.00001429, 0.00000755, -0.00000052, -0.00000760, -0.00001190, -0.00001258, -0.00000997, -0.00000530, -0.00000027, 0.00000350, 0.00000496, 0.00000394, 0.00000111, -0.00000226, -0.00000477, -0.00000533, -0.00000354, 0.00000018, 0.00000478, 0.00000892, 0.00001147, 0.00001182, 0.00001007, 0.00000695, 0.00000346, 0.00000052, -0.00000135, -0.00000212, -0.00000216, -0.00000198, -0.00000198, -0.00000228, -0.00000272, -0.00000299, -0.00000289, -0.00000244, -0.00000187, -0.00000148, -0.00000149, -0.00000184, -0.00000221, -0.00000217, -0.00000140, 0.00000012, 0.00000200, 0.00000363, 0.00000435, 0.00000379, 0.00000203, -0.00000042, -0.00000282, -0.00000452, -0.00000518, -0.00000490, -0.00000411, -0.00000336, -0.00000302, -0.00000311, -0.00000333, -0.00000323, },
		{ -0.00000392, -0.00000464, -0.00000516, -0.00000521, -0.00000459, -0.00000326, -0.00000138, 0.00000073, 0.00000269, 0.00000416, 0.00000486, 0.00000463, 0.00000338, 0.00000113, -0.00000194, -0.00000545, -0.00000880, -0.00001126, -0.00001214, -0.00001099, -0.00000784, -0.00000324, 0.00000187, 0.00000641, 0.00000948, 0.00001068, 0.00001012, 0.00000831, 0.00000593, 0.00000349, 0.00000121, -0.00000098, -0.00000322, -0.00000547, -0.00000736, -0.00000828, -0.00000759, -0.00000499, -0.00000082, 0.00000399, 0.00000811, 0.00001034, 0.00001007, 0.00000759, 0.00000398, 0.00000069, -0.00000110, -0.00000093, 0.00000065, 0.00000234, 0.00000265, 0.00000061, -0.00000371, -0.00000914, -0.00001377, -0.00001564, -0.00001357, -0.00000768, 0.00000056, 0.00000877, 0.00001448, 0.00001593, 0.00001271, 0.00000587, -0.00000246, -0.00000976, -0.00001397, -0.00001405, -0.00001031, -0.00000414, 0.00000241, 0.00000738, 0.00000945, 0.00000832, 0.00000469, -0.00000004, -0.00000423, -0.00000657, -0.00000646, -0.00000417, -0.00000071, 0.00000257, 0.00000444, 0.00000422, 0.00000204, -0.00000122, -0.00000428, -0.00000588, -0.00000528, -0.00000251, 0.00000166, 0.00000595, 0.00000910, 0.00001025, 0.00000926, 0.00000668, 0.00000350, 0.00000076, -0.00000086, -0.00000117, -0.00000051, 0.00000051, 0.00000132, 0.00000157, 0.00000129, 0.00000070, 0.00000011, -0.00000032, -0.00000061, -0.00000093, -0.00000145, -0.00000219, -0.00000294, -0.00000336, -0.00000313, -0.00000209, -0.00000042, 0.00000146, 0.00000304, 0.00000386, 0.00000375, 0.00000285, 0.00000150, 0.00000010, -0.00000107, -0.00000193, -0.00000259, -0.00000322, },
		{ 0.00000156, 0.00000076, -0.00000024, -0.00000137, -0.00000251, -0.00000352, -0.00000425, -0.00000459, -0.00000451, -0.00000403, -0.00000322, -0.00000213, -0.00000078, 0.00000076, 0.00000236, 0.00000369, 0.00000431, 0.00000376, 0.00000179, -0.00000151, -0.00000556, -0.00000938, -0.00001192, -0.00001234, -0.00001040, -0.00000656, -0.00000186, 0.00000241, 0.00000518, 0.00000594, 0.00000490, 0.00000277, 0.00000049, -0.00000123, -0.00000214, -0.00000248, -0.00000271, -0.00000320, -0.00000393, -0.00000447, -0.00000413, -0.00000238, 0.00000078, 0.00000471, 0.00000828, 0.00001026, 0.00000990, 0.00000725, 0.00000329, -0.00000047, -0.00000256, -0.00000211, 0.00000066, 0.00000457, 0.00000784, 0.00000882, 0.00000661, 0.00000150, -0.00000513, -0.00001127, -0.00001494, -0.00001496, -0.00001128, -0.00000500, 0.00000201, 0.00000774, 0.00001069, 0.00001032, 0.00000709, 0.00000230, -0.00000244, -0.00000571, -0.00000673, -0.00000551, -0.00000283, 0.00000013, 0.00000218, 0.00000255, 0.00000112, -0.00000154, -0.00000438, -0.00000628, -0.00000643, -0.00000464, -0.00000145, 0.00000207, 0.00000472, 0.00000558, 0.00000442, 0.00000174, -0.00000134, -0.00000353, -0.00000386, -0.00000204, 0.00000143, 0.00000542, 0.00000860, 0.00000994, 0.00000905, 0.00000628, 0.00000258, -0.00000092, -0.00000330, -0.00000414, -0.00000357, -0.00000214, -0.00000050, 0.00000083, 0.00000164, 0.00000200, 0.00000214, 0.00000226, 0.00000242, 0.00000250, 0.00000231, 0.00000173, 0.00000081, -0.00000024, -0.00000111, -0.00000153, -0.00000137, -0.00000070, 0.00000027, 0.00000124, 0.00000200, 0.00000241, 0.00000244, 0.00000214, },
		{ 0.00000162, 0.00000243, 0.00000313, 0.00000345, 0.00000324, 0.00000248, 0.00000128, -0.00000008, -0.00000130, -0.00000214, -0.00000252, -0.00000255, -0.00000246, -0.00000250, -0.00000276, -0.00000310, -0.00000321, -0.00000274, -0.00000154, 0.00000021, 0.00000191, 0.00000281, 0.00000223, -0.00000002, -0.00000353, -0.00000728, -0.00000997, -0.00001052, -0.00000849, -0.00000433, 0.00000075, 0.00000519, 0.00000760, 0.00000730, 0.00000452, 0.00000021, -0.00000429, -0.00000783, -0.00000983, -0.00001037, -0.00000998, -0.00000927, -0.00000855, -0.00000765, -0.00000604, -0.00000317, 0.00000110, 0.00000624, 0.00001112, 0.00001435, 0.00001485, 0.00001227, 0.00000727, 0.00000135, -0.00000366, -0.00000622, -0.00000560, -0.00000219, 0.00000271, 0.00000734, 0.00001009, 0.00001003, 0.00000720, 0.00000250, -0.00000264, -0.00000678, -0.00000888, -0.00000864, -0.00000644, -0.00000319, 0.00000002, 0.00000230, 0.00000317, 0.00000271, 0.00000140, -0.00000001, -0.00000088, -0.00000085, 0.00000002, 0.00000125, 0.00000213, 0.00000203, 0.00000066, -0.00000179, -0.00000465, -0.00000697, -0.00000790, -0.00000701, -0.00000449, -0.00000115, 0.00000191, 0.00000365, 0.00000355, 0.00000181, -0.00000069, -0.00000273, -0.00000322, -0.00000162, 0.00000182, 0.00000614, 0.00001001, 0.00001221, 0.00001206, 0.00000966, 0.00000573, 0.00000141, -0.00000223, -0.00000450, -0.00000526, -0.00000484, -0.00000381, -0.00000272, -0.00000190, -0.00000135, -0.00000090, -0.00000032, 0.00000054, 0.00000160, 0.00000266, 0.00000345, 0.00000376, 0.00000352, 0.00000285, 0.00000197, 0.00000116, 0.00000064, 0.00000056, 0.00000092, },
		{ 0.00000138, 0.00000064, 0.00000030, 0.00000064, 0.00000169, 0.00000321, 0.00000477, 0.00000586, 0.00000607, 0.00000528, 0.00000371, 0.00000185, 0.00000028, -0.00000053, -0.00000048, 0.00000017, 0.00000088, 0.00000110, 0.00000051, -0.00000077, -0.00000223, -0.00000320, -0.00000317, -0.00000214, -0.00000062, 0.00000044, 0.00000012, -0.00000204, -0.00000572, -0.00000981, -0.00001278, -0.00001319, -0.00001037, -0.00000466, 0.00000254, 0.00000930, 0.00001374, 0.00001473, 0.00001217, 0.00000699, 0.00000071, -0.00000517, -0.00000967, -0.00001254, -0.00001414, -0.00001506, -0.00001569, -0.00001593, -0.00001520, -0.00001280, -0.00000831, -0.00000196, 0.00000527, 0.00001188, 0.00001631, 0.00001749, 0.00001524, 0.00001029, 0.00000413, -0.00000156, -0.00000536, -0.00000652, -0.00000512, -0.00000192, 0.00000194, 0.00000535, 0.00000746, 0.00000795, 0.00000695, 0.00000492, 0.00000251, 0.00000029, -0.00000132, -0.00000218, -0.00000237, -0.00000214, -0.00000180, -0.00000158, -0.00000152, -0.00000153, -0.00000136, -0.00000080, 0.00000019, 0.00000140, 0.00000241, 0.00000272, 0.00000196, 0.00000012, -0.00000238, -0.00000481, -0.00000632, -0.00000633, -0.00000478, -0.00000219, 0.00000049, 0.00000224, 0.00000237, 0.00000084, -0.00000169, -0.00000410, -0.00000519, -0.00000421, -0.00000111, 0.00000339, 0.00000805, 0.00001158, 0.00001302, 0.00001213, 0.00000934, 0.00000560, 0.00000195, -0.00000080, -0.00000230, -0.00000269, -0.00000245, -0.00000212, -0.00000207, -0.00000239, -0.00000290, -0.00000325, -0.00000314, -0.00000243, -0.00000120, 0.00000025, 0.00000155, 0.00000238, 0.00000257, 0.00000215, },
		{ -0.00000029, 0.00000037, 0.00000037, -0.00000041, -0.00000176, -0.00000317, -0.00000404, -0.00000383, -0.00000234, 0.00000020, 0.00000317, 0.00000573, 0.00000714, 0.00000705, 0.00000562, 0.00000353, 0.00000166, 0.00000077, 0.00000117, 0.00000258, 0.00000427, 0.00000536, 0.00000527, 0.00000397, 0.00000202, 0.00000029, -0.00000044, 0.00000006, 0.00000130, 0.00000220, 0.00000158, -0.00000131, -0.00000631, -0.00001223, -0.00001725, -0.00001947, -0.00001772, -0.00001200, -0.00000360, 0.00000537, 0.00001263, 0.00001655, 0.00001658, 0.00001333, 0.00000821, 0.00000281, -0.00000173, -0.00000500, -0.00000732, -0.00000937, -0.00001169, -0.00001434, -0.00001677, -0.00001805, -0.00001731, -0.00001409, -0.00000868, -0.00000203, 0.00000452, 0.00000965, 0.00001243, 0.00001259, 0.00001049, 0.00000692, 0.00000288, -0.00000071, -0.00000322, -0.00000430, -0.00000395, -0.00000241, -0.00000010, 0.00000248, 0.00000479, 0.00000641, 0.00000708, 0.00000677, 0.00000569, 0.00000418, 0.00000261, 0.00000128, 0.00000030, -0.00000038, -0.00000084, -0.00000117, -0.00000133, -0.00000125, -0.00000083, -0.00000014, 0.00000059, 0.00000098, 0.00000067, -0.00000049, -0.00000231, -0.00000424, -0.00000556, -0.00000565, -0.00000429, -0.00000180, 0.00000101, 0.00000313, 0.00000375, 0.00000260, 0.00000008, -0.00000283, -0.00000493, -0.00000528, -0.00000352, -0.00000004, 0.00000417, 0.00000788, 0.00001007, 0.00001025, 0.00000858, 0.00000578, 0.00000283, 0.00000056, -0.00000056, -0.00000055, 0.00000017, 0.00000099, 0.00000140, 0.00000113, 0.00000025, -0.00000089, -0.00000186, -0.00000228, -0.00000203, -0.00000124, },
		{ 0.00000174, 0.00000135, 0.00000150, 0.00000215, 0.00000295, 0.00000336, 0.00000283, 0.00000110, -0.00000167, -0.00000486, -0.00000755, -0.00000886, -0.00000826, -0.00000580, -0.00000220, 0.00000146, 0.00000405, 0.00000488, 0.00000393, 0.00000190, -0.00000009, -0.00000100, -0.00000026, 0.00000194, 0.00000477, 0.00000710, 0.00000806, 0.00000744, 0.00000577, 0.00000410, 0.00000345, 0.00000430, 0.00000624, 0.00000806, 0.00000824, 0.00000564, 0.00000008, -0.00000741, -0.00001485, -0.00001998, -0.00002110, -0.00001770, -0.00001071, -0.00000215, 0.00000560, 0.00001060, 0.00001197, 0.00001011, 0.00000633, 0.00000228, -0.00000071, -0.00000210, -0.00000219, -0.00000183, -0.00000197, -0.00000317, -0.00000542, -0.00000815, -0.00001053, -0.00001181, -0.00001160, -0.00000993, -0.00000721, -0.00000401, -0.00000091, 0.00000167, 0.00000345, 0.00000430, 0.00000418, 0.00000316, 0.00000145, -0.00000055, -0.00000235, -0.00000341, -0.00000336, -0.00000214, -0.00000001, 0.00000247, 0.00000466, 0.00000605, 0.00000643, 0.00000594, 0.00000492, 0.00000382, 0.00000294, 0.00000239, 0.00000211, 0.00000194, 0.00000178, 0.00000162, 0.00000153, 0.00000153, 0.00000154, 0.00000134, 0.00000063, -0.00000076, -0.00000273, -0.00000486, -0.00000651, -0.00000704, -0.00000608, -0.00000378, -0.00000076, 0.00000202, 0.00000366, 0.00000368, 0.00000217, -0.00000016, -0.00000226, -0.00000316, -0.00000231, 0.00000016, 0.00000353, 0.00000672, 0.00000872, 0.00000890, 0.00000725, 0.00000436, 0.00000112, -0.00000151, -0.00000289, -0.00000285, -0.00000167, 0.00000006, 0.00000165, 0.00000262, 0.00000280, 0.00000237, },
		{ -0.00000179, -0.00000029, 0.00000065, 0.00000105, 0.00000123, 0.00000165, 0.00000268, 0.00000437, 0.00000634, 0.00000789, 0.00000821, 0.00000677, 0.00000354, -0.00000086, -0.00000533, -0.00000860, -0.00000972, -0.00000845, -0.00000537, -0.00000171, 0.00000112, 0.00000211, 0.00000100, -0.00000156, -0.00000430, -0.00000586, -0.00000539, -0.00000291, 0.00000071, 0.00000412, 0.00000617, 0.00000636, 0.00000516, 0.00000373, 0.00000337, 0.00000488, 0.00000804, 0.00001163, 0.00001388, 0.00001321, 0.00000898, 0.00000190, -0.00000617, -0.00001279, -0.00001590, -0.00001457, -0.00000936, -0.00000212, 0.00000473, 0.00000903, 0.00000965, 0.00000676, 0.00000164, -0.00000388, -0.00000815, -0.00001016, -0.00000981, -0.00000772, -0.00000483, -0.00000204, 0.00000008, 0.00000133, 0.00000179, 0.00000158, 0.00000083, -0.00000038, -0.00000193, -0.00000357, -0.00000494, -0.00000566, -0.00000550, -0.00000455, -0.00000318, -0.00000197, -0.00000142, -0.00000173, -0.00000272, -0.00000385, -0.00000446, -0.00000408, -0.00000259, -0.00000033, 0.00000210, 0.00000402, 0.00000503, 0.00000505, 0.00000435, 0.00000341, 0.00000265, 0.00000231, 0.00000238, 0.00000271, 0.00000309, 0.00000342, 0.00000368, 0.00000388, 0.00000400, 0.00000387, 0.00000324, 0.00000191, -0.00000013, -0.00000257, -0.00000484, -0.00000631, -0.00000647, -0.00000526, -0.00000306, -0.00000063, 0.00000119, 0.00000180, 0.00000112, -0.00000042, -0.00000195, -0.00000257, -0.00000169, 0.00000074, 0.00000417, 0.00000762, 0.00001006, 0.00001072, 0.00000937, 0.00000636, 0.00000251, -0.00000118, -0.00000387, -0.00000507, -0.00000481, -0.00000350, },
		{ -0.00000118, -0.00000341, -0.00000428, -0.00000399, -0.00000306, -0.00000209, -0.00000154, -0.00000148, -0.00000161, -0.00000137, -0.00000025, 0.00000196, 0.00000496, 0.00000797, 0.00000993, 0.00000995, 0.00000768, 0.00000354, -0.00000132, -0.00000541, -0.00000746, -0.00000687, -0.00000405, -0.00000028, 0.00000280, 0.00000382, 0.00000227, -0.00000131, -0.00000551, -0.00000867, -0.00000956, -0.00000792, -0.00000454, -0.00000090, 0.00000155, 0.00000202, 0.00000079, -0.00000095, -0.00000163, -0.00000013, 0.00000365, 0.00000863, 0.00001291, 0.00001458, 0.00001256, 0.00000714, -0.00000001, -0.00000643, -0.00000981, -0.00000887, -0.00000383, 0.00000365, 0.00001109, 0.00001610, 0.00001718, 0.00001408, 0.00000780, 0.00000004, -0.00000740, -0.00001318, -0.00001665, -0.00001779, -0.00001694, -0.00001454, -0.00001094, -0.00000648, -0.00000158, 0.00000318, 0.00000702, 0.00000923, 0.00000940, 0.00000760, 0.00000445, 0.00000091, -0.00000208, -0.00000393, -0.00000459, -0.00000450, -0.00000432, -0.00000461, -0.00000551, -0.00000674, -0.00000767, -0.00000771, -0.00000656, -0.00000433, -0.00000158, 0.00000100, 0.00000280, 0.00000358, 0.00000344, 0.00000280, 0.00000212, 0.00000172, 0.00000172, 0.00000202, 0.00000245, 0.00000287, 0.00000325, 0.00000363, 0.00000399, 0.00000425, 0.00000418, 0.00000354, 0.00000224, 0.00000040, -0.00000158, -0.00000315, -0.00000382, -0.00000340, -0.00000209, -0.00000044, 0.00000084, 0.00000116, 0.00000034, -0.00000129, -0.00000304, -0.00000403, -0.00000358, -0.00000147, 0.00000196, 0.00000588, 0.00000927, 0.00001122, 0.00001120, 0.00000926, 0.00000596, 0.00000217, },
		{ 0.00000975, 0.00000901, 0.00000676, 0.00000370, 0.00000067, -0.00000157, -0.00000261, -0.00000249, -0.00000166, -0.00000076, -0.00000036, -0.00000068, -0.00000149, -0.00000215, -0.00000197, -0.00000046, 0.00000230, 0.00000560, 0.00000830, 0.00000924, 0.00000778, 0.00000408, -0.00000077, -0.00000515, -0.00000746, -0.00000681, -0.00000342, 0.00000144, 0.00000586, 0.00000808, 0.00000718, 0.00000347, -0.00000163, -0.00000620, -0.00000865, -0.00000832, -0.00000581, -0.00000262, -0.00000042, -0.00000032, -0.00000227, -0.00000511, -0.00000710, -0.00000675, -0.00000358, 0.00000155, 0.00000678, 0.00000996, 0.00000962, 0.00000565, -0.00000055, -0.00000660, -0.00000999, -0.00000911, -0.00000383, 0.00000450, 0.00001351, 0.00002073, 0.00002440, 0.00002390, 0.00001979, 0.00001335, 0.00000603, -0.00000106, -0.00000732, -0.00001252, -0.00001656, -0.00001920, -0.00002003, -0.00001864, -0.00001494, -0.00000938, -0.00000296, 0.00000298, 0.00000724, 0.00000914, 0.00000874, 0.00000676, 0.00000425, 0.00000212, 0.00000082, 0.00000023, -0.00000020, -0.00000105, -0.00000264, -0.00000481, -0.00000701, -0.00000853, -0.00000885, -0.00000786, -0.00000588, -0.00000354, -0.00000147, -0.00000010, 0.00000052, 0.00000062, 0.00000057, 0.00000068, 0.00000105, 0.00000162, 0.00000220, 0.00000265, 0.00000292, 0.00000307, 0.00000316, 0.00000321, 0.00000310, 0.00000267, 0.00000179, 0.00000053, -0.00000087, -0.00000198, -0.00000238, -0.00000186, -0.00000055, 0.00000112, 0.00000249, 0.00000300, 0.00000235, 0.00000069, -0.00000141, -0.00000315, -0.00000378, -0.00000288, -0.00000050, 0.00000282, 0.00000620, 0.00000874, },
		{ 0.00000012, 0.00000246, 0.00000515, 0.00000739, 0.00000849, 0.00000804, 0.00000611, 0.00000320, 0.00000013, -0.00000229, -0.00000347, -0.00000327, -0.00000204, -0.00000048, 0.00000068, 0.00000098, 0.00000045, -0.00000036, -0.00000066, 0.00000024, 0.00000249, 0.00000555, 0.00000830, 0.00000942, 0.00000799, 0.00000397, -0.00000164, -0.00000706, -0.00001038, -0.00001032, -0.00000677, -0.00000091, 0.00000518, 0.00000932, 0.00001012, 0.00000751, 0.00000274, -0.00000216, -0.00000527, -0.00000557, -0.00000340, -0.00000021, 0.00000212, 0.00000218, -0.00000030, -0.00000428, -0.00000790, -0.00000938, -0.00000781, -0.00000369, 0.00000125, 0.00000476, 0.00000508, 0.00000175, -0.00000415, -0.00001035, -0.00001430, -0.00001415, -0.00000947, -0.00000137, 0.00000799, 0.00001623, 0.00002160, 0.00002347, 0.00002231, 0.00001926, 0.00001545, 0.00001157, 0.00000770, 0.00000352, -0.00000121, -0.00000633, -0.00001113, -0.00001457, -0.00001574, -0.00001431, -0.00001072, -0.00000606, -0.00000171, 0.00000125, 0.00000238, 0.00000199, 0.00000088, -0.00000002, -0.00000015, 0.00000050, 0.00000146, 0.00000202, 0.00000167, 0.00000029, -0.00000176, -0.00000384, -0.00000530, -0.00000578, -0.00000535, -0.00000437, -0.00000335, -0.00000265, -0.00000239, -0.00000240, -0.00000237, -0.00000203, -0.00000127, -0.00000020, 0.00000097, 0.00000200, 0.00000275, 0.00000320, 0.00000339, 0.00000336, 0.00000308, 0.00000246, 0.00000144, 0.00000007, -0.00000142, -0.00000266, -0.00000325, -0.00000293, -0.00000172, 0.00000006, 0.00000185, 0.00000308, 0.00000335, 0.00000259, 0.00000114, -0.00000041, -0.00000137, -0.00000125, },
		{ 0.00000066, -0.00000014, -0.00000063, -0.00000035, 0.00000092, 0.00000310, 0.00000571, 0.00000807, 0.00000946, 0.00000934, 0.00000757, 0.00000449, 0.00000083, -0.00000248, -0.00000463, -0.00000520, -0.00000434, -0.00000264, -0.00000092, 0.00000014, 0.00000033, 0.00000003, -0.00000000, 0.00000102, 0.00000345, 0.00000688, 0.00001019, 0.00001190, 0.00001082, 0.00000662, 0.00000014, -0.00000678, -0.00001189, -0.00001345, -0.00001089, -0.00000514, 0.00000172, 0.00000722, 0.00000952, 0.00000811, 0.00000396, -0.00000086, -0.00000418, -0.00000463, -0.00000224, 0.00000165, 0.00000499, 0.00000605, 0.00000416, 0.00000012, -0.00000428, -0.00000700, -0.00000680, -0.00000383, 0.00000038, 0.00000360, 0.00000386, 0.00000042, -0.00000584, -0.00001275, -0.00001766, -0.00001854, -0.00001478, -0.00000739, 0.00000144, 0.00000928, 0.00001431, 0.00001598, 0.00001500, 0.00001281, 0.00001085, 0.00000988, 0.00000979, 0.00000977, 0.00000887, 0.00000657, 0.00000312, -0.00000058, -0.00000339, -0.00000448, -0.00000370, -0.00000169, 0.00000045, 0.00000164, 0.00000127, -0.00000051, -0.00000298, -0.00000515, -0.00000624, -0.00000599, -0.00000469, -0.00000304, -0.00000174, -0.00000121, -0.00000145, -0.00000209, -0.00000265, -0.00000277, -0.00000240, -0.00000177, -0.00000126, -0.00000116, -0.00000155, -0.00000227, -0.00000297, -0.00000332, -0.00000313, -0.00000241, -0.00000131, -0.00000009, 0.00000107, 0.00000206, 0.00000283, 0.00000335, 0.00000356, 0.00000336, 0.00000266, 0.00000146, -0.00000005, -0.00000158, -0.00000273, -0.00000320, -0.00000287, -0.00000187, -0.00000055, 0.00000064, 0.00000129, 0.00000126, },
		{ -0.00000032, 0.00000005, 0.00000014, -0.00000018, -0.00000089, -0.00000172, -0.00000228, -0.00000216, -0.00000107, 0.00000104, 0.00000387, 0.00000684, 0.00000918, 0.00001019, 0.00000946, 0.00000703, 0.00000347, -0.00000033, -0.00000339, -0.00000505, -0.00000513, -0.00000404, -0.00000255, -0.00000144, -0.00000109, -0.00000129, -0.00000131, -0.00000027, 0.00000236, 0.00000635, 0.00001063, 0.00001358, 0.00001367, 0.00001019, 0.00000368, -0.00000409, -0.00001069, -0.00001395, -0.00001281, -0.00000779, -0.00000086, 0.00000536, 0.00000863, 0.00000797, 0.00000407, -0.00000108, -0.00000508, -0.00000618, -0.00000400, 0.00000030, 0.00000464, 0.00000698, 0.00000629, 0.00000302, -0.00000114, -0.00000405, -0.00000421, -0.00000146, 0.00000282, 0.00000633, 0.00000686, 0.00000337, -0.00000350, -0.00001155, -0.00001792, -0.00002020, -0.00001743, -0.00001047, -0.00000160, 0.00000639, 0.00001126, 0.00001216, 0.00000977, 0.00000580, 0.00000224, 0.00000044, 0.00000070, 0.00000231, 0.00000410, 0.00000505, 0.00000482, 0.00000384, 0.00000304, 0.00000329, 0.00000492, 0.00000752, 0.00001011, 0.00001154, 0.00001102, 0.00000846, 0.00000446, 0.00000006, -0.00000370, -0.00000617, -0.00000724, -0.00000730, -0.00000692, -0.00000658, -0.00000646, -0.00000644, -0.00000621, -0.00000553, -0.00000436, -0.00000294, -0.00000162, -0.00000076, -0.00000054, -0.00000087, -0.00000152, -0.00000215, -0.00000251, -0.00000248, -0.00000209, -0.00000147, -0.00000072, 0.00000008, 0.00000090, 0.00000169, 0.00000238, 0.00000285, 0.00000297, 0.00000266, 0.00000195, 0.00000100, 0.00000007, -0.00000061, -0.00000087, -0.00000072, },
		{ 0.00000150, 0.00000164, 0.00000190, 0.00000227, 0.00000264, 0.00000285, 0.00000273, 0.00000215, 0.00000112, -0.00000021, -0.00000152, -0.00000242, -0.00000249, -0.00000149, 0.00000054, 0.00000323, 0.00000591, 0.00000781, 0.00000827, 0.00000703, 0.00000434, 0.00000095, -0.00000217, -0.00000419, -0.00000472, -0.00000396, -0.00000260, -0.00000150, -0.00000118, -0.00000160, -0.00000207, -0.00000161, 0.00000056, 0.00000449, 0.00000925, 0.00001321, 0.00001457, 0.00001223, 0.00000632, -0.00000160, -0.00000908, -0.00001360, -0.00001364, -0.00000930, -0.00000231, 0.00000463, 0.00000891, 0.00000911, 0.00000551, -0.00000006, -0.00000503, -0.00000727, -0.00000597, -0.00000196, 0.00000271, 0.00000580, 0.00000592, 0.00000317, -0.00000088, -0.00000397, -0.00000428, -0.00000128, 0.00000383, 0.00000866, 0.00001066, 0.00000823, 0.00000160, -0.00000724, -0.00001522, -0.00001943, -0.00001830, -0.00001215, -0.00000310, 0.00000584, 0.00001193, 0.00001367, 0.00001123, 0.00000617, 0.00000065, -0.00000348, -0.00000537, -0.00000525, -0.00000412, -0.00000305, -0.00000263, -0.00000270, -0.00000249, -0.00000112, 0.00000189, 0.00000631, 0.00001121, 0.00001533, 0.00001756, 0.00001738, 0.00001501, 0.00001122, 0.00000703, 0.00000330, 0.00000042, -0.00000163, -0.00000317, -0.00000457, -0.00000598, -0.00000731, -0.00000829, -0.00000865, -0.00000828, -0.00000727, -0.00000594, -0.00000463, -0.00000361, -0.00000298, -0.00000267, -0.00000254, -0.00000242, -0.00000224, -0.00000199, -0.00000171, -0.00000144, -0.00000118, -0.00000089, -0.00000054, -0.00000010, 0.00000038, 0.00000082, 0.00000115, 0.00000135, 0.00000144, },
		{ -0.00000239, -0.00000231, -0.00000208, -0.00000167, -0.00000105, -0.00000022, 0.00000081, 0.00000198, 0.00000318, 0.00000427, 0.00000511, 0.00000551, 0.00000538, 0.00000467, 0.00000352, 0.00000222, 0.00000116, 0.00000076, 0.00000126, 0.00000260, 0.00000435, 0.00000583, 0.00000633, 0.00000537, 0.00000299, -0.00000028, -0.00000350, -0.00000574, -0.00000640, -0.00000553, -0.00000377, -0.00000204, -0.00000111, -0.00000117, -0.00000168, -0.00000164, -0.00000007, 0.00000338, 0.00000810, 0.00001252, 0.00001475, 0.00001334, 0.00000804, 0.00000009, -0.00000809, -0.00001377, -0.00001496, -0.00001130, -0.00000423, 0.00000356, 0.00000916, 0.00001062, 0.00000775, 0.00000213, -0.00000364, -0.00000705, -0.00000681, -0.00000335, 0.00000142, 0.00000508, 0.00000583, 0.00000336, -0.00000101, -0.00000495, -0.00000622, -0.00000379, 0.00000160, 0.00000769, 0.00001164, 0.00001131, 0.00000619, -0.00000223, -0.00001102, -0.00001699, -0.00001791, -0.00001341, -0.00000504, 0.00000436, 0.00001181, 0.00001527, 0.00001430, 0.00000999, 0.00000432, -0.00000074, -0.00000399, -0.00000532, -0.00000549, -0.00000554, -0.00000616, -0.00000735, -0.00000841, -0.00000836, -0.00000643, -0.00000251, 0.00000276, 0.00000824, 0.00001269, 0.00001525, 0.00001572, 0.00001449, 0.00001233, 0.00001001, 0.00000801, 0.00000643, 0.00000507, 0.00000361, 0.00000184, -0.00000023, -0.00000239, -0.00000432, -0.00000575, -0.00000659, -0.00000687, -0.00000677, -0.00000646, -0.00000605, -0.00000558, -0.00000504, -0.00000442, -0.00000375, -0.00000310, -0.00000256, -0.00000220, -0.00000203, -0.00000203, -0.00000213, -0.00000226, -0.00000236, },
		},
		{ { 0.00000268, 0.00000273, 0.00000287, 0.00000307, 0.00000328, 0.00000343, 0.00000346, 0.00000331, 0.00000293, 0.00000229, 0.00000140, 0.00000026, -0.00000106, -0.00000249, -0.00000395, -0.00000534, -0.00000656, -0.00000752, -0.00000816, -0.00000843, -0.00000831, -0.00000782, -0.00000701, -0.00000593, -0.00000467, -0.00000334, -0.00000202, -0.00000080, 0.00000024, 0.00000104, 0.00000159, 0.00000186, 0.00000189, 0.00000170, 0.00000136, 0.00000091, 0.00000042, -0.00000005, -0.00000047, -0.00000081, -0.00000105, -0.00000119, -0.00000126, -0.00000127, -0.00000125, -0.00000124, -0.00000126, -0.00000133, -0.00000144, -0.00000161, -0.00000181, -0.00000202, -0.00000222, -0.00000237, -0.00000247, -0.00000249, -0.00000242, -0.00000227, -0.00000205, -0.00000180, -0.00000152, -0.00000127, -0.00000106, -0.00000092, -0.00000088, -0.00000092, -0.00000106, -0.00000127, -0.00000152, -0.00000180, -0.00000205, -0.00000227, -0.00000242, -0.00000249, -0.00000247, -0.00000237, -0.00000222, -0.00000202, -0.00000181, -0.00000161, -0.00000144, -0.00000133, -0.00000126, -0.00000124, -0.00000125, -0.00000127, -0.00000126, -0.00000119, -0.00000105, -0.00000081, -0.00000047, -0.00000005, 0.00000042, 0.00000091, 0.00000136, 0.00000170, 0.00000189, 0.00000186, 0.00000159, 0.00000104, 0.00000024, -0.00000080, -0.00000202, -0.00000334, -0.00000467, -0.00000593, -0.00000701, -0.00000782, -0.00000831, -0.00000843, -0.00000816, -0.00000752, -0.00000656, -0.00000534, -0.00000395, -0.00000249, -0.00000106, 0.00000026, 0.00000140, 0.00000229, 0.00000293, 0.00000331, 0.00000346, 0.00000343, 0.00000328, 0.00000307, 0.00000287, 0.00000273, },
		{ 0.00000227, 0.00000262, 0.00000278, 0.00000281, 0.00000275, 0.00000267, 0.00000263, 0.00000267, 0.00000283, 0.00000311, 0.00000349, 0.00000395, 0.00000441, 0.00000481, 0.00000507, 0.00000511, 0.00000487, 0.00000432, 0.00000345, 0.00000225, 0.00000079, -0.00000086, -0.00000261, -0.00000434, -0.00000595, -0.00000732, -0.00000835, -0.00000899, -0.00000919, -0.00000894, -0.00000827, -0.00000724, -0.00000594, -0.00000445, -0.00000290, -0.00000139, -0.00000000, 0.00000117, 0.00000207, 0.00000269, 0.00000301, 0.00000306, 0.00000288, 0.00000254, 0.00000209, 0.00000159, 0.00000111, 0.00000069, 0.00000035, 0.00000010, -0.00000004, -0.00000012, -0.00000014, -0.00000014, -0.00000015, -0.00000019, -0.00000028, -0.00000043, -0.00000063, -0.00000086, -0.00000111, -0.00000135, -0.00000156, -0.00000171, -0.00000178, -0.00000178, -0.00000171, -0.00000157, -0.00000140, -0.00000121, -0.00000104, -0.00000091, -0.00000083, -0.00000083, -0.00000089, -0.00000102, -0.00000119, -0.00000139, -0.00000158, -0.00000174, -0.00000184, -0.00000188, -0.00000185, -0.00000174, -0.00000158, -0.00000138, -0.00000116, -0.00000096, -0.00000080, -0.00000070, -0.00000066, -0.00000068, -0.00000076, -0.00000086, -0.00000095, -0.00000102, -0.00000101, -0.00000092, -0.00000072, -0.00000042, -0.00000003, 0.00000042, 0.00000089, 0.00000131, 0.00000165, 0.00000183, 0.00000182, 0.00000158, 0.00000110, 0.00000040, -0.00000049, -0.00000153, -0.00000264, -0.00000373, -0.00000473, -0.00000555, -0.00000612, -0.00000640, -0.00000636, -0.00000600, -0.00000535, -0.00000446, -0.00000339, -0.00000224, -0.00000108, 0.00000000, 0.00000095, 0.00000171, },
		{ -0.00000342, -0.00000273, -0.00000189, -0.00000097, -0.00000004, 0.00000082, 0.00000155, 0.00000212, 0.00000250, 0.00000269, 0.00000272, 0.00000265, 0.00000252, 0.00000240, 0.00000236, 0.00000245, 0.00000270, 0.00000313, 0.00000372, 0.00000442, 0.00000518, 0.00000590, 0.00000649, 0.00000687, 0.00000695, 0.00000668, 0.00000600, 0.00000492, 0.00000348, 0.00000174, -0.00000020, -0.00000224, -0.00000423, -0.00000606, -0.00000759, -0.00000873, -0.00000941, -0.00000959, -0.00000927, -0.00000850, -0.00000733, -0.00000586, -0.00000422, -0.00000251, -0.00000086, 0.00000064, 0.00000189, 0.00000284, 0.00000346, 0.00000376, 0.00000375, 0.00000349, 0.00000303, 0.00000245, 0.00000182, 0.00000120, 0.00000064, 0.00000017, -0.00000017, -0.00000040, -0.00000051, -0.00000054, -0.00000050, -0.00000042, -0.00000033, -0.00000025, -0.00000019, -0.00000015, -0.00000013, -0.00000012, -0.00000011, -0.00000009, -0.00000005, 0.00000002, 0.00000011, 0.00000021, 0.00000032, 0.00000042, 0.00000048, 0.00000050, 0.00000045, 0.00000034, 0.00000017, -0.00000005, -0.00000030, -0.00000056, -0.00000080, -0.00000100, -0.00000113, -0.00000118, -0.00000115, -0.00000104, -0.00000085, -0.00000063, -0.00000038, -0.00000015, 0.00000003, 0.00000016, 0.00000020, 0.00000016, 0.00000004, -0.00000013, -0.00000033, -0.00000052, -0.00000066, -0.00000073, -0.00000069, -0.00000054, -0.00000028, 0.00000008, 0.00000049, 0.00000093, 0.00000133, 0.00000164, 0.00000182, 0.00000182, 0.00000163, 0.00000123, 0.00000065, -0.00000009, -0.00000092, -0.00000178, -0.00000260, -0.00000330, -0.00000382, -0.00000411, -0.00000414, -0.00000390, },
		{ -0.00000069, -0.00000133, -0.00000196, -0.00000251, -0.00000292, -0.00000315, -0.00000316, -0.00000295, -0.00000253, -0.00000194, -0.00000121, -0.00000043, 0.00000034, 0.00000103, 0.00000158, 0.00000196, 0.00000213, 0.00000212, 0.00000193, 0.00000164, 0.00000129, 0.00000098, 0.00000077, 0.00000073, 0.00000090, 0.00000132, 0.00000197, 0.00000282, 0.00000381, 0.00000485, 0.00000583, 0.00000664, 0.00000718, 0.00000735, 0.00000710, 0.00000639, 0.00000521, 0.00000363, 0.00000171, -0.00000043, -0.00000266, -0.00000484, -0.00000682, -0.00000847, -0.00000968, -0.00001037, -0.00001051, -0.00001010, -0.00000918, -0.00000781, -0.00000611, -0.00000420, -0.00000220, -0.00000024, 0.00000155, 0.00000309, 0.00000431, 0.00000516, 0.00000563, 0.00000574, 0.00000552, 0.00000503, 0.00000433, 0.00000350, 0.00000261, 0.00000171, 0.00000085, 0.00000007, -0.00000061, -0.00000117, -0.00000162, -0.00000195, -0.00000219, -0.00000233, -0.00000240, -0.00000240, -0.00000233, -0.00000220, -0.00000202, -0.00000179, -0.00000151, -0.00000121, -0.00000088, -0.00000056, -0.00000027, -0.00000001, 0.00000018, 0.00000029, 0.00000031, 0.00000024, 0.00000008, -0.00000014, -0.00000041, -0.00000070, -0.00000097, -0.00000119, -0.00000134, -0.00000140, -0.00000136, -0.00000122, -0.00000100, -0.00000072, -0.00000041, -0.00000011, 0.00000015, 0.00000034, 0.00000043, 0.00000042, 0.00000031, 0.00000011, -0.00000016, -0.00000045, -0.00000073, -0.00000095, -0.00000109, -0.00000113, -0.00000104, -0.00000084, -0.00000055, -0.00000019, 0.00000018, 0.00000052, 0.00000079, 0.00000093, 0.00000092, 0.00000073, 0.00000039, -0.00000010, },
		{ -0.00000109, -0.00000076, -0.00000047, -0.00000024, -0.00000012, -0.00000013, -0.00000028, -0.00000054, -0.00000091, -0.00000132, -0.00000174, -0.00000210, -0.00000236, -0.00000246, -0.00000237, -0.00000209, -0.00000161, -0.00000098, -0.00000025, 0.00000053, 0.00000129, 0.00000194, 0.00000244, 0.00000273, 0.00000279, 0.00000262, 0.00000224, 0.00000171, 0.00000110, 0.00000049, -0.00000002, -0.00000036, -0.00000046, -0.00000026, 0.00000024, 0.00000103, 0.00000206, 0.00000325, 0.00000449, 0.00000566, 0.00000665, 0.00000732, 0.00000759, 0.00000738, 0.00000665, 0.00000540, 0.00000369, 0.00000160, -0.00000076, -0.00000325, -0.00000572, -0.00000801, -0.00000999, -0.00001153, -0.00001255, -0.00001299, -0.00001284, -0.00001212, -0.00001090, -0.00000927, -0.00000733, -0.00000521, -0.00000304, -0.00000092, 0.00000103, 0.00000274, 0.00000414, 0.00000521, 0.00000592, 0.00000629, 0.00000636, 0.00000615, 0.00000572, 0.00000512, 0.00000441, 0.00000362, 0.00000279, 0.00000197, 0.00000119, 0.00000046, -0.00000020, -0.00000077, -0.00000123, -0.00000159, -0.00000183, -0.00000195, -0.00000195, -0.00000184, -0.00000164, -0.00000135, -0.00000101, -0.00000065, -0.00000030, 0.00000002, 0.00000026, 0.00000042, 0.00000046, 0.00000039, 0.00000021, -0.00000006, -0.00000039, -0.00000076, -0.00000112, -0.00000144, -0.00000169, -0.00000183, -0.00000185, -0.00000175, -0.00000154, -0.00000125, -0.00000090, -0.00000053, -0.00000019, 0.00000008, 0.00000026, 0.00000032, 0.00000026, 0.00000008, -0.00000020, -0.00000054, -0.00000091, -0.00000125, -0.00000154, -0.00000174, -0.00000183, -0.00000179, -0.00000164, -0.00000139, },
		{ 0.00000006, -0.00000040, -0.00000083, -0.00000120, -0.00000148, -0.00000165, -0.00000169, -0.00000162, -0.00000147, -0.00000125, -0.00000102, -0.00000081, -0.00000066, -0.00000059, -0.00000063, -0.00000076, -0.00000097, -0.00000124, -0.00000150, -0.00000172, -0.00000185, -0.00000184, -0.00000165, -0.00000128, -0.00000073, -0.00000002, 0.00000079, 0.00000166, 0.00000250, 0.00000324, 0.00000382, 0.00000417, 0.00000426, 0.00000409, 0.00000366, 0.00000301, 0.00000223, 0.00000138, 0.00000058, -0.00000009, -0.00000053, -0.00000066, -0.00000046, 0.00000011, 0.00000103, 0.00000222, 0.00000362, 0.00000512, 0.00000658, 0.00000788, 0.00000889, 0.00000952, 0.00000968, 0.00000933, 0.00000846, 0.00000710, 0.00000532, 0.00000322, 0.00000091, -0.00000146, -0.00000376, -0.00000586, -0.00000765, -0.00000904, -0.00000996, -0.00001039, -0.00001034, -0.00000983, -0.00000893, -0.00000771, -0.00000626, -0.00000468, -0.00000305, -0.00000147, 0.00000001, 0.00000131, 0.00000241, 0.00000328, 0.00000390, 0.00000429, 0.00000445, 0.00000440, 0.00000417, 0.00000379, 0.00000329, 0.00000269, 0.00000203, 0.00000135, 0.00000067, 0.00000003, -0.00000055, -0.00000104, -0.00000141, -0.00000165, -0.00000174, -0.00000169, -0.00000150, -0.00000119, -0.00000079, -0.00000033, 0.00000014, 0.00000060, 0.00000099, 0.00000128, 0.00000144, 0.00000147, 0.00000136, 0.00000112, 0.00000078, 0.00000038, -0.00000004, -0.00000045, -0.00000078, -0.00000102, -0.00000114, -0.00000111, -0.00000096, -0.00000069, -0.00000033, 0.00000007, 0.00000047, 0.00000082, 0.00000108, 0.00000123, 0.00000124, 0.00000111, 0.00000085, 0.00000049, },
		{ 0.00000026, 0.00000072, 0.00000112, 0.00000141, 0.00000158, 0.00000159, 0.00000145, 0.00000118, 0.00000080, 0.00000035, -0.00000012, -0.00000057, -0.00000096, -0.00000127, -0.00000146, -0.00000154, -0.00000152, -0.00000142, -0.00000126, -0.00000109, -0.00000093, -0.00000082, -0.00000078, -0.00000081, -0.00000090, -0.00000104, -0.00000118, -0.00000128, -0.00000131, -0.00000122, -0.00000097, -0.00000055, 0.00000005, 0.00000080, 0.00000167, 0.00000260, 0.00000352, 0.00000436, 0.00000505, 0.00000551, 0.00000571, 0.00000561, 0.00000521, 0.00000453, 0.00000363, 0.00000257, 0.00000145, 0.00000036, -0.00000060, -0.00000135, -0.00000181, -0.00000195, -0.00000174, -0.00000120, -0.00000037, 0.00000068, 0.00000185, 0.00000304, 0.00000415, 0.00000508, 0.00000574, 0.00000606, 0.00000602, 0.00000559, 0.00000482, 0.00000374, 0.00000242, 0.00000097, -0.00000054, -0.00000199, -0.00000331, -0.00000441, -0.00000523, -0.00000574, -0.00000593, -0.00000579, -0.00000535, -0.00000465, -0.00000376, -0.00000271, -0.00000159, -0.00000043, 0.00000070, 0.00000176, 0.00000271, 0.00000352, 0.00000417, 0.00000464, 0.00000494, 0.00000505, 0.00000499, 0.00000476, 0.00000436, 0.00000383, 0.00000319, 0.00000246, 0.00000167, 0.00000087, 0.00000009, -0.00000062, -0.00000123, -0.00000171, -0.00000202, -0.00000216, -0.00000212, -0.00000192, -0.00000157, -0.00000111, -0.00000059, -0.00000005, 0.00000045, 0.00000088, 0.00000118, 0.00000134, 0.00000134, 0.00000119, 0.00000091, 0.00000052, 0.00000008, -0.00000037, -0.00000078, -0.00000110, -0.00000130, -0.00000135, -0.00000126, -0.00000102, -0.00000066, -0.00000022, },
		{ -0.00000037, -0.00000080, -0.00000114, -0.00000134, -0.00000138, -0.00000127, -0.00000101, -0.00000064, -0.00000020, 0.00000027, 0.00000072, 0.00000109, 0.00000134, 0.00000146, 0.00000142, 0.00000122, 0.00000089, 0.00000045, -0.00000006, -0.00000060, -0.00000112, -0.00000159, -0.00000197, -0.00000226, -0.00000244, -0.00000253, -0.00000255, -0.00000251, -0.00000245, -0.00000239, -0.00000235, -0.00000234, -0.00000237, -0.00000242, -0.00000246, -0.00000246, -0.00000240, -0.00000223, -0.00000192, -0.00000145, -0.00000082, -0.00000003, 0.00000088, 0.00000188, 0.00000291, 0.00000391, 0.00000481, 0.00000555, 0.00000609, 0.00000638, 0.00000642, 0.00000620, 0.00000576, 0.00000514, 0.00000440, 0.00000361, 0.00000285, 0.00000218, 0.00000166, 0.00000132, 0.00000119, 0.00000126, 0.00000150, 0.00000187, 0.00000230, 0.00000272, 0.00000306, 0.00000326, 0.00000324, 0.00000298, 0.00000245, 0.00000167, 0.00000064, -0.00000057, -0.00000191, -0.00000331, -0.00000469, -0.00000598, -0.00000710, -0.00000800, -0.00000865, -0.00000900, -0.00000907, -0.00000886, -0.00000838, -0.00000769, -0.00000681, -0.00000579, -0.00000469, -0.00000353, -0.00000236, -0.00000122, -0.00000014, 0.00000087, 0.00000176, 0.00000254, 0.00000317, 0.00000364, 0.00000395, 0.00000408, 0.00000402, 0.00000379, 0.00000339, 0.00000284, 0.00000217, 0.00000142, 0.00000062, -0.00000017, -0.00000090, -0.00000152, -0.00000201, -0.00000231, -0.00000242, -0.00000234, -0.00000207, -0.00000165, -0.00000111, -0.00000051, 0.00000010, 0.00000065, 0.00000111, 0.00000142, 0.00000157, 0.00000155, 0.00000136, 0.00000103, 0.00000060, 0.00000011, },
		{ -0.00000002, 0.00000043, 0.00000073, 0.00000086, 0.00000080, 0.00000058, 0.00000022, -0.00000023, -0.00000073, -0.00000122, -0.00000165, -0.00000197, -0.00000216, -0.00000218, -0.00000205, -0.00000179, -0.00000141, -0.00000097, -0.00000051, -0.00000008, 0.00000028, 0.00000051, 0.00000061, 0.00000055, 0.00000034, 0.00000000, -0.00000044, -0.00000096, -0.00000152, -0.00000207, -0.00000259, -0.00000305, -0.00000343, -0.00000374, -0.00000397, -0.00000413, -0.00000424, -0.00000432, -0.00000438, -0.00000442, -0.00000445, -0.00000447, -0.00000447, -0.00000442, -0.00000430, -0.00000411, -0.00000381, -0.00000342, -0.00000292, -0.00000234, -0.00000170, -0.00000103, -0.00000037, 0.00000022, 0.00000072, 0.00000108, 0.00000128, 0.00000132, 0.00000119, 0.00000094, 0.00000059, 0.00000021, -0.00000014, -0.00000040, -0.00000051, -0.00000042, -0.00000010, 0.00000046, 0.00000125, 0.00000224, 0.00000337, 0.00000456, 0.00000573, 0.00000680, 0.00000768, 0.00000829, 0.00000858, 0.00000851, 0.00000807, 0.00000728, 0.00000616, 0.00000479, 0.00000324, 0.00000159, -0.00000008, -0.00000168, -0.00000314, -0.00000440, -0.00000541, -0.00000615, -0.00000661, -0.00000680, -0.00000674, -0.00000645, -0.00000598, -0.00000536, -0.00000463, -0.00000381, -0.00000294, -0.00000204, -0.00000113, -0.00000023, 0.00000063, 0.00000143, 0.00000216, 0.00000279, 0.00000329, 0.00000365, 0.00000383, 0.00000383, 0.00000364, 0.00000325, 0.00000269, 0.00000199, 0.00000118, 0.00000031, -0.00000056, -0.00000138, -0.00000208, -0.00000263, -0.00000298, -0.00000313, -0.00000306, -0.00000279, -0.00000236, -0.00000181, -0.00000119, -0.00000058, },
		{ -0.00000267, -0.00000302, -0.00000316, -0.00000307, -0.00000278, -0.00000232, -0.00000174, -0.00000110, -0.00000045, 0.00000013, 0.00000060, 0.00000092, 0.00000107, 0.00000105, 0.00000086, 0.00000055, 0.00000016, -0.00000027, -0.00000069, -0.00000104, -0.00000129, -0.00000140, -0.00000136, -0.00000118, -0.00000088, -0.00000047, -0.00000001, 0.00000048, 0.00000093, 0.00000132, 0.00000161, 0.00000178, 0.00000181, 0.00000172, 0.00000151, 0.00000119, 0.00000081, 0.00000037, -0.00000009, -0.00000055, -0.00000099, -0.00000139, -0.00000174, -0.00000204, -0.00000227, -0.00000243, -0.00000252, -0.00000251, -0.00000241, -0.00000218, -0.00000182, -0.00000132, -0.00000066, 0.00000015, 0.00000111, 0.00000217, 0.00000331, 0.00000447, 0.00000558, 0.00000658, 0.00000738, 0.00000793, 0.00000816, 0.00000804, 0.00000755, 0.00000669, 0.00000551, 0.00000406, 0.00000243, 0.00000073, -0.00000093, -0.00000244, -0.00000370, -0.00000461, -0.00000511, -0.00000517, -0.00000479, -0.00000400, -0.00000287, -0.00000148, 0.00000006, 0.00000163, 0.00000311, 0.00000439, 0.00000539, 0.00000604, 0.00000631, 0.00000617, 0.00000566, 0.00000482, 0.00000372, 0.00000244, 0.00000106, -0.00000032, -0.00000164, -0.00000283, -0.00000384, -0.00000464, -0.00000522, -0.00000557, -0.00000572, -0.00000569, -0.00000550, -0.00000518, -0.00000476, -0.00000425, -0.00000367, -0.00000303, -0.00000235, -0.00000162, -0.00000086, -0.00000008, 0.00000069, 0.00000143, 0.00000211, 0.00000270, 0.00000316, 0.00000345, 0.00000356, 0.00000346, 0.00000315, 0.00000264, 0.00000197, 0.00000118, 0.00000031, -0.00000057, -0.00000140, -0.00000212, },
		{ 0.00000421, 0.00000402, 0.00000363, 0.00000305, 0.00000233, 0.00000152, 0.00000070, -0.00000009, -0.00000077, -0.00000129, -0.00000162, -0.00000173, -0.00000164, -0.00000134, -0.00000089, -0.00000033, 0.00000028, 0.00000089, 0.00000143, 0.00000187, 0.00000217, 0.00000230, 0.00000228, 0.00000211, 0.00000182, 0.00000145, 0.00000105, 0.00000066, 0.00000031, 0.00000005, -0.00000011, -0.00000016, -0.00000010, 0.00000005, 0.00000028, 0.00000055, 0.00000084, 0.00000111, 0.00000133, 0.00000148, 0.00000155, 0.00000152, 0.00000139, 0.00000116, 0.00000083, 0.00000039, -0.00000015, -0.00000078, -0.00000152, -0.00000234, -0.00000326, -0.00000424, -0.00000528, -0.00000634, -0.00000737, -0.00000832, -0.00000913, -0.00000973, -0.00001006, -0.00001004, -0.00000964, -0.00000882, -0.00000756, -0.00000590, -0.00000388, -0.00000158, 0.00000090, 0.00000344, 0.00000589, 0.00000812, 0.00001001, 0.00001144, 0.00001234, 0.00001265, 0.00001238, 0.00001155, 0.00001024, 0.00000856, 0.00000664, 0.00000463, 0.00000268, 0.00000091, -0.00000053, -0.00000158, -0.00000217, -0.00000229, -0.00000196, -0.00000126, -0.00000026, 0.00000092, 0.00000217, 0.00000336, 0.00000439, 0.00000518, 0.00000565, 0.00000579, 0.00000559, 0.00000508, 0.00000430, 0.00000333, 0.00000224, 0.00000110, -0.00000002, -0.00000105, -0.00000194, -0.00000268, -0.00000324, -0.00000363, -0.00000387, -0.00000398, -0.00000398, -0.00000389, -0.00000372, -0.00000350, -0.00000322, -0.00000288, -0.00000247, -0.00000199, -0.00000143, -0.00000079, -0.00000008, 0.00000068, 0.00000146, 0.00000222, 0.00000291, 0.00000350, 0.00000393, 0.00000418, },
		{ -0.00000192, -0.00000141, -0.00000081, -0.00000014, 0.00000055, 0.00000120, 0.00000177, 0.00000221, 0.00000246, 0.00000251, 0.00000235, 0.00000197, 0.00000141, 0.00000071, -0.00000008, -0.00000090, -0.00000167, -0.00000235, -0.00000287, -0.00000322, -0.00000337, -0.00000332, -0.00000308, -0.00000271, -0.00000223, -0.00000169, -0.00000116, -0.00000067, -0.00000025, 0.00000005, 0.00000024, 0.00000031, 0.00000028, 0.00000017, -0.00000000, -0.00000019, -0.00000038, -0.00000053, -0.00000064, -0.00000069, -0.00000067, -0.00000059, -0.00000046, -0.00000027, -0.00000004, 0.00000023, 0.00000054, 0.00000089, 0.00000129, 0.00000173, 0.00000221, 0.00000273, 0.00000326, 0.00000378, 0.00000425, 0.00000462, 0.00000483, 0.00000484, 0.00000459, 0.00000403, 0.00000315, 0.00000193, 0.00000040, -0.00000138, -0.00000334, -0.00000539, -0.00000739, -0.00000921, -0.00001075, -0.00001186, -0.00001247, -0.00001250, -0.00001193, -0.00001079, -0.00000913, -0.00000705, -0.00000468, -0.00000218, 0.00000029, 0.00000257, 0.00000452, 0.00000601, 0.00000697, 0.00000737, 0.00000722, 0.00000657, 0.00000551, 0.00000418, 0.00000270, 0.00000122, -0.00000012, -0.00000120, -0.00000195, -0.00000230, -0.00000226, -0.00000184, -0.00000110, -0.00000014, 0.00000095, 0.00000206, 0.00000309, 0.00000394, 0.00000455, 0.00000488, 0.00000491, 0.00000466, 0.00000416, 0.00000347, 0.00000265, 0.00000178, 0.00000092, 0.00000012, -0.00000058, -0.00000116, -0.00000161, -0.00000193, -0.00000216, -0.00000232, -0.00000243, -0.00000252, -0.00000261, -0.00000269, -0.00000276, -0.00000281, -0.00000281, -0.00000275, -0.00000259, -0.00000231, },
		{ -0.00000164, -0.00000179, -0.00000200, -0.00000224, -0.00000250, -0.00000273, -0.00000288, -0.00000292, -0.00000282, -0.00000257, -0.00000217, -0.00000163, -0.00000099, -0.00000031, 0.00000037, 0.00000099, 0.00000148, 0.00000181, 0.00000195, 0.00000189, 0.00000162, 0.00000118, 0.00000061, -0.00000006, -0.00000075, -0.00000141, -0.00000200, -0.00000247, -0.00000279, -0.00000296, -0.00000298, -0.00000286, -0.00000264, -0.00000235, -0.00000202, -0.00000169, -0.00000138, -0.00000112, -0.00000092, -0.00000078, -0.00000070, -0.00000067, -0.00000068, -0.00000072, -0.00000079, -0.00000090, -0.00000105, -0.00000126, -0.00000155, -0.00000192, -0.00000240, -0.00000297, -0.00000362, -0.00000433, -0.00000506, -0.00000575, -0.00000633, -0.00000675, -0.00000694, -0.00000684, -0.00000644, -0.00000570, -0.00000466, -0.00000336, -0.00000188, -0.00000031, 0.00000124, 0.00000265, 0.00000380, 0.00000461, 0.00000498, 0.00000490, 0.00000435, 0.00000339, 0.00000208, 0.00000055, -0.00000108, -0.00000266, -0.00000404, -0.00000510, -0.00000573, -0.00000588, -0.00000553, -0.00000468, -0.00000343, -0.00000186, -0.00000011, 0.00000166, 0.00000333, 0.00000475, 0.00000580, 0.00000643, 0.00000658, 0.00000628, 0.00000557, 0.00000453, 0.00000328, 0.00000194, 0.00000063, -0.00000051, -0.00000141, -0.00000199, -0.00000222, -0.00000210, -0.00000168, -0.00000102, -0.00000020, 0.00000067, 0.00000151, 0.00000223, 0.00000276, 0.00000306, 0.00000311, 0.00000292, 0.00000253, 0.00000198, 0.00000133, 0.00000066, 0.00000002, -0.00000053, -0.00000098, -0.00000129, -0.00000148, -0.00000157, -0.00000158, -0.00000156, -0.00000154, -0.00000156, },
		{ -0.00000047, -0.00000065, -0.00000066, -0.00000056, -0.00000037, -0.00000016, 0.00000003, 0.00000013, 0.00000012, -0.00000002, -0.00000029, -0.00000069, -0.00000117, -0.00000168, -0.00000217, -0.00000258, -0.00000287, -0.00000301, -0.00000296, -0.00000273, -0.00000234, -0.00000183, -0.00000124, -0.00000064, -0.00000007, 0.00000041, 0.00000076, 0.00000095, 0.00000097, 0.00000083, 0.00000056, 0.00000020, -0.00000023, -0.00000067, -0.00000108, -0.00000142, -0.00000168, -0.00000184, -0.00000190, -0.00000187, -0.00000178, -0.00000163, -0.00000144, -0.00000123, -0.00000099, -0.00000072, -0.00000041, -0.00000002, 0.00000045, 0.00000104, 0.00000176, 0.00000260, 0.00000354, 0.00000456, 0.00000559, 0.00000656, 0.00000739, 0.00000800, 0.00000830, 0.00000824, 0.00000776, 0.00000686, 0.00000556, 0.00000390, 0.00000198, -0.00000011, -0.00000222, -0.00000423, -0.00000601, -0.00000744, -0.00000845, -0.00000898, -0.00000903, -0.00000864, -0.00000787, -0.00000684, -0.00000565, -0.00000445, -0.00000337, -0.00000250, -0.00000194, -0.00000172, -0.00000185, -0.00000230, -0.00000298, -0.00000380, -0.00000463, -0.00000537, -0.00000588, -0.00000608, -0.00000591, -0.00000535, -0.00000441, -0.00000314, -0.00000165, -0.00000004, 0.00000157, 0.00000304, 0.00000427, 0.00000517, 0.00000567, 0.00000578, 0.00000550, 0.00000489, 0.00000404, 0.00000305, 0.00000204, 0.00000111, 0.00000036, -0.00000015, -0.00000038, -0.00000031, 0.00000002, 0.00000056, 0.00000124, 0.00000197, 0.00000266, 0.00000322, 0.00000361, 0.00000378, 0.00000371, 0.00000341, 0.00000293, 0.00000231, 0.00000163, 0.00000096, 0.00000035, -0.00000014, },
		{ 0.00000334, 0.00000299, 0.00000245, 0.00000178, 0.00000106, 0.00000036, -0.00000023, -0.00000066, -0.00000088, -0.00000088, -0.00000067, -0.00000028, 0.00000023, 0.00000079, 0.00000133, 0.00000178, 0.00000209, 0.00000221, 0.00000213, 0.00000185, 0.00000141, 0.00000085, 0.00000023, -0.00000038, -0.00000093, -0.00000136, -0.00000163, -0.00000172, -0.00000164, -0.00000141, -0.00000107, -0.00000065, -0.00000023, 0.00000016, 0.00000048, 0.00000068, 0.00000077, 0.00000074, 0.00000061, 0.00000039, 0.00000013, -0.00000015, -0.00000042, -0.00000067, -0.00000090, -0.00000112, -0.00000136, -0.00000165, -0.00000202, -0.00000250, -0.00000311, -0.00000386, -0.00000471, -0.00000564, -0.00000658, -0.00000744, -0.00000814, -0.00000857, -0.00000866, -0.00000833, -0.00000754, -0.00000629, -0.00000460, -0.00000255, -0.00000025, 0.00000218, 0.00000459, 0.00000681, 0.00000872, 0.00001017, 0.00001109, 0.00001142, 0.00001116, 0.00001036, 0.00000911, 0.00000751, 0.00000571, 0.00000386, 0.00000210, 0.00000055, -0.00000070, -0.00000158, -0.00000210, -0.00000227, -0.00000216, -0.00000187, -0.00000148, -0.00000112, -0.00000087, -0.00000082, -0.00000100, -0.00000142, -0.00000206, -0.00000286, -0.00000373, -0.00000458, -0.00000531, -0.00000583, -0.00000606, -0.00000595, -0.00000551, -0.00000475, -0.00000373, -0.00000254, -0.00000129, -0.00000007, 0.00000101, 0.00000185, 0.00000240, 0.00000262, 0.00000252, 0.00000215, 0.00000156, 0.00000084, 0.00000010, -0.00000057, -0.00000109, -0.00000138, -0.00000142, -0.00000118, -0.00000071, -0.00000005, 0.00000073, 0.00000154, 0.00000228, 0.00000289, 0.00000329, 0.00000344, },
		{ 0.00000019, 0.00000103, 0.00000197, 0.00000292, 0.00000377, 0.00000443, 0.00000483, 0.00000493, 0.00000471, 0.00000420, 0.00000343, 0.00000249, 0.00000147, 0.00000047, -0.00000041, -0.00000109, -0.00000152, -0.00000166, -0.00000151, -0.00000111, -0.00000052, 0.00000018, 0.00000092, 0.00000159, 0.00000213, 0.00000248, 0.00000259, 0.00000246, 0.00000210, 0.00000157, 0.00000091, 0.00000020, -0.00000049, -0.00000111, -0.00000160, -0.00000194, -0.00000211, -0.00000212, -0.00000201, -0.00000181, -0.00000157, -0.00000134, -0.00000113, -0.00000097, -0.00000086, -0.00000078, -0.00000070, -0.00000057, -0.00000035, 0.00000001, 0.00000054, 0.00000123, 0.00000207, 0.00000302, 0.00000401, 0.00000494, 0.00000571, 0.00000622, 0.00000637, 0.00000608, 0.00000532, 0.00000408, 0.00000239, 0.00000034, -0.00000196, -0.00000436, -0.00000670, -0.00000882, -0.00001057, -0.00001183, -0.00001251, -0.00001257, -0.00001204, -0.00001096, -0.00000945, -0.00000764, -0.00000569, -0.00000376, -0.00000201, -0.00000057, 0.00000048, 0.00000109, 0.00000124, 0.00000099, 0.00000040, -0.00000042, -0.00000135, -0.00000227, -0.00000309, -0.00000371, -0.00000408, -0.00000418, -0.00000405, -0.00000371, -0.00000325, -0.00000276, -0.00000231, -0.00000198, -0.00000183, -0.00000190, -0.00000217, -0.00000262, -0.00000320, -0.00000382, -0.00000440, -0.00000485, -0.00000511, -0.00000511, -0.00000484, -0.00000429, -0.00000352, -0.00000257, -0.00000154, -0.00000052, 0.00000040, 0.00000114, 0.00000164, 0.00000187, 0.00000182, 0.00000153, 0.00000105, 0.00000048, -0.00000011, -0.00000061, -0.00000095, -0.00000105, -0.00000090, -0.00000047, },
		{ -0.00000140, -0.00000205, -0.00000267, -0.00000317, -0.00000345, -0.00000346, -0.00000314, -0.00000251, -0.00000159, -0.00000046, 0.00000080, 0.00000207, 0.00000323, 0.00000417, 0.00000480, 0.00000506, 0.00000492, 0.00000440, 0.00000354, 0.00000242, 0.00000117, -0.00000011, -0.00000130, -0.00000228, -0.00000298, -0.00000335, -0.00000336, -0.00000305, -0.00000246, -0.00000168, -0.00000082, 0.00000004, 0.00000079, 0.00000134, 0.00000166, 0.00000170, 0.00000149, 0.00000104, 0.00000043, -0.00000029, -0.00000104, -0.00000175, -0.00000237, -0.00000287, -0.00000325, -0.00000353, -0.00000376, -0.00000398, -0.00000426, -0.00000467, -0.00000524, -0.00000598, -0.00000689, -0.00000792, -0.00000900, -0.00001003, -0.00001089, -0.00001147, -0.00001166, -0.00001138, -0.00001057, -0.00000924, -0.00000741, -0.00000518, -0.00000267, -0.00000004, 0.00000256, 0.00000493, 0.00000694, 0.00000845, 0.00000937, 0.00000968, 0.00000939, 0.00000857, 0.00000732, 0.00000581, 0.00000417, 0.00000258, 0.00000119, 0.00000011, -0.00000057, -0.00000083, -0.00000066, -0.00000014, 0.00000066, 0.00000161, 0.00000258, 0.00000347, 0.00000415, 0.00000455, 0.00000463, 0.00000439, 0.00000385, 0.00000307, 0.00000215, 0.00000118, 0.00000025, -0.00000055, -0.00000116, -0.00000154, -0.00000169, -0.00000164, -0.00000144, -0.00000115, -0.00000086, -0.00000064, -0.00000054, -0.00000061, -0.00000087, -0.00000130, -0.00000186, -0.00000249, -0.00000313, -0.00000368, -0.00000410, -0.00000431, -0.00000429, -0.00000404, -0.00000358, -0.00000296, -0.00000225, -0.00000154, -0.00000091, -0.00000044, -0.00000018, -0.00000016, -0.00000038, -0.00000082, },
		{ -0.00000215, -0.00000160, -0.00000113, -0.00000082, -0.00000073, -0.00000089, -0.00000129, -0.00000191, -0.00000267, -0.00000349, -0.00000424, -0.00000483, -0.00000515, -0.00000512, -0.00000469, -0.00000385, -0.00000264, -0.00000112, 0.00000059, 0.00000237, 0.00000407, 0.00000555, 0.00000669, 0.00000740, 0.00000763, 0.00000738, 0.00000669, 0.00000564, 0.00000435, 0.00000296, 0.00000161, 0.00000044, -0.00000043, -0.00000093, -0.00000102, -0.00000071, -0.00000005, 0.00000089, 0.00000199, 0.00000314, 0.00000422, 0.00000513, 0.00000581, 0.00000621, 0.00000635, 0.00000627, 0.00000602, 0.00000572, 0.00000546, 0.00000533, 0.00000540, 0.00000571, 0.00000628, 0.00000706, 0.00000797, 0.00000891, 0.00000974, 0.00001034, 0.00001057, 0.00001033, 0.00000955, 0.00000822, 0.00000637, 0.00000408, 0.00000148, -0.00000126, -0.00000398, -0.00000648, -0.00000859, -0.00001019, -0.00001118, -0.00001152, -0.00001122, -0.00001036, -0.00000905, -0.00000742, -0.00000566, -0.00000391, -0.00000234, -0.00000105, -0.00000012, 0.00000041, 0.00000055, 0.00000037, -0.00000005, -0.00000060, -0.00000116, -0.00000162, -0.00000189, -0.00000191, -0.00000165, -0.00000113, -0.00000040, 0.00000048, 0.00000141, 0.00000229, 0.00000302, 0.00000353, 0.00000377, 0.00000371, 0.00000337, 0.00000280, 0.00000205, 0.00000122, 0.00000039, -0.00000036, -0.00000096, -0.00000136, -0.00000156, -0.00000155, -0.00000139, -0.00000112, -0.00000081, -0.00000055, -0.00000038, -0.00000035, -0.00000050, -0.00000081, -0.00000127, -0.00000182, -0.00000240, -0.00000293, -0.00000336, -0.00000362, -0.00000368, -0.00000353, -0.00000319, -0.00000271, },
		{ 0.00000023, -0.00000021, -0.00000056, -0.00000078, -0.00000080, -0.00000064, -0.00000029, 0.00000019, 0.00000074, 0.00000128, 0.00000171, 0.00000194, 0.00000192, 0.00000161, 0.00000100, 0.00000012, -0.00000095, -0.00000211, -0.00000323, -0.00000418, -0.00000483, -0.00000507, -0.00000483, -0.00000409, -0.00000286, -0.00000121, 0.00000074, 0.00000285, 0.00000494, 0.00000684, 0.00000838, 0.00000945, 0.00000995, 0.00000985, 0.00000917, 0.00000799, 0.00000643, 0.00000465, 0.00000283, 0.00000111, -0.00000033, -0.00000139, -0.00000201, -0.00000218, -0.00000193, -0.00000137, -0.00000061, 0.00000017, 0.00000084, 0.00000125, 0.00000129, 0.00000090, 0.00000007, -0.00000115, -0.00000266, -0.00000431, -0.00000594, -0.00000737, -0.00000841, -0.00000892, -0.00000882, -0.00000805, -0.00000664, -0.00000466, -0.00000225, 0.00000041, 0.00000313, 0.00000569, 0.00000791, 0.00000962, 0.00001071, 0.00001113, 0.00001089, 0.00001004, 0.00000870, 0.00000701, 0.00000514, 0.00000327, 0.00000154, 0.00000010, -0.00000099, -0.00000167, -0.00000196, -0.00000190, -0.00000157, -0.00000107, -0.00000052, -0.00000001, 0.00000037, 0.00000056, 0.00000054, 0.00000033, -0.00000003, -0.00000047, -0.00000092, -0.00000128, -0.00000149, -0.00000149, -0.00000125, -0.00000079, -0.00000011, 0.00000070, 0.00000157, 0.00000242, 0.00000315, 0.00000370, 0.00000400, 0.00000404, 0.00000380, 0.00000334, 0.00000270, 0.00000196, 0.00000120, 0.00000051, -0.00000004, -0.00000040, -0.00000056, -0.00000050, -0.00000027, 0.00000010, 0.00000052, 0.00000093, 0.00000127, 0.00000147, 0.00000151, 0.00000137, 0.00000108, 0.00000068, },
		{ -0.00000016, 0.00000020, 0.00000050, 0.00000068, 0.00000070, 0.00000052, 0.00000018, -0.00000031, -0.00000087, -0.00000143, -0.00000192, -0.00000228, -0.00000245, -0.00000240, -0.00000215, -0.00000173, -0.00000122, -0.00000071, -0.00000030, -0.00000009, -0.00000017, -0.00000059, -0.00000137, -0.00000249, -0.00000388, -0.00000543, -0.00000698, -0.00000839, -0.00000948, -0.00001011, -0.00001018, -0.00000960, -0.00000838, -0.00000657, -0.00000428, -0.00000167, 0.00000107, 0.00000372, 0.00000607, 0.00000795, 0.00000921, 0.00000979, 0.00000967, 0.00000891, 0.00000765, 0.00000605, 0.00000434, 0.00000274, 0.00000145, 0.00000066, 0.00000049, 0.00000100, 0.00000216, 0.00000388, 0.00000600, 0.00000831, 0.00001057, 0.00001253, 0.00001397, 0.00001469, 0.00001457, 0.00001356, 0.00001169, 0.00000906, 0.00000584, 0.00000225, -0.00000145, -0.00000501, -0.00000819, -0.00001078, -0.00001265, -0.00001371, -0.00001395, -0.00001343, -0.00001228, -0.00001065, -0.00000873, -0.00000672, -0.00000479, -0.00000310, -0.00000175, -0.00000081, -0.00000028, -0.00000013, -0.00000029, -0.00000066, -0.00000113, -0.00000160, -0.00000198, -0.00000221, -0.00000225, -0.00000211, -0.00000183, -0.00000144, -0.00000102, -0.00000064, -0.00000035, -0.00000019, -0.00000018, -0.00000032, -0.00000056, -0.00000086, -0.00000115, -0.00000135, -0.00000140, -0.00000127, -0.00000091, -0.00000033, 0.00000043, 0.00000132, 0.00000226, 0.00000317, 0.00000396, 0.00000455, 0.00000488, 0.00000492, 0.00000467, 0.00000416, 0.00000343, 0.00000258, 0.00000168, 0.00000083, 0.00000010, -0.00000045, -0.00000078, -0.00000088, -0.00000077, -0.00000051, },
		{ 0.00000127, 0.00000038, -0.00000040, -0.00000099, -0.00000134, -0.00000143, -0.00000128, -0.00000092, -0.00000043, 0.00000011, 0.00000061, 0.00000100, 0.00000122, 0.00000123, 0.00000102, 0.00000064, 0.00000013, -0.00000042, -0.00000094, -0.00000133, -0.00000152, -0.00000149, -0.00000121, -0.00000073, -0.00000010, 0.00000056, 0.00000115, 0.00000152, 0.00000158, 0.00000122, 0.00000040, -0.00000088, -0.00000258, -0.00000457, -0.00000671, -0.00000881, -0.00001066, -0.00001208, -0.00001288, -0.00001296, -0.00001226, -0.00001079, -0.00000864, -0.00000597, -0.00000302, -0.00000003, 0.00000271, 0.00000496, 0.00000647, 0.00000710, 0.00000676, 0.00000545, 0.00000328, 0.00000044, -0.00000281, -0.00000619, -0.00000935, -0.00001200, -0.00001387, -0.00001473, -0.00001448, -0.00001308, -0.00001061, -0.00000723, -0.00000319, 0.00000122, 0.00000569, 0.00000988, 0.00001352, 0.00001635, 0.00001823, 0.00001906, 0.00001886, 0.00001770, 0.00001576, 0.00001323, 0.00001036, 0.00000739, 0.00000454, 0.00000199, -0.00000010, -0.00000166, -0.00000267, -0.00000315, -0.00000320, -0.00000290, -0.00000240, -0.00000181, -0.00000124, -0.00000078, -0.00000048, -0.00000035, -0.00000040, -0.00000057, -0.00000082, -0.00000109, -0.00000133, -0.00000150, -0.00000158, -0.00000156, -0.00000146, -0.00000134, -0.00000122, -0.00000115, -0.00000119, -0.00000135, -0.00000164, -0.00000204, -0.00000251, -0.00000300, -0.00000343, -0.00000373, -0.00000384, -0.00000372, -0.00000333, -0.00000267, -0.00000179, -0.00000074, 0.00000040, 0.00000153, 0.00000255, 0.00000337, 0.00000392, 0.00000416, 0.00000406, 0.00000366, 0.00000301, 0.00000218, },
		{ 0.00000277, 0.00000378, 0.00000451, 0.00000489, 0.00000489, 0.00000452, 0.00000382, 0.00000287, 0.00000177, 0.00000065, -0.00000040, -0.00000126, -0.00000186, -0.00000216, -0.00000216, -0.00000189, -0.00000141, -0.00000082, -0.00000022, 0.00000031, 0.00000066, 0.00000079, 0.00000068, 0.00000034, -0.00000018, -0.00000079, -0.00000140, -0.00000190, -0.00000219, -0.00000220, -0.00000191, -0.00000130, -0.00000046, 0.00000054, 0.00000155, 0.00000242, 0.00000300, 0.00000314, 0.00000275, 0.00000177, 0.00000023, -0.00000178, -0.00000413, -0.00000660, -0.00000893, -0.00001088, -0.00001219, -0.00001264, -0.00001208, -0.00001043, -0.00000773, -0.00000408, 0.00000029, 0.00000511, 0.00001002, 0.00001466, 0.00001866, 0.00002169, 0.00002350, 0.00002393, 0.00002290, 0.00002048, 0.00001684, 0.00001223, 0.00000699, 0.00000151, -0.00000383, -0.00000866, -0.00001266, -0.00001558, -0.00001728, -0.00001772, -0.00001696, -0.00001513, -0.00001248, -0.00000928, -0.00000581, -0.00000238, 0.00000075, 0.00000338, 0.00000534, 0.00000658, 0.00000707, 0.00000689, 0.00000616, 0.00000501, 0.00000361, 0.00000214, 0.00000073, -0.00000050, -0.00000146, -0.00000214, -0.00000251, -0.00000263, -0.00000253, -0.00000230, -0.00000200, -0.00000168, -0.00000139, -0.00000115, -0.00000096, -0.00000082, -0.00000069, -0.00000055, -0.00000038, -0.00000015, 0.00000012, 0.00000043, 0.00000075, 0.00000102, 0.00000121, 0.00000126, 0.00000114, 0.00000083, 0.00000034, -0.00000032, -0.00000107, -0.00000186, -0.00000260, -0.00000320, -0.00000358, -0.00000367, -0.00000345, -0.00000289, -0.00000204, -0.00000095, 0.00000029, 0.00000157, },
		{ -0.00000581, -0.00000608, -0.00000597, -0.00000545, -0.00000453, -0.00000326, -0.00000173, -0.00000007, 0.00000158, 0.00000310, 0.00000435, 0.00000523, 0.00000568, 0.00000568, 0.00000525, 0.00000447, 0.00000343, 0.00000227, 0.00000112, 0.00000011, -0.00000065, -0.00000109, -0.00000119, -0.00000094, -0.00000042, 0.00000029, 0.00000107, 0.00000181, 0.00000239, 0.00000271, 0.00000272, 0.00000242, 0.00000183, 0.00000103, 0.00000014, -0.00000072, -0.00000142, -0.00000183, -0.00000189, -0.00000155, -0.00000083, 0.00000016, 0.00000131, 0.00000243, 0.00000331, 0.00000375, 0.00000355, 0.00000259, 0.00000078, -0.00000187, -0.00000526, -0.00000922, -0.00001349, -0.00001774, -0.00002164, -0.00002485, -0.00002703, -0.00002795, -0.00002743, -0.00002541, -0.00002195, -0.00001721, -0.00001147, -0.00000507, 0.00000159, 0.00000807, 0.00001396, 0.00001892, 0.00002265, 0.00002497, 0.00002580, 0.00002520, 0.00002331, 0.00002037, 0.00001668, 0.00001259, 0.00000844, 0.00000454, 0.00000118, -0.00000144, -0.00000321, -0.00000410, -0.00000415, -0.00000348, -0.00000226, -0.00000069, 0.00000101, 0.00000266, 0.00000407, 0.00000512, 0.00000574, 0.00000590, 0.00000562, 0.00000497, 0.00000403, 0.00000290, 0.00000169, 0.00000049, -0.00000062, -0.00000161, -0.00000243, -0.00000310, -0.00000363, -0.00000404, -0.00000435, -0.00000459, -0.00000478, -0.00000491, -0.00000498, -0.00000497, -0.00000487, -0.00000464, -0.00000428, -0.00000379, -0.00000320, -0.00000254, -0.00000186, -0.00000124, -0.00000073, -0.00000040, -0.00000032, -0.00000049, -0.00000094, -0.00000162, -0.00000248, -0.00000344, -0.00000439, -0.00000521, },
		{ 0.00000228, 0.00000228, 0.00000191, 0.00000117, 0.00000013, -0.00000114, -0.00000252, -0.00000386, -0.00000502, -0.00000587, -0.00000632, -0.00000628, -0.00000574, -0.00000473, -0.00000333, -0.00000165, 0.00000016, 0.00000193, 0.00000350, 0.00000472, 0.00000550, 0.00000576, 0.00000551, 0.00000478, 0.00000368, 0.00000233, 0.00000090, -0.00000047, -0.00000161, -0.00000242, -0.00000281, -0.00000278, -0.00000234, -0.00000158, -0.00000062, 0.00000039, 0.00000130, 0.00000198, 0.00000233, 0.00000228, 0.00000183, 0.00000105, 0.00000005, -0.00000103, -0.00000200, -0.00000268, -0.00000290, -0.00000255, -0.00000156, 0.00000007, 0.00000223, 0.00000478, 0.00000749, 0.00001009, 0.00001229, 0.00001382, 0.00001443, 0.00001395, 0.00001230, 0.00000947, 0.00000558, 0.00000084, -0.00000447, -0.00000999, -0.00001535, -0.00002018, -0.00002414, -0.00002696, -0.00002845, -0.00002854, -0.00002725, -0.00002474, -0.00002122, -0.00001699, -0.00001240, -0.00000780, -0.00000353, 0.00000014, 0.00000298, 0.00000485, 0.00000572, 0.00000564, 0.00000472, 0.00000316, 0.00000119, -0.00000096, -0.00000304, -0.00000485, -0.00000623, -0.00000708, -0.00000736, -0.00000708, -0.00000632, -0.00000518, -0.00000379, -0.00000230, -0.00000084, 0.00000049, 0.00000159, 0.00000241, 0.00000293, 0.00000316, 0.00000315, 0.00000292, 0.00000255, 0.00000208, 0.00000155, 0.00000100, 0.00000043, -0.00000013, -0.00000069, -0.00000125, -0.00000182, -0.00000239, -0.00000293, -0.00000343, -0.00000383, -0.00000411, -0.00000421, -0.00000411, -0.00000377, -0.00000321, -0.00000245, -0.00000154, -0.00000056, 0.00000042, 0.00000127, 0.00000192, },
		{ -0.00000623, -0.00000592, -0.00000526, -0.00000432, -0.00000315, -0.00000188, -0.00000063, 0.00000047, 0.00000128, 0.00000172, 0.00000170, 0.00000120, 0.00000024, -0.00000110, -0.00000270, -0.00000442, -0.00000607, -0.00000750, -0.00000854, -0.00000907, -0.00000903, -0.00000839, -0.00000719, -0.00000555, -0.00000360, -0.00000153, 0.00000046, 0.00000219, 0.00000349, 0.00000424, 0.00000439, 0.00000392, 0.00000291, 0.00000148, -0.00000021, -0.00000196, -0.00000359, -0.00000492, -0.00000582, -0.00000620, -0.00000606, -0.00000545, -0.00000448, -0.00000332, -0.00000216, -0.00000122, -0.00000067, -0.00000067, -0.00000131, -0.00000261, -0.00000452, -0.00000689, -0.00000955, -0.00001224, -0.00001470, -0.00001668, -0.00001795, -0.00001833, -0.00001774, -0.00001617, -0.00001369, -0.00001048, -0.00000677, -0.00000284, 0.00000100, 0.00000446, 0.00000727, 0.00000923, 0.00001021, 0.00001020, 0.00000924, 0.00000749, 0.00000515, 0.00000249, -0.00000021, -0.00000268, -0.00000467, -0.00000599, -0.00000652, -0.00000622, -0.00000514, -0.00000340, -0.00000117, 0.00000134, 0.00000389, 0.00000625, 0.00000823, 0.00000967, 0.00001047, 0.00001061, 0.00001010, 0.00000905, 0.00000757, 0.00000582, 0.00000396, 0.00000217, 0.00000057, -0.00000073, -0.00000166, -0.00000220, -0.00000236, -0.00000219, -0.00000176, -0.00000115, -0.00000044, 0.00000028, 0.00000096, 0.00000154, 0.00000201, 0.00000235, 0.00000258, 0.00000270, 0.00000272, 0.00000267, 0.00000254, 0.00000233, 0.00000202, 0.00000160, 0.00000106, 0.00000039, -0.00000042, -0.00000135, -0.00000235, -0.00000337, -0.00000434, -0.00000519, -0.00000583, -0.00000619, },
		{ 0.00000118, 0.00000032, -0.00000077, -0.00000203, -0.00000336, -0.00000466, -0.00000581, -0.00000669, -0.00000720, -0.00000727, -0.00000687, -0.00000601, -0.00000475, -0.00000320, -0.00000150, 0.00000018, 0.00000169, 0.00000285, 0.00000355, 0.00000369, 0.00000325, 0.00000226, 0.00000082, -0.00000094, -0.00000283, -0.00000463, -0.00000615, -0.00000720, -0.00000764, -0.00000737, -0.00000640, -0.00000477, -0.00000260, -0.00000008, 0.00000258, 0.00000514, 0.00000737, 0.00000909, 0.00001016, 0.00001050, 0.00001013, 0.00000913, 0.00000766, 0.00000593, 0.00000418, 0.00000266, 0.00000160, 0.00000119, 0.00000155, 0.00000270, 0.00000461, 0.00000713, 0.00001007, 0.00001315, 0.00001610, 0.00001863, 0.00002047, 0.00002142, 0.00002135, 0.00002020, 0.00001802, 0.00001494, 0.00001116, 0.00000695, 0.00000258, -0.00000166, -0.00000550, -0.00000874, -0.00001121, -0.00001285, -0.00001367, -0.00001375, -0.00001322, -0.00001227, -0.00001110, -0.00000991, -0.00000888, -0.00000814, -0.00000777, -0.00000779, -0.00000816, -0.00000878, -0.00000954, -0.00001027, -0.00001083, -0.00001107, -0.00001090, -0.00001025, -0.00000911, -0.00000754, -0.00000560, -0.00000344, -0.00000119, 0.00000098, 0.00000294, 0.00000455, 0.00000573, 0.00000643, 0.00000664, 0.00000639, 0.00000576, 0.00000484, 0.00000374, 0.00000257, 0.00000143, 0.00000040, -0.00000046, -0.00000112, -0.00000157, -0.00000182, -0.00000191, -0.00000187, -0.00000174, -0.00000156, -0.00000134, -0.00000109, -0.00000083, -0.00000054, -0.00000021, 0.00000015, 0.00000055, 0.00000098, 0.00000140, 0.00000178, 0.00000206, 0.00000218, 0.00000210, 0.00000178, },
		{ 0.00000101, 0.00000178, 0.00000268, 0.00000366, 0.00000463, 0.00000550, 0.00000616, 0.00000651, 0.00000650, 0.00000607, 0.00000524, 0.00000403, 0.00000255, 0.00000091, -0.00000074, -0.00000224, -0.00000344, -0.00000420, -0.00000444, -0.00000409, -0.00000319, -0.00000180, -0.00000005, 0.00000189, 0.00000382, 0.00000552, 0.00000680, 0.00000751, 0.00000754, 0.00000685, 0.00000548, 0.00000353, 0.00000117, -0.00000136, -0.00000384, -0.00000600, -0.00000763, -0.00000856, -0.00000868, -0.00000796, -0.00000646, -0.00000435, -0.00000182, 0.00000084, 0.00000336, 0.00000543, 0.00000681, 0.00000730, 0.00000679, 0.00000527, 0.00000280, -0.00000043, -0.00000417, -0.00000811, -0.00001191, -0.00001525, -0.00001780, -0.00001934, -0.00001970, -0.00001882, -0.00001673, -0.00001357, -0.00000955, -0.00000496, -0.00000011, 0.00000468, 0.00000910, 0.00001289, 0.00001586, 0.00001787, 0.00001890, 0.00001898, 0.00001822, 0.00001676, 0.00001481, 0.00001256, 0.00001020, 0.00000790, 0.00000577, 0.00000390, 0.00000232, 0.00000101, -0.00000007, -0.00000099, -0.00000182, -0.00000265, -0.00000351, -0.00000443, -0.00000542, -0.00000643, -0.00000741, -0.00000828, -0.00000896, -0.00000937, -0.00000945, -0.00000917, -0.00000851, -0.00000750, -0.00000620, -0.00000469, -0.00000306, -0.00000143, 0.00000011, 0.00000146, 0.00000257, 0.00000338, 0.00000389, 0.00000411, 0.00000408, 0.00000385, 0.00000349, 0.00000305, 0.00000260, 0.00000218, 0.00000180, 0.00000149, 0.00000123, 0.00000101, 0.00000082, 0.00000062, 0.00000042, 0.00000022, 0.00000003, -0.00000013, -0.00000020, -0.00000016, 0.00000004, 0.00000043, },
		{ 0.00000126, 0.00000053, -0.00000031, -0.00000120, -0.00000207, -0.00000282, -0.00000336, -0.00000363, -0.00000355, -0.00000313, -0.00000237, -0.00000133, -0.00000011, 0.00000117, 0.00000236, 0.00000331, 0.00000391, 0.00000404, 0.00000365, 0.00000272, 0.00000130, -0.00000050, -0.00000252, -0.00000459, -0.00000650, -0.00000806, -0.00000909, -0.00000947, -0.00000914, -0.00000809, -0.00000642, -0.00000427, -0.00000183, 0.00000065, 0.00000292, 0.00000474, 0.00000592, 0.00000631, 0.00000587, 0.00000461, 0.00000266, 0.00000022, -0.00000246, -0.00000507, -0.00000730, -0.00000887, -0.00000953, -0.00000911, -0.00000755, -0.00000488, -0.00000124, 0.00000313, 0.00000792, 0.00001277, 0.00001729, 0.00002112, 0.00002393, 0.00002548, 0.00002561, 0.00002429, 0.00002157, 0.00001764, 0.00001275, 0.00000724, 0.00000147, -0.00000419, -0.00000939, -0.00001383, -0.00001729, -0.00001963, -0.00002080, -0.00002083, -0.00001985, -0.00001803, -0.00001559, -0.00001278, -0.00000982, -0.00000692, -0.00000428, -0.00000201, -0.00000020, 0.00000113, 0.00000199, 0.00000243, 0.00000252, 0.00000233, 0.00000196, 0.00000146, 0.00000089, 0.00000029, -0.00000033, -0.00000097, -0.00000163, -0.00000232, -0.00000304, -0.00000379, -0.00000455, -0.00000528, -0.00000594, -0.00000650, -0.00000688, -0.00000706, -0.00000701, -0.00000670, -0.00000614, -0.00000537, -0.00000442, -0.00000336, -0.00000225, -0.00000117, -0.00000017, 0.00000070, 0.00000139, 0.00000190, 0.00000224, 0.00000243, 0.00000250, 0.00000251, 0.00000248, 0.00000246, 0.00000247, 0.00000251, 0.00000256, 0.00000260, 0.00000259, 0.00000248, 0.00000224, 0.00000184, },
		{ -0.00000056, 0.00000015, 0.00000096, 0.00000180, 0.00000259, 0.00000324, 0.00000366, 0.00000379, 0.00000357, 0.00000302, 0.00000215, 0.00000104, -0.00000021, -0.00000147, -0.00000260, -0.00000346, -0.00000394, -0.00000397, -0.00000350, -0.00000255, -0.00000120, 0.00000044, 0.00000219, 0.00000389, 0.00000533, 0.00000633, 0.00000677, 0.00000654, 0.00000562, 0.00000406, 0.00000197, -0.00000047, -0.00000305, -0.00000553, -0.00000765, -0.00000920, -0.00001002, -0.00001001, -0.00000915, -0.00000753, -0.00000530, -0.00000269, 0.00000001, 0.00000250, 0.00000446, 0.00000563, 0.00000580, 0.00000486, 0.00000277, -0.00000037, -0.00000437, -0.00000895, -0.00001375, -0.00001839, -0.00002247, -0.00002564, -0.00002760, -0.00002812, -0.00002710, -0.00002455, -0.00002059, -0.00001546, -0.00000947, -0.00000299, 0.00000357, 0.00000981, 0.00001537, 0.00001995, 0.00002334, 0.00002540, 0.00002612, 0.00002557, 0.00002390, 0.00002134, 0.00001816, 0.00001462, 0.00001100, 0.00000756, 0.00000449, 0.00000195, 0.00000001, -0.00000128, -0.00000196, -0.00000210, -0.00000179, -0.00000116, -0.00000033, 0.00000058, 0.00000148, 0.00000227, 0.00000290, 0.00000334, 0.00000357, 0.00000359, 0.00000341, 0.00000306, 0.00000254, 0.00000190, 0.00000113, 0.00000028, -0.00000066, -0.00000164, -0.00000265, -0.00000365, -0.00000460, -0.00000547, -0.00000621, -0.00000678, -0.00000715, -0.00000729, -0.00000721, -0.00000692, -0.00000643, -0.00000580, -0.00000508, -0.00000433, -0.00000361, -0.00000297, -0.00000246, -0.00000209, -0.00000187, -0.00000178, -0.00000178, -0.00000180, -0.00000180, -0.00000172, -0.00000150, -0.00000111, },
		{ -0.00000107, -0.00000173, -0.00000249, -0.00000326, -0.00000394, -0.00000442, -0.00000462, -0.00000450, -0.00000401, -0.00000317, -0.00000205, -0.00000073, 0.00000067, 0.00000201, 0.00000314, 0.00000394, 0.00000430, 0.00000417, 0.00000353, 0.00000243, 0.00000096, -0.00000074, -0.00000249, -0.00000412, -0.00000543, -0.00000627, -0.00000651, -0.00000610, -0.00000505, -0.00000342, -0.00000135, 0.00000096, 0.00000329, 0.00000541, 0.00000708, 0.00000812, 0.00000840, 0.00000786, 0.00000653, 0.00000452, 0.00000202, -0.00000071, -0.00000339, -0.00000570, -0.00000736, -0.00000812, -0.00000782, -0.00000639, -0.00000383, -0.00000030, 0.00000400, 0.00000874, 0.00001357, 0.00001810, 0.00002195, 0.00002478, 0.00002633, 0.00002643, 0.00002502, 0.00002215, 0.00001799, 0.00001281, 0.00000694, 0.00000077, -0.00000529, -0.00001087, -0.00001562, -0.00001928, -0.00002166, -0.00002270, -0.00002241, -0.00002091, -0.00001840, -0.00001514, -0.00001140, -0.00000750, -0.00000371, -0.00000028, 0.00000259, 0.00000478, 0.00000622, 0.00000692, 0.00000692, 0.00000635, 0.00000533, 0.00000403, 0.00000259, 0.00000117, -0.00000011, -0.00000115, -0.00000188, -0.00000229, -0.00000237, -0.00000213, -0.00000164, -0.00000094, -0.00000010, 0.00000080, 0.00000172, 0.00000259, 0.00000335, 0.00000398, 0.00000442, 0.00000468, 0.00000471, 0.00000453, 0.00000414, 0.00000356, 0.00000280, 0.00000192, 0.00000095, -0.00000005, -0.00000103, -0.00000191, -0.00000265, -0.00000320, -0.00000352, -0.00000361, -0.00000347, -0.00000313, -0.00000263, -0.00000204, -0.00000144, -0.00000090, -0.00000050, -0.00000029, -0.00000031, -0.00000058, },
		{ -0.00000321, -0.00000220, -0.00000119, -0.00000029, 0.00000041, 0.00000082, 0.00000088, 0.00000056, -0.00000011, -0.00000109, -0.00000228, -0.00000356, -0.00000479, -0.00000582, -0.00000651, -0.00000676, -0.00000649, -0.00000568, -0.00000435, -0.00000260, -0.00000055, 0.00000163, 0.00000375, 0.00000562, 0.00000705, 0.00000791, 0.00000811, 0.00000760, 0.00000644, 0.00000473, 0.00000263, 0.00000035, -0.00000187, -0.00000381, -0.00000525, -0.00000603, -0.00000605, -0.00000528, -0.00000379, -0.00000170, 0.00000075, 0.00000332, 0.00000571, 0.00000764, 0.00000882, 0.00000907, 0.00000825, 0.00000631, 0.00000333, -0.00000054, -0.00000505, -0.00000988, -0.00001466, -0.00001902, -0.00002261, -0.00002513, -0.00002633, -0.00002610, -0.00002440, -0.00002133, -0.00001705, -0.00001186, -0.00000609, -0.00000011, 0.00000570, 0.00001099, 0.00001545, 0.00001885, 0.00002106, 0.00002202, 0.00002179, 0.00002051, 0.00001837, 0.00001562, 0.00001254, 0.00000941, 0.00000648, 0.00000395, 0.00000199, 0.00000068, 0.00000005, 0.00000007, 0.00000064, 0.00000164, 0.00000290, 0.00000428, 0.00000560, 0.00000672, 0.00000754, 0.00000798, 0.00000799, 0.00000758, 0.00000679, 0.00000566, 0.00000429, 0.00000277, 0.00000119, -0.00000034, -0.00000176, -0.00000298, -0.00000396, -0.00000466, -0.00000508, -0.00000520, -0.00000506, -0.00000469, -0.00000414, -0.00000345, -0.00000270, -0.00000194, -0.00000123, -0.00000064, -0.00000022, 0.00000000, 0.00000000, -0.00000023, -0.00000067, -0.00000130, -0.00000207, -0.00000290, -0.00000373, -0.00000447, -0.00000505, -0.00000541, -0.00000551, -0.00000532, -0.00000484, -0.00000412, },
		{ 0.00000001, -0.00000082, -0.00000154, -0.00000206, -0.00000231, -0.00000225, -0.00000187, -0.00000119, -0.00000025, 0.00000083, 0.00000196, 0.00000301, 0.00000383, 0.00000431, 0.00000436, 0.00000391, 0.00000295, 0.00000151, -0.00000031, -0.00000240, -0.00000458, -0.00000667, -0.00000848, -0.00000982, -0.00001057, -0.00001061, -0.00000991, -0.00000850, -0.00000648, -0.00000400, -0.00000127, 0.00000149, 0.00000402, 0.00000610, 0.00000754, 0.00000820, 0.00000802, 0.00000702, 0.00000529, 0.00000300, 0.00000040, -0.00000225, -0.00000465, -0.00000652, -0.00000763, -0.00000778, -0.00000687, -0.00000491, -0.00000197, 0.00000175, 0.00000601, 0.00001049, 0.00001483, 0.00001869, 0.00002173, 0.00002369, 0.00002436, 0.00002365, 0.00002155, 0.00001816, 0.00001368, 0.00000838, 0.00000257, -0.00000340, -0.00000916, -0.00001441, -0.00001886, -0.00002232, -0.00002467, -0.00002586, -0.00002595, -0.00002507, -0.00002339, -0.00002113, -0.00001853, -0.00001583, -0.00001325, -0.00001097, -0.00000909, -0.00000771, -0.00000683, -0.00000642, -0.00000639, -0.00000663, -0.00000702, -0.00000741, -0.00000768, -0.00000774, -0.00000751, -0.00000695, -0.00000605, -0.00000485, -0.00000341, -0.00000180, -0.00000013, 0.00000151, 0.00000301, 0.00000430, 0.00000530, 0.00000597, 0.00000628, 0.00000622, 0.00000583, 0.00000514, 0.00000421, 0.00000311, 0.00000192, 0.00000073, -0.00000040, -0.00000139, -0.00000219, -0.00000274, -0.00000302, -0.00000302, -0.00000275, -0.00000224, -0.00000155, -0.00000073, 0.00000014, 0.00000100, 0.00000175, 0.00000234, 0.00000271, 0.00000282, 0.00000266, 0.00000226, 0.00000163, 0.00000086, },
		{ -0.00000157, -0.00000082, -0.00000021, 0.00000020, 0.00000036, 0.00000025, -0.00000012, -0.00000070, -0.00000143, -0.00000222, -0.00000297, -0.00000358, -0.00000395, -0.00000401, -0.00000372, -0.00000307, -0.00000210, -0.00000087, 0.00000050, 0.00000188, 0.00000310, 0.00000403, 0.00000451, 0.00000444, 0.00000375, 0.00000242, 0.00000050, -0.00000193, -0.00000470, -0.00000762, -0.00001046, -0.00001301, -0.00001505, -0.00001639, -0.00001689, -0.00001650, -0.00001523, -0.00001315, -0.00001044, -0.00000730, -0.00000402, -0.00000088, 0.00000183, 0.00000385, 0.00000495, 0.00000500, 0.00000393, 0.00000178, -0.00000134, -0.00000521, -0.00000957, -0.00001409, -0.00001843, -0.00002225, -0.00002525, -0.00002717, -0.00002785, -0.00002718, -0.00002520, -0.00002198, -0.00001772, -0.00001267, -0.00000713, -0.00000142, 0.00000415, 0.00000927, 0.00001369, 0.00001723, 0.00001976, 0.00002125, 0.00002173, 0.00002128, 0.00002007, 0.00001827, 0.00001608, 0.00001372, 0.00001136, 0.00000917, 0.00000727, 0.00000573, 0.00000458, 0.00000382, 0.00000339, 0.00000323, 0.00000325, 0.00000335, 0.00000344, 0.00000346, 0.00000336, 0.00000310, 0.00000270, 0.00000216, 0.00000154, 0.00000090, 0.00000029, -0.00000021, -0.00000054, -0.00000065, -0.00000052, -0.00000013, 0.00000051, 0.00000138, 0.00000241, 0.00000355, 0.00000473, 0.00000587, 0.00000689, 0.00000772, 0.00000830, 0.00000858, 0.00000852, 0.00000814, 0.00000742, 0.00000641, 0.00000517, 0.00000375, 0.00000223, 0.00000071, -0.00000073, -0.00000201, -0.00000306, -0.00000383, -0.00000428, -0.00000441, -0.00000423, -0.00000379, -0.00000314, -0.00000238, },
		{ -0.00000289, -0.00000420, -0.00000523, -0.00000591, -0.00000624, -0.00000620, -0.00000584, -0.00000521, -0.00000439, -0.00000347, -0.00000254, -0.00000170, -0.00000104, -0.00000059, -0.00000039, -0.00000043, -0.00000068, -0.00000107, -0.00000152, -0.00000191, -0.00000215, -0.00000214, -0.00000180, -0.00000108, 0.00000001, 0.00000146, 0.00000317, 0.00000505, 0.00000693, 0.00000866, 0.00001007, 0.00001101, 0.00001135, 0.00001100, 0.00000992, 0.00000816, 0.00000580, 0.00000298, -0.00000008, -0.00000316, -0.00000598, -0.00000831, -0.00000990, -0.00001055, -0.00001014, -0.00000860, -0.00000595, -0.00000229, 0.00000220, 0.00000728, 0.00001265, 0.00001799, 0.00002296, 0.00002725, 0.00003058, 0.00003274, 0.00003359, 0.00003306, 0.00003117, 0.00002804, 0.00002385, 0.00001884, 0.00001329, 0.00000750, 0.00000179, -0.00000358, -0.00000834, -0.00001232, -0.00001537, -0.00001744, -0.00001852, -0.00001867, -0.00001799, -0.00001665, -0.00001481, -0.00001264, -0.00001034, -0.00000805, -0.00000591, -0.00000402, -0.00000246, -0.00000124, -0.00000038, 0.00000017, 0.00000044, 0.00000051, 0.00000043, 0.00000028, 0.00000010, -0.00000006, -0.00000017, -0.00000022, -0.00000022, -0.00000017, -0.00000011, -0.00000006, -0.00000005, -0.00000010, -0.00000023, -0.00000045, -0.00000074, -0.00000110, -0.00000147, -0.00000184, -0.00000216, -0.00000237, -0.00000243, -0.00000232, -0.00000200, -0.00000147, -0.00000072, 0.00000021, 0.00000130, 0.00000247, 0.00000368, 0.00000483, 0.00000586, 0.00000669, 0.00000724, 0.00000748, 0.00000737, 0.00000688, 0.00000605, 0.00000490, 0.00000350, 0.00000192, 0.00000027, -0.00000137, },
		{ 0.00001107, 0.00001140, 0.00001132, 0.00001083, 0.00000995, 0.00000873, 0.00000726, 0.00000561, 0.00000391, 0.00000226, 0.00000076, -0.00000049, -0.00000142, -0.00000200, -0.00000220, -0.00000205, -0.00000159, -0.00000090, -0.00000006, 0.00000082, 0.00000166, 0.00000238, 0.00000289, 0.00000318, 0.00000323, 0.00000305, 0.00000270, 0.00000226, 0.00000181, 0.00000144, 0.00000126, 0.00000133, 0.00000172, 0.00000243, 0.00000346, 0.00000474, 0.00000617, 0.00000764, 0.00000898, 0.00001004, 0.00001065, 0.00001066, 0.00000996, 0.00000845, 0.00000613, 0.00000300, -0.00000083, -0.00000524, -0.00001002, -0.00001495, -0.00001978, -0.00002424, -0.00002808, -0.00003106, -0.00003300, -0.00003376, -0.00003326, -0.00003149, -0.00002851, -0.00002445, -0.00001950, -0.00001387, -0.00000784, -0.00000166, 0.00000436, 0.00000999, 0.00001499, 0.00001916, 0.00002239, 0.00002458, 0.00002571, 0.00002583, 0.00002502, 0.00002341, 0.00002115, 0.00001842, 0.00001542, 0.00001232, 0.00000930, 0.00000649, 0.00000401, 0.00000194, 0.00000031, -0.00000085, -0.00000157, -0.00000191, -0.00000193, -0.00000170, -0.00000132, -0.00000085, -0.00000038, 0.00000004, 0.00000037, 0.00000058, 0.00000066, 0.00000063, 0.00000050, 0.00000030, 0.00000007, -0.00000018, -0.00000041, -0.00000060, -0.00000074, -0.00000084, -0.00000089, -0.00000092, -0.00000095, -0.00000099, -0.00000106, -0.00000116, -0.00000131, -0.00000148, -0.00000166, -0.00000181, -0.00000190, -0.00000188, -0.00000171, -0.00000135, -0.00000078, 0.00000003, 0.00000106, 0.00000229, 0.00000367, 0.00000515, 0.00000665, 0.00000808, 0.00000934, 0.00001036, },
		{ -0.00000192, -0.00000129, -0.00000042, 0.00000065, 0.00000188, 0.00000319, 0.00000448, 0.00000566, 0.00000663, 0.00000730, 0.00000760, 0.00000749, 0.00000695, 0.00000598, 0.00000463, 0.00000297, 0.00000108, -0.00000091, -0.00000291, -0.00000480, -0.00000647, -0.00000784, -0.00000885, -0.00000947, -0.00000968, -0.00000953, -0.00000904, -0.00000830, -0.00000739, -0.00000637, -0.00000535, -0.00000439, -0.00000355, -0.00000286, -0.00000233, -0.00000195, -0.00000168, -0.00000148, -0.00000126, -0.00000097, -0.00000052, 0.00000016, 0.00000110, 0.00000235, 0.00000392, 0.00000577, 0.00000788, 0.00001015, 0.00001251, 0.00001482, 0.00001697, 0.00001881, 0.00002023, 0.00002110, 0.00002132, 0.00002081, 0.00001954, 0.00001750, 0.00001472, 0.00001127, 0.00000727, 0.00000285, -0.00000182, -0.00000655, -0.00001114, -0.00001541, -0.00001916, -0.00002225, -0.00002454, -0.00002595, -0.00002643, -0.00002601, -0.00002472, -0.00002267, -0.00002000, -0.00001687, -0.00001347, -0.00000999, -0.00000662, -0.00000353, -0.00000085, 0.00000130, 0.00000286, 0.00000381, 0.00000417, 0.00000400, 0.00000340, 0.00000248, 0.00000136, 0.00000019, -0.00000093, -0.00000189, -0.00000260, -0.00000303, -0.00000316, -0.00000298, -0.00000255, -0.00000192, -0.00000116, -0.00000035, 0.00000044, 0.00000115, 0.00000172, 0.00000213, 0.00000236, 0.00000243, 0.00000236, 0.00000217, 0.00000192, 0.00000163, 0.00000136, 0.00000111, 0.00000091, 0.00000076, 0.00000065, 0.00000055, 0.00000045, 0.00000031, 0.00000011, -0.00000016, -0.00000050, -0.00000090, -0.00000132, -0.00000174, -0.00000210, -0.00000235, -0.00000243, -0.00000230, },
		{ -0.00000148, -0.00000160, -0.00000182, -0.00000209, -0.00000240, -0.00000267, -0.00000284, -0.00000287, -0.00000269, -0.00000226, -0.00000155, -0.00000059, 0.00000062, 0.00000201, 0.00000352, 0.00000504, 0.00000649, 0.00000775, 0.00000874, 0.00000937, 0.00000959, 0.00000936, 0.00000868, 0.00000757, 0.00000608, 0.00000429, 0.00000229, 0.00000018, -0.00000195, -0.00000397, -0.00000582, -0.00000741, -0.00000867, -0.00000958, -0.00001012, -0.00001029, -0.00001012, -0.00000965, -0.00000893, -0.00000804, -0.00000705, -0.00000604, -0.00000509, -0.00000428, -0.00000369, -0.00000337, -0.00000337, -0.00000372, -0.00000443, -0.00000548, -0.00000684, -0.00000844, -0.00001019, -0.00001199, -0.00001370, -0.00001520, -0.00001636, -0.00001705, -0.00001716, -0.00001663, -0.00001541, -0.00001352, -0.00001098, -0.00000791, -0.00000444, -0.00000072, 0.00000304, 0.00000666, 0.00000994, 0.00001270, 0.00001480, 0.00001614, 0.00001669, 0.00001645, 0.00001549, 0.00001392, 0.00001190, 0.00000961, 0.00000724, 0.00000499, 0.00000303, 0.00000148, 0.00000045, -0.00000002, 0.00000005, 0.00000059, 0.00000152, 0.00000270, 0.00000397, 0.00000517, 0.00000618, 0.00000686, 0.00000713, 0.00000697, 0.00000635, 0.00000534, 0.00000401, 0.00000247, 0.00000084, -0.00000075, -0.00000219, -0.00000338, -0.00000425, -0.00000476, -0.00000492, -0.00000476, -0.00000433, -0.00000372, -0.00000300, -0.00000226, -0.00000159, -0.00000105, -0.00000067, -0.00000048, -0.00000046, -0.00000059, -0.00000083, -0.00000112, -0.00000142, -0.00000169, -0.00000187, -0.00000197, -0.00000197, -0.00000189, -0.00000177, -0.00000162, -0.00000151, -0.00000145, },
		{ -0.00000096, -0.00000138, -0.00000167, -0.00000185, -0.00000192, -0.00000193, -0.00000191, -0.00000192, -0.00000199, -0.00000216, -0.00000244, -0.00000282, -0.00000328, -0.00000377, -0.00000424, -0.00000463, -0.00000486, -0.00000487, -0.00000463, -0.00000411, -0.00000331, -0.00000224, -0.00000096, 0.00000047, 0.00000196, 0.00000344, 0.00000482, 0.00000599, 0.00000691, 0.00000750, 0.00000773, 0.00000758, 0.00000705, 0.00000616, 0.00000496, 0.00000349, 0.00000182, 0.00000002, -0.00000184, -0.00000366, -0.00000537, -0.00000686, -0.00000804, -0.00000882, -0.00000911, -0.00000884, -0.00000794, -0.00000636, -0.00000410, -0.00000118, 0.00000235, 0.00000636, 0.00001072, 0.00001523, 0.00001966, 0.00002377, 0.00002729, 0.00002999, 0.00003164, 0.00003207, 0.00003117, 0.00002889, 0.00002528, 0.00002047, 0.00001466, 0.00000811, 0.00000117, -0.00000582, -0.00001251, -0.00001854, -0.00002363, -0.00002755, -0.00003014, -0.00003136, -0.00003123, -0.00002989, -0.00002751, -0.00002436, -0.00002071, -0.00001687, -0.00001311, -0.00000966, -0.00000670, -0.00000437, -0.00000269, -0.00000164, -0.00000113, -0.00000104, -0.00000119, -0.00000142, -0.00000156, -0.00000148, -0.00000109, -0.00000034, 0.00000076, 0.00000214, 0.00000370, 0.00000532, 0.00000684, 0.00000814, 0.00000908, 0.00000958, 0.00000960, 0.00000912, 0.00000821, 0.00000693, 0.00000540, 0.00000376, 0.00000214, 0.00000068, -0.00000053, -0.00000140, -0.00000190, -0.00000204, -0.00000184, -0.00000137, -0.00000073, -0.00000001, 0.00000069, 0.00000128, 0.00000170, 0.00000190, 0.00000188, 0.00000164, 0.00000124, 0.00000071, 0.00000013, -0.00000045, },
		{ -0.00000209, -0.00000137, -0.00000088, -0.00000065, -0.00000063, -0.00000080, -0.00000108, -0.00000140, -0.00000170, -0.00000192, -0.00000202, -0.00000199, -0.00000184, -0.00000158, -0.00000126, -0.00000093, -0.00000064, -0.00000044, -0.00000034, -0.00000037, -0.00000051, -0.00000075, -0.00000103, -0.00000130, -0.00000152, -0.00000161, -0.00000155, -0.00000129, -0.00000083, -0.00000016, 0.00000068, 0.00000165, 0.00000271, 0.00000380, 0.00000486, 0.00000583, 0.00000667, 0.00000734, 0.00000779, 0.00000800, 0.00000793, 0.00000757, 0.00000686, 0.00000578, 0.00000429, 0.00000234, -0.00000009, -0.00000302, -0.00000644, -0.00001031, -0.00001455, -0.00001903, -0.00002358, -0.00002798, -0.00003198, -0.00003530, -0.00003766, -0.00003882, -0.00003855, -0.00003671, -0.00003322, -0.00002812, -0.00002154, -0.00001371, -0.00000498, 0.00000426, 0.00001353, 0.00002237, 0.00003029, 0.00003688, 0.00004180, 0.00004479, 0.00004575, 0.00004469, 0.00004175, 0.00003719, 0.00003135, 0.00002466, 0.00001756, 0.00001051, 0.00000391, -0.00000189, -0.00000664, -0.00001018, -0.00001248, -0.00001359, -0.00001366, -0.00001290, -0.00001157, -0.00000993, -0.00000823, -0.00000669, -0.00000546, -0.00000464, -0.00000424, -0.00000422, -0.00000449, -0.00000491, -0.00000533, -0.00000561, -0.00000564, -0.00000532, -0.00000464, -0.00000360, -0.00000228, -0.00000079, 0.00000073, 0.00000213, 0.00000326, 0.00000401, 0.00000429, 0.00000406, 0.00000334, 0.00000217, 0.00000067, -0.00000103, -0.00000279, -0.00000445, -0.00000588, -0.00000696, -0.00000764, -0.00000787, -0.00000767, -0.00000710, -0.00000624, -0.00000520, -0.00000409, -0.00000302, },
		{ -0.00000262, -0.00000340, -0.00000370, -0.00000356, -0.00000304, -0.00000225, -0.00000131, -0.00000034, 0.00000054, 0.00000123, 0.00000168, 0.00000186, 0.00000177, 0.00000145, 0.00000095, 0.00000035, -0.00000027, -0.00000086, -0.00000136, -0.00000174, -0.00000199, -0.00000212, -0.00000216, -0.00000215, -0.00000213, -0.00000214, -0.00000220, -0.00000233, -0.00000253, -0.00000275, -0.00000298, -0.00000317, -0.00000328, -0.00000327, -0.00000311, -0.00000279, -0.00000233, -0.00000173, -0.00000105, -0.00000031, 0.00000046, 0.00000123, 0.00000199, 0.00000276, 0.00000358, 0.00000450, 0.00000558, 0.00000689, 0.00000847, 0.00001034, 0.00001250, 0.00001487, 0.00001734, 0.00001973, 0.00002184, 0.00002342, 0.00002421, 0.00002399, 0.00002254, 0.00001973, 0.00001552, 0.00000995, 0.00000318, -0.00000452, -0.00001280, -0.00002122, -0.00002931, -0.00003658, -0.00004259, -0.00004695, -0.00004939, -0.00004974, -0.00004800, -0.00004429, -0.00003889, -0.00003219, -0.00002465, -0.00001680, -0.00000917, -0.00000225, 0.00000354, 0.00000789, 0.00001060, 0.00001164, 0.00001107, 0.00000910, 0.00000603, 0.00000223, -0.00000191, -0.00000601, -0.00000973, -0.00001279, -0.00001502, -0.00001631, -0.00001668, -0.00001623, -0.00001512, -0.00001355, -0.00001175, -0.00000993, -0.00000828, -0.00000691, -0.00000589, -0.00000523, -0.00000486, -0.00000469, -0.00000458, -0.00000439, -0.00000399, -0.00000328, -0.00000220, -0.00000076, 0.00000101, 0.00000298, 0.00000504, 0.00000700, 0.00000872, 0.00001003, 0.00001082, 0.00001102, 0.00001060, 0.00000961, 0.00000812, 0.00000628, 0.00000424, 0.00000217, 0.00000025, -0.00000138, },
		{ 0.00001094, 0.00001012, 0.00000867, 0.00000672, 0.00000442, 0.00000198, -0.00000041, -0.00000258, -0.00000438, -0.00000570, -0.00000650, -0.00000677, -0.00000657, -0.00000599, -0.00000516, -0.00000420, -0.00000325, -0.00000241, -0.00000178, -0.00000140, -0.00000129, -0.00000143, -0.00000178, -0.00000228, -0.00000285, -0.00000344, -0.00000400, -0.00000450, -0.00000494, -0.00000531, -0.00000565, -0.00000599, -0.00000636, -0.00000678, -0.00000726, -0.00000779, -0.00000834, -0.00000885, -0.00000927, -0.00000955, -0.00000965, -0.00000954, -0.00000924, -0.00000879, -0.00000828, -0.00000783, -0.00000757, -0.00000765, -0.00000817, -0.00000925, -0.00001092, -0.00001317, -0.00001589, -0.00001892, -0.00002203, -0.00002494, -0.00002735, -0.00002893, -0.00002942, -0.00002861, -0.00002637, -0.00002269, -0.00001766, -0.00001152, -0.00000458, 0.00000275, 0.00001000, 0.00001670, 0.00002240, 0.00002674, 0.00002943, 0.00003033, 0.00002944, 0.00002689, 0.00002297, 0.00001807, 0.00001263, 0.00000717, 0.00000215, -0.00000199, -0.00000491, -0.00000641, -0.00000639, -0.00000493, -0.00000222, 0.00000144, 0.00000567, 0.00001004, 0.00001414, 0.00001757, 0.00002006, 0.00002137, 0.00002145, 0.00002030, 0.00001807, 0.00001500, 0.00001136, 0.00000748, 0.00000368, 0.00000022, -0.00000266, -0.00000482, -0.00000622, -0.00000689, -0.00000692, -0.00000647, -0.00000574, -0.00000491, -0.00000415, -0.00000361, -0.00000335, -0.00000339, -0.00000369, -0.00000415, -0.00000463, -0.00000499, -0.00000508, -0.00000478, -0.00000401, -0.00000276, -0.00000106, 0.00000098, 0.00000322, 0.00000549, 0.00000758, 0.00000931, 0.00001051, 0.00001108, },
		{ -0.00000570, -0.00000340, -0.00000072, 0.00000215, 0.00000498, 0.00000753, 0.00000960, 0.00001102, 0.00001167, 0.00001153, 0.00001062, 0.00000905, 0.00000699, 0.00000462, 0.00000217, -0.00000015, -0.00000216, -0.00000371, -0.00000471, -0.00000514, -0.00000502, -0.00000443, -0.00000348, -0.00000230, -0.00000102, 0.00000023, 0.00000136, 0.00000232, 0.00000306, 0.00000362, 0.00000402, 0.00000432, 0.00000457, 0.00000483, 0.00000511, 0.00000544, 0.00000578, 0.00000610, 0.00000634, 0.00000646, 0.00000642, 0.00000620, 0.00000584, 0.00000540, 0.00000499, 0.00000476, 0.00000489, 0.00000554, 0.00000687, 0.00000898, 0.00001192, 0.00001566, 0.00002006, 0.00002491, 0.00002990, 0.00003466, 0.00003880, 0.00004191, 0.00004363, 0.00004367, 0.00004184, 0.00003809, 0.00003251, 0.00002534, 0.00001694, 0.00000777, -0.00000162, -0.00001068, -0.00001886, -0.00002570, -0.00003084, -0.00003405, -0.00003526, -0.00003456, -0.00003220, -0.00002853, -0.00002399, -0.00001907, -0.00001424, -0.00000994, -0.00000650, -0.00000414, -0.00000295, -0.00000287, -0.00000375, -0.00000529, -0.00000718, -0.00000905, -0.00001055, -0.00001140, -0.00001138, -0.00001041, -0.00000848, -0.00000573, -0.00000237, 0.00000131, 0.00000499, 0.00000835, 0.00001110, 0.00001300, 0.00001394, 0.00001385, 0.00001280, 0.00001094, 0.00000847, 0.00000566, 0.00000277, 0.00000006, -0.00000226, -0.00000404, -0.00000522, -0.00000579, -0.00000585, -0.00000552, -0.00000499, -0.00000442, -0.00000399, -0.00000383, -0.00000401, -0.00000455, -0.00000539, -0.00000642, -0.00000749, -0.00000840, -0.00000899, -0.00000909, -0.00000860, -0.00000747, },
		{ 0.00000180, 0.00000032, -0.00000130, -0.00000284, -0.00000406, -0.00000476, -0.00000476, -0.00000396, -0.00000234, 0.00000003, 0.00000300, 0.00000636, 0.00000985, 0.00001318, 0.00001610, 0.00001837, 0.00001982, 0.00002035, 0.00001996, 0.00001872, 0.00001677, 0.00001432, 0.00001158, 0.00000881, 0.00000621, 0.00000397, 0.00000220, 0.00000097, 0.00000025, 0.00000000, 0.00000010, 0.00000043, 0.00000085, 0.00000124, 0.00000151, 0.00000162, 0.00000156, 0.00000136, 0.00000108, 0.00000080, 0.00000057, 0.00000044, 0.00000041, 0.00000042, 0.00000037, 0.00000014, -0.00000047, -0.00000162, -0.00000347, -0.00000613, -0.00000965, -0.00001398, -0.00001896, -0.00002434, -0.00002976, -0.00003481, -0.00003904, -0.00004200, -0.00004328, -0.00004259, -0.00003973, -0.00003466, -0.00002754, -0.00001866, -0.00000846, 0.00000247, 0.00001350, 0.00002396, 0.00003320, 0.00004069, 0.00004600, 0.00004886, 0.00004920, 0.00004713, 0.00004294, 0.00003705, 0.00002998, 0.00002232, 0.00001464, 0.00000747, 0.00000124, -0.00000375, -0.00000733, -0.00000950, -0.00001038, -0.00001020, -0.00000928, -0.00000795, -0.00000655, -0.00000539, -0.00000466, -0.00000449, -0.00000489, -0.00000579, -0.00000701, -0.00000833, -0.00000952, -0.00001032, -0.00001055, -0.00001009, -0.00000888, -0.00000698, -0.00000452, -0.00000169, 0.00000125, 0.00000406, 0.00000647, 0.00000828, 0.00000935, 0.00000962, 0.00000913, 0.00000797, 0.00000632, 0.00000440, 0.00000246, 0.00000071, -0.00000065, -0.00000149, -0.00000175, -0.00000146, -0.00000072, 0.00000034, 0.00000151, 0.00000259, 0.00000339, 0.00000376, 0.00000362, 0.00000294, },
		{ -0.00000166, -0.00000018, 0.00000138, 0.00000276, 0.00000371, 0.00000405, 0.00000363, 0.00000242, 0.00000048, -0.00000204, -0.00000492, -0.00000786, -0.00001058, -0.00001277, -0.00001417, -0.00001461, -0.00001401, -0.00001237, -0.00000983, -0.00000658, -0.00000291, 0.00000087, 0.00000442, 0.00000745, 0.00000971, 0.00001106, 0.00001144, 0.00001087, 0.00000948, 0.00000748, 0.00000510, 0.00000261, 0.00000026, -0.00000174, -0.00000326, -0.00000422, -0.00000463, -0.00000457, -0.00000414, -0.00000351, -0.00000283, -0.00000220, -0.00000171, -0.00000136, -0.00000109, -0.00000077, -0.00000021, 0.00000077, 0.00000236, 0.00000472, 0.00000792, 0.00001194, 0.00001664, 0.00002177, 0.00002697, 0.00003180, 0.00003580, 0.00003848, 0.00003943, 0.00003830, 0.00003493, 0.00002928, 0.00002153, 0.00001202, 0.00000126, -0.00001012, -0.00002140, -0.00003187, -0.00004084, -0.00004773, -0.00005213, -0.00005378, -0.00005266, -0.00004896, -0.00004305, -0.00003544, -0.00002678, -0.00001774, -0.00000900, -0.00000115, 0.00000532, 0.00001007, 0.00001294, 0.00001395, 0.00001326, 0.00001118, 0.00000809, 0.00000444, 0.00000064, -0.00000291, -0.00000591, -0.00000816, -0.00000958, -0.00001018, -0.00001008, -0.00000947, -0.00000856, -0.00000758, -0.00000672, -0.00000612, -0.00000585, -0.00000590, -0.00000620, -0.00000661, -0.00000697, -0.00000711, -0.00000688, -0.00000618, -0.00000495, -0.00000323, -0.00000113, 0.00000121, 0.00000357, 0.00000574, 0.00000750, 0.00000867, 0.00000914, 0.00000887, 0.00000789, 0.00000632, 0.00000436, 0.00000222, 0.00000015, -0.00000161, -0.00000288, -0.00000352, -0.00000349, -0.00000283, },
		{ -0.00000064, -0.00000295, -0.00000507, -0.00000672, -0.00000766, -0.00000775, -0.00000692, -0.00000522, -0.00000280, 0.00000009, 0.00000315, 0.00000603, 0.00000840, 0.00000998, 0.00001054, 0.00000999, 0.00000833, 0.00000571, 0.00000237, -0.00000136, -0.00000512, -0.00000850, -0.00001115, -0.00001280, -0.00001325, -0.00001246, -0.00001050, -0.00000754, -0.00000390, 0.00000008, 0.00000400, 0.00000750, 0.00001025, 0.00001204, 0.00001272, 0.00001231, 0.00001092, 0.00000875, 0.00000607, 0.00000320, 0.00000041, -0.00000205, -0.00000402, -0.00000545, -0.00000638, -0.00000698, -0.00000747, -0.00000813, -0.00000925, -0.00001107, -0.00001375, -0.00001735, -0.00002178, -0.00002681, -0.00003208, -0.00003714, -0.00004147, -0.00004453, -0.00004584, -0.00004503, -0.00004185, -0.00003627, -0.00002843, -0.00001871, -0.00000762, 0.00000415, 0.00001585, 0.00002671, 0.00003600, 0.00004313, 0.00004764, 0.00004931, 0.00004813, 0.00004433, 0.00003833, 0.00003071, 0.00002217, 0.00001343, 0.00000520, -0.00000189, -0.00000736, -0.00001091, -0.00001243, -0.00001200, -0.00000986, -0.00000642, -0.00000214, 0.00000244, 0.00000682, 0.00001056, 0.00001333, 0.00001490, 0.00001520, 0.00001429, 0.00001234, 0.00000960, 0.00000636, 0.00000294, -0.00000038, -0.00000337, -0.00000587, -0.00000783, -0.00000923, -0.00001015, -0.00001070, -0.00001103, -0.00001127, -0.00001152, -0.00001183, -0.00001221, -0.00001261, -0.00001294, -0.00001309, -0.00001291, -0.00001231, -0.00001122, -0.00000962, -0.00000757, -0.00000519, -0.00000264, -0.00000014, 0.00000208, 0.00000381, 0.00000488, 0.00000517, 0.00000466, 0.00000341, 0.00000157, },
		{ 0.00000127, 0.00000356, 0.00000522, 0.00000604, 0.00000588, 0.00000470, 0.00000257, -0.00000032, -0.00000371, -0.00000726, -0.00001059, -0.00001333, -0.00001516, -0.00001584, -0.00001525, -0.00001340, -0.00001044, -0.00000665, -0.00000240, 0.00000188, 0.00000575, 0.00000878, 0.00001064, 0.00001111, 0.00001013, 0.00000778, 0.00000429, 0.00000002, -0.00000459, -0.00000904, -0.00001286, -0.00001563, -0.00001705, -0.00001695, -0.00001532, -0.00001233, -0.00000826, -0.00000351, 0.00000146, 0.00000618, 0.00001024, 0.00001333, 0.00001525, 0.00001597, 0.00001565, 0.00001453, 0.00001303, 0.00001157, 0.00001062, 0.00001057, 0.00001172, 0.00001419, 0.00001796, 0.00002279, 0.00002828, 0.00003389, 0.00003899, 0.00004293, 0.00004511, 0.00004503, 0.00004238, 0.00003703, 0.00002913, 0.00001904, 0.00000731, -0.00000531, -0.00001799, -0.00002987, -0.00004015, -0.00004815, -0.00005337, -0.00005555, -0.00005467, -0.00005097, -0.00004490, -0.00003708, -0.00002825, -0.00001921, -0.00001069, -0.00000334, 0.00000234, 0.00000604, 0.00000769, 0.00000739, 0.00000544, 0.00000228, -0.00000157, -0.00000556, -0.00000916, -0.00001194, -0.00001356, -0.00001387, -0.00001284, -0.00001060, -0.00000739, -0.00000354, 0.00000056, 0.00000453, 0.00000805, 0.00001085, 0.00001275, 0.00001367, 0.00001366, 0.00001282, 0.00001133, 0.00000939, 0.00000722, 0.00000501, 0.00000289, 0.00000098, -0.00000071, -0.00000218, -0.00000348, -0.00000469, -0.00000586, -0.00000705, -0.00000826, -0.00000946, -0.00001056, -0.00001147, -0.00001204, -0.00001215, -0.00001169, -0.00001062, -0.00000894, -0.00000673, -0.00000414, -0.00000138, },
		{ -0.00002021, -0.00002040, -0.00001982, -0.00001836, -0.00001603, -0.00001292, -0.00000920, -0.00000514, -0.00000108, 0.00000262, 0.00000562, 0.00000760, 0.00000835, 0.00000774, 0.00000583, 0.00000277, -0.00000114, -0.00000549, -0.00000980, -0.00001360, -0.00001642, -0.00001789, -0.00001779, -0.00001602, -0.00001268, -0.00000804, -0.00000252, 0.00000336, 0.00000902, 0.00001390, 0.00001747, 0.00001938, 0.00001938, 0.00001748, 0.00001384, 0.00000881, 0.00000291, -0.00000328, -0.00000913, -0.00001406, -0.00001761, -0.00001947, -0.00001952, -0.00001788, -0.00001485, -0.00001092, -0.00000670, -0.00000285, -0.00000001, 0.00000128, 0.00000064, -0.00000209, -0.00000685, -0.00001330, -0.00002089, -0.00002891, -0.00003653, -0.00004290, -0.00004724, -0.00004889, -0.00004745, -0.00004273, -0.00003488, -0.00002429, -0.00001165, 0.00000220, 0.00001628, 0.00002958, 0.00004116, 0.00005020, 0.00005614, 0.00005865, 0.00005771, 0.00005356, 0.00004671, 0.00003787, 0.00002786, 0.00001756, 0.00000780, -0.00000071, -0.00000740, -0.00001196, -0.00001428, -0.00001448, -0.00001290, -0.00001000, -0.00000634, -0.00000252, 0.00000092, 0.00000353, 0.00000498, 0.00000511, 0.00000394, 0.00000163, -0.00000155, -0.00000523, -0.00000899, -0.00001246, -0.00001528, -0.00001722, -0.00001813, -0.00001799, -0.00001688, -0.00001500, -0.00001259, -0.00000992, -0.00000728, -0.00000489, -0.00000295, -0.00000157, -0.00000079, -0.00000058, -0.00000087, -0.00000153, -0.00000245, -0.00000350, -0.00000460, -0.00000570, -0.00000678, -0.00000787, -0.00000901, -0.00001025, -0.00001164, -0.00001318, -0.00001483, -0.00001650, -0.00001807, -0.00001937, },
		{ 0.00000252, 0.00000131, -0.00000036, -0.00000250, -0.00000507, -0.00000794, -0.00001093, -0.00001378, -0.00001622, -0.00001797, -0.00001878, -0.00001847, -0.00001696, -0.00001430, -0.00001064, -0.00000627, -0.00000160, 0.00000295, 0.00000687, 0.00000974, 0.00001119, 0.00001098, 0.00000904, 0.00000548, 0.00000057, -0.00000524, -0.00001138, -0.00001724, -0.00002218, -0.00002566, -0.00002722, -0.00002664, -0.00002387, -0.00001911, -0.00001276, -0.00000540, 0.00000225, 0.00000946, 0.00001551, 0.00001979, 0.00002187, 0.00002157, 0.00001896, 0.00001435, 0.00000832, 0.00000160, -0.00000495, -0.00001049, -0.00001421, -0.00001552, -0.00001403, -0.00000966, -0.00000263, 0.00000652, 0.00001701, 0.00002786, 0.00003799, 0.00004633, 0.00005191, 0.00005397, 0.00005205, 0.00004601, 0.00003606, 0.00002277, 0.00000699, -0.00001019, -0.00002757, -0.00004393, -0.00005816, -0.00006931, -0.00007672, -0.00008004, -0.00007927, -0.00007475, -0.00006709, -0.00005715, -0.00004589, -0.00003433, -0.00002345, -0.00001406, -0.00000678, -0.00000196, 0.00000028, 0.00000012, -0.00000207, -0.00000574, -0.00001025, -0.00001493, -0.00001916, -0.00002245, -0.00002443, -0.00002494, -0.00002400, -0.00002177, -0.00001857, -0.00001481, -0.00001090, -0.00000728, -0.00000428, -0.00000216, -0.00000105, -0.00000093, -0.00000171, -0.00000316, -0.00000501, -0.00000697, -0.00000874, -0.00001008, -0.00001083, -0.00001090, -0.00001030, -0.00000909, -0.00000743, -0.00000548, -0.00000345, -0.00000149, 0.00000024, 0.00000167, 0.00000275, 0.00000350, 0.00000395, 0.00000420, 0.00000429, 0.00000430, 0.00000424, 0.00000409, 0.00000382, 0.00000332, },
		{ -0.00000321, -0.00000293, -0.00000250, -0.00000187, -0.00000108, -0.00000019, 0.00000068, 0.00000135, 0.00000162, 0.00000131, 0.00000027, -0.00000160, -0.00000429, -0.00000770, -0.00001163, -0.00001576, -0.00001973, -0.00002314, -0.00002559, -0.00002676, -0.00002643, -0.00002451, -0.00002108, -0.00001640, -0.00001087, -0.00000503, 0.00000053, 0.00000517, 0.00000834, 0.00000960, 0.00000868, 0.00000553, 0.00000032, -0.00000652, -0.00001440, -0.00002257, -0.00003022, -0.00003656, -0.00004089, -0.00004273, -0.00004179, -0.00003812, -0.00003205, -0.00002417, -0.00001533, -0.00000649, 0.00000131, 0.00000710, 0.00001005, 0.00000961, 0.00000553, -0.00000207, -0.00001270, -0.00002554, -0.00003949, -0.00005329, -0.00006561, -0.00007520, -0.00008099, -0.00008219, -0.00007840, -0.00006963, -0.00005629, -0.00003922, -0.00001955, 0.00000136, 0.00002205, 0.00004110, 0.00005720, 0.00006936, 0.00007690, 0.00007952, 0.00007736, 0.00007093, 0.00006103, 0.00004874, 0.00003523, 0.00002172, 0.00000929, -0.00000113, -0.00000888, -0.00001362, -0.00001531, -0.00001419, -0.00001079, -0.00000577, 0.00000010, 0.00000602, 0.00001130, 0.00001535, 0.00001780, 0.00001849, 0.00001743, 0.00001488, 0.00001120, 0.00000687, 0.00000237, -0.00000181, -0.00000528, -0.00000776, -0.00000912, -0.00000936, -0.00000861, -0.00000711, -0.00000516, -0.00000309, -0.00000119, 0.00000028, 0.00000115, 0.00000135, 0.00000089, -0.00000011, -0.00000150, -0.00000307, -0.00000462, -0.00000597, -0.00000697, -0.00000756, -0.00000772, -0.00000750, -0.00000698, -0.00000629, -0.00000554, -0.00000485, -0.00000428, -0.00000387, -0.00000359, -0.00000340, },
		{ -0.00000824, -0.00000777, -0.00000746, -0.00000739, -0.00000757, -0.00000797, -0.00000850, -0.00000902, -0.00000935, -0.00000934, -0.00000883, -0.00000775, -0.00000608, -0.00000390, -0.00000140, 0.00000117, 0.00000352, 0.00000531, 0.00000624, 0.00000607, 0.00000468, 0.00000207, -0.00000160, -0.00000604, -0.00001081, -0.00001540, -0.00001925, -0.00002184, -0.00002271, -0.00002159, -0.00001836, -0.00001314, -0.00000628, 0.00000167, 0.00001003, 0.00001799, 0.00002476, 0.00002962, 0.00003201, 0.00003161, 0.00002838, 0.00002260, 0.00001483, 0.00000591, -0.00000317, -0.00001129, -0.00001738, -0.00002050, -0.00001993, -0.00001531, -0.00000664, 0.00000565, 0.00002078, 0.00003760, 0.00005475, 0.00007073, 0.00008408, 0.00009347, 0.00009787, 0.00009662, 0.00008951, 0.00007682, 0.00005929, 0.00003806, 0.00001459, -0.00000949, -0.00003251, -0.00005287, -0.00006925, -0.00008066, -0.00008653, -0.00008675, -0.00008170, -0.00007212, -0.00005912, -0.00004400, -0.00002815, -0.00001295, 0.00000042, 0.00001100, 0.00001819, 0.00002173, 0.00002173, 0.00001863, 0.00001311, 0.00000603, -0.00000168, -0.00000910, -0.00001545, -0.00002013, -0.00002277, -0.00002325, -0.00002169, -0.00001844, -0.00001399, -0.00000892, -0.00000383, 0.00000072, 0.00000431, 0.00000662, 0.00000753, 0.00000708, 0.00000544, 0.00000291, -0.00000013, -0.00000332, -0.00000628, -0.00000874, -0.00001050, -0.00001147, -0.00001167, -0.00001124, -0.00001034, -0.00000919, -0.00000803, -0.00000705, -0.00000638, -0.00000611, -0.00000622, -0.00000667, -0.00000733, -0.00000807, -0.00000875, -0.00000926, -0.00000951, -0.00000948, -0.00000921, -0.00000875, },
		{ -0.00000267, -0.00000262, -0.00000232, -0.00000169, -0.00000070, 0.00000061, 0.00000214, 0.00000373, 0.00000519, 0.00000634, 0.00000699, 0.00000705, 0.00000646, 0.00000530, 0.00000370, 0.00000190, 0.00000017, -0.00000117, -0.00000186, -0.00000167, -0.00000048, 0.00000169, 0.00000469, 0.00000825, 0.00001195, 0.00001534, 0.00001792, 0.00001925, 0.00001897, 0.00001689, 0.00001300, 0.00000749, 0.00000078, -0.00000658, -0.00001386, -0.00002033, -0.00002528, -0.00002809, -0.00002836, -0.00002594, -0.00002098, -0.00001391, -0.00000546, 0.00000345, 0.00001174, 0.00001831, 0.00002216, 0.00002246, 0.00001872, 0.00001077, -0.00000112, -0.00001626, -0.00003362, -0.00005185, -0.00006944, -0.00008483, -0.00009654, -0.00010337, -0.00010442, -0.00009930, -0.00008808, -0.00007134, -0.00005014, -0.00002591, -0.00000032, 0.00002481, 0.00004773, 0.00006687, 0.00008098, 0.00008927, 0.00009140, 0.00008755, 0.00007838, 0.00006496, 0.00004863, 0.00003091, 0.00001333, -0.00000268, -0.00001594, -0.00002561, -0.00003124, -0.00003276, -0.00003053, -0.00002521, -0.00001770, -0.00000906, -0.00000034, 0.00000747, 0.00001357, 0.00001741, 0.00001871, 0.00001749, 0.00001405, 0.00000888, 0.00000264, -0.00000397, -0.00001023, -0.00001555, -0.00001945, -0.00002167, -0.00002214, -0.00002097, -0.00001845, -0.00001499, -0.00001105, -0.00000709, -0.00000355, -0.00000073, 0.00000115, 0.00000202, 0.00000194, 0.00000107, -0.00000034, -0.00000202, -0.00000370, -0.00000515, -0.00000618, -0.00000673, -0.00000677, -0.00000638, -0.00000568, -0.00000483, -0.00000399, -0.00000328, -0.00000278, -0.00000254, -0.00000250, -0.00000258, },
		{ -0.00000600, -0.00000623, -0.00000668, -0.00000745, -0.00000857, -0.00001000, -0.00001162, -0.00001327, -0.00001474, -0.00001582, -0.00001633, -0.00001616, -0.00001525, -0.00001369, -0.00001161, -0.00000928, -0.00000699, -0.00000506, -0.00000379, -0.00000343, -0.00000410, -0.00000581, -0.00000842, -0.00001165, -0.00001512, -0.00001836, -0.00002091, -0.00002232, -0.00002227, -0.00002059, -0.00001726, -0.00001251, -0.00000674, -0.00000051, 0.00000551, 0.00001062, 0.00001417, 0.00001564, 0.00001473, 0.00001136, 0.00000579, -0.00000148, -0.00000967, -0.00001783, -0.00002492, -0.00002988, -0.00003181, -0.00003004, -0.00002420, -0.00001433, -0.00000087, 0.00001535, 0.00003316, 0.00005113, 0.00006772, 0.00008142, 0.00009084, 0.00009491, 0.00009297, 0.00008481, 0.00007076, 0.00005162, 0.00002864, 0.00000339, -0.00002236, -0.00004680, -0.00006824, -0.00008524, -0.00009676, -0.00010222, -0.00010157, -0.00009524, -0.00008413, -0.00006948, -0.00005277, -0.00003557, -0.00001939, -0.00000555, 0.00000493, 0.00001144, 0.00001377, 0.00001214, 0.00000713, -0.00000037, -0.00000931, -0.00001853, -0.00002695, -0.00003364, -0.00003792, -0.00003939, -0.00003798, -0.00003394, -0.00002776, -0.00002014, -0.00001190, -0.00000384, 0.00000328, 0.00000887, 0.00001251, 0.00001405, 0.00001354, 0.00001126, 0.00000763, 0.00000320, -0.00000146, -0.00000581, -0.00000939, -0.00001188, -0.00001311, -0.00001309, -0.00001195, -0.00000995, -0.00000744, -0.00000476, -0.00000226, -0.00000021, 0.00000119, 0.00000186, 0.00000181, 0.00000116, 0.00000008, -0.00000123, -0.00000257, -0.00000376, -0.00000469, -0.00000533, -0.00000569, -0.00000587, },
		{ -0.00000658, -0.00000694, -0.00000710, -0.00000692, -0.00000635, -0.00000542, -0.00000423, -0.00000297, -0.00000185, -0.00000110, -0.00000090, -0.00000140, -0.00000262, -0.00000450, -0.00000687, -0.00000946, -0.00001194, -0.00001397, -0.00001523, -0.00001546, -0.00001454, -0.00001246, -0.00000938, -0.00000560, -0.00000154, 0.00000232, 0.00000547, 0.00000746, 0.00000793, 0.00000669, 0.00000376, -0.00000065, -0.00000614, -0.00001212, -0.00001792, -0.00002284, -0.00002624, -0.00002762, -0.00002667, -0.00002337, -0.00001799, -0.00001106, -0.00000336, 0.00000414, 0.00001042, 0.00001447, 0.00001545, 0.00001273, 0.00000606, -0.00000445, -0.00001828, -0.00003452, -0.00005197, -0.00006917, -0.00008461, -0.00009682, -0.00010450, -0.00010670, -0.00010288, -0.00009298, -0.00007745, -0.00005720, -0.00003356, -0.00000814, 0.00001731, 0.00004102, 0.00006141, 0.00007719, 0.00008745, 0.00009180, 0.00009032, 0.00008357, 0.00007255, 0.00005852, 0.00004294, 0.00002732, 0.00001301, 0.00000117, -0.00000739, -0.00001226, -0.00001343, -0.00001128, -0.00000651, -0.00000007, 0.00000700, 0.00001363, 0.00001888, 0.00002201, 0.00002256, 0.00002039, 0.00001568, 0.00000886, 0.00000063, -0.00000822, -0.00001684, -0.00002441, -0.00003029, -0.00003401, -0.00003536, -0.00003437, -0.00003132, -0.00002667, -0.00002103, -0.00001506, -0.00000939, -0.00000460, -0.00000110, 0.00000086, 0.00000124, 0.00000017, -0.00000208, -0.00000510, -0.00000846, -0.00001172, -0.00001448, -0.00001646, -0.00001750, -0.00001756, -0.00001675, -0.00001525, -0.00001333, -0.00001128, -0.00000935, -0.00000777, -0.00000666, -0.00000608, -0.00000596, -0.00000619, },
		{ -0.00000956, -0.00000844, -0.00000769, -0.00000749, -0.00000791, -0.00000891, -0.00001036, -0.00001204, -0.00001367, -0.00001497, -0.00001569, -0.00001568, -0.00001486, -0.00001330, -0.00001116, -0.00000875, -0.00000640, -0.00000450, -0.00000339, -0.00000334, -0.00000449, -0.00000684, -0.00001020, -0.00001424, -0.00001851, -0.00002248, -0.00002562, -0.00002743, -0.00002754, -0.00002577, -0.00002214, -0.00001689, -0.00001048, -0.00000352, 0.00000326, 0.00000912, 0.00001337, 0.00001548, 0.00001516, 0.00001239, 0.00000744, 0.00000088, -0.00000645, -0.00001357, -0.00001942, -0.00002296, -0.00002333, -0.00001994, -0.00001251, -0.00000122, 0.00001340, 0.00003040, 0.00004851, 0.00006628, 0.00008216, 0.00009469, 0.00010258, 0.00010491, 0.00010118, 0.00009137, 0.00007596, 0.00005592, 0.00003258, 0.00000755, -0.00001742, -0.00004062, -0.00006052, -0.00007587, -0.00008586, -0.00009014, -0.00008883, -0.00008253, -0.00007222, -0.00005913, -0.00004466, -0.00003020, -0.00001703, -0.00000618, 0.00000166, 0.00000615, 0.00000735, 0.00000565, 0.00000174, -0.00000353, -0.00000921, -0.00001440, -0.00001833, -0.00002044, -0.00002045, -0.00001836, -0.00001444, -0.00000920, -0.00000330, 0.00000253, 0.00000758, 0.00001122, 0.00001302, 0.00001276, 0.00001045, 0.00000634, 0.00000089, -0.00000533, -0.00001164, -0.00001741, -0.00002207, -0.00002518, -0.00002651, -0.00002601, -0.00002385, -0.00002035, -0.00001599, -0.00001127, -0.00000675, -0.00000287, -0.00000001, 0.00000161, 0.00000195, 0.00000110, -0.00000071, -0.00000314, -0.00000583, -0.00000839, -0.00001052, -0.00001199, -0.00001268, -0.00001261, -0.00001192, -0.00001081, },
		{ -0.00002372, -0.00002483, -0.00002521, -0.00002466, -0.00002316, -0.00002079, -0.00001779, -0.00001449, -0.00001127, -0.00000852, -0.00000655, -0.00000560, -0.00000574, -0.00000691, -0.00000888, -0.00001132, -0.00001378, -0.00001583, -0.00001704, -0.00001709, -0.00001581, -0.00001319, -0.00000942, -0.00000487, -0.00000004, 0.00000448, 0.00000811, 0.00001031, 0.00001068, 0.00000903, 0.00000538, 0.00000003, -0.00000653, -0.00001362, -0.00002044, -0.00002618, -0.00003012, -0.00003169, -0.00003060, -0.00002683, -0.00002071, -0.00001287, -0.00000422, 0.00000417, 0.00001114, 0.00001561, 0.00001666, 0.00001365, 0.00000631, -0.00000517, -0.00002020, -0.00003777, -0.00005654, -0.00007497, -0.00009143, -0.00010437, -0.00011248, -0.00011476, -0.00011070, -0.00010028, -0.00008398, -0.00006280, -0.00003811, -0.00001155, 0.00001509, 0.00004006, 0.00006179, 0.00007901, 0.00009086, 0.00009698, 0.00009744, 0.00009282, 0.00008403, 0.00007228, 0.00005890, 0.00004525, 0.00003254, 0.00002178, 0.00001364, 0.00000845, 0.00000618, 0.00000649, 0.00000876, 0.00001223, 0.00001605, 0.00001942, 0.00002167, 0.00002232, 0.00002115, 0.00001817, 0.00001363, 0.00000798, 0.00000176, -0.00000440, -0.00000995, -0.00001443, -0.00001754, -0.00001917, -0.00001942, -0.00001857, -0.00001704, -0.00001530, -0.00001388, -0.00001323, -0.00001369, -0.00001544, -0.00001849, -0.00002265, -0.00002761, -0.00003291, -0.00003804, -0.00004250, -0.00004584, -0.00004775, -0.00004806, -0.00004678, -0.00004408, -0.00004029, -0.00003583, -0.00003120, -0.00002685, -0.00002320, -0.00002055, -0.00001903, -0.00001865, -0.00001925, -0.00002054, -0.00002216, },
		{ -0.00001232, -0.00000922, -0.00000760, -0.00000756, -0.00000900, -0.00001164, -0.00001505, -0.00001871, -0.00002207, -0.00002462, -0.00002596, -0.00002583, -0.00002418, -0.00002117, -0.00001713, -0.00001254, -0.00000800, -0.00000408, -0.00000132, -0.00000012, -0.00000071, -0.00000306, -0.00000695, -0.00001192, -0.00001737, -0.00002260, -0.00002689, -0.00002963, -0.00003034, -0.00002880, -0.00002503, -0.00001935, -0.00001233, -0.00000475, 0.00000252, 0.00000857, 0.00001261, 0.00001401, 0.00001247, 0.00000799, 0.00000095, -0.00000793, -0.00001764, -0.00002700, -0.00003472, -0.00003964, -0.00004073, -0.00003734, -0.00002919, -0.00001650, 0.00000007, 0.00001940, 0.00004002, 0.00006025, 0.00007832, 0.00009256, 0.00010153, 0.00010418, 0.00009994, 0.00008877, 0.00007122, 0.00004832, 0.00002152, -0.00000740, -0.00003656, -0.00006407, -0.00008825, -0.00010773, -0.00012155, -0.00012926, -0.00013089, -0.00012696, -0.00011838, -0.00010635, -0.00009221, -0.00007733, -0.00006299, -0.00005023, -0.00003980, -0.00003209, -0.00002717, -0.00002477, -0.00002438, -0.00002530, -0.00002679, -0.00002812, -0.00002865, -0.00002795, -0.00002579, -0.00002217, -0.00001732, -0.00001160, -0.00000551, 0.00000041, 0.00000567, 0.00000986, 0.00001272, 0.00001413, 0.00001416, 0.00001302, 0.00001103, 0.00000855, 0.00000598, 0.00000362, 0.00000173, 0.00000039, -0.00000042, -0.00000085, -0.00000116, -0.00000165, -0.00000262, -0.00000435, -0.00000699, -0.00001060, -0.00001506, -0.00002013, -0.00002545, -0.00003058, -0.00003506, -0.00003844, -0.00004039, -0.00004069, -0.00003930, -0.00003634, -0.00003211, -0.00002706, -0.00002171, -0.00001663, },
		{ -0.00006514, -0.00006483, -0.00006202, -0.00005699, -0.00005023, -0.00004241, -0.00003430, -0.00002670, -0.00002036, -0.00001590, -0.00001370, -0.00001393, -0.00001647, -0.00002093, -0.00002672, -0.00003307, -0.00003916, -0.00004419, -0.00004748, -0.00004855, -0.00004719, -0.00004350, -0.00003783, -0.00003083, -0.00002331, -0.00001616, -0.00001029, -0.00000647, -0.00000524, -0.00000688, -0.00001130, -0.00001811, -0.00002659, -0.00003581, -0.00004470, -0.00005220, -0.00005736, -0.00005947, -0.00005816, -0.00005347, -0.00004583, -0.00003607, -0.00002535, -0.00001501, -0.00000649, -0.00000112, -0.00000001, -0.00000390, -0.00001306, -0.00002723, -0.00004564, -0.00006702, -0.00008973, -0.00011189, -0.00013155, -0.00014683, -0.00015617, -0.00015841, -0.00015293, -0.00013970, -0.00011932, -0.00009293, -0.00006214, -0.00002888, 0.00000476, 0.00003672, 0.00006513, 0.00008847, 0.00010566, 0.00011617, 0.00011998, 0.00011760, 0.00010997, 0.00009833, 0.00008413, 0.00006884, 0.00005384, 0.00004029, 0.00002906, 0.00002067, 0.00001527, 0.00001269, 0.00001248, 0.00001399, 0.00001652, 0.00001932, 0.00002176, 0.00002337, 0.00002386, 0.00002316, 0.00002142, 0.00001892, 0.00001607, 0.00001332, 0.00001110, 0.00000977, 0.00000954, 0.00001051, 0.00001261, 0.00001564, 0.00001930, 0.00002323, 0.00002707, 0.00003048, 0.00003321, 0.00003509, 0.00003608, 0.00003621, 0.00003561, 0.00003444, 0.00003286, 0.00003102, 0.00002898, 0.00002675, 0.00002422, 0.00002125, 0.00001762, 0.00001313, 0.00000760, 0.00000095, -0.00000681, -0.00001550, -0.00002481, -0.00003430, -0.00004344, -0.00005162, -0.00005827, -0.00006290, },
		{ -0.00000387, -0.00001801, -0.00003316, -0.00004859, -0.00006341, -0.00007672, -0.00008762, -0.00009533, -0.00009928, -0.00009919, -0.00009506, -0.00008725, -0.00007645, -0.00006357, -0.00004975, -0.00003619, -0.00002406, -0.00001438, -0.00000791, -0.00000508, -0.00000591, -0.00001007, -0.00001681, -0.00002515, -0.00003389, -0.00004179, -0.00004769, -0.00005063, -0.00005000, -0.00004557, -0.00003752, -0.00002649, -0.00001346, 0.00000035, 0.00001356, 0.00002487, 0.00003310, 0.00003745, 0.00003749, 0.00003331, 0.00002551, 0.00001510, 0.00000351, -0.00000764, -0.00001667, -0.00002199, -0.00002233, -0.00001686, -0.00000528, 0.00001208, 0.00003429, 0.00005988, 0.00008695, 0.00011334, 0.00013680, 0.00015523, 0.00016685, 0.00017036, 0.00016508, 0.00015102, 0.00012884, 0.00009985, 0.00006587, 0.00002909, -0.00000816, -0.00004357, -0.00007501, -0.00010079, -0.00011968, -0.00013105, -0.00013486, -0.00013166, -0.00012245, -0.00010860, -0.00009170, -0.00007339, -0.00005521, -0.00003849, -0.00002426, -0.00001313, -0.00000538, -0.00000088, 0.00000078, 0.00000024, -0.00000174, -0.00000439, -0.00000699, -0.00000898, -0.00000999, -0.00000984, -0.00000858, -0.00000642, -0.00000369, -0.00000078, 0.00000191, 0.00000407, 0.00000545, 0.00000596, 0.00000563, 0.00000463, 0.00000319, 0.00000164, 0.00000030, -0.00000055, -0.00000068, 0.00000005, 0.00000167, 0.00000413, 0.00000727, 0.00001092, 0.00001486, 0.00001886, 0.00002277, 0.00002645, 0.00002982, 0.00003284, 0.00003551, 0.00003782, 0.00003974, 0.00004117, 0.00004199, 0.00004196, 0.00004082, 0.00003825, 0.00003396, 0.00002767, 0.00001924, 0.00000867, },
		{ 0.00008161, 0.00008708, 0.00009082, 0.00009221, 0.00009065, 0.00008564, 0.00007684, 0.00006414, 0.00004772, 0.00002807, 0.00000598, -0.00001746, -0.00004097, -0.00006314, -0.00008255, -0.00009789, -0.00010811, -0.00011251, -0.00011082, -0.00010326, -0.00009054, -0.00007380, -0.00005454, -0.00003449, -0.00001543, 0.00000092, 0.00001315, 0.00002020, 0.00002155, 0.00001724, 0.00000787, -0.00000545, -0.00002121, -0.00003763, -0.00005289, -0.00006526, -0.00007335, -0.00007622, -0.00007350, -0.00006546, -0.00005297, -0.00003747, -0.00002079, -0.00000501, 0.00000779, 0.00001569, 0.00001721, 0.00001140, -0.00000199, -0.00002247, -0.00004884, -0.00007927, -0.00011143, -0.00014272, -0.00017048, -0.00019225, -0.00020595, -0.00021010, -0.00020395, -0.00018753, -0.00016170, -0.00012798, -0.00008853, -0.00004587, -0.00000271, 0.00003827, 0.00007465, 0.00010447, 0.00012634, 0.00013953, 0.00014400, 0.00014034, 0.00012971, 0.00011367, 0.00009402, 0.00007262, 0.00005127, 0.00003150, 0.00001451, 0.00000107, -0.00000846, -0.00001417, -0.00001645, -0.00001598, -0.00001360, -0.00001016, -0.00000649, -0.00000326, -0.00000096, 0.00000016, 0.00000006, -0.00000112, -0.00000308, -0.00000543, -0.00000778, -0.00000975, -0.00001109, -0.00001162, -0.00001130, -0.00001023, -0.00000859, -0.00000661, -0.00000457, -0.00000271, -0.00000123, -0.00000023, 0.00000026, 0.00000030, 0.00000003, -0.00000036, -0.00000066, -0.00000067, -0.00000019, 0.00000090, 0.00000268, 0.00000517, 0.00000835, 0.00001217, 0.00001657, 0.00002149, 0.00002690, 0.00003276, 0.00003907, 0.00004580, 0.00005292, 0.00006030, 0.00006776, 0.00007500, },
		{ 0.00001641, 0.00002006, 0.00002493, 0.00003114, 0.00003877, 0.00004776, 0.00005788, 0.00006872, 0.00007966, 0.00008990, 0.00009847, 0.00010431, 0.00010636, 0.00010365, 0.00009541, 0.00008117, 0.00006087, 0.00003488, 0.00000408, -0.00003022, -0.00006628, -0.00010210, -0.00013554, -0.00016446, -0.00018698, -0.00020156, -0.00020723, -0.00020365, -0.00019122, -0.00017101, -0.00014472, -0.00011458, -0.00008310, -0.00005292, -0.00002651, -0.00000599, 0.00000709, 0.00001191, 0.00000848, -0.00000240, -0.00001914, -0.00003953, -0.00006095, -0.00008060, -0.00009577, -0.00010412, -0.00010388, -0.00009408, -0.00007464, -0.00004640, -0.00001107, 0.00002890, 0.00007051, 0.00011052, 0.00014566, 0.00017296, 0.00019000, 0.00019512, 0.00018755, 0.00016749, 0.00013608, 0.00009532, 0.00004784, -0.00000327, -0.00005473, -0.00010336, -0.00014631, -0.00018126, -0.00020662, -0.00022157, -0.00022608, -0.00022090, -0.00020740, -0.00018741, -0.00016306, -0.00013656, -0.00011000, -0.00008520, -0.00006358, -0.00004606, -0.00003310, -0.00002466, -0.00002028, -0.00001922, -0.00002054, -0.00002322, -0.00002630, -0.00002897, -0.00003063, -0.00003090, -0.00002967, -0.00002706, -0.00002335, -0.00001894, -0.00001426, -0.00000975, -0.00000574, -0.00000247, -0.00000006, 0.00000153, 0.00000241, 0.00000280, 0.00000293, 0.00000304, 0.00000333, 0.00000395, 0.00000495, 0.00000632, 0.00000797, 0.00000975, 0.00001151, 0.00001307, 0.00001431, 0.00001512, 0.00001548, 0.00001540, 0.00001493, 0.00001419, 0.00001329, 0.00001234, 0.00001147, 0.00001075, 0.00001027, 0.00001008, 0.00001024, 0.00001084, 0.00001197, 0.00001377, },
		{ 0.00001206, 0.00001321, 0.00001418, 0.00001496, 0.00001560, 0.00001622, 0.00001700, 0.00001825, 0.00002035, 0.00002373, 0.00002887, 0.00003621, 0.00004608, 0.00005865, 0.00007385, 0.00009129, 0.00011028, 0.00012977, 0.00014841, 0.00016463, 0.00017672, 0.00018299, 0.00018190, 0.00017225, 0.00015332, 0.00012496, 0.00008775, 0.00004296, -0.00000747, -0.00006099, -0.00011466, -0.00016530, -0.00020978, -0.00024527, -0.00026948, -0.00028086, -0.00027884, -0.00026382, -0.00023725, -0.00020153, -0.00015981, -0.00011582, -0.00007348, -0.00003665, -0.00000876, 0.00000747, 0.00001029, -0.00000093, -0.00002561, -0.00006200, -0.00010730, -0.00015790, -0.00020968, -0.00025829, -0.00029960, -0.00032995, -0.00034653, -0.00034753, -0.00033233, -0.00030152, -0.00025684, -0.00020102, -0.00013752, -0.00007030, -0.00000343, 0.00005920, 0.00011415, 0.00015873, 0.00019111, 0.00021043, 0.00021682, 0.00021131, 0.00019564, 0.00017213, 0.00014339, 0.00011211, 0.00008084, 0.00005175, 0.00002656, 0.00000638, -0.00000825, -0.00001738, -0.00002152, -0.00002157, -0.00001865, -0.00001396, -0.00000868, -0.00000379, -0.00000005, 0.00000207, 0.00000237, 0.00000095, -0.00000190, -0.00000574, -0.00001006, -0.00001435, -0.00001817, -0.00002121, -0.00002327, -0.00002432, -0.00002445, -0.00002383, -0.00002271, -0.00002134, -0.00001996, -0.00001875, -0.00001783, -0.00001722, -0.00001688, -0.00001671, -0.00001659, -0.00001636, -0.00001591, -0.00001513, -0.00001399, -0.00001248, -0.00001065, -0.00000858, -0.00000635, -0.00000407, -0.00000182, 0.00000034, 0.00000239, 0.00000430, 0.00000608, 0.00000775, 0.00000931, 0.00001075, },
		{ 0.00000535, 0.00000569, 0.00000621, 0.00000693, 0.00000788, 0.00000904, 0.00001033, 0.00001166, 0.00001287, 0.00001377, 0.00001417, 0.00001392, 0.00001296, 0.00001133, 0.00000923, 0.00000705, 0.00000537, 0.00000491, 0.00000652, 0.00001107, 0.00001939, 0.00003212, 0.00004962, 0.00007184, 0.00009829, 0.00012791, 0.00015915, 0.00018997, 0.00021797, 0.00024053, 0.00025504, 0.00025911, 0.00025081, 0.00022888, 0.00019294, 0.00014361, 0.00008258, 0.00001254, -0.00006289, -0.00013942, -0.00021231, -0.00027676, -0.00032825, -0.00036292, -0.00037786, -0.00037142, -0.00034340, -0.00029505, -0.00022911, -0.00014961, -0.00006160, 0.00002919, 0.00011678, 0.00019535, 0.00025969, 0.00030558, 0.00033009, 0.00033176, 0.00031076, 0.00026874, 0.00020878, 0.00013506, 0.00005255, -0.00003338, -0.00011741, -0.00019459, -0.00026073, -0.00031266, -0.00034838, -0.00036716, -0.00036947, -0.00035687, -0.00033179, -0.00029729, -0.00025674, -0.00021353, -0.00017080, -0.00013124, -0.00009690, -0.00006910, -0.00004842, -0.00003476, -0.00002742, -0.00002527, -0.00002691, -0.00003083, -0.00003561, -0.00003999, -0.00004304, -0.00004414, -0.00004305, -0.00003984, -0.00003484, -0.00002857, -0.00002165, -0.00001469, -0.00000823, -0.00000270, 0.00000164, 0.00000469, 0.00000652, 0.00000731, 0.00000732, 0.00000686, 0.00000623, 0.00000566, 0.00000533, 0.00000533, 0.00000567, 0.00000628, 0.00000707, 0.00000790, 0.00000863, 0.00000917, 0.00000944, 0.00000943, 0.00000915, 0.00000866, 0.00000804, 0.00000737, 0.00000672, 0.00000614, 0.00000569, 0.00000536, 0.00000515, 0.00000505, 0.00000505, 0.00000515, },
		{ -0.00003970, -0.00003545, -0.00003117, -0.00002690, -0.00002266, -0.00001846, -0.00001428, -0.00001004, -0.00000566, -0.00000099, 0.00000410, 0.00000974, 0.00001599, 0.00002286, 0.00003026, 0.00003799, 0.00004573, 0.00005308, 0.00005961, 0.00006486, 0.00006848, 0.00007026, 0.00007021, 0.00006861, 0.00006608, 0.00006353, 0.00006213, 0.00006328, 0.00006842, 0.00007893, 0.00009592, 0.00012009, 0.00015154, 0.00018965, 0.00023298, 0.00027927, 0.00032552, 0.00036809, 0.00040291, 0.00042582, 0.00043281, 0.00042044, 0.00038610, 0.00032839, 0.00024728, 0.00014429, 0.00002249, -0.00011358, -0.00025810, -0.00040433, -0.00054502, -0.00067281, -0.00078084, -0.00086311, -0.00091499, -0.00093350, -0.00091758, -0.00086814, -0.00078803, -0.00068183, -0.00055553, -0.00041612, -0.00027111, -0.00012803, 0.00000608, 0.00012510, 0.00022422, 0.00030013, 0.00035120, 0.00037745, 0.00038042, 0.00036296, 0.00032890, 0.00028269, 0.00022901, 0.00017243, 0.00011706, 0.00006629, 0.00002267, -0.00001227, -0.00003791, -0.00005452, -0.00006309, -0.00006515, -0.00006251, -0.00005709, -0.00005071, -0.00004490, -0.00004085, -0.00003930, -0.00004054, -0.00004447, -0.00005067, -0.00005848, -0.00006713, -0.00007585, -0.00008392, -0.00009082, -0.00009617, -0.00009983, -0.00010184, -0.00010240, -0.00010181, -0.00010044, -0.00009865, -0.00009675, -0.00009499, -0.00009351, -0.00009236, -0.00009150, -0.00009082, -0.00009018, -0.00008943, -0.00008843, -0.00008707, -0.00008530, -0.00008309, -0.00008048, -0.00007752, -0.00007428, -0.00007084, -0.00006725, -0.00006356, -0.00005979, -0.00005594, -0.00005202, -0.00004800, -0.00004389, },
		{ 0.00003509, 0.00003713, 0.00003906, 0.00004090, 0.00004270, 0.00004448, 0.00004623, 0.00004793, 0.00004952, 0.00005092, 0.00005204, 0.00005278, 0.00005311, 0.00005302, 0.00005259, 0.00005197, 0.00005141, 0.00005122, 0.00005174, 0.00005334, 0.00005631, 0.00006086, 0.00006704, 0.00007472, 0.00008357, 0.00009307, 0.00010256, 0.00011127, 0.00011848, 0.00012358, 0.00012620, 0.00012635, 0.00012448, 0.00012155, 0.00011905, 0.00011892, 0.00012349, 0.00013525, 0.00015668, 0.00018995, 0.00023671, 0.00029775, 0.00037285, 0.00046058, 0.00055822, 0.00066177, 0.00076613, 0.00086525, 0.00095254, 0.00102125, 0.00106490, 0.00107783, 0.00105555, 0.00099521, 0.00089589, 0.00075874, 0.00058702, 0.00038602, 0.00016278, -0.00007432, -0.00031601, -0.00055271, -0.00077514, -0.00097480, -0.00114455, -0.00127899, -0.00137473, -0.00143052, -0.00144726, -0.00142780, -0.00137667, -0.00129968, -0.00120347, -0.00109502, -0.00098114, -0.00086809, -0.00076122, -0.00066466, -0.00058124, -0.00051243, -0.00045843, -0.00041831, -0.00039028, -0.00037195, -0.00036064, -0.00035361, -0.00034837, -0.00034281, -0.00033536, -0.00032505, -0.00031147, -0.00029474, -0.00027540, -0.00025426, -0.00023229, -0.00021047, -0.00018965, -0.00017052, -0.00015352, -0.00013883, -0.00012640, -0.00011597, -0.00010716, -0.00009951, -0.00009254, -0.00008585, -0.00007909, -0.00007207, -0.00006469, -0.00005697, -0.00004903, -0.00004103, -0.00003315, -0.00002559, -0.00001848, -0.00001193, -0.00000598, -0.00000064, 0.00000415, 0.00000846, 0.00001236, 0.00001594, 0.00001926, 0.00002236, 0.00002527, 0.00002800, 0.00003054, 0.00003290, },
		{ 0.00013328, 0.00013343, 0.00013388, 0.00013462, 0.00013564, 0.00013691, 0.00013845, 0.00014029, 0.00014248, 0.00014507, 0.00014817, 0.00015184, 0.00015615, 0.00016113, 0.00016673, 0.00017287, 0.00017937, 0.00018603, 0.00019259, 0.00019878, 0.00020440, 0.00020929, 0.00021343, 0.00021692, 0.00022003, 0.00022318, 0.00022687, 0.00023171, 0.00023827, 0.00024702, 0.00025827, 0.00027209, 0.00028820, 0.00030601, 0.00032460, 0.00034280, 0.00035926, 0.00037265, 0.00038176, 0.00038577, 0.00038440, 0.00037806, 0.00036799, 0.00035632, 0.00034603, 0.00034084, 0.00034502, 0.00036314, 0.00039967, 0.00045864, 0.00054326, 0.00065553, 0.00079595, 0.00096328, 0.00115447, 0.00136462, 0.00158720, 0.00181430, 0.00203708, 0.00224623, 0.00243255, 0.00258754, 0.00270390, 0.00277607, 0.00280052, 0.00277607, 0.00270390, 0.00258754, 0.00243255, 0.00224623, 0.00203708, 0.00181430, 0.00158720, 0.00136462, 0.00115447, 0.00096328, 0.00079595, 0.00065553, 0.00054326, 0.00045864, 0.00039967, 0.00036314, 0.00034502, 0.00034084, 0.00034603, 0.00035632, 0.00036799, 0.00037806, 0.00038440, 0.00038577, 0.00038176, 0.00037265, 0.00035926, 0.00034280, 0.00032460, 0.00030601, 0.00028820, 0.00027209, 0.00025827, 0.00024702, 0.00023827, 0.00023171, 0.00022687, 0.00022318, 0.00022003, 0.00021692, 0.00021343, 0.00020929, 0.00020440, 0.00019878, 0.00019259, 0.00018603, 0.00017937, 0.00017287, 0.00016673, 0.00016113, 0.00015615, 0.00015184, 0.00014817, 0.00014507, 0.00014248, 0.00014029, 0.00013845, 0.00013691, 0.00013564, 0.00013462, 0.00013388, 0.00013343, },
		{ 0.00003509, 0.00003290, 0.00003054, 0.00002800, 0.00002527, 0.00002236, 0.00001926, 0.00001594, 0.00001236, 0.00000846, 0.00000415, -0.00000064, -0.00000598, -0.00001193, -0.00001848, -0.00002559, -0.00003315, -0.00004103, -0.00004903, -0.00005697, -0.00006469, -0.00007207, -0.00007909, -0.00008585, -0.00009254, -0.00009951, -0.00010716, -0.00011597, -0.00012640, -0.00013883, -0.00015352, -0.00017052, -0.00018965, -0.00021047, -0.00023229, -0.00025426, -0.00027540, -0.00029474, -0.00031147, -0.00032505, -0.00033536, -0.00034281, -0.00034837, -0.00035361, -0.00036064, -0.00037195, -0.00039028, -0.00041831, -0.00045843, -0.00051243, -0.00058124, -0.00066466, -0.00076122, -0.00086809, -0.00098114, -0.00109502, -0.00120347, -0.00129968, -0.00137667, -0.00142780, -0.00144726, -0.00143052, -0.00137473, -0.00127899, -0.00114455, -0.00097480, -0.00077514, -0.00055271, -0.00031601, -0.00007432, 0.00016278, 0.00038602, 0.00058702, 0.00075874, 0.00089589, 0.00099521, 0.00105555, 0.00107783, 0.00106490, 0.00102125, 0.00095254, 0.00086525, 0.00076613, 0.00066177, 0.00055822, 0.00046058, 0.00037285, 0.00029775, 0.00023671, 0.00018995, 0.00015668, 0.00013525, 0.00012349, 0.00011892, 0.00011905, 0.00012155, 0.00012448, 0.00012635, 0.00012620, 0.00012358, 0.00011848, 0.00011127, 0.00010256, 0.00009307, 0.00008357, 0.00007472, 0.00006704, 0.00006086, 0.00005631, 0.00005334, 0.00005174, 0.00005122, 0.00005141, 0.00005197, 0.00005259, 0.00005302, 0.00005311, 0.00005278, 0.00005204, 0.00005092, 0.00004952, 0.00004793, 0.00004623, 0.00004448, 0.00004270, 0.00004090, 0.00003906, 0.00003713, },
		{ -0.00003970, -0.00004389, -0.00004800, -0.00005202, -0.00005594, -0.00005979, -0.00006356, -0.00006725, -0.00007084, -0.00007428, -0.00007752, -0.00008048, -0.00008309, -0.00008530, -0.00008707, -0.00008843, -0.00008943, -0.00009018, -0.00009082, -0.00009150, -0.00009236, -0.00009351, -0.00009499, -0.00009675, -0.00009865, -0.00010044, -0.00010181, -0.00010240, -0.00010184, -0.00009983, -0.00009617, -0.00009082, -0.00008392, -0.00007585, -0.00006713, -0.00005848, -0.00005067, -0.00004447, -0.00004054, -0.00003930, -0.00004085, -0.00004490, -0.00005071, -0.00005709, -0.00006251, -0.00006515, -0.00006309, -0.00005452, -0.00003791, -0.00001227, 0.00002267, 0.00006629, 0.00011706, 0.00017243, 0.00022901, 0.00028269, 0.00032890, 0.00036296, 0.00038042, 0.00037745, 0.00035120, 0.00030013, 0.00022422, 0.00012510, 0.00000608, -0.00012803, -0.00027111, -0.00041612, -0.00055553, -0.00068183, -0.00078803, -0.00086814, -0.00091758, -0.00093350, -0.00091499, -0.00086311, -0.00078084, -0.00067281, -0.00054502, -0.00040433, -0.00025810, -0.00011358, 0.00002249, 0.00014429, 0.00024728, 0.00032839, 0.00038610, 0.00042044, 0.00043281, 0.00042582, 0.00040291, 0.00036809, 0.00032552, 0.00027927, 0.00023298, 0.00018965, 0.00015154, 0.00012009, 0.00009592, 0.00007893, 0.00006842, 0.00006328, 0.00006213, 0.00006353, 0.00006608, 0.00006861, 0.00007021, 0.00007026, 0.00006848, 0.00006486, 0.00005961, 0.00005308, 0.00004573, 0.00003799, 0.00003026, 0.00002286, 0.00001599, 0.00000974, 0.00000410, -0.00000099, -0.00000566, -0.00001004, -0.00001428, -0.00001846, -0.00002266, -0.00002690, -0.00003117, -0.00003545, },
		{ 0.00000535, 0.00000515, 0.00000505, 0.00000505, 0.00000515, 0.00000536, 0.00000569, 0.00000614, 0.00000672, 0.00000737, 0.00000804, 0.00000866, 0.00000915, 0.00000943, 0.00000944, 0.00000917, 0.00000863, 0.00000790, 0.00000707, 0.00000628, 0.00000567, 0.00000533, 0.00000533, 0.00000566, 0.00000623, 0.00000686, 0.00000732, 0.00000731, 0.00000652, 0.00000469, 0.00000164, -0.00000270, -0.00000823, -0.00001469, -0.00002165, -0.00002857, -0.00003484, -0.00003984, -0.00004305, -0.00004414, -0.00004304, -0.00003999, -0.00003561, -0.00003083, -0.00002691, -0.00002527, -0.00002742, -0.00003476, -0.00004842, -0.00006910, -0.00009690, -0.00013124, -0.00017080, -0.00021353, -0.00025674, -0.00029729, -0.00033179, -0.00035687, -0.00036947, -0.00036716, -0.00034838, -0.00031266, -0.00026073, -0.00019459, -0.00011741, -0.00003338, 0.00005255, 0.00013506, 0.00020878, 0.00026874, 0.00031076, 0.00033176, 0.00033009, 0.00030558, 0.00025969, 0.00019535, 0.00011678, 0.00002919, -0.00006160, -0.00014961, -0.00022911, -0.00029505, -0.00034340, -0.00037142, -0.00037786, -0.00036292, -0.00032825, -0.00027676, -0.00021231, -0.00013942, -0.00006289, 0.00001254, 0.00008258, 0.00014361, 0.00019294, 0.00022888, 0.00025081, 0.00025911, 0.00025504, 0.00024053, 0.00021797, 0.00018997, 0.00015915, 0.00012791, 0.00009829, 0.00007184, 0.00004962, 0.00003212, 0.00001939, 0.00001107, 0.00000652, 0.00000491, 0.00000537, 0.00000705, 0.00000923, 0.00001133, 0.00001296, 0.00001392, 0.00001417, 0.00001377, 0.00001287, 0.00001166, 0.00001033, 0.00000904, 0.00000788, 0.00000693, 0.00000621, 0.00000569, },
		{ 0.00001206, 0.00001075, 0.00000931, 0.00000775, 0.00000608, 0.00000430, 0.00000239, 0.00000034, -0.00000182, -0.00000407, -0.00000635, -0.00000858, -0.00001065, -0.00001248, -0.00001399, -0.00001513, -0.00001591, -0.00001636, -0.00001659, -0.00001671, -0.00001688, -0.00001722, -0.00001783, -0.00001875, -0.00001996, -0.00002134, -0.00002271, -0.00002383, -0.00002445, -0.00002432, -0.00002327, -0.00002121, -0.00001817, -0.00001435, -0.00001006, -0.00000574, -0.00000190, 0.00000095, 0.00000237, 0.00000207, -0.00000005, -0.00000379, -0.00000868, -0.00001396, -0.00001865, -0.00002157, -0.00002152, -0.00001738, -0.00000825, 0.00000638, 0.00002656, 0.00005175, 0.00008084, 0.00011211, 0.00014339, 0.00017213, 0.00019564, 0.00021131, 0.00021682, 0.00021043, 0.00019111, 0.00015873, 0.00011415, 0.00005920, -0.00000343, -0.00007030, -0.00013752, -0.00020102, -0.00025684, -0.00030152, -0.00033233, -0.00034753, -0.00034653, -0.00032995, -0.00029960, -0.00025829, -0.00020968, -0.00015790, -0.00010730, -0.00006200, -0.00002561, -0.00000093, 0.00001029, 0.00000747, -0.00000876, -0.00003665, -0.00007348, -0.00011582, -0.00015981, -0.00020153, -0.00023725, -0.00026382, -0.00027884, -0.00028086, -0.00026948, -0.00024527, -0.00020978, -0.00016530, -0.00011466, -0.00006099, -0.00000747, 0.00004296, 0.00008775, 0.00012496, 0.00015332, 0.00017225, 0.00018190, 0.00018299, 0.00017672, 0.00016463, 0.00014841, 0.00012977, 0.00011028, 0.00009129, 0.00007385, 0.00005865, 0.00004608, 0.00003621, 0.00002887, 0.00002373, 0.00002035, 0.00001825, 0.00001700, 0.00001622, 0.00001560, 0.00001496, 0.00001418, 0.00001321, },
		{ 0.00001641, 0.00001377, 0.00001197, 0.00001084, 0.00001024, 0.00001008, 0.00001027, 0.00001075, 0.00001147, 0.00001234, 0.00001329, 0.00001419, 0.00001493, 0.00001540, 0.00001548, 0.00001512, 0.00001431, 0.00001307, 0.00001151, 0.00000975, 0.00000797, 0.00000632, 0.00000495, 0.00000395, 0.00000333, 0.00000304, 0.00000293, 0.00000280, 0.00000241, 0.00000153, -0.00000006, -0.00000247, -0.00000574, -0.00000975, -0.00001426, -0.00001894, -0.00002335, -0.00002706, -0.00002967, -0.00003090, -0.00003063, -0.00002897, -0.00002630, -0.00002322, -0.00002054, -0.00001922, -0.00002028, -0.00002466, -0.00003310, -0.00004606, -0.00006358, -0.00008520, -0.00011000, -0.00013656, -0.00016306, -0.00018741, -0.00020740, -0.00022090, -0.00022608, -0.00022157, -0.00020662, -0.00018126, -0.00014631, -0.00010336, -0.00005473, -0.00000327, 0.00004784, 0.00009532, 0.00013608, 0.00016749, 0.00018755, 0.00019512, 0.00019000, 0.00017296, 0.00014566, 0.00011052, 0.00007051, 0.00002890, -0.00001107, -0.00004640, -0.00007464, -0.00009408, -0.00010388, -0.00010412, -0.00009577, -0.00008060, -0.00006095, -0.00003953, -0.00001914, -0.00000240, 0.00000848, 0.00001191, 0.00000709, -0.00000599, -0.00002651, -0.00005292, -0.00008310, -0.00011458, -0.00014472, -0.00017101, -0.00019122, -0.00020365, -0.00020723, -0.00020156, -0.00018698, -0.00016446, -0.00013554, -0.00010210, -0.00006628, -0.00003022, 0.00000408, 0.00003488, 0.00006087, 0.00008117, 0.00009541, 0.00010365, 0.00010636, 0.00010431, 0.00009847, 0.00008990, 0.00007966, 0.00006872, 0.00005788, 0.00004776, 0.00003877, 0.00003114, 0.00002493, 0.00002006, },
		{ 0.00008161, 0.00007500, 0.00006776, 0.00006030, 0.00005292, 0.00004580, 0.00003907, 0.00003276, 0.00002690, 0.00002149, 0.00001657, 0.00001217, 0.00000835, 0.00000517, 0.00000268, 0.00000090, -0.00000019, -0.00000067, -0.00000066, -0.00000036, 0.00000003, 0.00000030, 0.00000026, -0.00000023, -0.00000123, -0.00000271, -0.00000457, -0.00000661, -0.00000859, -0.00001023, -0.00001130, -0.00001162, -0.00001109, -0.00000975, -0.00000778, -0.00000543, -0.00000308, -0.00000112, 0.00000006, 0.00000016, -0.00000096, -0.00000326, -0.00000649, -0.00001016, -0.00001360, -0.00001598, -0.00001645, -0.00001417, -0.00000846, 0.00000107, 0.00001451, 0.00003150, 0.00005127, 0.00007262, 0.00009402, 0.00011367, 0.00012971, 0.00014034, 0.00014400, 0.00013953, 0.00012634, 0.00010447, 0.00007465, 0.00003827, -0.00000271, -0.00004587, -0.00008853, -0.00012798, -0.00016170, -0.00018753, -0.00020395, -0.00021010, -0.00020595, -0.00019225, -0.00017048, -0.00014272, -0.00011143, -0.00007927, -0.00004884, -0.00002247, -0.00000199, 0.00001140, 0.00001721, 0.00001569, 0.00000779, -0.00000501, -0.00002079, -0.00003747, -0.00005297, -0.00006546, -0.00007350, -0.00007622, -0.00007335, -0.00006526, -0.00005289, -0.00003763, -0.00002121, -0.00000545, 0.00000787, 0.00001724, 0.00002155, 0.00002020, 0.00001315, 0.00000092, -0.00001543, -0.00003449, -0.00005454, -0.00007380, -0.00009054, -0.00010326, -0.00011082, -0.00011251, -0.00010811, -0.00009789, -0.00008255, -0.00006314, -0.00004097, -0.00001746, 0.00000598, 0.00002807, 0.00004772, 0.00006414, 0.00007684, 0.00008564, 0.00009065, 0.00009221, 0.00009082, 0.00008708, },
		{ -0.00000387, 0.00000867, 0.00001924, 0.00002767, 0.00003396, 0.00003825, 0.00004082, 0.00004196, 0.00004199, 0.00004117, 0.00003974, 0.00003782, 0.00003551, 0.00003284, 0.00002982, 0.00002645, 0.00002277, 0.00001886, 0.00001486, 0.00001092, 0.00000727, 0.00000413, 0.00000167, 0.00000005, -0.00000068, -0.00000055, 0.00000030, 0.00000164, 0.00000319, 0.00000463, 0.00000563, 0.00000596, 0.00000545, 0.00000407, 0.00000191, -0.00000078, -0.00000369, -0.00000642, -0.00000858, -0.00000984, -0.00000999, -0.00000898, -0.00000699, -0.00000439, -0.00000174, 0.00000024, 0.00000078, -0.00000088, -0.00000538, -0.00001313, -0.00002426, -0.00003849, -0.00005521, -0.00007339, -0.00009170, -0.00010860, -0.00012245, -0.00013166, -0.00013486, -0.00013105, -0.00011968, -0.00010079, -0.00007501, -0.00004357, -0.00000816, 0.00002909, 0.00006587, 0.00009985, 0.00012884, 0.00015102, 0.00016508, 0.00017036, 0.00016685, 0.00015523, 0.00013680, 0.00011334, 0.00008695, 0.00005988, 0.00003429, 0.00001208, -0.00000528, -0.00001686, -0.00002233, -0.00002199, -0.00001667, -0.00000764, 0.00000351, 0.00001510, 0.00002551, 0.00003331, 0.00003749, 0.00003745, 0.00003310, 0.00002487, 0.00001356, 0.00000035, -0.00001346, -0.00002649, -0.00003752, -0.00004557, -0.00005000, -0.00005063, -0.00004769, -0.00004179, -0.00003389, -0.00002515, -0.00001681, -0.00001007, -0.00000591, -0.00000508, -0.00000791, -0.00001438, -0.00002406, -0.00003619, -0.00004975, -0.00006357, -0.00007645, -0.00008725, -0.00009506, -0.00009919, -0.00009928, -0.00009533, -0.00008762, -0.00007672, -0.00006341, -0.00004859, -0.00003316, -0.00001801, },
		{ -0.00006514, -0.00006290, -0.00005827, -0.00005162, -0.00004344, -0.00003430, -0.00002481, -0.00001550, -0.00000681, 0.00000095, 0.00000760, 0.00001313, 0.00001762, 0.00002125, 0.00002422, 0.00002675, 0.00002898, 0.00003102, 0.00003286, 0.00003444, 0.00003561, 0.00003621, 0.00003608, 0.00003509, 0.00003321, 0.00003048, 0.00002707, 0.00002323, 0.00001930, 0.00001564, 0.00001261, 0.00001051, 0.00000954, 0.00000977, 0.00001110, 0.00001332, 0.00001607, 0.00001892, 0.00002142, 0.00002316, 0.00002386, 0.00002337, 0.00002176, 0.00001932, 0.00001652, 0.00001399, 0.00001248, 0.00001269, 0.00001527, 0.00002067, 0.00002906, 0.00004029, 0.00005384, 0.00006884, 0.00008413, 0.00009833, 0.00010997, 0.00011760, 0.00011998, 0.00011617, 0.00010566, 0.00008847, 0.00006513, 0.00003672, 0.00000476, -0.00002888, -0.00006214, -0.00009293, -0.00011932, -0.00013970, -0.00015293, -0.00015841, -0.00015617, -0.00014683, -0.00013155, -0.00011189, -0.00008973, -0.00006702, -0.00004564, -0.00002723, -0.00001306, -0.00000390, -0.00000001, -0.00000112, -0.00000649, -0.00001501, -0.00002535, -0.00003607, -0.00004583, -0.00005347, -0.00005816, -0.00005947, -0.00005736, -0.00005220, -0.00004470, -0.00003581, -0.00002659, -0.00001811, -0.00001130, -0.00000688, -0.00000524, -0.00000647, -0.00001029, -0.00001616, -0.00002331, -0.00003083, -0.00003783, -0.00004350, -0.00004719, -0.00004855, -0.00004748, -0.00004419, -0.00003916, -0.00003307, -0.00002672, -0.00002093, -0.00001647, -0.00001393, -0.00001370, -0.00001590, -0.00002036, -0.00002670, -0.00003430, -0.00004241, -0.00005023, -0.00005699, -0.00006202, -0.00006483, },
		{ -0.00001232, -0.00001663, -0.00002171, -0.00002706, -0.00003211, -0.00003634, -0.00003930, -0.00004069, -0.00004039, -0.00003844, -0.00003506, -0.00003058, -0.00002545, -0.00002013, -0.00001506, -0.00001060, -0.00000699, -0.00000435, -0.00000262, -0.00000165, -0.00000116, -0.00000085, -0.00000042, 0.00000039, 0.00000173, 0.00000362, 0.00000598, 0.00000855, 0.00001103, 0.00001302, 0.00001416, 0.00001413, 0.00001272, 0.00000986, 0.00000567, 0.00000041, -0.00000551, -0.00001160, -0.00001732, -0.00002217, -0.00002579, -0.00002795, -0.00002865, -0.00002812, -0.00002679, -0.00002530, -0.00002438, -0.00002477, -0.00002717, -0.00003209, -0.00003980, -0.00005023, -0.00006299, -0.00007733, -0.00009221, -0.00010635, -0.00011838, -0.00012696, -0.00013089, -0.00012926, -0.00012155, -0.00010773, -0.00008825, -0.00006407, -0.00003656, -0.00000740, 0.00002152, 0.00004832, 0.00007122, 0.00008877, 0.00009994, 0.00010418, 0.00010153, 0.00009256, 0.00007832, 0.00006025, 0.00004002, 0.00001940, 0.00000007, -0.00001650, -0.00002919, -0.00003734, -0.00004073, -0.00003964, -0.00003472, -0.00002700, -0.00001764, -0.00000793, 0.00000095, 0.00000799, 0.00001247, 0.00001401, 0.00001261, 0.00000857, 0.00000252, -0.00000475, -0.00001233, -0.00001935, -0.00002503, -0.00002880, -0.00003034, -0.00002963, -0.00002689, -0.00002260, -0.00001737, -0.00001192, -0.00000695, -0.00000306, -0.00000071, -0.00000012, -0.00000132, -0.00000408, -0.00000800, -0.00001254, -0.00001713, -0.00002117, -0.00002418, -0.00002583, -0.00002596, -0.00002462, -0.00002207, -0.00001871, -0.00001505, -0.00001164, -0.00000900, -0.00000756, -0.00000760, -0.00000922, },
		{ -0.00002372, -0.00002216, -0.00002054, -0.00001925, -0.00001865, -0.00001903, -0.00002055, -0.00002320, -0.00002685, -0.00003120, -0.00003583, -0.00004029, -0.00004408, -0.00004678, -0.00004806, -0.00004775, -0.00004584, -0.00004250, -0.00003804, -0.00003291, -0.00002761, -0.00002265, -0.00001849, -0.00001544, -0.00001369, -0.00001323, -0.00001388, -0.00001530, -0.00001704, -0.00001857, -0.00001942, -0.00001917, -0.00001754, -0.00001443, -0.00000995, -0.00000440, 0.00000176, 0.00000798, 0.00001363, 0.00001817, 0.00002115, 0.00002232, 0.00002167, 0.00001942, 0.00001605, 0.00001223, 0.00000876, 0.00000649, 0.00000618, 0.00000845, 0.00001364, 0.00002178, 0.00003254, 0.00004525, 0.00005890, 0.00007228, 0.00008403, 0.00009282, 0.00009744, 0.00009698, 0.00009086, 0.00007901, 0.00006179, 0.00004006, 0.00001509, -0.00001155, -0.00003811, -0.00006280, -0.00008398, -0.00010028, -0.00011070, -0.00011476, -0.00011248, -0.00010437, -0.00009143, -0.00007497, -0.00005654, -0.00003777, -0.00002020, -0.00000517, 0.00000631, 0.00001365, 0.00001666, 0.00001561, 0.00001114, 0.00000417, -0.00000422, -0.00001287, -0.00002071, -0.00002683, -0.00003060, -0.00003169, -0.00003012, -0.00002618, -0.00002044, -0.00001362, -0.00000653, 0.00000003, 0.00000538, 0.00000903, 0.00001068, 0.00001031, 0.00000811, 0.00000448, -0.00000004, -0.00000487, -0.00000942, -0.00001319, -0.00001581, -0.00001709, -0.00001704, -0.00001583, -0.00001378, -0.00001132, -0.00000888, -0.00000691, -0.00000574, -0.00000560, -0.00000655, -0.00000852, -0.00001127, -0.00001449, -0.00001779, -0.00002079, -0.00002316, -0.00002466, -0.00002521, -0.00002483, },
		{ -0.00000956, -0.00001081, -0.00001192, -0.00001261, -0.00001268, -0.00001199, -0.00001052, -0.00000839, -0.00000583, -0.00000314, -0.00000071, 0.00000110, 0.00000195, 0.00000161, -0.00000001, -0.00000287, -0.00000675, -0.00001127, -0.00001599, -0.00002035, -0.00002385, -0.00002601, -0.00002651, -0.00002518, -0.00002207, -0.00001741, -0.00001164, -0.00000533, 0.00000089, 0.00000634, 0.00001045, 0.00001276, 0.00001302, 0.00001122, 0.00000758, 0.00000253, -0.00000330, -0.00000920, -0.00001444, -0.00001836, -0.00002045, -0.00002044, -0.00001833, -0.00001440, -0.00000921, -0.00000353, 0.00000174, 0.00000565, 0.00000735, 0.00000615, 0.00000166, -0.00000618, -0.00001703, -0.00003020, -0.00004466, -0.00005913, -0.00007222, -0.00008253, -0.00008883, -0.00009014, -0.00008586, -0.00007587, -0.00006052, -0.00004062, -0.00001742, 0.00000755, 0.00003258, 0.00005592, 0.00007596, 0.00009137, 0.00010118, 0.00010491, 0.00010258, 0.00009469, 0.00008216, 0.00006628, 0.00004851, 0.00003040, 0.00001340, -0.00000122, -0.00001251, -0.00001994, -0.00002333, -0.00002296, -0.00001942, -0.00001357, -0.00000645, 0.00000088, 0.00000744, 0.00001239, 0.00001516, 0.00001548, 0.00001337, 0.00000912, 0.00000326, -0.00000352, -0.00001048, -0.00001689, -0.00002214, -0.00002577, -0.00002754, -0.00002743, -0.00002562, -0.00002248, -0.00001851, -0.00001424, -0.00001020, -0.00000684, -0.00000449, -0.00000334, -0.00000339, -0.00000450, -0.00000640, -0.00000875, -0.00001116, -0.00001330, -0.00001486, -0.00001568, -0.00001569, -0.00001497, -0.00001367, -0.00001204, -0.00001036, -0.00000891, -0.00000791, -0.00000749, -0.00000769, -0.00000844, },
		{ -0.00000658, -0.00000619, -0.00000596, -0.00000608, -0.00000666, -0.00000777, -0.00000935, -0.00001128, -0.00001333, -0.00001525, -0.00001675, -0.00001756, -0.00001750, -0.00001646, -0.00001448, -0.00001172, -0.00000846, -0.00000510, -0.00000208, 0.00000017, 0.00000124, 0.00000086, -0.00000110, -0.00000460, -0.00000939, -0.00001506, -0.00002103, -0.00002667, -0.00003132, -0.00003437, -0.00003536, -0.00003401, -0.00003029, -0.00002441, -0.00001684, -0.00000822, 0.00000063, 0.00000886, 0.00001568, 0.00002039, 0.00002256, 0.00002201, 0.00001888, 0.00001363, 0.00000700, -0.00000007, -0.00000651, -0.00001128, -0.00001343, -0.00001226, -0.00000739, 0.00000117, 0.00001301, 0.00002732, 0.00004294, 0.00005852, 0.00007255, 0.00008357, 0.00009032, 0.00009180, 0.00008745, 0.00007719, 0.00006141, 0.00004102, 0.00001731, -0.00000814, -0.00003356, -0.00005720, -0.00007745, -0.00009298, -0.00010288, -0.00010670, -0.00010450, -0.00009682, -0.00008461, -0.00006917, -0.00005197, -0.00003452, -0.00001828, -0.00000445, 0.00000606, 0.00001273, 0.00001545, 0.00001447, 0.00001042, 0.00000414, -0.00000336, -0.00001106, -0.00001799, -0.00002337, -0.00002667, -0.00002762, -0.00002624, -0.00002284, -0.00001792, -0.00001212, -0.00000614, -0.00000065, 0.00000376, 0.00000669, 0.00000793, 0.00000746, 0.00000547, 0.00000232, -0.00000154, -0.00000560, -0.00000938, -0.00001246, -0.00001454, -0.00001546, -0.00001523, -0.00001397, -0.00001194, -0.00000946, -0.00000687, -0.00000450, -0.00000262, -0.00000140, -0.00000090, -0.00000110, -0.00000185, -0.00000297, -0.00000423, -0.00000542, -0.00000635, -0.00000692, -0.00000710, -0.00000694, },
		{ -0.00000600, -0.00000587, -0.00000569, -0.00000533, -0.00000469, -0.00000376, -0.00000257, -0.00000123, 0.00000008, 0.00000116, 0.00000181, 0.00000186, 0.00000119, -0.00000021, -0.00000226, -0.00000476, -0.00000744, -0.00000995, -0.00001195, -0.00001309, -0.00001311, -0.00001188, -0.00000939, -0.00000581, -0.00000146, 0.00000320, 0.00000763, 0.00001126, 0.00001354, 0.00001405, 0.00001251, 0.00000887, 0.00000328, -0.00000384, -0.00001190, -0.00002014, -0.00002776, -0.00003394, -0.00003798, -0.00003939, -0.00003792, -0.00003364, -0.00002695, -0.00001853, -0.00000931, -0.00000037, 0.00000713, 0.00001214, 0.00001377, 0.00001144, 0.00000493, -0.00000555, -0.00001939, -0.00003557, -0.00005277, -0.00006948, -0.00008413, -0.00009524, -0.00010157, -0.00010222, -0.00009676, -0.00008524, -0.00006824, -0.00004680, -0.00002236, 0.00000339, 0.00002864, 0.00005162, 0.00007076, 0.00008481, 0.00009297, 0.00009491, 0.00009084, 0.00008142, 0.00006772, 0.00005113, 0.00003316, 0.00001535, -0.00000087, -0.00001433, -0.00002420, -0.00003004, -0.00003181, -0.00002988, -0.00002492, -0.00001783, -0.00000967, -0.00000148, 0.00000579, 0.00001136, 0.00001473, 0.00001564, 0.00001417, 0.00001062, 0.00000551, -0.00000051, -0.00000674, -0.00001251, -0.00001726, -0.00002059, -0.00002227, -0.00002232, -0.00002091, -0.00001836, -0.00001512, -0.00001165, -0.00000842, -0.00000581, -0.00000410, -0.00000343, -0.00000379, -0.00000506, -0.00000699, -0.00000928, -0.00001161, -0.00001369, -0.00001525, -0.00001616, -0.00001633, -0.00001582, -0.00001474, -0.00001327, -0.00001162, -0.00001000, -0.00000857, -0.00000745, -0.00000668, -0.00000623, },
		{ -0.00000267, -0.00000258, -0.00000250, -0.00000254, -0.00000278, -0.00000328, -0.00000399, -0.00000483, -0.00000568, -0.00000638, -0.00000677, -0.00000673, -0.00000618, -0.00000515, -0.00000370, -0.00000202, -0.00000034, 0.00000107, 0.00000194, 0.00000202, 0.00000115, -0.00000073, -0.00000355, -0.00000709, -0.00001105, -0.00001499, -0.00001845, -0.00002097, -0.00002214, -0.00002167, -0.00001945, -0.00001555, -0.00001023, -0.00000397, 0.00000264, 0.00000888, 0.00001405, 0.00001749, 0.00001871, 0.00001741, 0.00001357, 0.00000747, -0.00000034, -0.00000906, -0.00001770, -0.00002521, -0.00003053, -0.00003276, -0.00003124, -0.00002561, -0.00001594, -0.00000268, 0.00001333, 0.00003091, 0.00004863, 0.00006496, 0.00007838, 0.00008755, 0.00009140, 0.00008927, 0.00008098, 0.00006687, 0.00004773, 0.00002481, -0.00000032, -0.00002591, -0.00005014, -0.00007134, -0.00008808, -0.00009930, -0.00010442, -0.00010337, -0.00009654, -0.00008483, -0.00006944, -0.00005185, -0.00003362, -0.00001626, -0.00000112, 0.00001077, 0.00001872, 0.00002246, 0.00002216, 0.00001831, 0.00001174, 0.00000345, -0.00000546, -0.00001391, -0.00002098, -0.00002594, -0.00002836, -0.00002809, -0.00002528, -0.00002033, -0.00001386, -0.00000658, 0.00000078, 0.00000749, 0.00001300, 0.00001689, 0.00001897, 0.00001925, 0.00001792, 0.00001534, 0.00001195, 0.00000825, 0.00000469, 0.00000169, -0.00000048, -0.00000167, -0.00000186, -0.00000117, 0.00000017, 0.00000190, 0.00000370, 0.00000530, 0.00000646, 0.00000705, 0.00000699, 0.00000634, 0.00000519, 0.00000373, 0.00000214, 0.00000061, -0.00000070, -0.00000169, -0.00000232, -0.00000262, },
		{ -0.00000824, -0.00000875, -0.00000921, -0.00000948, -0.00000951, -0.00000926, -0.00000875, -0.00000807, -0.00000733, -0.00000667, -0.00000622, -0.00000611, -0.00000638, -0.00000705, -0.00000803, -0.00000919, -0.00001034, -0.00001124, -0.00001167, -0.00001147, -0.00001050, -0.00000874, -0.00000628, -0.00000332, -0.00000013, 0.00000291, 0.00000544, 0.00000708, 0.00000753, 0.00000662, 0.00000431, 0.00000072, -0.00000383, -0.00000892, -0.00001399, -0.00001844, -0.00002169, -0.00002325, -0.00002277, -0.00002013, -0.00001545, -0.00000910, -0.00000168, 0.00000603, 0.00001311, 0.00001863, 0.00002173, 0.00002173, 0.00001819, 0.00001100, 0.00000042, -0.00001295, -0.00002815, -0.00004400, -0.00005912, -0.00007212, -0.00008170, -0.00008675, -0.00008653, -0.00008066, -0.00006925, -0.00005287, -0.00003251, -0.00000949, 0.00001459, 0.00003806, 0.00005929, 0.00007682, 0.00008951, 0.00009662, 0.00009787, 0.00009347, 0.00008408, 0.00007073, 0.00005475, 0.00003760, 0.00002078, 0.00000565, -0.00000664, -0.00001531, -0.00001993, -0.00002050, -0.00001738, -0.00001129, -0.00000317, 0.00000591, 0.00001483, 0.00002260, 0.00002838, 0.00003161, 0.00003201, 0.00002962, 0.00002476, 0.00001799, 0.00001003, 0.00000167, -0.00000628, -0.00001314, -0.00001836, -0.00002159, -0.00002271, -0.00002184, -0.00001925, -0.00001540, -0.00001081, -0.00000604, -0.00000160, 0.00000207, 0.00000468, 0.00000607, 0.00000624, 0.00000531, 0.00000352, 0.00000117, -0.00000140, -0.00000390, -0.00000608, -0.00000775, -0.00000883, -0.00000934, -0.00000935, -0.00000902, -0.00000850, -0.00000797, -0.00000757, -0.00000739, -0.00000746, -0.00000777, },
		{ -0.00000321, -0.00000340, -0.00000359, -0.00000387, -0.00000428, -0.00000485, -0.00000554, -0.00000629, -0.00000698, -0.00000750, -0.00000772, -0.00000756, -0.00000697, -0.00000597, -0.00000462, -0.00000307, -0.00000150, -0.00000011, 0.00000089, 0.00000135, 0.00000115, 0.00000028, -0.00000119, -0.00000309, -0.00000516, -0.00000711, -0.00000861, -0.00000936, -0.00000912, -0.00000776, -0.00000528, -0.00000181, 0.00000237, 0.00000687, 0.00001120, 0.00001488, 0.00001743, 0.00001849, 0.00001780, 0.00001535, 0.00001130, 0.00000602, 0.00000010, -0.00000577, -0.00001079, -0.00001419, -0.00001531, -0.00001362, -0.00000888, -0.00000113, 0.00000929, 0.00002172, 0.00003523, 0.00004874, 0.00006103, 0.00007093, 0.00007736, 0.00007952, 0.00007690, 0.00006936, 0.00005720, 0.00004110, 0.00002205, 0.00000136, -0.00001955, -0.00003922, -0.00005629, -0.00006963, -0.00007840, -0.00008219, -0.00008099, -0.00007520, -0.00006561, -0.00005329, -0.00003949, -0.00002554, -0.00001270, -0.00000207, 0.00000553, 0.00000961, 0.00001005, 0.00000710, 0.00000131, -0.00000649, -0.00001533, -0.00002417, -0.00003205, -0.00003812, -0.00004179, -0.00004273, -0.00004089, -0.00003656, -0.00003022, -0.00002257, -0.00001440, -0.00000652, 0.00000032, 0.00000553, 0.00000868, 0.00000960, 0.00000834, 0.00000517, 0.00000053, -0.00000503, -0.00001087, -0.00001640, -0.00002108, -0.00002451, -0.00002643, -0.00002676, -0.00002559, -0.00002314, -0.00001973, -0.00001576, -0.00001163, -0.00000770, -0.00000429, -0.00000160, 0.00000027, 0.00000131, 0.00000162, 0.00000135, 0.00000068, -0.00000019, -0.00000108, -0.00000187, -0.00000250, -0.00000293, },
		{ 0.00000252, 0.00000332, 0.00000382, 0.00000409, 0.00000424, 0.00000430, 0.00000429, 0.00000420, 0.00000395, 0.00000350, 0.00000275, 0.00000167, 0.00000024, -0.00000149, -0.00000345, -0.00000548, -0.00000743, -0.00000909, -0.00001030, -0.00001090, -0.00001083, -0.00001008, -0.00000874, -0.00000697, -0.00000501, -0.00000316, -0.00000171, -0.00000093, -0.00000105, -0.00000216, -0.00000428, -0.00000728, -0.00001090, -0.00001481, -0.00001857, -0.00002177, -0.00002400, -0.00002494, -0.00002443, -0.00002245, -0.00001916, -0.00001493, -0.00001025, -0.00000574, -0.00000207, 0.00000012, 0.00000028, -0.00000196, -0.00000678, -0.00001406, -0.00002345, -0.00003433, -0.00004589, -0.00005715, -0.00006709, -0.00007475, -0.00007927, -0.00008004, -0.00007672, -0.00006931, -0.00005816, -0.00004393, -0.00002757, -0.00001019, 0.00000699, 0.00002277, 0.00003606, 0.00004601, 0.00005205, 0.00005397, 0.00005191, 0.00004633, 0.00003799, 0.00002786, 0.00001701, 0.00000652, -0.00000263, -0.00000966, -0.00001403, -0.00001552, -0.00001421, -0.00001049, -0.00000495, 0.00000160, 0.00000832, 0.00001435, 0.00001896, 0.00002157, 0.00002187, 0.00001979, 0.00001551, 0.00000946, 0.00000225, -0.00000540, -0.00001276, -0.00001911, -0.00002387, -0.00002664, -0.00002722, -0.00002566, -0.00002218, -0.00001724, -0.00001138, -0.00000524, 0.00000057, 0.00000548, 0.00000904, 0.00001098, 0.00001119, 0.00000974, 0.00000687, 0.00000295, -0.00000160, -0.00000627, -0.00001064, -0.00001430, -0.00001696, -0.00001847, -0.00001878, -0.00001797, -0.00001622, -0.00001378, -0.00001093, -0.00000794, -0.00000507, -0.00000250, -0.00000036, 0.00000131, },
		{ -0.00002021, -0.00001937, -0.00001807, -0.00001650, -0.00001483, -0.00001318, -0.00001164, -0.00001025, -0.00000901, -0.00000787, -0.00000678, -0.00000570, -0.00000460, -0.00000350, -0.00000245, -0.00000153, -0.00000087, -0.00000058, -0.00000079, -0.00000157, -0.00000295, -0.00000489, -0.00000728, -0.00000992, -0.00001259, -0.00001500, -0.00001688, -0.00001799, -0.00001813, -0.00001722, -0.00001528, -0.00001246, -0.00000899, -0.00000523, -0.00000155, 0.00000163, 0.00000394, 0.00000511, 0.00000498, 0.00000353, 0.00000092, -0.00000252, -0.00000634, -0.00001000, -0.00001290, -0.00001448, -0.00001428, -0.00001196, -0.00000740, -0.00000071, 0.00000780, 0.00001756, 0.00002786, 0.00003787, 0.00004671, 0.00005356, 0.00005771, 0.00005865, 0.00005614, 0.00005020, 0.00004116, 0.00002958, 0.00001628, 0.00000220, -0.00001165, -0.00002429, -0.00003488, -0.00004273, -0.00004745, -0.00004889, -0.00004724, -0.00004290, -0.00003653, -0.00002891, -0.00002089, -0.00001330, -0.00000685, -0.00000209, 0.00000064, 0.00000128, -0.00000001, -0.00000285, -0.00000670, -0.00001092, -0.00001485, -0.00001788, -0.00001952, -0.00001947, -0.00001761, -0.00001406, -0.00000913, -0.00000328, 0.00000291, 0.00000881, 0.00001384, 0.00001748, 0.00001938, 0.00001938, 0.00001747, 0.00001390, 0.00000902, 0.00000336, -0.00000252, -0.00000804, -0.00001268, -0.00001602, -0.00001779, -0.00001789, -0.00001642, -0.00001360, -0.00000980, -0.00000549, -0.00000114, 0.00000277, 0.00000583, 0.00000774, 0.00000835, 0.00000760, 0.00000562, 0.00000262, -0.00000108, -0.00000514, -0.00000920, -0.00001292, -0.00001603, -0.00001836, -0.00001982, -0.00002040, },
		{ 0.00000127, -0.00000138, -0.00000414, -0.00000673, -0.00000894, -0.00001062, -0.00001169, -0.00001215, -0.00001204, -0.00001147, -0.00001056, -0.00000946, -0.00000826, -0.00000705, -0.00000586, -0.00000469, -0.00000348, -0.00000218, -0.00000071, 0.00000098, 0.00000289, 0.00000501, 0.00000722, 0.00000939, 0.00001133, 0.00001282, 0.00001366, 0.00001367, 0.00001275, 0.00001085, 0.00000805, 0.00000453, 0.00000056, -0.00000354, -0.00000739, -0.00001060, -0.00001284, -0.00001387, -0.00001356, -0.00001194, -0.00000916, -0.00000556, -0.00000157, 0.00000228, 0.00000544, 0.00000739, 0.00000769, 0.00000604, 0.00000234, -0.00000334, -0.00001069, -0.00001921, -0.00002825, -0.00003708, -0.00004490, -0.00005097, -0.00005467, -0.00005555, -0.00005337, -0.00004815, -0.00004015, -0.00002987, -0.00001799, -0.00000531, 0.00000731, 0.00001904, 0.00002913, 0.00003703, 0.00004238, 0.00004503, 0.00004511, 0.00004293, 0.00003899, 0.00003389, 0.00002828, 0.00002279, 0.00001796, 0.00001419, 0.00001172, 0.00001057, 0.00001062, 0.00001157, 0.00001303, 0.00001453, 0.00001565, 0.00001597, 0.00001525, 0.00001333, 0.00001024, 0.00000618, 0.00000146, -0.00000351, -0.00000826, -0.00001233, -0.00001532, -0.00001695, -0.00001705, -0.00001563, -0.00001286, -0.00000904, -0.00000459, 0.00000002, 0.00000429, 0.00000778, 0.00001013, 0.00001111, 0.00001064, 0.00000878, 0.00000575, 0.00000188, -0.00000240, -0.00000665, -0.00001044, -0.00001340, -0.00001525, -0.00001584, -0.00001516, -0.00001333, -0.00001059, -0.00000726, -0.00000371, -0.00000032, 0.00000257, 0.00000470, 0.00000588, 0.00000604, 0.00000522, 0.00000356, },
		{ -0.00000064, 0.00000157, 0.00000341, 0.00000466, 0.00000517, 0.00000488, 0.00000381, 0.00000208, -0.00000014, -0.00000264, -0.00000519, -0.00000757, -0.00000962, -0.00001122, -0.00001231, -0.00001291, -0.00001309, -0.00001294, -0.00001261, -0.00001221, -0.00001183, -0.00001152, -0.00001127, -0.00001103, -0.00001070, -0.00001015, -0.00000923, -0.00000783, -0.00000587, -0.00000337, -0.00000038, 0.00000294, 0.00000636, 0.00000960, 0.00001234, 0.00001429, 0.00001520, 0.00001490, 0.00001333, 0.00001056, 0.00000682, 0.00000244, -0.00000214, -0.00000642, -0.00000986, -0.00001200, -0.00001243, -0.00001091, -0.00000736, -0.00000189, 0.00000520, 0.00001343, 0.00002217, 0.00003071, 0.00003833, 0.00004433, 0.00004813, 0.00004931, 0.00004764, 0.00004313, 0.00003600, 0.00002671, 0.00001585, 0.00000415, -0.00000762, -0.00001871, -0.00002843, -0.00003627, -0.00004185, -0.00004503, -0.00004584, -0.00004453, -0.00004147, -0.00003714, -0.00003208, -0.00002681, -0.00002178, -0.00001735, -0.00001375, -0.00001107, -0.00000925, -0.00000813, -0.00000747, -0.00000698, -0.00000638, -0.00000545, -0.00000402, -0.00000205, 0.00000041, 0.00000320, 0.00000607, 0.00000875, 0.00001092, 0.00001231, 0.00001272, 0.00001204, 0.00001025, 0.00000750, 0.00000400, 0.00000008, -0.00000390, -0.00000754, -0.00001050, -0.00001246, -0.00001325, -0.00001280, -0.00001115, -0.00000850, -0.00000512, -0.00000136, 0.00000237, 0.00000571, 0.00000833, 0.00000999, 0.00001054, 0.00000998, 0.00000840, 0.00000603, 0.00000315, 0.00000009, -0.00000280, -0.00000522, -0.00000692, -0.00000775, -0.00000766, -0.00000672, -0.00000507, -0.00000295, },
		{ -0.00000166, -0.00000283, -0.00000349, -0.00000352, -0.00000288, -0.00000161, 0.00000015, 0.00000222, 0.00000436, 0.00000632, 0.00000789, 0.00000887, 0.00000914, 0.00000867, 0.00000750, 0.00000574, 0.00000357, 0.00000121, -0.00000113, -0.00000323, -0.00000495, -0.00000618, -0.00000688, -0.00000711, -0.00000697, -0.00000661, -0.00000620, -0.00000590, -0.00000585, -0.00000612, -0.00000672, -0.00000758, -0.00000856, -0.00000947, -0.00001008, -0.00001018, -0.00000958, -0.00000816, -0.00000591, -0.00000291, 0.00000064, 0.00000444, 0.00000809, 0.00001118, 0.00001326, 0.00001395, 0.00001294, 0.00001007, 0.00000532, -0.00000115, -0.00000900, -0.00001774, -0.00002678, -0.00003544, -0.00004305, -0.00004896, -0.00005266, -0.00005378, -0.00005213, -0.00004773, -0.00004084, -0.00003187, -0.00002140, -0.00001012, 0.00000126, 0.00001202, 0.00002153, 0.00002928, 0.00003493, 0.00003830, 0.00003943, 0.00003848, 0.00003580, 0.00003180, 0.00002697, 0.00002177, 0.00001664, 0.00001194, 0.00000792, 0.00000472, 0.00000236, 0.00000077, -0.00000021, -0.00000077, -0.00000109, -0.00000136, -0.00000171, -0.00000220, -0.00000283, -0.00000351, -0.00000414, -0.00000457, -0.00000463, -0.00000422, -0.00000326, -0.00000174, 0.00000026, 0.00000261, 0.00000510, 0.00000748, 0.00000948, 0.00001087, 0.00001144, 0.00001106, 0.00000971, 0.00000745, 0.00000442, 0.00000087, -0.00000291, -0.00000658, -0.00000983, -0.00001237, -0.00001401, -0.00001461, -0.00001417, -0.00001277, -0.00001058, -0.00000786, -0.00000492, -0.00000204, 0.00000048, 0.00000242, 0.00000363, 0.00000405, 0.00000371, 0.00000276, 0.00000138, -0.00000018, },
		{ 0.00000180, 0.00000294, 0.00000362, 0.00000376, 0.00000339, 0.00000259, 0.00000151, 0.00000034, -0.00000072, -0.00000146, -0.00000175, -0.00000149, -0.00000065, 0.00000071, 0.00000246, 0.00000440, 0.00000632, 0.00000797, 0.00000913, 0.00000962, 0.00000935, 0.00000828, 0.00000647, 0.00000406, 0.00000125, -0.00000169, -0.00000452, -0.00000698, -0.00000888, -0.00001009, -0.00001055, -0.00001032, -0.00000952, -0.00000833, -0.00000701, -0.00000579, -0.00000489, -0.00000449, -0.00000466, -0.00000539, -0.00000655, -0.00000795, -0.00000928, -0.00001020, -0.00001038, -0.00000950, -0.00000733, -0.00000375, 0.00000124, 0.00000747, 0.00001464, 0.00002232, 0.00002998, 0.00003705, 0.00004294, 0.00004713, 0.00004920, 0.00004886, 0.00004600, 0.00004069, 0.00003320, 0.00002396, 0.00001350, 0.00000247, -0.00000846, -0.00001866, -0.00002754, -0.00003466, -0.00003973, -0.00004259, -0.00004328, -0.00004200, -0.00003904, -0.00003481, -0.00002976, -0.00002434, -0.00001896, -0.00001398, -0.00000965, -0.00000613, -0.00000347, -0.00000162, -0.00000047, 0.00000014, 0.00000037, 0.00000042, 0.00000041, 0.00000044, 0.00000057, 0.00000080, 0.00000108, 0.00000136, 0.00000156, 0.00000162, 0.00000151, 0.00000124, 0.00000085, 0.00000043, 0.00000010, 0.00000000, 0.00000025, 0.00000097, 0.00000220, 0.00000397, 0.00000621, 0.00000881, 0.00001158, 0.00001432, 0.00001677, 0.00001872, 0.00001996, 0.00002035, 0.00001982, 0.00001837, 0.00001610, 0.00001318, 0.00000985, 0.00000636, 0.00000300, 0.00000003, -0.00000234, -0.00000396, -0.00000476, -0.00000476, -0.00000406, -0.00000284, -0.00000130, 0.00000032, },
		{ -0.00000570, -0.00000747, -0.00000860, -0.00000909, -0.00000899, -0.00000840, -0.00000749, -0.00000642, -0.00000539, -0.00000455, -0.00000401, -0.00000383, -0.00000399, -0.00000442, -0.00000499, -0.00000552, -0.00000585, -0.00000579, -0.00000522, -0.00000404, -0.00000226, 0.00000006, 0.00000277, 0.00000566, 0.00000847, 0.00001094, 0.00001280, 0.00001385, 0.00001394, 0.00001300, 0.00001110, 0.00000835, 0.00000499, 0.00000131, -0.00000237, -0.00000573, -0.00000848, -0.00001041, -0.00001138, -0.00001140, -0.00001055, -0.00000905, -0.00000718, -0.00000529, -0.00000375, -0.00000287, -0.00000295, -0.00000414, -0.00000650, -0.00000994, -0.00001424, -0.00001907, -0.00002399, -0.00002853, -0.00003220, -0.00003456, -0.00003526, -0.00003405, -0.00003084, -0.00002570, -0.00001886, -0.00001068, -0.00000162, 0.00000777, 0.00001694, 0.00002534, 0.00003251, 0.00003809, 0.00004184, 0.00004367, 0.00004363, 0.00004191, 0.00003880, 0.00003466, 0.00002990, 0.00002491, 0.00002006, 0.00001566, 0.00001192, 0.00000898, 0.00000687, 0.00000554, 0.00000489, 0.00000476, 0.00000499, 0.00000540, 0.00000584, 0.00000620, 0.00000642, 0.00000646, 0.00000634, 0.00000610, 0.00000578, 0.00000544, 0.00000511, 0.00000483, 0.00000457, 0.00000432, 0.00000402, 0.00000362, 0.00000306, 0.00000232, 0.00000136, 0.00000023, -0.00000102, -0.00000230, -0.00000348, -0.00000443, -0.00000502, -0.00000514, -0.00000471, -0.00000371, -0.00000216, -0.00000015, 0.00000217, 0.00000462, 0.00000699, 0.00000905, 0.00001062, 0.00001153, 0.00001167, 0.00001102, 0.00000960, 0.00000753, 0.00000498, 0.00000215, -0.00000072, -0.00000340, },
		{ 0.00001094, 0.00001108, 0.00001051, 0.00000931, 0.00000758, 0.00000549, 0.00000322, 0.00000098, -0.00000106, -0.00000276, -0.00000401, -0.00000478, -0.00000508, -0.00000499, -0.00000463, -0.00000415, -0.00000369, -0.00000339, -0.00000335, -0.00000361, -0.00000415, -0.00000491, -0.00000574, -0.00000647, -0.00000692, -0.00000689, -0.00000622, -0.00000482, -0.00000266, 0.00000022, 0.00000368, 0.00000748, 0.00001136, 0.00001500, 0.00001807, 0.00002030, 0.00002145, 0.00002137, 0.00002006, 0.00001757, 0.00001414, 0.00001004, 0.00000567, 0.00000144, -0.00000222, -0.00000493, -0.00000639, -0.00000641, -0.00000491, -0.00000199, 0.00000215, 0.00000717, 0.00001263, 0.00001807, 0.00002297, 0.00002689, 0.00002944, 0.00003033, 0.00002943, 0.00002674, 0.00002240, 0.00001670, 0.00001000, 0.00000275, -0.00000458, -0.00001152, -0.00001766, -0.00002269, -0.00002637, -0.00002861, -0.00002942, -0.00002893, -0.00002735, -0.00002494, -0.00002203, -0.00001892, -0.00001589, -0.00001317, -0.00001092, -0.00000925, -0.00000817, -0.00000765, -0.00000757, -0.00000783, -0.00000828, -0.00000879, -0.00000924, -0.00000954, -0.00000965, -0.00000955, -0.00000927, -0.00000885, -0.00000834, -0.00000779, -0.00000726, -0.00000678, -0.00000636, -0.00000599, -0.00000565, -0.00000531, -0.00000494, -0.00000450, -0.00000400, -0.00000344, -0.00000285, -0.00000228, -0.00000178, -0.00000143, -0.00000129, -0.00000140, -0.00000178, -0.00000241, -0.00000325, -0.00000420, -0.00000516, -0.00000599, -0.00000657, -0.00000677, -0.00000650, -0.00000570, -0.00000438, -0.00000258, -0.00000041, 0.00000198, 0.00000442, 0.00000672, 0.00000867, 0.00001012, },
		{ -0.00000262, -0.00000138, 0.00000025, 0.00000217, 0.00000424, 0.00000628, 0.00000812, 0.00000961, 0.00001060, 0.00001102, 0.00001082, 0.00001003, 0.00000872, 0.00000700, 0.00000504, 0.00000298, 0.00000101, -0.00000076, -0.00000220, -0.00000328, -0.00000399, -0.00000439, -0.00000458, -0.00000469, -0.00000486, -0.00000523, -0.00000589, -0.00000691, -0.00000828, -0.00000993, -0.00001175, -0.00001355, -0.00001512, -0.00001623, -0.00001668, -0.00001631, -0.00001502, -0.00001279, -0.00000973, -0.00000601, -0.00000191, 0.00000223, 0.00000603, 0.00000910, 0.00001107, 0.00001164, 0.00001060, 0.00000789, 0.00000354, -0.00000225, -0.00000917, -0.00001680, -0.00002465, -0.00003219, -0.00003889, -0.00004429, -0.00004800, -0.00004974, -0.00004939, -0.00004695, -0.00004259, -0.00003658, -0.00002931, -0.00002122, -0.00001280, -0.00000452, 0.00000318, 0.00000995, 0.00001552, 0.00001973, 0.00002254, 0.00002399, 0.00002421, 0.00002342, 0.00002184, 0.00001973, 0.00001734, 0.00001487, 0.00001250, 0.00001034, 0.00000847, 0.00000689, 0.00000558, 0.00000450, 0.00000358, 0.00000276, 0.00000199, 0.00000123, 0.00000046, -0.00000031, -0.00000105, -0.00000173, -0.00000233, -0.00000279, -0.00000311, -0.00000327, -0.00000328, -0.00000317, -0.00000298, -0.00000275, -0.00000253, -0.00000233, -0.00000220, -0.00000214, -0.00000213, -0.00000215, -0.00000216, -0.00000212, -0.00000199, -0.00000174, -0.00000136, -0.00000086, -0.00000027, 0.00000035, 0.00000095, 0.00000145, 0.00000177, 0.00000186, 0.00000168, 0.00000123, 0.00000054, -0.00000034, -0.00000131, -0.00000225, -0.00000304, -0.00000356, -0.00000370, -0.00000340, },
		{ -0.00000209, -0.00000302, -0.00000409, -0.00000520, -0.00000624, -0.00000710, -0.00000767, -0.00000787, -0.00000764, -0.00000696, -0.00000588, -0.00000445, -0.00000279, -0.00000103, 0.00000067, 0.00000217, 0.00000334, 0.00000406, 0.00000429, 0.00000401, 0.00000326, 0.00000213, 0.00000073, -0.00000079, -0.00000228, -0.00000360, -0.00000464, -0.00000532, -0.00000564, -0.00000561, -0.00000533, -0.00000491, -0.00000449, -0.00000422, -0.00000424, -0.00000464, -0.00000546, -0.00000669, -0.00000823, -0.00000993, -0.00001157, -0.00001290, -0.00001366, -0.00001359, -0.00001248, -0.00001018, -0.00000664, -0.00000189, 0.00000391, 0.00001051, 0.00001756, 0.00002466, 0.00003135, 0.00003719, 0.00004175, 0.00004469, 0.00004575, 0.00004479, 0.00004180, 0.00003688, 0.00003029, 0.00002237, 0.00001353, 0.00000426, -0.00000498, -0.00001371, -0.00002154, -0.00002812, -0.00003322, -0.00003671, -0.00003855, -0.00003882, -0.00003766, -0.00003530, -0.00003198, -0.00002798, -0.00002358, -0.00001903, -0.00001455, -0.00001031, -0.00000644, -0.00000302, -0.00000009, 0.00000234, 0.00000429, 0.00000578, 0.00000686, 0.00000757, 0.00000793, 0.00000800, 0.00000779, 0.00000734, 0.00000667, 0.00000583, 0.00000486, 0.00000380, 0.00000271, 0.00000165, 0.00000068, -0.00000016, -0.00000083, -0.00000129, -0.00000155, -0.00000161, -0.00000152, -0.00000130, -0.00000103, -0.00000075, -0.00000051, -0.00000037, -0.00000034, -0.00000044, -0.00000064, -0.00000093, -0.00000126, -0.00000158, -0.00000184, -0.00000199, -0.00000202, -0.00000192, -0.00000170, -0.00000140, -0.00000108, -0.00000080, -0.00000063, -0.00000065, -0.00000088, -0.00000137, },
		{ -0.00000096, -0.00000045, 0.00000013, 0.00000071, 0.00000124, 0.00000164, 0.00000188, 0.00000190, 0.00000170, 0.00000128, 0.00000069, -0.00000001, -0.00000073, -0.00000137, -0.00000184, -0.00000204, -0.00000190, -0.00000140, -0.00000053, 0.00000068, 0.00000214, 0.00000376, 0.00000540, 0.00000693, 0.00000821, 0.00000912, 0.00000960, 0.00000958, 0.00000908, 0.00000814, 0.00000684, 0.00000532, 0.00000370, 0.00000214, 0.00000076, -0.00000034, -0.00000109, -0.00000148, -0.00000156, -0.00000142, -0.00000119, -0.00000104, -0.00000113, -0.00000164, -0.00000269, -0.00000437, -0.00000670, -0.00000966, -0.00001311, -0.00001687, -0.00002071, -0.00002436, -0.00002751, -0.00002989, -0.00003123, -0.00003136, -0.00003014, -0.00002755, -0.00002363, -0.00001854, -0.00001251, -0.00000582, 0.00000117, 0.00000811, 0.00001466, 0.00002047, 0.00002528, 0.00002889, 0.00003117, 0.00003207, 0.00003164, 0.00002999, 0.00002729, 0.00002377, 0.00001966, 0.00001523, 0.00001072, 0.00000636, 0.00000235, -0.00000118, -0.00000410, -0.00000636, -0.00000794, -0.00000884, -0.00000911, -0.00000882, -0.00000804, -0.00000686, -0.00000537, -0.00000366, -0.00000184, 0.00000002, 0.00000182, 0.00000349, 0.00000496, 0.00000616, 0.00000705, 0.00000758, 0.00000773, 0.00000750, 0.00000691, 0.00000599, 0.00000482, 0.00000344, 0.00000196, 0.00000047, -0.00000096, -0.00000224, -0.00000331, -0.00000411, -0.00000463, -0.00000487, -0.00000486, -0.00000463, -0.00000424, -0.00000377, -0.00000328, -0.00000282, -0.00000244, -0.00000216, -0.00000199, -0.00000192, -0.00000191, -0.00000193, -0.00000192, -0.00000185, -0.00000167, -0.00000138, },
		{ -0.00000148, -0.00000145, -0.00000151, -0.00000162, -0.00000177, -0.00000189, -0.00000197, -0.00000197, -0.00000187, -0.00000169, -0.00000142, -0.00000112, -0.00000083, -0.00000059, -0.00000046, -0.00000048, -0.00000067, -0.00000105, -0.00000159, -0.00000226, -0.00000300, -0.00000372, -0.00000433, -0.00000476, -0.00000492, -0.00000476, -0.00000425, -0.00000338, -0.00000219, -0.00000075, 0.00000084, 0.00000247, 0.00000401, 0.00000534, 0.00000635, 0.00000697, 0.00000713, 0.00000686, 0.00000618, 0.00000517, 0.00000397, 0.00000270, 0.00000152, 0.00000059, 0.00000005, -0.00000002, 0.00000045, 0.00000148, 0.00000303, 0.00000499, 0.00000724, 0.00000961, 0.00001190, 0.00001392, 0.00001549, 0.00001645, 0.00001669, 0.00001614, 0.00001480, 0.00001270, 0.00000994, 0.00000666, 0.00000304, -0.00000072, -0.00000444, -0.00000791, -0.00001098, -0.00001352, -0.00001541, -0.00001663, -0.00001716, -0.00001705, -0.00001636, -0.00001520, -0.00001370, -0.00001199, -0.00001019, -0.00000844, -0.00000684, -0.00000548, -0.00000443, -0.00000372, -0.00000337, -0.00000337, -0.00000369, -0.00000428, -0.00000509, -0.00000604, -0.00000705, -0.00000804, -0.00000893, -0.00000965, -0.00001012, -0.00001029, -0.00001012, -0.00000958, -0.00000867, -0.00000741, -0.00000582, -0.00000397, -0.00000195, 0.00000018, 0.00000229, 0.00000429, 0.00000608, 0.00000757, 0.00000868, 0.00000936, 0.00000959, 0.00000937, 0.00000874, 0.00000775, 0.00000649, 0.00000504, 0.00000352, 0.00000201, 0.00000062, -0.00000059, -0.00000155, -0.00000226, -0.00000269, -0.00000287, -0.00000284, -0.00000267, -0.00000240, -0.00000209, -0.00000182, -0.00000160, },
		{ -0.00000192, -0.00000230, -0.00000243, -0.00000235, -0.00000210, -0.00000174, -0.00000132, -0.00000090, -0.00000050, -0.00000016, 0.00000011, 0.00000031, 0.00000045, 0.00000055, 0.00000065, 0.00000076, 0.00000091, 0.00000111, 0.00000136, 0.00000163, 0.00000192, 0.00000217, 0.00000236, 0.00000243, 0.00000236, 0.00000213, 0.00000172, 0.00000115, 0.00000044, -0.00000035, -0.00000116, -0.00000192, -0.00000255, -0.00000298, -0.00000316, -0.00000303, -0.00000260, -0.00000189, -0.00000093, 0.00000019, 0.00000136, 0.00000248, 0.00000340, 0.00000400, 0.00000417, 0.00000381, 0.00000286, 0.00000130, -0.00000085, -0.00000353, -0.00000662, -0.00000999, -0.00001347, -0.00001687, -0.00002000, -0.00002267, -0.00002472, -0.00002601, -0.00002643, -0.00002595, -0.00002454, -0.00002225, -0.00001916, -0.00001541, -0.00001114, -0.00000655, -0.00000182, 0.00000285, 0.00000727, 0.00001127, 0.00001472, 0.00001750, 0.00001954, 0.00002081, 0.00002132, 0.00002110, 0.00002023, 0.00001881, 0.00001697, 0.00001482, 0.00001251, 0.00001015, 0.00000788, 0.00000577, 0.00000392, 0.00000235, 0.00000110, 0.00000016, -0.00000052, -0.00000097, -0.00000126, -0.00000148, -0.00000168, -0.00000195, -0.00000233, -0.00000286, -0.00000355, -0.00000439, -0.00000535, -0.00000637, -0.00000739, -0.00000830, -0.00000904, -0.00000953, -0.00000968, -0.00000947, -0.00000885, -0.00000784, -0.00000647, -0.00000480, -0.00000291, -0.00000091, 0.00000108, 0.00000297, 0.00000463, 0.00000598, 0.00000695, 0.00000749, 0.00000760, 0.00000730, 0.00000663, 0.00000566, 0.00000448, 0.00000319, 0.00000188, 0.00000065, -0.00000042, -0.00000129, },
		{ 0.00001107, 0.00001036, 0.00000934, 0.00000808, 0.00000665, 0.00000515, 0.00000367, 0.00000229, 0.00000106, 0.00000003, -0.00000078, -0.00000135, -0.00000171, -0.00000188, -0.00000190, -0.00000181, -0.00000166, -0.00000148, -0.00000131, -0.00000116, -0.00000106, -0.00000099, -0.00000095, -0.00000092, -0.00000089, -0.00000084, -0.00000074, -0.00000060, -0.00000041, -0.00000018, 0.00000007, 0.00000030, 0.00000050, 0.00000063, 0.00000066, 0.00000058, 0.00000037, 0.00000004, -0.00000038, -0.00000085, -0.00000132, -0.00000170, -0.00000193, -0.00000191, -0.00000157, -0.00000085, 0.00000031, 0.00000194, 0.00000401, 0.00000649, 0.00000930, 0.00001232, 0.00001542, 0.00001842, 0.00002115, 0.00002341, 0.00002502, 0.00002583, 0.00002571, 0.00002458, 0.00002239, 0.00001916, 0.00001499, 0.00000999, 0.00000436, -0.00000166, -0.00000784, -0.00001387, -0.00001950, -0.00002445, -0.00002851, -0.00003149, -0.00003326, -0.00003376, -0.00003300, -0.00003106, -0.00002808, -0.00002424, -0.00001978, -0.00001495, -0.00001002, -0.00000524, -0.00000083, 0.00000300, 0.00000613, 0.00000845, 0.00000996, 0.00001066, 0.00001065, 0.00001004, 0.00000898, 0.00000764, 0.00000617, 0.00000474, 0.00000346, 0.00000243, 0.00000172, 0.00000133, 0.00000126, 0.00000144, 0.00000181, 0.00000226, 0.00000270, 0.00000305, 0.00000323, 0.00000318, 0.00000289, 0.00000238, 0.00000166, 0.00000082, -0.00000006, -0.00000090, -0.00000159, -0.00000205, -0.00000220, -0.00000200, -0.00000142, -0.00000049, 0.00000076, 0.00000226, 0.00000391, 0.00000561, 0.00000726, 0.00000873, 0.00000995, 0.00001083, 0.00001132, 0.00001140, },
		{ -0.00000289, -0.00000137, 0.00000027, 0.00000192, 0.00000350, 0.00000490, 0.00000605, 0.00000688, 0.00000737, 0.00000748, 0.00000724, 0.00000669, 0.00000586, 0.00000483, 0.00000368, 0.00000247, 0.00000130, 0.00000021, -0.00000072, -0.00000147, -0.00000200, -0.00000232, -0.00000243, -0.00000237, -0.00000216, -0.00000184, -0.00000147, -0.00000110, -0.00000074, -0.00000045, -0.00000023, -0.00000010, -0.00000005, -0.00000006, -0.00000011, -0.00000017, -0.00000022, -0.00000022, -0.00000017, -0.00000006, 0.00000010, 0.00000028, 0.00000043, 0.00000051, 0.00000044, 0.00000017, -0.00000038, -0.00000124, -0.00000246, -0.00000402, -0.00000591, -0.00000805, -0.00001034, -0.00001264, -0.00001481, -0.00001665, -0.00001799, -0.00001867, -0.00001852, -0.00001744, -0.00001537, -0.00001232, -0.00000834, -0.00000358, 0.00000179, 0.00000750, 0.00001329, 0.00001884, 0.00002385, 0.00002804, 0.00003117, 0.00003306, 0.00003359, 0.00003274, 0.00003058, 0.00002725, 0.00002296, 0.00001799, 0.00001265, 0.00000728, 0.00000220, -0.00000229, -0.00000595, -0.00000860, -0.00001014, -0.00001055, -0.00000990, -0.00000831, -0.00000598, -0.00000316, -0.00000008, 0.00000298, 0.00000580, 0.00000816, 0.00000992, 0.00001100, 0.00001135, 0.00001101, 0.00001007, 0.00000866, 0.00000693, 0.00000505, 0.00000317, 0.00000146, 0.00000001, -0.00000108, -0.00000180, -0.00000214, -0.00000215, -0.00000191, -0.00000152, -0.00000107, -0.00000068, -0.00000043, -0.00000039, -0.00000059, -0.00000104, -0.00000170, -0.00000254, -0.00000347, -0.00000439, -0.00000521, -0.00000584, -0.00000620, -0.00000624, -0.00000591, -0.00000523, -0.00000420, },
		{ -0.00000157, -0.00000238, -0.00000314, -0.00000379, -0.00000423, -0.00000441, -0.00000428, -0.00000383, -0.00000306, -0.00000201, -0.00000073, 0.00000071, 0.00000223, 0.00000375, 0.00000517, 0.00000641, 0.00000742, 0.00000814, 0.00000852, 0.00000858, 0.00000830, 0.00000772, 0.00000689, 0.00000587, 0.00000473, 0.00000355, 0.00000241, 0.00000138, 0.00000051, -0.00000013, -0.00000052, -0.00000065, -0.00000054, -0.00000021, 0.00000029, 0.00000090, 0.00000154, 0.00000216, 0.00000270, 0.00000310, 0.00000336, 0.00000346, 0.00000344, 0.00000335, 0.00000325, 0.00000323, 0.00000339, 0.00000382, 0.00000458, 0.00000573, 0.00000727, 0.00000917, 0.00001136, 0.00001372, 0.00001608, 0.00001827, 0.00002007, 0.00002128, 0.00002173, 0.00002125, 0.00001976, 0.00001723, 0.00001369, 0.00000927, 0.00000415, -0.00000142, -0.00000713, -0.00001267, -0.00001772, -0.00002198, -0.00002520, -0.00002718, -0.00002785, -0.00002717, -0.00002525, -0.00002225, -0.00001843, -0.00001409, -0.00000957, -0.00000521, -0.00000134, 0.00000178, 0.00000393, 0.00000500, 0.00000495, 0.00000385, 0.00000183, -0.00000088, -0.00000402, -0.00000730, -0.00001044, -0.00001315, -0.00001523, -0.00001650, -0.00001689, -0.00001639, -0.00001505, -0.00001301, -0.00001046, -0.00000762, -0.00000470, -0.00000193, 0.00000050, 0.00000242, 0.00000375, 0.00000444, 0.00000451, 0.00000403, 0.00000310, 0.00000188, 0.00000050, -0.00000087, -0.00000210, -0.00000307, -0.00000372, -0.00000401, -0.00000395, -0.00000358, -0.00000297, -0.00000222, -0.00000143, -0.00000070, -0.00000012, 0.00000025, 0.00000036, 0.00000020, -0.00000021, -0.00000082, },
		{ 0.00000001, 0.00000086, 0.00000163, 0.00000226, 0.00000266, 0.00000282, 0.00000271, 0.00000234, 0.00000175, 0.00000100, 0.00000014, -0.00000073, -0.00000155, -0.00000224, -0.00000275, -0.00000302, -0.00000302, -0.00000274, -0.00000219, -0.00000139, -0.00000040, 0.00000073, 0.00000192, 0.00000311, 0.00000421, 0.00000514, 0.00000583, 0.00000622, 0.00000628, 0.00000597, 0.00000530, 0.00000430, 0.00000301, 0.00000151, -0.00000013, -0.00000180, -0.00000341, -0.00000485, -0.00000605, -0.00000695, -0.00000751, -0.00000774, -0.00000768, -0.00000741, -0.00000702, -0.00000663, -0.00000639, -0.00000642, -0.00000683, -0.00000771, -0.00000909, -0.00001097, -0.00001325, -0.00001583, -0.00001853, -0.00002113, -0.00002339, -0.00002507, -0.00002595, -0.00002586, -0.00002467, -0.00002232, -0.00001886, -0.00001441, -0.00000916, -0.00000340, 0.00000257, 0.00000838, 0.00001368, 0.00001816, 0.00002155, 0.00002365, 0.00002436, 0.00002369, 0.00002173, 0.00001869, 0.00001483, 0.00001049, 0.00000601, 0.00000175, -0.00000197, -0.00000491, -0.00000687, -0.00000778, -0.00000763, -0.00000652, -0.00000465, -0.00000225, 0.00000040, 0.00000300, 0.00000529, 0.00000702, 0.00000802, 0.00000820, 0.00000754, 0.00000610, 0.00000402, 0.00000149, -0.00000127, -0.00000400, -0.00000648, -0.00000850, -0.00000991, -0.00001061, -0.00001057, -0.00000982, -0.00000848, -0.00000667, -0.00000458, -0.00000240, -0.00000031, 0.00000151, 0.00000295, 0.00000391, 0.00000436, 0.00000431, 0.00000383, 0.00000301, 0.00000196, 0.00000083, -0.00000025, -0.00000119, -0.00000187, -0.00000225, -0.00000231, -0.00000206, -0.00000154, -0.00000082, },
		{ -0.00000321, -0.00000412, -0.00000484, -0.00000532, -0.00000551, -0.00000541, -0.00000505, -0.00000447, -0.00000373, -0.00000290, -0.00000207, -0.00000130, -0.00000067, -0.00000023, 0.00000000, 0.00000000, -0.00000022, -0.00000064, -0.00000123, -0.00000194, -0.00000270, -0.00000345, -0.00000414, -0.00000469, -0.00000506, -0.00000520, -0.00000508, -0.00000466, -0.00000396, -0.00000298, -0.00000176, -0.00000034, 0.00000119, 0.00000277, 0.00000429, 0.00000566, 0.00000679, 0.00000758, 0.00000799, 0.00000798, 0.00000754, 0.00000672, 0.00000560, 0.00000428, 0.00000290, 0.00000164, 0.00000064, 0.00000007, 0.00000005, 0.00000068, 0.00000199, 0.00000395, 0.00000648, 0.00000941, 0.00001254, 0.00001562, 0.00001837, 0.00002051, 0.00002179, 0.00002202, 0.00002106, 0.00001885, 0.00001545, 0.00001099, 0.00000570, -0.00000011, -0.00000609, -0.00001186, -0.00001705, -0.00002133, -0.00002440, -0.00002610, -0.00002633, -0.00002513, -0.00002261, -0.00001902, -0.00001466, -0.00000988, -0.00000505, -0.00000054, 0.00000333, 0.00000631, 0.00000825, 0.00000907, 0.00000882, 0.00000764, 0.00000571, 0.00000332, 0.00000075, -0.00000170, -0.00000379, -0.00000528, -0.00000605, -0.00000603, -0.00000525, -0.00000381, -0.00000187, 0.00000035, 0.00000263, 0.00000473, 0.00000644, 0.00000760, 0.00000811, 0.00000791, 0.00000705, 0.00000562, 0.00000375, 0.00000163, -0.00000055, -0.00000260, -0.00000435, -0.00000568, -0.00000649, -0.00000676, -0.00000651, -0.00000582, -0.00000479, -0.00000356, -0.00000228, -0.00000109, -0.00000011, 0.00000056, 0.00000088, 0.00000082, 0.00000041, -0.00000029, -0.00000119, -0.00000220, },
		{ -0.00000107, -0.00000058, -0.00000031, -0.00000029, -0.00000050, -0.00000090, -0.00000144, -0.00000204, -0.00000263, -0.00000313, -0.00000347, -0.00000361, -0.00000352, -0.00000320, -0.00000265, -0.00000191, -0.00000103, -0.00000005, 0.00000095, 0.00000192, 0.00000280, 0.00000356, 0.00000414, 0.00000453, 0.00000471, 0.00000468, 0.00000442, 0.00000398, 0.00000335, 0.00000259, 0.00000172, 0.00000080, -0.00000010, -0.00000094, -0.00000164, -0.00000213, -0.00000237, -0.00000229, -0.00000188, -0.00000115, -0.00000011, 0.00000117, 0.00000259, 0.00000403, 0.00000533, 0.00000635, 0.00000692, 0.00000692, 0.00000622, 0.00000478, 0.00000259, -0.00000028, -0.00000371, -0.00000750, -0.00001140, -0.00001514, -0.00001840, -0.00002091, -0.00002241, -0.00002270, -0.00002166, -0.00001928, -0.00001562, -0.00001087, -0.00000529, 0.00000077, 0.00000694, 0.00001281, 0.00001799, 0.00002215, 0.00002502, 0.00002643, 0.00002633, 0.00002478, 0.00002195, 0.00001810, 0.00001357, 0.00000874, 0.00000400, -0.00000030, -0.00000383, -0.00000639, -0.00000782, -0.00000812, -0.00000736, -0.00000570, -0.00000339, -0.00000071, 0.00000202, 0.00000452, 0.00000653, 0.00000786, 0.00000840, 0.00000812, 0.00000708, 0.00000541, 0.00000329, 0.00000096, -0.00000135, -0.00000342, -0.00000505, -0.00000610, -0.00000651, -0.00000627, -0.00000543, -0.00000412, -0.00000249, -0.00000074, 0.00000096, 0.00000243, 0.00000353, 0.00000417, 0.00000430, 0.00000394, 0.00000314, 0.00000201, 0.00000067, -0.00000073, -0.00000205, -0.00000317, -0.00000401, -0.00000450, -0.00000462, -0.00000442, -0.00000394, -0.00000326, -0.00000249, -0.00000173, },
		{ -0.00000056, -0.00000111, -0.00000150, -0.00000172, -0.00000180, -0.00000180, -0.00000178, -0.00000178, -0.00000187, -0.00000209, -0.00000246, -0.00000297, -0.00000361, -0.00000433, -0.00000508, -0.00000580, -0.00000643, -0.00000692, -0.00000721, -0.00000729, -0.00000715, -0.00000678, -0.00000621, -0.00000547, -0.00000460, -0.00000365, -0.00000265, -0.00000164, -0.00000066, 0.00000028, 0.00000113, 0.00000190, 0.00000254, 0.00000306, 0.00000341, 0.00000359, 0.00000357, 0.00000334, 0.00000290, 0.00000227, 0.00000148, 0.00000058, -0.00000033, -0.00000116, -0.00000179, -0.00000210, -0.00000196, -0.00000128, 0.00000001, 0.00000195, 0.00000449, 0.00000756, 0.00001100, 0.00001462, 0.00001816, 0.00002134, 0.00002390, 0.00002557, 0.00002612, 0.00002540, 0.00002334, 0.00001995, 0.00001537, 0.00000981, 0.00000357, -0.00000299, -0.00000947, -0.00001546, -0.00002059, -0.00002455, -0.00002710, -0.00002812, -0.00002760, -0.00002564, -0.00002247, -0.00001839, -0.00001375, -0.00000895, -0.00000437, -0.00000037, 0.00000277, 0.00000486, 0.00000580, 0.00000563, 0.00000446, 0.00000250, 0.00000001, -0.00000269, -0.00000530, -0.00000753, -0.00000915, -0.00001001, -0.00001002, -0.00000920, -0.00000765, -0.00000553, -0.00000305, -0.00000047, 0.00000197, 0.00000406, 0.00000562, 0.00000654, 0.00000677, 0.00000633, 0.00000533, 0.00000389, 0.00000219, 0.00000044, -0.00000120, -0.00000255, -0.00000350, -0.00000397, -0.00000394, -0.00000346, -0.00000260, -0.00000147, -0.00000021, 0.00000104, 0.00000215, 0.00000302, 0.00000357, 0.00000379, 0.00000366, 0.00000324, 0.00000259, 0.00000180, 0.00000096, 0.00000015, },
		{ 0.00000126, 0.00000184, 0.00000224, 0.00000248, 0.00000259, 0.00000260, 0.00000256, 0.00000251, 0.00000247, 0.00000246, 0.00000248, 0.00000251, 0.00000250, 0.00000243, 0.00000224, 0.00000190, 0.00000139, 0.00000070, -0.00000017, -0.00000117, -0.00000225, -0.00000336, -0.00000442, -0.00000537, -0.00000614, -0.00000670, -0.00000701, -0.00000706, -0.00000688, -0.00000650, -0.00000594, -0.00000528, -0.00000455, -0.00000379, -0.00000304, -0.00000232, -0.00000163, -0.00000097, -0.00000033, 0.00000029, 0.00000089, 0.00000146, 0.00000196, 0.00000233, 0.00000252, 0.00000243, 0.00000199, 0.00000113, -0.00000020, -0.00000201, -0.00000428, -0.00000692, -0.00000982, -0.00001278, -0.00001559, -0.00001803, -0.00001985, -0.00002083, -0.00002080, -0.00001963, -0.00001729, -0.00001383, -0.00000939, -0.00000419, 0.00000147, 0.00000724, 0.00001275, 0.00001764, 0.00002157, 0.00002429, 0.00002561, 0.00002548, 0.00002393, 0.00002112, 0.00001729, 0.00001277, 0.00000792, 0.00000313, -0.00000124, -0.00000488, -0.00000755, -0.00000911, -0.00000953, -0.00000887, -0.00000730, -0.00000507, -0.00000246, 0.00000022, 0.00000266, 0.00000461, 0.00000587, 0.00000631, 0.00000592, 0.00000474, 0.00000292, 0.00000065, -0.00000183, -0.00000427, -0.00000642, -0.00000809, -0.00000914, -0.00000947, -0.00000909, -0.00000806, -0.00000650, -0.00000459, -0.00000252, -0.00000050, 0.00000130, 0.00000272, 0.00000365, 0.00000404, 0.00000391, 0.00000331, 0.00000236, 0.00000117, -0.00000011, -0.00000133, -0.00000237, -0.00000313, -0.00000355, -0.00000363, -0.00000336, -0.00000282, -0.00000207, -0.00000120, -0.00000031, 0.00000053, },
		{ 0.00000101, 0.00000043, 0.00000004, -0.00000016, -0.00000020, -0.00000013, 0.00000003, 0.00000022, 0.00000042, 0.00000062, 0.00000082, 0.00000101, 0.00000123, 0.00000149, 0.00000180, 0.00000218, 0.00000260, 0.00000305, 0.00000349, 0.00000385, 0.00000408, 0.00000411, 0.00000389, 0.00000338, 0.00000257, 0.00000146, 0.00000011, -0.00000143, -0.00000306, -0.00000469, -0.00000620, -0.00000750, -0.00000851, -0.00000917, -0.00000945, -0.00000937, -0.00000896, -0.00000828, -0.00000741, -0.00000643, -0.00000542, -0.00000443, -0.00000351, -0.00000265, -0.00000182, -0.00000099, -0.00000007, 0.00000101, 0.00000232, 0.00000390, 0.00000577, 0.00000790, 0.00001020, 0.00001256, 0.00001481, 0.00001676, 0.00001822, 0.00001898, 0.00001890, 0.00001787, 0.00001586, 0.00001289, 0.00000910, 0.00000468, -0.00000011, -0.00000496, -0.00000955, -0.00001357, -0.00001673, -0.00001882, -0.00001970, -0.00001934, -0.00001780, -0.00001525, -0.00001191, -0.00000811, -0.00000417, -0.00000043, 0.00000280, 0.00000527, 0.00000679, 0.00000730, 0.00000681, 0.00000543, 0.00000336, 0.00000084, -0.00000182, -0.00000435, -0.00000646, -0.00000796, -0.00000868, -0.00000856, -0.00000763, -0.00000600, -0.00000384, -0.00000136, 0.00000117, 0.00000353, 0.00000548, 0.00000685, 0.00000754, 0.00000751, 0.00000680, 0.00000552, 0.00000382, 0.00000189, -0.00000005, -0.00000180, -0.00000319, -0.00000409, -0.00000444, -0.00000420, -0.00000344, -0.00000224, -0.00000074, 0.00000091, 0.00000255, 0.00000403, 0.00000524, 0.00000607, 0.00000650, 0.00000651, 0.00000616, 0.00000550, 0.00000463, 0.00000366, 0.00000268, 0.00000178, },
		{ 0.00000118, 0.00000178, 0.00000210, 0.00000218, 0.00000206, 0.00000178, 0.00000140, 0.00000098, 0.00000055, 0.00000015, -0.00000021, -0.00000054, -0.00000083, -0.00000109, -0.00000134, -0.00000156, -0.00000174, -0.00000187, -0.00000191, -0.00000182, -0.00000157, -0.00000112, -0.00000046, 0.00000040, 0.00000143, 0.00000257, 0.00000374, 0.00000484, 0.00000576, 0.00000639, 0.00000664, 0.00000643, 0.00000573, 0.00000455, 0.00000294, 0.00000098, -0.00000119, -0.00000344, -0.00000560, -0.00000754, -0.00000911, -0.00001025, -0.00001090, -0.00001107, -0.00001083, -0.00001027, -0.00000954, -0.00000878, -0.00000816, -0.00000779, -0.00000777, -0.00000814, -0.00000888, -0.00000991, -0.00001110, -0.00001227, -0.00001322, -0.00001375, -0.00001367, -0.00001285, -0.00001121, -0.00000874, -0.00000550, -0.00000166, 0.00000258, 0.00000695, 0.00001116, 0.00001494, 0.00001802, 0.00002020, 0.00002135, 0.00002142, 0.00002047, 0.00001863, 0.00001610, 0.00001315, 0.00001007, 0.00000713, 0.00000461, 0.00000270, 0.00000155, 0.00000119, 0.00000160, 0.00000266, 0.00000418, 0.00000593, 0.00000766, 0.00000913, 0.00001013, 0.00001050, 0.00001016, 0.00000909, 0.00000737, 0.00000514, 0.00000258, -0.00000008, -0.00000260, -0.00000477, -0.00000640, -0.00000737, -0.00000764, -0.00000720, -0.00000615, -0.00000463, -0.00000283, -0.00000094, 0.00000082, 0.00000226, 0.00000325, 0.00000369, 0.00000355, 0.00000285, 0.00000169, 0.00000018, -0.00000150, -0.00000320, -0.00000475, -0.00000601, -0.00000687, -0.00000727, -0.00000720, -0.00000669, -0.00000581, -0.00000466, -0.00000336, -0.00000203, -0.00000077, 0.00000032, },
		{ -0.00000623, -0.00000619, -0.00000583, -0.00000519, -0.00000434, -0.00000337, -0.00000235, -0.00000135, -0.00000042, 0.00000039, 0.00000106, 0.00000160, 0.00000202, 0.00000233, 0.00000254, 0.00000267, 0.00000272, 0.00000270, 0.00000258, 0.00000235, 0.00000201, 0.00000154, 0.00000096, 0.00000028, -0.00000044, -0.00000115, -0.00000176, -0.00000219, -0.00000236, -0.00000220, -0.00000166, -0.00000073, 0.00000057, 0.00000217, 0.00000396, 0.00000582, 0.00000757, 0.00000905, 0.00001010, 0.00001061, 0.00001047, 0.00000967, 0.00000823, 0.00000625, 0.00000389, 0.00000134, -0.00000117, -0.00000340, -0.00000514, -0.00000622, -0.00000652, -0.00000599, -0.00000467, -0.00000268, -0.00000021, 0.00000249, 0.00000515, 0.00000749, 0.00000924, 0.00001020, 0.00001021, 0.00000923, 0.00000727, 0.00000446, 0.00000100, -0.00000284, -0.00000677, -0.00001048, -0.00001369, -0.00001617, -0.00001774, -0.00001833, -0.00001795, -0.00001668, -0.00001470, -0.00001224, -0.00000955, -0.00000689, -0.00000452, -0.00000261, -0.00000131, -0.00000067, -0.00000067, -0.00000122, -0.00000216, -0.00000332, -0.00000448, -0.00000545, -0.00000606, -0.00000620, -0.00000582, -0.00000492, -0.00000359, -0.00000196, -0.00000021, 0.00000148, 0.00000291, 0.00000392, 0.00000439, 0.00000424, 0.00000349, 0.00000219, 0.00000046, -0.00000153, -0.00000360, -0.00000555, -0.00000719, -0.00000839, -0.00000903, -0.00000907, -0.00000854, -0.00000750, -0.00000607, -0.00000442, -0.00000270, -0.00000110, 0.00000024, 0.00000120, 0.00000170, 0.00000172, 0.00000128, 0.00000047, -0.00000063, -0.00000188, -0.00000315, -0.00000432, -0.00000526, -0.00000592, },
		{ 0.00000228, 0.00000192, 0.00000127, 0.00000042, -0.00000056, -0.00000154, -0.00000245, -0.00000321, -0.00000377, -0.00000411, -0.00000421, -0.00000411, -0.00000383, -0.00000343, -0.00000293, -0.00000239, -0.00000182, -0.00000125, -0.00000069, -0.00000013, 0.00000043, 0.00000100, 0.00000155, 0.00000208, 0.00000255, 0.00000292, 0.00000315, 0.00000316, 0.00000293, 0.00000241, 0.00000159, 0.00000049, -0.00000084, -0.00000230, -0.00000379, -0.00000518, -0.00000632, -0.00000708, -0.00000736, -0.00000708, -0.00000623, -0.00000485, -0.00000304, -0.00000096, 0.00000119, 0.00000316, 0.00000472, 0.00000564, 0.00000572, 0.00000485, 0.00000298, 0.00000014, -0.00000353, -0.00000780, -0.00001240, -0.00001699, -0.00002122, -0.00002474, -0.00002725, -0.00002854, -0.00002845, -0.00002696, -0.00002414, -0.00002018, -0.00001535, -0.00000999, -0.00000447, 0.00000084, 0.00000558, 0.00000947, 0.00001230, 0.00001395, 0.00001443, 0.00001382, 0.00001229, 0.00001009, 0.00000749, 0.00000478, 0.00000223, 0.00000007, -0.00000156, -0.00000255, -0.00000290, -0.00000268, -0.00000200, -0.00000103, 0.00000005, 0.00000105, 0.00000183, 0.00000228, 0.00000233, 0.00000198, 0.00000130, 0.00000039, -0.00000062, -0.00000158, -0.00000234, -0.00000278, -0.00000281, -0.00000242, -0.00000161, -0.00000047, 0.00000090, 0.00000233, 0.00000368, 0.00000478, 0.00000551, 0.00000576, 0.00000550, 0.00000472, 0.00000350, 0.00000193, 0.00000016, -0.00000165, -0.00000333, -0.00000473, -0.00000574, -0.00000628, -0.00000632, -0.00000587, -0.00000502, -0.00000386, -0.00000252, -0.00000114, 0.00000013, 0.00000117, 0.00000191, 0.00000228, },
		{ -0.00000581, -0.00000521, -0.00000439, -0.00000344, -0.00000248, -0.00000162, -0.00000094, -0.00000049, -0.00000032, -0.00000040, -0.00000073, -0.00000124, -0.00000186, -0.00000254, -0.00000320, -0.00000379, -0.00000428, -0.00000464, -0.00000487, -0.00000497, -0.00000498, -0.00000491, -0.00000478, -0.00000459, -0.00000435, -0.00000404, -0.00000363, -0.00000310, -0.00000243, -0.00000161, -0.00000062, 0.00000049, 0.00000169, 0.00000290, 0.00000403, 0.00000497, 0.00000562, 0.00000590, 0.00000574, 0.00000512, 0.00000407, 0.00000266, 0.00000101, -0.00000069, -0.00000226, -0.00000348, -0.00000415, -0.00000410, -0.00000321, -0.00000144, 0.00000118, 0.00000454, 0.00000844, 0.00001259, 0.00001668, 0.00002037, 0.00002331, 0.00002520, 0.00002580, 0.00002497, 0.00002265, 0.00001892, 0.00001396, 0.00000807, 0.00000159, -0.00000507, -0.00001147, -0.00001721, -0.00002195, -0.00002541, -0.00002743, -0.00002795, -0.00002703, -0.00002485, -0.00002164, -0.00001774, -0.00001349, -0.00000922, -0.00000526, -0.00000187, 0.00000078, 0.00000259, 0.00000355, 0.00000375, 0.00000331, 0.00000243, 0.00000131, 0.00000016, -0.00000083, -0.00000155, -0.00000189, -0.00000183, -0.00000142, -0.00000072, 0.00000014, 0.00000103, 0.00000183, 0.00000242, 0.00000272, 0.00000271, 0.00000239, 0.00000181, 0.00000107, 0.00000029, -0.00000042, -0.00000094, -0.00000119, -0.00000109, -0.00000065, 0.00000011, 0.00000112, 0.00000227, 0.00000343, 0.00000447, 0.00000525, 0.00000568, 0.00000568, 0.00000523, 0.00000435, 0.00000310, 0.00000158, -0.00000007, -0.00000173, -0.00000326, -0.00000453, -0.00000545, -0.00000597, -0.00000608, },
		{ 0.00000277, 0.00000157, 0.00000029, -0.00000095, -0.00000204, -0.00000289, -0.00000345, -0.00000367, -0.00000358, -0.00000320, -0.00000260, -0.00000186, -0.00000107, -0.00000032, 0.00000034, 0.00000083, 0.00000114, 0.00000126, 0.00000121, 0.00000102, 0.00000075, 0.00000043, 0.00000012, -0.00000015, -0.00000038, -0.00000055, -0.00000069, -0.00000082, -0.00000096, -0.00000115, -0.00000139, -0.00000168, -0.00000200, -0.00000230, -0.00000253, -0.00000263, -0.00000251, -0.00000214, -0.00000146, -0.00000050, 0.00000073, 0.00000214, 0.00000361, 0.00000501, 0.00000616, 0.00000689, 0.00000707, 0.00000658, 0.00000534, 0.00000338, 0.00000075, -0.00000238, -0.00000581, -0.00000928, -0.00001248, -0.00001513, -0.00001696, -0.00001772, -0.00001728, -0.00001558, -0.00001266, -0.00000866, -0.00000383, 0.00000151, 0.00000699, 0.00001223, 0.00001684, 0.00002048, 0.00002290, 0.00002393, 0.00002350, 0.00002169, 0.00001866, 0.00001466, 0.00001002, 0.00000511, 0.00000029, -0.00000408, -0.00000773, -0.00001043, -0.00001208, -0.00001264, -0.00001219, -0.00001088, -0.00000893, -0.00000660, -0.00000413, -0.00000178, 0.00000023, 0.00000177, 0.00000275, 0.00000314, 0.00000300, 0.00000242, 0.00000155, 0.00000054, -0.00000046, -0.00000130, -0.00000191, -0.00000220, -0.00000219, -0.00000190, -0.00000140, -0.00000079, -0.00000018, 0.00000034, 0.00000068, 0.00000079, 0.00000066, 0.00000031, -0.00000022, -0.00000082, -0.00000141, -0.00000189, -0.00000216, -0.00000216, -0.00000186, -0.00000126, -0.00000040, 0.00000065, 0.00000177, 0.00000287, 0.00000382, 0.00000452, 0.00000489, 0.00000489, 0.00000451, 0.00000378, },
		{ 0.00000127, 0.00000218, 0.00000301, 0.00000366, 0.00000406, 0.00000416, 0.00000392, 0.00000337, 0.00000255, 0.00000153, 0.00000040, -0.00000074, -0.00000179, -0.00000267, -0.00000333, -0.00000372, -0.00000384, -0.00000373, -0.00000343, -0.00000300, -0.00000251, -0.00000204, -0.00000164, -0.00000135, -0.00000119, -0.00000115, -0.00000122, -0.00000134, -0.00000146, -0.00000156, -0.00000158, -0.00000150, -0.00000133, -0.00000109, -0.00000082, -0.00000057, -0.00000040, -0.00000035, -0.00000048, -0.00000078, -0.00000124, -0.00000181, -0.00000240, -0.00000290, -0.00000320, -0.00000315, -0.00000267, -0.00000166, -0.00000010, 0.00000199, 0.00000454, 0.00000739, 0.00001036, 0.00001323, 0.00001576, 0.00001770, 0.00001886, 0.00001906, 0.00001823, 0.00001635, 0.00001352, 0.00000988, 0.00000569, 0.00000122, -0.00000319, -0.00000723, -0.00001061, -0.00001308, -0.00001448, -0.00001473, -0.00001387, -0.00001200, -0.00000935, -0.00000619, -0.00000281, 0.00000044, 0.00000328, 0.00000545, 0.00000676, 0.00000710, 0.00000647, 0.00000496, 0.00000271, -0.00000003, -0.00000302, -0.00000597, -0.00000864, -0.00001079, -0.00001226, -0.00001296, -0.00001288, -0.00001208, -0.00001066, -0.00000881, -0.00000671, -0.00000457, -0.00000258, -0.00000088, 0.00000040, 0.00000122, 0.00000158, 0.00000152, 0.00000115, 0.00000056, -0.00000010, -0.00000073, -0.00000121, -0.00000149, -0.00000152, -0.00000133, -0.00000094, -0.00000042, 0.00000013, 0.00000064, 0.00000102, 0.00000123, 0.00000122, 0.00000100, 0.00000061, 0.00000011, -0.00000043, -0.00000092, -0.00000128, -0.00000143, -0.00000134, -0.00000099, -0.00000040, 0.00000038, },
		{ -0.00000016, -0.00000051, -0.00000077, -0.00000088, -0.00000078, -0.00000045, 0.00000010, 0.00000083, 0.00000168, 0.00000258, 0.00000343, 0.00000416, 0.00000467, 0.00000492, 0.00000488, 0.00000455, 0.00000396, 0.00000317, 0.00000226, 0.00000132, 0.00000043, -0.00000033, -0.00000091, -0.00000127, -0.00000140, -0.00000135, -0.00000115, -0.00000086, -0.00000056, -0.00000032, -0.00000018, -0.00000019, -0.00000035, -0.00000064, -0.00000102, -0.00000144, -0.00000183, -0.00000211, -0.00000225, -0.00000221, -0.00000198, -0.00000160, -0.00000113, -0.00000066, -0.00000029, -0.00000013, -0.00000028, -0.00000081, -0.00000175, -0.00000310, -0.00000479, -0.00000672, -0.00000873, -0.00001065, -0.00001228, -0.00001343, -0.00001395, -0.00001371, -0.00001265, -0.00001078, -0.00000819, -0.00000501, -0.00000145, 0.00000225, 0.00000584, 0.00000906, 0.00001169, 0.00001356, 0.00001457, 0.00001469, 0.00001397, 0.00001253, 0.00001057, 0.00000831, 0.00000600, 0.00000388, 0.00000216, 0.00000100, 0.00000049, 0.00000066, 0.00000145, 0.00000274, 0.00000434, 0.00000605, 0.00000765, 0.00000891, 0.00000967, 0.00000979, 0.00000921, 0.00000795, 0.00000607, 0.00000372, 0.00000107, -0.00000167, -0.00000428, -0.00000657, -0.00000838, -0.00000960, -0.00001018, -0.00001011, -0.00000948, -0.00000839, -0.00000698, -0.00000543, -0.00000388, -0.00000249, -0.00000137, -0.00000059, -0.00000017, -0.00000009, -0.00000030, -0.00000071, -0.00000122, -0.00000173, -0.00000215, -0.00000240, -0.00000245, -0.00000228, -0.00000192, -0.00000143, -0.00000087, -0.00000031, 0.00000018, 0.00000052, 0.00000070, 0.00000068, 0.00000050, 0.00000020, },
		{ 0.00000023, 0.00000068, 0.00000108, 0.00000137, 0.00000151, 0.00000147, 0.00000127, 0.00000093, 0.00000052, 0.00000010, -0.00000027, -0.00000050, -0.00000056, -0.00000040, -0.00000004, 0.00000051, 0.00000120, 0.00000196, 0.00000270, 0.00000334, 0.00000380, 0.00000404, 0.00000400, 0.00000370, 0.00000315, 0.00000242, 0.00000157, 0.00000070, -0.00000011, -0.00000079, -0.00000125, -0.00000149, -0.00000149, -0.00000128, -0.00000092, -0.00000047, -0.00000003, 0.00000033, 0.00000054, 0.00000056, 0.00000037, -0.00000001, -0.00000052, -0.00000107, -0.00000157, -0.00000190, -0.00000196, -0.00000167, -0.00000099, 0.00000010, 0.00000154, 0.00000327, 0.00000514, 0.00000701, 0.00000870, 0.00001004, 0.00001089, 0.00001113, 0.00001071, 0.00000962, 0.00000791, 0.00000569, 0.00000313, 0.00000041, -0.00000225, -0.00000466, -0.00000664, -0.00000805, -0.00000882, -0.00000892, -0.00000841, -0.00000737, -0.00000594, -0.00000431, -0.00000266, -0.00000115, 0.00000007, 0.00000090, 0.00000129, 0.00000125, 0.00000084, 0.00000017, -0.00000061, -0.00000137, -0.00000193, -0.00000218, -0.00000201, -0.00000139, -0.00000033, 0.00000111, 0.00000283, 0.00000465, 0.00000643, 0.00000799, 0.00000917, 0.00000985, 0.00000995, 0.00000945, 0.00000838, 0.00000684, 0.00000494, 0.00000285, 0.00000074, -0.00000121, -0.00000286, -0.00000409, -0.00000483, -0.00000507, -0.00000483, -0.00000418, -0.00000323, -0.00000211, -0.00000095, 0.00000012, 0.00000100, 0.00000161, 0.00000192, 0.00000194, 0.00000171, 0.00000128, 0.00000074, 0.00000019, -0.00000029, -0.00000064, -0.00000080, -0.00000078, -0.00000056, -0.00000021, },
		{ -0.00000215, -0.00000271, -0.00000319, -0.00000353, -0.00000368, -0.00000362, -0.00000336, -0.00000293, -0.00000240, -0.00000182, -0.00000127, -0.00000081, -0.00000050, -0.00000035, -0.00000038, -0.00000055, -0.00000081, -0.00000112, -0.00000139, -0.00000155, -0.00000156, -0.00000136, -0.00000096, -0.00000036, 0.00000039, 0.00000122, 0.00000205, 0.00000280, 0.00000337, 0.00000371, 0.00000377, 0.00000353, 0.00000302, 0.00000229, 0.00000141, 0.00000048, -0.00000040, -0.00000113, -0.00000165, -0.00000191, -0.00000189, -0.00000162, -0.00000116, -0.00000060, -0.00000005, 0.00000037, 0.00000055, 0.00000041, -0.00000012, -0.00000105, -0.00000234, -0.00000391, -0.00000566, -0.00000742, -0.00000905, -0.00001036, -0.00001122, -0.00001152, -0.00001118, -0.00001019, -0.00000859, -0.00000648, -0.00000398, -0.00000126, 0.00000148, 0.00000408, 0.00000637, 0.00000822, 0.00000955, 0.00001033, 0.00001057, 0.00001034, 0.00000974, 0.00000891, 0.00000797, 0.00000706, 0.00000628, 0.00000571, 0.00000540, 0.00000533, 0.00000546, 0.00000572, 0.00000602, 0.00000627, 0.00000635, 0.00000621, 0.00000581, 0.00000513, 0.00000422, 0.00000314, 0.00000199, 0.00000089, -0.00000005, -0.00000071, -0.00000102, -0.00000093, -0.00000043, 0.00000044, 0.00000161, 0.00000296, 0.00000435, 0.00000564, 0.00000669, 0.00000738, 0.00000763, 0.00000740, 0.00000669, 0.00000555, 0.00000407, 0.00000237, 0.00000059, -0.00000112, -0.00000264, -0.00000385, -0.00000469, -0.00000512, -0.00000515, -0.00000483, -0.00000424, -0.00000349, -0.00000267, -0.00000191, -0.00000129, -0.00000089, -0.00000073, -0.00000082, -0.00000113, -0.00000160, },
		{ -0.00000140, -0.00000082, -0.00000038, -0.00000016, -0.00000018, -0.00000044, -0.00000091, -0.00000154, -0.00000225, -0.00000296, -0.00000358, -0.00000404, -0.00000429, -0.00000431, -0.00000410, -0.00000368, -0.00000313, -0.00000249, -0.00000186, -0.00000130, -0.00000087, -0.00000061, -0.00000054, -0.00000064, -0.00000086, -0.00000115, -0.00000144, -0.00000164, -0.00000169, -0.00000154, -0.00000116, -0.00000055, 0.00000025, 0.00000118, 0.00000215, 0.00000307, 0.00000385, 0.00000439, 0.00000463, 0.00000455, 0.00000415, 0.00000347, 0.00000258, 0.00000161, 0.00000066, -0.00000014, -0.00000066, -0.00000083, -0.00000057, 0.00000011, 0.00000119, 0.00000258, 0.00000417, 0.00000581, 0.00000732, 0.00000857, 0.00000939, 0.00000968, 0.00000937, 0.00000845, 0.00000694, 0.00000493, 0.00000256, -0.00000004, -0.00000267, -0.00000518, -0.00000741, -0.00000924, -0.00001057, -0.00001138, -0.00001166, -0.00001147, -0.00001089, -0.00001003, -0.00000900, -0.00000792, -0.00000689, -0.00000598, -0.00000524, -0.00000467, -0.00000426, -0.00000398, -0.00000376, -0.00000353, -0.00000325, -0.00000287, -0.00000237, -0.00000175, -0.00000104, -0.00000029, 0.00000043, 0.00000104, 0.00000149, 0.00000170, 0.00000166, 0.00000134, 0.00000079, 0.00000004, -0.00000082, -0.00000168, -0.00000246, -0.00000305, -0.00000336, -0.00000335, -0.00000298, -0.00000228, -0.00000130, -0.00000011, 0.00000117, 0.00000242, 0.00000354, 0.00000440, 0.00000492, 0.00000506, 0.00000480, 0.00000417, 0.00000323, 0.00000207, 0.00000080, -0.00000046, -0.00000159, -0.00000251, -0.00000314, -0.00000346, -0.00000345, -0.00000317, -0.00000267, -0.00000205, },
		{ 0.00000019, -0.00000047, -0.00000090, -0.00000105, -0.00000095, -0.00000061, -0.00000011, 0.00000048, 0.00000105, 0.00000153, 0.00000182, 0.00000187, 0.00000164, 0.00000114, 0.00000040, -0.00000052, -0.00000154, -0.00000257, -0.00000352, -0.00000429, -0.00000484, -0.00000511, -0.00000511, -0.00000485, -0.00000440, -0.00000382, -0.00000320, -0.00000262, -0.00000217, -0.00000190, -0.00000183, -0.00000198, -0.00000231, -0.00000276, -0.00000325, -0.00000371, -0.00000405, -0.00000418, -0.00000408, -0.00000371, -0.00000309, -0.00000227, -0.00000135, -0.00000042, 0.00000040, 0.00000099, 0.00000124, 0.00000109, 0.00000048, -0.00000057, -0.00000201, -0.00000376, -0.00000569, -0.00000764, -0.00000945, -0.00001096, -0.00001204, -0.00001257, -0.00001251, -0.00001183, -0.00001057, -0.00000882, -0.00000670, -0.00000436, -0.00000196, 0.00000034, 0.00000239, 0.00000408, 0.00000532, 0.00000608, 0.00000637, 0.00000622, 0.00000571, 0.00000494, 0.00000401, 0.00000302, 0.00000207, 0.00000123, 0.00000054, 0.00000001, -0.00000035, -0.00000057, -0.00000070, -0.00000078, -0.00000086, -0.00000097, -0.00000113, -0.00000134, -0.00000157, -0.00000181, -0.00000201, -0.00000212, -0.00000211, -0.00000194, -0.00000160, -0.00000111, -0.00000049, 0.00000020, 0.00000091, 0.00000157, 0.00000210, 0.00000246, 0.00000259, 0.00000248, 0.00000213, 0.00000159, 0.00000092, 0.00000018, -0.00000052, -0.00000111, -0.00000151, -0.00000166, -0.00000152, -0.00000109, -0.00000041, 0.00000047, 0.00000147, 0.00000249, 0.00000343, 0.00000420, 0.00000471, 0.00000493, 0.00000483, 0.00000443, 0.00000377, 0.00000292, 0.00000197, 0.00000103, },
		{ 0.00000334, 0.00000344, 0.00000329, 0.00000289, 0.00000228, 0.00000154, 0.00000073, -0.00000005, -0.00000071, -0.00000118, -0.00000142, -0.00000138, -0.00000109, -0.00000057, 0.00000010, 0.00000084, 0.00000156, 0.00000215, 0.00000252, 0.00000262, 0.00000240, 0.00000185, 0.00000101, -0.00000007, -0.00000129, -0.00000254, -0.00000373, -0.00000475, -0.00000551, -0.00000595, -0.00000606, -0.00000583, -0.00000531, -0.00000458, -0.00000373, -0.00000286, -0.00000206, -0.00000142, -0.00000100, -0.00000082, -0.00000087, -0.00000112, -0.00000148, -0.00000187, -0.00000216, -0.00000227, -0.00000210, -0.00000158, -0.00000070, 0.00000055, 0.00000210, 0.00000386, 0.00000571, 0.00000751, 0.00000911, 0.00001036, 0.00001116, 0.00001142, 0.00001109, 0.00001017, 0.00000872, 0.00000681, 0.00000459, 0.00000218, -0.00000025, -0.00000255, -0.00000460, -0.00000629, -0.00000754, -0.00000833, -0.00000866, -0.00000857, -0.00000814, -0.00000744, -0.00000658, -0.00000564, -0.00000471, -0.00000386, -0.00000311, -0.00000250, -0.00000202, -0.00000165, -0.00000136, -0.00000112, -0.00000090, -0.00000067, -0.00000042, -0.00000015, 0.00000013, 0.00000039, 0.00000061, 0.00000074, 0.00000077, 0.00000068, 0.00000048, 0.00000016, -0.00000023, -0.00000065, -0.00000107, -0.00000141, -0.00000164, -0.00000172, -0.00000163, -0.00000136, -0.00000093, -0.00000038, 0.00000023, 0.00000085, 0.00000141, 0.00000185, 0.00000213, 0.00000221, 0.00000209, 0.00000178, 0.00000133, 0.00000079, 0.00000023, -0.00000028, -0.00000067, -0.00000088, -0.00000088, -0.00000066, -0.00000023, 0.00000036, 0.00000106, 0.00000178, 0.00000245, 0.00000299, },
		{ -0.00000047, -0.00000014, 0.00000035, 0.00000096, 0.00000163, 0.00000231, 0.00000293, 0.00000341, 0.00000371, 0.00000378, 0.00000361, 0.00000322, 0.00000266, 0.00000197, 0.00000124, 0.00000056, 0.00000002, -0.00000031, -0.00000038, -0.00000015, 0.00000036, 0.00000111, 0.00000204, 0.00000305, 0.00000404, 0.00000489, 0.00000550, 0.00000578, 0.00000567, 0.00000517, 0.00000427, 0.00000304, 0.00000157, -0.00000004, -0.00000165, -0.00000314, -0.00000441, -0.00000535, -0.00000591, -0.00000608, -0.00000588, -0.00000537, -0.00000463, -0.00000380, -0.00000298, -0.00000230, -0.00000185, -0.00000172, -0.00000194, -0.00000250, -0.00000337, -0.00000445, -0.00000565, -0.00000684, -0.00000787, -0.00000864, -0.00000903, -0.00000898, -0.00000845, -0.00000744, -0.00000601, -0.00000423, -0.00000222, -0.00000011, 0.00000198, 0.00000390, 0.00000556, 0.00000686, 0.00000776, 0.00000824, 0.00000830, 0.00000800, 0.00000739, 0.00000656, 0.00000559, 0.00000456, 0.00000354, 0.00000260, 0.00000176, 0.00000104, 0.00000045, -0.00000002, -0.00000041, -0.00000072, -0.00000099, -0.00000123, -0.00000144, -0.00000163, -0.00000178, -0.00000187, -0.00000190, -0.00000184, -0.00000168, -0.00000142, -0.00000108, -0.00000067, -0.00000023, 0.00000020, 0.00000056, 0.00000083, 0.00000097, 0.00000095, 0.00000076, 0.00000041, -0.00000007, -0.00000064, -0.00000124, -0.00000183, -0.00000234, -0.00000273, -0.00000296, -0.00000301, -0.00000287, -0.00000258, -0.00000217, -0.00000168, -0.00000117, -0.00000069, -0.00000029, -0.00000002, 0.00000012, 0.00000013, 0.00000003, -0.00000016, -0.00000037, -0.00000056, -0.00000066, -0.00000065, },
		{ -0.00000164, -0.00000156, -0.00000154, -0.00000156, -0.00000158, -0.00000157, -0.00000148, -0.00000129, -0.00000098, -0.00000053, 0.00000002, 0.00000066, 0.00000133, 0.00000198, 0.00000253, 0.00000292, 0.00000311, 0.00000306, 0.00000276, 0.00000223, 0.00000151, 0.00000067, -0.00000020, -0.00000102, -0.00000168, -0.00000210, -0.00000222, -0.00000199, -0.00000141, -0.00000051, 0.00000063, 0.00000194, 0.00000328, 0.00000453, 0.00000557, 0.00000628, 0.00000658, 0.00000643, 0.00000580, 0.00000475, 0.00000333, 0.00000166, -0.00000011, -0.00000186, -0.00000343, -0.00000468, -0.00000553, -0.00000588, -0.00000573, -0.00000510, -0.00000404, -0.00000266, -0.00000108, 0.00000055, 0.00000208, 0.00000339, 0.00000435, 0.00000490, 0.00000498, 0.00000461, 0.00000380, 0.00000265, 0.00000124, -0.00000031, -0.00000188, -0.00000336, -0.00000466, -0.00000570, -0.00000644, -0.00000684, -0.00000694, -0.00000675, -0.00000633, -0.00000575, -0.00000506, -0.00000433, -0.00000362, -0.00000297, -0.00000240, -0.00000192, -0.00000155, -0.00000126, -0.00000105, -0.00000090, -0.00000079, -0.00000072, -0.00000068, -0.00000067, -0.00000070, -0.00000078, -0.00000092, -0.00000112, -0.00000138, -0.00000169, -0.00000202, -0.00000235, -0.00000264, -0.00000286, -0.00000298, -0.00000296, -0.00000279, -0.00000247, -0.00000200, -0.00000141, -0.00000075, -0.00000006, 0.00000061, 0.00000118, 0.00000162, 0.00000189, 0.00000195, 0.00000181, 0.00000148, 0.00000099, 0.00000037, -0.00000031, -0.00000099, -0.00000163, -0.00000217, -0.00000257, -0.00000282, -0.00000292, -0.00000288, -0.00000273, -0.00000250, -0.00000224, -0.00000200, -0.00000179, },
		{ -0.00000192, -0.00000231, -0.00000259, -0.00000275, -0.00000281, -0.00000281, -0.00000276, -0.00000269, -0.00000261, -0.00000252, -0.00000243, -0.00000232, -0.00000216, -0.00000193, -0.00000161, -0.00000116, -0.00000058, 0.00000012, 0.00000092, 0.00000178, 0.00000265, 0.00000347, 0.00000416, 0.00000466, 0.00000491, 0.00000488, 0.00000455, 0.00000394, 0.00000309, 0.00000206, 0.00000095, -0.00000014, -0.00000110, -0.00000184, -0.00000226, -0.00000230, -0.00000195, -0.00000120, -0.00000012, 0.00000122, 0.00000270, 0.00000418, 0.00000551, 0.00000657, 0.00000722, 0.00000737, 0.00000697, 0.00000601, 0.00000452, 0.00000257, 0.00000029, -0.00000218, -0.00000468, -0.00000705, -0.00000913, -0.00001079, -0.00001193, -0.00001250, -0.00001247, -0.00001186, -0.00001075, -0.00000921, -0.00000739, -0.00000539, -0.00000334, -0.00000138, 0.00000040, 0.00000193, 0.00000315, 0.00000403, 0.00000459, 0.00000484, 0.00000483, 0.00000462, 0.00000425, 0.00000378, 0.00000326, 0.00000273, 0.00000221, 0.00000173, 0.00000129, 0.00000089, 0.00000054, 0.00000023, -0.00000004, -0.00000027, -0.00000046, -0.00000059, -0.00000067, -0.00000069, -0.00000064, -0.00000053, -0.00000038, -0.00000019, -0.00000000, 0.00000017, 0.00000028, 0.00000031, 0.00000024, 0.00000005, -0.00000025, -0.00000067, -0.00000116, -0.00000169, -0.00000223, -0.00000271, -0.00000308, -0.00000332, -0.00000337, -0.00000322, -0.00000287, -0.00000235, -0.00000167, -0.00000090, -0.00000008, 0.00000071, 0.00000141, 0.00000197, 0.00000235, 0.00000251, 0.00000246, 0.00000221, 0.00000177, 0.00000120, 0.00000055, -0.00000014, -0.00000081, -0.00000141, },
		{ 0.00000421, 0.00000418, 0.00000393, 0.00000350, 0.00000291, 0.00000222, 0.00000146, 0.00000068, -0.00000008, -0.00000079, -0.00000143, -0.00000199, -0.00000247, -0.00000288, -0.00000322, -0.00000350, -0.00000372, -0.00000389, -0.00000398, -0.00000398, -0.00000387, -0.00000363, -0.00000324, -0.00000268, -0.00000194, -0.00000105, -0.00000002, 0.00000110, 0.00000224, 0.00000333, 0.00000430, 0.00000508, 0.00000559, 0.00000579, 0.00000565, 0.00000518, 0.00000439, 0.00000336, 0.00000217, 0.00000092, -0.00000026, -0.00000126, -0.00000196, -0.00000229, -0.00000217, -0.00000158, -0.00000053, 0.00000091, 0.00000268, 0.00000463, 0.00000664, 0.00000856, 0.00001024, 0.00001155, 0.00001238, 0.00001265, 0.00001234, 0.00001144, 0.00001001, 0.00000812, 0.00000589, 0.00000344, 0.00000090, -0.00000158, -0.00000388, -0.00000590, -0.00000756, -0.00000882, -0.00000964, -0.00001004, -0.00001006, -0.00000973, -0.00000913, -0.00000832, -0.00000737, -0.00000634, -0.00000528, -0.00000424, -0.00000326, -0.00000234, -0.00000152, -0.00000078, -0.00000015, 0.00000039, 0.00000083, 0.00000116, 0.00000139, 0.00000152, 0.00000155, 0.00000148, 0.00000133, 0.00000111, 0.00000084, 0.00000055, 0.00000028, 0.00000005, -0.00000010, -0.00000016, -0.00000011, 0.00000005, 0.00000031, 0.00000066, 0.00000105, 0.00000145, 0.00000182, 0.00000211, 0.00000228, 0.00000230, 0.00000217, 0.00000187, 0.00000143, 0.00000089, 0.00000028, -0.00000033, -0.00000089, -0.00000134, -0.00000164, -0.00000173, -0.00000162, -0.00000129, -0.00000077, -0.00000009, 0.00000070, 0.00000152, 0.00000233, 0.00000305, 0.00000363, 0.00000402, },
		{ -0.00000267, -0.00000212, -0.00000140, -0.00000057, 0.00000031, 0.00000118, 0.00000197, 0.00000264, 0.00000315, 0.00000346, 0.00000356, 0.00000345, 0.00000316, 0.00000270, 0.00000211, 0.00000143, 0.00000069, -0.00000008, -0.00000086, -0.00000162, -0.00000235, -0.00000303, -0.00000367, -0.00000425, -0.00000476, -0.00000518, -0.00000550, -0.00000569, -0.00000572, -0.00000557, -0.00000522, -0.00000464, -0.00000384, -0.00000283, -0.00000164, -0.00000032, 0.00000106, 0.00000244, 0.00000372, 0.00000482, 0.00000566, 0.00000617, 0.00000631, 0.00000604, 0.00000539, 0.00000439, 0.00000311, 0.00000163, 0.00000006, -0.00000148, -0.00000287, -0.00000400, -0.00000479, -0.00000517, -0.00000511, -0.00000461, -0.00000370, -0.00000244, -0.00000093, 0.00000073, 0.00000243, 0.00000406, 0.00000551, 0.00000669, 0.00000755, 0.00000804, 0.00000816, 0.00000793, 0.00000738, 0.00000658, 0.00000558, 0.00000447, 0.00000331, 0.00000217, 0.00000111, 0.00000015, -0.00000066, -0.00000132, -0.00000182, -0.00000218, -0.00000241, -0.00000251, -0.00000252, -0.00000243, -0.00000227, -0.00000204, -0.00000174, -0.00000139, -0.00000099, -0.00000055, -0.00000009, 0.00000037, 0.00000081, 0.00000119, 0.00000151, 0.00000172, 0.00000181, 0.00000178, 0.00000161, 0.00000132, 0.00000093, 0.00000048, -0.00000001, -0.00000047, -0.00000088, -0.00000118, -0.00000136, -0.00000140, -0.00000129, -0.00000104, -0.00000069, -0.00000027, 0.00000016, 0.00000055, 0.00000086, 0.00000105, 0.00000107, 0.00000092, 0.00000060, 0.00000013, -0.00000045, -0.00000110, -0.00000174, -0.00000232, -0.00000278, -0.00000307, -0.00000316, -0.00000302, },
		{ -0.00000002, -0.00000058, -0.00000119, -0.00000181, -0.00000236, -0.00000279, -0.00000306, -0.00000313, -0.00000298, -0.00000263, -0.00000208, -0.00000138, -0.00000056, 0.00000031, 0.00000118, 0.00000199, 0.00000269, 0.00000325, 0.00000364, 0.00000383, 0.00000383, 0.00000365, 0.00000329, 0.00000279, 0.00000216, 0.00000143, 0.00000063, -0.00000023, -0.00000113, -0.00000204, -0.00000294, -0.00000381, -0.00000463, -0.00000536, -0.00000598, -0.00000645, -0.00000674, -0.00000680, -0.00000661, -0.00000615, -0.00000541, -0.00000440, -0.00000314, -0.00000168, -0.00000008, 0.00000159, 0.00000324, 0.00000479, 0.00000616, 0.00000728, 0.00000807, 0.00000851, 0.00000858, 0.00000829, 0.00000768, 0.00000680, 0.00000573, 0.00000456, 0.00000337, 0.00000224, 0.00000125, 0.00000046, -0.00000010, -0.00000042, -0.00000051, -0.00000040, -0.00000014, 0.00000021, 0.00000059, 0.00000094, 0.00000119, 0.00000132, 0.00000128, 0.00000108, 0.00000072, 0.00000022, -0.00000037, -0.00000103, -0.00000170, -0.00000234, -0.00000292, -0.00000342, -0.00000381, -0.00000411, -0.00000430, -0.00000442, -0.00000447, -0.00000447, -0.00000445, -0.00000442, -0.00000438, -0.00000432, -0.00000424, -0.00000413, -0.00000397, -0.00000374, -0.00000343, -0.00000305, -0.00000259, -0.00000207, -0.00000152, -0.00000096, -0.00000044, 0.00000000, 0.00000034, 0.00000055, 0.00000061, 0.00000051, 0.00000028, -0.00000008, -0.00000051, -0.00000097, -0.00000141, -0.00000179, -0.00000205, -0.00000218, -0.00000216, -0.00000197, -0.00000165, -0.00000122, -0.00000073, -0.00000023, 0.00000022, 0.00000058, 0.00000080, 0.00000086, 0.00000073, 0.00000043, },
		{ -0.00000037, 0.00000011, 0.00000060, 0.00000103, 0.00000136, 0.00000155, 0.00000157, 0.00000142, 0.00000111, 0.00000065, 0.00000010, -0.00000051, -0.00000111, -0.00000165, -0.00000207, -0.00000234, -0.00000242, -0.00000231, -0.00000201, -0.00000152, -0.00000090, -0.00000017, 0.00000062, 0.00000142, 0.00000217, 0.00000284, 0.00000339, 0.00000379, 0.00000402, 0.00000408, 0.00000395, 0.00000364, 0.00000317, 0.00000254, 0.00000176, 0.00000087, -0.00000014, -0.00000122, -0.00000236, -0.00000353, -0.00000469, -0.00000579, -0.00000681, -0.00000769, -0.00000838, -0.00000886, -0.00000907, -0.00000900, -0.00000865, -0.00000800, -0.00000710, -0.00000598, -0.00000469, -0.00000331, -0.00000191, -0.00000057, 0.00000064, 0.00000167, 0.00000245, 0.00000298, 0.00000324, 0.00000326, 0.00000306, 0.00000272, 0.00000230, 0.00000187, 0.00000150, 0.00000126, 0.00000119, 0.00000132, 0.00000166, 0.00000218, 0.00000285, 0.00000361, 0.00000440, 0.00000514, 0.00000576, 0.00000620, 0.00000642, 0.00000638, 0.00000609, 0.00000555, 0.00000481, 0.00000391, 0.00000291, 0.00000188, 0.00000088, -0.00000003, -0.00000082, -0.00000145, -0.00000192, -0.00000223, -0.00000240, -0.00000246, -0.00000246, -0.00000242, -0.00000237, -0.00000234, -0.00000235, -0.00000239, -0.00000245, -0.00000251, -0.00000255, -0.00000253, -0.00000244, -0.00000226, -0.00000197, -0.00000159, -0.00000112, -0.00000060, -0.00000006, 0.00000045, 0.00000089, 0.00000122, 0.00000142, 0.00000146, 0.00000134, 0.00000109, 0.00000072, 0.00000027, -0.00000020, -0.00000064, -0.00000101, -0.00000127, -0.00000138, -0.00000134, -0.00000114, -0.00000080, },
		{ 0.00000026, -0.00000022, -0.00000066, -0.00000102, -0.00000126, -0.00000135, -0.00000130, -0.00000110, -0.00000078, -0.00000037, 0.00000008, 0.00000052, 0.00000091, 0.00000119, 0.00000134, 0.00000134, 0.00000118, 0.00000088, 0.00000045, -0.00000005, -0.00000059, -0.00000111, -0.00000157, -0.00000192, -0.00000212, -0.00000216, -0.00000202, -0.00000171, -0.00000123, -0.00000062, 0.00000009, 0.00000087, 0.00000167, 0.00000246, 0.00000319, 0.00000383, 0.00000436, 0.00000476, 0.00000499, 0.00000505, 0.00000494, 0.00000464, 0.00000417, 0.00000352, 0.00000271, 0.00000176, 0.00000070, -0.00000043, -0.00000159, -0.00000271, -0.00000376, -0.00000465, -0.00000535, -0.00000579, -0.00000593, -0.00000574, -0.00000523, -0.00000441, -0.00000331, -0.00000199, -0.00000054, 0.00000097, 0.00000242, 0.00000374, 0.00000482, 0.00000559, 0.00000602, 0.00000606, 0.00000574, 0.00000508, 0.00000415, 0.00000304, 0.00000185, 0.00000068, -0.00000037, -0.00000120, -0.00000174, -0.00000195, -0.00000181, -0.00000135, -0.00000060, 0.00000036, 0.00000145, 0.00000257, 0.00000363, 0.00000453, 0.00000521, 0.00000561, 0.00000571, 0.00000551, 0.00000505, 0.00000436, 0.00000352, 0.00000260, 0.00000167, 0.00000080, 0.00000005, -0.00000055, -0.00000097, -0.00000122, -0.00000131, -0.00000128, -0.00000118, -0.00000104, -0.00000090, -0.00000081, -0.00000078, -0.00000082, -0.00000093, -0.00000109, -0.00000126, -0.00000142, -0.00000152, -0.00000154, -0.00000146, -0.00000127, -0.00000096, -0.00000057, -0.00000012, 0.00000035, 0.00000080, 0.00000118, 0.00000145, 0.00000159, 0.00000158, 0.00000141, 0.00000112, 0.00000072, },
		{ 0.00000006, 0.00000049, 0.00000085, 0.00000111, 0.00000124, 0.00000123, 0.00000108, 0.00000082, 0.00000047, 0.00000007, -0.00000033, -0.00000069, -0.00000096, -0.00000111, -0.00000114, -0.00000102, -0.00000078, -0.00000045, -0.00000004, 0.00000038, 0.00000078, 0.00000112, 0.00000136, 0.00000147, 0.00000144, 0.00000128, 0.00000099, 0.00000060, 0.00000014, -0.00000033, -0.00000079, -0.00000119, -0.00000150, -0.00000169, -0.00000174, -0.00000165, -0.00000141, -0.00000104, -0.00000055, 0.00000003, 0.00000067, 0.00000135, 0.00000203, 0.00000269, 0.00000329, 0.00000379, 0.00000417, 0.00000440, 0.00000445, 0.00000429, 0.00000390, 0.00000328, 0.00000241, 0.00000131, 0.00000001, -0.00000147, -0.00000305, -0.00000468, -0.00000626, -0.00000771, -0.00000893, -0.00000983, -0.00001034, -0.00001039, -0.00000996, -0.00000904, -0.00000765, -0.00000586, -0.00000376, -0.00000146, 0.00000091, 0.00000322, 0.00000532, 0.00000710, 0.00000846, 0.00000933, 0.00000968, 0.00000952, 0.00000889, 0.00000788, 0.00000658, 0.00000512, 0.00000362, 0.00000222, 0.00000103, 0.00000011, -0.00000046, -0.00000066, -0.00000053, -0.00000009, 0.00000058, 0.00000138, 0.00000223, 0.00000301, 0.00000366, 0.00000409, 0.00000426, 0.00000417, 0.00000382, 0.00000324, 0.00000250, 0.00000166, 0.00000079, -0.00000002, -0.00000073, -0.00000128, -0.00000165, -0.00000184, -0.00000185, -0.00000172, -0.00000150, -0.00000124, -0.00000097, -0.00000076, -0.00000063, -0.00000059, -0.00000066, -0.00000081, -0.00000102, -0.00000125, -0.00000147, -0.00000162, -0.00000169, -0.00000165, -0.00000148, -0.00000120, -0.00000083, -0.00000040, },
		{ -0.00000109, -0.00000139, -0.00000164, -0.00000179, -0.00000183, -0.00000174, -0.00000154, -0.00000125, -0.00000091, -0.00000054, -0.00000020, 0.00000008, 0.00000026, 0.00000032, 0.00000026, 0.00000008, -0.00000019, -0.00000053, -0.00000090, -0.00000125, -0.00000154, -0.00000175, -0.00000185, -0.00000183, -0.00000169, -0.00000144, -0.00000112, -0.00000076, -0.00000039, -0.00000006, 0.00000021, 0.00000039, 0.00000046, 0.00000042, 0.00000026, 0.00000002, -0.00000030, -0.00000065, -0.00000101, -0.00000135, -0.00000164, -0.00000184, -0.00000195, -0.00000195, -0.00000183, -0.00000159, -0.00000123, -0.00000077, -0.00000020, 0.00000046, 0.00000119, 0.00000197, 0.00000279, 0.00000362, 0.00000441, 0.00000512, 0.00000572, 0.00000615, 0.00000636, 0.00000629, 0.00000592, 0.00000521, 0.00000414, 0.00000274, 0.00000103, -0.00000092, -0.00000304, -0.00000521, -0.00000733, -0.00000927, -0.00001090, -0.00001212, -0.00001284, -0.00001299, -0.00001255, -0.00001153, -0.00000999, -0.00000801, -0.00000572, -0.00000325, -0.00000076, 0.00000160, 0.00000369, 0.00000540, 0.00000665, 0.00000738, 0.00000759, 0.00000732, 0.00000665, 0.00000566, 0.00000449, 0.00000325, 0.00000206, 0.00000103, 0.00000024, -0.00000026, -0.00000046, -0.00000036, -0.00000002, 0.00000049, 0.00000110, 0.00000171, 0.00000224, 0.00000262, 0.00000279, 0.00000273, 0.00000244, 0.00000194, 0.00000129, 0.00000053, -0.00000025, -0.00000098, -0.00000161, -0.00000209, -0.00000237, -0.00000246, -0.00000236, -0.00000210, -0.00000174, -0.00000132, -0.00000091, -0.00000054, -0.00000028, -0.00000013, -0.00000012, -0.00000024, -0.00000047, -0.00000076, },
		{ -0.00000069, -0.00000010, 0.00000039, 0.00000073, 0.00000092, 0.00000093, 0.00000079, 0.00000052, 0.00000018, -0.00000019, -0.00000055, -0.00000084, -0.00000104, -0.00000113, -0.00000109, -0.00000095, -0.00000073, -0.00000045, -0.00000016, 0.00000011, 0.00000031, 0.00000042, 0.00000043, 0.00000034, 0.00000015, -0.00000011, -0.00000041, -0.00000072, -0.00000100, -0.00000122, -0.00000136, -0.00000140, -0.00000134, -0.00000119, -0.00000097, -0.00000070, -0.00000041, -0.00000014, 0.00000008, 0.00000024, 0.00000031, 0.00000029, 0.00000018, -0.00000001, -0.00000027, -0.00000056, -0.00000088, -0.00000121, -0.00000151, -0.00000179, -0.00000202, -0.00000220, -0.00000233, -0.00000240, -0.00000240, -0.00000233, -0.00000219, -0.00000195, -0.00000162, -0.00000117, -0.00000061, 0.00000007, 0.00000085, 0.00000171, 0.00000261, 0.00000350, 0.00000433, 0.00000503, 0.00000552, 0.00000574, 0.00000563, 0.00000516, 0.00000431, 0.00000309, 0.00000155, -0.00000024, -0.00000220, -0.00000420, -0.00000611, -0.00000781, -0.00000918, -0.00001010, -0.00001051, -0.00001037, -0.00000968, -0.00000847, -0.00000682, -0.00000484, -0.00000266, -0.00000043, 0.00000171, 0.00000363, 0.00000521, 0.00000639, 0.00000710, 0.00000735, 0.00000718, 0.00000664, 0.00000583, 0.00000485, 0.00000381, 0.00000282, 0.00000197, 0.00000132, 0.00000090, 0.00000073, 0.00000077, 0.00000098, 0.00000129, 0.00000164, 0.00000193, 0.00000212, 0.00000213, 0.00000196, 0.00000158, 0.00000103, 0.00000034, -0.00000043, -0.00000121, -0.00000194, -0.00000253, -0.00000295, -0.00000316, -0.00000315, -0.00000292, -0.00000251, -0.00000196, -0.00000133, },
		{ -0.00000342, -0.00000390, -0.00000414, -0.00000411, -0.00000382, -0.00000330, -0.00000260, -0.00000178, -0.00000092, -0.00000009, 0.00000065, 0.00000123, 0.00000163, 0.00000182, 0.00000182, 0.00000164, 0.00000133, 0.00000093, 0.00000049, 0.00000008, -0.00000028, -0.00000054, -0.00000069, -0.00000073, -0.00000066, -0.00000052, -0.00000033, -0.00000013, 0.00000004, 0.00000016, 0.00000020, 0.00000016, 0.00000003, -0.00000015, -0.00000038, -0.00000063, -0.00000085, -0.00000104, -0.00000115, -0.00000118, -0.00000113, -0.00000100, -0.00000080, -0.00000056, -0.00000030, -0.00000005, 0.00000017, 0.00000034, 0.00000045, 0.00000050, 0.00000048, 0.00000042, 0.00000032, 0.00000021, 0.00000011, 0.00000002, -0.00000005, -0.00000009, -0.00000011, -0.00000012, -0.00000013, -0.00000015, -0.00000019, -0.00000025, -0.00000033, -0.00000042, -0.00000050, -0.00000054, -0.00000051, -0.00000040, -0.00000017, 0.00000017, 0.00000064, 0.00000120, 0.00000182, 0.00000245, 0.00000303, 0.00000349, 0.00000375, 0.00000376, 0.00000346, 0.00000284, 0.00000189, 0.00000064, -0.00000086, -0.00000251, -0.00000422, -0.00000586, -0.00000733, -0.00000850, -0.00000927, -0.00000959, -0.00000941, -0.00000873, -0.00000759, -0.00000606, -0.00000423, -0.00000224, -0.00000020, 0.00000174, 0.00000348, 0.00000492, 0.00000600, 0.00000668, 0.00000695, 0.00000687, 0.00000649, 0.00000590, 0.00000518, 0.00000442, 0.00000372, 0.00000313, 0.00000270, 0.00000245, 0.00000236, 0.00000240, 0.00000252, 0.00000265, 0.00000272, 0.00000269, 0.00000250, 0.00000212, 0.00000155, 0.00000082, -0.00000004, -0.00000097, -0.00000189, -0.00000273, },
		{ 0.00000227, 0.00000171, 0.00000095, 0.00000000, -0.00000108, -0.00000224, -0.00000339, -0.00000446, -0.00000535, -0.00000600, -0.00000636, -0.00000640, -0.00000612, -0.00000555, -0.00000473, -0.00000373, -0.00000264, -0.00000153, -0.00000049, 0.00000040, 0.00000110, 0.00000158, 0.00000182, 0.00000183, 0.00000165, 0.00000131, 0.00000089, 0.00000042, -0.00000003, -0.00000042, -0.00000072, -0.00000092, -0.00000101, -0.00000102, -0.00000095, -0.00000086, -0.00000076, -0.00000068, -0.00000066, -0.00000070, -0.00000080, -0.00000096, -0.00000116, -0.00000138, -0.00000158, -0.00000174, -0.00000185, -0.00000188, -0.00000184, -0.00000174, -0.00000158, -0.00000139, -0.00000119, -0.00000102, -0.00000089, -0.00000083, -0.00000083, -0.00000091, -0.00000104, -0.00000121, -0.00000140, -0.00000157, -0.00000171, -0.00000178, -0.00000178, -0.00000171, -0.00000156, -0.00000135, -0.00000111, -0.00000086, -0.00000063, -0.00000043, -0.00000028, -0.00000019, -0.00000015, -0.00000014, -0.00000014, -0.00000012, -0.00000004, 0.00000010, 0.00000035, 0.00000069, 0.00000111, 0.00000159, 0.00000209, 0.00000254, 0.00000288, 0.00000306, 0.00000301, 0.00000269, 0.00000207, 0.00000117, -0.00000000, -0.00000139, -0.00000290, -0.00000445, -0.00000594, -0.00000724, -0.00000827, -0.00000894, -0.00000919, -0.00000899, -0.00000835, -0.00000732, -0.00000595, -0.00000434, -0.00000261, -0.00000086, 0.00000079, 0.00000225, 0.00000345, 0.00000432, 0.00000487, 0.00000511, 0.00000507, 0.00000481, 0.00000441, 0.00000395, 0.00000349, 0.00000311, 0.00000283, 0.00000267, 0.00000263, 0.00000267, 0.00000275, 0.00000281, 0.00000278, 0.00000262, },
		},
		{ { 0.00000194, 0.00000182, 0.00000151, 0.00000109, 0.00000069, 0.00000045, 0.00000049, 0.00000088, 0.00000161, 0.00000260, 0.00000372, 0.00000478, 0.00000560, 0.00000602, 0.00000593, 0.00000532, 0.00000424, 0.00000282, 0.00000125, -0.00000027, -0.00000158, -0.00000254, -0.00000310, -0.00000330, -0.00000324, -0.00000305, -0.00000290, -0.00000290, -0.00000314, -0.00000359, -0.00000416, -0.00000470, -0.00000501, -0.00000488, -0.00000418, -0.00000283, -0.00000086, 0.00000156, 0.00000419, 0.00000674, 0.00000888, 0.00001033, 0.00001089, 0.00001050, 0.00000922, 0.00000723, 0.00000482, 0.00000235, 0.00000014, -0.00000154, -0.00000250, -0.00000272, -0.00000226, -0.00000134, -0.00000022, 0.00000081, 0.00000149, 0.00000166, 0.00000125, 0.00000032, -0.00000096, -0.00000236, -0.00000360, -0.00000446, -0.00000476, -0.00000446, -0.00000360, -0.00000236, -0.00000096, 0.00000032, 0.00000125, 0.00000166, 0.00000149, 0.00000081, -0.00000022, -0.00000134, -0.00000226, -0.00000272, -0.00000250, -0.00000154, 0.00000014, 0.00000235, 0.00000482, 0.00000723, 0.00000922, 0.00001050, 0.00001089, 0.00001033, 0.00000888, 0.00000674, 0.00000419, 0.00000156, -0.00000086, -0.00000283, -0.00000418, -0.00000488, -0.00000501, -0.00000470, -0.00000416, -0.00000359, -0.00000314, -0.00000290, -0.00000290, -0.00000305, -0.00000324, -0.00000330, -0.00000310, -0.00000254, -0.00000158, -0.00000027, 0.00000125, 0.00000282, 0.00000424, 0.00000532, 0.00000593, 0.00000602, 0.00000560, 0.00000478, 0.00000372, 0.00000260, 0.00000161, 0.00000088, 0.00000049, 0.00000045, 0.00000069, 0.00000109, 0.00000151, 0.00000182, },
		{ 0.00000096, 0.00000111, 0.00000150, 0.00000208, 0.00000271, 0.00000327, 0.00000362, 0.00000363, 0.00000327, 0.00000254, 0.00000152, 0.00000034, -0.00000084, -0.00000189, -0.00000268, -0.00000316, -0.00000335, -0.00000331, -0.00000317, -0.00000304, -0.00000305, -0.00000326, -0.00000367, -0.00000419, -0.00000470, -0.00000501, -0.00000494, -0.00000434, -0.00000313, -0.00000133, 0.00000092, 0.00000340, 0.00000584, 0.00000792, 0.00000937, 0.00000998, 0.00000967, 0.00000848, 0.00000657, 0.00000422, 0.00000176, -0.00000049, -0.00000223, -0.00000329, -0.00000360, -0.00000322, -0.00000235, -0.00000125, -0.00000019, 0.00000054, 0.00000076, 0.00000039, -0.00000053, -0.00000185, -0.00000332, -0.00000468, -0.00000567, -0.00000610, -0.00000588, -0.00000505, -0.00000375, -0.00000220, -0.00000069, 0.00000052, 0.00000125, 0.00000139, 0.00000097, 0.00000012, -0.00000091, -0.00000184, -0.00000239, -0.00000234, -0.00000157, -0.00000009, 0.00000196, 0.00000432, 0.00000668, 0.00000868, 0.00001002, 0.00001049, 0.00001000, 0.00000859, 0.00000645, 0.00000386, 0.00000112, -0.00000144, -0.00000355, -0.00000505, -0.00000587, -0.00000607, -0.00000580, -0.00000526, -0.00000466, -0.00000417, -0.00000390, -0.00000387, -0.00000401, -0.00000420, -0.00000425, -0.00000401, -0.00000335, -0.00000223, -0.00000069, 0.00000113, 0.00000305, 0.00000481, 0.00000619, 0.00000700, 0.00000716, 0.00000664, 0.00000556, 0.00000409, 0.00000246, 0.00000091, -0.00000035, -0.00000116, -0.00000147, -0.00000129, -0.00000075, 0.00000002, 0.00000082, 0.00000150, 0.00000194, 0.00000209, 0.00000198, 0.00000169, 0.00000133, 0.00000106, },
		{ 0.00000094, 0.00000087, 0.00000054, -0.00000004, -0.00000080, -0.00000162, -0.00000240, -0.00000303, -0.00000345, -0.00000365, -0.00000367, -0.00000359, -0.00000351, -0.00000354, -0.00000372, -0.00000407, -0.00000453, -0.00000498, -0.00000526, -0.00000521, -0.00000469, -0.00000362, -0.00000200, 0.00000006, 0.00000237, 0.00000467, 0.00000667, 0.00000812, 0.00000879, 0.00000860, 0.00000755, 0.00000580, 0.00000358, 0.00000121, -0.00000099, -0.00000275, -0.00000387, -0.00000426, -0.00000397, -0.00000317, -0.00000211, -0.00000105, -0.00000028, 0.00000003, -0.00000022, -0.00000101, -0.00000218, -0.00000352, -0.00000475, -0.00000562, -0.00000593, -0.00000558, -0.00000461, -0.00000316, -0.00000147, 0.00000018, 0.00000150, 0.00000228, 0.00000238, 0.00000181, 0.00000071, -0.00000068, -0.00000205, -0.00000310, -0.00000355, -0.00000324, -0.00000213, -0.00000033, 0.00000192, 0.00000430, 0.00000645, 0.00000804, 0.00000880, 0.00000862, 0.00000748, 0.00000554, 0.00000306, 0.00000035, -0.00000226, -0.00000448, -0.00000611, -0.00000707, -0.00000737, -0.00000716, -0.00000663, -0.00000600, -0.00000544, -0.00000510, -0.00000500, -0.00000507, -0.00000520, -0.00000520, -0.00000489, -0.00000413, -0.00000286, -0.00000112, 0.00000095, 0.00000314, 0.00000517, 0.00000678, 0.00000775, 0.00000794, 0.00000732, 0.00000598, 0.00000413, 0.00000203, -0.00000002, -0.00000175, -0.00000293, -0.00000345, -0.00000331, -0.00000259, -0.00000147, -0.00000019, 0.00000103, 0.00000199, 0.00000257, 0.00000272, 0.00000248, 0.00000196, 0.00000132, 0.00000069, 0.00000022, -0.00000002, -0.00000001, 0.00000019, 0.00000050, 0.00000079, },
		{ -0.00000213, -0.00000250, -0.00000272, -0.00000281, -0.00000281, -0.00000279, -0.00000283, -0.00000298, -0.00000325, -0.00000362, -0.00000398, -0.00000422, -0.00000418, -0.00000374, -0.00000282, -0.00000140, 0.00000043, 0.00000252, 0.00000464, 0.00000653, 0.00000794, 0.00000865, 0.00000857, 0.00000768, 0.00000610, 0.00000403, 0.00000176, -0.00000041, -0.00000220, -0.00000341, -0.00000395, -0.00000384, -0.00000321, -0.00000228, -0.00000132, -0.00000057, -0.00000025, -0.00000043, -0.00000113, -0.00000221, -0.00000347, -0.00000465, -0.00000551, -0.00000586, -0.00000561, -0.00000476, -0.00000346, -0.00000194, -0.00000048, 0.00000065, 0.00000123, 0.00000113, 0.00000037, -0.00000091, -0.00000244, -0.00000392, -0.00000499, -0.00000537, -0.00000487, -0.00000345, -0.00000122, 0.00000157, 0.00000457, 0.00000738, 0.00000960, 0.00001092, 0.00001116, 0.00001030, 0.00000845, 0.00000587, 0.00000290, -0.00000009, -0.00000276, -0.00000487, -0.00000626, -0.00000694, -0.00000701, -0.00000665, -0.00000609, -0.00000553, -0.00000513, -0.00000494, -0.00000494, -0.00000499, -0.00000494, -0.00000458, -0.00000376, -0.00000242, -0.00000058, 0.00000163, 0.00000398, 0.00000618, 0.00000794, 0.00000899, 0.00000917, 0.00000844, 0.00000688, 0.00000469, 0.00000219, -0.00000029, -0.00000241, -0.00000389, -0.00000457, -0.00000440, -0.00000349, -0.00000201, -0.00000026, 0.00000148, 0.00000294, 0.00000391, 0.00000429, 0.00000408, 0.00000338, 0.00000234, 0.00000118, 0.00000009, -0.00000076, -0.00000129, -0.00000148, -0.00000138, -0.00000110, -0.00000075, -0.00000048, -0.00000037, -0.00000046, -0.00000075, -0.00000118, -0.00000167, },
		{ -0.00000248, -0.00000270, -0.00000297, -0.00000326, -0.00000345, -0.00000344, -0.00000310, -0.00000235, -0.00000117, 0.00000040, 0.00000224, 0.00000415, 0.00000590, 0.00000726, 0.00000802, 0.00000807, 0.00000737, 0.00000599, 0.00000412, 0.00000199, -0.00000010, -0.00000190, -0.00000320, -0.00000388, -0.00000393, -0.00000347, -0.00000267, -0.00000177, -0.00000102, -0.00000060, -0.00000064, -0.00000115, -0.00000202, -0.00000309, -0.00000411, -0.00000487, -0.00000516, -0.00000490, -0.00000409, -0.00000283, -0.00000135, 0.00000010, 0.00000126, 0.00000193, 0.00000198, 0.00000141, 0.00000038, -0.00000088, -0.00000205, -0.00000281, -0.00000289, -0.00000212, -0.00000046, 0.00000196, 0.00000486, 0.00000787, 0.00001055, 0.00001252, 0.00001343, 0.00001312, 0.00001156, 0.00000892, 0.00000548, 0.00000164, -0.00000217, -0.00000555, -0.00000820, -0.00000994, -0.00001075, -0.00001074, -0.00001011, -0.00000913, -0.00000807, -0.00000713, -0.00000643, -0.00000596, -0.00000565, -0.00000531, -0.00000476, -0.00000381, -0.00000237, -0.00000043, 0.00000189, 0.00000437, 0.00000671, 0.00000861, 0.00000978, 0.00001001, 0.00000925, 0.00000757, 0.00000517, 0.00000239, -0.00000039, -0.00000278, -0.00000447, -0.00000524, -0.00000502, -0.00000389, -0.00000207, 0.00000013, 0.00000235, 0.00000426, 0.00000558, 0.00000614, 0.00000592, 0.00000499, 0.00000354, 0.00000183, 0.00000011, -0.00000136, -0.00000243, -0.00000301, -0.00000311, -0.00000281, -0.00000227, -0.00000165, -0.00000108, -0.00000070, -0.00000054, -0.00000061, -0.00000085, -0.00000119, -0.00000155, -0.00000185, -0.00000206, -0.00000219, -0.00000227, -0.00000235, },
		{ -0.00000244, -0.00000147, -0.00000015, 0.00000143, 0.00000314, 0.00000477, 0.00000610, 0.00000694, 0.00000716, 0.00000668, 0.00000555, 0.00000391, 0.00000195, -0.00000005, -0.00000186, -0.00000326, -0.00000411, -0.00000437, -0.00000411, -0.00000348, -0.00000268, -0.00000195, -0.00000147, -0.00000137, -0.00000169, -0.00000238, -0.00000327, -0.00000418, -0.00000489, -0.00000522, -0.00000506, -0.00000439, -0.00000332, -0.00000201, -0.00000070, 0.00000036, 0.00000099, 0.00000105, 0.00000054, -0.00000042, -0.00000162, -0.00000277, -0.00000357, -0.00000376, -0.00000319, -0.00000183, 0.00000021, 0.00000268, 0.00000520, 0.00000739, 0.00000886, 0.00000933, 0.00000864, 0.00000681, 0.00000403, 0.00000060, -0.00000304, -0.00000646, -0.00000927, -0.00001117, -0.00001202, -0.00001182, -0.00001076, -0.00000911, -0.00000719, -0.00000534, -0.00000380, -0.00000274, -0.00000216, -0.00000195, -0.00000191, -0.00000179, -0.00000135, -0.00000043, 0.00000103, 0.00000295, 0.00000514, 0.00000730, 0.00000910, 0.00001021, 0.00001040, 0.00000956, 0.00000774, 0.00000515, 0.00000211, -0.00000097, -0.00000365, -0.00000557, -0.00000648, -0.00000627, -0.00000501, -0.00000294, -0.00000039, 0.00000221, 0.00000448, 0.00000606, 0.00000674, 0.00000645, 0.00000527, 0.00000340, 0.00000114, -0.00000118, -0.00000325, -0.00000482, -0.00000575, -0.00000600, -0.00000566, -0.00000488, -0.00000387, -0.00000283, -0.00000194, -0.00000132, -0.00000101, -0.00000100, -0.00000121, -0.00000155, -0.00000193, -0.00000225, -0.00000249, -0.00000264, -0.00000274, -0.00000282, -0.00000293, -0.00000307, -0.00000323, -0.00000334, -0.00000330, -0.00000303, },
		{ 0.00000602, 0.00000637, 0.00000610, 0.00000520, 0.00000376, 0.00000195, -0.00000000, -0.00000186, -0.00000340, -0.00000447, -0.00000498, -0.00000496, -0.00000453, -0.00000385, -0.00000314, -0.00000258, -0.00000233, -0.00000244, -0.00000289, -0.00000358, -0.00000434, -0.00000497, -0.00000531, -0.00000523, -0.00000471, -0.00000379, -0.00000262, -0.00000141, -0.00000037, 0.00000030, 0.00000049, 0.00000016, -0.00000058, -0.00000157, -0.00000254, -0.00000323, -0.00000339, -0.00000285, -0.00000159, 0.00000032, 0.00000263, 0.00000504, 0.00000718, 0.00000870, 0.00000933, 0.00000893, 0.00000751, 0.00000524, 0.00000240, -0.00000061, -0.00000341, -0.00000565, -0.00000708, -0.00000760, -0.00000726, -0.00000627, -0.00000491, -0.00000354, -0.00000246, -0.00000189, -0.00000195, -0.00000258, -0.00000360, -0.00000473, -0.00000564, -0.00000603, -0.00000567, -0.00000448, -0.00000254, -0.00000007, 0.00000258, 0.00000500, 0.00000679, 0.00000765, 0.00000739, 0.00000602, 0.00000372, 0.00000082, -0.00000222, -0.00000494, -0.00000691, -0.00000781, -0.00000751, -0.00000605, -0.00000366, -0.00000072, 0.00000232, 0.00000500, 0.00000691, 0.00000778, 0.00000750, 0.00000615, 0.00000394, 0.00000121, -0.00000163, -0.00000421, -0.00000621, -0.00000742, -0.00000778, -0.00000735, -0.00000630, -0.00000488, -0.00000336, -0.00000197, -0.00000090, -0.00000025, -0.00000002, -0.00000015, -0.00000052, -0.00000101, -0.00000150, -0.00000189, -0.00000216, -0.00000230, -0.00000236, -0.00000239, -0.00000243, -0.00000249, -0.00000254, -0.00000254, -0.00000238, -0.00000199, -0.00000130, -0.00000030, 0.00000098, 0.00000243, 0.00000388, 0.00000514, },
		{ -0.00000116, -0.00000281, -0.00000405, -0.00000478, -0.00000498, -0.00000470, -0.00000410, -0.00000336, -0.00000266, -0.00000219, -0.00000202, -0.00000219, -0.00000263, -0.00000320, -0.00000374, -0.00000409, -0.00000410, -0.00000373, -0.00000300, -0.00000200, -0.00000091, 0.00000008, 0.00000080, 0.00000110, 0.00000095, 0.00000040, -0.00000041, -0.00000125, -0.00000189, -0.00000210, -0.00000172, -0.00000068, 0.00000094, 0.00000296, 0.00000510, 0.00000703, 0.00000842, 0.00000903, 0.00000869, 0.00000740, 0.00000528, 0.00000258, -0.00000035, -0.00000317, -0.00000555, -0.00000727, -0.00000823, -0.00000847, -0.00000816, -0.00000754, -0.00000689, -0.00000648, -0.00000645, -0.00000686, -0.00000761, -0.00000846, -0.00000913, -0.00000927, -0.00000861, -0.00000698, -0.00000434, -0.00000087, 0.00000313, 0.00000720, 0.00001084, 0.00001359, 0.00001507, 0.00001508, 0.00001362, 0.00001092, 0.00000738, 0.00000353, -0.00000005, -0.00000284, -0.00000446, -0.00000469, -0.00000356, -0.00000130, 0.00000168, 0.00000487, 0.00000772, 0.00000975, 0.00001064, 0.00001023, 0.00000857, 0.00000589, 0.00000258, -0.00000090, -0.00000409, -0.00000662, -0.00000820, -0.00000875, -0.00000831, -0.00000707, -0.00000532, -0.00000340, -0.00000159, -0.00000016, 0.00000076, 0.00000114, 0.00000102, 0.00000054, -0.00000012, -0.00000081, -0.00000139, -0.00000179, -0.00000198, -0.00000202, -0.00000196, -0.00000187, -0.00000181, -0.00000179, -0.00000178, -0.00000171, -0.00000151, -0.00000107, -0.00000035, 0.00000065, 0.00000187, 0.00000318, 0.00000442, 0.00000540, 0.00000594, 0.00000593, 0.00000532, 0.00000415, 0.00000254, 0.00000070, },
		{ -0.00000333, -0.00000251, -0.00000181, -0.00000138, -0.00000126, -0.00000146, -0.00000186, -0.00000234, -0.00000273, -0.00000288, -0.00000271, -0.00000220, -0.00000140, -0.00000044, 0.00000050, 0.00000126, 0.00000171, 0.00000176, 0.00000145, 0.00000086, 0.00000017, -0.00000040, -0.00000064, -0.00000038, 0.00000044, 0.00000180, 0.00000356, 0.00000548, 0.00000728, 0.00000864, 0.00000933, 0.00000918, 0.00000816, 0.00000634, 0.00000395, 0.00000127, -0.00000138, -0.00000368, -0.00000543, -0.00000649, -0.00000688, -0.00000670, -0.00000616, -0.00000549, -0.00000490, -0.00000453, -0.00000441, -0.00000446, -0.00000450, -0.00000427, -0.00000352, -0.00000205, 0.00000025, 0.00000330, 0.00000689, 0.00001066, 0.00001413, 0.00001681, 0.00001825, 0.00001814, 0.00001636, 0.00001303, 0.00000847, 0.00000322, -0.00000210, -0.00000681, -0.00001033, -0.00001224, -0.00001234, -0.00001070, -0.00000763, -0.00000366, 0.00000060, 0.00000448, 0.00000741, 0.00000897, 0.00000896, 0.00000743, 0.00000464, 0.00000105, -0.00000282, -0.00000640, -0.00000924, -0.00001100, -0.00001154, -0.00001089, -0.00000928, -0.00000703, -0.00000454, -0.00000218, -0.00000026, 0.00000103, 0.00000163, 0.00000160, 0.00000109, 0.00000032, -0.00000051, -0.00000121, -0.00000166, -0.00000184, -0.00000177, -0.00000154, -0.00000126, -0.00000101, -0.00000085, -0.00000079, -0.00000077, -0.00000071, -0.00000050, -0.00000005, 0.00000069, 0.00000169, 0.00000288, 0.00000411, 0.00000520, 0.00000595, 0.00000622, 0.00000590, 0.00000500, 0.00000359, 0.00000183, -0.00000005, -0.00000183, -0.00000329, -0.00000427, -0.00000470, -0.00000461, -0.00000410, },
		{ -0.00000185, -0.00000231, -0.00000264, -0.00000271, -0.00000245, -0.00000189, -0.00000111, -0.00000023, 0.00000056, 0.00000114, 0.00000139, 0.00000129, 0.00000090, 0.00000034, -0.00000021, -0.00000054, -0.00000050, 0.00000002, 0.00000103, 0.00000244, 0.00000405, 0.00000562, 0.00000687, 0.00000757, 0.00000752, 0.00000668, 0.00000507, 0.00000287, 0.00000031, -0.00000230, -0.00000469, -0.00000663, -0.00000797, -0.00000867, -0.00000882, -0.00000856, -0.00000809, -0.00000759, -0.00000720, -0.00000697, -0.00000685, -0.00000672, -0.00000639, -0.00000566, -0.00000438, -0.00000250, -0.00000008, 0.00000268, 0.00000543, 0.00000779, 0.00000934, 0.00000973, 0.00000876, 0.00000641, 0.00000286, -0.00000149, -0.00000608, -0.00001028, -0.00001347, -0.00001512, -0.00001491, -0.00001278, -0.00000893, -0.00000381, 0.00000195, 0.00000758, 0.00001237, 0.00001570, 0.00001716, 0.00001660, 0.00001414, 0.00001018, 0.00000527, 0.00000010, -0.00000467, -0.00000847, -0.00001091, -0.00001182, -0.00001126, -0.00000950, -0.00000693, -0.00000404, -0.00000129, 0.00000095, 0.00000243, 0.00000306, 0.00000292, 0.00000219, 0.00000112, -0.00000001, -0.00000096, -0.00000159, -0.00000184, -0.00000174, -0.00000140, -0.00000096, -0.00000056, -0.00000030, -0.00000022, -0.00000030, -0.00000043, -0.00000049, -0.00000035, 0.00000008, 0.00000084, 0.00000188, 0.00000308, 0.00000426, 0.00000521, 0.00000574, 0.00000570, 0.00000505, 0.00000382, 0.00000213, 0.00000020, -0.00000173, -0.00000342, -0.00000467, -0.00000535, -0.00000545, -0.00000502, -0.00000423, -0.00000326, -0.00000233, -0.00000161, -0.00000121, -0.00000116, -0.00000142, },
		{ -0.00000179, -0.00000094, -0.00000022, 0.00000025, 0.00000040, 0.00000023, -0.00000018, -0.00000068, -0.00000109, -0.00000124, -0.00000100, -0.00000031, 0.00000077, 0.00000213, 0.00000354, 0.00000477, 0.00000557, 0.00000575, 0.00000521, 0.00000395, 0.00000207, -0.00000021, -0.00000266, -0.00000499, -0.00000696, -0.00000841, -0.00000926, -0.00000954, -0.00000934, -0.00000883, -0.00000818, -0.00000751, -0.00000691, -0.00000637, -0.00000581, -0.00000511, -0.00000411, -0.00000269, -0.00000081, 0.00000148, 0.00000401, 0.00000652, 0.00000868, 0.00001016, 0.00001068, 0.00001008, 0.00000837, 0.00000573, 0.00000252, -0.00000078, -0.00000362, -0.00000551, -0.00000605, -0.00000503, -0.00000248, 0.00000131, 0.00000585, 0.00001049, 0.00001454, 0.00001733, 0.00001836, 0.00001736, 0.00001434, 0.00000957, 0.00000358, -0.00000293, -0.00000922, -0.00001456, -0.00001838, -0.00002035, -0.00002037, -0.00001862, -0.00001550, -0.00001153, -0.00000733, -0.00000345, -0.00000035, 0.00000172, 0.00000266, 0.00000259, 0.00000176, 0.00000050, -0.00000083, -0.00000193, -0.00000261, -0.00000278, -0.00000249, -0.00000188, -0.00000113, -0.00000045, 0.00000002, 0.00000020, 0.00000010, -0.00000018, -0.00000048, -0.00000063, -0.00000049, 0.00000002, 0.00000092, 0.00000209, 0.00000336, 0.00000450, 0.00000528, 0.00000552, 0.00000510, 0.00000401, 0.00000236, 0.00000034, -0.00000179, -0.00000376, -0.00000532, -0.00000629, -0.00000661, -0.00000630, -0.00000550, -0.00000441, -0.00000327, -0.00000229, -0.00000163, -0.00000138, -0.00000152, -0.00000196, -0.00000254, -0.00000309, -0.00000344, -0.00000349, -0.00000319, -0.00000258, },
		{ 0.00000131, 0.00000086, 0.00000056, 0.00000054, 0.00000092, 0.00000171, 0.00000283, 0.00000411, 0.00000533, 0.00000627, 0.00000669, 0.00000647, 0.00000555, 0.00000399, 0.00000193, -0.00000039, -0.00000272, -0.00000482, -0.00000647, -0.00000757, -0.00000808, -0.00000806, -0.00000763, -0.00000695, -0.00000615, -0.00000534, -0.00000456, -0.00000380, -0.00000297, -0.00000197, -0.00000071, 0.00000085, 0.00000269, 0.00000467, 0.00000660, 0.00000821, 0.00000924, 0.00000946, 0.00000874, 0.00000708, 0.00000462, 0.00000167, -0.00000137, -0.00000405, -0.00000595, -0.00000674, -0.00000626, -0.00000455, -0.00000187, 0.00000137, 0.00000462, 0.00000730, 0.00000888, 0.00000901, 0.00000751, 0.00000450, 0.00000029, -0.00000457, -0.00000943, -0.00001362, -0.00001655, -0.00001779, -0.00001717, -0.00001475, -0.00001088, -0.00000607, -0.00000094, 0.00000386, 0.00000777, 0.00001040, 0.00001156, 0.00001130, 0.00000985, 0.00000762, 0.00000505, 0.00000260, 0.00000063, -0.00000062, -0.00000110, -0.00000089, -0.00000020, 0.00000070, 0.00000153, 0.00000206, 0.00000219, 0.00000190, 0.00000130, 0.00000058, -0.00000003, -0.00000033, -0.00000017, 0.00000049, 0.00000156, 0.00000288, 0.00000419, 0.00000522, 0.00000572, 0.00000552, 0.00000458, 0.00000296, 0.00000085, -0.00000147, -0.00000369, -0.00000553, -0.00000674, -0.00000722, -0.00000695, -0.00000606, -0.00000476, -0.00000331, -0.00000198, -0.00000098, -0.00000044, -0.00000039, -0.00000076, -0.00000139, -0.00000209, -0.00000266, -0.00000292, -0.00000280, -0.00000229, -0.00000147, -0.00000048, 0.00000049, 0.00000128, 0.00000176, 0.00000189, 0.00000170, },
		{ 0.00000336, 0.00000451, 0.00000551, 0.00000613, 0.00000618, 0.00000555, 0.00000425, 0.00000238, 0.00000013, -0.00000226, -0.00000451, -0.00000640, -0.00000776, -0.00000851, -0.00000865, -0.00000828, -0.00000752, -0.00000653, -0.00000544, -0.00000436, -0.00000331, -0.00000228, -0.00000121, -0.00000002, 0.00000132, 0.00000281, 0.00000438, 0.00000590, 0.00000717, 0.00000799, 0.00000818, 0.00000763, 0.00000633, 0.00000439, 0.00000206, -0.00000034, -0.00000243, -0.00000386, -0.00000435, -0.00000374, -0.00000206, 0.00000049, 0.00000356, 0.00000669, 0.00000940, 0.00001124, 0.00001190, 0.00001123, 0.00000932, 0.00000642, 0.00000298, -0.00000049, -0.00000346, -0.00000548, -0.00000625, -0.00000570, -0.00000395, -0.00000135, 0.00000163, 0.00000445, 0.00000662, 0.00000772, 0.00000756, 0.00000612, 0.00000362, 0.00000042, -0.00000299, -0.00000612, -0.00000852, -0.00000989, -0.00001011, -0.00000924, -0.00000751, -0.00000529, -0.00000299, -0.00000097, 0.00000047, 0.00000118, 0.00000118, 0.00000061, -0.00000024, -0.00000107, -0.00000159, -0.00000157, -0.00000092, 0.00000030, 0.00000191, 0.00000362, 0.00000511, 0.00000606, 0.00000626, 0.00000559, 0.00000410, 0.00000199, -0.00000047, -0.00000291, -0.00000498, -0.00000640, -0.00000700, -0.00000674, -0.00000574, -0.00000421, -0.00000245, -0.00000077, 0.00000056, 0.00000134, 0.00000152, 0.00000115, 0.00000038, -0.00000056, -0.00000143, -0.00000201, -0.00000215, -0.00000182, -0.00000105, -0.00000000, 0.00000114, 0.00000216, 0.00000289, 0.00000323, 0.00000315, 0.00000274, 0.00000214, 0.00000155, 0.00000117, 0.00000113, 0.00000151, 0.00000230, },
		{ 0.00000266, 0.00000056, -0.00000180, -0.00000414, -0.00000620, -0.00000776, -0.00000868, -0.00000892, -0.00000853, -0.00000763, -0.00000640, -0.00000501, -0.00000360, -0.00000227, -0.00000106, 0.00000004, 0.00000108, 0.00000212, 0.00000318, 0.00000424, 0.00000525, 0.00000608, 0.00000659, 0.00000666, 0.00000618, 0.00000513, 0.00000358, 0.00000171, -0.00000023, -0.00000196, -0.00000317, -0.00000363, -0.00000321, -0.00000190, 0.00000014, 0.00000261, 0.00000516, 0.00000734, 0.00000879, 0.00000921, 0.00000845, 0.00000656, 0.00000372, 0.00000030, -0.00000327, -0.00000656, -0.00000916, -0.00001081, -0.00001141, -0.00001103, -0.00000990, -0.00000837, -0.00000681, -0.00000557, -0.00000491, -0.00000495, -0.00000561, -0.00000669, -0.00000786, -0.00000873, -0.00000896, -0.00000829, -0.00000662, -0.00000400, -0.00000070, 0.00000291, 0.00000639, 0.00000927, 0.00001121, 0.00001197, 0.00001152, 0.00001002, 0.00000778, 0.00000523, 0.00000283, 0.00000099, -0.00000002, -0.00000006, 0.00000080, 0.00000231, 0.00000412, 0.00000580, 0.00000696, 0.00000729, 0.00000662, 0.00000499, 0.00000257, -0.00000030, -0.00000321, -0.00000575, -0.00000757, -0.00000845, -0.00000832, -0.00000728, -0.00000559, -0.00000357, -0.00000159, 0.00000001, 0.00000099, 0.00000125, 0.00000083, -0.00000011, -0.00000132, -0.00000250, -0.00000338, -0.00000376, -0.00000355, -0.00000278, -0.00000160, -0.00000021, 0.00000113, 0.00000219, 0.00000281, 0.00000292, 0.00000256, 0.00000189, 0.00000110, 0.00000042, 0.00000005, 0.00000013, 0.00000070, 0.00000166, 0.00000286, 0.00000406, 0.00000499, 0.00000542, 0.00000520, 0.00000426, },
		{ -0.00000917, -0.00000886, -0.00000793, -0.00000655, -0.00000493, -0.00000329, -0.00000178, -0.00000050, 0.00000053, 0.00000133, 0.00000197, 0.00000251, 0.00000301, 0.00000346, 0.00000384, 0.00000407, 0.00000405, 0.00000371, 0.00000299, 0.00000193, 0.00000063, -0.00000076, -0.00000201, -0.00000289, -0.00000322, -0.00000286, -0.00000180, -0.00000014, 0.00000192, 0.00000405, 0.00000594, 0.00000724, 0.00000770, 0.00000718, 0.00000568, 0.00000335, 0.00000048, -0.00000256, -0.00000540, -0.00000767, -0.00000911, -0.00000958, -0.00000912, -0.00000786, -0.00000608, -0.00000406, -0.00000211, -0.00000045, 0.00000078, 0.00000158, 0.00000207, 0.00000242, 0.00000285, 0.00000356, 0.00000466, 0.00000614, 0.00000786, 0.00000957, 0.00001094, 0.00001165, 0.00001139, 0.00000999, 0.00000746, 0.00000397, -0.00000014, -0.00000440, -0.00000828, -0.00001129, -0.00001303, -0.00001329, -0.00001208, -0.00000962, -0.00000631, -0.00000269, 0.00000067, 0.00000326, 0.00000468, 0.00000477, 0.00000354, 0.00000126, -0.00000168, -0.00000476, -0.00000748, -0.00000940, -0.00001024, -0.00000991, -0.00000851, -0.00000632, -0.00000373, -0.00000119, 0.00000092, 0.00000229, 0.00000278, 0.00000242, 0.00000139, -0.00000001, -0.00000145, -0.00000257, -0.00000313, -0.00000300, -0.00000217, -0.00000080, 0.00000086, 0.00000253, 0.00000391, 0.00000478, 0.00000503, 0.00000467, 0.00000384, 0.00000276, 0.00000170, 0.00000091, 0.00000061, 0.00000086, 0.00000165, 0.00000281, 0.00000411, 0.00000526, 0.00000597, 0.00000604, 0.00000534, 0.00000388, 0.00000179, -0.00000069, -0.00000327, -0.00000564, -0.00000752, -0.00000873, },
		{ 0.00000021, 0.00000126, 0.00000191, 0.00000224, 0.00000236, 0.00000239, 0.00000241, 0.00000246, 0.00000252, 0.00000256, 0.00000248, 0.00000222, 0.00000176, 0.00000111, 0.00000034, -0.00000040, -0.00000096, -0.00000118, -0.00000094, -0.00000020, 0.00000099, 0.00000251, 0.00000414, 0.00000561, 0.00000665, 0.00000704, 0.00000665, 0.00000543, 0.00000350, 0.00000106, -0.00000159, -0.00000412, -0.00000622, -0.00000765, -0.00000827, -0.00000806, -0.00000712, -0.00000566, -0.00000393, -0.00000220, -0.00000070, 0.00000042, 0.00000110, 0.00000138, 0.00000137, 0.00000122, 0.00000106, 0.00000099, 0.00000103, 0.00000112, 0.00000113, 0.00000087, 0.00000018, -0.00000105, -0.00000284, -0.00000508, -0.00000753, -0.00000984, -0.00001160, -0.00001242, -0.00001196, -0.00001006, -0.00000675, -0.00000224, 0.00000304, 0.00000851, 0.00001354, 0.00001752, 0.00001996, 0.00002056, 0.00001927, 0.00001630, 0.00001207, 0.00000718, 0.00000229, -0.00000196, -0.00000506, -0.00000670, -0.00000680, -0.00000553, -0.00000324, -0.00000045, 0.00000232, 0.00000457, 0.00000594, 0.00000626, 0.00000555, 0.00000403, 0.00000204, 0.00000000, -0.00000167, -0.00000268, -0.00000283, -0.00000213, -0.00000073, 0.00000108, 0.00000295, 0.00000452, 0.00000553, 0.00000580, 0.00000532, 0.00000423, 0.00000277, 0.00000126, 0.00000002, -0.00000070, -0.00000074, -0.00000012, 0.00000105, 0.00000251, 0.00000393, 0.00000498, 0.00000539, 0.00000498, 0.00000371, 0.00000170, -0.00000084, -0.00000357, -0.00000616, -0.00000826, -0.00000964, -0.00001015, -0.00000980, -0.00000868, -0.00000701, -0.00000505, -0.00000305, -0.00000125, },
		{ 0.00000147, 0.00000136, 0.00000143, 0.00000161, 0.00000180, 0.00000191, 0.00000187, 0.00000168, 0.00000138, 0.00000106, 0.00000085, 0.00000087, 0.00000119, 0.00000184, 0.00000276, 0.00000383, 0.00000488, 0.00000568, 0.00000606, 0.00000587, 0.00000506, 0.00000367, 0.00000183, -0.00000024, -0.00000227, -0.00000401, -0.00000523, -0.00000579, -0.00000564, -0.00000485, -0.00000356, -0.00000197, -0.00000033, 0.00000118, 0.00000237, 0.00000319, 0.00000362, 0.00000375, 0.00000368, 0.00000353, 0.00000343, 0.00000341, 0.00000349, 0.00000360, 0.00000365, 0.00000355, 0.00000323, 0.00000269, 0.00000200, 0.00000132, 0.00000087, 0.00000088, 0.00000157, 0.00000306, 0.00000536, 0.00000833, 0.00001165, 0.00001491, 0.00001759, 0.00001922, 0.00001936, 0.00001777, 0.00001441, 0.00000946, 0.00000336, -0.00000329, -0.00000978, -0.00001541, -0.00001957, -0.00002185, -0.00002206, -0.00002033, -0.00001701, -0.00001267, -0.00000796, -0.00000356, -0.00000005, 0.00000219, 0.00000297, 0.00000241, 0.00000078, -0.00000144, -0.00000373, -0.00000559, -0.00000664, -0.00000667, -0.00000566, -0.00000380, -0.00000144, 0.00000101, 0.00000312, 0.00000455, 0.00000509, 0.00000470, 0.00000354, 0.00000188, 0.00000009, -0.00000144, -0.00000239, -0.00000256, -0.00000191, -0.00000056, 0.00000120, 0.00000303, 0.00000453, 0.00000537, 0.00000531, 0.00000429, 0.00000237, -0.00000019, -0.00000306, -0.00000584, -0.00000814, -0.00000966, -0.00001022, -0.00000979, -0.00000848, -0.00000653, -0.00000422, -0.00000190, 0.00000015, 0.00000173, 0.00000273, 0.00000315, 0.00000310, 0.00000274, 0.00000224, 0.00000178, },
		{ 0.00000209, 0.00000216, 0.00000200, 0.00000171, 0.00000140, 0.00000119, 0.00000119, 0.00000143, 0.00000188, 0.00000244, 0.00000297, 0.00000331, 0.00000331, 0.00000290, 0.00000203, 0.00000077, -0.00000075, -0.00000234, -0.00000380, -0.00000493, -0.00000559, -0.00000570, -0.00000529, -0.00000443, -0.00000329, -0.00000204, -0.00000089, 0.00000004, 0.00000064, 0.00000089, 0.00000082, 0.00000052, 0.00000010, -0.00000034, -0.00000074, -0.00000104, -0.00000129, -0.00000152, -0.00000180, -0.00000219, -0.00000272, -0.00000334, -0.00000399, -0.00000453, -0.00000482, -0.00000474, -0.00000421, -0.00000325, -0.00000196, -0.00000056, 0.00000066, 0.00000140, 0.00000135, 0.00000030, -0.00000180, -0.00000487, -0.00000862, -0.00001263, -0.00001635, -0.00001924, -0.00002079, -0.00002064, -0.00001867, -0.00001495, -0.00000984, -0.00000388, 0.00000224, 0.00000781, 0.00001220, 0.00001492, 0.00001573, 0.00001466, 0.00001201, 0.00000828, 0.00000410, 0.00000015, -0.00000299, -0.00000490, -0.00000538, -0.00000448, -0.00000248, 0.00000016, 0.00000293, 0.00000528, 0.00000681, 0.00000727, 0.00000662, 0.00000502, 0.00000281, 0.00000044, -0.00000163, -0.00000303, -0.00000349, -0.00000295, -0.00000155, 0.00000042, 0.00000255, 0.00000440, 0.00000555, 0.00000572, 0.00000480, 0.00000286, 0.00000014, -0.00000297, -0.00000602, -0.00000858, -0.00001029, -0.00001092, -0.00001044, -0.00000895, -0.00000672, -0.00000411, -0.00000150, 0.00000075, 0.00000240, 0.00000330, 0.00000346, 0.00000299, 0.00000213, 0.00000111, 0.00000018, -0.00000047, -0.00000072, -0.00000058, -0.00000012, 0.00000053, 0.00000121, 0.00000176, },
		{ -0.00000154, -0.00000172, -0.00000161, -0.00000130, -0.00000091, -0.00000060, -0.00000050, -0.00000071, -0.00000126, -0.00000209, -0.00000311, -0.00000416, -0.00000508, -0.00000573, -0.00000599, -0.00000584, -0.00000530, -0.00000448, -0.00000351, -0.00000253, -0.00000170, -0.00000109, -0.00000078, -0.00000074, -0.00000092, -0.00000124, -0.00000160, -0.00000193, -0.00000217, -0.00000231, -0.00000237, -0.00000239, -0.00000241, -0.00000247, -0.00000256, -0.00000264, -0.00000264, -0.00000246, -0.00000201, -0.00000124, -0.00000015, 0.00000121, 0.00000268, 0.00000409, 0.00000522, 0.00000586, 0.00000587, 0.00000521, 0.00000394, 0.00000224, 0.00000042, -0.00000121, -0.00000228, -0.00000252, -0.00000177, -0.00000003, 0.00000250, 0.00000547, 0.00000842, 0.00001082, 0.00001218, 0.00001211, 0.00001044, 0.00000720, 0.00000268, -0.00000265, -0.00000814, -0.00001313, -0.00001696, -0.00001918, -0.00001951, -0.00001796, -0.00001481, -0.00001054, -0.00000580, -0.00000125, 0.00000248, 0.00000494, 0.00000590, 0.00000537, 0.00000363, 0.00000111, -0.00000161, -0.00000399, -0.00000555, -0.00000600, -0.00000526, -0.00000348, -0.00000102, 0.00000166, 0.00000403, 0.00000562, 0.00000610, 0.00000533, 0.00000337, 0.00000050, -0.00000285, -0.00000618, -0.00000898, -0.00001086, -0.00001155, -0.00001099, -0.00000932, -0.00000683, -0.00000395, -0.00000111, 0.00000128, 0.00000290, 0.00000361, 0.00000341, 0.00000245, 0.00000101, -0.00000059, -0.00000202, -0.00000303, -0.00000345, -0.00000326, -0.00000255, -0.00000149, -0.00000031, 0.00000075, 0.00000150, 0.00000182, 0.00000170, 0.00000120, 0.00000046, -0.00000035, -0.00000106, },
		{ 0.00000063, 0.00000096, 0.00000093, 0.00000059, 0.00000001, -0.00000065, -0.00000126, -0.00000168, -0.00000182, -0.00000165, -0.00000122, -0.00000059, 0.00000011, 0.00000077, 0.00000127, 0.00000155, 0.00000158, 0.00000139, 0.00000104, 0.00000060, 0.00000016, -0.00000022, -0.00000050, -0.00000068, -0.00000076, -0.00000079, -0.00000079, -0.00000077, -0.00000074, -0.00000065, -0.00000045, -0.00000009, 0.00000047, 0.00000123, 0.00000214, 0.00000312, 0.00000401, 0.00000467, 0.00000492, 0.00000466, 0.00000380, 0.00000239, 0.00000054, -0.00000156, -0.00000367, -0.00000554, -0.00000695, -0.00000776, -0.00000794, -0.00000759, -0.00000690, -0.00000615, -0.00000563, -0.00000560, -0.00000625, -0.00000760, -0.00000952, -0.00001172, -0.00001381, -0.00001531, -0.00001580, -0.00001491, -0.00001248, -0.00000855, -0.00000339, 0.00000253, 0.00000859, 0.00001410, 0.00001844, 0.00002109, 0.00002178, 0.00002050, 0.00001749, 0.00001324, 0.00000840, 0.00000367, -0.00000028, -0.00000296, -0.00000408, -0.00000362, -0.00000184, 0.00000081, 0.00000374, 0.00000633, 0.00000802, 0.00000845, 0.00000744, 0.00000510, 0.00000175, -0.00000213, -0.00000596, -0.00000918, -0.00001132, -0.00001212, -0.00001151, -0.00000965, -0.00000690, -0.00000373, -0.00000065, 0.00000186, 0.00000347, 0.00000401, 0.00000348, 0.00000210, 0.00000019, -0.00000185, -0.00000362, -0.00000481, -0.00000521, -0.00000478, -0.00000363, -0.00000200, -0.00000018, 0.00000151, 0.00000278, 0.00000345, 0.00000345, 0.00000284, 0.00000177, 0.00000047, -0.00000081, -0.00000183, -0.00000245, -0.00000259, -0.00000228, -0.00000162, -0.00000080, 0.00000001, },
		{ 0.00000049, -0.00000014, -0.00000053, -0.00000064, -0.00000052, -0.00000022, 0.00000013, 0.00000043, 0.00000059, 0.00000056, 0.00000032, -0.00000009, -0.00000061, -0.00000116, -0.00000165, -0.00000203, -0.00000227, -0.00000235, -0.00000229, -0.00000211, -0.00000185, -0.00000152, -0.00000113, -0.00000065, -0.00000008, 0.00000061, 0.00000141, 0.00000230, 0.00000319, 0.00000401, 0.00000462, 0.00000490, 0.00000476, 0.00000414, 0.00000306, 0.00000159, -0.00000011, -0.00000186, -0.00000343, -0.00000463, -0.00000531, -0.00000538, -0.00000487, -0.00000388, -0.00000258, -0.00000117, 0.00000014, 0.00000120, 0.00000194, 0.00000237, 0.00000263, 0.00000289, 0.00000339, 0.00000432, 0.00000581, 0.00000786, 0.00001034, 0.00001298, 0.00001540, 0.00001716, 0.00001783, 0.00001708, 0.00001474, 0.00001086, 0.00000569, -0.00000030, -0.00000649, -0.00001219, -0.00001673, -0.00001959, -0.00002045, -0.00001926, -0.00001624, -0.00001187, -0.00000681, -0.00000181, 0.00000242, 0.00000529, 0.00000645, 0.00000579, 0.00000352, 0.00000007, -0.00000397, -0.00000790, -0.00001109, -0.00001305, -0.00001348, -0.00001236, -0.00000990, -0.00000652, -0.00000276, 0.00000078, 0.00000360, 0.00000532, 0.00000577, 0.00000500, 0.00000326, 0.00000095, -0.00000145, -0.00000348, -0.00000476, -0.00000508, -0.00000438, -0.00000281, -0.00000065, 0.00000171, 0.00000387, 0.00000547, 0.00000628, 0.00000618, 0.00000523, 0.00000363, 0.00000166, -0.00000034, -0.00000205, -0.00000322, -0.00000372, -0.00000351, -0.00000270, -0.00000148, -0.00000010, 0.00000121, 0.00000223, 0.00000283, 0.00000296, 0.00000266, 0.00000205, 0.00000127, },
		{ -0.00000001, 0.00000019, 0.00000024, 0.00000009, -0.00000024, -0.00000071, -0.00000124, -0.00000175, -0.00000214, -0.00000236, -0.00000238, -0.00000219, -0.00000181, -0.00000129, -0.00000066, 0.00000003, 0.00000076, 0.00000151, 0.00000227, 0.00000303, 0.00000377, 0.00000443, 0.00000495, 0.00000526, 0.00000526, 0.00000490, 0.00000412, 0.00000294, 0.00000141, -0.00000034, -0.00000216, -0.00000386, -0.00000527, -0.00000623, -0.00000666, -0.00000653, -0.00000592, -0.00000494, -0.00000378, -0.00000261, -0.00000162, -0.00000093, -0.00000060, -0.00000061, -0.00000089, -0.00000130, -0.00000173, -0.00000205, -0.00000221, -0.00000224, -0.00000223, -0.00000234, -0.00000274, -0.00000360, -0.00000502, -0.00000699, -0.00000937, -0.00001190, -0.00001420, -0.00001585, -0.00001644, -0.00001563, -0.00001324, -0.00000930, -0.00000404, 0.00000208, 0.00000844, 0.00001435, 0.00001911, 0.00002216, 0.00002310, 0.00002182, 0.00001849, 0.00001355, 0.00000763, 0.00000150, -0.00000406, -0.00000839, -0.00001102, -0.00001174, -0.00001061, -0.00000797, -0.00000438, -0.00000049, 0.00000305, 0.00000567, 0.00000698, 0.00000686, 0.00000539, 0.00000290, -0.00000011, -0.00000311, -0.00000557, -0.00000707, -0.00000739, -0.00000649, -0.00000458, -0.00000200, 0.00000078, 0.00000328, 0.00000508, 0.00000589, 0.00000559, 0.00000424, 0.00000208, -0.00000052, -0.00000316, -0.00000542, -0.00000697, -0.00000762, -0.00000730, -0.00000612, -0.00000431, -0.00000216, 0.00000000, 0.00000188, 0.00000324, 0.00000398, 0.00000409, 0.00000365, 0.00000284, 0.00000187, 0.00000091, 0.00000014, -0.00000036, -0.00000055, -0.00000050, -0.00000028, },
		{ -0.00000182, -0.00000241, -0.00000280, -0.00000292, -0.00000276, -0.00000233, -0.00000168, -0.00000088, -0.00000000, 0.00000088, 0.00000173, 0.00000250, 0.00000317, 0.00000372, 0.00000411, 0.00000432, 0.00000431, 0.00000404, 0.00000347, 0.00000257, 0.00000137, -0.00000009, -0.00000171, -0.00000337, -0.00000491, -0.00000618, -0.00000704, -0.00000739, -0.00000720, -0.00000651, -0.00000541, -0.00000406, -0.00000261, -0.00000126, -0.00000015, 0.00000063, 0.00000105, 0.00000114, 0.00000099, 0.00000073, 0.00000048, 0.00000036, 0.00000042, 0.00000069, 0.00000113, 0.00000166, 0.00000219, 0.00000265, 0.00000299, 0.00000324, 0.00000347, 0.00000381, 0.00000442, 0.00000543, 0.00000693, 0.00000890, 0.00001122, 0.00001365, 0.00001584, 0.00001737, 0.00001785, 0.00001694, 0.00001444, 0.00001032, 0.00000479, -0.00000175, -0.00000871, -0.00001541, -0.00002116, -0.00002534, -0.00002750, -0.00002743, -0.00002519, -0.00002113, -0.00001580, -0.00000992, -0.00000424, 0.00000055, 0.00000392, 0.00000555, 0.00000542, 0.00000378, 0.00000106, -0.00000213, -0.00000515, -0.00000743, -0.00000853, -0.00000826, -0.00000664, -0.00000393, -0.00000057, 0.00000290, 0.00000592, 0.00000803, 0.00000891, 0.00000845, 0.00000676, 0.00000412, 0.00000098, -0.00000217, -0.00000484, -0.00000662, -0.00000729, -0.00000677, -0.00000516, -0.00000275, 0.00000012, 0.00000303, 0.00000561, 0.00000756, 0.00000871, 0.00000901, 0.00000855, 0.00000751, 0.00000613, 0.00000467, 0.00000333, 0.00000228, 0.00000157, 0.00000120, 0.00000109, 0.00000110, 0.00000112, 0.00000103, 0.00000075, 0.00000028, -0.00000037, -0.00000110, },
		{ -0.00000133, -0.00000005, 0.00000121, 0.00000236, 0.00000333, 0.00000410, 0.00000462, 0.00000491, 0.00000496, 0.00000477, 0.00000433, 0.00000365, 0.00000273, 0.00000159, 0.00000027, -0.00000115, -0.00000258, -0.00000390, -0.00000497, -0.00000569, -0.00000597, -0.00000577, -0.00000510, -0.00000403, -0.00000270, -0.00000126, 0.00000012, 0.00000129, 0.00000214, 0.00000259, 0.00000267, 0.00000242, 0.00000196, 0.00000142, 0.00000093, 0.00000059, 0.00000048, 0.00000060, 0.00000090, 0.00000132, 0.00000175, 0.00000209, 0.00000228, 0.00000228, 0.00000209, 0.00000176, 0.00000134, 0.00000091, 0.00000049, 0.00000009, -0.00000034, -0.00000091, -0.00000172, -0.00000288, -0.00000446, -0.00000647, -0.00000881, -0.00001126, -0.00001354, -0.00001529, -0.00001614, -0.00001577, -0.00001395, -0.00001063, -0.00000595, -0.00000023, 0.00000605, 0.00001228, 0.00001782, 0.00002207, 0.00002458, 0.00002510, 0.00002362, 0.00002038, 0.00001587, 0.00001070, 0.00000559, 0.00000121, -0.00000190, -0.00000339, -0.00000318, -0.00000142, 0.00000149, 0.00000499, 0.00000844, 0.00001123, 0.00001288, 0.00001307, 0.00001175, 0.00000908, 0.00000545, 0.00000135, -0.00000261, -0.00000590, -0.00000809, -0.00000890, -0.00000829, -0.00000639, -0.00000352, -0.00000013, 0.00000333, 0.00000640, 0.00000873, 0.00001009, 0.00001044, 0.00000985, 0.00000853, 0.00000676, 0.00000484, 0.00000302, 0.00000151, 0.00000042, -0.00000025, -0.00000057, -0.00000067, -0.00000070, -0.00000081, -0.00000108, -0.00000156, -0.00000222, -0.00000297, -0.00000370, -0.00000429, -0.00000462, -0.00000462, -0.00000425, -0.00000353, -0.00000252, },
		{ 0.00000647, 0.00000624, 0.00000576, 0.00000506, 0.00000418, 0.00000315, 0.00000198, 0.00000074, -0.00000055, -0.00000179, -0.00000292, -0.00000383, -0.00000443, -0.00000464, -0.00000442, -0.00000378, -0.00000277, -0.00000148, -0.00000005, 0.00000136, 0.00000261, 0.00000358, 0.00000418, 0.00000440, 0.00000427, 0.00000390, 0.00000340, 0.00000291, 0.00000257, 0.00000245, 0.00000261, 0.00000303, 0.00000365, 0.00000437, 0.00000506, 0.00000563, 0.00000599, 0.00000610, 0.00000595, 0.00000562, 0.00000517, 0.00000471, 0.00000432, 0.00000407, 0.00000401, 0.00000413, 0.00000440, 0.00000477, 0.00000521, 0.00000568, 0.00000618, 0.00000674, 0.00000742, 0.00000827, 0.00000933, 0.00001059, 0.00001199, 0.00001338, 0.00001457, 0.00001528, 0.00001525, 0.00001422, 0.00001204, 0.00000867, 0.00000420, -0.00000108, -0.00000677, -0.00001234, -0.00001725, -0.00002095, -0.00002304, -0.00002325, -0.00002157, -0.00001818, -0.00001351, -0.00000815, -0.00000275, 0.00000200, 0.00000553, 0.00000743, 0.00000751, 0.00000585, 0.00000275, -0.00000128, -0.00000563, -0.00000966, -0.00001277, -0.00001454, -0.00001473, -0.00001331, -0.00001051, -0.00000670, -0.00000238, 0.00000192, 0.00000571, 0.00000860, 0.00001037, 0.00001094, 0.00001043, 0.00000907, 0.00000716, 0.00000505, 0.00000303, 0.00000134, 0.00000010, -0.00000067, -0.00000104, -0.00000115, -0.00000117, -0.00000125, -0.00000150, -0.00000197, -0.00000262, -0.00000338, -0.00000412, -0.00000468, -0.00000493, -0.00000479, -0.00000421, -0.00000322, -0.00000189, -0.00000035, 0.00000126, 0.00000281, 0.00000417, 0.00000524, 0.00000599, 0.00000639, },
		{ -0.00000005, -0.00000122, -0.00000232, -0.00000329, -0.00000407, -0.00000460, -0.00000481, -0.00000466, -0.00000415, -0.00000330, -0.00000219, -0.00000093, 0.00000035, 0.00000151, 0.00000242, 0.00000300, 0.00000320, 0.00000304, 0.00000259, 0.00000196, 0.00000128, 0.00000070, 0.00000030, 0.00000017, 0.00000031, 0.00000067, 0.00000117, 0.00000168, 0.00000208, 0.00000225, 0.00000213, 0.00000170, 0.00000098, 0.00000008, -0.00000090, -0.00000183, -0.00000259, -0.00000309, -0.00000331, -0.00000326, -0.00000300, -0.00000263, -0.00000230, -0.00000210, -0.00000216, -0.00000253, -0.00000324, -0.00000430, -0.00000564, -0.00000722, -0.00000897, -0.00001080, -0.00001267, -0.00001453, -0.00001630, -0.00001793, -0.00001932, -0.00002034, -0.00002084, -0.00002065, -0.00001960, -0.00001755, -0.00001442, -0.00001024, -0.00000514, 0.00000061, 0.00000664, 0.00001247, 0.00001759, 0.00002152, 0.00002384, 0.00002430, 0.00002280, 0.00001949, 0.00001468, 0.00000890, 0.00000275, -0.00000310, -0.00000802, -0.00001152, -0.00001327, -0.00001316, -0.00001131, -0.00000802, -0.00000378, 0.00000086, 0.00000535, 0.00000916, 0.00001195, 0.00001349, 0.00001377, 0.00001291, 0.00001117, 0.00000889, 0.00000643, 0.00000410, 0.00000212, 0.00000064, -0.00000035, -0.00000091, -0.00000121, -0.00000140, -0.00000164, -0.00000205, -0.00000268, -0.00000347, -0.00000434, -0.00000515, -0.00000572, -0.00000591, -0.00000564, -0.00000486, -0.00000360, -0.00000198, -0.00000013, 0.00000178, 0.00000355, 0.00000506, 0.00000618, 0.00000686, 0.00000709, 0.00000691, 0.00000639, 0.00000560, 0.00000462, 0.00000352, 0.00000235, 0.00000115, },
		{ -0.00000718, -0.00000681, -0.00000613, -0.00000519, -0.00000407, -0.00000288, -0.00000176, -0.00000081, -0.00000015, 0.00000018, 0.00000016, -0.00000016, -0.00000070, -0.00000132, -0.00000189, -0.00000230, -0.00000247, -0.00000235, -0.00000197, -0.00000140, -0.00000076, -0.00000018, 0.00000023, 0.00000036, 0.00000020, -0.00000026, -0.00000093, -0.00000169, -0.00000240, -0.00000293, -0.00000315, -0.00000301, -0.00000251, -0.00000172, -0.00000076, 0.00000021, 0.00000102, 0.00000152, 0.00000160, 0.00000123, 0.00000044, -0.00000068, -0.00000197, -0.00000324, -0.00000427, -0.00000489, -0.00000496, -0.00000439, -0.00000314, -0.00000126, 0.00000116, 0.00000399, 0.00000706, 0.00001021, 0.00001326, 0.00001602, 0.00001833, 0.00002002, 0.00002094, 0.00002095, 0.00001994, 0.00001785, 0.00001465, 0.00001040, 0.00000525, -0.00000057, -0.00000674, -0.00001286, -0.00001850, -0.00002323, -0.00002664, -0.00002846, -0.00002850, -0.00002675, -0.00002338, -0.00001869, -0.00001312, -0.00000716, -0.00000135, 0.00000383, 0.00000800, 0.00001090, 0.00001245, 0.00001269, 0.00001183, 0.00001017, 0.00000803, 0.00000576, 0.00000364, 0.00000186, 0.00000053, -0.00000037, -0.00000094, -0.00000130, -0.00000163, -0.00000206, -0.00000268, -0.00000350, -0.00000447, -0.00000545, -0.00000631, -0.00000686, -0.00000697, -0.00000654, -0.00000555, -0.00000407, -0.00000221, -0.00000016, 0.00000190, 0.00000375, 0.00000523, 0.00000623, 0.00000668, 0.00000660, 0.00000605, 0.00000512, 0.00000393, 0.00000258, 0.00000118, -0.00000020, -0.00000152, -0.00000275, -0.00000388, -0.00000489, -0.00000577, -0.00000648, -0.00000698, -0.00000723, },
		{ 0.00000138, 0.00000230, 0.00000291, 0.00000318, 0.00000312, 0.00000279, 0.00000232, 0.00000182, 0.00000142, 0.00000121, 0.00000126, 0.00000157, 0.00000209, 0.00000271, 0.00000332, 0.00000379, 0.00000401, 0.00000393, 0.00000353, 0.00000288, 0.00000208, 0.00000128, 0.00000062, 0.00000023, 0.00000019, 0.00000051, 0.00000114, 0.00000194, 0.00000276, 0.00000339, 0.00000367, 0.00000346, 0.00000271, 0.00000146, -0.00000019, -0.00000204, -0.00000385, -0.00000539, -0.00000643, -0.00000681, -0.00000646, -0.00000542, -0.00000381, -0.00000185, 0.00000019, 0.00000203, 0.00000340, 0.00000408, 0.00000395, 0.00000294, 0.00000113, -0.00000137, -0.00000434, -0.00000755, -0.00001074, -0.00001366, -0.00001610, -0.00001788, -0.00001886, -0.00001899, -0.00001822, -0.00001657, -0.00001410, -0.00001089, -0.00000707, -0.00000279, 0.00000176, 0.00000637, 0.00001081, 0.00001485, 0.00001827, 0.00002089, 0.00002257, 0.00002327, 0.00002298, 0.00002181, 0.00001992, 0.00001751, 0.00001483, 0.00001211, 0.00000953, 0.00000725, 0.00000534, 0.00000378, 0.00000253, 0.00000147, 0.00000048, -0.00000056, -0.00000173, -0.00000304, -0.00000447, -0.00000590, -0.00000720, -0.00000820, -0.00000874, -0.00000872, -0.00000807, -0.00000681, -0.00000503, -0.00000289, -0.00000060, 0.00000163, 0.00000358, 0.00000509, 0.00000603, 0.00000638, 0.00000614, 0.00000541, 0.00000431, 0.00000298, 0.00000156, 0.00000015, -0.00000115, -0.00000230, -0.00000329, -0.00000411, -0.00000479, -0.00000533, -0.00000572, -0.00000595, -0.00000598, -0.00000578, -0.00000530, -0.00000456, -0.00000356, -0.00000237, -0.00000107, 0.00000022, },
		{ 0.00000187, 0.00000124, 0.00000075, 0.00000051, 0.00000053, 0.00000079, 0.00000121, 0.00000169, 0.00000209, 0.00000230, 0.00000225, 0.00000190, 0.00000129, 0.00000050, -0.00000032, -0.00000104, -0.00000151, -0.00000164, -0.00000140, -0.00000082, -0.00000002, 0.00000084, 0.00000157, 0.00000198, 0.00000193, 0.00000135, 0.00000025, -0.00000125, -0.00000294, -0.00000459, -0.00000591, -0.00000668, -0.00000671, -0.00000593, -0.00000439, -0.00000225, 0.00000026, 0.00000282, 0.00000512, 0.00000688, 0.00000789, 0.00000807, 0.00000746, 0.00000623, 0.00000464, 0.00000299, 0.00000163, 0.00000083, 0.00000080, 0.00000162, 0.00000326, 0.00000555, 0.00000823, 0.00001098, 0.00001346, 0.00001534, 0.00001640, 0.00001647, 0.00001551, 0.00001361, 0.00001094, 0.00000773, 0.00000426, 0.00000082, -0.00000234, -0.00000501, -0.00000705, -0.00000839, -0.00000904, -0.00000905, -0.00000855, -0.00000765, -0.00000654, -0.00000535, -0.00000425, -0.00000336, -0.00000280, -0.00000263, -0.00000289, -0.00000357, -0.00000463, -0.00000597, -0.00000745, -0.00000893, -0.00001023, -0.00001116, -0.00001159, -0.00001141, -0.00001055, -0.00000905, -0.00000697, -0.00000448, -0.00000176, 0.00000097, 0.00000348, 0.00000558, 0.00000712, 0.00000801, 0.00000825, 0.00000787, 0.00000700, 0.00000576, 0.00000432, 0.00000283, 0.00000140, 0.00000013, -0.00000096, -0.00000185, -0.00000257, -0.00000314, -0.00000359, -0.00000393, -0.00000416, -0.00000425, -0.00000416, -0.00000386, -0.00000331, -0.00000253, -0.00000154, -0.00000041, 0.00000075, 0.00000183, 0.00000272, 0.00000333, 0.00000360, 0.00000352, 0.00000314, 0.00000255, },
		{ 0.00000107, 0.00000165, 0.00000209, 0.00000230, 0.00000223, 0.00000187, 0.00000130, 0.00000063, -0.00000000, -0.00000046, -0.00000062, -0.00000043, 0.00000008, 0.00000085, 0.00000170, 0.00000246, 0.00000293, 0.00000296, 0.00000246, 0.00000142, -0.00000006, -0.00000180, -0.00000354, -0.00000500, -0.00000594, -0.00000616, -0.00000557, -0.00000421, -0.00000223, 0.00000011, 0.00000250, 0.00000460, 0.00000611, 0.00000683, 0.00000666, 0.00000567, 0.00000402, 0.00000201, -0.00000004, -0.00000180, -0.00000301, -0.00000353, -0.00000335, -0.00000261, -0.00000158, -0.00000063, -0.00000014, -0.00000046, -0.00000185, -0.00000440, -0.00000803, -0.00001248, -0.00001732, -0.00002202, -0.00002600, -0.00002872, -0.00002973, -0.00002878, -0.00002577, -0.00002085, -0.00001436, -0.00000679, 0.00000125, 0.00000912, 0.00001621, 0.00002200, 0.00002611, 0.00002833, 0.00002861, 0.00002710, 0.00002404, 0.00001982, 0.00001485, 0.00000957, 0.00000440, -0.00000030, -0.00000424, -0.00000722, -0.00000914, -0.00000998, -0.00000980, -0.00000873, -0.00000694, -0.00000466, -0.00000212, 0.00000045, 0.00000282, 0.00000480, 0.00000627, 0.00000712, 0.00000734, 0.00000697, 0.00000608, 0.00000479, 0.00000326, 0.00000161, -0.00000000, -0.00000150, -0.00000279, -0.00000387, -0.00000473, -0.00000540, -0.00000592, -0.00000632, -0.00000662, -0.00000682, -0.00000689, -0.00000680, -0.00000649, -0.00000595, -0.00000516, -0.00000414, -0.00000294, -0.00000167, -0.00000042, 0.00000068, 0.00000152, 0.00000204, 0.00000221, 0.00000204, 0.00000162, 0.00000105, 0.00000047, -0.00000000, -0.00000027, -0.00000027, -0.00000000, 0.00000048, },
		{ 0.00000171, 0.00000087, 0.00000014, -0.00000034, -0.00000049, -0.00000031, 0.00000015, 0.00000077, 0.00000135, 0.00000173, 0.00000172, 0.00000121, 0.00000018, -0.00000130, -0.00000307, -0.00000489, -0.00000647, -0.00000756, -0.00000793, -0.00000748, -0.00000620, -0.00000423, -0.00000182, 0.00000073, 0.00000307, 0.00000488, 0.00000593, 0.00000612, 0.00000546, 0.00000414, 0.00000241, 0.00000061, -0.00000093, -0.00000196, -0.00000233, -0.00000202, -0.00000119, -0.00000009, 0.00000091, 0.00000148, 0.00000129, 0.00000019, -0.00000183, -0.00000458, -0.00000768, -0.00001061, -0.00001280, -0.00001367, -0.00001277, -0.00000984, -0.00000484, 0.00000198, 0.00001009, 0.00001880, 0.00002725, 0.00003457, 0.00003995, 0.00004277, 0.00004264, 0.00003947, 0.00003347, 0.00002512, 0.00001510, 0.00000425, -0.00000658, -0.00001658, -0.00002507, -0.00003156, -0.00003578, -0.00003767, -0.00003737, -0.00003519, -0.00003154, -0.00002688, -0.00002168, -0.00001637, -0.00001131, -0.00000675, -0.00000289, 0.00000018, 0.00000245, 0.00000395, 0.00000476, 0.00000498, 0.00000473, 0.00000412, 0.00000329, 0.00000233, 0.00000134, 0.00000038, -0.00000048, -0.00000123, -0.00000185, -0.00000235, -0.00000274, -0.00000305, -0.00000329, -0.00000346, -0.00000354, -0.00000350, -0.00000331, -0.00000292, -0.00000231, -0.00000146, -0.00000039, 0.00000084, 0.00000215, 0.00000343, 0.00000456, 0.00000543, 0.00000597, 0.00000613, 0.00000593, 0.00000541, 0.00000469, 0.00000389, 0.00000315, 0.00000258, 0.00000227, 0.00000223, 0.00000244, 0.00000281, 0.00000323, 0.00000358, 0.00000372, 0.00000360, 0.00000319, 0.00000253, },
		{ -0.00000098, -0.00000025, 0.00000033, 0.00000060, 0.00000041, -0.00000029, -0.00000147, -0.00000298, -0.00000462, -0.00000610, -0.00000714, -0.00000752, -0.00000708, -0.00000579, -0.00000375, -0.00000117, 0.00000163, 0.00000429, 0.00000649, 0.00000794, 0.00000851, 0.00000816, 0.00000704, 0.00000537, 0.00000348, 0.00000170, 0.00000029, -0.00000055, -0.00000078, -0.00000052, 0.00000002, 0.00000051, 0.00000064, 0.00000014, -0.00000117, -0.00000326, -0.00000596, -0.00000889, -0.00001158, -0.00001351, -0.00001419, -0.00001327, -0.00001060, -0.00000629, -0.00000071, 0.00000554, 0.00001169, 0.00001694, 0.00002049, 0.00002176, 0.00002037, 0.00001630, 0.00000985, 0.00000160, -0.00000759, -0.00001676, -0.00002494, -0.00003124, -0.00003501, -0.00003587, -0.00003377, -0.00002896, -0.00002198, -0.00001354, -0.00000444, 0.00000449, 0.00001257, 0.00001923, 0.00002415, 0.00002722, 0.00002851, 0.00002826, 0.00002683, 0.00002460, 0.00002195, 0.00001921, 0.00001662, 0.00001432, 0.00001236, 0.00001075, 0.00000940, 0.00000824, 0.00000717, 0.00000613, 0.00000506, 0.00000395, 0.00000281, 0.00000166, 0.00000056, -0.00000046, -0.00000134, -0.00000202, -0.00000248, -0.00000267, -0.00000258, -0.00000220, -0.00000156, -0.00000069, 0.00000034, 0.00000144, 0.00000251, 0.00000343, 0.00000412, 0.00000449, 0.00000450, 0.00000415, 0.00000351, 0.00000265, 0.00000171, 0.00000081, 0.00000007, -0.00000040, -0.00000058, -0.00000046, -0.00000013, 0.00000033, 0.00000077, 0.00000108, 0.00000115, 0.00000095, 0.00000047, -0.00000019, -0.00000093, -0.00000162, -0.00000210, -0.00000228, -0.00000212, -0.00000165, },
		{ -0.00000148, -0.00000306, -0.00000458, -0.00000574, -0.00000628, -0.00000602, -0.00000487, -0.00000289, -0.00000028, 0.00000268, 0.00000562, 0.00000818, 0.00001004, 0.00001100, 0.00001098, 0.00001005, 0.00000843, 0.00000641, 0.00000432, 0.00000250, 0.00000116, 0.00000041, 0.00000022, 0.00000042, 0.00000076, 0.00000094, 0.00000070, -0.00000013, -0.00000158, -0.00000350, -0.00000560, -0.00000746, -0.00000862, -0.00000867, -0.00000730, -0.00000440, -0.00000009, 0.00000528, 0.00001116, 0.00001686, 0.00002167, 0.00002495, 0.00002624, 0.00002536, 0.00002240, 0.00001775, 0.00001208, 0.00000617, 0.00000087, -0.00000305, -0.00000502, -0.00000476, -0.00000231, 0.00000198, 0.00000747, 0.00001336, 0.00001877, 0.00002289, 0.00002509, 0.00002498, 0.00002249, 0.00001784, 0.00001152, 0.00000422, -0.00000330, -0.00001029, -0.00001612, -0.00002033, -0.00002272, -0.00002330, -0.00002230, -0.00002011, -0.00001720, -0.00001407, -0.00001115, -0.00000878, -0.00000716, -0.00000636, -0.00000632, -0.00000687, -0.00000781, -0.00000890, -0.00000992, -0.00001069, -0.00001109, -0.00001106, -0.00001059, -0.00000973, -0.00000854, -0.00000712, -0.00000555, -0.00000393, -0.00000238, -0.00000096, 0.00000023, 0.00000112, 0.00000169, 0.00000190, 0.00000178, 0.00000137, 0.00000076, 0.00000005, -0.00000064, -0.00000119, -0.00000152, -0.00000157, -0.00000134, -0.00000088, -0.00000029, 0.00000033, 0.00000085, 0.00000115, 0.00000118, 0.00000092, 0.00000043, -0.00000020, -0.00000084, -0.00000134, -0.00000159, -0.00000152, -0.00000112, -0.00000048, 0.00000027, 0.00000095, 0.00000138, 0.00000139, 0.00000090, -0.00000010, },
		{ -0.00000322, -0.00000064, 0.00000242, 0.00000558, 0.00000845, 0.00001066, 0.00001194, 0.00001214, 0.00001128, 0.00000952, 0.00000714, 0.00000451, 0.00000196, -0.00000018, -0.00000175, -0.00000267, -0.00000305, -0.00000308, -0.00000303, -0.00000316, -0.00000369, -0.00000470, -0.00000614, -0.00000780, -0.00000936, -0.00001045, -0.00001068, -0.00000979, -0.00000765, -0.00000435, -0.00000017, 0.00000442, 0.00000883, 0.00001244, 0.00001472, 0.00001527, 0.00001393, 0.00001081, 0.00000627, 0.00000087, -0.00000472, -0.00000980, -0.00001381, -0.00001635, -0.00001730, -0.00001682, -0.00001532, -0.00001340, -0.00001174, -0.00001095, -0.00001152, -0.00001369, -0.00001737, -0.00002222, -0.00002761, -0.00003274, -0.00003680, -0.00003901, -0.00003881, -0.00003594, -0.00003043, -0.00002268, -0.00001336, -0.00000331, 0.00000652, 0.00001524, 0.00002210, 0.00002661, 0.00002856, 0.00002801, 0.00002532, 0.00002101, 0.00001575, 0.00001019, 0.00000494, 0.00000048, -0.00000289, -0.00000503, -0.00000598, -0.00000589, -0.00000500, -0.00000360, -0.00000196, -0.00000030, 0.00000121, 0.00000245, 0.00000335, 0.00000392, 0.00000416, 0.00000411, 0.00000381, 0.00000330, 0.00000264, 0.00000189, 0.00000113, 0.00000042, -0.00000016, -0.00000055, -0.00000070, -0.00000060, -0.00000029, 0.00000017, 0.00000069, 0.00000114, 0.00000144, 0.00000151, 0.00000131, 0.00000087, 0.00000026, -0.00000039, -0.00000095, -0.00000131, -0.00000136, -0.00000109, -0.00000054, 0.00000018, 0.00000089, 0.00000141, 0.00000157, 0.00000125, 0.00000042, -0.00000086, -0.00000242, -0.00000399, -0.00000530, -0.00000605, -0.00000601, -0.00000507, },
		{ 0.00001011, 0.00000930, 0.00000739, 0.00000465, 0.00000146, -0.00000177, -0.00000467, -0.00000695, -0.00000846, -0.00000918, -0.00000927, -0.00000897, -0.00000855, -0.00000829, -0.00000837, -0.00000883, -0.00000959, -0.00001043, -0.00001105, -0.00001112, -0.00001037, -0.00000863, -0.00000590, -0.00000239, 0.00000157, 0.00000550, 0.00000889, 0.00001125, 0.00001223, 0.00001168, 0.00000964, 0.00000641, 0.00000244, -0.00000171, -0.00000548, -0.00000839, -0.00001012, -0.00001059, -0.00000992, -0.00000846, -0.00000667, -0.00000504, -0.00000401, -0.00000385, -0.00000460, -0.00000605, -0.00000777, -0.00000918, -0.00000964, -0.00000855, -0.00000553, -0.00000046, 0.00000647, 0.00001472, 0.00002349, 0.00003179, 0.00003860, 0.00004299, 0.00004426, 0.00004207, 0.00003643, 0.00002779, 0.00001691, 0.00000481, -0.00000741, -0.00001864, -0.00002795, -0.00003468, -0.00003848, -0.00003934, -0.00003758, -0.00003373, -0.00002848, -0.00002255, -0.00001660, -0.00001120, -0.00000669, -0.00000327, -0.00000094, 0.00000043, 0.00000102, 0.00000110, 0.00000087, 0.00000053, 0.00000019, -0.00000009, -0.00000027, -0.00000037, -0.00000041, -0.00000038, -0.00000029, -0.00000014, 0.00000010, 0.00000042, 0.00000080, 0.00000120, 0.00000156, 0.00000179, 0.00000185, 0.00000169, 0.00000131, 0.00000075, 0.00000010, -0.00000053, -0.00000101, -0.00000125, -0.00000121, -0.00000090, -0.00000040, 0.00000015, 0.00000057, 0.00000067, 0.00000033, -0.00000053, -0.00000187, -0.00000355, -0.00000533, -0.00000692, -0.00000802, -0.00000836, -0.00000776, -0.00000620, -0.00000377, -0.00000073, 0.00000255, 0.00000565, 0.00000815, 0.00000971, },
		{ -0.00000461, -0.00000642, -0.00000723, -0.00000710, -0.00000626, -0.00000503, -0.00000375, -0.00000269, -0.00000206, -0.00000188, -0.00000207, -0.00000239, -0.00000256, -0.00000229, -0.00000135, 0.00000036, 0.00000278, 0.00000569, 0.00000875, 0.00001153, 0.00001362, 0.00001466, 0.00001447, 0.00001301, 0.00001046, 0.00000715, 0.00000351, 0.00000001, -0.00000294, -0.00000507, -0.00000628, -0.00000663, -0.00000637, -0.00000585, -0.00000543, -0.00000544, -0.00000604, -0.00000724, -0.00000883, -0.00001045, -0.00001164, -0.00001195, -0.00001100, -0.00000864, -0.00000494, -0.00000027, 0.00000477, 0.00000938, 0.00001274, 0.00001409, 0.00001289, 0.00000891, 0.00000230, -0.00000637, -0.00001624, -0.00002619, -0.00003502, -0.00004162, -0.00004510, -0.00004494, -0.00004104, -0.00003372, -0.00002374, -0.00001208, 0.00000007, 0.00001156, 0.00002138, 0.00002880, 0.00003341, 0.00003516, 0.00003434, 0.00003148, 0.00002726, 0.00002240, 0.00001755, 0.00001323, 0.00000977, 0.00000729, 0.00000574, 0.00000495, 0.00000468, 0.00000467, 0.00000471, 0.00000465, 0.00000442, 0.00000404, 0.00000354, 0.00000301, 0.00000249, 0.00000203, 0.00000161, 0.00000120, 0.00000077, 0.00000028, -0.00000029, -0.00000091, -0.00000152, -0.00000203, -0.00000237, -0.00000247, -0.00000231, -0.00000195, -0.00000148, -0.00000106, -0.00000086, -0.00000102, -0.00000165, -0.00000274, -0.00000419, -0.00000580, -0.00000728, -0.00000833, -0.00000864, -0.00000800, -0.00000633, -0.00000368, -0.00000030, 0.00000347, 0.00000717, 0.00001034, 0.00001255, 0.00001351, 0.00001310, 0.00001138, 0.00000861, 0.00000516, 0.00000151, -0.00000188, },
		{ -0.00000217, -0.00000075, 0.00000003, 0.00000015, -0.00000027, -0.00000094, -0.00000156, -0.00000181, -0.00000146, -0.00000041, 0.00000127, 0.00000338, 0.00000558, 0.00000751, 0.00000881, 0.00000922, 0.00000861, 0.00000702, 0.00000465, 0.00000180, -0.00000114, -0.00000384, -0.00000601, -0.00000749, -0.00000829, -0.00000853, -0.00000844, -0.00000827, -0.00000824, -0.00000849, -0.00000900, -0.00000961, -0.00001006, -0.00000998, -0.00000906, -0.00000706, -0.00000392, 0.00000022, 0.00000499, 0.00000984, 0.00001412, 0.00001718, 0.00001850, 0.00001778, 0.00001503, 0.00001058, 0.00000507, -0.00000063, -0.00000554, -0.00000876, -0.00000957, -0.00000758, -0.00000280, 0.00000431, 0.00001294, 0.00002199, 0.00003023, 0.00003651, 0.00003987, 0.00003974, 0.00003595, 0.00002884, 0.00001913, 0.00000785, -0.00000381, -0.00001465, -0.00002366, -0.00003010, -0.00003361, -0.00003419, -0.00003221, -0.00002829, -0.00002320, -0.00001773, -0.00001257, -0.00000825, -0.00000506, -0.00000308, -0.00000218, -0.00000209, -0.00000250, -0.00000307, -0.00000357, -0.00000383, -0.00000381, -0.00000354, -0.00000312, -0.00000266, -0.00000224, -0.00000192, -0.00000167, -0.00000146, -0.00000122, -0.00000088, -0.00000044, 0.00000007, 0.00000057, 0.00000091, 0.00000096, 0.00000060, -0.00000020, -0.00000138, -0.00000280, -0.00000419, -0.00000524, -0.00000565, -0.00000516, -0.00000363, -0.00000108, 0.00000231, 0.00000619, 0.00001010, 0.00001351, 0.00001597, 0.00001708, 0.00001666, 0.00001472, 0.00001148, 0.00000735, 0.00000285, -0.00000145, -0.00000506, -0.00000759, -0.00000885, -0.00000886, -0.00000782, -0.00000608, -0.00000406, },
		{ -0.00000062, -0.00000139, -0.00000147, -0.00000074, 0.00000072, 0.00000268, 0.00000481, 0.00000673, 0.00000808, 0.00000861, 0.00000821, 0.00000693, 0.00000493, 0.00000250, -0.00000004, -0.00000240, -0.00000436, -0.00000581, -0.00000676, -0.00000730, -0.00000761, -0.00000782, -0.00000804, -0.00000830, -0.00000850, -0.00000847, -0.00000798, -0.00000682, -0.00000485, -0.00000208, 0.00000133, 0.00000505, 0.00000861, 0.00001147, 0.00001312, 0.00001319, 0.00001149, 0.00000810, 0.00000339, -0.00000207, -0.00000753, -0.00001221, -0.00001544, -0.00001675, -0.00001601, -0.00001342, -0.00000955, -0.00000521, -0.00000135, 0.00000109, 0.00000136, -0.00000098, -0.00000595, -0.00001310, -0.00002161, -0.00003036, -0.00003810, -0.00004363, -0.00004597, -0.00004453, -0.00003921, -0.00003040, -0.00001891, -0.00000590, 0.00000729, 0.00001937, 0.00002923, 0.00003607, 0.00003952, 0.00003965, 0.00003690, 0.00003201, 0.00002586, 0.00001936, 0.00001330, 0.00000827, 0.00000458, 0.00000228, 0.00000121, 0.00000105, 0.00000141, 0.00000193, 0.00000232, 0.00000243, 0.00000222, 0.00000175, 0.00000115, 0.00000053, -0.00000000, -0.00000044, -0.00000084, -0.00000129, -0.00000193, -0.00000284, -0.00000404, -0.00000545, -0.00000686, -0.00000803, -0.00000863, -0.00000839, -0.00000712, -0.00000477, -0.00000147, 0.00000245, 0.00000653, 0.00001024, 0.00001304, 0.00001448, 0.00001427, 0.00001238, 0.00000899, 0.00000451, -0.00000047, -0.00000532, -0.00000941, -0.00001225, -0.00001354, -0.00001324, -0.00001153, -0.00000880, -0.00000556, -0.00000236, 0.00000031, 0.00000212, 0.00000291, 0.00000273, 0.00000183, 0.00000058, },
		{ 0.00000449, 0.00000647, 0.00000783, 0.00000829, 0.00000773, 0.00000621, 0.00000391, 0.00000114, -0.00000179, -0.00000455, -0.00000692, -0.00000876, -0.00001006, -0.00001086, -0.00001127, -0.00001139, -0.00001128, -0.00001094, -0.00001030, -0.00000926, -0.00000771, -0.00000560, -0.00000294, 0.00000010, 0.00000326, 0.00000616, 0.00000840, 0.00000959, 0.00000946, 0.00000790, 0.00000504, 0.00000120, -0.00000308, -0.00000720, -0.00001050, -0.00001245, -0.00001273, -0.00001128, -0.00000834, -0.00000443, -0.00000024, 0.00000347, 0.00000599, 0.00000684, 0.00000585, 0.00000321, -0.00000054, -0.00000457, -0.00000790, -0.00000957, -0.00000880, -0.00000515, 0.00000135, 0.00001023, 0.00002055, 0.00003108, 0.00004042, 0.00004723, 0.00005044, 0.00004937, 0.00004390, 0.00003443, 0.00002189, 0.00000757, -0.00000708, -0.00002058, -0.00003167, -0.00003948, -0.00004357, -0.00004399, -0.00004124, -0.00003612, -0.00002962, -0.00002274, -0.00001634, -0.00001106, -0.00000724, -0.00000490, -0.00000386, -0.00000375, -0.00000414, -0.00000466, -0.00000504, -0.00000515, -0.00000501, -0.00000475, -0.00000452, -0.00000446, -0.00000462, -0.00000494, -0.00000523, -0.00000520, -0.00000457, -0.00000308, -0.00000060, 0.00000282, 0.00000694, 0.00001133, 0.00001544, 0.00001869, 0.00002056, 0.00002068, 0.00001893, 0.00001545, 0.00001066, 0.00000515, -0.00000033, -0.00000505, -0.00000839, -0.00000996, -0.00000962, -0.00000753, -0.00000413, -0.00000002, 0.00000413, 0.00000765, 0.00001004, 0.00001102, 0.00001056, 0.00000888, 0.00000640, 0.00000365, 0.00000117, -0.00000058, -0.00000133, -0.00000098, 0.00000033, 0.00000229, },
		{ 0.00000067, -0.00000243, -0.00000561, -0.00000850, -0.00001082, -0.00001242, -0.00001326, -0.00001340, -0.00001296, -0.00001206, -0.00001081, -0.00000927, -0.00000747, -0.00000541, -0.00000307, -0.00000052, 0.00000214, 0.00000472, 0.00000695, 0.00000856, 0.00000926, 0.00000888, 0.00000737, 0.00000483, 0.00000155, -0.00000204, -0.00000542, -0.00000811, -0.00000967, -0.00000986, -0.00000869, -0.00000638, -0.00000338, -0.00000029, 0.00000226, 0.00000372, 0.00000374, 0.00000224, -0.00000057, -0.00000418, -0.00000789, -0.00001091, -0.00001252, -0.00001222, -0.00000984, -0.00000561, -0.00000014, 0.00000565, 0.00001067, 0.00001381, 0.00001419, 0.00001129, 0.00000507, -0.00000399, -0.00001490, -0.00002630, -0.00003666, -0.00004450, -0.00004857, -0.00004810, -0.00004287, -0.00003328, -0.00002029, -0.00000526, 0.00001022, 0.00002455, 0.00003637, 0.00004470, 0.00004907, 0.00004952, 0.00004660, 0.00004118, 0.00003432, 0.00002710, 0.00002042, 0.00001494, 0.00001095, 0.00000845, 0.00000716, 0.00000666, 0.00000650, 0.00000631, 0.00000589, 0.00000525, 0.00000457, 0.00000417, 0.00000437, 0.00000544, 0.00000748, 0.00001038, 0.00001379, 0.00001720, 0.00001998, 0.00002153, 0.00002138, 0.00001929, 0.00001533, 0.00000987, 0.00000353, -0.00000288, -0.00000853, -0.00001266, -0.00001475, -0.00001458, -0.00001228, -0.00000831, -0.00000336, 0.00000174, 0.00000617, 0.00000925, 0.00001055, 0.00000995, 0.00000765, 0.00000413, 0.00000004, -0.00000391, -0.00000705, -0.00000893, -0.00000929, -0.00000817, -0.00000586, -0.00000283, 0.00000034, 0.00000308, 0.00000494, 0.00000562, 0.00000503, 0.00000329, },
		{ -0.00001360, -0.00001271, -0.00001115, -0.00000915, -0.00000691, -0.00000458, -0.00000225, 0.00000001, 0.00000218, 0.00000420, 0.00000599, 0.00000740, 0.00000829, 0.00000850, 0.00000793, 0.00000655, 0.00000445, 0.00000185, -0.00000092, -0.00000348, -0.00000544, -0.00000649, -0.00000644, -0.00000530, -0.00000326, -0.00000070, 0.00000192, 0.00000407, 0.00000532, 0.00000542, 0.00000432, 0.00000224, -0.00000037, -0.00000291, -0.00000472, -0.00000525, -0.00000412, -0.00000125, 0.00000311, 0.00000839, 0.00001378, 0.00001841, 0.00002142, 0.00002223, 0.00002057, 0.00001663, 0.00001104, 0.00000476, -0.00000100, -0.00000503, -0.00000630, -0.00000417, 0.00000151, 0.00001028, 0.00002114, 0.00003271, 0.00004333, 0.00005140, 0.00005557, 0.00005498, 0.00004934, 0.00003906, 0.00002514, 0.00000903, -0.00000755, -0.00002291, -0.00003558, -0.00004452, -0.00004927, -0.00004992, -0.00004709, -0.00004176, -0.00003510, -0.00002826, -0.00002217, -0.00001741, -0.00001419, -0.00001232, -0.00001132, -0.00001058, -0.00000950, -0.00000764, -0.00000483, -0.00000121, 0.00000278, 0.00000651, 0.00000925, 0.00001040, 0.00000954, 0.00000659, 0.00000184, -0.00000411, -0.00001039, -0.00001604, -0.00002018, -0.00002211, -0.00002151, -0.00001844, -0.00001335, -0.00000703, -0.00000045, 0.00000540, 0.00000964, 0.00001169, 0.00001133, 0.00000874, 0.00000444, -0.00000078, -0.00000602, -0.00001041, -0.00001327, -0.00001420, -0.00001313, -0.00001034, -0.00000637, -0.00000192, 0.00000227, 0.00000553, 0.00000738, 0.00000760, 0.00000624, 0.00000356, 0.00000003, -0.00000382, -0.00000747, -0.00001049, -0.00001257, -0.00001360, },
		{ 0.00000185, 0.00000344, 0.00000461, 0.00000540, 0.00000582, 0.00000585, 0.00000546, 0.00000460, 0.00000327, 0.00000150, -0.00000057, -0.00000274, -0.00000477, -0.00000637, -0.00000733, -0.00000751, -0.00000690, -0.00000564, -0.00000401, -0.00000237, -0.00000111, -0.00000058, -0.00000096, -0.00000229, -0.00000438, -0.00000686, -0.00000924, -0.00001099, -0.00001165, -0.00001095, -0.00000885, -0.00000560, -0.00000171, 0.00000215, 0.00000521, 0.00000679, 0.00000643, 0.00000397, -0.00000038, -0.00000605, -0.00001219, -0.00001782, -0.00002199, -0.00002399, -0.00002343, -0.00002043, -0.00001556, -0.00000981, -0.00000442, -0.00000068, 0.00000029, -0.00000225, -0.00000848, -0.00001797, -0.00002971, -0.00004222, -0.00005375, -0.00006258, -0.00006726, -0.00006685, -0.00006105, -0.00005031, -0.00003572, -0.00001889, -0.00000167, 0.00001408, 0.00002680, 0.00003543, 0.00003954, 0.00003938, 0.00003575, 0.00002989, 0.00002320, 0.00001703, 0.00001241, 0.00000993, 0.00000966, 0.00001115, 0.00001358, 0.00001595, 0.00001725, 0.00001672, 0.00001396, 0.00000906, 0.00000255, -0.00000467, -0.00001152, -0.00001690, -0.00001992, -0.00002006, -0.00001727, -0.00001200, -0.00000510, 0.00000232, 0.00000905, 0.00001401, 0.00001641, 0.00001593, 0.00001267, 0.00000723, 0.00000051, -0.00000637, -0.00001234, -0.00001648, -0.00001822, -0.00001739, -0.00001426, -0.00000944, -0.00000377, 0.00000178, 0.00000635, 0.00000928, 0.00001018, 0.00000903, 0.00000612, 0.00000199, -0.00000270, -0.00000724, -0.00001103, -0.00001364, -0.00001485, -0.00001468, -0.00001330, -0.00001105, -0.00000829, -0.00000539, -0.00000262, -0.00000017, },
		{ -0.00000366, -0.00000485, -0.00000615, -0.00000755, -0.00000897, -0.00001032, -0.00001143, -0.00001217, -0.00001242, -0.00001215, -0.00001142, -0.00001040, -0.00000932, -0.00000845, -0.00000805, -0.00000826, -0.00000912, -0.00001049, -0.00001209, -0.00001355, -0.00001445, -0.00001444, -0.00001329, -0.00001099, -0.00000775, -0.00000399, -0.00000029, 0.00000274, 0.00000453, 0.00000470, 0.00000314, 0.00000007, -0.00000402, -0.00000841, -0.00001226, -0.00001478, -0.00001540, -0.00001384, -0.00001025, -0.00000515, 0.00000063, 0.00000607, 0.00001017, 0.00001211, 0.00001147, 0.00000831, 0.00000319, -0.00000288, -0.00000859, -0.00001257, -0.00001360, -0.00001088, -0.00000418, 0.00000603, 0.00001867, 0.00003213, 0.00004451, 0.00005392, 0.00005877, 0.00005804, 0.00005144, 0.00003952, 0.00002355, 0.00000540, -0.00001279, -0.00002887, -0.00004107, -0.00004821, -0.00004985, -0.00004640, -0.00003897, -0.00002916, -0.00001886, -0.00000982, -0.00000346, -0.00000062, -0.00000144, -0.00000537, -0.00001130, -0.00001780, -0.00002336, -0.00002667, -0.00002687, -0.00002369, -0.00001746, -0.00000911, 0.00000009, 0.00000868, 0.00001531, 0.00001896, 0.00001909, 0.00001578, 0.00000965, 0.00000175, -0.00000657, -0.00001399, -0.00001933, -0.00002182, -0.00002118, -0.00001765, -0.00001192, -0.00000501, 0.00000191, 0.00000774, 0.00001160, 0.00001298, 0.00001179, 0.00000833, 0.00000325, -0.00000262, -0.00000838, -0.00001325, -0.00001664, -0.00001825, -0.00001807, -0.00001638, -0.00001360, -0.00001027, -0.00000691, -0.00000394, -0.00000166, -0.00000018, 0.00000051, 0.00000056, 0.00000012, -0.00000062, -0.00000154, -0.00000256, },
		{ -0.00000539, -0.00000544, -0.00000528, -0.00000488, -0.00000427, -0.00000355, -0.00000289, -0.00000244, -0.00000233, -0.00000262, -0.00000324, -0.00000404, -0.00000474, -0.00000507, -0.00000475, -0.00000361, -0.00000163, 0.00000104, 0.00000406, 0.00000699, 0.00000932, 0.00001059, 0.00001050, 0.00000893, 0.00000607, 0.00000231, -0.00000172, -0.00000535, -0.00000791, -0.00000893, -0.00000819, -0.00000581, -0.00000224, 0.00000179, 0.00000545, 0.00000789, 0.00000847, 0.00000687, 0.00000317, -0.00000211, -0.00000812, -0.00001379, -0.00001804, -0.00002001, -0.00001921, -0.00001568, -0.00001001, -0.00000329, 0.00000306, 0.00000752, 0.00000873, 0.00000580, -0.00000154, -0.00001281, -0.00002678, -0.00004163, -0.00005518, -0.00006526, -0.00007001, -0.00006819, -0.00005946, -0.00004440, -0.00002451, -0.00000198, 0.00002057, 0.00004053, 0.00005563, 0.00006432, 0.00006593, 0.00006082, 0.00005024, 0.00003616, 0.00002093, 0.00000691, -0.00000390, -0.00001014, -0.00001130, -0.00000772, -0.00000056, 0.00000852, 0.00001758, 0.00002479, 0.00002871, 0.00002853, 0.00002417, 0.00001629, 0.00000612, -0.00000474, -0.00001463, -0.00002208, -0.00002608, -0.00002619, -0.00002266, -0.00001627, -0.00000824, 0.00000000, 0.00000709, 0.00001191, 0.00001378, 0.00001255, 0.00000857, 0.00000259, -0.00000436, -0.00001117, -0.00001684, -0.00002064, -0.00002219, -0.00002148, -0.00001888, -0.00001496, -0.00001044, -0.00000604, -0.00000233, 0.00000031, 0.00000173, 0.00000201, 0.00000141, 0.00000025, -0.00000112, -0.00000240, -0.00000341, -0.00000408, -0.00000446, -0.00000463, -0.00000472, -0.00000484, -0.00000501, -0.00000522, },
		{ -0.00000516, -0.00000519, -0.00000533, -0.00000560, -0.00000597, -0.00000631, -0.00000646, -0.00000623, -0.00000547, -0.00000414, -0.00000231, -0.00000019, 0.00000189, 0.00000356, 0.00000445, 0.00000431, 0.00000303, 0.00000071, -0.00000233, -0.00000560, -0.00000856, -0.00001066, -0.00001148, -0.00001082, -0.00000878, -0.00000570, -0.00000217, 0.00000110, 0.00000344, 0.00000430, 0.00000341, 0.00000087, -0.00000291, -0.00000719, -0.00001107, -0.00001368, -0.00001430, -0.00001256, -0.00000855, -0.00000278, 0.00000381, 0.00001004, 0.00001474, 0.00001689, 0.00001593, 0.00001186, 0.00000532, -0.00000248, -0.00000992, -0.00001524, -0.00001682, -0.00001352, -0.00000494, 0.00000850, 0.00002546, 0.00004390, 0.00006128, 0.00007498, 0.00008267, 0.00008267, 0.00007431, 0.00005799, 0.00003522, 0.00000845, -0.00001932, -0.00004492, -0.00006546, -0.00007876, -0.00008365, -0.00008009, -0.00006923, -0.00005310, -0.00003437, -0.00001590, -0.00000032, 0.00001040, 0.00001513, 0.00001379, 0.00000724, -0.00000288, -0.00001445, -0.00002529, -0.00003343, -0.00003752, -0.00003693, -0.00003187, -0.00002327, -0.00001261, -0.00000164, 0.00000795, 0.00001477, 0.00001796, 0.00001731, 0.00001321, 0.00000662, -0.00000122, -0.00000893, -0.00001528, -0.00001935, -0.00002070, -0.00001932, -0.00001569, -0.00001058, -0.00000491, 0.00000040, 0.00000460, 0.00000722, 0.00000812, 0.00000746, 0.00000562, 0.00000315, 0.00000058, -0.00000162, -0.00000314, -0.00000389, -0.00000393, -0.00000347, -0.00000279, -0.00000217, -0.00000180, -0.00000181, -0.00000217, -0.00000279, -0.00000351, -0.00000420, -0.00000471, -0.00000502, -0.00000514, },
		{ -0.00000405, -0.00000321, -0.00000211, -0.00000071, 0.00000089, 0.00000252, 0.00000394, 0.00000486, 0.00000506, 0.00000441, 0.00000293, 0.00000082, -0.00000160, -0.00000389, -0.00000561, -0.00000642, -0.00000612, -0.00000472, -0.00000248, 0.00000016, 0.00000264, 0.00000439, 0.00000495, 0.00000409, 0.00000183, -0.00000148, -0.00000525, -0.00000875, -0.00001126, -0.00001218, -0.00001120, -0.00000839, -0.00000421, 0.00000055, 0.00000492, 0.00000788, 0.00000863, 0.00000671, 0.00000216, -0.00000446, -0.00001212, -0.00001946, -0.00002508, -0.00002774, -0.00002668, -0.00002173, -0.00001352, -0.00000333, 0.00000701, 0.00001539, 0.00001982, 0.00001873, 0.00001133, -0.00000223, -0.00002076, -0.00004213, -0.00006358, -0.00008206, -0.00009469, -0.00009922, -0.00009438, -0.00008012, -0.00005767, -0.00002937, 0.00000160, 0.00003168, 0.00005746, 0.00007609, 0.00008573, 0.00008579, 0.00007693, 0.00006100, 0.00004066, 0.00001902, -0.00000087, -0.00001646, -0.00002602, -0.00002887, -0.00002541, -0.00001697, -0.00000555, 0.00000655, 0.00001712, 0.00002437, 0.00002721, 0.00002537, 0.00001937, 0.00001037, -0.00000002, -0.00001012, -0.00001841, -0.00002378, -0.00002568, -0.00002419, -0.00001994, -0.00001392, -0.00000733, -0.00000134, 0.00000312, 0.00000548, 0.00000560, 0.00000376, 0.00000052, -0.00000335, -0.00000709, -0.00001005, -0.00001180, -0.00001222, -0.00001144, -0.00000982, -0.00000781, -0.00000590, -0.00000448, -0.00000379, -0.00000388, -0.00000463, -0.00000579, -0.00000708, -0.00000819, -0.00000891, -0.00000916, -0.00000894, -0.00000836, -0.00000757, -0.00000675, -0.00000599, -0.00000532, -0.00000471, },
		{ 0.00000518, 0.00000513, 0.00000455, 0.00000338, 0.00000167, -0.00000042, -0.00000258, -0.00000448, -0.00000581, -0.00000635, -0.00000603, -0.00000496, -0.00000345, -0.00000190, -0.00000076, -0.00000041, -0.00000111, -0.00000286, -0.00000545, -0.00000842, -0.00001122, -0.00001322, -0.00001394, -0.00001309, -0.00001070, -0.00000712, -0.00000298, 0.00000090, 0.00000368, 0.00000468, 0.00000351, 0.00000019, -0.00000480, -0.00001058, -0.00001603, -0.00001994, -0.00002127, -0.00001939, -0.00001418, -0.00000617, 0.00000354, 0.00001344, 0.00002181, 0.00002707, 0.00002806, 0.00002428, 0.00001607, 0.00000459, -0.00000827, -0.00002020, -0.00002883, -0.00003211, -0.00002868, -0.00001821, -0.00000146, 0.00001972, 0.00004260, 0.00006397, 0.00008059, 0.00008971, 0.00008943, 0.00007911, 0.00005944, 0.00003241, 0.00000100, -0.00003116, -0.00006038, -0.00008335, -0.00009769, -0.00010217, -0.00009693, -0.00008335, -0.00006387, -0.00004150, -0.00001946, -0.00000061, 0.00001287, 0.00001982, 0.00002012, 0.00001472, 0.00000530, -0.00000595, -0.00001682, -0.00002536, -0.00003019, -0.00003072, -0.00002711, -0.00002023, -0.00001141, -0.00000221, 0.00000591, 0.00001181, 0.00001481, 0.00001483, 0.00001229, 0.00000802, 0.00000305, -0.00000156, -0.00000497, -0.00000664, -0.00000640, -0.00000451, -0.00000147, 0.00000201, 0.00000522, 0.00000757, 0.00000869, 0.00000847, 0.00000709, 0.00000491, 0.00000242, 0.00000011, -0.00000163, -0.00000256, -0.00000262, -0.00000194, -0.00000077, 0.00000059, 0.00000186, 0.00000285, 0.00000346, 0.00000372, 0.00000377, 0.00000375, 0.00000382, 0.00000406, 0.00000446, 0.00000489, },
		{ -0.00000641, -0.00000762, -0.00000838, -0.00000853, -0.00000804, -0.00000703, -0.00000575, -0.00000454, -0.00000375, -0.00000362, -0.00000427, -0.00000561, -0.00000737, -0.00000912, -0.00001040, -0.00001079, -0.00001001, -0.00000804, -0.00000509, -0.00000165, 0.00000164, 0.00000408, 0.00000508, 0.00000428, 0.00000165, -0.00000245, -0.00000732, -0.00001202, -0.00001554, -0.00001700, -0.00001582, -0.00001189, -0.00000563, 0.00000205, 0.00000986, 0.00001635, 0.00002017, 0.00002029, 0.00001629, 0.00000843, -0.00000234, -0.00001448, -0.00002611, -0.00003529, -0.00004042, -0.00004051, -0.00003538, -0.00002580, -0.00001336, -0.00000028, 0.00001095, 0.00001798, 0.00001902, 0.00001318, 0.00000069, -0.00001709, -0.00003781, -0.00005842, -0.00007567, -0.00008657, -0.00008887, -0.00008142, -0.00006441, -0.00003934, -0.00000887, 0.00002361, 0.00005440, 0.00008006, 0.00009788, 0.00010623, 0.00010479, 0.00009452, 0.00007748, 0.00005650, 0.00003468, 0.00001497, -0.00000027, -0.00000957, -0.00001251, -0.00000968, -0.00000249, 0.00000709, 0.00001695, 0.00002517, 0.00003030, 0.00003161, 0.00002909, 0.00002343, 0.00001579, 0.00000760, 0.00000022, -0.00000523, -0.00000809, -0.00000824, -0.00000605, -0.00000231, 0.00000201, 0.00000591, 0.00000855, 0.00000943, 0.00000841, 0.00000575, 0.00000196, -0.00000223, -0.00000610, -0.00000905, -0.00001069, -0.00001092, -0.00000991, -0.00000803, -0.00000575, -0.00000357, -0.00000188, -0.00000091, -0.00000070, -0.00000110, -0.00000188, -0.00000272, -0.00000337, -0.00000366, -0.00000355, -0.00000314, -0.00000265, -0.00000230, -0.00000230, -0.00000278, -0.00000373, -0.00000502, },
		{ -0.00000554, -0.00000452, -0.00000377, -0.00000351, -0.00000382, -0.00000460, -0.00000561, -0.00000651, -0.00000693, -0.00000658, -0.00000535, -0.00000330, -0.00000075, 0.00000184, 0.00000392, 0.00000499, 0.00000468, 0.00000291, -0.00000012, -0.00000388, -0.00000762, -0.00001052, -0.00001181, -0.00001093, -0.00000775, -0.00000252, 0.00000403, 0.00001086, 0.00001673, 0.00002047, 0.00002119, 0.00001845, 0.00001240, 0.00000380, -0.00000606, -0.00001560, -0.00002316, -0.00002733, -0.00002718, -0.00002249, -0.00001378, -0.00000229, 0.00001023, 0.00002181, 0.00003061, 0.00003522, 0.00003500, 0.00003015, 0.00002180, 0.00001176, 0.00000227, -0.00000436, -0.00000623, -0.00000217, 0.00000799, 0.00002333, 0.00004189, 0.00006099, 0.00007759, 0.00008875, 0.00009214, 0.00008641, 0.00007143, 0.00004837, 0.00001954, -0.00001192, -0.00004249, -0.00006877, -0.00008797, -0.00009831, -0.00009921, -0.00009138, -0.00007662, -0.00005754, -0.00003712, -0.00001824, -0.00000329, 0.00000617, 0.00000959, 0.00000740, 0.00000088, -0.00000816, -0.00001766, -0.00002575, -0.00003100, -0.00003263, -0.00003059, -0.00002553, -0.00001856, -0.00001106, -0.00000440, 0.00000032, 0.00000245, 0.00000186, -0.00000107, -0.00000555, -0.00001059, -0.00001517, -0.00001842, -0.00001983, -0.00001924, -0.00001690, -0.00001334, -0.00000927, -0.00000543, -0.00000244, -0.00000066, -0.00000021, -0.00000090, -0.00000235, -0.00000406, -0.00000556, -0.00000647, -0.00000660, -0.00000597, -0.00000479, -0.00000340, -0.00000215, -0.00000139, -0.00000129, -0.00000188, -0.00000303, -0.00000446, -0.00000585, -0.00000690, -0.00000738, -0.00000723, -0.00000655, },
		{ -0.00000397, -0.00000469, -0.00000509, -0.00000494, -0.00000414, -0.00000279, -0.00000118, 0.00000030, 0.00000121, 0.00000117, -0.00000002, -0.00000230, -0.00000537, -0.00000870, -0.00001161, -0.00001345, -0.00001370, -0.00001211, -0.00000880, -0.00000427, 0.00000067, 0.00000501, 0.00000774, 0.00000802, 0.00000539, -0.00000014, -0.00000803, -0.00001728, -0.00002656, -0.00003446, -0.00003970, -0.00004141, -0.00003925, -0.00003355, -0.00002525, -0.00001574, -0.00000666, 0.00000039, 0.00000416, 0.00000396, -0.00000017, -0.00000751, -0.00001673, -0.00002616, -0.00003408, -0.00003905, -0.00004021, -0.00003746, -0.00003149, -0.00002374, -0.00001609, -0.00001057, -0.00000895, -0.00001241, -0.00002123, -0.00003472, -0.00005121, -0.00006829, -0.00008317, -0.00009312, -0.00009590, -0.00009016, -0.00007570, -0.00005356, -0.00002588, 0.00000438, 0.00003382, 0.00005915, 0.00007763, 0.00008748, 0.00008808, 0.00008006, 0.00006516, 0.00004592, 0.00002528, 0.00000610, -0.00000921, -0.00001908, -0.00002292, -0.00002111, -0.00001491, -0.00000615, 0.00000313, 0.00001102, 0.00001607, 0.00001746, 0.00001514, 0.00000974, 0.00000241, -0.00000545, -0.00001244, -0.00001741, -0.00001969, -0.00001911, -0.00001606, -0.00001132, -0.00000593, -0.00000094, 0.00000280, 0.00000472, 0.00000469, 0.00000295, 0.00000007, -0.00000322, -0.00000617, -0.00000817, -0.00000887, -0.00000821, -0.00000643, -0.00000402, -0.00000154, 0.00000045, 0.00000152, 0.00000148, 0.00000036, -0.00000156, -0.00000386, -0.00000606, -0.00000774, -0.00000862, -0.00000861, -0.00000781, -0.00000651, -0.00000506, -0.00000382, -0.00000306, -0.00000290, -0.00000327, },
		{ -0.00001204, -0.00001121, -0.00001080, -0.00001115, -0.00001241, -0.00001450, -0.00001713, -0.00001979, -0.00002192, -0.00002300, -0.00002266, -0.00002082, -0.00001773, -0.00001395, -0.00001025, -0.00000750, -0.00000647, -0.00000771, -0.00001137, -0.00001716, -0.00002436, -0.00003193, -0.00003868, -0.00004345, -0.00004535, -0.00004394, -0.00003929, -0.00003203, -0.00002321, -0.00001416, -0.00000625, -0.00000059, 0.00000209, 0.00000165, -0.00000149, -0.00000635, -0.00001165, -0.00001600, -0.00001820, -0.00001745, -0.00001352, -0.00000681, 0.00000172, 0.00001073, 0.00001876, 0.00002456, 0.00002729, 0.00002676, 0.00002349, 0.00001863, 0.00001380, 0.00001073, 0.00001099, 0.00001559, 0.00002475, 0.00003780, 0.00005316, 0.00006856, 0.00008140, 0.00008911, 0.00008962, 0.00008170, 0.00006527, 0.00004140, 0.00001228, -0.00001910, -0.00004936, -0.00007521, -0.00009392, -0.00010370, -0.00010398, -0.00009543, -0.00007982, -0.00005977, -0.00003829, -0.00001832, -0.00000234, 0.00000801, 0.00001211, 0.00001036, 0.00000403, -0.00000500, -0.00001460, -0.00002279, -0.00002804, -0.00002951, -0.00002710, -0.00002146, -0.00001375, -0.00000544, 0.00000202, 0.00000744, 0.00001011, 0.00000988, 0.00000714, 0.00000272, -0.00000231, -0.00000686, -0.00001003, -0.00001127, -0.00001048, -0.00000798, -0.00000440, -0.00000061, 0.00000257, 0.00000445, 0.00000464, 0.00000310, 0.00000013, -0.00000370, -0.00000771, -0.00001120, -0.00001365, -0.00001475, -0.00001452, -0.00001322, -0.00001129, -0.00000929, -0.00000771, -0.00000690, -0.00000700, -0.00000792, -0.00000939, -0.00001101, -0.00001239, -0.00001322, -0.00001336, -0.00001288, },
		{ -0.00000951, -0.00001144, -0.00001297, -0.00001359, -0.00001300, -0.00001114, -0.00000828, -0.00000498, -0.00000195, 0.00000005, 0.00000037, -0.00000135, -0.00000512, -0.00001057, -0.00001693, -0.00002325, -0.00002848, -0.00003172, -0.00003235, -0.00003017, -0.00002546, -0.00001892, -0.00001153, -0.00000442, 0.00000138, 0.00000510, 0.00000642, 0.00000547, 0.00000283, -0.00000057, -0.00000369, -0.00000556, -0.00000546, -0.00000312, 0.00000124, 0.00000696, 0.00001303, 0.00001832, 0.00002179, 0.00002271, 0.00002080, 0.00001632, 0.00001003, 0.00000301, -0.00000352, -0.00000851, -0.00001124, -0.00001157, -0.00000998, -0.00000749, -0.00000553, -0.00000562, -0.00000907, -0.00001668, -0.00002847, -0.00004363, -0.00006047, -0.00007669, -0.00008969, -0.00009695, -0.00009649, -0.00008718, -0.00006908, -0.00004340, -0.00001249, 0.00002054, 0.00005215, 0.00007896, 0.00009815, 0.00010791, 0.00010766, 0.00009812, 0.00008118, 0.00005957, 0.00003645, 0.00001493, -0.00000242, -0.00001385, -0.00001870, -0.00001736, -0.00001117, -0.00000209, 0.00000766, 0.00001598, 0.00002127, 0.00002263, 0.00001997, 0.00001397, 0.00000586, -0.00000281, -0.00001050, -0.00001598, -0.00001849, -0.00001792, -0.00001473, -0.00000985, -0.00000449, 0.00000014, 0.00000303, 0.00000359, 0.00000170, -0.00000227, -0.00000757, -0.00001321, -0.00001821, -0.00002173, -0.00002328, -0.00002275, -0.00002044, -0.00001694, -0.00001304, -0.00000951, -0.00000698, -0.00000583, -0.00000608, -0.00000750, -0.00000957, -0.00001172, -0.00001337, -0.00001414, -0.00001385, -0.00001261, -0.00001077, -0.00000880, -0.00000721, -0.00000641, -0.00000661, -0.00000774, },
		{ -0.00001468, -0.00001215, -0.00000982, -0.00000847, -0.00000874, -0.00001100, -0.00001521, -0.00002096, -0.00002753, -0.00003395, -0.00003926, -0.00004262, -0.00004351, -0.00004177, -0.00003768, -0.00003188, -0.00002522, -0.00001862, -0.00001290, -0.00000860, -0.00000592, -0.00000467, -0.00000435, -0.00000427, -0.00000370, -0.00000207, 0.00000093, 0.00000524, 0.00001047, 0.00001591, 0.00002074, 0.00002416, 0.00002557, 0.00002473, 0.00002178, 0.00001728, 0.00001208, 0.00000714, 0.00000339, 0.00000147, 0.00000166, 0.00000377, 0.00000720, 0.00001107, 0.00001440, 0.00001636, 0.00001644, 0.00001469, 0.00001168, 0.00000848, 0.00000647, 0.00000709, 0.00001152, 0.00002036, 0.00003346, 0.00004978, 0.00006748, 0.00008410, 0.00009691, 0.00010334, 0.00010139, 0.00009000, 0.00006932, 0.00004073, 0.00000675, -0.00002927, -0.00006360, -0.00009267, -0.00011352, -0.00012426, -0.00012431, -0.00011445, -0.00009670, -0.00007396, -0.00004960, -0.00002692, -0.00000866, 0.00000334, 0.00000835, 0.00000682, 0.00000016, -0.00000951, -0.00001982, -0.00002853, -0.00003392, -0.00003505, -0.00003185, -0.00002507, -0.00001611, -0.00000667, 0.00000153, 0.00000711, 0.00000923, 0.00000776, 0.00000321, -0.00000334, -0.00001055, -0.00001699, -0.00002148, -0.00002329, -0.00002223, -0.00001864, -0.00001333, -0.00000741, -0.00000198, 0.00000198, 0.00000388, 0.00000354, 0.00000125, -0.00000235, -0.00000640, -0.00001003, -0.00001251, -0.00001342, -0.00001272, -0.00001071, -0.00000799, -0.00000529, -0.00000331, -0.00000258, -0.00000333, -0.00000543, -0.00000849, -0.00001186, -0.00001484, -0.00001684, -0.00001746, -0.00001664, },
		{ -0.00001787, -0.00002394, -0.00003104, -0.00003809, -0.00004397, -0.00004775, -0.00004879, -0.00004689, -0.00004226, -0.00003548, -0.00002737, -0.00001884, -0.00001069, -0.00000351, 0.00000239, 0.00000704, 0.00001071, 0.00001381, 0.00001676, 0.00001986, 0.00002324, 0.00002672, 0.00002996, 0.00003244, 0.00003363, 0.00003314, 0.00003078, 0.00002667, 0.00002121, 0.00001508, 0.00000905, 0.00000388, 0.00000016, -0.00000182, -0.00000213, -0.00000116, 0.00000041, 0.00000182, 0.00000235, 0.00000149, -0.00000090, -0.00000459, -0.00000899, -0.00001330, -0.00001667, -0.00001844, -0.00001831, -0.00001648, -0.00001370, -0.00001115, -0.00001029, -0.00001257, -0.00001908, -0.00003034, -0.00004603, -0.00006491, -0.00008492, -0.00010340, -0.00011746, -0.00012438, -0.00012210, -0.00010959, -0.00008708, -0.00005614, -0.00001952, 0.00001915, 0.00005588, 0.00008682, 0.00010884, 0.00011991, 0.00011946, 0.00010835, 0.00008878, 0.00006390, 0.00003737, 0.00001273, -0.00000701, -0.00001988, -0.00002512, -0.00002322, -0.00001576, -0.00000506, 0.00000622, 0.00001562, 0.00002119, 0.00002189, 0.00001760, 0.00000917, -0.00000185, -0.00001352, -0.00002389, -0.00003134, -0.00003487, -0.00003424, -0.00002997, -0.00002321, -0.00001547, -0.00000834, -0.00000319, -0.00000091, -0.00000177, -0.00000543, -0.00001102, -0.00001732, -0.00002308, -0.00002719, -0.00002894, -0.00002814, -0.00002511, -0.00002062, -0.00001570, -0.00001143, -0.00000871, -0.00000810, -0.00000969, -0.00001310, -0.00001761, -0.00002225, -0.00002606, -0.00002826, -0.00002843, -0.00002660, -0.00002320, -0.00001907, -0.00001518, -0.00001253, -0.00001190, -0.00001369, },
		{ -0.00005882, -0.00005672, -0.00005085, -0.00004191, -0.00003093, -0.00001906, -0.00000737, 0.00000329, 0.00001236, 0.00001962, 0.00002513, 0.00002913, 0.00003194, 0.00003384, 0.00003499, 0.00003542, 0.00003501, 0.00003358, 0.00003097, 0.00002714, 0.00002221, 0.00001650, 0.00001051, 0.00000484, 0.00000009, -0.00000327, -0.00000499, -0.00000506, -0.00000379, -0.00000168, 0.00000062, 0.00000249, 0.00000344, 0.00000326, 0.00000200, 0.00000002, -0.00000212, -0.00000378, -0.00000441, -0.00000367, -0.00000155, 0.00000159, 0.00000513, 0.00000824, 0.00001013, 0.00001025, 0.00000842, 0.00000502, 0.00000095, -0.00000245, -0.00000362, -0.00000106, 0.00000632, 0.00001895, 0.00003634, 0.00005705, 0.00007877, 0.00009857, 0.00011330, 0.00012006, 0.00011667, 0.00010208, 0.00007661, 0.00004205, 0.00000143, -0.00004125, -0.00008163, -0.00011550, -0.00013945, -0.00015129, -0.00015040, -0.00013775, -0.00011577, -0.00008797, -0.00005838, -0.00003097, -0.00000905, 0.00000511, 0.00001065, 0.00000807, -0.00000088, -0.00001360, -0.00002713, -0.00003864, -0.00004591, -0.00004765, -0.00004368, -0.00003485, -0.00002286, -0.00000986, 0.00000194, 0.00001066, 0.00001511, 0.00001495, 0.00001070, 0.00000359, -0.00000466, -0.00001227, -0.00001766, -0.00001978, -0.00001831, -0.00001362, -0.00000674, 0.00000089, 0.00000773, 0.00001241, 0.00001403, 0.00001228, 0.00000754, 0.00000071, -0.00000688, -0.00001383, -0.00001888, -0.00002118, -0.00002046, -0.00001707, -0.00001192, -0.00000627, -0.00000155, 0.00000096, 0.00000036, -0.00000373, -0.00001106, -0.00002081, -0.00003172, -0.00004230, -0.00005108, -0.00005685, },
		{ -0.00000125, 0.00001471, 0.00002776, 0.00003756, 0.00004412, 0.00004780, 0.00004908, 0.00004851, 0.00004655, 0.00004356, 0.00003977, 0.00003535, 0.00003040, 0.00002508, 0.00001962, 0.00001433, 0.00000959, 0.00000577, 0.00000317, 0.00000197, 0.00000213, 0.00000341, 0.00000541, 0.00000758, 0.00000940, 0.00001041, 0.00001037, 0.00000922, 0.00000720, 0.00000470, 0.00000223, 0.00000026, -0.00000085, -0.00000099, -0.00000029, 0.00000086, 0.00000193, 0.00000235, 0.00000166, -0.00000034, -0.00000354, -0.00000747, -0.00001141, -0.00001452, -0.00001601, -0.00001539, -0.00001264, -0.00000829, -0.00000343, 0.00000041, 0.00000152, -0.00000172, -0.00001043, -0.00002497, -0.00004472, -0.00006799, -0.00009218, -0.00011402, -0.00013002, -0.00013700, -0.00013259, -0.00011567, -0.00008664, -0.00004750, -0.00000166, 0.00004645, 0.00009195, 0.00013014, 0.00015719, 0.00017063, 0.00016971, 0.00015547, 0.00013059, 0.00009894, 0.00006504, 0.00003337, 0.00000773, -0.00000929, -0.00001659, -0.00001465, -0.00000533, 0.00000849, 0.00002349, 0.00003646, 0.00004481, 0.00004703, 0.00004280, 0.00003301, 0.00001954, 0.00000481, -0.00000864, -0.00001865, -0.00002380, -0.00002365, -0.00001879, -0.00001064, -0.00000123, 0.00000727, 0.00001292, 0.00001438, 0.00001117, 0.00000367, -0.00000688, -0.00001871, -0.00002982, -0.00003838, -0.00004307, -0.00004331, -0.00003938, -0.00003237, -0.00002395, -0.00001609, -0.00001072, -0.00000932, -0.00001274, -0.00002097, -0.00003319, -0.00004788, -0.00006306, -0.00007661, -0.00008661, -0.00009159, -0.00009077, -0.00008410, -0.00007224, -0.00005641, -0.00003821, -0.00001930, },
		{ 0.00007226, 0.00006644, 0.00005940, 0.00005190, 0.00004446, 0.00003742, 0.00003098, 0.00002526, 0.00002035, 0.00001637, 0.00001341, 0.00001158, 0.00001088, 0.00001124, 0.00001245, 0.00001421, 0.00001610, 0.00001772, 0.00001869, 0.00001880, 0.00001799, 0.00001640, 0.00001432, 0.00001217, 0.00001038, 0.00000928, 0.00000908, 0.00000975, 0.00001108, 0.00001272, 0.00001422, 0.00001521, 0.00001544, 0.00001489, 0.00001378, 0.00001253, 0.00001169, 0.00001177, 0.00001316, 0.00001593, 0.00001985, 0.00002429, 0.00002843, 0.00003134, 0.00003220, 0.00003057, 0.00002652, 0.00002076, 0.00001457, 0.00000973, 0.00000816, 0.00001163, 0.00002136, 0.00003770, 0.00005988, 0.00008598, 0.00011302, 0.00013732, 0.00015498, 0.00016240, 0.00015692, 0.00013726, 0.00010384, 0.00005886, 0.00000612, -0.00004944, -0.00010231, -0.00014722, -0.00017977, -0.00019709, -0.00019822, -0.00018422, -0.00015799, -0.00012384, -0.00008683, -0.00005203, -0.00002379, -0.00000516, 0.00000252, -0.00000032, -0.00001165, -0.00002825, -0.00004632, -0.00006215, -0.00007274, -0.00007624, -0.00007223, -0.00006170, -0.00004681, -0.00003045, -0.00001566, -0.00000514, -0.00000069, -0.00000297, -0.00001138, -0.00002424, -0.00003906, -0.00005308, -0.00006371, -0.00006906, -0.00006824, -0.00006153, -0.00005033, -0.00003688, -0.00002388, -0.00001404, -0.00000955, -0.00001177, -0.00002091, -0.00003608, -0.00005534, -0.00007604, -0.00009524, -0.00011014, -0.00011847, -0.00011882, -0.00011083, -0.00009512, -0.00007325, -0.00004736, -0.00001990, 0.00000672, 0.00003045, 0.00004980, 0.00006392, 0.00007266, 0.00007639, 0.00007592, },
		{ 0.00002992, 0.00002532, 0.00002207, 0.00001993, 0.00001871, 0.00001821, 0.00001829, 0.00001876, 0.00001940, 0.00001996, 0.00002019, 0.00001986, 0.00001880, 0.00001698, 0.00001449, 0.00001154, 0.00000844, 0.00000553, 0.00000313, 0.00000142, 0.00000045, 0.00000010, 0.00000012, 0.00000016, -0.00000011, -0.00000097, -0.00000253, -0.00000474, -0.00000735, -0.00001002, -0.00001239, -0.00001413, -0.00001511, -0.00001542, -0.00001538, -0.00001549, -0.00001634, -0.00001844, -0.00002211, -0.00002734, -0.00003373, -0.00004054, -0.00004677, -0.00005136, -0.00005345, -0.00005258, -0.00004891, -0.00004329, -0.00003725, -0.00003280, -0.00003214, -0.00003729, -0.00004960, -0.00006946, -0.00009599, -0.00012703, -0.00015922, -0.00018842, -0.00021019, -0.00022045, -0.00021616, -0.00019580, -0.00015976, -0.00011046, -0.00005210, 0.00000974, 0.00006888, 0.00011927, 0.00015585, 0.00017528, 0.00017631, 0.00016003, 0.00012965, 0.00009003, 0.00004694, 0.00000624, -0.00002698, -0.00004910, -0.00005842, -0.00005535, -0.00004220, -0.00002271, -0.00000139, 0.00001727, 0.00002952, 0.00003299, 0.00002697, 0.00001252, -0.00000783, -0.00003054, -0.00005174, -0.00006789, -0.00007643, -0.00007619, -0.00006761, -0.00005263, -0.00003435, -0.00001651, -0.00000282, 0.00000363, 0.00000093, -0.00001133, -0.00003201, -0.00005857, -0.00008746, -0.00011465, -0.00013624, -0.00014900, -0.00015085, -0.00014110, -0.00012046, -0.00009095, -0.00005550, -0.00001757, 0.00001938, 0.00005231, 0.00007889, 0.00009776, 0.00010856, 0.00011177, 0.00010861, 0.00010069, 0.00008976, 0.00007746, 0.00006516, 0.00005384, 0.00004408, 0.00003611, },
		{ 0.00000817, 0.00000727, 0.00000612, 0.00000468, 0.00000290, 0.00000074, -0.00000177, -0.00000458, -0.00000755, -0.00001048, -0.00001313, -0.00001529, -0.00001678, -0.00001755, -0.00001765, -0.00001725, -0.00001661, -0.00001603, -0.00001573, -0.00001589, -0.00001652, -0.00001751, -0.00001863, -0.00001960, -0.00002013, -0.00002005, -0.00001932, -0.00001808, -0.00001660, -0.00001524, -0.00001437, -0.00001422, -0.00001484, -0.00001605, -0.00001740, -0.00001829, -0.00001806, -0.00001616, -0.00001229, -0.00000654, 0.00000056, 0.00000811, 0.00001494, 0.00001984, 0.00002182, 0.00002040, 0.00001576, 0.00000892, 0.00000162, -0.00000384, -0.00000497, 0.00000051, 0.00001414, 0.00003637, 0.00006622, 0.00010122, 0.00013759, 0.00017059, 0.00019519, 0.00020670, 0.00020158, 0.00017802, 0.00013641, 0.00007941, 0.00001180, -0.00006005, -0.00012898, -0.00018795, -0.00023100, -0.00025406, -0.00025553, -0.00023652, -0.00020064, -0.00015353, -0.00010203, -0.00005318, -0.00001322, 0.00001327, 0.00002397, 0.00001908, 0.00000117, -0.00002531, -0.00005482, -0.00008165, -0.00010088, -0.00010910, -0.00010500, -0.00008949, -0.00006551, -0.00003751, -0.00001071, 0.00000979, 0.00001987, 0.00001700, 0.00000073, -0.00002728, -0.00006349, -0.00010302, -0.00014036, -0.00017018, -0.00018808, -0.00019122, -0.00017859, -0.00015119, -0.00011173, -0.00006425, -0.00001350, 0.00003572, 0.00007913, 0.00011347, 0.00013679, 0.00014857, 0.00014956, 0.00014152, 0.00012688, 0.00010830, 0.00008832, 0.00006907, 0.00005205, 0.00003814, 0.00002759, 0.00002018, 0.00001538, 0.00001252, 0.00001093, 0.00001004, 0.00000944, 0.00000887, },
		{ 0.00001995, 0.00001875, 0.00001771, 0.00001693, 0.00001648, 0.00001644, 0.00001685, 0.00001766, 0.00001878, 0.00002003, 0.00002120, 0.00002208, 0.00002251, 0.00002242, 0.00002185, 0.00002094, 0.00001989, 0.00001893, 0.00001825, 0.00001794, 0.00001799, 0.00001824, 0.00001848, 0.00001843, 0.00001788, 0.00001671, 0.00001498, 0.00001288, 0.00001078, 0.00000906, 0.00000809, 0.00000809, 0.00000902, 0.00001059, 0.00001224, 0.00001323, 0.00001280, 0.00001034, 0.00000555, -0.00000146, -0.00001004, -0.00001914, -0.00002743, -0.00003353, -0.00003634, -0.00003533, -0.00003075, -0.00002378, -0.00001642, -0.00001131, -0.00001132, -0.00001903, -0.00003621, -0.00006336, -0.00009934, -0.00014130, -0.00018483, -0.00022444, -0.00025421, -0.00026861, -0.00026340, -0.00023632, -0.00018767, -0.00012046, -0.00004023, 0.00004554, 0.00012828, 0.00019945, 0.00025165, 0.00027964, 0.00028109, 0.00025688, 0.00021106, 0.00015024, 0.00008272, 0.00001728, -0.00003802, -0.00007697, -0.00009604, -0.00009484, -0.00007607, -0.00004505, -0.00000883, 0.00002486, 0.00004888, 0.00005770, 0.00004826, 0.00002039, -0.00002318, -0.00007724, -0.00013487, -0.00018846, -0.00023083, -0.00025613, -0.00026072, -0.00024350, -0.00020607, -0.00015233, -0.00008792, -0.00001941, 0.00004661, 0.00010431, 0.00014930, 0.00017901, 0.00019282, 0.00019191, 0.00017888, 0.00015723, 0.00013081, 0.00010327, 0.00007764, 0.00005606, 0.00003968, 0.00002872, 0.00002266, 0.00002048, 0.00002093, 0.00002278, 0.00002502, 0.00002692, 0.00002808, 0.00002841, 0.00002797, 0.00002698, 0.00002566, 0.00002420, 0.00002271, 0.00002129, },
		{ 0.00002012, 0.00001918, 0.00001830, 0.00001738, 0.00001633, 0.00001502, 0.00001341, 0.00001148, 0.00000932, 0.00000707, 0.00000490, 0.00000299, 0.00000147, 0.00000038, -0.00000031, -0.00000073, -0.00000107, -0.00000149, -0.00000215, -0.00000307, -0.00000419, -0.00000531, -0.00000620, -0.00000659, -0.00000631, -0.00000529, -0.00000367, -0.00000175, 0.00000005, 0.00000126, 0.00000148, 0.00000050, -0.00000160, -0.00000444, -0.00000735, -0.00000948, -0.00000998, -0.00000815, -0.00000369, 0.00000322, 0.00001182, 0.00002087, 0.00002885, 0.00003420, 0.00003568, 0.00003274, 0.00002572, 0.00001605, 0.00000611, -0.00000099, -0.00000185, 0.00000658, 0.00002642, 0.00005827, 0.00010082, 0.00015071, 0.00020268, 0.00025011, 0.00028579, 0.00030287, 0.00029593, 0.00026190, 0.00020070, 0.00011557, 0.00001293, -0.00009824, -0.00020742, -0.00030383, -0.00037780, -0.00042208, -0.00043287, -0.00041037, -0.00035883, -0.00028606, -0.00020242, -0.00011944, -0.00004833, 0.00000155, 0.00002387, 0.00001623, -0.00001966, -0.00007819, -0.00015072, -0.00022675, -0.00029536, -0.00034668, -0.00037312, -0.00037030, -0.00033752, -0.00027766, -0.00019669, -0.00010273, -0.00000498, 0.00008754, 0.00016705, 0.00022784, 0.00026676, 0.00028338, 0.00027966, 0.00025952, 0.00022801, 0.00019060, 0.00015241, 0.00011762, 0.00008911, 0.00006834, 0.00005543, 0.00004940, 0.00004858, 0.00005098, 0.00005468, 0.00005810, 0.00006017, 0.00006037, 0.00005867, 0.00005543, 0.00005117, 0.00004650, 0.00004190, 0.00003774, 0.00003419, 0.00003126, 0.00002887, 0.00002689, 0.00002520, 0.00002371, 0.00002237, 0.00002117, },
		{ 0.00002234, 0.00002211, 0.00002179, 0.00002146, 0.00002122, 0.00002119, 0.00002141, 0.00002191, 0.00002261, 0.00002340, 0.00002413, 0.00002464, 0.00002484, 0.00002471, 0.00002429, 0.00002372, 0.00002316, 0.00002275, 0.00002259, 0.00002263, 0.00002276, 0.00002273, 0.00002224, 0.00002102, 0.00001890, 0.00001585, 0.00001208, 0.00000794, 0.00000394, 0.00000061, -0.00000164, -0.00000260, -0.00000240, -0.00000153, -0.00000078, -0.00000113, -0.00000357, -0.00000882, -0.00001721, -0.00002843, -0.00004157, -0.00005511, -0.00006724, -0.00007612, -0.00008031, -0.00007918, -0.00007319, -0.00006409, -0.00005479, -0.00004913, -0.00005126, -0.00006504, -0.00009319, -0.00013665, -0.00019402, -0.00026137, -0.00033235, -0.00039879, -0.00045158, -0.00048184, -0.00048222, -0.00044806, -0.00037838, -0.00027637, -0.00014937, -0.00000835, 0.00013326, 0.00026104, 0.00036132, 0.00042289, 0.00043846, 0.00040573, 0.00032773, 0.00021265, 0.00007286, -0.00007647, -0.00021916, -0.00033996, -0.00042631, -0.00046970, -0.00046659, -0.00041863, -0.00033222, -0.00021757, -0.00008727, 0.00004529, 0.00016753, 0.00026909, 0.00034278, 0.00038515, 0.00039650, 0.00038038, 0.00034281, 0.00029122, 0.00023337, 0.00017633, 0.00012575, 0.00008536, 0.00005685, 0.00004001, 0.00003315, 0.00003361, 0.00003839, 0.00004462, 0.00005000, 0.00005302, 0.00005302, 0.00005011, 0.00004497, 0.00003857, 0.00003196, 0.00002602, 0.00002137, 0.00001828, 0.00001670, 0.00001637, 0.00001690, 0.00001787, 0.00001895, 0.00001991, 0.00002066, 0.00002118, 0.00002154, 0.00002181, 0.00002204, 0.00002224, 0.00002238, 0.00002242, },
		{ 0.00006207, 0.00005814, 0.00005435, 0.00005061, 0.00004685, 0.00004295, 0.00003886, 0.00003459, 0.00003017, 0.00002573, 0.00002140, 0.00001727, 0.00001344, 0.00000987, 0.00000650, 0.00000318, -0.00000027, -0.00000396, -0.00000795, -0.00001218, -0.00001644, -0.00002045, -0.00002387, -0.00002642, -0.00002793, -0.00002842, -0.00002814, -0.00002755, -0.00002723, -0.00002775, -0.00002955, -0.00003280, -0.00003727, -0.00004232, -0.00004697, -0.00005002, -0.00005031, -0.00004699, -0.00003973, -0.00002893, -0.00001577, -0.00000212, 0.00000976, 0.00001758, 0.00001954, 0.00001479, 0.00000386, -0.00001115, -0.00002665, -0.00003784, -0.00003940, -0.00002631, 0.00000520, 0.00005672, 0.00012693, 0.00021122, 0.00030172, 0.00038788, 0.00045746, 0.00049797, 0.00049823, 0.00045004, 0.00034953, 0.00019816, 0.00000307, -0.00022327, -0.00046405, -0.00069971, -0.00091002, -0.00107619, -0.00118300, -0.00122049, -0.00118507, -0.00107994, -0.00091465, -0.00070397, -0.00046627, -0.00022139, 0.00001145, 0.00021563, 0.00037868, 0.00049323, 0.00055737, 0.00057425, 0.00055122, 0.00049848, 0.00042759, 0.00034997, 0.00027561, 0.00021208, 0.00016409, 0.00013336, 0.00011899, 0.00011805, 0.00012644, 0.00013968, 0.00015362, 0.00016505, 0.00017195, 0.00017358, 0.00017030, 0.00016331, 0.00015422, 0.00014470, 0.00013616, 0.00012954, 0.00012524, 0.00012313, 0.00012273, 0.00012330, 0.00012409, 0.00012446, 0.00012396, 0.00012242, 0.00011985, 0.00011646, 0.00011252, 0.00010831, 0.00010402, 0.00009979, 0.00009565, 0.00009155, 0.00008744, 0.00008327, 0.00007901, 0.00007470, 0.00007039, 0.00006616, },
		{ -0.00001279, -0.00001467, -0.00001646, -0.00001811, -0.00001955, -0.00002072, -0.00002156, -0.00002211, -0.00002244, -0.00002266, -0.00002290, -0.00002328, -0.00002383, -0.00002454, -0.00002530, -0.00002596, -0.00002637, -0.00002643, -0.00002613, -0.00002559, -0.00002507, -0.00002493, -0.00002554, -0.00002722, -0.00003010, -0.00003410, -0.00003889, -0.00004389, -0.00004842, -0.00005182, -0.00005360, -0.00005367, -0.00005240, -0.00005065, -0.00004974, -0.00005119, -0.00005651, -0.00006685, -0.00008266, -0.00010352, -0.00012804, -0.00015395, -0.00017846, -0.00019871, -0.00021244, -0.00021855, -0.00021770, -0.00021258, -0.00020798, -0.00021038, -0.00022728, -0.00026611, -0.00033303, -0.00043168, -0.00056201, -0.00071961, -0.00089540, -0.00107601, -0.00124473, -0.00138307, -0.00147269, -0.00149744, -0.00144546, -0.00131076, -0.00109438, -0.00080464, -0.00045672, -0.00007133, 0.00032720, 0.00071328, 0.00106256, 0.00135427, 0.00157317, 0.00171075, 0.00176580, 0.00174398, 0.00165685, 0.00152015, 0.00135186, 0.00117016, 0.00099149, 0.00082914, 0.00069224, 0.00058542, 0.00050907, 0.00046004, 0.00043267, 0.00042002, 0.00041503, 0.00041150, 0.00040480, 0.00039218, 0.00037282, 0.00034748, 0.00031807, 0.00028707, 0.00025695, 0.00022974, 0.00020673, 0.00018840, 0.00017444, 0.00016393, 0.00015567, 0.00014839, 0.00014099, 0.00013275, 0.00012334, 0.00011281, 0.00010153, 0.00008999, 0.00007874, 0.00006822, 0.00005872, 0.00005032, 0.00004295, 0.00003645, 0.00003059, 0.00002517, 0.00002006, 0.00001520, 0.00001061, 0.00000635, 0.00000250, -0.00000091, -0.00000389, -0.00000647, -0.00000875, -0.00001083, },
		{ -0.00011173, -0.00011181, -0.00011206, -0.00011254, -0.00011335, -0.00011455, -0.00011619, -0.00011824, -0.00012064, -0.00012329, -0.00012608, -0.00012892, -0.00013181, -0.00013480, -0.00013803, -0.00014166, -0.00014584, -0.00015067, -0.00015608, -0.00016189, -0.00016776, -0.00017326, -0.00017795, -0.00018152, -0.00018388, -0.00018520, -0.00018599, -0.00018704, -0.00018924, -0.00019347, -0.00020034, -0.00021001, -0.00022200, -0.00023524, -0.00024809, -0.00025858, -0.00026480, -0.00026527, -0.00025935, -0.00024757, -0.00023170, -0.00021472, -0.00020040, -0.00019268, -0.00019492, -0.00020897, -0.00023441, -0.00026796, -0.00030322, -0.00033092, -0.00033961, -0.00031682, -0.00025056, -0.00013102, 0.00004779, 0.00028667, 0.00058037, 0.00091734, 0.00128028, 0.00164737, 0.00199416, 0.00229591, 0.00252998, 0.00267826, 0.00272904, 0.00267826, 0.00252998, 0.00229591, 0.00199416, 0.00164737, 0.00128028, 0.00091734, 0.00058037, 0.00028667, 0.00004779, -0.00013102, -0.00025056, -0.00031682, -0.00033961, -0.00033092, -0.00030322, -0.00026796, -0.00023441, -0.00020897, -0.00019492, -0.00019268, -0.00020040, -0.00021472, -0.00023170, -0.00024757, -0.00025935, -0.00026527, -0.00026480, -0.00025858, -0.00024809, -0.00023524, -0.00022200, -0.00021001, -0.00020034, -0.00019347, -0.00018924, -0.00018704, -0.00018599, -0.00018520, -0.00018388, -0.00018152, -0.00017795, -0.00017326, -0.00016776, -0.00016189, -0.00015608, -0.00015067, -0.00014584, -0.00014166, -0.00013803, -0.00013480, -0.00013181, -0.00012892, -0.00012608, -0.00012329, -0.00012064, -0.00011824, -0.00011619, -0.00011455, -0.00011335, -0.00011254, -0.00011206, -0.00011181, },
		{ -0.00001279, -0.00001083, -0.00000875, -0.00000647, -0.00000389, -0.00000091, 0.00000250, 0.00000635, 0.00001061, 0.00001520, 0.00002006, 0.00002517, 0.00003059, 0.00003645, 0.00004295, 0.00005032, 0.00005872, 0.00006822, 0.00007874, 0.00008999, 0.00010153, 0.00011281, 0.00012334, 0.00013275, 0.00014099, 0.00014839, 0.00015567, 0.00016393, 0.00017444, 0.00018840, 0.00020673, 0.00022974, 0.00025695, 0.00028707, 0.00031807, 0.00034748, 0.00037282, 0.00039218, 0.00040480, 0.00041150, 0.00041503, 0.00042002, 0.00043267, 0.00046004, 0.00050907, 0.00058542, 0.00069224, 0.00082914, 0.00099149, 0.00117016, 0.00135186, 0.00152015, 0.00165685, 0.00174398, 0.00176580, 0.00171075, 0.00157317, 0.00135427, 0.00106256, 0.00071328, 0.00032720, -0.00007133, -0.00045672, -0.00080464, -0.00109438, -0.00131076, -0.00144546, -0.00149744, -0.00147269, -0.00138307, -0.00124473, -0.00107601, -0.00089540, -0.00071961, -0.00056201, -0.00043168, -0.00033303, -0.00026611, -0.00022728, -0.00021038, -0.00020798, -0.00021258, -0.00021770, -0.00021855, -0.00021244, -0.00019871, -0.00017846, -0.00015395, -0.00012804, -0.00010352, -0.00008266, -0.00006685, -0.00005651, -0.00005119, -0.00004974, -0.00005065, -0.00005240, -0.00005367, -0.00005360, -0.00005182, -0.00004842, -0.00004389, -0.00003889, -0.00003410, -0.00003010, -0.00002722, -0.00002554, -0.00002493, -0.00002507, -0.00002559, -0.00002613, -0.00002643, -0.00002637, -0.00002596, -0.00002530, -0.00002454, -0.00002383, -0.00002328, -0.00002290, -0.00002266, -0.00002244, -0.00002211, -0.00002156, -0.00002072, -0.00001955, -0.00001811, -0.00001646, -0.00001467, },
		{ 0.00006207, 0.00006616, 0.00007039, 0.00007470, 0.00007901, 0.00008327, 0.00008744, 0.00009155, 0.00009565, 0.00009979, 0.00010402, 0.00010831, 0.00011252, 0.00011646, 0.00011985, 0.00012242, 0.00012396, 0.00012446, 0.00012409, 0.00012330, 0.00012273, 0.00012313, 0.00012524, 0.00012954, 0.00013616, 0.00014470, 0.00015422, 0.00016331, 0.00017030, 0.00017358, 0.00017195, 0.00016505, 0.00015362, 0.00013968, 0.00012644, 0.00011805, 0.00011899, 0.00013336, 0.00016409, 0.00021208, 0.00027561, 0.00034997, 0.00042759, 0.00049848, 0.00055122, 0.00057425, 0.00055737, 0.00049323, 0.00037868, 0.00021563, 0.00001145, -0.00022139, -0.00046627, -0.00070397, -0.00091465, -0.00107994, -0.00118507, -0.00122049, -0.00118300, -0.00107619, -0.00091002, -0.00069971, -0.00046405, -0.00022327, 0.00000307, 0.00019816, 0.00034953, 0.00045004, 0.00049823, 0.00049797, 0.00045746, 0.00038788, 0.00030172, 0.00021122, 0.00012693, 0.00005672, 0.00000520, -0.00002631, -0.00003940, -0.00003784, -0.00002665, -0.00001115, 0.00000386, 0.00001479, 0.00001954, 0.00001758, 0.00000976, -0.00000212, -0.00001577, -0.00002893, -0.00003973, -0.00004699, -0.00005031, -0.00005002, -0.00004697, -0.00004232, -0.00003727, -0.00003280, -0.00002955, -0.00002775, -0.00002723, -0.00002755, -0.00002814, -0.00002842, -0.00002793, -0.00002642, -0.00002387, -0.00002045, -0.00001644, -0.00001218, -0.00000795, -0.00000396, -0.00000027, 0.00000318, 0.00000650, 0.00000987, 0.00001344, 0.00001727, 0.00002140, 0.00002573, 0.00003017, 0.00003459, 0.00003886, 0.00004295, 0.00004685, 0.00005061, 0.00005435, 0.00005814, },
		{ 0.00002234, 0.00002242, 0.00002238, 0.00002224, 0.00002204, 0.00002181, 0.00002154, 0.00002118, 0.00002066, 0.00001991, 0.00001895, 0.00001787, 0.00001690, 0.00001637, 0.00001670, 0.00001828, 0.00002137, 0.00002602, 0.00003196, 0.00003857, 0.00004497, 0.00005011, 0.00005302, 0.00005302, 0.00005000, 0.00004462, 0.00003839, 0.00003361, 0.00003315, 0.00004001, 0.00005685, 0.00008536, 0.00012575, 0.00017633, 0.00023337, 0.00029122, 0.00034281, 0.00038038, 0.00039650, 0.00038515, 0.00034278, 0.00026909, 0.00016753, 0.00004529, -0.00008727, -0.00021757, -0.00033222, -0.00041863, -0.00046659, -0.00046970, -0.00042631, -0.00033996, -0.00021916, -0.00007647, 0.00007286, 0.00021265, 0.00032773, 0.00040573, 0.00043846, 0.00042289, 0.00036132, 0.00026104, 0.00013326, -0.00000835, -0.00014937, -0.00027637, -0.00037838, -0.00044806, -0.00048222, -0.00048184, -0.00045158, -0.00039879, -0.00033235, -0.00026137, -0.00019402, -0.00013665, -0.00009319, -0.00006504, -0.00005126, -0.00004913, -0.00005479, -0.00006409, -0.00007319, -0.00007918, -0.00008031, -0.00007612, -0.00006724, -0.00005511, -0.00004157, -0.00002843, -0.00001721, -0.00000882, -0.00000357, -0.00000113, -0.00000078, -0.00000153, -0.00000240, -0.00000260, -0.00000164, 0.00000061, 0.00000394, 0.00000794, 0.00001208, 0.00001585, 0.00001890, 0.00002102, 0.00002224, 0.00002273, 0.00002276, 0.00002263, 0.00002259, 0.00002275, 0.00002316, 0.00002372, 0.00002429, 0.00002471, 0.00002484, 0.00002464, 0.00002413, 0.00002340, 0.00002261, 0.00002191, 0.00002141, 0.00002119, 0.00002122, 0.00002146, 0.00002179, 0.00002211, },
		{ 0.00002012, 0.00002117, 0.00002237, 0.00002371, 0.00002520, 0.00002689, 0.00002887, 0.00003126, 0.00003419, 0.00003774, 0.00004190, 0.00004650, 0.00005117, 0.00005543, 0.00005867, 0.00006037, 0.00006017, 0.00005810, 0.00005468, 0.00005098, 0.00004858, 0.00004940, 0.00005543, 0.00006834, 0.00008911, 0.00011762, 0.00015241, 0.00019060, 0.00022801, 0.00025952, 0.00027966, 0.00028338, 0.00026676, 0.00022784, 0.00016705, 0.00008754, -0.00000498, -0.00010273, -0.00019669, -0.00027766, -0.00033752, -0.00037030, -0.00037312, -0.00034668, -0.00029536, -0.00022675, -0.00015072, -0.00007819, -0.00001966, 0.00001623, 0.00002387, 0.00000155, -0.00004833, -0.00011944, -0.00020242, -0.00028606, -0.00035883, -0.00041037, -0.00043287, -0.00042208, -0.00037780, -0.00030383, -0.00020742, -0.00009824, 0.00001293, 0.00011557, 0.00020070, 0.00026190, 0.00029593, 0.00030287, 0.00028579, 0.00025011, 0.00020268, 0.00015071, 0.00010082, 0.00005827, 0.00002642, 0.00000658, -0.00000185, -0.00000099, 0.00000611, 0.00001605, 0.00002572, 0.00003274, 0.00003568, 0.00003420, 0.00002885, 0.00002087, 0.00001182, 0.00000322, -0.00000369, -0.00000815, -0.00000998, -0.00000948, -0.00000735, -0.00000444, -0.00000160, 0.00000050, 0.00000148, 0.00000126, 0.00000005, -0.00000175, -0.00000367, -0.00000529, -0.00000631, -0.00000659, -0.00000620, -0.00000531, -0.00000419, -0.00000307, -0.00000215, -0.00000149, -0.00000107, -0.00000073, -0.00000031, 0.00000038, 0.00000147, 0.00000299, 0.00000490, 0.00000707, 0.00000932, 0.00001148, 0.00001341, 0.00001502, 0.00001633, 0.00001738, 0.00001830, 0.00001918, },
		{ 0.00001995, 0.00002129, 0.00002271, 0.00002420, 0.00002566, 0.00002698, 0.00002797, 0.00002841, 0.00002808, 0.00002692, 0.00002502, 0.00002278, 0.00002093, 0.00002048, 0.00002266, 0.00002872, 0.00003968, 0.00005606, 0.00007764, 0.00010327, 0.00013081, 0.00015723, 0.00017888, 0.00019191, 0.00019282, 0.00017901, 0.00014930, 0.00010431, 0.00004661, -0.00001941, -0.00008792, -0.00015233, -0.00020607, -0.00024350, -0.00026072, -0.00025613, -0.00023083, -0.00018846, -0.00013487, -0.00007724, -0.00002318, 0.00002039, 0.00004826, 0.00005770, 0.00004888, 0.00002486, -0.00000883, -0.00004505, -0.00007607, -0.00009484, -0.00009604, -0.00007697, -0.00003802, 0.00001728, 0.00008272, 0.00015024, 0.00021106, 0.00025688, 0.00028109, 0.00027964, 0.00025165, 0.00019945, 0.00012828, 0.00004554, -0.00004023, -0.00012046, -0.00018767, -0.00023632, -0.00026340, -0.00026861, -0.00025421, -0.00022444, -0.00018483, -0.00014130, -0.00009934, -0.00006336, -0.00003621, -0.00001903, -0.00001132, -0.00001131, -0.00001642, -0.00002378, -0.00003075, -0.00003533, -0.00003634, -0.00003353, -0.00002743, -0.00001914, -0.00001004, -0.00000146, 0.00000555, 0.00001034, 0.00001280, 0.00001323, 0.00001224, 0.00001059, 0.00000902, 0.00000809, 0.00000809, 0.00000906, 0.00001078, 0.00001288, 0.00001498, 0.00001671, 0.00001788, 0.00001843, 0.00001848, 0.00001824, 0.00001799, 0.00001794, 0.00001825, 0.00001893, 0.00001989, 0.00002094, 0.00002185, 0.00002242, 0.00002251, 0.00002208, 0.00002120, 0.00002003, 0.00001878, 0.00001766, 0.00001685, 0.00001644, 0.00001648, 0.00001693, 0.00001771, 0.00001875, },
		{ 0.00000817, 0.00000887, 0.00000944, 0.00001004, 0.00001093, 0.00001252, 0.00001538, 0.00002018, 0.00002759, 0.00003814, 0.00005205, 0.00006907, 0.00008832, 0.00010830, 0.00012688, 0.00014152, 0.00014956, 0.00014857, 0.00013679, 0.00011347, 0.00007913, 0.00003572, -0.00001350, -0.00006425, -0.00011173, -0.00015119, -0.00017859, -0.00019122, -0.00018808, -0.00017018, -0.00014036, -0.00010302, -0.00006349, -0.00002728, 0.00000073, 0.00001700, 0.00001987, 0.00000979, -0.00001071, -0.00003751, -0.00006551, -0.00008949, -0.00010500, -0.00010910, -0.00010088, -0.00008165, -0.00005482, -0.00002531, 0.00000117, 0.00001908, 0.00002397, 0.00001327, -0.00001322, -0.00005318, -0.00010203, -0.00015353, -0.00020064, -0.00023652, -0.00025553, -0.00025406, -0.00023100, -0.00018795, -0.00012898, -0.00006005, 0.00001180, 0.00007941, 0.00013641, 0.00017802, 0.00020158, 0.00020670, 0.00019519, 0.00017059, 0.00013759, 0.00010122, 0.00006622, 0.00003637, 0.00001414, 0.00000051, -0.00000497, -0.00000384, 0.00000162, 0.00000892, 0.00001576, 0.00002040, 0.00002182, 0.00001984, 0.00001494, 0.00000811, 0.00000056, -0.00000654, -0.00001229, -0.00001616, -0.00001806, -0.00001829, -0.00001740, -0.00001605, -0.00001484, -0.00001422, -0.00001437, -0.00001524, -0.00001660, -0.00001808, -0.00001932, -0.00002005, -0.00002013, -0.00001960, -0.00001863, -0.00001751, -0.00001652, -0.00001589, -0.00001573, -0.00001603, -0.00001661, -0.00001725, -0.00001765, -0.00001755, -0.00001678, -0.00001529, -0.00001313, -0.00001048, -0.00000755, -0.00000458, -0.00000177, 0.00000074, 0.00000290, 0.00000468, 0.00000612, 0.00000727, },
		{ 0.00002992, 0.00003611, 0.00004408, 0.00005384, 0.00006516, 0.00007746, 0.00008976, 0.00010069, 0.00010861, 0.00011177, 0.00010856, 0.00009776, 0.00007889, 0.00005231, 0.00001938, -0.00001757, -0.00005550, -0.00009095, -0.00012046, -0.00014110, -0.00015085, -0.00014900, -0.00013624, -0.00011465, -0.00008746, -0.00005857, -0.00003201, -0.00001133, 0.00000093, 0.00000363, -0.00000282, -0.00001651, -0.00003435, -0.00005263, -0.00006761, -0.00007619, -0.00007643, -0.00006789, -0.00005174, -0.00003054, -0.00000783, 0.00001252, 0.00002697, 0.00003299, 0.00002952, 0.00001727, -0.00000139, -0.00002271, -0.00004220, -0.00005535, -0.00005842, -0.00004910, -0.00002698, 0.00000624, 0.00004694, 0.00009003, 0.00012965, 0.00016003, 0.00017631, 0.00017528, 0.00015585, 0.00011927, 0.00006888, 0.00000974, -0.00005210, -0.00011046, -0.00015976, -0.00019580, -0.00021616, -0.00022045, -0.00021019, -0.00018842, -0.00015922, -0.00012703, -0.00009599, -0.00006946, -0.00004960, -0.00003729, -0.00003214, -0.00003280, -0.00003725, -0.00004329, -0.00004891, -0.00005258, -0.00005345, -0.00005136, -0.00004677, -0.00004054, -0.00003373, -0.00002734, -0.00002211, -0.00001844, -0.00001634, -0.00001549, -0.00001538, -0.00001542, -0.00001511, -0.00001413, -0.00001239, -0.00001002, -0.00000735, -0.00000474, -0.00000253, -0.00000097, -0.00000011, 0.00000016, 0.00000012, 0.00000010, 0.00000045, 0.00000142, 0.00000313, 0.00000553, 0.00000844, 0.00001154, 0.00001449, 0.00001698, 0.00001880, 0.00001986, 0.00002019, 0.00001996, 0.00001940, 0.00001876, 0.00001829, 0.00001821, 0.00001871, 0.00001993, 0.00002207, 0.00002532, },
		{ 0.00007226, 0.00007592, 0.00007639, 0.00007266, 0.00006392, 0.00004980, 0.00003045, 0.00000672, -0.00001990, -0.00004736, -0.00007325, -0.00009512, -0.00011083, -0.00011882, -0.00011847, -0.00011014, -0.00009524, -0.00007604, -0.00005534, -0.00003608, -0.00002091, -0.00001177, -0.00000955, -0.00001404, -0.00002388, -0.00003688, -0.00005033, -0.00006153, -0.00006824, -0.00006906, -0.00006371, -0.00005308, -0.00003906, -0.00002424, -0.00001138, -0.00000297, -0.00000069, -0.00000514, -0.00001566, -0.00003045, -0.00004681, -0.00006170, -0.00007223, -0.00007624, -0.00007274, -0.00006215, -0.00004632, -0.00002825, -0.00001165, -0.00000032, 0.00000252, -0.00000516, -0.00002379, -0.00005203, -0.00008683, -0.00012384, -0.00015799, -0.00018422, -0.00019822, -0.00019709, -0.00017977, -0.00014722, -0.00010231, -0.00004944, 0.00000612, 0.00005886, 0.00010384, 0.00013726, 0.00015692, 0.00016240, 0.00015498, 0.00013732, 0.00011302, 0.00008598, 0.00005988, 0.00003770, 0.00002136, 0.00001163, 0.00000816, 0.00000973, 0.00001457, 0.00002076, 0.00002652, 0.00003057, 0.00003220, 0.00003134, 0.00002843, 0.00002429, 0.00001985, 0.00001593, 0.00001316, 0.00001177, 0.00001169, 0.00001253, 0.00001378, 0.00001489, 0.00001544, 0.00001521, 0.00001422, 0.00001272, 0.00001108, 0.00000975, 0.00000908, 0.00000928, 0.00001038, 0.00001217, 0.00001432, 0.00001640, 0.00001799, 0.00001880, 0.00001869, 0.00001772, 0.00001610, 0.00001421, 0.00001245, 0.00001124, 0.00001088, 0.00001158, 0.00001341, 0.00001637, 0.00002035, 0.00002526, 0.00003098, 0.00003742, 0.00004446, 0.00005190, 0.00005940, 0.00006644, },
		{ -0.00000125, -0.00001930, -0.00003821, -0.00005641, -0.00007224, -0.00008410, -0.00009077, -0.00009159, -0.00008661, -0.00007661, -0.00006306, -0.00004788, -0.00003319, -0.00002097, -0.00001274, -0.00000932, -0.00001072, -0.00001609, -0.00002395, -0.00003237, -0.00003938, -0.00004331, -0.00004307, -0.00003838, -0.00002982, -0.00001871, -0.00000688, 0.00000367, 0.00001117, 0.00001438, 0.00001292, 0.00000727, -0.00000123, -0.00001064, -0.00001879, -0.00002365, -0.00002380, -0.00001865, -0.00000864, 0.00000481, 0.00001954, 0.00003301, 0.00004280, 0.00004703, 0.00004481, 0.00003646, 0.00002349, 0.00000849, -0.00000533, -0.00001465, -0.00001659, -0.00000929, 0.00000773, 0.00003337, 0.00006504, 0.00009894, 0.00013059, 0.00015547, 0.00016971, 0.00017063, 0.00015719, 0.00013014, 0.00009195, 0.00004645, -0.00000166, -0.00004750, -0.00008664, -0.00011567, -0.00013259, -0.00013700, -0.00013002, -0.00011402, -0.00009218, -0.00006799, -0.00004472, -0.00002497, -0.00001043, -0.00000172, 0.00000152, 0.00000041, -0.00000343, -0.00000829, -0.00001264, -0.00001539, -0.00001601, -0.00001452, -0.00001141, -0.00000747, -0.00000354, -0.00000034, 0.00000166, 0.00000235, 0.00000193, 0.00000086, -0.00000029, -0.00000099, -0.00000085, 0.00000026, 0.00000223, 0.00000470, 0.00000720, 0.00000922, 0.00001037, 0.00001041, 0.00000940, 0.00000758, 0.00000541, 0.00000341, 0.00000213, 0.00000197, 0.00000317, 0.00000577, 0.00000959, 0.00001433, 0.00001962, 0.00002508, 0.00003040, 0.00003535, 0.00003977, 0.00004356, 0.00004655, 0.00004851, 0.00004908, 0.00004780, 0.00004412, 0.00003756, 0.00002776, 0.00001471, },
		{ -0.00005882, -0.00005685, -0.00005108, -0.00004230, -0.00003172, -0.00002081, -0.00001106, -0.00000373, 0.00000036, 0.00000096, -0.00000155, -0.00000627, -0.00001192, -0.00001707, -0.00002046, -0.00002118, -0.00001888, -0.00001383, -0.00000688, 0.00000071, 0.00000754, 0.00001228, 0.00001403, 0.00001241, 0.00000773, 0.00000089, -0.00000674, -0.00001362, -0.00001831, -0.00001978, -0.00001766, -0.00001227, -0.00000466, 0.00000359, 0.00001070, 0.00001495, 0.00001511, 0.00001066, 0.00000194, -0.00000986, -0.00002286, -0.00003485, -0.00004368, -0.00004765, -0.00004591, -0.00003864, -0.00002713, -0.00001360, -0.00000088, 0.00000807, 0.00001065, 0.00000511, -0.00000905, -0.00003097, -0.00005838, -0.00008797, -0.00011577, -0.00013775, -0.00015040, -0.00015129, -0.00013945, -0.00011550, -0.00008163, -0.00004125, 0.00000143, 0.00004205, 0.00007661, 0.00010208, 0.00011667, 0.00012006, 0.00011330, 0.00009857, 0.00007877, 0.00005705, 0.00003634, 0.00001895, 0.00000632, -0.00000106, -0.00000362, -0.00000245, 0.00000095, 0.00000502, 0.00000842, 0.00001025, 0.00001013, 0.00000824, 0.00000513, 0.00000159, -0.00000155, -0.00000367, -0.00000441, -0.00000378, -0.00000212, 0.00000002, 0.00000200, 0.00000326, 0.00000344, 0.00000249, 0.00000062, -0.00000168, -0.00000379, -0.00000506, -0.00000499, -0.00000327, 0.00000009, 0.00000484, 0.00001051, 0.00001650, 0.00002221, 0.00002714, 0.00003097, 0.00003358, 0.00003501, 0.00003542, 0.00003499, 0.00003384, 0.00003194, 0.00002913, 0.00002513, 0.00001962, 0.00001236, 0.00000329, -0.00000737, -0.00001906, -0.00003093, -0.00004191, -0.00005085, -0.00005672, },
		{ -0.00001787, -0.00001369, -0.00001190, -0.00001253, -0.00001518, -0.00001907, -0.00002320, -0.00002660, -0.00002843, -0.00002826, -0.00002606, -0.00002225, -0.00001761, -0.00001310, -0.00000969, -0.00000810, -0.00000871, -0.00001143, -0.00001570, -0.00002062, -0.00002511, -0.00002814, -0.00002894, -0.00002719, -0.00002308, -0.00001732, -0.00001102, -0.00000543, -0.00000177, -0.00000091, -0.00000319, -0.00000834, -0.00001547, -0.00002321, -0.00002997, -0.00003424, -0.00003487, -0.00003134, -0.00002389, -0.00001352, -0.00000185, 0.00000917, 0.00001760, 0.00002189, 0.00002119, 0.00001562, 0.00000622, -0.00000506, -0.00001576, -0.00002322, -0.00002512, -0.00001988, -0.00000701, 0.00001273, 0.00003737, 0.00006390, 0.00008878, 0.00010835, 0.00011946, 0.00011991, 0.00010884, 0.00008682, 0.00005588, 0.00001915, -0.00001952, -0.00005614, -0.00008708, -0.00010959, -0.00012210, -0.00012438, -0.00011746, -0.00010340, -0.00008492, -0.00006491, -0.00004603, -0.00003034, -0.00001908, -0.00001257, -0.00001029, -0.00001115, -0.00001370, -0.00001648, -0.00001831, -0.00001844, -0.00001667, -0.00001330, -0.00000899, -0.00000459, -0.00000090, 0.00000149, 0.00000235, 0.00000182, 0.00000041, -0.00000116, -0.00000213, -0.00000182, 0.00000016, 0.00000388, 0.00000905, 0.00001508, 0.00002121, 0.00002667, 0.00003078, 0.00003314, 0.00003363, 0.00003244, 0.00002996, 0.00002672, 0.00002324, 0.00001986, 0.00001676, 0.00001381, 0.00001071, 0.00000704, 0.00000239, -0.00000351, -0.00001069, -0.00001884, -0.00002737, -0.00003548, -0.00004226, -0.00004689, -0.00004879, -0.00004775, -0.00004397, -0.00003809, -0.00003104, -0.00002394, },
		{ -0.00001468, -0.00001664, -0.00001746, -0.00001684, -0.00001484, -0.00001186, -0.00000849, -0.00000543, -0.00000333, -0.00000258, -0.00000331, -0.00000529, -0.00000799, -0.00001071, -0.00001272, -0.00001342, -0.00001251, -0.00001003, -0.00000640, -0.00000235, 0.00000125, 0.00000354, 0.00000388, 0.00000198, -0.00000198, -0.00000741, -0.00001333, -0.00001864, -0.00002223, -0.00002329, -0.00002148, -0.00001699, -0.00001055, -0.00000334, 0.00000321, 0.00000776, 0.00000923, 0.00000711, 0.00000153, -0.00000667, -0.00001611, -0.00002507, -0.00003185, -0.00003505, -0.00003392, -0.00002853, -0.00001982, -0.00000951, 0.00000016, 0.00000682, 0.00000835, 0.00000334, -0.00000866, -0.00002692, -0.00004960, -0.00007396, -0.00009670, -0.00011445, -0.00012431, -0.00012426, -0.00011352, -0.00009267, -0.00006360, -0.00002927, 0.00000675, 0.00004073, 0.00006932, 0.00009000, 0.00010139, 0.00010334, 0.00009691, 0.00008410, 0.00006748, 0.00004978, 0.00003346, 0.00002036, 0.00001152, 0.00000709, 0.00000647, 0.00000848, 0.00001168, 0.00001469, 0.00001644, 0.00001636, 0.00001440, 0.00001107, 0.00000720, 0.00000377, 0.00000166, 0.00000147, 0.00000339, 0.00000714, 0.00001208, 0.00001728, 0.00002178, 0.00002473, 0.00002557, 0.00002416, 0.00002074, 0.00001591, 0.00001047, 0.00000524, 0.00000093, -0.00000207, -0.00000370, -0.00000427, -0.00000435, -0.00000467, -0.00000592, -0.00000860, -0.00001290, -0.00001862, -0.00002522, -0.00003188, -0.00003768, -0.00004177, -0.00004351, -0.00004262, -0.00003926, -0.00003395, -0.00002753, -0.00002096, -0.00001521, -0.00001100, -0.00000874, -0.00000847, -0.00000982, -0.00001215, },
		{ -0.00000951, -0.00000774, -0.00000661, -0.00000641, -0.00000721, -0.00000880, -0.00001077, -0.00001261, -0.00001385, -0.00001414, -0.00001337, -0.00001172, -0.00000957, -0.00000750, -0.00000608, -0.00000583, -0.00000698, -0.00000951, -0.00001304, -0.00001694, -0.00002044, -0.00002275, -0.00002328, -0.00002173, -0.00001821, -0.00001321, -0.00000757, -0.00000227, 0.00000170, 0.00000359, 0.00000303, 0.00000014, -0.00000449, -0.00000985, -0.00001473, -0.00001792, -0.00001849, -0.00001598, -0.00001050, -0.00000281, 0.00000586, 0.00001397, 0.00001997, 0.00002263, 0.00002127, 0.00001598, 0.00000766, -0.00000209, -0.00001117, -0.00001736, -0.00001870, -0.00001385, -0.00000242, 0.00001493, 0.00003645, 0.00005957, 0.00008118, 0.00009812, 0.00010766, 0.00010791, 0.00009815, 0.00007896, 0.00005215, 0.00002054, -0.00001249, -0.00004340, -0.00006908, -0.00008718, -0.00009649, -0.00009695, -0.00008969, -0.00007669, -0.00006047, -0.00004363, -0.00002847, -0.00001668, -0.00000907, -0.00000562, -0.00000553, -0.00000749, -0.00000998, -0.00001157, -0.00001124, -0.00000851, -0.00000352, 0.00000301, 0.00001003, 0.00001632, 0.00002080, 0.00002271, 0.00002179, 0.00001832, 0.00001303, 0.00000696, 0.00000124, -0.00000312, -0.00000546, -0.00000556, -0.00000369, -0.00000057, 0.00000283, 0.00000547, 0.00000642, 0.00000510, 0.00000138, -0.00000442, -0.00001153, -0.00001892, -0.00002546, -0.00003017, -0.00003235, -0.00003172, -0.00002848, -0.00002325, -0.00001693, -0.00001057, -0.00000512, -0.00000135, 0.00000037, 0.00000005, -0.00000195, -0.00000498, -0.00000828, -0.00001114, -0.00001300, -0.00001359, -0.00001297, -0.00001144, },
		{ -0.00001204, -0.00001288, -0.00001336, -0.00001322, -0.00001239, -0.00001101, -0.00000939, -0.00000792, -0.00000700, -0.00000690, -0.00000771, -0.00000929, -0.00001129, -0.00001322, -0.00001452, -0.00001475, -0.00001365, -0.00001120, -0.00000771, -0.00000370, 0.00000013, 0.00000310, 0.00000464, 0.00000445, 0.00000257, -0.00000061, -0.00000440, -0.00000798, -0.00001048, -0.00001127, -0.00001003, -0.00000686, -0.00000231, 0.00000272, 0.00000714, 0.00000988, 0.00001011, 0.00000744, 0.00000202, -0.00000544, -0.00001375, -0.00002146, -0.00002710, -0.00002951, -0.00002804, -0.00002279, -0.00001460, -0.00000500, 0.00000403, 0.00001036, 0.00001211, 0.00000801, -0.00000234, -0.00001832, -0.00003829, -0.00005977, -0.00007982, -0.00009543, -0.00010398, -0.00010370, -0.00009392, -0.00007521, -0.00004936, -0.00001910, 0.00001228, 0.00004140, 0.00006527, 0.00008170, 0.00008962, 0.00008911, 0.00008140, 0.00006856, 0.00005316, 0.00003780, 0.00002475, 0.00001559, 0.00001099, 0.00001073, 0.00001380, 0.00001863, 0.00002349, 0.00002676, 0.00002729, 0.00002456, 0.00001876, 0.00001073, 0.00000172, -0.00000681, -0.00001352, -0.00001745, -0.00001820, -0.00001600, -0.00001165, -0.00000635, -0.00000149, 0.00000165, 0.00000209, -0.00000059, -0.00000625, -0.00001416, -0.00002321, -0.00003203, -0.00003929, -0.00004394, -0.00004535, -0.00004345, -0.00003868, -0.00003193, -0.00002436, -0.00001716, -0.00001137, -0.00000771, -0.00000647, -0.00000750, -0.00001025, -0.00001395, -0.00001773, -0.00002082, -0.00002266, -0.00002300, -0.00002192, -0.00001979, -0.00001713, -0.00001450, -0.00001241, -0.00001115, -0.00001080, -0.00001121, },
		{ -0.00000397, -0.00000327, -0.00000290, -0.00000306, -0.00000382, -0.00000506, -0.00000651, -0.00000781, -0.00000861, -0.00000862, -0.00000774, -0.00000606, -0.00000386, -0.00000156, 0.00000036, 0.00000148, 0.00000152, 0.00000045, -0.00000154, -0.00000402, -0.00000643, -0.00000821, -0.00000887, -0.00000817, -0.00000617, -0.00000322, 0.00000007, 0.00000295, 0.00000469, 0.00000472, 0.00000280, -0.00000094, -0.00000593, -0.00001132, -0.00001606, -0.00001911, -0.00001969, -0.00001741, -0.00001244, -0.00000545, 0.00000241, 0.00000974, 0.00001514, 0.00001746, 0.00001607, 0.00001102, 0.00000313, -0.00000615, -0.00001491, -0.00002111, -0.00002292, -0.00001908, -0.00000921, 0.00000610, 0.00002528, 0.00004592, 0.00006516, 0.00008006, 0.00008808, 0.00008748, 0.00007763, 0.00005915, 0.00003382, 0.00000438, -0.00002588, -0.00005356, -0.00007570, -0.00009016, -0.00009590, -0.00009312, -0.00008317, -0.00006829, -0.00005121, -0.00003472, -0.00002123, -0.00001241, -0.00000895, -0.00001057, -0.00001609, -0.00002374, -0.00003149, -0.00003746, -0.00004021, -0.00003905, -0.00003408, -0.00002616, -0.00001673, -0.00000751, -0.00000017, 0.00000396, 0.00000416, 0.00000039, -0.00000666, -0.00001574, -0.00002525, -0.00003355, -0.00003925, -0.00004141, -0.00003970, -0.00003446, -0.00002656, -0.00001728, -0.00000803, -0.00000014, 0.00000539, 0.00000802, 0.00000774, 0.00000501, 0.00000067, -0.00000427, -0.00000880, -0.00001211, -0.00001370, -0.00001345, -0.00001161, -0.00000870, -0.00000537, -0.00000230, -0.00000002, 0.00000117, 0.00000121, 0.00000030, -0.00000118, -0.00000279, -0.00000414, -0.00000494, -0.00000509, -0.00000469, },
		{ -0.00000554, -0.00000655, -0.00000723, -0.00000738, -0.00000690, -0.00000585, -0.00000446, -0.00000303, -0.00000188, -0.00000129, -0.00000139, -0.00000215, -0.00000340, -0.00000479, -0.00000597, -0.00000660, -0.00000647, -0.00000556, -0.00000406, -0.00000235, -0.00000090, -0.00000021, -0.00000066, -0.00000244, -0.00000543, -0.00000927, -0.00001334, -0.00001690, -0.00001924, -0.00001983, -0.00001842, -0.00001517, -0.00001059, -0.00000555, -0.00000107, 0.00000186, 0.00000245, 0.00000032, -0.00000440, -0.00001106, -0.00001856, -0.00002553, -0.00003059, -0.00003263, -0.00003100, -0.00002575, -0.00001766, -0.00000816, 0.00000088, 0.00000740, 0.00000959, 0.00000617, -0.00000329, -0.00001824, -0.00003712, -0.00005754, -0.00007662, -0.00009138, -0.00009921, -0.00009831, -0.00008797, -0.00006877, -0.00004249, -0.00001192, 0.00001954, 0.00004837, 0.00007143, 0.00008641, 0.00009214, 0.00008875, 0.00007759, 0.00006099, 0.00004189, 0.00002333, 0.00000799, -0.00000217, -0.00000623, -0.00000436, 0.00000227, 0.00001176, 0.00002180, 0.00003015, 0.00003500, 0.00003522, 0.00003061, 0.00002181, 0.00001023, -0.00000229, -0.00001378, -0.00002249, -0.00002718, -0.00002733, -0.00002316, -0.00001560, -0.00000606, 0.00000380, 0.00001240, 0.00001845, 0.00002119, 0.00002047, 0.00001673, 0.00001086, 0.00000403, -0.00000252, -0.00000775, -0.00001093, -0.00001181, -0.00001052, -0.00000762, -0.00000388, -0.00000012, 0.00000291, 0.00000468, 0.00000499, 0.00000392, 0.00000184, -0.00000075, -0.00000330, -0.00000535, -0.00000658, -0.00000693, -0.00000651, -0.00000561, -0.00000460, -0.00000382, -0.00000351, -0.00000377, -0.00000452, },
		{ -0.00000641, -0.00000502, -0.00000373, -0.00000278, -0.00000230, -0.00000230, -0.00000265, -0.00000314, -0.00000355, -0.00000366, -0.00000337, -0.00000272, -0.00000188, -0.00000110, -0.00000070, -0.00000091, -0.00000188, -0.00000357, -0.00000575, -0.00000803, -0.00000991, -0.00001092, -0.00001069, -0.00000905, -0.00000610, -0.00000223, 0.00000196, 0.00000575, 0.00000841, 0.00000943, 0.00000855, 0.00000591, 0.00000201, -0.00000231, -0.00000605, -0.00000824, -0.00000809, -0.00000523, 0.00000022, 0.00000760, 0.00001579, 0.00002343, 0.00002909, 0.00003161, 0.00003030, 0.00002517, 0.00001695, 0.00000709, -0.00000249, -0.00000968, -0.00001251, -0.00000957, -0.00000027, 0.00001497, 0.00003468, 0.00005650, 0.00007748, 0.00009452, 0.00010479, 0.00010623, 0.00009788, 0.00008006, 0.00005440, 0.00002361, -0.00000887, -0.00003934, -0.00006441, -0.00008142, -0.00008887, -0.00008657, -0.00007567, -0.00005842, -0.00003781, -0.00001709, 0.00000069, 0.00001318, 0.00001902, 0.00001798, 0.00001095, -0.00000028, -0.00001336, -0.00002580, -0.00003538, -0.00004051, -0.00004042, -0.00003529, -0.00002611, -0.00001448, -0.00000234, 0.00000843, 0.00001629, 0.00002029, 0.00002017, 0.00001635, 0.00000986, 0.00000205, -0.00000563, -0.00001189, -0.00001582, -0.00001700, -0.00001554, -0.00001202, -0.00000732, -0.00000245, 0.00000165, 0.00000428, 0.00000508, 0.00000408, 0.00000164, -0.00000165, -0.00000509, -0.00000804, -0.00001001, -0.00001079, -0.00001040, -0.00000912, -0.00000737, -0.00000561, -0.00000427, -0.00000362, -0.00000375, -0.00000454, -0.00000575, -0.00000703, -0.00000804, -0.00000853, -0.00000838, -0.00000762, },
		{ 0.00000518, 0.00000489, 0.00000446, 0.00000406, 0.00000382, 0.00000375, 0.00000377, 0.00000372, 0.00000346, 0.00000285, 0.00000186, 0.00000059, -0.00000077, -0.00000194, -0.00000262, -0.00000256, -0.00000163, 0.00000011, 0.00000242, 0.00000491, 0.00000709, 0.00000847, 0.00000869, 0.00000757, 0.00000522, 0.00000201, -0.00000147, -0.00000451, -0.00000640, -0.00000664, -0.00000497, -0.00000156, 0.00000305, 0.00000802, 0.00001229, 0.00001483, 0.00001481, 0.00001181, 0.00000591, -0.00000221, -0.00001141, -0.00002023, -0.00002711, -0.00003072, -0.00003019, -0.00002536, -0.00001682, -0.00000595, 0.00000530, 0.00001472, 0.00002012, 0.00001982, 0.00001287, -0.00000061, -0.00001946, -0.00004150, -0.00006387, -0.00008335, -0.00009693, -0.00010217, -0.00009769, -0.00008335, -0.00006038, -0.00003116, 0.00000100, 0.00003241, 0.00005944, 0.00007911, 0.00008943, 0.00008971, 0.00008059, 0.00006397, 0.00004260, 0.00001972, -0.00000146, -0.00001821, -0.00002868, -0.00003211, -0.00002883, -0.00002020, -0.00000827, 0.00000459, 0.00001607, 0.00002428, 0.00002806, 0.00002707, 0.00002181, 0.00001344, 0.00000354, -0.00000617, -0.00001418, -0.00001939, -0.00002127, -0.00001994, -0.00001603, -0.00001058, -0.00000480, 0.00000019, 0.00000351, 0.00000468, 0.00000368, 0.00000090, -0.00000298, -0.00000712, -0.00001070, -0.00001309, -0.00001394, -0.00001322, -0.00001122, -0.00000842, -0.00000545, -0.00000286, -0.00000111, -0.00000041, -0.00000076, -0.00000190, -0.00000345, -0.00000496, -0.00000603, -0.00000635, -0.00000581, -0.00000448, -0.00000258, -0.00000042, 0.00000167, 0.00000338, 0.00000455, 0.00000513, },
		{ -0.00000405, -0.00000471, -0.00000532, -0.00000599, -0.00000675, -0.00000757, -0.00000836, -0.00000894, -0.00000916, -0.00000891, -0.00000819, -0.00000708, -0.00000579, -0.00000463, -0.00000388, -0.00000379, -0.00000448, -0.00000590, -0.00000781, -0.00000982, -0.00001144, -0.00001222, -0.00001180, -0.00001005, -0.00000709, -0.00000335, 0.00000052, 0.00000376, 0.00000560, 0.00000548, 0.00000312, -0.00000134, -0.00000733, -0.00001392, -0.00001994, -0.00002419, -0.00002568, -0.00002378, -0.00001841, -0.00001012, -0.00000002, 0.00001037, 0.00001937, 0.00002537, 0.00002721, 0.00002437, 0.00001712, 0.00000655, -0.00000555, -0.00001697, -0.00002541, -0.00002887, -0.00002602, -0.00001646, -0.00000087, 0.00001902, 0.00004066, 0.00006100, 0.00007693, 0.00008579, 0.00008573, 0.00007609, 0.00005746, 0.00003168, 0.00000160, -0.00002937, -0.00005767, -0.00008012, -0.00009438, -0.00009922, -0.00009469, -0.00008206, -0.00006358, -0.00004213, -0.00002076, -0.00000223, 0.00001133, 0.00001873, 0.00001982, 0.00001539, 0.00000701, -0.00000333, -0.00001352, -0.00002173, -0.00002668, -0.00002774, -0.00002508, -0.00001946, -0.00001212, -0.00000446, 0.00000216, 0.00000671, 0.00000863, 0.00000788, 0.00000492, 0.00000055, -0.00000421, -0.00000839, -0.00001120, -0.00001218, -0.00001126, -0.00000875, -0.00000525, -0.00000148, 0.00000183, 0.00000409, 0.00000495, 0.00000439, 0.00000264, 0.00000016, -0.00000248, -0.00000472, -0.00000612, -0.00000642, -0.00000561, -0.00000389, -0.00000160, 0.00000082, 0.00000293, 0.00000441, 0.00000506, 0.00000486, 0.00000394, 0.00000252, 0.00000089, -0.00000071, -0.00000211, -0.00000321, },
		{ -0.00000516, -0.00000514, -0.00000502, -0.00000471, -0.00000420, -0.00000351, -0.00000279, -0.00000217, -0.00000181, -0.00000180, -0.00000217, -0.00000279, -0.00000347, -0.00000393, -0.00000389, -0.00000314, -0.00000162, 0.00000058, 0.00000315, 0.00000562, 0.00000746, 0.00000812, 0.00000722, 0.00000460, 0.00000040, -0.00000491, -0.00001058, -0.00001569, -0.00001932, -0.00002070, -0.00001935, -0.00001528, -0.00000893, -0.00000122, 0.00000662, 0.00001321, 0.00001731, 0.00001796, 0.00001477, 0.00000795, -0.00000164, -0.00001261, -0.00002327, -0.00003187, -0.00003693, -0.00003752, -0.00003343, -0.00002529, -0.00001445, -0.00000288, 0.00000724, 0.00001379, 0.00001513, 0.00001040, -0.00000032, -0.00001590, -0.00003437, -0.00005310, -0.00006923, -0.00008009, -0.00008365, -0.00007876, -0.00006546, -0.00004492, -0.00001932, 0.00000845, 0.00003522, 0.00005799, 0.00007431, 0.00008267, 0.00008267, 0.00007498, 0.00006128, 0.00004390, 0.00002546, 0.00000850, -0.00000494, -0.00001352, -0.00001682, -0.00001524, -0.00000992, -0.00000248, 0.00000532, 0.00001186, 0.00001593, 0.00001689, 0.00001474, 0.00001004, 0.00000381, -0.00000278, -0.00000855, -0.00001256, -0.00001430, -0.00001368, -0.00001107, -0.00000719, -0.00000291, 0.00000087, 0.00000341, 0.00000430, 0.00000344, 0.00000110, -0.00000217, -0.00000570, -0.00000878, -0.00001082, -0.00001148, -0.00001066, -0.00000856, -0.00000560, -0.00000233, 0.00000071, 0.00000303, 0.00000431, 0.00000445, 0.00000356, 0.00000189, -0.00000019, -0.00000231, -0.00000414, -0.00000547, -0.00000623, -0.00000646, -0.00000631, -0.00000597, -0.00000560, -0.00000533, -0.00000519, },
		{ -0.00000539, -0.00000522, -0.00000501, -0.00000484, -0.00000472, -0.00000463, -0.00000446, -0.00000408, -0.00000341, -0.00000240, -0.00000112, 0.00000025, 0.00000141, 0.00000201, 0.00000173, 0.00000031, -0.00000233, -0.00000604, -0.00001044, -0.00001496, -0.00001888, -0.00002148, -0.00002219, -0.00002064, -0.00001684, -0.00001117, -0.00000436, 0.00000259, 0.00000857, 0.00001255, 0.00001378, 0.00001191, 0.00000709, 0.00000000, -0.00000824, -0.00001627, -0.00002266, -0.00002619, -0.00002608, -0.00002208, -0.00001463, -0.00000474, 0.00000612, 0.00001629, 0.00002417, 0.00002853, 0.00002871, 0.00002479, 0.00001758, 0.00000852, -0.00000056, -0.00000772, -0.00001130, -0.00001014, -0.00000390, 0.00000691, 0.00002093, 0.00003616, 0.00005024, 0.00006082, 0.00006593, 0.00006432, 0.00005563, 0.00004053, 0.00002057, -0.00000198, -0.00002451, -0.00004440, -0.00005946, -0.00006819, -0.00007001, -0.00006526, -0.00005518, -0.00004163, -0.00002678, -0.00001281, -0.00000154, 0.00000580, 0.00000873, 0.00000752, 0.00000306, -0.00000329, -0.00001001, -0.00001568, -0.00001921, -0.00002001, -0.00001804, -0.00001379, -0.00000812, -0.00000211, 0.00000317, 0.00000687, 0.00000847, 0.00000789, 0.00000545, 0.00000179, -0.00000224, -0.00000581, -0.00000819, -0.00000893, -0.00000791, -0.00000535, -0.00000172, 0.00000231, 0.00000607, 0.00000893, 0.00001050, 0.00001059, 0.00000932, 0.00000699, 0.00000406, 0.00000104, -0.00000163, -0.00000361, -0.00000475, -0.00000507, -0.00000474, -0.00000404, -0.00000324, -0.00000262, -0.00000233, -0.00000244, -0.00000289, -0.00000355, -0.00000427, -0.00000488, -0.00000528, -0.00000544, },
		{ -0.00000366, -0.00000256, -0.00000154, -0.00000062, 0.00000012, 0.00000056, 0.00000051, -0.00000018, -0.00000166, -0.00000394, -0.00000691, -0.00001027, -0.00001360, -0.00001638, -0.00001807, -0.00001825, -0.00001664, -0.00001325, -0.00000838, -0.00000262, 0.00000325, 0.00000833, 0.00001179, 0.00001298, 0.00001160, 0.00000774, 0.00000191, -0.00000501, -0.00001192, -0.00001765, -0.00002118, -0.00002182, -0.00001933, -0.00001399, -0.00000657, 0.00000175, 0.00000965, 0.00001578, 0.00001909, 0.00001896, 0.00001531, 0.00000868, 0.00000009, -0.00000911, -0.00001746, -0.00002369, -0.00002687, -0.00002667, -0.00002336, -0.00001780, -0.00001130, -0.00000537, -0.00000144, -0.00000062, -0.00000346, -0.00000982, -0.00001886, -0.00002916, -0.00003897, -0.00004640, -0.00004985, -0.00004821, -0.00004107, -0.00002887, -0.00001279, 0.00000540, 0.00002355, 0.00003952, 0.00005144, 0.00005804, 0.00005877, 0.00005392, 0.00004451, 0.00003213, 0.00001867, 0.00000603, -0.00000418, -0.00001088, -0.00001360, -0.00001257, -0.00000859, -0.00000288, 0.00000319, 0.00000831, 0.00001147, 0.00001211, 0.00001017, 0.00000607, 0.00000063, -0.00000515, -0.00001025, -0.00001384, -0.00001540, -0.00001478, -0.00001226, -0.00000841, -0.00000402, 0.00000007, 0.00000314, 0.00000470, 0.00000453, 0.00000274, -0.00000029, -0.00000399, -0.00000775, -0.00001099, -0.00001329, -0.00001444, -0.00001445, -0.00001355, -0.00001209, -0.00001049, -0.00000912, -0.00000826, -0.00000805, -0.00000845, -0.00000932, -0.00001040, -0.00001142, -0.00001215, -0.00001242, -0.00001217, -0.00001143, -0.00001032, -0.00000897, -0.00000755, -0.00000615, -0.00000485, },
		{ 0.00000185, -0.00000017, -0.00000262, -0.00000539, -0.00000829, -0.00001105, -0.00001330, -0.00001468, -0.00001485, -0.00001364, -0.00001103, -0.00000724, -0.00000270, 0.00000199, 0.00000612, 0.00000903, 0.00001018, 0.00000928, 0.00000635, 0.00000178, -0.00000377, -0.00000944, -0.00001426, -0.00001739, -0.00001822, -0.00001648, -0.00001234, -0.00000637, 0.00000051, 0.00000723, 0.00001267, 0.00001593, 0.00001641, 0.00001401, 0.00000905, 0.00000232, -0.00000510, -0.00001200, -0.00001727, -0.00002006, -0.00001992, -0.00001690, -0.00001152, -0.00000467, 0.00000255, 0.00000906, 0.00001396, 0.00001672, 0.00001725, 0.00001595, 0.00001358, 0.00001115, 0.00000966, 0.00000993, 0.00001241, 0.00001703, 0.00002320, 0.00002989, 0.00003575, 0.00003938, 0.00003954, 0.00003543, 0.00002680, 0.00001408, -0.00000167, -0.00001889, -0.00003572, -0.00005031, -0.00006105, -0.00006685, -0.00006726, -0.00006258, -0.00005375, -0.00004222, -0.00002971, -0.00001797, -0.00000848, -0.00000225, 0.00000029, -0.00000068, -0.00000442, -0.00000981, -0.00001556, -0.00002043, -0.00002343, -0.00002399, -0.00002199, -0.00001782, -0.00001219, -0.00000605, -0.00000038, 0.00000397, 0.00000643, 0.00000679, 0.00000521, 0.00000215, -0.00000171, -0.00000560, -0.00000885, -0.00001095, -0.00001165, -0.00001099, -0.00000924, -0.00000686, -0.00000438, -0.00000229, -0.00000096, -0.00000058, -0.00000111, -0.00000237, -0.00000401, -0.00000564, -0.00000690, -0.00000751, -0.00000733, -0.00000637, -0.00000477, -0.00000274, -0.00000057, 0.00000150, 0.00000327, 0.00000460, 0.00000546, 0.00000585, 0.00000582, 0.00000540, 0.00000461, 0.00000344, },
		{ -0.00001360, -0.00001360, -0.00001257, -0.00001049, -0.00000747, -0.00000382, 0.00000003, 0.00000356, 0.00000624, 0.00000760, 0.00000738, 0.00000553, 0.00000227, -0.00000192, -0.00000637, -0.00001034, -0.00001313, -0.00001420, -0.00001327, -0.00001041, -0.00000602, -0.00000078, 0.00000444, 0.00000874, 0.00001133, 0.00001169, 0.00000964, 0.00000540, -0.00000045, -0.00000703, -0.00001335, -0.00001844, -0.00002151, -0.00002211, -0.00002018, -0.00001604, -0.00001039, -0.00000411, 0.00000184, 0.00000659, 0.00000954, 0.00001040, 0.00000925, 0.00000651, 0.00000278, -0.00000121, -0.00000483, -0.00000764, -0.00000950, -0.00001058, -0.00001132, -0.00001232, -0.00001419, -0.00001741, -0.00002217, -0.00002826, -0.00003510, -0.00004176, -0.00004709, -0.00004992, -0.00004927, -0.00004452, -0.00003558, -0.00002291, -0.00000755, 0.00000903, 0.00002514, 0.00003906, 0.00004934, 0.00005498, 0.00005557, 0.00005140, 0.00004333, 0.00003271, 0.00002114, 0.00001028, 0.00000151, -0.00000417, -0.00000630, -0.00000503, -0.00000100, 0.00000476, 0.00001104, 0.00001663, 0.00002057, 0.00002223, 0.00002142, 0.00001841, 0.00001378, 0.00000839, 0.00000311, -0.00000125, -0.00000412, -0.00000525, -0.00000472, -0.00000291, -0.00000037, 0.00000224, 0.00000432, 0.00000542, 0.00000532, 0.00000407, 0.00000192, -0.00000070, -0.00000326, -0.00000530, -0.00000644, -0.00000649, -0.00000544, -0.00000348, -0.00000092, 0.00000185, 0.00000445, 0.00000655, 0.00000793, 0.00000850, 0.00000829, 0.00000740, 0.00000599, 0.00000420, 0.00000218, 0.00000001, -0.00000225, -0.00000458, -0.00000691, -0.00000915, -0.00001115, -0.00001271, },
		{ 0.00000067, 0.00000329, 0.00000503, 0.00000562, 0.00000494, 0.00000308, 0.00000034, -0.00000283, -0.00000586, -0.00000817, -0.00000929, -0.00000893, -0.00000705, -0.00000391, 0.00000004, 0.00000413, 0.00000765, 0.00000995, 0.00001055, 0.00000925, 0.00000617, 0.00000174, -0.00000336, -0.00000831, -0.00001228, -0.00001458, -0.00001475, -0.00001266, -0.00000853, -0.00000288, 0.00000353, 0.00000987, 0.00001533, 0.00001929, 0.00002138, 0.00002153, 0.00001998, 0.00001720, 0.00001379, 0.00001038, 0.00000748, 0.00000544, 0.00000437, 0.00000417, 0.00000457, 0.00000525, 0.00000589, 0.00000631, 0.00000650, 0.00000666, 0.00000716, 0.00000845, 0.00001095, 0.00001494, 0.00002042, 0.00002710, 0.00003432, 0.00004118, 0.00004660, 0.00004952, 0.00004907, 0.00004470, 0.00003637, 0.00002455, 0.00001022, -0.00000526, -0.00002029, -0.00003328, -0.00004287, -0.00004810, -0.00004857, -0.00004450, -0.00003666, -0.00002630, -0.00001490, -0.00000399, 0.00000507, 0.00001129, 0.00001419, 0.00001381, 0.00001067, 0.00000565, -0.00000014, -0.00000561, -0.00000984, -0.00001222, -0.00001252, -0.00001091, -0.00000789, -0.00000418, -0.00000057, 0.00000224, 0.00000374, 0.00000372, 0.00000226, -0.00000029, -0.00000338, -0.00000638, -0.00000869, -0.00000986, -0.00000967, -0.00000811, -0.00000542, -0.00000204, 0.00000155, 0.00000483, 0.00000737, 0.00000888, 0.00000926, 0.00000856, 0.00000695, 0.00000472, 0.00000214, -0.00000052, -0.00000307, -0.00000541, -0.00000747, -0.00000927, -0.00001081, -0.00001206, -0.00001296, -0.00001340, -0.00001326, -0.00001242, -0.00001082, -0.00000850, -0.00000561, -0.00000243, },
		{ 0.00000449, 0.00000229, 0.00000033, -0.00000098, -0.00000133, -0.00000058, 0.00000117, 0.00000365, 0.00000640, 0.00000888, 0.00001056, 0.00001102, 0.00001004, 0.00000765, 0.00000413, -0.00000002, -0.00000413, -0.00000753, -0.00000962, -0.00000996, -0.00000839, -0.00000505, -0.00000033, 0.00000515, 0.00001066, 0.00001545, 0.00001893, 0.00002068, 0.00002056, 0.00001869, 0.00001544, 0.00001133, 0.00000694, 0.00000282, -0.00000060, -0.00000308, -0.00000457, -0.00000520, -0.00000523, -0.00000494, -0.00000462, -0.00000446, -0.00000452, -0.00000475, -0.00000501, -0.00000515, -0.00000504, -0.00000466, -0.00000414, -0.00000375, -0.00000386, -0.00000490, -0.00000724, -0.00001106, -0.00001634, -0.00002274, -0.00002962, -0.00003612, -0.00004124, -0.00004399, -0.00004357, -0.00003948, -0.00003167, -0.00002058, -0.00000708, 0.00000757, 0.00002189, 0.00003443, 0.00004390, 0.00004937, 0.00005044, 0.00004723, 0.00004042, 0.00003108, 0.00002055, 0.00001023, 0.00000135, -0.00000515, -0.00000880, -0.00000957, -0.00000790, -0.00000457, -0.00000054, 0.00000321, 0.00000585, 0.00000684, 0.00000599, 0.00000347, -0.00000024, -0.00000443, -0.00000834, -0.00001128, -0.00001273, -0.00001245, -0.00001050, -0.00000720, -0.00000308, 0.00000120, 0.00000504, 0.00000790, 0.00000946, 0.00000959, 0.00000840, 0.00000616, 0.00000326, 0.00000010, -0.00000294, -0.00000560, -0.00000771, -0.00000926, -0.00001030, -0.00001094, -0.00001128, -0.00001139, -0.00001127, -0.00001086, -0.00001006, -0.00000876, -0.00000692, -0.00000455, -0.00000179, 0.00000114, 0.00000391, 0.00000621, 0.00000773, 0.00000829, 0.00000783, 0.00000647, },
		{ -0.00000062, 0.00000058, 0.00000183, 0.00000273, 0.00000291, 0.00000212, 0.00000031, -0.00000236, -0.00000556, -0.00000880, -0.00001153, -0.00001324, -0.00001354, -0.00001225, -0.00000941, -0.00000532, -0.00000047, 0.00000451, 0.00000899, 0.00001238, 0.00001427, 0.00001448, 0.00001304, 0.00001024, 0.00000653, 0.00000245, -0.00000147, -0.00000477, -0.00000712, -0.00000839, -0.00000863, -0.00000803, -0.00000686, -0.00000545, -0.00000404, -0.00000284, -0.00000193, -0.00000129, -0.00000084, -0.00000044, -0.00000000, 0.00000053, 0.00000115, 0.00000175, 0.00000222, 0.00000243, 0.00000232, 0.00000193, 0.00000141, 0.00000105, 0.00000121, 0.00000228, 0.00000458, 0.00000827, 0.00001330, 0.00001936, 0.00002586, 0.00003201, 0.00003690, 0.00003965, 0.00003952, 0.00003607, 0.00002923, 0.00001937, 0.00000729, -0.00000590, -0.00001891, -0.00003040, -0.00003921, -0.00004453, -0.00004597, -0.00004363, -0.00003810, -0.00003036, -0.00002161, -0.00001310, -0.00000595, -0.00000098, 0.00000136, 0.00000109, -0.00000135, -0.00000521, -0.00000955, -0.00001342, -0.00001601, -0.00001675, -0.00001544, -0.00001221, -0.00000753, -0.00000207, 0.00000339, 0.00000810, 0.00001149, 0.00001319, 0.00001312, 0.00001147, 0.00000861, 0.00000505, 0.00000133, -0.00000208, -0.00000485, -0.00000682, -0.00000798, -0.00000847, -0.00000850, -0.00000830, -0.00000804, -0.00000782, -0.00000761, -0.00000730, -0.00000676, -0.00000581, -0.00000436, -0.00000240, -0.00000004, 0.00000250, 0.00000493, 0.00000693, 0.00000821, 0.00000861, 0.00000808, 0.00000673, 0.00000481, 0.00000268, 0.00000072, -0.00000074, -0.00000147, -0.00000139, },
		{ -0.00000217, -0.00000406, -0.00000608, -0.00000782, -0.00000886, -0.00000885, -0.00000759, -0.00000506, -0.00000145, 0.00000285, 0.00000735, 0.00001148, 0.00001472, 0.00001666, 0.00001708, 0.00001597, 0.00001351, 0.00001010, 0.00000619, 0.00000231, -0.00000108, -0.00000363, -0.00000516, -0.00000565, -0.00000524, -0.00000419, -0.00000280, -0.00000138, -0.00000020, 0.00000060, 0.00000096, 0.00000091, 0.00000057, 0.00000007, -0.00000044, -0.00000088, -0.00000122, -0.00000146, -0.00000167, -0.00000192, -0.00000224, -0.00000266, -0.00000312, -0.00000354, -0.00000381, -0.00000383, -0.00000357, -0.00000307, -0.00000250, -0.00000209, -0.00000218, -0.00000308, -0.00000506, -0.00000825, -0.00001257, -0.00001773, -0.00002320, -0.00002829, -0.00003221, -0.00003419, -0.00003361, -0.00003010, -0.00002366, -0.00001465, -0.00000381, 0.00000785, 0.00001913, 0.00002884, 0.00003595, 0.00003974, 0.00003987, 0.00003651, 0.00003023, 0.00002199, 0.00001294, 0.00000431, -0.00000280, -0.00000758, -0.00000957, -0.00000876, -0.00000554, -0.00000063, 0.00000507, 0.00001058, 0.00001503, 0.00001778, 0.00001850, 0.00001718, 0.00001412, 0.00000984, 0.00000499, 0.00000022, -0.00000392, -0.00000706, -0.00000906, -0.00000998, -0.00001006, -0.00000961, -0.00000900, -0.00000849, -0.00000824, -0.00000827, -0.00000844, -0.00000853, -0.00000829, -0.00000749, -0.00000601, -0.00000384, -0.00000114, 0.00000180, 0.00000465, 0.00000702, 0.00000861, 0.00000922, 0.00000881, 0.00000751, 0.00000558, 0.00000338, 0.00000127, -0.00000041, -0.00000146, -0.00000181, -0.00000156, -0.00000094, -0.00000027, 0.00000015, 0.00000003, -0.00000075, },
		{ -0.00000461, -0.00000188, 0.00000151, 0.00000516, 0.00000861, 0.00001138, 0.00001310, 0.00001351, 0.00001255, 0.00001034, 0.00000717, 0.00000347, -0.00000030, -0.00000368, -0.00000633, -0.00000800, -0.00000864, -0.00000833, -0.00000728, -0.00000580, -0.00000419, -0.00000274, -0.00000165, -0.00000102, -0.00000086, -0.00000106, -0.00000148, -0.00000195, -0.00000231, -0.00000247, -0.00000237, -0.00000203, -0.00000152, -0.00000091, -0.00000029, 0.00000028, 0.00000077, 0.00000120, 0.00000161, 0.00000203, 0.00000249, 0.00000301, 0.00000354, 0.00000404, 0.00000442, 0.00000465, 0.00000471, 0.00000467, 0.00000468, 0.00000495, 0.00000574, 0.00000729, 0.00000977, 0.00001323, 0.00001755, 0.00002240, 0.00002726, 0.00003148, 0.00003434, 0.00003516, 0.00003341, 0.00002880, 0.00002138, 0.00001156, 0.00000007, -0.00001208, -0.00002374, -0.00003372, -0.00004104, -0.00004494, -0.00004510, -0.00004162, -0.00003502, -0.00002619, -0.00001624, -0.00000637, 0.00000230, 0.00000891, 0.00001289, 0.00001409, 0.00001274, 0.00000938, 0.00000477, -0.00000027, -0.00000494, -0.00000864, -0.00001100, -0.00001195, -0.00001164, -0.00001045, -0.00000883, -0.00000724, -0.00000604, -0.00000544, -0.00000543, -0.00000585, -0.00000637, -0.00000663, -0.00000628, -0.00000507, -0.00000294, 0.00000001, 0.00000351, 0.00000715, 0.00001046, 0.00001301, 0.00001447, 0.00001466, 0.00001362, 0.00001153, 0.00000875, 0.00000569, 0.00000278, 0.00000036, -0.00000135, -0.00000229, -0.00000256, -0.00000239, -0.00000207, -0.00000188, -0.00000206, -0.00000269, -0.00000375, -0.00000503, -0.00000626, -0.00000710, -0.00000723, -0.00000642, },
		{ 0.00001011, 0.00000971, 0.00000815, 0.00000565, 0.00000255, -0.00000073, -0.00000377, -0.00000620, -0.00000776, -0.00000836, -0.00000802, -0.00000692, -0.00000533, -0.00000355, -0.00000187, -0.00000053, 0.00000033, 0.00000067, 0.00000057, 0.00000015, -0.00000040, -0.00000090, -0.00000121, -0.00000125, -0.00000101, -0.00000053, 0.00000010, 0.00000075, 0.00000131, 0.00000169, 0.00000185, 0.00000179, 0.00000156, 0.00000120, 0.00000080, 0.00000042, 0.00000010, -0.00000014, -0.00000029, -0.00000038, -0.00000041, -0.00000037, -0.00000027, -0.00000009, 0.00000019, 0.00000053, 0.00000087, 0.00000110, 0.00000102, 0.00000043, -0.00000094, -0.00000327, -0.00000669, -0.00001120, -0.00001660, -0.00002255, -0.00002848, -0.00003373, -0.00003758, -0.00003934, -0.00003848, -0.00003468, -0.00002795, -0.00001864, -0.00000741, 0.00000481, 0.00001691, 0.00002779, 0.00003643, 0.00004207, 0.00004426, 0.00004299, 0.00003860, 0.00003179, 0.00002349, 0.00001472, 0.00000647, -0.00000046, -0.00000553, -0.00000855, -0.00000964, -0.00000918, -0.00000777, -0.00000605, -0.00000460, -0.00000385, -0.00000401, -0.00000504, -0.00000667, -0.00000846, -0.00000992, -0.00001059, -0.00001012, -0.00000839, -0.00000548, -0.00000171, 0.00000244, 0.00000641, 0.00000964, 0.00001168, 0.00001223, 0.00001125, 0.00000889, 0.00000550, 0.00000157, -0.00000239, -0.00000590, -0.00000863, -0.00001037, -0.00001112, -0.00001105, -0.00001043, -0.00000959, -0.00000883, -0.00000837, -0.00000829, -0.00000855, -0.00000897, -0.00000927, -0.00000918, -0.00000846, -0.00000695, -0.00000467, -0.00000177, 0.00000146, 0.00000465, 0.00000739, 0.00000930, },
		{ -0.00000322, -0.00000507, -0.00000601, -0.00000605, -0.00000530, -0.00000399, -0.00000242, -0.00000086, 0.00000042, 0.00000125, 0.00000157, 0.00000141, 0.00000089, 0.00000018, -0.00000054, -0.00000109, -0.00000136, -0.00000131, -0.00000095, -0.00000039, 0.00000026, 0.00000087, 0.00000131, 0.00000151, 0.00000144, 0.00000114, 0.00000069, 0.00000017, -0.00000029, -0.00000060, -0.00000070, -0.00000055, -0.00000016, 0.00000042, 0.00000113, 0.00000189, 0.00000264, 0.00000330, 0.00000381, 0.00000411, 0.00000416, 0.00000392, 0.00000335, 0.00000245, 0.00000121, -0.00000030, -0.00000196, -0.00000360, -0.00000500, -0.00000589, -0.00000598, -0.00000503, -0.00000289, 0.00000048, 0.00000494, 0.00001019, 0.00001575, 0.00002101, 0.00002532, 0.00002801, 0.00002856, 0.00002661, 0.00002210, 0.00001524, 0.00000652, -0.00000331, -0.00001336, -0.00002268, -0.00003043, -0.00003594, -0.00003881, -0.00003901, -0.00003680, -0.00003274, -0.00002761, -0.00002222, -0.00001737, -0.00001369, -0.00001152, -0.00001095, -0.00001174, -0.00001340, -0.00001532, -0.00001682, -0.00001730, -0.00001635, -0.00001381, -0.00000980, -0.00000472, 0.00000087, 0.00000627, 0.00001081, 0.00001393, 0.00001527, 0.00001472, 0.00001244, 0.00000883, 0.00000442, -0.00000017, -0.00000435, -0.00000765, -0.00000979, -0.00001068, -0.00001045, -0.00000936, -0.00000780, -0.00000614, -0.00000470, -0.00000369, -0.00000316, -0.00000303, -0.00000308, -0.00000305, -0.00000267, -0.00000175, -0.00000018, 0.00000196, 0.00000451, 0.00000714, 0.00000952, 0.00001128, 0.00001214, 0.00001194, 0.00001066, 0.00000845, 0.00000558, 0.00000242, -0.00000064, },
		{ -0.00000148, -0.00000010, 0.00000090, 0.00000139, 0.00000138, 0.00000095, 0.00000027, -0.00000048, -0.00000112, -0.00000152, -0.00000159, -0.00000134, -0.00000084, -0.00000020, 0.00000043, 0.00000092, 0.00000118, 0.00000115, 0.00000085, 0.00000033, -0.00000029, -0.00000088, -0.00000134, -0.00000157, -0.00000152, -0.00000119, -0.00000064, 0.00000005, 0.00000076, 0.00000137, 0.00000178, 0.00000190, 0.00000169, 0.00000112, 0.00000023, -0.00000096, -0.00000238, -0.00000393, -0.00000555, -0.00000712, -0.00000854, -0.00000973, -0.00001059, -0.00001106, -0.00001109, -0.00001069, -0.00000992, -0.00000890, -0.00000781, -0.00000687, -0.00000632, -0.00000636, -0.00000716, -0.00000878, -0.00001115, -0.00001407, -0.00001720, -0.00002011, -0.00002230, -0.00002330, -0.00002272, -0.00002033, -0.00001612, -0.00001029, -0.00000330, 0.00000422, 0.00001152, 0.00001784, 0.00002249, 0.00002498, 0.00002509, 0.00002289, 0.00001877, 0.00001336, 0.00000747, 0.00000198, -0.00000231, -0.00000476, -0.00000502, -0.00000305, 0.00000087, 0.00000617, 0.00001208, 0.00001775, 0.00002240, 0.00002536, 0.00002624, 0.00002495, 0.00002167, 0.00001686, 0.00001116, 0.00000528, -0.00000009, -0.00000440, -0.00000730, -0.00000867, -0.00000862, -0.00000746, -0.00000560, -0.00000350, -0.00000158, -0.00000013, 0.00000070, 0.00000094, 0.00000076, 0.00000042, 0.00000022, 0.00000041, 0.00000116, 0.00000250, 0.00000432, 0.00000641, 0.00000843, 0.00001005, 0.00001098, 0.00001100, 0.00001004, 0.00000818, 0.00000562, 0.00000268, -0.00000028, -0.00000289, -0.00000487, -0.00000602, -0.00000628, -0.00000574, -0.00000458, -0.00000306, },
		{ -0.00000098, -0.00000165, -0.00000212, -0.00000228, -0.00000210, -0.00000162, -0.00000093, -0.00000019, 0.00000047, 0.00000095, 0.00000115, 0.00000108, 0.00000077, 0.00000033, -0.00000013, -0.00000046, -0.00000058, -0.00000040, 0.00000007, 0.00000081, 0.00000171, 0.00000265, 0.00000351, 0.00000415, 0.00000450, 0.00000449, 0.00000412, 0.00000343, 0.00000251, 0.00000144, 0.00000034, -0.00000069, -0.00000156, -0.00000220, -0.00000258, -0.00000267, -0.00000248, -0.00000202, -0.00000134, -0.00000046, 0.00000056, 0.00000166, 0.00000281, 0.00000395, 0.00000506, 0.00000613, 0.00000717, 0.00000824, 0.00000940, 0.00001075, 0.00001236, 0.00001432, 0.00001662, 0.00001921, 0.00002195, 0.00002460, 0.00002683, 0.00002826, 0.00002851, 0.00002722, 0.00002415, 0.00001923, 0.00001257, 0.00000449, -0.00000444, -0.00001354, -0.00002198, -0.00002896, -0.00003377, -0.00003587, -0.00003501, -0.00003124, -0.00002494, -0.00001676, -0.00000759, 0.00000160, 0.00000985, 0.00001630, 0.00002037, 0.00002176, 0.00002049, 0.00001694, 0.00001169, 0.00000554, -0.00000071, -0.00000629, -0.00001060, -0.00001327, -0.00001419, -0.00001351, -0.00001158, -0.00000889, -0.00000596, -0.00000326, -0.00000117, 0.00000014, 0.00000064, 0.00000051, 0.00000002, -0.00000052, -0.00000078, -0.00000055, 0.00000029, 0.00000170, 0.00000348, 0.00000537, 0.00000704, 0.00000816, 0.00000851, 0.00000794, 0.00000649, 0.00000429, 0.00000163, -0.00000117, -0.00000375, -0.00000579, -0.00000708, -0.00000752, -0.00000714, -0.00000610, -0.00000462, -0.00000298, -0.00000147, -0.00000029, 0.00000041, 0.00000060, 0.00000033, -0.00000025, },
		{ 0.00000171, 0.00000253, 0.00000319, 0.00000360, 0.00000372, 0.00000358, 0.00000323, 0.00000281, 0.00000244, 0.00000223, 0.00000227, 0.00000258, 0.00000315, 0.00000389, 0.00000469, 0.00000541, 0.00000593, 0.00000613, 0.00000597, 0.00000543, 0.00000456, 0.00000343, 0.00000215, 0.00000084, -0.00000039, -0.00000146, -0.00000231, -0.00000292, -0.00000331, -0.00000350, -0.00000354, -0.00000346, -0.00000329, -0.00000305, -0.00000274, -0.00000235, -0.00000185, -0.00000123, -0.00000048, 0.00000038, 0.00000134, 0.00000233, 0.00000329, 0.00000412, 0.00000473, 0.00000498, 0.00000476, 0.00000395, 0.00000245, 0.00000018, -0.00000289, -0.00000675, -0.00001131, -0.00001637, -0.00002168, -0.00002688, -0.00003154, -0.00003519, -0.00003737, -0.00003767, -0.00003578, -0.00003156, -0.00002507, -0.00001658, -0.00000658, 0.00000425, 0.00001510, 0.00002512, 0.00003347, 0.00003947, 0.00004264, 0.00004277, 0.00003995, 0.00003457, 0.00002725, 0.00001880, 0.00001009, 0.00000198, -0.00000484, -0.00000984, -0.00001277, -0.00001367, -0.00001280, -0.00001061, -0.00000768, -0.00000458, -0.00000183, 0.00000019, 0.00000129, 0.00000148, 0.00000091, -0.00000009, -0.00000119, -0.00000202, -0.00000233, -0.00000196, -0.00000093, 0.00000061, 0.00000241, 0.00000414, 0.00000546, 0.00000612, 0.00000593, 0.00000488, 0.00000307, 0.00000073, -0.00000182, -0.00000423, -0.00000620, -0.00000748, -0.00000793, -0.00000756, -0.00000647, -0.00000489, -0.00000307, -0.00000130, 0.00000018, 0.00000121, 0.00000172, 0.00000173, 0.00000135, 0.00000077, 0.00000015, -0.00000031, -0.00000049, -0.00000034, 0.00000014, 0.00000087, },
		{ 0.00000107, 0.00000048, -0.00000000, -0.00000027, -0.00000027, -0.00000000, 0.00000047, 0.00000105, 0.00000162, 0.00000204, 0.00000221, 0.00000204, 0.00000152, 0.00000068, -0.00000042, -0.00000167, -0.00000294, -0.00000414, -0.00000516, -0.00000595, -0.00000649, -0.00000680, -0.00000689, -0.00000682, -0.00000662, -0.00000632, -0.00000592, -0.00000540, -0.00000473, -0.00000387, -0.00000279, -0.00000150, -0.00000000, 0.00000161, 0.00000326, 0.00000479, 0.00000608, 0.00000697, 0.00000734, 0.00000712, 0.00000627, 0.00000480, 0.00000282, 0.00000045, -0.00000212, -0.00000466, -0.00000694, -0.00000873, -0.00000980, -0.00000998, -0.00000914, -0.00000722, -0.00000424, -0.00000030, 0.00000440, 0.00000957, 0.00001485, 0.00001982, 0.00002404, 0.00002710, 0.00002861, 0.00002833, 0.00002611, 0.00002200, 0.00001621, 0.00000912, 0.00000125, -0.00000679, -0.00001436, -0.00002085, -0.00002577, -0.00002878, -0.00002973, -0.00002872, -0.00002600, -0.00002202, -0.00001732, -0.00001248, -0.00000803, -0.00000440, -0.00000185, -0.00000046, -0.00000014, -0.00000063, -0.00000158, -0.00000261, -0.00000335, -0.00000353, -0.00000301, -0.00000180, -0.00000004, 0.00000201, 0.00000402, 0.00000567, 0.00000666, 0.00000683, 0.00000611, 0.00000460, 0.00000250, 0.00000011, -0.00000223, -0.00000421, -0.00000557, -0.00000616, -0.00000594, -0.00000500, -0.00000354, -0.00000180, -0.00000006, 0.00000142, 0.00000246, 0.00000296, 0.00000293, 0.00000246, 0.00000170, 0.00000085, 0.00000008, -0.00000043, -0.00000062, -0.00000046, -0.00000000, 0.00000063, 0.00000130, 0.00000187, 0.00000223, 0.00000230, 0.00000209, 0.00000165, },
		{ 0.00000187, 0.00000255, 0.00000314, 0.00000352, 0.00000360, 0.00000333, 0.00000272, 0.00000183, 0.00000075, -0.00000041, -0.00000154, -0.00000253, -0.00000331, -0.00000386, -0.00000416, -0.00000425, -0.00000416, -0.00000393, -0.00000359, -0.00000314, -0.00000257, -0.00000185, -0.00000096, 0.00000013, 0.00000140, 0.00000283, 0.00000432, 0.00000576, 0.00000700, 0.00000787, 0.00000825, 0.00000801, 0.00000712, 0.00000558, 0.00000348, 0.00000097, -0.00000176, -0.00000448, -0.00000697, -0.00000905, -0.00001055, -0.00001141, -0.00001159, -0.00001116, -0.00001023, -0.00000893, -0.00000745, -0.00000597, -0.00000463, -0.00000357, -0.00000289, -0.00000263, -0.00000280, -0.00000336, -0.00000425, -0.00000535, -0.00000654, -0.00000765, -0.00000855, -0.00000905, -0.00000904, -0.00000839, -0.00000705, -0.00000501, -0.00000234, 0.00000082, 0.00000426, 0.00000773, 0.00001094, 0.00001361, 0.00001551, 0.00001647, 0.00001640, 0.00001534, 0.00001346, 0.00001098, 0.00000823, 0.00000555, 0.00000326, 0.00000162, 0.00000080, 0.00000083, 0.00000163, 0.00000299, 0.00000464, 0.00000623, 0.00000746, 0.00000807, 0.00000789, 0.00000688, 0.00000512, 0.00000282, 0.00000026, -0.00000225, -0.00000439, -0.00000593, -0.00000671, -0.00000668, -0.00000591, -0.00000459, -0.00000294, -0.00000125, 0.00000025, 0.00000135, 0.00000193, 0.00000198, 0.00000157, 0.00000084, -0.00000002, -0.00000082, -0.00000140, -0.00000164, -0.00000151, -0.00000104, -0.00000032, 0.00000050, 0.00000129, 0.00000190, 0.00000225, 0.00000230, 0.00000209, 0.00000169, 0.00000121, 0.00000079, 0.00000053, 0.00000051, 0.00000075, 0.00000124, },
		{ 0.00000138, 0.00000022, -0.00000107, -0.00000237, -0.00000356, -0.00000456, -0.00000530, -0.00000578, -0.00000598, -0.00000595, -0.00000572, -0.00000533, -0.00000479, -0.00000411, -0.00000329, -0.00000230, -0.00000115, 0.00000015, 0.00000156, 0.00000298, 0.00000431, 0.00000541, 0.00000614, 0.00000638, 0.00000603, 0.00000509, 0.00000358, 0.00000163, -0.00000060, -0.00000289, -0.00000503, -0.00000681, -0.00000807, -0.00000872, -0.00000874, -0.00000820, -0.00000720, -0.00000590, -0.00000447, -0.00000304, -0.00000173, -0.00000056, 0.00000048, 0.00000147, 0.00000253, 0.00000378, 0.00000534, 0.00000725, 0.00000953, 0.00001211, 0.00001483, 0.00001751, 0.00001992, 0.00002181, 0.00002298, 0.00002327, 0.00002257, 0.00002089, 0.00001827, 0.00001485, 0.00001081, 0.00000637, 0.00000176, -0.00000279, -0.00000707, -0.00001089, -0.00001410, -0.00001657, -0.00001822, -0.00001899, -0.00001886, -0.00001788, -0.00001610, -0.00001366, -0.00001074, -0.00000755, -0.00000434, -0.00000137, 0.00000113, 0.00000294, 0.00000395, 0.00000408, 0.00000340, 0.00000203, 0.00000019, -0.00000185, -0.00000381, -0.00000542, -0.00000646, -0.00000681, -0.00000643, -0.00000539, -0.00000385, -0.00000204, -0.00000019, 0.00000146, 0.00000271, 0.00000346, 0.00000367, 0.00000339, 0.00000276, 0.00000194, 0.00000114, 0.00000051, 0.00000019, 0.00000023, 0.00000062, 0.00000128, 0.00000208, 0.00000288, 0.00000353, 0.00000393, 0.00000401, 0.00000379, 0.00000332, 0.00000271, 0.00000209, 0.00000157, 0.00000126, 0.00000121, 0.00000142, 0.00000182, 0.00000232, 0.00000279, 0.00000312, 0.00000318, 0.00000291, 0.00000230, },
		{ -0.00000718, -0.00000723, -0.00000698, -0.00000648, -0.00000577, -0.00000489, -0.00000388, -0.00000275, -0.00000152, -0.00000020, 0.00000118, 0.00000258, 0.00000393, 0.00000512, 0.00000605, 0.00000660, 0.00000668, 0.00000623, 0.00000523, 0.00000375, 0.00000190, -0.00000016, -0.00000221, -0.00000407, -0.00000555, -0.00000654, -0.00000697, -0.00000686, -0.00000631, -0.00000545, -0.00000447, -0.00000350, -0.00000268, -0.00000206, -0.00000163, -0.00000130, -0.00000094, -0.00000037, 0.00000053, 0.00000186, 0.00000364, 0.00000576, 0.00000803, 0.00001017, 0.00001183, 0.00001269, 0.00001245, 0.00001090, 0.00000800, 0.00000383, -0.00000135, -0.00000716, -0.00001312, -0.00001869, -0.00002338, -0.00002675, -0.00002850, -0.00002846, -0.00002664, -0.00002323, -0.00001850, -0.00001286, -0.00000674, -0.00000057, 0.00000525, 0.00001040, 0.00001465, 0.00001785, 0.00001994, 0.00002095, 0.00002094, 0.00002002, 0.00001833, 0.00001602, 0.00001326, 0.00001021, 0.00000706, 0.00000399, 0.00000116, -0.00000126, -0.00000314, -0.00000439, -0.00000496, -0.00000489, -0.00000427, -0.00000324, -0.00000197, -0.00000068, 0.00000044, 0.00000123, 0.00000160, 0.00000152, 0.00000102, 0.00000021, -0.00000076, -0.00000172, -0.00000251, -0.00000301, -0.00000315, -0.00000293, -0.00000240, -0.00000169, -0.00000093, -0.00000026, 0.00000020, 0.00000036, 0.00000023, -0.00000018, -0.00000076, -0.00000140, -0.00000197, -0.00000235, -0.00000247, -0.00000230, -0.00000189, -0.00000132, -0.00000070, -0.00000016, 0.00000016, 0.00000018, -0.00000015, -0.00000081, -0.00000176, -0.00000288, -0.00000407, -0.00000519, -0.00000613, -0.00000681, },
		{ -0.00000005, 0.00000115, 0.00000235, 0.00000352, 0.00000462, 0.00000560, 0.00000639, 0.00000691, 0.00000709, 0.00000686, 0.00000618, 0.00000506, 0.00000355, 0.00000178, -0.00000013, -0.00000198, -0.00000360, -0.00000486, -0.00000564, -0.00000591, -0.00000572, -0.00000515, -0.00000434, -0.00000347, -0.00000268, -0.00000205, -0.00000164, -0.00000140, -0.00000121, -0.00000091, -0.00000035, 0.00000064, 0.00000212, 0.00000410, 0.00000643, 0.00000889, 0.00001117, 0.00001291, 0.00001377, 0.00001349, 0.00001195, 0.00000916, 0.00000535, 0.00000086, -0.00000378, -0.00000802, -0.00001131, -0.00001316, -0.00001327, -0.00001152, -0.00000802, -0.00000310, 0.00000275, 0.00000890, 0.00001468, 0.00001949, 0.00002280, 0.00002430, 0.00002384, 0.00002152, 0.00001759, 0.00001247, 0.00000664, 0.00000061, -0.00000514, -0.00001024, -0.00001442, -0.00001755, -0.00001960, -0.00002065, -0.00002084, -0.00002034, -0.00001932, -0.00001793, -0.00001630, -0.00001453, -0.00001267, -0.00001080, -0.00000897, -0.00000722, -0.00000564, -0.00000430, -0.00000324, -0.00000253, -0.00000216, -0.00000210, -0.00000230, -0.00000263, -0.00000300, -0.00000326, -0.00000331, -0.00000309, -0.00000259, -0.00000183, -0.00000090, 0.00000008, 0.00000098, 0.00000170, 0.00000213, 0.00000225, 0.00000208, 0.00000168, 0.00000117, 0.00000067, 0.00000031, 0.00000017, 0.00000030, 0.00000070, 0.00000128, 0.00000196, 0.00000259, 0.00000304, 0.00000320, 0.00000300, 0.00000242, 0.00000151, 0.00000035, -0.00000093, -0.00000219, -0.00000330, -0.00000415, -0.00000466, -0.00000481, -0.00000460, -0.00000407, -0.00000329, -0.00000232, -0.00000122, },
		{ 0.00000647, 0.00000639, 0.00000599, 0.00000524, 0.00000417, 0.00000281, 0.00000126, -0.00000035, -0.00000189, -0.00000322, -0.00000421, -0.00000479, -0.00000493, -0.00000468, -0.00000412, -0.00000338, -0.00000262, -0.00000197, -0.00000150, -0.00000125, -0.00000117, -0.00000115, -0.00000104, -0.00000067, 0.00000010, 0.00000134, 0.00000303, 0.00000505, 0.00000716, 0.00000907, 0.00001043, 0.00001094, 0.00001037, 0.00000860, 0.00000571, 0.00000192, -0.00000238, -0.00000670, -0.00001051, -0.00001331, -0.00001473, -0.00001454, -0.00001277, -0.00000966, -0.00000563, -0.00000128, 0.00000275, 0.00000585, 0.00000751, 0.00000743, 0.00000553, 0.00000200, -0.00000275, -0.00000815, -0.00001351, -0.00001818, -0.00002157, -0.00002325, -0.00002304, -0.00002095, -0.00001725, -0.00001234, -0.00000677, -0.00000108, 0.00000420, 0.00000867, 0.00001204, 0.00001422, 0.00001525, 0.00001528, 0.00001457, 0.00001338, 0.00001199, 0.00001059, 0.00000933, 0.00000827, 0.00000742, 0.00000674, 0.00000618, 0.00000568, 0.00000521, 0.00000477, 0.00000440, 0.00000413, 0.00000401, 0.00000407, 0.00000432, 0.00000471, 0.00000517, 0.00000562, 0.00000595, 0.00000610, 0.00000599, 0.00000563, 0.00000506, 0.00000437, 0.00000365, 0.00000303, 0.00000261, 0.00000245, 0.00000257, 0.00000291, 0.00000340, 0.00000390, 0.00000427, 0.00000440, 0.00000418, 0.00000358, 0.00000261, 0.00000136, -0.00000005, -0.00000148, -0.00000277, -0.00000378, -0.00000442, -0.00000464, -0.00000443, -0.00000383, -0.00000292, -0.00000179, -0.00000055, 0.00000074, 0.00000198, 0.00000315, 0.00000418, 0.00000506, 0.00000576, 0.00000624, },
		{ -0.00000133, -0.00000252, -0.00000353, -0.00000425, -0.00000462, -0.00000462, -0.00000429, -0.00000370, -0.00000297, -0.00000222, -0.00000156, -0.00000108, -0.00000081, -0.00000070, -0.00000067, -0.00000057, -0.00000025, 0.00000042, 0.00000151, 0.00000302, 0.00000484, 0.00000676, 0.00000853, 0.00000985, 0.00001044, 0.00001009, 0.00000873, 0.00000640, 0.00000333, -0.00000013, -0.00000352, -0.00000639, -0.00000829, -0.00000890, -0.00000809, -0.00000590, -0.00000261, 0.00000135, 0.00000545, 0.00000908, 0.00001175, 0.00001307, 0.00001288, 0.00001123, 0.00000844, 0.00000499, 0.00000149, -0.00000142, -0.00000318, -0.00000339, -0.00000190, 0.00000121, 0.00000559, 0.00001070, 0.00001587, 0.00002038, 0.00002362, 0.00002510, 0.00002458, 0.00002207, 0.00001782, 0.00001228, 0.00000605, -0.00000023, -0.00000595, -0.00001063, -0.00001395, -0.00001577, -0.00001614, -0.00001529, -0.00001354, -0.00001126, -0.00000881, -0.00000647, -0.00000446, -0.00000288, -0.00000172, -0.00000091, -0.00000034, 0.00000009, 0.00000049, 0.00000091, 0.00000134, 0.00000176, 0.00000209, 0.00000228, 0.00000228, 0.00000209, 0.00000175, 0.00000132, 0.00000090, 0.00000060, 0.00000048, 0.00000059, 0.00000093, 0.00000142, 0.00000196, 0.00000242, 0.00000267, 0.00000259, 0.00000214, 0.00000129, 0.00000012, -0.00000126, -0.00000270, -0.00000403, -0.00000510, -0.00000577, -0.00000597, -0.00000569, -0.00000497, -0.00000390, -0.00000258, -0.00000115, 0.00000027, 0.00000159, 0.00000273, 0.00000365, 0.00000433, 0.00000477, 0.00000496, 0.00000491, 0.00000462, 0.00000410, 0.00000333, 0.00000236, 0.00000121, -0.00000005, },
		{ -0.00000182, -0.00000110, -0.00000037, 0.00000028, 0.00000075, 0.00000103, 0.00000112, 0.00000110, 0.00000109, 0.00000120, 0.00000157, 0.00000228, 0.00000333, 0.00000467, 0.00000613, 0.00000751, 0.00000855, 0.00000901, 0.00000871, 0.00000756, 0.00000561, 0.00000303, 0.00000012, -0.00000275, -0.00000516, -0.00000677, -0.00000729, -0.00000662, -0.00000484, -0.00000217, 0.00000098, 0.00000412, 0.00000676, 0.00000845, 0.00000891, 0.00000803, 0.00000592, 0.00000290, -0.00000057, -0.00000393, -0.00000664, -0.00000826, -0.00000853, -0.00000743, -0.00000515, -0.00000213, 0.00000106, 0.00000378, 0.00000542, 0.00000555, 0.00000392, 0.00000055, -0.00000424, -0.00000992, -0.00001580, -0.00002113, -0.00002519, -0.00002743, -0.00002750, -0.00002534, -0.00002116, -0.00001541, -0.00000871, -0.00000175, 0.00000479, 0.00001032, 0.00001444, 0.00001694, 0.00001785, 0.00001737, 0.00001584, 0.00001365, 0.00001122, 0.00000890, 0.00000693, 0.00000543, 0.00000442, 0.00000381, 0.00000347, 0.00000324, 0.00000299, 0.00000265, 0.00000219, 0.00000166, 0.00000113, 0.00000069, 0.00000042, 0.00000036, 0.00000048, 0.00000073, 0.00000099, 0.00000114, 0.00000105, 0.00000063, -0.00000015, -0.00000126, -0.00000261, -0.00000406, -0.00000541, -0.00000651, -0.00000720, -0.00000739, -0.00000704, -0.00000618, -0.00000491, -0.00000337, -0.00000171, -0.00000009, 0.00000137, 0.00000257, 0.00000347, 0.00000404, 0.00000431, 0.00000432, 0.00000411, 0.00000372, 0.00000317, 0.00000250, 0.00000173, 0.00000088, -0.00000000, -0.00000088, -0.00000168, -0.00000233, -0.00000276, -0.00000292, -0.00000280, -0.00000241, },
		{ -0.00000001, -0.00000028, -0.00000050, -0.00000055, -0.00000036, 0.00000014, 0.00000091, 0.00000187, 0.00000284, 0.00000365, 0.00000409, 0.00000398, 0.00000324, 0.00000188, 0.00000000, -0.00000216, -0.00000431, -0.00000612, -0.00000730, -0.00000762, -0.00000697, -0.00000542, -0.00000316, -0.00000052, 0.00000208, 0.00000424, 0.00000559, 0.00000589, 0.00000508, 0.00000328, 0.00000078, -0.00000200, -0.00000458, -0.00000649, -0.00000739, -0.00000707, -0.00000557, -0.00000311, -0.00000011, 0.00000290, 0.00000539, 0.00000686, 0.00000698, 0.00000567, 0.00000305, -0.00000049, -0.00000438, -0.00000797, -0.00001061, -0.00001174, -0.00001102, -0.00000839, -0.00000406, 0.00000150, 0.00000763, 0.00001355, 0.00001849, 0.00002182, 0.00002310, 0.00002216, 0.00001911, 0.00001435, 0.00000844, 0.00000208, -0.00000404, -0.00000930, -0.00001324, -0.00001563, -0.00001644, -0.00001585, -0.00001420, -0.00001190, -0.00000937, -0.00000699, -0.00000502, -0.00000360, -0.00000274, -0.00000234, -0.00000223, -0.00000224, -0.00000221, -0.00000205, -0.00000173, -0.00000130, -0.00000089, -0.00000061, -0.00000060, -0.00000093, -0.00000162, -0.00000261, -0.00000378, -0.00000494, -0.00000592, -0.00000653, -0.00000666, -0.00000623, -0.00000527, -0.00000386, -0.00000216, -0.00000034, 0.00000141, 0.00000294, 0.00000412, 0.00000490, 0.00000526, 0.00000526, 0.00000495, 0.00000443, 0.00000377, 0.00000303, 0.00000227, 0.00000151, 0.00000076, 0.00000003, -0.00000066, -0.00000129, -0.00000181, -0.00000219, -0.00000238, -0.00000236, -0.00000214, -0.00000175, -0.00000124, -0.00000071, -0.00000024, 0.00000009, 0.00000024, 0.00000019, },
		{ 0.00000049, 0.00000127, 0.00000205, 0.00000266, 0.00000296, 0.00000283, 0.00000223, 0.00000121, -0.00000010, -0.00000148, -0.00000270, -0.00000351, -0.00000372, -0.00000322, -0.00000205, -0.00000034, 0.00000166, 0.00000363, 0.00000523, 0.00000618, 0.00000628, 0.00000547, 0.00000387, 0.00000171, -0.00000065, -0.00000281, -0.00000438, -0.00000508, -0.00000476, -0.00000348, -0.00000145, 0.00000095, 0.00000326, 0.00000500, 0.00000577, 0.00000532, 0.00000360, 0.00000078, -0.00000276, -0.00000652, -0.00000990, -0.00001236, -0.00001348, -0.00001305, -0.00001109, -0.00000790, -0.00000397, 0.00000007, 0.00000352, 0.00000579, 0.00000645, 0.00000529, 0.00000242, -0.00000181, -0.00000681, -0.00001187, -0.00001624, -0.00001926, -0.00002045, -0.00001959, -0.00001673, -0.00001219, -0.00000649, -0.00000030, 0.00000569, 0.00001086, 0.00001474, 0.00001708, 0.00001783, 0.00001716, 0.00001540, 0.00001298, 0.00001034, 0.00000786, 0.00000581, 0.00000432, 0.00000339, 0.00000289, 0.00000263, 0.00000237, 0.00000194, 0.00000120, 0.00000014, -0.00000117, -0.00000258, -0.00000388, -0.00000487, -0.00000538, -0.00000531, -0.00000463, -0.00000343, -0.00000186, -0.00000011, 0.00000159, 0.00000306, 0.00000414, 0.00000476, 0.00000490, 0.00000462, 0.00000401, 0.00000319, 0.00000230, 0.00000141, 0.00000061, -0.00000008, -0.00000065, -0.00000113, -0.00000152, -0.00000185, -0.00000211, -0.00000229, -0.00000235, -0.00000227, -0.00000203, -0.00000165, -0.00000116, -0.00000061, -0.00000009, 0.00000032, 0.00000056, 0.00000059, 0.00000043, 0.00000013, -0.00000022, -0.00000052, -0.00000064, -0.00000053, -0.00000014, },
		{ 0.00000063, 0.00000001, -0.00000080, -0.00000162, -0.00000228, -0.00000259, -0.00000245, -0.00000183, -0.00000081, 0.00000047, 0.00000177, 0.00000284, 0.00000345, 0.00000345, 0.00000278, 0.00000151, -0.00000018, -0.00000200, -0.00000363, -0.00000478, -0.00000521, -0.00000481, -0.00000362, -0.00000185, 0.00000019, 0.00000210, 0.00000348, 0.00000401, 0.00000347, 0.00000186, -0.00000065, -0.00000373, -0.00000690, -0.00000965, -0.00001151, -0.00001212, -0.00001132, -0.00000918, -0.00000596, -0.00000213, 0.00000175, 0.00000510, 0.00000744, 0.00000845, 0.00000802, 0.00000633, 0.00000374, 0.00000081, -0.00000184, -0.00000362, -0.00000408, -0.00000296, -0.00000028, 0.00000367, 0.00000840, 0.00001324, 0.00001749, 0.00002050, 0.00002178, 0.00002109, 0.00001844, 0.00001410, 0.00000859, 0.00000253, -0.00000339, -0.00000855, -0.00001248, -0.00001491, -0.00001580, -0.00001531, -0.00001381, -0.00001172, -0.00000952, -0.00000760, -0.00000625, -0.00000560, -0.00000563, -0.00000615, -0.00000690, -0.00000759, -0.00000794, -0.00000776, -0.00000695, -0.00000554, -0.00000367, -0.00000156, 0.00000054, 0.00000239, 0.00000380, 0.00000466, 0.00000492, 0.00000467, 0.00000401, 0.00000312, 0.00000214, 0.00000123, 0.00000047, -0.00000009, -0.00000045, -0.00000065, -0.00000074, -0.00000077, -0.00000079, -0.00000079, -0.00000076, -0.00000068, -0.00000050, -0.00000022, 0.00000016, 0.00000060, 0.00000104, 0.00000139, 0.00000158, 0.00000155, 0.00000127, 0.00000077, 0.00000011, -0.00000059, -0.00000122, -0.00000165, -0.00000182, -0.00000168, -0.00000126, -0.00000065, 0.00000001, 0.00000059, 0.00000093, 0.00000096, },
		{ -0.00000154, -0.00000106, -0.00000035, 0.00000046, 0.00000120, 0.00000170, 0.00000182, 0.00000150, 0.00000075, -0.00000031, -0.00000149, -0.00000255, -0.00000326, -0.00000345, -0.00000303, -0.00000202, -0.00000059, 0.00000101, 0.00000245, 0.00000341, 0.00000361, 0.00000290, 0.00000128, -0.00000111, -0.00000395, -0.00000683, -0.00000932, -0.00001099, -0.00001155, -0.00001086, -0.00000898, -0.00000618, -0.00000285, 0.00000050, 0.00000337, 0.00000533, 0.00000610, 0.00000562, 0.00000403, 0.00000166, -0.00000102, -0.00000348, -0.00000526, -0.00000600, -0.00000555, -0.00000399, -0.00000161, 0.00000111, 0.00000363, 0.00000537, 0.00000590, 0.00000494, 0.00000248, -0.00000125, -0.00000580, -0.00001054, -0.00001481, -0.00001796, -0.00001951, -0.00001918, -0.00001696, -0.00001313, -0.00000814, -0.00000265, 0.00000268, 0.00000720, 0.00001044, 0.00001211, 0.00001218, 0.00001082, 0.00000842, 0.00000547, 0.00000250, -0.00000003, -0.00000177, -0.00000252, -0.00000228, -0.00000121, 0.00000042, 0.00000224, 0.00000394, 0.00000521, 0.00000587, 0.00000586, 0.00000522, 0.00000409, 0.00000268, 0.00000121, -0.00000015, -0.00000124, -0.00000201, -0.00000246, -0.00000264, -0.00000264, -0.00000256, -0.00000247, -0.00000241, -0.00000239, -0.00000237, -0.00000231, -0.00000217, -0.00000193, -0.00000160, -0.00000124, -0.00000092, -0.00000074, -0.00000078, -0.00000109, -0.00000170, -0.00000253, -0.00000351, -0.00000448, -0.00000530, -0.00000584, -0.00000599, -0.00000573, -0.00000508, -0.00000416, -0.00000311, -0.00000209, -0.00000126, -0.00000071, -0.00000050, -0.00000060, -0.00000091, -0.00000130, -0.00000161, -0.00000172, },
		{ 0.00000209, 0.00000176, 0.00000121, 0.00000053, -0.00000012, -0.00000058, -0.00000072, -0.00000047, 0.00000018, 0.00000111, 0.00000213, 0.00000299, 0.00000346, 0.00000330, 0.00000240, 0.00000075, -0.00000150, -0.00000411, -0.00000672, -0.00000895, -0.00001044, -0.00001092, -0.00001029, -0.00000858, -0.00000602, -0.00000297, 0.00000014, 0.00000286, 0.00000480, 0.00000572, 0.00000555, 0.00000440, 0.00000255, 0.00000042, -0.00000155, -0.00000295, -0.00000349, -0.00000303, -0.00000163, 0.00000044, 0.00000281, 0.00000502, 0.00000662, 0.00000727, 0.00000681, 0.00000528, 0.00000293, 0.00000016, -0.00000248, -0.00000448, -0.00000538, -0.00000490, -0.00000299, 0.00000015, 0.00000410, 0.00000828, 0.00001201, 0.00001466, 0.00001573, 0.00001492, 0.00001220, 0.00000781, 0.00000224, -0.00000388, -0.00000984, -0.00001495, -0.00001867, -0.00002064, -0.00002079, -0.00001924, -0.00001635, -0.00001263, -0.00000862, -0.00000487, -0.00000180, 0.00000030, 0.00000135, 0.00000140, 0.00000066, -0.00000056, -0.00000196, -0.00000325, -0.00000421, -0.00000474, -0.00000482, -0.00000453, -0.00000399, -0.00000334, -0.00000272, -0.00000219, -0.00000180, -0.00000152, -0.00000129, -0.00000104, -0.00000074, -0.00000034, 0.00000010, 0.00000052, 0.00000082, 0.00000089, 0.00000064, 0.00000004, -0.00000089, -0.00000204, -0.00000329, -0.00000443, -0.00000529, -0.00000570, -0.00000559, -0.00000493, -0.00000380, -0.00000234, -0.00000075, 0.00000077, 0.00000203, 0.00000290, 0.00000331, 0.00000331, 0.00000297, 0.00000244, 0.00000188, 0.00000143, 0.00000119, 0.00000119, 0.00000140, 0.00000171, 0.00000200, 0.00000216, },
		{ 0.00000147, 0.00000178, 0.00000224, 0.00000274, 0.00000310, 0.00000315, 0.00000273, 0.00000173, 0.00000015, -0.00000190, -0.00000422, -0.00000653, -0.00000848, -0.00000979, -0.00001022, -0.00000966, -0.00000814, -0.00000584, -0.00000306, -0.00000019, 0.00000237, 0.00000429, 0.00000531, 0.00000537, 0.00000453, 0.00000303, 0.00000120, -0.00000056, -0.00000191, -0.00000256, -0.00000239, -0.00000144, 0.00000009, 0.00000188, 0.00000354, 0.00000470, 0.00000509, 0.00000455, 0.00000312, 0.00000101, -0.00000144, -0.00000380, -0.00000566, -0.00000667, -0.00000664, -0.00000559, -0.00000373, -0.00000144, 0.00000078, 0.00000241, 0.00000297, 0.00000219, -0.00000005, -0.00000356, -0.00000796, -0.00001267, -0.00001701, -0.00002033, -0.00002206, -0.00002185, -0.00001957, -0.00001541, -0.00000978, -0.00000329, 0.00000336, 0.00000946, 0.00001441, 0.00001777, 0.00001936, 0.00001922, 0.00001759, 0.00001491, 0.00001165, 0.00000833, 0.00000536, 0.00000306, 0.00000157, 0.00000088, 0.00000087, 0.00000132, 0.00000200, 0.00000269, 0.00000323, 0.00000355, 0.00000365, 0.00000360, 0.00000349, 0.00000341, 0.00000343, 0.00000353, 0.00000368, 0.00000375, 0.00000362, 0.00000319, 0.00000237, 0.00000118, -0.00000033, -0.00000197, -0.00000356, -0.00000485, -0.00000564, -0.00000579, -0.00000523, -0.00000401, -0.00000227, -0.00000024, 0.00000183, 0.00000367, 0.00000506, 0.00000587, 0.00000606, 0.00000568, 0.00000488, 0.00000383, 0.00000276, 0.00000184, 0.00000119, 0.00000087, 0.00000085, 0.00000106, 0.00000138, 0.00000168, 0.00000187, 0.00000191, 0.00000180, 0.00000161, 0.00000143, 0.00000136, },
		{ 0.00000021, -0.00000125, -0.00000305, -0.00000505, -0.00000701, -0.00000868, -0.00000980, -0.00001015, -0.00000964, -0.00000826, -0.00000616, -0.00000357, -0.00000084, 0.00000170, 0.00000371, 0.00000498, 0.00000539, 0.00000498, 0.00000393, 0.00000251, 0.00000105, -0.00000012, -0.00000074, -0.00000070, 0.00000002, 0.00000126, 0.00000277, 0.00000423, 0.00000532, 0.00000580, 0.00000553, 0.00000452, 0.00000295, 0.00000108, -0.00000073, -0.00000213, -0.00000283, -0.00000268, -0.00000167, 0.00000000, 0.00000204, 0.00000403, 0.00000555, 0.00000626, 0.00000594, 0.00000457, 0.00000232, -0.00000045, -0.00000324, -0.00000553, -0.00000680, -0.00000670, -0.00000506, -0.00000196, 0.00000229, 0.00000718, 0.00001207, 0.00001630, 0.00001927, 0.00002056, 0.00001996, 0.00001752, 0.00001354, 0.00000851, 0.00000304, -0.00000224, -0.00000675, -0.00001006, -0.00001196, -0.00001242, -0.00001160, -0.00000984, -0.00000753, -0.00000508, -0.00000284, -0.00000105, 0.00000018, 0.00000087, 0.00000113, 0.00000112, 0.00000103, 0.00000099, 0.00000106, 0.00000122, 0.00000137, 0.00000138, 0.00000110, 0.00000042, -0.00000070, -0.00000220, -0.00000393, -0.00000566, -0.00000712, -0.00000806, -0.00000827, -0.00000765, -0.00000622, -0.00000412, -0.00000159, 0.00000106, 0.00000350, 0.00000543, 0.00000665, 0.00000704, 0.00000665, 0.00000561, 0.00000414, 0.00000251, 0.00000099, -0.00000020, -0.00000094, -0.00000118, -0.00000096, -0.00000040, 0.00000034, 0.00000111, 0.00000176, 0.00000222, 0.00000248, 0.00000256, 0.00000252, 0.00000246, 0.00000241, 0.00000239, 0.00000236, 0.00000224, 0.00000191, 0.00000126, },
		{ -0.00000917, -0.00000873, -0.00000752, -0.00000564, -0.00000327, -0.00000069, 0.00000179, 0.00000388, 0.00000534, 0.00000604, 0.00000597, 0.00000526, 0.00000411, 0.00000281, 0.00000165, 0.00000086, 0.00000061, 0.00000091, 0.00000170, 0.00000276, 0.00000384, 0.00000467, 0.00000503, 0.00000478, 0.00000391, 0.00000253, 0.00000086, -0.00000080, -0.00000217, -0.00000300, -0.00000313, -0.00000257, -0.00000145, -0.00000001, 0.00000139, 0.00000242, 0.00000278, 0.00000229, 0.00000092, -0.00000119, -0.00000373, -0.00000632, -0.00000851, -0.00000991, -0.00001024, -0.00000940, -0.00000748, -0.00000476, -0.00000168, 0.00000126, 0.00000354, 0.00000477, 0.00000468, 0.00000326, 0.00000067, -0.00000269, -0.00000631, -0.00000962, -0.00001208, -0.00001329, -0.00001303, -0.00001129, -0.00000828, -0.00000440, -0.00000014, 0.00000397, 0.00000746, 0.00000999, 0.00001139, 0.00001165, 0.00001094, 0.00000957, 0.00000786, 0.00000614, 0.00000466, 0.00000356, 0.00000285, 0.00000242, 0.00000207, 0.00000158, 0.00000078, -0.00000045, -0.00000211, -0.00000406, -0.00000608, -0.00000786, -0.00000912, -0.00000958, -0.00000911, -0.00000767, -0.00000540, -0.00000256, 0.00000048, 0.00000335, 0.00000568, 0.00000718, 0.00000770, 0.00000724, 0.00000594, 0.00000405, 0.00000192, -0.00000014, -0.00000180, -0.00000286, -0.00000322, -0.00000289, -0.00000201, -0.00000076, 0.00000063, 0.00000193, 0.00000299, 0.00000371, 0.00000405, 0.00000407, 0.00000384, 0.00000346, 0.00000301, 0.00000251, 0.00000197, 0.00000133, 0.00000053, -0.00000050, -0.00000178, -0.00000329, -0.00000493, -0.00000655, -0.00000793, -0.00000886, },
		{ 0.00000266, 0.00000426, 0.00000520, 0.00000542, 0.00000499, 0.00000406, 0.00000286, 0.00000166, 0.00000070, 0.00000013, 0.00000005, 0.00000042, 0.00000110, 0.00000189, 0.00000256, 0.00000292, 0.00000281, 0.00000219, 0.00000113, -0.00000021, -0.00000160, -0.00000278, -0.00000355, -0.00000376, -0.00000338, -0.00000250, -0.00000132, -0.00000011, 0.00000083, 0.00000125, 0.00000099, 0.00000001, -0.00000159, -0.00000357, -0.00000559, -0.00000728, -0.00000832, -0.00000845, -0.00000757, -0.00000575, -0.00000321, -0.00000030, 0.00000257, 0.00000499, 0.00000662, 0.00000729, 0.00000696, 0.00000580, 0.00000412, 0.00000231, 0.00000080, -0.00000006, -0.00000002, 0.00000099, 0.00000283, 0.00000523, 0.00000778, 0.00001002, 0.00001152, 0.00001197, 0.00001121, 0.00000927, 0.00000639, 0.00000291, -0.00000070, -0.00000400, -0.00000662, -0.00000829, -0.00000896, -0.00000873, -0.00000786, -0.00000669, -0.00000561, -0.00000495, -0.00000491, -0.00000557, -0.00000681, -0.00000837, -0.00000990, -0.00001103, -0.00001141, -0.00001081, -0.00000916, -0.00000656, -0.00000327, 0.00000030, 0.00000372, 0.00000656, 0.00000845, 0.00000921, 0.00000879, 0.00000734, 0.00000516, 0.00000261, 0.00000014, -0.00000190, -0.00000321, -0.00000363, -0.00000317, -0.00000196, -0.00000023, 0.00000171, 0.00000358, 0.00000513, 0.00000618, 0.00000666, 0.00000659, 0.00000608, 0.00000525, 0.00000424, 0.00000318, 0.00000212, 0.00000108, 0.00000004, -0.00000106, -0.00000227, -0.00000360, -0.00000501, -0.00000640, -0.00000763, -0.00000853, -0.00000892, -0.00000868, -0.00000776, -0.00000620, -0.00000414, -0.00000180, 0.00000056, },
		{ 0.00000336, 0.00000230, 0.00000151, 0.00000113, 0.00000117, 0.00000155, 0.00000214, 0.00000274, 0.00000315, 0.00000323, 0.00000289, 0.00000216, 0.00000114, -0.00000000, -0.00000105, -0.00000182, -0.00000215, -0.00000201, -0.00000143, -0.00000056, 0.00000038, 0.00000115, 0.00000152, 0.00000134, 0.00000056, -0.00000077, -0.00000245, -0.00000421, -0.00000574, -0.00000674, -0.00000700, -0.00000640, -0.00000498, -0.00000291, -0.00000047, 0.00000199, 0.00000410, 0.00000559, 0.00000626, 0.00000606, 0.00000511, 0.00000362, 0.00000191, 0.00000030, -0.00000092, -0.00000157, -0.00000159, -0.00000107, -0.00000024, 0.00000061, 0.00000118, 0.00000118, 0.00000047, -0.00000097, -0.00000299, -0.00000529, -0.00000751, -0.00000924, -0.00001011, -0.00000989, -0.00000852, -0.00000612, -0.00000299, 0.00000042, 0.00000362, 0.00000612, 0.00000756, 0.00000772, 0.00000662, 0.00000445, 0.00000163, -0.00000135, -0.00000395, -0.00000570, -0.00000625, -0.00000548, -0.00000346, -0.00000049, 0.00000298, 0.00000642, 0.00000932, 0.00001123, 0.00001190, 0.00001124, 0.00000940, 0.00000669, 0.00000356, 0.00000049, -0.00000206, -0.00000374, -0.00000435, -0.00000386, -0.00000243, -0.00000034, 0.00000206, 0.00000439, 0.00000633, 0.00000763, 0.00000818, 0.00000799, 0.00000717, 0.00000590, 0.00000438, 0.00000281, 0.00000132, -0.00000002, -0.00000121, -0.00000228, -0.00000331, -0.00000436, -0.00000544, -0.00000653, -0.00000752, -0.00000828, -0.00000865, -0.00000851, -0.00000776, -0.00000640, -0.00000451, -0.00000226, 0.00000013, 0.00000238, 0.00000425, 0.00000555, 0.00000618, 0.00000613, 0.00000551, 0.00000451, },
		{ 0.00000131, 0.00000170, 0.00000189, 0.00000176, 0.00000128, 0.00000049, -0.00000048, -0.00000147, -0.00000229, -0.00000280, -0.00000292, -0.00000266, -0.00000209, -0.00000139, -0.00000076, -0.00000039, -0.00000044, -0.00000098, -0.00000198, -0.00000331, -0.00000476, -0.00000606, -0.00000695, -0.00000722, -0.00000674, -0.00000553, -0.00000369, -0.00000147, 0.00000085, 0.00000296, 0.00000458, 0.00000552, 0.00000572, 0.00000522, 0.00000419, 0.00000288, 0.00000156, 0.00000049, -0.00000017, -0.00000033, -0.00000003, 0.00000058, 0.00000130, 0.00000190, 0.00000219, 0.00000206, 0.00000153, 0.00000070, -0.00000020, -0.00000089, -0.00000110, -0.00000062, 0.00000063, 0.00000260, 0.00000505, 0.00000762, 0.00000985, 0.00001130, 0.00001156, 0.00001040, 0.00000777, 0.00000386, -0.00000094, -0.00000607, -0.00001088, -0.00001475, -0.00001717, -0.00001779, -0.00001655, -0.00001362, -0.00000943, -0.00000457, 0.00000029, 0.00000450, 0.00000751, 0.00000901, 0.00000888, 0.00000730, 0.00000462, 0.00000137, -0.00000187, -0.00000455, -0.00000626, -0.00000674, -0.00000595, -0.00000405, -0.00000137, 0.00000167, 0.00000462, 0.00000708, 0.00000874, 0.00000946, 0.00000924, 0.00000821, 0.00000660, 0.00000467, 0.00000269, 0.00000085, -0.00000071, -0.00000197, -0.00000297, -0.00000380, -0.00000456, -0.00000534, -0.00000615, -0.00000695, -0.00000763, -0.00000806, -0.00000808, -0.00000757, -0.00000647, -0.00000482, -0.00000272, -0.00000039, 0.00000193, 0.00000399, 0.00000555, 0.00000647, 0.00000669, 0.00000627, 0.00000533, 0.00000411, 0.00000283, 0.00000171, 0.00000092, 0.00000054, 0.00000056, 0.00000086, },
		{ -0.00000179, -0.00000258, -0.00000319, -0.00000349, -0.00000344, -0.00000309, -0.00000254, -0.00000196, -0.00000152, -0.00000138, -0.00000163, -0.00000229, -0.00000327, -0.00000441, -0.00000550, -0.00000630, -0.00000661, -0.00000629, -0.00000532, -0.00000376, -0.00000179, 0.00000034, 0.00000236, 0.00000401, 0.00000510, 0.00000552, 0.00000528, 0.00000450, 0.00000336, 0.00000209, 0.00000092, 0.00000002, -0.00000049, -0.00000063, -0.00000048, -0.00000018, 0.00000010, 0.00000020, 0.00000002, -0.00000045, -0.00000113, -0.00000188, -0.00000249, -0.00000278, -0.00000261, -0.00000193, -0.00000083, 0.00000050, 0.00000176, 0.00000259, 0.00000266, 0.00000172, -0.00000035, -0.00000345, -0.00000733, -0.00001153, -0.00001550, -0.00001862, -0.00002037, -0.00002035, -0.00001838, -0.00001456, -0.00000922, -0.00000293, 0.00000358, 0.00000957, 0.00001434, 0.00001736, 0.00001836, 0.00001733, 0.00001454, 0.00001049, 0.00000585, 0.00000131, -0.00000248, -0.00000503, -0.00000605, -0.00000551, -0.00000362, -0.00000078, 0.00000252, 0.00000573, 0.00000837, 0.00001008, 0.00001068, 0.00001016, 0.00000868, 0.00000652, 0.00000401, 0.00000148, -0.00000081, -0.00000269, -0.00000411, -0.00000511, -0.00000581, -0.00000637, -0.00000691, -0.00000751, -0.00000818, -0.00000883, -0.00000934, -0.00000954, -0.00000926, -0.00000841, -0.00000696, -0.00000499, -0.00000266, -0.00000021, 0.00000207, 0.00000395, 0.00000521, 0.00000575, 0.00000557, 0.00000477, 0.00000354, 0.00000213, 0.00000077, -0.00000031, -0.00000100, -0.00000124, -0.00000109, -0.00000068, -0.00000018, 0.00000023, 0.00000040, 0.00000025, -0.00000022, -0.00000094, },
		{ -0.00000185, -0.00000142, -0.00000116, -0.00000121, -0.00000161, -0.00000233, -0.00000326, -0.00000423, -0.00000502, -0.00000545, -0.00000535, -0.00000467, -0.00000342, -0.00000173, 0.00000020, 0.00000213, 0.00000382, 0.00000505, 0.00000570, 0.00000574, 0.00000521, 0.00000426, 0.00000308, 0.00000188, 0.00000084, 0.00000008, -0.00000035, -0.00000049, -0.00000043, -0.00000030, -0.00000022, -0.00000030, -0.00000056, -0.00000096, -0.00000140, -0.00000174, -0.00000184, -0.00000159, -0.00000096, -0.00000001, 0.00000112, 0.00000219, 0.00000292, 0.00000306, 0.00000243, 0.00000095, -0.00000129, -0.00000404, -0.00000693, -0.00000950, -0.00001126, -0.00001182, -0.00001091, -0.00000847, -0.00000467, 0.00000010, 0.00000527, 0.00001018, 0.00001414, 0.00001660, 0.00001716, 0.00001570, 0.00001237, 0.00000758, 0.00000195, -0.00000381, -0.00000893, -0.00001278, -0.00001491, -0.00001512, -0.00001347, -0.00001028, -0.00000608, -0.00000149, 0.00000286, 0.00000641, 0.00000876, 0.00000973, 0.00000934, 0.00000779, 0.00000543, 0.00000268, -0.00000008, -0.00000250, -0.00000438, -0.00000566, -0.00000639, -0.00000672, -0.00000685, -0.00000697, -0.00000720, -0.00000759, -0.00000809, -0.00000856, -0.00000882, -0.00000867, -0.00000797, -0.00000663, -0.00000469, -0.00000230, 0.00000031, 0.00000287, 0.00000507, 0.00000668, 0.00000752, 0.00000757, 0.00000687, 0.00000562, 0.00000405, 0.00000244, 0.00000103, 0.00000002, -0.00000050, -0.00000054, -0.00000021, 0.00000034, 0.00000090, 0.00000129, 0.00000139, 0.00000114, 0.00000056, -0.00000023, -0.00000111, -0.00000189, -0.00000245, -0.00000271, -0.00000264, -0.00000231, },
		{ -0.00000333, -0.00000410, -0.00000461, -0.00000470, -0.00000427, -0.00000329, -0.00000183, -0.00000005, 0.00000183, 0.00000359, 0.00000500, 0.00000590, 0.00000622, 0.00000595, 0.00000520, 0.00000411, 0.00000288, 0.00000169, 0.00000069, -0.00000005, -0.00000050, -0.00000071, -0.00000077, -0.00000079, -0.00000085, -0.00000101, -0.00000126, -0.00000154, -0.00000177, -0.00000184, -0.00000166, -0.00000121, -0.00000051, 0.00000032, 0.00000109, 0.00000160, 0.00000163, 0.00000103, -0.00000026, -0.00000218, -0.00000454, -0.00000703, -0.00000928, -0.00001089, -0.00001154, -0.00001100, -0.00000924, -0.00000640, -0.00000282, 0.00000105, 0.00000464, 0.00000743, 0.00000896, 0.00000897, 0.00000741, 0.00000448, 0.00000060, -0.00000366, -0.00000763, -0.00001070, -0.00001234, -0.00001224, -0.00001033, -0.00000681, -0.00000210, 0.00000322, 0.00000847, 0.00001303, 0.00001636, 0.00001814, 0.00001825, 0.00001681, 0.00001413, 0.00001066, 0.00000689, 0.00000330, 0.00000025, -0.00000205, -0.00000352, -0.00000427, -0.00000450, -0.00000446, -0.00000441, -0.00000453, -0.00000490, -0.00000549, -0.00000616, -0.00000670, -0.00000688, -0.00000649, -0.00000543, -0.00000368, -0.00000138, 0.00000127, 0.00000395, 0.00000634, 0.00000816, 0.00000918, 0.00000933, 0.00000864, 0.00000728, 0.00000548, 0.00000356, 0.00000180, 0.00000044, -0.00000038, -0.00000064, -0.00000040, 0.00000017, 0.00000086, 0.00000145, 0.00000176, 0.00000171, 0.00000126, 0.00000050, -0.00000044, -0.00000140, -0.00000220, -0.00000271, -0.00000288, -0.00000273, -0.00000234, -0.00000186, -0.00000146, -0.00000126, -0.00000138, -0.00000181, -0.00000251, },
		{ -0.00000116, 0.00000070, 0.00000254, 0.00000415, 0.00000532, 0.00000593, 0.00000594, 0.00000540, 0.00000442, 0.00000318, 0.00000187, 0.00000065, -0.00000035, -0.00000107, -0.00000151, -0.00000171, -0.00000178, -0.00000179, -0.00000181, -0.00000187, -0.00000196, -0.00000202, -0.00000198, -0.00000179, -0.00000139, -0.00000081, -0.00000012, 0.00000054, 0.00000102, 0.00000114, 0.00000076, -0.00000016, -0.00000159, -0.00000340, -0.00000532, -0.00000707, -0.00000831, -0.00000875, -0.00000820, -0.00000662, -0.00000409, -0.00000090, 0.00000258, 0.00000589, 0.00000857, 0.00001023, 0.00001064, 0.00000975, 0.00000772, 0.00000487, 0.00000168, -0.00000130, -0.00000356, -0.00000469, -0.00000446, -0.00000284, -0.00000005, 0.00000353, 0.00000738, 0.00001092, 0.00001362, 0.00001508, 0.00001507, 0.00001359, 0.00001084, 0.00000720, 0.00000313, -0.00000087, -0.00000434, -0.00000698, -0.00000861, -0.00000927, -0.00000913, -0.00000846, -0.00000761, -0.00000686, -0.00000645, -0.00000648, -0.00000689, -0.00000754, -0.00000816, -0.00000847, -0.00000823, -0.00000727, -0.00000555, -0.00000317, -0.00000035, 0.00000258, 0.00000528, 0.00000740, 0.00000869, 0.00000903, 0.00000842, 0.00000703, 0.00000510, 0.00000296, 0.00000094, -0.00000068, -0.00000172, -0.00000210, -0.00000189, -0.00000125, -0.00000041, 0.00000040, 0.00000095, 0.00000110, 0.00000080, 0.00000008, -0.00000091, -0.00000200, -0.00000300, -0.00000373, -0.00000410, -0.00000409, -0.00000374, -0.00000320, -0.00000263, -0.00000219, -0.00000202, -0.00000219, -0.00000266, -0.00000336, -0.00000410, -0.00000470, -0.00000498, -0.00000478, -0.00000405, -0.00000281, },
		{ 0.00000602, 0.00000514, 0.00000388, 0.00000243, 0.00000098, -0.00000030, -0.00000130, -0.00000199, -0.00000238, -0.00000254, -0.00000254, -0.00000249, -0.00000243, -0.00000239, -0.00000236, -0.00000230, -0.00000216, -0.00000189, -0.00000150, -0.00000101, -0.00000052, -0.00000015, -0.00000002, -0.00000025, -0.00000090, -0.00000197, -0.00000336, -0.00000488, -0.00000630, -0.00000735, -0.00000778, -0.00000742, -0.00000621, -0.00000421, -0.00000163, 0.00000121, 0.00000394, 0.00000615, 0.00000750, 0.00000778, 0.00000691, 0.00000500, 0.00000232, -0.00000072, -0.00000366, -0.00000605, -0.00000751, -0.00000781, -0.00000691, -0.00000494, -0.00000222, 0.00000082, 0.00000372, 0.00000602, 0.00000739, 0.00000765, 0.00000679, 0.00000500, 0.00000258, -0.00000007, -0.00000254, -0.00000448, -0.00000567, -0.00000603, -0.00000564, -0.00000473, -0.00000360, -0.00000258, -0.00000195, -0.00000189, -0.00000246, -0.00000354, -0.00000491, -0.00000627, -0.00000726, -0.00000760, -0.00000708, -0.00000565, -0.00000341, -0.00000061, 0.00000240, 0.00000524, 0.00000751, 0.00000893, 0.00000933, 0.00000870, 0.00000718, 0.00000504, 0.00000263, 0.00000032, -0.00000159, -0.00000285, -0.00000339, -0.00000323, -0.00000254, -0.00000157, -0.00000058, 0.00000016, 0.00000049, 0.00000030, -0.00000037, -0.00000141, -0.00000262, -0.00000379, -0.00000471, -0.00000523, -0.00000531, -0.00000497, -0.00000434, -0.00000358, -0.00000289, -0.00000244, -0.00000233, -0.00000258, -0.00000314, -0.00000385, -0.00000453, -0.00000496, -0.00000498, -0.00000447, -0.00000340, -0.00000186, -0.00000000, 0.00000195, 0.00000376, 0.00000520, 0.00000610, 0.00000637, },
		{ -0.00000244, -0.00000303, -0.00000330, -0.00000334, -0.00000323, -0.00000307, -0.00000293, -0.00000282, -0.00000274, -0.00000264, -0.00000249, -0.00000225, -0.00000193, -0.00000155, -0.00000121, -0.00000100, -0.00000101, -0.00000132, -0.00000194, -0.00000283, -0.00000387, -0.00000488, -0.00000566, -0.00000600, -0.00000575, -0.00000482, -0.00000325, -0.00000118, 0.00000114, 0.00000340, 0.00000527, 0.00000645, 0.00000674, 0.00000606, 0.00000448, 0.00000221, -0.00000039, -0.00000294, -0.00000501, -0.00000627, -0.00000648, -0.00000557, -0.00000365, -0.00000097, 0.00000211, 0.00000515, 0.00000774, 0.00000956, 0.00001040, 0.00001021, 0.00000910, 0.00000730, 0.00000514, 0.00000295, 0.00000103, -0.00000043, -0.00000135, -0.00000179, -0.00000191, -0.00000195, -0.00000216, -0.00000274, -0.00000380, -0.00000534, -0.00000719, -0.00000911, -0.00001076, -0.00001182, -0.00001202, -0.00001117, -0.00000927, -0.00000646, -0.00000304, 0.00000060, 0.00000403, 0.00000681, 0.00000864, 0.00000933, 0.00000886, 0.00000739, 0.00000520, 0.00000268, 0.00000021, -0.00000183, -0.00000319, -0.00000376, -0.00000357, -0.00000277, -0.00000162, -0.00000042, 0.00000054, 0.00000105, 0.00000099, 0.00000036, -0.00000070, -0.00000201, -0.00000332, -0.00000439, -0.00000506, -0.00000522, -0.00000489, -0.00000418, -0.00000327, -0.00000238, -0.00000169, -0.00000137, -0.00000147, -0.00000195, -0.00000268, -0.00000348, -0.00000411, -0.00000437, -0.00000411, -0.00000326, -0.00000186, -0.00000005, 0.00000195, 0.00000391, 0.00000555, 0.00000668, 0.00000716, 0.00000694, 0.00000610, 0.00000477, 0.00000314, 0.00000143, -0.00000015, -0.00000147, },
		{ -0.00000248, -0.00000235, -0.00000227, -0.00000219, -0.00000206, -0.00000185, -0.00000155, -0.00000119, -0.00000085, -0.00000061, -0.00000054, -0.00000070, -0.00000108, -0.00000165, -0.00000227, -0.00000281, -0.00000311, -0.00000301, -0.00000243, -0.00000136, 0.00000011, 0.00000183, 0.00000354, 0.00000499, 0.00000592, 0.00000614, 0.00000558, 0.00000426, 0.00000235, 0.00000013, -0.00000207, -0.00000389, -0.00000502, -0.00000524, -0.00000447, -0.00000278, -0.00000039, 0.00000239, 0.00000517, 0.00000757, 0.00000925, 0.00001001, 0.00000978, 0.00000861, 0.00000671, 0.00000437, 0.00000189, -0.00000043, -0.00000237, -0.00000381, -0.00000476, -0.00000531, -0.00000565, -0.00000596, -0.00000643, -0.00000713, -0.00000807, -0.00000913, -0.00001011, -0.00001074, -0.00001075, -0.00000994, -0.00000820, -0.00000555, -0.00000217, 0.00000164, 0.00000548, 0.00000892, 0.00001156, 0.00001312, 0.00001343, 0.00001252, 0.00001055, 0.00000787, 0.00000486, 0.00000196, -0.00000046, -0.00000212, -0.00000289, -0.00000281, -0.00000205, -0.00000088, 0.00000038, 0.00000141, 0.00000198, 0.00000193, 0.00000126, 0.00000010, -0.00000135, -0.00000283, -0.00000409, -0.00000490, -0.00000516, -0.00000487, -0.00000411, -0.00000309, -0.00000202, -0.00000115, -0.00000064, -0.00000060, -0.00000102, -0.00000177, -0.00000267, -0.00000347, -0.00000393, -0.00000388, -0.00000320, -0.00000190, -0.00000010, 0.00000199, 0.00000412, 0.00000599, 0.00000737, 0.00000807, 0.00000802, 0.00000726, 0.00000590, 0.00000415, 0.00000224, 0.00000040, -0.00000117, -0.00000235, -0.00000310, -0.00000344, -0.00000345, -0.00000326, -0.00000297, -0.00000270, },
		{ -0.00000213, -0.00000167, -0.00000118, -0.00000075, -0.00000046, -0.00000037, -0.00000048, -0.00000075, -0.00000110, -0.00000138, -0.00000148, -0.00000129, -0.00000076, 0.00000009, 0.00000118, 0.00000234, 0.00000338, 0.00000408, 0.00000429, 0.00000391, 0.00000294, 0.00000148, -0.00000026, -0.00000201, -0.00000349, -0.00000440, -0.00000457, -0.00000389, -0.00000241, -0.00000029, 0.00000219, 0.00000469, 0.00000688, 0.00000844, 0.00000917, 0.00000899, 0.00000794, 0.00000618, 0.00000398, 0.00000163, -0.00000058, -0.00000242, -0.00000376, -0.00000458, -0.00000494, -0.00000499, -0.00000494, -0.00000494, -0.00000513, -0.00000553, -0.00000609, -0.00000665, -0.00000701, -0.00000694, -0.00000626, -0.00000487, -0.00000276, -0.00000009, 0.00000290, 0.00000587, 0.00000845, 0.00001030, 0.00001116, 0.00001092, 0.00000960, 0.00000738, 0.00000457, 0.00000157, -0.00000122, -0.00000345, -0.00000487, -0.00000537, -0.00000499, -0.00000392, -0.00000244, -0.00000091, 0.00000037, 0.00000113, 0.00000123, 0.00000065, -0.00000048, -0.00000194, -0.00000346, -0.00000476, -0.00000561, -0.00000586, -0.00000551, -0.00000465, -0.00000347, -0.00000221, -0.00000113, -0.00000043, -0.00000025, -0.00000057, -0.00000132, -0.00000228, -0.00000321, -0.00000384, -0.00000395, -0.00000341, -0.00000220, -0.00000041, 0.00000176, 0.00000403, 0.00000610, 0.00000768, 0.00000857, 0.00000865, 0.00000794, 0.00000653, 0.00000464, 0.00000252, 0.00000043, -0.00000140, -0.00000282, -0.00000374, -0.00000418, -0.00000422, -0.00000398, -0.00000362, -0.00000325, -0.00000298, -0.00000283, -0.00000279, -0.00000281, -0.00000281, -0.00000272, -0.00000250, },
		{ 0.00000094, 0.00000079, 0.00000050, 0.00000019, -0.00000001, -0.00000002, 0.00000022, 0.00000069, 0.00000132, 0.00000196, 0.00000248, 0.00000272, 0.00000257, 0.00000199, 0.00000103, -0.00000019, -0.00000147, -0.00000259, -0.00000331, -0.00000345, -0.00000293, -0.00000175, -0.00000002, 0.00000203, 0.00000413, 0.00000598, 0.00000732, 0.00000794, 0.00000775, 0.00000678, 0.00000517, 0.00000314, 0.00000095, -0.00000112, -0.00000286, -0.00000413, -0.00000489, -0.00000520, -0.00000520, -0.00000507, -0.00000500, -0.00000510, -0.00000544, -0.00000600, -0.00000663, -0.00000716, -0.00000737, -0.00000707, -0.00000611, -0.00000448, -0.00000226, 0.00000035, 0.00000306, 0.00000554, 0.00000748, 0.00000862, 0.00000880, 0.00000804, 0.00000645, 0.00000430, 0.00000192, -0.00000033, -0.00000213, -0.00000324, -0.00000355, -0.00000310, -0.00000205, -0.00000068, 0.00000071, 0.00000181, 0.00000238, 0.00000228, 0.00000150, 0.00000018, -0.00000147, -0.00000316, -0.00000461, -0.00000558, -0.00000593, -0.00000562, -0.00000475, -0.00000352, -0.00000218, -0.00000101, -0.00000022, 0.00000003, -0.00000028, -0.00000105, -0.00000211, -0.00000317, -0.00000397, -0.00000426, -0.00000387, -0.00000275, -0.00000099, 0.00000121, 0.00000358, 0.00000580, 0.00000755, 0.00000860, 0.00000879, 0.00000812, 0.00000667, 0.00000467, 0.00000237, 0.00000006, -0.00000200, -0.00000362, -0.00000469, -0.00000521, -0.00000526, -0.00000498, -0.00000453, -0.00000407, -0.00000372, -0.00000354, -0.00000351, -0.00000359, -0.00000367, -0.00000365, -0.00000345, -0.00000303, -0.00000240, -0.00000162, -0.00000080, -0.00000004, 0.00000054, 0.00000087, },
		{ 0.00000096, 0.00000106, 0.00000133, 0.00000169, 0.00000198, 0.00000209, 0.00000194, 0.00000150, 0.00000082, 0.00000002, -0.00000075, -0.00000129, -0.00000147, -0.00000116, -0.00000035, 0.00000091, 0.00000246, 0.00000409, 0.00000556, 0.00000664, 0.00000716, 0.00000700, 0.00000619, 0.00000481, 0.00000305, 0.00000113, -0.00000069, -0.00000223, -0.00000335, -0.00000401, -0.00000425, -0.00000420, -0.00000401, -0.00000387, -0.00000390, -0.00000417, -0.00000466, -0.00000526, -0.00000580, -0.00000607, -0.00000587, -0.00000505, -0.00000355, -0.00000144, 0.00000112, 0.00000386, 0.00000645, 0.00000859, 0.00001000, 0.00001049, 0.00001002, 0.00000868, 0.00000668, 0.00000432, 0.00000196, -0.00000009, -0.00000157, -0.00000234, -0.00000239, -0.00000184, -0.00000091, 0.00000012, 0.00000097, 0.00000139, 0.00000125, 0.00000052, -0.00000069, -0.00000220, -0.00000375, -0.00000505, -0.00000588, -0.00000610, -0.00000567, -0.00000468, -0.00000332, -0.00000185, -0.00000053, 0.00000039, 0.00000076, 0.00000054, -0.00000019, -0.00000125, -0.00000235, -0.00000322, -0.00000360, -0.00000329, -0.00000223, -0.00000049, 0.00000176, 0.00000422, 0.00000657, 0.00000848, 0.00000967, 0.00000998, 0.00000937, 0.00000792, 0.00000584, 0.00000340, 0.00000092, -0.00000133, -0.00000313, -0.00000434, -0.00000494, -0.00000501, -0.00000470, -0.00000419, -0.00000367, -0.00000326, -0.00000305, -0.00000304, -0.00000317, -0.00000331, -0.00000335, -0.00000316, -0.00000268, -0.00000189, -0.00000084, 0.00000034, 0.00000152, 0.00000254, 0.00000327, 0.00000363, 0.00000362, 0.00000327, 0.00000271, 0.00000208, 0.00000150, 0.00000111, },
		},
		{ { -0.00000211, -0.00000193, -0.00000142, -0.00000069, 0.00000000, 0.00000035, 0.00000009, -0.00000079, -0.00000198, -0.00000300, -0.00000344, -0.00000332, -0.00000308, -0.00000334, -0.00000442, -0.00000597, -0.00000697, -0.00000625, -0.00000318, 0.00000181, 0.00000729, 0.00001143, 0.00001298, 0.00001183, 0.00000910, 0.00000635, 0.00000472, 0.00000429, 0.00000421, 0.00000342, 0.00000158, -0.00000058, -0.00000162, -0.00000038, 0.00000311, 0.00000736, 0.00001014, 0.00000972, 0.00000589, 0.00000015, -0.00000516, -0.00000826, -0.00000892, -0.00000847, -0.00000872, -0.00001053, -0.00001295, -0.00001363, -0.00001028, -0.00000242, 0.00000784, 0.00001650, 0.00001966, 0.00001570, 0.00000631, -0.00000419, -0.00001110, -0.00001173, -0.00000671, 0.00000059, 0.00000614, 0.00000750, 0.00000509, 0.00000158, -0.00000002, 0.00000158, 0.00000509, 0.00000750, 0.00000614, 0.00000059, -0.00000671, -0.00001173, -0.00001110, -0.00000419, 0.00000631, 0.00001570, 0.00001966, 0.00001650, 0.00000784, -0.00000242, -0.00001028, -0.00001363, -0.00001295, -0.00001053, -0.00000872, -0.00000847, -0.00000892, -0.00000826, -0.00000516, 0.00000015, 0.00000589, 0.00000972, 0.00001014, 0.00000736, 0.00000311, -0.00000038, -0.00000162, -0.00000058, 0.00000158, 0.00000342, 0.00000421, 0.00000429, 0.00000472, 0.00000635, 0.00000910, 0.00001183, 0.00001298, 0.00001143, 0.00000729, 0.00000181, -0.00000318, -0.00000625, -0.00000697, -0.00000597, -0.00000442, -0.00000334, -0.00000308, -0.00000332, -0.00000344, -0.00000300, -0.00000198, -0.00000079, 0.00000009, 0.00000035, 0.00000000, -0.00000069, -0.00000142, -0.00000193, },
		{ -0.00000095, -0.00000086, -0.00000124, -0.00000209, -0.00000312, -0.00000387, -0.00000387, -0.00000302, -0.00000168, -0.00000052, -0.00000017, -0.00000069, -0.00000148, -0.00000151, 0.00000006, 0.00000330, 0.00000731, 0.00001064, 0.00001205, 0.00001122, 0.00000892, 0.00000654, 0.00000521, 0.00000511, 0.00000541, 0.00000485, 0.00000272, -0.00000062, -0.00000374, -0.00000508, -0.00000392, -0.00000101, 0.00000187, 0.00000285, 0.00000122, -0.00000218, -0.00000540, -0.00000668, -0.00000557, -0.00000324, -0.00000171, -0.00000242, -0.00000508, -0.00000763, -0.00000747, -0.00000312, 0.00000447, 0.00001214, 0.00001611, 0.00001410, 0.00000680, -0.00000230, -0.00000874, -0.00000945, -0.00000446, 0.00000319, 0.00000927, 0.00001070, 0.00000727, 0.00000154, -0.00000288, -0.00000360, -0.00000087, 0.00000269, 0.00000391, 0.00000125, -0.00000407, -0.00000864, -0.00000893, -0.00000352, 0.00000577, 0.00001464, 0.00001858, 0.00001520, 0.00000555, -0.00000654, -0.00001645, -0.00002107, -0.00002009, -0.00001561, -0.00001053, -0.00000682, -0.00000455, -0.00000237, 0.00000117, 0.00000632, 0.00001175, 0.00001528, 0.00001523, 0.00001147, 0.00000566, 0.00000029, -0.00000261, -0.00000248, -0.00000025, 0.00000241, 0.00000421, 0.00000493, 0.00000530, 0.00000625, 0.00000801, 0.00000980, 0.00001021, 0.00000808, 0.00000335, -0.00000275, -0.00000824, -0.00001134, -0.00001133, -0.00000880, -0.00000517, -0.00000199, -0.00000012, 0.00000042, 0.00000023, -0.00000000, 0.00000007, 0.00000035, 0.00000048, 0.00000018, -0.00000055, -0.00000146, -0.00000224, -0.00000266, -0.00000267, -0.00000237, -0.00000188, -0.00000136, },
		{ -0.00000171, -0.00000190, -0.00000151, -0.00000041, 0.00000114, 0.00000252, 0.00000306, 0.00000248, 0.00000118, 0.00000011, 0.00000030, 0.00000223, 0.00000548, 0.00000886, 0.00001106, 0.00001135, 0.00000997, 0.00000794, 0.00000628, 0.00000538, 0.00000466, 0.00000304, -0.00000024, -0.00000490, -0.00000950, -0.00001216, -0.00001163, -0.00000808, -0.00000310, 0.00000114, 0.00000312, 0.00000277, 0.00000137, 0.00000057, 0.00000120, 0.00000263, 0.00000321, 0.00000150, -0.00000251, -0.00000700, -0.00000918, -0.00000696, -0.00000039, 0.00000799, 0.00001435, 0.00001561, 0.00001125, 0.00000366, -0.00000311, -0.00000564, -0.00000308, 0.00000244, 0.00000713, 0.00000783, 0.00000394, -0.00000228, -0.00000707, -0.00000753, -0.00000343, 0.00000263, 0.00000680, 0.00000637, 0.00000148, -0.00000488, -0.00000865, -0.00000716, -0.00000083, 0.00000690, 0.00001149, 0.00000979, 0.00000185, -0.00000911, -0.00001842, -0.00002241, -0.00002009, -0.00001337, -0.00000558, 0.00000043, 0.00000372, 0.00000529, 0.00000694, 0.00000974, 0.00001324, 0.00001572, 0.00001544, 0.00001185, 0.00000611, 0.00000051, -0.00000279, -0.00000293, -0.00000068, 0.00000218, 0.00000396, 0.00000406, 0.00000310, 0.00000227, 0.00000238, 0.00000320, 0.00000362, 0.00000243, -0.00000083, -0.00000532, -0.00000932, -0.00001108, -0.00000976, -0.00000588, -0.00000103, 0.00000295, 0.00000486, 0.00000455, 0.00000277, 0.00000062, -0.00000106, -0.00000202, -0.00000244, -0.00000269, -0.00000298, -0.00000327, -0.00000338, -0.00000320, -0.00000275, -0.00000218, -0.00000164, -0.00000122, -0.00000095, -0.00000084, -0.00000096, -0.00000130, },
		{ 0.00000240, 0.00000334, 0.00000367, 0.00000305, 0.00000171, 0.00000042, 0.00000017, 0.00000163, 0.00000463, 0.00000816, 0.00001085, 0.00001163, 0.00001032, 0.00000763, 0.00000467, 0.00000216, 0.00000005, -0.00000236, -0.00000575, -0.00001000, -0.00001391, -0.00001571, -0.00001401, -0.00000875, -0.00000144, 0.00000547, 0.00000979, 0.00001072, 0.00000906, 0.00000654, 0.00000463, 0.00000362, 0.00000258, 0.00000027, -0.00000364, -0.00000789, -0.00001006, -0.00000801, -0.00000136, 0.00000782, 0.00001576, 0.00001892, 0.00001585, 0.00000814, -0.00000044, -0.00000598, -0.00000653, -0.00000310, 0.00000117, 0.00000301, 0.00000113, -0.00000302, -0.00000625, -0.00000571, -0.00000094, 0.00000565, 0.00001005, 0.00000905, 0.00000231, -0.00000719, -0.00001472, -0.00001637, -0.00001132, -0.00000236, 0.00000562, 0.00000833, 0.00000439, -0.00000403, -0.00001246, -0.00001665, -0.00001475, -0.00000802, 0.00000020, 0.00000652, 0.00000932, 0.00000930, 0.00000856, 0.00000901, 0.00001108, 0.00001350, 0.00001432, 0.00001226, 0.00000763, 0.00000218, -0.00000196, -0.00000354, -0.00000285, -0.00000137, -0.00000074, -0.00000170, -0.00000365, -0.00000520, -0.00000512, -0.00000325, -0.00000060, 0.00000125, 0.00000119, -0.00000070, -0.00000319, -0.00000458, -0.00000380, -0.00000098, 0.00000259, 0.00000520, 0.00000558, 0.00000360, 0.00000016, -0.00000336, -0.00000579, -0.00000664, -0.00000617, -0.00000501, -0.00000375, -0.00000267, -0.00000182, -0.00000111, -0.00000051, -0.00000008, 0.00000013, 0.00000016, 0.00000012, 0.00000013, 0.00000021, 0.00000028, 0.00000032, 0.00000039, 0.00000069, 0.00000139, },
		{ 0.00000228, 0.00000134, 0.00000102, 0.00000195, 0.00000419, 0.00000697, 0.00000907, 0.00000936, 0.00000747, 0.00000399, 0.00000009, -0.00000314, -0.00000532, -0.00000689, -0.00000856, -0.00001064, -0.00001250, -0.00001279, -0.00001020, -0.00000447, 0.00000318, 0.00001042, 0.00001488, 0.00001537, 0.00001241, 0.00000788, 0.00000381, 0.00000130, 0.00000003, -0.00000118, -0.00000319, -0.00000566, -0.00000696, -0.00000521, 0.00000034, 0.00000829, 0.00001549, 0.00001850, 0.00001549, 0.00000748, -0.00000211, -0.00000912, -0.00001090, -0.00000764, -0.00000207, 0.00000225, 0.00000312, 0.00000089, -0.00000193, -0.00000256, 0.00000006, 0.00000431, 0.00000678, 0.00000439, -0.00000340, -0.00001380, -0.00002192, -0.00002336, -0.00001673, -0.00000463, 0.00000757, 0.00001447, 0.00001344, 0.00000585, -0.00000379, -0.00001040, -0.00001095, -0.00000576, 0.00000203, 0.00000855, 0.00001138, 0.00001066, 0.00000852, 0.00000747, 0.00000869, 0.00001135, 0.00001329, 0.00001247, 0.00000833, 0.00000208, -0.00000407, -0.00000820, -0.00000975, -0.00000953, -0.00000900, -0.00000907, -0.00000954, -0.00000922, -0.00000695, -0.00000251, 0.00000298, 0.00000757, 0.00000953, 0.00000832, 0.00000490, 0.00000113, -0.00000128, -0.00000166, -0.00000065, 0.00000035, 0.00000007, -0.00000188, -0.00000481, -0.00000744, -0.00000860, -0.00000786, -0.00000561, -0.00000275, -0.00000020, 0.00000153, 0.00000236, 0.00000252, 0.00000229, 0.00000182, 0.00000125, 0.00000068, 0.00000029, 0.00000023, 0.00000050, 0.00000095, 0.00000134, 0.00000155, 0.00000166, 0.00000187, 0.00000231, 0.00000289, 0.00000325, 0.00000306, },
		{ 0.00000213, 0.00000350, 0.00000457, 0.00000442, 0.00000260, -0.00000048, -0.00000374, -0.00000606, -0.00000688, -0.00000657, -0.00000603, -0.00000602, -0.00000652, -0.00000657, -0.00000488, -0.00000074, 0.00000526, 0.00001132, 0.00001522, 0.00001554, 0.00001243, 0.00000753, 0.00000303, 0.00000045, -0.00000014, 0.00000010, -0.00000022, -0.00000157, -0.00000307, -0.00000301, -0.00000015, 0.00000514, 0.00001066, 0.00001346, 0.00001153, 0.00000514, -0.00000305, -0.00000928, -0.00001076, -0.00000719, -0.00000096, 0.00000431, 0.00000585, 0.00000330, -0.00000125, -0.00000473, -0.00000524, -0.00000326, -0.00000143, -0.00000270, -0.00000815, -0.00001582, -0.00002153, -0.00002109, -0.00001292, 0.00000066, 0.00001431, 0.00002229, 0.00002142, 0.00001273, 0.00000089, -0.00000829, -0.00001081, -0.00000635, 0.00000197, 0.00000961, 0.00001320, 0.00001216, 0.00000858, 0.00000554, 0.00000509, 0.00000701, 0.00000916, 0.00000893, 0.00000498, -0.00000198, -0.00000958, -0.00001522, -0.00001743, -0.00001641, -0.00001359, -0.00001046, -0.00000760, -0.00000455, -0.00000045, 0.00000495, 0.00001070, 0.00001493, 0.00001579, 0.00001257, 0.00000625, -0.00000096, -0.00000664, -0.00000932, -0.00000906, -0.00000718, -0.00000531, -0.00000450, -0.00000473, -0.00000514, -0.00000473, -0.00000301, -0.00000029, 0.00000255, 0.00000461, 0.00000537, 0.00000493, 0.00000373, 0.00000231, 0.00000104, 0.00000007, -0.00000054, -0.00000072, -0.00000044, 0.00000025, 0.00000114, 0.00000190, 0.00000228, 0.00000228, 0.00000213, 0.00000216, 0.00000250, 0.00000299, 0.00000325, 0.00000297, 0.00000223, 0.00000149, 0.00000136, },
		{ -0.00000076, -0.00000208, -0.00000355, -0.00000423, -0.00000365, -0.00000208, -0.00000052, 0.00000005, -0.00000067, -0.00000193, -0.00000229, -0.00000051, 0.00000358, 0.00000878, 0.00001306, 0.00001462, 0.00001295, 0.00000906, 0.00000489, 0.00000217, 0.00000139, 0.00000167, 0.00000155, 0.00000010, -0.00000230, -0.00000408, -0.00000356, -0.00000024, 0.00000464, 0.00000856, 0.00000918, 0.00000582, 0.00000001, -0.00000528, -0.00000733, -0.00000532, -0.00000088, 0.00000289, 0.00000328, -0.00000037, -0.00000620, -0.00001104, -0.00001239, -0.00000999, -0.00000604, -0.00000369, -0.00000492, -0.00000905, -0.00001283, -0.00001239, -0.00000571, 0.00000581, 0.00001770, 0.00002462, 0.00002328, 0.00001433, 0.00000213, -0.00000748, -0.00001027, -0.00000568, 0.00000307, 0.00001100, 0.00001411, 0.00001144, 0.00000528, -0.00000057, -0.00000309, -0.00000180, 0.00000120, 0.00000273, 0.00000058, -0.00000517, -0.00001223, -0.00001749, -0.00001874, -0.00001576, -0.00001012, -0.00000407, 0.00000084, 0.00000436, 0.00000720, 0.00001006, 0.00001277, 0.00001411, 0.00001261, 0.00000759, -0.00000009, -0.00000818, -0.00001402, -0.00001582, -0.00001346, -0.00000845, -0.00000303, 0.00000095, 0.00000281, 0.00000307, 0.00000283, 0.00000299, 0.00000377, 0.00000469, 0.00000507, 0.00000449, 0.00000306, 0.00000130, -0.00000024, -0.00000120, -0.00000156, -0.00000145, -0.00000101, -0.00000030, 0.00000065, 0.00000168, 0.00000252, 0.00000292, 0.00000280, 0.00000237, 0.00000202, 0.00000204, 0.00000241, 0.00000275, 0.00000259, 0.00000170, 0.00000031, -0.00000097, -0.00000156, -0.00000131, -0.00000062, -0.00000025, },
		{ 0.00000164, 0.00000294, 0.00000396, 0.00000380, 0.00000208, -0.00000057, -0.00000261, -0.00000245, 0.00000065, 0.00000593, 0.00001142, 0.00001490, 0.00001503, 0.00001209, 0.00000767, 0.00000372, 0.00000140, 0.00000053, -0.00000006, -0.00000146, -0.00000377, -0.00000585, -0.00000601, -0.00000328, 0.00000173, 0.00000688, 0.00000959, 0.00000837, 0.00000372, -0.00000207, -0.00000626, -0.00000728, -0.00000566, -0.00000361, -0.00000350, -0.00000627, -0.00001065, -0.00001389, -0.00001356, -0.00000915, -0.00000264, 0.00000261, 0.00000394, 0.00000110, -0.00000334, -0.00000549, -0.00000251, 0.00000550, 0.00001523, 0.00002190, 0.00002191, 0.00001501, 0.00000453, -0.00000441, -0.00000770, -0.00000450, 0.00000240, 0.00000827, 0.00000922, 0.00000439, -0.00000370, -0.00001072, -0.00001310, -0.00001008, -0.00000400, 0.00000121, 0.00000236, -0.00000117, -0.00000726, -0.00001234, -0.00001354, -0.00001016, -0.00000386, 0.00000253, 0.00000667, 0.00000788, 0.00000709, 0.00000589, 0.00000521, 0.00000478, 0.00000342, 0.00000013, -0.00000492, -0.00001018, -0.00001337, -0.00001279, -0.00000825, -0.00000131, 0.00000551, 0.00000995, 0.00001101, 0.00000922, 0.00000608, 0.00000318, 0.00000140, 0.00000071, 0.00000051, 0.00000020, -0.00000043, -0.00000117, -0.00000163, -0.00000161, -0.00000117, -0.00000057, -0.00000000, 0.00000052, 0.00000110, 0.00000178, 0.00000244, 0.00000282, 0.00000271, 0.00000218, 0.00000153, 0.00000110, 0.00000103, 0.00000108, 0.00000075, -0.00000030, -0.00000201, -0.00000381, -0.00000493, -0.00000484, -0.00000363, -0.00000191, -0.00000047, 0.00000028, 0.00000050, 0.00000080, },
		{ 0.00000193, -0.00000117, -0.00000373, -0.00000412, -0.00000132, 0.00000424, 0.00001071, 0.00001556, 0.00001686, 0.00001423, 0.00000899, 0.00000337, -0.00000072, -0.00000264, -0.00000308, -0.00000331, -0.00000410, -0.00000508, -0.00000505, -0.00000291, 0.00000129, 0.00000600, 0.00000886, 0.00000806, 0.00000345, -0.00000319, -0.00000912, -0.00001216, -0.00001187, -0.00000966, -0.00000772, -0.00000750, -0.00000867, -0.00000936, -0.00000749, -0.00000243, 0.00000431, 0.00000968, 0.00001092, 0.00000739, 0.00000126, -0.00000356, -0.00000357, 0.00000213, 0.00001113, 0.00001892, 0.00002133, 0.00001690, 0.00000768, -0.00000194, -0.00000775, -0.00000812, -0.00000472, -0.00000132, -0.00000136, -0.00000574, -0.00001221, -0.00001675, -0.00001604, -0.00000955, 0.00000001, 0.00000811, 0.00001100, 0.00000767, 0.00000039, -0.00000674, -0.00001008, -0.00000842, -0.00000337, 0.00000184, 0.00000432, 0.00000316, -0.00000031, -0.00000363, -0.00000481, -0.00000352, -0.00000107, 0.00000063, 0.00000043, -0.00000125, -0.00000273, -0.00000221, 0.00000102, 0.00000603, 0.00001072, 0.00001292, 0.00001162, 0.00000738, 0.00000200, -0.00000251, -0.00000488, -0.00000502, -0.00000375, -0.00000215, -0.00000096, -0.00000035, -0.00000009, 0.00000009, 0.00000029, 0.00000041, 0.00000037, 0.00000026, 0.00000033, 0.00000076, 0.00000147, 0.00000213, 0.00000232, 0.00000184, 0.00000084, -0.00000024, -0.00000106, -0.00000156, -0.00000202, -0.00000282, -0.00000405, -0.00000531, -0.00000588, -0.00000514, -0.00000297, 0.00000004, 0.00000296, 0.00000495, 0.00000578, 0.00000582, 0.00000566, 0.00000559, 0.00000533, 0.00000424, },
		{ -0.00000203, 0.00000319, 0.00000971, 0.00001505, 0.00001696, 0.00001456, 0.00000886, 0.00000220, -0.00000295, -0.00000524, -0.00000494, -0.00000351, -0.00000245, -0.00000234, -0.00000259, -0.00000207, -0.00000018, 0.00000249, 0.00000431, 0.00000357, -0.00000027, -0.00000609, -0.00001154, -0.00001432, -0.00001348, -0.00000991, -0.00000570, -0.00000278, -0.00000167, -0.00000128, 0.00000015, 0.00000365, 0.00000847, 0.00001233, 0.00001275, 0.00000885, 0.00000225, -0.00000352, -0.00000491, -0.00000049, 0.00000795, 0.00001616, 0.00001964, 0.00001606, 0.00000660, -0.00000478, -0.00001350, -0.00001683, -0.00001517, -0.00001142, -0.00000890, -0.00000919, -0.00001111, -0.00001169, -0.00000825, -0.00000044, 0.00000913, 0.00001621, 0.00001723, 0.00001139, 0.00000128, -0.00000855, -0.00001401, -0.00001359, -0.00000897, -0.00000375, -0.00000125, -0.00000256, -0.00000606, -0.00000864, -0.00000767, -0.00000264, 0.00000458, 0.00001089, 0.00001376, 0.00001261, 0.00000894, 0.00000529, 0.00000363, 0.00000428, 0.00000594, 0.00000661, 0.00000494, 0.00000106, -0.00000359, -0.00000708, -0.00000811, -0.00000661, -0.00000361, -0.00000058, 0.00000140, 0.00000201, 0.00000161, 0.00000082, 0.00000006, -0.00000050, -0.00000089, -0.00000107, -0.00000088, -0.00000022, 0.00000080, 0.00000174, 0.00000203, 0.00000131, -0.00000030, -0.00000226, -0.00000386, -0.00000467, -0.00000471, -0.00000438, -0.00000407, -0.00000379, -0.00000316, -0.00000165, 0.00000093, 0.00000418, 0.00000718, 0.00000900, 0.00000922, 0.00000814, 0.00000650, 0.00000502, 0.00000384, 0.00000256, 0.00000068, -0.00000179, -0.00000396, -0.00000442, },
		{ 0.00001537, 0.00001383, 0.00000863, 0.00000191, -0.00000364, -0.00000616, -0.00000547, -0.00000302, -0.00000084, -0.00000030, -0.00000141, -0.00000300, -0.00000376, -0.00000319, -0.00000201, -0.00000162, -0.00000306, -0.00000609, -0.00000924, -0.00001060, -0.00000905, -0.00000502, -0.00000021, 0.00000339, 0.00000482, 0.00000465, 0.00000451, 0.00000580, 0.00000855, 0.00001120, 0.00001157, 0.00000840, 0.00000250, -0.00000337, -0.00000595, -0.00000346, 0.00000307, 0.00001007, 0.00001326, 0.00001001, 0.00000091, -0.00001045, -0.00001937, -0.00002249, -0.00001946, -0.00001278, -0.00000613, -0.00000209, -0.00000081, -0.00000031, 0.00000181, 0.00000640, 0.00001187, 0.00001501, 0.00001298, 0.00000527, -0.00000561, -0.00001535, -0.00002005, -0.00001841, -0.00001229, -0.00000554, -0.00000169, -0.00000187, -0.00000428, -0.00000544, -0.00000243, 0.00000517, 0.00001486, 0.00002254, 0.00002474, 0.00002048, 0.00001175, 0.00000232, -0.00000427, -0.00000640, -0.00000490, -0.00000218, -0.00000061, -0.00000123, -0.00000340, -0.00000545, -0.00000591, -0.00000437, -0.00000160, 0.00000101, 0.00000233, 0.00000207, 0.00000075, -0.00000073, -0.00000173, -0.00000203, -0.00000184, -0.00000141, -0.00000088, -0.00000021, 0.00000050, 0.00000098, 0.00000075, -0.00000048, -0.00000257, -0.00000484, -0.00000638, -0.00000650, -0.00000515, -0.00000287, -0.00000049, 0.00000141, 0.00000279, 0.00000403, 0.00000556, 0.00000736, 0.00000892, 0.00000952, 0.00000873, 0.00000676, 0.00000440, 0.00000252, 0.00000149, 0.00000098, 0.00000028, -0.00000108, -0.00000275, -0.00000357, -0.00000224, 0.00000174, 0.00000746, 0.00001280, },
		{ -0.00000271, -0.00000601, -0.00000593, -0.00000355, -0.00000102, -0.00000026, -0.00000183, -0.00000472, -0.00000713, -0.00000766, -0.00000619, -0.00000384, -0.00000209, -0.00000178, -0.00000249, -0.00000295, -0.00000195, 0.00000075, 0.00000416, 0.00000674, 0.00000745, 0.00000649, 0.00000520, 0.00000504, 0.00000649, 0.00000850, 0.00000911, 0.00000672, 0.00000138, -0.00000488, -0.00000907, -0.00000901, -0.00000476, 0.00000111, 0.00000486, 0.00000371, -0.00000246, -0.00001089, -0.00001741, -0.00001866, -0.00001398, -0.00000566, 0.00000245, 0.00000714, 0.00000753, 0.00000521, 0.00000289, 0.00000243, 0.00000364, 0.00000443, 0.00000245, -0.00000312, -0.00001068, -0.00001691, -0.00001870, -0.00001505, -0.00000776, -0.00000044, 0.00000366, 0.00000360, 0.00000130, 0.00000035, 0.00000364, 0.00001137, 0.00002065, 0.00002686, 0.00002625, 0.00001802, 0.00000494, -0.00000802, -0.00001612, -0.00001713, -0.00001214, -0.00000464, 0.00000148, 0.00000387, 0.00000255, -0.00000059, -0.00000324, -0.00000400, -0.00000294, -0.00000119, -0.00000001, 0.00000001, -0.00000083, -0.00000173, -0.00000206, -0.00000170, -0.00000106, -0.00000063, -0.00000064, -0.00000099, -0.00000142, -0.00000181, -0.00000233, -0.00000320, -0.00000441, -0.00000547, -0.00000562, -0.00000425, -0.00000133, 0.00000241, 0.00000582, 0.00000795, 0.00000848, 0.00000787, 0.00000695, 0.00000634, 0.00000611, 0.00000585, 0.00000504, 0.00000363, 0.00000206, 0.00000103, 0.00000091, 0.00000138, 0.00000159, 0.00000078, -0.00000106, -0.00000293, -0.00000330, -0.00000104, 0.00000367, 0.00000915, 0.00001296, 0.00001313, 0.00000936, 0.00000319, },
		{ -0.00000309, -0.00000146, -0.00000220, -0.00000481, -0.00000746, -0.00000822, -0.00000626, -0.00000230, 0.00000189, 0.00000465, 0.00000536, 0.00000470, 0.00000396, 0.00000409, 0.00000508, 0.00000605, 0.00000606, 0.00000490, 0.00000334, 0.00000254, 0.00000312, 0.00000446, 0.00000493, 0.00000290, -0.00000201, -0.00000831, -0.00001328, -0.00001442, -0.00001100, -0.00000470, 0.00000118, 0.00000357, 0.00000140, -0.00000374, -0.00000850, -0.00000976, -0.00000645, -0.00000022, 0.00000558, 0.00000785, 0.00000557, 0.00000026, -0.00000504, -0.00000768, -0.00000699, -0.00000446, -0.00000259, -0.00000312, -0.00000578, -0.00000849, -0.00000881, -0.00000560, 0.00000001, 0.00000530, 0.00000768, 0.00000650, 0.00000354, 0.00000206, 0.00000463, 0.00001124, 0.00001890, 0.00002310, 0.00002027, 0.00001006, -0.00000416, -0.00001681, -0.00002283, -0.00002017, -0.00001074, 0.00000071, 0.00000907, 0.00001129, 0.00000755, 0.00000070, -0.00000557, -0.00000868, -0.00000811, -0.00000521, -0.00000200, 0.00000004, 0.00000057, 0.00000017, -0.00000039, -0.00000074, -0.00000108, -0.00000178, -0.00000296, -0.00000429, -0.00000517, -0.00000518, -0.00000436, -0.00000315, -0.00000200, -0.00000100, 0.00000021, 0.00000211, 0.00000481, 0.00000780, 0.00001006, 0.00001064, 0.00000926, 0.00000649, 0.00000352, 0.00000149, 0.00000088, 0.00000135, 0.00000214, 0.00000260, 0.00000261, 0.00000252, 0.00000270, 0.00000303, 0.00000289, 0.00000154, -0.00000114, -0.00000423, -0.00000604, -0.00000499, -0.00000069, 0.00000555, 0.00001118, 0.00001357, 0.00001151, 0.00000584, -0.00000091, -0.00000588, -0.00000742, -0.00000584, },
		{ -0.00000318, -0.00000421, -0.00000271, 0.00000124, 0.00000606, 0.00000965, 0.00001065, 0.00000910, 0.00000626, 0.00000367, 0.00000222, 0.00000175, 0.00000149, 0.00000079, -0.00000028, -0.00000107, -0.00000098, -0.00000021, 0.00000016, -0.00000122, -0.00000490, -0.00000993, -0.00001404, -0.00001479, -0.00001108, -0.00000401, 0.00000348, 0.00000811, 0.00000803, 0.00000385, -0.00000174, -0.00000560, -0.00000597, -0.00000342, -0.00000040, 0.00000040, -0.00000216, -0.00000694, -0.00001117, -0.00001218, -0.00000905, -0.00000317, 0.00000262, 0.00000579, 0.00000554, 0.00000318, 0.00000103, 0.00000083, 0.00000260, 0.00000474, 0.00000530, 0.00000356, 0.00000076, -0.00000059, 0.00000154, 0.00000708, 0.00001339, 0.00001647, 0.00001329, 0.00000380, -0.00000858, -0.00001847, -0.00002124, -0.00001550, -0.00000393, 0.00000805, 0.00001497, 0.00001405, 0.00000635, -0.00000404, -0.00001224, -0.00001503, -0.00001215, -0.00000592, 0.00000032, 0.00000400, 0.00000435, 0.00000226, -0.00000069, -0.00000326, -0.00000501, -0.00000613, -0.00000684, -0.00000700, -0.00000619, -0.00000407, -0.00000088, 0.00000260, 0.00000538, 0.00000688, 0.00000721, 0.00000706, 0.00000707, 0.00000742, 0.00000760, 0.00000689, 0.00000489, 0.00000201, -0.00000068, -0.00000203, -0.00000148, 0.00000059, 0.00000307, 0.00000484, 0.00000533, 0.00000474, 0.00000368, 0.00000253, 0.00000119, -0.00000084, -0.00000374, -0.00000688, -0.00000877, -0.00000775, -0.00000315, 0.00000403, 0.00001119, 0.00001525, 0.00001426, 0.00000845, 0.00000025, -0.00000690, -0.00001032, -0.00000941, -0.00000574, -0.00000198, -0.00000031, -0.00000116, },
		{ 0.00000660, 0.00000968, 0.00001050, 0.00000841, 0.00000431, -0.00000004, -0.00000308, -0.00000429, -0.00000421, -0.00000377, -0.00000360, -0.00000361, -0.00000335, -0.00000265, -0.00000198, -0.00000229, -0.00000421, -0.00000733, -0.00001009, -0.00001046, -0.00000727, -0.00000110, 0.00000565, 0.00000991, 0.00000957, 0.00000469, -0.00000246, -0.00000867, -0.00001152, -0.00001065, -0.00000777, -0.00000536, -0.00000507, -0.00000658, -0.00000791, -0.00000680, -0.00000228, 0.00000453, 0.00001092, 0.00001417, 0.00001307, 0.00000863, 0.00000331, -0.00000049, -0.00000181, -0.00000141, -0.00000098, -0.00000164, -0.00000304, -0.00000355, -0.00000153, 0.00000315, 0.00000855, 0.00001150, 0.00000947, 0.00000237, -0.00000693, -0.00001388, -0.00001469, -0.00000850, 0.00000183, 0.00001113, 0.00001456, 0.00001020, 0.00000005, -0.00001103, -0.00001788, -0.00001767, -0.00001111, -0.00000186, 0.00000559, 0.00000821, 0.00000564, -0.00000008, -0.00000596, -0.00000967, -0.00001044, -0.00000889, -0.00000618, -0.00000313, 0.00000007, 0.00000351, 0.00000704, 0.00000997, 0.00001136, 0.00001064, 0.00000806, 0.00000473, 0.00000200, 0.00000071, 0.00000079, 0.00000138, 0.00000158, 0.00000103, 0.00000020, -0.00000004, 0.00000094, 0.00000295, 0.00000503, 0.00000605, 0.00000538, 0.00000329, 0.00000069, -0.00000152, -0.00000304, -0.00000425, -0.00000567, -0.00000728, -0.00000816, -0.00000689, -0.00000255, 0.00000431, 0.00001146, 0.00001581, 0.00001504, 0.00000890, -0.00000033, -0.00000886, -0.00001320, -0.00001191, -0.00000625, 0.00000065, 0.00000554, 0.00000677, 0.00000499, 0.00000245, 0.00000152, 0.00000319, },
		{ 0.00000069, -0.00000365, -0.00000703, -0.00000829, -0.00000753, -0.00000576, -0.00000408, -0.00000300, -0.00000234, -0.00000169, -0.00000099, -0.00000074, -0.00000154, -0.00000348, -0.00000570, -0.00000673, -0.00000534, -0.00000159, 0.00000296, 0.00000586, 0.00000514, 0.00000051, -0.00000624, -0.00001217, -0.00001469, -0.00001298, -0.00000833, -0.00000329, -0.00000001, 0.00000093, 0.00000073, 0.00000142, 0.00000428, 0.00000886, 0.00001313, 0.00001467, 0.00001218, 0.00000637, -0.00000045, -0.00000566, -0.00000784, -0.00000734, -0.00000578, -0.00000475, -0.00000463, -0.00000446, -0.00000277, 0.00000103, 0.00000583, 0.00000926, 0.00000905, 0.00000471, -0.00000186, -0.00000721, -0.00000823, -0.00000418, 0.00000271, 0.00000826, 0.00000872, 0.00000300, -0.00000655, -0.00001528, -0.00001873, -0.00001511, -0.00000626, 0.00000334, 0.00000898, 0.00000819, 0.00000183, -0.00000654, -0.00001276, -0.00001411, -0.00001046, -0.00000385, 0.00000292, 0.00000783, 0.00001034, 0.00001104, 0.00001081, 0.00000997, 0.00000827, 0.00000533, 0.00000139, -0.00000249, -0.00000486, -0.00000473, -0.00000222, 0.00000143, 0.00000455, 0.00000594, 0.00000550, 0.00000409, 0.00000283, 0.00000229, 0.00000223, 0.00000185, 0.00000055, -0.00000154, -0.00000363, -0.00000479, -0.00000469, -0.00000385, -0.00000327, -0.00000356, -0.00000434, -0.00000430, -0.00000206, 0.00000268, 0.00000849, 0.00001271, 0.00001276, 0.00000781, -0.00000046, -0.00000842, -0.00001224, -0.00000991, -0.00000235, 0.00000700, 0.00001402, 0.00001598, 0.00001283, 0.00000700, 0.00000183, -0.00000033, 0.00000076, 0.00000334, 0.00000499, 0.00000408, },
		{ -0.00000735, -0.00000519, -0.00000276, -0.00000106, -0.00000032, -0.00000014, -0.00000008, -0.00000013, -0.00000072, -0.00000221, -0.00000435, -0.00000617, -0.00000652, -0.00000491, -0.00000205, 0.00000037, 0.00000060, -0.00000200, -0.00000638, -0.00001025, -0.00001134, -0.00000867, -0.00000322, 0.00000264, 0.00000652, 0.00000738, 0.00000600, 0.00000434, 0.00000420, 0.00000595, 0.00000832, 0.00000921, 0.00000713, 0.00000219, -0.00000384, -0.00000862, -0.00001055, -0.00000965, -0.00000734, -0.00000533, -0.00000442, -0.00000402, -0.00000274, 0.00000038, 0.00000492, 0.00000903, 0.00001049, 0.00000822, 0.00000313, -0.00000221, -0.00000510, -0.00000434, -0.00000114, 0.00000164, 0.00000119, -0.00000327, -0.00000985, -0.00001491, -0.00001519, -0.00000991, -0.00000141, 0.00000599, 0.00000835, 0.00000441, -0.00000368, -0.00001145, -0.00001453, -0.00001098, -0.00000220, 0.00000796, 0.00001533, 0.00001746, 0.00001458, 0.00000896, 0.00000334, -0.00000052, -0.00000247, -0.00000338, -0.00000409, -0.00000462, -0.00000417, -0.00000194, 0.00000203, 0.00000653, 0.00000969, 0.00001006, 0.00000743, 0.00000298, -0.00000142, -0.00000426, -0.00000512, -0.00000463, -0.00000381, -0.00000328, -0.00000293, -0.00000222, -0.00000091, 0.00000053, 0.00000107, -0.00000007, -0.00000264, -0.00000516, -0.00000565, -0.00000289, 0.00000256, 0.00000834, 0.00001144, 0.00001000, 0.00000459, -0.00000185, -0.00000548, -0.00000368, 0.00000333, 0.00001244, 0.00001919, 0.00002015, 0.00001471, 0.00000534, -0.00000381, -0.00000909, -0.00000918, -0.00000550, -0.00000107, 0.00000128, 0.00000042, -0.00000274, -0.00000614, -0.00000789, },
		{ -0.00000028, -0.00000029, -0.00000073, -0.00000117, -0.00000168, -0.00000267, -0.00000430, -0.00000612, -0.00000717, -0.00000667, -0.00000462, -0.00000205, -0.00000042, -0.00000071, -0.00000270, -0.00000491, -0.00000551, -0.00000344, 0.00000080, 0.00000537, 0.00000810, 0.00000781, 0.00000502, 0.00000162, -0.00000034, 0.00000010, 0.00000220, 0.00000406, 0.00000383, 0.00000092, -0.00000363, -0.00000781, -0.00000990, -0.00000943, -0.00000734, -0.00000514, -0.00000383, -0.00000316, -0.00000198, 0.00000079, 0.00000508, 0.00000941, 0.00001161, 0.00001018, 0.00000535, -0.00000090, -0.00000592, -0.00000792, -0.00000703, -0.00000511, -0.00000439, -0.00000595, -0.00000887, -0.00001078, -0.00000950, -0.00000468, 0.00000165, 0.00000610, 0.00000596, 0.00000101, -0.00000599, -0.00001072, -0.00000969, -0.00000232, 0.00000846, 0.00001770, 0.00002099, 0.00001677, 0.00000706, -0.00000372, -0.00001116, -0.00001293, -0.00000966, -0.00000406, 0.00000095, 0.00000377, 0.00000466, 0.00000497, 0.00000577, 0.00000701, 0.00000749, 0.00000590, 0.00000189, -0.00000343, -0.00000802, -0.00001002, -0.00000884, -0.00000536, -0.00000132, 0.00000173, 0.00000318, 0.00000338, 0.00000296, 0.00000212, 0.00000048, -0.00000243, -0.00000635, -0.00000982, -0.00001068, -0.00000733, 0.00000007, 0.00000915, 0.00001628, 0.00001845, 0.00001500, 0.00000815, 0.00000195, 0.00000000, 0.00000341, 0.00001005, 0.00001566, 0.00001627, 0.00001043, 0.00000012, -0.00001032, -0.00001643, -0.00001603, -0.00001016, -0.00000232, 0.00000352, 0.00000500, 0.00000232, -0.00000229, -0.00000608, -0.00000733, -0.00000605, -0.00000353, -0.00000133, },
		{ -0.00000144, -0.00000206, -0.00000305, -0.00000445, -0.00000568, -0.00000591, -0.00000469, -0.00000241, -0.00000019, 0.00000081, 0.00000019, -0.00000133, -0.00000236, -0.00000175, 0.00000058, 0.00000347, 0.00000518, 0.00000453, 0.00000169, -0.00000179, -0.00000392, -0.00000352, -0.00000092, 0.00000215, 0.00000363, 0.00000239, -0.00000115, -0.00000527, -0.00000808, -0.00000859, -0.00000716, -0.00000501, -0.00000329, -0.00000220, -0.00000101, 0.00000124, 0.00000471, 0.00000835, 0.00001024, 0.00000883, 0.00000397, -0.00000272, -0.00000866, -0.00001169, -0.00001125, -0.00000855, -0.00000570, -0.00000428, -0.00000436, -0.00000462, -0.00000348, -0.00000046, 0.00000323, 0.00000536, 0.00000421, 0.00000003, -0.00000474, -0.00000667, -0.00000346, 0.00000433, 0.00001317, 0.00001823, 0.00001604, 0.00000661, -0.00000631, -0.00001697, -0.00002057, -0.00001570, -0.00000494, 0.00000661, 0.00001408, 0.00001511, 0.00001074, 0.00000423, -0.00000104, -0.00000338, -0.00000336, -0.00000282, -0.00000335, -0.00000508, -0.00000672, -0.00000661, -0.00000397, 0.00000040, 0.00000462, 0.00000680, 0.00000617, 0.00000337, -0.00000021, -0.00000337, -0.00000580, -0.00000782, -0.00000962, -0.00001054, -0.00000902, -0.00000368, 0.00000543, 0.00001611, 0.00002454, 0.00002714, 0.00002250, 0.00001254, 0.00000171, -0.00000515, -0.00000541, -0.00000007, 0.00000680, 0.00001039, 0.00000783, -0.00000017, -0.00000973, -0.00001601, -0.00001586, -0.00000955, -0.00000047, 0.00000681, 0.00000897, 0.00000556, -0.00000110, -0.00000746, -0.00001072, -0.00001009, -0.00000681, -0.00000304, -0.00000050, 0.00000029, -0.00000011, -0.00000084, },
		{ -0.00000456, -0.00000535, -0.00000507, -0.00000361, -0.00000159, -0.00000009, 0.00000011, -0.00000090, -0.00000218, -0.00000259, -0.00000158, 0.00000034, 0.00000188, 0.00000185, 0.00000002, -0.00000259, -0.00000431, -0.00000383, -0.00000115, 0.00000238, 0.00000474, 0.00000449, 0.00000157, -0.00000265, -0.00000623, -0.00000778, -0.00000716, -0.00000530, -0.00000346, -0.00000225, -0.00000138, -0.00000006, 0.00000216, 0.00000471, 0.00000618, 0.00000515, 0.00000124, -0.00000439, -0.00000950, -0.00001193, -0.00001083, -0.00000708, -0.00000270, 0.00000039, 0.00000147, 0.00000124, 0.00000107, 0.00000185, 0.00000322, 0.00000387, 0.00000263, -0.00000040, -0.00000348, -0.00000423, -0.00000121, 0.00000469, 0.00001030, 0.00001173, 0.00000676, -0.00000347, -0.00001452, -0.00002076, -0.00001838, -0.00000762, 0.00000703, 0.00001894, 0.00002272, 0.00001691, 0.00000457, -0.00000843, -0.00001652, -0.00001711, -0.00001146, -0.00000346, 0.00000285, 0.00000535, 0.00000457, 0.00000266, 0.00000163, 0.00000206, 0.00000291, 0.00000258, 0.00000023, -0.00000354, -0.00000707, -0.00000881, -0.00000831, -0.00000631, -0.00000399, -0.00000183, 0.00000083, 0.00000527, 0.00001203, 0.00001984, 0.00002566, 0.00002608, 0.00001943, 0.00000724, -0.00000598, -0.00001472, -0.00001536, -0.00000815, 0.00000279, 0.00001154, 0.00001353, 0.00000793, -0.00000203, -0.00001096, -0.00001434, -0.00001092, -0.00000322, 0.00000410, 0.00000692, 0.00000387, -0.00000316, -0.00001029, -0.00001396, -0.00001273, -0.00000775, -0.00000173, 0.00000266, 0.00000421, 0.00000333, 0.00000141, -0.00000033, -0.00000146, -0.00000233, -0.00000336, },
		{ -0.00000386, -0.00000227, -0.00000141, -0.00000166, -0.00000259, -0.00000323, -0.00000279, -0.00000126, 0.00000051, 0.00000136, 0.00000072, -0.00000098, -0.00000243, -0.00000230, -0.00000019, 0.00000302, 0.00000558, 0.00000589, 0.00000347, -0.00000077, -0.00000504, -0.00000773, -0.00000824, -0.00000715, -0.00000563, -0.00000457, -0.00000400, -0.00000328, -0.00000177, 0.00000044, 0.00000239, 0.00000280, 0.00000101, -0.00000243, -0.00000587, -0.00000742, -0.00000613, -0.00000252, 0.00000168, 0.00000457, 0.00000520, 0.00000398, 0.00000221, 0.00000111, 0.00000090, 0.00000085, -0.00000006, -0.00000205, -0.00000413, -0.00000465, -0.00000255, 0.00000156, 0.00000527, 0.00000568, 0.00000134, -0.00000642, -0.00001365, -0.00001578, -0.00001026, 0.00000148, 0.00001427, 0.00002162, 0.00001912, 0.00000705, -0.00000947, -0.00002293, -0.00002715, -0.00002037, -0.00000599, 0.00000932, 0.00001913, 0.00002024, 0.00001379, 0.00000402, -0.00000443, -0.00000884, -0.00000932, -0.00000782, -0.00000640, -0.00000576, -0.00000515, -0.00000332, 0.00000012, 0.00000424, 0.00000738, 0.00000844, 0.00000782, 0.00000724, 0.00000846, 0.00001173, 0.00001516, 0.00001565, 0.00001087, 0.00000122, -0.00000974, -0.00001679, -0.00001590, -0.00000667, 0.00000697, 0.00001855, 0.00002230, 0.00001624, 0.00000318, -0.00001075, -0.00001936, -0.00001958, -0.00001276, -0.00000361, 0.00000264, 0.00000312, -0.00000145, -0.00000749, -0.00001097, -0.00000966, -0.00000423, 0.00000244, 0.00000714, 0.00000806, 0.00000553, 0.00000143, -0.00000217, -0.00000415, -0.00000471, -0.00000476, -0.00000507, -0.00000564, -0.00000591, -0.00000530, },
		{ -0.00000243, -0.00000262, -0.00000218, -0.00000076, 0.00000118, 0.00000267, 0.00000286, 0.00000173, 0.00000019, -0.00000047, 0.00000051, 0.00000278, 0.00000499, 0.00000552, 0.00000348, -0.00000071, -0.00000557, -0.00000937, -0.00001108, -0.00001080, -0.00000945, -0.00000803, -0.00000689, -0.00000565, -0.00000368, -0.00000082, 0.00000231, 0.00000450, 0.00000479, 0.00000317, 0.00000068, -0.00000109, -0.00000102, 0.00000084, 0.00000331, 0.00000485, 0.00000448, 0.00000241, -0.00000019, -0.00000209, -0.00000276, -0.00000269, -0.00000285, -0.00000378, -0.00000501, -0.00000534, -0.00000377, -0.00000053, 0.00000271, 0.00000367, 0.00000105, -0.00000426, -0.00000923, -0.00001031, -0.00000567, 0.00000321, 0.00001179, 0.00001474, 0.00000905, -0.00000383, -0.00001814, -0.00002652, -0.00002387, -0.00001025, 0.00000879, 0.00002485, 0.00003079, 0.00002426, 0.00000862, -0.00000899, -0.00002132, -0.00002446, -0.00001905, -0.00000917, 0.00000043, 0.00000677, 0.00000958, 0.00001037, 0.00001066, 0.00001074, 0.00000964, 0.00000647, 0.00000163, -0.00000284, -0.00000438, -0.00000173, 0.00000389, 0.00000907, 0.00001023, 0.00000593, -0.00000181, -0.00000836, -0.00000909, -0.00000232, 0.00000927, 0.00001967, 0.00002271, 0.00001550, 0.00000025, -0.00001664, -0.00002786, -0.00002890, -0.00002021, -0.00000672, 0.00000499, 0.00001019, 0.00000814, 0.00000194, -0.00000372, -0.00000535, -0.00000248, 0.00000253, 0.00000622, 0.00000623, 0.00000254, -0.00000282, -0.00000721, -0.00000897, -0.00000816, -0.00000620, -0.00000470, -0.00000445, -0.00000509, -0.00000565, -0.00000537, -0.00000428, -0.00000305, -0.00000239, },
		{ 0.00000329, 0.00000468, 0.00000495, 0.00000377, 0.00000168, -0.00000015, -0.00000067, 0.00000031, 0.00000189, 0.00000255, 0.00000109, -0.00000258, -0.00000739, -0.00001162, -0.00001389, -0.00001381, -0.00001198, -0.00000941, -0.00000682, -0.00000427, -0.00000136, 0.00000214, 0.00000584, 0.00000872, 0.00000973, 0.00000848, 0.00000562, 0.00000252, 0.00000049, 0.00000003, 0.00000058, 0.00000101, 0.00000037, -0.00000141, -0.00000352, -0.00000493, -0.00000511, -0.00000441, -0.00000374, -0.00000380, -0.00000445, -0.00000474, -0.00000366, -0.00000109, 0.00000181, 0.00000314, 0.00000154, -0.00000263, -0.00000716, -0.00000907, -0.00000658, -0.00000054, 0.00000562, 0.00000769, 0.00000334, -0.00000602, -0.00001545, -0.00001893, -0.00001283, 0.00000156, 0.00001811, 0.00002867, 0.00002732, 0.00001364, -0.00000676, -0.00002490, -0.00003286, -0.00002758, -0.00001209, 0.00000645, 0.00002053, 0.00002586, 0.00002269, 0.00001471, 0.00000627, -0.00000007, -0.00000418, -0.00000733, -0.00001025, -0.00001203, -0.00001061, -0.00000460, 0.00000503, 0.00001468, 0.00001981, 0.00001756, 0.00000885, -0.00000179, -0.00000856, -0.00000771, 0.00000002, 0.00000962, 0.00001451, 0.00001042, -0.00000200, -0.00001722, -0.00002770, -0.00002798, -0.00001760, -0.00000131, 0.00001345, 0.00002052, 0.00001798, 0.00000866, -0.00000183, -0.00000842, -0.00000910, -0.00000551, -0.00000134, 0.00000011, -0.00000212, -0.00000646, -0.00001009, -0.00001086, -0.00000858, -0.00000485, -0.00000194, -0.00000128, -0.00000277, -0.00000502, -0.00000645, -0.00000625, -0.00000472, -0.00000282, -0.00000136, -0.00000045, 0.00000038, 0.00000163, },
		{ 0.00000012, -0.00000267, -0.00000433, -0.00000426, -0.00000291, -0.00000159, -0.00000174, -0.00000398, -0.00000773, -0.00001147, -0.00001362, -0.00001333, -0.00001085, -0.00000716, -0.00000332, 0.00000014, 0.00000328, 0.00000633, 0.00000919, 0.00001121, 0.00001153, 0.00000966, 0.00000599, 0.00000167, -0.00000187, -0.00000378, -0.00000412, -0.00000372, -0.00000353, -0.00000399, -0.00000478, -0.00000520, -0.00000480, -0.00000380, -0.00000295, -0.00000290, -0.00000365, -0.00000440, -0.00000411, -0.00000239, 0.00000001, 0.00000147, 0.00000058, -0.00000276, -0.00000689, -0.00000924, -0.00000788, -0.00000299, 0.00000287, 0.00000613, 0.00000444, -0.00000159, -0.00000830, -0.00001092, -0.00000651, 0.00000380, 0.00001490, 0.00002019, 0.00001539, 0.00000138, -0.00001568, -0.00002725, -0.00002679, -0.00001345, 0.00000732, 0.00002638, 0.00003551, 0.00003136, 0.00001677, -0.00000121, -0.00001525, -0.00002129, -0.00001980, -0.00001434, -0.00000864, -0.00000425, -0.00000010, 0.00000570, 0.00001355, 0.00002109, 0.00002419, 0.00001960, 0.00000755, -0.00000751, -0.00001885, -0.00002107, -0.00001326, 0.00000016, 0.00001171, 0.00001498, 0.00000827, -0.00000445, -0.00001575, -0.00001892, -0.00001177, 0.00000216, 0.00001561, 0.00002166, 0.00001737, 0.00000513, -0.00000893, -0.00001848, -0.00002025, -0.00001531, -0.00000789, -0.00000253, -0.00000157, -0.00000418, -0.00000745, -0.00000854, -0.00000644, -0.00000248, 0.00000084, 0.00000151, -0.00000072, -0.00000434, -0.00000713, -0.00000753, -0.00000545, -0.00000207, 0.00000110, 0.00000312, 0.00000404, 0.00000445, 0.00000483, 0.00000507, 0.00000457, 0.00000285, },
		{ -0.00000607, -0.00000354, -0.00000189, -0.00000221, -0.00000448, -0.00000751, -0.00000966, -0.00000967, -0.00000736, -0.00000357, 0.00000047, 0.00000382, 0.00000621, 0.00000786, 0.00000901, 0.00000951, 0.00000887, 0.00000664, 0.00000297, -0.00000130, -0.00000493, -0.00000694, -0.00000713, -0.00000609, -0.00000481, -0.00000399, -0.00000376, -0.00000371, -0.00000345, -0.00000299, -0.00000277, -0.00000328, -0.00000449, -0.00000569, -0.00000588, -0.00000453, -0.00000210, -0.00000001, 0.00000026, -0.00000178, -0.00000503, -0.00000728, -0.00000655, -0.00000253, 0.00000290, 0.00000662, 0.00000611, 0.00000127, -0.00000524, -0.00000930, -0.00000790, -0.00000122, 0.00000709, 0.00001184, 0.00000949, 0.00000065, -0.00000978, -0.00001507, -0.00001061, 0.00000301, 0.00001972, 0.00003089, 0.00002991, 0.00001591, -0.00000550, -0.00002502, -0.00003442, -0.00003064, -0.00001691, -0.00000051, 0.00001149, 0.00001584, 0.00001412, 0.00001074, 0.00000943, 0.00001056, 0.00001114, 0.00000726, -0.00000249, -0.00001519, -0.00002463, -0.00002480, -0.00001376, 0.00000460, 0.00002212, 0.00003056, 0.00002609, 0.00001139, -0.00000578, -0.00001685, -0.00001710, -0.00000802, 0.00000383, 0.00001078, 0.00000835, -0.00000238, -0.00001574, -0.00002484, -0.00002546, -0.00001812, -0.00000730, 0.00000140, 0.00000443, 0.00000197, -0.00000281, -0.00000604, -0.00000559, -0.00000207, 0.00000186, 0.00000349, 0.00000179, -0.00000208, -0.00000563, -0.00000668, -0.00000463, -0.00000064, 0.00000329, 0.00000554, 0.00000571, 0.00000451, 0.00000296, 0.00000164, 0.00000037, -0.00000138, -0.00000381, -0.00000638, -0.00000803, -0.00000792, },
		{ 0.00000097, -0.00000188, -0.00000472, -0.00000596, -0.00000493, -0.00000211, 0.00000127, 0.00000398, 0.00000543, 0.00000577, 0.00000552, 0.00000499, 0.00000409, 0.00000247, -0.00000003, -0.00000303, -0.00000571, -0.00000721, -0.00000716, -0.00000592, -0.00000433, -0.00000317, -0.00000281, -0.00000303, -0.00000344, -0.00000380, -0.00000426, -0.00000505, -0.00000616, -0.00000704, -0.00000686, -0.00000510, -0.00000212, 0.00000078, 0.00000206, 0.00000095, -0.00000193, -0.00000469, -0.00000532, -0.00000300, 0.00000111, 0.00000439, 0.00000436, 0.00000041, -0.00000552, -0.00000982, -0.00000936, -0.00000353, 0.00000504, 0.00001186, 0.00001326, 0.00000873, 0.00000151, -0.00000327, -0.00000188, 0.00000540, 0.00001390, 0.00001709, 0.00001056, -0.00000483, -0.00002267, -0.00003412, -0.00003262, -0.00001780, 0.00000401, 0.00002293, 0.00003067, 0.00002483, 0.00000989, -0.00000570, -0.00001439, -0.00001365, -0.00000684, -0.00000044, 0.00000044, -0.00000444, -0.00001031, -0.00001059, -0.00000145, 0.00001507, 0.00003157, 0.00003916, 0.00003255, 0.00001336, -0.00001046, -0.00002873, -0.00003434, -0.00002686, -0.00001239, 0.00000018, 0.00000414, -0.00000158, -0.00001236, -0.00002101, -0.00002209, -0.00001491, -0.00000346, 0.00000624, 0.00000973, 0.00000654, -0.00000006, -0.00000534, -0.00000605, -0.00000220, 0.00000338, 0.00000704, 0.00000662, 0.00000263, -0.00000237, -0.00000546, -0.00000520, -0.00000226, 0.00000126, 0.00000323, 0.00000278, 0.00000052, -0.00000212, -0.00000398, -0.00000486, -0.00000536, -0.00000615, -0.00000729, -0.00000806, -0.00000749, -0.00000520, -0.00000185, 0.00000110, 0.00000222, },
		{ -0.00000542, -0.00000397, -0.00000138, 0.00000096, 0.00000213, 0.00000207, 0.00000136, 0.00000068, 0.00000028, -0.00000012, -0.00000095, -0.00000238, -0.00000407, -0.00000537, -0.00000580, -0.00000532, -0.00000442, -0.00000374, -0.00000365, -0.00000412, -0.00000478, -0.00000530, -0.00000563, -0.00000592, -0.00000622, -0.00000623, -0.00000540, -0.00000335, -0.00000037, 0.00000244, 0.00000365, 0.00000239, -0.00000095, -0.00000470, -0.00000675, -0.00000585, -0.00000251, 0.00000115, 0.00000269, 0.00000095, -0.00000295, -0.00000612, -0.00000558, -0.00000031, 0.00000782, 0.00001485, 0.00001698, 0.00001292, 0.00000482, -0.00000295, -0.00000643, -0.00000470, -0.00000054, 0.00000131, -0.00000254, -0.00001148, -0.00002059, -0.00002322, -0.00001525, 0.00000181, 0.00002063, 0.00003161, 0.00002809, 0.00001042, -0.00001361, -0.00003257, -0.00003731, -0.00002584, -0.00000432, 0.00001646, 0.00002708, 0.00002455, 0.00001352, 0.00000304, 0.00000078, 0.00000830, 0.00002007, 0.00002684, 0.00002136, 0.00000304, -0.00002129, -0.00004095, -0.00004720, -0.00003796, -0.00001881, 0.00000027, 0.00001044, 0.00000850, -0.00000191, -0.00001303, -0.00001771, -0.00001335, -0.00000291, 0.00000746, 0.00001226, 0.00000987, 0.00000308, -0.00000303, -0.00000438, -0.00000026, 0.00000645, 0.00001133, 0.00001120, 0.00000586, -0.00000194, -0.00000844, -0.00001100, -0.00000949, -0.00000593, -0.00000297, -0.00000225, -0.00000361, -0.00000562, -0.00000673, -0.00000636, -0.00000510, -0.00000402, -0.00000376, -0.00000397, -0.00000362, -0.00000186, 0.00000123, 0.00000444, 0.00000618, 0.00000538, 0.00000229, -0.00000167, -0.00000464, },
		{ -0.00000135, -0.00000082, -0.00000113, -0.00000146, -0.00000126, -0.00000066, -0.00000024, -0.00000060, -0.00000185, -0.00000356, -0.00000507, -0.00000588, -0.00000599, -0.00000573, -0.00000548, -0.00000541, -0.00000540, -0.00000523, -0.00000488, -0.00000448, -0.00000417, -0.00000386, -0.00000319, -0.00000184, 0.00000010, 0.00000188, 0.00000238, 0.00000084, -0.00000251, -0.00000622, -0.00000828, -0.00000728, -0.00000338, 0.00000160, 0.00000515, 0.00000556, 0.00000309, -0.00000006, -0.00000108, 0.00000153, 0.00000683, 0.00001171, 0.00001271, 0.00000822, -0.00000031, -0.00000896, -0.00001369, -0.00001272, -0.00000770, -0.00000263, -0.00000114, -0.00000396, -0.00000816, -0.00000907, -0.00000357, 0.00000706, 0.00001705, 0.00001912, 0.00000908, -0.00001074, -0.00003152, -0.00004199, -0.00003452, -0.00000983, 0.00002252, 0.00004849, 0.00005659, 0.00004389, 0.00001748, -0.00000933, -0.00002420, -0.00002222, -0.00000817, 0.00000673, 0.00001148, 0.00000166, -0.00001817, -0.00003730, -0.00004506, -0.00003696, -0.00001698, 0.00000481, 0.00001804, 0.00001765, 0.00000621, -0.00000810, -0.00001649, -0.00001450, -0.00000398, 0.00000859, 0.00001636, 0.00001605, 0.00000947, 0.00000182, -0.00000181, 0.00000039, 0.00000612, 0.00001067, 0.00001002, 0.00000330, -0.00000682, -0.00001581, -0.00001999, -0.00001841, -0.00001302, -0.00000717, -0.00000351, -0.00000271, -0.00000353, -0.00000407, -0.00000322, -0.00000124, 0.00000064, 0.00000137, 0.00000090, 0.00000025, 0.00000071, 0.00000277, 0.00000563, 0.00000759, 0.00000713, 0.00000386, -0.00000108, -0.00000570, -0.00000822, -0.00000802, -0.00000584, -0.00000318, },
		{ -0.00000112, -0.00000035, 0.00000046, 0.00000049, -0.00000071, -0.00000291, -0.00000535, -0.00000714, -0.00000780, -0.00000739, -0.00000634, -0.00000513, -0.00000405, -0.00000318, -0.00000256, -0.00000226, -0.00000236, -0.00000275, -0.00000306, -0.00000286, -0.00000201, -0.00000088, -0.00000027, -0.00000092, -0.00000282, -0.00000500, -0.00000588, -0.00000419, 0.00000005, 0.00000529, 0.00000909, 0.00000951, 0.00000635, 0.00000139, -0.00000252, -0.00000326, -0.00000076, 0.00000282, 0.00000438, 0.00000196, -0.00000389, -0.00001019, -0.00001333, -0.00001135, -0.00000530, 0.00000125, 0.00000443, 0.00000264, -0.00000236, -0.00000658, -0.00000656, -0.00000204, 0.00000342, 0.00000457, -0.00000179, -0.00001377, -0.00002434, -0.00002480, -0.00001011, 0.00001702, 0.00004601, 0.00006298, 0.00005785, 0.00003020, -0.00000957, -0.00004478, -0.00006067, -0.00005178, -0.00002467, 0.00000562, 0.00002355, 0.00002107, 0.00000138, -0.00002340, -0.00003923, -0.00003782, -0.00002068, 0.00000243, 0.00001942, 0.00002270, 0.00001290, -0.00000218, -0.00001265, -0.00001229, -0.00000178, 0.00001236, 0.00002202, 0.00002234, 0.00001416, 0.00000294, -0.00000495, -0.00000610, -0.00000183, 0.00000310, 0.00000375, -0.00000186, -0.00001164, -0.00002084, -0.00002496, -0.00002226, -0.00001454, -0.00000570, 0.00000062, 0.00000291, 0.00000220, 0.00000083, 0.00000070, 0.00000211, 0.00000394, 0.00000472, 0.00000392, 0.00000231, 0.00000142, 0.00000226, 0.00000452, 0.00000657, 0.00000650, 0.00000335, -0.00000217, -0.00000800, -0.00001186, -0.00001248, -0.00001020, -0.00000655, -0.00000334, -0.00000160, -0.00000122, -0.00000135, },
		{ -0.00000081, -0.00000287, -0.00000545, -0.00000750, -0.00000820, -0.00000741, -0.00000559, -0.00000354, -0.00000193, -0.00000110, -0.00000108, -0.00000174, -0.00000285, -0.00000408, -0.00000498, -0.00000507, -0.00000409, -0.00000229, -0.00000039, 0.00000076, 0.00000076, 0.00000000, -0.00000043, 0.00000056, 0.00000319, 0.00000643, 0.00000834, 0.00000721, 0.00000271, -0.00000359, -0.00000893, -0.00001083, -0.00000862, -0.00000390, 0.00000034, 0.00000152, -0.00000089, -0.00000501, -0.00000779, -0.00000705, -0.00000304, 0.00000159, 0.00000346, 0.00000082, -0.00000509, -0.00001057, -0.00001182, -0.00000755, -0.00000011, 0.00000588, 0.00000659, 0.00000200, -0.00000347, -0.00000343, 0.00000611, 0.00002309, 0.00003912, 0.00004326, 0.00002829, -0.00000413, -0.00004262, -0.00007074, -0.00007499, -0.00005214, -0.00001166, 0.00002840, 0.00005010, 0.00004450, 0.00001600, -0.00002020, -0.00004609, -0.00004994, -0.00003176, -0.00000235, 0.00002320, 0.00003369, 0.00002717, 0.00001088, -0.00000386, -0.00000843, -0.00000147, 0.00001116, 0.00002061, 0.00002052, 0.00001067, -0.00000347, -0.00001437, -0.00001712, -0.00001206, -0.00000412, 0.00000057, -0.00000154, -0.00000939, -0.00001828, -0.00002290, -0.00002047, -0.00001214, -0.00000198, 0.00000560, 0.00000832, 0.00000690, 0.00000406, 0.00000246, 0.00000308, 0.00000492, 0.00000610, 0.00000538, 0.00000311, 0.00000089, 0.00000028, 0.00000158, 0.00000351, 0.00000395, 0.00000140, -0.00000395, -0.00001019, -0.00001469, -0.00001562, -0.00001287, -0.00000800, -0.00000328, -0.00000037, 0.00000037, -0.00000018, -0.00000079, -0.00000072, -0.00000017, 0.00000006, },
		{ -0.00000693, -0.00000634, -0.00000460, -0.00000259, -0.00000121, -0.00000093, -0.00000171, -0.00000312, -0.00000462, -0.00000570, -0.00000592, -0.00000499, -0.00000285, 0.00000011, 0.00000311, 0.00000517, 0.00000562, 0.00000452, 0.00000268, 0.00000125, 0.00000092, 0.00000137, 0.00000139, -0.00000038, -0.00000437, -0.00000951, -0.00001353, -0.00001420, -0.00001070, -0.00000438, 0.00000184, 0.00000502, 0.00000393, -0.00000028, -0.00000475, -0.00000680, -0.00000557, -0.00000256, -0.00000051, -0.00000133, -0.00000457, -0.00000744, -0.00000664, -0.00000075, 0.00000832, 0.00001612, 0.00001822, 0.00001327, 0.00000420, -0.00000328, -0.00000435, 0.00000136, 0.00000871, 0.00000968, -0.00000157, -0.00002404, -0.00004886, -0.00006295, -0.00005599, -0.00002686, 0.00001420, 0.00004965, 0.00006298, 0.00004732, 0.00000948, -0.00003297, -0.00006022, -0.00005997, -0.00003331, 0.00000632, 0.00004035, 0.00005458, 0.00004574, 0.00002198, -0.00000262, -0.00001602, -0.00001430, -0.00000260, 0.00000906, 0.00001217, 0.00000461, -0.00000882, -0.00001992, -0.00002237, -0.00001542, -0.00000401, 0.00000466, 0.00000550, -0.00000154, -0.00001177, -0.00001889, -0.00001870, -0.00001135, -0.00000080, 0.00000783, 0.00001130, 0.00000969, 0.00000581, 0.00000298, 0.00000285, 0.00000468, 0.00000623, 0.00000561, 0.00000265, -0.00000105, -0.00000333, -0.00000309, -0.00000112, 0.00000040, -0.00000062, -0.00000472, -0.00001041, -0.00001499, -0.00001613, -0.00001322, -0.00000768, -0.00000207, 0.00000139, 0.00000192, 0.00000039, -0.00000152, -0.00000244, -0.00000209, -0.00000123, -0.00000101, -0.00000205, -0.00000405, -0.00000602, },
		{ -0.00000251, -0.00000224, -0.00000288, -0.00000391, -0.00000465, -0.00000458, -0.00000348, -0.00000140, 0.00000136, 0.00000428, 0.00000660, 0.00000752, 0.00000649, 0.00000360, -0.00000032, -0.00000399, -0.00000633, -0.00000709, -0.00000698, -0.00000721, -0.00000859, -0.00001082, -0.00001246, -0.00001177, -0.00000794, -0.00000187, 0.00000400, 0.00000697, 0.00000565, 0.00000089, -0.00000457, -0.00000771, -0.00000695, -0.00000309, 0.00000136, 0.00000392, 0.00000386, 0.00000270, 0.00000302, 0.00000637, 0.00001167, 0.00001537, 0.00001353, 0.00000462, -0.00000888, -0.00002131, -0.00002694, -0.00002356, -0.00001424, -0.00000588, -0.00000502, -0.00001339, -0.00002600, -0.00003341, -0.00002727, -0.00000596, 0.00002301, 0.00004593, 0.00004989, 0.00003030, -0.00000556, -0.00004122, -0.00005896, -0.00004896, -0.00001469, 0.00002845, 0.00006075, 0.00006805, 0.00004850, 0.00001292, -0.00002135, -0.00003959, -0.00003673, -0.00001882, 0.00000152, 0.00001261, 0.00000993, -0.00000237, -0.00001499, -0.00001956, -0.00001354, -0.00000121, 0.00000945, 0.00001191, 0.00000488, -0.00000719, -0.00001712, -0.00001931, -0.00001282, -0.00000145, 0.00000889, 0.00001375, 0.00001234, 0.00000738, 0.00000283, 0.00000126, 0.00000244, 0.00000404, 0.00000358, 0.00000031, -0.00000431, -0.00000765, -0.00000786, -0.00000517, -0.00000180, -0.00000049, -0.00000269, -0.00000757, -0.00001248, -0.00001460, -0.00001264, -0.00000753, -0.00000187, 0.00000172, 0.00000203, -0.00000024, -0.00000306, -0.00000451, -0.00000386, -0.00000184, -0.00000003, 0.00000027, -0.00000117, -0.00000350, -0.00000542, -0.00000599, -0.00000518, -0.00000369, },
		{ -0.00000131, -0.00000048, 0.00000096, 0.00000272, 0.00000433, 0.00000532, 0.00000527, 0.00000385, 0.00000094, -0.00000316, -0.00000766, -0.00001138, -0.00001328, -0.00001298, -0.00001106, -0.00000877, -0.00000734, -0.00000714, -0.00000746, -0.00000689, -0.00000439, -0.00000014, 0.00000430, 0.00000678, 0.00000596, 0.00000229, -0.00000204, -0.00000428, -0.00000278, 0.00000192, 0.00000733, 0.00001057, 0.00001007, 0.00000660, 0.00000255, 0.00000016, -0.00000025, -0.00000075, -0.00000434, -0.00001259, -0.00002380, -0.00003322, -0.00003546, -0.00002792, -0.00001296, 0.00000275, 0.00001169, 0.00000985, -0.00000051, -0.00001185, -0.00001561, -0.00000759, 0.00000897, 0.00002469, 0.00002928, 0.00001797, -0.00000482, -0.00002703, -0.00003524, -0.00002243, 0.00000735, 0.00004021, 0.00005928, 0.00005363, 0.00002431, -0.00001581, -0.00004853, -0.00005960, -0.00004576, -0.00001590, 0.00001420, 0.00003038, 0.00002719, 0.00000978, -0.00000998, -0.00002073, -0.00001771, -0.00000472, 0.00000910, 0.00001507, 0.00001001, -0.00000238, -0.00001430, -0.00001874, -0.00001347, -0.00000185, 0.00000958, 0.00001521, 0.00001339, 0.00000671, -0.00000011, -0.00000346, -0.00000279, -0.00000045, 0.00000041, -0.00000181, -0.00000606, -0.00000955, -0.00000974, -0.00000625, -0.00000120, 0.00000211, 0.00000134, -0.00000338, -0.00000949, -0.00001361, -0.00001350, -0.00000942, -0.00000376, 0.00000046, 0.00000141, -0.00000061, -0.00000356, -0.00000510, -0.00000407, -0.00000112, 0.00000186, 0.00000299, 0.00000151, -0.00000181, -0.00000531, -0.00000741, -0.00000746, -0.00000595, -0.00000396, -0.00000245, -0.00000177, -0.00000159, },
		{ 0.00000157, 0.00000004, -0.00000220, -0.00000498, -0.00000805, -0.00001110, -0.00001361, -0.00001497, -0.00001459, -0.00001231, -0.00000871, -0.00000495, -0.00000231, -0.00000144, -0.00000192, -0.00000249, -0.00000178, 0.00000070, 0.00000419, 0.00000700, 0.00000766, 0.00000590, 0.00000301, 0.00000108, 0.00000159, 0.00000436, 0.00000740, 0.00000808, 0.00000480, -0.00000186, -0.00000937, -0.00001483, -0.00001682, -0.00001633, -0.00001601, -0.00001810, -0.00002244, -0.00002601, -0.00002453, -0.00001543, -0.00000025, 0.00001534, 0.00002410, 0.00002139, 0.00000822, -0.00000862, -0.00002001, -0.00001948, -0.00000713, 0.00001023, 0.00002288, 0.00002386, 0.00001326, -0.00000167, -0.00001085, -0.00000765, 0.00000661, 0.00002285, 0.00002925, 0.00001847, -0.00000735, -0.00003677, -0.00005480, -0.00005101, -0.00002562, 0.00001015, 0.00003966, 0.00004947, 0.00003623, 0.00000810, -0.00001994, -0.00003420, -0.00002928, -0.00001015, 0.00001140, 0.00002367, 0.00002135, 0.00000762, -0.00000850, -0.00001784, -0.00001622, -0.00000612, 0.00000552, 0.00001192, 0.00001026, 0.00000281, -0.00000519, -0.00000911, -0.00000767, -0.00000327, 0.00000015, -0.00000009, -0.00000361, -0.00000754, -0.00000860, -0.00000556, -0.00000011, 0.00000417, 0.00000415, -0.00000077, -0.00000824, -0.00001433, -0.00001592, -0.00001243, -0.00000607, -0.00000030, 0.00000221, 0.00000108, -0.00000186, -0.00000394, -0.00000344, -0.00000060, 0.00000265, 0.00000402, 0.00000234, -0.00000178, -0.00000634, -0.00000919, -0.00000918, -0.00000661, -0.00000288, 0.00000047, 0.00000251, 0.00000322, 0.00000316, 0.00000292, 0.00000272, 0.00000239, },
		{ -0.00001384, -0.00001368, -0.00001269, -0.00001095, -0.00000848, -0.00000535, -0.00000193, 0.00000109, 0.00000287, 0.00000293, 0.00000159, -0.00000002, -0.00000052, 0.00000092, 0.00000393, 0.00000709, 0.00000872, 0.00000787, 0.00000500, 0.00000166, -0.00000058, -0.00000123, -0.00000131, -0.00000271, -0.00000677, -0.00001309, -0.00001949, -0.00002315, -0.00002238, -0.00001788, -0.00001237, -0.00000892, -0.00000881, -0.00001043, -0.00001023, -0.00000513, 0.00000493, 0.00001616, 0.00002259, 0.00001960, 0.00000712, -0.00000944, -0.00002162, -0.00002239, -0.00001032, 0.00000918, 0.00002649, 0.00003295, 0.00002562, 0.00000880, -0.00000853, -0.00001813, -0.00001719, -0.00000974, -0.00000369, -0.00000535, -0.00001490, -0.00002572, -0.00002833, -0.00001657, 0.00000766, 0.00003426, 0.00004986, 0.00004518, 0.00002070, -0.00001281, -0.00003955, -0.00004674, -0.00003135, -0.00000156, 0.00002765, 0.00004243, 0.00003693, 0.00001575, -0.00000940, -0.00002634, -0.00002875, -0.00001862, -0.00000398, 0.00000637, 0.00000780, 0.00000177, -0.00000608, -0.00001004, -0.00000790, -0.00000185, 0.00000350, 0.00000443, 0.00000059, -0.00000502, -0.00000828, -0.00000682, -0.00000167, 0.00000347, 0.00000470, 0.00000045, -0.00000738, -0.00001461, -0.00001728, -0.00001402, -0.00000672, 0.00000070, 0.00000469, 0.00000408, 0.00000043, -0.00000318, -0.00000422, -0.00000225, 0.00000104, 0.00000309, 0.00000220, -0.00000140, -0.00000575, -0.00000836, -0.00000774, -0.00000411, 0.00000082, 0.00000491, 0.00000672, 0.00000603, 0.00000363, 0.00000061, -0.00000227, -0.00000482, -0.00000718, -0.00000947, -0.00001155, -0.00001310, },
		{ 0.00000147, 0.00000303, 0.00000394, 0.00000428, 0.00000393, 0.00000275, 0.00000094, -0.00000081, -0.00000154, -0.00000056, 0.00000195, 0.00000479, 0.00000632, 0.00000533, 0.00000181, -0.00000296, -0.00000722, -0.00000978, -0.00001077, -0.00001140, -0.00001300, -0.00001580, -0.00001851, -0.00001903, -0.00001590, -0.00000959, -0.00000260, 0.00000190, 0.00000208, -0.00000116, -0.00000449, -0.00000413, 0.00000151, 0.00001039, 0.00001757, 0.00001807, 0.00001015, -0.00000301, -0.00001459, -0.00001779, -0.00000995, 0.00000541, 0.00002010, 0.00002554, 0.00001762, -0.00000089, -0.00002162, -0.00003521, -0.00003627, -0.00002601, -0.00001094, 0.00000137, 0.00000693, 0.00000725, 0.00000757, 0.00001250, 0.00002200, 0.00003062, 0.00003063, 0.00001739, -0.00000648, -0.00003115, -0.00004406, -0.00003686, -0.00001064, 0.00002364, 0.00005020, 0.00005630, 0.00003883, 0.00000578, -0.00002785, -0.00004782, -0.00004746, -0.00003029, -0.00000710, 0.00001040, 0.00001563, 0.00000971, -0.00000055, -0.00000750, -0.00000726, -0.00000146, 0.00000473, 0.00000640, 0.00000221, -0.00000494, -0.00001016, -0.00000994, -0.00000451, 0.00000235, 0.00000584, 0.00000326, -0.00000426, -0.00001243, -0.00001652, -0.00001418, -0.00000680, 0.00000148, 0.00000634, 0.00000578, 0.00000108, -0.00000424, -0.00000679, -0.00000526, -0.00000103, 0.00000296, 0.00000422, 0.00000221, -0.00000146, -0.00000421, -0.00000416, -0.00000126, 0.00000276, 0.00000549, 0.00000527, 0.00000193, -0.00000328, -0.00000861, -0.00001268, -0.00001502, -0.00001589, -0.00001577, -0.00001488, -0.00001319, -0.00001063, -0.00000738, -0.00000393, -0.00000086, },
		{ -0.00000132, -0.00000260, -0.00000343, -0.00000386, -0.00000367, -0.00000257, -0.00000064, 0.00000139, 0.00000231, 0.00000109, -0.00000243, -0.00000722, -0.00001154, -0.00001390, -0.00001390, -0.00001242, -0.00001095, -0.00001045, -0.00001060, -0.00001005, -0.00000741, -0.00000260, 0.00000274, 0.00000603, 0.00000545, 0.00000141, -0.00000321, -0.00000464, -0.00000054, 0.00000803, 0.00001677, 0.00002029, 0.00001531, 0.00000301, -0.00001118, -0.00002040, -0.00002027, -0.00001163, -0.00000033, 0.00000606, 0.00000263, -0.00000974, -0.00002461, -0.00003347, -0.00003049, -0.00001580, 0.00000462, 0.00002235, 0.00003111, 0.00002976, 0.00002214, 0.00001385, 0.00000847, 0.00000552, 0.00000167, -0.00000585, -0.00001600, -0.00002342, -0.00002130, -0.00000616, 0.00001856, 0.00004268, 0.00005384, 0.00004403, 0.00001460, -0.00002344, -0.00005429, -0.00006508, -0.00005216, -0.00002280, 0.00000888, 0.00002917, 0.00003143, 0.00001857, 0.00000030, -0.00001282, -0.00001515, -0.00000824, 0.00000128, 0.00000641, 0.00000415, -0.00000308, -0.00000964, -0.00001063, -0.00000517, 0.00000314, 0.00000869, 0.00000756, -0.00000002, -0.00000979, -0.00001618, -0.00001569, -0.00000893, -0.00000006, 0.00000588, 0.00000601, 0.00000114, -0.00000498, -0.00000815, -0.00000626, -0.00000044, 0.00000591, 0.00000932, 0.00000826, 0.00000387, -0.00000107, -0.00000400, -0.00000412, -0.00000270, -0.00000202, -0.00000384, -0.00000829, -0.00001386, -0.00001838, -0.00002023, -0.00001906, -0.00001569, -0.00001138, -0.00000709, -0.00000316, 0.00000045, 0.00000370, 0.00000624, 0.00000751, 0.00000719, 0.00000547, 0.00000299, 0.00000057, },
		{ -0.00000280, -0.00000171, -0.00000089, -0.00000037, -0.00000056, -0.00000201, -0.00000490, -0.00000860, -0.00001179, -0.00001311, -0.00001199, -0.00000909, -0.00000589, -0.00000379, -0.00000312, -0.00000303, -0.00000208, 0.00000052, 0.00000410, 0.00000681, 0.00000683, 0.00000384, -0.00000029, -0.00000243, -0.00000014, 0.00000635, 0.00001364, 0.00001668, 0.00001165, -0.00000147, -0.00001821, -0.00003168, -0.00003608, -0.00003001, -0.00001732, -0.00000499, 0.00000084, -0.00000159, -0.00000873, -0.00001386, -0.00001129, 0.00000016, 0.00001620, 0.00002953, 0.00003389, 0.00002754, 0.00001402, -0.00000014, -0.00000925, -0.00001125, -0.00000808, -0.00000345, 0.00000024, 0.00000343, 0.00000851, 0.00001698, 0.00002684, 0.00003243, 0.00002732, 0.00000869, -0.00001947, -0.00004681, -0.00006112, -0.00005460, -0.00002854, 0.00000647, 0.00003539, 0.00004607, 0.00003526, 0.00000998, -0.00001647, -0.00003157, -0.00002982, -0.00001490, 0.00000332, 0.00001479, 0.00001477, 0.00000576, -0.00000487, -0.00000990, -0.00000665, 0.00000180, 0.00000898, 0.00000944, 0.00000216, -0.00000890, -0.00001741, -0.00001863, -0.00001217, -0.00000205, 0.00000597, 0.00000793, 0.00000388, -0.00000243, -0.00000617, -0.00000447, 0.00000192, 0.00000922, 0.00001305, 0.00001101, 0.00000394, -0.00000492, -0.00001192, -0.00001514, -0.00001512, -0.00001402, -0.00001396, -0.00001557, -0.00001767, -0.00001826, -0.00001588, -0.00001054, -0.00000371, 0.00000261, 0.00000697, 0.00000902, 0.00000935, 0.00000871, 0.00000749, 0.00000557, 0.00000270, -0.00000097, -0.00000469, -0.00000745, -0.00000852, -0.00000790, -0.00000622, -0.00000433, },
		{ -0.00000407, -0.00000562, -0.00000725, -0.00000893, -0.00001012, -0.00000997, -0.00000797, -0.00000445, -0.00000062, 0.00000211, 0.00000302, 0.00000258, 0.00000213, 0.00000289, 0.00000497, 0.00000709, 0.00000744, 0.00000500, 0.00000054, -0.00000354, -0.00000470, -0.00000215, 0.00000217, 0.00000428, 0.00000045, -0.00001026, -0.00002471, -0.00003686, -0.00004074, -0.00003378, -0.00001852, -0.00000141, 0.00001051, 0.00001342, 0.00000859, 0.00000142, -0.00000199, 0.00000153, 0.00001036, 0.00001923, 0.00002257, 0.00001795, 0.00000760, -0.00000294, -0.00000814, -0.00000562, 0.00000275, 0.00001233, 0.00001859, 0.00001964, 0.00001656, 0.00001183, 0.00000695, 0.00000126, -0.00000714, -0.00001888, -0.00003134, -0.00003874, -0.00003506, -0.00001810, 0.00000778, 0.00003242, 0.00004436, 0.00003684, 0.00001199, -0.00001945, -0.00004302, -0.00004782, -0.00003203, -0.00000380, 0.00002308, 0.00003645, 0.00003179, 0.00001397, -0.00000611, -0.00001796, -0.00001703, -0.00000657, 0.00000510, 0.00001003, 0.00000524, -0.00000587, -0.00001610, -0.00001909, -0.00001309, -0.00000173, 0.00000853, 0.00001246, 0.00000904, 0.00000175, -0.00000400, -0.00000442, 0.00000041, 0.00000672, 0.00000954, 0.00000584, -0.00000374, -0.00001533, -0.00002422, -0.00002740, -0.00002490, -0.00001925, -0.00001357, -0.00000966, -0.00000724, -0.00000474, -0.00000073, 0.00000480, 0.00001031, 0.00001372, 0.00001366, 0.00001034, 0.00000526, 0.00000029, -0.00000338, -0.00000567, -0.00000722, -0.00000861, -0.00000980, -0.00001013, -0.00000895, -0.00000623, -0.00000278, 0.00000012, 0.00000150, 0.00000112, -0.00000043, -0.00000234, },
		{ -0.00000528, -0.00000337, -0.00000088, 0.00000227, 0.00000548, 0.00000767, 0.00000802, 0.00000659, 0.00000442, 0.00000284, 0.00000254, 0.00000291, 0.00000240, -0.00000033, -0.00000529, -0.00001073, -0.00001411, -0.00001370, -0.00001005, -0.00000605, -0.00000536, -0.00001008, -0.00001903, -0.00002789, -0.00003136, -0.00002623, -0.00001347, 0.00000187, 0.00001329, 0.00001627, 0.00001081, 0.00000136, -0.00000577, -0.00000602, 0.00000092, 0.00001116, 0.00001912, 0.00002094, 0.00001663, 0.00000981, 0.00000518, 0.00000554, 0.00001007, 0.00001501, 0.00001616, 0.00001150, 0.00000227, -0.00000790, -0.00001541, -0.00001856, -0.00001799, -0.00001540, -0.00001173, -0.00000634, 0.00000201, 0.00001308, 0.00002362, 0.00002798, 0.00002104, 0.00000197, -0.00002365, -0.00004515, -0.00005156, -0.00003749, -0.00000681, 0.00002820, 0.00005238, 0.00005510, 0.00003568, 0.00000376, -0.00002560, -0.00003958, -0.00003388, -0.00001422, 0.00000741, 0.00001958, 0.00001742, 0.00000446, -0.00001017, -0.00001754, -0.00001392, -0.00000237, 0.00000968, 0.00001526, 0.00001189, 0.00000253, -0.00000674, -0.00001080, -0.00000856, -0.00000334, -0.00000038, -0.00000339, -0.00001212, -0.00002256, -0.00002934, -0.00002887, -0.00002120, -0.00000969, 0.00000118, 0.00000824, 0.00001099, 0.00001109, 0.00001071, 0.00001086, 0.00001081, 0.00000899, 0.00000438, -0.00000245, -0.00000941, -0.00001410, -0.00001513, -0.00001285, -0.00000895, -0.00000535, -0.00000307, -0.00000192, -0.00000098, 0.00000039, 0.00000203, 0.00000295, 0.00000219, -0.00000048, -0.00000426, -0.00000774, -0.00000973, -0.00000989, -0.00000872, -0.00000702, },
		{ 0.00000863, 0.00000853, 0.00000721, 0.00000434, 0.00000046, -0.00000322, -0.00000564, -0.00000668, -0.00000731, -0.00000892, -0.00001212, -0.00001600, -0.00001845, -0.00001743, -0.00001259, -0.00000593, -0.00000102, -0.00000097, -0.00000645, -0.00001490, -0.00002173, -0.00002285, -0.00001716, -0.00000727, 0.00000185, 0.00000581, 0.00000342, -0.00000248, -0.00000674, -0.00000493, 0.00000390, 0.00001658, 0.00002751, 0.00003187, 0.00002821, 0.00001904, 0.00000907, 0.00000240, 0.00000024, 0.00000071, 0.00000053, -0.00000253, -0.00000801, -0.00001313, -0.00001463, -0.00001088, -0.00000291, 0.00000629, 0.00001354, 0.00001707, 0.00001685, 0.00001365, 0.00000778, -0.00000124, -0.00001360, -0.00002751, -0.00003829, -0.00003964, -0.00002694, -0.00000089, 0.00003091, 0.00005607, 0.00006280, 0.00004605, 0.00001104, -0.00002820, -0.00005506, -0.00005820, -0.00003722, -0.00000279, 0.00002874, 0.00004353, 0.00003689, 0.00001489, -0.00000963, -0.00002426, -0.00002341, -0.00001045, 0.00000517, 0.00001382, 0.00001105, -0.00000057, -0.00001382, -0.00002154, -0.00002087, -0.00001437, -0.00000773, -0.00000588, -0.00000977, -0.00001602, -0.00001926, -0.00001567, -0.00000536, 0.00000776, 0.00001825, 0.00002222, 0.00001920, 0.00001184, 0.00000390, -0.00000201, -0.00000554, -0.00000808, -0.00001110, -0.00001466, -0.00001726, -0.00001680, -0.00001227, -0.00000465, 0.00000341, 0.00000898, 0.00001042, 0.00000810, 0.00000394, 0.00000004, -0.00000250, -0.00000399, -0.00000545, -0.00000763, -0.00001020, -0.00001196, -0.00001161, -0.00000873, -0.00000409, 0.00000079, 0.00000455, 0.00000671, 0.00000768, 0.00000822, },
		{ -0.00000728, -0.00001074, -0.00001317, -0.00001381, -0.00001299, -0.00001191, -0.00001175, -0.00001266, -0.00001339, -0.00001212, -0.00000784, -0.00000155, 0.00000384, 0.00000493, -0.00000008, -0.00001002, -0.00002096, -0.00002814, -0.00002848, -0.00002222, -0.00001268, -0.00000418, 0.00000055, 0.00000186, 0.00000273, 0.00000653, 0.00001452, 0.00002463, 0.00003242, 0.00003358, 0.00002659, 0.00001373, -0.00000002, -0.00000962, -0.00001253, -0.00000966, -0.00000434, -0.00000000, 0.00000184, 0.00000210, 0.00000300, 0.00000613, 0.00001108, 0.00001555, 0.00001674, 0.00001308, 0.00000510, -0.00000500, -0.00001462, -0.00002186, -0.00002572, -0.00002545, -0.00001996, -0.00000813, 0.00000972, 0.00003026, 0.00004677, 0.00005124, 0.00003831, 0.00000917, -0.00002709, -0.00005625, -0.00006504, -0.00004786, -0.00001041, 0.00003205, 0.00006146, 0.00006545, 0.00004328, 0.00000606, -0.00002894, -0.00004695, -0.00004266, -0.00002207, 0.00000162, 0.00001544, 0.00001334, -0.00000147, -0.00001932, -0.00003013, -0.00002890, -0.00001783, -0.00000399, 0.00000537, 0.00000709, 0.00000335, -0.00000048, 0.00000024, 0.00000633, 0.00001437, 0.00001906, 0.00001659, 0.00000704, -0.00000582, -0.00001675, -0.00002207, -0.00002129, -0.00001676, -0.00001163, -0.00000780, -0.00000493, -0.00000139, 0.00000405, 0.00001075, 0.00001625, 0.00001768, 0.00001356, 0.00000496, -0.00000501, -0.00001284, -0.00001637, -0.00001573, -0.00001278, -0.00000965, -0.00000741, -0.00000564, -0.00000315, 0.00000084, 0.00000587, 0.00001042, 0.00001283, 0.00001235, 0.00000959, 0.00000599, 0.00000286, 0.00000059, -0.00000140, -0.00000394, },
		{ -0.00001134, -0.00000877, -0.00000661, -0.00000593, -0.00000653, -0.00000698, -0.00000573, -0.00000248, 0.00000112, 0.00000214, -0.00000187, -0.00001109, -0.00002280, -0.00003241, -0.00003576, -0.00003129, -0.00002080, -0.00000831, 0.00000214, 0.00000868, 0.00001212, 0.00001482, 0.00001862, 0.00002322, 0.00002619, 0.00002464, 0.00001735, 0.00000609, -0.00000492, -0.00001112, -0.00001007, -0.00000271, 0.00000716, 0.00001508, 0.00001820, 0.00001650, 0.00001221, 0.00000806, 0.00000552, 0.00000415, 0.00000235, -0.00000122, -0.00000646, -0.00001174, -0.00001490, -0.00001444, -0.00001016, -0.00000298, 0.00000580, 0.00001497, 0.00002318, 0.00002839, 0.00002771, 0.00001849, 0.00000041, -0.00002266, -0.00004265, -0.00004988, -0.00003779, -0.00000727, 0.00003189, 0.00006409, 0.00007468, 0.00005717, 0.00001714, -0.00002951, -0.00006373, -0.00007220, -0.00005366, -0.00001917, 0.00001374, 0.00002996, 0.00002405, 0.00000221, -0.00002193, -0.00003497, -0.00003059, -0.00001212, 0.00001033, 0.00002599, 0.00002918, 0.00002143, 0.00000945, 0.00000032, -0.00000267, -0.00000133, -0.00000049, -0.00000418, -0.00001269, -0.00002231, -0.00002768, -0.00002523, -0.00001535, -0.00000204, 0.00000949, 0.00001577, 0.00001663, 0.00001451, 0.00001233, 0.00001131, 0.00001031, 0.00000698, -0.00000013, -0.00001009, -0.00001962, -0.00002486, -0.00002358, -0.00001640, -0.00000649, 0.00000236, 0.00000764, 0.00000923, 0.00000885, 0.00000851, 0.00000907, 0.00000981, 0.00000918, 0.00000616, 0.00000113, -0.00000427, -0.00000818, -0.00000973, -0.00000951, -0.00000901, -0.00000946, -0.00001099, -0.00001255, -0.00001285, },
		{ -0.00000441, -0.00000647, -0.00000801, -0.00000749, -0.00000508, -0.00000290, -0.00000385, -0.00000968, -0.00001949, -0.00002980, -0.00003616, -0.00003545, -0.00002756, -0.00001531, -0.00000289, 0.00000642, 0.00001169, 0.00001422, 0.00001607, 0.00001831, 0.00002028, 0.00002020, 0.00001680, 0.00001069, 0.00000448, 0.00000145, 0.00000355, 0.00001002, 0.00001760, 0.00002219, 0.00002110, 0.00001450, 0.00000518, -0.00000304, -0.00000735, -0.00000720, -0.00000417, -0.00000063, 0.00000175, 0.00000270, 0.00000309, 0.00000391, 0.00000552, 0.00000738, 0.00000864, 0.00000867, 0.00000730, 0.00000451, 0.00000012, -0.00000604, -0.00001335, -0.00001951, -0.00002064, -0.00001286, 0.00000503, 0.00002924, 0.00005096, 0.00005937, 0.00004683, 0.00001374, -0.00002986, -0.00006755, -0.00008360, -0.00007068, -0.00003400, 0.00001053, 0.00004358, 0.00005165, 0.00003353, 0.00000057, -0.00002937, -0.00004112, -0.00002958, -0.00000169, 0.00002794, 0.00004495, 0.00004239, 0.00002337, -0.00000181, -0.00002188, -0.00003040, -0.00002813, -0.00002091, -0.00001507, -0.00001334, -0.00001371, -0.00001163, -0.00000378, 0.00000920, 0.00002270, 0.00003082, 0.00002994, 0.00002075, 0.00000769, -0.00000377, -0.00001026, -0.00001199, -0.00001182, -0.00001278, -0.00001571, -0.00001867, -0.00001834, -0.00001253, -0.00000196, 0.00000978, 0.00001816, 0.00002019, 0.00001598, 0.00000843, 0.00000131, -0.00000287, -0.00000403, -0.00000399, -0.00000480, -0.00000723, -0.00001027, -0.00001207, -0.00001137, -0.00000856, -0.00000541, -0.00000391, -0.00000487, -0.00000736, -0.00000947, -0.00000962, -0.00000769, -0.00000510, -0.00000370, },
		{ -0.00000880, -0.00000566, -0.00000370, -0.00000614, -0.00001418, -0.00002591, -0.00003696, -0.00004267, -0.00004043, -0.00003088, -0.00001733, -0.00000384, 0.00000682, 0.00001408, 0.00001904, 0.00002304, 0.00002639, 0.00002816, 0.00002717, 0.00002320, 0.00001765, 0.00001295, 0.00001110, 0.00001234, 0.00001480, 0.00001562, 0.00001271, 0.00000620, -0.00000144, -0.00000671, -0.00000700, -0.00000204, 0.00000596, 0.00001353, 0.00001769, 0.00001720, 0.00001282, 0.00000647, 0.00000013, -0.00000492, -0.00000816, -0.00000951, -0.00000896, -0.00000663, -0.00000299, 0.00000114, 0.00000491, 0.00000815, 0.00001155, 0.00001619, 0.00002217, 0.00002737, 0.00002756, 0.00001820, -0.00000244, -0.00003065, -0.00005710, -0.00007014, -0.00006127, -0.00003040, 0.00001249, 0.00005066, 0.00006801, 0.00005701, 0.00002295, -0.00001801, -0.00004645, -0.00004917, -0.00002574, 0.00001143, 0.00004348, 0.00005462, 0.00003978, 0.00000656, -0.00002939, -0.00005263, -0.00005528, -0.00003981, -0.00001613, 0.00000467, 0.00001615, 0.00001876, 0.00001770, 0.00001827, 0.00002200, 0.00002582, 0.00002461, 0.00001519, -0.00000100, -0.00001825, -0.00002974, -0.00003137, -0.00002393, -0.00001234, -0.00000245, 0.00000229, 0.00000248, 0.00000161, 0.00000323, 0.00000826, 0.00001432, 0.00001729, 0.00001415, 0.00000511, -0.00000634, -0.00001541, -0.00001873, -0.00001614, -0.00001041, -0.00000527, -0.00000313, -0.00000387, -0.00000543, -0.00000562, -0.00000377, -0.00000120, -0.00000015, -0.00000203, -0.00000626, -0.00001059, -0.00001255, -0.00001115, -0.00000754, -0.00000427, -0.00000350, -0.00000551, -0.00000855, -0.00001010, },
		{ -0.00001265, -0.00002530, -0.00004059, -0.00005255, -0.00005596, -0.00004876, -0.00003283, -0.00001275, 0.00000655, 0.00002176, 0.00003194, 0.00003777, 0.00004027, 0.00003983, 0.00003632, 0.00002988, 0.00002166, 0.00001379, 0.00000847, 0.00000673, 0.00000782, 0.00000961, 0.00000996, 0.00000813, 0.00000524, 0.00000356, 0.00000501, 0.00000975, 0.00001579, 0.00001999, 0.00001964, 0.00001393, 0.00000434, -0.00000609, -0.00001417, -0.00001774, -0.00001623, -0.00001043, -0.00000187, 0.00000761, 0.00001620, 0.00002219, 0.00002433, 0.00002222, 0.00001664, 0.00000927, 0.00000185, -0.00000482, -0.00001128, -0.00001875, -0.00002747, -0.00003526, -0.00003753, -0.00002926, -0.00000843, 0.00002120, 0.00004991, 0.00006546, 0.00005891, 0.00003007, -0.00001083, -0.00004674, -0.00006147, -0.00004784, -0.00001193, 0.00002917, 0.00005543, 0.00005367, 0.00002427, -0.00001906, -0.00005608, -0.00007001, -0.00005548, -0.00002037, 0.00001906, 0.00004670, 0.00005419, 0.00004364, 0.00002459, 0.00000754, -0.00000183, -0.00000465, -0.00000635, -0.00001179, -0.00002136, -0.00003056, -0.00003303, -0.00002501, -0.00000824, 0.00001070, 0.00002391, 0.00002663, 0.00001959, 0.00000812, -0.00000132, -0.00000498, -0.00000363, -0.00000133, -0.00000230, -0.00000785, -0.00001552, -0.00002072, -0.00001988, -0.00001278, -0.00000275, 0.00000530, 0.00000793, 0.00000494, -0.00000079, -0.00000545, -0.00000678, -0.00000532, -0.00000361, -0.00000418, -0.00000756, -0.00001188, -0.00001420, -0.00001262, -0.00000778, -0.00000251, -0.00000008, -0.00000194, -0.00000677, -0.00001130, -0.00001259, -0.00001019, -0.00000671, -0.00000644, },
		{ -0.00006699, -0.00006259, -0.00004499, -0.00001881, 0.00000897, 0.00003200, 0.00004674, 0.00005280, 0.00005194, 0.00004646, 0.00003827, 0.00002874, 0.00001923, 0.00001141, 0.00000691, 0.00000653, 0.00000961, 0.00001411, 0.00001754, 0.00001828, 0.00001636, 0.00001333, 0.00001111, 0.00001065, 0.00001132, 0.00001126, 0.00000864, 0.00000295, -0.00000444, -0.00001083, -0.00001349, -0.00001094, -0.00000362, 0.00000644, 0.00001643, 0.00002374, 0.00002659, 0.00002424, 0.00001702, 0.00000626, -0.00000581, -0.00001650, -0.00002334, -0.00002502, -0.00002182, -0.00001542, -0.00000778, 0.00000007, 0.00000854, 0.00001867, 0.00003047, 0.00004128, 0.00004592, 0.00003886, 0.00001791, -0.00001289, -0.00004320, -0.00006011, -0.00005438, -0.00002602, 0.00001411, 0.00004825, 0.00005990, 0.00004226, 0.00000249, -0.00004106, -0.00006732, -0.00006293, -0.00002892, 0.00001966, 0.00006120, 0.00007804, 0.00006467, 0.00002942, -0.00001077, -0.00003940, -0.00004822, -0.00003977, -0.00002391, -0.00001101, -0.00000594, -0.00000638, -0.00000582, 0.00000101, 0.00001396, 0.00002725, 0.00003307, 0.00002666, 0.00000957, -0.00001098, -0.00002612, -0.00003025, -0.00002384, -0.00001257, -0.00000347, -0.00000077, -0.00000374, -0.00000793, -0.00000856, -0.00000392, 0.00000358, 0.00000914, 0.00000887, 0.00000240, -0.00000686, -0.00001391, -0.00001534, -0.00001133, -0.00000532, -0.00000153, -0.00000222, -0.00000628, -0.00001026, -0.00001088, -0.00000743, -0.00000230, 0.00000058, -0.00000149, -0.00000803, -0.00001552, -0.00001969, -0.00001841, -0.00001344, -0.00000972, -0.00001247, -0.00002389, -0.00004137, -0.00005827, },
		{ 0.00000698, 0.00003769, 0.00005759, 0.00006475, 0.00006130, 0.00005151, 0.00003960, 0.00002855, 0.00001995, 0.00001457, 0.00001269, 0.00001411, 0.00001778, 0.00002183, 0.00002403, 0.00002287, 0.00001829, 0.00001185, 0.00000591, 0.00000234, 0.00000161, 0.00000269, 0.00000382, 0.00000378, 0.00000263, 0.00000162, 0.00000234, 0.00000558, 0.00001073, 0.00001593, 0.00001892, 0.00001804, 0.00001288, 0.00000435, -0.00000565, -0.00001484, -0.00002096, -0.00002226, -0.00001790, -0.00000835, 0.00000450, 0.00001769, 0.00002819, 0.00003397, 0.00003473, 0.00003163, 0.00002622, 0.00001911, 0.00000959, -0.00000344, -0.00001962, -0.00003552, -0.00004492, -0.00004140, -0.00002235, 0.00000797, 0.00003850, 0.00005558, 0.00004955, 0.00002066, -0.00001942, -0.00005208, -0.00006041, -0.00003811, 0.00000636, 0.00005299, 0.00007947, 0.00007218, 0.00003308, -0.00002110, -0.00006702, -0.00008603, -0.00007270, -0.00003636, 0.00000479, 0.00003345, 0.00004144, 0.00003213, 0.00001650, 0.00000548, 0.00000368, 0.00000760, 0.00000930, 0.00000252, -0.00001272, -0.00002975, -0.00003942, -0.00003577, -0.00001986, 0.00000079, 0.00001649, 0.00002085, 0.00001401, 0.00000185, -0.00000803, -0.00001097, -0.00000756, -0.00000251, -0.00000096, -0.00000482, -0.00001161, -0.00001632, -0.00001498, -0.00000748, 0.00000214, 0.00000830, 0.00000740, 0.00000011, -0.00000911, -0.00001499, -0.00001483, -0.00001011, -0.00000541, -0.00000516, -0.00001062, -0.00001888, -0.00002478, -0.00002446, -0.00001826, -0.00001108, -0.00000972, -0.00001880, -0.00003748, -0.00005915, -0.00007426, -0.00007480, -0.00005811, -0.00002804, },
		{ 0.00007375, 0.00006087, 0.00004556, 0.00003262, 0.00002421, 0.00002025, 0.00001959, 0.00002081, 0.00002258, 0.00002360, 0.00002266, 0.00001911, 0.00001329, 0.00000674, 0.00000167, -0.00000007, 0.00000206, 0.00000693, 0.00001231, 0.00001602, 0.00001691, 0.00001531, 0.00001249, 0.00000975, 0.00000765, 0.00000580, 0.00000334, -0.00000027, -0.00000475, -0.00000895, -0.00001131, -0.00001049, -0.00000585, 0.00000236, 0.00001299, 0.00002421, 0.00003372, 0.00003923, 0.00003909, 0.00003294, 0.00002207, 0.00000909, -0.00000307, -0.00001232, -0.00001818, -0.00002135, -0.00002266, -0.00002173, -0.00001682, -0.00000606, 0.00001033, 0.00002841, 0.00004094, 0.00004046, 0.00002368, -0.00000517, -0.00003451, -0.00005007, -0.00004204, -0.00001126, 0.00002945, 0.00006020, 0.00006356, 0.00003389, -0.00001854, -0.00007132, -0.00010019, -0.00009088, -0.00004631, 0.00001462, 0.00006613, 0.00008793, 0.00007438, 0.00003596, -0.00000735, -0.00003684, -0.00004396, -0.00003291, -0.00001607, -0.00000563, -0.00000647, -0.00001435, -0.00001989, -0.00001549, -0.00000062, 0.00001767, 0.00002916, 0.00002681, 0.00001111, -0.00001023, -0.00002686, -0.00003171, -0.00002461, -0.00001165, -0.00000085, 0.00000281, -0.00000005, -0.00000446, -0.00000507, -0.00000016, 0.00000717, 0.00001111, 0.00000728, -0.00000406, -0.00001781, -0.00002707, -0.00002743, -0.00001974, -0.00000961, -0.00000383, -0.00000606, -0.00001457, -0.00002345, -0.00002658, -0.00002193, -0.00001333, -0.00000856, -0.00001469, -0.00003339, -0.00005905, -0.00008085, -0.00008779, -0.00007404, -0.00004171, 0.00000014, 0.00003959, 0.00006685, 0.00007765, },
		{ 0.00002708, 0.00002302, 0.00002266, 0.00002344, 0.00002364, 0.00002252, 0.00002001, 0.00001638, 0.00001224, 0.00000861, 0.00000679, 0.00000776, 0.00001155, 0.00001692, 0.00002164, 0.00002348, 0.00002125, 0.00001538, 0.00000764, 0.00000032, -0.00000482, -0.00000708, -0.00000678, -0.00000465, -0.00000128, 0.00000319, 0.00000884, 0.00001561, 0.00002296, 0.00002983, 0.00003490, 0.00003696, 0.00003517, 0.00002933, 0.00001995, 0.00000840, -0.00000323, -0.00001254, -0.00001753, -0.00001740, -0.00001299, -0.00000637, 0.00000013, 0.00000516, 0.00000885, 0.00001228, 0.00001604, 0.00001899, 0.00001819, 0.00001062, -0.00000420, -0.00002233, -0.00003599, -0.00003699, -0.00002164, 0.00000555, 0.00003229, 0.00004360, 0.00002963, -0.00000775, -0.00005375, -0.00008637, -0.00008685, -0.00004979, 0.00001255, 0.00007470, 0.00010951, 0.00010111, 0.00005282, -0.00001434, -0.00007173, -0.00009682, -0.00008325, -0.00004246, 0.00000358, 0.00003423, 0.00004010, 0.00002590, 0.00000555, -0.00000718, -0.00000657, 0.00000283, 0.00001053, 0.00000763, -0.00000708, -0.00002640, -0.00003926, -0.00003776, -0.00002193, 0.00000031, 0.00001802, 0.00002363, 0.00001688, 0.00000419, -0.00000603, -0.00000877, -0.00000527, -0.00000142, -0.00000324, -0.00001235, -0.00002458, -0.00003260, -0.00003083, -0.00001940, -0.00000452, 0.00000511, 0.00000366, -0.00000813, -0.00002331, -0.00003297, -0.00003194, -0.00002230, -0.00001256, -0.00001270, -0.00002800, -0.00005510, -0.00008288, -0.00009751, -0.00008924, -0.00005724, -0.00001005, 0.00003833, 0.00007459, 0.00009123, 0.00008860, 0.00007317, 0.00005368, 0.00003720, },
		{ 0.00002658, 0.00002488, 0.00002204, 0.00001908, 0.00001678, 0.00001551, 0.00001546, 0.00001667, 0.00001884, 0.00002108, 0.00002204, 0.00002040, 0.00001562, 0.00000849, 0.00000097, -0.00000452, -0.00000612, -0.00000328, 0.00000319, 0.00001155, 0.00002000, 0.00002721, 0.00003250, 0.00003564, 0.00003653, 0.00003503, 0.00003109, 0.00002491, 0.00001707, 0.00000856, 0.00000060, -0.00000560, -0.00000901, -0.00000898, -0.00000558, 0.00000025, 0.00000673, 0.00001162, 0.00001313, 0.00001073, 0.00000548, -0.00000044, -0.00000498, -0.00000748, -0.00000903, -0.00001160, -0.00001624, -0.00002160, -0.00002403, -0.00001960, -0.00000725, 0.00000907, 0.00002106, 0.00001992, 0.00000191, -0.00002774, -0.00005531, -0.00006436, -0.00004443, 0.00000178, 0.00005720, 0.00009685, 0.00009956, 0.00005917, -0.00001069, -0.00008158, -0.00012295, -0.00011667, -0.00006581, 0.00000681, 0.00006964, 0.00009752, 0.00008298, 0.00003819, -0.00001281, -0.00004713, -0.00005406, -0.00003845, -0.00001554, -0.00000073, -0.00000078, -0.00001113, -0.00002038, -0.00001857, -0.00000387, 0.00001630, 0.00003019, 0.00002918, 0.00001318, -0.00000946, -0.00002713, -0.00003199, -0.00002429, -0.00001155, -0.00000325, -0.00000463, -0.00001359, -0.00002251, -0.00002374, -0.00001509, -0.00000161, 0.00000740, 0.00000447, -0.00001086, -0.00003118, -0.00004535, -0.00004536, -0.00003166, -0.00001363, -0.00000463, -0.00001429, -0.00004243, -0.00007808, -0.00010432, -0.00010644, -0.00007913, -0.00002902, 0.00002847, 0.00007611, 0.00010209, 0.00010391, 0.00008776, 0.00006425, 0.00004320, 0.00003003, 0.00002502, 0.00002509, 0.00002644, },
		{ 0.00002206, 0.00002163, 0.00002182, 0.00002197, 0.00002168, 0.00002076, 0.00001901, 0.00001623, 0.00001249, 0.00000845, 0.00000540, 0.00000488, 0.00000792, 0.00001444, 0.00002310, 0.00003177, 0.00003833, 0.00004145, 0.00004086, 0.00003718, 0.00003147, 0.00002476, 0.00001788, 0.00001145, 0.00000604, 0.00000213, 0.00000008, -0.00000002, 0.00000154, 0.00000414, 0.00000684, 0.00000854, 0.00000814, 0.00000483, -0.00000146, -0.00000978, -0.00001813, -0.00002410, -0.00002587, -0.00002320, -0.00001774, -0.00001222, -0.00000898, -0.00000847, -0.00000894, -0.00000766, -0.00000313, 0.00000323, 0.00000713, 0.00000391, -0.00000791, -0.00002405, -0.00003525, -0.00003166, -0.00000907, 0.00002660, 0.00005986, 0.00007205, 0.00005114, -0.00000020, -0.00006313, -0.00010972, -0.00011588, -0.00007400, 0.00000160, 0.00007985, 0.00012687, 0.00012193, 0.00006753, -0.00001189, -0.00008186, -0.00011434, -0.00010038, -0.00005265, 0.00000281, 0.00004076, 0.00004893, 0.00003213, 0.00000673, -0.00001031, -0.00001122, -0.00000059, 0.00000939, 0.00000787, -0.00000730, -0.00002822, -0.00004224, -0.00004027, -0.00002262, 0.00000101, 0.00001753, 0.00001838, 0.00000452, -0.00001461, -0.00002733, -0.00002718, -0.00001678, -0.00000579, -0.00000429, -0.00001578, -0.00003440, -0.00004835, -0.00004736, -0.00002994, -0.00000553, 0.00000995, 0.00000291, -0.00002960, -0.00007680, -0.00011845, -0.00013421, -0.00011327, -0.00005957, 0.00001022, 0.00007429, 0.00011505, 0.00012559, 0.00011070, 0.00008269, 0.00005482, 0.00003582, 0.00002774, 0.00002740, 0.00002972, 0.00003081, 0.00002944, 0.00002653, 0.00002372, },
		{ 0.00002468, 0.00002184, 0.00001912, 0.00001721, 0.00001642, 0.00001688, 0.00001873, 0.00002218, 0.00002729, 0.00003361, 0.00003995, 0.00004470, 0.00004641, 0.00004440, 0.00003909, 0.00003176, 0.00002401, 0.00001715, 0.00001186, 0.00000819, 0.00000581, 0.00000429, 0.00000324, 0.00000229, 0.00000109, -0.00000073, -0.00000352, -0.00000750, -0.00001265, -0.00001866, -0.00002477, -0.00002980, -0.00003231, -0.00003105, -0.00002562, -0.00001698, -0.00000746, -0.00000001, 0.00000313, 0.00000164, -0.00000258, -0.00000635, -0.00000713, -0.00000466, -0.00000132, -0.00000062, -0.00000456, -0.00001149, -0.00001626, -0.00001308, 0.00000023, 0.00001925, 0.00003362, 0.00003191, 0.00000878, -0.00002990, -0.00006730, -0.00008261, -0.00006189, -0.00000722, 0.00006138, 0.00011336, 0.00012182, 0.00007754, -0.00000471, -0.00009138, -0.00014520, -0.00014278, -0.00008585, -0.00000036, 0.00007622, 0.00011283, 0.00009894, 0.00004767, -0.00001294, -0.00005522, -0.00006536, -0.00004813, -0.00002111, -0.00000275, -0.00000160, -0.00001275, -0.00002296, -0.00002052, -0.00000348, 0.00001889, 0.00003214, 0.00002603, 0.00000141, -0.00002989, -0.00005198, -0.00005438, -0.00003817, -0.00001502, 0.00000026, -0.00000104, -0.00001622, -0.00003313, -0.00003826, -0.00002606, -0.00000357, 0.00001260, 0.00000572, -0.00003061, -0.00008696, -0.00014096, -0.00016735, -0.00015001, -0.00008986, -0.00000467, 0.00007889, 0.00013672, 0.00015667, 0.00014164, 0.00010590, 0.00006714, 0.00003875, 0.00002572, 0.00002519, 0.00003038, 0.00003507, 0.00003628, 0.00003441, 0.00003155, 0.00002946, 0.00002854, 0.00002801, 0.00002686, },
		{ 0.00002631, 0.00003178, 0.00003806, 0.00004409, 0.00004913, 0.00005272, 0.00005439, 0.00005359, 0.00004990, 0.00004335, 0.00003469, 0.00002523, 0.00001639, 0.00000919, 0.00000387, -0.00000003, -0.00000332, -0.00000670, -0.00001049, -0.00001462, -0.00001874, -0.00002249, -0.00002561, -0.00002798, -0.00002954, -0.00003022, -0.00002980, -0.00002795, -0.00002428, -0.00001859, -0.00001126, -0.00000343, 0.00000308, 0.00000634, 0.00000525, 0.00000034, -0.00000610, -0.00001093, -0.00001168, -0.00000809, -0.00000243, 0.00000160, 0.00000119, -0.00000362, -0.00000954, -0.00001185, -0.00000765, 0.00000156, 0.00000987, 0.00001011, -0.00000119, -0.00002010, -0.00003569, -0.00003527, -0.00001221, 0.00002798, 0.00006763, 0.00008422, 0.00006211, 0.00000283, -0.00007244, -0.00013066, -0.00014218, -0.00009618, -0.00000795, 0.00008639, 0.00014608, 0.00014497, 0.00008408, -0.00000892, -0.00009332, -0.00013493, -0.00012164, -0.00006721, -0.00000203, 0.00004366, 0.00005464, 0.00003602, 0.00000705, -0.00001211, -0.00001239, 0.00000017, 0.00001021, 0.00000435, -0.00001929, -0.00004938, -0.00006827, -0.00006327, -0.00003497, 0.00000238, 0.00002899, 0.00003117, 0.00000932, -0.00002236, -0.00004475, -0.00004567, -0.00002783, -0.00000784, -0.00000654, -0.00003618, -0.00009163, -0.00015095, -0.00018524, -0.00017281, -0.00011023, -0.00001453, 0.00008420, 0.00015603, 0.00018348, 0.00016721, 0.00012317, 0.00007364, 0.00003706, 0.00002158, 0.00002448, 0.00003652, 0.00004797, 0.00005315, 0.00005150, 0.00004593, 0.00003987, 0.00003521, 0.00003194, 0.00002909, 0.00002606, 0.00002328, 0.00002185, 0.00002278, },
		{ 0.00007347, 0.00006987, 0.00006278, 0.00005310, 0.00004162, 0.00002905, 0.00001622, 0.00000410, -0.00000634, -0.00001446, -0.00002026, -0.00002431, -0.00002743, -0.00003023, -0.00003286, -0.00003493, -0.00003582, -0.00003506, -0.00003258, -0.00002878, -0.00002431, -0.00001988, -0.00001593, -0.00001257, -0.00000955, -0.00000656, -0.00000349, -0.00000066, 0.00000113, 0.00000097, -0.00000154, -0.00000579, -0.00001007, -0.00001218, -0.00001062, -0.00000568, 0.00000035, 0.00000409, 0.00000294, -0.00000310, -0.00001100, -0.00001606, -0.00001463, -0.00000662, 0.00000396, 0.00001102, 0.00001021, 0.00000214, -0.00000716, -0.00000963, -0.00000063, 0.00001667, 0.00003128, 0.00002974, 0.00000461, -0.00003873, -0.00008172, -0.00010017, -0.00007701, -0.00001343, 0.00006797, 0.00013141, 0.00014440, 0.00009463, -0.00000177, -0.00010570, -0.00017253, -0.00017321, -0.00010829, -0.00000765, 0.00008415, 0.00012942, 0.00011462, 0.00005482, -0.00001645, -0.00006574, -0.00007639, -0.00005464, -0.00002269, -0.00000363, -0.00000756, -0.00002684, -0.00004269, -0.00003853, -0.00001156, 0.00002468, 0.00004823, 0.00004233, 0.00000632, -0.00004297, -0.00008051, -0.00008763, -0.00006347, -0.00002631, -0.00000372, -0.00001677, -0.00006720, -0.00013467, -0.00018598, -0.00019137, -0.00013945, -0.00004313, 0.00006633, 0.00015361, 0.00019467, 0.00018568, 0.00014205, 0.00008895, 0.00004912, 0.00003378, 0.00004056, 0.00005794, 0.00007285, 0.00007711, 0.00006989, 0.00005590, 0.00004153, 0.00003129, 0.00002640, 0.00002567, 0.00002731, 0.00003040, 0.00003512, 0.00004196, 0.00005070, 0.00006007, 0.00006808, 0.00007288, },
		{ 0.00000350, -0.00001272, -0.00002600, -0.00003618, -0.00004318, -0.00004702, -0.00004804, -0.00004692, -0.00004456, -0.00004168, -0.00003859, -0.00003513, -0.00003091, -0.00002569, -0.00001970, -0.00001368, -0.00000862, -0.00000536, -0.00000424, -0.00000493, -0.00000657, -0.00000809, -0.00000870, -0.00000814, -0.00000687, -0.00000574, -0.00000548, -0.00000621, -0.00000721, -0.00000724, -0.00000539, -0.00000185, 0.00000180, 0.00000322, 0.00000071, -0.00000542, -0.00001245, -0.00001628, -0.00001370, -0.00000452, 0.00000763, 0.00001687, 0.00001819, 0.00001062, -0.00000186, -0.00001234, -0.00001519, -0.00000998, -0.00000225, -0.00000036, -0.00000954, -0.00002700, -0.00004158, -0.00003924, -0.00001208, 0.00003423, 0.00007982, 0.00009867, 0.00007222, 0.00000170, -0.00008838, -0.00015893, -0.00017425, -0.00012057, -0.00001537, 0.00009842, 0.00017152, 0.00017161, 0.00009914, -0.00001272, -0.00011426, -0.00016328, -0.00014471, -0.00007584, 0.00000487, 0.00005896, 0.00006732, 0.00003754, -0.00000385, -0.00002896, -0.00002523, -0.00000169, 0.00001833, 0.00001327, -0.00002196, -0.00007166, -0.00010806, -0.00010822, -0.00006898, -0.00001094, 0.00003131, 0.00002786, -0.00002928, -0.00011987, -0.00020376, -0.00023975, -0.00020529, -0.00010708, 0.00002247, 0.00014046, 0.00021250, 0.00022606, 0.00019255, 0.00013846, 0.00009114, 0.00006683, 0.00006613, 0.00007772, 0.00008700, 0.00008439, 0.00006919, 0.00004807, 0.00003017, 0.00002203, 0.00002509, 0.00003636, 0.00005104, 0.00006523, 0.00007713, 0.00008652, 0.00009351, 0.00009748, 0.00009708, 0.00009102, 0.00007899, 0.00006206, 0.00004235, 0.00002221, },
		{ -0.00006494, -0.00005913, -0.00005193, -0.00004404, -0.00003621, -0.00002913, -0.00002321, -0.00001837, -0.00001425, -0.00001050, -0.00000709, -0.00000443, -0.00000317, -0.00000379, -0.00000624, -0.00000982, -0.00001328, -0.00001533, -0.00001512, -0.00001260, -0.00000861, -0.00000448, -0.00000154, -0.00000044, -0.00000091, -0.00000192, -0.00000233, -0.00000167, -0.00000055, -0.00000037, -0.00000235, -0.00000645, -0.00001089, -0.00001279, -0.00000981, -0.00000192, 0.00000788, 0.00001466, 0.00001412, 0.00000526, -0.00000846, -0.00002053, -0.00002485, -0.00001936, -0.00000755, 0.00000351, 0.00000755, 0.00000342, -0.00000364, -0.00000525, 0.00000400, 0.00002124, 0.00003473, 0.00002971, -0.00000204, -0.00005412, -0.00010477, -0.00012552, -0.00009616, -0.00001817, 0.00008123, 0.00015869, 0.00017451, 0.00011334, -0.00000526, -0.00013303, -0.00021437, -0.00021261, -0.00012818, 0.00000102, 0.00011774, 0.00017290, 0.00014850, 0.00006439, -0.00003427, -0.00010147, -0.00011361, -0.00007912, -0.00003005, -0.00000135, -0.00001002, -0.00004608, -0.00008022, -0.00008330, -0.00004596, 0.00001327, 0.00005660, 0.00004783, -0.00002599, -0.00014417, -0.00025946, -0.00031925, -0.00029022, -0.00017409, -0.00000681, 0.00015781, 0.00027141, 0.00030995, 0.00027992, 0.00021050, 0.00013714, 0.00008564, 0.00006381, 0.00006339, 0.00006909, 0.00006881, 0.00005964, 0.00004735, 0.00004116, 0.00004748, 0.00006639, 0.00009204, 0.00011621, 0.00013225, 0.00013738, 0.00013255, 0.00012044, 0.00010354, 0.00008308, 0.00005946, 0.00003329, 0.00000615, -0.00001947, -0.00004095, -0.00005650, -0.00006572, -0.00006933, -0.00006865, },
		{ -0.00001574, -0.00001135, -0.00000875, -0.00000772, -0.00000774, -0.00000813, -0.00000845, -0.00000868, -0.00000916, -0.00001030, -0.00001216, -0.00001428, -0.00001573, -0.00001554, -0.00001316, -0.00000885, -0.00000367, 0.00000088, 0.00000344, 0.00000342, 0.00000124, -0.00000180, -0.00000419, -0.00000502, -0.00000441, -0.00000350, -0.00000364, -0.00000545, -0.00000812, -0.00000974, -0.00000839, -0.00000359, 0.00000289, 0.00000759, 0.00000722, 0.00000085, -0.00000898, -0.00001710, -0.00001843, -0.00001112, 0.00000190, 0.00001409, 0.00001884, 0.00001348, 0.00000111, -0.00001118, -0.00001682, -0.00001446, -0.00000930, -0.00000993, -0.00002174, -0.00004131, -0.00005582, -0.00004931, -0.00001319, 0.00004501, 0.00010085, 0.00012255, 0.00008766, -0.00000186, -0.00011510, -0.00020275, -0.00021950, -0.00014760, -0.00000929, 0.00013987, 0.00023509, 0.00023278, 0.00013218, -0.00002329, -0.00016628, -0.00023757, -0.00021382, -0.00011694, -0.00000065, 0.00007837, 0.00008918, 0.00004007, -0.00002970, -0.00007466, -0.00007025, -0.00002704, 0.00001568, 0.00001332, -0.00005670, -0.00017882, -0.00030355, -0.00037013, -0.00033600, -0.00019813, 0.00000445, 0.00020764, 0.00034989, 0.00039699, 0.00035272, 0.00025210, 0.00014232, 0.00006179, 0.00002703, 0.00003199, 0.00005755, 0.00008443, 0.00010249, 0.00011255, 0.00012162, 0.00013576, 0.00015512, 0.00017359, 0.00018248, 0.00017543, 0.00015152, 0.00011518, 0.00007353, 0.00003311, -0.00000231, -0.00003173, -0.00005563, -0.00007438, -0.00008751, -0.00009410, -0.00009368, -0.00008702, -0.00007604, -0.00006318, -0.00005051, -0.00003923, -0.00002970, -0.00002189, },
		{ -0.00001328, -0.00001436, -0.00001459, -0.00001386, -0.00001252, -0.00001122, -0.00001049, -0.00001035, -0.00001032, -0.00000964, -0.00000775, -0.00000469, -0.00000120, 0.00000156, 0.00000253, 0.00000133, -0.00000155, -0.00000482, -0.00000705, -0.00000735, -0.00000591, -0.00000389, -0.00000284, -0.00000371, -0.00000617, -0.00000864, -0.00000911, -0.00000646, -0.00000138, 0.00000370, 0.00000591, 0.00000375, -0.00000172, -0.00000714, -0.00000870, -0.00000466, 0.00000312, 0.00000978, 0.00001011, 0.00000191, -0.00001217, -0.00002557, -0.00003148, -0.00002700, -0.00001521, -0.00000339, 0.00000173, -0.00000112, -0.00000625, -0.00000458, 0.00000934, 0.00003120, 0.00004642, 0.00003720, -0.00000579, -0.00007356, -0.00013792, -0.00016235, -0.00012092, -0.00001569, 0.00011770, 0.00022178, 0.00024283, 0.00015856, -0.00000663, -0.00018829, -0.00030927, -0.00031555, -0.00020304, -0.00002107, 0.00014995, 0.00023638, 0.00020602, 0.00008208, -0.00007053, -0.00017935, -0.00020120, -0.00014279, -0.00005464, -0.00000208, -0.00002889, -0.00013263, -0.00026459, -0.00035381, -0.00034272, -0.00021604, -0.00000879, 0.00020992, 0.00036823, 0.00042241, 0.00037270, 0.00025838, 0.00013631, 0.00005498, 0.00003639, 0.00007270, 0.00013661, 0.00019778, 0.00023654, 0.00024880, 0.00024200, 0.00022656, 0.00020837, 0.00018645, 0.00015583, 0.00011272, 0.00005853, 0.00000019, -0.00005280, -0.00009257, -0.00011580, -0.00012406, -0.00012181, -0.00011368, -0.00010265, -0.00008980, -0.00007527, -0.00005955, -0.00004393, -0.00003026, -0.00001999, -0.00001356, -0.00001037, -0.00000924, -0.00000917, -0.00000962, -0.00001051, -0.00001182, },
		{ -0.00000835, -0.00000614, -0.00000489, -0.00000493, -0.00000595, -0.00000710, -0.00000748, -0.00000664, -0.00000483, -0.00000287, -0.00000172, -0.00000197, -0.00000343, -0.00000525, -0.00000636, -0.00000604, -0.00000444, -0.00000255, -0.00000170, -0.00000276, -0.00000546, -0.00000833, -0.00000938, -0.00000724, -0.00000218, 0.00000381, 0.00000785, 0.00000780, 0.00000358, -0.00000259, -0.00000744, -0.00000859, -0.00000622, -0.00000311, -0.00000289, -0.00000752, -0.00001545, -0.00002205, -0.00002220, -0.00001372, 0.00000056, 0.00001385, 0.00001912, 0.00001353, 0.00000051, -0.00001212, -0.00001733, -0.00001416, -0.00000925, -0.00001274, -0.00003043, -0.00005691, -0.00007509, -0.00006426, -0.00001353, 0.00006672, 0.00014358, 0.00017387, 0.00012603, 0.00000087, -0.00016100, -0.00029195, -0.00032676, -0.00023577, -0.00004451, 0.00017196, 0.00032002, 0.00033050, 0.00019279, -0.00003731, -0.00026160, -0.00038477, -0.00035932, -0.00020798, -0.00001243, 0.00012658, 0.00013795, 0.00001423, -0.00018554, -0.00036475, -0.00043505, -0.00035656, -0.00015377, 0.00009909, 0.00031412, 0.00042879, 0.00042956, 0.00035085, 0.00025263, 0.00018978, 0.00018864, 0.00024040, 0.00031173, 0.00036467, 0.00037461, 0.00033817, 0.00026950, 0.00018928, 0.00011374, 0.00004924, -0.00000618, -0.00005692, -0.00010425, -0.00014435, -0.00017053, -0.00017759, -0.00016528, -0.00013888, -0.00010678, -0.00007685, -0.00005367, -0.00003789, -0.00002761, -0.00002045, -0.00001509, -0.00001150, -0.00001017, -0.00001113, -0.00001359, -0.00001619, -0.00001779, -0.00001803, -0.00001727, -0.00001620, -0.00001520, -0.00001421, -0.00001283, -0.00001080, },
		{ -0.00000743, -0.00000965, -0.00001103, -0.00001075, -0.00000876, -0.00000579, -0.00000297, -0.00000131, -0.00000112, -0.00000197, -0.00000299, -0.00000337, -0.00000292, -0.00000216, -0.00000202, -0.00000318, -0.00000550, -0.00000786, -0.00000864, -0.00000664, -0.00000201, 0.00000354, 0.00000738, 0.00000728, 0.00000277, -0.00000440, -0.00001104, -0.00001424, -0.00001301, -0.00000887, -0.00000488, -0.00000360, -0.00000528, -0.00000762, -0.00000728, -0.00000253, 0.00000500, 0.00001055, 0.00000906, -0.00000140, -0.00001749, -0.00003179, -0.00003688, -0.00003007, -0.00001562, -0.00000243, 0.00000176, -0.00000372, -0.00001084, -0.00000762, 0.00001280, 0.00004410, 0.00006612, 0.00005401, -0.00000604, -0.00010302, -0.00019883, -0.00024199, -0.00019317, -0.00005043, 0.00014034, 0.00029865, 0.00034356, 0.00023406, -0.00000450, -0.00028358, -0.00048612, -0.00051899, -0.00035872, -0.00006779, 0.00022855, 0.00039595, 0.00035120, 0.00010093, -0.00025934, -0.00058557, -0.00074622, -0.00067607, -0.00039912, -0.00001240, 0.00035939, 0.00061594, 0.00071636, 0.00068413, 0.00058380, 0.00048411, 0.00042671, 0.00041367, 0.00041644, 0.00039787, 0.00033414, 0.00022580, 0.00009457, -0.00002992, -0.00012454, -0.00018101, -0.00020507, -0.00020912, -0.00020382, -0.00019333, -0.00017608, -0.00014920, -0.00011299, -0.00007260, -0.00003609, -0.00001064, 0.00000074, 0.00000009, -0.00000726, -0.00001563, -0.00002144, -0.00002390, -0.00002419, -0.00002375, -0.00002324, -0.00002233, -0.00002039, -0.00001729, -0.00001364, -0.00001044, -0.00000840, -0.00000755, -0.00000725, -0.00000679, -0.00000588, -0.00000490, -0.00000459, -0.00000547, },
		{ -0.00000843, -0.00000468, -0.00000088, 0.00000166, 0.00000227, 0.00000114, -0.00000084, -0.00000261, -0.00000353, -0.00000368, -0.00000371, -0.00000431, -0.00000569, -0.00000724, -0.00000782, -0.00000643, -0.00000301, 0.00000126, 0.00000433, 0.00000431, 0.00000063, -0.00000538, -0.00001097, -0.00001338, -0.00001137, -0.00000601, -0.00000021, 0.00000300, 0.00000223, -0.00000146, -0.00000538, -0.00000710, -0.00000636, -0.00000538, -0.00000726, -0.00001342, -0.00002178, -0.00002717, -0.00002436, -0.00001181, 0.00000619, 0.00002100, 0.00002431, 0.00001347, -0.00000612, -0.00002373, -0.00003008, -0.00002427, -0.00001578, -0.00001913, -0.00004312, -0.00008126, -0.00011032, -0.00010060, -0.00003446, 0.00007679, 0.00018887, 0.00023954, 0.00017865, -0.00000039, -0.00024532, -0.00045789, -0.00053350, -0.00041189, -0.00011456, 0.00025177, 0.00053207, 0.00058715, 0.00035745, -0.00010308, -0.00064055, -0.00105831, -0.00119384, -0.00098238, -0.00048004, 0.00016193, 0.00075881, 0.00116159, 0.00130473, 0.00121496, 0.00098284, 0.00071458, 0.00048801, 0.00033085, 0.00022639, 0.00013795, 0.00003645, -0.00008315, -0.00020236, -0.00029296, -0.00033397, -0.00032236, -0.00027273, -0.00020821, -0.00014899, -0.00010484, -0.00007468, -0.00005187, -0.00003078, -0.00001081, 0.00000418, 0.00000947, 0.00000343, -0.00001084, -0.00002699, -0.00003863, -0.00004229, -0.00003848, -0.00003052, -0.00002222, -0.00001598, -0.00001215, -0.00000978, -0.00000784, -0.00000607, -0.00000498, -0.00000513, -0.00000653, -0.00000842, -0.00000977, -0.00000995, -0.00000913, -0.00000818, -0.00000801, -0.00000895, -0.00001048, -0.00001147, -0.00001090, },
		{ 0.00000317, 0.00000169, -0.00000097, -0.00000357, -0.00000516, -0.00000554, -0.00000521, -0.00000500, -0.00000546, -0.00000644, -0.00000712, -0.00000656, -0.00000433, -0.00000102, 0.00000193, 0.00000290, 0.00000111, -0.00000281, -0.00000692, -0.00000900, -0.00000783, -0.00000405, 0.00000004, 0.00000177, -0.00000032, -0.00000552, -0.00001123, -0.00001460, -0.00001424, -0.00001113, -0.00000786, -0.00000665, -0.00000755, -0.00000822, -0.00000557, 0.00000140, 0.00000978, 0.00001349, 0.00000673, -0.00001176, -0.00003639, -0.00005654, -0.00006206, -0.00004965, -0.00002568, -0.00000323, 0.00000591, -0.00000038, -0.00001181, -0.00001150, 0.00001156, 0.00005068, 0.00007896, 0.00006067, -0.00002685, -0.00017184, -0.00032135, -0.00039726, -0.00033233, -0.00011088, 0.00020729, 0.00049633, 0.00060936, 0.00044317, -0.00000573, -0.00062153, -0.00119982, -0.00151991, -0.00143034, -0.00091152, -0.00008806, 0.00081491, 0.00155322, 0.00194866, 0.00194344, 0.00160482, 0.00108320, 0.00054618, 0.00011832, -0.00014997, -0.00027743, -0.00032060, -0.00033571, -0.00035261, -0.00036992, -0.00036823, -0.00033016, -0.00025471, -0.00015918, -0.00006987, -0.00000855, 0.00001693, 0.00001372, -0.00000275, -0.00001797, -0.00002496, -0.00002510, -0.00002421, -0.00002705, -0.00003396, -0.00004108, -0.00004344, -0.00003840, -0.00002726, -0.00001434, -0.00000435, 0.00000008, -0.00000071, -0.00000429, -0.00000795, -0.00001016, -0.00001102, -0.00001146, -0.00001222, -0.00001318, -0.00001356, -0.00001264, -0.00001045, -0.00000788, -0.00000613, -0.00000592, -0.00000703, -0.00000833, -0.00000851, -0.00000684, -0.00000360, 0.00000003, 0.00000261, },
		{ -0.00000579, -0.00000624, -0.00000566, -0.00000494, -0.00000484, -0.00000550, -0.00000634, -0.00000638, -0.00000495, -0.00000217, 0.00000092, 0.00000283, 0.00000247, -0.00000018, -0.00000392, -0.00000694, -0.00000786, -0.00000659, -0.00000444, -0.00000334, -0.00000459, -0.00000786, -0.00001126, -0.00001242, -0.00001009, -0.00000508, 0.00000003, 0.00000247, 0.00000099, -0.00000324, -0.00000758, -0.00000994, -0.00001067, -0.00001263, -0.00001915, -0.00003090, -0.00004387, -0.00005040, -0.00004329, -0.00002095, 0.00000982, 0.00003587, 0.00004421, 0.00002970, -0.00000117, -0.00003330, -0.00005222, -0.00005412, -0.00004986, -0.00005871, -0.00009384, -0.00014766, -0.00018763, -0.00016897, -0.00006131, 0.00012313, 0.00031720, 0.00041434, 0.00031176, -0.00003508, -0.00057206, -0.00114375, -0.00153713, -0.00155907, -0.00111732, -0.00026967, 0.00078277, 0.00175909, 0.00239480, 0.00253257, 0.00217129, 0.00145557, 0.00061339, -0.00012949, -0.00062281, -0.00082569, -0.00079421, -0.00063566, -0.00045467, -0.00031519, -0.00023074, -0.00017989, -0.00013318, -0.00007520, -0.00001172, 0.00003803, 0.00005635, 0.00003901, -0.00000222, -0.00004659, -0.00007543, -0.00008069, -0.00006657, -0.00004468, -0.00002638, -0.00001715, -0.00001549, -0.00001602, -0.00001404, -0.00000857, -0.00000228, 0.00000094, -0.00000112, -0.00000758, -0.00001518, -0.00002041, -0.00002149, -0.00001900, -0.00001509, -0.00001184, -0.00001012, -0.00000938, -0.00000848, -0.00000675, -0.00000452, -0.00000290, -0.00000284, -0.00000443, -0.00000670, -0.00000814, -0.00000764, -0.00000510, -0.00000155, 0.00000143, 0.00000254, 0.00000146, -0.00000113, -0.00000393, },
		{ -0.00000449, -0.00000491, -0.00000578, -0.00000622, -0.00000538, -0.00000311, -0.00000014, 0.00000218, 0.00000257, 0.00000061, -0.00000299, -0.00000665, -0.00000880, -0.00000872, -0.00000693, -0.00000487, -0.00000393, -0.00000453, -0.00000585, -0.00000642, -0.00000528, -0.00000288, -0.00000109, -0.00000205, -0.00000671, -0.00001391, -0.00002079, -0.00002448, -0.00002397, -0.00002083, -0.00001810, -0.00001788, -0.00001934, -0.00001870, -0.00001179, 0.00000218, 0.00001823, 0.00002672, 0.00001830, -0.00000989, -0.00005048, -0.00008794, -0.00010605, -0.00009715, -0.00006733, -0.00003359, -0.00001335, -0.00001236, -0.00001923, -0.00001212, 0.00002379, 0.00007726, 0.00010364, 0.00003836, -0.00016977, -0.00051905, -0.00093020, -0.00125467, -0.00132060, -0.00100261, -0.00028683, 0.00069999, 0.00171550, 0.00247698, 0.00275934, 0.00247698, 0.00171550, 0.00069999, -0.00028683, -0.00100261, -0.00132060, -0.00125467, -0.00093020, -0.00051905, -0.00016977, 0.00003836, 0.00010364, 0.00007726, 0.00002379, -0.00001212, -0.00001923, -0.00001236, -0.00001335, -0.00003359, -0.00006733, -0.00009715, -0.00010605, -0.00008794, -0.00005048, -0.00000989, 0.00001830, 0.00002672, 0.00001823, 0.00000218, -0.00001179, -0.00001870, -0.00001934, -0.00001788, -0.00001810, -0.00002083, -0.00002397, -0.00002448, -0.00002079, -0.00001391, -0.00000671, -0.00000205, -0.00000109, -0.00000288, -0.00000528, -0.00000642, -0.00000585, -0.00000453, -0.00000393, -0.00000487, -0.00000693, -0.00000872, -0.00000880, -0.00000665, -0.00000299, 0.00000061, 0.00000257, 0.00000218, -0.00000014, -0.00000311, -0.00000538, -0.00000622, -0.00000578, -0.00000491, },
		{ -0.00000579, -0.00000393, -0.00000113, 0.00000146, 0.00000254, 0.00000143, -0.00000155, -0.00000510, -0.00000764, -0.00000814, -0.00000670, -0.00000443, -0.00000284, -0.00000290, -0.00000452, -0.00000675, -0.00000848, -0.00000938, -0.00001012, -0.00001184, -0.00001509, -0.00001900, -0.00002149, -0.00002041, -0.00001518, -0.00000758, -0.00000112, 0.00000094, -0.00000228, -0.00000857, -0.00001404, -0.00001602, -0.00001549, -0.00001715, -0.00002638, -0.00004468, -0.00006657, -0.00008069, -0.00007543, -0.00004659, -0.00000222, 0.00003901, 0.00005635, 0.00003803, -0.00001172, -0.00007520, -0.00013318, -0.00017989, -0.00023074, -0.00031519, -0.00045467, -0.00063566, -0.00079421, -0.00082569, -0.00062281, -0.00012949, 0.00061339, 0.00145557, 0.00217129, 0.00253257, 0.00239480, 0.00175909, 0.00078277, -0.00026967, -0.00111732, -0.00155907, -0.00153713, -0.00114375, -0.00057206, -0.00003508, 0.00031176, 0.00041434, 0.00031720, 0.00012313, -0.00006131, -0.00016897, -0.00018763, -0.00014766, -0.00009384, -0.00005871, -0.00004986, -0.00005412, -0.00005222, -0.00003330, -0.00000117, 0.00002970, 0.00004421, 0.00003587, 0.00000982, -0.00002095, -0.00004329, -0.00005040, -0.00004387, -0.00003090, -0.00001915, -0.00001263, -0.00001067, -0.00000994, -0.00000758, -0.00000324, 0.00000099, 0.00000247, 0.00000003, -0.00000508, -0.00001009, -0.00001242, -0.00001126, -0.00000786, -0.00000459, -0.00000334, -0.00000444, -0.00000659, -0.00000786, -0.00000694, -0.00000392, -0.00000018, 0.00000247, 0.00000283, 0.00000092, -0.00000217, -0.00000495, -0.00000638, -0.00000634, -0.00000550, -0.00000484, -0.00000494, -0.00000566, -0.00000624, },
		{ 0.00000317, 0.00000261, 0.00000003, -0.00000360, -0.00000684, -0.00000851, -0.00000833, -0.00000703, -0.00000592, -0.00000613, -0.00000788, -0.00001045, -0.00001264, -0.00001356, -0.00001318, -0.00001222, -0.00001146, -0.00001102, -0.00001016, -0.00000795, -0.00000429, -0.00000071, 0.00000008, -0.00000435, -0.00001434, -0.00002726, -0.00003840, -0.00004344, -0.00004108, -0.00003396, -0.00002705, -0.00002421, -0.00002510, -0.00002496, -0.00001797, -0.00000275, 0.00001372, 0.00001693, -0.00000855, -0.00006987, -0.00015918, -0.00025471, -0.00033016, -0.00036823, -0.00036992, -0.00035261, -0.00033571, -0.00032060, -0.00027743, -0.00014997, 0.00011832, 0.00054618, 0.00108320, 0.00160482, 0.00194344, 0.00194866, 0.00155322, 0.00081491, -0.00008806, -0.00091152, -0.00143034, -0.00151991, -0.00119982, -0.00062153, -0.00000573, 0.00044317, 0.00060936, 0.00049633, 0.00020729, -0.00011088, -0.00033233, -0.00039726, -0.00032135, -0.00017184, -0.00002685, 0.00006067, 0.00007896, 0.00005068, 0.00001156, -0.00001150, -0.00001181, -0.00000038, 0.00000591, -0.00000323, -0.00002568, -0.00004965, -0.00006206, -0.00005654, -0.00003639, -0.00001176, 0.00000673, 0.00001349, 0.00000978, 0.00000140, -0.00000557, -0.00000822, -0.00000755, -0.00000665, -0.00000786, -0.00001113, -0.00001424, -0.00001460, -0.00001123, -0.00000552, -0.00000032, 0.00000177, 0.00000004, -0.00000405, -0.00000783, -0.00000900, -0.00000692, -0.00000281, 0.00000111, 0.00000290, 0.00000193, -0.00000102, -0.00000433, -0.00000656, -0.00000712, -0.00000644, -0.00000546, -0.00000500, -0.00000521, -0.00000554, -0.00000516, -0.00000357, -0.00000097, 0.00000169, },
		{ -0.00000843, -0.00001090, -0.00001147, -0.00001048, -0.00000895, -0.00000801, -0.00000818, -0.00000913, -0.00000995, -0.00000977, -0.00000842, -0.00000653, -0.00000513, -0.00000498, -0.00000607, -0.00000784, -0.00000978, -0.00001215, -0.00001598, -0.00002222, -0.00003052, -0.00003848, -0.00004229, -0.00003863, -0.00002699, -0.00001084, 0.00000343, 0.00000947, 0.00000418, -0.00001081, -0.00003078, -0.00005187, -0.00007468, -0.00010484, -0.00014899, -0.00020821, -0.00027273, -0.00032236, -0.00033397, -0.00029296, -0.00020236, -0.00008315, 0.00003645, 0.00013795, 0.00022639, 0.00033085, 0.00048801, 0.00071458, 0.00098284, 0.00121496, 0.00130473, 0.00116159, 0.00075881, 0.00016193, -0.00048004, -0.00098238, -0.00119384, -0.00105831, -0.00064055, -0.00010308, 0.00035745, 0.00058715, 0.00053207, 0.00025177, -0.00011456, -0.00041189, -0.00053350, -0.00045789, -0.00024532, -0.00000039, 0.00017865, 0.00023954, 0.00018887, 0.00007679, -0.00003446, -0.00010060, -0.00011032, -0.00008126, -0.00004312, -0.00001913, -0.00001578, -0.00002427, -0.00003008, -0.00002373, -0.00000612, 0.00001347, 0.00002431, 0.00002100, 0.00000619, -0.00001181, -0.00002436, -0.00002717, -0.00002178, -0.00001342, -0.00000726, -0.00000538, -0.00000636, -0.00000710, -0.00000538, -0.00000146, 0.00000223, 0.00000300, -0.00000021, -0.00000601, -0.00001137, -0.00001338, -0.00001097, -0.00000538, 0.00000063, 0.00000431, 0.00000433, 0.00000126, -0.00000301, -0.00000643, -0.00000782, -0.00000724, -0.00000569, -0.00000431, -0.00000371, -0.00000368, -0.00000353, -0.00000261, -0.00000084, 0.00000114, 0.00000227, 0.00000166, -0.00000088, -0.00000468, },
		{ -0.00000743, -0.00000547, -0.00000459, -0.00000490, -0.00000588, -0.00000679, -0.00000725, -0.00000755, -0.00000840, -0.00001044, -0.00001364, -0.00001729, -0.00002039, -0.00002233, -0.00002324, -0.00002375, -0.00002419, -0.00002390, -0.00002144, -0.00001563, -0.00000726, 0.00000009, 0.00000074, -0.00001064, -0.00003609, -0.00007260, -0.00011299, -0.00014920, -0.00017608, -0.00019333, -0.00020382, -0.00020912, -0.00020507, -0.00018101, -0.00012454, -0.00002992, 0.00009457, 0.00022580, 0.00033414, 0.00039787, 0.00041644, 0.00041367, 0.00042671, 0.00048411, 0.00058380, 0.00068413, 0.00071636, 0.00061594, 0.00035939, -0.00001240, -0.00039912, -0.00067607, -0.00074622, -0.00058557, -0.00025934, 0.00010093, 0.00035120, 0.00039595, 0.00022855, -0.00006779, -0.00035872, -0.00051899, -0.00048612, -0.00028358, -0.00000450, 0.00023406, 0.00034356, 0.00029865, 0.00014034, -0.00005043, -0.00019317, -0.00024199, -0.00019883, -0.00010302, -0.00000604, 0.00005401, 0.00006612, 0.00004410, 0.00001280, -0.00000762, -0.00001084, -0.00000372, 0.00000176, -0.00000243, -0.00001562, -0.00003007, -0.00003688, -0.00003179, -0.00001749, -0.00000140, 0.00000906, 0.00001055, 0.00000500, -0.00000253, -0.00000728, -0.00000762, -0.00000528, -0.00000360, -0.00000488, -0.00000887, -0.00001301, -0.00001424, -0.00001104, -0.00000440, 0.00000277, 0.00000728, 0.00000738, 0.00000354, -0.00000201, -0.00000664, -0.00000864, -0.00000786, -0.00000550, -0.00000318, -0.00000202, -0.00000216, -0.00000292, -0.00000337, -0.00000299, -0.00000197, -0.00000112, -0.00000131, -0.00000297, -0.00000579, -0.00000876, -0.00001075, -0.00001103, -0.00000965, },
		{ -0.00000835, -0.00001080, -0.00001283, -0.00001421, -0.00001520, -0.00001620, -0.00001727, -0.00001803, -0.00001779, -0.00001619, -0.00001359, -0.00001113, -0.00001017, -0.00001150, -0.00001509, -0.00002045, -0.00002761, -0.00003789, -0.00005367, -0.00007685, -0.00010678, -0.00013888, -0.00016528, -0.00017759, -0.00017053, -0.00014435, -0.00010425, -0.00005692, -0.00000618, 0.00004924, 0.00011374, 0.00018928, 0.00026950, 0.00033817, 0.00037461, 0.00036467, 0.00031173, 0.00024040, 0.00018864, 0.00018978, 0.00025263, 0.00035085, 0.00042956, 0.00042879, 0.00031412, 0.00009909, -0.00015377, -0.00035656, -0.00043505, -0.00036475, -0.00018554, 0.00001423, 0.00013795, 0.00012658, -0.00001243, -0.00020798, -0.00035932, -0.00038477, -0.00026160, -0.00003731, 0.00019279, 0.00033050, 0.00032002, 0.00017196, -0.00004451, -0.00023577, -0.00032676, -0.00029195, -0.00016100, 0.00000087, 0.00012603, 0.00017387, 0.00014358, 0.00006672, -0.00001353, -0.00006426, -0.00007509, -0.00005691, -0.00003043, -0.00001274, -0.00000925, -0.00001416, -0.00001733, -0.00001212, 0.00000051, 0.00001353, 0.00001912, 0.00001385, 0.00000056, -0.00001372, -0.00002220, -0.00002205, -0.00001545, -0.00000752, -0.00000289, -0.00000311, -0.00000622, -0.00000859, -0.00000744, -0.00000259, 0.00000358, 0.00000780, 0.00000785, 0.00000381, -0.00000218, -0.00000724, -0.00000938, -0.00000833, -0.00000546, -0.00000276, -0.00000170, -0.00000255, -0.00000444, -0.00000604, -0.00000636, -0.00000525, -0.00000343, -0.00000197, -0.00000172, -0.00000287, -0.00000483, -0.00000664, -0.00000748, -0.00000710, -0.00000595, -0.00000493, -0.00000489, -0.00000614, },
		{ -0.00001328, -0.00001182, -0.00001051, -0.00000962, -0.00000917, -0.00000924, -0.00001037, -0.00001356, -0.00001999, -0.00003026, -0.00004393, -0.00005955, -0.00007527, -0.00008980, -0.00010265, -0.00011368, -0.00012181, -0.00012406, -0.00011580, -0.00009257, -0.00005280, 0.00000019, 0.00005853, 0.00011272, 0.00015583, 0.00018645, 0.00020837, 0.00022656, 0.00024200, 0.00024880, 0.00023654, 0.00019778, 0.00013661, 0.00007270, 0.00003639, 0.00005498, 0.00013631, 0.00025838, 0.00037270, 0.00042241, 0.00036823, 0.00020992, -0.00000879, -0.00021604, -0.00034272, -0.00035381, -0.00026459, -0.00013263, -0.00002889, -0.00000208, -0.00005464, -0.00014279, -0.00020120, -0.00017935, -0.00007053, 0.00008208, 0.00020602, 0.00023638, 0.00014995, -0.00002107, -0.00020304, -0.00031555, -0.00030927, -0.00018829, -0.00000663, 0.00015856, 0.00024283, 0.00022178, 0.00011770, -0.00001569, -0.00012092, -0.00016235, -0.00013792, -0.00007356, -0.00000579, 0.00003720, 0.00004642, 0.00003120, 0.00000934, -0.00000458, -0.00000625, -0.00000112, 0.00000173, -0.00000339, -0.00001521, -0.00002700, -0.00003148, -0.00002557, -0.00001217, 0.00000191, 0.00001011, 0.00000978, 0.00000312, -0.00000466, -0.00000870, -0.00000714, -0.00000172, 0.00000375, 0.00000591, 0.00000370, -0.00000138, -0.00000646, -0.00000911, -0.00000864, -0.00000617, -0.00000371, -0.00000284, -0.00000389, -0.00000591, -0.00000735, -0.00000705, -0.00000482, -0.00000155, 0.00000133, 0.00000253, 0.00000156, -0.00000120, -0.00000469, -0.00000775, -0.00000964, -0.00001032, -0.00001035, -0.00001049, -0.00001122, -0.00001252, -0.00001386, -0.00001459, -0.00001436, },
		{ -0.00001574, -0.00002189, -0.00002970, -0.00003923, -0.00005051, -0.00006318, -0.00007604, -0.00008702, -0.00009368, -0.00009410, -0.00008751, -0.00007438, -0.00005563, -0.00003173, -0.00000231, 0.00003311, 0.00007353, 0.00011518, 0.00015152, 0.00017543, 0.00018248, 0.00017359, 0.00015512, 0.00013576, 0.00012162, 0.00011255, 0.00010249, 0.00008443, 0.00005755, 0.00003199, 0.00002703, 0.00006179, 0.00014232, 0.00025210, 0.00035272, 0.00039699, 0.00034989, 0.00020764, 0.00000445, -0.00019813, -0.00033600, -0.00037013, -0.00030355, -0.00017882, -0.00005670, 0.00001332, 0.00001568, -0.00002704, -0.00007025, -0.00007466, -0.00002970, 0.00004007, 0.00008918, 0.00007837, -0.00000065, -0.00011694, -0.00021382, -0.00023757, -0.00016628, -0.00002329, 0.00013218, 0.00023278, 0.00023509, 0.00013987, -0.00000929, -0.00014760, -0.00021950, -0.00020275, -0.00011510, -0.00000186, 0.00008766, 0.00012255, 0.00010085, 0.00004501, -0.00001319, -0.00004931, -0.00005582, -0.00004131, -0.00002174, -0.00000993, -0.00000930, -0.00001446, -0.00001682, -0.00001118, 0.00000111, 0.00001348, 0.00001884, 0.00001409, 0.00000190, -0.00001112, -0.00001843, -0.00001710, -0.00000898, 0.00000085, 0.00000722, 0.00000759, 0.00000289, -0.00000359, -0.00000839, -0.00000974, -0.00000812, -0.00000545, -0.00000364, -0.00000350, -0.00000441, -0.00000502, -0.00000419, -0.00000180, 0.00000124, 0.00000342, 0.00000344, 0.00000088, -0.00000367, -0.00000885, -0.00001316, -0.00001554, -0.00001573, -0.00001428, -0.00001216, -0.00001030, -0.00000916, -0.00000868, -0.00000845, -0.00000813, -0.00000774, -0.00000772, -0.00000875, -0.00001135, },
		{ -0.00006494, -0.00006865, -0.00006933, -0.00006572, -0.00005650, -0.00004095, -0.00001947, 0.00000615, 0.00003329, 0.00005946, 0.00008308, 0.00010354, 0.00012044, 0.00013255, 0.00013738, 0.00013225, 0.00011621, 0.00009204, 0.00006639, 0.00004748, 0.00004116, 0.00004735, 0.00005964, 0.00006881, 0.00006909, 0.00006339, 0.00006381, 0.00008564, 0.00013714, 0.00021050, 0.00027992, 0.00030995, 0.00027141, 0.00015781, -0.00000681, -0.00017409, -0.00029022, -0.00031925, -0.00025946, -0.00014417, -0.00002599, 0.00004783, 0.00005660, 0.00001327, -0.00004596, -0.00008330, -0.00008022, -0.00004608, -0.00001002, -0.00000135, -0.00003005, -0.00007912, -0.00011361, -0.00010147, -0.00003427, 0.00006439, 0.00014850, 0.00017290, 0.00011774, 0.00000102, -0.00012818, -0.00021261, -0.00021437, -0.00013303, -0.00000526, 0.00011334, 0.00017451, 0.00015869, 0.00008123, -0.00001817, -0.00009616, -0.00012552, -0.00010477, -0.00005412, -0.00000204, 0.00002971, 0.00003473, 0.00002124, 0.00000400, -0.00000525, -0.00000364, 0.00000342, 0.00000755, 0.00000351, -0.00000755, -0.00001936, -0.00002485, -0.00002053, -0.00000846, 0.00000526, 0.00001412, 0.00001466, 0.00000788, -0.00000192, -0.00000981, -0.00001279, -0.00001089, -0.00000645, -0.00000235, -0.00000037, -0.00000055, -0.00000167, -0.00000233, -0.00000192, -0.00000091, -0.00000044, -0.00000154, -0.00000448, -0.00000861, -0.00001260, -0.00001512, -0.00001533, -0.00001328, -0.00000982, -0.00000624, -0.00000379, -0.00000317, -0.00000443, -0.00000709, -0.00001050, -0.00001425, -0.00001837, -0.00002321, -0.00002913, -0.00003621, -0.00004404, -0.00005193, -0.00005913, },
		{ 0.00000350, 0.00002221, 0.00004235, 0.00006206, 0.00007899, 0.00009102, 0.00009708, 0.00009748, 0.00009351, 0.00008652, 0.00007713, 0.00006523, 0.00005104, 0.00003636, 0.00002509, 0.00002203, 0.00003017, 0.00004807, 0.00006919, 0.00008439, 0.00008700, 0.00007772, 0.00006613, 0.00006683, 0.00009114, 0.00013846, 0.00019255, 0.00022606, 0.00021250, 0.00014046, 0.00002247, -0.00010708, -0.00020529, -0.00023975, -0.00020376, -0.00011987, -0.00002928, 0.00002786, 0.00003131, -0.00001094, -0.00006898, -0.00010822, -0.00010806, -0.00007166, -0.00002196, 0.00001327, 0.00001833, -0.00000169, -0.00002523, -0.00002896, -0.00000385, 0.00003754, 0.00006732, 0.00005896, 0.00000487, -0.00007584, -0.00014471, -0.00016328, -0.00011426, -0.00001272, 0.00009914, 0.00017161, 0.00017152, 0.00009842, -0.00001537, -0.00012057, -0.00017425, -0.00015893, -0.00008838, 0.00000170, 0.00007222, 0.00009867, 0.00007982, 0.00003423, -0.00001208, -0.00003924, -0.00004158, -0.00002700, -0.00000954, -0.00000036, -0.00000225, -0.00000998, -0.00001519, -0.00001234, -0.00000186, 0.00001062, 0.00001819, 0.00001687, 0.00000763, -0.00000452, -0.00001370, -0.00001628, -0.00001245, -0.00000542, 0.00000071, 0.00000322, 0.00000180, -0.00000185, -0.00000539, -0.00000724, -0.00000721, -0.00000621, -0.00000548, -0.00000574, -0.00000687, -0.00000814, -0.00000870, -0.00000809, -0.00000657, -0.00000493, -0.00000424, -0.00000536, -0.00000862, -0.00001368, -0.00001970, -0.00002569, -0.00003091, -0.00003513, -0.00003859, -0.00004168, -0.00004456, -0.00004692, -0.00004804, -0.00004702, -0.00004318, -0.00003618, -0.00002600, -0.00001272, },
		{ 0.00007347, 0.00007288, 0.00006808, 0.00006007, 0.00005070, 0.00004196, 0.00003512, 0.00003040, 0.00002731, 0.00002567, 0.00002640, 0.00003129, 0.00004153, 0.00005590, 0.00006989, 0.00007711, 0.00007285, 0.00005794, 0.00004056, 0.00003378, 0.00004912, 0.00008895, 0.00014205, 0.00018568, 0.00019467, 0.00015361, 0.00006633, -0.00004313, -0.00013945, -0.00019137, -0.00018598, -0.00013467, -0.00006720, -0.00001677, -0.00000372, -0.00002631, -0.00006347, -0.00008763, -0.00008051, -0.00004297, 0.00000632, 0.00004233, 0.00004823, 0.00002468, -0.00001156, -0.00003853, -0.00004269, -0.00002684, -0.00000756, -0.00000363, -0.00002269, -0.00005464, -0.00007639, -0.00006574, -0.00001645, 0.00005482, 0.00011462, 0.00012942, 0.00008415, -0.00000765, -0.00010829, -0.00017321, -0.00017253, -0.00010570, -0.00000177, 0.00009463, 0.00014440, 0.00013141, 0.00006797, -0.00001343, -0.00007701, -0.00010017, -0.00008172, -0.00003873, 0.00000461, 0.00002974, 0.00003128, 0.00001667, -0.00000063, -0.00000963, -0.00000716, 0.00000214, 0.00001021, 0.00001102, 0.00000396, -0.00000662, -0.00001463, -0.00001606, -0.00001100, -0.00000310, 0.00000294, 0.00000409, 0.00000035, -0.00000568, -0.00001062, -0.00001218, -0.00001007, -0.00000579, -0.00000154, 0.00000097, 0.00000113, -0.00000066, -0.00000349, -0.00000656, -0.00000955, -0.00001257, -0.00001593, -0.00001988, -0.00002431, -0.00002878, -0.00003258, -0.00003506, -0.00003582, -0.00003493, -0.00003286, -0.00003023, -0.00002743, -0.00002431, -0.00002026, -0.00001446, -0.00000634, 0.00000410, 0.00001622, 0.00002905, 0.00004162, 0.00005310, 0.00006278, 0.00006987, },
		{ 0.00002631, 0.00002278, 0.00002185, 0.00002328, 0.00002606, 0.00002909, 0.00003194, 0.00003521, 0.00003987, 0.00004593, 0.00005150, 0.00005315, 0.00004797, 0.00003652, 0.00002448, 0.00002158, 0.00003706, 0.00007364, 0.00012317, 0.00016721, 0.00018348, 0.00015603, 0.00008420, -0.00001453, -0.00011023, -0.00017281, -0.00018524, -0.00015095, -0.00009163, -0.00003618, -0.00000654, -0.00000784, -0.00002783, -0.00004567, -0.00004475, -0.00002236, 0.00000932, 0.00003117, 0.00002899, 0.00000238, -0.00003497, -0.00006327, -0.00006827, -0.00004938, -0.00001929, 0.00000435, 0.00001021, 0.00000017, -0.00001239, -0.00001211, 0.00000705, 0.00003602, 0.00005464, 0.00004366, -0.00000203, -0.00006721, -0.00012164, -0.00013493, -0.00009332, -0.00000892, 0.00008408, 0.00014497, 0.00014608, 0.00008639, -0.00000795, -0.00009618, -0.00014218, -0.00013066, -0.00007244, 0.00000283, 0.00006211, 0.00008422, 0.00006763, 0.00002798, -0.00001221, -0.00003527, -0.00003569, -0.00002010, -0.00000119, 0.00001011, 0.00000987, 0.00000156, -0.00000765, -0.00001185, -0.00000954, -0.00000362, 0.00000119, 0.00000160, -0.00000243, -0.00000809, -0.00001168, -0.00001093, -0.00000610, 0.00000034, 0.00000525, 0.00000634, 0.00000308, -0.00000343, -0.00001126, -0.00001859, -0.00002428, -0.00002795, -0.00002980, -0.00003022, -0.00002954, -0.00002798, -0.00002561, -0.00002249, -0.00001874, -0.00001462, -0.00001049, -0.00000670, -0.00000332, -0.00000003, 0.00000387, 0.00000919, 0.00001639, 0.00002523, 0.00003469, 0.00004335, 0.00004990, 0.00005359, 0.00005439, 0.00005272, 0.00004913, 0.00004409, 0.00003806, 0.00003178, },
		{ 0.00002468, 0.00002686, 0.00002801, 0.00002854, 0.00002946, 0.00003155, 0.00003441, 0.00003628, 0.00003507, 0.00003038, 0.00002519, 0.00002572, 0.00003875, 0.00006714, 0.00010590, 0.00014164, 0.00015667, 0.00013672, 0.00007889, -0.00000467, -0.00008986, -0.00015001, -0.00016735, -0.00014096, -0.00008696, -0.00003061, 0.00000572, 0.00001260, -0.00000357, -0.00002606, -0.00003826, -0.00003313, -0.00001622, -0.00000104, 0.00000026, -0.00001502, -0.00003817, -0.00005438, -0.00005198, -0.00002989, 0.00000141, 0.00002603, 0.00003214, 0.00001889, -0.00000348, -0.00002052, -0.00002296, -0.00001275, -0.00000160, -0.00000275, -0.00002111, -0.00004813, -0.00006536, -0.00005522, -0.00001294, 0.00004767, 0.00009894, 0.00011283, 0.00007622, -0.00000036, -0.00008585, -0.00014278, -0.00014520, -0.00009138, -0.00000471, 0.00007754, 0.00012182, 0.00011336, 0.00006138, -0.00000722, -0.00006189, -0.00008261, -0.00006730, -0.00002990, 0.00000878, 0.00003191, 0.00003362, 0.00001925, 0.00000023, -0.00001308, -0.00001626, -0.00001149, -0.00000456, -0.00000062, -0.00000132, -0.00000466, -0.00000713, -0.00000635, -0.00000258, 0.00000164, 0.00000313, -0.00000001, -0.00000746, -0.00001698, -0.00002562, -0.00003105, -0.00003231, -0.00002980, -0.00002477, -0.00001866, -0.00001265, -0.00000750, -0.00000352, -0.00000073, 0.00000109, 0.00000229, 0.00000324, 0.00000429, 0.00000581, 0.00000819, 0.00001186, 0.00001715, 0.00002401, 0.00003176, 0.00003909, 0.00004440, 0.00004641, 0.00004470, 0.00003995, 0.00003361, 0.00002729, 0.00002218, 0.00001873, 0.00001688, 0.00001642, 0.00001721, 0.00001912, 0.00002184, },
		{ 0.00002206, 0.00002372, 0.00002653, 0.00002944, 0.00003081, 0.00002972, 0.00002740, 0.00002774, 0.00003582, 0.00005482, 0.00008269, 0.00011070, 0.00012559, 0.00011505, 0.00007429, 0.00001022, -0.00005957, -0.00011327, -0.00013421, -0.00011845, -0.00007680, -0.00002960, 0.00000291, 0.00000995, -0.00000553, -0.00002994, -0.00004736, -0.00004835, -0.00003440, -0.00001578, -0.00000429, -0.00000579, -0.00001678, -0.00002718, -0.00002733, -0.00001461, 0.00000452, 0.00001838, 0.00001753, 0.00000101, -0.00002262, -0.00004027, -0.00004224, -0.00002822, -0.00000730, 0.00000787, 0.00000939, -0.00000059, -0.00001122, -0.00001031, 0.00000673, 0.00003213, 0.00004893, 0.00004076, 0.00000281, -0.00005265, -0.00010038, -0.00011434, -0.00008186, -0.00001189, 0.00006753, 0.00012193, 0.00012687, 0.00007985, 0.00000160, -0.00007400, -0.00011588, -0.00010972, -0.00006313, -0.00000020, 0.00005114, 0.00007205, 0.00005986, 0.00002660, -0.00000907, -0.00003166, -0.00003525, -0.00002405, -0.00000791, 0.00000391, 0.00000713, 0.00000323, -0.00000313, -0.00000766, -0.00000894, -0.00000847, -0.00000898, -0.00001222, -0.00001774, -0.00002320, -0.00002587, -0.00002410, -0.00001813, -0.00000978, -0.00000146, 0.00000483, 0.00000814, 0.00000854, 0.00000684, 0.00000414, 0.00000154, -0.00000002, 0.00000008, 0.00000213, 0.00000604, 0.00001145, 0.00001788, 0.00002476, 0.00003147, 0.00003718, 0.00004086, 0.00004145, 0.00003833, 0.00003177, 0.00002310, 0.00001444, 0.00000792, 0.00000488, 0.00000540, 0.00000845, 0.00001249, 0.00001623, 0.00001901, 0.00002076, 0.00002168, 0.00002197, 0.00002182, 0.00002163, },
		{ 0.00002658, 0.00002644, 0.00002509, 0.00002502, 0.00003003, 0.00004320, 0.00006425, 0.00008776, 0.00010391, 0.00010209, 0.00007611, 0.00002847, -0.00002902, -0.00007913, -0.00010644, -0.00010432, -0.00007808, -0.00004243, -0.00001429, -0.00000463, -0.00001363, -0.00003166, -0.00004536, -0.00004535, -0.00003118, -0.00001086, 0.00000447, 0.00000740, -0.00000161, -0.00001509, -0.00002374, -0.00002251, -0.00001359, -0.00000463, -0.00000325, -0.00001155, -0.00002429, -0.00003199, -0.00002713, -0.00000946, 0.00001318, 0.00002918, 0.00003019, 0.00001630, -0.00000387, -0.00001857, -0.00002038, -0.00001113, -0.00000078, -0.00000073, -0.00001554, -0.00003845, -0.00005406, -0.00004713, -0.00001281, 0.00003819, 0.00008298, 0.00009752, 0.00006964, 0.00000681, -0.00006581, -0.00011667, -0.00012295, -0.00008158, -0.00001069, 0.00005917, 0.00009956, 0.00009685, 0.00005720, 0.00000178, -0.00004443, -0.00006436, -0.00005531, -0.00002774, 0.00000191, 0.00001992, 0.00002106, 0.00000907, -0.00000725, -0.00001960, -0.00002403, -0.00002160, -0.00001624, -0.00001160, -0.00000903, -0.00000748, -0.00000498, -0.00000044, 0.00000548, 0.00001073, 0.00001313, 0.00001162, 0.00000673, 0.00000025, -0.00000558, -0.00000898, -0.00000901, -0.00000560, 0.00000060, 0.00000856, 0.00001707, 0.00002491, 0.00003109, 0.00003503, 0.00003653, 0.00003564, 0.00003250, 0.00002721, 0.00002000, 0.00001155, 0.00000319, -0.00000328, -0.00000612, -0.00000452, 0.00000097, 0.00000849, 0.00001562, 0.00002040, 0.00002204, 0.00002108, 0.00001884, 0.00001667, 0.00001546, 0.00001551, 0.00001678, 0.00001908, 0.00002204, 0.00002488, },
		{ 0.00002708, 0.00003720, 0.00005368, 0.00007317, 0.00008860, 0.00009123, 0.00007459, 0.00003833, -0.00001005, -0.00005724, -0.00008924, -0.00009751, -0.00008288, -0.00005510, -0.00002800, -0.00001270, -0.00001256, -0.00002230, -0.00003194, -0.00003297, -0.00002331, -0.00000813, 0.00000366, 0.00000511, -0.00000452, -0.00001940, -0.00003083, -0.00003260, -0.00002458, -0.00001235, -0.00000324, -0.00000142, -0.00000527, -0.00000877, -0.00000603, 0.00000419, 0.00001688, 0.00002363, 0.00001802, 0.00000031, -0.00002193, -0.00003776, -0.00003926, -0.00002640, -0.00000708, 0.00000763, 0.00001053, 0.00000283, -0.00000657, -0.00000718, 0.00000555, 0.00002590, 0.00004010, 0.00003423, 0.00000358, -0.00004246, -0.00008325, -0.00009682, -0.00007173, -0.00001434, 0.00005282, 0.00010111, 0.00010951, 0.00007470, 0.00001255, -0.00004979, -0.00008685, -0.00008637, -0.00005375, -0.00000775, 0.00002963, 0.00004360, 0.00003229, 0.00000555, -0.00002164, -0.00003699, -0.00003599, -0.00002233, -0.00000420, 0.00001062, 0.00001819, 0.00001899, 0.00001604, 0.00001228, 0.00000885, 0.00000516, 0.00000013, -0.00000637, -0.00001299, -0.00001740, -0.00001753, -0.00001254, -0.00000323, 0.00000840, 0.00001995, 0.00002933, 0.00003517, 0.00003696, 0.00003490, 0.00002983, 0.00002296, 0.00001561, 0.00000884, 0.00000319, -0.00000128, -0.00000465, -0.00000678, -0.00000708, -0.00000482, 0.00000032, 0.00000764, 0.00001538, 0.00002125, 0.00002348, 0.00002164, 0.00001692, 0.00001155, 0.00000776, 0.00000679, 0.00000861, 0.00001224, 0.00001638, 0.00002001, 0.00002252, 0.00002364, 0.00002344, 0.00002266, 0.00002302, },
		{ 0.00007375, 0.00007765, 0.00006685, 0.00003959, 0.00000014, -0.00004171, -0.00007404, -0.00008779, -0.00008085, -0.00005905, -0.00003339, -0.00001469, -0.00000856, -0.00001333, -0.00002193, -0.00002658, -0.00002345, -0.00001457, -0.00000606, -0.00000383, -0.00000961, -0.00001974, -0.00002743, -0.00002707, -0.00001781, -0.00000406, 0.00000728, 0.00001111, 0.00000717, -0.00000016, -0.00000507, -0.00000446, -0.00000005, 0.00000281, -0.00000085, -0.00001165, -0.00002461, -0.00003171, -0.00002686, -0.00001023, 0.00001111, 0.00002681, 0.00002916, 0.00001767, -0.00000062, -0.00001549, -0.00001989, -0.00001435, -0.00000647, -0.00000563, -0.00001607, -0.00003291, -0.00004396, -0.00003684, -0.00000735, 0.00003596, 0.00007438, 0.00008793, 0.00006613, 0.00001462, -0.00004631, -0.00009088, -0.00010019, -0.00007132, -0.00001854, 0.00003389, 0.00006356, 0.00006020, 0.00002945, -0.00001126, -0.00004204, -0.00005007, -0.00003451, -0.00000517, 0.00002368, 0.00004046, 0.00004094, 0.00002841, 0.00001033, -0.00000606, -0.00001682, -0.00002173, -0.00002266, -0.00002135, -0.00001818, -0.00001232, -0.00000307, 0.00000909, 0.00002207, 0.00003294, 0.00003909, 0.00003923, 0.00003372, 0.00002421, 0.00001299, 0.00000236, -0.00000585, -0.00001049, -0.00001131, -0.00000895, -0.00000475, -0.00000027, 0.00000334, 0.00000580, 0.00000765, 0.00000975, 0.00001249, 0.00001531, 0.00001691, 0.00001602, 0.00001231, 0.00000693, 0.00000206, -0.00000007, 0.00000167, 0.00000674, 0.00001329, 0.00001911, 0.00002266, 0.00002360, 0.00002258, 0.00002081, 0.00001959, 0.00002025, 0.00002421, 0.00003262, 0.00004556, 0.00006087, },
		{ 0.00000698, -0.00002804, -0.00005811, -0.00007480, -0.00007426, -0.00005915, -0.00003748, -0.00001880, -0.00000972, -0.00001108, -0.00001826, -0.00002446, -0.00002478, -0.00001888, -0.00001062, -0.00000516, -0.00000541, -0.00001011, -0.00001483, -0.00001499, -0.00000911, 0.00000011, 0.00000740, 0.00000830, 0.00000214, -0.00000748, -0.00001498, -0.00001632, -0.00001161, -0.00000482, -0.00000096, -0.00000251, -0.00000756, -0.00001097, -0.00000803, 0.00000185, 0.00001401, 0.00002085, 0.00001649, 0.00000079, -0.00001986, -0.00003577, -0.00003942, -0.00002975, -0.00001272, 0.00000252, 0.00000930, 0.00000760, 0.00000368, 0.00000548, 0.00001650, 0.00003213, 0.00004144, 0.00003345, 0.00000479, -0.00003636, -0.00007270, -0.00008603, -0.00006702, -0.00002110, 0.00003308, 0.00007218, 0.00007947, 0.00005299, 0.00000636, -0.00003811, -0.00006041, -0.00005208, -0.00001942, 0.00002066, 0.00004955, 0.00005558, 0.00003850, 0.00000797, -0.00002235, -0.00004140, -0.00004492, -0.00003552, -0.00001962, -0.00000344, 0.00000959, 0.00001911, 0.00002622, 0.00003163, 0.00003473, 0.00003397, 0.00002819, 0.00001769, 0.00000450, -0.00000835, -0.00001790, -0.00002226, -0.00002096, -0.00001484, -0.00000565, 0.00000435, 0.00001288, 0.00001804, 0.00001892, 0.00001593, 0.00001073, 0.00000558, 0.00000234, 0.00000162, 0.00000263, 0.00000378, 0.00000382, 0.00000269, 0.00000161, 0.00000234, 0.00000591, 0.00001185, 0.00001829, 0.00002287, 0.00002403, 0.00002183, 0.00001778, 0.00001411, 0.00001269, 0.00001457, 0.00001995, 0.00002855, 0.00003960, 0.00005151, 0.00006130, 0.00006475, 0.00005759, 0.00003769, },
		{ -0.00006699, -0.00005827, -0.00004137, -0.00002389, -0.00001247, -0.00000972, -0.00001344, -0.00001841, -0.00001969, -0.00001552, -0.00000803, -0.00000149, 0.00000058, -0.00000230, -0.00000743, -0.00001088, -0.00001026, -0.00000628, -0.00000222, -0.00000153, -0.00000532, -0.00001133, -0.00001534, -0.00001391, -0.00000686, 0.00000240, 0.00000887, 0.00000914, 0.00000358, -0.00000392, -0.00000856, -0.00000793, -0.00000374, -0.00000077, -0.00000347, -0.00001257, -0.00002384, -0.00003025, -0.00002612, -0.00001098, 0.00000957, 0.00002666, 0.00003307, 0.00002725, 0.00001396, 0.00000101, -0.00000582, -0.00000638, -0.00000594, -0.00001101, -0.00002391, -0.00003977, -0.00004822, -0.00003940, -0.00001077, 0.00002942, 0.00006467, 0.00007804, 0.00006120, 0.00001966, -0.00002892, -0.00006293, -0.00006732, -0.00004106, 0.00000249, 0.00004226, 0.00005990, 0.00004825, 0.00001411, -0.00002602, -0.00005438, -0.00006011, -0.00004320, -0.00001289, 0.00001791, 0.00003886, 0.00004592, 0.00004128, 0.00003047, 0.00001867, 0.00000854, 0.00000007, -0.00000778, -0.00001542, -0.00002182, -0.00002502, -0.00002334, -0.00001650, -0.00000581, 0.00000626, 0.00001702, 0.00002424, 0.00002659, 0.00002374, 0.00001643, 0.00000644, -0.00000362, -0.00001094, -0.00001349, -0.00001083, -0.00000444, 0.00000295, 0.00000864, 0.00001126, 0.00001132, 0.00001065, 0.00001111, 0.00001333, 0.00001636, 0.00001828, 0.00001754, 0.00001411, 0.00000961, 0.00000653, 0.00000691, 0.00001141, 0.00001923, 0.00002874, 0.00003827, 0.00004646, 0.00005194, 0.00005280, 0.00004674, 0.00003200, 0.00000897, -0.00001881, -0.00004499, -0.00006259, },
		{ -0.00001265, -0.00000644, -0.00000671, -0.00001019, -0.00001259, -0.00001130, -0.00000677, -0.00000194, -0.00000008, -0.00000251, -0.00000778, -0.00001262, -0.00001420, -0.00001188, -0.00000756, -0.00000418, -0.00000361, -0.00000532, -0.00000678, -0.00000545, -0.00000079, 0.00000494, 0.00000793, 0.00000530, -0.00000275, -0.00001278, -0.00001988, -0.00002072, -0.00001552, -0.00000785, -0.00000230, -0.00000133, -0.00000363, -0.00000498, -0.00000132, 0.00000812, 0.00001959, 0.00002663, 0.00002391, 0.00001070, -0.00000824, -0.00002501, -0.00003303, -0.00003056, -0.00002136, -0.00001179, -0.00000635, -0.00000465, -0.00000183, 0.00000754, 0.00002459, 0.00004364, 0.00005419, 0.00004670, 0.00001906, -0.00002037, -0.00005548, -0.00007001, -0.00005608, -0.00001906, 0.00002427, 0.00005367, 0.00005543, 0.00002917, -0.00001193, -0.00004784, -0.00006147, -0.00004674, -0.00001083, 0.00003007, 0.00005891, 0.00006546, 0.00004991, 0.00002120, -0.00000843, -0.00002926, -0.00003753, -0.00003526, -0.00002747, -0.00001875, -0.00001128, -0.00000482, 0.00000185, 0.00000927, 0.00001664, 0.00002222, 0.00002433, 0.00002219, 0.00001620, 0.00000761, -0.00000187, -0.00001043, -0.00001623, -0.00001774, -0.00001417, -0.00000609, 0.00000434, 0.00001393, 0.00001964, 0.00001999, 0.00001579, 0.00000975, 0.00000501, 0.00000356, 0.00000524, 0.00000813, 0.00000996, 0.00000961, 0.00000782, 0.00000673, 0.00000847, 0.00001379, 0.00002166, 0.00002988, 0.00003632, 0.00003983, 0.00004027, 0.00003777, 0.00003194, 0.00002176, 0.00000655, -0.00001275, -0.00003283, -0.00004876, -0.00005596, -0.00005255, -0.00004059, -0.00002530, },
		{ -0.00000880, -0.00001010, -0.00000855, -0.00000551, -0.00000350, -0.00000427, -0.00000754, -0.00001115, -0.00001255, -0.00001059, -0.00000626, -0.00000203, -0.00000015, -0.00000120, -0.00000377, -0.00000562, -0.00000543, -0.00000387, -0.00000313, -0.00000527, -0.00001041, -0.00001614, -0.00001873, -0.00001541, -0.00000634, 0.00000511, 0.00001415, 0.00001729, 0.00001432, 0.00000826, 0.00000323, 0.00000161, 0.00000248, 0.00000229, -0.00000245, -0.00001234, -0.00002393, -0.00003137, -0.00002974, -0.00001825, -0.00000100, 0.00001519, 0.00002461, 0.00002582, 0.00002200, 0.00001827, 0.00001770, 0.00001876, 0.00001615, 0.00000467, -0.00001613, -0.00003981, -0.00005528, -0.00005263, -0.00002939, 0.00000656, 0.00003978, 0.00005462, 0.00004348, 0.00001143, -0.00002574, -0.00004917, -0.00004645, -0.00001801, 0.00002295, 0.00005701, 0.00006801, 0.00005066, 0.00001249, -0.00003040, -0.00006127, -0.00007014, -0.00005710, -0.00003065, -0.00000244, 0.00001820, 0.00002756, 0.00002737, 0.00002217, 0.00001619, 0.00001155, 0.00000815, 0.00000491, 0.00000114, -0.00000299, -0.00000663, -0.00000896, -0.00000951, -0.00000816, -0.00000492, 0.00000013, 0.00000647, 0.00001282, 0.00001720, 0.00001769, 0.00001353, 0.00000596, -0.00000204, -0.00000700, -0.00000671, -0.00000144, 0.00000620, 0.00001271, 0.00001562, 0.00001480, 0.00001234, 0.00001110, 0.00001295, 0.00001765, 0.00002320, 0.00002717, 0.00002816, 0.00002639, 0.00002304, 0.00001904, 0.00001408, 0.00000682, -0.00000384, -0.00001733, -0.00003088, -0.00004043, -0.00004267, -0.00003696, -0.00002591, -0.00001418, -0.00000614, -0.00000370, -0.00000566, },
		{ -0.00000441, -0.00000370, -0.00000510, -0.00000769, -0.00000962, -0.00000947, -0.00000736, -0.00000487, -0.00000391, -0.00000541, -0.00000856, -0.00001137, -0.00001207, -0.00001027, -0.00000723, -0.00000480, -0.00000399, -0.00000403, -0.00000287, 0.00000131, 0.00000843, 0.00001598, 0.00002019, 0.00001816, 0.00000978, -0.00000196, -0.00001253, -0.00001834, -0.00001867, -0.00001571, -0.00001278, -0.00001182, -0.00001199, -0.00001026, -0.00000377, 0.00000769, 0.00002075, 0.00002994, 0.00003082, 0.00002270, 0.00000920, -0.00000378, -0.00001163, -0.00001371, -0.00001334, -0.00001507, -0.00002091, -0.00002813, -0.00003040, -0.00002188, -0.00000181, 0.00002337, 0.00004239, 0.00004495, 0.00002794, -0.00000169, -0.00002958, -0.00004112, -0.00002937, 0.00000057, 0.00003353, 0.00005165, 0.00004358, 0.00001053, -0.00003400, -0.00007068, -0.00008360, -0.00006755, -0.00002986, 0.00001374, 0.00004683, 0.00005937, 0.00005096, 0.00002924, 0.00000503, -0.00001286, -0.00002064, -0.00001951, -0.00001335, -0.00000604, 0.00000012, 0.00000451, 0.00000730, 0.00000867, 0.00000864, 0.00000738, 0.00000552, 0.00000391, 0.00000309, 0.00000270, 0.00000175, -0.00000063, -0.00000417, -0.00000720, -0.00000735, -0.00000304, 0.00000518, 0.00001450, 0.00002110, 0.00002219, 0.00001760, 0.00001002, 0.00000355, 0.00000145, 0.00000448, 0.00001069, 0.00001680, 0.00002020, 0.00002028, 0.00001831, 0.00001607, 0.00001422, 0.00001169, 0.00000642, -0.00000289, -0.00001531, -0.00002756, -0.00003545, -0.00003616, -0.00002980, -0.00001949, -0.00000968, -0.00000385, -0.00000290, -0.00000508, -0.00000749, -0.00000801, -0.00000647, },
		{ -0.00001134, -0.00001285, -0.00001255, -0.00001099, -0.00000946, -0.00000901, -0.00000951, -0.00000973, -0.00000818, -0.00000427, 0.00000113, 0.00000616, 0.00000918, 0.00000981, 0.00000907, 0.00000851, 0.00000885, 0.00000923, 0.00000764, 0.00000236, -0.00000649, -0.00001640, -0.00002358, -0.00002486, -0.00001962, -0.00001009, -0.00000013, 0.00000698, 0.00001031, 0.00001131, 0.00001233, 0.00001451, 0.00001663, 0.00001577, 0.00000949, -0.00000204, -0.00001535, -0.00002523, -0.00002768, -0.00002231, -0.00001269, -0.00000418, -0.00000049, -0.00000133, -0.00000267, 0.00000032, 0.00000945, 0.00002143, 0.00002918, 0.00002599, 0.00001033, -0.00001212, -0.00003059, -0.00003497, -0.00002193, 0.00000221, 0.00002405, 0.00002996, 0.00001374, -0.00001917, -0.00005366, -0.00007220, -0.00006373, -0.00002951, 0.00001714, 0.00005717, 0.00007468, 0.00006409, 0.00003189, -0.00000727, -0.00003779, -0.00004988, -0.00004265, -0.00002266, 0.00000041, 0.00001849, 0.00002771, 0.00002839, 0.00002318, 0.00001497, 0.00000580, -0.00000298, -0.00001016, -0.00001444, -0.00001490, -0.00001174, -0.00000646, -0.00000122, 0.00000235, 0.00000415, 0.00000552, 0.00000806, 0.00001221, 0.00001650, 0.00001820, 0.00001508, 0.00000716, -0.00000271, -0.00001007, -0.00001112, -0.00000492, 0.00000609, 0.00001735, 0.00002464, 0.00002619, 0.00002322, 0.00001862, 0.00001482, 0.00001212, 0.00000868, 0.00000214, -0.00000831, -0.00002080, -0.00003129, -0.00003576, -0.00003241, -0.00002280, -0.00001109, -0.00000187, 0.00000214, 0.00000112, -0.00000248, -0.00000573, -0.00000698, -0.00000653, -0.00000593, -0.00000661, -0.00000877, },
		{ -0.00000728, -0.00000394, -0.00000140, 0.00000059, 0.00000286, 0.00000599, 0.00000959, 0.00001235, 0.00001283, 0.00001042, 0.00000587, 0.00000084, -0.00000315, -0.00000564, -0.00000741, -0.00000965, -0.00001278, -0.00001573, -0.00001637, -0.00001284, -0.00000501, 0.00000496, 0.00001356, 0.00001768, 0.00001625, 0.00001075, 0.00000405, -0.00000139, -0.00000493, -0.00000780, -0.00001163, -0.00001676, -0.00002129, -0.00002207, -0.00001675, -0.00000582, 0.00000704, 0.00001659, 0.00001906, 0.00001437, 0.00000633, 0.00000024, -0.00000048, 0.00000335, 0.00000709, 0.00000537, -0.00000399, -0.00001783, -0.00002890, -0.00003013, -0.00001932, -0.00000147, 0.00001334, 0.00001544, 0.00000162, -0.00002207, -0.00004266, -0.00004695, -0.00002894, 0.00000606, 0.00004328, 0.00006545, 0.00006146, 0.00003205, -0.00001041, -0.00004786, -0.00006504, -0.00005625, -0.00002709, 0.00000917, 0.00003831, 0.00005124, 0.00004677, 0.00003026, 0.00000972, -0.00000813, -0.00001996, -0.00002545, -0.00002572, -0.00002186, -0.00001462, -0.00000500, 0.00000510, 0.00001308, 0.00001674, 0.00001555, 0.00001108, 0.00000613, 0.00000300, 0.00000210, 0.00000184, -0.00000000, -0.00000434, -0.00000966, -0.00001253, -0.00000962, -0.00000002, 0.00001373, 0.00002659, 0.00003358, 0.00003242, 0.00002463, 0.00001452, 0.00000653, 0.00000273, 0.00000186, 0.00000055, -0.00000418, -0.00001268, -0.00002222, -0.00002848, -0.00002814, -0.00002096, -0.00001002, -0.00000008, 0.00000493, 0.00000384, -0.00000155, -0.00000784, -0.00001212, -0.00001339, -0.00001266, -0.00001175, -0.00001191, -0.00001299, -0.00001381, -0.00001317, -0.00001074, },
		{ 0.00000863, 0.00000822, 0.00000768, 0.00000671, 0.00000455, 0.00000079, -0.00000409, -0.00000873, -0.00001161, -0.00001196, -0.00001020, -0.00000763, -0.00000545, -0.00000399, -0.00000250, 0.00000004, 0.00000394, 0.00000810, 0.00001042, 0.00000898, 0.00000341, -0.00000465, -0.00001227, -0.00001680, -0.00001726, -0.00001466, -0.00001110, -0.00000808, -0.00000554, -0.00000201, 0.00000390, 0.00001184, 0.00001920, 0.00002222, 0.00001825, 0.00000776, -0.00000536, -0.00001567, -0.00001926, -0.00001602, -0.00000977, -0.00000588, -0.00000773, -0.00001437, -0.00002087, -0.00002154, -0.00001382, -0.00000057, 0.00001105, 0.00001382, 0.00000517, -0.00001045, -0.00002341, -0.00002426, -0.00000963, 0.00001489, 0.00003689, 0.00004353, 0.00002874, -0.00000279, -0.00003722, -0.00005820, -0.00005506, -0.00002820, 0.00001104, 0.00004605, 0.00006280, 0.00005607, 0.00003091, -0.00000089, -0.00002694, -0.00003964, -0.00003829, -0.00002751, -0.00001360, -0.00000124, 0.00000778, 0.00001365, 0.00001685, 0.00001707, 0.00001354, 0.00000629, -0.00000291, -0.00001088, -0.00001463, -0.00001313, -0.00000801, -0.00000253, 0.00000053, 0.00000071, 0.00000024, 0.00000240, 0.00000907, 0.00001904, 0.00002821, 0.00003187, 0.00002751, 0.00001658, 0.00000390, -0.00000493, -0.00000674, -0.00000248, 0.00000342, 0.00000581, 0.00000185, -0.00000727, -0.00001716, -0.00002285, -0.00002173, -0.00001490, -0.00000645, -0.00000097, -0.00000102, -0.00000593, -0.00001259, -0.00001743, -0.00001845, -0.00001600, -0.00001212, -0.00000892, -0.00000731, -0.00000668, -0.00000564, -0.00000322, 0.00000046, 0.00000434, 0.00000721, 0.00000853, },
		{ -0.00000528, -0.00000702, -0.00000872, -0.00000989, -0.00000973, -0.00000774, -0.00000426, -0.00000048, 0.00000219, 0.00000295, 0.00000203, 0.00000039, -0.00000098, -0.00000192, -0.00000307, -0.00000535, -0.00000895, -0.00001285, -0.00001513, -0.00001410, -0.00000941, -0.00000245, 0.00000438, 0.00000899, 0.00001081, 0.00001086, 0.00001071, 0.00001109, 0.00001099, 0.00000824, 0.00000118, -0.00000969, -0.00002120, -0.00002887, -0.00002934, -0.00002256, -0.00001212, -0.00000339, -0.00000038, -0.00000334, -0.00000856, -0.00001080, -0.00000674, 0.00000253, 0.00001189, 0.00001526, 0.00000968, -0.00000237, -0.00001392, -0.00001754, -0.00001017, 0.00000446, 0.00001742, 0.00001958, 0.00000741, -0.00001422, -0.00003388, -0.00003958, -0.00002560, 0.00000376, 0.00003568, 0.00005510, 0.00005238, 0.00002820, -0.00000681, -0.00003749, -0.00005156, -0.00004515, -0.00002365, 0.00000197, 0.00002104, 0.00002798, 0.00002362, 0.00001308, 0.00000201, -0.00000634, -0.00001173, -0.00001540, -0.00001799, -0.00001856, -0.00001541, -0.00000790, 0.00000227, 0.00001150, 0.00001616, 0.00001501, 0.00001007, 0.00000554, 0.00000518, 0.00000981, 0.00001663, 0.00002094, 0.00001912, 0.00001116, 0.00000092, -0.00000602, -0.00000577, 0.00000136, 0.00001081, 0.00001627, 0.00001329, 0.00000187, -0.00001347, -0.00002623, -0.00003136, -0.00002789, -0.00001903, -0.00001008, -0.00000536, -0.00000605, -0.00001005, -0.00001370, -0.00001411, -0.00001073, -0.00000529, -0.00000033, 0.00000240, 0.00000291, 0.00000254, 0.00000284, 0.00000442, 0.00000659, 0.00000802, 0.00000767, 0.00000548, 0.00000227, -0.00000088, -0.00000337, },
		{ -0.00000407, -0.00000234, -0.00000043, 0.00000112, 0.00000150, 0.00000012, -0.00000278, -0.00000623, -0.00000895, -0.00001013, -0.00000980, -0.00000861, -0.00000722, -0.00000567, -0.00000338, 0.00000029, 0.00000526, 0.00001034, 0.00001366, 0.00001372, 0.00001031, 0.00000480, -0.00000073, -0.00000474, -0.00000724, -0.00000966, -0.00001357, -0.00001925, -0.00002490, -0.00002740, -0.00002422, -0.00001533, -0.00000374, 0.00000584, 0.00000954, 0.00000672, 0.00000041, -0.00000442, -0.00000400, 0.00000175, 0.00000904, 0.00001246, 0.00000853, -0.00000173, -0.00001309, -0.00001909, -0.00001610, -0.00000587, 0.00000524, 0.00001003, 0.00000510, -0.00000657, -0.00001703, -0.00001796, -0.00000611, 0.00001397, 0.00003179, 0.00003645, 0.00002308, -0.00000380, -0.00003203, -0.00004782, -0.00004302, -0.00001945, 0.00001199, 0.00003684, 0.00004436, 0.00003242, 0.00000778, -0.00001810, -0.00003506, -0.00003874, -0.00003134, -0.00001888, -0.00000714, 0.00000126, 0.00000695, 0.00001183, 0.00001656, 0.00001964, 0.00001859, 0.00001233, 0.00000275, -0.00000562, -0.00000814, -0.00000294, 0.00000760, 0.00001795, 0.00002257, 0.00001923, 0.00001036, 0.00000153, -0.00000199, 0.00000142, 0.00000859, 0.00001342, 0.00001051, -0.00000141, -0.00001852, -0.00003378, -0.00004074, -0.00003686, -0.00002471, -0.00001026, 0.00000045, 0.00000428, 0.00000217, -0.00000215, -0.00000470, -0.00000354, 0.00000054, 0.00000500, 0.00000744, 0.00000709, 0.00000497, 0.00000289, 0.00000213, 0.00000258, 0.00000302, 0.00000211, -0.00000062, -0.00000445, -0.00000797, -0.00000997, -0.00001012, -0.00000893, -0.00000725, -0.00000562, },
		{ -0.00000280, -0.00000433, -0.00000622, -0.00000790, -0.00000852, -0.00000745, -0.00000469, -0.00000097, 0.00000270, 0.00000557, 0.00000749, 0.00000871, 0.00000935, 0.00000902, 0.00000697, 0.00000261, -0.00000371, -0.00001054, -0.00001588, -0.00001826, -0.00001767, -0.00001557, -0.00001396, -0.00001402, -0.00001512, -0.00001514, -0.00001192, -0.00000492, 0.00000394, 0.00001101, 0.00001305, 0.00000922, 0.00000192, -0.00000447, -0.00000617, -0.00000243, 0.00000388, 0.00000793, 0.00000597, -0.00000205, -0.00001217, -0.00001863, -0.00001741, -0.00000890, 0.00000216, 0.00000944, 0.00000898, 0.00000180, -0.00000665, -0.00000990, -0.00000487, 0.00000576, 0.00001477, 0.00001479, 0.00000332, -0.00001490, -0.00002982, -0.00003157, -0.00001647, 0.00000998, 0.00003526, 0.00004607, 0.00003539, 0.00000647, -0.00002854, -0.00005460, -0.00006112, -0.00004681, -0.00001947, 0.00000869, 0.00002732, 0.00003243, 0.00002684, 0.00001698, 0.00000851, 0.00000343, 0.00000024, -0.00000345, -0.00000808, -0.00001125, -0.00000925, -0.00000014, 0.00001402, 0.00002754, 0.00003389, 0.00002953, 0.00001620, 0.00000016, -0.00001129, -0.00001386, -0.00000873, -0.00000159, 0.00000084, -0.00000499, -0.00001732, -0.00003001, -0.00003608, -0.00003168, -0.00001821, -0.00000147, 0.00001165, 0.00001668, 0.00001364, 0.00000635, -0.00000014, -0.00000243, -0.00000029, 0.00000384, 0.00000683, 0.00000681, 0.00000410, 0.00000052, -0.00000208, -0.00000303, -0.00000312, -0.00000379, -0.00000589, -0.00000909, -0.00001199, -0.00001311, -0.00001179, -0.00000860, -0.00000490, -0.00000201, -0.00000056, -0.00000037, -0.00000089, -0.00000171, },
		{ -0.00000132, 0.00000057, 0.00000299, 0.00000547, 0.00000719, 0.00000751, 0.00000624, 0.00000370, 0.00000045, -0.00000316, -0.00000709, -0.00001138, -0.00001569, -0.00001906, -0.00002023, -0.00001838, -0.00001386, -0.00000829, -0.00000384, -0.00000202, -0.00000270, -0.00000412, -0.00000400, -0.00000107, 0.00000387, 0.00000826, 0.00000932, 0.00000591, -0.00000044, -0.00000626, -0.00000815, -0.00000498, 0.00000114, 0.00000601, 0.00000588, -0.00000006, -0.00000893, -0.00001569, -0.00001618, -0.00000979, -0.00000002, 0.00000756, 0.00000869, 0.00000314, -0.00000517, -0.00001063, -0.00000964, -0.00000308, 0.00000415, 0.00000641, 0.00000128, -0.00000824, -0.00001515, -0.00001282, 0.00000030, 0.00001857, 0.00003143, 0.00002917, 0.00000888, -0.00002280, -0.00005216, -0.00006508, -0.00005429, -0.00002344, 0.00001460, 0.00004403, 0.00005384, 0.00004268, 0.00001856, -0.00000616, -0.00002130, -0.00002342, -0.00001600, -0.00000585, 0.00000167, 0.00000552, 0.00000847, 0.00001385, 0.00002214, 0.00002976, 0.00003111, 0.00002235, 0.00000462, -0.00001580, -0.00003049, -0.00003347, -0.00002461, -0.00000974, 0.00000263, 0.00000606, -0.00000033, -0.00001163, -0.00002027, -0.00002040, -0.00001118, 0.00000301, 0.00001531, 0.00002029, 0.00001677, 0.00000803, -0.00000054, -0.00000464, -0.00000321, 0.00000141, 0.00000545, 0.00000603, 0.00000274, -0.00000260, -0.00000741, -0.00001005, -0.00001060, -0.00001045, -0.00001095, -0.00001242, -0.00001390, -0.00001390, -0.00001154, -0.00000722, -0.00000243, 0.00000109, 0.00000231, 0.00000139, -0.00000064, -0.00000257, -0.00000367, -0.00000386, -0.00000343, -0.00000260, },
		{ 0.00000147, -0.00000086, -0.00000393, -0.00000738, -0.00001063, -0.00001319, -0.00001488, -0.00001577, -0.00001589, -0.00001502, -0.00001268, -0.00000861, -0.00000328, 0.00000193, 0.00000527, 0.00000549, 0.00000276, -0.00000126, -0.00000416, -0.00000421, -0.00000146, 0.00000221, 0.00000422, 0.00000296, -0.00000103, -0.00000526, -0.00000679, -0.00000424, 0.00000108, 0.00000578, 0.00000634, 0.00000148, -0.00000680, -0.00001418, -0.00001652, -0.00001243, -0.00000426, 0.00000326, 0.00000584, 0.00000235, -0.00000451, -0.00000994, -0.00001016, -0.00000494, 0.00000221, 0.00000640, 0.00000473, -0.00000146, -0.00000726, -0.00000750, -0.00000055, 0.00000971, 0.00001563, 0.00001040, -0.00000710, -0.00003029, -0.00004746, -0.00004782, -0.00002785, 0.00000578, 0.00003883, 0.00005630, 0.00005020, 0.00002364, -0.00001064, -0.00003686, -0.00004406, -0.00003115, -0.00000648, 0.00001739, 0.00003063, 0.00003062, 0.00002200, 0.00001250, 0.00000757, 0.00000725, 0.00000693, 0.00000137, -0.00001094, -0.00002601, -0.00003627, -0.00003521, -0.00002162, -0.00000089, 0.00001762, 0.00002554, 0.00002010, 0.00000541, -0.00000995, -0.00001779, -0.00001459, -0.00000301, 0.00001015, 0.00001807, 0.00001757, 0.00001039, 0.00000151, -0.00000413, -0.00000449, -0.00000116, 0.00000208, 0.00000190, -0.00000260, -0.00000959, -0.00001590, -0.00001903, -0.00001851, -0.00001580, -0.00001300, -0.00001140, -0.00001077, -0.00000978, -0.00000722, -0.00000296, 0.00000181, 0.00000533, 0.00000632, 0.00000479, 0.00000195, -0.00000056, -0.00000154, -0.00000081, 0.00000094, 0.00000275, 0.00000393, 0.00000428, 0.00000394, 0.00000303, },
		{ -0.00001384, -0.00001310, -0.00001155, -0.00000947, -0.00000718, -0.00000482, -0.00000227, 0.00000061, 0.00000363, 0.00000603, 0.00000672, 0.00000491, 0.00000082, -0.00000411, -0.00000774, -0.00000836, -0.00000575, -0.00000140, 0.00000220, 0.00000309, 0.00000104, -0.00000225, -0.00000422, -0.00000318, 0.00000043, 0.00000408, 0.00000469, 0.00000070, -0.00000672, -0.00001402, -0.00001728, -0.00001461, -0.00000738, 0.00000045, 0.00000470, 0.00000347, -0.00000167, -0.00000682, -0.00000828, -0.00000502, 0.00000059, 0.00000443, 0.00000350, -0.00000185, -0.00000790, -0.00001004, -0.00000608, 0.00000177, 0.00000780, 0.00000637, -0.00000398, -0.00001862, -0.00002875, -0.00002634, -0.00000940, 0.00001575, 0.00003693, 0.00004243, 0.00002765, -0.00000156, -0.00003135, -0.00004674, -0.00003955, -0.00001281, 0.00002070, 0.00004518, 0.00004986, 0.00003426, 0.00000766, -0.00001657, -0.00002833, -0.00002572, -0.00001490, -0.00000535, -0.00000369, -0.00000974, -0.00001719, -0.00001813, -0.00000853, 0.00000880, 0.00002562, 0.00003295, 0.00002649, 0.00000918, -0.00001032, -0.00002239, -0.00002162, -0.00000944, 0.00000712, 0.00001960, 0.00002259, 0.00001616, 0.00000493, -0.00000513, -0.00001023, -0.00001043, -0.00000881, -0.00000892, -0.00001237, -0.00001788, -0.00002238, -0.00002315, -0.00001949, -0.00001309, -0.00000677, -0.00000271, -0.00000131, -0.00000123, -0.00000058, 0.00000166, 0.00000500, 0.00000787, 0.00000872, 0.00000709, 0.00000393, 0.00000092, -0.00000052, -0.00000002, 0.00000159, 0.00000293, 0.00000287, 0.00000109, -0.00000193, -0.00000535, -0.00000848, -0.00001095, -0.00001269, -0.00001368, },
		{ 0.00000157, 0.00000239, 0.00000272, 0.00000292, 0.00000316, 0.00000322, 0.00000251, 0.00000047, -0.00000288, -0.00000661, -0.00000918, -0.00000919, -0.00000634, -0.00000178, 0.00000234, 0.00000402, 0.00000265, -0.00000060, -0.00000344, -0.00000394, -0.00000186, 0.00000108, 0.00000221, -0.00000030, -0.00000607, -0.00001243, -0.00001592, -0.00001433, -0.00000824, -0.00000077, 0.00000415, 0.00000417, -0.00000011, -0.00000556, -0.00000860, -0.00000754, -0.00000361, -0.00000009, 0.00000015, -0.00000327, -0.00000767, -0.00000911, -0.00000519, 0.00000281, 0.00001026, 0.00001192, 0.00000552, -0.00000612, -0.00001622, -0.00001784, -0.00000850, 0.00000762, 0.00002135, 0.00002367, 0.00001140, -0.00001015, -0.00002928, -0.00003420, -0.00001994, 0.00000810, 0.00003623, 0.00004947, 0.00003966, 0.00001015, -0.00002562, -0.00005101, -0.00005480, -0.00003677, -0.00000735, 0.00001847, 0.00002925, 0.00002285, 0.00000661, -0.00000765, -0.00001085, -0.00000167, 0.00001326, 0.00002386, 0.00002288, 0.00001023, -0.00000713, -0.00001948, -0.00002001, -0.00000862, 0.00000822, 0.00002139, 0.00002410, 0.00001534, -0.00000025, -0.00001543, -0.00002453, -0.00002601, -0.00002244, -0.00001810, -0.00001601, -0.00001633, -0.00001682, -0.00001483, -0.00000937, -0.00000186, 0.00000480, 0.00000808, 0.00000740, 0.00000436, 0.00000159, 0.00000108, 0.00000301, 0.00000590, 0.00000766, 0.00000700, 0.00000419, 0.00000070, -0.00000178, -0.00000249, -0.00000192, -0.00000144, -0.00000231, -0.00000495, -0.00000871, -0.00001231, -0.00001459, -0.00001497, -0.00001361, -0.00001110, -0.00000805, -0.00000498, -0.00000220, 0.00000004, },
		{ -0.00000131, -0.00000159, -0.00000177, -0.00000245, -0.00000396, -0.00000595, -0.00000746, -0.00000741, -0.00000531, -0.00000181, 0.00000151, 0.00000299, 0.00000186, -0.00000112, -0.00000407, -0.00000510, -0.00000356, -0.00000061, 0.00000141, 0.00000046, -0.00000376, -0.00000942, -0.00001350, -0.00001361, -0.00000949, -0.00000338, 0.00000134, 0.00000211, -0.00000120, -0.00000625, -0.00000974, -0.00000955, -0.00000606, -0.00000181, 0.00000041, -0.00000045, -0.00000279, -0.00000346, -0.00000011, 0.00000671, 0.00001339, 0.00001521, 0.00000958, -0.00000185, -0.00001347, -0.00001874, -0.00001430, -0.00000238, 0.00001001, 0.00001507, 0.00000910, -0.00000472, -0.00001771, -0.00002073, -0.00000998, 0.00000978, 0.00002719, 0.00003038, 0.00001420, -0.00001590, -0.00004576, -0.00005960, -0.00004853, -0.00001581, 0.00002431, 0.00005363, 0.00005928, 0.00004021, 0.00000735, -0.00002243, -0.00003524, -0.00002703, -0.00000482, 0.00001797, 0.00002928, 0.00002469, 0.00000897, -0.00000759, -0.00001561, -0.00001185, -0.00000051, 0.00000985, 0.00001169, 0.00000275, -0.00001296, -0.00002792, -0.00003546, -0.00003322, -0.00002380, -0.00001259, -0.00000434, -0.00000075, -0.00000025, 0.00000016, 0.00000255, 0.00000660, 0.00001007, 0.00001057, 0.00000733, 0.00000192, -0.00000278, -0.00000428, -0.00000204, 0.00000229, 0.00000596, 0.00000678, 0.00000430, -0.00000014, -0.00000439, -0.00000689, -0.00000746, -0.00000714, -0.00000734, -0.00000877, -0.00001106, -0.00001298, -0.00001328, -0.00001138, -0.00000766, -0.00000316, 0.00000094, 0.00000385, 0.00000527, 0.00000532, 0.00000433, 0.00000272, 0.00000096, -0.00000048, },
		{ -0.00000251, -0.00000369, -0.00000518, -0.00000599, -0.00000542, -0.00000350, -0.00000117, 0.00000027, -0.00000003, -0.00000184, -0.00000386, -0.00000451, -0.00000306, -0.00000024, 0.00000203, 0.00000172, -0.00000187, -0.00000753, -0.00001264, -0.00001460, -0.00001248, -0.00000757, -0.00000269, -0.00000049, -0.00000180, -0.00000517, -0.00000786, -0.00000765, -0.00000431, 0.00000031, 0.00000358, 0.00000404, 0.00000244, 0.00000126, 0.00000283, 0.00000738, 0.00001234, 0.00001375, 0.00000889, -0.00000145, -0.00001282, -0.00001931, -0.00001712, -0.00000719, 0.00000488, 0.00001191, 0.00000945, -0.00000121, -0.00001354, -0.00001956, -0.00001499, -0.00000237, 0.00000993, 0.00001261, 0.00000152, -0.00001882, -0.00003673, -0.00003959, -0.00002135, 0.00001292, 0.00004850, 0.00006805, 0.00006075, 0.00002845, -0.00001469, -0.00004896, -0.00005896, -0.00004122, -0.00000556, 0.00003030, 0.00004989, 0.00004593, 0.00002301, -0.00000596, -0.00002727, -0.00003341, -0.00002600, -0.00001339, -0.00000502, -0.00000588, -0.00001424, -0.00002356, -0.00002694, -0.00002131, -0.00000888, 0.00000462, 0.00001353, 0.00001537, 0.00001167, 0.00000637, 0.00000302, 0.00000270, 0.00000386, 0.00000392, 0.00000136, -0.00000309, -0.00000695, -0.00000771, -0.00000457, 0.00000089, 0.00000565, 0.00000697, 0.00000400, -0.00000187, -0.00000794, -0.00001177, -0.00001246, -0.00001082, -0.00000859, -0.00000721, -0.00000698, -0.00000709, -0.00000633, -0.00000399, -0.00000032, 0.00000360, 0.00000649, 0.00000752, 0.00000660, 0.00000428, 0.00000136, -0.00000140, -0.00000348, -0.00000458, -0.00000465, -0.00000391, -0.00000288, -0.00000224, },
		{ -0.00000693, -0.00000602, -0.00000405, -0.00000205, -0.00000101, -0.00000123, -0.00000209, -0.00000244, -0.00000152, 0.00000039, 0.00000192, 0.00000139, -0.00000207, -0.00000768, -0.00001322, -0.00001613, -0.00001499, -0.00001041, -0.00000472, -0.00000062, 0.00000040, -0.00000112, -0.00000309, -0.00000333, -0.00000105, 0.00000265, 0.00000561, 0.00000623, 0.00000468, 0.00000285, 0.00000298, 0.00000581, 0.00000969, 0.00001130, 0.00000783, -0.00000080, -0.00001135, -0.00001870, -0.00001889, -0.00001177, -0.00000154, 0.00000550, 0.00000466, -0.00000401, -0.00001542, -0.00002237, -0.00001992, -0.00000882, 0.00000461, 0.00001217, 0.00000906, -0.00000260, -0.00001430, -0.00001602, -0.00000262, 0.00002198, 0.00004574, 0.00005458, 0.00004035, 0.00000632, -0.00003331, -0.00005997, -0.00006022, -0.00003297, 0.00000948, 0.00004732, 0.00006298, 0.00004965, 0.00001420, -0.00002686, -0.00005599, -0.00006295, -0.00004886, -0.00002404, -0.00000157, 0.00000968, 0.00000871, 0.00000136, -0.00000435, -0.00000328, 0.00000420, 0.00001327, 0.00001822, 0.00001612, 0.00000832, -0.00000075, -0.00000664, -0.00000744, -0.00000457, -0.00000133, -0.00000051, -0.00000256, -0.00000557, -0.00000680, -0.00000475, -0.00000028, 0.00000393, 0.00000502, 0.00000184, -0.00000438, -0.00001070, -0.00001420, -0.00001353, -0.00000951, -0.00000437, -0.00000038, 0.00000139, 0.00000137, 0.00000092, 0.00000125, 0.00000268, 0.00000452, 0.00000562, 0.00000517, 0.00000311, 0.00000011, -0.00000285, -0.00000499, -0.00000592, -0.00000570, -0.00000462, -0.00000312, -0.00000171, -0.00000093, -0.00000121, -0.00000259, -0.00000460, -0.00000634, },
		{ -0.00000081, 0.00000006, -0.00000017, -0.00000072, -0.00000079, -0.00000018, 0.00000037, -0.00000037, -0.00000328, -0.00000800, -0.00001287, -0.00001562, -0.00001469, -0.00001019, -0.00000395, 0.00000140, 0.00000395, 0.00000351, 0.00000158, 0.00000028, 0.00000089, 0.00000311, 0.00000538, 0.00000610, 0.00000492, 0.00000308, 0.00000246, 0.00000406, 0.00000690, 0.00000832, 0.00000560, -0.00000198, -0.00001214, -0.00002047, -0.00002290, -0.00001828, -0.00000939, -0.00000154, 0.00000057, -0.00000412, -0.00001206, -0.00001712, -0.00001437, -0.00000347, 0.00001067, 0.00002052, 0.00002061, 0.00001116, -0.00000147, -0.00000843, -0.00000386, 0.00001088, 0.00002717, 0.00003369, 0.00002320, -0.00000235, -0.00003176, -0.00004994, -0.00004609, -0.00002020, 0.00001600, 0.00004450, 0.00005010, 0.00002840, -0.00001166, -0.00005214, -0.00007499, -0.00007074, -0.00004262, -0.00000413, 0.00002829, 0.00004326, 0.00003912, 0.00002309, 0.00000611, -0.00000343, -0.00000347, 0.00000200, 0.00000659, 0.00000588, -0.00000011, -0.00000755, -0.00001182, -0.00001057, -0.00000509, 0.00000082, 0.00000346, 0.00000159, -0.00000304, -0.00000705, -0.00000779, -0.00000501, -0.00000089, 0.00000152, 0.00000034, -0.00000390, -0.00000862, -0.00001083, -0.00000893, -0.00000359, 0.00000271, 0.00000721, 0.00000834, 0.00000643, 0.00000319, 0.00000056, -0.00000043, 0.00000000, 0.00000076, 0.00000076, -0.00000039, -0.00000229, -0.00000409, -0.00000507, -0.00000498, -0.00000408, -0.00000285, -0.00000174, -0.00000108, -0.00000110, -0.00000193, -0.00000354, -0.00000559, -0.00000741, -0.00000820, -0.00000750, -0.00000545, -0.00000287, },
		{ -0.00000112, -0.00000135, -0.00000122, -0.00000160, -0.00000334, -0.00000655, -0.00001020, -0.00001248, -0.00001186, -0.00000800, -0.00000217, 0.00000335, 0.00000650, 0.00000657, 0.00000452, 0.00000226, 0.00000142, 0.00000231, 0.00000392, 0.00000472, 0.00000394, 0.00000211, 0.00000070, 0.00000083, 0.00000220, 0.00000291, 0.00000062, -0.00000570, -0.00001454, -0.00002226, -0.00002496, -0.00002084, -0.00001164, -0.00000186, 0.00000375, 0.00000310, -0.00000183, -0.00000610, -0.00000495, 0.00000294, 0.00001416, 0.00002234, 0.00002202, 0.00001236, -0.00000178, -0.00001229, -0.00001265, -0.00000218, 0.00001290, 0.00002270, 0.00001942, 0.00000243, -0.00002068, -0.00003782, -0.00003923, -0.00002340, 0.00000138, 0.00002107, 0.00002355, 0.00000562, -0.00002467, -0.00005178, -0.00006067, -0.00004478, -0.00000957, 0.00003020, 0.00005785, 0.00006298, 0.00004601, 0.00001702, -0.00001011, -0.00002480, -0.00002434, -0.00001377, -0.00000179, 0.00000457, 0.00000342, -0.00000204, -0.00000656, -0.00000658, -0.00000236, 0.00000264, 0.00000443, 0.00000125, -0.00000530, -0.00001135, -0.00001333, -0.00001019, -0.00000389, 0.00000196, 0.00000438, 0.00000282, -0.00000076, -0.00000326, -0.00000252, 0.00000139, 0.00000635, 0.00000951, 0.00000909, 0.00000529, 0.00000005, -0.00000419, -0.00000588, -0.00000500, -0.00000282, -0.00000092, -0.00000027, -0.00000088, -0.00000201, -0.00000286, -0.00000306, -0.00000275, -0.00000236, -0.00000226, -0.00000256, -0.00000318, -0.00000405, -0.00000513, -0.00000634, -0.00000739, -0.00000780, -0.00000714, -0.00000535, -0.00000291, -0.00000071, 0.00000049, 0.00000046, -0.00000035, },
		{ -0.00000135, -0.00000318, -0.00000584, -0.00000802, -0.00000822, -0.00000570, -0.00000108, 0.00000386, 0.00000713, 0.00000759, 0.00000563, 0.00000277, 0.00000071, 0.00000025, 0.00000090, 0.00000137, 0.00000064, -0.00000124, -0.00000322, -0.00000407, -0.00000353, -0.00000271, -0.00000351, -0.00000717, -0.00001302, -0.00001841, -0.00001999, -0.00001581, -0.00000682, 0.00000330, 0.00001002, 0.00001067, 0.00000612, 0.00000039, -0.00000181, 0.00000182, 0.00000947, 0.00001605, 0.00001636, 0.00000859, -0.00000398, -0.00001450, -0.00001649, -0.00000810, 0.00000621, 0.00001765, 0.00001804, 0.00000481, -0.00001698, -0.00003696, -0.00004506, -0.00003730, -0.00001817, 0.00000166, 0.00001148, 0.00000673, -0.00000817, -0.00002222, -0.00002420, -0.00000933, 0.00001748, 0.00004389, 0.00005659, 0.00004849, 0.00002252, -0.00000983, -0.00003452, -0.00004199, -0.00003152, -0.00001074, 0.00000908, 0.00001912, 0.00001705, 0.00000706, -0.00000357, -0.00000907, -0.00000816, -0.00000396, -0.00000114, -0.00000263, -0.00000770, -0.00001272, -0.00001369, -0.00000896, -0.00000031, 0.00000822, 0.00001271, 0.00001171, 0.00000683, 0.00000153, -0.00000108, -0.00000006, 0.00000309, 0.00000556, 0.00000515, 0.00000160, -0.00000338, -0.00000728, -0.00000828, -0.00000622, -0.00000251, 0.00000084, 0.00000238, 0.00000188, 0.00000010, -0.00000184, -0.00000319, -0.00000386, -0.00000417, -0.00000448, -0.00000488, -0.00000523, -0.00000540, -0.00000541, -0.00000548, -0.00000573, -0.00000599, -0.00000588, -0.00000507, -0.00000356, -0.00000185, -0.00000060, -0.00000024, -0.00000066, -0.00000126, -0.00000146, -0.00000113, -0.00000082, },
		{ -0.00000542, -0.00000464, -0.00000167, 0.00000229, 0.00000538, 0.00000618, 0.00000444, 0.00000123, -0.00000186, -0.00000362, -0.00000397, -0.00000376, -0.00000402, -0.00000510, -0.00000636, -0.00000673, -0.00000562, -0.00000361, -0.00000225, -0.00000297, -0.00000593, -0.00000949, -0.00001100, -0.00000844, -0.00000194, 0.00000586, 0.00001120, 0.00001133, 0.00000645, -0.00000026, -0.00000438, -0.00000303, 0.00000308, 0.00000987, 0.00001226, 0.00000746, -0.00000291, -0.00001335, -0.00001771, -0.00001303, -0.00000191, 0.00000850, 0.00001044, 0.00000027, -0.00001881, -0.00003796, -0.00004720, -0.00004095, -0.00002129, 0.00000304, 0.00002136, 0.00002684, 0.00002007, 0.00000830, 0.00000078, 0.00000304, 0.00001352, 0.00002455, 0.00002708, 0.00001646, -0.00000432, -0.00002584, -0.00003731, -0.00003257, -0.00001361, 0.00001042, 0.00002809, 0.00003161, 0.00002063, 0.00000181, -0.00001525, -0.00002322, -0.00002059, -0.00001148, -0.00000254, 0.00000131, -0.00000054, -0.00000470, -0.00000643, -0.00000295, 0.00000482, 0.00001292, 0.00001698, 0.00001485, 0.00000782, -0.00000031, -0.00000558, -0.00000612, -0.00000295, 0.00000095, 0.00000269, 0.00000115, -0.00000251, -0.00000585, -0.00000675, -0.00000470, -0.00000095, 0.00000239, 0.00000365, 0.00000244, -0.00000037, -0.00000335, -0.00000540, -0.00000623, -0.00000622, -0.00000592, -0.00000563, -0.00000530, -0.00000478, -0.00000412, -0.00000365, -0.00000374, -0.00000442, -0.00000532, -0.00000580, -0.00000537, -0.00000407, -0.00000238, -0.00000095, -0.00000012, 0.00000028, 0.00000068, 0.00000136, 0.00000207, 0.00000213, 0.00000096, -0.00000138, -0.00000397, },
		{ 0.00000097, 0.00000222, 0.00000110, -0.00000185, -0.00000520, -0.00000749, -0.00000806, -0.00000729, -0.00000615, -0.00000536, -0.00000486, -0.00000398, -0.00000212, 0.00000052, 0.00000278, 0.00000323, 0.00000126, -0.00000226, -0.00000520, -0.00000546, -0.00000237, 0.00000263, 0.00000662, 0.00000704, 0.00000338, -0.00000220, -0.00000605, -0.00000534, -0.00000006, 0.00000654, 0.00000973, 0.00000624, -0.00000346, -0.00001491, -0.00002209, -0.00002101, -0.00001236, -0.00000158, 0.00000414, 0.00000018, -0.00001239, -0.00002686, -0.00003434, -0.00002873, -0.00001046, 0.00001336, 0.00003255, 0.00003916, 0.00003157, 0.00001507, -0.00000145, -0.00001059, -0.00001031, -0.00000444, 0.00000044, -0.00000044, -0.00000684, -0.00001365, -0.00001439, -0.00000570, 0.00000989, 0.00002483, 0.00003067, 0.00002293, 0.00000401, -0.00001780, -0.00003262, -0.00003412, -0.00002267, -0.00000483, 0.00001056, 0.00001709, 0.00001390, 0.00000540, -0.00000188, -0.00000327, 0.00000151, 0.00000873, 0.00001326, 0.00001186, 0.00000504, -0.00000353, -0.00000936, -0.00000982, -0.00000552, 0.00000041, 0.00000436, 0.00000439, 0.00000111, -0.00000300, -0.00000532, -0.00000469, -0.00000193, 0.00000095, 0.00000206, 0.00000078, -0.00000212, -0.00000510, -0.00000686, -0.00000704, -0.00000616, -0.00000505, -0.00000426, -0.00000380, -0.00000344, -0.00000303, -0.00000281, -0.00000317, -0.00000433, -0.00000592, -0.00000716, -0.00000721, -0.00000571, -0.00000303, -0.00000003, 0.00000247, 0.00000409, 0.00000499, 0.00000552, 0.00000577, 0.00000543, 0.00000398, 0.00000127, -0.00000211, -0.00000493, -0.00000596, -0.00000472, -0.00000188, },
		{ -0.00000607, -0.00000792, -0.00000803, -0.00000638, -0.00000381, -0.00000138, 0.00000037, 0.00000164, 0.00000296, 0.00000451, 0.00000571, 0.00000554, 0.00000329, -0.00000064, -0.00000463, -0.00000668, -0.00000563, -0.00000208, 0.00000179, 0.00000349, 0.00000186, -0.00000207, -0.00000559, -0.00000604, -0.00000281, 0.00000197, 0.00000443, 0.00000140, -0.00000730, -0.00001812, -0.00002546, -0.00002484, -0.00001574, -0.00000238, 0.00000835, 0.00001078, 0.00000383, -0.00000802, -0.00001710, -0.00001685, -0.00000578, 0.00001139, 0.00002609, 0.00003056, 0.00002212, 0.00000460, -0.00001376, -0.00002480, -0.00002463, -0.00001519, -0.00000249, 0.00000726, 0.00001114, 0.00001056, 0.00000943, 0.00001074, 0.00001412, 0.00001584, 0.00001149, -0.00000051, -0.00001691, -0.00003064, -0.00003442, -0.00002502, -0.00000550, 0.00001591, 0.00002991, 0.00003089, 0.00001972, 0.00000301, -0.00001061, -0.00001507, -0.00000978, 0.00000065, 0.00000949, 0.00001184, 0.00000709, -0.00000122, -0.00000790, -0.00000930, -0.00000524, 0.00000127, 0.00000611, 0.00000662, 0.00000290, -0.00000253, -0.00000655, -0.00000728, -0.00000503, -0.00000178, 0.00000026, -0.00000001, -0.00000210, -0.00000453, -0.00000588, -0.00000569, -0.00000449, -0.00000328, -0.00000277, -0.00000299, -0.00000345, -0.00000371, -0.00000376, -0.00000399, -0.00000481, -0.00000609, -0.00000713, -0.00000694, -0.00000493, -0.00000130, 0.00000297, 0.00000664, 0.00000887, 0.00000951, 0.00000901, 0.00000786, 0.00000621, 0.00000382, 0.00000047, -0.00000357, -0.00000736, -0.00000967, -0.00000966, -0.00000751, -0.00000448, -0.00000221, -0.00000189, -0.00000354, },
		{ 0.00000012, 0.00000285, 0.00000457, 0.00000507, 0.00000483, 0.00000445, 0.00000404, 0.00000312, 0.00000110, -0.00000207, -0.00000545, -0.00000753, -0.00000713, -0.00000434, -0.00000072, 0.00000151, 0.00000084, -0.00000248, -0.00000644, -0.00000854, -0.00000745, -0.00000418, -0.00000157, -0.00000253, -0.00000789, -0.00001531, -0.00002025, -0.00001848, -0.00000893, 0.00000513, 0.00001737, 0.00002166, 0.00001561, 0.00000216, -0.00001177, -0.00001892, -0.00001575, -0.00000445, 0.00000827, 0.00001498, 0.00001171, 0.00000016, -0.00001326, -0.00002107, -0.00001885, -0.00000751, 0.00000755, 0.00001960, 0.00002419, 0.00002109, 0.00001355, 0.00000570, -0.00000010, -0.00000425, -0.00000864, -0.00001434, -0.00001980, -0.00002129, -0.00001525, -0.00000121, 0.00001677, 0.00003136, 0.00003551, 0.00002638, 0.00000732, -0.00001345, -0.00002679, -0.00002725, -0.00001568, 0.00000138, 0.00001539, 0.00002019, 0.00001490, 0.00000380, -0.00000651, -0.00001092, -0.00000830, -0.00000159, 0.00000444, 0.00000613, 0.00000287, -0.00000299, -0.00000788, -0.00000924, -0.00000689, -0.00000276, 0.00000058, 0.00000147, 0.00000001, -0.00000239, -0.00000411, -0.00000440, -0.00000365, -0.00000290, -0.00000295, -0.00000380, -0.00000480, -0.00000520, -0.00000478, -0.00000399, -0.00000353, -0.00000372, -0.00000412, -0.00000378, -0.00000187, 0.00000167, 0.00000599, 0.00000966, 0.00001153, 0.00001121, 0.00000919, 0.00000633, 0.00000328, 0.00000014, -0.00000332, -0.00000716, -0.00001085, -0.00001333, -0.00001362, -0.00001147, -0.00000773, -0.00000398, -0.00000174, -0.00000159, -0.00000291, -0.00000426, -0.00000433, -0.00000267, },
		{ 0.00000329, 0.00000163, 0.00000038, -0.00000045, -0.00000136, -0.00000282, -0.00000472, -0.00000625, -0.00000645, -0.00000502, -0.00000277, -0.00000128, -0.00000194, -0.00000485, -0.00000858, -0.00001086, -0.00001009, -0.00000646, -0.00000212, 0.00000011, -0.00000134, -0.00000551, -0.00000910, -0.00000842, -0.00000183, 0.00000866, 0.00001798, 0.00002052, 0.00001345, -0.00000131, -0.00001760, -0.00002798, -0.00002770, -0.00001722, -0.00000200, 0.00001042, 0.00001451, 0.00000962, 0.00000002, -0.00000771, -0.00000856, -0.00000179, 0.00000885, 0.00001756, 0.00001981, 0.00001468, 0.00000503, -0.00000460, -0.00001061, -0.00001203, -0.00001025, -0.00000733, -0.00000418, -0.00000007, 0.00000627, 0.00001471, 0.00002269, 0.00002586, 0.00002053, 0.00000645, -0.00001209, -0.00002758, -0.00003286, -0.00002490, -0.00000676, 0.00001364, 0.00002732, 0.00002867, 0.00001811, 0.00000156, -0.00001283, -0.00001893, -0.00001545, -0.00000602, 0.00000334, 0.00000769, 0.00000562, -0.00000054, -0.00000658, -0.00000907, -0.00000716, -0.00000263, 0.00000154, 0.00000314, 0.00000181, -0.00000109, -0.00000366, -0.00000474, -0.00000445, -0.00000380, -0.00000374, -0.00000441, -0.00000511, -0.00000493, -0.00000352, -0.00000141, 0.00000037, 0.00000101, 0.00000058, 0.00000003, 0.00000049, 0.00000252, 0.00000562, 0.00000848, 0.00000973, 0.00000872, 0.00000584, 0.00000214, -0.00000136, -0.00000427, -0.00000682, -0.00000941, -0.00001198, -0.00001381, -0.00001389, -0.00001162, -0.00000739, -0.00000258, 0.00000109, 0.00000255, 0.00000189, 0.00000031, -0.00000067, -0.00000015, 0.00000168, 0.00000377, 0.00000495, 0.00000468, },
		{ -0.00000243, -0.00000239, -0.00000305, -0.00000428, -0.00000537, -0.00000565, -0.00000509, -0.00000445, -0.00000470, -0.00000620, -0.00000816, -0.00000897, -0.00000721, -0.00000282, 0.00000254, 0.00000623, 0.00000622, 0.00000253, -0.00000248, -0.00000535, -0.00000372, 0.00000194, 0.00000814, 0.00001019, 0.00000499, -0.00000672, -0.00002021, -0.00002890, -0.00002786, -0.00001664, 0.00000025, 0.00001550, 0.00002271, 0.00001967, 0.00000927, -0.00000232, -0.00000909, -0.00000836, -0.00000181, 0.00000593, 0.00001023, 0.00000907, 0.00000389, -0.00000173, -0.00000438, -0.00000284, 0.00000163, 0.00000647, 0.00000964, 0.00001074, 0.00001066, 0.00001037, 0.00000958, 0.00000677, 0.00000043, -0.00000917, -0.00001905, -0.00002446, -0.00002132, -0.00000899, 0.00000862, 0.00002426, 0.00003079, 0.00002485, 0.00000879, -0.00001025, -0.00002387, -0.00002652, -0.00001814, -0.00000383, 0.00000905, 0.00001474, 0.00001179, 0.00000321, -0.00000567, -0.00001031, -0.00000923, -0.00000426, 0.00000105, 0.00000367, 0.00000271, -0.00000053, -0.00000377, -0.00000534, -0.00000501, -0.00000378, -0.00000285, -0.00000269, -0.00000276, -0.00000209, -0.00000019, 0.00000241, 0.00000448, 0.00000485, 0.00000331, 0.00000084, -0.00000102, -0.00000109, 0.00000068, 0.00000317, 0.00000479, 0.00000450, 0.00000231, -0.00000082, -0.00000368, -0.00000565, -0.00000689, -0.00000803, -0.00000945, -0.00001080, -0.00001108, -0.00000937, -0.00000557, -0.00000071, 0.00000348, 0.00000552, 0.00000499, 0.00000278, 0.00000051, -0.00000047, 0.00000019, 0.00000173, 0.00000286, 0.00000267, 0.00000118, -0.00000076, -0.00000218, -0.00000262, },
		{ -0.00000386, -0.00000530, -0.00000591, -0.00000564, -0.00000507, -0.00000476, -0.00000471, -0.00000415, -0.00000217, 0.00000143, 0.00000553, 0.00000806, 0.00000714, 0.00000244, -0.00000423, -0.00000966, -0.00001097, -0.00000749, -0.00000145, 0.00000312, 0.00000264, -0.00000361, -0.00001276, -0.00001958, -0.00001936, -0.00001075, 0.00000318, 0.00001624, 0.00002230, 0.00001855, 0.00000697, -0.00000667, -0.00001590, -0.00001679, -0.00000974, 0.00000122, 0.00001087, 0.00001565, 0.00001516, 0.00001173, 0.00000846, 0.00000724, 0.00000782, 0.00000844, 0.00000738, 0.00000424, 0.00000012, -0.00000332, -0.00000515, -0.00000576, -0.00000640, -0.00000782, -0.00000932, -0.00000884, -0.00000443, 0.00000402, 0.00001379, 0.00002024, 0.00001913, 0.00000932, -0.00000599, -0.00002037, -0.00002715, -0.00002293, -0.00000947, 0.00000705, 0.00001912, 0.00002162, 0.00001427, 0.00000148, -0.00001026, -0.00001578, -0.00001365, -0.00000642, 0.00000134, 0.00000568, 0.00000527, 0.00000156, -0.00000255, -0.00000465, -0.00000413, -0.00000205, -0.00000006, 0.00000085, 0.00000090, 0.00000111, 0.00000221, 0.00000398, 0.00000520, 0.00000457, 0.00000168, -0.00000252, -0.00000613, -0.00000742, -0.00000587, -0.00000243, 0.00000101, 0.00000280, 0.00000239, 0.00000044, -0.00000177, -0.00000328, -0.00000400, -0.00000457, -0.00000563, -0.00000715, -0.00000824, -0.00000773, -0.00000504, -0.00000077, 0.00000347, 0.00000589, 0.00000558, 0.00000302, -0.00000019, -0.00000230, -0.00000243, -0.00000098, 0.00000072, 0.00000136, 0.00000051, -0.00000126, -0.00000279, -0.00000323, -0.00000259, -0.00000166, -0.00000141, -0.00000227, },
		{ -0.00000456, -0.00000336, -0.00000233, -0.00000146, -0.00000033, 0.00000141, 0.00000333, 0.00000421, 0.00000266, -0.00000173, -0.00000775, -0.00001273, -0.00001396, -0.00001029, -0.00000316, 0.00000387, 0.00000692, 0.00000410, -0.00000322, -0.00001092, -0.00001434, -0.00001096, -0.00000203, 0.00000793, 0.00001353, 0.00001154, 0.00000279, -0.00000815, -0.00001536, -0.00001472, -0.00000598, 0.00000724, 0.00001943, 0.00002608, 0.00002566, 0.00001984, 0.00001203, 0.00000527, 0.00000083, -0.00000183, -0.00000399, -0.00000631, -0.00000831, -0.00000881, -0.00000707, -0.00000354, 0.00000023, 0.00000258, 0.00000291, 0.00000206, 0.00000163, 0.00000266, 0.00000457, 0.00000535, 0.00000285, -0.00000346, -0.00001146, -0.00001711, -0.00001652, -0.00000843, 0.00000457, 0.00001691, 0.00002272, 0.00001894, 0.00000703, -0.00000762, -0.00001838, -0.00002076, -0.00001452, -0.00000347, 0.00000676, 0.00001173, 0.00001030, 0.00000469, -0.00000121, -0.00000423, -0.00000348, -0.00000040, 0.00000263, 0.00000387, 0.00000322, 0.00000185, 0.00000107, 0.00000124, 0.00000147, 0.00000039, -0.00000270, -0.00000708, -0.00001083, -0.00001193, -0.00000950, -0.00000439, 0.00000124, 0.00000515, 0.00000618, 0.00000471, 0.00000216, -0.00000006, -0.00000138, -0.00000225, -0.00000346, -0.00000530, -0.00000716, -0.00000778, -0.00000623, -0.00000265, 0.00000157, 0.00000449, 0.00000474, 0.00000238, -0.00000115, -0.00000383, -0.00000431, -0.00000259, 0.00000002, 0.00000185, 0.00000188, 0.00000034, -0.00000158, -0.00000259, -0.00000218, -0.00000090, 0.00000011, -0.00000009, -0.00000159, -0.00000361, -0.00000507, -0.00000535, },
		{ -0.00000144, -0.00000084, -0.00000011, 0.00000029, -0.00000050, -0.00000304, -0.00000681, -0.00001009, -0.00001072, -0.00000746, -0.00000110, 0.00000556, 0.00000897, 0.00000681, -0.00000047, -0.00000955, -0.00001586, -0.00001601, -0.00000973, -0.00000017, 0.00000783, 0.00001039, 0.00000680, -0.00000007, -0.00000541, -0.00000515, 0.00000171, 0.00001254, 0.00002250, 0.00002714, 0.00002454, 0.00001611, 0.00000543, -0.00000368, -0.00000902, -0.00001054, -0.00000962, -0.00000782, -0.00000580, -0.00000337, -0.00000021, 0.00000337, 0.00000617, 0.00000680, 0.00000462, 0.00000040, -0.00000397, -0.00000661, -0.00000672, -0.00000508, -0.00000335, -0.00000282, -0.00000336, -0.00000338, -0.00000104, 0.00000423, 0.00001074, 0.00001511, 0.00001408, 0.00000661, -0.00000494, -0.00001570, -0.00002057, -0.00001697, -0.00000631, 0.00000661, 0.00001604, 0.00001823, 0.00001317, 0.00000433, -0.00000346, -0.00000667, -0.00000474, 0.00000003, 0.00000421, 0.00000536, 0.00000323, -0.00000046, -0.00000348, -0.00000462, -0.00000436, -0.00000428, -0.00000570, -0.00000855, -0.00001125, -0.00001169, -0.00000866, -0.00000272, 0.00000397, 0.00000883, 0.00001024, 0.00000835, 0.00000471, 0.00000124, -0.00000101, -0.00000220, -0.00000329, -0.00000501, -0.00000716, -0.00000859, -0.00000808, -0.00000527, -0.00000115, 0.00000239, 0.00000363, 0.00000215, -0.00000092, -0.00000352, -0.00000392, -0.00000179, 0.00000169, 0.00000453, 0.00000518, 0.00000347, 0.00000058, -0.00000175, -0.00000236, -0.00000133, 0.00000019, 0.00000081, -0.00000019, -0.00000241, -0.00000469, -0.00000591, -0.00000568, -0.00000445, -0.00000305, -0.00000206, },
		{ -0.00000028, -0.00000133, -0.00000353, -0.00000605, -0.00000733, -0.00000608, -0.00000229, 0.00000232, 0.00000500, 0.00000352, -0.00000232, -0.00001016, -0.00001603, -0.00001643, -0.00001032, 0.00000012, 0.00001043, 0.00001627, 0.00001566, 0.00001005, 0.00000341, 0.00000000, 0.00000195, 0.00000815, 0.00001500, 0.00001845, 0.00001628, 0.00000915, 0.00000007, -0.00000733, -0.00001068, -0.00000982, -0.00000635, -0.00000243, 0.00000048, 0.00000212, 0.00000296, 0.00000338, 0.00000318, 0.00000173, -0.00000132, -0.00000536, -0.00000884, -0.00001002, -0.00000802, -0.00000343, 0.00000189, 0.00000590, 0.00000749, 0.00000701, 0.00000577, 0.00000497, 0.00000466, 0.00000377, 0.00000095, -0.00000406, -0.00000966, -0.00001293, -0.00001116, -0.00000372, 0.00000706, 0.00001677, 0.00002099, 0.00001770, 0.00000846, -0.00000232, -0.00000969, -0.00001072, -0.00000599, 0.00000101, 0.00000596, 0.00000610, 0.00000165, -0.00000468, -0.00000950, -0.00001078, -0.00000887, -0.00000595, -0.00000439, -0.00000511, -0.00000703, -0.00000792, -0.00000592, -0.00000090, 0.00000535, 0.00001018, 0.00001161, 0.00000941, 0.00000508, 0.00000079, -0.00000198, -0.00000316, -0.00000383, -0.00000514, -0.00000734, -0.00000943, -0.00000990, -0.00000781, -0.00000363, 0.00000092, 0.00000383, 0.00000406, 0.00000220, 0.00000010, -0.00000034, 0.00000162, 0.00000502, 0.00000781, 0.00000810, 0.00000537, 0.00000080, -0.00000344, -0.00000551, -0.00000491, -0.00000270, -0.00000071, -0.00000042, -0.00000205, -0.00000462, -0.00000667, -0.00000717, -0.00000612, -0.00000430, -0.00000267, -0.00000168, -0.00000117, -0.00000073, -0.00000029, },
		{ -0.00000735, -0.00000789, -0.00000614, -0.00000274, 0.00000042, 0.00000128, -0.00000107, -0.00000550, -0.00000918, -0.00000909, -0.00000381, 0.00000534, 0.00001471, 0.00002015, 0.00001919, 0.00001244, 0.00000333, -0.00000368, -0.00000548, -0.00000185, 0.00000459, 0.00001000, 0.00001144, 0.00000834, 0.00000256, -0.00000289, -0.00000565, -0.00000516, -0.00000264, -0.00000007, 0.00000107, 0.00000053, -0.00000091, -0.00000222, -0.00000293, -0.00000328, -0.00000381, -0.00000463, -0.00000512, -0.00000426, -0.00000142, 0.00000298, 0.00000743, 0.00001006, 0.00000969, 0.00000653, 0.00000203, -0.00000194, -0.00000417, -0.00000462, -0.00000409, -0.00000338, -0.00000247, -0.00000052, 0.00000334, 0.00000896, 0.00001458, 0.00001746, 0.00001533, 0.00000796, -0.00000220, -0.00001098, -0.00001453, -0.00001145, -0.00000368, 0.00000441, 0.00000835, 0.00000599, -0.00000141, -0.00000991, -0.00001519, -0.00001491, -0.00000985, -0.00000327, 0.00000119, 0.00000164, -0.00000114, -0.00000434, -0.00000510, -0.00000221, 0.00000313, 0.00000822, 0.00001049, 0.00000903, 0.00000492, 0.00000038, -0.00000274, -0.00000402, -0.00000442, -0.00000533, -0.00000734, -0.00000965, -0.00001055, -0.00000862, -0.00000384, 0.00000219, 0.00000713, 0.00000921, 0.00000832, 0.00000595, 0.00000420, 0.00000434, 0.00000600, 0.00000738, 0.00000652, 0.00000264, -0.00000322, -0.00000867, -0.00001134, -0.00001025, -0.00000638, -0.00000200, 0.00000060, 0.00000037, -0.00000205, -0.00000491, -0.00000652, -0.00000617, -0.00000435, -0.00000221, -0.00000072, -0.00000013, -0.00000008, -0.00000014, -0.00000032, -0.00000106, -0.00000276, -0.00000519, },
		{ 0.00000069, 0.00000408, 0.00000499, 0.00000334, 0.00000076, -0.00000033, 0.00000183, 0.00000700, 0.00001283, 0.00001598, 0.00001402, 0.00000700, -0.00000235, -0.00000991, -0.00001224, -0.00000842, -0.00000046, 0.00000781, 0.00001276, 0.00001271, 0.00000849, 0.00000268, -0.00000206, -0.00000430, -0.00000434, -0.00000356, -0.00000327, -0.00000385, -0.00000469, -0.00000479, -0.00000363, -0.00000154, 0.00000055, 0.00000185, 0.00000223, 0.00000229, 0.00000283, 0.00000409, 0.00000550, 0.00000594, 0.00000455, 0.00000143, -0.00000222, -0.00000473, -0.00000486, -0.00000249, 0.00000139, 0.00000533, 0.00000827, 0.00000997, 0.00001081, 0.00001104, 0.00001034, 0.00000783, 0.00000292, -0.00000385, -0.00001046, -0.00001411, -0.00001276, -0.00000654, 0.00000183, 0.00000819, 0.00000898, 0.00000334, -0.00000626, -0.00001511, -0.00001873, -0.00001528, -0.00000655, 0.00000300, 0.00000872, 0.00000826, 0.00000271, -0.00000418, -0.00000823, -0.00000721, -0.00000186, 0.00000471, 0.00000905, 0.00000926, 0.00000583, 0.00000103, -0.00000277, -0.00000446, -0.00000463, -0.00000475, -0.00000578, -0.00000734, -0.00000784, -0.00000566, -0.00000045, 0.00000637, 0.00001218, 0.00001467, 0.00001313, 0.00000886, 0.00000428, 0.00000142, 0.00000073, 0.00000093, -0.00000001, -0.00000329, -0.00000833, -0.00001298, -0.00001469, -0.00001217, -0.00000624, 0.00000051, 0.00000514, 0.00000586, 0.00000296, -0.00000159, -0.00000534, -0.00000673, -0.00000570, -0.00000348, -0.00000154, -0.00000074, -0.00000099, -0.00000169, -0.00000234, -0.00000300, -0.00000408, -0.00000576, -0.00000753, -0.00000829, -0.00000703, -0.00000365, },
		{ 0.00000660, 0.00000319, 0.00000152, 0.00000245, 0.00000499, 0.00000677, 0.00000554, 0.00000065, -0.00000625, -0.00001191, -0.00001320, -0.00000886, -0.00000033, 0.00000890, 0.00001504, 0.00001581, 0.00001146, 0.00000431, -0.00000255, -0.00000689, -0.00000816, -0.00000728, -0.00000567, -0.00000425, -0.00000304, -0.00000152, 0.00000069, 0.00000329, 0.00000538, 0.00000605, 0.00000503, 0.00000295, 0.00000094, -0.00000004, 0.00000020, 0.00000103, 0.00000158, 0.00000138, 0.00000079, 0.00000071, 0.00000200, 0.00000473, 0.00000806, 0.00001064, 0.00001136, 0.00000997, 0.00000704, 0.00000351, 0.00000007, -0.00000313, -0.00000618, -0.00000889, -0.00001044, -0.00000967, -0.00000596, -0.00000008, 0.00000564, 0.00000821, 0.00000559, -0.00000186, -0.00001111, -0.00001767, -0.00001788, -0.00001103, 0.00000005, 0.00001020, 0.00001456, 0.00001113, 0.00000183, -0.00000850, -0.00001469, -0.00001388, -0.00000693, 0.00000237, 0.00000947, 0.00001150, 0.00000855, 0.00000315, -0.00000153, -0.00000355, -0.00000304, -0.00000164, -0.00000098, -0.00000141, -0.00000181, -0.00000049, 0.00000331, 0.00000863, 0.00001307, 0.00001417, 0.00001092, 0.00000453, -0.00000228, -0.00000680, -0.00000791, -0.00000658, -0.00000507, -0.00000536, -0.00000777, -0.00001065, -0.00001152, -0.00000867, -0.00000246, 0.00000469, 0.00000957, 0.00000991, 0.00000565, -0.00000110, -0.00000727, -0.00001046, -0.00001009, -0.00000733, -0.00000421, -0.00000229, -0.00000198, -0.00000265, -0.00000335, -0.00000361, -0.00000360, -0.00000377, -0.00000421, -0.00000429, -0.00000308, -0.00000004, 0.00000431, 0.00000841, 0.00001050, 0.00000968, },
		{ -0.00000318, -0.00000116, -0.00000031, -0.00000198, -0.00000574, -0.00000941, -0.00001032, -0.00000690, 0.00000025, 0.00000845, 0.00001426, 0.00001525, 0.00001119, 0.00000403, -0.00000315, -0.00000775, -0.00000877, -0.00000688, -0.00000374, -0.00000084, 0.00000119, 0.00000253, 0.00000368, 0.00000474, 0.00000533, 0.00000484, 0.00000307, 0.00000059, -0.00000148, -0.00000203, -0.00000068, 0.00000201, 0.00000489, 0.00000689, 0.00000760, 0.00000742, 0.00000707, 0.00000706, 0.00000721, 0.00000688, 0.00000538, 0.00000260, -0.00000088, -0.00000407, -0.00000619, -0.00000700, -0.00000684, -0.00000613, -0.00000501, -0.00000326, -0.00000069, 0.00000226, 0.00000435, 0.00000400, 0.00000032, -0.00000592, -0.00001215, -0.00001503, -0.00001224, -0.00000404, 0.00000635, 0.00001405, 0.00001497, 0.00000805, -0.00000393, -0.00001550, -0.00002124, -0.00001847, -0.00000858, 0.00000380, 0.00001329, 0.00001647, 0.00001339, 0.00000708, 0.00000154, -0.00000059, 0.00000076, 0.00000356, 0.00000530, 0.00000474, 0.00000260, 0.00000083, 0.00000103, 0.00000318, 0.00000554, 0.00000579, 0.00000262, -0.00000317, -0.00000905, -0.00001218, -0.00001117, -0.00000694, -0.00000216, 0.00000040, -0.00000040, -0.00000342, -0.00000597, -0.00000560, -0.00000174, 0.00000385, 0.00000803, 0.00000811, 0.00000348, -0.00000401, -0.00001108, -0.00001479, -0.00001404, -0.00000993, -0.00000490, -0.00000122, 0.00000016, -0.00000021, -0.00000098, -0.00000107, -0.00000028, 0.00000079, 0.00000149, 0.00000175, 0.00000222, 0.00000367, 0.00000626, 0.00000910, 0.00001065, 0.00000965, 0.00000606, 0.00000124, -0.00000271, -0.00000421, },
		{ -0.00000309, -0.00000584, -0.00000742, -0.00000588, -0.00000091, 0.00000584, 0.00001151, 0.00001357, 0.00001118, 0.00000555, -0.00000069, -0.00000499, -0.00000604, -0.00000423, -0.00000114, 0.00000154, 0.00000289, 0.00000303, 0.00000270, 0.00000252, 0.00000261, 0.00000260, 0.00000214, 0.00000135, 0.00000088, 0.00000149, 0.00000352, 0.00000649, 0.00000926, 0.00001064, 0.00001006, 0.00000780, 0.00000481, 0.00000211, 0.00000021, -0.00000100, -0.00000200, -0.00000315, -0.00000436, -0.00000518, -0.00000517, -0.00000429, -0.00000296, -0.00000178, -0.00000108, -0.00000074, -0.00000039, 0.00000017, 0.00000057, 0.00000004, -0.00000200, -0.00000521, -0.00000811, -0.00000868, -0.00000557, 0.00000070, 0.00000755, 0.00001129, 0.00000907, 0.00000071, -0.00001074, -0.00002017, -0.00002283, -0.00001681, -0.00000416, 0.00001006, 0.00002027, 0.00002310, 0.00001890, 0.00001124, 0.00000463, 0.00000206, 0.00000354, 0.00000650, 0.00000768, 0.00000530, 0.00000001, -0.00000560, -0.00000881, -0.00000849, -0.00000578, -0.00000312, -0.00000259, -0.00000446, -0.00000699, -0.00000768, -0.00000504, 0.00000026, 0.00000557, 0.00000785, 0.00000558, -0.00000022, -0.00000645, -0.00000976, -0.00000850, -0.00000374, 0.00000140, 0.00000357, 0.00000118, -0.00000470, -0.00001100, -0.00001442, -0.00001328, -0.00000831, -0.00000201, 0.00000290, 0.00000493, 0.00000446, 0.00000312, 0.00000254, 0.00000334, 0.00000490, 0.00000606, 0.00000605, 0.00000508, 0.00000409, 0.00000396, 0.00000470, 0.00000536, 0.00000465, 0.00000189, -0.00000230, -0.00000626, -0.00000822, -0.00000746, -0.00000481, -0.00000220, -0.00000146, },
		{ -0.00000271, 0.00000319, 0.00000936, 0.00001313, 0.00001296, 0.00000915, 0.00000367, -0.00000104, -0.00000330, -0.00000293, -0.00000106, 0.00000078, 0.00000159, 0.00000138, 0.00000091, 0.00000103, 0.00000206, 0.00000363, 0.00000504, 0.00000585, 0.00000611, 0.00000634, 0.00000695, 0.00000787, 0.00000848, 0.00000795, 0.00000582, 0.00000241, -0.00000133, -0.00000425, -0.00000562, -0.00000547, -0.00000441, -0.00000320, -0.00000233, -0.00000181, -0.00000142, -0.00000099, -0.00000064, -0.00000063, -0.00000106, -0.00000170, -0.00000206, -0.00000173, -0.00000083, 0.00000001, -0.00000001, -0.00000119, -0.00000294, -0.00000400, -0.00000324, -0.00000059, 0.00000255, 0.00000387, 0.00000148, -0.00000464, -0.00001214, -0.00001713, -0.00001612, -0.00000802, 0.00000494, 0.00001802, 0.00002625, 0.00002686, 0.00002065, 0.00001137, 0.00000364, 0.00000035, 0.00000130, 0.00000360, 0.00000366, -0.00000044, -0.00000776, -0.00001505, -0.00001870, -0.00001691, -0.00001068, -0.00000312, 0.00000245, 0.00000443, 0.00000364, 0.00000243, 0.00000289, 0.00000521, 0.00000753, 0.00000714, 0.00000245, -0.00000566, -0.00001398, -0.00001866, -0.00001741, -0.00001089, -0.00000246, 0.00000371, 0.00000486, 0.00000111, -0.00000476, -0.00000901, -0.00000907, -0.00000488, 0.00000138, 0.00000672, 0.00000911, 0.00000850, 0.00000649, 0.00000504, 0.00000520, 0.00000649, 0.00000745, 0.00000674, 0.00000416, 0.00000075, -0.00000195, -0.00000295, -0.00000249, -0.00000178, -0.00000209, -0.00000384, -0.00000619, -0.00000766, -0.00000713, -0.00000472, -0.00000183, -0.00000026, -0.00000102, -0.00000355, -0.00000593, -0.00000601, },
		{ 0.00001537, 0.00001280, 0.00000746, 0.00000174, -0.00000224, -0.00000357, -0.00000275, -0.00000108, 0.00000028, 0.00000098, 0.00000149, 0.00000252, 0.00000440, 0.00000676, 0.00000873, 0.00000952, 0.00000892, 0.00000736, 0.00000556, 0.00000403, 0.00000279, 0.00000141, -0.00000049, -0.00000287, -0.00000515, -0.00000650, -0.00000638, -0.00000484, -0.00000257, -0.00000048, 0.00000075, 0.00000098, 0.00000050, -0.00000021, -0.00000088, -0.00000141, -0.00000184, -0.00000203, -0.00000173, -0.00000073, 0.00000075, 0.00000207, 0.00000233, 0.00000101, -0.00000160, -0.00000437, -0.00000591, -0.00000545, -0.00000340, -0.00000123, -0.00000061, -0.00000218, -0.00000490, -0.00000640, -0.00000427, 0.00000232, 0.00001175, 0.00002048, 0.00002474, 0.00002254, 0.00001486, 0.00000517, -0.00000243, -0.00000544, -0.00000428, -0.00000187, -0.00000169, -0.00000554, -0.00001229, -0.00001841, -0.00002005, -0.00001535, -0.00000561, 0.00000527, 0.00001298, 0.00001501, 0.00001187, 0.00000640, 0.00000181, -0.00000031, -0.00000081, -0.00000209, -0.00000613, -0.00001278, -0.00001946, -0.00002249, -0.00001937, -0.00001045, 0.00000091, 0.00001001, 0.00001326, 0.00001007, 0.00000307, -0.00000346, -0.00000595, -0.00000337, 0.00000250, 0.00000840, 0.00001157, 0.00001120, 0.00000855, 0.00000580, 0.00000451, 0.00000465, 0.00000482, 0.00000339, -0.00000021, -0.00000502, -0.00000905, -0.00001060, -0.00000924, -0.00000609, -0.00000306, -0.00000162, -0.00000201, -0.00000319, -0.00000376, -0.00000300, -0.00000141, -0.00000030, -0.00000084, -0.00000302, -0.00000547, -0.00000616, -0.00000364, 0.00000191, 0.00000863, 0.00001383, },
		{ -0.00000203, -0.00000442, -0.00000396, -0.00000179, 0.00000068, 0.00000256, 0.00000384, 0.00000502, 0.00000650, 0.00000814, 0.00000922, 0.00000900, 0.00000718, 0.00000418, 0.00000093, -0.00000165, -0.00000316, -0.00000379, -0.00000407, -0.00000438, -0.00000471, -0.00000467, -0.00000386, -0.00000226, -0.00000030, 0.00000131, 0.00000203, 0.00000174, 0.00000080, -0.00000022, -0.00000088, -0.00000107, -0.00000089, -0.00000050, 0.00000006, 0.00000082, 0.00000161, 0.00000201, 0.00000140, -0.00000058, -0.00000361, -0.00000661, -0.00000811, -0.00000708, -0.00000359, 0.00000106, 0.00000494, 0.00000661, 0.00000594, 0.00000428, 0.00000363, 0.00000529, 0.00000894, 0.00001261, 0.00001376, 0.00001089, 0.00000458, -0.00000264, -0.00000767, -0.00000864, -0.00000606, -0.00000256, -0.00000125, -0.00000375, -0.00000897, -0.00001359, -0.00001401, -0.00000855, 0.00000128, 0.00001139, 0.00001723, 0.00001621, 0.00000913, -0.00000044, -0.00000825, -0.00001169, -0.00001111, -0.00000919, -0.00000890, -0.00001142, -0.00001517, -0.00001683, -0.00001350, -0.00000478, 0.00000660, 0.00001606, 0.00001964, 0.00001616, 0.00000795, -0.00000049, -0.00000491, -0.00000352, 0.00000225, 0.00000885, 0.00001275, 0.00001233, 0.00000847, 0.00000365, 0.00000015, -0.00000128, -0.00000167, -0.00000278, -0.00000570, -0.00000991, -0.00001348, -0.00001432, -0.00001154, -0.00000609, -0.00000027, 0.00000357, 0.00000431, 0.00000249, -0.00000018, -0.00000207, -0.00000259, -0.00000234, -0.00000245, -0.00000351, -0.00000494, -0.00000524, -0.00000295, 0.00000220, 0.00000886, 0.00001456, 0.00001696, 0.00001505, 0.00000971, 0.00000319, },
		{ 0.00000193, 0.00000424, 0.00000533, 0.00000559, 0.00000566, 0.00000582, 0.00000578, 0.00000495, 0.00000296, 0.00000004, -0.00000297, -0.00000514, -0.00000588, -0.00000531, -0.00000405, -0.00000282, -0.00000202, -0.00000156, -0.00000106, -0.00000024, 0.00000084, 0.00000184, 0.00000232, 0.00000213, 0.00000147, 0.00000076, 0.00000033, 0.00000026, 0.00000037, 0.00000041, 0.00000029, 0.00000009, -0.00000009, -0.00000035, -0.00000096, -0.00000215, -0.00000375, -0.00000502, -0.00000488, -0.00000251, 0.00000200, 0.00000738, 0.00001162, 0.00001292, 0.00001072, 0.00000603, 0.00000102, -0.00000221, -0.00000273, -0.00000125, 0.00000043, 0.00000063, -0.00000107, -0.00000352, -0.00000481, -0.00000363, -0.00000031, 0.00000316, 0.00000432, 0.00000184, -0.00000337, -0.00000842, -0.00001008, -0.00000674, 0.00000039, 0.00000767, 0.00001100, 0.00000811, 0.00000001, -0.00000955, -0.00001604, -0.00001675, -0.00001221, -0.00000574, -0.00000136, -0.00000132, -0.00000472, -0.00000812, -0.00000775, -0.00000194, 0.00000768, 0.00001690, 0.00002133, 0.00001892, 0.00001113, 0.00000213, -0.00000357, -0.00000356, 0.00000126, 0.00000739, 0.00001092, 0.00000968, 0.00000431, -0.00000243, -0.00000749, -0.00000936, -0.00000867, -0.00000750, -0.00000772, -0.00000966, -0.00001187, -0.00001216, -0.00000912, -0.00000319, 0.00000345, 0.00000806, 0.00000886, 0.00000600, 0.00000129, -0.00000291, -0.00000505, -0.00000508, -0.00000410, -0.00000331, -0.00000308, -0.00000264, -0.00000072, 0.00000337, 0.00000899, 0.00001423, 0.00001686, 0.00001556, 0.00001071, 0.00000424, -0.00000132, -0.00000412, -0.00000373, -0.00000117, },
		{ 0.00000164, 0.00000080, 0.00000050, 0.00000028, -0.00000047, -0.00000191, -0.00000363, -0.00000484, -0.00000493, -0.00000381, -0.00000201, -0.00000030, 0.00000075, 0.00000108, 0.00000103, 0.00000110, 0.00000153, 0.00000218, 0.00000271, 0.00000282, 0.00000244, 0.00000178, 0.00000110, 0.00000052, -0.00000000, -0.00000057, -0.00000117, -0.00000161, -0.00000163, -0.00000117, -0.00000043, 0.00000020, 0.00000051, 0.00000071, 0.00000140, 0.00000318, 0.00000608, 0.00000922, 0.00001101, 0.00000995, 0.00000551, -0.00000131, -0.00000825, -0.00001279, -0.00001337, -0.00001018, -0.00000492, 0.00000013, 0.00000342, 0.00000478, 0.00000521, 0.00000589, 0.00000709, 0.00000788, 0.00000667, 0.00000253, -0.00000386, -0.00001016, -0.00001354, -0.00001234, -0.00000726, -0.00000117, 0.00000236, 0.00000121, -0.00000400, -0.00001008, -0.00001310, -0.00001072, -0.00000370, 0.00000439, 0.00000922, 0.00000827, 0.00000240, -0.00000450, -0.00000770, -0.00000441, 0.00000453, 0.00001501, 0.00002191, 0.00002190, 0.00001523, 0.00000550, -0.00000251, -0.00000549, -0.00000334, 0.00000110, 0.00000394, 0.00000261, -0.00000264, -0.00000915, -0.00001356, -0.00001389, -0.00001065, -0.00000627, -0.00000350, -0.00000361, -0.00000566, -0.00000728, -0.00000626, -0.00000207, 0.00000372, 0.00000837, 0.00000959, 0.00000688, 0.00000173, -0.00000328, -0.00000601, -0.00000585, -0.00000377, -0.00000146, -0.00000006, 0.00000053, 0.00000140, 0.00000372, 0.00000767, 0.00001209, 0.00001503, 0.00001490, 0.00001142, 0.00000593, 0.00000065, -0.00000245, -0.00000261, -0.00000057, 0.00000208, 0.00000380, 0.00000396, 0.00000294, },
		{ -0.00000076, -0.00000025, -0.00000062, -0.00000131, -0.00000156, -0.00000097, 0.00000031, 0.00000170, 0.00000259, 0.00000275, 0.00000241, 0.00000204, 0.00000202, 0.00000237, 0.00000280, 0.00000292, 0.00000252, 0.00000168, 0.00000065, -0.00000030, -0.00000101, -0.00000145, -0.00000156, -0.00000120, -0.00000024, 0.00000130, 0.00000306, 0.00000449, 0.00000507, 0.00000469, 0.00000377, 0.00000299, 0.00000283, 0.00000307, 0.00000281, 0.00000095, -0.00000303, -0.00000845, -0.00001346, -0.00001582, -0.00001402, -0.00000818, -0.00000009, 0.00000759, 0.00001261, 0.00001411, 0.00001277, 0.00001006, 0.00000720, 0.00000436, 0.00000084, -0.00000407, -0.00001012, -0.00001576, -0.00001874, -0.00001749, -0.00001223, -0.00000517, 0.00000058, 0.00000273, 0.00000120, -0.00000180, -0.00000309, -0.00000057, 0.00000528, 0.00001144, 0.00001411, 0.00001100, 0.00000307, -0.00000568, -0.00001027, -0.00000748, 0.00000213, 0.00001433, 0.00002328, 0.00002462, 0.00001770, 0.00000581, -0.00000571, -0.00001239, -0.00001283, -0.00000905, -0.00000492, -0.00000369, -0.00000604, -0.00000999, -0.00001239, -0.00001104, -0.00000620, -0.00000037, 0.00000328, 0.00000289, -0.00000088, -0.00000532, -0.00000733, -0.00000528, 0.00000001, 0.00000582, 0.00000918, 0.00000856, 0.00000464, -0.00000024, -0.00000356, -0.00000408, -0.00000230, 0.00000010, 0.00000155, 0.00000167, 0.00000139, 0.00000217, 0.00000489, 0.00000906, 0.00001295, 0.00001462, 0.00001306, 0.00000878, 0.00000358, -0.00000051, -0.00000229, -0.00000193, -0.00000067, 0.00000005, -0.00000052, -0.00000208, -0.00000365, -0.00000423, -0.00000355, -0.00000208, },
		{ 0.00000213, 0.00000136, 0.00000149, 0.00000223, 0.00000297, 0.00000325, 0.00000299, 0.00000250, 0.00000216, 0.00000213, 0.00000228, 0.00000228, 0.00000190, 0.00000114, 0.00000025, -0.00000044, -0.00000072, -0.00000054, 0.00000007, 0.00000104, 0.00000231, 0.00000373, 0.00000493, 0.00000537, 0.00000461, 0.00000255, -0.00000029, -0.00000301, -0.00000473, -0.00000514, -0.00000473, -0.00000450, -0.00000531, -0.00000718, -0.00000906, -0.00000932, -0.00000664, -0.00000096, 0.00000625, 0.00001257, 0.00001579, 0.00001493, 0.00001070, 0.00000495, -0.00000045, -0.00000455, -0.00000760, -0.00001046, -0.00001359, -0.00001641, -0.00001743, -0.00001522, -0.00000958, -0.00000198, 0.00000498, 0.00000893, 0.00000916, 0.00000701, 0.00000509, 0.00000554, 0.00000858, 0.00001216, 0.00001320, 0.00000961, 0.00000197, -0.00000635, -0.00001081, -0.00000829, 0.00000089, 0.00001273, 0.00002142, 0.00002229, 0.00001431, 0.00000066, -0.00001292, -0.00002109, -0.00002153, -0.00001582, -0.00000815, -0.00000270, -0.00000143, -0.00000326, -0.00000524, -0.00000473, -0.00000125, 0.00000330, 0.00000585, 0.00000431, -0.00000096, -0.00000719, -0.00001076, -0.00000928, -0.00000305, 0.00000514, 0.00001153, 0.00001346, 0.00001066, 0.00000514, -0.00000015, -0.00000301, -0.00000307, -0.00000157, -0.00000022, 0.00000010, -0.00000014, 0.00000045, 0.00000303, 0.00000753, 0.00001243, 0.00001554, 0.00001522, 0.00001132, 0.00000526, -0.00000074, -0.00000488, -0.00000657, -0.00000652, -0.00000602, -0.00000603, -0.00000657, -0.00000688, -0.00000606, -0.00000374, -0.00000048, 0.00000260, 0.00000442, 0.00000457, 0.00000350, },
		{ 0.00000228, 0.00000306, 0.00000325, 0.00000289, 0.00000231, 0.00000187, 0.00000166, 0.00000155, 0.00000134, 0.00000095, 0.00000050, 0.00000023, 0.00000029, 0.00000068, 0.00000125, 0.00000182, 0.00000229, 0.00000252, 0.00000236, 0.00000153, -0.00000020, -0.00000275, -0.00000561, -0.00000786, -0.00000860, -0.00000744, -0.00000481, -0.00000188, 0.00000007, 0.00000035, -0.00000065, -0.00000166, -0.00000128, 0.00000113, 0.00000490, 0.00000832, 0.00000953, 0.00000757, 0.00000298, -0.00000251, -0.00000695, -0.00000922, -0.00000954, -0.00000907, -0.00000900, -0.00000953, -0.00000975, -0.00000820, -0.00000407, 0.00000208, 0.00000833, 0.00001247, 0.00001329, 0.00001135, 0.00000869, 0.00000747, 0.00000852, 0.00001066, 0.00001138, 0.00000855, 0.00000203, -0.00000576, -0.00001095, -0.00001040, -0.00000379, 0.00000585, 0.00001344, 0.00001447, 0.00000757, -0.00000463, -0.00001673, -0.00002336, -0.00002192, -0.00001380, -0.00000340, 0.00000439, 0.00000678, 0.00000431, 0.00000006, -0.00000256, -0.00000193, 0.00000089, 0.00000312, 0.00000225, -0.00000207, -0.00000764, -0.00001090, -0.00000912, -0.00000211, 0.00000748, 0.00001549, 0.00001850, 0.00001549, 0.00000829, 0.00000034, -0.00000521, -0.00000696, -0.00000566, -0.00000319, -0.00000118, 0.00000003, 0.00000130, 0.00000381, 0.00000788, 0.00001241, 0.00001537, 0.00001488, 0.00001042, 0.00000318, -0.00000447, -0.00001020, -0.00001279, -0.00001250, -0.00001064, -0.00000856, -0.00000689, -0.00000532, -0.00000314, 0.00000009, 0.00000399, 0.00000747, 0.00000936, 0.00000907, 0.00000697, 0.00000419, 0.00000195, 0.00000102, 0.00000134, },
		{ 0.00000240, 0.00000139, 0.00000069, 0.00000039, 0.00000032, 0.00000028, 0.00000021, 0.00000013, 0.00000012, 0.00000016, 0.00000013, -0.00000008, -0.00000051, -0.00000111, -0.00000182, -0.00000267, -0.00000375, -0.00000501, -0.00000617, -0.00000664, -0.00000579, -0.00000336, 0.00000016, 0.00000360, 0.00000558, 0.00000520, 0.00000259, -0.00000098, -0.00000380, -0.00000458, -0.00000319, -0.00000070, 0.00000119, 0.00000125, -0.00000060, -0.00000325, -0.00000512, -0.00000520, -0.00000365, -0.00000170, -0.00000074, -0.00000137, -0.00000285, -0.00000354, -0.00000196, 0.00000218, 0.00000763, 0.00001226, 0.00001432, 0.00001350, 0.00001108, 0.00000901, 0.00000856, 0.00000930, 0.00000932, 0.00000652, 0.00000020, -0.00000802, -0.00001475, -0.00001665, -0.00001246, -0.00000403, 0.00000439, 0.00000833, 0.00000562, -0.00000236, -0.00001132, -0.00001637, -0.00001472, -0.00000719, 0.00000231, 0.00000905, 0.00001005, 0.00000565, -0.00000094, -0.00000571, -0.00000625, -0.00000302, 0.00000113, 0.00000301, 0.00000117, -0.00000310, -0.00000653, -0.00000598, -0.00000044, 0.00000814, 0.00001585, 0.00001892, 0.00001576, 0.00000782, -0.00000136, -0.00000801, -0.00001006, -0.00000789, -0.00000364, 0.00000027, 0.00000258, 0.00000362, 0.00000463, 0.00000654, 0.00000906, 0.00001072, 0.00000979, 0.00000547, -0.00000144, -0.00000875, -0.00001401, -0.00001571, -0.00001391, -0.00001000, -0.00000575, -0.00000236, 0.00000005, 0.00000216, 0.00000467, 0.00000763, 0.00001032, 0.00001163, 0.00001085, 0.00000816, 0.00000463, 0.00000163, 0.00000017, 0.00000042, 0.00000171, 0.00000305, 0.00000367, 0.00000334, },
		{ -0.00000171, -0.00000130, -0.00000096, -0.00000084, -0.00000095, -0.00000122, -0.00000164, -0.00000218, -0.00000275, -0.00000320, -0.00000338, -0.00000327, -0.00000298, -0.00000269, -0.00000244, -0.00000202, -0.00000106, 0.00000062, 0.00000277, 0.00000455, 0.00000486, 0.00000295, -0.00000103, -0.00000588, -0.00000976, -0.00001108, -0.00000932, -0.00000532, -0.00000083, 0.00000243, 0.00000362, 0.00000320, 0.00000238, 0.00000227, 0.00000310, 0.00000406, 0.00000396, 0.00000218, -0.00000068, -0.00000293, -0.00000279, 0.00000051, 0.00000611, 0.00001185, 0.00001544, 0.00001572, 0.00001324, 0.00000974, 0.00000694, 0.00000529, 0.00000372, 0.00000043, -0.00000558, -0.00001337, -0.00002009, -0.00002241, -0.00001842, -0.00000911, 0.00000185, 0.00000979, 0.00001149, 0.00000690, -0.00000083, -0.00000716, -0.00000865, -0.00000488, 0.00000148, 0.00000637, 0.00000680, 0.00000263, -0.00000343, -0.00000753, -0.00000707, -0.00000228, 0.00000394, 0.00000783, 0.00000713, 0.00000244, -0.00000308, -0.00000564, -0.00000311, 0.00000366, 0.00001125, 0.00001561, 0.00001435, 0.00000799, -0.00000039, -0.00000696, -0.00000918, -0.00000700, -0.00000251, 0.00000150, 0.00000321, 0.00000263, 0.00000120, 0.00000057, 0.00000137, 0.00000277, 0.00000312, 0.00000114, -0.00000310, -0.00000808, -0.00001163, -0.00001216, -0.00000950, -0.00000490, -0.00000024, 0.00000304, 0.00000466, 0.00000538, 0.00000628, 0.00000794, 0.00000997, 0.00001135, 0.00001106, 0.00000886, 0.00000548, 0.00000223, 0.00000030, 0.00000011, 0.00000118, 0.00000248, 0.00000306, 0.00000252, 0.00000114, -0.00000041, -0.00000151, -0.00000190, },
		{ -0.00000095, -0.00000136, -0.00000188, -0.00000237, -0.00000267, -0.00000266, -0.00000224, -0.00000146, -0.00000055, 0.00000018, 0.00000048, 0.00000035, 0.00000007, -0.00000000, 0.00000023, 0.00000042, -0.00000012, -0.00000199, -0.00000517, -0.00000880, -0.00001133, -0.00001134, -0.00000824, -0.00000275, 0.00000335, 0.00000808, 0.00001021, 0.00000980, 0.00000801, 0.00000625, 0.00000530, 0.00000493, 0.00000421, 0.00000241, -0.00000025, -0.00000248, -0.00000261, 0.00000029, 0.00000566, 0.00001147, 0.00001523, 0.00001528, 0.00001175, 0.00000632, 0.00000117, -0.00000237, -0.00000455, -0.00000682, -0.00001053, -0.00001561, -0.00002009, -0.00002107, -0.00001645, -0.00000654, 0.00000555, 0.00001520, 0.00001858, 0.00001464, 0.00000577, -0.00000352, -0.00000893, -0.00000864, -0.00000407, 0.00000125, 0.00000391, 0.00000269, -0.00000087, -0.00000360, -0.00000288, 0.00000154, 0.00000727, 0.00001070, 0.00000927, 0.00000319, -0.00000446, -0.00000945, -0.00000874, -0.00000230, 0.00000680, 0.00001410, 0.00001611, 0.00001214, 0.00000447, -0.00000312, -0.00000747, -0.00000763, -0.00000508, -0.00000242, -0.00000171, -0.00000324, -0.00000557, -0.00000668, -0.00000540, -0.00000218, 0.00000122, 0.00000285, 0.00000187, -0.00000101, -0.00000392, -0.00000508, -0.00000374, -0.00000062, 0.00000272, 0.00000485, 0.00000541, 0.00000511, 0.00000521, 0.00000654, 0.00000892, 0.00001122, 0.00001205, 0.00001064, 0.00000731, 0.00000330, 0.00000006, -0.00000151, -0.00000148, -0.00000069, -0.00000017, -0.00000052, -0.00000168, -0.00000302, -0.00000387, -0.00000387, -0.00000312, -0.00000209, -0.00000124, -0.00000086, },
		},
		{ { -0.00000230, -0.00000220, -0.00000184, -0.00000112, -0.00000012, 0.00000090, 0.00000157, 0.00000159, 0.00000100, 0.00000025, -0.00000012, -0.00000030, -0.00000129, -0.00000373, -0.00000612, -0.00000520, 0.00000055, 0.00000724, 0.00000719, -0.00000368, -0.00001949, -0.00002743, -0.00001972, -0.00000276, 0.00000766, 0.00000216, -0.00001180, -0.00001719, -0.00000587, 0.00001203, 0.00001879, 0.00000859, -0.00000616, -0.00000905, 0.00000182, 0.00001231, 0.00000907, -0.00000506, -0.00001505, -0.00001148, -0.00000083, 0.00000355, -0.00000201, -0.00000771, -0.00000317, 0.00000944, 0.00001789, 0.00001435, 0.00000357, -0.00000429, -0.00000611, -0.00000748, -0.00001258, -0.00001657, -0.00001109, 0.00000380, 0.00001712, 0.00001789, 0.00000682, -0.00000503, -0.00000836, -0.00000403, 0.00000089, 0.00000244, 0.00000233, 0.00000244, 0.00000089, -0.00000403, -0.00000836, -0.00000503, 0.00000682, 0.00001789, 0.00001712, 0.00000380, -0.00001109, -0.00001657, -0.00001258, -0.00000748, -0.00000611, -0.00000429, 0.00000357, 0.00001435, 0.00001789, 0.00000944, -0.00000317, -0.00000771, -0.00000201, 0.00000355, -0.00000083, -0.00001148, -0.00001505, -0.00000506, 0.00000907, 0.00001231, 0.00000182, -0.00000905, -0.00000616, 0.00000859, 0.00001879, 0.00001203, -0.00000587, -0.00001719, -0.00001180, 0.00000216, 0.00000766, -0.00000276, -0.00001972, -0.00002743, -0.00001949, -0.00000368, 0.00000719, 0.00000724, 0.00000055, -0.00000520, -0.00000612, -0.00000373, -0.00000129, -0.00000030, -0.00000012, 0.00000025, 0.00000100, 0.00000159, 0.00000157, 0.00000090, -0.00000012, -0.00000112, -0.00000184, -0.00000220, },
		{ -0.00000175, -0.00000092, 0.00000032, 0.00000149, 0.00000191, 0.00000121, -0.00000020, -0.00000137, -0.00000163, -0.00000142, -0.00000189, -0.00000322, -0.00000362, -0.00000103, 0.00000344, 0.00000453, -0.00000280, -0.00001663, -0.00002692, -0.00002372, -0.00000796, 0.00000725, 0.00000846, -0.00000394, -0.00001517, -0.00001113, 0.00000606, 0.00001963, 0.00001598, -0.00000064, -0.00001296, -0.00000979, 0.00000247, 0.00000835, 0.00000144, -0.00000917, -0.00001102, -0.00000322, 0.00000356, 0.00000150, -0.00000435, -0.00000276, 0.00000870, 0.00001935, 0.00001726, 0.00000293, -0.00001159, -0.00001628, -0.00001267, -0.00000875, -0.00000737, -0.00000373, 0.00000569, 0.00001542, 0.00001591, 0.00000525, -0.00000713, -0.00001021, -0.00000306, 0.00000523, 0.00000646, 0.00000149, -0.00000320, -0.00000418, -0.00000344, -0.00000273, 0.00000033, 0.00000779, 0.00001510, 0.00001345, 0.00000013, -0.00001585, -0.00002161, -0.00001369, -0.00000126, 0.00000478, 0.00000363, 0.00000319, 0.00000800, 0.00001265, 0.00000924, -0.00000108, -0.00000817, -0.00000549, 0.00000106, 0.00000013, -0.00000993, -0.00001723, -0.00001007, 0.00000780, 0.00001941, 0.00001334, -0.00000280, -0.00001046, -0.00000112, 0.00001396, 0.00001659, 0.00000222, -0.00001514, -0.00001875, -0.00000743, 0.00000402, 0.00000176, -0.00001233, -0.00002354, -0.00002041, -0.00000582, 0.00000733, 0.00000942, 0.00000199, -0.00000627, -0.00000871, -0.00000551, -0.00000104, 0.00000132, 0.00000140, 0.00000079, 0.00000062, 0.00000095, 0.00000129, 0.00000126, 0.00000070, -0.00000028, -0.00000133, -0.00000208, -0.00000236, -0.00000232, -0.00000214, },
		{ 0.00000097, 0.00000180, 0.00000128, -0.00000040, -0.00000203, -0.00000238, -0.00000138, -0.00000038, -0.00000080, -0.00000226, -0.00000268, -0.00000094, 0.00000054, -0.00000290, -0.00001267, -0.00002294, -0.00002420, -0.00001277, 0.00000337, 0.00001056, 0.00000313, -0.00000963, -0.00001219, 0.00000053, 0.00001660, 0.00001906, 0.00000411, -0.00001426, -0.00001905, -0.00000793, 0.00000547, 0.00000751, -0.00000105, -0.00000810, -0.00000563, 0.00000167, 0.00000428, 0.00000095, -0.00000002, 0.00000720, 0.00001648, 0.00001565, 0.00000121, -0.00001624, -0.00002274, -0.00001537, -0.00000357, 0.00000290, 0.00000389, 0.00000549, 0.00000951, 0.00001050, 0.00000411, -0.00000518, -0.00000800, -0.00000142, 0.00000688, 0.00000729, -0.00000063, -0.00000810, -0.00000767, -0.00000101, 0.00000491, 0.00000677, 0.00000710, 0.00000769, 0.00000513, -0.00000422, -0.00001631, -0.00002066, -0.00001155, 0.00000414, 0.00001317, 0.00001008, 0.00000234, 0.00000045, 0.00000520, 0.00000787, 0.00000254, -0.00000584, -0.00000811, -0.00000350, -0.00000107, -0.00000721, -0.00001525, -0.00001200, 0.00000524, 0.00002229, 0.00002225, 0.00000483, -0.00001181, -0.00001051, 0.00000596, 0.00001794, 0.00001070, -0.00000955, -0.00002305, -0.00001819, -0.00000314, 0.00000453, -0.00000267, -0.00001533, -0.00001880, -0.00000908, 0.00000425, 0.00000919, 0.00000322, -0.00000609, -0.00001010, -0.00000688, -0.00000093, 0.00000273, 0.00000282, 0.00000125, 0.00000016, 0.00000020, 0.00000085, 0.00000141, 0.00000141, 0.00000065, -0.00000062, -0.00000183, -0.00000247, -0.00000254, -0.00000237, -0.00000215, -0.00000160, -0.00000046, },
		{ -0.00000026, -0.00000201, -0.00000250, -0.00000124, 0.00000048, 0.00000060, -0.00000157, -0.00000426, -0.00000494, -0.00000352, -0.00000351, -0.00000850, -0.00001702, -0.00002181, -0.00001630, -0.00000253, 0.00000876, 0.00000815, -0.00000204, -0.00000906, -0.00000318, 0.00001072, 0.00001710, 0.00000644, -0.00001341, -0.00002432, -0.00001666, 0.00000096, 0.00001145, 0.00000738, -0.00000286, -0.00000651, -0.00000092, 0.00000573, 0.00000642, 0.00000389, 0.00000517, 0.00001010, 0.00000999, -0.00000118, -0.00001702, -0.00002375, -0.00001496, 0.00000119, 0.00001104, 0.00000963, 0.00000379, 0.00000186, 0.00000387, 0.00000395, 0.00000002, -0.00000289, -0.00000015, 0.00000498, 0.00000467, -0.00000298, -0.00001015, -0.00000796, 0.00000272, 0.00001189, 0.00001165, 0.00000421, -0.00000261, -0.00000546, -0.00000779, -0.00001205, -0.00001378, -0.00000671, 0.00000712, 0.00001672, 0.00001355, 0.00000185, -0.00000569, -0.00000240, 0.00000513, 0.00000591, -0.00000182, -0.00000918, -0.00000875, -0.00000419, -0.00000436, -0.00000983, -0.00001021, 0.00000246, 0.00002066, 0.00002665, 0.00001236, -0.00000938, -0.00001700, -0.00000360, 0.00001490, 0.00001678, -0.00000191, -0.00002332, -0.00002727, -0.00001220, 0.00000478, 0.00000726, -0.00000403, -0.00001450, -0.00001278, -0.00000189, 0.00000618, 0.00000412, -0.00000418, -0.00000959, -0.00000757, -0.00000127, 0.00000339, 0.00000378, 0.00000159, -0.00000024, -0.00000044, 0.00000045, 0.00000144, 0.00000188, 0.00000146, 0.00000022, -0.00000135, -0.00000250, -0.00000285, -0.00000269, -0.00000245, -0.00000212, -0.00000128, 0.00000011, 0.00000127, 0.00000121, },
		{ -0.00000115, 0.00000056, 0.00000081, -0.00000166, -0.00000559, -0.00000786, -0.00000670, -0.00000420, -0.00000489, -0.00001077, -0.00001766, -0.00001800, -0.00000866, 0.00000431, 0.00001049, 0.00000568, -0.00000300, -0.00000447, 0.00000353, 0.00001076, 0.00000548, -0.00001135, -0.00002507, -0.00002197, -0.00000413, 0.00001240, 0.00001426, 0.00000391, -0.00000483, -0.00000286, 0.00000557, 0.00001014, 0.00000734, 0.00000287, 0.00000207, 0.00000207, -0.00000354, -0.00001394, -0.00001933, -0.00001182, 0.00000407, 0.00001483, 0.00001222, 0.00000169, -0.00000466, -0.00000172, 0.00000461, 0.00000636, 0.00000304, 0.00000013, 0.00000023, -0.00000022, -0.00000437, -0.00000827, -0.00000472, 0.00000650, 0.00001564, 0.00001282, -0.00000058, -0.00001295, -0.00001513, -0.00000935, -0.00000419, -0.00000291, -0.00000055, 0.00000696, 0.00001470, 0.00001354, 0.00000238, -0.00000843, -0.00000832, 0.00000094, 0.00000715, 0.00000228, -0.00000822, -0.00001277, -0.00000811, -0.00000230, -0.00000272, -0.00000518, 0.00000061, 0.00001548, 0.00002586, 0.00001816, -0.00000354, -0.00001890, -0.00001272, 0.00000741, 0.00001819, 0.00000573, -0.00001903, -0.00003187, -0.00002079, 0.00000224, 0.00001502, 0.00000809, -0.00000788, -0.00001577, -0.00000993, 0.00000064, 0.00000440, -0.00000059, -0.00000689, -0.00000739, -0.00000232, 0.00000284, 0.00000400, 0.00000186, -0.00000038, -0.00000080, 0.00000021, 0.00000141, 0.00000209, 0.00000200, 0.00000104, -0.00000063, -0.00000231, -0.00000321, -0.00000322, -0.00000286, -0.00000251, -0.00000194, -0.00000083, 0.00000048, 0.00000100, 0.00000014, -0.00000141, -0.00000212, },
		{ -0.00000237, -0.00000650, -0.00000942, -0.00000853, -0.00000466, -0.00000227, -0.00000543, -0.00001289, -0.00001774, -0.00001344, -0.00000111, 0.00001016, 0.00001166, 0.00000397, -0.00000358, -0.00000333, 0.00000177, 0.00000142, -0.00000921, -0.00002193, -0.00002304, -0.00000847, 0.00001047, 0.00001823, 0.00001092, -0.00000052, -0.00000318, 0.00000399, 0.00001081, 0.00000888, 0.00000069, -0.00000527, -0.00000566, -0.00000520, -0.00000820, -0.00001120, -0.00000704, 0.00000444, 0.00001369, 0.00001155, 0.00000036, -0.00000789, -0.00000482, 0.00000534, 0.00001132, 0.00000766, -0.00000056, -0.00000545, -0.00000533, -0.00000398, -0.00000294, 0.00000092, 0.00000887, 0.00001462, 0.00000953, -0.00000612, -0.00002036, -0.00002093, -0.00000834, 0.00000497, 0.00000873, 0.00000528, 0.00000409, 0.00000826, 0.00001059, 0.00000410, -0.00000708, -0.00001146, -0.00000447, 0.00000502, 0.00000496, -0.00000518, -0.00001378, -0.00001131, -0.00000169, 0.00000378, 0.00000162, 0.00000054, 0.00000875, 0.00002051, 0.00002080, 0.00000409, -0.00001592, -0.00001945, -0.00000328, 0.00001390, 0.00001128, -0.00001075, -0.00003009, -0.00002642, -0.00000268, 0.00001842, 0.00001828, 0.00000015, -0.00001671, -0.00001792, -0.00000654, 0.00000365, 0.00000395, -0.00000228, -0.00000624, -0.00000402, 0.00000089, 0.00000333, 0.00000212, -0.00000009, -0.00000078, 0.00000014, 0.00000136, 0.00000208, 0.00000223, 0.00000169, 0.00000022, -0.00000185, -0.00000351, -0.00000398, -0.00000349, -0.00000281, -0.00000227, -0.00000151, -0.00000036, 0.00000059, 0.00000051, -0.00000054, -0.00000145, -0.00000116, -0.00000007, 0.00000002, },
		{ -0.00000898, -0.00000480, -0.00000078, -0.00000167, -0.00000841, -0.00001572, -0.00001584, -0.00000595, 0.00000770, 0.00001457, 0.00000966, -0.00000148, -0.00000857, -0.00000760, -0.00000464, -0.00000779, -0.00001631, -0.00002004, -0.00001091, 0.00000660, 0.00001872, 0.00001657, 0.00000548, -0.00000154, 0.00000161, 0.00000822, 0.00000785, -0.00000105, -0.00000995, -0.00001099, -0.00000578, -0.00000177, -0.00000201, -0.00000198, 0.00000271, 0.00000875, 0.00000851, 0.00000046, -0.00000710, -0.00000509, 0.00000524, 0.00001294, 0.00000957, -0.00000154, -0.00000963, -0.00000855, -0.00000179, 0.00000396, 0.00000681, 0.00000876, 0.00000894, 0.00000267, -0.00001083, -0.00002270, -0.00002121, -0.00000514, 0.00001269, 0.00001783, 0.00000940, -0.00000011, -0.00000012, 0.00000595, 0.00000648, -0.00000250, -0.00001171, -0.00001043, -0.00000052, 0.00000533, -0.00000049, -0.00001080, -0.00001243, -0.00000266, 0.00000766, 0.00000828, 0.00000265, 0.00000284, 0.00001234, 0.00001944, 0.00001107, -0.00000920, -0.00002231, -0.00001476, 0.00000459, 0.00001314, -0.00000018, -0.00002163, -0.00002716, -0.00000884, 0.00001622, 0.00002398, 0.00000823, -0.00001456, -0.00002359, -0.00001399, 0.00000149, 0.00000815, 0.00000343, -0.00000405, -0.00000599, -0.00000227, 0.00000173, 0.00000237, 0.00000071, -0.00000035, 0.00000019, 0.00000130, 0.00000199, 0.00000222, 0.00000203, 0.00000095, -0.00000125, -0.00000363, -0.00000480, -0.00000431, -0.00000308, -0.00000219, -0.00000172, -0.00000107, -0.00000009, 0.00000054, 0.00000020, -0.00000071, -0.00000122, -0.00000110, -0.00000139, -0.00000341, -0.00000694, -0.00000968, },
		{ 0.00000042, -0.00000470, -0.00001243, -0.00001559, -0.00000916, 0.00000393, 0.00001387, 0.00001234, 0.00000054, -0.00001159, -0.00001515, -0.00001091, -0.00000732, -0.00000987, -0.00001412, -0.00001088, 0.00000208, 0.00001590, 0.00001948, 0.00001154, 0.00000175, -0.00000054, 0.00000343, 0.00000454, -0.00000218, -0.00001107, -0.00001270, -0.00000529, 0.00000340, 0.00000560, 0.00000221, 0.00000008, 0.00000229, 0.00000448, 0.00000184, -0.00000317, -0.00000325, 0.00000378, 0.00001063, 0.00000854, -0.00000187, -0.00001041, -0.00000854, 0.00000145, 0.00000989, 0.00001066, 0.00000591, 0.00000032, -0.00000558, -0.00001328, -0.00001944, -0.00001615, -0.00000084, 0.00001674, 0.00002201, 0.00001112, -0.00000377, -0.00000751, 0.00000109, 0.00000850, 0.00000331, -0.00000979, -0.00001584, -0.00000820, 0.00000335, 0.00000493, -0.00000412, -0.00001070, -0.00000481, 0.00000778, 0.00001300, 0.00000666, -0.00000026, 0.00000362, 0.00001402, 0.00001492, -0.00000095, -0.00002064, -0.00002406, -0.00000752, 0.00001063, 0.00001011, -0.00000802, -0.00002215, -0.00001451, 0.00000864, 0.00002356, 0.00001477, -0.00000899, -0.00002496, -0.00002006, -0.00000223, 0.00001054, 0.00000899, -0.00000091, -0.00000754, -0.00000606, -0.00000071, 0.00000242, 0.00000193, 0.00000047, 0.00000032, 0.00000121, 0.00000191, 0.00000210, 0.00000201, 0.00000128, -0.00000072, -0.00000350, -0.00000535, -0.00000507, -0.00000332, -0.00000182, -0.00000147, -0.00000157, -0.00000107, -0.00000004, 0.00000050, -0.00000003, -0.00000115, -0.00000210, -0.00000295, -0.00000451, -0.00000697, -0.00000897, -0.00000838, -0.00000464, -0.00000028, },
		{ -0.00001307, -0.00001035, -0.00000033, 0.00000982, 0.00001113, 0.00000126, -0.00001245, -0.00001931, -0.00001549, -0.00000738, -0.00000404, -0.00000697, -0.00000868, -0.00000200, 0.00001060, 0.00001887, 0.00001599, 0.00000597, -0.00000138, -0.00000169, 0.00000002, -0.00000281, -0.00000904, -0.00001086, -0.00000400, 0.00000607, 0.00000989, 0.00000492, -0.00000201, -0.00000327, 0.00000078, 0.00000364, 0.00000189, -0.00000063, 0.00000106, 0.00000531, 0.00000548, -0.00000082, -0.00000714, -0.00000526, 0.00000448, 0.00001262, 0.00001093, 0.00000107, -0.00000850, -0.00001253, -0.00001274, -0.00001183, -0.00000759, 0.00000314, 0.00001618, 0.00002058, 0.00001057, -0.00000535, -0.00001159, -0.00000282, 0.00000898, 0.00000794, -0.00000682, -0.00001946, -0.00001583, -0.00000024, 0.00000982, 0.00000472, -0.00000627, -0.00000724, 0.00000433, 0.00001467, 0.00001163, 0.00000031, -0.00000359, 0.00000535, 0.00001393, 0.00000633, -0.00001473, -0.00002845, -0.00001929, 0.00000407, 0.00001734, 0.00000769, -0.00001190, -0.00001775, -0.00000252, 0.00001676, 0.00001819, -0.00000071, -0.00002096, -0.00002315, -0.00000717, 0.00000992, 0.00001297, 0.00000286, -0.00000787, -0.00000961, -0.00000379, 0.00000201, 0.00000330, 0.00000165, 0.00000057, 0.00000107, 0.00000185, 0.00000196, 0.00000164, 0.00000103, -0.00000049, -0.00000306, -0.00000528, -0.00000543, -0.00000352, -0.00000148, -0.00000105, -0.00000189, -0.00000225, -0.00000122, 0.00000022, 0.00000050, -0.00000079, -0.00000271, -0.00000428, -0.00000550, -0.00000676, -0.00000773, -0.00000716, -0.00000429, -0.00000045, 0.00000117, -0.00000192, -0.00000845, },
		{ 0.00000326, 0.00000615, -0.00000006, -0.00001168, -0.00001938, -0.00001681, -0.00000708, 0.00000029, -0.00000054, -0.00000541, -0.00000506, 0.00000400, 0.00001492, 0.00001777, 0.00001030, -0.00000011, -0.00000515, -0.00000429, -0.00000329, -0.00000535, -0.00000668, -0.00000224, 0.00000615, 0.00001053, 0.00000607, -0.00000257, -0.00000625, -0.00000179, 0.00000475, 0.00000588, 0.00000150, -0.00000211, -0.00000097, 0.00000190, 0.00000170, -0.00000073, 0.00000014, 0.00000615, 0.00001112, 0.00000741, -0.00000433, -0.00001479, -0.00001601, -0.00000943, -0.00000236, 0.00000169, 0.00000566, 0.00001157, 0.00001461, 0.00000848, -0.00000417, -0.00001138, -0.00000517, 0.00000716, 0.00000952, -0.00000396, -0.00002025, -0.00002107, -0.00000425, 0.00001293, 0.00001339, -0.00000006, -0.00000880, -0.00000127, 0.00001309, 0.00001616, 0.00000404, -0.00000796, -0.00000498, 0.00000778, 0.00001059, -0.00000577, -0.00002631, -0.00002757, -0.00000521, 0.00001928, 0.00002169, 0.00000183, -0.00001693, -0.00001429, 0.00000483, 0.00001731, 0.00000858, -0.00001179, -0.00002204, -0.00001241, 0.00000575, 0.00001410, 0.00000667, -0.00000639, -0.00001194, -0.00000710, 0.00000084, 0.00000442, 0.00000310, 0.00000102, 0.00000086, 0.00000170, 0.00000179, 0.00000104, 0.00000025, -0.00000065, -0.00000234, -0.00000436, -0.00000507, -0.00000366, -0.00000154, -0.00000092, -0.00000216, -0.00000338, -0.00000270, -0.00000057, 0.00000080, -0.00000013, -0.00000271, -0.00000511, -0.00000627, -0.00000649, -0.00000638, -0.00000573, -0.00000384, -0.00000094, 0.00000113, 0.00000011, -0.00000431, -0.00000913, -0.00000982, -0.00000447, },
		{ -0.00000998, -0.00001561, -0.00001411, -0.00000555, 0.00000280, 0.00000354, -0.00000267, -0.00000717, -0.00000256, 0.00000883, 0.00001682, 0.00001402, 0.00000337, -0.00000563, -0.00000720, -0.00000404, -0.00000191, -0.00000190, -0.00000028, 0.00000461, 0.00000846, 0.00000599, -0.00000147, -0.00000624, -0.00000306, 0.00000431, 0.00000724, 0.00000237, -0.00000459, -0.00000590, -0.00000072, 0.00000505, 0.00000671, 0.00000580, 0.00000587, 0.00000591, 0.00000126, -0.00000865, -0.00001676, -0.00001513, -0.00000446, 0.00000605, 0.00000895, 0.00000619, 0.00000477, 0.00000643, 0.00000578, -0.00000070, -0.00000748, -0.00000594, 0.00000302, 0.00000738, -0.00000198, -0.00001780, -0.00002224, -0.00000745, 0.00001338, 0.00001954, 0.00000653, -0.00000848, -0.00000690, 0.00000894, 0.00001872, 0.00000919, -0.00000920, -0.00001489, -0.00000220, 0.00001086, 0.00000437, -0.00001774, -0.00003016, -0.00001524, 0.00001497, 0.00003055, 0.00001641, -0.00001124, -0.00002335, -0.00000970, 0.00001170, 0.00001654, 0.00000102, -0.00001621, -0.00001660, -0.00000155, 0.00001150, 0.00000962, -0.00000297, -0.00001214, -0.00001002, -0.00000127, 0.00000480, 0.00000461, 0.00000178, 0.00000061, 0.00000130, 0.00000152, 0.00000040, -0.00000077, -0.00000111, -0.00000144, -0.00000263, -0.00000389, -0.00000373, -0.00000230, -0.00000151, -0.00000255, -0.00000414, -0.00000396, -0.00000157, 0.00000080, 0.00000064, -0.00000214, -0.00000530, -0.00000673, -0.00000628, -0.00000523, -0.00000440, -0.00000337, -0.00000146, 0.00000081, 0.00000164, -0.00000044, -0.00000475, -0.00000833, -0.00000823, -0.00000466, -0.00000160, -0.00000341, },
		{ -0.00000240, 0.00000359, 0.00000426, -0.00000190, -0.00000873, -0.00000779, 0.00000249, 0.00001404, 0.00001665, 0.00000833, -0.00000301, -0.00000809, -0.00000524, -0.00000019, 0.00000190, 0.00000177, 0.00000293, 0.00000533, 0.00000510, 0.00000056, -0.00000411, -0.00000337, 0.00000219, 0.00000566, 0.00000197, -0.00000540, -0.00000783, -0.00000153, 0.00000812, 0.00001275, 0.00000988, 0.00000384, -0.00000112, -0.00000544, -0.00001072, -0.00001420, -0.00001048, 0.00000062, 0.00001112, 0.00001238, 0.00000498, -0.00000186, -0.00000130, 0.00000329, 0.00000369, -0.00000166, -0.00000556, -0.00000255, 0.00000218, -0.00000114, -0.00001252, -0.00001888, -0.00000895, 0.00001101, 0.00002157, 0.00001205, -0.00000578, -0.00001048, 0.00000343, 0.00001810, 0.00001342, -0.00000791, -0.00002231, -0.00001338, 0.00000744, 0.00001356, -0.00000461, -0.00002639, -0.00002382, 0.00000514, 0.00003210, 0.00002892, -0.00000112, -0.00002697, -0.00002367, 0.00000190, 0.00002093, 0.00001499, -0.00000606, -0.00001824, -0.00001067, 0.00000502, 0.00001070, 0.00000199, -0.00000962, -0.00001179, -0.00000420, 0.00000396, 0.00000581, 0.00000286, 0.00000044, 0.00000055, 0.00000100, -0.00000008, -0.00000157, -0.00000158, -0.00000053, -0.00000049, -0.00000210, -0.00000371, -0.00000379, -0.00000302, -0.00000314, -0.00000426, -0.00000445, -0.00000242, 0.00000039, 0.00000113, -0.00000125, -0.00000482, -0.00000674, -0.00000614, -0.00000447, -0.00000340, -0.00000291, -0.00000175, 0.00000064, 0.00000294, 0.00000283, -0.00000089, -0.00000683, -0.00001168, -0.00001277, -0.00001051, -0.00000801, -0.00000784, -0.00000899, -0.00000776, },
		{ -0.00000377, -0.00001011, -0.00001074, -0.00000214, 0.00001079, 0.00001788, 0.00001332, 0.00000160, -0.00000701, -0.00000665, -0.00000062, 0.00000386, 0.00000385, 0.00000239, 0.00000275, 0.00000376, 0.00000227, -0.00000143, -0.00000325, -0.00000094, 0.00000209, 0.00000083, -0.00000422, -0.00000637, -0.00000054, 0.00000980, 0.00001549, 0.00001128, 0.00000101, -0.00000747, -0.00001053, -0.00001017, -0.00000841, -0.00000387, 0.00000438, 0.00001188, 0.00001163, 0.00000279, -0.00000628, -0.00000647, 0.00000140, 0.00000723, 0.00000381, -0.00000456, -0.00000804, -0.00000430, -0.00000129, -0.00000564, -0.00001191, -0.00000841, 0.00000650, 0.00001912, 0.00001537, -0.00000098, -0.00001075, -0.00000205, 0.00001383, 0.00001463, -0.00000529, -0.00002562, -0.00002263, 0.00000173, 0.00001983, 0.00000987, -0.00001733, -0.00002898, -0.00000790, 0.00002610, 0.00003679, 0.00001165, -0.00002391, -0.00003398, -0.00001043, 0.00002027, 0.00002720, 0.00000701, -0.00001624, -0.00001968, -0.00000467, 0.00000910, 0.00000754, -0.00000429, -0.00001162, -0.00000756, 0.00000156, 0.00000615, 0.00000412, 0.00000052, -0.00000047, 0.00000017, -0.00000028, -0.00000170, -0.00000172, 0.00000018, 0.00000144, -0.00000019, -0.00000360, -0.00000571, -0.00000527, -0.00000395, -0.00000363, -0.00000384, -0.00000280, -0.00000045, 0.00000097, -0.00000043, -0.00000375, -0.00000615, -0.00000595, -0.00000417, -0.00000285, -0.00000250, -0.00000167, 0.00000086, 0.00000414, 0.00000525, 0.00000170, -0.00000621, -0.00001499, -0.00002003, -0.00001872, -0.00001248, -0.00000545, -0.00000101, 0.00000080, 0.00000196, 0.00000298, 0.00000162, },
		{ -0.00000373, 0.00000830, 0.00001748, 0.00001661, 0.00000640, -0.00000449, -0.00000756, -0.00000244, 0.00000399, 0.00000575, 0.00000344, 0.00000161, 0.00000218, 0.00000257, 0.00000027, -0.00000308, -0.00000389, -0.00000180, -0.00000030, -0.00000153, -0.00000246, 0.00000149, 0.00000949, 0.00001435, 0.00000995, -0.00000143, -0.00001101, -0.00001251, -0.00000752, -0.00000166, 0.00000244, 0.00000597, 0.00000889, 0.00000773, 0.00000060, -0.00000746, -0.00000814, 0.00000011, 0.00000852, 0.00000709, -0.00000341, -0.00001186, -0.00000993, -0.00000197, 0.00000118, -0.00000323, -0.00000603, 0.00000119, 0.00001316, 0.00001592, 0.00000481, -0.00000769, -0.00000638, 0.00000644, 0.00001177, -0.00000263, -0.00002414, -0.00002728, -0.00000436, 0.00002190, 0.00002206, -0.00000536, -0.00002946, -0.00002077, 0.00001426, 0.00003838, 0.00002423, -0.00001473, -0.00003831, -0.00002280, 0.00001442, 0.00003506, 0.00002077, -0.00001036, -0.00002651, -0.00001603, 0.00000453, 0.00001251, 0.00000331, -0.00000904, -0.00001077, -0.00000254, 0.00000507, 0.00000533, 0.00000114, -0.00000141, -0.00000089, -0.00000022, -0.00000097, -0.00000125, 0.00000055, 0.00000249, 0.00000121, -0.00000337, -0.00000745, -0.00000768, -0.00000488, -0.00000248, -0.00000218, -0.00000253, -0.00000164, -0.00000006, -0.00000003, -0.00000225, -0.00000483, -0.00000552, -0.00000427, -0.00000281, -0.00000211, -0.00000111, 0.00000157, 0.00000516, 0.00000651, 0.00000251, -0.00000698, -0.00001824, -0.00002551, -0.00002451, -0.00001534, -0.00000282, 0.00000660, 0.00000929, 0.00000634, 0.00000133, -0.00000340, -0.00000777, -0.00001128, -0.00001092, },
		{ 0.00001685, 0.00000960, -0.00000117, -0.00000697, -0.00000410, 0.00000301, 0.00000704, 0.00000545, 0.00000189, 0.00000056, 0.00000111, 0.00000021, -0.00000303, -0.00000553, -0.00000440, -0.00000086, 0.00000158, 0.00000213, 0.00000355, 0.00000737, 0.00001014, 0.00000678, -0.00000251, -0.00001074, -0.00001103, -0.00000390, 0.00000375, 0.00000654, 0.00000535, 0.00000383, 0.00000250, -0.00000087, -0.00000565, -0.00000669, -0.00000079, 0.00000701, 0.00000729, -0.00000240, -0.00001288, -0.00001294, -0.00000265, 0.00000636, 0.00000490, -0.00000240, -0.00000335, 0.00000557, 0.00001376, 0.00000995, -0.00000249, -0.00000893, -0.00000261, 0.00000518, -0.00000091, -0.00001836, -0.00002599, -0.00000896, 0.00001947, 0.00002895, 0.00000655, -0.00002500, -0.00003003, -0.00000030, 0.00003331, 0.00003349, -0.00000176, -0.00003568, -0.00003232, 0.00000472, 0.00003692, 0.00003247, -0.00000144, -0.00002950, -0.00002689, -0.00000246, 0.00001576, 0.00001202, -0.00000411, -0.00001321, -0.00000807, 0.00000223, 0.00000619, 0.00000256, -0.00000176, -0.00000205, -0.00000013, 0.00000047, -0.00000009, 0.00000058, 0.00000222, 0.00000154, -0.00000304, -0.00000832, -0.00000947, -0.00000575, -0.00000122, 0.00000004, -0.00000163, -0.00000288, -0.00000189, -0.00000032, -0.00000064, -0.00000283, -0.00000466, -0.00000459, -0.00000321, -0.00000174, -0.00000006, 0.00000264, 0.00000572, 0.00000626, 0.00000129, -0.00000918, -0.00002111, -0.00002845, -0.00002663, -0.00001549, -0.00000001, 0.00001216, 0.00001525, 0.00000917, -0.00000090, -0.00000891, -0.00001222, -0.00001170, -0.00000850, -0.00000219, 0.00000702, 0.00001531, },
		{ -0.00000408, -0.00000399, 0.00000181, 0.00000712, 0.00000705, 0.00000279, -0.00000084, -0.00000151, -0.00000138, -0.00000304, -0.00000535, -0.00000486, -0.00000069, 0.00000392, 0.00000560, 0.00000478, 0.00000431, 0.00000470, 0.00000316, -0.00000191, -0.00000710, -0.00000691, -0.00000054, 0.00000631, 0.00000751, 0.00000318, -0.00000125, -0.00000213, -0.00000133, -0.00000211, -0.00000344, -0.00000158, 0.00000319, 0.00000468, -0.00000161, -0.00001085, -0.00001251, -0.00000288, 0.00000891, 0.00001062, 0.00000163, -0.00000582, -0.00000156, 0.00000942, 0.00001278, 0.00000291, -0.00000956, -0.00001125, -0.00000347, -0.00000058, -0.00000984, -0.00001915, -0.00001085, 0.00001338, 0.00002908, 0.00001568, -0.00001645, -0.00003318, -0.00001397, 0.00002259, 0.00003678, 0.00001166, -0.00002665, -0.00003649, -0.00000633, 0.00003240, 0.00003951, 0.00000866, -0.00002777, -0.00003481, -0.00001049, 0.00001652, 0.00002016, 0.00000239, -0.00001443, -0.00001435, -0.00000227, 0.00000653, 0.00000497, -0.00000105, -0.00000312, -0.00000042, 0.00000208, 0.00000166, 0.00000046, 0.00000066, 0.00000053, -0.00000272, -0.00000781, -0.00000986, -0.00000628, -0.00000042, 0.00000208, -0.00000033, -0.00000380, -0.00000407, -0.00000133, 0.00000069, -0.00000042, -0.00000326, -0.00000480, -0.00000382, -0.00000137, 0.00000129, 0.00000377, 0.00000543, 0.00000425, -0.00000183, -0.00001225, -0.00002290, -0.00002823, -0.00002463, -0.00001273, 0.00000259, 0.00001429, 0.00001671, 0.00000918, -0.00000314, -0.00001266, -0.00001477, -0.00001059, -0.00000435, 0.00000126, 0.00000644, 0.00001137, 0.00001362, 0.00001027, 0.00000243, },
		{ 0.00000549, 0.00000677, 0.00000328, -0.00000163, -0.00000394, -0.00000348, -0.00000301, -0.00000370, -0.00000338, -0.00000000, 0.00000474, 0.00000687, 0.00000488, 0.00000151, 0.00000005, 0.00000036, -0.00000007, -0.00000168, -0.00000172, 0.00000169, 0.00000575, 0.00000558, 0.00000044, -0.00000470, -0.00000483, -0.00000097, 0.00000150, -0.00000004, -0.00000238, -0.00000165, 0.00000047, -0.00000104, -0.00000648, -0.00000890, -0.00000237, 0.00000876, 0.00001297, 0.00000520, -0.00000570, -0.00000658, 0.00000379, 0.00001218, 0.00000658, -0.00000860, -0.00001739, -0.00001180, -0.00000154, -0.00000081, -0.00000875, -0.00000966, 0.00000538, 0.00002290, 0.00002024, -0.00000563, -0.00002939, -0.00002360, 0.00000842, 0.00003287, 0.00002206, -0.00001321, -0.00003388, -0.00001583, 0.00002255, 0.00004002, 0.00001755, -0.00002148, -0.00003766, -0.00001757, 0.00001460, 0.00002587, 0.00000900, -0.00001427, -0.00002030, -0.00000771, 0.00000636, 0.00000825, 0.00000092, -0.00000403, -0.00000156, 0.00000314, 0.00000373, 0.00000062, -0.00000166, -0.00000176, -0.00000261, -0.00000585, -0.00000836, -0.00000618, -0.00000046, 0.00000311, 0.00000097, -0.00000400, -0.00000593, -0.00000287, 0.00000131, 0.00000192, -0.00000130, -0.00000446, -0.00000426, -0.00000105, 0.00000254, 0.00000442, 0.00000388, 0.00000044, -0.00000630, -0.00001518, -0.00002264, -0.00002452, -0.00001899, -0.00000793, 0.00000413, 0.00001220, 0.00001288, 0.00000599, -0.00000462, -0.00001260, -0.00001329, -0.00000732, 0.00000014, 0.00000441, 0.00000532, 0.00000594, 0.00000778, 0.00000869, 0.00000608, 0.00000125, -0.00000126, 0.00000109, },
		{ -0.00000143, -0.00000471, -0.00000479, -0.00000297, -0.00000172, -0.00000108, 0.00000080, 0.00000409, 0.00000598, 0.00000407, -0.00000010, -0.00000241, -0.00000092, 0.00000212, 0.00000345, 0.00000287, 0.00000254, 0.00000297, 0.00000191, -0.00000197, -0.00000580, -0.00000523, -0.00000025, 0.00000377, 0.00000216, -0.00000314, -0.00000597, -0.00000376, -0.00000061, -0.00000110, -0.00000329, -0.00000102, 0.00000668, 0.00001208, 0.00000758, -0.00000343, -0.00000879, -0.00000212, 0.00000800, 0.00000725, -0.00000663, -0.00001960, -0.00001749, -0.00000327, 0.00000645, 0.00000230, -0.00000590, -0.00000229, 0.00001261, 0.00001975, 0.00000506, -0.00001980, -0.00002732, -0.00000624, 0.00002235, 0.00002686, 0.00000165, -0.00002475, -0.00002129, 0.00000955, 0.00003353, 0.00002301, -0.00001178, -0.00003415, -0.00002162, 0.00001043, 0.00002752, 0.00001392, -0.00001289, -0.00002459, -0.00001279, 0.00000590, 0.00001195, 0.00000381, -0.00000486, -0.00000385, 0.00000306, 0.00000588, 0.00000159, -0.00000388, -0.00000494, -0.00000304, -0.00000287, -0.00000491, -0.00000512, -0.00000140, 0.00000255, 0.00000175, -0.00000329, -0.00000677, -0.00000454, 0.00000093, 0.00000362, 0.00000108, -0.00000316, -0.00000412, -0.00000088, 0.00000306, 0.00000395, 0.00000089, -0.00000472, -0.00001113, -0.00001675, -0.00001952, -0.00001758, -0.00001109, -0.00000286, 0.00000350, 0.00000595, 0.00000455, 0.00000049, -0.00000443, -0.00000766, -0.00000680, -0.00000194, 0.00000337, 0.00000494, 0.00000233, -0.00000031, 0.00000121, 0.00000619, 0.00000970, 0.00000822, 0.00000370, 0.00000102, 0.00000207, 0.00000387, 0.00000264, },
		{ -0.00000304, -0.00000072, 0.00000063, 0.00000141, 0.00000278, 0.00000396, 0.00000294, -0.00000024, -0.00000241, -0.00000072, 0.00000368, 0.00000664, 0.00000566, 0.00000232, -0.00000044, -0.00000194, -0.00000348, -0.00000488, -0.00000384, 0.00000035, 0.00000404, 0.00000253, -0.00000374, -0.00000853, -0.00000662, -0.00000021, 0.00000385, 0.00000274, 0.00000106, 0.00000392, 0.00000889, 0.00000827, -0.00000021, -0.00000850, -0.00000734, 0.00000121, 0.00000473, -0.00000433, -0.00001755, -0.00001910, -0.00000510, 0.00001030, 0.00001114, -0.00000069, -0.00000757, 0.00000136, 0.00001503, 0.00001320, -0.00000714, -0.00002488, -0.00001837, 0.00000758, 0.00002499, 0.00001457, -0.00001107, -0.00002140, -0.00000376, 0.00002123, 0.00002364, -0.00000055, -0.00002439, -0.00002117, 0.00000496, 0.00002423, 0.00001551, -0.00001073, -0.00002596, -0.00001597, 0.00000551, 0.00001525, 0.00000674, -0.00000588, -0.00000726, 0.00000169, 0.00000797, 0.00000380, -0.00000508, -0.00000846, -0.00000437, 0.00000033, 0.00000011, -0.00000276, -0.00000287, 0.00000017, 0.00000147, -0.00000183, -0.00000611, -0.00000574, -0.00000043, 0.00000423, 0.00000358, -0.00000069, -0.00000308, -0.00000107, 0.00000216, 0.00000178, -0.00000338, -0.00001023, -0.00001496, -0.00001589, -0.00001332, -0.00000839, -0.00000296, 0.00000036, -0.00000015, -0.00000341, -0.00000600, -0.00000528, -0.00000183, 0.00000181, 0.00000395, 0.00000466, 0.00000406, 0.00000135, -0.00000320, -0.00000633, -0.00000383, 0.00000454, 0.00001314, 0.00001515, 0.00000936, 0.00000134, -0.00000249, -0.00000122, 0.00000082, -0.00000009, -0.00000299, -0.00000442, },
		{ 0.00000174, 0.00000196, 0.00000227, 0.00000208, 0.00000063, -0.00000077, 0.00000032, 0.00000407, 0.00000718, 0.00000623, 0.00000160, -0.00000297, -0.00000466, -0.00000400, -0.00000293, -0.00000168, 0.00000052, 0.00000253, 0.00000123, -0.00000405, -0.00000881, -0.00000734, 0.00000028, 0.00000729, 0.00000774, 0.00000354, 0.00000167, 0.00000475, 0.00000712, 0.00000247, -0.00000673, -0.00001110, -0.00000642, -0.00000008, -0.00000236, -0.00001215, -0.00001649, -0.00000637, 0.00001020, 0.00001579, 0.00000461, -0.00000917, -0.00000774, 0.00000775, 0.00001705, 0.00000516, -0.00001755, -0.00002552, -0.00000793, 0.00001713, 0.00002261, 0.00000379, -0.00001639, -0.00001454, 0.00000581, 0.00001936, 0.00001004, -0.00000996, -0.00001578, -0.00000058, 0.00001625, 0.00001289, -0.00000830, -0.00002358, -0.00001593, 0.00000550, 0.00001714, 0.00000851, -0.00000737, -0.00001130, -0.00000056, 0.00001001, 0.00000734, -0.00000472, -0.00001178, -0.00000681, 0.00000284, 0.00000602, 0.00000109, -0.00000419, -0.00000378, -0.00000038, -0.00000015, -0.00000377, -0.00000578, -0.00000233, 0.00000352, 0.00000572, 0.00000274, -0.00000110, -0.00000179, -0.00000069, -0.00000234, -0.00000828, -0.00001469, -0.00001643, -0.00001210, -0.00000474, 0.00000132, 0.00000321, 0.00000005, -0.00000694, -0.00001394, -0.00001558, -0.00000886, 0.00000342, 0.00001390, 0.00001644, 0.00001086, 0.00000174, -0.00000619, -0.00001076, -0.00001097, -0.00000556, 0.00000503, 0.00001599, 0.00001999, 0.00001362, 0.00000163, -0.00000688, -0.00000689, -0.00000159, 0.00000208, 0.00000092, -0.00000228, -0.00000335, -0.00000154, 0.00000075, },
		{ 0.00000178, 0.00000152, 0.00000102, 0.00000134, 0.00000334, 0.00000552, 0.00000498, 0.00000092, -0.00000379, -0.00000550, -0.00000371, -0.00000108, 0.00000013, 0.00000019, 0.00000004, -0.00000103, -0.00000404, -0.00000709, -0.00000587, 0.00000106, 0.00000887, 0.00001079, 0.00000589, 0.00000061, 0.00000087, 0.00000437, 0.00000340, -0.00000472, -0.00001296, -0.00001287, -0.00000563, -0.00000118, -0.00000520, -0.00001074, -0.00000663, 0.00000684, 0.00001572, 0.00000877, -0.00000697, -0.00001254, 0.00000000, 0.00001597, 0.00001399, -0.00000764, -0.00002644, -0.00002051, 0.00000536, 0.00002392, 0.00001600, -0.00000796, -0.00002058, -0.00000912, 0.00001144, 0.00001794, 0.00000627, -0.00000636, -0.00000488, 0.00000509, 0.00000634, -0.00000608, -0.00001746, -0.00001209, 0.00000601, 0.00001673, 0.00000800, -0.00000948, -0.00001507, -0.00000278, 0.00001209, 0.00001184, -0.00000280, -0.00001449, -0.00001026, 0.00000401, 0.00001204, 0.00000637, -0.00000466, -0.00000877, -0.00000408, 0.00000105, 0.00000007, -0.00000400, -0.00000399, 0.00000155, 0.00000682, 0.00000643, 0.00000150, -0.00000309, -0.00000545, -0.00000812, -0.00001285, -0.00001663, -0.00001451, -0.00000571, 0.00000450, 0.00000930, 0.00000568, -0.00000435, -0.00001583, -0.00002306, -0.00002107, -0.00000823, 0.00001071, 0.00002553, 0.00002713, 0.00001484, -0.00000292, -0.00001573, -0.00001865, -0.00001331, -0.00000353, 0.00000782, 0.00001759, 0.00002103, 0.00001476, 0.00000164, -0.00000946, -0.00001094, -0.00000365, 0.00000422, 0.00000556, 0.00000087, -0.00000378, -0.00000398, -0.00000082, 0.00000191, 0.00000250, 0.00000204, },
		{ 0.00000165, 0.00000203, 0.00000287, 0.00000279, 0.00000048, -0.00000293, -0.00000450, -0.00000288, -0.00000008, 0.00000087, -0.00000058, -0.00000244, -0.00000329, -0.00000372, -0.00000414, -0.00000277, 0.00000227, 0.00000890, 0.00001160, 0.00000751, 0.00000069, -0.00000198, 0.00000066, 0.00000215, -0.00000345, -0.00001289, -0.00001659, -0.00001029, -0.00000104, 0.00000117, -0.00000374, -0.00000570, 0.00000187, 0.00001178, 0.00001093, -0.00000206, -0.00001246, -0.00000634, 0.00001070, 0.00001752, 0.00000222, -0.00002142, -0.00002744, -0.00000744, 0.00001821, 0.00002232, 0.00000133, -0.00002086, -0.00002000, 0.00000212, 0.00002173, 0.00002081, 0.00000507, -0.00000689, -0.00000670, -0.00000266, -0.00000443, -0.00000864, -0.00000494, 0.00000692, 0.00001361, 0.00000465, -0.00001209, -0.00001748, -0.00000390, 0.00001427, 0.00001647, 0.00000004, -0.00001634, -0.00001420, 0.00000375, 0.00001741, 0.00001263, -0.00000387, -0.00001407, -0.00000954, 0.00000114, 0.00000500, 0.00000007, -0.00000450, -0.00000140, 0.00000612, 0.00000920, 0.00000414, -0.00000469, -0.00001139, -0.00001463, -0.00001603, -0.00001507, -0.00000898, 0.00000207, 0.00001207, 0.00001337, 0.00000368, -0.00001179, -0.00002458, -0.00002826, -0.00002045, -0.00000281, 0.00001850, 0.00003333, 0.00003254, 0.00001533, -0.00000823, -0.00002399, -0.00002444, -0.00001308, 0.00000108, 0.00001164, 0.00001720, 0.00001772, 0.00001222, 0.00000151, -0.00000885, -0.00001149, -0.00000460, 0.00000519, 0.00000883, 0.00000381, -0.00000391, -0.00000678, -0.00000323, 0.00000208, 0.00000442, 0.00000349, 0.00000196, 0.00000150, 0.00000162, },
		{ 0.00000061, 0.00000019, -0.00000115, -0.00000227, -0.00000171, -0.00000006, 0.00000031, -0.00000167, -0.00000421, -0.00000476, -0.00000307, -0.00000087, 0.00000107, 0.00000396, 0.00000806, 0.00001047, 0.00000783, 0.00000127, -0.00000359, -0.00000313, -0.00000083, -0.00000333, -0.00001131, -0.00001697, -0.00001289, -0.00000187, 0.00000526, 0.00000251, -0.00000371, -0.00000278, 0.00000581, 0.00001078, 0.00000369, -0.00000839, -0.00000996, 0.00000308, 0.00001558, 0.00000973, -0.00001225, -0.00002767, -0.00001843, 0.00000689, 0.00002108, 0.00000874, -0.00001584, -0.00002437, -0.00000610, 0.00002092, 0.00003050, 0.00001597, -0.00000595, -0.00001609, -0.00001175, -0.00000354, 0.00000100, 0.00000390, 0.00000788, 0.00000814, -0.00000116, -0.00001470, -0.00001770, -0.00000316, 0.00001638, 0.00002012, 0.00000285, -0.00001719, -0.00001767, 0.00000253, 0.00002157, 0.00001899, -0.00000192, -0.00001894, -0.00001616, -0.00000007, 0.00001048, 0.00000647, -0.00000322, -0.00000500, 0.00000289, 0.00000979, 0.00000620, -0.00000592, -0.00001701, -0.00002051, -0.00001705, -0.00000996, -0.00000070, 0.00000943, 0.00001572, 0.00001212, -0.00000220, -0.00001987, -0.00003035, -0.00002773, -0.00001358, 0.00000617, 0.00002459, 0.00003467, 0.00003058, 0.00001211, -0.00001190, -0.00002750, -0.00002593, -0.00001083, 0.00000573, 0.00001427, 0.00001436, 0.00001097, 0.00000702, 0.00000173, -0.00000462, -0.00000788, -0.00000420, 0.00000400, 0.00000893, 0.00000545, -0.00000303, -0.00000803, -0.00000506, 0.00000220, 0.00000678, 0.00000581, 0.00000227, 0.00000021, 0.00000040, 0.00000099, 0.00000086, 0.00000059, },
		{ 0.00000033, -0.00000028, -0.00000060, -0.00000123, -0.00000302, -0.00000505, -0.00000499, -0.00000197, 0.00000210, 0.00000475, 0.00000587, 0.00000699, 0.00000803, 0.00000658, 0.00000150, -0.00000416, -0.00000609, -0.00000438, -0.00000420, -0.00000896, -0.00001451, -0.00001307, -0.00000328, 0.00000633, 0.00000656, -0.00000107, -0.00000556, -0.00000014, 0.00000852, 0.00000826, -0.00000230, -0.00001037, -0.00000444, 0.00000962, 0.00001343, -0.00000174, -0.00002210, -0.00002547, -0.00000725, 0.00001270, 0.00001217, -0.00000736, -0.00002164, -0.00001120, 0.00001609, 0.00003339, 0.00002398, -0.00000200, -0.00002043, -0.00001837, -0.00000344, 0.00000907, 0.00001215, 0.00000849, 0.00000148, -0.00000811, -0.00001662, -0.00001541, -0.00000043, 0.00001803, 0.00002177, 0.00000472, -0.00001688, -0.00001957, 0.00000121, 0.00002410, 0.00002423, 0.00000056, -0.00002273, -0.00002280, -0.00000219, 0.00001597, 0.00001461, -0.00000011, -0.00000903, -0.00000324, 0.00000729, 0.00000732, -0.00000581, -0.00002046, -0.00002431, -0.00001584, -0.00000249, 0.00000858, 0.00001453, 0.00001409, 0.00000546, -0.00001019, -0.00002547, -0.00003046, -0.00002101, -0.00000236, 0.00001580, 0.00002675, 0.00002859, 0.00002143, 0.00000632, -0.00001190, -0.00002392, -0.00002197, -0.00000781, 0.00000749, 0.00001338, 0.00000925, 0.00000305, 0.00000136, 0.00000286, 0.00000238, -0.00000105, -0.00000282, 0.00000037, 0.00000520, 0.00000532, -0.00000047, -0.00000606, -0.00000498, 0.00000202, 0.00000796, 0.00000749, 0.00000231, -0.00000192, -0.00000223, -0.00000048, 0.00000024, -0.00000045, -0.00000083, -0.00000015, 0.00000055, },
		{ -0.00000318, -0.00000456, -0.00000512, -0.00000392, -0.00000024, 0.00000452, 0.00000753, 0.00000750, 0.00000599, 0.00000499, 0.00000393, 0.00000081, -0.00000412, -0.00000755, -0.00000721, -0.00000550, -0.00000669, -0.00001055, -0.00001126, -0.00000469, 0.00000478, 0.00000787, 0.00000161, -0.00000586, -0.00000448, 0.00000481, 0.00001019, 0.00000361, -0.00000788, -0.00000961, 0.00000202, 0.00001293, 0.00000709, -0.00001315, -0.00002758, -0.00002073, -0.00000043, 0.00001069, 0.00000175, -0.00001333, -0.00001227, 0.00000874, 0.00002939, 0.00002771, 0.00000442, -0.00001825, -0.00002053, -0.00000403, 0.00001342, 0.00001748, 0.00000842, -0.00000462, -0.00001424, -0.00001714, -0.00001108, 0.00000358, 0.00001870, 0.00002091, 0.00000518, -0.00001526, -0.00001896, 0.00000063, 0.00002473, 0.00002705, 0.00000267, -0.00002489, -0.00002794, -0.00000429, 0.00002095, 0.00002323, 0.00000408, -0.00001340, -0.00001201, 0.00000166, 0.00000763, -0.00000332, -0.00002007, -0.00002506, -0.00001316, 0.00000518, 0.00001655, 0.00001614, 0.00000735, -0.00000510, -0.00001775, -0.00002583, -0.00002345, -0.00000923, 0.00000971, 0.00002240, 0.00002343, 0.00001631, 0.00000765, 0.00000009, -0.00000721, -0.00001309, -0.00001312, -0.00000552, 0.00000421, 0.00000759, 0.00000271, -0.00000322, -0.00000215, 0.00000514, 0.00001005, 0.00000665, -0.00000123, -0.00000492, -0.00000137, 0.00000379, 0.00000395, -0.00000027, -0.00000240, 0.00000113, 0.00000641, 0.00000702, 0.00000188, -0.00000382, -0.00000503, -0.00000232, 0.00000005, -0.00000019, -0.00000125, -0.00000060, 0.00000145, 0.00000249, 0.00000129, -0.00000110, },
		{ -0.00000180, 0.00000224, 0.00000650, 0.00000900, 0.00000816, 0.00000491, 0.00000195, 0.00000057, -0.00000079, -0.00000385, -0.00000737, -0.00000842, -0.00000660, -0.00000519, -0.00000664, -0.00000835, -0.00000548, 0.00000183, 0.00000678, 0.00000362, -0.00000411, -0.00000650, 0.00000059, 0.00000907, 0.00000764, -0.00000347, -0.00001110, -0.00000476, 0.00000887, 0.00001183, -0.00000391, -0.00002500, -0.00003028, -0.00001471, 0.00000467, 0.00000847, -0.00000265, -0.00000969, 0.00000093, 0.00002032, 0.00002704, 0.00001224, -0.00000979, -0.00001745, -0.00000503, 0.00001285, 0.00001829, 0.00000761, -0.00000840, -0.00001757, -0.00001583, -0.00000585, 0.00000755, 0.00001795, 0.00001773, 0.00000446, -0.00001220, -0.00001546, 0.00000127, 0.00002330, 0.00002653, 0.00000358, -0.00002496, -0.00003002, -0.00000525, 0.00002487, 0.00003051, 0.00000793, -0.00001805, -0.00002233, -0.00000606, 0.00000776, 0.00000224, -0.00001503, -0.00002267, -0.00001046, 0.00001046, 0.00002116, 0.00001413, -0.00000264, -0.00001683, -0.00002245, -0.00001964, -0.00000975, 0.00000508, 0.00001871, 0.00002275, 0.00001438, 0.00000093, -0.00000650, -0.00000413, 0.00000162, 0.00000263, -0.00000158, -0.00000509, -0.00000445, -0.00000277, -0.00000390, -0.00000554, -0.00000166, 0.00000822, 0.00001574, 0.00001232, -0.00000014, -0.00001004, -0.00000832, 0.00000191, 0.00000961, 0.00000832, 0.00000219, -0.00000075, 0.00000146, 0.00000358, 0.00000099, -0.00000427, -0.00000659, -0.00000405, -0.00000029, 0.00000065, -0.00000072, -0.00000089, 0.00000144, 0.00000366, 0.00000270, -0.00000114, -0.00000485, -0.00000606, -0.00000475, },
		{ 0.00000898, 0.00000727, 0.00000337, -0.00000059, -0.00000250, -0.00000273, -0.00000359, -0.00000600, -0.00000789, -0.00000708, -0.00000471, -0.00000395, -0.00000531, -0.00000523, -0.00000101, 0.00000424, 0.00000437, -0.00000156, -0.00000643, -0.00000314, 0.00000559, 0.00000901, 0.00000152, -0.00000886, -0.00000896, 0.00000265, 0.00001143, 0.00000287, -0.00001903, -0.00003378, -0.00002637, -0.00000440, 0.00001035, 0.00000650, -0.00000500, -0.00000526, 0.00000935, 0.00002309, 0.00002001, 0.00000301, -0.00000978, -0.00000603, 0.00000758, 0.00001429, 0.00000621, -0.00000863, -0.00001676, -0.00001267, -0.00000114, 0.00001002, 0.00001559, 0.00001318, 0.00000334, -0.00000772, -0.00000950, 0.00000296, 0.00001986, 0.00002246, 0.00000302, -0.00002262, -0.00002794, -0.00000420, 0.00002709, 0.00003449, 0.00000980, -0.00002274, -0.00003228, -0.00001384, 0.00000864, 0.00001071, -0.00000593, -0.00001809, -0.00000942, 0.00001127, 0.00002133, 0.00000967, -0.00001276, -0.00002651, -0.00002278, -0.00000763, 0.00000825, 0.00001869, 0.00002157, 0.00001527, 0.00000092, -0.00001346, -0.00001645, -0.00000446, 0.00001243, 0.00001890, 0.00000947, -0.00000675, -0.00001654, -0.00001553, -0.00000918, -0.00000314, 0.00000306, 0.00001119, 0.00001722, 0.00001372, 0.00000018, -0.00001269, -0.00001260, 0.00000088, 0.00001505, 0.00001719, 0.00000737, -0.00000340, -0.00000613, -0.00000249, -0.00000020, -0.00000265, -0.00000595, -0.00000532, -0.00000149, 0.00000106, 0.00000046, -0.00000053, 0.00000098, 0.00000363, 0.00000346, -0.00000075, -0.00000567, -0.00000708, -0.00000409, 0.00000089, 0.00000530, 0.00000813, },
		{ -0.00000235, -0.00000426, -0.00000410, -0.00000347, -0.00000433, -0.00000626, -0.00000679, -0.00000493, -0.00000284, -0.00000289, -0.00000395, -0.00000265, 0.00000132, 0.00000356, 0.00000037, -0.00000515, -0.00000559, 0.00000123, 0.00000795, 0.00000569, -0.00000414, -0.00000995, -0.00000394, 0.00000648, 0.00000559, -0.00001162, -0.00003085, -0.00003268, -0.00001435, 0.00000637, 0.00001074, -0.00000031, -0.00000842, -0.00000005, 0.00001778, 0.00002631, 0.00001704, 0.00000066, -0.00000645, -0.00000072, 0.00000663, 0.00000457, -0.00000516, -0.00001166, -0.00000812, 0.00000182, 0.00000996, 0.00001192, 0.00000862, 0.00000287, -0.00000216, -0.00000227, 0.00000501, 0.00001478, 0.00001562, 0.00000137, -0.00001779, -0.00002150, -0.00000101, 0.00002694, 0.00003361, 0.00000847, -0.00002689, -0.00003948, -0.00001919, 0.00001104, 0.00002090, 0.00000521, -0.00001306, -0.00001129, 0.00000687, 0.00001739, 0.00000489, -0.00001964, -0.00003150, -0.00001875, 0.00000744, 0.00002670, 0.00002867, 0.00001715, 0.00000076, -0.00001440, -0.00002303, -0.00001880, -0.00000028, 0.00002209, 0.00003073, 0.00001682, -0.00000962, -0.00002824, -0.00002720, -0.00001223, 0.00000271, 0.00001030, 0.00001288, 0.00001361, 0.00001019, 0.00000014, -0.00001087, -0.00001172, 0.00000148, 0.00001831, 0.00002313, 0.00001131, -0.00000606, -0.00001424, -0.00000974, -0.00000161, 0.00000062, -0.00000309, -0.00000591, -0.00000382, 0.00000027, 0.00000187, 0.00000093, 0.00000078, 0.00000246, 0.00000297, -0.00000032, -0.00000527, -0.00000691, -0.00000300, 0.00000360, 0.00000834, 0.00000930, 0.00000758, 0.00000466, 0.00000111, },
		{ -0.00000351, -0.00000336, -0.00000461, -0.00000587, -0.00000513, -0.00000278, -0.00000139, -0.00000212, -0.00000282, -0.00000114, 0.00000138, 0.00000070, -0.00000366, -0.00000638, -0.00000242, 0.00000536, 0.00000798, 0.00000111, -0.00000824, -0.00000922, -0.00000104, 0.00000414, -0.00000479, -0.00002294, -0.00003266, -0.00002292, -0.00000275, 0.00000829, 0.00000227, -0.00000824, -0.00000527, 0.00001310, 0.00003004, 0.00002896, 0.00001146, -0.00000563, -0.00000928, -0.00000242, 0.00000299, 0.00000100, -0.00000338, -0.00000294, 0.00000250, 0.00000724, 0.00000761, 0.00000537, 0.00000381, 0.00000376, 0.00000460, 0.00000650, 0.00000878, 0.00000762, -0.00000038, -0.00001080, -0.00001168, 0.00000356, 0.00002394, 0.00002733, 0.00000375, -0.00002952, -0.00004167, -0.00002000, 0.00001513, 0.00003066, 0.00001543, -0.00000951, -0.00001627, -0.00000173, 0.00001110, 0.00000216, -0.00002077, -0.00003060, -0.00001190, 0.00002155, 0.00004146, 0.00003321, 0.00000671, -0.00001779, -0.00002847, -0.00002549, -0.00001251, 0.00000754, 0.00002748, 0.00003409, 0.00001835, -0.00001196, -0.00003504, -0.00003410, -0.00001301, 0.00000895, 0.00001689, 0.00001236, 0.00000602, 0.00000318, 0.00000061, -0.00000388, -0.00000491, 0.00000368, 0.00001764, 0.00002341, 0.00001251, -0.00000754, -0.00001999, -0.00001610, -0.00000339, 0.00000405, 0.00000093, -0.00000569, -0.00000683, -0.00000180, 0.00000309, 0.00000347, 0.00000133, 0.00000048, 0.00000089, -0.00000032, -0.00000350, -0.00000484, -0.00000112, 0.00000576, 0.00001038, 0.00000958, 0.00000515, 0.00000084, -0.00000170, -0.00000317, -0.00000416, -0.00000424, },
		{ -0.00000455, -0.00000447, -0.00000274, -0.00000072, -0.00000059, -0.00000213, -0.00000281, -0.00000151, -0.00000066, -0.00000274, -0.00000568, -0.00000433, 0.00000234, 0.00000789, 0.00000493, -0.00000510, -0.00001198, -0.00000849, -0.00000025, -0.00000021, -0.00001288, -0.00002724, -0.00002831, -0.00001466, -0.00000043, 0.00000127, -0.00000577, -0.00000547, 0.00001043, 0.00003083, 0.00003632, 0.00002050, -0.00000303, -0.00001519, -0.00001021, 0.00000168, 0.00000790, 0.00000596, 0.00000209, 0.00000132, 0.00000272, 0.00000359, 0.00000419, 0.00000629, 0.00000911, 0.00000959, 0.00000673, 0.00000297, 0.00000057, -0.00000108, -0.00000249, -0.00000051, 0.00000800, 0.00001798, 0.00001652, -0.00000333, -0.00002945, -0.00003745, -0.00001533, 0.00002021, 0.00003740, 0.00002169, -0.00000881, -0.00002325, -0.00001188, 0.00000502, 0.00000317, -0.00001549, -0.00002453, -0.00000486, 0.00003074, 0.00004934, 0.00003225, -0.00000647, -0.00003600, -0.00003862, -0.00002085, 0.00000075, 0.00001694, 0.00002662, 0.00002728, 0.00001363, -0.00001183, -0.00003330, -0.00003358, -0.00001234, 0.00001170, 0.00001929, 0.00000939, -0.00000268, -0.00000419, 0.00000252, 0.00000715, 0.00000638, 0.00000664, 0.00001232, 0.00001695, 0.00001040, -0.00000672, -0.00002086, -0.00001964, -0.00000592, 0.00000554, 0.00000448, -0.00000453, -0.00000926, -0.00000443, 0.00000369, 0.00000638, 0.00000259, -0.00000184, -0.00000251, -0.00000096, -0.00000045, -0.00000056, 0.00000179, 0.00000685, 0.00001036, 0.00000838, 0.00000221, -0.00000309, -0.00000439, -0.00000311, -0.00000238, -0.00000302, -0.00000371, -0.00000373, -0.00000390, },
		{ -0.00000062, -0.00000022, -0.00000170, -0.00000377, -0.00000416, -0.00000295, -0.00000268, -0.00000419, -0.00000440, -0.00000023, 0.00000567, 0.00000612, -0.00000193, -0.00001182, -0.00001348, -0.00000562, 0.00000120, -0.00000391, -0.00001880, -0.00002958, -0.00002605, -0.00001305, -0.00000376, -0.00000309, -0.00000193, 0.00001009, 0.00002906, 0.00003831, 0.00002654, 0.00000161, -0.00001626, -0.00001454, 0.00000069, 0.00001335, 0.00001402, 0.00000636, -0.00000052, -0.00000199, 0.00000068, 0.00000499, 0.00000969, 0.00001295, 0.00001182, 0.00000562, -0.00000142, -0.00000365, 0.00000011, 0.00000578, 0.00000940, 0.00001065, 0.00000970, 0.00000342, -0.00001063, -0.00002576, -0.00002692, -0.00000588, 0.00002483, 0.00003890, 0.00002196, -0.00001117, -0.00002997, -0.00002014, 0.00000164, 0.00000822, -0.00000531, -0.00001568, -0.00000037, 0.00003238, 0.00004908, 0.00002730, -0.00001822, -0.00004952, -0.00004338, -0.00001135, 0.00001714, 0.00002542, 0.00001942, 0.00001157, 0.00000400, -0.00000796, -0.00002160, -0.00002491, -0.00001155, 0.00000777, 0.00001485, 0.00000464, -0.00000866, -0.00000816, 0.00000633, 0.00001929, 0.00001870, 0.00000904, 0.00000318, 0.00000498, 0.00000554, -0.00000314, -0.00001576, -0.00001932, -0.00000948, 0.00000331, 0.00000591, -0.00000226, -0.00000950, -0.00000644, 0.00000322, 0.00000832, 0.00000394, -0.00000381, -0.00000621, -0.00000192, 0.00000346, 0.00000552, 0.00000557, 0.00000653, 0.00000759, 0.00000536, -0.00000053, -0.00000572, -0.00000601, -0.00000247, 0.00000010, -0.00000106, -0.00000401, -0.00000535, -0.00000442, -0.00000308, -0.00000245, -0.00000179, },
		{ -0.00000437, -0.00000565, -0.00000493, -0.00000330, -0.00000292, -0.00000346, -0.00000200, 0.00000220, 0.00000465, 0.00000022, -0.00000923, -0.00001472, -0.00001004, -0.00000047, 0.00000139, -0.00001026, -0.00002679, -0.00003370, -0.00002600, -0.00001218, -0.00000269, 0.00000247, 0.00001125, 0.00002580, 0.00003595, 0.00002956, 0.00000814, -0.00001174, -0.00001426, -0.00000000, 0.00001570, 0.00001892, 0.00000961, -0.00000149, -0.00000513, -0.00000064, 0.00000692, 0.00001285, 0.00001470, 0.00001137, 0.00000382, -0.00000331, -0.00000398, 0.00000323, 0.00001245, 0.00001566, 0.00001030, 0.00000046, -0.00000883, -0.00001561, -0.00001818, -0.00001186, 0.00000603, 0.00002724, 0.00003410, 0.00001601, -0.00001548, -0.00003375, -0.00002344, 0.00000237, 0.00001595, 0.00000650, -0.00000727, -0.00000027, 0.00002614, 0.00004168, 0.00002088, -0.00002487, -0.00005537, -0.00004285, -0.00000070, 0.00003192, 0.00003089, 0.00000782, -0.00000916, -0.00000794, -0.00000039, -0.00000147, -0.00000955, -0.00001180, -0.00000410, 0.00000292, -0.00000057, -0.00000852, -0.00000570, 0.00001167, 0.00002859, 0.00002767, 0.00000970, -0.00000729, -0.00000909, -0.00000060, 0.00000254, -0.00000573, -0.00001529, -0.00001396, -0.00000318, 0.00000434, 0.00000108, -0.00000637, -0.00000673, 0.00000132, 0.00000778, 0.00000446, -0.00000467, -0.00000858, -0.00000242, 0.00000750, 0.00001209, 0.00000944, 0.00000475, 0.00000224, 0.00000072, -0.00000242, -0.00000587, -0.00000579, -0.00000171, 0.00000183, 0.00000054, -0.00000428, -0.00000734, -0.00000569, -0.00000166, 0.00000071, 0.00000038, -0.00000078, -0.00000149, -0.00000251, },
		{ -0.00000415, -0.00000276, -0.00000281, -0.00000309, -0.00000129, 0.00000158, 0.00000093, -0.00000540, -0.00001239, -0.00001223, -0.00000424, 0.00000192, -0.00000411, -0.00002103, -0.00003558, -0.00003574, -0.00002243, -0.00000653, 0.00000426, 0.00001222, 0.00002229, 0.00003158, 0.00003066, 0.00001597, -0.00000254, -0.00000959, -0.00000035, 0.00001445, 0.00001990, 0.00001189, -0.00000038, -0.00000542, -0.00000036, 0.00000872, 0.00001456, 0.00001433, 0.00000927, 0.00000242, -0.00000220, -0.00000061, 0.00000748, 0.00001613, 0.00001688, 0.00000680, -0.00000782, -0.00001708, -0.00001618, -0.00000748, 0.00000461, 0.00001718, 0.00002608, 0.00002372, 0.00000565, -0.00001961, -0.00003246, -0.00002024, 0.00000697, 0.00002371, 0.00001602, -0.00000217, -0.00000480, 0.00001405, 0.00002997, 0.00001552, -0.00002449, -0.00005277, -0.00003848, 0.00000730, 0.00004103, 0.00003233, -0.00000483, -0.00002962, -0.00001929, 0.00000933, 0.00002291, 0.00000928, -0.00001351, -0.00002244, -0.00001480, -0.00000468, -0.00000058, 0.00000422, 0.00001735, 0.00003128, 0.00002948, 0.00000808, -0.00001559, -0.00002069, -0.00000620, 0.00000845, 0.00000620, -0.00000887, -0.00001851, -0.00001274, -0.00000007, 0.00000510, 0.00000019, -0.00000479, -0.00000210, 0.00000377, 0.00000343, -0.00000370, -0.00000791, -0.00000159, 0.00001068, 0.00001712, 0.00001223, 0.00000195, -0.00000446, -0.00000461, -0.00000302, -0.00000323, -0.00000358, -0.00000142, 0.00000148, 0.00000073, -0.00000423, -0.00000830, -0.00000660, -0.00000046, 0.00000405, 0.00000319, -0.00000098, -0.00000415, -0.00000490, -0.00000498, -0.00000553, -0.00000552, },
		{ -0.00000391, -0.00000370, -0.00000152, 0.00000041, -0.00000185, -0.00000799, -0.00001187, -0.00000847, -0.00000160, -0.00000168, -0.00001412, -0.00003153, -0.00003990, -0.00003239, -0.00001521, 0.00000066, 0.00001100, 0.00001935, 0.00002782, 0.00003145, 0.00002426, 0.00000923, -0.00000184, -0.00000014, 0.00001043, 0.00001744, 0.00001340, 0.00000317, -0.00000254, 0.00000113, 0.00000934, 0.00001413, 0.00001243, 0.00000707, 0.00000253, 0.00000162, 0.00000509, 0.00001145, 0.00001610, 0.00001320, 0.00000123, -0.00001312, -0.00001913, -0.00001154, 0.00000457, 0.00001886, 0.00002451, 0.00002100, 0.00001018, -0.00000584, -0.00002127, -0.00002543, -0.00001121, 0.00001354, 0.00002845, 0.00002000, -0.00000191, -0.00001247, -0.00000016, 0.00001753, 0.00001284, -0.00001753, -0.00004331, -0.00003243, 0.00001000, 0.00004231, 0.00003002, -0.00001487, -0.00004483, -0.00002772, 0.00001853, 0.00004590, 0.00002765, -0.00001603, -0.00004329, -0.00003466, -0.00000660, 0.00001424, 0.00002017, 0.00002171, 0.00002517, 0.00002219, 0.00000456, -0.00001826, -0.00002554, -0.00000978, 0.00001194, 0.00001581, -0.00000199, -0.00002153, -0.00002247, -0.00000609, 0.00000876, 0.00000868, -0.00000106, -0.00000663, -0.00000348, 0.00000076, -0.00000052, -0.00000317, 0.00000114, 0.00001191, 0.00001851, 0.00001271, -0.00000100, -0.00001040, -0.00000915, -0.00000238, 0.00000132, -0.00000005, -0.00000199, -0.00000148, -0.00000098, -0.00000353, -0.00000694, -0.00000595, 0.00000032, 0.00000613, 0.00000542, -0.00000113, -0.00000720, -0.00000826, -0.00000583, -0.00000409, -0.00000436, -0.00000477, -0.00000411, -0.00000354, },
		{ -0.00000077, 0.00000011, -0.00000358, -0.00000950, -0.00001139, -0.00000702, -0.00000291, -0.00000818, -0.00002340, -0.00003795, -0.00003989, -0.00002747, -0.00000930, 0.00000589, 0.00001688, 0.00002643, 0.00003342, 0.00003229, 0.00002114, 0.00000719, 0.00000099, 0.00000548, 0.00001304, 0.00001435, 0.00000850, 0.00000278, 0.00000331, 0.00000840, 0.00001169, 0.00001000, 0.00000616, 0.00000471, 0.00000706, 0.00001106, 0.00001372, 0.00001266, 0.00000631, -0.00000441, -0.00001399, -0.00001449, -0.00000250, 0.00001568, 0.00002788, 0.00002620, 0.00001300, -0.00000304, -0.00001475, -0.00001895, -0.00001387, 0.00000087, 0.00001920, 0.00002784, 0.00001706, -0.00000613, -0.00002058, -0.00001236, 0.00000753, 0.00001298, -0.00000659, -0.00003016, -0.00002666, 0.00000686, 0.00003607, 0.00002522, -0.00001961, -0.00005155, -0.00003211, 0.00002432, 0.00006212, 0.00004201, -0.00001779, -0.00006129, -0.00005215, -0.00000613, 0.00003252, 0.00003850, 0.00002323, 0.00001054, 0.00000659, 0.00000030, -0.00001312, -0.00002110, -0.00001070, 0.00001044, 0.00001908, 0.00000351, -0.00002101, -0.00002862, -0.00001192, 0.00001052, 0.00001624, 0.00000322, -0.00001117, -0.00001224, -0.00000284, 0.00000463, 0.00000534, 0.00000552, 0.00001041, 0.00001489, 0.00001023, -0.00000293, -0.00001320, -0.00001137, -0.00000109, 0.00000586, 0.00000337, -0.00000345, -0.00000637, -0.00000416, -0.00000197, -0.00000273, -0.00000314, 0.00000048, 0.00000561, 0.00000572, -0.00000116, -0.00000920, -0.00001127, -0.00000698, -0.00000233, -0.00000207, -0.00000487, -0.00000659, -0.00000595, -0.00000490, -0.00000449, -0.00000328, },
		{ -0.00000615, -0.00001168, -0.00001163, -0.00000659, -0.00000506, -0.00001440, -0.00003123, -0.00004316, -0.00004041, -0.00002436, -0.00000402, 0.00001362, 0.00002742, 0.00003725, 0.00003968, 0.00003138, 0.00001591, 0.00000343, 0.00000172, 0.00000859, 0.00001482, 0.00001422, 0.00000911, 0.00000573, 0.00000642, 0.00000820, 0.00000809, 0.00000725, 0.00000871, 0.00001259, 0.00001531, 0.00001349, 0.00000712, -0.00000110, -0.00000791, -0.00001005, -0.00000451, 0.00000876, 0.00002370, 0.00003024, 0.00002231, 0.00000417, -0.00001250, -0.00001834, -0.00001264, -0.00000071, 0.00001202, 0.00002115, 0.00002119, 0.00000816, -0.00001278, -0.00002620, -0.00001942, 0.00000164, 0.00001471, 0.00000464, -0.00001696, -0.00002227, -0.00000053, 0.00002471, 0.00001954, -0.00001822, -0.00004917, -0.00003259, 0.00002466, 0.00006805, 0.00005023, -0.00001692, -0.00007135, -0.00006326, -0.00000410, 0.00004927, 0.00005446, 0.00002112, -0.00000977, -0.00001401, -0.00000322, -0.00000015, -0.00000751, -0.00000925, 0.00000246, 0.00001362, 0.00000645, -0.00001522, -0.00002781, -0.00001586, 0.00000857, 0.00001942, 0.00000648, -0.00001402, -0.00001950, -0.00000602, 0.00001063, 0.00001565, 0.00001050, 0.00000609, 0.00000641, 0.00000505, -0.00000279, -0.00001115, -0.00001030, -0.00000020, 0.00000800, 0.00000508, -0.00000527, -0.00001135, -0.00000751, 0.00000035, 0.00000361, 0.00000147, -0.00000009, 0.00000196, 0.00000331, -0.00000106, -0.00000886, -0.00001245, -0.00000833, -0.00000168, -0.00000000, -0.00000438, -0.00000910, -0.00000936, -0.00000634, -0.00000398, -0.00000322, -0.00000197, -0.00000011, -0.00000094, },
		{ -0.00001096, -0.00000575, -0.00000792, -0.00002236, -0.00004132, -0.00005035, -0.00004135, -0.00001853, 0.00000752, 0.00002887, 0.00004247, 0.00004657, 0.00003951, 0.00002357, 0.00000738, 0.00000070, 0.00000567, 0.00001486, 0.00001879, 0.00001489, 0.00000821, 0.00000458, 0.00000509, 0.00000739, 0.00001007, 0.00001345, 0.00001673, 0.00001670, 0.00001095, 0.00000147, -0.00000624, -0.00000768, -0.00000218, 0.00000801, 0.00001943, 0.00002740, 0.00002659, 0.00001474, -0.00000305, -0.00001599, -0.00001582, -0.00000396, 0.00001034, 0.00001844, 0.00001779, 0.00000978, -0.00000374, -0.00001896, -0.00002729, -0.00002019, -0.00000023, 0.00001601, 0.00001263, -0.00000659, -0.00001922, -0.00000912, 0.00001177, 0.00001429, -0.00001196, -0.00003966, -0.00003021, 0.00001916, 0.00006306, 0.00005188, -0.00001209, -0.00007030, -0.00006563, -0.00000197, 0.00005948, 0.00006365, 0.00001566, -0.00003105, -0.00003471, -0.00000480, 0.00001824, 0.00001237, -0.00000646, -0.00001162, -0.00000043, 0.00000663, -0.00000349, -0.00001830, -0.00001688, 0.00000156, 0.00001539, 0.00000742, -0.00001325, -0.00002187, -0.00000733, 0.00001557, 0.00002447, 0.00001440, -0.00000014, -0.00000499, -0.00000153, -0.00000015, -0.00000406, -0.00000606, -0.00000080, 0.00000583, 0.00000393, -0.00000646, -0.00001394, -0.00000935, 0.00000287, 0.00001012, 0.00000654, -0.00000128, -0.00000411, -0.00000156, -0.00000086, -0.00000568, -0.00001089, -0.00000971, -0.00000331, 0.00000033, -0.00000344, -0.00001019, -0.00001230, -0.00000823, -0.00000330, -0.00000196, -0.00000272, -0.00000206, -0.00000076, -0.00000294, -0.00000904, -0.00001334, },
		{ -0.00001431, -0.00003470, -0.00005406, -0.00005618, -0.00003622, -0.00000352, 0.00002754, 0.00004742, 0.00005337, 0.00004643, 0.00003038, 0.00001278, 0.00000290, 0.00000512, 0.00001460, 0.00002124, 0.00001901, 0.00001079, 0.00000414, 0.00000364, 0.00000804, 0.00001356, 0.00001755, 0.00001863, 0.00001532, 0.00000727, -0.00000226, -0.00000718, -0.00000330, 0.00000777, 0.00001990, 0.00002717, 0.00002693, 0.00001931, 0.00000645, -0.00000679, -0.00001339, -0.00000871, 0.00000429, 0.00001599, 0.00001786, 0.00000934, -0.00000351, -0.00001479, -0.00002207, -0.00002343, -0.00001574, 0.00000032, 0.00001501, 0.00001530, -0.00000040, -0.00001658, -0.00001561, 0.00000064, 0.00001002, -0.00000365, -0.00002661, -0.00002629, 0.00000932, 0.00004943, 0.00004807, -0.00000320, -0.00005794, -0.00005915, -0.00000125, 0.00005979, 0.00006340, 0.00000808, -0.00004815, -0.00005054, -0.00000398, 0.00003784, 0.00003379, -0.00000353, -0.00002944, -0.00002029, 0.00000473, 0.00001323, -0.00000082, -0.00001496, -0.00001082, 0.00000306, 0.00000554, -0.00000733, -0.00001667, -0.00000583, 0.00001712, 0.00002809, 0.00001560, -0.00000635, -0.00001579, -0.00000751, 0.00000442, 0.00000631, 0.00000008, -0.00000348, -0.00000110, -0.00000016, -0.00000594, -0.00001208, -0.00000837, 0.00000461, 0.00001417, 0.00001019, -0.00000261, -0.00001061, -0.00000754, -0.00000075, -0.00000045, -0.00000674, -0.00001083, -0.00000752, -0.00000197, -0.00000235, -0.00000873, -0.00001323, -0.00001041, -0.00000390, -0.00000094, -0.00000283, -0.00000439, -0.00000241, -0.00000068, -0.00000435, -0.00001149, -0.00001437, -0.00001001, -0.00000629, },
		{ -0.00006516, -0.00005462, -0.00002057, 0.00002009, 0.00004966, 0.00006023, 0.00005384, 0.00003743, 0.00001935, 0.00000767, 0.00000689, 0.00001438, 0.00002160, 0.00002120, 0.00001357, 0.00000585, 0.00000451, 0.00000969, 0.00001643, 0.00001988, 0.00001829, 0.00001238, 0.00000417, -0.00000271, -0.00000348, 0.00000435, 0.00001763, 0.00002866, 0.00003090, 0.00002371, 0.00001164, 0.00000026, -0.00000656, -0.00000676, -0.00000030, 0.00000917, 0.00001459, 0.00001038, -0.00000212, -0.00001515, -0.00002164, -0.00002080, -0.00001594, -0.00000877, 0.00000113, 0.00001084, 0.00001257, 0.00000194, -0.00001316, -0.00001766, -0.00000652, 0.00000649, 0.00000343, -0.00001397, -0.00002187, -0.00000191, 0.00003145, 0.00004081, 0.00000835, -0.00003712, -0.00004581, -0.00000276, 0.00004958, 0.00005351, 0.00000010, -0.00005709, -0.00005797, -0.00000122, 0.00005385, 0.00005159, -0.00000137, -0.00004710, -0.00004132, 0.00000187, 0.00003225, 0.00002146, -0.00001094, -0.00002734, -0.00001625, 0.00000135, 0.00000421, -0.00000312, -0.00000154, 0.00001328, 0.00002357, 0.00001316, -0.00000999, -0.00002190, -0.00001090, 0.00000921, 0.00001652, 0.00000601, -0.00000789, -0.00001127, -0.00000607, -0.00000306, -0.00000509, -0.00000423, 0.00000453, 0.00001346, 0.00001077, -0.00000329, -0.00001488, -0.00001262, -0.00000115, 0.00000487, -0.00000130, -0.00001133, -0.00001334, -0.00000670, -0.00000147, -0.00000456, -0.00001129, -0.00001255, -0.00000693, -0.00000172, -0.00000234, -0.00000533, -0.00000424, 0.00000019, 0.00000043, -0.00000713, -0.00001577, -0.00001624, -0.00000985, -0.00000940, -0.00002494, -0.00005008, },
		{ 0.00000439, 0.00004644, 0.00006643, 0.00006310, 0.00004611, 0.00002693, 0.00001379, 0.00001029, 0.00001475, 0.00002087, 0.00002195, 0.00001649, 0.00000951, 0.00000746, 0.00001176, 0.00001786, 0.00002008, 0.00001656, 0.00000962, 0.00000312, 0.00000044, 0.00000389, 0.00001360, 0.00002574, 0.00003338, 0.00003112, 0.00001983, 0.00000630, -0.00000234, -0.00000358, 0.00000036, 0.00000563, 0.00000873, 0.00000674, -0.00000158, -0.00001345, -0.00002227, -0.00002264, -0.00001558, -0.00000735, -0.00000244, 0.00000041, 0.00000397, 0.00000619, 0.00000192, -0.00000822, -0.00001471, -0.00000921, 0.00000305, 0.00000673, -0.00000472, -0.00001741, -0.00001114, 0.00001399, 0.00003219, 0.00001993, -0.00001286, -0.00002893, -0.00000627, 0.00003124, 0.00003623, -0.00000666, -0.00005630, -0.00005584, 0.00000237, 0.00006234, 0.00006170, -0.00000028, -0.00006030, -0.00005841, -0.00000088, 0.00004977, 0.00004374, -0.00000613, -0.00004524, -0.00003893, -0.00000383, 0.00002027, 0.00001714, 0.00000443, 0.00000305, 0.00000994, 0.00000731, -0.00000862, -0.00001997, -0.00001050, 0.00001171, 0.00002233, 0.00000955, -0.00001250, -0.00002134, -0.00001171, 0.00000190, 0.00000569, 0.00000211, 0.00000200, 0.00000709, 0.00000759, -0.00000252, -0.00001459, -0.00001489, -0.00000262, 0.00000767, 0.00000341, -0.00001071, -0.00001861, -0.00001246, -0.00000131, 0.00000113, -0.00000681, -0.00001421, -0.00001243, -0.00000512, -0.00000151, -0.00000353, -0.00000446, -0.00000031, 0.00000302, -0.00000256, -0.00001487, -0.00002189, -0.00001675, -0.00000871, -0.00001473, -0.00003941, -0.00006617, -0.00007060, -0.00004272, },
		{ 0.00007418, 0.00005750, 0.00003580, 0.00002035, 0.00001455, 0.00001623, 0.00002048, 0.00002220, 0.00001922, 0.00001409, 0.00001162, 0.00001388, 0.00001791, 0.00001877, 0.00001466, 0.00000849, 0.00000484, 0.00000643, 0.00001312, 0.00002276, 0.00003157, 0.00003482, 0.00002940, 0.00001706, 0.00000442, -0.00000184, -0.00000037, 0.00000429, 0.00000625, 0.00000283, -0.00000490, -0.00001414, -0.00002156, -0.00002357, -0.00001829, -0.00000837, -0.00000042, 0.00000083, -0.00000245, -0.00000391, -0.00000107, 0.00000130, -0.00000196, -0.00000802, -0.00000842, -0.00000084, 0.00000533, -0.00000010, -0.00001287, -0.00001572, 0.00000096, 0.00002377, 0.00002854, 0.00000920, -0.00001212, -0.00001048, 0.00000939, 0.00001553, -0.00001128, -0.00004697, -0.00004562, 0.00000547, 0.00006147, 0.00006231, 0.00000006, -0.00006567, -0.00006752, -0.00000285, 0.00006181, 0.00006093, -0.00000180, -0.00006076, -0.00005999, -0.00000838, 0.00003818, 0.00004022, 0.00001024, -0.00001286, -0.00001123, -0.00000055, -0.00000075, -0.00000858, -0.00000637, 0.00000935, 0.00002011, 0.00000925, -0.00001484, -0.00002699, -0.00001481, 0.00000743, 0.00001693, 0.00000861, -0.00000270, -0.00000375, 0.00000131, 0.00000009, -0.00000879, -0.00001336, -0.00000563, 0.00000577, 0.00000546, -0.00000844, -0.00002067, -0.00001716, -0.00000243, 0.00000595, -0.00000134, -0.00001489, -0.00001904, -0.00001067, -0.00000072, 0.00000118, -0.00000205, -0.00000192, 0.00000167, -0.00000028, -0.00001206, -0.00002421, -0.00002367, -0.00001191, -0.00000726, -0.00002546, -0.00005925, -0.00008087, -0.00006679, -0.00001938, 0.00003538, 0.00006986, },
		{ 0.00002752, 0.00001923, 0.00001901, 0.00002148, 0.00002279, 0.00002139, 0.00001824, 0.00001591, 0.00001611, 0.00001757, 0.00001718, 0.00001366, 0.00000957, 0.00000906, 0.00001400, 0.00002254, 0.00003093, 0.00003576, 0.00003483, 0.00002753, 0.00001595, 0.00000500, -0.00000045, 0.00000027, 0.00000258, 0.00000078, -0.00000676, -0.00001655, -0.00002347, -0.00002458, -0.00001978, -0.00001096, -0.00000177, 0.00000291, 0.00000025, -0.00000677, -0.00001062, -0.00000661, 0.00000133, 0.00000452, -0.00000010, -0.00000592, -0.00000518, 0.00000018, 0.00000041, -0.00000805, -0.00001461, -0.00000571, 0.00001636, 0.00003180, 0.00002448, 0.00000170, -0.00001354, -0.00001055, -0.00000418, -0.00001372, -0.00003262, -0.00003073, 0.00000710, 0.00005200, 0.00005419, 0.00000026, -0.00006179, -0.00006682, -0.00000393, 0.00006551, 0.00006925, 0.00000129, -0.00007021, -0.00007446, -0.00001098, 0.00005422, 0.00006095, 0.00001395, -0.00003204, -0.00003597, -0.00000820, 0.00001338, 0.00001097, 0.00000008, 0.00000040, 0.00000814, 0.00000504, -0.00001226, -0.00002437, -0.00001380, 0.00001102, 0.00002415, 0.00001287, -0.00000810, -0.00001578, -0.00000619, 0.00000407, 0.00000133, -0.00000842, -0.00001015, -0.00000155, 0.00000382, -0.00000422, -0.00001739, -0.00001883, -0.00000537, 0.00000709, 0.00000286, -0.00001395, -0.00002417, -0.00001663, -0.00000049, 0.00000733, 0.00000256, -0.00000414, -0.00000392, -0.00000184, -0.00000816, -0.00002132, -0.00002716, -0.00001760, -0.00000532, -0.00001300, -0.00004653, -0.00008276, -0.00008782, -0.00004851, 0.00001508, 0.00006718, 0.00008502, 0.00007185, 0.00004696, },
		{ 0.00002392, 0.00002411, 0.00002287, 0.00002114, 0.00001969, 0.00001884, 0.00001816, 0.00001659, 0.00001400, 0.00001243, 0.00001485, 0.00002211, 0.00003144, 0.00003830, 0.00003965, 0.00003516, 0.00002633, 0.00001543, 0.00000535, -0.00000115, -0.00000363, -0.00000497, -0.00000914, -0.00001696, -0.00002465, -0.00002711, -0.00002244, -0.00001321, -0.00000383, 0.00000201, 0.00000220, -0.00000329, -0.00001096, -0.00001435, -0.00000891, 0.00000225, 0.00000957, 0.00000639, -0.00000339, -0.00000925, -0.00000641, -0.00000144, -0.00000299, -0.00000871, -0.00000634, 0.00001009, 0.00002883, 0.00003066, 0.00001091, -0.00001381, -0.00002404, -0.00001920, -0.00001470, -0.00001783, -0.00001540, 0.00000692, 0.00003690, 0.00004025, 0.00000087, -0.00004970, -0.00005707, -0.00000436, 0.00005999, 0.00006720, 0.00000298, -0.00007104, -0.00007892, -0.00001107, 0.00006464, 0.00007434, 0.00001419, -0.00005076, -0.00005893, -0.00001339, 0.00003165, 0.00003485, 0.00000654, -0.00001515, -0.00001250, -0.00000171, -0.00000285, -0.00001152, -0.00000847, 0.00000986, 0.00002316, 0.00001306, -0.00001172, -0.00002441, -0.00001245, 0.00000790, 0.00001246, -0.00000175, -0.00001536, -0.00001281, -0.00000116, 0.00000167, -0.00000832, -0.00001648, -0.00001038, 0.00000243, 0.00000376, -0.00001084, -0.00002481, -0.00002059, -0.00000152, 0.00001208, 0.00000751, -0.00000629, -0.00001214, -0.00000700, -0.00000415, -0.00001313, -0.00002503, -0.00002384, -0.00001031, -0.00000650, -0.00003150, -0.00007495, -0.00010044, -0.00007829, -0.00001430, 0.00005455, 0.00009173, 0.00008819, 0.00006191, 0.00003685, 0.00002460, 0.00002269, },
		{ 0.00002309, 0.00002302, 0.00002248, 0.00002053, 0.00001765, 0.00001548, 0.00001617, 0.00002137, 0.00003058, 0.00004034, 0.00004584, 0.00004426, 0.00003632, 0.00002501, 0.00001310, 0.00000230, -0.00000631, -0.00001220, -0.00001622, -0.00002016, -0.00002455, -0.00002719, -0.00002482, -0.00001686, -0.00000680, 0.00000041, 0.00000202, -0.00000149, -0.00000773, -0.00001337, -0.00001438, -0.00000801, 0.00000341, 0.00001179, 0.00000959, -0.00000179, -0.00001192, -0.00001175, -0.00000363, 0.00000185, -0.00000044, -0.00000295, 0.00000483, 0.00002055, 0.00002810, 0.00001525, -0.00001057, -0.00002868, -0.00002744, -0.00001520, -0.00000670, -0.00000336, 0.00000523, 0.00002032, 0.00002444, 0.00000208, -0.00003252, -0.00004120, -0.00000450, 0.00004674, 0.00005600, 0.00000364, -0.00006273, -0.00007249, -0.00000898, 0.00006674, 0.00007713, 0.00001069, -0.00006497, -0.00007495, -0.00001465, 0.00005034, 0.00005765, 0.00001062, -0.00003528, -0.00003801, -0.00000857, 0.00001364, 0.00001071, -0.00000020, 0.00000176, 0.00001155, 0.00000862, -0.00001077, -0.00002518, -0.00001534, 0.00000919, 0.00001997, 0.00000426, -0.00001942, -0.00002439, -0.00000777, 0.00000804, 0.00000466, -0.00001063, -0.00001703, -0.00000828, 0.00000046, -0.00000531, -0.00001878, -0.00002045, -0.00000448, 0.00001210, 0.00001013, -0.00000754, -0.00001969, -0.00001367, -0.00000104, -0.00000176, -0.00001745, -0.00002874, -0.00002144, -0.00000850, -0.00001799, -0.00005821, -0.00010068, -0.00010324, -0.00005074, 0.00002904, 0.00008959, 0.00010423, 0.00008190, 0.00005076, 0.00003158, 0.00002641, 0.00002669, 0.00002601, 0.00002414, },
		{ 0.00002440, 0.00002014, 0.00001772, 0.00002034, 0.00002865, 0.00004026, 0.00005049, 0.00005435, 0.00004933, 0.00003690, 0.00002127, 0.00000629, -0.00000635, -0.00001639, -0.00002355, -0.00002747, -0.00002849, -0.00002739, -0.00002428, -0.00001852, -0.00001052, -0.00000298, 0.00000056, -0.00000117, -0.00000611, -0.00001088, -0.00001303, -0.00001132, -0.00000521, 0.00000379, 0.00001057, 0.00000910, -0.00000114, -0.00001213, -0.00001378, -0.00000467, 0.00000562, 0.00000718, 0.00000169, 0.00000045, 0.00000939, 0.00001938, 0.00001550, -0.00000438, -0.00002477, -0.00002864, -0.00001593, -0.00000166, 0.00000314, 0.00000278, 0.00000622, 0.00001048, 0.00000362, -0.00001442, -0.00002328, -0.00000450, 0.00002918, 0.00003901, 0.00000378, -0.00004707, -0.00005703, -0.00000568, 0.00005981, 0.00006870, 0.00000432, -0.00007151, -0.00008051, -0.00001177, 0.00006517, 0.00007393, 0.00001075, -0.00005624, -0.00006279, -0.00001297, 0.00003531, 0.00003853, 0.00000849, -0.00001381, -0.00001023, 0.00000064, -0.00000304, -0.00001534, -0.00001391, 0.00000533, 0.00001934, 0.00000743, -0.00001983, -0.00003140, -0.00001359, 0.00001278, 0.00001748, -0.00000318, -0.00002393, -0.00002276, -0.00000630, 0.00000217, -0.00000601, -0.00001547, -0.00000981, 0.00000493, 0.00000813, -0.00000700, -0.00002263, -0.00001868, 0.00000022, 0.00000882, -0.00000696, -0.00003083, -0.00003550, -0.00001845, -0.00000916, -0.00003574, -0.00008759, -0.00011856, -0.00009011, -0.00000977, 0.00007453, 0.00011646, 0.00010682, 0.00007144, 0.00004153, 0.00002930, 0.00002811, 0.00002804, 0.00002638, 0.00002560, 0.00002659, 0.00002692, },
		{ 0.00002580, 0.00003791, 0.00005245, 0.00006271, 0.00006335, 0.00005278, 0.00003386, 0.00001229, -0.00000665, -0.00002053, -0.00002952, -0.00003431, -0.00003493, -0.00003149, -0.00002522, -0.00001804, -0.00001138, -0.00000597, -0.00000274, -0.00000275, -0.00000579, -0.00000959, -0.00001137, -0.00001012, -0.00000677, -0.00000236, 0.00000259, 0.00000646, 0.00000588, -0.00000082, -0.00000946, -0.00001181, -0.00000405, 0.00000746, 0.00001163, 0.00000515, -0.00000292, -0.00000154, 0.00000819, 0.00001299, 0.00000308, -0.00001488, -0.00002416, -0.00001694, -0.00000280, 0.00000369, 0.00000027, -0.00000286, 0.00000087, 0.00000494, 0.00000062, -0.00000721, -0.00000420, 0.00001163, 0.00002066, 0.00000379, -0.00002771, -0.00003657, -0.00000240, 0.00004540, 0.00005130, -0.00000322, -0.00006904, -0.00007484, -0.00000590, 0.00007251, 0.00007990, 0.00000673, -0.00007341, -0.00008105, -0.00001316, 0.00005833, 0.00006614, 0.00001474, -0.00003524, -0.00003866, -0.00000852, 0.00001208, 0.00000503, -0.00000876, -0.00000557, 0.00000781, 0.00000664, -0.00001401, -0.00002924, -0.00001639, 0.00001309, 0.00002482, 0.00000324, -0.00002868, -0.00003641, -0.00001428, 0.00001009, 0.00001082, -0.00000673, -0.00001721, -0.00000959, 0.00000073, -0.00000402, -0.00001792, -0.00001907, -0.00000120, 0.00001390, 0.00000233, -0.00002923, -0.00004752, -0.00003274, -0.00000692, -0.00001291, -0.00006407, -0.00012134, -0.00012650, -0.00005859, 0.00004462, 0.00012045, 0.00013402, 0.00009936, 0.00005590, 0.00003144, 0.00002712, 0.00002995, 0.00003109, 0.00003076, 0.00003125, 0.00003130, 0.00002829, 0.00002319, 0.00002087, },
		{ 0.00007610, 0.00007033, 0.00005103, 0.00002372, -0.00000384, -0.00002518, -0.00003756, -0.00004200, -0.00004098, -0.00003625, -0.00002870, -0.00001971, -0.00001159, -0.00000641, -0.00000465, -0.00000543, -0.00000752, -0.00000973, -0.00001065, -0.00000921, -0.00000587, -0.00000263, -0.00000094, -0.00000020, 0.00000091, 0.00000171, -0.00000010, -0.00000443, -0.00000669, -0.00000220, 0.00000690, 0.00001174, 0.00000611, -0.00000501, -0.00000951, -0.00000193, 0.00000902, 0.00000960, -0.00000284, -0.00001622, -0.00001764, -0.00000821, -0.00000045, -0.00000192, -0.00000639, -0.00000366, 0.00000551, 0.00001006, 0.00000432, -0.00000329, -0.00000196, 0.00000505, 0.00000375, -0.00000907, -0.00001602, -0.00000008, 0.00002690, 0.00002929, -0.00001002, -0.00005850, -0.00006005, 0.00000091, 0.00007048, 0.00007446, -0.00000012, -0.00008270, -0.00008853, -0.00000881, 0.00007784, 0.00008728, 0.00001644, -0.00005882, -0.00006792, -0.00001602, 0.00003303, 0.00003265, -0.00000182, -0.00002375, -0.00001444, 0.00000227, -0.00000034, -0.00001523, -0.00001504, 0.00000630, 0.00002183, 0.00000627, -0.00002824, -0.00004337, -0.00002051, 0.00001582, 0.00002647, 0.00000298, -0.00002527, -0.00002859, -0.00001077, 0.00000130, -0.00000497, -0.00001351, -0.00000573, 0.00000986, 0.00000638, -0.00002370, -0.00005209, -0.00004590, -0.00001167, 0.00000373, -0.00003644, -0.00011118, -0.00015291, -0.00011032, 0.00000102, 0.00011135, 0.00015772, 0.00013220, 0.00007590, 0.00003474, 0.00002447, 0.00003202, 0.00003930, 0.00003999, 0.00003729, 0.00003416, 0.00003024, 0.00002616, 0.00002624, 0.00003488, 0.00005114, 0.00006787, },
		{ 0.00000433, -0.00002668, -0.00004625, -0.00005293, -0.00004968, -0.00004128, -0.00003147, -0.00002201, -0.00001370, -0.00000770, -0.00000528, -0.00000637, -0.00000918, -0.00001135, -0.00001145, -0.00000928, -0.00000559, -0.00000196, -0.00000031, -0.00000148, -0.00000381, -0.00000423, -0.00000167, 0.00000135, 0.00000161, -0.00000026, -0.00000004, 0.00000420, 0.00000793, 0.00000451, -0.00000565, -0.00001301, -0.00000859, 0.00000465, 0.00001325, 0.00000757, -0.00000725, -0.00001708, -0.00001479, -0.00000699, -0.00000387, -0.00000571, -0.00000395, 0.00000514, 0.00001331, 0.00001038, -0.00000149, -0.00000919, -0.00000510, 0.00000343, 0.00000491, 0.00000028, 0.00000084, 0.00000861, 0.00000782, -0.00001458, -0.00004285, -0.00004045, 0.00000639, 0.00005958, 0.00005955, -0.00000751, -0.00008170, -0.00008375, -0.00000127, 0.00008925, 0.00009710, 0.00001295, -0.00007965, -0.00009153, -0.00001952, 0.00005620, 0.00006181, 0.00000434, -0.00004730, -0.00004397, -0.00000392, 0.00002104, 0.00001016, -0.00001000, -0.00000916, 0.00000582, 0.00000480, -0.00001974, -0.00003842, -0.00002244, 0.00001603, 0.00003441, 0.00001016, -0.00003135, -0.00004633, -0.00002311, 0.00000774, 0.00001340, -0.00000306, -0.00001303, -0.00000417, 0.00000290, -0.00001467, -0.00004539, -0.00005253, -0.00002247, 0.00000822, -0.00001267, -0.00009009, -0.00016264, -0.00015518, -0.00005200, 0.00008474, 0.00016942, 0.00016459, 0.00010224, 0.00004334, 0.00002352, 0.00003515, 0.00005060, 0.00005284, 0.00004413, 0.00003436, 0.00002922, 0.00002959, 0.00003635, 0.00005072, 0.00006992, 0.00008539, 0.00008724, 0.00007096, 0.00004021, },
		{ -0.00006186, -0.00004971, -0.00003521, -0.00002306, -0.00001483, -0.00000997, -0.00000767, -0.00000768, -0.00000958, -0.00001190, -0.00001259, -0.00001058, -0.00000655, -0.00000231, 0.00000020, -0.00000023, -0.00000339, -0.00000698, -0.00000750, -0.00000325, 0.00000326, 0.00000696, 0.00000540, 0.00000157, 0.00000033, 0.00000192, 0.00000128, -0.00000497, -0.00001211, -0.00001103, 0.00000057, 0.00001309, 0.00001380, 0.00000077, -0.00001451, -0.00001950, -0.00001351, -0.00000594, -0.00000324, -0.00000188, 0.00000401, 0.00001165, 0.00001166, 0.00000119, -0.00000975, -0.00000902, 0.00000257, 0.00001192, 0.00000966, 0.00000052, -0.00000564, -0.00000870, -0.00001624, -0.00002610, -0.00002110, 0.00000889, 0.00004253, 0.00003940, -0.00001294, -0.00007053, -0.00006841, 0.00000699, 0.00008962, 0.00009353, 0.00000535, -0.00009291, -0.00010414, -0.00001806, 0.00007670, 0.00008594, 0.00000750, -0.00007225, -0.00007483, -0.00000953, 0.00004768, 0.00004302, -0.00000310, -0.00003287, -0.00002240, -0.00000056, -0.00000156, -0.00001884, -0.00001901, 0.00000750, 0.00002883, 0.00001186, -0.00003189, -0.00005560, -0.00003219, 0.00001270, 0.00003114, 0.00000890, -0.00002161, -0.00002521, -0.00000755, -0.00000346, -0.00002692, -0.00004937, -0.00003728, -0.00000282, 0.00000011, -0.00006186, -0.00015106, -0.00018319, -0.00010793, 0.00003777, 0.00015970, 0.00018892, 0.00013483, 0.00006304, 0.00002990, 0.00004064, 0.00006256, 0.00006635, 0.00005041, 0.00003221, 0.00002664, 0.00003584, 0.00005419, 0.00007551, 0.00009382, 0.00010133, 0.00009058, 0.00006022, 0.00001786, -0.00002358, -0.00005279, -0.00006491, },
		{ -0.00001484, -0.00001057, -0.00001006, -0.00001085, -0.00001164, -0.00001220, -0.00001225, -0.00001101, -0.00000801, -0.00000402, -0.00000079, 0.00000012, -0.00000177, -0.00000545, -0.00000861, -0.00000831, -0.00000309, 0.00000484, 0.00001016, 0.00000880, 0.00000228, -0.00000337, -0.00000407, -0.00000227, -0.00000348, -0.00000823, -0.00000995, -0.00000277, 0.00000945, 0.00001499, 0.00000656, -0.00000979, -0.00002037, -0.00001806, -0.00000848, -0.00000148, 0.00000050, 0.00000263, 0.00000754, 0.00000996, 0.00000433, -0.00000525, -0.00000776, 0.00000106, 0.00001182, 0.00001166, -0.00000056, -0.00001352, -0.00001773, -0.00001526, -0.00001211, -0.00000633, 0.00000782, 0.00002341, 0.00001918, -0.00001462, -0.00005179, -0.00004677, 0.00001318, 0.00007858, 0.00007785, -0.00000396, -0.00009531, -0.00010311, -0.00001232, 0.00008971, 0.00009948, 0.00000661, -0.00009350, -0.00010018, -0.00001241, 0.00007513, 0.00007681, 0.00000311, -0.00006198, -0.00005833, -0.00000769, 0.00002575, 0.00001422, -0.00001114, -0.00001166, 0.00000676, 0.00000694, -0.00002337, -0.00004988, -0.00003453, 0.00001250, 0.00004036, 0.00001778, -0.00002860, -0.00004657, -0.00002155, 0.00000756, -0.00000062, -0.00003671, -0.00005314, -0.00002838, -0.00000191, -0.00003132, -0.00011755, -0.00018719, -0.00015944, -0.00002945, 0.00012104, 0.00019712, 0.00017229, 0.00009944, 0.00004998, 0.00004972, 0.00007107, 0.00007568, 0.00005487, 0.00003058, 0.00002714, 0.00004882, 0.00008164, 0.00010854, 0.00011920, 0.00010972, 0.00007993, 0.00003452, -0.00001544, -0.00005576, -0.00007632, -0.00007595, -0.00006132, -0.00004190, -0.00002523, },
		{ -0.00001506, -0.00001456, -0.00001263, -0.00001024, -0.00000782, -0.00000532, -0.00000292, -0.00000142, -0.00000167, -0.00000370, -0.00000649, -0.00000826, -0.00000710, -0.00000205, 0.00000529, 0.00001056, 0.00000950, 0.00000227, -0.00000563, -0.00000818, -0.00000507, -0.00000198, -0.00000350, -0.00000689, -0.00000510, 0.00000377, 0.00001197, 0.00000963, -0.00000365, -0.00001704, -0.00001961, -0.00001146, -0.00000202, 0.00000159, 0.00000154, 0.00000350, 0.00000729, 0.00000724, 0.00000149, -0.00000351, -0.00000088, 0.00000644, 0.00000783, -0.00000182, -0.00001478, -0.00001912, -0.00001255, -0.00000340, 0.00000145, 0.00000384, 0.00000643, 0.00000338, -0.00001206, -0.00002981, -0.00002431, 0.00001521, 0.00005848, 0.00005418, -0.00001208, -0.00008607, -0.00008922, -0.00000438, 0.00009191, 0.00009952, 0.00000218, -0.00010614, -0.00011422, -0.00001155, 0.00009782, 0.00010417, 0.00000643, -0.00009149, -0.00009472, -0.00001374, 0.00005894, 0.00005554, -0.00000149, -0.00004098, -0.00003047, -0.00000337, -0.00000347, -0.00002573, -0.00002872, 0.00000327, 0.00003371, 0.00001950, -0.00003007, -0.00005940, -0.00003392, 0.00001490, 0.00002548, -0.00001847, -0.00006632, -0.00006232, -0.00001773, -0.00000359, -0.00006692, -0.00016548, -0.00019980, -0.00011247, 0.00005092, 0.00018275, 0.00021128, 0.00015525, 0.00008855, 0.00006325, 0.00007186, 0.00007604, 0.00005758, 0.00003518, 0.00003805, 0.00007267, 0.00011765, 0.00014455, 0.00013888, 0.00010372, 0.00005093, -0.00000631, -0.00005535, -0.00008522, -0.00009071, -0.00007576, -0.00005138, -0.00002909, -0.00001534, -0.00001045, -0.00001117, -0.00001362, },
		{ -0.00000715, -0.00000499, -0.00000395, -0.00000351, -0.00000345, -0.00000390, -0.00000491, -0.00000604, -0.00000640, -0.00000499, -0.00000121, 0.00000420, 0.00000848, 0.00000810, 0.00000209, -0.00000580, -0.00000951, -0.00000660, -0.00000123, 0.00000016, -0.00000352, -0.00000620, -0.00000199, 0.00000660, 0.00001008, 0.00000256, -0.00001065, -0.00001810, -0.00001436, -0.00000509, 0.00000068, 0.00000101, 0.00000129, 0.00000524, 0.00000920, 0.00000758, 0.00000130, -0.00000277, -0.00000121, 0.00000093, -0.00000294, -0.00001112, -0.00001477, -0.00000925, -0.00000055, 0.00000227, -0.00000159, -0.00000531, -0.00000545, -0.00000621, -0.00000939, -0.00000601, 0.00001238, 0.00003370, 0.00002820, -0.00001666, -0.00006722, -0.00006624, 0.00000312, 0.00008252, 0.00008656, -0.00000401, -0.00010691, -0.00011393, -0.00000742, 0.00011060, 0.00011895, 0.00000579, -0.00011532, -0.00012349, -0.00001630, 0.00009244, 0.00009689, 0.00000651, -0.00007705, -0.00007693, -0.00001578, 0.00002766, 0.00001556, -0.00001615, -0.00001774, 0.00000705, 0.00001211, -0.00002186, -0.00005530, -0.00003966, 0.00001426, 0.00004103, -0.00000090, -0.00007241, -0.00009401, -0.00004159, 0.00001621, -0.00000965, -0.00012310, -0.00022386, -0.00020160, -0.00004573, 0.00014278, 0.00024562, 0.00022734, 0.00014636, 0.00008197, 0.00006315, 0.00006561, 0.00006103, 0.00005349, 0.00006715, 0.00010963, 0.00015687, 0.00017362, 0.00014338, 0.00007798, 0.00000358, -0.00005757, -0.00009429, -0.00010425, -0.00009108, -0.00006395, -0.00003551, -0.00001628, -0.00000960, -0.00001168, -0.00001615, -0.00001856, -0.00001772, -0.00001452, -0.00001055, },
		{ -0.00000508, -0.00000537, -0.00000525, -0.00000491, -0.00000456, -0.00000414, -0.00000324, -0.00000129, 0.00000187, 0.00000503, 0.00000574, 0.00000221, -0.00000402, -0.00000828, -0.00000690, -0.00000160, 0.00000159, -0.00000105, -0.00000605, -0.00000611, 0.00000097, 0.00000843, 0.00000730, -0.00000323, -0.00001421, -0.00001634, -0.00000965, -0.00000201, 0.00000089, 0.00000137, 0.00000452, 0.00000977, 0.00001094, 0.00000474, -0.00000406, -0.00000804, -0.00000609, -0.00000384, -0.00000534, -0.00000774, -0.00000628, -0.00000226, -0.00000180, -0.00000671, -0.00001063, -0.00000718, 0.00000096, 0.00000562, 0.00000496, 0.00000555, 0.00000941, 0.00000543, -0.00001661, -0.00004298, -0.00003977, 0.00000790, 0.00006357, 0.00006421, -0.00000963, -0.00009527, -0.00009984, -0.00000172, 0.00011021, 0.00011805, 0.00000175, -0.00012794, -0.00013806, -0.00001428, 0.00011951, 0.00012925, 0.00000995, -0.00011383, -0.00012335, -0.00002601, 0.00006669, 0.00006757, -0.00000072, -0.00005081, -0.00003866, -0.00000317, -0.00000094, -0.00002922, -0.00003593, 0.00000145, 0.00003634, 0.00000966, -0.00006677, -0.00011069, -0.00006481, 0.00002324, 0.00003942, -0.00007106, -0.00022830, -0.00028224, -0.00015608, 0.00007815, 0.00026568, 0.00030481, 0.00021858, 0.00010722, 0.00004855, 0.00004830, 0.00007050, 0.00009228, 0.00011925, 0.00015772, 0.00018960, 0.00018291, 0.00012357, 0.00003115, -0.00005635, -0.00010995, -0.00012337, -0.00010716, -0.00007634, -0.00004387, -0.00001952, -0.00000874, -0.00001045, -0.00001772, -0.00002279, -0.00002208, -0.00001708, -0.00001127, -0.00000701, -0.00000482, -0.00000421, -0.00000452, },
		{ -0.00000407, -0.00000319, -0.00000272, -0.00000261, -0.00000224, -0.00000083, 0.00000160, 0.00000352, 0.00000277, -0.00000115, -0.00000550, -0.00000639, -0.00000298, 0.00000076, 0.00000002, -0.00000484, -0.00000766, -0.00000329, 0.00000542, 0.00000942, 0.00000309, -0.00000902, -0.00001650, -0.00001394, -0.00000565, 0.00000071, 0.00000298, 0.00000497, 0.00000890, 0.00001088, 0.00000588, -0.00000438, -0.00001187, -0.00001098, -0.00000466, -0.00000033, -0.00000116, -0.00000414, -0.00000604, -0.00000761, -0.00001011, -0.00001036, -0.00000386, 0.00000717, 0.00001321, 0.00000797, -0.00000321, -0.00000988, -0.00001028, -0.00001243, -0.00001855, -0.00001549, 0.00000883, 0.00003936, 0.00003804, -0.00001265, -0.00007364, -0.00007580, 0.00000329, 0.00009638, 0.00010223, -0.00000386, -0.00012599, -0.00013524, -0.00000844, 0.00013400, 0.00014532, 0.00000760, -0.00014389, -0.00015954, -0.00003042, 0.00010730, 0.00012003, 0.00001243, -0.00009286, -0.00009678, -0.00002220, 0.00003314, 0.00001828, -0.00002329, -0.00002631, 0.00000568, 0.00000923, -0.00004534, -0.00010096, -0.00007868, 0.00001314, 0.00006407, -0.00002353, -0.00021128, -0.00033664, -0.00026117, -0.00000679, 0.00025854, 0.00036904, 0.00029498, 0.00014214, 0.00003928, 0.00003538, 0.00009317, 0.00015429, 0.00019237, 0.00020928, 0.00020343, 0.00016024, 0.00007396, -0.00003250, -0.00011758, -0.00015145, -0.00013510, -0.00009273, -0.00005031, -0.00002178, -0.00000941, -0.00000991, -0.00001761, -0.00002542, -0.00002730, -0.00002192, -0.00001318, -0.00000640, -0.00000399, -0.00000466, -0.00000593, -0.00000648, -0.00000633, -0.00000583, -0.00000505, },
		{ -0.00000305, -0.00000333, -0.00000279, -0.00000077, 0.00000201, 0.00000342, 0.00000168, -0.00000241, -0.00000546, -0.00000474, -0.00000153, -0.00000019, -0.00000304, -0.00000672, -0.00000540, 0.00000191, 0.00000862, 0.00000682, -0.00000373, -0.00001440, -0.00001630, -0.00000911, -0.00000009, 0.00000467, 0.00000581, 0.00000700, 0.00000804, 0.00000484, -0.00000373, -0.00001197, -0.00001266, -0.00000553, 0.00000197, 0.00000277, -0.00000287, -0.00000925, -0.00001213, -0.00001116, -0.00000678, 0.00000131, 0.00001056, 0.00001361, 0.00000473, -0.00001137, -0.00002148, -0.00001754, -0.00000587, 0.00000139, 0.00000238, 0.00000622, 0.00001514, 0.00001379, -0.00001217, -0.00004674, -0.00004764, 0.00000574, 0.00007202, 0.00007579, -0.00000865, -0.00010938, -0.00011615, -0.00000100, 0.00013216, 0.00014166, 0.00000038, -0.00016028, -0.00017717, -0.00002733, 0.00014144, 0.00016144, 0.00001869, -0.00013748, -0.00015622, -0.00003921, 0.00007731, 0.00008032, -0.00000572, -0.00006896, -0.00005023, -0.00000225, -0.00000596, -0.00005854, -0.00007757, -0.00001736, 0.00005084, 0.00000625, -0.00017200, -0.00034726, -0.00034016, -0.00010610, 0.00020938, 0.00039646, 0.00036212, 0.00019274, 0.00005438, 0.00004478, 0.00013601, 0.00023496, 0.00027527, 0.00025108, 0.00018648, 0.00009856, -0.00000424, -0.00010300, -0.00016589, -0.00017075, -0.00012606, -0.00006529, -0.00002099, -0.00000485, -0.00000895, -0.00001969, -0.00002810, -0.00003064, -0.00002654, -0.00001748, -0.00000791, -0.00000270, -0.00000332, -0.00000684, -0.00000911, -0.00000844, -0.00000618, -0.00000434, -0.00000360, -0.00000337, -0.00000309, -0.00000288, },
		{ -0.00000174, 0.00000109, 0.00000360, 0.00000348, 0.00000007, -0.00000432, -0.00000611, -0.00000412, -0.00000124, -0.00000129, -0.00000421, -0.00000544, -0.00000130, 0.00000548, 0.00000748, 0.00000065, -0.00001025, -0.00001577, -0.00001141, -0.00000182, 0.00000500, 0.00000608, 0.00000465, 0.00000390, 0.00000244, -0.00000246, -0.00000901, -0.00001129, -0.00000638, 0.00000107, 0.00000326, -0.00000223, -0.00001010, -0.00001341, -0.00001001, -0.00000261, 0.00000520, 0.00001028, 0.00000882, -0.00000165, -0.00001676, -0.00002494, -0.00001743, 0.00000076, 0.00001388, 0.00001198, 0.00000155, -0.00000433, -0.00000416, -0.00000861, -0.00002024, -0.00002145, 0.00000487, 0.00004243, 0.00004516, -0.00001067, -0.00008153, -0.00008580, 0.00000521, 0.00011389, 0.00011970, -0.00000896, -0.00015862, -0.00017231, -0.00001771, 0.00016133, 0.00018227, 0.00001570, -0.00017617, -0.00020428, -0.00004876, 0.00012488, 0.00014493, 0.00001071, -0.00012257, -0.00012342, -0.00002065, 0.00005031, 0.00001496, -0.00006110, -0.00006935, -0.00000676, 0.00001014, -0.00011046, -0.00030117, -0.00037582, -0.00021421, 0.00010439, 0.00036382, 0.00040702, 0.00026799, 0.00011820, 0.00009782, 0.00020306, 0.00032073, 0.00034726, 0.00026648, 0.00013147, -0.00000004, -0.00010156, -0.00016667, -0.00018819, -0.00016094, -0.00009842, -0.00003383, 0.00000171, 0.00000130, -0.00001798, -0.00003419, -0.00003729, -0.00003016, -0.00001966, -0.00001026, -0.00000413, -0.00000273, -0.00000578, -0.00001012, -0.00001162, -0.00000888, -0.00000442, -0.00000175, -0.00000206, -0.00000374, -0.00000472, -0.00000447, -0.00000388, -0.00000364, -0.00000327, },
		{ 0.00000387, 0.00000161, -0.00000297, -0.00000645, -0.00000602, -0.00000261, -0.00000020, -0.00000139, -0.00000405, -0.00000355, 0.00000127, 0.00000556, 0.00000342, -0.00000497, -0.00001240, -0.00001195, -0.00000435, 0.00000319, 0.00000500, 0.00000233, 0.00000002, -0.00000021, -0.00000111, -0.00000458, -0.00000790, -0.00000691, -0.00000214, 0.00000088, -0.00000198, -0.00000815, -0.00001110, -0.00000745, -0.00000003, 0.00000579, 0.00000661, 0.00000172, -0.00000787, -0.00001826, -0.00002188, -0.00001268, 0.00000543, 0.00001822, 0.00001394, -0.00000317, -0.00001609, -0.00001382, -0.00000346, 0.00000086, -0.00000185, 0.00000123, 0.00001387, 0.00001697, -0.00000919, -0.00004845, -0.00005146, 0.00000790, 0.00008300, 0.00008526, -0.00001686, -0.00013847, -0.00014651, -0.00000488, 0.00016167, 0.00017772, 0.00000422, -0.00020035, -0.00023012, -0.00004777, 0.00016672, 0.00019805, 0.00001998, -0.00017940, -0.00020000, -0.00003910, 0.00011822, 0.00010920, -0.00003445, -0.00014058, -0.00010521, -0.00001220, -0.00002796, -0.00019454, -0.00035980, -0.00032755, -0.00006526, 0.00025482, 0.00042132, 0.00037862, 0.00025505, 0.00021372, 0.00029277, 0.00038980, 0.00038124, 0.00023926, 0.00003922, -0.00012236, -0.00020115, -0.00020961, -0.00017705, -0.00012249, -0.00005979, -0.00000931, 0.00000926, -0.00000526, -0.00003220, -0.00004692, -0.00004075, -0.00002331, -0.00000878, -0.00000328, -0.00000432, -0.00000750, -0.00001060, -0.00001238, -0.00001131, -0.00000702, -0.00000199, 0.00000016, -0.00000181, -0.00000550, -0.00000746, -0.00000650, -0.00000419, -0.00000263, -0.00000223, -0.00000167, 0.00000023, 0.00000283, },
		{ -0.00000559, -0.00000675, -0.00000392, -0.00000010, 0.00000068, -0.00000208, -0.00000448, -0.00000263, 0.00000219, 0.00000433, 0.00000016, -0.00000715, -0.00001062, -0.00000709, -0.00000067, 0.00000226, 0.00000023, -0.00000253, -0.00000218, -0.00000009, -0.00000036, -0.00000382, -0.00000675, -0.00000591, -0.00000292, -0.00000202, -0.00000432, -0.00000631, -0.00000451, -0.00000002, 0.00000271, 0.00000075, -0.00000496, -0.00001152, -0.00001595, -0.00001506, -0.00000649, 0.00000702, 0.00001585, 0.00001066, -0.00000622, -0.00001954, -0.00001571, 0.00000136, 0.00001349, 0.00000906, -0.00000356, -0.00000826, -0.00000381, -0.00000493, -0.00001709, -0.00002053, 0.00000618, 0.00004620, 0.00004670, -0.00002034, -0.00010379, -0.00010637, 0.00000670, 0.00014141, 0.00014914, -0.00001206, -0.00020381, -0.00022789, -0.00003665, 0.00019346, 0.00022740, 0.00001843, -0.00022905, -0.00026359, -0.00005046, 0.00018835, 0.00020688, -0.00000667, -0.00022444, -0.00023027, -0.00005285, 0.00007182, -0.00003572, -0.00029598, -0.00044477, -0.00029912, 0.00006636, 0.00040449, 0.00053424, 0.00048217, 0.00040345, 0.00039682, 0.00041601, 0.00034954, 0.00015822, -0.00007962, -0.00024524, -0.00028191, -0.00022188, -0.00013412, -0.00006418, -0.00002115, -0.00000079, -0.00000225, -0.00002198, -0.00004465, -0.00005118, -0.00003610, -0.00001254, 0.00000160, 0.00000015, -0.00000915, -0.00001567, -0.00001563, -0.00001171, -0.00000723, -0.00000332, -0.00000059, -0.00000053, -0.00000373, -0.00000800, -0.00000978, -0.00000758, -0.00000341, -0.00000059, -0.00000042, -0.00000134, -0.00000101, 0.00000110, 0.00000306, 0.00000238, -0.00000137, },
		{ -0.00000092, 0.00000148, -0.00000030, -0.00000413, -0.00000515, -0.00000135, 0.00000358, 0.00000393, -0.00000152, -0.00000781, -0.00000926, -0.00000567, -0.00000194, -0.00000179, -0.00000350, -0.00000293, 0.00000044, 0.00000247, -0.00000021, -0.00000563, -0.00000852, -0.00000649, -0.00000239, -0.00000032, -0.00000099, -0.00000207, -0.00000224, -0.00000292, -0.00000566, -0.00000946, -0.00001172, -0.00001098, -0.00000741, -0.00000146, 0.00000556, 0.00000950, 0.00000518, -0.00000674, -0.00001628, -0.00001220, 0.00000429, 0.00001720, 0.00001181, -0.00000757, -0.00002080, -0.00001502, 0.00000030, 0.00000661, 0.00000157, 0.00000093, 0.00001135, 0.00001249, -0.00001814, -0.00006194, -0.00006149, 0.00001307, 0.00010430, 0.00010366, -0.00002786, -0.00018474, -0.00019849, -0.00001919, 0.00019806, 0.00022604, 0.00000615, -0.00026086, -0.00029972, -0.00005002, 0.00024852, 0.00028773, 0.00001243, -0.00030995, -0.00035930, -0.00009801, 0.00018161, 0.00015431, -0.00020056, -0.00056555, -0.00058422, -0.00018765, 0.00036485, 0.00073944, 0.00080265, 0.00066449, 0.00050083, 0.00037509, 0.00023142, 0.00002111, -0.00020573, -0.00034008, -0.00032347, -0.00020083, -0.00007105, -0.00000159, 0.00000652, -0.00001184, -0.00003019, -0.00004255, -0.00004842, -0.00004284, -0.00002415, -0.00000270, 0.00000623, -0.00000205, -0.00001688, -0.00002377, -0.00001830, -0.00000755, -0.00000041, 0.00000056, -0.00000184, -0.00000478, -0.00000762, -0.00000989, -0.00000992, -0.00000664, -0.00000174, 0.00000140, 0.00000099, -0.00000142, -0.00000264, -0.00000110, 0.00000171, 0.00000275, 0.00000039, -0.00000384, -0.00000639, -0.00000492, },
		{ -0.00000294, -0.00000589, -0.00000400, 0.00000173, 0.00000561, 0.00000304, -0.00000432, -0.00001025, -0.00001053, -0.00000694, -0.00000402, -0.00000335, -0.00000243, 0.00000068, 0.00000361, 0.00000224, -0.00000360, -0.00000889, -0.00000857, -0.00000314, 0.00000213, 0.00000286, -0.00000072, -0.00000546, -0.00000908, -0.00001109, -0.00001125, -0.00000904, -0.00000501, -0.00000113, 0.00000111, 0.00000205, 0.00000215, 0.00000011, -0.00000492, -0.00000927, -0.00000636, 0.00000469, 0.00001389, 0.00000891, -0.00000933, -0.00002351, -0.00001740, 0.00000434, 0.00001928, 0.00001262, -0.00000564, -0.00001481, -0.00001122, -0.00001113, -0.00002175, -0.00002170, 0.00001218, 0.00005805, 0.00005202, -0.00003815, -0.00014649, -0.00014932, -0.00000126, 0.00017807, 0.00019443, -0.00001294, -0.00026679, -0.00029996, -0.00003734, 0.00028629, 0.00033378, 0.00001581, -0.00038307, -0.00046613, -0.00013234, 0.00028997, 0.00034465, -0.00009857, -0.00068840, -0.00089427, -0.00047642, 0.00031794, 0.00098989, 0.00120065, 0.00097853, 0.00058728, 0.00025194, 0.00002001, -0.00016336, -0.00031511, -0.00037965, -0.00031160, -0.00015253, -0.00000734, 0.00004668, 0.00001570, -0.00003857, -0.00006552, -0.00005909, -0.00003969, -0.00002238, -0.00000876, 0.00000098, 0.00000144, -0.00000972, -0.00002381, -0.00002750, -0.00001681, -0.00000132, 0.00000628, 0.00000243, -0.00000626, -0.00001183, -0.00001214, -0.00000973, -0.00000682, -0.00000354, -0.00000005, 0.00000196, 0.00000095, -0.00000209, -0.00000408, -0.00000291, 0.00000047, 0.00000290, 0.00000195, -0.00000183, -0.00000530, -0.00000544, -0.00000221, 0.00000107, 0.00000079, },
		{ -0.00000037, 0.00000520, 0.00000549, -0.00000116, -0.00000991, -0.00001411, -0.00001179, -0.00000662, -0.00000291, -0.00000107, 0.00000104, 0.00000351, 0.00000340, -0.00000109, -0.00000695, -0.00000864, -0.00000423, 0.00000204, 0.00000393, -0.00000065, -0.00000825, -0.00001374, -0.00001444, -0.00001085, -0.00000504, 0.00000036, 0.00000276, 0.00000127, -0.00000207, -0.00000389, -0.00000310, -0.00000181, -0.00000168, -0.00000087, 0.00000296, 0.00000685, 0.00000361, -0.00000838, -0.00001871, -0.00001382, 0.00000564, 0.00002085, 0.00001375, -0.00001088, -0.00002841, -0.00002200, -0.00000196, 0.00000898, 0.00000574, 0.00000460, 0.00001234, 0.00000548, -0.00003993, -0.00009694, -0.00009204, 0.00001130, 0.00013671, 0.00014036, -0.00003244, -0.00024391, -0.00026445, -0.00001660, 0.00029223, 0.00033439, 0.00000182, -0.00042931, -0.00052750, -0.00014395, 0.00038192, 0.00050038, -0.00001747, -0.00080829, -0.00119202, -0.00075697, 0.00028232, 0.00126988, 0.00164058, 0.00131297, 0.00063996, 0.00004678, -0.00027326, -0.00037627, -0.00038514, -0.00034571, -0.00024230, -0.00009047, 0.00003462, 0.00006403, 0.00000488, -0.00007000, -0.00009511, -0.00006500, -0.00001990, 0.00000451, 0.00000449, -0.00000474, -0.00001301, -0.00002025, -0.00002574, -0.00002358, -0.00001107, 0.00000392, 0.00000905, 0.00000059, -0.00001272, -0.00001916, -0.00001528, -0.00000686, -0.00000091, 0.00000079, 0.00000056, 0.00000013, -0.00000092, -0.00000294, -0.00000446, -0.00000354, -0.00000033, 0.00000266, 0.00000283, -0.00000013, -0.00000384, -0.00000528, -0.00000337, -0.00000013, 0.00000095, -0.00000158, -0.00000510, -0.00000516, },
		{ 0.00000070, -0.00000866, -0.00001571, -0.00001548, -0.00000937, -0.00000286, 0.00000054, 0.00000178, 0.00000296, 0.00000353, 0.00000127, -0.00000354, -0.00000692, -0.00000538, -0.00000055, 0.00000190, -0.00000183, -0.00000962, -0.00001554, -0.00001527, -0.00000924, -0.00000139, 0.00000396, 0.00000430, 0.00000002, -0.00000543, -0.00000747, -0.00000425, 0.00000125, 0.00000408, 0.00000271, 0.00000026, -0.00000046, -0.00000128, -0.00000540, -0.00001023, -0.00000794, 0.00000381, 0.00001418, 0.00000850, -0.00001279, -0.00002932, -0.00002135, 0.00000601, 0.00002550, 0.00001783, -0.00000679, -0.00002391, -0.00002736, -0.00003319, -0.00004641, -0.00003921, 0.00001472, 0.00008228, 0.00007663, -0.00004579, -0.00019574, -0.00020194, 0.00000487, 0.00026280, 0.00028934, -0.00002502, -0.00043734, -0.00052929, -0.00012849, 0.00044119, 0.00059020, 0.00002022, -0.00091531, -0.00143511, -0.00098259, 0.00027421, 0.00155239, 0.00207138, 0.00162637, 0.00064863, -0.00022187, -0.00061847, -0.00059196, -0.00040091, -0.00023483, -0.00012356, -0.00003185, 0.00003591, 0.00004059, -0.00002097, -0.00009110, -0.00010558, -0.00005735, 0.00000394, 0.00002832, 0.00001045, -0.00001913, -0.00003316, -0.00002949, -0.00002001, -0.00001153, -0.00000299, 0.00000461, 0.00000489, -0.00000526, -0.00001861, -0.00002331, -0.00001524, -0.00000196, 0.00000583, 0.00000467, -0.00000059, -0.00000393, -0.00000407, -0.00000331, -0.00000315, -0.00000271, -0.00000085, 0.00000163, 0.00000261, 0.00000099, -0.00000211, -0.00000431, -0.00000396, -0.00000166, 0.00000013, -0.00000069, -0.00000342, -0.00000469, -0.00000189, 0.00000330, 0.00000545, },
		{ -0.00001714, -0.00001165, -0.00000350, 0.00000173, 0.00000287, 0.00000267, 0.00000312, 0.00000290, 0.00000008, -0.00000415, -0.00000620, -0.00000465, -0.00000253, -0.00000403, -0.00000942, -0.00001423, -0.00001368, -0.00000732, 0.00000071, 0.00000536, 0.00000418, -0.00000142, -0.00000720, -0.00000860, -0.00000405, 0.00000313, 0.00000688, 0.00000409, -0.00000205, -0.00000550, -0.00000416, -0.00000178, -0.00000197, -0.00000283, -0.00000025, 0.00000379, 0.00000120, -0.00001074, -0.00002081, -0.00001357, 0.00001000, 0.00002726, 0.00001616, -0.00001792, -0.00004488, -0.00004259, -0.00002079, -0.00000485, -0.00000074, 0.00000857, 0.00002605, 0.00001731, -0.00004863, -0.00013173, -0.00012730, 0.00001920, 0.00020195, 0.00020912, -0.00005555, -0.00040232, -0.00047056, -0.00009071, 0.00045379, 0.00059413, 0.00000336, -0.00099520, -0.00158410, -0.00111301, 0.00030243, 0.00180250, 0.00243518, 0.00187647, 0.00061241, -0.00051872, -0.00097161, -0.00078342, -0.00035999, -0.00006094, 0.00002608, 0.00000742, -0.00001293, -0.00002373, -0.00005206, -0.00008751, -0.00008752, -0.00003806, 0.00002105, 0.00003854, 0.00000639, -0.00003696, -0.00005159, -0.00003336, -0.00000714, 0.00000557, 0.00000408, -0.00000159, -0.00000687, -0.00001365, -0.00002076, -0.00002130, -0.00001133, 0.00000297, 0.00001039, 0.00000631, -0.00000319, -0.00000876, -0.00000721, -0.00000269, -0.00000032, -0.00000082, -0.00000150, -0.00000047, 0.00000128, 0.00000167, -0.00000003, -0.00000252, -0.00000389, -0.00000326, -0.00000153, -0.00000061, -0.00000149, -0.00000286, -0.00000227, 0.00000077, 0.00000310, 0.00000052, -0.00000731, -0.00001527, },
		{ 0.00000196, 0.00000385, 0.00000290, 0.00000265, 0.00000354, 0.00000279, -0.00000126, -0.00000635, -0.00000881, -0.00000801, -0.00000696, -0.00000823, -0.00001056, -0.00001020, -0.00000550, 0.00000098, 0.00000470, 0.00000319, -0.00000215, -0.00000720, -0.00000800, -0.00000347, 0.00000332, 0.00000688, 0.00000391, -0.00000302, -0.00000724, -0.00000480, 0.00000100, 0.00000332, -0.00000015, -0.00000480, -0.00000576, -0.00000453, -0.00000573, -0.00000811, -0.00000396, 0.00000882, 0.00001761, 0.00000574, -0.00002510, -0.00004924, -0.00004159, -0.00000729, 0.00002090, 0.00001855, -0.00000425, -0.00002082, -0.00002624, -0.00004039, -0.00006496, -0.00005722, 0.00002088, 0.00012080, 0.00011212, -0.00007910, -0.00032814, -0.00036426, -0.00004314, 0.00041194, 0.00050892, -0.00006450, -0.00103194, -0.00161111, -0.00112376, 0.00036545, 0.00198356, 0.00267866, 0.00202903, 0.00054074, -0.00079782, -0.00128286, -0.00092801, -0.00027394, 0.00014670, 0.00018160, 0.00001757, -0.00010842, -0.00011894, -0.00007713, -0.00005050, -0.00003863, -0.00001328, 0.00002035, 0.00002710, -0.00000671, -0.00004972, -0.00005964, -0.00002957, 0.00000852, 0.00002151, 0.00000672, -0.00001429, -0.00002371, -0.00002172, -0.00001691, -0.00001211, -0.00000442, 0.00000531, 0.00000999, 0.00000472, -0.00000591, -0.00001199, -0.00000886, -0.00000127, 0.00000290, 0.00000113, -0.00000257, -0.00000336, -0.00000067, 0.00000227, 0.00000246, -0.00000022, -0.00000348, -0.00000489, -0.00000368, -0.00000121, 0.00000029, -0.00000020, -0.00000149, -0.00000166, -0.00000088, -0.00000185, -0.00000664, -0.00001321, -0.00001630, -0.00001264, -0.00000463, },
		{ 0.00000245, 0.00000337, 0.00000401, 0.00000093, -0.00000575, -0.00001171, -0.00001302, -0.00001021, -0.00000704, -0.00000593, -0.00000564, -0.00000371, -0.00000008, 0.00000259, 0.00000184, -0.00000195, -0.00000580, -0.00000655, -0.00000323, 0.00000199, 0.00000503, 0.00000307, -0.00000240, -0.00000609, -0.00000392, 0.00000199, 0.00000470, 0.00000003, -0.00000805, -0.00001135, -0.00000685, 0.00000019, 0.00000341, 0.00000308, 0.00000325, 0.00000242, -0.00000709, -0.00002628, -0.00004014, -0.00002976, 0.00000283, 0.00002874, 0.00001973, -0.00001871, -0.00005031, -0.00004827, -0.00002443, -0.00000872, -0.00000556, 0.00000880, 0.00003510, 0.00001953, -0.00008731, -0.00022737, -0.00023409, -0.00000211, 0.00031720, 0.00034963, -0.00016586, -0.00101185, -0.00150668, -0.00101210, 0.00045130, 0.00206489, 0.00276433, 0.00206489, 0.00045130, -0.00101210, -0.00150668, -0.00101185, -0.00016586, 0.00034963, 0.00031720, -0.00000211, -0.00023409, -0.00022737, -0.00008731, 0.00001953, 0.00003510, 0.00000880, -0.00000556, -0.00000872, -0.00002443, -0.00004827, -0.00005031, -0.00001871, 0.00001973, 0.00002874, 0.00000283, -0.00002976, -0.00004014, -0.00002628, -0.00000709, 0.00000242, 0.00000325, 0.00000308, 0.00000341, 0.00000019, -0.00000685, -0.00001135, -0.00000805, 0.00000003, 0.00000470, 0.00000199, -0.00000392, -0.00000609, -0.00000240, 0.00000307, 0.00000503, 0.00000199, -0.00000323, -0.00000655, -0.00000580, -0.00000195, 0.00000184, 0.00000259, -0.00000008, -0.00000371, -0.00000564, -0.00000593, -0.00000704, -0.00001021, -0.00001302, -0.00001171, -0.00000575, 0.00000093, 0.00000401, 0.00000337, },
		{ 0.00000196, -0.00000463, -0.00001264, -0.00001630, -0.00001321, -0.00000664, -0.00000185, -0.00000088, -0.00000166, -0.00000149, -0.00000020, 0.00000029, -0.00000121, -0.00000368, -0.00000489, -0.00000348, -0.00000022, 0.00000246, 0.00000227, -0.00000067, -0.00000336, -0.00000257, 0.00000113, 0.00000290, -0.00000127, -0.00000886, -0.00001199, -0.00000591, 0.00000472, 0.00000999, 0.00000531, -0.00000442, -0.00001211, -0.00001691, -0.00002172, -0.00002371, -0.00001429, 0.00000672, 0.00002151, 0.00000852, -0.00002957, -0.00005964, -0.00004972, -0.00000671, 0.00002710, 0.00002035, -0.00001328, -0.00003863, -0.00005050, -0.00007713, -0.00011894, -0.00010842, 0.00001757, 0.00018160, 0.00014670, -0.00027394, -0.00092801, -0.00128286, -0.00079782, 0.00054074, 0.00202903, 0.00267866, 0.00198356, 0.00036545, -0.00112376, -0.00161111, -0.00103194, -0.00006450, 0.00050892, 0.00041194, -0.00004314, -0.00036426, -0.00032814, -0.00007910, 0.00011212, 0.00012080, 0.00002088, -0.00005722, -0.00006496, -0.00004039, -0.00002624, -0.00002082, -0.00000425, 0.00001855, 0.00002090, -0.00000729, -0.00004159, -0.00004924, -0.00002510, 0.00000574, 0.00001761, 0.00000882, -0.00000396, -0.00000811, -0.00000573, -0.00000453, -0.00000576, -0.00000480, -0.00000015, 0.00000332, 0.00000100, -0.00000480, -0.00000724, -0.00000302, 0.00000391, 0.00000688, 0.00000332, -0.00000347, -0.00000800, -0.00000720, -0.00000215, 0.00000319, 0.00000470, 0.00000098, -0.00000550, -0.00001020, -0.00001056, -0.00000823, -0.00000696, -0.00000801, -0.00000881, -0.00000635, -0.00000126, 0.00000279, 0.00000354, 0.00000265, 0.00000290, 0.00000385, },
		{ -0.00001714, -0.00001527, -0.00000731, 0.00000052, 0.00000310, 0.00000077, -0.00000227, -0.00000286, -0.00000149, -0.00000061, -0.00000153, -0.00000326, -0.00000389, -0.00000252, -0.00000003, 0.00000167, 0.00000128, -0.00000047, -0.00000150, -0.00000082, -0.00000032, -0.00000269, -0.00000721, -0.00000876, -0.00000319, 0.00000631, 0.00001039, 0.00000297, -0.00001133, -0.00002130, -0.00002076, -0.00001365, -0.00000687, -0.00000159, 0.00000408, 0.00000557, -0.00000714, -0.00003336, -0.00005159, -0.00003696, 0.00000639, 0.00003854, 0.00002105, -0.00003806, -0.00008752, -0.00008751, -0.00005206, -0.00002373, -0.00001293, 0.00000742, 0.00002608, -0.00006094, -0.00035999, -0.00078342, -0.00097161, -0.00051872, 0.00061241, 0.00187647, 0.00243518, 0.00180250, 0.00030243, -0.00111301, -0.00158410, -0.00099520, 0.00000336, 0.00059413, 0.00045379, -0.00009071, -0.00047056, -0.00040232, -0.00005555, 0.00020912, 0.00020195, 0.00001920, -0.00012730, -0.00013173, -0.00004863, 0.00001731, 0.00002605, 0.00000857, -0.00000074, -0.00000485, -0.00002079, -0.00004259, -0.00004488, -0.00001792, 0.00001616, 0.00002726, 0.00001000, -0.00001357, -0.00002081, -0.00001074, 0.00000120, 0.00000379, -0.00000025, -0.00000283, -0.00000197, -0.00000178, -0.00000416, -0.00000550, -0.00000205, 0.00000409, 0.00000688, 0.00000313, -0.00000405, -0.00000860, -0.00000720, -0.00000142, 0.00000418, 0.00000536, 0.00000071, -0.00000732, -0.00001368, -0.00001423, -0.00000942, -0.00000403, -0.00000253, -0.00000465, -0.00000620, -0.00000415, 0.00000008, 0.00000290, 0.00000312, 0.00000267, 0.00000287, 0.00000173, -0.00000350, -0.00001165, },
		{ 0.00000070, 0.00000545, 0.00000330, -0.00000189, -0.00000469, -0.00000342, -0.00000069, 0.00000013, -0.00000166, -0.00000396, -0.00000431, -0.00000211, 0.00000099, 0.00000261, 0.00000163, -0.00000085, -0.00000271, -0.00000315, -0.00000331, -0.00000407, -0.00000393, -0.00000059, 0.00000467, 0.00000583, -0.00000196, -0.00001524, -0.00002331, -0.00001861, -0.00000526, 0.00000489, 0.00000461, -0.00000299, -0.00001153, -0.00002001, -0.00002949, -0.00003316, -0.00001913, 0.00001045, 0.00002832, 0.00000394, -0.00005735, -0.00010558, -0.00009110, -0.00002097, 0.00004059, 0.00003591, -0.00003185, -0.00012356, -0.00023483, -0.00040091, -0.00059196, -0.00061847, -0.00022187, 0.00064863, 0.00162637, 0.00207138, 0.00155239, 0.00027421, -0.00098259, -0.00143511, -0.00091531, 0.00002022, 0.00059020, 0.00044119, -0.00012849, -0.00052929, -0.00043734, -0.00002502, 0.00028934, 0.00026280, 0.00000487, -0.00020194, -0.00019574, -0.00004579, 0.00007663, 0.00008228, 0.00001472, -0.00003921, -0.00004641, -0.00003319, -0.00002736, -0.00002391, -0.00000679, 0.00001783, 0.00002550, 0.00000601, -0.00002135, -0.00002932, -0.00001279, 0.00000850, 0.00001418, 0.00000381, -0.00000794, -0.00001023, -0.00000540, -0.00000128, -0.00000046, 0.00000026, 0.00000271, 0.00000408, 0.00000125, -0.00000425, -0.00000747, -0.00000543, 0.00000002, 0.00000430, 0.00000396, -0.00000139, -0.00000924, -0.00001527, -0.00001554, -0.00000962, -0.00000183, 0.00000190, -0.00000055, -0.00000538, -0.00000692, -0.00000354, 0.00000127, 0.00000353, 0.00000296, 0.00000178, 0.00000054, -0.00000286, -0.00000937, -0.00001548, -0.00001571, -0.00000866, },
		{ -0.00000037, -0.00000516, -0.00000510, -0.00000158, 0.00000095, -0.00000013, -0.00000337, -0.00000528, -0.00000384, -0.00000013, 0.00000283, 0.00000266, -0.00000033, -0.00000354, -0.00000446, -0.00000294, -0.00000092, 0.00000013, 0.00000056, 0.00000079, -0.00000091, -0.00000686, -0.00001528, -0.00001916, -0.00001272, 0.00000059, 0.00000905, 0.00000392, -0.00001107, -0.00002358, -0.00002574, -0.00002025, -0.00001301, -0.00000474, 0.00000449, 0.00000451, -0.00001990, -0.00006500, -0.00009511, -0.00007000, 0.00000488, 0.00006403, 0.00003462, -0.00009047, -0.00024230, -0.00034571, -0.00038514, -0.00037627, -0.00027326, 0.00004678, 0.00063996, 0.00131297, 0.00164058, 0.00126988, 0.00028232, -0.00075697, -0.00119202, -0.00080829, -0.00001747, 0.00050038, 0.00038192, -0.00014395, -0.00052750, -0.00042931, 0.00000182, 0.00033439, 0.00029223, -0.00001660, -0.00026445, -0.00024391, -0.00003244, 0.00014036, 0.00013671, 0.00001130, -0.00009204, -0.00009694, -0.00003993, 0.00000548, 0.00001234, 0.00000460, 0.00000574, 0.00000898, -0.00000196, -0.00002200, -0.00002841, -0.00001088, 0.00001375, 0.00002085, 0.00000564, -0.00001382, -0.00001871, -0.00000838, 0.00000361, 0.00000685, 0.00000296, -0.00000087, -0.00000168, -0.00000181, -0.00000310, -0.00000389, -0.00000207, 0.00000127, 0.00000276, 0.00000036, -0.00000504, -0.00001085, -0.00001444, -0.00001374, -0.00000825, -0.00000065, 0.00000393, 0.00000204, -0.00000423, -0.00000864, -0.00000695, -0.00000109, 0.00000340, 0.00000351, 0.00000104, -0.00000107, -0.00000291, -0.00000662, -0.00001179, -0.00001411, -0.00000991, -0.00000116, 0.00000549, 0.00000520, },
		{ -0.00000294, 0.00000079, 0.00000107, -0.00000221, -0.00000544, -0.00000530, -0.00000183, 0.00000195, 0.00000290, 0.00000047, -0.00000291, -0.00000408, -0.00000209, 0.00000095, 0.00000196, -0.00000005, -0.00000354, -0.00000682, -0.00000973, -0.00001214, -0.00001183, -0.00000626, 0.00000243, 0.00000628, -0.00000132, -0.00001681, -0.00002750, -0.00002381, -0.00000972, 0.00000144, 0.00000098, -0.00000876, -0.00002238, -0.00003969, -0.00005909, -0.00006552, -0.00003857, 0.00001570, 0.00004668, -0.00000734, -0.00015253, -0.00031160, -0.00037965, -0.00031511, -0.00016336, 0.00002001, 0.00025194, 0.00058728, 0.00097853, 0.00120065, 0.00098989, 0.00031794, -0.00047642, -0.00089427, -0.00068840, -0.00009857, 0.00034465, 0.00028997, -0.00013234, -0.00046613, -0.00038307, 0.00001581, 0.00033378, 0.00028629, -0.00003734, -0.00029996, -0.00026679, -0.00001294, 0.00019443, 0.00017807, -0.00000126, -0.00014932, -0.00014649, -0.00003815, 0.00005202, 0.00005805, 0.00001218, -0.00002170, -0.00002175, -0.00001113, -0.00001122, -0.00001481, -0.00000564, 0.00001262, 0.00001928, 0.00000434, -0.00001740, -0.00002351, -0.00000933, 0.00000891, 0.00001389, 0.00000469, -0.00000636, -0.00000927, -0.00000492, 0.00000011, 0.00000215, 0.00000205, 0.00000111, -0.00000113, -0.00000501, -0.00000904, -0.00001125, -0.00001109, -0.00000908, -0.00000546, -0.00000072, 0.00000286, 0.00000213, -0.00000314, -0.00000857, -0.00000889, -0.00000360, 0.00000224, 0.00000361, 0.00000068, -0.00000243, -0.00000335, -0.00000402, -0.00000694, -0.00001053, -0.00001025, -0.00000432, 0.00000304, 0.00000561, 0.00000173, -0.00000400, -0.00000589, },
		{ -0.00000092, -0.00000492, -0.00000639, -0.00000384, 0.00000039, 0.00000275, 0.00000171, -0.00000110, -0.00000264, -0.00000142, 0.00000099, 0.00000140, -0.00000174, -0.00000664, -0.00000992, -0.00000989, -0.00000762, -0.00000478, -0.00000184, 0.00000056, -0.00000041, -0.00000755, -0.00001830, -0.00002377, -0.00001688, -0.00000205, 0.00000623, -0.00000270, -0.00002415, -0.00004284, -0.00004842, -0.00004255, -0.00003019, -0.00001184, 0.00000652, -0.00000159, -0.00007105, -0.00020083, -0.00032347, -0.00034008, -0.00020573, 0.00002111, 0.00023142, 0.00037509, 0.00050083, 0.00066449, 0.00080265, 0.00073944, 0.00036485, -0.00018765, -0.00058422, -0.00056555, -0.00020056, 0.00015431, 0.00018161, -0.00009801, -0.00035930, -0.00030995, 0.00001243, 0.00028773, 0.00024852, -0.00005002, -0.00029972, -0.00026086, 0.00000615, 0.00022604, 0.00019806, -0.00001919, -0.00019849, -0.00018474, -0.00002786, 0.00010366, 0.00010430, 0.00001307, -0.00006149, -0.00006194, -0.00001814, 0.00001249, 0.00001135, 0.00000093, 0.00000157, 0.00000661, 0.00000030, -0.00001502, -0.00002080, -0.00000757, 0.00001181, 0.00001720, 0.00000429, -0.00001220, -0.00001628, -0.00000674, 0.00000518, 0.00000950, 0.00000556, -0.00000146, -0.00000741, -0.00001098, -0.00001172, -0.00000946, -0.00000566, -0.00000292, -0.00000224, -0.00000207, -0.00000099, -0.00000032, -0.00000239, -0.00000649, -0.00000852, -0.00000563, -0.00000021, 0.00000247, 0.00000044, -0.00000293, -0.00000350, -0.00000179, -0.00000194, -0.00000567, -0.00000926, -0.00000781, -0.00000152, 0.00000393, 0.00000358, -0.00000135, -0.00000515, -0.00000413, -0.00000030, 0.00000148, },
		{ -0.00000559, -0.00000137, 0.00000238, 0.00000306, 0.00000110, -0.00000101, -0.00000134, -0.00000042, -0.00000059, -0.00000341, -0.00000758, -0.00000978, -0.00000800, -0.00000373, -0.00000053, -0.00000059, -0.00000332, -0.00000723, -0.00001171, -0.00001563, -0.00001567, -0.00000915, 0.00000015, 0.00000160, -0.00001254, -0.00003610, -0.00005118, -0.00004465, -0.00002198, -0.00000225, -0.00000079, -0.00002115, -0.00006418, -0.00013412, -0.00022188, -0.00028191, -0.00024524, -0.00007962, 0.00015822, 0.00034954, 0.00041601, 0.00039682, 0.00040345, 0.00048217, 0.00053424, 0.00040449, 0.00006636, -0.00029912, -0.00044477, -0.00029598, -0.00003572, 0.00007182, -0.00005285, -0.00023027, -0.00022444, -0.00000667, 0.00020688, 0.00018835, -0.00005046, -0.00026359, -0.00022905, 0.00001843, 0.00022740, 0.00019346, -0.00003665, -0.00022789, -0.00020381, -0.00001206, 0.00014914, 0.00014141, 0.00000670, -0.00010637, -0.00010379, -0.00002034, 0.00004670, 0.00004620, 0.00000618, -0.00002053, -0.00001709, -0.00000493, -0.00000381, -0.00000826, -0.00000356, 0.00000906, 0.00001349, 0.00000136, -0.00001571, -0.00001954, -0.00000622, 0.00001066, 0.00001585, 0.00000702, -0.00000649, -0.00001506, -0.00001595, -0.00001152, -0.00000496, 0.00000075, 0.00000271, -0.00000002, -0.00000451, -0.00000631, -0.00000432, -0.00000202, -0.00000292, -0.00000591, -0.00000675, -0.00000382, -0.00000036, -0.00000009, -0.00000218, -0.00000253, 0.00000023, 0.00000226, -0.00000067, -0.00000709, -0.00001062, -0.00000715, 0.00000016, 0.00000433, 0.00000219, -0.00000263, -0.00000448, -0.00000208, 0.00000068, -0.00000010, -0.00000392, -0.00000675, },
		{ 0.00000387, 0.00000283, 0.00000023, -0.00000167, -0.00000223, -0.00000263, -0.00000419, -0.00000650, -0.00000746, -0.00000550, -0.00000181, 0.00000016, -0.00000199, -0.00000702, -0.00001131, -0.00001238, -0.00001060, -0.00000750, -0.00000432, -0.00000328, -0.00000878, -0.00002331, -0.00004075, -0.00004692, -0.00003220, -0.00000526, 0.00000926, -0.00000931, -0.00005979, -0.00012249, -0.00017705, -0.00020961, -0.00020115, -0.00012236, 0.00003922, 0.00023926, 0.00038124, 0.00038980, 0.00029277, 0.00021372, 0.00025505, 0.00037862, 0.00042132, 0.00025482, -0.00006526, -0.00032755, -0.00035980, -0.00019454, -0.00002796, -0.00001220, -0.00010521, -0.00014058, -0.00003445, 0.00010920, 0.00011822, -0.00003910, -0.00020000, -0.00017940, 0.00001998, 0.00019805, 0.00016672, -0.00004777, -0.00023012, -0.00020035, 0.00000422, 0.00017772, 0.00016167, -0.00000488, -0.00014651, -0.00013847, -0.00001686, 0.00008526, 0.00008300, 0.00000790, -0.00005146, -0.00004845, -0.00000919, 0.00001697, 0.00001387, 0.00000123, -0.00000185, 0.00000086, -0.00000346, -0.00001382, -0.00001609, -0.00000317, 0.00001394, 0.00001822, 0.00000543, -0.00001268, -0.00002188, -0.00001826, -0.00000787, 0.00000172, 0.00000661, 0.00000579, -0.00000003, -0.00000745, -0.00001110, -0.00000815, -0.00000198, 0.00000088, -0.00000214, -0.00000691, -0.00000790, -0.00000458, -0.00000111, -0.00000021, 0.00000002, 0.00000233, 0.00000500, 0.00000319, -0.00000435, -0.00001195, -0.00001240, -0.00000497, 0.00000342, 0.00000556, 0.00000127, -0.00000355, -0.00000405, -0.00000139, -0.00000020, -0.00000261, -0.00000602, -0.00000645, -0.00000297, 0.00000161, },
		{ -0.00000174, -0.00000327, -0.00000364, -0.00000388, -0.00000447, -0.00000472, -0.00000374, -0.00000206, -0.00000175, -0.00000442, -0.00000888, -0.00001162, -0.00001012, -0.00000578, -0.00000273, -0.00000413, -0.00001026, -0.00001966, -0.00003016, -0.00003729, -0.00003419, -0.00001798, 0.00000130, 0.00000171, -0.00003383, -0.00009842, -0.00016094, -0.00018819, -0.00016667, -0.00010156, -0.00000004, 0.00013147, 0.00026648, 0.00034726, 0.00032073, 0.00020306, 0.00009782, 0.00011820, 0.00026799, 0.00040702, 0.00036382, 0.00010439, -0.00021421, -0.00037582, -0.00030117, -0.00011046, 0.00001014, -0.00000676, -0.00006935, -0.00006110, 0.00001496, 0.00005031, -0.00002065, -0.00012342, -0.00012257, 0.00001071, 0.00014493, 0.00012488, -0.00004876, -0.00020428, -0.00017617, 0.00001570, 0.00018227, 0.00016133, -0.00001771, -0.00017231, -0.00015862, -0.00000896, 0.00011970, 0.00011389, 0.00000521, -0.00008580, -0.00008153, -0.00001067, 0.00004516, 0.00004243, 0.00000487, -0.00002145, -0.00002024, -0.00000861, -0.00000416, -0.00000433, 0.00000155, 0.00001198, 0.00001388, 0.00000076, -0.00001743, -0.00002494, -0.00001676, -0.00000165, 0.00000882, 0.00001028, 0.00000520, -0.00000261, -0.00001001, -0.00001341, -0.00001010, -0.00000223, 0.00000326, 0.00000107, -0.00000638, -0.00001129, -0.00000901, -0.00000246, 0.00000244, 0.00000390, 0.00000465, 0.00000608, 0.00000500, -0.00000182, -0.00001141, -0.00001577, -0.00001025, 0.00000065, 0.00000748, 0.00000548, -0.00000130, -0.00000544, -0.00000421, -0.00000129, -0.00000124, -0.00000412, -0.00000611, -0.00000432, 0.00000007, 0.00000348, 0.00000360, 0.00000109, },
		{ -0.00000305, -0.00000288, -0.00000309, -0.00000337, -0.00000360, -0.00000434, -0.00000618, -0.00000844, -0.00000911, -0.00000684, -0.00000332, -0.00000270, -0.00000791, -0.00001748, -0.00002654, -0.00003064, -0.00002810, -0.00001969, -0.00000895, -0.00000485, -0.00002099, -0.00006529, -0.00012606, -0.00017075, -0.00016589, -0.00010300, -0.00000424, 0.00009856, 0.00018648, 0.00025108, 0.00027527, 0.00023496, 0.00013601, 0.00004478, 0.00005438, 0.00019274, 0.00036212, 0.00039646, 0.00020938, -0.00010610, -0.00034016, -0.00034726, -0.00017200, 0.00000625, 0.00005084, -0.00001736, -0.00007757, -0.00005854, -0.00000596, -0.00000225, -0.00005023, -0.00006896, -0.00000572, 0.00008032, 0.00007731, -0.00003921, -0.00015622, -0.00013748, 0.00001869, 0.00016144, 0.00014144, -0.00002733, -0.00017717, -0.00016028, 0.00000038, 0.00014166, 0.00013216, -0.00000100, -0.00011615, -0.00010938, -0.00000865, 0.00007579, 0.00007202, 0.00000574, -0.00004764, -0.00004674, -0.00001217, 0.00001379, 0.00001514, 0.00000622, 0.00000238, 0.00000139, -0.00000587, -0.00001754, -0.00002148, -0.00001137, 0.00000473, 0.00001361, 0.00001056, 0.00000131, -0.00000678, -0.00001116, -0.00001213, -0.00000925, -0.00000287, 0.00000277, 0.00000197, -0.00000553, -0.00001266, -0.00001197, -0.00000373, 0.00000484, 0.00000804, 0.00000700, 0.00000581, 0.00000467, -0.00000009, -0.00000911, -0.00001630, -0.00001440, -0.00000373, 0.00000682, 0.00000862, 0.00000191, -0.00000540, -0.00000672, -0.00000304, -0.00000019, -0.00000153, -0.00000474, -0.00000546, -0.00000241, 0.00000168, 0.00000342, 0.00000201, -0.00000077, -0.00000279, -0.00000333, },
		{ -0.00000407, -0.00000505, -0.00000583, -0.00000633, -0.00000648, -0.00000593, -0.00000466, -0.00000399, -0.00000640, -0.00001318, -0.00002192, -0.00002730, -0.00002542, -0.00001761, -0.00000991, -0.00000941, -0.00002178, -0.00005031, -0.00009273, -0.00013510, -0.00015145, -0.00011758, -0.00003250, 0.00007396, 0.00016024, 0.00020343, 0.00020928, 0.00019237, 0.00015429, 0.00009317, 0.00003538, 0.00003928, 0.00014214, 0.00029498, 0.00036904, 0.00025854, -0.00000679, -0.00026117, -0.00033664, -0.00021128, -0.00002353, 0.00006407, 0.00001314, -0.00007868, -0.00010096, -0.00004534, 0.00000923, 0.00000568, -0.00002631, -0.00002329, 0.00001828, 0.00003314, -0.00002220, -0.00009678, -0.00009286, 0.00001243, 0.00012003, 0.00010730, -0.00003042, -0.00015954, -0.00014389, 0.00000760, 0.00014532, 0.00013400, -0.00000844, -0.00013524, -0.00012599, -0.00000386, 0.00010223, 0.00009638, 0.00000329, -0.00007580, -0.00007364, -0.00001265, 0.00003804, 0.00003936, 0.00000883, -0.00001549, -0.00001855, -0.00001243, -0.00001028, -0.00000988, -0.00000321, 0.00000797, 0.00001321, 0.00000717, -0.00000386, -0.00001036, -0.00001011, -0.00000761, -0.00000604, -0.00000414, -0.00000116, -0.00000033, -0.00000466, -0.00001098, -0.00001187, -0.00000438, 0.00000588, 0.00001088, 0.00000890, 0.00000497, 0.00000298, 0.00000071, -0.00000565, -0.00001394, -0.00001650, -0.00000902, 0.00000309, 0.00000942, 0.00000542, -0.00000329, -0.00000766, -0.00000484, 0.00000002, 0.00000076, -0.00000298, -0.00000639, -0.00000550, -0.00000115, 0.00000277, 0.00000352, 0.00000160, -0.00000083, -0.00000224, -0.00000261, -0.00000272, -0.00000319, },
		{ -0.00000508, -0.00000452, -0.00000421, -0.00000482, -0.00000701, -0.00001127, -0.00001708, -0.00002208, -0.00002279, -0.00001772, -0.00001045, -0.00000874, -0.00001952, -0.00004387, -0.00007634, -0.00010716, -0.00012337, -0.00010995, -0.00005635, 0.00003115, 0.00012357, 0.00018291, 0.00018960, 0.00015772, 0.00011925, 0.00009228, 0.00007050, 0.00004830, 0.00004855, 0.00010722, 0.00021858, 0.00030481, 0.00026568, 0.00007815, -0.00015608, -0.00028224, -0.00022830, -0.00007106, 0.00003942, 0.00002324, -0.00006481, -0.00011069, -0.00006677, 0.00000966, 0.00003634, 0.00000145, -0.00003593, -0.00002922, -0.00000094, -0.00000317, -0.00003866, -0.00005081, -0.00000072, 0.00006757, 0.00006669, -0.00002601, -0.00012335, -0.00011383, 0.00000995, 0.00012925, 0.00011951, -0.00001428, -0.00013806, -0.00012794, 0.00000175, 0.00011805, 0.00011021, -0.00000172, -0.00009984, -0.00009527, -0.00000963, 0.00006421, 0.00006357, 0.00000790, -0.00003977, -0.00004298, -0.00001661, 0.00000543, 0.00000941, 0.00000555, 0.00000496, 0.00000562, 0.00000096, -0.00000718, -0.00001063, -0.00000671, -0.00000180, -0.00000226, -0.00000628, -0.00000774, -0.00000534, -0.00000384, -0.00000609, -0.00000804, -0.00000406, 0.00000474, 0.00001094, 0.00000977, 0.00000452, 0.00000137, 0.00000089, -0.00000201, -0.00000965, -0.00001634, -0.00001421, -0.00000323, 0.00000730, 0.00000843, 0.00000097, -0.00000611, -0.00000605, -0.00000105, 0.00000159, -0.00000160, -0.00000690, -0.00000828, -0.00000402, 0.00000221, 0.00000574, 0.00000503, 0.00000187, -0.00000129, -0.00000324, -0.00000414, -0.00000456, -0.00000491, -0.00000525, -0.00000537, },
		{ -0.00000715, -0.00001055, -0.00001452, -0.00001772, -0.00001856, -0.00001615, -0.00001168, -0.00000960, -0.00001628, -0.00003551, -0.00006395, -0.00009108, -0.00010425, -0.00009429, -0.00005757, 0.00000358, 0.00007798, 0.00014338, 0.00017362, 0.00015687, 0.00010963, 0.00006715, 0.00005349, 0.00006103, 0.00006561, 0.00006315, 0.00008197, 0.00014636, 0.00022734, 0.00024562, 0.00014278, -0.00004573, -0.00020160, -0.00022386, -0.00012310, -0.00000965, 0.00001621, -0.00004159, -0.00009401, -0.00007241, -0.00000090, 0.00004103, 0.00001426, -0.00003966, -0.00005530, -0.00002186, 0.00001211, 0.00000705, -0.00001774, -0.00001615, 0.00001556, 0.00002766, -0.00001578, -0.00007693, -0.00007705, 0.00000651, 0.00009689, 0.00009244, -0.00001630, -0.00012349, -0.00011532, 0.00000579, 0.00011895, 0.00011060, -0.00000742, -0.00011393, -0.00010691, -0.00000401, 0.00008656, 0.00008252, 0.00000312, -0.00006624, -0.00006722, -0.00001666, 0.00002820, 0.00003370, 0.00001238, -0.00000601, -0.00000939, -0.00000621, -0.00000545, -0.00000531, -0.00000159, 0.00000227, -0.00000055, -0.00000925, -0.00001477, -0.00001112, -0.00000294, 0.00000093, -0.00000121, -0.00000277, 0.00000130, 0.00000758, 0.00000920, 0.00000524, 0.00000129, 0.00000101, 0.00000068, -0.00000509, -0.00001436, -0.00001810, -0.00001065, 0.00000256, 0.00001008, 0.00000660, -0.00000199, -0.00000620, -0.00000352, 0.00000016, -0.00000123, -0.00000660, -0.00000951, -0.00000580, 0.00000209, 0.00000810, 0.00000848, 0.00000420, -0.00000121, -0.00000499, -0.00000640, -0.00000604, -0.00000491, -0.00000390, -0.00000345, -0.00000351, -0.00000395, -0.00000499, },
		{ -0.00001506, -0.00001362, -0.00001117, -0.00001045, -0.00001534, -0.00002909, -0.00005138, -0.00007576, -0.00009071, -0.00008522, -0.00005535, -0.00000631, 0.00005093, 0.00010372, 0.00013888, 0.00014455, 0.00011765, 0.00007267, 0.00003805, 0.00003518, 0.00005758, 0.00007604, 0.00007186, 0.00006325, 0.00008855, 0.00015525, 0.00021128, 0.00018275, 0.00005092, -0.00011247, -0.00019980, -0.00016548, -0.00006692, -0.00000359, -0.00001773, -0.00006232, -0.00006632, -0.00001847, 0.00002548, 0.00001490, -0.00003392, -0.00005940, -0.00003007, 0.00001950, 0.00003371, 0.00000327, -0.00002872, -0.00002573, -0.00000347, -0.00000337, -0.00003047, -0.00004098, -0.00000149, 0.00005554, 0.00005894, -0.00001374, -0.00009472, -0.00009149, 0.00000643, 0.00010417, 0.00009782, -0.00001155, -0.00011422, -0.00010614, 0.00000218, 0.00009952, 0.00009191, -0.00000438, -0.00008922, -0.00008607, -0.00001208, 0.00005418, 0.00005848, 0.00001521, -0.00002431, -0.00002981, -0.00001206, 0.00000338, 0.00000643, 0.00000384, 0.00000145, -0.00000340, -0.00001255, -0.00001912, -0.00001478, -0.00000182, 0.00000783, 0.00000644, -0.00000088, -0.00000351, 0.00000149, 0.00000724, 0.00000729, 0.00000350, 0.00000154, 0.00000159, -0.00000202, -0.00001146, -0.00001961, -0.00001704, -0.00000365, 0.00000963, 0.00001197, 0.00000377, -0.00000510, -0.00000689, -0.00000350, -0.00000198, -0.00000507, -0.00000818, -0.00000563, 0.00000227, 0.00000950, 0.00001056, 0.00000529, -0.00000205, -0.00000710, -0.00000826, -0.00000649, -0.00000370, -0.00000167, -0.00000142, -0.00000292, -0.00000532, -0.00000782, -0.00001024, -0.00001263, -0.00001456, },
		{ -0.00001484, -0.00002523, -0.00004190, -0.00006132, -0.00007595, -0.00007632, -0.00005576, -0.00001544, 0.00003452, 0.00007993, 0.00010972, 0.00011920, 0.00010854, 0.00008164, 0.00004882, 0.00002714, 0.00003058, 0.00005487, 0.00007568, 0.00007107, 0.00004972, 0.00004998, 0.00009944, 0.00017229, 0.00019712, 0.00012104, -0.00002945, -0.00015944, -0.00018719, -0.00011755, -0.00003132, -0.00000191, -0.00002838, -0.00005314, -0.00003671, -0.00000062, 0.00000756, -0.00002155, -0.00004657, -0.00002860, 0.00001778, 0.00004036, 0.00001250, -0.00003453, -0.00004988, -0.00002337, 0.00000694, 0.00000676, -0.00001166, -0.00001114, 0.00001422, 0.00002575, -0.00000769, -0.00005833, -0.00006198, 0.00000311, 0.00007681, 0.00007513, -0.00001241, -0.00010018, -0.00009350, 0.00000661, 0.00009948, 0.00008971, -0.00001232, -0.00010311, -0.00009531, -0.00000396, 0.00007785, 0.00007858, 0.00001318, -0.00004677, -0.00005179, -0.00001462, 0.00001918, 0.00002341, 0.00000782, -0.00000633, -0.00001211, -0.00001526, -0.00001773, -0.00001352, -0.00000056, 0.00001166, 0.00001182, 0.00000106, -0.00000776, -0.00000525, 0.00000433, 0.00000996, 0.00000754, 0.00000263, 0.00000050, -0.00000148, -0.00000848, -0.00001806, -0.00002037, -0.00000979, 0.00000656, 0.00001499, 0.00000945, -0.00000277, -0.00000995, -0.00000823, -0.00000348, -0.00000227, -0.00000407, -0.00000337, 0.00000228, 0.00000880, 0.00001016, 0.00000484, -0.00000309, -0.00000831, -0.00000861, -0.00000545, -0.00000177, 0.00000012, -0.00000079, -0.00000402, -0.00000801, -0.00001101, -0.00001225, -0.00001220, -0.00001164, -0.00001085, -0.00001006, -0.00001057, },
		{ -0.00006186, -0.00006491, -0.00005279, -0.00002358, 0.00001786, 0.00006022, 0.00009058, 0.00010133, 0.00009382, 0.00007551, 0.00005419, 0.00003584, 0.00002664, 0.00003221, 0.00005041, 0.00006635, 0.00006256, 0.00004064, 0.00002990, 0.00006304, 0.00013483, 0.00018892, 0.00015970, 0.00003777, -0.00010793, -0.00018319, -0.00015106, -0.00006186, 0.00000011, -0.00000282, -0.00003728, -0.00004937, -0.00002692, -0.00000346, -0.00000755, -0.00002521, -0.00002161, 0.00000890, 0.00003114, 0.00001270, -0.00003219, -0.00005560, -0.00003189, 0.00001186, 0.00002883, 0.00000750, -0.00001901, -0.00001884, -0.00000156, -0.00000056, -0.00002240, -0.00003287, -0.00000310, 0.00004302, 0.00004768, -0.00000953, -0.00007483, -0.00007225, 0.00000750, 0.00008594, 0.00007670, -0.00001806, -0.00010414, -0.00009291, 0.00000535, 0.00009353, 0.00008962, 0.00000699, -0.00006841, -0.00007053, -0.00001294, 0.00003940, 0.00004253, 0.00000889, -0.00002110, -0.00002610, -0.00001624, -0.00000870, -0.00000564, 0.00000052, 0.00000966, 0.00001192, 0.00000257, -0.00000902, -0.00000975, 0.00000119, 0.00001166, 0.00001165, 0.00000401, -0.00000188, -0.00000324, -0.00000594, -0.00001351, -0.00001950, -0.00001451, 0.00000077, 0.00001380, 0.00001309, 0.00000057, -0.00001103, -0.00001211, -0.00000497, 0.00000128, 0.00000192, 0.00000033, 0.00000157, 0.00000540, 0.00000696, 0.00000326, -0.00000325, -0.00000750, -0.00000698, -0.00000339, -0.00000023, 0.00000020, -0.00000231, -0.00000655, -0.00001058, -0.00001259, -0.00001190, -0.00000958, -0.00000768, -0.00000767, -0.00000997, -0.00001483, -0.00002306, -0.00003521, -0.00004971, },
		{ 0.00000433, 0.00004021, 0.00007096, 0.00008724, 0.00008539, 0.00006992, 0.00005072, 0.00003635, 0.00002959, 0.00002922, 0.00003436, 0.00004413, 0.00005284, 0.00005060, 0.00003515, 0.00002352, 0.00004334, 0.00010224, 0.00016459, 0.00016942, 0.00008474, -0.00005200, -0.00015518, -0.00016264, -0.00009009, -0.00001267, 0.00000822, -0.00002247, -0.00005253, -0.00004539, -0.00001467, 0.00000290, -0.00000417, -0.00001303, -0.00000306, 0.00001340, 0.00000774, -0.00002311, -0.00004633, -0.00003135, 0.00001016, 0.00003441, 0.00001603, -0.00002244, -0.00003842, -0.00001974, 0.00000480, 0.00000582, -0.00000916, -0.00001000, 0.00001016, 0.00002104, -0.00000392, -0.00004397, -0.00004730, 0.00000434, 0.00006181, 0.00005620, -0.00001952, -0.00009153, -0.00007965, 0.00001295, 0.00009710, 0.00008925, -0.00000127, -0.00008375, -0.00008170, -0.00000751, 0.00005955, 0.00005958, 0.00000639, -0.00004045, -0.00004285, -0.00001458, 0.00000782, 0.00000861, 0.00000084, 0.00000028, 0.00000491, 0.00000343, -0.00000510, -0.00000919, -0.00000149, 0.00001038, 0.00001331, 0.00000514, -0.00000395, -0.00000571, -0.00000387, -0.00000699, -0.00001479, -0.00001708, -0.00000725, 0.00000757, 0.00001325, 0.00000465, -0.00000859, -0.00001301, -0.00000565, 0.00000451, 0.00000793, 0.00000420, -0.00000004, -0.00000026, 0.00000161, 0.00000135, -0.00000167, -0.00000423, -0.00000381, -0.00000148, -0.00000031, -0.00000196, -0.00000559, -0.00000928, -0.00001145, -0.00001135, -0.00000918, -0.00000637, -0.00000528, -0.00000770, -0.00001370, -0.00002201, -0.00003147, -0.00004128, -0.00004968, -0.00005293, -0.00004625, -0.00002668, },
		{ 0.00007610, 0.00006787, 0.00005114, 0.00003488, 0.00002624, 0.00002616, 0.00003024, 0.00003416, 0.00003729, 0.00003999, 0.00003930, 0.00003202, 0.00002447, 0.00003474, 0.00007590, 0.00013220, 0.00015772, 0.00011135, 0.00000102, -0.00011032, -0.00015291, -0.00011118, -0.00003644, 0.00000373, -0.00001167, -0.00004590, -0.00005209, -0.00002370, 0.00000638, 0.00000986, -0.00000573, -0.00001351, -0.00000497, 0.00000130, -0.00001077, -0.00002859, -0.00002527, 0.00000298, 0.00002647, 0.00001582, -0.00002051, -0.00004337, -0.00002824, 0.00000627, 0.00002183, 0.00000630, -0.00001504, -0.00001523, -0.00000034, 0.00000227, -0.00001444, -0.00002375, -0.00000182, 0.00003265, 0.00003303, -0.00001602, -0.00006792, -0.00005882, 0.00001644, 0.00008728, 0.00007784, -0.00000881, -0.00008853, -0.00008270, -0.00000012, 0.00007446, 0.00007048, 0.00000091, -0.00006005, -0.00005850, -0.00001002, 0.00002929, 0.00002690, -0.00000008, -0.00001602, -0.00000907, 0.00000375, 0.00000505, -0.00000196, -0.00000329, 0.00000432, 0.00001006, 0.00000551, -0.00000366, -0.00000639, -0.00000192, -0.00000045, -0.00000821, -0.00001764, -0.00001622, -0.00000284, 0.00000960, 0.00000902, -0.00000193, -0.00000951, -0.00000501, 0.00000611, 0.00001174, 0.00000690, -0.00000220, -0.00000669, -0.00000443, -0.00000010, 0.00000171, 0.00000091, -0.00000020, -0.00000094, -0.00000263, -0.00000587, -0.00000921, -0.00001065, -0.00000973, -0.00000752, -0.00000543, -0.00000465, -0.00000641, -0.00001159, -0.00001971, -0.00002870, -0.00003625, -0.00004098, -0.00004200, -0.00003756, -0.00002518, -0.00000384, 0.00002372, 0.00005103, 0.00007033, },
		{ 0.00002580, 0.00002087, 0.00002319, 0.00002829, 0.00003130, 0.00003125, 0.00003076, 0.00003109, 0.00002995, 0.00002712, 0.00003144, 0.00005590, 0.00009936, 0.00013402, 0.00012045, 0.00004462, -0.00005859, -0.00012650, -0.00012134, -0.00006407, -0.00001291, -0.00000692, -0.00003274, -0.00004752, -0.00002923, 0.00000233, 0.00001390, -0.00000120, -0.00001907, -0.00001792, -0.00000402, 0.00000073, -0.00000959, -0.00001721, -0.00000673, 0.00001082, 0.00001009, -0.00001428, -0.00003641, -0.00002868, 0.00000324, 0.00002482, 0.00001309, -0.00001639, -0.00002924, -0.00001401, 0.00000664, 0.00000781, -0.00000557, -0.00000876, 0.00000503, 0.00001208, -0.00000852, -0.00003866, -0.00003524, 0.00001474, 0.00006614, 0.00005833, -0.00001316, -0.00008105, -0.00007341, 0.00000673, 0.00007990, 0.00007251, -0.00000590, -0.00007484, -0.00006904, -0.00000322, 0.00005130, 0.00004540, -0.00000240, -0.00003657, -0.00002771, 0.00000379, 0.00002066, 0.00001163, -0.00000420, -0.00000721, 0.00000062, 0.00000494, 0.00000087, -0.00000286, 0.00000027, 0.00000369, -0.00000280, -0.00001694, -0.00002416, -0.00001488, 0.00000308, 0.00001299, 0.00000819, -0.00000154, -0.00000292, 0.00000515, 0.00001163, 0.00000746, -0.00000405, -0.00001181, -0.00000946, -0.00000082, 0.00000588, 0.00000646, 0.00000259, -0.00000236, -0.00000677, -0.00001012, -0.00001137, -0.00000959, -0.00000579, -0.00000275, -0.00000274, -0.00000597, -0.00001138, -0.00001804, -0.00002522, -0.00003149, -0.00003493, -0.00003431, -0.00002952, -0.00002053, -0.00000665, 0.00001229, 0.00003386, 0.00005278, 0.00006335, 0.00006271, 0.00005245, 0.00003791, },
		{ 0.00002440, 0.00002692, 0.00002659, 0.00002560, 0.00002638, 0.00002804, 0.00002811, 0.00002930, 0.00004153, 0.00007144, 0.00010682, 0.00011646, 0.00007453, -0.00000977, -0.00009011, -0.00011856, -0.00008759, -0.00003574, -0.00000916, -0.00001845, -0.00003550, -0.00003083, -0.00000696, 0.00000882, 0.00000022, -0.00001868, -0.00002263, -0.00000700, 0.00000813, 0.00000493, -0.00000981, -0.00001547, -0.00000601, 0.00000217, -0.00000630, -0.00002276, -0.00002393, -0.00000318, 0.00001748, 0.00001278, -0.00001359, -0.00003140, -0.00001983, 0.00000743, 0.00001934, 0.00000533, -0.00001391, -0.00001534, -0.00000304, 0.00000064, -0.00001023, -0.00001381, 0.00000849, 0.00003853, 0.00003531, -0.00001297, -0.00006279, -0.00005624, 0.00001075, 0.00007393, 0.00006517, -0.00001177, -0.00008051, -0.00007151, 0.00000432, 0.00006870, 0.00005981, -0.00000568, -0.00005703, -0.00004707, 0.00000378, 0.00003901, 0.00002918, -0.00000450, -0.00002328, -0.00001442, 0.00000362, 0.00001048, 0.00000622, 0.00000278, 0.00000314, -0.00000166, -0.00001593, -0.00002864, -0.00002477, -0.00000438, 0.00001550, 0.00001938, 0.00000939, 0.00000045, 0.00000169, 0.00000718, 0.00000562, -0.00000467, -0.00001378, -0.00001213, -0.00000114, 0.00000910, 0.00001057, 0.00000379, -0.00000521, -0.00001132, -0.00001303, -0.00001088, -0.00000611, -0.00000117, 0.00000056, -0.00000298, -0.00001052, -0.00001852, -0.00002428, -0.00002739, -0.00002849, -0.00002747, -0.00002355, -0.00001639, -0.00000635, 0.00000629, 0.00002127, 0.00003690, 0.00004933, 0.00005435, 0.00005049, 0.00004026, 0.00002865, 0.00002034, 0.00001772, 0.00002014, },
		{ 0.00002309, 0.00002414, 0.00002601, 0.00002669, 0.00002641, 0.00003158, 0.00005076, 0.00008190, 0.00010423, 0.00008959, 0.00002904, -0.00005074, -0.00010324, -0.00010068, -0.00005821, -0.00001799, -0.00000850, -0.00002144, -0.00002874, -0.00001745, -0.00000176, -0.00000104, -0.00001367, -0.00001969, -0.00000754, 0.00001013, 0.00001210, -0.00000448, -0.00002045, -0.00001878, -0.00000531, 0.00000046, -0.00000828, -0.00001703, -0.00001063, 0.00000466, 0.00000804, -0.00000777, -0.00002439, -0.00001942, 0.00000426, 0.00001997, 0.00000919, -0.00001534, -0.00002518, -0.00001077, 0.00000862, 0.00001155, 0.00000176, -0.00000020, 0.00001071, 0.00001364, -0.00000857, -0.00003801, -0.00003528, 0.00001062, 0.00005765, 0.00005034, -0.00001465, -0.00007495, -0.00006497, 0.00001069, 0.00007713, 0.00006674, -0.00000898, -0.00007249, -0.00006273, 0.00000364, 0.00005600, 0.00004674, -0.00000450, -0.00004120, -0.00003252, 0.00000208, 0.00002444, 0.00002032, 0.00000523, -0.00000336, -0.00000670, -0.00001520, -0.00002744, -0.00002868, -0.00001057, 0.00001525, 0.00002810, 0.00002055, 0.00000483, -0.00000295, -0.00000044, 0.00000185, -0.00000363, -0.00001175, -0.00001192, -0.00000179, 0.00000959, 0.00001179, 0.00000341, -0.00000801, -0.00001438, -0.00001337, -0.00000773, -0.00000149, 0.00000202, 0.00000041, -0.00000680, -0.00001686, -0.00002482, -0.00002719, -0.00002455, -0.00002016, -0.00001622, -0.00001220, -0.00000631, 0.00000230, 0.00001310, 0.00002501, 0.00003632, 0.00004426, 0.00004584, 0.00004034, 0.00003058, 0.00002137, 0.00001617, 0.00001548, 0.00001765, 0.00002053, 0.00002248, 0.00002302, },
		{ 0.00002392, 0.00002269, 0.00002460, 0.00003685, 0.00006191, 0.00008819, 0.00009173, 0.00005455, -0.00001430, -0.00007829, -0.00010044, -0.00007495, -0.00003150, -0.00000650, -0.00001031, -0.00002384, -0.00002503, -0.00001313, -0.00000415, -0.00000700, -0.00001214, -0.00000629, 0.00000751, 0.00001208, -0.00000152, -0.00002059, -0.00002481, -0.00001084, 0.00000376, 0.00000243, -0.00001038, -0.00001648, -0.00000832, 0.00000167, -0.00000116, -0.00001281, -0.00001536, -0.00000175, 0.00001246, 0.00000790, -0.00001245, -0.00002441, -0.00001172, 0.00001306, 0.00002316, 0.00000986, -0.00000847, -0.00001152, -0.00000285, -0.00000171, -0.00001250, -0.00001515, 0.00000654, 0.00003485, 0.00003165, -0.00001339, -0.00005893, -0.00005076, 0.00001419, 0.00007434, 0.00006464, -0.00001107, -0.00007892, -0.00007104, 0.00000298, 0.00006720, 0.00005999, -0.00000436, -0.00005707, -0.00004970, 0.00000087, 0.00004025, 0.00003690, 0.00000692, -0.00001540, -0.00001783, -0.00001470, -0.00001920, -0.00002404, -0.00001381, 0.00001091, 0.00003066, 0.00002883, 0.00001009, -0.00000634, -0.00000871, -0.00000299, -0.00000144, -0.00000641, -0.00000925, -0.00000339, 0.00000639, 0.00000957, 0.00000225, -0.00000891, -0.00001435, -0.00001096, -0.00000329, 0.00000220, 0.00000201, -0.00000383, -0.00001321, -0.00002244, -0.00002711, -0.00002465, -0.00001696, -0.00000914, -0.00000497, -0.00000363, -0.00000115, 0.00000535, 0.00001543, 0.00002633, 0.00003516, 0.00003965, 0.00003830, 0.00003144, 0.00002211, 0.00001485, 0.00001243, 0.00001400, 0.00001659, 0.00001816, 0.00001884, 0.00001969, 0.00002114, 0.00002287, 0.00002411, },
		{ 0.00002752, 0.00004696, 0.00007185, 0.00008502, 0.00006718, 0.00001508, -0.00004851, -0.00008782, -0.00008276, -0.00004653, -0.00001300, -0.00000532, -0.00001760, -0.00002716, -0.00002132, -0.00000816, -0.00000184, -0.00000392, -0.00000414, 0.00000256, 0.00000733, -0.00000049, -0.00001663, -0.00002417, -0.00001395, 0.00000286, 0.00000709, -0.00000537, -0.00001883, -0.00001739, -0.00000422, 0.00000382, -0.00000155, -0.00001015, -0.00000842, 0.00000133, 0.00000407, -0.00000619, -0.00001578, -0.00000810, 0.00001287, 0.00002415, 0.00001102, -0.00001380, -0.00002437, -0.00001226, 0.00000504, 0.00000814, 0.00000040, 0.00000008, 0.00001097, 0.00001338, -0.00000820, -0.00003597, -0.00003204, 0.00001395, 0.00006095, 0.00005422, -0.00001098, -0.00007446, -0.00007021, 0.00000129, 0.00006925, 0.00006551, -0.00000393, -0.00006682, -0.00006179, 0.00000026, 0.00005419, 0.00005200, 0.00000710, -0.00003073, -0.00003262, -0.00001372, -0.00000418, -0.00001055, -0.00001354, 0.00000170, 0.00002448, 0.00003180, 0.00001636, -0.00000571, -0.00001461, -0.00000805, 0.00000041, 0.00000018, -0.00000518, -0.00000592, -0.00000010, 0.00000452, 0.00000133, -0.00000661, -0.00001062, -0.00000677, 0.00000025, 0.00000291, -0.00000177, -0.00001096, -0.00001978, -0.00002458, -0.00002347, -0.00001655, -0.00000676, 0.00000078, 0.00000258, 0.00000027, -0.00000045, 0.00000500, 0.00001595, 0.00002753, 0.00003483, 0.00003576, 0.00003093, 0.00002254, 0.00001400, 0.00000906, 0.00000957, 0.00001366, 0.00001718, 0.00001757, 0.00001611, 0.00001591, 0.00001824, 0.00002139, 0.00002279, 0.00002148, 0.00001901, 0.00001923, },
		{ 0.00007418, 0.00006986, 0.00003538, -0.00001938, -0.00006679, -0.00008087, -0.00005925, -0.00002546, -0.00000726, -0.00001191, -0.00002367, -0.00002421, -0.00001206, -0.00000028, 0.00000167, -0.00000192, -0.00000205, 0.00000118, -0.00000072, -0.00001067, -0.00001904, -0.00001489, -0.00000134, 0.00000595, -0.00000243, -0.00001716, -0.00002067, -0.00000844, 0.00000546, 0.00000577, -0.00000563, -0.00001336, -0.00000879, 0.00000009, 0.00000131, -0.00000375, -0.00000270, 0.00000861, 0.00001693, 0.00000743, -0.00001481, -0.00002699, -0.00001484, 0.00000925, 0.00002011, 0.00000935, -0.00000637, -0.00000858, -0.00000075, -0.00000055, -0.00001123, -0.00001286, 0.00001024, 0.00004022, 0.00003818, -0.00000838, -0.00005999, -0.00006076, -0.00000180, 0.00006093, 0.00006181, -0.00000285, -0.00006752, -0.00006567, 0.00000006, 0.00006231, 0.00006147, 0.00000547, -0.00004562, -0.00004697, -0.00001128, 0.00001553, 0.00000939, -0.00001048, -0.00001212, 0.00000920, 0.00002854, 0.00002377, 0.00000096, -0.00001572, -0.00001287, -0.00000010, 0.00000533, -0.00000084, -0.00000842, -0.00000802, -0.00000196, 0.00000130, -0.00000107, -0.00000391, -0.00000245, 0.00000083, -0.00000042, -0.00000837, -0.00001829, -0.00002357, -0.00002156, -0.00001414, -0.00000490, 0.00000283, 0.00000625, 0.00000429, -0.00000037, -0.00000184, 0.00000442, 0.00001706, 0.00002940, 0.00003482, 0.00003157, 0.00002276, 0.00001312, 0.00000643, 0.00000484, 0.00000849, 0.00001466, 0.00001877, 0.00001791, 0.00001388, 0.00001162, 0.00001409, 0.00001922, 0.00002220, 0.00002048, 0.00001623, 0.00001455, 0.00002035, 0.00003580, 0.00005750, },
		{ 0.00000439, -0.00004272, -0.00007060, -0.00006617, -0.00003941, -0.00001473, -0.00000871, -0.00001675, -0.00002189, -0.00001487, -0.00000256, 0.00000302, -0.00000031, -0.00000446, -0.00000353, -0.00000151, -0.00000512, -0.00001243, -0.00001421, -0.00000681, 0.00000113, -0.00000131, -0.00001246, -0.00001861, -0.00001071, 0.00000341, 0.00000767, -0.00000262, -0.00001489, -0.00001459, -0.00000252, 0.00000759, 0.00000709, 0.00000200, 0.00000211, 0.00000569, 0.00000190, -0.00001171, -0.00002134, -0.00001250, 0.00000955, 0.00002233, 0.00001171, -0.00001050, -0.00001997, -0.00000862, 0.00000731, 0.00000994, 0.00000305, 0.00000443, 0.00001714, 0.00002027, -0.00000383, -0.00003893, -0.00004524, -0.00000613, 0.00004374, 0.00004977, -0.00000088, -0.00005841, -0.00006030, -0.00000028, 0.00006170, 0.00006234, 0.00000237, -0.00005584, -0.00005630, -0.00000666, 0.00003623, 0.00003124, -0.00000627, -0.00002893, -0.00001286, 0.00001993, 0.00003219, 0.00001399, -0.00001114, -0.00001741, -0.00000472, 0.00000673, 0.00000305, -0.00000921, -0.00001471, -0.00000822, 0.00000192, 0.00000619, 0.00000397, 0.00000041, -0.00000244, -0.00000735, -0.00001558, -0.00002264, -0.00002227, -0.00001345, -0.00000158, 0.00000674, 0.00000873, 0.00000563, 0.00000036, -0.00000358, -0.00000234, 0.00000630, 0.00001983, 0.00003112, 0.00003338, 0.00002574, 0.00001360, 0.00000389, 0.00000044, 0.00000312, 0.00000962, 0.00001656, 0.00002008, 0.00001786, 0.00001176, 0.00000746, 0.00000951, 0.00001649, 0.00002195, 0.00002087, 0.00001475, 0.00001029, 0.00001379, 0.00002693, 0.00004611, 0.00006310, 0.00006643, 0.00004644, },
		{ -0.00006516, -0.00005008, -0.00002494, -0.00000940, -0.00000985, -0.00001624, -0.00001577, -0.00000713, 0.00000043, 0.00000019, -0.00000424, -0.00000533, -0.00000234, -0.00000172, -0.00000693, -0.00001255, -0.00001129, -0.00000456, -0.00000147, -0.00000670, -0.00001334, -0.00001133, -0.00000130, 0.00000487, -0.00000115, -0.00001262, -0.00001488, -0.00000329, 0.00001077, 0.00001346, 0.00000453, -0.00000423, -0.00000509, -0.00000306, -0.00000607, -0.00001127, -0.00000789, 0.00000601, 0.00001652, 0.00000921, -0.00001090, -0.00002190, -0.00000999, 0.00001316, 0.00002357, 0.00001328, -0.00000154, -0.00000312, 0.00000421, 0.00000135, -0.00001625, -0.00002734, -0.00001094, 0.00002146, 0.00003225, 0.00000187, -0.00004132, -0.00004710, -0.00000137, 0.00005159, 0.00005385, -0.00000122, -0.00005797, -0.00005709, 0.00000010, 0.00005351, 0.00004958, -0.00000276, -0.00004581, -0.00003712, 0.00000835, 0.00004081, 0.00003145, -0.00000191, -0.00002187, -0.00001397, 0.00000343, 0.00000649, -0.00000652, -0.00001766, -0.00001316, 0.00000194, 0.00001257, 0.00001084, 0.00000113, -0.00000877, -0.00001594, -0.00002080, -0.00002164, -0.00001515, -0.00000212, 0.00001038, 0.00001459, 0.00000917, -0.00000030, -0.00000676, -0.00000656, 0.00000026, 0.00001164, 0.00002371, 0.00003090, 0.00002866, 0.00001763, 0.00000435, -0.00000348, -0.00000271, 0.00000417, 0.00001238, 0.00001829, 0.00001988, 0.00001643, 0.00000969, 0.00000451, 0.00000585, 0.00001357, 0.00002120, 0.00002160, 0.00001438, 0.00000689, 0.00000767, 0.00001935, 0.00003743, 0.00005384, 0.00006023, 0.00004966, 0.00002009, -0.00002057, -0.00005462, },
		{ -0.00001431, -0.00000629, -0.00001001, -0.00001437, -0.00001149, -0.00000435, -0.00000068, -0.00000241, -0.00000439, -0.00000283, -0.00000094, -0.00000390, -0.00001041, -0.00001323, -0.00000873, -0.00000235, -0.00000197, -0.00000752, -0.00001083, -0.00000674, -0.00000045, -0.00000075, -0.00000754, -0.00001061, -0.00000261, 0.00001019, 0.00001417, 0.00000461, -0.00000837, -0.00001208, -0.00000594, -0.00000016, -0.00000110, -0.00000348, 0.00000008, 0.00000631, 0.00000442, -0.00000751, -0.00001579, -0.00000635, 0.00001560, 0.00002809, 0.00001712, -0.00000583, -0.00001667, -0.00000733, 0.00000554, 0.00000306, -0.00001082, -0.00001496, -0.00000082, 0.00001323, 0.00000473, -0.00002029, -0.00002944, -0.00000353, 0.00003379, 0.00003784, -0.00000398, -0.00005054, -0.00004815, 0.00000808, 0.00006340, 0.00005979, -0.00000125, -0.00005915, -0.00005794, -0.00000320, 0.00004807, 0.00004943, 0.00000932, -0.00002629, -0.00002661, -0.00000365, 0.00001002, 0.00000064, -0.00001561, -0.00001658, -0.00000040, 0.00001530, 0.00001501, 0.00000032, -0.00001574, -0.00002343, -0.00002207, -0.00001479, -0.00000351, 0.00000934, 0.00001786, 0.00001599, 0.00000429, -0.00000871, -0.00001339, -0.00000679, 0.00000645, 0.00001931, 0.00002693, 0.00002717, 0.00001990, 0.00000777, -0.00000330, -0.00000718, -0.00000226, 0.00000727, 0.00001532, 0.00001863, 0.00001755, 0.00001356, 0.00000804, 0.00000364, 0.00000414, 0.00001079, 0.00001901, 0.00002124, 0.00001460, 0.00000512, 0.00000290, 0.00001278, 0.00003038, 0.00004643, 0.00005337, 0.00004742, 0.00002754, -0.00000352, -0.00003622, -0.00005618, -0.00005406, -0.00003470, },
		{ -0.00001096, -0.00001334, -0.00000904, -0.00000294, -0.00000076, -0.00000206, -0.00000272, -0.00000196, -0.00000330, -0.00000823, -0.00001230, -0.00001019, -0.00000344, 0.00000033, -0.00000331, -0.00000971, -0.00001089, -0.00000568, -0.00000086, -0.00000156, -0.00000411, -0.00000128, 0.00000654, 0.00001012, 0.00000287, -0.00000935, -0.00001394, -0.00000646, 0.00000393, 0.00000583, -0.00000080, -0.00000606, -0.00000406, -0.00000015, -0.00000153, -0.00000499, -0.00000014, 0.00001440, 0.00002447, 0.00001557, -0.00000733, -0.00002187, -0.00001325, 0.00000742, 0.00001539, 0.00000156, -0.00001688, -0.00001830, -0.00000349, 0.00000663, -0.00000043, -0.00001162, -0.00000646, 0.00001237, 0.00001824, -0.00000480, -0.00003471, -0.00003105, 0.00001566, 0.00006365, 0.00005948, -0.00000197, -0.00006563, -0.00007030, -0.00001209, 0.00005188, 0.00006306, 0.00001916, -0.00003021, -0.00003966, -0.00001196, 0.00001429, 0.00001177, -0.00000912, -0.00001922, -0.00000659, 0.00001263, 0.00001601, -0.00000023, -0.00002019, -0.00002729, -0.00001896, -0.00000374, 0.00000978, 0.00001779, 0.00001844, 0.00001034, -0.00000396, -0.00001582, -0.00001599, -0.00000305, 0.00001474, 0.00002659, 0.00002740, 0.00001943, 0.00000801, -0.00000218, -0.00000768, -0.00000624, 0.00000147, 0.00001095, 0.00001670, 0.00001673, 0.00001345, 0.00001007, 0.00000739, 0.00000509, 0.00000458, 0.00000821, 0.00001489, 0.00001879, 0.00001486, 0.00000567, 0.00000070, 0.00000738, 0.00002357, 0.00003951, 0.00004657, 0.00004247, 0.00002887, 0.00000752, -0.00001853, -0.00004135, -0.00005035, -0.00004132, -0.00002236, -0.00000792, -0.00000575, },
		{ -0.00000615, -0.00000094, -0.00000011, -0.00000197, -0.00000322, -0.00000398, -0.00000634, -0.00000936, -0.00000910, -0.00000438, -0.00000000, -0.00000168, -0.00000833, -0.00001245, -0.00000886, -0.00000106, 0.00000331, 0.00000196, -0.00000009, 0.00000147, 0.00000361, 0.00000035, -0.00000751, -0.00001135, -0.00000527, 0.00000508, 0.00000800, -0.00000020, -0.00001030, -0.00001115, -0.00000279, 0.00000505, 0.00000641, 0.00000609, 0.00001050, 0.00001565, 0.00001063, -0.00000602, -0.00001950, -0.00001402, 0.00000648, 0.00001942, 0.00000857, -0.00001586, -0.00002781, -0.00001522, 0.00000645, 0.00001362, 0.00000246, -0.00000925, -0.00000751, -0.00000015, -0.00000322, -0.00001401, -0.00000977, 0.00002112, 0.00005446, 0.00004927, -0.00000410, -0.00006326, -0.00007135, -0.00001692, 0.00005023, 0.00006805, 0.00002466, -0.00003259, -0.00004917, -0.00001822, 0.00001954, 0.00002471, -0.00000053, -0.00002227, -0.00001696, 0.00000464, 0.00001471, 0.00000164, -0.00001942, -0.00002620, -0.00001278, 0.00000816, 0.00002119, 0.00002115, 0.00001202, -0.00000071, -0.00001264, -0.00001834, -0.00001250, 0.00000417, 0.00002231, 0.00003024, 0.00002370, 0.00000876, -0.00000451, -0.00001005, -0.00000791, -0.00000110, 0.00000712, 0.00001349, 0.00001531, 0.00001259, 0.00000871, 0.00000725, 0.00000809, 0.00000820, 0.00000642, 0.00000573, 0.00000911, 0.00001422, 0.00001482, 0.00000859, 0.00000172, 0.00000343, 0.00001591, 0.00003138, 0.00003968, 0.00003725, 0.00002742, 0.00001362, -0.00000402, -0.00002436, -0.00004041, -0.00004316, -0.00003123, -0.00001440, -0.00000506, -0.00000659, -0.00001163, -0.00001168, },
		{ -0.00000077, -0.00000328, -0.00000449, -0.00000490, -0.00000595, -0.00000659, -0.00000487, -0.00000207, -0.00000233, -0.00000698, -0.00001127, -0.00000920, -0.00000116, 0.00000572, 0.00000561, 0.00000048, -0.00000314, -0.00000273, -0.00000197, -0.00000416, -0.00000637, -0.00000345, 0.00000337, 0.00000586, -0.00000109, -0.00001137, -0.00001320, -0.00000293, 0.00001023, 0.00001489, 0.00001041, 0.00000552, 0.00000534, 0.00000463, -0.00000284, -0.00001224, -0.00001117, 0.00000322, 0.00001624, 0.00001052, -0.00001192, -0.00002862, -0.00002101, 0.00000351, 0.00001908, 0.00001044, -0.00001070, -0.00002110, -0.00001312, 0.00000030, 0.00000659, 0.00001054, 0.00002323, 0.00003850, 0.00003252, -0.00000613, -0.00005215, -0.00006129, -0.00001779, 0.00004201, 0.00006212, 0.00002432, -0.00003211, -0.00005155, -0.00001961, 0.00002522, 0.00003607, 0.00000686, -0.00002666, -0.00003016, -0.00000659, 0.00001298, 0.00000753, -0.00001236, -0.00002058, -0.00000613, 0.00001706, 0.00002784, 0.00001920, 0.00000087, -0.00001387, -0.00001895, -0.00001475, -0.00000304, 0.00001300, 0.00002620, 0.00002788, 0.00001568, -0.00000250, -0.00001449, -0.00001399, -0.00000441, 0.00000631, 0.00001266, 0.00001372, 0.00001106, 0.00000706, 0.00000471, 0.00000616, 0.00001000, 0.00001169, 0.00000840, 0.00000331, 0.00000278, 0.00000850, 0.00001435, 0.00001304, 0.00000548, 0.00000099, 0.00000719, 0.00002114, 0.00003229, 0.00003342, 0.00002643, 0.00001688, 0.00000589, -0.00000930, -0.00002747, -0.00003989, -0.00003795, -0.00002340, -0.00000818, -0.00000291, -0.00000702, -0.00001139, -0.00000950, -0.00000358, 0.00000011, },
		{ -0.00000391, -0.00000354, -0.00000411, -0.00000477, -0.00000436, -0.00000409, -0.00000583, -0.00000826, -0.00000720, -0.00000113, 0.00000542, 0.00000613, 0.00000032, -0.00000595, -0.00000694, -0.00000353, -0.00000098, -0.00000148, -0.00000199, -0.00000005, 0.00000132, -0.00000238, -0.00000915, -0.00001040, -0.00000100, 0.00001271, 0.00001851, 0.00001191, 0.00000114, -0.00000317, -0.00000052, 0.00000076, -0.00000348, -0.00000663, -0.00000106, 0.00000868, 0.00000876, -0.00000609, -0.00002247, -0.00002153, -0.00000199, 0.00001581, 0.00001194, -0.00000978, -0.00002554, -0.00001826, 0.00000456, 0.00002219, 0.00002517, 0.00002171, 0.00002017, 0.00001424, -0.00000660, -0.00003466, -0.00004329, -0.00001603, 0.00002765, 0.00004590, 0.00001853, -0.00002772, -0.00004483, -0.00001487, 0.00003002, 0.00004231, 0.00001000, -0.00003243, -0.00004331, -0.00001753, 0.00001284, 0.00001753, -0.00000016, -0.00001247, -0.00000191, 0.00002000, 0.00002845, 0.00001354, -0.00001121, -0.00002543, -0.00002127, -0.00000584, 0.00001018, 0.00002100, 0.00002451, 0.00001886, 0.00000457, -0.00001154, -0.00001913, -0.00001312, 0.00000123, 0.00001320, 0.00001610, 0.00001145, 0.00000509, 0.00000162, 0.00000253, 0.00000707, 0.00001243, 0.00001413, 0.00000934, 0.00000113, -0.00000254, 0.00000317, 0.00001340, 0.00001744, 0.00001043, -0.00000014, -0.00000184, 0.00000923, 0.00002426, 0.00003145, 0.00002782, 0.00001935, 0.00001100, 0.00000066, -0.00001521, -0.00003239, -0.00003990, -0.00003153, -0.00001412, -0.00000168, -0.00000160, -0.00000847, -0.00001187, -0.00000799, -0.00000185, 0.00000041, -0.00000152, -0.00000370, },
		{ -0.00000415, -0.00000552, -0.00000553, -0.00000498, -0.00000490, -0.00000415, -0.00000098, 0.00000319, 0.00000405, -0.00000046, -0.00000660, -0.00000830, -0.00000423, 0.00000073, 0.00000148, -0.00000142, -0.00000358, -0.00000323, -0.00000302, -0.00000461, -0.00000446, 0.00000195, 0.00001223, 0.00001712, 0.00001068, -0.00000159, -0.00000791, -0.00000370, 0.00000343, 0.00000377, -0.00000210, -0.00000479, 0.00000019, 0.00000510, -0.00000007, -0.00001274, -0.00001851, -0.00000887, 0.00000620, 0.00000845, -0.00000620, -0.00002069, -0.00001559, 0.00000808, 0.00002948, 0.00003128, 0.00001735, 0.00000422, -0.00000058, -0.00000468, -0.00001480, -0.00002244, -0.00001351, 0.00000928, 0.00002291, 0.00000933, -0.00001929, -0.00002962, -0.00000483, 0.00003233, 0.00004103, 0.00000730, -0.00003848, -0.00005277, -0.00002449, 0.00001552, 0.00002997, 0.00001405, -0.00000480, -0.00000217, 0.00001602, 0.00002371, 0.00000697, -0.00002024, -0.00003246, -0.00001961, 0.00000565, 0.00002372, 0.00002608, 0.00001718, 0.00000461, -0.00000748, -0.00001618, -0.00001708, -0.00000782, 0.00000680, 0.00001688, 0.00001613, 0.00000748, -0.00000061, -0.00000220, 0.00000242, 0.00000927, 0.00001433, 0.00001456, 0.00000872, -0.00000036, -0.00000542, -0.00000038, 0.00001189, 0.00001990, 0.00001445, -0.00000035, -0.00000959, -0.00000254, 0.00001597, 0.00003066, 0.00003158, 0.00002229, 0.00001222, 0.00000426, -0.00000653, -0.00002243, -0.00003574, -0.00003558, -0.00002103, -0.00000411, 0.00000192, -0.00000424, -0.00001223, -0.00001239, -0.00000540, 0.00000093, 0.00000158, -0.00000129, -0.00000309, -0.00000281, -0.00000276, },
		{ -0.00000437, -0.00000251, -0.00000149, -0.00000078, 0.00000038, 0.00000071, -0.00000166, -0.00000569, -0.00000734, -0.00000428, 0.00000054, 0.00000183, -0.00000171, -0.00000579, -0.00000587, -0.00000242, 0.00000072, 0.00000224, 0.00000475, 0.00000944, 0.00001209, 0.00000750, -0.00000242, -0.00000858, -0.00000467, 0.00000446, 0.00000778, 0.00000132, -0.00000673, -0.00000637, 0.00000108, 0.00000434, -0.00000318, -0.00001396, -0.00001529, -0.00000573, 0.00000254, -0.00000060, -0.00000909, -0.00000729, 0.00000970, 0.00002767, 0.00002859, 0.00001167, -0.00000570, -0.00000852, -0.00000057, 0.00000292, -0.00000410, -0.00001180, -0.00000955, -0.00000147, -0.00000039, -0.00000794, -0.00000916, 0.00000782, 0.00003089, 0.00003192, -0.00000070, -0.00004285, -0.00005537, -0.00002487, 0.00002088, 0.00004168, 0.00002614, -0.00000027, -0.00000727, 0.00000650, 0.00001595, 0.00000237, -0.00002344, -0.00003375, -0.00001548, 0.00001601, 0.00003410, 0.00002724, 0.00000603, -0.00001186, -0.00001818, -0.00001561, -0.00000883, 0.00000046, 0.00001030, 0.00001566, 0.00001245, 0.00000323, -0.00000398, -0.00000331, 0.00000382, 0.00001137, 0.00001470, 0.00001285, 0.00000692, -0.00000064, -0.00000513, -0.00000149, 0.00000961, 0.00001892, 0.00001570, -0.00000000, -0.00001426, -0.00001174, 0.00000814, 0.00002956, 0.00003595, 0.00002580, 0.00001125, 0.00000247, -0.00000269, -0.00001218, -0.00002600, -0.00003370, -0.00002679, -0.00001026, 0.00000139, -0.00000047, -0.00001004, -0.00001472, -0.00000923, 0.00000022, 0.00000465, 0.00000220, -0.00000200, -0.00000346, -0.00000292, -0.00000330, -0.00000493, -0.00000565, },
		{ -0.00000062, -0.00000179, -0.00000245, -0.00000308, -0.00000442, -0.00000535, -0.00000401, -0.00000106, 0.00000010, -0.00000247, -0.00000601, -0.00000572, -0.00000053, 0.00000536, 0.00000759, 0.00000653, 0.00000557, 0.00000552, 0.00000346, -0.00000192, -0.00000621, -0.00000381, 0.00000394, 0.00000832, 0.00000322, -0.00000644, -0.00000950, -0.00000226, 0.00000591, 0.00000331, -0.00000948, -0.00001932, -0.00001576, -0.00000314, 0.00000554, 0.00000498, 0.00000318, 0.00000904, 0.00001870, 0.00001929, 0.00000633, -0.00000816, -0.00000866, 0.00000464, 0.00001485, 0.00000777, -0.00001155, -0.00002491, -0.00002160, -0.00000796, 0.00000400, 0.00001157, 0.00001942, 0.00002542, 0.00001714, -0.00001135, -0.00004338, -0.00004952, -0.00001822, 0.00002730, 0.00004908, 0.00003238, -0.00000037, -0.00001568, -0.00000531, 0.00000822, 0.00000164, -0.00002014, -0.00002997, -0.00001117, 0.00002196, 0.00003890, 0.00002483, -0.00000588, -0.00002692, -0.00002576, -0.00001063, 0.00000342, 0.00000970, 0.00001065, 0.00000940, 0.00000578, 0.00000011, -0.00000365, -0.00000142, 0.00000562, 0.00001182, 0.00001295, 0.00000969, 0.00000499, 0.00000068, -0.00000199, -0.00000052, 0.00000636, 0.00001402, 0.00001335, 0.00000069, -0.00001454, -0.00001626, 0.00000161, 0.00002654, 0.00003831, 0.00002906, 0.00001009, -0.00000193, -0.00000309, -0.00000376, -0.00001305, -0.00002605, -0.00002958, -0.00001880, -0.00000391, 0.00000120, -0.00000562, -0.00001348, -0.00001182, -0.00000193, 0.00000612, 0.00000567, -0.00000023, -0.00000440, -0.00000419, -0.00000268, -0.00000295, -0.00000416, -0.00000377, -0.00000170, -0.00000022, },
		{ -0.00000455, -0.00000390, -0.00000373, -0.00000371, -0.00000302, -0.00000238, -0.00000311, -0.00000439, -0.00000309, 0.00000221, 0.00000838, 0.00001036, 0.00000685, 0.00000179, -0.00000056, -0.00000045, -0.00000096, -0.00000251, -0.00000184, 0.00000259, 0.00000638, 0.00000369, -0.00000443, -0.00000926, -0.00000453, 0.00000448, 0.00000554, -0.00000592, -0.00001964, -0.00002086, -0.00000672, 0.00001040, 0.00001695, 0.00001232, 0.00000664, 0.00000638, 0.00000715, 0.00000252, -0.00000419, -0.00000268, 0.00000939, 0.00001929, 0.00001170, -0.00001234, -0.00003358, -0.00003330, -0.00001183, 0.00001363, 0.00002728, 0.00002662, 0.00001694, 0.00000075, -0.00002085, -0.00003862, -0.00003600, -0.00000647, 0.00003225, 0.00004934, 0.00003074, -0.00000486, -0.00002453, -0.00001549, 0.00000317, 0.00000502, -0.00001188, -0.00002325, -0.00000881, 0.00002169, 0.00003740, 0.00002021, -0.00001533, -0.00003745, -0.00002945, -0.00000333, 0.00001652, 0.00001798, 0.00000800, -0.00000051, -0.00000249, -0.00000108, 0.00000057, 0.00000297, 0.00000673, 0.00000959, 0.00000911, 0.00000629, 0.00000419, 0.00000359, 0.00000272, 0.00000132, 0.00000209, 0.00000596, 0.00000790, 0.00000168, -0.00001021, -0.00001519, -0.00000303, 0.00002050, 0.00003632, 0.00003083, 0.00001043, -0.00000547, -0.00000577, 0.00000127, -0.00000043, -0.00001466, -0.00002831, -0.00002724, -0.00001288, -0.00000021, -0.00000025, -0.00000849, -0.00001198, -0.00000510, 0.00000493, 0.00000789, 0.00000234, -0.00000433, -0.00000568, -0.00000274, -0.00000066, -0.00000151, -0.00000281, -0.00000213, -0.00000059, -0.00000072, -0.00000274, -0.00000447, },
		{ -0.00000351, -0.00000424, -0.00000416, -0.00000317, -0.00000170, 0.00000084, 0.00000515, 0.00000958, 0.00001038, 0.00000576, -0.00000112, -0.00000484, -0.00000350, -0.00000032, 0.00000089, 0.00000048, 0.00000133, 0.00000347, 0.00000309, -0.00000180, -0.00000683, -0.00000569, 0.00000093, 0.00000405, -0.00000339, -0.00001610, -0.00001999, -0.00000754, 0.00001251, 0.00002341, 0.00001764, 0.00000368, -0.00000491, -0.00000388, 0.00000061, 0.00000318, 0.00000602, 0.00001236, 0.00001689, 0.00000895, -0.00001301, -0.00003410, -0.00003504, -0.00001196, 0.00001835, 0.00003409, 0.00002748, 0.00000754, -0.00001251, -0.00002549, -0.00002847, -0.00001779, 0.00000671, 0.00003321, 0.00004146, 0.00002155, -0.00001190, -0.00003060, -0.00002077, 0.00000216, 0.00001110, -0.00000173, -0.00001627, -0.00000951, 0.00001543, 0.00003066, 0.00001513, -0.00002000, -0.00004167, -0.00002952, 0.00000375, 0.00002733, 0.00002394, 0.00000356, -0.00001168, -0.00001080, -0.00000038, 0.00000762, 0.00000878, 0.00000650, 0.00000460, 0.00000376, 0.00000381, 0.00000537, 0.00000761, 0.00000724, 0.00000250, -0.00000294, -0.00000338, 0.00000100, 0.00000299, -0.00000242, -0.00000928, -0.00000563, 0.00001146, 0.00002896, 0.00003004, 0.00001310, -0.00000527, -0.00000824, 0.00000227, 0.00000829, -0.00000275, -0.00002292, -0.00003266, -0.00002294, -0.00000479, 0.00000414, -0.00000104, -0.00000922, -0.00000824, 0.00000111, 0.00000798, 0.00000536, -0.00000242, -0.00000638, -0.00000366, 0.00000070, 0.00000138, -0.00000114, -0.00000282, -0.00000212, -0.00000139, -0.00000278, -0.00000513, -0.00000587, -0.00000461, -0.00000336, },
		{ -0.00000235, 0.00000111, 0.00000466, 0.00000758, 0.00000930, 0.00000834, 0.00000360, -0.00000300, -0.00000691, -0.00000527, -0.00000032, 0.00000297, 0.00000246, 0.00000078, 0.00000093, 0.00000187, 0.00000027, -0.00000382, -0.00000591, -0.00000309, 0.00000062, -0.00000161, -0.00000974, -0.00001424, -0.00000606, 0.00001131, 0.00002313, 0.00001831, 0.00000148, -0.00001172, -0.00001087, 0.00000014, 0.00001019, 0.00001361, 0.00001288, 0.00001030, 0.00000271, -0.00001223, -0.00002720, -0.00002824, -0.00000962, 0.00001682, 0.00003073, 0.00002209, -0.00000028, -0.00001880, -0.00002303, -0.00001440, 0.00000076, 0.00001715, 0.00002867, 0.00002670, 0.00000744, -0.00001875, -0.00003150, -0.00001964, 0.00000489, 0.00001739, 0.00000687, -0.00001129, -0.00001306, 0.00000521, 0.00002090, 0.00001104, -0.00001919, -0.00003948, -0.00002689, 0.00000847, 0.00003361, 0.00002694, -0.00000101, -0.00002150, -0.00001779, 0.00000137, 0.00001562, 0.00001478, 0.00000501, -0.00000227, -0.00000216, 0.00000287, 0.00000862, 0.00001192, 0.00000996, 0.00000182, -0.00000812, -0.00001166, -0.00000516, 0.00000457, 0.00000663, -0.00000072, -0.00000645, 0.00000066, 0.00001704, 0.00002631, 0.00001778, -0.00000005, -0.00000842, -0.00000031, 0.00001074, 0.00000637, -0.00001435, -0.00003268, -0.00003085, -0.00001162, 0.00000559, 0.00000648, -0.00000394, -0.00000995, -0.00000414, 0.00000569, 0.00000795, 0.00000123, -0.00000559, -0.00000515, 0.00000037, 0.00000356, 0.00000132, -0.00000265, -0.00000395, -0.00000289, -0.00000284, -0.00000493, -0.00000679, -0.00000626, -0.00000433, -0.00000347, -0.00000410, -0.00000426, },
		{ 0.00000898, 0.00000813, 0.00000530, 0.00000089, -0.00000409, -0.00000708, -0.00000567, -0.00000075, 0.00000346, 0.00000363, 0.00000098, -0.00000053, 0.00000046, 0.00000106, -0.00000149, -0.00000532, -0.00000595, -0.00000265, -0.00000020, -0.00000249, -0.00000613, -0.00000340, 0.00000737, 0.00001719, 0.00001505, 0.00000088, -0.00001260, -0.00001269, 0.00000018, 0.00001372, 0.00001722, 0.00001119, 0.00000306, -0.00000314, -0.00000918, -0.00001553, -0.00001654, -0.00000675, 0.00000947, 0.00001890, 0.00001243, -0.00000446, -0.00001645, -0.00001346, 0.00000092, 0.00001527, 0.00002157, 0.00001869, 0.00000825, -0.00000763, -0.00002278, -0.00002651, -0.00001276, 0.00000967, 0.00002133, 0.00001127, -0.00000942, -0.00001809, -0.00000593, 0.00001071, 0.00000864, -0.00001384, -0.00003228, -0.00002274, 0.00000980, 0.00003449, 0.00002709, -0.00000420, -0.00002794, -0.00002262, 0.00000302, 0.00002246, 0.00001986, 0.00000296, -0.00000950, -0.00000772, 0.00000334, 0.00001318, 0.00001559, 0.00001002, -0.00000114, -0.00001267, -0.00001676, -0.00000863, 0.00000621, 0.00001429, 0.00000758, -0.00000603, -0.00000978, 0.00000301, 0.00002001, 0.00002309, 0.00000935, -0.00000526, -0.00000500, 0.00000650, 0.00001035, -0.00000440, -0.00002637, -0.00003378, -0.00001903, 0.00000287, 0.00001143, 0.00000265, -0.00000896, -0.00000886, 0.00000152, 0.00000901, 0.00000559, -0.00000314, -0.00000643, -0.00000156, 0.00000437, 0.00000424, -0.00000101, -0.00000523, -0.00000531, -0.00000395, -0.00000471, -0.00000708, -0.00000789, -0.00000600, -0.00000359, -0.00000273, -0.00000250, -0.00000059, 0.00000337, 0.00000727, },
		{ -0.00000180, -0.00000475, -0.00000606, -0.00000485, -0.00000114, 0.00000270, 0.00000366, 0.00000144, -0.00000089, -0.00000072, 0.00000065, -0.00000029, -0.00000405, -0.00000659, -0.00000427, 0.00000099, 0.00000358, 0.00000146, -0.00000075, 0.00000219, 0.00000832, 0.00000961, 0.00000191, -0.00000832, -0.00001004, -0.00000014, 0.00001232, 0.00001574, 0.00000822, -0.00000166, -0.00000554, -0.00000390, -0.00000277, -0.00000445, -0.00000509, -0.00000158, 0.00000263, 0.00000162, -0.00000413, -0.00000650, 0.00000093, 0.00001438, 0.00002275, 0.00001871, 0.00000508, -0.00000975, -0.00001964, -0.00002245, -0.00001683, -0.00000264, 0.00001413, 0.00002116, 0.00001046, -0.00001046, -0.00002267, -0.00001503, 0.00000224, 0.00000776, -0.00000606, -0.00002233, -0.00001805, 0.00000793, 0.00003051, 0.00002487, -0.00000525, -0.00003002, -0.00002496, 0.00000358, 0.00002653, 0.00002330, 0.00000127, -0.00001546, -0.00001220, 0.00000446, 0.00001773, 0.00001795, 0.00000755, -0.00000585, -0.00001583, -0.00001757, -0.00000840, 0.00000761, 0.00001829, 0.00001285, -0.00000503, -0.00001745, -0.00000979, 0.00001224, 0.00002704, 0.00002032, 0.00000093, -0.00000969, -0.00000265, 0.00000847, 0.00000467, -0.00001471, -0.00003028, -0.00002500, -0.00000391, 0.00001183, 0.00000887, -0.00000476, -0.00001110, -0.00000347, 0.00000764, 0.00000907, 0.00000059, -0.00000650, -0.00000411, 0.00000362, 0.00000678, 0.00000183, -0.00000548, -0.00000835, -0.00000664, -0.00000519, -0.00000660, -0.00000842, -0.00000737, -0.00000385, -0.00000079, 0.00000057, 0.00000195, 0.00000491, 0.00000816, 0.00000900, 0.00000650, 0.00000224, },
		{ -0.00000318, -0.00000110, 0.00000129, 0.00000249, 0.00000145, -0.00000060, -0.00000125, -0.00000019, 0.00000005, -0.00000232, -0.00000503, -0.00000382, 0.00000188, 0.00000702, 0.00000641, 0.00000113, -0.00000240, -0.00000027, 0.00000395, 0.00000379, -0.00000137, -0.00000492, -0.00000123, 0.00000665, 0.00001005, 0.00000514, -0.00000215, -0.00000322, 0.00000271, 0.00000759, 0.00000421, -0.00000552, -0.00001312, -0.00001309, -0.00000721, 0.00000009, 0.00000765, 0.00001631, 0.00002343, 0.00002240, 0.00000971, -0.00000923, -0.00002345, -0.00002583, -0.00001775, -0.00000510, 0.00000735, 0.00001614, 0.00001655, 0.00000518, -0.00001316, -0.00002506, -0.00002007, -0.00000332, 0.00000763, 0.00000166, -0.00001201, -0.00001340, 0.00000408, 0.00002323, 0.00002095, -0.00000429, -0.00002794, -0.00002489, 0.00000267, 0.00002705, 0.00002473, 0.00000063, -0.00001896, -0.00001526, 0.00000518, 0.00002091, 0.00001870, 0.00000358, -0.00001108, -0.00001714, -0.00001424, -0.00000462, 0.00000842, 0.00001748, 0.00001342, -0.00000403, -0.00002053, -0.00001825, 0.00000442, 0.00002771, 0.00002939, 0.00000874, -0.00001227, -0.00001333, 0.00000175, 0.00001069, -0.00000043, -0.00002073, -0.00002758, -0.00001315, 0.00000709, 0.00001293, 0.00000202, -0.00000961, -0.00000788, 0.00000361, 0.00001019, 0.00000481, -0.00000448, -0.00000586, 0.00000161, 0.00000787, 0.00000478, -0.00000469, -0.00001126, -0.00001055, -0.00000669, -0.00000550, -0.00000721, -0.00000755, -0.00000412, 0.00000081, 0.00000393, 0.00000499, 0.00000599, 0.00000750, 0.00000753, 0.00000452, -0.00000024, -0.00000392, -0.00000512, -0.00000456, },
		{ 0.00000033, 0.00000055, -0.00000015, -0.00000083, -0.00000045, 0.00000024, -0.00000048, -0.00000223, -0.00000192, 0.00000231, 0.00000749, 0.00000796, 0.00000202, -0.00000498, -0.00000606, -0.00000047, 0.00000532, 0.00000520, 0.00000037, -0.00000282, -0.00000105, 0.00000238, 0.00000286, 0.00000136, 0.00000305, 0.00000925, 0.00001338, 0.00000749, -0.00000781, -0.00002197, -0.00002392, -0.00001190, 0.00000632, 0.00002143, 0.00002859, 0.00002675, 0.00001580, -0.00000236, -0.00002101, -0.00003046, -0.00002547, -0.00001019, 0.00000546, 0.00001409, 0.00001453, 0.00000858, -0.00000249, -0.00001584, -0.00002431, -0.00002046, -0.00000581, 0.00000732, 0.00000729, -0.00000324, -0.00000903, -0.00000011, 0.00001461, 0.00001597, -0.00000219, -0.00002280, -0.00002273, 0.00000056, 0.00002423, 0.00002410, 0.00000121, -0.00001957, -0.00001688, 0.00000472, 0.00002177, 0.00001803, -0.00000043, -0.00001541, -0.00001662, -0.00000811, 0.00000148, 0.00000849, 0.00001215, 0.00000907, -0.00000344, -0.00001837, -0.00002043, -0.00000200, 0.00002398, 0.00003339, 0.00001609, -0.00001120, -0.00002164, -0.00000736, 0.00001217, 0.00001270, -0.00000725, -0.00002547, -0.00002210, -0.00000174, 0.00001343, 0.00000962, -0.00000444, -0.00001037, -0.00000230, 0.00000826, 0.00000852, -0.00000014, -0.00000556, -0.00000107, 0.00000656, 0.00000633, -0.00000328, -0.00001307, -0.00001451, -0.00000896, -0.00000420, -0.00000438, -0.00000609, -0.00000416, 0.00000150, 0.00000658, 0.00000803, 0.00000699, 0.00000587, 0.00000475, 0.00000210, -0.00000197, -0.00000499, -0.00000505, -0.00000302, -0.00000123, -0.00000060, -0.00000028, },
		{ 0.00000061, 0.00000059, 0.00000086, 0.00000099, 0.00000040, 0.00000021, 0.00000227, 0.00000581, 0.00000678, 0.00000220, -0.00000506, -0.00000803, -0.00000303, 0.00000545, 0.00000893, 0.00000400, -0.00000420, -0.00000788, -0.00000462, 0.00000173, 0.00000702, 0.00001097, 0.00001436, 0.00001427, 0.00000573, -0.00001083, -0.00002593, -0.00002750, -0.00001190, 0.00001211, 0.00003058, 0.00003467, 0.00002459, 0.00000617, -0.00001358, -0.00002773, -0.00003035, -0.00001987, -0.00000220, 0.00001212, 0.00001572, 0.00000943, -0.00000070, -0.00000996, -0.00001705, -0.00002051, -0.00001701, -0.00000592, 0.00000620, 0.00000979, 0.00000289, -0.00000500, -0.00000322, 0.00000647, 0.00001048, -0.00000007, -0.00001616, -0.00001894, -0.00000192, 0.00001899, 0.00002157, 0.00000253, -0.00001767, -0.00001719, 0.00000285, 0.00002012, 0.00001638, -0.00000316, -0.00001770, -0.00001470, -0.00000116, 0.00000814, 0.00000788, 0.00000390, 0.00000100, -0.00000354, -0.00001175, -0.00001609, -0.00000595, 0.00001597, 0.00003050, 0.00002092, -0.00000610, -0.00002437, -0.00001584, 0.00000874, 0.00002108, 0.00000689, -0.00001843, -0.00002767, -0.00001225, 0.00000973, 0.00001558, 0.00000308, -0.00000996, -0.00000839, 0.00000369, 0.00001078, 0.00000581, -0.00000278, -0.00000371, 0.00000251, 0.00000526, -0.00000187, -0.00001289, -0.00001697, -0.00001131, -0.00000333, -0.00000083, -0.00000313, -0.00000359, 0.00000127, 0.00000783, 0.00001047, 0.00000806, 0.00000396, 0.00000107, -0.00000087, -0.00000307, -0.00000476, -0.00000421, -0.00000167, 0.00000031, -0.00000006, -0.00000171, -0.00000227, -0.00000115, 0.00000019, },
		{ 0.00000165, 0.00000162, 0.00000150, 0.00000196, 0.00000349, 0.00000442, 0.00000208, -0.00000323, -0.00000678, -0.00000391, 0.00000381, 0.00000883, 0.00000519, -0.00000460, -0.00001149, -0.00000885, 0.00000151, 0.00001222, 0.00001772, 0.00001720, 0.00001164, 0.00000108, -0.00001308, -0.00002444, -0.00002399, -0.00000823, 0.00001533, 0.00003254, 0.00003333, 0.00001850, -0.00000281, -0.00002045, -0.00002826, -0.00002458, -0.00001179, 0.00000368, 0.00001337, 0.00001207, 0.00000207, -0.00000898, -0.00001507, -0.00001603, -0.00001463, -0.00001139, -0.00000469, 0.00000414, 0.00000920, 0.00000612, -0.00000140, -0.00000450, 0.00000007, 0.00000500, 0.00000114, -0.00000954, -0.00001407, -0.00000387, 0.00001263, 0.00001741, 0.00000375, -0.00001420, -0.00001634, 0.00000004, 0.00001647, 0.00001427, -0.00000390, -0.00001748, -0.00001209, 0.00000465, 0.00001361, 0.00000692, -0.00000494, -0.00000864, -0.00000443, -0.00000266, -0.00000670, -0.00000689, 0.00000507, 0.00002081, 0.00002173, 0.00000212, -0.00002000, -0.00002086, 0.00000133, 0.00002232, 0.00001821, -0.00000744, -0.00002744, -0.00002142, 0.00000222, 0.00001752, 0.00001070, -0.00000634, -0.00001246, -0.00000206, 0.00001093, 0.00001178, 0.00000187, -0.00000570, -0.00000374, 0.00000117, -0.00000104, -0.00001029, -0.00001659, -0.00001289, -0.00000345, 0.00000215, 0.00000066, -0.00000198, 0.00000069, 0.00000751, 0.00001160, 0.00000890, 0.00000227, -0.00000277, -0.00000414, -0.00000372, -0.00000329, -0.00000244, -0.00000058, 0.00000087, -0.00000008, -0.00000288, -0.00000450, -0.00000293, 0.00000048, 0.00000279, 0.00000287, 0.00000203, },
		{ 0.00000178, 0.00000204, 0.00000250, 0.00000191, -0.00000082, -0.00000398, -0.00000378, 0.00000087, 0.00000556, 0.00000422, -0.00000365, -0.00001094, -0.00000946, 0.00000164, 0.00001476, 0.00002103, 0.00001759, 0.00000782, -0.00000353, -0.00001331, -0.00001865, -0.00001573, -0.00000292, 0.00001484, 0.00002713, 0.00002553, 0.00001071, -0.00000823, -0.00002107, -0.00002306, -0.00001583, -0.00000435, 0.00000568, 0.00000930, 0.00000450, -0.00000571, -0.00001451, -0.00001663, -0.00001285, -0.00000812, -0.00000545, -0.00000309, 0.00000150, 0.00000643, 0.00000682, 0.00000155, -0.00000399, -0.00000400, 0.00000007, 0.00000105, -0.00000408, -0.00000877, -0.00000466, 0.00000637, 0.00001204, 0.00000401, -0.00001026, -0.00001449, -0.00000280, 0.00001184, 0.00001209, -0.00000278, -0.00001507, -0.00000948, 0.00000800, 0.00001673, 0.00000601, -0.00001209, -0.00001746, -0.00000608, 0.00000634, 0.00000509, -0.00000488, -0.00000636, 0.00000627, 0.00001794, 0.00001144, -0.00000912, -0.00002058, -0.00000796, 0.00001600, 0.00002392, 0.00000536, -0.00002051, -0.00002644, -0.00000764, 0.00001399, 0.00001597, 0.00000000, -0.00001254, -0.00000697, 0.00000877, 0.00001572, 0.00000684, -0.00000663, -0.00001074, -0.00000520, -0.00000118, -0.00000563, -0.00001287, -0.00001296, -0.00000472, 0.00000340, 0.00000437, 0.00000087, 0.00000061, 0.00000589, 0.00001079, 0.00000887, 0.00000106, -0.00000587, -0.00000709, -0.00000404, -0.00000103, 0.00000004, 0.00000019, 0.00000013, -0.00000108, -0.00000371, -0.00000550, -0.00000379, 0.00000092, 0.00000498, 0.00000552, 0.00000334, 0.00000134, 0.00000102, 0.00000152, },
		{ 0.00000174, 0.00000075, -0.00000154, -0.00000335, -0.00000228, 0.00000092, 0.00000208, -0.00000159, -0.00000689, -0.00000688, 0.00000163, 0.00001362, 0.00001999, 0.00001599, 0.00000503, -0.00000556, -0.00001097, -0.00001076, -0.00000619, 0.00000174, 0.00001086, 0.00001644, 0.00001390, 0.00000342, -0.00000886, -0.00001558, -0.00001394, -0.00000694, 0.00000005, 0.00000321, 0.00000132, -0.00000474, -0.00001210, -0.00001643, -0.00001469, -0.00000828, -0.00000234, -0.00000069, -0.00000179, -0.00000110, 0.00000274, 0.00000572, 0.00000352, -0.00000233, -0.00000578, -0.00000377, -0.00000015, -0.00000038, -0.00000378, -0.00000419, 0.00000109, 0.00000602, 0.00000284, -0.00000681, -0.00001178, -0.00000472, 0.00000734, 0.00001001, -0.00000056, -0.00001130, -0.00000737, 0.00000851, 0.00001714, 0.00000550, -0.00001593, -0.00002358, -0.00000830, 0.00001289, 0.00001625, -0.00000058, -0.00001578, -0.00000996, 0.00001004, 0.00001936, 0.00000581, -0.00001454, -0.00001639, 0.00000379, 0.00002261, 0.00001713, -0.00000793, -0.00002552, -0.00001755, 0.00000516, 0.00001705, 0.00000775, -0.00000774, -0.00000917, 0.00000461, 0.00001579, 0.00001020, -0.00000637, -0.00001649, -0.00001215, -0.00000236, -0.00000008, -0.00000642, -0.00001110, -0.00000673, 0.00000247, 0.00000712, 0.00000475, 0.00000167, 0.00000354, 0.00000774, 0.00000729, 0.00000028, -0.00000734, -0.00000881, -0.00000405, 0.00000123, 0.00000253, 0.00000052, -0.00000168, -0.00000293, -0.00000400, -0.00000466, -0.00000297, 0.00000160, 0.00000623, 0.00000718, 0.00000407, 0.00000032, -0.00000077, 0.00000063, 0.00000208, 0.00000227, 0.00000196, },
		{ -0.00000304, -0.00000442, -0.00000299, -0.00000009, 0.00000082, -0.00000122, -0.00000249, 0.00000134, 0.00000936, 0.00001515, 0.00001314, 0.00000454, -0.00000383, -0.00000633, -0.00000320, 0.00000135, 0.00000406, 0.00000466, 0.00000395, 0.00000181, -0.00000183, -0.00000528, -0.00000600, -0.00000341, -0.00000015, 0.00000036, -0.00000296, -0.00000839, -0.00001332, -0.00001589, -0.00001496, -0.00001023, -0.00000338, 0.00000178, 0.00000216, -0.00000107, -0.00000308, -0.00000069, 0.00000358, 0.00000423, -0.00000043, -0.00000574, -0.00000611, -0.00000183, 0.00000147, 0.00000017, -0.00000287, -0.00000276, 0.00000011, 0.00000033, -0.00000437, -0.00000846, -0.00000508, 0.00000380, 0.00000797, 0.00000169, -0.00000726, -0.00000588, 0.00000674, 0.00001525, 0.00000551, -0.00001597, -0.00002596, -0.00001073, 0.00001551, 0.00002423, 0.00000496, -0.00002117, -0.00002439, -0.00000055, 0.00002364, 0.00002123, -0.00000376, -0.00002140, -0.00001107, 0.00001457, 0.00002499, 0.00000758, -0.00001837, -0.00002488, -0.00000714, 0.00001320, 0.00001503, 0.00000136, -0.00000757, -0.00000069, 0.00001114, 0.00001030, -0.00000510, -0.00001910, -0.00001755, -0.00000433, 0.00000473, 0.00000121, -0.00000734, -0.00000850, -0.00000021, 0.00000827, 0.00000889, 0.00000392, 0.00000106, 0.00000274, 0.00000385, -0.00000021, -0.00000662, -0.00000853, -0.00000374, 0.00000253, 0.00000404, 0.00000035, -0.00000384, -0.00000488, -0.00000348, -0.00000194, -0.00000044, 0.00000232, 0.00000566, 0.00000664, 0.00000368, -0.00000072, -0.00000241, -0.00000024, 0.00000294, 0.00000396, 0.00000278, 0.00000141, 0.00000063, -0.00000072, },
		{ -0.00000143, 0.00000264, 0.00000387, 0.00000207, 0.00000102, 0.00000370, 0.00000822, 0.00000970, 0.00000619, 0.00000121, -0.00000031, 0.00000233, 0.00000494, 0.00000337, -0.00000194, -0.00000680, -0.00000766, -0.00000443, 0.00000049, 0.00000455, 0.00000595, 0.00000350, -0.00000286, -0.00001109, -0.00001758, -0.00001952, -0.00001675, -0.00001113, -0.00000472, 0.00000089, 0.00000395, 0.00000306, -0.00000088, -0.00000412, -0.00000316, 0.00000108, 0.00000362, 0.00000093, -0.00000454, -0.00000677, -0.00000329, 0.00000175, 0.00000255, -0.00000140, -0.00000512, -0.00000491, -0.00000287, -0.00000304, -0.00000494, -0.00000388, 0.00000159, 0.00000588, 0.00000306, -0.00000385, -0.00000486, 0.00000381, 0.00001195, 0.00000590, -0.00001279, -0.00002459, -0.00001289, 0.00001392, 0.00002752, 0.00001043, -0.00002162, -0.00003415, -0.00001178, 0.00002301, 0.00003353, 0.00000955, -0.00002129, -0.00002475, 0.00000165, 0.00002686, 0.00002235, -0.00000624, -0.00002732, -0.00001980, 0.00000506, 0.00001975, 0.00001261, -0.00000229, -0.00000590, 0.00000230, 0.00000645, -0.00000327, -0.00001749, -0.00001960, -0.00000663, 0.00000725, 0.00000800, -0.00000212, -0.00000879, -0.00000343, 0.00000758, 0.00001208, 0.00000668, -0.00000102, -0.00000329, -0.00000110, -0.00000061, -0.00000376, -0.00000597, -0.00000314, 0.00000216, 0.00000377, -0.00000025, -0.00000523, -0.00000580, -0.00000197, 0.00000191, 0.00000297, 0.00000254, 0.00000287, 0.00000345, 0.00000212, -0.00000092, -0.00000241, -0.00000010, 0.00000407, 0.00000598, 0.00000409, 0.00000080, -0.00000108, -0.00000172, -0.00000297, -0.00000479, -0.00000471, },
		{ 0.00000549, 0.00000109, -0.00000126, 0.00000125, 0.00000608, 0.00000869, 0.00000778, 0.00000594, 0.00000532, 0.00000441, 0.00000014, -0.00000732, -0.00001329, -0.00001260, -0.00000462, 0.00000599, 0.00001288, 0.00001220, 0.00000413, -0.00000793, -0.00001899, -0.00002452, -0.00002264, -0.00001518, -0.00000630, 0.00000044, 0.00000388, 0.00000442, 0.00000254, -0.00000105, -0.00000426, -0.00000446, -0.00000130, 0.00000192, 0.00000131, -0.00000287, -0.00000593, -0.00000400, 0.00000097, 0.00000311, -0.00000046, -0.00000618, -0.00000836, -0.00000585, -0.00000261, -0.00000176, -0.00000166, 0.00000062, 0.00000373, 0.00000314, -0.00000156, -0.00000403, 0.00000092, 0.00000825, 0.00000636, -0.00000771, -0.00002030, -0.00001427, 0.00000900, 0.00002587, 0.00001460, -0.00001757, -0.00003766, -0.00002148, 0.00001755, 0.00004002, 0.00002255, -0.00001583, -0.00003388, -0.00001321, 0.00002206, 0.00003287, 0.00000842, -0.00002360, -0.00002939, -0.00000563, 0.00002024, 0.00002290, 0.00000538, -0.00000966, -0.00000875, -0.00000081, -0.00000154, -0.00001180, -0.00001739, -0.00000860, 0.00000658, 0.00001218, 0.00000379, -0.00000658, -0.00000570, 0.00000520, 0.00001297, 0.00000876, -0.00000237, -0.00000890, -0.00000648, -0.00000104, 0.00000047, -0.00000165, -0.00000238, -0.00000004, 0.00000150, -0.00000097, -0.00000483, -0.00000470, 0.00000044, 0.00000558, 0.00000575, 0.00000169, -0.00000172, -0.00000168, -0.00000007, 0.00000036, 0.00000005, 0.00000151, 0.00000488, 0.00000687, 0.00000474, -0.00000000, -0.00000338, -0.00000370, -0.00000301, -0.00000348, -0.00000394, -0.00000163, 0.00000328, 0.00000677, },
		{ -0.00000408, 0.00000243, 0.00001027, 0.00001362, 0.00001137, 0.00000644, 0.00000126, -0.00000435, -0.00001059, -0.00001477, -0.00001266, -0.00000314, 0.00000918, 0.00001671, 0.00001429, 0.00000259, -0.00001273, -0.00002463, -0.00002823, -0.00002290, -0.00001225, -0.00000183, 0.00000425, 0.00000543, 0.00000377, 0.00000129, -0.00000137, -0.00000382, -0.00000480, -0.00000326, -0.00000042, 0.00000069, -0.00000133, -0.00000407, -0.00000380, -0.00000033, 0.00000208, -0.00000042, -0.00000628, -0.00000986, -0.00000781, -0.00000272, 0.00000053, 0.00000066, 0.00000046, 0.00000166, 0.00000208, -0.00000042, -0.00000312, -0.00000105, 0.00000497, 0.00000653, -0.00000227, -0.00001435, -0.00001443, 0.00000239, 0.00002016, 0.00001652, -0.00001049, -0.00003481, -0.00002777, 0.00000866, 0.00003951, 0.00003240, -0.00000633, -0.00003649, -0.00002665, 0.00001166, 0.00003678, 0.00002259, -0.00001397, -0.00003318, -0.00001645, 0.00001568, 0.00002908, 0.00001338, -0.00001085, -0.00001915, -0.00000984, -0.00000058, -0.00000347, -0.00001125, -0.00000956, 0.00000291, 0.00001278, 0.00000942, -0.00000156, -0.00000582, 0.00000163, 0.00001062, 0.00000891, -0.00000288, -0.00001251, -0.00001085, -0.00000161, 0.00000468, 0.00000319, -0.00000158, -0.00000344, -0.00000211, -0.00000133, -0.00000213, -0.00000125, 0.00000318, 0.00000751, 0.00000631, -0.00000054, -0.00000691, -0.00000710, -0.00000191, 0.00000316, 0.00000470, 0.00000431, 0.00000478, 0.00000560, 0.00000392, -0.00000069, -0.00000486, -0.00000535, -0.00000304, -0.00000138, -0.00000151, -0.00000084, 0.00000279, 0.00000705, 0.00000712, 0.00000181, -0.00000399, },
		{ 0.00001685, 0.00001531, 0.00000702, -0.00000219, -0.00000850, -0.00001170, -0.00001222, -0.00000891, -0.00000090, 0.00000917, 0.00001525, 0.00001216, -0.00000001, -0.00001549, -0.00002663, -0.00002845, -0.00002111, -0.00000918, 0.00000129, 0.00000626, 0.00000572, 0.00000264, -0.00000006, -0.00000174, -0.00000321, -0.00000459, -0.00000466, -0.00000283, -0.00000064, -0.00000032, -0.00000189, -0.00000288, -0.00000163, 0.00000004, -0.00000122, -0.00000575, -0.00000947, -0.00000832, -0.00000304, 0.00000154, 0.00000222, 0.00000058, -0.00000009, 0.00000047, -0.00000013, -0.00000205, -0.00000176, 0.00000256, 0.00000619, 0.00000223, -0.00000807, -0.00001321, -0.00000411, 0.00001202, 0.00001576, -0.00000246, -0.00002689, -0.00002950, -0.00000144, 0.00003247, 0.00003692, 0.00000472, -0.00003232, -0.00003568, -0.00000176, 0.00003349, 0.00003331, -0.00000030, -0.00003003, -0.00002500, 0.00000655, 0.00002895, 0.00001947, -0.00000896, -0.00002599, -0.00001836, -0.00000091, 0.00000518, -0.00000261, -0.00000893, -0.00000249, 0.00000995, 0.00001376, 0.00000557, -0.00000335, -0.00000240, 0.00000490, 0.00000636, -0.00000265, -0.00001294, -0.00001288, -0.00000240, 0.00000729, 0.00000701, -0.00000079, -0.00000669, -0.00000565, -0.00000087, 0.00000250, 0.00000383, 0.00000535, 0.00000654, 0.00000375, -0.00000390, -0.00001103, -0.00001074, -0.00000251, 0.00000678, 0.00001014, 0.00000737, 0.00000355, 0.00000213, 0.00000158, -0.00000086, -0.00000440, -0.00000553, -0.00000303, 0.00000021, 0.00000111, 0.00000056, 0.00000189, 0.00000545, 0.00000704, 0.00000301, -0.00000410, -0.00000697, -0.00000117, 0.00000960, },
		{ -0.00000373, -0.00001092, -0.00001128, -0.00000777, -0.00000340, 0.00000133, 0.00000634, 0.00000929, 0.00000660, -0.00000282, -0.00001534, -0.00002451, -0.00002551, -0.00001824, -0.00000698, 0.00000251, 0.00000651, 0.00000516, 0.00000157, -0.00000111, -0.00000211, -0.00000281, -0.00000427, -0.00000552, -0.00000483, -0.00000225, -0.00000003, -0.00000006, -0.00000164, -0.00000253, -0.00000218, -0.00000248, -0.00000488, -0.00000768, -0.00000745, -0.00000337, 0.00000121, 0.00000249, 0.00000055, -0.00000125, -0.00000097, -0.00000022, -0.00000089, -0.00000141, 0.00000114, 0.00000533, 0.00000507, -0.00000254, -0.00001077, -0.00000904, 0.00000331, 0.00001251, 0.00000453, -0.00001603, -0.00002651, -0.00001036, 0.00002077, 0.00003506, 0.00001442, -0.00002280, -0.00003831, -0.00001473, 0.00002423, 0.00003838, 0.00001426, -0.00002077, -0.00002946, -0.00000536, 0.00002206, 0.00002190, -0.00000436, -0.00002728, -0.00002414, -0.00000263, 0.00001177, 0.00000644, -0.00000638, -0.00000769, 0.00000481, 0.00001592, 0.00001316, 0.00000119, -0.00000603, -0.00000323, 0.00000118, -0.00000197, -0.00000993, -0.00001186, -0.00000341, 0.00000709, 0.00000852, 0.00000011, -0.00000814, -0.00000746, 0.00000060, 0.00000773, 0.00000889, 0.00000597, 0.00000244, -0.00000166, -0.00000752, -0.00001251, -0.00001101, -0.00000143, 0.00000995, 0.00001435, 0.00000949, 0.00000149, -0.00000246, -0.00000153, -0.00000030, -0.00000180, -0.00000389, -0.00000308, 0.00000027, 0.00000257, 0.00000218, 0.00000161, 0.00000344, 0.00000575, 0.00000399, -0.00000244, -0.00000756, -0.00000449, 0.00000640, 0.00001661, 0.00001748, 0.00000830, },
		{ -0.00000377, 0.00000162, 0.00000298, 0.00000196, 0.00000080, -0.00000101, -0.00000545, -0.00001248, -0.00001872, -0.00002003, -0.00001499, -0.00000621, 0.00000170, 0.00000525, 0.00000414, 0.00000086, -0.00000167, -0.00000250, -0.00000285, -0.00000417, -0.00000595, -0.00000615, -0.00000375, -0.00000043, 0.00000097, -0.00000045, -0.00000280, -0.00000384, -0.00000363, -0.00000395, -0.00000527, -0.00000571, -0.00000360, -0.00000019, 0.00000144, 0.00000018, -0.00000172, -0.00000170, -0.00000028, 0.00000017, -0.00000047, 0.00000052, 0.00000412, 0.00000615, 0.00000156, -0.00000756, -0.00001162, -0.00000429, 0.00000754, 0.00000910, -0.00000467, -0.00001968, -0.00001624, 0.00000701, 0.00002720, 0.00002027, -0.00001043, -0.00003398, -0.00002391, 0.00001165, 0.00003679, 0.00002610, -0.00000790, -0.00002898, -0.00001733, 0.00000987, 0.00001983, 0.00000173, -0.00002263, -0.00002562, -0.00000529, 0.00001463, 0.00001383, -0.00000205, -0.00001075, -0.00000098, 0.00001537, 0.00001912, 0.00000650, -0.00000841, -0.00001191, -0.00000564, -0.00000129, -0.00000430, -0.00000804, -0.00000456, 0.00000381, 0.00000723, 0.00000140, -0.00000647, -0.00000628, 0.00000279, 0.00001163, 0.00001188, 0.00000438, -0.00000387, -0.00000841, -0.00001017, -0.00001053, -0.00000747, 0.00000101, 0.00001128, 0.00001549, 0.00000980, -0.00000054, -0.00000637, -0.00000422, 0.00000083, 0.00000209, -0.00000094, -0.00000325, -0.00000143, 0.00000227, 0.00000376, 0.00000275, 0.00000239, 0.00000385, 0.00000386, -0.00000062, -0.00000665, -0.00000701, 0.00000160, 0.00001332, 0.00001788, 0.00001079, -0.00000214, -0.00001074, -0.00001011, },
		{ -0.00000240, -0.00000776, -0.00000899, -0.00000784, -0.00000801, -0.00001051, -0.00001277, -0.00001168, -0.00000683, -0.00000089, 0.00000283, 0.00000294, 0.00000064, -0.00000175, -0.00000291, -0.00000340, -0.00000447, -0.00000614, -0.00000674, -0.00000482, -0.00000125, 0.00000113, 0.00000039, -0.00000242, -0.00000445, -0.00000426, -0.00000314, -0.00000302, -0.00000379, -0.00000371, -0.00000210, -0.00000049, -0.00000053, -0.00000158, -0.00000157, -0.00000008, 0.00000100, 0.00000055, 0.00000044, 0.00000286, 0.00000581, 0.00000396, -0.00000420, -0.00001179, -0.00000962, 0.00000199, 0.00001070, 0.00000502, -0.00001067, -0.00001824, -0.00000606, 0.00001499, 0.00002093, 0.00000190, -0.00002367, -0.00002697, -0.00000112, 0.00002892, 0.00003210, 0.00000514, -0.00002382, -0.00002639, -0.00000461, 0.00001356, 0.00000744, -0.00001338, -0.00002231, -0.00000791, 0.00001342, 0.00001810, 0.00000343, -0.00001048, -0.00000578, 0.00001205, 0.00002157, 0.00001101, -0.00000895, -0.00001888, -0.00001252, -0.00000114, 0.00000218, -0.00000255, -0.00000556, -0.00000166, 0.00000369, 0.00000329, -0.00000130, -0.00000186, 0.00000498, 0.00001238, 0.00001112, 0.00000062, -0.00001048, -0.00001420, -0.00001072, -0.00000544, -0.00000112, 0.00000384, 0.00000988, 0.00001275, 0.00000812, -0.00000153, -0.00000783, -0.00000540, 0.00000197, 0.00000566, 0.00000219, -0.00000337, -0.00000411, 0.00000056, 0.00000510, 0.00000533, 0.00000293, 0.00000177, 0.00000190, -0.00000019, -0.00000524, -0.00000809, -0.00000301, 0.00000833, 0.00001665, 0.00001404, 0.00000249, -0.00000779, -0.00000873, -0.00000190, 0.00000426, 0.00000359, },
		{ -0.00000998, -0.00000341, -0.00000160, -0.00000466, -0.00000823, -0.00000833, -0.00000475, -0.00000044, 0.00000164, 0.00000081, -0.00000146, -0.00000337, -0.00000440, -0.00000523, -0.00000628, -0.00000673, -0.00000530, -0.00000214, 0.00000064, 0.00000080, -0.00000157, -0.00000396, -0.00000414, -0.00000255, -0.00000151, -0.00000230, -0.00000373, -0.00000389, -0.00000263, -0.00000144, -0.00000111, -0.00000077, 0.00000040, 0.00000152, 0.00000130, 0.00000061, 0.00000178, 0.00000461, 0.00000480, -0.00000127, -0.00001002, -0.00001214, -0.00000297, 0.00000962, 0.00001150, -0.00000155, -0.00001660, -0.00001621, 0.00000102, 0.00001654, 0.00001170, -0.00000970, -0.00002335, -0.00001124, 0.00001641, 0.00003055, 0.00001497, -0.00001524, -0.00003016, -0.00001774, 0.00000437, 0.00001086, -0.00000220, -0.00001489, -0.00000920, 0.00000919, 0.00001872, 0.00000894, -0.00000690, -0.00000848, 0.00000653, 0.00001954, 0.00001338, -0.00000745, -0.00002224, -0.00001780, -0.00000198, 0.00000738, 0.00000302, -0.00000594, -0.00000748, -0.00000070, 0.00000578, 0.00000643, 0.00000477, 0.00000619, 0.00000895, 0.00000605, -0.00000446, -0.00001513, -0.00001676, -0.00000865, 0.00000126, 0.00000591, 0.00000587, 0.00000580, 0.00000671, 0.00000505, -0.00000072, -0.00000590, -0.00000459, 0.00000237, 0.00000724, 0.00000431, -0.00000306, -0.00000624, -0.00000147, 0.00000599, 0.00000846, 0.00000461, -0.00000028, -0.00000190, -0.00000191, -0.00000404, -0.00000720, -0.00000563, 0.00000337, 0.00001402, 0.00001682, 0.00000883, -0.00000256, -0.00000717, -0.00000267, 0.00000354, 0.00000280, -0.00000555, -0.00001411, -0.00001561, },
		{ 0.00000326, -0.00000447, -0.00000982, -0.00000913, -0.00000431, 0.00000011, 0.00000113, -0.00000094, -0.00000384, -0.00000573, -0.00000638, -0.00000649, -0.00000627, -0.00000511, -0.00000271, -0.00000013, 0.00000080, -0.00000057, -0.00000270, -0.00000338, -0.00000216, -0.00000092, -0.00000154, -0.00000366, -0.00000507, -0.00000436, -0.00000234, -0.00000065, 0.00000025, 0.00000104, 0.00000179, 0.00000170, 0.00000086, 0.00000102, 0.00000310, 0.00000442, 0.00000084, -0.00000710, -0.00001194, -0.00000639, 0.00000667, 0.00001410, 0.00000575, -0.00001241, -0.00002204, -0.00001179, 0.00000858, 0.00001731, 0.00000483, -0.00001429, -0.00001693, 0.00000183, 0.00002169, 0.00001928, -0.00000521, -0.00002757, -0.00002631, -0.00000577, 0.00001059, 0.00000778, -0.00000498, -0.00000796, 0.00000404, 0.00001616, 0.00001309, -0.00000127, -0.00000880, -0.00000006, 0.00001339, 0.00001293, -0.00000425, -0.00002107, -0.00002025, -0.00000396, 0.00000952, 0.00000716, -0.00000517, -0.00001138, -0.00000417, 0.00000848, 0.00001461, 0.00001157, 0.00000566, 0.00000169, -0.00000236, -0.00000943, -0.00001601, -0.00001479, -0.00000433, 0.00000741, 0.00001112, 0.00000615, 0.00000014, -0.00000073, 0.00000170, 0.00000190, -0.00000097, -0.00000211, 0.00000150, 0.00000588, 0.00000475, -0.00000179, -0.00000625, -0.00000257, 0.00000607, 0.00001053, 0.00000615, -0.00000224, -0.00000668, -0.00000535, -0.00000329, -0.00000429, -0.00000515, -0.00000011, 0.00001030, 0.00001777, 0.00001492, 0.00000400, -0.00000506, -0.00000541, -0.00000054, 0.00000029, -0.00000708, -0.00001681, -0.00001938, -0.00001168, -0.00000006, 0.00000615, },
		{ -0.00001307, -0.00000845, -0.00000192, 0.00000117, -0.00000045, -0.00000429, -0.00000716, -0.00000773, -0.00000676, -0.00000550, -0.00000428, -0.00000271, -0.00000079, 0.00000050, 0.00000022, -0.00000122, -0.00000225, -0.00000189, -0.00000105, -0.00000148, -0.00000352, -0.00000543, -0.00000528, -0.00000306, -0.00000049, 0.00000103, 0.00000164, 0.00000196, 0.00000185, 0.00000107, 0.00000057, 0.00000165, 0.00000330, 0.00000201, -0.00000379, -0.00000961, -0.00000787, 0.00000286, 0.00001297, 0.00000992, -0.00000717, -0.00002315, -0.00002096, -0.00000071, 0.00001819, 0.00001676, -0.00000252, -0.00001775, -0.00001190, 0.00000769, 0.00001734, 0.00000407, -0.00001929, -0.00002845, -0.00001473, 0.00000633, 0.00001393, 0.00000535, -0.00000359, 0.00000031, 0.00001163, 0.00001467, 0.00000433, -0.00000724, -0.00000627, 0.00000472, 0.00000982, -0.00000024, -0.00001583, -0.00001946, -0.00000682, 0.00000794, 0.00000898, -0.00000282, -0.00001159, -0.00000535, 0.00001057, 0.00002058, 0.00001618, 0.00000314, -0.00000759, -0.00001183, -0.00001274, -0.00001253, -0.00000850, 0.00000107, 0.00001093, 0.00001262, 0.00000448, -0.00000526, -0.00000714, -0.00000082, 0.00000548, 0.00000531, 0.00000106, -0.00000063, 0.00000189, 0.00000364, 0.00000078, -0.00000327, -0.00000201, 0.00000492, 0.00000989, 0.00000607, -0.00000400, -0.00001086, -0.00000904, -0.00000281, 0.00000002, -0.00000169, -0.00000138, 0.00000597, 0.00001599, 0.00001887, 0.00001060, -0.00000200, -0.00000868, -0.00000697, -0.00000404, -0.00000738, -0.00001549, -0.00001931, -0.00001245, 0.00000126, 0.00001113, 0.00000982, -0.00000033, -0.00001035, },
		{ 0.00000042, -0.00000028, -0.00000464, -0.00000838, -0.00000897, -0.00000697, -0.00000451, -0.00000295, -0.00000210, -0.00000115, -0.00000003, 0.00000050, -0.00000004, -0.00000107, -0.00000157, -0.00000147, -0.00000182, -0.00000332, -0.00000507, -0.00000535, -0.00000350, -0.00000072, 0.00000128, 0.00000201, 0.00000210, 0.00000191, 0.00000121, 0.00000032, 0.00000047, 0.00000193, 0.00000242, -0.00000071, -0.00000606, -0.00000754, -0.00000091, 0.00000899, 0.00001054, -0.00000223, -0.00002006, -0.00002496, -0.00000899, 0.00001477, 0.00002356, 0.00000864, -0.00001451, -0.00002215, -0.00000802, 0.00001011, 0.00001063, -0.00000752, -0.00002406, -0.00002064, -0.00000095, 0.00001492, 0.00001402, 0.00000362, -0.00000026, 0.00000666, 0.00001300, 0.00000778, -0.00000481, -0.00001070, -0.00000412, 0.00000493, 0.00000335, -0.00000820, -0.00001584, -0.00000979, 0.00000331, 0.00000850, 0.00000109, -0.00000751, -0.00000377, 0.00001112, 0.00002201, 0.00001674, -0.00000084, -0.00001615, -0.00001944, -0.00001328, -0.00000558, 0.00000032, 0.00000591, 0.00001066, 0.00000989, 0.00000145, -0.00000854, -0.00001041, -0.00000187, 0.00000854, 0.00001063, 0.00000378, -0.00000325, -0.00000317, 0.00000184, 0.00000448, 0.00000229, 0.00000008, 0.00000221, 0.00000560, 0.00000340, -0.00000529, -0.00001270, -0.00001107, -0.00000218, 0.00000454, 0.00000343, -0.00000054, 0.00000175, 0.00001154, 0.00001948, 0.00001590, 0.00000208, -0.00001088, -0.00001412, -0.00000987, -0.00000732, -0.00001091, -0.00001515, -0.00001159, 0.00000054, 0.00001234, 0.00001387, 0.00000393, -0.00000916, -0.00001559, -0.00001243, -0.00000470, },
		{ -0.00000898, -0.00000968, -0.00000694, -0.00000341, -0.00000139, -0.00000110, -0.00000122, -0.00000071, 0.00000020, 0.00000054, -0.00000009, -0.00000107, -0.00000172, -0.00000219, -0.00000308, -0.00000431, -0.00000480, -0.00000363, -0.00000125, 0.00000095, 0.00000203, 0.00000222, 0.00000199, 0.00000130, 0.00000019, -0.00000035, 0.00000071, 0.00000237, 0.00000173, -0.00000227, -0.00000599, -0.00000405, 0.00000343, 0.00000815, 0.00000149, -0.00001399, -0.00002359, -0.00001456, 0.00000823, 0.00002398, 0.00001622, -0.00000884, -0.00002716, -0.00002163, -0.00000018, 0.00001314, 0.00000459, -0.00001476, -0.00002231, -0.00000920, 0.00001107, 0.00001944, 0.00001234, 0.00000284, 0.00000265, 0.00000828, 0.00000766, -0.00000266, -0.00001243, -0.00001080, -0.00000049, 0.00000533, -0.00000052, -0.00001043, -0.00001171, -0.00000250, 0.00000648, 0.00000595, -0.00000012, -0.00000011, 0.00000940, 0.00001783, 0.00001269, -0.00000514, -0.00002121, -0.00002270, -0.00001083, 0.00000267, 0.00000894, 0.00000876, 0.00000681, 0.00000396, -0.00000179, -0.00000855, -0.00000963, -0.00000154, 0.00000957, 0.00001294, 0.00000524, -0.00000509, -0.00000710, 0.00000046, 0.00000851, 0.00000875, 0.00000271, -0.00000198, -0.00000201, -0.00000177, -0.00000578, -0.00001099, -0.00000995, -0.00000105, 0.00000785, 0.00000822, 0.00000161, -0.00000154, 0.00000548, 0.00001657, 0.00001872, 0.00000660, -0.00001091, -0.00002004, -0.00001631, -0.00000779, -0.00000464, -0.00000760, -0.00000857, -0.00000148, 0.00000966, 0.00001457, 0.00000770, -0.00000595, -0.00001584, -0.00001572, -0.00000841, -0.00000167, -0.00000078, -0.00000480, },
		{ -0.00000237, 0.00000002, -0.00000007, -0.00000116, -0.00000145, -0.00000054, 0.00000051, 0.00000059, -0.00000036, -0.00000151, -0.00000227, -0.00000281, -0.00000349, -0.00000398, -0.00000351, -0.00000185, 0.00000022, 0.00000169, 0.00000223, 0.00000208, 0.00000136, 0.00000014, -0.00000078, -0.00000009, 0.00000212, 0.00000333, 0.00000089, -0.00000402, -0.00000624, -0.00000228, 0.00000395, 0.00000365, -0.00000654, -0.00001792, -0.00001671, 0.00000015, 0.00001828, 0.00001842, -0.00000268, -0.00002642, -0.00003009, -0.00001075, 0.00001128, 0.00001390, -0.00000328, -0.00001945, -0.00001592, 0.00000409, 0.00002080, 0.00002051, 0.00000875, 0.00000054, 0.00000162, 0.00000378, -0.00000169, -0.00001131, -0.00001378, -0.00000518, 0.00000496, 0.00000502, -0.00000447, -0.00001146, -0.00000708, 0.00000410, 0.00001059, 0.00000826, 0.00000409, 0.00000528, 0.00000873, 0.00000497, -0.00000834, -0.00002093, -0.00002036, -0.00000612, 0.00000953, 0.00001462, 0.00000887, 0.00000092, -0.00000294, -0.00000398, -0.00000533, -0.00000545, -0.00000056, 0.00000766, 0.00001132, 0.00000534, -0.00000482, -0.00000789, 0.00000036, 0.00001155, 0.00001369, 0.00000444, -0.00000704, -0.00001120, -0.00000820, -0.00000520, -0.00000566, -0.00000527, 0.00000069, 0.00000888, 0.00001081, 0.00000399, -0.00000318, -0.00000052, 0.00001092, 0.00001823, 0.00001047, -0.00000847, -0.00002304, -0.00002193, -0.00000921, 0.00000142, 0.00000177, -0.00000333, -0.00000358, 0.00000397, 0.00001166, 0.00001016, -0.00000111, -0.00001344, -0.00001774, -0.00001289, -0.00000543, -0.00000227, -0.00000466, -0.00000853, -0.00000942, -0.00000650, },
		{ -0.00000115, -0.00000212, -0.00000141, 0.00000014, 0.00000100, 0.00000048, -0.00000083, -0.00000194, -0.00000251, -0.00000286, -0.00000322, -0.00000321, -0.00000231, -0.00000063, 0.00000104, 0.00000200, 0.00000209, 0.00000141, 0.00000021, -0.00000080, -0.00000038, 0.00000186, 0.00000400, 0.00000284, -0.00000232, -0.00000739, -0.00000689, -0.00000059, 0.00000440, 0.00000064, -0.00000993, -0.00001577, -0.00000788, 0.00000809, 0.00001502, 0.00000224, -0.00002079, -0.00003187, -0.00001903, 0.00000573, 0.00001819, 0.00000741, -0.00001272, -0.00001890, -0.00000354, 0.00001816, 0.00002586, 0.00001548, 0.00000061, -0.00000518, -0.00000272, -0.00000230, -0.00000811, -0.00001277, -0.00000822, 0.00000228, 0.00000715, 0.00000094, -0.00000832, -0.00000843, 0.00000238, 0.00001354, 0.00001470, 0.00000696, -0.00000055, -0.00000291, -0.00000419, -0.00000935, -0.00001513, -0.00001295, -0.00000058, 0.00001282, 0.00001564, 0.00000650, -0.00000472, -0.00000827, -0.00000437, -0.00000022, 0.00000023, 0.00000013, 0.00000304, 0.00000636, 0.00000461, -0.00000172, -0.00000466, 0.00000169, 0.00001222, 0.00001483, 0.00000407, -0.00001182, -0.00001933, -0.00001394, -0.00000354, 0.00000207, 0.00000207, 0.00000287, 0.00000734, 0.00001014, 0.00000557, -0.00000286, -0.00000483, 0.00000391, 0.00001426, 0.00001240, -0.00000413, -0.00002197, -0.00002507, -0.00001135, 0.00000548, 0.00001076, 0.00000353, -0.00000447, -0.00000300, 0.00000568, 0.00001049, 0.00000431, -0.00000866, -0.00001800, -0.00001766, -0.00001077, -0.00000489, -0.00000420, -0.00000670, -0.00000786, -0.00000559, -0.00000166, 0.00000081, 0.00000056, },
		{ -0.00000026, 0.00000121, 0.00000127, 0.00000011, -0.00000128, -0.00000212, -0.00000245, -0.00000269, -0.00000285, -0.00000250, -0.00000135, 0.00000022, 0.00000146, 0.00000188, 0.00000144, 0.00000045, -0.00000044, -0.00000024, 0.00000159, 0.00000378, 0.00000339, -0.00000127, -0.00000757, -0.00000959, -0.00000418, 0.00000412, 0.00000618, -0.00000189, -0.00001278, -0.00001450, -0.00000403, 0.00000726, 0.00000478, -0.00001220, -0.00002727, -0.00002332, -0.00000191, 0.00001678, 0.00001490, -0.00000360, -0.00001700, -0.00000938, 0.00001236, 0.00002665, 0.00002066, 0.00000246, -0.00001021, -0.00000983, -0.00000436, -0.00000419, -0.00000875, -0.00000918, -0.00000182, 0.00000591, 0.00000513, -0.00000240, -0.00000569, 0.00000185, 0.00001355, 0.00001672, 0.00000712, -0.00000671, -0.00001378, -0.00001205, -0.00000779, -0.00000546, -0.00000261, 0.00000421, 0.00001165, 0.00001189, 0.00000272, -0.00000796, -0.00001015, -0.00000298, 0.00000467, 0.00000498, -0.00000015, -0.00000289, 0.00000002, 0.00000395, 0.00000387, 0.00000186, 0.00000379, 0.00000963, 0.00001104, 0.00000119, -0.00001496, -0.00002375, -0.00001702, -0.00000118, 0.00000999, 0.00001010, 0.00000517, 0.00000389, 0.00000642, 0.00000573, -0.00000092, -0.00000651, -0.00000286, 0.00000738, 0.00001145, 0.00000096, -0.00001666, -0.00002432, -0.00001341, 0.00000644, 0.00001710, 0.00001072, -0.00000318, -0.00000906, -0.00000204, 0.00000815, 0.00000876, -0.00000253, -0.00001630, -0.00002181, -0.00001702, -0.00000850, -0.00000351, -0.00000352, -0.00000494, -0.00000426, -0.00000157, 0.00000060, 0.00000048, -0.00000124, -0.00000250, -0.00000201, },
		{ 0.00000097, -0.00000046, -0.00000160, -0.00000215, -0.00000237, -0.00000254, -0.00000247, -0.00000183, -0.00000062, 0.00000065, 0.00000141, 0.00000141, 0.00000085, 0.00000020, 0.00000016, 0.00000125, 0.00000282, 0.00000273, -0.00000093, -0.00000688, -0.00001010, -0.00000609, 0.00000322, 0.00000919, 0.00000425, -0.00000908, -0.00001880, -0.00001533, -0.00000267, 0.00000453, -0.00000314, -0.00001819, -0.00002305, -0.00000955, 0.00001070, 0.00001794, 0.00000596, -0.00001051, -0.00001181, 0.00000483, 0.00002225, 0.00002229, 0.00000524, -0.00001200, -0.00001525, -0.00000721, -0.00000107, -0.00000350, -0.00000811, -0.00000584, 0.00000254, 0.00000787, 0.00000520, 0.00000045, 0.00000234, 0.00001008, 0.00001317, 0.00000414, -0.00001155, -0.00002066, -0.00001631, -0.00000422, 0.00000513, 0.00000769, 0.00000710, 0.00000677, 0.00000491, -0.00000101, -0.00000767, -0.00000810, -0.00000063, 0.00000729, 0.00000688, -0.00000142, -0.00000800, -0.00000518, 0.00000411, 0.00001050, 0.00000951, 0.00000549, 0.00000389, 0.00000290, -0.00000357, -0.00001537, -0.00002274, -0.00001624, 0.00000121, 0.00001565, 0.00001648, 0.00000720, -0.00000002, 0.00000095, 0.00000428, 0.00000167, -0.00000563, -0.00000810, -0.00000105, 0.00000751, 0.00000547, -0.00000793, -0.00001905, -0.00001426, 0.00000411, 0.00001906, 0.00001660, 0.00000053, -0.00001219, -0.00000963, 0.00000313, 0.00001056, 0.00000337, -0.00001277, -0.00002420, -0.00002294, -0.00001267, -0.00000290, 0.00000054, -0.00000094, -0.00000268, -0.00000226, -0.00000080, -0.00000038, -0.00000138, -0.00000238, -0.00000203, -0.00000040, 0.00000128, 0.00000180, },
		{ -0.00000175, -0.00000214, -0.00000232, -0.00000236, -0.00000208, -0.00000133, -0.00000028, 0.00000070, 0.00000126, 0.00000129, 0.00000095, 0.00000062, 0.00000079, 0.00000140, 0.00000132, -0.00000104, -0.00000551, -0.00000871, -0.00000627, 0.00000199, 0.00000942, 0.00000733, -0.00000582, -0.00002041, -0.00002354, -0.00001233, 0.00000176, 0.00000402, -0.00000743, -0.00001875, -0.00001514, 0.00000222, 0.00001659, 0.00001396, -0.00000112, -0.00001046, -0.00000280, 0.00001334, 0.00001941, 0.00000780, -0.00001007, -0.00001723, -0.00000993, 0.00000013, 0.00000106, -0.00000549, -0.00000817, -0.00000108, 0.00000924, 0.00001265, 0.00000800, 0.00000319, 0.00000363, 0.00000478, -0.00000126, -0.00001369, -0.00002161, -0.00001585, 0.00000013, 0.00001345, 0.00001510, 0.00000779, 0.00000033, -0.00000273, -0.00000344, -0.00000418, -0.00000320, 0.00000149, 0.00000646, 0.00000523, -0.00000306, -0.00001021, -0.00000713, 0.00000525, 0.00001591, 0.00001542, 0.00000569, -0.00000373, -0.00000737, -0.00000875, -0.00001267, -0.00001628, -0.00001159, 0.00000293, 0.00001726, 0.00001935, 0.00000870, -0.00000276, -0.00000435, 0.00000150, 0.00000356, -0.00000322, -0.00001102, -0.00000917, 0.00000144, 0.00000835, 0.00000247, -0.00000979, -0.00001296, -0.00000064, 0.00001598, 0.00001963, 0.00000606, -0.00001113, -0.00001517, -0.00000394, 0.00000846, 0.00000725, -0.00000796, -0.00002372, -0.00002692, -0.00001663, -0.00000280, 0.00000453, 0.00000344, -0.00000103, -0.00000362, -0.00000322, -0.00000189, -0.00000142, -0.00000163, -0.00000137, -0.00000020, 0.00000121, 0.00000191, 0.00000149, 0.00000032, -0.00000092, },
		},
		{ { 0.00000163, 0.00000196, 0.00000145, -0.00000080, -0.00000238, -0.00000213, -0.00000205, -0.00000246, -0.00000143, -0.00000023, -0.00000207, -0.00000518, -0.00000443, 0.00000036, 0.00000448, 0.00000507, 0.00000202, -0.00000356, -0.00000637, 0.00000048, 0.00001359, 0.00001837, 0.00000755, -0.00000794, -0.00001308, -0.00000540, 0.00000662, 0.00001254, 0.00000546, -0.00001030, -0.00001766, -0.00000588, 0.00001136, 0.00001209, -0.00000370, -0.00001662, -0.00001591, -0.00000723, 0.00000173, 0.00000592, 0.00000109, -0.00000770, -0.00000657, 0.00000435, 0.00000597, -0.00000795, -0.00001665, -0.00000671, 0.00000607, 0.00000659, 0.00000309, 0.00000285, -0.00000189, -0.00000850, -0.00000230, 0.00000706, -0.00000572, -0.00002596, -0.00001133, 0.00002622, 0.00002388, -0.00002265, -0.00003712, 0.00001096, 0.00004628, 0.00001096, -0.00003712, -0.00002265, 0.00002388, 0.00002622, -0.00001133, -0.00002596, -0.00000572, 0.00000706, -0.00000230, -0.00000850, -0.00000189, 0.00000285, 0.00000309, 0.00000659, 0.00000607, -0.00000671, -0.00001665, -0.00000795, 0.00000597, 0.00000435, -0.00000657, -0.00000770, 0.00000109, 0.00000592, 0.00000173, -0.00000723, -0.00001591, -0.00001662, -0.00000370, 0.00001209, 0.00001136, -0.00000588, -0.00001766, -0.00001030, 0.00000546, 0.00001254, 0.00000662, -0.00000540, -0.00001308, -0.00000794, 0.00000755, 0.00001837, 0.00001359, 0.00000048, -0.00000637, -0.00000356, 0.00000202, 0.00000507, 0.00000448, 0.00000036, -0.00000443, -0.00000518, -0.00000207, -0.00000023, -0.00000143, -0.00000246, -0.00000205, -0.00000213, -0.00000238, -0.00000080, 0.00000145, 0.00000196, },
		{ 0.00000098, -0.00000120, -0.00000298, -0.00000264, -0.00000192, -0.00000180, -0.00000093, -0.00000021, -0.00000225, -0.00000502, -0.00000342, 0.00000205, 0.00000530, 0.00000324, -0.00000159, -0.00000500, -0.00000269, 0.00000712, 0.00001713, 0.00001479, -0.00000063, -0.00001331, -0.00001034, 0.00000254, 0.00001068, 0.00000723, -0.00000419, -0.00001378, -0.00001089, 0.00000398, 0.00001382, 0.00000448, -0.00001384, -0.00001978, -0.00001037, -0.00000017, 0.00000283, 0.00000179, -0.00000133, -0.00000495, -0.00000303, 0.00000306, 0.00000087, -0.00001064, -0.00001354, -0.00000121, 0.00000781, 0.00000260, -0.00000215, 0.00000231, 0.00000234, -0.00000574, -0.00000540, 0.00000172, -0.00000687, -0.00002280, -0.00000927, 0.00002481, 0.00002226, -0.00002195, -0.00003595, 0.00001044, 0.00004513, 0.00001102, -0.00003602, -0.00002170, 0.00002388, 0.00002473, -0.00001363, -0.00002625, -0.00000212, 0.00001147, -0.00000219, -0.00001253, -0.00000407, 0.00000628, 0.00000841, 0.00000750, 0.00000158, -0.00001137, -0.00001632, -0.00000332, 0.00000937, 0.00000274, -0.00001041, -0.00000789, 0.00000524, 0.00000812, -0.00000302, -0.00001486, -0.00001724, -0.00000845, 0.00000666, 0.00001440, 0.00000344, -0.00001571, -0.00001907, -0.00000188, 0.00001462, 0.00001291, -0.00000128, -0.00001213, -0.00001118, 0.00000046, 0.00001417, 0.00001766, 0.00000730, -0.00000522, -0.00000703, -0.00000008, 0.00000526, 0.00000518, 0.00000246, -0.00000124, -0.00000468, -0.00000490, -0.00000201, -0.00000052, -0.00000197, -0.00000293, -0.00000199, -0.00000150, -0.00000169, -0.00000033, 0.00000187, 0.00000229, 0.00000151, 0.00000145, },
		{ -0.00000348, -0.00000299, -0.00000157, -0.00000104, -0.00000060, -0.00000043, -0.00000227, -0.00000422, -0.00000199, 0.00000310, 0.00000452, 0.00000034, -0.00000421, -0.00000370, 0.00000303, 0.00001263, 0.00001680, 0.00000812, -0.00000786, -0.00001442, -0.00000451, 0.00000886, 0.00000995, -0.00000077, -0.00001111, -0.00001161, -0.00000152, 0.00001046, 0.00001030, -0.00000557, -0.00002067, -0.00001777, -0.00000316, 0.00000417, 0.00000079, -0.00000219, -0.00000115, -0.00000101, -0.00000148, -0.00000084, -0.00000389, -0.00001012, -0.00000813, 0.00000288, 0.00000611, -0.00000281, -0.00000531, 0.00000457, 0.00000663, -0.00000563, -0.00001052, -0.00000248, -0.00000504, -0.00001833, -0.00000907, 0.00002086, 0.00002064, -0.00001900, -0.00003366, 0.00000830, 0.00004220, 0.00001208, -0.00003242, -0.00002057, 0.00002109, 0.00002125, -0.00001421, -0.00002317, 0.00000218, 0.00001313, -0.00000465, -0.00001567, -0.00000303, 0.00001126, 0.00001159, 0.00000493, -0.00000371, -0.00001320, -0.00001286, 0.00000115, 0.00000978, -0.00000074, -0.00001262, -0.00000531, 0.00000922, 0.00000680, -0.00001033, -0.00002021, -0.00001303, 0.00000261, 0.00001367, 0.00001024, -0.00000737, -0.00002182, -0.00001417, 0.00000883, 0.00002001, 0.00000783, -0.00001049, -0.00001497, -0.00000461, 0.00000947, 0.00001703, 0.00001283, 0.00000030, -0.00000824, -0.00000475, 0.00000401, 0.00000695, 0.00000356, -0.00000001, -0.00000216, -0.00000421, -0.00000450, -0.00000227, -0.00000101, -0.00000239, -0.00000318, -0.00000177, -0.00000075, -0.00000090, 0.00000014, 0.00000212, 0.00000235, 0.00000119, 0.00000087, 0.00000053, -0.00000155, },
		{ -0.00000105, -0.00000032, -0.00000048, -0.00000071, -0.00000191, -0.00000290, -0.00000069, 0.00000294, 0.00000238, -0.00000249, -0.00000473, 0.00000015, 0.00000910, 0.00001534, 0.00001274, 0.00000044, -0.00001215, -0.00001159, 0.00000197, 0.00001161, 0.00000528, -0.00000848, -0.00001333, -0.00000515, 0.00000704, 0.00001160, 0.00000234, -0.00001472, -0.00002255, -0.00001212, 0.00000308, 0.00000502, -0.00000278, -0.00000439, 0.00000117, 0.00000230, -0.00000260, -0.00000587, -0.00000647, -0.00000654, -0.00000248, 0.00000399, 0.00000176, -0.00000746, -0.00000547, 0.00000822, 0.00000905, -0.00000821, -0.00001579, -0.00000411, -0.00000097, -0.00001446, -0.00001130, 0.00001569, 0.00002031, -0.00001429, -0.00003172, 0.00000427, 0.00003882, 0.00001514, -0.00002720, -0.00002091, 0.00001544, 0.00001759, -0.00001164, -0.00001764, 0.00000499, 0.00001136, -0.00000794, -0.00001596, 0.00000089, 0.00001548, 0.00001116, 0.00000006, -0.00000740, -0.00001145, -0.00000794, 0.00000345, 0.00000729, -0.00000412, -0.00001199, -0.00000115, 0.00001085, 0.00000174, -0.00001774, -0.00002078, -0.00000390, 0.00001310, 0.00001461, 0.00000068, -0.00001720, -0.00002160, -0.00000429, 0.00001813, 0.00001949, -0.00000117, -0.00001773, -0.00001247, 0.00000493, 0.00001626, 0.00001511, 0.00000543, -0.00000510, -0.00000781, -0.00000062, 0.00000717, 0.00000657, 0.00000085, -0.00000207, -0.00000216, -0.00000332, -0.00000433, -0.00000284, -0.00000152, -0.00000260, -0.00000322, -0.00000142, 0.00000011, -0.00000002, 0.00000051, 0.00000210, 0.00000216, 0.00000078, 0.00000033, 0.00000012, -0.00000187, -0.00000387, -0.00000311, },
		{ -0.00000051, -0.00000083, -0.00000109, -0.00000139, -0.00000011, 0.00000145, -0.00000039, -0.00000406, -0.00000275, 0.00000544, 0.00001371, 0.00001451, 0.00000625, -0.00000607, -0.00001271, -0.00000638, 0.00000674, 0.00001033, -0.00000116, -0.00001325, -0.00001069, 0.00000286, 0.00001189, 0.00000741, -0.00000719, -0.00002030, -0.00001970, -0.00000531, 0.00000655, 0.00000308, -0.00000597, -0.00000422, 0.00000426, 0.00000360, -0.00000581, -0.00001009, -0.00000600, -0.00000129, 0.00000134, 0.00000188, -0.00000341, -0.00000967, -0.00000312, 0.00001129, 0.00000862, -0.00001225, -0.00001931, -0.00000298, 0.00000366, -0.00001253, -0.00001536, 0.00001089, 0.00002164, -0.00000909, -0.00003114, -0.00000090, 0.00003645, 0.00002020, -0.00002200, -0.00002374, 0.00000811, 0.00001578, -0.00000573, -0.00001168, 0.00000462, 0.00000701, -0.00000971, -0.00001255, 0.00000589, 0.00001671, 0.00000722, -0.00000478, -0.00000771, -0.00000694, -0.00000389, 0.00000260, 0.00000338, -0.00000550, -0.00000865, 0.00000245, 0.00000875, -0.00000558, -0.00002227, -0.00001558, 0.00000760, 0.00001935, 0.00000875, -0.00001110, -0.00002222, -0.00001460, 0.00000729, 0.00002256, 0.00001301, -0.00001099, -0.00002023, -0.00000500, 0.00001457, 0.00001860, 0.00000895, -0.00000212, -0.00000721, -0.00000434, 0.00000364, 0.00000828, 0.00000440, -0.00000198, -0.00000316, -0.00000142, -0.00000246, -0.00000452, -0.00000359, -0.00000190, -0.00000260, -0.00000311, -0.00000094, 0.00000110, 0.00000084, 0.00000065, 0.00000174, 0.00000179, 0.00000042, -0.00000013, -0.00000029, -0.00000217, -0.00000409, -0.00000300, -0.00000042, 0.00000027, },
		{ 0.00000003, -0.00000020, -0.00000062, -0.00000102, -0.00000273, -0.00000359, 0.00000129, 0.00001059, 0.00001556, 0.00001053, -0.00000078, -0.00000988, -0.00001006, -0.00000093, 0.00000841, 0.00000605, -0.00000676, -0.00001374, -0.00000480, 0.00000936, 0.00001168, -0.00000034, -0.00001526, -0.00002117, -0.00001378, 0.00000029, 0.00000678, -0.00000008, -0.00000742, -0.00000227, 0.00000651, 0.00000252, -0.00000961, -0.00001198, -0.00000295, 0.00000361, 0.00000216, -0.00000236, -0.00000737, -0.00000897, 0.00000016, 0.00001214, 0.00000567, -0.00001589, -0.00002002, -0.00000008, 0.00000712, -0.00001286, -0.00001989, 0.00000757, 0.00002398, -0.00000483, -0.00003202, -0.00000580, 0.00003598, 0.00002617, -0.00001847, -0.00002890, 0.00000113, 0.00001709, 0.00000226, -0.00000757, 0.00000071, 0.00000217, -0.00000809, -0.00000625, 0.00000944, 0.00001382, 0.00000148, -0.00000711, -0.00000435, -0.00000179, -0.00000259, -0.00000080, 0.00000020, -0.00000393, -0.00000420, 0.00000333, 0.00000310, -0.00001244, -0.00002163, -0.00000572, 0.00001772, 0.00001892, -0.00000225, -0.00002090, -0.00002012, -0.00000281, 0.00001645, 0.00002025, 0.00000269, -0.00001803, -0.00001674, 0.00000524, 0.00002123, 0.00001571, 0.00000055, -0.00000737, -0.00000549, 0.00000068, 0.00000639, 0.00000701, 0.00000133, -0.00000402, -0.00000317, -0.00000039, -0.00000195, -0.00000503, -0.00000430, -0.00000207, -0.00000242, -0.00000287, -0.00000032, 0.00000217, 0.00000155, 0.00000044, 0.00000111, 0.00000142, 0.00000020, -0.00000054, -0.00000076, -0.00000243, -0.00000407, -0.00000266, 0.00000019, 0.00000068, -0.00000056, -0.00000063, },
		{ -0.00000217, -0.00000369, -0.00000370, -0.00000096, 0.00000623, 0.00001407, 0.00001424, 0.00000469, -0.00000654, -0.00001054, -0.00000567, 0.00000294, 0.00000691, 0.00000082, -0.00000948, -0.00001027, 0.00000175, 0.00001206, 0.00000686, -0.00000888, -0.00001969, -0.00001784, -0.00000714, 0.00000322, 0.00000458, -0.00000279, -0.00000683, 0.00000015, 0.00000686, -0.00000019, -0.00001228, -0.00001106, 0.00000117, 0.00000639, 0.00000011, -0.00000686, -0.00000880, -0.00000614, 0.00000251, 0.00001019, 0.00000161, -0.00001749, -0.00001805, 0.00000293, 0.00000834, -0.00001466, -0.00002341, 0.00000586, 0.00002603, -0.00000236, -0.00003349, -0.00000904, 0.00003723, 0.00003139, -0.00001748, -0.00003502, -0.00000353, 0.00002147, 0.00001012, -0.00000685, -0.00000551, -0.00000071, -0.00000260, 0.00000073, 0.00000936, 0.00000742, -0.00000341, -0.00000542, 0.00000130, 0.00000152, -0.00000459, -0.00000485, -0.00000040, -0.00000002, -0.00000099, 0.00000044, -0.00000421, -0.00001586, -0.00001532, 0.00000582, 0.00002289, 0.00001163, -0.00001471, -0.00002495, -0.00001100, 0.00000993, 0.00001972, 0.00001164, -0.00000784, -0.00001946, -0.00000810, 0.00001494, 0.00002275, 0.00000859, -0.00000734, -0.00000872, -0.00000077, 0.00000528, 0.00000658, 0.00000395, -0.00000146, -0.00000477, -0.00000247, 0.00000041, -0.00000189, -0.00000562, -0.00000480, -0.00000206, -0.00000214, -0.00000247, 0.00000047, 0.00000320, 0.00000195, -0.00000015, 0.00000035, 0.00000119, 0.00000014, -0.00000098, -0.00000131, -0.00000256, -0.00000376, -0.00000215, 0.00000070, 0.00000095, -0.00000047, -0.00000014, 0.00000109, 0.00000022, },
		{ -0.00000277, 0.00000321, 0.00001066, 0.00001494, 0.00001031, -0.00000135, -0.00000990, -0.00000867, -0.00000131, 0.00000429, 0.00000339, -0.00000324, -0.00000859, -0.00000456, 0.00000636, 0.00001017, -0.00000066, -0.00001556, -0.00001974, -0.00001201, -0.00000183, 0.00000328, 0.00000139, -0.00000401, -0.00000483, 0.00000172, 0.00000524, -0.00000310, -0.00001267, -0.00000803, 0.00000450, 0.00000626, -0.00000354, -0.00000984, -0.00000759, -0.00000285, 0.00000258, 0.00000609, -0.00000176, -0.00001635, -0.00001460, 0.00000452, 0.00000731, -0.00001657, -0.00002510, 0.00000501, 0.00002655, -0.00000158, -0.00003419, -0.00000994, 0.00003904, 0.00003432, -0.00001868, -0.00004006, -0.00000475, 0.00002742, 0.00001558, -0.00000958, -0.00001171, 0.00000012, 0.00000555, 0.00000566, 0.00000486, -0.00000031, -0.00000493, 0.00000006, 0.00000668, 0.00000130, -0.00000874, -0.00000717, 0.00000213, 0.00000421, -0.00000104, -0.00000554, -0.00001020, -0.00001392, -0.00000499, 0.00001515, 0.00002112, -0.00000027, -0.00002427, -0.00002135, 0.00000251, 0.00001923, 0.00001556, -0.00000052, -0.00001467, -0.00001435, 0.00000296, 0.00002087, 0.00001868, -0.00000033, -0.00001218, -0.00000597, 0.00000507, 0.00000774, 0.00000419, 0.00000030, -0.00000304, -0.00000433, -0.00000167, 0.00000067, -0.00000212, -0.00000604, -0.00000508, -0.00000196, -0.00000175, -0.00000185, 0.00000137, 0.00000400, 0.00000191, -0.00000100, -0.00000033, 0.00000119, 0.00000014, -0.00000155, -0.00000191, -0.00000245, -0.00000312, -0.00000157, 0.00000103, 0.00000117, -0.00000017, 0.00000047, 0.00000170, -0.00000030, -0.00000428, -0.00000563, },
		{ 0.00001336, 0.00001307, 0.00000498, -0.00000614, -0.00001066, -0.00000556, 0.00000170, 0.00000338, -0.00000041, -0.00000471, -0.00000476, 0.00000104, 0.00000732, 0.00000448, -0.00000836, -0.00001868, -0.00001613, -0.00000580, 0.00000101, 0.00000127, -0.00000137, -0.00000357, -0.00000256, 0.00000173, 0.00000247, -0.00000480, -0.00001065, -0.00000441, 0.00000563, 0.00000365, -0.00000701, -0.00001032, -0.00000473, -0.00000079, 0.00000015, 0.00000135, -0.00000307, -0.00001290, -0.00001129, 0.00000393, 0.00000497, -0.00001732, -0.00002504, 0.00000377, 0.00002495, -0.00000153, -0.00003296, -0.00000889, 0.00003983, 0.00003436, -0.00002062, -0.00004219, -0.00000286, 0.00003269, 0.00001729, -0.00001428, -0.00001539, 0.00000484, 0.00001382, 0.00000646, -0.00000300, -0.00000640, -0.00000190, 0.00000738, 0.00000921, -0.00000251, -0.00001268, -0.00000608, 0.00000662, 0.00000621, -0.00000500, -0.00001223, -0.00001200, -0.00000671, 0.00000597, 0.00001898, 0.00001288, -0.00001273, -0.00002741, -0.00001083, 0.00001613, 0.00002175, 0.00000498, -0.00001214, -0.00001508, -0.00000406, 0.00001277, 0.00002100, 0.00001049, -0.00000799, -0.00001255, -0.00000038, 0.00000986, 0.00000721, 0.00000013, -0.00000265, -0.00000307, -0.00000324, -0.00000132, 0.00000039, -0.00000232, -0.00000616, -0.00000523, -0.00000186, -0.00000118, -0.00000097, 0.00000223, 0.00000436, 0.00000143, -0.00000192, -0.00000074, 0.00000140, 0.00000005, -0.00000231, -0.00000244, -0.00000200, -0.00000221, -0.00000104, 0.00000119, 0.00000142, 0.00000030, 0.00000089, 0.00000156, -0.00000157, -0.00000620, -0.00000611, -0.00000008, 0.00000780, },
		{ -0.00000036, -0.00000890, -0.00000945, -0.00000255, 0.00000286, 0.00000129, -0.00000287, -0.00000337, 0.00000031, 0.00000445, 0.00000436, -0.00000311, -0.00001400, -0.00001784, -0.00001050, -0.00000096, 0.00000131, -0.00000153, -0.00000288, -0.00000210, -0.00000105, 0.00000028, -0.00000017, -0.00000451, -0.00000710, -0.00000180, 0.00000418, -0.00000007, -0.00000881, -0.00000851, -0.00000187, -0.00000098, -0.00000391, -0.00000234, -0.00000191, -0.00000846, -0.00000944, 0.00000142, 0.00000267, -0.00001636, -0.00002416, 0.00000111, 0.00002159, -0.00000094, -0.00002942, -0.00000713, 0.00003832, 0.00003209, -0.00002133, -0.00004062, 0.00000051, 0.00003515, 0.00001550, -0.00001849, -0.00001503, 0.00001189, 0.00001941, 0.00000267, -0.00001160, -0.00000840, 0.00000497, 0.00001363, 0.00000746, -0.00000818, -0.00001388, -0.00000150, 0.00001062, 0.00000414, -0.00001159, -0.00001642, -0.00000816, 0.00000350, 0.00001367, 0.00001596, 0.00000101, -0.00002144, -0.00002283, 0.00000342, 0.00002544, 0.00001641, -0.00000876, -0.00001920, -0.00000851, 0.00000821, 0.00001796, 0.00001525, 0.00000107, -0.00001183, -0.00000865, 0.00000580, 0.00001187, 0.00000397, -0.00000415, -0.00000398, -0.00000184, -0.00000227, -0.00000169, -0.00000012, -0.00000221, -0.00000602, -0.00000540, -0.00000172, -0.00000031, 0.00000009, 0.00000278, 0.00000415, 0.00000067, -0.00000260, -0.00000081, 0.00000164, -0.00000029, -0.00000320, -0.00000272, -0.00000114, -0.00000114, -0.00000067, 0.00000119, 0.00000173, 0.00000075, 0.00000083, 0.00000064, -0.00000307, -0.00000716, -0.00000490, 0.00000365, 0.00001159, 0.00001368, 0.00000908, },
		{ -0.00000731, -0.00000052, 0.00000257, -0.00000057, -0.00000317, -0.00000011, 0.00000460, 0.00000456, -0.00000151, -0.00001036, -0.00001631, -0.00001390, -0.00000473, 0.00000163, -0.00000017, -0.00000396, -0.00000310, -0.00000051, -0.00000075, -0.00000181, -0.00000162, -0.00000240, -0.00000343, -0.00000121, 0.00000081, -0.00000332, -0.00000837, -0.00000554, -0.00000043, -0.00000335, -0.00000809, -0.00000390, 0.00000105, -0.00000457, -0.00000955, -0.00000196, 0.00000148, -0.00001402, -0.00002364, -0.00000327, 0.00001756, 0.00000115, -0.00002421, -0.00000617, 0.00003423, 0.00002910, -0.00001925, -0.00003599, 0.00000315, 0.00003381, 0.00001191, -0.00001985, -0.00001076, 0.00001858, 0.00002043, -0.00000425, -0.00001785, -0.00000544, 0.00001324, 0.00001617, 0.00000185, -0.00001294, -0.00001084, 0.00000485, 0.00001141, -0.00000215, -0.00001791, -0.00001544, 0.00000050, 0.00001292, 0.00001524, 0.00000730, -0.00001020, -0.00002338, -0.00001195, 0.00001689, 0.00002743, 0.00000477, -0.00002126, -0.00001914, 0.00000324, 0.00001847, 0.00001669, 0.00000557, -0.00000627, -0.00001065, -0.00000220, 0.00001019, 0.00001051, -0.00000068, -0.00000721, -0.00000355, -0.00000015, -0.00000200, -0.00000262, -0.00000046, -0.00000169, -0.00000582, -0.00000569, -0.00000139, 0.00000092, 0.00000115, 0.00000278, 0.00000339, -0.00000006, -0.00000284, -0.00000062, 0.00000168, -0.00000093, -0.00000402, -0.00000256, 0.00000010, -0.00000007, -0.00000053, 0.00000107, 0.00000204, 0.00000097, 0.00000014, -0.00000077, -0.00000412, -0.00000668, -0.00000228, 0.00000739, 0.00001363, 0.00001170, 0.00000407, -0.00000463, -0.00000962, },
		{ 0.00000179, -0.00000116, -0.00000149, 0.00000348, 0.00000658, 0.00000147, -0.00000848, -0.00001547, -0.00001520, -0.00000835, -0.00000023, 0.00000199, -0.00000239, -0.00000542, -0.00000251, 0.00000048, -0.00000142, -0.00000352, -0.00000145, 0.00000060, -0.00000092, -0.00000270, -0.00000315, -0.00000496, -0.00000611, -0.00000290, -0.00000111, -0.00000690, -0.00001093, -0.00000320, 0.00000450, -0.00000230, -0.00001117, -0.00000494, 0.00000172, -0.00001130, -0.00002433, -0.00000878, 0.00001409, 0.00000482, -0.00001872, -0.00000706, 0.00002843, 0.00002727, -0.00001397, -0.00003006, 0.00000318, 0.00002923, 0.00000897, -0.00001714, -0.00000440, 0.00002219, 0.00001676, -0.00001144, -0.00001943, 0.00000134, 0.00001977, 0.00001385, -0.00000538, -0.00001412, -0.00000395, 0.00001005, 0.00000733, -0.00001076, -0.00002055, -0.00000840, 0.00001104, 0.00001776, 0.00001012, -0.00000357, -0.00001668, -0.00001801, 0.00000156, 0.00002535, 0.00002160, -0.00000948, -0.00002854, -0.00001181, 0.00001657, 0.00002336, 0.00000928, -0.00000473, -0.00000896, -0.00000508, 0.00000417, 0.00001113, 0.00000644, -0.00000501, -0.00000826, -0.00000193, 0.00000112, -0.00000258, -0.00000366, -0.00000031, -0.00000093, -0.00000579, -0.00000601, -0.00000069, 0.00000245, 0.00000185, 0.00000204, 0.00000231, -0.00000040, -0.00000256, -0.00000044, 0.00000128, -0.00000178, -0.00000450, -0.00000188, 0.00000156, 0.00000082, -0.00000066, 0.00000084, 0.00000224, 0.00000076, -0.00000111, -0.00000216, -0.00000413, -0.00000470, 0.00000098, 0.00001011, 0.00001356, 0.00000808, -0.00000084, -0.00000731, -0.00000886, -0.00000515, 0.00000040, },
		{ 0.00000117, 0.00000584, 0.00000568, -0.00000367, -0.00001468, -0.00001756, -0.00001153, -0.00000279, 0.00000231, 0.00000074, -0.00000443, -0.00000584, -0.00000176, 0.00000066, -0.00000237, -0.00000416, -0.00000003, 0.00000328, -0.00000022, -0.00000546, -0.00000634, -0.00000471, -0.00000322, -0.00000174, -0.00000357, -0.00001020, -0.00001163, -0.00000103, 0.00000708, -0.00000192, -0.00001322, -0.00000666, 0.00000291, -0.00000931, -0.00002632, -0.00001429, 0.00001195, 0.00000926, -0.00001437, -0.00000982, 0.00002260, 0.00002791, -0.00000637, -0.00002501, -0.00000004, 0.00002330, 0.00000882, -0.00001086, 0.00000132, 0.00002115, 0.00001020, -0.00001584, -0.00001584, 0.00000917, 0.00002197, 0.00000754, -0.00001112, -0.00001046, 0.00000454, 0.00001124, -0.00000127, -0.00001827, -0.00001703, 0.00000314, 0.00001933, 0.00001571, 0.00000036, -0.00001221, -0.00001613, -0.00000751, 0.00001314, 0.00002639, 0.00001008, -0.00002190, -0.00002842, 0.00000058, 0.00002735, 0.00002125, -0.00000194, -0.00001218, -0.00000606, 0.00000271, 0.00000791, 0.00000830, 0.00000136, -0.00000751, -0.00000739, -0.00000017, 0.00000140, -0.00000367, -0.00000429, 0.00000031, -0.00000024, -0.00000601, -0.00000613, 0.00000051, 0.00000400, 0.00000188, 0.00000063, 0.00000133, -0.00000009, -0.00000195, -0.00000062, 0.00000039, -0.00000256, -0.00000439, -0.00000070, 0.00000298, 0.00000136, -0.00000105, 0.00000050, 0.00000219, 0.00000011, -0.00000256, -0.00000294, -0.00000281, -0.00000166, 0.00000395, 0.00001114, 0.00001158, 0.00000375, -0.00000489, -0.00000850, -0.00000735, -0.00000342, 0.00000067, 0.00000149, -0.00000018, },
		{ 0.00000231, -0.00000928, -0.00001874, -0.00001674, -0.00000662, 0.00000165, 0.00000290, -0.00000130, -0.00000582, -0.00000556, -0.00000126, 0.00000030, -0.00000292, -0.00000360, 0.00000176, 0.00000472, -0.00000110, -0.00000828, -0.00000798, -0.00000322, -0.00000099, -0.00000242, -0.00000676, -0.00001206, -0.00001033, 0.00000135, 0.00000796, -0.00000284, -0.00001463, -0.00000701, 0.00000406, -0.00000870, -0.00002894, -0.00001868, 0.00001107, 0.00001310, -0.00001202, -0.00001343, 0.00001850, 0.00003118, 0.00000165, -0.00002245, -0.00000566, 0.00001843, 0.00001231, -0.00000305, 0.00000387, 0.00001573, 0.00000371, -0.00001542, -0.00000860, 0.00001480, 0.00001902, -0.00000014, -0.00001269, -0.00000274, 0.00001134, 0.00000697, -0.00001190, -0.00002110, -0.00000706, 0.00001556, 0.00002171, 0.00000707, -0.00001001, -0.00001499, -0.00000895, 0.00000403, 0.00001904, 0.00002023, -0.00000318, -0.00002879, -0.00002118, 0.00001423, 0.00003235, 0.00001279, -0.00001339, -0.00001423, 0.00000144, 0.00000977, 0.00000743, 0.00000278, -0.00000271, -0.00000752, -0.00000553, 0.00000071, 0.00000071, -0.00000460, -0.00000422, 0.00000110, 0.00000007, -0.00000632, -0.00000574, 0.00000210, 0.00000513, 0.00000102, -0.00000113, 0.00000089, 0.00000091, -0.00000138, -0.00000144, -0.00000083, -0.00000288, -0.00000354, 0.00000074, 0.00000409, 0.00000148, -0.00000165, -0.00000000, 0.00000181, -0.00000085, -0.00000376, -0.00000268, -0.00000028, 0.00000170, 0.00000583, 0.00001033, 0.00000832, -0.00000047, -0.00000780, -0.00000861, -0.00000558, -0.00000203, 0.00000100, 0.00000216, 0.00000188, 0.00000347, 0.00000600, },
		{ -0.00002011, -0.00001380, -0.00000174, 0.00000440, 0.00000200, -0.00000345, -0.00000646, -0.00000490, -0.00000106, -0.00000015, -0.00000274, -0.00000234, 0.00000303, 0.00000465, -0.00000275, -0.00001014, -0.00000809, -0.00000157, -0.00000024, -0.00000444, -0.00000942, -0.00001200, -0.00000790, 0.00000281, 0.00000716, -0.00000406, -0.00001478, -0.00000656, 0.00000425, -0.00000937, -0.00003110, -0.00002136, 0.00001057, 0.00001505, -0.00001152, -0.00001627, 0.00001719, 0.00003601, 0.00000809, -0.00002271, -0.00001167, 0.00001646, 0.00001861, 0.00000361, 0.00000209, 0.00000801, 0.00000020, -0.00001008, -0.00000070, 0.00001588, 0.00001218, -0.00000589, -0.00000906, 0.00000636, 0.00001355, -0.00000197, -0.00002074, -0.00001699, 0.00000712, 0.00002443, 0.00001647, -0.00000525, -0.00001653, -0.00001045, 0.00000187, 0.00001218, 0.00001764, 0.00000959, -0.00001388, -0.00002853, -0.00000936, 0.00002509, 0.00003024, 0.00000060, -0.00002151, -0.00001017, 0.00001102, 0.00001337, 0.00000265, -0.00000334, -0.00000416, -0.00000540, -0.00000400, 0.00000018, -0.00000034, -0.00000474, -0.00000359, 0.00000157, -0.00000006, -0.00000634, -0.00000468, 0.00000372, 0.00000543, -0.00000063, -0.00000270, 0.00000131, 0.00000233, -0.00000133, -0.00000298, -0.00000199, -0.00000237, -0.00000202, 0.00000206, 0.00000463, 0.00000121, -0.00000237, -0.00000069, 0.00000112, -0.00000179, -0.00000423, -0.00000128, 0.00000296, 0.00000457, 0.00000618, 0.00000797, 0.00000453, -0.00000407, -0.00000966, -0.00000805, -0.00000368, -0.00000052, 0.00000191, 0.00000365, 0.00000400, 0.00000433, 0.00000384, -0.00000239, -0.00001387, },
		{ 0.00000217, 0.00000541, 0.00000020, -0.00000528, -0.00000642, -0.00000404, -0.00000095, -0.00000033, -0.00000197, -0.00000109, 0.00000324, 0.00000349, -0.00000418, -0.00001061, -0.00000737, -0.00000077, -0.00000105, -0.00000676, -0.00001064, -0.00001035, -0.00000548, 0.00000283, 0.00000533, -0.00000469, -0.00001375, -0.00000623, 0.00000308, -0.00001058, -0.00003188, -0.00002261, 0.00000920, 0.00001456, -0.00001179, -0.00001683, 0.00001858, 0.00004058, 0.00001165, -0.00002474, -0.00001587, 0.00001790, 0.00002558, 0.00000689, -0.00000338, 0.00000094, 0.00000131, -0.00000170, 0.00000470, 0.00001196, 0.00000436, -0.00000717, -0.00000143, 0.00001336, 0.00000996, -0.00001264, -0.00002412, -0.00000615, 0.00002133, 0.00002626, 0.00000473, -0.00001670, -0.00001602, -0.00000002, 0.00001185, 0.00001382, 0.00001009, -0.00000141, -0.00001891, -0.00002192, 0.00000319, 0.00003025, 0.00002197, -0.00001168, -0.00002390, -0.00000123, 0.00001948, 0.00001197, -0.00000501, -0.00000775, -0.00000243, -0.00000246, -0.00000401, -0.00000153, -0.00000084, -0.00000383, -0.00000293, 0.00000131, -0.00000034, -0.00000563, -0.00000309, 0.00000480, 0.00000464, -0.00000261, -0.00000352, 0.00000258, 0.00000362, -0.00000216, -0.00000497, -0.00000254, -0.00000087, -0.00000019, 0.00000286, 0.00000451, 0.00000067, -0.00000316, -0.00000158, 0.00000030, -0.00000231, -0.00000369, 0.00000105, 0.00000620, 0.00000635, 0.00000501, 0.00000461, 0.00000077, -0.00000692, -0.00001070, -0.00000695, -0.00000138, 0.00000148, 0.00000337, 0.00000524, 0.00000520, 0.00000333, 0.00000000, -0.00000712, -0.00001650, -0.00001901, -0.00000981, },
		{ -0.00000198, -0.00000655, -0.00000575, -0.00000294, -0.00000076, -0.00000026, -0.00000110, -0.00000041, 0.00000246, 0.00000198, -0.00000472, -0.00000994, -0.00000674, -0.00000134, -0.00000281, -0.00000835, -0.00001022, -0.00000798, -0.00000395, 0.00000158, 0.00000337, -0.00000436, -0.00001220, -0.00000673, 0.00000082, -0.00001138, -0.00003100, -0.00002333, 0.00000598, 0.00001200, -0.00001138, -0.00001442, 0.00002148, 0.00004318, 0.00001233, -0.00002659, -0.00001676, 0.00002165, 0.00003070, 0.00000604, -0.00001027, -0.00000281, 0.00000671, 0.00000666, 0.00000558, 0.00000468, -0.00000115, -0.00000327, 0.00000729, 0.00001546, 0.00000173, -0.00002100, -0.00002007, 0.00000859, 0.00003097, 0.00001993, -0.00000981, -0.00002279, -0.00000797, 0.00001247, 0.00001679, 0.00000838, -0.00000027, -0.00000881, -0.00001738, -0.00001183, 0.00001276, 0.00002882, 0.00001027, -0.00002062, -0.00002003, 0.00000985, 0.00002403, 0.00000567, -0.00001313, -0.00000872, 0.00000186, -0.00000036, -0.00000609, -0.00000352, 0.00000002, -0.00000215, -0.00000298, 0.00000027, -0.00000020, -0.00000393, -0.00000145, 0.00000477, 0.00000294, -0.00000423, -0.00000324, 0.00000429, 0.00000419, -0.00000389, -0.00000686, -0.00000203, 0.00000146, 0.00000144, 0.00000280, 0.00000381, 0.00000006, -0.00000395, -0.00000262, -0.00000038, -0.00000204, -0.00000213, 0.00000376, 0.00000876, 0.00000679, 0.00000265, 0.00000081, -0.00000269, -0.00000907, -0.00001099, -0.00000513, 0.00000161, 0.00000402, 0.00000486, 0.00000606, 0.00000491, 0.00000072, -0.00000449, -0.00001080, -0.00001690, -0.00001609, -0.00000577, 0.00000462, 0.00000494, },
		{ -0.00000443, -0.00000157, -0.00000051, -0.00000027, -0.00000062, -0.00000043, 0.00000120, 0.00000083, -0.00000428, -0.00000882, -0.00000690, -0.00000309, -0.00000462, -0.00000863, -0.00000860, -0.00000584, -0.00000360, -0.00000028, 0.00000196, -0.00000331, -0.00001092, -0.00000824, -0.00000175, -0.00001105, -0.00002889, -0.00002463, 0.00000065, 0.00000849, -0.00000911, -0.00000945, 0.00002422, 0.00004294, 0.00001136, -0.00002635, -0.00001426, 0.00002557, 0.00003216, 0.00000207, -0.00001575, -0.00000208, 0.00001427, 0.00001204, 0.00000190, -0.00000291, -0.00000205, 0.00000427, 0.00001357, 0.00001175, -0.00000788, -0.00002346, -0.00000925, 0.00002271, 0.00003279, 0.00000727, -0.00002222, -0.00002092, 0.00000521, 0.00002223, 0.00001435, -0.00000202, -0.00000921, -0.00001029, -0.00001077, -0.00000195, 0.00001702, 0.00002203, -0.00000135, -0.00002408, -0.00001130, 0.00001983, 0.00002316, -0.00000390, -0.00001914, -0.00000568, 0.00000709, -0.00000052, -0.00000985, -0.00000456, 0.00000257, -0.00000051, -0.00000431, -0.00000112, 0.00000104, -0.00000140, -0.00000054, 0.00000332, 0.00000091, -0.00000478, -0.00000192, 0.00000571, 0.00000360, -0.00000615, -0.00000794, -0.00000028, 0.00000415, 0.00000230, 0.00000181, 0.00000280, -0.00000042, -0.00000473, -0.00000369, -0.00000062, -0.00000080, 0.00000016, 0.00000621, 0.00001020, 0.00000600, -0.00000044, -0.00000310, -0.00000578, -0.00001052, -0.00001032, -0.00000229, 0.00000534, 0.00000661, 0.00000558, 0.00000552, 0.00000317, -0.00000275, -0.00000866, -0.00001288, -0.00001528, -0.00001216, -0.00000242, 0.00000544, 0.00000333, -0.00000412, -0.00000706, },
		{ -0.00000051, -0.00000073, -0.00000069, -0.00000085, 0.00000005, 0.00000032, -0.00000328, -0.00000795, -0.00000803, -0.00000539, -0.00000572, -0.00000761, -0.00000657, -0.00000455, -0.00000414, -0.00000199, 0.00000128, -0.00000217, -0.00001044, -0.00001037, -0.00000379, -0.00000955, -0.00002649, -0.00002722, -0.00000618, 0.00000536, -0.00000466, -0.00000322, 0.00002542, 0.00004023, 0.00001058, -0.00002298, -0.00000969, 0.00002743, 0.00002967, -0.00000271, -0.00001758, 0.00000241, 0.00002097, 0.00001284, -0.00000438, -0.00000765, 0.00000194, 0.00001237, 0.00001486, 0.00000378, -0.00001480, -0.00001842, 0.00000509, 0.00003163, 0.00002624, -0.00000747, -0.00002817, -0.00001141, 0.00001899, 0.00002536, 0.00000501, -0.00001341, -0.00001324, -0.00000587, -0.00000225, 0.00000452, 0.00001565, 0.00001263, -0.00000985, -0.00002177, -0.00000040, 0.00002599, 0.00001703, -0.00001417, -0.00002118, 0.00000060, 0.00001129, -0.00000357, -0.00001412, -0.00000357, 0.00000639, 0.00000005, -0.00000703, -0.00000203, 0.00000379, 0.00000133, -0.00000115, 0.00000065, -0.00000052, -0.00000381, -0.00000015, 0.00000605, 0.00000186, -0.00000816, -0.00000765, 0.00000243, 0.00000645, 0.00000202, 0.00000009, 0.00000183, -0.00000065, -0.00000548, -0.00000461, -0.00000014, 0.00000134, 0.00000264, 0.00000785, 0.00001041, 0.00000433, -0.00000388, -0.00000699, -0.00000847, -0.00001103, -0.00000833, 0.00000168, 0.00000934, 0.00000844, 0.00000488, 0.00000355, 0.00000054, -0.00000624, -0.00001184, -0.00001326, -0.00001213, -0.00000795, -0.00000026, 0.00000465, 0.00000100, -0.00000575, -0.00000658, -0.00000244, -0.00000009, },
		{ -0.00000109, -0.00000117, -0.00000064, 0.00000028, -0.00000230, -0.00000769, -0.00000978, -0.00000752, -0.00000583, -0.00000580, -0.00000484, -0.00000419, -0.00000500, -0.00000312, 0.00000102, -0.00000157, -0.00001076, -0.00001238, -0.00000482, -0.00000741, -0.00002476, -0.00003113, -0.00001334, 0.00000351, 0.00000129, 0.00000272, 0.00002458, 0.00003642, 0.00001161, -0.00001681, -0.00000515, 0.00002590, 0.00002455, -0.00000570, -0.00001508, 0.00000835, 0.00002421, 0.00000950, -0.00001018, -0.00000764, 0.00000901, 0.00001764, 0.00001067, -0.00000503, -0.00001570, -0.00000701, 0.00001830, 0.00003251, 0.00001373, -0.00001921, -0.00002564, 0.00000246, 0.00002840, 0.00002033, -0.00000814, -0.00002148, -0.00001091, 0.00000219, 0.00000468, 0.00000602, 0.00001024, 0.00000378, -0.00001355, -0.00001516, 0.00000959, 0.00002694, 0.00000732, -0.00002246, -0.00001864, 0.00000832, 0.00001276, -0.00000909, -0.00001734, -0.00000008, 0.00001045, -0.00000132, -0.00001064, -0.00000150, 0.00000798, 0.00000324, -0.00000374, -0.00000248, -0.00000040, -0.00000144, 0.00000109, 0.00000477, -0.00000045, -0.00000900, -0.00000584, 0.00000539, 0.00000766, 0.00000057, -0.00000192, 0.00000120, -0.00000064, -0.00000619, -0.00000515, 0.00000120, 0.00000409, 0.00000475, 0.00000835, 0.00000958, 0.00000213, -0.00000750, -0.00001084, -0.00001053, -0.00001012, -0.00000472, 0.00000641, 0.00001275, 0.00000874, 0.00000261, 0.00000064, -0.00000222, -0.00000914, -0.00001379, -0.00001211, -0.00000799, -0.00000410, 0.00000036, 0.00000240, -0.00000153, -0.00000637, -0.00000494, -0.00000000, 0.00000109, -0.00000114, -0.00000179, },
		{ -0.00000092, 0.00000025, -0.00000176, -0.00000795, -0.00001159, -0.00000904, -0.00000514, -0.00000381, -0.00000375, -0.00000444, -0.00000566, -0.00000367, 0.00000066, -0.00000179, -0.00001144, -0.00001363, -0.00000492, -0.00000551, -0.00002421, -0.00003564, -0.00001968, 0.00000301, 0.00000743, 0.00000723, 0.00002234, 0.00003327, 0.00001507, -0.00000934, -0.00000256, 0.00002114, 0.00001914, -0.00000513, -0.00000939, 0.00001290, 0.00002293, 0.00000419, -0.00001265, -0.00000303, 0.00001602, 0.00001791, 0.00000285, -0.00001073, -0.00000945, 0.00000729, 0.00002606, 0.00002535, -0.00000020, -0.00002398, -0.00001575, 0.00001570, 0.00002996, 0.00000861, -0.00002058, -0.00002326, -0.00000323, 0.00001041, 0.00000749, 0.00000298, 0.00000352, -0.00000206, -0.00001251, -0.00000674, 0.00001627, 0.00002287, -0.00000345, -0.00002685, -0.00001225, 0.00001530, 0.00001071, -0.00001577, -0.00001814, 0.00000552, 0.00001337, -0.00000493, -0.00001408, 0.00000120, 0.00001288, 0.00000325, -0.00000816, -0.00000488, 0.00000188, 0.00000153, 0.00000073, 0.00000192, -0.00000228, -0.00000796, -0.00000296, 0.00000761, 0.00000737, -0.00000165, -0.00000367, 0.00000106, -0.00000048, -0.00000676, -0.00000508, 0.00000331, 0.00000694, 0.00000600, 0.00000769, 0.00000803, -0.00000034, -0.00001130, -0.00001452, -0.00001146, -0.00000728, 0.00000037, 0.00001101, 0.00001454, 0.00000713, -0.00000080, -0.00000245, -0.00000455, -0.00001119, -0.00001451, -0.00000968, -0.00000336, -0.00000116, -0.00000076, -0.00000094, -0.00000354, -0.00000551, -0.00000225, 0.00000233, 0.00000143, -0.00000257, -0.00000320, -0.00000135, -0.00000104, },
		{ -0.00000168, -0.00000831, -0.00001298, -0.00000991, -0.00000412, -0.00000210, -0.00000321, -0.00000486, -0.00000591, -0.00000395, -0.00000020, -0.00000270, -0.00001188, -0.00001386, -0.00000466, -0.00000456, -0.00002467, -0.00003972, -0.00002456, 0.00000311, 0.00001238, 0.00001011, 0.00002010, 0.00003208, 0.00002035, -0.00000257, -0.00000282, 0.00001471, 0.00001573, -0.00000085, -0.00000289, 0.00001392, 0.00001798, -0.00000011, -0.00001038, 0.00000411, 0.00001986, 0.00001313, -0.00000511, -0.00001041, 0.00000245, 0.00001989, 0.00002604, 0.00001299, -0.00001059, -0.00002044, -0.00000240, 0.00002342, 0.00002295, -0.00000590, -0.00002803, -0.00001817, 0.00000678, 0.00001552, 0.00000546, -0.00000258, -0.00000180, -0.00000388, -0.00000825, 0.00000083, 0.00001848, 0.00001533, -0.00001275, -0.00002664, -0.00000379, 0.00001968, 0.00000543, -0.00002183, -0.00001578, 0.00001207, 0.00001397, -0.00001032, -0.00001607, 0.00000615, 0.00001727, 0.00000062, -0.00001356, -0.00000528, 0.00000622, 0.00000385, -0.00000195, -0.00000172, -0.00000240, -0.00000492, -0.00000003, 0.00000819, 0.00000569, -0.00000391, -0.00000475, 0.00000130, -0.00000031, -0.00000704, -0.00000427, 0.00000591, 0.00000936, 0.00000616, 0.00000609, 0.00000604, -0.00000307, -0.00001525, -0.00001758, -0.00001057, -0.00000231, 0.00000619, 0.00001427, 0.00001404, 0.00000394, -0.00000448, -0.00000506, -0.00000624, -0.00001241, -0.00001405, -0.00000622, 0.00000117, 0.00000028, -0.00000363, -0.00000467, -0.00000424, -0.00000295, 0.00000095, 0.00000380, 0.00000057, -0.00000456, -0.00000440, -0.00000108, -0.00000051, -0.00000119, -0.00000016, },
		{ -0.00001383, -0.00001043, -0.00000316, -0.00000075, -0.00000295, -0.00000518, -0.00000591, -0.00000435, -0.00000161, -0.00000384, -0.00001169, -0.00001332, -0.00000474, -0.00000481, -0.00002546, -0.00004249, -0.00002809, 0.00000264, 0.00001534, 0.00001207, 0.00001932, 0.00003315, 0.00002595, 0.00000195, -0.00000541, 0.00000885, 0.00001559, 0.00000570, 0.00000182, 0.00001087, 0.00001173, -0.00000092, -0.00000394, 0.00001070, 0.00001882, 0.00000543, -0.00000953, -0.00000343, 0.00001635, 0.00002690, 0.00001877, 0.00000004, -0.00001405, -0.00001040, 0.00000925, 0.00002272, 0.00000973, -0.00001830, -0.00002809, -0.00000816, 0.00001537, 0.00001562, -0.00000017, -0.00000793, -0.00000398, -0.00000222, -0.00000303, 0.00000564, 0.00001651, 0.00000656, -0.00001876, -0.00002244, 0.00000459, 0.00002044, -0.00000181, -0.00002563, -0.00001048, 0.00001799, 0.00001164, -0.00001634, -0.00001554, 0.00001274, 0.00001975, -0.00000470, -0.00001851, -0.00000279, 0.00001174, 0.00000407, -0.00000696, -0.00000475, 0.00000012, -0.00000053, 0.00000160, 0.00000667, 0.00000332, -0.00000542, -0.00000506, 0.00000159, -0.00000016, -0.00000676, -0.00000273, 0.00000854, 0.00001090, 0.00000524, 0.00000387, 0.00000375, -0.00000610, -0.00001909, -0.00001929, -0.00000727, 0.00000440, 0.00001146, 0.00001507, 0.00001123, 0.00000008, -0.00000755, -0.00000693, -0.00000742, -0.00001281, -0.00001237, -0.00000208, 0.00000488, -0.00000019, -0.00000784, -0.00000780, -0.00000296, 0.00000100, 0.00000372, 0.00000370, -0.00000137, -0.00000644, -0.00000485, -0.00000028, 0.00000001, -0.00000184, -0.00000101, -0.00000174, -0.00000837, },
		{ -0.00000241, 0.00000039, -0.00000272, -0.00000543, -0.00000594, -0.00000502, -0.00000330, -0.00000479, -0.00001086, -0.00001254, -0.00000558, -0.00000594, -0.00002576, -0.00004364, -0.00003101, 0.00000060, 0.00001636, 0.00001431, 0.00002092, 0.00003573, 0.00003016, 0.00000374, -0.00000871, 0.00000547, 0.00001847, 0.00001216, 0.00000309, 0.00000502, 0.00000699, 0.00000266, 0.00000438, 0.00001400, 0.00001331, -0.00000176, -0.00000802, 0.00000834, 0.00002795, 0.00002661, 0.00000739, -0.00000897, -0.00000996, 0.00000185, 0.00001465, 0.00001372, -0.00000519, -0.00002456, -0.00002103, 0.00000313, 0.00001952, 0.00001088, -0.00000692, -0.00001083, -0.00000275, 0.00000127, 0.00000108, 0.00000699, 0.00001173, -0.00000124, -0.00002077, -0.00001577, 0.00001100, 0.00001767, -0.00000923, -0.00002618, -0.00000331, 0.00002181, 0.00000662, -0.00002150, -0.00001207, 0.00001965, 0.00001924, -0.00001192, -0.00002143, 0.00000268, 0.00001690, 0.00000118, -0.00001337, -0.00000555, 0.00000549, 0.00000381, 0.00000073, 0.00000335, 0.00000138, -0.00000564, -0.00000496, 0.00000159, 0.00000014, -0.00000568, -0.00000070, 0.00001063, 0.00001133, 0.00000351, 0.00000129, 0.00000113, -0.00000939, -0.00002222, -0.00001880, -0.00000149, 0.00001171, 0.00001468, 0.00001288, 0.00000687, -0.00000331, -0.00000951, -0.00000828, -0.00000828, -0.00001225, -0.00000946, 0.00000216, 0.00000697, -0.00000270, -0.00001246, -0.00000922, 0.00000049, 0.00000545, 0.00000498, 0.00000180, -0.00000371, -0.00000739, -0.00000438, 0.00000060, -0.00000000, -0.00000299, -0.00000199, -0.00000156, -0.00000799, -0.00001433, -0.00001090, },
		{ -0.00000251, -0.00000580, -0.00000618, -0.00000581, -0.00000488, -0.00000539, -0.00000974, -0.00001201, -0.00000717, -0.00000731, -0.00002505, -0.00004350, -0.00003423, -0.00000339, 0.00001617, 0.00001790, 0.00002483, 0.00003845, 0.00003191, 0.00000355, -0.00001076, 0.00000534, 0.00002275, 0.00001630, 0.00000089, -0.00000124, 0.00000578, 0.00000965, 0.00001166, 0.00001282, 0.00000566, -0.00000516, -0.00000046, 0.00002131, 0.00003395, 0.00002003, -0.00000371, -0.00001138, -0.00000077, 0.00001125, 0.00001167, -0.00000047, -0.00001684, -0.00002290, -0.00000958, 0.00001180, 0.00001798, 0.00000324, -0.00001210, -0.00001025, 0.00000079, 0.00000460, 0.00000279, 0.00000534, 0.00000600, -0.00000658, -0.00001923, -0.00000856, 0.00001434, 0.00001238, -0.00001512, -0.00002341, 0.00000415, 0.00002259, -0.00000007, -0.00002444, -0.00000602, 0.00002530, 0.00001535, -0.00001954, -0.00002106, 0.00001031, 0.00001990, -0.00000504, -0.00001949, -0.00000284, 0.00001290, 0.00000630, -0.00000321, -0.00000069, 0.00000103, -0.00000459, -0.00000511, 0.00000109, 0.00000098, -0.00000364, 0.00000135, 0.00001167, 0.00001069, 0.00000131, -0.00000156, -0.00000194, -0.00001261, -0.00002372, -0.00001552, 0.00000611, 0.00001802, 0.00001473, 0.00000806, 0.00000230, -0.00000528, -0.00001046, -0.00000956, -0.00000881, -0.00001042, -0.00000552, 0.00000560, 0.00000679, -0.00000672, -0.00001617, -0.00000814, 0.00000552, 0.00000903, 0.00000399, -0.00000140, -0.00000540, -0.00000690, -0.00000335, 0.00000085, -0.00000081, -0.00000431, -0.00000262, -0.00000098, -0.00000738, -0.00001476, -0.00001145, -0.00000177, 0.00000151, },
		{ -0.00000654, -0.00000642, -0.00000613, -0.00000582, -0.00000876, -0.00001190, -0.00000906, -0.00000836, -0.00002337, -0.00004279, -0.00003835, -0.00000900, 0.00001578, 0.00002324, 0.00003016, 0.00004004, 0.00003120, 0.00000296, -0.00001018, 0.00000783, 0.00002629, 0.00001702, -0.00000330, -0.00000538, 0.00000852, 0.00001771, 0.00001560, 0.00000793, -0.00000087, -0.00000293, 0.00001106, 0.00003161, 0.00003330, 0.00001040, -0.00001065, -0.00000739, 0.00000912, 0.00001393, 0.00000130, -0.00001500, -0.00002162, -0.00001440, 0.00000220, 0.00001516, 0.00001166, -0.00000444, -0.00001386, -0.00000666, 0.00000481, 0.00000623, 0.00000195, 0.00000196, 0.00000100, -0.00000890, -0.00001536, -0.00000249, 0.00001445, 0.00000614, -0.00001832, -0.00001811, 0.00001030, 0.00002023, -0.00000693, -0.00002435, 0.00000145, 0.00002829, 0.00000861, -0.00002578, -0.00001697, 0.00001843, 0.00001932, -0.00001370, -0.00002330, 0.00000382, 0.00002062, 0.00000531, -0.00000982, -0.00000383, 0.00000301, -0.00000293, -0.00000625, 0.00000028, 0.00000288, -0.00000077, 0.00000274, 0.00001133, 0.00000929, -0.00000102, -0.00000475, -0.00000543, -0.00001503, -0.00002264, -0.00000950, 0.00001407, 0.00002164, 0.00001128, 0.00000182, -0.00000108, -0.00000556, -0.00001100, -0.00001110, -0.00000865, -0.00000705, -0.00000120, 0.00000718, 0.00000417, -0.00001109, -0.00001761, -0.00000449, 0.00001080, 0.00001044, 0.00000075, -0.00000477, -0.00000546, -0.00000509, -0.00000264, -0.00000008, -0.00000216, -0.00000519, -0.00000262, -0.00000022, -0.00000688, -0.00001525, -0.00001192, -0.00000105, 0.00000257, -0.00000252, -0.00000647, },
		{ -0.00000715, -0.00000644, -0.00000817, -0.00001197, -0.00001076, -0.00000889, -0.00002117, -0.00004216, -0.00004336, -0.00001543, 0.00001581, 0.00002988, 0.00003571, 0.00003997, 0.00002902, 0.00000352, -0.00000675, 0.00001130, 0.00002737, 0.00001471, -0.00000716, -0.00000590, 0.00001398, 0.00002418, 0.00001547, 0.00000174, -0.00000349, 0.00000461, 0.00002312, 0.00003664, 0.00002740, 0.00000166, -0.00001168, 0.00000013, 0.00001508, 0.00000867, -0.00001278, -0.00002513, -0.00001851, -0.00000238, 0.00001046, 0.00001260, 0.00000308, -0.00000954, -0.00001184, -0.00000163, 0.00000757, 0.00000556, -0.00000070, -0.00000170, -0.00000221, -0.00000857, -0.00001067, 0.00000141, 0.00001206, 0.00000052, -0.00001847, -0.00001165, 0.00001399, 0.00001544, -0.00001248, -0.00002124, 0.00000877, 0.00002786, 0.00000035, -0.00002906, -0.00000974, 0.00002505, 0.00001462, -0.00002299, -0.00002307, 0.00001365, 0.00002636, -0.00000002, -0.00001763, -0.00000437, 0.00000729, -0.00000183, -0.00000884, -0.00000020, 0.00000633, 0.00000243, 0.00000273, 0.00000958, 0.00000760, -0.00000331, -0.00000843, -0.00000901, -0.00001564, -0.00001838, -0.00000165, 0.00002047, 0.00002143, 0.00000506, -0.00000412, -0.00000238, -0.00000462, -0.00001188, -0.00001282, -0.00000718, -0.00000224, 0.00000239, 0.00000604, -0.00000025, -0.00001413, -0.00001589, 0.00000081, 0.00001458, 0.00000896, -0.00000381, -0.00000691, -0.00000353, -0.00000278, -0.00000315, -0.00000221, -0.00000337, -0.00000509, -0.00000214, 0.00000022, -0.00000671, -0.00001556, -0.00001202, -0.00000028, 0.00000332, -0.00000294, -0.00000736, -0.00000673, -0.00000662, },
		{ -0.00000786, -0.00001181, -0.00001196, -0.00000914, -0.00001907, -0.00004185, -0.00004872, -0.00002193, 0.00001625, 0.00003684, 0.00004062, 0.00003866, 0.00002681, 0.00000605, -0.00000142, 0.00001391, 0.00002544, 0.00001100, -0.00000864, -0.00000279, 0.00001990, 0.00002728, 0.00001234, -0.00000281, -0.00000091, 0.00001525, 0.00003244, 0.00003595, 0.00001937, -0.00000317, -0.00000776, 0.00000687, 0.00001414, -0.00000271, -0.00002565, -0.00002789, -0.00000930, 0.00000890, 0.00001295, 0.00000565, -0.00000469, -0.00001064, -0.00000711, 0.00000286, 0.00000802, 0.00000295, -0.00000397, -0.00000449, -0.00000339, -0.00000651, -0.00000649, 0.00000288, 0.00000834, -0.00000334, -0.00001603, -0.00000551, 0.00001480, 0.00000950, -0.00001563, -0.00001590, 0.00001441, 0.00002414, -0.00000765, -0.00002856, -0.00000085, 0.00002840, 0.00000641, -0.00003061, -0.00001800, 0.00002474, 0.00002798, -0.00000932, -0.00002448, -0.00000118, 0.00001287, -0.00000259, -0.00001276, 0.00000062, 0.00001145, 0.00000507, 0.00000070, 0.00000678, 0.00000619, -0.00000558, -0.00001268, -0.00001196, -0.00001339, -0.00001103, 0.00000632, 0.00002346, 0.00001730, -0.00000225, -0.00000817, -0.00000159, -0.00000348, -0.00001355, -0.00001406, -0.00000390, 0.00000325, 0.00000394, 0.00000200, -0.00000499, -0.00001425, -0.00001114, 0.00000597, 0.00001538, 0.00000492, -0.00000806, -0.00000677, -0.00000012, -0.00000121, -0.00000535, -0.00000491, -0.00000358, -0.00000393, -0.00000176, -0.00000009, -0.00000672, -0.00001527, -0.00001165, 0.00000023, 0.00000341, -0.00000375, -0.00000811, -0.00000649, -0.00000646, -0.00000823, -0.00000747, },
		{ -0.00001280, -0.00000961, -0.00001741, -0.00004162, -0.00005377, -0.00002821, 0.00001646, 0.00004317, 0.00004478, 0.00003721, 0.00002573, 0.00001031, 0.00000424, 0.00001451, 0.00002129, 0.00000794, -0.00000675, 0.00000254, 0.00002406, 0.00002671, 0.00000850, -0.00000349, 0.00000622, 0.00002598, 0.00003714, 0.00003118, 0.00001248, -0.00000332, -0.00000205, 0.00000893, 0.00000613, -0.00001610, -0.00003307, -0.00002311, 0.00000226, 0.00001587, 0.00000970, -0.00000279, -0.00000922, -0.00000794, -0.00000163, 0.00000528, 0.00000613, -0.00000057, -0.00000674, -0.00000586, -0.00000290, -0.00000381, -0.00000361, 0.00000236, 0.00000456, -0.00000493, -0.00001201, -0.00000089, 0.00001306, 0.00000381, -0.00001604, -0.00000966, 0.00001735, 0.00001809, -0.00001372, -0.00002446, 0.00000768, 0.00002753, -0.00000362, -0.00003445, -0.00000860, 0.00003448, 0.00002421, -0.00002095, -0.00002806, 0.00000581, 0.00001802, -0.00000615, -0.00001711, 0.00000374, 0.00001782, 0.00000602, -0.00000359, 0.00000363, 0.00000550, -0.00000801, -0.00001733, -0.00001327, -0.00000760, -0.00000165, 0.00001220, 0.00002189, 0.00001036, -0.00000857, -0.00000950, 0.00000042, -0.00000316, -0.00001573, -0.00001389, 0.00000115, 0.00000808, 0.00000241, -0.00000409, -0.00000805, -0.00001056, -0.00000468, 0.00000889, 0.00001260, -0.00000023, -0.00001030, -0.00000423, 0.00000346, -0.00000143, -0.00000893, -0.00000708, -0.00000231, -0.00000223, -0.00000222, -0.00000121, -0.00000635, -0.00001401, -0.00001109, -0.00000005, 0.00000274, -0.00000455, -0.00000832, -0.00000582, -0.00000631, -0.00000966, -0.00000884, -0.00000746, -0.00001114, },
		{ -0.00001607, -0.00004092, -0.00005811, -0.00003459, 0.00001560, 0.00004839, 0.00004882, 0.00003680, 0.00002613, 0.00001519, 0.00000877, 0.00001306, 0.00001658, 0.00000711, -0.00000191, 0.00000805, 0.00002519, 0.00002373, 0.00000642, 0.00000046, 0.00001582, 0.00003422, 0.00003727, 0.00002506, 0.00000874, -0.00000039, 0.00000152, 0.00000445, -0.00000633, -0.00002679, -0.00003293, -0.00001325, 0.00001198, 0.00001688, 0.00000274, -0.00000964, -0.00000955, -0.00000298, 0.00000261, 0.00000504, 0.00000272, -0.00000386, -0.00000832, -0.00000586, -0.00000144, -0.00000137, -0.00000227, 0.00000065, 0.00000168, -0.00000443, -0.00000762, 0.00000163, 0.00000969, -0.00000052, -0.00001407, -0.00000392, 0.00001731, 0.00001121, -0.00001675, -0.00001788, 0.00001397, 0.00002266, -0.00001316, -0.00003330, 0.00000329, 0.00004028, 0.00001513, -0.00003230, -0.00002671, 0.00001546, 0.00002076, -0.00001261, -0.00002043, 0.00000979, 0.00002434, 0.00000415, -0.00000985, 0.00000110, 0.00000575, -0.00001086, -0.00002188, -0.00001186, 0.00000169, 0.00000786, 0.00001394, 0.00001583, 0.00000272, -0.00001212, -0.00000840, 0.00000219, -0.00000415, -0.00001748, -0.00001159, 0.00000705, 0.00001064, -0.00000234, -0.00001043, -0.00000762, -0.00000342, 0.00000125, 0.00000796, 0.00000698, -0.00000441, -0.00000951, -0.00000029, 0.00000561, -0.00000374, -0.00001279, -0.00000771, 0.00000021, -0.00000096, -0.00000393, -0.00000261, -0.00000494, -0.00001185, -0.00001096, -0.00000151, 0.00000160, -0.00000476, -0.00000785, -0.00000512, -0.00000658, -0.00001140, -0.00001014, -0.00000665, -0.00001006, -0.00001370, -0.00001060, },
		{ -0.00006187, -0.00004169, 0.00001314, 0.00005276, 0.00005370, 0.00003811, 0.00002746, 0.00001935, 0.00001154, 0.00001058, 0.00001301, 0.00000897, 0.00000447, 0.00001203, 0.00002344, 0.00002035, 0.00000773, 0.00000822, 0.00002543, 0.00003879, 0.00003447, 0.00001999, 0.00000800, 0.00000240, 0.00000005, -0.00000566, -0.00001894, -0.00003127, -0.00002596, -0.00000229, 0.00001678, 0.00001257, -0.00000484, -0.00001281, -0.00000636, 0.00000209, 0.00000435, 0.00000257, -0.00000100, -0.00000609, -0.00000862, -0.00000492, 0.00000028, 0.00000031, -0.00000224, -0.00000136, 0.00000021, -0.00000248, -0.00000389, 0.00000206, 0.00000579, -0.00000288, -0.00001066, 0.00000021, 0.00001483, 0.00000506, -0.00001653, -0.00001055, 0.00001680, 0.00001511, -0.00001995, -0.00002731, 0.00001500, 0.00004041, 0.00000234, -0.00004053, -0.00002000, 0.00002558, 0.00001950, -0.00002099, -0.00002101, 0.00001860, 0.00002940, -0.00000125, -0.00001716, 0.00000014, 0.00000685, -0.00001431, -0.00002553, -0.00000691, 0.00001351, 0.00001512, 0.00001030, 0.00000668, -0.00000314, -0.00001209, -0.00000618, 0.00000240, -0.00000601, -0.00001751, -0.00000717, 0.00001218, 0.00000975, -0.00000924, -0.00001468, -0.00000287, 0.00000539, 0.00000422, 0.00000280, 0.00000050, -0.00000568, -0.00000590, 0.00000325, 0.00000523, -0.00000739, -0.00001545, -0.00000647, 0.00000302, -0.00000103, -0.00000666, -0.00000339, -0.00000214, -0.00000939, -0.00001201, -0.00000409, 0.00000071, -0.00000393, -0.00000698, -0.00000498, -0.00000741, -0.00001299, -0.00001089, -0.00000548, -0.00000905, -0.00001503, -0.00001204, -0.00001457, -0.00003939, },
		{ 0.00000916, 0.00005696, 0.00006009, 0.00004098, 0.00002870, 0.00002190, 0.00001294, 0.00000854, 0.00001161, 0.00001286, 0.00001068, 0.00001375, 0.00002004, 0.00001849, 0.00001274, 0.00001802, 0.00003329, 0.00004010, 0.00003078, 0.00001683, 0.00000815, 0.00000185, -0.00000711, -0.00001815, -0.00002745, -0.00002846, -0.00001520, 0.00000586, 0.00001572, 0.00000537, -0.00001017, -0.00001184, -0.00000150, 0.00000542, 0.00000339, -0.00000100, -0.00000397, -0.00000697, -0.00000794, -0.00000358, 0.00000173, 0.00000108, -0.00000300, -0.00000297, 0.00000013, 0.00000006, -0.00000143, 0.00000094, 0.00000238, -0.00000333, -0.00000693, 0.00000223, 0.00001097, 0.00000078, -0.00001374, -0.00000422, 0.00001600, 0.00000694, -0.00002253, -0.00001795, 0.00002377, 0.00003458, -0.00001142, -0.00004344, -0.00000899, 0.00003357, 0.00001370, -0.00002942, -0.00001758, 0.00002905, 0.00003121, -0.00001018, -0.00002408, 0.00000149, 0.00000839, -0.00001836, -0.00002733, 0.00000177, 0.00002601, 0.00001787, 0.00000132, -0.00000307, -0.00000516, -0.00000901, -0.00000478, 0.00000046, -0.00000743, -0.00001481, -0.00000169, 0.00001461, 0.00000528, -0.00001611, -0.00001485, 0.00000553, 0.00001316, 0.00000259, -0.00000536, -0.00000430, -0.00000311, -0.00000099, 0.00000449, 0.00000227, -0.00001079, -0.00001583, -0.00000396, 0.00000479, -0.00000274, -0.00000945, -0.00000269, 0.00000172, -0.00000762, -0.00001459, -0.00000715, 0.00000084, -0.00000209, -0.00000643, -0.00000585, -0.00000844, -0.00001378, -0.00001093, -0.00000445, -0.00000866, -0.00001676, -0.00001339, -0.00001251, -0.00003715, -0.00006563, -0.00004999, },
		{ 0.00006798, 0.00004468, 0.00002901, 0.00002280, 0.00001398, 0.00000813, 0.00001243, 0.00001753, 0.00001558, 0.00001358, 0.00001663, 0.00001919, 0.00002064, 0.00002816, 0.00003888, 0.00003946, 0.00002748, 0.00001458, 0.00000632, -0.00000373, -0.00001801, -0.00002878, -0.00002931, -0.00001997, -0.00000467, 0.00000880, 0.00001004, -0.00000163, -0.00001163, -0.00000783, 0.00000287, 0.00000605, 0.00000052, -0.00000441, -0.00000562, -0.00000673, -0.00000679, -0.00000228, 0.00000268, 0.00000108, -0.00000402, -0.00000379, 0.00000110, 0.00000241, -0.00000039, -0.00000095, 0.00000007, -0.00000238, -0.00000382, 0.00000230, 0.00000695, -0.00000118, -0.00000965, -0.00000014, 0.00001246, 0.00000024, -0.00002065, -0.00000767, 0.00002755, 0.00002406, -0.00002304, -0.00004010, 0.00000397, 0.00003718, 0.00000415, -0.00003557, -0.00000974, 0.00003917, 0.00002837, -0.00002170, -0.00002897, 0.00000540, 0.00000975, -0.00002272, -0.00002640, 0.00001357, 0.00003683, 0.00001461, -0.00001152, -0.00001054, -0.00000240, -0.00000461, -0.00000592, -0.00000304, -0.00000668, -0.00000929, 0.00000287, 0.00001294, -0.00000143, -0.00002025, -0.00001017, 0.00001530, 0.00001716, -0.00000373, -0.00001397, -0.00000532, 0.00000264, 0.00000291, 0.00000235, -0.00000201, -0.00001214, -0.00001379, -0.00000165, 0.00000447, -0.00000548, -0.00001104, -0.00000028, 0.00000561, -0.00000743, -0.00001838, -0.00000963, 0.00000241, 0.00000010, -0.00000699, -0.00000765, -0.00000888, -0.00001325, -0.00001067, -0.00000436, -0.00000909, -0.00001831, -0.00001399, -0.00000990, -0.00003486, -0.00006996, -0.00005944, 0.00000431, 0.00006161, },
		{ 0.00002820, 0.00002275, 0.00001574, 0.00000977, 0.00001474, 0.00002182, 0.00001887, 0.00001255, 0.00001445, 0.00002248, 0.00003023, 0.00003770, 0.00004270, 0.00003798, 0.00002426, 0.00001094, 0.00000031, -0.00001366, -0.00002899, -0.00003403, -0.00002452, -0.00000923, 0.00000224, 0.00000634, 0.00000243, -0.00000592, -0.00000929, -0.00000281, 0.00000515, 0.00000418, -0.00000298, -0.00000673, -0.00000598, -0.00000584, -0.00000555, -0.00000123, 0.00000308, 0.00000059, -0.00000484, -0.00000374, 0.00000257, 0.00000405, -0.00000054, -0.00000285, -0.00000097, -0.00000079, -0.00000190, 0.00000104, 0.00000377, -0.00000112, -0.00000565, 0.00000131, 0.00000774, -0.00000351, -0.00001537, 0.00000091, 0.00002570, 0.00001143, -0.00002990, -0.00003128, 0.00001589, 0.00003527, -0.00000714, -0.00003741, 0.00000161, 0.00004654, 0.00002032, -0.00003400, -0.00003031, 0.00001151, 0.00001023, -0.00002687, -0.00002221, 0.00002706, 0.00004358, 0.00000511, -0.00002555, -0.00001326, 0.00000456, -0.00000132, -0.00001044, -0.00000645, -0.00000235, -0.00000214, 0.00000420, 0.00000703, -0.00000778, -0.00001943, -0.00000168, 0.00002323, 0.00001575, -0.00001295, -0.00002007, -0.00000185, 0.00000947, 0.00000368, -0.00000283, -0.00000552, -0.00001035, -0.00001044, -0.00000121, 0.00000197, -0.00000783, -0.00001042, 0.00000316, 0.00000818, -0.00000910, -0.00002229, -0.00001060, 0.00000507, 0.00000158, -0.00000904, -0.00000963, -0.00000786, -0.00001152, -0.00001102, -0.00000576, -0.00000989, -0.00001878, -0.00001360, -0.00000743, -0.00003334, -0.00007494, -0.00006931, -0.00000068, 0.00006675, 0.00007666, 0.00004839, },
		{ 0.00001875, 0.00001312, 0.00001767, 0.00002518, 0.00002099, 0.00001171, 0.00001405, 0.00002786, 0.00004069, 0.00004663, 0.00004543, 0.00003561, 0.00001939, 0.00000368, -0.00001010, -0.00002509, -0.00003617, -0.00003251, -0.00001547, -0.00000008, 0.00000389, 0.00000025, -0.00000416, -0.00000644, -0.00000459, 0.00000122, 0.00000477, 0.00000084, -0.00000590, -0.00000768, -0.00000547, -0.00000480, -0.00000443, -0.00000049, 0.00000301, -0.00000010, -0.00000521, -0.00000298, 0.00000401, 0.00000475, -0.00000138, -0.00000421, -0.00000095, 0.00000072, -0.00000124, -0.00000070, 0.00000190, 0.00000016, -0.00000280, 0.00000061, 0.00000358, -0.00000386, -0.00000869, 0.00000572, 0.00001911, -0.00000026, -0.00003071, -0.00001918, 0.00002405, 0.00002824, -0.00001746, -0.00003391, 0.00001442, 0.00004900, 0.00000773, -0.00004476, -0.00002722, 0.00001886, 0.00000930, -0.00003009, -0.00001482, 0.00004020, 0.00004449, -0.00000941, -0.00003762, -0.00000994, 0.00001375, -0.00000142, -0.00001780, -0.00000754, 0.00000590, 0.00000436, 0.00000055, -0.00000161, -0.00001065, -0.00001291, 0.00000782, 0.00002625, 0.00000921, -0.00002189, -0.00002143, 0.00000492, 0.00001457, 0.00000036, -0.00000924, -0.00000620, -0.00000562, -0.00000774, -0.00000370, -0.00000165, -0.00000815, -0.00000756, 0.00000614, 0.00000841, -0.00001192, -0.00002495, -0.00000978, 0.00000774, 0.00000142, -0.00001216, -0.00001064, -0.00000498, -0.00000944, -0.00001298, -0.00000859, -0.00001003, -0.00001748, -0.00001269, -0.00000618, -0.00003304, -0.00007989, -0.00007859, -0.00000557, 0.00007163, 0.00008521, 0.00005186, 0.00002687, 0.00002272, },
		{ 0.00002073, 0.00002773, 0.00002254, 0.00001157, 0.00001541, 0.00003494, 0.00005192, 0.00005521, 0.00004691, 0.00003088, 0.00001071, -0.00000803, -0.00002287, -0.00003420, -0.00003710, -0.00002534, -0.00000581, 0.00000478, 0.00000084, -0.00000668, -0.00000774, -0.00000374, 0.00000043, 0.00000299, 0.00000229, -0.00000264, -0.00000756, -0.00000750, -0.00000466, -0.00000385, -0.00000348, -0.00000004, 0.00000260, -0.00000073, -0.00000501, -0.00000183, 0.00000501, 0.00000460, -0.00000240, -0.00000478, -0.00000032, 0.00000169, -0.00000153, -0.00000223, 0.00000130, 0.00000169, -0.00000152, -0.00000111, 0.00000122, -0.00000152, -0.00000287, 0.00000595, 0.00000995, -0.00000832, -0.00002580, -0.00000680, 0.00002679, 0.00001791, -0.00002422, -0.00002546, 0.00002589, 0.00004528, -0.00000751, -0.00005164, -0.00001973, 0.00002606, 0.00000673, -0.00003166, -0.00000491, 0.00005080, 0.00003884, -0.00002649, -0.00004484, -0.00000088, 0.00002235, -0.00000622, -0.00002618, -0.00000438, 0.00001700, 0.00000743, -0.00000844, -0.00000999, -0.00000745, -0.00000203, 0.00001453, 0.00002269, -0.00000004, -0.00002711, -0.00001766, 0.00001222, 0.00001562, -0.00000621, -0.00001425, -0.00000304, 0.00000037, -0.00000766, -0.00000892, -0.00000446, -0.00000541, -0.00000357, 0.00000696, 0.00000623, -0.00001435, -0.00002529, -0.00000789, 0.00000901, -0.00000059, -0.00001512, -0.00000966, -0.00000077, -0.00000835, -0.00001701, -0.00001186, -0.00000834, -0.00001445, -0.00001247, -0.00000709, -0.00003356, -0.00008357, -0.00008661, -0.00001094, 0.00007511, 0.00009309, 0.00005561, 0.00002602, 0.00002326, 0.00002280, 0.00001754, },
		{ 0.00002378, 0.00001224, 0.00001860, 0.00004396, 0.00006415, 0.00006312, 0.00004563, 0.00002158, -0.00000299, -0.00002278, -0.00003441, -0.00003783, -0.00003170, -0.00001557, 0.00000111, 0.00000460, -0.00000475, -0.00001187, -0.00000783, 0.00000050, 0.00000399, 0.00000237, -0.00000113, -0.00000525, -0.00000791, -0.00000676, -0.00000391, -0.00000303, -0.00000263, 0.00000014, 0.00000197, -0.00000112, -0.00000429, -0.00000064, 0.00000536, 0.00000386, -0.00000315, -0.00000461, 0.00000045, 0.00000194, -0.00000228, -0.00000314, 0.00000149, 0.00000272, -0.00000153, -0.00000263, 0.00000110, 0.00000197, 0.00000030, 0.00000219, 0.00000097, -0.00001126, -0.00001705, 0.00000297, 0.00002404, 0.00000705, -0.00002580, -0.00001390, 0.00003327, 0.00003554, -0.00002265, -0.00005295, -0.00000884, 0.00003165, 0.00000275, -0.00003103, 0.00000624, 0.00005696, 0.00002728, -0.00004299, -0.00004539, 0.00001208, 0.00002752, -0.00001557, -0.00003297, 0.00000395, 0.00002851, 0.00000482, -0.00002128, -0.00001449, 0.00000272, 0.00000990, 0.00001493, 0.00001320, -0.00000810, -0.00002626, -0.00001057, 0.00001667, 0.00001189, -0.00001352, -0.00001556, 0.00000326, 0.00000512, -0.00001120, -0.00001523, -0.00000449, 0.00000013, -0.00000046, 0.00000455, 0.00000278, -0.00001467, -0.00002326, -0.00000651, 0.00000784, -0.00000358, -0.00001637, -0.00000652, 0.00000339, -0.00000940, -0.00002256, -0.00001405, -0.00000423, -0.00001077, -0.00001435, -0.00001027, -0.00003363, -0.00008481, -0.00009361, -0.00001809, 0.00007623, 0.00010063, 0.00006073, 0.00002638, 0.00002424, 0.00002730, 0.00002266, 0.00002401, 0.00002978, },
		{ 0.00002433, 0.00005558, 0.00007710, 0.00006862, 0.00003906, 0.00000608, -0.00002084, -0.00003722, -0.00004107, -0.00003482, -0.00002218, -0.00000678, 0.00000345, 0.00000064, -0.00001018, -0.00001392, -0.00000541, 0.00000435, 0.00000528, 0.00000015, -0.00000425, -0.00000655, -0.00000741, -0.00000598, -0.00000335, -0.00000224, -0.00000188, 0.00000003, 0.00000126, -0.00000114, -0.00000323, 0.00000028, 0.00000504, 0.00000289, -0.00000340, -0.00000392, 0.00000099, 0.00000158, -0.00000306, -0.00000339, 0.00000190, 0.00000295, -0.00000209, -0.00000302, 0.00000269, 0.00000485, 0.00000007, -0.00000379, -0.00000538, -0.00000914, -0.00000720, 0.00000819, 0.00001725, -0.00000155, -0.00002204, -0.00000204, 0.00003468, 0.00002143, -0.00003470, -0.00004812, 0.00000360, 0.00003446, -0.00000202, -0.00002804, 0.00001698, 0.00005752, 0.00001163, -0.00005584, -0.00003899, 0.00002604, 0.00002732, -0.00002782, -0.00003563, 0.00001678, 0.00003732, -0.00000433, -0.00003483, -0.00001204, 0.00001845, 0.00001846, 0.00000718, 0.00000080, -0.00001092, -0.00001922, -0.00000374, 0.00001570, 0.00000484, -0.00001834, -0.00001234, 0.00001035, 0.00000636, -0.00001769, -0.00002014, -0.00000076, 0.00000680, -0.00000024, -0.00000084, 0.00000012, -0.00001178, -0.00002005, -0.00000741, 0.00000428, -0.00000581, -0.00001481, -0.00000221, 0.00000570, -0.00001286, -0.00002806, -0.00001378, 0.00000176, -0.00000820, -0.00001917, -0.00001467, -0.00003163, -0.00008330, -0.00010082, -0.00002839, 0.00007496, 0.00010895, 0.00006824, 0.00002793, 0.00002497, 0.00003185, 0.00002872, 0.00002787, 0.00003133, 0.00002450, 0.00001387, },
		{ 0.00008901, 0.00006862, 0.00002484, -0.00001535, -0.00003969, -0.00004730, -0.00004076, -0.00002651, -0.00001206, -0.00000167, 0.00000141, -0.00000469, -0.00001340, -0.00001296, -0.00000221, 0.00000645, 0.00000453, -0.00000246, -0.00000628, -0.00000675, -0.00000664, -0.00000539, -0.00000286, -0.00000143, -0.00000132, -0.00000034, 0.00000064, -0.00000076, -0.00000205, 0.00000072, 0.00000419, 0.00000203, -0.00000307, -0.00000302, 0.00000112, 0.00000085, -0.00000357, -0.00000321, 0.00000207, 0.00000256, -0.00000239, -0.00000204, 0.00000484, 0.00000573, -0.00000309, -0.00000971, -0.00000769, -0.00000347, 0.00000101, 0.00000845, 0.00000895, -0.00000592, -0.00001436, 0.00000712, 0.00002968, 0.00000578, -0.00004125, -0.00003791, 0.00001538, 0.00003395, -0.00000670, -0.00002297, 0.00002563, 0.00005236, -0.00000544, -0.00006273, -0.00002691, 0.00003785, 0.00002119, -0.00004030, -0.00003254, 0.00003193, 0.00004063, -0.00001893, -0.00004508, -0.00000128, 0.00003612, 0.00001955, -0.00000798, -0.00000984, -0.00000585, -0.00000847, -0.00000127, 0.00000882, -0.00000220, -0.00001807, -0.00000589, 0.00001506, 0.00000325, -0.00002480, -0.00002137, 0.00000602, 0.00001202, -0.00000394, -0.00000751, 0.00000031, -0.00000596, -0.00001779, -0.00001163, -0.00000028, -0.00000534, -0.00001055, 0.00000124, 0.00000487, -0.00001760, -0.00003143, -0.00001068, 0.00000784, -0.00000843, -0.00002650, -0.00001836, -0.00002645, -0.00007998, -0.00011004, -0.00004244, 0.00007234, 0.00011939, 0.00007835, 0.00002987, 0.00002487, 0.00003678, 0.00003632, 0.00003229, 0.00003177, 0.00002453, 0.00001741, 0.00003395, 0.00007008, },
		{ 0.00000230, -0.00003986, -0.00005503, -0.00004993, -0.00003384, -0.00001611, -0.00000460, -0.00000101, -0.00000315, -0.00000901, -0.00001373, -0.00001029, 0.00000010, 0.00000645, 0.00000271, -0.00000441, -0.00000708, -0.00000641, -0.00000601, -0.00000496, -0.00000228, -0.00000063, -0.00000102, -0.00000087, 0.00000027, -0.00000005, -0.00000103, 0.00000061, 0.00000306, 0.00000152, -0.00000228, -0.00000217, 0.00000083, 0.00000001, -0.00000377, -0.00000293, 0.00000186, 0.00000204, -0.00000187, -0.00000014, 0.00000622, 0.00000410, -0.00000775, -0.00001352, -0.00000593, 0.00000341, 0.00000568, 0.00000484, 0.00000182, -0.00000549, -0.00000532, 0.00001133, 0.00001952, -0.00000815, -0.00004108, -0.00002425, 0.00002444, 0.00003031, -0.00001037, -0.00001659, 0.00003085, 0.00004243, -0.00002105, -0.00006267, -0.00001167, 0.00004496, 0.00001016, -0.00005013, -0.00002359, 0.00004631, 0.00003688, -0.00003610, -0.00004846, 0.00001670, 0.00005088, 0.00001084, -0.00002710, -0.00001392, 0.00000714, 0.00000155, -0.00000613, -0.00000188, -0.00000520, -0.00001203, 0.00000058, 0.00001477, -0.00000296, -0.00002935, -0.00001808, 0.00001345, 0.00001345, -0.00001086, -0.00001280, 0.00000437, 0.00000098, -0.00001858, -0.00001877, -0.00000350, -0.00000104, -0.00000518, 0.00000162, 0.00000098, -0.00002135, -0.00003113, -0.00000592, 0.00001162, -0.00001209, -0.00003447, -0.00001933, -0.00001834, -0.00007692, -0.00012265, -0.00005948, 0.00007011, 0.00013262, 0.00009019, 0.00003121, 0.00002415, 0.00004310, 0.00004582, 0.00003642, 0.00003024, 0.00002452, 0.00002482, 0.00004877, 0.00008617, 0.00009630, 0.00005968, },
		{ -0.00006266, -0.00004443, -0.00002311, -0.00000731, -0.00000156, -0.00000369, -0.00000777, -0.00001094, -0.00001184, -0.00000757, 0.00000067, 0.00000491, 0.00000094, -0.00000523, -0.00000704, -0.00000603, -0.00000561, -0.00000447, -0.00000157, -0.00000000, -0.00000105, -0.00000140, 0.00000025, 0.00000081, -0.00000038, 0.00000000, 0.00000196, 0.00000149, -0.00000122, -0.00000158, 0.00000022, -0.00000075, -0.00000375, -0.00000277, 0.00000142, 0.00000190, -0.00000049, 0.00000177, 0.00000587, 0.00000043, -0.00001203, -0.00001406, -0.00000134, 0.00000910, 0.00000627, -0.00000050, -0.00000214, -0.00000128, 0.00000222, 0.00000975, 0.00000677, -0.00001749, -0.00003451, -0.00000978, 0.00002940, 0.00002444, -0.00001236, -0.00000996, 0.00003191, 0.00002955, -0.00003272, -0.00005611, 0.00000369, 0.00004603, -0.00000351, -0.00005503, -0.00001025, 0.00005679, 0.00002638, -0.00005187, -0.00004296, 0.00003846, 0.00005801, -0.00000719, -0.00004489, -0.00000819, 0.00002514, 0.00000584, -0.00001884, -0.00001214, -0.00000105, -0.00000218, 0.00000305, 0.00000878, -0.00000902, -0.00002866, -0.00001155, 0.00001826, 0.00001019, -0.00001853, -0.00001431, 0.00001154, 0.00000631, -0.00002339, -0.00002678, -0.00000309, 0.00000657, -0.00000136, -0.00000238, -0.00000422, -0.00002156, -0.00002714, -0.00000206, 0.00001119, -0.00001806, -0.00004038, -0.00001666, -0.00000922, -0.00007643, -0.00013864, -0.00007749, 0.00006972, 0.00014799, 0.00010220, 0.00003169, 0.00002412, 0.00005185, 0.00005651, 0.00003868, 0.00002658, 0.00002660, 0.00003862, 0.00006870, 0.00010028, 0.00009428, 0.00003964, -0.00002639, -0.00006249, },
		{ -0.00001257, -0.00000277, -0.00000259, -0.00000739, -0.00001057, -0.00001052, -0.00000924, -0.00000602, -0.00000040, 0.00000285, -0.00000005, -0.00000506, -0.00000671, -0.00000588, -0.00000527, -0.00000377, -0.00000085, 0.00000029, -0.00000132, -0.00000171, 0.00000057, 0.00000157, -0.00000023, -0.00000089, 0.00000113, 0.00000189, -0.00000014, -0.00000132, -0.00000056, -0.00000136, -0.00000365, -0.00000276, 0.00000108, 0.00000241, 0.00000132, 0.00000271, 0.00000355, -0.00000403, -0.00001431, -0.00001139, 0.00000408, 0.00001193, 0.00000365, -0.00000518, -0.00000222, 0.00000455, 0.00000595, 0.00000320, -0.00000536, -0.00002057, -0.00002329, 0.00000286, 0.00002989, 0.00001767, -0.00001244, -0.00000430, 0.00002887, 0.00001600, -0.00003889, -0.00004476, 0.00001637, 0.00004119, -0.00001702, -0.00005399, 0.00000489, 0.00006118, 0.00001126, -0.00006239, -0.00002894, 0.00005908, 0.00005454, -0.00003115, -0.00005583, 0.00000770, 0.00004298, 0.00000072, -0.00003685, -0.00001695, 0.00001098, 0.00000731, -0.00000168, -0.00000102, -0.00001086, -0.00002199, -0.00000511, 0.00001767, 0.00000355, -0.00002357, -0.00001116, 0.00001925, 0.00000757, -0.00003130, -0.00003266, 0.00000198, 0.00001516, -0.00000170, -0.00001023, -0.00000771, -0.00001665, -0.00002136, -0.00000208, 0.00000628, -0.00002355, -0.00004191, -0.00001134, -0.00000205, -0.00007979, -0.00015623, -0.00009405, 0.00007131, 0.00016350, 0.00011315, 0.00003242, 0.00002668, 0.00006309, 0.00006640, 0.00003768, 0.00002236, 0.00003425, 0.00006045, 0.00009099, 0.00010679, 0.00007884, 0.00000932, -0.00005612, -0.00007775, -0.00006050, -0.00003293, },
		{ -0.00000568, -0.00000987, -0.00001099, -0.00000892, -0.00000732, -0.00000596, -0.00000226, 0.00000119, -0.00000012, -0.00000439, -0.00000649, -0.00000588, -0.00000475, -0.00000290, -0.00000034, 0.00000018, -0.00000162, -0.00000167, 0.00000107, 0.00000201, -0.00000054, -0.00000178, 0.00000073, 0.00000258, 0.00000076, -0.00000139, -0.00000138, -0.00000183, -0.00000356, -0.00000278, 0.00000113, 0.00000347, 0.00000279, 0.00000207, -0.00000019, -0.00000775, -0.00001379, -0.00000663, 0.00000833, 0.00001145, -0.00000040, -0.00000741, 0.00000084, 0.00000944, 0.00000488, -0.00000609, -0.00001394, -0.00001731, -0.00001009, 0.00001167, 0.00002652, 0.00001135, -0.00001093, -0.00000059, 0.00002258, 0.00000402, -0.00003920, -0.00003111, 0.00002441, 0.00003188, -0.00002780, -0.00004749, 0.00001885, 0.00005887, -0.00000507, -0.00006509, -0.00000916, 0.00007360, 0.00004028, -0.00005561, -0.00005591, 0.00003086, 0.00005479, -0.00001474, -0.00005514, -0.00001238, 0.00002852, 0.00001136, -0.00001442, -0.00001039, -0.00000516, -0.00001128, -0.00000301, 0.00001087, -0.00000304, -0.00002312, -0.00000474, 0.00002395, 0.00000393, -0.00003942, -0.00003375, 0.00001060, 0.00002136, -0.00000740, -0.00001936, -0.00000650, -0.00000710, -0.00001719, -0.00000797, -0.00000116, -0.00002516, -0.00003848, -0.00000630, 0.00000038, -0.00008629, -0.00017234, -0.00010764, 0.00007312, 0.00017676, 0.00012321, 0.00003591, 0.00003327, 0.00007514, 0.00007283, 0.00003359, 0.00002132, 0.00005108, 0.00008938, 0.00010983, 0.00009946, 0.00004856, -0.00002677, -0.00008013, -0.00008171, -0.00004964, -0.00001966, -0.00000584, -0.00000303, },
		{ -0.00000980, -0.00000763, -0.00000677, -0.00000684, -0.00000399, 0.00000028, 0.00000034, -0.00000370, -0.00000641, -0.00000577, -0.00000397, -0.00000205, -0.00000023, -0.00000022, -0.00000174, -0.00000128, 0.00000151, 0.00000197, -0.00000114, -0.00000243, 0.00000077, 0.00000333, 0.00000134, -0.00000168, -0.00000212, -0.00000218, -0.00000345, -0.00000260, 0.00000167, 0.00000463, 0.00000321, -0.00000016, -0.00000422, -0.00000948, -0.00001065, -0.00000149, 0.00001007, 0.00000842, -0.00000385, -0.00000653, 0.00000524, 0.00001132, -0.00000041, -0.00001522, -0.00001712, -0.00000916, 0.00000221, 0.00001572, 0.00002069, 0.00000650, -0.00000857, 0.00000060, 0.00001447, -0.00000465, -0.00003443, -0.00001779, 0.00002706, 0.00002034, -0.00003414, -0.00003721, 0.00002917, 0.00005095, -0.00001901, -0.00005947, 0.00001204, 0.00007847, 0.00001799, -0.00007463, -0.00004406, 0.00005587, 0.00005599, -0.00003806, -0.00006773, 0.00000278, 0.00004657, 0.00000587, -0.00003286, -0.00001408, 0.00000897, -0.00000092, -0.00000873, -0.00000016, -0.00000517, -0.00001637, 0.00000134, 0.00002273, -0.00000297, -0.00004391, -0.00002922, 0.00001953, 0.00002215, -0.00001727, -0.00002595, 0.00000087, 0.00000427, -0.00001814, -0.00001945, -0.00000739, -0.00002040, -0.00003195, -0.00000540, -0.00000294, -0.00009299, -0.00018391, -0.00011884, 0.00007200, 0.00018632, 0.00013460, 0.00004504, 0.00004347, 0.00008466, 0.00007398, 0.00002931, 0.00002867, 0.00007877, 0.00012061, 0.00011729, 0.00007391, 0.00000627, -0.00006109, -0.00009226, -0.00007372, -0.00003414, -0.00000924, -0.00000457, -0.00000654, -0.00000830, -0.00000997, },
		{ -0.00000730, -0.00000771, -0.00000501, -0.00000009, 0.00000085, -0.00000316, -0.00000625, -0.00000535, -0.00000306, -0.00000149, -0.00000055, -0.00000067, -0.00000151, -0.00000068, 0.00000167, 0.00000149, -0.00000176, -0.00000265, 0.00000110, 0.00000391, 0.00000157, -0.00000209, -0.00000274, -0.00000241, -0.00000324, -0.00000209, 0.00000248, 0.00000527, 0.00000224, -0.00000336, -0.00000727, -0.00000872, -0.00000598, 0.00000237, 0.00000901, 0.00000438, -0.00000520, -0.00000318, 0.00000879, 0.00000928, -0.00000803, -0.00002144, -0.00001462, 0.00000138, 0.00001131, 0.00001531, 0.00001407, 0.00000353, -0.00000629, -0.00000069, 0.00000623, -0.00000908, -0.00002626, -0.00000695, 0.00002482, 0.00000898, -0.00003556, -0.00002547, 0.00003451, 0.00003978, -0.00002779, -0.00004722, 0.00002995, 0.00007270, -0.00000736, -0.00008360, -0.00002257, 0.00007635, 0.00004484, -0.00006392, -0.00006967, 0.00002626, 0.00005898, -0.00001055, -0.00005188, -0.00000778, 0.00002919, 0.00000361, -0.00002327, -0.00001089, 0.00000077, -0.00000541, 0.00000248, 0.00001484, -0.00000918, -0.00004162, -0.00002094, 0.00002446, 0.00001655, -0.00002768, -0.00002651, 0.00001320, 0.00001309, -0.00002614, -0.00003353, -0.00000831, -0.00000924, -0.00002639, -0.00001167, -0.00001035, -0.00009580, -0.00018959, -0.00013058, 0.00006477, 0.00019302, 0.00015106, 0.00006143, 0.00005441, 0.00008788, 0.00007069, 0.00003057, 0.00004906, 0.00011500, 0.00014557, 0.00010571, 0.00003000, -0.00004072, -0.00008517, -0.00008931, -0.00005700, -0.00001960, -0.00000480, -0.00000793, -0.00001052, -0.00000866, -0.00000806, -0.00000852, -0.00000761, },
		{ -0.00000538, -0.00000042, 0.00000109, -0.00000261, -0.00000574, -0.00000467, -0.00000231, -0.00000138, -0.00000111, -0.00000094, -0.00000098, -0.00000013, 0.00000140, 0.00000074, -0.00000215, -0.00000245, 0.00000152, 0.00000418, 0.00000151, -0.00000246, -0.00000317, -0.00000251, -0.00000281, -0.00000129, 0.00000315, 0.00000489, 0.00000003, -0.00000652, -0.00000841, -0.00000579, -0.00000127, 0.00000390, 0.00000589, 0.00000106, -0.00000395, 0.00000108, 0.00000973, 0.00000380, -0.00001540, -0.00002304, -0.00000765, 0.00001152, 0.00001600, 0.00001169, 0.00000816, 0.00000222, -0.00000493, -0.00000385, -0.00000056, -0.00000930, -0.00001676, 0.00000009, 0.00001910, -0.00000025, -0.00003269, -0.00001455, 0.00003493, 0.00002822, -0.00003024, -0.00003164, 0.00004091, 0.00005806, -0.00003007, -0.00008060, 0.00000356, 0.00008700, 0.00002320, -0.00008573, -0.00005885, 0.00005275, 0.00006055, -0.00003570, -0.00006508, 0.00001006, 0.00005019, -0.00000228, -0.00004425, -0.00001557, 0.00001585, 0.00000497, -0.00000509, 0.00000266, -0.00000968, -0.00003182, -0.00001336, 0.00002182, 0.00000663, -0.00003375, -0.00001982, 0.00002645, 0.00001528, -0.00004007, -0.00004533, -0.00000147, 0.00000526, -0.00002648, -0.00002560, -0.00001767, -0.00009123, -0.00019092, -0.00014719, 0.00005007, 0.00020028, 0.00017627, 0.00008393, 0.00006127, 0.00008286, 0.00006762, 0.00004447, 0.00008388, 0.00015233, 0.00015384, 0.00007084, -0.00002675, -0.00008222, -0.00009255, -0.00007255, -0.00003765, -0.00001094, -0.00000675, -0.00001310, -0.00001241, -0.00000655, -0.00000563, -0.00000842, -0.00000886, -0.00000800, -0.00000787, },
		{ 0.00000118, -0.00000173, -0.00000480, -0.00000398, -0.00000199, -0.00000165, -0.00000165, -0.00000090, -0.00000031, 0.00000015, 0.00000077, -0.00000001, -0.00000216, -0.00000197, 0.00000178, 0.00000408, 0.00000130, -0.00000269, -0.00000340, -0.00000242, -0.00000215, -0.00000044, 0.00000322, 0.00000335, -0.00000279, -0.00000859, -0.00000734, -0.00000173, 0.00000209, 0.00000300, 0.00000209, -0.00000034, -0.00000072, 0.00000441, 0.00000734, -0.00000342, -0.00002014, -0.00001970, 0.00000152, 0.00001896, 0.00001633, 0.00000657, 0.00000384, 0.00000189, -0.00000496, -0.00000782, -0.00000480, -0.00000617, -0.00000791, 0.00000303, 0.00001176, -0.00000620, -0.00002695, -0.00000605, 0.00003162, 0.00001872, -0.00002701, -0.00001659, 0.00004333, 0.00003849, -0.00004537, -0.00006696, 0.00002809, 0.00008516, -0.00000404, -0.00009774, -0.00003684, 0.00007543, 0.00004888, -0.00006413, -0.00006692, 0.00003726, 0.00006524, -0.00002034, -0.00006617, -0.00000938, 0.00003756, 0.00000874, -0.00002237, -0.00000859, -0.00000042, -0.00001719, -0.00001210, 0.00001075, -0.00000264, -0.00003136, -0.00000805, 0.00003502, 0.00000913, -0.00005558, -0.00005006, 0.00001228, 0.00001773, -0.00003532, -0.00004422, -0.00001975, -0.00007847, -0.00019225, -0.00017237, 0.00002973, 0.00021309, 0.00021165, 0.00010811, 0.00005940, 0.00007160, 0.00007286, 0.00007654, 0.00012881, 0.00017897, 0.00013644, 0.00001447, -0.00008576, -0.00010821, -0.00008130, -0.00004768, -0.00002245, -0.00001046, -0.00001266, -0.00001668, -0.00001115, -0.00000332, -0.00000435, -0.00000984, -0.00001047, -0.00000798, -0.00000723, -0.00000564, -0.00000113, },
		{ -0.00000362, -0.00000362, -0.00000215, -0.00000206, -0.00000194, -0.00000063, 0.00000023, 0.00000008, -0.00000002, -0.00000053, -0.00000182, -0.00000142, 0.00000176, 0.00000369, 0.00000110, -0.00000272, -0.00000340, -0.00000213, -0.00000136, 0.00000011, 0.00000236, 0.00000088, -0.00000532, -0.00000887, -0.00000451, 0.00000213, 0.00000333, 0.00000040, -0.00000092, 0.00000051, 0.00000312, 0.00000537, 0.00000216, -0.00001015, -0.00002075, -0.00001253, 0.00001033, 0.00002243, 0.00001337, 0.00000161, 0.00000130, 0.00000169, -0.00000633, -0.00001141, -0.00000606, -0.00000111, -0.00000119, 0.00000250, 0.00000462, -0.00000864, -0.00002001, -0.00000064, 0.00002638, 0.00001268, -0.00002016, -0.00000530, 0.00003799, 0.00001886, -0.00005077, -0.00004671, 0.00004549, 0.00007165, -0.00003039, -0.00009682, -0.00000854, 0.00008818, 0.00002541, -0.00008871, -0.00005472, 0.00006815, 0.00006839, -0.00004839, -0.00008195, 0.00000956, 0.00006010, 0.00000085, -0.00004657, -0.00001235, 0.00001964, -0.00000358, -0.00002169, -0.00000578, -0.00000491, -0.00001939, 0.00000322, 0.00003408, -0.00000318, -0.00006636, -0.00004542, 0.00002832, 0.00002261, -0.00005245, -0.00006179, -0.00001295, -0.00006046, -0.00019908, -0.00020694, 0.00000875, 0.00023564, 0.00025471, 0.00012756, 0.00004714, 0.00006115, 0.00009549, 0.00012733, 0.00017308, 0.00018181, 0.00008973, -0.00005458, -0.00013402, -0.00011233, -0.00005516, -0.00002289, -0.00001627, -0.00001694, -0.00001849, -0.00001640, -0.00000764, -0.00000108, -0.00000511, -0.00001199, -0.00001130, -0.00000691, -0.00000631, -0.00000632, -0.00000219, 0.00000147, -0.00000030, },
		{ -0.00000262, -0.00000231, -0.00000194, -0.00000037, 0.00000046, -0.00000024, -0.00000071, -0.00000072, -0.00000129, -0.00000103, 0.00000143, 0.00000316, 0.00000101, -0.00000251, -0.00000319, -0.00000167, -0.00000069, 0.00000001, 0.00000054, -0.00000189, -0.00000672, -0.00000726, -0.00000088, 0.00000467, 0.00000244, -0.00000265, -0.00000218, 0.00000299, 0.00000598, 0.00000344, -0.00000431, -0.00001434, -0.00001705, -0.00000358, 0.00001665, 0.00002189, 0.00000877, -0.00000209, 0.00000009, 0.00000094, -0.00000851, -0.00001358, -0.00000463, 0.00000430, 0.00000267, -0.00000031, -0.00000098, -0.00000813, -0.00001334, 0.00000189, 0.00002098, 0.00001018, -0.00001238, 0.00000052, 0.00002761, 0.00000347, -0.00004664, -0.00002524, 0.00005249, 0.00005039, -0.00004981, -0.00008345, 0.00001932, 0.00008748, -0.00000485, -0.00010287, -0.00002985, 0.00009522, 0.00005665, -0.00008059, -0.00008531, 0.00003902, 0.00007602, -0.00002065, -0.00007138, -0.00000315, 0.00004733, 0.00000171, -0.00004335, -0.00002124, 0.00000506, -0.00000102, 0.00000674, 0.00002199, -0.00001554, -0.00006667, -0.00003347, 0.00003959, 0.00001687, -0.00007312, -0.00007204, 0.00000259, -0.00004332, -0.00021542, -0.00024753, -0.00000657, 0.00026861, 0.00029895, 0.00013686, 0.00002847, 0.00006249, 0.00014168, 0.00018974, 0.00020121, 0.00015081, 0.00001813, -0.00012198, -0.00016012, -0.00009449, -0.00002261, -0.00000597, -0.00002015, -0.00002625, -0.00002050, -0.00001220, -0.00000425, -0.00000148, -0.00000743, -0.00001341, -0.00001064, -0.00000530, -0.00000589, -0.00000759, -0.00000321, 0.00000234, 0.00000156, -0.00000261, -0.00000378, },
		{ -0.00000180, -0.00000035, 0.00000037, -0.00000058, -0.00000113, -0.00000066, -0.00000079, -0.00000089, 0.00000090, 0.00000264, 0.00000111, -0.00000207, -0.00000279, -0.00000121, -0.00000041, -0.00000091, -0.00000191, -0.00000418, -0.00000650, -0.00000425, 0.00000240, 0.00000529, 0.00000017, -0.00000492, -0.00000149, 0.00000589, 0.00000670, -0.00000080, -0.00001008, -0.00001478, -0.00001011, 0.00000482, 0.00001933, 0.00001832, 0.00000412, -0.00000415, -0.00000053, -0.00000062, -0.00001067, -0.00001376, -0.00000139, 0.00000871, 0.00000370, -0.00000399, -0.00000431, -0.00000564, -0.00000790, 0.00000236, 0.00001660, 0.00001022, -0.00000604, 0.00000098, 0.00001584, -0.00000515, -0.00003580, -0.00000765, 0.00004894, 0.00002704, -0.00005850, -0.00006148, 0.00004027, 0.00007363, -0.00003483, -0.00010272, 0.00000254, 0.00011138, 0.00003127, -0.00010902, -0.00007297, 0.00007286, 0.00007865, -0.00005328, -0.00008872, 0.00002104, 0.00007565, -0.00000744, -0.00007361, -0.00002741, 0.00002879, 0.00001642, -0.00000365, 0.00000208, -0.00001970, -0.00005426, -0.00002082, 0.00003934, 0.00000229, -0.00008958, -0.00007128, 0.00002194, -0.00003394, -0.00024105, -0.00028705, -0.00001171, 0.00030738, 0.00033597, 0.00013526, 0.00001373, 0.00008692, 0.00021041, 0.00024885, 0.00019724, 0.00008355, -0.00006554, -0.00017178, -0.00015831, -0.00006144, 0.00000600, -0.00000133, -0.00003092, -0.00003320, -0.00001700, -0.00000626, -0.00000353, -0.00000474, -0.00000971, -0.00001285, -0.00000874, -0.00000412, -0.00000645, -0.00000906, -0.00000361, 0.00000383, 0.00000342, -0.00000210, -0.00000439, -0.00000308, -0.00000225, },
		{ 0.00000014, -0.00000071, -0.00000124, -0.00000055, -0.00000050, -0.00000096, 0.00000032, 0.00000225, 0.00000139, -0.00000145, -0.00000228, -0.00000092, -0.00000076, -0.00000260, -0.00000443, -0.00000535, -0.00000468, -0.00000072, 0.00000439, 0.00000409, -0.00000240, -0.00000564, 0.00000055, 0.00000786, 0.00000496, -0.00000598, -0.00001344, -0.00001144, -0.00000181, 0.00001080, 0.00001836, 0.00001324, 0.00000059, -0.00000485, -0.00000127, -0.00000279, -0.00001198, -0.00001192, 0.00000251, 0.00001124, 0.00000258, -0.00000725, -0.00000532, -0.00000228, -0.00000405, 0.00000171, 0.00001361, 0.00001126, -0.00000251, -0.00000225, 0.00000603, -0.00000675, -0.00002244, 0.00000275, 0.00003751, 0.00000720, -0.00005589, -0.00003681, 0.00005009, 0.00005052, -0.00005758, -0.00008837, 0.00003495, 0.00011234, -0.00000244, -0.00012614, -0.00004615, 0.00010271, 0.00006455, -0.00009038, -0.00009137, 0.00005745, 0.00009564, -0.00003333, -0.00010465, -0.00001731, 0.00006252, 0.00002357, -0.00003002, -0.00001759, -0.00000837, -0.00003242, -0.00001685, 0.00002446, -0.00001409, -0.00009403, -0.00006103, 0.00003659, -0.00003650, -0.00027025, -0.00031748, -0.00000739, 0.00034249, 0.00035956, 0.00012961, 0.00001759, 0.00014101, 0.00029084, 0.00028478, 0.00015036, -0.00001212, -0.00014275, -0.00019146, -0.00013096, -0.00002472, 0.00002212, -0.00000829, -0.00004256, -0.00003394, -0.00000926, -0.00000194, -0.00000683, -0.00000942, -0.00001001, -0.00001002, -0.00000672, -0.00000432, -0.00000778, -0.00000998, -0.00000309, 0.00000554, 0.00000479, -0.00000215, -0.00000512, -0.00000326, -0.00000196, -0.00000180, -0.00000068, },
		{ -0.00000120, -0.00000059, -0.00000046, -0.00000112, -0.00000015, 0.00000200, 0.00000180, -0.00000071, -0.00000180, -0.00000105, -0.00000185, -0.00000471, -0.00000636, -0.00000507, -0.00000177, 0.00000233, 0.00000469, 0.00000178, -0.00000424, -0.00000471, 0.00000288, 0.00000799, 0.00000132, -0.00001039, -0.00001346, -0.00000536, 0.00000578, 0.00001339, 0.00001465, 0.00000816, -0.00000141, -0.00000483, -0.00000253, -0.00000499, -0.00001184, -0.00000860, 0.00000593, 0.00001166, 0.00000032, -0.00000923, -0.00000443, 0.00000105, -0.00000172, 0.00000069, 0.00001170, 0.00001189, -0.00000191, -0.00000674, 0.00000027, -0.00000312, -0.00001066, 0.00000520, 0.00002262, -0.00000527, -0.00004454, -0.00001544, 0.00004795, 0.00002435, -0.00006843, -0.00006380, 0.00005991, 0.00009795, -0.00003660, -0.00012725, -0.00001037, 0.00012057, 0.00003504, -0.00012290, -0.00007566, 0.00009871, 0.00009929, -0.00007291, -0.00012641, 0.00001182, 0.00009753, 0.00001229, -0.00006847, -0.00002629, 0.00002130, -0.00001007, -0.00003023, -0.00000192, -0.00002170, -0.00008249, -0.00004889, 0.00003766, -0.00004927, -0.00029296, -0.00033401, -0.00000134, 0.00036278, 0.00037034, 0.00013475, 0.00005418, 0.00022152, 0.00036277, 0.00027840, 0.00006005, -0.00011851, -0.00019448, -0.00017645, -0.00008832, 0.00000317, 0.00002227, -0.00002139, -0.00004888, -0.00002778, -0.00000091, -0.00000207, -0.00001344, -0.00001304, -0.00000720, -0.00000588, -0.00000600, -0.00000611, -0.00000904, -0.00000968, -0.00000187, 0.00000683, 0.00000537, -0.00000248, -0.00000556, -0.00000313, -0.00000172, -0.00000213, -0.00000124, 0.00000001, -0.00000053, },
		{ -0.00000054, -0.00000119, -0.00000047, 0.00000185, 0.00000229, 0.00000008, -0.00000150, -0.00000174, -0.00000353, -0.00000672, -0.00000719, -0.00000343, 0.00000142, 0.00000417, 0.00000346, -0.00000071, -0.00000472, -0.00000267, 0.00000440, 0.00000608, -0.00000297, -0.00001261, -0.00001021, 0.00000174, 0.00001104, 0.00001263, 0.00000962, 0.00000415, -0.00000206, -0.00000479, -0.00000427, -0.00000656, -0.00001013, -0.00000463, 0.00000804, 0.00001029, -0.00000204, -0.00000959, -0.00000237, 0.00000375, -0.00000058, -0.00000033, 0.00001019, 0.00001129, -0.00000324, -0.00001023, -0.00000105, 0.00000278, -0.00000328, 0.00000134, 0.00000887, -0.00000916, -0.00002906, -0.00000167, 0.00003639, 0.00000159, -0.00006630, -0.00003560, 0.00007222, 0.00007228, -0.00006323, -0.00011204, 0.00002606, 0.00012144, -0.00000390, -0.00014222, -0.00004308, 0.00013484, 0.00008250, -0.00011800, -0.00012984, 0.00005669, 0.00012236, -0.00002091, -0.00010928, -0.00001443, 0.00006537, 0.00000054, -0.00006488, -0.00003011, -0.00001033, -0.00005786, -0.00004675, 0.00002044, -0.00006332, -0.00029853, -0.00033912, -0.00000729, 0.00036075, 0.00037902, 0.00017039, 0.00013055, 0.00031272, 0.00040090, 0.00021829, -0.00006139, -0.00021223, -0.00020738, -0.00013229, -0.00004537, 0.00001325, 0.00000931, -0.00003266, -0.00004624, -0.00001757, 0.00000374, -0.00000739, -0.00002076, -0.00001334, -0.00000164, -0.00000225, -0.00000749, -0.00000880, -0.00000916, -0.00000800, -0.00000064, 0.00000713, 0.00000532, -0.00000258, -0.00000554, -0.00000294, -0.00000181, -0.00000276, -0.00000178, 0.00000015, -0.00000013, -0.00000123, -0.00000086, },
		{ -0.00000068, 0.00000171, 0.00000280, 0.00000083, -0.00000155, -0.00000304, -0.00000549, -0.00000810, -0.00000671, -0.00000095, 0.00000406, 0.00000449, 0.00000131, -0.00000251, -0.00000382, -0.00000043, 0.00000438, 0.00000269, -0.00000648, -0.00001193, -0.00000470, 0.00000805, 0.00001319, 0.00000942, 0.00000466, 0.00000165, -0.00000197, -0.00000516, -0.00000607, -0.00000696, -0.00000714, -0.00000094, 0.00000851, 0.00000784, -0.00000366, -0.00000842, 0.00000014, 0.00000551, -0.00000029, -0.00000135, 0.00000845, 0.00000947, -0.00000502, -0.00001131, 0.00000089, 0.00000797, -0.00000117, -0.00000569, -0.00000041, -0.00000599, -0.00001438, 0.00000302, 0.00002025, -0.00001300, -0.00005376, -0.00001075, 0.00007048, 0.00004222, -0.00007665, -0.00008468, 0.00005463, 0.00010494, -0.00004335, -0.00014293, -0.00000021, 0.00015631, 0.00004690, -0.00015744, -0.00011022, 0.00010814, 0.00012655, -0.00007264, -0.00013937, 0.00002252, 0.00011308, -0.00001205, -0.00011708, -0.00004611, 0.00002523, -0.00003076, -0.00006657, -0.00001194, -0.00006438, -0.00028105, -0.00034459, -0.00004066, 0.00033830, 0.00040642, 0.00025462, 0.00024079, 0.00038767, 0.00038223, 0.00010650, -0.00019036, -0.00027093, -0.00017869, -0.00007344, -0.00001673, 0.00000346, -0.00000868, -0.00003505, -0.00003527, -0.00000856, 0.00000179, -0.00001602, -0.00002545, -0.00000951, 0.00000482, -0.00000092, -0.00001100, -0.00001101, -0.00000746, -0.00000555, -0.00000030, 0.00000631, 0.00000519, -0.00000205, -0.00000521, -0.00000308, -0.00000234, -0.00000342, -0.00000201, 0.00000055, 0.00000023, -0.00000148, -0.00000130, -0.00000057, -0.00000110, },
		{ 0.00000330, 0.00000141, -0.00000207, -0.00000480, -0.00000728, -0.00000844, -0.00000508, 0.00000165, 0.00000554, 0.00000349, -0.00000091, -0.00000313, -0.00000200, 0.00000112, 0.00000272, -0.00000113, -0.00000808, -0.00000855, 0.00000153, 0.00001222, 0.00001233, 0.00000504, 0.00000074, 0.00000048, -0.00000178, -0.00000598, -0.00000740, -0.00000600, -0.00000353, 0.00000174, 0.00000747, 0.00000515, -0.00000410, -0.00000615, 0.00000245, 0.00000628, -0.00000062, -0.00000254, 0.00000619, 0.00000708, -0.00000587, -0.00000980, 0.00000410, 0.00001040, -0.00000331, -0.00001245, -0.00000395, 0.00000100, -0.00000424, 0.00000007, 0.00000489, -0.00001769, -0.00003596, 0.00000546, 0.00005725, 0.00001527, -0.00007518, -0.00005242, 0.00006928, 0.00007555, -0.00007414, -0.00012471, 0.00004311, 0.00015714, -0.00000020, -0.00018051, -0.00006925, 0.00015352, 0.00010440, -0.00013273, -0.00014645, 0.00008128, 0.00014941, -0.00005384, -0.00017503, -0.00003626, 0.00008161, -0.00001724, -0.00011492, -0.00004704, -0.00003759, -0.00024431, -0.00036991, -0.00011171, 0.00031054, 0.00047913, 0.00039590, 0.00036322, 0.00041412, 0.00029443, -0.00003917, -0.00029764, -0.00028062, -0.00011801, -0.00001870, -0.00001141, -0.00002059, -0.00002169, -0.00002548, -0.00002078, -0.00000594, -0.00000661, -0.00002403, -0.00002505, -0.00000279, 0.00000961, -0.00000261, -0.00001516, -0.00001140, -0.00000412, -0.00000342, -0.00000138, 0.00000483, 0.00000563, -0.00000083, -0.00000502, -0.00000386, -0.00000310, -0.00000374, -0.00000187, 0.00000095, 0.00000032, -0.00000193, -0.00000169, -0.00000043, -0.00000088, -0.00000090, 0.00000156, },
		{ -0.00000306, -0.00000673, -0.00000844, -0.00000762, -0.00000278, 0.00000367, 0.00000565, 0.00000174, -0.00000250, -0.00000256, -0.00000003, 0.00000142, -0.00000007, -0.00000421, -0.00000727, -0.00000339, 0.00000696, 0.00001362, 0.00000931, 0.00000075, -0.00000175, 0.00000010, -0.00000191, -0.00000695, -0.00000776, -0.00000387, -0.00000005, 0.00000304, 0.00000543, 0.00000294, -0.00000331, -0.00000338, 0.00000414, 0.00000613, -0.00000146, -0.00000404, 0.00000360, 0.00000503, -0.00000505, -0.00000660, 0.00000655, 0.00000948, -0.00000749, -0.00001644, -0.00000261, 0.00000809, -0.00000016, -0.00000706, -0.00000558, -0.00001390, -0.00001872, 0.00001094, 0.00003792, -0.00000278, -0.00006135, -0.00002323, 0.00006814, 0.00004103, -0.00008973, -0.00009244, 0.00007681, 0.00013698, -0.00004783, -0.00018045, -0.00001514, 0.00018050, 0.00005753, -0.00018672, -0.00012345, 0.00015058, 0.00015975, -0.00012199, -0.00022149, 0.00000736, 0.00014587, -0.00003358, -0.00018850, -0.00006551, 0.00002590, -0.00020409, -0.00043676, -0.00021859, 0.00030544, 0.00062139, 0.00058606, 0.00046259, 0.00036371, 0.00014243, -0.00018784, -0.00035690, -0.00024094, -0.00004480, 0.00001512, -0.00002949, -0.00004778, -0.00002199, -0.00000638, -0.00000969, -0.00001237, -0.00001810, -0.00002725, -0.00001933, 0.00000392, 0.00001070, -0.00000650, -0.00001801, -0.00000941, -0.00000025, -0.00000269, -0.00000375, 0.00000355, 0.00000703, 0.00000066, -0.00000550, -0.00000522, -0.00000365, -0.00000346, -0.00000155, 0.00000099, 0.00000007, -0.00000234, -0.00000182, -0.00000012, -0.00000069, -0.00000118, 0.00000140, 0.00000377, 0.00000171, },
		{ -0.00000865, -0.00000582, -0.00000044, 0.00000461, 0.00000459, -0.00000006, -0.00000304, -0.00000119, 0.00000140, 0.00000043, -0.00000313, -0.00000567, -0.00000429, 0.00000225, 0.00001047, 0.00001242, 0.00000527, -0.00000261, -0.00000296, -0.00000008, -0.00000246, -0.00000760, -0.00000693, -0.00000107, 0.00000266, 0.00000298, 0.00000304, 0.00000165, -0.00000156, -0.00000068, 0.00000497, 0.00000520, -0.00000272, -0.00000577, 0.00000120, 0.00000406, -0.00000268, -0.00000310, 0.00000694, 0.00000599, -0.00001132, -0.00001670, 0.00000135, 0.00001246, -0.00000133, -0.00001437, -0.00000938, -0.00000532, -0.00000665, 0.00000701, 0.00001875, -0.00000950, -0.00004079, -0.00000331, 0.00005388, 0.00000994, -0.00008815, -0.00005466, 0.00009372, 0.00010129, -0.00008472, -0.00015679, 0.00003971, 0.00018114, -0.00000497, -0.00022023, -0.00007161, 0.00021385, 0.00013507, -0.00020419, -0.00023899, 0.00008266, 0.00019819, -0.00008999, -0.00027248, -0.00004682, 0.00012295, -0.00018624, -0.00056067, -0.00034276, 0.00035839, 0.00084529, 0.00079723, 0.00049757, 0.00022204, -0.00004959, -0.00030303, -0.00035325, -0.00016642, 0.00001757, 0.00001921, -0.00006194, -0.00006592, -0.00000751, 0.00001517, -0.00000796, -0.00002648, -0.00002733, -0.00002324, -0.00001064, 0.00000729, 0.00000771, -0.00001041, -0.00001798, -0.00000569, 0.00000261, -0.00000382, -0.00000652, 0.00000328, 0.00000919, 0.00000171, -0.00000688, -0.00000673, -0.00000351, -0.00000266, -0.00000148, 0.00000047, -0.00000032, -0.00000240, -0.00000164, 0.00000016, -0.00000068, -0.00000152, 0.00000129, 0.00000411, 0.00000160, -0.00000442, -0.00000843, },
		{ 0.00000133, 0.00000432, 0.00000287, -0.00000130, -0.00000253, 0.00000037, 0.00000186, -0.00000144, -0.00000554, -0.00000518, 0.00000006, 0.00000710, 0.00001160, 0.00000936, 0.00000130, -0.00000468, -0.00000337, -0.00000047, -0.00000320, -0.00000750, -0.00000504, 0.00000179, 0.00000422, 0.00000193, 0.00000090, 0.00000139, 0.00000068, 0.00000150, 0.00000491, 0.00000364, -0.00000427, -0.00000742, -0.00000045, 0.00000445, 0.00000046, -0.00000066, 0.00000504, 0.00000152, -0.00001308, -0.00001387, 0.00000530, 0.00001296, -0.00000537, -0.00001870, -0.00000729, 0.00000363, -0.00000186, -0.00000244, 0.00000482, -0.00000610, -0.00002015, 0.00000465, 0.00003249, -0.00001106, -0.00007233, -0.00002075, 0.00009168, 0.00005956, -0.00010282, -0.00011567, 0.00008239, 0.00015465, -0.00006892, -0.00022366, -0.00000093, 0.00025407, 0.00007581, -0.00028144, -0.00021593, 0.00017637, 0.00021763, -0.00018523, -0.00034306, 0.00002298, 0.00023559, -0.00022005, -0.00074267, -0.00044945, 0.00050240, 0.00114246, 0.00098489, 0.00043201, -0.00000388, -0.00024206, -0.00035303, -0.00028901, -0.00008274, 0.00004973, -0.00000387, -0.00009401, -0.00006675, 0.00001701, 0.00003038, -0.00001772, -0.00004310, -0.00002937, -0.00001264, -0.00000308, 0.00000519, 0.00000232, -0.00001178, -0.00001477, -0.00000198, 0.00000318, -0.00000625, -0.00000847, 0.00000431, 0.00001135, 0.00000167, -0.00000891, -0.00000771, -0.00000250, -0.00000175, -0.00000209, -0.00000057, -0.00000046, -0.00000190, -0.00000129, 0.00000017, -0.00000090, -0.00000179, 0.00000126, 0.00000420, 0.00000103, -0.00000589, -0.00000951, -0.00000782, -0.00000348, },
		{ 0.00000109, -0.00000163, -0.00000127, 0.00000157, 0.00000129, -0.00000353, -0.00000665, -0.00000297, 0.00000470, 0.00001027, 0.00001052, 0.00000544, -0.00000185, -0.00000559, -0.00000349, -0.00000118, -0.00000376, -0.00000643, -0.00000244, 0.00000413, 0.00000458, 0.00000036, -0.00000058, 0.00000200, 0.00000294, 0.00000289, 0.00000401, 0.00000166, -0.00000584, -0.00000857, -0.00000098, 0.00000588, 0.00000331, -0.00000004, 0.00000154, -0.00000224, -0.00001223, -0.00000957, 0.00000731, 0.00001014, -0.00000941, -0.00001871, -0.00000193, 0.00000939, -0.00000367, -0.00001265, -0.00000143, 0.00000340, -0.00000487, 0.00000188, 0.00001108, -0.00001908, -0.00004883, 0.00000202, 0.00007402, 0.00002225, -0.00009974, -0.00006787, 0.00010359, 0.00010784, -0.00011936, -0.00019556, 0.00007234, 0.00025934, -0.00000725, -0.00033348, -0.00015136, 0.00026667, 0.00018929, -0.00030450, -0.00037396, 0.00014421, 0.00033307, -0.00032867, -0.00096381, -0.00049398, 0.00075650, 0.00148067, 0.00109682, 0.00024680, -0.00028368, -0.00038935, -0.00032087, -0.00018438, -0.00001892, 0.00004304, -0.00004087, -0.00011090, -0.00004964, 0.00004169, 0.00003262, -0.00003588, -0.00005529, -0.00002205, 0.00000077, -0.00000069, -0.00000211, -0.00000235, -0.00000891, -0.00000971, -0.00000034, 0.00000127, -0.00000858, -0.00000865, 0.00000609, 0.00001249, 0.00000043, -0.00001081, -0.00000761, -0.00000088, -0.00000138, -0.00000356, -0.00000169, 0.00000007, -0.00000089, -0.00000113, -0.00000030, -0.00000123, -0.00000184, 0.00000124, 0.00000389, 0.00000009, -0.00000709, -0.00000967, -0.00000615, -0.00000118, 0.00000210, 0.00000299, },
		{ 0.00000027, 0.00000205, 0.00000001, -0.00000515, -0.00000623, 0.00000029, 0.00000862, 0.00001142, 0.00000787, 0.00000158, -0.00000390, -0.00000572, -0.00000370, -0.00000207, -0.00000375, -0.00000441, 0.00000038, 0.00000561, 0.00000395, -0.00000123, -0.00000126, 0.00000313, 0.00000481, 0.00000340, 0.00000247, -0.00000048, -0.00000710, -0.00000886, -0.00000040, 0.00000763, 0.00000497, -0.00000120, -0.00000229, -0.00000418, -0.00000933, -0.00000564, 0.00000665, 0.00000576, -0.00001127, -0.00001497, 0.00000353, 0.00001034, -0.00000935, -0.00001966, -0.00000052, 0.00001396, 0.00000234, -0.00000739, -0.00000451, -0.00001550, -0.00002541, 0.00001055, 0.00004803, -0.00000255, -0.00007925, -0.00002530, 0.00010043, 0.00005300, -0.00014521, -0.00014334, 0.00013060, 0.00022700, -0.00009555, -0.00034525, -0.00005660, 0.00032932, 0.00011041, -0.00042199, -0.00034513, 0.00030000, 0.00037912, -0.00051928, -0.00118529, -0.00043299, 0.00111602, 0.00180711, 0.00108637, -0.00005157, -0.00056852, -0.00045207, -0.00021074, -0.00007240, 0.00000194, 0.00000305, -0.00007247, -0.00010377, -0.00002229, 0.00005553, 0.00002066, -0.00005503, -0.00005740, -0.00000730, 0.00001151, -0.00000547, -0.00001184, -0.00000311, -0.00000197, -0.00000528, -0.00000211, -0.00000205, -0.00000918, -0.00000698, 0.00000741, 0.00001188, -0.00000143, -0.00001167, -0.00000636, 0.00000060, -0.00000205, -0.00000559, -0.00000231, 0.00000143, 0.00000027, -0.00000152, -0.00000120, -0.00000140, -0.00000159, 0.00000105, 0.00000305, -0.00000099, -0.00000766, -0.00000885, -0.00000404, 0.00000054, 0.00000176, 0.00000107, -0.00000022, -0.00000101, },
		{ -0.00000152, -0.00000583, -0.00000450, 0.00000376, 0.00001115, 0.00001071, 0.00000447, -0.00000160, -0.00000491, -0.00000553, -0.00000414, -0.00000283, -0.00000294, -0.00000170, 0.00000292, 0.00000608, 0.00000265, -0.00000251, -0.00000121, 0.00000442, 0.00000602, 0.00000308, 0.00000054, -0.00000246, -0.00000769, -0.00000815, 0.00000087, 0.00000886, 0.00000503, -0.00000344, -0.00000524, -0.00000405, -0.00000565, -0.00000338, 0.00000388, 0.00000183, -0.00001013, -0.00000943, 0.00000656, 0.00000696, -0.00001542, -0.00002155, 0.00000474, 0.00002134, 0.00000197, -0.00001778, -0.00001140, -0.00000495, -0.00000842, 0.00000644, 0.00002235, -0.00001143, -0.00004967, 0.00000275, 0.00007710, 0.00000394, -0.00014273, -0.00008110, 0.00016032, 0.00016489, -0.00016813, -0.00031217, 0.00004721, 0.00034527, -0.00000701, -0.00050766, -0.00025067, 0.00045811, 0.00034257, -0.00077657, -0.00135530, -0.00023766, 0.00154832, 0.00205876, 0.00092614, -0.00042880, -0.00080216, -0.00040870, -0.00004826, 0.00001056, -0.00002947, -0.00005199, -0.00008040, -0.00007362, 0.00000187, 0.00005110, -0.00000018, -0.00006630, -0.00004783, 0.00000926, 0.00001485, -0.00001623, -0.00001973, 0.00000169, 0.00000681, -0.00000405, -0.00000712, -0.00000473, -0.00000703, -0.00000447, 0.00000692, 0.00000959, -0.00000276, -0.00001086, -0.00000454, 0.00000107, -0.00000380, -0.00000749, -0.00000192, 0.00000338, 0.00000102, -0.00000265, -0.00000225, -0.00000116, -0.00000113, 0.00000045, 0.00000173, -0.00000186, -0.00000731, -0.00000724, -0.00000206, 0.00000129, 0.00000049, -0.00000090, -0.00000072, 0.00000034, 0.00000166, 0.00000177, },
		{ -0.00000195, 0.00000669, 0.00001202, 0.00000864, 0.00000107, -0.00000382, -0.00000517, -0.00000535, -0.00000473, -0.00000315, -0.00000128, 0.00000133, 0.00000484, 0.00000559, 0.00000104, -0.00000329, -0.00000062, 0.00000555, 0.00000647, 0.00000208, -0.00000149, -0.00000393, -0.00000739, -0.00000662, 0.00000216, 0.00000896, 0.00000367, -0.00000577, -0.00000660, -0.00000241, -0.00000251, -0.00000314, 0.00000034, -0.00000021, -0.00000657, -0.00000436, 0.00000614, 0.00000131, -0.00001906, -0.00001868, 0.00001069, 0.00002338, -0.00000304, -0.00002476, -0.00001015, 0.00000673, -0.00000093, -0.00000514, 0.00000399, -0.00000631, -0.00002085, 0.00001235, 0.00004317, -0.00002838, -0.00011647, -0.00002509, 0.00015633, 0.00008911, -0.00020814, -0.00024204, 0.00013509, 0.00030723, -0.00013837, -0.00053637, -0.00010316, 0.00057833, 0.00020828, -0.00106232, -0.00142117, 0.00009508, 0.00199620, 0.00217719, 0.00061790, -0.00082839, -0.00093519, -0.00026277, 0.00012575, 0.00003581, -0.00010514, -0.00009709, -0.00005441, -0.00003167, 0.00000886, 0.00002822, -0.00001958, -0.00006319, -0.00003025, 0.00002047, 0.00000909, -0.00002869, -0.00002196, 0.00001114, 0.00001402, -0.00000735, -0.00001353, -0.00000472, -0.00000239, -0.00000288, 0.00000386, 0.00000664, -0.00000233, -0.00000850, -0.00000326, -0.00000003, -0.00000607, -0.00000843, -0.00000045, 0.00000529, 0.00000090, -0.00000435, -0.00000302, -0.00000038, -0.00000073, -0.00000073, 0.00000016, -0.00000210, -0.00000595, -0.00000524, -0.00000071, 0.00000099, -0.00000129, -0.00000243, -0.00000037, 0.00000205, 0.00000258, 0.00000087, -0.00000283, -0.00000543, },
		{ 0.00001133, 0.00000585, -0.00000180, -0.00000509, -0.00000506, -0.00000533, -0.00000526, -0.00000277, 0.00000108, 0.00000429, 0.00000590, 0.00000430, -0.00000060, -0.00000354, 0.00000026, 0.00000628, 0.00000617, 0.00000065, -0.00000327, -0.00000462, -0.00000623, -0.00000472, 0.00000282, 0.00000774, 0.00000156, -0.00000729, -0.00000629, -0.00000026, -0.00000076, -0.00000439, -0.00000256, 0.00000003, -0.00000211, -0.00000145, 0.00000282, -0.00000410, -0.00001900, -0.00001314, 0.00001427, 0.00002042, -0.00000882, -0.00002606, -0.00000397, 0.00001464, -0.00000223, -0.00001771, -0.00000374, 0.00000687, -0.00000079, 0.00000569, 0.00001006, -0.00003917, -0.00007744, 0.00001144, 0.00012330, 0.00001899, -0.00020815, -0.00015307, 0.00018594, 0.00022279, -0.00025310, -0.00049648, 0.00006761, 0.00062362, -0.00001535, -0.00132215, -0.00134386, 0.00053732, 0.00238883, 0.00212398, 0.00019540, -0.00118403, -0.00093826, -0.00004314, 0.00026636, -0.00000877, -0.00020185, -0.00010901, 0.00000336, 0.00000426, -0.00001021, -0.00000512, -0.00002609, -0.00004475, -0.00001269, 0.00002071, -0.00000321, -0.00003715, -0.00001707, 0.00002180, 0.00001653, -0.00001438, -0.00001851, -0.00000102, 0.00000307, -0.00000391, -0.00000138, 0.00000467, 0.00000048, -0.00000559, -0.00000365, -0.00000256, -0.00000781, -0.00000783, 0.00000159, 0.00000638, -0.00000022, -0.00000609, -0.00000312, 0.00000074, -0.00000076, -0.00000251, -0.00000126, -0.00000141, -0.00000379, -0.00000342, -0.00000034, -0.00000017, -0.00000306, -0.00000320, 0.00000071, 0.00000376, 0.00000292, -0.00000034, -0.00000360, -0.00000407, 0.00000083, 0.00000856, },
		{ -0.00000386, -0.00000560, -0.00000486, -0.00000548, -0.00000545, -0.00000160, 0.00000389, 0.00000679, 0.00000599, 0.00000243, -0.00000205, -0.00000329, 0.00000124, 0.00000648, 0.00000523, -0.00000090, -0.00000447, -0.00000440, -0.00000446, -0.00000306, 0.00000245, 0.00000548, -0.00000049, -0.00000751, -0.00000483, 0.00000144, -0.00000054, -0.00000608, -0.00000387, 0.00000196, 0.00000160, -0.00000124, -0.00000178, -0.00000732, -0.00001568, -0.00000759, 0.00001406, 0.00001461, -0.00001211, -0.00002207, 0.00000293, 0.00001638, -0.00000872, -0.00002590, -0.00000179, 0.00002067, 0.00000676, -0.00001029, -0.00001281, -0.00003082, -0.00003915, 0.00002228, 0.00007384, -0.00002903, -0.00017234, -0.00006821, 0.00018914, 0.00011274, -0.00032385, -0.00039474, 0.00022180, 0.00057164, -0.00029518, -0.00149799, -0.00111031, 0.00103307, 0.00265771, 0.00189204, -0.00028145, -0.00143544, -0.00081040, 0.00020311, 0.00033717, -0.00011475, -0.00028831, -0.00007449, 0.00007833, 0.00001605, -0.00005495, -0.00003501, -0.00001177, -0.00001676, -0.00000459, 0.00000868, -0.00001579, -0.00003697, -0.00000705, 0.00002902, 0.00001317, -0.00002228, -0.00001954, 0.00000555, 0.00000693, -0.00000828, -0.00000722, 0.00000518, 0.00000525, -0.00000367, -0.00000630, -0.00000559, -0.00000792, -0.00000578, 0.00000324, 0.00000607, -0.00000195, -0.00000721, -0.00000245, 0.00000176, -0.00000152, -0.00000459, -0.00000203, 0.00000030, -0.00000129, -0.00000227, -0.00000103, -0.00000179, -0.00000436, -0.00000311, 0.00000222, 0.00000518, 0.00000272, -0.00000157, -0.00000369, -0.00000207, 0.00000331, 0.00000916, 0.00000945, 0.00000294, },
		{ -0.00000478, -0.00000564, -0.00000504, 0.00000035, 0.00000682, 0.00000851, 0.00000510, 0.00000021, -0.00000313, -0.00000261, 0.00000211, 0.00000611, 0.00000382, -0.00000226, -0.00000487, -0.00000335, -0.00000255, -0.00000215, 0.00000097, 0.00000277, -0.00000180, -0.00000646, -0.00000301, 0.00000213, -0.00000136, -0.00000722, -0.00000348, 0.00000437, 0.00000345, -0.00000313, -0.00000580, -0.00000763, -0.00001079, -0.00000416, 0.00001047, 0.00000879, -0.00001142, -0.00001512, 0.00000709, 0.00001247, -0.00001575, -0.00002717, 0.00000564, 0.00002898, 0.00000301, -0.00002691, -0.00002097, -0.00001142, -0.00001297, 0.00000976, 0.00002410, -0.00004648, -0.00011488, -0.00000795, 0.00014796, 0.00000490, -0.00033509, -0.00025560, 0.00032047, 0.00042299, -0.00057953, -0.00154318, -0.00073972, 0.00150926, 0.00275336, 0.00150926, -0.00073972, -0.00154318, -0.00057953, 0.00042299, 0.00032047, -0.00025560, -0.00033509, 0.00000490, 0.00014796, -0.00000795, -0.00011488, -0.00004648, 0.00002410, 0.00000976, -0.00001297, -0.00001142, -0.00002097, -0.00002691, 0.00000301, 0.00002898, 0.00000564, -0.00002717, -0.00001575, 0.00001247, 0.00000709, -0.00001512, -0.00001142, 0.00000879, 0.00001047, -0.00000416, -0.00001079, -0.00000763, -0.00000580, -0.00000313, 0.00000345, 0.00000437, -0.00000348, -0.00000722, -0.00000136, 0.00000213, -0.00000301, -0.00000646, -0.00000180, 0.00000277, 0.00000097, -0.00000215, -0.00000255, -0.00000335, -0.00000487, -0.00000226, 0.00000382, 0.00000611, 0.00000211, -0.00000261, -0.00000313, 0.00000021, 0.00000510, 0.00000851, 0.00000682, 0.00000035, -0.00000504, -0.00000564, },
		{ -0.00000386, 0.00000294, 0.00000945, 0.00000916, 0.00000331, -0.00000207, -0.00000369, -0.00000157, 0.00000272, 0.00000518, 0.00000222, -0.00000311, -0.00000436, -0.00000179, -0.00000103, -0.00000227, -0.00000129, 0.00000030, -0.00000203, -0.00000459, -0.00000152, 0.00000176, -0.00000245, -0.00000721, -0.00000195, 0.00000607, 0.00000324, -0.00000578, -0.00000792, -0.00000559, -0.00000630, -0.00000367, 0.00000525, 0.00000518, -0.00000722, -0.00000828, 0.00000693, 0.00000555, -0.00001954, -0.00002228, 0.00001317, 0.00002902, -0.00000705, -0.00003697, -0.00001579, 0.00000868, -0.00000459, -0.00001676, -0.00001177, -0.00003501, -0.00005495, 0.00001605, 0.00007833, -0.00007449, -0.00028831, -0.00011475, 0.00033717, 0.00020311, -0.00081040, -0.00143544, -0.00028145, 0.00189204, 0.00265771, 0.00103307, -0.00111031, -0.00149799, -0.00029518, 0.00057164, 0.00022180, -0.00039474, -0.00032385, 0.00011274, 0.00018914, -0.00006821, -0.00017234, -0.00002903, 0.00007384, 0.00002228, -0.00003915, -0.00003082, -0.00001281, -0.00001029, 0.00000676, 0.00002067, -0.00000179, -0.00002590, -0.00000872, 0.00001638, 0.00000293, -0.00002207, -0.00001211, 0.00001461, 0.00001406, -0.00000759, -0.00001568, -0.00000732, -0.00000178, -0.00000124, 0.00000160, 0.00000196, -0.00000387, -0.00000608, -0.00000054, 0.00000144, -0.00000483, -0.00000751, -0.00000049, 0.00000548, 0.00000245, -0.00000306, -0.00000446, -0.00000440, -0.00000447, -0.00000090, 0.00000523, 0.00000648, 0.00000124, -0.00000329, -0.00000205, 0.00000243, 0.00000599, 0.00000679, 0.00000389, -0.00000160, -0.00000545, -0.00000548, -0.00000486, -0.00000560, },
		{ 0.00001133, 0.00000856, 0.00000083, -0.00000407, -0.00000360, -0.00000034, 0.00000292, 0.00000376, 0.00000071, -0.00000320, -0.00000306, -0.00000017, -0.00000034, -0.00000342, -0.00000379, -0.00000141, -0.00000126, -0.00000251, -0.00000076, 0.00000074, -0.00000312, -0.00000609, -0.00000022, 0.00000638, 0.00000159, -0.00000783, -0.00000781, -0.00000256, -0.00000365, -0.00000559, 0.00000048, 0.00000467, -0.00000138, -0.00000391, 0.00000307, -0.00000102, -0.00001851, -0.00001438, 0.00001653, 0.00002180, -0.00001707, -0.00003715, -0.00000321, 0.00002071, -0.00001269, -0.00004475, -0.00002609, -0.00000512, -0.00001021, 0.00000426, 0.00000336, -0.00010901, -0.00020185, -0.00000877, 0.00026636, -0.00004314, -0.00093826, -0.00118403, 0.00019540, 0.00212398, 0.00238883, 0.00053732, -0.00134386, -0.00132215, -0.00001535, 0.00062362, 0.00006761, -0.00049648, -0.00025310, 0.00022279, 0.00018594, -0.00015307, -0.00020815, 0.00001899, 0.00012330, 0.00001144, -0.00007744, -0.00003917, 0.00001006, 0.00000569, -0.00000079, 0.00000687, -0.00000374, -0.00001771, -0.00000223, 0.00001464, -0.00000397, -0.00002606, -0.00000882, 0.00002042, 0.00001427, -0.00001314, -0.00001900, -0.00000410, 0.00000282, -0.00000145, -0.00000211, 0.00000003, -0.00000256, -0.00000439, -0.00000076, -0.00000026, -0.00000629, -0.00000729, 0.00000156, 0.00000774, 0.00000282, -0.00000472, -0.00000623, -0.00000462, -0.00000327, 0.00000065, 0.00000617, 0.00000628, 0.00000026, -0.00000354, -0.00000060, 0.00000430, 0.00000590, 0.00000429, 0.00000108, -0.00000277, -0.00000526, -0.00000533, -0.00000506, -0.00000509, -0.00000180, 0.00000585, },
		{ -0.00000195, -0.00000543, -0.00000283, 0.00000087, 0.00000258, 0.00000205, -0.00000037, -0.00000243, -0.00000129, 0.00000099, -0.00000071, -0.00000524, -0.00000595, -0.00000210, 0.00000016, -0.00000073, -0.00000073, -0.00000038, -0.00000302, -0.00000435, 0.00000090, 0.00000529, -0.00000045, -0.00000843, -0.00000607, -0.00000003, -0.00000326, -0.00000850, -0.00000233, 0.00000664, 0.00000386, -0.00000288, -0.00000239, -0.00000472, -0.00001353, -0.00000735, 0.00001402, 0.00001114, -0.00002196, -0.00002869, 0.00000909, 0.00002047, -0.00003025, -0.00006319, -0.00001958, 0.00002822, 0.00000886, -0.00003167, -0.00005441, -0.00009709, -0.00010514, 0.00003581, 0.00012575, -0.00026277, -0.00093519, -0.00082839, 0.00061790, 0.00217719, 0.00199620, 0.00009508, -0.00142117, -0.00106232, 0.00020828, 0.00057833, -0.00010316, -0.00053637, -0.00013837, 0.00030723, 0.00013509, -0.00024204, -0.00020814, 0.00008911, 0.00015633, -0.00002509, -0.00011647, -0.00002838, 0.00004317, 0.00001235, -0.00002085, -0.00000631, 0.00000399, -0.00000514, -0.00000093, 0.00000673, -0.00001015, -0.00002476, -0.00000304, 0.00002338, 0.00001069, -0.00001868, -0.00001906, 0.00000131, 0.00000614, -0.00000436, -0.00000657, -0.00000021, 0.00000034, -0.00000314, -0.00000251, -0.00000241, -0.00000660, -0.00000577, 0.00000367, 0.00000896, 0.00000216, -0.00000662, -0.00000739, -0.00000393, -0.00000149, 0.00000208, 0.00000647, 0.00000555, -0.00000062, -0.00000329, 0.00000104, 0.00000559, 0.00000484, 0.00000133, -0.00000128, -0.00000315, -0.00000473, -0.00000535, -0.00000517, -0.00000382, 0.00000107, 0.00000864, 0.00001202, 0.00000669, },
		{ -0.00000152, 0.00000177, 0.00000166, 0.00000034, -0.00000072, -0.00000090, 0.00000049, 0.00000129, -0.00000206, -0.00000724, -0.00000731, -0.00000186, 0.00000173, 0.00000045, -0.00000113, -0.00000116, -0.00000225, -0.00000265, 0.00000102, 0.00000338, -0.00000192, -0.00000749, -0.00000380, 0.00000107, -0.00000454, -0.00001086, -0.00000276, 0.00000959, 0.00000692, -0.00000447, -0.00000703, -0.00000473, -0.00000712, -0.00000405, 0.00000681, 0.00000169, -0.00001973, -0.00001623, 0.00001485, 0.00000926, -0.00004783, -0.00006630, -0.00000018, 0.00005110, 0.00000187, -0.00007362, -0.00008040, -0.00005199, -0.00002947, 0.00001056, -0.00004826, -0.00040870, -0.00080216, -0.00042880, 0.00092614, 0.00205876, 0.00154832, -0.00023766, -0.00135530, -0.00077657, 0.00034257, 0.00045811, -0.00025067, -0.00050766, -0.00000701, 0.00034527, 0.00004721, -0.00031217, -0.00016813, 0.00016489, 0.00016032, -0.00008110, -0.00014273, 0.00000394, 0.00007710, 0.00000275, -0.00004967, -0.00001143, 0.00002235, 0.00000644, -0.00000842, -0.00000495, -0.00001140, -0.00001778, 0.00000197, 0.00002134, 0.00000474, -0.00002155, -0.00001542, 0.00000696, 0.00000656, -0.00000943, -0.00001013, 0.00000183, 0.00000388, -0.00000338, -0.00000565, -0.00000405, -0.00000524, -0.00000344, 0.00000503, 0.00000886, 0.00000087, -0.00000815, -0.00000769, -0.00000246, 0.00000054, 0.00000308, 0.00000602, 0.00000442, -0.00000121, -0.00000251, 0.00000265, 0.00000608, 0.00000292, -0.00000170, -0.00000294, -0.00000283, -0.00000414, -0.00000553, -0.00000491, -0.00000160, 0.00000447, 0.00001071, 0.00001115, 0.00000376, -0.00000450, -0.00000583, },
		{ 0.00000027, -0.00000101, -0.00000022, 0.00000107, 0.00000176, 0.00000054, -0.00000404, -0.00000885, -0.00000766, -0.00000099, 0.00000305, 0.00000105, -0.00000159, -0.00000140, -0.00000120, -0.00000152, 0.00000027, 0.00000143, -0.00000231, -0.00000559, -0.00000205, 0.00000060, -0.00000636, -0.00001167, -0.00000143, 0.00001188, 0.00000741, -0.00000698, -0.00000918, -0.00000205, -0.00000211, -0.00000528, -0.00000197, -0.00000311, -0.00001184, -0.00000547, 0.00001151, -0.00000730, -0.00005740, -0.00005503, 0.00002066, 0.00005553, -0.00002229, -0.00010377, -0.00007247, 0.00000305, 0.00000194, -0.00007240, -0.00021074, -0.00045207, -0.00056852, -0.00005157, 0.00108637, 0.00180711, 0.00111602, -0.00043299, -0.00118529, -0.00051928, 0.00037912, 0.00030000, -0.00034513, -0.00042199, 0.00011041, 0.00032932, -0.00005660, -0.00034525, -0.00009555, 0.00022700, 0.00013060, -0.00014334, -0.00014521, 0.00005300, 0.00010043, -0.00002530, -0.00007925, -0.00000255, 0.00004803, 0.00001055, -0.00002541, -0.00001550, -0.00000451, -0.00000739, 0.00000234, 0.00001396, -0.00000052, -0.00001966, -0.00000935, 0.00001034, 0.00000353, -0.00001497, -0.00001127, 0.00000576, 0.00000665, -0.00000564, -0.00000933, -0.00000418, -0.00000229, -0.00000120, 0.00000497, 0.00000763, -0.00000040, -0.00000886, -0.00000710, -0.00000048, 0.00000247, 0.00000340, 0.00000481, 0.00000313, -0.00000126, -0.00000123, 0.00000395, 0.00000561, 0.00000038, -0.00000441, -0.00000375, -0.00000207, -0.00000370, -0.00000572, -0.00000390, 0.00000158, 0.00000787, 0.00001142, 0.00000862, 0.00000029, -0.00000623, -0.00000515, 0.00000001, 0.00000205, },
		{ 0.00000109, 0.00000299, 0.00000210, -0.00000118, -0.00000615, -0.00000967, -0.00000709, 0.00000009, 0.00000389, 0.00000124, -0.00000184, -0.00000123, -0.00000030, -0.00000113, -0.00000089, 0.00000007, -0.00000169, -0.00000356, -0.00000138, -0.00000088, -0.00000761, -0.00001081, 0.00000043, 0.00001249, 0.00000609, -0.00000865, -0.00000858, 0.00000127, -0.00000034, -0.00000971, -0.00000891, -0.00000235, -0.00000211, -0.00000069, 0.00000077, -0.00002205, -0.00005529, -0.00003588, 0.00003262, 0.00004169, -0.00004964, -0.00011090, -0.00004087, 0.00004304, -0.00001892, -0.00018438, -0.00032087, -0.00038935, -0.00028368, 0.00024680, 0.00109682, 0.00148067, 0.00075650, -0.00049398, -0.00096381, -0.00032867, 0.00033307, 0.00014421, -0.00037396, -0.00030450, 0.00018929, 0.00026667, -0.00015136, -0.00033348, -0.00000725, 0.00025934, 0.00007234, -0.00019556, -0.00011936, 0.00010784, 0.00010359, -0.00006787, -0.00009974, 0.00002225, 0.00007402, 0.00000202, -0.00004883, -0.00001908, 0.00001108, 0.00000188, -0.00000487, 0.00000340, -0.00000143, -0.00001265, -0.00000367, 0.00000939, -0.00000193, -0.00001871, -0.00000941, 0.00001014, 0.00000731, -0.00000957, -0.00001223, -0.00000224, 0.00000154, -0.00000004, 0.00000331, 0.00000588, -0.00000098, -0.00000857, -0.00000584, 0.00000166, 0.00000401, 0.00000289, 0.00000294, 0.00000200, -0.00000058, 0.00000036, 0.00000458, 0.00000413, -0.00000244, -0.00000643, -0.00000376, -0.00000118, -0.00000349, -0.00000559, -0.00000185, 0.00000544, 0.00001052, 0.00001027, 0.00000470, -0.00000297, -0.00000665, -0.00000353, 0.00000129, 0.00000157, -0.00000127, -0.00000163, },
		{ 0.00000133, -0.00000348, -0.00000782, -0.00000951, -0.00000589, 0.00000103, 0.00000420, 0.00000126, -0.00000179, -0.00000090, 0.00000017, -0.00000129, -0.00000190, -0.00000046, -0.00000057, -0.00000209, -0.00000175, -0.00000250, -0.00000771, -0.00000891, 0.00000167, 0.00001135, 0.00000431, -0.00000847, -0.00000625, 0.00000318, -0.00000198, -0.00001477, -0.00001178, 0.00000232, 0.00000519, -0.00000308, -0.00001264, -0.00002937, -0.00004310, -0.00001772, 0.00003038, 0.00001701, -0.00006675, -0.00009401, -0.00000387, 0.00004973, -0.00008274, -0.00028901, -0.00035303, -0.00024206, -0.00000388, 0.00043201, 0.00098489, 0.00114246, 0.00050240, -0.00044945, -0.00074267, -0.00022005, 0.00023559, 0.00002298, -0.00034306, -0.00018523, 0.00021763, 0.00017637, -0.00021593, -0.00028144, 0.00007581, 0.00025407, -0.00000093, -0.00022366, -0.00006892, 0.00015465, 0.00008239, -0.00011567, -0.00010282, 0.00005956, 0.00009168, -0.00002075, -0.00007233, -0.00001106, 0.00003249, 0.00000465, -0.00002015, -0.00000610, 0.00000482, -0.00000244, -0.00000186, 0.00000363, -0.00000729, -0.00001870, -0.00000537, 0.00001296, 0.00000530, -0.00001387, -0.00001308, 0.00000152, 0.00000504, -0.00000066, 0.00000046, 0.00000445, -0.00000045, -0.00000742, -0.00000427, 0.00000364, 0.00000491, 0.00000150, 0.00000068, 0.00000139, 0.00000090, 0.00000193, 0.00000422, 0.00000179, -0.00000504, -0.00000750, -0.00000320, -0.00000047, -0.00000337, -0.00000468, 0.00000130, 0.00000936, 0.00001160, 0.00000710, 0.00000006, -0.00000518, -0.00000554, -0.00000144, 0.00000186, 0.00000037, -0.00000253, -0.00000130, 0.00000287, 0.00000432, },
		{ -0.00000865, -0.00000843, -0.00000442, 0.00000160, 0.00000411, 0.00000129, -0.00000152, -0.00000068, 0.00000016, -0.00000164, -0.00000240, -0.00000032, 0.00000047, -0.00000148, -0.00000266, -0.00000351, -0.00000673, -0.00000688, 0.00000171, 0.00000919, 0.00000328, -0.00000652, -0.00000382, 0.00000261, -0.00000569, -0.00001798, -0.00001041, 0.00000771, 0.00000729, -0.00001064, -0.00002324, -0.00002733, -0.00002648, -0.00000796, 0.00001517, -0.00000751, -0.00006592, -0.00006194, 0.00001921, 0.00001757, -0.00016642, -0.00035325, -0.00030303, -0.00004959, 0.00022204, 0.00049757, 0.00079723, 0.00084529, 0.00035839, -0.00034276, -0.00056067, -0.00018624, 0.00012295, -0.00004682, -0.00027248, -0.00008999, 0.00019819, 0.00008266, -0.00023899, -0.00020419, 0.00013507, 0.00021385, -0.00007161, -0.00022023, -0.00000497, 0.00018114, 0.00003971, -0.00015679, -0.00008472, 0.00010129, 0.00009372, -0.00005466, -0.00008815, 0.00000994, 0.00005388, -0.00000331, -0.00004079, -0.00000950, 0.00001875, 0.00000701, -0.00000665, -0.00000532, -0.00000938, -0.00001437, -0.00000133, 0.00001246, 0.00000135, -0.00001670, -0.00001132, 0.00000599, 0.00000694, -0.00000310, -0.00000268, 0.00000406, 0.00000120, -0.00000577, -0.00000272, 0.00000520, 0.00000497, -0.00000068, -0.00000156, 0.00000165, 0.00000304, 0.00000298, 0.00000266, -0.00000107, -0.00000693, -0.00000760, -0.00000246, -0.00000008, -0.00000296, -0.00000261, 0.00000527, 0.00001242, 0.00001047, 0.00000225, -0.00000429, -0.00000567, -0.00000313, 0.00000043, 0.00000140, -0.00000119, -0.00000304, -0.00000006, 0.00000459, 0.00000461, -0.00000044, -0.00000582, },
		{ -0.00000306, 0.00000171, 0.00000377, 0.00000140, -0.00000118, -0.00000069, -0.00000012, -0.00000182, -0.00000234, 0.00000007, 0.00000099, -0.00000155, -0.00000346, -0.00000365, -0.00000522, -0.00000550, 0.00000066, 0.00000703, 0.00000355, -0.00000375, -0.00000269, -0.00000025, -0.00000941, -0.00001801, -0.00000650, 0.00001070, 0.00000392, -0.00001933, -0.00002725, -0.00001810, -0.00001237, -0.00000969, -0.00000638, -0.00002199, -0.00004778, -0.00002949, 0.00001512, -0.00004480, -0.00024094, -0.00035690, -0.00018784, 0.00014243, 0.00036371, 0.00046259, 0.00058606, 0.00062139, 0.00030544, -0.00021859, -0.00043676, -0.00020409, 0.00002590, -0.00006551, -0.00018850, -0.00003358, 0.00014587, 0.00000736, -0.00022149, -0.00012199, 0.00015975, 0.00015058, -0.00012345, -0.00018672, 0.00005753, 0.00018050, -0.00001514, -0.00018045, -0.00004783, 0.00013698, 0.00007681, -0.00009244, -0.00008973, 0.00004103, 0.00006814, -0.00002323, -0.00006135, -0.00000278, 0.00003792, 0.00001094, -0.00001872, -0.00001390, -0.00000558, -0.00000706, -0.00000016, 0.00000809, -0.00000261, -0.00001644, -0.00000749, 0.00000948, 0.00000655, -0.00000660, -0.00000505, 0.00000503, 0.00000360, -0.00000404, -0.00000146, 0.00000613, 0.00000414, -0.00000338, -0.00000331, 0.00000294, 0.00000543, 0.00000304, -0.00000005, -0.00000387, -0.00000776, -0.00000695, -0.00000191, 0.00000010, -0.00000175, 0.00000075, 0.00000931, 0.00001362, 0.00000696, -0.00000339, -0.00000727, -0.00000421, -0.00000007, 0.00000142, -0.00000003, -0.00000256, -0.00000250, 0.00000174, 0.00000565, 0.00000367, -0.00000278, -0.00000762, -0.00000844, -0.00000673, },
		{ 0.00000330, 0.00000156, -0.00000090, -0.00000088, -0.00000043, -0.00000169, -0.00000193, 0.00000032, 0.00000095, -0.00000187, -0.00000374, -0.00000310, -0.00000386, -0.00000502, -0.00000083, 0.00000563, 0.00000483, -0.00000138, -0.00000342, -0.00000412, -0.00001140, -0.00001516, -0.00000261, 0.00000961, -0.00000279, -0.00002505, -0.00002403, -0.00000661, -0.00000594, -0.00002078, -0.00002548, -0.00002169, -0.00002059, -0.00001141, -0.00001870, -0.00011801, -0.00028062, -0.00029764, -0.00003917, 0.00029443, 0.00041412, 0.00036322, 0.00039590, 0.00047913, 0.00031054, -0.00011171, -0.00036991, -0.00024431, -0.00003759, -0.00004704, -0.00011492, -0.00001724, 0.00008161, -0.00003626, -0.00017503, -0.00005384, 0.00014941, 0.00008128, -0.00014645, -0.00013273, 0.00010440, 0.00015352, -0.00006925, -0.00018051, -0.00000020, 0.00015714, 0.00004311, -0.00012471, -0.00007414, 0.00007555, 0.00006928, -0.00005242, -0.00007518, 0.00001527, 0.00005725, 0.00000546, -0.00003596, -0.00001769, 0.00000489, 0.00000007, -0.00000424, 0.00000100, -0.00000395, -0.00001245, -0.00000331, 0.00001040, 0.00000410, -0.00000980, -0.00000587, 0.00000708, 0.00000619, -0.00000254, -0.00000062, 0.00000628, 0.00000245, -0.00000615, -0.00000410, 0.00000515, 0.00000747, 0.00000174, -0.00000353, -0.00000600, -0.00000740, -0.00000598, -0.00000178, 0.00000048, 0.00000074, 0.00000504, 0.00001233, 0.00001222, 0.00000153, -0.00000855, -0.00000808, -0.00000113, 0.00000272, 0.00000112, -0.00000200, -0.00000313, -0.00000091, 0.00000349, 0.00000554, 0.00000165, -0.00000508, -0.00000844, -0.00000728, -0.00000480, -0.00000207, 0.00000141, },
		{ -0.00000068, -0.00000110, -0.00000057, -0.00000130, -0.00000148, 0.00000023, 0.00000055, -0.00000201, -0.00000342, -0.00000234, -0.00000308, -0.00000521, -0.00000205, 0.00000519, 0.00000631, -0.00000030, -0.00000555, -0.00000746, -0.00001101, -0.00001100, -0.00000092, 0.00000482, -0.00000951, -0.00002545, -0.00001602, 0.00000179, -0.00000856, -0.00003527, -0.00003505, -0.00000868, 0.00000346, -0.00001673, -0.00007344, -0.00017869, -0.00027093, -0.00019036, 0.00010650, 0.00038223, 0.00038767, 0.00024079, 0.00025462, 0.00040642, 0.00033830, -0.00004066, -0.00034459, -0.00028105, -0.00006438, -0.00001194, -0.00006657, -0.00003076, 0.00002523, -0.00004611, -0.00011708, -0.00001205, 0.00011308, 0.00002252, -0.00013937, -0.00007264, 0.00012655, 0.00010814, -0.00011022, -0.00015744, 0.00004690, 0.00015631, -0.00000021, -0.00014293, -0.00004335, 0.00010494, 0.00005463, -0.00008468, -0.00007665, 0.00004222, 0.00007048, -0.00001075, -0.00005376, -0.00001300, 0.00002025, 0.00000302, -0.00001438, -0.00000599, -0.00000041, -0.00000569, -0.00000117, 0.00000797, 0.00000089, -0.00001131, -0.00000502, 0.00000947, 0.00000845, -0.00000135, -0.00000029, 0.00000551, 0.00000014, -0.00000842, -0.00000366, 0.00000784, 0.00000851, -0.00000094, -0.00000714, -0.00000696, -0.00000607, -0.00000516, -0.00000197, 0.00000165, 0.00000466, 0.00000942, 0.00001319, 0.00000805, -0.00000470, -0.00001193, -0.00000648, 0.00000269, 0.00000438, -0.00000043, -0.00000382, -0.00000251, 0.00000131, 0.00000449, 0.00000406, -0.00000095, -0.00000671, -0.00000810, -0.00000549, -0.00000304, -0.00000155, 0.00000083, 0.00000280, 0.00000171, },
		{ -0.00000054, -0.00000086, -0.00000123, -0.00000013, 0.00000015, -0.00000178, -0.00000276, -0.00000181, -0.00000294, -0.00000554, -0.00000258, 0.00000532, 0.00000713, -0.00000064, -0.00000800, -0.00000916, -0.00000880, -0.00000749, -0.00000225, -0.00000164, -0.00001334, -0.00002076, -0.00000739, 0.00000374, -0.00001757, -0.00004624, -0.00003266, 0.00000931, 0.00001325, -0.00004537, -0.00013229, -0.00020738, -0.00021223, -0.00006139, 0.00021829, 0.00040090, 0.00031272, 0.00013055, 0.00017039, 0.00037902, 0.00036075, -0.00000729, -0.00033912, -0.00029853, -0.00006332, 0.00002044, -0.00004675, -0.00005786, -0.00001033, -0.00003011, -0.00006488, 0.00000054, 0.00006537, -0.00001443, -0.00010928, -0.00002091, 0.00012236, 0.00005669, -0.00012984, -0.00011800, 0.00008250, 0.00013484, -0.00004308, -0.00014222, -0.00000390, 0.00012144, 0.00002606, -0.00011204, -0.00006323, 0.00007228, 0.00007222, -0.00003560, -0.00006630, 0.00000159, 0.00003639, -0.00000167, -0.00002906, -0.00000916, 0.00000887, 0.00000134, -0.00000328, 0.00000278, -0.00000105, -0.00001023, -0.00000324, 0.00001129, 0.00001019, -0.00000033, -0.00000058, 0.00000375, -0.00000237, -0.00000959, -0.00000204, 0.00001029, 0.00000804, -0.00000463, -0.00001013, -0.00000656, -0.00000427, -0.00000479, -0.00000206, 0.00000415, 0.00000962, 0.00001263, 0.00001104, 0.00000174, -0.00001021, -0.00001261, -0.00000297, 0.00000608, 0.00000440, -0.00000267, -0.00000472, -0.00000071, 0.00000346, 0.00000417, 0.00000142, -0.00000343, -0.00000719, -0.00000672, -0.00000353, -0.00000174, -0.00000150, 0.00000008, 0.00000229, 0.00000185, -0.00000047, -0.00000119, },
		{ -0.00000120, -0.00000053, 0.00000001, -0.00000124, -0.00000213, -0.00000172, -0.00000313, -0.00000556, -0.00000248, 0.00000537, 0.00000683, -0.00000187, -0.00000968, -0.00000904, -0.00000611, -0.00000600, -0.00000588, -0.00000720, -0.00001304, -0.00001344, -0.00000207, -0.00000091, -0.00002778, -0.00004888, -0.00002139, 0.00002227, 0.00000317, -0.00008832, -0.00017645, -0.00019448, -0.00011851, 0.00006005, 0.00027840, 0.00036277, 0.00022152, 0.00005418, 0.00013475, 0.00037034, 0.00036278, -0.00000134, -0.00033401, -0.00029296, -0.00004927, 0.00003766, -0.00004889, -0.00008249, -0.00002170, -0.00000192, -0.00003023, -0.00001007, 0.00002130, -0.00002629, -0.00006847, 0.00001229, 0.00009753, 0.00001182, -0.00012641, -0.00007291, 0.00009929, 0.00009871, -0.00007566, -0.00012290, 0.00003504, 0.00012057, -0.00001037, -0.00012725, -0.00003660, 0.00009795, 0.00005991, -0.00006380, -0.00006843, 0.00002435, 0.00004795, -0.00001544, -0.00004454, -0.00000527, 0.00002262, 0.00000520, -0.00001066, -0.00000312, 0.00000027, -0.00000674, -0.00000191, 0.00001189, 0.00001170, 0.00000069, -0.00000172, 0.00000105, -0.00000443, -0.00000923, 0.00000032, 0.00001166, 0.00000593, -0.00000860, -0.00001184, -0.00000499, -0.00000253, -0.00000483, -0.00000141, 0.00000816, 0.00001465, 0.00001339, 0.00000578, -0.00000536, -0.00001346, -0.00001039, 0.00000132, 0.00000799, 0.00000288, -0.00000471, -0.00000424, 0.00000178, 0.00000469, 0.00000233, -0.00000177, -0.00000507, -0.00000636, -0.00000471, -0.00000185, -0.00000105, -0.00000180, -0.00000071, 0.00000180, 0.00000200, -0.00000015, -0.00000112, -0.00000046, -0.00000059, },
		{ 0.00000014, -0.00000068, -0.00000180, -0.00000196, -0.00000326, -0.00000512, -0.00000215, 0.00000479, 0.00000554, -0.00000309, -0.00000998, -0.00000778, -0.00000432, -0.00000672, -0.00001002, -0.00001001, -0.00000942, -0.00000683, -0.00000194, -0.00000926, -0.00003394, -0.00004256, -0.00000829, 0.00002212, -0.00002472, -0.00013096, -0.00019146, -0.00014275, -0.00001212, 0.00015036, 0.00028478, 0.00029084, 0.00014101, 0.00001759, 0.00012961, 0.00035956, 0.00034249, -0.00000739, -0.00031748, -0.00027025, -0.00003650, 0.00003659, -0.00006103, -0.00009403, -0.00001409, 0.00002446, -0.00001685, -0.00003242, -0.00000837, -0.00001759, -0.00003002, 0.00002357, 0.00006252, -0.00001731, -0.00010465, -0.00003333, 0.00009564, 0.00005745, -0.00009137, -0.00009038, 0.00006455, 0.00010271, -0.00004615, -0.00012614, -0.00000244, 0.00011234, 0.00003495, -0.00008837, -0.00005758, 0.00005052, 0.00005009, -0.00003681, -0.00005589, 0.00000720, 0.00003751, 0.00000275, -0.00002244, -0.00000675, 0.00000603, -0.00000225, -0.00000251, 0.00001126, 0.00001361, 0.00000171, -0.00000405, -0.00000228, -0.00000532, -0.00000725, 0.00000258, 0.00001124, 0.00000251, -0.00001192, -0.00001198, -0.00000279, -0.00000127, -0.00000485, 0.00000059, 0.00001324, 0.00001836, 0.00001080, -0.00000181, -0.00001144, -0.00001344, -0.00000598, 0.00000496, 0.00000786, 0.00000055, -0.00000564, -0.00000240, 0.00000409, 0.00000439, -0.00000072, -0.00000468, -0.00000535, -0.00000443, -0.00000260, -0.00000076, -0.00000092, -0.00000228, -0.00000145, 0.00000139, 0.00000225, 0.00000032, -0.00000096, -0.00000050, -0.00000055, -0.00000124, -0.00000071, },
		{ -0.00000180, -0.00000225, -0.00000308, -0.00000439, -0.00000210, 0.00000342, 0.00000383, -0.00000361, -0.00000906, -0.00000645, -0.00000412, -0.00000874, -0.00001285, -0.00000971, -0.00000474, -0.00000353, -0.00000626, -0.00001700, -0.00003320, -0.00003092, -0.00000133, 0.00000600, -0.00006144, -0.00015831, -0.00017178, -0.00006554, 0.00008355, 0.00019724, 0.00024885, 0.00021041, 0.00008692, 0.00001373, 0.00013526, 0.00033597, 0.00030738, -0.00001171, -0.00028705, -0.00024105, -0.00003394, 0.00002194, -0.00007128, -0.00008958, 0.00000229, 0.00003934, -0.00002082, -0.00005426, -0.00001970, 0.00000208, -0.00000365, 0.00001642, 0.00002879, -0.00002741, -0.00007361, -0.00000744, 0.00007565, 0.00002104, -0.00008872, -0.00005328, 0.00007865, 0.00007286, -0.00007297, -0.00010902, 0.00003127, 0.00011138, 0.00000254, -0.00010272, -0.00003483, 0.00007363, 0.00004027, -0.00006148, -0.00005850, 0.00002704, 0.00004894, -0.00000765, -0.00003580, -0.00000515, 0.00001584, 0.00000098, -0.00000604, 0.00001022, 0.00001660, 0.00000236, -0.00000790, -0.00000564, -0.00000431, -0.00000399, 0.00000370, 0.00000871, -0.00000139, -0.00001376, -0.00001067, -0.00000062, -0.00000053, -0.00000415, 0.00000412, 0.00001832, 0.00001933, 0.00000482, -0.00001011, -0.00001478, -0.00001008, -0.00000080, 0.00000670, 0.00000589, -0.00000149, -0.00000492, 0.00000017, 0.00000529, 0.00000240, -0.00000425, -0.00000650, -0.00000418, -0.00000191, -0.00000091, -0.00000041, -0.00000121, -0.00000279, -0.00000207, 0.00000111, 0.00000264, 0.00000090, -0.00000089, -0.00000079, -0.00000066, -0.00000113, -0.00000058, 0.00000037, -0.00000035, },
		{ -0.00000262, -0.00000378, -0.00000261, 0.00000156, 0.00000234, -0.00000321, -0.00000759, -0.00000589, -0.00000530, -0.00001064, -0.00001341, -0.00000743, -0.00000148, -0.00000425, -0.00001220, -0.00002050, -0.00002625, -0.00002015, -0.00000597, -0.00002261, -0.00009449, -0.00016012, -0.00012198, 0.00001813, 0.00015081, 0.00020121, 0.00018974, 0.00014168, 0.00006249, 0.00002847, 0.00013686, 0.00029895, 0.00026861, -0.00000657, -0.00024753, -0.00021542, -0.00004332, 0.00000259, -0.00007204, -0.00007312, 0.00001687, 0.00003959, -0.00003347, -0.00006667, -0.00001554, 0.00002199, 0.00000674, -0.00000102, 0.00000506, -0.00002124, -0.00004335, 0.00000171, 0.00004733, -0.00000315, -0.00007138, -0.00002065, 0.00007602, 0.00003902, -0.00008531, -0.00008059, 0.00005665, 0.00009522, -0.00002985, -0.00010287, -0.00000485, 0.00008748, 0.00001932, -0.00008345, -0.00004981, 0.00005039, 0.00005249, -0.00002524, -0.00004664, 0.00000347, 0.00002761, 0.00000052, -0.00001238, 0.00001018, 0.00002098, 0.00000189, -0.00001334, -0.00000813, -0.00000098, -0.00000031, 0.00000267, 0.00000430, -0.00000463, -0.00001358, -0.00000851, 0.00000094, 0.00000009, -0.00000209, 0.00000877, 0.00002189, 0.00001665, -0.00000358, -0.00001705, -0.00001434, -0.00000431, 0.00000344, 0.00000598, 0.00000299, -0.00000218, -0.00000265, 0.00000244, 0.00000467, -0.00000088, -0.00000726, -0.00000672, -0.00000189, 0.00000054, 0.00000001, -0.00000069, -0.00000167, -0.00000319, -0.00000251, 0.00000101, 0.00000316, 0.00000143, -0.00000103, -0.00000129, -0.00000072, -0.00000071, -0.00000024, 0.00000046, -0.00000037, -0.00000194, -0.00000231, },
		{ -0.00000362, -0.00000030, 0.00000147, -0.00000219, -0.00000632, -0.00000631, -0.00000691, -0.00001130, -0.00001199, -0.00000511, -0.00000108, -0.00000764, -0.00001640, -0.00001849, -0.00001694, -0.00001627, -0.00002289, -0.00005516, -0.00011233, -0.00013402, -0.00005458, 0.00008973, 0.00018181, 0.00017308, 0.00012733, 0.00009549, 0.00006115, 0.00004714, 0.00012756, 0.00025471, 0.00023564, 0.00000875, -0.00020694, -0.00019908, -0.00006046, -0.00001295, -0.00006179, -0.00005245, 0.00002261, 0.00002832, -0.00004542, -0.00006636, -0.00000318, 0.00003408, 0.00000322, -0.00001939, -0.00000491, -0.00000578, -0.00002169, -0.00000358, 0.00001964, -0.00001235, -0.00004657, 0.00000085, 0.00006010, 0.00000956, -0.00008195, -0.00004839, 0.00006839, 0.00006815, -0.00005472, -0.00008871, 0.00002541, 0.00008818, -0.00000854, -0.00009682, -0.00003039, 0.00007165, 0.00004549, -0.00004671, -0.00005077, 0.00001886, 0.00003799, -0.00000530, -0.00002016, 0.00001268, 0.00002638, -0.00000064, -0.00002001, -0.00000864, 0.00000462, 0.00000250, -0.00000119, -0.00000111, -0.00000606, -0.00001141, -0.00000633, 0.00000169, 0.00000130, 0.00000161, 0.00001337, 0.00002243, 0.00001033, -0.00001253, -0.00002075, -0.00001015, 0.00000216, 0.00000537, 0.00000312, 0.00000051, -0.00000092, 0.00000040, 0.00000333, 0.00000213, -0.00000451, -0.00000887, -0.00000532, 0.00000088, 0.00000236, 0.00000011, -0.00000136, -0.00000213, -0.00000340, -0.00000272, 0.00000110, 0.00000369, 0.00000176, -0.00000142, -0.00000182, -0.00000053, -0.00000002, 0.00000008, 0.00000023, -0.00000063, -0.00000194, -0.00000206, -0.00000215, -0.00000362, },
		{ 0.00000118, -0.00000113, -0.00000564, -0.00000723, -0.00000798, -0.00001047, -0.00000984, -0.00000435, -0.00000332, -0.00001115, -0.00001668, -0.00001266, -0.00001046, -0.00002245, -0.00004768, -0.00008130, -0.00010821, -0.00008576, 0.00001447, 0.00013644, 0.00017897, 0.00012881, 0.00007654, 0.00007286, 0.00007160, 0.00005940, 0.00010811, 0.00021165, 0.00021309, 0.00002973, -0.00017237, -0.00019225, -0.00007847, -0.00001975, -0.00004422, -0.00003532, 0.00001773, 0.00001228, -0.00005006, -0.00005558, 0.00000913, 0.00003502, -0.00000805, -0.00003136, -0.00000264, 0.00001075, -0.00001210, -0.00001719, -0.00000042, -0.00000859, -0.00002237, 0.00000874, 0.00003756, -0.00000938, -0.00006617, -0.00002034, 0.00006524, 0.00003726, -0.00006692, -0.00006413, 0.00004888, 0.00007543, -0.00003684, -0.00009774, -0.00000404, 0.00008516, 0.00002809, -0.00006696, -0.00004537, 0.00003849, 0.00004333, -0.00001659, -0.00002701, 0.00001872, 0.00003162, -0.00000605, -0.00002695, -0.00000620, 0.00001176, 0.00000303, -0.00000791, -0.00000617, -0.00000480, -0.00000782, -0.00000496, 0.00000189, 0.00000384, 0.00000657, 0.00001633, 0.00001896, 0.00000152, -0.00001970, -0.00002014, -0.00000342, 0.00000734, 0.00000441, -0.00000072, -0.00000034, 0.00000209, 0.00000300, 0.00000209, -0.00000173, -0.00000734, -0.00000859, -0.00000279, 0.00000335, 0.00000322, -0.00000044, -0.00000215, -0.00000242, -0.00000340, -0.00000269, 0.00000130, 0.00000408, 0.00000178, -0.00000197, -0.00000216, -0.00000001, 0.00000077, 0.00000015, -0.00000031, -0.00000090, -0.00000165, -0.00000165, -0.00000199, -0.00000398, -0.00000480, -0.00000173, },
		{ -0.00000538, -0.00000787, -0.00000800, -0.00000886, -0.00000842, -0.00000563, -0.00000655, -0.00001241, -0.00001310, -0.00000675, -0.00001094, -0.00003765, -0.00007255, -0.00009255, -0.00008222, -0.00002675, 0.00007084, 0.00015384, 0.00015233, 0.00008388, 0.00004447, 0.00006762, 0.00008286, 0.00006127, 0.00008393, 0.00017627, 0.00020028, 0.00005007, -0.00014719, -0.00019092, -0.00009123, -0.00001767, -0.00002560, -0.00002648, 0.00000526, -0.00000147, -0.00004533, -0.00004007, 0.00001528, 0.00002645, -0.00001982, -0.00003375, 0.00000663, 0.00002182, -0.00001336, -0.00003182, -0.00000968, 0.00000266, -0.00000509, 0.00000497, 0.00001585, -0.00001557, -0.00004425, -0.00000228, 0.00005019, 0.00001006, -0.00006508, -0.00003570, 0.00006055, 0.00005275, -0.00005885, -0.00008573, 0.00002320, 0.00008700, 0.00000356, -0.00008060, -0.00003007, 0.00005806, 0.00004091, -0.00003164, -0.00003024, 0.00002822, 0.00003493, -0.00001455, -0.00003269, -0.00000025, 0.00001910, 0.00000009, -0.00001676, -0.00000930, -0.00000056, -0.00000385, -0.00000493, 0.00000222, 0.00000816, 0.00001169, 0.00001600, 0.00001152, -0.00000765, -0.00002304, -0.00001540, 0.00000380, 0.00000973, 0.00000108, -0.00000395, 0.00000106, 0.00000589, 0.00000390, -0.00000127, -0.00000579, -0.00000841, -0.00000652, 0.00000003, 0.00000489, 0.00000315, -0.00000129, -0.00000281, -0.00000251, -0.00000317, -0.00000246, 0.00000151, 0.00000418, 0.00000152, -0.00000245, -0.00000215, 0.00000074, 0.00000140, -0.00000013, -0.00000098, -0.00000094, -0.00000111, -0.00000138, -0.00000231, -0.00000467, -0.00000574, -0.00000261, 0.00000109, -0.00000042, },
		{ -0.00000730, -0.00000761, -0.00000852, -0.00000806, -0.00000866, -0.00001052, -0.00000793, -0.00000480, -0.00001960, -0.00005700, -0.00008931, -0.00008517, -0.00004072, 0.00003000, 0.00010571, 0.00014557, 0.00011500, 0.00004906, 0.00003057, 0.00007069, 0.00008788, 0.00005441, 0.00006143, 0.00015106, 0.00019302, 0.00006477, -0.00013058, -0.00018959, -0.00009580, -0.00001035, -0.00001167, -0.00002639, -0.00000924, -0.00000831, -0.00003353, -0.00002614, 0.00001309, 0.00001320, -0.00002651, -0.00002768, 0.00001655, 0.00002446, -0.00002094, -0.00004162, -0.00000918, 0.00001484, 0.00000248, -0.00000541, 0.00000077, -0.00001089, -0.00002327, 0.00000361, 0.00002919, -0.00000778, -0.00005188, -0.00001055, 0.00005898, 0.00002626, -0.00006967, -0.00006392, 0.00004484, 0.00007635, -0.00002257, -0.00008360, -0.00000736, 0.00007270, 0.00002995, -0.00004722, -0.00002779, 0.00003978, 0.00003451, -0.00002547, -0.00003556, 0.00000898, 0.00002482, -0.00000695, -0.00002626, -0.00000908, 0.00000623, -0.00000069, -0.00000629, 0.00000353, 0.00001407, 0.00001531, 0.00001131, 0.00000138, -0.00001462, -0.00002144, -0.00000803, 0.00000928, 0.00000879, -0.00000318, -0.00000520, 0.00000438, 0.00000901, 0.00000237, -0.00000598, -0.00000872, -0.00000727, -0.00000336, 0.00000224, 0.00000527, 0.00000248, -0.00000209, -0.00000324, -0.00000241, -0.00000274, -0.00000209, 0.00000157, 0.00000391, 0.00000110, -0.00000265, -0.00000176, 0.00000149, 0.00000167, -0.00000068, -0.00000151, -0.00000067, -0.00000055, -0.00000149, -0.00000306, -0.00000535, -0.00000625, -0.00000316, 0.00000085, -0.00000009, -0.00000501, -0.00000771, },
		{ -0.00000980, -0.00000997, -0.00000830, -0.00000654, -0.00000457, -0.00000924, -0.00003414, -0.00007372, -0.00009226, -0.00006109, 0.00000627, 0.00007391, 0.00011729, 0.00012061, 0.00007877, 0.00002867, 0.00002931, 0.00007398, 0.00008466, 0.00004347, 0.00004504, 0.00013460, 0.00018632, 0.00007200, -0.00011884, -0.00018391, -0.00009299, -0.00000294, -0.00000540, -0.00003195, -0.00002040, -0.00000739, -0.00001945, -0.00001814, 0.00000427, 0.00000087, -0.00002595, -0.00001727, 0.00002215, 0.00001953, -0.00002922, -0.00004391, -0.00000297, 0.00002273, 0.00000134, -0.00001637, -0.00000517, -0.00000016, -0.00000873, -0.00000092, 0.00000897, -0.00001408, -0.00003286, 0.00000587, 0.00004657, 0.00000278, -0.00006773, -0.00003806, 0.00005599, 0.00005587, -0.00004406, -0.00007463, 0.00001799, 0.00007847, 0.00001204, -0.00005947, -0.00001901, 0.00005095, 0.00002917, -0.00003721, -0.00003414, 0.00002034, 0.00002706, -0.00001779, -0.00003443, -0.00000465, 0.00001447, 0.00000060, -0.00000857, 0.00000650, 0.00002069, 0.00001572, 0.00000221, -0.00000916, -0.00001712, -0.00001522, -0.00000041, 0.00001132, 0.00000524, -0.00000653, -0.00000385, 0.00000842, 0.00001007, -0.00000149, -0.00001065, -0.00000948, -0.00000422, -0.00000016, 0.00000321, 0.00000463, 0.00000167, -0.00000260, -0.00000345, -0.00000218, -0.00000212, -0.00000168, 0.00000134, 0.00000333, 0.00000077, -0.00000243, -0.00000114, 0.00000197, 0.00000151, -0.00000128, -0.00000174, -0.00000022, -0.00000023, -0.00000205, -0.00000397, -0.00000577, -0.00000641, -0.00000370, 0.00000034, 0.00000028, -0.00000399, -0.00000684, -0.00000677, -0.00000763, },
		{ -0.00000568, -0.00000303, -0.00000584, -0.00001966, -0.00004964, -0.00008171, -0.00008013, -0.00002677, 0.00004856, 0.00009946, 0.00010983, 0.00008938, 0.00005108, 0.00002132, 0.00003359, 0.00007283, 0.00007514, 0.00003327, 0.00003591, 0.00012321, 0.00017676, 0.00007312, -0.00010764, -0.00017234, -0.00008629, 0.00000038, -0.00000630, -0.00003848, -0.00002516, -0.00000116, -0.00000797, -0.00001719, -0.00000710, -0.00000650, -0.00001936, -0.00000740, 0.00002136, 0.00001060, -0.00003375, -0.00003942, 0.00000393, 0.00002395, -0.00000474, -0.00002312, -0.00000304, 0.00001087, -0.00000301, -0.00001128, -0.00000516, -0.00001039, -0.00001442, 0.00001136, 0.00002852, -0.00001238, -0.00005514, -0.00001474, 0.00005479, 0.00003086, -0.00005591, -0.00005561, 0.00004028, 0.00007360, -0.00000916, -0.00006509, -0.00000507, 0.00005887, 0.00001885, -0.00004749, -0.00002780, 0.00003188, 0.00002441, -0.00003111, -0.00003920, 0.00000402, 0.00002258, -0.00000059, -0.00001093, 0.00001135, 0.00002652, 0.00001167, -0.00001009, -0.00001731, -0.00001394, -0.00000609, 0.00000488, 0.00000944, 0.00000084, -0.00000741, -0.00000040, 0.00001145, 0.00000833, -0.00000663, -0.00001379, -0.00000775, -0.00000019, 0.00000207, 0.00000279, 0.00000347, 0.00000113, -0.00000278, -0.00000356, -0.00000183, -0.00000138, -0.00000139, 0.00000076, 0.00000258, 0.00000073, -0.00000178, -0.00000054, 0.00000201, 0.00000107, -0.00000167, -0.00000162, 0.00000018, -0.00000034, -0.00000290, -0.00000475, -0.00000588, -0.00000649, -0.00000439, -0.00000012, 0.00000119, -0.00000226, -0.00000596, -0.00000732, -0.00000892, -0.00001099, -0.00000987, },
		{ -0.00001257, -0.00003293, -0.00006050, -0.00007775, -0.00005612, 0.00000932, 0.00007884, 0.00010679, 0.00009099, 0.00006045, 0.00003425, 0.00002236, 0.00003768, 0.00006640, 0.00006309, 0.00002668, 0.00003242, 0.00011315, 0.00016350, 0.00007131, -0.00009405, -0.00015623, -0.00007979, -0.00000205, -0.00001134, -0.00004191, -0.00002355, 0.00000628, -0.00000208, -0.00002136, -0.00001665, -0.00000771, -0.00001023, -0.00000170, 0.00001516, 0.00000198, -0.00003266, -0.00003130, 0.00000757, 0.00001925, -0.00001116, -0.00002357, 0.00000355, 0.00001767, -0.00000511, -0.00002199, -0.00001086, -0.00000102, -0.00000168, 0.00000731, 0.00001098, -0.00001695, -0.00003685, 0.00000072, 0.00004298, 0.00000770, -0.00005583, -0.00003115, 0.00005454, 0.00005908, -0.00002894, -0.00006239, 0.00001126, 0.00006118, 0.00000489, -0.00005399, -0.00001702, 0.00004119, 0.00001637, -0.00004476, -0.00003889, 0.00001600, 0.00002887, -0.00000430, -0.00001244, 0.00001767, 0.00002989, 0.00000286, -0.00002329, -0.00002057, -0.00000536, 0.00000320, 0.00000595, 0.00000455, -0.00000222, -0.00000518, 0.00000365, 0.00001193, 0.00000408, -0.00001139, -0.00001431, -0.00000403, 0.00000355, 0.00000271, 0.00000132, 0.00000241, 0.00000108, -0.00000276, -0.00000365, -0.00000136, -0.00000056, -0.00000132, -0.00000014, 0.00000189, 0.00000113, -0.00000089, -0.00000023, 0.00000157, 0.00000057, -0.00000171, -0.00000132, 0.00000029, -0.00000085, -0.00000377, -0.00000527, -0.00000588, -0.00000671, -0.00000506, -0.00000005, 0.00000285, -0.00000040, -0.00000602, -0.00000924, -0.00001052, -0.00001057, -0.00000739, -0.00000259, -0.00000277, },
		{ -0.00006266, -0.00006249, -0.00002639, 0.00003964, 0.00009428, 0.00010028, 0.00006870, 0.00003862, 0.00002660, 0.00002658, 0.00003868, 0.00005651, 0.00005185, 0.00002412, 0.00003169, 0.00010220, 0.00014799, 0.00006972, -0.00007749, -0.00013864, -0.00007643, -0.00000922, -0.00001666, -0.00004038, -0.00001806, 0.00001119, -0.00000206, -0.00002714, -0.00002156, -0.00000422, -0.00000238, -0.00000136, 0.00000657, -0.00000309, -0.00002678, -0.00002339, 0.00000631, 0.00001154, -0.00001431, -0.00001853, 0.00001019, 0.00001826, -0.00001155, -0.00002866, -0.00000902, 0.00000878, 0.00000305, -0.00000218, -0.00000105, -0.00001214, -0.00001884, 0.00000584, 0.00002514, -0.00000819, -0.00004489, -0.00000719, 0.00005801, 0.00003846, -0.00004296, -0.00005187, 0.00002638, 0.00005679, -0.00001025, -0.00005503, -0.00000351, 0.00004603, 0.00000369, -0.00005611, -0.00003272, 0.00002955, 0.00003191, -0.00000996, -0.00001236, 0.00002444, 0.00002940, -0.00000978, -0.00003451, -0.00001749, 0.00000677, 0.00000975, 0.00000222, -0.00000128, -0.00000214, -0.00000050, 0.00000627, 0.00000910, -0.00000134, -0.00001406, -0.00001203, 0.00000043, 0.00000587, 0.00000177, -0.00000049, 0.00000190, 0.00000142, -0.00000277, -0.00000375, -0.00000075, 0.00000022, -0.00000158, -0.00000122, 0.00000149, 0.00000196, 0.00000000, -0.00000038, 0.00000081, 0.00000025, -0.00000140, -0.00000105, -0.00000000, -0.00000157, -0.00000447, -0.00000561, -0.00000603, -0.00000704, -0.00000523, 0.00000094, 0.00000491, 0.00000067, -0.00000757, -0.00001184, -0.00001094, -0.00000777, -0.00000369, -0.00000156, -0.00000731, -0.00002311, -0.00004443, },
		{ 0.00000230, 0.00005968, 0.00009630, 0.00008617, 0.00004877, 0.00002482, 0.00002452, 0.00003024, 0.00003642, 0.00004582, 0.00004310, 0.00002415, 0.00003121, 0.00009019, 0.00013262, 0.00007011, -0.00005948, -0.00012265, -0.00007692, -0.00001834, -0.00001933, -0.00003447, -0.00001209, 0.00001162, -0.00000592, -0.00003113, -0.00002135, 0.00000098, 0.00000162, -0.00000518, -0.00000104, -0.00000350, -0.00001877, -0.00001858, 0.00000098, 0.00000437, -0.00001280, -0.00001086, 0.00001345, 0.00001345, -0.00001808, -0.00002935, -0.00000296, 0.00001477, 0.00000058, -0.00001203, -0.00000520, -0.00000188, -0.00000613, 0.00000155, 0.00000714, -0.00001392, -0.00002710, 0.00001084, 0.00005088, 0.00001670, -0.00004846, -0.00003610, 0.00003688, 0.00004631, -0.00002359, -0.00005013, 0.00001016, 0.00004496, -0.00001167, -0.00006267, -0.00002105, 0.00004243, 0.00003085, -0.00001659, -0.00001037, 0.00003031, 0.00002444, -0.00002425, -0.00004108, -0.00000815, 0.00001952, 0.00001133, -0.00000532, -0.00000549, 0.00000182, 0.00000484, 0.00000568, 0.00000341, -0.00000593, -0.00001352, -0.00000775, 0.00000410, 0.00000622, -0.00000014, -0.00000187, 0.00000204, 0.00000186, -0.00000293, -0.00000377, 0.00000001, 0.00000083, -0.00000217, -0.00000228, 0.00000152, 0.00000306, 0.00000061, -0.00000103, -0.00000005, 0.00000027, -0.00000087, -0.00000102, -0.00000063, -0.00000228, -0.00000496, -0.00000601, -0.00000641, -0.00000708, -0.00000441, 0.00000271, 0.00000645, 0.00000010, -0.00001029, -0.00001373, -0.00000901, -0.00000315, -0.00000101, -0.00000460, -0.00001611, -0.00003384, -0.00004993, -0.00005503, -0.00003986, },
		{ 0.00008901, 0.00007008, 0.00003395, 0.00001741, 0.00002453, 0.00003177, 0.00003229, 0.00003632, 0.00003678, 0.00002487, 0.00002987, 0.00007835, 0.00011939, 0.00007234, -0.00004244, -0.00011004, -0.00007998, -0.00002645, -0.00001836, -0.00002650, -0.00000843, 0.00000784, -0.00001068, -0.00003143, -0.00001760, 0.00000487, 0.00000124, -0.00001055, -0.00000534, -0.00000028, -0.00001163, -0.00001779, -0.00000596, 0.00000031, -0.00000751, -0.00000394, 0.00001202, 0.00000602, -0.00002137, -0.00002480, 0.00000325, 0.00001506, -0.00000589, -0.00001807, -0.00000220, 0.00000882, -0.00000127, -0.00000847, -0.00000585, -0.00000984, -0.00000798, 0.00001955, 0.00003612, -0.00000128, -0.00004508, -0.00001893, 0.00004063, 0.00003193, -0.00003254, -0.00004030, 0.00002119, 0.00003785, -0.00002691, -0.00006273, -0.00000544, 0.00005236, 0.00002563, -0.00002297, -0.00000670, 0.00003395, 0.00001538, -0.00003791, -0.00004125, 0.00000578, 0.00002968, 0.00000712, -0.00001436, -0.00000592, 0.00000895, 0.00000845, 0.00000101, -0.00000347, -0.00000769, -0.00000971, -0.00000309, 0.00000573, 0.00000484, -0.00000204, -0.00000239, 0.00000256, 0.00000207, -0.00000321, -0.00000357, 0.00000085, 0.00000112, -0.00000302, -0.00000307, 0.00000203, 0.00000419, 0.00000072, -0.00000205, -0.00000076, 0.00000064, -0.00000034, -0.00000132, -0.00000143, -0.00000286, -0.00000539, -0.00000664, -0.00000675, -0.00000628, -0.00000246, 0.00000453, 0.00000645, -0.00000221, -0.00001296, -0.00001340, -0.00000469, 0.00000141, -0.00000167, -0.00001206, -0.00002651, -0.00004076, -0.00004730, -0.00003969, -0.00001535, 0.00002484, 0.00006862, },
		{ 0.00002433, 0.00001387, 0.00002450, 0.00003133, 0.00002787, 0.00002872, 0.00003185, 0.00002497, 0.00002793, 0.00006824, 0.00010895, 0.00007496, -0.00002839, -0.00010082, -0.00008330, -0.00003163, -0.00001467, -0.00001917, -0.00000820, 0.00000176, -0.00001378, -0.00002806, -0.00001286, 0.00000570, -0.00000221, -0.00001481, -0.00000581, 0.00000428, -0.00000741, -0.00002005, -0.00001178, 0.00000012, -0.00000084, -0.00000024, 0.00000680, -0.00000076, -0.00002014, -0.00001769, 0.00000636, 0.00001035, -0.00001234, -0.00001834, 0.00000484, 0.00001570, -0.00000374, -0.00001922, -0.00001092, 0.00000080, 0.00000718, 0.00001846, 0.00001845, -0.00001204, -0.00003483, -0.00000433, 0.00003732, 0.00001678, -0.00003563, -0.00002782, 0.00002732, 0.00002604, -0.00003899, -0.00005584, 0.00001163, 0.00005752, 0.00001698, -0.00002804, -0.00000202, 0.00003446, 0.00000360, -0.00004812, -0.00003470, 0.00002143, 0.00003468, -0.00000204, -0.00002204, -0.00000155, 0.00001725, 0.00000819, -0.00000720, -0.00000914, -0.00000538, -0.00000379, 0.00000007, 0.00000485, 0.00000269, -0.00000302, -0.00000209, 0.00000295, 0.00000190, -0.00000339, -0.00000306, 0.00000158, 0.00000099, -0.00000392, -0.00000340, 0.00000289, 0.00000504, 0.00000028, -0.00000323, -0.00000114, 0.00000126, 0.00000003, -0.00000188, -0.00000224, -0.00000335, -0.00000598, -0.00000741, -0.00000655, -0.00000425, 0.00000015, 0.00000528, 0.00000435, -0.00000541, -0.00001392, -0.00001018, 0.00000064, 0.00000345, -0.00000678, -0.00002218, -0.00003482, -0.00004107, -0.00003722, -0.00002084, 0.00000608, 0.00003906, 0.00006862, 0.00007710, 0.00005558, },
		{ 0.00002378, 0.00002978, 0.00002401, 0.00002266, 0.00002730, 0.00002424, 0.00002638, 0.00006073, 0.00010063, 0.00007623, -0.00001809, -0.00009361, -0.00008481, -0.00003363, -0.00001027, -0.00001435, -0.00001077, -0.00000423, -0.00001405, -0.00002256, -0.00000940, 0.00000339, -0.00000652, -0.00001637, -0.00000358, 0.00000784, -0.00000651, -0.00002326, -0.00001467, 0.00000278, 0.00000455, -0.00000046, 0.00000013, -0.00000449, -0.00001523, -0.00001120, 0.00000512, 0.00000326, -0.00001556, -0.00001352, 0.00001189, 0.00001667, -0.00001057, -0.00002626, -0.00000810, 0.00001320, 0.00001493, 0.00000990, 0.00000272, -0.00001449, -0.00002128, 0.00000482, 0.00002851, 0.00000395, -0.00003297, -0.00001557, 0.00002752, 0.00001208, -0.00004539, -0.00004299, 0.00002728, 0.00005696, 0.00000624, -0.00003103, 0.00000275, 0.00003165, -0.00000884, -0.00005295, -0.00002265, 0.00003554, 0.00003327, -0.00001390, -0.00002580, 0.00000705, 0.00002404, 0.00000297, -0.00001705, -0.00001126, 0.00000097, 0.00000219, 0.00000030, 0.00000197, 0.00000110, -0.00000263, -0.00000153, 0.00000272, 0.00000149, -0.00000314, -0.00000228, 0.00000194, 0.00000045, -0.00000461, -0.00000315, 0.00000386, 0.00000536, -0.00000064, -0.00000429, -0.00000112, 0.00000197, 0.00000014, -0.00000263, -0.00000303, -0.00000391, -0.00000676, -0.00000791, -0.00000525, -0.00000113, 0.00000237, 0.00000399, 0.00000050, -0.00000783, -0.00001187, -0.00000475, 0.00000460, 0.00000111, -0.00001557, -0.00003170, -0.00003783, -0.00003441, -0.00002278, -0.00000299, 0.00002158, 0.00004563, 0.00006312, 0.00006415, 0.00004396, 0.00001860, 0.00001224, },
		{ 0.00002073, 0.00001754, 0.00002280, 0.00002326, 0.00002602, 0.00005561, 0.00009309, 0.00007511, -0.00001094, -0.00008661, -0.00008357, -0.00003356, -0.00000709, -0.00001247, -0.00001445, -0.00000834, -0.00001186, -0.00001701, -0.00000835, -0.00000077, -0.00000966, -0.00001512, -0.00000059, 0.00000901, -0.00000789, -0.00002529, -0.00001435, 0.00000623, 0.00000696, -0.00000357, -0.00000541, -0.00000446, -0.00000892, -0.00000766, 0.00000037, -0.00000304, -0.00001425, -0.00000621, 0.00001562, 0.00001222, -0.00001766, -0.00002711, -0.00000004, 0.00002269, 0.00001453, -0.00000203, -0.00000745, -0.00000999, -0.00000844, 0.00000743, 0.00001700, -0.00000438, -0.00002618, -0.00000622, 0.00002235, -0.00000088, -0.00004484, -0.00002649, 0.00003884, 0.00005080, -0.00000491, -0.00003166, 0.00000673, 0.00002606, -0.00001973, -0.00005164, -0.00000751, 0.00004528, 0.00002589, -0.00002546, -0.00002422, 0.00001791, 0.00002679, -0.00000680, -0.00002580, -0.00000832, 0.00000995, 0.00000595, -0.00000287, -0.00000152, 0.00000122, -0.00000111, -0.00000152, 0.00000169, 0.00000130, -0.00000223, -0.00000153, 0.00000169, -0.00000032, -0.00000478, -0.00000240, 0.00000460, 0.00000501, -0.00000183, -0.00000501, -0.00000073, 0.00000260, -0.00000004, -0.00000348, -0.00000385, -0.00000466, -0.00000750, -0.00000756, -0.00000264, 0.00000229, 0.00000299, 0.00000043, -0.00000374, -0.00000774, -0.00000668, 0.00000084, 0.00000478, -0.00000581, -0.00002534, -0.00003710, -0.00003420, -0.00002287, -0.00000803, 0.00001071, 0.00003088, 0.00004691, 0.00005521, 0.00005192, 0.00003494, 0.00001541, 0.00001157, 0.00002254, 0.00002773, },
		{ 0.00001875, 0.00002272, 0.00002687, 0.00005186, 0.00008521, 0.00007163, -0.00000557, -0.00007859, -0.00007989, -0.00003304, -0.00000618, -0.00001269, -0.00001748, -0.00001003, -0.00000859, -0.00001298, -0.00000944, -0.00000498, -0.00001064, -0.00001216, 0.00000142, 0.00000774, -0.00000978, -0.00002495, -0.00001192, 0.00000841, 0.00000614, -0.00000756, -0.00000815, -0.00000165, -0.00000370, -0.00000774, -0.00000562, -0.00000620, -0.00000924, 0.00000036, 0.00001457, 0.00000492, -0.00002143, -0.00002189, 0.00000921, 0.00002625, 0.00000782, -0.00001291, -0.00001065, -0.00000161, 0.00000055, 0.00000436, 0.00000590, -0.00000754, -0.00001780, -0.00000142, 0.00001375, -0.00000994, -0.00003762, -0.00000941, 0.00004449, 0.00004020, -0.00001482, -0.00003009, 0.00000930, 0.00001886, -0.00002722, -0.00004476, 0.00000773, 0.00004900, 0.00001442, -0.00003391, -0.00001746, 0.00002824, 0.00002405, -0.00001918, -0.00003071, -0.00000026, 0.00001911, 0.00000572, -0.00000869, -0.00000386, 0.00000358, 0.00000061, -0.00000280, 0.00000016, 0.00000190, -0.00000070, -0.00000124, 0.00000072, -0.00000095, -0.00000421, -0.00000138, 0.00000475, 0.00000401, -0.00000298, -0.00000521, -0.00000010, 0.00000301, -0.00000049, -0.00000443, -0.00000480, -0.00000547, -0.00000768, -0.00000590, 0.00000084, 0.00000477, 0.00000122, -0.00000459, -0.00000644, -0.00000416, 0.00000025, 0.00000389, -0.00000008, -0.00001547, -0.00003251, -0.00003617, -0.00002509, -0.00001010, 0.00000368, 0.00001939, 0.00003561, 0.00004543, 0.00004663, 0.00004069, 0.00002786, 0.00001405, 0.00001171, 0.00002099, 0.00002518, 0.00001767, 0.00001312, },
		{ 0.00002820, 0.00004839, 0.00007666, 0.00006675, -0.00000068, -0.00006931, -0.00007494, -0.00003334, -0.00000743, -0.00001360, -0.00001878, -0.00000989, -0.00000576, -0.00001102, -0.00001152, -0.00000786, -0.00000963, -0.00000904, 0.00000158, 0.00000507, -0.00001060, -0.00002229, -0.00000910, 0.00000818, 0.00000316, -0.00001042, -0.00000783, 0.00000197, -0.00000121, -0.00001044, -0.00001035, -0.00000552, -0.00000283, 0.00000368, 0.00000947, -0.00000185, -0.00002007, -0.00001295, 0.00001575, 0.00002323, -0.00000168, -0.00001943, -0.00000778, 0.00000703, 0.00000420, -0.00000214, -0.00000235, -0.00000645, -0.00001044, -0.00000132, 0.00000456, -0.00001326, -0.00002555, 0.00000511, 0.00004358, 0.00002706, -0.00002221, -0.00002687, 0.00001023, 0.00001151, -0.00003031, -0.00003400, 0.00002032, 0.00004654, 0.00000161, -0.00003741, -0.00000714, 0.00003527, 0.00001589, -0.00003128, -0.00002990, 0.00001143, 0.00002570, 0.00000091, -0.00001537, -0.00000351, 0.00000774, 0.00000131, -0.00000565, -0.00000112, 0.00000377, 0.00000104, -0.00000190, -0.00000079, -0.00000097, -0.00000285, -0.00000054, 0.00000405, 0.00000257, -0.00000374, -0.00000484, 0.00000059, 0.00000308, -0.00000123, -0.00000555, -0.00000584, -0.00000598, -0.00000673, -0.00000298, 0.00000418, 0.00000515, -0.00000281, -0.00000929, -0.00000592, 0.00000243, 0.00000634, 0.00000224, -0.00000923, -0.00002452, -0.00003403, -0.00002899, -0.00001366, 0.00000031, 0.00001094, 0.00002426, 0.00003798, 0.00004270, 0.00003770, 0.00003023, 0.00002248, 0.00001445, 0.00001255, 0.00001887, 0.00002182, 0.00001474, 0.00000977, 0.00001574, 0.00002275, },
		{ 0.00006798, 0.00006161, 0.00000431, -0.00005944, -0.00006996, -0.00003486, -0.00000990, -0.00001399, -0.00001831, -0.00000909, -0.00000436, -0.00001067, -0.00001325, -0.00000888, -0.00000765, -0.00000699, 0.00000010, 0.00000241, -0.00000963, -0.00001838, -0.00000743, 0.00000561, -0.00000028, -0.00001104, -0.00000548, 0.00000447, -0.00000165, -0.00001379, -0.00001214, -0.00000201, 0.00000235, 0.00000291, 0.00000264, -0.00000532, -0.00001397, -0.00000373, 0.00001716, 0.00001530, -0.00001017, -0.00002025, -0.00000143, 0.00001294, 0.00000287, -0.00000929, -0.00000668, -0.00000304, -0.00000592, -0.00000461, -0.00000240, -0.00001054, -0.00001152, 0.00001461, 0.00003683, 0.00001357, -0.00002640, -0.00002272, 0.00000975, 0.00000540, -0.00002897, -0.00002170, 0.00002837, 0.00003917, -0.00000974, -0.00003557, 0.00000415, 0.00003718, 0.00000397, -0.00004010, -0.00002304, 0.00002406, 0.00002755, -0.00000767, -0.00002065, 0.00000024, 0.00001246, -0.00000014, -0.00000965, -0.00000118, 0.00000695, 0.00000230, -0.00000382, -0.00000238, 0.00000007, -0.00000095, -0.00000039, 0.00000241, 0.00000110, -0.00000379, -0.00000402, 0.00000108, 0.00000268, -0.00000228, -0.00000679, -0.00000673, -0.00000562, -0.00000441, 0.00000052, 0.00000605, 0.00000287, -0.00000783, -0.00001163, -0.00000163, 0.00001004, 0.00000880, -0.00000467, -0.00001997, -0.00002931, -0.00002878, -0.00001801, -0.00000373, 0.00000632, 0.00001458, 0.00002748, 0.00003946, 0.00003888, 0.00002816, 0.00002064, 0.00001919, 0.00001663, 0.00001358, 0.00001558, 0.00001753, 0.00001243, 0.00000813, 0.00001398, 0.00002280, 0.00002901, 0.00004468, },
		{ 0.00000916, -0.00004999, -0.00006563, -0.00003715, -0.00001251, -0.00001339, -0.00001676, -0.00000866, -0.00000445, -0.00001093, -0.00001378, -0.00000844, -0.00000585, -0.00000643, -0.00000209, 0.00000084, -0.00000715, -0.00001459, -0.00000762, 0.00000172, -0.00000269, -0.00000945, -0.00000274, 0.00000479, -0.00000396, -0.00001583, -0.00001079, 0.00000227, 0.00000449, -0.00000099, -0.00000311, -0.00000430, -0.00000536, 0.00000259, 0.00001316, 0.00000553, -0.00001485, -0.00001611, 0.00000528, 0.00001461, -0.00000169, -0.00001481, -0.00000743, 0.00000046, -0.00000478, -0.00000901, -0.00000516, -0.00000307, 0.00000132, 0.00001787, 0.00002601, 0.00000177, -0.00002733, -0.00001836, 0.00000839, 0.00000149, -0.00002408, -0.00001018, 0.00003121, 0.00002905, -0.00001758, -0.00002942, 0.00001370, 0.00003357, -0.00000899, -0.00004344, -0.00001142, 0.00003458, 0.00002377, -0.00001795, -0.00002253, 0.00000694, 0.00001600, -0.00000422, -0.00001374, 0.00000078, 0.00001097, 0.00000223, -0.00000693, -0.00000333, 0.00000238, 0.00000094, -0.00000143, 0.00000006, 0.00000013, -0.00000297, -0.00000300, 0.00000108, 0.00000173, -0.00000358, -0.00000794, -0.00000697, -0.00000397, -0.00000100, 0.00000339, 0.00000542, -0.00000150, -0.00001184, -0.00001017, 0.00000537, 0.00001572, 0.00000586, -0.00001520, -0.00002846, -0.00002745, -0.00001815, -0.00000711, 0.00000185, 0.00000815, 0.00001683, 0.00003078, 0.00004010, 0.00003329, 0.00001802, 0.00001274, 0.00001849, 0.00002004, 0.00001375, 0.00001068, 0.00001286, 0.00001161, 0.00000854, 0.00001294, 0.00002190, 0.00002870, 0.00004098, 0.00006009, 0.00005696, },
		{ -0.00006187, -0.00003939, -0.00001457, -0.00001204, -0.00001503, -0.00000905, -0.00000548, -0.00001089, -0.00001299, -0.00000741, -0.00000498, -0.00000698, -0.00000393, 0.00000071, -0.00000409, -0.00001201, -0.00000939, -0.00000214, -0.00000339, -0.00000666, -0.00000103, 0.00000302, -0.00000647, -0.00001545, -0.00000739, 0.00000523, 0.00000325, -0.00000590, -0.00000568, 0.00000050, 0.00000280, 0.00000422, 0.00000539, -0.00000287, -0.00001468, -0.00000924, 0.00000975, 0.00001218, -0.00000717, -0.00001751, -0.00000601, 0.00000240, -0.00000618, -0.00001209, -0.00000314, 0.00000668, 0.00001030, 0.00001512, 0.00001351, -0.00000691, -0.00002553, -0.00001431, 0.00000685, 0.00000014, -0.00001716, -0.00000125, 0.00002940, 0.00001860, -0.00002101, -0.00002099, 0.00001950, 0.00002558, -0.00002000, -0.00004053, 0.00000234, 0.00004041, 0.00001500, -0.00002731, -0.00001995, 0.00001511, 0.00001680, -0.00001055, -0.00001653, 0.00000506, 0.00001483, 0.00000021, -0.00001066, -0.00000288, 0.00000579, 0.00000206, -0.00000389, -0.00000248, 0.00000021, -0.00000136, -0.00000224, 0.00000031, 0.00000028, -0.00000492, -0.00000862, -0.00000609, -0.00000100, 0.00000257, 0.00000435, 0.00000209, -0.00000636, -0.00001281, -0.00000484, 0.00001257, 0.00001678, -0.00000229, -0.00002596, -0.00003127, -0.00001894, -0.00000566, 0.00000005, 0.00000240, 0.00000800, 0.00001999, 0.00003447, 0.00003879, 0.00002543, 0.00000822, 0.00000773, 0.00002035, 0.00002344, 0.00001203, 0.00000447, 0.00000897, 0.00001301, 0.00001058, 0.00001154, 0.00001935, 0.00002746, 0.00003811, 0.00005370, 0.00005276, 0.00001314, -0.00004169, },
		{ -0.00001607, -0.00001060, -0.00001370, -0.00001006, -0.00000665, -0.00001014, -0.00001140, -0.00000658, -0.00000512, -0.00000785, -0.00000476, 0.00000160, -0.00000151, -0.00001096, -0.00001185, -0.00000494, -0.00000261, -0.00000393, -0.00000096, 0.00000021, -0.00000771, -0.00001279, -0.00000374, 0.00000561, -0.00000029, -0.00000951, -0.00000441, 0.00000698, 0.00000796, 0.00000125, -0.00000342, -0.00000762, -0.00001043, -0.00000234, 0.00001064, 0.00000705, -0.00001159, -0.00001748, -0.00000415, 0.00000219, -0.00000840, -0.00001212, 0.00000272, 0.00001583, 0.00001394, 0.00000786, 0.00000169, -0.00001186, -0.00002188, -0.00001086, 0.00000575, 0.00000110, -0.00000985, 0.00000415, 0.00002434, 0.00000979, -0.00002043, -0.00001261, 0.00002076, 0.00001546, -0.00002671, -0.00003230, 0.00001513, 0.00004028, 0.00000329, -0.00003330, -0.00001316, 0.00002266, 0.00001397, -0.00001788, -0.00001675, 0.00001121, 0.00001731, -0.00000392, -0.00001407, -0.00000052, 0.00000969, 0.00000163, -0.00000762, -0.00000443, 0.00000168, 0.00000065, -0.00000227, -0.00000137, -0.00000144, -0.00000586, -0.00000832, -0.00000386, 0.00000272, 0.00000504, 0.00000261, -0.00000298, -0.00000955, -0.00000964, 0.00000274, 0.00001688, 0.00001198, -0.00001325, -0.00003293, -0.00002679, -0.00000633, 0.00000445, 0.00000152, -0.00000039, 0.00000874, 0.00002506, 0.00003727, 0.00003422, 0.00001582, 0.00000046, 0.00000642, 0.00002373, 0.00002519, 0.00000805, -0.00000191, 0.00000711, 0.00001658, 0.00001306, 0.00000877, 0.00001519, 0.00002613, 0.00003680, 0.00004882, 0.00004839, 0.00001560, -0.00003459, -0.00005811, -0.00004092, },
		{ -0.00001280, -0.00001114, -0.00000746, -0.00000884, -0.00000966, -0.00000631, -0.00000582, -0.00000832, -0.00000455, 0.00000274, -0.00000005, -0.00001109, -0.00001401, -0.00000635, -0.00000121, -0.00000222, -0.00000223, -0.00000231, -0.00000708, -0.00000893, -0.00000143, 0.00000346, -0.00000423, -0.00001030, -0.00000023, 0.00001260, 0.00000889, -0.00000468, -0.00001056, -0.00000805, -0.00000409, 0.00000241, 0.00000808, 0.00000115, -0.00001389, -0.00001573, -0.00000316, 0.00000042, -0.00000950, -0.00000857, 0.00001036, 0.00002189, 0.00001220, -0.00000165, -0.00000760, -0.00001327, -0.00001733, -0.00000801, 0.00000550, 0.00000363, -0.00000359, 0.00000602, 0.00001782, 0.00000374, -0.00001711, -0.00000615, 0.00001802, 0.00000581, -0.00002806, -0.00002095, 0.00002421, 0.00003448, -0.00000860, -0.00003445, -0.00000362, 0.00002753, 0.00000768, -0.00002446, -0.00001372, 0.00001809, 0.00001735, -0.00000966, -0.00001604, 0.00000381, 0.00001306, -0.00000089, -0.00001201, -0.00000493, 0.00000456, 0.00000236, -0.00000361, -0.00000381, -0.00000290, -0.00000586, -0.00000674, -0.00000057, 0.00000613, 0.00000528, -0.00000163, -0.00000794, -0.00000922, -0.00000279, 0.00000970, 0.00001587, 0.00000226, -0.00002311, -0.00003307, -0.00001610, 0.00000613, 0.00000893, -0.00000205, -0.00000332, 0.00001248, 0.00003118, 0.00003714, 0.00002598, 0.00000622, -0.00000349, 0.00000850, 0.00002671, 0.00002406, 0.00000254, -0.00000675, 0.00000794, 0.00002129, 0.00001451, 0.00000424, 0.00001031, 0.00002573, 0.00003721, 0.00004478, 0.00004317, 0.00001646, -0.00002821, -0.00005377, -0.00004162, -0.00001741, -0.00000961, },
		{ -0.00000786, -0.00000747, -0.00000823, -0.00000646, -0.00000649, -0.00000811, -0.00000375, 0.00000341, 0.00000023, -0.00001165, -0.00001527, -0.00000672, -0.00000009, -0.00000176, -0.00000393, -0.00000358, -0.00000491, -0.00000535, -0.00000121, -0.00000012, -0.00000677, -0.00000806, 0.00000492, 0.00001538, 0.00000597, -0.00001114, -0.00001425, -0.00000499, 0.00000200, 0.00000394, 0.00000325, -0.00000390, -0.00001406, -0.00001355, -0.00000348, -0.00000159, -0.00000817, -0.00000225, 0.00001730, 0.00002346, 0.00000632, -0.00001103, -0.00001339, -0.00001196, -0.00001268, -0.00000558, 0.00000619, 0.00000678, 0.00000070, 0.00000507, 0.00001145, 0.00000062, -0.00001276, -0.00000259, 0.00001287, -0.00000118, -0.00002448, -0.00000932, 0.00002798, 0.00002474, -0.00001800, -0.00003061, 0.00000641, 0.00002840, -0.00000085, -0.00002856, -0.00000765, 0.00002414, 0.00001441, -0.00001590, -0.00001563, 0.00000950, 0.00001480, -0.00000551, -0.00001603, -0.00000334, 0.00000834, 0.00000288, -0.00000649, -0.00000651, -0.00000339, -0.00000449, -0.00000397, 0.00000295, 0.00000802, 0.00000286, -0.00000711, -0.00001064, -0.00000469, 0.00000565, 0.00001295, 0.00000890, -0.00000930, -0.00002789, -0.00002565, -0.00000271, 0.00001414, 0.00000687, -0.00000776, -0.00000317, 0.00001937, 0.00003595, 0.00003244, 0.00001525, -0.00000091, -0.00000281, 0.00001234, 0.00002728, 0.00001990, -0.00000279, -0.00000864, 0.00001100, 0.00002544, 0.00001391, -0.00000142, 0.00000605, 0.00002681, 0.00003866, 0.00004062, 0.00003684, 0.00001625, -0.00002193, -0.00004872, -0.00004185, -0.00001907, -0.00000914, -0.00001196, -0.00001181, },
		{ -0.00000715, -0.00000662, -0.00000673, -0.00000736, -0.00000294, 0.00000332, -0.00000028, -0.00001202, -0.00001556, -0.00000671, 0.00000022, -0.00000214, -0.00000509, -0.00000337, -0.00000221, -0.00000315, -0.00000278, -0.00000353, -0.00000691, -0.00000381, 0.00000896, 0.00001458, 0.00000081, -0.00001589, -0.00001413, -0.00000025, 0.00000604, 0.00000239, -0.00000224, -0.00000718, -0.00001282, -0.00001188, -0.00000462, -0.00000238, -0.00000412, 0.00000506, 0.00002143, 0.00002047, -0.00000165, -0.00001838, -0.00001564, -0.00000901, -0.00000843, -0.00000331, 0.00000760, 0.00000958, 0.00000273, 0.00000243, 0.00000633, -0.00000020, -0.00000884, -0.00000183, 0.00000729, -0.00000437, -0.00001763, -0.00000002, 0.00002636, 0.00001365, -0.00002307, -0.00002299, 0.00001462, 0.00002505, -0.00000974, -0.00002906, 0.00000035, 0.00002786, 0.00000877, -0.00002124, -0.00001248, 0.00001544, 0.00001399, -0.00001165, -0.00001847, 0.00000052, 0.00001206, 0.00000141, -0.00001067, -0.00000857, -0.00000221, -0.00000170, -0.00000070, 0.00000556, 0.00000757, -0.00000163, -0.00001184, -0.00000954, 0.00000308, 0.00001260, 0.00001046, -0.00000238, -0.00001851, -0.00002513, -0.00001278, 0.00000867, 0.00001508, 0.00000013, -0.00001168, 0.00000166, 0.00002740, 0.00003664, 0.00002312, 0.00000461, -0.00000349, 0.00000174, 0.00001547, 0.00002418, 0.00001398, -0.00000590, -0.00000716, 0.00001471, 0.00002737, 0.00001130, -0.00000675, 0.00000352, 0.00002902, 0.00003997, 0.00003571, 0.00002988, 0.00001581, -0.00001543, -0.00004336, -0.00004216, -0.00002117, -0.00000889, -0.00001076, -0.00001197, -0.00000817, -0.00000644, },
		{ -0.00000654, -0.00000647, -0.00000252, 0.00000257, -0.00000105, -0.00001192, -0.00001525, -0.00000688, -0.00000022, -0.00000262, -0.00000519, -0.00000216, -0.00000008, -0.00000264, -0.00000509, -0.00000546, -0.00000477, 0.00000075, 0.00001044, 0.00001080, -0.00000449, -0.00001761, -0.00001109, 0.00000417, 0.00000718, -0.00000120, -0.00000705, -0.00000865, -0.00001110, -0.00001100, -0.00000556, -0.00000108, 0.00000182, 0.00001128, 0.00002164, 0.00001407, -0.00000950, -0.00002264, -0.00001503, -0.00000543, -0.00000475, -0.00000102, 0.00000929, 0.00001133, 0.00000274, -0.00000077, 0.00000288, 0.00000028, -0.00000625, -0.00000293, 0.00000301, -0.00000383, -0.00000982, 0.00000531, 0.00002062, 0.00000382, -0.00002330, -0.00001370, 0.00001932, 0.00001843, -0.00001697, -0.00002578, 0.00000861, 0.00002829, 0.00000145, -0.00002435, -0.00000693, 0.00002023, 0.00001030, -0.00001811, -0.00001832, 0.00000614, 0.00001445, -0.00000249, -0.00001536, -0.00000890, 0.00000100, 0.00000196, 0.00000195, 0.00000623, 0.00000481, -0.00000666, -0.00001386, -0.00000444, 0.00001166, 0.00001516, 0.00000220, -0.00001440, -0.00002162, -0.00001500, 0.00000130, 0.00001393, 0.00000912, -0.00000739, -0.00001065, 0.00001040, 0.00003330, 0.00003161, 0.00001106, -0.00000293, -0.00000087, 0.00000793, 0.00001560, 0.00001771, 0.00000852, -0.00000538, -0.00000330, 0.00001702, 0.00002629, 0.00000783, -0.00001018, 0.00000296, 0.00003120, 0.00004004, 0.00003016, 0.00002324, 0.00001578, -0.00000900, -0.00003835, -0.00004279, -0.00002337, -0.00000836, -0.00000906, -0.00001190, -0.00000876, -0.00000582, -0.00000613, -0.00000642, },
		{ -0.00000251, 0.00000151, -0.00000177, -0.00001145, -0.00001476, -0.00000738, -0.00000098, -0.00000262, -0.00000431, -0.00000081, 0.00000085, -0.00000335, -0.00000690, -0.00000540, -0.00000140, 0.00000399, 0.00000903, 0.00000552, -0.00000814, -0.00001617, -0.00000672, 0.00000679, 0.00000560, -0.00000552, -0.00001042, -0.00000881, -0.00000956, -0.00001046, -0.00000528, 0.00000230, 0.00000806, 0.00001473, 0.00001802, 0.00000611, -0.00001552, -0.00002372, -0.00001261, -0.00000194, -0.00000156, 0.00000131, 0.00001069, 0.00001167, 0.00000135, -0.00000364, 0.00000098, 0.00000109, -0.00000511, -0.00000459, 0.00000103, -0.00000069, -0.00000321, 0.00000630, 0.00001290, -0.00000284, -0.00001949, -0.00000504, 0.00001990, 0.00001031, -0.00002106, -0.00001954, 0.00001535, 0.00002530, -0.00000602, -0.00002444, -0.00000007, 0.00002259, 0.00000415, -0.00002341, -0.00001512, 0.00001238, 0.00001434, -0.00000856, -0.00001923, -0.00000658, 0.00000600, 0.00000534, 0.00000279, 0.00000460, 0.00000079, -0.00001025, -0.00001210, 0.00000324, 0.00001798, 0.00001180, -0.00000958, -0.00002290, -0.00001684, -0.00000047, 0.00001167, 0.00001125, -0.00000077, -0.00001138, -0.00000371, 0.00002003, 0.00003395, 0.00002131, -0.00000046, -0.00000516, 0.00000566, 0.00001282, 0.00001166, 0.00000965, 0.00000578, -0.00000124, 0.00000089, 0.00001630, 0.00002275, 0.00000534, -0.00001076, 0.00000355, 0.00003191, 0.00003845, 0.00002483, 0.00001790, 0.00001617, -0.00000339, -0.00003423, -0.00004350, -0.00002505, -0.00000731, -0.00000717, -0.00001201, -0.00000974, -0.00000539, -0.00000488, -0.00000581, -0.00000618, -0.00000580, },
		{ -0.00000241, -0.00001090, -0.00001433, -0.00000799, -0.00000156, -0.00000199, -0.00000299, -0.00000000, 0.00000060, -0.00000438, -0.00000739, -0.00000371, 0.00000180, 0.00000498, 0.00000545, 0.00000049, -0.00000922, -0.00001246, -0.00000270, 0.00000697, 0.00000216, -0.00000946, -0.00001225, -0.00000828, -0.00000828, -0.00000951, -0.00000331, 0.00000687, 0.00001288, 0.00001468, 0.00001171, -0.00000149, -0.00001880, -0.00002222, -0.00000939, 0.00000113, 0.00000129, 0.00000351, 0.00001133, 0.00001063, -0.00000070, -0.00000568, 0.00000014, 0.00000159, -0.00000496, -0.00000564, 0.00000138, 0.00000335, 0.00000073, 0.00000381, 0.00000549, -0.00000555, -0.00001337, 0.00000118, 0.00001690, 0.00000268, -0.00002143, -0.00001192, 0.00001924, 0.00001965, -0.00001207, -0.00002150, 0.00000662, 0.00002181, -0.00000331, -0.00002618, -0.00000923, 0.00001767, 0.00001100, -0.00001577, -0.00002077, -0.00000124, 0.00001173, 0.00000699, 0.00000108, 0.00000127, -0.00000275, -0.00001083, -0.00000692, 0.00001088, 0.00001952, 0.00000313, -0.00002103, -0.00002456, -0.00000519, 0.00001372, 0.00001465, 0.00000185, -0.00000996, -0.00000897, 0.00000739, 0.00002661, 0.00002795, 0.00000834, -0.00000802, -0.00000176, 0.00001331, 0.00001400, 0.00000438, 0.00000266, 0.00000699, 0.00000502, 0.00000309, 0.00001216, 0.00001847, 0.00000547, -0.00000871, 0.00000374, 0.00003016, 0.00003573, 0.00002092, 0.00001431, 0.00001636, 0.00000060, -0.00003101, -0.00004364, -0.00002576, -0.00000594, -0.00000558, -0.00001254, -0.00001086, -0.00000479, -0.00000330, -0.00000502, -0.00000594, -0.00000543, -0.00000272, 0.00000039, },
		{ -0.00001383, -0.00000837, -0.00000174, -0.00000101, -0.00000184, 0.00000001, -0.00000028, -0.00000485, -0.00000644, -0.00000137, 0.00000370, 0.00000372, 0.00000100, -0.00000296, -0.00000780, -0.00000784, -0.00000019, 0.00000488, -0.00000208, -0.00001237, -0.00001281, -0.00000742, -0.00000693, -0.00000755, 0.00000008, 0.00001123, 0.00001507, 0.00001146, 0.00000440, -0.00000727, -0.00001929, -0.00001909, -0.00000610, 0.00000375, 0.00000387, 0.00000524, 0.00001090, 0.00000854, -0.00000273, -0.00000676, -0.00000016, 0.00000159, -0.00000506, -0.00000542, 0.00000332, 0.00000667, 0.00000160, -0.00000053, 0.00000012, -0.00000475, -0.00000696, 0.00000407, 0.00001174, -0.00000279, -0.00001851, -0.00000470, 0.00001975, 0.00001274, -0.00001554, -0.00001634, 0.00001164, 0.00001799, -0.00001048, -0.00002563, -0.00000181, 0.00002044, 0.00000459, -0.00002244, -0.00001876, 0.00000656, 0.00001651, 0.00000564, -0.00000303, -0.00000222, -0.00000398, -0.00000793, -0.00000017, 0.00001562, 0.00001537, -0.00000816, -0.00002809, -0.00001830, 0.00000973, 0.00002272, 0.00000925, -0.00001040, -0.00001405, 0.00000004, 0.00001877, 0.00002690, 0.00001635, -0.00000343, -0.00000953, 0.00000543, 0.00001882, 0.00001070, -0.00000394, -0.00000092, 0.00001173, 0.00001087, 0.00000182, 0.00000570, 0.00001559, 0.00000885, -0.00000541, 0.00000195, 0.00002595, 0.00003315, 0.00001932, 0.00001207, 0.00001534, 0.00000264, -0.00002809, -0.00004249, -0.00002546, -0.00000481, -0.00000474, -0.00001332, -0.00001169, -0.00000384, -0.00000161, -0.00000435, -0.00000591, -0.00000518, -0.00000295, -0.00000075, -0.00000316, -0.00001043, },
		{ -0.00000168, -0.00000016, -0.00000119, -0.00000051, -0.00000108, -0.00000440, -0.00000456, 0.00000057, 0.00000380, 0.00000095, -0.00000295, -0.00000424, -0.00000467, -0.00000363, 0.00000028, 0.00000117, -0.00000622, -0.00001405, -0.00001241, -0.00000624, -0.00000506, -0.00000448, 0.00000394, 0.00001404, 0.00001427, 0.00000619, -0.00000231, -0.00001057, -0.00001758, -0.00001525, -0.00000307, 0.00000604, 0.00000609, 0.00000616, 0.00000936, 0.00000591, -0.00000427, -0.00000704, -0.00000031, 0.00000130, -0.00000475, -0.00000391, 0.00000569, 0.00000819, -0.00000003, -0.00000492, -0.00000240, -0.00000172, -0.00000195, 0.00000385, 0.00000622, -0.00000528, -0.00001356, 0.00000062, 0.00001727, 0.00000615, -0.00001607, -0.00001032, 0.00001397, 0.00001207, -0.00001578, -0.00002183, 0.00000543, 0.00001968, -0.00000379, -0.00002664, -0.00001275, 0.00001533, 0.00001848, 0.00000083, -0.00000825, -0.00000388, -0.00000180, -0.00000258, 0.00000546, 0.00001552, 0.00000678, -0.00001817, -0.00002803, -0.00000590, 0.00002295, 0.00002342, -0.00000240, -0.00002044, -0.00001059, 0.00001299, 0.00002604, 0.00001989, 0.00000245, -0.00001041, -0.00000511, 0.00001313, 0.00001986, 0.00000411, -0.00001038, -0.00000011, 0.00001798, 0.00001392, -0.00000289, -0.00000085, 0.00001573, 0.00001471, -0.00000282, -0.00000257, 0.00002035, 0.00003208, 0.00002010, 0.00001011, 0.00001238, 0.00000311, -0.00002456, -0.00003972, -0.00002467, -0.00000456, -0.00000466, -0.00001386, -0.00001188, -0.00000270, -0.00000020, -0.00000395, -0.00000591, -0.00000486, -0.00000321, -0.00000210, -0.00000412, -0.00000991, -0.00001298, -0.00000831, },
		{ -0.00000092, -0.00000104, -0.00000135, -0.00000320, -0.00000257, 0.00000143, 0.00000233, -0.00000225, -0.00000551, -0.00000354, -0.00000094, -0.00000076, -0.00000116, -0.00000336, -0.00000968, -0.00001451, -0.00001119, -0.00000455, -0.00000245, -0.00000080, 0.00000713, 0.00001454, 0.00001101, 0.00000037, -0.00000728, -0.00001146, -0.00001452, -0.00001130, -0.00000034, 0.00000803, 0.00000769, 0.00000600, 0.00000694, 0.00000331, -0.00000508, -0.00000676, -0.00000048, 0.00000106, -0.00000367, -0.00000165, 0.00000737, 0.00000761, -0.00000296, -0.00000796, -0.00000228, 0.00000192, 0.00000073, 0.00000153, 0.00000188, -0.00000488, -0.00000816, 0.00000325, 0.00001288, 0.00000120, -0.00001408, -0.00000493, 0.00001337, 0.00000552, -0.00001814, -0.00001577, 0.00001071, 0.00001530, -0.00001225, -0.00002685, -0.00000345, 0.00002287, 0.00001627, -0.00000674, -0.00001251, -0.00000206, 0.00000352, 0.00000298, 0.00000749, 0.00001041, -0.00000323, -0.00002326, -0.00002058, 0.00000861, 0.00002996, 0.00001570, -0.00001575, -0.00002398, -0.00000020, 0.00002535, 0.00002606, 0.00000729, -0.00000945, -0.00001073, 0.00000285, 0.00001791, 0.00001602, -0.00000303, -0.00001265, 0.00000419, 0.00002293, 0.00001290, -0.00000939, -0.00000513, 0.00001914, 0.00002114, -0.00000256, -0.00000934, 0.00001507, 0.00003327, 0.00002234, 0.00000723, 0.00000743, 0.00000301, -0.00001968, -0.00003564, -0.00002421, -0.00000551, -0.00000492, -0.00001363, -0.00001144, -0.00000179, 0.00000066, -0.00000367, -0.00000566, -0.00000444, -0.00000375, -0.00000381, -0.00000514, -0.00000904, -0.00001159, -0.00000795, -0.00000176, 0.00000025, },
		{ -0.00000109, -0.00000179, -0.00000114, 0.00000109, -0.00000000, -0.00000494, -0.00000637, -0.00000153, 0.00000240, 0.00000036, -0.00000410, -0.00000799, -0.00001211, -0.00001379, -0.00000914, -0.00000222, 0.00000064, 0.00000261, 0.00000874, 0.00001275, 0.00000641, -0.00000472, -0.00001012, -0.00001053, -0.00001084, -0.00000750, 0.00000213, 0.00000958, 0.00000835, 0.00000475, 0.00000409, 0.00000120, -0.00000515, -0.00000619, -0.00000064, 0.00000120, -0.00000192, 0.00000057, 0.00000766, 0.00000539, -0.00000584, -0.00000900, -0.00000045, 0.00000477, 0.00000109, -0.00000144, -0.00000040, -0.00000248, -0.00000374, 0.00000324, 0.00000798, -0.00000150, -0.00001064, -0.00000132, 0.00001045, -0.00000008, -0.00001734, -0.00000909, 0.00001276, 0.00000832, -0.00001864, -0.00002246, 0.00000732, 0.00002694, 0.00000959, -0.00001516, -0.00001355, 0.00000378, 0.00001024, 0.00000602, 0.00000468, 0.00000219, -0.00001091, -0.00002148, -0.00000814, 0.00002033, 0.00002840, 0.00000246, -0.00002564, -0.00001921, 0.00001373, 0.00003251, 0.00001830, -0.00000701, -0.00001570, -0.00000503, 0.00001067, 0.00001764, 0.00000901, -0.00000764, -0.00001018, 0.00000950, 0.00002421, 0.00000835, -0.00001508, -0.00000570, 0.00002455, 0.00002590, -0.00000515, -0.00001681, 0.00001161, 0.00003642, 0.00002458, 0.00000272, 0.00000129, 0.00000351, -0.00001334, -0.00003113, -0.00002476, -0.00000741, -0.00000482, -0.00001238, -0.00001076, -0.00000157, 0.00000102, -0.00000312, -0.00000500, -0.00000419, -0.00000484, -0.00000580, -0.00000583, -0.00000752, -0.00000978, -0.00000769, -0.00000230, 0.00000028, -0.00000064, -0.00000117, },
		{ -0.00000051, -0.00000009, -0.00000244, -0.00000658, -0.00000575, 0.00000100, 0.00000465, -0.00000026, -0.00000795, -0.00001213, -0.00001326, -0.00001184, -0.00000624, 0.00000054, 0.00000355, 0.00000488, 0.00000844, 0.00000934, 0.00000168, -0.00000833, -0.00001103, -0.00000847, -0.00000699, -0.00000388, 0.00000433, 0.00001041, 0.00000785, 0.00000264, 0.00000134, -0.00000014, -0.00000461, -0.00000548, -0.00000065, 0.00000183, 0.00000009, 0.00000202, 0.00000645, 0.00000243, -0.00000765, -0.00000816, 0.00000186, 0.00000605, -0.00000015, -0.00000381, -0.00000052, 0.00000065, -0.00000115, 0.00000133, 0.00000379, -0.00000203, -0.00000703, 0.00000005, 0.00000639, -0.00000357, -0.00001412, -0.00000357, 0.00001129, 0.00000060, -0.00002118, -0.00001417, 0.00001703, 0.00002599, -0.00000040, -0.00002177, -0.00000985, 0.00001263, 0.00001565, 0.00000452, -0.00000225, -0.00000587, -0.00001324, -0.00001341, 0.00000501, 0.00002536, 0.00001899, -0.00001141, -0.00002817, -0.00000747, 0.00002624, 0.00003163, 0.00000509, -0.00001842, -0.00001480, 0.00000378, 0.00001486, 0.00001237, 0.00000194, -0.00000765, -0.00000438, 0.00001284, 0.00002097, 0.00000241, -0.00001758, -0.00000271, 0.00002967, 0.00002743, -0.00000969, -0.00002298, 0.00001058, 0.00004023, 0.00002542, -0.00000322, -0.00000466, 0.00000536, -0.00000618, -0.00002722, -0.00002649, -0.00000955, -0.00000379, -0.00001037, -0.00001044, -0.00000217, 0.00000128, -0.00000199, -0.00000414, -0.00000455, -0.00000657, -0.00000761, -0.00000572, -0.00000539, -0.00000803, -0.00000795, -0.00000328, 0.00000032, 0.00000005, -0.00000085, -0.00000069, -0.00000073, },
		{ -0.00000443, -0.00000706, -0.00000412, 0.00000333, 0.00000544, -0.00000242, -0.00001216, -0.00001528, -0.00001288, -0.00000866, -0.00000275, 0.00000317, 0.00000552, 0.00000558, 0.00000661, 0.00000534, -0.00000229, -0.00001032, -0.00001052, -0.00000578, -0.00000310, -0.00000044, 0.00000600, 0.00001020, 0.00000621, 0.00000016, -0.00000080, -0.00000062, -0.00000369, -0.00000473, -0.00000042, 0.00000280, 0.00000181, 0.00000230, 0.00000415, -0.00000028, -0.00000794, -0.00000615, 0.00000360, 0.00000571, -0.00000192, -0.00000478, 0.00000091, 0.00000332, -0.00000054, -0.00000140, 0.00000104, -0.00000112, -0.00000431, -0.00000051, 0.00000257, -0.00000456, -0.00000985, -0.00000052, 0.00000709, -0.00000568, -0.00001914, -0.00000390, 0.00002316, 0.00001983, -0.00001130, -0.00002408, -0.00000135, 0.00002203, 0.00001702, -0.00000195, -0.00001077, -0.00001029, -0.00000921, -0.00000202, 0.00001435, 0.00002223, 0.00000521, -0.00002092, -0.00002222, 0.00000727, 0.00003279, 0.00002271, -0.00000925, -0.00002346, -0.00000788, 0.00001175, 0.00001357, 0.00000427, -0.00000205, -0.00000291, 0.00000190, 0.00001204, 0.00001427, -0.00000208, -0.00001575, 0.00000207, 0.00003216, 0.00002557, -0.00001426, -0.00002635, 0.00001136, 0.00004294, 0.00002422, -0.00000945, -0.00000911, 0.00000849, 0.00000065, -0.00002463, -0.00002889, -0.00001105, -0.00000175, -0.00000824, -0.00001092, -0.00000331, 0.00000196, -0.00000028, -0.00000360, -0.00000584, -0.00000860, -0.00000863, -0.00000462, -0.00000309, -0.00000690, -0.00000882, -0.00000428, 0.00000083, 0.00000120, -0.00000043, -0.00000062, -0.00000027, -0.00000051, -0.00000157, },
		{ -0.00000198, 0.00000494, 0.00000462, -0.00000577, -0.00001609, -0.00001690, -0.00001080, -0.00000449, 0.00000072, 0.00000491, 0.00000606, 0.00000486, 0.00000402, 0.00000161, -0.00000513, -0.00001099, -0.00000907, -0.00000269, 0.00000081, 0.00000265, 0.00000679, 0.00000876, 0.00000376, -0.00000213, -0.00000204, -0.00000038, -0.00000262, -0.00000395, 0.00000006, 0.00000381, 0.00000280, 0.00000144, 0.00000146, -0.00000203, -0.00000686, -0.00000389, 0.00000419, 0.00000429, -0.00000324, -0.00000423, 0.00000294, 0.00000477, -0.00000145, -0.00000393, -0.00000020, 0.00000027, -0.00000298, -0.00000215, 0.00000002, -0.00000352, -0.00000609, -0.00000036, 0.00000186, -0.00000872, -0.00001313, 0.00000567, 0.00002403, 0.00000985, -0.00002003, -0.00002062, 0.00001027, 0.00002882, 0.00001276, -0.00001183, -0.00001738, -0.00000881, -0.00000027, 0.00000838, 0.00001679, 0.00001247, -0.00000797, -0.00002279, -0.00000981, 0.00001993, 0.00003097, 0.00000859, -0.00002007, -0.00002100, 0.00000173, 0.00001546, 0.00000729, -0.00000327, -0.00000115, 0.00000468, 0.00000558, 0.00000666, 0.00000671, -0.00000281, -0.00001027, 0.00000604, 0.00003070, 0.00002165, -0.00001676, -0.00002659, 0.00001233, 0.00004318, 0.00002148, -0.00001442, -0.00001138, 0.00001200, 0.00000598, -0.00002333, -0.00003100, -0.00001138, 0.00000082, -0.00000673, -0.00001220, -0.00000436, 0.00000337, 0.00000158, -0.00000395, -0.00000798, -0.00001022, -0.00000835, -0.00000281, -0.00000134, -0.00000674, -0.00000994, -0.00000472, 0.00000198, 0.00000246, -0.00000041, -0.00000110, -0.00000026, -0.00000076, -0.00000294, -0.00000575, -0.00000655, },
		{ 0.00000217, -0.00000981, -0.00001901, -0.00001650, -0.00000712, 0.00000000, 0.00000333, 0.00000520, 0.00000524, 0.00000337, 0.00000148, -0.00000138, -0.00000695, -0.00001070, -0.00000692, 0.00000077, 0.00000461, 0.00000501, 0.00000635, 0.00000620, 0.00000105, -0.00000369, -0.00000231, 0.00000030, -0.00000158, -0.00000316, 0.00000067, 0.00000451, 0.00000286, -0.00000019, -0.00000087, -0.00000254, -0.00000497, -0.00000216, 0.00000362, 0.00000258, -0.00000352, -0.00000261, 0.00000464, 0.00000480, -0.00000309, -0.00000563, -0.00000034, 0.00000131, -0.00000293, -0.00000383, -0.00000084, -0.00000153, -0.00000401, -0.00000246, -0.00000243, -0.00000775, -0.00000501, 0.00001197, 0.00001948, -0.00000123, -0.00002390, -0.00001168, 0.00002197, 0.00003025, 0.00000319, -0.00002192, -0.00001891, -0.00000141, 0.00001009, 0.00001382, 0.00001185, -0.00000002, -0.00001602, -0.00001670, 0.00000473, 0.00002626, 0.00002133, -0.00000615, -0.00002412, -0.00001264, 0.00000996, 0.00001336, -0.00000143, -0.00000717, 0.00000436, 0.00001196, 0.00000470, -0.00000170, 0.00000131, 0.00000094, -0.00000338, 0.00000689, 0.00002558, 0.00001790, -0.00001587, -0.00002474, 0.00001165, 0.00004058, 0.00001858, -0.00001683, -0.00001179, 0.00001456, 0.00000920, -0.00002261, -0.00003188, -0.00001058, 0.00000308, -0.00000623, -0.00001375, -0.00000469, 0.00000533, 0.00000283, -0.00000548, -0.00001035, -0.00001064, -0.00000676, -0.00000105, -0.00000077, -0.00000737, -0.00001061, -0.00000418, 0.00000349, 0.00000324, -0.00000109, -0.00000197, -0.00000033, -0.00000095, -0.00000404, -0.00000642, -0.00000528, 0.00000020, 0.00000541, },
		{ -0.00002011, -0.00001387, -0.00000239, 0.00000384, 0.00000433, 0.00000400, 0.00000365, 0.00000191, -0.00000052, -0.00000368, -0.00000805, -0.00000966, -0.00000407, 0.00000453, 0.00000797, 0.00000618, 0.00000457, 0.00000296, -0.00000128, -0.00000423, -0.00000179, 0.00000112, -0.00000069, -0.00000237, 0.00000121, 0.00000463, 0.00000206, -0.00000202, -0.00000237, -0.00000199, -0.00000298, -0.00000133, 0.00000233, 0.00000131, -0.00000270, -0.00000063, 0.00000543, 0.00000372, -0.00000468, -0.00000634, -0.00000006, 0.00000157, -0.00000359, -0.00000474, -0.00000034, 0.00000018, -0.00000400, -0.00000540, -0.00000416, -0.00000334, 0.00000265, 0.00001337, 0.00001102, -0.00001017, -0.00002151, 0.00000060, 0.00003024, 0.00002509, -0.00000936, -0.00002853, -0.00001388, 0.00000959, 0.00001764, 0.00001218, 0.00000187, -0.00001045, -0.00001653, -0.00000525, 0.00001647, 0.00002443, 0.00000712, -0.00001699, -0.00002074, -0.00000197, 0.00001355, 0.00000636, -0.00000906, -0.00000589, 0.00001218, 0.00001588, -0.00000070, -0.00001008, 0.00000020, 0.00000801, 0.00000209, 0.00000361, 0.00001861, 0.00001646, -0.00001167, -0.00002271, 0.00000809, 0.00003601, 0.00001719, -0.00001627, -0.00001152, 0.00001505, 0.00001057, -0.00002136, -0.00003110, -0.00000937, 0.00000425, -0.00000656, -0.00001478, -0.00000406, 0.00000716, 0.00000281, -0.00000790, -0.00001200, -0.00000942, -0.00000444, -0.00000024, -0.00000157, -0.00000809, -0.00001014, -0.00000275, 0.00000465, 0.00000303, -0.00000234, -0.00000274, -0.00000015, -0.00000106, -0.00000490, -0.00000646, -0.00000345, 0.00000200, 0.00000440, -0.00000174, -0.00001380, },
		{ 0.00000231, 0.00000600, 0.00000347, 0.00000188, 0.00000216, 0.00000100, -0.00000203, -0.00000558, -0.00000861, -0.00000780, -0.00000047, 0.00000832, 0.00001033, 0.00000583, 0.00000170, -0.00000028, -0.00000268, -0.00000376, -0.00000085, 0.00000181, -0.00000000, -0.00000165, 0.00000148, 0.00000409, 0.00000074, -0.00000354, -0.00000288, -0.00000083, -0.00000144, -0.00000138, 0.00000091, 0.00000089, -0.00000113, 0.00000102, 0.00000513, 0.00000210, -0.00000574, -0.00000632, 0.00000007, 0.00000110, -0.00000422, -0.00000460, 0.00000071, 0.00000071, -0.00000553, -0.00000752, -0.00000271, 0.00000278, 0.00000743, 0.00000977, 0.00000144, -0.00001423, -0.00001339, 0.00001279, 0.00003235, 0.00001423, -0.00002118, -0.00002879, -0.00000318, 0.00002023, 0.00001904, 0.00000403, -0.00000895, -0.00001499, -0.00001001, 0.00000707, 0.00002171, 0.00001556, -0.00000706, -0.00002110, -0.00001190, 0.00000697, 0.00001134, -0.00000274, -0.00001269, -0.00000014, 0.00001902, 0.00001480, -0.00000860, -0.00001542, 0.00000371, 0.00001573, 0.00000387, -0.00000305, 0.00001231, 0.00001843, -0.00000566, -0.00002245, 0.00000165, 0.00003118, 0.00001850, -0.00001343, -0.00001202, 0.00001310, 0.00001107, -0.00001868, -0.00002894, -0.00000870, 0.00000406, -0.00000701, -0.00001463, -0.00000284, 0.00000796, 0.00000135, -0.00001033, -0.00001206, -0.00000676, -0.00000242, -0.00000099, -0.00000322, -0.00000798, -0.00000828, -0.00000110, 0.00000472, 0.00000176, -0.00000360, -0.00000292, 0.00000030, -0.00000126, -0.00000556, -0.00000582, -0.00000130, 0.00000290, 0.00000165, -0.00000662, -0.00001674, -0.00001874, -0.00000928, },
		{ 0.00000117, -0.00000018, 0.00000149, 0.00000067, -0.00000342, -0.00000735, -0.00000850, -0.00000489, 0.00000375, 0.00001158, 0.00001114, 0.00000395, -0.00000166, -0.00000281, -0.00000294, -0.00000256, 0.00000011, 0.00000219, 0.00000050, -0.00000105, 0.00000136, 0.00000298, -0.00000070, -0.00000439, -0.00000256, 0.00000039, -0.00000062, -0.00000195, -0.00000009, 0.00000133, 0.00000063, 0.00000188, 0.00000400, 0.00000051, -0.00000613, -0.00000601, -0.00000024, 0.00000031, -0.00000429, -0.00000367, 0.00000140, -0.00000017, -0.00000739, -0.00000751, 0.00000136, 0.00000830, 0.00000791, 0.00000271, -0.00000606, -0.00001218, -0.00000194, 0.00002125, 0.00002735, 0.00000058, -0.00002842, -0.00002190, 0.00001008, 0.00002639, 0.00001314, -0.00000751, -0.00001613, -0.00001221, 0.00000036, 0.00001571, 0.00001933, 0.00000314, -0.00001703, -0.00001827, -0.00000127, 0.00001124, 0.00000454, -0.00001046, -0.00001112, 0.00000754, 0.00002197, 0.00000917, -0.00001584, -0.00001584, 0.00001020, 0.00002115, 0.00000132, -0.00001086, 0.00000882, 0.00002330, -0.00000004, -0.00002501, -0.00000637, 0.00002791, 0.00002260, -0.00000982, -0.00001437, 0.00000926, 0.00001195, -0.00001429, -0.00002632, -0.00000931, 0.00000291, -0.00000666, -0.00001322, -0.00000192, 0.00000708, -0.00000103, -0.00001163, -0.00001020, -0.00000357, -0.00000174, -0.00000322, -0.00000471, -0.00000634, -0.00000546, -0.00000022, 0.00000328, -0.00000003, -0.00000416, -0.00000237, 0.00000066, -0.00000176, -0.00000584, -0.00000443, 0.00000074, 0.00000231, -0.00000279, -0.00001153, -0.00001756, -0.00001468, -0.00000367, 0.00000568, 0.00000584, },
		{ 0.00000179, 0.00000040, -0.00000515, -0.00000886, -0.00000731, -0.00000084, 0.00000808, 0.00001356, 0.00001011, 0.00000098, -0.00000470, -0.00000413, -0.00000216, -0.00000111, 0.00000076, 0.00000224, 0.00000084, -0.00000066, 0.00000082, 0.00000156, -0.00000188, -0.00000450, -0.00000178, 0.00000128, -0.00000044, -0.00000256, -0.00000040, 0.00000231, 0.00000204, 0.00000185, 0.00000245, -0.00000069, -0.00000601, -0.00000579, -0.00000093, -0.00000031, -0.00000366, -0.00000258, 0.00000112, -0.00000193, -0.00000826, -0.00000501, 0.00000644, 0.00001113, 0.00000417, -0.00000508, -0.00000896, -0.00000473, 0.00000928, 0.00002336, 0.00001657, -0.00001181, -0.00002854, -0.00000948, 0.00002160, 0.00002535, 0.00000156, -0.00001801, -0.00001668, -0.00000357, 0.00001012, 0.00001776, 0.00001104, -0.00000840, -0.00002055, -0.00001076, 0.00000733, 0.00001005, -0.00000395, -0.00001412, -0.00000538, 0.00001385, 0.00001977, 0.00000134, -0.00001943, -0.00001144, 0.00001676, 0.00002219, -0.00000440, -0.00001714, 0.00000897, 0.00002923, 0.00000318, -0.00003006, -0.00001397, 0.00002727, 0.00002843, -0.00000706, -0.00001872, 0.00000482, 0.00001409, -0.00000878, -0.00002433, -0.00001130, 0.00000172, -0.00000494, -0.00001117, -0.00000230, 0.00000450, -0.00000320, -0.00001093, -0.00000690, -0.00000111, -0.00000290, -0.00000611, -0.00000496, -0.00000315, -0.00000270, -0.00000092, 0.00000060, -0.00000145, -0.00000352, -0.00000142, 0.00000048, -0.00000251, -0.00000542, -0.00000239, 0.00000199, -0.00000023, -0.00000835, -0.00001520, -0.00001547, -0.00000848, 0.00000147, 0.00000658, 0.00000348, -0.00000149, -0.00000116, },
		{ -0.00000731, -0.00000962, -0.00000463, 0.00000407, 0.00001170, 0.00001363, 0.00000739, -0.00000228, -0.00000668, -0.00000412, -0.00000077, 0.00000014, 0.00000097, 0.00000204, 0.00000107, -0.00000053, -0.00000007, 0.00000010, -0.00000256, -0.00000402, -0.00000093, 0.00000168, -0.00000062, -0.00000284, -0.00000006, 0.00000339, 0.00000278, 0.00000115, 0.00000092, -0.00000139, -0.00000569, -0.00000582, -0.00000169, -0.00000046, -0.00000262, -0.00000200, -0.00000015, -0.00000355, -0.00000721, -0.00000068, 0.00001051, 0.00001019, -0.00000220, -0.00001065, -0.00000627, 0.00000557, 0.00001669, 0.00001847, 0.00000324, -0.00001914, -0.00002126, 0.00000477, 0.00002743, 0.00001689, -0.00001195, -0.00002338, -0.00001020, 0.00000730, 0.00001524, 0.00001292, 0.00000050, -0.00001544, -0.00001791, -0.00000215, 0.00001141, 0.00000485, -0.00001084, -0.00001294, 0.00000185, 0.00001617, 0.00001324, -0.00000544, -0.00001785, -0.00000425, 0.00002043, 0.00001858, -0.00001076, -0.00001985, 0.00001191, 0.00003381, 0.00000315, -0.00003599, -0.00001925, 0.00002910, 0.00003423, -0.00000617, -0.00002421, 0.00000115, 0.00001756, -0.00000327, -0.00002364, -0.00001402, 0.00000148, -0.00000196, -0.00000955, -0.00000457, 0.00000105, -0.00000390, -0.00000809, -0.00000335, -0.00000043, -0.00000554, -0.00000837, -0.00000332, 0.00000081, -0.00000121, -0.00000343, -0.00000240, -0.00000162, -0.00000181, -0.00000075, -0.00000051, -0.00000310, -0.00000396, -0.00000017, 0.00000163, -0.00000473, -0.00001390, -0.00001631, -0.00001036, -0.00000151, 0.00000456, 0.00000460, -0.00000011, -0.00000317, -0.00000057, 0.00000257, -0.00000052, },
		{ -0.00000036, 0.00000908, 0.00001368, 0.00001159, 0.00000365, -0.00000490, -0.00000716, -0.00000307, 0.00000064, 0.00000083, 0.00000075, 0.00000173, 0.00000119, -0.00000067, -0.00000114, -0.00000114, -0.00000272, -0.00000320, -0.00000029, 0.00000164, -0.00000081, -0.00000260, 0.00000067, 0.00000415, 0.00000278, 0.00000009, -0.00000031, -0.00000172, -0.00000540, -0.00000602, -0.00000221, -0.00000012, -0.00000169, -0.00000227, -0.00000184, -0.00000398, -0.00000415, 0.00000397, 0.00001187, 0.00000580, -0.00000865, -0.00001183, 0.00000107, 0.00001525, 0.00001796, 0.00000821, -0.00000851, -0.00001920, -0.00000876, 0.00001641, 0.00002544, 0.00000342, -0.00002283, -0.00002144, 0.00000101, 0.00001596, 0.00001367, 0.00000350, -0.00000816, -0.00001642, -0.00001159, 0.00000414, 0.00001062, -0.00000150, -0.00001388, -0.00000818, 0.00000746, 0.00001363, 0.00000497, -0.00000840, -0.00001160, 0.00000267, 0.00001941, 0.00001189, -0.00001503, -0.00001849, 0.00001550, 0.00003515, 0.00000051, -0.00004062, -0.00002133, 0.00003209, 0.00003832, -0.00000713, -0.00002942, -0.00000094, 0.00002159, 0.00000111, -0.00002416, -0.00001636, 0.00000267, 0.00000142, -0.00000944, -0.00000846, -0.00000191, -0.00000234, -0.00000391, -0.00000098, -0.00000187, -0.00000851, -0.00000881, -0.00000007, 0.00000418, -0.00000180, -0.00000710, -0.00000451, -0.00000017, 0.00000028, -0.00000105, -0.00000210, -0.00000288, -0.00000153, 0.00000131, -0.00000096, -0.00001050, -0.00001784, -0.00001400, -0.00000311, 0.00000436, 0.00000445, 0.00000031, -0.00000337, -0.00000287, 0.00000129, 0.00000286, -0.00000255, -0.00000945, -0.00000890, },
		{ 0.00001336, 0.00000780, -0.00000008, -0.00000611, -0.00000620, -0.00000157, 0.00000156, 0.00000089, 0.00000030, 0.00000142, 0.00000119, -0.00000104, -0.00000221, -0.00000200, -0.00000244, -0.00000231, 0.00000005, 0.00000140, -0.00000074, -0.00000192, 0.00000143, 0.00000436, 0.00000223, -0.00000097, -0.00000118, -0.00000186, -0.00000523, -0.00000616, -0.00000232, 0.00000039, -0.00000132, -0.00000324, -0.00000307, -0.00000265, 0.00000013, 0.00000721, 0.00000986, -0.00000038, -0.00001255, -0.00000799, 0.00001049, 0.00002100, 0.00001277, -0.00000406, -0.00001508, -0.00001214, 0.00000498, 0.00002175, 0.00001613, -0.00001083, -0.00002741, -0.00001273, 0.00001288, 0.00001898, 0.00000597, -0.00000671, -0.00001200, -0.00001223, -0.00000500, 0.00000621, 0.00000662, -0.00000608, -0.00001268, -0.00000251, 0.00000921, 0.00000738, -0.00000190, -0.00000640, -0.00000300, 0.00000646, 0.00001382, 0.00000484, -0.00001539, -0.00001428, 0.00001729, 0.00003269, -0.00000286, -0.00004219, -0.00002062, 0.00003436, 0.00003983, -0.00000889, -0.00003296, -0.00000153, 0.00002495, 0.00000377, -0.00002504, -0.00001732, 0.00000497, 0.00000393, -0.00001129, -0.00001290, -0.00000307, 0.00000135, 0.00000015, -0.00000079, -0.00000473, -0.00001032, -0.00000701, 0.00000365, 0.00000563, -0.00000441, -0.00001065, -0.00000480, 0.00000247, 0.00000173, -0.00000256, -0.00000357, -0.00000137, 0.00000127, 0.00000101, -0.00000580, -0.00001613, -0.00001868, -0.00000836, 0.00000448, 0.00000732, 0.00000104, -0.00000476, -0.00000471, -0.00000041, 0.00000338, 0.00000170, -0.00000556, -0.00001066, -0.00000614, 0.00000498, 0.00001307, },
		{ -0.00000277, -0.00000563, -0.00000428, -0.00000030, 0.00000170, 0.00000047, -0.00000017, 0.00000117, 0.00000103, -0.00000157, -0.00000312, -0.00000245, -0.00000191, -0.00000155, 0.00000014, 0.00000119, -0.00000033, -0.00000100, 0.00000191, 0.00000400, 0.00000137, -0.00000185, -0.00000175, -0.00000196, -0.00000508, -0.00000604, -0.00000212, 0.00000067, -0.00000167, -0.00000433, -0.00000304, 0.00000030, 0.00000419, 0.00000774, 0.00000507, -0.00000597, -0.00001218, -0.00000033, 0.00001868, 0.00002087, 0.00000296, -0.00001435, -0.00001467, -0.00000052, 0.00001556, 0.00001923, 0.00000251, -0.00002135, -0.00002427, -0.00000027, 0.00002112, 0.00001515, -0.00000499, -0.00001392, -0.00001020, -0.00000554, -0.00000104, 0.00000421, 0.00000213, -0.00000717, -0.00000874, 0.00000130, 0.00000668, 0.00000006, -0.00000493, -0.00000031, 0.00000486, 0.00000566, 0.00000555, 0.00000012, -0.00001171, -0.00000958, 0.00001558, 0.00002742, -0.00000475, -0.00004006, -0.00001868, 0.00003432, 0.00003904, -0.00000994, -0.00003419, -0.00000158, 0.00002655, 0.00000501, -0.00002510, -0.00001657, 0.00000731, 0.00000452, -0.00001460, -0.00001635, -0.00000176, 0.00000609, 0.00000258, -0.00000285, -0.00000759, -0.00000984, -0.00000354, 0.00000626, 0.00000450, -0.00000803, -0.00001267, -0.00000310, 0.00000524, 0.00000172, -0.00000483, -0.00000401, 0.00000139, 0.00000328, -0.00000183, -0.00001201, -0.00001974, -0.00001556, -0.00000066, 0.00001017, 0.00000636, -0.00000456, -0.00000859, -0.00000324, 0.00000339, 0.00000429, -0.00000131, -0.00000867, -0.00000990, -0.00000135, 0.00001031, 0.00001494, 0.00001066, 0.00000321, },
		{ -0.00000217, 0.00000022, 0.00000109, -0.00000014, -0.00000047, 0.00000095, 0.00000070, -0.00000215, -0.00000376, -0.00000256, -0.00000131, -0.00000098, 0.00000014, 0.00000119, 0.00000035, -0.00000015, 0.00000195, 0.00000320, 0.00000047, -0.00000247, -0.00000214, -0.00000206, -0.00000480, -0.00000562, -0.00000189, 0.00000041, -0.00000247, -0.00000477, -0.00000146, 0.00000395, 0.00000658, 0.00000528, -0.00000077, -0.00000872, -0.00000734, 0.00000859, 0.00002275, 0.00001494, -0.00000810, -0.00001946, -0.00000784, 0.00001164, 0.00001972, 0.00000993, -0.00001100, -0.00002495, -0.00001471, 0.00001163, 0.00002289, 0.00000582, -0.00001532, -0.00001586, -0.00000421, 0.00000044, -0.00000099, -0.00000002, -0.00000040, -0.00000485, -0.00000459, 0.00000152, 0.00000130, -0.00000542, -0.00000341, 0.00000742, 0.00000936, 0.00000073, -0.00000260, -0.00000071, -0.00000551, -0.00000685, 0.00001012, 0.00002147, -0.00000353, -0.00003502, -0.00001748, 0.00003139, 0.00003723, -0.00000904, -0.00003349, -0.00000236, 0.00002603, 0.00000586, -0.00002341, -0.00001466, 0.00000834, 0.00000293, -0.00001805, -0.00001749, 0.00000161, 0.00001019, 0.00000251, -0.00000614, -0.00000880, -0.00000686, 0.00000011, 0.00000639, 0.00000117, -0.00001106, -0.00001228, -0.00000019, 0.00000686, 0.00000015, -0.00000683, -0.00000279, 0.00000458, 0.00000322, -0.00000714, -0.00001784, -0.00001969, -0.00000888, 0.00000686, 0.00001206, 0.00000175, -0.00001027, -0.00000948, 0.00000082, 0.00000691, 0.00000294, -0.00000567, -0.00001054, -0.00000654, 0.00000469, 0.00001424, 0.00001407, 0.00000623, -0.00000096, -0.00000370, -0.00000369, },
		{ 0.00000003, -0.00000063, -0.00000056, 0.00000068, 0.00000019, -0.00000266, -0.00000407, -0.00000243, -0.00000076, -0.00000054, 0.00000020, 0.00000142, 0.00000111, 0.00000044, 0.00000155, 0.00000217, -0.00000032, -0.00000287, -0.00000242, -0.00000207, -0.00000430, -0.00000503, -0.00000195, -0.00000039, -0.00000317, -0.00000402, 0.00000133, 0.00000701, 0.00000639, 0.00000068, -0.00000549, -0.00000737, 0.00000055, 0.00001571, 0.00002123, 0.00000524, -0.00001674, -0.00001803, 0.00000269, 0.00002025, 0.00001645, -0.00000281, -0.00002012, -0.00002090, -0.00000225, 0.00001892, 0.00001772, -0.00000572, -0.00002163, -0.00001244, 0.00000310, 0.00000333, -0.00000420, -0.00000393, 0.00000020, -0.00000080, -0.00000259, -0.00000179, -0.00000435, -0.00000711, 0.00000148, 0.00001382, 0.00000944, -0.00000625, -0.00000809, 0.00000217, 0.00000071, -0.00000757, 0.00000226, 0.00001709, 0.00000113, -0.00002890, -0.00001847, 0.00002617, 0.00003598, -0.00000580, -0.00003202, -0.00000483, 0.00002398, 0.00000757, -0.00001989, -0.00001286, 0.00000712, -0.00000008, -0.00002002, -0.00001589, 0.00000567, 0.00001214, 0.00000016, -0.00000897, -0.00000737, -0.00000236, 0.00000216, 0.00000361, -0.00000295, -0.00001198, -0.00000961, 0.00000252, 0.00000651, -0.00000227, -0.00000742, -0.00000008, 0.00000678, 0.00000029, -0.00001378, -0.00002117, -0.00001526, -0.00000034, 0.00001168, 0.00000936, -0.00000480, -0.00001374, -0.00000676, 0.00000605, 0.00000841, -0.00000093, -0.00001006, -0.00000988, -0.00000078, 0.00001053, 0.00001556, 0.00001059, 0.00000129, -0.00000359, -0.00000273, -0.00000102, -0.00000062, -0.00000020, },
		{ -0.00000051, 0.00000027, -0.00000042, -0.00000300, -0.00000409, -0.00000217, -0.00000029, -0.00000013, 0.00000042, 0.00000179, 0.00000174, 0.00000065, 0.00000084, 0.00000110, -0.00000094, -0.00000311, -0.00000260, -0.00000190, -0.00000359, -0.00000452, -0.00000246, -0.00000142, -0.00000316, -0.00000198, 0.00000440, 0.00000828, 0.00000364, -0.00000434, -0.00000721, -0.00000212, 0.00000895, 0.00001860, 0.00001457, -0.00000500, -0.00002023, -0.00001099, 0.00001301, 0.00002256, 0.00000729, -0.00001460, -0.00002222, -0.00001110, 0.00000875, 0.00001935, 0.00000760, -0.00001558, -0.00002227, -0.00000558, 0.00000875, 0.00000245, -0.00000865, -0.00000550, 0.00000338, 0.00000260, -0.00000389, -0.00000694, -0.00000771, -0.00000478, 0.00000722, 0.00001671, 0.00000589, -0.00001255, -0.00000971, 0.00000701, 0.00000462, -0.00001168, -0.00000573, 0.00001578, 0.00000811, -0.00002374, -0.00002200, 0.00002020, 0.00003645, -0.00000090, -0.00003114, -0.00000909, 0.00002164, 0.00001089, -0.00001536, -0.00001253, 0.00000366, -0.00000298, -0.00001931, -0.00001225, 0.00000862, 0.00001129, -0.00000312, -0.00000967, -0.00000341, 0.00000188, 0.00000134, -0.00000129, -0.00000600, -0.00001009, -0.00000581, 0.00000360, 0.00000426, -0.00000422, -0.00000597, 0.00000308, 0.00000655, -0.00000531, -0.00001970, -0.00002030, -0.00000719, 0.00000741, 0.00001189, 0.00000286, -0.00001069, -0.00001325, -0.00000116, 0.00001033, 0.00000674, -0.00000638, -0.00001271, -0.00000607, 0.00000625, 0.00001451, 0.00001371, 0.00000544, -0.00000275, -0.00000406, -0.00000039, 0.00000145, -0.00000011, -0.00000139, -0.00000109, -0.00000083, },
		{ -0.00000105, -0.00000311, -0.00000387, -0.00000187, 0.00000012, 0.00000033, 0.00000078, 0.00000216, 0.00000210, 0.00000051, -0.00000002, 0.00000011, -0.00000142, -0.00000322, -0.00000260, -0.00000152, -0.00000284, -0.00000433, -0.00000332, -0.00000216, -0.00000207, 0.00000085, 0.00000657, 0.00000717, -0.00000062, -0.00000781, -0.00000510, 0.00000543, 0.00001511, 0.00001626, 0.00000493, -0.00001247, -0.00001773, -0.00000117, 0.00001949, 0.00001813, -0.00000429, -0.00002160, -0.00001720, 0.00000068, 0.00001461, 0.00001310, -0.00000390, -0.00002078, -0.00001774, 0.00000174, 0.00001085, -0.00000115, -0.00001199, -0.00000412, 0.00000729, 0.00000345, -0.00000794, -0.00001145, -0.00000740, 0.00000006, 0.00001116, 0.00001548, 0.00000089, -0.00001596, -0.00000794, 0.00001136, 0.00000499, -0.00001764, -0.00001164, 0.00001759, 0.00001544, -0.00002091, -0.00002720, 0.00001514, 0.00003882, 0.00000427, -0.00003172, -0.00001429, 0.00002031, 0.00001569, -0.00001130, -0.00001446, -0.00000097, -0.00000411, -0.00001579, -0.00000821, 0.00000905, 0.00000822, -0.00000547, -0.00000746, 0.00000176, 0.00000399, -0.00000248, -0.00000654, -0.00000647, -0.00000587, -0.00000260, 0.00000230, 0.00000117, -0.00000439, -0.00000278, 0.00000502, 0.00000308, -0.00001212, -0.00002255, -0.00001472, 0.00000234, 0.00001160, 0.00000704, -0.00000515, -0.00001333, -0.00000848, 0.00000528, 0.00001161, 0.00000197, -0.00001159, -0.00001215, 0.00000044, 0.00001274, 0.00001534, 0.00000910, 0.00000015, -0.00000473, -0.00000249, 0.00000238, 0.00000294, -0.00000069, -0.00000290, -0.00000191, -0.00000071, -0.00000048, -0.00000032, },
		{ -0.00000348, -0.00000155, 0.00000053, 0.00000087, 0.00000119, 0.00000235, 0.00000212, 0.00000014, -0.00000090, -0.00000075, -0.00000177, -0.00000318, -0.00000239, -0.00000101, -0.00000227, -0.00000450, -0.00000421, -0.00000216, -0.00000001, 0.00000356, 0.00000695, 0.00000401, -0.00000475, -0.00000824, 0.00000030, 0.00001283, 0.00001703, 0.00000947, -0.00000461, -0.00001497, -0.00001049, 0.00000783, 0.00002001, 0.00000883, -0.00001417, -0.00002182, -0.00000737, 0.00001024, 0.00001367, 0.00000261, -0.00001303, -0.00002021, -0.00001033, 0.00000680, 0.00000922, -0.00000531, -0.00001262, -0.00000074, 0.00000978, 0.00000115, -0.00001286, -0.00001320, -0.00000371, 0.00000493, 0.00001159, 0.00001126, -0.00000303, -0.00001567, -0.00000465, 0.00001313, 0.00000218, -0.00002317, -0.00001421, 0.00002125, 0.00002109, -0.00002057, -0.00003242, 0.00001208, 0.00004220, 0.00000830, -0.00003366, -0.00001900, 0.00002064, 0.00002086, -0.00000907, -0.00001833, -0.00000504, -0.00000248, -0.00001052, -0.00000563, 0.00000663, 0.00000457, -0.00000531, -0.00000281, 0.00000611, 0.00000288, -0.00000813, -0.00001012, -0.00000389, -0.00000084, -0.00000148, -0.00000101, -0.00000115, -0.00000219, 0.00000079, 0.00000417, -0.00000316, -0.00001777, -0.00002067, -0.00000557, 0.00001030, 0.00001046, -0.00000152, -0.00001161, -0.00001111, -0.00000077, 0.00000995, 0.00000886, -0.00000451, -0.00001442, -0.00000786, 0.00000812, 0.00001680, 0.00001263, 0.00000303, -0.00000370, -0.00000421, 0.00000034, 0.00000452, 0.00000310, -0.00000199, -0.00000422, -0.00000227, -0.00000043, -0.00000060, -0.00000104, -0.00000157, -0.00000299, },
		{ 0.00000098, 0.00000145, 0.00000151, 0.00000229, 0.00000187, -0.00000033, -0.00000169, -0.00000150, -0.00000199, -0.00000293, -0.00000197, -0.00000052, -0.00000201, -0.00000490, -0.00000468, -0.00000124, 0.00000246, 0.00000518, 0.00000526, -0.00000008, -0.00000703, -0.00000522, 0.00000730, 0.00001766, 0.00001417, 0.00000046, -0.00001118, -0.00001213, -0.00000128, 0.00001291, 0.00001462, -0.00000188, -0.00001907, -0.00001571, 0.00000344, 0.00001440, 0.00000666, -0.00000845, -0.00001724, -0.00001486, -0.00000302, 0.00000812, 0.00000524, -0.00000789, -0.00001041, 0.00000274, 0.00000937, -0.00000332, -0.00001632, -0.00001137, 0.00000158, 0.00000750, 0.00000841, 0.00000628, -0.00000407, -0.00001253, -0.00000219, 0.00001147, -0.00000212, -0.00002625, -0.00001363, 0.00002473, 0.00002388, -0.00002170, -0.00003602, 0.00001102, 0.00004513, 0.00001044, -0.00003595, -0.00002195, 0.00002226, 0.00002481, -0.00000927, -0.00002280, -0.00000687, 0.00000172, -0.00000540, -0.00000574, 0.00000234, 0.00000231, -0.00000215, 0.00000260, 0.00000781, -0.00000121, -0.00001354, -0.00001064, 0.00000087, 0.00000306, -0.00000303, -0.00000495, -0.00000133, 0.00000179, 0.00000283, -0.00000017, -0.00001037, -0.00001978, -0.00001384, 0.00000448, 0.00001382, 0.00000398, -0.00001089, -0.00001378, -0.00000419, 0.00000723, 0.00001068, 0.00000254, -0.00001034, -0.00001331, -0.00000063, 0.00001479, 0.00001713, 0.00000712, -0.00000269, -0.00000500, -0.00000159, 0.00000324, 0.00000530, 0.00000205, -0.00000342, -0.00000502, -0.00000225, -0.00000021, -0.00000093, -0.00000180, -0.00000192, -0.00000264, -0.00000298, -0.00000120, },
		},
		{ { -0.00000192, -0.00000188, -0.00000063, 0.00000141, 0.00000180, 0.00000139, 0.00000088, -0.00000128, -0.00000218, -0.00000005, -0.00000036, -0.00000105, 0.00000405, 0.00000497, -0.00000371, -0.00000661, -0.00000236, -0.00000303, -0.00000086, 0.00000573, 0.00000051, -0.00000581, 0.00000186, 0.00000358, -0.00000405, -0.00000010, 0.00000345, -0.00000607, -0.00000995, -0.00000781, -0.00000573, 0.00000561, 0.00000337, -0.00002135, -0.00001738, 0.00000855, -0.00000189, -0.00001110, 0.00002035, 0.00002327, -0.00000791, 0.00000276, 0.00002244, -0.00000085, -0.00000727, 0.00001406, 0.00000402, -0.00000815, 0.00000984, 0.00000613, -0.00001727, -0.00001028, 0.00000785, 0.00001075, 0.00000801, -0.00000386, -0.00000622, 0.00001858, 0.00001598, -0.00002706, -0.00002612, 0.00002587, 0.00003637, -0.00001093, -0.00004110, -0.00001093, 0.00003637, 0.00002587, -0.00002612, -0.00002706, 0.00001598, 0.00001858, -0.00000622, -0.00000386, 0.00000801, 0.00001075, 0.00000785, -0.00001028, -0.00001727, 0.00000613, 0.00000984, -0.00000815, 0.00000402, 0.00001406, -0.00000727, -0.00000085, 0.00002244, 0.00000276, -0.00000791, 0.00002327, 0.00002035, -0.00001110, -0.00000189, 0.00000855, -0.00001738, -0.00002135, 0.00000337, 0.00000561, -0.00000573, -0.00000781, -0.00000995, -0.00000607, 0.00000345, -0.00000010, -0.00000405, 0.00000358, 0.00000186, -0.00000581, 0.00000051, 0.00000573, -0.00000086, -0.00000303, -0.00000236, -0.00000661, -0.00000371, 0.00000497, 0.00000405, -0.00000105, -0.00000036, -0.00000005, -0.00000218, -0.00000128, 0.00000088, 0.00000139, 0.00000180, 0.00000141, -0.00000063, -0.00000188, },
		{ -0.00000130, 0.00000103, 0.00000198, 0.00000118, 0.00000101, -0.00000010, -0.00000226, -0.00000106, 0.00000082, -0.00000073, 0.00000099, 0.00000514, 0.00000037, -0.00000668, -0.00000445, -0.00000212, -0.00000234, 0.00000309, 0.00000342, -0.00000418, -0.00000123, 0.00000446, -0.00000230, -0.00000282, 0.00000485, -0.00000197, -0.00001135, -0.00000836, -0.00000639, -0.00000041, 0.00000747, -0.00001002, -0.00002548, -0.00000393, 0.00000625, -0.00000930, 0.00000650, 0.00002757, 0.00000448, -0.00000433, 0.00001978, 0.00000958, -0.00001084, 0.00000879, 0.00001387, -0.00000854, 0.00000078, 0.00001374, -0.00001070, -0.00001935, 0.00000571, 0.00001601, 0.00000660, -0.00000274, -0.00000312, 0.00001457, 0.00001729, -0.00001961, -0.00002912, 0.00001950, 0.00003883, -0.00000765, -0.00004138, -0.00001316, 0.00003316, 0.00002964, -0.00002010, -0.00003251, 0.00001066, 0.00002341, -0.00000650, -0.00000808, 0.00001046, 0.00000842, 0.00000418, -0.00000143, -0.00001563, -0.00000542, 0.00001276, -0.00000016, -0.00000474, 0.00001162, 0.00000105, -0.00000748, 0.00001707, 0.00001278, -0.00001256, 0.00001039, 0.00003028, -0.00000292, -0.00001316, 0.00001144, -0.00000207, -0.00002591, -0.00000715, 0.00000957, -0.00000196, -0.00000866, -0.00000795, -0.00000775, -0.00000034, 0.00000182, -0.00000359, 0.00000123, 0.00000427, -0.00000487, -0.00000351, 0.00000622, 0.00000234, -0.00000378, -0.00000166, -0.00000451, -0.00000686, 0.00000211, 0.00000686, 0.00000055, -0.00000209, 0.00000014, -0.00000113, -0.00000230, 0.00000003, 0.00000177, 0.00000188, 0.00000143, -0.00000006, -0.00000147, -0.00000185, -0.00000209, },
		{ 0.00000203, 0.00000131, 0.00000081, 0.00000064, -0.00000153, -0.00000199, 0.00000087, 0.00000059, -0.00000082, 0.00000309, 0.00000342, -0.00000436, -0.00000655, -0.00000244, -0.00000209, 0.00000015, 0.00000388, -0.00000099, -0.00000314, 0.00000314, 0.00000047, -0.00000421, 0.00000337, 0.00000251, -0.00001022, -0.00001052, -0.00000583, -0.00000464, 0.00000504, 0.00000116, -0.00002302, -0.00001891, 0.00000493, -0.00000102, -0.00000330, 0.00002168, 0.00001788, -0.00000311, 0.00001143, 0.00001736, -0.00000675, 0.00000006, 0.00001851, -0.00000119, -0.00000801, 0.00001326, 0.00000032, -0.00002304, -0.00000216, 0.00002035, 0.00000870, -0.00000422, 0.00000011, 0.00001366, 0.00001551, -0.00001324, -0.00002805, 0.00001233, 0.00003892, -0.00000317, -0.00004174, -0.00001529, 0.00003061, 0.00003071, -0.00001358, -0.00003400, 0.00000287, 0.00002598, -0.00000302, -0.00001340, 0.00001077, 0.00001034, -0.00000186, 0.00000214, -0.00000702, -0.00001414, 0.00000682, 0.00001023, -0.00000640, 0.00000243, 0.00000773, -0.00000632, 0.00000693, 0.00001857, -0.00000759, -0.00000411, 0.00002997, 0.00001200, -0.00001888, 0.00000337, 0.00001234, -0.00001972, -0.00001937, 0.00000745, 0.00000430, -0.00000881, -0.00000729, -0.00000670, -0.00000442, 0.00000117, -0.00000139, -0.00000083, 0.00000460, -0.00000196, -0.00000647, 0.00000381, 0.00000576, -0.00000287, -0.00000249, -0.00000196, -0.00000760, -0.00000238, 0.00000763, 0.00000389, -0.00000308, -0.00000106, 0.00000029, -0.00000239, -0.00000151, 0.00000176, 0.00000245, 0.00000137, 0.00000015, -0.00000093, -0.00000159, -0.00000216, -0.00000186, 0.00000036, },
		{ 0.00000073, 0.00000087, -0.00000056, -0.00000225, 0.00000011, 0.00000176, -0.00000091, 0.00000030, 0.00000415, -0.00000070, -0.00000706, -0.00000420, -0.00000136, -0.00000150, 0.00000215, 0.00000181, -0.00000280, 0.00000050, 0.00000236, -0.00000324, 0.00000043, 0.00000499, -0.00000632, -0.00001257, -0.00000635, -0.00000576, -0.00000042, 0.00000677, -0.00001279, -0.00002819, -0.00000595, 0.00000604, -0.00000464, 0.00001008, 0.00002503, 0.00000649, 0.00000365, 0.00001804, 0.00000264, -0.00000581, 0.00001545, 0.00000909, -0.00001057, 0.00000573, 0.00000939, -0.00001882, -0.00001177, 0.00001992, 0.00001407, -0.00000569, 0.00000101, 0.00001586, 0.00001338, -0.00001017, -0.00002411, 0.00000695, 0.00003609, 0.00000142, -0.00004116, -0.00001806, 0.00002901, 0.00003044, -0.00000883, -0.00003177, -0.00000435, 0.00002450, 0.00000263, -0.00001685, 0.00000745, 0.00001469, -0.00000577, -0.00000107, 0.00000331, -0.00001486, -0.00000516, 0.00001594, 0.00000066, -0.00000779, 0.00000721, 0.00000128, -0.00000134, 0.00001599, 0.00000274, -0.00001224, 0.00001909, 0.00002545, -0.00001370, -0.00001008, 0.00001771, -0.00000485, -0.00002621, -0.00000168, 0.00000958, -0.00000607, -0.00000828, -0.00000449, -0.00000653, -0.00000215, 0.00000069, -0.00000101, 0.00000299, 0.00000107, -0.00000676, -0.00000049, 0.00000728, 0.00000010, -0.00000358, -0.00000059, -0.00000592, -0.00000650, 0.00000549, 0.00000756, -0.00000211, -0.00000325, 0.00000098, -0.00000123, -0.00000304, 0.00000082, 0.00000330, 0.00000169, -0.00000005, -0.00000054, -0.00000106, -0.00000208, -0.00000227, -0.00000039, 0.00000176, 0.00000158, },
		{ 0.00000018, -0.00000186, -0.00000077, 0.00000213, 0.00000016, -0.00000171, 0.00000261, 0.00000246, -0.00000528, -0.00000631, -0.00000157, -0.00000150, -0.00000025, 0.00000271, -0.00000069, -0.00000175, 0.00000220, -0.00000067, -0.00000171, 0.00000446, -0.00000140, -0.00001239, -0.00000863, -0.00000524, -0.00000483, 0.00000570, -0.00000129, -0.00002742, -0.00002045, 0.00000526, 0.00000065, -0.00000002, 0.00002281, 0.00001927, 0.00000246, 0.00001217, 0.00001149, -0.00000467, 0.00000745, 0.00001590, -0.00000522, -0.00000340, 0.00001148, -0.00000898, -0.00001762, 0.00001368, 0.00001966, -0.00000440, -0.00000081, 0.00001903, 0.00001357, -0.00001034, -0.00002007, 0.00000502, 0.00003155, 0.00000448, -0.00003912, -0.00002131, 0.00002761, 0.00003041, -0.00000658, -0.00002796, -0.00000843, 0.00001965, 0.00000743, -0.00001659, 0.00000165, 0.00001808, -0.00000502, -0.00000841, 0.00000944, -0.00000716, -0.00001654, 0.00001257, 0.00001245, -0.00001218, -0.00000137, 0.00000941, -0.00000227, 0.00000667, 0.00001087, -0.00001004, 0.00000372, 0.00002945, 0.00000077, -0.00001965, 0.00001101, 0.00001109, -0.00002255, -0.00001406, 0.00000956, -0.00000036, -0.00000920, -0.00000321, -0.00000584, -0.00000643, 0.00000066, 0.00000080, 0.00000095, 0.00000241, -0.00000442, -0.00000443, 0.00000565, 0.00000387, -0.00000324, -0.00000092, -0.00000322, -0.00000843, 0.00000109, 0.00000962, 0.00000101, -0.00000512, 0.00000013, 0.00000064, -0.00000363, -0.00000108, 0.00000375, 0.00000269, -0.00000032, -0.00000060, -0.00000039, -0.00000173, -0.00000260, -0.00000105, 0.00000125, 0.00000165, 0.00000086, 0.00000090, },
		{ -0.00000127, 0.00000179, 0.00000141, -0.00000229, -0.00000002, 0.00000370, -0.00000194, -0.00000722, -0.00000318, -0.00000084, -0.00000174, 0.00000159, 0.00000161, -0.00000225, 0.00000029, 0.00000169, -0.00000158, 0.00000193, 0.00000204, -0.00000924, -0.00001127, -0.00000544, -0.00000630, 0.00000083, 0.00000562, -0.00001823, -0.00003059, -0.00000457, 0.00000609, -0.00000359, 0.00001382, 0.00002825, 0.00000957, 0.00000487, 0.00001462, 0.00000270, 0.00000038, 0.00001551, 0.00000458, -0.00000790, 0.00000609, 0.00000076, -0.00001623, 0.00000448, 0.00002159, 0.00000028, -0.00000340, 0.00002056, 0.00001688, -0.00001156, -0.00001850, 0.00000611, 0.00002765, 0.00000498, -0.00003624, -0.00002409, 0.00002550, 0.00003129, -0.00000592, -0.00002503, -0.00000882, 0.00001405, 0.00000896, -0.00001327, -0.00000377, 0.00001806, -0.00000040, -0.00001519, 0.00000835, 0.00000460, -0.00002096, 0.00000118, 0.00002181, -0.00000727, -0.00001341, 0.00001151, 0.00000469, -0.00000314, 0.00001093, -0.00000010, -0.00000734, 0.00002142, 0.00001669, -0.00001816, -0.00000347, 0.00001915, -0.00000895, -0.00002272, 0.00000230, 0.00000545, -0.00000770, -0.00000367, -0.00000344, -0.00000935, -0.00000216, 0.00000316, 0.00000023, 0.00000152, -0.00000106, -0.00000588, 0.00000138, 0.00000620, -0.00000066, -0.00000197, -0.00000130, -0.00000766, -0.00000363, 0.00000871, 0.00000513, -0.00000520, -0.00000214, 0.00000205, -0.00000276, -0.00000323, 0.00000311, 0.00000412, -0.00000009, -0.00000115, 0.00000010, -0.00000103, -0.00000279, -0.00000167, 0.00000074, 0.00000144, 0.00000096, 0.00000104, 0.00000066, -0.00000125, },
		{ 0.00000223, -0.00000171, -0.00000236, 0.00000276, 0.00000138, -0.00000603, -0.00000538, -0.00000087, -0.00000175, -0.00000039, 0.00000257, -0.00000099, -0.00000190, 0.00000230, 0.00000057, -0.00000040, 0.00000254, -0.00000451, -0.00001204, -0.00000747, -0.00000581, -0.00000374, 0.00000594, -0.00000654, -0.00003135, -0.00001867, 0.00000524, -0.00000107, 0.00000415, 0.00002922, 0.00002144, 0.00000221, 0.00001102, 0.00001145, -0.00000082, 0.00000916, 0.00001273, -0.00000479, -0.00000249, 0.00000488, -0.00000849, -0.00000257, 0.00001797, 0.00000615, -0.00000405, 0.00001936, 0.00002182, -0.00001104, -0.00002009, 0.00000792, 0.00002639, 0.00000352, -0.00003402, -0.00002563, 0.00002253, 0.00003262, -0.00000528, -0.00002426, -0.00000705, 0.00001050, 0.00000698, -0.00000943, -0.00000637, 0.00001471, 0.00000498, -0.00001778, 0.00000137, 0.00001413, -0.00001615, -0.00001238, 0.00002283, 0.00000470, -0.00002175, 0.00000478, 0.00001473, -0.00000680, 0.00000252, 0.00001024, -0.00000807, 0.00000593, 0.00002479, -0.00000517, -0.00001638, 0.00001453, 0.00000788, -0.00002160, -0.00000955, 0.00000699, -0.00000301, -0.00000463, -0.00000136, -0.00000959, -0.00000659, 0.00000402, 0.00000158, -0.00000055, 0.00000122, -0.00000420, -0.00000337, 0.00000533, 0.00000329, -0.00000198, -0.00000112, -0.00000529, -0.00000658, 0.00000504, 0.00000821, -0.00000285, -0.00000459, 0.00000191, -0.00000079, -0.00000456, 0.00000124, 0.00000523, 0.00000100, -0.00000181, 0.00000004, -0.00000011, -0.00000264, -0.00000229, 0.00000030, 0.00000110, 0.00000075, 0.00000125, 0.00000115, -0.00000072, -0.00000139, 0.00000125, },
		{ -0.00000362, 0.00000051, 0.00000320, -0.00000314, -0.00000672, -0.00000222, -0.00000102, -0.00000177, 0.00000179, 0.00000084, -0.00000292, 0.00000098, 0.00000311, -0.00000069, 0.00000062, -0.00000070, -0.00000982, -0.00001025, -0.00000570, -0.00000566, 0.00000188, 0.00000154, -0.00002335, -0.00002949, -0.00000388, 0.00000238, -0.00000111, 0.00002311, 0.00003185, 0.00000718, 0.00000440, 0.00001610, 0.00000459, 0.00000196, 0.00001464, 0.00000389, -0.00000813, 0.00000160, 0.00000076, -0.00000337, 0.00001055, 0.00000957, -0.00000145, 0.00001667, 0.00002587, -0.00000737, -0.00002323, 0.00000777, 0.00002797, 0.00000208, -0.00003371, -0.00002614, 0.00001942, 0.00003346, -0.00000366, -0.00002521, -0.00000553, 0.00001025, 0.00000347, -0.00000777, -0.00000583, 0.00001032, 0.00000806, -0.00001566, -0.00000717, 0.00001700, -0.00000501, -0.00002109, 0.00001456, 0.00001716, -0.00002126, -0.00000790, 0.00002083, -0.00000170, -0.00000919, 0.00001366, 0.00000111, -0.00000805, 0.00002013, 0.00001255, -0.00001893, -0.00000018, 0.00001859, -0.00000964, -0.00001956, 0.00000145, 0.00000272, -0.00000380, -0.00000086, -0.00000766, -0.00001042, 0.00000222, 0.00000408, -0.00000201, 0.00000123, -0.00000054, -0.00000609, 0.00000118, 0.00000642, 0.00000030, -0.00000208, -0.00000318, -0.00000675, 0.00000030, 0.00000853, 0.00000110, -0.00000558, 0.00000005, 0.00000107, -0.00000431, -0.00000119, 0.00000520, 0.00000272, -0.00000191, -0.00000063, 0.00000063, -0.00000203, -0.00000283, -0.00000018, 0.00000087, 0.00000024, 0.00000122, 0.00000183, -0.00000018, -0.00000144, 0.00000082, 0.00000260, -0.00000068, },
		{ 0.00000306, 0.00000008, -0.00000618, -0.00000430, -0.00000079, -0.00000188, 0.00000009, 0.00000183, -0.00000234, -0.00000118, 0.00000431, 0.00000129, -0.00000162, 0.00000036, -0.00000558, -0.00001154, -0.00000736, -0.00000545, -0.00000259, 0.00000333, -0.00001188, -0.00003162, -0.00001744, 0.00000081, -0.00000138, 0.00001461, 0.00003605, 0.00001794, 0.00000046, 0.00001435, 0.00001291, -0.00000078, 0.00001009, 0.00001266, -0.00000689, -0.00000610, 0.00000590, 0.00000239, 0.00000380, 0.00000807, 0.00000305, 0.00001498, 0.00002742, -0.00000137, -0.00002525, 0.00000442, 0.00003063, 0.00000276, -0.00003523, -0.00002682, 0.00001703, 0.00003337, -0.00000121, -0.00002647, -0.00000590, 0.00001239, 0.00000118, -0.00000924, -0.00000402, 0.00000752, 0.00000789, -0.00001118, -0.00001294, 0.00001296, 0.00000645, -0.00002096, 0.00000139, 0.00002327, -0.00001208, -0.00001980, 0.00001843, 0.00000908, -0.00001696, 0.00000750, 0.00001384, -0.00001268, 0.00000514, 0.00002502, -0.00000834, -0.00001621, 0.00001670, 0.00000781, -0.00002098, -0.00000981, 0.00000535, 0.00000013, -0.00000130, -0.00000537, -0.00001195, -0.00000169, 0.00000586, -0.00000163, -0.00000061, 0.00000290, -0.00000541, -0.00000443, 0.00000655, 0.00000447, -0.00000249, -0.00000269, -0.00000480, -0.00000324, 0.00000575, 0.00000468, -0.00000409, -0.00000250, 0.00000153, -0.00000268, -0.00000300, 0.00000373, 0.00000425, -0.00000101, -0.00000147, 0.00000073, -0.00000109, -0.00000301, -0.00000079, 0.00000076, -0.00000029, 0.00000068, 0.00000251, 0.00000067, -0.00000153, 0.00000043, 0.00000282, 0.00000029, -0.00000389, -0.00000188, },
		{ -0.00000395, -0.00000579, -0.00000177, -0.00000124, -0.00000123, 0.00000141, -0.00000097, -0.00000273, 0.00000352, 0.00000418, -0.00000211, -0.00000126, -0.00000176, -0.00000989, -0.00001005, -0.00000533, -0.00000478, 0.00000037, -0.00000307, -0.00002490, -0.00002845, -0.00000809, -0.00000062, 0.00000840, 0.00003351, 0.00002934, 0.00000305, 0.00000844, 0.00001884, 0.00000295, 0.00000316, 0.00001642, 0.00000060, -0.00001252, 0.00000407, 0.00001095, 0.00000212, 0.00000234, 0.00000620, 0.00001605, 0.00002700, 0.00000472, -0.00002425, -0.00000119, 0.00003173, 0.00000630, -0.00003700, -0.00002901, 0.00001551, 0.00003282, 0.00000122, -0.00002682, -0.00000820, 0.00001475, 0.00000179, -0.00001255, -0.00000350, 0.00000735, 0.00000593, -0.00000751, -0.00001419, 0.00000542, 0.00001291, -0.00001309, -0.00000973, 0.00002014, 0.00000052, -0.00002467, 0.00000831, 0.00001889, -0.00001627, -0.00000484, 0.00002199, -0.00000586, -0.00001188, 0.00002507, 0.00001039, -0.00002364, 0.00000272, 0.00002153, -0.00001115, -0.00002119, 0.00000145, 0.00000587, -0.00000064, -0.00000426, -0.00001105, -0.00000582, 0.00000538, 0.00000035, -0.00000270, 0.00000436, -0.00000186, -0.00000877, 0.00000293, 0.00000857, -0.00000078, -0.00000371, -0.00000249, -0.00000410, 0.00000119, 0.00000588, -0.00000051, -0.00000404, 0.00000010, -0.00000077, -0.00000325, 0.00000143, 0.00000464, 0.00000069, -0.00000173, 0.00000012, -0.00000033, -0.00000261, -0.00000140, 0.00000057, -0.00000058, -0.00000025, 0.00000274, 0.00000193, -0.00000140, -0.00000016, 0.00000301, 0.00000117, -0.00000363, -0.00000362, 0.00000155, 0.00000222, },
		{ -0.00000352, -0.00000100, -0.00000149, 0.00000016, -0.00000002, -0.00000298, 0.00000147, 0.00000617, -0.00000019, -0.00000369, -0.00000048, -0.00000591, -0.00001157, -0.00000689, -0.00000482, -0.00000357, -0.00000016, -0.00001403, -0.00003133, -0.00002118, -0.00000404, 0.00000561, 0.00002750, 0.00003659, 0.00001185, 0.00000335, 0.00001915, 0.00001079, -0.00000084, 0.00001369, 0.00000975, -0.00001303, -0.00000303, 0.00001686, 0.00000689, -0.00000388, 0.00000522, 0.00001929, 0.00002673, 0.00000896, -0.00002036, -0.00000660, 0.00002954, 0.00001145, -0.00003691, -0.00003302, 0.00001392, 0.00003264, 0.00000302, -0.00002610, -0.00001123, 0.00001548, 0.00000497, -0.00001521, -0.00000571, 0.00000866, 0.00000459, -0.00000634, -0.00001223, -0.00000119, 0.00001257, -0.00000252, -0.00001362, 0.00001032, 0.00000959, -0.00002066, -0.00000404, 0.00002196, -0.00000805, -0.00001642, 0.00002073, 0.00000780, -0.00002185, 0.00001255, 0.00002741, -0.00001733, -0.00001585, 0.00002370, 0.00000626, -0.00002577, -0.00000890, 0.00000980, 0.00000274, -0.00000435, -0.00000909, -0.00000824, 0.00000251, 0.00000241, -0.00000347, 0.00000353, 0.00000255, -0.00000980, -0.00000313, 0.00001017, 0.00000323, -0.00000477, -0.00000148, -0.00000239, -0.00000290, 0.00000383, 0.00000341, -0.00000319, -0.00000246, -0.00000001, -0.00000190, -0.00000050, 0.00000345, 0.00000233, -0.00000086, -0.00000073, -0.00000028, -0.00000180, -0.00000163, 0.00000013, -0.00000065, -0.00000117, 0.00000223, 0.00000324, -0.00000064, -0.00000094, 0.00000295, 0.00000215, -0.00000314, -0.00000468, -0.00000037, 0.00000271, -0.00000118, -0.00000564, },
		{ -0.00000111, -0.00000082, -0.00000010, -0.00000239, -0.00000053, 0.00000627, 0.00000311, -0.00000469, -0.00000208, -0.00000189, -0.00001020, -0.00000968, -0.00000477, -0.00000559, -0.00000215, -0.00000507, -0.00002541, -0.00003195, -0.00001389, 0.00000323, 0.00002189, 0.00003815, 0.00002300, 0.00000330, 0.00001463, 0.00001793, 0.00000110, 0.00000736, 0.00001530, -0.00000702, -0.00001037, 0.00001648, 0.00001541, -0.00000616, -0.00000009, 0.00002202, 0.00002859, 0.00001122, -0.00001542, -0.00000961, 0.00002437, 0.00001572, -0.00003379, -0.00003765, 0.00001092, 0.00003320, 0.00000438, -0.00002500, -0.00001371, 0.00001411, 0.00000886, -0.00001528, -0.00000995, 0.00000916, 0.00000533, -0.00000704, -0.00000989, -0.00000424, 0.00000761, 0.00000513, -0.00000961, -0.00000009, 0.00001081, -0.00001122, -0.00001208, 0.00001702, 0.00000227, -0.00002139, 0.00001119, 0.00002030, -0.00002024, -0.00000582, 0.00003386, 0.00000007, -0.00002874, 0.00001263, 0.00002286, -0.00001953, -0.00002131, 0.00000825, 0.00000841, -0.00000412, -0.00000770, -0.00000820, -0.00000138, 0.00000281, -0.00000248, 0.00000155, 0.00000563, -0.00000727, -0.00000896, 0.00000789, 0.00000798, -0.00000414, -0.00000222, 0.00000039, -0.00000462, -0.00000067, 0.00000550, 0.00000009, -0.00000451, -0.00000120, 0.00000007, -0.00000095, 0.00000119, 0.00000285, 0.00000099, -0.00000094, -0.00000108, -0.00000115, -0.00000118, -0.00000041, -0.00000083, -0.00000170, 0.00000114, 0.00000395, 0.00000085, -0.00000142, 0.00000231, 0.00000316, -0.00000235, -0.00000533, -0.00000204, 0.00000193, 0.00000083, -0.00000389, -0.00000482, -0.00000187, },
		{ -0.00000078, -0.00000192, -0.00000164, 0.00000482, 0.00000600, -0.00000321, -0.00000494, -0.00000021, -0.00000620, -0.00001156, -0.00000639, -0.00000550, -0.00000550, -0.00000165, -0.00001491, -0.00003485, -0.00002738, -0.00000314, 0.00001783, 0.00003593, 0.00003216, 0.00000921, 0.00000925, 0.00002059, 0.00000791, 0.00000239, 0.00001474, 0.00000200, -0.00001303, 0.00001029, 0.00002269, -0.00000234, -0.00000681, 0.00002133, 0.00003265, 0.00001315, -0.00001148, -0.00000971, 0.00001826, 0.00001705, -0.00002833, -0.00004086, 0.00000588, 0.00003382, 0.00000609, -0.00002422, -0.00001518, 0.00001155, 0.00001152, -0.00001260, -0.00001403, 0.00000720, 0.00000758, -0.00000758, -0.00000923, -0.00000403, 0.00000204, 0.00000698, -0.00000143, -0.00000537, 0.00000490, -0.00000267, -0.00001246, 0.00000783, 0.00000857, -0.00001853, -0.00000089, 0.00002524, -0.00000926, -0.00002070, 0.00002693, 0.00001979, -0.00002886, -0.00000609, 0.00003012, -0.00000427, -0.00002922, 0.00000023, 0.00001354, -0.00000192, -0.00000748, -0.00000639, -0.00000436, 0.00000080, -0.00000066, 0.00000013, 0.00000627, -0.00000275, -0.00001196, 0.00000231, 0.00001100, -0.00000109, -0.00000364, 0.00000243, -0.00000349, -0.00000553, 0.00000440, 0.00000442, -0.00000447, -0.00000390, 0.00000122, 0.00000036, -0.00000095, 0.00000167, 0.00000291, 0.00000014, -0.00000221, -0.00000133, -0.00000016, -0.00000054, -0.00000139, -0.00000190, 0.00000008, 0.00000366, 0.00000254, -0.00000098, 0.00000121, 0.00000371, -0.00000109, -0.00000565, -0.00000340, 0.00000066, 0.00000148, -0.00000152, -0.00000464, -0.00000350, -0.00000115, -0.00000089, },
		{ -0.00000204, 0.00000295, 0.00000724, 0.00000005, -0.00000674, -0.00000156, -0.00000175, -0.00001059, -0.00000946, -0.00000524, -0.00000717, -0.00000320, -0.00000587, -0.00002888, -0.00003829, -0.00001541, 0.00001290, 0.00003271, 0.00003726, 0.00001861, 0.00000719, 0.00001834, 0.00001562, 0.00000242, 0.00000972, 0.00000894, -0.00000929, 0.00000244, 0.00002460, 0.00000600, -0.00001073, 0.00001632, 0.00003683, 0.00001678, -0.00000918, -0.00000824, 0.00001354, 0.00001510, -0.00002271, -0.00004114, -0.00000042, 0.00003321, 0.00000867, -0.00002375, -0.00001614, 0.00000912, 0.00001212, -0.00000868, -0.00001588, 0.00000298, 0.00000936, -0.00000636, -0.00001025, -0.00000282, -0.00000103, 0.00000394, 0.00000561, -0.00000354, -0.00000328, -0.00000000, -0.00000684, 0.00000036, 0.00000808, -0.00001153, -0.00000925, 0.00002145, 0.00000415, -0.00002565, 0.00001108, 0.00003222, -0.00001630, -0.00002273, 0.00002444, 0.00001299, -0.00002782, -0.00001122, 0.00001470, 0.00000240, -0.00000766, -0.00000420, -0.00000514, -0.00000299, 0.00000034, 0.00000039, 0.00000495, 0.00000136, -0.00001103, -0.00000417, 0.00001035, 0.00000340, -0.00000402, 0.00000278, -0.00000053, -0.00000864, 0.00000034, 0.00000767, -0.00000181, -0.00000664, 0.00000051, 0.00000263, -0.00000173, -0.00000083, 0.00000372, 0.00000241, -0.00000266, -0.00000259, 0.00000074, 0.00000019, -0.00000219, -0.00000224, -0.00000045, 0.00000258, 0.00000355, 0.00000056, 0.00000036, 0.00000332, 0.00000041, -0.00000535, -0.00000462, -0.00000060, 0.00000107, 0.00000025, -0.00000293, -0.00000472, -0.00000232, -0.00000046, -0.00000111, -0.00000214, },
		{ 0.00000684, 0.00000337, -0.00000613, -0.00000461, 0.00000063, -0.00000669, -0.00001187, -0.00000673, -0.00000673, -0.00000632, -0.00000219, -0.00001805, -0.00004111, -0.00003072, 0.00000338, 0.00002923, 0.00003907, 0.00002796, 0.00001022, 0.00001397, 0.00002026, 0.00000762, 0.00000458, 0.00001058, -0.00000153, -0.00000216, 0.00002053, 0.00001450, -0.00000909, 0.00000905, 0.00003828, 0.00002270, -0.00000737, -0.00000717, 0.00001122, 0.00001140, -0.00001918, -0.00003860, -0.00000614, 0.00003045, 0.00001174, -0.00002290, -0.00001732, 0.00000748, 0.00001128, -0.00000549, -0.00001494, -0.00000163, 0.00000891, -0.00000351, -0.00001133, -0.00000278, -0.00000114, -0.00000057, 0.00000799, 0.00000301, -0.00000812, -0.00000394, -0.00000044, -0.00000125, 0.00000280, -0.00000598, -0.00001108, 0.00001288, 0.00001301, -0.00002045, -0.00000514, 0.00003235, 0.00000202, -0.00002908, 0.00000925, 0.00002387, -0.00001726, -0.00002040, 0.00001034, 0.00000694, -0.00000700, -0.00000257, -0.00000359, -0.00000702, -0.00000078, 0.00000217, 0.00000323, 0.00000321, -0.00000704, -0.00000865, 0.00000583, 0.00000705, -0.00000218, 0.00000192, 0.00000246, -0.00000889, -0.00000499, 0.00000809, 0.00000252, -0.00000776, -0.00000197, 0.00000445, -0.00000070, -0.00000347, 0.00000273, 0.00000494, -0.00000163, -0.00000434, 0.00000067, 0.00000173, -0.00000260, -0.00000316, -0.00000047, 0.00000151, 0.00000325, 0.00000254, 0.00000069, 0.00000210, 0.00000138, -0.00000416, -0.00000559, -0.00000192, 0.00000023, 0.00000087, -0.00000061, -0.00000448, -0.00000428, -0.00000060, -0.00000055, -0.00000266, -0.00000249, 0.00000153, },
		{ -0.00000348, -0.00000699, -0.00000014, -0.00000190, -0.00001144, -0.00000991, -0.00000607, -0.00000787, -0.00000359, -0.00000845, -0.00003470, -0.00004286, -0.00001209, 0.00002318, 0.00003926, 0.00003517, 0.00001716, 0.00001121, 0.00002048, 0.00001497, 0.00000325, 0.00000740, 0.00000567, -0.00000094, 0.00001360, 0.00001883, -0.00000246, 0.00000334, 0.00003556, 0.00002944, -0.00000412, -0.00000751, 0.00001048, 0.00000818, -0.00001860, -0.00003497, -0.00000953, 0.00002585, 0.00001409, -0.00002099, -0.00001890, 0.00000632, 0.00001026, -0.00000415, -0.00001235, -0.00000450, 0.00000601, -0.00000071, -0.00001081, -0.00000432, -0.00000002, -0.00000339, 0.00000566, 0.00000956, -0.00000679, -0.00001086, 0.00000167, 0.00000271, -0.00000228, -0.00000542, -0.00000831, 0.00000531, 0.00001438, -0.00001026, -0.00001435, 0.00002222, 0.00001674, -0.00002304, -0.00000701, 0.00002353, -0.00000296, -0.00002244, 0.00000241, 0.00000891, -0.00000523, -0.00000145, -0.00000041, -0.00000974, -0.00000428, 0.00000429, 0.00000272, 0.00000235, -0.00000225, -0.00000914, -0.00000075, 0.00000756, 0.00000157, 0.00000137, 0.00000413, -0.00000674, -0.00000940, 0.00000533, 0.00000659, -0.00000637, -0.00000501, 0.00000468, 0.00000155, -0.00000496, 0.00000022, 0.00000643, 0.00000085, -0.00000544, -0.00000073, 0.00000330, -0.00000196, -0.00000456, -0.00000060, 0.00000117, 0.00000187, 0.00000378, 0.00000255, 0.00000098, 0.00000106, -0.00000239, -0.00000572, -0.00000348, -0.00000072, 0.00000063, 0.00000118, -0.00000270, -0.00000584, -0.00000203, 0.00000045, -0.00000258, -0.00000343, 0.00000053, 0.00000572, 0.00000535, },
		{ -0.00000295, 0.00000106, -0.00000780, -0.00001270, -0.00000728, -0.00000724, -0.00000674, -0.00000424, -0.00002305, -0.00004618, -0.00003027, 0.00001122, 0.00003772, 0.00004028, 0.00002552, 0.00001225, 0.00001789, 0.00002076, 0.00000684, 0.00000288, 0.00000866, 0.00000482, 0.00000846, 0.00001750, 0.00000557, 0.00000228, 0.00002990, 0.00003436, 0.00000158, -0.00000860, 0.00000952, 0.00000677, -0.00002001, -0.00003246, -0.00001017, 0.00002094, 0.00001459, -0.00001811, -0.00002030, 0.00000491, 0.00000989, -0.00000432, -0.00001004, -0.00000469, 0.00000220, 0.00000019, -0.00000847, -0.00000616, 0.00000022, -0.00000350, 0.00000137, 0.00001231, -0.00000063, -0.00001557, -0.00000214, 0.00000812, -0.00000291, -0.00000914, -0.00000554, 0.00000234, 0.00001038, -0.00000176, -0.00001436, 0.00000889, 0.00002159, -0.00000947, -0.00001596, 0.00001359, 0.00000747, -0.00001652, -0.00000447, 0.00000686, -0.00000366, -0.00000041, 0.00000367, -0.00001022, -0.00000943, 0.00000526, 0.00000422, -0.00000022, 0.00000087, -0.00000553, -0.00000647, 0.00000390, 0.00000541, 0.00000259, 0.00000433, -0.00000369, -0.00001133, 0.00000051, 0.00000849, -0.00000293, -0.00000702, 0.00000313, 0.00000374, -0.00000472, -0.00000267, 0.00000606, 0.00000378, -0.00000492, -0.00000289, 0.00000384, -0.00000020, -0.00000572, -0.00000150, 0.00000164, 0.00000038, 0.00000338, 0.00000524, 0.00000123, -0.00000059, -0.00000108, -0.00000450, -0.00000500, -0.00000204, 0.00000014, 0.00000194, -0.00000031, -0.00000589, -0.00000440, 0.00000073, -0.00000142, -0.00000444, -0.00000065, 0.00000471, 0.00000576, -0.00000040, -0.00000709, },
		{ -0.00000288, -0.00001265, -0.00001050, -0.00000635, -0.00000829, -0.00000535, -0.00001252, -0.00003951, -0.00004455, -0.00000746, 0.00003174, 0.00004389, 0.00003363, 0.00001696, 0.00001545, 0.00002301, 0.00001343, 0.00000091, 0.00000693, 0.00001126, 0.00000818, 0.00001260, 0.00001081, 0.00000616, 0.00002456, 0.00003569, 0.00000885, -0.00000870, 0.00000708, 0.00000686, -0.00002139, -0.00003234, -0.00000929, 0.00001753, 0.00001314, -0.00001524, -0.00002073, 0.00000287, 0.00001008, -0.00000486, -0.00000934, -0.00000300, -0.00000038, -0.00000139, -0.00000571, -0.00000663, -0.00000116, -0.00000215, -0.00000174, 0.00001051, 0.00000628, -0.00001478, -0.00000927, 0.00001014, 0.00000150, -0.00001315, -0.00000643, 0.00000334, 0.00000584, 0.00000117, -0.00000854, -0.00000025, 0.00001661, 0.00000328, -0.00001414, 0.00000096, 0.00000886, -0.00000676, -0.00000586, 0.00000229, -0.00000437, 0.00000032, 0.00000833, -0.00000817, -0.00001502, 0.00000400, 0.00000749, -0.00000276, 0.00000080, 0.00000048, -0.00000860, -0.00000301, 0.00000696, 0.00000595, 0.00000408, -0.00000134, -0.00001052, -0.00000434, 0.00000732, 0.00000099, -0.00000689, 0.00000068, 0.00000471, -0.00000318, -0.00000464, 0.00000405, 0.00000576, -0.00000269, -0.00000450, 0.00000278, 0.00000184, -0.00000571, -0.00000324, 0.00000216, -0.00000018, 0.00000145, 0.00000728, 0.00000354, -0.00000249, -0.00000140, -0.00000220, -0.00000560, -0.00000405, -0.00000043, 0.00000204, 0.00000154, -0.00000433, -0.00000643, -0.00000053, 0.00000026, -0.00000455, -0.00000233, 0.00000385, 0.00000535, 0.00000162, -0.00000501, -0.00000543, 0.00000081, },
		{ -0.00001359, -0.00000753, -0.00000750, -0.00000837, -0.00000763, -0.00002703, -0.00004890, -0.00002895, 0.00001800, 0.00004485, 0.00004130, 0.00002399, 0.00001537, 0.00002229, 0.00001982, 0.00000329, 0.00000305, 0.00001464, 0.00001248, 0.00000824, 0.00001112, 0.00001228, 0.00002269, 0.00003390, 0.00001547, -0.00000652, 0.00000334, 0.00000687, -0.00002120, -0.00003411, -0.00000893, 0.00001641, 0.00001085, -0.00001361, -0.00001998, 0.00000063, 0.00001010, -0.00000476, -0.00001025, -0.00000122, -0.00000049, -0.00000428, -0.00000441, -0.00000517, -0.00000319, -0.00000141, -0.00000225, 0.00000632, 0.00001000, -0.00000898, -0.00001495, 0.00000657, 0.00000770, -0.00001334, -0.00001105, 0.00000467, 0.00000448, -0.00000070, -0.00000273, -0.00000187, 0.00000728, 0.00000851, -0.00000468, -0.00000643, 0.00000169, 0.00000015, -0.00000075, -0.00000084, -0.00000825, -0.00000070, 0.00001323, -0.00000342, -0.00001979, 0.00000007, 0.00001147, -0.00000361, -0.00000257, 0.00000622, -0.00000584, -0.00001068, 0.00000446, 0.00001042, 0.00000485, -0.00000051, -0.00000796, -0.00000734, 0.00000356, 0.00000353, -0.00000463, -0.00000126, 0.00000410, -0.00000148, -0.00000503, 0.00000154, 0.00000581, 0.00000019, -0.00000431, 0.00000066, 0.00000285, -0.00000426, -0.00000497, 0.00000178, 0.00000043, -0.00000088, 0.00000735, 0.00000727, -0.00000295, -0.00000368, -0.00000008, -0.00000438, -0.00000645, -0.00000160, 0.00000217, 0.00000236, -0.00000213, -0.00000686, -0.00000298, 0.00000113, -0.00000326, -0.00000398, 0.00000258, 0.00000489, 0.00000216, -0.00000226, -0.00000573, -0.00000180, 0.00000043, -0.00000907, },
		{ -0.00000640, -0.00000962, -0.00000848, -0.00001560, -0.00004201, -0.00004581, -0.00000396, 0.00003957, 0.00004832, 0.00003250, 0.00001820, 0.00002065, 0.00002379, 0.00000909, 0.00000055, 0.00001370, 0.00001831, 0.00000772, 0.00000766, 0.00001677, 0.00002524, 0.00003159, 0.00001950, -0.00000235, -0.00000026, 0.00000535, -0.00001934, -0.00003606, -0.00001057, 0.00001685, 0.00000942, -0.00001366, -0.00001875, -0.00000101, 0.00000942, -0.00000392, -0.00001174, -0.00000088, 0.00000148, -0.00000642, -0.00000549, -0.00000286, -0.00000418, -0.00000233, -0.00000101, 0.00000283, 0.00000923, -0.00000172, -0.00001558, -0.00000079, 0.00001106, -0.00000856, -0.00001612, 0.00000282, 0.00000646, -0.00000349, -0.00000109, 0.00000204, 0.00000048, 0.00000503, 0.00000511, -0.00000454, -0.00000824, -0.00000083, 0.00000732, 0.00000144, -0.00001353, -0.00000532, 0.00001725, 0.00000411, -0.00002251, -0.00000636, 0.00001477, -0.00000173, -0.00000804, 0.00000918, 0.00000123, -0.00001604, -0.00000230, 0.00001394, 0.00000766, -0.00000093, -0.00000519, -0.00000752, -0.00000123, 0.00000338, -0.00000132, -0.00000156, 0.00000251, -0.00000068, -0.00000422, -0.00000013, 0.00000414, 0.00000208, -0.00000212, -0.00000099, 0.00000198, -0.00000221, -0.00000549, 0.00000022, 0.00000138, -0.00000219, 0.00000534, 0.00001065, -0.00000083, -0.00000680, 0.00000021, -0.00000139, -0.00000806, -0.00000397, 0.00000245, 0.00000269, -0.00000060, -0.00000552, -0.00000519, 0.00000017, -0.00000124, -0.00000448, 0.00000065, 0.00000439, 0.00000186, -0.00000041, -0.00000371, -0.00000435, 0.00000039, -0.00000377, -0.00001377, -0.00001116, },
		{ -0.00001139, -0.00001035, -0.00002842, -0.00005126, -0.00002942, 0.00002418, 0.00005239, 0.00004253, 0.00002366, 0.00001989, 0.00002510, 0.00001585, 0.00000162, 0.00001010, 0.00002212, 0.00001156, 0.00000386, 0.00001713, 0.00003032, 0.00003174, 0.00002086, 0.00000209, -0.00000232, 0.00000205, -0.00001705, -0.00003651, -0.00001418, 0.00001706, 0.00000992, -0.00001468, -0.00001821, -0.00000160, 0.00000824, -0.00000298, -0.00001271, -0.00000229, 0.00000389, -0.00000629, -0.00000819, -0.00000156, -0.00000329, -0.00000425, -0.00000001, 0.00000182, 0.00000558, 0.00000301, -0.00001091, -0.00000754, 0.00000891, -0.00000143, -0.00001762, -0.00000277, 0.00000876, -0.00000358, -0.00000368, 0.00000643, 0.00000008, -0.00000286, 0.00000859, 0.00000454, -0.00001348, -0.00000966, 0.00001210, 0.00001000, -0.00001613, -0.00001403, 0.00001808, 0.00001398, -0.00002176, -0.00001459, 0.00001592, 0.00000304, -0.00001365, 0.00000778, 0.00001050, -0.00001668, -0.00001185, 0.00001441, 0.00001243, -0.00000154, -0.00000361, -0.00000503, -0.00000512, 0.00000031, 0.00000145, 0.00000012, 0.00000103, -0.00000117, -0.00000320, -0.00000032, 0.00000196, 0.00000187, 0.00000093, -0.00000062, -0.00000033, -0.00000116, -0.00000421, -0.00000165, 0.00000140, -0.00000196, 0.00000250, 0.00001191, 0.00000346, -0.00000868, -0.00000212, 0.00000198, -0.00000750, -0.00000738, 0.00000213, 0.00000343, -0.00000026, -0.00000351, -0.00000565, -0.00000233, -0.00000004, -0.00000319, -0.00000129, 0.00000327, 0.00000146, -0.00000002, -0.00000082, -0.00000473, -0.00000237, 0.00000012, -0.00000996, -0.00001502, -0.00000772, -0.00000816, },
		{ -0.00001583, -0.00004361, -0.00004970, -0.00000175, 0.00004865, 0.00005349, 0.00003201, 0.00002083, 0.00002482, 0.00002147, 0.00000618, 0.00000682, 0.00002219, 0.00001752, 0.00000301, 0.00001377, 0.00003453, 0.00003569, 0.00002148, 0.00000494, -0.00000269, -0.00000203, -0.00001585, -0.00003496, -0.00001834, 0.00001542, 0.00001203, -0.00001516, -0.00001907, -0.00000154, 0.00000732, -0.00000257, -0.00001283, -0.00000462, 0.00000511, -0.00000390, -0.00001058, -0.00000246, -0.00000121, -0.00000556, -0.00000058, 0.00000288, 0.00000206, 0.00000333, -0.00000396, -0.00000976, 0.00000232, 0.00000357, -0.00001393, -0.00000930, 0.00000793, -0.00000035, -0.00000723, 0.00000710, 0.00000494, -0.00000873, 0.00000386, 0.00001417, -0.00000967, -0.00002101, 0.00000853, 0.00002124, -0.00001191, -0.00002453, 0.00001319, 0.00002439, -0.00001619, -0.00002315, 0.00001361, 0.00000992, -0.00001722, 0.00000184, 0.00001902, -0.00001160, -0.00002153, 0.00001053, 0.00001806, -0.00000110, -0.00000394, -0.00000096, -0.00000649, -0.00000485, 0.00000224, 0.00000326, 0.00000065, -0.00000256, -0.00000291, 0.00000058, 0.00000069, -0.00000027, 0.00000294, 0.00000224, -0.00000235, -0.00000238, -0.00000191, -0.00000229, -0.00000015, -0.00000103, 0.00000063, 0.00001052, 0.00000793, -0.00000760, -0.00000611, 0.00000358, -0.00000441, -0.00001045, 0.00000015, 0.00000481, -0.00000041, -0.00000236, -0.00000403, -0.00000474, -0.00000103, -0.00000080, -0.00000187, 0.00000120, 0.00000095, -0.00000047, 0.00000126, -0.00000259, -0.00000538, 0.00000044, -0.00000384, -0.00001589, -0.00001216, -0.00000622, -0.00001218, -0.00001158, },
		{ -0.00005673, -0.00003271, 0.00003207, 0.00006203, 0.00004406, 0.00002407, 0.00002402, 0.00002521, 0.00001260, 0.00000612, 0.00001932, 0.00002244, 0.00000617, 0.00000941, 0.00003529, 0.00004226, 0.00002406, 0.00000577, -0.00000257, -0.00000558, -0.00001637, -0.00003224, -0.00002135, 0.00001157, 0.00001425, -0.00001387, -0.00002094, -0.00000188, 0.00000724, -0.00000270, -0.00001258, -0.00000674, 0.00000451, -0.00000063, -0.00001099, -0.00000521, 0.00000034, -0.00000514, -0.00000246, 0.00000415, 0.00000080, 0.00000032, 0.00000125, -0.00000650, -0.00000471, 0.00000328, -0.00000703, -0.00001275, 0.00000295, 0.00000359, -0.00000812, 0.00000342, 0.00001051, -0.00000836, -0.00000550, 0.00001762, 0.00000167, -0.00002742, -0.00000365, 0.00002865, 0.00000016, -0.00003193, 0.00000168, 0.00003193, -0.00000534, -0.00002968, 0.00000712, 0.00001721, -0.00001696, -0.00000740, 0.00002392, -0.00000150, -0.00002840, 0.00000230, 0.00002281, 0.00000126, -0.00000609, 0.00000314, -0.00000446, -0.00001060, 0.00000019, 0.00000684, 0.00000179, -0.00000411, -0.00000357, 0.00000161, 0.00000092, -0.00000298, 0.00000252, 0.00000635, -0.00000208, -0.00000565, -0.00000049, -0.00000066, -0.00000247, -0.00000105, 0.00000071, 0.00000777, 0.00001025, -0.00000357, -0.00000932, 0.00000201, -0.00000012, -0.00001127, -0.00000363, 0.00000586, 0.00000018, -0.00000278, -0.00000152, -0.00000521, -0.00000409, 0.00000083, -0.00000015, -0.00000105, -0.00000042, -0.00000098, 0.00000186, 0.00000067, -0.00000622, -0.00000246, 0.00000118, -0.00001217, -0.00001743, -0.00000708, -0.00000955, -0.00001521, -0.00001028, -0.00002782, },
		{ 0.00000172, 0.00006171, 0.00005935, 0.00003106, 0.00002336, 0.00002725, 0.00001924, 0.00000853, 0.00001582, 0.00002440, 0.00001188, 0.00000711, 0.00003239, 0.00004858, 0.00003012, 0.00000597, -0.00000359, -0.00000819, -0.00001805, -0.00002977, -0.00002237, 0.00000654, 0.00001492, -0.00001076, -0.00002256, -0.00000348, 0.00000787, -0.00000274, -0.00001264, -0.00000811, 0.00000267, 0.00000184, -0.00000903, -0.00000815, -0.00000007, -0.00000328, -0.00000418, 0.00000405, 0.00000184, -0.00000322, 0.00000233, -0.00000015, -0.00000799, -0.00000203, -0.00000122, -0.00001088, -0.00000384, 0.00000471, -0.00000526, -0.00000172, 0.00001236, -0.00000222, -0.00001317, 0.00001233, 0.00001415, -0.00002397, -0.00001923, 0.00002651, 0.00001643, -0.00003115, -0.00001407, 0.00003265, 0.00000927, -0.00003131, -0.00000304, 0.00002248, -0.00001202, -0.00001749, 0.00002316, 0.00001135, -0.00003001, -0.00000895, 0.00002490, 0.00000583, -0.00000937, 0.00000563, 0.00000096, -0.00001509, -0.00000490, 0.00000975, 0.00000445, -0.00000531, -0.00000480, 0.00000203, 0.00000214, -0.00000468, -0.00000020, 0.00000950, 0.00000130, -0.00000905, -0.00000174, 0.00000266, -0.00000364, -0.00000313, 0.00000199, 0.00000589, 0.00000937, 0.00000134, -0.00000936, -0.00000212, 0.00000275, -0.00000884, -0.00000778, 0.00000506, 0.00000202, -0.00000399, -0.00000002, -0.00000308, -0.00000743, -0.00000013, 0.00000326, -0.00000179, -0.00000316, -0.00000160, 0.00000162, 0.00000317, -0.00000428, -0.00000601, 0.00000240, -0.00000535, -0.00001981, -0.00001196, -0.00000593, -0.00001628, -0.00001246, -0.00001273, -0.00004805, -0.00005845, },
		{ 0.00007380, 0.00004383, 0.00002392, 0.00002772, 0.00002524, 0.00001325, 0.00001373, 0.00002355, 0.00001754, 0.00000826, 0.00002789, 0.00005213, 0.00003901, 0.00000788, -0.00000635, -0.00001066, -0.00001984, -0.00002836, -0.00002178, 0.00000190, 0.00001333, -0.00000697, -0.00002264, -0.00000628, 0.00000832, -0.00000201, -0.00001319, -0.00000902, 0.00000068, 0.00000263, -0.00000577, -0.00000950, -0.00000247, -0.00000148, -0.00000442, 0.00000245, 0.00000359, -0.00000484, -0.00000029, 0.00000537, -0.00000584, -0.00000888, -0.00000017, -0.00000498, -0.00000843, 0.00000139, -0.00000081, -0.00000452, 0.00000925, 0.00000539, -0.00001436, 0.00000129, 0.00002056, -0.00001146, -0.00003027, 0.00001378, 0.00002976, -0.00002010, -0.00002844, 0.00002425, 0.00002369, -0.00002581, -0.00001459, 0.00002325, -0.00000312, -0.00002535, 0.00001621, 0.00002367, -0.00002513, -0.00002083, 0.00002298, 0.00001217, -0.00001253, 0.00000516, 0.00000883, -0.00001646, -0.00001254, 0.00001085, 0.00000847, -0.00000600, -0.00000620, 0.00000200, 0.00000320, -0.00000478, -0.00000361, 0.00000988, 0.00000658, -0.00001007, -0.00000583, 0.00000552, -0.00000206, -0.00000669, 0.00000243, 0.00000646, 0.00000644, 0.00000420, -0.00000562, -0.00000617, 0.00000204, -0.00000427, -0.00000987, 0.00000173, 0.00000410, -0.00000433, -0.00000059, 0.00000043, -0.00000872, -0.00000391, 0.00000622, 0.00000036, -0.00000650, -0.00000329, 0.00000148, 0.00000411, -0.00000097, -0.00000741, -0.00000015, 0.00000086, -0.00001699, -0.00001863, -0.00000531, -0.00001288, -0.00001785, -0.00000572, -0.00002888, -0.00006900, -0.00003636, 0.00004576, },
		{ 0.00002830, 0.00002665, 0.00003011, 0.00001936, 0.00001387, 0.00002148, 0.00002127, 0.00001202, 0.00002443, 0.00005230, 0.00004844, 0.00001309, -0.00000987, -0.00001435, -0.00002125, -0.00002785, -0.00002073, -0.00000127, 0.00001007, -0.00000408, -0.00002075, -0.00000929, 0.00000767, -0.00000042, -0.00001374, -0.00001014, -0.00000077, 0.00000205, -0.00000283, -0.00000856, -0.00000553, -0.00000128, -0.00000310, 0.00000055, 0.00000429, -0.00000402, -0.00000408, 0.00000720, 0.00000008, -0.00001295, -0.00000443, 0.00000100, -0.00000799, -0.00000467, 0.00000160, -0.00000323, 0.00000362, 0.00000964, -0.00000874, -0.00000909, 0.00001735, 0.00000392, -0.00003073, -0.00000472, 0.00003318, -0.00000184, -0.00003493, 0.00000816, 0.00003259, -0.00001329, -0.00002374, 0.00001817, 0.00000733, -0.00002820, 0.00000443, 0.00003194, -0.00001423, -0.00003042, 0.00001650, 0.00001912, -0.00001408, 0.00000099, 0.00001726, -0.00001316, -0.00002146, 0.00000899, 0.00001364, -0.00000605, -0.00000791, 0.00000227, 0.00000353, -0.00000417, -0.00000576, 0.00000755, 0.00001108, -0.00000735, -0.00001078, 0.00000548, 0.00000212, -0.00000954, 0.00000029, 0.00000900, 0.00000412, 0.00000322, -0.00000009, -0.00000698, -0.00000225, -0.00000051, -0.00000813, -0.00000287, 0.00000437, -0.00000282, -0.00000247, 0.00000301, -0.00000681, -0.00000867, 0.00000631, 0.00000509, -0.00000842, -0.00000691, 0.00000144, 0.00000421, 0.00000158, -0.00000571, -0.00000383, 0.00000313, -0.00001012, -0.00002275, -0.00000971, -0.00000754, -0.00002046, -0.00000826, -0.00000948, -0.00006051, -0.00007015, 0.00001198, 0.00007927, 0.00006270, },
		{ 0.00003299, 0.00002626, 0.00001614, 0.00001973, 0.00002286, 0.00001644, 0.00002346, 0.00005054, 0.00005611, 0.00002136, -0.00001219, -0.00001994, -0.00002280, -0.00002746, -0.00002006, -0.00000300, 0.00000646, -0.00000304, -0.00001768, -0.00001125, 0.00000566, 0.00000136, -0.00001357, -0.00001181, -0.00000179, 0.00000107, -0.00000129, -0.00000604, -0.00000748, -0.00000309, -0.00000143, -0.00000024, 0.00000342, -0.00000201, -0.00000645, 0.00000513, 0.00000611, -0.00001190, -0.00001123, 0.00000305, -0.00000294, -0.00000963, -0.00000028, 0.00000060, -0.00000056, 0.00000845, -0.00000022, -0.00001321, 0.00000666, 0.00001428, -0.00002021, -0.00002053, 0.00002415, 0.00001619, -0.00002967, -0.00001008, 0.00003171, 0.00000294, -0.00002650, 0.00000815, 0.00001573, -0.00002473, -0.00000900, 0.00003347, 0.00000042, -0.00003492, 0.00000611, 0.00002492, -0.00001263, -0.00000649, 0.00002363, -0.00000459, -0.00002950, 0.00000316, 0.00001937, -0.00000484, -0.00001068, 0.00000332, 0.00000386, -0.00000450, -0.00000601, 0.00000452, 0.00001244, -0.00000199, -0.00001345, 0.00000178, 0.00000656, -0.00000933, -0.00000429, 0.00001118, 0.00000475, -0.00000081, 0.00000371, -0.00000309, -0.00000744, -0.00000061, -0.00000313, -0.00000586, 0.00000145, -0.00000035, -0.00000361, 0.00000307, -0.00000281, -0.00001155, 0.00000273, 0.00001007, -0.00000694, -0.00001187, 0.00000031, 0.00000469, 0.00000231, -0.00000235, -0.00000536, 0.00000098, -0.00000322, -0.00002115, -0.00001710, -0.00000508, -0.00001738, -0.00001548, 0.00000094, -0.00003815, -0.00008717, -0.00003349, 0.00006669, 0.00008337, 0.00004007, 0.00002513, },
		{ 0.00002042, 0.00001906, 0.00002324, 0.00001997, 0.00002461, 0.00004899, 0.00006126, 0.00003093, -0.00001180, -0.00002677, -0.00002561, -0.00002669, -0.00001983, -0.00000409, 0.00000362, -0.00000367, -0.00001476, -0.00001150, 0.00000295, 0.00000240, -0.00001236, -0.00001368, -0.00000302, 0.00000049, -0.00000109, -0.00000343, -0.00000735, -0.00000589, -0.00000095, 0.00000051, 0.00000188, -0.00000062, -0.00000638, 0.00000124, 0.00000906, -0.00000658, -0.00001633, -0.00000023, 0.00000329, -0.00001016, -0.00000558, 0.00000346, -0.00000079, 0.00000356, 0.00000599, -0.00000962, -0.00000523, 0.00001473, -0.00000428, -0.00002624, 0.00000639, 0.00002574, -0.00001418, -0.00002288, 0.00002058, 0.00001703, -0.00002085, -0.00000336, 0.00001879, -0.00001599, -0.00002002, 0.00002747, 0.00001506, -0.00003253, -0.00000619, 0.00002757, -0.00000753, -0.00001555, 0.00002526, 0.00000836, -0.00003379, -0.00000695, 0.00002426, -0.00000120, -0.00001504, 0.00000437, 0.00000589, -0.00000661, -0.00000577, 0.00000345, 0.00001043, 0.00000277, -0.00001168, -0.00000365, 0.00000812, -0.00000560, -0.00000893, 0.00001047, 0.00000838, -0.00000486, 0.00000306, 0.00000384, -0.00000959, -0.00000544, 0.00000201, -0.00000454, -0.00000383, 0.00000062, -0.00000235, 0.00000112, 0.00000059, -0.00001073, -0.00000279, 0.00001222, -0.00000183, -0.00001591, -0.00000327, 0.00000578, 0.00000203, 0.00000018, -0.00000330, -0.00000287, -0.00000039, -0.00001441, -0.00002256, -0.00000859, -0.00001100, -0.00002008, -0.00000040, -0.00001365, -0.00008142, -0.00007691, 0.00003182, 0.00009613, 0.00006097, 0.00002663, 0.00003290, 0.00003318, },
		{ 0.00002332, 0.00002232, 0.00002647, 0.00004888, 0.00006487, 0.00003990, -0.00000867, -0.00003328, -0.00003037, -0.00002591, -0.00001949, -0.00000531, 0.00000175, -0.00000496, -0.00001305, -0.00001041, 0.00000067, 0.00000220, -0.00001055, -0.00001503, -0.00000490, 0.00000035, -0.00000140, -0.00000192, -0.00000553, -0.00000803, -0.00000234, 0.00000192, 0.00000105, -0.00000069, -0.00000473, -0.00000182, 0.00000809, -0.00000024, -0.00001682, -0.00000685, 0.00000658, -0.00000582, -0.00001086, 0.00000252, 0.00000223, -0.00000112, 0.00000680, -0.00000150, -0.00001191, 0.00000610, 0.00000873, -0.00001971, -0.00001166, 0.00002227, 0.00000469, -0.00002442, 0.00000354, 0.00002306, -0.00000838, -0.00001148, 0.00001530, -0.00000533, -0.00002521, 0.00001573, 0.00002553, -0.00002341, -0.00001733, 0.00002553, 0.00000056, -0.00002341, 0.00002049, 0.00002288, -0.00003158, -0.00002014, 0.00002600, 0.00000585, -0.00002029, 0.00000338, 0.00001097, -0.00000931, -0.00000774, 0.00000549, 0.00000763, 0.00000356, -0.00000628, -0.00000691, 0.00000510, -0.00000068, -0.00001047, 0.00000632, 0.00001250, -0.00000571, -0.00000188, 0.00000983, -0.00000619, -0.00001259, 0.00000332, 0.00000136, -0.00000833, -0.00000221, 0.00000084, -0.00000039, 0.00000107, -0.00000696, -0.00000690, 0.00000996, 0.00000457, -0.00001634, -0.00000915, 0.00000610, 0.00000238, 0.00000041, 0.00000093, -0.00000456, -0.00000280, -0.00000680, -0.00002180, -0.00001655, -0.00000705, -0.00001776, -0.00000851, 0.00000140, -0.00005744, -0.00010278, -0.00001989, 0.00008919, 0.00008714, 0.00003644, 0.00003040, 0.00003837, 0.00002678, 0.00001981, },
		{ 0.00002799, 0.00005000, 0.00006857, 0.00004761, -0.00000414, -0.00003819, -0.00003664, -0.00002599, -0.00001861, -0.00000682, 0.00000035, -0.00000591, -0.00001264, -0.00000906, -0.00000040, 0.00000110, -0.00000899, -0.00001538, -0.00000727, 0.00000008, -0.00000139, -0.00000170, -0.00000333, -0.00000838, -0.00000495, 0.00000252, 0.00000162, -0.00000169, -0.00000321, -0.00000261, 0.00000476, 0.00000379, -0.00001279, -0.00001289, 0.00000479, 0.00000067, -0.00001251, -0.00000217, 0.00000534, -0.00000243, 0.00000300, 0.00000580, -0.00001058, -0.00000591, 0.00001254, -0.00000527, -0.00002142, 0.00000773, 0.00001793, -0.00001402, -0.00001169, 0.00001830, 0.00000575, -0.00001231, 0.00000715, 0.00000289, -0.00002342, 0.00000216, 0.00002877, -0.00000996, -0.00002400, 0.00001866, 0.00000949, -0.00002719, 0.00000968, 0.00003483, -0.00002157, -0.00003337, 0.00002216, 0.00001595, -0.00002414, -0.00000188, 0.00001849, -0.00000954, -0.00001388, 0.00000889, 0.00000794, -0.00000056, -0.00000115, -0.00000465, -0.00000081, 0.00000138, -0.00000748, 0.00000114, 0.00001368, -0.00000251, -0.00000790, 0.00001107, 0.00000182, -0.00001746, -0.00000124, 0.00000889, -0.00000834, -0.00000877, 0.00000319, 0.00000101, -0.00000134, -0.00000327, -0.00000691, 0.00000463, 0.00000855, -0.00001216, -0.00001499, 0.00000370, 0.00000400, -0.00000087, 0.00000437, -0.00000196, -0.00000778, -0.00000335, -0.00001457, -0.00002350, -0.00001001, -0.00001042, -0.00001464, 0.00000230, -0.00002834, -0.00010207, -0.00007322, 0.00005543, 0.00010768, 0.00005827, 0.00002943, 0.00003959, 0.00003454, 0.00002264, 0.00002351, 0.00002398, },
		{ 0.00007324, 0.00005486, 0.00000037, -0.00004147, -0.00004325, -0.00002750, -0.00001742, -0.00000826, -0.00000122, -0.00000616, -0.00001283, -0.00000841, -0.00000034, 0.00000001, -0.00000830, -0.00001491, -0.00000946, -0.00000083, -0.00000086, -0.00000208, -0.00000196, -0.00000703, -0.00000730, 0.00000143, 0.00000303, -0.00000233, -0.00000294, -0.00000159, 0.00000163, 0.00000409, -0.00000692, -0.00001507, -0.00000093, 0.00000529, -0.00000930, -0.00000780, 0.00000539, -0.00000016, -0.00000187, 0.00000830, -0.00000334, -0.00001433, 0.00000631, 0.00000884, -0.00001883, -0.00001029, 0.00001913, 0.00000296, -0.00001779, 0.00000511, 0.00001492, -0.00000522, -0.00000117, 0.00000551, -0.00001648, -0.00000860, 0.00002425, 0.00000362, -0.00002402, 0.00000874, 0.00001630, -0.00002529, -0.00000440, 0.00004011, -0.00000505, -0.00004229, 0.00001164, 0.00002672, -0.00002350, -0.00001199, 0.00002537, -0.00000405, -0.00002339, 0.00000950, 0.00001392, -0.00000675, -0.00000120, 0.00000322, -0.00000491, -0.00000245, -0.00000203, -0.00000094, 0.00001036, 0.00000225, -0.00001079, 0.00000675, 0.00001015, -0.00001625, -0.00000990, 0.00001322, -0.00000239, -0.00001595, 0.00000142, 0.00000561, -0.00000399, -0.00000286, -0.00000293, 0.00000001, 0.00000748, -0.00000540, -0.00001736, -0.00000197, 0.00000553, -0.00000154, 0.00000494, 0.00000370, -0.00001080, -0.00000620, -0.00000524, -0.00002384, -0.00001926, -0.00000463, -0.00001261, -0.00000634, -0.00000854, -0.00007786, -0.00010926, -0.00000133, 0.00010789, 0.00008952, 0.00003692, 0.00003635, 0.00004134, 0.00002849, 0.00002420, 0.00002524, 0.00002909, 0.00005133, },
		{ 0.00000447, -0.00004422, -0.00004928, -0.00003009, -0.00001654, -0.00000932, -0.00000319, -0.00000601, -0.00001278, -0.00000876, 0.00000006, -0.00000032, -0.00000839, -0.00001432, -0.00001091, -0.00000239, -0.00000025, -0.00000224, -0.00000181, -0.00000503, -0.00000817, -0.00000101, 0.00000399, -0.00000176, -0.00000372, -0.00000027, 0.00000043, 0.00000168, -0.00000252, -0.00001274, -0.00000714, 0.00000536, -0.00000314, -0.00001074, 0.00000160, 0.00000315, -0.00000425, 0.00000566, 0.00000472, -0.00001509, -0.00000527, 0.00001524, -0.00000646, -0.00002256, 0.00000806, 0.00001759, -0.00001184, -0.00000968, 0.00001453, 0.00000622, -0.00000475, 0.00000234, -0.00000846, -0.00001333, 0.00001439, 0.00001291, -0.00001753, -0.00000090, 0.00001859, -0.00001834, -0.00001750, 0.00003649, 0.00001389, -0.00004296, -0.00000404, 0.00003414, -0.00001631, -0.00002478, 0.00002720, 0.00000814, -0.00003217, 0.00000320, 0.00002439, -0.00000947, -0.00000906, 0.00001245, -0.00000202, -0.00001143, 0.00000087, 0.00000297, 0.00000467, 0.00000411, -0.00000849, 0.00000005, 0.00001379, -0.00000885, -0.00001762, 0.00001089, 0.00000709, -0.00001909, -0.00000556, 0.00001064, -0.00000345, -0.00000655, 0.00000185, -0.00000017, 0.00000204, -0.00000037, -0.00001422, -0.00000875, 0.00000435, -0.00000025, 0.00000321, 0.00000894, -0.00000879, -0.00001296, 0.00000016, -0.00001610, -0.00002898, -0.00000659, -0.00000328, -0.00001425, -0.00000546, -0.00004432, -0.00011505, -0.00006497, 0.00007689, 0.00011869, 0.00005906, 0.00003248, 0.00004358, 0.00003729, 0.00002671, 0.00002589, 0.00003011, 0.00005232, 0.00007852, 0.00006292, },
		{ -0.00005490, -0.00003282, -0.00001633, -0.00001020, -0.00000535, -0.00000596, -0.00001209, -0.00000963, -0.00000002, 0.00000025, -0.00000863, -0.00001424, -0.00001158, -0.00000407, -0.00000013, -0.00000184, -0.00000239, -0.00000360, -0.00000738, -0.00000356, 0.00000350, -0.00000022, -0.00000447, 0.00000006, 0.00000115, -0.00000110, -0.00000145, -0.00000794, -0.00001040, 0.00000118, 0.00000225, -0.00000909, -0.00000393, 0.00000427, -0.00000300, 0.00000078, 0.00000894, -0.00000896, -0.00001516, 0.00001137, 0.00000813, -0.00002331, -0.00000915, 0.00002214, 0.00000282, -0.00001814, 0.00000474, 0.00001574, -0.00000113, -0.00000347, -0.00000358, -0.00001179, 0.00000367, 0.00001522, -0.00000732, -0.00000672, 0.00001584, -0.00000912, -0.00002561, 0.00002494, 0.00002961, -0.00003387, -0.00002060, 0.00003442, -0.00000330, -0.00003569, 0.00002073, 0.00002427, -0.00003462, -0.00001107, 0.00003418, -0.00000367, -0.00002265, 0.00001620, 0.00000988, -0.00002035, -0.00000385, 0.00001174, 0.00000178, -0.00000008, -0.00000317, -0.00000367, 0.00001056, 0.00000044, -0.00001915, 0.00000258, 0.00001466, -0.00001522, -0.00001507, 0.00001184, 0.00000168, -0.00001176, 0.00000323, 0.00000497, -0.00000379, -0.00000089, -0.00000698, -0.00001254, -0.00000104, 0.00000184, 0.00000183, 0.00001083, -0.00000254, -0.00001837, -0.00000182, -0.00000441, -0.00003211, -0.00001709, 0.00000611, -0.00001285, -0.00001553, -0.00001904, -0.00009114, -0.00011301, 0.00001577, 0.00012817, 0.00009538, 0.00003635, 0.00003921, 0.00004636, 0.00003304, 0.00002609, 0.00003080, 0.00005331, 0.00008357, 0.00007229, 0.00000908, -0.00004753, },
		{ -0.00001630, -0.00001135, -0.00000760, -0.00000617, -0.00001088, -0.00001038, -0.00000094, 0.00000118, -0.00000833, -0.00001480, -0.00001199, -0.00000526, -0.00000063, -0.00000121, -0.00000291, -0.00000326, -0.00000575, -0.00000500, 0.00000161, 0.00000122, -0.00000421, -0.00000072, 0.00000251, -0.00000234, -0.00000324, -0.00000381, -0.00000942, -0.00000438, 0.00000384, -0.00000394, -0.00000776, 0.00000175, 0.00000006, -0.00000264, 0.00000788, -0.00000042, -0.00001821, 0.00000055, 0.00001690, -0.00001319, -0.00002335, 0.00001423, 0.00001814, -0.00001550, -0.00000915, 0.00001768, 0.00000806, -0.00000712, -0.00000382, -0.00000675, -0.00000357, 0.00001087, 0.00000225, -0.00000673, 0.00000995, -0.00000131, -0.00002686, 0.00000941, 0.00003723, -0.00001719, -0.00003241, 0.00002617, 0.00001172, -0.00003990, 0.00000612, 0.00003847, -0.00002674, -0.00002973, 0.00003655, 0.00001166, -0.00003529, 0.00000899, 0.00002729, -0.00002184, -0.00001748, 0.00001976, 0.00000674, -0.00000918, -0.00000039, -0.00000061, 0.00000322, 0.00000525, -0.00001284, -0.00000648, 0.00001522, -0.00000576, -0.00002168, 0.00000665, 0.00000905, -0.00001412, -0.00000082, 0.00001240, -0.00000513, -0.00000740, -0.00000029, -0.00001002, -0.00000902, 0.00000153, 0.00000305, 0.00000947, 0.00000399, -0.00001821, -0.00000938, 0.00000414, -0.00002551, -0.00003034, 0.00000738, -0.00000130, -0.00002716, -0.00001308, -0.00005208, -0.00012759, -0.00005936, 0.00010263, 0.00013465, 0.00005694, 0.00003128, 0.00005083, 0.00004407, 0.00002752, 0.00003019, 0.00005474, 0.00008824, 0.00008230, 0.00001537, -0.00005142, -0.00006127, -0.00003507, },
		{ -0.00001023, -0.00000659, -0.00000943, -0.00001063, -0.00000246, 0.00000181, -0.00000727, -0.00001558, -0.00001273, -0.00000580, -0.00000135, -0.00000089, -0.00000296, -0.00000371, -0.00000441, -0.00000493, -0.00000070, 0.00000153, -0.00000291, -0.00000176, 0.00000317, -0.00000165, -0.00000589, -0.00000252, -0.00000559, -0.00000793, 0.00000119, 0.00000144, -0.00000753, -0.00000310, 0.00000188, -0.00000273, 0.00000357, 0.00000552, -0.00001380, -0.00001056, 0.00001577, 0.00000144, -0.00002746, -0.00000183, 0.00002570, -0.00000297, -0.00001939, 0.00001029, 0.00001758, -0.00000498, -0.00000772, -0.00000240, -0.00000526, 0.00000308, 0.00000725, -0.00000166, 0.00000436, 0.00000249, -0.00002236, -0.00000481, 0.00003486, 0.00000173, -0.00003506, 0.00001166, 0.00002326, -0.00003497, -0.00001244, 0.00004445, -0.00000894, -0.00004557, 0.00002703, 0.00003195, -0.00003897, -0.00000958, 0.00004206, -0.00001072, -0.00003554, 0.00001957, 0.00002060, -0.00001736, -0.00000561, 0.00000820, -0.00000168, 0.00000139, -0.00000258, -0.00000944, 0.00000785, 0.00000351, -0.00002068, -0.00000317, 0.00001360, -0.00001078, -0.00000823, 0.00001697, 0.00000030, -0.00001607, 0.00000069, -0.00000150, -0.00001503, -0.00000390, 0.00000632, 0.00000794, 0.00000678, -0.00001250, -0.00001669, 0.00000449, -0.00001264, -0.00003776, -0.00000227, 0.00001296, -0.00002890, -0.00002485, -0.00001878, -0.00010541, -0.00012260, 0.00003910, 0.00015669, 0.00009682, 0.00002909, 0.00004665, 0.00005742, 0.00003329, 0.00002780, 0.00005602, 0.00009342, 0.00009206, 0.00002353, -0.00005471, -0.00006941, -0.00003737, -0.00001549, -0.00001288, },
		{ -0.00000779, -0.00001040, -0.00000412, 0.00000181, -0.00000574, -0.00001601, -0.00001405, -0.00000601, -0.00000173, -0.00000110, -0.00000274, -0.00000423, -0.00000396, -0.00000396, -0.00000228, 0.00000053, -0.00000142, -0.00000208, 0.00000273, -0.00000001, -0.00000741, -0.00000411, -0.00000174, -0.00000771, -0.00000352, 0.00000390, -0.00000358, -0.00000720, 0.00000051, -0.00000044, -0.00000043, 0.00000641, -0.00000552, -0.00001602, 0.00000649, 0.00001239, -0.00002024, -0.00001755, 0.00002132, 0.00001277, -0.00001999, -0.00000291, 0.00002147, 0.00000296, -0.00001126, -0.00000178, -0.00000253, -0.00000377, 0.00000617, 0.00000528, 0.00000223, 0.00000208, -0.00001549, -0.00001358, 0.00002453, 0.00001652, -0.00002768, -0.00000370, 0.00002693, -0.00002254, -0.00002844, 0.00003890, 0.00001342, -0.00005123, 0.00000659, 0.00004869, -0.00002890, -0.00003347, 0.00004531, 0.00001221, -0.00004908, 0.00000644, 0.00003843, -0.00001699, -0.00001999, 0.00001650, 0.00000220, -0.00000965, 0.00000385, -0.00000241, -0.00000256, 0.00000564, -0.00001187, -0.00001171, 0.00001100, -0.00000316, -0.00001385, 0.00001509, 0.00001008, -0.00002092, -0.00000606, 0.00000841, -0.00001429, -0.00001387, 0.00000802, 0.00000968, 0.00000509, -0.00000552, -0.00001806, -0.00000261, -0.00000153, -0.00003410, -0.00001762, 0.00001966, -0.00001735, -0.00004190, -0.00000686, -0.00006038, -0.00015038, -0.00004760, 0.00014117, 0.00014614, 0.00004456, 0.00003531, 0.00006710, 0.00004571, 0.00002535, 0.00005528, 0.00010000, 0.00010172, 0.00003249, -0.00005613, -0.00007905, -0.00004116, -0.00001344, -0.00001402, -0.00001372, -0.00000729, },
		{ -0.00000569, 0.00000127, -0.00000422, -0.00001585, -0.00001572, -0.00000641, -0.00000149, -0.00000156, -0.00000274, -0.00000440, -0.00000423, -0.00000307, -0.00000263, -0.00000106, -0.00000070, -0.00000145, 0.00000177, 0.00000127, -0.00000706, -0.00000696, -0.00000017, -0.00000446, -0.00000710, 0.00000237, 0.00000145, -0.00000795, -0.00000349, 0.00000152, -0.00000167, 0.00000329, 0.00000160, -0.00001390, -0.00000501, 0.00001429, -0.00000657, -0.00002504, 0.00000737, 0.00002322, -0.00001018, -0.00001487, 0.00001659, 0.00001282, -0.00001042, -0.00000495, 0.00000118, -0.00000648, 0.00000062, 0.00001004, 0.00000458, -0.00000031, -0.00000992, -0.00001565, 0.00001114, 0.00002272, -0.00001362, -0.00001389, 0.00002186, -0.00000773, -0.00003631, 0.00002344, 0.00003224, -0.00004322, -0.00001811, 0.00005370, -0.00000671, -0.00005266, 0.00003246, 0.00003933, -0.00004916, -0.00001790, 0.00005098, -0.00000340, -0.00003841, 0.00001611, 0.00001633, -0.00002059, -0.00000056, 0.00001208, -0.00000748, -0.00000270, -0.00000086, -0.00001226, 0.00000144, 0.00000313, -0.00001306, 0.00000779, 0.00001830, -0.00001789, -0.00001736, 0.00001322, -0.00000551, -0.00002373, 0.00000402, 0.00001521, 0.00000223, -0.00000263, -0.00001231, -0.00001090, 0.00000083, -0.00002162, -0.00002870, 0.00001335, 0.00000053, -0.00004959, -0.00001791, -0.00001657, -0.00013404, -0.00012869, 0.00007969, 0.00018287, 0.00008409, 0.00002632, 0.00006686, 0.00006325, 0.00002693, 0.00005106, 0.00010726, 0.00011288, 0.00004107, -0.00005569, -0.00008867, -0.00004757, -0.00001095, -0.00001361, -0.00001804, -0.00000883, -0.00000586, -0.00000975, },
		{ -0.00000291, -0.00001525, -0.00001736, -0.00000730, -0.00000079, -0.00000174, -0.00000323, -0.00000431, -0.00000462, -0.00000287, -0.00000215, -0.00000221, -0.00000106, -0.00000045, 0.00000119, 0.00000158, -0.00000551, -0.00000913, -0.00000141, -0.00000058, -0.00000747, -0.00000166, 0.00000443, -0.00000501, -0.00000755, 0.00000093, -0.00000028, -0.00000065, 0.00000427, -0.00000668, -0.00001239, 0.00000731, 0.00000565, -0.00002135, -0.00000873, 0.00002284, 0.00000498, -0.00001892, 0.00000473, 0.00001885, -0.00000389, -0.00000894, 0.00000244, -0.00000477, -0.00000571, 0.00000999, 0.00000974, -0.00000151, -0.00000760, -0.00001297, -0.00000001, 0.00001972, 0.00000107, -0.00001525, 0.00001136, 0.00000343, -0.00003423, 0.00000408, 0.00004069, -0.00002405, -0.00003768, 0.00004362, 0.00001961, -0.00005819, 0.00000641, 0.00005940, -0.00003209, -0.00004491, 0.00004948, 0.00002131, -0.00005131, 0.00000200, 0.00003525, -0.00002210, -0.00001711, 0.00002537, 0.00000014, -0.00001793, 0.00000348, -0.00000213, -0.00000921, 0.00000170, -0.00000571, 0.00000072, 0.00001949, -0.00000821, -0.00002632, 0.00000895, 0.00000729, -0.00002727, -0.00000668, 0.00002089, 0.00000299, -0.00000591, -0.00000378, -0.00001303, -0.00000640, -0.00000899, -0.00002771, -0.00000205, 0.00001243, -0.00004106, -0.00003942, 0.00000508, -0.00008542, -0.00017429, -0.00001579, 0.00018145, 0.00014050, 0.00003428, 0.00005585, 0.00007938, 0.00003657, 0.00004457, 0.00011251, 0.00012718, 0.00004959, -0.00005497, -0.00009675, -0.00005619, -0.00000982, -0.00001109, -0.00002218, -0.00001187, -0.00000384, -0.00000850, -0.00000721, 0.00000044, },
		{ -0.00001880, -0.00000863, -0.00000005, -0.00000133, -0.00000403, -0.00000440, -0.00000471, -0.00000330, -0.00000167, -0.00000240, -0.00000199, 0.00000008, 0.00000135, 0.00000125, -0.00000393, -0.00000963, -0.00000423, 0.00000158, -0.00000479, -0.00000534, 0.00000385, -0.00000038, -0.00000904, -0.00000231, 0.00000157, -0.00000263, 0.00000254, 0.00000087, -0.00001252, -0.00000362, 0.00001017, -0.00000972, -0.00001881, 0.00001207, 0.00001741, -0.00001263, -0.00000810, 0.00001695, 0.00000573, -0.00000978, 0.00000010, -0.00000105, -0.00000923, 0.00000534, 0.00001428, 0.00000054, -0.00000787, -0.00000903, -0.00000574, 0.00001077, 0.00001054, -0.00000836, 0.00000121, 0.00000718, -0.00002503, -0.00001189, 0.00003656, -0.00000129, -0.00004469, 0.00002207, 0.00003955, -0.00004699, -0.00002326, 0.00006323, -0.00000225, -0.00006296, 0.00003094, 0.00004822, -0.00004975, -0.00002369, 0.00004871, -0.00000857, -0.00003935, 0.00002711, 0.00002075, -0.00003043, -0.00000562, 0.00001435, -0.00001210, -0.00000979, 0.00000258, 0.00000061, 0.00001295, 0.00000173, -0.00002679, -0.00000252, 0.00001652, -0.00002109, -0.00001992, 0.00002115, 0.00000951, -0.00001228, 0.00000069, -0.00000601, -0.00001692, -0.00000544, -0.00001499, -0.00001545, 0.00000965, -0.00002189, -0.00005349, -0.00000164, -0.00003113, -0.00016886, -0.00011465, 0.00012677, 0.00019098, 0.00007008, 0.00004282, 0.00008561, 0.00005427, 0.00004051, 0.00011297, 0.00014424, 0.00006034, -0.00005573, -0.00010341, -0.00006511, -0.00001146, -0.00000741, -0.00002462, -0.00001646, -0.00000253, -0.00000635, -0.00000859, -0.00000070, -0.00000173, -0.00001444, },
		{ 0.00000046, -0.00000036, -0.00000469, -0.00000496, -0.00000454, -0.00000386, -0.00000177, -0.00000189, -0.00000264, -0.00000022, 0.00000189, 0.00000102, -0.00000305, -0.00000880, -0.00000685, 0.00000118, -0.00000101, -0.00000646, 0.00000055, 0.00000304, -0.00000708, -0.00000625, 0.00000160, -0.00000203, -0.00000104, 0.00000491, -0.00000669, -0.00001215, 0.00000561, 0.00000265, -0.00001846, -0.00000305, 0.00002062, 0.00000067, -0.00001502, 0.00000737, 0.00001327, -0.00000528, -0.00000391, 0.00000130, -0.00000858, -0.00000115, 0.00001513, 0.00000543, -0.00000824, -0.00000672, -0.00000617, 0.00000100, 0.00001196, 0.00000239, -0.00000343, 0.00000393, -0.00001440, -0.00001953, 0.00002327, 0.00001628, -0.00003743, -0.00000205, 0.00004541, -0.00002380, -0.00004490, 0.00004864, 0.00002938, -0.00006335, 0.00000064, 0.00006564, -0.00003086, -0.00005169, 0.00004714, 0.00001777, -0.00005572, 0.00001148, 0.00004649, -0.00002944, -0.00002769, 0.00002691, -0.00000092, -0.00002664, 0.00000319, 0.00001018, 0.00000429, 0.00000423, -0.00001805, -0.00001373, 0.00001583, -0.00000795, -0.00002823, 0.00001293, 0.00001875, -0.00001557, -0.00000290, 0.00000624, -0.00002154, -0.00001408, 0.00000059, -0.00001613, -0.00000610, -0.00000610, -0.00004884, -0.00002532, 0.00000188, -0.00012032, -0.00018045, 0.00002606, 0.00020487, 0.00013054, 0.00004456, 0.00007809, 0.00007366, 0.00004474, 0.00010876, 0.00016083, 0.00007609, -0.00005788, -0.00011085, -0.00007233, -0.00001543, -0.00000477, -0.00002456, -0.00002141, -0.00000291, -0.00000345, -0.00000941, -0.00000233, -0.00000061, -0.00001343, -0.00002011, -0.00001021, },
		{ -0.00000487, -0.00000589, -0.00000448, -0.00000415, -0.00000244, -0.00000136, -0.00000252, -0.00000100, 0.00000212, 0.00000129, -0.00000278, -0.00000763, -0.00000814, -0.00000098, 0.00000167, -0.00000470, -0.00000307, 0.00000346, -0.00000305, -0.00000838, -0.00000082, -0.00000037, -0.00000359, 0.00000448, 0.00000091, -0.00001418, -0.00000434, 0.00000897, -0.00000925, -0.00001449, 0.00001330, 0.00001374, -0.00001224, -0.00000512, 0.00001404, 0.00000317, -0.00000598, 0.00000041, -0.00000520, -0.00000594, 0.00001148, 0.00001058, -0.00000638, -0.00000665, -0.00000388, -0.00000539, 0.00000653, 0.00001121, -0.00000074, -0.00000222, -0.00000763, -0.00001850, 0.00000770, 0.00002293, -0.00002067, -0.00001899, 0.00003622, 0.00000121, -0.00005067, 0.00002187, 0.00005053, -0.00004514, -0.00002986, 0.00006526, -0.00000023, -0.00006994, 0.00002744, 0.00004695, -0.00005569, -0.00001851, 0.00006448, -0.00000963, -0.00005379, 0.00002477, 0.00002314, -0.00003842, -0.00000996, 0.00002500, 0.00000252, -0.00000418, -0.00000637, -0.00001604, 0.00000500, 0.00000382, -0.00002569, -0.00000087, 0.00002386, -0.00001148, -0.00001233, 0.00001550, -0.00001486, -0.00002924, 0.00000742, -0.00000161, -0.00002319, -0.00000620, -0.00002843, -0.00004498, 0.00000123, -0.00005678, -0.00018913, -0.00008971, 0.00015921, 0.00019347, 0.00007685, 0.00006382, 0.00008423, 0.00005938, 0.00010471, 0.00017274, 0.00009736, -0.00005856, -0.00012176, -0.00007788, -0.00001930, -0.00000500, -0.00002301, -0.00002485, -0.00000524, -0.00000078, -0.00000906, -0.00000439, 0.00000019, -0.00001201, -0.00002136, -0.00001201, 0.00000069, 0.00000091, },
		{ -0.00000482, -0.00000419, -0.00000325, -0.00000133, -0.00000180, -0.00000157, 0.00000170, 0.00000185, -0.00000256, -0.00000687, -0.00000815, -0.00000334, 0.00000218, -0.00000151, -0.00000483, 0.00000116, 0.00000053, -0.00000748, -0.00000434, 0.00000023, -0.00000386, 0.00000138, 0.00000612, -0.00000977, -0.00001360, 0.00000653, 0.00000265, -0.00001677, -0.00000010, 0.00001956, -0.00000138, -0.00001382, 0.00000694, 0.00001115, -0.00000335, -0.00000279, -0.00000215, -0.00000695, 0.00000519, 0.00001292, -0.00000182, -0.00000725, -0.00000189, -0.00000700, -0.00000164, 0.00001400, 0.00000696, -0.00000610, -0.00000678, -0.00001263, -0.00000353, 0.00001855, -0.00000289, -0.00002314, 0.00001806, 0.00001771, -0.00004043, -0.00000558, 0.00005380, -0.00001599, -0.00004844, 0.00004687, 0.00003039, -0.00006997, -0.00000459, 0.00006607, -0.00003591, -0.00005159, 0.00006356, 0.00002477, -0.00007026, 0.00000320, 0.00005038, -0.00003450, -0.00003537, 0.00003477, 0.00001425, -0.00001921, -0.00000190, -0.00000567, -0.00000862, 0.00000505, -0.00001286, -0.00001185, 0.00001922, -0.00000174, -0.00002030, 0.00001525, 0.00000045, -0.00003998, -0.00000071, 0.00001984, -0.00002771, -0.00002387, -0.00000737, -0.00004398, -0.00002438, -0.00001159, -0.00014254, -0.00017769, 0.00005436, 0.00022361, 0.00014188, 0.00006098, 0.00007866, 0.00007892, 0.00010832, 0.00017856, 0.00012084, -0.00005376, -0.00013678, -0.00008448, -0.00002033, -0.00000777, -0.00002259, -0.00002558, -0.00000831, 0.00000034, -0.00000752, -0.00000626, 0.00000022, -0.00001013, -0.00002232, -0.00001420, 0.00000067, 0.00000227, -0.00000455, -0.00000681, },
		{ -0.00000387, -0.00000180, -0.00000106, -0.00000151, 0.00000089, 0.00000218, -0.00000201, -0.00000658, -0.00000765, -0.00000475, 0.00000096, 0.00000111, -0.00000406, -0.00000196, 0.00000176, -0.00000437, -0.00000678, -0.00000122, -0.00000268, -0.00000166, 0.00000712, -0.00000233, -0.00001742, -0.00000233, 0.00001031, -0.00000983, -0.00001209, 0.00001531, 0.00001144, -0.00001395, -0.00000459, 0.00001376, 0.00000372, -0.00000519, -0.00000177, -0.00000472, -0.00000047, 0.00001096, 0.00000365, -0.00000630, -0.00000163, -0.00000531, -0.00000817, 0.00001025, 0.00001488, -0.00000453, -0.00000994, -0.00000716, -0.00000738, 0.00000797, 0.00000821, -0.00001546, 0.00000050, 0.00002045, -0.00002144, -0.00002289, 0.00004032, 0.00001173, -0.00004865, 0.00001870, 0.00004905, -0.00005168, -0.00003679, 0.00006613, -0.00000266, -0.00007359, 0.00004071, 0.00006100, -0.00006598, -0.00003252, 0.00006678, -0.00001069, -0.00006304, 0.00002888, 0.00003838, -0.00003012, -0.00001224, 0.00001259, -0.00001367, -0.00000781, 0.00000185, -0.00001105, 0.00000552, 0.00000593, -0.00001981, 0.00000604, 0.00001498, -0.00003785, -0.00001964, 0.00003430, -0.00001338, -0.00004789, -0.00000240, -0.00002112, -0.00005152, -0.00000499, -0.00006858, -0.00020379, -0.00008039, 0.00018920, 0.00021975, 0.00009041, 0.00006109, 0.00009133, 0.00012428, 0.00018262, 0.00014103, -0.00004164, -0.00015297, -0.00009603, -0.00001799, -0.00001017, -0.00002580, -0.00002453, -0.00000982, -0.00000071, -0.00000579, -0.00000702, -0.00000061, -0.00000825, -0.00002254, -0.00001676, 0.00000021, 0.00000372, -0.00000386, -0.00000743, -0.00000547, -0.00000431, },
		{ -0.00000073, -0.00000098, 0.00000024, 0.00000201, -0.00000126, -0.00000639, -0.00000734, -0.00000506, -0.00000070, 0.00000203, -0.00000179, -0.00000387, 0.00000048, -0.00000112, -0.00000674, -0.00000395, -0.00000187, -0.00000296, 0.00000488, 0.00000398, -0.00001470, -0.00001215, 0.00000992, 0.00000140, -0.00001651, 0.00000381, 0.00001915, -0.00000552, -0.00001438, 0.00000878, 0.00001165, -0.00000371, -0.00000400, -0.00000183, -0.00000292, 0.00000573, 0.00000714, -0.00000279, -0.00000232, -0.00000301, -0.00001071, 0.00000276, 0.00001856, 0.00000199, -0.00001288, -0.00000555, -0.00000525, -0.00000222, 0.00000943, -0.00000232, -0.00000852, 0.00001150, -0.00000409, -0.00002526, 0.00001870, 0.00002730, -0.00003289, -0.00000683, 0.00004976, -0.00002342, -0.00005684, 0.00004690, 0.00003104, -0.00007483, 0.00000334, 0.00008388, -0.00003880, -0.00006868, 0.00006140, 0.00002704, -0.00007883, 0.00000326, 0.00006500, -0.00002558, -0.00003670, 0.00002680, -0.00000180, -0.00002900, 0.00000625, 0.00000450, -0.00000865, 0.00000274, -0.00001024, -0.00000407, 0.00001921, -0.00002334, -0.00003706, 0.00003173, 0.00001327, -0.00006123, -0.00001988, 0.00000861, -0.00005743, -0.00003317, -0.00000692, -0.00016106, -0.00019655, 0.00008132, 0.00027170, 0.00016102, 0.00005001, 0.00008545, 0.00014886, 0.00019388, 0.00015463, -0.00002505, -0.00016520, -0.00011437, -0.00001523, -0.00000842, -0.00003274, -0.00002477, -0.00000799, -0.00000277, -0.00000559, -0.00000629, -0.00000161, -0.00000712, -0.00002186, -0.00001927, -0.00000100, 0.00000517, -0.00000286, -0.00000771, -0.00000613, -0.00000475, -0.00000433, -0.00000239, },
		{ 0.00000008, 0.00000157, -0.00000069, -0.00000601, -0.00000737, -0.00000481, -0.00000173, 0.00000139, 0.00000031, -0.00000373, -0.00000187, 0.00000043, -0.00000456, -0.00000615, -0.00000257, -0.00000271, 0.00000176, 0.00000666, -0.00000791, -0.00001780, 0.00000249, 0.00001023, -0.00001186, -0.00000833, 0.00001773, 0.00000667, -0.00001703, -0.00000171, 0.00001565, 0.00000229, -0.00000636, -0.00000086, -0.00000193, 0.00000011, 0.00000670, 0.00000214, -0.00000201, -0.00000194, -0.00000974, -0.00000440, 0.00001633, 0.00000991, -0.00001203, -0.00000773, -0.00000131, -0.00000740, 0.00000294, 0.00000842, -0.00000655, -0.00000162, 0.00000376, -0.00001582, 0.00000008, 0.00002690, -0.00001071, -0.00001964, 0.00003541, 0.00000244, -0.00005860, 0.00001703, 0.00005204, -0.00005518, -0.00003415, 0.00008355, 0.00000249, -0.00008972, 0.00003288, 0.00006441, -0.00007203, -0.00003591, 0.00007984, -0.00000083, -0.00006505, 0.00002428, 0.00002611, -0.00004512, -0.00000825, 0.00002798, -0.00001098, -0.00001431, 0.00000057, -0.00000480, 0.00001087, -0.00000654, -0.00004081, 0.00001363, 0.00003642, -0.00005280, -0.00005039, 0.00002412, -0.00003487, -0.00007158, 0.00001259, -0.00007478, -0.00024871, -0.00007514, 0.00025650, 0.00025739, 0.00007111, 0.00006132, 0.00016917, 0.00022003, 0.00016491, -0.00001061, -0.00017006, -0.00013701, -0.00001703, -0.00000094, -0.00004023, -0.00002972, -0.00000344, -0.00000314, -0.00000803, -0.00000505, -0.00000149, -0.00000715, -0.00002083, -0.00002098, -0.00000304, 0.00000625, -0.00000148, -0.00000773, -0.00000657, -0.00000541, -0.00000496, -0.00000282, -0.00000073, -0.00000049, },
		{ -0.00000045, -0.00000553, -0.00000752, -0.00000460, -0.00000185, 0.00000023, 0.00000112, -0.00000218, -0.00000353, -0.00000016, -0.00000189, -0.00000645, -0.00000445, -0.00000226, -0.00000035, 0.00000582, -0.00000097, -0.00001720, -0.00000737, 0.00001205, -0.00000169, -0.00001468, 0.00000841, 0.00001586, -0.00001114, -0.00001229, 0.00001266, 0.00001020, -0.00000574, -0.00000252, 0.00000054, -0.00000307, 0.00000261, 0.00000591, 0.00000060, -0.00000196, -0.00000746, -0.00000839, 0.00000983, 0.00001500, -0.00000667, -0.00001079, 0.00000060, -0.00000688, -0.00000582, 0.00001163, 0.00000281, -0.00001054, 0.00000053, -0.00000295, -0.00000785, 0.00001490, 0.00000658, -0.00001701, 0.00001567, 0.00001613, -0.00004491, -0.00001070, 0.00005378, -0.00002401, -0.00005760, 0.00006067, 0.00004227, -0.00008621, -0.00000911, 0.00008570, -0.00004169, -0.00007412, 0.00007198, 0.00003823, -0.00008210, -0.00000058, 0.00005871, -0.00004229, -0.00004009, 0.00004483, 0.00000689, -0.00003817, 0.00000027, 0.00000845, -0.00000205, -0.00000065, -0.00002799, -0.00000689, 0.00004145, -0.00002602, -0.00007421, 0.00001309, 0.00000294, -0.00009086, -0.00001377, 0.00000992, -0.00021811, -0.00022768, 0.00015352, 0.00033892, 0.00014130, 0.00003625, 0.00017008, 0.00026016, 0.00018188, -0.00000382, -0.00016955, -0.00015801, -0.00002680, 0.00000963, -0.00004340, -0.00004033, 0.00000046, 0.00000073, -0.00001223, -0.00000537, 0.00000057, -0.00000763, -0.00002055, -0.00002150, -0.00000534, 0.00000641, 0.00000022, -0.00000741, -0.00000680, -0.00000599, -0.00000595, -0.00000320, -0.00000069, -0.00000034, 0.00000018, 0.00000133, },
		{ -0.00000762, -0.00000458, -0.00000142, -0.00000052, 0.00000063, -0.00000059, -0.00000361, -0.00000187, -0.00000038, -0.00000488, -0.00000614, -0.00000273, -0.00000110, 0.00000330, 0.00000320, -0.00001180, -0.00001449, 0.00000662, 0.00000783, -0.00001253, -0.00000332, 0.00001722, -0.00000010, -0.00001734, 0.00000350, 0.00001554, -0.00000076, -0.00000501, 0.00000184, -0.00000284, -0.00000272, 0.00000623, 0.00000494, -0.00000147, -0.00000585, -0.00000888, 0.00000257, 0.00001492, 0.00000083, -0.00001134, -0.00000070, -0.00000337, -0.00001175, 0.00000714, 0.00001276, -0.00001063, -0.00000907, 0.00000480, -0.00000436, 0.00000057, 0.00001201, -0.00000426, 0.00000133, 0.00001530, -0.00002532, -0.00002619, 0.00003911, 0.00000502, -0.00006014, 0.00002583, 0.00006582, -0.00005983, -0.00004833, 0.00008184, 0.00000199, -0.00009553, 0.00004041, 0.00007692, -0.00007588, -0.00004189, 0.00007920, -0.00001448, -0.00007645, 0.00004035, 0.00004259, -0.00005389, -0.00001981, 0.00002956, -0.00000555, -0.00001249, -0.00000836, -0.00001330, 0.00002567, 0.00000211, -0.00007446, -0.00001863, 0.00003104, -0.00007587, -0.00006239, 0.00005213, -0.00012408, -0.00031882, -0.00002024, 0.00035494, 0.00025203, 0.00004002, 0.00014635, 0.00030164, 0.00021658, -0.00000334, -0.00017117, -0.00017219, -0.00004292, 0.00001743, -0.00003926, -0.00005357, -0.00000071, 0.00000893, -0.00001519, -0.00000905, 0.00000383, -0.00000697, -0.00002182, -0.00002137, -0.00000684, 0.00000539, 0.00000172, -0.00000652, -0.00000688, -0.00000631, -0.00000716, -0.00000391, -0.00000038, -0.00000037, 0.00000005, 0.00000147, -0.00000027, -0.00000526, },
		{ -0.00000088, -0.00000055, -0.00000037, 0.00000004, -0.00000252, -0.00000324, -0.00000058, -0.00000266, -0.00000643, -0.00000411, -0.00000132, 0.00000096, 0.00000392, -0.00000503, -0.00001600, -0.00000244, 0.00001165, -0.00000410, -0.00001112, 0.00001054, 0.00000995, -0.00001429, -0.00000738, 0.00001471, 0.00000691, -0.00000537, 0.00000069, -0.00000056, -0.00000638, 0.00000278, 0.00000876, 0.00000101, -0.00000528, -0.00000754, -0.00000244, 0.00001035, 0.00000674, -0.00000794, -0.00000352, -0.00000045, -0.00001270, -0.00000122, 0.00001743, -0.00000296, -0.00001755, 0.00000348, 0.00000494, -0.00000751, 0.00000599, 0.00000903, -0.00000157, 0.00000532, -0.00001037, -0.00002683, 0.00001789, 0.00002117, -0.00004502, -0.00000619, 0.00006658, -0.00002235, -0.00007008, 0.00005516, 0.00004244, -0.00009116, -0.00000457, 0.00009899, -0.00004469, -0.00008409, 0.00007421, 0.00003163, -0.00009888, 0.00000865, 0.00008199, -0.00004648, -0.00005736, 0.00004338, 0.00001114, -0.00003620, 0.00000074, 0.00000276, 0.00000184, 0.00001152, -0.00004957, -0.00004941, 0.00002968, -0.00003576, -0.00009745, 0.00003754, -0.00001547, -0.00031488, -0.00021085, 0.00027075, 0.00036330, 0.00009836, 0.00011231, 0.00032583, 0.00027208, 0.00000114, -0.00018311, -0.00018017, -0.00005885, 0.00001742, -0.00002967, -0.00006365, -0.00000915, 0.00001809, -0.00001348, -0.00001577, 0.00000594, -0.00000373, -0.00002416, -0.00002201, -0.00000677, 0.00000380, 0.00000220, -0.00000502, -0.00000665, -0.00000647, -0.00000825, -0.00000513, -0.00000000, -0.00000013, -0.00000055, 0.00000166, 0.00000028, -0.00000521, -0.00000783, -0.00000465, },
		{ -0.00000112, -0.00000035, -0.00000133, -0.00000347, -0.00000179, -0.00000119, -0.00000518, -0.00000543, -0.00000195, -0.00000043, 0.00000240, -0.00000002, -0.00001236, -0.00001021, 0.00000844, 0.00000526, -0.00001140, 0.00000003, 0.00001373, -0.00000514, -0.00001440, 0.00000741, 0.00001328, -0.00000170, -0.00000181, 0.00000126, -0.00000685, -0.00000258, 0.00000966, 0.00000536, -0.00000447, -0.00000633, -0.00000427, 0.00000412, 0.00000840, -0.00000197, -0.00000495, -0.00000003, -0.00000998, -0.00000855, 0.00001490, 0.00000724, -0.00001928, -0.00000481, 0.00001227, -0.00000604, -0.00000498, 0.00001481, 0.00000563, -0.00000465, -0.00000583, -0.00001779, 0.00000101, 0.00002156, -0.00002282, -0.00002398, 0.00004871, 0.00001067, -0.00006844, 0.00001800, 0.00006500, -0.00006359, -0.00004620, 0.00009499, 0.00000131, -0.00010960, 0.00004143, 0.00007890, -0.00009291, -0.00004248, 0.00010534, -0.00001030, -0.00009774, 0.00003411, 0.00004702, -0.00005487, -0.00001448, 0.00003411, -0.00000862, -0.00000702, -0.00001637, -0.00005549, -0.00000115, 0.00000145, -0.00009375, -0.00001402, 0.00005473, -0.00022508, -0.00034877, 0.00009145, 0.00041568, 0.00021215, 0.00010053, 0.00032139, 0.00033761, 0.00002385, -0.00020700, -0.00018991, -0.00006738, 0.00000988, -0.00002105, -0.00006601, -0.00002239, 0.00002289, -0.00000597, -0.00002249, 0.00000436, 0.00000186, -0.00002570, -0.00002477, -0.00000552, 0.00000300, 0.00000110, -0.00000350, -0.00000569, -0.00000660, -0.00000905, -0.00000667, -0.00000002, 0.00000062, -0.00000132, 0.00000134, 0.00000132, -0.00000498, -0.00000840, -0.00000478, -0.00000034, -0.00000010, },
		{ -0.00000083, -0.00000285, -0.00000287, -0.00000100, -0.00000330, -0.00000572, -0.00000311, -0.00000120, 0.00000028, 0.00000193, -0.00000636, -0.00001327, 0.00000076, 0.00001036, -0.00000508, -0.00000834, 0.00000979, 0.00000483, -0.00001414, -0.00000294, 0.00001449, 0.00000520, -0.00000299, 0.00000094, -0.00000483, -0.00000700, 0.00000676, 0.00000969, -0.00000192, -0.00000592, -0.00000404, -0.00000078, 0.00000572, 0.00000346, -0.00000321, -0.00000145, -0.00000655, -0.00001180, 0.00000742, 0.00001408, -0.00001339, -0.00001402, 0.00001245, 0.00000202, -0.00001287, 0.00001078, 0.00001604, -0.00000742, -0.00001031, -0.00000790, -0.00000540, 0.00001143, -0.00000520, -0.00002473, 0.00002388, 0.00002778, -0.00004850, -0.00001390, 0.00006405, -0.00002539, -0.00006965, 0.00006723, 0.00004491, -0.00010776, -0.00000849, 0.00010787, -0.00005628, -0.00009410, 0.00009749, 0.00004582, -0.00012040, -0.00000449, 0.00008740, -0.00004999, -0.00005408, 0.00006022, 0.00001083, -0.00004375, -0.00000098, -0.00002741, -0.00003955, 0.00000706, -0.00005331, -0.00006097, 0.00005804, -0.00009916, -0.00038211, -0.00013122, 0.00035821, 0.00034493, 0.00014604, 0.00029788, 0.00039202, 0.00007274, -0.00023366, -0.00021247, -0.00006689, 0.00000168, -0.00002026, -0.00006126, -0.00003403, 0.00001968, 0.00000437, -0.00002494, -0.00000147, 0.00000740, -0.00002432, -0.00002960, -0.00000476, 0.00000426, -0.00000110, -0.00000307, -0.00000375, -0.00000649, -0.00000976, -0.00000809, -0.00000069, 0.00000160, -0.00000166, 0.00000027, 0.00000233, -0.00000408, -0.00000927, -0.00000525, 0.00000029, 0.00000055, -0.00000139, -0.00000115, },
		{ -0.00000320, -0.00000160, -0.00000190, -0.00000483, -0.00000417, -0.00000198, -0.00000143, 0.00000133, -0.00000106, -0.00001120, -0.00000693, 0.00000900, 0.00000343, -0.00001045, 0.00000108, 0.00001023, -0.00000729, -0.00001111, 0.00000927, 0.00001196, -0.00000063, -0.00000109, -0.00000259, -0.00000853, 0.00000139, 0.00001156, 0.00000262, -0.00000537, -0.00000356, -0.00000304, 0.00000087, 0.00000565, 0.00000112, -0.00000244, -0.00000470, -0.00001106, -0.00000069, 0.00001447, -0.00000354, -0.00001828, 0.00000558, 0.00001019, -0.00001300, 0.00000084, 0.00002198, -0.00000159, -0.00001730, -0.00000408, -0.00000239, -0.00000010, 0.00000130, -0.00001420, 0.00000448, 0.00002675, -0.00002270, -0.00002943, 0.00004504, 0.00000761, -0.00006919, 0.00002826, 0.00007055, -0.00008021, -0.00005702, 0.00010657, -0.00000087, -0.00012503, 0.00005662, 0.00010141, -0.00010970, -0.00006346, 0.00011047, -0.00001311, -0.00010181, 0.00005852, 0.00005991, -0.00007345, -0.00002240, 0.00002131, -0.00005399, -0.00002848, -0.00000533, -0.00006555, 0.00000632, -0.00000281, -0.00030466, -0.00031650, 0.00017934, 0.00042984, 0.00026176, 0.00028916, 0.00041751, 0.00014042, -0.00024649, -0.00025355, -0.00006485, 0.00000231, -0.00002895, -0.00005608, -0.00003773, 0.00000988, 0.00001149, -0.00002076, -0.00000880, 0.00000960, -0.00001950, -0.00003459, -0.00000644, 0.00000759, -0.00000277, -0.00000471, -0.00000135, -0.00000550, -0.00001068, -0.00000922, -0.00000175, 0.00000221, -0.00000126, -0.00000113, 0.00000261, -0.00000244, -0.00000986, -0.00000639, 0.00000089, 0.00000144, -0.00000135, -0.00000177, -0.00000104, -0.00000220, },
		{ -0.00000138, -0.00000347, -0.00000443, -0.00000289, -0.00000265, -0.00000052, 0.00000177, -0.00000603, -0.00001096, 0.00000272, 0.00000908, -0.00000607, -0.00000711, 0.00000864, 0.00000187, -0.00001310, -0.00000001, 0.00001464, 0.00000524, -0.00000257, -0.00000207, -0.00000727, -0.00000377, 0.00000965, 0.00000758, -0.00000320, -0.00000366, -0.00000332, -0.00000342, 0.00000397, 0.00000560, -0.00000117, -0.00000457, -0.00000852, -0.00000592, 0.00000924, 0.00000483, -0.00001534, -0.00000379, 0.00001289, -0.00000629, -0.00000814, 0.00001968, 0.00000843, -0.00001994, -0.00000764, 0.00000391, -0.00000576, -0.00000240, -0.00000216, -0.00000271, 0.00001422, -0.00000365, -0.00002701, 0.00002057, 0.00002439, -0.00005024, -0.00000587, 0.00007183, -0.00003962, -0.00008702, 0.00007645, 0.00005283, -0.00012252, -0.00000394, 0.00013395, -0.00006412, -0.00012135, 0.00009913, 0.00004730, -0.00013259, 0.00001825, 0.00011889, -0.00006926, -0.00007891, 0.00005964, -0.00002318, -0.00008670, 0.00001119, -0.00001783, -0.00005499, 0.00001680, -0.00016245, -0.00039072, -0.00007674, 0.00039743, 0.00041864, 0.00034003, 0.00041721, 0.00020434, -0.00023245, -0.00030665, -0.00007472, 0.00001661, -0.00004065, -0.00005934, -0.00003235, 0.00000027, 0.00001014, -0.00001212, -0.00001268, 0.00000686, -0.00001339, -0.00003696, -0.00001123, 0.00001125, -0.00000206, -0.00000825, 0.00000014, -0.00000307, -0.00001161, -0.00001036, -0.00000263, 0.00000210, -0.00000044, -0.00000206, 0.00000184, -0.00000070, -0.00000952, -0.00000806, 0.00000102, 0.00000268, -0.00000114, -0.00000212, -0.00000142, -0.00000213, -0.00000304, -0.00000208, },
		{ -0.00000388, -0.00000348, -0.00000362, -0.00000259, 0.00000201, -0.00000057, -0.00001016, -0.00000465, 0.00000909, 0.00000136, -0.00001039, 0.00000178, 0.00000813, -0.00000844, -0.00000863, 0.00001124, 0.00001192, -0.00000109, -0.00000344, -0.00000514, -0.00000644, 0.00000478, 0.00001055, 0.00000111, -0.00000366, -0.00000319, -0.00000562, -0.00000003, 0.00000770, 0.00000236, -0.00000460, -0.00000646, -0.00000734, 0.00000187, 0.00000787, -0.00000739, -0.00001002, 0.00000873, 0.00000220, -0.00001094, 0.00001102, 0.00001601, -0.00001539, -0.00001454, 0.00000698, -0.00000384, -0.00001003, 0.00000360, 0.00000108, 0.00000069, 0.00000220, -0.00001389, 0.00000302, 0.00002349, -0.00002533, -0.00002385, 0.00005374, -0.00000272, -0.00009083, 0.00003108, 0.00008585, -0.00008847, -0.00006180, 0.00013019, 0.00000180, -0.00015447, 0.00005118, 0.00010853, -0.00012599, -0.00005144, 0.00015767, -0.00001908, -0.00014590, 0.00005620, 0.00004856, -0.00012947, -0.00002787, 0.00005632, -0.00007315, -0.00004308, -0.00003322, -0.00032485, -0.00032050, 0.00021313, 0.00054491, 0.00047900, 0.00042456, 0.00023986, -0.00019449, -0.00035406, -0.00010673, 0.00003899, -0.00004373, -0.00007542, -0.00002426, -0.00000093, 0.00000019, -0.00000552, -0.00000933, 0.00000129, -0.00000987, -0.00003524, -0.00001739, 0.00001265, 0.00000162, -0.00001197, -0.00000076, 0.00000052, -0.00001175, -0.00001202, -0.00000305, 0.00000155, 0.00000001, -0.00000206, 0.00000054, 0.00000019, -0.00000813, -0.00000954, 0.00000022, 0.00000413, -0.00000058, -0.00000243, -0.00000154, -0.00000247, -0.00000314, -0.00000205, -0.00000127, -0.00000249, },
		{ -0.00000429, -0.00000450, 0.00000048, 0.00000317, -0.00000579, -0.00000944, 0.00000405, 0.00000722, -0.00000766, -0.00000589, 0.00000841, -0.00000035, -0.00001231, 0.00000313, 0.00001562, 0.00000394, -0.00000488, -0.00000423, -0.00000619, -0.00000059, 0.00000985, 0.00000630, -0.00000211, -0.00000355, -0.00000605, -0.00000393, 0.00000650, 0.00000632, -0.00000316, -0.00000561, -0.00000625, -0.00000403, 0.00000516, 0.00000086, -0.00000991, 0.00000082, 0.00000693, -0.00000690, 0.00000156, 0.00001671, -0.00000607, -0.00001888, 0.00000411, 0.00000204, -0.00001478, 0.00000115, 0.00000905, -0.00000572, -0.00000324, -0.00000084, -0.00000162, 0.00001215, -0.00000683, -0.00002389, 0.00002878, 0.00001825, -0.00007300, -0.00001122, 0.00008956, -0.00003802, -0.00009647, 0.00009270, 0.00006406, -0.00014890, -0.00001832, 0.00014469, -0.00007810, -0.00012452, 0.00015189, 0.00006540, -0.00018758, -0.00000169, 0.00013041, -0.00012035, -0.00011462, 0.00010859, -0.00001943, -0.00013843, 0.00001212, -0.00015301, -0.00045335, -0.00009486, 0.00055081, 0.00069034, 0.00049456, 0.00024123, -0.00015572, -0.00037778, -0.00015827, 0.00005467, -0.00002951, -0.00009933, -0.00002422, 0.00000987, -0.00001173, -0.00000801, 0.00000035, -0.00000170, -0.00001177, -0.00003102, -0.00002163, 0.00001030, 0.00000672, -0.00001344, -0.00000439, 0.00000376, -0.00001013, -0.00001421, -0.00000341, 0.00000133, -0.00000044, -0.00000147, -0.00000027, -0.00000023, -0.00000645, -0.00000993, -0.00000144, 0.00000514, 0.00000063, -0.00000278, -0.00000149, -0.00000264, -0.00000387, -0.00000194, -0.00000086, -0.00000209, -0.00000318, -0.00000335, },
		{ -0.00000196, 0.00000452, -0.00000022, -0.00000992, -0.00000292, 0.00000837, -0.00000133, -0.00000999, 0.00000331, 0.00000640, -0.00000963, -0.00000566, 0.00001378, 0.00001054, -0.00000398, -0.00000521, -0.00000451, -0.00000399, 0.00000581, 0.00001000, 0.00000173, -0.00000387, -0.00000608, -0.00000622, 0.00000309, 0.00000853, 0.00000005, -0.00000512, -0.00000465, -0.00000675, -0.00000080, 0.00000521, -0.00000428, -0.00000574, 0.00000546, 0.00000027, -0.00000337, 0.00001082, 0.00000251, -0.00001714, -0.00000253, 0.00000638, -0.00001357, -0.00000588, 0.00001393, -0.00000321, -0.00001252, 0.00000425, 0.00000426, 0.00000091, -0.00000069, -0.00001296, 0.00000957, 0.00002068, -0.00004635, -0.00003658, 0.00006869, 0.00000854, -0.00009891, 0.00003822, 0.00010092, -0.00010761, -0.00008399, 0.00014007, -0.00000458, -0.00017078, 0.00009663, 0.00015284, -0.00017598, -0.00009844, 0.00017992, -0.00004671, -0.00021015, 0.00009570, 0.00009246, -0.00020173, -0.00005375, 0.00003966, -0.00041384, -0.00043502, 0.00037091, 0.00090313, 0.00067597, 0.00023823, -0.00014880, -0.00037511, -0.00021154, 0.00004896, -0.00000041, -0.00011811, -0.00003996, 0.00002756, -0.00001558, -0.00002186, 0.00000983, 0.00000328, -0.00001803, -0.00002861, -0.00002137, 0.00000541, 0.00001015, -0.00001128, -0.00000913, 0.00000466, -0.00000665, -0.00001602, -0.00000449, 0.00000192, -0.00000150, -0.00000130, 0.00000011, -0.00000141, -0.00000571, -0.00000899, -0.00000307, 0.00000503, 0.00000233, -0.00000280, -0.00000167, -0.00000232, -0.00000487, -0.00000248, 0.00000002, -0.00000170, -0.00000298, -0.00000270, -0.00000428, -0.00000609, },
		{ 0.00000448, -0.00000662, -0.00000835, 0.00000464, 0.00000450, -0.00000879, -0.00000360, 0.00000830, -0.00000274, -0.00001079, 0.00000685, 0.00001516, 0.00000037, -0.00000675, -0.00000355, -0.00000448, 0.00000059, 0.00001023, 0.00000690, -0.00000270, -0.00000662, -0.00000699, -0.00000050, 0.00000796, 0.00000373, -0.00000369, -0.00000360, -0.00000672, -0.00000664, 0.00000421, 0.00000306, -0.00000717, -0.00000048, 0.00000544, -0.00000195, 0.00000264, 0.00000568, -0.00001027, -0.00000812, 0.00000587, -0.00000811, -0.00001179, 0.00001230, 0.00000419, -0.00001801, 0.00000011, 0.00001301, -0.00000277, -0.00000469, -0.00000140, 0.00000252, 0.00001112, -0.00002440, -0.00004136, 0.00003730, 0.00003579, -0.00007467, -0.00001124, 0.00010307, -0.00004857, -0.00012264, 0.00009711, 0.00006687, -0.00017127, 0.00001002, 0.00020756, -0.00010675, -0.00019627, 0.00016570, 0.00007054, -0.00026392, 0.00000542, 0.00021152, -0.00017866, -0.00019935, 0.00015581, -0.00021059, -0.00068416, 0.00000430, 0.00100867, 0.00097579, 0.00029474, -0.00019258, -0.00036849, -0.00024305, 0.00001967, 0.00002857, -0.00011910, -0.00006914, 0.00004006, -0.00000472, -0.00004141, 0.00001029, 0.00001713, -0.00002310, -0.00003221, -0.00001730, 0.00000166, 0.00000940, -0.00000661, -0.00001207, 0.00000223, -0.00000275, -0.00001611, -0.00000662, 0.00000294, -0.00000221, -0.00000233, 0.00000139, -0.00000207, -0.00000660, -0.00000754, -0.00000353, 0.00000375, 0.00000375, -0.00000197, -0.00000215, -0.00000177, -0.00000540, -0.00000392, 0.00000074, -0.00000063, -0.00000325, -0.00000227, -0.00000339, -0.00000702, -0.00000475, 0.00000378, },
		{ -0.00001019, -0.00000161, 0.00000660, -0.00000379, -0.00000822, 0.00000507, 0.00000416, -0.00001011, -0.00000188, 0.00001491, 0.00000685, -0.00000654, -0.00000436, -0.00000311, -0.00000319, 0.00000685, 0.00001108, 0.00000090, -0.00000720, -0.00000747, -0.00000281, 0.00000539, 0.00000605, -0.00000096, -0.00000280, -0.00000545, -0.00001014, -0.00000067, 0.00000795, -0.00000313, -0.00000655, 0.00000544, 0.00000345, -0.00000273, 0.00000253, -0.00000264, -0.00000887, 0.00000128, -0.00000280, -0.00001293, 0.00000586, 0.00001034, -0.00001634, -0.00000865, 0.00001716, 0.00000197, -0.00001148, 0.00000336, 0.00000588, -0.00000006, -0.00001455, -0.00003192, 0.00001053, 0.00003957, -0.00003948, -0.00003915, 0.00007693, 0.00000468, -0.00012433, 0.00003469, 0.00011037, -0.00012759, -0.00007566, 0.00020751, -0.00000243, -0.00025376, 0.00008494, 0.00018575, -0.00024084, -0.00013411, 0.00027583, -0.00005830, -0.00034946, 0.00013239, 0.00007631, -0.00073944, -0.00046968, 0.00090398, 0.00133559, 0.00048372, -0.00026785, -0.00039958, -0.00024188, -0.00001612, 0.00003816, -0.00010057, -0.00009843, 0.00003607, 0.00001802, -0.00005527, -0.00000327, 0.00003398, -0.00002002, -0.00004226, -0.00001382, 0.00000282, 0.00000484, -0.00000289, -0.00001091, -0.00000228, -0.00000089, -0.00001382, -0.00000896, 0.00000333, -0.00000177, -0.00000439, 0.00000236, -0.00000116, -0.00000865, -0.00000709, -0.00000229, 0.00000227, 0.00000390, -0.00000032, -0.00000242, -0.00000162, -0.00000510, -0.00000568, 0.00000039, 0.00000111, -0.00000323, -0.00000267, -0.00000206, -0.00000681, -0.00000743, 0.00000149, 0.00000727, -0.00000141, },
		{ 0.00000424, 0.00000153, -0.00000840, -0.00000064, 0.00000746, -0.00000477, -0.00000821, 0.00000945, 0.00001230, -0.00000308, -0.00000613, -0.00000196, -0.00000416, 0.00000168, 0.00001204, 0.00000626, -0.00000635, -0.00000857, -0.00000389, 0.00000257, 0.00000605, 0.00000210, -0.00000139, -0.00000404, -0.00001100, -0.00000655, 0.00000819, 0.00000359, -0.00000879, 0.00000081, 0.00000833, -0.00000265, -0.00000390, 0.00000117, -0.00000458, -0.00000360, -0.00000118, -0.00000965, -0.00000088, 0.00001127, -0.00000951, -0.00001555, 0.00001383, 0.00001035, -0.00001405, 0.00000029, 0.00001268, -0.00000508, -0.00001581, -0.00001919, -0.00000279, 0.00002687, -0.00001020, -0.00004141, 0.00003966, 0.00003466, -0.00009589, -0.00002202, 0.00011419, -0.00006024, -0.00012999, 0.00015491, 0.00009852, -0.00024670, -0.00003409, 0.00025114, -0.00014064, -0.00026712, 0.00024547, 0.00011963, -0.00042291, -0.00002855, 0.00032827, -0.00056633, -0.00091246, 0.00054315, 0.00163690, 0.00084623, -0.00030841, -0.00050588, -0.00022595, -0.00002840, 0.00001972, -0.00007685, -0.00011111, 0.00001430, 0.00004008, -0.00005347, -0.00002707, 0.00004363, -0.00000555, -0.00005373, -0.00001651, 0.00000957, 0.00000024, -0.00000373, -0.00000599, -0.00000560, -0.00000286, -0.00001033, -0.00000979, 0.00000234, -0.00000041, -0.00000633, 0.00000188, 0.00000119, -0.00001030, -0.00000873, -0.00000016, 0.00000206, 0.00000252, 0.00000126, -0.00000170, -0.00000208, -0.00000441, -0.00000675, -0.00000126, 0.00000262, -0.00000210, -0.00000368, -0.00000128, -0.00000533, -0.00000929, -0.00000183, 0.00000792, 0.00000384, -0.00000835, -0.00000726, },
		{ -0.00000492, -0.00000517, 0.00000614, 0.00000173, -0.00000949, 0.00000128, 0.00001365, 0.00000292, -0.00000671, -0.00000243, -0.00000293, -0.00000259, 0.00000916, 0.00001114, -0.00000288, -0.00000985, -0.00000491, 0.00000082, 0.00000438, 0.00000400, 0.00000088, -0.00000259, -0.00001012, -0.00001103, 0.00000428, 0.00000912, -0.00000595, -0.00000502, 0.00000919, 0.00000170, -0.00000901, -0.00000061, 0.00000151, -0.00000502, -0.00000352, -0.00000529, -0.00000411, 0.00000723, -0.00000244, -0.00001660, 0.00000553, 0.00001604, -0.00000994, -0.00000619, 0.00001608, -0.00000235, -0.00002152, -0.00001194, -0.00000298, 0.00000998, 0.00000334, -0.00002606, 0.00000947, 0.00003727, -0.00005570, -0.00005412, 0.00008503, 0.00000268, -0.00013872, 0.00007403, 0.00016073, -0.00017962, -0.00014606, 0.00024066, 0.00000118, -0.00033804, 0.00012389, 0.00028436, -0.00037321, -0.00026097, 0.00044064, -0.00021831, -0.00117829, -0.00002892, 0.00173995, 0.00135435, -0.00021880, -0.00068959, -0.00024364, 0.00000643, -0.00001267, -0.00007237, -0.00009995, -0.00001241, 0.00004645, -0.00003566, -0.00004906, 0.00003805, 0.00001596, -0.00005832, -0.00002782, 0.00001793, 0.00000058, -0.00001000, -0.00000087, -0.00000435, -0.00000808, -0.00000846, -0.00000781, 0.00000057, 0.00000058, -0.00000698, -0.00000014, 0.00000362, -0.00001010, -0.00001215, 0.00000105, 0.00000400, 0.00000056, 0.00000149, 0.00000023, -0.00000253, -0.00000423, -0.00000662, -0.00000341, 0.00000274, 0.00000008, -0.00000425, -0.00000173, -0.00000328, -0.00000958, -0.00000548, 0.00000655, 0.00000787, -0.00000412, -0.00001013, -0.00000090, 0.00000409, },
		{ 0.00000202, 0.00000583, -0.00000599, -0.00000571, 0.00000988, 0.00000877, -0.00000435, -0.00000415, -0.00000141, -0.00000421, 0.00000392, 0.00001307, 0.00000282, -0.00000976, -0.00000679, 0.00000004, 0.00000259, 0.00000396, 0.00000320, -0.00000066, -0.00000843, -0.00001320, -0.00000152, 0.00001085, 0.00000018, -0.00000820, 0.00000560, 0.00000664, -0.00000970, -0.00000613, 0.00000518, -0.00000195, -0.00000706, -0.00000331, -0.00000314, 0.00000166, 0.00000075, -0.00001229, -0.00000246, 0.00001549, -0.00000184, -0.00001026, 0.00001352, 0.00000424, -0.00002466, -0.00001269, 0.00000312, -0.00000085, 0.00000148, -0.00000692, -0.00000329, 0.00002145, -0.00002314, -0.00005704, 0.00004241, 0.00003960, -0.00010872, -0.00000206, 0.00016830, -0.00008229, -0.00021089, 0.00016220, 0.00013115, -0.00031901, -0.00004365, 0.00036942, -0.00021090, -0.00046698, 0.00036892, 0.00018298, -0.00117469, -0.00068028, 0.00154200, 0.00189990, 0.00008381, -0.00089639, -0.00035401, 0.00008280, -0.00002316, -0.00010543, -0.00007680, -0.00002291, 0.00003102, -0.00001404, -0.00005582, 0.00001824, 0.00003349, -0.00005012, -0.00004391, 0.00002116, 0.00000832, -0.00001834, -0.00000029, 0.00000222, -0.00001310, -0.00001096, -0.00000369, 0.00000002, -0.00000008, -0.00000628, -0.00000243, 0.00000456, -0.00000784, -0.00001562, -0.00000021, 0.00000744, -0.00000019, -0.00000028, 0.00000247, -0.00000189, -0.00000504, -0.00000584, -0.00000472, 0.00000109, 0.00000212, -0.00000331, -0.00000311, -0.00000192, -0.00000811, -0.00000843, 0.00000348, 0.00001003, 0.00000095, -0.00000970, -0.00000618, 0.00000278, -0.00000057, -0.00000622, },
		{ -0.00000048, -0.00000852, 0.00000274, 0.00001132, 0.00000081, -0.00000525, -0.00000128, -0.00000332, -0.00000102, 0.00001097, 0.00000861, -0.00000685, -0.00000915, -0.00000098, 0.00000187, 0.00000261, 0.00000432, 0.00000177, -0.00000622, -0.00001336, -0.00000691, 0.00000846, 0.00000624, -0.00000688, 0.00000007, 0.00000863, -0.00000616, -0.00001146, 0.00000411, 0.00000345, -0.00000836, -0.00000475, -0.00000031, -0.00000178, -0.00000075, -0.00000642, -0.00000597, 0.00000969, 0.00000511, -0.00000880, 0.00000733, 0.00000917, -0.00002218, -0.00001810, 0.00000776, -0.00000211, -0.00000800, 0.00000454, 0.00000057, 0.00000217, -0.00000913, -0.00004022, 0.00000734, 0.00004463, -0.00006184, -0.00004792, 0.00013016, 0.00000644, -0.00021225, 0.00005187, 0.00020406, -0.00022230, -0.00019313, 0.00034473, 0.00000334, -0.00056330, 0.00014627, 0.00049872, -0.00090272, -0.00123582, 0.00102907, 0.00232110, 0.00062098, -0.00102071, -0.00059160, 0.00015456, 0.00002833, -0.00017052, -0.00007092, -0.00000192, 0.00000455, -0.00000765, -0.00004267, -0.00000381, 0.00003578, -0.00003122, -0.00005566, 0.00001452, 0.00002052, -0.00002276, -0.00000675, 0.00001095, -0.00001351, -0.00001821, -0.00000040, 0.00000276, -0.00000222, -0.00000567, -0.00000335, 0.00000357, -0.00000494, -0.00001719, -0.00000398, 0.00001035, 0.00000161, -0.00000322, 0.00000342, 0.00000042, -0.00000620, -0.00000564, -0.00000441, -0.00000138, 0.00000249, -0.00000086, -0.00000406, -0.00000222, -0.00000573, -0.00000953, -0.00000056, 0.00000995, 0.00000558, -0.00000669, -0.00000948, -0.00000130, 0.00000181, -0.00000398, -0.00000177, 0.00000605, },
		{ -0.00000428, 0.00000904, 0.00000654, -0.00000380, -0.00000261, -0.00000165, -0.00000353, 0.00000593, 0.00001180, -0.00000117, -0.00001030, -0.00000327, 0.00000200, 0.00000145, 0.00000366, 0.00000387, -0.00000340, -0.00001210, -0.00001057, 0.00000349, 0.00000947, -0.00000195, -0.00000391, 0.00000657, -0.00000123, -0.00001339, -0.00000081, 0.00000751, -0.00000583, -0.00000783, 0.00000127, -0.00000166, -0.00000458, -0.00000296, -0.00000437, 0.00000278, 0.00000710, -0.00000305, 0.00000215, 0.00000934, -0.00001584, -0.00002272, 0.00000691, 0.00000307, -0.00001599, 0.00000443, 0.00001212, -0.00000833, -0.00001256, -0.00001959, -0.00000817, 0.00002746, -0.00002199, -0.00005626, 0.00007246, 0.00005756, -0.00016258, -0.00004543, 0.00020353, -0.00009177, -0.00027300, 0.00022877, 0.00018987, -0.00051818, -0.00013543, 0.00063061, -0.00045727, -0.00154362, 0.00029445, 0.00246054, 0.00132269, -0.00094202, -0.00093307, 0.00014643, 0.00015852, -0.00023148, -0.00011358, 0.00004545, -0.00000762, -0.00002956, -0.00002009, -0.00001096, 0.00002002, -0.00001279, -0.00005428, -0.00000005, 0.00002973, -0.00001878, -0.00001787, 0.00001614, -0.00000703, -0.00002706, -0.00000168, 0.00000896, -0.00000372, -0.00000703, -0.00000242, 0.00000192, -0.00000343, -0.00001620, -0.00000845, 0.00001064, 0.00000566, -0.00000529, 0.00000197, 0.00000362, -0.00000634, -0.00000684, -0.00000300, -0.00000296, 0.00000055, 0.00000172, -0.00000314, -0.00000389, -0.00000406, -0.00000835, -0.00000421, 0.00000760, 0.00000883, -0.00000226, -0.00001001, -0.00000598, 0.00000094, -0.00000065, -0.00000295, 0.00000347, 0.00000373, -0.00000672, },
		{ 0.00000979, 0.00000056, -0.00000382, -0.00000104, -0.00000332, 0.00000053, 0.00001086, 0.00000529, -0.00000847, -0.00000652, 0.00000165, 0.00000160, 0.00000198, 0.00000450, -0.00000025, -0.00000970, -0.00001224, -0.00000203, 0.00000891, 0.00000387, -0.00000413, 0.00000208, 0.00000181, -0.00001132, -0.00000645, 0.00000768, -0.00000074, -0.00000958, 0.00000006, 0.00000063, -0.00000743, -0.00000351, -0.00000036, -0.00000113, 0.00000401, 0.00000294, 0.00000130, 0.00000568, -0.00000986, -0.00002307, 0.00000163, 0.00000874, -0.00001737, -0.00000360, 0.00002107, -0.00000620, -0.00002371, -0.00000817, -0.00000502, 0.00000531, -0.00000334, -0.00003790, 0.00002351, 0.00006381, -0.00009343, -0.00009782, 0.00014624, 0.00002064, -0.00026689, 0.00007585, 0.00028821, -0.00036121, -0.00037009, 0.00055827, 0.00001310, -0.00153390, -0.00048803, 0.00223040, 0.00203494, -0.00057949, -0.00128603, -0.00001190, 0.00033858, -0.00023353, -0.00021579, 0.00008909, 0.00002066, -0.00007463, -0.00001093, 0.00000770, -0.00000313, -0.00000959, -0.00003859, -0.00001318, 0.00002867, -0.00000761, -0.00002700, 0.00001343, 0.00000407, -0.00003216, -0.00000925, 0.00001583, -0.00000169, -0.00001090, -0.00000105, 0.00000176, -0.00000428, -0.00001401, -0.00001115, 0.00000779, 0.00000978, -0.00000460, -0.00000135, 0.00000586, -0.00000443, -0.00000898, -0.00000204, -0.00000243, -0.00000275, 0.00000246, -0.00000010, -0.00000531, -0.00000445, -0.00000577, -0.00000591, 0.00000357, 0.00000990, 0.00000251, -0.00000805, -0.00000936, -0.00000248, 0.00000126, -0.00000136, 0.00000067, 0.00000486, -0.00000236, -0.00000804, 0.00000298, },
		{ -0.00000290, -0.00000198, -0.00000184, -0.00000278, 0.00000641, 0.00000965, -0.00000332, -0.00000891, -0.00000046, 0.00000273, 0.00000086, 0.00000325, 0.00000223, -0.00000631, -0.00001208, -0.00000662, 0.00000539, 0.00000787, -0.00000075, -0.00000183, 0.00000138, -0.00000726, -0.00000966, 0.00000404, 0.00000388, -0.00000815, -0.00000296, 0.00000246, -0.00000745, -0.00000666, 0.00000247, -0.00000064, -0.00000112, 0.00000543, 0.00000462, 0.00000176, -0.00000742, -0.00001955, -0.00000413, 0.00001060, -0.00001268, -0.00001248, 0.00002167, 0.00000410, -0.00003155, -0.00000991, 0.00000654, -0.00000753, -0.00000562, -0.00001256, 0.00000004, 0.00003898, -0.00003753, -0.00009918, 0.00006943, 0.00007989, -0.00019664, -0.00005284, 0.00028218, -0.00016324, -0.00048443, 0.00033963, 0.00036751, -0.00124349, -0.00112628, 0.00165456, 0.00256776, 0.00005995, -0.00151243, -0.00034338, 0.00049304, -0.00013186, -0.00035240, 0.00008527, 0.00009752, -0.00011644, -0.00003768, 0.00004646, -0.00001290, -0.00002995, -0.00001883, -0.00001289, 0.00001650, 0.00000302, -0.00002745, 0.00000363, 0.00001348, -0.00002951, -0.00002071, 0.00001877, 0.00000516, -0.00001535, -0.00000179, 0.00000425, -0.00000634, -0.00001313, -0.00001083, 0.00000352, 0.00001133, -0.00000097, -0.00000435, 0.00000549, -0.00000095, -0.00001043, -0.00000283, -0.00000015, -0.00000531, 0.00000026, 0.00000358, -0.00000457, -0.00000681, -0.00000376, -0.00000480, -0.00000059, 0.00000826, 0.00000656, -0.00000429, -0.00001057, -0.00000672, 0.00000051, 0.00000128, -0.00000025, 0.00000342, 0.00000161, -0.00000754, -0.00000384, 0.00000860, 0.00000608, },
		{ -0.00000105, -0.00000323, 0.00000093, 0.00000986, 0.00000339, -0.00000829, -0.00000416, 0.00000327, 0.00000147, 0.00000107, 0.00000286, -0.00000248, -0.00001018, -0.00000957, 0.00000051, 0.00000870, 0.00000418, -0.00000285, -0.00000164, -0.00000415, -0.00000936, -0.00000098, 0.00000548, -0.00000408, -0.00000538, 0.00000210, -0.00000522, -0.00000964, 0.00000209, 0.00000259, -0.00000454, 0.00000345, 0.00000912, 0.00000082, -0.00000860, -0.00001523, -0.00000697, 0.00000821, -0.00000612, -0.00001676, 0.00001468, 0.00001449, -0.00003012, -0.00001994, 0.00001479, -0.00000643, -0.00001795, 0.00000305, 0.00000192, 0.00000681, -0.00001319, -0.00006722, 0.00001000, 0.00008123, -0.00010633, -0.00011805, 0.00020083, 0.00000071, -0.00046466, 0.00007702, 0.00052627, -0.00079523, -0.00148086, 0.00086615, 0.00276537, 0.00086615, -0.00148086, -0.00079523, 0.00052627, 0.00007702, -0.00046466, 0.00000071, 0.00020083, -0.00011805, -0.00010633, 0.00008123, 0.00001000, -0.00006722, -0.00001319, 0.00000681, 0.00000192, 0.00000305, -0.00001795, -0.00000643, 0.00001479, -0.00001994, -0.00003012, 0.00001449, 0.00001468, -0.00001676, -0.00000612, 0.00000821, -0.00000697, -0.00001523, -0.00000860, 0.00000082, 0.00000912, 0.00000345, -0.00000454, 0.00000259, 0.00000209, -0.00000964, -0.00000522, 0.00000210, -0.00000538, -0.00000408, 0.00000548, -0.00000098, -0.00000936, -0.00000415, -0.00000164, -0.00000285, 0.00000418, 0.00000870, 0.00000051, -0.00000957, -0.00001018, -0.00000248, 0.00000286, 0.00000107, 0.00000147, 0.00000327, -0.00000416, -0.00000829, 0.00000339, 0.00000986, 0.00000093, -0.00000323, },
		{ -0.00000290, 0.00000608, 0.00000860, -0.00000384, -0.00000754, 0.00000161, 0.00000342, -0.00000025, 0.00000128, 0.00000051, -0.00000672, -0.00001057, -0.00000429, 0.00000656, 0.00000826, -0.00000059, -0.00000480, -0.00000376, -0.00000681, -0.00000457, 0.00000358, 0.00000026, -0.00000531, -0.00000015, -0.00000283, -0.00001043, -0.00000095, 0.00000549, -0.00000435, -0.00000097, 0.00001133, 0.00000352, -0.00001083, -0.00001313, -0.00000634, 0.00000425, -0.00000179, -0.00001535, 0.00000516, 0.00001877, -0.00002071, -0.00002951, 0.00001348, 0.00000363, -0.00002745, 0.00000302, 0.00001650, -0.00001289, -0.00001883, -0.00002995, -0.00001290, 0.00004646, -0.00003768, -0.00011644, 0.00009752, 0.00008527, -0.00035240, -0.00013186, 0.00049304, -0.00034338, -0.00151243, 0.00005995, 0.00256776, 0.00165456, -0.00112628, -0.00124349, 0.00036751, 0.00033963, -0.00048443, -0.00016324, 0.00028218, -0.00005284, -0.00019664, 0.00007989, 0.00006943, -0.00009918, -0.00003753, 0.00003898, 0.00000004, -0.00001256, -0.00000562, -0.00000753, 0.00000654, -0.00000991, -0.00003155, 0.00000410, 0.00002167, -0.00001248, -0.00001268, 0.00001060, -0.00000413, -0.00001955, -0.00000742, 0.00000176, 0.00000462, 0.00000543, -0.00000112, -0.00000064, 0.00000247, -0.00000666, -0.00000745, 0.00000246, -0.00000296, -0.00000815, 0.00000388, 0.00000404, -0.00000966, -0.00000726, 0.00000138, -0.00000183, -0.00000075, 0.00000787, 0.00000539, -0.00000662, -0.00001208, -0.00000631, 0.00000223, 0.00000325, 0.00000086, 0.00000273, -0.00000046, -0.00000891, -0.00000332, 0.00000965, 0.00000641, -0.00000278, -0.00000184, -0.00000198, },
		{ 0.00000979, 0.00000298, -0.00000804, -0.00000236, 0.00000486, 0.00000067, -0.00000136, 0.00000126, -0.00000248, -0.00000936, -0.00000805, 0.00000251, 0.00000990, 0.00000357, -0.00000591, -0.00000577, -0.00000445, -0.00000531, -0.00000010, 0.00000246, -0.00000275, -0.00000243, -0.00000204, -0.00000898, -0.00000443, 0.00000586, -0.00000135, -0.00000460, 0.00000978, 0.00000779, -0.00001115, -0.00001401, -0.00000428, 0.00000176, -0.00000105, -0.00001090, -0.00000169, 0.00001583, -0.00000925, -0.00003216, 0.00000407, 0.00001343, -0.00002700, -0.00000761, 0.00002867, -0.00001318, -0.00003859, -0.00000959, -0.00000313, 0.00000770, -0.00001093, -0.00007463, 0.00002066, 0.00008909, -0.00021579, -0.00023353, 0.00033858, -0.00001190, -0.00128603, -0.00057949, 0.00203494, 0.00223040, -0.00048803, -0.00153390, 0.00001310, 0.00055827, -0.00037009, -0.00036121, 0.00028821, 0.00007585, -0.00026689, 0.00002064, 0.00014624, -0.00009782, -0.00009343, 0.00006381, 0.00002351, -0.00003790, -0.00000334, 0.00000531, -0.00000502, -0.00000817, -0.00002371, -0.00000620, 0.00002107, -0.00000360, -0.00001737, 0.00000874, 0.00000163, -0.00002307, -0.00000986, 0.00000568, 0.00000130, 0.00000294, 0.00000401, -0.00000113, -0.00000036, -0.00000351, -0.00000743, 0.00000063, 0.00000006, -0.00000958, -0.00000074, 0.00000768, -0.00000645, -0.00001132, 0.00000181, 0.00000208, -0.00000413, 0.00000387, 0.00000891, -0.00000203, -0.00001224, -0.00000970, -0.00000025, 0.00000450, 0.00000198, 0.00000160, 0.00000165, -0.00000652, -0.00000847, 0.00000529, 0.00001086, 0.00000053, -0.00000332, -0.00000104, -0.00000382, 0.00000056, },
		{ -0.00000428, -0.00000672, 0.00000373, 0.00000347, -0.00000295, -0.00000065, 0.00000094, -0.00000598, -0.00001001, -0.00000226, 0.00000883, 0.00000760, -0.00000421, -0.00000835, -0.00000406, -0.00000389, -0.00000314, 0.00000172, 0.00000055, -0.00000296, -0.00000300, -0.00000684, -0.00000634, 0.00000362, 0.00000197, -0.00000529, 0.00000566, 0.00001064, -0.00000845, -0.00001620, -0.00000343, 0.00000192, -0.00000242, -0.00000703, -0.00000372, 0.00000896, -0.00000168, -0.00002706, -0.00000703, 0.00001614, -0.00001787, -0.00001878, 0.00002973, -0.00000005, -0.00005428, -0.00001279, 0.00002002, -0.00001096, -0.00002009, -0.00002956, -0.00000762, 0.00004545, -0.00011358, -0.00023148, 0.00015852, 0.00014643, -0.00093307, -0.00094202, 0.00132269, 0.00246054, 0.00029445, -0.00154362, -0.00045727, 0.00063061, -0.00013543, -0.00051818, 0.00018987, 0.00022877, -0.00027300, -0.00009177, 0.00020353, -0.00004543, -0.00016258, 0.00005756, 0.00007246, -0.00005626, -0.00002199, 0.00002746, -0.00000817, -0.00001959, -0.00001256, -0.00000833, 0.00001212, 0.00000443, -0.00001599, 0.00000307, 0.00000691, -0.00002272, -0.00001584, 0.00000934, 0.00000215, -0.00000305, 0.00000710, 0.00000278, -0.00000437, -0.00000296, -0.00000458, -0.00000166, 0.00000127, -0.00000783, -0.00000583, 0.00000751, -0.00000081, -0.00001339, -0.00000123, 0.00000657, -0.00000391, -0.00000195, 0.00000947, 0.00000349, -0.00001057, -0.00001210, -0.00000340, 0.00000387, 0.00000366, 0.00000145, 0.00000200, -0.00000327, -0.00001030, -0.00000117, 0.00001180, 0.00000593, -0.00000353, -0.00000165, -0.00000261, -0.00000380, 0.00000654, 0.00000904, },
		{ -0.00000048, 0.00000605, -0.00000177, -0.00000398, 0.00000181, -0.00000130, -0.00000948, -0.00000669, 0.00000558, 0.00000995, -0.00000056, -0.00000953, -0.00000573, -0.00000222, -0.00000406, -0.00000086, 0.00000249, -0.00000138, -0.00000441, -0.00000564, -0.00000620, 0.00000042, 0.00000342, -0.00000322, 0.00000161, 0.00001035, -0.00000398, -0.00001719, -0.00000494, 0.00000357, -0.00000335, -0.00000567, -0.00000222, 0.00000276, -0.00000040, -0.00001821, -0.00001351, 0.00001095, -0.00000675, -0.00002276, 0.00002052, 0.00001452, -0.00005566, -0.00003122, 0.00003578, -0.00000381, -0.00004267, -0.00000765, 0.00000455, -0.00000192, -0.00007092, -0.00017052, 0.00002833, 0.00015456, -0.00059160, -0.00102071, 0.00062098, 0.00232110, 0.00102907, -0.00123582, -0.00090272, 0.00049872, 0.00014627, -0.00056330, 0.00000334, 0.00034473, -0.00019313, -0.00022230, 0.00020406, 0.00005187, -0.00021225, 0.00000644, 0.00013016, -0.00004792, -0.00006184, 0.00004463, 0.00000734, -0.00004022, -0.00000913, 0.00000217, 0.00000057, 0.00000454, -0.00000800, -0.00000211, 0.00000776, -0.00001810, -0.00002218, 0.00000917, 0.00000733, -0.00000880, 0.00000511, 0.00000969, -0.00000597, -0.00000642, -0.00000075, -0.00000178, -0.00000031, -0.00000475, -0.00000836, 0.00000345, 0.00000411, -0.00001146, -0.00000616, 0.00000863, 0.00000007, -0.00000688, 0.00000624, 0.00000846, -0.00000691, -0.00001336, -0.00000622, 0.00000177, 0.00000432, 0.00000261, 0.00000187, -0.00000098, -0.00000915, -0.00000685, 0.00000861, 0.00001097, -0.00000102, -0.00000332, -0.00000128, -0.00000525, 0.00000081, 0.00001132, 0.00000274, -0.00000852, },
		{ 0.00000202, -0.00000622, -0.00000057, 0.00000278, -0.00000618, -0.00000970, 0.00000095, 0.00001003, 0.00000348, -0.00000843, -0.00000811, -0.00000192, -0.00000311, -0.00000331, 0.00000212, 0.00000109, -0.00000472, -0.00000584, -0.00000504, -0.00000189, 0.00000247, -0.00000028, -0.00000019, 0.00000744, -0.00000021, -0.00001562, -0.00000784, 0.00000456, -0.00000243, -0.00000628, -0.00000008, 0.00000002, -0.00000369, -0.00001096, -0.00001310, 0.00000222, -0.00000029, -0.00001834, 0.00000832, 0.00002116, -0.00004391, -0.00005012, 0.00003349, 0.00001824, -0.00005582, -0.00001404, 0.00003102, -0.00002291, -0.00007680, -0.00010543, -0.00002316, 0.00008280, -0.00035401, -0.00089639, 0.00008381, 0.00189990, 0.00154200, -0.00068028, -0.00117469, 0.00018298, 0.00036892, -0.00046698, -0.00021090, 0.00036942, -0.00004365, -0.00031901, 0.00013115, 0.00016220, -0.00021089, -0.00008229, 0.00016830, -0.00000206, -0.00010872, 0.00003960, 0.00004241, -0.00005704, -0.00002314, 0.00002145, -0.00000329, -0.00000692, 0.00000148, -0.00000085, 0.00000312, -0.00001269, -0.00002466, 0.00000424, 0.00001352, -0.00001026, -0.00000184, 0.00001549, -0.00000246, -0.00001229, 0.00000075, 0.00000166, -0.00000314, -0.00000331, -0.00000706, -0.00000195, 0.00000518, -0.00000613, -0.00000970, 0.00000664, 0.00000560, -0.00000820, 0.00000018, 0.00001085, -0.00000152, -0.00001320, -0.00000843, -0.00000066, 0.00000320, 0.00000396, 0.00000259, 0.00000004, -0.00000679, -0.00000976, 0.00000282, 0.00001307, 0.00000392, -0.00000421, -0.00000141, -0.00000415, -0.00000435, 0.00000877, 0.00000988, -0.00000571, -0.00000599, 0.00000583, },
		{ -0.00000492, 0.00000409, -0.00000090, -0.00001013, -0.00000412, 0.00000787, 0.00000655, -0.00000548, -0.00000958, -0.00000328, -0.00000173, -0.00000425, 0.00000008, 0.00000274, -0.00000341, -0.00000662, -0.00000423, -0.00000253, 0.00000023, 0.00000149, 0.00000056, 0.00000400, 0.00000105, -0.00001215, -0.00001010, 0.00000362, -0.00000014, -0.00000698, 0.00000058, 0.00000057, -0.00000781, -0.00000846, -0.00000808, -0.00000435, -0.00000087, -0.00001000, 0.00000058, 0.00001793, -0.00002782, -0.00005832, 0.00001596, 0.00003805, -0.00004906, -0.00003566, 0.00004645, -0.00001241, -0.00009995, -0.00007237, -0.00001267, 0.00000643, -0.00024364, -0.00068959, -0.00021880, 0.00135435, 0.00173995, -0.00002892, -0.00117829, -0.00021831, 0.00044064, -0.00026097, -0.00037321, 0.00028436, 0.00012389, -0.00033804, 0.00000118, 0.00024066, -0.00014606, -0.00017962, 0.00016073, 0.00007403, -0.00013872, 0.00000268, 0.00008503, -0.00005412, -0.00005570, 0.00003727, 0.00000947, -0.00002606, 0.00000334, 0.00000998, -0.00000298, -0.00001194, -0.00002152, -0.00000235, 0.00001608, -0.00000619, -0.00000994, 0.00001604, 0.00000553, -0.00001660, -0.00000244, 0.00000723, -0.00000411, -0.00000529, -0.00000352, -0.00000502, 0.00000151, -0.00000061, -0.00000901, 0.00000170, 0.00000919, -0.00000502, -0.00000595, 0.00000912, 0.00000428, -0.00001103, -0.00001012, -0.00000259, 0.00000088, 0.00000400, 0.00000438, 0.00000082, -0.00000491, -0.00000985, -0.00000288, 0.00001114, 0.00000916, -0.00000259, -0.00000293, -0.00000243, -0.00000671, 0.00000292, 0.00001365, 0.00000128, -0.00000949, 0.00000173, 0.00000614, -0.00000517, },
		{ 0.00000424, -0.00000726, -0.00000835, 0.00000384, 0.00000792, -0.00000183, -0.00000929, -0.00000533, -0.00000128, -0.00000368, -0.00000210, 0.00000262, -0.00000126, -0.00000675, -0.00000441, -0.00000208, -0.00000170, 0.00000126, 0.00000252, 0.00000206, -0.00000016, -0.00000873, -0.00001030, 0.00000119, 0.00000188, -0.00000633, -0.00000041, 0.00000234, -0.00000979, -0.00001033, -0.00000286, -0.00000560, -0.00000599, -0.00000373, 0.00000024, 0.00000957, -0.00001651, -0.00005373, -0.00000555, 0.00004363, -0.00002707, -0.00005347, 0.00004008, 0.00001430, -0.00011111, -0.00007685, 0.00001972, -0.00002840, -0.00022595, -0.00050588, -0.00030841, 0.00084623, 0.00163690, 0.00054315, -0.00091246, -0.00056633, 0.00032827, -0.00002855, -0.00042291, 0.00011963, 0.00024547, -0.00026712, -0.00014064, 0.00025114, -0.00003409, -0.00024670, 0.00009852, 0.00015491, -0.00012999, -0.00006024, 0.00011419, -0.00002202, -0.00009589, 0.00003466, 0.00003966, -0.00004141, -0.00001020, 0.00002687, -0.00000279, -0.00001919, -0.00001581, -0.00000508, 0.00001268, 0.00000029, -0.00001405, 0.00001035, 0.00001383, -0.00001555, -0.00000951, 0.00001127, -0.00000088, -0.00000965, -0.00000118, -0.00000360, -0.00000458, 0.00000117, -0.00000390, -0.00000265, 0.00000833, 0.00000081, -0.00000879, 0.00000359, 0.00000819, -0.00000655, -0.00001100, -0.00000404, -0.00000139, 0.00000210, 0.00000605, 0.00000257, -0.00000389, -0.00000857, -0.00000635, 0.00000626, 0.00001204, 0.00000168, -0.00000416, -0.00000196, -0.00000613, -0.00000308, 0.00001230, 0.00000945, -0.00000821, -0.00000477, 0.00000746, -0.00000064, -0.00000840, 0.00000153, },
		{ -0.00001019, -0.00000141, 0.00000727, 0.00000149, -0.00000743, -0.00000681, -0.00000206, -0.00000267, -0.00000323, 0.00000111, 0.00000039, -0.00000568, -0.00000510, -0.00000162, -0.00000242, -0.00000032, 0.00000390, 0.00000227, -0.00000229, -0.00000709, -0.00000865, -0.00000116, 0.00000236, -0.00000439, -0.00000177, 0.00000333, -0.00000896, -0.00001382, -0.00000089, -0.00000228, -0.00001091, -0.00000289, 0.00000484, 0.00000282, -0.00001382, -0.00004226, -0.00002002, 0.00003398, -0.00000327, -0.00005527, 0.00001802, 0.00003607, -0.00009843, -0.00010057, 0.00003816, -0.00001612, -0.00024188, -0.00039958, -0.00026785, 0.00048372, 0.00133559, 0.00090398, -0.00046968, -0.00073944, 0.00007631, 0.00013239, -0.00034946, -0.00005830, 0.00027583, -0.00013411, -0.00024084, 0.00018575, 0.00008494, -0.00025376, -0.00000243, 0.00020751, -0.00007566, -0.00012759, 0.00011037, 0.00003469, -0.00012433, 0.00000468, 0.00007693, -0.00003915, -0.00003948, 0.00003957, 0.00001053, -0.00003192, -0.00001455, -0.00000006, 0.00000588, 0.00000336, -0.00001148, 0.00000197, 0.00001716, -0.00000865, -0.00001634, 0.00001034, 0.00000586, -0.00001293, -0.00000280, 0.00000128, -0.00000887, -0.00000264, 0.00000253, -0.00000273, 0.00000345, 0.00000544, -0.00000655, -0.00000313, 0.00000795, -0.00000067, -0.00001014, -0.00000545, -0.00000280, -0.00000096, 0.00000605, 0.00000539, -0.00000281, -0.00000747, -0.00000720, 0.00000090, 0.00001108, 0.00000685, -0.00000319, -0.00000311, -0.00000436, -0.00000654, 0.00000685, 0.00001491, -0.00000188, -0.00001011, 0.00000416, 0.00000507, -0.00000822, -0.00000379, 0.00000660, -0.00000161, },
		{ 0.00000448, 0.00000378, -0.00000475, -0.00000702, -0.00000339, -0.00000227, -0.00000325, -0.00000063, 0.00000074, -0.00000392, -0.00000540, -0.00000177, -0.00000215, -0.00000197, 0.00000375, 0.00000375, -0.00000353, -0.00000754, -0.00000660, -0.00000207, 0.00000139, -0.00000233, -0.00000221, 0.00000294, -0.00000662, -0.00001611, -0.00000275, 0.00000223, -0.00001207, -0.00000661, 0.00000940, 0.00000166, -0.00001730, -0.00003221, -0.00002310, 0.00001713, 0.00001029, -0.00004141, -0.00000472, 0.00004006, -0.00006914, -0.00011910, 0.00002857, 0.00001967, -0.00024305, -0.00036849, -0.00019258, 0.00029474, 0.00097579, 0.00100867, 0.00000430, -0.00068416, -0.00021059, 0.00015581, -0.00019935, -0.00017866, 0.00021152, 0.00000542, -0.00026392, 0.00007054, 0.00016570, -0.00019627, -0.00010675, 0.00020756, 0.00001002, -0.00017127, 0.00006687, 0.00009711, -0.00012264, -0.00004857, 0.00010307, -0.00001124, -0.00007467, 0.00003579, 0.00003730, -0.00004136, -0.00002440, 0.00001112, 0.00000252, -0.00000140, -0.00000469, -0.00000277, 0.00001301, 0.00000011, -0.00001801, 0.00000419, 0.00001230, -0.00001179, -0.00000811, 0.00000587, -0.00000812, -0.00001027, 0.00000568, 0.00000264, -0.00000195, 0.00000544, -0.00000048, -0.00000717, 0.00000306, 0.00000421, -0.00000664, -0.00000672, -0.00000360, -0.00000369, 0.00000373, 0.00000796, -0.00000050, -0.00000699, -0.00000662, -0.00000270, 0.00000690, 0.00001023, 0.00000059, -0.00000448, -0.00000355, -0.00000675, 0.00000037, 0.00001516, 0.00000685, -0.00001079, -0.00000274, 0.00000830, -0.00000360, -0.00000879, 0.00000450, 0.00000464, -0.00000835, -0.00000662, },
		{ -0.00000196, -0.00000609, -0.00000428, -0.00000270, -0.00000298, -0.00000170, 0.00000002, -0.00000248, -0.00000487, -0.00000232, -0.00000167, -0.00000280, 0.00000233, 0.00000503, -0.00000307, -0.00000899, -0.00000571, -0.00000141, 0.00000011, -0.00000130, -0.00000150, 0.00000192, -0.00000449, -0.00001602, -0.00000665, 0.00000466, -0.00000913, -0.00001128, 0.00001015, 0.00000541, -0.00002137, -0.00002861, -0.00001803, 0.00000328, 0.00000983, -0.00002186, -0.00001558, 0.00002756, -0.00003996, -0.00011811, -0.00000041, 0.00004896, -0.00021154, -0.00037511, -0.00014880, 0.00023823, 0.00067597, 0.00090313, 0.00037091, -0.00043502, -0.00041384, 0.00003966, -0.00005375, -0.00020173, 0.00009246, 0.00009570, -0.00021015, -0.00004671, 0.00017992, -0.00009844, -0.00017598, 0.00015284, 0.00009663, -0.00017078, -0.00000458, 0.00014007, -0.00008399, -0.00010761, 0.00010092, 0.00003822, -0.00009891, 0.00000854, 0.00006869, -0.00003658, -0.00004635, 0.00002068, 0.00000957, -0.00001296, -0.00000069, 0.00000091, 0.00000426, 0.00000425, -0.00001252, -0.00000321, 0.00001393, -0.00000588, -0.00001357, 0.00000638, -0.00000253, -0.00001714, 0.00000251, 0.00001082, -0.00000337, 0.00000027, 0.00000546, -0.00000574, -0.00000428, 0.00000521, -0.00000080, -0.00000675, -0.00000465, -0.00000512, 0.00000005, 0.00000853, 0.00000309, -0.00000622, -0.00000608, -0.00000387, 0.00000173, 0.00001000, 0.00000581, -0.00000399, -0.00000451, -0.00000521, -0.00000398, 0.00001054, 0.00001378, -0.00000566, -0.00000963, 0.00000640, 0.00000331, -0.00000999, -0.00000133, 0.00000837, -0.00000292, -0.00000992, -0.00000022, 0.00000452, },
		{ -0.00000429, -0.00000335, -0.00000318, -0.00000209, -0.00000086, -0.00000194, -0.00000387, -0.00000264, -0.00000149, -0.00000278, 0.00000063, 0.00000514, -0.00000144, -0.00000993, -0.00000645, -0.00000023, -0.00000027, -0.00000147, -0.00000044, 0.00000133, -0.00000341, -0.00001421, -0.00001013, 0.00000376, -0.00000439, -0.00001344, 0.00000672, 0.00001030, -0.00002163, -0.00003102, -0.00001177, -0.00000170, 0.00000035, -0.00000801, -0.00001173, 0.00000987, -0.00002422, -0.00009933, -0.00002951, 0.00005467, -0.00015827, -0.00037778, -0.00015572, 0.00024123, 0.00049456, 0.00069034, 0.00055081, -0.00009486, -0.00045335, -0.00015301, 0.00001212, -0.00013843, -0.00001943, 0.00010859, -0.00011462, -0.00012035, 0.00013041, -0.00000169, -0.00018758, 0.00006540, 0.00015189, -0.00012452, -0.00007810, 0.00014469, -0.00001832, -0.00014890, 0.00006406, 0.00009270, -0.00009647, -0.00003802, 0.00008956, -0.00001122, -0.00007300, 0.00001825, 0.00002878, -0.00002389, -0.00000683, 0.00001215, -0.00000162, -0.00000084, -0.00000324, -0.00000572, 0.00000905, 0.00000115, -0.00001478, 0.00000204, 0.00000411, -0.00001888, -0.00000607, 0.00001671, 0.00000156, -0.00000690, 0.00000693, 0.00000082, -0.00000991, 0.00000086, 0.00000516, -0.00000403, -0.00000625, -0.00000561, -0.00000316, 0.00000632, 0.00000650, -0.00000393, -0.00000605, -0.00000355, -0.00000211, 0.00000630, 0.00000985, -0.00000059, -0.00000619, -0.00000423, -0.00000488, 0.00000394, 0.00001562, 0.00000313, -0.00001231, -0.00000035, 0.00000841, -0.00000589, -0.00000766, 0.00000722, 0.00000405, -0.00000944, -0.00000579, 0.00000317, 0.00000048, -0.00000450, },
		{ -0.00000388, -0.00000249, -0.00000127, -0.00000205, -0.00000314, -0.00000247, -0.00000154, -0.00000243, -0.00000058, 0.00000413, 0.00000022, -0.00000954, -0.00000813, 0.00000019, 0.00000054, -0.00000206, 0.00000001, 0.00000155, -0.00000305, -0.00001202, -0.00001175, 0.00000052, -0.00000076, -0.00001197, 0.00000162, 0.00001265, -0.00001739, -0.00003524, -0.00000987, 0.00000129, -0.00000933, -0.00000552, 0.00000019, -0.00000093, -0.00002426, -0.00007542, -0.00004373, 0.00003899, -0.00010673, -0.00035406, -0.00019449, 0.00023986, 0.00042456, 0.00047900, 0.00054491, 0.00021313, -0.00032050, -0.00032485, -0.00003322, -0.00004308, -0.00007315, 0.00005632, -0.00002787, -0.00012947, 0.00004856, 0.00005620, -0.00014590, -0.00001908, 0.00015767, -0.00005144, -0.00012599, 0.00010853, 0.00005118, -0.00015447, 0.00000180, 0.00013019, -0.00006180, -0.00008847, 0.00008585, 0.00003108, -0.00009083, -0.00000272, 0.00005374, -0.00002385, -0.00002533, 0.00002349, 0.00000302, -0.00001389, 0.00000220, 0.00000069, 0.00000108, 0.00000360, -0.00001003, -0.00000384, 0.00000698, -0.00001454, -0.00001539, 0.00001601, 0.00001102, -0.00001094, 0.00000220, 0.00000873, -0.00001002, -0.00000739, 0.00000787, 0.00000187, -0.00000734, -0.00000646, -0.00000460, 0.00000236, 0.00000770, -0.00000003, -0.00000562, -0.00000319, -0.00000366, 0.00000111, 0.00001055, 0.00000478, -0.00000644, -0.00000514, -0.00000344, -0.00000109, 0.00001192, 0.00001124, -0.00000863, -0.00000844, 0.00000813, 0.00000178, -0.00001039, 0.00000136, 0.00000909, -0.00000465, -0.00001016, -0.00000057, 0.00000201, -0.00000259, -0.00000362, -0.00000348, },
		{ -0.00000138, -0.00000208, -0.00000304, -0.00000213, -0.00000142, -0.00000212, -0.00000114, 0.00000268, 0.00000102, -0.00000806, -0.00000952, -0.00000070, 0.00000184, -0.00000206, -0.00000044, 0.00000210, -0.00000263, -0.00001036, -0.00001161, -0.00000307, 0.00000014, -0.00000825, -0.00000206, 0.00001125, -0.00001123, -0.00003696, -0.00001339, 0.00000686, -0.00001268, -0.00001212, 0.00001014, 0.00000027, -0.00003235, -0.00005934, -0.00004065, 0.00001661, -0.00007472, -0.00030665, -0.00023245, 0.00020434, 0.00041721, 0.00034003, 0.00041864, 0.00039743, -0.00007674, -0.00039072, -0.00016245, 0.00001680, -0.00005499, -0.00001783, 0.00001119, -0.00008670, -0.00002318, 0.00005964, -0.00007891, -0.00006926, 0.00011889, 0.00001825, -0.00013259, 0.00004730, 0.00009913, -0.00012135, -0.00006412, 0.00013395, -0.00000394, -0.00012252, 0.00005283, 0.00007645, -0.00008702, -0.00003962, 0.00007183, -0.00000587, -0.00005024, 0.00002439, 0.00002057, -0.00002701, -0.00000365, 0.00001422, -0.00000271, -0.00000216, -0.00000240, -0.00000576, 0.00000391, -0.00000764, -0.00001994, 0.00000843, 0.00001968, -0.00000814, -0.00000629, 0.00001289, -0.00000379, -0.00001534, 0.00000483, 0.00000924, -0.00000592, -0.00000852, -0.00000457, -0.00000117, 0.00000560, 0.00000397, -0.00000342, -0.00000332, -0.00000366, -0.00000320, 0.00000758, 0.00000965, -0.00000377, -0.00000727, -0.00000207, -0.00000257, 0.00000524, 0.00001464, -0.00000001, -0.00001310, 0.00000187, 0.00000864, -0.00000711, -0.00000607, 0.00000908, 0.00000272, -0.00001096, -0.00000603, 0.00000177, -0.00000052, -0.00000265, -0.00000289, -0.00000443, -0.00000347, },
		{ -0.00000320, -0.00000220, -0.00000104, -0.00000177, -0.00000135, 0.00000144, 0.00000089, -0.00000639, -0.00000986, -0.00000244, 0.00000261, -0.00000113, -0.00000126, 0.00000221, -0.00000175, -0.00000922, -0.00001068, -0.00000550, -0.00000135, -0.00000471, -0.00000277, 0.00000759, -0.00000644, -0.00003459, -0.00001950, 0.00000960, -0.00000880, -0.00002076, 0.00001149, 0.00000988, -0.00003773, -0.00005608, -0.00002895, 0.00000231, -0.00006485, -0.00025355, -0.00024649, 0.00014042, 0.00041751, 0.00028916, 0.00026176, 0.00042984, 0.00017934, -0.00031650, -0.00030466, -0.00000281, 0.00000632, -0.00006555, -0.00000533, -0.00002848, -0.00005399, 0.00002131, -0.00002240, -0.00007345, 0.00005991, 0.00005852, -0.00010181, -0.00001311, 0.00011047, -0.00006346, -0.00010970, 0.00010141, 0.00005662, -0.00012503, -0.00000087, 0.00010657, -0.00005702, -0.00008021, 0.00007055, 0.00002826, -0.00006919, 0.00000761, 0.00004504, -0.00002943, -0.00002270, 0.00002675, 0.00000448, -0.00001420, 0.00000130, -0.00000010, -0.00000239, -0.00000408, -0.00001730, -0.00000159, 0.00002198, 0.00000084, -0.00001300, 0.00001019, 0.00000558, -0.00001828, -0.00000354, 0.00001447, -0.00000069, -0.00001106, -0.00000470, -0.00000244, 0.00000112, 0.00000565, 0.00000087, -0.00000304, -0.00000356, -0.00000537, 0.00000262, 0.00001156, 0.00000139, -0.00000853, -0.00000259, -0.00000109, -0.00000063, 0.00001196, 0.00000927, -0.00001111, -0.00000729, 0.00001023, 0.00000108, -0.00001045, 0.00000343, 0.00000900, -0.00000693, -0.00001120, -0.00000106, 0.00000133, -0.00000143, -0.00000198, -0.00000417, -0.00000483, -0.00000190, -0.00000160, },
		{ -0.00000083, -0.00000115, -0.00000139, 0.00000055, 0.00000029, -0.00000525, -0.00000927, -0.00000408, 0.00000233, 0.00000027, -0.00000166, 0.00000160, -0.00000069, -0.00000809, -0.00000976, -0.00000649, -0.00000375, -0.00000307, -0.00000110, 0.00000426, -0.00000476, -0.00002960, -0.00002432, 0.00000740, -0.00000147, -0.00002494, 0.00000437, 0.00001968, -0.00003403, -0.00006126, -0.00002026, 0.00000168, -0.00006689, -0.00021247, -0.00023366, 0.00007274, 0.00039202, 0.00029788, 0.00014604, 0.00034493, 0.00035821, -0.00013122, -0.00038211, -0.00009916, 0.00005804, -0.00006097, -0.00005331, 0.00000706, -0.00003955, -0.00002741, -0.00000098, -0.00004375, 0.00001083, 0.00006022, -0.00005408, -0.00004999, 0.00008740, -0.00000449, -0.00012040, 0.00004582, 0.00009749, -0.00009410, -0.00005628, 0.00010787, -0.00000849, -0.00010776, 0.00004491, 0.00006723, -0.00006965, -0.00002539, 0.00006405, -0.00001390, -0.00004850, 0.00002778, 0.00002388, -0.00002473, -0.00000520, 0.00001143, -0.00000540, -0.00000790, -0.00001031, -0.00000742, 0.00001604, 0.00001078, -0.00001287, 0.00000202, 0.00001245, -0.00001402, -0.00001339, 0.00001408, 0.00000742, -0.00001180, -0.00000655, -0.00000145, -0.00000321, 0.00000346, 0.00000572, -0.00000078, -0.00000404, -0.00000592, -0.00000192, 0.00000969, 0.00000676, -0.00000700, -0.00000483, 0.00000094, -0.00000299, 0.00000520, 0.00001449, -0.00000294, -0.00001414, 0.00000483, 0.00000979, -0.00000834, -0.00000508, 0.00001036, 0.00000076, -0.00001327, -0.00000636, 0.00000193, 0.00000028, -0.00000120, -0.00000311, -0.00000572, -0.00000330, -0.00000100, -0.00000287, -0.00000285, },
		{ -0.00000112, -0.00000010, -0.00000034, -0.00000478, -0.00000840, -0.00000498, 0.00000132, 0.00000134, -0.00000132, 0.00000062, -0.00000002, -0.00000667, -0.00000905, -0.00000660, -0.00000569, -0.00000350, 0.00000110, 0.00000300, -0.00000552, -0.00002477, -0.00002570, 0.00000186, 0.00000436, -0.00002249, -0.00000597, 0.00002289, -0.00002239, -0.00006601, -0.00002105, 0.00000988, -0.00006738, -0.00018991, -0.00020700, 0.00002385, 0.00033761, 0.00032139, 0.00010053, 0.00021215, 0.00041568, 0.00009145, -0.00034877, -0.00022508, 0.00005473, -0.00001402, -0.00009375, 0.00000145, -0.00000115, -0.00005549, -0.00001637, -0.00000702, -0.00000862, 0.00003411, -0.00001448, -0.00005487, 0.00004702, 0.00003411, -0.00009774, -0.00001030, 0.00010534, -0.00004248, -0.00009291, 0.00007890, 0.00004143, -0.00010960, 0.00000131, 0.00009499, -0.00004620, -0.00006359, 0.00006500, 0.00001800, -0.00006844, 0.00001067, 0.00004871, -0.00002398, -0.00002282, 0.00002156, 0.00000101, -0.00001779, -0.00000583, -0.00000465, 0.00000563, 0.00001481, -0.00000498, -0.00000604, 0.00001227, -0.00000481, -0.00001928, 0.00000724, 0.00001490, -0.00000855, -0.00000998, -0.00000003, -0.00000495, -0.00000197, 0.00000840, 0.00000412, -0.00000427, -0.00000633, -0.00000447, 0.00000536, 0.00000966, -0.00000258, -0.00000685, 0.00000126, -0.00000181, -0.00000170, 0.00001328, 0.00000741, -0.00001440, -0.00000514, 0.00001373, 0.00000003, -0.00001140, 0.00000526, 0.00000844, -0.00001021, -0.00001236, -0.00000002, 0.00000240, -0.00000043, -0.00000195, -0.00000543, -0.00000518, -0.00000119, -0.00000179, -0.00000347, -0.00000133, -0.00000035, },
		{ -0.00000088, -0.00000465, -0.00000783, -0.00000521, 0.00000028, 0.00000166, -0.00000055, -0.00000013, -0.00000000, -0.00000513, -0.00000825, -0.00000647, -0.00000665, -0.00000502, 0.00000220, 0.00000380, -0.00000677, -0.00002201, -0.00002416, -0.00000373, 0.00000594, -0.00001577, -0.00001348, 0.00001809, -0.00000915, -0.00006365, -0.00002967, 0.00001742, -0.00005885, -0.00018017, -0.00018311, 0.00000114, 0.00027208, 0.00032583, 0.00011231, 0.00009836, 0.00036330, 0.00027075, -0.00021085, -0.00031488, -0.00001547, 0.00003754, -0.00009745, -0.00003576, 0.00002968, -0.00004941, -0.00004957, 0.00001152, 0.00000184, 0.00000276, 0.00000074, -0.00003620, 0.00001114, 0.00004338, -0.00005736, -0.00004648, 0.00008199, 0.00000865, -0.00009888, 0.00003163, 0.00007421, -0.00008409, -0.00004469, 0.00009899, -0.00000457, -0.00009116, 0.00004244, 0.00005516, -0.00007008, -0.00002235, 0.00006658, -0.00000619, -0.00004502, 0.00002117, 0.00001789, -0.00002683, -0.00001037, 0.00000532, -0.00000157, 0.00000903, 0.00000599, -0.00000751, 0.00000494, 0.00000348, -0.00001755, -0.00000296, 0.00001743, -0.00000122, -0.00001270, -0.00000045, -0.00000352, -0.00000794, 0.00000674, 0.00001035, -0.00000244, -0.00000754, -0.00000528, 0.00000101, 0.00000876, 0.00000278, -0.00000638, -0.00000056, 0.00000069, -0.00000537, 0.00000691, 0.00001471, -0.00000738, -0.00001429, 0.00000995, 0.00001054, -0.00001112, -0.00000410, 0.00001165, -0.00000244, -0.00001600, -0.00000503, 0.00000392, 0.00000096, -0.00000132, -0.00000411, -0.00000643, -0.00000266, -0.00000058, -0.00000324, -0.00000252, 0.00000004, -0.00000037, -0.00000055, },
		{ -0.00000762, -0.00000526, -0.00000027, 0.00000147, 0.00000005, -0.00000037, -0.00000038, -0.00000391, -0.00000716, -0.00000631, -0.00000688, -0.00000652, 0.00000172, 0.00000539, -0.00000684, -0.00002137, -0.00002182, -0.00000697, 0.00000383, -0.00000905, -0.00001519, 0.00000893, -0.00000071, -0.00005357, -0.00003926, 0.00001743, -0.00004292, -0.00017219, -0.00017117, -0.00000334, 0.00021658, 0.00030164, 0.00014635, 0.00004002, 0.00025203, 0.00035494, -0.00002024, -0.00031882, -0.00012408, 0.00005213, -0.00006239, -0.00007587, 0.00003104, -0.00001863, -0.00007446, 0.00000211, 0.00002567, -0.00001330, -0.00000836, -0.00001249, -0.00000555, 0.00002956, -0.00001981, -0.00005389, 0.00004259, 0.00004035, -0.00007645, -0.00001448, 0.00007920, -0.00004189, -0.00007588, 0.00007692, 0.00004041, -0.00009553, 0.00000199, 0.00008184, -0.00004833, -0.00005983, 0.00006582, 0.00002583, -0.00006014, 0.00000502, 0.00003911, -0.00002619, -0.00002532, 0.00001530, 0.00000133, -0.00000426, 0.00001201, 0.00000057, -0.00000436, 0.00000480, -0.00000907, -0.00001063, 0.00001276, 0.00000714, -0.00001175, -0.00000337, -0.00000070, -0.00001134, 0.00000083, 0.00001492, 0.00000257, -0.00000888, -0.00000585, -0.00000147, 0.00000494, 0.00000623, -0.00000272, -0.00000284, 0.00000184, -0.00000501, -0.00000076, 0.00001554, 0.00000350, -0.00001734, -0.00000010, 0.00001722, -0.00000332, -0.00001253, 0.00000783, 0.00000662, -0.00001449, -0.00001180, 0.00000320, 0.00000330, -0.00000110, -0.00000273, -0.00000614, -0.00000488, -0.00000038, -0.00000187, -0.00000361, -0.00000059, 0.00000063, -0.00000052, -0.00000142, -0.00000458, },
		{ -0.00000045, 0.00000133, 0.00000018, -0.00000034, -0.00000069, -0.00000320, -0.00000595, -0.00000599, -0.00000680, -0.00000741, 0.00000022, 0.00000641, -0.00000534, -0.00002150, -0.00002055, -0.00000763, 0.00000057, -0.00000537, -0.00001223, 0.00000073, 0.00000046, -0.00004033, -0.00004340, 0.00000963, -0.00002680, -0.00015801, -0.00016955, -0.00000382, 0.00018188, 0.00026016, 0.00017008, 0.00003625, 0.00014130, 0.00033892, 0.00015352, -0.00022768, -0.00021811, 0.00000992, -0.00001377, -0.00009086, 0.00000294, 0.00001309, -0.00007421, -0.00002602, 0.00004145, -0.00000689, -0.00002799, -0.00000065, -0.00000205, 0.00000845, 0.00000027, -0.00003817, 0.00000689, 0.00004483, -0.00004009, -0.00004229, 0.00005871, -0.00000058, -0.00008210, 0.00003823, 0.00007198, -0.00007412, -0.00004169, 0.00008570, -0.00000911, -0.00008621, 0.00004227, 0.00006067, -0.00005760, -0.00002401, 0.00005378, -0.00001070, -0.00004491, 0.00001613, 0.00001567, -0.00001701, 0.00000658, 0.00001490, -0.00000785, -0.00000295, 0.00000053, -0.00001054, 0.00000281, 0.00001163, -0.00000582, -0.00000688, 0.00000060, -0.00001079, -0.00000667, 0.00001500, 0.00000983, -0.00000839, -0.00000746, -0.00000196, 0.00000060, 0.00000591, 0.00000261, -0.00000307, 0.00000054, -0.00000252, -0.00000574, 0.00001020, 0.00001266, -0.00001229, -0.00001114, 0.00001586, 0.00000841, -0.00001468, -0.00000169, 0.00001205, -0.00000737, -0.00001720, -0.00000097, 0.00000582, -0.00000035, -0.00000226, -0.00000445, -0.00000645, -0.00000189, -0.00000016, -0.00000353, -0.00000218, 0.00000112, 0.00000023, -0.00000185, -0.00000460, -0.00000752, -0.00000553, },
		{ 0.00000008, -0.00000049, -0.00000073, -0.00000282, -0.00000496, -0.00000541, -0.00000657, -0.00000773, -0.00000148, 0.00000625, -0.00000304, -0.00002098, -0.00002083, -0.00000715, -0.00000149, -0.00000505, -0.00000803, -0.00000314, -0.00000344, -0.00002972, -0.00004023, -0.00000094, -0.00001703, -0.00013701, -0.00017006, -0.00001061, 0.00016491, 0.00022003, 0.00016917, 0.00006132, 0.00007111, 0.00025739, 0.00025650, -0.00007514, -0.00024871, -0.00007478, 0.00001259, -0.00007158, -0.00003487, 0.00002412, -0.00005039, -0.00005280, 0.00003642, 0.00001363, -0.00004081, -0.00000654, 0.00001087, -0.00000480, 0.00000057, -0.00001431, -0.00001098, 0.00002798, -0.00000825, -0.00004512, 0.00002611, 0.00002428, -0.00006505, -0.00000083, 0.00007984, -0.00003591, -0.00007203, 0.00006441, 0.00003288, -0.00008972, 0.00000249, 0.00008355, -0.00003415, -0.00005518, 0.00005204, 0.00001703, -0.00005860, 0.00000244, 0.00003541, -0.00001964, -0.00001071, 0.00002690, 0.00000008, -0.00001582, 0.00000376, -0.00000162, -0.00000655, 0.00000842, 0.00000294, -0.00000740, -0.00000131, -0.00000773, -0.00001203, 0.00000991, 0.00001633, -0.00000440, -0.00000974, -0.00000194, -0.00000201, 0.00000214, 0.00000670, 0.00000011, -0.00000193, -0.00000086, -0.00000636, 0.00000229, 0.00001565, -0.00000171, -0.00001703, 0.00000667, 0.00001773, -0.00000833, -0.00001186, 0.00001023, 0.00000249, -0.00001780, -0.00000791, 0.00000666, 0.00000176, -0.00000271, -0.00000257, -0.00000615, -0.00000456, 0.00000043, -0.00000187, -0.00000373, 0.00000031, 0.00000139, -0.00000173, -0.00000481, -0.00000737, -0.00000601, -0.00000069, 0.00000157, },
		{ -0.00000073, -0.00000239, -0.00000433, -0.00000475, -0.00000613, -0.00000771, -0.00000286, 0.00000517, -0.00000100, -0.00001927, -0.00002186, -0.00000712, -0.00000161, -0.00000629, -0.00000559, -0.00000277, -0.00000799, -0.00002477, -0.00003274, -0.00000842, -0.00001523, -0.00011437, -0.00016520, -0.00002505, 0.00015463, 0.00019388, 0.00014886, 0.00008545, 0.00005001, 0.00016102, 0.00027170, 0.00008132, -0.00019655, -0.00016106, -0.00000692, -0.00003317, -0.00005743, 0.00000861, -0.00001988, -0.00006123, 0.00001327, 0.00003173, -0.00003706, -0.00002334, 0.00001921, -0.00000407, -0.00001024, 0.00000274, -0.00000865, 0.00000450, 0.00000625, -0.00002900, -0.00000180, 0.00002680, -0.00003670, -0.00002558, 0.00006500, 0.00000326, -0.00007883, 0.00002704, 0.00006140, -0.00006868, -0.00003880, 0.00008388, 0.00000334, -0.00007483, 0.00003104, 0.00004690, -0.00005684, -0.00002342, 0.00004976, -0.00000683, -0.00003289, 0.00002730, 0.00001870, -0.00002526, -0.00000409, 0.00001150, -0.00000852, -0.00000232, 0.00000943, -0.00000222, -0.00000525, -0.00000555, -0.00001288, 0.00000199, 0.00001856, 0.00000276, -0.00001071, -0.00000301, -0.00000232, -0.00000279, 0.00000714, 0.00000573, -0.00000292, -0.00000183, -0.00000400, -0.00000371, 0.00001165, 0.00000878, -0.00001438, -0.00000552, 0.00001915, 0.00000381, -0.00001651, 0.00000140, 0.00000992, -0.00001215, -0.00001470, 0.00000398, 0.00000488, -0.00000296, -0.00000187, -0.00000395, -0.00000674, -0.00000112, 0.00000048, -0.00000387, -0.00000179, 0.00000203, -0.00000070, -0.00000506, -0.00000734, -0.00000639, -0.00000126, 0.00000201, 0.00000024, -0.00000098, },
		{ -0.00000387, -0.00000431, -0.00000547, -0.00000743, -0.00000386, 0.00000372, 0.00000021, -0.00001676, -0.00002254, -0.00000825, -0.00000061, -0.00000702, -0.00000579, -0.00000071, -0.00000982, -0.00002453, -0.00002580, -0.00001017, -0.00001799, -0.00009603, -0.00015297, -0.00004164, 0.00014103, 0.00018262, 0.00012428, 0.00009133, 0.00006109, 0.00009041, 0.00021975, 0.00018920, -0.00008039, -0.00020379, -0.00006858, -0.00000499, -0.00005152, -0.00002112, -0.00000240, -0.00004789, -0.00001338, 0.00003430, -0.00001964, -0.00003785, 0.00001498, 0.00000604, -0.00001981, 0.00000593, 0.00000552, -0.00001105, 0.00000185, -0.00000781, -0.00001367, 0.00001259, -0.00001224, -0.00003012, 0.00003838, 0.00002888, -0.00006304, -0.00001069, 0.00006678, -0.00003252, -0.00006598, 0.00006100, 0.00004071, -0.00007359, -0.00000266, 0.00006613, -0.00003679, -0.00005168, 0.00004905, 0.00001870, -0.00004865, 0.00001173, 0.00004032, -0.00002289, -0.00002144, 0.00002045, 0.00000050, -0.00001546, 0.00000821, 0.00000797, -0.00000738, -0.00000716, -0.00000994, -0.00000453, 0.00001488, 0.00001025, -0.00000817, -0.00000531, -0.00000163, -0.00000630, 0.00000365, 0.00001096, -0.00000047, -0.00000472, -0.00000177, -0.00000519, 0.00000372, 0.00001376, -0.00000459, -0.00001395, 0.00001144, 0.00001531, -0.00001209, -0.00000983, 0.00001031, -0.00000233, -0.00001742, -0.00000233, 0.00000712, -0.00000166, -0.00000268, -0.00000122, -0.00000678, -0.00000437, 0.00000176, -0.00000196, -0.00000406, 0.00000111, 0.00000096, -0.00000475, -0.00000765, -0.00000658, -0.00000201, 0.00000218, 0.00000089, -0.00000151, -0.00000106, -0.00000180, },
		{ -0.00000482, -0.00000681, -0.00000455, 0.00000227, 0.00000067, -0.00001420, -0.00002232, -0.00001013, 0.00000022, -0.00000626, -0.00000752, 0.00000034, -0.00000831, -0.00002558, -0.00002259, -0.00000777, -0.00002033, -0.00008448, -0.00013678, -0.00005376, 0.00012084, 0.00017856, 0.00010832, 0.00007892, 0.00007866, 0.00006098, 0.00014188, 0.00022361, 0.00005436, -0.00017769, -0.00014254, -0.00001159, -0.00002438, -0.00004398, -0.00000737, -0.00002387, -0.00002771, 0.00001984, -0.00000071, -0.00003998, 0.00000045, 0.00001525, -0.00002030, -0.00000174, 0.00001922, -0.00001185, -0.00001286, 0.00000505, -0.00000862, -0.00000567, -0.00000190, -0.00001921, 0.00001425, 0.00003477, -0.00003537, -0.00003450, 0.00005038, 0.00000320, -0.00007026, 0.00002477, 0.00006356, -0.00005159, -0.00003591, 0.00006607, -0.00000459, -0.00006997, 0.00003039, 0.00004687, -0.00004844, -0.00001599, 0.00005380, -0.00000558, -0.00004043, 0.00001771, 0.00001806, -0.00002314, -0.00000289, 0.00001855, -0.00000353, -0.00001263, -0.00000678, -0.00000610, 0.00000696, 0.00001400, -0.00000164, -0.00000700, -0.00000189, -0.00000725, -0.00000182, 0.00001292, 0.00000519, -0.00000695, -0.00000215, -0.00000279, -0.00000335, 0.00001115, 0.00000694, -0.00001382, -0.00000138, 0.00001956, -0.00000010, -0.00001677, 0.00000265, 0.00000653, -0.00001360, -0.00000977, 0.00000612, 0.00000138, -0.00000386, 0.00000023, -0.00000434, -0.00000748, 0.00000053, 0.00000116, -0.00000483, -0.00000151, 0.00000218, -0.00000334, -0.00000815, -0.00000687, -0.00000256, 0.00000185, 0.00000170, -0.00000157, -0.00000180, -0.00000133, -0.00000325, -0.00000419, },
		{ -0.00000487, 0.00000091, 0.00000069, -0.00001201, -0.00002136, -0.00001201, 0.00000019, -0.00000439, -0.00000906, -0.00000078, -0.00000524, -0.00002485, -0.00002301, -0.00000500, -0.00001930, -0.00007788, -0.00012176, -0.00005856, 0.00009736, 0.00017274, 0.00010471, 0.00005938, 0.00008423, 0.00006382, 0.00007685, 0.00019347, 0.00015921, -0.00008971, -0.00018913, -0.00005678, 0.00000123, -0.00004498, -0.00002843, -0.00000620, -0.00002319, -0.00000161, 0.00000742, -0.00002924, -0.00001486, 0.00001550, -0.00001233, -0.00001148, 0.00002386, -0.00000087, -0.00002569, 0.00000382, 0.00000500, -0.00001604, -0.00000637, -0.00000418, 0.00000252, 0.00002500, -0.00000996, -0.00003842, 0.00002314, 0.00002477, -0.00005379, -0.00000963, 0.00006448, -0.00001851, -0.00005569, 0.00004695, 0.00002744, -0.00006994, -0.00000023, 0.00006526, -0.00002986, -0.00004514, 0.00005053, 0.00002187, -0.00005067, 0.00000121, 0.00003622, -0.00001899, -0.00002067, 0.00002293, 0.00000770, -0.00001850, -0.00000763, -0.00000222, -0.00000074, 0.00001121, 0.00000653, -0.00000539, -0.00000388, -0.00000665, -0.00000638, 0.00001058, 0.00001148, -0.00000594, -0.00000520, 0.00000041, -0.00000598, 0.00000317, 0.00001404, -0.00000512, -0.00001224, 0.00001374, 0.00001330, -0.00001449, -0.00000925, 0.00000897, -0.00000434, -0.00001418, 0.00000091, 0.00000448, -0.00000359, -0.00000037, -0.00000082, -0.00000838, -0.00000305, 0.00000346, -0.00000307, -0.00000470, 0.00000167, -0.00000098, -0.00000814, -0.00000763, -0.00000278, 0.00000129, 0.00000212, -0.00000100, -0.00000252, -0.00000136, -0.00000244, -0.00000415, -0.00000448, -0.00000589, },
		{ 0.00000046, -0.00001021, -0.00002011, -0.00001343, -0.00000061, -0.00000233, -0.00000941, -0.00000345, -0.00000291, -0.00002141, -0.00002456, -0.00000477, -0.00001543, -0.00007233, -0.00011085, -0.00005788, 0.00007609, 0.00016083, 0.00010876, 0.00004474, 0.00007366, 0.00007809, 0.00004456, 0.00013054, 0.00020487, 0.00002606, -0.00018045, -0.00012032, 0.00000188, -0.00002532, -0.00004884, -0.00000610, -0.00000610, -0.00001613, 0.00000059, -0.00001408, -0.00002154, 0.00000624, -0.00000290, -0.00001557, 0.00001875, 0.00001293, -0.00002823, -0.00000795, 0.00001583, -0.00001373, -0.00001805, 0.00000423, 0.00000429, 0.00001018, 0.00000319, -0.00002664, -0.00000092, 0.00002691, -0.00002769, -0.00002944, 0.00004649, 0.00001148, -0.00005572, 0.00001777, 0.00004714, -0.00005169, -0.00003086, 0.00006564, 0.00000064, -0.00006335, 0.00002938, 0.00004864, -0.00004490, -0.00002380, 0.00004541, -0.00000205, -0.00003743, 0.00001628, 0.00002327, -0.00001953, -0.00001440, 0.00000393, -0.00000343, 0.00000239, 0.00001196, 0.00000100, -0.00000617, -0.00000672, -0.00000824, 0.00000543, 0.00001513, -0.00000115, -0.00000858, 0.00000130, -0.00000391, -0.00000528, 0.00001327, 0.00000737, -0.00001502, 0.00000067, 0.00002062, -0.00000305, -0.00001846, 0.00000265, 0.00000561, -0.00001215, -0.00000669, 0.00000491, -0.00000104, -0.00000203, 0.00000160, -0.00000625, -0.00000708, 0.00000304, 0.00000055, -0.00000646, -0.00000101, 0.00000118, -0.00000685, -0.00000880, -0.00000305, 0.00000102, 0.00000189, -0.00000022, -0.00000264, -0.00000189, -0.00000177, -0.00000386, -0.00000454, -0.00000496, -0.00000469, -0.00000036, },
		{ -0.00001880, -0.00001444, -0.00000173, -0.00000070, -0.00000859, -0.00000635, -0.00000253, -0.00001646, -0.00002462, -0.00000741, -0.00001146, -0.00006511, -0.00010341, -0.00005573, 0.00006034, 0.00014424, 0.00011297, 0.00004051, 0.00005427, 0.00008561, 0.00004282, 0.00007008, 0.00019098, 0.00012677, -0.00011465, -0.00016886, -0.00003113, -0.00000164, -0.00005349, -0.00002189, 0.00000965, -0.00001545, -0.00001499, -0.00000544, -0.00001692, -0.00000601, 0.00000069, -0.00001228, 0.00000951, 0.00002115, -0.00001992, -0.00002109, 0.00001652, -0.00000252, -0.00002679, 0.00000173, 0.00001295, 0.00000061, 0.00000258, -0.00000979, -0.00001210, 0.00001435, -0.00000562, -0.00003043, 0.00002075, 0.00002711, -0.00003935, -0.00000857, 0.00004871, -0.00002369, -0.00004975, 0.00004822, 0.00003094, -0.00006296, -0.00000225, 0.00006323, -0.00002326, -0.00004699, 0.00003955, 0.00002207, -0.00004469, -0.00000129, 0.00003656, -0.00001189, -0.00002503, 0.00000718, 0.00000121, -0.00000836, 0.00001054, 0.00001077, -0.00000574, -0.00000903, -0.00000787, 0.00000054, 0.00001428, 0.00000534, -0.00000923, -0.00000105, 0.00000010, -0.00000978, 0.00000573, 0.00001695, -0.00000810, -0.00001263, 0.00001741, 0.00001207, -0.00001881, -0.00000972, 0.00001017, -0.00000362, -0.00001252, 0.00000087, 0.00000254, -0.00000263, 0.00000157, -0.00000231, -0.00000904, -0.00000038, 0.00000385, -0.00000534, -0.00000479, 0.00000158, -0.00000423, -0.00000963, -0.00000393, 0.00000125, 0.00000135, 0.00000008, -0.00000199, -0.00000240, -0.00000167, -0.00000330, -0.00000471, -0.00000440, -0.00000403, -0.00000133, -0.00000005, -0.00000863, },
		{ -0.00000291, 0.00000044, -0.00000721, -0.00000850, -0.00000384, -0.00001187, -0.00002218, -0.00001109, -0.00000982, -0.00005619, -0.00009675, -0.00005497, 0.00004959, 0.00012718, 0.00011251, 0.00004457, 0.00003657, 0.00007938, 0.00005585, 0.00003428, 0.00014050, 0.00018145, -0.00001579, -0.00017429, -0.00008542, 0.00000508, -0.00003942, -0.00004106, 0.00001243, -0.00000205, -0.00002771, -0.00000899, -0.00000640, -0.00001303, -0.00000378, -0.00000591, 0.00000299, 0.00002089, -0.00000668, -0.00002727, 0.00000729, 0.00000895, -0.00002632, -0.00000821, 0.00001949, 0.00000072, -0.00000571, 0.00000170, -0.00000921, -0.00000213, 0.00000348, -0.00001793, 0.00000014, 0.00002537, -0.00001711, -0.00002210, 0.00003525, 0.00000200, -0.00005131, 0.00002131, 0.00004948, -0.00004491, -0.00003209, 0.00005940, 0.00000641, -0.00005819, 0.00001961, 0.00004362, -0.00003768, -0.00002405, 0.00004069, 0.00000408, -0.00003423, 0.00000343, 0.00001136, -0.00001525, 0.00000107, 0.00001972, -0.00000001, -0.00001297, -0.00000760, -0.00000151, 0.00000974, 0.00000999, -0.00000571, -0.00000477, 0.00000244, -0.00000894, -0.00000389, 0.00001885, 0.00000473, -0.00001892, 0.00000498, 0.00002284, -0.00000873, -0.00002135, 0.00000565, 0.00000731, -0.00001239, -0.00000668, 0.00000427, -0.00000065, -0.00000028, 0.00000093, -0.00000755, -0.00000501, 0.00000443, -0.00000166, -0.00000747, -0.00000058, -0.00000141, -0.00000913, -0.00000551, 0.00000158, 0.00000119, -0.00000045, -0.00000106, -0.00000221, -0.00000215, -0.00000287, -0.00000462, -0.00000431, -0.00000323, -0.00000174, -0.00000079, -0.00000730, -0.00001736, -0.00001525, },
		{ -0.00000569, -0.00000975, -0.00000586, -0.00000883, -0.00001804, -0.00001361, -0.00001095, -0.00004757, -0.00008867, -0.00005569, 0.00004107, 0.00011288, 0.00010726, 0.00005106, 0.00002693, 0.00006325, 0.00006686, 0.00002632, 0.00008409, 0.00018287, 0.00007969, -0.00012869, -0.00013404, -0.00001657, -0.00001791, -0.00004959, 0.00000053, 0.00001335, -0.00002870, -0.00002162, 0.00000083, -0.00001090, -0.00001231, -0.00000263, 0.00000223, 0.00001521, 0.00000402, -0.00002373, -0.00000551, 0.00001322, -0.00001736, -0.00001789, 0.00001830, 0.00000779, -0.00001306, 0.00000313, 0.00000144, -0.00001226, -0.00000086, -0.00000270, -0.00000748, 0.00001208, -0.00000056, -0.00002059, 0.00001633, 0.00001611, -0.00003841, -0.00000340, 0.00005098, -0.00001790, -0.00004916, 0.00003933, 0.00003246, -0.00005266, -0.00000671, 0.00005370, -0.00001811, -0.00004322, 0.00003224, 0.00002344, -0.00003631, -0.00000773, 0.00002186, -0.00001389, -0.00001362, 0.00002272, 0.00001114, -0.00001565, -0.00000992, -0.00000031, 0.00000458, 0.00001004, 0.00000062, -0.00000648, 0.00000118, -0.00000495, -0.00001042, 0.00001282, 0.00001659, -0.00001487, -0.00001018, 0.00002322, 0.00000737, -0.00002504, -0.00000657, 0.00001429, -0.00000501, -0.00001390, 0.00000160, 0.00000329, -0.00000167, 0.00000152, -0.00000349, -0.00000795, 0.00000145, 0.00000237, -0.00000710, -0.00000446, -0.00000017, -0.00000696, -0.00000706, 0.00000127, 0.00000177, -0.00000145, -0.00000070, -0.00000106, -0.00000263, -0.00000307, -0.00000423, -0.00000440, -0.00000274, -0.00000156, -0.00000149, -0.00000641, -0.00001572, -0.00001585, -0.00000422, 0.00000127, },
		{ -0.00000779, -0.00000729, -0.00001372, -0.00001402, -0.00001344, -0.00004116, -0.00007905, -0.00005613, 0.00003249, 0.00010172, 0.00010000, 0.00005528, 0.00002535, 0.00004571, 0.00006710, 0.00003531, 0.00004456, 0.00014614, 0.00014117, -0.00004760, -0.00015038, -0.00006038, -0.00000686, -0.00004190, -0.00001735, 0.00001966, -0.00001762, -0.00003410, -0.00000153, -0.00000261, -0.00001806, -0.00000552, 0.00000509, 0.00000968, 0.00000802, -0.00001387, -0.00001429, 0.00000841, -0.00000606, -0.00002092, 0.00001008, 0.00001509, -0.00001385, -0.00000316, 0.00001100, -0.00001171, -0.00001187, 0.00000564, -0.00000256, -0.00000241, 0.00000385, -0.00000965, 0.00000220, 0.00001650, -0.00001999, -0.00001699, 0.00003843, 0.00000644, -0.00004908, 0.00001221, 0.00004531, -0.00003347, -0.00002890, 0.00004869, 0.00000659, -0.00005123, 0.00001342, 0.00003890, -0.00002844, -0.00002254, 0.00002693, -0.00000370, -0.00002768, 0.00001652, 0.00002453, -0.00001358, -0.00001549, 0.00000208, 0.00000223, 0.00000528, 0.00000617, -0.00000377, -0.00000253, -0.00000178, -0.00001126, 0.00000296, 0.00002147, -0.00000291, -0.00001999, 0.00001277, 0.00002132, -0.00001755, -0.00002024, 0.00001239, 0.00000649, -0.00001602, -0.00000552, 0.00000641, -0.00000043, -0.00000044, 0.00000051, -0.00000720, -0.00000358, 0.00000390, -0.00000352, -0.00000771, -0.00000174, -0.00000411, -0.00000741, -0.00000001, 0.00000273, -0.00000208, -0.00000142, 0.00000053, -0.00000228, -0.00000396, -0.00000396, -0.00000423, -0.00000274, -0.00000110, -0.00000173, -0.00000601, -0.00001405, -0.00001601, -0.00000574, 0.00000181, -0.00000412, -0.00001040, },
		{ -0.00001023, -0.00001288, -0.00001549, -0.00003737, -0.00006941, -0.00005471, 0.00002353, 0.00009206, 0.00009342, 0.00005602, 0.00002780, 0.00003329, 0.00005742, 0.00004665, 0.00002909, 0.00009682, 0.00015669, 0.00003910, -0.00012260, -0.00010541, -0.00001878, -0.00002485, -0.00002890, 0.00001296, -0.00000227, -0.00003776, -0.00001264, 0.00000449, -0.00001669, -0.00001250, 0.00000678, 0.00000794, 0.00000632, -0.00000390, -0.00001503, -0.00000150, 0.00000069, -0.00001607, 0.00000030, 0.00001697, -0.00000823, -0.00001078, 0.00001360, -0.00000317, -0.00002068, 0.00000351, 0.00000785, -0.00000944, -0.00000258, 0.00000139, -0.00000168, 0.00000820, -0.00000561, -0.00001736, 0.00002060, 0.00001957, -0.00003554, -0.00001072, 0.00004206, -0.00000958, -0.00003897, 0.00003195, 0.00002703, -0.00004557, -0.00000894, 0.00004445, -0.00001244, -0.00003497, 0.00002326, 0.00001166, -0.00003506, 0.00000173, 0.00003486, -0.00000481, -0.00002236, 0.00000249, 0.00000436, -0.00000166, 0.00000725, 0.00000308, -0.00000526, -0.00000240, -0.00000772, -0.00000498, 0.00001758, 0.00001029, -0.00001939, -0.00000297, 0.00002570, -0.00000183, -0.00002746, 0.00000144, 0.00001577, -0.00001056, -0.00001380, 0.00000552, 0.00000357, -0.00000273, 0.00000188, -0.00000310, -0.00000753, 0.00000144, 0.00000119, -0.00000793, -0.00000559, -0.00000252, -0.00000589, -0.00000165, 0.00000317, -0.00000176, -0.00000291, 0.00000153, -0.00000070, -0.00000493, -0.00000441, -0.00000371, -0.00000296, -0.00000089, -0.00000135, -0.00000580, -0.00001273, -0.00001558, -0.00000727, 0.00000181, -0.00000246, -0.00001063, -0.00000943, -0.00000659, },
		{ -0.00001630, -0.00003507, -0.00006127, -0.00005142, 0.00001537, 0.00008230, 0.00008824, 0.00005474, 0.00003019, 0.00002752, 0.00004407, 0.00005083, 0.00003128, 0.00005694, 0.00013465, 0.00010263, -0.00005936, -0.00012759, -0.00005208, -0.00001308, -0.00002716, -0.00000130, 0.00000738, -0.00003034, -0.00002551, 0.00000414, -0.00000938, -0.00001821, 0.00000399, 0.00000947, 0.00000305, 0.00000153, -0.00000902, -0.00001002, -0.00000029, -0.00000740, -0.00000513, 0.00001240, -0.00000082, -0.00001412, 0.00000905, 0.00000665, -0.00002168, -0.00000576, 0.00001522, -0.00000648, -0.00001284, 0.00000525, 0.00000322, -0.00000061, -0.00000039, -0.00000918, 0.00000674, 0.00001976, -0.00001748, -0.00002184, 0.00002729, 0.00000899, -0.00003529, 0.00001166, 0.00003655, -0.00002973, -0.00002674, 0.00003847, 0.00000612, -0.00003990, 0.00001172, 0.00002617, -0.00003241, -0.00001719, 0.00003723, 0.00000941, -0.00002686, -0.00000131, 0.00000995, -0.00000673, 0.00000225, 0.00001087, -0.00000357, -0.00000675, -0.00000382, -0.00000712, 0.00000806, 0.00001768, -0.00000915, -0.00001550, 0.00001814, 0.00001423, -0.00002335, -0.00001319, 0.00001690, 0.00000055, -0.00001821, -0.00000042, 0.00000788, -0.00000264, 0.00000006, 0.00000175, -0.00000776, -0.00000394, 0.00000384, -0.00000438, -0.00000942, -0.00000381, -0.00000324, -0.00000234, 0.00000251, -0.00000072, -0.00000421, 0.00000122, 0.00000161, -0.00000500, -0.00000575, -0.00000326, -0.00000291, -0.00000121, -0.00000063, -0.00000526, -0.00001199, -0.00001480, -0.00000833, 0.00000118, -0.00000094, -0.00001038, -0.00001088, -0.00000617, -0.00000760, -0.00001135, },
		{ -0.00005490, -0.00004753, 0.00000908, 0.00007229, 0.00008357, 0.00005331, 0.00003080, 0.00002609, 0.00003304, 0.00004636, 0.00003921, 0.00003635, 0.00009538, 0.00012817, 0.00001577, -0.00011301, -0.00009114, -0.00001904, -0.00001553, -0.00001285, 0.00000611, -0.00001709, -0.00003211, -0.00000441, -0.00000182, -0.00001837, -0.00000254, 0.00001083, 0.00000183, 0.00000184, -0.00000104, -0.00001254, -0.00000698, -0.00000089, -0.00000379, 0.00000497, 0.00000323, -0.00001176, 0.00000168, 0.00001184, -0.00001507, -0.00001522, 0.00001466, 0.00000258, -0.00001915, 0.00000044, 0.00001056, -0.00000367, -0.00000317, -0.00000008, 0.00000178, 0.00001174, -0.00000385, -0.00002035, 0.00000988, 0.00001620, -0.00002265, -0.00000367, 0.00003418, -0.00001107, -0.00003462, 0.00002427, 0.00002073, -0.00003569, -0.00000330, 0.00003442, -0.00002060, -0.00003387, 0.00002961, 0.00002494, -0.00002561, -0.00000912, 0.00001584, -0.00000672, -0.00000732, 0.00001522, 0.00000367, -0.00001179, -0.00000358, -0.00000347, -0.00000113, 0.00001574, 0.00000474, -0.00001814, 0.00000282, 0.00002214, -0.00000915, -0.00002331, 0.00000813, 0.00001137, -0.00001516, -0.00000896, 0.00000894, 0.00000078, -0.00000300, 0.00000427, -0.00000393, -0.00000909, 0.00000225, 0.00000118, -0.00001040, -0.00000794, -0.00000145, -0.00000110, 0.00000115, 0.00000006, -0.00000447, -0.00000022, 0.00000350, -0.00000356, -0.00000738, -0.00000360, -0.00000239, -0.00000184, -0.00000013, -0.00000407, -0.00001158, -0.00001424, -0.00000863, 0.00000025, -0.00000002, -0.00000963, -0.00001209, -0.00000596, -0.00000535, -0.00001020, -0.00001633, -0.00003282, },
		{ 0.00000447, 0.00006292, 0.00007852, 0.00005232, 0.00003011, 0.00002589, 0.00002671, 0.00003729, 0.00004358, 0.00003248, 0.00005906, 0.00011869, 0.00007689, -0.00006497, -0.00011505, -0.00004432, -0.00000546, -0.00001425, -0.00000328, -0.00000659, -0.00002898, -0.00001610, 0.00000016, -0.00001296, -0.00000879, 0.00000894, 0.00000321, -0.00000025, 0.00000435, -0.00000875, -0.00001422, -0.00000037, 0.00000204, -0.00000017, 0.00000185, -0.00000655, -0.00000345, 0.00001064, -0.00000556, -0.00001909, 0.00000709, 0.00001089, -0.00001762, -0.00000885, 0.00001379, 0.00000005, -0.00000849, 0.00000411, 0.00000467, 0.00000297, 0.00000087, -0.00001143, -0.00000202, 0.00001245, -0.00000906, -0.00000947, 0.00002439, 0.00000320, -0.00003217, 0.00000814, 0.00002720, -0.00002478, -0.00001631, 0.00003414, -0.00000404, -0.00004296, 0.00001389, 0.00003649, -0.00001750, -0.00001834, 0.00001859, -0.00000090, -0.00001753, 0.00001291, 0.00001439, -0.00001333, -0.00000846, 0.00000234, -0.00000475, 0.00000622, 0.00001453, -0.00000968, -0.00001184, 0.00001759, 0.00000806, -0.00002256, -0.00000646, 0.00001524, -0.00000527, -0.00001509, 0.00000472, 0.00000566, -0.00000425, 0.00000315, 0.00000160, -0.00001074, -0.00000314, 0.00000536, -0.00000714, -0.00001274, -0.00000252, 0.00000168, 0.00000043, -0.00000027, -0.00000372, -0.00000176, 0.00000399, -0.00000101, -0.00000817, -0.00000503, -0.00000181, -0.00000224, -0.00000025, -0.00000239, -0.00001091, -0.00001432, -0.00000839, -0.00000032, 0.00000006, -0.00000876, -0.00001278, -0.00000601, -0.00000319, -0.00000932, -0.00001654, -0.00003009, -0.00004928, -0.00004422, },
		{ 0.00007324, 0.00005133, 0.00002909, 0.00002524, 0.00002420, 0.00002849, 0.00004134, 0.00003635, 0.00003692, 0.00008952, 0.00010789, -0.00000133, -0.00010926, -0.00007786, -0.00000854, -0.00000634, -0.00001261, -0.00000463, -0.00001926, -0.00002384, -0.00000524, -0.00000620, -0.00001080, 0.00000370, 0.00000494, -0.00000154, 0.00000553, -0.00000197, -0.00001736, -0.00000540, 0.00000748, 0.00000001, -0.00000293, -0.00000286, -0.00000399, 0.00000561, 0.00000142, -0.00001595, -0.00000239, 0.00001322, -0.00000990, -0.00001625, 0.00001015, 0.00000675, -0.00001079, 0.00000225, 0.00001036, -0.00000094, -0.00000203, -0.00000245, -0.00000491, 0.00000322, -0.00000120, -0.00000675, 0.00001392, 0.00000950, -0.00002339, -0.00000405, 0.00002537, -0.00001199, -0.00002350, 0.00002672, 0.00001164, -0.00004229, -0.00000505, 0.00004011, -0.00000440, -0.00002529, 0.00001630, 0.00000874, -0.00002402, 0.00000362, 0.00002425, -0.00000860, -0.00001648, 0.00000551, -0.00000117, -0.00000522, 0.00001492, 0.00000511, -0.00001779, 0.00000296, 0.00001913, -0.00001029, -0.00001883, 0.00000884, 0.00000631, -0.00001433, -0.00000334, 0.00000830, -0.00000187, -0.00000016, 0.00000539, -0.00000780, -0.00000930, 0.00000529, -0.00000093, -0.00001507, -0.00000692, 0.00000409, 0.00000163, -0.00000159, -0.00000294, -0.00000233, 0.00000303, 0.00000143, -0.00000730, -0.00000703, -0.00000196, -0.00000208, -0.00000086, -0.00000083, -0.00000946, -0.00001491, -0.00000830, 0.00000001, -0.00000034, -0.00000841, -0.00001283, -0.00000616, -0.00000122, -0.00000826, -0.00001742, -0.00002750, -0.00004325, -0.00004147, 0.00000037, 0.00005486, },
		{ 0.00002799, 0.00002398, 0.00002351, 0.00002264, 0.00003454, 0.00003959, 0.00002943, 0.00005827, 0.00010768, 0.00005543, -0.00007322, -0.00010207, -0.00002834, 0.00000230, -0.00001464, -0.00001042, -0.00001001, -0.00002350, -0.00001457, -0.00000335, -0.00000778, -0.00000196, 0.00000437, -0.00000087, 0.00000400, 0.00000370, -0.00001499, -0.00001216, 0.00000855, 0.00000463, -0.00000691, -0.00000327, -0.00000134, 0.00000101, 0.00000319, -0.00000877, -0.00000834, 0.00000889, -0.00000124, -0.00001746, 0.00000182, 0.00001107, -0.00000790, -0.00000251, 0.00001368, 0.00000114, -0.00000748, 0.00000138, -0.00000081, -0.00000465, -0.00000115, -0.00000056, 0.00000794, 0.00000889, -0.00001388, -0.00000954, 0.00001849, -0.00000188, -0.00002414, 0.00001595, 0.00002216, -0.00003337, -0.00002157, 0.00003483, 0.00000968, -0.00002719, 0.00000949, 0.00001866, -0.00002400, -0.00000996, 0.00002877, 0.00000216, -0.00002342, 0.00000289, 0.00000715, -0.00001231, 0.00000575, 0.00001830, -0.00001169, -0.00001402, 0.00001793, 0.00000773, -0.00002142, -0.00000527, 0.00001254, -0.00000591, -0.00001058, 0.00000580, 0.00000300, -0.00000243, 0.00000534, -0.00000217, -0.00001251, 0.00000067, 0.00000479, -0.00001289, -0.00001279, 0.00000379, 0.00000476, -0.00000261, -0.00000321, -0.00000169, 0.00000162, 0.00000252, -0.00000495, -0.00000838, -0.00000333, -0.00000170, -0.00000139, 0.00000008, -0.00000727, -0.00001538, -0.00000899, 0.00000110, -0.00000040, -0.00000906, -0.00001264, -0.00000591, 0.00000035, -0.00000682, -0.00001861, -0.00002599, -0.00003664, -0.00003819, -0.00000414, 0.00004761, 0.00006857, 0.00005000, },
		{ 0.00002332, 0.00001981, 0.00002678, 0.00003837, 0.00003040, 0.00003644, 0.00008714, 0.00008919, -0.00001989, -0.00010278, -0.00005744, 0.00000140, -0.00000851, -0.00001776, -0.00000705, -0.00001655, -0.00002180, -0.00000680, -0.00000280, -0.00000456, 0.00000093, 0.00000041, 0.00000238, 0.00000610, -0.00000915, -0.00001634, 0.00000457, 0.00000996, -0.00000690, -0.00000696, 0.00000107, -0.00000039, 0.00000084, -0.00000221, -0.00000833, 0.00000136, 0.00000332, -0.00001259, -0.00000619, 0.00000983, -0.00000188, -0.00000571, 0.00001250, 0.00000632, -0.00001047, -0.00000068, 0.00000510, -0.00000691, -0.00000628, 0.00000356, 0.00000763, 0.00000549, -0.00000774, -0.00000931, 0.00001097, 0.00000338, -0.00002029, 0.00000585, 0.00002600, -0.00002014, -0.00003158, 0.00002288, 0.00002049, -0.00002341, 0.00000056, 0.00002553, -0.00001733, -0.00002341, 0.00002553, 0.00001573, -0.00002521, -0.00000533, 0.00001530, -0.00001148, -0.00000838, 0.00002306, 0.00000354, -0.00002442, 0.00000469, 0.00002227, -0.00001166, -0.00001971, 0.00000873, 0.00000610, -0.00001191, -0.00000150, 0.00000680, -0.00000112, 0.00000223, 0.00000252, -0.00001086, -0.00000582, 0.00000658, -0.00000685, -0.00001682, -0.00000024, 0.00000809, -0.00000182, -0.00000473, -0.00000069, 0.00000105, 0.00000192, -0.00000234, -0.00000803, -0.00000553, -0.00000192, -0.00000140, 0.00000035, -0.00000490, -0.00001503, -0.00001055, 0.00000220, 0.00000067, -0.00001041, -0.00001305, -0.00000496, 0.00000175, -0.00000531, -0.00001949, -0.00002591, -0.00003037, -0.00003328, -0.00000867, 0.00003990, 0.00006487, 0.00004888, 0.00002647, 0.00002232, },
		{ 0.00002042, 0.00003318, 0.00003290, 0.00002663, 0.00006097, 0.00009613, 0.00003182, -0.00007691, -0.00008142, -0.00001365, -0.00000040, -0.00002008, -0.00001100, -0.00000859, -0.00002256, -0.00001441, -0.00000039, -0.00000287, -0.00000330, 0.00000018, 0.00000203, 0.00000578, -0.00000327, -0.00001591, -0.00000183, 0.00001222, -0.00000279, -0.00001073, 0.00000059, 0.00000112, -0.00000235, 0.00000062, -0.00000383, -0.00000454, 0.00000201, -0.00000544, -0.00000959, 0.00000384, 0.00000306, -0.00000486, 0.00000838, 0.00001047, -0.00000893, -0.00000560, 0.00000812, -0.00000365, -0.00001168, 0.00000277, 0.00001043, 0.00000345, -0.00000577, -0.00000661, 0.00000589, 0.00000437, -0.00001504, -0.00000120, 0.00002426, -0.00000695, -0.00003379, 0.00000836, 0.00002526, -0.00001555, -0.00000753, 0.00002757, -0.00000619, -0.00003253, 0.00001506, 0.00002747, -0.00002002, -0.00001599, 0.00001879, -0.00000336, -0.00002085, 0.00001703, 0.00002058, -0.00002288, -0.00001418, 0.00002574, 0.00000639, -0.00002624, -0.00000428, 0.00001473, -0.00000523, -0.00000962, 0.00000599, 0.00000356, -0.00000079, 0.00000346, -0.00000558, -0.00001016, 0.00000329, -0.00000023, -0.00001633, -0.00000658, 0.00000906, 0.00000124, -0.00000638, -0.00000062, 0.00000188, 0.00000051, -0.00000095, -0.00000589, -0.00000735, -0.00000343, -0.00000109, 0.00000049, -0.00000302, -0.00001368, -0.00001236, 0.00000240, 0.00000295, -0.00001150, -0.00001476, -0.00000367, 0.00000362, -0.00000409, -0.00001983, -0.00002669, -0.00002561, -0.00002677, -0.00001180, 0.00003093, 0.00006126, 0.00004899, 0.00002461, 0.00001997, 0.00002324, 0.00001906, },
		{ 0.00003299, 0.00002513, 0.00004007, 0.00008337, 0.00006669, -0.00003349, -0.00008717, -0.00003815, 0.00000094, -0.00001548, -0.00001738, -0.00000508, -0.00001710, -0.00002115, -0.00000322, 0.00000098, -0.00000536, -0.00000235, 0.00000231, 0.00000469, 0.00000031, -0.00001187, -0.00000694, 0.00001007, 0.00000273, -0.00001155, -0.00000281, 0.00000307, -0.00000361, -0.00000035, 0.00000145, -0.00000586, -0.00000313, -0.00000061, -0.00000744, -0.00000309, 0.00000371, -0.00000081, 0.00000475, 0.00001118, -0.00000429, -0.00000933, 0.00000656, 0.00000178, -0.00001345, -0.00000199, 0.00001244, 0.00000452, -0.00000601, -0.00000450, 0.00000386, 0.00000332, -0.00001068, -0.00000484, 0.00001937, 0.00000316, -0.00002950, -0.00000459, 0.00002363, -0.00000649, -0.00001263, 0.00002492, 0.00000611, -0.00003492, 0.00000042, 0.00003347, -0.00000900, -0.00002473, 0.00001573, 0.00000815, -0.00002650, 0.00000294, 0.00003171, -0.00001008, -0.00002967, 0.00001619, 0.00002415, -0.00002053, -0.00002021, 0.00001428, 0.00000666, -0.00001321, -0.00000022, 0.00000845, -0.00000056, 0.00000060, -0.00000028, -0.00000963, -0.00000294, 0.00000305, -0.00001123, -0.00001190, 0.00000611, 0.00000513, -0.00000645, -0.00000201, 0.00000342, -0.00000024, -0.00000143, -0.00000309, -0.00000748, -0.00000604, -0.00000129, 0.00000107, -0.00000179, -0.00001181, -0.00001357, 0.00000136, 0.00000566, -0.00001125, -0.00001768, -0.00000304, 0.00000646, -0.00000300, -0.00002006, -0.00002746, -0.00002280, -0.00001994, -0.00001219, 0.00002136, 0.00005611, 0.00005054, 0.00002346, 0.00001644, 0.00002286, 0.00001973, 0.00001614, 0.00002626, },
		{ 0.00002830, 0.00006270, 0.00007927, 0.00001198, -0.00007015, -0.00006051, -0.00000948, -0.00000826, -0.00002046, -0.00000754, -0.00000971, -0.00002275, -0.00001012, 0.00000313, -0.00000383, -0.00000571, 0.00000158, 0.00000421, 0.00000144, -0.00000691, -0.00000842, 0.00000509, 0.00000631, -0.00000867, -0.00000681, 0.00000301, -0.00000247, -0.00000282, 0.00000437, -0.00000287, -0.00000813, -0.00000051, -0.00000225, -0.00000698, -0.00000009, 0.00000322, 0.00000412, 0.00000900, 0.00000029, -0.00000954, 0.00000212, 0.00000548, -0.00001078, -0.00000735, 0.00001108, 0.00000755, -0.00000576, -0.00000417, 0.00000353, 0.00000227, -0.00000791, -0.00000605, 0.00001364, 0.00000899, -0.00002146, -0.00001316, 0.00001726, 0.00000099, -0.00001408, 0.00001912, 0.00001650, -0.00003042, -0.00001423, 0.00003194, 0.00000443, -0.00002820, 0.00000733, 0.00001817, -0.00002374, -0.00001329, 0.00003259, 0.00000816, -0.00003493, -0.00000184, 0.00003318, -0.00000472, -0.00003073, 0.00000392, 0.00001735, -0.00000909, -0.00000874, 0.00000964, 0.00000362, -0.00000323, 0.00000160, -0.00000467, -0.00000799, 0.00000100, -0.00000443, -0.00001295, 0.00000008, 0.00000720, -0.00000408, -0.00000402, 0.00000429, 0.00000055, -0.00000310, -0.00000128, -0.00000553, -0.00000856, -0.00000283, 0.00000205, -0.00000077, -0.00001014, -0.00001374, -0.00000042, 0.00000767, -0.00000929, -0.00002075, -0.00000408, 0.00001007, -0.00000127, -0.00002073, -0.00002785, -0.00002125, -0.00001435, -0.00000987, 0.00001309, 0.00004844, 0.00005230, 0.00002443, 0.00001202, 0.00002127, 0.00002148, 0.00001387, 0.00001936, 0.00003011, 0.00002665, },
		{ 0.00007380, 0.00004576, -0.00003636, -0.00006900, -0.00002888, -0.00000572, -0.00001785, -0.00001288, -0.00000531, -0.00001863, -0.00001699, 0.00000086, -0.00000015, -0.00000741, -0.00000097, 0.00000411, 0.00000148, -0.00000329, -0.00000650, 0.00000036, 0.00000622, -0.00000391, -0.00000872, 0.00000043, -0.00000059, -0.00000433, 0.00000410, 0.00000173, -0.00000987, -0.00000427, 0.00000204, -0.00000617, -0.00000562, 0.00000420, 0.00000644, 0.00000646, 0.00000243, -0.00000669, -0.00000206, 0.00000552, -0.00000583, -0.00001007, 0.00000658, 0.00000988, -0.00000361, -0.00000478, 0.00000320, 0.00000200, -0.00000620, -0.00000600, 0.00000847, 0.00001085, -0.00001254, -0.00001646, 0.00000883, 0.00000516, -0.00001253, 0.00001217, 0.00002298, -0.00002083, -0.00002513, 0.00002367, 0.00001621, -0.00002535, -0.00000312, 0.00002325, -0.00001459, -0.00002581, 0.00002369, 0.00002425, -0.00002844, -0.00002010, 0.00002976, 0.00001378, -0.00003027, -0.00001146, 0.00002056, 0.00000129, -0.00001436, 0.00000539, 0.00000925, -0.00000452, -0.00000081, 0.00000139, -0.00000843, -0.00000498, -0.00000017, -0.00000888, -0.00000584, 0.00000537, -0.00000029, -0.00000484, 0.00000359, 0.00000245, -0.00000442, -0.00000148, -0.00000247, -0.00000950, -0.00000577, 0.00000263, 0.00000068, -0.00000902, -0.00001319, -0.00000201, 0.00000832, -0.00000628, -0.00002264, -0.00000697, 0.00001333, 0.00000190, -0.00002178, -0.00002836, -0.00001984, -0.00001066, -0.00000635, 0.00000788, 0.00003901, 0.00005213, 0.00002789, 0.00000826, 0.00001754, 0.00002355, 0.00001373, 0.00001325, 0.00002524, 0.00002772, 0.00002392, 0.00004383, },
		{ 0.00000172, -0.00005845, -0.00004805, -0.00001273, -0.00001246, -0.00001628, -0.00000593, -0.00001196, -0.00001981, -0.00000535, 0.00000240, -0.00000601, -0.00000428, 0.00000317, 0.00000162, -0.00000160, -0.00000316, -0.00000179, 0.00000326, -0.00000013, -0.00000743, -0.00000308, -0.00000002, -0.00000399, 0.00000202, 0.00000506, -0.00000778, -0.00000884, 0.00000275, -0.00000212, -0.00000936, 0.00000134, 0.00000937, 0.00000589, 0.00000199, -0.00000313, -0.00000364, 0.00000266, -0.00000174, -0.00000905, 0.00000130, 0.00000950, -0.00000020, -0.00000468, 0.00000214, 0.00000203, -0.00000480, -0.00000531, 0.00000445, 0.00000975, -0.00000490, -0.00001509, 0.00000096, 0.00000563, -0.00000937, 0.00000583, 0.00002490, -0.00000895, -0.00003001, 0.00001135, 0.00002316, -0.00001749, -0.00001202, 0.00002248, -0.00000304, -0.00003131, 0.00000927, 0.00003265, -0.00001407, -0.00003115, 0.00001643, 0.00002651, -0.00001923, -0.00002397, 0.00001415, 0.00001233, -0.00001317, -0.00000222, 0.00001236, -0.00000172, -0.00000526, 0.00000471, -0.00000384, -0.00001088, -0.00000122, -0.00000203, -0.00000799, -0.00000015, 0.00000233, -0.00000322, 0.00000184, 0.00000405, -0.00000418, -0.00000328, -0.00000007, -0.00000815, -0.00000903, 0.00000184, 0.00000267, -0.00000811, -0.00001264, -0.00000274, 0.00000787, -0.00000348, -0.00002256, -0.00001076, 0.00001492, 0.00000654, -0.00002237, -0.00002977, -0.00001805, -0.00000819, -0.00000359, 0.00000597, 0.00003012, 0.00004858, 0.00003239, 0.00000711, 0.00001188, 0.00002440, 0.00001582, 0.00000853, 0.00001924, 0.00002725, 0.00002336, 0.00003106, 0.00005935, 0.00006171, },
		{ -0.00005673, -0.00002782, -0.00001028, -0.00001521, -0.00000955, -0.00000708, -0.00001743, -0.00001217, 0.00000118, -0.00000246, -0.00000622, 0.00000067, 0.00000186, -0.00000098, -0.00000042, -0.00000105, -0.00000015, 0.00000083, -0.00000409, -0.00000521, -0.00000152, -0.00000278, 0.00000018, 0.00000586, -0.00000363, -0.00001127, -0.00000012, 0.00000201, -0.00000932, -0.00000357, 0.00001025, 0.00000777, 0.00000071, -0.00000105, -0.00000247, -0.00000066, -0.00000049, -0.00000565, -0.00000208, 0.00000635, 0.00000252, -0.00000298, 0.00000092, 0.00000161, -0.00000357, -0.00000411, 0.00000179, 0.00000684, 0.00000019, -0.00001060, -0.00000446, 0.00000314, -0.00000609, 0.00000126, 0.00002281, 0.00000230, -0.00002840, -0.00000150, 0.00002392, -0.00000740, -0.00001696, 0.00001721, 0.00000712, -0.00002968, -0.00000534, 0.00003193, 0.00000168, -0.00003193, 0.00000016, 0.00002865, -0.00000365, -0.00002742, 0.00000167, 0.00001762, -0.00000550, -0.00000836, 0.00001051, 0.00000342, -0.00000812, 0.00000359, 0.00000295, -0.00001275, -0.00000703, 0.00000328, -0.00000471, -0.00000650, 0.00000125, 0.00000032, 0.00000080, 0.00000415, -0.00000246, -0.00000514, 0.00000034, -0.00000521, -0.00001099, -0.00000063, 0.00000451, -0.00000674, -0.00001258, -0.00000270, 0.00000724, -0.00000188, -0.00002094, -0.00001387, 0.00001425, 0.00001157, -0.00002135, -0.00003224, -0.00001637, -0.00000558, -0.00000257, 0.00000577, 0.00002406, 0.00004226, 0.00003529, 0.00000941, 0.00000617, 0.00002244, 0.00001932, 0.00000612, 0.00001260, 0.00002521, 0.00002402, 0.00002407, 0.00004406, 0.00006203, 0.00003207, -0.00003271, },
		{ -0.00001583, -0.00001158, -0.00001218, -0.00000622, -0.00001216, -0.00001589, -0.00000384, 0.00000044, -0.00000538, -0.00000259, 0.00000126, -0.00000047, 0.00000095, 0.00000120, -0.00000187, -0.00000080, -0.00000103, -0.00000474, -0.00000403, -0.00000236, -0.00000041, 0.00000481, 0.00000015, -0.00001045, -0.00000441, 0.00000358, -0.00000611, -0.00000760, 0.00000793, 0.00001052, 0.00000063, -0.00000103, -0.00000015, -0.00000229, -0.00000191, -0.00000238, -0.00000235, 0.00000224, 0.00000294, -0.00000027, 0.00000069, 0.00000058, -0.00000291, -0.00000256, 0.00000065, 0.00000326, 0.00000224, -0.00000485, -0.00000649, -0.00000096, -0.00000394, -0.00000110, 0.00001806, 0.00001053, -0.00002153, -0.00001160, 0.00001902, 0.00000184, -0.00001722, 0.00000992, 0.00001361, -0.00002315, -0.00001619, 0.00002439, 0.00001319, -0.00002453, -0.00001191, 0.00002124, 0.00000853, -0.00002101, -0.00000967, 0.00001417, 0.00000386, -0.00000873, 0.00000494, 0.00000710, -0.00000723, -0.00000035, 0.00000793, -0.00000930, -0.00001393, 0.00000357, 0.00000232, -0.00000976, -0.00000396, 0.00000333, 0.00000206, 0.00000288, -0.00000058, -0.00000556, -0.00000121, -0.00000246, -0.00001058, -0.00000390, 0.00000511, -0.00000462, -0.00001283, -0.00000257, 0.00000732, -0.00000154, -0.00001907, -0.00001516, 0.00001203, 0.00001542, -0.00001834, -0.00003496, -0.00001585, -0.00000203, -0.00000269, 0.00000494, 0.00002148, 0.00003569, 0.00003453, 0.00001377, 0.00000301, 0.00001752, 0.00002219, 0.00000682, 0.00000618, 0.00002147, 0.00002482, 0.00002083, 0.00003201, 0.00005349, 0.00004865, -0.00000175, -0.00004970, -0.00004361, },
		{ -0.00001139, -0.00000816, -0.00000772, -0.00001502, -0.00000996, 0.00000012, -0.00000237, -0.00000473, -0.00000082, -0.00000002, 0.00000146, 0.00000327, -0.00000129, -0.00000319, -0.00000004, -0.00000233, -0.00000565, -0.00000351, -0.00000026, 0.00000343, 0.00000213, -0.00000738, -0.00000750, 0.00000198, -0.00000212, -0.00000868, 0.00000346, 0.00001191, 0.00000250, -0.00000196, 0.00000140, -0.00000165, -0.00000421, -0.00000116, -0.00000033, -0.00000062, 0.00000093, 0.00000187, 0.00000196, -0.00000032, -0.00000320, -0.00000117, 0.00000103, 0.00000012, 0.00000145, 0.00000031, -0.00000512, -0.00000503, -0.00000361, -0.00000154, 0.00001243, 0.00001441, -0.00001185, -0.00001668, 0.00001050, 0.00000778, -0.00001365, 0.00000304, 0.00001592, -0.00001459, -0.00002176, 0.00001398, 0.00001808, -0.00001403, -0.00001613, 0.00001000, 0.00001210, -0.00000966, -0.00001348, 0.00000454, 0.00000859, -0.00000286, 0.00000008, 0.00000643, -0.00000368, -0.00000358, 0.00000876, -0.00000277, -0.00001762, -0.00000143, 0.00000891, -0.00000754, -0.00001091, 0.00000301, 0.00000558, 0.00000182, -0.00000001, -0.00000425, -0.00000329, -0.00000156, -0.00000819, -0.00000629, 0.00000389, -0.00000229, -0.00001271, -0.00000298, 0.00000824, -0.00000160, -0.00001821, -0.00001468, 0.00000992, 0.00001706, -0.00001418, -0.00003651, -0.00001705, 0.00000205, -0.00000232, 0.00000209, 0.00002086, 0.00003174, 0.00003032, 0.00001713, 0.00000386, 0.00001156, 0.00002212, 0.00001010, 0.00000162, 0.00001585, 0.00002510, 0.00001989, 0.00002366, 0.00004253, 0.00005239, 0.00002418, -0.00002942, -0.00005126, -0.00002842, -0.00001035, },
		{ -0.00000640, -0.00001116, -0.00001377, -0.00000377, 0.00000039, -0.00000435, -0.00000371, -0.00000041, 0.00000186, 0.00000439, 0.00000065, -0.00000448, -0.00000124, 0.00000017, -0.00000519, -0.00000552, -0.00000060, 0.00000269, 0.00000245, -0.00000397, -0.00000806, -0.00000139, 0.00000021, -0.00000680, -0.00000083, 0.00001065, 0.00000534, -0.00000219, 0.00000138, 0.00000022, -0.00000549, -0.00000221, 0.00000198, -0.00000099, -0.00000212, 0.00000208, 0.00000414, -0.00000013, -0.00000422, -0.00000068, 0.00000251, -0.00000156, -0.00000132, 0.00000338, -0.00000123, -0.00000752, -0.00000519, -0.00000093, 0.00000766, 0.00001394, -0.00000230, -0.00001604, 0.00000123, 0.00000918, -0.00000804, -0.00000173, 0.00001477, -0.00000636, -0.00002251, 0.00000411, 0.00001725, -0.00000532, -0.00001353, 0.00000144, 0.00000732, -0.00000083, -0.00000824, -0.00000454, 0.00000511, 0.00000503, 0.00000048, 0.00000204, -0.00000109, -0.00000349, 0.00000646, 0.00000282, -0.00001612, -0.00000856, 0.00001106, -0.00000079, -0.00001558, -0.00000172, 0.00000923, 0.00000283, -0.00000101, -0.00000233, -0.00000418, -0.00000286, -0.00000549, -0.00000642, 0.00000148, -0.00000088, -0.00001174, -0.00000392, 0.00000942, -0.00000101, -0.00001875, -0.00001366, 0.00000942, 0.00001685, -0.00001057, -0.00003606, -0.00001934, 0.00000535, -0.00000026, -0.00000235, 0.00001950, 0.00003159, 0.00002524, 0.00001677, 0.00000766, 0.00000772, 0.00001831, 0.00001370, 0.00000055, 0.00000909, 0.00002379, 0.00002065, 0.00001820, 0.00003250, 0.00004832, 0.00003957, -0.00000396, -0.00004581, -0.00004201, -0.00001560, -0.00000848, -0.00000962, },
		{ -0.00001359, -0.00000907, 0.00000043, -0.00000180, -0.00000573, -0.00000226, 0.00000216, 0.00000489, 0.00000258, -0.00000398, -0.00000326, 0.00000113, -0.00000298, -0.00000686, -0.00000213, 0.00000236, 0.00000217, -0.00000160, -0.00000645, -0.00000438, -0.00000008, -0.00000368, -0.00000295, 0.00000727, 0.00000735, -0.00000088, 0.00000043, 0.00000178, -0.00000497, -0.00000426, 0.00000285, 0.00000066, -0.00000431, 0.00000019, 0.00000581, 0.00000154, -0.00000503, -0.00000148, 0.00000410, -0.00000126, -0.00000463, 0.00000353, 0.00000356, -0.00000734, -0.00000796, -0.00000051, 0.00000485, 0.00001042, 0.00000446, -0.00001068, -0.00000584, 0.00000622, -0.00000257, -0.00000361, 0.00001147, 0.00000007, -0.00001979, -0.00000342, 0.00001323, -0.00000070, -0.00000825, -0.00000084, -0.00000075, 0.00000015, 0.00000169, -0.00000643, -0.00000468, 0.00000851, 0.00000728, -0.00000187, -0.00000273, -0.00000070, 0.00000448, 0.00000467, -0.00001105, -0.00001334, 0.00000770, 0.00000657, -0.00001495, -0.00000898, 0.00001000, 0.00000632, -0.00000225, -0.00000141, -0.00000319, -0.00000517, -0.00000441, -0.00000428, -0.00000049, -0.00000122, -0.00001025, -0.00000476, 0.00001010, 0.00000063, -0.00001998, -0.00001361, 0.00001085, 0.00001641, -0.00000893, -0.00003411, -0.00002120, 0.00000687, 0.00000334, -0.00000652, 0.00001547, 0.00003390, 0.00002269, 0.00001228, 0.00001112, 0.00000824, 0.00001248, 0.00001464, 0.00000305, 0.00000329, 0.00001982, 0.00002229, 0.00001537, 0.00002399, 0.00004130, 0.00004485, 0.00001800, -0.00002895, -0.00004890, -0.00002703, -0.00000763, -0.00000837, -0.00000750, -0.00000753, },
		{ -0.00000288, 0.00000081, -0.00000543, -0.00000501, 0.00000162, 0.00000535, 0.00000385, -0.00000233, -0.00000455, 0.00000026, -0.00000053, -0.00000643, -0.00000433, 0.00000154, 0.00000204, -0.00000043, -0.00000405, -0.00000560, -0.00000220, -0.00000140, -0.00000249, 0.00000354, 0.00000728, 0.00000145, -0.00000018, 0.00000216, -0.00000324, -0.00000571, 0.00000184, 0.00000278, -0.00000450, -0.00000269, 0.00000576, 0.00000405, -0.00000464, -0.00000318, 0.00000471, 0.00000068, -0.00000689, 0.00000099, 0.00000732, -0.00000434, -0.00001052, -0.00000134, 0.00000408, 0.00000595, 0.00000696, -0.00000301, -0.00000860, 0.00000048, 0.00000080, -0.00000276, 0.00000749, 0.00000400, -0.00001502, -0.00000817, 0.00000833, 0.00000032, -0.00000437, 0.00000229, -0.00000586, -0.00000676, 0.00000886, 0.00000096, -0.00001414, 0.00000328, 0.00001661, -0.00000025, -0.00000854, 0.00000117, 0.00000584, 0.00000334, -0.00000643, -0.00001315, 0.00000150, 0.00001014, -0.00000927, -0.00001478, 0.00000628, 0.00001051, -0.00000174, -0.00000215, -0.00000116, -0.00000663, -0.00000571, -0.00000139, -0.00000038, -0.00000300, -0.00000934, -0.00000486, 0.00001008, 0.00000287, -0.00002073, -0.00001524, 0.00001314, 0.00001753, -0.00000929, -0.00003234, -0.00002139, 0.00000686, 0.00000708, -0.00000870, 0.00000885, 0.00003569, 0.00002456, 0.00000616, 0.00001081, 0.00001260, 0.00000818, 0.00001126, 0.00000693, 0.00000091, 0.00001343, 0.00002301, 0.00001545, 0.00001696, 0.00003363, 0.00004389, 0.00003174, -0.00000746, -0.00004455, -0.00003951, -0.00001252, -0.00000535, -0.00000829, -0.00000635, -0.00001050, -0.00001265, },
		{ -0.00000295, -0.00000709, -0.00000040, 0.00000576, 0.00000471, -0.00000065, -0.00000444, -0.00000142, 0.00000073, -0.00000440, -0.00000589, -0.00000031, 0.00000194, 0.00000014, -0.00000204, -0.00000500, -0.00000450, -0.00000108, -0.00000059, 0.00000123, 0.00000524, 0.00000338, 0.00000038, 0.00000164, -0.00000150, -0.00000572, -0.00000020, 0.00000384, -0.00000289, -0.00000492, 0.00000378, 0.00000606, -0.00000267, -0.00000472, 0.00000374, 0.00000313, -0.00000702, -0.00000293, 0.00000849, 0.00000051, -0.00001133, -0.00000369, 0.00000433, 0.00000259, 0.00000541, 0.00000390, -0.00000647, -0.00000553, 0.00000087, -0.00000022, 0.00000422, 0.00000526, -0.00000943, -0.00001022, 0.00000367, -0.00000041, -0.00000366, 0.00000686, -0.00000447, -0.00001652, 0.00000747, 0.00001359, -0.00001596, -0.00000947, 0.00002159, 0.00000889, -0.00001436, -0.00000176, 0.00001038, 0.00000234, -0.00000554, -0.00000914, -0.00000291, 0.00000812, -0.00000214, -0.00001557, -0.00000063, 0.00001231, 0.00000137, -0.00000350, 0.00000022, -0.00000616, -0.00000847, 0.00000019, 0.00000220, -0.00000469, -0.00001004, -0.00000432, 0.00000989, 0.00000491, -0.00002030, -0.00001811, 0.00001459, 0.00002094, -0.00001017, -0.00003246, -0.00002001, 0.00000677, 0.00000952, -0.00000860, 0.00000158, 0.00003436, 0.00002990, 0.00000228, 0.00000557, 0.00001750, 0.00000846, 0.00000482, 0.00000866, 0.00000288, 0.00000684, 0.00002076, 0.00001789, 0.00001225, 0.00002552, 0.00004028, 0.00003772, 0.00001122, -0.00003027, -0.00004618, -0.00002305, -0.00000424, -0.00000674, -0.00000724, -0.00000728, -0.00001270, -0.00000780, 0.00000106, },
		{ -0.00000348, 0.00000535, 0.00000572, 0.00000053, -0.00000343, -0.00000258, 0.00000045, -0.00000203, -0.00000584, -0.00000270, 0.00000118, 0.00000063, -0.00000072, -0.00000348, -0.00000572, -0.00000239, 0.00000106, 0.00000098, 0.00000255, 0.00000378, 0.00000187, 0.00000117, -0.00000060, -0.00000456, -0.00000196, 0.00000330, -0.00000073, -0.00000544, 0.00000085, 0.00000643, 0.00000022, -0.00000496, 0.00000155, 0.00000468, -0.00000501, -0.00000637, 0.00000659, 0.00000533, -0.00000940, -0.00000674, 0.00000413, 0.00000137, 0.00000157, 0.00000756, -0.00000075, -0.00000914, -0.00000225, 0.00000235, 0.00000272, 0.00000429, -0.00000428, -0.00000974, -0.00000041, -0.00000145, -0.00000523, 0.00000891, 0.00000241, -0.00002244, -0.00000296, 0.00002353, -0.00000701, -0.00002304, 0.00001674, 0.00002222, -0.00001435, -0.00001026, 0.00001438, 0.00000531, -0.00000831, -0.00000542, -0.00000228, 0.00000271, 0.00000167, -0.00001086, -0.00000679, 0.00000956, 0.00000566, -0.00000339, -0.00000002, -0.00000432, -0.00001081, -0.00000071, 0.00000601, -0.00000450, -0.00001235, -0.00000415, 0.00001026, 0.00000632, -0.00001890, -0.00002099, 0.00001409, 0.00002585, -0.00000953, -0.00003497, -0.00001860, 0.00000818, 0.00001048, -0.00000751, -0.00000412, 0.00002944, 0.00003556, 0.00000334, -0.00000246, 0.00001883, 0.00001360, -0.00000094, 0.00000567, 0.00000740, 0.00000325, 0.00001497, 0.00002048, 0.00001121, 0.00001716, 0.00003517, 0.00003926, 0.00002318, -0.00001209, -0.00004286, -0.00003470, -0.00000845, -0.00000359, -0.00000787, -0.00000607, -0.00000991, -0.00001144, -0.00000190, -0.00000014, -0.00000699, },
		{ 0.00000684, 0.00000153, -0.00000249, -0.00000266, -0.00000055, -0.00000060, -0.00000428, -0.00000448, -0.00000061, 0.00000087, 0.00000023, -0.00000192, -0.00000559, -0.00000416, 0.00000138, 0.00000210, 0.00000069, 0.00000254, 0.00000325, 0.00000151, -0.00000047, -0.00000316, -0.00000260, 0.00000173, 0.00000067, -0.00000434, -0.00000163, 0.00000494, 0.00000273, -0.00000347, -0.00000070, 0.00000445, -0.00000197, -0.00000776, 0.00000252, 0.00000809, -0.00000499, -0.00000889, 0.00000246, 0.00000192, -0.00000218, 0.00000705, 0.00000583, -0.00000865, -0.00000704, 0.00000321, 0.00000323, 0.00000217, -0.00000078, -0.00000702, -0.00000359, -0.00000257, -0.00000700, 0.00000694, 0.00001034, -0.00002040, -0.00001726, 0.00002387, 0.00000925, -0.00002908, 0.00000202, 0.00003235, -0.00000514, -0.00002045, 0.00001301, 0.00001288, -0.00001108, -0.00000598, 0.00000280, -0.00000125, -0.00000044, -0.00000394, -0.00000812, 0.00000301, 0.00000799, -0.00000057, -0.00000114, -0.00000278, -0.00001133, -0.00000351, 0.00000891, -0.00000163, -0.00001494, -0.00000549, 0.00001128, 0.00000748, -0.00001732, -0.00002290, 0.00001174, 0.00003045, -0.00000614, -0.00003860, -0.00001918, 0.00001140, 0.00001122, -0.00000717, -0.00000737, 0.00002270, 0.00003828, 0.00000905, -0.00000909, 0.00001450, 0.00002053, -0.00000216, -0.00000153, 0.00001058, 0.00000458, 0.00000762, 0.00002026, 0.00001397, 0.00001022, 0.00002796, 0.00003907, 0.00002923, 0.00000338, -0.00003072, -0.00004111, -0.00001805, -0.00000219, -0.00000632, -0.00000673, -0.00000673, -0.00001187, -0.00000669, 0.00000063, -0.00000461, -0.00000613, 0.00000337, },
		{ -0.00000204, -0.00000214, -0.00000111, -0.00000046, -0.00000232, -0.00000472, -0.00000293, 0.00000025, 0.00000107, -0.00000060, -0.00000462, -0.00000535, 0.00000041, 0.00000332, 0.00000036, 0.00000056, 0.00000355, 0.00000258, -0.00000045, -0.00000224, -0.00000219, 0.00000019, 0.00000074, -0.00000259, -0.00000266, 0.00000241, 0.00000372, -0.00000083, -0.00000173, 0.00000263, 0.00000051, -0.00000664, -0.00000181, 0.00000767, 0.00000034, -0.00000864, -0.00000053, 0.00000278, -0.00000402, 0.00000340, 0.00001035, -0.00000417, -0.00001103, 0.00000136, 0.00000495, 0.00000039, 0.00000034, -0.00000299, -0.00000514, -0.00000420, -0.00000766, 0.00000240, 0.00001470, -0.00001122, -0.00002782, 0.00001299, 0.00002444, -0.00002273, -0.00001630, 0.00003222, 0.00001108, -0.00002565, 0.00000415, 0.00002145, -0.00000925, -0.00001153, 0.00000808, 0.00000036, -0.00000684, -0.00000000, -0.00000328, -0.00000354, 0.00000561, 0.00000394, -0.00000103, -0.00000282, -0.00001025, -0.00000636, 0.00000936, 0.00000298, -0.00001588, -0.00000868, 0.00001212, 0.00000912, -0.00001614, -0.00002375, 0.00000867, 0.00003321, -0.00000042, -0.00004114, -0.00002271, 0.00001510, 0.00001354, -0.00000824, -0.00000918, 0.00001678, 0.00003683, 0.00001632, -0.00001073, 0.00000600, 0.00002460, 0.00000244, -0.00000929, 0.00000894, 0.00000972, 0.00000242, 0.00001562, 0.00001834, 0.00000719, 0.00001861, 0.00003726, 0.00003271, 0.00001290, -0.00001541, -0.00003829, -0.00002888, -0.00000587, -0.00000320, -0.00000717, -0.00000524, -0.00000946, -0.00001059, -0.00000175, -0.00000156, -0.00000674, 0.00000005, 0.00000724, 0.00000295, },
		{ -0.00000078, -0.00000089, -0.00000115, -0.00000350, -0.00000464, -0.00000152, 0.00000148, 0.00000066, -0.00000340, -0.00000565, -0.00000109, 0.00000371, 0.00000121, -0.00000098, 0.00000254, 0.00000366, 0.00000008, -0.00000190, -0.00000139, -0.00000054, -0.00000016, -0.00000133, -0.00000221, 0.00000014, 0.00000291, 0.00000167, -0.00000095, 0.00000036, 0.00000122, -0.00000390, -0.00000447, 0.00000442, 0.00000440, -0.00000553, -0.00000349, 0.00000243, -0.00000364, -0.00000109, 0.00001100, 0.00000231, -0.00001196, -0.00000275, 0.00000627, 0.00000013, -0.00000066, 0.00000080, -0.00000436, -0.00000639, -0.00000748, -0.00000192, 0.00001354, 0.00000023, -0.00002922, -0.00000427, 0.00003012, -0.00000609, -0.00002886, 0.00001979, 0.00002693, -0.00002070, -0.00000926, 0.00002524, -0.00000089, -0.00001853, 0.00000857, 0.00000783, -0.00001246, -0.00000267, 0.00000490, -0.00000537, -0.00000143, 0.00000698, 0.00000204, -0.00000403, -0.00000923, -0.00000758, 0.00000758, 0.00000720, -0.00001403, -0.00001260, 0.00001152, 0.00001155, -0.00001518, -0.00002422, 0.00000609, 0.00003382, 0.00000588, -0.00004086, -0.00002833, 0.00001705, 0.00001826, -0.00000971, -0.00001148, 0.00001315, 0.00003265, 0.00002133, -0.00000681, -0.00000234, 0.00002269, 0.00001029, -0.00001303, 0.00000200, 0.00001474, 0.00000239, 0.00000791, 0.00002059, 0.00000925, 0.00000921, 0.00003216, 0.00003593, 0.00001783, -0.00000314, -0.00002738, -0.00003485, -0.00001491, -0.00000165, -0.00000550, -0.00000550, -0.00000639, -0.00001156, -0.00000620, -0.00000021, -0.00000494, -0.00000321, 0.00000600, 0.00000482, -0.00000164, -0.00000192, },
		{ -0.00000111, -0.00000187, -0.00000482, -0.00000389, 0.00000083, 0.00000193, -0.00000204, -0.00000533, -0.00000235, 0.00000316, 0.00000231, -0.00000142, 0.00000085, 0.00000395, 0.00000114, -0.00000170, -0.00000083, -0.00000041, -0.00000118, -0.00000115, -0.00000108, -0.00000094, 0.00000099, 0.00000285, 0.00000119, -0.00000095, 0.00000007, -0.00000120, -0.00000451, 0.00000009, 0.00000550, -0.00000067, -0.00000462, 0.00000039, -0.00000222, -0.00000414, 0.00000798, 0.00000789, -0.00000896, -0.00000727, 0.00000563, 0.00000155, -0.00000248, 0.00000281, -0.00000138, -0.00000820, -0.00000770, -0.00000412, 0.00000841, 0.00000825, -0.00002131, -0.00001953, 0.00002286, 0.00001263, -0.00002874, 0.00000007, 0.00003386, -0.00000582, -0.00002024, 0.00002030, 0.00001119, -0.00002139, 0.00000227, 0.00001702, -0.00001208, -0.00001122, 0.00001081, -0.00000009, -0.00000961, 0.00000513, 0.00000761, -0.00000424, -0.00000989, -0.00000704, 0.00000533, 0.00000916, -0.00000995, -0.00001528, 0.00000886, 0.00001411, -0.00001371, -0.00002500, 0.00000438, 0.00003320, 0.00001092, -0.00003765, -0.00003379, 0.00001572, 0.00002437, -0.00000961, -0.00001542, 0.00001122, 0.00002859, 0.00002202, -0.00000009, -0.00000616, 0.00001541, 0.00001648, -0.00001037, -0.00000702, 0.00001530, 0.00000736, 0.00000110, 0.00001793, 0.00001463, 0.00000330, 0.00002300, 0.00003815, 0.00002189, 0.00000323, -0.00001389, -0.00003195, -0.00002541, -0.00000507, -0.00000215, -0.00000559, -0.00000477, -0.00000968, -0.00001020, -0.00000189, -0.00000208, -0.00000469, 0.00000311, 0.00000627, -0.00000053, -0.00000239, -0.00000010, -0.00000082, },
		{ -0.00000352, -0.00000564, -0.00000118, 0.00000271, -0.00000037, -0.00000468, -0.00000314, 0.00000215, 0.00000295, -0.00000094, -0.00000064, 0.00000324, 0.00000223, -0.00000117, -0.00000065, 0.00000013, -0.00000163, -0.00000180, -0.00000028, -0.00000073, -0.00000086, 0.00000233, 0.00000345, -0.00000050, -0.00000190, -0.00000001, -0.00000246, -0.00000319, 0.00000341, 0.00000383, -0.00000290, -0.00000239, -0.00000148, -0.00000477, 0.00000323, 0.00001017, -0.00000313, -0.00000980, 0.00000255, 0.00000353, -0.00000347, 0.00000241, 0.00000251, -0.00000824, -0.00000909, -0.00000435, 0.00000274, 0.00000980, -0.00000890, -0.00002577, 0.00000626, 0.00002370, -0.00001585, -0.00001733, 0.00002741, 0.00001255, -0.00002185, 0.00000780, 0.00002073, -0.00001642, -0.00000805, 0.00002196, -0.00000404, -0.00002066, 0.00000959, 0.00001032, -0.00001362, -0.00000252, 0.00001257, -0.00000119, -0.00001223, -0.00000634, 0.00000459, 0.00000866, -0.00000571, -0.00001521, 0.00000497, 0.00001548, -0.00001123, -0.00002610, 0.00000302, 0.00003264, 0.00001392, -0.00003302, -0.00003691, 0.00001145, 0.00002954, -0.00000660, -0.00002036, 0.00000896, 0.00002673, 0.00001929, 0.00000522, -0.00000388, 0.00000689, 0.00001686, -0.00000303, -0.00001303, 0.00000975, 0.00001369, -0.00000084, 0.00001079, 0.00001915, 0.00000335, 0.00001185, 0.00003659, 0.00002750, 0.00000561, -0.00000404, -0.00002118, -0.00003133, -0.00001403, -0.00000016, -0.00000357, -0.00000482, -0.00000689, -0.00001157, -0.00000591, -0.00000048, -0.00000369, -0.00000019, 0.00000617, 0.00000147, -0.00000298, -0.00000002, 0.00000016, -0.00000149, -0.00000100, },
		{ -0.00000395, 0.00000222, 0.00000155, -0.00000362, -0.00000363, 0.00000117, 0.00000301, -0.00000016, -0.00000140, 0.00000193, 0.00000274, -0.00000025, -0.00000058, 0.00000057, -0.00000140, -0.00000261, -0.00000033, 0.00000012, -0.00000173, 0.00000069, 0.00000464, 0.00000143, -0.00000325, -0.00000077, 0.00000010, -0.00000404, -0.00000051, 0.00000588, 0.00000119, -0.00000410, -0.00000249, -0.00000371, -0.00000078, 0.00000857, 0.00000293, -0.00000877, -0.00000186, 0.00000436, -0.00000270, 0.00000035, 0.00000538, -0.00000582, -0.00001105, -0.00000426, -0.00000064, 0.00000587, 0.00000145, -0.00002119, -0.00001115, 0.00002153, 0.00000272, -0.00002364, 0.00001039, 0.00002507, -0.00001188, -0.00000586, 0.00002199, -0.00000484, -0.00001627, 0.00001889, 0.00000831, -0.00002467, 0.00000052, 0.00002014, -0.00000973, -0.00001309, 0.00001291, 0.00000542, -0.00001419, -0.00000751, 0.00000593, 0.00000735, -0.00000350, -0.00001255, 0.00000179, 0.00001475, -0.00000820, -0.00002682, 0.00000122, 0.00003282, 0.00001551, -0.00002901, -0.00003700, 0.00000630, 0.00003173, -0.00000119, -0.00002425, 0.00000472, 0.00002700, 0.00001605, 0.00000620, 0.00000234, 0.00000212, 0.00001095, 0.00000407, -0.00001252, 0.00000060, 0.00001642, 0.00000316, 0.00000295, 0.00001884, 0.00000844, 0.00000305, 0.00002934, 0.00003351, 0.00000840, -0.00000062, -0.00000809, -0.00002845, -0.00002490, -0.00000307, 0.00000037, -0.00000478, -0.00000533, -0.00001005, -0.00000989, -0.00000176, -0.00000126, -0.00000211, 0.00000418, 0.00000352, -0.00000273, -0.00000097, 0.00000141, -0.00000123, -0.00000124, -0.00000177, -0.00000579, },
		{ 0.00000306, -0.00000188, -0.00000389, 0.00000029, 0.00000282, 0.00000043, -0.00000153, 0.00000067, 0.00000251, 0.00000068, -0.00000029, 0.00000076, -0.00000079, -0.00000301, -0.00000109, 0.00000073, -0.00000147, -0.00000101, 0.00000425, 0.00000373, -0.00000300, -0.00000268, 0.00000153, -0.00000250, -0.00000409, 0.00000468, 0.00000575, -0.00000324, -0.00000480, -0.00000269, -0.00000249, 0.00000447, 0.00000655, -0.00000443, -0.00000541, 0.00000290, -0.00000061, -0.00000163, 0.00000586, -0.00000169, -0.00001195, -0.00000537, -0.00000130, 0.00000013, 0.00000535, -0.00000981, -0.00002098, 0.00000781, 0.00001670, -0.00001621, -0.00000834, 0.00002502, 0.00000514, -0.00001268, 0.00001384, 0.00000750, -0.00001696, 0.00000908, 0.00001843, -0.00001980, -0.00001208, 0.00002327, 0.00000139, -0.00002096, 0.00000645, 0.00001296, -0.00001294, -0.00001118, 0.00000789, 0.00000752, -0.00000402, -0.00000924, 0.00000118, 0.00001239, -0.00000590, -0.00002647, -0.00000121, 0.00003337, 0.00001703, -0.00002682, -0.00003523, 0.00000276, 0.00003063, 0.00000442, -0.00002525, -0.00000137, 0.00002742, 0.00001498, 0.00000305, 0.00000807, 0.00000380, 0.00000239, 0.00000590, -0.00000610, -0.00000689, 0.00001266, 0.00001009, -0.00000078, 0.00001291, 0.00001435, 0.00000046, 0.00001794, 0.00003605, 0.00001461, -0.00000138, 0.00000081, -0.00001744, -0.00003162, -0.00001188, 0.00000333, -0.00000259, -0.00000545, -0.00000736, -0.00001154, -0.00000558, 0.00000036, -0.00000162, 0.00000129, 0.00000431, -0.00000118, -0.00000234, 0.00000183, 0.00000009, -0.00000188, -0.00000079, -0.00000430, -0.00000618, 0.00000008, },
		{ -0.00000362, -0.00000068, 0.00000260, 0.00000082, -0.00000144, -0.00000018, 0.00000183, 0.00000122, 0.00000024, 0.00000087, -0.00000018, -0.00000283, -0.00000203, 0.00000063, -0.00000063, -0.00000191, 0.00000272, 0.00000520, -0.00000119, -0.00000431, 0.00000107, 0.00000005, -0.00000558, 0.00000110, 0.00000853, 0.00000030, -0.00000675, -0.00000318, -0.00000208, 0.00000030, 0.00000642, 0.00000118, -0.00000609, -0.00000054, 0.00000123, -0.00000201, 0.00000408, 0.00000222, -0.00001042, -0.00000766, -0.00000086, -0.00000380, 0.00000272, 0.00000145, -0.00001956, -0.00000964, 0.00001859, -0.00000018, -0.00001893, 0.00001255, 0.00002013, -0.00000805, 0.00000111, 0.00001366, -0.00000919, -0.00000170, 0.00002083, -0.00000790, -0.00002126, 0.00001716, 0.00001456, -0.00002109, -0.00000501, 0.00001700, -0.00000717, -0.00001566, 0.00000806, 0.00001032, -0.00000583, -0.00000777, 0.00000347, 0.00001025, -0.00000553, -0.00002521, -0.00000366, 0.00003346, 0.00001942, -0.00002614, -0.00003371, 0.00000208, 0.00002797, 0.00000777, -0.00002323, -0.00000737, 0.00002587, 0.00001667, -0.00000145, 0.00000957, 0.00001055, -0.00000337, 0.00000076, 0.00000160, -0.00000813, 0.00000389, 0.00001464, 0.00000196, 0.00000459, 0.00001610, 0.00000440, 0.00000718, 0.00003185, 0.00002311, -0.00000111, 0.00000238, -0.00000388, -0.00002949, -0.00002335, 0.00000154, 0.00000188, -0.00000566, -0.00000570, -0.00001025, -0.00000982, -0.00000070, 0.00000062, -0.00000069, 0.00000311, 0.00000098, -0.00000292, 0.00000084, 0.00000179, -0.00000177, -0.00000102, -0.00000222, -0.00000672, -0.00000314, 0.00000320, 0.00000051, },
		{ 0.00000223, 0.00000125, -0.00000139, -0.00000072, 0.00000115, 0.00000125, 0.00000075, 0.00000110, 0.00000030, -0.00000229, -0.00000264, -0.00000011, 0.00000004, -0.00000181, 0.00000100, 0.00000523, 0.00000124, -0.00000456, -0.00000079, 0.00000191, -0.00000459, -0.00000285, 0.00000821, 0.00000504, -0.00000658, -0.00000529, -0.00000112, -0.00000198, 0.00000329, 0.00000533, -0.00000337, -0.00000420, 0.00000122, -0.00000055, 0.00000158, 0.00000402, -0.00000659, -0.00000959, -0.00000136, -0.00000463, -0.00000301, 0.00000699, -0.00000955, -0.00002160, 0.00000788, 0.00001453, -0.00001638, -0.00000517, 0.00002479, 0.00000593, -0.00000807, 0.00001024, 0.00000252, -0.00000680, 0.00001473, 0.00000478, -0.00002175, 0.00000470, 0.00002283, -0.00001238, -0.00001615, 0.00001413, 0.00000137, -0.00001778, 0.00000498, 0.00001471, -0.00000637, -0.00000943, 0.00000698, 0.00001050, -0.00000705, -0.00002426, -0.00000528, 0.00003262, 0.00002253, -0.00002563, -0.00003402, 0.00000352, 0.00002639, 0.00000792, -0.00002009, -0.00001104, 0.00002182, 0.00001936, -0.00000405, 0.00000615, 0.00001797, -0.00000257, -0.00000849, 0.00000488, -0.00000249, -0.00000479, 0.00001273, 0.00000916, -0.00000082, 0.00001145, 0.00001102, 0.00000221, 0.00002144, 0.00002922, 0.00000415, -0.00000107, 0.00000524, -0.00001867, -0.00003135, -0.00000654, 0.00000594, -0.00000374, -0.00000581, -0.00000747, -0.00001204, -0.00000451, 0.00000254, -0.00000040, 0.00000057, 0.00000230, -0.00000190, -0.00000099, 0.00000257, -0.00000039, -0.00000175, -0.00000087, -0.00000538, -0.00000603, 0.00000138, 0.00000276, -0.00000236, -0.00000171, },
		{ -0.00000127, -0.00000125, 0.00000066, 0.00000104, 0.00000096, 0.00000144, 0.00000074, -0.00000167, -0.00000279, -0.00000103, 0.00000010, -0.00000115, -0.00000009, 0.00000412, 0.00000311, -0.00000323, -0.00000276, 0.00000205, -0.00000214, -0.00000520, 0.00000513, 0.00000871, -0.00000363, -0.00000766, -0.00000130, -0.00000197, -0.00000066, 0.00000620, 0.00000138, -0.00000588, -0.00000106, 0.00000152, 0.00000023, 0.00000316, -0.00000216, -0.00000935, -0.00000344, -0.00000367, -0.00000770, 0.00000545, 0.00000230, -0.00002272, -0.00000895, 0.00001915, -0.00000347, -0.00001816, 0.00001669, 0.00002142, -0.00000734, -0.00000010, 0.00001093, -0.00000314, 0.00000469, 0.00001151, -0.00001341, -0.00000727, 0.00002181, 0.00000118, -0.00002096, 0.00000460, 0.00000835, -0.00001519, -0.00000040, 0.00001806, -0.00000377, -0.00001327, 0.00000896, 0.00001405, -0.00000882, -0.00002503, -0.00000592, 0.00003129, 0.00002550, -0.00002409, -0.00003624, 0.00000498, 0.00002765, 0.00000611, -0.00001850, -0.00001156, 0.00001688, 0.00002056, -0.00000340, 0.00000028, 0.00002159, 0.00000448, -0.00001623, 0.00000076, 0.00000609, -0.00000790, 0.00000458, 0.00001551, 0.00000038, 0.00000270, 0.00001462, 0.00000487, 0.00000957, 0.00002825, 0.00001382, -0.00000359, 0.00000609, -0.00000457, -0.00003059, -0.00001823, 0.00000562, 0.00000083, -0.00000630, -0.00000544, -0.00001127, -0.00000924, 0.00000204, 0.00000193, -0.00000158, 0.00000169, 0.00000029, -0.00000225, 0.00000161, 0.00000159, -0.00000174, -0.00000084, -0.00000318, -0.00000722, -0.00000194, 0.00000370, -0.00000002, -0.00000229, 0.00000141, 0.00000179, },
		{ 0.00000018, 0.00000090, 0.00000086, 0.00000165, 0.00000125, -0.00000105, -0.00000260, -0.00000173, -0.00000039, -0.00000060, -0.00000032, 0.00000269, 0.00000375, -0.00000108, -0.00000363, 0.00000064, 0.00000013, -0.00000512, 0.00000101, 0.00000962, 0.00000109, -0.00000843, -0.00000322, -0.00000092, -0.00000324, 0.00000387, 0.00000565, -0.00000443, -0.00000442, 0.00000241, 0.00000095, 0.00000080, 0.00000066, -0.00000643, -0.00000584, -0.00000321, -0.00000920, -0.00000036, 0.00000956, -0.00001406, -0.00002255, 0.00001109, 0.00001101, -0.00001965, 0.00000077, 0.00002945, 0.00000372, -0.00001004, 0.00001087, 0.00000667, -0.00000227, 0.00000941, -0.00000137, -0.00001218, 0.00001245, 0.00001257, -0.00001654, -0.00000716, 0.00000944, -0.00000841, -0.00000502, 0.00001808, 0.00000165, -0.00001659, 0.00000743, 0.00001965, -0.00000843, -0.00002796, -0.00000658, 0.00003041, 0.00002761, -0.00002131, -0.00003912, 0.00000448, 0.00003155, 0.00000502, -0.00002007, -0.00001034, 0.00001357, 0.00001903, -0.00000081, -0.00000440, 0.00001966, 0.00001368, -0.00001762, -0.00000898, 0.00001148, -0.00000340, -0.00000522, 0.00001590, 0.00000745, -0.00000467, 0.00001149, 0.00001217, 0.00000246, 0.00001927, 0.00002281, -0.00000002, 0.00000065, 0.00000526, -0.00002045, -0.00002742, -0.00000129, 0.00000570, -0.00000483, -0.00000524, -0.00000863, -0.00001239, -0.00000140, 0.00000446, -0.00000171, -0.00000067, 0.00000220, -0.00000175, -0.00000069, 0.00000271, -0.00000025, -0.00000150, -0.00000157, -0.00000631, -0.00000528, 0.00000246, 0.00000261, -0.00000171, 0.00000016, 0.00000213, -0.00000077, -0.00000186, },
		{ 0.00000073, 0.00000158, 0.00000176, -0.00000039, -0.00000227, -0.00000208, -0.00000106, -0.00000054, -0.00000005, 0.00000169, 0.00000330, 0.00000082, -0.00000304, -0.00000123, 0.00000098, -0.00000325, -0.00000211, 0.00000756, 0.00000549, -0.00000650, -0.00000592, -0.00000059, -0.00000358, 0.00000010, 0.00000728, -0.00000049, -0.00000676, 0.00000107, 0.00000299, -0.00000101, 0.00000069, -0.00000215, -0.00000653, -0.00000449, -0.00000828, -0.00000607, 0.00000958, -0.00000168, -0.00002621, -0.00000485, 0.00001771, -0.00001008, -0.00001370, 0.00002545, 0.00001909, -0.00001224, 0.00000274, 0.00001599, -0.00000134, 0.00000128, 0.00000721, -0.00000779, 0.00000066, 0.00001594, -0.00000516, -0.00001486, 0.00000331, -0.00000107, -0.00000577, 0.00001469, 0.00000745, -0.00001685, 0.00000263, 0.00002450, -0.00000435, -0.00003177, -0.00000883, 0.00003044, 0.00002901, -0.00001806, -0.00004116, 0.00000142, 0.00003609, 0.00000695, -0.00002411, -0.00001017, 0.00001338, 0.00001586, 0.00000101, -0.00000569, 0.00001407, 0.00001992, -0.00001177, -0.00001882, 0.00000939, 0.00000573, -0.00001057, 0.00000909, 0.00001545, -0.00000581, 0.00000264, 0.00001804, 0.00000365, 0.00000649, 0.00002503, 0.00001008, -0.00000464, 0.00000604, -0.00000595, -0.00002819, -0.00001279, 0.00000677, -0.00000042, -0.00000576, -0.00000635, -0.00001257, -0.00000632, 0.00000499, 0.00000043, -0.00000324, 0.00000236, 0.00000050, -0.00000280, 0.00000181, 0.00000215, -0.00000150, -0.00000136, -0.00000420, -0.00000706, -0.00000070, 0.00000415, 0.00000030, -0.00000091, 0.00000176, 0.00000011, -0.00000225, -0.00000056, 0.00000087, },
		{ 0.00000203, 0.00000036, -0.00000186, -0.00000216, -0.00000159, -0.00000093, 0.00000015, 0.00000137, 0.00000245, 0.00000176, -0.00000151, -0.00000239, 0.00000029, -0.00000106, -0.00000308, 0.00000389, 0.00000763, -0.00000238, -0.00000760, -0.00000196, -0.00000249, -0.00000287, 0.00000576, 0.00000381, -0.00000647, -0.00000196, 0.00000460, -0.00000083, -0.00000139, 0.00000117, -0.00000442, -0.00000670, -0.00000729, -0.00000881, 0.00000430, 0.00000745, -0.00001937, -0.00001972, 0.00001234, 0.00000337, -0.00001888, 0.00001200, 0.00002997, -0.00000411, -0.00000759, 0.00001857, 0.00000693, -0.00000632, 0.00000773, 0.00000243, -0.00000640, 0.00001023, 0.00000682, -0.00001414, -0.00000702, 0.00000214, -0.00000186, 0.00001034, 0.00001077, -0.00001340, -0.00000302, 0.00002598, 0.00000287, -0.00003400, -0.00001358, 0.00003071, 0.00003061, -0.00001529, -0.00004174, -0.00000317, 0.00003892, 0.00001233, -0.00002805, -0.00001324, 0.00001551, 0.00001366, 0.00000011, -0.00000422, 0.00000870, 0.00002035, -0.00000216, -0.00002304, 0.00000032, 0.00001326, -0.00000801, -0.00000119, 0.00001851, 0.00000006, -0.00000675, 0.00001736, 0.00001143, -0.00000311, 0.00001788, 0.00002168, -0.00000330, -0.00000102, 0.00000493, -0.00001891, -0.00002302, 0.00000116, 0.00000504, -0.00000464, -0.00000583, -0.00001052, -0.00001022, 0.00000251, 0.00000337, -0.00000421, 0.00000047, 0.00000314, -0.00000314, -0.00000099, 0.00000388, 0.00000015, -0.00000209, -0.00000244, -0.00000655, -0.00000436, 0.00000342, 0.00000309, -0.00000082, 0.00000059, 0.00000087, -0.00000199, -0.00000153, 0.00000064, 0.00000081, 0.00000131, },
		{ -0.00000130, -0.00000209, -0.00000185, -0.00000147, -0.00000006, 0.00000143, 0.00000188, 0.00000177, 0.00000003, -0.00000230, -0.00000113, 0.00000014, -0.00000209, 0.00000055, 0.00000686, 0.00000211, -0.00000686, -0.00000451, -0.00000166, -0.00000378, 0.00000234, 0.00000622, -0.00000351, -0.00000487, 0.00000427, 0.00000123, -0.00000359, 0.00000182, -0.00000034, -0.00000775, -0.00000795, -0.00000866, -0.00000196, 0.00000957, -0.00000715, -0.00002591, -0.00000207, 0.00001144, -0.00001316, -0.00000292, 0.00003028, 0.00001039, -0.00001256, 0.00001278, 0.00001707, -0.00000748, 0.00000105, 0.00001162, -0.00000474, -0.00000016, 0.00001276, -0.00000542, -0.00001563, -0.00000143, 0.00000418, 0.00000842, 0.00001046, -0.00000808, -0.00000650, 0.00002341, 0.00001066, -0.00003251, -0.00002010, 0.00002964, 0.00003316, -0.00001316, -0.00004138, -0.00000765, 0.00003883, 0.00001950, -0.00002912, -0.00001961, 0.00001729, 0.00001457, -0.00000312, -0.00000274, 0.00000660, 0.00001601, 0.00000571, -0.00001935, -0.00001070, 0.00001374, 0.00000078, -0.00000854, 0.00001387, 0.00000879, -0.00001084, 0.00000958, 0.00001978, -0.00000433, 0.00000448, 0.00002757, 0.00000650, -0.00000930, 0.00000625, -0.00000393, -0.00002548, -0.00001002, 0.00000747, -0.00000041, -0.00000639, -0.00000836, -0.00001135, -0.00000197, 0.00000485, -0.00000282, -0.00000230, 0.00000446, -0.00000123, -0.00000418, 0.00000342, 0.00000309, -0.00000234, -0.00000212, -0.00000445, -0.00000668, 0.00000037, 0.00000514, 0.00000099, -0.00000073, 0.00000082, -0.00000106, -0.00000226, -0.00000010, 0.00000101, 0.00000118, 0.00000198, 0.00000103, },
		},
		{ { -0.00000125, -0.00000087, -0.00000585, -0.00000786, -0.00000214, -0.00000305, -0.00000482, 0.00000004, 0.00000227, -0.00000365, -0.00000386, -0.00000003, -0.00000772, -0.00000736, -0.00000314, -0.00000532, 0.00000032, -0.00001590, -0.00003121, -0.00001091, 0.00000457, 0.00002709, 0.00002993, -0.00000482, 0.00000737, 0.00001499, -0.00000680, 0.00001656, 0.00001465, -0.00000382, 0.00001470, 0.00000117, -0.00000166, 0.00002956, 0.00001939, -0.00000334, -0.00000240, 0.00000250, -0.00001178, -0.00002207, 0.00000394, 0.00000303, -0.00001475, 0.00000555, 0.00000144, -0.00001130, 0.00000767, 0.00000509, -0.00000668, 0.00000111, 0.00001199, 0.00000543, -0.00001563, -0.00000767, 0.00000636, 0.00000206, 0.00001052, -0.00000545, -0.00002399, 0.00001164, 0.00001935, -0.00001073, -0.00000272, 0.00000373, -0.00000666, 0.00000373, -0.00000272, -0.00001073, 0.00001935, 0.00001164, -0.00002399, -0.00000545, 0.00001052, 0.00000206, 0.00000636, -0.00000767, -0.00001563, 0.00000543, 0.00001199, 0.00000111, -0.00000668, 0.00000509, 0.00000767, -0.00001130, 0.00000144, 0.00000555, -0.00001475, 0.00000303, 0.00000394, -0.00002207, -0.00001178, 0.00000250, -0.00000240, -0.00000334, 0.00001939, 0.00002956, -0.00000166, 0.00000117, 0.00001470, -0.00000382, 0.00001465, 0.00001656, -0.00000680, 0.00001499, 0.00000737, -0.00000482, 0.00002993, 0.00002709, 0.00000457, -0.00001091, -0.00003121, -0.00001590, 0.00000032, -0.00000532, -0.00000314, -0.00000736, -0.00000772, -0.00000003, -0.00000386, -0.00000365, 0.00000227, 0.00000004, -0.00000482, -0.00000305, -0.00000214, -0.00000786, -0.00000585, -0.00000087, },
		{ -0.00000205, -0.00000771, -0.00000649, -0.00000188, -0.00000409, -0.00000279, 0.00000171, -0.00000009, -0.00000437, -0.00000108, -0.00000336, -0.00001059, -0.00000296, -0.00000310, -0.00000569, -0.00000162, -0.00002426, -0.00003089, -0.00000471, 0.00001397, 0.00003305, 0.00001852, -0.00000471, 0.00001490, 0.00000663, -0.00000033, 0.00002281, 0.00000508, 0.00000167, 0.00001286, -0.00000528, 0.00001554, 0.00003270, 0.00000393, -0.00000252, 0.00000355, -0.00000873, -0.00002164, -0.00000807, 0.00000796, -0.00001035, -0.00000589, 0.00001033, -0.00001062, -0.00000308, 0.00001374, -0.00000607, -0.00000529, 0.00001089, 0.00000903, -0.00000510, -0.00001394, -0.00000154, 0.00000442, 0.00000628, 0.00000863, -0.00001681, -0.00001422, 0.00002128, 0.00000766, -0.00001253, 0.00000307, -0.00000030, -0.00000509, 0.00000509, -0.00000896, -0.00000198, 0.00002520, -0.00000443, -0.00002359, 0.00000606, 0.00000544, 0.00000267, 0.00000626, -0.00001707, -0.00000812, 0.00001534, 0.00000425, -0.00000488, 0.00000130, 0.00000755, -0.00000326, -0.00000694, 0.00000789, -0.00000753, -0.00000911, 0.00001226, -0.00001337, -0.00002177, 0.00000518, -0.00000035, -0.00000885, 0.00000773, 0.00002828, 0.00001442, -0.00000836, 0.00001195, 0.00000870, -0.00000298, 0.00002192, 0.00000380, -0.00000423, 0.00002026, -0.00000409, 0.00000412, 0.00003730, 0.00001466, -0.00000140, -0.00001455, -0.00002969, -0.00000802, 0.00000130, -0.00000598, -0.00000561, -0.00000835, -0.00000284, -0.00000077, -0.00000607, -0.00000057, 0.00000273, -0.00000288, -0.00000523, -0.00000151, -0.00000342, -0.00000840, -0.00000352, -0.00000044, -0.00000135, },
		{ -0.00000871, -0.00000494, -0.00000225, -0.00000378, -0.00000050, 0.00000126, -0.00000211, -0.00000277, -0.00000097, -0.00000847, -0.00000861, 0.00000105, -0.00000619, -0.00000489, -0.00000558, -0.00003294, -0.00002669, 0.00000424, 0.00002245, 0.00003248, 0.00000922, 0.00000042, 0.00001626, 0.00000198, 0.00000980, 0.00002070, -0.00000030, 0.00000742, 0.00000545, 0.00000041, 0.00003176, 0.00002104, -0.00000471, 0.00000574, -0.00000282, -0.00002268, -0.00001555, 0.00000412, -0.00000158, -0.00001309, 0.00000699, 0.00000168, -0.00001437, 0.00001118, 0.00000659, -0.00001392, 0.00000608, 0.00001325, -0.00000020, -0.00000876, -0.00000958, 0.00000028, 0.00000556, 0.00000979, 0.00000025, -0.00002095, -0.00000043, 0.00002137, -0.00000320, -0.00000865, 0.00000538, -0.00000432, -0.00000151, 0.00000270, -0.00001140, 0.00001029, 0.00002083, -0.00001882, -0.00001351, 0.00001033, -0.00000056, 0.00000800, -0.00000124, -0.00002168, 0.00000754, 0.00001431, -0.00000699, 0.00000034, 0.00000750, -0.00000005, -0.00000664, 0.00000244, 0.00000161, -0.00001416, 0.00000595, 0.00000461, -0.00002797, -0.00000369, 0.00001167, -0.00001286, -0.00000346, 0.00002150, 0.00002331, -0.00000081, -0.00000296, 0.00001761, 0.00000021, 0.00000462, 0.00002070, -0.00000777, 0.00000616, 0.00001644, -0.00001176, 0.00002017, 0.00003416, 0.00000128, -0.00000188, -0.00001844, -0.00002633, -0.00000015, 0.00000008, -0.00000853, -0.00000721, -0.00000551, -0.00000011, -0.00000412, -0.00000522, 0.00000250, 0.00000062, -0.00000517, -0.00000394, -0.00000062, -0.00000533, -0.00000770, -0.00000136, -0.00000074, -0.00000163, -0.00000377, },
		{ -0.00000356, -0.00000248, -0.00000239, 0.00000058, -0.00000030, -0.00000224, -0.00000129, -0.00000433, -0.00001117, -0.00000296, 0.00000099, -0.00000973, -0.00000271, -0.00001337, -0.00003931, -0.00001759, 0.00001324, 0.00002786, 0.00002858, 0.00000382, 0.00000595, 0.00001443, 0.00000304, 0.00001723, 0.00001430, -0.00000042, 0.00000836, 0.00000114, 0.00001619, 0.00003510, 0.00000436, -0.00000059, 0.00000778, -0.00001958, -0.00002460, -0.00000135, 0.00000315, -0.00001146, -0.00000341, 0.00001013, -0.00001094, -0.00000432, 0.00001746, -0.00000849, -0.00000879, 0.00001686, 0.00000519, -0.00000621, -0.00000709, -0.00000774, 0.00000214, 0.00000910, 0.00000802, -0.00000864, -0.00001685, 0.00001013, 0.00001483, -0.00000909, -0.00000288, 0.00000408, -0.00000634, 0.00000179, -0.00000175, -0.00000785, 0.00001979, 0.00000797, -0.00002423, -0.00000089, 0.00000624, -0.00000075, 0.00001129, -0.00001416, -0.00001374, 0.00002042, 0.00000068, -0.00000890, 0.00001131, 0.00000269, -0.00000716, 0.00000000, 0.00000410, -0.00000963, -0.00000619, 0.00001380, -0.00001577, -0.00002325, 0.00001734, -0.00000050, -0.00001889, 0.00001299, 0.00002510, 0.00000984, -0.00000658, 0.00000717, 0.00001527, -0.00000412, 0.00001322, 0.00001106, -0.00001083, 0.00001668, 0.00000449, -0.00000842, 0.00003458, 0.00002002, -0.00000555, 0.00000001, -0.00002409, -0.00001916, 0.00000600, -0.00000473, -0.00001077, -0.00000562, -0.00000192, -0.00000118, -0.00000655, -0.00000179, 0.00000312, -0.00000263, -0.00000555, -0.00000200, -0.00000099, -0.00000706, -0.00000584, -0.00000002, -0.00000154, -0.00000211, -0.00000574, -0.00000878, },
		{ -0.00000216, -0.00000102, 0.00000017, -0.00000117, -0.00000115, -0.00000232, -0.00000866, -0.00000913, 0.00000213, -0.00000338, -0.00001053, -0.00000181, -0.00002454, -0.00003994, -0.00000576, 0.00001991, 0.00003119, 0.00002351, 0.00000108, 0.00001059, 0.00001292, 0.00000710, 0.00001970, 0.00000835, 0.00000104, 0.00000605, 0.00000656, 0.00003096, 0.00002401, -0.00000415, 0.00000837, -0.00000465, -0.00003241, -0.00001189, 0.00000591, -0.00000695, -0.00001029, 0.00000728, 0.00000110, -0.00001403, 0.00001071, 0.00000894, -0.00001753, 0.00000638, 0.00001649, -0.00000563, -0.00000487, -0.00000636, -0.00000718, 0.00000679, 0.00001040, 0.00000217, -0.00001280, -0.00000904, 0.00001455, 0.00000678, -0.00001005, 0.00000175, 0.00000078, -0.00000612, 0.00000349, -0.00000536, -0.00000000, 0.00002182, -0.00000667, -0.00001958, 0.00000640, -0.00000037, 0.00000528, 0.00000599, -0.00002267, 0.00000467, 0.00001857, -0.00001422, 0.00000335, 0.00001435, -0.00000929, -0.00000362, 0.00000637, -0.00000491, -0.00001176, 0.00000815, 0.00000368, -0.00002990, 0.00000037, 0.00002136, -0.00002058, -0.00000742, 0.00002683, 0.00001543, -0.00000068, -0.00000308, 0.00001422, 0.00000880, -0.00000264, 0.00001673, -0.00000075, -0.00000461, 0.00001965, -0.00000764, 0.00000573, 0.00003788, 0.00000229, -0.00000286, -0.00000158, -0.00002873, -0.00000782, 0.00000676, -0.00001156, -0.00000954, -0.00000204, -0.00000083, -0.00000413, -0.00000588, 0.00000148, 0.00000097, -0.00000486, -0.00000427, -0.00000056, -0.00000240, -0.00000786, -0.00000340, 0.00000015, -0.00000243, -0.00000284, -0.00000770, -0.00000797, -0.00000230, },
		{ -0.00000051, -0.00000061, -0.00000066, -0.00000102, -0.00000575, -0.00001045, -0.00000385, 0.00000310, -0.00000892, -0.00000814, -0.00000523, -0.00003556, -0.00003392, 0.00000508, 0.00002490, 0.00003366, 0.00001704, 0.00000053, 0.00001527, 0.00001228, 0.00001101, 0.00001886, 0.00000403, 0.00000136, 0.00000680, 0.00001945, 0.00003428, 0.00000839, -0.00000068, 0.00000832, -0.00002490, -0.00002914, 0.00000352, -0.00000001, -0.00001306, -0.00000008, 0.00000779, -0.00000966, -0.00000427, 0.00001680, -0.00000694, -0.00001183, 0.00001878, 0.00000497, -0.00000962, -0.00000076, -0.00000922, -0.00000354, 0.00001163, 0.00000706, -0.00000286, -0.00001225, -0.00000238, 0.00001461, 0.00000022, -0.00000813, 0.00000424, -0.00000309, -0.00000419, 0.00000377, -0.00000648, 0.00000822, 0.00001634, -0.00001643, -0.00000999, 0.00000595, -0.00000259, 0.00001102, -0.00000654, -0.00001779, 0.00002059, 0.00000151, -0.00001526, 0.00001952, 0.00000237, -0.00001461, 0.00000776, 0.00000153, -0.00001364, -0.00000075, 0.00001219, -0.00001658, -0.00002229, 0.00002374, 0.00000303, -0.00002839, 0.00001541, 0.00002540, 0.00000187, -0.00000229, 0.00000309, 0.00001629, 0.00000281, 0.00000140, 0.00001401, -0.00000818, 0.00000528, 0.00001383, -0.00001158, 0.00002240, 0.00002760, -0.00000891, 0.00000477, -0.00000931, -0.00002739, 0.00000371, 0.00000036, -0.00001577, -0.00000444, 0.00000010, -0.00000256, -0.00000585, -0.00000300, 0.00000226, -0.00000213, -0.00000504, -0.00000251, -0.00000012, -0.00000422, -0.00000741, -0.00000119, -0.00000076, -0.00000304, -0.00000403, -0.00000926, -0.00000620, -0.00000110, -0.00000165, },
		{ -0.00000060, 0.00000006, -0.00000322, -0.00000902, -0.00000832, 0.00000090, -0.00000027, -0.00001222, -0.00000513, -0.00001334, -0.00004242, -0.00002393, 0.00001335, 0.00003055, 0.00003400, 0.00000907, 0.00000359, 0.00001976, 0.00001119, 0.00001416, 0.00001635, 0.00000005, 0.00000250, 0.00001446, 0.00003051, 0.00002551, -0.00000034, 0.00000587, -0.00000609, -0.00003717, -0.00001215, 0.00000772, -0.00001097, -0.00000783, 0.00000753, -0.00000109, -0.00001076, 0.00000864, 0.00000898, -0.00001608, 0.00000373, 0.00001828, -0.00000774, -0.00000489, -0.00000074, -0.00001190, 0.00000403, 0.00001209, 0.00000198, -0.00000423, -0.00001070, 0.00000215, 0.00001301, -0.00000465, -0.00000473, 0.00000443, -0.00000663, -0.00000066, 0.00000302, -0.00000546, 0.00001381, 0.00000694, -0.00001901, -0.00000187, 0.00000159, 0.00000113, 0.00000974, -0.00001693, -0.00000107, 0.00002083, -0.00001688, 0.00000129, 0.00002285, -0.00001580, -0.00000468, 0.00001292, -0.00001213, -0.00000946, 0.00001143, -0.00000076, -0.00002710, 0.00000274, 0.00002644, -0.00002214, -0.00001393, 0.00003144, 0.00000992, -0.00000399, 0.00000131, 0.00000744, 0.00001567, -0.00000181, 0.00000491, 0.00000822, -0.00000945, 0.00001265, 0.00000442, -0.00000548, 0.00003171, 0.00001028, -0.00000870, 0.00000921, -0.00001951, -0.00001820, 0.00000907, -0.00000987, -0.00001353, 0.00000106, -0.00000116, -0.00000463, -0.00000512, -0.00000049, 0.00000056, -0.00000399, -0.00000376, -0.00000126, -0.00000053, -0.00000584, -0.00000590, 0.00000004, -0.00000221, -0.00000334, -0.00000586, -0.00000980, -0.00000355, -0.00000034, -0.00000151, -0.00000061, },
		{ -0.00000076, -0.00000678, -0.00000955, -0.00000402, 0.00000236, -0.00000539, -0.00001239, -0.00000448, -0.00002358, -0.00004382, -0.00001342, 0.00002092, 0.00003722, 0.00002957, 0.00000229, 0.00001105, 0.00002174, 0.00000965, 0.00001720, 0.00001179, -0.00000313, 0.00000861, 0.00002572, 0.00003218, 0.00001304, -0.00000015, 0.00000355, -0.00002580, -0.00003180, 0.00000397, -0.00000119, -0.00001422, 0.00000296, 0.00000474, -0.00000862, -0.00000175, 0.00001266, -0.00000450, -0.00001133, 0.00001598, 0.00000645, -0.00001190, 0.00000192, -0.00000609, -0.00000932, 0.00001105, 0.00000745, -0.00000014, -0.00000406, -0.00001045, 0.00000652, 0.00001023, -0.00000811, -0.00000044, 0.00000192, -0.00000840, 0.00000418, 0.00000070, -0.00000302, 0.00001604, -0.00000258, -0.00001636, 0.00000184, -0.00000144, 0.00000634, 0.00000156, -0.00001632, 0.00001503, 0.00000465, -0.00001982, 0.00002223, 0.00000685, -0.00002258, 0.00001269, 0.00000216, -0.00001954, 0.00000614, 0.00000921, -0.00001861, -0.00001602, 0.00002400, 0.00000634, -0.00003197, 0.00001265, 0.00002784, -0.00000597, 0.00000018, 0.00000337, 0.00001107, 0.00001318, -0.00000601, 0.00000792, 0.00000208, -0.00000667, 0.00001562, -0.00000268, 0.00000549, 0.00002975, -0.00000417, -0.00000055, 0.00000585, -0.00002532, -0.00000547, 0.00000510, -0.00001701, -0.00000584, 0.00000269, -0.00000421, -0.00000479, -0.00000332, -0.00000002, -0.00000175, -0.00000383, -0.00000238, -0.00000071, -0.00000142, -0.00000683, -0.00000387, -0.00000001, -0.00000352, -0.00000373, -0.00000805, -0.00000858, -0.00000074, -0.00000059, -0.00000163, -0.00000044, 0.00000009, },
		{ -0.00000926, -0.00000702, -0.00000057, 0.00000053, -0.00000965, -0.00001088, -0.00000710, -0.00003312, -0.00004140, -0.00000320, 0.00003042, 0.00004169, 0.00002011, 0.00000098, 0.00002038, 0.00001940, 0.00000966, 0.00001921, 0.00000479, -0.00000190, 0.00002028, 0.00003307, 0.00002511, 0.00000495, 0.00000104, -0.00001006, -0.00003593, -0.00001442, 0.00000691, -0.00001218, -0.00000612, 0.00000773, -0.00000418, -0.00000650, 0.00000740, 0.00000557, -0.00001122, 0.00000152, 0.00001607, -0.00000618, -0.00000668, 0.00000338, -0.00001156, -0.00000142, 0.00001248, 0.00000222, 0.00000167, -0.00000631, -0.00000974, 0.00001158, 0.00000519, -0.00000915, 0.00000341, -0.00000304, -0.00000634, 0.00000841, -0.00000362, 0.00000114, 0.00001525, -0.00001033, -0.00001147, 0.00000216, -0.00000122, 0.00000808, -0.00000666, -0.00000523, 0.00001781, -0.00001417, -0.00000339, 0.00002846, -0.00001646, -0.00000953, 0.00001937, -0.00001705, -0.00000919, 0.00001661, -0.00000831, -0.00002193, 0.00000747, 0.00002365, -0.00001841, -0.00001821, 0.00003064, 0.00000886, -0.00000984, 0.00000669, 0.00000262, 0.00001607, 0.00000671, -0.00000834, 0.00001114, -0.00000439, -0.00000142, 0.00001561, -0.00000587, 0.00001490, 0.00002007, -0.00001003, 0.00000794, -0.00000306, -0.00002314, 0.00000338, -0.00000477, -0.00001613, 0.00000180, -0.00000046, -0.00000572, -0.00000331, -0.00000249, -0.00000111, -0.00000281, -0.00000249, -0.00000178, -0.00000047, -0.00000257, -0.00000707, -0.00000201, -0.00000110, -0.00000431, -0.00000476, -0.00000960, -0.00000550, 0.00000090, -0.00000176, -0.00000121, 0.00000025, 0.00000015, -0.00000352, },
		{ -0.00000340, 0.00000033, -0.00000275, -0.00001223, -0.00000936, -0.00001209, -0.00004116, -0.00003677, 0.00000891, 0.00004139, 0.00003978, 0.00000971, 0.00000710, 0.00002667, 0.00001419, 0.00001264, 0.00001767, -0.00000188, 0.00000698, 0.00003244, 0.00003210, 0.00001573, 0.00000137, -0.00000488, -0.00002543, -0.00003004, 0.00000011, -0.00000210, -0.00001343, 0.00000465, 0.00000206, -0.00000848, 0.00000238, 0.00000762, -0.00000391, -0.00000679, 0.00001075, 0.00000649, -0.00001074, 0.00000082, -0.00000137, -0.00001158, 0.00000649, 0.00000837, 0.00000135, 0.00000338, -0.00001165, -0.00000528, 0.00001510, -0.00000172, -0.00000670, 0.00000444, -0.00000805, 0.00000014, 0.00000867, -0.00000827, 0.00000749, 0.00001108, -0.00001559, -0.00000608, 0.00000103, 0.00000066, 0.00000597, -0.00000844, 0.00000658, 0.00000607, -0.00001925, 0.00001886, 0.00001308, -0.00002680, 0.00001237, 0.00000569, -0.00002528, 0.00001193, 0.00000740, -0.00002368, -0.00000485, 0.00002160, 0.00000436, -0.00002696, 0.00000671, 0.00002785, -0.00000927, -0.00000169, 0.00000756, 0.00000387, 0.00001994, -0.00000435, -0.00000513, 0.00001224, -0.00001094, 0.00000690, 0.00001292, -0.00000670, 0.00002028, 0.00000872, -0.00000821, 0.00001169, -0.00001174, -0.00001560, 0.00000411, -0.00001298, -0.00000850, 0.00000412, -0.00000501, -0.00000420, -0.00000239, -0.00000314, -0.00000209, -0.00000217, -0.00000141, -0.00000185, -0.00000014, -0.00000395, -0.00000660, -0.00000080, -0.00000267, -0.00000478, -0.00000639, -0.00000931, -0.00000174, 0.00000044, -0.00000262, 0.00000017, 0.00000055, -0.00000138, -0.00000679, -0.00000895, },
		{ -0.00000090, -0.00000598, -0.00001362, -0.00000833, -0.00001891, -0.00004832, -0.00002891, 0.00002478, 0.00004948, 0.00003084, 0.00000440, 0.00001771, 0.00002658, 0.00001010, 0.00001694, 0.00001159, -0.00000289, 0.00002196, 0.00003849, 0.00002488, 0.00000850, -0.00000372, -0.00001695, -0.00003087, -0.00001556, 0.00000210, -0.00001104, -0.00000433, 0.00000714, -0.00000689, -0.00000303, 0.00000786, 0.00000012, -0.00000616, 0.00000222, 0.00001011, -0.00000329, -0.00000712, 0.00000371, -0.00000712, -0.00000613, 0.00000936, 0.00000383, 0.00000489, 0.00000030, -0.00001612, 0.00000328, 0.00001393, -0.00000768, -0.00000184, 0.00000132, -0.00000923, 0.00000820, 0.00000304, -0.00000941, 0.00001447, 0.00000309, -0.00001724, -0.00000119, -0.00000081, 0.00000265, 0.00000348, -0.00000418, 0.00000946, -0.00000876, -0.00000652, 0.00002791, -0.00001144, -0.00001504, 0.00002215, -0.00001753, -0.00001118, 0.00002243, -0.00001622, -0.00001806, 0.00001700, 0.00001406, -0.00001420, -0.00001508, 0.00002277, 0.00001018, -0.00001398, 0.00000829, 0.00000318, 0.00001053, 0.00001677, -0.00001493, 0.00000430, 0.00000698, -0.00001369, 0.00001756, 0.00000552, -0.00000542, 0.00002258, -0.00000096, -0.00000252, 0.00001061, -0.00001668, -0.00000846, -0.00000101, -0.00001420, -0.00000033, 0.00000060, -0.00000664, -0.00000147, -0.00000355, -0.00000400, -0.00000180, -0.00000073, -0.00000147, -0.00000180, 0.00000019, -0.00000564, -0.00000552, -0.00000044, -0.00000432, -0.00000533, -0.00000774, -0.00000687, 0.00000071, -0.00000154, -0.00000182, 0.00000159, -0.00000062, -0.00000398, -0.00000859, -0.00000625, -0.00000106, },
		{ -0.00000884, -0.00001408, -0.00000785, -0.00002837, -0.00005393, -0.00001489, 0.00004251, 0.00004992, 0.00001927, 0.00000735, 0.00002679, 0.00002120, 0.00001008, 0.00001882, 0.00000414, 0.00000533, 0.00003662, 0.00003590, 0.00001632, 0.00000178, -0.00001379, -0.00002618, -0.00002416, -0.00000546, -0.00000486, -0.00001020, 0.00000507, -0.00000021, -0.00000867, 0.00000623, 0.00000394, -0.00000581, -0.00000038, 0.00000643, 0.00000326, -0.00000665, -0.00000154, 0.00000142, -0.00000896, -0.00000002, 0.00000746, 0.00000362, 0.00000788, -0.00000769, -0.00001492, 0.00001212, 0.00000770, -0.00000967, 0.00000210, -0.00000406, -0.00000444, 0.00001246, -0.00000615, -0.00000413, 0.00001806, -0.00000711, -0.00001425, 0.00000147, -0.00000283, 0.00000533, 0.00000292, -0.00000005, 0.00000288, -0.00001348, 0.00001236, 0.00001664, -0.00002465, 0.00000737, 0.00000951, -0.00002785, 0.00001431, 0.00000797, -0.00003091, 0.00000755, 0.00002132, -0.00000533, -0.00001568, 0.00000304, 0.00002094, -0.00000653, -0.00000621, 0.00001127, 0.00000059, 0.00001883, 0.00000444, -0.00001723, 0.00001407, -0.00000482, -0.00000717, 0.00002496, -0.00000685, 0.00000056, 0.00002187, -0.00000842, 0.00000476, 0.00000660, -0.00001860, -0.00000487, -0.00000609, -0.00000931, 0.00000282, -0.00000452, -0.00000419, -0.00000072, -0.00000599, -0.00000359, -0.00000035, 0.00000016, -0.00000247, -0.00000085, -0.00000013, -0.00000744, -0.00000383, -0.00000103, -0.00000592, -0.00000584, -0.00000774, -0.00000353, 0.00000066, -0.00000298, 0.00000064, 0.00000138, -0.00000308, -0.00000614, -0.00000810, -0.00000302, -0.00000096, -0.00000284, },
		{ -0.00001321, -0.00000915, -0.00004131, -0.00005453, 0.00000659, 0.00005656, 0.00004193, 0.00001161, 0.00001633, 0.00002971, 0.00001493, 0.00001340, 0.00001614, 0.00000119, 0.00002096, 0.00004431, 0.00002736, 0.00000826, -0.00000833, -0.00002450, -0.00002539, -0.00001384, -0.00000547, -0.00000963, -0.00000144, 0.00000538, -0.00000849, -0.00000092, 0.00000912, -0.00000485, -0.00000334, 0.00000576, 0.00000313, -0.00000205, -0.00000465, 0.00000098, -0.00000191, -0.00000704, 0.00000281, 0.00000541, 0.00000730, 0.00000546, -0.00001575, -0.00000717, 0.00001646, -0.00000043, -0.00000746, 0.00000247, -0.00000739, 0.00000399, 0.00000921, -0.00001294, 0.00000614, 0.00001470, -0.00001534, -0.00000789, 0.00000030, -0.00000289, 0.00000949, 0.00000236, -0.00000076, -0.00000486, -0.00000522, 0.00002135, -0.00000416, -0.00001720, 0.00001925, -0.00001395, -0.00001354, 0.00002615, -0.00002128, -0.00001768, 0.00003005, 0.00000261, -0.00001491, -0.00000296, 0.00001100, 0.00000878, -0.00001206, 0.00000408, 0.00000775, 0.00000443, 0.00002085, -0.00001108, -0.00000850, 0.00001564, -0.00001595, 0.00000879, 0.00002125, -0.00001848, 0.00001241, 0.00001568, -0.00001227, 0.00001279, -0.00000009, -0.00001933, -0.00000346, -0.00000735, -0.00000351, 0.00000047, -0.00000648, -0.00000045, -0.00000332, -0.00000743, -0.00000156, 0.00000126, -0.00000035, -0.00000332, 0.00000100, -0.00000192, -0.00000862, -0.00000177, -0.00000282, -0.00000713, -0.00000567, -0.00000634, -0.00000122, -0.00000108, -0.00000214, 0.00000283, -0.00000122, -0.00000535, -0.00000684, -0.00000599, -0.00000102, -0.00000255, -0.00000456, -0.00001158, },
		{ -0.00001466, -0.00005608, -0.00004552, 0.00003243, 0.00006147, 0.00002991, 0.00001174, 0.00002556, 0.00002637, 0.00001181, 0.00001665, 0.00001095, 0.00000698, 0.00003759, 0.00004256, 0.00001663, -0.00000153, -0.00002109, -0.00002794, -0.00001696, -0.00000966, -0.00001001, -0.00000581, 0.00000523, -0.00000248, -0.00000833, 0.00000839, 0.00000225, -0.00000870, 0.00000484, 0.00000506, -0.00000239, -0.00000231, -0.00000230, 0.00000080, -0.00000325, -0.00000493, 0.00000315, 0.00000676, 0.00001007, -0.00000237, -0.00001846, 0.00000328, 0.00001454, -0.00000651, -0.00000366, -0.00000016, -0.00000566, 0.00001047, -0.00000004, -0.00001220, 0.00001568, 0.00000495, -0.00001764, -0.00000223, -0.00000363, 0.00000170, 0.00001293, -0.00000150, -0.00000440, -0.00000573, 0.00000699, 0.00001477, -0.00001692, 0.00000101, 0.00001033, -0.00002540, 0.00001293, 0.00000978, -0.00003710, 0.00001664, 0.00002569, -0.00002040, -0.00000423, 0.00000675, 0.00000641, -0.00000081, -0.00000855, 0.00000964, 0.00000411, 0.00001162, 0.00001296, -0.00002013, 0.00000507, 0.00000633, -0.00001659, 0.00002539, 0.00000427, -0.00002004, 0.00002523, 0.00000293, -0.00000909, 0.00001891, -0.00001012, -0.00001823, -0.00000151, -0.00000579, -0.00000104, -0.00000351, -0.00000418, 0.00000071, -0.00000755, -0.00000621, 0.00000117, 0.00000195, -0.00000209, -0.00000271, 0.00000262, -0.00000536, -0.00000810, -0.00000027, -0.00000573, -0.00000706, -0.00000457, -0.00000475, -0.00000081, -0.00000228, 0.00000065, 0.00000245, -0.00000492, -0.00000602, -0.00000627, -0.00000349, -0.00000101, -0.00000461, -0.00000596, -0.00001401, -0.00001098, },
		{ -0.00006784, -0.00002468, 0.00005583, 0.00005601, 0.00002049, 0.00001833, 0.00002995, 0.00002042, 0.00001260, 0.00001709, 0.00000821, 0.00002087, 0.00004838, 0.00003302, 0.00000468, -0.00001452, -0.00002993, -0.00002151, -0.00001015, -0.00001279, -0.00000984, 0.00000237, 0.00000306, -0.00000914, 0.00000031, 0.00000967, -0.00000754, -0.00000301, 0.00000986, -0.00000177, -0.00000376, -0.00000036, -0.00000204, 0.00000094, -0.00000371, -0.00000447, 0.00000464, 0.00001029, 0.00000793, -0.00001104, -0.00001439, 0.00001140, 0.00000846, -0.00000892, -0.00000100, -0.00000263, -0.00000005, 0.00001058, -0.00000930, -0.00000395, 0.00001871, -0.00000619, -0.00001403, -0.00000110, -0.00000571, 0.00001051, 0.00001110, -0.00000812, -0.00000504, 0.00000005, 0.00001221, 0.00000102, -0.00001391, 0.00001173, -0.00000897, -0.00001353, 0.00002552, -0.00002205, -0.00001988, 0.00004221, -0.00000503, -0.00002264, 0.00001442, 0.00000175, 0.00000030, -0.00000342, -0.00000329, 0.00001116, 0.00000387, 0.00001552, -0.00000029, -0.00001820, 0.00001357, -0.00000698, -0.00000379, 0.00003012, -0.00001755, -0.00000699, 0.00002975, -0.00001153, 0.00000261, 0.00001790, -0.00002142, -0.00001293, 0.00000102, -0.00000463, -0.00000191, -0.00000510, -0.00000041, -0.00000222, -0.00001027, -0.00000254, 0.00000328, 0.00000109, -0.00000389, -0.00000021, 0.00000221, -0.00000923, -0.00000546, -0.00000077, -0.00000858, -0.00000494, -0.00000352, -0.00000414, -0.00000123, -0.00000146, 0.00000302, -0.00000095, -0.00000734, -0.00000512, -0.00000519, -0.00000158, -0.00000285, -0.00000604, -0.00000738, -0.00001554, -0.00000875, -0.00002631, },
		{ 0.00000540, 0.00007000, 0.00004414, 0.00001812, 0.00002629, 0.00002837, 0.00001612, 0.00001513, 0.00001518, 0.00001232, 0.00003781, 0.00004942, 0.00001819, -0.00000931, -0.00002701, -0.00002873, -0.00001149, -0.00001123, -0.00001582, -0.00000254, 0.00000594, -0.00000473, -0.00000722, 0.00000867, 0.00000163, -0.00001047, 0.00000701, 0.00000615, -0.00000771, -0.00000037, -0.00000047, -0.00000188, 0.00000206, -0.00000577, -0.00000362, 0.00000886, 0.00001182, 0.00000140, -0.00001587, -0.00000637, 0.00001468, 0.00000148, -0.00000858, 0.00000016, -0.00000256, 0.00000518, 0.00000473, -0.00001288, 0.00000664, 0.00001403, -0.00001329, -0.00000880, -0.00000402, -0.00000136, 0.00001806, 0.00000203, -0.00001209, -0.00000015, 0.00000534, 0.00000815, -0.00000740, -0.00000285, 0.00000658, -0.00001864, 0.00000966, 0.00001026, -0.00003894, 0.00002001, 0.00003319, -0.00003488, 0.00000134, 0.00001875, -0.00001127, 0.00000202, -0.00000446, 0.00000137, 0.00001196, 0.00000470, 0.00001360, -0.00001128, -0.00000925, 0.00001259, -0.00001371, 0.00001433, 0.00001754, -0.00003008, 0.00001444, 0.00002045, -0.00001840, 0.00001781, 0.00000637, -0.00002805, -0.00000349, 0.00000146, -0.00000516, -0.00000336, -0.00000353, 0.00000112, -0.00000742, -0.00000931, 0.00000189, 0.00000374, -0.00000107, -0.00000414, 0.00000296, -0.00000123, -0.00001127, -0.00000198, -0.00000394, -0.00000914, -0.00000154, -0.00000410, -0.00000411, -0.00000080, 0.00000064, 0.00000273, -0.00000544, -0.00000714, -0.00000385, -0.00000400, -0.00000099, -0.00000576, -0.00000615, -0.00000927, -0.00001581, -0.00000903, -0.00004347, -0.00007043, },
		{ 0.00007196, 0.00003255, 0.00002244, 0.00003056, 0.00002358, 0.00001533, 0.00001674, 0.00001433, 0.00002422, 0.00005115, 0.00003974, -0.00000033, -0.00002372, -0.00003217, -0.00001890, -0.00000708, -0.00001721, -0.00001177, 0.00000552, 0.00000132, -0.00000923, 0.00000125, 0.00000866, -0.00000783, -0.00000370, 0.00001196, -0.00000319, -0.00000721, 0.00000308, -0.00000284, 0.00000064, 0.00000151, -0.00000902, 0.00000059, 0.00001323, 0.00000893, -0.00000566, -0.00001562, 0.00000180, 0.00001361, -0.00000446, -0.00000675, 0.00000112, -0.00000059, 0.00000667, -0.00000262, -0.00000967, 0.00001376, 0.00000535, -0.00001496, -0.00000640, -0.00000587, 0.00000883, 0.00001749, -0.00000979, -0.00000845, 0.00000594, 0.00000468, 0.00000164, -0.00000640, 0.00000311, -0.00000577, -0.00000964, 0.00002046, -0.00001967, -0.00002147, 0.00004871, -0.00000617, -0.00003447, 0.00002995, 0.00000039, -0.00001519, 0.00000779, -0.00000825, 0.00000853, 0.00001097, 0.00000393, 0.00000892, -0.00001680, -0.00000018, 0.00000621, -0.00000955, 0.00002494, -0.00000520, -0.00002465, 0.00003089, 0.00000186, -0.00001169, 0.00002640, -0.00001201, -0.00002454, 0.00000555, -0.00000179, -0.00000581, -0.00000338, -0.00000089, -0.00000113, -0.00001152, -0.00000476, 0.00000504, 0.00000233, -0.00000327, -0.00000206, 0.00000443, -0.00000660, -0.00000995, -0.00000027, -0.00000822, -0.00000591, 0.00000048, -0.00000644, -0.00000292, 0.00000069, 0.00000188, -0.00000027, -0.00000840, -0.00000498, -0.00000318, -0.00000280, -0.00000232, -0.00000824, -0.00000501, -0.00001216, -0.00001499, -0.00001426, -0.00006238, -0.00005949, 0.00003799, },
		{ 0.00002660, 0.00002911, 0.00002933, 0.00001971, 0.00001697, 0.00001675, 0.00001870, 0.00004076, 0.00005481, 0.00002026, -0.00001998, -0.00003306, -0.00002689, -0.00000894, -0.00001127, -0.00001958, -0.00000184, 0.00000676, -0.00000619, -0.00000569, 0.00000770, 0.00000086, -0.00000999, 0.00000639, 0.00000753, -0.00001017, -0.00000152, 0.00000251, -0.00000389, 0.00000437, -0.00000283, -0.00000979, 0.00000781, 0.00001394, 0.00000326, -0.00001012, -0.00001210, 0.00000807, 0.00000961, -0.00000872, -0.00000354, 0.00000235, 0.00000083, 0.00000471, -0.00000727, -0.00000292, 0.00001530, -0.00000275, -0.00001406, -0.00000687, -0.00000171, 0.00001810, 0.00000714, -0.00001592, 0.00000113, 0.00000669, 0.00000026, -0.00000032, -0.00000239, -0.00000039, -0.00001043, 0.00000720, 0.00000739, -0.00003522, 0.00001865, 0.00003926, -0.00004286, -0.00000098, 0.00003452, -0.00002493, -0.00000246, 0.00000607, -0.00000990, 0.00001817, 0.00000431, 0.00000392, 0.00000417, -0.00001886, 0.00000655, 0.00000097, 0.00000033, 0.00002146, -0.00002399, -0.00000511, 0.00003250, -0.00001413, 0.00000480, 0.00002097, -0.00002723, -0.00001170, 0.00000846, -0.00000657, -0.00000477, -0.00000227, 0.00000035, -0.00000599, -0.00001182, 0.00000134, 0.00000553, -0.00000020, -0.00000399, 0.00000128, 0.00000251, -0.00001111, -0.00000619, -0.00000206, -0.00001015, -0.00000028, -0.00000141, -0.00000817, 0.00000031, 0.00000157, 0.00000113, -0.00000410, -0.00000866, -0.00000254, -0.00000306, -0.00000210, -0.00000544, -0.00000852, -0.00000400, -0.00001622, -0.00001368, -0.00002569, -0.00007729, -0.00003471, 0.00006512, 0.00006391, },
		{ 0.00003297, 0.00002485, 0.00001925, 0.00001842, 0.00001745, 0.00003056, 0.00005534, 0.00004463, -0.00000575, -0.00003526, -0.00003235, -0.00001597, -0.00000664, -0.00001823, -0.00001364, 0.00000615, 0.00000066, -0.00000868, 0.00000200, 0.00000628, -0.00000684, -0.00000366, 0.00001101, -0.00000211, -0.00000984, 0.00000361, -0.00000108, -0.00000112, 0.00000531, -0.00000906, -0.00000517, 0.00001401, 0.00001012, -0.00000174, -0.00001185, -0.00000721, 0.00001200, 0.00000356, -0.00001042, 0.00000103, 0.00000243, 0.00000064, 0.00000194, -0.00000836, 0.00000360, 0.00001292, -0.00000873, -0.00001357, -0.00000592, 0.00000769, 0.00001884, -0.00000659, -0.00001154, 0.00000882, 0.00000094, -0.00000140, 0.00000160, -0.00000289, -0.00000576, -0.00000278, 0.00001307, -0.00001680, -0.00001932, 0.00004745, -0.00000253, -0.00004422, 0.00003742, 0.00000749, -0.00003250, 0.00001512, -0.00000650, -0.00000088, 0.00002225, -0.00000653, 0.00000897, -0.00000309, -0.00001838, 0.00001298, -0.00000179, 0.00000777, 0.00000807, -0.00002917, 0.00001628, 0.00002035, -0.00001806, 0.00001971, 0.00000375, -0.00003029, 0.00000277, 0.00000381, -0.00000885, -0.00000253, -0.00000135, -0.00000086, -0.00001077, -0.00000765, 0.00000614, 0.00000329, -0.00000221, -0.00000267, 0.00000361, -0.00000221, -0.00001241, -0.00000294, -0.00000610, -0.00000726, 0.00000355, -0.00000666, -0.00000618, 0.00000370, 0.00000030, -0.00000067, -0.00000699, -0.00000685, -0.00000089, -0.00000317, -0.00000283, -0.00000876, -0.00000593, -0.00000545, -0.00002046, -0.00001298, -0.00004283, -0.00008234, -0.00000013, 0.00008107, 0.00005170, 0.00002747, },
		{ 0.00002146, 0.00002139, 0.00001832, 0.00002319, 0.00004796, 0.00005956, 0.00002020, -0.00003120, -0.00003988, -0.00002279, -0.00000828, -0.00001185, -0.00001970, -0.00000320, 0.00000626, -0.00000609, -0.00000407, 0.00000610, -0.00000062, -0.00000771, 0.00000466, 0.00000719, -0.00000922, -0.00000417, 0.00000421, -0.00000336, 0.00000359, 0.00000119, -0.00001230, 0.00000340, 0.00001529, 0.00000465, -0.00000449, -0.00001229, -0.00000149, 0.00001305, -0.00000345, -0.00000827, 0.00000523, 0.00000007, 0.00000036, -0.00000002, -0.00000749, 0.00000868, 0.00000858, -0.00001383, -0.00001278, -0.00000011, 0.00001566, 0.00000983, -0.00001402, -0.00000058, 0.00000797, -0.00000522, 0.00000230, 0.00000108, -0.00000719, -0.00000387, 0.00000704, 0.00000127, -0.00002745, 0.00001593, 0.00003948, -0.00004203, -0.00000801, 0.00004674, -0.00002919, -0.00001379, 0.00001828, -0.00001807, 0.00001773, 0.00001197, -0.00001162, 0.00001672, -0.00001613, -0.00001114, 0.00001874, -0.00000562, 0.00001017, -0.00000628, -0.00002127, 0.00002923, 0.00000377, -0.00001022, 0.00002346, -0.00001475, -0.00002059, 0.00001027, -0.00000393, -0.00000691, -0.00000114, -0.00000120, -0.00000420, -0.00001272, -0.00000057, 0.00000723, -0.00000003, -0.00000236, -0.00000051, 0.00000334, -0.00000725, -0.00001074, -0.00000253, -0.00000861, -0.00000091, 0.00000182, -0.00001095, -0.00000016, 0.00000392, -0.00000222, -0.00000206, -0.00000859, -0.00000397, -0.00000026, -0.00000379, -0.00000514, -0.00000991, -0.00000218, -0.00001100, -0.00002274, -0.00001497, -0.00006311, -0.00007334, 0.00003723, 0.00008446, 0.00004146, 0.00003200, 0.00003162, },
		{ 0.00002297, 0.00001885, 0.00003737, 0.00006345, 0.00004648, -0.00001265, -0.00004669, -0.00003223, -0.00001208, -0.00000868, -0.00001741, -0.00001341, 0.00000374, 0.00000031, -0.00000697, 0.00000237, 0.00000330, -0.00000580, -0.00000228, 0.00000788, -0.00000071, -0.00000983, 0.00000130, 0.00000152, -0.00000183, 0.00000582, -0.00000564, -0.00000946, 0.00001102, 0.00001169, 0.00000096, -0.00000623, -0.00001177, 0.00000478, 0.00001028, -0.00000893, -0.00000241, 0.00000607, -0.00000348, 0.00000178, -0.00000211, -0.00000551, 0.00001273, 0.00000212, -0.00001827, -0.00000834, 0.00000853, 0.00001576, -0.00000206, -0.00001081, 0.00000745, -0.00000001, -0.00000496, 0.00000643, -0.00000479, -0.00000769, 0.00000434, 0.00000613, -0.00001556, -0.00001256, 0.00004003, 0.00000126, -0.00004643, 0.00003524, 0.00001829, -0.00004435, 0.00001425, 0.00000161, -0.00001284, 0.00003110, -0.00000948, -0.00000172, 0.00001632, -0.00003024, 0.00000660, 0.00001725, -0.00001133, 0.00001079, -0.00001658, -0.00000677, 0.00003151, -0.00000883, 0.00000156, 0.00001555, -0.00002436, -0.00000596, 0.00000824, -0.00000860, -0.00000304, -0.00000160, -0.00000175, -0.00000842, -0.00001009, 0.00000600, 0.00000422, -0.00000179, -0.00000101, 0.00000068, 0.00000098, -0.00001060, -0.00000837, -0.00000423, -0.00000672, 0.00000397, -0.00000477, -0.00000958, 0.00000588, -0.00000024, -0.00000323, -0.00000315, -0.00000914, -0.00000051, -0.00000104, -0.00000536, -0.00000756, -0.00000778, -0.00000082, -0.00001958, -0.00002139, -0.00002247, -0.00008179, -0.00004930, 0.00006988, 0.00007823, 0.00003637, 0.00003527, 0.00002698, 0.00002209, },
		{ 0.00002559, 0.00005803, 0.00006582, 0.00001571, -0.00004207, -0.00004611, -0.00001846, -0.00000800, -0.00001388, -0.00001683, -0.00000468, 0.00000329, -0.00000464, -0.00000217, 0.00000415, -0.00000255, -0.00000492, 0.00000336, 0.00000472, -0.00000636, -0.00000595, 0.00000326, -0.00000062, 0.00000174, 0.00000336, -0.00001025, -0.00000216, 0.00001380, 0.00000678, -0.00000035, -0.00000865, -0.00000915, 0.00000997, 0.00000406, -0.00000995, 0.00000412, 0.00000221, -0.00000482, 0.00000414, -0.00000553, -0.00000130, 0.00001478, -0.00000716, -0.00001914, 0.00000073, 0.00001368, 0.00000830, -0.00000800, -0.00000185, 0.00000637, -0.00000692, 0.00000152, 0.00000431, -0.00001073, -0.00000024, 0.00000853, -0.00000604, -0.00001863, 0.00001507, 0.00003235, -0.00003473, -0.00001369, 0.00004973, -0.00002397, -0.00002660, 0.00002569, -0.00001892, 0.00001022, 0.00002351, -0.00002494, 0.00001883, -0.00000099, -0.00003188, 0.00002851, 0.00000224, -0.00001242, 0.00001129, -0.00002314, 0.00000967, 0.00002587, -0.00001532, 0.00001024, 0.00000297, -0.00002276, 0.00000454, 0.00000122, -0.00000784, -0.00000091, -0.00000290, -0.00000316, -0.00001140, -0.00000333, 0.00000838, -0.00000036, -0.00000049, -0.00000028, 0.00000047, -0.00000182, -0.00001238, -0.00000691, -0.00000494, -0.00000170, 0.00000314, -0.00001079, -0.00000228, 0.00000665, -0.00000562, -0.00000140, -0.00000547, -0.00000785, 0.00000267, -0.00000393, -0.00000710, -0.00000818, -0.00000395, -0.00000462, -0.00002749, -0.00001722, -0.00003746, -0.00009266, -0.00001304, 0.00009242, 0.00006741, 0.00003542, 0.00003444, 0.00002365, 0.00002566, 0.00002200, },
		{ 0.00007525, 0.00004636, -0.00002314, -0.00005577, -0.00003220, -0.00000849, -0.00001146, -0.00001655, -0.00001060, -0.00000031, -0.00000111, -0.00000396, 0.00000245, 0.00000031, -0.00000524, -0.00000020, 0.00000437, -0.00000062, -0.00000730, -0.00000191, 0.00000275, -0.00000028, 0.00000368, -0.00000169, -0.00000980, 0.00000494, 0.00001205, 0.00000385, -0.00000143, -0.00001132, -0.00000384, 0.00001168, -0.00000277, -0.00000585, 0.00000696, -0.00000385, -0.00000184, 0.00000425, -0.00000892, 0.00000583, 0.00001195, -0.00001697, -0.00001308, 0.00001004, 0.00001118, 0.00000025, -0.00000539, 0.00000375, -0.00000101, -0.00000624, 0.00000685, -0.00000393, -0.00000881, 0.00000912, 0.00000122, -0.00001611, -0.00000293, 0.00003030, 0.00000096, -0.00003963, 0.00002708, 0.00002569, -0.00004599, 0.00000604, 0.00001120, -0.00002122, 0.00003254, -0.00000381, -0.00001783, 0.00003132, -0.00002856, -0.00001102, 0.00003761, -0.00002114, -0.00000162, 0.00000676, -0.00002527, 0.00002643, 0.00001382, -0.00001698, 0.00001447, -0.00000748, -0.00001488, 0.00000748, -0.00000445, -0.00000426, -0.00000178, -0.00000361, -0.00000564, -0.00001072, 0.00000426, 0.00000527, -0.00000241, 0.00000226, -0.00000183, 0.00000039, -0.00000467, -0.00001377, -0.00000535, -0.00000290, 0.00000190, -0.00000260, -0.00001070, 0.00000559, 0.00000081, -0.00000723, 0.00000093, -0.00000921, -0.00000349, 0.00000329, -0.00000843, -0.00000709, -0.00000682, -0.00000168, -0.00001328, -0.00003077, -0.00001402, -0.00005914, -0.00009001, 0.00002967, 0.00010288, 0.00005604, 0.00003533, 0.00003100, 0.00002480, 0.00002802, 0.00002138, 0.00004318, },
		{ 0.00000676, -0.00005358, -0.00004948, -0.00001569, -0.00000789, -0.00001629, -0.00001308, -0.00000470, -0.00000164, -0.00000319, 0.00000037, 0.00000208, -0.00000441, -0.00000274, 0.00000295, 0.00000129, -0.00000374, -0.00000556, 0.00000034, 0.00000234, 0.00000105, 0.00000258, -0.00000546, -0.00000572, 0.00000861, 0.00000908, 0.00000303, -0.00000419, -0.00001211, 0.00000256, 0.00000910, -0.00000653, 0.00000037, 0.00000413, -0.00000733, 0.00000357, 0.00000010, -0.00000868, 0.00001320, 0.00000267, -0.00002176, -0.00000123, 0.00001306, 0.00000386, -0.00000162, 0.00000015, 0.00000163, -0.00000635, 0.00000015, 0.00000460, -0.00001022, 0.00000153, 0.00000956, -0.00001204, -0.00001113, 0.00001705, 0.00002003, -0.00002599, -0.00001308, 0.00004279, -0.00001433, -0.00003401, 0.00002455, -0.00001312, -0.00000000, 0.00003220, -0.00002971, 0.00000964, 0.00001899, -0.00004459, 0.00002420, 0.00002038, -0.00003516, 0.00001745, -0.00000831, -0.00001715, 0.00003938, -0.00000453, -0.00001299, 0.00001572, -0.00001402, -0.00000643, 0.00000538, -0.00000608, -0.00000178, -0.00000405, -0.00000376, -0.00000818, -0.00000563, 0.00000807, -0.00000042, 0.00000028, 0.00000262, -0.00000434, 0.00000164, -0.00000901, -0.00001419, -0.00000172, -0.00000013, 0.00000091, -0.00000782, -0.00000430, 0.00000752, -0.00000699, -0.00000328, -0.00000021, -0.00001146, 0.00000293, -0.00000098, -0.00001156, -0.00000456, -0.00000554, -0.00000313, -0.00002362, -0.00002792, -0.00001669, -0.00008329, -0.00007052, 0.00007204, 0.00010192, 0.00004551, 0.00003426, 0.00002912, 0.00002881, 0.00002674, 0.00002870, 0.00006859, 0.00007434, },
		{ -0.00006225, -0.00003094, -0.00000731, -0.00001348, -0.00001596, -0.00000674, -0.00000376, -0.00000418, -0.00000063, 0.00000280, -0.00000235, -0.00000510, 0.00000147, 0.00000184, -0.00000192, -0.00000399, -0.00000386, 0.00000180, 0.00000265, 0.00000116, 0.00000025, -0.00000621, -0.00000145, 0.00000934, 0.00000746, 0.00000204, -0.00000808, -0.00000965, 0.00000724, 0.00000428, -0.00000574, 0.00000413, -0.00000192, -0.00000540, 0.00000677, -0.00000625, -0.00000283, 0.00001565, -0.00000972, -0.00001720, 0.00000964, 0.00000772, -0.00000073, 0.00000239, 0.00000117, -0.00000412, -0.00000479, 0.00000438, -0.00000308, -0.00000699, 0.00001077, -0.00000201, -0.00001694, 0.00000663, 0.00002179, -0.00000465, -0.00002690, 0.00001910, 0.00002481, -0.00003812, -0.00000403, 0.00001625, -0.00002226, 0.00002647, 0.00000604, -0.00003105, 0.00003476, -0.00001422, -0.00003054, 0.00004825, -0.00001674, -0.00002393, 0.00002830, -0.00002895, 0.00000589, 0.00003907, -0.00002458, -0.00000104, 0.00001299, -0.00001756, 0.00000050, 0.00000182, -0.00000506, -0.00000191, -0.00000572, -0.00000416, -0.00000851, 0.00000127, 0.00000581, -0.00000328, 0.00000520, -0.00000165, -0.00000414, 0.00000235, -0.00001547, -0.00001081, 0.00000319, -0.00000037, -0.00000283, -0.00000782, 0.00000280, 0.00000207, -0.00001013, 0.00000240, -0.00000573, -0.00000854, 0.00000693, -0.00000917, -0.00000963, -0.00000188, -0.00000632, -0.00000794, -0.00003197, -0.00002066, -0.00002889, -0.00010325, -0.00003395, 0.00010769, 0.00009083, 0.00003607, 0.00003389, 0.00003099, 0.00003084, 0.00002599, 0.00004972, 0.00008802, 0.00004577, -0.00003796, },
		{ -0.00001442, -0.00000938, -0.00001719, -0.00001006, -0.00000347, -0.00000646, -0.00000319, 0.00000325, 0.00000048, -0.00000591, -0.00000164, 0.00000287, -0.00000152, -0.00000264, -0.00000352, -0.00000254, 0.00000363, 0.00000220, 0.00000007, -0.00000097, -0.00000540, 0.00000124, 0.00000957, 0.00000684, -0.00000075, -0.00001083, -0.00000490, 0.00000873, 0.00000037, -0.00000248, 0.00000317, -0.00000633, 0.00000021, 0.00000477, -0.00000995, 0.00000595, 0.00001015, -0.00001803, -0.00000528, 0.00001224, -0.00000055, 0.00000205, 0.00000550, -0.00000376, -0.00000622, 0.00000027, 0.00000162, -0.00000731, 0.00000359, 0.00000844, -0.00001548, -0.00000549, 0.00002049, 0.00000661, -0.00002016, -0.00000537, 0.00003023, -0.00000712, -0.00003216, 0.00001632, -0.00000549, -0.00000716, 0.00003344, -0.00002528, -0.00000504, 0.00003372, -0.00004285, 0.00000868, 0.00003842, -0.00004710, 0.00000865, 0.00001561, -0.00003794, 0.00003654, 0.00001813, -0.00003460, 0.00001610, 0.00000340, -0.00001699, 0.00000574, -0.00000186, -0.00000341, -0.00000385, -0.00000629, -0.00000450, -0.00000556, 0.00000537, 0.00000021, -0.00000015, 0.00000665, -0.00000782, 0.00000060, -0.00000157, -0.00002074, -0.00000206, 0.00000499, -0.00000417, -0.00000436, -0.00000350, 0.00000464, -0.00000571, -0.00000637, 0.00000394, -0.00001142, -0.00000070, 0.00000376, -0.00001573, -0.00000258, -0.00000312, -0.00000850, -0.00001470, -0.00003592, -0.00001285, -0.00005174, -0.00011128, 0.00001699, 0.00013040, 0.00007116, 0.00002988, 0.00003705, 0.00003363, 0.00002913, 0.00003541, 0.00007939, 0.00008291, -0.00000468, -0.00006689, -0.00004888, },
		{ -0.00001503, -0.00001430, -0.00000398, -0.00000540, -0.00000781, 0.00000129, 0.00000382, -0.00000423, -0.00000524, 0.00000201, 0.00000048, -0.00000349, -0.00000175, -0.00000373, -0.00000033, 0.00000495, 0.00000018, -0.00000030, -0.00000106, -0.00000508, 0.00000339, 0.00001049, 0.00000528, -0.00000466, -0.00001099, -0.00000006, 0.00000796, -0.00000118, -0.00000024, -0.00000073, -0.00000615, 0.00000488, -0.00000092, -0.00000791, 0.00001187, -0.00000043, -0.00001696, 0.00000598, 0.00000571, -0.00000334, 0.00000888, 0.00000194, -0.00000886, -0.00000256, 0.00000166, -0.00000415, -0.00000231, 0.00001053, -0.00000502, -0.00001632, 0.00001391, 0.00001585, -0.00001344, -0.00001354, 0.00001594, 0.00001579, -0.00002618, -0.00000962, 0.00001382, -0.00001704, 0.00001722, 0.00001357, -0.00003554, 0.00002652, 0.00000475, -0.00004131, 0.00004409, -0.00000247, -0.00004476, 0.00003735, -0.00001669, -0.00002011, 0.00005441, -0.00001696, -0.00002391, 0.00002751, -0.00001140, -0.00000974, 0.00000743, -0.00000509, -0.00000188, -0.00000674, -0.00000567, -0.00000348, -0.00000140, 0.00000429, -0.00000288, 0.00000615, 0.00000132, -0.00000979, 0.00000626, -0.00001156, -0.00001880, 0.00000820, 0.00000002, -0.00000693, -0.00000155, -0.00000025, 0.00000081, -0.00000938, 0.00000015, -0.00000045, -0.00001175, 0.00000637, -0.00000618, -0.00001435, 0.00000408, -0.00000948, -0.00000955, -0.00002297, -0.00003398, -0.00000964, -0.00008319, -0.00009925, 0.00007578, 0.00013393, 0.00004711, 0.00003130, 0.00004263, 0.00003210, 0.00003022, 0.00006075, 0.00009927, 0.00004405, -0.00005529, -0.00006752, -0.00002695, -0.00000946, },
		{ -0.00000752, -0.00000276, -0.00000909, -0.00000483, 0.00000513, 0.00000015, -0.00000689, -0.00000158, 0.00000240, -0.00000288, -0.00000283, -0.00000148, -0.00000366, 0.00000290, 0.00000404, -0.00000193, 0.00000110, -0.00000195, -0.00000516, 0.00000667, 0.00001091, 0.00000194, -0.00000779, -0.00000903, 0.00000364, 0.00000668, -0.00000159, -0.00000061, -0.00000385, -0.00000255, 0.00000549, -0.00000588, -0.00000180, 0.00001119, -0.00000911, -0.00000825, 0.00000900, -0.00000292, 0.00000254, 0.00001157, -0.00000661, -0.00000802, 0.00000164, -0.00000270, -0.00000454, 0.00000726, 0.00000515, -0.00001682, -0.00000081, 0.00002308, -0.00000443, -0.00001855, 0.00000612, 0.00001825, -0.00000654, -0.00002252, 0.00000655, -0.00000147, -0.00000816, 0.00002753, -0.00001621, -0.00001714, 0.00003683, -0.00002850, -0.00000854, 0.00004608, -0.00004437, -0.00000914, 0.00003614, -0.00004347, 0.00001945, 0.00004133, -0.00004444, 0.00000431, 0.00002190, -0.00002199, 0.00000244, 0.00000310, -0.00000572, -0.00000150, -0.00000994, -0.00000283, -0.00000136, 0.00000036, 0.00000066, -0.00000015, 0.00000896, -0.00000753, -0.00000428, 0.00000582, -0.00002235, -0.00000707, 0.00001189, -0.00000902, -0.00000350, 0.00000172, -0.00000160, -0.00000372, -0.00000742, 0.00000355, -0.00000662, -0.00000590, 0.00000617, -0.00001585, -0.00000451, 0.00000307, -0.00001662, -0.00000896, -0.00003263, -0.00002520, -0.00001751, -0.00011647, -0.00006056, 0.00013041, 0.00011497, 0.00002784, 0.00004215, 0.00004419, 0.00002744, 0.00004543, 0.00009329, 0.00008771, -0.00001720, -0.00008120, -0.00004553, -0.00001309, -0.00001390, -0.00001588, },
		{ -0.00000598, -0.00000983, 0.00000058, 0.00000493, -0.00000457, -0.00000567, 0.00000131, -0.00000019, -0.00000425, -0.00000124, -0.00000235, -0.00000183, 0.00000525, 0.00000092, -0.00000181, 0.00000273, -0.00000474, -0.00000360, 0.00001075, 0.00000914, -0.00000203, -0.00000919, -0.00000615, 0.00000650, 0.00000532, -0.00000270, -0.00000219, -0.00000419, 0.00000107, 0.00000279, -0.00000714, 0.00000372, 0.00000568, -0.00001081, 0.00000055, 0.00000343, -0.00000499, 0.00001168, 0.00000551, -0.00001261, -0.00000199, 0.00000023, -0.00000683, 0.00000323, 0.00001031, -0.00000877, -0.00001374, 0.00001845, 0.00001164, -0.00002176, -0.00000377, 0.00001814, 0.00000319, -0.00001669, -0.00000771, 0.00000594, -0.00001029, 0.00001004, 0.00001478, -0.00003115, 0.00001267, 0.00001869, -0.00003809, 0.00002888, 0.00001150, -0.00005420, 0.00003155, 0.00000337, -0.00003886, 0.00005351, 0.00000068, -0.00004298, 0.00003047, 0.00000037, -0.00001831, 0.00001093, -0.00000521, -0.00000188, -0.00000480, -0.00001103, 0.00000264, -0.00000111, -0.00000094, -0.00000040, 0.00000542, 0.00000444, -0.00001237, 0.00000458, -0.00000393, -0.00002469, 0.00000849, 0.00000409, -0.00001345, 0.00000495, -0.00000040, -0.00000486, -0.00000475, -0.00000375, 0.00000240, -0.00000922, 0.00000076, -0.00000185, -0.00001702, 0.00000627, -0.00000741, -0.00001800, -0.00001059, -0.00004077, -0.00001142, -0.00004266, -0.00013840, 0.00000492, 0.00016441, 0.00007879, 0.00002424, 0.00005613, 0.00003638, 0.00003069, 0.00007829, 0.00010959, 0.00003776, -0.00007211, -0.00007239, -0.00002068, -0.00001367, -0.00001768, -0.00001032, -0.00000315, },
		{ -0.00000658, 0.00000436, 0.00000127, -0.00000686, -0.00000257, 0.00000129, -0.00000297, -0.00000332, -0.00000068, -0.00000284, 0.00000137, 0.00000484, -0.00000202, 0.00000085, 0.00000201, -0.00000787, 0.00000094, 0.00001340, 0.00000516, -0.00000503, -0.00000931, -0.00000274, 0.00000887, 0.00000284, -0.00000455, -0.00000274, -0.00000277, 0.00000265, -0.00000030, -0.00000551, 0.00000575, 0.00000057, -0.00000689, 0.00000289, -0.00000361, 0.00000175, 0.00001515, -0.00000545, -0.00001077, 0.00000227, -0.00000600, -0.00000332, 0.00001190, 0.00000123, -0.00001732, 0.00000391, 0.00002332, -0.00001183, -0.00001942, 0.00001644, 0.00001106, -0.00001201, -0.00001060, 0.00000094, -0.00000318, -0.00000445, 0.00001865, -0.00000855, -0.00002148, 0.00002898, -0.00000978, -0.00001772, 0.00003972, -0.00003199, -0.00002510, 0.00004396, -0.00003383, -0.00000173, 0.00005380, -0.00003997, -0.00001214, 0.00003345, -0.00002045, -0.00000150, 0.00000718, -0.00000951, 0.00000329, -0.00001253, -0.00000585, 0.00000733, -0.00000511, -0.00000141, 0.00000273, 0.00000755, -0.00000405, -0.00000894, 0.00000817, -0.00001735, -0.00001404, 0.00001626, -0.00001019, -0.00000657, 0.00000999, -0.00000809, -0.00000438, -0.00000373, -0.00000199, 0.00000011, -0.00000771, 0.00000228, -0.00001088, -0.00000831, 0.00000816, -0.00002062, -0.00001198, -0.00001914, -0.00004099, -0.00000068, -0.00008541, -0.00013177, 0.00008500, 0.00016432, 0.00004169, 0.00003984, 0.00006050, 0.00002488, 0.00005514, 0.00011370, 0.00008676, -0.00003217, -0.00009333, -0.00004174, -0.00001077, -0.00002097, -0.00001370, -0.00000405, -0.00000407, -0.00000910, },
		{ 0.00000464, -0.00000276, -0.00000621, -0.00000049, -0.00000067, -0.00000394, -0.00000187, -0.00000113, -0.00000166, 0.00000380, 0.00000207, -0.00000235, 0.00000372, -0.00000179, -0.00000842, 0.00000719, 0.00001278, 0.00000067, -0.00000666, -0.00000869, 0.00000158, 0.00000980, -0.00000130, -0.00000541, -0.00000190, -0.00000158, 0.00000286, -0.00000227, -0.00000366, 0.00000558, -0.00000092, -0.00000323, -0.00000092, -0.00000447, 0.00001113, 0.00000913, -0.00001280, -0.00000381, -0.00000054, -0.00000920, 0.00000737, 0.00001096, -0.00001279, -0.00001002, 0.00002102, 0.00000757, -0.00002684, 0.00000126, 0.00002263, -0.00000731, -0.00001316, -0.00000023, -0.00000210, -0.00000666, 0.00000733, 0.00001026, -0.00002243, 0.00000085, 0.00002222, -0.00002368, 0.00001205, 0.00001624, -0.00004984, 0.00001726, 0.00001967, -0.00004281, 0.00003963, 0.00001751, -0.00005008, 0.00002357, 0.00001086, -0.00002195, 0.00001344, -0.00000660, -0.00000319, 0.00000180, -0.00001924, 0.00000651, 0.00000478, -0.00001023, 0.00000295, 0.00000569, 0.00000382, -0.00000926, -0.00000112, 0.00000206, -0.00002366, 0.00000320, 0.00000928, -0.00001829, 0.00000788, 0.00000392, -0.00001378, 0.00000074, -0.00000515, -0.00000079, -0.00000059, -0.00000636, -0.00000138, -0.00001330, 0.00000274, -0.00000193, -0.00002655, -0.00000488, -0.00003363, -0.00002865, -0.00000625, -0.00013402, -0.00008479, 0.00015545, 0.00013068, 0.00002385, 0.00006367, 0.00004775, 0.00002783, 0.00009991, 0.00012229, 0.00002335, -0.00008714, -0.00007528, -0.00001470, -0.00001769, -0.00002225, -0.00000461, -0.00000306, -0.00000726, -0.00000890, -0.00000218, },
		{ -0.00000512, -0.00000451, -0.00000034, -0.00000242, -0.00000341, -0.00000115, -0.00000120, 0.00000054, 0.00000388, -0.00000060, -0.00000003, 0.00000399, -0.00000650, -0.00000495, 0.00001224, 0.00000920, -0.00000268, -0.00000762, -0.00000690, 0.00000639, 0.00000776, -0.00000560, -0.00000406, -0.00000124, -0.00000097, 0.00000305, -0.00000403, -0.00000236, 0.00000622, -0.00000062, -0.00000392, -0.00000462, 0.00000189, 0.00001480, -0.00000161, -0.00001192, 0.00000040, -0.00000741, -0.00000303, 0.00001559, -0.00000166, -0.00001764, 0.00000872, 0.00002124, -0.00001630, -0.00001989, 0.00002216, 0.00000889, -0.00001918, -0.00000212, 0.00000164, -0.00000830, -0.00000032, 0.00001135, -0.00000571, -0.00001832, 0.00001628, 0.00000433, -0.00001490, 0.00002394, -0.00001957, -0.00003093, 0.00003810, -0.00001730, -0.00001575, 0.00005119, -0.00002718, -0.00002359, 0.00003477, -0.00001615, -0.00000264, 0.00001065, -0.00001608, 0.00000937, -0.00001092, -0.00001501, 0.00001845, -0.00000747, -0.00000863, 0.00000976, 0.00000328, -0.00000129, -0.00000815, 0.00000306, -0.00000865, -0.00001725, 0.00001389, -0.00000674, -0.00001137, 0.00001701, -0.00001076, -0.00000963, 0.00000406, -0.00000925, 0.00000352, -0.00000217, -0.00000788, -0.00000450, -0.00000846, 0.00000650, -0.00001714, -0.00002125, -0.00000575, -0.00004527, -0.00000828, -0.00003801, -0.00016547, -0.00000236, 0.00019125, 0.00008261, 0.00003425, 0.00007546, 0.00002824, 0.00006172, 0.00014033, 0.00008202, -0.00005310, -0.00010019, -0.00003827, -0.00000880, -0.00002801, -0.00001274, 0.00000093, -0.00000697, -0.00000807, -0.00000649, 0.00000054, 0.00000278, },
		{ -0.00000330, -0.00000104, -0.00000308, -0.00000265, -0.00000079, -0.00000025, 0.00000200, 0.00000227, -0.00000120, 0.00000261, 0.00000109, -0.00000903, 0.00000117, 0.00001398, 0.00000457, -0.00000469, -0.00000820, -0.00000323, 0.00000976, 0.00000275, -0.00000756, -0.00000158, -0.00000192, 0.00000039, 0.00000279, -0.00000657, 0.00000031, 0.00000805, -0.00000264, -0.00000712, -0.00000285, 0.00000946, 0.00001014, -0.00000896, -0.00000647, -0.00000243, -0.00000959, 0.00000933, 0.00001173, -0.00001566, -0.00000634, 0.00002244, 0.00000280, -0.00002804, 0.00000318, 0.00002529, -0.00001241, -0.00001455, 0.00000769, -0.00000559, -0.00000745, 0.00000755, 0.00000378, -0.00001475, -0.00000449, 0.00001658, -0.00000623, 0.00000224, 0.00000955, -0.00003728, 0.00000439, 0.00002475, -0.00003345, 0.00002255, 0.00002431, -0.00004477, 0.00001416, 0.00001441, -0.00002152, 0.00001764, -0.00000910, -0.00000801, 0.00001332, -0.00002703, 0.00000425, 0.00001715, -0.00002135, 0.00000378, 0.00001054, -0.00000292, -0.00000188, -0.00000534, 0.00000054, -0.00001427, -0.00000404, 0.00001025, -0.00001729, 0.00000586, 0.00001054, -0.00002153, 0.00000268, -0.00000128, -0.00000955, 0.00000973, -0.00000891, -0.00000874, -0.00000308, -0.00000300, 0.00000033, -0.00002716, -0.00001050, -0.00001726, -0.00004420, 0.00000532, -0.00009176, -0.00015720, 0.00009131, 0.00018323, 0.00004762, 0.00006093, 0.00006373, 0.00002824, 0.00012031, 0.00013955, 0.00000079, -0.00010248, -0.00007286, -0.00001166, -0.00002070, -0.00002722, 0.00000027, -0.00000194, -0.00001000, -0.00000582, -0.00000490, 0.00000166, 0.00000072, -0.00000617, },
		{ -0.00000160, -0.00000321, -0.00000200, 0.00000003, 0.00000084, 0.00000197, 0.00000090, 0.00000001, 0.00000335, -0.00000302, -0.00000795, 0.00000714, 0.00001251, 0.00000053, -0.00000603, -0.00000768, 0.00000174, 0.00000969, -0.00000293, -0.00000624, -0.00000033, -0.00000288, 0.00000304, 0.00000024, -0.00000833, 0.00000584, 0.00000775, -0.00000798, -0.00000746, 0.00000350, 0.00001170, 0.00000199, -0.00000959, -0.00000377, -0.00000801, -0.00000206, 0.00001676, -0.00000307, -0.00001731, 0.00001276, 0.00001781, -0.00001920, -0.00001804, 0.00002284, 0.00000885, -0.00002335, 0.00000045, 0.00000634, -0.00001215, 0.00000063, 0.00000731, -0.00000696, -0.00001184, 0.00000564, 0.00000944, -0.00000353, 0.00000775, -0.00001465, -0.00002569, 0.00002526, -0.00000409, -0.00001720, 0.00003793, -0.00001581, -0.00002454, 0.00002912, -0.00001466, -0.00000019, 0.00001578, -0.00002572, 0.00001368, -0.00000234, -0.00002905, 0.00002755, -0.00000294, -0.00002189, 0.00001926, 0.00000010, -0.00000437, 0.00000083, -0.00000686, -0.00000301, -0.00001118, 0.00000413, -0.00000205, -0.00001272, 0.00001809, -0.00000793, -0.00001742, 0.00001139, -0.00001286, -0.00000005, 0.00000978, -0.00001896, -0.00000309, -0.00000055, -0.00000290, -0.00000997, -0.00002760, -0.00000375, -0.00003261, -0.00002934, -0.00000354, -0.00014714, -0.00010278, 0.00016524, 0.00014589, 0.00004150, 0.00007897, 0.00004149, 0.00006823, 0.00016755, 0.00007762, -0.00008204, -0.00010183, -0.00003313, -0.00001018, -0.00003236, -0.00001351, 0.00000574, -0.00000920, -0.00000796, -0.00000406, -0.00000472, 0.00000284, -0.00000146, -0.00000685, -0.00000241, },
		{ -0.00000317, -0.00000095, 0.00000111, 0.00000102, 0.00000138, 0.00000075, 0.00000124, 0.00000196, -0.00000595, -0.00000423, 0.00001106, 0.00000923, -0.00000256, -0.00000685, -0.00000532, 0.00000592, 0.00000612, -0.00000637, -0.00000336, -0.00000127, -0.00000196, 0.00000507, -0.00000489, -0.00000608, 0.00001202, 0.00000216, -0.00001263, -0.00000228, 0.00000863, 0.00000795, -0.00000322, -0.00000741, -0.00000645, -0.00000801, 0.00001001, 0.00001107, -0.00001629, -0.00000319, 0.00002274, -0.00000235, -0.00002642, 0.00000442, 0.00002376, -0.00001259, -0.00001686, 0.00001148, -0.00000349, -0.00001012, 0.00000710, -0.00000115, -0.00001079, -0.00000386, 0.00000752, 0.00000635, 0.00000220, -0.00000348, -0.00002536, -0.00000045, 0.00001910, -0.00001900, 0.00001168, 0.00001908, -0.00003367, 0.00000985, 0.00001052, -0.00002133, 0.00002505, -0.00001027, -0.00001820, 0.00002739, -0.00002961, -0.00000640, 0.00003363, -0.00002960, -0.00000180, 0.00002180, -0.00001422, 0.00000435, -0.00000172, -0.00001080, -0.00000061, -0.00000624, 0.00000216, -0.00000960, 0.00000188, 0.00001393, -0.00002294, -0.00000108, 0.00000617, -0.00001861, 0.00001504, -0.00000304, -0.00002237, 0.00000803, -0.00000404, -0.00000652, -0.00001775, -0.00002216, -0.00000512, -0.00004223, -0.00001091, -0.00003892, -0.00017899, -0.00001735, 0.00020138, 0.00010632, 0.00005588, 0.00007393, 0.00004086, 0.00013703, 0.00015813, -0.00002534, -0.00012164, -0.00006332, -0.00001155, -0.00002452, -0.00002943, 0.00000191, 0.00000037, -0.00001265, -0.00000308, -0.00000562, -0.00000355, 0.00000402, -0.00000470, -0.00000678, -0.00000133, -0.00000222, },
		{ 0.00000056, 0.00000146, 0.00000051, 0.00000144, 0.00000111, 0.00000136, -0.00000011, -0.00000682, 0.00000030, 0.00001266, 0.00000531, -0.00000487, -0.00000654, -0.00000172, 0.00000738, 0.00000130, -0.00000659, -0.00000153, -0.00000278, 0.00000125, 0.00000370, -0.00000955, 0.00000127, 0.00001378, -0.00000724, -0.00001145, 0.00000511, 0.00000831, 0.00000308, -0.00000425, -0.00000801, -0.00000949, 0.00000051, 0.00001543, -0.00000395, -0.00001534, 0.00001523, 0.00001383, -0.00002124, -0.00001384, 0.00002037, 0.00000752, -0.00002245, -0.00000119, 0.00001042, -0.00001132, -0.00000184, 0.00000485, -0.00000924, -0.00000645, 0.00000057, 0.00000717, 0.00000888, -0.00000158, -0.00001813, -0.00001522, 0.00001418, 0.00000026, -0.00000889, 0.00002319, -0.00001264, -0.00001683, 0.00002348, -0.00001891, 0.00000281, 0.00002489, -0.00003616, 0.00001226, 0.00001354, -0.00004347, 0.00002933, 0.00001011, -0.00003863, 0.00002658, 0.00000305, -0.00001611, 0.00001564, -0.00001449, -0.00000731, 0.00000535, -0.00000827, -0.00000264, -0.00000568, 0.00001182, -0.00000258, -0.00002187, 0.00001251, -0.00000977, -0.00000879, 0.00002155, -0.00002344, -0.00001044, 0.00001390, -0.00001424, -0.00000756, -0.00002266, -0.00001586, -0.00001174, -0.00004284, -0.00000201, -0.00008995, -0.00017322, 0.00007361, 0.00020397, 0.00008221, 0.00006814, 0.00005963, 0.00008428, 0.00018801, 0.00007476, -0.00011606, -0.00010209, -0.00002329, -0.00001671, -0.00003381, -0.00001391, 0.00000722, -0.00000875, -0.00000895, -0.00000114, -0.00000863, 0.00000036, 0.00000293, -0.00000837, -0.00000493, -0.00000067, -0.00000319, -0.00000247, },
		{ 0.00000084, 0.00000045, 0.00000207, 0.00000080, 0.00000087, -0.00000168, -0.00000626, 0.00000458, 0.00001240, 0.00000117, -0.00000611, -0.00000479, 0.00000140, 0.00000608, -0.00000225, -0.00000509, -0.00000169, -0.00000249, 0.00000435, -0.00000138, -0.00000948, 0.00001013, 0.00000800, -0.00001443, -0.00000389, 0.00000876, 0.00000424, 0.00000136, -0.00000527, -0.00001159, -0.00000609, 0.00001089, 0.00000823, -0.00001493, -0.00000006, 0.00002149, -0.00000702, -0.00002339, 0.00000644, 0.00001838, -0.00001088, -0.00001634, 0.00000999, 0.00000136, -0.00001069, 0.00000371, -0.00000376, -0.00000969, -0.00000085, 0.00000074, 0.00001084, 0.00000824, -0.00001483, -0.00002000, 0.00000245, 0.00000935, -0.00000810, 0.00001027, 0.00000723, -0.00002516, 0.00001322, 0.00000401, -0.00002549, 0.00003373, -0.00000647, -0.00003313, 0.00003897, -0.00002294, -0.00002406, 0.00004860, -0.00002982, -0.00001550, 0.00003740, -0.00002459, 0.00000282, 0.00001273, -0.00002815, 0.00000898, 0.00000300, -0.00001542, 0.00000038, 0.00000237, 0.00000846, -0.00001618, -0.00000760, 0.00001088, -0.00002100, 0.00001159, 0.00000867, -0.00003461, 0.00001205, 0.00000393, -0.00002159, -0.00000490, -0.00002830, -0.00000875, -0.00002051, -0.00003812, -0.00000848, -0.00014039, -0.00013198, 0.00015230, 0.00018903, 0.00006982, 0.00006721, 0.00006892, 0.00015527, 0.00016928, -0.00004938, -0.00014515, -0.00004919, -0.00001124, -0.00003194, -0.00002679, 0.00000041, 0.00000151, -0.00001240, -0.00000270, -0.00000451, -0.00000849, 0.00000484, -0.00000179, -0.00000988, -0.00000185, -0.00000153, -0.00000369, -0.00000077, 0.00000157, },
		{ 0.00000131, 0.00000213, -0.00000015, 0.00000080, -0.00000290, -0.00000490, 0.00000845, 0.00001042, -0.00000272, -0.00000568, -0.00000244, 0.00000274, 0.00000373, -0.00000378, -0.00000400, -0.00000238, 0.00000000, 0.00000436, -0.00000687, -0.00000342, 0.00001444, -0.00000269, -0.00001404, 0.00000479, 0.00000630, 0.00000163, 0.00000155, -0.00000980, -0.00001236, 0.00000389, 0.00001303, -0.00000555, -0.00001142, 0.00001617, 0.00000946, -0.00002226, -0.00000851, 0.00001714, 0.00000334, -0.00001779, -0.00000318, 0.00001049, -0.00000601, -0.00000490, 0.00000119, -0.00001019, -0.00000385, 0.00000025, 0.00000257, 0.00001595, -0.00000260, -0.00002497, -0.00000708, 0.00000971, -0.00000260, 0.00000144, 0.00001433, -0.00001749, -0.00000742, 0.00002322, -0.00002662, 0.00000162, 0.00003755, -0.00004268, 0.00000252, 0.00003323, -0.00005175, 0.00002048, 0.00002922, -0.00005306, 0.00002630, 0.00001478, -0.00003431, 0.00002914, -0.00001205, -0.00002390, 0.00002704, -0.00001481, -0.00001425, 0.00001047, 0.00000219, -0.00000164, -0.00001668, 0.00000501, -0.00000283, -0.00001532, 0.00002477, -0.00001750, -0.00002317, 0.00002669, -0.00001876, -0.00001570, -0.00000665, -0.00003310, 0.00000085, -0.00003335, -0.00003080, -0.00002886, -0.00017990, -0.00006370, 0.00021468, 0.00016466, 0.00005627, 0.00006958, 0.00011971, 0.00019961, 0.00006787, -0.00014761, -0.00010527, -0.00000856, -0.00002647, -0.00003505, -0.00001082, 0.00000348, -0.00000615, -0.00000883, -0.00000065, -0.00000936, -0.00000359, 0.00000556, -0.00000789, -0.00000751, 0.00000005, -0.00000351, -0.00000253, 0.00000100, 0.00000132, 0.00000021, },
		{ 0.00000101, -0.00000059, 0.00000122, -0.00000444, -0.00000253, 0.00001159, 0.00000667, -0.00000535, -0.00000375, -0.00000095, 0.00000267, 0.00000192, -0.00000440, -0.00000380, -0.00000177, 0.00000256, 0.00000089, -0.00000850, 0.00000499, 0.00001094, -0.00001141, -0.00000653, 0.00000833, 0.00000136, 0.00000295, -0.00000108, -0.00001530, -0.00000547, 0.00001239, 0.00000396, -0.00001292, 0.00000370, 0.00001854, -0.00001073, -0.00001973, 0.00000912, 0.00001134, -0.00001026, -0.00001196, 0.00000561, 0.00000478, -0.00000700, -0.00000182, -0.00000591, -0.00000906, 0.00000130, -0.00000141, 0.00000930, 0.00001388, -0.00001908, -0.00002094, 0.00000753, 0.00000266, -0.00000432, 0.00001458, -0.00000310, -0.00002356, 0.00002051, 0.00000124, -0.00003423, 0.00003866, 0.00000431, -0.00004861, 0.00004227, -0.00000637, -0.00004312, 0.00005530, -0.00001935, -0.00003471, 0.00005142, -0.00002736, -0.00000998, 0.00003478, -0.00004270, 0.00000634, 0.00002483, -0.00003658, 0.00000462, 0.00001330, -0.00000726, -0.00000529, -0.00000936, 0.00000569, -0.00001346, 0.00000263, 0.00001722, -0.00003620, 0.00000531, 0.00001736, -0.00003606, 0.00000064, -0.00002070, -0.00002767, 0.00000725, -0.00005176, -0.00001801, -0.00006378, -0.00020302, 0.00002877, 0.00025845, 0.00012551, 0.00004440, 0.00010412, 0.00018647, 0.00016452, -0.00007063, -0.00016752, -0.00003675, -0.00000747, -0.00004339, -0.00002067, -0.00000126, -0.00000166, -0.00000779, -0.00000412, -0.00000404, -0.00001013, 0.00000235, 0.00000092, -0.00001094, -0.00000296, -0.00000103, -0.00000444, 0.00000002, 0.00000136, 0.00000048, 0.00000059, 0.00000221, },
		{ 0.00000032, 0.00000103, -0.00000616, 0.00000132, 0.00001304, 0.00000181, -0.00000575, -0.00000161, -0.00000088, 0.00000256, 0.00000067, -0.00000522, -0.00000333, 0.00000014, 0.00000285, -0.00000325, -0.00000506, 0.00001010, 0.00000226, -0.00001286, 0.00000210, 0.00000539, -0.00000053, 0.00000497, -0.00000830, -0.00001516, 0.00000620, 0.00001096, -0.00000818, -0.00000647, 0.00001640, 0.00000455, -0.00002193, -0.00000330, 0.00001442, -0.00000289, -0.00001249, -0.00000249, 0.00000650, 0.00000046, -0.00000519, -0.00000477, -0.00001001, -0.00000295, 0.00000163, 0.00000001, 0.00001600, 0.00000080, -0.00002859, -0.00000521, 0.00001048, -0.00000662, 0.00000710, 0.00001257, -0.00002442, -0.00000302, 0.00002820, -0.00003181, -0.00000639, 0.00004938, -0.00004039, -0.00001424, 0.00004991, -0.00004850, 0.00000226, 0.00004786, -0.00005858, 0.00001564, 0.00003309, -0.00005203, 0.00003308, 0.00000394, -0.00004802, 0.00004389, -0.00000789, -0.00003734, 0.00003021, -0.00000273, -0.00001191, 0.00000006, -0.00000696, -0.00000029, -0.00001103, 0.00001560, -0.00000551, -0.00003102, 0.00002799, -0.00001217, -0.00003113, 0.00001007, -0.00004112, -0.00000583, -0.00000329, -0.00006630, 0.00000027, -0.00012007, -0.00019704, 0.00014424, 0.00026902, 0.00007075, 0.00006089, 0.00017497, 0.00020970, 0.00004642, -0.00016991, -0.00011102, 0.00000606, -0.00003451, -0.00003902, -0.00000335, -0.00000394, -0.00000458, -0.00000480, -0.00000402, -0.00000782, -0.00000603, 0.00000408, -0.00000558, -0.00000886, -0.00000015, -0.00000387, -0.00000263, 0.00000189, 0.00000021, 0.00000050, 0.00000160, 0.00000199, -0.00000047, },
		{ -0.00000075, -0.00000686, 0.00000628, 0.00001186, -0.00000266, -0.00000409, -0.00000074, -0.00000133, 0.00000307, -0.00000100, -0.00000593, -0.00000166, 0.00000157, 0.00000087, -0.00000481, 0.00000048, 0.00000911, -0.00000542, -0.00000771, 0.00000590, 0.00000039, 0.00000202, 0.00000255, -0.00001545, -0.00000625, 0.00001342, -0.00000010, -0.00001148, 0.00000798, 0.00001455, -0.00001359, -0.00001555, 0.00001152, 0.00000478, -0.00001181, -0.00000555, 0.00000203, 0.00000433, -0.00000031, -0.00000638, -0.00000972, -0.00000761, 0.00000181, -0.00000023, 0.00000615, 0.00001458, -0.00001635, -0.00002328, 0.00000900, 0.00000182, -0.00000510, 0.00001787, -0.00000689, -0.00002671, 0.00002490, 0.00000517, -0.00004291, 0.00003492, 0.00001988, -0.00005794, 0.00003400, 0.00001576, -0.00005565, 0.00004938, -0.00000049, -0.00005280, 0.00005693, -0.00001876, -0.00003152, 0.00005572, -0.00004513, -0.00001128, 0.00005303, -0.00005121, -0.00000393, 0.00003555, -0.00002806, 0.00000099, 0.00000133, -0.00001191, -0.00000026, -0.00000212, 0.00001254, -0.00002343, -0.00000686, 0.00002523, -0.00003890, -0.00000691, -0.00000267, -0.00004818, 0.00002130, -0.00003789, -0.00005758, 0.00000757, -0.00019779, -0.00013745, 0.00026464, 0.00022670, 0.00002650, 0.00013062, 0.00023940, 0.00014361, -0.00009639, -0.00017940, -0.00003113, -0.00000115, -0.00005485, -0.00001530, 0.00000012, -0.00000992, -0.00000072, -0.00000438, -0.00000703, -0.00000735, -0.00000176, 0.00000071, -0.00000878, -0.00000429, -0.00000121, -0.00000511, 0.00000086, 0.00000130, -0.00000073, 0.00000178, 0.00000189, 0.00000072, -0.00000092, 0.00000197, },
		{ -0.00000528, 0.00001092, 0.00000813, -0.00000525, -0.00000169, -0.00000145, -0.00000097, 0.00000351, -0.00000343, -0.00000535, 0.00000047, 0.00000120, -0.00000114, -0.00000329, 0.00000393, 0.00000423, -0.00000780, -0.00000129, 0.00000402, -0.00000148, 0.00000492, -0.00000495, -0.00001552, 0.00000620, 0.00000936, -0.00001082, -0.00000204, 0.00001661, -0.00000065, -0.00002034, 0.00000124, 0.00001216, -0.00000914, -0.00000911, 0.00000115, 0.00000126, 0.00000435, -0.00000161, -0.00001137, -0.00001065, -0.00000192, 0.00000246, 0.00000062, 0.00001167, 0.00000316, -0.00002574, -0.00000799, 0.00001145, -0.00000665, 0.00000666, 0.00001352, -0.00002686, -0.00000527, 0.00003315, -0.00002924, -0.00001888, 0.00005384, -0.00002773, -0.00003218, 0.00005629, -0.00003275, -0.00001899, 0.00005810, -0.00005152, -0.00000308, 0.00005113, -0.00006095, 0.00002298, 0.00002973, -0.00006930, 0.00004630, 0.00001409, -0.00006521, 0.00004384, 0.00000457, -0.00003506, 0.00002300, -0.00001382, -0.00001096, 0.00000738, -0.00000090, 0.00000094, -0.00002300, 0.00001427, -0.00000015, -0.00004163, 0.00001403, -0.00003200, -0.00002719, 0.00002757, -0.00008046, -0.00001618, -0.00002714, -0.00027026, -0.00000698, 0.00034604, 0.00013629, 0.00004038, 0.00023662, 0.00023106, 0.00000285, -0.00018357, -0.00011221, 0.00001263, -0.00003704, -0.00004510, 0.00000530, -0.00001068, -0.00000721, 0.00000350, -0.00000936, -0.00000714, -0.00000423, -0.00000166, -0.00000345, -0.00000722, -0.00000178, -0.00000415, -0.00000284, 0.00000293, -0.00000102, 0.00000031, 0.00000291, 0.00000080, -0.00000031, 0.00000009, 0.00000288, -0.00000383, },
		{ 0.00001360, 0.00000299, -0.00000546, -0.00000006, -0.00000279, 0.00000057, 0.00000272, -0.00000562, -0.00000318, 0.00000143, -0.00000023, -0.00000134, -0.00000107, 0.00000385, -0.00000005, -0.00000566, 0.00000173, 0.00000049, 0.00000050, 0.00000339, -0.00001197, -0.00000673, 0.00001236, -0.00000284, -0.00001071, 0.00001171, 0.00001062, -0.00001584, -0.00001090, 0.00001289, -0.00000033, -0.00001460, -0.00000002, 0.00000155, 0.00000080, 0.00000588, -0.00000728, -0.00001530, -0.00000621, 0.00000192, 0.00000167, 0.00000458, 0.00001012, -0.00001091, -0.00002197, 0.00000510, 0.00000433, -0.00000520, 0.00001577, -0.00000506, -0.00002861, 0.00002187, 0.00001307, -0.00004495, 0.00002163, 0.00003378, -0.00005542, 0.00001611, 0.00003522, -0.00005515, 0.00003184, 0.00001972, -0.00006206, 0.00005004, -0.00000081, -0.00005373, 0.00006571, -0.00003120, -0.00003854, 0.00007617, -0.00005028, -0.00002637, 0.00006285, -0.00004395, -0.00000593, 0.00002665, -0.00003630, 0.00000792, 0.00000775, -0.00000911, -0.00000282, -0.00001140, 0.00001580, -0.00002592, -0.00002166, 0.00001119, -0.00005275, 0.00001153, -0.00000473, -0.00009583, 0.00003531, -0.00012242, -0.00028407, 0.00017349, 0.00034225, 0.00004863, 0.00013938, 0.00030939, 0.00011744, -0.00013789, -0.00017462, -0.00003090, 0.00000118, -0.00005994, -0.00001404, 0.00000449, -0.00001996, 0.00000442, -0.00000074, -0.00001376, -0.00000231, -0.00000409, -0.00000432, -0.00000405, -0.00000439, -0.00000292, -0.00000507, 0.00000111, 0.00000168, -0.00000230, 0.00000287, 0.00000220, -0.00000061, -0.00000007, 0.00000166, 0.00000204, -0.00000681, -0.00000121, },
		{ -0.00000198, -0.00000385, 0.00000015, -0.00000353, 0.00000243, 0.00000054, -0.00000624, -0.00000067, 0.00000064, -0.00000081, -0.00000029, -0.00000055, 0.00000228, -0.00000133, -0.00000291, 0.00000118, -0.00000083, 0.00000250, -0.00000233, -0.00001206, 0.00000444, 0.00000706, -0.00001213, 0.00000090, 0.00001665, -0.00000528, -0.00001800, 0.00000511, 0.00000976, -0.00001380, -0.00000819, 0.00000570, -0.00000195, 0.00000419, 0.00000385, -0.00001554, -0.00001353, -0.00000033, 0.00000261, 0.00000279, 0.00000690, 0.00000150, -0.00001804, -0.00001057, 0.00000857, -0.00000205, 0.00000361, 0.00001167, -0.00002242, -0.00001021, 0.00003190, -0.00001853, -0.00002921, 0.00004631, -0.00000859, -0.00004308, 0.00004867, -0.00000950, -0.00003426, 0.00005473, -0.00003383, -0.00002276, 0.00006075, -0.00005622, 0.00000140, 0.00005543, -0.00007705, 0.00003090, 0.00004449, -0.00008643, 0.00004267, 0.00002578, -0.00006501, 0.00004119, -0.00000592, -0.00003761, 0.00003391, -0.00001225, -0.00001016, 0.00000424, -0.00000678, 0.00000319, -0.00003386, -0.00000044, -0.00001249, -0.00004419, 0.00003601, -0.00006143, -0.00005974, 0.00004656, -0.00025269, -0.00019339, 0.00033703, 0.00025093, 0.00003900, 0.00028489, 0.00027018, -0.00005893, -0.00019837, -0.00009902, 0.00000646, -0.00003596, -0.00004787, 0.00000956, -0.00001313, -0.00001379, 0.00001251, -0.00001261, -0.00000997, 0.00000173, -0.00000886, -0.00000440, -0.00000183, -0.00000426, -0.00000492, -0.00000255, 0.00000312, -0.00000139, -0.00000065, 0.00000448, -0.00000019, -0.00000073, 0.00000104, 0.00000262, -0.00000042, -0.00000820, 0.00000443, 0.00001334, },
		{ -0.00000137, -0.00000085, -0.00000315, 0.00000354, -0.00000203, -0.00000515, 0.00000062, -0.00000065, 0.00000022, -0.00000014, -0.00000150, 0.00000197, -0.00000097, -0.00000211, 0.00000002, -0.00000007, 0.00000118, -0.00000690, -0.00000481, 0.00000871, -0.00000463, -0.00000967, 0.00001366, 0.00000740, -0.00001730, -0.00000619, 0.00001317, -0.00000420, -0.00001690, 0.00000276, 0.00000378, -0.00000287, 0.00000849, -0.00000443, -0.00002041, -0.00000666, 0.00000284, 0.00000281, 0.00000491, 0.00000358, -0.00000737, -0.00001557, -0.00000125, 0.00000554, -0.00000137, 0.00000902, -0.00000257, -0.00002443, 0.00001250, 0.00001842, -0.00003656, 0.00000362, 0.00003836, -0.00004014, -0.00000414, 0.00004368, -0.00004048, 0.00000953, 0.00003255, -0.00005799, 0.00003264, 0.00001910, -0.00006706, 0.00005935, -0.00000480, -0.00006406, 0.00008348, -0.00003146, -0.00005578, 0.00008407, -0.00004551, -0.00002861, 0.00006019, -0.00005543, 0.00000013, 0.00003579, -0.00004055, 0.00001095, 0.00000457, -0.00001401, -0.00000532, -0.00002774, 0.00000457, -0.00003394, -0.00001275, 0.00002133, -0.00010116, 0.00001067, -0.00002264, -0.00034808, -0.00000564, 0.00040921, 0.00013784, 0.00014458, 0.00037586, 0.00010083, -0.00020026, -0.00015714, -0.00002712, -0.00000884, -0.00005521, -0.00001540, 0.00000724, -0.00002615, 0.00000421, 0.00000612, -0.00002079, 0.00000092, -0.00000226, -0.00001318, 0.00000045, -0.00000152, -0.00000650, -0.00000425, 0.00000087, 0.00000173, -0.00000291, 0.00000273, 0.00000324, -0.00000216, 0.00000079, 0.00000187, 0.00000250, -0.00000365, -0.00000709, 0.00001011, 0.00001006, -0.00000549, },
		{ -0.00000258, -0.00000184, 0.00000359, -0.00000396, -0.00000339, 0.00000045, -0.00000079, 0.00000149, -0.00000193, -0.00000163, 0.00000316, -0.00000153, -0.00000241, 0.00000014, -0.00000017, -0.00000214, -0.00000560, 0.00000300, 0.00000287, -0.00001171, 0.00000259, 0.00001568, -0.00000846, -0.00001535, 0.00000828, 0.00000690, -0.00001644, -0.00000861, 0.00000867, -0.00000171, 0.00000144, 0.00000774, -0.00001540, -0.00001785, 0.00000051, 0.00000322, 0.00000445, 0.00000396, -0.00000331, -0.00001000, -0.00000913, 0.00000193, 0.00000363, 0.00000257, 0.00000483, -0.00001409, -0.00001179, 0.00002221, -0.00000495, -0.00003052, 0.00002776, 0.00000874, -0.00004078, 0.00002970, 0.00001197, -0.00003639, 0.00003860, -0.00001282, -0.00003497, 0.00005715, -0.00003952, -0.00002295, 0.00007005, -0.00006628, 0.00000311, 0.00007105, -0.00009076, 0.00002459, 0.00005454, -0.00009013, 0.00004389, 0.00001984, -0.00007336, 0.00005509, -0.00000565, -0.00004257, 0.00004172, -0.00001882, -0.00001591, -0.00000398, -0.00002442, -0.00000179, -0.00003605, 0.00001316, -0.00002396, -0.00009072, 0.00006051, -0.00016056, -0.00033900, 0.00020784, 0.00037073, 0.00010084, 0.00031212, 0.00031600, -0.00012271, -0.00022653, -0.00006762, -0.00000841, -0.00003923, -0.00004057, 0.00000629, -0.00001223, -0.00001965, 0.00001669, -0.00001095, -0.00001549, 0.00000824, -0.00001347, -0.00000980, 0.00000575, -0.00000566, -0.00000711, -0.00000108, 0.00000188, -0.00000091, -0.00000134, 0.00000473, 0.00000002, -0.00000177, 0.00000256, 0.00000183, 0.00000150, -0.00000672, -0.00000329, 0.00001413, 0.00000446, -0.00000660, 0.00000102, },
		{ 0.00000020, 0.00000279, -0.00000505, -0.00000202, -0.00000006, 0.00000037, 0.00000105, -0.00000419, 0.00000057, 0.00000366, -0.00000344, -0.00000174, 0.00000024, -0.00000240, -0.00000287, 0.00000018, 0.00000385, -0.00000685, -0.00000750, 0.00001350, 0.00000466, -0.00001742, -0.00000198, 0.00001262, -0.00000737, -0.00001766, 0.00000287, 0.00000652, -0.00000400, 0.00000704, -0.00000039, -0.00002224, -0.00000959, 0.00000420, 0.00000353, 0.00000554, -0.00000170, -0.00000750, -0.00000678, -0.00000537, 0.00000282, 0.00000523, 0.00000199, -0.00000392, -0.00001420, 0.00000251, 0.00001579, -0.00001991, -0.00001076, 0.00002953, -0.00001765, -0.00001738, 0.00003730, -0.00001724, -0.00000788, 0.00003227, -0.00004220, 0.00001201, 0.00003202, -0.00006574, 0.00003900, 0.00002350, -0.00007616, 0.00007087, -0.00000128, -0.00008004, 0.00008911, -0.00002974, -0.00006079, 0.00008904, -0.00005726, -0.00002735, 0.00007445, -0.00006612, 0.00000340, 0.00004427, -0.00005382, 0.00000518, -0.00000935, -0.00002776, -0.00000246, -0.00002678, 0.00001355, -0.00006442, -0.00003844, 0.00003984, -0.00030156, -0.00021205, 0.00035959, 0.00028571, 0.00018830, 0.00041429, 0.00009805, -0.00027409, -0.00014079, -0.00000967, -0.00003386, -0.00004463, -0.00001280, 0.00000237, -0.00002507, -0.00000017, 0.00001132, -0.00002307, -0.00000023, 0.00000255, -0.00002182, 0.00000144, 0.00000464, -0.00001077, -0.00000362, 0.00000106, 0.00000033, -0.00000171, 0.00000153, 0.00000369, -0.00000240, 0.00000071, 0.00000299, 0.00000141, -0.00000030, -0.00000864, 0.00000272, 0.00001490, -0.00000193, -0.00000491, 0.00000226, -0.00000451, },
		{ 0.00000118, -0.00000562, -0.00000102, 0.00000011, 0.00000135, -0.00000136, -0.00000446, 0.00000385, 0.00000172, -0.00000472, -0.00000011, -0.00000181, -0.00000402, 0.00000127, 0.00000384, -0.00000299, -0.00001065, 0.00000405, 0.00001307, -0.00001000, -0.00001234, 0.00001050, 0.00000379, -0.00001759, -0.00000883, 0.00000889, 0.00000092, -0.00000093, 0.00000778, -0.00001181, -0.00002098, -0.00000091, 0.00000476, 0.00000493, 0.00000261, -0.00000824, -0.00000533, -0.00000362, -0.00000475, 0.00000618, 0.00000523, -0.00000450, -0.00000738, -0.00000672, 0.00000760, 0.00000400, -0.00002044, 0.00000520, 0.00001584, -0.00002515, 0.00000782, 0.00002266, -0.00002422, 0.00001684, 0.00000238, -0.00003463, 0.00004181, -0.00001846, -0.00003968, 0.00006746, -0.00004078, -0.00002514, 0.00008219, -0.00007506, -0.00000382, 0.00007896, -0.00009923, 0.00002709, 0.00005670, -0.00010391, 0.00005733, 0.00002333, -0.00008725, 0.00007128, -0.00000640, -0.00006035, 0.00003682, -0.00003778, -0.00001917, 0.00000383, -0.00002635, -0.00000154, -0.00007420, 0.00001167, -0.00005278, -0.00037405, -0.00002202, 0.00041805, 0.00025520, 0.00034058, 0.00034001, -0.00017076, -0.00027069, -0.00002677, -0.00002003, -0.00005513, -0.00002144, -0.00000138, -0.00001368, -0.00001865, 0.00001290, -0.00000602, -0.00001904, 0.00001042, -0.00001331, -0.00001761, 0.00001177, -0.00000398, -0.00001083, 0.00000124, -0.00000020, -0.00000080, -0.00000029, 0.00000281, 0.00000099, -0.00000212, 0.00000317, 0.00000197, 0.00000121, -0.00000299, -0.00000820, 0.00000940, 0.00001154, -0.00000681, -0.00000102, 0.00000137, -0.00000577, 0.00000279, },
		{ -0.00000556, 0.00000012, 0.00000064, 0.00000079, -0.00000389, -0.00000195, 0.00000550, -0.00000174, -0.00000381, 0.00000008, -0.00000526, -0.00000117, 0.00000676, 0.00000024, -0.00001046, -0.00000452, 0.00001211, 0.00000165, -0.00001586, 0.00000146, 0.00001144, -0.00001007, -0.00001690, 0.00000173, 0.00000765, -0.00000201, 0.00000361, 0.00000159, -0.00001976, -0.00001339, 0.00000434, 0.00000474, 0.00000514, -0.00000441, -0.00000980, 0.00000018, -0.00000463, -0.00000245, 0.00001052, -0.00000138, -0.00000960, -0.00000256, -0.00000213, 0.00000488, -0.00000229, -0.00001456, 0.00000990, 0.00000239, -0.00001728, 0.00001940, 0.00000464, -0.00001290, 0.00001961, -0.00002201, -0.00000306, 0.00003248, -0.00005109, 0.00001368, 0.00004201, -0.00006925, 0.00004364, 0.00002767, -0.00008812, 0.00007479, -0.00000256, -0.00008985, 0.00010042, -0.00003718, -0.00006709, 0.00010770, -0.00007022, -0.00002998, 0.00009654, -0.00008355, -0.00001095, 0.00003718, -0.00007336, 0.00001730, -0.00000637, -0.00003519, -0.00000770, -0.00006362, 0.00002283, -0.00017198, -0.00035544, 0.00015991, 0.00043428, 0.00032401, 0.00041878, 0.00009126, -0.00033713, -0.00013965, 0.00002258, -0.00006582, -0.00003840, 0.00000040, -0.00001194, -0.00001914, -0.00000291, 0.00000889, -0.00001814, -0.00000401, 0.00000558, -0.00002558, -0.00000213, 0.00001166, -0.00001310, -0.00000443, 0.00000232, -0.00000274, 0.00000064, 0.00000081, 0.00000191, -0.00000067, 0.00000014, 0.00000386, 0.00000070, 0.00000093, -0.00000610, -0.00000463, 0.00001425, 0.00000487, -0.00000805, 0.00000309, -0.00000181, -0.00000511, 0.00000520, -0.00000146, },
		{ 0.00000130, 0.00000050, -0.00000096, -0.00000452, 0.00000162, 0.00000423, -0.00000419, -0.00000186, -0.00000274, -0.00000610, 0.00000606, 0.00000714, -0.00000857, -0.00001058, 0.00000586, 0.00000920, -0.00001056, -0.00000868, 0.00001154, 0.00000076, -0.00001790, -0.00000812, 0.00000699, 0.00000363, -0.00000081, 0.00000434, -0.00000778, -0.00002044, -0.00000448, 0.00000613, 0.00000502, 0.00000144, -0.00001086, -0.00000464, 0.00000277, -0.00000742, 0.00000426, 0.00000930, -0.00001164, -0.00000582, 0.00000314, -0.00000505, 0.00000361, -0.00000317, -0.00001183, 0.00000921, -0.00000295, -0.00000694, 0.00001883, -0.00000371, -0.00000066, 0.00000581, -0.00002301, 0.00002190, -0.00000241, -0.00004243, 0.00004972, -0.00001135, -0.00004164, 0.00007329, -0.00004555, -0.00003028, 0.00008797, -0.00008801, -0.00000338, 0.00009052, -0.00011497, 0.00003543, 0.00006635, -0.00012568, 0.00007550, 0.00003434, -0.00011732, 0.00006667, -0.00002155, -0.00006959, 0.00006023, -0.00005132, -0.00002451, -0.00000436, -0.00006167, -0.00000069, -0.00027426, -0.00027552, 0.00031542, 0.00048602, 0.00042437, 0.00031313, -0.00020128, -0.00031578, 0.00000481, -0.00001732, -0.00008380, 0.00000230, -0.00000507, -0.00002348, -0.00000851, 0.00000379, -0.00000376, -0.00001526, 0.00000614, -0.00001051, -0.00002285, 0.00001256, 0.00000080, -0.00001432, 0.00000283, -0.00000172, -0.00000224, 0.00000287, -0.00000035, 0.00000095, -0.00000028, 0.00000225, 0.00000289, 0.00000014, -0.00000020, -0.00000824, 0.00000146, 0.00001506, -0.00000243, -0.00000516, 0.00000489, -0.00000579, -0.00000189, 0.00000603, -0.00000472, -0.00000429, },
		{ -0.00000060, -0.00000239, -0.00000304, 0.00000387, 0.00000128, -0.00000441, -0.00000166, -0.00000650, -0.00000111, 0.00001226, -0.00000042, -0.00001479, -0.00000227, 0.00001069, -0.00000188, -0.00001307, 0.00000435, 0.00000959, -0.00001219, -0.00001529, 0.00000074, 0.00000654, 0.00000158, 0.00000105, -0.00000005, -0.00001460, -0.00001520, 0.00000204, 0.00000629, 0.00000388, -0.00000538, -0.00001154, 0.00000279, -0.00000014, -0.00000635, 0.00001115, -0.00000067, -0.00001609, 0.00000503, 0.00000055, -0.00000909, 0.00000835, -0.00000671, -0.00001204, 0.00001128, -0.00000439, -0.00000091, 0.00001596, -0.00000451, 0.00000168, -0.00000654, -0.00000718, 0.00002154, -0.00003067, -0.00000587, 0.00004412, -0.00004687, 0.00001378, 0.00004343, -0.00007624, 0.00004713, 0.00002396, -0.00010339, 0.00008889, -0.00000167, -0.00010333, 0.00011953, -0.00004697, -0.00007917, 0.00014011, -0.00008463, -0.00005731, 0.00009790, -0.00010726, 0.00000232, 0.00005437, -0.00010431, 0.00002771, -0.00002651, -0.00007013, -0.00003242, -0.00035925, -0.00015897, 0.00048998, 0.00058272, 0.00042102, 0.00004384, -0.00036914, -0.00015398, 0.00005591, -0.00008894, -0.00004525, 0.00002372, -0.00002982, -0.00001569, 0.00000217, -0.00000314, -0.00000907, -0.00000479, 0.00000182, -0.00002280, -0.00000740, 0.00001510, -0.00001121, -0.00000672, 0.00000390, -0.00000615, 0.00000233, 0.00000209, -0.00000230, 0.00000185, 0.00000077, 0.00000304, 0.00000147, 0.00000010, -0.00000232, -0.00000800, 0.00000784, 0.00001145, -0.00000719, -0.00000000, 0.00000309, -0.00000810, 0.00000286, 0.00000414, -0.00000722, -0.00000177, 0.00000159, },
		{ -0.00000255, -0.00000093, 0.00000404, -0.00000122, -0.00000383, -0.00000416, -0.00000670, 0.00000795, 0.00001112, -0.00001149, -0.00001220, 0.00000734, 0.00000538, -0.00001098, -0.00000444, 0.00001157, -0.00000206, -0.00001765, -0.00000664, 0.00000445, 0.00000465, 0.00000175, -0.00000013, -0.00000622, -0.00001596, -0.00000804, 0.00000551, 0.00000547, -0.00000023, -0.00001090, -0.00000589, 0.00000646, -0.00000483, 0.00000023, 0.00001081, -0.00001364, -0.00000859, 0.00001237, -0.00001026, -0.00000484, 0.00001284, -0.00001680, -0.00000730, 0.00001547, -0.00000843, 0.00000428, 0.00001501, -0.00000654, -0.00000184, -0.00000714, 0.00000614, 0.00000281, -0.00003214, 0.00002578, 0.00001012, -0.00004042, 0.00004925, -0.00001464, -0.00004335, 0.00007750, -0.00006170, -0.00003452, 0.00010707, -0.00010046, -0.00000335, 0.00010568, -0.00013971, 0.00004847, 0.00009111, -0.00016337, 0.00006559, 0.00002870, -0.00013729, 0.00010294, -0.00003340, -0.00009843, 0.00008874, -0.00010159, -0.00004816, -0.00007933, -0.00044696, 0.00003034, 0.00071963, 0.00062703, 0.00023246, -0.00023977, -0.00033384, 0.00001130, -0.00000157, -0.00011285, 0.00001742, 0.00000210, -0.00004156, 0.00000589, -0.00000327, -0.00001011, -0.00000287, -0.00000148, -0.00001059, -0.00002085, 0.00000756, 0.00000565, -0.00001444, 0.00000202, -0.00000193, -0.00000500, 0.00000660, -0.00000258, -0.00000161, 0.00000372, 0.00000092, 0.00000303, 0.00000028, -0.00000008, -0.00000459, -0.00000522, 0.00001213, 0.00000526, -0.00000780, 0.00000438, -0.00000129, -0.00000701, 0.00000678, -0.00000007, -0.00000750, 0.00000082, 0.00000042, -0.00000164, },
		{ 0.00000055, 0.00000312, -0.00000272, -0.00000460, -0.00000677, -0.00000093, 0.00001468, 0.00000153, -0.00001776, -0.00000227, 0.00000967, -0.00000521, -0.00001007, 0.00000701, 0.00000715, -0.00001361, -0.00001341, 0.00000048, 0.00000406, 0.00000452, 0.00000094, -0.00000430, -0.00000990, -0.00001305, -0.00000198, 0.00000668, 0.00000285, -0.00000592, -0.00001122, 0.00000165, 0.00000463, -0.00000579, 0.00000713, 0.00000109, -0.00001904, 0.00000619, 0.00000738, -0.00001891, 0.00000837, 0.00000719, -0.00002649, 0.00000636, 0.00001379, -0.00001355, 0.00001341, 0.00001117, -0.00001190, -0.00000152, -0.00000133, 0.00000594, -0.00001377, -0.00001303, 0.00003156, -0.00001955, -0.00000648, 0.00004094, -0.00004998, 0.00001815, 0.00003816, -0.00009601, 0.00005962, 0.00003458, -0.00012016, 0.00010325, -0.00000471, -0.00012375, 0.00015494, -0.00005174, -0.00011981, 0.00014715, -0.00010654, -0.00005299, 0.00013598, -0.00015564, 0.00000698, 0.00007554, -0.00018966, 0.00003521, -0.00020290, -0.00050232, 0.00036484, 0.00093971, 0.00048892, -0.00007768, -0.00037203, -0.00016279, 0.00006701, -0.00009104, -0.00006295, 0.00004634, -0.00004044, -0.00002095, 0.00001564, -0.00001982, -0.00000320, 0.00000261, -0.00000969, -0.00001684, -0.00000882, 0.00001138, -0.00000543, -0.00000830, 0.00000369, -0.00000792, 0.00000201, 0.00000513, -0.00000708, 0.00000283, 0.00000388, 0.00000053, 0.00000289, -0.00000084, -0.00000045, -0.00000616, -0.00000100, 0.00001331, -0.00000085, -0.00000503, 0.00000588, -0.00000566, -0.00000293, 0.00000780, -0.00000462, -0.00000545, 0.00000188, -0.00000142, -0.00000161, -0.00000200, },
		{ 0.00000200, -0.00000446, -0.00000664, -0.00000557, 0.00000804, 0.00001335, -0.00001067, -0.00001445, 0.00000697, 0.00000272, -0.00001142, -0.00000036, 0.00001108, -0.00000470, -0.00001687, -0.00000497, 0.00000248, 0.00000376, 0.00000500, -0.00000307, -0.00000798, -0.00000969, -0.00000886, 0.00000212, 0.00000578, -0.00000172, -0.00000969, -0.00000662, 0.00000617, 0.00000083, -0.00000209, 0.00000751, -0.00001136, -0.00001163, 0.00001610, -0.00000749, -0.00001511, 0.00002003, -0.00001000, -0.00002446, 0.00002189, 0.00000203, -0.00001166, 0.00002321, -0.00000017, -0.00001375, 0.00000445, -0.00000018, -0.00000215, -0.00001592, 0.00000619, 0.00001690, -0.00002457, 0.00002156, 0.00000571, -0.00003864, 0.00005274, -0.00003221, -0.00005315, 0.00009976, -0.00006629, -0.00004250, 0.00012364, -0.00012202, 0.00000033, 0.00014066, -0.00017711, 0.00002835, 0.00009750, -0.00019278, 0.00010452, 0.00003360, -0.00019717, 0.00015535, -0.00006532, -0.00018537, 0.00012257, -0.00045183, -0.00039463, 0.00084472, 0.00098190, 0.00014084, -0.00033217, -0.00030289, -0.00000427, 0.00000857, -0.00012286, 0.00001458, 0.00001758, -0.00005898, 0.00001501, -0.00000189, -0.00002535, 0.00001287, -0.00000564, -0.00001828, -0.00001115, 0.00000065, 0.00000575, -0.00000894, -0.00000107, -0.00000190, -0.00000706, 0.00000843, -0.00000231, -0.00000617, 0.00000775, 0.00000130, 0.00000117, 0.00000211, -0.00000175, -0.00000070, -0.00000704, 0.00000355, 0.00001188, -0.00000523, -0.00000087, 0.00000450, -0.00000793, 0.00000206, 0.00000568, -0.00000752, -0.00000258, 0.00000099, -0.00000236, -0.00000074, -0.00000176, 0.00000151, },
		{ -0.00000716, -0.00000721, 0.00000052, 0.00001400, 0.00000405, -0.00001743, -0.00000459, 0.00000816, -0.00000714, -0.00000770, 0.00000929, 0.00000452, -0.00001442, -0.00001147, 0.00000072, 0.00000159, 0.00000550, 0.00000273, -0.00000862, -0.00000804, -0.00000742, -0.00000533, 0.00000432, 0.00000261, -0.00000635, -0.00000935, -0.00000075, 0.00000675, -0.00000088, 0.00000167, 0.00000027, -0.00001671, 0.00000299, 0.00001271, -0.00002020, 0.00000052, 0.00001833, -0.00002799, -0.00000745, 0.00002699, -0.00001318, 0.00000129, 0.00002435, -0.00001454, -0.00000613, 0.00000804, -0.00000620, -0.00000616, -0.00000836, 0.00001275, 0.00000124, -0.00001106, 0.00002443, -0.00002110, -0.00000065, 0.00003513, -0.00007210, 0.00002561, 0.00005644, -0.00011035, 0.00006568, 0.00003800, -0.00014384, 0.00013368, 0.00000270, -0.00017289, 0.00016169, -0.00006499, -0.00013083, 0.00020284, -0.00015588, -0.00007508, 0.00020506, -0.00026757, -0.00001594, 0.00006199, -0.00074765, 0.00002594, 0.00130453, 0.00069813, -0.00027286, -0.00038086, -0.00013201, 0.00003927, -0.00007636, -0.00007496, 0.00005213, -0.00003647, -0.00003306, 0.00002985, -0.00003171, -0.00000655, 0.00001782, -0.00002377, -0.00001417, -0.00000220, 0.00000099, 0.00000047, -0.00000578, -0.00000010, -0.00000732, 0.00000035, 0.00000760, -0.00000957, 0.00000087, 0.00000860, -0.00000170, 0.00000303, -0.00000007, -0.00000145, -0.00000137, -0.00000743, 0.00000823, 0.00000850, -0.00000754, 0.00000323, 0.00000142, -0.00000764, 0.00000597, 0.00000177, -0.00000826, -0.00000057, -0.00000059, -0.00000172, -0.00000034, -0.00000179, 0.00000256, 0.00000011, },
		{ -0.00000377, 0.00000803, 0.00001257, -0.00000720, -0.00001522, 0.00000380, 0.00000119, -0.00001112, 0.00000258, 0.00001045, -0.00000705, -0.00001562, -0.00000347, 0.00000137, 0.00000196, 0.00000700, -0.00000355, -0.00001154, -0.00000465, -0.00000570, -0.00000273, 0.00000441, -0.00000201, -0.00000849, -0.00000597, 0.00000339, 0.00000591, -0.00000054, 0.00000073, -0.00000841, -0.00001115, 0.00001339, -0.00000101, -0.00002031, 0.00001630, 0.00000188, -0.00003300, 0.00001493, 0.00001649, -0.00001956, 0.00001846, 0.00001218, -0.00002107, 0.00000676, 0.00000203, -0.00001124, -0.00000247, -0.00000262, 0.00001038, -0.00000384, 0.00000188, 0.00001025, -0.00002081, 0.00002501, -0.00001144, -0.00005290, 0.00007298, -0.00002726, -0.00006519, 0.00011237, -0.00008059, -0.00004761, 0.00016122, -0.00015168, -0.00003093, 0.00015666, -0.00020935, 0.00005755, 0.00012743, -0.00027605, 0.00015526, 0.00004823, -0.00035251, 0.00023786, -0.00027142, -0.00084713, 0.00076060, 0.00145656, 0.00012193, -0.00051777, -0.00022974, -0.00001363, -0.00001465, -0.00010130, -0.00000136, 0.00002705, -0.00006244, 0.00001220, 0.00000682, -0.00004191, 0.00002279, -0.00000118, -0.00003284, 0.00000107, -0.00000198, -0.00000197, 0.00000103, -0.00000372, -0.00000411, -0.00000619, 0.00000703, -0.00000032, -0.00000997, 0.00000895, 0.00000438, -0.00000192, 0.00000402, -0.00000280, 0.00000056, -0.00000384, -0.00000633, 0.00001285, 0.00000312, -0.00000752, 0.00000668, -0.00000245, -0.00000534, 0.00000806, -0.00000261, -0.00000765, 0.00000030, -0.00000137, -0.00000051, -0.00000103, -0.00000109, 0.00000319, -0.00000349, -0.00000929, },
		{ 0.00001181, 0.00000492, -0.00001359, -0.00000757, 0.00000483, -0.00000708, -0.00000584, 0.00001050, 0.00000224, -0.00001478, -0.00000946, 0.00000102, -0.00000004, 0.00000473, 0.00000458, -0.00001092, -0.00000912, -0.00000085, -0.00000565, -0.00000083, 0.00000231, -0.00000593, -0.00000741, -0.00000220, 0.00000585, 0.00000543, -0.00000158, -0.00000432, -0.00001115, -0.00000014, 0.00001260, -0.00001397, -0.00000829, 0.00002017, -0.00001843, -0.00002065, 0.00002815, -0.00000211, -0.00001119, 0.00002729, -0.00000640, -0.00001380, 0.00001341, -0.00001023, -0.00000735, 0.00000169, -0.00000253, 0.00000912, -0.00000258, 0.00000434, -0.00000086, -0.00000445, 0.00001893, -0.00004292, 0.00000044, 0.00005633, -0.00007911, 0.00002348, 0.00006246, -0.00013107, 0.00008621, 0.00005463, -0.00019333, 0.00013445, 0.00000179, -0.00019860, 0.00022320, -0.00009722, -0.00018850, 0.00030442, -0.00025543, -0.00017184, 0.00033288, -0.00080060, -0.00047465, 0.00154717, 0.00108691, -0.00048755, -0.00045307, -0.00003743, -0.00001978, -0.00006933, -0.00005844, 0.00003164, -0.00002272, -0.00004009, 0.00003301, -0.00003210, -0.00001776, 0.00003311, -0.00003202, -0.00001931, 0.00001131, -0.00001068, 0.00000123, 0.00000271, -0.00000680, -0.00000593, -0.00000014, 0.00000665, -0.00000834, -0.00000280, 0.00001196, -0.00000118, 0.00000093, 0.00000215, -0.00000354, 0.00000251, -0.00000843, -0.00000174, 0.00001562, -0.00000397, -0.00000426, 0.00000853, -0.00000665, -0.00000135, 0.00000811, -0.00000700, -0.00000623, 0.00000091, -0.00000120, -0.00000013, -0.00000185, 0.00000066, 0.00000176, -0.00000802, -0.00000803, 0.00000290, },
		{ -0.00000356, -0.00001300, -0.00000139, -0.00000049, -0.00000921, 0.00000399, 0.00000945, -0.00000876, -0.00001410, -0.00000215, 0.00000080, 0.00000042, 0.00000698, -0.00000258, -0.00001442, -0.00000256, 0.00000018, -0.00000649, 0.00000029, -0.00000108, -0.00000732, -0.00000475, 0.00000075, 0.00000811, 0.00000385, -0.00000588, -0.00000805, -0.00000657, 0.00000738, 0.00000354, -0.00001739, 0.00000582, 0.00001011, -0.00002907, 0.00000059, 0.00002523, -0.00001534, 0.00000571, 0.00002131, -0.00001841, 0.00000116, 0.00000693, -0.00001735, 0.00000305, -0.00000112, -0.00000181, 0.00001135, -0.00000418, 0.00000246, -0.00000027, 0.00000518, -0.00000500, -0.00003703, 0.00003923, 0.00000105, -0.00006324, 0.00007806, -0.00003423, -0.00007450, 0.00014445, -0.00009909, -0.00008588, 0.00018074, -0.00017872, -0.00001964, 0.00020862, -0.00029701, 0.00008208, 0.00019625, -0.00047178, 0.00021483, 0.00003350, -0.00118798, 0.00043086, 0.00194545, 0.00026572, -0.00079484, -0.00016004, 0.00002510, -0.00008866, -0.00005633, -0.00000918, 0.00001170, -0.00004460, 0.00000124, 0.00001248, -0.00004833, 0.00002068, 0.00000989, -0.00004685, 0.00000799, 0.00000350, -0.00001556, 0.00001051, -0.00000214, -0.00001019, -0.00000204, 0.00000369, 0.00000041, -0.00000982, 0.00000620, 0.00000856, -0.00000355, 0.00000376, -0.00000166, -0.00000090, 0.00000113, -0.00001242, 0.00000681, 0.00001349, -0.00001036, 0.00000252, 0.00000699, -0.00000994, 0.00000416, 0.00000538, -0.00001092, -0.00000355, 0.00000162, -0.00000120, -0.00000061, -0.00000148, 0.00000193, -0.00000244, -0.00001052, -0.00000240, 0.00000872, 0.00000950, },
		{ -0.00000890, -0.00000068, -0.00000585, -0.00000360, 0.00001003, 0.00000043, -0.00001462, -0.00000746, 0.00000120, -0.00000081, 0.00000331, 0.00000493, -0.00001090, -0.00001121, 0.00000386, -0.00000249, -0.00000683, 0.00000043, -0.00000407, -0.00000623, -0.00000232, 0.00000383, 0.00000973, -0.00000128, -0.00001057, -0.00000624, -0.00000002, 0.00000733, -0.00000534, -0.00001174, 0.00001225, -0.00000539, -0.00002496, 0.00001742, 0.00001138, -0.00001557, 0.00001879, 0.00000601, -0.00001670, 0.00001094, -0.00000732, -0.00001197, 0.00000943, -0.00000858, 0.00000546, 0.00001027, -0.00000901, 0.00000582, 0.00000315, -0.00000051, -0.00002002, -0.00000874, 0.00003764, -0.00004083, -0.00000606, 0.00005994, -0.00009281, 0.00003405, 0.00008143, -0.00017315, 0.00007893, 0.00006455, -0.00022689, 0.00018751, -0.00000295, -0.00028395, 0.00032822, -0.00015167, -0.00036497, 0.00050490, -0.00064127, -0.00100994, 0.00155278, 0.00161878, -0.00062255, -0.00063640, 0.00011027, -0.00007331, -0.00010324, -0.00000071, -0.00000686, -0.00001735, -0.00002648, 0.00001839, -0.00002388, -0.00002670, 0.00003757, -0.00002924, -0.00002993, 0.00002446, -0.00001639, -0.00000576, 0.00001489, -0.00001272, -0.00000674, 0.00000282, 0.00000142, -0.00000490, -0.00000472, 0.00001088, 0.00000270, -0.00000189, 0.00000350, -0.00000395, 0.00000299, -0.00000491, -0.00001133, 0.00001598, 0.00000516, -0.00001189, 0.00001039, 0.00000100, -0.00000984, 0.00000959, -0.00000083, -0.00001268, 0.00000062, 0.00000139, -0.00000190, -0.00000060, -0.00000020, 0.00000061, -0.00000745, -0.00000819, 0.00000496, 0.00000984, 0.00000344, -0.00000808, },
		{ -0.00000365, -0.00000577, 0.00000454, 0.00000747, -0.00000958, -0.00001242, -0.00000094, 0.00000039, -0.00000034, 0.00000549, -0.00000193, -0.00001505, -0.00000287, 0.00000592, -0.00000717, -0.00000584, -0.00000027, -0.00000549, -0.00000413, -0.00000015, 0.00000774, 0.00000794, -0.00000919, -0.00001063, -0.00000059, 0.00000255, 0.00000309, -0.00000866, -0.00000411, 0.00000941, -0.00001608, -0.00001166, 0.00002262, -0.00000241, -0.00000605, 0.00002075, -0.00000700, -0.00000599, 0.00000814, -0.00001684, 0.00000132, 0.00000490, -0.00001104, 0.00001557, 0.00000097, -0.00000859, 0.00001310, -0.00000198, -0.00001006, -0.00001553, 0.00001187, 0.00001033, -0.00004090, 0.00003696, -0.00000202, -0.00007114, 0.00009967, -0.00004345, -0.00011244, 0.00016126, -0.00011582, -0.00009297, 0.00024232, -0.00025019, -0.00003256, 0.00031550, -0.00048903, 0.00006877, 0.00034920, -0.00131224, -0.00008731, 0.00230185, 0.00061841, -0.00110078, -0.00016369, 0.00014051, -0.00020170, -0.00001925, 0.00001813, -0.00003780, -0.00001310, -0.00000261, 0.00000123, -0.00003786, 0.00000967, 0.00001754, -0.00005035, 0.00000518, 0.00001436, -0.00002789, 0.00001351, 0.00000521, -0.00001861, 0.00000283, 0.00000157, -0.00000302, -0.00000484, 0.00000157, 0.00001020, -0.00000112, 0.00000075, 0.00000067, -0.00000249, 0.00000372, -0.00001241, -0.00000301, 0.00001991, -0.00000618, -0.00000598, 0.00001454, -0.00000745, -0.00000456, 0.00001155, -0.00000926, -0.00001009, 0.00000461, -0.00000084, -0.00000211, 0.00000059, -0.00000009, -0.00000338, -0.00000955, -0.00000142, 0.00000953, 0.00000602, -0.00000180, -0.00000826, -0.00000627, },
		{ -0.00000030, 0.00000813, -0.00000142, -0.00001371, -0.00000555, 0.00000134, -0.00000113, 0.00000222, 0.00000348, -0.00000988, -0.00001201, 0.00000556, 0.00000227, -0.00001093, -0.00000363, -0.00000144, -0.00000541, -0.00000202, 0.00000292, 0.00001058, 0.00000093, -0.00001495, -0.00000458, 0.00000314, 0.00000062, 0.00000031, -0.00000801, -0.00000011, 0.00000280, -0.00001816, 0.00000171, 0.00001835, -0.00000918, 0.00000449, 0.00001415, -0.00001080, 0.00000275, -0.00000288, -0.00001403, 0.00001070, -0.00000617, -0.00000289, 0.00001850, -0.00001098, 0.00000198, 0.00001337, -0.00001364, -0.00000914, -0.00000392, 0.00001169, -0.00001080, -0.00001413, 0.00003567, -0.00004741, -0.00000167, 0.00007402, -0.00012365, 0.00002138, 0.00009679, -0.00020572, 0.00011361, 0.00008550, -0.00031997, 0.00026814, -0.00000164, -0.00051046, 0.00052074, -0.00033531, -0.00142704, 0.00127884, 0.00217777, -0.00057369, -0.00093126, 0.00025645, -0.00007032, -0.00019783, 0.00008617, -0.00003660, -0.00004245, 0.00001501, -0.00000773, -0.00002048, -0.00002006, 0.00002512, -0.00001909, -0.00003651, 0.00002812, -0.00001258, -0.00001717, 0.00002442, -0.00001339, -0.00001129, 0.00000882, -0.00000576, -0.00000298, -0.00000174, 0.00000492, 0.00000749, -0.00000220, 0.00000163, -0.00000155, 0.00000091, -0.00000098, -0.00001548, 0.00000918, 0.00001492, -0.00001392, 0.00000511, 0.00001128, -0.00001336, 0.00000417, 0.00000708, -0.00001577, -0.00000310, 0.00000541, -0.00000402, -0.00000019, 0.00000146, -0.00000276, -0.00000723, -0.00000649, 0.00000593, 0.00000862, 0.00000071, -0.00000359, -0.00000737, -0.00000691, -0.00000523, },
		{ 0.00000422, -0.00000980, -0.00001041, 0.00000014, 0.00000004, -0.00000048, 0.00000364, -0.00000279, -0.00001364, -0.00000348, 0.00000906, -0.00000494, -0.00001150, -0.00000117, -0.00000283, -0.00000432, 0.00000048, 0.00000680, 0.00000907, -0.00000879, -0.00001372, 0.00000324, 0.00000135, -0.00000120, 0.00000039, -0.00000797, 0.00000075, -0.00000213, -0.00001475, 0.00001029, 0.00001082, -0.00000950, 0.00001056, 0.00000639, -0.00000785, 0.00000309, -0.00001141, -0.00000279, 0.00000889, -0.00001274, 0.00001046, 0.00000962, -0.00001425, 0.00001532, 0.00000065, -0.00001845, 0.00000124, -0.00000175, 0.00000283, -0.00001138, 0.00000592, 0.00000673, -0.00004394, 0.00004780, -0.00000753, -0.00010183, 0.00011072, -0.00004983, -0.00013164, 0.00021723, -0.00015867, -0.00013695, 0.00035683, -0.00040101, -0.00012874, 0.00056217, -0.00119659, -0.00065758, 0.00239825, 0.00114374, -0.00132444, -0.00029820, 0.00032094, -0.00030914, -0.00003010, 0.00009456, -0.00010979, 0.00000930, 0.00001898, -0.00003358, -0.00001575, 0.00000270, 0.00000930, -0.00003841, -0.00000348, 0.00002147, -0.00003068, 0.00000769, 0.00001497, -0.00002452, 0.00000462, 0.00000303, -0.00001066, 0.00000229, -0.00000079, 0.00000637, 0.00000526, -0.00000269, 0.00000128, -0.00000132, 0.00000209, -0.00000834, -0.00001066, 0.00001820, 0.00000334, -0.00001291, 0.00001488, 0.00000175, -0.00001246, 0.00001107, -0.00000285, -0.00001587, 0.00000467, 0.00000162, -0.00000502, 0.00000324, -0.00000059, -0.00000695, -0.00000739, 0.00000021, 0.00000943, 0.00000362, -0.00000223, -0.00000321, -0.00000848, -0.00000827, -0.00000227, 0.00000543, },
		{ -0.00001207, -0.00000368, 0.00000147, -0.00000130, 0.00000172, 0.00000100, -0.00000946, -0.00001051, 0.00000528, 0.00000577, -0.00001155, -0.00000857, 0.00000017, -0.00000415, -0.00000212, 0.00000387, 0.00000903, 0.00000214, -0.00001534, -0.00000571, 0.00000647, -0.00000351, 0.00000086, -0.00000028, -0.00000980, 0.00000221, -0.00000489, -0.00000993, 0.00001471, 0.00000359, -0.00000703, 0.00001287, 0.00000162, -0.00000497, -0.00000219, -0.00001102, 0.00000660, -0.00000055, -0.00000856, 0.00001796, -0.00000419, -0.00000459, 0.00001812, -0.00001647, -0.00000834, 0.00000653, -0.00000927, 0.00000260, -0.00000392, 0.00000459, -0.00001341, -0.00001119, 0.00004065, -0.00006813, -0.00001407, 0.00008893, -0.00014891, 0.00003575, 0.00013167, -0.00028494, 0.00015587, 0.00012858, -0.00054995, 0.00039476, -0.00000299, -0.00161225, 0.00078340, 0.00260554, -0.00028607, -0.00127122, 0.00032071, 0.00002391, -0.00033936, 0.00016449, -0.00002539, -0.00010815, 0.00007384, -0.00002616, -0.00003827, 0.00000871, 0.00000115, -0.00001409, -0.00002782, 0.00001815, -0.00000342, -0.00002427, 0.00002446, -0.00000739, -0.00001723, 0.00001409, -0.00001073, -0.00000551, 0.00000609, -0.00000269, 0.00000897, 0.00000238, -0.00000335, 0.00000184, -0.00000048, -0.00000074, -0.00001283, -0.00000041, 0.00001892, -0.00000782, -0.00000388, 0.00001744, -0.00000831, -0.00000522, 0.00001108, -0.00001275, -0.00000886, 0.00000801, -0.00000420, -0.00000167, 0.00000478, -0.00000600, -0.00000898, -0.00000297, 0.00000629, 0.00000752, -0.00000151, -0.00000181, -0.00000387, -0.00001129, -0.00000653, 0.00000353, 0.00000601, -0.00000400, },
		{ 0.00000110, -0.00000021, -0.00000038, 0.00000216, -0.00000508, -0.00001161, -0.00000251, 0.00000904, -0.00000194, -0.00001406, -0.00000403, -0.00000050, -0.00000441, 0.00000153, 0.00000695, 0.00000691, -0.00000687, -0.00001435, 0.00000335, 0.00000283, -0.00000558, 0.00000516, -0.00000514, -0.00001000, 0.00000536, -0.00000802, -0.00000436, 0.00001649, -0.00000337, -0.00000313, 0.00001404, -0.00000175, -0.00000541, -0.00000605, -0.00000435, 0.00000775, -0.00000813, 0.00000228, 0.00001428, -0.00001120, 0.00000942, 0.00000573, -0.00002234, 0.00000815, -0.00000283, -0.00001168, 0.00001124, -0.00000633, -0.00000386, -0.00000975, 0.00000931, -0.00000091, -0.00006407, 0.00005435, -0.00000746, -0.00012434, 0.00015006, -0.00006420, -0.00018786, 0.00031013, -0.00025336, -0.00028917, 0.00061050, -0.00090786, -0.00111850, 0.00218577, 0.00172089, -0.00135090, -0.00056846, 0.00050654, -0.00035178, -0.00011258, 0.00020588, -0.00017220, -0.00000364, 0.00007394, -0.00008313, 0.00000093, 0.00001528, -0.00002049, -0.00001677, -0.00000633, 0.00001464, -0.00002073, -0.00000237, 0.00001949, -0.00002263, 0.00000170, 0.00000687, -0.00001885, 0.00000684, 0.00000219, -0.00000232, 0.00001203, -0.00000285, -0.00000200, 0.00000303, -0.00000189, -0.00000496, -0.00001150, 0.00000915, 0.00001223, -0.00001245, 0.00000727, 0.00001214, -0.00001320, 0.00000306, 0.00000401, -0.00001648, 0.00000066, 0.00000473, -0.00000693, 0.00000414, 0.00000114, -0.00001154, -0.00000609, 0.00000325, 0.00000823, 0.00000240, -0.00000360, -0.00000030, -0.00000727, -0.00001255, -0.00000069, 0.00000721, 0.00000041, -0.00001025, -0.00000789, },
		{ -0.00000137, 0.00000157, -0.00000154, -0.00001034, -0.00000736, 0.00000498, 0.00000613, -0.00000933, -0.00001182, -0.00000071, -0.00000251, -0.00000230, 0.00000583, 0.00000719, 0.00000063, -0.00001245, -0.00000697, 0.00000715, -0.00000371, -0.00000183, 0.00000621, -0.00001249, -0.00000498, 0.00000687, -0.00001193, 0.00000334, 0.00001461, -0.00000978, 0.00000342, 0.00001375, -0.00000645, -0.00000663, -0.00000509, 0.00000158, 0.00000280, -0.00000817, 0.00001071, 0.00000450, -0.00000686, 0.00001465, -0.00001168, -0.00001095, 0.00001389, -0.00001856, 0.00000093, 0.00001301, -0.00001792, 0.00000011, 0.00000031, 0.00000056, -0.00002852, -0.00001804, 0.00005193, -0.00008452, -0.00001492, 0.00012284, -0.00020084, 0.00004351, 0.00018788, -0.00047821, 0.00019941, 0.00023732, -0.00153849, 0.00020678, 0.00276597, 0.00020678, -0.00153849, 0.00023732, 0.00019941, -0.00047821, 0.00018788, 0.00004351, -0.00020084, 0.00012284, -0.00001492, -0.00008452, 0.00005193, -0.00001804, -0.00002852, 0.00000056, 0.00000031, 0.00000011, -0.00001792, 0.00001301, 0.00000093, -0.00001856, 0.00001389, -0.00001095, -0.00001168, 0.00001465, -0.00000686, 0.00000450, 0.00001071, -0.00000817, 0.00000280, 0.00000158, -0.00000509, -0.00000663, -0.00000645, 0.00001375, 0.00000342, -0.00000978, 0.00001461, 0.00000334, -0.00001193, 0.00000687, -0.00000498, -0.00001249, 0.00000621, -0.00000183, -0.00000371, 0.00000715, -0.00000697, -0.00001245, 0.00000063, 0.00000719, 0.00000583, -0.00000230, -0.00000251, -0.00000071, -0.00001182, -0.00000933, 0.00000613, 0.00000498, -0.00000736, -0.00001034, -0.00000154, 0.00000157, },
		{ 0.00000110, -0.00000789, -0.00001025, 0.00000041, 0.00000721, -0.00000069, -0.00001255, -0.00000727, -0.00000030, -0.00000360, 0.00000240, 0.00000823, 0.00000325, -0.00000609, -0.00001154, 0.00000114, 0.00000414, -0.00000693, 0.00000473, 0.00000066, -0.00001648, 0.00000401, 0.00000306, -0.00001320, 0.00001214, 0.00000727, -0.00001245, 0.00001223, 0.00000915, -0.00001150, -0.00000496, -0.00000189, 0.00000303, -0.00000200, -0.00000285, 0.00001203, -0.00000232, 0.00000219, 0.00000684, -0.00001885, 0.00000687, 0.00000170, -0.00002263, 0.00001949, -0.00000237, -0.00002073, 0.00001464, -0.00000633, -0.00001677, -0.00002049, 0.00001528, 0.00000093, -0.00008313, 0.00007394, -0.00000364, -0.00017220, 0.00020588, -0.00011258, -0.00035178, 0.00050654, -0.00056846, -0.00135090, 0.00172089, 0.00218577, -0.00111850, -0.00090786, 0.00061050, -0.00028917, -0.00025336, 0.00031013, -0.00018786, -0.00006420, 0.00015006, -0.00012434, -0.00000746, 0.00005435, -0.00006407, -0.00000091, 0.00000931, -0.00000975, -0.00000386, -0.00000633, 0.00001124, -0.00001168, -0.00000283, 0.00000815, -0.00002234, 0.00000573, 0.00000942, -0.00001120, 0.00001428, 0.00000228, -0.00000813, 0.00000775, -0.00000435, -0.00000605, -0.00000541, -0.00000175, 0.00001404, -0.00000313, -0.00000337, 0.00001649, -0.00000436, -0.00000802, 0.00000536, -0.00001000, -0.00000514, 0.00000516, -0.00000558, 0.00000283, 0.00000335, -0.00001435, -0.00000687, 0.00000691, 0.00000695, 0.00000153, -0.00000441, -0.00000050, -0.00000403, -0.00001406, -0.00000194, 0.00000904, -0.00000251, -0.00001161, -0.00000508, 0.00000216, -0.00000038, -0.00000021, },
		{ -0.00001207, -0.00000400, 0.00000601, 0.00000353, -0.00000653, -0.00001129, -0.00000387, -0.00000181, -0.00000151, 0.00000752, 0.00000629, -0.00000297, -0.00000898, -0.00000600, 0.00000478, -0.00000167, -0.00000420, 0.00000801, -0.00000886, -0.00001275, 0.00001108, -0.00000522, -0.00000831, 0.00001744, -0.00000388, -0.00000782, 0.00001892, -0.00000041, -0.00001283, -0.00000074, -0.00000048, 0.00000184, -0.00000335, 0.00000238, 0.00000897, -0.00000269, 0.00000609, -0.00000551, -0.00001073, 0.00001409, -0.00001723, -0.00000739, 0.00002446, -0.00002427, -0.00000342, 0.00001815, -0.00002782, -0.00001409, 0.00000115, 0.00000871, -0.00003827, -0.00002616, 0.00007384, -0.00010815, -0.00002539, 0.00016449, -0.00033936, 0.00002391, 0.00032071, -0.00127122, -0.00028607, 0.00260554, 0.00078340, -0.00161225, -0.00000299, 0.00039476, -0.00054995, 0.00012858, 0.00015587, -0.00028494, 0.00013167, 0.00003575, -0.00014891, 0.00008893, -0.00001407, -0.00006813, 0.00004065, -0.00001119, -0.00001341, 0.00000459, -0.00000392, 0.00000260, -0.00000927, 0.00000653, -0.00000834, -0.00001647, 0.00001812, -0.00000459, -0.00000419, 0.00001796, -0.00000856, -0.00000055, 0.00000660, -0.00001102, -0.00000219, -0.00000497, 0.00000162, 0.00001287, -0.00000703, 0.00000359, 0.00001471, -0.00000993, -0.00000489, 0.00000221, -0.00000980, -0.00000028, 0.00000086, -0.00000351, 0.00000647, -0.00000571, -0.00001534, 0.00000214, 0.00000903, 0.00000387, -0.00000212, -0.00000415, 0.00000017, -0.00000857, -0.00001155, 0.00000577, 0.00000528, -0.00001051, -0.00000946, 0.00000100, 0.00000172, -0.00000130, 0.00000147, -0.00000368, },
		{ 0.00000422, 0.00000543, -0.00000227, -0.00000827, -0.00000848, -0.00000321, -0.00000223, 0.00000362, 0.00000943, 0.00000021, -0.00000739, -0.00000695, -0.00000059, 0.00000324, -0.00000502, 0.00000162, 0.00000467, -0.00001587, -0.00000285, 0.00001107, -0.00001246, 0.00000175, 0.00001488, -0.00001291, 0.00000334, 0.00001820, -0.00001066, -0.00000834, 0.00000209, -0.00000132, 0.00000128, -0.00000269, 0.00000526, 0.00000637, -0.00000079, 0.00000229, -0.00001066, 0.00000303, 0.00000462, -0.00002452, 0.00001497, 0.00000769, -0.00003068, 0.00002147, -0.00000348, -0.00003841, 0.00000930, 0.00000270, -0.00001575, -0.00003358, 0.00001898, 0.00000930, -0.00010979, 0.00009456, -0.00003010, -0.00030914, 0.00032094, -0.00029820, -0.00132444, 0.00114374, 0.00239825, -0.00065758, -0.00119659, 0.00056217, -0.00012874, -0.00040101, 0.00035683, -0.00013695, -0.00015867, 0.00021723, -0.00013164, -0.00004983, 0.00011072, -0.00010183, -0.00000753, 0.00004780, -0.00004394, 0.00000673, 0.00000592, -0.00001138, 0.00000283, -0.00000175, 0.00000124, -0.00001845, 0.00000065, 0.00001532, -0.00001425, 0.00000962, 0.00001046, -0.00001274, 0.00000889, -0.00000279, -0.00001141, 0.00000309, -0.00000785, 0.00000639, 0.00001056, -0.00000950, 0.00001082, 0.00001029, -0.00001475, -0.00000213, 0.00000075, -0.00000797, 0.00000039, -0.00000120, 0.00000135, 0.00000324, -0.00001372, -0.00000879, 0.00000907, 0.00000680, 0.00000048, -0.00000432, -0.00000283, -0.00000117, -0.00001150, -0.00000494, 0.00000906, -0.00000348, -0.00001364, -0.00000279, 0.00000364, -0.00000048, 0.00000004, 0.00000014, -0.00001041, -0.00000980, },
		{ -0.00000030, -0.00000523, -0.00000691, -0.00000737, -0.00000359, 0.00000071, 0.00000862, 0.00000593, -0.00000649, -0.00000723, -0.00000276, 0.00000146, -0.00000019, -0.00000402, 0.00000541, -0.00000310, -0.00001577, 0.00000708, 0.00000417, -0.00001336, 0.00001128, 0.00000511, -0.00001392, 0.00001492, 0.00000918, -0.00001548, -0.00000098, 0.00000091, -0.00000155, 0.00000163, -0.00000220, 0.00000749, 0.00000492, -0.00000174, -0.00000298, -0.00000576, 0.00000882, -0.00001129, -0.00001339, 0.00002442, -0.00001717, -0.00001258, 0.00002812, -0.00003651, -0.00001909, 0.00002512, -0.00002006, -0.00002048, -0.00000773, 0.00001501, -0.00004245, -0.00003660, 0.00008617, -0.00019783, -0.00007032, 0.00025645, -0.00093126, -0.00057369, 0.00217777, 0.00127884, -0.00142704, -0.00033531, 0.00052074, -0.00051046, -0.00000164, 0.00026814, -0.00031997, 0.00008550, 0.00011361, -0.00020572, 0.00009679, 0.00002138, -0.00012365, 0.00007402, -0.00000167, -0.00004741, 0.00003567, -0.00001413, -0.00001080, 0.00001169, -0.00000392, -0.00000914, -0.00001364, 0.00001337, 0.00000198, -0.00001098, 0.00001850, -0.00000289, -0.00000617, 0.00001070, -0.00001403, -0.00000288, 0.00000275, -0.00001080, 0.00001415, 0.00000449, -0.00000918, 0.00001835, 0.00000171, -0.00001816, 0.00000280, -0.00000011, -0.00000801, 0.00000031, 0.00000062, 0.00000314, -0.00000458, -0.00001495, 0.00000093, 0.00001058, 0.00000292, -0.00000202, -0.00000541, -0.00000144, -0.00000363, -0.00001093, 0.00000227, 0.00000556, -0.00001201, -0.00000988, 0.00000348, 0.00000222, -0.00000113, 0.00000134, -0.00000555, -0.00001371, -0.00000142, 0.00000813, },
		{ -0.00000365, -0.00000627, -0.00000826, -0.00000180, 0.00000602, 0.00000953, -0.00000142, -0.00000955, -0.00000338, -0.00000009, 0.00000059, -0.00000211, -0.00000084, 0.00000461, -0.00001009, -0.00000926, 0.00001155, -0.00000456, -0.00000745, 0.00001454, -0.00000598, -0.00000618, 0.00001991, -0.00000301, -0.00001241, 0.00000372, -0.00000249, 0.00000067, 0.00000075, -0.00000112, 0.00001020, 0.00000157, -0.00000484, -0.00000302, 0.00000157, 0.00000283, -0.00001861, 0.00000521, 0.00001351, -0.00002789, 0.00001436, 0.00000518, -0.00005035, 0.00001754, 0.00000967, -0.00003786, 0.00000123, -0.00000261, -0.00001310, -0.00003780, 0.00001813, -0.00001925, -0.00020170, 0.00014051, -0.00016369, -0.00110078, 0.00061841, 0.00230185, -0.00008731, -0.00131224, 0.00034920, 0.00006877, -0.00048903, 0.00031550, -0.00003256, -0.00025019, 0.00024232, -0.00009297, -0.00011582, 0.00016126, -0.00011244, -0.00004345, 0.00009967, -0.00007114, -0.00000202, 0.00003696, -0.00004090, 0.00001033, 0.00001187, -0.00001553, -0.00001006, -0.00000198, 0.00001310, -0.00000859, 0.00000097, 0.00001557, -0.00001104, 0.00000490, 0.00000132, -0.00001684, 0.00000814, -0.00000599, -0.00000700, 0.00002075, -0.00000605, -0.00000241, 0.00002262, -0.00001166, -0.00001608, 0.00000941, -0.00000411, -0.00000866, 0.00000309, 0.00000255, -0.00000059, -0.00001063, -0.00000919, 0.00000794, 0.00000774, -0.00000015, -0.00000413, -0.00000549, -0.00000027, -0.00000584, -0.00000717, 0.00000592, -0.00000287, -0.00001505, -0.00000193, 0.00000549, -0.00000034, 0.00000039, -0.00000094, -0.00001242, -0.00000958, 0.00000747, 0.00000454, -0.00000577, },
		{ -0.00000890, -0.00000808, 0.00000344, 0.00000984, 0.00000496, -0.00000819, -0.00000745, 0.00000061, -0.00000020, -0.00000060, -0.00000190, 0.00000139, 0.00000062, -0.00001268, -0.00000083, 0.00000959, -0.00000984, 0.00000100, 0.00001039, -0.00001189, 0.00000516, 0.00001598, -0.00001133, -0.00000491, 0.00000299, -0.00000395, 0.00000350, -0.00000189, 0.00000270, 0.00001088, -0.00000472, -0.00000490, 0.00000142, 0.00000282, -0.00000674, -0.00001272, 0.00001489, -0.00000576, -0.00001639, 0.00002446, -0.00002993, -0.00002924, 0.00003757, -0.00002670, -0.00002388, 0.00001839, -0.00002648, -0.00001735, -0.00000686, -0.00000071, -0.00010324, -0.00007331, 0.00011027, -0.00063640, -0.00062255, 0.00161878, 0.00155278, -0.00100994, -0.00064127, 0.00050490, -0.00036497, -0.00015167, 0.00032822, -0.00028395, -0.00000295, 0.00018751, -0.00022689, 0.00006455, 0.00007893, -0.00017315, 0.00008143, 0.00003405, -0.00009281, 0.00005994, -0.00000606, -0.00004083, 0.00003764, -0.00000874, -0.00002002, -0.00000051, 0.00000315, 0.00000582, -0.00000901, 0.00001027, 0.00000546, -0.00000858, 0.00000943, -0.00001197, -0.00000732, 0.00001094, -0.00001670, 0.00000601, 0.00001879, -0.00001557, 0.00001138, 0.00001742, -0.00002496, -0.00000539, 0.00001225, -0.00001174, -0.00000534, 0.00000733, -0.00000002, -0.00000624, -0.00001057, -0.00000128, 0.00000973, 0.00000383, -0.00000232, -0.00000623, -0.00000407, 0.00000043, -0.00000683, -0.00000249, 0.00000386, -0.00001121, -0.00001090, 0.00000493, 0.00000331, -0.00000081, 0.00000120, -0.00000746, -0.00001462, 0.00000043, 0.00001003, -0.00000360, -0.00000585, -0.00000068, },
		{ -0.00000356, 0.00000950, 0.00000872, -0.00000240, -0.00001052, -0.00000244, 0.00000193, -0.00000148, -0.00000061, -0.00000120, 0.00000162, -0.00000355, -0.00001092, 0.00000538, 0.00000416, -0.00000994, 0.00000699, 0.00000252, -0.00001036, 0.00001349, 0.00000681, -0.00001242, 0.00000113, -0.00000090, -0.00000166, 0.00000376, -0.00000355, 0.00000856, 0.00000620, -0.00000982, 0.00000041, 0.00000369, -0.00000204, -0.00001019, -0.00000214, 0.00001051, -0.00001556, 0.00000350, 0.00000799, -0.00004685, 0.00000989, 0.00002068, -0.00004833, 0.00001248, 0.00000124, -0.00004460, 0.00001170, -0.00000918, -0.00005633, -0.00008866, 0.00002510, -0.00016004, -0.00079484, 0.00026572, 0.00194545, 0.00043086, -0.00118798, 0.00003350, 0.00021483, -0.00047178, 0.00019625, 0.00008208, -0.00029701, 0.00020862, -0.00001964, -0.00017872, 0.00018074, -0.00008588, -0.00009909, 0.00014445, -0.00007450, -0.00003423, 0.00007806, -0.00006324, 0.00000105, 0.00003923, -0.00003703, -0.00000500, 0.00000518, -0.00000027, 0.00000246, -0.00000418, 0.00001135, -0.00000181, -0.00000112, 0.00000305, -0.00001735, 0.00000693, 0.00000116, -0.00001841, 0.00002131, 0.00000571, -0.00001534, 0.00002523, 0.00000059, -0.00002907, 0.00001011, 0.00000582, -0.00001739, 0.00000354, 0.00000738, -0.00000657, -0.00000805, -0.00000588, 0.00000385, 0.00000811, 0.00000075, -0.00000475, -0.00000732, -0.00000108, 0.00000029, -0.00000649, 0.00000018, -0.00000256, -0.00001442, -0.00000258, 0.00000698, 0.00000042, 0.00000080, -0.00000215, -0.00001410, -0.00000876, 0.00000945, 0.00000399, -0.00000921, -0.00000049, -0.00000139, -0.00001300, },
		{ 0.00001181, 0.00000290, -0.00000803, -0.00000802, 0.00000176, 0.00000066, -0.00000185, -0.00000013, -0.00000120, 0.00000091, -0.00000623, -0.00000700, 0.00000811, -0.00000135, -0.00000665, 0.00000853, -0.00000426, -0.00000397, 0.00001562, -0.00000174, -0.00000843, 0.00000251, -0.00000354, 0.00000215, 0.00000093, -0.00000118, 0.00001196, -0.00000280, -0.00000834, 0.00000665, -0.00000014, -0.00000593, -0.00000680, 0.00000271, 0.00000123, -0.00001068, 0.00001131, -0.00001931, -0.00003202, 0.00003311, -0.00001776, -0.00003210, 0.00003301, -0.00004009, -0.00002272, 0.00003164, -0.00005844, -0.00006933, -0.00001978, -0.00003743, -0.00045307, -0.00048755, 0.00108691, 0.00154717, -0.00047465, -0.00080060, 0.00033288, -0.00017184, -0.00025543, 0.00030442, -0.00018850, -0.00009722, 0.00022320, -0.00019860, 0.00000179, 0.00013445, -0.00019333, 0.00005463, 0.00008621, -0.00013107, 0.00006246, 0.00002348, -0.00007911, 0.00005633, 0.00000044, -0.00004292, 0.00001893, -0.00000445, -0.00000086, 0.00000434, -0.00000258, 0.00000912, -0.00000253, 0.00000169, -0.00000735, -0.00001023, 0.00001341, -0.00001380, -0.00000640, 0.00002729, -0.00001119, -0.00000211, 0.00002815, -0.00002065, -0.00001843, 0.00002017, -0.00000829, -0.00001397, 0.00001260, -0.00000014, -0.00001115, -0.00000432, -0.00000158, 0.00000543, 0.00000585, -0.00000220, -0.00000741, -0.00000593, 0.00000231, -0.00000083, -0.00000565, -0.00000085, -0.00000912, -0.00001092, 0.00000458, 0.00000473, -0.00000004, 0.00000102, -0.00000946, -0.00001478, 0.00000224, 0.00001050, -0.00000584, -0.00000708, 0.00000483, -0.00000757, -0.00001359, 0.00000492, },
		{ -0.00000377, -0.00000929, -0.00000349, 0.00000319, -0.00000109, -0.00000103, -0.00000051, -0.00000137, 0.00000030, -0.00000765, -0.00000261, 0.00000806, -0.00000534, -0.00000245, 0.00000668, -0.00000752, 0.00000312, 0.00001285, -0.00000633, -0.00000384, 0.00000056, -0.00000280, 0.00000402, -0.00000192, 0.00000438, 0.00000895, -0.00000997, -0.00000032, 0.00000703, -0.00000619, -0.00000411, -0.00000372, 0.00000103, -0.00000197, -0.00000198, 0.00000107, -0.00003284, -0.00000118, 0.00002279, -0.00004191, 0.00000682, 0.00001220, -0.00006244, 0.00002705, -0.00000136, -0.00010130, -0.00001465, -0.00001363, -0.00022974, -0.00051777, 0.00012193, 0.00145656, 0.00076060, -0.00084713, -0.00027142, 0.00023786, -0.00035251, 0.00004823, 0.00015526, -0.00027605, 0.00012743, 0.00005755, -0.00020935, 0.00015666, -0.00003093, -0.00015168, 0.00016122, -0.00004761, -0.00008059, 0.00011237, -0.00006519, -0.00002726, 0.00007298, -0.00005290, -0.00001144, 0.00002501, -0.00002081, 0.00001025, 0.00000188, -0.00000384, 0.00001038, -0.00000262, -0.00000247, -0.00001124, 0.00000203, 0.00000676, -0.00002107, 0.00001218, 0.00001846, -0.00001956, 0.00001649, 0.00001493, -0.00003300, 0.00000188, 0.00001630, -0.00002031, -0.00000101, 0.00001339, -0.00001115, -0.00000841, 0.00000073, -0.00000054, 0.00000591, 0.00000339, -0.00000597, -0.00000849, -0.00000201, 0.00000441, -0.00000273, -0.00000570, -0.00000465, -0.00001154, -0.00000355, 0.00000700, 0.00000196, 0.00000137, -0.00000347, -0.00001562, -0.00000705, 0.00001045, 0.00000258, -0.00001112, 0.00000119, 0.00000380, -0.00001522, -0.00000720, 0.00001257, 0.00000803, },
		{ -0.00000716, 0.00000011, 0.00000256, -0.00000179, -0.00000034, -0.00000172, -0.00000059, -0.00000057, -0.00000826, 0.00000177, 0.00000597, -0.00000764, 0.00000142, 0.00000323, -0.00000754, 0.00000850, 0.00000823, -0.00000743, -0.00000137, -0.00000145, -0.00000007, 0.00000303, -0.00000170, 0.00000860, 0.00000087, -0.00000957, 0.00000760, 0.00000035, -0.00000732, -0.00000010, -0.00000578, 0.00000047, 0.00000099, -0.00000220, -0.00001417, -0.00002377, 0.00001782, -0.00000655, -0.00003171, 0.00002985, -0.00003306, -0.00003647, 0.00005213, -0.00007496, -0.00007636, 0.00003927, -0.00013201, -0.00038086, -0.00027286, 0.00069813, 0.00130453, 0.00002594, -0.00074765, 0.00006199, -0.00001594, -0.00026757, 0.00020506, -0.00007508, -0.00015588, 0.00020284, -0.00013083, -0.00006499, 0.00016169, -0.00017289, 0.00000270, 0.00013368, -0.00014384, 0.00003800, 0.00006568, -0.00011035, 0.00005644, 0.00002561, -0.00007210, 0.00003513, -0.00000065, -0.00002110, 0.00002443, -0.00001106, 0.00000124, 0.00001275, -0.00000836, -0.00000616, -0.00000620, 0.00000804, -0.00000613, -0.00001454, 0.00002435, 0.00000129, -0.00001318, 0.00002699, -0.00000745, -0.00002799, 0.00001833, 0.00000052, -0.00002020, 0.00001271, 0.00000299, -0.00001671, 0.00000027, 0.00000167, -0.00000088, 0.00000675, -0.00000075, -0.00000935, -0.00000635, 0.00000261, 0.00000432, -0.00000533, -0.00000742, -0.00000804, -0.00000862, 0.00000273, 0.00000550, 0.00000159, 0.00000072, -0.00001147, -0.00001442, 0.00000452, 0.00000929, -0.00000770, -0.00000714, 0.00000816, -0.00000459, -0.00001743, 0.00000405, 0.00001400, 0.00000052, -0.00000721, },
		{ 0.00000200, 0.00000151, -0.00000176, -0.00000074, -0.00000236, 0.00000099, -0.00000258, -0.00000752, 0.00000568, 0.00000206, -0.00000793, 0.00000450, -0.00000087, -0.00000523, 0.00001188, 0.00000355, -0.00000704, -0.00000070, -0.00000175, 0.00000211, 0.00000117, 0.00000130, 0.00000775, -0.00000617, -0.00000231, 0.00000843, -0.00000706, -0.00000190, -0.00000107, -0.00000894, 0.00000575, 0.00000065, -0.00001115, -0.00001828, -0.00000564, 0.00001287, -0.00002535, -0.00000189, 0.00001501, -0.00005898, 0.00001758, 0.00001458, -0.00012286, 0.00000857, -0.00000427, -0.00030289, -0.00033217, 0.00014084, 0.00098190, 0.00084472, -0.00039463, -0.00045183, 0.00012257, -0.00018537, -0.00006532, 0.00015535, -0.00019717, 0.00003360, 0.00010452, -0.00019278, 0.00009750, 0.00002835, -0.00017711, 0.00014066, 0.00000033, -0.00012202, 0.00012364, -0.00004250, -0.00006629, 0.00009976, -0.00005315, -0.00003221, 0.00005274, -0.00003864, 0.00000571, 0.00002156, -0.00002457, 0.00001690, 0.00000619, -0.00001592, -0.00000215, -0.00000018, 0.00000445, -0.00001375, -0.00000017, 0.00002321, -0.00001166, 0.00000203, 0.00002189, -0.00002446, -0.00001000, 0.00002003, -0.00001511, -0.00000749, 0.00001610, -0.00001163, -0.00001136, 0.00000751, -0.00000209, 0.00000083, 0.00000617, -0.00000662, -0.00000969, -0.00000172, 0.00000578, 0.00000212, -0.00000886, -0.00000969, -0.00000798, -0.00000307, 0.00000500, 0.00000376, 0.00000248, -0.00000497, -0.00001687, -0.00000470, 0.00001108, -0.00000036, -0.00001142, 0.00000272, 0.00000697, -0.00001445, -0.00001067, 0.00001335, 0.00000804, -0.00000557, -0.00000664, -0.00000446, },
		{ 0.00000055, -0.00000200, -0.00000161, -0.00000142, 0.00000188, -0.00000545, -0.00000462, 0.00000780, -0.00000293, -0.00000566, 0.00000588, -0.00000503, -0.00000085, 0.00001331, -0.00000100, -0.00000616, -0.00000045, -0.00000084, 0.00000289, 0.00000053, 0.00000388, 0.00000283, -0.00000708, 0.00000513, 0.00000201, -0.00000792, 0.00000369, -0.00000830, -0.00000543, 0.00001138, -0.00000882, -0.00001684, -0.00000969, 0.00000261, -0.00000320, -0.00001982, 0.00001564, -0.00002095, -0.00004044, 0.00004634, -0.00006295, -0.00009104, 0.00006701, -0.00016279, -0.00037203, -0.00007768, 0.00048892, 0.00093971, 0.00036484, -0.00050232, -0.00020290, 0.00003521, -0.00018966, 0.00007554, 0.00000698, -0.00015564, 0.00013598, -0.00005299, -0.00010654, 0.00014715, -0.00011981, -0.00005174, 0.00015494, -0.00012375, -0.00000471, 0.00010325, -0.00012016, 0.00003458, 0.00005962, -0.00009601, 0.00003816, 0.00001815, -0.00004998, 0.00004094, -0.00000648, -0.00001955, 0.00003156, -0.00001303, -0.00001377, 0.00000594, -0.00000133, -0.00000152, -0.00001190, 0.00001117, 0.00001341, -0.00001355, 0.00001379, 0.00000636, -0.00002649, 0.00000719, 0.00000837, -0.00001891, 0.00000738, 0.00000619, -0.00001904, 0.00000109, 0.00000713, -0.00000579, 0.00000463, 0.00000165, -0.00001122, -0.00000592, 0.00000285, 0.00000668, -0.00000198, -0.00001305, -0.00000990, -0.00000430, 0.00000094, 0.00000452, 0.00000406, 0.00000048, -0.00001341, -0.00001361, 0.00000715, 0.00000701, -0.00001007, -0.00000521, 0.00000967, -0.00000227, -0.00001776, 0.00000153, 0.00001468, -0.00000093, -0.00000677, -0.00000460, -0.00000272, 0.00000312, },
		{ -0.00000255, -0.00000164, 0.00000042, 0.00000082, -0.00000750, -0.00000007, 0.00000678, -0.00000701, -0.00000129, 0.00000438, -0.00000780, 0.00000526, 0.00001213, -0.00000522, -0.00000459, -0.00000008, 0.00000028, 0.00000303, 0.00000092, 0.00000372, -0.00000161, -0.00000258, 0.00000660, -0.00000500, -0.00000193, 0.00000202, -0.00001444, 0.00000565, 0.00000756, -0.00002085, -0.00001059, -0.00000148, -0.00000287, -0.00001011, -0.00000327, 0.00000589, -0.00004156, 0.00000210, 0.00001742, -0.00011285, -0.00000157, 0.00001130, -0.00033384, -0.00023977, 0.00023246, 0.00062703, 0.00071963, 0.00003034, -0.00044696, -0.00007933, -0.00004816, -0.00010159, 0.00008874, -0.00009843, -0.00003340, 0.00010294, -0.00013729, 0.00002870, 0.00006559, -0.00016337, 0.00009111, 0.00004847, -0.00013971, 0.00010568, -0.00000335, -0.00010046, 0.00010707, -0.00003452, -0.00006170, 0.00007750, -0.00004335, -0.00001464, 0.00004925, -0.00004042, 0.00001012, 0.00002578, -0.00003214, 0.00000281, 0.00000614, -0.00000714, -0.00000184, -0.00000654, 0.00001501, 0.00000428, -0.00000843, 0.00001547, -0.00000730, -0.00001680, 0.00001284, -0.00000484, -0.00001026, 0.00001237, -0.00000859, -0.00001364, 0.00001081, 0.00000023, -0.00000483, 0.00000646, -0.00000589, -0.00001090, -0.00000023, 0.00000547, 0.00000551, -0.00000804, -0.00001596, -0.00000622, -0.00000013, 0.00000175, 0.00000465, 0.00000445, -0.00000664, -0.00001765, -0.00000206, 0.00001157, -0.00000444, -0.00001098, 0.00000538, 0.00000734, -0.00001220, -0.00001149, 0.00001112, 0.00000795, -0.00000670, -0.00000416, -0.00000383, -0.00000122, 0.00000404, -0.00000093, },
		{ -0.00000060, 0.00000159, -0.00000177, -0.00000722, 0.00000414, 0.00000286, -0.00000810, 0.00000309, -0.00000000, -0.00000719, 0.00001145, 0.00000784, -0.00000800, -0.00000232, 0.00000010, 0.00000147, 0.00000304, 0.00000077, 0.00000185, -0.00000230, 0.00000209, 0.00000233, -0.00000615, 0.00000390, -0.00000672, -0.00001121, 0.00001510, -0.00000740, -0.00002280, 0.00000182, -0.00000479, -0.00000907, -0.00000314, 0.00000217, -0.00001569, -0.00002982, 0.00002372, -0.00004525, -0.00008894, 0.00005591, -0.00015398, -0.00036914, 0.00004384, 0.00042102, 0.00058272, 0.00048998, -0.00015897, -0.00035925, -0.00003242, -0.00007013, -0.00002651, 0.00002771, -0.00010431, 0.00005437, 0.00000232, -0.00010726, 0.00009790, -0.00005731, -0.00008463, 0.00014011, -0.00007917, -0.00004697, 0.00011953, -0.00010333, -0.00000167, 0.00008889, -0.00010339, 0.00002396, 0.00004713, -0.00007624, 0.00004343, 0.00001378, -0.00004687, 0.00004412, -0.00000587, -0.00003067, 0.00002154, -0.00000718, -0.00000654, 0.00000168, -0.00000451, 0.00001596, -0.00000091, -0.00000439, 0.00001128, -0.00001204, -0.00000671, 0.00000835, -0.00000909, 0.00000055, 0.00000503, -0.00001609, -0.00000067, 0.00001115, -0.00000635, -0.00000014, 0.00000279, -0.00001154, -0.00000538, 0.00000388, 0.00000629, 0.00000204, -0.00001520, -0.00001460, -0.00000005, 0.00000105, 0.00000158, 0.00000654, 0.00000074, -0.00001529, -0.00001219, 0.00000959, 0.00000435, -0.00001307, -0.00000188, 0.00001069, -0.00000227, -0.00001479, -0.00000042, 0.00001226, -0.00000111, -0.00000650, -0.00000166, -0.00000441, 0.00000128, 0.00000387, -0.00000304, -0.00000239, },
		{ 0.00000130, -0.00000429, -0.00000472, 0.00000603, -0.00000189, -0.00000579, 0.00000489, -0.00000516, -0.00000243, 0.00001506, 0.00000146, -0.00000824, -0.00000020, 0.00000014, 0.00000289, 0.00000225, -0.00000028, 0.00000095, -0.00000035, 0.00000287, -0.00000224, -0.00000172, 0.00000283, -0.00001432, 0.00000080, 0.00001256, -0.00002285, -0.00001051, 0.00000614, -0.00001526, -0.00000376, 0.00000379, -0.00000851, -0.00002348, -0.00000507, 0.00000230, -0.00008380, -0.00001732, 0.00000481, -0.00031578, -0.00020128, 0.00031313, 0.00042437, 0.00048602, 0.00031542, -0.00027552, -0.00027426, -0.00000069, -0.00006167, -0.00000436, -0.00002451, -0.00005132, 0.00006023, -0.00006959, -0.00002155, 0.00006667, -0.00011732, 0.00003434, 0.00007550, -0.00012568, 0.00006635, 0.00003543, -0.00011497, 0.00009052, -0.00000338, -0.00008801, 0.00008797, -0.00003028, -0.00004555, 0.00007329, -0.00004164, -0.00001135, 0.00004972, -0.00004243, -0.00000241, 0.00002190, -0.00002301, 0.00000581, -0.00000066, -0.00000371, 0.00001883, -0.00000694, -0.00000295, 0.00000921, -0.00001183, -0.00000317, 0.00000361, -0.00000505, 0.00000314, -0.00000582, -0.00001164, 0.00000930, 0.00000426, -0.00000742, 0.00000277, -0.00000464, -0.00001086, 0.00000144, 0.00000502, 0.00000613, -0.00000448, -0.00002044, -0.00000778, 0.00000434, -0.00000081, 0.00000363, 0.00000699, -0.00000812, -0.00001790, 0.00000076, 0.00001154, -0.00000868, -0.00001056, 0.00000920, 0.00000586, -0.00001058, -0.00000857, 0.00000714, 0.00000606, -0.00000610, -0.00000274, -0.00000186, -0.00000419, 0.00000423, 0.00000162, -0.00000452, -0.00000096, 0.00000050, },
		{ -0.00000556, -0.00000146, 0.00000520, -0.00000511, -0.00000181, 0.00000309, -0.00000805, 0.00000487, 0.00001425, -0.00000463, -0.00000610, 0.00000093, 0.00000070, 0.00000386, 0.00000014, -0.00000067, 0.00000191, 0.00000081, 0.00000064, -0.00000274, 0.00000232, -0.00000443, -0.00001310, 0.00001166, -0.00000213, -0.00002558, 0.00000558, -0.00000401, -0.00001814, 0.00000889, -0.00000291, -0.00001914, -0.00001194, 0.00000040, -0.00003840, -0.00006582, 0.00002258, -0.00013965, -0.00033713, 0.00009126, 0.00041878, 0.00032401, 0.00043428, 0.00015991, -0.00035544, -0.00017198, 0.00002283, -0.00006362, -0.00000770, -0.00003519, -0.00000637, 0.00001730, -0.00007336, 0.00003718, -0.00001095, -0.00008355, 0.00009654, -0.00002998, -0.00007022, 0.00010770, -0.00006709, -0.00003718, 0.00010042, -0.00008985, -0.00000256, 0.00007479, -0.00008812, 0.00002767, 0.00004364, -0.00006925, 0.00004201, 0.00001368, -0.00005109, 0.00003248, -0.00000306, -0.00002201, 0.00001961, -0.00001290, 0.00000464, 0.00001940, -0.00001728, 0.00000239, 0.00000990, -0.00001456, -0.00000229, 0.00000488, -0.00000213, -0.00000256, -0.00000960, -0.00000138, 0.00001052, -0.00000245, -0.00000463, 0.00000018, -0.00000980, -0.00000441, 0.00000514, 0.00000474, 0.00000434, -0.00001339, -0.00001976, 0.00000159, 0.00000361, -0.00000201, 0.00000765, 0.00000173, -0.00001690, -0.00001007, 0.00001144, 0.00000146, -0.00001586, 0.00000165, 0.00001211, -0.00000452, -0.00001046, 0.00000024, 0.00000676, -0.00000117, -0.00000526, 0.00000008, -0.00000381, -0.00000174, 0.00000550, -0.00000195, -0.00000389, 0.00000079, 0.00000064, 0.00000012, },
		{ 0.00000118, 0.00000279, -0.00000577, 0.00000137, -0.00000102, -0.00000681, 0.00001154, 0.00000940, -0.00000820, -0.00000299, 0.00000121, 0.00000197, 0.00000317, -0.00000212, 0.00000099, 0.00000281, -0.00000029, -0.00000080, -0.00000020, 0.00000124, -0.00001083, -0.00000398, 0.00001177, -0.00001761, -0.00001331, 0.00001042, -0.00001904, -0.00000602, 0.00001290, -0.00001865, -0.00001368, -0.00000138, -0.00002144, -0.00005513, -0.00002003, -0.00002677, -0.00027069, -0.00017076, 0.00034001, 0.00034058, 0.00025520, 0.00041805, -0.00002202, -0.00037405, -0.00005278, 0.00001167, -0.00007420, -0.00000154, -0.00002635, 0.00000383, -0.00001917, -0.00003778, 0.00003682, -0.00006035, -0.00000640, 0.00007128, -0.00008725, 0.00002333, 0.00005733, -0.00010391, 0.00005670, 0.00002709, -0.00009923, 0.00007896, -0.00000382, -0.00007506, 0.00008219, -0.00002514, -0.00004078, 0.00006746, -0.00003968, -0.00001846, 0.00004181, -0.00003463, 0.00000238, 0.00001684, -0.00002422, 0.00002266, 0.00000782, -0.00002515, 0.00001584, 0.00000520, -0.00002044, 0.00000400, 0.00000760, -0.00000672, -0.00000738, -0.00000450, 0.00000523, 0.00000618, -0.00000475, -0.00000362, -0.00000533, -0.00000824, 0.00000261, 0.00000493, 0.00000476, -0.00000091, -0.00002098, -0.00001181, 0.00000778, -0.00000093, 0.00000092, 0.00000889, -0.00000883, -0.00001759, 0.00000379, 0.00001050, -0.00001234, -0.00001000, 0.00001307, 0.00000405, -0.00001065, -0.00000299, 0.00000384, 0.00000127, -0.00000402, -0.00000181, -0.00000011, -0.00000472, 0.00000172, 0.00000385, -0.00000446, -0.00000136, 0.00000135, 0.00000011, -0.00000102, -0.00000562, },
		{ 0.00000020, -0.00000451, 0.00000226, -0.00000491, -0.00000193, 0.00001490, 0.00000272, -0.00000864, -0.00000030, 0.00000141, 0.00000299, 0.00000071, -0.00000240, 0.00000369, 0.00000153, -0.00000171, 0.00000033, 0.00000106, -0.00000362, -0.00001077, 0.00000464, 0.00000144, -0.00002182, 0.00000255, -0.00000023, -0.00002307, 0.00001132, -0.00000017, -0.00002507, 0.00000237, -0.00001280, -0.00004463, -0.00003386, -0.00000967, -0.00014079, -0.00027409, 0.00009805, 0.00041429, 0.00018830, 0.00028571, 0.00035959, -0.00021205, -0.00030156, 0.00003984, -0.00003844, -0.00006442, 0.00001355, -0.00002678, -0.00000246, -0.00002776, -0.00000935, 0.00000518, -0.00005382, 0.00004427, 0.00000340, -0.00006612, 0.00007445, -0.00002735, -0.00005726, 0.00008904, -0.00006079, -0.00002974, 0.00008911, -0.00008004, -0.00000128, 0.00007087, -0.00007616, 0.00002350, 0.00003900, -0.00006574, 0.00003202, 0.00001201, -0.00004220, 0.00003227, -0.00000788, -0.00001724, 0.00003730, -0.00001738, -0.00001765, 0.00002953, -0.00001076, -0.00001991, 0.00001579, 0.00000251, -0.00001420, -0.00000392, 0.00000199, 0.00000523, 0.00000282, -0.00000537, -0.00000678, -0.00000750, -0.00000170, 0.00000554, 0.00000353, 0.00000420, -0.00000959, -0.00002224, -0.00000039, 0.00000704, -0.00000400, 0.00000652, 0.00000287, -0.00001766, -0.00000737, 0.00001262, -0.00000198, -0.00001742, 0.00000466, 0.00001350, -0.00000750, -0.00000685, 0.00000385, 0.00000018, -0.00000287, -0.00000240, 0.00000024, -0.00000174, -0.00000344, 0.00000366, 0.00000057, -0.00000419, 0.00000105, 0.00000037, -0.00000006, -0.00000202, -0.00000505, 0.00000279, },
		{ -0.00000258, 0.00000102, -0.00000660, 0.00000446, 0.00001413, -0.00000329, -0.00000672, 0.00000150, 0.00000183, 0.00000256, -0.00000177, 0.00000002, 0.00000473, -0.00000134, -0.00000091, 0.00000188, -0.00000108, -0.00000711, -0.00000566, 0.00000575, -0.00000980, -0.00001347, 0.00000824, -0.00001549, -0.00001095, 0.00001669, -0.00001965, -0.00001223, 0.00000629, -0.00004057, -0.00003923, -0.00000841, -0.00006762, -0.00022653, -0.00012271, 0.00031600, 0.00031212, 0.00010084, 0.00037073, 0.00020784, -0.00033900, -0.00016056, 0.00006051, -0.00009072, -0.00002396, 0.00001316, -0.00003605, -0.00000179, -0.00002442, -0.00000398, -0.00001591, -0.00001882, 0.00004172, -0.00004257, -0.00000565, 0.00005509, -0.00007336, 0.00001984, 0.00004389, -0.00009013, 0.00005454, 0.00002459, -0.00009076, 0.00007105, 0.00000311, -0.00006628, 0.00007005, -0.00002295, -0.00003952, 0.00005715, -0.00003497, -0.00001282, 0.00003860, -0.00003639, 0.00001197, 0.00002970, -0.00004078, 0.00000874, 0.00002776, -0.00003052, -0.00000495, 0.00002221, -0.00001179, -0.00001409, 0.00000483, 0.00000257, 0.00000363, 0.00000193, -0.00000913, -0.00001000, -0.00000331, 0.00000396, 0.00000445, 0.00000322, 0.00000051, -0.00001785, -0.00001540, 0.00000774, 0.00000144, -0.00000171, 0.00000867, -0.00000861, -0.00001644, 0.00000690, 0.00000828, -0.00001535, -0.00000846, 0.00001568, 0.00000259, -0.00001171, 0.00000287, 0.00000300, -0.00000560, -0.00000214, -0.00000017, 0.00000014, -0.00000241, -0.00000153, 0.00000316, -0.00000163, -0.00000193, 0.00000149, -0.00000079, 0.00000045, -0.00000339, -0.00000396, 0.00000359, -0.00000184, },
		{ -0.00000137, -0.00000549, 0.00001006, 0.00001011, -0.00000709, -0.00000365, 0.00000250, 0.00000187, 0.00000079, -0.00000216, 0.00000324, 0.00000273, -0.00000291, 0.00000173, 0.00000087, -0.00000425, -0.00000650, -0.00000152, 0.00000045, -0.00001318, -0.00000226, 0.00000092, -0.00002079, 0.00000612, 0.00000421, -0.00002615, 0.00000724, -0.00001540, -0.00005521, -0.00000884, -0.00002712, -0.00015714, -0.00020026, 0.00010083, 0.00037586, 0.00014458, 0.00013784, 0.00040921, -0.00000564, -0.00034808, -0.00002264, 0.00001067, -0.00010116, 0.00002133, -0.00001275, -0.00003394, 0.00000457, -0.00002774, -0.00000532, -0.00001401, 0.00000457, 0.00001095, -0.00004055, 0.00003579, 0.00000013, -0.00005543, 0.00006019, -0.00002861, -0.00004551, 0.00008407, -0.00005578, -0.00003146, 0.00008348, -0.00006406, -0.00000480, 0.00005935, -0.00006706, 0.00001910, 0.00003264, -0.00005799, 0.00003255, 0.00000953, -0.00004048, 0.00004368, -0.00000414, -0.00004014, 0.00003836, 0.00000362, -0.00003656, 0.00001842, 0.00001250, -0.00002443, -0.00000257, 0.00000902, -0.00000137, 0.00000554, -0.00000125, -0.00001557, -0.00000737, 0.00000358, 0.00000491, 0.00000281, 0.00000284, -0.00000666, -0.00002041, -0.00000443, 0.00000849, -0.00000287, 0.00000378, 0.00000276, -0.00001690, -0.00000420, 0.00001317, -0.00000619, -0.00001730, 0.00000740, 0.00001366, -0.00000967, -0.00000463, 0.00000871, -0.00000481, -0.00000690, 0.00000118, -0.00000007, 0.00000002, -0.00000211, -0.00000097, 0.00000197, -0.00000150, -0.00000014, 0.00000022, -0.00000065, 0.00000062, -0.00000515, -0.00000203, 0.00000354, -0.00000315, -0.00000085, },
		{ -0.00000198, 0.00001334, 0.00000443, -0.00000820, -0.00000042, 0.00000262, 0.00000104, -0.00000073, -0.00000019, 0.00000448, -0.00000065, -0.00000139, 0.00000312, -0.00000255, -0.00000492, -0.00000426, -0.00000183, -0.00000440, -0.00000886, 0.00000173, -0.00000997, -0.00001261, 0.00001251, -0.00001379, -0.00001313, 0.00000956, -0.00004787, -0.00003596, 0.00000646, -0.00009902, -0.00019837, -0.00005893, 0.00027018, 0.00028489, 0.00003900, 0.00025093, 0.00033703, -0.00019339, -0.00025269, 0.00004656, -0.00005974, -0.00006143, 0.00003601, -0.00004419, -0.00001249, -0.00000044, -0.00003386, 0.00000319, -0.00000678, 0.00000424, -0.00001016, -0.00001225, 0.00003391, -0.00003761, -0.00000592, 0.00004119, -0.00006501, 0.00002578, 0.00004267, -0.00008643, 0.00004449, 0.00003090, -0.00007705, 0.00005543, 0.00000140, -0.00005622, 0.00006075, -0.00002276, -0.00003383, 0.00005473, -0.00003426, -0.00000950, 0.00004867, -0.00004308, -0.00000859, 0.00004631, -0.00002921, -0.00001853, 0.00003190, -0.00001021, -0.00002242, 0.00001167, 0.00000361, -0.00000205, 0.00000857, -0.00001057, -0.00001804, 0.00000150, 0.00000690, 0.00000279, 0.00000261, -0.00000033, -0.00001353, -0.00001554, 0.00000385, 0.00000419, -0.00000195, 0.00000570, -0.00000819, -0.00001380, 0.00000976, 0.00000511, -0.00001800, -0.00000528, 0.00001665, 0.00000090, -0.00001213, 0.00000706, 0.00000444, -0.00001206, -0.00000233, 0.00000250, -0.00000083, 0.00000118, -0.00000291, -0.00000133, 0.00000228, -0.00000055, -0.00000029, -0.00000081, 0.00000064, -0.00000067, -0.00000624, 0.00000054, 0.00000243, -0.00000353, 0.00000015, -0.00000385, },
		{ 0.00001360, -0.00000121, -0.00000681, 0.00000204, 0.00000166, -0.00000007, -0.00000061, 0.00000220, 0.00000287, -0.00000230, 0.00000168, 0.00000111, -0.00000507, -0.00000292, -0.00000439, -0.00000405, -0.00000432, -0.00000409, -0.00000231, -0.00001376, -0.00000074, 0.00000442, -0.00001996, 0.00000449, -0.00001404, -0.00005994, 0.00000118, -0.00003090, -0.00017462, -0.00013789, 0.00011744, 0.00030939, 0.00013938, 0.00004863, 0.00034225, 0.00017349, -0.00028407, -0.00012242, 0.00003531, -0.00009583, -0.00000473, 0.00001153, -0.00005275, 0.00001119, -0.00002166, -0.00002592, 0.00001580, -0.00001140, -0.00000282, -0.00000911, 0.00000775, 0.00000792, -0.00003630, 0.00002665, -0.00000593, -0.00004395, 0.00006285, -0.00002637, -0.00005028, 0.00007617, -0.00003854, -0.00003120, 0.00006571, -0.00005373, -0.00000081, 0.00005004, -0.00006206, 0.00001972, 0.00003184, -0.00005515, 0.00003522, 0.00001611, -0.00005542, 0.00003378, 0.00002163, -0.00004495, 0.00001307, 0.00002187, -0.00002861, -0.00000506, 0.00001577, -0.00000520, 0.00000433, 0.00000510, -0.00002197, -0.00001091, 0.00001012, 0.00000458, 0.00000167, 0.00000192, -0.00000621, -0.00001530, -0.00000728, 0.00000588, 0.00000080, 0.00000155, -0.00000002, -0.00001460, -0.00000033, 0.00001289, -0.00001090, -0.00001584, 0.00001062, 0.00001171, -0.00001071, -0.00000284, 0.00001236, -0.00000673, -0.00001197, 0.00000339, 0.00000050, 0.00000049, 0.00000173, -0.00000566, -0.00000005, 0.00000385, -0.00000107, -0.00000134, -0.00000023, 0.00000143, -0.00000318, -0.00000562, 0.00000272, 0.00000057, -0.00000279, -0.00000006, -0.00000546, 0.00000299, },
		{ -0.00000528, -0.00000383, 0.00000288, 0.00000009, -0.00000031, 0.00000080, 0.00000291, 0.00000031, -0.00000102, 0.00000293, -0.00000284, -0.00000415, -0.00000178, -0.00000722, -0.00000345, -0.00000166, -0.00000423, -0.00000714, -0.00000936, 0.00000350, -0.00000721, -0.00001068, 0.00000530, -0.00004510, -0.00003704, 0.00001263, -0.00011221, -0.00018357, 0.00000285, 0.00023106, 0.00023662, 0.00004038, 0.00013629, 0.00034604, -0.00000698, -0.00027026, -0.00002714, -0.00001618, -0.00008046, 0.00002757, -0.00002719, -0.00003200, 0.00001403, -0.00004163, -0.00000015, 0.00001427, -0.00002300, 0.00000094, -0.00000090, 0.00000738, -0.00001096, -0.00001382, 0.00002300, -0.00003506, 0.00000457, 0.00004384, -0.00006521, 0.00001409, 0.00004630, -0.00006930, 0.00002973, 0.00002298, -0.00006095, 0.00005113, -0.00000308, -0.00005152, 0.00005810, -0.00001899, -0.00003275, 0.00005629, -0.00003218, -0.00002773, 0.00005384, -0.00001888, -0.00002924, 0.00003315, -0.00000527, -0.00002686, 0.00001352, 0.00000666, -0.00000665, 0.00001145, -0.00000799, -0.00002574, 0.00000316, 0.00001167, 0.00000062, 0.00000246, -0.00000192, -0.00001065, -0.00001137, -0.00000161, 0.00000435, 0.00000126, 0.00000115, -0.00000911, -0.00000914, 0.00001216, 0.00000124, -0.00002034, -0.00000065, 0.00001661, -0.00000204, -0.00001082, 0.00000936, 0.00000620, -0.00001552, -0.00000495, 0.00000492, -0.00000148, 0.00000402, -0.00000129, -0.00000780, 0.00000423, 0.00000393, -0.00000329, -0.00000114, 0.00000120, 0.00000047, -0.00000535, -0.00000343, 0.00000351, -0.00000097, -0.00000145, -0.00000169, -0.00000525, 0.00000813, 0.00001092, },
		{ -0.00000075, 0.00000197, -0.00000092, 0.00000072, 0.00000189, 0.00000178, -0.00000073, 0.00000130, 0.00000086, -0.00000511, -0.00000121, -0.00000429, -0.00000878, 0.00000071, -0.00000176, -0.00000735, -0.00000703, -0.00000438, -0.00000072, -0.00000992, 0.00000012, -0.00001530, -0.00005485, -0.00000115, -0.00003113, -0.00017940, -0.00009639, 0.00014361, 0.00023940, 0.00013062, 0.00002650, 0.00022670, 0.00026464, -0.00013745, -0.00019779, 0.00000757, -0.00005758, -0.00003789, 0.00002130, -0.00004818, -0.00000267, -0.00000691, -0.00003890, 0.00002523, -0.00000686, -0.00002343, 0.00001254, -0.00000212, -0.00000026, -0.00001191, 0.00000133, 0.00000099, -0.00002806, 0.00003555, -0.00000393, -0.00005121, 0.00005303, -0.00001128, -0.00004513, 0.00005572, -0.00003152, -0.00001876, 0.00005693, -0.00005280, -0.00000049, 0.00004938, -0.00005565, 0.00001576, 0.00003400, -0.00005794, 0.00001988, 0.00003492, -0.00004291, 0.00000517, 0.00002490, -0.00002671, -0.00000689, 0.00001787, -0.00000510, 0.00000182, 0.00000900, -0.00002328, -0.00001635, 0.00001458, 0.00000615, -0.00000023, 0.00000181, -0.00000761, -0.00000972, -0.00000638, -0.00000031, 0.00000433, 0.00000203, -0.00000555, -0.00001181, 0.00000478, 0.00001152, -0.00001555, -0.00001359, 0.00001455, 0.00000798, -0.00001148, -0.00000010, 0.00001342, -0.00000625, -0.00001545, 0.00000255, 0.00000202, 0.00000039, 0.00000590, -0.00000771, -0.00000542, 0.00000911, 0.00000048, -0.00000481, 0.00000087, 0.00000157, -0.00000166, -0.00000593, -0.00000100, 0.00000307, -0.00000133, -0.00000074, -0.00000409, -0.00000266, 0.00001186, 0.00000628, -0.00000686, },
		{ 0.00000032, -0.00000047, 0.00000199, 0.00000160, 0.00000050, 0.00000021, 0.00000189, -0.00000263, -0.00000387, -0.00000015, -0.00000886, -0.00000558, 0.00000408, -0.00000603, -0.00000782, -0.00000402, -0.00000480, -0.00000458, -0.00000394, -0.00000335, -0.00003902, -0.00003451, 0.00000606, -0.00011102, -0.00016991, 0.00004642, 0.00020970, 0.00017497, 0.00006089, 0.00007075, 0.00026902, 0.00014424, -0.00019704, -0.00012007, 0.00000027, -0.00006630, -0.00000329, -0.00000583, -0.00004112, 0.00001007, -0.00003113, -0.00001217, 0.00002799, -0.00003102, -0.00000551, 0.00001560, -0.00001103, -0.00000029, -0.00000696, 0.00000006, -0.00001191, -0.00000273, 0.00003021, -0.00003734, -0.00000789, 0.00004389, -0.00004802, 0.00000394, 0.00003308, -0.00005203, 0.00003309, 0.00001564, -0.00005858, 0.00004786, 0.00000226, -0.00004850, 0.00004991, -0.00001424, -0.00004039, 0.00004938, -0.00000639, -0.00003181, 0.00002820, -0.00000302, -0.00002442, 0.00001257, 0.00000710, -0.00000662, 0.00001048, -0.00000521, -0.00002859, 0.00000080, 0.00001600, 0.00000001, 0.00000163, -0.00000295, -0.00001001, -0.00000477, -0.00000519, 0.00000046, 0.00000650, -0.00000249, -0.00001249, -0.00000289, 0.00001442, -0.00000330, -0.00002193, 0.00000455, 0.00001640, -0.00000647, -0.00000818, 0.00001096, 0.00000620, -0.00001516, -0.00000830, 0.00000497, -0.00000053, 0.00000539, 0.00000210, -0.00001286, 0.00000226, 0.00001010, -0.00000506, -0.00000325, 0.00000285, 0.00000014, -0.00000333, -0.00000522, 0.00000067, 0.00000256, -0.00000088, -0.00000161, -0.00000575, 0.00000181, 0.00001304, 0.00000132, -0.00000616, 0.00000103, },
		{ 0.00000101, 0.00000221, 0.00000059, 0.00000048, 0.00000136, 0.00000002, -0.00000444, -0.00000103, -0.00000296, -0.00001094, 0.00000092, 0.00000235, -0.00001013, -0.00000404, -0.00000412, -0.00000779, -0.00000166, -0.00000126, -0.00002067, -0.00004339, -0.00000747, -0.00003675, -0.00016752, -0.00007063, 0.00016452, 0.00018647, 0.00010412, 0.00004440, 0.00012551, 0.00025845, 0.00002877, -0.00020302, -0.00006378, -0.00001801, -0.00005176, 0.00000725, -0.00002767, -0.00002070, 0.00000064, -0.00003606, 0.00001736, 0.00000531, -0.00003620, 0.00001722, 0.00000263, -0.00001346, 0.00000569, -0.00000936, -0.00000529, -0.00000726, 0.00001330, 0.00000462, -0.00003658, 0.00002483, 0.00000634, -0.00004270, 0.00003478, -0.00000998, -0.00002736, 0.00005142, -0.00003471, -0.00001935, 0.00005530, -0.00004312, -0.00000637, 0.00004227, -0.00004861, 0.00000431, 0.00003866, -0.00003423, 0.00000124, 0.00002051, -0.00002356, -0.00000310, 0.00001458, -0.00000432, 0.00000266, 0.00000753, -0.00002094, -0.00001908, 0.00001388, 0.00000930, -0.00000141, 0.00000130, -0.00000906, -0.00000591, -0.00000182, -0.00000700, 0.00000478, 0.00000561, -0.00001196, -0.00001026, 0.00001134, 0.00000912, -0.00001973, -0.00001073, 0.00001854, 0.00000370, -0.00001292, 0.00000396, 0.00001239, -0.00000547, -0.00001530, -0.00000108, 0.00000295, 0.00000136, 0.00000833, -0.00000653, -0.00001141, 0.00001094, 0.00000499, -0.00000850, 0.00000089, 0.00000256, -0.00000177, -0.00000380, -0.00000440, 0.00000192, 0.00000267, -0.00000095, -0.00000375, -0.00000535, 0.00000667, 0.00001159, -0.00000253, -0.00000444, 0.00000122, -0.00000059, },
		{ 0.00000131, 0.00000021, 0.00000132, 0.00000100, -0.00000253, -0.00000351, 0.00000005, -0.00000751, -0.00000789, 0.00000556, -0.00000359, -0.00000936, -0.00000065, -0.00000883, -0.00000615, 0.00000348, -0.00001082, -0.00003505, -0.00002647, -0.00000856, -0.00010527, -0.00014761, 0.00006787, 0.00019961, 0.00011971, 0.00006958, 0.00005627, 0.00016466, 0.00021468, -0.00006370, -0.00017990, -0.00002886, -0.00003080, -0.00003335, 0.00000085, -0.00003310, -0.00000665, -0.00001570, -0.00001876, 0.00002669, -0.00002317, -0.00001750, 0.00002477, -0.00001532, -0.00000283, 0.00000501, -0.00001668, -0.00000164, 0.00000219, 0.00001047, -0.00001425, -0.00001481, 0.00002704, -0.00002390, -0.00001205, 0.00002914, -0.00003431, 0.00001478, 0.00002630, -0.00005306, 0.00002922, 0.00002048, -0.00005175, 0.00003323, 0.00000252, -0.00004268, 0.00003755, 0.00000162, -0.00002662, 0.00002322, -0.00000742, -0.00001749, 0.00001433, 0.00000144, -0.00000260, 0.00000971, -0.00000708, -0.00002497, -0.00000260, 0.00001595, 0.00000257, 0.00000025, -0.00000385, -0.00001019, 0.00000119, -0.00000490, -0.00000601, 0.00001049, -0.00000318, -0.00001779, 0.00000334, 0.00001714, -0.00000851, -0.00002226, 0.00000946, 0.00001617, -0.00001142, -0.00000555, 0.00001303, 0.00000389, -0.00001236, -0.00000980, 0.00000155, 0.00000163, 0.00000630, 0.00000479, -0.00001404, -0.00000269, 0.00001444, -0.00000342, -0.00000687, 0.00000436, 0.00000000, -0.00000238, -0.00000400, -0.00000378, 0.00000373, 0.00000274, -0.00000244, -0.00000568, -0.00000272, 0.00001042, 0.00000845, -0.00000490, -0.00000290, 0.00000080, -0.00000015, 0.00000213, },
		{ 0.00000084, 0.00000157, -0.00000077, -0.00000369, -0.00000153, -0.00000185, -0.00000988, -0.00000179, 0.00000484, -0.00000849, -0.00000451, -0.00000270, -0.00001240, 0.00000151, 0.00000041, -0.00002679, -0.00003194, -0.00001124, -0.00004919, -0.00014515, -0.00004938, 0.00016928, 0.00015527, 0.00006892, 0.00006721, 0.00006982, 0.00018903, 0.00015230, -0.00013198, -0.00014039, -0.00000848, -0.00003812, -0.00002051, -0.00000875, -0.00002830, -0.00000490, -0.00002159, 0.00000393, 0.00001205, -0.00003461, 0.00000867, 0.00001159, -0.00002100, 0.00001088, -0.00000760, -0.00001618, 0.00000846, 0.00000237, 0.00000038, -0.00001542, 0.00000300, 0.00000898, -0.00002815, 0.00001273, 0.00000282, -0.00002459, 0.00003740, -0.00001550, -0.00002982, 0.00004860, -0.00002406, -0.00002294, 0.00003897, -0.00003313, -0.00000647, 0.00003373, -0.00002549, 0.00000401, 0.00001322, -0.00002516, 0.00000723, 0.00001027, -0.00000810, 0.00000935, 0.00000245, -0.00002000, -0.00001483, 0.00000824, 0.00001084, 0.00000074, -0.00000085, -0.00000969, -0.00000376, 0.00000371, -0.00001069, 0.00000136, 0.00000999, -0.00001634, -0.00001088, 0.00001838, 0.00000644, -0.00002339, -0.00000702, 0.00002149, -0.00000006, -0.00001493, 0.00000823, 0.00001089, -0.00000609, -0.00001159, -0.00000527, 0.00000136, 0.00000424, 0.00000876, -0.00000389, -0.00001443, 0.00000800, 0.00001013, -0.00000948, -0.00000138, 0.00000435, -0.00000249, -0.00000169, -0.00000509, -0.00000225, 0.00000608, 0.00000140, -0.00000479, -0.00000611, 0.00000117, 0.00001240, 0.00000458, -0.00000626, -0.00000168, 0.00000087, 0.00000080, 0.00000207, 0.00000045, },
		{ 0.00000056, -0.00000247, -0.00000319, -0.00000067, -0.00000493, -0.00000837, 0.00000293, 0.00000036, -0.00000863, -0.00000114, -0.00000895, -0.00000875, 0.00000722, -0.00001391, -0.00003381, -0.00001671, -0.00002329, -0.00010209, -0.00011606, 0.00007476, 0.00018801, 0.00008428, 0.00005963, 0.00006814, 0.00008221, 0.00020397, 0.00007361, -0.00017322, -0.00008995, -0.00000201, -0.00004284, -0.00001174, -0.00001586, -0.00002266, -0.00000756, -0.00001424, 0.00001390, -0.00001044, -0.00002344, 0.00002155, -0.00000879, -0.00000977, 0.00001251, -0.00002187, -0.00000258, 0.00001182, -0.00000568, -0.00000264, -0.00000827, 0.00000535, -0.00000731, -0.00001449, 0.00001564, -0.00001611, 0.00000305, 0.00002658, -0.00003863, 0.00001011, 0.00002933, -0.00004347, 0.00001354, 0.00001226, -0.00003616, 0.00002489, 0.00000281, -0.00001891, 0.00002348, -0.00001683, -0.00001264, 0.00002319, -0.00000889, 0.00000026, 0.00001418, -0.00001522, -0.00001813, -0.00000158, 0.00000888, 0.00000717, 0.00000057, -0.00000645, -0.00000924, 0.00000485, -0.00000184, -0.00001132, 0.00001042, -0.00000119, -0.00002245, 0.00000752, 0.00002037, -0.00001384, -0.00002124, 0.00001383, 0.00001523, -0.00001534, -0.00000395, 0.00001543, 0.00000051, -0.00000949, -0.00000801, -0.00000425, 0.00000308, 0.00000831, 0.00000511, -0.00001145, -0.00000724, 0.00001378, 0.00000127, -0.00000955, 0.00000370, 0.00000125, -0.00000278, -0.00000153, -0.00000659, 0.00000130, 0.00000738, -0.00000172, -0.00000654, -0.00000487, 0.00000531, 0.00001266, 0.00000030, -0.00000682, -0.00000011, 0.00000136, 0.00000111, 0.00000144, 0.00000051, 0.00000146, },
		{ -0.00000317, -0.00000222, -0.00000133, -0.00000678, -0.00000470, 0.00000402, -0.00000355, -0.00000562, -0.00000308, -0.00001265, 0.00000037, 0.00000191, -0.00002943, -0.00002452, -0.00001155, -0.00006332, -0.00012164, -0.00002534, 0.00015813, 0.00013703, 0.00004086, 0.00007393, 0.00005588, 0.00010632, 0.00020138, -0.00001735, -0.00017899, -0.00003892, -0.00001091, -0.00004223, -0.00000512, -0.00002216, -0.00001775, -0.00000652, -0.00000404, 0.00000803, -0.00002237, -0.00000304, 0.00001504, -0.00001861, 0.00000617, -0.00000108, -0.00002294, 0.00001393, 0.00000188, -0.00000960, 0.00000216, -0.00000624, -0.00000061, -0.00001080, -0.00000172, 0.00000435, -0.00001422, 0.00002180, -0.00000180, -0.00002960, 0.00003363, -0.00000640, -0.00002961, 0.00002739, -0.00001820, -0.00001027, 0.00002505, -0.00002133, 0.00001052, 0.00000985, -0.00003367, 0.00001908, 0.00001168, -0.00001900, 0.00001910, -0.00000045, -0.00002536, -0.00000348, 0.00000220, 0.00000635, 0.00000752, -0.00000386, -0.00001079, -0.00000115, 0.00000710, -0.00001012, -0.00000349, 0.00001148, -0.00001686, -0.00001259, 0.00002376, 0.00000442, -0.00002642, -0.00000235, 0.00002274, -0.00000319, -0.00001629, 0.00001107, 0.00001001, -0.00000801, -0.00000645, -0.00000741, -0.00000322, 0.00000795, 0.00000863, -0.00000228, -0.00001263, 0.00000216, 0.00001202, -0.00000608, -0.00000489, 0.00000507, -0.00000196, -0.00000127, -0.00000336, -0.00000637, 0.00000612, 0.00000592, -0.00000532, -0.00000685, -0.00000256, 0.00000923, 0.00001106, -0.00000423, -0.00000595, 0.00000196, 0.00000124, 0.00000075, 0.00000138, 0.00000102, 0.00000111, -0.00000095, },
		{ -0.00000160, -0.00000241, -0.00000685, -0.00000146, 0.00000284, -0.00000472, -0.00000406, -0.00000796, -0.00000920, 0.00000574, -0.00001351, -0.00003236, -0.00001018, -0.00003313, -0.00010183, -0.00008204, 0.00007762, 0.00016755, 0.00006823, 0.00004149, 0.00007897, 0.00004150, 0.00014589, 0.00016524, -0.00010278, -0.00014714, -0.00000354, -0.00002934, -0.00003261, -0.00000375, -0.00002760, -0.00000997, -0.00000290, -0.00000055, -0.00000309, -0.00001896, 0.00000978, -0.00000005, -0.00001286, 0.00001139, -0.00001742, -0.00000793, 0.00001809, -0.00001272, -0.00000205, 0.00000413, -0.00001118, -0.00000301, -0.00000686, 0.00000083, -0.00000437, 0.00000010, 0.00001926, -0.00002189, -0.00000294, 0.00002755, -0.00002905, -0.00000234, 0.00001368, -0.00002572, 0.00001578, -0.00000019, -0.00001466, 0.00002912, -0.00002454, -0.00001581, 0.00003793, -0.00001720, -0.00000409, 0.00002526, -0.00002569, -0.00001465, 0.00000775, -0.00000353, 0.00000944, 0.00000564, -0.00001184, -0.00000696, 0.00000731, 0.00000063, -0.00001215, 0.00000634, 0.00000045, -0.00002335, 0.00000885, 0.00002284, -0.00001804, -0.00001920, 0.00001781, 0.00001276, -0.00001731, -0.00000307, 0.00001676, -0.00000206, -0.00000801, -0.00000377, -0.00000959, 0.00000199, 0.00001170, 0.00000350, -0.00000746, -0.00000798, 0.00000775, 0.00000584, -0.00000833, 0.00000024, 0.00000304, -0.00000288, -0.00000033, -0.00000624, -0.00000293, 0.00000969, 0.00000174, -0.00000768, -0.00000603, 0.00000053, 0.00001251, 0.00000714, -0.00000795, -0.00000302, 0.00000335, 0.00000001, 0.00000090, 0.00000197, 0.00000084, 0.00000003, -0.00000200, -0.00000321, },
		{ -0.00000330, -0.00000617, 0.00000072, 0.00000166, -0.00000490, -0.00000582, -0.00001000, -0.00000194, 0.00000027, -0.00002722, -0.00002070, -0.00001166, -0.00007286, -0.00010248, 0.00000079, 0.00013955, 0.00012031, 0.00002824, 0.00006373, 0.00006093, 0.00004762, 0.00018323, 0.00009131, -0.00015720, -0.00009176, 0.00000532, -0.00004420, -0.00001726, -0.00001050, -0.00002716, 0.00000033, -0.00000300, -0.00000308, -0.00000874, -0.00000891, 0.00000973, -0.00000955, -0.00000128, 0.00000268, -0.00002153, 0.00001054, 0.00000586, -0.00001729, 0.00001025, -0.00000404, -0.00001427, 0.00000054, -0.00000534, -0.00000188, -0.00000292, 0.00001054, 0.00000378, -0.00002135, 0.00001715, 0.00000425, -0.00002703, 0.00001332, -0.00000801, -0.00000910, 0.00001764, -0.00002152, 0.00001441, 0.00001416, -0.00004477, 0.00002431, 0.00002255, -0.00003345, 0.00002475, 0.00000439, -0.00003728, 0.00000955, 0.00000224, -0.00000623, 0.00001658, -0.00000449, -0.00001475, 0.00000378, 0.00000755, -0.00000745, -0.00000559, 0.00000769, -0.00001455, -0.00001241, 0.00002529, 0.00000318, -0.00002804, 0.00000280, 0.00002244, -0.00000634, -0.00001566, 0.00001173, 0.00000933, -0.00000959, -0.00000243, -0.00000647, -0.00000896, 0.00001014, 0.00000946, -0.00000285, -0.00000712, -0.00000264, 0.00000805, 0.00000031, -0.00000657, 0.00000279, 0.00000039, -0.00000192, -0.00000158, -0.00000756, 0.00000275, 0.00000976, -0.00000323, -0.00000820, -0.00000469, 0.00000457, 0.00001398, 0.00000117, -0.00000903, 0.00000109, 0.00000261, -0.00000120, 0.00000227, 0.00000200, -0.00000025, -0.00000079, -0.00000265, -0.00000308, -0.00000104, },
		{ -0.00000512, 0.00000278, 0.00000054, -0.00000649, -0.00000807, -0.00000697, 0.00000093, -0.00001274, -0.00002801, -0.00000880, -0.00003827, -0.00010019, -0.00005310, 0.00008202, 0.00014033, 0.00006172, 0.00002824, 0.00007546, 0.00003425, 0.00008261, 0.00019125, -0.00000236, -0.00016547, -0.00003801, -0.00000828, -0.00004527, -0.00000575, -0.00002125, -0.00001714, 0.00000650, -0.00000846, -0.00000450, -0.00000788, -0.00000217, 0.00000352, -0.00000925, 0.00000406, -0.00000963, -0.00001076, 0.00001701, -0.00001137, -0.00000674, 0.00001389, -0.00001725, -0.00000865, 0.00000306, -0.00000815, -0.00000129, 0.00000328, 0.00000976, -0.00000863, -0.00000747, 0.00001845, -0.00001501, -0.00001092, 0.00000937, -0.00001608, 0.00001065, -0.00000264, -0.00001615, 0.00003477, -0.00002359, -0.00002718, 0.00005119, -0.00001575, -0.00001730, 0.00003810, -0.00003093, -0.00001957, 0.00002394, -0.00001490, 0.00000433, 0.00001628, -0.00001832, -0.00000571, 0.00001135, -0.00000032, -0.00000830, 0.00000164, -0.00000212, -0.00001918, 0.00000889, 0.00002216, -0.00001989, -0.00001630, 0.00002124, 0.00000872, -0.00001764, -0.00000166, 0.00001559, -0.00000303, -0.00000741, 0.00000040, -0.00001192, -0.00000161, 0.00001480, 0.00000189, -0.00000462, -0.00000392, -0.00000062, 0.00000622, -0.00000236, -0.00000403, 0.00000305, -0.00000097, -0.00000124, -0.00000406, -0.00000560, 0.00000776, 0.00000639, -0.00000690, -0.00000762, -0.00000268, 0.00000920, 0.00001224, -0.00000495, -0.00000650, 0.00000399, -0.00000003, -0.00000060, 0.00000388, 0.00000054, -0.00000120, -0.00000115, -0.00000341, -0.00000242, -0.00000034, -0.00000451, },
		{ 0.00000464, -0.00000218, -0.00000890, -0.00000726, -0.00000306, -0.00000461, -0.00002225, -0.00001769, -0.00001470, -0.00007528, -0.00008714, 0.00002335, 0.00012229, 0.00009991, 0.00002783, 0.00004775, 0.00006367, 0.00002385, 0.00013068, 0.00015545, -0.00008479, -0.00013402, -0.00000625, -0.00002865, -0.00003363, -0.00000488, -0.00002655, -0.00000193, 0.00000274, -0.00001330, -0.00000138, -0.00000636, -0.00000059, -0.00000079, -0.00000515, 0.00000074, -0.00001378, 0.00000392, 0.00000788, -0.00001829, 0.00000928, 0.00000320, -0.00002366, 0.00000206, -0.00000112, -0.00000926, 0.00000382, 0.00000569, 0.00000295, -0.00001023, 0.00000478, 0.00000651, -0.00001924, 0.00000180, -0.00000319, -0.00000660, 0.00001344, -0.00002195, 0.00001086, 0.00002357, -0.00005008, 0.00001751, 0.00003963, -0.00004281, 0.00001967, 0.00001726, -0.00004984, 0.00001624, 0.00001205, -0.00002368, 0.00002222, 0.00000085, -0.00002243, 0.00001026, 0.00000733, -0.00000666, -0.00000210, -0.00000023, -0.00001316, -0.00000731, 0.00002263, 0.00000126, -0.00002684, 0.00000757, 0.00002102, -0.00001002, -0.00001279, 0.00001096, 0.00000737, -0.00000920, -0.00000054, -0.00000381, -0.00001280, 0.00000913, 0.00001113, -0.00000447, -0.00000092, -0.00000323, -0.00000092, 0.00000558, -0.00000366, -0.00000227, 0.00000286, -0.00000158, -0.00000190, -0.00000541, -0.00000130, 0.00000980, 0.00000158, -0.00000869, -0.00000666, 0.00000067, 0.00001278, 0.00000719, -0.00000842, -0.00000179, 0.00000372, -0.00000235, 0.00000207, 0.00000380, -0.00000166, -0.00000113, -0.00000187, -0.00000394, -0.00000067, -0.00000049, -0.00000621, -0.00000276, },
		{ -0.00000658, -0.00000910, -0.00000407, -0.00000405, -0.00001370, -0.00002097, -0.00001077, -0.00004174, -0.00009333, -0.00003217, 0.00008676, 0.00011370, 0.00005514, 0.00002488, 0.00006050, 0.00003984, 0.00004169, 0.00016432, 0.00008500, -0.00013177, -0.00008541, -0.00000068, -0.00004099, -0.00001914, -0.00001198, -0.00002062, 0.00000816, -0.00000831, -0.00001088, 0.00000228, -0.00000771, 0.00000011, -0.00000199, -0.00000373, -0.00000438, -0.00000809, 0.00000999, -0.00000657, -0.00001019, 0.00001626, -0.00001404, -0.00001735, 0.00000817, -0.00000894, -0.00000405, 0.00000755, 0.00000273, -0.00000141, -0.00000511, 0.00000733, -0.00000585, -0.00001253, 0.00000329, -0.00000951, 0.00000718, -0.00000150, -0.00002045, 0.00003345, -0.00001214, -0.00003997, 0.00005380, -0.00000173, -0.00003383, 0.00004396, -0.00002510, -0.00003199, 0.00003972, -0.00001772, -0.00000978, 0.00002898, -0.00002148, -0.00000855, 0.00001865, -0.00000445, -0.00000318, 0.00000094, -0.00001060, -0.00001201, 0.00001106, 0.00001644, -0.00001942, -0.00001183, 0.00002332, 0.00000391, -0.00001732, 0.00000123, 0.00001190, -0.00000332, -0.00000600, 0.00000227, -0.00001077, -0.00000545, 0.00001515, 0.00000175, -0.00000361, 0.00000289, -0.00000689, 0.00000057, 0.00000575, -0.00000551, -0.00000030, 0.00000265, -0.00000277, -0.00000274, -0.00000455, 0.00000284, 0.00000887, -0.00000274, -0.00000931, -0.00000503, 0.00000516, 0.00001340, 0.00000094, -0.00000787, 0.00000201, 0.00000085, -0.00000202, 0.00000484, 0.00000137, -0.00000284, -0.00000068, -0.00000332, -0.00000297, 0.00000129, -0.00000257, -0.00000686, 0.00000127, 0.00000436, },
		{ -0.00000598, -0.00000315, -0.00001032, -0.00001768, -0.00001367, -0.00002068, -0.00007239, -0.00007211, 0.00003776, 0.00010959, 0.00007829, 0.00003069, 0.00003638, 0.00005613, 0.00002424, 0.00007879, 0.00016441, 0.00000492, -0.00013840, -0.00004266, -0.00001142, -0.00004077, -0.00001059, -0.00001800, -0.00000741, 0.00000627, -0.00001702, -0.00000185, 0.00000076, -0.00000922, 0.00000240, -0.00000375, -0.00000475, -0.00000486, -0.00000040, 0.00000495, -0.00001345, 0.00000409, 0.00000849, -0.00002469, -0.00000393, 0.00000458, -0.00001237, 0.00000444, 0.00000542, -0.00000040, -0.00000094, -0.00000111, 0.00000264, -0.00001103, -0.00000480, -0.00000188, -0.00000521, 0.00001093, -0.00001831, 0.00000037, 0.00003047, -0.00004298, 0.00000068, 0.00005351, -0.00003886, 0.00000337, 0.00003155, -0.00005420, 0.00001150, 0.00002888, -0.00003809, 0.00001869, 0.00001267, -0.00003115, 0.00001478, 0.00001004, -0.00001029, 0.00000594, -0.00000771, -0.00001669, 0.00000319, 0.00001814, -0.00000377, -0.00002176, 0.00001164, 0.00001845, -0.00001374, -0.00000877, 0.00001031, 0.00000323, -0.00000683, 0.00000023, -0.00000199, -0.00001261, 0.00000551, 0.00001168, -0.00000499, 0.00000343, 0.00000055, -0.00001081, 0.00000568, 0.00000372, -0.00000714, 0.00000279, 0.00000107, -0.00000419, -0.00000219, -0.00000270, 0.00000532, 0.00000650, -0.00000615, -0.00000919, -0.00000203, 0.00000914, 0.00001075, -0.00000360, -0.00000474, 0.00000273, -0.00000181, 0.00000092, 0.00000525, -0.00000183, -0.00000235, -0.00000124, -0.00000425, -0.00000019, 0.00000131, -0.00000567, -0.00000457, 0.00000493, 0.00000058, -0.00000983, },
		{ -0.00000752, -0.00001588, -0.00001390, -0.00001309, -0.00004553, -0.00008120, -0.00001720, 0.00008771, 0.00009329, 0.00004543, 0.00002744, 0.00004419, 0.00004215, 0.00002784, 0.00011497, 0.00013041, -0.00006056, -0.00011647, -0.00001751, -0.00002520, -0.00003263, -0.00000896, -0.00001662, 0.00000307, -0.00000451, -0.00001585, 0.00000617, -0.00000590, -0.00000662, 0.00000355, -0.00000742, -0.00000372, -0.00000160, 0.00000172, -0.00000350, -0.00000902, 0.00001189, -0.00000707, -0.00002235, 0.00000582, -0.00000428, -0.00000753, 0.00000896, -0.00000015, 0.00000066, 0.00000036, -0.00000136, -0.00000283, -0.00000994, -0.00000150, -0.00000572, 0.00000310, 0.00000244, -0.00002199, 0.00002190, 0.00000431, -0.00004444, 0.00004133, 0.00001945, -0.00004347, 0.00003614, -0.00000914, -0.00004437, 0.00004608, -0.00000854, -0.00002850, 0.00003683, -0.00001714, -0.00001621, 0.00002753, -0.00000816, -0.00000147, 0.00000655, -0.00002252, -0.00000654, 0.00001825, 0.00000612, -0.00001855, -0.00000443, 0.00002308, -0.00000081, -0.00001682, 0.00000515, 0.00000726, -0.00000454, -0.00000270, 0.00000164, -0.00000802, -0.00000661, 0.00001157, 0.00000254, -0.00000292, 0.00000900, -0.00000825, -0.00000911, 0.00001119, -0.00000180, -0.00000588, 0.00000549, -0.00000255, -0.00000385, -0.00000061, -0.00000159, 0.00000668, 0.00000364, -0.00000903, -0.00000779, 0.00000194, 0.00001091, 0.00000667, -0.00000516, -0.00000195, 0.00000110, -0.00000193, 0.00000404, 0.00000290, -0.00000366, -0.00000148, -0.00000283, -0.00000288, 0.00000240, -0.00000158, -0.00000689, 0.00000015, 0.00000513, -0.00000483, -0.00000909, -0.00000276, },
		{ -0.00001503, -0.00000946, -0.00002695, -0.00006752, -0.00005529, 0.00004405, 0.00009927, 0.00006075, 0.00003022, 0.00003210, 0.00004263, 0.00003130, 0.00004711, 0.00013393, 0.00007578, -0.00009925, -0.00008319, -0.00000964, -0.00003398, -0.00002297, -0.00000955, -0.00000948, 0.00000408, -0.00001435, -0.00000618, 0.00000637, -0.00001175, -0.00000045, 0.00000015, -0.00000938, 0.00000081, -0.00000025, -0.00000155, -0.00000693, 0.00000002, 0.00000820, -0.00001880, -0.00001156, 0.00000626, -0.00000979, 0.00000132, 0.00000615, -0.00000288, 0.00000429, -0.00000140, -0.00000348, -0.00000567, -0.00000674, -0.00000188, -0.00000509, 0.00000743, -0.00000974, -0.00001140, 0.00002751, -0.00002391, -0.00001696, 0.00005441, -0.00002011, -0.00001669, 0.00003735, -0.00004476, -0.00000247, 0.00004409, -0.00004131, 0.00000475, 0.00002652, -0.00003554, 0.00001357, 0.00001722, -0.00001704, 0.00001382, -0.00000962, -0.00002618, 0.00001579, 0.00001594, -0.00001354, -0.00001344, 0.00001585, 0.00001391, -0.00001632, -0.00000502, 0.00001053, -0.00000231, -0.00000415, 0.00000166, -0.00000256, -0.00000886, 0.00000194, 0.00000888, -0.00000334, 0.00000571, 0.00000598, -0.00001696, -0.00000043, 0.00001187, -0.00000791, -0.00000092, 0.00000488, -0.00000615, -0.00000073, -0.00000024, -0.00000118, 0.00000796, -0.00000006, -0.00001099, -0.00000466, 0.00000528, 0.00001049, 0.00000339, -0.00000508, -0.00000106, -0.00000030, 0.00000018, 0.00000495, -0.00000033, -0.00000373, -0.00000175, -0.00000349, 0.00000048, 0.00000201, -0.00000524, -0.00000423, 0.00000382, 0.00000129, -0.00000781, -0.00000540, -0.00000398, -0.00001430, },
		{ -0.00001442, -0.00004888, -0.00006689, -0.00000468, 0.00008291, 0.00007939, 0.00003541, 0.00002913, 0.00003363, 0.00003705, 0.00002988, 0.00007116, 0.00013040, 0.00001699, -0.00011128, -0.00005174, -0.00001285, -0.00003592, -0.00001470, -0.00000850, -0.00000312, -0.00000258, -0.00001573, 0.00000376, -0.00000070, -0.00001142, 0.00000394, -0.00000637, -0.00000571, 0.00000464, -0.00000350, -0.00000436, -0.00000417, 0.00000499, -0.00000206, -0.00002074, -0.00000157, 0.00000060, -0.00000782, 0.00000665, -0.00000015, 0.00000021, 0.00000537, -0.00000556, -0.00000450, -0.00000629, -0.00000385, -0.00000341, -0.00000186, 0.00000574, -0.00001699, 0.00000340, 0.00001610, -0.00003460, 0.00001813, 0.00003654, -0.00003794, 0.00001561, 0.00000865, -0.00004710, 0.00003842, 0.00000868, -0.00004285, 0.00003372, -0.00000504, -0.00002528, 0.00003344, -0.00000716, -0.00000549, 0.00001632, -0.00003216, -0.00000712, 0.00003023, -0.00000537, -0.00002016, 0.00000661, 0.00002049, -0.00000549, -0.00001548, 0.00000844, 0.00000359, -0.00000731, 0.00000162, 0.00000027, -0.00000622, -0.00000376, 0.00000550, 0.00000205, -0.00000055, 0.00001224, -0.00000528, -0.00001803, 0.00001015, 0.00000595, -0.00000995, 0.00000477, 0.00000021, -0.00000633, 0.00000317, -0.00000248, 0.00000037, 0.00000873, -0.00000490, -0.00001083, -0.00000075, 0.00000684, 0.00000957, 0.00000124, -0.00000540, -0.00000097, 0.00000007, 0.00000220, 0.00000363, -0.00000254, -0.00000352, -0.00000264, -0.00000152, 0.00000287, -0.00000164, -0.00000591, 0.00000048, 0.00000325, -0.00000319, -0.00000646, -0.00000347, -0.00001006, -0.00001719, -0.00000938, },
		{ -0.00006225, -0.00003796, 0.00004577, 0.00008802, 0.00004972, 0.00002599, 0.00003084, 0.00003099, 0.00003389, 0.00003607, 0.00009083, 0.00010769, -0.00003395, -0.00010325, -0.00002889, -0.00002066, -0.00003197, -0.00000794, -0.00000632, -0.00000188, -0.00000963, -0.00000917, 0.00000693, -0.00000854, -0.00000573, 0.00000240, -0.00001013, 0.00000207, 0.00000280, -0.00000782, -0.00000283, -0.00000037, 0.00000319, -0.00001081, -0.00001547, 0.00000235, -0.00000414, -0.00000165, 0.00000520, -0.00000328, 0.00000581, 0.00000127, -0.00000851, -0.00000416, -0.00000572, -0.00000191, -0.00000506, 0.00000182, 0.00000050, -0.00001756, 0.00001299, -0.00000104, -0.00002458, 0.00003907, 0.00000589, -0.00002895, 0.00002830, -0.00002393, -0.00001674, 0.00004825, -0.00003054, -0.00001422, 0.00003476, -0.00003105, 0.00000604, 0.00002647, -0.00002226, 0.00001625, -0.00000403, -0.00003812, 0.00002481, 0.00001910, -0.00002690, -0.00000465, 0.00002179, 0.00000663, -0.00001694, -0.00000201, 0.00001077, -0.00000699, -0.00000308, 0.00000438, -0.00000479, -0.00000412, 0.00000117, 0.00000239, -0.00000073, 0.00000772, 0.00000964, -0.00001720, -0.00000972, 0.00001565, -0.00000283, -0.00000625, 0.00000677, -0.00000540, -0.00000192, 0.00000413, -0.00000574, 0.00000428, 0.00000724, -0.00000965, -0.00000808, 0.00000204, 0.00000746, 0.00000934, -0.00000145, -0.00000621, 0.00000025, 0.00000116, 0.00000265, 0.00000180, -0.00000386, -0.00000399, -0.00000192, 0.00000184, 0.00000147, -0.00000510, -0.00000235, 0.00000280, -0.00000063, -0.00000418, -0.00000376, -0.00000674, -0.00001596, -0.00001348, -0.00000731, -0.00003094, },
		{ 0.00000676, 0.00007434, 0.00006859, 0.00002870, 0.00002674, 0.00002881, 0.00002912, 0.00003426, 0.00004551, 0.00010192, 0.00007204, -0.00007052, -0.00008329, -0.00001669, -0.00002792, -0.00002362, -0.00000313, -0.00000554, -0.00000456, -0.00001156, -0.00000098, 0.00000293, -0.00001146, -0.00000021, -0.00000328, -0.00000699, 0.00000752, -0.00000430, -0.00000782, 0.00000091, -0.00000013, -0.00000172, -0.00001419, -0.00000901, 0.00000164, -0.00000434, 0.00000262, 0.00000028, -0.00000042, 0.00000807, -0.00000563, -0.00000818, -0.00000376, -0.00000405, -0.00000178, -0.00000608, 0.00000538, -0.00000643, -0.00001402, 0.00001572, -0.00001299, -0.00000453, 0.00003938, -0.00001715, -0.00000831, 0.00001745, -0.00003516, 0.00002038, 0.00002420, -0.00004459, 0.00001899, 0.00000964, -0.00002971, 0.00003220, -0.00000000, -0.00001312, 0.00002455, -0.00003401, -0.00001433, 0.00004279, -0.00001308, -0.00002599, 0.00002003, 0.00001705, -0.00001113, -0.00001204, 0.00000956, 0.00000153, -0.00001022, 0.00000460, 0.00000015, -0.00000635, 0.00000163, 0.00000015, -0.00000162, 0.00000386, 0.00001306, -0.00000123, -0.00002176, 0.00000267, 0.00001320, -0.00000868, 0.00000010, 0.00000357, -0.00000733, 0.00000413, 0.00000037, -0.00000653, 0.00000910, 0.00000256, -0.00001211, -0.00000419, 0.00000303, 0.00000908, 0.00000861, -0.00000572, -0.00000546, 0.00000258, 0.00000105, 0.00000234, 0.00000034, -0.00000556, -0.00000374, 0.00000129, 0.00000295, -0.00000274, -0.00000441, 0.00000208, 0.00000037, -0.00000319, -0.00000164, -0.00000470, -0.00001308, -0.00001629, -0.00000789, -0.00001569, -0.00004948, -0.00005358, },
		{ 0.00007525, 0.00004318, 0.00002138, 0.00002802, 0.00002480, 0.00003100, 0.00003533, 0.00005604, 0.00010288, 0.00002967, -0.00009001, -0.00005914, -0.00001402, -0.00003077, -0.00001328, -0.00000168, -0.00000682, -0.00000709, -0.00000843, 0.00000329, -0.00000349, -0.00000921, 0.00000093, -0.00000723, 0.00000081, 0.00000559, -0.00001070, -0.00000260, 0.00000190, -0.00000290, -0.00000535, -0.00001377, -0.00000467, 0.00000039, -0.00000183, 0.00000226, -0.00000241, 0.00000527, 0.00000426, -0.00001072, -0.00000564, -0.00000361, -0.00000178, -0.00000426, -0.00000445, 0.00000748, -0.00001488, -0.00000748, 0.00001447, -0.00001698, 0.00001382, 0.00002643, -0.00002527, 0.00000676, -0.00000162, -0.00002114, 0.00003761, -0.00001102, -0.00002856, 0.00003132, -0.00001783, -0.00000381, 0.00003254, -0.00002122, 0.00001120, 0.00000604, -0.00004599, 0.00002569, 0.00002708, -0.00003963, 0.00000096, 0.00003030, -0.00000293, -0.00001611, 0.00000122, 0.00000912, -0.00000881, -0.00000393, 0.00000685, -0.00000624, -0.00000101, 0.00000375, -0.00000539, 0.00000025, 0.00001118, 0.00001004, -0.00001308, -0.00001697, 0.00001195, 0.00000583, -0.00000892, 0.00000425, -0.00000184, -0.00000385, 0.00000696, -0.00000585, -0.00000277, 0.00001168, -0.00000384, -0.00001132, -0.00000143, 0.00000385, 0.00001205, 0.00000494, -0.00000980, -0.00000169, 0.00000368, -0.00000028, 0.00000275, -0.00000191, -0.00000730, -0.00000062, 0.00000437, -0.00000020, -0.00000524, 0.00000031, 0.00000245, -0.00000396, -0.00000111, -0.00000031, -0.00001060, -0.00001655, -0.00001146, -0.00000849, -0.00003220, -0.00005577, -0.00002314, 0.00004636, },
		{ 0.00002559, 0.00002200, 0.00002566, 0.00002365, 0.00003444, 0.00003542, 0.00006741, 0.00009242, -0.00001304, -0.00009266, -0.00003746, -0.00001722, -0.00002749, -0.00000462, -0.00000395, -0.00000818, -0.00000710, -0.00000393, 0.00000267, -0.00000785, -0.00000547, -0.00000140, -0.00000562, 0.00000665, -0.00000228, -0.00001079, 0.00000314, -0.00000170, -0.00000494, -0.00000691, -0.00001238, -0.00000182, 0.00000047, -0.00000028, -0.00000049, -0.00000036, 0.00000838, -0.00000333, -0.00001140, -0.00000316, -0.00000290, -0.00000091, -0.00000784, 0.00000122, 0.00000454, -0.00002276, 0.00000297, 0.00001024, -0.00001532, 0.00002587, 0.00000967, -0.00002314, 0.00001129, -0.00001242, 0.00000224, 0.00002851, -0.00003188, -0.00000099, 0.00001883, -0.00002494, 0.00002351, 0.00001022, -0.00001892, 0.00002569, -0.00002660, -0.00002397, 0.00004973, -0.00001369, -0.00003473, 0.00003235, 0.00001507, -0.00001863, -0.00000604, 0.00000853, -0.00000024, -0.00001073, 0.00000431, 0.00000152, -0.00000692, 0.00000637, -0.00000185, -0.00000800, 0.00000830, 0.00001368, 0.00000073, -0.00001914, -0.00000716, 0.00001478, -0.00000130, -0.00000553, 0.00000414, -0.00000482, 0.00000221, 0.00000412, -0.00000995, 0.00000406, 0.00000997, -0.00000915, -0.00000865, -0.00000035, 0.00000678, 0.00001380, -0.00000216, -0.00001025, 0.00000336, 0.00000174, -0.00000062, 0.00000326, -0.00000595, -0.00000636, 0.00000472, 0.00000336, -0.00000492, -0.00000255, 0.00000415, -0.00000217, -0.00000464, 0.00000329, -0.00000468, -0.00001683, -0.00001388, -0.00000800, -0.00001846, -0.00004611, -0.00004207, 0.00001571, 0.00006582, 0.00005803, },
		{ 0.00002297, 0.00002209, 0.00002698, 0.00003527, 0.00003637, 0.00007823, 0.00006988, -0.00004930, -0.00008179, -0.00002247, -0.00002139, -0.00001958, -0.00000082, -0.00000778, -0.00000756, -0.00000536, -0.00000104, -0.00000051, -0.00000914, -0.00000315, -0.00000323, -0.00000024, 0.00000588, -0.00000958, -0.00000477, 0.00000397, -0.00000672, -0.00000423, -0.00000837, -0.00001060, 0.00000098, 0.00000068, -0.00000101, -0.00000179, 0.00000422, 0.00000600, -0.00001009, -0.00000842, -0.00000175, -0.00000160, -0.00000304, -0.00000860, 0.00000824, -0.00000596, -0.00002436, 0.00001555, 0.00000156, -0.00000883, 0.00003151, -0.00000677, -0.00001658, 0.00001079, -0.00001133, 0.00001725, 0.00000660, -0.00003024, 0.00001632, -0.00000172, -0.00000948, 0.00003110, -0.00001284, 0.00000161, 0.00001425, -0.00004435, 0.00001829, 0.00003524, -0.00004643, 0.00000126, 0.00004003, -0.00001256, -0.00001556, 0.00000613, 0.00000434, -0.00000769, -0.00000479, 0.00000643, -0.00000496, -0.00000001, 0.00000745, -0.00001081, -0.00000206, 0.00001576, 0.00000853, -0.00000834, -0.00001827, 0.00000212, 0.00001273, -0.00000551, -0.00000211, 0.00000178, -0.00000348, 0.00000607, -0.00000241, -0.00000893, 0.00001028, 0.00000478, -0.00001177, -0.00000623, 0.00000096, 0.00001169, 0.00001102, -0.00000946, -0.00000564, 0.00000582, -0.00000183, 0.00000152, 0.00000130, -0.00000983, -0.00000071, 0.00000788, -0.00000228, -0.00000580, 0.00000330, 0.00000237, -0.00000697, 0.00000031, 0.00000374, -0.00001341, -0.00001741, -0.00000868, -0.00001208, -0.00003223, -0.00004669, -0.00001265, 0.00004648, 0.00006345, 0.00003737, 0.00001885, },
		{ 0.00002146, 0.00003162, 0.00003200, 0.00004146, 0.00008446, 0.00003723, -0.00007334, -0.00006311, -0.00001497, -0.00002274, -0.00001100, -0.00000218, -0.00000991, -0.00000514, -0.00000379, -0.00000026, -0.00000397, -0.00000859, -0.00000206, -0.00000222, 0.00000392, -0.00000016, -0.00001095, 0.00000182, -0.00000091, -0.00000861, -0.00000253, -0.00001074, -0.00000725, 0.00000334, -0.00000051, -0.00000236, -0.00000003, 0.00000723, -0.00000057, -0.00001272, -0.00000420, -0.00000120, -0.00000114, -0.00000691, -0.00000393, 0.00001027, -0.00002059, -0.00001475, 0.00002346, -0.00001022, 0.00000377, 0.00002923, -0.00002127, -0.00000628, 0.00001017, -0.00000562, 0.00001874, -0.00001114, -0.00001613, 0.00001672, -0.00001162, 0.00001197, 0.00001773, -0.00001807, 0.00001828, -0.00001379, -0.00002919, 0.00004674, -0.00000801, -0.00004203, 0.00003948, 0.00001593, -0.00002745, 0.00000127, 0.00000704, -0.00000387, -0.00000719, 0.00000108, 0.00000230, -0.00000522, 0.00000797, -0.00000058, -0.00001402, 0.00000983, 0.00001566, -0.00000011, -0.00001278, -0.00001383, 0.00000858, 0.00000868, -0.00000749, -0.00000002, 0.00000036, 0.00000007, 0.00000523, -0.00000827, -0.00000345, 0.00001305, -0.00000149, -0.00001229, -0.00000449, 0.00000465, 0.00001529, 0.00000340, -0.00001230, 0.00000119, 0.00000359, -0.00000336, 0.00000421, -0.00000417, -0.00000922, 0.00000719, 0.00000466, -0.00000771, -0.00000062, 0.00000610, -0.00000407, -0.00000609, 0.00000626, -0.00000320, -0.00001970, -0.00001185, -0.00000828, -0.00002279, -0.00003988, -0.00003120, 0.00002020, 0.00005956, 0.00004796, 0.00002319, 0.00001832, 0.00002139, },
		{ 0.00003297, 0.00002747, 0.00005170, 0.00008107, -0.00000013, -0.00008234, -0.00004283, -0.00001298, -0.00002046, -0.00000545, -0.00000593, -0.00000876, -0.00000283, -0.00000317, -0.00000089, -0.00000685, -0.00000699, -0.00000067, 0.00000030, 0.00000370, -0.00000618, -0.00000666, 0.00000355, -0.00000726, -0.00000610, -0.00000294, -0.00001241, -0.00000221, 0.00000361, -0.00000267, -0.00000221, 0.00000329, 0.00000614, -0.00000765, -0.00001077, -0.00000086, -0.00000135, -0.00000253, -0.00000885, 0.00000381, 0.00000277, -0.00003029, 0.00000375, 0.00001971, -0.00001806, 0.00002035, 0.00001628, -0.00002917, 0.00000807, 0.00000777, -0.00000179, 0.00001298, -0.00001838, -0.00000309, 0.00000897, -0.00000653, 0.00002225, -0.00000088, -0.00000650, 0.00001512, -0.00003250, 0.00000749, 0.00003742, -0.00004422, -0.00000253, 0.00004745, -0.00001932, -0.00001680, 0.00001307, -0.00000278, -0.00000576, -0.00000289, 0.00000160, -0.00000140, 0.00000094, 0.00000882, -0.00001154, -0.00000659, 0.00001884, 0.00000769, -0.00000592, -0.00001357, -0.00000873, 0.00001292, 0.00000360, -0.00000836, 0.00000194, 0.00000064, 0.00000243, 0.00000103, -0.00001042, 0.00000356, 0.00001200, -0.00000721, -0.00001185, -0.00000174, 0.00001012, 0.00001401, -0.00000517, -0.00000906, 0.00000531, -0.00000112, -0.00000108, 0.00000361, -0.00000984, -0.00000211, 0.00001101, -0.00000366, -0.00000684, 0.00000628, 0.00000200, -0.00000868, 0.00000066, 0.00000615, -0.00001364, -0.00001823, -0.00000664, -0.00001597, -0.00003235, -0.00003526, -0.00000575, 0.00004463, 0.00005534, 0.00003056, 0.00001745, 0.00001842, 0.00001925, 0.00002485, },
		{ 0.00002660, 0.00006391, 0.00006512, -0.00003471, -0.00007729, -0.00002569, -0.00001368, -0.00001622, -0.00000400, -0.00000852, -0.00000544, -0.00000210, -0.00000306, -0.00000254, -0.00000866, -0.00000410, 0.00000113, 0.00000157, 0.00000031, -0.00000817, -0.00000141, -0.00000028, -0.00001015, -0.00000206, -0.00000619, -0.00001111, 0.00000251, 0.00000128, -0.00000399, -0.00000020, 0.00000553, 0.00000134, -0.00001182, -0.00000599, 0.00000035, -0.00000227, -0.00000477, -0.00000657, 0.00000846, -0.00001170, -0.00002723, 0.00002097, 0.00000480, -0.00001413, 0.00003250, -0.00000511, -0.00002399, 0.00002146, 0.00000033, 0.00000097, 0.00000655, -0.00001886, 0.00000417, 0.00000392, 0.00000431, 0.00001817, -0.00000990, 0.00000607, -0.00000246, -0.00002493, 0.00003452, -0.00000098, -0.00004286, 0.00003926, 0.00001865, -0.00003522, 0.00000739, 0.00000720, -0.00001043, -0.00000039, -0.00000239, -0.00000032, 0.00000026, 0.00000669, 0.00000113, -0.00001592, 0.00000714, 0.00001810, -0.00000171, -0.00000687, -0.00001406, -0.00000275, 0.00001530, -0.00000292, -0.00000727, 0.00000471, 0.00000083, 0.00000235, -0.00000354, -0.00000872, 0.00000961, 0.00000807, -0.00001210, -0.00001012, 0.00000326, 0.00001394, 0.00000781, -0.00000979, -0.00000283, 0.00000437, -0.00000389, 0.00000251, -0.00000152, -0.00001017, 0.00000753, 0.00000639, -0.00000999, 0.00000086, 0.00000770, -0.00000569, -0.00000619, 0.00000676, -0.00000184, -0.00001958, -0.00001127, -0.00000894, -0.00002689, -0.00003306, -0.00001998, 0.00002026, 0.00005481, 0.00004076, 0.00001870, 0.00001675, 0.00001697, 0.00001971, 0.00002933, 0.00002911, },
		{ 0.00007196, 0.00003799, -0.00005949, -0.00006238, -0.00001426, -0.00001499, -0.00001216, -0.00000501, -0.00000824, -0.00000232, -0.00000280, -0.00000318, -0.00000498, -0.00000840, -0.00000027, 0.00000188, 0.00000069, -0.00000292, -0.00000644, 0.00000048, -0.00000591, -0.00000822, -0.00000027, -0.00000995, -0.00000660, 0.00000443, -0.00000206, -0.00000327, 0.00000233, 0.00000504, -0.00000476, -0.00001152, -0.00000113, -0.00000089, -0.00000338, -0.00000581, -0.00000179, 0.00000555, -0.00002454, -0.00001201, 0.00002640, -0.00001169, 0.00000186, 0.00003089, -0.00002465, -0.00000520, 0.00002494, -0.00000955, 0.00000621, -0.00000018, -0.00001680, 0.00000892, 0.00000393, 0.00001097, 0.00000853, -0.00000825, 0.00000779, -0.00001519, 0.00000039, 0.00002995, -0.00003447, -0.00000617, 0.00004871, -0.00002147, -0.00001967, 0.00002046, -0.00000964, -0.00000577, 0.00000311, -0.00000640, 0.00000164, 0.00000468, 0.00000594, -0.00000845, -0.00000979, 0.00001749, 0.00000883, -0.00000587, -0.00000640, -0.00001496, 0.00000535, 0.00001376, -0.00000967, -0.00000262, 0.00000667, -0.00000059, 0.00000112, -0.00000675, -0.00000446, 0.00001361, 0.00000180, -0.00001562, -0.00000566, 0.00000893, 0.00001323, 0.00000059, -0.00000902, 0.00000151, 0.00000064, -0.00000284, 0.00000308, -0.00000721, -0.00000319, 0.00001196, -0.00000370, -0.00000783, 0.00000866, 0.00000125, -0.00000923, 0.00000132, 0.00000552, -0.00001177, -0.00001721, -0.00000708, -0.00001890, -0.00003217, -0.00002372, -0.00000033, 0.00003974, 0.00005115, 0.00002422, 0.00001433, 0.00001674, 0.00001533, 0.00002358, 0.00003056, 0.00002244, 0.00003255, },
		{ 0.00000540, -0.00007043, -0.00004347, -0.00000903, -0.00001581, -0.00000927, -0.00000615, -0.00000576, -0.00000099, -0.00000400, -0.00000385, -0.00000714, -0.00000544, 0.00000273, 0.00000064, -0.00000080, -0.00000411, -0.00000410, -0.00000154, -0.00000914, -0.00000394, -0.00000198, -0.00001127, -0.00000123, 0.00000296, -0.00000414, -0.00000107, 0.00000374, 0.00000189, -0.00000931, -0.00000742, 0.00000112, -0.00000353, -0.00000336, -0.00000516, 0.00000146, -0.00000349, -0.00002805, 0.00000637, 0.00001781, -0.00001840, 0.00002045, 0.00001444, -0.00003008, 0.00001754, 0.00001433, -0.00001371, 0.00001259, -0.00000925, -0.00001128, 0.00001360, 0.00000470, 0.00001196, 0.00000137, -0.00000446, 0.00000202, -0.00001127, 0.00001875, 0.00000134, -0.00003488, 0.00003319, 0.00002001, -0.00003894, 0.00001026, 0.00000966, -0.00001864, 0.00000658, -0.00000285, -0.00000740, 0.00000815, 0.00000534, -0.00000015, -0.00001209, 0.00000203, 0.00001806, -0.00000136, -0.00000402, -0.00000880, -0.00001329, 0.00001403, 0.00000664, -0.00001288, 0.00000473, 0.00000518, -0.00000256, 0.00000016, -0.00000858, 0.00000148, 0.00001468, -0.00000637, -0.00001587, 0.00000140, 0.00001182, 0.00000886, -0.00000362, -0.00000577, 0.00000206, -0.00000188, -0.00000047, -0.00000037, -0.00000771, 0.00000615, 0.00000701, -0.00001047, 0.00000163, 0.00000867, -0.00000722, -0.00000473, 0.00000594, -0.00000254, -0.00001582, -0.00001123, -0.00001149, -0.00002873, -0.00002701, -0.00000931, 0.00001819, 0.00004942, 0.00003781, 0.00001232, 0.00001518, 0.00001513, 0.00001612, 0.00002837, 0.00002629, 0.00001812, 0.00004414, 0.00007000, },
		{ -0.00006784, -0.00002631, -0.00000875, -0.00001554, -0.00000738, -0.00000604, -0.00000285, -0.00000158, -0.00000519, -0.00000512, -0.00000734, -0.00000095, 0.00000302, -0.00000146, -0.00000123, -0.00000414, -0.00000352, -0.00000494, -0.00000858, -0.00000077, -0.00000546, -0.00000923, 0.00000221, -0.00000021, -0.00000389, 0.00000109, 0.00000328, -0.00000254, -0.00001027, -0.00000222, -0.00000041, -0.00000510, -0.00000191, -0.00000463, 0.00000102, -0.00001293, -0.00002142, 0.00001790, 0.00000261, -0.00001153, 0.00002975, -0.00000699, -0.00001755, 0.00003012, -0.00000379, -0.00000698, 0.00001357, -0.00001820, -0.00000029, 0.00001552, 0.00000387, 0.00001116, -0.00000329, -0.00000342, 0.00000030, 0.00000175, 0.00001442, -0.00002264, -0.00000503, 0.00004221, -0.00001988, -0.00002205, 0.00002552, -0.00001353, -0.00000897, 0.00001173, -0.00001391, 0.00000102, 0.00001221, 0.00000005, -0.00000504, -0.00000812, 0.00001110, 0.00001051, -0.00000571, -0.00000110, -0.00001403, -0.00000619, 0.00001871, -0.00000395, -0.00000930, 0.00001058, -0.00000005, -0.00000263, -0.00000100, -0.00000892, 0.00000846, 0.00001140, -0.00001439, -0.00001104, 0.00000793, 0.00001029, 0.00000464, -0.00000447, -0.00000371, 0.00000094, -0.00000204, -0.00000036, -0.00000376, -0.00000177, 0.00000986, -0.00000301, -0.00000754, 0.00000967, 0.00000031, -0.00000914, 0.00000306, 0.00000237, -0.00000984, -0.00001279, -0.00001015, -0.00002151, -0.00002993, -0.00001452, 0.00000468, 0.00003302, 0.00004838, 0.00002087, 0.00000821, 0.00001709, 0.00001260, 0.00002042, 0.00002995, 0.00001833, 0.00002049, 0.00005601, 0.00005583, -0.00002468, },
		{ -0.00001466, -0.00001098, -0.00001401, -0.00000596, -0.00000461, -0.00000101, -0.00000349, -0.00000627, -0.00000602, -0.00000492, 0.00000245, 0.00000065, -0.00000228, -0.00000081, -0.00000475, -0.00000457, -0.00000706, -0.00000573, -0.00000027, -0.00000810, -0.00000536, 0.00000262, -0.00000271, -0.00000209, 0.00000195, 0.00000117, -0.00000621, -0.00000755, 0.00000071, -0.00000418, -0.00000351, -0.00000104, -0.00000579, -0.00000151, -0.00001823, -0.00001012, 0.00001891, -0.00000909, 0.00000293, 0.00002523, -0.00002004, 0.00000427, 0.00002539, -0.00001659, 0.00000633, 0.00000507, -0.00002013, 0.00001296, 0.00001162, 0.00000411, 0.00000964, -0.00000855, -0.00000081, 0.00000641, 0.00000675, -0.00000423, -0.00002040, 0.00002569, 0.00001664, -0.00003710, 0.00000978, 0.00001293, -0.00002540, 0.00001033, 0.00000101, -0.00001692, 0.00001477, 0.00000699, -0.00000573, -0.00000440, -0.00000150, 0.00001293, 0.00000170, -0.00000363, -0.00000223, -0.00001764, 0.00000495, 0.00001568, -0.00001220, -0.00000004, 0.00001047, -0.00000566, -0.00000016, -0.00000366, -0.00000651, 0.00001454, 0.00000328, -0.00001846, -0.00000237, 0.00001007, 0.00000676, 0.00000315, -0.00000493, -0.00000325, 0.00000080, -0.00000230, -0.00000231, -0.00000239, 0.00000506, 0.00000484, -0.00000870, 0.00000225, 0.00000839, -0.00000833, -0.00000248, 0.00000523, -0.00000581, -0.00001001, -0.00000966, -0.00001696, -0.00002794, -0.00002109, -0.00000153, 0.00001663, 0.00004256, 0.00003759, 0.00000698, 0.00001095, 0.00001665, 0.00001181, 0.00002637, 0.00002556, 0.00001174, 0.00002991, 0.00006147, 0.00003243, -0.00004552, -0.00005608, },
		{ -0.00001321, -0.00001158, -0.00000456, -0.00000255, -0.00000102, -0.00000599, -0.00000684, -0.00000535, -0.00000122, 0.00000283, -0.00000214, -0.00000108, -0.00000122, -0.00000634, -0.00000567, -0.00000713, -0.00000282, -0.00000177, -0.00000862, -0.00000192, 0.00000100, -0.00000332, -0.00000035, 0.00000126, -0.00000156, -0.00000743, -0.00000332, -0.00000045, -0.00000648, 0.00000047, -0.00000351, -0.00000735, -0.00000346, -0.00001933, -0.00000009, 0.00001279, -0.00001227, 0.00001568, 0.00001241, -0.00001848, 0.00002125, 0.00000879, -0.00001595, 0.00001564, -0.00000850, -0.00001108, 0.00002085, 0.00000443, 0.00000775, 0.00000408, -0.00001206, 0.00000878, 0.00001100, -0.00000296, -0.00001491, 0.00000261, 0.00003005, -0.00001768, -0.00002128, 0.00002615, -0.00001354, -0.00001395, 0.00001925, -0.00001720, -0.00000416, 0.00002135, -0.00000522, -0.00000486, -0.00000076, 0.00000236, 0.00000949, -0.00000289, 0.00000030, -0.00000789, -0.00001534, 0.00001470, 0.00000614, -0.00001294, 0.00000921, 0.00000399, -0.00000739, 0.00000247, -0.00000746, -0.00000043, 0.00001646, -0.00000717, -0.00001575, 0.00000546, 0.00000730, 0.00000541, 0.00000281, -0.00000704, -0.00000191, 0.00000098, -0.00000465, -0.00000205, 0.00000313, 0.00000576, -0.00000334, -0.00000485, 0.00000912, -0.00000092, -0.00000849, 0.00000538, -0.00000144, -0.00000963, -0.00000547, -0.00001384, -0.00002539, -0.00002450, -0.00000833, 0.00000826, 0.00002736, 0.00004431, 0.00002096, 0.00000119, 0.00001614, 0.00001340, 0.00001493, 0.00002971, 0.00001633, 0.00001161, 0.00004193, 0.00005656, 0.00000659, -0.00005453, -0.00004131, -0.00000915, },
		{ -0.00000884, -0.00000284, -0.00000096, -0.00000302, -0.00000810, -0.00000614, -0.00000308, 0.00000138, 0.00000064, -0.00000298, 0.00000066, -0.00000353, -0.00000774, -0.00000584, -0.00000592, -0.00000103, -0.00000383, -0.00000744, -0.00000013, -0.00000085, -0.00000247, 0.00000016, -0.00000035, -0.00000359, -0.00000599, -0.00000072, -0.00000419, -0.00000452, 0.00000282, -0.00000931, -0.00000609, -0.00000487, -0.00001860, 0.00000660, 0.00000476, -0.00000842, 0.00002187, 0.00000056, -0.00000685, 0.00002496, -0.00000717, -0.00000482, 0.00001407, -0.00001723, 0.00000444, 0.00001883, 0.00000059, 0.00001127, -0.00000621, -0.00000653, 0.00002094, 0.00000304, -0.00001568, -0.00000533, 0.00002132, 0.00000755, -0.00003091, 0.00000797, 0.00001431, -0.00002785, 0.00000951, 0.00000737, -0.00002465, 0.00001664, 0.00001236, -0.00001348, 0.00000288, -0.00000005, 0.00000292, 0.00000533, -0.00000283, 0.00000147, -0.00001425, -0.00000711, 0.00001806, -0.00000413, -0.00000615, 0.00001246, -0.00000444, -0.00000406, 0.00000210, -0.00000967, 0.00000770, 0.00001212, -0.00001492, -0.00000769, 0.00000788, 0.00000362, 0.00000746, -0.00000002, -0.00000896, 0.00000142, -0.00000154, -0.00000665, 0.00000326, 0.00000643, -0.00000038, -0.00000581, 0.00000394, 0.00000623, -0.00000867, -0.00000021, 0.00000507, -0.00001020, -0.00000486, -0.00000546, -0.00002416, -0.00002618, -0.00001379, 0.00000178, 0.00001632, 0.00003590, 0.00003662, 0.00000533, 0.00000414, 0.00001882, 0.00001008, 0.00002120, 0.00002679, 0.00000735, 0.00001927, 0.00004992, 0.00004251, -0.00001489, -0.00005393, -0.00002837, -0.00000785, -0.00001408, },
		{ -0.00000090, -0.00000106, -0.00000625, -0.00000859, -0.00000398, -0.00000062, 0.00000159, -0.00000182, -0.00000154, 0.00000071, -0.00000687, -0.00000774, -0.00000533, -0.00000432, -0.00000044, -0.00000552, -0.00000564, 0.00000019, -0.00000180, -0.00000147, -0.00000073, -0.00000180, -0.00000400, -0.00000355, -0.00000147, -0.00000664, 0.00000060, -0.00000033, -0.00001420, -0.00000101, -0.00000846, -0.00001668, 0.00001061, -0.00000252, -0.00000096, 0.00002258, -0.00000542, 0.00000552, 0.00001756, -0.00001369, 0.00000698, 0.00000430, -0.00001493, 0.00001677, 0.00001053, 0.00000318, 0.00000829, -0.00001398, 0.00001018, 0.00002277, -0.00001508, -0.00001420, 0.00001406, 0.00001700, -0.00001806, -0.00001622, 0.00002243, -0.00001118, -0.00001753, 0.00002215, -0.00001504, -0.00001144, 0.00002791, -0.00000652, -0.00000876, 0.00000946, -0.00000418, 0.00000348, 0.00000265, -0.00000081, -0.00000119, -0.00001724, 0.00000309, 0.00001447, -0.00000941, 0.00000304, 0.00000820, -0.00000923, 0.00000132, -0.00000184, -0.00000768, 0.00001393, 0.00000328, -0.00001612, 0.00000030, 0.00000489, 0.00000383, 0.00000936, -0.00000613, -0.00000712, 0.00000371, -0.00000712, -0.00000329, 0.00001011, 0.00000222, -0.00000616, 0.00000012, 0.00000786, -0.00000303, -0.00000689, 0.00000714, -0.00000433, -0.00001104, 0.00000210, -0.00001556, -0.00003087, -0.00001695, -0.00000372, 0.00000850, 0.00002488, 0.00003849, 0.00002196, -0.00000289, 0.00001159, 0.00001694, 0.00001010, 0.00002658, 0.00001771, 0.00000440, 0.00003084, 0.00004948, 0.00002478, -0.00002891, -0.00004832, -0.00001891, -0.00000833, -0.00001362, -0.00000598, },
		{ -0.00000340, -0.00000895, -0.00000679, -0.00000138, 0.00000055, 0.00000017, -0.00000262, 0.00000044, -0.00000174, -0.00000931, -0.00000639, -0.00000478, -0.00000267, -0.00000080, -0.00000660, -0.00000395, -0.00000014, -0.00000185, -0.00000141, -0.00000217, -0.00000209, -0.00000314, -0.00000239, -0.00000420, -0.00000501, 0.00000412, -0.00000850, -0.00001298, 0.00000411, -0.00001560, -0.00001174, 0.00001169, -0.00000821, 0.00000872, 0.00002028, -0.00000670, 0.00001292, 0.00000690, -0.00001094, 0.00001224, -0.00000513, -0.00000435, 0.00001994, 0.00000387, 0.00000756, -0.00000169, -0.00000927, 0.00002785, 0.00000671, -0.00002696, 0.00000436, 0.00002160, -0.00000485, -0.00002368, 0.00000740, 0.00001193, -0.00002528, 0.00000569, 0.00001237, -0.00002680, 0.00001308, 0.00001886, -0.00001925, 0.00000607, 0.00000658, -0.00000844, 0.00000597, 0.00000066, 0.00000103, -0.00000608, -0.00001559, 0.00001108, 0.00000749, -0.00000827, 0.00000867, 0.00000014, -0.00000805, 0.00000444, -0.00000670, -0.00000172, 0.00001510, -0.00000528, -0.00001165, 0.00000338, 0.00000135, 0.00000837, 0.00000649, -0.00001158, -0.00000137, 0.00000082, -0.00001074, 0.00000649, 0.00001075, -0.00000679, -0.00000391, 0.00000762, 0.00000238, -0.00000848, 0.00000206, 0.00000465, -0.00001343, -0.00000210, 0.00000011, -0.00003004, -0.00002543, -0.00000488, 0.00000137, 0.00001573, 0.00003210, 0.00003244, 0.00000698, -0.00000188, 0.00001767, 0.00001264, 0.00001419, 0.00002667, 0.00000710, 0.00000971, 0.00003978, 0.00004139, 0.00000891, -0.00003677, -0.00004116, -0.00001209, -0.00000936, -0.00001223, -0.00000275, 0.00000033, },
		{ -0.00000926, -0.00000352, 0.00000015, 0.00000025, -0.00000121, -0.00000176, 0.00000090, -0.00000550, -0.00000960, -0.00000476, -0.00000431, -0.00000110, -0.00000201, -0.00000707, -0.00000257, -0.00000047, -0.00000178, -0.00000249, -0.00000281, -0.00000111, -0.00000249, -0.00000331, -0.00000572, -0.00000046, 0.00000180, -0.00001613, -0.00000477, 0.00000338, -0.00002314, -0.00000306, 0.00000794, -0.00001003, 0.00002007, 0.00001490, -0.00000587, 0.00001561, -0.00000142, -0.00000439, 0.00001114, -0.00000834, 0.00000671, 0.00001607, 0.00000262, 0.00000669, -0.00000984, 0.00000886, 0.00003064, -0.00001821, -0.00001841, 0.00002365, 0.00000747, -0.00002193, -0.00000831, 0.00001661, -0.00000919, -0.00001705, 0.00001937, -0.00000953, -0.00001646, 0.00002846, -0.00000339, -0.00001417, 0.00001781, -0.00000523, -0.00000666, 0.00000808, -0.00000122, 0.00000216, -0.00001147, -0.00001033, 0.00001525, 0.00000114, -0.00000362, 0.00000841, -0.00000634, -0.00000304, 0.00000341, -0.00000915, 0.00000519, 0.00001158, -0.00000974, -0.00000631, 0.00000167, 0.00000222, 0.00001248, -0.00000142, -0.00001156, 0.00000338, -0.00000668, -0.00000618, 0.00001607, 0.00000152, -0.00001122, 0.00000557, 0.00000740, -0.00000650, -0.00000418, 0.00000773, -0.00000612, -0.00001218, 0.00000691, -0.00001442, -0.00003593, -0.00001006, 0.00000104, 0.00000495, 0.00002511, 0.00003307, 0.00002028, -0.00000190, 0.00000479, 0.00001921, 0.00000966, 0.00001940, 0.00002038, 0.00000098, 0.00002011, 0.00004169, 0.00003042, -0.00000320, -0.00004140, -0.00003312, -0.00000710, -0.00001088, -0.00000965, 0.00000053, -0.00000057, -0.00000702, },
		{ -0.00000076, 0.00000009, -0.00000044, -0.00000163, -0.00000059, -0.00000074, -0.00000858, -0.00000805, -0.00000373, -0.00000352, -0.00000001, -0.00000387, -0.00000683, -0.00000142, -0.00000071, -0.00000238, -0.00000383, -0.00000175, -0.00000002, -0.00000332, -0.00000479, -0.00000421, 0.00000269, -0.00000584, -0.00001701, 0.00000510, -0.00000547, -0.00002532, 0.00000585, -0.00000055, -0.00000417, 0.00002975, 0.00000549, -0.00000268, 0.00001562, -0.00000667, 0.00000208, 0.00000792, -0.00000601, 0.00001318, 0.00001107, 0.00000337, 0.00000018, -0.00000597, 0.00002784, 0.00001265, -0.00003197, 0.00000634, 0.00002400, -0.00001602, -0.00001861, 0.00000921, 0.00000614, -0.00001954, 0.00000216, 0.00001269, -0.00002258, 0.00000685, 0.00002223, -0.00001982, 0.00000465, 0.00001503, -0.00001632, 0.00000156, 0.00000634, -0.00000144, 0.00000184, -0.00001636, -0.00000258, 0.00001604, -0.00000302, 0.00000070, 0.00000418, -0.00000840, 0.00000192, -0.00000044, -0.00000811, 0.00001023, 0.00000652, -0.00001045, -0.00000406, -0.00000014, 0.00000745, 0.00001105, -0.00000932, -0.00000609, 0.00000192, -0.00001190, 0.00000645, 0.00001598, -0.00001133, -0.00000450, 0.00001266, -0.00000175, -0.00000862, 0.00000474, 0.00000296, -0.00001422, -0.00000119, 0.00000397, -0.00003180, -0.00002580, 0.00000355, -0.00000015, 0.00001304, 0.00003218, 0.00002572, 0.00000861, -0.00000313, 0.00001179, 0.00001720, 0.00000965, 0.00002174, 0.00001105, 0.00000229, 0.00002957, 0.00003722, 0.00002092, -0.00001342, -0.00004382, -0.00002358, -0.00000448, -0.00001239, -0.00000539, 0.00000236, -0.00000402, -0.00000955, -0.00000678, },
		{ -0.00000060, -0.00000061, -0.00000151, -0.00000034, -0.00000355, -0.00000980, -0.00000586, -0.00000334, -0.00000221, 0.00000004, -0.00000590, -0.00000584, -0.00000053, -0.00000126, -0.00000376, -0.00000399, 0.00000056, -0.00000049, -0.00000512, -0.00000463, -0.00000116, 0.00000106, -0.00001353, -0.00000987, 0.00000907, -0.00001820, -0.00001951, 0.00000921, -0.00000870, 0.00001028, 0.00003171, -0.00000548, 0.00000442, 0.00001265, -0.00000945, 0.00000822, 0.00000491, -0.00000181, 0.00001567, 0.00000744, 0.00000131, -0.00000399, 0.00000992, 0.00003144, -0.00001393, -0.00002214, 0.00002644, 0.00000274, -0.00002710, -0.00000076, 0.00001143, -0.00000946, -0.00001213, 0.00001292, -0.00000468, -0.00001580, 0.00002285, 0.00000129, -0.00001688, 0.00002083, -0.00000107, -0.00001693, 0.00000974, 0.00000113, 0.00000159, -0.00000187, -0.00001901, 0.00000694, 0.00001381, -0.00000546, 0.00000302, -0.00000066, -0.00000663, 0.00000443, -0.00000473, -0.00000465, 0.00001301, 0.00000215, -0.00001070, -0.00000423, 0.00000198, 0.00001209, 0.00000403, -0.00001190, -0.00000074, -0.00000489, -0.00000774, 0.00001828, 0.00000373, -0.00001608, 0.00000898, 0.00000864, -0.00001076, -0.00000109, 0.00000753, -0.00000783, -0.00001097, 0.00000772, -0.00001215, -0.00003717, -0.00000609, 0.00000587, -0.00000034, 0.00002551, 0.00003051, 0.00001446, 0.00000250, 0.00000005, 0.00001635, 0.00001416, 0.00001119, 0.00001976, 0.00000359, 0.00000907, 0.00003400, 0.00003055, 0.00001335, -0.00002393, -0.00004242, -0.00001334, -0.00000513, -0.00001222, -0.00000027, 0.00000090, -0.00000832, -0.00000902, -0.00000322, 0.00000006, },
		{ -0.00000051, -0.00000165, -0.00000110, -0.00000620, -0.00000926, -0.00000403, -0.00000304, -0.00000076, -0.00000119, -0.00000741, -0.00000422, -0.00000012, -0.00000251, -0.00000504, -0.00000213, 0.00000226, -0.00000300, -0.00000585, -0.00000256, 0.00000010, -0.00000444, -0.00001577, 0.00000036, 0.00000371, -0.00002739, -0.00000931, 0.00000477, -0.00000891, 0.00002760, 0.00002240, -0.00001158, 0.00001383, 0.00000528, -0.00000818, 0.00001401, 0.00000140, 0.00000281, 0.00001629, 0.00000309, -0.00000229, 0.00000187, 0.00002540, 0.00001541, -0.00002839, 0.00000303, 0.00002374, -0.00002229, -0.00001658, 0.00001219, -0.00000075, -0.00001364, 0.00000153, 0.00000776, -0.00001461, 0.00000237, 0.00001952, -0.00001526, 0.00000151, 0.00002059, -0.00001779, -0.00000654, 0.00001102, -0.00000259, 0.00000595, -0.00000999, -0.00001643, 0.00001634, 0.00000822, -0.00000648, 0.00000377, -0.00000419, -0.00000309, 0.00000424, -0.00000813, 0.00000022, 0.00001461, -0.00000238, -0.00001225, -0.00000286, 0.00000706, 0.00001163, -0.00000354, -0.00000922, -0.00000076, -0.00000962, 0.00000497, 0.00001878, -0.00001183, -0.00000694, 0.00001680, -0.00000427, -0.00000966, 0.00000779, -0.00000008, -0.00001306, -0.00000001, 0.00000352, -0.00002914, -0.00002490, 0.00000832, -0.00000068, 0.00000839, 0.00003428, 0.00001945, 0.00000680, 0.00000136, 0.00000403, 0.00001886, 0.00001101, 0.00001228, 0.00001527, 0.00000053, 0.00001704, 0.00003366, 0.00002490, 0.00000508, -0.00003392, -0.00003556, -0.00000523, -0.00000814, -0.00000892, 0.00000310, -0.00000385, -0.00001045, -0.00000575, -0.00000102, -0.00000066, -0.00000061, },
		{ -0.00000216, -0.00000230, -0.00000797, -0.00000770, -0.00000284, -0.00000243, 0.00000015, -0.00000340, -0.00000786, -0.00000240, -0.00000056, -0.00000427, -0.00000486, 0.00000097, 0.00000148, -0.00000588, -0.00000413, -0.00000083, -0.00000204, -0.00000954, -0.00001156, 0.00000676, -0.00000782, -0.00002873, -0.00000158, -0.00000286, 0.00000229, 0.00003788, 0.00000573, -0.00000764, 0.00001965, -0.00000461, -0.00000075, 0.00001673, -0.00000264, 0.00000880, 0.00001422, -0.00000308, -0.00000068, 0.00001543, 0.00002683, -0.00000742, -0.00002058, 0.00002136, 0.00000037, -0.00002990, 0.00000368, 0.00000815, -0.00001176, -0.00000491, 0.00000637, -0.00000362, -0.00000929, 0.00001435, 0.00000335, -0.00001422, 0.00001857, 0.00000467, -0.00002267, 0.00000599, 0.00000528, -0.00000037, 0.00000640, -0.00001958, -0.00000667, 0.00002182, -0.00000000, -0.00000536, 0.00000349, -0.00000612, 0.00000078, 0.00000175, -0.00001005, 0.00000678, 0.00001455, -0.00000904, -0.00001280, 0.00000217, 0.00001040, 0.00000679, -0.00000718, -0.00000636, -0.00000487, -0.00000563, 0.00001649, 0.00000638, -0.00001753, 0.00000894, 0.00001071, -0.00001403, 0.00000110, 0.00000728, -0.00001029, -0.00000695, 0.00000591, -0.00001189, -0.00003241, -0.00000465, 0.00000837, -0.00000415, 0.00002401, 0.00003096, 0.00000656, 0.00000605, 0.00000104, 0.00000835, 0.00001970, 0.00000710, 0.00001292, 0.00001059, 0.00000108, 0.00002351, 0.00003119, 0.00001991, -0.00000576, -0.00003994, -0.00002454, -0.00000181, -0.00001053, -0.00000338, 0.00000213, -0.00000913, -0.00000866, -0.00000232, -0.00000115, -0.00000117, 0.00000017, -0.00000102, },
		{ -0.00000356, -0.00000878, -0.00000574, -0.00000211, -0.00000154, -0.00000002, -0.00000584, -0.00000706, -0.00000099, -0.00000200, -0.00000555, -0.00000263, 0.00000312, -0.00000179, -0.00000655, -0.00000118, -0.00000192, -0.00000562, -0.00001077, -0.00000473, 0.00000600, -0.00001916, -0.00002409, 0.00000001, -0.00000555, 0.00002002, 0.00003458, -0.00000842, 0.00000449, 0.00001668, -0.00001083, 0.00001106, 0.00001322, -0.00000412, 0.00001527, 0.00000717, -0.00000658, 0.00000984, 0.00002510, 0.00001299, -0.00001889, -0.00000050, 0.00001734, -0.00002325, -0.00001577, 0.00001380, -0.00000619, -0.00000963, 0.00000410, 0.00000000, -0.00000716, 0.00000269, 0.00001131, -0.00000890, 0.00000068, 0.00002042, -0.00001374, -0.00001416, 0.00001129, -0.00000075, 0.00000624, -0.00000089, -0.00002423, 0.00000797, 0.00001979, -0.00000785, -0.00000175, 0.00000179, -0.00000634, 0.00000408, -0.00000288, -0.00000909, 0.00001483, 0.00001013, -0.00001685, -0.00000864, 0.00000802, 0.00000910, 0.00000214, -0.00000774, -0.00000709, -0.00000621, 0.00000519, 0.00001686, -0.00000879, -0.00000849, 0.00001746, -0.00000432, -0.00001094, 0.00001013, -0.00000341, -0.00001146, 0.00000315, -0.00000135, -0.00002460, -0.00001958, 0.00000778, -0.00000059, 0.00000436, 0.00003510, 0.00001619, 0.00000114, 0.00000836, -0.00000042, 0.00001430, 0.00001723, 0.00000304, 0.00001443, 0.00000595, 0.00000382, 0.00002858, 0.00002786, 0.00001324, -0.00001759, -0.00003931, -0.00001337, -0.00000271, -0.00000973, 0.00000099, -0.00000296, -0.00001117, -0.00000433, -0.00000129, -0.00000224, -0.00000030, 0.00000058, -0.00000239, -0.00000248, },
		{ -0.00000871, -0.00000377, -0.00000163, -0.00000074, -0.00000136, -0.00000770, -0.00000533, -0.00000062, -0.00000394, -0.00000517, 0.00000062, 0.00000250, -0.00000522, -0.00000412, -0.00000011, -0.00000551, -0.00000721, -0.00000853, 0.00000008, -0.00000015, -0.00002633, -0.00001844, -0.00000188, 0.00000128, 0.00003416, 0.00002017, -0.00001176, 0.00001644, 0.00000616, -0.00000777, 0.00002070, 0.00000462, 0.00000021, 0.00001761, -0.00000296, -0.00000081, 0.00002331, 0.00002150, -0.00000346, -0.00001286, 0.00001167, -0.00000369, -0.00002797, 0.00000461, 0.00000595, -0.00001416, 0.00000161, 0.00000244, -0.00000664, -0.00000005, 0.00000750, 0.00000034, -0.00000699, 0.00001431, 0.00000754, -0.00002168, -0.00000124, 0.00000800, -0.00000056, 0.00001033, -0.00001351, -0.00001882, 0.00002083, 0.00001029, -0.00001140, 0.00000270, -0.00000151, -0.00000432, 0.00000538, -0.00000865, -0.00000320, 0.00002137, -0.00000043, -0.00002095, 0.00000025, 0.00000979, 0.00000556, 0.00000028, -0.00000958, -0.00000876, -0.00000020, 0.00001325, 0.00000608, -0.00001392, 0.00000659, 0.00001118, -0.00001437, 0.00000168, 0.00000699, -0.00001309, -0.00000158, 0.00000412, -0.00001555, -0.00002268, -0.00000282, 0.00000574, -0.00000471, 0.00002104, 0.00003176, 0.00000041, 0.00000545, 0.00000742, -0.00000030, 0.00002070, 0.00000980, 0.00000198, 0.00001626, 0.00000042, 0.00000922, 0.00003248, 0.00002245, 0.00000424, -0.00002669, -0.00003294, -0.00000558, -0.00000489, -0.00000619, 0.00000105, -0.00000861, -0.00000847, -0.00000097, -0.00000277, -0.00000211, 0.00000126, -0.00000050, -0.00000378, -0.00000225, -0.00000494, },
		{ -0.00000205, -0.00000135, -0.00000044, -0.00000352, -0.00000840, -0.00000342, -0.00000151, -0.00000523, -0.00000288, 0.00000273, -0.00000057, -0.00000607, -0.00000077, -0.00000284, -0.00000835, -0.00000561, -0.00000598, 0.00000130, -0.00000802, -0.00002969, -0.00001455, -0.00000140, 0.00001466, 0.00003730, 0.00000412, -0.00000409, 0.00002026, -0.00000423, 0.00000380, 0.00002192, -0.00000298, 0.00000870, 0.00001195, -0.00000836, 0.00001442, 0.00002828, 0.00000773, -0.00000885, -0.00000035, 0.00000518, -0.00002177, -0.00001337, 0.00001226, -0.00000911, -0.00000753, 0.00000789, -0.00000694, -0.00000326, 0.00000755, 0.00000130, -0.00000488, 0.00000425, 0.00001534, -0.00000812, -0.00001707, 0.00000626, 0.00000267, 0.00000544, 0.00000606, -0.00002359, -0.00000443, 0.00002520, -0.00000198, -0.00000896, 0.00000509, -0.00000509, -0.00000030, 0.00000307, -0.00001253, 0.00000766, 0.00002128, -0.00001422, -0.00001681, 0.00000863, 0.00000628, 0.00000442, -0.00000154, -0.00001394, -0.00000510, 0.00000903, 0.00001089, -0.00000529, -0.00000607, 0.00001374, -0.00000308, -0.00001062, 0.00001033, -0.00000589, -0.00001035, 0.00000796, -0.00000807, -0.00002164, -0.00000873, 0.00000355, -0.00000252, 0.00000393, 0.00003270, 0.00001554, -0.00000528, 0.00001286, 0.00000167, 0.00000508, 0.00002281, -0.00000033, 0.00000663, 0.00001490, -0.00000471, 0.00001852, 0.00003305, 0.00001397, -0.00000471, -0.00003089, -0.00002426, -0.00000162, -0.00000569, -0.00000310, -0.00000296, -0.00001059, -0.00000336, -0.00000108, -0.00000437, -0.00000009, 0.00000171, -0.00000279, -0.00000409, -0.00000188, -0.00000649, -0.00000771, },
		},
		{ { -0.00000583, -0.00000622, -0.00000437, -0.00000295, -0.00001528, -0.00000963, -0.00003057, -0.00004612, 0.00000853, 0.00004572, 0.00002990, 0.00000836, 0.00002154, 0.00000643, 0.00001063, 0.00002024, 0.00000469, 0.00002727, 0.00002653, 0.00000906, 0.00000502, -0.00001234, -0.00003656, -0.00000210, 0.00000484, -0.00001169, 0.00000702, -0.00001526, -0.00000472, 0.00001103, -0.00000835, -0.00000396, 0.00000227, -0.00001007, -0.00000437, -0.00000246, -0.00000172, 0.00000995, 0.00000200, 0.00000419, -0.00000906, -0.00000275, -0.00000152, -0.00000905, 0.00000688, -0.00000221, -0.00000124, 0.00000527, -0.00000025, -0.00000899, 0.00000266, -0.00000210, 0.00000728, 0.00001189, -0.00001730, -0.00000317, -0.00000981, 0.00000004, 0.00002439, -0.00000137, -0.00001767, 0.00000335, 0.00000381, -0.00000497, 0.00000613, -0.00000497, 0.00000381, 0.00000335, -0.00001767, -0.00000137, 0.00002439, 0.00000004, -0.00000981, -0.00000317, -0.00001730, 0.00001189, 0.00000728, -0.00000210, 0.00000266, -0.00000899, -0.00000025, 0.00000527, -0.00000124, -0.00000221, 0.00000688, -0.00000905, -0.00000152, -0.00000275, -0.00000906, 0.00000419, 0.00000200, 0.00000995, -0.00000172, -0.00000246, -0.00000437, -0.00001007, 0.00000227, -0.00000396, -0.00000835, 0.00001103, -0.00000472, -0.00001526, 0.00000702, -0.00001169, 0.00000484, -0.00000210, -0.00003656, -0.00001234, 0.00000502, 0.00000906, 0.00002653, 0.00002727, 0.00000469, 0.00002024, 0.00001063, 0.00000643, 0.00002154, 0.00000836, 0.00002990, 0.00004572, 0.00000853, -0.00004612, -0.00003057, -0.00000963, -0.00001528, -0.00000295, -0.00000437, -0.00000622, },
		{ -0.00000590, -0.00000330, -0.00000518, -0.00001644, -0.00000922, -0.00003620, -0.00004548, 0.00001808, 0.00004936, 0.00002542, 0.00001078, 0.00002154, 0.00000467, 0.00001607, 0.00001674, 0.00000758, 0.00003489, 0.00002272, 0.00000560, 0.00000082, -0.00002198, -0.00002941, 0.00000418, -0.00000404, -0.00000504, 0.00000501, -0.00001960, 0.00000360, 0.00000630, -0.00001051, -0.00000017, -0.00000059, -0.00000946, -0.00000388, -0.00000334, 0.00000296, 0.00000674, 0.00000119, 0.00000462, -0.00001112, -0.00000016, -0.00000405, -0.00000650, 0.00000725, -0.00000501, 0.00000205, 0.00000483, -0.00000389, -0.00000623, 0.00000235, -0.00000254, 0.00001228, 0.00000528, -0.00001805, -0.00000196, -0.00001016, 0.00000779, 0.00002073, -0.00000948, -0.00001276, 0.00000685, -0.00000130, -0.00000103, 0.00000442, -0.00000624, 0.00000865, -0.00000351, -0.00001855, 0.00000798, 0.00002389, -0.00000740, -0.00000772, -0.00000541, -0.00001409, 0.00001632, 0.00000136, 0.00000068, 0.00000063, -0.00001011, 0.00000410, 0.00000320, -0.00000278, 0.00000089, 0.00000481, -0.00001007, 0.00000015, -0.00000488, -0.00000683, 0.00000286, 0.00000529, 0.00001039, -0.00000630, 0.00000019, -0.00000675, -0.00000877, 0.00000374, -0.00000815, -0.00000277, 0.00001140, -0.00001191, -0.00000766, 0.00000346, -0.00001376, 0.00001358, -0.00001432, -0.00003715, -0.00000180, 0.00000636, 0.00001368, 0.00002719, 0.00001916, 0.00000523, 0.00002167, 0.00000536, 0.00000979, 0.00001971, 0.00000744, 0.00003392, 0.00004023, 0.00000055, -0.00004546, -0.00002581, -0.00001027, -0.00001326, -0.00000155, -0.00000576, -0.00000610, -0.00000592, },
		{ -0.00000291, -0.00000782, -0.00001667, -0.00000946, -0.00004252, -0.00004295, 0.00002863, 0.00005072, 0.00002142, 0.00001396, 0.00001995, 0.00000513, 0.00002023, 0.00001259, 0.00001411, 0.00004012, 0.00001707, 0.00000246, -0.00000634, -0.00002729, -0.00001920, 0.00000385, -0.00000922, 0.00000256, -0.00000147, -0.00001887, 0.00000976, -0.00000056, -0.00000908, 0.00000157, -0.00000317, -0.00000809, -0.00000480, -0.00000178, 0.00000583, 0.00000245, 0.00000263, 0.00000338, -0.00001202, 0.00000198, -0.00000667, -0.00000286, 0.00000549, -0.00000610, 0.00000561, 0.00000205, -0.00000550, -0.00000347, 0.00000067, -0.00000046, 0.00001486, -0.00000195, -0.00001676, -0.00000183, -0.00000815, 0.00001408, 0.00001350, -0.00001416, -0.00000585, 0.00000631, -0.00000429, 0.00000334, 0.00000013, -0.00000381, 0.00001064, -0.00001156, -0.00001479, 0.00001622, 0.00001978, -0.00001318, -0.00000483, -0.00000814, -0.00000868, 0.00001751, -0.00000365, 0.00000459, -0.00000369, -0.00000836, 0.00000730, -0.00000038, -0.00000201, 0.00000289, 0.00000211, -0.00000988, 0.00000078, -0.00000599, -0.00000547, 0.00000205, 0.00001005, 0.00000734, -0.00000880, 0.00000296, -0.00001032, -0.00000534, 0.00000253, -0.00001043, 0.00000426, 0.00000697, -0.00001499, -0.00000030, -0.00000417, -0.00000938, 0.00001771, -0.00002895, -0.00003009, 0.00000597, 0.00000667, 0.00001901, 0.00002412, 0.00001269, 0.00000791, 0.00002062, 0.00000152, 0.00001364, 0.00001645, 0.00000833, 0.00003658, 0.00003370, -0.00000570, -0.00004406, -0.00002190, -0.00001076, -0.00001068, -0.00000124, -0.00000702, -0.00000569, -0.00000613, -0.00000515, },
		{ -0.00001039, -0.00001612, -0.00001071, -0.00004916, -0.00003811, 0.00003937, 0.00004985, 0.00001860, 0.00001701, 0.00001761, 0.00000771, 0.00002214, 0.00000958, 0.00002333, 0.00004188, 0.00001069, -0.00000176, -0.00001467, -0.00002670, -0.00001016, -0.00000100, -0.00000911, 0.00000755, -0.00000946, -0.00001353, 0.00001162, -0.00000673, -0.00000544, 0.00000095, -0.00000431, -0.00000732, -0.00000581, 0.00000163, 0.00000583, -0.00000104, 0.00000518, 0.00000061, -0.00001132, 0.00000289, -0.00000840, 0.00000080, 0.00000213, -0.00000476, 0.00000779, -0.00000183, -0.00000482, -0.00000204, -0.00000091, 0.00000331, 0.00001437, -0.00000830, -0.00001433, -0.00000218, -0.00000388, 0.00001717, 0.00000449, -0.00001431, 0.00000041, 0.00000281, -0.00000399, 0.00000581, -0.00000454, 0.00000158, 0.00000822, -0.00001793, -0.00000741, 0.00002159, 0.00001335, -0.00001665, -0.00000204, -0.00001053, -0.00000202, 0.00001525, -0.00000620, 0.00000786, -0.00000913, -0.00000356, 0.00000770, -0.00000367, 0.00000048, 0.00000308, -0.00000013, -0.00000928, 0.00000088, -0.00000623, -0.00000539, 0.00000316, 0.00001434, 0.00000155, -0.00000812, 0.00000407, -0.00001349, -0.00000086, -0.00000141, -0.00000894, 0.00000972, -0.00000046, -0.00001275, 0.00000326, -0.00001218, 0.00000040, 0.00001414, -0.00004091, -0.00001732, 0.00000881, 0.00000838, 0.00002328, 0.00001803, 0.00000931, 0.00001088, 0.00001765, -0.00000029, 0.00001688, 0.00001257, 0.00001074, 0.00003737, 0.00002703, -0.00001041, -0.00004233, -0.00001867, -0.00001080, -0.00000796, -0.00000201, -0.00000776, -0.00000527, -0.00000628, -0.00000412, -0.00000341, },
		{ -0.00001514, -0.00001316, -0.00005560, -0.00003083, 0.00004944, 0.00004717, 0.00001721, 0.00001924, 0.00001553, 0.00001158, 0.00002165, 0.00000932, 0.00003355, 0.00004002, 0.00000394, -0.00000789, -0.00002129, -0.00002110, -0.00000540, -0.00000680, -0.00000468, 0.00000797, -0.00001568, -0.00000595, 0.00000904, -0.00001005, -0.00000171, -0.00000113, -0.00000394, -0.00000790, -0.00000547, 0.00000534, 0.00000322, -0.00000246, 0.00000742, -0.00000284, -0.00000931, 0.00000240, -0.00000862, 0.00000328, -0.00000158, -0.00000130, 0.00000747, -0.00000513, -0.00000269, -0.00000235, -0.00000108, 0.00000727, 0.00001113, -0.00001279, -0.00001168, -0.00000209, 0.00000164, 0.00001610, -0.00000383, -0.00001036, 0.00000385, -0.00000147, -0.00000089, 0.00000506, -0.00000713, 0.00000767, 0.00000158, -0.00002021, 0.00000143, 0.00002346, 0.00000602, -0.00001779, 0.00000008, -0.00001181, 0.00000453, 0.00001030, -0.00000559, 0.00000883, -0.00001375, 0.00000317, 0.00000465, -0.00000484, 0.00000319, 0.00000176, -0.00000131, -0.00000904, 0.00000127, -0.00000646, -0.00000607, 0.00000677, 0.00001611, -0.00000503, -0.00000457, 0.00000246, -0.00001458, 0.00000274, -0.00000655, -0.00000341, 0.00001089, -0.00000754, -0.00000657, 0.00000117, -0.00001639, 0.00001182, 0.00000277, -0.00004566, -0.00000311, 0.00000706, 0.00001309, 0.00002429, 0.00001097, 0.00000920, 0.00001255, 0.00001402, -0.00000030, 0.00001891, 0.00000891, 0.00001396, 0.00003630, 0.00002093, -0.00001400, -0.00004048, -0.00001592, -0.00001025, -0.00000562, -0.00000361, -0.00000774, -0.00000510, -0.00000613, -0.00000318, -0.00000476, -0.00001246, },
		{ -0.00001678, -0.00006133, -0.00002123, 0.00005819, 0.00004332, 0.00001708, 0.00002045, 0.00001452, 0.00001557, 0.00001952, 0.00001262, 0.00004294, 0.00003492, -0.00000356, -0.00001544, -0.00002368, -0.00001341, -0.00000568, -0.00001024, 0.00000131, 0.00000416, -0.00001796, 0.00000088, 0.00000364, -0.00000993, 0.00000041, -0.00000319, -0.00000290, -0.00000960, -0.00000305, 0.00000766, -0.00000074, -0.00000161, 0.00000831, -0.00000590, -0.00000678, 0.00000098, -0.00000735, 0.00000382, -0.00000413, 0.00000298, 0.00000460, -0.00000649, -0.00000053, -0.00000383, 0.00000070, 0.00000991, 0.00000625, -0.00001519, -0.00000936, -0.00000070, 0.00000676, 0.00001109, -0.00000912, -0.00000416, 0.00000359, -0.00000422, 0.00000316, 0.00000140, -0.00000608, 0.00001163, -0.00000725, -0.00001745, 0.00000927, 0.00002231, -0.00000107, -0.00001694, 0.00000134, -0.00001160, 0.00000966, 0.00000414, -0.00000218, 0.00000663, -0.00001562, 0.00000972, -0.00000107, -0.00000287, 0.00000442, 0.00000007, -0.00000174, -0.00000934, 0.00000249, -0.00000770, -0.00000614, 0.00001219, 0.00001417, -0.00001003, 0.00000022, -0.00000154, -0.00001280, 0.00000366, -0.00001035, 0.00000441, 0.00000680, -0.00001082, 0.00000014, -0.00000566, -0.00001410, 0.00001987, -0.00001310, -0.00004127, 0.00000778, 0.00000338, 0.00002048, 0.00002057, 0.00000555, 0.00001111, 0.00001233, 0.00001090, 0.00000073, 0.00001976, 0.00000596, 0.00001727, 0.00003379, 0.00001576, -0.00001688, -0.00003858, -0.00001339, -0.00000913, -0.00000413, -0.00000553, -0.00000700, -0.00000536, -0.00000545, -0.00000277, -0.00000665, -0.00001381, -0.00001415, },
		{ -0.00006594, -0.00000960, 0.00006517, 0.00003888, 0.00001786, 0.00002082, 0.00001493, 0.00001863, 0.00001704, 0.00001945, 0.00004999, 0.00002699, -0.00001228, -0.00002251, -0.00002095, -0.00000712, -0.00000943, -0.00000971, 0.00000584, -0.00000177, -0.00001612, 0.00000469, -0.00000210, -0.00000728, 0.00000030, -0.00000405, -0.00000238, -0.00001136, 0.00000109, 0.00000747, -0.00000440, 0.00000078, 0.00000761, -0.00000787, -0.00000456, -0.00000055, -0.00000531, 0.00000252, -0.00000451, 0.00000641, 0.00000021, -0.00000545, 0.00000034, -0.00000524, 0.00000400, 0.00001028, 0.00000111, -0.00001592, -0.00000731, 0.00000229, 0.00000965, 0.00000364, -0.00000999, 0.00000174, 0.00000037, -0.00000407, 0.00000590, -0.00000337, -0.00000149, 0.00001129, -0.00001523, -0.00001048, 0.00001430, 0.00001925, -0.00000733, -0.00001454, 0.00000179, -0.00000992, 0.00001244, -0.00000155, 0.00000272, 0.00000151, -0.00001355, 0.00001373, -0.00000750, 0.00000189, 0.00000306, -0.00000059, -0.00000234, -0.00000961, 0.00000436, -0.00001050, -0.00000404, 0.00001754, 0.00000879, -0.00001175, 0.00000405, -0.00000638, -0.00000880, 0.00000120, -0.00001038, 0.00001128, -0.00000109, -0.00000851, 0.00000342, -0.00001372, -0.00000559, 0.00002061, -0.00002805, -0.00002932, 0.00001224, 0.00000144, 0.00002825, 0.00001242, 0.00000382, 0.00001301, 0.00001077, 0.00000884, 0.00000200, 0.00001985, 0.00000370, 0.00002019, 0.00003042, 0.00001156, -0.00001939, -0.00003661, -0.00001083, -0.00000772, -0.00000376, -0.00000716, -0.00000586, -0.00000599, -0.00000419, -0.00000329, -0.00000855, -0.00001451, -0.00001350, -0.00002137, },
		{ 0.00000377, 0.00007012, 0.00003429, 0.00001920, 0.00002080, 0.00001657, 0.00002018, 0.00001571, 0.00002901, 0.00005357, 0.00001636, -0.00002207, -0.00002668, -0.00001431, -0.00000473, -0.00001373, -0.00000581, 0.00000706, -0.00000712, -0.00001172, 0.00000481, -0.00000605, -0.00000388, -0.00000149, -0.00000337, -0.00000322, -0.00001186, 0.00000572, 0.00000474, -0.00000644, 0.00000352, 0.00000576, -0.00000863, -0.00000313, -0.00000146, -0.00000354, 0.00000026, -0.00000261, 0.00000762, -0.00000407, -0.00000264, -0.00000067, -0.00000527, 0.00000761, 0.00000830, -0.00000315, -0.00001568, -0.00000499, 0.00000633, 0.00000903, -0.00000383, -0.00000660, 0.00000507, -0.00000384, -0.00000115, 0.00000579, -0.00000682, 0.00000480, 0.00000617, -0.00001945, -0.00000161, 0.00001597, 0.00001544, -0.00001253, -0.00001095, 0.00000158, -0.00000720, 0.00001260, -0.00000535, 0.00000735, -0.00000512, -0.00000766, 0.00001353, -0.00001212, 0.00000779, -0.00000080, 0.00000065, -0.00000419, -0.00000875, 0.00000583, -0.00001435, 0.00000114, 0.00002052, 0.00000175, -0.00000995, 0.00000530, -0.00001002, -0.00000443, -0.00000356, -0.00000569, 0.00001390, -0.00000927, -0.00000159, 0.00000082, -0.00001850, 0.00000568, 0.00001308, -0.00003681, -0.00001431, 0.00001014, 0.00000400, 0.00003311, 0.00000203, 0.00000630, 0.00001302, 0.00000917, 0.00000753, 0.00000319, 0.00001959, 0.00000182, 0.00002266, 0.00002660, 0.00000819, -0.00002173, -0.00003444, -0.00000805, -0.00000647, -0.00000446, -0.00000799, -0.00000479, -0.00000667, -0.00000257, -0.00000493, -0.00000987, -0.00001495, -0.00001328, -0.00002673, -0.00006919, },
		{ 0.00007280, 0.00002992, 0.00002089, 0.00002082, 0.00001894, 0.00002018, 0.00001676, 0.00004000, 0.00005282, 0.00000309, -0.00003165, -0.00002610, -0.00000657, -0.00000671, -0.00001583, -0.00000067, 0.00000494, -0.00000990, -0.00000712, 0.00000220, -0.00000725, -0.00000142, -0.00000371, -0.00000173, -0.00000541, -0.00001028, 0.00000926, 0.00000041, -0.00000641, 0.00000559, 0.00000356, -0.00000862, -0.00000236, -0.00000152, -0.00000287, -0.00000164, 0.00000066, 0.00000616, -0.00000666, 0.00000057, -0.00000319, -0.00000323, 0.00001009, 0.00000473, -0.00000602, -0.00001505, -0.00000172, 0.00001012, 0.00000473, -0.00000887, -0.00000056, 0.00000475, -0.00000681, 0.00000301, 0.00000269, -0.00000706, 0.00001001, -0.00000224, -0.00001831, 0.00000637, 0.00001499, 0.00001156, -0.00001657, -0.00000649, 0.00000083, -0.00000401, 0.00001049, -0.00000651, 0.00001010, -0.00001125, 0.00000058, 0.00000876, -0.00001286, 0.00001247, -0.00000584, 0.00000366, -0.00000778, -0.00000578, 0.00000543, -0.00001768, 0.00000892, 0.00001945, -0.00000437, -0.00000598, 0.00000365, -0.00001105, -0.00000189, -0.00000810, 0.00000246, 0.00001053, -0.00001374, 0.00000649, -0.00000720, -0.00001668, 0.00001456, 0.00000001, -0.00003658, -0.00000168, 0.00000435, 0.00001154, 0.00003236, -0.00000728, 0.00001172, 0.00001039, 0.00000882, 0.00000611, 0.00000459, 0.00001897, 0.00000005, 0.00002489, 0.00002251, 0.00000555, -0.00002406, -0.00003189, -0.00000501, -0.00000593, -0.00000576, -0.00000783, -0.00000426, -0.00000698, -0.00000108, -0.00000751, -0.00001022, -0.00001568, -0.00001328, -0.00003278, -0.00007087, 0.00001859, },
		{ 0.00002612, 0.00002281, 0.00002118, 0.00002142, 0.00001916, 0.00002105, 0.00005084, 0.00004698, -0.00001233, -0.00003886, -0.00002056, -0.00000087, -0.00001152, -0.00001442, 0.00000323, 0.00000105, -0.00000969, -0.00000430, -0.00000126, -0.00000611, -0.00000079, -0.00000515, -0.00000017, -0.00000815, -0.00000667, 0.00001052, -0.00000406, -0.00000468, 0.00000653, 0.00000166, -0.00000841, -0.00000178, -0.00000109, -0.00000349, -0.00000211, 0.00000384, 0.00000267, -0.00000677, 0.00000270, -0.00000602, 0.00000067, 0.00001035, 0.00000080, -0.00000764, -0.00001412, 0.00000282, 0.00001198, -0.00000211, -0.00000982, 0.00000564, 0.00000124, -0.00000701, 0.00000628, -0.00000198, -0.00000365, 0.00001162, -0.00001104, -0.00001216, 0.00001125, 0.00001280, 0.00000769, -0.00001921, -0.00000158, -0.00000038, -0.00000091, 0.00000693, -0.00000511, 0.00001005, -0.00001493, 0.00000879, 0.00000069, -0.00000892, 0.00001376, -0.00000997, 0.00000721, -0.00001258, -0.00000052, 0.00000199, -0.00001837, 0.00001734, 0.00001421, -0.00000740, -0.00000211, 0.00000034, -0.00000951, -0.00000249, -0.00000960, 0.00001090, 0.00000204, -0.00001189, 0.00001123, -0.00001720, -0.00000798, 0.00001653, -0.00001321, -0.00002837, 0.00000461, -0.00000075, 0.00002186, 0.00002527, -0.00001247, 0.00001757, 0.00000579, 0.00001022, 0.00000378, 0.00000686, 0.00001750, -0.00000143, 0.00002702, 0.00001817, 0.00000355, -0.00002650, -0.00002861, -0.00000198, -0.00000645, -0.00000699, -0.00000690, -0.00000444, -0.00000652, -0.00000037, -0.00001042, -0.00000963, -0.00001710, -0.00001311, -0.00003975, -0.00007058, 0.00003454, 0.00007300, },
		{ 0.00002483, 0.00002201, 0.00002344, 0.00001797, 0.00002892, 0.00005961, 0.00003549, -0.00002848, -0.00004149, -0.00001173, 0.00000070, -0.00001650, -0.00001005, 0.00000432, -0.00000245, -0.00000755, -0.00000403, -0.00000376, -0.00000395, -0.00000180, -0.00000523, 0.00000041, -0.00001028, -0.00000197, 0.00000917, -0.00000737, -0.00000219, 0.00000649, 0.00000027, -0.00000836, -0.00000085, -0.00000091, -0.00000484, -0.00000077, 0.00000552, -0.00000144, -0.00000463, 0.00000281, -0.00000775, 0.00000535, 0.00000813, -0.00000237, -0.00000854, -0.00001252, 0.00000822, 0.00001060, -0.00000936, -0.00000643, 0.00000961, -0.00000370, -0.00000429, 0.00000697, -0.00000598, 0.00000211, 0.00000848, -0.00001700, -0.00000311, 0.00001215, 0.00001084, 0.00000351, -0.00002004, 0.00000318, -0.00000193, 0.00000167, 0.00000289, -0.00000191, 0.00000722, -0.00001495, 0.00001450, -0.00000823, -0.00000117, 0.00001060, -0.00001122, 0.00000946, -0.00001710, 0.00000607, -0.00000457, -0.00001469, 0.00002358, 0.00000645, -0.00000660, -0.00000040, -0.00000257, -0.00000692, -0.00000579, -0.00000636, 0.00001580, -0.00000816, -0.00000398, 0.00000930, -0.00002416, 0.00000433, 0.00001003, -0.00002093, -0.00001652, 0.00000380, -0.00000136, 0.00003115, 0.00001351, -0.00001202, 0.00002123, 0.00000093, 0.00001279, 0.00000040, 0.00001045, 0.00001443, -0.00000203, 0.00002886, 0.00001366, 0.00000209, -0.00002910, -0.00002426, 0.00000043, -0.00000803, -0.00000751, -0.00000576, -0.00000513, -0.00000524, -0.00000097, -0.00001285, -0.00000859, -0.00001925, -0.00001247, -0.00004801, -0.00006770, 0.00005105, 0.00007060, 0.00002333, },
		{ 0.00002337, 0.00002457, 0.00001780, 0.00004011, 0.00006410, 0.00001847, -0.00004297, -0.00003836, -0.00000249, -0.00000209, -0.00001918, -0.00000462, 0.00000257, -0.00000396, -0.00000531, -0.00000572, -0.00000434, -0.00000221, -0.00000357, -0.00000416, -0.00000027, -0.00001089, 0.00000244, 0.00000582, -0.00000886, 0.00000017, 0.00000599, -0.00000083, -0.00000841, 0.00000059, -0.00000166, -0.00000592, 0.00000186, 0.00000497, -0.00000461, -0.00000136, 0.00000090, -0.00000733, 0.00000928, 0.00000401, -0.00000411, -0.00000926, -0.00000966, 0.00001333, 0.00000560, -0.00001456, -0.00000001, 0.00000995, -0.00000781, 0.00000011, 0.00000462, -0.00000728, 0.00000774, 0.00000140, -0.00001792, 0.00000583, 0.00000975, 0.00000987, -0.00000122, -0.00001865, 0.00000702, -0.00000357, 0.00000346, -0.00000070, 0.00000196, 0.00000252, -0.00001126, 0.00001595, -0.00001514, 0.00000813, 0.00000350, -0.00000857, 0.00000874, -0.00001943, 0.00001201, -0.00001285, -0.00000633, 0.00002511, -0.00000092, -0.00000307, -0.00000162, -0.00000327, -0.00000557, -0.00000956, 0.00000106, 0.00001458, -0.00001554, 0.00000667, 0.00000038, -0.00002387, 0.00001471, -0.00000250, -0.00001985, -0.00000680, -0.00000141, 0.00000388, 0.00003569, 0.00000055, -0.00000659, 0.00002114, -0.00000226, 0.00001517, -0.00000328, 0.00001513, 0.00000935, -0.00000092, 0.00002988, 0.00000931, 0.00000090, -0.00003164, -0.00001869, 0.00000147, -0.00001012, -0.00000701, -0.00000502, -0.00000582, -0.00000352, -0.00000299, -0.00001407, -0.00000792, -0.00002162, -0.00001146, -0.00005783, -0.00006139, 0.00006721, 0.00006574, 0.00002196, 0.00002679, },
		{ 0.00002458, 0.00002000, 0.00005352, 0.00006197, -0.00000276, -0.00005296, -0.00003005, 0.00000421, -0.00000770, -0.00001840, -0.00000033, -0.00000070, -0.00000307, -0.00000456, -0.00000792, -0.00000317, -0.00000175, -0.00000506, -0.00000268, -0.00000182, -0.00000972, 0.00000537, 0.00000162, -0.00000862, 0.00000186, 0.00000551, -0.00000203, -0.00000816, 0.00000220, -0.00000351, -0.00000576, 0.00000460, 0.00000245, -0.00000575, 0.00000156, -0.00000211, -0.00000458, 0.00001112, -0.00000080, -0.00000445, -0.00000996, -0.00000516, 0.00001658, -0.00000215, -0.00001589, 0.00000713, 0.00000678, -0.00000938, 0.00000415, 0.00000029, -0.00000504, 0.00001060, -0.00000715, -0.00001342, 0.00001191, 0.00000579, 0.00000978, -0.00000634, -0.00001495, 0.00000923, -0.00000494, 0.00000430, -0.00000321, 0.00000536, -0.00000263, -0.00000498, 0.00001273, -0.00001776, 0.00001601, -0.00000556, -0.00000244, 0.00000438, -0.00001806, 0.00001505, -0.00002027, 0.00000517, 0.00002085, -0.00000494, 0.00000066, -0.00000477, -0.00000136, -0.00000718, -0.00001080, 0.00000978, 0.00000726, -0.00001652, 0.00001511, -0.00001238, -0.00001517, 0.00001781, -0.00001543, -0.00001071, -0.00000363, -0.00000630, 0.00001325, 0.00003369, -0.00000991, 0.00000144, 0.00001741, -0.00000267, 0.00001594, -0.00000598, 0.00001989, 0.00000253, 0.00000242, 0.00002937, 0.00000563, -0.00000047, -0.00003361, -0.00001214, 0.00000052, -0.00001181, -0.00000575, -0.00000499, -0.00000596, -0.00000204, -0.00000601, -0.00001379, -0.00000843, -0.00002337, -0.00001074, -0.00006904, -0.00005089, 0.00008180, 0.00005898, 0.00002220, 0.00002852, 0.00002521, },
		{ 0.00002599, 0.00006694, 0.00005139, -0.00002537, -0.00005622, -0.00001884, 0.00000664, -0.00001365, -0.00001459, 0.00000138, -0.00000363, -0.00000067, -0.00000595, -0.00000916, -0.00000123, -0.00000263, -0.00000566, -0.00000152, -0.00000340, -0.00000729, 0.00000625, -0.00000223, -0.00000725, 0.00000287, 0.00000520, -0.00000356, -0.00000717, 0.00000330, -0.00000599, -0.00000395, 0.00000620, -0.00000099, -0.00000474, 0.00000298, -0.00000484, -0.00000025, 0.00001020, -0.00000498, -0.00000392, -0.00001035, 0.00000072, 0.00001656, -0.00001071, -0.00001279, 0.00001255, 0.00000155, -0.00000791, 0.00000604, -0.00000401, -0.00000007, 0.00000914, -0.00001402, -0.00000505, 0.00001365, 0.00000223, 0.00000976, -0.00001114, -0.00000938, 0.00000939, -0.00000565, 0.00000418, -0.00000429, 0.00000747, -0.00000676, 0.00000202, 0.00000589, -0.00001518, 0.00001989, -0.00001380, 0.00000538, -0.00000296, -0.00001262, 0.00001358, -0.00002395, 0.00001668, 0.00001190, -0.00000396, 0.00000186, -0.00000753, 0.00000180, -0.00001178, -0.00000727, 0.00001578, -0.00000322, -0.00001036, 0.00001696, -0.00002353, -0.00000087, 0.00001114, -0.00002238, 0.00000188, -0.00000805, -0.00000649, 0.00002286, 0.00002599, -0.00001546, 0.00000914, 0.00001154, -0.00000040, 0.00001439, -0.00000653, 0.00002323, -0.00000494, 0.00000788, 0.00002691, 0.00000310, -0.00000255, -0.00003422, -0.00000545, -0.00000251, -0.00001219, -0.00000446, -0.00000548, -0.00000534, -0.00000149, -0.00000923, -0.00001231, -0.00001046, -0.00002362, -0.00001143, -0.00008080, -0.00003591, 0.00009359, 0.00005124, 0.00002394, 0.00003002, 0.00002732, 0.00002360, },
		{ 0.00007699, 0.00003192, -0.00004541, -0.00005219, -0.00000784, 0.00000480, -0.00001772, -0.00000940, 0.00000038, -0.00000476, 0.00000169, -0.00000899, -0.00000862, 0.00000030, -0.00000424, -0.00000532, -0.00000106, -0.00000426, -0.00000458, 0.00000528, -0.00000495, -0.00000549, 0.00000351, 0.00000487, -0.00000527, -0.00000531, 0.00000325, -0.00000820, -0.00000086, 0.00000587, -0.00000396, -0.00000229, 0.00000249, -0.00000606, 0.00000425, 0.00000675, -0.00000757, -0.00000318, -0.00000982, 0.00000700, 0.00001264, -0.00001770, -0.00000617, 0.00001458, -0.00000359, -0.00000428, 0.00000498, -0.00000619, 0.00000549, 0.00000355, -0.00001662, 0.00000435, 0.00001121, 0.00000043, 0.00000886, -0.00001458, -0.00000297, 0.00000759, -0.00000550, 0.00000316, -0.00000394, 0.00000796, -0.00000887, 0.00000775, -0.00000244, -0.00000813, 0.00001842, -0.00001862, 0.00001238, -0.00001136, -0.00000422, 0.00000750, -0.00002189, 0.00002455, 0.00000131, 0.00000148, -0.00000106, -0.00000743, 0.00000361, -0.00001729, 0.00000109, 0.00001575, -0.00001213, 0.00000030, 0.00001070, -0.00002779, 0.00001317, -0.00000340, -0.00001917, 0.00001146, -0.00001715, -0.00000012, 0.00002856, 0.00001565, -0.00001587, 0.00001434, 0.00000559, 0.00000335, 0.00001087, -0.00000450, 0.00002386, -0.00001149, 0.00001460, 0.00002258, 0.00000192, -0.00000558, -0.00003276, 0.00000012, -0.00000695, -0.00001080, -0.00000397, -0.00000586, -0.00000427, -0.00000216, -0.00001176, -0.00001044, -0.00001371, -0.00002198, -0.00001471, -0.00009168, -0.00001686, 0.00010167, 0.00004358, 0.00002678, 0.00003148, 0.00002922, 0.00002235, 0.00003672, },
		{ 0.00000533, -0.00005882, -0.00004245, 0.00000022, 0.00000025, -0.00001887, -0.00000472, -0.00000230, -0.00000366, 0.00000259, -0.00001245, -0.00000640, 0.00000063, -0.00000575, -0.00000447, -0.00000115, -0.00000411, -0.00000255, 0.00000322, -0.00000635, -0.00000384, 0.00000410, 0.00000416, -0.00000672, -0.00000289, 0.00000179, -0.00000917, 0.00000251, 0.00000366, -0.00000537, 0.00000033, 0.00000060, -0.00000517, 0.00000745, 0.00000176, -0.00000827, -0.00000269, -0.00000782, 0.00001218, 0.00000538, -0.00002109, 0.00000190, 0.00001291, -0.00000675, -0.00000031, 0.00000164, -0.00000509, 0.00000910, -0.00000424, -0.00001395, 0.00001179, 0.00000615, 0.00000068, 0.00000656, -0.00001575, 0.00000298, 0.00000445, -0.00000455, 0.00000146, -0.00000241, 0.00000696, -0.00000862, 0.00001071, -0.00000975, 0.00000128, 0.00001192, -0.00001849, 0.00001621, -0.00001836, 0.00000486, -0.00000162, -0.00001392, 0.00002600, -0.00000696, 0.00000877, -0.00000762, -0.00000324, 0.00000155, -0.00002031, 0.00001166, 0.00000889, -0.00001513, 0.00001039, -0.00000133, -0.00002267, 0.00002057, -0.00002000, -0.00000595, 0.00001262, -0.00002548, 0.00001120, 0.00002799, 0.00000640, -0.00001288, 0.00001641, 0.00000109, 0.00000709, 0.00000652, -0.00000048, 0.00002131, -0.00001569, 0.00002130, 0.00001699, 0.00000180, -0.00000936, -0.00002895, 0.00000333, -0.00001148, -0.00000798, -0.00000471, -0.00000541, -0.00000350, -0.00000374, -0.00001317, -0.00000908, -0.00001734, -0.00001876, -0.00002133, -0.00010004, 0.00000519, 0.00010575, 0.00003683, 0.00003027, 0.00003318, 0.00003024, 0.00002235, 0.00005203, 0.00007965, },
		{ -0.00006291, -0.00003022, 0.00000408, -0.00000485, -0.00001742, -0.00000180, -0.00000517, -0.00000095, 0.00000137, -0.00001501, -0.00000332, -0.00000038, -0.00000659, -0.00000365, -0.00000129, -0.00000323, -0.00000173, 0.00000095, -0.00000672, -0.00000242, 0.00000474, 0.00000289, -0.00000741, -0.00000060, -0.00000072, -0.00000840, 0.00000496, 0.00000042, -0.00000487, 0.00000199, -0.00000158, -0.00000251, 0.00000834, -0.00000342, -0.00000735, -0.00000246, -0.00000427, 0.00001474, -0.00000359, -0.00001997, 0.00000899, 0.00000863, -0.00000710, 0.00000217, -0.00000225, -0.00000109, 0.00000890, -0.00001132, -0.00000693, 0.00001525, 0.00000064, 0.00000230, 0.00000311, -0.00001432, 0.00000724, 0.00000091, -0.00000318, -0.00000055, -0.00000009, 0.00000493, -0.00000642, 0.00001033, -0.00001395, 0.00001024, 0.00000226, -0.00001347, 0.00001554, -0.00002183, 0.00001199, -0.00001091, -0.00000203, 0.00002033, -0.00000962, 0.00001417, -0.00001531, 0.00000416, -0.00000531, -0.00001767, 0.00002001, -0.00000232, -0.00001046, 0.00001480, -0.00001378, -0.00000996, 0.00001743, -0.00003132, 0.00001253, 0.00000365, -0.00002760, 0.00002313, 0.00002168, 0.00000085, -0.00000907, 0.00001626, -0.00000158, 0.00000980, 0.00000261, 0.00000424, 0.00001606, -0.00001678, 0.00002677, 0.00001100, 0.00000216, -0.00001313, -0.00002331, 0.00000345, -0.00001459, -0.00000477, -0.00000644, -0.00000381, -0.00000384, -0.00000540, -0.00001367, -0.00000870, -0.00002037, -0.00001495, -0.00003126, -0.00010455, 0.00002882, 0.00010613, 0.00003141, 0.00003423, 0.00003513, 0.00002989, 0.00002579, 0.00006989, 0.00007135, -0.00002423, },
		{ -0.00001922, 0.00000430, -0.00000879, -0.00001456, -0.00000088, -0.00000702, 0.00000218, -0.00000171, -0.00001582, -0.00000042, -0.00000231, -0.00000661, -0.00000315, -0.00000100, -0.00000224, -0.00000209, -0.00000086, -0.00000651, -0.00000112, 0.00000522, 0.00000115, -0.00000708, 0.00000083, -0.00000344, -0.00000609, 0.00000565, -0.00000264, -0.00000295, 0.00000213, -0.00000289, 0.00000091, 0.00000669, -0.00000759, -0.00000545, -0.00000213, 0.00000023, 0.00001376, -0.00001199, -0.00001484, 0.00001321, 0.00000365, -0.00000511, 0.00000211, -0.00000465, 0.00000406, 0.00000461, -0.00001498, 0.00000202, 0.00001428, -0.00000350, 0.00000409, -0.00000059, -0.00001072, 0.00000914, -0.00000208, -0.00000197, -0.00000234, 0.00000248, 0.00000252, -0.00000317, 0.00000711, -0.00001404, 0.00001620, -0.00000780, -0.00000509, 0.00001045, -0.00002075, 0.00001513, -0.00001726, 0.00001024, 0.00000935, -0.00000538, 0.00001443, -0.00002058, 0.00001189, -0.00001545, -0.00000821, 0.00002182, -0.00001288, -0.00000011, 0.00001090, -0.00002067, 0.00000478, 0.00000445, -0.00003174, 0.00002888, -0.00001260, -0.00002072, 0.00003071, 0.00001275, -0.00000060, -0.00000642, 0.00001542, -0.00000308, 0.00001145, -0.00000015, 0.00000831, 0.00000937, -0.00001488, 0.00003028, 0.00000533, 0.00000243, -0.00001592, -0.00001708, 0.00000074, -0.00001521, -0.00000254, -0.00000820, -0.00000148, -0.00000561, -0.00000624, -0.00001402, -0.00000911, -0.00002223, -0.00001173, -0.00004376, -0.00010450, 0.00005275, 0.00010336, 0.00002738, 0.00003881, 0.00003682, 0.00002842, 0.00003477, 0.00008619, 0.00005045, -0.00005091, -0.00005752, },
		{ 0.00000274, -0.00001101, -0.00001158, -0.00000134, -0.00000739, 0.00000459, -0.00000574, -0.00001479, 0.00000150, -0.00000448, -0.00000604, -0.00000292, -0.00000009, -0.00000183, -0.00000316, -0.00000198, -0.00000604, 0.00000024, 0.00000526, -0.00000064, -0.00000591, 0.00000099, -0.00000542, -0.00000308, 0.00000445, -0.00000442, -0.00000065, 0.00000100, -0.00000261, 0.00000376, 0.00000305, -0.00001003, -0.00000320, -0.00000121, 0.00000453, 0.00000931, -0.00001771, -0.00000741, 0.00001388, -0.00000014, -0.00000228, -0.00000022, -0.00000419, 0.00000796, -0.00000232, -0.00001378, 0.00000998, 0.00000997, -0.00000538, 0.00000504, -0.00000341, -0.00000607, 0.00000874, -0.00000393, -0.00000142, -0.00000340, 0.00000480, 0.00000030, 0.00000004, 0.00000232, -0.00001029, 0.00001759, -0.00001551, 0.00000415, 0.00000244, -0.00001554, 0.00001363, -0.00001848, 0.00001898, -0.00000315, 0.00000436, 0.00000823, -0.00002041, 0.00001622, -0.00002515, 0.00000616, 0.00001499, -0.00001747, 0.00001082, 0.00000014, -0.00001834, 0.00001457, -0.00001310, -0.00002007, 0.00003621, -0.00002988, -0.00000608, 0.00003070, 0.00000529, 0.00000031, -0.00000538, 0.00001496, -0.00000438, 0.00001272, -0.00000200, 0.00001102, 0.00000254, -0.00001074, 0.00003172, 0.00000024, 0.00000246, -0.00001706, -0.00001184, -0.00000355, -0.00001328, -0.00000228, -0.00000875, 0.00000049, -0.00000834, -0.00000589, -0.00001506, -0.00000950, -0.00002312, -0.00000995, -0.00005773, -0.00009983, 0.00007614, 0.00009787, 0.00002483, 0.00004429, 0.00003724, 0.00002736, 0.00005025, 0.00009535, 0.00001853, -0.00006886, -0.00004532, -0.00001218, },
		{ -0.00001208, -0.00000915, -0.00000230, -0.00000656, 0.00000566, -0.00000971, -0.00001239, 0.00000210, -0.00000632, -0.00000522, -0.00000265, 0.00000123, -0.00000239, -0.00000433, -0.00000249, -0.00000542, 0.00000162, 0.00000466, -0.00000197, -0.00000444, 0.00000002, -0.00000603, -0.00000046, 0.00000200, -0.00000447, 0.00000097, -0.00000056, -0.00000078, 0.00000494, -0.00000147, -0.00001060, -0.00000096, 0.00000051, 0.00000732, 0.00000255, -0.00001946, 0.00000009, 0.00001163, -0.00000163, -0.00000033, -0.00000341, -0.00000087, 0.00000862, -0.00000932, -0.00000811, 0.00001465, 0.00000430, -0.00000526, 0.00000480, -0.00000449, -0.00000172, 0.00000676, -0.00000459, -0.00000173, -0.00000336, 0.00000642, -0.00000131, 0.00000231, -0.00000243, -0.00000413, 0.00001431, -0.00001895, 0.00001165, -0.00000624, -0.00000785, 0.00000843, -0.00001415, 0.00002137, -0.00001271, 0.00001599, -0.00000313, -0.00001381, 0.00001426, -0.00002989, 0.00002081, 0.00000110, -0.00001295, 0.00001631, -0.00001225, -0.00000754, 0.00001437, -0.00002745, -0.00000027, 0.00003125, -0.00004128, 0.00001159, 0.00002316, 0.00000240, 0.00000083, -0.00000485, 0.00001482, -0.00000590, 0.00001439, -0.00000381, 0.00001255, -0.00000359, -0.00000521, 0.00003133, -0.00000436, 0.00000268, -0.00001663, -0.00000878, -0.00000763, -0.00000978, -0.00000404, -0.00000728, 0.00000079, -0.00001082, -0.00000481, -0.00001710, -0.00000901, -0.00002387, -0.00000971, -0.00007233, -0.00009076, 0.00009854, 0.00008974, 0.00002423, 0.00005054, 0.00003536, 0.00002951, 0.00007075, 0.00009184, -0.00001926, -0.00007419, -0.00003102, -0.00000972, 0.00000140, },
		{ -0.00000714, -0.00000322, -0.00000516, 0.00000540, -0.00001287, -0.00000937, 0.00000149, -0.00000754, -0.00000440, -0.00000206, 0.00000248, -0.00000389, -0.00000512, -0.00000265, -0.00000459, 0.00000282, 0.00000351, -0.00000252, -0.00000334, -0.00000147, -0.00000523, 0.00000086, -0.00000061, -0.00000308, 0.00000127, -0.00000148, 0.00000183, 0.00000399, -0.00000573, -0.00000962, 0.00000123, 0.00000279, 0.00000761, -0.00000465, -0.00001724, 0.00000565, 0.00000804, -0.00000095, -0.00000043, -0.00000558, 0.00000395, 0.00000535, -0.00001366, 0.00000007, 0.00001513, -0.00000075, -0.00000413, 0.00000383, -0.00000371, 0.00000124, 0.00000424, -0.00000450, -0.00000272, -0.00000213, 0.00000709, -0.00000217, 0.00000319, -0.00000570, 0.00000242, 0.00000757, -0.00001755, 0.00001553, -0.00001327, 0.00000010, 0.00000163, -0.00000581, 0.00001671, -0.00001585, 0.00002501, -0.00001615, -0.00000240, 0.00000548, -0.00002637, 0.00002999, -0.00001482, -0.00000021, 0.00001233, -0.00001952, 0.00000652, 0.00000366, -0.00003187, 0.00002004, 0.00001579, -0.00004234, 0.00002618, 0.00001131, 0.00000468, -0.00000117, -0.00000309, 0.00001390, -0.00000706, 0.00001648, -0.00000635, 0.00001368, -0.00000895, 0.00000123, 0.00002921, -0.00000853, 0.00000379, -0.00001558, -0.00000813, -0.00000997, -0.00000635, -0.00000688, -0.00000395, -0.00000138, -0.00001165, -0.00000426, -0.00001952, -0.00000739, -0.00002542, -0.00001048, -0.00008730, -0.00007721, 0.00011948, 0.00007886, 0.00002659, 0.00005651, 0.00003110, 0.00003811, 0.00009176, 0.00007234, -0.00005521, -0.00006659, -0.00001956, -0.00001033, 0.00000140, -0.00001301, },
		{ -0.00000408, -0.00000367, 0.00000418, -0.00001496, -0.00000636, 0.00000002, -0.00000807, -0.00000362, -0.00000101, 0.00000315, -0.00000599, -0.00000531, -0.00000261, -0.00000350, 0.00000353, 0.00000217, -0.00000236, -0.00000304, -0.00000269, -0.00000359, 0.00000059, -0.00000233, -0.00000114, 0.00000035, -0.00000103, 0.00000410, 0.00000115, -0.00000882, -0.00000759, 0.00000351, 0.00000491, 0.00000520, -0.00001036, -0.00001223, 0.00000825, 0.00000488, 0.00000078, -0.00000262, -0.00000528, 0.00000809, -0.00000087, -0.00001362, 0.00000802, 0.00001205, -0.00000400, -0.00000306, 0.00000302, -0.00000170, 0.00000237, 0.00000212, -0.00000433, -0.00000385, -0.00000001, 0.00000678, -0.00000236, 0.00000272, -0.00000671, 0.00000739, -0.00000050, -0.00001215, 0.00001506, -0.00001702, 0.00000628, -0.00000425, 0.00000360, 0.00000670, -0.00001132, 0.00002774, -0.00002644, 0.00001008, -0.00000776, -0.00001422, 0.00002934, -0.00002610, 0.00001573, -0.00000085, -0.00001665, 0.00001627, -0.00001286, -0.00002389, 0.00003302, -0.00000404, -0.00003298, 0.00003304, -0.00000001, 0.00001018, -0.00000601, 0.00000100, 0.00001087, -0.00000651, 0.00001801, -0.00000957, 0.00001508, -0.00001387, 0.00000852, 0.00002513, -0.00001187, 0.00000624, -0.00001531, -0.00000897, -0.00001002, -0.00000445, -0.00000921, 0.00000009, -0.00000578, -0.00000998, -0.00000566, -0.00002087, -0.00000542, -0.00002807, -0.00001165, -0.00010289, -0.00005853, 0.00013790, 0.00006553, 0.00003303, 0.00006018, 0.00002618, 0.00005524, 0.00010631, 0.00003766, -0.00008119, -0.00004963, -0.00001424, -0.00001124, 0.00000241, -0.00001449, -0.00000499, },
		{ -0.00000219, 0.00000241, -0.00001601, -0.00000378, -0.00000194, -0.00000796, -0.00000283, 0.00000033, 0.00000293, -0.00000823, -0.00000492, -0.00000243, -0.00000225, 0.00000359, 0.00000108, -0.00000187, -0.00000356, -0.00000301, -0.00000203, -0.00000089, -0.00000258, 0.00000035, -0.00000108, 0.00000091, 0.00000495, -0.00000275, -0.00001034, -0.00000494, 0.00000595, 0.00000596, 0.00000077, -0.00001318, -0.00000630, 0.00000806, 0.00000337, 0.00000202, -0.00000582, -0.00000217, 0.00000951, -0.00000780, -0.00000909, 0.00001332, 0.00000713, -0.00000534, -0.00000265, 0.00000327, 0.00000044, 0.00000198, 0.00000082, -0.00000461, -0.00000446, 0.00000247, 0.00000569, -0.00000213, 0.00000142, -0.00000550, 0.00000947, -0.00000758, -0.00000462, 0.00001081, -0.00001692, 0.00000952, -0.00000731, 0.00001083, -0.00000517, -0.00000055, 0.00002267, -0.00003043, 0.00001921, -0.00002089, 0.00000342, 0.00001793, -0.00002723, 0.00002761, -0.00001822, -0.00000310, 0.00001535, -0.00002707, -0.00000667, 0.00003407, -0.00002066, -0.00001735, 0.00003089, -0.00000665, 0.00001550, -0.00001213, 0.00000706, 0.00000527, -0.00000302, 0.00001730, -0.00001239, 0.00001667, -0.00001850, 0.00001671, 0.00001856, -0.00001333, 0.00000956, -0.00001693, -0.00000966, -0.00000847, -0.00000462, -0.00000969, 0.00000307, -0.00001103, -0.00000612, -0.00000978, -0.00001957, -0.00000481, -0.00003099, -0.00001327, -0.00011920, -0.00003372, 0.00015193, 0.00005111, 0.00004394, 0.00005929, 0.00002438, 0.00008004, 0.00010691, -0.00000659, -0.00009150, -0.00002964, -0.00001547, -0.00000992, 0.00000299, -0.00001628, -0.00000231, -0.00000526, },
		{ 0.00000025, -0.00001619, -0.00000186, -0.00000408, -0.00000724, -0.00000197, 0.00000171, 0.00000177, -0.00001021, -0.00000406, -0.00000209, -0.00000109, 0.00000309, 0.00000055, -0.00000158, -0.00000449, -0.00000230, -0.00000134, -0.00000271, -0.00000147, 0.00000067, -0.00000203, 0.00000373, 0.00000380, -0.00000666, -0.00001030, -0.00000192, 0.00000833, 0.00000519, -0.00000432, -0.00001276, -0.00000131, 0.00000624, 0.00000373, 0.00000155, -0.00000833, 0.00000273, 0.00000721, -0.00001282, -0.00000162, 0.00001473, 0.00000231, -0.00000551, -0.00000271, 0.00000490, 0.00000162, 0.00000088, 0.00000022, -0.00000552, -0.00000407, 0.00000466, 0.00000418, -0.00000179, 0.00000000, -0.00000282, 0.00000834, -0.00001188, 0.00000279, 0.00000431, -0.00001352, 0.00000972, -0.00000684, 0.00001350, -0.00001485, 0.00001292, 0.00001100, -0.00002678, 0.00002172, -0.00002895, 0.00002083, -0.00000091, -0.00001647, 0.00002902, -0.00003186, 0.00001625, 0.00000191, -0.00003109, 0.00001214, 0.00002404, -0.00002848, -0.00000176, 0.00002223, -0.00000702, 0.00001766, -0.00001701, 0.00001331, -0.00000192, 0.00000355, 0.00001301, -0.00001308, 0.00001741, -0.00002227, 0.00002533, 0.00000926, -0.00001154, 0.00001229, -0.00002045, -0.00000871, -0.00000686, -0.00000617, -0.00000795, 0.00000358, -0.00001520, -0.00000161, -0.00001608, -0.00001482, -0.00000741, -0.00003239, -0.00001661, -0.00013541, -0.00000217, 0.00015929, 0.00003820, 0.00005805, 0.00005266, 0.00003061, 0.00010758, 0.00008834, -0.00005100, -0.00008491, -0.00001335, -0.00002077, -0.00000539, 0.00000157, -0.00001734, 0.00000064, -0.00000703, -0.00000052, },
		{ -0.00001559, -0.00000073, -0.00000607, -0.00000596, -0.00000113, 0.00000286, -0.00000018, -0.00001167, -0.00000281, -0.00000162, -0.00000026, 0.00000232, 0.00000056, -0.00000186, -0.00000520, -0.00000094, -0.00000191, -0.00000390, 0.00000031, -0.00000026, -0.00000168, 0.00000638, 0.00000071, -0.00000966, -0.00000903, 0.00000134, 0.00001013, 0.00000242, -0.00000848, -0.00000980, 0.00000159, 0.00000430, 0.00000516, -0.00000090, -0.00000865, 0.00000750, 0.00000171, -0.00001396, 0.00000631, 0.00001250, -0.00000103, -0.00000555, -0.00000250, 0.00000751, 0.00000126, -0.00000003, -0.00000020, -0.00000674, -0.00000259, 0.00000605, 0.00000262, -0.00000157, -0.00000087, 0.00000015, 0.00000471, -0.00001257, 0.00000812, -0.00000248, -0.00000807, 0.00000772, -0.00000350, 0.00001086, -0.00001921, 0.00002477, -0.00000388, -0.00001677, 0.00001681, -0.00002860, 0.00003185, -0.00002058, 0.00000305, 0.00001744, -0.00003432, 0.00003277, -0.00001987, -0.00002113, 0.00002377, 0.00000877, -0.00002638, 0.00000838, 0.00001177, -0.00000246, 0.00001555, -0.00001868, 0.00001750, -0.00000855, 0.00001179, 0.00000499, -0.00001025, 0.00001580, -0.00002387, 0.00003307, -0.00000208, -0.00000562, 0.00001235, -0.00002452, -0.00000570, -0.00000669, -0.00000758, -0.00000488, 0.00000130, -0.00001673, 0.00000147, -0.00002271, -0.00000733, -0.00001421, -0.00003030, -0.00002394, -0.00014921, 0.00003542, 0.00015820, 0.00003010, 0.00007221, 0.00004165, 0.00004881, 0.00012923, 0.00005038, -0.00008519, -0.00006536, -0.00000539, -0.00002608, 0.00000130, -0.00000252, -0.00001646, 0.00000299, -0.00000920, 0.00000153, -0.00000230, },
		{ -0.00000051, -0.00000758, -0.00000424, -0.00000045, 0.00000364, -0.00000273, -0.00001246, -0.00000121, -0.00000120, 0.00000010, 0.00000166, 0.00000082, -0.00000276, -0.00000519, 0.00000038, -0.00000351, -0.00000387, 0.00000184, -0.00000190, 0.00000025, 0.00000776, -0.00000366, -0.00001117, -0.00000692, 0.00000473, 0.00001070, -0.00000185, -0.00001050, -0.00000571, 0.00000221, 0.00000345, 0.00000636, -0.00000455, -0.00000621, 0.00001011, -0.00000515, -0.00001070, 0.00001221, 0.00000808, -0.00000254, -0.00000618, -0.00000113, 0.00001010, -0.00000052, -0.00000016, -0.00000106, -0.00000767, -0.00000032, 0.00000644, 0.00000129, -0.00000151, -0.00000083, 0.00000220, -0.00000002, -0.00000996, 0.00001023, -0.00000779, -0.00000214, 0.00000481, 0.00000107, 0.00000397, -0.00001694, 0.00003137, -0.00001772, -0.00000377, 0.00000644, -0.00001956, 0.00003235, -0.00003374, 0.00002422, -0.00000415, -0.00002215, 0.00003766, -0.00004096, 0.00000026, 0.00002253, -0.00000360, -0.00001808, 0.00001106, 0.00000396, 0.00000378, 0.00001031, -0.00001681, 0.00001820, -0.00001231, 0.00001912, -0.00000523, -0.00000374, 0.00001079, -0.00002182, 0.00003789, -0.00001353, 0.00000404, 0.00000806, -0.00002689, -0.00000165, -0.00000844, -0.00000739, -0.00000219, -0.00000275, -0.00001529, 0.00000152, -0.00002726, 0.00000064, -0.00002449, -0.00002373, -0.00003764, -0.00015711, 0.00007644, 0.00014861, 0.00002950, 0.00008213, 0.00003085, 0.00007944, 0.00013497, -0.00000098, -0.00010126, -0.00004054, -0.00000663, -0.00002774, 0.00000767, -0.00000851, -0.00001307, 0.00000367, -0.00001106, 0.00000380, -0.00000526, -0.00001430, },
		{ -0.00000823, -0.00000237, -0.00000001, 0.00000401, -0.00000570, -0.00001239, 0.00000062, -0.00000106, 0.00000013, 0.00000134, 0.00000094, -0.00000396, -0.00000431, 0.00000097, -0.00000547, -0.00000255, 0.00000238, -0.00000335, 0.00000341, 0.00000709, -0.00000828, -0.00001106, -0.00000429, 0.00000796, 0.00000950, -0.00000654, -0.00000995, -0.00000203, 0.00000131, 0.00000410, 0.00000617, -0.00000792, -0.00000163, 0.00000931, -0.00001090, -0.00000410, 0.00001456, 0.00000337, -0.00000280, -0.00000737, 0.00000188, 0.00001152, -0.00000299, 0.00000050, -0.00000264, -0.00000772, 0.00000210, 0.00000594, 0.00000031, -0.00000152, 0.00000004, 0.00000257, -0.00000421, -0.00000528, 0.00000902, -0.00001041, 0.00000295, 0.00000220, 0.00000497, -0.00000479, -0.00000890, 0.00003095, -0.00002691, 0.00000822, -0.00000550, -0.00000484, 0.00002199, -0.00003522, 0.00003872, -0.00002809, 0.00000201, 0.00002600, -0.00005112, 0.00002469, 0.00000864, -0.00000666, -0.00001008, 0.00000838, 0.00000080, 0.00000856, 0.00000438, -0.00001284, 0.00001572, -0.00001200, 0.00002298, -0.00001494, 0.00000500, 0.00000257, -0.00001538, 0.00003786, -0.00002230, 0.00001536, -0.00000081, -0.00002548, 0.00000133, -0.00001116, -0.00000521, -0.00000145, -0.00000666, -0.00001204, -0.00000167, -0.00002786, 0.00000618, -0.00003578, -0.00001359, -0.00005893, -0.00015543, 0.00011674, 0.00013279, 0.00003696, 0.00008421, 0.00002724, 0.00011758, 0.00011696, -0.00005452, -0.00009681, -0.00001904, -0.00001413, -0.00002416, 0.00001114, -0.00001449, -0.00000775, 0.00000212, -0.00001172, 0.00000577, -0.00000840, -0.00001246, -0.00000119, },
		{ -0.00000078, 0.00000024, 0.00000392, -0.00000891, -0.00001131, 0.00000242, -0.00000138, 0.00000010, 0.00000132, 0.00000062, -0.00000498, -0.00000284, 0.00000048, -0.00000698, -0.00000044, 0.00000167, -0.00000374, 0.00000687, 0.00000418, -0.00001212, -0.00000955, -0.00000140, 0.00001060, 0.00000642, -0.00001038, -0.00000726, 0.00000013, 0.00000007, 0.00000578, 0.00000412, -0.00000956, 0.00000347, 0.00000517, -0.00001346, 0.00000367, 0.00001319, -0.00000009, -0.00000286, -0.00000843, 0.00000630, 0.00001101, -0.00000516, 0.00000145, -0.00000481, -0.00000656, 0.00000403, 0.00000493, -0.00000037, -0.00000134, 0.00000126, 0.00000114, -0.00000655, -0.00000025, 0.00000535, -0.00001001, 0.00000631, 0.00000065, 0.00000669, -0.00001265, 0.00000239, 0.00002392, -0.00002957, 0.00001600, -0.00001479, 0.00001051, 0.00000439, -0.00002416, 0.00004038, -0.00004492, 0.00002991, -0.00000046, -0.00004346, 0.00004122, -0.00001139, 0.00000112, -0.00000823, 0.00000511, 0.00000105, 0.00001067, -0.00000019, -0.00000900, 0.00001194, -0.00000834, 0.00002207, -0.00002143, 0.00001343, -0.00000718, -0.00000529, 0.00003219, -0.00002572, 0.00002501, -0.00001250, -0.00001957, 0.00000119, -0.00001284, -0.00000211, -0.00000307, -0.00000874, -0.00000900, -0.00000670, -0.00002431, 0.00000695, -0.00004476, -0.00000276, -0.00008685, -0.00014179, 0.00015183, 0.00011495, 0.00005007, 0.00007780, 0.00003767, 0.00015291, 0.00007322, -0.00009705, -0.00007594, -0.00000702, -0.00002277, -0.00001649, 0.00001039, -0.00001834, -0.00000202, -0.00000135, -0.00001063, 0.00000676, -0.00001128, -0.00001045, -0.00000256, -0.00000779, },
		{ 0.00000054, 0.00000323, -0.00001205, -0.00000909, 0.00000378, -0.00000213, 0.00000029, 0.00000134, -0.00000012, -0.00000544, -0.00000132, -0.00000098, -0.00000747, 0.00000169, 0.00000001, -0.00000258, 0.00000952, -0.00000056, -0.00001436, -0.00000706, 0.00000155, 0.00001210, 0.00000189, -0.00001232, -0.00000352, 0.00000045, -0.00000043, 0.00000745, 0.00000071, -0.00000867, 0.00000718, -0.00000092, -0.00001190, 0.00001017, 0.00000924, -0.00000166, -0.00000361, -0.00000838, 0.00001119, 0.00000845, -0.00000625, 0.00000202, -0.00000701, -0.00000433, 0.00000503, 0.00000383, -0.00000083, -0.00000079, 0.00000214, -0.00000152, -0.00000644, 0.00000352, 0.00000069, -0.00000704, 0.00000766, 0.00000029, 0.00000564, -0.00001734, 0.00001370, 0.00001254, -0.00002591, 0.00001820, -0.00001859, 0.00002134, -0.00001420, -0.00000449, 0.00002788, -0.00004730, 0.00005073, -0.00003330, -0.00001794, 0.00004114, -0.00002739, 0.00001529, -0.00001459, 0.00000589, 0.00000135, 0.00001131, -0.00000310, -0.00000676, 0.00000920, -0.00000361, 0.00001713, -0.00002329, 0.00001910, -0.00001591, 0.00000622, 0.00002192, -0.00002257, 0.00002965, -0.00002369, -0.00001049, -0.00000287, -0.00001151, -0.00000023, -0.00000587, -0.00000857, -0.00000792, -0.00001119, -0.00001827, 0.00000246, -0.00004868, 0.00000484, -0.00011813, -0.00011622, 0.00017861, 0.00009967, 0.00006397, 0.00006660, 0.00006542, 0.00017215, 0.00000977, -0.00011788, -0.00004781, -0.00000605, -0.00002775, -0.00000785, 0.00000610, -0.00001887, 0.00000232, -0.00000555, -0.00000794, 0.00000638, -0.00001342, -0.00000876, -0.00000411, -0.00000631, 0.00000014, },
		{ 0.00000170, -0.00001467, -0.00000588, 0.00000429, -0.00000300, 0.00000081, 0.00000111, -0.00000102, -0.00000524, -0.00000028, -0.00000292, -0.00000678, 0.00000315, -0.00000190, 0.00000006, 0.00001039, -0.00000619, -0.00001462, -0.00000409, 0.00000431, 0.00001206, -0.00000322, -0.00001190, -0.00000001, -0.00000061, 0.00000030, 0.00000807, -0.00000291, -0.00000551, 0.00000808, -0.00000678, -0.00000679, 0.00001365, 0.00000447, -0.00000170, -0.00000528, -0.00000645, 0.00001531, 0.00000445, -0.00000600, 0.00000166, -0.00000853, -0.00000158, 0.00000505, 0.00000295, -0.00000116, 0.00000020, 0.00000205, -0.00000437, -0.00000415, 0.00000498, -0.00000333, -0.00000252, 0.00000721, 0.00000079, 0.00000219, -0.00001776, 0.00002219, -0.00000010, -0.00001781, 0.00001550, -0.00001635, 0.00002444, -0.00002746, 0.00001667, 0.00000541, -0.00003320, 0.00005542, -0.00006035, 0.00001786, 0.00002207, -0.00003006, 0.00002736, -0.00002611, 0.00001241, -0.00000151, 0.00001317, -0.00000607, -0.00000564, 0.00000872, -0.00000045, 0.00001055, -0.00002109, 0.00002086, -0.00002145, 0.00001622, 0.00000984, -0.00001386, 0.00002744, -0.00003077, -0.00000131, -0.00000960, -0.00000653, -0.00000158, -0.00000759, -0.00000741, -0.00000911, -0.00001314, -0.00001258, -0.00000564, -0.00004675, 0.00000585, -0.00014826, -0.00008129, 0.00019663, 0.00008981, 0.00007341, 0.00005831, 0.00010727, 0.00016346, -0.00006009, -0.00011285, -0.00002314, -0.00001286, -0.00002682, -0.00000160, 0.00000053, -0.00001644, 0.00000415, -0.00000903, -0.00000448, 0.00000473, -0.00001463, -0.00000780, -0.00000516, -0.00000425, 0.00000027, 0.00000110, },
		{ -0.00001617, -0.00000210, 0.00000371, -0.00000357, 0.00000149, 0.00000047, -0.00000172, -0.00000460, -0.00000000, -0.00000472, -0.00000518, 0.00000356, -0.00000326, 0.00000358, 0.00000894, -0.00001160, -0.00001292, -0.00000112, 0.00000660, 0.00001034, -0.00000783, -0.00000937, 0.00000225, -0.00000212, 0.00000203, 0.00000710, -0.00000538, -0.00000131, 0.00000586, -0.00001033, 0.00000012, 0.00001356, 0.00000055, -0.00000118, -0.00000739, -0.00000253, 0.00001754, 0.00000004, -0.00000470, 0.00000028, -0.00000886, 0.00000099, 0.00000435, 0.00000240, -0.00000133, 0.00000141, 0.00000067, -0.00000627, -0.00000069, 0.00000396, -0.00000545, 0.00000223, 0.00000548, 0.00000156, -0.00000258, -0.00001412, 0.00002612, -0.00001106, -0.00000786, 0.00000991, -0.00000978, 0.00001960, -0.00003138, 0.00003190, -0.00001893, -0.00000702, 0.00004069, -0.00007050, 0.00005124, -0.00001041, -0.00001533, 0.00002871, -0.00003589, 0.00002216, -0.00000822, 0.00001833, -0.00001170, -0.00000331, 0.00000958, -0.00000018, 0.00000502, -0.00001693, 0.00001941, -0.00002311, 0.00002250, -0.00000058, -0.00000271, 0.00001900, -0.00003147, 0.00000449, -0.00001607, 0.00000070, -0.00000675, -0.00000617, -0.00000765, -0.00001101, -0.00001223, -0.00000967, -0.00001433, -0.00004051, -0.00000110, -0.00017318, -0.00004090, 0.00020816, 0.00008498, 0.00007563, 0.00006199, 0.00015260, 0.00012140, -0.00011944, -0.00008608, -0.00001005, -0.00002120, -0.00002113, 0.00000045, -0.00000369, -0.00001274, 0.00000359, -0.00001080, -0.00000133, 0.00000241, -0.00001515, -0.00000759, -0.00000517, -0.00000231, -0.00000015, 0.00000191, -0.00000080, },
		{ 0.00000154, 0.00000214, -0.00000348, 0.00000197, -0.00000048, -0.00000196, -0.00000391, -0.00000043, -0.00000590, -0.00000323, 0.00000298, -0.00000345, 0.00000703, 0.00000520, -0.00001570, -0.00000967, 0.00000145, 0.00000813, 0.00000719, -0.00001097, -0.00000554, 0.00000282, -0.00000316, 0.00000398, 0.00000477, -0.00000585, 0.00000227, 0.00000144, -0.00001039, 0.00000660, 0.00001062, -0.00000163, -0.00000113, -0.00000897, 0.00000275, 0.00001737, -0.00000372, -0.00000305, -0.00000177, -0.00000786, 0.00000280, 0.00000341, 0.00000212, -0.00000124, 0.00000242, -0.00000178, -0.00000641, 0.00000259, 0.00000113, -0.00000504, 0.00000602, 0.00000316, 0.00000206, -0.00000733, -0.00000768, 0.00002517, -0.00001835, 0.00000153, 0.00000377, -0.00000184, 0.00000947, -0.00002566, 0.00003638, -0.00003642, 0.00002199, 0.00001064, -0.00005833, 0.00006924, -0.00004445, 0.00001327, 0.00001484, -0.00003657, 0.00002946, -0.00001599, 0.00002631, -0.00002147, 0.00000312, 0.00000890, -0.00000182, 0.00000196, -0.00001319, 0.00001681, -0.00002212, 0.00002471, -0.00000682, 0.00000691, 0.00000741, -0.00002607, 0.00000470, -0.00001895, 0.00000707, -0.00001411, -0.00000111, -0.00001136, -0.00001088, -0.00001028, -0.00001019, -0.00002086, -0.00003305, -0.00001486, -0.00019089, 0.00000166, 0.00021667, 0.00008147, 0.00007276, 0.00008382, 0.00018565, 0.00005033, -0.00015331, -0.00004864, -0.00001100, -0.00002499, -0.00001445, -0.00000102, -0.00000505, -0.00000986, 0.00000191, -0.00001085, 0.00000077, 0.00000014, -0.00001557, -0.00000768, -0.00000404, -0.00000116, -0.00000057, 0.00000266, -0.00000414, -0.00001609, },
		{ 0.00000006, -0.00000266, 0.00000191, -0.00000140, -0.00000175, -0.00000356, -0.00000122, -0.00000629, -0.00000148, 0.00000184, -0.00000229, 0.00000939, -0.00000022, -0.00001769, -0.00000555, 0.00000333, 0.00000872, 0.00000319, -0.00001212, -0.00000150, 0.00000193, -0.00000319, 0.00000528, 0.00000195, -0.00000430, 0.00000382, -0.00000342, -0.00000707, 0.00001081, 0.00000634, -0.00000209, -0.00000206, -0.00000906, 0.00000822, 0.00001499, -0.00000618, -0.00000175, -0.00000385, -0.00000588, 0.00000357, 0.00000266, 0.00000195, -0.00000082, 0.00000273, -0.00000464, -0.00000463, 0.00000445, -0.00000224, -0.00000230, 0.00000800, 0.00000086, 0.00000197, -0.00001093, -0.00000019, 0.00002020, -0.00002115, 0.00000879, -0.00000124, 0.00000464, -0.00000188, -0.00001351, 0.00002965, -0.00004120, 0.00004340, -0.00002451, -0.00002668, 0.00006392, -0.00006653, 0.00004563, -0.00001206, -0.00002409, 0.00002828, -0.00001956, 0.00003350, -0.00003401, 0.00001500, 0.00000336, -0.00000235, 0.00000049, -0.00001088, 0.00001501, -0.00002081, 0.00002460, -0.00000860, 0.00001198, -0.00000316, -0.00001741, -0.00000039, -0.00001624, 0.00000915, -0.00002033, 0.00000572, -0.00001881, -0.00000639, -0.00001050, -0.00001233, -0.00002437, -0.00002727, -0.00003270, -0.00020186, 0.00004545, 0.00022439, 0.00007437, 0.00007212, 0.00012300, 0.00019052, -0.00003557, -0.00015406, -0.00001443, -0.00002210, -0.00002130, -0.00001095, -0.00000329, -0.00000399, -0.00000907, 0.00000076, -0.00001007, 0.00000184, -0.00000171, -0.00001643, -0.00000725, -0.00000226, -0.00000106, -0.00000050, 0.00000279, -0.00000776, -0.00001428, 0.00000428, },
		{ -0.00000138, 0.00000119, -0.00000193, -0.00000136, -0.00000366, -0.00000195, -0.00000604, -0.00000027, 0.00000073, -0.00000004, 0.00000990, -0.00000632, -0.00001721, -0.00000137, 0.00000438, 0.00000834, -0.00000092, -0.00001125, 0.00000180, 0.00000022, -0.00000222, 0.00000537, -0.00000025, -0.00000156, 0.00000279, -0.00000682, -0.00000165, 0.00001189, 0.00000234, -0.00000162, -0.00000379, -0.00000721, 0.00001267, 0.00001120, -0.00000726, -0.00000122, -0.00000530, -0.00000354, 0.00000344, 0.00000231, 0.00000179, -0.00000013, 0.00000197, -0.00000696, -0.00000144, 0.00000424, -0.00000475, 0.00000188, 0.00000788, -0.00000092, 0.00000119, -0.00001271, 0.00000664, 0.00001283, -0.00001975, 0.00001327, -0.00000447, 0.00000810, -0.00001069, 0.00000005, 0.00001553, -0.00003259, 0.00004977, -0.00005235, 0.00001411, 0.00003580, -0.00006688, 0.00006889, -0.00004344, -0.00000021, 0.00001563, -0.00001388, 0.00003436, -0.00004479, 0.00003045, -0.00000855, 0.00000169, -0.00000227, -0.00000879, 0.00001440, -0.00002112, 0.00002510, -0.00000812, 0.00001201, -0.00000937, -0.00000954, -0.00000794, -0.00000853, 0.00000504, -0.00002183, 0.00001067, -0.00002767, 0.00000264, -0.00001568, -0.00001269, -0.00002659, -0.00002410, -0.00005220, -0.00020795, 0.00009226, 0.00023020, 0.00006091, 0.00008372, 0.00016973, 0.00015745, -0.00011596, -0.00012434, 0.00000510, -0.00003530, -0.00001158, -0.00001285, -0.00000314, -0.00000265, -0.00001004, 0.00000118, -0.00000972, 0.00000257, -0.00000341, -0.00001781, -0.00000562, -0.00000071, -0.00000169, 0.00000024, 0.00000182, -0.00001085, -0.00001111, 0.00000561, -0.00000190, },
		{ -0.00000004, -0.00000187, -0.00000114, -0.00000410, -0.00000235, -0.00000548, 0.00000033, 0.00000019, 0.00000267, 0.00000819, -0.00001190, -0.00001442, 0.00000210, 0.00000463, 0.00000715, -0.00000444, -0.00000878, 0.00000378, -0.00000152, -0.00000076, 0.00000427, -0.00000099, 0.00000100, -0.00000029, -0.00000746, 0.00000396, 0.00001018, -0.00000029, -0.00000120, -0.00000555, -0.00000365, 0.00001524, 0.00000702, -0.00000731, -0.00000146, -0.00000579, -0.00000148, 0.00000280, 0.00000239, 0.00000163, 0.00000060, 0.00000013, -0.00000788, 0.00000209, 0.00000206, -0.00000531, 0.00000622, 0.00000593, -0.00000191, -0.00000012, -0.00001254, 0.00001157, 0.00000486, -0.00001507, 0.00001497, -0.00000616, 0.00000856, -0.00001504, 0.00001039, 0.00000013, -0.00001505, 0.00003961, -0.00006316, 0.00005006, -0.00000603, -0.00004368, 0.00007278, -0.00006769, 0.00002758, -0.00000624, 0.00000287, 0.00002412, -0.00004756, 0.00004420, -0.00002490, 0.00001190, -0.00000945, -0.00000396, 0.00001302, -0.00002291, 0.00002838, -0.00000895, 0.00000931, -0.00001048, -0.00000576, -0.00001390, 0.00000100, -0.00000438, -0.00001676, 0.00000997, -0.00003364, 0.00001345, -0.00002621, -0.00000825, -0.00003097, -0.00002177, -0.00007315, -0.00021015, 0.00014507, 0.00022920, 0.00004369, 0.00011549, 0.00020684, 0.00008771, -0.00017095, -0.00007631, 0.00000523, -0.00004216, -0.00000089, -0.00001915, 0.00000121, -0.00000369, -0.00001098, 0.00000291, -0.00001054, 0.00000383, -0.00000573, -0.00001910, -0.00000264, -0.00000027, -0.00000230, 0.00000131, -0.00000036, -0.00001260, -0.00000740, 0.00000551, -0.00000318, -0.00000010, },
		{ -0.00000130, -0.00000138, -0.00000458, -0.00000234, -0.00000495, 0.00000055, 0.00000050, 0.00000500, 0.00000445, -0.00001585, -0.00000997, 0.00000429, 0.00000430, 0.00000539, -0.00000688, -0.00000543, 0.00000432, -0.00000273, 0.00000052, 0.00000259, -0.00000009, 0.00000211, -0.00000406, -0.00000515, 0.00000798, 0.00000687, -0.00000132, -0.00000151, -0.00000642, 0.00000077, 0.00001571, 0.00000328, -0.00000693, -0.00000210, -0.00000536, -0.00000010, 0.00000212, 0.00000271, 0.00000153, 0.00000099, -0.00000237, -0.00000699, 0.00000474, -0.00000120, -0.00000352, 0.00000946, 0.00000282, -0.00000208, -0.00000173, -0.00001078, 0.00001401, -0.00000210, -0.00000838, 0.00001417, -0.00000682, 0.00000720, -0.00001522, 0.00001520, -0.00001091, 0.00000390, 0.00001779, -0.00005369, 0.00006885, -0.00004729, -0.00000393, 0.00005393, -0.00007496, 0.00004895, -0.00003005, 0.00002740, 0.00000175, -0.00003734, 0.00004932, -0.00004012, 0.00002633, -0.00002224, 0.00000652, 0.00000730, -0.00002340, 0.00003414, -0.00001384, 0.00000783, -0.00000873, -0.00000673, -0.00001513, 0.00000800, -0.00001552, -0.00000628, 0.00000185, -0.00003245, 0.00002097, -0.00003914, 0.00000147, -0.00004052, -0.00001685, -0.00009824, -0.00020638, 0.00020516, 0.00021480, 0.00003170, 0.00016792, 0.00021520, -0.00000506, -0.00018805, -0.00002713, -0.00001057, -0.00003785, 0.00000480, -0.00002595, 0.00000895, -0.00000863, -0.00000961, 0.00000445, -0.00001220, 0.00000599, -0.00000944, -0.00001918, 0.00000103, -0.00000130, -0.00000211, 0.00000193, -0.00000327, -0.00001260, -0.00000409, 0.00000442, -0.00000359, 0.00000073, -0.00000137, },
		{ -0.00000209, -0.00000478, -0.00000206, -0.00000465, 0.00000074, 0.00000163, 0.00000616, -0.00000064, -0.00001738, -0.00000487, 0.00000496, 0.00000371, 0.00000335, -0.00000801, -0.00000192, 0.00000370, -0.00000320, 0.00000111, 0.00000118, 0.00000191, 0.00000113, -0.00000687, -0.00000078, 0.00000941, 0.00000344, -0.00000130, -0.00000262, -0.00000575, 0.00000495, 0.00001448, 0.00000041, -0.00000659, -0.00000266, -0.00000438, 0.00000052, 0.00000172, 0.00000309, 0.00000154, 0.00000068, -0.00000479, -0.00000448, 0.00000560, -0.00000426, 0.00000020, 0.00001073, -0.00000054, -0.00000160, -0.00000333, -0.00000806, 0.00001398, -0.00000690, -0.00000104, 0.00001124, -0.00000672, 0.00000539, -0.00001308, 0.00001504, -0.00001475, 0.00001705, -0.00000666, -0.00002835, 0.00006460, -0.00007392, 0.00003908, 0.00001751, -0.00006140, 0.00005508, -0.00004621, 0.00005181, -0.00002847, -0.00001330, 0.00004027, -0.00004689, 0.00003937, -0.00003818, 0.00002318, -0.00000587, -0.00001809, 0.00003891, -0.00002273, 0.00001084, -0.00000814, -0.00001007, -0.00001121, 0.00000923, -0.00002367, 0.00000554, -0.00001201, -0.00002216, 0.00002010, -0.00004888, 0.00001327, -0.00005546, -0.00000663, -0.00013177, -0.00019077, 0.00026915, 0.00018282, 0.00003808, 0.00023080, 0.00018119, -0.00009826, -0.00016674, 0.00000749, -0.00003265, -0.00002324, 0.00000173, -0.00002843, 0.00001680, -0.00001677, -0.00000454, 0.00000386, -0.00001341, 0.00000845, -0.00001467, -0.00001700, 0.00000412, -0.00000335, -0.00000079, 0.00000132, -0.00000596, -0.00001109, -0.00000187, 0.00000305, -0.00000333, 0.00000092, -0.00000243, -0.00000052, },
		{ -0.00000455, -0.00000177, -0.00000462, 0.00000127, 0.00000322, 0.00000559, -0.00000599, -0.00001626, -0.00000024, 0.00000430, 0.00000315, 0.00000126, -0.00000785, 0.00000115, 0.00000237, -0.00000309, 0.00000089, 0.00000078, 0.00000394, -0.00000171, -0.00000746, 0.00000407, 0.00000829, 0.00000101, -0.00000110, -0.00000397, -0.00000352, 0.00000796, 0.00001231, -0.00000164, -0.00000648, -0.00000278, -0.00000336, 0.00000062, 0.00000172, 0.00000342, 0.00000155, -0.00000048, -0.00000630, -0.00000113, 0.00000443, -0.00000587, 0.00000479, 0.00000977, -0.00000333, -0.00000078, -0.00000471, -0.00000509, 0.00001206, -0.00000890, 0.00000563, 0.00000666, -0.00000570, 0.00000382, -0.00001070, 0.00001253, -0.00001228, 0.00002054, -0.00002439, 0.00000279, 0.00003989, -0.00007742, 0.00007091, -0.00002475, -0.00003098, 0.00004262, -0.00004689, 0.00006638, -0.00005769, 0.00001964, 0.00001605, -0.00003936, 0.00004371, -0.00005125, 0.00004269, -0.00002670, -0.00000321, 0.00003719, -0.00003192, 0.00001864, -0.00001213, -0.00001158, -0.00000494, 0.00000444, -0.00002541, 0.00001352, -0.00002678, -0.00000480, 0.00000838, -0.00004964, 0.00002100, -0.00007189, 0.00000813, -0.00017668, -0.00015539, 0.00032802, 0.00013584, 0.00007448, 0.00028411, 0.00010337, -0.00016769, -0.00011865, 0.00001905, -0.00004972, -0.00000431, -0.00000965, -0.00002338, 0.00002054, -0.00002528, 0.00000375, -0.00000006, -0.00001256, 0.00000981, -0.00002052, -0.00001225, 0.00000533, -0.00000532, 0.00000124, -0.00000080, -0.00000739, -0.00000891, -0.00000089, 0.00000199, -0.00000285, 0.00000061, -0.00000302, 0.00000013, -0.00000309, },
		{ -0.00000168, -0.00000471, 0.00000240, 0.00000464, 0.00000315, -0.00001038, -0.00001291, 0.00000297, 0.00000286, 0.00000281, -0.00000073, -0.00000655, 0.00000340, 0.00000078, -0.00000274, 0.00000018, 0.00000167, 0.00000488, -0.00000537, -0.00000542, 0.00000777, 0.00000555, -0.00000002, -0.00000145, -0.00000473, -0.00000030, 0.00000936, 0.00000999, -0.00000319, -0.00000645, -0.00000241, -0.00000266, 0.00000057, 0.00000202, 0.00000368, 0.00000134, -0.00000230, -0.00000632, 0.00000192, 0.00000176, -0.00000527, 0.00000894, 0.00000700, -0.00000503, 0.00000006, -0.00000572, -0.00000252, 0.00000913, -0.00000808, 0.00001046, 0.00000119, -0.00000340, 0.00000220, -0.00000917, 0.00001062, -0.00000717, 0.00001522, -0.00002966, 0.00002816, 0.00000453, -0.00005787, 0.00008150, -0.00005906, 0.00000597, 0.00001537, -0.00002968, 0.00006365, -0.00007549, 0.00005177, -0.00001823, -0.00001635, 0.00003363, -0.00005396, 0.00005805, -0.00005096, 0.00002161, 0.00002417, -0.00003526, 0.00002763, -0.00002107, -0.00000777, -0.00000093, -0.00000320, -0.00002062, 0.00001407, -0.00003647, 0.00001392, -0.00001218, -0.00003847, 0.00001827, -0.00008273, 0.00002180, -0.00023125, -0.00009404, 0.00036920, 0.00008539, 0.00014435, 0.00030367, -0.00000471, -0.00019610, -0.00006309, 0.00000899, -0.00005390, 0.00001084, -0.00002474, -0.00001100, 0.00001723, -0.00003036, 0.00001289, -0.00000696, -0.00000873, 0.00000851, -0.00002528, -0.00000571, 0.00000402, -0.00000598, 0.00000292, -0.00000398, -0.00000698, -0.00000711, -0.00000071, 0.00000148, -0.00000257, 0.00000014, -0.00000325, 0.00000044, -0.00000409, -0.00000392, },
		{ -0.00000468, 0.00000418, 0.00000517, -0.00000076, -0.00001278, -0.00000832, 0.00000426, 0.00000134, 0.00000261, -0.00000245, -0.00000439, 0.00000465, -0.00000077, -0.00000249, -0.00000047, 0.00000354, 0.00000406, -0.00000845, -0.00000132, 0.00000924, 0.00000248, 0.00000002, -0.00000257, -0.00000421, 0.00000297, 0.00000932, 0.00000799, -0.00000458, -0.00000622, -0.00000178, -0.00000240, 0.00000067, 0.00000245, 0.00000389, 0.00000063, -0.00000424, -0.00000477, 0.00000366, -0.00000132, -0.00000249, 0.00001149, 0.00000327, -0.00000549, 0.00000069, -0.00000636, -0.00000073, 0.00000624, -0.00000498, 0.00001257, -0.00000405, 0.00000021, -0.00000034, -0.00000802, 0.00001089, -0.00000373, 0.00000585, -0.00002241, 0.00003933, -0.00002862, -0.00002345, 0.00006873, -0.00007495, 0.00003684, -0.00001695, 0.00000087, 0.00004193, -0.00007417, 0.00007186, -0.00005239, 0.00001704, 0.00000826, -0.00004080, 0.00006105, -0.00007045, 0.00005157, -0.00000117, -0.00002708, 0.00003145, -0.00003127, 0.00000164, -0.00000298, -0.00000861, -0.00001288, 0.00000734, -0.00003696, 0.00002690, -0.00003529, -0.00001738, 0.00000183, -0.00008074, 0.00002503, -0.00028738, -0.00000660, 0.00038158, 0.00005012, 0.00023776, 0.00027005, -0.00011728, -0.00017972, -0.00001952, -0.00001284, -0.00004400, 0.00001596, -0.00003691, 0.00000491, 0.00000675, -0.00002914, 0.00001955, -0.00001485, -0.00000240, 0.00000379, -0.00002710, 0.00000085, 0.00000064, -0.00000467, 0.00000306, -0.00000703, -0.00000494, -0.00000649, -0.00000058, 0.00000126, -0.00000258, -0.00000020, -0.00000338, 0.00000048, -0.00000489, -0.00000304, -0.00000192, },
		{ 0.00000638, 0.00000415, -0.00000521, -0.00001264, -0.00000377, 0.00000376, 0.00000038, 0.00000233, -0.00000369, -0.00000171, 0.00000488, -0.00000212, -0.00000245, -0.00000052, 0.00000563, 0.00000147, -0.00000972, 0.00000354, 0.00000829, 0.00000023, 0.00000032, -0.00000405, -0.00000219, 0.00000536, 0.00000839, 0.00000642, -0.00000600, -0.00000554, -0.00000127, -0.00000241, 0.00000105, 0.00000287, 0.00000401, -0.00000075, -0.00000555, -0.00000218, 0.00000356, -0.00000351, 0.00000168, 0.00001183, -0.00000043, -0.00000493, 0.00000101, -0.00000679, 0.00000019, 0.00000432, -0.00000068, 0.00001166, -0.00000774, 0.00000441, -0.00000446, -0.00000573, 0.00001268, -0.00000457, -0.00000146, -0.00000791, 0.00003415, -0.00004868, 0.00001310, 0.00003879, -0.00006904, 0.00005171, -0.00004261, 0.00003476, 0.00000660, -0.00005223, 0.00007158, -0.00007462, 0.00005047, -0.00002688, -0.00001149, 0.00004592, -0.00007575, 0.00007734, -0.00003410, -0.00000562, 0.00002407, -0.00003615, 0.00001261, -0.00001135, -0.00000772, -0.00000776, -0.00000242, -0.00002844, 0.00002908, -0.00005291, 0.00000680, -0.00002600, -0.00006242, 0.00000830, -0.00033186, 0.00009847, 0.00036165, 0.00004953, 0.00033107, 0.00017746, -0.00020442, -0.00012975, -0.00000014, -0.00003315, -0.00002560, 0.00000966, -0.00004077, 0.00001865, -0.00000779, -0.00002113, 0.00002110, -0.00002097, 0.00000451, -0.00000371, -0.00002494, 0.00000542, -0.00000335, -0.00000174, 0.00000100, -0.00000858, -0.00000227, -0.00000718, 0.00000013, 0.00000086, -0.00000265, -0.00000039, -0.00000367, 0.00000047, -0.00000544, -0.00000206, -0.00000251, -0.00000422, },
		{ 0.00000128, -0.00000898, -0.00001016, -0.00000050, 0.00000215, 0.00000027, 0.00000168, -0.00000420, 0.00000110, 0.00000418, -0.00000321, -0.00000253, 0.00000029, 0.00000705, -0.00000219, -0.00000859, 0.00000768, 0.00000560, -0.00000067, 0.00000013, -0.00000514, 0.00000094, 0.00000631, 0.00000728, 0.00000506, -0.00000733, -0.00000437, -0.00000117, -0.00000238, 0.00000161, 0.00000325, 0.00000387, -0.00000272, -0.00000564, 0.00000050, 0.00000183, -0.00000384, 0.00000597, 0.00001009, -0.00000331, -0.00000377, 0.00000104, -0.00000724, 0.00000054, 0.00000395, 0.00000347, 0.00000819, -0.00000881, 0.00000783, -0.00000980, -0.00000086, 0.00001354, -0.00000921, -0.00000230, 0.00000594, 0.00001705, -0.00005078, 0.00003966, 0.00000329, -0.00004594, 0.00004711, -0.00005255, 0.00006030, -0.00003153, -0.00001564, 0.00004924, -0.00007603, 0.00007191, -0.00006105, 0.00002756, 0.00001281, -0.00006034, 0.00008809, -0.00006471, 0.00002481, 0.00000339, -0.00002943, 0.00001811, -0.00002168, -0.00000031, -0.00000962, -0.00000879, -0.00001585, 0.00002025, -0.00005905, 0.00002516, -0.00005748, -0.00003079, -0.00003332, -0.00035082, 0.00020542, 0.00031773, 0.00009522, 0.00039247, 0.00003873, -0.00024256, -0.00006812, -0.00000564, -0.00004141, -0.00000791, -0.00000408, -0.00003473, 0.00002526, -0.00002143, -0.00000867, 0.00001685, -0.00002310, 0.00000951, -0.00001197, -0.00001924, 0.00000667, -0.00000612, 0.00000148, -0.00000291, -0.00000770, -0.00000033, -0.00000857, 0.00000158, -0.00000014, -0.00000236, -0.00000073, -0.00000408, 0.00000061, -0.00000584, -0.00000107, -0.00000342, -0.00000304, 0.00000850, },
		{ -0.00001083, -0.00000625, 0.00000078, 0.00000046, 0.00000082, 0.00000054, -0.00000373, 0.00000360, 0.00000271, -0.00000402, -0.00000251, 0.00000186, 0.00000715, -0.00000580, -0.00000533, 0.00000995, 0.00000230, -0.00000037, -0.00000093, -0.00000506, 0.00000433, 0.00000581, 0.00000653, 0.00000360, -0.00000831, -0.00000295, -0.00000156, -0.00000204, 0.00000220, 0.00000362, 0.00000324, -0.00000486, -0.00000428, 0.00000225, -0.00000068, -0.00000202, 0.00000911, 0.00000697, -0.00000498, -0.00000239, 0.00000082, -0.00000796, 0.00000092, 0.00000515, 0.00000616, 0.00000338, -0.00000696, 0.00000889, -0.00001483, 0.00000675, 0.00001079, -0.00001442, 0.00000396, 0.00001232, -0.00000330, -0.00003734, 0.00004922, -0.00002566, -0.00001578, 0.00002696, -0.00004390, 0.00006879, -0.00006024, 0.00002399, 0.00001101, -0.00005461, 0.00007238, -0.00008198, 0.00006466, -0.00003089, -0.00002428, 0.00007604, -0.00008112, 0.00005419, -0.00002634, -0.00000897, 0.00001146, -0.00002642, 0.00000884, -0.00001846, -0.00000678, -0.00000670, 0.00000562, -0.00005303, 0.00003092, -0.00008248, 0.00000468, -0.00009681, -0.00033556, 0.00029661, 0.00026957, 0.00018316, 0.00039252, -0.00011630, -0.00022341, -0.00001893, -0.00002588, -0.00003451, 0.00000093, -0.00001793, -0.00002181, 0.00002294, -0.00002959, 0.00000417, 0.00000849, -0.00002074, 0.00001070, -0.00001838, -0.00001187, 0.00000465, -0.00000640, 0.00000329, -0.00000731, -0.00000445, -0.00000022, -0.00000964, 0.00000331, -0.00000170, -0.00000141, -0.00000164, -0.00000429, 0.00000089, -0.00000619, -0.00000014, -0.00000454, -0.00000094, 0.00000978, -0.00000325, },
		{ -0.00000237, 0.00000011, -0.00000043, 0.00000148, -0.00000090, -0.00000222, 0.00000533, 0.00000075, -0.00000449, -0.00000213, 0.00000372, 0.00000574, -0.00000825, -0.00000086, 0.00000993, -0.00000046, 0.00000045, -0.00000265, -0.00000341, 0.00000701, 0.00000433, 0.00000633, 0.00000182, -0.00000867, -0.00000163, -0.00000228, -0.00000128, 0.00000265, 0.00000399, 0.00000191, -0.00000651, -0.00000179, 0.00000238, -0.00000272, 0.00000145, 0.00001026, 0.00000348, -0.00000553, -0.00000105, 0.00000031, -0.00000901, 0.00000203, 0.00000732, 0.00000659, -0.00000112, -0.00000292, 0.00000667, -0.00001735, 0.00001542, 0.00000319, -0.00001588, 0.00001387, 0.00000843, -0.00001801, -0.00001632, 0.00004196, -0.00003999, 0.00001011, 0.00000053, -0.00002087, 0.00005789, -0.00007061, 0.00005365, -0.00003096, -0.00001646, 0.00004988, -0.00008075, 0.00008682, -0.00007219, 0.00002429, 0.00004061, -0.00007442, 0.00007020, -0.00005489, 0.00002068, -0.00000955, -0.00001852, 0.00001156, -0.00002875, 0.00000361, -0.00000697, -0.00000647, -0.00004028, 0.00002317, -0.00009330, 0.00003163, -0.00017111, -0.00028708, 0.00036034, 0.00024220, 0.00029089, 0.00031581, -0.00024928, -0.00015748, 0.00000099, -0.00004521, -0.00001767, -0.00000204, -0.00002505, -0.00000805, 0.00001381, -0.00003022, 0.00001339, -0.00000075, -0.00001539, 0.00000782, -0.00002097, -0.00000539, 0.00000084, -0.00000415, 0.00000246, -0.00001037, 0.00000006, -0.00000223, -0.00000948, 0.00000455, -0.00000335, 0.00000005, -0.00000340, -0.00000383, 0.00000107, -0.00000646, 0.00000068, -0.00000568, 0.00000211, 0.00000938, -0.00000857, -0.00001000, },
		{ -0.00000168, -0.00000009, 0.00000159, -0.00000215, 0.00000016, 0.00000594, -0.00000136, -0.00000458, -0.00000129, 0.00000529, 0.00000318, -0.00000886, 0.00000354, 0.00000797, -0.00000197, 0.00000097, -0.00000435, -0.00000034, 0.00000825, 0.00000258, 0.00000652, -0.00000025, -0.00000829, -0.00000078, -0.00000298, -0.00000021, 0.00000292, 0.00000424, -0.00000013, -0.00000703, 0.00000102, 0.00000085, -0.00000321, 0.00000544, 0.00000927, 0.00000054, -0.00000531, 0.00000018, -0.00000067, -0.00001014, 0.00000440, 0.00000937, 0.00000478, -0.00000381, 0.00000166, 0.00000149, -0.00001562, 0.00002217, -0.00000796, -0.00001060, 0.00002171, -0.00000341, -0.00002170, 0.00000261, 0.00002430, -0.00003828, 0.00002372, -0.00002171, 0.00000730, 0.00003241, -0.00006051, 0.00006388, -0.00006295, 0.00002606, 0.00001071, -0.00005575, 0.00008490, -0.00009691, 0.00007104, -0.00000993, -0.00004320, 0.00006339, -0.00006957, 0.00004870, -0.00003998, 0.00000423, 0.00000032, -0.00003127, 0.00001611, -0.00001712, -0.00000916, -0.00002975, 0.00000777, -0.00008892, 0.00004044, -0.00024129, -0.00021659, 0.00039752, 0.00025556, 0.00038200, 0.00016927, -0.00032608, -0.00007058, -0.00001159, -0.00005004, -0.00000141, -0.00001326, -0.00002254, 0.00000055, 0.00000287, -0.00002462, 0.00001700, -0.00000774, -0.00000979, 0.00000249, -0.00001939, -0.00000189, -0.00000251, -0.00000070, -0.00000103, -0.00001063, 0.00000409, -0.00000568, -0.00000778, 0.00000466, -0.00000434, 0.00000142, -0.00000584, -0.00000238, 0.00000073, -0.00000644, 0.00000121, -0.00000649, 0.00000577, 0.00000665, -0.00001330, -0.00000653, 0.00000003, },
		{ 0.00000123, 0.00000075, -0.00000259, 0.00000289, 0.00000532, -0.00000324, -0.00000424, -0.00000006, 0.00000608, 0.00000021, -0.00000755, 0.00000677, 0.00000496, -0.00000216, 0.00000067, -0.00000521, 0.00000346, 0.00000779, 0.00000125, 0.00000676, -0.00000241, -0.00000729, -0.00000055, -0.00000333, 0.00000096, 0.00000306, 0.00000417, -0.00000256, -0.00000608, 0.00000314, -0.00000168, -0.00000160, 0.00000862, 0.00000669, -0.00000133, -0.00000476, 0.00000133, -0.00000235, -0.00001079, 0.00000806, 0.00001012, 0.00000161, -0.00000394, 0.00000485, -0.00000496, -0.00000932, 0.00002399, -0.00001907, 0.00000136, 0.00002225, -0.00001710, -0.00001450, 0.00001222, 0.00000548, -0.00002534, 0.00002302, -0.00003230, 0.00003034, 0.00000208, -0.00003522, 0.00005237, -0.00007485, 0.00005891, -0.00003251, -0.00001395, 0.00005767, -0.00009519, 0.00010024, -0.00006039, 0.00000452, 0.00003202, -0.00006061, 0.00006170, -0.00006824, 0.00003633, -0.00002695, -0.00001757, 0.00002015, -0.00003046, -0.00000158, -0.00002896, -0.00000538, -0.00007612, 0.00002932, -0.00029545, -0.00014049, 0.00042361, 0.00031377, 0.00041753, -0.00001711, -0.00032894, 0.00000584, -0.00004569, -0.00003508, 0.00000403, -0.00002458, -0.00001292, 0.00000110, -0.00000465, -0.00001666, 0.00001579, -0.00001096, -0.00000651, -0.00000254, -0.00001515, -0.00000191, -0.00000351, 0.00000188, -0.00000589, -0.00000775, 0.00000607, -0.00000925, -0.00000504, 0.00000350, -0.00000412, 0.00000190, -0.00000831, -0.00000003, -0.00000040, -0.00000586, 0.00000121, -0.00000651, 0.00000930, 0.00000149, -0.00001589, -0.00000142, 0.00000005, -0.00000332, },
		{ -0.00000092, -0.00000177, 0.00000527, 0.00000367, -0.00000457, -0.00000351, 0.00000130, 0.00000587, -0.00000234, -0.00000485, 0.00000822, 0.00000196, -0.00000149, -0.00000049, -0.00000464, 0.00000702, 0.00000590, 0.00000075, 0.00000669, -0.00000433, -0.00000596, -0.00000089, -0.00000314, 0.00000199, 0.00000316, 0.00000354, -0.00000484, -0.00000377, 0.00000376, -0.00000409, 0.00000184, 0.00000995, 0.00000352, -0.00000210, -0.00000415, 0.00000236, -0.00000483, -0.00001021, 0.00001243, 0.00000878, -0.00000144, -0.00000188, 0.00000528, -0.00001021, -0.00000009, 0.00001939, -0.00002570, 0.00001652, 0.00001339, -0.00002571, -0.00000146, 0.00001046, -0.00000656, -0.00000937, 0.00001190, -0.00002941, 0.00004114, -0.00002255, -0.00000512, 0.00002464, -0.00006391, 0.00007162, -0.00006550, 0.00003133, 0.00001298, -0.00006591, 0.00010087, -0.00009375, 0.00005314, -0.00001576, -0.00002639, 0.00004947, -0.00008003, 0.00006511, -0.00006385, 0.00001463, 0.00000590, -0.00003528, 0.00000915, -0.00003883, -0.00000808, -0.00006635, 0.00000578, -0.00033113, -0.00007147, 0.00046350, 0.00039907, 0.00037065, -0.00019822, -0.00026326, 0.00004649, -0.00008153, -0.00000571, -0.00000593, -0.00002757, -0.00000280, -0.00000469, -0.00000592, -0.00001067, 0.00001271, -0.00001127, -0.00000650, -0.00000481, -0.00001095, -0.00000424, -0.00000165, 0.00000181, -0.00000996, -0.00000274, 0.00000532, -0.00001164, -0.00000225, 0.00000164, -0.00000271, 0.00000089, -0.00000993, 0.00000268, -0.00000227, -0.00000459, 0.00000054, -0.00000534, 0.00001166, -0.00000543, -0.00001518, 0.00000364, -0.00000229, -0.00000362, 0.00000267, },
		{ 0.00000034, 0.00000662, 0.00000144, -0.00000520, -0.00000249, 0.00000249, 0.00000478, -0.00000383, -0.00000167, 0.00000789, -0.00000025, -0.00000069, -0.00000205, -0.00000249, 0.00000945, 0.00000321, 0.00000115, 0.00000607, -0.00000577, -0.00000465, -0.00000153, -0.00000242, 0.00000274, 0.00000320, 0.00000224, -0.00000630, -0.00000076, 0.00000261, -0.00000518, 0.00000616, 0.00000906, 0.00000077, -0.00000214, -0.00000353, 0.00000308, -0.00000795, -0.00000775, 0.00001637, 0.00000532, -0.00000296, 0.00000100, 0.00000283, -0.00001212, 0.00000900, 0.00000937, -0.00002462, 0.00002932, -0.00000275, -0.00002472, 0.00001010, 0.00000064, -0.00000828, 0.00000193, -0.00000235, -0.00001684, 0.00003822, -0.00003433, 0.00001864, -0.00000837, -0.00003554, 0.00006121, -0.00007777, 0.00006531, -0.00003472, -0.00001776, 0.00007130, -0.00009774, 0.00008490, -0.00006379, 0.00002431, 0.00001048, -0.00006458, 0.00007510, -0.00009608, 0.00005801, -0.00002961, -0.00002045, 0.00001019, -0.00005133, 0.00000033, -0.00006905, -0.00001707, -0.00035759, -0.00000938, 0.00054004, 0.00047411, 0.00023886, -0.00032841, -0.00015582, 0.00004161, -0.00009959, 0.00002483, -0.00002769, -0.00001805, 0.00000061, -0.00001142, -0.00000200, -0.00000905, 0.00001106, -0.00001127, -0.00000853, -0.00000347, -0.00000926, -0.00000646, 0.00000194, -0.00000141, -0.00001129, 0.00000247, 0.00000232, -0.00001220, -0.00000038, 0.00000001, -0.00000077, -0.00000161, -0.00000998, 0.00000493, -0.00000445, -0.00000278, -0.00000070, -0.00000283, 0.00001174, -0.00001265, -0.00001087, 0.00000676, -0.00000595, -0.00000202, 0.00000321, -0.00000277, },
		{ 0.00000657, -0.00000076, -0.00000515, -0.00000135, 0.00000326, 0.00000323, -0.00000403, 0.00000110, 0.00000634, -0.00000137, -0.00000037, -0.00000328, 0.00000081, 0.00001018, 0.00000052, 0.00000220, 0.00000490, -0.00000660, -0.00000363, -0.00000216, -0.00000133, 0.00000319, 0.00000309, 0.00000040, -0.00000647, 0.00000202, 0.00000009, -0.00000420, 0.00001002, 0.00000630, -0.00000084, -0.00000196, -0.00000276, 0.00000314, -0.00001115, -0.00000323, 0.00001858, 0.00000064, -0.00000217, 0.00000296, -0.00000123, -0.00000984, 0.00001468, -0.00000271, -0.00001542, 0.00003455, -0.00002060, -0.00001399, 0.00001409, -0.00001066, -0.00000138, 0.00000452, -0.00001265, -0.00000159, 0.00002566, -0.00003198, 0.00002852, -0.00003477, -0.00000091, 0.00003304, -0.00006651, 0.00007736, -0.00006969, 0.00003343, 0.00002068, -0.00007005, 0.00008682, -0.00009357, 0.00007424, -0.00004562, -0.00002054, 0.00005483, -0.00010633, 0.00009634, -0.00007854, 0.00001772, -0.00001083, -0.00005191, 0.00001012, -0.00008462, -0.00002865, -0.00039143, 0.00006334, 0.00066038, 0.00049342, 0.00004847, -0.00037738, -0.00004493, -0.00000003, -0.00008924, 0.00004192, -0.00005098, 0.00000144, -0.00000658, -0.00001290, 0.00000282, -0.00001105, 0.00001252, -0.00001368, -0.00000983, 0.00000017, -0.00001105, -0.00000624, 0.00000492, -0.00000664, -0.00000906, 0.00000590, -0.00000150, -0.00001125, 0.00000021, -0.00000057, 0.00000066, -0.00000491, -0.00000829, 0.00000602, -0.00000629, -0.00000088, -0.00000210, 0.00000063, 0.00000887, -0.00001823, -0.00000388, 0.00000661, -0.00000921, 0.00000112, 0.00000215, -0.00000383, 0.00000323, },
		{ -0.00000239, -0.00000463, -0.00000022, 0.00000347, 0.00000170, -0.00000316, 0.00000284, 0.00000437, -0.00000161, -0.00000077, -0.00000348, 0.00000447, 0.00000917, -0.00000151, 0.00000351, 0.00000332, -0.00000689, -0.00000299, -0.00000252, -0.00000010, 0.00000338, 0.00000266, -0.00000158, -0.00000526, 0.00000367, -0.00000282, -0.00000110, 0.00001222, 0.00000260, -0.00000115, -0.00000197, -0.00000171, 0.00000219, -0.00001356, 0.00000288, 0.00001806, -0.00000374, 0.00000061, 0.00000265, -0.00000477, -0.00000440, 0.00001491, -0.00001242, -0.00000100, 0.00002963, -0.00003359, 0.00000225, 0.00000827, -0.00001688, 0.00000875, -0.00000087, -0.00001511, 0.00000959, 0.00001066, -0.00002002, 0.00002323, -0.00004637, 0.00002767, -0.00000164, -0.00003743, 0.00006504, -0.00008088, 0.00007049, -0.00003403, -0.00001990, 0.00005629, -0.00009159, 0.00010371, -0.00010020, 0.00004140, 0.00000322, -0.00008184, 0.00010980, -0.00012278, 0.00007081, -0.00005806, -0.00002653, 0.00000391, -0.00010085, -0.00002995, -0.00044613, 0.00017782, 0.00080593, 0.00042084, -0.00015175, -0.00034113, 0.00003392, -0.00005590, -0.00005327, 0.00003664, -0.00006376, 0.00002252, -0.00002259, -0.00000550, 0.00000334, -0.00001314, 0.00001604, -0.00001958, -0.00000761, 0.00000320, -0.00001518, -0.00000267, 0.00000504, -0.00001152, -0.00000408, 0.00000646, -0.00000458, -0.00000982, -0.00000006, 0.00000011, 0.00000069, -0.00000790, -0.00000537, 0.00000571, -0.00000725, 0.00000058, -0.00000299, 0.00000410, 0.00000319, -0.00002040, 0.00000387, 0.00000300, -0.00001040, 0.00000457, -0.00000049, -0.00000329, 0.00000601, 0.00000520, },
		{ -0.00000391, 0.00000077, 0.00000314, 0.00000066, -0.00000175, 0.00000336, 0.00000268, -0.00000148, -0.00000167, -0.00000230, 0.00000756, 0.00000683, -0.00000250, 0.00000466, 0.00000157, -0.00000679, -0.00000269, -0.00000249, 0.00000106, 0.00000335, 0.00000179, -0.00000315, -0.00000306, 0.00000368, -0.00000492, 0.00000340, 0.00001208, -0.00000090, -0.00000049, -0.00000227, -0.00000041, 0.00000010, -0.00001426, 0.00000944, 0.00001462, -0.00000624, 0.00000405, -0.00000011, -0.00000576, 0.00000168, 0.00000993, -0.00001615, 0.00001352, 0.00001589, -0.00003691, 0.00001735, -0.00000501, -0.00001443, 0.00001598, -0.00000988, -0.00001027, 0.00001310, -0.00000000, -0.00000610, 0.00000757, -0.00004144, 0.00004154, -0.00003040, -0.00000215, 0.00003479, -0.00006616, 0.00008142, -0.00007419, 0.00003498, 0.00000290, -0.00005530, 0.00009839, -0.00013202, 0.00010084, -0.00006755, -0.00002148, 0.00008356, -0.00013988, 0.00011864, -0.00012193, 0.00002938, -0.00003512, -0.00009600, -0.00003858, -0.00051944, 0.00036614, 0.00093167, 0.00024637, -0.00030629, -0.00024339, 0.00006106, -0.00009990, -0.00000613, 0.00001025, -0.00005818, 0.00003492, -0.00004046, 0.00000959, -0.00000355, -0.00001106, 0.00001826, -0.00002747, -0.00000079, 0.00000267, -0.00001887, 0.00000320, 0.00000133, -0.00001362, 0.00000161, 0.00000444, -0.00000604, -0.00000902, -0.00000017, 0.00000141, -0.00000089, -0.00000958, -0.00000221, 0.00000434, -0.00000720, 0.00000131, -0.00000284, 0.00000629, -0.00000413, -0.00001824, 0.00000995, -0.00000295, -0.00000856, 0.00000676, -0.00000384, -0.00000087, 0.00000767, 0.00000307, -0.00000321, },
		{ 0.00000158, 0.00000245, 0.00000032, -0.00000040, 0.00000296, 0.00000165, -0.00000154, -0.00000250, 0.00000012, 0.00000933, 0.00000386, -0.00000243, 0.00000536, -0.00000014, -0.00000648, -0.00000255, -0.00000211, 0.00000205, 0.00000307, 0.00000052, -0.00000382, -0.00000063, 0.00000217, -0.00000527, 0.00000807, 0.00000967, -0.00000329, 0.00000053, -0.00000270, 0.00000086, -0.00000288, -0.00001266, 0.00001495, 0.00000903, -0.00000593, 0.00000636, -0.00000418, -0.00000334, 0.00000546, 0.00000227, -0.00001268, 0.00002285, -0.00000192, -0.00002960, 0.00002526, -0.00002008, -0.00000401, 0.00001629, -0.00001709, -0.00000208, 0.00000964, -0.00000288, 0.00000259, -0.00001019, -0.00002473, 0.00003856, -0.00004446, 0.00002667, -0.00000095, -0.00003289, 0.00006391, -0.00008655, 0.00007481, -0.00005464, 0.00000486, 0.00005571, -0.00012567, 0.00013468, -0.00013464, 0.00006092, 0.00001582, -0.00011258, 0.00013579, -0.00017933, 0.00010438, -0.00011252, -0.00004867, -0.00008576, -0.00058446, 0.00064345, 0.00097718, -0.00000453, -0.00037264, -0.00012648, 0.00003908, -0.00011372, 0.00003298, -0.00002633, -0.00003461, 0.00003150, -0.00005124, 0.00002662, -0.00001692, -0.00000239, 0.00001533, -0.00003369, 0.00000909, -0.00000279, -0.00001921, 0.00000867, -0.00000524, -0.00001163, 0.00000564, 0.00000132, -0.00000605, -0.00000933, 0.00000083, 0.00000221, -0.00000345, -0.00000953, 0.00000026, 0.00000264, -0.00000651, 0.00000148, -0.00000156, 0.00000605, -0.00001110, -0.00001218, 0.00001239, -0.00000909, -0.00000399, 0.00000649, -0.00000650, 0.00000290, 0.00000759, 0.00000102, -0.00000331, -0.00000320, },
		{ 0.00000163, 0.00000065, 0.00000037, 0.00000215, 0.00000123, -0.00000210, -0.00000262, 0.00000320, 0.00000945, 0.00000103, -0.00000154, 0.00000550, -0.00000168, -0.00000606, -0.00000244, -0.00000148, 0.00000282, 0.00000245, -0.00000087, -0.00000340, 0.00000119, -0.00000016, -0.00000351, 0.00001159, 0.00000575, -0.00000413, 0.00000133, -0.00000297, 0.00000170, -0.00000609, -0.00000879, 0.00001804, 0.00000285, -0.00000295, 0.00000609, -0.00000751, 0.00000169, 0.00000496, -0.00000430, -0.00000385, 0.00002378, -0.00001758, -0.00001486, 0.00002300, -0.00003069, 0.00001011, 0.00000938, -0.00001876, 0.00000454, 0.00000322, 0.00000103, 0.00000244, -0.00002219, -0.00000472, 0.00002324, -0.00004171, 0.00004031, -0.00002893, 0.00000516, 0.00002608, -0.00006827, 0.00008537, -0.00009537, 0.00006870, -0.00001293, -0.00007851, 0.00012623, -0.00017228, 0.00013949, -0.00007833, -0.00003807, 0.00010230, -0.00020082, 0.00017089, -0.00021437, 0.00004863, -0.00020428, -0.00059037, 0.00099197, 0.00088732, -0.00027440, -0.00033758, -0.00003485, -0.00001000, -0.00009436, 0.00004918, -0.00005758, -0.00000172, 0.00001199, -0.00004837, 0.00003785, -0.00003191, 0.00001178, 0.00000523, -0.00003420, 0.00001830, -0.00001198, -0.00001469, 0.00001083, -0.00001205, -0.00000613, 0.00000652, -0.00000109, -0.00000577, -0.00001028, 0.00000319, 0.00000145, -0.00000579, -0.00000820, 0.00000159, 0.00000131, -0.00000575, 0.00000165, 0.00000022, 0.00000300, -0.00001556, -0.00000404, 0.00001041, -0.00001310, 0.00000183, 0.00000347, -0.00000711, 0.00000683, 0.00000578, -0.00000023, -0.00000309, -0.00000257, 0.00000215, },
		{ 0.00000139, 0.00000038, 0.00000150, 0.00000105, -0.00000306, -0.00000158, 0.00000610, 0.00000806, -0.00000108, -0.00000023, 0.00000509, -0.00000296, -0.00000559, -0.00000225, -0.00000073, 0.00000334, 0.00000143, -0.00000198, -0.00000209, 0.00000181, -0.00000226, -0.00000004, 0.00001303, 0.00000143, -0.00000356, 0.00000159, -0.00000288, 0.00000173, -0.00000864, -0.00000338, 0.00001803, -0.00000209, 0.00000144, 0.00000292, -0.00000807, 0.00000701, 0.00000020, -0.00000645, 0.00000624, 0.00001649, -0.00002598, 0.00000143, 0.00001181, -0.00003269, 0.00002232, -0.00000166, -0.00001438, 0.00000621, -0.00000133, 0.00000740, -0.00000542, -0.00002422, 0.00001024, 0.00000402, -0.00002673, 0.00003701, -0.00004034, 0.00003357, -0.00001699, -0.00002779, 0.00006343, -0.00010415, 0.00011361, -0.00008563, -0.00000293, 0.00007282, -0.00016190, 0.00018483, -0.00016980, 0.00006730, 0.00001436, -0.00016281, 0.00019333, -0.00030646, 0.00017829, -0.00041049, -0.00047531, 0.00135441, 0.00063554, -0.00049016, -0.00022316, 0.00000215, -0.00005562, -0.00005446, 0.00003877, -0.00007070, 0.00002740, -0.00001658, -0.00003102, 0.00003739, -0.00004187, 0.00002688, -0.00001071, -0.00002678, 0.00002262, -0.00002145, -0.00000624, 0.00000801, -0.00001605, 0.00000068, 0.00000436, -0.00000163, -0.00000651, -0.00001066, 0.00000620, -0.00000117, -0.00000671, -0.00000662, 0.00000207, 0.00000062, -0.00000537, 0.00000248, 0.00000131, -0.00000214, -0.00001604, 0.00000356, 0.00000485, -0.00001339, 0.00000675, -0.00000138, -0.00000498, 0.00000952, 0.00000296, -0.00000034, -0.00000300, -0.00000191, 0.00000245, 0.00000094, },
		{ -0.00000028, 0.00000136, 0.00000066, -0.00000399, 0.00000067, 0.00000803, 0.00000570, -0.00000218, 0.00000110, 0.00000426, -0.00000396, -0.00000505, -0.00000200, 0.00000007, 0.00000355, 0.00000006, -0.00000238, -0.00000044, 0.00000112, -0.00000315, 0.00000418, 0.00001214, -0.00000221, -0.00000212, 0.00000131, -0.00000243, 0.00000087, -0.00000973, 0.00000226, 0.00001518, -0.00000447, 0.00000534, -0.00000214, -0.00000490, 0.00000985, -0.00000669, -0.00000282, 0.00001295, 0.00000445, -0.00002514, 0.00001325, -0.00000368, -0.00002565, 0.00002826, -0.00001220, -0.00000651, 0.00000265, -0.00000082, 0.00001101, -0.00001625, -0.00001708, 0.00001543, -0.00001073, -0.00000802, 0.00002188, -0.00003392, 0.00004284, -0.00004929, 0.00001874, 0.00001782, -0.00007751, 0.00012314, -0.00013778, 0.00007720, -0.00001169, -0.00010051, 0.00017541, -0.00022501, 0.00017092, -0.00010994, -0.00006023, 0.00014242, -0.00034353, 0.00029496, -0.00068739, -0.00018913, 0.00164217, 0.00024105, -0.00058767, -0.00007936, -0.00001867, -0.00007279, -0.00001551, 0.00001059, -0.00006168, 0.00004141, -0.00004295, -0.00000483, 0.00002415, -0.00004168, 0.00003687, -0.00002791, -0.00001251, 0.00001943, -0.00002713, 0.00000302, 0.00000088, -0.00001528, 0.00000598, 0.00000074, -0.00000042, -0.00000898, -0.00000920, 0.00000840, -0.00000491, -0.00000575, -0.00000587, 0.00000249, 0.00000021, -0.00000530, 0.00000427, 0.00000042, -0.00000760, -0.00001241, 0.00000813, -0.00000209, -0.00000978, 0.00000887, -0.00000627, -0.00000048, 0.00000990, 0.00000027, 0.00000048, -0.00000335, -0.00000105, 0.00000242, 0.00000057, 0.00000217, },
		{ 0.00000174, -0.00000029, -0.00000429, 0.00000370, 0.00000848, 0.00000308, -0.00000229, 0.00000214, 0.00000316, -0.00000467, -0.00000444, -0.00000172, 0.00000088, 0.00000335, -0.00000143, -0.00000187, 0.00000083, -0.00000042, -0.00000230, 0.00000789, 0.00000935, -0.00000448, -0.00000048, 0.00000072, -0.00000185, -0.00000060, -0.00000900, 0.00000678, 0.00001068, -0.00000392, 0.00000701, -0.00000702, 0.00000136, 0.00000821, -0.00001230, 0.00000527, 0.00001314, -0.00000714, -0.00001685, 0.00001686, -0.00001759, -0.00001263, 0.00002655, -0.00001838, 0.00000089, -0.00000357, 0.00000473, 0.00000853, -0.00002426, -0.00000551, 0.00001117, -0.00001619, 0.00000609, 0.00000390, -0.00001572, 0.00003186, -0.00006018, 0.00005396, -0.00003418, -0.00002496, 0.00009337, -0.00015137, 0.00013509, -0.00010088, -0.00000370, 0.00010718, -0.00021872, 0.00023442, -0.00023345, 0.00008638, 0.00001056, -0.00028655, 0.00033530, -0.00097673, 0.00028123, 0.00175816, -0.00022833, -0.00053556, 0.00003433, -0.00007416, -0.00005336, 0.00000301, -0.00001850, -0.00003708, 0.00003611, -0.00005690, 0.00002063, 0.00000265, -0.00003042, 0.00003716, -0.00004041, 0.00000431, 0.00000928, -0.00002628, 0.00000918, -0.00000771, -0.00000992, 0.00000771, -0.00000213, 0.00000121, -0.00001267, -0.00000539, 0.00000834, -0.00000821, -0.00000346, -0.00000644, 0.00000367, -0.00000066, -0.00000488, 0.00000649, -0.00000310, -0.00001113, -0.00000614, 0.00000831, -0.00000764, -0.00000368, 0.00000730, -0.00000919, 0.00000503, 0.00000775, -0.00000118, 0.00000157, -0.00000410, 0.00000011, 0.00000202, 0.00000066, 0.00000258, -0.00000123, },
		{ -0.00000179, -0.00000342, 0.00000676, 0.00000739, 0.00000091, -0.00000173, 0.00000271, 0.00000196, -0.00000507, -0.00000377, -0.00000145, 0.00000167, 0.00000265, -0.00000267, -0.00000055, 0.00000116, -0.00000194, 0.00000017, 0.00001008, 0.00000561, -0.00000524, 0.00000085, 0.00000013, -0.00000146, -0.00000211, -0.00000673, 0.00000918, 0.00000614, -0.00000125, 0.00000568, -0.00000952, 0.00000849, 0.00000189, -0.00001341, 0.00001427, 0.00000657, -0.00001360, -0.00000572, 0.00001214, -0.00002528, 0.00000144, 0.00001900, -0.00001879, 0.00000454, -0.00000861, 0.00001242, -0.00000001, -0.00002552, 0.00000446, 0.00000190, -0.00001256, 0.00001096, -0.00000837, 0.00000415, 0.00000756, -0.00004808, 0.00006604, -0.00007387, 0.00003485, 0.00003441, -0.00012185, 0.00015106, -0.00016505, 0.00009877, -0.00000288, -0.00014513, 0.00022812, -0.00031151, 0.00023339, -0.00017875, -0.00012239, 0.00023917, -0.00118642, 0.00089366, 0.00162831, -0.00067063, -0.00034909, 0.00007175, -0.00012577, -0.00000939, -0.00000671, -0.00003225, -0.00001069, 0.00001641, -0.00005407, 0.00003630, -0.00001904, -0.00001206, 0.00002705, -0.00004365, 0.00001790, -0.00000417, -0.00001904, 0.00000948, -0.00001398, -0.00000228, 0.00000562, -0.00000263, 0.00000147, -0.00001605, 0.00000001, 0.00000541, -0.00000946, -0.00000124, -0.00000787, 0.00000589, -0.00000259, -0.00000321, 0.00000787, -0.00000877, -0.00001096, 0.00000015, 0.00000454, -0.00000965, 0.00000245, 0.00000267, -0.00000883, 0.00000966, 0.00000380, -0.00000078, 0.00000210, -0.00000488, 0.00000151, 0.00000126, 0.00000117, 0.00000240, -0.00000202, 0.00000237, },
		{ -0.00000118, 0.00000900, 0.00000514, -0.00000036, -0.00000090, 0.00000284, 0.00000080, -0.00000517, -0.00000312, -0.00000119, 0.00000237, 0.00000145, -0.00000326, 0.00000111, 0.00000034, -0.00000257, 0.00000354, 0.00001026, 0.00000196, -0.00000482, 0.00000165, -0.00000030, -0.00000147, -0.00000302, -0.00000373, 0.00000930, 0.00000303, 0.00000193, 0.00000189, -0.00000822, 0.00001362, -0.00000717, -0.00000854, 0.00001994, -0.00000401, -0.00001274, 0.00000301, 0.00000229, -0.00002511, 0.00001201, 0.00000934, -0.00001475, 0.00000342, -0.00000947, 0.00001812, -0.00001143, -0.00001962, 0.00000845, -0.00000650, -0.00000416, 0.00000713, -0.00001053, 0.00001630, -0.00001861, -0.00002033, 0.00005284, -0.00008825, 0.00008150, -0.00003365, -0.00005985, 0.00012010, -0.00018239, 0.00017322, -0.00012134, -0.00002236, 0.00014407, -0.00030870, 0.00032716, -0.00037275, 0.00012256, -0.00002425, -0.00121366, 0.00154903, 0.00123100, -0.00097849, -0.00008797, 0.00001817, -0.00013694, 0.00003282, -0.00003666, -0.00002333, 0.00000348, -0.00000616, -0.00003797, 0.00003789, -0.00003313, 0.00000634, 0.00001032, -0.00003674, 0.00002366, -0.00001564, -0.00000859, 0.00000381, -0.00001506, 0.00000429, 0.00000139, -0.00000062, -0.00000077, -0.00001731, 0.00000520, 0.00000034, -0.00000785, -0.00000063, -0.00000895, 0.00000857, -0.00000555, 0.00000028, 0.00000682, -0.00001480, -0.00000667, 0.00000373, -0.00000110, -0.00000744, 0.00000616, -0.00000310, -0.00000508, 0.00001176, -0.00000048, 0.00000130, 0.00000150, -0.00000516, 0.00000286, 0.00000028, 0.00000198, 0.00000156, -0.00000227, 0.00000280, -0.00000350, },
		{ 0.00000975, 0.00000245, -0.00000064, -0.00000020, 0.00000263, -0.00000023, -0.00000498, -0.00000258, -0.00000091, 0.00000284, -0.00000007, -0.00000294, 0.00000249, -0.00000130, -0.00000169, 0.00000673, 0.00000859, -0.00000079, -0.00000379, 0.00000201, -0.00000059, -0.00000182, -0.00000292, -0.00000105, 0.00000778, 0.00000202, 0.00000393, -0.00000267, -0.00000327, 0.00001437, -0.00001565, 0.00000131, 0.00001920, -0.00001406, -0.00000565, 0.00000564, -0.00000783, -0.00001868, 0.00001673, 0.00000132, -0.00000931, -0.00000097, -0.00000549, 0.00001864, -0.00002116, -0.00000954, 0.00000555, -0.00000967, 0.00000330, -0.00000077, -0.00000370, 0.00001598, -0.00003588, 0.00001060, 0.00002190, -0.00007466, 0.00010058, -0.00008860, 0.00001311, 0.00005373, -0.00014731, 0.00019457, -0.00020972, 0.00011218, 0.00000175, -0.00021415, 0.00032355, -0.00050407, 0.00038199, -0.00043500, -0.00097841, 0.00210940, 0.00061336, -0.00107303, 0.00016221, -0.00010345, -0.00008991, 0.00004554, -0.00006661, 0.00000264, -0.00000153, -0.00001934, -0.00001808, 0.00002788, -0.00003591, 0.00001800, -0.00000642, -0.00002308, 0.00002033, -0.00002075, 0.00000032, -0.00000506, -0.00001042, 0.00000700, -0.00000235, 0.00000244, -0.00000525, -0.00001542, 0.00000816, -0.00000501, -0.00000388, -0.00000248, -0.00000835, 0.00001050, -0.00000865, 0.00000523, 0.00000225, -0.00001868, 0.00000037, 0.00000293, -0.00000567, -0.00000220, 0.00000598, -0.00000761, 0.00000079, 0.00001060, -0.00000350, 0.00000417, -0.00000030, -0.00000450, 0.00000377, -0.00000068, 0.00000286, 0.00000026, -0.00000185, 0.00000262, -0.00000481, 0.00000217, },
		{ 0.00000011, -0.00000015, 0.00000013, 0.00000225, -0.00000107, -0.00000456, -0.00000221, -0.00000055, 0.00000296, -0.00000159, -0.00000173, 0.00000299, -0.00000304, 0.00000074, 0.00000872, 0.00000579, -0.00000229, -0.00000271, 0.00000215, -0.00000091, -0.00000219, -0.00000186, 0.00000046, 0.00000581, 0.00000280, 0.00000372, -0.00000588, 0.00000360, 0.00001000, -0.00002011, 0.00001289, 0.00001151, -0.00001922, 0.00000387, 0.00000151, -0.00001393, -0.00000971, 0.00001605, -0.00000296, -0.00000548, -0.00000565, 0.00000146, 0.00001331, -0.00002551, 0.00000008, -0.00000166, -0.00000661, 0.00000566, -0.00000686, 0.00000663, 0.00000470, -0.00003854, 0.00003267, -0.00001347, -0.00004090, 0.00008868, -0.00011444, 0.00007303, -0.00002456, -0.00007269, 0.00015594, -0.00023866, 0.00021448, -0.00015657, -0.00004908, 0.00020682, -0.00051427, 0.00056528, -0.00091532, -0.00045621, 0.00243311, -0.00011339, -0.00093094, 0.00031874, -0.00024103, 0.00000546, 0.00001342, -0.00007393, 0.00002930, -0.00002168, -0.00001650, -0.00000500, 0.00001402, -0.00002946, 0.00001986, -0.00001689, -0.00000882, 0.00001074, -0.00001817, 0.00000358, -0.00001271, -0.00000226, 0.00000506, -0.00000335, 0.00000432, -0.00001031, -0.00001082, 0.00000767, -0.00000844, 0.00000076, -0.00000644, -0.00000535, 0.00001046, -0.00001044, 0.00001018, -0.00000563, -0.00001815, 0.00000737, -0.00000202, -0.00000663, 0.00000354, 0.00000214, -0.00000892, 0.00000667, 0.00000667, -0.00000415, 0.00000662, -0.00000279, -0.00000280, 0.00000387, -0.00000135, 0.00000354, -0.00000119, -0.00000088, 0.00000165, -0.00000507, 0.00000595, 0.00000873, },
		{ 0.00000065, 0.00000004, 0.00000188, -0.00000168, -0.00000401, -0.00000204, -0.00000012, 0.00000265, -0.00000274, 0.00000005, 0.00000234, -0.00000403, 0.00000417, 0.00000890, 0.00000283, -0.00000268, -0.00000188, 0.00000228, -0.00000143, -0.00000216, -0.00000036, 0.00000050, 0.00000457, 0.00000423, 0.00000140, -0.00000615, 0.00000974, 0.00000176, -0.00001851, 0.00002200, -0.00000082, -0.00001714, 0.00001108, -0.00000700, -0.00001399, -0.00000215, 0.00001245, -0.00000370, -0.00000479, -0.00000784, 0.00000808, 0.00000418, -0.00002337, 0.00000528, -0.00000884, 0.00000027, 0.00000222, -0.00000715, 0.00001384, -0.00001121, -0.00002778, 0.00003911, -0.00004013, -0.00000120, 0.00005374, -0.00010693, 0.00010254, -0.00008855, 0.00001519, 0.00007131, -0.00019906, 0.00024995, -0.00027990, 0.00013767, -0.00000018, -0.00037805, 0.00058598, -0.00134262, 0.00030186, 0.00241668, -0.00080002, -0.00059481, 0.00033044, -0.00033362, 0.00011469, -0.00005783, -0.00004507, 0.00003750, -0.00004347, -0.00000036, -0.00000476, 0.00000515, -0.00002045, 0.00001409, -0.00001825, 0.00000008, 0.00000041, -0.00001032, -0.00000014, -0.00001515, 0.00000551, 0.00000005, -0.00000104, 0.00000329, -0.00001366, -0.00000540, 0.00000409, -0.00000849, 0.00000393, -0.00001100, -0.00000039, 0.00000796, -0.00000965, 0.00001302, -0.00001505, -0.00001235, 0.00001113, -0.00000895, -0.00000302, 0.00000692, -0.00000353, -0.00000642, 0.00001042, 0.00000150, -0.00000230, 0.00000762, -0.00000510, -0.00000040, 0.00000305, -0.00000151, 0.00000377, -0.00000242, 0.00000026, 0.00000006, -0.00000387, 0.00000926, 0.00000619, -0.00000126, },
		{ -0.00000034, 0.00000162, -0.00000209, -0.00000344, -0.00000204, 0.00000036, 0.00000194, -0.00000322, 0.00000184, 0.00000068, -0.00000358, 0.00000762, 0.00000726, 0.00000057, -0.00000244, -0.00000134, 0.00000244, -0.00000217, -0.00000138, 0.00000083, -0.00000052, 0.00000473, 0.00000492, -0.00000183, -0.00000316, 0.00001268, -0.00000746, -0.00001110, 0.00002519, -0.00001352, -0.00000854, 0.00001239, -0.00001572, -0.00000888, 0.00000154, 0.00000894, -0.00000280, -0.00000668, -0.00000638, 0.00001134, -0.00000506, -0.00001655, 0.00000457, -0.00001217, 0.00000683, -0.00000436, -0.00000136, 0.00001349, -0.00002400, -0.00001027, 0.00003055, -0.00005040, 0.00003016, 0.00001079, -0.00007401, 0.00009686, -0.00011928, 0.00008681, -0.00002975, -0.00010565, 0.00020660, -0.00032695, 0.00028533, -0.00023962, -0.00011837, 0.00039398, -0.00158080, 0.00117127, 0.00202965, -0.00129985, -0.00016187, 0.00019569, -0.00033546, 0.00019246, -0.00014213, 0.00001648, 0.00001509, -0.00005034, 0.00001794, -0.00001546, 0.00000631, -0.00001641, 0.00000697, -0.00001264, 0.00000085, -0.00000519, -0.00000224, -0.00000863, -0.00001088, 0.00000921, -0.00000487, 0.00000312, -0.00000078, -0.00001370, -0.00000160, -0.00000069, -0.00000526, 0.00000409, -0.00001423, 0.00000509, 0.00000367, -0.00000599, 0.00001185, -0.00002304, -0.00000246, 0.00000946, -0.00001450, 0.00000405, 0.00000607, -0.00000835, -0.00000109, 0.00001078, -0.00000301, 0.00000125, 0.00000673, -0.00000637, 0.00000205, 0.00000148, -0.00000110, 0.00000347, -0.00000313, 0.00000113, -0.00000166, -0.00000118, 0.00001120, 0.00000281, -0.00000138, 0.00000125, },
		{ 0.00000149, -0.00000232, -0.00000294, -0.00000214, 0.00000080, 0.00000099, -0.00000288, 0.00000307, -0.00000141, -0.00000147, 0.00001006, 0.00000436, -0.00000055, -0.00000211, -0.00000089, 0.00000251, -0.00000292, 0.00000013, 0.00000100, -0.00000170, 0.00000612, 0.00000394, -0.00000423, 0.00000195, 0.00001117, -0.00001433, -0.00000037, 0.00002118, -0.00002213, 0.00000323, 0.00000682, -0.00002061, -0.00000156, 0.00000120, 0.00000752, -0.00000259, -0.00000915, -0.00000218, 0.00001003, -0.00001098, -0.00000872, -0.00000057, -0.00001019, 0.00000952, -0.00000984, 0.00000720, 0.00000540, -0.00002829, 0.00000544, 0.00001370, -0.00004443, 0.00004446, -0.00002475, -0.00003141, 0.00006440, -0.00011164, 0.00012065, -0.00011353, 0.00000854, 0.00009981, -0.00028185, 0.00034291, -0.00043634, 0.00019460, -0.00000021, -0.00152139, 0.00198020, 0.00132994, -0.00151077, 0.00024539, -0.00003718, -0.00023445, 0.00020161, -0.00020308, 0.00008991, -0.00003589, -0.00003126, 0.00002418, -0.00002783, 0.00001566, -0.00002137, 0.00000543, -0.00000605, -0.00000463, -0.00000348, 0.00000098, -0.00001715, -0.00000165, 0.00000721, -0.00000669, 0.00000641, -0.00000621, -0.00001046, -0.00000121, -0.00000410, -0.00000039, 0.00000111, -0.00001470, 0.00000927, -0.00000079, -0.00000053, 0.00000606, -0.00002658, 0.00000845, 0.00000221, -0.00001555, 0.00001174, 0.00000095, -0.00000996, 0.00000488, 0.00000786, -0.00000531, 0.00000509, 0.00000433, -0.00000614, 0.00000380, -0.00000040, -0.00000022, 0.00000268, -0.00000324, 0.00000141, -0.00000284, 0.00000251, 0.00001121, -0.00000050, -0.00000042, 0.00000128, -0.00000073, },
		{ -0.00000241, -0.00000261, -0.00000228, 0.00000114, 0.00000005, -0.00000188, 0.00000334, -0.00000315, 0.00000195, 0.00001072, 0.00000113, -0.00000054, -0.00000206, -0.00000030, 0.00000232, -0.00000335, 0.00000202, -0.00000011, -0.00000206, 0.00000784, 0.00000125, -0.00000431, 0.00000704, 0.00000579, -0.00001646, 0.00000989, 0.00001142, -0.00002381, 0.00001356, -0.00000371, -0.00001950, 0.00000443, -0.00000136, 0.00000839, -0.00000447, -0.00000996, 0.00000235, 0.00000519, -0.00001197, -0.00000347, -0.00000678, -0.00000430, 0.00000721, -0.00001082, 0.00001368, -0.00000660, -0.00002333, 0.00001312, -0.00000249, -0.00002899, 0.00004136, -0.00004323, 0.00000458, 0.00002195, -0.00007514, 0.00011073, -0.00015485, 0.00010489, -0.00003336, -0.00016071, 0.00028846, -0.00052386, 0.00046240, -0.00051932, -0.00112099, 0.00255339, 0.00045319, -0.00140486, 0.00051715, -0.00028714, -0.00005764, 0.00012935, -0.00020923, 0.00014474, -0.00009829, 0.00001282, 0.00000854, -0.00002983, 0.00002499, -0.00003314, 0.00001308, -0.00000514, -0.00001077, 0.00000382, -0.00000312, -0.00002076, 0.00000828, 0.00000079, -0.00000414, 0.00000639, -0.00001026, -0.00000582, -0.00000426, -0.00000414, 0.00000362, -0.00000364, -0.00001228, 0.00001094, -0.00000356, 0.00000448, -0.00000312, -0.00002386, 0.00001651, -0.00000837, -0.00001059, 0.00001662, -0.00000654, -0.00000735, 0.00000908, 0.00000305, -0.00000483, 0.00000783, 0.00000139, -0.00000452, 0.00000436, -0.00000207, 0.00000083, 0.00000161, -0.00000289, 0.00000100, -0.00000289, 0.00000636, 0.00000927, -0.00000288, 0.00000113, 0.00000062, -0.00000088, 0.00000142, },
		{ -0.00000250, -0.00000238, 0.00000134, -0.00000062, -0.00000056, 0.00000258, -0.00000379, 0.00000588, 0.00000937, -0.00000150, 0.00000018, -0.00000238, 0.00000055, 0.00000175, -0.00000311, 0.00000365, -0.00000221, -0.00000100, 0.00000871, -0.00000218, -0.00000161, 0.00000986, -0.00000133, -0.00001334, 0.00001617, -0.00000056, -0.00001838, 0.00001856, -0.00001532, -0.00001281, 0.00000661, -0.00000360, 0.00001019, -0.00000822, -0.00000795, 0.00000472, -0.00000049, -0.00000885, -0.00000268, -0.00001061, 0.00000230, 0.00000153, -0.00000653, 0.00001428, -0.00001684, -0.00001292, 0.00001140, -0.00001122, -0.00001346, 0.00002764, -0.00004377, 0.00002370, -0.00001305, -0.00002858, 0.00006765, -0.00014600, 0.00015367, -0.00014819, -0.00000598, 0.00013838, -0.00046922, 0.00058961, -0.00103545, -0.00042179, 0.00276050, -0.00042179, -0.00103545, 0.00058961, -0.00046922, 0.00013838, -0.00000598, -0.00014819, 0.00015367, -0.00014600, 0.00006765, -0.00002858, -0.00001305, 0.00002370, -0.00004377, 0.00002764, -0.00001346, -0.00001122, 0.00001140, -0.00001292, -0.00001684, 0.00001428, -0.00000653, 0.00000153, 0.00000230, -0.00001061, -0.00000268, -0.00000885, -0.00000049, 0.00000472, -0.00000795, -0.00000822, 0.00001019, -0.00000360, 0.00000661, -0.00001281, -0.00001532, 0.00001856, -0.00001838, -0.00000056, 0.00001617, -0.00001334, -0.00000133, 0.00000986, -0.00000161, -0.00000218, 0.00000871, -0.00000100, -0.00000221, 0.00000365, -0.00000311, 0.00000175, 0.00000055, -0.00000238, 0.00000018, -0.00000150, 0.00000937, 0.00000588, -0.00000379, 0.00000258, -0.00000056, -0.00000062, 0.00000134, -0.00000238, },
		{ -0.00000241, 0.00000142, -0.00000088, 0.00000062, 0.00000113, -0.00000288, 0.00000927, 0.00000636, -0.00000289, 0.00000100, -0.00000289, 0.00000161, 0.00000083, -0.00000207, 0.00000436, -0.00000452, 0.00000139, 0.00000783, -0.00000483, 0.00000305, 0.00000908, -0.00000735, -0.00000654, 0.00001662, -0.00001059, -0.00000837, 0.00001651, -0.00002386, -0.00000312, 0.00000448, -0.00000356, 0.00001094, -0.00001228, -0.00000364, 0.00000362, -0.00000414, -0.00000426, -0.00000582, -0.00001026, 0.00000639, -0.00000414, 0.00000079, 0.00000828, -0.00002076, -0.00000312, 0.00000382, -0.00001077, -0.00000514, 0.00001308, -0.00003314, 0.00002499, -0.00002983, 0.00000854, 0.00001282, -0.00009829, 0.00014474, -0.00020923, 0.00012935, -0.00005764, -0.00028714, 0.00051715, -0.00140486, 0.00045319, 0.00255339, -0.00112099, -0.00051932, 0.00046240, -0.00052386, 0.00028846, -0.00016071, -0.00003336, 0.00010489, -0.00015485, 0.00011073, -0.00007514, 0.00002195, 0.00000458, -0.00004323, 0.00004136, -0.00002899, -0.00000249, 0.00001312, -0.00002333, -0.00000660, 0.00001368, -0.00001082, 0.00000721, -0.00000430, -0.00000678, -0.00000347, -0.00001197, 0.00000519, 0.00000235, -0.00000996, -0.00000447, 0.00000839, -0.00000136, 0.00000443, -0.00001950, -0.00000371, 0.00001356, -0.00002381, 0.00001142, 0.00000989, -0.00001646, 0.00000579, 0.00000704, -0.00000431, 0.00000125, 0.00000784, -0.00000206, -0.00000011, 0.00000202, -0.00000335, 0.00000232, -0.00000030, -0.00000206, -0.00000054, 0.00000113, 0.00001072, 0.00000195, -0.00000315, 0.00000334, -0.00000188, 0.00000005, 0.00000114, -0.00000228, -0.00000261, },
		{ 0.00000149, -0.00000073, 0.00000128, -0.00000042, -0.00000050, 0.00001121, 0.00000251, -0.00000284, 0.00000141, -0.00000324, 0.00000268, -0.00000022, -0.00000040, 0.00000380, -0.00000614, 0.00000433, 0.00000509, -0.00000531, 0.00000786, 0.00000488, -0.00000996, 0.00000095, 0.00001174, -0.00001555, 0.00000221, 0.00000845, -0.00002658, 0.00000606, -0.00000053, -0.00000079, 0.00000927, -0.00001470, 0.00000111, -0.00000039, -0.00000410, -0.00000121, -0.00001046, -0.00000621, 0.00000641, -0.00000669, 0.00000721, -0.00000165, -0.00001715, 0.00000098, -0.00000348, -0.00000463, -0.00000605, 0.00000543, -0.00002137, 0.00001566, -0.00002783, 0.00002418, -0.00003126, -0.00003589, 0.00008991, -0.00020308, 0.00020161, -0.00023445, -0.00003718, 0.00024539, -0.00151077, 0.00132994, 0.00198020, -0.00152139, -0.00000021, 0.00019460, -0.00043634, 0.00034291, -0.00028185, 0.00009981, 0.00000854, -0.00011353, 0.00012065, -0.00011164, 0.00006440, -0.00003141, -0.00002475, 0.00004446, -0.00004443, 0.00001370, 0.00000544, -0.00002829, 0.00000540, 0.00000720, -0.00000984, 0.00000952, -0.00001019, -0.00000057, -0.00000872, -0.00001098, 0.00001003, -0.00000218, -0.00000915, -0.00000259, 0.00000752, 0.00000120, -0.00000156, -0.00002061, 0.00000682, 0.00000323, -0.00002213, 0.00002118, -0.00000037, -0.00001433, 0.00001117, 0.00000195, -0.00000423, 0.00000394, 0.00000612, -0.00000170, 0.00000100, 0.00000013, -0.00000292, 0.00000251, -0.00000089, -0.00000211, -0.00000055, 0.00000436, 0.00001006, -0.00000147, -0.00000141, 0.00000307, -0.00000288, 0.00000099, 0.00000080, -0.00000214, -0.00000294, -0.00000232, },
		{ -0.00000034, 0.00000125, -0.00000138, 0.00000281, 0.00001120, -0.00000118, -0.00000166, 0.00000113, -0.00000313, 0.00000347, -0.00000110, 0.00000148, 0.00000205, -0.00000637, 0.00000673, 0.00000125, -0.00000301, 0.00001078, -0.00000109, -0.00000835, 0.00000607, 0.00000405, -0.00001450, 0.00000946, -0.00000246, -0.00002304, 0.00001185, -0.00000599, 0.00000367, 0.00000509, -0.00001423, 0.00000409, -0.00000526, -0.00000069, -0.00000160, -0.00001370, -0.00000078, 0.00000312, -0.00000487, 0.00000921, -0.00001088, -0.00000863, -0.00000224, -0.00000519, 0.00000085, -0.00001264, 0.00000697, -0.00001641, 0.00000631, -0.00001546, 0.00001794, -0.00005034, 0.00001509, 0.00001648, -0.00014213, 0.00019246, -0.00033546, 0.00019569, -0.00016187, -0.00129985, 0.00202965, 0.00117127, -0.00158080, 0.00039398, -0.00011837, -0.00023962, 0.00028533, -0.00032695, 0.00020660, -0.00010565, -0.00002975, 0.00008681, -0.00011928, 0.00009686, -0.00007401, 0.00001079, 0.00003016, -0.00005040, 0.00003055, -0.00001027, -0.00002400, 0.00001349, -0.00000136, -0.00000436, 0.00000683, -0.00001217, 0.00000457, -0.00001655, -0.00000506, 0.00001134, -0.00000638, -0.00000668, -0.00000280, 0.00000894, 0.00000154, -0.00000888, -0.00001572, 0.00001239, -0.00000854, -0.00001352, 0.00002519, -0.00001110, -0.00000746, 0.00001268, -0.00000316, -0.00000183, 0.00000492, 0.00000473, -0.00000052, 0.00000083, -0.00000138, -0.00000217, 0.00000244, -0.00000134, -0.00000244, 0.00000057, 0.00000726, 0.00000762, -0.00000358, 0.00000068, 0.00000184, -0.00000322, 0.00000194, 0.00000036, -0.00000204, -0.00000344, -0.00000209, 0.00000162, },
		{ 0.00000065, -0.00000126, 0.00000619, 0.00000926, -0.00000387, 0.00000006, 0.00000026, -0.00000242, 0.00000377, -0.00000151, 0.00000305, -0.00000040, -0.00000510, 0.00000762, -0.00000230, 0.00000150, 0.00001042, -0.00000642, -0.00000353, 0.00000692, -0.00000302, -0.00000895, 0.00001113, -0.00001235, -0.00001505, 0.00001302, -0.00000965, 0.00000796, -0.00000039, -0.00001100, 0.00000393, -0.00000849, 0.00000409, -0.00000540, -0.00001366, 0.00000329, -0.00000104, 0.00000005, 0.00000551, -0.00001515, -0.00000014, -0.00001032, 0.00000041, 0.00000008, -0.00001825, 0.00001409, -0.00002045, 0.00000515, -0.00000476, -0.00000036, -0.00004347, 0.00003750, -0.00004507, -0.00005783, 0.00011469, -0.00033362, 0.00033044, -0.00059481, -0.00080002, 0.00241668, 0.00030186, -0.00134262, 0.00058598, -0.00037805, -0.00000018, 0.00013767, -0.00027990, 0.00024995, -0.00019906, 0.00007131, 0.00001519, -0.00008855, 0.00010254, -0.00010693, 0.00005374, -0.00000120, -0.00004013, 0.00003911, -0.00002778, -0.00001121, 0.00001384, -0.00000715, 0.00000222, 0.00000027, -0.00000884, 0.00000528, -0.00002337, 0.00000418, 0.00000808, -0.00000784, -0.00000479, -0.00000370, 0.00001245, -0.00000215, -0.00001399, -0.00000700, 0.00001108, -0.00001714, -0.00000082, 0.00002200, -0.00001851, 0.00000176, 0.00000974, -0.00000615, 0.00000140, 0.00000423, 0.00000457, 0.00000050, -0.00000036, -0.00000216, -0.00000143, 0.00000228, -0.00000188, -0.00000268, 0.00000283, 0.00000890, 0.00000417, -0.00000403, 0.00000234, 0.00000005, -0.00000274, 0.00000265, -0.00000012, -0.00000204, -0.00000401, -0.00000168, 0.00000188, 0.00000004, },
		{ 0.00000011, 0.00000873, 0.00000595, -0.00000507, 0.00000165, -0.00000088, -0.00000119, 0.00000354, -0.00000135, 0.00000387, -0.00000280, -0.00000279, 0.00000662, -0.00000415, 0.00000667, 0.00000667, -0.00000892, 0.00000214, 0.00000354, -0.00000663, -0.00000202, 0.00000737, -0.00001815, -0.00000563, 0.00001018, -0.00001044, 0.00001046, -0.00000535, -0.00000644, 0.00000076, -0.00000844, 0.00000767, -0.00001082, -0.00001031, 0.00000432, -0.00000335, 0.00000506, -0.00000226, -0.00001271, 0.00000358, -0.00001817, 0.00001074, -0.00000882, -0.00001689, 0.00001986, -0.00002946, 0.00001402, -0.00000500, -0.00001650, -0.00002168, 0.00002930, -0.00007393, 0.00001342, 0.00000546, -0.00024103, 0.00031874, -0.00093094, -0.00011339, 0.00243311, -0.00045621, -0.00091532, 0.00056528, -0.00051427, 0.00020682, -0.00004908, -0.00015657, 0.00021448, -0.00023866, 0.00015594, -0.00007269, -0.00002456, 0.00007303, -0.00011444, 0.00008868, -0.00004090, -0.00001347, 0.00003267, -0.00003854, 0.00000470, 0.00000663, -0.00000686, 0.00000566, -0.00000661, -0.00000166, 0.00000008, -0.00002551, 0.00001331, 0.00000146, -0.00000565, -0.00000548, -0.00000296, 0.00001605, -0.00000971, -0.00001393, 0.00000151, 0.00000387, -0.00001922, 0.00001151, 0.00001289, -0.00002011, 0.00001000, 0.00000360, -0.00000588, 0.00000372, 0.00000280, 0.00000581, 0.00000046, -0.00000186, -0.00000219, -0.00000091, 0.00000215, -0.00000271, -0.00000229, 0.00000579, 0.00000872, 0.00000074, -0.00000304, 0.00000299, -0.00000173, -0.00000159, 0.00000296, -0.00000055, -0.00000221, -0.00000456, -0.00000107, 0.00000225, 0.00000013, -0.00000015, },
		{ 0.00000975, 0.00000217, -0.00000481, 0.00000262, -0.00000185, 0.00000026, 0.00000286, -0.00000068, 0.00000377, -0.00000450, -0.00000030, 0.00000417, -0.00000350, 0.00001060, 0.00000079, -0.00000761, 0.00000598, -0.00000220, -0.00000567, 0.00000293, 0.00000037, -0.00001868, 0.00000225, 0.00000523, -0.00000865, 0.00001050, -0.00000835, -0.00000248, -0.00000388, -0.00000501, 0.00000816, -0.00001542, -0.00000525, 0.00000244, -0.00000235, 0.00000700, -0.00001042, -0.00000506, 0.00000032, -0.00002075, 0.00002033, -0.00002308, -0.00000642, 0.00001800, -0.00003591, 0.00002788, -0.00001808, -0.00001934, -0.00000153, 0.00000264, -0.00006661, 0.00004554, -0.00008991, -0.00010345, 0.00016221, -0.00107303, 0.00061336, 0.00210940, -0.00097841, -0.00043500, 0.00038199, -0.00050407, 0.00032355, -0.00021415, 0.00000175, 0.00011218, -0.00020972, 0.00019457, -0.00014731, 0.00005373, 0.00001311, -0.00008860, 0.00010058, -0.00007466, 0.00002190, 0.00001060, -0.00003588, 0.00001598, -0.00000370, -0.00000077, 0.00000330, -0.00000967, 0.00000555, -0.00000954, -0.00002116, 0.00001864, -0.00000549, -0.00000097, -0.00000931, 0.00000132, 0.00001673, -0.00001868, -0.00000783, 0.00000564, -0.00000565, -0.00001406, 0.00001920, 0.00000131, -0.00001565, 0.00001437, -0.00000327, -0.00000267, 0.00000393, 0.00000202, 0.00000778, -0.00000105, -0.00000292, -0.00000182, -0.00000059, 0.00000201, -0.00000379, -0.00000079, 0.00000859, 0.00000673, -0.00000169, -0.00000130, 0.00000249, -0.00000294, -0.00000007, 0.00000284, -0.00000091, -0.00000258, -0.00000498, -0.00000023, 0.00000263, -0.00000020, -0.00000064, 0.00000245, },
		{ -0.00000118, -0.00000350, 0.00000280, -0.00000227, 0.00000156, 0.00000198, 0.00000028, 0.00000286, -0.00000516, 0.00000150, 0.00000130, -0.00000048, 0.00001176, -0.00000508, -0.00000310, 0.00000616, -0.00000744, -0.00000110, 0.00000373, -0.00000667, -0.00001480, 0.00000682, 0.00000028, -0.00000555, 0.00000857, -0.00000895, -0.00000063, -0.00000785, 0.00000034, 0.00000520, -0.00001731, -0.00000077, -0.00000062, 0.00000139, 0.00000429, -0.00001506, 0.00000381, -0.00000859, -0.00001564, 0.00002366, -0.00003674, 0.00001032, 0.00000634, -0.00003313, 0.00003789, -0.00003797, -0.00000616, 0.00000348, -0.00002333, -0.00003666, 0.00003282, -0.00013694, 0.00001817, -0.00008797, -0.00097849, 0.00123100, 0.00154903, -0.00121366, -0.00002425, 0.00012256, -0.00037275, 0.00032716, -0.00030870, 0.00014407, -0.00002236, -0.00012134, 0.00017322, -0.00018239, 0.00012010, -0.00005985, -0.00003365, 0.00008150, -0.00008825, 0.00005284, -0.00002033, -0.00001861, 0.00001630, -0.00001053, 0.00000713, -0.00000416, -0.00000650, 0.00000845, -0.00001962, -0.00001143, 0.00001812, -0.00000947, 0.00000342, -0.00001475, 0.00000934, 0.00001201, -0.00002511, 0.00000229, 0.00000301, -0.00001274, -0.00000401, 0.00001994, -0.00000854, -0.00000717, 0.00001362, -0.00000822, 0.00000189, 0.00000193, 0.00000303, 0.00000930, -0.00000373, -0.00000302, -0.00000147, -0.00000030, 0.00000165, -0.00000482, 0.00000196, 0.00001026, 0.00000354, -0.00000257, 0.00000034, 0.00000111, -0.00000326, 0.00000145, 0.00000237, -0.00000119, -0.00000312, -0.00000517, 0.00000080, 0.00000284, -0.00000090, -0.00000036, 0.00000514, 0.00000900, },
		{ -0.00000179, 0.00000237, -0.00000202, 0.00000240, 0.00000117, 0.00000126, 0.00000151, -0.00000488, 0.00000210, -0.00000078, 0.00000380, 0.00000966, -0.00000883, 0.00000267, 0.00000245, -0.00000965, 0.00000454, 0.00000015, -0.00001096, -0.00000877, 0.00000787, -0.00000321, -0.00000259, 0.00000589, -0.00000787, -0.00000124, -0.00000946, 0.00000541, 0.00000001, -0.00001605, 0.00000147, -0.00000263, 0.00000562, -0.00000228, -0.00001398, 0.00000948, -0.00001904, -0.00000417, 0.00001790, -0.00004365, 0.00002705, -0.00001206, -0.00001904, 0.00003630, -0.00005407, 0.00001641, -0.00001069, -0.00003225, -0.00000671, -0.00000939, -0.00012577, 0.00007175, -0.00034909, -0.00067063, 0.00162831, 0.00089366, -0.00118642, 0.00023917, -0.00012239, -0.00017875, 0.00023339, -0.00031151, 0.00022812, -0.00014513, -0.00000288, 0.00009877, -0.00016505, 0.00015106, -0.00012185, 0.00003441, 0.00003485, -0.00007387, 0.00006604, -0.00004808, 0.00000756, 0.00000415, -0.00000837, 0.00001096, -0.00001256, 0.00000190, 0.00000446, -0.00002552, -0.00000001, 0.00001242, -0.00000861, 0.00000454, -0.00001879, 0.00001900, 0.00000144, -0.00002528, 0.00001214, -0.00000572, -0.00001360, 0.00000657, 0.00001427, -0.00001341, 0.00000189, 0.00000849, -0.00000952, 0.00000568, -0.00000125, 0.00000614, 0.00000918, -0.00000673, -0.00000211, -0.00000146, 0.00000013, 0.00000085, -0.00000524, 0.00000561, 0.00001008, 0.00000017, -0.00000194, 0.00000116, -0.00000055, -0.00000267, 0.00000265, 0.00000167, -0.00000145, -0.00000377, -0.00000507, 0.00000196, 0.00000271, -0.00000173, 0.00000091, 0.00000739, 0.00000676, -0.00000342, },
		{ 0.00000174, -0.00000123, 0.00000258, 0.00000066, 0.00000202, 0.00000011, -0.00000410, 0.00000157, -0.00000118, 0.00000775, 0.00000503, -0.00000919, 0.00000730, -0.00000368, -0.00000764, 0.00000831, -0.00000614, -0.00001113, -0.00000310, 0.00000649, -0.00000488, -0.00000066, 0.00000367, -0.00000644, -0.00000346, -0.00000821, 0.00000834, -0.00000539, -0.00001267, 0.00000121, -0.00000213, 0.00000771, -0.00000992, -0.00000771, 0.00000918, -0.00002628, 0.00000928, 0.00000431, -0.00004041, 0.00003716, -0.00003042, 0.00000265, 0.00002063, -0.00005690, 0.00003611, -0.00003708, -0.00001850, 0.00000301, -0.00005336, -0.00007416, 0.00003433, -0.00053556, -0.00022833, 0.00175816, 0.00028123, -0.00097673, 0.00033530, -0.00028655, 0.00001056, 0.00008638, -0.00023345, 0.00023442, -0.00021872, 0.00010718, -0.00000370, -0.00010088, 0.00013509, -0.00015137, 0.00009337, -0.00002496, -0.00003418, 0.00005396, -0.00006018, 0.00003186, -0.00001572, 0.00000390, 0.00000609, -0.00001619, 0.00001117, -0.00000551, -0.00002426, 0.00000853, 0.00000473, -0.00000357, 0.00000089, -0.00001838, 0.00002655, -0.00001263, -0.00001759, 0.00001686, -0.00001685, -0.00000714, 0.00001314, 0.00000527, -0.00001230, 0.00000821, 0.00000136, -0.00000702, 0.00000701, -0.00000392, 0.00001068, 0.00000678, -0.00000900, -0.00000060, -0.00000185, 0.00000072, -0.00000048, -0.00000448, 0.00000935, 0.00000789, -0.00000230, -0.00000042, 0.00000083, -0.00000187, -0.00000143, 0.00000335, 0.00000088, -0.00000172, -0.00000444, -0.00000467, 0.00000316, 0.00000214, -0.00000229, 0.00000308, 0.00000848, 0.00000370, -0.00000429, -0.00000029, },
		{ -0.00000028, 0.00000217, 0.00000057, 0.00000242, -0.00000105, -0.00000335, 0.00000048, 0.00000027, 0.00000990, -0.00000048, -0.00000627, 0.00000887, -0.00000978, -0.00000209, 0.00000813, -0.00001241, -0.00000760, 0.00000042, 0.00000427, -0.00000530, 0.00000021, 0.00000249, -0.00000587, -0.00000575, -0.00000491, 0.00000840, -0.00000920, -0.00000898, -0.00000042, 0.00000074, 0.00000598, -0.00001528, 0.00000088, 0.00000302, -0.00002713, 0.00001943, -0.00001251, -0.00002791, 0.00003687, -0.00004168, 0.00002415, -0.00000483, -0.00004295, 0.00004141, -0.00006168, 0.00001059, -0.00001551, -0.00007279, -0.00001867, -0.00007936, -0.00058767, 0.00024105, 0.00164217, -0.00018913, -0.00068739, 0.00029496, -0.00034353, 0.00014242, -0.00006023, -0.00010994, 0.00017092, -0.00022501, 0.00017541, -0.00010051, -0.00001169, 0.00007720, -0.00013778, 0.00012314, -0.00007751, 0.00001782, 0.00001874, -0.00004929, 0.00004284, -0.00003392, 0.00002188, -0.00000802, -0.00001073, 0.00001543, -0.00001708, -0.00001625, 0.00001101, -0.00000082, 0.00000265, -0.00000651, -0.00001220, 0.00002826, -0.00002565, -0.00000368, 0.00001325, -0.00002514, 0.00000445, 0.00001295, -0.00000282, -0.00000669, 0.00000985, -0.00000490, -0.00000214, 0.00000534, -0.00000447, 0.00001518, 0.00000226, -0.00000973, 0.00000087, -0.00000243, 0.00000131, -0.00000212, -0.00000221, 0.00001214, 0.00000418, -0.00000315, 0.00000112, -0.00000044, -0.00000238, 0.00000006, 0.00000355, 0.00000007, -0.00000200, -0.00000505, -0.00000396, 0.00000426, 0.00000110, -0.00000218, 0.00000570, 0.00000803, 0.00000067, -0.00000399, 0.00000066, 0.00000136, },
		{ 0.00000139, 0.00000094, 0.00000245, -0.00000191, -0.00000300, -0.00000034, 0.00000296, 0.00000952, -0.00000498, -0.00000138, 0.00000675, -0.00001339, 0.00000485, 0.00000356, -0.00001604, -0.00000214, 0.00000131, 0.00000248, -0.00000537, 0.00000062, 0.00000207, -0.00000662, -0.00000671, -0.00000117, 0.00000620, -0.00001066, -0.00000651, -0.00000163, 0.00000436, 0.00000068, -0.00001605, 0.00000801, -0.00000624, -0.00002145, 0.00002262, -0.00002678, -0.00001071, 0.00002688, -0.00004187, 0.00003739, -0.00003102, -0.00001658, 0.00002740, -0.00007070, 0.00003877, -0.00005446, -0.00005562, 0.00000215, -0.00022316, -0.00049016, 0.00063554, 0.00135441, -0.00047531, -0.00041049, 0.00017829, -0.00030646, 0.00019333, -0.00016281, 0.00001436, 0.00006730, -0.00016980, 0.00018483, -0.00016190, 0.00007282, -0.00000293, -0.00008563, 0.00011361, -0.00010415, 0.00006343, -0.00002779, -0.00001699, 0.00003357, -0.00004034, 0.00003701, -0.00002673, 0.00000402, 0.00001024, -0.00002422, -0.00000542, 0.00000740, -0.00000133, 0.00000621, -0.00001438, -0.00000166, 0.00002232, -0.00003269, 0.00001181, 0.00000143, -0.00002598, 0.00001649, 0.00000624, -0.00000645, 0.00000020, 0.00000701, -0.00000807, 0.00000292, 0.00000144, -0.00000209, 0.00001803, -0.00000338, -0.00000864, 0.00000173, -0.00000288, 0.00000159, -0.00000356, 0.00000143, 0.00001303, -0.00000004, -0.00000226, 0.00000181, -0.00000209, -0.00000198, 0.00000143, 0.00000334, -0.00000073, -0.00000225, -0.00000559, -0.00000296, 0.00000509, -0.00000023, -0.00000108, 0.00000806, 0.00000610, -0.00000158, -0.00000306, 0.00000105, 0.00000150, 0.00000038, },
		{ 0.00000163, 0.00000215, -0.00000257, -0.00000309, -0.00000023, 0.00000578, 0.00000683, -0.00000711, 0.00000347, 0.00000183, -0.00001310, 0.00001041, -0.00000404, -0.00001556, 0.00000300, 0.00000022, 0.00000165, -0.00000575, 0.00000131, 0.00000159, -0.00000820, -0.00000579, 0.00000145, 0.00000319, -0.00001028, -0.00000577, -0.00000109, 0.00000652, -0.00000613, -0.00001205, 0.00001083, -0.00001469, -0.00001198, 0.00001830, -0.00003420, 0.00000523, 0.00001178, -0.00003191, 0.00003785, -0.00004837, 0.00001199, -0.00000172, -0.00005758, 0.00004918, -0.00009436, -0.00001000, -0.00003485, -0.00033758, -0.00027440, 0.00088732, 0.00099197, -0.00059037, -0.00020428, 0.00004863, -0.00021437, 0.00017089, -0.00020082, 0.00010230, -0.00003807, -0.00007833, 0.00013949, -0.00017228, 0.00012623, -0.00007851, -0.00001293, 0.00006870, -0.00009537, 0.00008537, -0.00006827, 0.00002608, 0.00000516, -0.00002893, 0.00004031, -0.00004171, 0.00002324, -0.00000472, -0.00002219, 0.00000244, 0.00000103, 0.00000322, 0.00000454, -0.00001876, 0.00000938, 0.00001011, -0.00003069, 0.00002300, -0.00001486, -0.00001758, 0.00002378, -0.00000385, -0.00000430, 0.00000496, 0.00000169, -0.00000751, 0.00000609, -0.00000295, 0.00000285, 0.00001804, -0.00000879, -0.00000609, 0.00000170, -0.00000297, 0.00000133, -0.00000413, 0.00000575, 0.00001159, -0.00000351, -0.00000016, 0.00000119, -0.00000340, -0.00000087, 0.00000245, 0.00000282, -0.00000148, -0.00000244, -0.00000606, -0.00000168, 0.00000550, -0.00000154, 0.00000103, 0.00000945, 0.00000320, -0.00000262, -0.00000210, 0.00000123, 0.00000215, 0.00000037, 0.00000065, },
		{ 0.00000158, -0.00000320, -0.00000331, 0.00000102, 0.00000759, 0.00000290, -0.00000650, 0.00000649, -0.00000399, -0.00000909, 0.00001239, -0.00001218, -0.00001110, 0.00000605, -0.00000156, 0.00000148, -0.00000651, 0.00000264, 0.00000026, -0.00000953, -0.00000345, 0.00000221, 0.00000083, -0.00000933, -0.00000605, 0.00000132, 0.00000564, -0.00001163, -0.00000524, 0.00000867, -0.00001921, -0.00000279, 0.00000909, -0.00003369, 0.00001533, -0.00000239, -0.00001692, 0.00002662, -0.00005124, 0.00003150, -0.00003461, -0.00002633, 0.00003298, -0.00011372, 0.00003908, -0.00012648, -0.00037264, -0.00000453, 0.00097718, 0.00064345, -0.00058446, -0.00008576, -0.00004867, -0.00011252, 0.00010438, -0.00017933, 0.00013579, -0.00011258, 0.00001582, 0.00006092, -0.00013464, 0.00013468, -0.00012567, 0.00005571, 0.00000486, -0.00005464, 0.00007481, -0.00008655, 0.00006391, -0.00003289, -0.00000095, 0.00002667, -0.00004446, 0.00003856, -0.00002473, -0.00001019, 0.00000259, -0.00000288, 0.00000964, -0.00000208, -0.00001709, 0.00001629, -0.00000401, -0.00002008, 0.00002526, -0.00002960, -0.00000192, 0.00002285, -0.00001268, 0.00000227, 0.00000546, -0.00000334, -0.00000418, 0.00000636, -0.00000593, 0.00000903, 0.00001495, -0.00001266, -0.00000288, 0.00000086, -0.00000270, 0.00000053, -0.00000329, 0.00000967, 0.00000807, -0.00000527, 0.00000217, -0.00000063, -0.00000382, 0.00000052, 0.00000307, 0.00000205, -0.00000211, -0.00000255, -0.00000648, -0.00000014, 0.00000536, -0.00000243, 0.00000386, 0.00000933, 0.00000012, -0.00000250, -0.00000154, 0.00000165, 0.00000296, -0.00000040, 0.00000032, 0.00000245, },
		{ -0.00000391, -0.00000321, 0.00000307, 0.00000767, -0.00000087, -0.00000384, 0.00000676, -0.00000856, -0.00000295, 0.00000995, -0.00001824, -0.00000413, 0.00000629, -0.00000284, 0.00000131, -0.00000720, 0.00000434, -0.00000221, -0.00000958, -0.00000089, 0.00000141, -0.00000017, -0.00000902, -0.00000604, 0.00000444, 0.00000161, -0.00001362, 0.00000133, 0.00000320, -0.00001887, 0.00000267, -0.00000079, -0.00002747, 0.00001826, -0.00001106, -0.00000355, 0.00000959, -0.00004046, 0.00003492, -0.00005818, 0.00001025, -0.00000613, -0.00009990, 0.00006106, -0.00024339, -0.00030629, 0.00024637, 0.00093167, 0.00036614, -0.00051944, -0.00003858, -0.00009600, -0.00003512, 0.00002938, -0.00012193, 0.00011864, -0.00013988, 0.00008356, -0.00002148, -0.00006755, 0.00010084, -0.00013202, 0.00009839, -0.00005530, 0.00000290, 0.00003498, -0.00007419, 0.00008142, -0.00006616, 0.00003479, -0.00000215, -0.00003040, 0.00004154, -0.00004144, 0.00000757, -0.00000610, -0.00000000, 0.00001310, -0.00001027, -0.00000988, 0.00001598, -0.00001443, -0.00000501, 0.00001735, -0.00003691, 0.00001589, 0.00001352, -0.00001615, 0.00000993, 0.00000168, -0.00000576, -0.00000011, 0.00000405, -0.00000624, 0.00001462, 0.00000944, -0.00001426, 0.00000010, -0.00000041, -0.00000227, -0.00000049, -0.00000090, 0.00001208, 0.00000340, -0.00000492, 0.00000368, -0.00000306, -0.00000315, 0.00000179, 0.00000335, 0.00000106, -0.00000249, -0.00000269, -0.00000679, 0.00000157, 0.00000466, -0.00000250, 0.00000683, 0.00000756, -0.00000230, -0.00000167, -0.00000148, 0.00000268, 0.00000336, -0.00000175, 0.00000066, 0.00000314, 0.00000077, },
		{ -0.00000239, 0.00000520, 0.00000601, -0.00000329, -0.00000049, 0.00000457, -0.00001040, 0.00000300, 0.00000387, -0.00002040, 0.00000319, 0.00000410, -0.00000299, 0.00000058, -0.00000725, 0.00000571, -0.00000537, -0.00000790, 0.00000069, 0.00000011, -0.00000006, -0.00000982, -0.00000458, 0.00000646, -0.00000408, -0.00001152, 0.00000504, -0.00000267, -0.00001518, 0.00000320, -0.00000761, -0.00001958, 0.00001604, -0.00001314, 0.00000334, -0.00000550, -0.00002259, 0.00002252, -0.00006376, 0.00003664, -0.00005327, -0.00005590, 0.00003392, -0.00034113, -0.00015175, 0.00042084, 0.00080593, 0.00017782, -0.00044613, -0.00002995, -0.00010085, 0.00000391, -0.00002653, -0.00005806, 0.00007081, -0.00012278, 0.00010980, -0.00008184, 0.00000322, 0.00004140, -0.00010020, 0.00010371, -0.00009159, 0.00005629, -0.00001990, -0.00003403, 0.00007049, -0.00008088, 0.00006504, -0.00003743, -0.00000164, 0.00002767, -0.00004637, 0.00002323, -0.00002002, 0.00001066, 0.00000959, -0.00001511, -0.00000087, 0.00000875, -0.00001688, 0.00000827, 0.00000225, -0.00003359, 0.00002963, -0.00000100, -0.00001242, 0.00001491, -0.00000440, -0.00000477, 0.00000265, 0.00000061, -0.00000374, 0.00001806, 0.00000288, -0.00001356, 0.00000219, -0.00000171, -0.00000197, -0.00000115, 0.00000260, 0.00001222, -0.00000110, -0.00000282, 0.00000367, -0.00000526, -0.00000158, 0.00000266, 0.00000338, -0.00000010, -0.00000252, -0.00000299, -0.00000689, 0.00000332, 0.00000351, -0.00000151, 0.00000917, 0.00000447, -0.00000348, -0.00000077, -0.00000161, 0.00000437, 0.00000284, -0.00000316, 0.00000170, 0.00000347, -0.00000022, -0.00000463, },
		{ 0.00000657, 0.00000323, -0.00000383, 0.00000215, 0.00000112, -0.00000921, 0.00000661, -0.00000388, -0.00001823, 0.00000887, 0.00000063, -0.00000210, -0.00000088, -0.00000629, 0.00000602, -0.00000829, -0.00000491, 0.00000066, -0.00000057, 0.00000021, -0.00001125, -0.00000150, 0.00000590, -0.00000906, -0.00000664, 0.00000492, -0.00000624, -0.00001105, 0.00000017, -0.00000983, -0.00001368, 0.00001252, -0.00001105, 0.00000282, -0.00001290, -0.00000658, 0.00000144, -0.00005098, 0.00004192, -0.00008924, -0.00000003, -0.00004493, -0.00037738, 0.00004847, 0.00049342, 0.00066038, 0.00006334, -0.00039143, -0.00002865, -0.00008462, 0.00001012, -0.00005191, -0.00001083, 0.00001772, -0.00007854, 0.00009634, -0.00010633, 0.00005483, -0.00002054, -0.00004562, 0.00007424, -0.00009357, 0.00008682, -0.00007005, 0.00002068, 0.00003343, -0.00006969, 0.00007736, -0.00006651, 0.00003304, -0.00000091, -0.00003477, 0.00002852, -0.00003198, 0.00002566, -0.00000159, -0.00001265, 0.00000452, -0.00000138, -0.00001066, 0.00001409, -0.00001399, -0.00002060, 0.00003455, -0.00001542, -0.00000271, 0.00001468, -0.00000984, -0.00000123, 0.00000296, -0.00000217, 0.00000064, 0.00001858, -0.00000323, -0.00001115, 0.00000314, -0.00000276, -0.00000196, -0.00000084, 0.00000630, 0.00001002, -0.00000420, 0.00000009, 0.00000202, -0.00000647, 0.00000040, 0.00000309, 0.00000319, -0.00000133, -0.00000216, -0.00000363, -0.00000660, 0.00000490, 0.00000220, 0.00000052, 0.00001018, 0.00000081, -0.00000328, -0.00000037, -0.00000137, 0.00000634, 0.00000110, -0.00000403, 0.00000323, 0.00000326, -0.00000135, -0.00000515, -0.00000076, },
		{ 0.00000034, -0.00000277, 0.00000321, -0.00000202, -0.00000595, 0.00000676, -0.00001087, -0.00001265, 0.00001174, -0.00000283, -0.00000070, -0.00000278, -0.00000445, 0.00000493, -0.00000998, -0.00000161, -0.00000077, 0.00000001, -0.00000038, -0.00001220, 0.00000232, 0.00000247, -0.00001129, -0.00000141, 0.00000194, -0.00000646, -0.00000926, -0.00000347, -0.00000853, -0.00001127, 0.00001106, -0.00000905, -0.00000200, -0.00001142, 0.00000061, -0.00001805, -0.00002769, 0.00002483, -0.00009959, 0.00004161, -0.00015582, -0.00032841, 0.00023886, 0.00047411, 0.00054004, -0.00000938, -0.00035759, -0.00001707, -0.00006905, 0.00000033, -0.00005133, 0.00001019, -0.00002045, -0.00002961, 0.00005801, -0.00009608, 0.00007510, -0.00006458, 0.00001048, 0.00002431, -0.00006379, 0.00008490, -0.00009774, 0.00007130, -0.00001776, -0.00003472, 0.00006531, -0.00007777, 0.00006121, -0.00003554, -0.00000837, 0.00001864, -0.00003433, 0.00003822, -0.00001684, -0.00000235, 0.00000193, -0.00000828, 0.00000064, 0.00001010, -0.00002472, -0.00000275, 0.00002932, -0.00002462, 0.00000937, 0.00000900, -0.00001212, 0.00000283, 0.00000100, -0.00000296, 0.00000532, 0.00001637, -0.00000775, -0.00000795, 0.00000308, -0.00000353, -0.00000214, 0.00000077, 0.00000906, 0.00000616, -0.00000518, 0.00000261, -0.00000076, -0.00000630, 0.00000224, 0.00000320, 0.00000274, -0.00000242, -0.00000153, -0.00000465, -0.00000577, 0.00000607, 0.00000115, 0.00000321, 0.00000945, -0.00000249, -0.00000205, -0.00000069, -0.00000025, 0.00000789, -0.00000167, -0.00000383, 0.00000478, 0.00000249, -0.00000249, -0.00000520, 0.00000144, 0.00000662, },
		{ -0.00000092, 0.00000267, -0.00000362, -0.00000229, 0.00000364, -0.00001518, -0.00000543, 0.00001166, -0.00000534, 0.00000054, -0.00000459, -0.00000227, 0.00000268, -0.00000993, 0.00000089, -0.00000271, 0.00000164, -0.00000225, -0.00001164, 0.00000532, -0.00000274, -0.00000996, 0.00000181, -0.00000165, -0.00000424, -0.00001095, -0.00000481, -0.00000650, -0.00001127, 0.00001271, -0.00001067, -0.00000592, -0.00000469, -0.00000280, -0.00002757, -0.00000593, -0.00000571, -0.00008153, 0.00004649, -0.00026326, -0.00019822, 0.00037065, 0.00039907, 0.00046350, -0.00007147, -0.00033113, 0.00000578, -0.00006635, -0.00000808, -0.00003883, 0.00000915, -0.00003528, 0.00000590, 0.00001463, -0.00006385, 0.00006511, -0.00008003, 0.00004947, -0.00002639, -0.00001576, 0.00005314, -0.00009375, 0.00010087, -0.00006591, 0.00001298, 0.00003133, -0.00006550, 0.00007162, -0.00006391, 0.00002464, -0.00000512, -0.00002255, 0.00004114, -0.00002941, 0.00001190, -0.00000937, -0.00000656, 0.00001046, -0.00000146, -0.00002571, 0.00001339, 0.00001652, -0.00002570, 0.00001939, -0.00000009, -0.00001021, 0.00000528, -0.00000188, -0.00000144, 0.00000878, 0.00001243, -0.00001021, -0.00000483, 0.00000236, -0.00000415, -0.00000210, 0.00000352, 0.00000995, 0.00000184, -0.00000409, 0.00000376, -0.00000377, -0.00000484, 0.00000354, 0.00000316, 0.00000199, -0.00000314, -0.00000089, -0.00000596, -0.00000433, 0.00000669, 0.00000075, 0.00000590, 0.00000702, -0.00000464, -0.00000049, -0.00000149, 0.00000196, 0.00000822, -0.00000485, -0.00000234, 0.00000587, 0.00000130, -0.00000351, -0.00000457, 0.00000367, 0.00000527, -0.00000177, },
		{ 0.00000123, -0.00000332, 0.00000005, -0.00000142, -0.00001589, 0.00000149, 0.00000930, -0.00000651, 0.00000121, -0.00000586, -0.00000040, -0.00000003, -0.00000831, 0.00000190, -0.00000412, 0.00000350, -0.00000504, -0.00000925, 0.00000607, -0.00000775, -0.00000589, 0.00000188, -0.00000351, -0.00000191, -0.00001515, -0.00000254, -0.00000651, -0.00001096, 0.00001579, -0.00001666, -0.00000465, 0.00000110, -0.00001292, -0.00002458, 0.00000403, -0.00003508, -0.00004569, 0.00000584, -0.00032894, -0.00001711, 0.00041753, 0.00031377, 0.00042361, -0.00014049, -0.00029545, 0.00002932, -0.00007612, -0.00000538, -0.00002896, -0.00000158, -0.00003046, 0.00002015, -0.00001757, -0.00002695, 0.00003633, -0.00006824, 0.00006170, -0.00006061, 0.00003202, 0.00000452, -0.00006039, 0.00010024, -0.00009519, 0.00005767, -0.00001395, -0.00003251, 0.00005891, -0.00007485, 0.00005237, -0.00003522, 0.00000208, 0.00003034, -0.00003230, 0.00002302, -0.00002534, 0.00000548, 0.00001222, -0.00001450, -0.00001710, 0.00002225, 0.00000136, -0.00001907, 0.00002399, -0.00000932, -0.00000496, 0.00000485, -0.00000394, 0.00000161, 0.00001012, 0.00000806, -0.00001079, -0.00000235, 0.00000133, -0.00000476, -0.00000133, 0.00000669, 0.00000862, -0.00000160, -0.00000168, 0.00000314, -0.00000608, -0.00000256, 0.00000417, 0.00000306, 0.00000096, -0.00000333, -0.00000055, -0.00000729, -0.00000241, 0.00000676, 0.00000125, 0.00000779, 0.00000346, -0.00000521, 0.00000067, -0.00000216, 0.00000496, 0.00000677, -0.00000755, 0.00000021, 0.00000608, -0.00000006, -0.00000424, -0.00000324, 0.00000532, 0.00000289, -0.00000259, 0.00000075, },
		{ -0.00000168, 0.00000003, -0.00000653, -0.00001330, 0.00000665, 0.00000577, -0.00000649, 0.00000121, -0.00000644, 0.00000073, -0.00000238, -0.00000584, 0.00000142, -0.00000434, 0.00000466, -0.00000778, -0.00000568, 0.00000409, -0.00001063, -0.00000103, -0.00000070, -0.00000251, -0.00000189, -0.00001939, 0.00000249, -0.00000979, -0.00000774, 0.00001700, -0.00002462, 0.00000287, 0.00000055, -0.00002254, -0.00001326, -0.00000141, -0.00005004, -0.00001159, -0.00007058, -0.00032608, 0.00016927, 0.00038200, 0.00025556, 0.00039752, -0.00021659, -0.00024129, 0.00004044, -0.00008892, 0.00000777, -0.00002975, -0.00000916, -0.00001712, 0.00001611, -0.00003127, 0.00000032, 0.00000423, -0.00003998, 0.00004870, -0.00006957, 0.00006339, -0.00004320, -0.00000993, 0.00007104, -0.00009691, 0.00008490, -0.00005575, 0.00001071, 0.00002606, -0.00006295, 0.00006388, -0.00006051, 0.00003241, 0.00000730, -0.00002171, 0.00002372, -0.00003828, 0.00002430, 0.00000261, -0.00002170, -0.00000341, 0.00002171, -0.00001060, -0.00000796, 0.00002217, -0.00001562, 0.00000149, 0.00000166, -0.00000381, 0.00000478, 0.00000937, 0.00000440, -0.00001014, -0.00000067, 0.00000018, -0.00000531, 0.00000054, 0.00000927, 0.00000544, -0.00000321, 0.00000085, 0.00000102, -0.00000703, -0.00000013, 0.00000424, 0.00000292, -0.00000021, -0.00000298, -0.00000078, -0.00000829, -0.00000025, 0.00000652, 0.00000258, 0.00000825, -0.00000034, -0.00000435, 0.00000097, -0.00000197, 0.00000797, 0.00000354, -0.00000886, 0.00000318, 0.00000529, -0.00000129, -0.00000458, -0.00000136, 0.00000594, 0.00000016, -0.00000215, 0.00000159, -0.00000009, },
		{ -0.00000237, -0.00001000, -0.00000857, 0.00000938, 0.00000211, -0.00000568, 0.00000068, -0.00000646, 0.00000107, -0.00000383, -0.00000340, 0.00000005, -0.00000335, 0.00000455, -0.00000948, -0.00000223, 0.00000006, -0.00001037, 0.00000246, -0.00000415, 0.00000084, -0.00000539, -0.00002097, 0.00000782, -0.00001539, -0.00000075, 0.00001339, -0.00003022, 0.00001381, -0.00000805, -0.00002505, -0.00000204, -0.00001767, -0.00004521, 0.00000099, -0.00015748, -0.00024928, 0.00031581, 0.00029089, 0.00024220, 0.00036034, -0.00028708, -0.00017111, 0.00003163, -0.00009330, 0.00002317, -0.00004028, -0.00000647, -0.00000697, 0.00000361, -0.00002875, 0.00001156, -0.00001852, -0.00000955, 0.00002068, -0.00005489, 0.00007020, -0.00007442, 0.00004061, 0.00002429, -0.00007219, 0.00008682, -0.00008075, 0.00004988, -0.00001646, -0.00003096, 0.00005365, -0.00007061, 0.00005789, -0.00002087, 0.00000053, 0.00001011, -0.00003999, 0.00004196, -0.00001632, -0.00001801, 0.00000843, 0.00001387, -0.00001588, 0.00000319, 0.00001542, -0.00001735, 0.00000667, -0.00000292, -0.00000112, 0.00000659, 0.00000732, 0.00000203, -0.00000901, 0.00000031, -0.00000105, -0.00000553, 0.00000348, 0.00001026, 0.00000145, -0.00000272, 0.00000238, -0.00000179, -0.00000651, 0.00000191, 0.00000399, 0.00000265, -0.00000128, -0.00000228, -0.00000163, -0.00000867, 0.00000182, 0.00000633, 0.00000433, 0.00000701, -0.00000341, -0.00000265, 0.00000045, -0.00000046, 0.00000993, -0.00000086, -0.00000825, 0.00000574, 0.00000372, -0.00000213, -0.00000449, 0.00000075, 0.00000533, -0.00000222, -0.00000090, 0.00000148, -0.00000043, 0.00000011, },
		{ -0.00001083, -0.00000325, 0.00000978, -0.00000094, -0.00000454, -0.00000014, -0.00000619, 0.00000089, -0.00000429, -0.00000164, -0.00000141, -0.00000170, 0.00000331, -0.00000964, -0.00000022, -0.00000445, -0.00000731, 0.00000329, -0.00000640, 0.00000465, -0.00001187, -0.00001838, 0.00001070, -0.00002074, 0.00000849, 0.00000417, -0.00002959, 0.00002294, -0.00002181, -0.00001793, 0.00000093, -0.00003451, -0.00002588, -0.00001893, -0.00022341, -0.00011630, 0.00039252, 0.00018316, 0.00026957, 0.00029661, -0.00033556, -0.00009681, 0.00000468, -0.00008248, 0.00003092, -0.00005303, 0.00000562, -0.00000670, -0.00000678, -0.00001846, 0.00000884, -0.00002642, 0.00001146, -0.00000897, -0.00002634, 0.00005419, -0.00008112, 0.00007604, -0.00002428, -0.00003089, 0.00006466, -0.00008198, 0.00007238, -0.00005461, 0.00001101, 0.00002399, -0.00006024, 0.00006879, -0.00004390, 0.00002696, -0.00001578, -0.00002566, 0.00004922, -0.00003734, -0.00000330, 0.00001232, 0.00000396, -0.00001442, 0.00001079, 0.00000675, -0.00001483, 0.00000889, -0.00000696, 0.00000338, 0.00000616, 0.00000515, 0.00000092, -0.00000796, 0.00000082, -0.00000239, -0.00000498, 0.00000697, 0.00000911, -0.00000202, -0.00000068, 0.00000225, -0.00000428, -0.00000486, 0.00000324, 0.00000362, 0.00000220, -0.00000204, -0.00000156, -0.00000295, -0.00000831, 0.00000360, 0.00000653, 0.00000581, 0.00000433, -0.00000506, -0.00000093, -0.00000037, 0.00000230, 0.00000995, -0.00000533, -0.00000580, 0.00000715, 0.00000186, -0.00000251, -0.00000402, 0.00000271, 0.00000360, -0.00000373, 0.00000054, 0.00000082, 0.00000046, 0.00000078, -0.00000625, },
		{ 0.00000128, 0.00000850, -0.00000304, -0.00000342, -0.00000107, -0.00000584, 0.00000061, -0.00000408, -0.00000073, -0.00000236, -0.00000014, 0.00000158, -0.00000857, -0.00000033, -0.00000770, -0.00000291, 0.00000148, -0.00000612, 0.00000667, -0.00001924, -0.00001197, 0.00000951, -0.00002310, 0.00001685, -0.00000867, -0.00002143, 0.00002526, -0.00003473, -0.00000408, -0.00000791, -0.00004141, -0.00000564, -0.00006812, -0.00024256, 0.00003873, 0.00039247, 0.00009522, 0.00031773, 0.00020542, -0.00035082, -0.00003332, -0.00003079, -0.00005748, 0.00002516, -0.00005905, 0.00002025, -0.00001585, -0.00000879, -0.00000962, -0.00000031, -0.00002168, 0.00001811, -0.00002943, 0.00000339, 0.00002481, -0.00006471, 0.00008809, -0.00006034, 0.00001281, 0.00002756, -0.00006105, 0.00007191, -0.00007603, 0.00004924, -0.00001564, -0.00003153, 0.00006030, -0.00005255, 0.00004711, -0.00004594, 0.00000329, 0.00003966, -0.00005078, 0.00001705, 0.00000594, -0.00000230, -0.00000921, 0.00001354, -0.00000086, -0.00000980, 0.00000783, -0.00000881, 0.00000819, 0.00000347, 0.00000395, 0.00000054, -0.00000724, 0.00000104, -0.00000377, -0.00000331, 0.00001009, 0.00000597, -0.00000384, 0.00000183, 0.00000050, -0.00000564, -0.00000272, 0.00000387, 0.00000325, 0.00000161, -0.00000238, -0.00000117, -0.00000437, -0.00000733, 0.00000506, 0.00000728, 0.00000631, 0.00000094, -0.00000514, 0.00000013, -0.00000067, 0.00000560, 0.00000768, -0.00000859, -0.00000219, 0.00000705, 0.00000029, -0.00000253, -0.00000321, 0.00000418, 0.00000110, -0.00000420, 0.00000168, 0.00000027, 0.00000215, -0.00000050, -0.00001016, -0.00000898, },
		{ 0.00000638, -0.00000422, -0.00000251, -0.00000206, -0.00000544, 0.00000047, -0.00000367, -0.00000039, -0.00000265, 0.00000086, 0.00000013, -0.00000718, -0.00000227, -0.00000858, 0.00000100, -0.00000174, -0.00000335, 0.00000542, -0.00002494, -0.00000371, 0.00000451, -0.00002097, 0.00002110, -0.00002113, -0.00000779, 0.00001865, -0.00004077, 0.00000966, -0.00002560, -0.00003315, -0.00000014, -0.00012975, -0.00020442, 0.00017746, 0.00033107, 0.00004953, 0.00036165, 0.00009847, -0.00033186, 0.00000830, -0.00006242, -0.00002600, 0.00000680, -0.00005291, 0.00002908, -0.00002844, -0.00000242, -0.00000776, -0.00000772, -0.00001135, 0.00001261, -0.00003615, 0.00002407, -0.00000562, -0.00003410, 0.00007734, -0.00007575, 0.00004592, -0.00001149, -0.00002688, 0.00005047, -0.00007462, 0.00007158, -0.00005223, 0.00000660, 0.00003476, -0.00004261, 0.00005171, -0.00006904, 0.00003879, 0.00001310, -0.00004868, 0.00003415, -0.00000791, -0.00000146, -0.00000457, 0.00001268, -0.00000573, -0.00000446, 0.00000441, -0.00000774, 0.00001166, -0.00000068, 0.00000432, 0.00000019, -0.00000679, 0.00000101, -0.00000493, -0.00000043, 0.00001183, 0.00000168, -0.00000351, 0.00000356, -0.00000218, -0.00000555, -0.00000075, 0.00000401, 0.00000287, 0.00000105, -0.00000241, -0.00000127, -0.00000554, -0.00000600, 0.00000642, 0.00000839, 0.00000536, -0.00000219, -0.00000405, 0.00000032, 0.00000023, 0.00000829, 0.00000354, -0.00000972, 0.00000147, 0.00000563, -0.00000052, -0.00000245, -0.00000212, 0.00000488, -0.00000171, -0.00000369, 0.00000233, 0.00000038, 0.00000376, -0.00000377, -0.00001264, -0.00000521, 0.00000415, },
		{ -0.00000468, -0.00000192, -0.00000304, -0.00000489, 0.00000048, -0.00000338, -0.00000020, -0.00000258, 0.00000126, -0.00000058, -0.00000649, -0.00000494, -0.00000703, 0.00000306, -0.00000467, 0.00000064, 0.00000085, -0.00002710, 0.00000379, -0.00000240, -0.00001485, 0.00001955, -0.00002914, 0.00000675, 0.00000491, -0.00003691, 0.00001596, -0.00004400, -0.00001284, -0.00001952, -0.00017972, -0.00011728, 0.00027005, 0.00023776, 0.00005012, 0.00038158, -0.00000660, -0.00028738, 0.00002503, -0.00008074, 0.00000183, -0.00001738, -0.00003529, 0.00002690, -0.00003696, 0.00000734, -0.00001288, -0.00000861, -0.00000298, 0.00000164, -0.00003127, 0.00003145, -0.00002708, -0.00000117, 0.00005157, -0.00007045, 0.00006105, -0.00004080, 0.00000826, 0.00001704, -0.00005239, 0.00007186, -0.00007417, 0.00004193, 0.00000087, -0.00001695, 0.00003684, -0.00007495, 0.00006873, -0.00002345, -0.00002862, 0.00003933, -0.00002241, 0.00000585, -0.00000373, 0.00001089, -0.00000802, -0.00000034, 0.00000021, -0.00000405, 0.00001257, -0.00000498, 0.00000624, -0.00000073, -0.00000636, 0.00000069, -0.00000549, 0.00000327, 0.00001149, -0.00000249, -0.00000132, 0.00000366, -0.00000477, -0.00000424, 0.00000063, 0.00000389, 0.00000245, 0.00000067, -0.00000240, -0.00000178, -0.00000622, -0.00000458, 0.00000799, 0.00000932, 0.00000297, -0.00000421, -0.00000257, 0.00000002, 0.00000248, 0.00000924, -0.00000132, -0.00000845, 0.00000406, 0.00000354, -0.00000047, -0.00000249, -0.00000077, 0.00000465, -0.00000439, -0.00000245, 0.00000261, 0.00000134, 0.00000426, -0.00000832, -0.00001278, -0.00000076, 0.00000517, 0.00000418, },
		{ -0.00000168, -0.00000392, -0.00000409, 0.00000044, -0.00000325, 0.00000014, -0.00000257, 0.00000148, -0.00000071, -0.00000711, -0.00000698, -0.00000398, 0.00000292, -0.00000598, 0.00000402, -0.00000571, -0.00002528, 0.00000851, -0.00000873, -0.00000696, 0.00001289, -0.00003036, 0.00001723, -0.00001100, -0.00002474, 0.00001084, -0.00005390, 0.00000899, -0.00006309, -0.00019610, -0.00000471, 0.00030367, 0.00014435, 0.00008539, 0.00036920, -0.00009404, -0.00023125, 0.00002180, -0.00008273, 0.00001827, -0.00003847, -0.00001218, 0.00001392, -0.00003647, 0.00001407, -0.00002062, -0.00000320, -0.00000093, -0.00000777, -0.00002107, 0.00002763, -0.00003526, 0.00002417, 0.00002161, -0.00005096, 0.00005805, -0.00005396, 0.00003363, -0.00001635, -0.00001823, 0.00005177, -0.00007549, 0.00006365, -0.00002968, 0.00001537, 0.00000597, -0.00005906, 0.00008150, -0.00005787, 0.00000453, 0.00002816, -0.00002966, 0.00001522, -0.00000717, 0.00001062, -0.00000917, 0.00000220, -0.00000340, 0.00000119, 0.00001046, -0.00000808, 0.00000913, -0.00000252, -0.00000572, 0.00000006, -0.00000503, 0.00000700, 0.00000894, -0.00000527, 0.00000176, 0.00000192, -0.00000632, -0.00000230, 0.00000134, 0.00000368, 0.00000202, 0.00000057, -0.00000266, -0.00000241, -0.00000645, -0.00000319, 0.00000999, 0.00000936, -0.00000030, -0.00000473, -0.00000145, -0.00000002, 0.00000555, 0.00000777, -0.00000542, -0.00000537, 0.00000488, 0.00000167, 0.00000018, -0.00000274, 0.00000078, 0.00000340, -0.00000655, -0.00000073, 0.00000281, 0.00000286, 0.00000297, -0.00001291, -0.00001038, 0.00000315, 0.00000464, 0.00000240, -0.00000471, },
		{ -0.00000455, -0.00000309, 0.00000013, -0.00000302, 0.00000061, -0.00000285, 0.00000199, -0.00000089, -0.00000891, -0.00000739, -0.00000080, 0.00000124, -0.00000532, 0.00000533, -0.00001225, -0.00002052, 0.00000981, -0.00001256, -0.00000006, 0.00000375, -0.00002528, 0.00002054, -0.00002338, -0.00000965, -0.00000431, -0.00004972, 0.00001905, -0.00011865, -0.00016769, 0.00010337, 0.00028411, 0.00007448, 0.00013584, 0.00032802, -0.00015539, -0.00017668, 0.00000813, -0.00007189, 0.00002100, -0.00004964, 0.00000838, -0.00000480, -0.00002678, 0.00001352, -0.00002541, 0.00000444, -0.00000494, -0.00001158, -0.00001213, 0.00001864, -0.00003192, 0.00003719, -0.00000321, -0.00002670, 0.00004269, -0.00005125, 0.00004371, -0.00003936, 0.00001605, 0.00001964, -0.00005769, 0.00006638, -0.00004689, 0.00004262, -0.00003098, -0.00002475, 0.00007091, -0.00007742, 0.00003989, 0.00000279, -0.00002439, 0.00002054, -0.00001228, 0.00001253, -0.00001070, 0.00000382, -0.00000570, 0.00000666, 0.00000563, -0.00000890, 0.00001206, -0.00000509, -0.00000471, -0.00000078, -0.00000333, 0.00000977, 0.00000479, -0.00000587, 0.00000443, -0.00000113, -0.00000630, -0.00000048, 0.00000155, 0.00000342, 0.00000172, 0.00000062, -0.00000336, -0.00000278, -0.00000648, -0.00000164, 0.00001231, 0.00000796, -0.00000352, -0.00000397, -0.00000110, 0.00000101, 0.00000829, 0.00000407, -0.00000746, -0.00000171, 0.00000394, 0.00000078, 0.00000089, -0.00000309, 0.00000237, 0.00000115, -0.00000785, 0.00000126, 0.00000315, 0.00000430, -0.00000024, -0.00001626, -0.00000599, 0.00000559, 0.00000322, 0.00000127, -0.00000462, -0.00000177, },
		{ -0.00000209, -0.00000052, -0.00000243, 0.00000092, -0.00000333, 0.00000305, -0.00000187, -0.00001109, -0.00000596, 0.00000132, -0.00000079, -0.00000335, 0.00000412, -0.00001700, -0.00001467, 0.00000845, -0.00001341, 0.00000386, -0.00000454, -0.00001677, 0.00001680, -0.00002843, 0.00000173, -0.00002324, -0.00003265, 0.00000749, -0.00016674, -0.00009826, 0.00018119, 0.00023080, 0.00003808, 0.00018282, 0.00026915, -0.00019077, -0.00013177, -0.00000663, -0.00005546, 0.00001327, -0.00004888, 0.00002010, -0.00002216, -0.00001201, 0.00000554, -0.00002367, 0.00000923, -0.00001121, -0.00001007, -0.00000814, 0.00001084, -0.00002273, 0.00003891, -0.00001809, -0.00000587, 0.00002318, -0.00003818, 0.00003937, -0.00004689, 0.00004027, -0.00001330, -0.00002847, 0.00005181, -0.00004621, 0.00005508, -0.00006140, 0.00001751, 0.00003908, -0.00007392, 0.00006460, -0.00002835, -0.00000666, 0.00001705, -0.00001475, 0.00001504, -0.00001308, 0.00000539, -0.00000672, 0.00001124, -0.00000104, -0.00000690, 0.00001398, -0.00000806, -0.00000333, -0.00000160, -0.00000054, 0.00001073, 0.00000020, -0.00000426, 0.00000560, -0.00000448, -0.00000479, 0.00000068, 0.00000154, 0.00000309, 0.00000172, 0.00000052, -0.00000438, -0.00000266, -0.00000659, 0.00000041, 0.00001448, 0.00000495, -0.00000575, -0.00000262, -0.00000130, 0.00000344, 0.00000941, -0.00000078, -0.00000687, 0.00000113, 0.00000191, 0.00000118, 0.00000111, -0.00000320, 0.00000370, -0.00000192, -0.00000801, 0.00000335, 0.00000371, 0.00000496, -0.00000487, -0.00001738, -0.00000064, 0.00000616, 0.00000163, 0.00000074, -0.00000465, -0.00000206, -0.00000478, },
		{ -0.00000130, -0.00000137, 0.00000073, -0.00000359, 0.00000442, -0.00000409, -0.00001260, -0.00000327, 0.00000193, -0.00000211, -0.00000130, 0.00000103, -0.00001918, -0.00000944, 0.00000599, -0.00001220, 0.00000445, -0.00000961, -0.00000863, 0.00000895, -0.00002595, 0.00000480, -0.00003785, -0.00001057, -0.00002713, -0.00018805, -0.00000506, 0.00021520, 0.00016792, 0.00003170, 0.00021480, 0.00020516, -0.00020638, -0.00009824, -0.00001685, -0.00004052, 0.00000147, -0.00003914, 0.00002097, -0.00003245, 0.00000185, -0.00000628, -0.00001552, 0.00000800, -0.00001513, -0.00000673, -0.00000873, 0.00000783, -0.00001384, 0.00003414, -0.00002340, 0.00000730, 0.00000652, -0.00002224, 0.00002633, -0.00004012, 0.00004932, -0.00003734, 0.00000175, 0.00002740, -0.00003005, 0.00004895, -0.00007496, 0.00005393, -0.00000393, -0.00004729, 0.00006885, -0.00005369, 0.00001779, 0.00000390, -0.00001091, 0.00001520, -0.00001522, 0.00000720, -0.00000682, 0.00001417, -0.00000838, -0.00000210, 0.00001401, -0.00001078, -0.00000173, -0.00000208, 0.00000282, 0.00000946, -0.00000352, -0.00000120, 0.00000474, -0.00000699, -0.00000237, 0.00000099, 0.00000153, 0.00000271, 0.00000212, -0.00000010, -0.00000536, -0.00000210, -0.00000693, 0.00000328, 0.00001571, 0.00000077, -0.00000642, -0.00000151, -0.00000132, 0.00000687, 0.00000798, -0.00000515, -0.00000406, 0.00000211, -0.00000009, 0.00000259, 0.00000052, -0.00000273, 0.00000432, -0.00000543, -0.00000688, 0.00000539, 0.00000430, 0.00000429, -0.00000997, -0.00001585, 0.00000445, 0.00000500, 0.00000050, 0.00000055, -0.00000495, -0.00000234, -0.00000458, -0.00000138, },
		{ -0.00000004, -0.00000010, -0.00000318, 0.00000551, -0.00000740, -0.00001260, -0.00000036, 0.00000131, -0.00000230, -0.00000027, -0.00000264, -0.00001910, -0.00000573, 0.00000383, -0.00001054, 0.00000291, -0.00001098, -0.00000369, 0.00000121, -0.00001915, -0.00000089, -0.00004216, 0.00000523, -0.00007631, -0.00017095, 0.00008771, 0.00020684, 0.00011549, 0.00004369, 0.00022920, 0.00014507, -0.00021015, -0.00007315, -0.00002177, -0.00003097, -0.00000825, -0.00002621, 0.00001345, -0.00003364, 0.00000997, -0.00001676, -0.00000438, 0.00000100, -0.00001390, -0.00000576, -0.00001048, 0.00000931, -0.00000895, 0.00002838, -0.00002291, 0.00001302, -0.00000396, -0.00000945, 0.00001190, -0.00002490, 0.00004420, -0.00004756, 0.00002412, 0.00000287, -0.00000624, 0.00002758, -0.00006769, 0.00007278, -0.00004368, -0.00000603, 0.00005006, -0.00006316, 0.00003961, -0.00001505, 0.00000013, 0.00001039, -0.00001504, 0.00000856, -0.00000616, 0.00001497, -0.00001507, 0.00000486, 0.00001157, -0.00001254, -0.00000012, -0.00000191, 0.00000593, 0.00000622, -0.00000531, 0.00000206, 0.00000209, -0.00000788, 0.00000013, 0.00000060, 0.00000163, 0.00000239, 0.00000280, -0.00000148, -0.00000579, -0.00000146, -0.00000731, 0.00000702, 0.00001524, -0.00000365, -0.00000555, -0.00000120, -0.00000029, 0.00001018, 0.00000396, -0.00000746, -0.00000029, 0.00000100, -0.00000099, 0.00000427, -0.00000076, -0.00000152, 0.00000378, -0.00000878, -0.00000444, 0.00000715, 0.00000463, 0.00000210, -0.00001442, -0.00001190, 0.00000819, 0.00000267, 0.00000019, 0.00000033, -0.00000548, -0.00000235, -0.00000410, -0.00000114, -0.00000187, },
		{ -0.00000138, -0.00000190, 0.00000561, -0.00001111, -0.00001085, 0.00000182, 0.00000024, -0.00000169, -0.00000071, -0.00000562, -0.00001781, -0.00000341, 0.00000257, -0.00000972, 0.00000118, -0.00001004, -0.00000265, -0.00000314, -0.00001285, -0.00001158, -0.00003530, 0.00000510, -0.00012434, -0.00011596, 0.00015745, 0.00016973, 0.00008372, 0.00006091, 0.00023020, 0.00009226, -0.00020795, -0.00005220, -0.00002410, -0.00002659, -0.00001269, -0.00001568, 0.00000264, -0.00002767, 0.00001067, -0.00002183, 0.00000504, -0.00000853, -0.00000794, -0.00000954, -0.00000937, 0.00001201, -0.00000812, 0.00002510, -0.00002112, 0.00001440, -0.00000879, -0.00000227, 0.00000169, -0.00000855, 0.00003045, -0.00004479, 0.00003436, -0.00001388, 0.00001563, -0.00000021, -0.00004344, 0.00006889, -0.00006688, 0.00003580, 0.00001411, -0.00005235, 0.00004977, -0.00003259, 0.00001553, 0.00000005, -0.00001069, 0.00000810, -0.00000447, 0.00001327, -0.00001975, 0.00001283, 0.00000664, -0.00001271, 0.00000119, -0.00000092, 0.00000788, 0.00000188, -0.00000475, 0.00000424, -0.00000144, -0.00000696, 0.00000197, -0.00000013, 0.00000179, 0.00000231, 0.00000344, -0.00000354, -0.00000530, -0.00000122, -0.00000726, 0.00001120, 0.00001267, -0.00000721, -0.00000379, -0.00000162, 0.00000234, 0.00001189, -0.00000165, -0.00000682, 0.00000279, -0.00000156, -0.00000025, 0.00000537, -0.00000222, 0.00000022, 0.00000180, -0.00001125, -0.00000092, 0.00000834, 0.00000438, -0.00000137, -0.00001721, -0.00000632, 0.00000990, -0.00000004, 0.00000073, -0.00000027, -0.00000604, -0.00000195, -0.00000366, -0.00000136, -0.00000193, 0.00000119, },
		{ 0.00000006, 0.00000428, -0.00001428, -0.00000776, 0.00000279, -0.00000050, -0.00000106, -0.00000226, -0.00000725, -0.00001643, -0.00000171, 0.00000184, -0.00001007, 0.00000076, -0.00000907, -0.00000399, -0.00000329, -0.00001095, -0.00002130, -0.00002210, -0.00001443, -0.00015406, -0.00003557, 0.00019052, 0.00012300, 0.00007212, 0.00007437, 0.00022439, 0.00004545, -0.00020186, -0.00003270, -0.00002727, -0.00002437, -0.00001233, -0.00001050, -0.00000639, -0.00001881, 0.00000572, -0.00002033, 0.00000915, -0.00001624, -0.00000039, -0.00001741, -0.00000316, 0.00001198, -0.00000860, 0.00002460, -0.00002081, 0.00001501, -0.00001088, 0.00000049, -0.00000235, 0.00000336, 0.00001500, -0.00003401, 0.00003350, -0.00001956, 0.00002828, -0.00002409, -0.00001206, 0.00004563, -0.00006653, 0.00006392, -0.00002668, -0.00002451, 0.00004340, -0.00004120, 0.00002965, -0.00001351, -0.00000188, 0.00000464, -0.00000124, 0.00000879, -0.00002115, 0.00002020, -0.00000019, -0.00001093, 0.00000197, 0.00000086, 0.00000800, -0.00000230, -0.00000224, 0.00000445, -0.00000463, -0.00000464, 0.00000273, -0.00000082, 0.00000195, 0.00000266, 0.00000357, -0.00000588, -0.00000385, -0.00000175, -0.00000618, 0.00001499, 0.00000822, -0.00000906, -0.00000206, -0.00000209, 0.00000634, 0.00001081, -0.00000707, -0.00000342, 0.00000382, -0.00000430, 0.00000195, 0.00000528, -0.00000319, 0.00000193, -0.00000150, -0.00001212, 0.00000319, 0.00000872, 0.00000333, -0.00000555, -0.00001769, -0.00000022, 0.00000939, -0.00000229, 0.00000184, -0.00000148, -0.00000629, -0.00000122, -0.00000356, -0.00000175, -0.00000140, 0.00000191, -0.00000266, },
		{ 0.00000154, -0.00001609, -0.00000414, 0.00000266, -0.00000057, -0.00000116, -0.00000404, -0.00000768, -0.00001557, 0.00000014, 0.00000077, -0.00001085, 0.00000191, -0.00000986, -0.00000505, -0.00000102, -0.00001445, -0.00002499, -0.00001100, -0.00004864, -0.00015331, 0.00005033, 0.00018565, 0.00008382, 0.00007276, 0.00008147, 0.00021667, 0.00000166, -0.00019089, -0.00001486, -0.00003305, -0.00002086, -0.00001019, -0.00001028, -0.00001088, -0.00001136, -0.00000111, -0.00001411, 0.00000707, -0.00001895, 0.00000470, -0.00002607, 0.00000741, 0.00000691, -0.00000682, 0.00002471, -0.00002212, 0.00001681, -0.00001319, 0.00000196, -0.00000182, 0.00000890, 0.00000312, -0.00002147, 0.00002631, -0.00001599, 0.00002946, -0.00003657, 0.00001484, 0.00001327, -0.00004445, 0.00006924, -0.00005833, 0.00001064, 0.00002199, -0.00003642, 0.00003638, -0.00002566, 0.00000947, -0.00000184, 0.00000377, 0.00000153, -0.00001835, 0.00002517, -0.00000768, -0.00000733, 0.00000206, 0.00000316, 0.00000602, -0.00000504, 0.00000113, 0.00000259, -0.00000641, -0.00000178, 0.00000242, -0.00000124, 0.00000212, 0.00000341, 0.00000280, -0.00000786, -0.00000177, -0.00000305, -0.00000372, 0.00001737, 0.00000275, -0.00000897, -0.00000113, -0.00000163, 0.00001062, 0.00000660, -0.00001039, 0.00000144, 0.00000227, -0.00000585, 0.00000477, 0.00000398, -0.00000316, 0.00000282, -0.00000554, -0.00001097, 0.00000719, 0.00000813, 0.00000145, -0.00000967, -0.00001570, 0.00000520, 0.00000703, -0.00000345, 0.00000298, -0.00000323, -0.00000590, -0.00000043, -0.00000391, -0.00000196, -0.00000048, 0.00000197, -0.00000348, 0.00000214, },
		{ -0.00001617, -0.00000080, 0.00000191, -0.00000015, -0.00000231, -0.00000517, -0.00000759, -0.00001515, 0.00000241, -0.00000133, -0.00001080, 0.00000359, -0.00001274, -0.00000369, 0.00000045, -0.00002113, -0.00002120, -0.00001005, -0.00008608, -0.00011944, 0.00012140, 0.00015260, 0.00006199, 0.00007563, 0.00008498, 0.00020816, -0.00004090, -0.00017318, -0.00000110, -0.00004051, -0.00001433, -0.00000967, -0.00001223, -0.00001101, -0.00000765, -0.00000617, -0.00000675, 0.00000070, -0.00001607, 0.00000449, -0.00003147, 0.00001900, -0.00000271, -0.00000058, 0.00002250, -0.00002311, 0.00001941, -0.00001693, 0.00000502, -0.00000018, 0.00000958, -0.00000331, -0.00001170, 0.00001833, -0.00000822, 0.00002216, -0.00003589, 0.00002871, -0.00001533, -0.00001041, 0.00005124, -0.00007050, 0.00004069, -0.00000702, -0.00001893, 0.00003190, -0.00003138, 0.00001960, -0.00000978, 0.00000991, -0.00000786, -0.00001106, 0.00002612, -0.00001412, -0.00000258, 0.00000156, 0.00000548, 0.00000223, -0.00000545, 0.00000396, -0.00000069, -0.00000627, 0.00000067, 0.00000141, -0.00000133, 0.00000240, 0.00000435, 0.00000099, -0.00000886, 0.00000028, -0.00000470, 0.00000004, 0.00001754, -0.00000253, -0.00000739, -0.00000118, 0.00000055, 0.00001356, 0.00000012, -0.00001033, 0.00000586, -0.00000131, -0.00000538, 0.00000710, 0.00000203, -0.00000212, 0.00000225, -0.00000937, -0.00000783, 0.00001034, 0.00000660, -0.00000112, -0.00001292, -0.00001160, 0.00000894, 0.00000358, -0.00000326, 0.00000356, -0.00000518, -0.00000472, -0.00000000, -0.00000460, -0.00000172, 0.00000047, 0.00000149, -0.00000357, 0.00000371, -0.00000210, },
		{ 0.00000170, 0.00000110, 0.00000027, -0.00000425, -0.00000516, -0.00000780, -0.00001463, 0.00000473, -0.00000448, -0.00000903, 0.00000415, -0.00001644, 0.00000053, -0.00000160, -0.00002682, -0.00001286, -0.00002314, -0.00011285, -0.00006009, 0.00016346, 0.00010727, 0.00005831, 0.00007341, 0.00008981, 0.00019663, -0.00008129, -0.00014826, 0.00000585, -0.00004675, -0.00000564, -0.00001258, -0.00001314, -0.00000911, -0.00000741, -0.00000759, -0.00000158, -0.00000653, -0.00000960, -0.00000131, -0.00003077, 0.00002744, -0.00001386, 0.00000984, 0.00001622, -0.00002145, 0.00002086, -0.00002109, 0.00001055, -0.00000045, 0.00000872, -0.00000564, -0.00000607, 0.00001317, -0.00000151, 0.00001241, -0.00002611, 0.00002736, -0.00003006, 0.00002207, 0.00001786, -0.00006035, 0.00005542, -0.00003320, 0.00000541, 0.00001667, -0.00002746, 0.00002444, -0.00001635, 0.00001550, -0.00001781, -0.00000010, 0.00002219, -0.00001776, 0.00000219, 0.00000079, 0.00000721, -0.00000252, -0.00000333, 0.00000498, -0.00000415, -0.00000437, 0.00000205, 0.00000020, -0.00000116, 0.00000295, 0.00000505, -0.00000158, -0.00000853, 0.00000166, -0.00000600, 0.00000445, 0.00001531, -0.00000645, -0.00000528, -0.00000170, 0.00000447, 0.00001365, -0.00000679, -0.00000678, 0.00000808, -0.00000551, -0.00000291, 0.00000807, 0.00000030, -0.00000061, -0.00000001, -0.00001190, -0.00000322, 0.00001206, 0.00000431, -0.00000409, -0.00001462, -0.00000619, 0.00001039, 0.00000006, -0.00000190, 0.00000315, -0.00000678, -0.00000292, -0.00000028, -0.00000524, -0.00000102, 0.00000111, 0.00000081, -0.00000300, 0.00000429, -0.00000588, -0.00001467, },
		{ 0.00000054, 0.00000014, -0.00000631, -0.00000411, -0.00000876, -0.00001342, 0.00000638, -0.00000794, -0.00000555, 0.00000232, -0.00001887, 0.00000610, -0.00000785, -0.00002775, -0.00000605, -0.00004781, -0.00011788, 0.00000977, 0.00017215, 0.00006542, 0.00006660, 0.00006397, 0.00009967, 0.00017861, -0.00011622, -0.00011813, 0.00000484, -0.00004868, 0.00000246, -0.00001827, -0.00001119, -0.00000792, -0.00000857, -0.00000587, -0.00000023, -0.00001151, -0.00000287, -0.00001049, -0.00002369, 0.00002965, -0.00002257, 0.00002192, 0.00000622, -0.00001591, 0.00001910, -0.00002329, 0.00001713, -0.00000361, 0.00000920, -0.00000676, -0.00000310, 0.00001131, 0.00000135, 0.00000589, -0.00001459, 0.00001529, -0.00002739, 0.00004114, -0.00001794, -0.00003330, 0.00005073, -0.00004730, 0.00002788, -0.00000449, -0.00001420, 0.00002134, -0.00001859, 0.00001820, -0.00002591, 0.00001254, 0.00001370, -0.00001734, 0.00000564, 0.00000029, 0.00000766, -0.00000704, 0.00000069, 0.00000352, -0.00000644, -0.00000152, 0.00000214, -0.00000079, -0.00000083, 0.00000383, 0.00000503, -0.00000433, -0.00000701, 0.00000202, -0.00000625, 0.00000845, 0.00001119, -0.00000838, -0.00000361, -0.00000166, 0.00000924, 0.00001017, -0.00001190, -0.00000092, 0.00000718, -0.00000867, 0.00000071, 0.00000745, -0.00000043, 0.00000045, -0.00000352, -0.00001232, 0.00000189, 0.00001210, 0.00000155, -0.00000706, -0.00001436, -0.00000056, 0.00000952, -0.00000258, 0.00000001, 0.00000169, -0.00000747, -0.00000098, -0.00000132, -0.00000544, -0.00000012, 0.00000134, 0.00000029, -0.00000213, 0.00000378, -0.00000909, -0.00001205, 0.00000323, },
		{ -0.00000078, -0.00000779, -0.00000256, -0.00001045, -0.00001128, 0.00000676, -0.00001063, -0.00000135, -0.00000202, -0.00001834, 0.00001039, -0.00001649, -0.00002277, -0.00000702, -0.00007594, -0.00009705, 0.00007322, 0.00015291, 0.00003767, 0.00007780, 0.00005007, 0.00011495, 0.00015183, -0.00014179, -0.00008685, -0.00000276, -0.00004476, 0.00000695, -0.00002431, -0.00000670, -0.00000900, -0.00000874, -0.00000307, -0.00000211, -0.00001284, 0.00000119, -0.00001957, -0.00001250, 0.00002501, -0.00002572, 0.00003219, -0.00000529, -0.00000718, 0.00001343, -0.00002143, 0.00002207, -0.00000834, 0.00001194, -0.00000900, -0.00000019, 0.00001067, 0.00000105, 0.00000511, -0.00000823, 0.00000112, -0.00001139, 0.00004122, -0.00004346, -0.00000046, 0.00002991, -0.00004492, 0.00004038, -0.00002416, 0.00000439, 0.00001051, -0.00001479, 0.00001600, -0.00002957, 0.00002392, 0.00000239, -0.00001265, 0.00000669, 0.00000065, 0.00000631, -0.00001001, 0.00000535, -0.00000025, -0.00000655, 0.00000114, 0.00000126, -0.00000134, -0.00000037, 0.00000493, 0.00000403, -0.00000656, -0.00000481, 0.00000145, -0.00000516, 0.00001101, 0.00000630, -0.00000843, -0.00000286, -0.00000009, 0.00001319, 0.00000367, -0.00001346, 0.00000517, 0.00000347, -0.00000956, 0.00000412, 0.00000578, 0.00000007, 0.00000013, -0.00000726, -0.00001038, 0.00000642, 0.00001060, -0.00000140, -0.00000955, -0.00001212, 0.00000418, 0.00000687, -0.00000374, 0.00000167, -0.00000044, -0.00000698, 0.00000048, -0.00000284, -0.00000498, 0.00000062, 0.00000132, 0.00000010, -0.00000138, 0.00000242, -0.00001131, -0.00000891, 0.00000392, 0.00000024, },
		{ -0.00000823, -0.00000119, -0.00001246, -0.00000840, 0.00000577, -0.00001172, 0.00000212, -0.00000775, -0.00001449, 0.00001114, -0.00002416, -0.00001413, -0.00001904, -0.00009681, -0.00005452, 0.00011696, 0.00011758, 0.00002724, 0.00008421, 0.00003696, 0.00013279, 0.00011674, -0.00015543, -0.00005893, -0.00001359, -0.00003578, 0.00000618, -0.00002786, -0.00000167, -0.00001204, -0.00000666, -0.00000145, -0.00000521, -0.00001116, 0.00000133, -0.00002548, -0.00000081, 0.00001536, -0.00002230, 0.00003786, -0.00001538, 0.00000257, 0.00000500, -0.00001494, 0.00002298, -0.00001200, 0.00001572, -0.00001284, 0.00000438, 0.00000856, 0.00000080, 0.00000838, -0.00001008, -0.00000666, 0.00000864, 0.00002469, -0.00005112, 0.00002600, 0.00000201, -0.00002809, 0.00003872, -0.00003522, 0.00002199, -0.00000484, -0.00000550, 0.00000822, -0.00002691, 0.00003095, -0.00000890, -0.00000479, 0.00000497, 0.00000220, 0.00000295, -0.00001041, 0.00000902, -0.00000528, -0.00000421, 0.00000257, 0.00000004, -0.00000152, 0.00000031, 0.00000594, 0.00000210, -0.00000772, -0.00000264, 0.00000050, -0.00000299, 0.00001152, 0.00000188, -0.00000737, -0.00000280, 0.00000337, 0.00001456, -0.00000410, -0.00001090, 0.00000931, -0.00000163, -0.00000792, 0.00000617, 0.00000410, 0.00000131, -0.00000203, -0.00000995, -0.00000654, 0.00000950, 0.00000796, -0.00000429, -0.00001106, -0.00000828, 0.00000709, 0.00000341, -0.00000335, 0.00000238, -0.00000255, -0.00000547, 0.00000097, -0.00000431, -0.00000396, 0.00000094, 0.00000134, 0.00000013, -0.00000106, 0.00000062, -0.00001239, -0.00000570, 0.00000401, -0.00000001, -0.00000237, },
		{ -0.00000051, -0.00001430, -0.00000526, 0.00000380, -0.00001106, 0.00000367, -0.00001307, -0.00000851, 0.00000767, -0.00002774, -0.00000663, -0.00004054, -0.00010126, -0.00000098, 0.00013497, 0.00007944, 0.00003085, 0.00008213, 0.00002950, 0.00014861, 0.00007644, -0.00015711, -0.00003764, -0.00002373, -0.00002449, 0.00000064, -0.00002726, 0.00000152, -0.00001529, -0.00000275, -0.00000219, -0.00000739, -0.00000844, -0.00000165, -0.00002689, 0.00000806, 0.00000404, -0.00001353, 0.00003789, -0.00002182, 0.00001079, -0.00000374, -0.00000523, 0.00001912, -0.00001231, 0.00001820, -0.00001681, 0.00001031, 0.00000378, 0.00000396, 0.00001106, -0.00001808, -0.00000360, 0.00002253, 0.00000026, -0.00004096, 0.00003766, -0.00002215, -0.00000415, 0.00002422, -0.00003374, 0.00003235, -0.00001956, 0.00000644, -0.00000377, -0.00001772, 0.00003137, -0.00001694, 0.00000397, 0.00000107, 0.00000481, -0.00000214, -0.00000779, 0.00001023, -0.00000996, -0.00000002, 0.00000220, -0.00000083, -0.00000151, 0.00000129, 0.00000644, -0.00000032, -0.00000767, -0.00000106, -0.00000016, -0.00000052, 0.00001010, -0.00000113, -0.00000618, -0.00000254, 0.00000808, 0.00001221, -0.00001070, -0.00000515, 0.00001011, -0.00000621, -0.00000455, 0.00000636, 0.00000345, 0.00000221, -0.00000571, -0.00001050, -0.00000185, 0.00001070, 0.00000473, -0.00000692, -0.00001117, -0.00000366, 0.00000776, 0.00000025, -0.00000190, 0.00000184, -0.00000387, -0.00000351, 0.00000038, -0.00000519, -0.00000276, 0.00000082, 0.00000166, 0.00000010, -0.00000120, -0.00000121, -0.00001246, -0.00000273, 0.00000364, -0.00000045, -0.00000424, -0.00000758, },
		{ -0.00001559, -0.00000230, 0.00000153, -0.00000920, 0.00000299, -0.00001646, -0.00000252, 0.00000130, -0.00002608, -0.00000539, -0.00006536, -0.00008519, 0.00005038, 0.00012923, 0.00004881, 0.00004165, 0.00007221, 0.00003010, 0.00015820, 0.00003542, -0.00014921, -0.00002394, -0.00003030, -0.00001421, -0.00000733, -0.00002271, 0.00000147, -0.00001673, 0.00000130, -0.00000488, -0.00000758, -0.00000669, -0.00000570, -0.00002452, 0.00001235, -0.00000562, -0.00000208, 0.00003307, -0.00002387, 0.00001580, -0.00001025, 0.00000499, 0.00001179, -0.00000855, 0.00001750, -0.00001868, 0.00001555, -0.00000246, 0.00001177, 0.00000838, -0.00002638, 0.00000877, 0.00002377, -0.00002113, -0.00001987, 0.00003277, -0.00003432, 0.00001744, 0.00000305, -0.00002058, 0.00003185, -0.00002860, 0.00001681, -0.00001677, -0.00000388, 0.00002477, -0.00001921, 0.00001086, -0.00000350, 0.00000772, -0.00000807, -0.00000248, 0.00000812, -0.00001257, 0.00000471, 0.00000015, -0.00000087, -0.00000157, 0.00000262, 0.00000605, -0.00000259, -0.00000674, -0.00000020, -0.00000003, 0.00000126, 0.00000751, -0.00000250, -0.00000555, -0.00000103, 0.00001250, 0.00000631, -0.00001396, 0.00000171, 0.00000750, -0.00000865, -0.00000090, 0.00000516, 0.00000430, 0.00000159, -0.00000980, -0.00000848, 0.00000242, 0.00001013, 0.00000134, -0.00000903, -0.00000966, 0.00000071, 0.00000638, -0.00000168, -0.00000026, 0.00000031, -0.00000390, -0.00000191, -0.00000094, -0.00000520, -0.00000186, 0.00000056, 0.00000232, -0.00000026, -0.00000162, -0.00000281, -0.00001167, -0.00000018, 0.00000286, -0.00000113, -0.00000596, -0.00000607, -0.00000073, },
		{ 0.00000025, -0.00000052, -0.00000703, 0.00000064, -0.00001734, 0.00000157, -0.00000539, -0.00002077, -0.00001335, -0.00008491, -0.00005100, 0.00008834, 0.00010758, 0.00003061, 0.00005266, 0.00005805, 0.00003820, 0.00015929, -0.00000217, -0.00013541, -0.00001661, -0.00003239, -0.00000741, -0.00001482, -0.00001608, -0.00000161, -0.00001520, 0.00000358, -0.00000795, -0.00000617, -0.00000686, -0.00000871, -0.00002045, 0.00001229, -0.00001154, 0.00000926, 0.00002533, -0.00002227, 0.00001741, -0.00001308, 0.00001301, 0.00000355, -0.00000192, 0.00001331, -0.00001701, 0.00001766, -0.00000702, 0.00002223, -0.00000176, -0.00002848, 0.00002404, 0.00001214, -0.00003109, 0.00000191, 0.00001625, -0.00003186, 0.00002902, -0.00001647, -0.00000091, 0.00002083, -0.00002895, 0.00002172, -0.00002678, 0.00001100, 0.00001292, -0.00001485, 0.00001350, -0.00000684, 0.00000972, -0.00001352, 0.00000431, 0.00000279, -0.00001188, 0.00000834, -0.00000282, 0.00000000, -0.00000179, 0.00000418, 0.00000466, -0.00000407, -0.00000552, 0.00000022, 0.00000088, 0.00000162, 0.00000490, -0.00000271, -0.00000551, 0.00000231, 0.00001473, -0.00000162, -0.00001282, 0.00000721, 0.00000273, -0.00000833, 0.00000155, 0.00000373, 0.00000624, -0.00000131, -0.00001276, -0.00000432, 0.00000519, 0.00000833, -0.00000192, -0.00001030, -0.00000666, 0.00000380, 0.00000373, -0.00000203, 0.00000067, -0.00000147, -0.00000271, -0.00000134, -0.00000230, -0.00000449, -0.00000158, 0.00000055, 0.00000309, -0.00000109, -0.00000209, -0.00000406, -0.00001021, 0.00000177, 0.00000171, -0.00000197, -0.00000724, -0.00000408, -0.00000186, -0.00001619, },
		{ -0.00000219, -0.00000526, -0.00000231, -0.00001628, 0.00000299, -0.00000992, -0.00001547, -0.00002964, -0.00009150, -0.00000659, 0.00010691, 0.00008004, 0.00002438, 0.00005929, 0.00004394, 0.00005111, 0.00015193, -0.00003372, -0.00011920, -0.00001327, -0.00003099, -0.00000481, -0.00001957, -0.00000978, -0.00000612, -0.00001103, 0.00000307, -0.00000969, -0.00000462, -0.00000847, -0.00000966, -0.00001693, 0.00000956, -0.00001333, 0.00001856, 0.00001671, -0.00001850, 0.00001667, -0.00001239, 0.00001730, -0.00000302, 0.00000527, 0.00000706, -0.00001213, 0.00001550, -0.00000665, 0.00003089, -0.00001735, -0.00002066, 0.00003407, -0.00000667, -0.00002707, 0.00001535, -0.00000310, -0.00001822, 0.00002761, -0.00002723, 0.00001793, 0.00000342, -0.00002089, 0.00001921, -0.00003043, 0.00002267, -0.00000055, -0.00000517, 0.00001083, -0.00000731, 0.00000952, -0.00001692, 0.00001081, -0.00000462, -0.00000758, 0.00000947, -0.00000550, 0.00000142, -0.00000213, 0.00000569, 0.00000247, -0.00000446, -0.00000461, 0.00000082, 0.00000198, 0.00000044, 0.00000327, -0.00000265, -0.00000534, 0.00000713, 0.00001332, -0.00000909, -0.00000780, 0.00000951, -0.00000217, -0.00000582, 0.00000202, 0.00000337, 0.00000806, -0.00000630, -0.00001318, 0.00000077, 0.00000596, 0.00000595, -0.00000494, -0.00001034, -0.00000275, 0.00000495, 0.00000091, -0.00000108, 0.00000035, -0.00000258, -0.00000089, -0.00000203, -0.00000301, -0.00000356, -0.00000187, 0.00000108, 0.00000359, -0.00000225, -0.00000243, -0.00000492, -0.00000823, 0.00000293, 0.00000033, -0.00000283, -0.00000796, -0.00000194, -0.00000378, -0.00001601, 0.00000241, },
		{ -0.00000408, -0.00000499, -0.00001449, 0.00000241, -0.00001124, -0.00001424, -0.00004963, -0.00008119, 0.00003766, 0.00010631, 0.00005524, 0.00002618, 0.00006018, 0.00003303, 0.00006553, 0.00013790, -0.00005853, -0.00010289, -0.00001165, -0.00002807, -0.00000542, -0.00002087, -0.00000566, -0.00000998, -0.00000578, 0.00000009, -0.00000921, -0.00000445, -0.00001002, -0.00000897, -0.00001531, 0.00000624, -0.00001187, 0.00002513, 0.00000852, -0.00001387, 0.00001508, -0.00000957, 0.00001801, -0.00000651, 0.00001087, 0.00000100, -0.00000601, 0.00001018, -0.00000001, 0.00003304, -0.00003298, -0.00000404, 0.00003302, -0.00002389, -0.00001286, 0.00001627, -0.00001665, -0.00000085, 0.00001573, -0.00002610, 0.00002934, -0.00001422, -0.00000776, 0.00001008, -0.00002644, 0.00002774, -0.00001132, 0.00000670, 0.00000360, -0.00000425, 0.00000628, -0.00001702, 0.00001506, -0.00001215, -0.00000050, 0.00000739, -0.00000671, 0.00000272, -0.00000236, 0.00000678, -0.00000001, -0.00000385, -0.00000433, 0.00000212, 0.00000237, -0.00000170, 0.00000302, -0.00000306, -0.00000400, 0.00001205, 0.00000802, -0.00001362, -0.00000087, 0.00000809, -0.00000528, -0.00000262, 0.00000078, 0.00000488, 0.00000825, -0.00001223, -0.00001036, 0.00000520, 0.00000491, 0.00000351, -0.00000759, -0.00000882, 0.00000115, 0.00000410, -0.00000103, 0.00000035, -0.00000114, -0.00000233, 0.00000059, -0.00000359, -0.00000269, -0.00000304, -0.00000236, 0.00000217, 0.00000353, -0.00000350, -0.00000261, -0.00000531, -0.00000599, 0.00000315, -0.00000101, -0.00000362, -0.00000807, 0.00000002, -0.00000636, -0.00001496, 0.00000418, -0.00000367, },
		{ -0.00000714, -0.00001301, 0.00000140, -0.00001033, -0.00001956, -0.00006659, -0.00005521, 0.00007234, 0.00009176, 0.00003811, 0.00003110, 0.00005651, 0.00002659, 0.00007886, 0.00011948, -0.00007721, -0.00008730, -0.00001048, -0.00002542, -0.00000739, -0.00001952, -0.00000426, -0.00001165, -0.00000138, -0.00000395, -0.00000688, -0.00000635, -0.00000997, -0.00000813, -0.00001558, 0.00000379, -0.00000853, 0.00002921, 0.00000123, -0.00000895, 0.00001368, -0.00000635, 0.00001648, -0.00000706, 0.00001390, -0.00000309, -0.00000117, 0.00000468, 0.00001131, 0.00002618, -0.00004234, 0.00001579, 0.00002004, -0.00003187, 0.00000366, 0.00000652, -0.00001952, 0.00001233, -0.00000021, -0.00001482, 0.00002999, -0.00002637, 0.00000548, -0.00000240, -0.00001615, 0.00002501, -0.00001585, 0.00001671, -0.00000581, 0.00000163, 0.00000010, -0.00001327, 0.00001553, -0.00001755, 0.00000757, 0.00000242, -0.00000570, 0.00000319, -0.00000217, 0.00000709, -0.00000213, -0.00000272, -0.00000450, 0.00000424, 0.00000124, -0.00000371, 0.00000383, -0.00000413, -0.00000075, 0.00001513, 0.00000007, -0.00001366, 0.00000535, 0.00000395, -0.00000558, -0.00000043, -0.00000095, 0.00000804, 0.00000565, -0.00001724, -0.00000465, 0.00000761, 0.00000279, 0.00000123, -0.00000962, -0.00000573, 0.00000399, 0.00000183, -0.00000148, 0.00000127, -0.00000308, -0.00000061, 0.00000086, -0.00000523, -0.00000147, -0.00000334, -0.00000252, 0.00000351, 0.00000282, -0.00000459, -0.00000265, -0.00000512, -0.00000389, 0.00000248, -0.00000206, -0.00000440, -0.00000754, 0.00000149, -0.00000937, -0.00001287, 0.00000540, -0.00000516, -0.00000322, },
		{ -0.00001208, 0.00000140, -0.00000972, -0.00003102, -0.00007419, -0.00001926, 0.00009184, 0.00007075, 0.00002951, 0.00003536, 0.00005054, 0.00002423, 0.00008974, 0.00009854, -0.00009076, -0.00007233, -0.00000971, -0.00002387, -0.00000901, -0.00001710, -0.00000481, -0.00001082, 0.00000079, -0.00000728, -0.00000404, -0.00000978, -0.00000763, -0.00000878, -0.00001663, 0.00000268, -0.00000436, 0.00003133, -0.00000521, -0.00000359, 0.00001255, -0.00000381, 0.00001439, -0.00000590, 0.00001482, -0.00000485, 0.00000083, 0.00000240, 0.00002316, 0.00001159, -0.00004128, 0.00003125, -0.00000027, -0.00002745, 0.00001437, -0.00000754, -0.00001225, 0.00001631, -0.00001295, 0.00000110, 0.00002081, -0.00002989, 0.00001426, -0.00001381, -0.00000313, 0.00001599, -0.00001271, 0.00002137, -0.00001415, 0.00000843, -0.00000785, -0.00000624, 0.00001165, -0.00001895, 0.00001431, -0.00000413, -0.00000243, 0.00000231, -0.00000131, 0.00000642, -0.00000336, -0.00000173, -0.00000459, 0.00000676, -0.00000172, -0.00000449, 0.00000480, -0.00000526, 0.00000430, 0.00001465, -0.00000811, -0.00000932, 0.00000862, -0.00000087, -0.00000341, -0.00000033, -0.00000163, 0.00001163, 0.00000009, -0.00001946, 0.00000255, 0.00000732, 0.00000051, -0.00000096, -0.00001060, -0.00000147, 0.00000494, -0.00000078, -0.00000056, 0.00000097, -0.00000447, 0.00000200, -0.00000046, -0.00000603, 0.00000002, -0.00000444, -0.00000197, 0.00000466, 0.00000162, -0.00000542, -0.00000249, -0.00000433, -0.00000239, 0.00000123, -0.00000265, -0.00000522, -0.00000632, 0.00000210, -0.00001239, -0.00000971, 0.00000566, -0.00000656, -0.00000230, -0.00000915, },
		{ 0.00000274, -0.00001218, -0.00004532, -0.00006886, 0.00001853, 0.00009535, 0.00005025, 0.00002736, 0.00003724, 0.00004429, 0.00002483, 0.00009787, 0.00007614, -0.00009983, -0.00005773, -0.00000995, -0.00002312, -0.00000950, -0.00001506, -0.00000589, -0.00000834, 0.00000049, -0.00000875, -0.00000228, -0.00001328, -0.00000355, -0.00001184, -0.00001706, 0.00000246, 0.00000024, 0.00003172, -0.00001074, 0.00000254, 0.00001102, -0.00000200, 0.00001272, -0.00000438, 0.00001496, -0.00000538, 0.00000031, 0.00000529, 0.00003070, -0.00000608, -0.00002988, 0.00003621, -0.00002007, -0.00001310, 0.00001457, -0.00001834, 0.00000014, 0.00001082, -0.00001747, 0.00001499, 0.00000616, -0.00002515, 0.00001622, -0.00002041, 0.00000823, 0.00000436, -0.00000315, 0.00001898, -0.00001848, 0.00001363, -0.00001554, 0.00000244, 0.00000415, -0.00001551, 0.00001759, -0.00001029, 0.00000232, 0.00000004, 0.00000030, 0.00000480, -0.00000340, -0.00000142, -0.00000393, 0.00000874, -0.00000607, -0.00000341, 0.00000504, -0.00000538, 0.00000997, 0.00000998, -0.00001378, -0.00000232, 0.00000796, -0.00000419, -0.00000022, -0.00000228, -0.00000014, 0.00001388, -0.00000741, -0.00001771, 0.00000931, 0.00000453, -0.00000121, -0.00000320, -0.00001003, 0.00000305, 0.00000376, -0.00000261, 0.00000100, -0.00000065, -0.00000442, 0.00000445, -0.00000308, -0.00000542, 0.00000099, -0.00000591, -0.00000064, 0.00000526, 0.00000024, -0.00000604, -0.00000198, -0.00000316, -0.00000183, -0.00000009, -0.00000292, -0.00000604, -0.00000448, 0.00000150, -0.00001479, -0.00000574, 0.00000459, -0.00000739, -0.00000134, -0.00001158, -0.00001101, },
		{ -0.00001922, -0.00005752, -0.00005091, 0.00005045, 0.00008619, 0.00003477, 0.00002842, 0.00003682, 0.00003881, 0.00002738, 0.00010336, 0.00005275, -0.00010450, -0.00004376, -0.00001173, -0.00002223, -0.00000911, -0.00001402, -0.00000624, -0.00000561, -0.00000148, -0.00000820, -0.00000254, -0.00001521, 0.00000074, -0.00001708, -0.00001592, 0.00000243, 0.00000533, 0.00003028, -0.00001488, 0.00000937, 0.00000831, -0.00000015, 0.00001145, -0.00000308, 0.00001542, -0.00000642, -0.00000060, 0.00001275, 0.00003071, -0.00002072, -0.00001260, 0.00002888, -0.00003174, 0.00000445, 0.00000478, -0.00002067, 0.00001090, -0.00000011, -0.00001288, 0.00002182, -0.00000821, -0.00001545, 0.00001189, -0.00002058, 0.00001443, -0.00000538, 0.00000935, 0.00001024, -0.00001726, 0.00001513, -0.00002075, 0.00001045, -0.00000509, -0.00000780, 0.00001620, -0.00001404, 0.00000711, -0.00000317, 0.00000252, 0.00000248, -0.00000234, -0.00000197, -0.00000208, 0.00000914, -0.00001072, -0.00000059, 0.00000409, -0.00000350, 0.00001428, 0.00000202, -0.00001498, 0.00000461, 0.00000406, -0.00000465, 0.00000211, -0.00000511, 0.00000365, 0.00001321, -0.00001484, -0.00001199, 0.00001376, 0.00000023, -0.00000213, -0.00000545, -0.00000759, 0.00000669, 0.00000091, -0.00000289, 0.00000213, -0.00000295, -0.00000264, 0.00000565, -0.00000609, -0.00000344, 0.00000083, -0.00000708, 0.00000115, 0.00000522, -0.00000112, -0.00000651, -0.00000086, -0.00000209, -0.00000224, -0.00000100, -0.00000315, -0.00000661, -0.00000231, -0.00000042, -0.00001582, -0.00000171, 0.00000218, -0.00000702, -0.00000088, -0.00001456, -0.00000879, 0.00000430, },
		{ -0.00006291, -0.00002423, 0.00007135, 0.00006989, 0.00002579, 0.00002989, 0.00003513, 0.00003423, 0.00003141, 0.00010613, 0.00002882, -0.00010455, -0.00003126, -0.00001495, -0.00002037, -0.00000870, -0.00001367, -0.00000540, -0.00000384, -0.00000381, -0.00000644, -0.00000477, -0.00001459, 0.00000345, -0.00002331, -0.00001313, 0.00000216, 0.00001100, 0.00002677, -0.00001678, 0.00001606, 0.00000424, 0.00000261, 0.00000980, -0.00000158, 0.00001626, -0.00000907, 0.00000085, 0.00002168, 0.00002313, -0.00002760, 0.00000365, 0.00001253, -0.00003132, 0.00001743, -0.00000996, -0.00001378, 0.00001480, -0.00001046, -0.00000232, 0.00002001, -0.00001767, -0.00000531, 0.00000416, -0.00001531, 0.00001417, -0.00000962, 0.00002033, -0.00000203, -0.00001091, 0.00001199, -0.00002183, 0.00001554, -0.00001347, 0.00000226, 0.00001024, -0.00001395, 0.00001033, -0.00000642, 0.00000493, -0.00000009, -0.00000055, -0.00000318, 0.00000091, 0.00000724, -0.00001432, 0.00000311, 0.00000230, 0.00000064, 0.00001525, -0.00000693, -0.00001132, 0.00000890, -0.00000109, -0.00000225, 0.00000217, -0.00000710, 0.00000863, 0.00000899, -0.00001997, -0.00000359, 0.00001474, -0.00000427, -0.00000246, -0.00000735, -0.00000342, 0.00000834, -0.00000251, -0.00000158, 0.00000199, -0.00000487, 0.00000042, 0.00000496, -0.00000840, -0.00000072, -0.00000060, -0.00000741, 0.00000289, 0.00000474, -0.00000242, -0.00000672, 0.00000095, -0.00000173, -0.00000323, -0.00000129, -0.00000365, -0.00000659, -0.00000038, -0.00000332, -0.00001501, 0.00000137, -0.00000095, -0.00000517, -0.00000180, -0.00001742, -0.00000485, 0.00000408, -0.00003022, },
		{ 0.00000533, 0.00007965, 0.00005203, 0.00002235, 0.00003024, 0.00003318, 0.00003027, 0.00003683, 0.00010575, 0.00000519, -0.00010004, -0.00002133, -0.00001876, -0.00001734, -0.00000908, -0.00001317, -0.00000374, -0.00000350, -0.00000541, -0.00000471, -0.00000798, -0.00001148, 0.00000333, -0.00002895, -0.00000936, 0.00000180, 0.00001699, 0.00002130, -0.00001569, 0.00002131, -0.00000048, 0.00000652, 0.00000709, 0.00000109, 0.00001641, -0.00001288, 0.00000640, 0.00002799, 0.00001120, -0.00002548, 0.00001262, -0.00000595, -0.00002000, 0.00002057, -0.00002267, -0.00000133, 0.00001039, -0.00001513, 0.00000889, 0.00001166, -0.00002031, 0.00000155, -0.00000324, -0.00000762, 0.00000877, -0.00000696, 0.00002600, -0.00001392, -0.00000162, 0.00000486, -0.00001836, 0.00001621, -0.00001849, 0.00001192, 0.00000128, -0.00000975, 0.00001071, -0.00000862, 0.00000696, -0.00000241, 0.00000146, -0.00000455, 0.00000445, 0.00000298, -0.00001575, 0.00000656, 0.00000068, 0.00000615, 0.00001179, -0.00001395, -0.00000424, 0.00000910, -0.00000509, 0.00000164, -0.00000031, -0.00000675, 0.00001291, 0.00000190, -0.00002109, 0.00000538, 0.00001218, -0.00000782, -0.00000269, -0.00000827, 0.00000176, 0.00000745, -0.00000517, 0.00000060, 0.00000033, -0.00000537, 0.00000366, 0.00000251, -0.00000917, 0.00000179, -0.00000289, -0.00000672, 0.00000416, 0.00000410, -0.00000384, -0.00000635, 0.00000322, -0.00000255, -0.00000411, -0.00000115, -0.00000447, -0.00000575, 0.00000063, -0.00000640, -0.00001245, 0.00000259, -0.00000366, -0.00000230, -0.00000472, -0.00001887, 0.00000025, 0.00000022, -0.00004245, -0.00005882, },
		{ 0.00007699, 0.00003672, 0.00002235, 0.00002922, 0.00003148, 0.00002678, 0.00004358, 0.00010167, -0.00001686, -0.00009168, -0.00001471, -0.00002198, -0.00001371, -0.00001044, -0.00001176, -0.00000216, -0.00000427, -0.00000586, -0.00000397, -0.00001080, -0.00000695, 0.00000012, -0.00003276, -0.00000558, 0.00000192, 0.00002258, 0.00001460, -0.00001149, 0.00002386, -0.00000450, 0.00001087, 0.00000335, 0.00000559, 0.00001434, -0.00001587, 0.00001565, 0.00002856, -0.00000012, -0.00001715, 0.00001146, -0.00001917, -0.00000340, 0.00001317, -0.00002779, 0.00001070, 0.00000030, -0.00001213, 0.00001575, 0.00000109, -0.00001729, 0.00000361, -0.00000743, -0.00000106, 0.00000148, 0.00000131, 0.00002455, -0.00002189, 0.00000750, -0.00000422, -0.00001136, 0.00001238, -0.00001862, 0.00001842, -0.00000813, -0.00000244, 0.00000775, -0.00000887, 0.00000796, -0.00000394, 0.00000316, -0.00000550, 0.00000759, -0.00000297, -0.00001458, 0.00000886, 0.00000043, 0.00001121, 0.00000435, -0.00001662, 0.00000355, 0.00000549, -0.00000619, 0.00000498, -0.00000428, -0.00000359, 0.00001458, -0.00000617, -0.00001770, 0.00001264, 0.00000700, -0.00000982, -0.00000318, -0.00000757, 0.00000675, 0.00000425, -0.00000606, 0.00000249, -0.00000229, -0.00000396, 0.00000587, -0.00000086, -0.00000820, 0.00000325, -0.00000531, -0.00000527, 0.00000487, 0.00000351, -0.00000549, -0.00000495, 0.00000528, -0.00000458, -0.00000426, -0.00000106, -0.00000532, -0.00000424, 0.00000030, -0.00000862, -0.00000899, 0.00000169, -0.00000476, 0.00000038, -0.00000940, -0.00001772, 0.00000480, -0.00000784, -0.00005219, -0.00004541, 0.00003192, },
		{ 0.00002599, 0.00002360, 0.00002732, 0.00003002, 0.00002394, 0.00005124, 0.00009359, -0.00003591, -0.00008080, -0.00001143, -0.00002362, -0.00001046, -0.00001231, -0.00000923, -0.00000149, -0.00000534, -0.00000548, -0.00000446, -0.00001219, -0.00000251, -0.00000545, -0.00003422, -0.00000255, 0.00000310, 0.00002691, 0.00000788, -0.00000494, 0.00002323, -0.00000653, 0.00001439, -0.00000040, 0.00001154, 0.00000914, -0.00001546, 0.00002599, 0.00002286, -0.00000649, -0.00000805, 0.00000188, -0.00002238, 0.00001114, -0.00000087, -0.00002353, 0.00001696, -0.00001036, -0.00000322, 0.00001578, -0.00000727, -0.00001178, 0.00000180, -0.00000753, 0.00000186, -0.00000396, 0.00001190, 0.00001668, -0.00002395, 0.00001358, -0.00001262, -0.00000296, 0.00000538, -0.00001380, 0.00001989, -0.00001518, 0.00000589, 0.00000202, -0.00000676, 0.00000747, -0.00000429, 0.00000418, -0.00000565, 0.00000939, -0.00000938, -0.00001114, 0.00000976, 0.00000223, 0.00001365, -0.00000505, -0.00001402, 0.00000914, -0.00000007, -0.00000401, 0.00000604, -0.00000791, 0.00000155, 0.00001255, -0.00001279, -0.00001071, 0.00001656, 0.00000072, -0.00001035, -0.00000392, -0.00000498, 0.00001020, -0.00000025, -0.00000484, 0.00000298, -0.00000474, -0.00000099, 0.00000620, -0.00000395, -0.00000599, 0.00000330, -0.00000717, -0.00000356, 0.00000520, 0.00000287, -0.00000725, -0.00000223, 0.00000625, -0.00000729, -0.00000340, -0.00000152, -0.00000566, -0.00000263, -0.00000123, -0.00000916, -0.00000595, -0.00000067, -0.00000363, 0.00000138, -0.00001459, -0.00001365, 0.00000664, -0.00001884, -0.00005622, -0.00002537, 0.00005139, 0.00006694, },
		{ 0.00002458, 0.00002521, 0.00002852, 0.00002220, 0.00005898, 0.00008180, -0.00005089, -0.00006904, -0.00001074, -0.00002337, -0.00000843, -0.00001379, -0.00000601, -0.00000204, -0.00000596, -0.00000499, -0.00000575, -0.00001181, 0.00000052, -0.00001214, -0.00003361, -0.00000047, 0.00000563, 0.00002937, 0.00000242, 0.00000253, 0.00001989, -0.00000598, 0.00001594, -0.00000267, 0.00001741, 0.00000144, -0.00000991, 0.00003369, 0.00001325, -0.00000630, -0.00000363, -0.00001071, -0.00001543, 0.00001781, -0.00001517, -0.00001238, 0.00001511, -0.00001652, 0.00000726, 0.00000978, -0.00001080, -0.00000718, -0.00000136, -0.00000477, 0.00000066, -0.00000494, 0.00002085, 0.00000517, -0.00002027, 0.00001505, -0.00001806, 0.00000438, -0.00000244, -0.00000556, 0.00001601, -0.00001776, 0.00001273, -0.00000498, -0.00000263, 0.00000536, -0.00000321, 0.00000430, -0.00000494, 0.00000923, -0.00001495, -0.00000634, 0.00000978, 0.00000579, 0.00001191, -0.00001342, -0.00000715, 0.00001060, -0.00000504, 0.00000029, 0.00000415, -0.00000938, 0.00000678, 0.00000713, -0.00001589, -0.00000215, 0.00001658, -0.00000516, -0.00000996, -0.00000445, -0.00000080, 0.00001112, -0.00000458, -0.00000211, 0.00000156, -0.00000575, 0.00000245, 0.00000460, -0.00000576, -0.00000351, 0.00000220, -0.00000816, -0.00000203, 0.00000551, 0.00000186, -0.00000862, 0.00000162, 0.00000537, -0.00000972, -0.00000182, -0.00000268, -0.00000506, -0.00000175, -0.00000317, -0.00000792, -0.00000456, -0.00000307, -0.00000070, -0.00000033, -0.00001840, -0.00000770, 0.00000421, -0.00003005, -0.00005296, -0.00000276, 0.00006197, 0.00005352, 0.00002000, },
		{ 0.00002337, 0.00002679, 0.00002196, 0.00006574, 0.00006721, -0.00006139, -0.00005783, -0.00001146, -0.00002162, -0.00000792, -0.00001407, -0.00000299, -0.00000352, -0.00000582, -0.00000502, -0.00000701, -0.00001012, 0.00000147, -0.00001869, -0.00003164, 0.00000090, 0.00000931, 0.00002988, -0.00000092, 0.00000935, 0.00001513, -0.00000328, 0.00001517, -0.00000226, 0.00002114, -0.00000659, 0.00000055, 0.00003569, 0.00000388, -0.00000141, -0.00000680, -0.00001985, -0.00000250, 0.00001471, -0.00002387, 0.00000038, 0.00000667, -0.00001554, 0.00001458, 0.00000106, -0.00000956, -0.00000557, -0.00000327, -0.00000162, -0.00000307, -0.00000092, 0.00002511, -0.00000633, -0.00001285, 0.00001201, -0.00001943, 0.00000874, -0.00000857, 0.00000350, 0.00000813, -0.00001514, 0.00001595, -0.00001126, 0.00000252, 0.00000196, -0.00000070, 0.00000346, -0.00000357, 0.00000702, -0.00001865, -0.00000122, 0.00000987, 0.00000975, 0.00000583, -0.00001792, 0.00000140, 0.00000774, -0.00000728, 0.00000462, 0.00000011, -0.00000781, 0.00000995, -0.00000001, -0.00001456, 0.00000560, 0.00001333, -0.00000966, -0.00000926, -0.00000411, 0.00000401, 0.00000928, -0.00000733, 0.00000090, -0.00000136, -0.00000461, 0.00000497, 0.00000186, -0.00000592, -0.00000166, 0.00000059, -0.00000841, -0.00000083, 0.00000599, 0.00000017, -0.00000886, 0.00000582, 0.00000244, -0.00001089, -0.00000027, -0.00000416, -0.00000357, -0.00000221, -0.00000434, -0.00000572, -0.00000531, -0.00000396, 0.00000257, -0.00000462, -0.00001918, -0.00000209, -0.00000249, -0.00003836, -0.00004297, 0.00001847, 0.00006410, 0.00004011, 0.00001780, 0.00002457, },
		{ 0.00002483, 0.00002333, 0.00007060, 0.00005105, -0.00006770, -0.00004801, -0.00001247, -0.00001925, -0.00000859, -0.00001285, -0.00000097, -0.00000524, -0.00000513, -0.00000576, -0.00000751, -0.00000803, 0.00000043, -0.00002426, -0.00002910, 0.00000209, 0.00001366, 0.00002886, -0.00000203, 0.00001443, 0.00001045, 0.00000040, 0.00001279, 0.00000093, 0.00002123, -0.00001202, 0.00001351, 0.00003115, -0.00000136, 0.00000380, -0.00001652, -0.00002093, 0.00001003, 0.00000433, -0.00002416, 0.00000930, -0.00000398, -0.00000816, 0.00001580, -0.00000636, -0.00000579, -0.00000692, -0.00000257, -0.00000040, -0.00000660, 0.00000645, 0.00002358, -0.00001469, -0.00000457, 0.00000607, -0.00001710, 0.00000946, -0.00001122, 0.00001060, -0.00000117, -0.00000823, 0.00001450, -0.00001495, 0.00000722, -0.00000191, 0.00000289, 0.00000167, -0.00000193, 0.00000318, -0.00002004, 0.00000351, 0.00001084, 0.00001215, -0.00000311, -0.00001700, 0.00000848, 0.00000211, -0.00000598, 0.00000697, -0.00000429, -0.00000370, 0.00000961, -0.00000643, -0.00000936, 0.00001060, 0.00000822, -0.00001252, -0.00000854, -0.00000237, 0.00000813, 0.00000535, -0.00000775, 0.00000281, -0.00000463, -0.00000144, 0.00000552, -0.00000077, -0.00000484, -0.00000091, -0.00000085, -0.00000836, 0.00000027, 0.00000649, -0.00000219, -0.00000737, 0.00000917, -0.00000197, -0.00001028, 0.00000041, -0.00000523, -0.00000180, -0.00000395, -0.00000376, -0.00000403, -0.00000755, -0.00000245, 0.00000432, -0.00001005, -0.00001650, 0.00000070, -0.00001173, -0.00004149, -0.00002848, 0.00003549, 0.00005961, 0.00002892, 0.00001797, 0.00002344, 0.00002201, },
		{ 0.00002612, 0.00007300, 0.00003454, -0.00007058, -0.00003975, -0.00001311, -0.00001710, -0.00000963, -0.00001042, -0.00000037, -0.00000652, -0.00000444, -0.00000690, -0.00000699, -0.00000645, -0.00000198, -0.00002861, -0.00002650, 0.00000355, 0.00001817, 0.00002702, -0.00000143, 0.00001750, 0.00000686, 0.00000378, 0.00001022, 0.00000579, 0.00001757, -0.00001247, 0.00002527, 0.00002186, -0.00000075, 0.00000461, -0.00002837, -0.00001321, 0.00001653, -0.00000798, -0.00001720, 0.00001123, -0.00001189, 0.00000204, 0.00001090, -0.00000960, -0.00000249, -0.00000951, 0.00000034, -0.00000211, -0.00000740, 0.00001421, 0.00001734, -0.00001837, 0.00000199, -0.00000052, -0.00001258, 0.00000721, -0.00000997, 0.00001376, -0.00000892, 0.00000069, 0.00000879, -0.00001493, 0.00001005, -0.00000511, 0.00000693, -0.00000091, -0.00000038, -0.00000158, -0.00001921, 0.00000769, 0.00001280, 0.00001125, -0.00001216, -0.00001104, 0.00001162, -0.00000365, -0.00000198, 0.00000628, -0.00000701, 0.00000124, 0.00000564, -0.00000982, -0.00000211, 0.00001198, 0.00000282, -0.00001412, -0.00000764, 0.00000080, 0.00001035, 0.00000067, -0.00000602, 0.00000270, -0.00000677, 0.00000267, 0.00000384, -0.00000211, -0.00000349, -0.00000109, -0.00000178, -0.00000841, 0.00000166, 0.00000653, -0.00000468, -0.00000406, 0.00001052, -0.00000667, -0.00000815, -0.00000017, -0.00000515, -0.00000079, -0.00000611, -0.00000126, -0.00000430, -0.00000969, 0.00000105, 0.00000323, -0.00001442, -0.00001152, -0.00000087, -0.00002056, -0.00003886, -0.00001233, 0.00004698, 0.00005084, 0.00002105, 0.00001916, 0.00002142, 0.00002118, 0.00002281, },
		{ 0.00007280, 0.00001859, -0.00007087, -0.00003278, -0.00001328, -0.00001568, -0.00001022, -0.00000751, -0.00000108, -0.00000698, -0.00000426, -0.00000783, -0.00000576, -0.00000593, -0.00000501, -0.00003189, -0.00002406, 0.00000555, 0.00002251, 0.00002489, 0.00000005, 0.00001897, 0.00000459, 0.00000611, 0.00000882, 0.00001039, 0.00001172, -0.00000728, 0.00003236, 0.00001154, 0.00000435, -0.00000168, -0.00003658, 0.00000001, 0.00001456, -0.00001668, -0.00000720, 0.00000649, -0.00001374, 0.00001053, 0.00000246, -0.00000810, -0.00000189, -0.00001105, 0.00000365, -0.00000598, -0.00000437, 0.00001945, 0.00000892, -0.00001768, 0.00000543, -0.00000578, -0.00000778, 0.00000366, -0.00000584, 0.00001247, -0.00001286, 0.00000876, 0.00000058, -0.00001125, 0.00001010, -0.00000651, 0.00001049, -0.00000401, 0.00000083, -0.00000649, -0.00001657, 0.00001156, 0.00001499, 0.00000637, -0.00001831, -0.00000224, 0.00001001, -0.00000706, 0.00000269, 0.00000301, -0.00000681, 0.00000475, -0.00000056, -0.00000887, 0.00000473, 0.00001012, -0.00000172, -0.00001505, -0.00000602, 0.00000473, 0.00001009, -0.00000323, -0.00000319, 0.00000057, -0.00000666, 0.00000616, 0.00000066, -0.00000164, -0.00000287, -0.00000152, -0.00000236, -0.00000862, 0.00000356, 0.00000559, -0.00000641, 0.00000041, 0.00000926, -0.00001028, -0.00000541, -0.00000173, -0.00000371, -0.00000142, -0.00000725, 0.00000220, -0.00000712, -0.00000990, 0.00000494, -0.00000067, -0.00001583, -0.00000671, -0.00000657, -0.00002610, -0.00003165, 0.00000309, 0.00005282, 0.00004000, 0.00001676, 0.00002018, 0.00001894, 0.00002082, 0.00002089, 0.00002992, },
		{ 0.00000377, -0.00006919, -0.00002673, -0.00001328, -0.00001495, -0.00000987, -0.00000493, -0.00000257, -0.00000667, -0.00000479, -0.00000799, -0.00000446, -0.00000647, -0.00000805, -0.00003444, -0.00002173, 0.00000819, 0.00002660, 0.00002266, 0.00000182, 0.00001959, 0.00000319, 0.00000753, 0.00000917, 0.00001302, 0.00000630, 0.00000203, 0.00003311, 0.00000400, 0.00001014, -0.00001431, -0.00003681, 0.00001308, 0.00000568, -0.00001850, 0.00000082, -0.00000159, -0.00000927, 0.00001390, -0.00000569, -0.00000356, -0.00000443, -0.00001002, 0.00000530, -0.00000995, 0.00000175, 0.00002052, 0.00000114, -0.00001435, 0.00000583, -0.00000875, -0.00000419, 0.00000065, -0.00000080, 0.00000779, -0.00001212, 0.00001353, -0.00000766, -0.00000512, 0.00000735, -0.00000535, 0.00001260, -0.00000720, 0.00000158, -0.00001095, -0.00001253, 0.00001544, 0.00001597, -0.00000161, -0.00001945, 0.00000617, 0.00000480, -0.00000682, 0.00000579, -0.00000115, -0.00000384, 0.00000507, -0.00000660, -0.00000383, 0.00000903, 0.00000633, -0.00000499, -0.00001568, -0.00000315, 0.00000830, 0.00000761, -0.00000527, -0.00000067, -0.00000264, -0.00000407, 0.00000762, -0.00000261, 0.00000026, -0.00000354, -0.00000146, -0.00000313, -0.00000863, 0.00000576, 0.00000352, -0.00000644, 0.00000474, 0.00000572, -0.00001186, -0.00000322, -0.00000337, -0.00000149, -0.00000388, -0.00000605, 0.00000481, -0.00001172, -0.00000712, 0.00000706, -0.00000581, -0.00001373, -0.00000473, -0.00001431, -0.00002668, -0.00002207, 0.00001636, 0.00005357, 0.00002901, 0.00001571, 0.00002018, 0.00001657, 0.00002080, 0.00001920, 0.00003429, 0.00007012, },
		{ -0.00006594, -0.00002137, -0.00001350, -0.00001451, -0.00000855, -0.00000329, -0.00000419, -0.00000599, -0.00000586, -0.00000716, -0.00000376, -0.00000772, -0.00001083, -0.00003661, -0.00001939, 0.00001156, 0.00003042, 0.00002019, 0.00000370, 0.00001985, 0.00000200, 0.00000884, 0.00001077, 0.00001301, 0.00000382, 0.00001242, 0.00002825, 0.00000144, 0.00001224, -0.00002932, -0.00002805, 0.00002061, -0.00000559, -0.00001372, 0.00000342, -0.00000851, -0.00000109, 0.00001128, -0.00001038, 0.00000120, -0.00000880, -0.00000638, 0.00000405, -0.00001175, 0.00000879, 0.00001754, -0.00000404, -0.00001050, 0.00000436, -0.00000961, -0.00000234, -0.00000059, 0.00000306, 0.00000189, -0.00000750, 0.00001373, -0.00001355, 0.00000151, 0.00000272, -0.00000155, 0.00001244, -0.00000992, 0.00000179, -0.00001454, -0.00000733, 0.00001925, 0.00001430, -0.00001048, -0.00001523, 0.00001129, -0.00000149, -0.00000337, 0.00000590, -0.00000407, 0.00000037, 0.00000174, -0.00000999, 0.00000364, 0.00000965, 0.00000229, -0.00000731, -0.00001592, 0.00000111, 0.00001028, 0.00000400, -0.00000524, 0.00000034, -0.00000545, 0.00000021, 0.00000641, -0.00000451, 0.00000252, -0.00000531, -0.00000055, -0.00000456, -0.00000787, 0.00000761, 0.00000078, -0.00000440, 0.00000747, 0.00000109, -0.00001136, -0.00000238, -0.00000405, 0.00000030, -0.00000728, -0.00000210, 0.00000469, -0.00001612, -0.00000177, 0.00000584, -0.00000971, -0.00000943, -0.00000712, -0.00002095, -0.00002251, -0.00001228, 0.00002699, 0.00004999, 0.00001945, 0.00001704, 0.00001863, 0.00001493, 0.00002082, 0.00001786, 0.00003888, 0.00006517, -0.00000960, },
		{ -0.00001678, -0.00001415, -0.00001381, -0.00000665, -0.00000277, -0.00000545, -0.00000536, -0.00000700, -0.00000553, -0.00000413, -0.00000913, -0.00001339, -0.00003858, -0.00001688, 0.00001576, 0.00003379, 0.00001727, 0.00000596, 0.00001976, 0.00000073, 0.00001090, 0.00001233, 0.00001111, 0.00000555, 0.00002057, 0.00002048, 0.00000338, 0.00000778, -0.00004127, -0.00001310, 0.00001987, -0.00001410, -0.00000566, 0.00000014, -0.00001082, 0.00000680, 0.00000441, -0.00001035, 0.00000366, -0.00001280, -0.00000154, 0.00000022, -0.00001003, 0.00001417, 0.00001219, -0.00000614, -0.00000770, 0.00000249, -0.00000934, -0.00000174, 0.00000007, 0.00000442, -0.00000287, -0.00000107, 0.00000972, -0.00001562, 0.00000663, -0.00000218, 0.00000414, 0.00000966, -0.00001160, 0.00000134, -0.00001694, -0.00000107, 0.00002231, 0.00000927, -0.00001745, -0.00000725, 0.00001163, -0.00000608, 0.00000140, 0.00000316, -0.00000422, 0.00000359, -0.00000416, -0.00000912, 0.00001109, 0.00000676, -0.00000070, -0.00000936, -0.00001519, 0.00000625, 0.00000991, 0.00000070, -0.00000383, -0.00000053, -0.00000649, 0.00000460, 0.00000298, -0.00000413, 0.00000382, -0.00000735, 0.00000098, -0.00000678, -0.00000590, 0.00000831, -0.00000161, -0.00000074, 0.00000766, -0.00000305, -0.00000960, -0.00000290, -0.00000319, 0.00000041, -0.00000993, 0.00000364, 0.00000088, -0.00001796, 0.00000416, 0.00000131, -0.00001024, -0.00000568, -0.00001341, -0.00002368, -0.00001544, -0.00000356, 0.00003492, 0.00004294, 0.00001262, 0.00001952, 0.00001557, 0.00001452, 0.00002045, 0.00001708, 0.00004332, 0.00005819, -0.00002123, -0.00006133, },
		{ -0.00001514, -0.00001246, -0.00000476, -0.00000318, -0.00000613, -0.00000510, -0.00000774, -0.00000361, -0.00000562, -0.00001025, -0.00001592, -0.00004048, -0.00001400, 0.00002093, 0.00003630, 0.00001396, 0.00000891, 0.00001891, -0.00000030, 0.00001402, 0.00001255, 0.00000920, 0.00001097, 0.00002429, 0.00001309, 0.00000706, -0.00000311, -0.00004566, 0.00000277, 0.00001182, -0.00001639, 0.00000117, -0.00000657, -0.00000754, 0.00001089, -0.00000341, -0.00000655, 0.00000274, -0.00001458, 0.00000246, -0.00000457, -0.00000503, 0.00001611, 0.00000677, -0.00000607, -0.00000646, 0.00000127, -0.00000904, -0.00000131, 0.00000176, 0.00000319, -0.00000484, 0.00000465, 0.00000317, -0.00001375, 0.00000883, -0.00000559, 0.00001030, 0.00000453, -0.00001181, 0.00000008, -0.00001779, 0.00000602, 0.00002346, 0.00000143, -0.00002021, 0.00000158, 0.00000767, -0.00000713, 0.00000506, -0.00000089, -0.00000147, 0.00000385, -0.00001036, -0.00000383, 0.00001610, 0.00000164, -0.00000209, -0.00001168, -0.00001279, 0.00001113, 0.00000727, -0.00000108, -0.00000235, -0.00000269, -0.00000513, 0.00000747, -0.00000130, -0.00000158, 0.00000328, -0.00000862, 0.00000240, -0.00000931, -0.00000284, 0.00000742, -0.00000246, 0.00000322, 0.00000534, -0.00000547, -0.00000790, -0.00000394, -0.00000113, -0.00000171, -0.00001005, 0.00000904, -0.00000595, -0.00001568, 0.00000797, -0.00000468, -0.00000680, -0.00000540, -0.00002110, -0.00002129, -0.00000789, 0.00000394, 0.00004002, 0.00003355, 0.00000932, 0.00002165, 0.00001158, 0.00001553, 0.00001924, 0.00001721, 0.00004717, 0.00004944, -0.00003083, -0.00005560, -0.00001316, },
		{ -0.00001039, -0.00000341, -0.00000412, -0.00000628, -0.00000527, -0.00000776, -0.00000201, -0.00000796, -0.00001080, -0.00001867, -0.00004233, -0.00001041, 0.00002703, 0.00003737, 0.00001074, 0.00001257, 0.00001688, -0.00000029, 0.00001765, 0.00001088, 0.00000931, 0.00001803, 0.00002328, 0.00000838, 0.00000881, -0.00001732, -0.00004091, 0.00001414, 0.00000040, -0.00001218, 0.00000326, -0.00001275, -0.00000046, 0.00000972, -0.00000894, -0.00000141, -0.00000086, -0.00001349, 0.00000407, -0.00000812, 0.00000155, 0.00001434, 0.00000316, -0.00000539, -0.00000623, 0.00000088, -0.00000928, -0.00000013, 0.00000308, 0.00000048, -0.00000367, 0.00000770, -0.00000356, -0.00000913, 0.00000786, -0.00000620, 0.00001525, -0.00000202, -0.00001053, -0.00000204, -0.00001665, 0.00001335, 0.00002159, -0.00000741, -0.00001793, 0.00000822, 0.00000158, -0.00000454, 0.00000581, -0.00000399, 0.00000281, 0.00000041, -0.00001431, 0.00000449, 0.00001717, -0.00000388, -0.00000218, -0.00001433, -0.00000830, 0.00001437, 0.00000331, -0.00000091, -0.00000204, -0.00000482, -0.00000183, 0.00000779, -0.00000476, 0.00000213, 0.00000080, -0.00000840, 0.00000289, -0.00001132, 0.00000061, 0.00000518, -0.00000104, 0.00000583, 0.00000163, -0.00000581, -0.00000732, -0.00000431, 0.00000095, -0.00000544, -0.00000673, 0.00001162, -0.00001353, -0.00000946, 0.00000755, -0.00000911, -0.00000100, -0.00001016, -0.00002670, -0.00001467, -0.00000176, 0.00001069, 0.00004188, 0.00002333, 0.00000958, 0.00002214, 0.00000771, 0.00001761, 0.00001701, 0.00001860, 0.00004985, 0.00003937, -0.00003811, -0.00004916, -0.00001071, -0.00001612, },
		{ -0.00000291, -0.00000515, -0.00000613, -0.00000569, -0.00000702, -0.00000124, -0.00001068, -0.00001076, -0.00002190, -0.00004406, -0.00000570, 0.00003370, 0.00003658, 0.00000833, 0.00001645, 0.00001364, 0.00000152, 0.00002062, 0.00000791, 0.00001269, 0.00002412, 0.00001901, 0.00000667, 0.00000597, -0.00003009, -0.00002895, 0.00001771, -0.00000938, -0.00000417, -0.00000030, -0.00001499, 0.00000697, 0.00000426, -0.00001043, 0.00000253, -0.00000534, -0.00001032, 0.00000296, -0.00000880, 0.00000734, 0.00001005, 0.00000205, -0.00000547, -0.00000599, 0.00000078, -0.00000988, 0.00000211, 0.00000289, -0.00000201, -0.00000038, 0.00000730, -0.00000836, -0.00000369, 0.00000459, -0.00000365, 0.00001751, -0.00000868, -0.00000814, -0.00000483, -0.00001318, 0.00001978, 0.00001622, -0.00001479, -0.00001156, 0.00001064, -0.00000381, 0.00000013, 0.00000334, -0.00000429, 0.00000631, -0.00000585, -0.00001416, 0.00001350, 0.00001408, -0.00000815, -0.00000183, -0.00001676, -0.00000195, 0.00001486, -0.00000046, 0.00000067, -0.00000347, -0.00000550, 0.00000205, 0.00000561, -0.00000610, 0.00000549, -0.00000286, -0.00000667, 0.00000198, -0.00001202, 0.00000338, 0.00000263, 0.00000245, 0.00000583, -0.00000178, -0.00000480, -0.00000809, -0.00000317, 0.00000157, -0.00000908, -0.00000056, 0.00000976, -0.00001887, -0.00000147, 0.00000256, -0.00000922, 0.00000385, -0.00001920, -0.00002729, -0.00000634, 0.00000246, 0.00001707, 0.00004012, 0.00001411, 0.00001259, 0.00002023, 0.00000513, 0.00001995, 0.00001396, 0.00002142, 0.00005072, 0.00002863, -0.00004295, -0.00004252, -0.00000946, -0.00001667, -0.00000782, },
		{ -0.00000590, -0.00000592, -0.00000610, -0.00000576, -0.00000155, -0.00001326, -0.00001027, -0.00002581, -0.00004546, 0.00000055, 0.00004023, 0.00003392, 0.00000744, 0.00001971, 0.00000979, 0.00000536, 0.00002167, 0.00000523, 0.00001916, 0.00002719, 0.00001368, 0.00000636, -0.00000180, -0.00003715, -0.00001432, 0.00001358, -0.00001376, 0.00000346, -0.00000766, -0.00001191, 0.00001140, -0.00000277, -0.00000815, 0.00000374, -0.00000877, -0.00000675, 0.00000019, -0.00000630, 0.00001039, 0.00000529, 0.00000286, -0.00000683, -0.00000488, 0.00000015, -0.00001007, 0.00000481, 0.00000089, -0.00000278, 0.00000320, 0.00000410, -0.00001011, 0.00000063, 0.00000068, 0.00000136, 0.00001632, -0.00001409, -0.00000541, -0.00000772, -0.00000740, 0.00002389, 0.00000798, -0.00001855, -0.00000351, 0.00000865, -0.00000624, 0.00000442, -0.00000103, -0.00000130, 0.00000685, -0.00001276, -0.00000948, 0.00002073, 0.00000779, -0.00001016, -0.00000196, -0.00001805, 0.00000528, 0.00001228, -0.00000254, 0.00000235, -0.00000623, -0.00000389, 0.00000483, 0.00000205, -0.00000501, 0.00000725, -0.00000650, -0.00000405, -0.00000016, -0.00001112, 0.00000462, 0.00000119, 0.00000674, 0.00000296, -0.00000334, -0.00000388, -0.00000946, -0.00000059, -0.00000017, -0.00001051, 0.00000630, 0.00000360, -0.00001960, 0.00000501, -0.00000504, -0.00000404, 0.00000418, -0.00002941, -0.00002198, 0.00000082, 0.00000560, 0.00002272, 0.00003489, 0.00000758, 0.00001674, 0.00001607, 0.00000467, 0.00002154, 0.00001078, 0.00002542, 0.00004936, 0.00001808, -0.00004548, -0.00003620, -0.00000922, -0.00001644, -0.00000518, -0.00000330, },
		},
		{ { 0.00013395, 0.00002970, -0.00004047, -0.00000236, 0.00001171, 0.00000286, 0.00004692, 0.00003792, -0.00000620, -0.00004067, -0.00002933, -0.00000603, -0.00001230, -0.00000669, 0.00000387, -0.00001365, -0.00000087, -0.00001017, 0.00000433, -0.00000530, -0.00000738, -0.00001151, 0.00001313, 0.00000119, -0.00000631, -0.00000528, -0.00000116, 0.00000453, 0.00000132, -0.00000425, 0.00000543, -0.00000406, 0.00000013, 0.00000231, 0.00000618, -0.00000574, 0.00000602, -0.00000348, -0.00000370, -0.00001075, 0.00001108, -0.00000366, 0.00000149, 0.00000274, -0.00000174, -0.00000143, 0.00000011, -0.00000174, 0.00000006, 0.00000065, 0.00000171, -0.00000554, -0.00000037, -0.00000751, 0.00001302, -0.00000565, 0.00000634, 0.00000245, -0.00000398, -0.00000803, 0.00000626, -0.00000001, -0.00000045, -0.00000173, -0.00000020, -0.00000173, -0.00000045, -0.00000001, 0.00000626, -0.00000803, -0.00000398, 0.00000245, 0.00000634, -0.00000565, 0.00001302, -0.00000751, -0.00000037, -0.00000554, 0.00000171, 0.00000065, 0.00000006, -0.00000174, 0.00000011, -0.00000143, -0.00000174, 0.00000274, 0.00000149, -0.00000366, 0.00001108, -0.00001075, -0.00000370, -0.00000348, 0.00000602, -0.00000574, 0.00000618, 0.00000231, 0.00000013, -0.00000406, 0.00000543, -0.00000425, 0.00000132, 0.00000453, -0.00000116, -0.00000528, -0.00000631, 0.00000119, 0.00001313, -0.00001151, -0.00000738, -0.00000530, 0.00000433, -0.00001017, -0.00000087, -0.00001365, 0.00000387, -0.00000669, -0.00001230, -0.00000603, -0.00002933, -0.00004067, -0.00000620, 0.00003792, 0.00004692, 0.00000286, 0.00001171, -0.00000236, -0.00004047, 0.00002970, },
		{ 0.00003166, -0.00003622, -0.00000234, 0.00001340, 0.00000296, 0.00005357, 0.00003697, -0.00001489, -0.00004287, -0.00002448, -0.00000789, -0.00001230, -0.00000543, 0.00000243, -0.00001234, -0.00000244, -0.00000876, 0.00000479, -0.00000709, -0.00000766, -0.00001046, 0.00001501, -0.00000161, -0.00000587, -0.00000515, 0.00000014, 0.00000441, 0.00000026, -0.00000365, 0.00000530, -0.00000416, 0.00000002, 0.00000363, 0.00000492, -0.00000540, 0.00000617, -0.00000432, -0.00000444, -0.00000882, 0.00001121, -0.00000519, 0.00000334, 0.00000121, -0.00000134, -0.00000174, 0.00000030, -0.00000162, -0.00000020, 0.00000162, 0.00000030, -0.00000456, -0.00000187, -0.00000524, 0.00001228, -0.00000618, 0.00000741, 0.00000139, -0.00000502, -0.00000663, 0.00000666, -0.00000072, -0.00000039, -0.00000166, -0.00000025, -0.00000172, -0.00000053, 0.00000079, 0.00000552, -0.00000921, -0.00000273, 0.00000328, 0.00000517, -0.00000472, 0.00001321, -0.00000950, 0.00000119, -0.00000664, 0.00000333, -0.00000071, 0.00000074, -0.00000224, 0.00000026, -0.00000143, -0.00000181, 0.00000404, -0.00000042, -0.00000171, 0.00001027, -0.00001232, -0.00000287, -0.00000265, 0.00000566, -0.00000581, 0.00000729, 0.00000087, 0.00000046, -0.00000404, 0.00000554, -0.00000486, 0.00000262, 0.00000419, -0.00000213, -0.00000571, -0.00000628, 0.00000392, 0.00001064, -0.00001203, -0.00000718, -0.00000326, 0.00000328, -0.00001113, 0.00000052, -0.00001492, 0.00000553, -0.00000859, -0.00001134, -0.00000512, -0.00003401, -0.00003703, 0.00000156, 0.00003776, 0.00004067, 0.00000314, 0.00000989, -0.00000227, -0.00004465, 0.00002757, 0.00013410, },
		{ -0.00003191, -0.00000221, 0.00001496, 0.00000350, 0.00006058, 0.00003461, -0.00002420, -0.00004341, -0.00001994, -0.00001035, -0.00001143, -0.00000488, 0.00000137, -0.00001121, -0.00000399, -0.00000706, 0.00000474, -0.00000865, -0.00000797, -0.00000891, 0.00001614, -0.00000422, -0.00000513, -0.00000521, 0.00000172, 0.00000380, -0.00000050, -0.00000313, 0.00000518, -0.00000433, 0.00000015, 0.00000472, 0.00000359, -0.00000485, 0.00000611, -0.00000519, -0.00000499, -0.00000665, 0.00001071, -0.00000619, 0.00000496, -0.00000040, -0.00000074, -0.00000226, 0.00000075, -0.00000182, -0.00000003, 0.00000213, -0.00000081, -0.00000379, -0.00000318, -0.00000283, 0.00001105, -0.00000626, 0.00000829, 0.00000014, -0.00000578, -0.00000509, 0.00000675, -0.00000130, -0.00000035, -0.00000153, -0.00000039, -0.00000164, -0.00000063, 0.00000168, 0.00000443, -0.00001007, -0.00000133, 0.00000383, 0.00000398, -0.00000343, 0.00001281, -0.00001106, 0.00000264, -0.00000772, 0.00000504, -0.00000237, 0.00000182, -0.00000314, 0.00000081, -0.00000180, -0.00000144, 0.00000493, -0.00000220, 0.00000051, 0.00000884, -0.00001344, -0.00000202, -0.00000183, 0.00000508, -0.00000556, 0.00000815, -0.00000061, 0.00000096, -0.00000407, 0.00000559, -0.00000540, 0.00000407, 0.00000341, -0.00000274, -0.00000650, -0.00000562, 0.00000632, 0.00000773, -0.00001206, -0.00000706, -0.00000102, 0.00000162, -0.00001148, 0.00000152, -0.00001595, 0.00000718, -0.00001093, -0.00000945, -0.00000537, -0.00003804, -0.00003230, 0.00000821, 0.00003677, 0.00003485, 0.00000373, 0.00000794, -0.00000205, -0.00004876, 0.00002527, 0.00013456, 0.00003349, },
		{ -0.00000195, 0.00001640, 0.00000459, 0.00006785, 0.00003052, -0.00003369, -0.00004225, -0.00001612, -0.00001305, -0.00000989, -0.00000503, 0.00000079, -0.00001037, -0.00000534, -0.00000527, 0.00000430, -0.00001003, -0.00000821, -0.00000695, 0.00001644, -0.00000643, -0.00000430, -0.00000533, 0.00000346, 0.00000276, -0.00000090, -0.00000273, 0.00000508, -0.00000456, 0.00000054, 0.00000553, 0.00000229, -0.00000416, 0.00000587, -0.00000609, -0.00000528, -0.00000439, 0.00000965, -0.00000661, 0.00000621, -0.00000194, -0.00000004, -0.00000287, 0.00000136, -0.00000226, 0.00000049, 0.00000219, -0.00000155, -0.00000332, -0.00000417, -0.00000045, 0.00000945, -0.00000590, 0.00000890, -0.00000124, -0.00000623, -0.00000351, 0.00000657, -0.00000178, -0.00000032, -0.00000136, -0.00000059, -0.00000149, -0.00000074, 0.00000261, 0.00000300, -0.00001053, 0.00000012, 0.00000413, 0.00000285, -0.00000189, 0.00001182, -0.00001211, 0.00000387, -0.00000865, 0.00000669, -0.00000422, 0.00000321, -0.00000439, 0.00000175, -0.00000258, -0.00000055, 0.00000526, -0.00000364, 0.00000279, 0.00000686, -0.00001406, -0.00000120, -0.00000102, 0.00000431, -0.00000498, 0.00000868, -0.00000203, 0.00000157, -0.00000413, 0.00000556, -0.00000581, 0.00000555, 0.00000229, -0.00000301, -0.00000764, -0.00000422, 0.00000813, 0.00000466, -0.00001170, -0.00000696, 0.00000135, -0.00000061, -0.00001115, 0.00000198, -0.00001650, 0.00000854, -0.00001343, -0.00000680, -0.00000690, -0.00004100, -0.00002690, 0.00001374, 0.00003517, 0.00002947, 0.00000457, 0.00000584, -0.00000169, -0.00005280, 0.00002279, 0.00013530, 0.00003520, -0.00002755, },
		{ 0.00001769, 0.00000637, 0.00007517, 0.00002446, -0.00004283, -0.00003952, -0.00001331, -0.00001561, -0.00000794, -0.00000576, 0.00000069, -0.00000992, -0.00000635, -0.00000355, 0.00000360, -0.00001130, -0.00000828, -0.00000475, 0.00001596, -0.00000809, -0.00000355, -0.00000537, 0.00000522, 0.00000138, -0.00000096, -0.00000249, 0.00000501, -0.00000485, 0.00000118, 0.00000600, 0.00000111, -0.00000338, 0.00000547, -0.00000700, -0.00000526, -0.00000219, 0.00000817, -0.00000648, 0.00000703, -0.00000329, 0.00000062, -0.00000348, 0.00000202, -0.00000281, 0.00000129, 0.00000184, -0.00000193, -0.00000320, -0.00000474, 0.00000174, 0.00000761, -0.00000513, 0.00000919, -0.00000268, -0.00000635, -0.00000196, 0.00000616, -0.00000214, -0.00000030, -0.00000117, -0.00000085, -0.00000128, -0.00000084, 0.00000353, 0.00000127, -0.00001054, 0.00000152, 0.00000417, 0.00000186, -0.00000019, 0.00001029, -0.00001258, 0.00000476, -0.00000930, 0.00000815, -0.00000609, 0.00000480, -0.00000592, 0.00000305, -0.00000377, 0.00000089, 0.00000493, -0.00000452, 0.00000489, 0.00000450, -0.00001420, -0.00000045, -0.00000025, 0.00000337, -0.00000405, 0.00000881, -0.00000329, 0.00000221, -0.00000419, 0.00000542, -0.00000602, 0.00000694, 0.00000096, -0.00000303, -0.00000907, -0.00000205, 0.00000914, 0.00000170, -0.00001112, -0.00000677, 0.00000368, -0.00000331, -0.00001014, 0.00000178, -0.00001639, 0.00000933, -0.00001575, -0.00000366, -0.00000964, -0.00004255, -0.00002131, 0.00001823, 0.00003310, 0.00002455, 0.00000559, 0.00000360, -0.00000118, -0.00005679, 0.00002014, 0.00013631, 0.00003685, -0.00002319, -0.00000153, },
		{ 0.00000900, 0.00008225, 0.00001627, -0.00005101, -0.00003552, -0.00001165, -0.00001772, -0.00000587, -0.00000691, 0.00000100, -0.00000985, -0.00000693, -0.00000205, 0.00000274, -0.00001252, -0.00000804, -0.00000251, 0.00001480, -0.00000913, -0.00000300, -0.00000519, 0.00000684, -0.00000024, -0.00000071, -0.00000241, 0.00000497, -0.00000515, 0.00000204, 0.00000611, 0.00000011, -0.00000258, 0.00000493, -0.00000792, -0.00000488, -0.00000020, 0.00000642, -0.00000587, 0.00000738, -0.00000438, 0.00000118, -0.00000399, 0.00000262, -0.00000337, 0.00000224, 0.00000116, -0.00000196, -0.00000342, -0.00000482, 0.00000360, 0.00000566, -0.00000401, 0.00000912, -0.00000410, -0.00000615, -0.00000053, 0.00000559, -0.00000240, -0.00000026, -0.00000100, -0.00000112, -0.00000103, -0.00000090, 0.00000438, -0.00000071, -0.00001007, 0.00000277, 0.00000401, 0.00000107, 0.00000155, 0.00000831, -0.00001245, 0.00000523, -0.00000959, 0.00000927, -0.00000783, 0.00000643, -0.00000761, 0.00000462, -0.00000530, 0.00000287, 0.00000386, -0.00000469, 0.00000657, 0.00000195, -0.00001391, 0.00000022, 0.00000046, 0.00000233, -0.00000282, 0.00000852, -0.00000431, 0.00000283, -0.00000422, 0.00000515, -0.00000595, 0.00000811, -0.00000046, -0.00000292, -0.00001066, 0.00000083, 0.00000919, -0.00000092, -0.00001047, -0.00000637, 0.00000579, -0.00000627, -0.00000854, 0.00000091, -0.00001551, 0.00000931, -0.00001754, -0.00000042, -0.00001339, -0.00004253, -0.00001597, 0.00002186, 0.00003065, 0.00002013, 0.00000669, 0.00000127, -0.00000054, -0.00006071, 0.00001730, 0.00013759, 0.00003846, -0.00001886, -0.00000092, 0.00001881, },
		{ 0.00008868, 0.00000593, -0.00005766, -0.00003068, -0.00001116, -0.00001915, -0.00000393, -0.00000828, 0.00000160, -0.00001010, -0.00000709, -0.00000085, 0.00000182, -0.00001373, -0.00000739, -0.00000046, 0.00001314, -0.00000957, -0.00000274, -0.00000470, 0.00000819, -0.00000196, -0.00000020, -0.00000248, 0.00000496, -0.00000543, 0.00000308, 0.00000586, -0.00000065, -0.00000182, 0.00000426, -0.00000880, -0.00000414, 0.00000145, 0.00000457, -0.00000488, 0.00000730, -0.00000516, 0.00000157, -0.00000435, 0.00000309, -0.00000383, 0.00000321, 0.00000024, -0.00000172, -0.00000395, -0.00000440, 0.00000500, 0.00000378, -0.00000263, 0.00000868, -0.00000543, -0.00000565, 0.00000074, 0.00000490, -0.00000257, -0.00000020, -0.00000087, -0.00000138, -0.00000078, -0.00000089, 0.00000510, -0.00000285, -0.00000910, 0.00000380, 0.00000369, 0.00000051, 0.00000321, 0.00000598, -0.00001175, 0.00000524, -0.00000942, 0.00000994, -0.00000927, 0.00000795, -0.00000932, 0.00000633, -0.00000706, 0.00000529, 0.00000206, -0.00000402, 0.00000761, -0.00000056, -0.00001331, 0.00000085, 0.00000104, 0.00000126, -0.00000134, 0.00000781, -0.00000503, 0.00000337, -0.00000420, 0.00000475, -0.00000559, 0.00000895, -0.00000181, -0.00000281, -0.00001223, 0.00000428, 0.00000824, -0.00000298, -0.00000993, -0.00000561, 0.00000748, -0.00000926, -0.00000651, -0.00000056, -0.00001382, 0.00000833, -0.00001846, 0.00000247, -0.00001777, -0.00004092, -0.00001124, 0.00002483, 0.00002785, 0.00001627, 0.00000777, -0.00000111, 0.00000019, -0.00006455, 0.00001425, 0.00013913, 0.00004006, -0.00001458, -0.00000009, 0.00001973, 0.00001267, },
		{ -0.00000637, -0.00006229, -0.00002554, -0.00001165, -0.00001979, -0.00000236, -0.00000965, 0.00000233, -0.00001056, -0.00000687, 0.00000001, 0.00000089, -0.00001495, -0.00000626, 0.00000121, 0.00001119, -0.00000948, -0.00000278, -0.00000384, 0.00000914, -0.00000364, 0.00000049, -0.00000269, 0.00000496, -0.00000565, 0.00000422, 0.00000527, -0.00000114, -0.00000115, 0.00000349, -0.00000961, -0.00000306, 0.00000267, 0.00000277, -0.00000365, 0.00000683, -0.00000564, 0.00000178, -0.00000452, 0.00000337, -0.00000409, 0.00000409, -0.00000079, -0.00000129, -0.00000472, -0.00000350, 0.00000588, 0.00000208, -0.00000110, 0.00000788, -0.00000660, -0.00000490, 0.00000180, 0.00000413, -0.00000266, -0.00000012, -0.00000080, -0.00000160, -0.00000054, -0.00000080, 0.00000562, -0.00000505, -0.00000767, 0.00000452, 0.00000329, 0.00000019, 0.00000470, 0.00000342, -0.00001054, 0.00000478, -0.00000878, 0.00001006, -0.00001028, 0.00000920, -0.00001088, 0.00000804, -0.00000892, 0.00000802, -0.00000039, -0.00000246, 0.00000782, -0.00000279, -0.00001252, 0.00000149, 0.00000144, 0.00000028, 0.00000029, 0.00000671, -0.00000542, 0.00000377, -0.00000412, 0.00000421, -0.00000491, 0.00000939, -0.00000297, -0.00000287, -0.00001357, 0.00000809, 0.00000632, -0.00000437, -0.00000963, -0.00000438, 0.00000852, -0.00001201, -0.00000430, -0.00000246, -0.00001143, 0.00000631, -0.00001825, 0.00000457, -0.00002232, -0.00003789, -0.00000736, 0.00002736, 0.00002471, 0.00001303, 0.00000869, -0.00000344, 0.00000097, -0.00006826, 0.00001094, 0.00014097, 0.00004165, -0.00001036, 0.00000098, 0.00002045, 0.00001757, 0.00009395, },
		{ -0.00006454, -0.00002064, -0.00001287, -0.00001965, -0.00000127, -0.00001086, 0.00000304, -0.00001109, -0.00000635, 0.00000054, -0.00000005, -0.00001615, -0.00000464, 0.00000230, 0.00000919, -0.00000899, -0.00000308, -0.00000260, 0.00000961, -0.00000516, 0.00000127, -0.00000299, 0.00000495, -0.00000576, 0.00000540, 0.00000439, -0.00000136, -0.00000062, 0.00000264, -0.00001029, -0.00000170, 0.00000342, 0.00000117, -0.00000231, 0.00000607, -0.00000585, 0.00000182, -0.00000450, 0.00000343, -0.00000410, 0.00000478, -0.00000181, -0.00000078, -0.00000562, -0.00000218, 0.00000620, 0.00000069, 0.00000047, 0.00000677, -0.00000754, -0.00000394, 0.00000263, 0.00000333, -0.00000269, -0.00000002, -0.00000079, -0.00000175, -0.00000033, -0.00000059, 0.00000586, -0.00000720, -0.00000584, 0.00000489, 0.00000288, 0.00000010, 0.00000593, 0.00000079, -0.00000891, 0.00000391, -0.00000766, 0.00000959, -0.00001076, 0.00001002, -0.00001213, 0.00000956, -0.00001067, 0.00001087, -0.00000336, -0.00000005, 0.00000706, -0.00000454, -0.00001170, 0.00000220, 0.00000157, -0.00000052, 0.00000193, 0.00000531, -0.00000548, 0.00000403, -0.00000397, 0.00000357, -0.00000394, 0.00000938, -0.00000384, -0.00000321, -0.00001447, 0.00001197, 0.00000358, -0.00000504, -0.00000964, -0.00000260, 0.00000874, -0.00001423, -0.00000219, -0.00000453, -0.00000851, 0.00000335, -0.00001677, 0.00000549, -0.00002651, -0.00003376, -0.00000444, 0.00002959, 0.00002124, 0.00001046, 0.00000936, -0.00000563, 0.00000171, -0.00007178, 0.00000731, 0.00014315, 0.00004320, -0.00000618, 0.00000227, 0.00002102, 0.00002383, 0.00009748, -0.00002025, },
		{ -0.00001652, -0.00001443, -0.00001887, -0.00000072, -0.00001180, 0.00000360, -0.00001156, -0.00000566, 0.00000079, -0.00000101, -0.00001723, -0.00000257, 0.00000271, 0.00000737, -0.00000828, -0.00000354, -0.00000105, 0.00000956, -0.00000641, 0.00000206, -0.00000335, 0.00000491, -0.00000571, 0.00000654, 0.00000327, -0.00000129, -0.00000025, 0.00000173, -0.00001080, -0.00000014, 0.00000369, -0.00000013, -0.00000099, 0.00000511, -0.00000585, 0.00000173, -0.00000433, 0.00000328, -0.00000382, 0.00000518, -0.00000273, -0.00000029, -0.00000653, -0.00000055, 0.00000596, -0.00000029, 0.00000195, 0.00000540, -0.00000823, -0.00000284, 0.00000322, 0.00000253, -0.00000264, 0.00000010, -0.00000086, -0.00000181, -0.00000018, -0.00000027, 0.00000578, -0.00000916, -0.00000369, 0.00000488, 0.00000254, 0.00000021, 0.00000683, -0.00000179, -0.00000699, 0.00000268, -0.00000611, 0.00000855, -0.00001064, 0.00001031, -0.00001293, 0.00001074, -0.00001214, 0.00001361, -0.00000661, 0.00000310, 0.00000529, -0.00000561, -0.00001100, 0.00000307, 0.00000135, -0.00000098, 0.00000346, 0.00000371, -0.00000526, 0.00000413, -0.00000376, 0.00000287, -0.00000273, 0.00000893, -0.00000436, -0.00000394, -0.00001470, 0.00001560, 0.00000023, -0.00000502, -0.00000997, -0.00000028, 0.00000804, -0.00001568, -0.00000048, -0.00000648, -0.00000536, -0.00000037, -0.00001402, 0.00000494, -0.00002982, -0.00002897, -0.00000244, 0.00003163, 0.00001747, 0.00000858, 0.00000966, -0.00000759, 0.00000233, -0.00007503, 0.00000328, 0.00014575, 0.00004464, -0.00000196, 0.00000373, 0.00002156, 0.00003153, 0.00009864, -0.00003515, -0.00006422, },
		{ -0.00001594, -0.00001766, -0.00000065, -0.00001242, 0.00000393, -0.00001184, -0.00000491, 0.00000083, -0.00000204, -0.00001810, -0.00000018, 0.00000239, 0.00000592, -0.00000754, -0.00000402, 0.00000072, 0.00000899, -0.00000730, 0.00000278, -0.00000373, 0.00000483, -0.00000548, 0.00000756, 0.00000199, -0.00000097, -0.00000006, 0.00000079, -0.00001110, 0.00000152, 0.00000352, -0.00000106, 0.00000018, 0.00000407, -0.00000570, 0.00000155, -0.00000406, 0.00000298, -0.00000327, 0.00000526, -0.00000345, 0.00000008, -0.00000733, 0.00000125, 0.00000523, -0.00000083, 0.00000324, 0.00000385, -0.00000864, -0.00000165, 0.00000357, 0.00000177, -0.00000254, 0.00000023, -0.00000099, -0.00000180, -0.00000008, 0.00000018, 0.00000533, -0.00001082, -0.00000135, 0.00000451, 0.00000231, 0.00000047, 0.00000738, -0.00000419, -0.00000490, 0.00000119, -0.00000422, 0.00000696, -0.00000990, 0.00001000, -0.00001315, 0.00001142, -0.00001313, 0.00001601, -0.00000990, 0.00000678, 0.00000256, -0.00000591, -0.00001056, 0.00000416, 0.00000072, -0.00000101, 0.00000470, 0.00000205, -0.00000482, 0.00000411, -0.00000352, 0.00000217, -0.00000135, 0.00000806, -0.00000451, -0.00000510, -0.00001412, 0.00001868, -0.00000342, -0.00000444, -0.00001055, 0.00000252, 0.00000638, -0.00001618, 0.00000055, -0.00000798, -0.00000234, -0.00000450, -0.00001018, 0.00000280, -0.00003178, -0.00002402, -0.00000118, 0.00003347, 0.00001348, 0.00000738, 0.00000953, -0.00000924, 0.00000276, -0.00007796, -0.00000123, 0.00014884, 0.00004591, 0.00000239, 0.00000523, 0.00002225, 0.00004059, 0.00009684, -0.00005031, -0.00006141, -0.00001359, },
		{ -0.00001628, -0.00000093, -0.00001275, 0.00000399, -0.00001186, -0.00000422, 0.00000074, -0.00000319, -0.00001862, 0.00000236, 0.00000137, 0.00000498, -0.00000693, -0.00000438, 0.00000258, 0.00000797, -0.00000782, 0.00000335, -0.00000409, 0.00000469, -0.00000503, 0.00000839, 0.00000062, -0.00000043, -0.00000008, -0.00000015, -0.00001115, 0.00000318, 0.00000296, -0.00000160, 0.00000113, 0.00000304, -0.00000547, 0.00000134, -0.00000374, 0.00000259, -0.00000250, 0.00000502, -0.00000392, 0.00000026, -0.00000791, 0.00000308, 0.00000410, -0.00000090, 0.00000424, 0.00000219, -0.00000875, -0.00000044, 0.00000370, 0.00000106, -0.00000238, 0.00000033, -0.00000116, -0.00000170, -0.00000003, 0.00000072, 0.00000450, -0.00001206, 0.00000107, 0.00000381, 0.00000226, 0.00000082, 0.00000755, -0.00000632, -0.00000277, -0.00000043, -0.00000210, 0.00000494, -0.00000860, 0.00000908, -0.00001273, 0.00001149, -0.00001349, 0.00001783, -0.00001293, 0.00001073, -0.00000101, -0.00000538, -0.00001045, 0.00000551, -0.00000039, -0.00000050, 0.00000550, 0.00000050, -0.00000428, 0.00000402, -0.00000329, 0.00000153, 0.00000009, 0.00000686, -0.00000434, -0.00000666, -0.00001262, 0.00002090, -0.00000707, -0.00000350, -0.00001124, 0.00000564, 0.00000383, -0.00001563, 0.00000068, -0.00000873, 0.00000018, -0.00000865, -0.00000555, -0.00000088, -0.00003206, -0.00001942, -0.00000042, 0.00003505, 0.00000940, 0.00000676, 0.00000896, -0.00001053, 0.00000293, -0.00008049, -0.00000630, 0.00015252, 0.00004690, 0.00000700, 0.00000662, 0.00002339, 0.00005081, 0.00009156, -0.00006487, -0.00005638, -0.00001213, -0.00001703, },
		{ -0.00000139, -0.00001289, 0.00000382, -0.00001159, -0.00000365, 0.00000056, -0.00000450, -0.00001867, 0.00000485, -0.00000021, 0.00000459, -0.00000658, -0.00000447, 0.00000436, 0.00000660, -0.00000795, 0.00000376, -0.00000440, 0.00000450, -0.00000436, 0.00000897, -0.00000076, 0.00000028, -0.00000030, -0.00000105, -0.00001094, 0.00000473, 0.00000213, -0.00000176, 0.00000180, 0.00000207, -0.00000522, 0.00000116, -0.00000346, 0.00000219, -0.00000158, 0.00000448, -0.00000413, 0.00000020, -0.00000818, 0.00000480, 0.00000270, -0.00000054, 0.00000490, 0.00000053, -0.00000858, 0.00000075, 0.00000362, 0.00000042, -0.00000217, 0.00000041, -0.00000136, -0.00000154, -0.00000002, 0.00000132, 0.00000331, -0.00001280, 0.00000341, 0.00000283, 0.00000240, 0.00000120, 0.00000737, -0.00000808, -0.00000072, -0.00000208, 0.00000013, 0.00000259, -0.00000680, 0.00000757, -0.00001166, 0.00001089, -0.00001309, 0.00001886, -0.00001541, 0.00001462, -0.00000519, -0.00000406, -0.00001073, 0.00000715, -0.00000198, 0.00000063, 0.00000572, -0.00000077, -0.00000375, 0.00000392, -0.00000313, 0.00000105, 0.00000148, 0.00000546, -0.00000395, -0.00000854, -0.00001020, 0.00002205, -0.00001038, -0.00000243, -0.00001186, 0.00000886, 0.00000057, -0.00001406, -0.00000019, -0.00000849, 0.00000182, -0.00001233, -0.00000061, -0.00000586, -0.00003050, -0.00001562, 0.00000016, 0.00003622, 0.00000542, 0.00000660, 0.00000798, -0.00001147, 0.00000284, -0.00008260, -0.00001196, 0.00015683, 0.00004755, 0.00001201, 0.00000772, 0.00002536, 0.00006177, 0.00008248, -0.00007791, -0.00004962, -0.00001224, -0.00001741, -0.00001503, },
		{ -0.00001296, 0.00000352, -0.00001107, -0.00000324, 0.00000033, -0.00000598, -0.00001813, 0.00000704, -0.00000218, 0.00000474, -0.00000656, -0.00000417, 0.00000589, 0.00000501, -0.00000773, 0.00000397, -0.00000466, 0.00000426, -0.00000347, 0.00000924, -0.00000206, 0.00000111, -0.00000073, -0.00000187, -0.00001047, 0.00000608, 0.00000111, -0.00000158, 0.00000218, 0.00000123, -0.00000498, 0.00000105, -0.00000326, 0.00000186, -0.00000061, 0.00000370, -0.00000409, -0.00000011, -0.00000806, 0.00000626, 0.00000115, 0.00000021, 0.00000517, -0.00000107, -0.00000816, 0.00000186, 0.00000336, -0.00000012, -0.00000192, 0.00000044, -0.00000155, -0.00000134, -0.00000001, 0.00000193, 0.00000179, -0.00001297, 0.00000556, 0.00000168, 0.00000274, 0.00000155, 0.00000687, -0.00000943, 0.00000115, -0.00000364, 0.00000232, 0.00000007, -0.00000464, 0.00000556, -0.00000998, 0.00000962, -0.00001188, 0.00001894, -0.00001705, 0.00001811, -0.00000971, -0.00000208, -0.00001137, 0.00000904, -0.00000403, 0.00000240, 0.00000525, -0.00000162, -0.00000337, 0.00000390, -0.00000311, 0.00000080, 0.00000267, 0.00000400, -0.00000346, -0.00001057, -0.00000694, 0.00002198, -0.00001307, -0.00000147, -0.00001217, 0.00001189, -0.00000315, -0.00001161, -0.00000207, -0.00000713, 0.00000229, -0.00001508, 0.00000412, -0.00001173, -0.00002714, -0.00001295, 0.00000086, 0.00003682, 0.00000173, 0.00000673, 0.00000668, -0.00001209, 0.00000250, -0.00008430, -0.00001823, 0.00016180, 0.00004778, 0.00001754, 0.00000834, 0.00002862, 0.00007285, 0.00006949, -0.00008853, -0.00004180, -0.00001382, -0.00001688, -0.00001413, -0.00000183, },
		{ 0.00000320, -0.00001037, -0.00000297, 0.00000002, -0.00000759, -0.00001697, 0.00000873, -0.00000430, 0.00000534, -0.00000690, -0.00000341, 0.00000702, 0.00000335, -0.00000724, 0.00000399, -0.00000485, 0.00000401, -0.00000239, 0.00000919, -0.00000323, 0.00000200, -0.00000133, -0.00000257, -0.00000977, 0.00000718, 0.00000004, -0.00000115, 0.00000227, 0.00000054, -0.00000479, 0.00000103, -0.00000318, 0.00000168, 0.00000031, 0.00000278, -0.00000386, -0.00000064, -0.00000755, 0.00000736, -0.00000039, 0.00000124, 0.00000505, -0.00000253, -0.00000751, 0.00000287, 0.00000296, -0.00000057, -0.00000163, 0.00000040, -0.00000171, -0.00000113, 0.00000002, 0.00000249, 0.00000001, -0.00001256, 0.00000738, 0.00000045, 0.00000327, 0.00000180, 0.00000612, -0.00001036, 0.00000276, -0.00000501, 0.00000436, -0.00000248, -0.00000226, 0.00000318, -0.00000777, 0.00000773, -0.00000986, 0.00001799, -0.00001764, 0.00002086, -0.00001422, 0.00000039, -0.00001230, 0.00001111, -0.00000645, 0.00000477, 0.00000405, -0.00000191, -0.00000327, 0.00000406, -0.00000329, 0.00000087, 0.00000354, 0.00000262, -0.00000304, -0.00001255, -0.00000304, 0.00002066, -0.00001492, -0.00000085, -0.00001197, 0.00001442, -0.00000697, -0.00000852, -0.00000483, -0.00000465, 0.00000141, -0.00001649, 0.00000805, -0.00001791, -0.00002223, -0.00001158, 0.00000194, 0.00003669, -0.00000144, 0.00000693, 0.00000520, -0.00001250, 0.00000198, -0.00008566, -0.00002505, 0.00016741, 0.00004761, 0.00002367, 0.00000833, 0.00003362, 0.00008324, 0.00005276, -0.00009597, -0.00003369, -0.00001659, -0.00001541, -0.00001376, -0.00000207, -0.00001310, },
		{ -0.00000960, -0.00000277, -0.00000040, -0.00000925, -0.00001517, 0.00000975, -0.00000635, 0.00000622, -0.00000753, -0.00000215, 0.00000764, 0.00000177, -0.00000655, 0.00000385, -0.00000499, 0.00000376, -0.00000115, 0.00000880, -0.00000419, 0.00000288, -0.00000210, -0.00000311, -0.00000889, 0.00000796, -0.00000100, -0.00000054, 0.00000211, -0.00000000, -0.00000465, 0.00000109, -0.00000326, 0.00000169, 0.00000110, 0.00000181, -0.00000350, -0.00000133, -0.00000666, 0.00000803, -0.00000179, 0.00000245, 0.00000456, -0.00000379, -0.00000668, 0.00000373, 0.00000242, -0.00000091, -0.00000133, 0.00000031, -0.00000181, -0.00000093, 0.00000010, 0.00000293, -0.00000192, -0.00001156, 0.00000878, -0.00000076, 0.00000395, 0.00000193, 0.00000518, -0.00001086, 0.00000406, -0.00000612, 0.00000612, -0.00000489, 0.00000017, 0.00000059, -0.00000518, 0.00000533, -0.00000711, 0.00001599, -0.00001702, 0.00002256, -0.00001837, 0.00000309, -0.00001337, 0.00001323, -0.00000913, 0.00000767, 0.00000209, -0.00000155, -0.00000355, 0.00000449, -0.00000372, 0.00000131, 0.00000398, 0.00000149, -0.00000286, -0.00001422, 0.00000125, 0.00001818, -0.00001582, -0.00000073, -0.00001107, 0.00001615, -0.00001053, -0.00000513, -0.00000822, -0.00000118, -0.00000086, -0.00001629, 0.00001064, -0.00002376, -0.00001620, -0.00001152, 0.00000354, 0.00003570, -0.00000390, 0.00000699, 0.00000371, -0.00001282, 0.00000138, -0.00008678, -0.00003233, 0.00017354, 0.00004711, 0.00003040, 0.00000761, 0.00004081, 0.00009197, 0.00003279, -0.00009964, -0.00002609, -0.00002012, -0.00001310, -0.00001400, -0.00000200, -0.00001346, 0.00000299, },
		{ -0.00000258, -0.00000104, -0.00001082, -0.00001283, 0.00000998, -0.00000807, 0.00000721, -0.00000835, -0.00000046, 0.00000768, 0.00000041, -0.00000577, 0.00000360, -0.00000510, 0.00000357, 0.00000019, 0.00000809, -0.00000490, 0.00000369, -0.00000298, -0.00000346, -0.00000788, 0.00000844, -0.00000192, 0.00000016, 0.00000175, -0.00000041, -0.00000454, 0.00000122, -0.00000346, 0.00000194, 0.00000166, 0.00000088, -0.00000311, -0.00000211, -0.00000544, 0.00000822, -0.00000294, 0.00000372, 0.00000374, -0.00000480, -0.00000572, 0.00000443, 0.00000180, -0.00000113, -0.00000103, 0.00000015, -0.00000185, -0.00000077, 0.00000025, 0.00000317, -0.00000390, -0.00001004, 0.00000968, -0.00000185, 0.00000471, 0.00000189, 0.00000413, -0.00001098, 0.00000503, -0.00000692, 0.00000754, -0.00000704, 0.00000249, -0.00000205, -0.00000237, 0.00000260, -0.00000378, 0.00001303, -0.00001515, 0.00002297, -0.00002183, 0.00000576, -0.00001443, 0.00001524, -0.00001187, 0.00001095, -0.00000055, -0.00000051, -0.00000432, 0.00000525, -0.00000447, 0.00000219, 0.00000389, 0.00000074, -0.00000309, -0.00001531, 0.00000556, 0.00001474, -0.00001574, -0.00000120, -0.00000934, 0.00001684, -0.00001345, -0.00000184, -0.00001190, 0.00000300, -0.00000440, -0.00001435, 0.00001147, -0.00002863, -0.00000960, -0.00001256, 0.00000570, 0.00003382, -0.00000552, 0.00000672, 0.00000236, -0.00001318, 0.00000083, -0.00008780, -0.00003993, 0.00018005, 0.00004644, 0.00003764, 0.00000623, 0.00005052, 0.00009799, 0.00001037, -0.00009927, -0.00001975, -0.00002388, -0.00001018, -0.00001480, -0.00000154, -0.00001412, 0.00000302, -0.00000888, },
		{ -0.00000197, -0.00001214, -0.00001010, 0.00000939, -0.00000928, 0.00000808, -0.00000919, 0.00000159, 0.00000711, -0.00000063, -0.00000500, 0.00000328, -0.00000521, 0.00000347, 0.00000157, 0.00000708, -0.00000533, 0.00000437, -0.00000394, -0.00000359, -0.00000682, 0.00000861, -0.00000267, 0.00000089, 0.00000124, -0.00000070, -0.00000444, 0.00000138, -0.00000377, 0.00000241, 0.00000196, 0.00000008, -0.00000275, -0.00000287, -0.00000398, 0.00000795, -0.00000374, 0.00000493, 0.00000267, -0.00000554, -0.00000465, 0.00000496, 0.00000113, -0.00000124, -0.00000075, -0.00000007, -0.00000182, -0.00000065, 0.00000050, 0.00000316, -0.00000580, -0.00000808, 0.00001007, -0.00000271, 0.00000550, 0.00000169, 0.00000303, -0.00001075, 0.00000566, -0.00000739, 0.00000854, -0.00000883, 0.00000456, -0.00000456, 0.00000048, -0.00000030, -0.00000007, 0.00000928, -0.00001205, 0.00002196, -0.00002428, 0.00000810, -0.00001525, 0.00001694, -0.00001447, 0.00001440, -0.00000375, 0.00000120, -0.00000559, 0.00000640, -0.00000556, 0.00000352, 0.00000322, 0.00000047, -0.00000386, -0.00001557, 0.00000953, 0.00001065, -0.00001479, -0.00000224, -0.00000678, 0.00001631, -0.00001540, 0.00000097, -0.00001545, 0.00000750, -0.00000889, -0.00001074, 0.00001025, -0.00003191, -0.00000310, -0.00001438, 0.00000831, 0.00003109, -0.00000624, 0.00000599, 0.00000130, -0.00001373, 0.00000047, -0.00008886, -0.00004765, 0.00018671, 0.00004585, 0.00004519, 0.00000440, 0.00006292, 0.00010024, -0.00001342, -0.00009486, -0.00001529, -0.00002734, -0.00000698, -0.00001605, -0.00000073, -0.00001511, 0.00000338, -0.00000831, -0.00000230, },
		{ -0.00001301, -0.00000720, 0.00000806, -0.00000984, 0.00000867, -0.00000988, 0.00000383, 0.00000596, -0.00000127, -0.00000433, 0.00000295, -0.00000534, 0.00000351, 0.00000290, 0.00000583, -0.00000548, 0.00000488, -0.00000491, -0.00000351, -0.00000577, 0.00000851, -0.00000321, 0.00000156, 0.00000063, -0.00000091, -0.00000431, 0.00000153, -0.00000412, 0.00000307, 0.00000197, -0.00000053, -0.00000253, -0.00000351, -0.00000240, 0.00000727, -0.00000415, 0.00000598, 0.00000141, -0.00000600, -0.00000355, 0.00000530, 0.00000042, -0.00000122, -0.00000052, -0.00000031, -0.00000171, -0.00000059, 0.00000085, 0.00000285, -0.00000748, -0.00000581, 0.00000995, -0.00000328, 0.00000623, 0.00000133, 0.00000195, -0.00001025, 0.00000599, -0.00000755, 0.00000912, -0.00001017, 0.00000626, -0.00000679, 0.00000318, -0.00000314, 0.00000377, 0.00000497, -0.00000788, 0.00001947, -0.00002551, 0.00000985, -0.00001566, 0.00001814, -0.00001670, 0.00001777, -0.00000730, 0.00000347, -0.00000736, 0.00000794, -0.00000700, 0.00000529, 0.00000193, 0.00000074, -0.00000526, -0.00001478, 0.00001275, 0.00000628, -0.00001317, -0.00000372, -0.00000347, 0.00001451, -0.00001613, 0.00000290, -0.00001844, 0.00001183, -0.00001390, -0.00000573, 0.00000694, -0.00003313, 0.00000265, -0.00001651, 0.00001114, 0.00002768, -0.00000608, 0.00000472, 0.00000061, -0.00001456, 0.00000039, -0.00009012, -0.00005532, 0.00019326, 0.00004563, 0.00005274, 0.00000251, 0.00007793, 0.00009774, -0.00003729, -0.00008678, -0.00001309, -0.00002997, -0.00000388, -0.00001750, 0.00000031, -0.00001640, 0.00000407, -0.00000798, -0.00000187, -0.00000330, },
		{ -0.00000437, 0.00000612, -0.00000973, 0.00000882, -0.00001024, 0.00000610, 0.00000434, -0.00000148, -0.00000381, 0.00000264, -0.00000552, 0.00000372, 0.00000411, 0.00000441, -0.00000536, 0.00000517, -0.00000585, -0.00000322, -0.00000479, 0.00000821, -0.00000353, 0.00000215, -0.00000003, -0.00000106, -0.00000415, 0.00000161, -0.00000446, 0.00000386, 0.00000171, -0.00000091, -0.00000247, -0.00000394, -0.00000082, 0.00000626, -0.00000415, 0.00000677, 0.00000006, -0.00000619, -0.00000243, 0.00000546, -0.00000028, -0.00000110, -0.00000034, -0.00000057, -0.00000154, -0.00000057, 0.00000128, 0.00000220, -0.00000883, -0.00000336, 0.00000935, -0.00000352, 0.00000685, 0.00000083, 0.00000095, -0.00000953, 0.00000605, -0.00000743, 0.00000928, -0.00001105, 0.00000752, -0.00000861, 0.00000557, -0.00000572, 0.00000749, 0.00000039, -0.00000288, 0.00001557, -0.00002538, 0.00001081, -0.00001545, 0.00001866, -0.00001831, 0.00002080, -0.00001097, 0.00000616, -0.00000955, 0.00000984, -0.00000874, 0.00000747, 0.00000005, 0.00000156, -0.00000731, -0.00001281, 0.00001487, 0.00000205, -0.00001118, -0.00000544, 0.00000039, 0.00001152, -0.00001552, 0.00000370, -0.00002046, 0.00001551, -0.00001888, 0.00000022, 0.00000169, -0.00003202, 0.00000706, -0.00001846, 0.00001386, 0.00002385, -0.00000518, 0.00000296, 0.00000032, -0.00001570, 0.00000067, -0.00009167, -0.00006276, 0.00019946, 0.00004614, 0.00005988, 0.00000112, 0.00009518, 0.00008972, -0.00005982, -0.00007569, -0.00001332, -0.00003136, -0.00000126, -0.00001889, 0.00000143, -0.00001786, 0.00000508, -0.00000789, -0.00000125, -0.00000508, -0.00001326, },
		{ 0.00000381, -0.00000896, 0.00000843, -0.00001011, 0.00000819, 0.00000237, -0.00000128, -0.00000349, 0.00000240, -0.00000575, 0.00000413, 0.00000512, 0.00000290, -0.00000499, 0.00000523, -0.00000668, -0.00000275, -0.00000394, 0.00000777, -0.00000366, 0.00000261, -0.00000071, -0.00000117, -0.00000392, 0.00000159, -0.00000470, 0.00000471, 0.00000121, -0.00000106, -0.00000263, -0.00000409, 0.00000063, 0.00000503, -0.00000377, 0.00000725, -0.00000129, -0.00000613, -0.00000133, 0.00000544, -0.00000094, -0.00000088, -0.00000025, -0.00000080, -0.00000134, -0.00000058, 0.00000177, 0.00000124, -0.00000974, -0.00000088, 0.00000837, -0.00000341, 0.00000729, 0.00000023, 0.00000005, -0.00000867, 0.00000590, -0.00000707, 0.00000907, -0.00001146, 0.00000829, -0.00000994, 0.00000753, -0.00000788, 0.00001085, -0.00000414, 0.00000263, 0.00001046, -0.00002385, 0.00001082, -0.00001449, 0.00001833, -0.00001909, 0.00002319, -0.00001447, 0.00000905, -0.00001205, 0.00001203, -0.00001074, 0.00000998, -0.00000236, 0.00000290, -0.00000998, -0.00000958, 0.00001558, -0.00000161, -0.00000915, -0.00000710, 0.00000450, 0.00000754, -0.00001358, 0.00000319, -0.00002122, 0.00001807, -0.00002322, 0.00000650, -0.00000510, -0.00002854, 0.00000968, -0.00001971, 0.00001609, 0.00001991, -0.00000378, 0.00000081, 0.00000036, -0.00001714, 0.00000128, -0.00009356, -0.00006987, 0.00020507, 0.00004774, 0.00006617, 0.00000096, 0.00011397, 0.00007572, -0.00007960, -0.00006254, -0.00001586, -0.00003128, 0.00000055, -0.00001994, 0.00000239, -0.00001933, 0.00000629, -0.00000803, -0.00000045, -0.00000732, -0.00001274, -0.00000190, },
		{ -0.00000767, 0.00000750, -0.00000939, 0.00000991, 0.00000022, -0.00000072, -0.00000339, 0.00000224, -0.00000604, 0.00000477, 0.00000587, 0.00000138, -0.00000443, 0.00000504, -0.00000736, -0.00000213, -0.00000325, 0.00000724, -0.00000361, 0.00000296, -0.00000139, -0.00000123, -0.00000363, 0.00000144, -0.00000479, 0.00000552, 0.00000055, -0.00000101, -0.00000300, -0.00000391, 0.00000184, 0.00000371, -0.00000307, 0.00000739, -0.00000258, -0.00000584, -0.00000030, 0.00000525, -0.00000153, -0.00000059, -0.00000024, -0.00000100, -0.00000112, -0.00000059, 0.00000227, -0.00000003, -0.00001014, 0.00000146, 0.00000710, -0.00000297, 0.00000749, -0.00000042, -0.00000070, -0.00000773, 0.00000559, -0.00000654, 0.00000856, -0.00001146, 0.00000857, -0.00001075, 0.00000897, -0.00000947, 0.00001361, -0.00000829, 0.00000828, 0.00000441, -0.00002100, 0.00000983, -0.00001271, 0.00001705, -0.00001886, 0.00002470, -0.00001750, 0.00001189, -0.00001466, 0.00001439, -0.00001287, 0.00001270, -0.00000519, 0.00000467, -0.00001315, -0.00000514, 0.00001468, -0.00000432, -0.00000746, -0.00000836, 0.00000847, 0.00000293, -0.00001049, 0.00000136, -0.00002052, 0.00001912, -0.00002636, 0.00001240, -0.00001282, -0.00002290, 0.00001026, -0.00001987, 0.00001746, 0.00001625, -0.00000216, -0.00000154, 0.00000058, -0.00001875, 0.00000215, -0.00009573, -0.00007663, 0.00020996, 0.00005073, 0.00007118, 0.00000287, 0.00013325, 0.00005567, -0.00009532, -0.00004844, -0.00002032, -0.00002967, 0.00000133, -0.00002041, 0.00000298, -0.00002061, 0.00000756, -0.00000830, 0.00000044, -0.00000997, -0.00001138, -0.00000005, 0.00000138, },
		{ 0.00000609, -0.00000801, 0.00001112, -0.00000196, 0.00000010, -0.00000349, 0.00000216, -0.00000637, 0.00000561, 0.00000630, -0.00000006, -0.00000373, 0.00000462, -0.00000784, -0.00000144, -0.00000273, 0.00000668, -0.00000343, 0.00000318, -0.00000207, -0.00000123, -0.00000330, 0.00000116, -0.00000467, 0.00000621, -0.00000020, -0.00000081, -0.00000354, -0.00000340, 0.00000273, 0.00000243, -0.00000214, 0.00000719, -0.00000373, -0.00000537, 0.00000066, 0.00000489, -0.00000203, -0.00000024, -0.00000033, -0.00000113, -0.00000092, -0.00000058, 0.00000272, -0.00000154, -0.00000999, 0.00000354, 0.00000567, -0.00000225, 0.00000745, -0.00000109, -0.00000129, -0.00000677, 0.00000519, -0.00000591, 0.00000782, -0.00001109, 0.00000842, -0.00001103, 0.00000985, -0.00001041, 0.00001561, -0.00001177, 0.00001368, -0.00000222, -0.00001702, 0.00000787, -0.00001009, 0.00001479, -0.00001749, 0.00002509, -0.00001978, 0.00001441, -0.00001719, 0.00001675, -0.00001502, 0.00001549, -0.00000830, 0.00000671, -0.00001663, 0.00000035, 0.00001208, -0.00000579, -0.00000644, -0.00000888, 0.00001187, -0.00000189, -0.00000656, -0.00000161, -0.00001836, 0.00001844, -0.00002778, 0.00001723, -0.00002073, -0.00001557, 0.00000879, -0.00001864, 0.00001765, 0.00001322, -0.00000065, -0.00000383, 0.00000080, -0.00002036, 0.00000309, -0.00009804, -0.00008313, 0.00021410, 0.00005531, 0.00007453, 0.00000775, 0.00015168, 0.00002994, -0.00010594, -0.00003461, -0.00002611, -0.00002669, 0.00000099, -0.00002013, 0.00000300, -0.00002149, 0.00000868, -0.00000855, 0.00000131, -0.00001292, -0.00000916, 0.00000098, -0.00000086, -0.00000606, },
		{ -0.00000601, 0.00001168, -0.00000397, 0.00000110, -0.00000375, 0.00000216, -0.00000672, 0.00000664, 0.00000637, -0.00000134, -0.00000297, 0.00000400, -0.00000809, -0.00000073, -0.00000238, 0.00000613, -0.00000316, 0.00000330, -0.00000274, -0.00000117, -0.00000296, 0.00000075, -0.00000433, 0.00000670, -0.00000094, -0.00000052, -0.00000420, -0.00000258, 0.00000323, 0.00000129, -0.00000108, 0.00000666, -0.00000470, -0.00000475, 0.00000151, 0.00000440, -0.00000241, 0.00000013, -0.00000050, -0.00000118, -0.00000076, -0.00000052, 0.00000308, -0.00000321, -0.00000928, 0.00000524, 0.00000421, -0.00000130, 0.00000714, -0.00000173, -0.00000174, -0.00000582, 0.00000472, -0.00000524, 0.00000694, -0.00001047, 0.00000790, -0.00001084, 0.00001020, -0.00001068, 0.00001673, -0.00001434, 0.00001844, -0.00000901, -0.00001217, 0.00000506, -0.00000671, 0.00001158, -0.00001497, 0.00002422, -0.00002105, 0.00001634, -0.00001939, 0.00001894, -0.00001702, 0.00001818, -0.00001152, 0.00000888, -0.00002019, 0.00000666, 0.00000785, -0.00000580, -0.00000635, -0.00000835, 0.00001426, -0.00000639, -0.00000226, -0.00000541, -0.00001491, 0.00001598, -0.00002717, 0.00002032, -0.00002800, -0.00000721, 0.00000548, -0.00001593, 0.00001646, 0.00001112, 0.00000046, -0.00000581, 0.00000076, -0.00002173, 0.00000386, -0.00010025, -0.00008958, 0.00021759, 0.00006154, 0.00007603, 0.00001639, 0.00016768, -0.00000059, -0.00011073, -0.00002222, -0.00003246, -0.00002266, -0.00000038, -0.00001903, 0.00000232, -0.00002179, 0.00000948, -0.00000862, 0.00000198, -0.00001597, -0.00000617, 0.00000103, -0.00000264, -0.00000435, 0.00000434, },
		{ 0.00001155, -0.00000570, 0.00000214, -0.00000411, 0.00000219, -0.00000705, 0.00000782, 0.00000609, -0.00000239, -0.00000223, 0.00000324, -0.00000808, -0.00000005, -0.00000216, 0.00000563, -0.00000283, 0.00000333, -0.00000341, -0.00000102, -0.00000265, 0.00000027, -0.00000377, 0.00000695, -0.00000160, -0.00000023, -0.00000491, -0.00000151, 0.00000334, 0.00000041, 0.00000001, 0.00000587, -0.00000546, -0.00000402, 0.00000224, 0.00000379, -0.00000266, 0.00000050, -0.00000073, -0.00000114, -0.00000065, -0.00000037, 0.00000326, -0.00000493, -0.00000806, 0.00000647, 0.00000283, -0.00000022, 0.00000659, -0.00000231, -0.00000203, -0.00000492, 0.00000425, -0.00000458, 0.00000600, -0.00000966, 0.00000712, -0.00001026, 0.00001009, -0.00001033, 0.00001696, -0.00001584, 0.00002224, -0.00001553, -0.00000680, 0.00000161, -0.00000272, 0.00000754, -0.00001132, 0.00002201, -0.00002112, 0.00001742, -0.00002103, 0.00002074, -0.00001869, 0.00002059, -0.00001467, 0.00001097, -0.00002355, 0.00001342, 0.00000216, -0.00000431, -0.00000737, -0.00000654, 0.00001526, -0.00001007, 0.00000192, -0.00000959, -0.00001051, 0.00001193, -0.00002438, 0.00002117, -0.00003384, 0.00000139, 0.00000077, -0.00001186, 0.00001381, 0.00001015, 0.00000089, -0.00000722, 0.00000023, -0.00002260, 0.00000417, -0.00010203, -0.00009630, 0.00022071, 0.00006927, 0.00007569, 0.00002945, 0.00017954, -0.00003459, -0.00010945, -0.00001227, -0.00003855, -0.00001807, -0.00000258, -0.00001717, 0.00000091, -0.00002139, 0.00000978, -0.00000834, 0.00000225, -0.00001887, -0.00000260, 0.00000008, -0.00000373, -0.00000282, 0.00000243, -0.00000347, },
		{ -0.00000701, 0.00000312, -0.00000452, 0.00000223, -0.00000731, 0.00000908, 0.00000546, -0.00000316, -0.00000157, 0.00000240, -0.00000784, 0.00000053, -0.00000204, 0.00000518, -0.00000246, 0.00000327, -0.00000408, -0.00000076, -0.00000242, -0.00000023, -0.00000301, 0.00000691, -0.00000210, -0.00000001, -0.00000557, -0.00000027, 0.00000307, -0.00000017, 0.00000102, 0.00000488, -0.00000599, -0.00000320, 0.00000283, 0.00000309, -0.00000276, 0.00000082, -0.00000101, -0.00000103, -0.00000061, -0.00000014, 0.00000322, -0.00000659, -0.00000640, 0.00000718, 0.00000166, 0.00000089, 0.00000583, -0.00000281, -0.00000220, -0.00000408, 0.00000378, -0.00000397, 0.00000506, -0.00000878, 0.00000617, -0.00000940, 0.00000961, -0.00000945, 0.00001635, -0.00001618, 0.00002481, -0.00002140, -0.00000128, -0.00000223, 0.00000167, 0.00000288, -0.00000670, 0.00001848, -0.00001987, 0.00001745, -0.00002187, 0.00002195, -0.00001986, 0.00002254, -0.00001755, 0.00001280, -0.00002642, 0.00002024, -0.00000467, -0.00000138, -0.00000955, -0.00000331, 0.00001455, -0.00001244, 0.00000540, -0.00001359, -0.00000569, 0.00000665, -0.00001954, 0.00001950, -0.00003754, 0.00000937, -0.00000472, -0.00000670, 0.00000982, 0.00001038, 0.00000050, -0.00000789, -0.00000100, -0.00002273, 0.00000371, -0.00010304, -0.00010372, 0.00022385, 0.00007820, 0.00007378, 0.00004729, 0.00018557, -0.00007036, -0.00010229, -0.00000555, -0.00004358, -0.00001351, -0.00000525, -0.00001474, -0.00000117, -0.00002025, 0.00000946, -0.00000758, 0.00000194, -0.00002134, 0.00000130, -0.00000180, -0.00000397, -0.00000169, 0.00000059, -0.00000055, 0.00001073, },
		{ 0.00000396, -0.00000492, 0.00000224, -0.00000744, 0.00001035, 0.00000451, -0.00000361, -0.00000107, 0.00000154, -0.00000738, 0.00000098, -0.00000199, 0.00000477, -0.00000206, 0.00000314, -0.00000474, -0.00000038, -0.00000232, -0.00000068, -0.00000212, 0.00000660, -0.00000241, 0.00000009, -0.00000610, 0.00000103, 0.00000247, -0.00000040, 0.00000186, 0.00000376, -0.00000631, -0.00000233, 0.00000326, 0.00000233, -0.00000271, 0.00000107, -0.00000131, -0.00000085, -0.00000063, 0.00000020, 0.00000290, -0.00000807, -0.00000442, 0.00000739, 0.00000077, 0.00000193, 0.00000491, -0.00000321, -0.00000224, -0.00000334, 0.00000334, -0.00000344, 0.00000419, -0.00000790, 0.00000516, -0.00000837, 0.00000889, -0.00000819, 0.00001502, -0.00001541, 0.00002598, -0.00002625, 0.00000402, -0.00000616, 0.00000620, -0.00000215, -0.00000133, 0.00001375, -0.00001727, 0.00001629, -0.00002175, 0.00002240, -0.00002034, 0.00002385, -0.00001999, 0.00001418, -0.00002852, 0.00002666, -0.00001222, 0.00000276, -0.00001280, 0.00000132, 0.00001198, -0.00001310, 0.00000767, -0.00001684, -0.00000105, 0.00000073, -0.00001302, 0.00001528, -0.00003860, 0.00001593, -0.00001028, -0.00000093, 0.00000474, 0.00001171, -0.00000074, -0.00000773, -0.00000306, -0.00002193, 0.00000222, -0.00010290, -0.00011230, 0.00022755, 0.00008778, 0.00007093, 0.00006986, 0.00018425, -0.00010588, -0.00008998, -0.00000246, -0.00004687, -0.00000956, -0.00000797, -0.00001201, -0.00000375, -0.00001843, 0.00000851, -0.00000624, 0.00000090, -0.00002311, 0.00000518, -0.00000444, -0.00000330, -0.00000118, -0.00000094, 0.00000256, 0.00000928, -0.00000786, },
		{ -0.00000524, 0.00000218, -0.00000739, 0.00001153, 0.00000331, -0.00000374, -0.00000079, 0.00000076, -0.00000676, 0.00000128, -0.00000195, 0.00000439, -0.00000164, 0.00000293, -0.00000537, 0.00000011, -0.00000237, -0.00000101, -0.00000116, 0.00000605, -0.00000250, 0.00000002, -0.00000643, 0.00000228, 0.00000164, -0.00000028, 0.00000246, 0.00000259, -0.00000641, -0.00000143, 0.00000353, 0.00000156, -0.00000252, 0.00000123, -0.00000159, -0.00000062, -0.00000069, 0.00000062, 0.00000230, -0.00000925, -0.00000223, 0.00000711, 0.00000022, 0.00000283, 0.00000389, -0.00000350, -0.00000218, -0.00000268, 0.00000294, -0.00000301, 0.00000342, -0.00000708, 0.00000418, -0.00000730, 0.00000806, -0.00000671, 0.00001315, -0.00001364, 0.00002569, -0.00002981, 0.00000875, -0.00000988, 0.00001058, -0.00000724, 0.00000449, 0.00000805, -0.00001340, 0.00001391, -0.00002053, 0.00002192, -0.00002001, 0.00002435, -0.00002180, 0.00001495, -0.00002960, 0.00003224, -0.00002000, 0.00000777, -0.00001691, 0.00000720, 0.00000752, -0.00001181, 0.00000829, -0.00001875, 0.00000274, -0.00000514, -0.00000542, 0.00000880, -0.00003680, 0.00002040, -0.00001518, 0.00000489, -0.00000100, 0.00001393, -0.00000275, -0.00000677, -0.00000597, -0.00002013, -0.00000049, -0.00010131, -0.00012250, 0.00023239, 0.00009736, 0.00006804, 0.00009665, 0.00017435, -0.00013900, -0.00007365, -0.00000305, -0.00004794, -0.00000676, -0.00001029, -0.00000935, -0.00000655, -0.00001609, 0.00000697, -0.00000431, -0.00000097, -0.00002393, 0.00000869, -0.00000757, -0.00000177, -0.00000140, -0.00000196, 0.00000563, 0.00000732, -0.00000821, 0.00000457, },
		{ 0.00000203, -0.00000712, 0.00001255, 0.00000194, -0.00000356, -0.00000075, 0.00000012, -0.00000604, 0.00000143, -0.00000189, 0.00000403, -0.00000119, 0.00000263, -0.00000595, 0.00000071, -0.00000261, -0.00000116, -0.00000023, 0.00000532, -0.00000240, -0.00000024, -0.00000650, 0.00000338, 0.00000068, 0.00000013, 0.00000278, 0.00000144, -0.00000633, -0.00000052, 0.00000361, 0.00000081, -0.00000221, 0.00000127, -0.00000182, -0.00000038, -0.00000079, 0.00000110, 0.00000140, -0.00001004, 0.00000000, 0.00000643, 0.00000003, 0.00000350, 0.00000284, -0.00000369, -0.00000204, -0.00000212, 0.00000258, -0.00000268, 0.00000278, -0.00000637, 0.00000330, -0.00000626, 0.00000726, -0.00000520, 0.00001098, -0.00001107, 0.00002402, -0.00003192, 0.00001262, -0.00001309, 0.00001451, -0.00001206, 0.00001042, 0.00000168, -0.00000845, 0.00001034, -0.00001818, 0.00002043, -0.00001874, 0.00002393, -0.00002287, 0.00001501, -0.00002946, 0.00003656, -0.00002748, 0.00001324, -0.00002154, 0.00001401, 0.00000135, -0.00000849, 0.00000696, -0.00001887, 0.00000505, -0.00001016, 0.00000248, 0.00000060, -0.00003219, 0.00002231, -0.00001875, 0.00001016, -0.00000692, 0.00001668, -0.00000529, -0.00000517, -0.00000962, -0.00001735, -0.00000448, -0.00009806, -0.00013472, 0.00023894, 0.00010619, 0.00006630, 0.00012663, 0.00015516, -0.00016761, -0.00005481, -0.00000696, -0.00004658, -0.00000551, -0.00001179, -0.00000713, -0.00000925, -0.00001349, 0.00000499, -0.00000189, -0.00000366, -0.00002362, 0.00001148, -0.00001085, 0.00000047, -0.00000240, -0.00000228, 0.00000842, 0.00000503, -0.00000811, 0.00000493, -0.00000545, },
		{ -0.00000660, 0.00001332, 0.00000049, -0.00000312, -0.00000097, -0.00000031, -0.00000529, 0.00000144, -0.00000180, 0.00000366, -0.00000071, 0.00000221, -0.00000643, 0.00000136, -0.00000303, -0.00000107, 0.00000059, 0.00000448, -0.00000213, -0.00000067, -0.00000630, 0.00000424, -0.00000031, 0.00000078, 0.00000281, 0.00000038, -0.00000609, 0.00000037, 0.00000352, 0.00000011, -0.00000181, 0.00000121, -0.00000197, -0.00000016, -0.00000088, 0.00000158, 0.00000023, -0.00001037, 0.00000215, 0.00000544, 0.00000019, 0.00000390, 0.00000181, -0.00000381, -0.00000181, -0.00000165, 0.00000226, -0.00000243, 0.00000225, -0.00000579, 0.00000257, -0.00000535, 0.00000660, -0.00000384, 0.00000875, -0.00000798, 0.00002111, -0.00003252, 0.00001542, -0.00001556, 0.00001775, -0.00001630, 0.00001607, -0.00000501, -0.00000267, 0.00000573, -0.00001472, 0.00001790, -0.00001648, 0.00002249, -0.00002308, 0.00001428, -0.00002797, 0.00003926, -0.00003414, 0.00001867, -0.00002625, 0.00002133, -0.00000622, -0.00000323, 0.00000357, -0.00001687, 0.00000531, -0.00001356, 0.00000975, -0.00000851, -0.00002516, 0.00002151, -0.00002049, 0.00001433, -0.00001248, 0.00001953, -0.00000803, -0.00000321, -0.00001378, -0.00001379, -0.00000965, -0.00009312, -0.00014920, 0.00024769, 0.00011352, 0.00006709, 0.00015822, 0.00012656, -0.00018979, -0.00003517, -0.00001347, -0.00004289, -0.00000602, -0.00001214, -0.00000569, -0.00001151, -0.00001091, 0.00000280, 0.00000086, -0.00000707, -0.00002210, 0.00001324, -0.00001390, 0.00000318, -0.00000412, -0.00000181, 0.00001072, 0.00000259, -0.00000762, 0.00000502, -0.00000554, 0.00000177, },
		{ 0.00001375, -0.00000094, -0.00000247, -0.00000143, -0.00000048, -0.00000460, 0.00000135, -0.00000167, 0.00000327, -0.00000018, 0.00000167, -0.00000678, 0.00000202, -0.00000360, -0.00000072, 0.00000122, 0.00000362, -0.00000175, -0.00000123, -0.00000582, 0.00000481, -0.00000121, 0.00000159, 0.00000256, -0.00000056, -0.00000571, 0.00000123, 0.00000325, -0.00000048, -0.00000135, 0.00000103, -0.00000204, 0.00000003, -0.00000094, 0.00000203, -0.00000115, -0.00001020, 0.00000407, 0.00000426, 0.00000064, 0.00000401, 0.00000087, -0.00000386, -0.00000152, -0.00000128, 0.00000198, -0.00000225, 0.00000183, -0.00000533, 0.00000201, -0.00000461, 0.00000615, -0.00000276, 0.00000671, -0.00000467, 0.00001723, -0.00003167, 0.00001704, -0.00001714, 0.00002009, -0.00001968, 0.00002110, -0.00001160, 0.00000358, 0.00000030, -0.00001027, 0.00001436, -0.00001323, 0.00002000, -0.00002238, 0.00001276, -0.00002512, 0.00004010, -0.00003950, 0.00002358, -0.00003058, 0.00002863, -0.00001469, 0.00000364, -0.00000179, -0.00001261, 0.00000316, -0.00001468, 0.00001548, -0.00001754, -0.00001641, 0.00001813, -0.00002014, 0.00001694, -0.00001719, 0.00002203, -0.00001058, -0.00000126, -0.00001810, -0.00000977, -0.00001575, -0.00008663, -0.00016601, 0.00025895, 0.00011875, 0.00007186, 0.00018940, 0.00008909, -0.00020403, -0.00001653, -0.00002158, -0.00003722, -0.00000829, -0.00001117, -0.00000531, -0.00001299, -0.00000869, 0.00000068, 0.00000367, -0.00001100, -0.00001940, 0.00001374, -0.00001636, 0.00000605, -0.00000641, -0.00000053, 0.00001237, 0.00000020, -0.00000685, 0.00000488, -0.00000549, 0.00000141, -0.00000581, },
		{ -0.00000226, -0.00000171, -0.00000210, -0.00000039, -0.00000404, 0.00000123, -0.00000152, 0.00000288, 0.00000039, 0.00000099, -0.00000695, 0.00000261, -0.00000426, -0.00000012, 0.00000160, 0.00000281, -0.00000134, -0.00000186, -0.00000510, 0.00000504, -0.00000192, 0.00000244, 0.00000206, -0.00000135, -0.00000523, 0.00000204, 0.00000282, -0.00000094, -0.00000086, 0.00000076, -0.00000200, 0.00000014, -0.00000093, 0.00000238, -0.00000266, -0.00000953, 0.00000566, 0.00000301, 0.00000132, 0.00000382, 0.00000005, -0.00000386, -0.00000117, -0.00000100, 0.00000173, -0.00000212, 0.00000150, -0.00000497, 0.00000160, -0.00000404, 0.00000595, -0.00000206, 0.00000505, -0.00000148, 0.00001271, -0.00002955, 0.00001748, -0.00001773, 0.00002141, -0.00002198, 0.00002516, -0.00001771, 0.00000993, -0.00000565, -0.00000504, 0.00000994, -0.00000905, 0.00001649, -0.00002077, 0.00001049, -0.00002095, 0.00003892, -0.00004318, 0.00002750, -0.00003403, 0.00003535, -0.00002346, 0.00001165, -0.00000883, -0.00000621, -0.00000158, -0.00001299, 0.00001880, -0.00002544, -0.00000687, 0.00001263, -0.00001768, 0.00001776, -0.00002068, 0.00002377, -0.00001255, 0.00000030, -0.00002216, -0.00000572, -0.00002236, -0.00007895, -0.00018496, 0.00027276, 0.00012153, 0.00008199, 0.00021776, 0.00004404, -0.00020936, -0.00000058, -0.00003011, -0.00003022, -0.00001208, -0.00000887, -0.00000613, -0.00001346, -0.00000712, -0.00000105, 0.00000625, -0.00001513, -0.00001571, 0.00001289, -0.00001789, 0.00000873, -0.00000905, 0.00000149, 0.00001326, -0.00000198, -0.00000592, 0.00000454, -0.00000534, 0.00000099, -0.00000477, 0.00001381, },
		{ -0.00000090, -0.00000291, -0.00000002, -0.00000367, 0.00000112, -0.00000137, 0.00000248, 0.00000098, 0.00000018, -0.00000690, 0.00000308, -0.00000495, 0.00000071, 0.00000169, 0.00000214, -0.00000096, -0.00000250, -0.00000422, 0.00000494, -0.00000239, 0.00000326, 0.00000139, -0.00000196, -0.00000467, 0.00000277, 0.00000224, -0.00000124, -0.00000040, 0.00000042, -0.00000186, 0.00000018, -0.00000083, 0.00000256, -0.00000420, -0.00000839, 0.00000683, 0.00000182, 0.00000214, 0.00000336, -0.00000061, -0.00000384, -0.00000076, -0.00000083, 0.00000152, -0.00000204, 0.00000124, -0.00000469, 0.00000132, -0.00000364, 0.00000599, -0.00000180, 0.00000395, 0.00000129, 0.00000790, -0.00002640, 0.00001681, -0.00001734, 0.00002165, -0.00002307, 0.00002800, -0.00002295, 0.00001598, -0.00001175, 0.00000070, 0.00000482, -0.00000409, 0.00001202, -0.00001829, 0.00000757, -0.00001561, 0.00003572, -0.00004491, 0.00003004, -0.00003617, 0.00004089, -0.00003186, 0.00002018, -0.00001705, 0.00000201, -0.00000879, -0.00000825, 0.00001902, -0.00003116, 0.00000238, 0.00000576, -0.00001340, 0.00001675, -0.00002272, 0.00002447, -0.00001361, 0.00000107, -0.00002552, -0.00000214, -0.00002892, -0.00007063, -0.00020561, 0.00028887, 0.00012183, 0.00009860, 0.00024068, -0.00000665, -0.00020544, 0.00001123, -0.00003780, -0.00002269, -0.00001694, -0.00000545, -0.00000813, -0.00001278, -0.00000645, -0.00000213, 0.00000827, -0.00001910, -0.00001129, 0.00001071, -0.00001828, 0.00001090, -0.00001174, 0.00000411, 0.00001334, -0.00000377, -0.00000495, 0.00000407, -0.00000510, 0.00000053, -0.00000351, 0.00001345, -0.00000336, },
		{ -0.00000378, 0.00000059, -0.00000352, 0.00000110, -0.00000126, 0.00000212, 0.00000156, -0.00000074, -0.00000661, 0.00000335, -0.00000558, 0.00000170, 0.00000150, 0.00000165, -0.00000068, -0.00000305, -0.00000324, 0.00000456, -0.00000256, 0.00000396, 0.00000059, -0.00000241, -0.00000403, 0.00000340, 0.00000153, -0.00000137, 0.00000000, 0.00000005, -0.00000164, 0.00000013, -0.00000063, 0.00000254, -0.00000568, -0.00000687, 0.00000754, 0.00000079, 0.00000299, 0.00000268, -0.00000111, -0.00000381, -0.00000030, -0.00000075, 0.00000134, -0.00000198, 0.00000101, -0.00000445, 0.00000114, -0.00000335, 0.00000623, -0.00000196, 0.00000350, 0.00000340, 0.00000318, -0.00002254, 0.00001523, -0.00001607, 0.00002086, -0.00002291, 0.00002944, -0.00002701, 0.00002133, -0.00001763, 0.00000663, -0.00000075, 0.00000148, 0.00000675, -0.00001502, 0.00000414, -0.00000934, 0.00003062, -0.00004456, 0.00003093, -0.00003664, 0.00004473, -0.00003922, 0.00002852, -0.00002579, 0.00001151, -0.00001813, -0.00000050, 0.00001571, -0.00003378, 0.00001019, -0.00000155, -0.00000787, 0.00001417, -0.00002331, 0.00002398, -0.00001355, 0.00000077, -0.00002778, 0.00000045, -0.00003483, -0.00006236, -0.00022728, 0.00030668, 0.00012013, 0.00012238, 0.00025555, -0.00006044, -0.00019266, 0.00001788, -0.00004349, -0.00001555, -0.00002227, -0.00000124, -0.00001114, -0.00001094, -0.00000681, -0.00000234, 0.00000942, -0.00002251, -0.00000654, 0.00000737, -0.00001738, 0.00001227, -0.00001420, 0.00000712, 0.00001265, -0.00000510, -0.00000406, 0.00000355, -0.00000484, 0.00000010, -0.00000209, 0.00001267, -0.00000419, -0.00000015, },
		{ 0.00000136, -0.00000360, 0.00000121, -0.00000123, 0.00000183, 0.00000210, -0.00000174, -0.00000610, 0.00000338, -0.00000607, 0.00000277, 0.00000106, 0.00000136, -0.00000054, -0.00000347, -0.00000226, 0.00000396, -0.00000244, 0.00000446, -0.00000026, -0.00000271, -0.00000334, 0.00000390, 0.00000075, -0.00000133, 0.00000030, -0.00000032, -0.00000136, 0.00000000, -0.00000032, 0.00000226, -0.00000698, -0.00000505, 0.00000779, -0.00000000, 0.00000378, 0.00000186, -0.00000146, -0.00000377, 0.00000019, -0.00000077, 0.00000122, -0.00000196, 0.00000082, -0.00000422, 0.00000099, -0.00000313, 0.00000661, -0.00000248, 0.00000372, 0.00000466, -0.00000110, -0.00001832, 0.00001297, -0.00001410, 0.00001918, -0.00002157, 0.00002941, -0.00002966, 0.00002565, -0.00002291, 0.00001239, -0.00000646, 0.00000740, 0.00000085, -0.00001109, 0.00000038, -0.00000243, 0.00002386, -0.00004213, 0.00003004, -0.00003520, 0.00004645, -0.00004489, 0.00003594, -0.00003432, 0.00002156, -0.00002897, 0.00000992, 0.00000875, -0.00003265, 0.00001550, -0.00000823, -0.00000185, 0.00001047, -0.00002268, 0.00002238, -0.00001233, -0.00000076, -0.00002867, 0.00000161, -0.00003947, -0.00005493, -0.00024909, 0.00032528, 0.00011739, 0.00015343, 0.00025998, -0.00011437, -0.00017209, 0.00001883, -0.00004625, -0.00000968, -0.00002737, 0.00000323, -0.00001484, -0.00000811, -0.00000820, -0.00000154, 0.00000947, -0.00002496, -0.00000193, 0.00000317, -0.00001523, 0.00001262, -0.00001612, 0.00001025, 0.00001130, -0.00000589, -0.00000334, 0.00000304, -0.00000460, -0.00000025, -0.00000059, 0.00001150, -0.00000467, 0.00000046, -0.00000463, },
		{ -0.00000390, 0.00000147, -0.00000134, 0.00000166, 0.00000254, -0.00000276, -0.00000537, 0.00000313, -0.00000632, 0.00000382, 0.00000041, 0.00000128, -0.00000057, -0.00000368, -0.00000138, 0.00000322, -0.00000205, 0.00000473, -0.00000109, -0.00000289, -0.00000259, 0.00000426, -0.00000008, -0.00000112, 0.00000049, -0.00000064, -0.00000105, -0.00000018, 0.00000008, 0.00000172, -0.00000800, -0.00000306, 0.00000761, -0.00000050, 0.00000443, 0.00000095, -0.00000167, -0.00000371, 0.00000070, -0.00000088, 0.00000114, -0.00000196, 0.00000064, -0.00000397, 0.00000085, -0.00000290, 0.00000701, -0.00000326, 0.00000454, 0.00000496, -0.00000466, -0.00001407, 0.00001030, -0.00001163, 0.00001682, -0.00001920, 0.00002796, -0.00003077, 0.00002867, -0.00002727, 0.00001764, -0.00001200, 0.00001339, -0.00000545, -0.00000667, -0.00000353, 0.00000478, 0.00001578, -0.00003779, 0.00002735, -0.00003176, 0.00004577, -0.00004836, 0.00004175, -0.00004186, 0.00003133, -0.00004050, 0.00002234, -0.00000159, -0.00002742, 0.00001742, -0.00001323, 0.00000375, 0.00000634, -0.00002127, 0.00001993, -0.00001011, -0.00000348, -0.00002806, 0.00000105, -0.00004233, -0.00004910, -0.00027009, 0.00034347, 0.00011507, 0.00019113, 0.00025203, -0.00016527, -0.00014545, 0.00001412, -0.00004544, -0.00000586, -0.00003152, 0.00000739, -0.00001876, -0.00000458, -0.00001048, 0.00000023, 0.00000829, -0.00002613, 0.00000209, -0.00000148, -0.00001196, 0.00001185, -0.00001725, 0.00001323, 0.00000945, -0.00000616, -0.00000285, 0.00000259, -0.00000441, -0.00000044, 0.00000091, 0.00000999, -0.00000480, 0.00000087, -0.00000536, 0.00000223, },
		{ 0.00000190, -0.00000159, 0.00000163, 0.00000283, -0.00000373, -0.00000449, 0.00000260, -0.00000629, 0.00000477, -0.00000036, 0.00000137, -0.00000077, -0.00000366, -0.00000068, 0.00000245, -0.00000144, 0.00000474, -0.00000186, -0.00000295, -0.00000180, 0.00000444, -0.00000089, -0.00000077, 0.00000053, -0.00000089, -0.00000074, -0.00000039, 0.00000053, 0.00000091, -0.00000867, -0.00000103, 0.00000707, -0.00000070, 0.00000487, 0.00000005, -0.00000177, -0.00000363, 0.00000121, -0.00000108, 0.00000113, -0.00000201, 0.00000048, -0.00000370, 0.00000067, -0.00000262, 0.00000736, -0.00000415, 0.00000586, 0.00000429, -0.00000729, -0.00001012, 0.00000751, -0.00000894, 0.00001403, -0.00001604, 0.00002523, -0.00003033, 0.00003022, -0.00003041, 0.00002206, -0.00001705, 0.00001915, -0.00001188, -0.00000193, -0.00000738, 0.00001194, 0.00000680, -0.00003181, 0.00002301, -0.00002639, 0.00004258, -0.00004926, 0.00004537, -0.00004765, 0.00003994, -0.00005174, 0.00003586, -0.00001471, -0.00001814, 0.00001538, -0.00001561, 0.00000799, 0.00000257, -0.00001971, 0.00001713, -0.00000727, -0.00000712, -0.00002608, -0.00000131, -0.00004306, -0.00004552, -0.00028931, 0.00035994, 0.00011515, 0.00023404, 0.00023045, -0.00021002, -0.00011495, 0.00000436, -0.00004083, -0.00000463, -0.00003410, 0.00001064, -0.00002239, -0.00000077, -0.00001338, 0.00000285, 0.00000588, -0.00002580, 0.00000506, -0.00000611, -0.00000785, 0.00000997, -0.00001741, 0.00001579, 0.00000729, -0.00000596, -0.00000263, 0.00000225, -0.00000432, -0.00000044, 0.00000231, 0.00000821, -0.00000457, 0.00000103, -0.00000588, 0.00000308, -0.00000436, },
		{ -0.00000201, 0.00000179, 0.00000291, -0.00000458, -0.00000353, 0.00000183, -0.00000593, 0.00000551, -0.00000117, 0.00000159, -0.00000111, -0.00000341, -0.00000021, 0.00000175, -0.00000070, 0.00000452, -0.00000253, -0.00000293, -0.00000098, 0.00000444, -0.00000165, -0.00000032, 0.00000044, -0.00000104, -0.00000048, -0.00000059, 0.00000100, -0.00000013, -0.00000891, 0.00000091, 0.00000625, -0.00000060, 0.00000504, -0.00000080, -0.00000181, -0.00000352, 0.00000170, -0.00000133, 0.00000117, -0.00000209, 0.00000035, -0.00000340, 0.00000045, -0.00000224, 0.00000755, -0.00000499, 0.00000749, 0.00000274, -0.00000886, -0.00000673, 0.00000488, -0.00000629, 0.00001110, -0.00001241, 0.00002147, -0.00002846, 0.00003026, -0.00003217, 0.00002537, -0.00002131, 0.00002436, -0.00001817, 0.00000294, -0.00001098, 0.00001869, -0.00000260, -0.00002458, 0.00001730, -0.00001931, 0.00003695, -0.00004740, 0.00004637, -0.00005109, 0.00004658, -0.00006168, 0.00004937, -0.00002967, -0.00000525, 0.00000914, -0.00001466, 0.00001001, 0.00000001, -0.00001873, 0.00001461, -0.00000439, -0.00001121, -0.00002308, -0.00000531, -0.00004160, -0.00004463, -0.00030593, 0.00037332, 0.00011993, 0.00027993, 0.00019486, -0.00024577, -0.00008312, -0.00000934, -0.00003263, -0.00000627, -0.00003465, 0.00001245, -0.00002518, 0.00000283, -0.00001651, 0.00000601, 0.00000239, -0.00002388, 0.00000662, -0.00001023, -0.00000328, 0.00000713, -0.00001652, 0.00001770, 0.00000504, -0.00000539, -0.00000266, 0.00000204, -0.00000433, -0.00000019, 0.00000350, 0.00000626, -0.00000402, 0.00000092, -0.00000612, 0.00000382, -0.00000492, 0.00000247, },
		{ 0.00000214, 0.00000274, -0.00000522, -0.00000257, 0.00000086, -0.00000524, 0.00000597, -0.00000193, 0.00000185, -0.00000154, -0.00000295, -0.00000002, 0.00000119, 0.00000010, 0.00000408, -0.00000306, -0.00000285, -0.00000015, 0.00000425, -0.00000230, 0.00000020, 0.00000022, -0.00000107, -0.00000028, -0.00000074, 0.00000144, -0.00000134, -0.00000871, 0.00000265, 0.00000526, -0.00000025, 0.00000495, -0.00000152, -0.00000181, -0.00000336, 0.00000213, -0.00000163, 0.00000126, -0.00000223, 0.00000027, -0.00000309, 0.00000016, -0.00000174, 0.00000751, -0.00000564, 0.00000921, 0.00000047, -0.00000936, -0.00000407, 0.00000262, -0.00000391, 0.00000832, -0.00000862, 0.00001701, -0.00002535, 0.00002883, -0.00003244, 0.00002739, -0.00002453, 0.00002876, -0.00002403, 0.00000773, -0.00001417, 0.00002469, -0.00001195, -0.00001654, 0.00001060, -0.00001089, 0.00002914, -0.00004281, 0.00004453, -0.00005170, 0.00005053, -0.00006935, 0.00006166, -0.00004525, 0.00001038, -0.00000109, -0.00001000, 0.00000913, -0.00000057, -0.00001908, 0.00001309, -0.00000215, -0.00001506, -0.00001969, -0.00001044, -0.00003821, -0.00004652, -0.00031941, 0.00038243, 0.00013191, 0.00032576, 0.00014588, -0.00027023, -0.00005257, -0.00002547, -0.00002146, -0.00001069, -0.00003292, 0.00001244, -0.00002664, 0.00000573, -0.00001942, 0.00000934, -0.00000187, -0.00002045, 0.00000654, -0.00001339, 0.00000132, 0.00000356, -0.00001459, 0.00001879, 0.00000291, -0.00000455, -0.00000291, 0.00000193, -0.00000444, 0.00000032, 0.00000440, 0.00000425, -0.00000320, 0.00000055, -0.00000604, 0.00000434, -0.00000548, 0.00000314, -0.00000258, },
		{ 0.00000230, -0.00000558, -0.00000171, -0.00000021, -0.00000424, 0.00000612, -0.00000256, 0.00000211, -0.00000201, -0.00000233, -0.00000009, 0.00000085, 0.00000087, 0.00000349, -0.00000347, -0.00000270, 0.00000069, 0.00000386, -0.00000280, 0.00000073, -0.00000009, -0.00000098, -0.00000018, -0.00000082, 0.00000178, -0.00000266, -0.00000807, 0.00000409, 0.00000421, 0.00000027, 0.00000459, -0.00000209, -0.00000180, -0.00000313, 0.00000249, -0.00000194, 0.00000140, -0.00000243, 0.00000026, -0.00000279, -0.00000015, -0.00000112, 0.00000721, -0.00000596, 0.00001081, -0.00000229, -0.00000886, -0.00000224, 0.00000090, -0.00000201, 0.00000595, -0.00000503, 0.00001221, -0.00002133, 0.00002611, -0.00003126, 0.00002801, -0.00002652, 0.00003209, -0.00002922, 0.00001227, -0.00001681, 0.00002969, -0.00002078, -0.00000819, 0.00000335, -0.00000159, 0.00001958, -0.00003573, 0.00003986, -0.00004927, 0.00005130, -0.00007389, 0.00007156, -0.00006012, 0.00002758, -0.00001467, -0.00000165, 0.00000494, 0.00000144, -0.00002145, 0.00001330, -0.00000136, -0.00001785, -0.00001670, -0.00001597, -0.00003348, -0.00005089, -0.00032964, 0.00038645, 0.00015357, 0.00036791, 0.00008518, -0.00028188, -0.00002581, -0.00004226, -0.00000829, -0.00001750, -0.00002897, 0.00001039, -0.00002641, 0.00000746, -0.00002162, 0.00001235, -0.00000647, -0.00001577, 0.00000474, -0.00001520, 0.00000551, -0.00000041, -0.00001174, 0.00001897, 0.00000109, -0.00000359, -0.00000333, 0.00000192, -0.00000462, 0.00000110, 0.00000495, 0.00000229, -0.00000220, -0.00000003, -0.00000563, 0.00000456, -0.00000594, 0.00000384, -0.00000325, 0.00000266, },
		{ -0.00000562, -0.00000106, -0.00000127, -0.00000300, 0.00000593, -0.00000300, 0.00000228, -0.00000243, -0.00000162, -0.00000040, 0.00000076, 0.00000152, 0.00000280, -0.00000375, -0.00000250, 0.00000149, 0.00000330, -0.00000312, 0.00000122, -0.00000047, -0.00000079, -0.00000018, -0.00000080, 0.00000197, -0.00000400, -0.00000703, 0.00000517, 0.00000322, 0.00000088, 0.00000401, -0.00000248, -0.00000182, -0.00000282, 0.00000275, -0.00000223, 0.00000157, -0.00000269, 0.00000033, -0.00000253, -0.00000047, -0.00000041, 0.00000663, -0.00000587, 0.00001207, -0.00000530, -0.00000750, -0.00000124, -0.00000019, -0.00000071, 0.00000419, -0.00000194, 0.00000746, -0.00001674, 0.00002236, -0.00002876, 0.00002724, -0.00002718, 0.00003418, -0.00003350, 0.00001638, -0.00001881, 0.00003347, -0.00002869, 0.00000001, -0.00000397, 0.00000804, 0.00000883, -0.00002659, 0.00003258, -0.00004381, 0.00004863, -0.00007470, 0.00007799, -0.00007286, 0.00004494, -0.00003061, 0.00000999, -0.00000260, 0.00000639, -0.00002635, 0.00001588, -0.00000276, -0.00001873, -0.00001509, -0.00002089, -0.00002836, -0.00005701, -0.00033702, 0.00038507, 0.00018708, 0.00040232, 0.00001547, -0.00028011, -0.00000494, -0.00005787, 0.00000564, -0.00002598, -0.00002309, 0.00000635, -0.00002430, 0.00000768, -0.00002270, 0.00001456, -0.00001090, -0.00001024, 0.00000132, -0.00001544, 0.00000886, -0.00000439, -0.00000819, 0.00001822, -0.00000028, -0.00000264, -0.00000383, 0.00000196, -0.00000483, 0.00000212, 0.00000508, 0.00000048, -0.00000110, -0.00000076, -0.00000490, 0.00000442, -0.00000621, 0.00000449, -0.00000397, 0.00000332, 0.00000160, },
		{ -0.00000070, -0.00000222, -0.00000161, 0.00000543, -0.00000321, 0.00000232, -0.00000274, -0.00000091, -0.00000089, 0.00000093, 0.00000199, 0.00000208, -0.00000392, -0.00000225, 0.00000224, 0.00000260, -0.00000325, 0.00000163, -0.00000087, -0.00000051, -0.00000029, -0.00000067, 0.00000196, -0.00000525, -0.00000565, 0.00000585, 0.00000235, 0.00000150, 0.00000325, -0.00000269, -0.00000187, -0.00000244, 0.00000288, -0.00000247, 0.00000174, -0.00000298, 0.00000049, -0.00000234, -0.00000075, 0.00000034, 0.00000581, -0.00000531, 0.00001282, -0.00000826, -0.00000552, -0.00000100, -0.00000065, -0.00000007, 0.00000318, 0.00000039, 0.00000312, -0.00001198, 0.00001792, -0.00002516, 0.00002518, -0.00002650, 0.00003491, -0.00003668, 0.00001991, -0.00002012, 0.00003591, -0.00003533, 0.00000762, -0.00001091, 0.00001744, -0.00000249, -0.00001596, 0.00002314, -0.00003558, 0.00004255, -0.00007142, 0.00008015, -0.00008218, 0.00006091, -0.00004758, 0.00002411, -0.00001316, 0.00001431, -0.00003406, 0.00002128, -0.00000700, -0.00001685, -0.00001586, -0.00002404, -0.00002410, -0.00006369, -0.00034252, 0.00037873, 0.00023408, 0.00042480, -0.00005964, -0.00026533, 0.00000849, -0.00007059, 0.00001896, -0.00003520, -0.00001587, 0.00000059, -0.00002034, 0.00000619, -0.00002231, 0.00001552, -0.00001461, -0.00000439, -0.00000344, -0.00001403, 0.00001104, -0.00000801, -0.00000421, 0.00001663, -0.00000111, -0.00000180, -0.00000436, 0.00000200, -0.00000502, 0.00000331, 0.00000479, -0.00000108, -0.00000000, -0.00000155, -0.00000392, 0.00000390, -0.00000619, 0.00000499, -0.00000465, 0.00000406, 0.00000066, -0.00000531, },
		{ -0.00000296, -0.00000018, 0.00000468, -0.00000319, 0.00000219, -0.00000289, -0.00000028, -0.00000147, 0.00000134, 0.00000224, 0.00000139, -0.00000402, -0.00000193, 0.00000289, 0.00000179, -0.00000318, 0.00000192, -0.00000124, -0.00000017, -0.00000048, -0.00000042, 0.00000173, -0.00000633, -0.00000404, 0.00000613, 0.00000169, 0.00000204, 0.00000240, -0.00000275, -0.00000195, -0.00000198, 0.00000289, -0.00000264, 0.00000187, -0.00000330, 0.00000074, -0.00000225, -0.00000095, 0.00000105, 0.00000481, -0.00000430, 0.00001292, -0.00001093, -0.00000315, -0.00000139, -0.00000054, -0.00000007, 0.00000297, 0.00000180, -0.00000050, -0.00000745, 0.00001318, -0.00002079, 0.00002202, -0.00002458, 0.00003426, -0.00003865, 0.00002277, -0.00002073, 0.00003696, -0.00004048, 0.00001430, -0.00001706, 0.00002609, -0.00001370, -0.00000453, 0.00001215, -0.00002508, 0.00003337, -0.00006408, 0.00007753, -0.00008697, 0.00007399, -0.00006407, 0.00003955, -0.00002605, 0.00002486, -0.00004450, 0.00002967, -0.00001448, -0.00001154, -0.00001996, -0.00002421, -0.00002214, -0.00006939, -0.00034772, 0.00036866, 0.00029537, 0.00043138, -0.00013585, -0.00023901, 0.00001367, -0.00007901, 0.00003036, -0.00004409, -0.00000807, -0.00000637, -0.00001476, 0.00000301, -0.00002028, 0.00001489, -0.00001705, 0.00000116, -0.00000908, -0.00001105, 0.00001184, -0.00001089, -0.00000013, 0.00001433, -0.00000138, -0.00000117, -0.00000484, 0.00000199, -0.00000512, 0.00000462, 0.00000408, -0.00000232, 0.00000100, -0.00000230, -0.00000277, 0.00000300, -0.00000583, 0.00000525, -0.00000521, 0.00000479, -0.00000045, -0.00000466, -0.00000068, },
		{ 0.00000119, 0.00000374, -0.00000297, 0.00000188, -0.00000284, 0.00000019, -0.00000206, 0.00000193, 0.00000226, 0.00000077, -0.00000408, -0.00000154, 0.00000341, 0.00000092, -0.00000292, 0.00000205, -0.00000155, 0.00000018, -0.00000073, -0.00000008, 0.00000126, -0.00000715, -0.00000231, 0.00000605, 0.00000126, 0.00000243, 0.00000153, -0.00000268, -0.00000206, -0.00000146, 0.00000275, -0.00000271, 0.00000195, -0.00000361, 0.00000108, -0.00000227, -0.00000103, 0.00000167, 0.00000375, -0.00000290, 0.00001232, -0.00001307, -0.00000067, -0.00000221, 0.00000003, -0.00000061, 0.00000352, 0.00000220, -0.00000316, -0.00000348, 0.00000851, -0.00001598, 0.00001804, -0.00002158, 0.00003229, -0.00003933, 0.00002488, -0.00002069, 0.00003667, -0.00004400, 0.00001976, -0.00002208, 0.00003350, -0.00002416, 0.00000698, 0.00000032, -0.00001299, 0.00002169, -0.00005300, 0.00006999, -0.00008648, 0.00008282, -0.00007853, 0.00005495, -0.00004023, 0.00003738, -0.00005729, 0.00004090, -0.00002533, -0.00000240, -0.00002815, -0.00002029, -0.00002397, -0.00007229, -0.00035471, 0.00035699, 0.00037074, 0.00041859, -0.00020843, -0.00020353, 0.00001057, -0.00008220, 0.00003868, -0.00005159, -0.00000057, -0.00001386, -0.00000802, -0.00000165, -0.00001661, 0.00001246, -0.00001780, 0.00000580, -0.00001505, -0.00000678, 0.00001118, -0.00001277, 0.00000373, 0.00001153, -0.00000110, -0.00000079, -0.00000521, 0.00000190, -0.00000507, 0.00000594, 0.00000299, -0.00000320, 0.00000183, -0.00000291, -0.00000157, 0.00000180, -0.00000512, 0.00000522, -0.00000556, 0.00000540, -0.00000164, -0.00000371, -0.00000103, -0.00000337, },
		{ 0.00000272, -0.00000258, 0.00000142, -0.00000258, 0.00000045, -0.00000256, 0.00000263, 0.00000206, 0.00000027, -0.00000410, -0.00000107, 0.00000376, 0.00000005, -0.00000251, 0.00000203, -0.00000176, 0.00000050, -0.00000101, 0.00000032, 0.00000057, -0.00000764, -0.00000056, 0.00000567, 0.00000106, 0.00000261, 0.00000070, -0.00000254, -0.00000217, -0.00000091, 0.00000250, -0.00000266, 0.00000195, -0.00000389, 0.00000147, -0.00000240, -0.00000095, 0.00000212, 0.00000273, -0.00000122, 0.00001102, -0.00001454, 0.00000169, -0.00000325, 0.00000088, -0.00000156, 0.00000471, 0.00000162, -0.00000472, -0.00000036, 0.00000427, -0.00001111, 0.00001357, -0.00001776, 0.00002915, -0.00003875, 0.00002620, -0.00002007, 0.00003514, -0.00004586, 0.00002386, -0.00002577, 0.00003931, -0.00003330, 0.00001785, -0.00001157, -0.00000012, 0.00000830, -0.00003886, 0.00005783, -0.00008034, 0.00008634, -0.00008945, 0.00006881, -0.00005444, 0.00005086, -0.00007166, 0.00005445, -0.00003930, 0.00001066, -0.00004086, -0.00001137, -0.00003100, -0.00007049, -0.00036601, 0.00034670, 0.00045873, 0.00038385, -0.00027258, -0.00016206, 0.00000003, -0.00007979, 0.00004310, -0.00005675, 0.00000571, -0.00002109, -0.00000074, -0.00000738, -0.00001152, 0.00000823, -0.00001656, 0.00000897, -0.00002072, -0.00000163, 0.00000914, -0.00001346, 0.00000708, 0.00000846, -0.00000037, -0.00000069, -0.00000546, 0.00000168, -0.00000480, 0.00000718, 0.00000159, -0.00000371, 0.00000243, -0.00000328, -0.00000045, 0.00000038, -0.00000405, 0.00000485, -0.00000562, 0.00000581, -0.00000280, -0.00000255, -0.00000173, -0.00000341, 0.00000236, },
		{ -0.00000209, 0.00000084, -0.00000214, 0.00000047, -0.00000289, 0.00000335, 0.00000168, -0.00000011, -0.00000412, -0.00000052, 0.00000393, -0.00000077, -0.00000198, 0.00000186, -0.00000184, 0.00000076, -0.00000127, 0.00000074, -0.00000032, -0.00000775, 0.00000110, 0.00000508, 0.00000109, 0.00000256, -0.00000003, -0.00000235, -0.00000225, -0.00000035, 0.00000214, -0.00000249, 0.00000184, -0.00000408, 0.00000189, -0.00000264, -0.00000071, 0.00000233, 0.00000187, 0.00000058, 0.00000911, -0.00001523, 0.00000371, -0.00000432, 0.00000183, -0.00000271, 0.00000636, 0.00000015, -0.00000515, 0.00000172, 0.00000078, -0.00000655, 0.00000896, -0.00001342, 0.00002504, -0.00003697, 0.00002674, -0.00001897, 0.00003255, -0.00004613, 0.00002654, -0.00002801, 0.00004329, -0.00004067, 0.00002747, -0.00002276, 0.00001268, -0.00000585, -0.00002259, 0.00004172, -0.00006867, 0.00008389, -0.00009561, 0.00007968, -0.00006728, 0.00006408, -0.00008659, 0.00006945, -0.00005574, 0.00002728, -0.00005813, 0.00000306, -0.00004436, -0.00006214, -0.00038433, 0.00034151, 0.00055663, 0.00032573, -0.00032379, -0.00011829, -0.00001637, -0.00007204, 0.00004320, -0.00005886, 0.00000998, -0.00002724, 0.00000636, -0.00001360, -0.00000542, 0.00000240, -0.00001322, 0.00001023, -0.00002547, 0.00000387, 0.00000596, -0.00001290, 0.00000966, 0.00000535, 0.00000072, -0.00000083, -0.00000557, 0.00000134, -0.00000428, 0.00000824, -0.00000003, -0.00000386, 0.00000278, -0.00000334, 0.00000045, -0.00000114, -0.00000269, 0.00000413, -0.00000535, 0.00000592, -0.00000381, -0.00000128, -0.00000274, -0.00000305, 0.00000326, 0.00000172, },
		{ 0.00000021, -0.00000155, 0.00000025, -0.00000299, 0.00000401, 0.00000116, -0.00000036, -0.00000413, 0.00000009, 0.00000389, -0.00000147, -0.00000138, 0.00000154, -0.00000178, 0.00000092, -0.00000148, 0.00000112, -0.00000135, -0.00000749, 0.00000256, 0.00000436, 0.00000129, 0.00000229, -0.00000060, -0.00000217, -0.00000229, 0.00000017, 0.00000172, -0.00000222, 0.00000160, -0.00000417, 0.00000229, -0.00000294, -0.00000031, 0.00000229, 0.00000127, 0.00000234, 0.00000670, -0.00001513, 0.00000524, -0.00000522, 0.00000271, -0.00000387, 0.00000823, -0.00000199, -0.00000451, 0.00000266, -0.00000173, -0.00000261, 0.00000455, -0.00000890, 0.00002025, -0.00003414, 0.00002655, -0.00001751, 0.00002912, -0.00004495, 0.00002785, -0.00002882, 0.00004534, -0.00004596, 0.00003531, -0.00003258, 0.00002458, -0.00001976, -0.00000532, 0.00002268, -0.00005203, 0.00007530, -0.00009610, 0.00008628, -0.00007736, 0.00007569, -0.00010080, 0.00008474, -0.00007364, 0.00004668, -0.00007953, 0.00002305, -0.00006474, -0.00004574, -0.00041237, 0.00034570, 0.00066046, 0.00024422, -0.00035818, -0.00007617, -0.00003644, -0.00005983, 0.00003907, -0.00005755, 0.00001164, -0.00003158, 0.00001252, -0.00001964, 0.00000114, -0.00000463, -0.00000792, 0.00000931, -0.00002874, 0.00000916, 0.00000199, -0.00001117, 0.00001132, 0.00000243, 0.00000203, -0.00000117, -0.00000555, 0.00000088, -0.00000348, 0.00000903, -0.00000175, -0.00000371, 0.00000287, -0.00000305, 0.00000103, -0.00000262, -0.00000114, 0.00000310, -0.00000471, 0.00000565, -0.00000456, -0.00000002, -0.00000397, -0.00000229, 0.00000381, 0.00000083, -0.00000159, },
		{ -0.00000089, -0.00000019, -0.00000282, 0.00000452, 0.00000059, -0.00000051, -0.00000413, 0.00000076, 0.00000364, -0.00000202, -0.00000078, 0.00000113, -0.00000159, 0.00000096, -0.00000161, 0.00000142, -0.00000245, -0.00000685, 0.00000375, 0.00000362, 0.00000158, 0.00000181, -0.00000099, -0.00000203, -0.00000225, 0.00000061, 0.00000126, -0.00000186, 0.00000124, -0.00000412, 0.00000262, -0.00000328, 0.00000021, 0.00000197, 0.00000103, 0.00000389, 0.00000400, -0.00001429, 0.00000620, -0.00000583, 0.00000335, -0.00000484, 0.00001009, -0.00000457, -0.00000298, 0.00000250, -0.00000315, 0.00000046, 0.00000067, -0.00000455, 0.00001509, -0.00003043, 0.00002569, -0.00001583, 0.00002510, -0.00004255, 0.00002793, -0.00002832, 0.00004553, -0.00004903, 0.00004103, -0.00004050, 0.00003483, -0.00003244, 0.00001174, 0.00000199, -0.00003140, 0.00006089, -0.00009049, 0.00008763, -0.00008344, 0.00008437, -0.00011291, 0.00009894, -0.00009167, 0.00006767, -0.00010412, 0.00004806, -0.00009227, -0.00002030, -0.00045252, 0.00036382, 0.00076509, 0.00014083, -0.00037286, -0.00003956, -0.00005754, -0.00004459, 0.00003129, -0.00005286, 0.00001034, -0.00003356, 0.00001706, -0.00002479, 0.00000749, -0.00001229, -0.00000098, 0.00000616, -0.00003010, 0.00001365, -0.00000231, -0.00000844, 0.00001201, -0.00000013, 0.00000342, -0.00000167, -0.00000544, 0.00000032, -0.00000239, 0.00000945, -0.00000348, -0.00000331, 0.00000275, -0.00000240, 0.00000118, -0.00000390, 0.00000049, 0.00000182, -0.00000374, 0.00000498, -0.00000494, 0.00000107, -0.00000529, -0.00000120, 0.00000398, 0.00000012, -0.00000115, -0.00000041, },
		{ -0.00000077, -0.00000237, 0.00000482, 0.00000001, -0.00000059, -0.00000410, 0.00000144, 0.00000319, -0.00000238, -0.00000022, 0.00000065, -0.00000128, 0.00000088, -0.00000162, 0.00000158, -0.00000354, -0.00000588, 0.00000463, 0.00000292, 0.00000191, 0.00000116, -0.00000120, -0.00000195, -0.00000211, 0.00000095, 0.00000082, -0.00000144, 0.00000078, -0.00000392, 0.00000286, -0.00000359, 0.00000080, 0.00000141, 0.00000119, 0.00000507, 0.00000120, -0.00001284, 0.00000656, -0.00000606, 0.00000366, -0.00000546, 0.00001167, -0.00000728, -0.00000080, 0.00000134, -0.00000343, 0.00000249, -0.00000242, -0.00000067, 0.00000989, -0.00002609, 0.00002427, -0.00001406, 0.00002074, -0.00003918, 0.00002698, -0.00002674, 0.00004403, -0.00004993, 0.00004446, -0.00004615, 0.00004287, -0.00004304, 0.00002738, -0.00001892, -0.00000813, 0.00004147, -0.00007884, 0.00008316, -0.00008453, 0.00008889, -0.00012158, 0.00011058, -0.00010829, 0.00008872, -0.00013049, 0.00007697, -0.00012634, 0.00001446, -0.00050656, 0.00040034, 0.00086449, 0.00001864, -0.00036617, -0.00001188, -0.00007688, -0.00002816, 0.00002091, -0.00004522, 0.00000612, -0.00003286, 0.00001947, -0.00002842, 0.00001293, -0.00001987, 0.00000704, 0.00000095, -0.00002929, 0.00001685, -0.00000646, -0.00000503, 0.00001175, -0.00000220, 0.00000478, -0.00000224, -0.00000526, -0.00000030, -0.00000105, 0.00000947, -0.00000509, -0.00000277, 0.00000249, -0.00000143, 0.00000084, -0.00000485, 0.00000207, 0.00000041, -0.00000249, 0.00000390, -0.00000489, 0.00000189, -0.00000657, 0.00000015, 0.00000378, -0.00000035, -0.00000082, -0.00000094, -0.00000024, },
		{ -0.00000168, 0.00000487, -0.00000049, -0.00000064, -0.00000402, 0.00000209, 0.00000258, -0.00000252, 0.00000024, 0.00000017, -0.00000089, 0.00000068, -0.00000152, 0.00000156, -0.00000453, -0.00000466, 0.00000518, 0.00000235, 0.00000218, 0.00000043, -0.00000123, -0.00000195, -0.00000187, 0.00000116, 0.00000043, -0.00000102, 0.00000023, -0.00000356, 0.00000295, -0.00000384, 0.00000142, 0.00000067, 0.00000174, 0.00000575, -0.00000148, -0.00001095, 0.00000639, -0.00000588, 0.00000357, -0.00000560, 0.00001278, -0.00000985, 0.00000176, -0.00000060, -0.00000266, 0.00000342, -0.00000454, 0.00000245, 0.00000498, -0.00002137, 0.00002241, -0.00001233, 0.00001629, -0.00003512, 0.00002524, -0.00002434, 0.00004114, -0.00004883, 0.00004563, -0.00004939, 0.00004834, -0.00005089, 0.00004053, -0.00003859, 0.00001621, 0.00001829, -0.00006173, 0.00007278, -0.00008006, 0.00008835, -0.00012561, 0.00011821, -0.00012186, 0.00010804, -0.00015680, 0.00010809, -0.00016564, 0.00005792, -0.00057542, 0.00045931, 0.00095206, -0.00011780, -0.00033790, 0.00000415, -0.00009176, -0.00001260, 0.00000931, -0.00003546, -0.00000066, -0.00002948, 0.00001942, -0.00003001, 0.00001680, -0.00002662, 0.00001545, -0.00000589, -0.00002630, 0.00001838, -0.00000999, -0.00000127, 0.00001067, -0.00000374, 0.00000600, -0.00000284, -0.00000507, -0.00000090, 0.00000051, 0.00000903, -0.00000649, -0.00000217, 0.00000216, -0.00000020, -0.00000000, -0.00000537, 0.00000343, -0.00000102, -0.00000105, 0.00000245, -0.00000437, 0.00000232, -0.00000766, 0.00000162, 0.00000326, -0.00000056, -0.00000066, -0.00000133, 0.00000033, -0.00000142, },
		{ 0.00000466, -0.00000087, -0.00000069, -0.00000388, 0.00000269, 0.00000184, -0.00000245, 0.00000056, -0.00000028, -0.00000045, 0.00000038, -0.00000130, 0.00000135, -0.00000534, -0.00000328, 0.00000542, 0.00000193, 0.00000235, -0.00000033, -0.00000113, -0.00000200, -0.00000154, 0.00000122, 0.00000014, -0.00000062, -0.00000037, -0.00000305, 0.00000288, -0.00000397, 0.00000197, -0.00000019, 0.00000267, 0.00000587, -0.00000385, -0.00000881, 0.00000577, -0.00000534, 0.00000308, -0.00000521, 0.00001326, -0.00001199, 0.00000437, -0.00000306, -0.00000101, 0.00000330, -0.00000561, 0.00000460, 0.00000064, -0.00001653, 0.00002023, -0.00001073, 0.00001198, -0.00003064, 0.00002296, -0.00002143, 0.00003721, -0.00004604, 0.00004471, -0.00005026, 0.00005111, -0.00005559, 0.00005033, -0.00005567, 0.00003993, -0.00000709, -0.00004020, 0.00005690, -0.00006989, 0.00008218, -0.00012409, 0.00012059, -0.00013079, 0.00012377, -0.00018099, 0.00013921, -0.00020812, 0.00010855, -0.00065891, 0.00054393, 0.00102105, -0.00026269, -0.00028933, 0.00000678, -0.00009981, -0.00000002, -0.00000185, -0.00002474, -0.00000928, -0.00002375, 0.00001691, -0.00002928, 0.00001858, -0.00003182, 0.00002346, -0.00001377, -0.00002131, 0.00001801, -0.00001248, 0.00000243, 0.00000897, -0.00000474, 0.00000704, -0.00000340, -0.00000490, -0.00000145, 0.00000221, 0.00000815, -0.00000759, -0.00000161, 0.00000187, 0.00000117, -0.00000131, -0.00000536, 0.00000447, -0.00000231, 0.00000044, 0.00000073, -0.00000340, 0.00000230, -0.00000845, 0.00000309, 0.00000249, -0.00000053, -0.00000068, -0.00000154, 0.00000073, -0.00000205, -0.00000081, },
		{ -0.00000109, -0.00000078, -0.00000363, 0.00000317, 0.00000103, -0.00000218, 0.00000071, -0.00000064, -0.00000001, 0.00000002, -0.00000098, 0.00000093, -0.00000591, -0.00000182, 0.00000538, 0.00000168, 0.00000234, -0.00000105, -0.00000094, -0.00000210, -0.00000112, 0.00000112, -0.00000002, -0.00000030, -0.00000097, -0.00000241, 0.00000263, -0.00000394, 0.00000240, -0.00000105, 0.00000389, 0.00000540, -0.00000575, -0.00000662, 0.00000485, -0.00000451, 0.00000226, -0.00000432, 0.00001301, -0.00001349, 0.00000674, -0.00000573, 0.00000127, 0.00000224, -0.00000563, 0.00000567, -0.00000287, -0.00001182, 0.00001788, -0.00000937, 0.00000799, -0.00002599, 0.00002039, -0.00001831, 0.00003265, -0.00004196, 0.00004205, -0.00004899, 0.00005128, -0.00005700, 0.00005621, -0.00006900, 0.00006141, -0.00003291, -0.00001568, 0.00003642, -0.00005437, 0.00007028, -0.00011651, 0.00011678, -0.00013372, 0.00013408, -0.00020082, 0.00016781, -0.00025111, 0.00016380, -0.00075554, 0.00065620, 0.00106499, -0.00040928, -0.00022318, -0.00000462, -0.00009929, 0.00000769, -0.00001090, -0.00001440, -0.00001872, -0.00001630, 0.00001222, -0.00002619, 0.00001797, -0.00003488, 0.00003026, -0.00002194, -0.00001472, 0.00001573, -0.00001362, 0.00000572, 0.00000690, -0.00000527, 0.00000786, -0.00000390, -0.00000477, -0.00000188, 0.00000397, 0.00000684, -0.00000833, -0.00000117, 0.00000173, 0.00000255, -0.00000300, -0.00000483, 0.00000506, -0.00000334, 0.00000185, -0.00000115, -0.00000204, 0.00000180, -0.00000881, 0.00000441, 0.00000157, -0.00000029, -0.00000087, -0.00000155, 0.00000091, -0.00000254, 0.00000017, 0.00000420, },
		{ -0.00000093, -0.00000327, 0.00000351, 0.00000020, -0.00000175, 0.00000070, -0.00000089, 0.00000038, -0.00000037, -0.00000061, 0.00000033, -0.00000618, -0.00000038, 0.00000512, 0.00000161, 0.00000214, -0.00000166, -0.00000070, -0.00000222, -0.00000064, 0.00000088, -0.00000004, -0.00000009, -0.00000152, -0.00000170, 0.00000223, -0.00000373, 0.00000266, -0.00000183, 0.00000528, 0.00000437, -0.00000706, -0.00000457, 0.00000378, -0.00000350, 0.00000119, -0.00000299, 0.00001203, -0.00001421, 0.00000860, -0.00000829, 0.00000392, 0.00000046, -0.00000471, 0.00000562, -0.00000537, -0.00000748, 0.00001547, -0.00000830, 0.00000448, -0.00002140, 0.00001774, -0.00001528, 0.00002786, -0.00003705, 0.00003804, -0.00004595, 0.00004919, -0.00005529, 0.00005796, -0.00007778, 0.00007918, -0.00005737, 0.00001014, 0.00001262, -0.00003432, 0.00005304, -0.00010283, 0.00010627, -0.00012963, 0.00013744, -0.00021420, 0.00019119, -0.00029144, 0.00022028, -0.00086251, 0.00079662, 0.00107820, -0.00055031, -0.00014349, -0.00002947, -0.00008923, 0.00000903, -0.00001626, -0.00000588, -0.00002780, -0.00000803, 0.00000591, -0.00002101, 0.00001490, -0.00003539, 0.00003513, -0.00002959, -0.00000713, 0.00001171, -0.00001326, 0.00000827, 0.00000471, -0.00000543, 0.00000847, -0.00000433, -0.00000471, -0.00000216, 0.00000571, 0.00000516, -0.00000867, -0.00000092, 0.00000183, 0.00000377, -0.00000495, -0.00000378, 0.00000516, -0.00000400, 0.00000303, -0.00000304, -0.00000039, 0.00000085, -0.00000869, 0.00000546, 0.00000062, 0.00000007, -0.00000120, -0.00000138, 0.00000083, -0.00000282, 0.00000115, 0.00000354, -0.00000115, },
		{ -0.00000279, 0.00000365, -0.00000058, -0.00000120, 0.00000052, -0.00000099, 0.00000069, -0.00000074, -0.00000021, -0.00000043, -0.00000613, 0.00000094, 0.00000472, 0.00000166, 0.00000175, -0.00000211, -0.00000047, -0.00000232, -0.00000016, 0.00000053, 0.00000008, -0.00000002, -0.00000196, -0.00000096, 0.00000169, -0.00000334, 0.00000268, -0.00000242, 0.00000670, 0.00000286, -0.00000773, -0.00000281, 0.00000272, -0.00000246, 0.00000004, -0.00000136, 0.00001039, -0.00001409, 0.00000976, -0.00001047, 0.00000660, -0.00000177, -0.00000301, 0.00000452, -0.00000678, -0.00000369, 0.00001312, -0.00000756, 0.00000156, -0.00001702, 0.00001518, -0.00001256, 0.00002321, -0.00003175, 0.00003319, -0.00004165, 0.00004531, -0.00005088, 0.00005572, -0.00008157, 0.00009211, -0.00007877, 0.00003543, -0.00001289, -0.00001099, 0.00003132, -0.00008351, 0.00008908, -0.00011799, 0.00013270, -0.00021926, 0.00020674, -0.00032574, 0.00027388, -0.00097566, 0.00096396, 0.00105620, -0.00067845, -0.00005529, -0.00006602, -0.00006953, 0.00000307, -0.00001665, -0.00000051, -0.00003525, -0.00000001, -0.00000119, -0.00001427, 0.00000961, -0.00003319, 0.00003753, -0.00003593, 0.00000077, 0.00000632, -0.00001138, 0.00000984, 0.00000265, -0.00000536, 0.00000892, -0.00000470, -0.00000470, -0.00000223, 0.00000734, 0.00000317, -0.00000862, -0.00000089, 0.00000225, 0.00000470, -0.00000699, -0.00000230, 0.00000475, -0.00000419, 0.00000385, -0.00000477, 0.00000140, -0.00000049, -0.00000806, 0.00000614, -0.00000025, 0.00000049, -0.00000160, -0.00000107, 0.00000051, -0.00000283, 0.00000205, 0.00000275, -0.00000106, -0.00000115, },
		{ 0.00000358, -0.00000124, -0.00000058, 0.00000021, -0.00000094, 0.00000088, -0.00000106, 0.00000015, -0.00000128, -0.00000577, 0.00000207, 0.00000426, 0.00000179, 0.00000117, -0.00000238, -0.00000029, -0.00000237, 0.00000031, 0.00000010, 0.00000033, -0.00000011, -0.00000227, -0.00000025, 0.00000106, -0.00000280, 0.00000247, -0.00000273, 0.00000803, 0.00000102, -0.00000777, -0.00000146, 0.00000180, -0.00000150, -0.00000107, 0.00000040, 0.00000823, -0.00001314, 0.00001010, -0.00001204, 0.00000905, -0.00000417, -0.00000080, 0.00000250, -0.00000706, -0.00000059, 0.00001094, -0.00000718, -0.00000071, -0.00001301, 0.00001282, -0.00001032, 0.00001899, -0.00002648, 0.00002796, -0.00003659, 0.00004026, -0.00004442, 0.00004995, -0.00008036, 0.00009945, -0.00009570, 0.00005844, -0.00003837, 0.00001411, 0.00000637, -0.00005949, 0.00006574, -0.00009883, 0.00011925, -0.00021464, 0.00021215, -0.00035062, 0.00032003, -0.00108966, 0.00115514, 0.00099615, -0.00078680, 0.00003567, -0.00011148, -0.00004104, -0.00001046, -0.00001124, 0.00000059, -0.00003989, 0.00000661, -0.00000810, -0.00000676, 0.00000262, -0.00002843, 0.00003712, -0.00004026, 0.00000823, 0.00000006, -0.00000814, 0.00001033, 0.00000093, -0.00000518, 0.00000926, -0.00000504, -0.00000472, -0.00000210, 0.00000877, 0.00000098, -0.00000820, -0.00000109, 0.00000304, 0.00000518, -0.00000892, -0.00000053, 0.00000388, -0.00000389, 0.00000420, -0.00000620, 0.00000317, -0.00000210, -0.00000695, 0.00000639, -0.00000093, 0.00000087, -0.00000200, -0.00000069, -0.00000003, -0.00000251, 0.00000276, 0.00000190, -0.00000084, -0.00000143, -0.00000219, },
		{ -0.00000175, 0.00000003, -0.00000021, -0.00000075, 0.00000093, -0.00000128, 0.00000044, -0.00000216, -0.00000511, 0.00000297, 0.00000380, 0.00000193, 0.00000045, -0.00000244, -0.00000019, -0.00000233, 0.00000070, -0.00000036, 0.00000068, -0.00000035, -0.00000240, 0.00000039, 0.00000038, -0.00000212, 0.00000203, -0.00000272, 0.00000910, -0.00000101, -0.00000725, -0.00000057, 0.00000111, -0.00000075, -0.00000198, 0.00000208, 0.00000574, -0.00001147, 0.00000959, -0.00001282, 0.00001099, -0.00000645, 0.00000167, -0.00000021, -0.00000628, 0.00000172, 0.00000898, -0.00000713, -0.00000231, -0.00000943, 0.00001074, -0.00000865, 0.00001543, -0.00002159, 0.00002278, -0.00003131, 0.00003468, -0.00003668, 0.00004141, -0.00007457, 0.00010093, -0.00010709, 0.00007759, -0.00006205, 0.00003924, -0.00002023, -0.00003215, 0.00003733, -0.00007280, 0.00009713, -0.00019962, 0.00020566, -0.00036305, 0.00035405, -0.00119825, 0.00136529, 0.00089711, -0.00086938, 0.00012340, -0.00016222, -0.00000549, -0.00003106, 0.00000026, -0.00000330, -0.00004071, 0.00001074, -0.00001374, 0.00000057, -0.00000530, -0.00002155, 0.00003387, -0.00004209, 0.00001453, -0.00000648, -0.00000383, 0.00000973, -0.00000031, -0.00000502, 0.00000954, -0.00000538, -0.00000474, -0.00000177, 0.00000992, -0.00000132, -0.00000746, -0.00000151, 0.00000420, 0.00000511, -0.00001057, 0.00000137, 0.00000263, -0.00000310, 0.00000403, -0.00000718, 0.00000475, -0.00000381, -0.00000544, 0.00000618, -0.00000136, 0.00000113, -0.00000234, -0.00000033, -0.00000072, -0.00000188, 0.00000321, 0.00000108, -0.00000054, -0.00000176, -0.00000149, 0.00000329, },
		{ 0.00000059, -0.00000067, -0.00000044, 0.00000084, -0.00000139, 0.00000060, -0.00000300, -0.00000421, 0.00000360, 0.00000342, 0.00000202, -0.00000034, -0.00000232, -0.00000019, -0.00000219, 0.00000099, -0.00000080, 0.00000110, -0.00000074, -0.00000233, 0.00000089, -0.00000028, -0.00000138, 0.00000139, -0.00000235, 0.00000982, -0.00000305, -0.00000627, -0.00000013, 0.00000073, -0.00000027, -0.00000257, 0.00000351, 0.00000314, -0.00000926, 0.00000828, -0.00001276, 0.00001225, -0.00000834, 0.00000408, -0.00000335, -0.00000459, 0.00000322, 0.00000731, -0.00000740, -0.00000325, -0.00000632, 0.00000895, -0.00000757, 0.00001265, -0.00001733, 0.00001802, -0.00002626, 0.00002921, -0.00002854, 0.00003107, -0.00006498, 0.00009675, -0.00011236, 0.00009164, -0.00008233, 0.00006267, -0.00004673, -0.00000317, 0.00000537, -0.00004109, 0.00006708, -0.00017419, 0.00018625, -0.00036059, 0.00037149, -0.00129460, 0.00158787, 0.00076023, -0.00092155, 0.00020211, -0.00021412, 0.00003470, -0.00005754, 0.00001753, -0.00001245, -0.00003704, 0.00001144, -0.00001709, 0.00000668, -0.00001321, -0.00001326, 0.00002805, -0.00004116, 0.00001906, -0.00001269, 0.00000115, 0.00000816, -0.00000101, -0.00000496, 0.00000981, -0.00000575, -0.00000472, -0.00000126, 0.00001075, -0.00000363, -0.00000646, -0.00000208, 0.00000570, 0.00000442, -0.00001176, 0.00000321, 0.00000115, -0.00000191, 0.00000332, -0.00000762, 0.00000599, -0.00000547, -0.00000366, 0.00000554, -0.00000148, 0.00000122, -0.00000253, -0.00000007, -0.00000146, -0.00000096, 0.00000337, 0.00000035, -0.00000022, -0.00000209, -0.00000073, 0.00000280, -0.00000206, },
		{ -0.00000114, -0.00000005, 0.00000063, -0.00000138, 0.00000062, -0.00000372, -0.00000314, 0.00000398, 0.00000313, 0.00000200, -0.00000115, -0.00000204, -0.00000029, -0.00000193, 0.00000114, -0.00000118, 0.00000152, -0.00000122, -0.00000207, 0.00000121, -0.00000087, -0.00000063, 0.00000061, -0.00000165, 0.00001009, -0.00000492, -0.00000499, -0.00000011, 0.00000064, -0.00000010, -0.00000276, 0.00000452, 0.00000068, -0.00000672, 0.00000633, -0.00001185, 0.00001270, -0.00000964, 0.00000618, -0.00000663, -0.00000216, 0.00000394, 0.00000594, -0.00000793, -0.00000358, -0.00000371, 0.00000743, -0.00000704, 0.00001070, -0.00001387, 0.00001392, -0.00002181, 0.00002439, -0.00002079, 0.00002001, -0.00005266, 0.00008757, -0.00011144, 0.00009978, -0.00009788, 0.00008278, -0.00007134, 0.00002556, -0.00002827, -0.00000544, 0.00003048, -0.00013918, 0.00015381, -0.00034167, 0.00036852, -0.00137174, 0.00181497, 0.00058878, -0.00094035, 0.00026657, -0.00026286, 0.00007652, -0.00008804, 0.00003964, -0.00002662, -0.00002861, 0.00000810, -0.00001729, 0.00001058, -0.00002005, -0.00000450, 0.00002023, -0.00003749, 0.00002141, -0.00001796, 0.00000633, 0.00000581, -0.00000115, -0.00000505, 0.00001010, -0.00000617, -0.00000461, -0.00000062, 0.00001120, -0.00000584, -0.00000529, -0.00000275, 0.00000745, 0.00000309, -0.00001233, 0.00000480, -0.00000040, -0.00000044, 0.00000212, -0.00000749, 0.00000675, -0.00000689, -0.00000174, 0.00000454, -0.00000131, 0.00000111, -0.00000253, 0.00000000, -0.00000215, 0.00000017, 0.00000321, -0.00000021, 0.00000008, -0.00000239, 0.00000005, 0.00000212, -0.00000215, 0.00000104, },
		{ 0.00000037, 0.00000031, -0.00000125, 0.00000046, -0.00000426, -0.00000198, 0.00000413, 0.00000297, 0.00000182, -0.00000190, -0.00000166, -0.00000048, -0.00000157, 0.00000114, -0.00000144, 0.00000191, -0.00000176, -0.00000163, 0.00000134, -0.00000134, 0.00000006, -0.00000022, -0.00000067, 0.00000987, -0.00000649, -0.00000356, -0.00000040, 0.00000083, -0.00000024, -0.00000251, 0.00000499, -0.00000144, -0.00000411, 0.00000392, -0.00001021, 0.00001232, -0.00001020, 0.00000771, -0.00000977, 0.00000077, 0.00000396, 0.00000487, -0.00000866, -0.00000335, -0.00000156, 0.00000614, -0.00000697, 0.00000952, -0.00001124, 0.00001063, -0.00001817, 0.00002061, -0.00001418, 0.00000930, -0.00003886, 0.00007442, -0.00010473, 0.00010169, -0.00010779, 0.00009824, -0.00009242, 0.00005215, -0.00006153, 0.00003206, -0.00001069, -0.00009620, 0.00010917, -0.00030582, 0.00034229, -0.00142304, 0.00203775, 0.00038806, -0.00092469, 0.00031256, -0.00030427, 0.00011671, -0.00012026, 0.00006511, -0.00004503, -0.00001564, 0.00000045, -0.00001373, 0.00001141, -0.00002478, 0.00000365, 0.00001124, -0.00003142, 0.00002136, -0.00002185, 0.00001125, 0.00000296, -0.00000083, -0.00000529, 0.00001041, -0.00000666, -0.00000437, 0.00000012, 0.00001127, -0.00000785, -0.00000403, -0.00000341, 0.00000936, 0.00000116, -0.00001220, 0.00000595, -0.00000182, 0.00000117, 0.00000054, -0.00000678, 0.00000695, -0.00000793, 0.00000015, 0.00000329, -0.00000087, 0.00000078, -0.00000233, -0.00000018, -0.00000268, 0.00000143, 0.00000276, -0.00000056, 0.00000030, -0.00000262, 0.00000080, 0.00000130, -0.00000202, 0.00000134, -0.00000155, },
		{ -0.00000008, -0.00000103, 0.00000014, -0.00000457, -0.00000082, 0.00000410, 0.00000293, 0.00000146, -0.00000252, -0.00000121, -0.00000073, -0.00000111, 0.00000098, -0.00000157, 0.00000221, -0.00000231, -0.00000104, 0.00000127, -0.00000166, 0.00000064, -0.00000103, 0.00000052, 0.00000916, -0.00000763, -0.00000214, -0.00000089, 0.00000122, -0.00000062, -0.00000186, 0.00000485, -0.00000304, -0.00000168, 0.00000129, -0.00000799, 0.00001117, -0.00000996, 0.00000852, -0.00001248, 0.00000393, 0.00000339, 0.00000405, -0.00000953, -0.00000265, 0.00000015, 0.00000500, -0.00000723, 0.00000901, -0.00000940, 0.00000816, -0.00001543, 0.00001810, -0.00000925, -0.00000008, -0.00002491, 0.00005863, -0.00009311, 0.00009756, -0.00011161, 0.00010809, -0.00010860, 0.00007484, -0.00009232, 0.00006908, -0.00005403, -0.00004756, 0.00005417, -0.00025379, 0.00029127, -0.00144268, 0.00224690, 0.00016508, -0.00087551, 0.00033715, -0.00033472, 0.00015202, -0.00015159, 0.00009205, -0.00006641, 0.00000120, -0.00001133, -0.00000616, 0.00000856, -0.00002649, 0.00001011, 0.00000209, -0.00002354, 0.00001901, -0.00002403, 0.00001551, -0.00000007, -0.00000016, -0.00000566, 0.00001073, -0.00000719, -0.00000398, 0.00000089, 0.00001095, -0.00000961, -0.00000276, -0.00000397, 0.00001126, -0.00000126, -0.00001132, 0.00000653, -0.00000293, 0.00000271, -0.00000127, -0.00000558, 0.00000656, -0.00000847, 0.00000184, 0.00000192, -0.00000023, 0.00000027, -0.00000191, -0.00000065, -0.00000295, 0.00000269, 0.00000206, -0.00000070, 0.00000040, -0.00000274, 0.00000146, 0.00000040, -0.00000168, 0.00000146, -0.00000185, 0.00000077, },
		{ -0.00000075, -0.00000032, -0.00000462, 0.00000027, 0.00000395, 0.00000297, 0.00000091, -0.00000296, -0.00000077, -0.00000100, -0.00000060, 0.00000068, -0.00000155, 0.00000238, -0.00000280, -0.00000036, 0.00000100, -0.00000180, 0.00000106, -0.00000172, 0.00000180, 0.00000802, -0.00000828, -0.00000089, -0.00000146, 0.00000171, -0.00000117, -0.00000086, 0.00000410, -0.00000398, 0.00000037, -0.00000131, -0.00000542, 0.00000938, -0.00000896, 0.00000851, -0.00001456, 0.00000707, 0.00000240, 0.00000345, -0.00001045, -0.00000156, 0.00000145, 0.00000395, -0.00000768, 0.00000901, -0.00000826, 0.00000644, -0.00001355, 0.00001691, -0.00000632, -0.00000734, -0.00001207, 0.00004167, -0.00007781, 0.00008807, -0.00010938, 0.00011184, -0.00011889, 0.00009217, -0.00011871, 0.00010329, -0.00009685, 0.00000390, -0.00000849, -0.00018755, 0.00021542, -0.00142613, 0.00243322, -0.00007175, -0.00079563, 0.00033889, -0.00035138, 0.00017948, -0.00017938, 0.00011829, -0.00008912, 0.00002079, -0.00002665, 0.00000527, 0.00000177, -0.00002451, 0.00001386, -0.00000612, -0.00001470, 0.00001466, -0.00002440, 0.00001877, -0.00000301, 0.00000070, -0.00000609, 0.00001101, -0.00000773, -0.00000344, 0.00000164, 0.00001027, -0.00001103, -0.00000157, -0.00000433, 0.00001301, -0.00000403, -0.00000973, 0.00000645, -0.00000357, 0.00000400, -0.00000313, -0.00000402, 0.00000561, -0.00000842, 0.00000320, 0.00000055, 0.00000053, -0.00000035, -0.00000134, -0.00000142, -0.00000287, 0.00000385, 0.00000120, -0.00000063, 0.00000037, -0.00000272, 0.00000199, -0.00000052, -0.00000119, 0.00000141, -0.00000202, 0.00000112, -0.00000049, },
		{ -0.00000089, -0.00000441, 0.00000122, 0.00000374, 0.00000304, 0.00000020, -0.00000320, -0.00000038, -0.00000126, -0.00000007, 0.00000027, -0.00000136, 0.00000239, -0.00000317, 0.00000037, 0.00000057, -0.00000176, 0.00000127, -0.00000221, 0.00000305, 0.00000653, -0.00000841, 0.00000009, -0.00000199, 0.00000220, -0.00000175, 0.00000035, 0.00000280, -0.00000422, 0.00000185, -0.00000363, -0.00000273, 0.00000713, -0.00000728, 0.00000765, -0.00001584, 0.00000995, 0.00000113, 0.00000299, -0.00001133, -0.00000018, 0.00000240, 0.00000295, -0.00000818, 0.00000936, -0.00000764, 0.00000532, -0.00001236, 0.00001690, -0.00000547, -0.00001195, -0.00000144, 0.00002502, -0.00006028, 0.00007429, -0.00010167, 0.00010951, -0.00012279, 0.00010311, -0.00013909, 0.00013250, -0.00013645, 0.00005498, -0.00007538, -0.00011030, 0.00011638, -0.00137053, 0.00258821, -0.00031319, -0.00068964, 0.00031794, -0.00035245, 0.00019666, -0.00020114, 0.00014160, -0.00011129, 0.00004168, -0.00004449, 0.00002000, -0.00000886, -0.00001848, 0.00001409, -0.00001231, -0.00000586, 0.00000889, -0.00002306, 0.00002085, -0.00000562, 0.00000161, -0.00000651, 0.00001122, -0.00000824, -0.00000275, 0.00000231, 0.00000925, -0.00001210, -0.00000050, -0.00000443, 0.00001442, -0.00000695, -0.00000753, 0.00000567, -0.00000363, 0.00000486, -0.00000484, -0.00000226, 0.00000420, -0.00000778, 0.00000410, -0.00000067, 0.00000129, -0.00000103, -0.00000067, -0.00000244, -0.00000239, 0.00000480, 0.00000026, -0.00000038, 0.00000020, -0.00000254, 0.00000235, -0.00000140, -0.00000057, 0.00000119, -0.00000204, 0.00000136, -0.00000088, -0.00000045, },
		{ -0.00000395, 0.00000199, 0.00000354, 0.00000310, -0.00000064, -0.00000321, -0.00000009, -0.00000145, 0.00000044, -0.00000021, -0.00000104, 0.00000222, -0.00000339, 0.00000106, 0.00000003, -0.00000157, 0.00000129, -0.00000245, 0.00000416, 0.00000482, -0.00000805, 0.00000073, -0.00000237, 0.00000258, -0.00000226, 0.00000164, 0.00000107, -0.00000377, 0.00000265, -0.00000547, -0.00000016, 0.00000466, -0.00000510, 0.00000601, -0.00001623, 0.00001237, -0.00000025, 0.00000260, -0.00001212, 0.00000141, 0.00000302, 0.00000196, -0.00000863, 0.00000988, -0.00000737, 0.00000462, -0.00001162, 0.00001777, -0.00000653, -0.00001369, 0.00000616, 0.00001006, -0.00004208, 0.00005759, -0.00008942, 0.00010161, -0.00012033, 0.00010711, -0.00015232, 0.00015489, -0.00017028, 0.00010242, -0.00014261, -0.00002622, -0.00000254, -0.00127500, 0.00270458, -0.00054965, -0.00056354, 0.00027595, -0.00033731, 0.00020190, -0.00021482, 0.00015987, -0.00013094, 0.00006217, -0.00006352, 0.00003709, -0.00002279, -0.00000843, 0.00001029, -0.00001553, 0.00000194, 0.00000243, -0.00002033, 0.00002173, -0.00000772, 0.00000245, -0.00000686, 0.00001128, -0.00000867, -0.00000192, 0.00000286, 0.00000796, -0.00001279, 0.00000042, -0.00000421, 0.00001535, -0.00000980, -0.00000486, 0.00000424, -0.00000304, 0.00000515, -0.00000619, -0.00000050, 0.00000246, -0.00000659, 0.00000450, -0.00000165, 0.00000197, -0.00000166, 0.00000000, -0.00000366, -0.00000151, 0.00000543, -0.00000066, -0.00000001, -0.00000008, -0.00000222, 0.00000249, -0.00000217, 0.00000010, 0.00000083, -0.00000190, 0.00000146, -0.00000121, -0.00000018, -0.00000150, },
		{ 0.00000255, 0.00000340, 0.00000307, -0.00000155, -0.00000301, 0.00000008, -0.00000156, 0.00000088, -0.00000072, -0.00000061, 0.00000188, -0.00000340, 0.00000167, -0.00000058, -0.00000125, 0.00000113, -0.00000240, 0.00000502, 0.00000302, -0.00000728, 0.00000101, -0.00000254, 0.00000279, -0.00000258, 0.00000284, -0.00000092, -0.00000272, 0.00000275, -0.00000670, 0.00000205, 0.00000220, -0.00000263, 0.00000374, -0.00001572, 0.00001418, -0.00000159, 0.00000220, -0.00001271, 0.00000310, 0.00000334, 0.00000096, -0.00000893, 0.00001042, -0.00000727, 0.00000413, -0.00001103, 0.00001914, -0.00000913, -0.00001263, 0.00001024, -0.00000208, -0.00002475, 0.00003950, -0.00007393, 0.00008905, -0.00011203, 0.00010422, -0.00015780, 0.00016917, -0.00019622, 0.00014314, -0.00020609, 0.00005976, -0.00013653, -0.00114078, 0.00277674, -0.00077183, -0.00042431, 0.00021601, -0.00030655, 0.00019444, -0.00021892, 0.00017131, -0.00014622, 0.00008051, -0.00008219, 0.00005529, -0.00003913, 0.00000519, 0.00000230, -0.00001509, 0.00000773, -0.00000387, -0.00001668, 0.00002153, -0.00000928, 0.00000314, -0.00000709, 0.00001117, -0.00000895, -0.00000101, 0.00000323, 0.00000646, -0.00001310, 0.00000116, -0.00000364, 0.00001567, -0.00001235, -0.00000194, 0.00000227, -0.00000181, 0.00000478, -0.00000703, 0.00000104, 0.00000060, -0.00000494, 0.00000439, -0.00000231, 0.00000248, -0.00000216, 0.00000057, -0.00000495, -0.00000026, 0.00000569, -0.00000146, 0.00000041, -0.00000045, -0.00000175, 0.00000242, -0.00000279, 0.00000077, 0.00000038, -0.00000161, 0.00000141, -0.00000144, 0.00000002, -0.00000210, -0.00000327, },
		{ 0.00000335, 0.00000290, -0.00000245, -0.00000262, 0.00000011, -0.00000156, 0.00000122, -0.00000120, -0.00000012, 0.00000138, -0.00000320, 0.00000214, -0.00000119, -0.00000086, 0.00000081, -0.00000206, 0.00000555, 0.00000128, -0.00000620, 0.00000093, -0.00000247, 0.00000276, -0.00000261, 0.00000380, -0.00000299, -0.00000120, 0.00000218, -0.00000722, 0.00000374, -0.00000002, -0.00000009, 0.00000101, -0.00001438, 0.00001529, -0.00000275, 0.00000174, -0.00001306, 0.00000481, 0.00000340, -0.00000003, -0.00000904, 0.00001087, -0.00000721, 0.00000368, -0.00001033, 0.00002054, -0.00001273, -0.00000915, 0.00001068, -0.00001060, -0.00000960, 0.00002156, -0.00005665, 0.00007312, -0.00009888, 0.00009500, -0.00015556, 0.00017464, -0.00021271, 0.00017447, -0.00026190, 0.00014229, -0.00027946, -0.00097125, 0.00280119, -0.00097125, -0.00027946, 0.00014229, -0.00026190, 0.00017447, -0.00021271, 0.00017464, -0.00015556, 0.00009500, -0.00009888, 0.00007312, -0.00005665, 0.00002156, -0.00000960, -0.00001060, 0.00001068, -0.00000915, -0.00001273, 0.00002054, -0.00001033, 0.00000368, -0.00000721, 0.00001087, -0.00000904, -0.00000003, 0.00000340, 0.00000481, -0.00001306, 0.00000174, -0.00000275, 0.00001529, -0.00001438, 0.00000101, -0.00000009, -0.00000002, 0.00000374, -0.00000722, 0.00000218, -0.00000120, -0.00000299, 0.00000380, -0.00000261, 0.00000276, -0.00000247, 0.00000093, -0.00000620, 0.00000128, 0.00000555, -0.00000206, 0.00000081, -0.00000086, -0.00000119, 0.00000214, -0.00000320, 0.00000138, -0.00000012, -0.00000120, 0.00000122, -0.00000156, 0.00000011, -0.00000262, -0.00000245, 0.00000290, },
		{ 0.00000255, -0.00000327, -0.00000210, 0.00000002, -0.00000144, 0.00000141, -0.00000161, 0.00000038, 0.00000077, -0.00000279, 0.00000242, -0.00000175, -0.00000045, 0.00000041, -0.00000146, 0.00000569, -0.00000026, -0.00000495, 0.00000057, -0.00000216, 0.00000248, -0.00000231, 0.00000439, -0.00000494, 0.00000060, 0.00000104, -0.00000703, 0.00000478, -0.00000181, 0.00000227, -0.00000194, -0.00001235, 0.00001567, -0.00000364, 0.00000116, -0.00001310, 0.00000646, 0.00000323, -0.00000101, -0.00000895, 0.00001117, -0.00000709, 0.00000314, -0.00000928, 0.00002153, -0.00001668, -0.00000387, 0.00000773, -0.00001509, 0.00000230, 0.00000519, -0.00003913, 0.00005529, -0.00008219, 0.00008051, -0.00014622, 0.00017131, -0.00021892, 0.00019444, -0.00030655, 0.00021601, -0.00042431, -0.00077183, 0.00277674, -0.00114078, -0.00013653, 0.00005976, -0.00020609, 0.00014314, -0.00019622, 0.00016917, -0.00015780, 0.00010422, -0.00011203, 0.00008905, -0.00007393, 0.00003950, -0.00002475, -0.00000208, 0.00001024, -0.00001263, -0.00000913, 0.00001914, -0.00001103, 0.00000413, -0.00000727, 0.00001042, -0.00000893, 0.00000096, 0.00000334, 0.00000310, -0.00001271, 0.00000220, -0.00000159, 0.00001418, -0.00001572, 0.00000374, -0.00000263, 0.00000220, 0.00000205, -0.00000670, 0.00000275, -0.00000272, -0.00000092, 0.00000284, -0.00000258, 0.00000279, -0.00000254, 0.00000101, -0.00000728, 0.00000302, 0.00000502, -0.00000240, 0.00000113, -0.00000125, -0.00000058, 0.00000167, -0.00000340, 0.00000188, -0.00000061, -0.00000072, 0.00000088, -0.00000156, 0.00000008, -0.00000301, -0.00000155, 0.00000307, 0.00000340, },
		{ -0.00000395, -0.00000150, -0.00000018, -0.00000121, 0.00000146, -0.00000190, 0.00000083, 0.00000010, -0.00000217, 0.00000249, -0.00000222, -0.00000008, -0.00000001, -0.00000066, 0.00000543, -0.00000151, -0.00000366, 0.00000000, -0.00000166, 0.00000197, -0.00000165, 0.00000450, -0.00000659, 0.00000246, -0.00000050, -0.00000619, 0.00000515, -0.00000304, 0.00000424, -0.00000486, -0.00000980, 0.00001535, -0.00000421, 0.00000042, -0.00001279, 0.00000796, 0.00000286, -0.00000192, -0.00000867, 0.00001128, -0.00000686, 0.00000245, -0.00000772, 0.00002173, -0.00002033, 0.00000243, 0.00000194, -0.00001553, 0.00001029, -0.00000843, -0.00002279, 0.00003709, -0.00006352, 0.00006217, -0.00013094, 0.00015987, -0.00021482, 0.00020190, -0.00033731, 0.00027595, -0.00056354, -0.00054965, 0.00270458, -0.00127500, -0.00000254, -0.00002622, -0.00014261, 0.00010242, -0.00017028, 0.00015489, -0.00015232, 0.00010711, -0.00012033, 0.00010161, -0.00008942, 0.00005759, -0.00004208, 0.00001006, 0.00000616, -0.00001369, -0.00000653, 0.00001777, -0.00001162, 0.00000462, -0.00000737, 0.00000988, -0.00000863, 0.00000196, 0.00000302, 0.00000141, -0.00001212, 0.00000260, -0.00000025, 0.00001237, -0.00001623, 0.00000601, -0.00000510, 0.00000466, -0.00000016, -0.00000547, 0.00000265, -0.00000377, 0.00000107, 0.00000164, -0.00000226, 0.00000258, -0.00000237, 0.00000073, -0.00000805, 0.00000482, 0.00000416, -0.00000245, 0.00000129, -0.00000157, 0.00000003, 0.00000106, -0.00000339, 0.00000222, -0.00000104, -0.00000021, 0.00000044, -0.00000145, -0.00000009, -0.00000321, -0.00000064, 0.00000310, 0.00000354, 0.00000199, },
		{ -0.00000089, -0.00000045, -0.00000088, 0.00000136, -0.00000204, 0.00000119, -0.00000057, -0.00000140, 0.00000235, -0.00000254, 0.00000020, -0.00000038, 0.00000026, 0.00000480, -0.00000239, -0.00000244, -0.00000067, -0.00000103, 0.00000129, -0.00000067, 0.00000410, -0.00000778, 0.00000420, -0.00000226, -0.00000484, 0.00000486, -0.00000363, 0.00000567, -0.00000753, -0.00000695, 0.00001442, -0.00000443, -0.00000050, -0.00001210, 0.00000925, 0.00000231, -0.00000275, -0.00000824, 0.00001122, -0.00000651, 0.00000161, -0.00000562, 0.00002085, -0.00002306, 0.00000889, -0.00000586, -0.00001231, 0.00001409, -0.00001848, -0.00000886, 0.00002000, -0.00004449, 0.00004168, -0.00011129, 0.00014160, -0.00020114, 0.00019666, -0.00035245, 0.00031794, -0.00068964, -0.00031319, 0.00258821, -0.00137053, 0.00011638, -0.00011030, -0.00007538, 0.00005498, -0.00013645, 0.00013250, -0.00013909, 0.00010311, -0.00012279, 0.00010951, -0.00010167, 0.00007429, -0.00006028, 0.00002502, -0.00000144, -0.00001195, -0.00000547, 0.00001690, -0.00001236, 0.00000532, -0.00000764, 0.00000936, -0.00000818, 0.00000295, 0.00000240, -0.00000018, -0.00001133, 0.00000299, 0.00000113, 0.00000995, -0.00001584, 0.00000765, -0.00000728, 0.00000713, -0.00000273, -0.00000363, 0.00000185, -0.00000422, 0.00000280, 0.00000035, -0.00000175, 0.00000220, -0.00000199, 0.00000009, -0.00000841, 0.00000653, 0.00000305, -0.00000221, 0.00000127, -0.00000176, 0.00000057, 0.00000037, -0.00000317, 0.00000239, -0.00000136, 0.00000027, -0.00000007, -0.00000126, -0.00000038, -0.00000320, 0.00000020, 0.00000304, 0.00000374, 0.00000122, -0.00000441, },
		{ -0.00000075, -0.00000049, 0.00000112, -0.00000202, 0.00000141, -0.00000119, -0.00000052, 0.00000199, -0.00000272, 0.00000037, -0.00000063, 0.00000120, 0.00000385, -0.00000287, -0.00000142, -0.00000134, -0.00000035, 0.00000053, 0.00000055, 0.00000320, -0.00000842, 0.00000561, -0.00000402, -0.00000313, 0.00000400, -0.00000357, 0.00000645, -0.00000973, -0.00000403, 0.00001301, -0.00000433, -0.00000157, -0.00001103, 0.00001027, 0.00000164, -0.00000344, -0.00000773, 0.00001101, -0.00000609, 0.00000070, -0.00000301, 0.00001877, -0.00002440, 0.00001466, -0.00001470, -0.00000612, 0.00001386, -0.00002451, 0.00000177, 0.00000527, -0.00002665, 0.00002079, -0.00008912, 0.00011829, -0.00017938, 0.00017948, -0.00035138, 0.00033889, -0.00079563, -0.00007175, 0.00243322, -0.00142613, 0.00021542, -0.00018755, -0.00000849, 0.00000390, -0.00009685, 0.00010329, -0.00011871, 0.00009217, -0.00011889, 0.00011184, -0.00010938, 0.00008807, -0.00007781, 0.00004167, -0.00001207, -0.00000734, -0.00000632, 0.00001691, -0.00001355, 0.00000644, -0.00000826, 0.00000901, -0.00000768, 0.00000395, 0.00000145, -0.00000156, -0.00001045, 0.00000345, 0.00000240, 0.00000707, -0.00001456, 0.00000851, -0.00000896, 0.00000938, -0.00000542, -0.00000131, 0.00000037, -0.00000398, 0.00000410, -0.00000086, -0.00000117, 0.00000171, -0.00000146, -0.00000089, -0.00000828, 0.00000802, 0.00000180, -0.00000172, 0.00000106, -0.00000180, 0.00000100, -0.00000036, -0.00000280, 0.00000238, -0.00000155, 0.00000068, -0.00000060, -0.00000100, -0.00000077, -0.00000296, 0.00000091, 0.00000297, 0.00000395, 0.00000027, -0.00000462, -0.00000032, },
		{ -0.00000008, 0.00000077, -0.00000185, 0.00000146, -0.00000168, 0.00000040, 0.00000146, -0.00000274, 0.00000040, -0.00000070, 0.00000206, 0.00000269, -0.00000295, -0.00000065, -0.00000191, 0.00000027, -0.00000023, 0.00000192, 0.00000184, -0.00000847, 0.00000656, -0.00000558, -0.00000127, 0.00000271, -0.00000293, 0.00000653, -0.00001132, -0.00000126, 0.00001126, -0.00000397, -0.00000276, -0.00000961, 0.00001095, 0.00000089, -0.00000398, -0.00000719, 0.00001073, -0.00000566, -0.00000016, -0.00000007, 0.00001551, -0.00002403, 0.00001901, -0.00002354, 0.00000209, 0.00001011, -0.00002649, 0.00000856, -0.00000616, -0.00001133, 0.00000120, -0.00006641, 0.00009205, -0.00015159, 0.00015202, -0.00033472, 0.00033715, -0.00087551, 0.00016508, 0.00224690, -0.00144268, 0.00029127, -0.00025379, 0.00005417, -0.00004756, -0.00005403, 0.00006908, -0.00009232, 0.00007484, -0.00010860, 0.00010809, -0.00011161, 0.00009756, -0.00009311, 0.00005863, -0.00002491, -0.00000008, -0.00000925, 0.00001810, -0.00001543, 0.00000816, -0.00000940, 0.00000901, -0.00000723, 0.00000500, 0.00000015, -0.00000265, -0.00000953, 0.00000405, 0.00000339, 0.00000393, -0.00001248, 0.00000852, -0.00000996, 0.00001117, -0.00000799, 0.00000129, -0.00000168, -0.00000304, 0.00000485, -0.00000186, -0.00000062, 0.00000122, -0.00000089, -0.00000214, -0.00000763, 0.00000916, 0.00000052, -0.00000103, 0.00000064, -0.00000166, 0.00000127, -0.00000104, -0.00000231, 0.00000221, -0.00000157, 0.00000098, -0.00000111, -0.00000073, -0.00000121, -0.00000252, 0.00000146, 0.00000293, 0.00000410, -0.00000082, -0.00000457, 0.00000014, -0.00000103, },
		{ 0.00000037, -0.00000155, 0.00000134, -0.00000202, 0.00000130, 0.00000080, -0.00000262, 0.00000030, -0.00000056, 0.00000276, 0.00000143, -0.00000268, -0.00000018, -0.00000233, 0.00000078, -0.00000087, 0.00000329, 0.00000015, -0.00000793, 0.00000695, -0.00000678, 0.00000054, 0.00000117, -0.00000182, 0.00000595, -0.00001220, 0.00000116, 0.00000936, -0.00000341, -0.00000403, -0.00000785, 0.00001127, 0.00000012, -0.00000437, -0.00000666, 0.00001041, -0.00000529, -0.00000083, 0.00000296, 0.00001125, -0.00002185, 0.00002136, -0.00003142, 0.00001124, 0.00000365, -0.00002478, 0.00001141, -0.00001373, 0.00000045, -0.00001564, -0.00004503, 0.00006511, -0.00012026, 0.00011671, -0.00030427, 0.00031256, -0.00092469, 0.00038806, 0.00203775, -0.00142304, 0.00034229, -0.00030582, 0.00010917, -0.00009620, -0.00001069, 0.00003206, -0.00006153, 0.00005215, -0.00009242, 0.00009824, -0.00010779, 0.00010169, -0.00010473, 0.00007442, -0.00003886, 0.00000930, -0.00001418, 0.00002061, -0.00001817, 0.00001063, -0.00001124, 0.00000952, -0.00000697, 0.00000614, -0.00000156, -0.00000335, -0.00000866, 0.00000487, 0.00000396, 0.00000077, -0.00000977, 0.00000771, -0.00001020, 0.00001232, -0.00001021, 0.00000392, -0.00000411, -0.00000144, 0.00000499, -0.00000251, -0.00000024, 0.00000083, -0.00000040, -0.00000356, -0.00000649, 0.00000987, -0.00000067, -0.00000022, 0.00000006, -0.00000134, 0.00000134, -0.00000163, -0.00000176, 0.00000191, -0.00000144, 0.00000114, -0.00000157, -0.00000048, -0.00000166, -0.00000190, 0.00000182, 0.00000297, 0.00000413, -0.00000198, -0.00000426, 0.00000046, -0.00000125, 0.00000031, },
		{ -0.00000114, 0.00000104, -0.00000215, 0.00000212, 0.00000005, -0.00000239, 0.00000008, -0.00000021, 0.00000321, 0.00000017, -0.00000215, 0.00000000, -0.00000253, 0.00000111, -0.00000131, 0.00000454, -0.00000174, -0.00000689, 0.00000675, -0.00000749, 0.00000212, -0.00000044, -0.00000040, 0.00000480, -0.00001233, 0.00000309, 0.00000745, -0.00000275, -0.00000529, -0.00000584, 0.00001120, -0.00000062, -0.00000461, -0.00000617, 0.00001010, -0.00000505, -0.00000115, 0.00000581, 0.00000633, -0.00001796, 0.00002141, -0.00003749, 0.00002023, -0.00000450, -0.00002005, 0.00001058, -0.00001729, 0.00000810, -0.00002861, -0.00002662, 0.00003964, -0.00008804, 0.00007652, -0.00026286, 0.00026657, -0.00094035, 0.00058878, 0.00181497, -0.00137174, 0.00036852, -0.00034167, 0.00015381, -0.00013918, 0.00003048, -0.00000544, -0.00002827, 0.00002556, -0.00007134, 0.00008278, -0.00009788, 0.00009978, -0.00011144, 0.00008757, -0.00005266, 0.00002001, -0.00002079, 0.00002439, -0.00002181, 0.00001392, -0.00001387, 0.00001070, -0.00000704, 0.00000743, -0.00000371, -0.00000358, -0.00000793, 0.00000594, 0.00000394, -0.00000216, -0.00000663, 0.00000618, -0.00000964, 0.00001270, -0.00001185, 0.00000633, -0.00000672, 0.00000068, 0.00000452, -0.00000276, -0.00000010, 0.00000064, -0.00000011, -0.00000499, -0.00000492, 0.00001009, -0.00000165, 0.00000061, -0.00000063, -0.00000087, 0.00000121, -0.00000207, -0.00000122, 0.00000152, -0.00000118, 0.00000114, -0.00000193, -0.00000029, -0.00000204, -0.00000115, 0.00000200, 0.00000313, 0.00000398, -0.00000314, -0.00000372, 0.00000062, -0.00000138, 0.00000063, -0.00000005, },
		{ 0.00000059, -0.00000206, 0.00000280, -0.00000073, -0.00000209, -0.00000022, 0.00000035, 0.00000337, -0.00000096, -0.00000146, -0.00000007, -0.00000253, 0.00000122, -0.00000148, 0.00000554, -0.00000366, -0.00000547, 0.00000599, -0.00000762, 0.00000332, -0.00000191, 0.00000115, 0.00000321, -0.00001176, 0.00000442, 0.00000570, -0.00000208, -0.00000646, -0.00000363, 0.00001075, -0.00000126, -0.00000472, -0.00000575, 0.00000981, -0.00000496, -0.00000101, 0.00000816, 0.00000115, -0.00001269, 0.00001906, -0.00004116, 0.00002805, -0.00001326, -0.00001321, 0.00000668, -0.00001709, 0.00001144, -0.00003704, -0.00001245, 0.00001753, -0.00005754, 0.00003470, -0.00021412, 0.00020211, -0.00092155, 0.00076023, 0.00158787, -0.00129460, 0.00037149, -0.00036059, 0.00018625, -0.00017419, 0.00006708, -0.00004109, 0.00000537, -0.00000317, -0.00004673, 0.00006267, -0.00008233, 0.00009164, -0.00011236, 0.00009675, -0.00006498, 0.00003107, -0.00002854, 0.00002921, -0.00002626, 0.00001802, -0.00001733, 0.00001265, -0.00000757, 0.00000895, -0.00000632, -0.00000325, -0.00000740, 0.00000731, 0.00000322, -0.00000459, -0.00000335, 0.00000408, -0.00000834, 0.00001225, -0.00001276, 0.00000828, -0.00000926, 0.00000314, 0.00000351, -0.00000257, -0.00000027, 0.00000073, -0.00000013, -0.00000627, -0.00000305, 0.00000982, -0.00000235, 0.00000139, -0.00000138, -0.00000028, 0.00000089, -0.00000233, -0.00000074, 0.00000110, -0.00000080, 0.00000099, -0.00000219, -0.00000019, -0.00000232, -0.00000034, 0.00000202, 0.00000342, 0.00000360, -0.00000421, -0.00000300, 0.00000060, -0.00000139, 0.00000084, -0.00000044, -0.00000067, },
		{ -0.00000175, 0.00000329, -0.00000149, -0.00000176, -0.00000054, 0.00000108, 0.00000321, -0.00000188, -0.00000072, -0.00000033, -0.00000234, 0.00000113, -0.00000136, 0.00000618, -0.00000544, -0.00000381, 0.00000475, -0.00000718, 0.00000403, -0.00000310, 0.00000263, 0.00000137, -0.00001057, 0.00000511, 0.00000420, -0.00000151, -0.00000746, -0.00000132, 0.00000992, -0.00000177, -0.00000474, -0.00000538, 0.00000954, -0.00000502, -0.00000031, 0.00000973, -0.00000383, -0.00000648, 0.00001453, -0.00004209, 0.00003387, -0.00002155, -0.00000530, 0.00000057, -0.00001374, 0.00001074, -0.00004071, -0.00000330, 0.00000026, -0.00003106, -0.00000549, -0.00016222, 0.00012340, -0.00086938, 0.00089711, 0.00136529, -0.00119825, 0.00035405, -0.00036305, 0.00020566, -0.00019962, 0.00009713, -0.00007280, 0.00003733, -0.00003215, -0.00002023, 0.00003924, -0.00006205, 0.00007759, -0.00010709, 0.00010093, -0.00007457, 0.00004141, -0.00003668, 0.00003468, -0.00003131, 0.00002278, -0.00002159, 0.00001543, -0.00000865, 0.00001074, -0.00000943, -0.00000231, -0.00000713, 0.00000898, 0.00000172, -0.00000628, -0.00000021, 0.00000167, -0.00000645, 0.00001099, -0.00001282, 0.00000959, -0.00001147, 0.00000574, 0.00000208, -0.00000198, -0.00000075, 0.00000111, -0.00000057, -0.00000725, -0.00000101, 0.00000910, -0.00000272, 0.00000203, -0.00000212, 0.00000038, 0.00000039, -0.00000240, -0.00000035, 0.00000068, -0.00000036, 0.00000070, -0.00000233, -0.00000019, -0.00000244, 0.00000045, 0.00000193, 0.00000380, 0.00000297, -0.00000511, -0.00000216, 0.00000044, -0.00000128, 0.00000093, -0.00000075, -0.00000021, 0.00000003, },
		{ 0.00000358, -0.00000219, -0.00000143, -0.00000084, 0.00000190, 0.00000276, -0.00000251, -0.00000003, -0.00000069, -0.00000200, 0.00000087, -0.00000093, 0.00000639, -0.00000695, -0.00000210, 0.00000317, -0.00000620, 0.00000420, -0.00000389, 0.00000388, -0.00000053, -0.00000892, 0.00000518, 0.00000304, -0.00000109, -0.00000820, 0.00000098, 0.00000877, -0.00000210, -0.00000472, -0.00000504, 0.00000926, -0.00000518, 0.00000093, 0.00001033, -0.00000814, 0.00000006, 0.00000823, -0.00004026, 0.00003712, -0.00002843, 0.00000262, -0.00000676, -0.00000810, 0.00000661, -0.00003989, 0.00000059, -0.00001124, -0.00001046, -0.00004104, -0.00011148, 0.00003567, -0.00078680, 0.00099615, 0.00115514, -0.00108966, 0.00032003, -0.00035062, 0.00021215, -0.00021464, 0.00011925, -0.00009883, 0.00006574, -0.00005949, 0.00000637, 0.00001411, -0.00003837, 0.00005844, -0.00009570, 0.00009945, -0.00008036, 0.00004995, -0.00004442, 0.00004026, -0.00003659, 0.00002796, -0.00002648, 0.00001899, -0.00001032, 0.00001282, -0.00001301, -0.00000071, -0.00000718, 0.00001094, -0.00000059, -0.00000706, 0.00000250, -0.00000080, -0.00000417, 0.00000905, -0.00001204, 0.00001010, -0.00001314, 0.00000823, 0.00000040, -0.00000107, -0.00000150, 0.00000180, -0.00000146, -0.00000777, 0.00000102, 0.00000803, -0.00000273, 0.00000247, -0.00000280, 0.00000106, -0.00000025, -0.00000227, -0.00000011, 0.00000033, 0.00000010, 0.00000031, -0.00000237, -0.00000029, -0.00000238, 0.00000117, 0.00000179, 0.00000426, 0.00000207, -0.00000577, -0.00000128, 0.00000015, -0.00000106, 0.00000088, -0.00000094, 0.00000021, -0.00000058, -0.00000124, },
		{ -0.00000279, -0.00000115, -0.00000106, 0.00000275, 0.00000205, -0.00000283, 0.00000051, -0.00000107, -0.00000160, 0.00000049, -0.00000025, 0.00000614, -0.00000806, -0.00000049, 0.00000140, -0.00000477, 0.00000385, -0.00000419, 0.00000475, -0.00000230, -0.00000699, 0.00000470, 0.00000225, -0.00000089, -0.00000862, 0.00000317, 0.00000734, -0.00000223, -0.00000470, -0.00000470, 0.00000892, -0.00000536, 0.00000265, 0.00000984, -0.00001138, 0.00000632, 0.00000077, -0.00003593, 0.00003753, -0.00003319, 0.00000961, -0.00001427, -0.00000119, -0.00000001, -0.00003525, -0.00000051, -0.00001665, 0.00000307, -0.00006953, -0.00006602, -0.00005529, -0.00067845, 0.00105620, 0.00096396, -0.00097566, 0.00027388, -0.00032574, 0.00020674, -0.00021926, 0.00013270, -0.00011799, 0.00008908, -0.00008351, 0.00003132, -0.00001099, -0.00001289, 0.00003543, -0.00007877, 0.00009211, -0.00008157, 0.00005572, -0.00005088, 0.00004531, -0.00004165, 0.00003319, -0.00003175, 0.00002321, -0.00001256, 0.00001518, -0.00001702, 0.00000156, -0.00000756, 0.00001312, -0.00000369, -0.00000678, 0.00000452, -0.00000301, -0.00000177, 0.00000660, -0.00001047, 0.00000976, -0.00001409, 0.00001039, -0.00000136, 0.00000004, -0.00000246, 0.00000272, -0.00000281, -0.00000773, 0.00000286, 0.00000670, -0.00000242, 0.00000268, -0.00000334, 0.00000169, -0.00000096, -0.00000196, -0.00000002, 0.00000008, 0.00000053, -0.00000016, -0.00000232, -0.00000047, -0.00000211, 0.00000175, 0.00000166, 0.00000472, 0.00000094, -0.00000613, -0.00000043, -0.00000021, -0.00000074, 0.00000069, -0.00000099, 0.00000052, -0.00000120, -0.00000058, 0.00000365, },
		{ -0.00000093, -0.00000115, 0.00000354, 0.00000115, -0.00000282, 0.00000083, -0.00000138, -0.00000120, 0.00000007, 0.00000062, 0.00000546, -0.00000869, 0.00000085, -0.00000039, -0.00000304, 0.00000303, -0.00000400, 0.00000516, -0.00000378, -0.00000495, 0.00000377, 0.00000183, -0.00000092, -0.00000867, 0.00000516, 0.00000571, -0.00000216, -0.00000471, -0.00000433, 0.00000847, -0.00000543, 0.00000471, 0.00000827, -0.00001326, 0.00001171, -0.00000713, -0.00002959, 0.00003513, -0.00003539, 0.00001490, -0.00002101, 0.00000591, -0.00000803, -0.00002780, -0.00000588, -0.00001626, 0.00000903, -0.00008923, -0.00002947, -0.00014349, -0.00055031, 0.00107820, 0.00079662, -0.00086251, 0.00022028, -0.00029144, 0.00019119, -0.00021420, 0.00013744, -0.00012963, 0.00010627, -0.00010283, 0.00005304, -0.00003432, 0.00001262, 0.00001014, -0.00005737, 0.00007918, -0.00007778, 0.00005796, -0.00005529, 0.00004919, -0.00004595, 0.00003804, -0.00003705, 0.00002786, -0.00001528, 0.00001774, -0.00002140, 0.00000448, -0.00000830, 0.00001547, -0.00000748, -0.00000537, 0.00000562, -0.00000471, 0.00000046, 0.00000392, -0.00000829, 0.00000860, -0.00001421, 0.00001203, -0.00000299, 0.00000119, -0.00000350, 0.00000378, -0.00000457, -0.00000706, 0.00000437, 0.00000528, -0.00000183, 0.00000266, -0.00000373, 0.00000223, -0.00000170, -0.00000152, -0.00000009, -0.00000004, 0.00000088, -0.00000064, -0.00000222, -0.00000070, -0.00000166, 0.00000214, 0.00000161, 0.00000512, -0.00000038, -0.00000618, 0.00000033, -0.00000061, -0.00000037, 0.00000038, -0.00000089, 0.00000070, -0.00000175, 0.00000020, 0.00000351, -0.00000327, },
		{ -0.00000109, 0.00000420, 0.00000017, -0.00000254, 0.00000091, -0.00000155, -0.00000087, -0.00000029, 0.00000157, 0.00000441, -0.00000881, 0.00000180, -0.00000204, -0.00000115, 0.00000185, -0.00000334, 0.00000506, -0.00000483, -0.00000300, 0.00000255, 0.00000173, -0.00000117, -0.00000833, 0.00000684, 0.00000397, -0.00000188, -0.00000477, -0.00000390, 0.00000786, -0.00000527, 0.00000690, 0.00000572, -0.00001362, 0.00001573, -0.00001472, -0.00002194, 0.00003026, -0.00003488, 0.00001797, -0.00002619, 0.00001222, -0.00001630, -0.00001872, -0.00001440, -0.00001090, 0.00000769, -0.00009929, -0.00000462, -0.00022318, -0.00040928, 0.00106499, 0.00065620, -0.00075554, 0.00016380, -0.00025111, 0.00016781, -0.00020082, 0.00013408, -0.00013372, 0.00011678, -0.00011651, 0.00007028, -0.00005437, 0.00003642, -0.00001568, -0.00003291, 0.00006141, -0.00006900, 0.00005621, -0.00005700, 0.00005128, -0.00004899, 0.00004205, -0.00004196, 0.00003265, -0.00001831, 0.00002039, -0.00002599, 0.00000799, -0.00000937, 0.00001788, -0.00001182, -0.00000287, 0.00000567, -0.00000563, 0.00000224, 0.00000127, -0.00000573, 0.00000674, -0.00001349, 0.00001301, -0.00000432, 0.00000226, -0.00000451, 0.00000485, -0.00000662, -0.00000575, 0.00000540, 0.00000389, -0.00000105, 0.00000240, -0.00000394, 0.00000263, -0.00000241, -0.00000097, -0.00000030, -0.00000002, 0.00000112, -0.00000112, -0.00000210, -0.00000094, -0.00000105, 0.00000234, 0.00000168, 0.00000538, -0.00000182, -0.00000591, 0.00000093, -0.00000098, 0.00000002, -0.00000001, -0.00000064, 0.00000071, -0.00000218, 0.00000103, 0.00000317, -0.00000363, -0.00000078, },
		{ 0.00000466, -0.00000081, -0.00000205, 0.00000073, -0.00000154, -0.00000068, -0.00000053, 0.00000249, 0.00000309, -0.00000845, 0.00000230, -0.00000340, 0.00000073, 0.00000044, -0.00000231, 0.00000447, -0.00000536, -0.00000131, 0.00000117, 0.00000187, -0.00000161, -0.00000759, 0.00000815, 0.00000221, -0.00000145, -0.00000490, -0.00000340, 0.00000704, -0.00000474, 0.00000897, 0.00000243, -0.00001248, 0.00001801, -0.00002131, -0.00001377, 0.00002346, -0.00003182, 0.00001858, -0.00002928, 0.00001691, -0.00002375, -0.00000928, -0.00002474, -0.00000185, -0.00000002, -0.00009981, 0.00000678, -0.00028933, -0.00026269, 0.00102105, 0.00054393, -0.00065891, 0.00010855, -0.00020812, 0.00013921, -0.00018099, 0.00012377, -0.00013079, 0.00012059, -0.00012409, 0.00008218, -0.00006989, 0.00005690, -0.00004020, -0.00000709, 0.00003993, -0.00005567, 0.00005033, -0.00005559, 0.00005111, -0.00005026, 0.00004471, -0.00004604, 0.00003721, -0.00002143, 0.00002296, -0.00003064, 0.00001198, -0.00001073, 0.00002023, -0.00001653, 0.00000064, 0.00000460, -0.00000561, 0.00000330, -0.00000101, -0.00000306, 0.00000437, -0.00001199, 0.00001326, -0.00000521, 0.00000308, -0.00000534, 0.00000577, -0.00000881, -0.00000385, 0.00000587, 0.00000267, -0.00000019, 0.00000197, -0.00000397, 0.00000288, -0.00000305, -0.00000037, -0.00000062, 0.00000014, 0.00000122, -0.00000154, -0.00000200, -0.00000113, -0.00000033, 0.00000235, 0.00000193, 0.00000542, -0.00000328, -0.00000534, 0.00000135, -0.00000130, 0.00000038, -0.00000045, -0.00000028, 0.00000056, -0.00000245, 0.00000184, 0.00000269, -0.00000388, -0.00000069, -0.00000087, },
		{ -0.00000168, -0.00000142, 0.00000033, -0.00000133, -0.00000066, -0.00000056, 0.00000326, 0.00000162, -0.00000766, 0.00000232, -0.00000437, 0.00000245, -0.00000105, -0.00000102, 0.00000343, -0.00000537, -0.00000000, -0.00000020, 0.00000216, -0.00000217, -0.00000649, 0.00000903, 0.00000051, -0.00000090, -0.00000507, -0.00000284, 0.00000600, -0.00000374, 0.00001067, -0.00000127, -0.00000999, 0.00001838, -0.00002630, -0.00000589, 0.00001545, -0.00002662, 0.00001680, -0.00003001, 0.00001942, -0.00002948, -0.00000066, -0.00003546, 0.00000931, -0.00001260, -0.00009176, 0.00000415, -0.00033790, -0.00011780, 0.00095206, 0.00045931, -0.00057542, 0.00005792, -0.00016564, 0.00010809, -0.00015680, 0.00010804, -0.00012186, 0.00011821, -0.00012561, 0.00008835, -0.00008006, 0.00007278, -0.00006173, 0.00001829, 0.00001621, -0.00003859, 0.00004053, -0.00005089, 0.00004834, -0.00004939, 0.00004563, -0.00004883, 0.00004114, -0.00002434, 0.00002524, -0.00003512, 0.00001629, -0.00001233, 0.00002241, -0.00002137, 0.00000498, 0.00000245, -0.00000454, 0.00000342, -0.00000266, -0.00000060, 0.00000176, -0.00000985, 0.00001278, -0.00000560, 0.00000357, -0.00000588, 0.00000639, -0.00001095, -0.00000148, 0.00000575, 0.00000174, 0.00000067, 0.00000142, -0.00000384, 0.00000295, -0.00000356, 0.00000023, -0.00000102, 0.00000043, 0.00000116, -0.00000187, -0.00000195, -0.00000123, 0.00000043, 0.00000218, 0.00000235, 0.00000518, -0.00000466, -0.00000453, 0.00000156, -0.00000152, 0.00000068, -0.00000089, 0.00000017, 0.00000024, -0.00000252, 0.00000258, 0.00000209, -0.00000402, -0.00000064, -0.00000049, 0.00000487, },
		{ -0.00000077, -0.00000024, -0.00000094, -0.00000082, -0.00000035, 0.00000378, 0.00000015, -0.00000657, 0.00000189, -0.00000489, 0.00000390, -0.00000249, 0.00000041, 0.00000207, -0.00000485, 0.00000084, -0.00000143, 0.00000249, -0.00000277, -0.00000509, 0.00000947, -0.00000105, -0.00000030, -0.00000526, -0.00000224, 0.00000478, -0.00000220, 0.00001175, -0.00000503, -0.00000646, 0.00001685, -0.00002929, 0.00000095, 0.00000704, -0.00001987, 0.00001293, -0.00002842, 0.00001947, -0.00003286, 0.00000612, -0.00004522, 0.00002091, -0.00002816, -0.00007688, -0.00001188, -0.00036617, 0.00001864, 0.00086449, 0.00040034, -0.00050656, 0.00001446, -0.00012634, 0.00007697, -0.00013049, 0.00008872, -0.00010829, 0.00011058, -0.00012158, 0.00008889, -0.00008453, 0.00008316, -0.00007884, 0.00004147, -0.00000813, -0.00001892, 0.00002738, -0.00004304, 0.00004287, -0.00004615, 0.00004446, -0.00004993, 0.00004403, -0.00002674, 0.00002698, -0.00003918, 0.00002074, -0.00001406, 0.00002427, -0.00002609, 0.00000989, -0.00000067, -0.00000242, 0.00000249, -0.00000343, 0.00000134, -0.00000080, -0.00000728, 0.00001167, -0.00000546, 0.00000366, -0.00000606, 0.00000656, -0.00001284, 0.00000120, 0.00000507, 0.00000119, 0.00000141, 0.00000080, -0.00000359, 0.00000286, -0.00000392, 0.00000078, -0.00000144, 0.00000082, 0.00000095, -0.00000211, -0.00000195, -0.00000120, 0.00000116, 0.00000191, 0.00000292, 0.00000463, -0.00000588, -0.00000354, 0.00000158, -0.00000162, 0.00000088, -0.00000128, 0.00000065, -0.00000022, -0.00000238, 0.00000319, 0.00000144, -0.00000410, -0.00000059, 0.00000001, 0.00000482, -0.00000237, },
		{ -0.00000089, -0.00000041, -0.00000115, 0.00000012, 0.00000398, -0.00000120, -0.00000529, 0.00000107, -0.00000494, 0.00000498, -0.00000374, 0.00000182, 0.00000049, -0.00000390, 0.00000118, -0.00000240, 0.00000275, -0.00000331, -0.00000348, 0.00000945, -0.00000239, 0.00000032, -0.00000544, -0.00000167, 0.00000342, -0.00000013, 0.00001201, -0.00000844, -0.00000231, 0.00001365, -0.00003010, 0.00000616, -0.00000098, -0.00001229, 0.00000749, -0.00002479, 0.00001706, -0.00003356, 0.00001034, -0.00005286, 0.00003129, -0.00004459, -0.00005754, -0.00003956, -0.00037286, 0.00014083, 0.00076509, 0.00036382, -0.00045252, -0.00002030, -0.00009227, 0.00004806, -0.00010412, 0.00006767, -0.00009167, 0.00009894, -0.00011291, 0.00008437, -0.00008344, 0.00008763, -0.00009049, 0.00006089, -0.00003140, 0.00000199, 0.00001174, -0.00003244, 0.00003483, -0.00004050, 0.00004103, -0.00004903, 0.00004553, -0.00002832, 0.00002793, -0.00004255, 0.00002510, -0.00001583, 0.00002569, -0.00003043, 0.00001509, -0.00000455, 0.00000067, 0.00000046, -0.00000315, 0.00000250, -0.00000298, -0.00000457, 0.00001009, -0.00000484, 0.00000335, -0.00000583, 0.00000620, -0.00001429, 0.00000400, 0.00000389, 0.00000103, 0.00000197, 0.00000021, -0.00000328, 0.00000262, -0.00000412, 0.00000124, -0.00000186, 0.00000126, 0.00000061, -0.00000225, -0.00000203, -0.00000099, 0.00000181, 0.00000158, 0.00000362, 0.00000375, -0.00000685, -0.00000245, 0.00000142, -0.00000161, 0.00000096, -0.00000159, 0.00000113, -0.00000078, -0.00000202, 0.00000364, 0.00000076, -0.00000413, -0.00000051, 0.00000059, 0.00000452, -0.00000282, -0.00000019, },
		{ 0.00000021, -0.00000159, 0.00000083, 0.00000381, -0.00000229, -0.00000397, -0.00000002, -0.00000456, 0.00000565, -0.00000471, 0.00000310, -0.00000114, -0.00000262, 0.00000103, -0.00000305, 0.00000287, -0.00000371, -0.00000175, 0.00000903, -0.00000348, 0.00000088, -0.00000555, -0.00000117, 0.00000203, 0.00000243, 0.00001132, -0.00001117, 0.00000199, 0.00000916, -0.00002874, 0.00000931, -0.00000792, -0.00000463, 0.00000114, -0.00001964, 0.00001252, -0.00003158, 0.00001164, -0.00005755, 0.00003907, -0.00005983, -0.00003644, -0.00007617, -0.00035818, 0.00024422, 0.00066046, 0.00034570, -0.00041237, -0.00004574, -0.00006474, 0.00002305, -0.00007953, 0.00004668, -0.00007364, 0.00008474, -0.00010080, 0.00007569, -0.00007736, 0.00008628, -0.00009610, 0.00007530, -0.00005203, 0.00002268, -0.00000532, -0.00001976, 0.00002458, -0.00003258, 0.00003531, -0.00004596, 0.00004534, -0.00002882, 0.00002785, -0.00004495, 0.00002912, -0.00001751, 0.00002655, -0.00003414, 0.00002025, -0.00000890, 0.00000455, -0.00000261, -0.00000173, 0.00000266, -0.00000451, -0.00000199, 0.00000823, -0.00000387, 0.00000271, -0.00000522, 0.00000524, -0.00001513, 0.00000670, 0.00000234, 0.00000127, 0.00000229, -0.00000031, -0.00000294, 0.00000229, -0.00000417, 0.00000160, -0.00000222, 0.00000172, 0.00000017, -0.00000229, -0.00000217, -0.00000060, 0.00000229, 0.00000129, 0.00000436, 0.00000256, -0.00000749, -0.00000135, 0.00000112, -0.00000148, 0.00000092, -0.00000178, 0.00000154, -0.00000138, -0.00000147, 0.00000389, 0.00000009, -0.00000413, -0.00000036, 0.00000116, 0.00000401, -0.00000299, 0.00000025, -0.00000155, },
		{ -0.00000209, 0.00000172, 0.00000326, -0.00000305, -0.00000274, -0.00000128, -0.00000381, 0.00000592, -0.00000535, 0.00000413, -0.00000269, -0.00000114, 0.00000045, -0.00000334, 0.00000278, -0.00000386, -0.00000003, 0.00000824, -0.00000428, 0.00000134, -0.00000557, -0.00000083, 0.00000072, 0.00000535, 0.00000966, -0.00001290, 0.00000596, 0.00000387, -0.00002547, 0.00001023, -0.00001322, 0.00000240, -0.00000542, -0.00001360, 0.00000636, -0.00002724, 0.00000998, -0.00005886, 0.00004320, -0.00007204, -0.00001637, -0.00011829, -0.00032379, 0.00032573, 0.00055663, 0.00034151, -0.00038433, -0.00006214, -0.00004436, 0.00000306, -0.00005813, 0.00002728, -0.00005574, 0.00006945, -0.00008659, 0.00006408, -0.00006728, 0.00007968, -0.00009561, 0.00008389, -0.00006867, 0.00004172, -0.00002259, -0.00000585, 0.00001268, -0.00002276, 0.00002747, -0.00004067, 0.00004329, -0.00002801, 0.00002654, -0.00004613, 0.00003255, -0.00001897, 0.00002674, -0.00003697, 0.00002504, -0.00001342, 0.00000896, -0.00000655, 0.00000078, 0.00000172, -0.00000515, 0.00000015, 0.00000636, -0.00000271, 0.00000183, -0.00000432, 0.00000371, -0.00001523, 0.00000911, 0.00000058, 0.00000187, 0.00000233, -0.00000071, -0.00000264, 0.00000189, -0.00000408, 0.00000184, -0.00000249, 0.00000214, -0.00000035, -0.00000225, -0.00000235, -0.00000003, 0.00000256, 0.00000109, 0.00000508, 0.00000110, -0.00000775, -0.00000032, 0.00000074, -0.00000127, 0.00000076, -0.00000184, 0.00000186, -0.00000198, -0.00000077, 0.00000393, -0.00000052, -0.00000412, -0.00000011, 0.00000168, 0.00000335, -0.00000289, 0.00000047, -0.00000214, 0.00000084, },
		{ 0.00000272, 0.00000236, -0.00000341, -0.00000173, -0.00000255, -0.00000280, 0.00000581, -0.00000562, 0.00000485, -0.00000405, 0.00000038, -0.00000045, -0.00000328, 0.00000243, -0.00000371, 0.00000159, 0.00000718, -0.00000480, 0.00000168, -0.00000546, -0.00000069, -0.00000037, 0.00000846, 0.00000708, -0.00001346, 0.00000914, -0.00000163, -0.00002072, 0.00000897, -0.00001656, 0.00000823, -0.00001152, -0.00000738, -0.00000074, -0.00002109, 0.00000571, -0.00005675, 0.00004310, -0.00007979, 0.00000003, -0.00016206, -0.00027258, 0.00038385, 0.00045873, 0.00034670, -0.00036601, -0.00007049, -0.00003100, -0.00001137, -0.00004086, 0.00001066, -0.00003930, 0.00005445, -0.00007166, 0.00005086, -0.00005444, 0.00006881, -0.00008945, 0.00008634, -0.00008034, 0.00005783, -0.00003886, 0.00000830, -0.00000012, -0.00001157, 0.00001785, -0.00003330, 0.00003931, -0.00002577, 0.00002386, -0.00004586, 0.00003514, -0.00002007, 0.00002620, -0.00003875, 0.00002915, -0.00001776, 0.00001357, -0.00001111, 0.00000427, -0.00000036, -0.00000472, 0.00000162, 0.00000471, -0.00000156, 0.00000088, -0.00000325, 0.00000169, -0.00001454, 0.00001102, -0.00000122, 0.00000273, 0.00000212, -0.00000095, -0.00000240, 0.00000147, -0.00000389, 0.00000195, -0.00000266, 0.00000250, -0.00000091, -0.00000217, -0.00000254, 0.00000070, 0.00000261, 0.00000106, 0.00000567, -0.00000056, -0.00000764, 0.00000057, 0.00000032, -0.00000101, 0.00000050, -0.00000176, 0.00000203, -0.00000251, 0.00000005, 0.00000376, -0.00000107, -0.00000410, 0.00000027, 0.00000206, 0.00000263, -0.00000256, 0.00000045, -0.00000258, 0.00000142, -0.00000258, },
		{ 0.00000119, -0.00000337, -0.00000103, -0.00000371, -0.00000164, 0.00000540, -0.00000556, 0.00000522, -0.00000512, 0.00000180, -0.00000157, -0.00000291, 0.00000183, -0.00000320, 0.00000299, 0.00000594, -0.00000507, 0.00000190, -0.00000521, -0.00000079, -0.00000110, 0.00001153, 0.00000373, -0.00001277, 0.00001118, -0.00000678, -0.00001505, 0.00000580, -0.00001780, 0.00001246, -0.00001661, -0.00000165, -0.00000802, -0.00001386, -0.00000057, -0.00005159, 0.00003868, -0.00008220, 0.00001057, -0.00020353, -0.00020843, 0.00041859, 0.00037074, 0.00035699, -0.00035471, -0.00007229, -0.00002397, -0.00002029, -0.00002815, -0.00000240, -0.00002533, 0.00004090, -0.00005729, 0.00003738, -0.00004023, 0.00005495, -0.00007853, 0.00008282, -0.00008648, 0.00006999, -0.00005300, 0.00002169, -0.00001299, 0.00000032, 0.00000698, -0.00002416, 0.00003350, -0.00002208, 0.00001976, -0.00004400, 0.00003667, -0.00002069, 0.00002488, -0.00003933, 0.00003229, -0.00002158, 0.00001804, -0.00001598, 0.00000851, -0.00000348, -0.00000316, 0.00000220, 0.00000352, -0.00000061, 0.00000003, -0.00000221, -0.00000067, -0.00001307, 0.00001232, -0.00000290, 0.00000375, 0.00000167, -0.00000103, -0.00000227, 0.00000108, -0.00000361, 0.00000195, -0.00000271, 0.00000275, -0.00000146, -0.00000206, -0.00000268, 0.00000153, 0.00000243, 0.00000126, 0.00000605, -0.00000231, -0.00000715, 0.00000126, -0.00000008, -0.00000073, 0.00000018, -0.00000155, 0.00000205, -0.00000292, 0.00000092, 0.00000341, -0.00000154, -0.00000408, 0.00000077, 0.00000226, 0.00000193, -0.00000206, 0.00000019, -0.00000284, 0.00000188, -0.00000297, 0.00000374, },
		{ -0.00000296, -0.00000068, -0.00000466, -0.00000045, 0.00000479, -0.00000521, 0.00000525, -0.00000583, 0.00000300, -0.00000277, -0.00000230, 0.00000100, -0.00000232, 0.00000408, 0.00000462, -0.00000512, 0.00000199, -0.00000484, -0.00000117, -0.00000138, 0.00001433, -0.00000013, -0.00001089, 0.00001184, -0.00001105, -0.00000908, 0.00000116, -0.00001705, 0.00001489, -0.00002028, 0.00000301, -0.00001476, -0.00000637, -0.00000807, -0.00004409, 0.00003036, -0.00007901, 0.00001367, -0.00023901, -0.00013585, 0.00043138, 0.00029537, 0.00036866, -0.00034772, -0.00006939, -0.00002214, -0.00002421, -0.00001996, -0.00001154, -0.00001448, 0.00002967, -0.00004450, 0.00002486, -0.00002605, 0.00003955, -0.00006407, 0.00007399, -0.00008697, 0.00007753, -0.00006408, 0.00003337, -0.00002508, 0.00001215, -0.00000453, -0.00001370, 0.00002609, -0.00001706, 0.00001430, -0.00004048, 0.00003696, -0.00002073, 0.00002277, -0.00003865, 0.00003426, -0.00002458, 0.00002202, -0.00002079, 0.00001318, -0.00000745, -0.00000050, 0.00000180, 0.00000297, -0.00000007, -0.00000054, -0.00000139, -0.00000315, -0.00001093, 0.00001292, -0.00000430, 0.00000481, 0.00000105, -0.00000095, -0.00000225, 0.00000074, -0.00000330, 0.00000187, -0.00000264, 0.00000289, -0.00000198, -0.00000195, -0.00000275, 0.00000240, 0.00000204, 0.00000169, 0.00000613, -0.00000404, -0.00000633, 0.00000173, -0.00000042, -0.00000048, -0.00000017, -0.00000124, 0.00000192, -0.00000318, 0.00000179, 0.00000289, -0.00000193, -0.00000402, 0.00000139, 0.00000224, 0.00000134, -0.00000147, -0.00000028, -0.00000289, 0.00000219, -0.00000319, 0.00000468, -0.00000018, },
		{ -0.00000070, -0.00000531, 0.00000066, 0.00000406, -0.00000465, 0.00000499, -0.00000619, 0.00000390, -0.00000392, -0.00000155, -0.00000000, -0.00000108, 0.00000479, 0.00000331, -0.00000502, 0.00000200, -0.00000436, -0.00000180, -0.00000111, 0.00001663, -0.00000421, -0.00000801, 0.00001104, -0.00001403, -0.00000344, -0.00000439, -0.00001461, 0.00001552, -0.00002231, 0.00000619, -0.00002034, 0.00000059, -0.00001587, -0.00003520, 0.00001896, -0.00007059, 0.00000849, -0.00026533, -0.00005964, 0.00042480, 0.00023408, 0.00037873, -0.00034252, -0.00006369, -0.00002410, -0.00002404, -0.00001586, -0.00001685, -0.00000700, 0.00002128, -0.00003406, 0.00001431, -0.00001316, 0.00002411, -0.00004758, 0.00006091, -0.00008218, 0.00008015, -0.00007142, 0.00004255, -0.00003558, 0.00002314, -0.00001596, -0.00000249, 0.00001744, -0.00001091, 0.00000762, -0.00003533, 0.00003591, -0.00002012, 0.00001991, -0.00003668, 0.00003491, -0.00002650, 0.00002518, -0.00002516, 0.00001792, -0.00001198, 0.00000312, 0.00000039, 0.00000318, -0.00000007, -0.00000065, -0.00000100, -0.00000552, -0.00000826, 0.00001282, -0.00000531, 0.00000581, 0.00000034, -0.00000075, -0.00000234, 0.00000049, -0.00000298, 0.00000174, -0.00000247, 0.00000288, -0.00000244, -0.00000187, -0.00000269, 0.00000325, 0.00000150, 0.00000235, 0.00000585, -0.00000565, -0.00000525, 0.00000196, -0.00000067, -0.00000029, -0.00000051, -0.00000087, 0.00000163, -0.00000325, 0.00000260, 0.00000224, -0.00000225, -0.00000392, 0.00000208, 0.00000199, 0.00000093, -0.00000089, -0.00000091, -0.00000274, 0.00000232, -0.00000321, 0.00000543, -0.00000161, -0.00000222, },
		{ -0.00000562, 0.00000160, 0.00000332, -0.00000397, 0.00000449, -0.00000621, 0.00000442, -0.00000490, -0.00000076, -0.00000110, 0.00000048, 0.00000508, 0.00000212, -0.00000483, 0.00000196, -0.00000383, -0.00000264, -0.00000028, 0.00001822, -0.00000819, -0.00000439, 0.00000886, -0.00001544, 0.00000132, -0.00001024, -0.00001090, 0.00001456, -0.00002270, 0.00000768, -0.00002430, 0.00000635, -0.00002309, -0.00002598, 0.00000564, -0.00005787, -0.00000494, -0.00028011, 0.00001547, 0.00040232, 0.00018708, 0.00038507, -0.00033702, -0.00005701, -0.00002836, -0.00002089, -0.00001509, -0.00001873, -0.00000276, 0.00001588, -0.00002635, 0.00000639, -0.00000260, 0.00000999, -0.00003061, 0.00004494, -0.00007286, 0.00007799, -0.00007470, 0.00004863, -0.00004381, 0.00003258, -0.00002659, 0.00000883, 0.00000804, -0.00000397, 0.00000001, -0.00002869, 0.00003347, -0.00001881, 0.00001638, -0.00003350, 0.00003418, -0.00002718, 0.00002724, -0.00002876, 0.00002236, -0.00001674, 0.00000746, -0.00000194, 0.00000419, -0.00000071, -0.00000019, -0.00000124, -0.00000750, -0.00000530, 0.00001207, -0.00000587, 0.00000663, -0.00000041, -0.00000047, -0.00000253, 0.00000033, -0.00000269, 0.00000157, -0.00000223, 0.00000275, -0.00000282, -0.00000182, -0.00000248, 0.00000401, 0.00000088, 0.00000322, 0.00000517, -0.00000703, -0.00000400, 0.00000197, -0.00000080, -0.00000018, -0.00000079, -0.00000047, 0.00000122, -0.00000312, 0.00000330, 0.00000149, -0.00000250, -0.00000375, 0.00000280, 0.00000152, 0.00000076, -0.00000040, -0.00000162, -0.00000243, 0.00000228, -0.00000300, 0.00000593, -0.00000300, -0.00000127, -0.00000106, },
		{ 0.00000230, 0.00000266, -0.00000325, 0.00000384, -0.00000594, 0.00000456, -0.00000563, -0.00000003, -0.00000220, 0.00000229, 0.00000495, 0.00000110, -0.00000462, 0.00000192, -0.00000333, -0.00000359, 0.00000109, 0.00001897, -0.00001174, -0.00000041, 0.00000551, -0.00001520, 0.00000474, -0.00001577, -0.00000647, 0.00001235, -0.00002162, 0.00000746, -0.00002641, 0.00001039, -0.00002897, -0.00001750, -0.00000829, -0.00004226, -0.00002581, -0.00028188, 0.00008518, 0.00036791, 0.00015357, 0.00038645, -0.00032964, -0.00005089, -0.00003348, -0.00001597, -0.00001670, -0.00001785, -0.00000136, 0.00001330, -0.00002145, 0.00000144, 0.00000494, -0.00000165, -0.00001467, 0.00002758, -0.00006012, 0.00007156, -0.00007389, 0.00005130, -0.00004927, 0.00003986, -0.00003573, 0.00001958, -0.00000159, 0.00000335, -0.00000819, -0.00002078, 0.00002969, -0.00001681, 0.00001227, -0.00002922, 0.00003209, -0.00002652, 0.00002801, -0.00003126, 0.00002611, -0.00002133, 0.00001221, -0.00000503, 0.00000595, -0.00000201, 0.00000090, -0.00000224, -0.00000886, -0.00000229, 0.00001081, -0.00000596, 0.00000721, -0.00000112, -0.00000015, -0.00000279, 0.00000026, -0.00000243, 0.00000140, -0.00000194, 0.00000249, -0.00000313, -0.00000180, -0.00000209, 0.00000459, 0.00000027, 0.00000421, 0.00000409, -0.00000807, -0.00000266, 0.00000178, -0.00000082, -0.00000018, -0.00000098, -0.00000009, 0.00000073, -0.00000280, 0.00000386, 0.00000069, -0.00000270, -0.00000347, 0.00000349, 0.00000087, 0.00000085, -0.00000009, -0.00000233, -0.00000201, 0.00000211, -0.00000256, 0.00000612, -0.00000424, -0.00000021, -0.00000171, -0.00000558, },
		{ 0.00000214, -0.00000258, 0.00000314, -0.00000548, 0.00000434, -0.00000604, 0.00000055, -0.00000320, 0.00000425, 0.00000440, 0.00000032, -0.00000444, 0.00000193, -0.00000291, -0.00000455, 0.00000291, 0.00001879, -0.00001459, 0.00000356, 0.00000132, -0.00001339, 0.00000654, -0.00002045, -0.00000187, 0.00000934, -0.00001942, 0.00000573, -0.00002664, 0.00001244, -0.00003292, -0.00001069, -0.00002146, -0.00002547, -0.00005257, -0.00027023, 0.00014588, 0.00032576, 0.00013191, 0.00038243, -0.00031941, -0.00004652, -0.00003821, -0.00001044, -0.00001969, -0.00001506, -0.00000215, 0.00001309, -0.00001908, -0.00000057, 0.00000913, -0.00001000, -0.00000109, 0.00001038, -0.00004525, 0.00006166, -0.00006935, 0.00005053, -0.00005170, 0.00004453, -0.00004281, 0.00002914, -0.00001089, 0.00001060, -0.00001654, -0.00001195, 0.00002469, -0.00001417, 0.00000773, -0.00002403, 0.00002876, -0.00002453, 0.00002739, -0.00003244, 0.00002883, -0.00002535, 0.00001701, -0.00000862, 0.00000832, -0.00000391, 0.00000262, -0.00000407, -0.00000936, 0.00000047, 0.00000921, -0.00000564, 0.00000751, -0.00000174, 0.00000016, -0.00000309, 0.00000027, -0.00000223, 0.00000126, -0.00000163, 0.00000213, -0.00000336, -0.00000181, -0.00000152, 0.00000495, -0.00000025, 0.00000526, 0.00000265, -0.00000871, -0.00000134, 0.00000144, -0.00000074, -0.00000028, -0.00000107, 0.00000022, 0.00000020, -0.00000230, 0.00000425, -0.00000015, -0.00000285, -0.00000306, 0.00000408, 0.00000010, 0.00000119, -0.00000002, -0.00000295, -0.00000154, 0.00000185, -0.00000193, 0.00000597, -0.00000524, 0.00000086, -0.00000257, -0.00000522, 0.00000274, },
		{ -0.00000201, 0.00000247, -0.00000492, 0.00000382, -0.00000612, 0.00000092, -0.00000402, 0.00000626, 0.00000350, -0.00000019, -0.00000433, 0.00000204, -0.00000266, -0.00000539, 0.00000504, 0.00001770, -0.00001652, 0.00000713, -0.00000328, -0.00001023, 0.00000662, -0.00002388, 0.00000239, 0.00000601, -0.00001651, 0.00000283, -0.00002518, 0.00001245, -0.00003465, -0.00000627, -0.00003263, -0.00000934, -0.00008312, -0.00024577, 0.00019486, 0.00027993, 0.00011993, 0.00037332, -0.00030593, -0.00004463, -0.00004160, -0.00000531, -0.00002308, -0.00001121, -0.00000439, 0.00001461, -0.00001873, 0.00000001, 0.00001001, -0.00001466, 0.00000914, -0.00000525, -0.00002967, 0.00004937, -0.00006168, 0.00004658, -0.00005109, 0.00004637, -0.00004740, 0.00003695, -0.00001931, 0.00001730, -0.00002458, -0.00000260, 0.00001869, -0.00001098, 0.00000294, -0.00001817, 0.00002436, -0.00002131, 0.00002537, -0.00003217, 0.00003026, -0.00002846, 0.00002147, -0.00001241, 0.00001110, -0.00000629, 0.00000488, -0.00000673, -0.00000886, 0.00000274, 0.00000749, -0.00000499, 0.00000755, -0.00000224, 0.00000045, -0.00000340, 0.00000035, -0.00000209, 0.00000117, -0.00000133, 0.00000170, -0.00000352, -0.00000181, -0.00000080, 0.00000504, -0.00000060, 0.00000625, 0.00000091, -0.00000891, -0.00000013, 0.00000100, -0.00000059, -0.00000048, -0.00000104, 0.00000044, -0.00000032, -0.00000165, 0.00000444, -0.00000098, -0.00000293, -0.00000253, 0.00000452, -0.00000070, 0.00000175, -0.00000021, -0.00000341, -0.00000111, 0.00000159, -0.00000117, 0.00000551, -0.00000593, 0.00000183, -0.00000353, -0.00000458, 0.00000291, 0.00000179, },
		{ 0.00000190, -0.00000436, 0.00000308, -0.00000588, 0.00000103, -0.00000457, 0.00000821, 0.00000231, -0.00000044, -0.00000432, 0.00000225, -0.00000263, -0.00000596, 0.00000729, 0.00001579, -0.00001741, 0.00000997, -0.00000785, -0.00000611, 0.00000506, -0.00002580, 0.00000588, 0.00000285, -0.00001338, -0.00000077, -0.00002239, 0.00001064, -0.00003410, -0.00000463, -0.00004083, 0.00000436, -0.00011495, -0.00021002, 0.00023045, 0.00023404, 0.00011515, 0.00035994, -0.00028931, -0.00004552, -0.00004306, -0.00000131, -0.00002608, -0.00000712, -0.00000727, 0.00001713, -0.00001971, 0.00000257, 0.00000799, -0.00001561, 0.00001538, -0.00001814, -0.00001471, 0.00003586, -0.00005174, 0.00003994, -0.00004765, 0.00004537, -0.00004926, 0.00004258, -0.00002639, 0.00002301, -0.00003181, 0.00000680, 0.00001194, -0.00000738, -0.00000193, -0.00001188, 0.00001915, -0.00001705, 0.00002206, -0.00003041, 0.00003022, -0.00003033, 0.00002523, -0.00001604, 0.00001403, -0.00000894, 0.00000751, -0.00001012, -0.00000729, 0.00000429, 0.00000586, -0.00000415, 0.00000736, -0.00000262, 0.00000067, -0.00000370, 0.00000048, -0.00000201, 0.00000113, -0.00000108, 0.00000121, -0.00000363, -0.00000177, 0.00000005, 0.00000487, -0.00000070, 0.00000707, -0.00000103, -0.00000867, 0.00000091, 0.00000053, -0.00000039, -0.00000074, -0.00000089, 0.00000053, -0.00000077, -0.00000089, 0.00000444, -0.00000180, -0.00000295, -0.00000186, 0.00000474, -0.00000144, 0.00000245, -0.00000068, -0.00000366, -0.00000077, 0.00000137, -0.00000036, 0.00000477, -0.00000629, 0.00000260, -0.00000449, -0.00000373, 0.00000283, 0.00000163, -0.00000159, },
		{ -0.00000390, 0.00000223, -0.00000536, 0.00000087, -0.00000480, 0.00000999, 0.00000091, -0.00000044, -0.00000441, 0.00000259, -0.00000285, -0.00000616, 0.00000945, 0.00001323, -0.00001725, 0.00001185, -0.00001196, -0.00000148, 0.00000209, -0.00002613, 0.00000829, 0.00000023, -0.00001048, -0.00000458, -0.00001876, 0.00000739, -0.00003152, -0.00000586, -0.00004544, 0.00001412, -0.00014545, -0.00016527, 0.00025203, 0.00019113, 0.00011507, 0.00034347, -0.00027009, -0.00004910, -0.00004233, 0.00000105, -0.00002806, -0.00000348, -0.00001011, 0.00001993, -0.00002127, 0.00000634, 0.00000375, -0.00001323, 0.00001742, -0.00002742, -0.00000159, 0.00002234, -0.00004050, 0.00003133, -0.00004186, 0.00004175, -0.00004836, 0.00004577, -0.00003176, 0.00002735, -0.00003779, 0.00001578, 0.00000478, -0.00000353, -0.00000667, -0.00000545, 0.00001339, -0.00001200, 0.00001764, -0.00002727, 0.00002867, -0.00003077, 0.00002796, -0.00001920, 0.00001682, -0.00001163, 0.00001030, -0.00001407, -0.00000466, 0.00000496, 0.00000454, -0.00000326, 0.00000701, -0.00000290, 0.00000085, -0.00000397, 0.00000064, -0.00000196, 0.00000114, -0.00000088, 0.00000070, -0.00000371, -0.00000167, 0.00000095, 0.00000443, -0.00000050, 0.00000761, -0.00000306, -0.00000800, 0.00000172, 0.00000008, -0.00000018, -0.00000105, -0.00000064, 0.00000049, -0.00000112, -0.00000008, 0.00000426, -0.00000259, -0.00000289, -0.00000109, 0.00000473, -0.00000205, 0.00000322, -0.00000138, -0.00000368, -0.00000057, 0.00000128, 0.00000041, 0.00000382, -0.00000632, 0.00000313, -0.00000537, -0.00000276, 0.00000254, 0.00000166, -0.00000134, 0.00000147, },
		{ 0.00000136, -0.00000463, 0.00000046, -0.00000467, 0.00001150, -0.00000059, -0.00000025, -0.00000460, 0.00000304, -0.00000334, -0.00000589, 0.00001130, 0.00001025, -0.00001612, 0.00001262, -0.00001523, 0.00000317, -0.00000193, -0.00002496, 0.00000947, -0.00000154, -0.00000820, -0.00000811, -0.00001484, 0.00000323, -0.00002737, -0.00000968, -0.00004625, 0.00001883, -0.00017209, -0.00011437, 0.00025998, 0.00015343, 0.00011739, 0.00032528, -0.00024909, -0.00005493, -0.00003947, 0.00000161, -0.00002867, -0.00000076, -0.00001233, 0.00002238, -0.00002268, 0.00001047, -0.00000185, -0.00000823, 0.00001550, -0.00003265, 0.00000875, 0.00000992, -0.00002897, 0.00002156, -0.00003432, 0.00003594, -0.00004489, 0.00004645, -0.00003520, 0.00003004, -0.00004213, 0.00002386, -0.00000243, 0.00000038, -0.00001109, 0.00000085, 0.00000740, -0.00000646, 0.00001239, -0.00002291, 0.00002565, -0.00002966, 0.00002941, -0.00002157, 0.00001918, -0.00001410, 0.00001297, -0.00001832, -0.00000110, 0.00000466, 0.00000372, -0.00000248, 0.00000661, -0.00000313, 0.00000099, -0.00000422, 0.00000082, -0.00000196, 0.00000122, -0.00000077, 0.00000019, -0.00000377, -0.00000146, 0.00000186, 0.00000378, -0.00000000, 0.00000779, -0.00000505, -0.00000698, 0.00000226, -0.00000032, 0.00000000, -0.00000136, -0.00000032, 0.00000030, -0.00000133, 0.00000075, 0.00000390, -0.00000334, -0.00000271, -0.00000026, 0.00000446, -0.00000244, 0.00000396, -0.00000226, -0.00000347, -0.00000054, 0.00000136, 0.00000106, 0.00000277, -0.00000607, 0.00000338, -0.00000610, -0.00000174, 0.00000210, 0.00000183, -0.00000123, 0.00000121, -0.00000360, },
		{ -0.00000378, -0.00000015, -0.00000419, 0.00001267, -0.00000209, 0.00000010, -0.00000484, 0.00000355, -0.00000406, -0.00000510, 0.00001265, 0.00000712, -0.00001420, 0.00001227, -0.00001738, 0.00000737, -0.00000654, -0.00002251, 0.00000942, -0.00000234, -0.00000681, -0.00001094, -0.00001114, -0.00000124, -0.00002227, -0.00001555, -0.00004349, 0.00001788, -0.00019266, -0.00006044, 0.00025555, 0.00012238, 0.00012013, 0.00030668, -0.00022728, -0.00006236, -0.00003483, 0.00000045, -0.00002778, 0.00000077, -0.00001355, 0.00002398, -0.00002331, 0.00001417, -0.00000787, -0.00000155, 0.00001019, -0.00003378, 0.00001571, -0.00000050, -0.00001813, 0.00001151, -0.00002579, 0.00002852, -0.00003922, 0.00004473, -0.00003664, 0.00003093, -0.00004456, 0.00003062, -0.00000934, 0.00000414, -0.00001502, 0.00000675, 0.00000148, -0.00000075, 0.00000663, -0.00001763, 0.00002133, -0.00002701, 0.00002944, -0.00002291, 0.00002086, -0.00001607, 0.00001523, -0.00002254, 0.00000318, 0.00000340, 0.00000350, -0.00000196, 0.00000623, -0.00000335, 0.00000114, -0.00000445, 0.00000101, -0.00000198, 0.00000134, -0.00000075, -0.00000030, -0.00000381, -0.00000111, 0.00000268, 0.00000299, 0.00000079, 0.00000754, -0.00000687, -0.00000568, 0.00000254, -0.00000063, 0.00000013, -0.00000164, 0.00000005, 0.00000000, -0.00000137, 0.00000153, 0.00000340, -0.00000403, -0.00000241, 0.00000059, 0.00000396, -0.00000256, 0.00000456, -0.00000324, -0.00000305, -0.00000068, 0.00000165, 0.00000150, 0.00000170, -0.00000558, 0.00000335, -0.00000661, -0.00000074, 0.00000156, 0.00000212, -0.00000126, 0.00000110, -0.00000352, 0.00000059, },
		{ -0.00000090, -0.00000336, 0.00001345, -0.00000351, 0.00000053, -0.00000510, 0.00000407, -0.00000495, -0.00000377, 0.00001334, 0.00000411, -0.00001174, 0.00001090, -0.00001828, 0.00001071, -0.00001129, -0.00001910, 0.00000827, -0.00000213, -0.00000645, -0.00001278, -0.00000813, -0.00000545, -0.00001694, -0.00002269, -0.00003780, 0.00001123, -0.00020544, -0.00000665, 0.00024068, 0.00009860, 0.00012183, 0.00028887, -0.00020561, -0.00007063, -0.00002892, -0.00000214, -0.00002552, 0.00000107, -0.00001361, 0.00002447, -0.00002272, 0.00001675, -0.00001340, 0.00000576, 0.00000238, -0.00003116, 0.00001902, -0.00000825, -0.00000879, 0.00000201, -0.00001705, 0.00002018, -0.00003186, 0.00004089, -0.00003617, 0.00003004, -0.00004491, 0.00003572, -0.00001561, 0.00000757, -0.00001829, 0.00001202, -0.00000409, 0.00000482, 0.00000070, -0.00001175, 0.00001598, -0.00002295, 0.00002800, -0.00002307, 0.00002165, -0.00001734, 0.00001681, -0.00002640, 0.00000790, 0.00000129, 0.00000395, -0.00000180, 0.00000599, -0.00000364, 0.00000132, -0.00000469, 0.00000124, -0.00000204, 0.00000152, -0.00000083, -0.00000076, -0.00000384, -0.00000061, 0.00000336, 0.00000214, 0.00000182, 0.00000683, -0.00000839, -0.00000420, 0.00000256, -0.00000083, 0.00000018, -0.00000186, 0.00000042, -0.00000040, -0.00000124, 0.00000224, 0.00000277, -0.00000467, -0.00000196, 0.00000139, 0.00000326, -0.00000239, 0.00000494, -0.00000422, -0.00000250, -0.00000096, 0.00000214, 0.00000169, 0.00000071, -0.00000495, 0.00000308, -0.00000690, 0.00000018, 0.00000098, 0.00000248, -0.00000137, 0.00000112, -0.00000367, -0.00000002, -0.00000291, },
		{ -0.00000226, 0.00001381, -0.00000477, 0.00000099, -0.00000534, 0.00000454, -0.00000592, -0.00000198, 0.00001326, 0.00000149, -0.00000905, 0.00000873, -0.00001789, 0.00001289, -0.00001571, -0.00001513, 0.00000625, -0.00000105, -0.00000712, -0.00001346, -0.00000613, -0.00000887, -0.00001208, -0.00003022, -0.00003011, -0.00000058, -0.00020936, 0.00004404, 0.00021776, 0.00008199, 0.00012153, 0.00027276, -0.00018496, -0.00007895, -0.00002236, -0.00000572, -0.00002216, 0.00000030, -0.00001255, 0.00002377, -0.00002068, 0.00001776, -0.00001768, 0.00001263, -0.00000687, -0.00002544, 0.00001880, -0.00001299, -0.00000158, -0.00000621, -0.00000883, 0.00001165, -0.00002346, 0.00003535, -0.00003403, 0.00002750, -0.00004318, 0.00003892, -0.00002095, 0.00001049, -0.00002077, 0.00001649, -0.00000905, 0.00000994, -0.00000504, -0.00000565, 0.00000993, -0.00001771, 0.00002516, -0.00002198, 0.00002141, -0.00001773, 0.00001748, -0.00002955, 0.00001271, -0.00000148, 0.00000505, -0.00000206, 0.00000595, -0.00000404, 0.00000160, -0.00000497, 0.00000150, -0.00000212, 0.00000173, -0.00000100, -0.00000117, -0.00000386, 0.00000005, 0.00000382, 0.00000132, 0.00000301, 0.00000566, -0.00000953, -0.00000266, 0.00000238, -0.00000093, 0.00000014, -0.00000200, 0.00000076, -0.00000086, -0.00000094, 0.00000282, 0.00000204, -0.00000523, -0.00000135, 0.00000206, 0.00000244, -0.00000192, 0.00000504, -0.00000510, -0.00000186, -0.00000134, 0.00000281, 0.00000160, -0.00000012, -0.00000426, 0.00000261, -0.00000695, 0.00000099, 0.00000039, 0.00000288, -0.00000152, 0.00000123, -0.00000404, -0.00000039, -0.00000210, -0.00000171, },
		{ 0.00001375, -0.00000581, 0.00000141, -0.00000549, 0.00000488, -0.00000685, 0.00000020, 0.00001237, -0.00000053, -0.00000641, 0.00000605, -0.00001636, 0.00001374, -0.00001940, -0.00001100, 0.00000367, 0.00000068, -0.00000869, -0.00001299, -0.00000531, -0.00001117, -0.00000829, -0.00003722, -0.00002158, -0.00001653, -0.00020403, 0.00008909, 0.00018940, 0.00007186, 0.00011875, 0.00025895, -0.00016601, -0.00008663, -0.00001575, -0.00000977, -0.00001810, -0.00000126, -0.00001058, 0.00002203, -0.00001719, 0.00001694, -0.00002014, 0.00001813, -0.00001641, -0.00001754, 0.00001548, -0.00001468, 0.00000316, -0.00001261, -0.00000179, 0.00000364, -0.00001469, 0.00002863, -0.00003058, 0.00002358, -0.00003950, 0.00004010, -0.00002512, 0.00001276, -0.00002238, 0.00002000, -0.00001323, 0.00001436, -0.00001027, 0.00000030, 0.00000358, -0.00001160, 0.00002110, -0.00001968, 0.00002009, -0.00001714, 0.00001704, -0.00003167, 0.00001723, -0.00000467, 0.00000671, -0.00000276, 0.00000615, -0.00000461, 0.00000201, -0.00000533, 0.00000183, -0.00000225, 0.00000198, -0.00000128, -0.00000152, -0.00000386, 0.00000087, 0.00000401, 0.00000064, 0.00000426, 0.00000407, -0.00001020, -0.00000115, 0.00000203, -0.00000094, 0.00000003, -0.00000204, 0.00000103, -0.00000135, -0.00000048, 0.00000325, 0.00000123, -0.00000571, -0.00000056, 0.00000256, 0.00000159, -0.00000121, 0.00000481, -0.00000582, -0.00000123, -0.00000175, 0.00000362, 0.00000122, -0.00000072, -0.00000360, 0.00000202, -0.00000678, 0.00000167, -0.00000018, 0.00000327, -0.00000167, 0.00000135, -0.00000460, -0.00000048, -0.00000143, -0.00000247, -0.00000094, },
		{ -0.00000660, 0.00000177, -0.00000554, 0.00000502, -0.00000762, 0.00000259, 0.00001072, -0.00000181, -0.00000412, 0.00000318, -0.00001390, 0.00001324, -0.00002210, -0.00000707, 0.00000086, 0.00000280, -0.00001091, -0.00001151, -0.00000569, -0.00001214, -0.00000602, -0.00004289, -0.00001347, -0.00003517, -0.00018979, 0.00012656, 0.00015822, 0.00006709, 0.00011352, 0.00024769, -0.00014920, -0.00009312, -0.00000965, -0.00001379, -0.00001378, -0.00000321, -0.00000803, 0.00001953, -0.00001248, 0.00001433, -0.00002049, 0.00002151, -0.00002516, -0.00000851, 0.00000975, -0.00001356, 0.00000531, -0.00001687, 0.00000357, -0.00000323, -0.00000622, 0.00002133, -0.00002625, 0.00001867, -0.00003414, 0.00003926, -0.00002797, 0.00001428, -0.00002308, 0.00002249, -0.00001648, 0.00001790, -0.00001472, 0.00000573, -0.00000267, -0.00000501, 0.00001607, -0.00001630, 0.00001775, -0.00001556, 0.00001542, -0.00003252, 0.00002111, -0.00000798, 0.00000875, -0.00000384, 0.00000660, -0.00000535, 0.00000257, -0.00000579, 0.00000225, -0.00000243, 0.00000226, -0.00000165, -0.00000181, -0.00000381, 0.00000181, 0.00000390, 0.00000019, 0.00000544, 0.00000215, -0.00001037, 0.00000023, 0.00000158, -0.00000088, -0.00000016, -0.00000197, 0.00000121, -0.00000181, 0.00000011, 0.00000352, 0.00000037, -0.00000609, 0.00000038, 0.00000281, 0.00000078, -0.00000031, 0.00000424, -0.00000630, -0.00000067, -0.00000213, 0.00000448, 0.00000059, -0.00000107, -0.00000303, 0.00000136, -0.00000643, 0.00000221, -0.00000071, 0.00000366, -0.00000180, 0.00000144, -0.00000529, -0.00000031, -0.00000097, -0.00000312, 0.00000049, 0.00001332, },
		{ 0.00000203, -0.00000545, 0.00000493, -0.00000811, 0.00000503, 0.00000842, -0.00000228, -0.00000240, 0.00000047, -0.00001085, 0.00001148, -0.00002362, -0.00000366, -0.00000189, 0.00000499, -0.00001349, -0.00000925, -0.00000713, -0.00001179, -0.00000551, -0.00004658, -0.00000696, -0.00005481, -0.00016761, 0.00015516, 0.00012663, 0.00006630, 0.00010619, 0.00023894, -0.00013472, -0.00009806, -0.00000448, -0.00001735, -0.00000962, -0.00000517, -0.00000529, 0.00001668, -0.00000692, 0.00001016, -0.00001875, 0.00002231, -0.00003219, 0.00000060, 0.00000248, -0.00001016, 0.00000505, -0.00001887, 0.00000696, -0.00000849, 0.00000135, 0.00001401, -0.00002154, 0.00001324, -0.00002748, 0.00003656, -0.00002946, 0.00001501, -0.00002287, 0.00002393, -0.00001874, 0.00002043, -0.00001818, 0.00001034, -0.00000845, 0.00000168, 0.00001042, -0.00001206, 0.00001451, -0.00001309, 0.00001262, -0.00003192, 0.00002402, -0.00001107, 0.00001098, -0.00000520, 0.00000726, -0.00000626, 0.00000330, -0.00000637, 0.00000278, -0.00000268, 0.00000258, -0.00000212, -0.00000204, -0.00000369, 0.00000284, 0.00000350, 0.00000003, 0.00000643, 0.00000000, -0.00001004, 0.00000140, 0.00000110, -0.00000079, -0.00000038, -0.00000182, 0.00000127, -0.00000221, 0.00000081, 0.00000361, -0.00000052, -0.00000633, 0.00000144, 0.00000278, 0.00000013, 0.00000068, 0.00000338, -0.00000650, -0.00000024, -0.00000240, 0.00000532, -0.00000023, -0.00000116, -0.00000261, 0.00000071, -0.00000595, 0.00000263, -0.00000119, 0.00000403, -0.00000189, 0.00000143, -0.00000604, 0.00000012, -0.00000075, -0.00000356, 0.00000194, 0.00001255, -0.00000712, },
		{ -0.00000524, 0.00000457, -0.00000821, 0.00000732, 0.00000563, -0.00000196, -0.00000140, -0.00000177, -0.00000757, 0.00000869, -0.00002393, -0.00000097, -0.00000431, 0.00000697, -0.00001609, -0.00000655, -0.00000935, -0.00001029, -0.00000676, -0.00004794, -0.00000305, -0.00007365, -0.00013900, 0.00017435, 0.00009665, 0.00006804, 0.00009736, 0.00023239, -0.00012250, -0.00010131, -0.00000049, -0.00002013, -0.00000597, -0.00000677, -0.00000275, 0.00001393, -0.00000100, 0.00000489, -0.00001518, 0.00002040, -0.00003680, 0.00000880, -0.00000542, -0.00000514, 0.00000274, -0.00001875, 0.00000829, -0.00001181, 0.00000752, 0.00000720, -0.00001691, 0.00000777, -0.00002000, 0.00003224, -0.00002960, 0.00001495, -0.00002180, 0.00002435, -0.00002001, 0.00002192, -0.00002053, 0.00001391, -0.00001340, 0.00000805, 0.00000449, -0.00000724, 0.00001058, -0.00000988, 0.00000875, -0.00002981, 0.00002569, -0.00001364, 0.00001315, -0.00000671, 0.00000806, -0.00000730, 0.00000418, -0.00000708, 0.00000342, -0.00000301, 0.00000294, -0.00000268, -0.00000218, -0.00000350, 0.00000389, 0.00000283, 0.00000022, 0.00000711, -0.00000223, -0.00000925, 0.00000230, 0.00000062, -0.00000069, -0.00000062, -0.00000159, 0.00000123, -0.00000252, 0.00000156, 0.00000353, -0.00000143, -0.00000641, 0.00000259, 0.00000246, -0.00000028, 0.00000164, 0.00000228, -0.00000643, 0.00000002, -0.00000250, 0.00000605, -0.00000116, -0.00000101, -0.00000237, 0.00000011, -0.00000537, 0.00000293, -0.00000164, 0.00000439, -0.00000195, 0.00000128, -0.00000676, 0.00000076, -0.00000079, -0.00000374, 0.00000331, 0.00001153, -0.00000739, 0.00000218, },
		{ 0.00000396, -0.00000786, 0.00000928, 0.00000256, -0.00000094, -0.00000118, -0.00000330, -0.00000444, 0.00000518, -0.00002311, 0.00000090, -0.00000624, 0.00000851, -0.00001843, -0.00000375, -0.00001201, -0.00000797, -0.00000956, -0.00004687, -0.00000246, -0.00008998, -0.00010588, 0.00018425, 0.00006986, 0.00007093, 0.00008778, 0.00022755, -0.00011230, -0.00010290, 0.00000222, -0.00002193, -0.00000306, -0.00000773, -0.00000074, 0.00001171, 0.00000474, -0.00000093, -0.00001028, 0.00001593, -0.00003860, 0.00001528, -0.00001302, 0.00000073, -0.00000105, -0.00001684, 0.00000767, -0.00001310, 0.00001198, 0.00000132, -0.00001280, 0.00000276, -0.00001222, 0.00002666, -0.00002852, 0.00001418, -0.00001999, 0.00002385, -0.00002034, 0.00002240, -0.00002175, 0.00001629, -0.00001727, 0.00001375, -0.00000133, -0.00000215, 0.00000620, -0.00000616, 0.00000402, -0.00002625, 0.00002598, -0.00001541, 0.00001502, -0.00000819, 0.00000889, -0.00000837, 0.00000516, -0.00000790, 0.00000419, -0.00000344, 0.00000334, -0.00000334, -0.00000224, -0.00000321, 0.00000491, 0.00000193, 0.00000077, 0.00000739, -0.00000442, -0.00000807, 0.00000290, 0.00000020, -0.00000063, -0.00000085, -0.00000131, 0.00000107, -0.00000271, 0.00000233, 0.00000326, -0.00000233, -0.00000631, 0.00000376, 0.00000186, -0.00000040, 0.00000247, 0.00000103, -0.00000610, 0.00000009, -0.00000241, 0.00000660, -0.00000212, -0.00000068, -0.00000232, -0.00000038, -0.00000474, 0.00000314, -0.00000206, 0.00000477, -0.00000199, 0.00000098, -0.00000738, 0.00000154, -0.00000107, -0.00000361, 0.00000451, 0.00001035, -0.00000744, 0.00000224, -0.00000492, },
		{ -0.00000701, 0.00001073, -0.00000055, 0.00000059, -0.00000169, -0.00000397, -0.00000180, 0.00000130, -0.00002134, 0.00000194, -0.00000758, 0.00000946, -0.00002025, -0.00000117, -0.00001474, -0.00000525, -0.00001351, -0.00004358, -0.00000555, -0.00010229, -0.00007036, 0.00018557, 0.00004729, 0.00007378, 0.00007820, 0.00022385, -0.00010372, -0.00010304, 0.00000371, -0.00002273, -0.00000100, -0.00000789, 0.00000050, 0.00001038, 0.00000982, -0.00000670, -0.00000472, 0.00000937, -0.00003754, 0.00001950, -0.00001954, 0.00000665, -0.00000569, -0.00001359, 0.00000540, -0.00001244, 0.00001455, -0.00000331, -0.00000955, -0.00000138, -0.00000467, 0.00002024, -0.00002642, 0.00001280, -0.00001755, 0.00002254, -0.00001986, 0.00002195, -0.00002187, 0.00001745, -0.00001987, 0.00001848, -0.00000670, 0.00000288, 0.00000167, -0.00000223, -0.00000128, -0.00002140, 0.00002481, -0.00001618, 0.00001635, -0.00000945, 0.00000961, -0.00000940, 0.00000617, -0.00000878, 0.00000506, -0.00000397, 0.00000378, -0.00000408, -0.00000220, -0.00000281, 0.00000583, 0.00000089, 0.00000166, 0.00000718, -0.00000640, -0.00000659, 0.00000322, -0.00000014, -0.00000061, -0.00000103, -0.00000101, 0.00000082, -0.00000276, 0.00000309, 0.00000283, -0.00000320, -0.00000599, 0.00000488, 0.00000102, -0.00000017, 0.00000307, -0.00000027, -0.00000557, -0.00000001, -0.00000210, 0.00000691, -0.00000301, -0.00000023, -0.00000242, -0.00000076, -0.00000408, 0.00000327, -0.00000246, 0.00000518, -0.00000204, 0.00000053, -0.00000784, 0.00000240, -0.00000157, -0.00000316, 0.00000546, 0.00000908, -0.00000731, 0.00000223, -0.00000452, 0.00000312, },
		{ 0.00001155, -0.00000347, 0.00000243, -0.00000282, -0.00000373, 0.00000008, -0.00000260, -0.00001887, 0.00000225, -0.00000834, 0.00000978, -0.00002139, 0.00000091, -0.00001717, -0.00000258, -0.00001807, -0.00003855, -0.00001227, -0.00010945, -0.00003459, 0.00017954, 0.00002945, 0.00007569, 0.00006927, 0.00022071, -0.00009630, -0.00010203, 0.00000417, -0.00002260, 0.00000023, -0.00000722, 0.00000089, 0.00001015, 0.00001381, -0.00001186, 0.00000077, 0.00000139, -0.00003384, 0.00002117, -0.00002438, 0.00001193, -0.00001051, -0.00000959, 0.00000192, -0.00001007, 0.00001526, -0.00000654, -0.00000737, -0.00000431, 0.00000216, 0.00001342, -0.00002355, 0.00001097, -0.00001467, 0.00002059, -0.00001869, 0.00002074, -0.00002103, 0.00001742, -0.00002112, 0.00002201, -0.00001132, 0.00000754, -0.00000272, 0.00000161, -0.00000680, -0.00001553, 0.00002224, -0.00001584, 0.00001696, -0.00001033, 0.00001009, -0.00001026, 0.00000712, -0.00000966, 0.00000600, -0.00000458, 0.00000425, -0.00000492, -0.00000203, -0.00000231, 0.00000659, -0.00000022, 0.00000283, 0.00000647, -0.00000806, -0.00000493, 0.00000326, -0.00000037, -0.00000065, -0.00000114, -0.00000073, 0.00000050, -0.00000266, 0.00000379, 0.00000224, -0.00000402, -0.00000546, 0.00000587, 0.00000001, 0.00000041, 0.00000334, -0.00000151, -0.00000491, -0.00000023, -0.00000160, 0.00000695, -0.00000377, 0.00000027, -0.00000265, -0.00000102, -0.00000341, 0.00000333, -0.00000283, 0.00000563, -0.00000216, -0.00000005, -0.00000808, 0.00000324, -0.00000223, -0.00000239, 0.00000609, 0.00000782, -0.00000705, 0.00000219, -0.00000411, 0.00000214, -0.00000570, },
		{ -0.00000601, 0.00000434, -0.00000435, -0.00000264, 0.00000103, -0.00000617, -0.00001597, 0.00000198, -0.00000862, 0.00000948, -0.00002179, 0.00000232, -0.00001903, -0.00000038, -0.00002266, -0.00003246, -0.00002222, -0.00011073, -0.00000059, 0.00016768, 0.00001639, 0.00007603, 0.00006154, 0.00021759, -0.00008958, -0.00010025, 0.00000386, -0.00002173, 0.00000076, -0.00000581, 0.00000046, 0.00001112, 0.00001646, -0.00001593, 0.00000548, -0.00000721, -0.00002800, 0.00002032, -0.00002717, 0.00001598, -0.00001491, -0.00000541, -0.00000226, -0.00000639, 0.00001426, -0.00000835, -0.00000635, -0.00000580, 0.00000785, 0.00000666, -0.00002019, 0.00000888, -0.00001152, 0.00001818, -0.00001702, 0.00001894, -0.00001939, 0.00001634, -0.00002105, 0.00002422, -0.00001497, 0.00001158, -0.00000671, 0.00000506, -0.00001217, -0.00000901, 0.00001844, -0.00001434, 0.00001673, -0.00001068, 0.00001020, -0.00001084, 0.00000790, -0.00001047, 0.00000694, -0.00000524, 0.00000472, -0.00000582, -0.00000174, -0.00000173, 0.00000714, -0.00000130, 0.00000421, 0.00000524, -0.00000928, -0.00000321, 0.00000308, -0.00000052, -0.00000076, -0.00000118, -0.00000050, 0.00000013, -0.00000241, 0.00000440, 0.00000151, -0.00000475, -0.00000470, 0.00000666, -0.00000108, 0.00000129, 0.00000323, -0.00000258, -0.00000420, -0.00000052, -0.00000094, 0.00000670, -0.00000433, 0.00000075, -0.00000296, -0.00000117, -0.00000274, 0.00000330, -0.00000316, 0.00000613, -0.00000238, -0.00000073, -0.00000809, 0.00000400, -0.00000297, -0.00000134, 0.00000637, 0.00000664, -0.00000672, 0.00000216, -0.00000375, 0.00000110, -0.00000397, 0.00001168, },
		{ 0.00000609, -0.00000606, -0.00000086, 0.00000098, -0.00000916, -0.00001292, 0.00000131, -0.00000855, 0.00000868, -0.00002149, 0.00000300, -0.00002013, 0.00000099, -0.00002669, -0.00002611, -0.00003461, -0.00010594, 0.00002994, 0.00015168, 0.00000775, 0.00007453, 0.00005531, 0.00021410, -0.00008313, -0.00009804, 0.00000309, -0.00002036, 0.00000080, -0.00000383, -0.00000065, 0.00001322, 0.00001765, -0.00001864, 0.00000879, -0.00001557, -0.00002073, 0.00001723, -0.00002778, 0.00001844, -0.00001836, -0.00000161, -0.00000656, -0.00000189, 0.00001187, -0.00000888, -0.00000644, -0.00000579, 0.00001208, 0.00000035, -0.00001663, 0.00000671, -0.00000830, 0.00001549, -0.00001502, 0.00001675, -0.00001719, 0.00001441, -0.00001978, 0.00002509, -0.00001749, 0.00001479, -0.00001009, 0.00000787, -0.00001702, -0.00000222, 0.00001368, -0.00001177, 0.00001561, -0.00001041, 0.00000985, -0.00001103, 0.00000842, -0.00001109, 0.00000782, -0.00000591, 0.00000519, -0.00000677, -0.00000129, -0.00000109, 0.00000745, -0.00000225, 0.00000567, 0.00000354, -0.00000999, -0.00000154, 0.00000272, -0.00000058, -0.00000092, -0.00000113, -0.00000033, -0.00000024, -0.00000203, 0.00000489, 0.00000066, -0.00000537, -0.00000373, 0.00000719, -0.00000214, 0.00000243, 0.00000273, -0.00000340, -0.00000354, -0.00000081, -0.00000020, 0.00000621, -0.00000467, 0.00000116, -0.00000330, -0.00000123, -0.00000207, 0.00000318, -0.00000343, 0.00000668, -0.00000273, -0.00000144, -0.00000784, 0.00000462, -0.00000373, -0.00000006, 0.00000630, 0.00000561, -0.00000637, 0.00000216, -0.00000349, 0.00000010, -0.00000196, 0.00001112, -0.00000801, },
		{ -0.00000767, 0.00000138, -0.00000005, -0.00001138, -0.00000997, 0.00000044, -0.00000830, 0.00000756, -0.00002061, 0.00000298, -0.00002041, 0.00000133, -0.00002967, -0.00002032, -0.00004844, -0.00009532, 0.00005567, 0.00013325, 0.00000287, 0.00007118, 0.00005073, 0.00020996, -0.00007663, -0.00009573, 0.00000215, -0.00001875, 0.00000058, -0.00000154, -0.00000216, 0.00001625, 0.00001746, -0.00001987, 0.00001026, -0.00002290, -0.00001282, 0.00001240, -0.00002636, 0.00001912, -0.00002052, 0.00000136, -0.00001049, 0.00000293, 0.00000847, -0.00000836, -0.00000746, -0.00000432, 0.00001468, -0.00000514, -0.00001315, 0.00000467, -0.00000519, 0.00001270, -0.00001287, 0.00001439, -0.00001466, 0.00001189, -0.00001750, 0.00002470, -0.00001886, 0.00001705, -0.00001271, 0.00000983, -0.00002100, 0.00000441, 0.00000828, -0.00000829, 0.00001361, -0.00000947, 0.00000897, -0.00001075, 0.00000857, -0.00001146, 0.00000856, -0.00000654, 0.00000559, -0.00000773, -0.00000070, -0.00000042, 0.00000749, -0.00000297, 0.00000710, 0.00000146, -0.00001014, -0.00000003, 0.00000227, -0.00000059, -0.00000112, -0.00000100, -0.00000024, -0.00000059, -0.00000153, 0.00000525, -0.00000030, -0.00000584, -0.00000258, 0.00000739, -0.00000307, 0.00000371, 0.00000184, -0.00000391, -0.00000300, -0.00000101, 0.00000055, 0.00000552, -0.00000479, 0.00000144, -0.00000363, -0.00000123, -0.00000139, 0.00000296, -0.00000361, 0.00000724, -0.00000325, -0.00000213, -0.00000736, 0.00000504, -0.00000443, 0.00000138, 0.00000587, 0.00000477, -0.00000604, 0.00000224, -0.00000339, -0.00000072, 0.00000022, 0.00000991, -0.00000939, 0.00000750, },
		{ 0.00000381, -0.00000190, -0.00001274, -0.00000732, -0.00000045, -0.00000803, 0.00000629, -0.00001933, 0.00000239, -0.00001994, 0.00000055, -0.00003128, -0.00001586, -0.00006254, -0.00007960, 0.00007572, 0.00011397, 0.00000096, 0.00006617, 0.00004774, 0.00020507, -0.00006987, -0.00009356, 0.00000128, -0.00001714, 0.00000036, 0.00000081, -0.00000378, 0.00001991, 0.00001609, -0.00001971, 0.00000968, -0.00002854, -0.00000510, 0.00000650, -0.00002322, 0.00001807, -0.00002122, 0.00000319, -0.00001358, 0.00000754, 0.00000450, -0.00000710, -0.00000915, -0.00000161, 0.00001558, -0.00000958, -0.00000998, 0.00000290, -0.00000236, 0.00000998, -0.00001074, 0.00001203, -0.00001205, 0.00000905, -0.00001447, 0.00002319, -0.00001909, 0.00001833, -0.00001449, 0.00001082, -0.00002385, 0.00001046, 0.00000263, -0.00000414, 0.00001085, -0.00000788, 0.00000753, -0.00000994, 0.00000829, -0.00001146, 0.00000907, -0.00000707, 0.00000590, -0.00000867, 0.00000005, 0.00000023, 0.00000729, -0.00000341, 0.00000837, -0.00000088, -0.00000974, 0.00000124, 0.00000177, -0.00000058, -0.00000134, -0.00000080, -0.00000025, -0.00000088, -0.00000094, 0.00000544, -0.00000133, -0.00000613, -0.00000129, 0.00000725, -0.00000377, 0.00000503, 0.00000063, -0.00000409, -0.00000263, -0.00000106, 0.00000121, 0.00000471, -0.00000470, 0.00000159, -0.00000392, -0.00000117, -0.00000071, 0.00000261, -0.00000366, 0.00000777, -0.00000394, -0.00000275, -0.00000668, 0.00000523, -0.00000499, 0.00000290, 0.00000512, 0.00000413, -0.00000575, 0.00000240, -0.00000349, -0.00000128, 0.00000237, 0.00000819, -0.00001011, 0.00000843, -0.00000896, },
		{ -0.00000437, -0.00001326, -0.00000508, -0.00000125, -0.00000789, 0.00000508, -0.00001786, 0.00000143, -0.00001889, -0.00000126, -0.00003136, -0.00001332, -0.00007569, -0.00005982, 0.00008972, 0.00009518, 0.00000112, 0.00005988, 0.00004614, 0.00019946, -0.00006276, -0.00009167, 0.00000067, -0.00001570, 0.00000032, 0.00000296, -0.00000518, 0.00002385, 0.00001386, -0.00001846, 0.00000706, -0.00003202, 0.00000169, 0.00000022, -0.00001888, 0.00001551, -0.00002046, 0.00000370, -0.00001552, 0.00001152, 0.00000039, -0.00000544, -0.00001118, 0.00000205, 0.00001487, -0.00001281, -0.00000731, 0.00000156, 0.00000005, 0.00000747, -0.00000874, 0.00000984, -0.00000955, 0.00000616, -0.00001097, 0.00002080, -0.00001831, 0.00001866, -0.00001545, 0.00001081, -0.00002538, 0.00001557, -0.00000288, 0.00000039, 0.00000749, -0.00000572, 0.00000557, -0.00000861, 0.00000752, -0.00001105, 0.00000928, -0.00000743, 0.00000605, -0.00000953, 0.00000095, 0.00000083, 0.00000685, -0.00000352, 0.00000935, -0.00000336, -0.00000883, 0.00000220, 0.00000128, -0.00000057, -0.00000154, -0.00000057, -0.00000034, -0.00000110, -0.00000028, 0.00000546, -0.00000243, -0.00000619, 0.00000006, 0.00000677, -0.00000415, 0.00000626, -0.00000082, -0.00000394, -0.00000247, -0.00000091, 0.00000171, 0.00000386, -0.00000446, 0.00000161, -0.00000415, -0.00000106, -0.00000003, 0.00000215, -0.00000353, 0.00000821, -0.00000479, -0.00000322, -0.00000585, 0.00000517, -0.00000536, 0.00000441, 0.00000411, 0.00000372, -0.00000552, 0.00000264, -0.00000381, -0.00000148, 0.00000434, 0.00000610, -0.00001024, 0.00000882, -0.00000973, 0.00000612, },
		{ -0.00001301, -0.00000330, -0.00000187, -0.00000798, 0.00000407, -0.00001640, 0.00000031, -0.00001750, -0.00000388, -0.00002997, -0.00001309, -0.00008678, -0.00003729, 0.00009774, 0.00007793, 0.00000251, 0.00005274, 0.00004563, 0.00019326, -0.00005532, -0.00009012, 0.00000039, -0.00001456, 0.00000061, 0.00000472, -0.00000608, 0.00002768, 0.00001114, -0.00001651, 0.00000265, -0.00003313, 0.00000694, -0.00000573, -0.00001390, 0.00001183, -0.00001844, 0.00000290, -0.00001613, 0.00001451, -0.00000347, -0.00000372, -0.00001317, 0.00000628, 0.00001275, -0.00001478, -0.00000526, 0.00000074, 0.00000193, 0.00000529, -0.00000700, 0.00000794, -0.00000736, 0.00000347, -0.00000730, 0.00001777, -0.00001670, 0.00001814, -0.00001566, 0.00000985, -0.00002551, 0.00001947, -0.00000788, 0.00000497, 0.00000377, -0.00000314, 0.00000318, -0.00000679, 0.00000626, -0.00001017, 0.00000912, -0.00000755, 0.00000599, -0.00001025, 0.00000195, 0.00000133, 0.00000623, -0.00000328, 0.00000995, -0.00000581, -0.00000748, 0.00000285, 0.00000085, -0.00000059, -0.00000171, -0.00000031, -0.00000052, -0.00000122, 0.00000042, 0.00000530, -0.00000355, -0.00000600, 0.00000141, 0.00000598, -0.00000415, 0.00000727, -0.00000240, -0.00000351, -0.00000253, -0.00000053, 0.00000197, 0.00000307, -0.00000412, 0.00000153, -0.00000431, -0.00000091, 0.00000063, 0.00000156, -0.00000321, 0.00000851, -0.00000577, -0.00000351, -0.00000491, 0.00000488, -0.00000548, 0.00000583, 0.00000290, 0.00000351, -0.00000534, 0.00000295, -0.00000433, -0.00000127, 0.00000596, 0.00000383, -0.00000988, 0.00000867, -0.00000984, 0.00000806, -0.00000720, },
		{ -0.00000197, -0.00000230, -0.00000831, 0.00000338, -0.00001511, -0.00000073, -0.00001605, -0.00000698, -0.00002734, -0.00001529, -0.00009486, -0.00001342, 0.00010024, 0.00006292, 0.00000440, 0.00004519, 0.00004585, 0.00018671, -0.00004765, -0.00008886, 0.00000047, -0.00001373, 0.00000130, 0.00000599, -0.00000624, 0.00003109, 0.00000831, -0.00001438, -0.00000310, -0.00003191, 0.00001025, -0.00001074, -0.00000889, 0.00000750, -0.00001545, 0.00000097, -0.00001540, 0.00001631, -0.00000678, -0.00000224, -0.00001479, 0.00001065, 0.00000953, -0.00001557, -0.00000386, 0.00000047, 0.00000322, 0.00000352, -0.00000556, 0.00000640, -0.00000559, 0.00000120, -0.00000375, 0.00001440, -0.00001447, 0.00001694, -0.00001525, 0.00000810, -0.00002428, 0.00002196, -0.00001205, 0.00000928, -0.00000007, -0.00000030, 0.00000048, -0.00000456, 0.00000456, -0.00000883, 0.00000854, -0.00000739, 0.00000566, -0.00001075, 0.00000303, 0.00000169, 0.00000550, -0.00000271, 0.00001007, -0.00000808, -0.00000580, 0.00000316, 0.00000050, -0.00000065, -0.00000182, -0.00000007, -0.00000075, -0.00000124, 0.00000113, 0.00000496, -0.00000465, -0.00000554, 0.00000267, 0.00000493, -0.00000374, 0.00000795, -0.00000398, -0.00000287, -0.00000275, 0.00000008, 0.00000196, 0.00000241, -0.00000377, 0.00000138, -0.00000444, -0.00000070, 0.00000124, 0.00000089, -0.00000267, 0.00000861, -0.00000682, -0.00000359, -0.00000394, 0.00000437, -0.00000533, 0.00000708, 0.00000157, 0.00000347, -0.00000521, 0.00000328, -0.00000500, -0.00000063, 0.00000711, 0.00000159, -0.00000919, 0.00000808, -0.00000928, 0.00000939, -0.00001010, -0.00001214, },
		{ -0.00000258, -0.00000888, 0.00000302, -0.00001412, -0.00000154, -0.00001480, -0.00001018, -0.00002388, -0.00001975, -0.00009927, 0.00001037, 0.00009799, 0.00005052, 0.00000623, 0.00003764, 0.00004644, 0.00018005, -0.00003993, -0.00008780, 0.00000083, -0.00001318, 0.00000236, 0.00000672, -0.00000552, 0.00003382, 0.00000570, -0.00001256, -0.00000960, -0.00002863, 0.00001147, -0.00001435, -0.00000440, 0.00000300, -0.00001190, -0.00000184, -0.00001345, 0.00001684, -0.00000934, -0.00000120, -0.00001574, 0.00001474, 0.00000556, -0.00001531, -0.00000309, 0.00000074, 0.00000389, 0.00000219, -0.00000447, 0.00000525, -0.00000432, -0.00000051, -0.00000055, 0.00001095, -0.00001187, 0.00001524, -0.00001443, 0.00000576, -0.00002183, 0.00002297, -0.00001515, 0.00001303, -0.00000378, 0.00000260, -0.00000237, -0.00000205, 0.00000249, -0.00000704, 0.00000754, -0.00000692, 0.00000503, -0.00001098, 0.00000413, 0.00000189, 0.00000471, -0.00000185, 0.00000968, -0.00001004, -0.00000390, 0.00000317, 0.00000025, -0.00000077, -0.00000185, 0.00000015, -0.00000103, -0.00000113, 0.00000180, 0.00000443, -0.00000572, -0.00000480, 0.00000374, 0.00000372, -0.00000294, 0.00000822, -0.00000544, -0.00000211, -0.00000311, 0.00000088, 0.00000166, 0.00000194, -0.00000346, 0.00000122, -0.00000454, -0.00000041, 0.00000175, 0.00000016, -0.00000192, 0.00000844, -0.00000788, -0.00000346, -0.00000298, 0.00000369, -0.00000490, 0.00000809, 0.00000019, 0.00000357, -0.00000510, 0.00000360, -0.00000577, 0.00000041, 0.00000768, -0.00000046, -0.00000835, 0.00000721, -0.00000807, 0.00000998, -0.00001283, -0.00001082, -0.00000104, },
		{ -0.00000960, 0.00000299, -0.00001346, -0.00000200, -0.00001400, -0.00001310, -0.00002012, -0.00002609, -0.00009964, 0.00003279, 0.00009197, 0.00004081, 0.00000761, 0.00003040, 0.00004711, 0.00017354, -0.00003233, -0.00008678, 0.00000138, -0.00001282, 0.00000371, 0.00000699, -0.00000390, 0.00003570, 0.00000354, -0.00001152, -0.00001620, -0.00002376, 0.00001064, -0.00001629, -0.00000086, -0.00000118, -0.00000822, -0.00000513, -0.00001053, 0.00001615, -0.00001107, -0.00000073, -0.00001582, 0.00001818, 0.00000125, -0.00001422, -0.00000286, 0.00000149, 0.00000398, 0.00000131, -0.00000372, 0.00000449, -0.00000355, -0.00000155, 0.00000209, 0.00000767, -0.00000913, 0.00001323, -0.00001337, 0.00000309, -0.00001837, 0.00002256, -0.00001702, 0.00001599, -0.00000711, 0.00000533, -0.00000518, 0.00000059, 0.00000017, -0.00000489, 0.00000612, -0.00000612, 0.00000406, -0.00001086, 0.00000518, 0.00000193, 0.00000395, -0.00000076, 0.00000878, -0.00001156, -0.00000192, 0.00000293, 0.00000010, -0.00000093, -0.00000181, 0.00000031, -0.00000133, -0.00000091, 0.00000242, 0.00000373, -0.00000668, -0.00000379, 0.00000456, 0.00000245, -0.00000179, 0.00000803, -0.00000666, -0.00000133, -0.00000350, 0.00000181, 0.00000110, 0.00000169, -0.00000326, 0.00000109, -0.00000465, -0.00000000, 0.00000211, -0.00000054, -0.00000100, 0.00000796, -0.00000889, -0.00000311, -0.00000210, 0.00000288, -0.00000419, 0.00000880, -0.00000115, 0.00000376, -0.00000499, 0.00000385, -0.00000655, 0.00000177, 0.00000764, -0.00000215, -0.00000753, 0.00000622, -0.00000635, 0.00000975, -0.00001517, -0.00000925, -0.00000040, -0.00000277, },
		{ 0.00000320, -0.00001310, -0.00000207, -0.00001376, -0.00001541, -0.00001659, -0.00003369, -0.00009597, 0.00005276, 0.00008324, 0.00003362, 0.00000833, 0.00002367, 0.00004761, 0.00016741, -0.00002505, -0.00008566, 0.00000198, -0.00001250, 0.00000520, 0.00000693, -0.00000144, 0.00003669, 0.00000194, -0.00001158, -0.00002223, -0.00001791, 0.00000805, -0.00001649, 0.00000141, -0.00000465, -0.00000483, -0.00000852, -0.00000697, 0.00001442, -0.00001197, -0.00000085, -0.00001492, 0.00002066, -0.00000304, -0.00001255, -0.00000304, 0.00000262, 0.00000354, 0.00000087, -0.00000329, 0.00000406, -0.00000327, -0.00000191, 0.00000405, 0.00000477, -0.00000645, 0.00001111, -0.00001230, 0.00000039, -0.00001422, 0.00002086, -0.00001764, 0.00001799, -0.00000986, 0.00000773, -0.00000777, 0.00000318, -0.00000226, -0.00000248, 0.00000436, -0.00000501, 0.00000276, -0.00001036, 0.00000612, 0.00000180, 0.00000327, 0.00000045, 0.00000738, -0.00001256, 0.00000001, 0.00000249, 0.00000002, -0.00000113, -0.00000171, 0.00000040, -0.00000163, -0.00000057, 0.00000296, 0.00000287, -0.00000751, -0.00000253, 0.00000505, 0.00000124, -0.00000039, 0.00000736, -0.00000755, -0.00000064, -0.00000386, 0.00000278, 0.00000031, 0.00000168, -0.00000318, 0.00000103, -0.00000479, 0.00000054, 0.00000227, -0.00000115, 0.00000004, 0.00000718, -0.00000977, -0.00000257, -0.00000133, 0.00000200, -0.00000323, 0.00000919, -0.00000239, 0.00000401, -0.00000485, 0.00000399, -0.00000724, 0.00000335, 0.00000702, -0.00000341, -0.00000690, 0.00000534, -0.00000430, 0.00000873, -0.00001697, -0.00000759, 0.00000002, -0.00000297, -0.00001037, },
		{ -0.00001296, -0.00000183, -0.00001413, -0.00001688, -0.00001382, -0.00004180, -0.00008853, 0.00006949, 0.00007285, 0.00002862, 0.00000834, 0.00001754, 0.00004778, 0.00016180, -0.00001823, -0.00008430, 0.00000250, -0.00001209, 0.00000668, 0.00000673, 0.00000173, 0.00003682, 0.00000086, -0.00001295, -0.00002714, -0.00001173, 0.00000412, -0.00001508, 0.00000229, -0.00000713, -0.00000207, -0.00001161, -0.00000315, 0.00001189, -0.00001217, -0.00000147, -0.00001307, 0.00002198, -0.00000694, -0.00001057, -0.00000346, 0.00000400, 0.00000267, 0.00000080, -0.00000311, 0.00000390, -0.00000337, -0.00000162, 0.00000525, 0.00000240, -0.00000403, 0.00000904, -0.00001137, -0.00000208, -0.00000971, 0.00001811, -0.00001705, 0.00001894, -0.00001188, 0.00000962, -0.00000998, 0.00000556, -0.00000464, 0.00000007, 0.00000232, -0.00000364, 0.00000115, -0.00000943, 0.00000687, 0.00000155, 0.00000274, 0.00000168, 0.00000556, -0.00001297, 0.00000179, 0.00000193, -0.00000001, -0.00000134, -0.00000155, 0.00000044, -0.00000192, -0.00000012, 0.00000336, 0.00000186, -0.00000816, -0.00000107, 0.00000517, 0.00000021, 0.00000115, 0.00000626, -0.00000806, -0.00000011, -0.00000409, 0.00000370, -0.00000061, 0.00000186, -0.00000326, 0.00000105, -0.00000498, 0.00000123, 0.00000218, -0.00000158, 0.00000111, 0.00000608, -0.00001047, -0.00000187, -0.00000073, 0.00000111, -0.00000206, 0.00000924, -0.00000347, 0.00000426, -0.00000466, 0.00000397, -0.00000773, 0.00000501, 0.00000589, -0.00000417, -0.00000656, 0.00000474, -0.00000218, 0.00000704, -0.00001813, -0.00000598, 0.00000033, -0.00000324, -0.00001107, 0.00000352, },
		{ -0.00000139, -0.00001503, -0.00001741, -0.00001224, -0.00004962, -0.00007791, 0.00008248, 0.00006177, 0.00002536, 0.00000772, 0.00001201, 0.00004755, 0.00015683, -0.00001196, -0.00008260, 0.00000284, -0.00001147, 0.00000798, 0.00000660, 0.00000542, 0.00003622, 0.00000016, -0.00001562, -0.00003050, -0.00000586, -0.00000061, -0.00001233, 0.00000182, -0.00000849, -0.00000019, -0.00001406, 0.00000057, 0.00000886, -0.00001186, -0.00000243, -0.00001038, 0.00002205, -0.00001020, -0.00000854, -0.00000395, 0.00000546, 0.00000148, 0.00000105, -0.00000313, 0.00000392, -0.00000375, -0.00000077, 0.00000572, 0.00000063, -0.00000198, 0.00000715, -0.00001073, -0.00000406, -0.00000519, 0.00001462, -0.00001541, 0.00001886, -0.00001309, 0.00001089, -0.00001166, 0.00000757, -0.00000680, 0.00000259, 0.00000013, -0.00000208, -0.00000072, -0.00000808, 0.00000737, 0.00000120, 0.00000240, 0.00000283, 0.00000341, -0.00001280, 0.00000331, 0.00000132, -0.00000002, -0.00000154, -0.00000136, 0.00000041, -0.00000217, 0.00000042, 0.00000362, 0.00000075, -0.00000858, 0.00000053, 0.00000490, -0.00000054, 0.00000270, 0.00000480, -0.00000818, 0.00000020, -0.00000413, 0.00000448, -0.00000158, 0.00000219, -0.00000346, 0.00000116, -0.00000522, 0.00000207, 0.00000180, -0.00000176, 0.00000213, 0.00000473, -0.00001094, -0.00000105, -0.00000030, 0.00000028, -0.00000076, 0.00000897, -0.00000436, 0.00000450, -0.00000440, 0.00000376, -0.00000795, 0.00000660, 0.00000436, -0.00000447, -0.00000658, 0.00000459, -0.00000021, 0.00000485, -0.00001867, -0.00000450, 0.00000056, -0.00000365, -0.00001159, 0.00000382, -0.00001289, },
		{ -0.00001628, -0.00001703, -0.00001213, -0.00005638, -0.00006487, 0.00009156, 0.00005081, 0.00002339, 0.00000662, 0.00000700, 0.00004690, 0.00015252, -0.00000630, -0.00008049, 0.00000293, -0.00001053, 0.00000896, 0.00000676, 0.00000940, 0.00003505, -0.00000042, -0.00001942, -0.00003206, -0.00000088, -0.00000555, -0.00000865, 0.00000018, -0.00000873, 0.00000068, -0.00001563, 0.00000383, 0.00000564, -0.00001124, -0.00000350, -0.00000707, 0.00002090, -0.00001262, -0.00000666, -0.00000434, 0.00000686, 0.00000009, 0.00000153, -0.00000329, 0.00000402, -0.00000428, 0.00000050, 0.00000550, -0.00000050, -0.00000039, 0.00000551, -0.00001045, -0.00000538, -0.00000101, 0.00001073, -0.00001293, 0.00001783, -0.00001349, 0.00001149, -0.00001273, 0.00000908, -0.00000860, 0.00000494, -0.00000210, -0.00000043, -0.00000277, -0.00000632, 0.00000755, 0.00000082, 0.00000226, 0.00000381, 0.00000107, -0.00001206, 0.00000450, 0.00000072, -0.00000003, -0.00000170, -0.00000116, 0.00000033, -0.00000238, 0.00000106, 0.00000370, -0.00000044, -0.00000875, 0.00000219, 0.00000424, -0.00000090, 0.00000410, 0.00000308, -0.00000791, 0.00000026, -0.00000392, 0.00000502, -0.00000250, 0.00000259, -0.00000374, 0.00000134, -0.00000547, 0.00000304, 0.00000113, -0.00000160, 0.00000296, 0.00000318, -0.00001115, -0.00000015, -0.00000008, -0.00000043, 0.00000062, 0.00000839, -0.00000503, 0.00000469, -0.00000409, 0.00000335, -0.00000782, 0.00000797, 0.00000258, -0.00000438, -0.00000693, 0.00000498, 0.00000137, 0.00000236, -0.00001862, -0.00000319, 0.00000074, -0.00000422, -0.00001186, 0.00000399, -0.00001275, -0.00000093, },
		{ -0.00001594, -0.00001359, -0.00006141, -0.00005031, 0.00009684, 0.00004059, 0.00002225, 0.00000523, 0.00000239, 0.00004591, 0.00014884, -0.00000123, -0.00007796, 0.00000276, -0.00000924, 0.00000953, 0.00000738, 0.00001348, 0.00003347, -0.00000118, -0.00002402, -0.00003178, 0.00000280, -0.00001018, -0.00000450, -0.00000234, -0.00000798, 0.00000055, -0.00001618, 0.00000638, 0.00000252, -0.00001055, -0.00000444, -0.00000342, 0.00001868, -0.00001412, -0.00000510, -0.00000451, 0.00000806, -0.00000135, 0.00000217, -0.00000352, 0.00000411, -0.00000482, 0.00000205, 0.00000470, -0.00000101, 0.00000072, 0.00000416, -0.00001056, -0.00000591, 0.00000256, 0.00000678, -0.00000990, 0.00001601, -0.00001313, 0.00001142, -0.00001315, 0.00001000, -0.00000990, 0.00000696, -0.00000422, 0.00000119, -0.00000490, -0.00000419, 0.00000738, 0.00000047, 0.00000231, 0.00000451, -0.00000135, -0.00001082, 0.00000533, 0.00000018, -0.00000008, -0.00000180, -0.00000099, 0.00000023, -0.00000254, 0.00000177, 0.00000357, -0.00000165, -0.00000864, 0.00000385, 0.00000324, -0.00000083, 0.00000523, 0.00000125, -0.00000733, 0.00000008, -0.00000345, 0.00000526, -0.00000327, 0.00000298, -0.00000406, 0.00000155, -0.00000570, 0.00000407, 0.00000018, -0.00000106, 0.00000352, 0.00000152, -0.00001110, 0.00000079, -0.00000006, -0.00000097, 0.00000199, 0.00000756, -0.00000548, 0.00000483, -0.00000373, 0.00000278, -0.00000730, 0.00000899, 0.00000072, -0.00000402, -0.00000754, 0.00000592, 0.00000239, -0.00000018, -0.00001810, -0.00000204, 0.00000083, -0.00000491, -0.00001184, 0.00000393, -0.00001242, -0.00000065, -0.00001766, },
		{ -0.00001652, -0.00006422, -0.00003515, 0.00009864, 0.00003153, 0.00002156, 0.00000373, -0.00000196, 0.00004464, 0.00014575, 0.00000328, -0.00007503, 0.00000233, -0.00000759, 0.00000966, 0.00000858, 0.00001747, 0.00003163, -0.00000244, -0.00002897, -0.00002982, 0.00000494, -0.00001402, -0.00000037, -0.00000536, -0.00000648, -0.00000048, -0.00001568, 0.00000804, -0.00000028, -0.00000997, -0.00000502, 0.00000023, 0.00001560, -0.00001470, -0.00000394, -0.00000436, 0.00000893, -0.00000273, 0.00000287, -0.00000376, 0.00000413, -0.00000526, 0.00000371, 0.00000346, -0.00000098, 0.00000135, 0.00000307, -0.00001100, -0.00000561, 0.00000529, 0.00000310, -0.00000661, 0.00001361, -0.00001214, 0.00001074, -0.00001293, 0.00001031, -0.00001064, 0.00000855, -0.00000611, 0.00000268, -0.00000699, -0.00000179, 0.00000683, 0.00000021, 0.00000254, 0.00000488, -0.00000369, -0.00000916, 0.00000578, -0.00000027, -0.00000018, -0.00000181, -0.00000086, 0.00000010, -0.00000264, 0.00000253, 0.00000322, -0.00000284, -0.00000823, 0.00000540, 0.00000195, -0.00000029, 0.00000596, -0.00000055, -0.00000653, -0.00000029, -0.00000273, 0.00000518, -0.00000382, 0.00000328, -0.00000433, 0.00000173, -0.00000585, 0.00000511, -0.00000099, -0.00000013, 0.00000369, -0.00000014, -0.00001080, 0.00000173, -0.00000025, -0.00000129, 0.00000327, 0.00000654, -0.00000571, 0.00000491, -0.00000335, 0.00000206, -0.00000641, 0.00000956, -0.00000105, -0.00000354, -0.00000828, 0.00000737, 0.00000271, -0.00000257, -0.00001723, -0.00000101, 0.00000079, -0.00000566, -0.00001156, 0.00000360, -0.00001180, -0.00000072, -0.00001887, -0.00001443, },
		{ -0.00006454, -0.00002025, 0.00009748, 0.00002383, 0.00002102, 0.00000227, -0.00000618, 0.00004320, 0.00014315, 0.00000731, -0.00007178, 0.00000171, -0.00000563, 0.00000936, 0.00001046, 0.00002124, 0.00002959, -0.00000444, -0.00003376, -0.00002651, 0.00000549, -0.00001677, 0.00000335, -0.00000851, -0.00000453, -0.00000219, -0.00001423, 0.00000874, -0.00000260, -0.00000964, -0.00000504, 0.00000358, 0.00001197, -0.00001447, -0.00000321, -0.00000384, 0.00000938, -0.00000394, 0.00000357, -0.00000397, 0.00000403, -0.00000548, 0.00000531, 0.00000193, -0.00000052, 0.00000157, 0.00000220, -0.00001170, -0.00000454, 0.00000706, -0.00000005, -0.00000336, 0.00001087, -0.00001067, 0.00000956, -0.00001213, 0.00001002, -0.00001076, 0.00000959, -0.00000766, 0.00000391, -0.00000891, 0.00000079, 0.00000593, 0.00000010, 0.00000288, 0.00000489, -0.00000584, -0.00000720, 0.00000586, -0.00000059, -0.00000033, -0.00000175, -0.00000079, -0.00000002, -0.00000269, 0.00000333, 0.00000263, -0.00000394, -0.00000754, 0.00000677, 0.00000047, 0.00000069, 0.00000620, -0.00000218, -0.00000562, -0.00000078, -0.00000181, 0.00000478, -0.00000410, 0.00000343, -0.00000450, 0.00000182, -0.00000585, 0.00000607, -0.00000231, 0.00000117, 0.00000342, -0.00000170, -0.00001029, 0.00000264, -0.00000062, -0.00000136, 0.00000439, 0.00000540, -0.00000576, 0.00000495, -0.00000299, 0.00000127, -0.00000516, 0.00000961, -0.00000260, -0.00000308, -0.00000899, 0.00000919, 0.00000230, -0.00000464, -0.00001615, -0.00000005, 0.00000054, -0.00000635, -0.00001109, 0.00000304, -0.00001086, -0.00000127, -0.00001965, -0.00001287, -0.00002064, },
		{ -0.00000637, 0.00009395, 0.00001757, 0.00002045, 0.00000098, -0.00001036, 0.00004165, 0.00014097, 0.00001094, -0.00006826, 0.00000097, -0.00000344, 0.00000869, 0.00001303, 0.00002471, 0.00002736, -0.00000736, -0.00003789, -0.00002232, 0.00000457, -0.00001825, 0.00000631, -0.00001143, -0.00000246, -0.00000430, -0.00001201, 0.00000852, -0.00000438, -0.00000963, -0.00000437, 0.00000632, 0.00000809, -0.00001357, -0.00000287, -0.00000297, 0.00000939, -0.00000491, 0.00000421, -0.00000412, 0.00000377, -0.00000542, 0.00000671, 0.00000029, 0.00000028, 0.00000144, 0.00000149, -0.00001252, -0.00000279, 0.00000782, -0.00000246, -0.00000039, 0.00000802, -0.00000892, 0.00000804, -0.00001088, 0.00000920, -0.00001028, 0.00001006, -0.00000878, 0.00000478, -0.00001054, 0.00000342, 0.00000470, 0.00000019, 0.00000329, 0.00000452, -0.00000767, -0.00000505, 0.00000562, -0.00000080, -0.00000054, -0.00000160, -0.00000080, -0.00000012, -0.00000266, 0.00000413, 0.00000180, -0.00000490, -0.00000660, 0.00000788, -0.00000110, 0.00000208, 0.00000588, -0.00000350, -0.00000472, -0.00000129, -0.00000079, 0.00000409, -0.00000409, 0.00000337, -0.00000452, 0.00000178, -0.00000564, 0.00000683, -0.00000365, 0.00000277, 0.00000267, -0.00000306, -0.00000961, 0.00000349, -0.00000115, -0.00000114, 0.00000527, 0.00000422, -0.00000565, 0.00000496, -0.00000269, 0.00000049, -0.00000364, 0.00000914, -0.00000384, -0.00000278, -0.00000948, 0.00001119, 0.00000121, -0.00000626, -0.00001495, 0.00000089, 0.00000001, -0.00000687, -0.00001056, 0.00000233, -0.00000965, -0.00000236, -0.00001979, -0.00001165, -0.00002554, -0.00006229, },
		{ 0.00008868, 0.00001267, 0.00001973, -0.00000009, -0.00001458, 0.00004006, 0.00013913, 0.00001425, -0.00006455, 0.00000019, -0.00000111, 0.00000777, 0.00001627, 0.00002785, 0.00002483, -0.00001124, -0.00004092, -0.00001777, 0.00000247, -0.00001846, 0.00000833, -0.00001382, -0.00000056, -0.00000651, -0.00000926, 0.00000748, -0.00000561, -0.00000993, -0.00000298, 0.00000824, 0.00000428, -0.00001223, -0.00000281, -0.00000181, 0.00000895, -0.00000559, 0.00000475, -0.00000420, 0.00000337, -0.00000503, 0.00000781, -0.00000134, 0.00000126, 0.00000104, 0.00000085, -0.00001331, -0.00000056, 0.00000761, -0.00000402, 0.00000206, 0.00000529, -0.00000706, 0.00000633, -0.00000932, 0.00000795, -0.00000927, 0.00000994, -0.00000942, 0.00000524, -0.00001175, 0.00000598, 0.00000321, 0.00000051, 0.00000369, 0.00000380, -0.00000910, -0.00000285, 0.00000510, -0.00000089, -0.00000078, -0.00000138, -0.00000087, -0.00000020, -0.00000257, 0.00000490, 0.00000074, -0.00000565, -0.00000543, 0.00000868, -0.00000263, 0.00000378, 0.00000500, -0.00000440, -0.00000395, -0.00000172, 0.00000024, 0.00000321, -0.00000383, 0.00000309, -0.00000435, 0.00000157, -0.00000516, 0.00000730, -0.00000488, 0.00000457, 0.00000145, -0.00000414, -0.00000880, 0.00000426, -0.00000182, -0.00000065, 0.00000586, 0.00000308, -0.00000543, 0.00000496, -0.00000248, -0.00000020, -0.00000196, 0.00000819, -0.00000470, -0.00000274, -0.00000957, 0.00001314, -0.00000046, -0.00000739, -0.00001373, 0.00000182, -0.00000085, -0.00000709, -0.00001010, 0.00000160, -0.00000828, -0.00000393, -0.00001915, -0.00001116, -0.00003068, -0.00005766, 0.00000593, },
		{ 0.00000900, 0.00001881, -0.00000092, -0.00001886, 0.00003846, 0.00013759, 0.00001730, -0.00006071, -0.00000054, 0.00000127, 0.00000669, 0.00002013, 0.00003065, 0.00002186, -0.00001597, -0.00004253, -0.00001339, -0.00000042, -0.00001754, 0.00000931, -0.00001551, 0.00000091, -0.00000854, -0.00000627, 0.00000579, -0.00000637, -0.00001047, -0.00000092, 0.00000919, 0.00000083, -0.00001066, -0.00000292, -0.00000046, 0.00000811, -0.00000595, 0.00000515, -0.00000422, 0.00000283, -0.00000431, 0.00000852, -0.00000282, 0.00000233, 0.00000046, 0.00000022, -0.00001391, 0.00000195, 0.00000657, -0.00000469, 0.00000386, 0.00000287, -0.00000530, 0.00000462, -0.00000761, 0.00000643, -0.00000783, 0.00000927, -0.00000959, 0.00000523, -0.00001245, 0.00000831, 0.00000155, 0.00000107, 0.00000401, 0.00000277, -0.00001007, -0.00000071, 0.00000438, -0.00000090, -0.00000103, -0.00000112, -0.00000100, -0.00000026, -0.00000240, 0.00000559, -0.00000053, -0.00000615, -0.00000410, 0.00000912, -0.00000401, 0.00000566, 0.00000360, -0.00000482, -0.00000342, -0.00000196, 0.00000116, 0.00000224, -0.00000337, 0.00000262, -0.00000399, 0.00000118, -0.00000438, 0.00000738, -0.00000587, 0.00000642, -0.00000020, -0.00000488, -0.00000792, 0.00000493, -0.00000258, 0.00000011, 0.00000611, 0.00000204, -0.00000515, 0.00000497, -0.00000241, -0.00000071, -0.00000024, 0.00000684, -0.00000519, -0.00000300, -0.00000913, 0.00001480, -0.00000251, -0.00000804, -0.00001252, 0.00000274, -0.00000205, -0.00000693, -0.00000985, 0.00000100, -0.00000691, -0.00000587, -0.00001772, -0.00001165, -0.00003552, -0.00005101, 0.00001627, 0.00008225, },
		{ 0.00001769, -0.00000153, -0.00002319, 0.00003685, 0.00013631, 0.00002014, -0.00005679, -0.00000118, 0.00000360, 0.00000559, 0.00002455, 0.00003310, 0.00001823, -0.00002131, -0.00004255, -0.00000964, -0.00000366, -0.00001575, 0.00000933, -0.00001639, 0.00000178, -0.00001014, -0.00000331, 0.00000368, -0.00000677, -0.00001112, 0.00000170, 0.00000914, -0.00000205, -0.00000907, -0.00000303, 0.00000096, 0.00000694, -0.00000602, 0.00000542, -0.00000419, 0.00000221, -0.00000329, 0.00000881, -0.00000405, 0.00000337, -0.00000025, -0.00000045, -0.00001420, 0.00000450, 0.00000489, -0.00000452, 0.00000493, 0.00000089, -0.00000377, 0.00000305, -0.00000592, 0.00000480, -0.00000609, 0.00000815, -0.00000930, 0.00000476, -0.00001258, 0.00001029, -0.00000019, 0.00000186, 0.00000417, 0.00000152, -0.00001054, 0.00000127, 0.00000353, -0.00000084, -0.00000128, -0.00000085, -0.00000117, -0.00000030, -0.00000214, 0.00000616, -0.00000196, -0.00000635, -0.00000268, 0.00000919, -0.00000513, 0.00000761, 0.00000174, -0.00000474, -0.00000320, -0.00000193, 0.00000184, 0.00000129, -0.00000281, 0.00000202, -0.00000348, 0.00000062, -0.00000329, 0.00000703, -0.00000648, 0.00000817, -0.00000219, -0.00000526, -0.00000700, 0.00000547, -0.00000338, 0.00000111, 0.00000600, 0.00000118, -0.00000485, 0.00000501, -0.00000249, -0.00000096, 0.00000138, 0.00000522, -0.00000537, -0.00000355, -0.00000809, 0.00001596, -0.00000475, -0.00000828, -0.00001130, 0.00000360, -0.00000355, -0.00000635, -0.00000992, 0.00000069, -0.00000576, -0.00000794, -0.00001561, -0.00001331, -0.00003952, -0.00004283, 0.00002446, 0.00007517, 0.00000637, },
		{ -0.00000195, -0.00002755, 0.00003520, 0.00013530, 0.00002279, -0.00005280, -0.00000169, 0.00000584, 0.00000457, 0.00002947, 0.00003517, 0.00001374, -0.00002690, -0.00004100, -0.00000690, -0.00000680, -0.00001343, 0.00000854, -0.00001650, 0.00000198, -0.00001115, -0.00000061, 0.00000135, -0.00000696, -0.00001170, 0.00000466, 0.00000813, -0.00000422, -0.00000764, -0.00000301, 0.00000229, 0.00000555, -0.00000581, 0.00000556, -0.00000413, 0.00000157, -0.00000203, 0.00000868, -0.00000498, 0.00000431, -0.00000102, -0.00000120, -0.00001406, 0.00000686, 0.00000279, -0.00000364, 0.00000526, -0.00000055, -0.00000258, 0.00000175, -0.00000439, 0.00000321, -0.00000422, 0.00000669, -0.00000865, 0.00000387, -0.00001211, 0.00001182, -0.00000189, 0.00000285, 0.00000413, 0.00000012, -0.00001053, 0.00000300, 0.00000261, -0.00000074, -0.00000149, -0.00000059, -0.00000136, -0.00000032, -0.00000178, 0.00000657, -0.00000351, -0.00000623, -0.00000124, 0.00000890, -0.00000590, 0.00000945, -0.00000045, -0.00000417, -0.00000332, -0.00000155, 0.00000219, 0.00000049, -0.00000226, 0.00000136, -0.00000287, -0.00000004, -0.00000194, 0.00000621, -0.00000661, 0.00000965, -0.00000439, -0.00000528, -0.00000609, 0.00000587, -0.00000416, 0.00000229, 0.00000553, 0.00000054, -0.00000456, 0.00000508, -0.00000273, -0.00000090, 0.00000276, 0.00000346, -0.00000533, -0.00000430, -0.00000643, 0.00001644, -0.00000695, -0.00000821, -0.00001003, 0.00000430, -0.00000527, -0.00000534, -0.00001037, 0.00000079, -0.00000503, -0.00000989, -0.00001305, -0.00001612, -0.00004225, -0.00003369, 0.00003052, 0.00006785, 0.00000459, 0.00001640, },
		{ -0.00003191, 0.00003349, 0.00013456, 0.00002527, -0.00004876, -0.00000205, 0.00000794, 0.00000373, 0.00003485, 0.00003677, 0.00000821, -0.00003230, -0.00003804, -0.00000537, -0.00000945, -0.00001093, 0.00000718, -0.00001595, 0.00000152, -0.00001148, 0.00000162, -0.00000102, -0.00000706, -0.00001206, 0.00000773, 0.00000632, -0.00000562, -0.00000650, -0.00000274, 0.00000341, 0.00000407, -0.00000540, 0.00000559, -0.00000407, 0.00000096, -0.00000061, 0.00000815, -0.00000556, 0.00000508, -0.00000183, -0.00000202, -0.00001344, 0.00000884, 0.00000051, -0.00000220, 0.00000493, -0.00000144, -0.00000180, 0.00000081, -0.00000314, 0.00000182, -0.00000237, 0.00000504, -0.00000772, 0.00000264, -0.00001106, 0.00001281, -0.00000343, 0.00000398, 0.00000383, -0.00000133, -0.00001007, 0.00000443, 0.00000168, -0.00000063, -0.00000164, -0.00000039, -0.00000153, -0.00000035, -0.00000130, 0.00000675, -0.00000509, -0.00000578, 0.00000014, 0.00000829, -0.00000626, 0.00001105, -0.00000283, -0.00000318, -0.00000379, -0.00000081, 0.00000213, -0.00000003, -0.00000182, 0.00000075, -0.00000226, -0.00000074, -0.00000040, 0.00000496, -0.00000619, 0.00001071, -0.00000665, -0.00000499, -0.00000519, 0.00000611, -0.00000485, 0.00000359, 0.00000472, 0.00000015, -0.00000433, 0.00000518, -0.00000313, -0.00000050, 0.00000380, 0.00000172, -0.00000521, -0.00000513, -0.00000422, 0.00001614, -0.00000891, -0.00000797, -0.00000865, 0.00000474, -0.00000706, -0.00000399, -0.00001121, 0.00000137, -0.00000488, -0.00001143, -0.00001035, -0.00001994, -0.00004341, -0.00002420, 0.00003461, 0.00006058, 0.00000350, 0.00001496, -0.00000221, },
		{ 0.00003166, 0.00013410, 0.00002757, -0.00004465, -0.00000227, 0.00000989, 0.00000314, 0.00004067, 0.00003776, 0.00000156, -0.00003703, -0.00003401, -0.00000512, -0.00001134, -0.00000859, 0.00000553, -0.00001492, 0.00000052, -0.00001113, 0.00000328, -0.00000326, -0.00000718, -0.00001203, 0.00001064, 0.00000392, -0.00000628, -0.00000571, -0.00000213, 0.00000419, 0.00000262, -0.00000486, 0.00000554, -0.00000404, 0.00000046, 0.00000087, 0.00000729, -0.00000581, 0.00000566, -0.00000265, -0.00000287, -0.00001232, 0.00001027, -0.00000171, -0.00000042, 0.00000404, -0.00000181, -0.00000143, 0.00000026, -0.00000224, 0.00000074, -0.00000071, 0.00000333, -0.00000664, 0.00000119, -0.00000950, 0.00001321, -0.00000472, 0.00000517, 0.00000328, -0.00000273, -0.00000921, 0.00000552, 0.00000079, -0.00000053, -0.00000172, -0.00000025, -0.00000166, -0.00000039, -0.00000072, 0.00000666, -0.00000663, -0.00000502, 0.00000139, 0.00000741, -0.00000618, 0.00001228, -0.00000524, -0.00000187, -0.00000456, 0.00000030, 0.00000162, -0.00000020, -0.00000162, 0.00000030, -0.00000174, -0.00000134, 0.00000121, 0.00000334, -0.00000519, 0.00001121, -0.00000882, -0.00000444, -0.00000432, 0.00000617, -0.00000540, 0.00000492, 0.00000363, 0.00000002, -0.00000416, 0.00000530, -0.00000365, 0.00000026, 0.00000441, 0.00000014, -0.00000515, -0.00000587, -0.00000161, 0.00001501, -0.00001046, -0.00000766, -0.00000709, 0.00000479, -0.00000876, -0.00000244, -0.00001234, 0.00000243, -0.00000543, -0.00001230, -0.00000789, -0.00002448, -0.00004287, -0.00001489, 0.00003697, 0.00005357, 0.00000296, 0.00001340, -0.00000234, -0.00003622, },
		},
		{ { -0.00011238, -0.00001293, 0.00006250, 0.00002245, 0.00001985, 0.00002009, 0.00000864, 0.00002981, 0.00007165, -0.00000136, -0.00005835, -0.00001773, -0.00001494, -0.00000939, -0.00001160, -0.00000412, -0.00000619, -0.00000655, 0.00000562, -0.00000393, -0.00000543, -0.00000526, -0.00000320, 0.00000174, -0.00001420, 0.00000057, 0.00000495, -0.00000049, -0.00000245, -0.00000450, 0.00001054, -0.00000338, -0.00000214, -0.00000113, 0.00000214, 0.00000118, 0.00000160, 0.00000151, -0.00000671, -0.00000016, 0.00000586, -0.00000144, -0.00000135, 0.00000012, 0.00000022, 0.00000075, -0.00000110, 0.00000195, 0.00000082, 0.00000007, -0.00000873, 0.00000278, 0.00000310, 0.00000144, -0.00000132, -0.00000196, -0.00000394, -0.00000127, 0.00000649, -0.00000230, -0.00000275, -0.00000201, 0.00000138, 0.00000081, 0.00000129, 0.00000081, 0.00000138, -0.00000201, -0.00000275, -0.00000230, 0.00000649, -0.00000127, -0.00000394, -0.00000196, -0.00000132, 0.00000144, 0.00000310, 0.00000278, -0.00000873, 0.00000007, 0.00000082, 0.00000195, -0.00000110, 0.00000075, 0.00000022, 0.00000012, -0.00000135, -0.00000144, 0.00000586, -0.00000016, -0.00000671, 0.00000151, 0.00000160, 0.00000118, 0.00000214, -0.00000113, -0.00000214, -0.00000338, 0.00001054, -0.00000450, -0.00000245, -0.00000049, 0.00000495, 0.00000057, -0.00001420, 0.00000174, -0.00000320, -0.00000526, -0.00000543, -0.00000393, 0.00000562, -0.00000655, -0.00000619, -0.00000412, -0.00001160, -0.00000939, -0.00001494, -0.00001773, -0.00005835, -0.00000136, 0.00007165, 0.00002981, 0.00000864, 0.00002009, 0.00001985, 0.00002245, 0.00006250, -0.00001293, },
		{ -0.00001481, 0.00005858, 0.00002223, 0.00001891, 0.00001888, 0.00000774, 0.00002521, 0.00006583, 0.00001460, -0.00005626, -0.00002380, -0.00001242, -0.00001132, -0.00001077, -0.00000483, -0.00000516, -0.00000776, 0.00000556, -0.00000309, -0.00000546, -0.00000531, -0.00000439, 0.00000333, -0.00001332, -0.00000254, 0.00000693, -0.00000126, -0.00000102, -0.00000631, 0.00000973, -0.00000079, -0.00000372, -0.00000040, 0.00000130, 0.00000176, 0.00000096, 0.00000243, -0.00000634, -0.00000133, 0.00000563, -0.00000016, -0.00000195, 0.00000032, -0.00000041, 0.00000108, -0.00000128, 0.00000202, 0.00000071, 0.00000112, -0.00000843, 0.00000069, 0.00000425, 0.00000100, -0.00000048, -0.00000242, -0.00000312, -0.00000291, 0.00000684, -0.00000134, -0.00000297, -0.00000238, 0.00000131, 0.00000096, 0.00000117, 0.00000091, 0.00000123, -0.00000155, -0.00000262, -0.00000290, 0.00000560, 0.00000059, -0.00000470, -0.00000152, -0.00000212, 0.00000184, 0.00000203, 0.00000438, -0.00000829, -0.00000139, 0.00000113, 0.00000162, -0.00000063, 0.00000013, 0.00000100, -0.00000014, -0.00000064, -0.00000263, 0.00000578, 0.00000104, -0.00000677, 0.00000035, 0.00000228, 0.00000059, 0.00000296, -0.00000181, -0.00000075, -0.00000522, 0.00001014, -0.00000177, -0.00000434, 0.00000071, 0.00000275, 0.00000318, -0.00001421, -0.00000028, -0.00000210, -0.00000508, -0.00000541, -0.00000459, 0.00000533, -0.00000516, -0.00000719, -0.00000342, -0.00001244, -0.00000762, -0.00001691, -0.00001355, -0.00005638, -0.00001941, 0.00007532, 0.00003600, 0.00000933, 0.00002142, 0.00002090, 0.00002254, 0.00006659, -0.00001097, -0.00011246, },
		{ 0.00005478, 0.00002191, 0.00001803, 0.00001785, 0.00000659, 0.00002196, 0.00005880, 0.00002766, -0.00005038, -0.00003090, -0.00001008, -0.00001285, -0.00001037, -0.00000523, -0.00000442, -0.00000853, 0.00000499, -0.00000199, -0.00000560, -0.00000514, -0.00000569, 0.00000450, -0.00001176, -0.00000572, 0.00000829, -0.00000134, -0.00000024, -0.00000711, 0.00000782, 0.00000227, -0.00000524, 0.00000018, 0.00000057, 0.00000221, 0.00000048, 0.00000305, -0.00000567, -0.00000243, 0.00000515, 0.00000110, -0.00000234, 0.00000037, -0.00000080, 0.00000105, -0.00000118, 0.00000186, 0.00000078, 0.00000177, -0.00000749, -0.00000168, 0.00000525, 0.00000069, 0.00000025, -0.00000274, -0.00000242, -0.00000416, 0.00000656, -0.00000002, -0.00000324, -0.00000260, 0.00000097, 0.00000136, 0.00000086, 0.00000119, 0.00000094, -0.00000106, -0.00000254, -0.00000317, 0.00000434, 0.00000244, -0.00000522, -0.00000127, -0.00000272, 0.00000203, 0.00000125, 0.00000532, -0.00000708, -0.00000320, 0.00000160, 0.00000106, 0.00000008, -0.00000068, 0.00000178, -0.00000036, 0.00000010, -0.00000364, 0.00000538, 0.00000224, -0.00000652, -0.00000094, 0.00000286, 0.00000011, 0.00000362, -0.00000227, 0.00000024, -0.00000616, 0.00000858, 0.00000162, -0.00000636, 0.00000196, 0.00000079, 0.00000492, -0.00001318, -0.00000272, -0.00000108, -0.00000487, -0.00000529, -0.00000520, 0.00000489, -0.00000387, -0.00000788, -0.00000304, -0.00001292, -0.00000648, -0.00001772, -0.00001176, -0.00005061, -0.00003831, 0.00007579, 0.00004397, 0.00000991, 0.00002285, 0.00002210, 0.00002250, 0.00007082, -0.00000890, -0.00011271, -0.00001660, },
		{ 0.00002158, 0.00001712, 0.00001707, 0.00000515, 0.00001983, 0.00005130, 0.00003745, -0.00004144, -0.00003795, -0.00000874, -0.00001347, -0.00001071, -0.00000508, -0.00000416, -0.00000868, 0.00000382, -0.00000059, -0.00000587, -0.00000474, -0.00000708, 0.00000529, -0.00000976, -0.00000861, 0.00000875, -0.00000061, -0.00000013, -0.00000698, 0.00000508, 0.00000543, -0.00000640, 0.00000045, 0.00000009, 0.00000242, 0.00000023, 0.00000331, -0.00000473, -0.00000340, 0.00000445, 0.00000225, -0.00000246, 0.00000023, -0.00000091, 0.00000071, -0.00000086, 0.00000156, 0.00000096, 0.00000210, -0.00000611, -0.00000402, 0.00000586, 0.00000068, 0.00000072, -0.00000281, -0.00000198, -0.00000489, 0.00000566, 0.00000157, -0.00000353, -0.00000269, 0.00000039, 0.00000193, 0.00000044, 0.00000154, 0.00000063, -0.00000063, -0.00000246, -0.00000320, 0.00000289, 0.00000404, -0.00000531, -0.00000132, -0.00000302, 0.00000190, 0.00000086, 0.00000554, -0.00000520, -0.00000519, 0.00000209, 0.00000039, 0.00000089, -0.00000150, 0.00000240, -0.00000042, 0.00000074, -0.00000435, 0.00000463, 0.00000341, -0.00000602, -0.00000224, 0.00000324, -0.00000015, 0.00000403, -0.00000243, 0.00000073, -0.00000620, 0.00000608, 0.00000527, -0.00000810, 0.00000286, -0.00000052, 0.00000551, -0.00001110, -0.00000549, -0.00000016, -0.00000470, -0.00000498, -0.00000586, 0.00000450, -0.00000292, -0.00000803, -0.00000320, -0.00001278, -0.00000629, -0.00001710, -0.00001239, -0.00004183, -0.00005652, 0.00007205, 0.00005373, 0.00001051, 0.00002433, 0.00002344, 0.00002235, 0.00007513, -0.00000661, -0.00011319, -0.00001826, 0.00005105, },
		{ 0.00001606, 0.00001662, 0.00000336, 0.00001860, 0.00004386, 0.00004402, -0.00003046, -0.00004383, -0.00000901, -0.00001287, -0.00001197, -0.00000428, -0.00000447, -0.00000819, 0.00000210, 0.00000101, -0.00000624, -0.00000413, -0.00000851, 0.00000571, -0.00000752, -0.00001093, 0.00000819, 0.00000085, -0.00000054, -0.00000615, 0.00000189, 0.00000830, -0.00000694, 0.00000026, -0.00000006, 0.00000234, 0.00000025, 0.00000325, -0.00000361, -0.00000418, 0.00000357, 0.00000323, -0.00000229, -0.00000010, -0.00000079, 0.00000013, -0.00000047, 0.00000125, 0.00000115, 0.00000222, -0.00000450, -0.00000608, 0.00000591, 0.00000106, 0.00000086, -0.00000256, -0.00000187, -0.00000508, 0.00000422, 0.00000328, -0.00000372, -0.00000269, -0.00000036, 0.00000257, 0.00000004, 0.00000183, 0.00000042, -0.00000034, -0.00000233, -0.00000309, 0.00000145, 0.00000522, -0.00000488, -0.00000172, -0.00000298, 0.00000141, 0.00000090, 0.00000511, -0.00000283, -0.00000715, 0.00000245, -0.00000026, 0.00000164, -0.00000216, 0.00000270, -0.00000022, 0.00000122, -0.00000473, 0.00000356, 0.00000451, -0.00000531, -0.00000343, 0.00000332, -0.00000015, 0.00000415, -0.00000225, 0.00000072, -0.00000545, 0.00000298, 0.00000872, -0.00000914, 0.00000304, -0.00000087, 0.00000483, -0.00000808, -0.00000840, 0.00000058, -0.00000459, -0.00000446, -0.00000663, 0.00000426, -0.00000244, -0.00000755, -0.00000396, -0.00001195, -0.00000709, -0.00001511, -0.00001504, -0.00003125, -0.00007234, 0.00006332, 0.00006505, 0.00001139, 0.00002580, 0.00002493, 0.00002216, 0.00007945, -0.00000403, -0.00011400, -0.00001970, 0.00004728, 0.00002134, },
		{ 0.00001658, 0.00000121, 0.00001811, 0.00003682, 0.00004770, -0.00001859, -0.00004761, -0.00001126, -0.00001102, -0.00001407, -0.00000294, -0.00000525, -0.00000717, 0.00000002, 0.00000265, -0.00000658, -0.00000342, -0.00000985, 0.00000574, -0.00000519, -0.00001253, 0.00000667, 0.00000281, -0.00000122, -0.00000492, -0.00000134, 0.00001051, -0.00000667, -0.00000044, 0.00000012, 0.00000199, 0.00000051, 0.00000293, -0.00000242, -0.00000471, 0.00000254, 0.00000399, -0.00000186, -0.00000057, -0.00000049, -0.00000053, -0.00000016, 0.00000105, 0.00000126, 0.00000225, -0.00000286, -0.00000764, 0.00000529, 0.00000185, 0.00000069, -0.00000200, -0.00000206, -0.00000481, 0.00000241, 0.00000490, -0.00000371, -0.00000268, -0.00000119, 0.00000313, -0.00000020, 0.00000194, 0.00000041, -0.00000025, -0.00000212, -0.00000293, 0.00000017, 0.00000583, -0.00000389, -0.00000244, -0.00000262, 0.00000062, 0.00000129, 0.00000418, -0.00000025, -0.00000882, 0.00000250, -0.00000073, 0.00000214, -0.00000247, 0.00000257, 0.00000027, 0.00000149, -0.00000473, 0.00000220, 0.00000549, -0.00000443, -0.00000443, 0.00000305, 0.00000011, 0.00000401, -0.00000177, 0.00000030, -0.00000414, -0.00000030, 0.00001150, -0.00000913, 0.00000225, -0.00000012, 0.00000297, -0.00000443, -0.00001116, 0.00000102, -0.00000449, -0.00000378, -0.00000745, 0.00000419, -0.00000244, -0.00000650, -0.00000520, -0.00001058, -0.00000868, -0.00001212, -0.00001893, -0.00002035, -0.00008420, 0.00004919, 0.00007735, 0.00001299, 0.00002712, 0.00002662, 0.00002193, 0.00008370, -0.00000106, -0.00011520, -0.00002086, 0.00004338, 0.00002131, 0.00001475, },
		{ -0.00000131, 0.00001819, 0.00003037, 0.00004898, -0.00000690, -0.00004866, -0.00001547, -0.00000816, -0.00001669, -0.00000132, -0.00000626, -0.00000589, -0.00000214, 0.00000406, -0.00000673, -0.00000275, -0.00001097, 0.00000535, -0.00000286, -0.00001337, 0.00000437, 0.00000494, -0.00000184, -0.00000363, -0.00000425, 0.00001178, -0.00000552, -0.00000162, 0.00000058, 0.00000142, 0.00000094, 0.00000245, -0.00000130, -0.00000492, 0.00000137, 0.00000451, -0.00000121, -0.00000111, -0.00000014, -0.00000114, -0.00000007, 0.00000105, 0.00000122, 0.00000226, -0.00000135, -0.00000856, 0.00000398, 0.00000297, 0.00000028, -0.00000121, -0.00000247, -0.00000421, 0.00000046, 0.00000624, -0.00000337, -0.00000271, -0.00000196, 0.00000347, -0.00000016, 0.00000179, 0.00000065, -0.00000036, -0.00000182, -0.00000279, -0.00000084, 0.00000583, -0.00000244, -0.00000337, -0.00000207, -0.00000035, 0.00000188, 0.00000298, 0.00000223, -0.00000994, 0.00000208, -0.00000087, 0.00000226, -0.00000233, 0.00000197, 0.00000105, 0.00000159, -0.00000440, 0.00000065, 0.00000628, -0.00000342, -0.00000517, 0.00000245, 0.00000058, 0.00000366, -0.00000109, -0.00000038, -0.00000257, -0.00000334, 0.00001321, -0.00000786, 0.00000045, 0.00000163, 0.00000023, -0.00000058, -0.00001341, 0.00000098, -0.00000432, -0.00000306, -0.00000824, 0.00000420, -0.00000279, -0.00000511, -0.00000665, -0.00000895, -0.00001064, -0.00000875, -0.00002307, -0.00001060, -0.00009088, 0.00002985, 0.00008965, 0.00001585, 0.00002811, 0.00002860, 0.00002166, 0.00008788, 0.00000235, -0.00011684, -0.00002171, 0.00003930, 0.00002153, 0.00001314, 0.00001698, },
		{ 0.00001865, 0.00002465, 0.00004840, 0.00000376, -0.00004675, -0.00002123, -0.00000486, -0.00001935, 0.00000016, -0.00000716, -0.00000469, -0.00000405, 0.00000498, -0.00000650, -0.00000231, -0.00001171, 0.00000449, -0.00000060, -0.00001351, 0.00000160, 0.00000686, -0.00000208, -0.00000258, -0.00000652, 0.00001198, -0.00000355, -0.00000313, 0.00000119, 0.00000074, 0.00000141, 0.00000195, -0.00000035, -0.00000477, 0.00000013, 0.00000480, -0.00000042, -0.00000161, 0.00000016, -0.00000156, -0.00000027, 0.00000129, 0.00000103, 0.00000231, -0.00000006, -0.00000880, 0.00000211, 0.00000425, -0.00000021, -0.00000034, -0.00000295, -0.00000346, -0.00000140, 0.00000708, -0.00000262, -0.00000286, -0.00000259, 0.00000349, 0.00000023, 0.00000135, 0.00000112, -0.00000064, -0.00000146, -0.00000268, -0.00000153, 0.00000529, -0.00000066, -0.00000433, -0.00000149, -0.00000133, 0.00000247, 0.00000178, 0.00000431, -0.00001030, 0.00000108, -0.00000061, 0.00000193, -0.00000171, 0.00000094, 0.00000200, 0.00000157, -0.00000381, -0.00000096, 0.00000680, -0.00000229, -0.00000564, 0.00000156, 0.00000117, 0.00000324, -0.00000034, -0.00000114, -0.00000102, -0.00000577, 0.00001362, -0.00000533, -0.00000223, 0.00000411, -0.00000294, 0.00000295, -0.00001478, 0.00000028, -0.00000395, -0.00000243, -0.00000882, 0.00000416, -0.00000329, -0.00000368, -0.00000796, -0.00000748, -0.00001249, -0.00000570, -0.00002646, -0.00000327, -0.00009170, 0.00000611, 0.00010059, 0.00002065, 0.00002854, 0.00003099, 0.00002130, 0.00009199, 0.00000621, -0.00011889, -0.00002226, 0.00003502, 0.00002203, 0.00001121, 0.00001780, -0.00000412, },
		{ 0.00001975, 0.00004644, 0.00001283, -0.00004212, -0.00002779, -0.00000183, -0.00002149, 0.00000106, -0.00000757, -0.00000389, -0.00000538, 0.00000518, -0.00000574, -0.00000220, -0.00001196, 0.00000316, 0.00000157, -0.00001307, -0.00000132, 0.00000821, -0.00000173, -0.00000194, -0.00000803, 0.00001112, -0.00000094, -0.00000477, 0.00000178, 0.00000011, 0.00000182, 0.00000155, 0.00000032, -0.00000426, -0.00000116, 0.00000485, 0.00000046, -0.00000201, 0.00000032, -0.00000170, -0.00000082, 0.00000173, 0.00000073, 0.00000238, 0.00000096, -0.00000841, -0.00000014, 0.00000547, -0.00000063, 0.00000046, -0.00000333, -0.00000277, -0.00000294, 0.00000729, -0.00000144, -0.00000314, -0.00000301, 0.00000312, 0.00000095, 0.00000067, 0.00000175, -0.00000098, -0.00000112, -0.00000260, -0.00000192, 0.00000431, 0.00000122, -0.00000513, -0.00000105, -0.00000215, 0.00000288, 0.00000082, 0.00000578, -0.00000978, -0.00000050, 0.00000003, 0.00000119, -0.00000069, -0.00000037, 0.00000298, 0.00000155, -0.00000308, -0.00000250, 0.00000698, -0.00000106, -0.00000585, 0.00000048, 0.00000174, 0.00000287, 0.00000032, -0.00000178, 0.00000026, -0.00000733, 0.00001266, -0.00000173, -0.00000543, 0.00000686, -0.00000597, 0.00000563, -0.00001496, -0.00000120, -0.00000327, -0.00000207, -0.00000904, 0.00000389, -0.00000369, -0.00000253, -0.00000875, -0.00000656, -0.00001372, -0.00000359, -0.00002830, 0.00000083, -0.00008671, -0.00002051, 0.00010851, 0.00002806, 0.00002822, 0.00003392, 0.00002078, 0.00009608, 0.00001047, -0.00012129, -0.00002259, 0.00003061, 0.00002273, 0.00000906, 0.00001892, -0.00000709, 0.00001929, },
		{ 0.00004345, 0.00002009, -0.00003534, -0.00003422, 0.00000017, -0.00002256, 0.00000102, -0.00000723, -0.00000377, -0.00000591, 0.00000453, -0.00000441, -0.00000249, -0.00001169, 0.00000139, 0.00000360, -0.00001217, -0.00000409, 0.00000874, -0.00000068, -0.00000177, -0.00000875, 0.00000936, 0.00000202, -0.00000625, 0.00000216, -0.00000034, 0.00000203, 0.00000134, 0.00000064, -0.00000341, -0.00000240, 0.00000466, 0.00000135, -0.00000223, 0.00000029, -0.00000153, -0.00000166, 0.00000229, 0.00000041, 0.00000241, 0.00000177, -0.00000751, -0.00000252, 0.00000640, -0.00000077, 0.00000103, -0.00000349, -0.00000229, -0.00000400, 0.00000682, 0.00000013, -0.00000350, -0.00000321, 0.00000239, 0.00000195, -0.00000013, 0.00000240, -0.00000126, -0.00000088, -0.00000251, -0.00000207, 0.00000307, 0.00000298, -0.00000555, -0.00000091, -0.00000267, 0.00000296, 0.00000026, 0.00000648, -0.00000841, -0.00000255, 0.00000096, 0.00000013, 0.00000059, -0.00000175, 0.00000379, 0.00000166, -0.00000233, -0.00000383, 0.00000675, 0.00000026, -0.00000582, -0.00000069, 0.00000216, 0.00000266, 0.00000079, -0.00000217, 0.00000110, -0.00000793, 0.00001045, 0.00000258, -0.00000867, 0.00000934, -0.00000828, 0.00000699, -0.00001375, -0.00000348, -0.00000226, -0.00000207, -0.00000879, 0.00000328, -0.00000380, -0.00000193, -0.00000877, -0.00000646, -0.00001402, -0.00000285, -0.00002812, 0.00000143, -0.00007671, -0.00004797, 0.00011167, 0.00003860, 0.00002706, 0.00003747, 0.00002003, 0.00010023, 0.00001506, -0.00012394, -0.00002281, 0.00002617, 0.00002352, 0.00000680, 0.00002017, -0.00001002, 0.00001986, 0.00001576, },
		{ 0.00002560, -0.00002723, -0.00003953, 0.00000050, -0.00002222, -0.00000016, -0.00000599, -0.00000442, -0.00000559, 0.00000305, -0.00000258, -0.00000311, -0.00001096, -0.00000068, 0.00000538, -0.00001092, -0.00000645, 0.00000835, 0.00000100, -0.00000195, -0.00000884, 0.00000699, 0.00000496, -0.00000729, 0.00000214, -0.00000050, 0.00000198, 0.00000139, 0.00000062, -0.00000230, -0.00000353, 0.00000422, 0.00000220, -0.00000224, 0.00000005, -0.00000110, -0.00000268, 0.00000282, 0.00000021, 0.00000233, 0.00000240, -0.00000628, -0.00000478, 0.00000683, -0.00000053, 0.00000128, -0.00000331, -0.00000213, -0.00000451, 0.00000569, 0.00000197, -0.00000386, -0.00000323, 0.00000138, 0.00000307, -0.00000089, 0.00000292, -0.00000136, -0.00000081, -0.00000236, -0.00000208, 0.00000176, 0.00000439, -0.00000546, -0.00000117, -0.00000279, 0.00000263, 0.00000017, 0.00000641, -0.00000630, -0.00000487, 0.00000198, -0.00000105, 0.00000189, -0.00000297, 0.00000422, 0.00000203, -0.00000167, -0.00000482, 0.00000607, 0.00000164, -0.00000559, -0.00000181, 0.00000232, 0.00000270, 0.00000100, -0.00000225, 0.00000142, -0.00000759, 0.00000729, 0.00000707, -0.00001140, 0.00001102, -0.00000939, 0.00000677, -0.00001114, -0.00000644, -0.00000098, -0.00000244, -0.00000807, 0.00000230, -0.00000351, -0.00000203, -0.00000789, -0.00000727, -0.00001325, -0.00000358, -0.00002592, -0.00000108, -0.00006316, -0.00007386, 0.00010845, 0.00005252, 0.00002516, 0.00004163, 0.00001907, 0.00010446, 0.00001992, -0.00012673, -0.00002305, 0.00002183, 0.00002425, 0.00000463, 0.00002134, -0.00001267, 0.00002009, 0.00001281, 0.00003967, },
		{ -0.00001870, -0.00004289, -0.00000123, -0.00002038, -0.00000244, -0.00000395, -0.00000576, -0.00000453, 0.00000094, -0.00000046, -0.00000390, -0.00000993, -0.00000285, 0.00000679, -0.00000938, -0.00000830, 0.00000706, 0.00000310, -0.00000228, -0.00000854, 0.00000435, 0.00000752, -0.00000767, 0.00000164, -0.00000032, 0.00000163, 0.00000170, 0.00000030, -0.00000104, -0.00000444, 0.00000354, 0.00000297, -0.00000205, -0.00000036, -0.00000047, -0.00000373, 0.00000316, 0.00000022, 0.00000208, 0.00000295, -0.00000489, -0.00000667, 0.00000661, 0.00000015, 0.00000119, -0.00000280, -0.00000230, -0.00000450, 0.00000404, 0.00000388, -0.00000410, -0.00000316, 0.00000020, 0.00000413, -0.00000144, 0.00000315, -0.00000117, -0.00000096, -0.00000212, -0.00000202, 0.00000054, 0.00000530, -0.00000477, -0.00000182, -0.00000252, 0.00000189, 0.00000054, 0.00000569, -0.00000372, -0.00000717, 0.00000285, -0.00000211, 0.00000296, -0.00000378, 0.00000412, 0.00000274, -0.00000119, -0.00000540, 0.00000495, 0.00000304, -0.00000519, -0.00000280, 0.00000216, 0.00000301, 0.00000092, -0.00000200, 0.00000126, -0.00000649, 0.00000359, 0.00001121, -0.00001311, 0.00001148, -0.00000904, 0.00000492, -0.00000735, -0.00000980, 0.00000038, -0.00000306, -0.00000695, 0.00000103, -0.00000287, -0.00000280, -0.00000620, -0.00000886, -0.00001160, -0.00000556, -0.00002211, -0.00000580, -0.00004798, -0.00009573, 0.00009766, 0.00006953, 0.00002292, 0.00004623, 0.00001799, 0.00010874, 0.00002503, -0.00012957, -0.00002342, 0.00001771, 0.00002476, 0.00000272, 0.00002221, -0.00001482, 0.00001975, 0.00001097, 0.00003524, 0.00002960, },
		{ -0.00004377, -0.00000500, -0.00001730, -0.00000551, -0.00000140, -0.00000751, -0.00000301, -0.00000147, 0.00000162, -0.00000461, -0.00000885, -0.00000487, 0.00000768, -0.00000758, -0.00000960, 0.00000506, 0.00000531, -0.00000245, -0.00000812, 0.00000181, 0.00000938, -0.00000723, 0.00000061, 0.00000020, 0.00000101, 0.00000222, -0.00000024, 0.00000024, -0.00000504, 0.00000262, 0.00000364, -0.00000168, -0.00000088, 0.00000023, -0.00000465, 0.00000317, 0.00000054, 0.00000162, 0.00000344, -0.00000348, -0.00000803, 0.00000569, 0.00000124, 0.00000079, -0.00000200, -0.00000273, -0.00000406, 0.00000209, 0.00000563, -0.00000406, -0.00000308, -0.00000099, 0.00000495, -0.00000161, 0.00000300, -0.00000064, -0.00000135, -0.00000179, -0.00000196, -0.00000046, 0.00000561, -0.00000353, -0.00000280, -0.00000196, 0.00000087, 0.00000122, 0.00000455, -0.00000098, -0.00000913, 0.00000331, -0.00000283, 0.00000357, -0.00000399, 0.00000338, 0.00000379, -0.00000092, -0.00000554, 0.00000344, 0.00000439, -0.00000466, -0.00000359, 0.00000164, 0.00000358, 0.00000062, -0.00000150, 0.00000074, -0.00000490, -0.00000018, 0.00001444, -0.00001341, 0.00001050, -0.00000716, 0.00000166, -0.00000281, -0.00001313, 0.00000154, -0.00000374, -0.00000567, -0.00000034, -0.00000202, -0.00000404, -0.00000400, -0.00001086, -0.00000945, -0.00000826, -0.00001747, -0.00001145, -0.00003330, -0.00011143, 0.00007878, 0.00008879, 0.00002107, 0.00005091, 0.00001702, 0.00011296, 0.00003045, -0.00013246, -0.00002397, 0.00001387, 0.00002496, 0.00000120, 0.00002264, -0.00001631, 0.00001870, 0.00001027, 0.00003029, 0.00003241, -0.00001055, },
		{ -0.00001044, -0.00001351, -0.00000884, 0.00000119, -0.00000927, -0.00000146, -0.00000376, 0.00000329, -0.00000493, -0.00000799, -0.00000648, 0.00000789, -0.00000552, -0.00001040, 0.00000263, 0.00000723, -0.00000217, -0.00000786, -0.00000034, 0.00001034, -0.00000594, -0.00000088, 0.00000096, 0.00000023, 0.00000284, -0.00000086, 0.00000140, -0.00000525, 0.00000148, 0.00000418, -0.00000115, -0.00000142, 0.00000089, -0.00000529, 0.00000275, 0.00000119, 0.00000096, 0.00000390, -0.00000215, -0.00000878, 0.00000413, 0.00000259, 0.00000023, -0.00000105, -0.00000331, -0.00000339, 0.00000008, 0.00000699, -0.00000362, -0.00000310, -0.00000203, 0.00000537, -0.00000131, 0.00000242, 0.00000021, -0.00000191, -0.00000142, -0.00000192, -0.00000118, 0.00000535, -0.00000184, -0.00000395, -0.00000126, -0.00000027, 0.00000201, 0.00000325, 0.00000155, -0.00001044, 0.00000315, -0.00000301, 0.00000357, -0.00000349, 0.00000201, 0.00000513, -0.00000081, -0.00000529, 0.00000167, 0.00000558, -0.00000398, -0.00000413, 0.00000079, 0.00000432, 0.00000017, -0.00000086, 0.00000002, -0.00000312, -0.00000357, 0.00001639, -0.00001212, 0.00000811, -0.00000402, -0.00000253, 0.00000188, -0.00001591, 0.00000215, -0.00000420, -0.00000451, -0.00000151, -0.00000125, -0.00000544, -0.00000170, -0.00001278, -0.00000737, -0.00001097, -0.00001296, -0.00001661, -0.00002108, -0.00011943, 0.00005220, 0.00010877, 0.00002062, 0.00005516, 0.00001649, 0.00011690, 0.00003631, -0.00013545, -0.00002468, 0.00001031, 0.00002483, 0.00000011, 0.00002256, -0.00001708, 0.00001688, 0.00001063, 0.00002498, 0.00003431, -0.00000338, -0.00004204, },
		{ -0.00000981, -0.00001175, 0.00000328, -0.00001055, -0.00000032, -0.00000549, 0.00000419, -0.00000461, -0.00000758, -0.00000744, 0.00000732, -0.00000318, -0.00001081, 0.00000009, 0.00000853, -0.00000123, -0.00000794, -0.00000190, 0.00001032, -0.00000390, -0.00000264, 0.00000181, -0.00000059, 0.00000345, -0.00000143, 0.00000231, -0.00000503, 0.00000017, 0.00000457, -0.00000053, -0.00000192, 0.00000139, -0.00000555, 0.00000188, 0.00000211, 0.00000020, 0.00000428, -0.00000094, -0.00000892, 0.00000207, 0.00000401, -0.00000031, -0.00000011, -0.00000385, -0.00000267, -0.00000173, 0.00000775, -0.00000270, -0.00000329, -0.00000283, 0.00000528, -0.00000050, 0.00000146, 0.00000130, -0.00000254, -0.00000108, -0.00000189, -0.00000161, 0.00000459, 0.00000010, -0.00000503, -0.00000062, -0.00000132, 0.00000269, 0.00000208, 0.00000357, -0.00001087, 0.00000225, -0.00000259, 0.00000290, -0.00000232, 0.00000014, 0.00000659, -0.00000078, -0.00000473, -0.00000024, 0.00000651, -0.00000315, -0.00000444, -0.00000031, 0.00000512, -0.00000028, -0.00000023, -0.00000069, -0.00000144, -0.00000621, 0.00001681, -0.00000928, 0.00000459, -0.00000007, -0.00000698, 0.00000601, -0.00001761, 0.00000187, -0.00000416, -0.00000376, -0.00000218, -0.00000084, -0.00000662, 0.00000022, -0.00001408, -0.00000596, -0.00001298, -0.00000955, -0.00001999, -0.00001284, -0.00011907, 0.00001928, 0.00012734, 0.00002280, 0.00005841, 0.00001682, 0.00012028, 0.00004281, -0.00013868, -0.00002544, 0.00000694, 0.00002441, -0.00000058, 0.00002199, -0.00001718, 0.00001438, 0.00001185, 0.00001952, 0.00003546, 0.00000253, -0.00003795, -0.00001681, },
		{ -0.00001359, 0.00000434, -0.00001094, 0.00000002, -0.00000630, 0.00000404, -0.00000347, -0.00000780, -0.00000762, 0.00000594, -0.00000063, -0.00001093, -0.00000227, 0.00000894, 0.00000047, -0.00000840, -0.00000283, 0.00000939, -0.00000132, -0.00000446, 0.00000257, -0.00000131, 0.00000392, -0.00000184, 0.00000289, -0.00000439, -0.00000126, 0.00000479, 0.00000016, -0.00000230, 0.00000167, -0.00000540, 0.00000062, 0.00000319, -0.00000054, 0.00000451, 0.00000016, -0.00000854, -0.00000025, 0.00000523, -0.00000065, 0.00000066, -0.00000419, -0.00000212, -0.00000313, 0.00000780, -0.00000129, -0.00000364, -0.00000331, 0.00000467, 0.00000076, 0.00000024, 0.00000246, -0.00000308, -0.00000086, -0.00000183, -0.00000182, 0.00000351, 0.00000203, -0.00000583, -0.00000025, -0.00000208, 0.00000304, 0.00000130, 0.00000483, -0.00001031, 0.00000061, -0.00000158, 0.00000163, -0.00000061, -0.00000202, 0.00000797, -0.00000067, -0.00000399, -0.00000209, 0.00000706, -0.00000217, -0.00000452, -0.00000155, 0.00000584, -0.00000062, 0.00000027, -0.00000124, -0.00000010, -0.00000789, 0.00001569, -0.00000519, 0.00000045, 0.00000402, -0.00001095, 0.00000892, -0.00001778, 0.00000044, -0.00000341, -0.00000367, -0.00000212, -0.00000105, -0.00000725, 0.00000134, -0.00001431, -0.00000570, -0.00001369, -0.00000796, -0.00002071, -0.00000943, -0.00011074, -0.00001768, 0.00014199, 0.00002886, 0.00006010, 0.00001840, 0.00012285, 0.00005017, -0.00014231, -0.00002611, 0.00000361, 0.00002384, -0.00000100, 0.00002108, -0.00001678, 0.00001143, 0.00001360, 0.00001423, 0.00003589, 0.00000718, -0.00003215, -0.00002313, -0.00000706, },
		{ 0.00000403, -0.00001016, -0.00000067, -0.00000600, 0.00000276, -0.00000149, -0.00000865, -0.00000700, 0.00000384, 0.00000203, -0.00001082, -0.00000423, 0.00000833, 0.00000289, -0.00000916, -0.00000320, 0.00000777, 0.00000147, -0.00000604, 0.00000305, -0.00000178, 0.00000414, -0.00000201, 0.00000309, -0.00000338, -0.00000269, 0.00000478, 0.00000089, -0.00000254, 0.00000170, -0.00000487, -0.00000089, 0.00000423, -0.00000110, 0.00000449, 0.00000121, -0.00000778, -0.00000258, 0.00000604, -0.00000060, 0.00000110, -0.00000421, -0.00000186, -0.00000397, 0.00000710, 0.00000055, -0.00000409, -0.00000350, 0.00000359, 0.00000231, -0.00000104, 0.00000350, -0.00000338, -0.00000087, -0.00000169, -0.00000189, 0.00000228, 0.00000371, -0.00000614, -0.00000030, -0.00000242, 0.00000293, 0.00000104, 0.00000524, -0.00000879, -0.00000165, -0.00000015, -0.00000006, 0.00000139, -0.00000417, 0.00000901, -0.00000035, -0.00000323, -0.00000371, 0.00000714, -0.00000102, -0.00000443, -0.00000283, 0.00000635, -0.00000073, 0.00000053, -0.00000151, 0.00000076, -0.00000852, 0.00001324, -0.00000034, -0.00000367, 0.00000754, -0.00001374, 0.00001007, -0.00001617, -0.00000219, -0.00000189, -0.00000436, -0.00000119, -0.00000202, -0.00000712, 0.00000138, -0.00001321, -0.00000686, -0.00001278, -0.00000858, -0.00001841, -0.00001082, -0.00009585, -0.00005561, 0.00015002, 0.00003982, 0.00005990, 0.00002149, 0.00012440, 0.00005857, -0.00014649, -0.00002652, 0.00000017, 0.00002328, -0.00000133, 0.00002003, -0.00001615, 0.00000833, 0.00001549, 0.00000948, 0.00003547, 0.00001085, -0.00002549, -0.00002836, -0.00000603, -0.00001384, },
		{ -0.00000818, -0.00000243, -0.00000460, 0.00000044, 0.00000117, -0.00001002, -0.00000575, 0.00000124, 0.00000461, -0.00001048, -0.00000568, 0.00000675, 0.00000581, -0.00001000, -0.00000323, 0.00000575, 0.00000414, -0.00000713, 0.00000308, -0.00000192, 0.00000406, -0.00000189, 0.00000293, -0.00000209, -0.00000401, 0.00000451, 0.00000164, -0.00000262, 0.00000151, -0.00000405, -0.00000249, 0.00000503, -0.00000132, 0.00000414, 0.00000224, -0.00000679, -0.00000468, 0.00000622, -0.00000008, 0.00000116, -0.00000384, -0.00000197, -0.00000423, 0.00000572, 0.00000264, -0.00000454, -0.00000346, 0.00000217, 0.00000394, -0.00000215, 0.00000420, -0.00000328, -0.00000118, -0.00000143, -0.00000189, 0.00000109, 0.00000495, -0.00000583, -0.00000084, -0.00000228, 0.00000231, 0.00000135, 0.00000483, -0.00000649, -0.00000426, 0.00000145, -0.00000188, 0.00000336, -0.00000599, 0.00000947, 0.00000031, -0.00000258, -0.00000497, 0.00000669, 0.00000028, -0.00000420, -0.00000402, 0.00000656, -0.00000055, 0.00000050, -0.00000146, 0.00000110, -0.00000821, 0.00000982, 0.00000465, -0.00000707, 0.00000984, -0.00001481, 0.00000917, -0.00001278, -0.00000590, 0.00000031, -0.00000578, 0.00000055, -0.00000371, -0.00000621, 0.00000031, -0.00001076, -0.00000938, -0.00001030, -0.00001129, -0.00001336, -0.00001620, -0.00007664, -0.00009105, 0.00014904, 0.00005620, 0.00005783, 0.00002614, 0.00012489, 0.00006808, -0.00015132, -0.00002658, -0.00000353, 0.00002287, -0.00000176, 0.00001907, -0.00001556, 0.00000543, 0.00001711, 0.00000566, 0.00003404, 0.00001395, -0.00001889, -0.00003160, -0.00000727, -0.00001225, 0.00000226, },
		{ -0.00000501, -0.00000236, -0.00000259, 0.00000420, -0.00001163, -0.00000411, -0.00000153, 0.00000684, -0.00000984, -0.00000663, 0.00000438, 0.00000886, -0.00001062, -0.00000318, 0.00000367, 0.00000634, -0.00000751, 0.00000257, -0.00000168, 0.00000366, -0.00000151, 0.00000248, -0.00000066, -0.00000508, 0.00000393, 0.00000241, -0.00000256, 0.00000116, -0.00000307, -0.00000395, 0.00000541, -0.00000109, 0.00000343, 0.00000330, -0.00000571, -0.00000633, 0.00000568, 0.00000093, 0.00000084, -0.00000311, -0.00000243, -0.00000397, 0.00000385, 0.00000476, -0.00000482, -0.00000331, 0.00000060, 0.00000541, -0.00000287, 0.00000441, -0.00000270, -0.00000180, -0.00000103, -0.00000191, 0.00000008, 0.00000560, -0.00000485, -0.00000184, -0.00000170, 0.00000125, 0.00000213, 0.00000378, -0.00000371, -0.00000687, 0.00000289, -0.00000351, 0.00000497, -0.00000717, 0.00000917, 0.00000141, -0.00000211, -0.00000575, 0.00000569, 0.00000169, -0.00000386, -0.00000505, 0.00000640, -0.00000008, 0.00000019, -0.00000111, 0.00000100, -0.00000717, 0.00000592, 0.00000912, -0.00000916, 0.00001044, -0.00001388, 0.00000625, -0.00000792, -0.00001031, 0.00000288, -0.00000769, 0.00000286, -0.00000589, -0.00000471, -0.00000168, -0.00000727, -0.00001292, -0.00000667, -0.00001556, -0.00000642, -0.00002405, -0.00005594, -0.00012057, 0.00013726, 0.00007778, 0.00005441, 0.00003208, 0.00012453, 0.00007860, -0.00015673, -0.00002627, -0.00000752, 0.00002271, -0.00000242, 0.00001839, -0.00001526, 0.00000302, 0.00001809, 0.00000306, 0.00003144, 0.00001689, -0.00001317, -0.00003223, -0.00001093, -0.00000894, -0.00000076, -0.00000523, },
		{ 0.00000028, -0.00000587, 0.00000713, -0.00001309, -0.00000248, -0.00000409, 0.00000845, -0.00000880, -0.00000717, 0.00000153, 0.00001165, -0.00001069, -0.00000331, 0.00000184, 0.00000779, -0.00000705, 0.00000153, -0.00000110, 0.00000301, -0.00000094, 0.00000185, 0.00000075, -0.00000580, 0.00000304, 0.00000317, -0.00000238, 0.00000072, -0.00000210, -0.00000508, 0.00000522, -0.00000035, 0.00000237, 0.00000436, -0.00000463, -0.00000743, 0.00000442, 0.00000233, 0.00000025, -0.00000211, -0.00000312, -0.00000334, 0.00000173, 0.00000665, -0.00000478, -0.00000319, -0.00000093, 0.00000650, -0.00000302, 0.00000403, -0.00000163, -0.00000270, -0.00000055, -0.00000197, -0.00000065, 0.00000563, -0.00000330, -0.00000318, -0.00000083, -0.00000009, 0.00000319, 0.00000236, -0.00000084, -0.00000910, 0.00000384, -0.00000466, 0.00000591, -0.00000748, 0.00000803, 0.00000291, -0.00000186, -0.00000602, 0.00000421, 0.00000311, -0.00000341, -0.00000584, 0.00000586, 0.00000065, -0.00000032, -0.00000054, 0.00000058, -0.00000568, 0.00000204, 0.00001251, -0.00000950, 0.00000914, -0.00001102, 0.00000167, -0.00000216, -0.00001482, 0.00000535, -0.00000970, 0.00000535, -0.00000817, -0.00000299, -0.00000416, -0.00000327, -0.00001682, -0.00000262, -0.00002048, 0.00000118, -0.00003247, -0.00003669, -0.00014120, 0.00011393, 0.00010341, 0.00005071, 0.00003869, 0.00012374, 0.00008984, -0.00016255, -0.00002573, -0.00001174, 0.00002275, -0.00000334, 0.00001808, -0.00001542, 0.00000131, 0.00001820, 0.00000186, 0.00002760, 0.00002000, -0.00000887, -0.00003005, -0.00001672, -0.00000441, -0.00000452, -0.00000179, -0.00000799, },
		{ -0.00000883, 0.00000946, -0.00001399, -0.00000122, -0.00000605, 0.00000915, -0.00000725, -0.00000747, -0.00000142, 0.00001373, -0.00000994, -0.00000384, 0.00000050, 0.00000836, -0.00000577, 0.00000005, -0.00000030, 0.00000221, -0.00000030, 0.00000117, 0.00000200, -0.00000608, 0.00000183, 0.00000390, -0.00000212, 0.00000028, -0.00000126, -0.00000574, 0.00000441, 0.00000085, 0.00000107, 0.00000537, -0.00000358, -0.00000794, 0.00000254, 0.00000394, -0.00000043, -0.00000102, -0.00000387, -0.00000254, -0.00000037, 0.00000805, -0.00000426, -0.00000320, -0.00000223, 0.00000701, -0.00000250, 0.00000306, -0.00000016, -0.00000373, -0.00000006, -0.00000206, -0.00000110, 0.00000510, -0.00000133, -0.00000462, 0.00000011, -0.00000148, 0.00000427, 0.00000091, 0.00000172, -0.00001058, 0.00000405, -0.00000509, 0.00000601, -0.00000683, 0.00000607, 0.00000473, -0.00000178, -0.00000583, 0.00000236, 0.00000444, -0.00000284, -0.00000638, 0.00000499, 0.00000155, -0.00000094, 0.00000011, 0.00000003, -0.00000408, -0.00000135, 0.00001441, -0.00000793, 0.00000606, -0.00000663, -0.00000388, 0.00000371, -0.00001874, 0.00000719, -0.00001132, 0.00000753, -0.00001006, -0.00000154, -0.00000658, 0.00000057, -0.00002032, 0.00000098, -0.00002497, 0.00000800, -0.00003949, -0.00002152, -0.00015096, 0.00007959, 0.00013095, 0.00004831, 0.00004509, 0.00012316, 0.00010138, -0.00016841, -0.00002522, -0.00001600, 0.00002288, -0.00000445, 0.00001813, -0.00001605, 0.00000034, 0.00001739, 0.00000202, 0.00002267, 0.00002337, -0.00000619, -0.00002534, -0.00002392, 0.00000052, -0.00000827, 0.00000150, -0.00001078, 0.00000276, },
		{ 0.00001073, -0.00001397, -0.00000069, -0.00000710, 0.00000877, -0.00000514, -0.00000769, -0.00000411, 0.00001478, -0.00000820, -0.00000485, -0.00000025, 0.00000801, -0.00000381, -0.00000168, 0.00000057, 0.00000141, 0.00000028, 0.00000059, 0.00000297, -0.00000588, 0.00000037, 0.00000456, -0.00000179, -0.00000010, -0.00000066, -0.00000585, 0.00000302, 0.00000237, -0.00000032, 0.00000620, -0.00000255, -0.00000792, 0.00000025, 0.00000551, -0.00000100, -0.00000002, -0.00000451, -0.00000181, -0.00000217, 0.00000877, -0.00000318, -0.00000341, -0.00000319, 0.00000686, -0.00000131, 0.00000160, 0.00000156, -0.00000475, 0.00000032, -0.00000212, -0.00000132, 0.00000415, 0.00000080, -0.00000592, 0.00000088, -0.00000266, 0.00000511, -0.00000026, 0.00000364, -0.00001107, 0.00000334, -0.00000469, 0.00000520, -0.00000528, 0.00000350, 0.00000665, -0.00000176, -0.00000526, 0.00000030, 0.00000554, -0.00000213, -0.00000668, 0.00000386, 0.00000250, -0.00000154, 0.00000071, -0.00000047, -0.00000262, -0.00000390, 0.00001461, -0.00000458, 0.00000163, -0.00000139, -0.00000955, 0.00000879, -0.00002135, 0.00000785, -0.00001210, 0.00000891, -0.00001107, -0.00000086, -0.00000835, 0.00000354, -0.00002263, 0.00000328, -0.00002800, 0.00001275, -0.00004342, -0.00001237, -0.00014911, 0.00003619, 0.00015737, 0.00004913, 0.00005023, 0.00012357, 0.00011267, -0.00017391, -0.00002508, -0.00002001, 0.00002285, -0.00000558, 0.00001838, -0.00001705, -0.00000001, 0.00001579, 0.00000331, 0.00001697, 0.00002686, -0.00000493, -0.00001879, -0.00003150, 0.00000487, -0.00001117, 0.00000394, -0.00001279, 0.00000451, -0.00001093, },
		{ -0.00001283, -0.00000107, -0.00000705, 0.00000726, -0.00000248, -0.00000791, -0.00000628, 0.00001458, -0.00000547, -0.00000629, -0.00000044, 0.00000689, -0.00000139, -0.00000342, 0.00000130, 0.00000075, 0.00000069, 0.00000019, 0.00000357, -0.00000521, -0.00000125, 0.00000509, -0.00000140, -0.00000038, -0.00000034, -0.00000543, 0.00000118, 0.00000400, -0.00000157, 0.00000672, -0.00000147, -0.00000750, -0.00000219, 0.00000677, -0.00000124, 0.00000069, -0.00000484, -0.00000133, -0.00000347, 0.00000869, -0.00000157, -0.00000381, -0.00000375, 0.00000604, 0.00000041, -0.00000014, 0.00000327, -0.00000553, 0.00000045, -0.00000209, -0.00000138, 0.00000297, 0.00000283, -0.00000681, 0.00000125, -0.00000343, 0.00000547, -0.00000089, 0.00000467, -0.00001043, 0.00000171, -0.00000350, 0.00000360, -0.00000302, 0.00000058, 0.00000843, -0.00000165, -0.00000445, -0.00000175, 0.00000627, -0.00000124, -0.00000678, 0.00000258, 0.00000335, -0.00000200, 0.00000115, -0.00000078, -0.00000153, -0.00000543, 0.00001317, 0.00000014, -0.00000347, 0.00000383, -0.00001437, 0.00001225, -0.00002205, 0.00000695, -0.00001168, 0.00000912, -0.00001083, -0.00000131, -0.00000901, 0.00000508, -0.00002316, 0.00000361, -0.00002880, 0.00001450, -0.00004318, -0.00001016, -0.00013634, -0.00001303, 0.00017902, 0.00005516, 0.00005314, 0.00012567, 0.00012319, -0.00017860, -0.00002569, -0.00002344, 0.00002236, -0.00000647, 0.00001862, -0.00001817, 0.00000001, 0.00001372, 0.00000530, 0.00001098, 0.00003009, -0.00000461, -0.00001141, -0.00003824, 0.00000759, -0.00001245, 0.00000494, -0.00001350, 0.00000507, -0.00001174, 0.00001063, },
		{ -0.00000240, -0.00000591, 0.00000472, 0.00000056, -0.00000817, -0.00000777, 0.00001312, -0.00000196, -0.00000795, -0.00000023, 0.00000522, 0.00000116, -0.00000489, 0.00000171, 0.00000036, 0.00000083, 0.00000006, 0.00000379, -0.00000414, -0.00000291, 0.00000539, -0.00000092, -0.00000056, -0.00000030, -0.00000457, -0.00000089, 0.00000546, -0.00000245, 0.00000678, -0.00000029, -0.00000681, -0.00000452, 0.00000746, -0.00000099, 0.00000100, -0.00000477, -0.00000123, -0.00000415, 0.00000780, 0.00000049, -0.00000434, -0.00000395, 0.00000466, 0.00000246, -0.00000190, 0.00000472, -0.00000587, 0.00000022, -0.00000189, -0.00000139, 0.00000178, 0.00000448, -0.00000708, 0.00000108, -0.00000364, 0.00000522, -0.00000084, 0.00000472, -0.00000872, -0.00000067, -0.00000173, 0.00000144, -0.00000039, -0.00000228, 0.00000974, -0.00000128, -0.00000358, -0.00000361, 0.00000651, -0.00000015, -0.00000670, 0.00000127, 0.00000400, -0.00000222, 0.00000135, -0.00000082, -0.00000091, -0.00000592, 0.00001038, 0.00000561, -0.00000842, 0.00000813, -0.00001750, 0.00001345, -0.00002050, 0.00000433, -0.00000993, 0.00000801, -0.00000919, -0.00000308, -0.00000832, 0.00000489, -0.00002161, 0.00000172, -0.00002705, 0.00001288, -0.00003849, -0.00001464, -0.00011476, -0.00006379, 0.00019205, 0.00006808, 0.00005314, 0.00012997, 0.00013261, -0.00018217, -0.00002736, -0.00002598, 0.00002114, -0.00000686, 0.00001857, -0.00001913, 0.00000006, 0.00001157, 0.00000748, 0.00000531, 0.00003257, -0.00000453, -0.00000430, -0.00004301, 0.00000788, -0.00001158, 0.00000413, -0.00001265, 0.00000421, -0.00001109, 0.00000907, -0.00001053, },
		{ -0.00000387, 0.00000144, 0.00000372, -0.00000837, -0.00000857, 0.00001057, 0.00000200, -0.00000951, 0.00000010, 0.00000333, 0.00000350, -0.00000582, 0.00000166, 0.00000032, 0.00000066, 0.00000020, 0.00000366, -0.00000281, -0.00000445, 0.00000540, -0.00000035, -0.00000064, -0.00000048, -0.00000343, -0.00000292, 0.00000651, -0.00000278, 0.00000630, 0.00000105, -0.00000601, -0.00000649, 0.00000742, -0.00000017, 0.00000085, -0.00000424, -0.00000156, -0.00000420, 0.00000621, 0.00000280, -0.00000485, -0.00000389, 0.00000290, 0.00000456, -0.00000337, 0.00000565, -0.00000561, -0.00000044, -0.00000149, -0.00000145, 0.00000074, 0.00000557, -0.00000661, 0.00000029, -0.00000326, 0.00000434, -0.00000013, 0.00000388, -0.00000616, -0.00000351, 0.00000031, -0.00000092, 0.00000222, -0.00000470, 0.00001033, -0.00000051, -0.00000279, -0.00000509, 0.00000616, 0.00000113, -0.00000651, 0.00000004, 0.00000434, -0.00000217, 0.00000129, -0.00000058, -0.00000074, -0.00000551, 0.00000666, 0.00001112, -0.00001239, 0.00001072, -0.00001833, 0.00001206, -0.00001670, 0.00000013, -0.00000697, 0.00000566, -0.00000624, -0.00000608, -0.00000632, 0.00000301, -0.00001809, -0.00000225, -0.00002294, 0.00000819, -0.00002993, -0.00002449, -0.00008757, -0.00011127, 0.00019296, 0.00008884, 0.00005012, 0.00013659, 0.00014085, -0.00018453, -0.00003025, -0.00002749, 0.00001902, -0.00000658, 0.00001802, -0.00001966, -0.00000022, 0.00000977, 0.00000929, 0.00000056, 0.00003377, -0.00000397, 0.00000150, -0.00004491, 0.00000524, -0.00000839, 0.00000150, -0.00001026, 0.00000195, -0.00000905, 0.00000620, -0.00000729, -0.00000449, },
		{ -0.00000215, 0.00000662, -0.00000834, -0.00000881, 0.00000726, 0.00000593, -0.00001060, 0.00000028, 0.00000154, 0.00000531, -0.00000605, 0.00000107, 0.00000064, 0.00000020, 0.00000056, 0.00000329, -0.00000137, -0.00000572, 0.00000503, 0.00000034, -0.00000067, -0.00000080, -0.00000219, -0.00000466, 0.00000690, -0.00000242, 0.00000525, 0.00000254, -0.00000520, -0.00000794, 0.00000657, 0.00000120, 0.00000029, -0.00000332, -0.00000224, -0.00000374, 0.00000415, 0.00000510, -0.00000516, -0.00000370, 0.00000099, 0.00000642, -0.00000428, 0.00000587, -0.00000468, -0.00000151, -0.00000091, -0.00000161, -0.00000002, 0.00000599, -0.00000539, -0.00000104, -0.00000238, 0.00000296, 0.00000112, 0.00000238, -0.00000311, -0.00000640, 0.00000222, -0.00000308, 0.00000437, -0.00000630, 0.00000999, 0.00000073, -0.00000216, -0.00000608, 0.00000522, 0.00000256, -0.00000621, -0.00000103, 0.00000433, -0.00000184, 0.00000099, -0.00000010, -0.00000094, -0.00000446, 0.00000258, 0.00001591, -0.00001469, 0.00001108, -0.00001659, 0.00000820, -0.00001103, -0.00000518, -0.00000323, 0.00000244, -0.00000237, -0.00000992, -0.00000336, -0.00000017, -0.00001309, -0.00000767, -0.00001719, 0.00000136, -0.00001881, -0.00003748, -0.00005868, -0.00015072, 0.00017914, 0.00011735, 0.00004474, 0.00014513, 0.00014824, -0.00018585, -0.00003425, -0.00002798, 0.00001597, -0.00000556, 0.00001685, -0.00001958, -0.00000108, 0.00000868, 0.00001031, -0.00000281, 0.00003328, -0.00000234, 0.00000522, -0.00004350, -0.00000029, -0.00000317, -0.00000260, -0.00000668, -0.00000135, -0.00000597, 0.00000245, -0.00000353, -0.00000697, -0.00000131, },
		{ 0.00000886, -0.00000785, -0.00000871, 0.00000363, 0.00000932, -0.00001083, 0.00000004, 0.00000014, 0.00000636, -0.00000546, -0.00000002, 0.00000127, -0.00000047, 0.00000106, 0.00000279, 0.00000002, -0.00000657, 0.00000425, 0.00000114, -0.00000067, -0.00000117, -0.00000103, -0.00000588, 0.00000650, -0.00000136, 0.00000371, 0.00000411, -0.00000443, -0.00000880, 0.00000496, 0.00000295, -0.00000051, -0.00000216, -0.00000314, -0.00000294, 0.00000188, 0.00000711, -0.00000509, -0.00000355, -0.00000084, 0.00000775, -0.00000445, 0.00000531, -0.00000311, -0.00000289, -0.00000023, -0.00000186, -0.00000046, 0.00000575, -0.00000356, -0.00000271, -0.00000120, 0.00000130, 0.00000262, 0.00000055, -0.00000000, -0.00000888, 0.00000360, -0.00000465, 0.00000572, -0.00000683, 0.00000862, 0.00000242, -0.00000175, -0.00000650, 0.00000371, 0.00000405, -0.00000581, -0.00000188, 0.00000397, -0.00000129, 0.00000053, 0.00000053, -0.00000136, -0.00000307, -0.00000134, 0.00001939, -0.00001486, 0.00000903, -0.00001245, 0.00000237, -0.00000422, -0.00001085, 0.00000064, -0.00000104, 0.00000182, -0.00001398, -0.00000007, -0.00000397, -0.00000745, -0.00001360, -0.00001088, -0.00000628, -0.00000698, -0.00005094, -0.00003212, -0.00017813, 0.00014944, 0.00015214, 0.00003851, 0.00015465, 0.00015553, -0.00018664, -0.00003904, -0.00002771, 0.00001220, -0.00000394, 0.00001511, -0.00001886, -0.00000264, 0.00000847, 0.00001026, -0.00000452, 0.00003092, 0.00000066, 0.00000654, -0.00003885, -0.00000818, 0.00000339, -0.00000746, -0.00000255, -0.00000513, -0.00000244, -0.00000159, 0.00000018, -0.00000935, 0.00000131, -0.00000553, },
		{ -0.00000669, -0.00000854, 0.00000013, 0.00001168, -0.00000994, -0.00000079, -0.00000070, 0.00000655, -0.00000410, -0.00000152, 0.00000207, -0.00000123, 0.00000157, 0.00000230, 0.00000119, -0.00000693, 0.00000307, 0.00000203, -0.00000065, -0.00000149, -0.00000010, -0.00000644, 0.00000529, 0.00000030, 0.00000183, 0.00000563, -0.00000366, -0.00000907, 0.00000276, 0.00000488, -0.00000136, -0.00000095, -0.00000405, -0.00000204, -0.00000029, 0.00000855, -0.00000448, -0.00000356, -0.00000238, 0.00000837, -0.00000377, 0.00000399, -0.00000105, -0.00000442, 0.00000044, -0.00000215, -0.00000060, 0.00000497, -0.00000133, -0.00000447, 0.00000001, -0.00000037, 0.00000408, -0.00000121, 0.00000271, -0.00001055, 0.00000413, -0.00000534, 0.00000602, -0.00000616, 0.00000629, 0.00000444, -0.00000151, -0.00000640, 0.00000176, 0.00000549, -0.00000529, -0.00000249, 0.00000328, -0.00000060, 0.00000002, 0.00000118, -0.00000183, -0.00000166, -0.00000463, 0.00002114, -0.00001277, 0.00000479, -0.00000648, -0.00000455, 0.00000272, -0.00001596, 0.00000388, -0.00000407, 0.00000560, -0.00001755, 0.00000281, -0.00000754, -0.00000215, -0.00001890, -0.00000529, -0.00001315, 0.00000357, -0.00006214, -0.00001143, -0.00019075, 0.00010445, 0.00019033, 0.00003373, 0.00016374, 0.00016379, -0.00018769, -0.00004404, -0.00002712, 0.00000806, -0.00000202, 0.00001302, -0.00001761, -0.00000484, 0.00000914, 0.00000912, -0.00000459, 0.00002680, 0.00000497, 0.00000559, -0.00003159, -0.00001742, 0.00001021, -0.00001217, 0.00000133, -0.00000863, 0.00000084, -0.00000521, 0.00000320, -0.00001110, 0.00000346, -0.00000822, 0.00001005, },
		{ -0.00000851, -0.00000283, 0.00001266, -0.00000781, -0.00000224, -0.00000093, 0.00000589, -0.00000212, -0.00000322, 0.00000289, -0.00000194, 0.00000197, 0.00000196, 0.00000203, -0.00000674, 0.00000155, 0.00000293, -0.00000062, -0.00000173, 0.00000049, -0.00000629, 0.00000336, 0.00000235, -0.00000011, 0.00000690, -0.00000283, -0.00000888, 0.00000021, 0.00000667, -0.00000200, 0.00000009, -0.00000475, -0.00000129, -0.00000208, 0.00000922, -0.00000328, -0.00000379, -0.00000350, 0.00000819, -0.00000229, 0.00000208, 0.00000127, -0.00000584, 0.00000091, -0.00000238, -0.00000054, 0.00000382, 0.00000099, -0.00000600, 0.00000095, -0.00000174, 0.00000518, -0.00000255, 0.00000466, -0.00001111, 0.00000359, -0.00000503, 0.00000521, -0.00000437, 0.00000322, 0.00000655, -0.00000132, -0.00000584, -0.00000047, 0.00000672, -0.00000462, -0.00000288, 0.00000235, 0.00000010, -0.00000044, 0.00000173, -0.00000220, -0.00000047, -0.00000698, 0.00002102, -0.00000864, -0.00000105, 0.00000040, -0.00001145, 0.00000870, -0.00001959, 0.00000572, -0.00000597, 0.00000827, -0.00001989, 0.00000455, -0.00001005, 0.00000182, -0.00002249, -0.00000163, -0.00001784, 0.00001106, -0.00006884, 0.00000083, -0.00018762, 0.00004675, 0.00022774, 0.00003327, 0.00017074, 0.00017429, -0.00018989, -0.00004857, -0.00002679, 0.00000406, -0.00000022, 0.00001091, -0.00001613, -0.00000745, 0.00001048, 0.00000710, -0.00000332, 0.00002135, 0.00001020, 0.00000296, -0.00002277, -0.00002670, 0.00001608, -0.00001569, 0.00000412, -0.00001114, 0.00000317, -0.00000778, 0.00000499, -0.00001176, 0.00000471, -0.00000978, 0.00000992, -0.00000472, },
		{ -0.00000496, 0.00001210, -0.00000450, -0.00000416, -0.00000067, 0.00000457, 0.00000023, -0.00000486, 0.00000352, -0.00000247, 0.00000214, 0.00000184, 0.00000248, -0.00000605, -0.00000021, 0.00000374, -0.00000053, -0.00000187, 0.00000074, -0.00000550, 0.00000092, 0.00000449, -0.00000184, 0.00000772, -0.00000183, -0.00000837, -0.00000241, 0.00000804, -0.00000221, 0.00000077, -0.00000508, -0.00000087, -0.00000330, 0.00000903, -0.00000148, -0.00000424, -0.00000416, 0.00000722, -0.00000017, -0.00000014, 0.00000354, -0.00000688, 0.00000103, -0.00000245, -0.00000040, 0.00000255, 0.00000310, -0.00000701, 0.00000138, -0.00000256, 0.00000565, -0.00000320, 0.00000558, -0.00001042, 0.00000198, -0.00000375, 0.00000341, -0.00000171, -0.00000024, 0.00000846, -0.00000102, -0.00000499, -0.00000276, 0.00000760, -0.00000376, -0.00000307, 0.00000129, 0.00000071, -0.00000075, 0.00000212, -0.00000236, 0.00000033, -0.00000826, 0.00001915, -0.00000299, -0.00000764, 0.00000712, -0.00001718, 0.00001269, -0.00002096, 0.00000560, -0.00000620, 0.00000929, -0.00002048, 0.00000458, -0.00001084, 0.00000372, -0.00002356, -0.00000077, -0.00001932, 0.00001428, -0.00006966, 0.00000352, -0.00016971, -0.00001927, 0.00025925, 0.00004013, 0.00017401, 0.00018826, -0.00019412, -0.00005196, -0.00002731, 0.00000073, 0.00000099, 0.00000920, -0.00001478, -0.00001013, 0.00001211, 0.00000460, -0.00000122, 0.00001522, 0.00001564, -0.00000045, -0.00001373, -0.00003460, 0.00001982, -0.00001714, 0.00000512, -0.00001205, 0.00000403, -0.00000879, 0.00000516, -0.00001106, 0.00000481, -0.00000997, 0.00000836, -0.00000195, -0.00000876, },
		{ 0.00001007, -0.00000032, -0.00000625, -0.00000014, 0.00000284, 0.00000262, -0.00000619, 0.00000380, -0.00000269, 0.00000202, 0.00000200, 0.00000256, -0.00000495, -0.00000202, 0.00000435, -0.00000033, -0.00000193, 0.00000067, -0.00000421, -0.00000173, 0.00000638, -0.00000305, 0.00000791, -0.00000057, -0.00000771, -0.00000480, 0.00000873, -0.00000182, 0.00000095, -0.00000492, -0.00000091, -0.00000383, 0.00000798, 0.00000077, -0.00000481, -0.00000440, 0.00000561, 0.00000231, -0.00000234, 0.00000541, -0.00000732, 0.00000066, -0.00000227, -0.00000033, 0.00000138, 0.00000472, -0.00000727, 0.00000111, -0.00000270, 0.00000538, -0.00000303, 0.00000540, -0.00000855, -0.00000053, -0.00000172, 0.00000091, 0.00000144, -0.00000363, 0.00000982, -0.00000046, -0.00000400, -0.00000490, 0.00000797, -0.00000268, -0.00000311, 0.00000019, 0.00000112, -0.00000085, 0.00000228, -0.00000226, 0.00000068, -0.00000850, 0.00001590, 0.00000342, -0.00001396, 0.00001256, -0.00002071, 0.00001392, -0.00001962, 0.00000324, -0.00000454, 0.00000841, -0.00001907, 0.00000266, -0.00000959, 0.00000316, -0.00002175, -0.00000305, -0.00001719, 0.00001281, -0.00006431, -0.00000293, -0.00013989, -0.00008778, 0.00027939, 0.00005697, 0.00017239, 0.00020659, -0.00020099, -0.00005375, -0.00002912, -0.00000152, 0.00000121, 0.00000823, -0.00001390, -0.00001249, 0.00001362, 0.00000212, 0.00000108, 0.00000919, 0.00002047, -0.00000357, -0.00000581, -0.00003985, 0.00002054, -0.00001596, 0.00000394, -0.00001108, 0.00000314, -0.00000805, 0.00000361, -0.00000896, 0.00000371, -0.00000880, 0.00000550, 0.00000146, -0.00000927, -0.00000616, },
		{ 0.00000426, -0.00000811, 0.00000036, 0.00000104, 0.00000471, -0.00000695, 0.00000359, -0.00000255, 0.00000159, 0.00000242, 0.00000231, -0.00000359, -0.00000373, 0.00000463, 0.00000003, -0.00000195, 0.00000037, -0.00000262, -0.00000426, 0.00000768, -0.00000351, 0.00000736, 0.00000099, -0.00000704, -0.00000673, 0.00000858, -0.00000079, 0.00000063, -0.00000426, -0.00000142, -0.00000372, 0.00000623, 0.00000326, -0.00000535, -0.00000435, 0.00000357, 0.00000481, -0.00000416, 0.00000659, -0.00000696, -0.00000026, -0.00000181, -0.00000041, 0.00000049, 0.00000566, -0.00000667, 0.00000013, -0.00000213, 0.00000438, -0.00000209, 0.00000425, -0.00000574, -0.00000361, 0.00000068, -0.00000187, 0.00000458, -0.00000650, 0.00001033, 0.00000053, -0.00000304, -0.00000668, 0.00000774, -0.00000138, -0.00000303, -0.00000084, 0.00000125, -0.00000070, 0.00000222, -0.00000192, 0.00000064, -0.00000790, 0.00001179, 0.00000976, -0.00001905, 0.00001582, -0.00002135, 0.00001204, -0.00001555, -0.00000122, -0.00000113, 0.00000576, -0.00001581, -0.00000108, -0.00000642, 0.00000026, -0.00001725, -0.00000820, -0.00001180, 0.00000716, -0.00005368, -0.00001661, -0.00010255, -0.00015219, 0.00028311, 0.00008548, 0.00016549, 0.00022959, -0.00021066, -0.00005382, -0.00003237, -0.00000248, 0.00000023, 0.00000823, -0.00001375, -0.00001423, 0.00001461, 0.00000016, 0.00000295, 0.00000402, 0.00002389, -0.00000543, -0.00000015, -0.00004155, 0.00001780, -0.00001204, 0.00000063, -0.00000827, 0.00000060, -0.00000567, 0.00000053, -0.00000571, 0.00000163, -0.00000649, 0.00000167, 0.00000518, -0.00000989, -0.00000651, 0.00000684, },
		{ -0.00000928, 0.00000049, -0.00000050, 0.00000622, -0.00000698, 0.00000284, -0.00000205, 0.00000087, 0.00000304, 0.00000185, -0.00000215, -0.00000513, 0.00000449, 0.00000059, -0.00000198, -0.00000005, -0.00000097, -0.00000637, 0.00000814, -0.00000309, 0.00000606, 0.00000282, -0.00000644, -0.00000807, 0.00000755, 0.00000083, -0.00000012, -0.00000318, -0.00000229, -0.00000309, 0.00000402, 0.00000569, -0.00000566, -0.00000416, 0.00000139, 0.00000700, -0.00000529, 0.00000688, -0.00000575, -0.00000170, -0.00000111, -0.00000066, -0.00000003, 0.00000585, -0.00000525, -0.00000147, -0.00000101, 0.00000280, -0.00000056, 0.00000241, -0.00000241, -0.00000678, 0.00000299, -0.00000445, 0.00000722, -0.00000839, 0.00000976, 0.00000201, -0.00000222, -0.00000794, 0.00000685, 0.00000011, -0.00000286, -0.00000171, 0.00000103, -0.00000032, 0.00000199, -0.00000140, 0.00000029, -0.00000673, 0.00000740, 0.00001522, -0.00002212, 0.00001631, -0.00001886, 0.00000722, -0.00000920, -0.00000721, 0.00000349, 0.00000187, -0.00001124, -0.00000608, -0.00000188, -0.00000437, -0.00001081, -0.00001533, -0.00000420, -0.00000134, -0.00003967, -0.00003445, -0.00006302, -0.00020593, 0.00026650, 0.00012587, 0.00015406, 0.00025680, -0.00022265, -0.00005255, -0.00003684, -0.00000228, -0.00000187, 0.00000916, -0.00001438, -0.00001522, 0.00001484, -0.00000096, 0.00000391, 0.00000030, 0.00002531, -0.00000534, 0.00000253, -0.00003939, 0.00001175, -0.00000577, -0.00000436, -0.00000409, -0.00000318, -0.00000211, -0.00000356, -0.00000181, -0.00000098, -0.00000349, -0.00000262, 0.00000874, -0.00001033, -0.00000626, 0.00000287, 0.00000868, },
		{ -0.00000002, -0.00000153, 0.00000694, -0.00000621, 0.00000159, -0.00000126, -0.00000003, 0.00000376, 0.00000131, -0.00000080, -0.00000609, 0.00000387, 0.00000135, -0.00000204, -0.00000049, 0.00000053, -0.00000780, 0.00000762, -0.00000178, 0.00000413, 0.00000481, -0.00000590, -0.00000878, 0.00000574, 0.00000285, -0.00000111, -0.00000187, -0.00000335, -0.00000216, 0.00000164, 0.00000777, -0.00000553, -0.00000401, -0.00000068, 0.00000856, -0.00000551, 0.00000620, -0.00000375, -0.00000350, -0.00000029, -0.00000106, -0.00000017, 0.00000535, -0.00000317, -0.00000345, 0.00000042, 0.00000093, 0.00000123, 0.00000028, 0.00000094, -0.00000953, 0.00000473, -0.00000636, 0.00000892, -0.00000901, 0.00000799, 0.00000399, -0.00000160, -0.00000859, 0.00000531, 0.00000173, -0.00000263, -0.00000235, 0.00000047, 0.00000026, 0.00000163, -0.00000080, -0.00000020, -0.00000531, 0.00000328, 0.00001918, -0.00002272, 0.00001390, -0.00001352, 0.00000014, -0.00000149, -0.00001379, 0.00000846, -0.00000245, -0.00000620, -0.00001147, 0.00000315, -0.00000973, -0.00000361, -0.00002307, 0.00000406, -0.00001075, -0.00002485, -0.00005273, -0.00002681, -0.00024337, 0.00022757, 0.00017645, 0.00014011, 0.00028693, -0.00023589, -0.00005080, -0.00004189, -0.00000141, -0.00000471, 0.00001073, -0.00001559, -0.00001553, 0.00001429, -0.00000110, 0.00000373, -0.00000168, 0.00002446, -0.00000300, 0.00000208, -0.00003370, 0.00000316, 0.00000191, -0.00001015, 0.00000067, -0.00000745, 0.00000193, -0.00000795, 0.00000204, -0.00000351, -0.00000040, -0.00000673, 0.00001160, -0.00001026, -0.00000573, -0.00000128, 0.00001229, -0.00000933, },
		{ -0.00000190, 0.00000678, -0.00000467, -0.00000006, -0.00000030, -0.00000101, 0.00000445, 0.00000082, 0.00000031, -0.00000652, 0.00000279, 0.00000226, -0.00000213, -0.00000088, 0.00000173, -0.00000841, 0.00000612, 0.00000026, 0.00000180, 0.00000673, -0.00000534, -0.00000893, 0.00000334, 0.00000499, -0.00000208, -0.00000057, -0.00000438, -0.00000120, -0.00000056, 0.00000922, -0.00000483, -0.00000404, -0.00000242, 0.00000929, -0.00000474, 0.00000461, -0.00000117, -0.00000543, 0.00000049, -0.00000150, -0.00000001, 0.00000433, -0.00000074, -0.00000547, 0.00000183, -0.00000088, 0.00000289, -0.00000169, 0.00000381, -0.00001139, 0.00000550, -0.00000725, 0.00000938, -0.00000820, 0.00000510, 0.00000632, -0.00000117, -0.00000861, 0.00000321, 0.00000337, -0.00000231, -0.00000273, -0.00000043, 0.00000097, 0.00000123, -0.00000018, -0.00000072, -0.00000391, -0.00000014, 0.00002127, -0.00002079, 0.00000894, -0.00000611, -0.00000811, 0.00000635, -0.00001982, 0.00001273, -0.00000620, -0.00000171, -0.00001620, 0.00000758, -0.00001460, 0.00000295, -0.00002983, 0.00001116, -0.00001889, -0.00001199, -0.00006771, 0.00000120, -0.00026058, 0.00016678, 0.00023348, 0.00012688, 0.00031793, -0.00024874, -0.00004988, -0.00004653, -0.00000066, -0.00000762, 0.00001237, -0.00001694, -0.00001549, 0.00001318, -0.00000040, 0.00000247, -0.00000199, 0.00002151, 0.00000136, -0.00000105, -0.00002539, -0.00000671, 0.00000972, -0.00001559, 0.00000504, -0.00001134, 0.00000558, -0.00001179, 0.00000510, -0.00000533, 0.00000215, -0.00001003, 0.00001326, -0.00000934, -0.00000532, -0.00000505, 0.00001457, -0.00000796, -0.00000132, },
		{ 0.00000578, -0.00000252, -0.00000191, 0.00000067, -0.00000194, 0.00000502, 0.00000049, 0.00000110, -0.00000640, 0.00000132, 0.00000324, -0.00000221, -0.00000119, 0.00000254, -0.00000820, 0.00000379, 0.00000274, -0.00000060, 0.00000835, -0.00000464, -0.00000867, 0.00000066, 0.00000692, -0.00000277, 0.00000050, -0.00000513, -0.00000045, -0.00000232, 0.00000984, -0.00000348, -0.00000431, -0.00000369, 0.00000911, -0.00000305, 0.00000235, 0.00000168, -0.00000717, 0.00000100, -0.00000185, 0.00000029, 0.00000302, 0.00000172, -0.00000716, 0.00000286, -0.00000227, 0.00000405, -0.00000310, 0.00000576, -0.00001200, 0.00000505, -0.00000694, 0.00000849, -0.00000601, 0.00000131, 0.00000878, -0.00000084, -0.00000806, 0.00000070, 0.00000491, -0.00000192, -0.00000282, -0.00000162, 0.00000174, 0.00000085, 0.00000039, -0.00000116, -0.00000271, -0.00000262, 0.00002142, -0.00001665, 0.00000221, 0.00000222, -0.00001613, 0.00001294, -0.00002407, 0.00001527, -0.00000838, 0.00000121, -0.00001925, 0.00001032, -0.00001780, 0.00000749, -0.00003410, 0.00001542, -0.00002376, -0.00000357, -0.00007630, 0.00001747, -0.00025600, 0.00008727, 0.00029134, 0.00011848, 0.00034733, -0.00025923, -0.00005134, -0.00004958, -0.00000101, -0.00000975, 0.00001336, -0.00001783, -0.00001560, 0.00001193, 0.00000076, 0.00000049, -0.00000102, 0.00001702, 0.00000708, -0.00000591, -0.00001588, -0.00001625, 0.00001621, -0.00001950, 0.00000801, -0.00001395, 0.00000803, -0.00001432, 0.00000668, -0.00000586, 0.00000361, -0.00001199, 0.00001332, -0.00000734, -0.00000532, -0.00000796, 0.00001511, -0.00000506, -0.00000342, -0.00000159, },
		{ -0.00000001, -0.00000372, 0.00000148, -0.00000270, 0.00000538, 0.00000037, 0.00000152, -0.00000578, -0.00000038, 0.00000413, -0.00000220, -0.00000143, 0.00000298, -0.00000726, 0.00000092, 0.00000528, -0.00000270, 0.00000938, -0.00000364, -0.00000820, -0.00000198, 0.00000832, -0.00000292, 0.00000112, -0.00000543, -0.00000013, -0.00000343, 0.00000953, -0.00000152, -0.00000480, -0.00000445, 0.00000805, -0.00000066, -0.00000026, 0.00000440, -0.00000841, 0.00000103, -0.00000195, 0.00000057, 0.00000170, 0.00000384, -0.00000820, 0.00000322, -0.00000297, 0.00000444, -0.00000363, 0.00000654, -0.00001120, 0.00000333, -0.00000543, 0.00000638, -0.00000272, -0.00000299, 0.00001106, -0.00000047, -0.00000706, -0.00000203, 0.00000619, -0.00000142, -0.00000263, -0.00000303, 0.00000249, 0.00000053, 0.00000089, -0.00000149, -0.00000180, -0.00000411, 0.00001987, -0.00001100, -0.00000520, 0.00001011, -0.00002252, 0.00001704, -0.00002556, 0.00001525, -0.00000823, 0.00000180, -0.00001983, 0.00001055, -0.00001837, 0.00000897, -0.00003473, 0.00001558, -0.00002390, -0.00000129, -0.00007654, 0.00002033, -0.00023069, -0.00000525, 0.00034293, 0.00011942, 0.00037267, -0.00026545, -0.00005666, -0.00004988, -0.00000345, -0.00001025, 0.00001294, -0.00001759, -0.00001645, 0.00001108, 0.00000183, -0.00000165, 0.00000055, 0.00001181, 0.00001315, -0.00001121, -0.00000681, -0.00002381, 0.00002002, -0.00002084, 0.00000875, -0.00001457, 0.00000861, -0.00001493, 0.00000632, -0.00000473, 0.00000364, -0.00001227, 0.00001162, -0.00000419, -0.00000593, -0.00000970, 0.00001378, -0.00000075, -0.00000611, -0.00000076, 0.00000414, },
		{ -0.00000526, 0.00000198, -0.00000320, 0.00000549, 0.00000049, 0.00000161, -0.00000481, -0.00000213, 0.00000479, -0.00000202, -0.00000166, 0.00000310, -0.00000580, -0.00000213, 0.00000747, -0.00000413, 0.00000960, -0.00000222, -0.00000770, -0.00000429, 0.00000892, -0.00000239, 0.00000118, -0.00000517, -0.00000031, -0.00000382, 0.00000833, 0.00000091, -0.00000541, -0.00000477, 0.00000630, 0.00000212, -0.00000280, 0.00000661, -0.00000885, 0.00000043, -0.00000170, 0.00000067, 0.00000063, 0.00000532, -0.00000832, 0.00000273, -0.00000282, 0.00000390, -0.00000317, 0.00000606, -0.00000906, 0.00000052, -0.00000298, 0.00000336, 0.00000125, -0.00000731, 0.00001280, 0.00000009, -0.00000577, -0.00000475, 0.00000708, -0.00000080, -0.00000217, -0.00000459, 0.00000315, 0.00000029, 0.00000132, -0.00000174, -0.00000116, -0.00000474, 0.00001709, -0.00000472, -0.00001211, 0.00001625, -0.00002606, 0.00001769, -0.00002365, 0.00001224, -0.00000537, -0.00000033, -0.00001756, 0.00000788, -0.00001586, 0.00000684, -0.00003120, 0.00001113, -0.00001875, -0.00000574, -0.00006800, 0.00001026, -0.00018833, -0.00010300, 0.00038050, 0.00013380, 0.00039204, -0.00026592, -0.00006699, -0.00004656, -0.00000870, -0.00000842, 0.00001048, -0.00001569, -0.00001855, 0.00001117, 0.00000224, -0.00000331, 0.00000196, 0.00000688, 0.00001844, -0.00001557, 0.00000024, -0.00002798, 0.00002015, -0.00001895, 0.00000683, -0.00001283, 0.00000700, -0.00001338, 0.00000386, -0.00000186, 0.00000213, -0.00001082, 0.00000824, -0.00000005, -0.00000712, -0.00001016, 0.00001066, 0.00000463, -0.00000904, 0.00000034, 0.00000213, 0.00000255, },
		{ 0.00000206, -0.00000342, 0.00000534, 0.00000080, 0.00000146, -0.00000364, -0.00000370, 0.00000505, -0.00000158, -0.00000194, 0.00000303, -0.00000407, -0.00000496, 0.00000891, -0.00000461, 0.00000888, -0.00000034, -0.00000730, -0.00000604, 0.00000859, -0.00000112, 0.00000067, -0.00000436, -0.00000101, -0.00000354, 0.00000643, 0.00000354, -0.00000595, -0.00000477, 0.00000410, 0.00000490, -0.00000488, 0.00000797, -0.00000831, -0.00000087, -0.00000107, 0.00000049, -0.00000003, 0.00000599, -0.00000745, 0.00000135, -0.00000182, 0.00000247, -0.00000178, 0.00000446, -0.00000584, -0.00000303, 0.00000002, -0.00000011, 0.00000534, -0.00001112, 0.00001366, 0.00000099, -0.00000433, -0.00000725, 0.00000746, -0.00000005, -0.00000149, -0.00000620, 0.00000366, 0.00000013, 0.00000172, -0.00000195, -0.00000071, -0.00000476, 0.00001368, 0.00000123, -0.00001738, 0.00001956, -0.00002594, 0.00001450, -0.00001829, 0.00000635, 0.00000008, -0.00000505, -0.00001258, 0.00000245, -0.00001038, 0.00000127, -0.00002375, 0.00000240, -0.00000874, -0.00001627, -0.00005185, -0.00001025, -0.00013473, -0.00019695, 0.00039662, 0.00016453, 0.00040465, -0.00026000, -0.00008281, -0.00003929, -0.00001709, -0.00000396, 0.00000568, -0.00001182, -0.00002221, 0.00001255, 0.00000156, -0.00000394, 0.00000248, 0.00000312, 0.00002192, -0.00001777, 0.00000401, -0.00002783, 0.00001615, -0.00001375, 0.00000228, -0.00000881, 0.00000331, -0.00000979, -0.00000049, 0.00000250, -0.00000068, -0.00000788, 0.00000352, 0.00000471, -0.00000871, -0.00000949, 0.00000612, 0.00001050, -0.00001173, 0.00000134, 0.00000008, 0.00000485, -0.00000629, },
		{ -0.00000337, 0.00000501, 0.00000121, 0.00000120, -0.00000248, -0.00000490, 0.00000478, -0.00000081, -0.00000234, 0.00000289, -0.00000235, -0.00000723, 0.00000933, -0.00000401, 0.00000721, 0.00000195, -0.00000707, -0.00000710, 0.00000729, 0.00000078, -0.00000029, -0.00000310, -0.00000209, -0.00000274, 0.00000408, 0.00000609, -0.00000622, -0.00000464, 0.00000175, 0.00000730, -0.00000613, 0.00000824, -0.00000672, -0.00000279, -0.00000011, 0.00000002, -0.00000019, 0.00000580, -0.00000563, -0.00000075, -0.00000014, 0.00000036, 0.00000030, 0.00000209, -0.00000201, -0.00000679, 0.00000304, -0.00000347, 0.00000898, -0.00001392, 0.00001338, 0.00000233, -0.00000291, -0.00000932, 0.00000723, 0.00000081, -0.00000061, -0.00000777, 0.00000396, 0.00000005, 0.00000214, -0.00000219, -0.00000031, -0.00000448, 0.00001027, 0.00000598, -0.00002016, 0.00001942, -0.00002195, 0.00000768, -0.00001000, -0.00000177, 0.00000745, -0.00001171, -0.00000560, -0.00000501, -0.00000269, -0.00000694, -0.00001338, -0.00000939, 0.00000471, -0.00003105, -0.00003065, -0.00003704, -0.00007710, -0.00027793, 0.00038527, 0.00021252, 0.00041135, -0.00024822, -0.00010367, -0.00002849, -0.00002831, 0.00000295, -0.00000132, -0.00000607, -0.00002744, 0.00001533, -0.00000044, -0.00000320, 0.00000163, 0.00000120, 0.00002283, -0.00001701, 0.00000382, -0.00002314, 0.00000829, -0.00000574, -0.00000434, -0.00000305, -0.00000197, -0.00000468, -0.00000616, 0.00000778, -0.00000429, -0.00000397, -0.00000194, 0.00000956, -0.00001034, -0.00000803, 0.00000071, 0.00001620, -0.00001366, 0.00000191, -0.00000169, 0.00000660, -0.00000667, 0.00000169, },
		{ 0.00000456, 0.00000164, 0.00000095, -0.00000149, -0.00000557, 0.00000392, 0.00000029, -0.00000286, 0.00000278, -0.00000085, -0.00000867, 0.00000857, -0.00000233, 0.00000476, 0.00000448, -0.00000696, -0.00000748, 0.00000517, 0.00000310, -0.00000149, -0.00000162, -0.00000335, -0.00000167, 0.00000161, 0.00000823, -0.00000602, -0.00000456, -0.00000046, 0.00000898, -0.00000634, 0.00000737, -0.00000420, -0.00000515, 0.00000102, -0.00000067, 0.00000011, 0.00000484, -0.00000309, -0.00000330, 0.00000190, -0.00000209, 0.00000267, -0.00000059, 0.00000187, -0.00001017, 0.00000552, -0.00000618, 0.00001164, -0.00001534, 0.00001184, 0.00000410, -0.00000159, -0.00001083, 0.00000638, 0.00000177, 0.00000040, -0.00000920, 0.00000401, 0.00000002, 0.00000261, -0.00000252, 0.00000013, -0.00000416, 0.00000737, 0.00000893, -0.00002002, 0.00001578, -0.00001450, -0.00000191, 0.00000010, -0.00001098, 0.00001565, -0.00001921, 0.00000226, -0.00001332, 0.00000598, -0.00001637, -0.00000171, -0.00002239, 0.00001944, -0.00004742, -0.00000793, -0.00006504, -0.00002305, -0.00033779, 0.00034290, 0.00027604, 0.00041488, -0.00023235, -0.00012818, -0.00001534, -0.00004145, 0.00001155, -0.00000990, 0.00000103, -0.00003383, 0.00001924, -0.00000364, -0.00000109, -0.00000076, 0.00000139, 0.00002092, -0.00001308, -0.00000031, -0.00001443, -0.00000249, 0.00000398, -0.00001200, 0.00000354, -0.00000798, 0.00000110, -0.00001230, 0.00001317, -0.00000800, 0.00000022, -0.00000740, 0.00001384, -0.00001153, -0.00000624, -0.00000487, 0.00002101, -0.00001434, 0.00000172, -0.00000290, 0.00000761, -0.00000633, 0.00000090, -0.00000311, },
		{ 0.00000198, 0.00000082, -0.00000080, -0.00000565, 0.00000251, 0.00000164, -0.00000349, 0.00000276, 0.00000027, -0.00000915, 0.00000668, 0.00000022, 0.00000181, 0.00000699, -0.00000683, -0.00000731, 0.00000248, 0.00000550, -0.00000263, -0.00000017, -0.00000453, -0.00000062, -0.00000063, 0.00000968, -0.00000519, -0.00000467, -0.00000230, 0.00000974, -0.00000544, 0.00000546, -0.00000100, -0.00000764, 0.00000208, -0.00000141, 0.00000072, 0.00000336, -0.00000018, -0.00000588, 0.00000388, -0.00000445, 0.00000487, -0.00000305, 0.00000522, -0.00001263, 0.00000699, -0.00000780, 0.00001296, -0.00001515, 0.00000905, 0.00000622, -0.00000043, -0.00001168, 0.00000492, 0.00000276, 0.00000151, -0.00001039, 0.00000377, 0.00000006, 0.00000312, -0.00000293, 0.00000067, -0.00000400, 0.00000533, 0.00000979, -0.00001701, 0.00000915, -0.00000461, -0.00001288, 0.00001050, -0.00001979, 0.00002329, -0.00002618, 0.00000960, -0.00002102, 0.00001409, -0.00002534, 0.00000931, -0.00003439, 0.00003291, -0.00006231, 0.00001242, -0.00008902, 0.00002052, -0.00037057, 0.00026921, 0.00035041, 0.00041987, -0.00021537, -0.00015409, -0.00000168, -0.00005499, 0.00002061, -0.00001901, 0.00000858, -0.00004064, 0.00002369, -0.00000757, 0.00000206, -0.00000445, 0.00000350, 0.00001645, -0.00000637, -0.00000765, -0.00000294, -0.00001463, 0.00001387, -0.00001934, 0.00000978, -0.00001365, 0.00000654, -0.00001793, 0.00001780, -0.00001102, 0.00000393, -0.00001208, 0.00001691, -0.00001184, -0.00000461, -0.00000996, 0.00002429, -0.00001342, 0.00000062, -0.00000342, 0.00000780, -0.00000529, -0.00000022, -0.00000274, 0.00000410, },
		{ 0.00000089, -0.00000047, -0.00000514, 0.00000066, 0.00000312, -0.00000413, 0.00000284, 0.00000095, -0.00000868, 0.00000384, 0.00000329, -0.00000123, 0.00000915, -0.00000650, -0.00000677, -0.00000046, 0.00000764, -0.00000343, 0.00000100, -0.00000539, 0.00000016, -0.00000238, 0.00001024, -0.00000370, -0.00000501, -0.00000364, 0.00000951, -0.00000351, 0.00000279, 0.00000248, -0.00000988, 0.00000281, -0.00000203, 0.00000144, 0.00000164, 0.00000269, -0.00000807, 0.00000541, -0.00000631, 0.00000647, -0.00000482, 0.00000756, -0.00001375, 0.00000712, -0.00000807, 0.00001277, -0.00001338, 0.00000524, 0.00000849, 0.00000061, -0.00001187, 0.00000293, 0.00000372, 0.00000266, -0.00001125, 0.00000320, 0.00000016, 0.00000366, -0.00000339, 0.00000128, -0.00000406, 0.00000426, 0.00000864, -0.00001163, 0.00000056, 0.00000626, -0.00002354, 0.00001949, -0.00002667, 0.00002895, -0.00003124, 0.00001500, -0.00002666, 0.00002010, -0.00003211, 0.00001774, -0.00004322, 0.00004269, -0.00007284, 0.00002687, -0.00010454, 0.00004840, -0.00037339, 0.00016765, 0.00042802, 0.00043252, -0.00020105, -0.00017860, 0.00001020, -0.00006712, 0.00002858, -0.00002729, 0.00001541, -0.00004687, 0.00002783, -0.00001152, 0.00000560, -0.00000885, 0.00000693, 0.00001016, 0.00000215, -0.00001687, 0.00000958, -0.00002625, 0.00002225, -0.00002496, 0.00001447, -0.00001791, 0.00001063, -0.00002210, 0.00002081, -0.00001263, 0.00000645, -0.00001531, 0.00001823, -0.00001089, -0.00000358, -0.00001396, 0.00002559, -0.00001075, -0.00000140, -0.00000324, 0.00000719, -0.00000368, -0.00000151, -0.00000241, 0.00000371, 0.00000217, },
		{ -0.00000048, -0.00000415, -0.00000144, 0.00000453, -0.00000468, 0.00000295, 0.00000124, -0.00000743, 0.00000042, 0.00000642, -0.00000391, 0.00001063, -0.00000577, -0.00000610, -0.00000327, 0.00000916, -0.00000363, 0.00000166, -0.00000575, 0.00000047, -0.00000344, 0.00000985, -0.00000159, -0.00000556, -0.00000446, 0.00000834, -0.00000083, -0.00000026, 0.00000578, -0.00001150, 0.00000296, -0.00000232, 0.00000204, 0.00000003, 0.00000511, -0.00000947, 0.00000612, -0.00000732, 0.00000713, -0.00000556, 0.00000857, -0.00001332, 0.00000580, -0.00000696, 0.00001112, -0.00001027, 0.00000075, 0.00001063, 0.00000161, -0.00001144, 0.00000056, 0.00000455, 0.00000380, -0.00001171, 0.00000229, 0.00000034, 0.00000415, -0.00000381, 0.00000188, -0.00000429, 0.00000406, 0.00000590, -0.00000478, -0.00000864, 0.00001643, -0.00003214, 0.00002549, -0.00003028, 0.00003146, -0.00003327, 0.00001731, -0.00002907, 0.00002275, -0.00003532, 0.00002202, -0.00004719, 0.00004692, -0.00007685, 0.00003288, -0.00010864, 0.00005784, -0.00034695, 0.00004541, 0.00049891, 0.00045989, -0.00019333, -0.00019886, 0.00001802, -0.00007600, 0.00003393, -0.00003339, 0.00002030, -0.00005147, 0.00003073, -0.00001462, 0.00000871, -0.00001316, 0.00001081, 0.00000314, 0.00001116, -0.00002630, 0.00002116, -0.00003544, 0.00002751, -0.00002762, 0.00001662, -0.00001988, 0.00001257, -0.00002409, 0.00002162, -0.00001233, 0.00000730, -0.00001662, 0.00001751, -0.00000852, -0.00000342, -0.00001650, 0.00002470, -0.00000644, -0.00000415, -0.00000249, 0.00000596, -0.00000172, -0.00000277, -0.00000221, 0.00000346, 0.00000217, 0.00000116, },
		{ -0.00000284, -0.00000355, 0.00000565, -0.00000497, 0.00000300, 0.00000123, -0.00000564, -0.00000315, 0.00000913, -0.00000581, 0.00001115, -0.00000449, -0.00000551, -0.00000565, 0.00000980, -0.00000306, 0.00000171, -0.00000549, 0.00000021, -0.00000375, 0.00000856, 0.00000097, -0.00000620, -0.00000482, 0.00000644, 0.00000224, -0.00000320, 0.00000846, -0.00001214, 0.00000232, -0.00000214, 0.00000233, -0.00000119, 0.00000674, -0.00000980, 0.00000580, -0.00000729, 0.00000667, -0.00000511, 0.00000814, -0.00001136, 0.00000319, -0.00000469, 0.00000833, -0.00000624, -0.00000389, 0.00001229, 0.00000266, -0.00001050, -0.00000200, 0.00000516, 0.00000491, -0.00001174, 0.00000106, 0.00000062, 0.00000454, -0.00000408, 0.00000235, -0.00000455, 0.00000447, 0.00000217, 0.00000244, -0.00001700, 0.00002431, -0.00003720, 0.00002733, -0.00002975, 0.00003016, -0.00003165, 0.00001592, -0.00002760, 0.00002139, -0.00003419, 0.00002133, -0.00004544, 0.00004471, -0.00007335, 0.00002942, -0.00010041, 0.00004902, -0.00029563, -0.00008715, 0.00055165, 0.00050893, -0.00019557, -0.00021259, 0.00001997, -0.00008019, 0.00003541, -0.00003620, 0.00002229, -0.00005356, 0.00003159, -0.00001611, 0.00001060, -0.00001653, 0.00001414, -0.00000340, 0.00001920, -0.00003422, 0.00002996, -0.00004057, 0.00002850, -0.00002655, 0.00001566, -0.00001908, 0.00001194, -0.00002354, 0.00001996, -0.00000995, 0.00000631, -0.00001588, 0.00001475, -0.00000482, -0.00000417, -0.00001745, 0.00002174, -0.00000087, -0.00000725, -0.00000147, 0.00000436, 0.00000032, -0.00000381, -0.00000227, 0.00000340, 0.00000198, 0.00000160, -0.00000075, },
		{ -0.00000542, 0.00000630, -0.00000489, 0.00000290, 0.00000107, -0.00000362, -0.00000644, 0.00001097, -0.00000660, 0.00001055, -0.00000260, -0.00000514, -0.00000738, 0.00000940, -0.00000169, 0.00000115, -0.00000464, -0.00000057, -0.00000337, 0.00000658, 0.00000371, -0.00000673, -0.00000488, 0.00000410, 0.00000528, -0.00000559, 0.00001012, -0.00001161, 0.00000084, -0.00000146, 0.00000220, -0.00000183, 0.00000741, -0.00000896, 0.00000443, -0.00000624, 0.00000514, -0.00000355, 0.00000645, -0.00000817, -0.00000035, -0.00000166, 0.00000488, -0.00000189, -0.00000813, 0.00001315, 0.00000392, -0.00000921, -0.00000454, 0.00000541, 0.00000598, -0.00001134, -0.00000045, 0.00000096, 0.00000476, -0.00000410, 0.00000256, -0.00000469, 0.00000514, -0.00000182, 0.00000896, -0.00002322, 0.00002867, -0.00003779, 0.00002449, -0.00002492, 0.00002502, -0.00002640, 0.00001088, -0.00002235, 0.00001610, -0.00002879, 0.00001575, -0.00003817, 0.00003635, -0.00006276, 0.00001716, -0.00008119, 0.00002500, -0.00022701, -0.00021745, 0.00057469, 0.00058527, -0.00020962, -0.00021870, 0.00001522, -0.00007906, 0.00003247, -0.00003519, 0.00002086, -0.00005269, 0.00002996, -0.00001550, 0.00001071, -0.00001830, 0.00001609, -0.00000838, 0.00002500, -0.00003919, 0.00003457, -0.00004065, 0.00002471, -0.00002161, 0.00001159, -0.00001555, 0.00000877, -0.00002054, 0.00001603, -0.00000572, 0.00000367, -0.00001329, 0.00001030, -0.00000015, -0.00000562, -0.00001697, 0.00001710, 0.00000538, -0.00001018, -0.00000051, 0.00000272, 0.00000216, -0.00000443, -0.00000264, 0.00000352, 0.00000165, 0.00000213, -0.00000117, -0.00000144, },
		{ 0.00000631, -0.00000436, 0.00000258, 0.00000091, -0.00000167, -0.00000904, 0.00001163, -0.00000612, 0.00000884, -0.00000018, -0.00000502, -0.00000834, 0.00000798, 0.00000035, 0.00000011, -0.00000334, -0.00000175, -0.00000250, 0.00000417, 0.00000631, -0.00000694, -0.00000482, 0.00000162, 0.00000788, -0.00000705, 0.00001054, -0.00000985, -0.00000140, -0.00000036, 0.00000166, -0.00000185, 0.00000708, -0.00000704, 0.00000218, -0.00000438, 0.00000278, -0.00000118, 0.00000386, -0.00000423, -0.00000425, 0.00000152, 0.00000138, 0.00000214, -0.00001142, 0.00001291, 0.00000547, -0.00000773, -0.00000683, 0.00000519, 0.00000702, -0.00001057, -0.00000211, 0.00000130, 0.00000482, -0.00000384, 0.00000246, -0.00000458, 0.00000578, -0.00000544, 0.00001386, -0.00002641, 0.00002885, -0.00003370, 0.00001724, -0.00001639, 0.00001681, -0.00001831, 0.00000299, -0.00001416, 0.00000778, -0.00002008, 0.00000636, -0.00002666, 0.00002339, -0.00004693, -0.00000149, -0.00005435, -0.00000870, -0.00015099, -0.00033210, 0.00055780, 0.00069209, -0.00023506, -0.00021785, 0.00000430, -0.00007307, 0.00002546, -0.00003061, 0.00001623, -0.00004902, 0.00002592, -0.00001275, 0.00000888, -0.00001817, 0.00001618, -0.00001112, 0.00002773, -0.00004035, 0.00003435, -0.00003553, 0.00001650, -0.00001340, 0.00000505, -0.00000987, 0.00000365, -0.00001567, 0.00001043, -0.00000025, -0.00000008, -0.00000942, 0.00000480, 0.00000488, -0.00000734, -0.00001547, 0.00001142, 0.00001154, -0.00001237, -0.00000002, 0.00000136, 0.00000353, -0.00000450, -0.00000335, 0.00000379, 0.00000123, 0.00000266, -0.00000159, -0.00000013, -0.00000683, },
		{ -0.00000339, 0.00000204, 0.00000085, -0.00000002, -0.00001069, 0.00001097, -0.00000442, 0.00000620, 0.00000257, -0.00000507, -0.00000858, 0.00000570, 0.00000281, -0.00000115, -0.00000182, -0.00000309, -0.00000139, 0.00000170, 0.00000845, -0.00000663, -0.00000482, -0.00000070, 0.00000970, -0.00000735, 0.00000965, -0.00000701, -0.00000415, 0.00000097, 0.00000084, -0.00000134, 0.00000593, -0.00000433, -0.00000059, -0.00000209, 0.00000002, 0.00000155, 0.00000094, -0.00000020, -0.00000784, 0.00000424, -0.00000152, 0.00000524, -0.00001327, 0.00001137, 0.00000738, -0.00000624, -0.00000861, 0.00000438, 0.00000809, -0.00000955, -0.00000376, 0.00000153, 0.00000478, -0.00000335, 0.00000206, -0.00000420, 0.00000620, -0.00000825, 0.00001661, -0.00002621, 0.00002493, -0.00002555, 0.00000667, -0.00000552, 0.00000695, -0.00000880, -0.00000629, -0.00000456, -0.00000197, -0.00000977, -0.00000492, -0.00001313, 0.00000839, -0.00002885, -0.00002281, -0.00002484, -0.00004491, -0.00007846, -0.00041851, 0.00049367, 0.00082899, -0.00026861, -0.00021273, -0.00001072, -0.00006397, 0.00001578, -0.00002364, 0.00000939, -0.00004339, 0.00002015, -0.00000840, 0.00000548, -0.00001634, 0.00001443, -0.00001145, 0.00002720, -0.00003760, 0.00002951, -0.00002594, 0.00000503, -0.00000321, -0.00000275, -0.00000315, -0.00000242, -0.00000992, 0.00000415, 0.00000554, -0.00000411, -0.00000507, -0.00000090, 0.00000950, -0.00000875, -0.00001356, 0.00000551, 0.00001678, -0.00001324, -0.00000035, 0.00000056, 0.00000422, -0.00000393, -0.00000441, 0.00000416, 0.00000080, 0.00000311, -0.00000191, 0.00000093, -0.00000764, 0.00000565, },
		{ 0.00000135, 0.00000089, 0.00000122, -0.00001129, 0.00000905, -0.00000173, 0.00000299, 0.00000533, -0.00000510, -0.00000826, 0.00000287, 0.00000533, -0.00000232, -0.00000036, -0.00000432, -0.00000034, -0.00000052, 0.00000985, -0.00000568, -0.00000501, -0.00000264, 0.00001054, -0.00000644, 0.00000761, -0.00000342, -0.00000704, 0.00000223, -0.00000006, -0.00000051, 0.00000424, -0.00000124, -0.00000339, 0.00000013, -0.00000263, 0.00000406, -0.00000172, 0.00000325, -0.00001047, 0.00000589, -0.00000329, 0.00000690, -0.00001338, 0.00000846, 0.00000966, -0.00000490, -0.00000969, 0.00000288, 0.00000925, -0.00000847, -0.00000516, 0.00000145, 0.00000479, -0.00000277, 0.00000155, -0.00000368, 0.00000639, -0.00001011, 0.00001714, -0.00002289, 0.00001772, -0.00001472, -0.00000543, 0.00000574, -0.00000264, 0.00000023, -0.00001505, 0.00000447, -0.00001105, -0.00000010, -0.00001562, -0.00000041, -0.00000544, -0.00001226, -0.00004231, 0.00000164, -0.00007594, -0.00001993, -0.00046647, 0.00037911, 0.00099134, -0.00030387, -0.00020812, -0.00002621, -0.00005467, 0.00000584, -0.00001628, 0.00000209, -0.00003735, 0.00001397, -0.00000353, 0.00000141, -0.00001356, 0.00001141, -0.00000985, 0.00002393, -0.00003164, 0.00002115, -0.00001350, -0.00000783, 0.00000713, -0.00001019, 0.00000320, -0.00000813, -0.00000453, -0.00000161, 0.00001056, -0.00000744, -0.00000121, -0.00000582, 0.00001286, -0.00000921, -0.00001189, 0.00000022, 0.00002034, -0.00001235, -0.00000173, 0.00000052, 0.00000408, -0.00000268, -0.00000575, 0.00000460, 0.00000039, 0.00000346, -0.00000207, 0.00000167, -0.00000782, 0.00000437, -0.00000211, },
		{ 0.00000098, 0.00000202, -0.00001091, 0.00000615, 0.00000151, -0.00000031, 0.00000768, -0.00000488, -0.00000764, -0.00000015, 0.00000752, -0.00000308, 0.00000077, -0.00000519, 0.00000039, -0.00000219, 0.00001034, -0.00000405, -0.00000541, -0.00000408, 0.00001034, -0.00000447, 0.00000476, 0.00000044, -0.00000960, 0.00000306, -0.00000075, 0.00000035, 0.00000243, 0.00000170, -0.00000567, 0.00000176, -0.00000463, 0.00000581, -0.00000350, 0.00000552, -0.00001160, 0.00000601, -0.00000350, 0.00000682, -0.00001163, 0.00000430, 0.00001224, -0.00000384, -0.00000987, 0.00000061, 0.00001060, -0.00000753, -0.00000604, 0.00000086, 0.00000507, -0.00000237, 0.00000119, -0.00000329, 0.00000655, -0.00001119, 0.00001584, -0.00001734, 0.00000866, -0.00000315, -0.00001685, 0.00001515, -0.00000982, 0.00000676, -0.00002125, 0.00001080, -0.00001724, 0.00000655, -0.00002308, 0.00000854, -0.00001475, -0.00000093, -0.00005546, 0.00001955, -0.00009471, 0.00001596, -0.00046958, 0.00021607, 0.00117001, -0.00033157, -0.00021053, -0.00003741, -0.00004901, -0.00000126, -0.00001118, -0.00000338, -0.00003290, 0.00000912, 0.00000031, -0.00000199, -0.00001101, 0.00000821, -0.00000737, 0.00001907, -0.00002388, 0.00001111, -0.00000042, -0.00001977, 0.00001551, -0.00001551, 0.00000765, -0.00001210, -0.00000079, -0.00000564, 0.00001370, -0.00000911, 0.00000122, -0.00000904, 0.00001421, -0.00000812, -0.00001110, -0.00000370, 0.00002160, -0.00000941, -0.00000429, 0.00000134, 0.00000308, -0.00000080, -0.00000733, 0.00000507, -0.00000002, 0.00000370, -0.00000210, 0.00000210, -0.00000747, 0.00000268, -0.00000071, 0.00000067, },
		{ 0.00000250, -0.00000978, 0.00000271, 0.00000476, -0.00000316, 0.00000923, -0.00000413, -0.00000700, -0.00000295, 0.00000900, -0.00000316, 0.00000135, -0.00000549, 0.00000061, -0.00000316, 0.00000987, -0.00000183, -0.00000597, -0.00000502, 0.00000923, -0.00000176, 0.00000158, 0.00000404, -0.00001137, 0.00000313, -0.00000096, 0.00000091, 0.00000092, 0.00000398, -0.00000694, 0.00000233, -0.00000552, 0.00000633, -0.00000396, 0.00000618, -0.00001089, 0.00000438, -0.00000201, 0.00000492, -0.00000813, -0.00000089, 0.00001496, -0.00000316, -0.00000903, -0.00000246, 0.00001221, -0.00000697, -0.00000613, -0.00000051, 0.00000585, -0.00000245, 0.00000135, -0.00000340, 0.00000705, -0.00001193, 0.00001347, -0.00001084, -0.00000042, 0.00000697, -0.00002529, 0.00002056, -0.00001265, 0.00000895, -0.00002306, 0.00001255, -0.00001858, 0.00000809, -0.00002498, 0.00001111, -0.00001670, 0.00000192, -0.00005853, 0.00002444, -0.00009591, 0.00002360, -0.00042619, 0.00001189, 0.00135172, -0.00034026, -0.00022742, -0.00003897, -0.00005114, -0.00000212, -0.00001118, -0.00000451, -0.00003225, 0.00000755, 0.00000141, -0.00000315, -0.00001016, 0.00000620, -0.00000541, 0.00001423, -0.00001623, 0.00000163, 0.00001080, -0.00002840, 0.00001994, -0.00001709, 0.00000887, -0.00001313, 0.00000018, -0.00000691, 0.00001408, -0.00000834, 0.00000149, -0.00000984, 0.00001300, -0.00000510, -0.00001167, -0.00000567, 0.00002022, -0.00000441, -0.00000792, 0.00000298, 0.00000126, 0.00000162, -0.00000908, 0.00000557, -0.00000045, 0.00000393, -0.00000209, 0.00000236, -0.00000678, 0.00000085, 0.00000052, 0.00000022, 0.00000098, },
		{ -0.00000825, -0.00000076, 0.00000743, -0.00000505, 0.00000963, -0.00000265, -0.00000658, -0.00000519, 0.00000946, -0.00000239, 0.00000125, -0.00000515, 0.00000024, -0.00000337, 0.00000853, 0.00000078, -0.00000653, -0.00000558, 0.00000744, 0.00000129, -0.00000140, 0.00000682, -0.00001193, 0.00000218, -0.00000048, 0.00000091, 0.00000006, 0.00000521, -0.00000684, 0.00000154, -0.00000504, 0.00000538, -0.00000284, 0.00000502, -0.00000826, 0.00000102, 0.00000110, 0.00000139, -0.00000321, -0.00000670, 0.00001764, -0.00000290, -0.00000711, -0.00000632, 0.00001416, -0.00000702, -0.00000519, -0.00000284, 0.00000740, -0.00000336, 0.00000241, -0.00000444, 0.00000834, -0.00001293, 0.00001104, -0.00000490, -0.00000759, 0.00001352, -0.00002875, 0.00002025, -0.00000971, 0.00000552, -0.00001923, 0.00000845, -0.00001373, 0.00000307, -0.00001975, 0.00000558, -0.00000940, -0.00000576, -0.00004920, 0.00001374, -0.00007683, 0.00000128, -0.00033984, -0.00022096, 0.00152000, -0.00031747, -0.00026625, -0.00002588, -0.00006492, 0.00000631, -0.00001889, 0.00000097, -0.00003739, 0.00001102, -0.00000182, -0.00000060, -0.00001243, 0.00000683, -0.00000550, 0.00001117, -0.00001071, -0.00000501, 0.00001783, -0.00003167, 0.00001885, -0.00001379, 0.00000594, -0.00001041, -0.00000235, -0.00000478, 0.00001118, -0.00000469, -0.00000085, -0.00000785, 0.00000902, -0.00000003, -0.00001384, -0.00000541, 0.00001615, 0.00000241, -0.00001237, 0.00000527, -0.00000124, 0.00000445, -0.00001091, 0.00000613, -0.00000102, 0.00000427, -0.00000220, 0.00000263, -0.00000603, -0.00000077, 0.00000125, 0.00000023, 0.00000073, 0.00000285, },
		{ -0.00000373, 0.00000902, -0.00000558, 0.00000866, -0.00000036, -0.00000656, -0.00000661, 0.00000878, -0.00000073, 0.00000049, -0.00000418, -0.00000068, -0.00000291, 0.00000653, 0.00000349, -0.00000689, -0.00000592, 0.00000528, 0.00000418, -0.00000367, 0.00000835, -0.00001101, 0.00000012, 0.00000077, 0.00000020, 0.00000010, 0.00000512, -0.00000520, -0.00000069, -0.00000314, 0.00000292, -0.00000016, 0.00000215, -0.00000393, -0.00000377, 0.00000548, -0.00000336, 0.00000264, -0.00001266, 0.00002005, -0.00000307, -0.00000412, -0.00001088, 0.00001647, -0.00000782, -0.00000304, -0.00000626, 0.00000988, -0.00000534, 0.00000471, -0.00000677, 0.00001084, -0.00001480, 0.00000955, -0.00000098, -0.00001116, 0.00001487, -0.00002590, 0.00001331, -0.00000041, -0.00000394, -0.00000936, -0.00000191, -0.00000230, -0.00000892, -0.00000687, -0.00000859, 0.00000762, -0.00002440, -0.00002709, -0.00001275, -0.00003788, -0.00004860, -0.00021904, -0.00046584, 0.00165671, -0.00025121, -0.00033318, 0.00000563, -0.00009307, 0.00002615, -0.00003608, 0.00001461, -0.00004970, 0.00002076, -0.00001053, 0.00000679, -0.00001894, 0.00001125, -0.00000895, 0.00001143, -0.00000909, -0.00000688, 0.00001888, -0.00002824, 0.00001145, -0.00000521, -0.00000141, -0.00000372, -0.00000858, 0.00000090, 0.00000496, 0.00000181, -0.00000581, -0.00000308, 0.00000242, 0.00000690, -0.00001753, -0.00000296, 0.00000969, 0.00001052, -0.00001721, 0.00000796, -0.00000421, 0.00000752, -0.00001278, 0.00000681, -0.00000182, 0.00000488, -0.00000261, 0.00000312, -0.00000551, -0.00000184, 0.00000120, 0.00000092, 0.00000004, 0.00000329, -0.00000668, },
		{ 0.00000915, -0.00000456, 0.00000630, 0.00000269, -0.00000697, -0.00000713, 0.00000695, 0.00000169, -0.00000079, -0.00000273, -0.00000199, -0.00000199, 0.00000418, 0.00000598, -0.00000682, -0.00000623, 0.00000309, 0.00000648, -0.00000480, 0.00000836, -0.00000857, -0.00000299, 0.00000274, -0.00000124, 0.00000111, 0.00000369, -0.00000210, -0.00000421, 0.00000001, -0.00000081, 0.00000379, -0.00000208, 0.00000164, -0.00000946, 0.00001059, -0.00000876, 0.00000879, -0.00001823, 0.00002194, -0.00000364, -0.00000018, -0.00001594, 0.00001906, -0.00000943, 0.00000032, -0.00001077, 0.00001335, -0.00000852, 0.00000840, -0.00001060, 0.00001483, -0.00001802, 0.00000983, -0.00000016, -0.00001001, 0.00001013, -0.00001634, -0.00000018, 0.00001483, -0.00001889, 0.00000596, -0.00001789, 0.00001505, -0.00002718, 0.00001287, -0.00003050, 0.00003327, -0.00005263, 0.00000613, -0.00005271, 0.00001742, -0.00011971, -0.00007635, -0.00070354, 0.00174384, -0.00013167, -0.00043182, 0.00005715, -0.00013653, 0.00005800, -0.00006322, 0.00003684, -0.00006956, 0.00003710, -0.00002507, 0.00001941, -0.00003020, 0.00002010, -0.00001656, 0.00001602, -0.00001255, -0.00000282, 0.00001304, -0.00001777, -0.00000211, 0.00000823, -0.00001268, 0.00000650, -0.00001808, 0.00000967, -0.00000410, 0.00001069, -0.00001297, 0.00000404, -0.00000626, 0.00001515, -0.00002237, 0.00000133, 0.00000145, 0.00001923, -0.00002190, 0.00001071, -0.00000742, 0.00001068, -0.00001464, 0.00000766, -0.00000299, 0.00000589, -0.00000347, 0.00000405, -0.00000548, -0.00000208, 0.00000016, 0.00000241, -0.00000119, 0.00000400, -0.00000545, -0.00000574, },
		{ -0.00000202, 0.00000275, 0.00000629, -0.00000771, -0.00000680, 0.00000416, 0.00000459, -0.00000233, -0.00000103, -0.00000347, -0.00000087, 0.00000181, 0.00000792, -0.00000614, -0.00000670, 0.00000116, 0.00000785, -0.00000457, 0.00000681, -0.00000477, -0.00000686, 0.00000519, -0.00000325, 0.00000295, 0.00000111, 0.00000215, -0.00000860, 0.00000396, -0.00000536, 0.00000852, -0.00000708, 0.00000777, -0.00001534, 0.00001576, -0.00001412, 0.00001457, -0.00002292, 0.00002311, -0.00000452, 0.00000452, -0.00002125, 0.00002180, -0.00001181, 0.00000479, -0.00001618, 0.00001766, -0.00001284, 0.00001343, -0.00001588, 0.00002031, -0.00002278, 0.00001230, -0.00000300, -0.00000377, -0.00000058, -0.00000075, -0.00001902, 0.00003454, -0.00003777, 0.00002513, -0.00003785, 0.00003658, -0.00004987, 0.00003751, -0.00005791, 0.00006494, -0.00008744, 0.00004683, -0.00010157, 0.00008286, -0.00020269, 0.00007298, -0.00091421, 0.00176565, 0.00004714, -0.00056216, 0.00012737, -0.00019390, 0.00010055, -0.00009920, 0.00006668, -0.00009610, 0.00005928, -0.00004482, 0.00003680, -0.00004589, 0.00003320, -0.00002835, 0.00002519, -0.00002138, 0.00000735, 0.00000055, -0.00000102, -0.00002064, 0.00002519, -0.00002665, 0.00001914, -0.00002982, 0.00002053, -0.00001501, 0.00002101, -0.00002148, 0.00001267, -0.00001613, 0.00002392, -0.00002776, 0.00000682, -0.00000774, 0.00002768, -0.00002588, 0.00001318, -0.00001061, 0.00001372, -0.00001641, 0.00000872, -0.00000457, 0.00000739, -0.00000489, 0.00000554, -0.00000613, -0.00000133, -0.00000195, 0.00000471, -0.00000299, 0.00000510, -0.00000479, -0.00000652, 0.00000765, },
		{ -0.00000160, 0.00001005, -0.00000857, -0.00000580, 0.00000074, 0.00000760, -0.00000380, 0.00000061, -0.00000483, 0.00000016, -0.00000024, 0.00000905, -0.00000475, -0.00000740, -0.00000030, 0.00000811, -0.00000295, 0.00000388, -0.00000000, -0.00001107, 0.00000775, -0.00000556, 0.00000536, -0.00000226, 0.00000703, -0.00001331, 0.00000813, -0.00001011, 0.00001336, -0.00001214, 0.00001368, -0.00002067, 0.00002027, -0.00001879, 0.00001938, -0.00002629, 0.00002340, -0.00000562, 0.00000969, -0.00002645, 0.00002445, -0.00001473, 0.00001004, -0.00002212, 0.00002251, -0.00001800, 0.00001949, -0.00002228, 0.00002699, -0.00002887, 0.00001692, -0.00000935, 0.00000704, -0.00001617, 0.00001915, -0.00004107, 0.00005635, -0.00005819, 0.00004578, -0.00005934, 0.00005969, -0.00007423, 0.00006404, -0.00008751, 0.00009883, -0.00012445, 0.00008992, -0.00015307, 0.00015038, -0.00028633, 0.00021277, -0.00107951, 0.00171061, 0.00028602, -0.00071976, 0.00021165, -0.00026125, 0.00015044, -0.00014116, 0.00010169, -0.00012714, 0.00008537, -0.00006810, 0.00005752, -0.00006477, 0.00004952, -0.00004350, 0.00003824, -0.00003487, 0.00002268, -0.00001724, 0.00002015, -0.00004201, 0.00004363, -0.00004149, 0.00003259, -0.00004232, 0.00003210, -0.00002641, 0.00003154, -0.00003023, 0.00002171, -0.00002607, 0.00003222, -0.00003290, 0.00001270, -0.00001691, 0.00003500, -0.00002860, 0.00001507, -0.00001353, 0.00001648, -0.00001804, 0.00000998, -0.00000658, 0.00000936, -0.00000685, 0.00000759, -0.00000748, 0.00000040, -0.00000501, 0.00000768, -0.00000523, 0.00000657, -0.00000483, -0.00000596, 0.00000464, 0.00000177, },
		{ 0.00001352, -0.00000923, -0.00000445, -0.00000290, 0.00001028, -0.00000487, 0.00000194, -0.00000581, 0.00000084, -0.00000172, 0.00000924, -0.00000264, -0.00000834, -0.00000122, 0.00000726, -0.00000015, -0.00000000, 0.00000517, -0.00001503, 0.00000999, -0.00000778, 0.00000790, -0.00000587, 0.00001193, -0.00001766, 0.00001186, -0.00001438, 0.00001761, -0.00001651, 0.00001863, -0.00002473, 0.00002351, -0.00002218, 0.00002269, -0.00002803, 0.00002271, -0.00000681, 0.00001497, -0.00003111, 0.00002667, -0.00001786, 0.00001560, -0.00002805, 0.00002737, -0.00002347, 0.00002599, -0.00002916, 0.00003421, -0.00003571, 0.00002310, -0.00001839, 0.00002106, -0.00003464, 0.00004079, -0.00006343, 0.00007734, -0.00007727, 0.00006502, -0.00007939, 0.00008130, -0.00009696, 0.00008891, -0.00011530, 0.00013048, -0.00015859, 0.00012955, -0.00020017, 0.00021119, -0.00035910, 0.00032785, -0.00118464, 0.00157302, 0.00057972, -0.00089555, 0.00030216, -0.00033223, 0.00020241, -0.00018470, 0.00013805, -0.00015933, 0.00011241, -0.00009229, 0.00007924, -0.00008478, 0.00006722, -0.00006034, 0.00005359, -0.00005135, 0.00004124, -0.00003795, 0.00004303, -0.00006346, 0.00006101, -0.00005505, 0.00004497, -0.00005386, 0.00004272, -0.00003677, 0.00004088, -0.00003797, 0.00002996, -0.00003490, 0.00003903, -0.00003695, 0.00001811, -0.00002509, 0.00004038, -0.00002962, 0.00001612, -0.00001597, 0.00001879, -0.00001943, 0.00001138, -0.00000891, 0.00001169, -0.00000923, 0.00001007, -0.00000940, 0.00000293, -0.00000877, 0.00001100, -0.00000768, 0.00000829, -0.00000549, -0.00000422, 0.00000043, 0.00000631, -0.00000619, },
		{ -0.00000938, -0.00000307, -0.00000633, 0.00001225, -0.00000525, 0.00000271, -0.00000624, 0.00000101, -0.00000249, 0.00000847, 0.00000003, -0.00000940, -0.00000165, 0.00000546, 0.00000343, -0.00000426, 0.00001007, -0.00001816, 0.00001143, -0.00000950, 0.00001014, -0.00000918, 0.00001615, -0.00002098, 0.00001452, -0.00001753, 0.00002062, -0.00001953, 0.00002196, -0.00002696, 0.00002499, -0.00002386, 0.00002419, -0.00002799, 0.00002102, -0.00000793, 0.00001994, -0.00003476, 0.00002811, -0.00002077, 0.00002086, -0.00003330, 0.00003160, -0.00002856, 0.00003214, -0.00003566, 0.00004107, -0.00004237, 0.00002978, -0.00002870, 0.00003629, -0.00005336, 0.00006112, -0.00008292, 0.00009437, -0.00009203, 0.00007992, -0.00009499, 0.00009824, -0.00011472, 0.00010849, -0.00013728, 0.00015537, -0.00018482, 0.00015992, -0.00023605, 0.00025702, -0.00041064, 0.00040585, -0.00122005, 0.00135413, 0.00091669, -0.00107615, 0.00038831, -0.00039867, 0.00024984, -0.00022431, 0.00017106, -0.00018852, 0.00013672, -0.00011412, 0.00009904, -0.00010327, 0.00008383, -0.00007657, 0.00006900, -0.00006843, 0.00006035, -0.00005857, 0.00006440, -0.00008194, 0.00007471, -0.00006513, 0.00005438, -0.00006269, 0.00005079, -0.00004461, 0.00004769, -0.00004350, 0.00003623, -0.00004150, 0.00004342, -0.00003916, 0.00002224, -0.00003139, 0.00004320, -0.00002866, 0.00001619, -0.00001774, 0.00002048, -0.00002045, 0.00001277, -0.00001137, 0.00001411, -0.00001176, 0.00001271, -0.00001160, 0.00000591, -0.00001277, 0.00001426, -0.00000999, 0.00001000, -0.00000657, -0.00000162, -0.00000443, 0.00001096, -0.00001039, 0.00001620, },
		{ -0.00000199, -0.00000913, 0.00001316, -0.00000476, 0.00000281, -0.00000603, 0.00000060, -0.00000254, 0.00000688, 0.00000302, -0.00001038, -0.00000177, 0.00000304, 0.00000727, -0.00000824, 0.00001404, -0.00001990, 0.00001170, -0.00001037, 0.00001164, -0.00001164, 0.00001913, -0.00002271, 0.00001558, -0.00001907, 0.00002190, -0.00002072, 0.00002323, -0.00002704, 0.00002447, -0.00002365, 0.00002373, -0.00002618, 0.00001840, -0.00000882, 0.00002416, -0.00003694, 0.00002835, -0.00002296, 0.00002516, -0.00003715, 0.00003446, -0.00003248, 0.00003704, -0.00004078, 0.00004649, -0.00004770, 0.00003564, -0.00003850, 0.00005037, -0.00006949, 0.00007705, -0.00009649, 0.00010464, -0.00009986, 0.00008793, -0.00010355, 0.00010778, -0.00012458, 0.00011960, -0.00014993, 0.00016961, -0.00019882, 0.00017620, -0.00025507, 0.00028122, -0.00043314, 0.00043858, -0.00118256, 0.00106241, 0.00127963, -0.00124487, 0.00045790, -0.00045146, 0.00028552, -0.00025407, 0.00019565, -0.00021029, 0.00015437, -0.00013013, 0.00011377, -0.00011732, 0.00009664, -0.00008957, 0.00008184, -0.00008332, 0.00007694, -0.00007582, 0.00008103, -0.00009457, 0.00008240, -0.00006987, 0.00005923, -0.00006737, 0.00005497, -0.00004868, 0.00005090, -0.00004583, 0.00003960, -0.00004499, 0.00004469, -0.00003896, 0.00002444, -0.00003516, 0.00004307, -0.00002566, 0.00001524, -0.00001873, 0.00002140, -0.00002095, 0.00001396, -0.00001365, 0.00001630, -0.00001406, 0.00001513, -0.00001369, 0.00000886, -0.00001650, 0.00001695, -0.00001175, 0.00001138, -0.00000773, 0.00000136, -0.00000929, 0.00001500, -0.00001357, 0.00001764, -0.00000872, },
		{ -0.00001103, 0.00001285, -0.00000334, 0.00000225, -0.00000519, -0.00000033, -0.00000199, 0.00000473, 0.00000599, -0.00001101, -0.00000181, 0.00000039, 0.00001081, -0.00001131, 0.00001649, -0.00001988, 0.00001053, -0.00001016, 0.00001209, -0.00001285, 0.00002042, -0.00002250, 0.00001477, -0.00001874, 0.00002121, -0.00001986, 0.00002229, -0.00002488, 0.00002196, -0.00002156, 0.00002141, -0.00002277, 0.00001498, -0.00000933, 0.00002721, -0.00003724, 0.00002707, -0.00002396, 0.00002786, -0.00003891, 0.00003528, -0.00003446, 0.00003979, -0.00004353, 0.00004941, -0.00005053, 0.00003927, -0.00004594, 0.00006096, -0.00008036, 0.00008591, -0.00010173, 0.00010609, -0.00009896, 0.00008734, -0.00010326, 0.00010803, -0.00012453, 0.00012005, -0.00015083, 0.00017053, -0.00019769, 0.00017517, -0.00025359, 0.00027978, -0.00042235, 0.00042301, -0.00107575, 0.00071314, 0.00164672, -0.00138322, 0.00049840, -0.00048172, 0.00030260, -0.00026847, 0.00020717, -0.00022056, 0.00016180, -0.00013711, 0.00012053, -0.00012424, 0.00010307, -0.00009683, 0.00008955, -0.00009327, 0.00008810, -0.00008672, 0.00009014, -0.00009910, 0.00008240, -0.00006806, 0.00005850, -0.00006696, 0.00005437, -0.00004821, 0.00004983, -0.00004440, 0.00003946, -0.00004483, 0.00004250, -0.00003609, 0.00002433, -0.00003602, 0.00003990, -0.00002074, 0.00001334, -0.00001886, 0.00002141, -0.00002076, 0.00001467, -0.00001540, 0.00001784, -0.00001572, 0.00001689, -0.00001519, 0.00001126, -0.00001938, 0.00001857, -0.00001256, 0.00001208, -0.00000861, 0.00000419, -0.00001348, 0.00001779, -0.00001522, 0.00001753, -0.00000708, -0.00000143, },
		{ 0.00001130, -0.00000110, 0.00000115, -0.00000389, -0.00000162, -0.00000105, 0.00000235, 0.00000857, -0.00001102, -0.00000202, -0.00000207, 0.00001351, -0.00001295, 0.00001706, -0.00001792, 0.00000791, -0.00000879, 0.00001133, -0.00001259, 0.00001981, -0.00002022, 0.00001205, -0.00001653, 0.00001856, -0.00001700, 0.00001925, -0.00002070, 0.00001771, -0.00001786, 0.00001748, -0.00001804, 0.00001094, -0.00000931, 0.00002873, -0.00003535, 0.00002400, -0.00002338, 0.00002840, -0.00003805, 0.00003352, -0.00003388, 0.00003966, -0.00004311, 0.00004896, -0.00004988, 0.00003944, -0.00004939, 0.00006607, -0.00008391, 0.00008585, -0.00009725, 0.00009773, -0.00008862, 0.00007749, -0.00009348, 0.00009827, -0.00011379, 0.00010898, -0.00013898, 0.00015709, -0.00018037, 0.00015575, -0.00023053, 0.00025179, -0.00037806, 0.00036144, -0.00090958, 0.00032706, 0.00199351, -0.00147283, 0.00049867, -0.00048210, 0.00029567, -0.00026326, 0.00020204, -0.00021627, 0.00015632, -0.00013270, 0.00011714, -0.00012196, 0.00010112, -0.00009636, 0.00009006, -0.00009604, 0.00009149, -0.00008901, 0.00008987, -0.00009426, 0.00007404, -0.00005932, 0.00005191, -0.00006116, 0.00004873, -0.00004298, 0.00004436, -0.00003908, 0.00003568, -0.00004092, 0.00003686, -0.00003058, 0.00002183, -0.00003392, 0.00003390, -0.00001424, 0.00001066, -0.00001809, 0.00002041, -0.00001971, 0.00001464, -0.00001625, 0.00001832, -0.00001631, 0.00001756, -0.00001568, 0.00001261, -0.00002093, 0.00001872, -0.00001210, 0.00001182, -0.00000884, 0.00000635, -0.00001641, 0.00001883, -0.00001502, 0.00001576, -0.00000445, -0.00000149, -0.00001188, },
		{ 0.00000169, -0.00000024, -0.00000235, -0.00000301, -0.00000002, 0.00000010, 0.00001042, -0.00001021, -0.00000260, -0.00000401, 0.00001497, -0.00001284, 0.00001560, -0.00001409, 0.00000399, -0.00000639, 0.00000940, -0.00001085, 0.00001738, -0.00001606, 0.00000767, -0.00001269, 0.00001422, -0.00001245, 0.00001448, -0.00001495, 0.00001217, -0.00001295, 0.00001236, -0.00001240, 0.00000650, -0.00000867, 0.00002844, -0.00003113, 0.00001908, -0.00002099, 0.00002646, -0.00003425, 0.00002891, -0.00003038, 0.00003620, -0.00003902, 0.00004459, -0.00004513, 0.00003532, -0.00004774, 0.00006445, -0.00007903, 0.00007621, -0.00008291, 0.00007992, -0.00006942, 0.00005901, -0.00007477, 0.00007908, -0.00009294, 0.00008696, -0.00011504, 0.00013003, -0.00014783, 0.00011916, -0.00018749, 0.00019959, -0.00030410, 0.00026116, -0.00069928, -0.00007148, 0.00229526, -0.00149759, 0.00045047, -0.00044794, 0.00026163, -0.00023618, 0.00017849, -0.00019591, 0.00013666, -0.00011577, 0.00010254, -0.00010945, 0.00008973, -0.00008706, 0.00008214, -0.00009030, 0.00008576, -0.00008156, 0.00007955, -0.00008000, 0.00005772, -0.00004426, 0.00003998, -0.00005042, 0.00003845, -0.00003339, 0.00003489, -0.00003026, 0.00002857, -0.00003361, 0.00002822, -0.00002284, 0.00001718, -0.00002912, 0.00002555, -0.00000667, 0.00000745, -0.00001644, 0.00001831, -0.00001766, 0.00001361, -0.00001587, 0.00001741, -0.00001550, 0.00001681, -0.00001479, 0.00001255, -0.00002079, 0.00001713, -0.00001021, 0.00001043, -0.00000817, 0.00000746, -0.00001765, 0.00001783, -0.00001289, 0.00001242, -0.00000097, -0.00000212, -0.00001169, 0.00000865, },
		{ -0.00000162, -0.00000084, -0.00000425, 0.00000082, -0.00000169, 0.00001128, -0.00000847, -0.00000367, -0.00000520, 0.00001496, -0.00001094, 0.00001227, -0.00000875, -0.00000080, -0.00000326, 0.00000651, -0.00000784, 0.00001339, -0.00001043, 0.00000209, -0.00000771, 0.00000871, -0.00000676, 0.00000857, -0.00000824, 0.00000595, -0.00000738, 0.00000653, -0.00000628, 0.00000189, -0.00000732, 0.00002622, -0.00002464, 0.00001241, -0.00001677, 0.00002195, -0.00002752, 0.00002149, -0.00002394, 0.00002936, -0.00003121, 0.00003626, -0.00003618, 0.00002669, -0.00004061, 0.00005577, -0.00006573, 0.00005758, -0.00005994, 0.00005425, -0.00004314, 0.00003367, -0.00004892, 0.00005227, -0.00006387, 0.00005602, -0.00008116, 0.00009184, -0.00010292, 0.00006877, -0.00012851, 0.00012842, -0.00020769, 0.00013338, -0.00046361, -0.00045687, 0.00252933, -0.00144560, 0.00034996, -0.00037826, 0.00020043, -0.00018754, 0.00013688, -0.00015987, 0.00010323, -0.00008674, 0.00007708, -0.00008694, 0.00006905, -0.00006896, 0.00006570, -0.00007585, 0.00007078, -0.00006455, 0.00005988, -0.00005755, 0.00003496, -0.00002437, 0.00002402, -0.00003583, 0.00002453, -0.00002040, 0.00002236, -0.00001878, 0.00001885, -0.00002362, 0.00001734, -0.00001351, 0.00001087, -0.00002213, 0.00001553, 0.00000137, 0.00000398, -0.00001397, 0.00001511, -0.00001453, 0.00001143, -0.00001406, 0.00001490, -0.00001311, 0.00001448, -0.00001236, 0.00001088, -0.00001881, 0.00001376, -0.00000689, 0.00000790, -0.00000649, 0.00000729, -0.00001703, 0.00001480, -0.00000903, 0.00000787, 0.00000302, -0.00000314, -0.00001063, 0.00000521, 0.00000469, },
		{ 0.00000038, -0.00000511, 0.00000124, -0.00000280, 0.00001104, -0.00000582, -0.00000520, -0.00000556, 0.00001348, -0.00000741, 0.00000748, -0.00000247, -0.00000593, 0.00000015, 0.00000304, -0.00000396, 0.00000836, -0.00000394, -0.00000403, -0.00000221, 0.00000265, -0.00000057, 0.00000221, -0.00000128, -0.00000033, -0.00000169, 0.00000050, -0.00000011, -0.00000266, -0.00000528, 0.00002211, -0.00001615, 0.00000434, -0.00001094, 0.00001509, -0.00001821, 0.00001167, -0.00001493, 0.00001950, -0.00002011, 0.00002444, -0.00002352, 0.00001397, -0.00002841, 0.00004066, -0.00004519, 0.00003181, -0.00003073, 0.00002346, -0.00001257, 0.00000423, -0.00001866, 0.00002066, -0.00002953, 0.00001929, -0.00004078, 0.00004635, -0.00005004, 0.00000964, -0.00005958, 0.00004568, -0.00009851, -0.00000823, -0.00022284, -0.00080479, 0.00267761, -0.00131091, 0.00019860, -0.00027625, 0.00011530, -0.00012033, 0.00007987, -0.00011056, 0.00005826, -0.00004760, 0.00004252, -0.00005600, 0.00004046, -0.00004328, 0.00004184, -0.00005370, 0.00004772, -0.00003949, 0.00003284, -0.00002925, 0.00000818, -0.00000185, 0.00000586, -0.00001900, 0.00000842, -0.00000537, 0.00000803, -0.00000577, 0.00000757, -0.00001197, 0.00000524, -0.00000346, 0.00000357, -0.00001369, 0.00000468, 0.00000924, 0.00000054, -0.00001076, 0.00001086, -0.00001035, 0.00000806, -0.00001074, 0.00001078, -0.00000917, 0.00001059, -0.00000843, 0.00000764, -0.00001509, 0.00000881, -0.00000238, 0.00000440, -0.00000388, 0.00000586, -0.00001462, 0.00001003, -0.00000391, 0.00000261, 0.00000709, -0.00000427, -0.00000897, 0.00000137, 0.00000750, -0.00000266, },
		{ -0.00000541, 0.00000110, -0.00000311, 0.00000972, -0.00000244, -0.00000705, -0.00000518, 0.00001074, -0.00000270, 0.00000184, 0.00000405, -0.00001074, 0.00000335, -0.00000058, 0.00000029, 0.00000289, 0.00000271, -0.00000998, 0.00000311, -0.00000327, 0.00000542, -0.00000391, 0.00000525, -0.00000606, 0.00000359, -0.00000525, 0.00000571, -0.00000694, -0.00000262, 0.00001633, -0.00000615, -0.00000460, -0.00000395, 0.00000637, -0.00000698, 0.00000018, -0.00000408, 0.00000742, -0.00000662, 0.00001011, -0.00000816, -0.00000178, -0.00001232, 0.00002070, -0.00001959, 0.00000172, 0.00000144, -0.00000901, 0.00001889, -0.00002602, 0.00001272, -0.00001236, 0.00000648, -0.00001938, 0.00000190, -0.00000176, 0.00000551, -0.00005221, 0.00001227, -0.00004009, 0.00001266, -0.00014925, 0.00000350, -0.00109452, 0.00272839, -0.00109452, 0.00000350, -0.00014925, 0.00001266, -0.00004009, 0.00001227, -0.00005221, 0.00000551, -0.00000176, 0.00000190, -0.00001938, 0.00000648, -0.00001236, 0.00001272, -0.00002602, 0.00001889, -0.00000901, 0.00000144, 0.00000172, -0.00001959, 0.00002070, -0.00001232, -0.00000178, -0.00000816, 0.00001011, -0.00000662, 0.00000742, -0.00000408, 0.00000018, -0.00000698, 0.00000637, -0.00000395, -0.00000460, -0.00000615, 0.00001633, -0.00000262, -0.00000694, 0.00000571, -0.00000525, 0.00000359, -0.00000606, 0.00000525, -0.00000391, 0.00000542, -0.00000327, 0.00000311, -0.00000998, 0.00000271, 0.00000289, 0.00000029, -0.00000058, 0.00000335, -0.00001074, 0.00000405, 0.00000184, -0.00000270, 0.00001074, -0.00000518, -0.00000705, -0.00000244, 0.00000972, -0.00000311, 0.00000110, },
		{ 0.00000038, -0.00000266, 0.00000750, 0.00000137, -0.00000897, -0.00000427, 0.00000709, 0.00000261, -0.00000391, 0.00001003, -0.00001462, 0.00000586, -0.00000388, 0.00000440, -0.00000238, 0.00000881, -0.00001509, 0.00000764, -0.00000843, 0.00001059, -0.00000917, 0.00001078, -0.00001074, 0.00000806, -0.00001035, 0.00001086, -0.00001076, 0.00000054, 0.00000924, 0.00000468, -0.00001369, 0.00000357, -0.00000346, 0.00000524, -0.00001197, 0.00000757, -0.00000577, 0.00000803, -0.00000537, 0.00000842, -0.00001900, 0.00000586, -0.00000185, 0.00000818, -0.00002925, 0.00003284, -0.00003949, 0.00004772, -0.00005370, 0.00004184, -0.00004328, 0.00004046, -0.00005600, 0.00004252, -0.00004760, 0.00005826, -0.00011056, 0.00007987, -0.00012033, 0.00011530, -0.00027625, 0.00019860, -0.00131091, 0.00267761, -0.00080479, -0.00022284, -0.00000823, -0.00009851, 0.00004568, -0.00005958, 0.00000964, -0.00005004, 0.00004635, -0.00004078, 0.00001929, -0.00002953, 0.00002066, -0.00001866, 0.00000423, -0.00001257, 0.00002346, -0.00003073, 0.00003181, -0.00004519, 0.00004066, -0.00002841, 0.00001397, -0.00002352, 0.00002444, -0.00002011, 0.00001950, -0.00001493, 0.00001167, -0.00001821, 0.00001509, -0.00001094, 0.00000434, -0.00001615, 0.00002211, -0.00000528, -0.00000266, -0.00000011, 0.00000050, -0.00000169, -0.00000033, -0.00000128, 0.00000221, -0.00000057, 0.00000265, -0.00000221, -0.00000403, -0.00000394, 0.00000836, -0.00000396, 0.00000304, 0.00000015, -0.00000593, -0.00000247, 0.00000748, -0.00000741, 0.00001348, -0.00000556, -0.00000520, -0.00000582, 0.00001104, -0.00000280, 0.00000124, -0.00000511, },
		{ -0.00000162, 0.00000469, 0.00000521, -0.00001063, -0.00000314, 0.00000302, 0.00000787, -0.00000903, 0.00001480, -0.00001703, 0.00000729, -0.00000649, 0.00000790, -0.00000689, 0.00001376, -0.00001881, 0.00001088, -0.00001236, 0.00001448, -0.00001311, 0.00001490, -0.00001406, 0.00001143, -0.00001453, 0.00001511, -0.00001397, 0.00000398, 0.00000137, 0.00001553, -0.00002213, 0.00001087, -0.00001351, 0.00001734, -0.00002362, 0.00001885, -0.00001878, 0.00002236, -0.00002040, 0.00002453, -0.00003583, 0.00002402, -0.00002437, 0.00003496, -0.00005755, 0.00005988, -0.00006455, 0.00007078, -0.00007585, 0.00006570, -0.00006896, 0.00006905, -0.00008694, 0.00007708, -0.00008674, 0.00010323, -0.00015987, 0.00013688, -0.00018754, 0.00020043, -0.00037826, 0.00034996, -0.00144560, 0.00252933, -0.00045687, -0.00046361, 0.00013338, -0.00020769, 0.00012842, -0.00012851, 0.00006877, -0.00010292, 0.00009184, -0.00008116, 0.00005602, -0.00006387, 0.00005227, -0.00004892, 0.00003367, -0.00004314, 0.00005425, -0.00005994, 0.00005758, -0.00006573, 0.00005577, -0.00004061, 0.00002669, -0.00003618, 0.00003626, -0.00003121, 0.00002936, -0.00002394, 0.00002149, -0.00002752, 0.00002195, -0.00001677, 0.00001241, -0.00002464, 0.00002622, -0.00000732, 0.00000189, -0.00000628, 0.00000653, -0.00000738, 0.00000595, -0.00000824, 0.00000857, -0.00000676, 0.00000871, -0.00000771, 0.00000209, -0.00001043, 0.00001339, -0.00000784, 0.00000651, -0.00000326, -0.00000080, -0.00000875, 0.00001227, -0.00001094, 0.00001496, -0.00000520, -0.00000367, -0.00000847, 0.00001128, -0.00000169, 0.00000082, -0.00000425, -0.00000084, },
		{ 0.00000169, 0.00000865, -0.00001169, -0.00000212, -0.00000097, 0.00001242, -0.00001289, 0.00001783, -0.00001765, 0.00000746, -0.00000817, 0.00001043, -0.00001021, 0.00001713, -0.00002079, 0.00001255, -0.00001479, 0.00001681, -0.00001550, 0.00001741, -0.00001587, 0.00001361, -0.00001766, 0.00001831, -0.00001644, 0.00000745, -0.00000667, 0.00002555, -0.00002912, 0.00001718, -0.00002284, 0.00002822, -0.00003361, 0.00002857, -0.00003026, 0.00003489, -0.00003339, 0.00003845, -0.00005042, 0.00003998, -0.00004426, 0.00005772, -0.00008000, 0.00007955, -0.00008156, 0.00008576, -0.00009030, 0.00008214, -0.00008706, 0.00008973, -0.00010945, 0.00010254, -0.00011577, 0.00013666, -0.00019591, 0.00017849, -0.00023618, 0.00026163, -0.00044794, 0.00045047, -0.00149759, 0.00229526, -0.00007148, -0.00069928, 0.00026116, -0.00030410, 0.00019959, -0.00018749, 0.00011916, -0.00014783, 0.00013003, -0.00011504, 0.00008696, -0.00009294, 0.00007908, -0.00007477, 0.00005901, -0.00006942, 0.00007992, -0.00008291, 0.00007621, -0.00007903, 0.00006445, -0.00004774, 0.00003532, -0.00004513, 0.00004459, -0.00003902, 0.00003620, -0.00003038, 0.00002891, -0.00003425, 0.00002646, -0.00002099, 0.00001908, -0.00003113, 0.00002844, -0.00000867, 0.00000650, -0.00001240, 0.00001236, -0.00001295, 0.00001217, -0.00001495, 0.00001448, -0.00001245, 0.00001422, -0.00001269, 0.00000767, -0.00001606, 0.00001738, -0.00001085, 0.00000940, -0.00000639, 0.00000399, -0.00001409, 0.00001560, -0.00001284, 0.00001497, -0.00000401, -0.00000260, -0.00001021, 0.00001042, 0.00000010, -0.00000002, -0.00000301, -0.00000235, -0.00000024, },
		{ 0.00001130, -0.00001188, -0.00000149, -0.00000445, 0.00001576, -0.00001502, 0.00001883, -0.00001641, 0.00000635, -0.00000884, 0.00001182, -0.00001210, 0.00001872, -0.00002093, 0.00001261, -0.00001568, 0.00001756, -0.00001631, 0.00001832, -0.00001625, 0.00001464, -0.00001971, 0.00002041, -0.00001809, 0.00001066, -0.00001424, 0.00003390, -0.00003392, 0.00002183, -0.00003058, 0.00003686, -0.00004092, 0.00003568, -0.00003908, 0.00004436, -0.00004298, 0.00004873, -0.00006116, 0.00005191, -0.00005932, 0.00007404, -0.00009426, 0.00008987, -0.00008901, 0.00009149, -0.00009604, 0.00009006, -0.00009636, 0.00010112, -0.00012196, 0.00011714, -0.00013270, 0.00015632, -0.00021627, 0.00020204, -0.00026326, 0.00029567, -0.00048210, 0.00049867, -0.00147283, 0.00199351, 0.00032706, -0.00090958, 0.00036144, -0.00037806, 0.00025179, -0.00023053, 0.00015575, -0.00018037, 0.00015709, -0.00013898, 0.00010898, -0.00011379, 0.00009827, -0.00009348, 0.00007749, -0.00008862, 0.00009773, -0.00009725, 0.00008585, -0.00008391, 0.00006607, -0.00004939, 0.00003944, -0.00004988, 0.00004896, -0.00004311, 0.00003966, -0.00003388, 0.00003352, -0.00003805, 0.00002840, -0.00002338, 0.00002400, -0.00003535, 0.00002873, -0.00000931, 0.00001094, -0.00001804, 0.00001748, -0.00001786, 0.00001771, -0.00002070, 0.00001925, -0.00001700, 0.00001856, -0.00001653, 0.00001205, -0.00002022, 0.00001981, -0.00001259, 0.00001133, -0.00000879, 0.00000791, -0.00001792, 0.00001706, -0.00001295, 0.00001351, -0.00000207, -0.00000202, -0.00001102, 0.00000857, 0.00000235, -0.00000105, -0.00000162, -0.00000389, 0.00000115, -0.00000110, },
		{ -0.00001103, -0.00000143, -0.00000708, 0.00001753, -0.00001522, 0.00001779, -0.00001348, 0.00000419, -0.00000861, 0.00001208, -0.00001256, 0.00001857, -0.00001938, 0.00001126, -0.00001519, 0.00001689, -0.00001572, 0.00001784, -0.00001540, 0.00001467, -0.00002076, 0.00002141, -0.00001886, 0.00001334, -0.00002074, 0.00003990, -0.00003602, 0.00002433, -0.00003609, 0.00004250, -0.00004483, 0.00003946, -0.00004440, 0.00004983, -0.00004821, 0.00005437, -0.00006696, 0.00005850, -0.00006806, 0.00008240, -0.00009910, 0.00009014, -0.00008672, 0.00008810, -0.00009327, 0.00008955, -0.00009683, 0.00010307, -0.00012424, 0.00012053, -0.00013711, 0.00016180, -0.00022056, 0.00020717, -0.00026847, 0.00030260, -0.00048172, 0.00049840, -0.00138322, 0.00164672, 0.00071314, -0.00107575, 0.00042301, -0.00042235, 0.00027978, -0.00025359, 0.00017517, -0.00019769, 0.00017053, -0.00015083, 0.00012005, -0.00012453, 0.00010803, -0.00010326, 0.00008734, -0.00009896, 0.00010609, -0.00010173, 0.00008591, -0.00008036, 0.00006096, -0.00004594, 0.00003927, -0.00005053, 0.00004941, -0.00004353, 0.00003979, -0.00003446, 0.00003528, -0.00003891, 0.00002786, -0.00002396, 0.00002707, -0.00003724, 0.00002721, -0.00000933, 0.00001498, -0.00002277, 0.00002141, -0.00002156, 0.00002196, -0.00002488, 0.00002229, -0.00001986, 0.00002121, -0.00001874, 0.00001477, -0.00002250, 0.00002042, -0.00001285, 0.00001209, -0.00001016, 0.00001053, -0.00001988, 0.00001649, -0.00001131, 0.00001081, 0.00000039, -0.00000181, -0.00001101, 0.00000599, 0.00000473, -0.00000199, -0.00000033, -0.00000519, 0.00000225, -0.00000334, 0.00001285, },
		{ -0.00000199, -0.00000872, 0.00001764, -0.00001357, 0.00001500, -0.00000929, 0.00000136, -0.00000773, 0.00001138, -0.00001175, 0.00001695, -0.00001650, 0.00000886, -0.00001369, 0.00001513, -0.00001406, 0.00001630, -0.00001365, 0.00001396, -0.00002095, 0.00002140, -0.00001873, 0.00001524, -0.00002566, 0.00004307, -0.00003516, 0.00002444, -0.00003896, 0.00004469, -0.00004499, 0.00003960, -0.00004583, 0.00005090, -0.00004868, 0.00005497, -0.00006737, 0.00005923, -0.00006987, 0.00008240, -0.00009457, 0.00008103, -0.00007582, 0.00007694, -0.00008332, 0.00008184, -0.00008957, 0.00009664, -0.00011732, 0.00011377, -0.00013013, 0.00015437, -0.00021029, 0.00019565, -0.00025407, 0.00028552, -0.00045146, 0.00045790, -0.00124487, 0.00127963, 0.00106241, -0.00118256, 0.00043858, -0.00043314, 0.00028122, -0.00025507, 0.00017620, -0.00019882, 0.00016961, -0.00014993, 0.00011960, -0.00012458, 0.00010778, -0.00010355, 0.00008793, -0.00009986, 0.00010464, -0.00009649, 0.00007705, -0.00006949, 0.00005037, -0.00003850, 0.00003564, -0.00004770, 0.00004649, -0.00004078, 0.00003704, -0.00003248, 0.00003446, -0.00003715, 0.00002516, -0.00002296, 0.00002835, -0.00003694, 0.00002416, -0.00000882, 0.00001840, -0.00002618, 0.00002373, -0.00002365, 0.00002447, -0.00002704, 0.00002323, -0.00002072, 0.00002190, -0.00001907, 0.00001558, -0.00002271, 0.00001913, -0.00001164, 0.00001164, -0.00001037, 0.00001170, -0.00001990, 0.00001404, -0.00000824, 0.00000727, 0.00000304, -0.00000177, -0.00001038, 0.00000302, 0.00000688, -0.00000254, 0.00000060, -0.00000603, 0.00000281, -0.00000476, 0.00001316, -0.00000913, },
		{ -0.00000938, 0.00001620, -0.00001039, 0.00001096, -0.00000443, -0.00000162, -0.00000657, 0.00001000, -0.00000999, 0.00001426, -0.00001277, 0.00000591, -0.00001160, 0.00001271, -0.00001176, 0.00001411, -0.00001137, 0.00001277, -0.00002045, 0.00002048, -0.00001774, 0.00001619, -0.00002866, 0.00004320, -0.00003139, 0.00002224, -0.00003916, 0.00004342, -0.00004150, 0.00003623, -0.00004350, 0.00004769, -0.00004461, 0.00005079, -0.00006269, 0.00005438, -0.00006513, 0.00007471, -0.00008194, 0.00006440, -0.00005857, 0.00006035, -0.00006843, 0.00006900, -0.00007657, 0.00008383, -0.00010327, 0.00009904, -0.00011412, 0.00013672, -0.00018852, 0.00017106, -0.00022431, 0.00024984, -0.00039867, 0.00038831, -0.00107615, 0.00091669, 0.00135413, -0.00122005, 0.00040585, -0.00041064, 0.00025702, -0.00023605, 0.00015992, -0.00018482, 0.00015537, -0.00013728, 0.00010849, -0.00011472, 0.00009824, -0.00009499, 0.00007992, -0.00009203, 0.00009437, -0.00008292, 0.00006112, -0.00005336, 0.00003629, -0.00002870, 0.00002978, -0.00004237, 0.00004107, -0.00003566, 0.00003214, -0.00002856, 0.00003160, -0.00003330, 0.00002086, -0.00002077, 0.00002811, -0.00003476, 0.00001994, -0.00000793, 0.00002102, -0.00002799, 0.00002419, -0.00002386, 0.00002499, -0.00002696, 0.00002196, -0.00001953, 0.00002062, -0.00001753, 0.00001452, -0.00002098, 0.00001615, -0.00000918, 0.00001014, -0.00000950, 0.00001143, -0.00001816, 0.00001007, -0.00000426, 0.00000343, 0.00000546, -0.00000165, -0.00000940, 0.00000003, 0.00000847, -0.00000249, 0.00000101, -0.00000624, 0.00000271, -0.00000525, 0.00001225, -0.00000633, -0.00000307, },
		{ 0.00001352, -0.00000619, 0.00000631, 0.00000043, -0.00000422, -0.00000549, 0.00000829, -0.00000768, 0.00001100, -0.00000877, 0.00000293, -0.00000940, 0.00001007, -0.00000923, 0.00001169, -0.00000891, 0.00001138, -0.00001943, 0.00001879, -0.00001597, 0.00001612, -0.00002962, 0.00004038, -0.00002509, 0.00001811, -0.00003695, 0.00003903, -0.00003490, 0.00002996, -0.00003797, 0.00004088, -0.00003677, 0.00004272, -0.00005386, 0.00004497, -0.00005505, 0.00006101, -0.00006346, 0.00004303, -0.00003795, 0.00004124, -0.00005135, 0.00005359, -0.00006034, 0.00006722, -0.00008478, 0.00007924, -0.00009229, 0.00011241, -0.00015933, 0.00013805, -0.00018470, 0.00020241, -0.00033223, 0.00030216, -0.00089555, 0.00057972, 0.00157302, -0.00118464, 0.00032785, -0.00035910, 0.00021119, -0.00020017, 0.00012955, -0.00015859, 0.00013048, -0.00011530, 0.00008891, -0.00009696, 0.00008130, -0.00007939, 0.00006502, -0.00007727, 0.00007734, -0.00006343, 0.00004079, -0.00003464, 0.00002106, -0.00001839, 0.00002310, -0.00003571, 0.00003421, -0.00002916, 0.00002599, -0.00002347, 0.00002737, -0.00002805, 0.00001560, -0.00001786, 0.00002667, -0.00003111, 0.00001497, -0.00000681, 0.00002271, -0.00002803, 0.00002269, -0.00002218, 0.00002351, -0.00002473, 0.00001863, -0.00001651, 0.00001761, -0.00001438, 0.00001186, -0.00001766, 0.00001193, -0.00000587, 0.00000790, -0.00000778, 0.00000999, -0.00001503, 0.00000517, -0.00000000, -0.00000015, 0.00000726, -0.00000122, -0.00000834, -0.00000264, 0.00000924, -0.00000172, 0.00000084, -0.00000581, 0.00000194, -0.00000487, 0.00001028, -0.00000290, -0.00000445, -0.00000923, },
		{ -0.00000160, 0.00000177, 0.00000464, -0.00000596, -0.00000483, 0.00000657, -0.00000523, 0.00000768, -0.00000501, 0.00000040, -0.00000748, 0.00000759, -0.00000685, 0.00000936, -0.00000658, 0.00000998, -0.00001804, 0.00001648, -0.00001353, 0.00001507, -0.00002860, 0.00003500, -0.00001691, 0.00001270, -0.00003290, 0.00003222, -0.00002607, 0.00002171, -0.00003023, 0.00003154, -0.00002641, 0.00003210, -0.00004232, 0.00003259, -0.00004149, 0.00004363, -0.00004201, 0.00002015, -0.00001724, 0.00002268, -0.00003487, 0.00003824, -0.00004350, 0.00004952, -0.00006477, 0.00005752, -0.00006810, 0.00008537, -0.00012714, 0.00010169, -0.00014116, 0.00015044, -0.00026125, 0.00021165, -0.00071976, 0.00028602, 0.00171061, -0.00107951, 0.00021277, -0.00028633, 0.00015038, -0.00015307, 0.00008992, -0.00012445, 0.00009883, -0.00008751, 0.00006404, -0.00007423, 0.00005969, -0.00005934, 0.00004578, -0.00005819, 0.00005635, -0.00004107, 0.00001915, -0.00001617, 0.00000704, -0.00000935, 0.00001692, -0.00002887, 0.00002699, -0.00002228, 0.00001949, -0.00001800, 0.00002251, -0.00002212, 0.00001004, -0.00001473, 0.00002445, -0.00002645, 0.00000969, -0.00000562, 0.00002340, -0.00002629, 0.00001938, -0.00001879, 0.00002027, -0.00002067, 0.00001368, -0.00001214, 0.00001336, -0.00001011, 0.00000813, -0.00001331, 0.00000703, -0.00000226, 0.00000536, -0.00000556, 0.00000775, -0.00001107, -0.00000000, 0.00000388, -0.00000295, 0.00000811, -0.00000030, -0.00000740, -0.00000475, 0.00000905, -0.00000024, 0.00000016, -0.00000483, 0.00000061, -0.00000380, 0.00000760, 0.00000074, -0.00000580, -0.00000857, 0.00001005, },
		{ -0.00000202, 0.00000765, -0.00000652, -0.00000479, 0.00000510, -0.00000299, 0.00000471, -0.00000195, -0.00000133, -0.00000613, 0.00000554, -0.00000489, 0.00000739, -0.00000457, 0.00000872, -0.00001641, 0.00001372, -0.00001061, 0.00001318, -0.00002588, 0.00002768, -0.00000774, 0.00000682, -0.00002776, 0.00002392, -0.00001613, 0.00001267, -0.00002148, 0.00002101, -0.00001501, 0.00002053, -0.00002982, 0.00001914, -0.00002665, 0.00002519, -0.00002064, -0.00000102, 0.00000055, 0.00000735, -0.00002138, 0.00002519, -0.00002835, 0.00003320, -0.00004589, 0.00003680, -0.00004482, 0.00005928, -0.00009610, 0.00006668, -0.00009920, 0.00010055, -0.00019390, 0.00012737, -0.00056216, 0.00004714, 0.00176565, -0.00091421, 0.00007298, -0.00020269, 0.00008286, -0.00010157, 0.00004683, -0.00008744, 0.00006494, -0.00005791, 0.00003751, -0.00004987, 0.00003658, -0.00003785, 0.00002513, -0.00003777, 0.00003454, -0.00001902, -0.00000075, -0.00000058, -0.00000377, -0.00000300, 0.00001230, -0.00002278, 0.00002031, -0.00001588, 0.00001343, -0.00001284, 0.00001766, -0.00001618, 0.00000479, -0.00001181, 0.00002180, -0.00002125, 0.00000452, -0.00000452, 0.00002311, -0.00002292, 0.00001457, -0.00001412, 0.00001576, -0.00001534, 0.00000777, -0.00000708, 0.00000852, -0.00000536, 0.00000396, -0.00000860, 0.00000215, 0.00000111, 0.00000295, -0.00000325, 0.00000519, -0.00000686, -0.00000477, 0.00000681, -0.00000457, 0.00000785, 0.00000116, -0.00000670, -0.00000614, 0.00000792, 0.00000181, -0.00000087, -0.00000347, -0.00000103, -0.00000233, 0.00000459, 0.00000416, -0.00000680, -0.00000771, 0.00000629, 0.00000275, },
		{ 0.00000915, -0.00000574, -0.00000545, 0.00000400, -0.00000119, 0.00000241, 0.00000016, -0.00000208, -0.00000548, 0.00000405, -0.00000347, 0.00000589, -0.00000299, 0.00000766, -0.00001464, 0.00001068, -0.00000742, 0.00001071, -0.00002190, 0.00001923, 0.00000145, 0.00000133, -0.00002237, 0.00001515, -0.00000626, 0.00000404, -0.00001297, 0.00001069, -0.00000410, 0.00000967, -0.00001808, 0.00000650, -0.00001268, 0.00000823, -0.00000211, -0.00001777, 0.00001304, -0.00000282, -0.00001255, 0.00001602, -0.00001656, 0.00002010, -0.00003020, 0.00001941, -0.00002507, 0.00003710, -0.00006956, 0.00003684, -0.00006322, 0.00005800, -0.00013653, 0.00005715, -0.00043182, -0.00013167, 0.00174384, -0.00070354, -0.00007635, -0.00011971, 0.00001742, -0.00005271, 0.00000613, -0.00005263, 0.00003327, -0.00003050, 0.00001287, -0.00002718, 0.00001505, -0.00001789, 0.00000596, -0.00001889, 0.00001483, -0.00000018, -0.00001634, 0.00001013, -0.00001001, -0.00000016, 0.00000983, -0.00001802, 0.00001483, -0.00001060, 0.00000840, -0.00000852, 0.00001335, -0.00001077, 0.00000032, -0.00000943, 0.00001906, -0.00001594, -0.00000018, -0.00000364, 0.00002194, -0.00001823, 0.00000879, -0.00000876, 0.00001059, -0.00000946, 0.00000164, -0.00000208, 0.00000379, -0.00000081, 0.00000001, -0.00000421, -0.00000210, 0.00000369, 0.00000111, -0.00000124, 0.00000274, -0.00000299, -0.00000857, 0.00000836, -0.00000480, 0.00000648, 0.00000309, -0.00000623, -0.00000682, 0.00000598, 0.00000418, -0.00000199, -0.00000199, -0.00000273, -0.00000079, 0.00000169, 0.00000695, -0.00000713, -0.00000697, 0.00000269, 0.00000630, -0.00000456, },
		{ -0.00000373, -0.00000668, 0.00000329, 0.00000004, 0.00000092, 0.00000120, -0.00000184, -0.00000551, 0.00000312, -0.00000261, 0.00000488, -0.00000182, 0.00000681, -0.00001278, 0.00000752, -0.00000421, 0.00000796, -0.00001721, 0.00001052, 0.00000969, -0.00000296, -0.00001753, 0.00000690, 0.00000242, -0.00000308, -0.00000581, 0.00000181, 0.00000496, 0.00000090, -0.00000858, -0.00000372, -0.00000141, -0.00000521, 0.00001145, -0.00002824, 0.00001888, -0.00000688, -0.00000909, 0.00001143, -0.00000895, 0.00001125, -0.00001894, 0.00000679, -0.00001053, 0.00002076, -0.00004970, 0.00001461, -0.00003608, 0.00002615, -0.00009307, 0.00000563, -0.00033318, -0.00025121, 0.00165671, -0.00046584, -0.00021904, -0.00004860, -0.00003788, -0.00001275, -0.00002709, -0.00002440, 0.00000762, -0.00000859, -0.00000687, -0.00000892, -0.00000230, -0.00000191, -0.00000936, -0.00000394, -0.00000041, 0.00001331, -0.00002590, 0.00001487, -0.00001116, -0.00000098, 0.00000955, -0.00001480, 0.00001084, -0.00000677, 0.00000471, -0.00000534, 0.00000988, -0.00000626, -0.00000304, -0.00000782, 0.00001647, -0.00001088, -0.00000412, -0.00000307, 0.00002005, -0.00001266, 0.00000264, -0.00000336, 0.00000548, -0.00000377, -0.00000393, 0.00000215, -0.00000016, 0.00000292, -0.00000314, -0.00000069, -0.00000520, 0.00000512, 0.00000010, 0.00000020, 0.00000077, 0.00000012, -0.00001101, 0.00000835, -0.00000367, 0.00000418, 0.00000528, -0.00000592, -0.00000689, 0.00000349, 0.00000653, -0.00000291, -0.00000068, -0.00000418, 0.00000049, -0.00000073, 0.00000878, -0.00000661, -0.00000656, -0.00000036, 0.00000866, -0.00000558, 0.00000902, },
		{ -0.00000825, 0.00000285, 0.00000073, 0.00000023, 0.00000125, -0.00000077, -0.00000603, 0.00000263, -0.00000220, 0.00000427, -0.00000102, 0.00000613, -0.00001091, 0.00000445, -0.00000124, 0.00000527, -0.00001237, 0.00000241, 0.00001615, -0.00000541, -0.00001384, -0.00000003, 0.00000902, -0.00000785, -0.00000085, -0.00000469, 0.00001118, -0.00000478, -0.00000235, -0.00001041, 0.00000594, -0.00001379, 0.00001885, -0.00003167, 0.00001783, -0.00000501, -0.00001071, 0.00001117, -0.00000550, 0.00000683, -0.00001243, -0.00000060, -0.00000182, 0.00001102, -0.00003739, 0.00000097, -0.00001889, 0.00000631, -0.00006492, -0.00002588, -0.00026625, -0.00031747, 0.00152000, -0.00022096, -0.00033984, 0.00000128, -0.00007683, 0.00001374, -0.00004920, -0.00000576, -0.00000940, 0.00000558, -0.00001975, 0.00000307, -0.00001373, 0.00000845, -0.00001923, 0.00000552, -0.00000971, 0.00002025, -0.00002875, 0.00001352, -0.00000759, -0.00000490, 0.00001104, -0.00001293, 0.00000834, -0.00000444, 0.00000241, -0.00000336, 0.00000740, -0.00000284, -0.00000519, -0.00000702, 0.00001416, -0.00000632, -0.00000711, -0.00000290, 0.00001764, -0.00000670, -0.00000321, 0.00000139, 0.00000110, 0.00000102, -0.00000826, 0.00000502, -0.00000284, 0.00000538, -0.00000504, 0.00000154, -0.00000684, 0.00000521, 0.00000006, 0.00000091, -0.00000048, 0.00000218, -0.00001193, 0.00000682, -0.00000140, 0.00000129, 0.00000744, -0.00000558, -0.00000653, 0.00000078, 0.00000853, -0.00000337, 0.00000024, -0.00000515, 0.00000125, -0.00000239, 0.00000946, -0.00000519, -0.00000658, -0.00000265, 0.00000963, -0.00000505, 0.00000743, -0.00000076, },
		{ 0.00000250, 0.00000098, 0.00000022, 0.00000052, 0.00000085, -0.00000678, 0.00000236, -0.00000209, 0.00000393, -0.00000045, 0.00000557, -0.00000908, 0.00000162, 0.00000126, 0.00000298, -0.00000792, -0.00000441, 0.00002022, -0.00000567, -0.00001167, -0.00000510, 0.00001300, -0.00000984, 0.00000149, -0.00000834, 0.00001408, -0.00000691, 0.00000018, -0.00001313, 0.00000887, -0.00001709, 0.00001994, -0.00002840, 0.00001080, 0.00000163, -0.00001623, 0.00001423, -0.00000541, 0.00000620, -0.00001016, -0.00000315, 0.00000141, 0.00000755, -0.00003225, -0.00000451, -0.00001118, -0.00000212, -0.00005114, -0.00003897, -0.00022742, -0.00034026, 0.00135172, 0.00001189, -0.00042619, 0.00002360, -0.00009591, 0.00002444, -0.00005853, 0.00000192, -0.00001670, 0.00001111, -0.00002498, 0.00000809, -0.00001858, 0.00001255, -0.00002306, 0.00000895, -0.00001265, 0.00002056, -0.00002529, 0.00000697, -0.00000042, -0.00001084, 0.00001347, -0.00001193, 0.00000705, -0.00000340, 0.00000135, -0.00000245, 0.00000585, -0.00000051, -0.00000613, -0.00000697, 0.00001221, -0.00000246, -0.00000903, -0.00000316, 0.00001496, -0.00000089, -0.00000813, 0.00000492, -0.00000201, 0.00000438, -0.00001089, 0.00000618, -0.00000396, 0.00000633, -0.00000552, 0.00000233, -0.00000694, 0.00000398, 0.00000092, 0.00000091, -0.00000096, 0.00000313, -0.00001137, 0.00000404, 0.00000158, -0.00000176, 0.00000923, -0.00000502, -0.00000597, -0.00000183, 0.00000987, -0.00000316, 0.00000061, -0.00000549, 0.00000135, -0.00000316, 0.00000900, -0.00000295, -0.00000700, -0.00000413, 0.00000923, -0.00000316, 0.00000476, 0.00000271, -0.00000978, },
		{ 0.00000098, 0.00000067, -0.00000071, 0.00000268, -0.00000747, 0.00000210, -0.00000210, 0.00000370, -0.00000002, 0.00000507, -0.00000733, -0.00000080, 0.00000308, 0.00000134, -0.00000429, -0.00000941, 0.00002160, -0.00000370, -0.00001110, -0.00000812, 0.00001421, -0.00000904, 0.00000122, -0.00000911, 0.00001370, -0.00000564, -0.00000079, -0.00001210, 0.00000765, -0.00001551, 0.00001551, -0.00001977, -0.00000042, 0.00001111, -0.00002388, 0.00001907, -0.00000737, 0.00000821, -0.00001101, -0.00000199, 0.00000031, 0.00000912, -0.00003290, -0.00000338, -0.00001118, -0.00000126, -0.00004901, -0.00003741, -0.00021053, -0.00033157, 0.00117001, 0.00021607, -0.00046958, 0.00001596, -0.00009471, 0.00001955, -0.00005546, -0.00000093, -0.00001475, 0.00000854, -0.00002308, 0.00000655, -0.00001724, 0.00001080, -0.00002125, 0.00000676, -0.00000982, 0.00001515, -0.00001685, -0.00000315, 0.00000866, -0.00001734, 0.00001584, -0.00001119, 0.00000655, -0.00000329, 0.00000119, -0.00000237, 0.00000507, 0.00000086, -0.00000604, -0.00000753, 0.00001060, 0.00000061, -0.00000987, -0.00000384, 0.00001224, 0.00000430, -0.00001163, 0.00000682, -0.00000350, 0.00000601, -0.00001160, 0.00000552, -0.00000350, 0.00000581, -0.00000463, 0.00000176, -0.00000567, 0.00000170, 0.00000243, 0.00000035, -0.00000075, 0.00000306, -0.00000960, 0.00000044, 0.00000476, -0.00000447, 0.00001034, -0.00000408, -0.00000541, -0.00000405, 0.00001034, -0.00000219, 0.00000039, -0.00000519, 0.00000077, -0.00000308, 0.00000752, -0.00000015, -0.00000764, -0.00000488, 0.00000768, -0.00000031, 0.00000151, 0.00000615, -0.00001091, 0.00000202, },
		{ 0.00000135, -0.00000211, 0.00000437, -0.00000782, 0.00000167, -0.00000207, 0.00000346, 0.00000039, 0.00000460, -0.00000575, -0.00000268, 0.00000408, 0.00000052, -0.00000173, -0.00001235, 0.00002034, 0.00000022, -0.00001189, -0.00000921, 0.00001286, -0.00000582, -0.00000121, -0.00000744, 0.00001056, -0.00000161, -0.00000453, -0.00000813, 0.00000320, -0.00001019, 0.00000713, -0.00000783, -0.00001350, 0.00002115, -0.00003164, 0.00002393, -0.00000985, 0.00001141, -0.00001356, 0.00000141, -0.00000353, 0.00001397, -0.00003735, 0.00000209, -0.00001628, 0.00000584, -0.00005467, -0.00002621, -0.00020812, -0.00030387, 0.00099134, 0.00037911, -0.00046647, -0.00001993, -0.00007594, 0.00000164, -0.00004231, -0.00001226, -0.00000544, -0.00000041, -0.00001562, -0.00000010, -0.00001105, 0.00000447, -0.00001505, 0.00000023, -0.00000264, 0.00000574, -0.00000543, -0.00001472, 0.00001772, -0.00002289, 0.00001714, -0.00001011, 0.00000639, -0.00000368, 0.00000155, -0.00000277, 0.00000479, 0.00000145, -0.00000516, -0.00000847, 0.00000925, 0.00000288, -0.00000969, -0.00000490, 0.00000966, 0.00000846, -0.00001338, 0.00000690, -0.00000329, 0.00000589, -0.00001047, 0.00000325, -0.00000172, 0.00000406, -0.00000263, 0.00000013, -0.00000339, -0.00000124, 0.00000424, -0.00000051, -0.00000006, 0.00000223, -0.00000704, -0.00000342, 0.00000761, -0.00000644, 0.00001054, -0.00000264, -0.00000501, -0.00000568, 0.00000985, -0.00000052, -0.00000034, -0.00000432, -0.00000036, -0.00000232, 0.00000533, 0.00000287, -0.00000826, -0.00000510, 0.00000533, 0.00000299, -0.00000173, 0.00000905, -0.00001129, 0.00000122, 0.00000089, },
		{ -0.00000339, 0.00000565, -0.00000764, 0.00000093, -0.00000191, 0.00000311, 0.00000080, 0.00000416, -0.00000441, -0.00000393, 0.00000422, 0.00000056, -0.00000035, -0.00001324, 0.00001678, 0.00000551, -0.00001356, -0.00000875, 0.00000950, -0.00000090, -0.00000507, -0.00000411, 0.00000554, 0.00000415, -0.00000992, -0.00000242, -0.00000315, -0.00000275, -0.00000321, 0.00000503, -0.00002594, 0.00002951, -0.00003760, 0.00002720, -0.00001145, 0.00001443, -0.00001634, 0.00000548, -0.00000840, 0.00002015, -0.00004339, 0.00000939, -0.00002364, 0.00001578, -0.00006397, -0.00001072, -0.00021273, -0.00026861, 0.00082899, 0.00049367, -0.00041851, -0.00007846, -0.00004491, -0.00002484, -0.00002281, -0.00002885, 0.00000839, -0.00001313, -0.00000492, -0.00000977, -0.00000197, -0.00000456, -0.00000629, -0.00000880, 0.00000695, -0.00000552, 0.00000667, -0.00002555, 0.00002493, -0.00002621, 0.00001661, -0.00000825, 0.00000620, -0.00000420, 0.00000206, -0.00000335, 0.00000478, 0.00000153, -0.00000376, -0.00000955, 0.00000809, 0.00000438, -0.00000861, -0.00000624, 0.00000738, 0.00001137, -0.00001327, 0.00000524, -0.00000152, 0.00000424, -0.00000784, -0.00000020, 0.00000094, 0.00000155, 0.00000002, -0.00000209, -0.00000059, -0.00000433, 0.00000593, -0.00000134, 0.00000084, 0.00000097, -0.00000415, -0.00000701, 0.00000965, -0.00000735, 0.00000970, -0.00000070, -0.00000482, -0.00000663, 0.00000845, 0.00000170, -0.00000139, -0.00000309, -0.00000182, -0.00000115, 0.00000281, 0.00000570, -0.00000858, -0.00000507, 0.00000257, 0.00000620, -0.00000442, 0.00001097, -0.00001069, -0.00000002, 0.00000085, 0.00000204, },
		{ 0.00000631, -0.00000683, -0.00000013, -0.00000159, 0.00000266, 0.00000123, 0.00000379, -0.00000335, -0.00000450, 0.00000353, 0.00000136, -0.00000002, -0.00001237, 0.00001154, 0.00001142, -0.00001547, -0.00000734, 0.00000488, 0.00000480, -0.00000942, -0.00000008, -0.00000025, 0.00001043, -0.00001567, 0.00000365, -0.00000987, 0.00000505, -0.00001340, 0.00001650, -0.00003553, 0.00003435, -0.00004035, 0.00002773, -0.00001112, 0.00001618, -0.00001817, 0.00000888, -0.00001275, 0.00002592, -0.00004902, 0.00001623, -0.00003061, 0.00002546, -0.00007307, 0.00000430, -0.00021785, -0.00023506, 0.00069209, 0.00055780, -0.00033210, -0.00015099, -0.00000870, -0.00005435, -0.00000149, -0.00004693, 0.00002339, -0.00002666, 0.00000636, -0.00002008, 0.00000778, -0.00001416, 0.00000299, -0.00001831, 0.00001681, -0.00001639, 0.00001724, -0.00003370, 0.00002885, -0.00002641, 0.00001386, -0.00000544, 0.00000578, -0.00000458, 0.00000246, -0.00000384, 0.00000482, 0.00000130, -0.00000211, -0.00001057, 0.00000702, 0.00000519, -0.00000683, -0.00000773, 0.00000547, 0.00001291, -0.00001142, 0.00000214, 0.00000138, 0.00000152, -0.00000425, -0.00000423, 0.00000386, -0.00000118, 0.00000278, -0.00000438, 0.00000218, -0.00000704, 0.00000708, -0.00000185, 0.00000166, -0.00000036, -0.00000140, -0.00000985, 0.00001054, -0.00000705, 0.00000788, 0.00000162, -0.00000482, -0.00000694, 0.00000631, 0.00000417, -0.00000250, -0.00000175, -0.00000334, 0.00000011, 0.00000035, 0.00000798, -0.00000834, -0.00000502, -0.00000018, 0.00000884, -0.00000612, 0.00001163, -0.00000904, -0.00000167, 0.00000091, 0.00000258, -0.00000436, },
		{ -0.00000542, -0.00000144, -0.00000117, 0.00000213, 0.00000165, 0.00000352, -0.00000264, -0.00000443, 0.00000216, 0.00000272, -0.00000051, -0.00001018, 0.00000538, 0.00001710, -0.00001697, -0.00000562, -0.00000015, 0.00001030, -0.00001329, 0.00000367, -0.00000572, 0.00001603, -0.00002054, 0.00000877, -0.00001555, 0.00001159, -0.00002161, 0.00002471, -0.00004065, 0.00003457, -0.00003919, 0.00002500, -0.00000838, 0.00001609, -0.00001830, 0.00001071, -0.00001550, 0.00002996, -0.00005269, 0.00002086, -0.00003519, 0.00003247, -0.00007906, 0.00001522, -0.00021870, -0.00020962, 0.00058527, 0.00057469, -0.00021745, -0.00022701, 0.00002500, -0.00008119, 0.00001716, -0.00006276, 0.00003635, -0.00003817, 0.00001575, -0.00002879, 0.00001610, -0.00002235, 0.00001088, -0.00002640, 0.00002502, -0.00002492, 0.00002449, -0.00003779, 0.00002867, -0.00002322, 0.00000896, -0.00000182, 0.00000514, -0.00000469, 0.00000256, -0.00000410, 0.00000476, 0.00000096, -0.00000045, -0.00001134, 0.00000598, 0.00000541, -0.00000454, -0.00000921, 0.00000392, 0.00001315, -0.00000813, -0.00000189, 0.00000488, -0.00000166, -0.00000035, -0.00000817, 0.00000645, -0.00000355, 0.00000514, -0.00000624, 0.00000443, -0.00000896, 0.00000741, -0.00000183, 0.00000220, -0.00000146, 0.00000084, -0.00001161, 0.00001012, -0.00000559, 0.00000528, 0.00000410, -0.00000488, -0.00000673, 0.00000371, 0.00000658, -0.00000337, -0.00000057, -0.00000464, 0.00000115, -0.00000169, 0.00000940, -0.00000738, -0.00000514, -0.00000260, 0.00001055, -0.00000660, 0.00001097, -0.00000644, -0.00000362, 0.00000107, 0.00000290, -0.00000489, 0.00000630, },
		{ -0.00000284, -0.00000075, 0.00000160, 0.00000198, 0.00000340, -0.00000227, -0.00000381, 0.00000032, 0.00000436, -0.00000147, -0.00000725, -0.00000087, 0.00002174, -0.00001745, -0.00000417, -0.00000482, 0.00001475, -0.00001588, 0.00000631, -0.00000995, 0.00001996, -0.00002354, 0.00001194, -0.00001908, 0.00001566, -0.00002655, 0.00002850, -0.00004057, 0.00002996, -0.00003422, 0.00001920, -0.00000340, 0.00001414, -0.00001653, 0.00001060, -0.00001611, 0.00003159, -0.00005356, 0.00002229, -0.00003620, 0.00003541, -0.00008019, 0.00001997, -0.00021259, -0.00019557, 0.00050893, 0.00055165, -0.00008715, -0.00029563, 0.00004902, -0.00010041, 0.00002942, -0.00007335, 0.00004471, -0.00004544, 0.00002133, -0.00003419, 0.00002139, -0.00002760, 0.00001592, -0.00003165, 0.00003016, -0.00002975, 0.00002733, -0.00003720, 0.00002431, -0.00001700, 0.00000244, 0.00000217, 0.00000447, -0.00000455, 0.00000235, -0.00000408, 0.00000454, 0.00000062, 0.00000106, -0.00001174, 0.00000491, 0.00000516, -0.00000200, -0.00001050, 0.00000266, 0.00001229, -0.00000389, -0.00000624, 0.00000833, -0.00000469, 0.00000319, -0.00001136, 0.00000814, -0.00000511, 0.00000667, -0.00000729, 0.00000580, -0.00000980, 0.00000674, -0.00000119, 0.00000233, -0.00000214, 0.00000232, -0.00001214, 0.00000846, -0.00000320, 0.00000224, 0.00000644, -0.00000482, -0.00000620, 0.00000097, 0.00000856, -0.00000375, 0.00000021, -0.00000549, 0.00000171, -0.00000306, 0.00000980, -0.00000565, -0.00000551, -0.00000449, 0.00001115, -0.00000581, 0.00000913, -0.00000315, -0.00000564, 0.00000123, 0.00000300, -0.00000497, 0.00000565, -0.00000355, },
		{ -0.00000048, 0.00000116, 0.00000217, 0.00000346, -0.00000221, -0.00000277, -0.00000172, 0.00000596, -0.00000249, -0.00000415, -0.00000644, 0.00002470, -0.00001650, -0.00000342, -0.00000852, 0.00001751, -0.00001662, 0.00000730, -0.00001233, 0.00002162, -0.00002409, 0.00001257, -0.00001988, 0.00001662, -0.00002762, 0.00002751, -0.00003544, 0.00002116, -0.00002630, 0.00001116, 0.00000314, 0.00001081, -0.00001316, 0.00000871, -0.00001462, 0.00003073, -0.00005147, 0.00002030, -0.00003339, 0.00003393, -0.00007600, 0.00001802, -0.00019886, -0.00019333, 0.00045989, 0.00049891, 0.00004541, -0.00034695, 0.00005784, -0.00010864, 0.00003288, -0.00007685, 0.00004692, -0.00004719, 0.00002202, -0.00003532, 0.00002275, -0.00002907, 0.00001731, -0.00003327, 0.00003146, -0.00003028, 0.00002549, -0.00003214, 0.00001643, -0.00000864, -0.00000478, 0.00000590, 0.00000406, -0.00000429, 0.00000188, -0.00000381, 0.00000415, 0.00000034, 0.00000229, -0.00001171, 0.00000380, 0.00000455, 0.00000056, -0.00001144, 0.00000161, 0.00001063, 0.00000075, -0.00001027, 0.00001112, -0.00000696, 0.00000580, -0.00001332, 0.00000857, -0.00000556, 0.00000713, -0.00000732, 0.00000612, -0.00000947, 0.00000511, 0.00000003, 0.00000204, -0.00000232, 0.00000296, -0.00001150, 0.00000578, -0.00000026, -0.00000083, 0.00000834, -0.00000446, -0.00000556, -0.00000159, 0.00000985, -0.00000344, 0.00000047, -0.00000575, 0.00000166, -0.00000363, 0.00000916, -0.00000327, -0.00000610, -0.00000577, 0.00001063, -0.00000391, 0.00000642, 0.00000042, -0.00000743, 0.00000124, 0.00000295, -0.00000468, 0.00000453, -0.00000144, -0.00000415, },
		{ 0.00000089, 0.00000217, 0.00000371, -0.00000241, -0.00000151, -0.00000368, 0.00000719, -0.00000324, -0.00000140, -0.00001075, 0.00002559, -0.00001396, -0.00000358, -0.00001089, 0.00001823, -0.00001531, 0.00000645, -0.00001263, 0.00002081, -0.00002210, 0.00001063, -0.00001791, 0.00001447, -0.00002496, 0.00002225, -0.00002625, 0.00000958, -0.00001687, 0.00000215, 0.00001016, 0.00000693, -0.00000885, 0.00000560, -0.00001152, 0.00002783, -0.00004687, 0.00001541, -0.00002729, 0.00002858, -0.00006712, 0.00001020, -0.00017860, -0.00020105, 0.00043252, 0.00042802, 0.00016765, -0.00037339, 0.00004840, -0.00010454, 0.00002687, -0.00007284, 0.00004269, -0.00004322, 0.00001774, -0.00003211, 0.00002010, -0.00002666, 0.00001500, -0.00003124, 0.00002895, -0.00002667, 0.00001949, -0.00002354, 0.00000626, 0.00000056, -0.00001163, 0.00000864, 0.00000426, -0.00000406, 0.00000128, -0.00000339, 0.00000366, 0.00000016, 0.00000320, -0.00001125, 0.00000266, 0.00000372, 0.00000293, -0.00001187, 0.00000061, 0.00000849, 0.00000524, -0.00001338, 0.00001277, -0.00000807, 0.00000712, -0.00001375, 0.00000756, -0.00000482, 0.00000647, -0.00000631, 0.00000541, -0.00000807, 0.00000269, 0.00000164, 0.00000144, -0.00000203, 0.00000281, -0.00000988, 0.00000248, 0.00000279, -0.00000351, 0.00000951, -0.00000364, -0.00000501, -0.00000370, 0.00001024, -0.00000238, 0.00000016, -0.00000539, 0.00000100, -0.00000343, 0.00000764, -0.00000046, -0.00000677, -0.00000650, 0.00000915, -0.00000123, 0.00000329, 0.00000384, -0.00000868, 0.00000095, 0.00000284, -0.00000413, 0.00000312, 0.00000066, -0.00000514, -0.00000047, },
		{ 0.00000198, 0.00000410, -0.00000274, -0.00000022, -0.00000529, 0.00000780, -0.00000342, 0.00000062, -0.00001342, 0.00002429, -0.00000996, -0.00000461, -0.00001184, 0.00001691, -0.00001208, 0.00000393, -0.00001102, 0.00001780, -0.00001793, 0.00000654, -0.00001365, 0.00000978, -0.00001934, 0.00001387, -0.00001463, -0.00000294, -0.00000765, -0.00000637, 0.00001645, 0.00000350, -0.00000445, 0.00000206, -0.00000757, 0.00002369, -0.00004064, 0.00000858, -0.00001901, 0.00002061, -0.00005499, -0.00000168, -0.00015409, -0.00021537, 0.00041987, 0.00035041, 0.00026921, -0.00037057, 0.00002052, -0.00008902, 0.00001242, -0.00006231, 0.00003291, -0.00003439, 0.00000931, -0.00002534, 0.00001409, -0.00002102, 0.00000960, -0.00002618, 0.00002329, -0.00001979, 0.00001050, -0.00001288, -0.00000461, 0.00000915, -0.00001701, 0.00000979, 0.00000533, -0.00000400, 0.00000067, -0.00000293, 0.00000312, 0.00000006, 0.00000377, -0.00001039, 0.00000151, 0.00000276, 0.00000492, -0.00001168, -0.00000043, 0.00000622, 0.00000905, -0.00001515, 0.00001296, -0.00000780, 0.00000699, -0.00001263, 0.00000522, -0.00000305, 0.00000487, -0.00000445, 0.00000388, -0.00000588, -0.00000018, 0.00000336, 0.00000072, -0.00000141, 0.00000208, -0.00000764, -0.00000100, 0.00000546, -0.00000544, 0.00000974, -0.00000230, -0.00000467, -0.00000519, 0.00000968, -0.00000063, -0.00000062, -0.00000453, -0.00000017, -0.00000263, 0.00000550, 0.00000248, -0.00000731, -0.00000683, 0.00000699, 0.00000181, 0.00000022, 0.00000668, -0.00000915, 0.00000027, 0.00000276, -0.00000349, 0.00000164, 0.00000251, -0.00000565, -0.00000080, 0.00000082, },
		{ 0.00000456, -0.00000311, 0.00000090, -0.00000633, 0.00000761, -0.00000290, 0.00000172, -0.00001434, 0.00002101, -0.00000487, -0.00000624, -0.00001153, 0.00001384, -0.00000740, 0.00000022, -0.00000800, 0.00001317, -0.00001230, 0.00000110, -0.00000798, 0.00000354, -0.00001200, 0.00000398, -0.00000249, -0.00001443, -0.00000031, -0.00001308, 0.00002092, 0.00000139, -0.00000076, -0.00000109, -0.00000364, 0.00001924, -0.00003383, 0.00000103, -0.00000990, 0.00001155, -0.00004145, -0.00001534, -0.00012818, -0.00023235, 0.00041488, 0.00027604, 0.00034290, -0.00033779, -0.00002305, -0.00006504, -0.00000793, -0.00004742, 0.00001944, -0.00002239, -0.00000171, -0.00001637, 0.00000598, -0.00001332, 0.00000226, -0.00001921, 0.00001565, -0.00001098, 0.00000010, -0.00000191, -0.00001450, 0.00001578, -0.00002002, 0.00000893, 0.00000737, -0.00000416, 0.00000013, -0.00000252, 0.00000261, 0.00000002, 0.00000401, -0.00000920, 0.00000040, 0.00000177, 0.00000638, -0.00001083, -0.00000159, 0.00000410, 0.00001184, -0.00001534, 0.00001164, -0.00000618, 0.00000552, -0.00001017, 0.00000187, -0.00000059, 0.00000267, -0.00000209, 0.00000190, -0.00000330, -0.00000309, 0.00000484, 0.00000011, -0.00000067, 0.00000102, -0.00000515, -0.00000420, 0.00000737, -0.00000634, 0.00000898, -0.00000046, -0.00000456, -0.00000602, 0.00000823, 0.00000161, -0.00000167, -0.00000335, -0.00000162, -0.00000149, 0.00000310, 0.00000517, -0.00000748, -0.00000696, 0.00000448, 0.00000476, -0.00000233, 0.00000857, -0.00000867, -0.00000085, 0.00000278, -0.00000286, 0.00000029, 0.00000392, -0.00000557, -0.00000149, 0.00000095, 0.00000164, },
		{ -0.00000337, 0.00000169, -0.00000667, 0.00000660, -0.00000169, 0.00000191, -0.00001366, 0.00001620, 0.00000071, -0.00000803, -0.00001034, 0.00000956, -0.00000194, -0.00000397, -0.00000429, 0.00000778, -0.00000616, -0.00000468, -0.00000197, -0.00000305, -0.00000434, -0.00000574, 0.00000829, -0.00002314, 0.00000382, -0.00001701, 0.00002283, 0.00000120, 0.00000163, -0.00000320, -0.00000044, 0.00001533, -0.00002744, -0.00000607, -0.00000132, 0.00000295, -0.00002831, -0.00002849, -0.00010367, -0.00024822, 0.00041135, 0.00021252, 0.00038527, -0.00027793, -0.00007710, -0.00003704, -0.00003065, -0.00003105, 0.00000471, -0.00000939, -0.00001338, -0.00000694, -0.00000269, -0.00000501, -0.00000560, -0.00001171, 0.00000745, -0.00000177, -0.00001000, 0.00000768, -0.00002195, 0.00001942, -0.00002016, 0.00000598, 0.00001027, -0.00000448, -0.00000031, -0.00000219, 0.00000214, 0.00000005, 0.00000396, -0.00000777, -0.00000061, 0.00000081, 0.00000723, -0.00000932, -0.00000291, 0.00000233, 0.00001338, -0.00001392, 0.00000898, -0.00000347, 0.00000304, -0.00000679, -0.00000201, 0.00000209, 0.00000030, 0.00000036, -0.00000014, -0.00000075, -0.00000563, 0.00000580, -0.00000019, 0.00000002, -0.00000011, -0.00000279, -0.00000672, 0.00000824, -0.00000613, 0.00000730, 0.00000175, -0.00000464, -0.00000622, 0.00000609, 0.00000408, -0.00000274, -0.00000209, -0.00000310, -0.00000029, 0.00000078, 0.00000729, -0.00000710, -0.00000707, 0.00000195, 0.00000721, -0.00000401, 0.00000933, -0.00000723, -0.00000235, 0.00000289, -0.00000234, -0.00000081, 0.00000478, -0.00000490, -0.00000248, 0.00000120, 0.00000121, 0.00000501, },
		{ 0.00000206, -0.00000629, 0.00000485, 0.00000008, 0.00000134, -0.00001173, 0.00001050, 0.00000612, -0.00000949, -0.00000871, 0.00000471, 0.00000352, -0.00000788, -0.00000068, 0.00000250, -0.00000049, -0.00000979, 0.00000331, -0.00000881, 0.00000228, -0.00001375, 0.00001615, -0.00002783, 0.00000401, -0.00001777, 0.00002192, 0.00000312, 0.00000248, -0.00000394, 0.00000156, 0.00001255, -0.00002221, -0.00001182, 0.00000568, -0.00000396, -0.00001709, -0.00003929, -0.00008281, -0.00026000, 0.00040465, 0.00016453, 0.00039662, -0.00019695, -0.00013473, -0.00001025, -0.00005185, -0.00001627, -0.00000874, 0.00000240, -0.00002375, 0.00000127, -0.00001038, 0.00000245, -0.00001258, -0.00000505, 0.00000008, 0.00000635, -0.00001829, 0.00001450, -0.00002594, 0.00001956, -0.00001738, 0.00000123, 0.00001368, -0.00000476, -0.00000071, -0.00000195, 0.00000172, 0.00000013, 0.00000366, -0.00000620, -0.00000149, -0.00000005, 0.00000746, -0.00000725, -0.00000433, 0.00000099, 0.00001366, -0.00001112, 0.00000534, -0.00000011, 0.00000002, -0.00000303, -0.00000584, 0.00000446, -0.00000178, 0.00000247, -0.00000182, 0.00000135, -0.00000745, 0.00000599, -0.00000003, 0.00000049, -0.00000107, -0.00000087, -0.00000831, 0.00000797, -0.00000488, 0.00000490, 0.00000410, -0.00000477, -0.00000595, 0.00000354, 0.00000643, -0.00000354, -0.00000101, -0.00000436, 0.00000067, -0.00000112, 0.00000859, -0.00000604, -0.00000730, -0.00000034, 0.00000888, -0.00000461, 0.00000891, -0.00000496, -0.00000407, 0.00000303, -0.00000194, -0.00000158, 0.00000505, -0.00000370, -0.00000364, 0.00000146, 0.00000080, 0.00000534, -0.00000342, },
		{ -0.00000526, 0.00000255, 0.00000213, 0.00000034, -0.00000904, 0.00000463, 0.00001066, -0.00001016, -0.00000712, -0.00000005, 0.00000824, -0.00001082, 0.00000213, -0.00000186, 0.00000386, -0.00001338, 0.00000700, -0.00001283, 0.00000683, -0.00001895, 0.00002015, -0.00002798, 0.00000024, -0.00001557, 0.00001844, 0.00000688, 0.00000196, -0.00000331, 0.00000224, 0.00001117, -0.00001855, -0.00001569, 0.00001048, -0.00000842, -0.00000870, -0.00004656, -0.00006699, -0.00026592, 0.00039204, 0.00013380, 0.00038050, -0.00010300, -0.00018833, 0.00001026, -0.00006800, -0.00000574, -0.00001875, 0.00001113, -0.00003120, 0.00000684, -0.00001586, 0.00000788, -0.00001756, -0.00000033, -0.00000537, 0.00001224, -0.00002365, 0.00001769, -0.00002606, 0.00001625, -0.00001211, -0.00000472, 0.00001709, -0.00000474, -0.00000116, -0.00000174, 0.00000132, 0.00000029, 0.00000315, -0.00000459, -0.00000217, -0.00000080, 0.00000708, -0.00000475, -0.00000577, 0.00000009, 0.00001280, -0.00000731, 0.00000125, 0.00000336, -0.00000298, 0.00000052, -0.00000906, 0.00000606, -0.00000317, 0.00000390, -0.00000282, 0.00000273, -0.00000832, 0.00000532, 0.00000063, 0.00000067, -0.00000170, 0.00000043, -0.00000885, 0.00000661, -0.00000280, 0.00000212, 0.00000630, -0.00000477, -0.00000541, 0.00000091, 0.00000833, -0.00000382, -0.00000031, -0.00000517, 0.00000118, -0.00000239, 0.00000892, -0.00000429, -0.00000770, -0.00000222, 0.00000960, -0.00000413, 0.00000747, -0.00000213, -0.00000580, 0.00000310, -0.00000166, -0.00000202, 0.00000479, -0.00000213, -0.00000481, 0.00000161, 0.00000049, 0.00000549, -0.00000320, 0.00000198, },
		{ -0.00000001, 0.00000414, -0.00000076, -0.00000611, -0.00000075, 0.00001378, -0.00000970, -0.00000593, -0.00000419, 0.00001162, -0.00001227, 0.00000364, -0.00000473, 0.00000632, -0.00001493, 0.00000861, -0.00001457, 0.00000875, -0.00002084, 0.00002002, -0.00002381, -0.00000681, -0.00001121, 0.00001315, 0.00001181, 0.00000055, -0.00000165, 0.00000183, 0.00001108, -0.00001645, -0.00001759, 0.00001294, -0.00001025, -0.00000345, -0.00004988, -0.00005666, -0.00026545, 0.00037267, 0.00011942, 0.00034293, -0.00000525, -0.00023069, 0.00002033, -0.00007654, -0.00000129, -0.00002390, 0.00001558, -0.00003473, 0.00000897, -0.00001837, 0.00001055, -0.00001983, 0.00000180, -0.00000823, 0.00001525, -0.00002556, 0.00001704, -0.00002252, 0.00001011, -0.00000520, -0.00001100, 0.00001987, -0.00000411, -0.00000180, -0.00000149, 0.00000089, 0.00000053, 0.00000249, -0.00000303, -0.00000263, -0.00000142, 0.00000619, -0.00000203, -0.00000706, -0.00000047, 0.00001106, -0.00000299, -0.00000272, 0.00000638, -0.00000543, 0.00000333, -0.00001120, 0.00000654, -0.00000363, 0.00000444, -0.00000297, 0.00000322, -0.00000820, 0.00000384, 0.00000170, 0.00000057, -0.00000195, 0.00000103, -0.00000841, 0.00000440, -0.00000026, -0.00000066, 0.00000805, -0.00000445, -0.00000480, -0.00000152, 0.00000953, -0.00000343, -0.00000013, -0.00000543, 0.00000112, -0.00000292, 0.00000832, -0.00000198, -0.00000820, -0.00000364, 0.00000938, -0.00000270, 0.00000528, 0.00000092, -0.00000726, 0.00000298, -0.00000143, -0.00000220, 0.00000413, -0.00000038, -0.00000578, 0.00000152, 0.00000037, 0.00000538, -0.00000270, 0.00000148, -0.00000372, },
		{ 0.00000578, -0.00000159, -0.00000342, -0.00000506, 0.00001511, -0.00000796, -0.00000532, -0.00000734, 0.00001332, -0.00001199, 0.00000361, -0.00000586, 0.00000668, -0.00001432, 0.00000803, -0.00001395, 0.00000801, -0.00001950, 0.00001621, -0.00001625, -0.00001588, -0.00000591, 0.00000708, 0.00001702, -0.00000102, 0.00000049, 0.00000076, 0.00001193, -0.00001560, -0.00001783, 0.00001336, -0.00000975, -0.00000101, -0.00004958, -0.00005134, -0.00025923, 0.00034733, 0.00011848, 0.00029134, 0.00008727, -0.00025600, 0.00001747, -0.00007630, -0.00000357, -0.00002376, 0.00001542, -0.00003410, 0.00000749, -0.00001780, 0.00001032, -0.00001925, 0.00000121, -0.00000838, 0.00001527, -0.00002407, 0.00001294, -0.00001613, 0.00000222, 0.00000221, -0.00001665, 0.00002142, -0.00000262, -0.00000271, -0.00000116, 0.00000039, 0.00000085, 0.00000174, -0.00000162, -0.00000282, -0.00000192, 0.00000491, 0.00000070, -0.00000806, -0.00000084, 0.00000878, 0.00000131, -0.00000601, 0.00000849, -0.00000694, 0.00000505, -0.00001200, 0.00000576, -0.00000310, 0.00000405, -0.00000227, 0.00000286, -0.00000716, 0.00000172, 0.00000302, 0.00000029, -0.00000185, 0.00000100, -0.00000717, 0.00000168, 0.00000235, -0.00000305, 0.00000911, -0.00000369, -0.00000431, -0.00000348, 0.00000984, -0.00000232, -0.00000045, -0.00000513, 0.00000050, -0.00000277, 0.00000692, 0.00000066, -0.00000867, -0.00000464, 0.00000835, -0.00000060, 0.00000274, 0.00000379, -0.00000820, 0.00000254, -0.00000119, -0.00000221, 0.00000324, 0.00000132, -0.00000640, 0.00000110, 0.00000049, 0.00000502, -0.00000194, 0.00000067, -0.00000191, -0.00000252, },
		{ -0.00000190, -0.00000132, -0.00000796, 0.00001457, -0.00000505, -0.00000532, -0.00000934, 0.00001326, -0.00001003, 0.00000215, -0.00000533, 0.00000510, -0.00001179, 0.00000558, -0.00001134, 0.00000504, -0.00001559, 0.00000972, -0.00000671, -0.00002539, -0.00000105, 0.00000136, 0.00002151, -0.00000199, 0.00000247, -0.00000040, 0.00001318, -0.00001549, -0.00001694, 0.00001237, -0.00000762, -0.00000066, -0.00004653, -0.00004988, -0.00024874, 0.00031793, 0.00012688, 0.00023348, 0.00016678, -0.00026058, 0.00000120, -0.00006771, -0.00001199, -0.00001889, 0.00001116, -0.00002983, 0.00000295, -0.00001460, 0.00000758, -0.00001620, -0.00000171, -0.00000620, 0.00001273, -0.00001982, 0.00000635, -0.00000811, -0.00000611, 0.00000894, -0.00002079, 0.00002127, -0.00000014, -0.00000391, -0.00000072, -0.00000018, 0.00000123, 0.00000097, -0.00000043, -0.00000273, -0.00000231, 0.00000337, 0.00000321, -0.00000861, -0.00000117, 0.00000632, 0.00000510, -0.00000820, 0.00000938, -0.00000725, 0.00000550, -0.00001139, 0.00000381, -0.00000169, 0.00000289, -0.00000088, 0.00000183, -0.00000547, -0.00000074, 0.00000433, -0.00000001, -0.00000150, 0.00000049, -0.00000543, -0.00000117, 0.00000461, -0.00000474, 0.00000929, -0.00000242, -0.00000404, -0.00000483, 0.00000922, -0.00000056, -0.00000120, -0.00000438, -0.00000057, -0.00000208, 0.00000499, 0.00000334, -0.00000893, -0.00000534, 0.00000673, 0.00000180, 0.00000026, 0.00000612, -0.00000841, 0.00000173, -0.00000088, -0.00000213, 0.00000226, 0.00000279, -0.00000652, 0.00000031, 0.00000082, 0.00000445, -0.00000101, -0.00000030, -0.00000006, -0.00000467, 0.00000678, },
		{ -0.00000002, -0.00000933, 0.00001229, -0.00000128, -0.00000573, -0.00001026, 0.00001160, -0.00000673, -0.00000040, -0.00000351, 0.00000204, -0.00000795, 0.00000193, -0.00000745, 0.00000067, -0.00001015, 0.00000191, 0.00000316, -0.00003370, 0.00000208, -0.00000300, 0.00002446, -0.00000168, 0.00000373, -0.00000110, 0.00001429, -0.00001553, -0.00001559, 0.00001073, -0.00000471, -0.00000141, -0.00004189, -0.00005080, -0.00023589, 0.00028693, 0.00014011, 0.00017645, 0.00022757, -0.00024337, -0.00002681, -0.00005273, -0.00002485, -0.00001075, 0.00000406, -0.00002307, -0.00000361, -0.00000973, 0.00000315, -0.00001147, -0.00000620, -0.00000245, 0.00000846, -0.00001379, -0.00000149, 0.00000014, -0.00001352, 0.00001390, -0.00002272, 0.00001918, 0.00000328, -0.00000531, -0.00000020, -0.00000080, 0.00000163, 0.00000026, 0.00000047, -0.00000235, -0.00000263, 0.00000173, 0.00000531, -0.00000859, -0.00000160, 0.00000399, 0.00000799, -0.00000901, 0.00000892, -0.00000636, 0.00000473, -0.00000953, 0.00000094, 0.00000028, 0.00000123, 0.00000093, 0.00000042, -0.00000345, -0.00000317, 0.00000535, -0.00000017, -0.00000106, -0.00000029, -0.00000350, -0.00000375, 0.00000620, -0.00000551, 0.00000856, -0.00000068, -0.00000401, -0.00000553, 0.00000777, 0.00000164, -0.00000216, -0.00000335, -0.00000187, -0.00000111, 0.00000285, 0.00000574, -0.00000878, -0.00000590, 0.00000481, 0.00000413, -0.00000178, 0.00000762, -0.00000780, 0.00000053, -0.00000049, -0.00000204, 0.00000135, 0.00000387, -0.00000609, -0.00000080, 0.00000131, 0.00000376, -0.00000003, -0.00000126, 0.00000159, -0.00000621, 0.00000694, -0.00000153, },
		{ -0.00000928, 0.00000868, 0.00000287, -0.00000626, -0.00001033, 0.00000874, -0.00000262, -0.00000349, -0.00000098, -0.00000181, -0.00000356, -0.00000211, -0.00000318, -0.00000409, -0.00000436, -0.00000577, 0.00001175, -0.00003939, 0.00000253, -0.00000534, 0.00002531, 0.00000030, 0.00000391, -0.00000096, 0.00001484, -0.00001522, -0.00001438, 0.00000916, -0.00000187, -0.00000228, -0.00003684, -0.00005255, -0.00022265, 0.00025680, 0.00015406, 0.00012587, 0.00026650, -0.00020593, -0.00006302, -0.00003445, -0.00003967, -0.00000134, -0.00000420, -0.00001533, -0.00001081, -0.00000437, -0.00000188, -0.00000608, -0.00001124, 0.00000187, 0.00000349, -0.00000721, -0.00000920, 0.00000722, -0.00001886, 0.00001631, -0.00002212, 0.00001522, 0.00000740, -0.00000673, 0.00000029, -0.00000140, 0.00000199, -0.00000032, 0.00000103, -0.00000171, -0.00000286, 0.00000011, 0.00000685, -0.00000794, -0.00000222, 0.00000201, 0.00000976, -0.00000839, 0.00000722, -0.00000445, 0.00000299, -0.00000678, -0.00000241, 0.00000241, -0.00000056, 0.00000280, -0.00000101, -0.00000147, -0.00000525, 0.00000585, -0.00000003, -0.00000066, -0.00000111, -0.00000170, -0.00000575, 0.00000688, -0.00000529, 0.00000700, 0.00000139, -0.00000416, -0.00000566, 0.00000569, 0.00000402, -0.00000309, -0.00000229, -0.00000318, -0.00000012, 0.00000083, 0.00000755, -0.00000807, -0.00000644, 0.00000282, 0.00000606, -0.00000309, 0.00000814, -0.00000637, -0.00000097, -0.00000005, -0.00000198, 0.00000059, 0.00000449, -0.00000513, -0.00000215, 0.00000185, 0.00000304, 0.00000087, -0.00000205, 0.00000284, -0.00000698, 0.00000622, -0.00000050, 0.00000049, },
		{ 0.00000426, 0.00000684, -0.00000651, -0.00000989, 0.00000518, 0.00000167, -0.00000649, 0.00000163, -0.00000571, 0.00000053, -0.00000567, 0.00000060, -0.00000827, 0.00000063, -0.00001204, 0.00001780, -0.00004155, -0.00000015, -0.00000543, 0.00002389, 0.00000402, 0.00000295, 0.00000016, 0.00001461, -0.00001423, -0.00001375, 0.00000823, 0.00000023, -0.00000248, -0.00003237, -0.00005382, -0.00021066, 0.00022959, 0.00016549, 0.00008548, 0.00028311, -0.00015219, -0.00010255, -0.00001661, -0.00005368, 0.00000716, -0.00001180, -0.00000820, -0.00001725, 0.00000026, -0.00000642, -0.00000108, -0.00001581, 0.00000576, -0.00000113, -0.00000122, -0.00001555, 0.00001204, -0.00002135, 0.00001582, -0.00001905, 0.00000976, 0.00001179, -0.00000790, 0.00000064, -0.00000192, 0.00000222, -0.00000070, 0.00000125, -0.00000084, -0.00000303, -0.00000138, 0.00000774, -0.00000668, -0.00000304, 0.00000053, 0.00001033, -0.00000650, 0.00000458, -0.00000187, 0.00000068, -0.00000361, -0.00000574, 0.00000425, -0.00000209, 0.00000438, -0.00000213, 0.00000013, -0.00000667, 0.00000566, 0.00000049, -0.00000041, -0.00000181, -0.00000026, -0.00000696, 0.00000659, -0.00000416, 0.00000481, 0.00000357, -0.00000435, -0.00000535, 0.00000326, 0.00000623, -0.00000372, -0.00000142, -0.00000426, 0.00000063, -0.00000079, 0.00000858, -0.00000673, -0.00000704, 0.00000099, 0.00000736, -0.00000351, 0.00000768, -0.00000426, -0.00000262, 0.00000037, -0.00000195, 0.00000003, 0.00000463, -0.00000373, -0.00000359, 0.00000231, 0.00000242, 0.00000159, -0.00000255, 0.00000359, -0.00000695, 0.00000471, 0.00000104, 0.00000036, -0.00000811, },
		{ 0.00001007, -0.00000616, -0.00000927, 0.00000146, 0.00000550, -0.00000880, 0.00000371, -0.00000896, 0.00000361, -0.00000805, 0.00000314, -0.00001108, 0.00000394, -0.00001596, 0.00002054, -0.00003985, -0.00000581, -0.00000357, 0.00002047, 0.00000919, 0.00000108, 0.00000212, 0.00001362, -0.00001249, -0.00001390, 0.00000823, 0.00000121, -0.00000152, -0.00002912, -0.00005375, -0.00020099, 0.00020659, 0.00017239, 0.00005697, 0.00027939, -0.00008778, -0.00013989, -0.00000293, -0.00006431, 0.00001281, -0.00001719, -0.00000305, -0.00002175, 0.00000316, -0.00000959, 0.00000266, -0.00001907, 0.00000841, -0.00000454, 0.00000324, -0.00001962, 0.00001392, -0.00002071, 0.00001256, -0.00001396, 0.00000342, 0.00001590, -0.00000850, 0.00000068, -0.00000226, 0.00000228, -0.00000085, 0.00000112, 0.00000019, -0.00000311, -0.00000268, 0.00000797, -0.00000490, -0.00000400, -0.00000046, 0.00000982, -0.00000363, 0.00000144, 0.00000091, -0.00000172, -0.00000053, -0.00000855, 0.00000540, -0.00000303, 0.00000538, -0.00000270, 0.00000111, -0.00000727, 0.00000472, 0.00000138, -0.00000033, -0.00000227, 0.00000066, -0.00000732, 0.00000541, -0.00000234, 0.00000231, 0.00000561, -0.00000440, -0.00000481, 0.00000077, 0.00000798, -0.00000383, -0.00000091, -0.00000492, 0.00000095, -0.00000182, 0.00000873, -0.00000480, -0.00000771, -0.00000057, 0.00000791, -0.00000305, 0.00000638, -0.00000173, -0.00000421, 0.00000067, -0.00000193, -0.00000033, 0.00000435, -0.00000202, -0.00000495, 0.00000256, 0.00000200, 0.00000202, -0.00000269, 0.00000380, -0.00000619, 0.00000262, 0.00000284, -0.00000014, -0.00000625, -0.00000032, },
		{ -0.00000496, -0.00000876, -0.00000195, 0.00000836, -0.00000997, 0.00000481, -0.00001106, 0.00000516, -0.00000879, 0.00000403, -0.00001205, 0.00000512, -0.00001714, 0.00001982, -0.00003460, -0.00001373, -0.00000045, 0.00001564, 0.00001522, -0.00000122, 0.00000460, 0.00001211, -0.00001013, -0.00001478, 0.00000920, 0.00000099, 0.00000073, -0.00002731, -0.00005196, -0.00019412, 0.00018826, 0.00017401, 0.00004013, 0.00025925, -0.00001927, -0.00016971, 0.00000352, -0.00006966, 0.00001428, -0.00001932, -0.00000077, -0.00002356, 0.00000372, -0.00001084, 0.00000458, -0.00002048, 0.00000929, -0.00000620, 0.00000560, -0.00002096, 0.00001269, -0.00001718, 0.00000712, -0.00000764, -0.00000299, 0.00001915, -0.00000826, 0.00000033, -0.00000236, 0.00000212, -0.00000075, 0.00000071, 0.00000129, -0.00000307, -0.00000376, 0.00000760, -0.00000276, -0.00000499, -0.00000102, 0.00000846, -0.00000024, -0.00000171, 0.00000341, -0.00000375, 0.00000198, -0.00001042, 0.00000558, -0.00000320, 0.00000565, -0.00000256, 0.00000138, -0.00000701, 0.00000310, 0.00000255, -0.00000040, -0.00000245, 0.00000103, -0.00000688, 0.00000354, -0.00000014, -0.00000017, 0.00000722, -0.00000416, -0.00000424, -0.00000148, 0.00000903, -0.00000330, -0.00000087, -0.00000508, 0.00000077, -0.00000221, 0.00000804, -0.00000241, -0.00000837, -0.00000183, 0.00000772, -0.00000184, 0.00000449, 0.00000092, -0.00000550, 0.00000074, -0.00000187, -0.00000053, 0.00000374, -0.00000021, -0.00000605, 0.00000248, 0.00000184, 0.00000214, -0.00000247, 0.00000352, -0.00000486, 0.00000023, 0.00000457, -0.00000067, -0.00000416, -0.00000450, 0.00001210, },
		{ -0.00000851, -0.00000472, 0.00000992, -0.00000978, 0.00000471, -0.00001176, 0.00000499, -0.00000778, 0.00000317, -0.00001114, 0.00000412, -0.00001569, 0.00001608, -0.00002670, -0.00002277, 0.00000296, 0.00001020, 0.00002135, -0.00000332, 0.00000710, 0.00001048, -0.00000745, -0.00001613, 0.00001091, -0.00000022, 0.00000406, -0.00002679, -0.00004857, -0.00018989, 0.00017429, 0.00017074, 0.00003327, 0.00022774, 0.00004675, -0.00018762, 0.00000083, -0.00006884, 0.00001106, -0.00001784, -0.00000163, -0.00002249, 0.00000182, -0.00001005, 0.00000455, -0.00001989, 0.00000827, -0.00000597, 0.00000572, -0.00001959, 0.00000870, -0.00001145, 0.00000040, -0.00000105, -0.00000864, 0.00002102, -0.00000698, -0.00000047, -0.00000220, 0.00000173, -0.00000044, 0.00000010, 0.00000235, -0.00000288, -0.00000462, 0.00000672, -0.00000047, -0.00000584, -0.00000132, 0.00000655, 0.00000322, -0.00000437, 0.00000521, -0.00000503, 0.00000359, -0.00001111, 0.00000466, -0.00000255, 0.00000518, -0.00000174, 0.00000095, -0.00000600, 0.00000099, 0.00000382, -0.00000054, -0.00000238, 0.00000091, -0.00000584, 0.00000127, 0.00000208, -0.00000229, 0.00000819, -0.00000350, -0.00000379, -0.00000328, 0.00000922, -0.00000208, -0.00000129, -0.00000475, 0.00000009, -0.00000200, 0.00000667, 0.00000021, -0.00000888, -0.00000283, 0.00000690, -0.00000011, 0.00000235, 0.00000336, -0.00000629, 0.00000049, -0.00000173, -0.00000062, 0.00000293, 0.00000155, -0.00000674, 0.00000203, 0.00000196, 0.00000197, -0.00000194, 0.00000289, -0.00000322, -0.00000212, 0.00000589, -0.00000093, -0.00000224, -0.00000781, 0.00001266, -0.00000283, },
		{ -0.00000669, 0.00001005, -0.00000822, 0.00000346, -0.00001110, 0.00000320, -0.00000521, 0.00000084, -0.00000863, 0.00000133, -0.00001217, 0.00001021, -0.00001742, -0.00003159, 0.00000559, 0.00000497, 0.00002680, -0.00000459, 0.00000912, 0.00000914, -0.00000484, -0.00001761, 0.00001302, -0.00000202, 0.00000806, -0.00002712, -0.00004404, -0.00018769, 0.00016379, 0.00016374, 0.00003373, 0.00019033, 0.00010445, -0.00019075, -0.00001143, -0.00006214, 0.00000357, -0.00001315, -0.00000529, -0.00001890, -0.00000215, -0.00000754, 0.00000281, -0.00001755, 0.00000560, -0.00000407, 0.00000388, -0.00001596, 0.00000272, -0.00000455, -0.00000648, 0.00000479, -0.00001277, 0.00002114, -0.00000463, -0.00000166, -0.00000183, 0.00000118, 0.00000002, -0.00000060, 0.00000328, -0.00000249, -0.00000529, 0.00000549, 0.00000176, -0.00000640, -0.00000151, 0.00000444, 0.00000629, -0.00000616, 0.00000602, -0.00000534, 0.00000413, -0.00001055, 0.00000271, -0.00000121, 0.00000408, -0.00000037, 0.00000001, -0.00000447, -0.00000133, 0.00000497, -0.00000060, -0.00000215, 0.00000044, -0.00000442, -0.00000105, 0.00000399, -0.00000377, 0.00000837, -0.00000238, -0.00000356, -0.00000448, 0.00000855, -0.00000029, -0.00000204, -0.00000405, -0.00000095, -0.00000136, 0.00000488, 0.00000276, -0.00000907, -0.00000366, 0.00000563, 0.00000183, 0.00000030, 0.00000529, -0.00000644, -0.00000010, -0.00000149, -0.00000065, 0.00000203, 0.00000307, -0.00000693, 0.00000119, 0.00000230, 0.00000157, -0.00000123, 0.00000207, -0.00000152, -0.00000410, 0.00000655, -0.00000070, -0.00000079, -0.00000994, 0.00001168, 0.00000013, -0.00000854, },
		{ 0.00000886, -0.00000553, 0.00000131, -0.00000935, 0.00000018, -0.00000159, -0.00000244, -0.00000513, -0.00000255, -0.00000746, 0.00000339, -0.00000818, -0.00003885, 0.00000654, 0.00000066, 0.00003092, -0.00000452, 0.00001026, 0.00000847, -0.00000264, -0.00001886, 0.00001511, -0.00000394, 0.00001220, -0.00002771, -0.00003904, -0.00018664, 0.00015553, 0.00015465, 0.00003851, 0.00015214, 0.00014944, -0.00017813, -0.00003212, -0.00005094, -0.00000698, -0.00000628, -0.00001088, -0.00001360, -0.00000745, -0.00000397, -0.00000007, -0.00001398, 0.00000182, -0.00000104, 0.00000064, -0.00001085, -0.00000422, 0.00000237, -0.00001245, 0.00000903, -0.00001486, 0.00001939, -0.00000134, -0.00000307, -0.00000136, 0.00000053, 0.00000053, -0.00000129, 0.00000397, -0.00000188, -0.00000581, 0.00000405, 0.00000371, -0.00000650, -0.00000175, 0.00000242, 0.00000862, -0.00000683, 0.00000572, -0.00000465, 0.00000360, -0.00000888, -0.00000000, 0.00000055, 0.00000262, 0.00000130, -0.00000120, -0.00000271, -0.00000356, 0.00000575, -0.00000046, -0.00000186, -0.00000023, -0.00000289, -0.00000311, 0.00000531, -0.00000445, 0.00000775, -0.00000084, -0.00000355, -0.00000509, 0.00000711, 0.00000188, -0.00000294, -0.00000314, -0.00000216, -0.00000051, 0.00000295, 0.00000496, -0.00000880, -0.00000443, 0.00000411, 0.00000371, -0.00000136, 0.00000650, -0.00000588, -0.00000103, -0.00000117, -0.00000067, 0.00000114, 0.00000425, -0.00000657, 0.00000002, 0.00000279, 0.00000106, -0.00000047, 0.00000127, -0.00000002, -0.00000546, 0.00000636, 0.00000014, 0.00000004, -0.00001083, 0.00000932, 0.00000363, -0.00000871, -0.00000785, },
		{ -0.00000215, -0.00000131, -0.00000697, -0.00000353, 0.00000245, -0.00000597, -0.00000135, -0.00000668, -0.00000260, -0.00000317, -0.00000029, -0.00004350, 0.00000522, -0.00000234, 0.00003328, -0.00000281, 0.00001031, 0.00000868, -0.00000108, -0.00001958, 0.00001685, -0.00000556, 0.00001597, -0.00002798, -0.00003425, -0.00018585, 0.00014824, 0.00014513, 0.00004474, 0.00011735, 0.00017914, -0.00015072, -0.00005868, -0.00003748, -0.00001881, 0.00000136, -0.00001719, -0.00000767, -0.00001309, -0.00000017, -0.00000336, -0.00000992, -0.00000237, 0.00000244, -0.00000323, -0.00000518, -0.00001103, 0.00000820, -0.00001659, 0.00001108, -0.00001469, 0.00001591, 0.00000258, -0.00000446, -0.00000094, -0.00000010, 0.00000099, -0.00000184, 0.00000433, -0.00000103, -0.00000621, 0.00000256, 0.00000522, -0.00000608, -0.00000216, 0.00000073, 0.00000999, -0.00000630, 0.00000437, -0.00000308, 0.00000222, -0.00000640, -0.00000311, 0.00000238, 0.00000112, 0.00000296, -0.00000238, -0.00000104, -0.00000539, 0.00000599, -0.00000002, -0.00000161, -0.00000091, -0.00000151, -0.00000468, 0.00000587, -0.00000428, 0.00000642, 0.00000099, -0.00000370, -0.00000516, 0.00000510, 0.00000415, -0.00000374, -0.00000224, -0.00000332, 0.00000029, 0.00000120, 0.00000657, -0.00000794, -0.00000520, 0.00000254, 0.00000525, -0.00000242, 0.00000690, -0.00000466, -0.00000219, -0.00000080, -0.00000067, 0.00000034, 0.00000503, -0.00000572, -0.00000137, 0.00000329, 0.00000056, 0.00000020, 0.00000064, 0.00000107, -0.00000605, 0.00000531, 0.00000154, 0.00000028, -0.00001060, 0.00000593, 0.00000726, -0.00000881, -0.00000834, 0.00000662, },
		{ -0.00000387, -0.00000449, -0.00000729, 0.00000620, -0.00000905, 0.00000195, -0.00001026, 0.00000150, -0.00000839, 0.00000524, -0.00004491, 0.00000150, -0.00000397, 0.00003377, 0.00000056, 0.00000929, 0.00000977, -0.00000022, -0.00001966, 0.00001802, -0.00000658, 0.00001902, -0.00002749, -0.00003025, -0.00018453, 0.00014085, 0.00013659, 0.00005012, 0.00008884, 0.00019296, -0.00011127, -0.00008757, -0.00002449, -0.00002993, 0.00000819, -0.00002294, -0.00000225, -0.00001809, 0.00000301, -0.00000632, -0.00000608, -0.00000624, 0.00000566, -0.00000697, 0.00000013, -0.00001670, 0.00001206, -0.00001833, 0.00001072, -0.00001239, 0.00001112, 0.00000666, -0.00000551, -0.00000074, -0.00000058, 0.00000129, -0.00000217, 0.00000434, 0.00000004, -0.00000651, 0.00000113, 0.00000616, -0.00000509, -0.00000279, -0.00000051, 0.00001033, -0.00000470, 0.00000222, -0.00000092, 0.00000031, -0.00000351, -0.00000616, 0.00000388, -0.00000013, 0.00000434, -0.00000326, 0.00000029, -0.00000661, 0.00000557, 0.00000074, -0.00000145, -0.00000149, -0.00000044, -0.00000561, 0.00000565, -0.00000337, 0.00000456, 0.00000290, -0.00000389, -0.00000485, 0.00000280, 0.00000621, -0.00000420, -0.00000156, -0.00000424, 0.00000085, -0.00000017, 0.00000742, -0.00000649, -0.00000601, 0.00000105, 0.00000630, -0.00000278, 0.00000651, -0.00000292, -0.00000343, -0.00000048, -0.00000064, -0.00000035, 0.00000540, -0.00000445, -0.00000281, 0.00000366, 0.00000020, 0.00000066, 0.00000032, 0.00000166, -0.00000582, 0.00000350, 0.00000333, 0.00000010, -0.00000951, 0.00000200, 0.00001057, -0.00000857, -0.00000837, 0.00000372, 0.00000144, },
		{ -0.00000240, -0.00001053, 0.00000907, -0.00001109, 0.00000421, -0.00001265, 0.00000413, -0.00001158, 0.00000788, -0.00004301, -0.00000430, -0.00000453, 0.00003257, 0.00000531, 0.00000748, 0.00001157, 0.00000006, -0.00001913, 0.00001857, -0.00000686, 0.00002114, -0.00002598, -0.00002736, -0.00018217, 0.00013261, 0.00012997, 0.00005314, 0.00006808, 0.00019205, -0.00006379, -0.00011476, -0.00001464, -0.00003849, 0.00001288, -0.00002705, 0.00000172, -0.00002161, 0.00000489, -0.00000832, -0.00000308, -0.00000919, 0.00000801, -0.00000993, 0.00000433, -0.00002050, 0.00001345, -0.00001750, 0.00000813, -0.00000842, 0.00000561, 0.00001038, -0.00000592, -0.00000091, -0.00000082, 0.00000135, -0.00000222, 0.00000400, 0.00000127, -0.00000670, -0.00000015, 0.00000651, -0.00000361, -0.00000358, -0.00000128, 0.00000974, -0.00000228, -0.00000039, 0.00000144, -0.00000173, -0.00000067, -0.00000872, 0.00000472, -0.00000084, 0.00000522, -0.00000364, 0.00000108, -0.00000708, 0.00000448, 0.00000178, -0.00000139, -0.00000189, 0.00000022, -0.00000587, 0.00000472, -0.00000190, 0.00000246, 0.00000466, -0.00000395, -0.00000434, 0.00000049, 0.00000780, -0.00000415, -0.00000123, -0.00000477, 0.00000100, -0.00000099, 0.00000746, -0.00000452, -0.00000681, -0.00000029, 0.00000678, -0.00000245, 0.00000546, -0.00000089, -0.00000457, -0.00000030, -0.00000056, -0.00000092, 0.00000539, -0.00000291, -0.00000414, 0.00000379, 0.00000006, 0.00000083, 0.00000036, 0.00000171, -0.00000489, 0.00000116, 0.00000522, -0.00000023, -0.00000795, -0.00000196, 0.00001312, -0.00000777, -0.00000817, 0.00000056, 0.00000472, -0.00000591, },
		{ -0.00001283, 0.00001063, -0.00001174, 0.00000507, -0.00001350, 0.00000494, -0.00001245, 0.00000759, -0.00003824, -0.00001141, -0.00000461, 0.00003009, 0.00001098, 0.00000530, 0.00001372, 0.00000001, -0.00001817, 0.00001862, -0.00000647, 0.00002236, -0.00002344, -0.00002569, -0.00017860, 0.00012319, 0.00012567, 0.00005314, 0.00005516, 0.00017902, -0.00001303, -0.00013634, -0.00001016, -0.00004318, 0.00001450, -0.00002880, 0.00000361, -0.00002316, 0.00000508, -0.00000901, -0.00000131, -0.00001083, 0.00000912, -0.00001168, 0.00000695, -0.00002205, 0.00001225, -0.00001437, 0.00000383, -0.00000347, 0.00000014, 0.00001317, -0.00000543, -0.00000153, -0.00000078, 0.00000115, -0.00000200, 0.00000335, 0.00000258, -0.00000678, -0.00000124, 0.00000627, -0.00000175, -0.00000445, -0.00000165, 0.00000843, 0.00000058, -0.00000302, 0.00000360, -0.00000350, 0.00000171, -0.00001043, 0.00000467, -0.00000089, 0.00000547, -0.00000343, 0.00000125, -0.00000681, 0.00000283, 0.00000297, -0.00000138, -0.00000209, 0.00000045, -0.00000553, 0.00000327, -0.00000014, 0.00000041, 0.00000604, -0.00000375, -0.00000381, -0.00000157, 0.00000869, -0.00000347, -0.00000133, -0.00000484, 0.00000069, -0.00000124, 0.00000677, -0.00000219, -0.00000750, -0.00000147, 0.00000672, -0.00000157, 0.00000400, 0.00000118, -0.00000543, -0.00000034, -0.00000038, -0.00000140, 0.00000509, -0.00000125, -0.00000521, 0.00000357, 0.00000019, 0.00000069, 0.00000075, 0.00000130, -0.00000342, -0.00000139, 0.00000689, -0.00000044, -0.00000629, -0.00000547, 0.00001458, -0.00000628, -0.00000791, -0.00000248, 0.00000726, -0.00000705, -0.00000107, },
		{ 0.00001073, -0.00001093, 0.00000451, -0.00001279, 0.00000394, -0.00001117, 0.00000487, -0.00003150, -0.00001879, -0.00000493, 0.00002686, 0.00001697, 0.00000331, 0.00001579, -0.00000001, -0.00001705, 0.00001838, -0.00000558, 0.00002285, -0.00002001, -0.00002508, -0.00017391, 0.00011267, 0.00012357, 0.00005023, 0.00004913, 0.00015737, 0.00003619, -0.00014911, -0.00001237, -0.00004342, 0.00001275, -0.00002800, 0.00000328, -0.00002263, 0.00000354, -0.00000835, -0.00000086, -0.00001107, 0.00000891, -0.00001210, 0.00000785, -0.00002135, 0.00000879, -0.00000955, -0.00000139, 0.00000163, -0.00000458, 0.00001461, -0.00000390, -0.00000262, -0.00000047, 0.00000071, -0.00000154, 0.00000250, 0.00000386, -0.00000668, -0.00000213, 0.00000554, 0.00000030, -0.00000526, -0.00000176, 0.00000665, 0.00000350, -0.00000528, 0.00000520, -0.00000469, 0.00000334, -0.00001107, 0.00000364, -0.00000026, 0.00000511, -0.00000266, 0.00000088, -0.00000592, 0.00000080, 0.00000415, -0.00000132, -0.00000212, 0.00000032, -0.00000475, 0.00000156, 0.00000160, -0.00000131, 0.00000686, -0.00000319, -0.00000341, -0.00000318, 0.00000877, -0.00000217, -0.00000181, -0.00000451, -0.00000002, -0.00000100, 0.00000551, 0.00000025, -0.00000792, -0.00000255, 0.00000620, -0.00000032, 0.00000237, 0.00000302, -0.00000585, -0.00000066, -0.00000010, -0.00000179, 0.00000456, 0.00000037, -0.00000588, 0.00000297, 0.00000059, 0.00000028, 0.00000141, 0.00000057, -0.00000168, -0.00000381, 0.00000801, -0.00000025, -0.00000485, -0.00000820, 0.00001478, -0.00000411, -0.00000769, -0.00000514, 0.00000877, -0.00000710, -0.00000069, -0.00001397, },
		{ -0.00000883, 0.00000276, -0.00001078, 0.00000150, -0.00000827, 0.00000052, -0.00002392, -0.00002534, -0.00000619, 0.00002337, 0.00002267, 0.00000202, 0.00001739, 0.00000034, -0.00001605, 0.00001813, -0.00000445, 0.00002288, -0.00001600, -0.00002522, -0.00016841, 0.00010138, 0.00012316, 0.00004509, 0.00004831, 0.00013095, 0.00007959, -0.00015096, -0.00002152, -0.00003949, 0.00000800, -0.00002497, 0.00000098, -0.00002032, 0.00000057, -0.00000658, -0.00000154, -0.00001006, 0.00000753, -0.00001132, 0.00000719, -0.00001874, 0.00000371, -0.00000388, -0.00000663, 0.00000606, -0.00000793, 0.00001441, -0.00000135, -0.00000408, 0.00000003, 0.00000011, -0.00000094, 0.00000155, 0.00000499, -0.00000638, -0.00000284, 0.00000444, 0.00000236, -0.00000583, -0.00000178, 0.00000473, 0.00000607, -0.00000683, 0.00000601, -0.00000509, 0.00000405, -0.00001058, 0.00000172, 0.00000091, 0.00000427, -0.00000148, 0.00000011, -0.00000462, -0.00000133, 0.00000510, -0.00000110, -0.00000206, -0.00000006, -0.00000373, -0.00000016, 0.00000306, -0.00000250, 0.00000701, -0.00000223, -0.00000320, -0.00000426, 0.00000805, -0.00000037, -0.00000254, -0.00000387, -0.00000102, -0.00000043, 0.00000394, 0.00000254, -0.00000794, -0.00000358, 0.00000537, 0.00000107, 0.00000085, 0.00000441, -0.00000574, -0.00000126, 0.00000028, -0.00000212, 0.00000390, 0.00000183, -0.00000608, 0.00000200, 0.00000117, -0.00000030, 0.00000221, -0.00000030, 0.00000005, -0.00000577, 0.00000836, 0.00000050, -0.00000384, -0.00000994, 0.00001373, -0.00000142, -0.00000747, -0.00000725, 0.00000915, -0.00000605, -0.00000122, -0.00001399, 0.00000946, },
		{ 0.00000028, -0.00000799, -0.00000179, -0.00000452, -0.00000441, -0.00001672, -0.00003005, -0.00000887, 0.00002000, 0.00002760, 0.00000186, 0.00001820, 0.00000131, -0.00001542, 0.00001808, -0.00000334, 0.00002275, -0.00001174, -0.00002573, -0.00016255, 0.00008984, 0.00012374, 0.00003869, 0.00005071, 0.00010341, 0.00011393, -0.00014120, -0.00003669, -0.00003247, 0.00000118, -0.00002048, -0.00000262, -0.00001682, -0.00000327, -0.00000416, -0.00000299, -0.00000817, 0.00000535, -0.00000970, 0.00000535, -0.00001482, -0.00000216, 0.00000167, -0.00001102, 0.00000914, -0.00000950, 0.00001251, 0.00000204, -0.00000568, 0.00000058, -0.00000054, -0.00000032, 0.00000065, 0.00000586, -0.00000584, -0.00000341, 0.00000311, 0.00000421, -0.00000602, -0.00000186, 0.00000291, 0.00000803, -0.00000748, 0.00000591, -0.00000466, 0.00000384, -0.00000910, -0.00000084, 0.00000236, 0.00000319, -0.00000009, -0.00000083, -0.00000318, -0.00000330, 0.00000563, -0.00000065, -0.00000197, -0.00000055, -0.00000270, -0.00000163, 0.00000403, -0.00000302, 0.00000650, -0.00000093, -0.00000319, -0.00000478, 0.00000665, 0.00000173, -0.00000334, -0.00000312, -0.00000211, 0.00000025, 0.00000233, 0.00000442, -0.00000743, -0.00000463, 0.00000436, 0.00000237, -0.00000035, 0.00000522, -0.00000508, -0.00000210, 0.00000072, -0.00000238, 0.00000317, 0.00000304, -0.00000580, 0.00000075, 0.00000185, -0.00000094, 0.00000301, -0.00000110, 0.00000153, -0.00000705, 0.00000779, 0.00000184, -0.00000331, -0.00001069, 0.00001165, 0.00000153, -0.00000717, -0.00000880, 0.00000845, -0.00000409, -0.00000248, -0.00001309, 0.00000713, -0.00000587, },
		{ -0.00000501, -0.00000523, -0.00000076, -0.00000894, -0.00001093, -0.00003223, -0.00001317, 0.00001689, 0.00003144, 0.00000306, 0.00001809, 0.00000302, -0.00001526, 0.00001839, -0.00000242, 0.00002271, -0.00000752, -0.00002627, -0.00015673, 0.00007860, 0.00012453, 0.00003208, 0.00005441, 0.00007778, 0.00013726, -0.00012057, -0.00005594, -0.00002405, -0.00000642, -0.00001556, -0.00000667, -0.00001292, -0.00000727, -0.00000168, -0.00000471, -0.00000589, 0.00000286, -0.00000769, 0.00000288, -0.00001031, -0.00000792, 0.00000625, -0.00001388, 0.00001044, -0.00000916, 0.00000912, 0.00000592, -0.00000717, 0.00000100, -0.00000111, 0.00000019, -0.00000008, 0.00000640, -0.00000505, -0.00000386, 0.00000169, 0.00000569, -0.00000575, -0.00000211, 0.00000141, 0.00000917, -0.00000717, 0.00000497, -0.00000351, 0.00000289, -0.00000687, -0.00000371, 0.00000378, 0.00000213, 0.00000125, -0.00000170, -0.00000184, -0.00000485, 0.00000560, 0.00000008, -0.00000191, -0.00000103, -0.00000180, -0.00000270, 0.00000441, -0.00000287, 0.00000541, 0.00000060, -0.00000331, -0.00000482, 0.00000476, 0.00000385, -0.00000397, -0.00000243, -0.00000311, 0.00000084, 0.00000093, 0.00000568, -0.00000633, -0.00000571, 0.00000330, 0.00000343, -0.00000109, 0.00000541, -0.00000395, -0.00000307, 0.00000116, -0.00000256, 0.00000241, 0.00000393, -0.00000508, -0.00000066, 0.00000248, -0.00000151, 0.00000366, -0.00000168, 0.00000257, -0.00000751, 0.00000634, 0.00000367, -0.00000318, -0.00001062, 0.00000886, 0.00000438, -0.00000663, -0.00000984, 0.00000684, -0.00000153, -0.00000411, -0.00001163, 0.00000420, -0.00000259, -0.00000236, },
		{ -0.00000818, 0.00000226, -0.00001225, -0.00000727, -0.00003160, -0.00001889, 0.00001395, 0.00003404, 0.00000566, 0.00001711, 0.00000543, -0.00001556, 0.00001907, -0.00000176, 0.00002287, -0.00000353, -0.00002658, -0.00015132, 0.00006808, 0.00012489, 0.00002614, 0.00005783, 0.00005620, 0.00014904, -0.00009105, -0.00007664, -0.00001620, -0.00001336, -0.00001129, -0.00001030, -0.00000938, -0.00001076, 0.00000031, -0.00000621, -0.00000371, 0.00000055, -0.00000578, 0.00000031, -0.00000590, -0.00001278, 0.00000917, -0.00001481, 0.00000984, -0.00000707, 0.00000465, 0.00000982, -0.00000821, 0.00000110, -0.00000146, 0.00000050, -0.00000055, 0.00000656, -0.00000402, -0.00000420, 0.00000028, 0.00000669, -0.00000497, -0.00000258, 0.00000031, 0.00000947, -0.00000599, 0.00000336, -0.00000188, 0.00000145, -0.00000426, -0.00000649, 0.00000483, 0.00000135, 0.00000231, -0.00000228, -0.00000084, -0.00000583, 0.00000495, 0.00000109, -0.00000189, -0.00000143, -0.00000118, -0.00000328, 0.00000420, -0.00000215, 0.00000394, 0.00000217, -0.00000346, -0.00000454, 0.00000264, 0.00000572, -0.00000423, -0.00000197, -0.00000384, 0.00000116, -0.00000008, 0.00000622, -0.00000468, -0.00000679, 0.00000224, 0.00000414, -0.00000132, 0.00000503, -0.00000249, -0.00000405, 0.00000151, -0.00000262, 0.00000164, 0.00000451, -0.00000401, -0.00000209, 0.00000293, -0.00000189, 0.00000406, -0.00000192, 0.00000308, -0.00000713, 0.00000414, 0.00000575, -0.00000323, -0.00001000, 0.00000581, 0.00000675, -0.00000568, -0.00001048, 0.00000461, 0.00000124, -0.00000575, -0.00001002, 0.00000117, 0.00000044, -0.00000460, -0.00000243, },
		{ 0.00000403, -0.00001384, -0.00000603, -0.00002836, -0.00002549, 0.00001085, 0.00003547, 0.00000948, 0.00001549, 0.00000833, -0.00001615, 0.00002003, -0.00000133, 0.00002328, 0.00000017, -0.00002652, -0.00014649, 0.00005857, 0.00012440, 0.00002149, 0.00005990, 0.00003982, 0.00015002, -0.00005561, -0.00009585, -0.00001082, -0.00001841, -0.00000858, -0.00001278, -0.00000686, -0.00001321, 0.00000138, -0.00000712, -0.00000202, -0.00000119, -0.00000436, -0.00000189, -0.00000219, -0.00001617, 0.00001007, -0.00001374, 0.00000754, -0.00000367, -0.00000034, 0.00001324, -0.00000852, 0.00000076, -0.00000151, 0.00000053, -0.00000073, 0.00000635, -0.00000283, -0.00000443, -0.00000102, 0.00000714, -0.00000371, -0.00000323, -0.00000035, 0.00000901, -0.00000417, 0.00000139, -0.00000006, -0.00000015, -0.00000165, -0.00000879, 0.00000524, 0.00000104, 0.00000293, -0.00000242, -0.00000030, -0.00000614, 0.00000371, 0.00000228, -0.00000189, -0.00000169, -0.00000087, -0.00000338, 0.00000350, -0.00000104, 0.00000231, 0.00000359, -0.00000350, -0.00000409, 0.00000055, 0.00000710, -0.00000397, -0.00000186, -0.00000421, 0.00000110, -0.00000060, 0.00000604, -0.00000258, -0.00000778, 0.00000121, 0.00000449, -0.00000110, 0.00000423, -0.00000089, -0.00000487, 0.00000170, -0.00000254, 0.00000089, 0.00000478, -0.00000269, -0.00000338, 0.00000309, -0.00000201, 0.00000414, -0.00000178, 0.00000305, -0.00000604, 0.00000147, 0.00000777, -0.00000320, -0.00000916, 0.00000289, 0.00000833, -0.00000423, -0.00001082, 0.00000203, 0.00000384, -0.00000700, -0.00000865, -0.00000149, 0.00000276, -0.00000600, -0.00000067, -0.00001016, },
		{ -0.00001359, -0.00000706, -0.00002313, -0.00003215, 0.00000718, 0.00003589, 0.00001423, 0.00001360, 0.00001143, -0.00001678, 0.00002108, -0.00000100, 0.00002384, 0.00000361, -0.00002611, -0.00014231, 0.00005017, 0.00012285, 0.00001840, 0.00006010, 0.00002886, 0.00014199, -0.00001768, -0.00011074, -0.00000943, -0.00002071, -0.00000796, -0.00001369, -0.00000570, -0.00001431, 0.00000134, -0.00000725, -0.00000105, -0.00000212, -0.00000367, -0.00000341, 0.00000044, -0.00001778, 0.00000892, -0.00001095, 0.00000402, 0.00000045, -0.00000519, 0.00001569, -0.00000789, -0.00000010, -0.00000124, 0.00000027, -0.00000062, 0.00000584, -0.00000155, -0.00000452, -0.00000217, 0.00000706, -0.00000209, -0.00000399, -0.00000067, 0.00000797, -0.00000202, -0.00000061, 0.00000163, -0.00000158, 0.00000061, -0.00001031, 0.00000483, 0.00000130, 0.00000304, -0.00000208, -0.00000025, -0.00000583, 0.00000203, 0.00000351, -0.00000182, -0.00000183, -0.00000086, -0.00000308, 0.00000246, 0.00000024, 0.00000076, 0.00000467, -0.00000331, -0.00000364, -0.00000129, 0.00000780, -0.00000313, -0.00000212, -0.00000419, 0.00000066, -0.00000065, 0.00000523, -0.00000025, -0.00000854, 0.00000016, 0.00000451, -0.00000054, 0.00000319, 0.00000062, -0.00000540, 0.00000167, -0.00000230, 0.00000016, 0.00000479, -0.00000126, -0.00000439, 0.00000289, -0.00000184, 0.00000392, -0.00000131, 0.00000257, -0.00000446, -0.00000132, 0.00000939, -0.00000283, -0.00000840, 0.00000047, 0.00000894, -0.00000227, -0.00001093, -0.00000063, 0.00000594, -0.00000762, -0.00000780, -0.00000347, 0.00000404, -0.00000630, 0.00000002, -0.00001094, 0.00000434, },
		{ -0.00000981, -0.00001681, -0.00003795, 0.00000253, 0.00003546, 0.00001952, 0.00001185, 0.00001438, -0.00001718, 0.00002199, -0.00000058, 0.00002441, 0.00000694, -0.00002544, -0.00013868, 0.00004281, 0.00012028, 0.00001682, 0.00005841, 0.00002280, 0.00012734, 0.00001928, -0.00011907, -0.00001284, -0.00001999, -0.00000955, -0.00001298, -0.00000596, -0.00001408, 0.00000022, -0.00000662, -0.00000084, -0.00000218, -0.00000376, -0.00000416, 0.00000187, -0.00001761, 0.00000601, -0.00000698, -0.00000007, 0.00000459, -0.00000928, 0.00001681, -0.00000621, -0.00000144, -0.00000069, -0.00000023, -0.00000028, 0.00000512, -0.00000031, -0.00000444, -0.00000315, 0.00000651, -0.00000024, -0.00000473, -0.00000078, 0.00000659, 0.00000014, -0.00000232, 0.00000290, -0.00000259, 0.00000225, -0.00001087, 0.00000357, 0.00000208, 0.00000269, -0.00000132, -0.00000062, -0.00000503, 0.00000010, 0.00000459, -0.00000161, -0.00000189, -0.00000108, -0.00000254, 0.00000130, 0.00000146, -0.00000050, 0.00000528, -0.00000283, -0.00000329, -0.00000270, 0.00000775, -0.00000173, -0.00000267, -0.00000385, -0.00000011, -0.00000031, 0.00000401, 0.00000207, -0.00000892, -0.00000094, 0.00000428, 0.00000020, 0.00000211, 0.00000188, -0.00000555, 0.00000139, -0.00000192, -0.00000053, 0.00000457, 0.00000017, -0.00000503, 0.00000231, -0.00000143, 0.00000345, -0.00000059, 0.00000181, -0.00000264, -0.00000390, 0.00001032, -0.00000190, -0.00000794, -0.00000123, 0.00000853, 0.00000009, -0.00001081, -0.00000318, 0.00000732, -0.00000744, -0.00000758, -0.00000461, 0.00000419, -0.00000549, -0.00000032, -0.00001055, 0.00000328, -0.00001175, },
		{ -0.00001044, -0.00004204, -0.00000338, 0.00003431, 0.00002498, 0.00001063, 0.00001688, -0.00001708, 0.00002256, 0.00000011, 0.00002483, 0.00001031, -0.00002468, -0.00013545, 0.00003631, 0.00011690, 0.00001649, 0.00005516, 0.00002062, 0.00010877, 0.00005220, -0.00011943, -0.00002108, -0.00001661, -0.00001296, -0.00001097, -0.00000737, -0.00001278, -0.00000170, -0.00000544, -0.00000125, -0.00000151, -0.00000451, -0.00000420, 0.00000215, -0.00001591, 0.00000188, -0.00000253, -0.00000402, 0.00000811, -0.00001212, 0.00001639, -0.00000357, -0.00000312, 0.00000002, -0.00000086, 0.00000017, 0.00000432, 0.00000079, -0.00000413, -0.00000398, 0.00000558, 0.00000167, -0.00000529, -0.00000081, 0.00000513, 0.00000201, -0.00000349, 0.00000357, -0.00000301, 0.00000315, -0.00001044, 0.00000155, 0.00000325, 0.00000201, -0.00000027, -0.00000126, -0.00000395, -0.00000184, 0.00000535, -0.00000118, -0.00000192, -0.00000142, -0.00000191, 0.00000021, 0.00000242, -0.00000131, 0.00000537, -0.00000203, -0.00000310, -0.00000362, 0.00000699, 0.00000008, -0.00000339, -0.00000331, -0.00000105, 0.00000023, 0.00000259, 0.00000413, -0.00000878, -0.00000215, 0.00000390, 0.00000096, 0.00000119, 0.00000275, -0.00000529, 0.00000089, -0.00000142, -0.00000115, 0.00000418, 0.00000148, -0.00000525, 0.00000140, -0.00000086, 0.00000284, 0.00000023, 0.00000096, -0.00000088, -0.00000594, 0.00001034, -0.00000034, -0.00000786, -0.00000217, 0.00000723, 0.00000263, -0.00001040, -0.00000552, 0.00000789, -0.00000648, -0.00000799, -0.00000493, 0.00000329, -0.00000376, -0.00000146, -0.00000927, 0.00000119, -0.00000884, -0.00001351, },
		{ -0.00004377, -0.00001055, 0.00003241, 0.00003029, 0.00001027, 0.00001870, -0.00001631, 0.00002264, 0.00000120, 0.00002496, 0.00001387, -0.00002397, -0.00013246, 0.00003045, 0.00011296, 0.00001702, 0.00005091, 0.00002107, 0.00008879, 0.00007878, -0.00011143, -0.00003330, -0.00001145, -0.00001747, -0.00000826, -0.00000945, -0.00001086, -0.00000400, -0.00000404, -0.00000202, -0.00000034, -0.00000567, -0.00000374, 0.00000154, -0.00001313, -0.00000281, 0.00000166, -0.00000716, 0.00001050, -0.00001341, 0.00001444, -0.00000018, -0.00000490, 0.00000074, -0.00000150, 0.00000062, 0.00000358, 0.00000164, -0.00000359, -0.00000466, 0.00000439, 0.00000344, -0.00000554, -0.00000092, 0.00000379, 0.00000338, -0.00000399, 0.00000357, -0.00000283, 0.00000331, -0.00000913, -0.00000098, 0.00000455, 0.00000122, 0.00000087, -0.00000196, -0.00000280, -0.00000353, 0.00000561, -0.00000046, -0.00000196, -0.00000179, -0.00000135, -0.00000064, 0.00000300, -0.00000161, 0.00000495, -0.00000099, -0.00000308, -0.00000406, 0.00000563, 0.00000209, -0.00000406, -0.00000273, -0.00000200, 0.00000079, 0.00000124, 0.00000569, -0.00000803, -0.00000348, 0.00000344, 0.00000162, 0.00000054, 0.00000317, -0.00000465, 0.00000023, -0.00000088, -0.00000168, 0.00000364, 0.00000262, -0.00000504, 0.00000024, -0.00000024, 0.00000222, 0.00000101, 0.00000020, 0.00000061, -0.00000723, 0.00000938, 0.00000181, -0.00000812, -0.00000245, 0.00000531, 0.00000506, -0.00000960, -0.00000758, 0.00000768, -0.00000487, -0.00000885, -0.00000461, 0.00000162, -0.00000147, -0.00000301, -0.00000751, -0.00000140, -0.00000551, -0.00001730, -0.00000500, },
		{ -0.00001870, 0.00002960, 0.00003524, 0.00001097, 0.00001975, -0.00001482, 0.00002221, 0.00000272, 0.00002476, 0.00001771, -0.00002342, -0.00012957, 0.00002503, 0.00010874, 0.00001799, 0.00004623, 0.00002292, 0.00006953, 0.00009766, -0.00009573, -0.00004798, -0.00000580, -0.00002211, -0.00000556, -0.00001160, -0.00000886, -0.00000620, -0.00000280, -0.00000287, 0.00000103, -0.00000695, -0.00000306, 0.00000038, -0.00000980, -0.00000735, 0.00000492, -0.00000904, 0.00001148, -0.00001311, 0.00001121, 0.00000359, -0.00000649, 0.00000126, -0.00000200, 0.00000092, 0.00000301, 0.00000216, -0.00000280, -0.00000519, 0.00000304, 0.00000495, -0.00000540, -0.00000119, 0.00000274, 0.00000412, -0.00000378, 0.00000296, -0.00000211, 0.00000285, -0.00000717, -0.00000372, 0.00000569, 0.00000054, 0.00000189, -0.00000252, -0.00000182, -0.00000477, 0.00000530, 0.00000054, -0.00000202, -0.00000212, -0.00000096, -0.00000117, 0.00000315, -0.00000144, 0.00000413, 0.00000020, -0.00000316, -0.00000410, 0.00000388, 0.00000404, -0.00000450, -0.00000230, -0.00000280, 0.00000119, 0.00000015, 0.00000661, -0.00000667, -0.00000489, 0.00000295, 0.00000208, 0.00000022, 0.00000316, -0.00000373, -0.00000047, -0.00000036, -0.00000205, 0.00000297, 0.00000354, -0.00000444, -0.00000104, 0.00000030, 0.00000170, 0.00000163, -0.00000032, 0.00000164, -0.00000767, 0.00000752, 0.00000435, -0.00000854, -0.00000228, 0.00000310, 0.00000706, -0.00000830, -0.00000938, 0.00000679, -0.00000285, -0.00000993, -0.00000390, -0.00000046, 0.00000094, -0.00000453, -0.00000576, -0.00000395, -0.00000244, -0.00002038, -0.00000123, -0.00004289, },
		{ 0.00002560, 0.00003967, 0.00001281, 0.00002009, -0.00001267, 0.00002134, 0.00000463, 0.00002425, 0.00002183, -0.00002305, -0.00012673, 0.00001992, 0.00010446, 0.00001907, 0.00004163, 0.00002516, 0.00005252, 0.00010845, -0.00007386, -0.00006316, -0.00000108, -0.00002592, -0.00000358, -0.00001325, -0.00000727, -0.00000789, -0.00000203, -0.00000351, 0.00000230, -0.00000807, -0.00000244, -0.00000098, -0.00000644, -0.00001114, 0.00000677, -0.00000939, 0.00001102, -0.00001140, 0.00000707, 0.00000729, -0.00000759, 0.00000142, -0.00000225, 0.00000100, 0.00000270, 0.00000232, -0.00000181, -0.00000559, 0.00000164, 0.00000607, -0.00000482, -0.00000167, 0.00000203, 0.00000422, -0.00000297, 0.00000189, -0.00000105, 0.00000198, -0.00000487, -0.00000630, 0.00000641, 0.00000017, 0.00000263, -0.00000279, -0.00000117, -0.00000546, 0.00000439, 0.00000176, -0.00000208, -0.00000236, -0.00000081, -0.00000136, 0.00000292, -0.00000089, 0.00000307, 0.00000138, -0.00000323, -0.00000386, 0.00000197, 0.00000569, -0.00000451, -0.00000213, -0.00000331, 0.00000128, -0.00000053, 0.00000683, -0.00000478, -0.00000628, 0.00000240, 0.00000233, 0.00000021, 0.00000282, -0.00000268, -0.00000110, 0.00000005, -0.00000224, 0.00000220, 0.00000422, -0.00000353, -0.00000230, 0.00000062, 0.00000139, 0.00000198, -0.00000050, 0.00000214, -0.00000729, 0.00000496, 0.00000699, -0.00000884, -0.00000195, 0.00000100, 0.00000835, -0.00000645, -0.00001092, 0.00000538, -0.00000068, -0.00001096, -0.00000311, -0.00000258, 0.00000305, -0.00000559, -0.00000442, -0.00000599, -0.00000016, -0.00002222, 0.00000050, -0.00003953, -0.00002723, },
		{ 0.00004345, 0.00001576, 0.00001986, -0.00001002, 0.00002017, 0.00000680, 0.00002352, 0.00002617, -0.00002281, -0.00012394, 0.00001506, 0.00010023, 0.00002003, 0.00003747, 0.00002706, 0.00003860, 0.00011167, -0.00004797, -0.00007671, 0.00000143, -0.00002812, -0.00000285, -0.00001402, -0.00000646, -0.00000877, -0.00000193, -0.00000380, 0.00000328, -0.00000879, -0.00000207, -0.00000226, -0.00000348, -0.00001375, 0.00000699, -0.00000828, 0.00000934, -0.00000867, 0.00000258, 0.00001045, -0.00000793, 0.00000110, -0.00000217, 0.00000079, 0.00000266, 0.00000216, -0.00000069, -0.00000582, 0.00000026, 0.00000675, -0.00000383, -0.00000233, 0.00000166, 0.00000379, -0.00000175, 0.00000059, 0.00000013, 0.00000096, -0.00000255, -0.00000841, 0.00000648, 0.00000026, 0.00000296, -0.00000267, -0.00000091, -0.00000555, 0.00000298, 0.00000307, -0.00000207, -0.00000251, -0.00000088, -0.00000126, 0.00000240, -0.00000013, 0.00000195, 0.00000239, -0.00000321, -0.00000350, 0.00000013, 0.00000682, -0.00000400, -0.00000229, -0.00000349, 0.00000103, -0.00000077, 0.00000640, -0.00000252, -0.00000751, 0.00000177, 0.00000241, 0.00000041, 0.00000229, -0.00000166, -0.00000153, 0.00000029, -0.00000223, 0.00000135, 0.00000466, -0.00000240, -0.00000341, 0.00000064, 0.00000134, 0.00000203, -0.00000034, 0.00000216, -0.00000625, 0.00000202, 0.00000936, -0.00000875, -0.00000177, -0.00000068, 0.00000874, -0.00000409, -0.00001217, 0.00000360, 0.00000139, -0.00001169, -0.00000249, -0.00000441, 0.00000453, -0.00000591, -0.00000377, -0.00000723, 0.00000102, -0.00002256, 0.00000017, -0.00003422, -0.00003534, 0.00002009, },
		{ 0.00001975, 0.00001929, -0.00000709, 0.00001892, 0.00000906, 0.00002273, 0.00003061, -0.00002259, -0.00012129, 0.00001047, 0.00009608, 0.00002078, 0.00003392, 0.00002822, 0.00002806, 0.00010851, -0.00002051, -0.00008671, 0.00000083, -0.00002830, -0.00000359, -0.00001372, -0.00000656, -0.00000875, -0.00000253, -0.00000369, 0.00000389, -0.00000904, -0.00000207, -0.00000327, -0.00000120, -0.00001496, 0.00000563, -0.00000597, 0.00000686, -0.00000543, -0.00000173, 0.00001266, -0.00000733, 0.00000026, -0.00000178, 0.00000032, 0.00000287, 0.00000174, 0.00000048, -0.00000585, -0.00000106, 0.00000698, -0.00000250, -0.00000308, 0.00000155, 0.00000298, -0.00000037, -0.00000069, 0.00000119, 0.00000003, -0.00000050, -0.00000978, 0.00000578, 0.00000082, 0.00000288, -0.00000215, -0.00000105, -0.00000513, 0.00000122, 0.00000431, -0.00000192, -0.00000260, -0.00000112, -0.00000098, 0.00000175, 0.00000067, 0.00000095, 0.00000312, -0.00000301, -0.00000314, -0.00000144, 0.00000729, -0.00000294, -0.00000277, -0.00000333, 0.00000046, -0.00000063, 0.00000547, -0.00000014, -0.00000841, 0.00000096, 0.00000238, 0.00000073, 0.00000173, -0.00000082, -0.00000170, 0.00000032, -0.00000201, 0.00000046, 0.00000485, -0.00000116, -0.00000426, 0.00000032, 0.00000155, 0.00000182, 0.00000011, 0.00000178, -0.00000477, -0.00000094, 0.00001112, -0.00000803, -0.00000194, -0.00000173, 0.00000821, -0.00000132, -0.00001307, 0.00000157, 0.00000316, -0.00001196, -0.00000220, -0.00000574, 0.00000518, -0.00000538, -0.00000389, -0.00000757, 0.00000106, -0.00002149, -0.00000183, -0.00002779, -0.00004212, 0.00001283, 0.00004644, },
		{ 0.00001865, -0.00000412, 0.00001780, 0.00001121, 0.00002203, 0.00003502, -0.00002226, -0.00011889, 0.00000621, 0.00009199, 0.00002130, 0.00003099, 0.00002854, 0.00002065, 0.00010059, 0.00000611, -0.00009170, -0.00000327, -0.00002646, -0.00000570, -0.00001249, -0.00000748, -0.00000796, -0.00000368, -0.00000329, 0.00000416, -0.00000882, -0.00000243, -0.00000395, 0.00000028, -0.00001478, 0.00000295, -0.00000294, 0.00000411, -0.00000223, -0.00000533, 0.00001362, -0.00000577, -0.00000102, -0.00000114, -0.00000034, 0.00000324, 0.00000117, 0.00000156, -0.00000564, -0.00000229, 0.00000680, -0.00000096, -0.00000381, 0.00000157, 0.00000200, 0.00000094, -0.00000171, 0.00000193, -0.00000061, 0.00000108, -0.00001030, 0.00000431, 0.00000178, 0.00000247, -0.00000133, -0.00000149, -0.00000433, -0.00000066, 0.00000529, -0.00000153, -0.00000268, -0.00000146, -0.00000064, 0.00000112, 0.00000135, 0.00000023, 0.00000349, -0.00000259, -0.00000286, -0.00000262, 0.00000708, -0.00000140, -0.00000346, -0.00000295, -0.00000034, -0.00000021, 0.00000425, 0.00000211, -0.00000880, -0.00000006, 0.00000231, 0.00000103, 0.00000129, -0.00000027, -0.00000156, 0.00000016, -0.00000161, -0.00000042, 0.00000480, 0.00000013, -0.00000477, -0.00000035, 0.00000195, 0.00000141, 0.00000074, 0.00000119, -0.00000313, -0.00000355, 0.00001198, -0.00000652, -0.00000258, -0.00000208, 0.00000686, 0.00000160, -0.00001351, -0.00000060, 0.00000449, -0.00001171, -0.00000231, -0.00000650, 0.00000498, -0.00000405, -0.00000469, -0.00000716, 0.00000016, -0.00001935, -0.00000486, -0.00002123, -0.00004675, 0.00000376, 0.00004840, 0.00002465, },
		{ -0.00000131, 0.00001698, 0.00001314, 0.00002153, 0.00003930, -0.00002171, -0.00011684, 0.00000235, 0.00008788, 0.00002166, 0.00002860, 0.00002811, 0.00001585, 0.00008965, 0.00002985, -0.00009088, -0.00001060, -0.00002307, -0.00000875, -0.00001064, -0.00000895, -0.00000665, -0.00000511, -0.00000279, 0.00000420, -0.00000824, -0.00000306, -0.00000432, 0.00000098, -0.00001341, -0.00000058, 0.00000023, 0.00000163, 0.00000045, -0.00000786, 0.00001321, -0.00000334, -0.00000257, -0.00000038, -0.00000109, 0.00000366, 0.00000058, 0.00000245, -0.00000517, -0.00000342, 0.00000628, 0.00000065, -0.00000440, 0.00000159, 0.00000105, 0.00000197, -0.00000233, 0.00000226, -0.00000087, 0.00000208, -0.00000994, 0.00000223, 0.00000298, 0.00000188, -0.00000035, -0.00000207, -0.00000337, -0.00000244, 0.00000583, -0.00000084, -0.00000279, -0.00000182, -0.00000036, 0.00000065, 0.00000179, -0.00000016, 0.00000347, -0.00000196, -0.00000271, -0.00000337, 0.00000624, 0.00000046, -0.00000421, -0.00000247, -0.00000121, 0.00000028, 0.00000297, 0.00000398, -0.00000856, -0.00000135, 0.00000226, 0.00000122, 0.00000105, -0.00000007, -0.00000114, -0.00000014, -0.00000111, -0.00000121, 0.00000451, 0.00000137, -0.00000492, -0.00000130, 0.00000245, 0.00000094, 0.00000142, 0.00000058, -0.00000162, -0.00000552, 0.00001178, -0.00000425, -0.00000363, -0.00000184, 0.00000494, 0.00000437, -0.00001337, -0.00000286, 0.00000535, -0.00001097, -0.00000275, -0.00000673, 0.00000406, -0.00000214, -0.00000589, -0.00000626, -0.00000132, -0.00001669, -0.00000816, -0.00001547, -0.00004866, -0.00000690, 0.00004898, 0.00003037, 0.00001819, },
		{ 0.00001658, 0.00001475, 0.00002131, 0.00004338, -0.00002086, -0.00011520, -0.00000106, 0.00008370, 0.00002193, 0.00002662, 0.00002712, 0.00001299, 0.00007735, 0.00004919, -0.00008420, -0.00002035, -0.00001893, -0.00001212, -0.00000868, -0.00001058, -0.00000520, -0.00000650, -0.00000244, 0.00000419, -0.00000745, -0.00000378, -0.00000449, 0.00000102, -0.00001116, -0.00000443, 0.00000297, -0.00000012, 0.00000225, -0.00000913, 0.00001150, -0.00000030, -0.00000414, 0.00000030, -0.00000177, 0.00000401, 0.00000011, 0.00000305, -0.00000443, -0.00000443, 0.00000549, 0.00000220, -0.00000473, 0.00000149, 0.00000027, 0.00000257, -0.00000247, 0.00000214, -0.00000073, 0.00000250, -0.00000882, -0.00000025, 0.00000418, 0.00000129, 0.00000062, -0.00000262, -0.00000244, -0.00000389, 0.00000583, 0.00000017, -0.00000293, -0.00000212, -0.00000025, 0.00000041, 0.00000194, -0.00000020, 0.00000313, -0.00000119, -0.00000268, -0.00000371, 0.00000490, 0.00000241, -0.00000481, -0.00000206, -0.00000200, 0.00000069, 0.00000185, 0.00000529, -0.00000764, -0.00000286, 0.00000225, 0.00000126, 0.00000105, -0.00000016, -0.00000053, -0.00000049, -0.00000057, -0.00000186, 0.00000399, 0.00000254, -0.00000471, -0.00000242, 0.00000293, 0.00000051, 0.00000199, 0.00000012, -0.00000044, -0.00000667, 0.00001051, -0.00000134, -0.00000492, -0.00000122, 0.00000281, 0.00000667, -0.00001253, -0.00000519, 0.00000574, -0.00000985, -0.00000342, -0.00000658, 0.00000265, 0.00000002, -0.00000717, -0.00000525, -0.00000294, -0.00001407, -0.00001102, -0.00001126, -0.00004761, -0.00001859, 0.00004770, 0.00003682, 0.00001811, 0.00000121, },
		{ 0.00001606, 0.00002134, 0.00004728, -0.00001970, -0.00011400, -0.00000403, 0.00007945, 0.00002216, 0.00002493, 0.00002580, 0.00001139, 0.00006505, 0.00006332, -0.00007234, -0.00003125, -0.00001504, -0.00001511, -0.00000709, -0.00001195, -0.00000396, -0.00000755, -0.00000244, 0.00000426, -0.00000663, -0.00000446, -0.00000459, 0.00000058, -0.00000840, -0.00000808, 0.00000483, -0.00000087, 0.00000304, -0.00000914, 0.00000872, 0.00000298, -0.00000545, 0.00000072, -0.00000225, 0.00000415, -0.00000015, 0.00000332, -0.00000343, -0.00000531, 0.00000451, 0.00000356, -0.00000473, 0.00000122, -0.00000022, 0.00000270, -0.00000216, 0.00000164, -0.00000026, 0.00000245, -0.00000715, -0.00000283, 0.00000511, 0.00000090, 0.00000141, -0.00000298, -0.00000172, -0.00000488, 0.00000522, 0.00000145, -0.00000309, -0.00000233, -0.00000034, 0.00000042, 0.00000183, 0.00000004, 0.00000257, -0.00000036, -0.00000269, -0.00000372, 0.00000328, 0.00000422, -0.00000508, -0.00000187, -0.00000256, 0.00000086, 0.00000106, 0.00000591, -0.00000608, -0.00000450, 0.00000222, 0.00000115, 0.00000125, -0.00000047, 0.00000013, -0.00000079, -0.00000010, -0.00000229, 0.00000323, 0.00000357, -0.00000418, -0.00000361, 0.00000325, 0.00000025, 0.00000234, -0.00000006, 0.00000026, -0.00000694, 0.00000830, 0.00000189, -0.00000615, -0.00000054, 0.00000085, 0.00000819, -0.00001093, -0.00000752, 0.00000571, -0.00000851, -0.00000413, -0.00000624, 0.00000101, 0.00000210, -0.00000819, -0.00000447, -0.00000428, -0.00001197, -0.00001287, -0.00000901, -0.00004383, -0.00003046, 0.00004402, 0.00004386, 0.00001860, 0.00000336, 0.00001662, },
		{ 0.00002158, 0.00005105, -0.00001826, -0.00011319, -0.00000661, 0.00007513, 0.00002235, 0.00002344, 0.00002433, 0.00001051, 0.00005373, 0.00007205, -0.00005652, -0.00004183, -0.00001239, -0.00001710, -0.00000629, -0.00001278, -0.00000320, -0.00000803, -0.00000292, 0.00000450, -0.00000586, -0.00000498, -0.00000470, -0.00000016, -0.00000549, -0.00001110, 0.00000551, -0.00000052, 0.00000286, -0.00000810, 0.00000527, 0.00000608, -0.00000620, 0.00000073, -0.00000243, 0.00000403, -0.00000015, 0.00000324, -0.00000224, -0.00000602, 0.00000341, 0.00000463, -0.00000435, 0.00000074, -0.00000042, 0.00000240, -0.00000150, 0.00000089, 0.00000039, 0.00000209, -0.00000519, -0.00000520, 0.00000554, 0.00000086, 0.00000190, -0.00000302, -0.00000132, -0.00000531, 0.00000404, 0.00000289, -0.00000320, -0.00000246, -0.00000063, 0.00000063, 0.00000154, 0.00000044, 0.00000193, 0.00000039, -0.00000269, -0.00000353, 0.00000157, 0.00000566, -0.00000489, -0.00000198, -0.00000281, 0.00000072, 0.00000068, 0.00000586, -0.00000402, -0.00000611, 0.00000210, 0.00000096, 0.00000156, -0.00000086, 0.00000071, -0.00000091, 0.00000023, -0.00000246, 0.00000225, 0.00000445, -0.00000340, -0.00000473, 0.00000331, 0.00000023, 0.00000242, 0.00000009, 0.00000045, -0.00000640, 0.00000543, 0.00000508, -0.00000698, -0.00000013, -0.00000061, 0.00000875, -0.00000861, -0.00000976, 0.00000529, -0.00000708, -0.00000474, -0.00000587, -0.00000059, 0.00000382, -0.00000868, -0.00000416, -0.00000508, -0.00001071, -0.00001347, -0.00000874, -0.00003795, -0.00004144, 0.00003745, 0.00005130, 0.00001983, 0.00000515, 0.00001707, 0.00001712, },
		{ 0.00005478, -0.00001660, -0.00011271, -0.00000890, 0.00007082, 0.00002250, 0.00002210, 0.00002285, 0.00000991, 0.00004397, 0.00007579, -0.00003831, -0.00005061, -0.00001176, -0.00001772, -0.00000648, -0.00001292, -0.00000304, -0.00000788, -0.00000387, 0.00000489, -0.00000520, -0.00000529, -0.00000487, -0.00000108, -0.00000272, -0.00001318, 0.00000492, 0.00000079, 0.00000196, -0.00000636, 0.00000162, 0.00000858, -0.00000616, 0.00000024, -0.00000227, 0.00000362, 0.00000011, 0.00000286, -0.00000094, -0.00000652, 0.00000224, 0.00000538, -0.00000364, 0.00000010, -0.00000036, 0.00000178, -0.00000068, 0.00000008, 0.00000106, 0.00000160, -0.00000320, -0.00000708, 0.00000532, 0.00000125, 0.00000203, -0.00000272, -0.00000127, -0.00000522, 0.00000244, 0.00000434, -0.00000317, -0.00000254, -0.00000106, 0.00000094, 0.00000119, 0.00000086, 0.00000136, 0.00000097, -0.00000260, -0.00000324, -0.00000002, 0.00000656, -0.00000416, -0.00000242, -0.00000274, 0.00000025, 0.00000069, 0.00000525, -0.00000168, -0.00000749, 0.00000177, 0.00000078, 0.00000186, -0.00000118, 0.00000105, -0.00000080, 0.00000037, -0.00000234, 0.00000110, 0.00000515, -0.00000243, -0.00000567, 0.00000305, 0.00000048, 0.00000221, 0.00000057, 0.00000018, -0.00000524, 0.00000227, 0.00000782, -0.00000711, -0.00000024, -0.00000134, 0.00000829, -0.00000572, -0.00001176, 0.00000450, -0.00000569, -0.00000514, -0.00000560, -0.00000199, 0.00000499, -0.00000853, -0.00000442, -0.00000523, -0.00001037, -0.00001285, -0.00001008, -0.00003090, -0.00005038, 0.00002766, 0.00005880, 0.00002196, 0.00000659, 0.00001785, 0.00001803, 0.00002191, },
		{ -0.00001481, -0.00011246, -0.00001097, 0.00006659, 0.00002254, 0.00002090, 0.00002142, 0.00000933, 0.00003600, 0.00007532, -0.00001941, -0.00005638, -0.00001355, -0.00001691, -0.00000762, -0.00001244, -0.00000342, -0.00000719, -0.00000516, 0.00000533, -0.00000459, -0.00000541, -0.00000508, -0.00000210, -0.00000028, -0.00001421, 0.00000318, 0.00000275, 0.00000071, -0.00000434, -0.00000177, 0.00001014, -0.00000522, -0.00000075, -0.00000181, 0.00000296, 0.00000059, 0.00000228, 0.00000035, -0.00000677, 0.00000104, 0.00000578, -0.00000263, -0.00000064, -0.00000014, 0.00000100, 0.00000013, -0.00000063, 0.00000162, 0.00000113, -0.00000139, -0.00000829, 0.00000438, 0.00000203, 0.00000184, -0.00000212, -0.00000152, -0.00000470, 0.00000059, 0.00000560, -0.00000290, -0.00000262, -0.00000155, 0.00000123, 0.00000091, 0.00000117, 0.00000096, 0.00000131, -0.00000238, -0.00000297, -0.00000134, 0.00000684, -0.00000291, -0.00000312, -0.00000242, -0.00000048, 0.00000100, 0.00000425, 0.00000069, -0.00000843, 0.00000112, 0.00000071, 0.00000202, -0.00000128, 0.00000108, -0.00000041, 0.00000032, -0.00000195, -0.00000016, 0.00000563, -0.00000133, -0.00000634, 0.00000243, 0.00000096, 0.00000176, 0.00000130, -0.00000040, -0.00000372, -0.00000079, 0.00000973, -0.00000631, -0.00000102, -0.00000126, 0.00000693, -0.00000254, -0.00001332, 0.00000333, -0.00000439, -0.00000531, -0.00000546, -0.00000309, 0.00000556, -0.00000776, -0.00000516, -0.00000483, -0.00001077, -0.00001132, -0.00001242, -0.00002380, -0.00005626, 0.00001460, 0.00006583, 0.00002521, 0.00000774, 0.00001888, 0.00001891, 0.00002223, 0.00005858, },
		},
		{ { -0.00000449, -0.00000579, 0.00000317, -0.00000844, -0.00000743, -0.00000836, -0.00001328, -0.00001574, -0.00006494, 0.00000350, 0.00007347, 0.00002631, 0.00002468, 0.00002206, 0.00002657, 0.00002708, 0.00007375, 0.00000698, -0.00006699, -0.00001265, -0.00000880, -0.00000441, -0.00001134, -0.00000728, 0.00000863, -0.00000528, -0.00000407, -0.00000280, -0.00000132, 0.00000148, -0.00001383, 0.00000157, -0.00000130, -0.00000250, -0.00000693, -0.00000081, -0.00000112, -0.00000135, -0.00000541, 0.00000098, -0.00000607, 0.00000012, 0.00000329, -0.00000243, -0.00000386, -0.00000456, -0.00000144, -0.00000028, -0.00000735, 0.00000069, 0.00000660, -0.00000318, -0.00000309, -0.00000271, 0.00001537, -0.00000203, 0.00000193, 0.00000164, -0.00000076, 0.00000213, 0.00000227, 0.00000240, -0.00000172, -0.00000095, -0.00000212, -0.00000095, -0.00000172, 0.00000240, 0.00000227, 0.00000213, -0.00000076, 0.00000164, 0.00000193, -0.00000203, 0.00001537, -0.00000271, -0.00000309, -0.00000318, 0.00000660, 0.00000069, -0.00000735, -0.00000028, -0.00000144, -0.00000456, -0.00000386, -0.00000243, 0.00000329, 0.00000012, -0.00000607, 0.00000098, -0.00000541, -0.00000135, -0.00000112, -0.00000081, -0.00000693, -0.00000250, -0.00000130, 0.00000157, -0.00001383, 0.00000148, -0.00000132, -0.00000280, -0.00000407, -0.00000528, 0.00000863, -0.00000728, -0.00001134, -0.00000441, -0.00000880, -0.00001265, -0.00006699, 0.00000698, 0.00007375, 0.00002708, 0.00002657, 0.00002206, 0.00002468, 0.00002631, 0.00007347, 0.00000350, -0.00006494, -0.00001574, -0.00001328, -0.00000836, -0.00000743, -0.00000844, 0.00000317, -0.00000579, },
		{ -0.00000624, 0.00000169, -0.00000468, -0.00000965, -0.00000614, -0.00001436, -0.00001135, -0.00005913, -0.00001272, 0.00006987, 0.00003178, 0.00002184, 0.00002163, 0.00002488, 0.00002302, 0.00006087, 0.00003769, -0.00006259, -0.00002530, -0.00000566, -0.00000647, -0.00000877, -0.00001074, 0.00000853, -0.00000337, -0.00000562, -0.00000170, -0.00000259, 0.00000303, -0.00001367, 0.00000004, -0.00000047, -0.00000224, -0.00000634, -0.00000287, -0.00000035, -0.00000082, -0.00000397, -0.00000188, -0.00000354, -0.00000267, 0.00000468, -0.00000262, -0.00000227, -0.00000534, -0.00000206, -0.00000029, -0.00000519, -0.00000365, 0.00000968, -0.00000421, -0.00000146, -0.00000601, 0.00001383, 0.00000319, -0.00000117, 0.00000294, -0.00000209, 0.00000349, 0.00000134, 0.00000334, -0.00000190, -0.00000086, -0.00000193, -0.00000136, -0.00000130, 0.00000139, 0.00000306, 0.00000135, -0.00000025, 0.00000079, 0.00000424, -0.00000442, 0.00001279, 0.00000319, -0.00000584, -0.00000117, 0.00000319, 0.00000408, -0.00000789, -0.00000133, -0.00000084, -0.00000336, -0.00000530, -0.00000239, 0.00000163, 0.00000285, -0.00000791, 0.00000222, -0.00000464, -0.00000318, -0.00000135, 0.00000006, -0.00000602, -0.00000369, -0.00000158, 0.00000239, -0.00001309, -0.00000085, 0.00000057, -0.00000433, -0.00000234, -0.00000702, 0.00000822, -0.00000394, -0.00001284, -0.00000370, -0.00001010, -0.00000644, -0.00005827, -0.00002804, 0.00007765, 0.00003720, 0.00002644, 0.00002372, 0.00002686, 0.00002278, 0.00007288, 0.00002221, -0.00006865, -0.00002189, -0.00001182, -0.00001081, -0.00000547, -0.00001090, 0.00000261, -0.00000393, -0.00000491, },
		{ -0.00000097, -0.00000089, -0.00001103, -0.00000489, -0.00001460, -0.00000875, -0.00005194, -0.00002600, 0.00006278, 0.00003806, 0.00001912, 0.00002182, 0.00002203, 0.00002266, 0.00004556, 0.00005759, -0.00004499, -0.00004059, -0.00000370, -0.00000801, -0.00000661, -0.00001316, 0.00000721, -0.00000088, -0.00000724, -0.00000089, -0.00000342, 0.00000394, -0.00001268, -0.00000220, 0.00000097, -0.00000287, -0.00000459, -0.00000545, 0.00000046, -0.00000113, -0.00000138, -0.00000471, -0.00000189, -0.00000433, 0.00000496, -0.00000218, -0.00000141, -0.00000507, -0.00000305, -0.00000073, -0.00000276, -0.00000703, 0.00001050, -0.00000271, -0.00000220, -0.00000593, 0.00000863, 0.00000970, -0.00000373, 0.00000396, -0.00000355, 0.00000457, 0.00000101, 0.00000366, -0.00000151, -0.00000125, -0.00000142, -0.00000189, -0.00000096, 0.00000069, 0.00000325, 0.00000148, -0.00000062, 0.00000050, 0.00000533, -0.00000396, 0.00000746, 0.00000935, -0.00000742, -0.00000031, 0.00000152, 0.00000499, -0.00000614, -0.00000353, -0.00000011, -0.00000233, -0.00000590, -0.00000305, 0.00000038, 0.00000457, -0.00000803, 0.00000110, -0.00000166, -0.00000584, -0.00000122, -0.00000016, -0.00000405, -0.00000517, -0.00000176, 0.00000273, -0.00001154, -0.00000393, 0.00000300, -0.00000622, -0.00000043, -0.00000871, 0.00000769, -0.00000140, -0.00001255, -0.00000510, -0.00000855, -0.00000671, -0.00004137, -0.00005811, 0.00006685, 0.00005368, 0.00002509, 0.00002653, 0.00002801, 0.00002185, 0.00006808, 0.00004234, -0.00006933, -0.00002970, -0.00001051, -0.00001283, -0.00000459, -0.00001147, 0.00000003, -0.00000113, -0.00000578, -0.00000567, },
		{ 0.00000166, -0.00001076, -0.00000494, -0.00001386, -0.00000772, -0.00004404, -0.00003618, 0.00005309, 0.00004409, 0.00001721, 0.00002197, 0.00001908, 0.00002344, 0.00003262, 0.00006475, -0.00001881, -0.00005255, -0.00000614, -0.00000749, -0.00000592, -0.00001381, 0.00000434, 0.00000227, -0.00000892, -0.00000037, -0.00000386, 0.00000428, -0.00001094, -0.00000497, 0.00000272, -0.00000390, -0.00000259, -0.00000749, 0.00000049, -0.00000146, 0.00000096, -0.00000596, -0.00000221, -0.00000426, 0.00000378, -0.00000076, -0.00000166, -0.00000361, -0.00000445, -0.00000117, -0.00000106, -0.00000829, 0.00000841, 0.00000124, -0.00000481, -0.00000355, 0.00000191, 0.00001505, -0.00000412, 0.00000380, -0.00000424, 0.00000442, 0.00000195, 0.00000305, -0.00000041, -0.00000209, -0.00000070, -0.00000237, -0.00000085, 0.00000038, 0.00000288, 0.00000222, -0.00000131, 0.00000027, 0.00000559, -0.00000179, 0.00000173, 0.00001313, -0.00000588, -0.00000198, 0.00000245, 0.00000334, -0.00000274, -0.00000605, 0.00000029, -0.00000146, -0.00000564, -0.00000428, -0.00000045, 0.00000507, -0.00000638, -0.00000185, 0.00000229, -0.00000802, -0.00000160, -0.00000072, -0.00000204, -0.00000599, -0.00000245, 0.00000292, -0.00000946, -0.00000738, 0.00000547, -0.00000790, 0.00000113, -0.00000989, 0.00000671, 0.00000059, -0.00001098, -0.00000769, -0.00000551, -0.00001019, -0.00002389, -0.00007480, 0.00003959, 0.00007317, 0.00002502, 0.00002944, 0.00002854, 0.00002328, 0.00006007, 0.00006205, -0.00006572, -0.00003923, -0.00000962, -0.00001422, -0.00000491, -0.00001048, -0.00000360, 0.00000146, -0.00000622, -0.00000494, -0.00000357, },
		{ -0.00000876, -0.00000595, -0.00001252, -0.00000774, -0.00003621, -0.00004318, 0.00004161, 0.00004913, 0.00001642, 0.00002168, 0.00001678, 0.00002364, 0.00002421, 0.00006130, 0.00000897, -0.00005596, -0.00001418, -0.00000508, -0.00000652, -0.00001298, 0.00000047, 0.00000548, -0.00001012, -0.00000056, -0.00000367, 0.00000393, -0.00000847, -0.00000805, 0.00000433, -0.00000464, -0.00000120, -0.00000820, -0.00000070, -0.00000126, 0.00000213, -0.00000493, -0.00000447, -0.00000291, 0.00000168, 0.00000118, -0.00000259, -0.00000159, -0.00000568, -0.00000168, -0.00000032, -0.00000753, 0.00000431, 0.00000606, -0.00000746, -0.00000102, -0.00000364, 0.00001696, -0.00000133, 0.00000207, -0.00000365, 0.00000260, 0.00000418, 0.00000171, 0.00000114, -0.00000313, -0.00000000, -0.00000267, -0.00000095, 0.00000031, 0.00000231, 0.00000297, -0.00000156, -0.00000047, 0.00000566, 0.00000068, -0.00000224, 0.00001295, -0.00000091, -0.00000574, 0.00000499, 0.00000076, 0.00000042, -0.00000733, -0.00000050, -0.00000033, -0.00000506, -0.00000536, -0.00000135, 0.00000484, -0.00000381, -0.00000520, 0.00000538, -0.00000822, -0.00000334, -0.00000079, -0.00000101, -0.00000541, -0.00000395, 0.00000316, -0.00000717, -0.00001063, 0.00000720, -0.00000852, 0.00000150, -0.00000973, 0.00000456, 0.00000286, -0.00000946, -0.00000961, -0.00000350, -0.00001259, -0.00001247, -0.00007426, 0.00000014, 0.00008860, 0.00003003, 0.00003081, 0.00002946, 0.00002606, 0.00005070, 0.00007898, -0.00005651, -0.00005051, -0.00000917, -0.00001521, -0.00000588, -0.00000895, -0.00000684, 0.00000254, -0.00000539, -0.00000484, -0.00000516, 0.00000227, },
		{ -0.00000710, -0.00001123, -0.00000813, -0.00002914, -0.00004702, 0.00002905, 0.00005272, 0.00001688, 0.00002076, 0.00001551, 0.00002252, 0.00002025, 0.00005151, 0.00003200, -0.00004876, -0.00002590, -0.00000290, -0.00000698, -0.00001191, -0.00000322, 0.00000767, -0.00000997, -0.00000201, -0.00000257, 0.00000276, -0.00000535, -0.00001109, 0.00000532, -0.00000457, -0.00000093, -0.00000740, -0.00000291, -0.00000065, 0.00000207, -0.00000210, -0.00000751, -0.00000159, -0.00000015, 0.00000267, -0.00000323, -0.00000009, -0.00000591, -0.00000267, -0.00000014, -0.00000576, -0.00000004, 0.00000965, -0.00000822, -0.00000026, -0.00000616, 0.00001456, 0.00000424, -0.00000057, -0.00000209, -0.00000048, 0.00000697, 0.00000041, 0.00000252, -0.00000387, 0.00000034, -0.00000266, -0.00000122, 0.00000028, 0.00000186, 0.00000324, -0.00000097, -0.00000191, 0.00000582, 0.00000256, -0.00000357, 0.00000915, 0.00000584, -0.00000941, 0.00000677, -0.00000033, 0.00000128, -0.00000608, -0.00000304, 0.00000141, -0.00000476, -0.00000564, -0.00000282, 0.00000445, -0.00000138, -0.00000749, 0.00000618, -0.00000569, -0.00000655, -0.00000018, -0.00000123, -0.00000349, -0.00000595, 0.00000322, -0.00000481, -0.00001319, 0.00000752, -0.00000744, 0.00000013, -0.00000774, 0.00000080, 0.00000599, -0.00000901, -0.00000947, -0.00000427, -0.00001129, -0.00000972, -0.00005915, -0.00004171, 0.00009123, 0.00004320, 0.00002971, 0.00003155, 0.00002909, 0.00004195, 0.00009102, -0.00004095, -0.00006319, -0.00000925, -0.00001620, -0.00000679, -0.00000801, -0.00000852, 0.00000142, -0.00000311, -0.00000550, -0.00000554, 0.00000114, -0.00000579, },
		{ -0.00001049, -0.00000845, -0.00002321, -0.00004804, 0.00001622, 0.00005439, 0.00001873, 0.00001901, 0.00001546, 0.00002001, 0.00001959, 0.00003960, 0.00004674, -0.00003283, -0.00003696, -0.00000385, -0.00000573, -0.00001175, -0.00000564, 0.00000802, -0.00000796, -0.00000490, -0.00000064, 0.00000095, -0.00000192, -0.00001361, 0.00000527, -0.00000347, -0.00000170, -0.00000559, -0.00000535, -0.00000024, 0.00000136, 0.00000127, -0.00000965, -0.00000173, -0.00000067, 0.00000286, -0.00000279, 0.00000011, -0.00000469, -0.00000430, -0.00000008, -0.00000409, -0.00000308, 0.00001065, -0.00000626, -0.00000184, -0.00000547, 0.00000886, 0.00001071, -0.00000261, -0.00000052, -0.00000374, 0.00000907, 0.00000017, 0.00000305, -0.00000388, 0.00000009, -0.00000224, -0.00000164, 0.00000020, 0.00000166, 0.00000299, 0.00000031, -0.00000363, 0.00000578, 0.00000384, -0.00000275, 0.00000367, 0.00001150, -0.00001032, 0.00000554, 0.00000183, -0.00000107, -0.00000229, -0.00000681, 0.00000333, -0.00000471, -0.00000509, -0.00000472, 0.00000404, 0.00000037, -0.00000805, 0.00000445, -0.00000107, -0.00001019, 0.00000038, -0.00000208, -0.00000116, -0.00000746, 0.00000251, -0.00000227, -0.00001488, 0.00000624, -0.00000469, -0.00000278, -0.00000426, -0.00000408, 0.00000959, -0.00000951, -0.00000736, -0.00000754, -0.00000677, -0.00001344, -0.00003748, -0.00007404, 0.00007459, 0.00006425, 0.00002740, 0.00003441, 0.00003194, 0.00003512, 0.00009708, -0.00001947, -0.00007605, -0.00001037, -0.00001728, -0.00000725, -0.00000818, -0.00000833, -0.00000155, -0.00000014, -0.00000634, -0.00000521, -0.00000084, -0.00000298, -0.00000748, },
		{ -0.00000868, -0.00001837, -0.00004693, 0.00000410, 0.00005359, 0.00002217, 0.00001623, 0.00001667, 0.00001638, 0.00002081, 0.00002855, 0.00005280, -0.00001275, -0.00004267, -0.00000968, -0.00000248, -0.00001266, -0.00000667, 0.00000659, -0.00000444, -0.00000860, 0.00000139, -0.00000081, 0.00000110, -0.00001497, 0.00000386, -0.00000139, -0.00000312, -0.00000354, -0.00000714, -0.00000059, 0.00000069, 0.00000398, -0.00000966, -0.00000398, 0.00000031, 0.00000173, -0.00000126, -0.00000090, -0.00000241, -0.00000612, -0.00000013, -0.00000301, -0.00000429, 0.00000910, -0.00000230, -0.00000473, -0.00000302, 0.00000220, 0.00001556, -0.00000245, 0.00000005, -0.00000606, 0.00000935, 0.00000162, 0.00000248, -0.00000303, -0.00000079, -0.00000147, -0.00000218, 0.00000013, 0.00000155, 0.00000250, 0.00000169, -0.00000484, 0.00000495, 0.00000502, -0.00000108, -0.00000104, 0.00001357, -0.00000690, 0.00000065, 0.00000700, -0.00000550, 0.00000232, -0.00001009, 0.00000421, -0.00000415, -0.00000445, -0.00000625, 0.00000313, 0.00000165, -0.00000729, 0.00000123, 0.00000387, -0.00001248, -0.00000037, -0.00000244, 0.00000028, -0.00000740, 0.00000047, 0.00000061, -0.00001576, 0.00000370, -0.00000097, -0.00000622, -0.00000048, -0.00000872, 0.00001236, -0.00000973, -0.00000486, -0.00001115, -0.00000194, -0.00001841, -0.00001880, -0.00008779, 0.00003833, 0.00008776, 0.00002774, 0.00003628, 0.00003521, 0.00003040, 0.00009748, 0.00000615, -0.00008702, -0.00001356, -0.00001803, -0.00000755, -0.00000913, -0.00000703, -0.00000510, 0.00000217, -0.00000639, -0.00000500, -0.00000261, -0.00000131, -0.00000665, -0.00001036, },
		{ -0.00001426, -0.00004456, -0.00000634, 0.00004990, 0.00002729, 0.00001249, 0.00001884, 0.00001224, 0.00002258, 0.00001995, 0.00005194, 0.00000655, -0.00004043, -0.00001949, 0.00000112, -0.00001339, -0.00000731, 0.00000442, -0.00000062, -0.00001179, 0.00000231, -0.00000154, 0.00000287, -0.00001458, 0.00000095, 0.00000137, -0.00000462, -0.00000193, -0.00000780, -0.00000185, 0.00000028, 0.00000543, -0.00000736, -0.00000772, 0.00000189, 0.00000020, 0.00000051, -0.00000218, -0.00000019, -0.00000717, -0.00000072, -0.00000234, -0.00000421, 0.00000626, 0.00000189, -0.00000713, -0.00000084, -0.00000295, 0.00001686, 0.00000064, -0.00000067, -0.00000688, 0.00000747, 0.00000462, 0.00000118, -0.00000168, -0.00000199, -0.00000055, -0.00000275, 0.00000012, 0.00000134, 0.00000215, 0.00000258, -0.00000493, 0.00000295, 0.00000650, 0.00000028, -0.00000330, 0.00001117, 0.00000025, -0.00000625, 0.00001283, -0.00000918, 0.00000500, -0.00001072, 0.00000266, -0.00000217, -0.00000470, -0.00000645, 0.00000110, 0.00000297, -0.00000615, -0.00000186, 0.00000713, -0.00001186, -0.00000328, -0.00000152, -0.00000003, -0.00000531, -0.00000288, 0.00000363, -0.00001589, 0.00000045, 0.00000270, -0.00000894, 0.00000220, -0.00001161, 0.00001283, -0.00000817, -0.00000391, -0.00001255, -0.00000008, -0.00001969, -0.00000972, -0.00008085, -0.00001005, 0.00010391, 0.00003581, 0.00003507, 0.00003987, 0.00002731, 0.00009351, 0.00003329, -0.00009368, -0.00001999, -0.00001780, -0.00000841, -0.00000995, -0.00000593, -0.00000764, 0.00000257, -0.00000496, -0.00000546, -0.00000353, -0.00000112, -0.00000483, -0.00001033, -0.00000916, },
		{ -0.00004168, -0.00001446, 0.00004335, 0.00003361, 0.00000845, 0.00002108, 0.00000861, 0.00002360, 0.00001457, 0.00004646, 0.00002176, -0.00003088, -0.00002980, 0.00000215, -0.00001211, -0.00000892, 0.00000285, 0.00000211, -0.00001310, 0.00000109, -0.00000056, 0.00000293, -0.00001231, -0.00000316, 0.00000428, -0.00000570, -0.00000110, -0.00000738, -0.00000356, -0.00000012, 0.00000577, -0.00000357, -0.00001147, 0.00000255, -0.00000047, 0.00000136, -0.00000259, 0.00000081, -0.00000667, -0.00000221, -0.00000169, -0.00000377, 0.00000367, 0.00000465, -0.00000766, -0.00000030, -0.00000524, 0.00001423, 0.00000593, -0.00000193, -0.00000657, 0.00000398, 0.00000816, 0.00000011, -0.00000053, -0.00000300, 0.00000018, -0.00000320, 0.00000016, 0.00000094, 0.00000213, 0.00000275, -0.00000381, 0.00000004, 0.00000813, 0.00000098, -0.00000293, 0.00000555, 0.00000844, -0.00001191, 0.00001598, -0.00000909, 0.00000352, -0.00000746, -0.00000173, 0.00000143, -0.00000620, -0.00000502, -0.00000207, 0.00000451, -0.00000536, -0.00000362, 0.00000760, -0.00000799, -0.00000800, 0.00000039, -0.00000184, -0.00000181, -0.00000661, 0.00000603, -0.00001502, -0.00000316, 0.00000557, -0.00001013, 0.00000296, -0.00001196, 0.00001043, -0.00000427, -0.00000541, -0.00001059, -0.00000251, -0.00001552, -0.00001108, -0.00005905, -0.00005724, 0.00010209, 0.00005482, 0.00003038, 0.00004593, 0.00002567, 0.00008652, 0.00005946, -0.00009410, -0.00003026, -0.00001619, -0.00001044, -0.00000977, -0.00000613, -0.00000814, 0.00000061, -0.00000217, -0.00000644, -0.00000369, -0.00000197, -0.00000287, -0.00000964, -0.00001030, -0.00001051, },
		{ -0.00002027, 0.00003469, 0.00003995, 0.00000540, 0.00002204, 0.00000679, 0.00002266, 0.00001269, 0.00003827, 0.00003194, -0.00001733, -0.00003616, -0.00000187, -0.00000784, -0.00001211, 0.00000255, 0.00000302, -0.00001199, -0.00000243, 0.00000195, 0.00000159, -0.00000870, -0.00000766, 0.00000660, -0.00000592, -0.00000108, -0.00000634, -0.00000506, -0.00000095, 0.00000552, 0.00000047, -0.00001362, 0.00000110, 0.00000051, 0.00000072, -0.00000158, 0.00000019, -0.00000462, -0.00000435, -0.00000099, -0.00000360, 0.00000222, 0.00000536, -0.00000620, -0.00000141, -0.00000494, 0.00000898, 0.00001142, -0.00000229, -0.00000603, 0.00000009, 0.00001085, 0.00000030, -0.00000017, -0.00000344, 0.00000047, -0.00000338, 0.00000013, 0.00000050, 0.00000228, 0.00000240, -0.00000202, -0.00000298, 0.00000922, 0.00000148, -0.00000106, -0.00000069, 0.00001426, -0.00001320, 0.00001402, -0.00000381, -0.00000232, -0.00000110, -0.00000774, 0.00000553, -0.00000816, -0.00000277, -0.00000545, 0.00000572, -0.00000486, -0.00000397, 0.00000563, -0.00000216, -0.00001287, 0.00000193, -0.00000385, 0.00000152, -0.00000917, 0.00000672, -0.00001267, -0.00000709, 0.00000750, -0.00000979, 0.00000203, -0.00001020, 0.00000587, 0.00000113, -0.00000856, -0.00000626, -0.00000778, -0.00000803, -0.00001826, -0.00003339, -0.00008924, 0.00007611, 0.00008269, 0.00002519, 0.00005150, 0.00002640, 0.00007713, 0.00008308, -0.00008751, -0.00004394, -0.00001359, -0.00001364, -0.00000842, -0.00000789, -0.00000670, -0.00000299, 0.00000092, -0.00000713, -0.00000371, -0.00000299, -0.00000173, -0.00000775, -0.00001216, -0.00000709, -0.00003859, },
		{ 0.00002523, 0.00004470, 0.00000487, 0.00002040, 0.00000776, 0.00001911, 0.00001411, 0.00002874, 0.00003778, -0.00000384, -0.00003545, -0.00001109, -0.00000155, -0.00001600, 0.00000291, 0.00000258, -0.00000909, -0.00000722, 0.00000479, -0.00000002, -0.00000495, -0.00001138, 0.00000752, -0.00000498, -0.00000174, -0.00000513, -0.00000588, -0.00000238, 0.00000499, 0.00000382, -0.00001333, -0.00000258, 0.00000279, -0.00000098, 0.00000034, -0.00000133, -0.00000205, -0.00000617, -0.00000074, -0.00000361, 0.00000175, 0.00000470, -0.00000384, -0.00000300, -0.00000351, 0.00000336, 0.00001490, -0.00000052, -0.00000603, -0.00000314, 0.00001163, 0.00000223, -0.00000070, -0.00000332, 0.00000034, -0.00000327, -0.00000008, 0.00000023, 0.00000228, 0.00000204, -0.00000031, -0.00000514, 0.00000900, 0.00000252, 0.00000078, -0.00000499, 0.00001525, -0.00000886, 0.00000700, 0.00000534, -0.00001016, 0.00000556, -0.00001273, 0.00000806, -0.00000896, -0.00000128, -0.00000753, 0.00000554, -0.00000398, -0.00000375, 0.00000277, 0.00000336, -0.00001562, 0.00000139, -0.00000451, 0.00000299, -0.00000919, 0.00000491, -0.00000860, -0.00001138, 0.00000871, -0.00000861, 0.00000040, -0.00000763, 0.00000084, 0.00000616, -0.00001137, -0.00000203, -0.00001262, -0.00000149, -0.00002446, -0.00001469, -0.00009751, 0.00002847, 0.00011070, 0.00002572, 0.00005315, 0.00003129, 0.00006523, 0.00010354, -0.00007438, -0.00005955, -0.00001114, -0.00001729, -0.00000653, -0.00001046, -0.00000443, -0.00000666, 0.00000283, -0.00000656, -0.00000432, -0.00000337, -0.00000197, -0.00000469, -0.00001428, -0.00000443, -0.00003514, -0.00002431, },
		{ 0.00004641, 0.00000792, 0.00001562, 0.00001155, 0.00001329, 0.00001779, 0.00001924, 0.00004027, 0.00000682, -0.00002756, -0.00002280, 0.00000384, -0.00001844, 0.00000241, 0.00000213, -0.00000589, -0.00001154, 0.00000632, -0.00000051, -0.00000231, -0.00001328, 0.00000649, -0.00000285, -0.00000285, -0.00000405, -0.00000599, -0.00000406, 0.00000409, 0.00000621, -0.00001085, -0.00000738, 0.00000500, -0.00000243, 0.00000188, -0.00000236, -0.00000042, -0.00000652, -0.00000154, -0.00000335, 0.00000149, 0.00000396, -0.00000210, -0.00000376, -0.00000245, -0.00000073, 0.00001503, 0.00000357, -0.00000653, -0.00000533, 0.00001031, 0.00000548, -0.00000149, -0.00000308, 0.00000007, -0.00000299, -0.00000051, 0.00000029, 0.00000190, 0.00000202, 0.00000075, -0.00000589, 0.00000717, 0.00000440, 0.00000159, -0.00000604, 0.00001118, -0.00000033, -0.00000235, 0.00001471, -0.00001603, 0.00000897, -0.00001396, 0.00000714, -0.00000721, -0.00000193, -0.00000713, 0.00000329, -0.00000211, -0.00000402, 0.00000072, 0.00000650, -0.00001469, -0.00000207, -0.00000306, 0.00000186, -0.00000633, 0.00000083, -0.00000328, -0.00001569, 0.00000935, -0.00000722, -0.00000098, -0.00000545, -0.00000315, 0.00000918, -0.00001206, -0.00000015, -0.00001420, 0.00000058, -0.00002478, -0.00000856, -0.00008288, -0.00002902, 0.00012558, 0.00003875, 0.00004797, 0.00004153, 0.00005104, 0.00012044, -0.00005563, -0.00007528, -0.00001017, -0.00002039, -0.00000513, -0.00001264, -0.00000284, -0.00000881, 0.00000247, -0.00000433, -0.00000569, -0.00000292, -0.00000343, -0.00000120, -0.00001573, -0.00000317, -0.00003091, -0.00002743, 0.00001639, },
		{ 0.00001444, 0.00000849, 0.00001692, 0.00000674, 0.00002183, 0.00001142, 0.00003983, 0.00001408, -0.00001531, -0.00003241, 0.00000493, -0.00001743, -0.00000033, 0.00000290, -0.00000378, -0.00001390, 0.00000533, 0.00000092, -0.00000143, -0.00001298, 0.00000360, 0.00000011, -0.00000408, -0.00000318, -0.00000572, -0.00000537, 0.00000247, 0.00000786, -0.00000716, -0.00001162, 0.00000552, -0.00000230, 0.00000185, -0.00000175, -0.00000071, -0.00000491, -0.00000348, -0.00000265, 0.00000079, 0.00000409, -0.00000178, -0.00000319, -0.00000234, -0.00000264, 0.00001209, 0.00000878, -0.00000657, -0.00000689, 0.00000763, 0.00000886, -0.00000151, -0.00000334, -0.00000001, -0.00000270, -0.00000111, 0.00000067, 0.00000114, 0.00000237, 0.00000107, -0.00000531, 0.00000417, 0.00000676, 0.00000138, -0.00000423, 0.00000403, 0.00000890, -0.00000991, 0.00002015, -0.00001643, 0.00000681, -0.00001029, 0.00000244, -0.00000282, -0.00000485, -0.00000434, -0.00000064, 0.00000053, -0.00000509, 0.00000026, 0.00000658, -0.00001019, -0.00000768, -0.00000024, -0.00000112, -0.00000177, -0.00000410, 0.00000194, -0.00001906, 0.00000903, -0.00000567, -0.00000191, -0.00000398, -0.00000563, 0.00000981, -0.00001027, -0.00000120, -0.00001188, -0.00000230, -0.00001888, -0.00001333, -0.00005510, -0.00007913, 0.00011505, 0.00006714, 0.00003652, 0.00005590, 0.00003636, 0.00013254, -0.00003173, -0.00008980, -0.00001150, -0.00002233, -0.00000498, -0.00001356, -0.00000290, -0.00000872, -0.00000018, -0.00000102, -0.00000724, -0.00000216, -0.00000526, 0.00000156, -0.00001554, -0.00000379, -0.00002569, -0.00003024, 0.00000919, 0.00004440, },
		{ 0.00000097, 0.00002164, 0.00000167, 0.00002403, 0.00000691, 0.00003632, 0.00001904, -0.00000289, -0.00003576, -0.00000008, -0.00001259, -0.00000528, 0.00000497, -0.00000312, -0.00001390, 0.00000181, 0.00000393, -0.00000192, -0.00001105, -0.00000032, 0.00000311, -0.00000498, -0.00000255, -0.00000548, -0.00000580, -0.00000003, 0.00000901, -0.00000332, -0.00001389, 0.00000348, -0.00000019, 0.00000002, 0.00000058, -0.00000270, -0.00000205, -0.00000570, -0.00000199, -0.00000028, 0.00000508, -0.00000249, -0.00000201, -0.00000259, -0.00000308, 0.00000767, 0.00001305, -0.00000488, -0.00000857, 0.00000466, 0.00001105, 0.00000006, -0.00000443, 0.00000022, -0.00000244, -0.00000183, 0.00000124, 0.00000025, 0.00000280, 0.00000103, -0.00000405, 0.00000093, 0.00000873, 0.00000091, -0.00000114, -0.00000315, 0.00001504, -0.00001224, 0.00001919, -0.00001032, -0.00000047, -0.00000315, -0.00000423, 0.00000254, -0.00000857, -0.00000072, -0.00000463, 0.00000278, -0.00000636, 0.00000090, 0.00000452, -0.00000395, -0.00001322, 0.00000203, -0.00000407, 0.00000234, -0.00000773, 0.00000527, -0.00002022, 0.00000697, -0.00000338, -0.00000307, -0.00000250, -0.00000740, 0.00000907, -0.00000723, -0.00000377, -0.00000756, -0.00000743, -0.00001062, -0.00002193, -0.00002800, -0.00010644, 0.00007429, 0.00010590, 0.00002448, 0.00006989, 0.00002509, 0.00013738, -0.00000231, -0.00010265, -0.00001509, -0.00002324, -0.00000608, -0.00001318, -0.00000453, -0.00000693, -0.00000392, 0.00000193, -0.00000782, -0.00000202, -0.00000636, 0.00000253, -0.00001316, -0.00000624, -0.00001970, -0.00003286, 0.00000387, 0.00003909, 0.00002310, },
		{ 0.00002348, -0.00000007, 0.00002287, 0.00000653, 0.00002988, 0.00002304, 0.00000642, -0.00003129, -0.00001002, -0.00000593, -0.00001073, 0.00000710, -0.00000303, -0.00001242, -0.00000296, 0.00000710, -0.00000248, -0.00000877, -0.00000399, 0.00000517, -0.00000507, -0.00000226, -0.00000541, -0.00000532, -0.00000303, 0.00000952, 0.00000014, -0.00001381, -0.00000071, 0.00000302, -0.00000259, 0.00000347, -0.00000491, 0.00000037, -0.00000673, -0.00000229, -0.00000107, 0.00000605, -0.00000296, -0.00000162, -0.00000208, -0.00000331, 0.00000372, 0.00001462, -0.00000074, -0.00001065, 0.00000216, 0.00001134, 0.00000330, -0.00000597, 0.00000041, -0.00000202, -0.00000268, 0.00000182, -0.00000044, 0.00000292, 0.00000110, -0.00000283, -0.00000166, 0.00000952, 0.00000103, 0.00000154, -0.00000776, 0.00001581, -0.00000842, 0.00001244, 0.00000012, -0.00000955, 0.00000387, -0.00000966, 0.00000623, -0.00001086, 0.00000151, -0.00000668, 0.00000324, -0.00000673, 0.00000137, 0.00000226, 0.00000140, -0.00001612, 0.00000173, -0.00000509, 0.00000402, -0.00000836, 0.00000550, -0.00001837, 0.00000262, 0.00000029, -0.00000534, 0.00000004, -0.00000965, 0.00000851, -0.00000480, -0.00000562, -0.00000418, -0.00001088, -0.00000516, -0.00002658, -0.00001270, -0.00010432, 0.00001022, 0.00014164, 0.00002158, 0.00007711, 0.00002203, 0.00013224, 0.00003310, -0.00011368, -0.00002046, -0.00002376, -0.00000784, -0.00001222, -0.00000675, -0.00000487, -0.00000695, 0.00000290, -0.00000643, -0.00000318, -0.00000604, 0.00000132, -0.00000885, -0.00000982, -0.00001368, -0.00003493, -0.00000003, 0.00003176, 0.00003177, -0.00000452, },
		{ 0.00000206, 0.00001829, 0.00000961, 0.00002166, 0.00002639, 0.00001169, -0.00002079, -0.00002096, -0.00000102, -0.00001411, 0.00000744, -0.00000208, -0.00001095, -0.00000722, 0.00000872, -0.00000178, -0.00000733, -0.00000633, 0.00000562, -0.00000409, -0.00000236, -0.00000539, -0.00000442, -0.00000571, 0.00000887, 0.00000328, -0.00001197, -0.00000557, 0.00000558, -0.00000431, 0.00000518, -0.00000551, 0.00000060, -0.00000535, -0.00000421, -0.00000098, 0.00000606, -0.00000195, -0.00000306, -0.00000018, -0.00000410, 0.00000140, 0.00001294, 0.00000526, -0.00001251, 0.00000005, 0.00000997, 0.00000731, -0.00000697, -0.00000012, -0.00000107, -0.00000375, 0.00000228, -0.00000072, 0.00000252, 0.00000152, -0.00000203, -0.00000317, 0.00000892, 0.00000206, 0.00000289, -0.00000877, 0.00001145, -0.00000046, 0.00000333, 0.00001043, -0.00001586, 0.00000692, -0.00001097, 0.00000622, -0.00001009, 0.00000084, -0.00000563, 0.00000126, -0.00000562, 0.00000065, 0.00000142, 0.00000395, -0.00001499, -0.00000186, -0.00000355, 0.00000265, -0.00000574, 0.00000276, -0.00001385, -0.00000370, 0.00000526, -0.00000895, 0.00000394, -0.00001278, 0.00000885, -0.00000399, -0.00000543, -0.00000361, -0.00001026, -0.00000541, -0.00002345, -0.00001256, -0.00007808, -0.00005957, 0.00015667, 0.00003706, 0.00007285, 0.00003016, 0.00011621, 0.00007353, -0.00012181, -0.00002762, -0.00002419, -0.00000978, -0.00001146, -0.00000849, -0.00000393, -0.00000787, 0.00000110, -0.00000301, -0.00000551, -0.00000444, -0.00000155, -0.00000367, -0.00001328, -0.00000862, -0.00003582, -0.00000332, 0.00002401, 0.00003833, -0.00000613, 0.00002125, },
		{ 0.00001185, 0.00001411, 0.00001379, 0.00002817, 0.00001422, -0.00000831, -0.00002814, -0.00000097, -0.00001369, 0.00000500, 0.00000052, -0.00001045, -0.00000978, 0.00000787, 0.00000071, -0.00000714, -0.00000709, 0.00000452, -0.00000229, -0.00000274, -0.00000523, -0.00000373, -0.00000720, 0.00000664, 0.00000633, -0.00000940, -0.00000937, 0.00000589, -0.00000383, 0.00000453, -0.00000344, -0.00000200, -0.00000159, -0.00000734, -0.00000021, 0.00000490, 0.00000075, -0.00000609, 0.00000249, -0.00000509, 0.00000053, 0.00000905, 0.00001132, -0.00001279, -0.00000237, 0.00000793, 0.00001063, -0.00000625, -0.00000199, 0.00000062, -0.00000501, 0.00000252, -0.00000054, 0.00000167, 0.00000218, -0.00000156, -0.00000380, 0.00000736, 0.00000363, 0.00000303, -0.00000689, 0.00000431, 0.00000781, -0.00000368, 0.00001627, -0.00001601, 0.00000410, -0.00000749, 0.00000253, -0.00000646, -0.00000248, -0.00000208, -0.00000226, -0.00000361, -0.00000124, 0.00000232, 0.00000351, -0.00001041, -0.00000753, -0.00000061, -0.00000060, -0.00000140, -0.00000126, -0.00000828, -0.00001053, 0.00001034, -0.00001285, 0.00000811, -0.00001573, 0.00000924, -0.00000403, -0.00000387, -0.00000532, -0.00000628, -0.00001011, -0.00001457, -0.00002230, -0.00004243, -0.00011328, 0.00013672, 0.00007364, 0.00005794, 0.00004807, 0.00009204, 0.00011517, -0.00012406, -0.00003790, -0.00002391, -0.00001216, -0.00001102, -0.00000939, -0.00000454, -0.00000659, -0.00000282, 0.00000126, -0.00000787, -0.00000255, -0.00000482, 0.00000088, -0.00001533, -0.00000537, -0.00003506, -0.00000670, 0.00001715, 0.00004145, -0.00000328, 0.00001538, 0.00000693, },
		{ 0.00001755, 0.00000847, 0.00002717, 0.00001607, 0.00000214, -0.00002848, -0.00000645, -0.00001005, 0.00000054, 0.00000410, -0.00001060, -0.00001076, 0.00000500, 0.00000419, -0.00000746, -0.00000698, 0.00000268, -0.00000039, -0.00000306, -0.00000488, -0.00000365, -0.00000716, 0.00000297, 0.00000919, -0.00000682, -0.00001107, 0.00000347, -0.00000114, 0.00000169, 0.00000080, -0.00000638, 0.00000296, -0.00001009, 0.00000016, 0.00000334, 0.00000415, -0.00000924, 0.00000430, -0.00000505, -0.00000006, 0.00000489, 0.00001522, -0.00001020, -0.00000576, 0.00000628, 0.00001204, -0.00000318, -0.00000518, 0.00000277, -0.00000617, 0.00000236, 0.00000007, 0.00000064, 0.00000271, -0.00000107, -0.00000407, 0.00000555, 0.00000504, 0.00000270, -0.00000374, -0.00000255, 0.00001276, -0.00000548, 0.00001566, -0.00000973, -0.00000322, -0.00000145, -0.00000248, -0.00000212, -0.00000644, 0.00000179, -0.00000520, -0.00000224, -0.00000322, 0.00000392, 0.00000159, -0.00000472, -0.00001264, 0.00000141, -0.00000343, 0.00000220, -0.00000415, -0.00000384, -0.00001587, 0.00001367, -0.00001513, 0.00001042, -0.00001637, 0.00000764, -0.00000287, -0.00000312, -0.00000678, -0.00000222, -0.00001483, -0.00000606, -0.00003194, -0.00001429, -0.00013421, 0.00007889, 0.00012317, 0.00004056, 0.00006919, 0.00006639, 0.00015152, -0.00011580, -0.00005367, -0.00002144, -0.00001598, -0.00001016, -0.00001013, -0.00000586, -0.00000444, -0.00000693, 0.00000433, -0.00000864, -0.00000170, -0.00000705, 0.00000344, -0.00001512, -0.00000424, -0.00003259, -0.00001050, 0.00001186, 0.00004086, 0.00000319, 0.00000764, 0.00001231, 0.00000591, },
		{ 0.00000673, 0.00002320, 0.00001831, 0.00000868, -0.00002222, -0.00001490, -0.00000605, -0.00000353, 0.00000682, -0.00001004, -0.00001140, 0.00000166, 0.00000700, -0.00000689, -0.00000720, 0.00000125, 0.00000077, -0.00000286, -0.00000447, -0.00000412, -0.00000592, -0.00000129, 0.00001121, -0.00000427, -0.00001079, -0.00000077, 0.00000238, -0.00000179, 0.00000537, -0.00001025, 0.00000586, -0.00001046, -0.00000122, 0.00000254, 0.00000673, -0.00001060, 0.00000357, -0.00000291, -0.00000146, 0.00000217, 0.00001554, -0.00000447, -0.00001000, 0.00000538, 0.00001122, 0.00000181, -0.00000880, 0.00000454, -0.00000664, 0.00000152, 0.00000103, -0.00000030, 0.00000282, -0.00000025, -0.00000438, 0.00000403, 0.00000584, 0.00000252, -0.00000084, -0.00000689, 0.00001271, -0.00000185, 0.00001005, -0.00000017, -0.00001092, 0.00000312, -0.00000535, 0.00000011, -0.00000853, 0.00000349, -0.00000546, -0.00000297, -0.00000407, 0.00000473, 0.00000028, -0.00000062, -0.00001460, 0.00000047, -0.00000394, 0.00000309, -0.00000421, -0.00000202, -0.00001826, 0.00001372, -0.00001410, 0.00000898, -0.00001283, 0.00000236, 0.00000131, -0.00000526, -0.00000545, -0.00000153, -0.00001499, -0.00000383, -0.00003297, -0.00000463, -0.00011846, -0.00000467, 0.00016721, 0.00003378, 0.00008439, 0.00004748, 0.00017543, -0.00009257, -0.00007686, -0.00001564, -0.00002222, -0.00000795, -0.00001184, -0.00000642, -0.00000335, -0.00000900, 0.00000430, -0.00000665, -0.00000276, -0.00000736, 0.00000342, -0.00001261, -0.00000493, -0.00002878, -0.00001462, 0.00000819, 0.00003718, 0.00001155, 0.00000032, 0.00001602, 0.00000234, 0.00001828, },
		{ 0.00001765, 0.00002028, 0.00001212, -0.00001268, -0.00002173, -0.00000536, -0.00000470, 0.00000683, -0.00000741, -0.00001300, -0.00000058, 0.00000766, -0.00000438, -0.00000859, 0.00000092, 0.00000076, -0.00000201, -0.00000416, -0.00000478, -0.00000432, -0.00000493, 0.00001153, -0.00000136, -0.00000945, -0.00000504, 0.00000474, -0.00000392, 0.00000810, -0.00001134, 0.00000514, -0.00000727, -0.00000490, 0.00000312, 0.00000744, -0.00000906, -0.00000027, 0.00000129, -0.00000377, 0.00000138, 0.00001243, 0.00000318, -0.00001391, 0.00000466, 0.00000892, 0.00000729, -0.00001134, 0.00000486, -0.00000579, -0.00000020, 0.00000231, -0.00000101, 0.00000244, 0.00000084, -0.00000471, 0.00000278, 0.00000611, 0.00000261, 0.00000118, -0.00000816, 0.00000849, 0.00000459, 0.00000341, 0.00000783, -0.00001434, 0.00000265, -0.00000372, -0.00000134, -0.00000745, 0.00000186, -0.00000237, -0.00000593, -0.00000353, 0.00000394, 0.00000089, 0.00000040, -0.00001248, -0.00000375, -0.00000186, 0.00000104, -0.00000146, -0.00000270, -0.00001767, 0.00001032, -0.00000941, 0.00000341, -0.00000501, -0.00000648, 0.00000843, -0.00001041, -0.00000079, -0.00000532, -0.00000911, -0.00000961, -0.00002331, -0.00001363, -0.00007680, -0.00008986, 0.00018348, 0.00004912, 0.00008700, 0.00004115, 0.00018248, -0.00005281, -0.00010679, -0.00000726, -0.00003052, -0.00000430, -0.00001509, -0.00000528, -0.00000460, -0.00000783, 0.00000063, -0.00000201, -0.00000546, -0.00000591, 0.00000124, -0.00000861, -0.00000657, -0.00002431, -0.00001874, 0.00000581, 0.00003147, 0.00002000, -0.00000482, 0.00001691, 0.00000161, 0.00001636, 0.00000782, },
		{ 0.00002021, 0.00001482, -0.00000418, -0.00002285, -0.00001008, -0.00000215, 0.00000384, -0.00000260, -0.00001580, -0.00000122, 0.00000590, -0.00000013, -0.00001082, 0.00000137, 0.00000000, -0.00000088, -0.00000386, -0.00000530, -0.00000317, -0.00000694, 0.00000967, 0.00000214, -0.00000803, -0.00000773, 0.00000449, -0.00000352, 0.00000781, -0.00000867, 0.00000051, -0.00000110, -0.00000993, 0.00000446, 0.00000649, -0.00000502, -0.00000610, 0.00000600, -0.00000585, 0.00000167, 0.00000753, 0.00001041, -0.00001571, 0.00000304, 0.00000654, 0.00001143, -0.00001134, 0.00000295, -0.00000337, -0.00000275, 0.00000373, -0.00000145, 0.00000178, 0.00000183, -0.00000467, 0.00000141, 0.00000634, 0.00000260, 0.00000253, -0.00000728, 0.00000268, 0.00001000, 0.00000000, 0.00001039, -0.00001096, -0.00000361, 0.00000194, -0.00000551, -0.00000418, -0.00000207, 0.00000263, -0.00000949, -0.00000271, 0.00000211, 0.00000312, -0.00000112, -0.00000756, -0.00000941, 0.00000108, -0.00000225, 0.00000221, -0.00000412, -0.00001556, 0.00000480, -0.00000244, -0.00000465, 0.00000496, -0.00001640, 0.00001598, -0.00001614, 0.00000494, -0.00001133, 0.00000011, -0.00001974, -0.00000813, -0.00003166, -0.00002960, -0.00015001, 0.00015603, 0.00008895, 0.00007771, 0.00004735, 0.00017359, 0.00000018, -0.00013888, 0.00000009, -0.00003848, -0.00000071, -0.00001900, -0.00000289, -0.00000787, -0.00000405, -0.00000538, 0.00000354, -0.00000833, -0.00000389, -0.00000180, -0.00000448, -0.00000810, -0.00001988, -0.00002249, 0.00000429, 0.00002476, 0.00002721, -0.00000708, 0.00001531, 0.00000269, 0.00001333, 0.00000961, 0.00001295, },
		{ 0.00001863, 0.00000055, -0.00001715, -0.00001903, 0.00000217, -0.00000029, 0.00000274, -0.00001851, -0.00000131, 0.00000302, 0.00000430, -0.00001246, 0.00000139, -0.00000043, -0.00000027, -0.00000319, -0.00000563, -0.00000281, -0.00000713, 0.00000599, 0.00000584, -0.00000689, -0.00000824, 0.00000157, -0.00000092, 0.00000502, -0.00000322, -0.00000624, 0.00000565, -0.00001404, 0.00000492, 0.00000520, -0.00000021, -0.00001155, 0.00000886, -0.00000601, 0.00000155, 0.00000303, 0.00001487, -0.00001401, -0.00000024, 0.00000520, 0.00001297, -0.00000824, -0.00000104, 0.00000015, -0.00000561, 0.00000492, -0.00000156, 0.00000109, 0.00000231, -0.00000387, -0.00000049, 0.00000695, 0.00000214, 0.00000367, -0.00000567, -0.00000206, 0.00001144, 0.00000195, 0.00000680, -0.00000203, -0.00001276, 0.00000814, -0.00000910, -0.00000157, -0.00000558, 0.00000663, -0.00001100, -0.00000351, 0.00000070, 0.00000538, -0.00000308, -0.00000269, -0.00001350, 0.00000222, -0.00000422, 0.00000422, -0.00000399, -0.00001396, -0.00000072, 0.00000438, -0.00001227, 0.00001356, -0.00002357, 0.00002019, -0.00001873, 0.00000793, -0.00001534, 0.00000740, -0.00002743, 0.00000366, -0.00004536, 0.00000291, -0.00016735, 0.00008419, 0.00014204, 0.00006613, 0.00005964, 0.00015512, 0.00005853, -0.00016528, 0.00000074, -0.00004229, 0.00000008, -0.00002149, -0.00000109, -0.00001126, 0.00000004, -0.00001098, 0.00000737, -0.00000938, -0.00000284, -0.00000420, -0.00000154, -0.00000870, -0.00001593, -0.00002561, 0.00000324, 0.00001788, 0.00003250, -0.00000678, 0.00001249, 0.00000382, 0.00001111, 0.00000997, 0.00001110, 0.00001681, },
		{ 0.00000186, -0.00000727, -0.00002788, 0.00000428, -0.00000243, 0.00000604, -0.00001902, -0.00000271, 0.00000108, 0.00000678, -0.00001177, -0.00000037, 0.00000056, -0.00000091, -0.00000184, -0.00000592, -0.00000303, -0.00000609, 0.00000168, 0.00000872, -0.00000565, -0.00000715, -0.00000265, 0.00000215, 0.00000162, 0.00000263, -0.00001217, 0.00000991, -0.00001479, 0.00000289, 0.00000504, 0.00000339, -0.00001432, 0.00000806, -0.00000328, 0.00000009, 0.00000045, 0.00001536, -0.00000875, -0.00000490, 0.00000510, 0.00001183, -0.00000276, -0.00000588, 0.00000360, -0.00000787, 0.00000537, -0.00000121, 0.00000052, 0.00000213, -0.00000227, -0.00000287, 0.00000787, 0.00000135, 0.00000474, -0.00000425, -0.00000430, 0.00000834, 0.00000815, -0.00000007, 0.00000793, -0.00001958, 0.00001019, -0.00000842, -0.00000253, -0.00000604, 0.00000704, -0.00000844, -0.00000717, 0.00000084, 0.00000610, -0.00000332, -0.00000049, -0.00001360, -0.00000029, -0.00000317, 0.00000296, -0.00000107, -0.00001402, -0.00000474, 0.00000900, -0.00001680, 0.00001768, -0.00002486, 0.00001816, -0.00001541, 0.00000531, -0.00001391, 0.00000830, -0.00002707, 0.00000511, -0.00004535, 0.00000995, -0.00014096, -0.00001453, 0.00018568, 0.00006683, 0.00006881, 0.00013576, 0.00011272, -0.00017759, -0.00001064, -0.00003863, -0.00000435, -0.00002042, -0.00000205, -0.00001242, 0.00000177, -0.00001339, 0.00000728, -0.00000725, -0.00000371, -0.00000502, -0.00000044, -0.00000814, -0.00001257, -0.00002798, 0.00000229, 0.00001145, 0.00003564, -0.00000465, 0.00000975, 0.00000378, 0.00001065, 0.00000813, 0.00001234, 0.00001069, 0.00002322, },
		{ 0.00000185, -0.00003136, 0.00000045, -0.00000014, 0.00000545, -0.00001590, -0.00000676, 0.00000160, 0.00000596, -0.00000794, -0.00000437, 0.00000319, -0.00000282, 0.00000010, -0.00000621, -0.00000344, -0.00000481, -0.00000187, 0.00000973, -0.00000368, -0.00000563, -0.00000623, 0.00000363, -0.00000034, 0.00000652, -0.00001469, 0.00000957, -0.00001108, -0.00000201, 0.00000648, 0.00000482, -0.00001348, 0.00000345, 0.00000173, -0.00000230, -0.00000014, 0.00001241, -0.00000144, -0.00000950, 0.00000540, 0.00000909, 0.00000334, -0.00000976, 0.00000558, -0.00000861, 0.00000460, -0.00000024, -0.00000001, 0.00000147, -0.00000031, -0.00000515, 0.00000848, 0.00000088, 0.00000533, -0.00000304, -0.00000434, 0.00000256, 0.00001500, -0.00000541, 0.00001353, -0.00001936, 0.00000500, -0.00000183, -0.00000789, -0.00000281, 0.00000339, -0.00000194, -0.00001302, 0.00000221, 0.00000492, -0.00000105, -0.00000180, -0.00000949, -0.00000606, 0.00000043, -0.00000103, 0.00000387, -0.00001512, -0.00000724, 0.00001082, -0.00001725, 0.00001626, -0.00001962, 0.00000978, -0.00000634, -0.00000275, -0.00000686, 0.00000214, -0.00001781, -0.00000452, -0.00003118, -0.00000553, -0.00008696, -0.00011023, 0.00019466, 0.00009114, 0.00006909, 0.00012162, 0.00015583, -0.00017053, -0.00003609, -0.00002700, -0.00001435, -0.00001518, -0.00000672, -0.00001009, -0.00000033, -0.00001137, 0.00000277, -0.00000218, -0.00000618, -0.00000441, -0.00000092, -0.00000687, -0.00000955, -0.00002954, 0.00000109, 0.00000604, 0.00003653, -0.00000128, 0.00000765, 0.00000263, 0.00001133, 0.00000524, 0.00001480, 0.00000448, 0.00002620, 0.00000273, },
		{ -0.00002622, -0.00001025, 0.00000635, 0.00000142, -0.00000959, -0.00001309, 0.00000437, 0.00000229, -0.00000187, -0.00000951, 0.00000643, -0.00000500, 0.00000188, -0.00000623, -0.00000380, -0.00000399, -0.00000378, 0.00000849, -0.00000082, -0.00000457, -0.00000778, 0.00000239, 0.00000010, 0.00000738, -0.00001298, 0.00000469, -0.00000401, -0.00000831, 0.00000850, 0.00000465, -0.00000991, -0.00000319, 0.00000687, -0.00000408, 0.00000010, 0.00000787, 0.00000546, -0.00001217, 0.00000485, 0.00000634, 0.00000807, -0.00001109, 0.00000519, -0.00000744, 0.00000255, 0.00000129, -0.00000058, 0.00000075, 0.00000131, -0.00000650, 0.00000795, 0.00000149, 0.00000483, -0.00000152, -0.00000356, -0.00000289, 0.00001845, -0.00000515, 0.00001154, -0.00001075, -0.00000672, 0.00000866, -0.00001531, 0.00000198, -0.00000220, 0.00000587, -0.00001841, 0.00000291, 0.00000308, 0.00000265, -0.00000517, -0.00000337, -0.00001243, 0.00000408, -0.00000526, 0.00000826, -0.00001514, -0.00000965, 0.00001086, -0.00001466, 0.00001075, -0.00001009, -0.00000196, 0.00000511, -0.00001277, 0.00000240, -0.00000748, -0.00000406, -0.00001940, -0.00001086, -0.00002995, -0.00003061, -0.00017282, 0.00015361, 0.00013846, 0.00006339, 0.00011255, 0.00018645, -0.00014435, -0.00007260, -0.00001084, -0.00002727, -0.00000758, -0.00001391, -0.00000508, -0.00000552, -0.00000601, -0.00000440, 0.00000380, -0.00000865, -0.00000350, -0.00000192, -0.00000574, -0.00000657, -0.00003022, -0.00000073, 0.00000213, 0.00003503, 0.00000319, 0.00000580, 0.00000162, 0.00001126, 0.00000356, 0.00001562, 0.00000145, 0.00002464, 0.00000653, 0.00000581, },
		{ -0.00002471, 0.00001365, -0.00000321, -0.00000260, -0.00001949, 0.00000741, -0.00000204, 0.00000400, -0.00001353, 0.00000834, -0.00000588, 0.00000239, -0.00000540, -0.00000425, -0.00000375, -0.00000412, 0.00000563, 0.00000231, -0.00000400, -0.00000716, -0.00000115, 0.00000220, 0.00000600, -0.00000834, -0.00000246, 0.00000348, -0.00001328, 0.00000911, 0.00000451, -0.00000571, -0.00000913, 0.00000959, -0.00000357, -0.00000022, 0.00000381, 0.00000979, -0.00001164, 0.00000271, 0.00000471, 0.00001020, -0.00000933, 0.00000259, -0.00000481, -0.00000029, 0.00000306, -0.00000118, 0.00000033, 0.00000203, -0.00000638, 0.00000582, 0.00000352, 0.00000307, 0.00000069, -0.00000327, -0.00000565, 0.00001628, 0.00000171, 0.00000279, 0.00000318, -0.00002021, 0.00001798, -0.00002024, 0.00000444, -0.00000605, 0.00001120, -0.00001999, 0.00000062, 0.00000246, 0.00000561, -0.00000786, 0.00000134, -0.00001592, 0.00000470, -0.00000679, 0.00000932, -0.00001192, -0.00001357, 0.00001071, -0.00001109, 0.00000405, -0.00000013, -0.00001252, 0.00001415, -0.00001988, 0.00000887, -0.00001497, 0.00000728, -0.00003083, 0.00000447, -0.00004736, 0.00000572, -0.00018524, 0.00006633, 0.00019255, 0.00006381, 0.00010249, 0.00020836, -0.00010425, -0.00011299, 0.00000343, -0.00003840, -0.00000112, -0.00002079, 0.00000003, -0.00001124, -0.00000021, -0.00001104, 0.00000784, -0.00000911, -0.00000364, -0.00000233, -0.00000548, -0.00000349, -0.00002980, -0.00000352, 0.00000008, 0.00003109, 0.00000884, 0.00000334, 0.00000234, 0.00000864, 0.00000502, 0.00001271, 0.00000355, 0.00001735, 0.00001452, 0.00000343, -0.00001347, },
		{ 0.00001669, -0.00000464, 0.00000190, -0.00002314, 0.00000808, -0.00000427, 0.00000697, -0.00001420, 0.00000721, -0.00000419, 0.00000084, -0.00000335, -0.00000505, -0.00000371, -0.00000371, 0.00000252, 0.00000450, -0.00000328, -0.00000530, -0.00000527, 0.00000406, 0.00000434, -0.00000329, -0.00000867, 0.00000811, -0.00001442, 0.00000672, 0.00000580, -0.00000278, -0.00001217, 0.00000837, -0.00000024, -0.00000158, 0.00000129, 0.00001072, -0.00000808, -0.00000062, 0.00000429, 0.00000980, -0.00000532, -0.00000098, -0.00000188, -0.00000301, 0.00000449, -0.00000161, 0.00000026, 0.00000173, -0.00000484, 0.00000241, 0.00000648, 0.00000059, 0.00000329, -0.00000385, -0.00000516, 0.00000915, 0.00001254, -0.00000814, 0.00001624, -0.00002890, 0.00002052, -0.00001848, 0.00000140, -0.00000533, 0.00001133, -0.00001581, -0.00000569, 0.00000406, 0.00000623, -0.00000765, 0.00000211, -0.00001433, 0.00000070, -0.00000424, 0.00000592, -0.00000491, -0.00001925, 0.00001109, -0.00000808, -0.00000139, 0.00000699, -0.00001833, 0.00001729, -0.00002072, 0.00000914, -0.00001632, 0.00001111, -0.00003260, 0.00000740, -0.00004835, 0.00001260, -0.00015095, -0.00004313, 0.00022606, 0.00008564, 0.00008443, 0.00022656, -0.00005692, -0.00014920, 0.00000946, -0.00004345, 0.00000094, -0.00002448, 0.00000246, -0.00001460, 0.00000300, -0.00001424, 0.00000779, -0.00000646, -0.00000545, -0.00000167, -0.00000621, -0.00000066, -0.00002795, -0.00000750, -0.00000002, 0.00002491, 0.00001561, -0.00000027, 0.00000558, 0.00000295, 0.00000975, 0.00000620, 0.00001003, 0.00000609, 0.00002463, -0.00000248, 0.00000188, -0.00003686, },
		{ -0.00000054, 0.00000209, -0.00002238, 0.00000480, -0.00000278, 0.00000565, -0.00001070, 0.00000271, 0.00000005, -0.00000251, -0.00000037, -0.00000616, -0.00000345, -0.00000353, 0.00000049, 0.00000479, -0.00000177, -0.00000346, -0.00000808, 0.00000383, 0.00000420, -0.00000001, -0.00001152, 0.00000803, -0.00001100, 0.00000138, 0.00000855, -0.00000167, -0.00001187, 0.00000372, 0.00000464, -0.00000308, 0.00000002, 0.00000906, -0.00000310, -0.00000374, 0.00000420, 0.00000801, -0.00000083, -0.00000381, 0.00000007, -0.00000473, 0.00000507, -0.00000163, 0.00000037, 0.00000080, -0.00000257, -0.00000133, 0.00000925, -0.00000148, 0.00000538, -0.00000469, -0.00000264, 0.00000007, 0.00002250, -0.00001536, 0.00002230, -0.00002786, 0.00001345, -0.00000893, -0.00000730, -0.00000006, 0.00000645, -0.00000682, -0.00001454, 0.00000691, 0.00000468, -0.00000431, -0.00000120, -0.00000823, -0.00000672, 0.00000109, -0.00000044, 0.00000394, -0.00002489, 0.00001099, -0.00000554, -0.00000493, 0.00001031, -0.00001867, 0.00001432, -0.00001551, 0.00000358, -0.00001161, 0.00000717, -0.00002458, -0.00000161, -0.00003440, -0.00000357, -0.00009163, -0.00013945, 0.00021250, 0.00013714, 0.00005755, 0.00024200, -0.00000618, -0.00017608, 0.00000417, -0.00004109, -0.00000228, -0.00002397, 0.00000099, -0.00001424, 0.00000223, -0.00001301, 0.00000358, -0.00000139, -0.00000812, -0.00000055, -0.00000721, 0.00000113, -0.00002428, -0.00001265, 0.00000154, 0.00001707, 0.00002296, -0.00000475, 0.00001073, -0.00000444, 0.00001579, -0.00000144, 0.00001760, -0.00000492, 0.00003242, -0.00000674, 0.00001329, -0.00004073, 0.00001165, },
		{ -0.00000116, -0.00001788, -0.00000186, 0.00000192, 0.00000090, -0.00000438, -0.00000359, 0.00000529, -0.00000622, 0.00000244, -0.00000704, -0.00000299, -0.00000399, 0.00000003, 0.00000317, 0.00000044, -0.00000225, -0.00000859, 0.00000092, 0.00000595, 0.00000093, -0.00001065, 0.00000385, -0.00000470, -0.00000488, 0.00001120, -0.00000129, -0.00000966, -0.00000208, 0.00000856, -0.00000301, -0.00000118, 0.00000654, 0.00000113, -0.00000508, 0.00000342, 0.00000625, 0.00000242, -0.00000459, 0.00000035, -0.00000514, 0.00000469, -0.00000117, 0.00000041, -0.00000022, -0.00000048, -0.00000425, 0.00001064, -0.00000203, 0.00000605, -0.00000479, -0.00000007, -0.00000733, 0.00002714, -0.00001472, 0.00001855, -0.00001664, -0.00000131, 0.00000513, -0.00001812, 0.00000654, -0.00000026, 0.00000330, -0.00002226, 0.00000833, 0.00000285, 0.00000031, -0.00000624, -0.00000077, -0.00001402, 0.00000578, -0.00000626, 0.00001102, -0.00002739, 0.00000825, -0.00000200, -0.00000779, 0.00001131, -0.00001571, 0.00000826, -0.00000785, -0.00000392, -0.00000482, -0.00000015, -0.00001235, -0.00001509, -0.00001578, -0.00002606, -0.00003618, -0.00019137, 0.00014046, 0.00021049, 0.00003199, 0.00024879, 0.00004923, -0.00019333, -0.00001081, -0.00003396, -0.00000857, -0.00002084, -0.00000324, -0.00001114, -0.00000147, -0.00000888, -0.00000260, 0.00000369, -0.00000975, -0.00000037, -0.00000724, 0.00000097, -0.00001860, -0.00001866, 0.00000414, 0.00000856, 0.00002983, -0.00000895, 0.00001593, -0.00001083, 0.00001999, -0.00000671, 0.00002219, -0.00001112, 0.00003359, -0.00000493, 0.00001627, -0.00003378, -0.00000146, 0.00000803, },
		{ -0.00001237, -0.00000937, 0.00000734, -0.00000457, 0.00000184, -0.00000893, 0.00000909, -0.00000828, 0.00000366, -0.00000686, -0.00000277, -0.00000477, 0.00000059, 0.00000068, 0.00000239, -0.00000138, -0.00000715, -0.00000363, 0.00000832, 0.00000073, -0.00000777, -0.00000174, 0.00000118, -0.00000907, 0.00001157, 0.00000015, -0.00000772, -0.00000626, 0.00000918, -0.00000015, -0.00000319, 0.00000463, 0.00000311, -0.00000393, 0.00000158, 0.00000530, 0.00000362, -0.00000319, -0.00000065, -0.00000473, 0.00000377, -0.00000043, 0.00000029, -0.00000089, 0.00000075, -0.00000563, 0.00001006, -0.00000068, 0.00000503, -0.00000363, 0.00000107, -0.00001068, 0.00002454, -0.00000598, 0.00000698, 0.00000025, -0.00001759, 0.00001737, -0.00002546, 0.00000973, -0.00000437, 0.00001002, -0.00002495, 0.00000561, 0.00000298, 0.00000359, -0.00000973, 0.00000415, -0.00001728, 0.00000634, -0.00000814, 0.00001305, -0.00002422, 0.00000119, 0.00000390, -0.00001163, 0.00001233, -0.00001278, 0.00000323, -0.00000230, -0.00000856, -0.00000096, -0.00000507, -0.00000324, -0.00002374, -0.00000429, -0.00003826, -0.00000654, -0.00018598, 0.00002247, 0.00027992, 0.00002703, 0.00023654, 0.00011374, -0.00020383, -0.00003079, -0.00002706, -0.00001405, -0.00001810, -0.00000758, -0.00000786, -0.00000538, -0.00000489, -0.00000744, 0.00000591, -0.00000839, -0.00000235, -0.00000539, -0.00000154, -0.00001127, -0.00002477, 0.00000684, 0.00000060, 0.00003490, -0.00001131, 0.00001892, -0.00001349, 0.00001964, -0.00000700, 0.00002110, -0.00001006, 0.00002660, 0.00000391, 0.00001081, -0.00001852, -0.00001821, 0.00001678, -0.00000449, },
		{ -0.00001483, 0.00001057, -0.00000771, 0.00000503, -0.00001083, 0.00000951, -0.00000728, 0.00000239, -0.00000509, -0.00000328, -0.00000519, 0.00000101, -0.00000109, 0.00000281, -0.00000006, -0.00000501, -0.00000781, 0.00000921, 0.00000142, -0.00000537, -0.00000560, 0.00000357, -0.00000901, 0.00000840, 0.00000364, -0.00000751, -0.00000729, 0.00000582, 0.00000513, -0.00000566, 0.00000361, 0.00000276, -0.00000101, -0.00000058, 0.00000493, 0.00000319, -0.00000071, -0.00000167, -0.00000450, 0.00000299, 0.00000020, 0.00000009, -0.00000107, 0.00000098, -0.00000547, 0.00000779, 0.00000201, 0.00000294, -0.00000154, 0.00000053, -0.00000982, 0.00001611, 0.00000724, -0.00000667, 0.00001550, -0.00002798, 0.00002166, -0.00002484, 0.00000624, -0.00000303, 0.00001067, -0.00002084, -0.00000198, 0.00000582, 0.00000405, -0.00000954, 0.00000417, -0.00001461, 0.00000148, -0.00000498, 0.00000922, -0.00001533, -0.00000968, 0.00001184, -0.00001675, 0.00001451, -0.00001182, 0.00000161, -0.00000133, -0.00000793, -0.00000251, -0.00000446, -0.00000142, -0.00002251, -0.00000579, -0.00003313, -0.00000784, -0.00013467, -0.00010708, 0.00030995, 0.00006179, 0.00019778, 0.00018928, -0.00020913, -0.00005187, -0.00002421, -0.00001602, -0.00001788, -0.00000995, -0.00000665, -0.00000710, -0.00000360, -0.00000859, 0.00000375, -0.00000359, -0.00000646, -0.00000185, -0.00000579, -0.00000343, -0.00002980, 0.00000854, -0.00000560, 0.00003696, -0.00001049, 0.00001804, -0.00001094, 0.00001393, -0.00000204, 0.00001450, -0.00000271, 0.00001374, 0.00001658, 0.00000136, -0.00000141, -0.00003167, 0.00002029, -0.00000413, -0.00000892, },
		{ 0.00001008, -0.00000695, 0.00000393, -0.00000862, 0.00000635, -0.00000338, -0.00000095, -0.00000212, -0.00000448, -0.00000480, 0.00000037, -0.00000102, 0.00000101, 0.00000216, -0.00000329, -0.00000990, 0.00000713, 0.00000428, -0.00000507, -0.00000597, 0.00000140, -0.00000476, 0.00000250, 0.00000847, -0.00000868, -0.00000567, 0.00000000, 0.00001065, -0.00000697, 0.00000257, 0.00000136, 0.00000186, -0.00000162, 0.00000421, 0.00000237, 0.00000119, -0.00000128, -0.00000531, 0.00000282, 0.00000051, -0.00000010, -0.00000089, 0.00000050, -0.00000441, 0.00000481, 0.00000489, 0.00000094, 0.00000055, -0.00000091, -0.00000635, 0.00000543, 0.00001943, -0.00001590, 0.00002271, -0.00002770, 0.00001561, -0.00001574, -0.00000345, 0.00000309, 0.00000613, -0.00001164, -0.00001214, 0.00000969, 0.00000244, -0.00000606, -0.00000011, -0.00000738, -0.00000680, 0.00000115, 0.00000192, -0.00000373, -0.00002120, 0.00001921, -0.00002129, 0.00001663, -0.00001199, 0.00000248, -0.00000363, -0.00000374, -0.00000756, -0.00000005, -0.00000527, -0.00001359, -0.00001678, -0.00001622, -0.00002784, -0.00006720, -0.00020529, 0.00027140, 0.00014232, 0.00013660, 0.00026950, -0.00020507, -0.00007469, -0.00002511, -0.00001549, -0.00001934, -0.00001067, -0.00000756, -0.00000637, -0.00000529, -0.00000623, -0.00000173, 0.00000289, -0.00001089, 0.00000180, -0.00001007, 0.00000308, -0.00003231, 0.00000814, -0.00000901, 0.00003517, -0.00000584, 0.00001288, -0.00000361, 0.00000434, 0.00000596, 0.00000518, 0.00000716, -0.00000002, 0.00002751, -0.00000577, 0.00001052, -0.00003608, 0.00001531, 0.00000152, -0.00000880, -0.00001682, },
		{ -0.00000309, -0.00000027, -0.00000390, 0.00000139, 0.00000160, -0.00000470, 0.00000078, -0.00000569, -0.00000380, -0.00000140, 0.00000084, -0.00000243, 0.00000472, -0.00000220, -0.00000943, 0.00000219, 0.00000886, -0.00000658, -0.00000342, -0.00000374, 0.00000111, -0.00000338, 0.00001232, -0.00000936, -0.00000361, -0.00000529, 0.00001346, -0.00000521, 0.00000026, 0.00000057, 0.00000285, -0.00000038, 0.00000241, 0.00000227, 0.00000125, 0.00000113, -0.00000718, 0.00000306, 0.00000071, -0.00000036, -0.00000050, -0.00000022, -0.00000320, 0.00000210, 0.00000689, -0.00000004, 0.00000185, -0.00000222, -0.00000243, -0.00000368, 0.00002608, -0.00001679, 0.00001967, -0.00001721, 0.00000216, -0.00000238, -0.00001491, 0.00000988, 0.00000040, -0.00000186, -0.00002047, 0.00001130, 0.00000126, -0.00000180, -0.00000556, 0.00000045, -0.00001418, 0.00000602, -0.00000447, 0.00000584, -0.00002887, 0.00002222, -0.00002207, 0.00001578, -0.00001026, 0.00000229, -0.00000498, -0.00000077, -0.00001097, 0.00000281, -0.00000877, -0.00000463, -0.00002718, -0.00000104, -0.00004567, -0.00001677, -0.00023975, 0.00015781, 0.00025210, 0.00007270, 0.00033817, -0.00018101, -0.00010484, -0.00002496, -0.00001715, -0.00001870, -0.00001263, -0.00000822, -0.00000538, -0.00000762, -0.00000311, -0.00000714, 0.00000759, -0.00001279, 0.00000322, -0.00001218, 0.00000634, -0.00003105, 0.00000483, -0.00000898, 0.00002933, 0.00000236, 0.00000435, 0.00000644, -0.00000609, 0.00001353, -0.00000304, 0.00001508, -0.00000962, 0.00003187, -0.00000602, 0.00001342, -0.00003001, 0.00000301, 0.00001040, -0.00001042, -0.00001633, 0.00000660, },
		{ -0.00000475, 0.00000034, -0.00000252, 0.00000515, -0.00000675, 0.00000207, -0.00000588, -0.00000294, -0.00000352, 0.00000331, -0.00000586, 0.00000619, -0.00000101, -0.00000734, -0.00000384, 0.00001313, -0.00000791, -0.00000040, -0.00000851, 0.00000486, -0.00000595, 0.00001275, -0.00000750, -0.00000351, -0.00000733, 0.00001153, 0.00000033, -0.00000365, 0.00000120, 0.00000121, 0.00000310, -0.00000025, 0.00000310, -0.00000060, 0.00000490, -0.00000906, 0.00000281, 0.00000140, -0.00000096, 0.00000006, -0.00000088, -0.00000233, 0.00000021, 0.00000760, 0.00000019, 0.00000223, -0.00000293, 0.00000048, -0.00000902, 0.00002566, -0.00000974, 0.00000927, -0.00000200, -0.00001176, 0.00000835, -0.00002209, 0.00001226, -0.00000181, 0.00000376, -0.00002290, 0.00000784, 0.00000284, 0.00000041, -0.00000860, 0.00000470, -0.00001652, 0.00000589, -0.00000617, 0.00000954, -0.00002934, 0.00001825, -0.00001675, 0.00000949, -0.00000377, -0.00000245, -0.00000132, -0.00000347, -0.00000803, -0.00000084, -0.00000603, -0.00000325, -0.00002733, 0.00000026, -0.00004475, -0.00000372, -0.00020376, -0.00000682, 0.00035271, 0.00003639, 0.00037461, -0.00012454, -0.00014899, -0.00001798, -0.00002638, -0.00001179, -0.00001915, -0.00000557, -0.00000726, -0.00000729, -0.00000289, -0.00000870, 0.00000722, -0.00000981, 0.00000071, -0.00001062, 0.00000525, -0.00002562, -0.00000146, -0.00000558, 0.00001995, 0.00001299, -0.00000565, 0.00001643, -0.00001416, 0.00001769, -0.00000735, 0.00001820, -0.00001253, 0.00002822, 0.00000092, 0.00000860, -0.00001732, -0.00001117, 0.00001757, -0.00001023, -0.00001600, 0.00000255, 0.00000137, },
		{ 0.00000152, -0.00000326, 0.00000556, -0.00000585, 0.00000095, -0.00000453, -0.00000290, -0.00000493, 0.00000485, -0.00000742, 0.00000515, 0.00000124, -0.00000514, -0.00000862, 0.00001467, -0.00000680, 0.00000040, -0.00000976, 0.00000371, -0.00000346, 0.00000885, -0.00000243, -0.00000627, -0.00000532, 0.00000514, 0.00000829, -0.00000789, 0.00000263, -0.00000218, 0.00000735, -0.00000248, 0.00000406, -0.00000325, 0.00000832, -0.00000932, 0.00000094, 0.00000318, -0.00000215, 0.00000081, -0.00000142, -0.00000181, -0.00000100, 0.00000742, 0.00000103, 0.00000229, -0.00000328, 0.00000212, -0.00001054, 0.00001984, 0.00000122, -0.00000232, 0.00001042, -0.00001891, 0.00001078, -0.00002101, 0.00000746, 0.00000182, 0.00000311, -0.00001828, -0.00000080, 0.00000738, -0.00000045, -0.00000753, 0.00000348, -0.00001243, -0.00000006, -0.00000243, 0.00000672, -0.00002256, 0.00000776, -0.00000581, -0.00000204, 0.00000769, -0.00001234, 0.00000812, -0.00001256, 0.00000185, -0.00001165, 0.00000419, -0.00001155, -0.00001461, -0.00001502, -0.00002236, -0.00002631, -0.00011987, -0.00017410, 0.00039698, 0.00005498, 0.00036467, -0.00002992, -0.00020821, -0.00000275, -0.00004468, 0.00000218, -0.00003090, 0.00000139, -0.00001342, -0.00000253, -0.00000752, -0.00000467, 0.00000084, -0.00000193, -0.00000542, -0.00000568, 0.00000034, -0.00001698, -0.00000978, 0.00000025, 0.00000840, 0.00002421, -0.00001484, 0.00002374, -0.00001774, 0.00001720, -0.00000720, 0.00001650, -0.00000966, 0.00001904, 0.00001116, 0.00000142, -0.00000499, -0.00002039, 0.00001807, -0.00000513, -0.00001809, 0.00000017, 0.00000393, -0.00000680, },
		{ -0.00000075, 0.00000309, -0.00000251, -0.00000193, -0.00000210, -0.00000365, -0.00000511, 0.00000448, -0.00000613, 0.00000124, 0.00000471, -0.00000383, -0.00001055, 0.00001218, -0.00000228, -0.00000217, -0.00000645, -0.00000246, 0.00000307, 0.00000225, 0.00000431, -0.00001065, -0.00000088, -0.00000305, 0.00001549, -0.00001006, 0.00000321, -0.00000541, 0.00001014, -0.00000262, 0.00000395, -0.00000512, 0.00000953, -0.00000664, -0.00000304, 0.00000608, -0.00000375, 0.00000161, -0.00000184, -0.00000142, -0.00000200, 0.00000707, 0.00000158, 0.00000283, -0.00000381, 0.00000296, -0.00000962, 0.00001203, 0.00001087, -0.00000909, 0.00001451, -0.00001575, 0.00000383, -0.00001236, -0.00000290, 0.00000947, -0.00000182, -0.00000939, -0.00001135, 0.00001234, -0.00000279, -0.00000361, -0.00000167, -0.00000425, -0.00000893, 0.00000389, 0.00000042, -0.00001212, -0.00000535, 0.00000704, -0.00001534, 0.00002075, -0.00002393, 0.00001959, -0.00002384, 0.00001401, -0.00002461, 0.00001688, -0.00002429, 0.00000452, -0.00003818, 0.00000932, -0.00006347, -0.00002928, -0.00029022, 0.00034989, 0.00013631, 0.00031172, 0.00009456, -0.00027273, 0.00001371, -0.00006658, 0.00001823, -0.00004387, 0.00000977, -0.00002178, 0.00000499, -0.00001546, 0.00000312, -0.00000898, 0.00000787, -0.00001245, 0.00000035, -0.00000610, -0.00000746, -0.00001813, 0.00000673, -0.00000323, 0.00003372, -0.00002096, 0.00002659, -0.00001623, 0.00001282, -0.00000417, 0.00001221, -0.00000434, 0.00000907, 0.00001912, -0.00000199, 0.00000085, -0.00002027, 0.00001016, 0.00000493, -0.00002244, -0.00000025, 0.00000386, -0.00000557, -0.00000088, },
		{ -0.00000005, 0.00000115, -0.00000469, -0.00000001, -0.00000440, -0.00000441, 0.00000241, -0.00000252, -0.00000439, 0.00000835, -0.00000316, -0.00000965, 0.00000637, 0.00000453, -0.00000694, -0.00000022, -0.00001090, 0.00001007, -0.00000352, 0.00000968, -0.00001389, 0.00000289, -0.00000928, 0.00001849, -0.00000801, 0.00000150, -0.00000669, 0.00000971, 0.00000029, 0.00000217, -0.00000520, 0.00000757, -0.00000097, -0.00000845, 0.00000921, -0.00000503, 0.00000201, -0.00000203, -0.00000100, -0.00000315, 0.00000706, 0.00000138, 0.00000409, -0.00000463, 0.00000338, -0.00000781, 0.00000527, 0.00001565, -0.00000836, 0.00000962, -0.00000445, -0.00000802, -0.00000158, -0.00001335, 0.00001606, -0.00000610, -0.00000154, -0.00001870, 0.00001375, -0.00000346, -0.00000008, -0.00000682, 0.00000326, -0.00001568, 0.00000793, -0.00000441, -0.00000339, -0.00001567, 0.00001659, -0.00002523, 0.00002994, -0.00003137, 0.00002663, -0.00003025, 0.00002085, -0.00003171, 0.00002363, -0.00003199, 0.00001838, -0.00005438, 0.00003116, -0.00008763, 0.00002786, -0.00031925, 0.00020763, 0.00025838, 0.00024039, 0.00022580, -0.00032236, 0.00001693, -0.00008069, 0.00002672, -0.00005040, 0.00001349, -0.00002717, 0.00001055, -0.00002206, 0.00000978, -0.00001710, 0.00001466, -0.00001628, 0.00000408, -0.00001093, -0.00000001, -0.00002410, 0.00001162, -0.00001254, 0.00003923, -0.00002226, 0.00002424, -0.00001043, 0.00000647, -0.00000062, 0.00000806, 0.00000000, 0.00000240, 0.00002094, 0.00000154, -0.00000159, -0.00001163, -0.00000301, 0.00001616, -0.00002601, -0.00000075, 0.00000270, -0.00000256, -0.00000500, 0.00000282, },
		{ 0.00000269, -0.00000531, 0.00000026, -0.00000411, -0.00000374, -0.00000019, 0.00000168, -0.00000950, 0.00001024, -0.00000198, -0.00000734, -0.00000045, 0.00001092, -0.00001117, 0.00000558, -0.00001741, 0.00001325, -0.00000491, 0.00001092, -0.00001356, 0.00000327, -0.00001076, 0.00001549, -0.00000136, -0.00000251, -0.00000557, 0.00000589, 0.00000566, -0.00000068, -0.00000365, 0.00000298, 0.00000624, -0.00001346, 0.00001101, -0.00000489, 0.00000140, -0.00000173, -0.00000065, -0.00000436, 0.00000721, 0.00000078, 0.00000550, -0.00000512, 0.00000318, -0.00000580, 0.00000084, 0.00001516, -0.00000181, 0.00000003, 0.00000827, -0.00001710, 0.00000414, -0.00001770, 0.00001637, -0.00000495, 0.00000057, -0.00001888, 0.00000890, -0.00000011, 0.00000015, -0.00000828, 0.00000584, -0.00001617, 0.00000597, -0.00000400, -0.00000038, -0.00001926, 0.00001906, -0.00002767, 0.00003083, -0.00002974, 0.00002391, -0.00002612, 0.00001649, -0.00002686, 0.00001802, -0.00002713, 0.00001753, -0.00005198, 0.00002899, -0.00008051, 0.00003131, -0.00025947, 0.00000445, 0.00037270, 0.00018864, 0.00033413, -0.00033397, -0.00000856, -0.00007543, 0.00001829, -0.00004329, 0.00000673, -0.00002436, 0.00000906, -0.00002220, 0.00001011, -0.00001844, 0.00001412, -0.00001370, 0.00000294, -0.00001168, 0.00000313, -0.00002587, 0.00001313, -0.00001753, 0.00003909, -0.00001790, 0.00001702, -0.00000187, 0.00000013, 0.00000175, 0.00000552, 0.00000184, 0.00000025, 0.00001663, 0.00001037, -0.00000873, -0.00000032, -0.00001459, 0.00002259, -0.00002453, -0.00000434, 0.00000302, -0.00000051, -0.00000779, 0.00000439, -0.00000108, },
		{ -0.00000299, -0.00000178, -0.00000239, -0.00000380, -0.00000209, 0.00000458, -0.00001193, 0.00000883, 0.00000079, -0.00000533, -0.00000566, 0.00001416, -0.00001218, 0.00000785, -0.00001866, 0.00001001, -0.00000050, 0.00000739, -0.00000915, -0.00000037, -0.00000719, 0.00000747, 0.00000781, -0.00000700, -0.00000324, 0.00000015, 0.00001147, -0.00000293, -0.00000170, -0.00000251, 0.00001257, -0.00001582, 0.00000995, -0.00000251, -0.00000058, -0.00000074, -0.00000063, -0.00000518, 0.00000687, 0.00000071, 0.00000594, -0.00000426, 0.00000173, -0.00000337, -0.00000183, 0.00001173, 0.00000594, -0.00000771, 0.00001499, -0.00001684, 0.00000019, -0.00001302, 0.00000860, 0.00000294, -0.00000411, -0.00001177, -0.00000145, 0.00000672, -0.00000327, -0.00000502, 0.00000235, -0.00000979, -0.00000205, 0.00000176, -0.00000333, -0.00001601, 0.00001437, -0.00002230, 0.00002270, -0.00001825, 0.00001070, -0.00001098, 0.00000079, -0.00001023, 0.00000031, -0.00000946, 0.00000101, -0.00002989, 0.00000238, -0.00004297, -0.00001094, -0.00014417, -0.00019813, 0.00042240, 0.00018978, 0.00039787, -0.00029296, -0.00006987, -0.00004659, -0.00000989, -0.00002095, -0.00001176, -0.00001181, -0.00000140, -0.00001372, 0.00000190, -0.00001112, 0.00000526, -0.00000452, -0.00000310, -0.00000809, 0.00000164, -0.00002320, 0.00001072, -0.00001740, 0.00003294, -0.00000835, 0.00000626, 0.00000761, -0.00000492, 0.00000271, 0.00000415, 0.00000210, 0.00000072, 0.00000981, 0.00001924, -0.00001386, 0.00000606, -0.00001779, 0.00001960, -0.00001543, -0.00001259, 0.00000637, -0.00000133, -0.00000705, 0.00000196, 0.00000153, 0.00000095, },
		{ -0.00000503, 0.00000001, -0.00000445, -0.00000276, 0.00000521, -0.00001083, 0.00000398, 0.00000508, -0.00000442, -0.00000784, 0.00001307, -0.00000905, 0.00000557, -0.00001398, 0.00000091, 0.00000794, 0.00000125, -0.00000264, -0.00000620, -0.00000096, -0.00000211, 0.00001576, -0.00000919, -0.00000171, -0.00000517, 0.00001522, -0.00000280, -0.00000075, -0.00000695, 0.00001578, -0.00001402, 0.00000551, 0.00000200, -0.00000361, 0.00000075, -0.00000106, -0.00000517, 0.00000538, 0.00000200, 0.00000455, -0.00000142, -0.00000132, -0.00000021, -0.00000398, 0.00000846, 0.00001023, -0.00000855, 0.00001171, -0.00000577, -0.00001239, -0.00000191, -0.00000398, 0.00001416, -0.00001206, -0.00000153, -0.00001282, 0.00001339, -0.00000767, 0.00000059, -0.00000451, -0.00000002, -0.00001217, 0.00000905, -0.00000856, -0.00000977, 0.00000633, -0.00001269, 0.00000921, -0.00000100, -0.00000824, 0.00000957, -0.00001986, 0.00001111, -0.00002193, 0.00001318, -0.00002262, 0.00000141, -0.00003497, 0.00000632, -0.00006898, -0.00002599, -0.00033600, 0.00036823, 0.00025262, 0.00041644, -0.00020236, -0.00015918, -0.00000223, -0.00005048, 0.00000982, -0.00003640, 0.00000618, -0.00001749, 0.00000056, -0.00001218, 0.00000190, -0.00000846, 0.00000763, -0.00001100, -0.00000243, -0.00000258, -0.00001774, 0.00000547, -0.00001299, 0.00002207, 0.00000450, -0.00000581, 0.00001620, -0.00000816, 0.00000309, 0.00000235, 0.00000300, 0.00000053, 0.00000518, 0.00002257, -0.00001128, 0.00000264, -0.00000995, 0.00000712, -0.00000024, -0.00002380, 0.00001167, -0.00000456, -0.00000304, -0.00000389, 0.00000683, -0.00000295, 0.00000111, },
		{ 0.00000147, -0.00000474, -0.00000269, 0.00000398, -0.00000708, -0.00000271, 0.00000941, -0.00000402, -0.00000734, 0.00000863, -0.00000317, 0.00000026, -0.00000566, -0.00001045, 0.00001616, -0.00000356, 0.00000261, -0.00001105, 0.00000431, -0.00000913, 0.00001891, -0.00000696, -0.00000242, -0.00000827, 0.00001528, 0.00000051, -0.00000137, -0.00000923, 0.00001493, -0.00000818, -0.00000131, 0.00000738, -0.00000661, 0.00000207, -0.00000170, -0.00000429, 0.00000259, 0.00000473, 0.00000143, 0.00000298, -0.00000536, 0.00000337, -0.00000631, 0.00000724, 0.00000908, -0.00000179, 0.00000016, 0.00001139, -0.00002686, 0.00000850, -0.00001449, 0.00002234, -0.00001712, 0.00000550, -0.00001931, 0.00001522, -0.00000910, 0.00000443, -0.00000994, 0.00000756, -0.00001862, 0.00001246, -0.00001079, -0.00000588, 0.00000024, -0.00000418, -0.00000378, 0.00001519, -0.00002501, 0.00002666, -0.00003577, 0.00002681, -0.00003776, 0.00002918, -0.00004027, 0.00002603, -0.00006327, 0.00004233, -0.00010822, 0.00004782, -0.00037013, 0.00020992, 0.00035085, 0.00041366, -0.00008315, -0.00025471, 0.00003901, -0.00008795, 0.00003586, -0.00005654, 0.00002100, -0.00003179, 0.00001384, -0.00002558, 0.00001409, -0.00002053, 0.00001686, -0.00001606, 0.00000160, -0.00000636, -0.00001222, -0.00000044, -0.00000637, 0.00000909, 0.00001769, -0.00001650, 0.00002219, -0.00000951, 0.00000392, -0.00000122, 0.00000613, -0.00000252, 0.00000554, 0.00001795, 0.00000016, -0.00000973, 0.00000541, -0.00000944, 0.00001534, -0.00003321, 0.00001537, -0.00000744, 0.00000159, -0.00001019, 0.00001172, -0.00000611, 0.00000439, -0.00000727, },
		{ -0.00000366, -0.00000285, 0.00000221, -0.00000270, -0.00000866, 0.00001161, -0.00000274, -0.00000578, 0.00000331, 0.00000262, -0.00000504, 0.00000245, -0.00001937, 0.00001964, -0.00000357, 0.00000393, -0.00001239, 0.00000584, -0.00001091, 0.00001585, -0.00000039, -0.00000508, -0.00000893, 0.00001175, 0.00000611, -0.00000285, -0.00000954, 0.00001070, -0.00000009, -0.00000825, 0.00001162, -0.00000811, 0.00000233, -0.00000206, -0.00000296, -0.00000088, 0.00000806, -0.00000222, 0.00000743, -0.00000884, 0.00000617, -0.00000831, 0.00000782, 0.00000389, 0.00000885, -0.00001326, 0.00002609, -0.00003433, 0.00001044, -0.00001649, 0.00002202, -0.00001437, 0.00000466, -0.00001712, 0.00000958, -0.00000519, 0.00000350, -0.00001016, 0.00000870, -0.00001741, 0.00000854, -0.00000673, -0.00000773, -0.00000048, -0.00000049, -0.00001163, 0.00002461, -0.00003303, 0.00003307, -0.00003942, 0.00002916, -0.00003926, 0.00003019, -0.00004224, 0.00003214, -0.00006827, 0.00004823, -0.00010806, 0.00005660, -0.00030355, -0.00000879, 0.00042956, 0.00042671, 0.00003645, -0.00033016, 0.00005635, -0.00010606, 0.00004421, -0.00006207, 0.00002431, -0.00003688, 0.00001912, -0.00003148, 0.00001884, -0.00002485, 0.00001819, -0.00001463, 0.00000119, -0.00000713, -0.00000898, -0.00000498, 0.00000013, -0.00000307, 0.00002819, -0.00002334, 0.00002433, -0.00000896, 0.00000552, -0.00000645, 0.00001108, -0.00000801, 0.00001007, 0.00000761, 0.00001621, -0.00002461, 0.00002010, -0.00002162, 0.00002410, -0.00003545, 0.00001353, -0.00000664, 0.00000346, -0.00001333, 0.00001272, -0.00000558, 0.00000436, -0.00000655, 0.00000058, },
		{ -0.00000377, 0.00000111, 0.00000039, -0.00001169, 0.00001018, 0.00000038, -0.00000475, -0.00000050, 0.00000579, -0.00000768, 0.00000714, -0.00002249, 0.00001606, 0.00000213, 0.00000110, -0.00001000, 0.00000329, -0.00000764, 0.00000814, 0.00000799, -0.00000764, -0.00000847, 0.00000631, 0.00001185, -0.00000355, -0.00000908, 0.00000494, 0.00000759, -0.00001279, 0.00001292, -0.00000709, 0.00000101, -0.00000173, -0.00000178, -0.00000408, 0.00001063, -0.00000473, 0.00001006, -0.00001002, 0.00000680, -0.00000881, 0.00000845, -0.00000173, 0.00001757, -0.00002106, 0.00003056, -0.00002873, 0.00000028, -0.00000809, 0.00001236, -0.00000346, -0.00000400, -0.00000719, -0.00000185, 0.00000281, -0.00000185, -0.00000494, 0.00000314, -0.00000890, -0.00000173, 0.00000254, -0.00001436, 0.00000335, -0.00000133, -0.00001371, 0.00002582, -0.00003056, 0.00002725, -0.00002975, 0.00001767, -0.00002640, 0.00001630, -0.00002822, 0.00001889, -0.00004938, 0.00002468, -0.00007166, 0.00001327, -0.00017882, -0.00021605, 0.00042878, 0.00048411, 0.00013795, -0.00036823, 0.00003803, -0.00009715, 0.00002969, -0.00004965, 0.00001347, -0.00003007, 0.00001352, -0.00002701, 0.00001348, -0.00001937, 0.00001062, -0.00000662, -0.00000362, -0.00000466, -0.00000847, -0.00000748, 0.00000516, -0.00001232, 0.00003397, -0.00002501, 0.00002223, -0.00000663, 0.00000739, -0.00001174, 0.00001555, -0.00001313, 0.00001501, -0.00000293, 0.00002953, -0.00003347, 0.00002555, -0.00002238, 0.00002139, -0.00002792, 0.00000462, -0.00000075, 0.00000082, -0.00001134, 0.00000822, -0.00000031, 0.00000041, -0.00000253, -0.00000276, -0.00000109, },
		{ 0.00000091, 0.00000147, -0.00001124, 0.00000535, 0.00000492, -0.00000464, -0.00000181, 0.00000554, -0.00000699, 0.00000753, -0.00001946, 0.00000659, 0.00001113, -0.00000334, -0.00000604, -0.00000125, -0.00000207, -0.00000045, 0.00001434, -0.00000747, -0.00000872, 0.00000116, 0.00001544, -0.00000196, -0.00000900, -0.00000046, 0.00001261, -0.00001337, 0.00001072, -0.00000359, -0.00000160, -0.00000083, -0.00000108, -0.00000619, 0.00001136, -0.00000486, 0.00000969, -0.00000802, 0.00000462, -0.00000707, 0.00000739, -0.00000438, 0.00001981, -0.00001885, 0.00002212, -0.00001046, -0.00001881, 0.00000621, -0.00000178, 0.00001067, -0.00001542, 0.00000488, -0.00001346, 0.00001026, -0.00000790, 0.00000222, -0.00000517, 0.00000216, -0.00001309, 0.00001189, -0.00002087, 0.00000709, -0.00000267, -0.00001334, 0.00002200, -0.00002136, 0.00001396, -0.00001272, -0.00000062, -0.00000708, -0.00000387, -0.00000730, -0.00000348, -0.00001929, -0.00001156, -0.00002197, -0.00004596, -0.00005670, -0.00034272, 0.00031412, 0.00058380, 0.00022639, -0.00036992, -0.00001172, -0.00006733, -0.00000118, -0.00002568, -0.00000612, -0.00001562, 0.00000050, -0.00001521, 0.00000111, -0.00000755, -0.00000186, 0.00000396, -0.00000954, -0.00000132, -0.00000894, -0.00000903, 0.00000885, -0.00001818, 0.00003473, -0.00002182, 0.00001665, -0.00000299, 0.00000865, -0.00001490, 0.00001674, -0.00001463, 0.00001617, -0.00000814, 0.00003389, -0.00003049, 0.00001762, -0.00001031, 0.00000823, -0.00001296, -0.00000888, 0.00000833, -0.00000509, -0.00000529, -0.00000030, 0.00000782, -0.00000552, 0.00000290, -0.00000689, 0.00000181, -0.00000501, },
		{ 0.00000124, -0.00000855, -0.00000090, 0.00000903, -0.00000446, -0.00000141, 0.00000318, -0.00000446, 0.00000521, -0.00001278, -0.00000479, 0.00001892, -0.00000550, -0.00000369, -0.00000474, 0.00000225, -0.00000598, 0.00001561, -0.00000312, -0.00001053, -0.00000237, 0.00001572, 0.00000218, -0.00000953, -0.00000456, 0.00001411, -0.00001018, 0.00000603, 0.00000106, -0.00000437, 0.00000001, -0.00000074, -0.00000700, 0.00000997, -0.00000249, 0.00000653, -0.00000343, 0.00000040, -0.00000354, 0.00000424, -0.00000284, 0.00001468, -0.00000751, 0.00000460, 0.00001337, -0.00003795, 0.00001765, -0.00001229, 0.00002053, -0.00002236, 0.00001191, -0.00001874, 0.00001192, -0.00001004, 0.00000640, -0.00001063, 0.00000944, -0.00001909, 0.00001527, -0.00002154, 0.00000537, 0.00000032, -0.00001507, 0.00001827, -0.00001178, 0.00000101, 0.00000252, -0.00001549, 0.00000763, -0.00001857, 0.00000787, -0.00002052, 0.00000435, -0.00003853, 0.00001327, -0.00008330, 0.00001332, -0.00035381, 0.00009909, 0.00068413, 0.00033085, -0.00035261, -0.00007520, -0.00003359, -0.00003330, -0.00000324, -0.00002373, -0.00000243, -0.00001212, -0.00000340, -0.00001118, 0.00000351, -0.00001234, 0.00001102, -0.00001185, -0.00000063, -0.00000767, -0.00001160, 0.00001228, -0.00002135, 0.00003163, -0.00001542, 0.00000927, 0.00000114, 0.00000868, -0.00001444, 0.00001308, -0.00001088, 0.00001150, -0.00000562, 0.00002754, -0.00001580, -0.00000088, 0.00000918, -0.00000862, 0.00000276, -0.00002131, 0.00001612, -0.00001056, 0.00000125, -0.00000896, 0.00001485, -0.00000982, 0.00000662, -0.00000924, 0.00000314, -0.00000534, 0.00000085, },
		{ -0.00000570, -0.00000592, 0.00001049, -0.00000277, -0.00000098, 0.00000103, -0.00000259, 0.00000289, -0.00000613, -0.00001350, 0.00002133, -0.00000251, -0.00000493, -0.00000524, 0.00000312, -0.00000653, 0.00001124, 0.00000446, -0.00001296, -0.00000455, 0.00001323, 0.00000763, -0.00000975, -0.00000761, 0.00001277, -0.00000492, 0.00000102, 0.00000494, -0.00000592, -0.00000001, -0.00000039, -0.00000684, 0.00000704, 0.00000139, 0.00000203, 0.00000189, -0.00000397, 0.00000023, 0.00000012, 0.00000163, 0.00000503, 0.00000756, -0.00001376, 0.00003256, -0.00004719, 0.00001804, -0.00001265, 0.00002061, -0.00001992, 0.00000945, -0.00001430, 0.00000553, -0.00000607, 0.00000473, -0.00000964, 0.00000899, -0.00001610, 0.00000968, -0.00001382, -0.00000399, 0.00000945, -0.00002091, 0.00001770, -0.00000635, -0.00000581, 0.00000930, -0.00001989, 0.00001053, -0.00002038, 0.00000939, -0.00002296, 0.00001021, -0.00004269, 0.00001833, -0.00008022, 0.00001568, -0.00026459, -0.00015377, 0.00071635, 0.00048800, -0.00033571, -0.00013319, -0.00001335, -0.00005222, 0.00000591, -0.00003008, 0.00000176, -0.00001733, 0.00000173, -0.00001683, 0.00000755, -0.00001520, 0.00001021, -0.00000765, -0.00000456, -0.00000313, -0.00001624, 0.00001604, -0.00002266, 0.00002622, -0.00000778, 0.00000185, 0.00000492, 0.00000730, -0.00001016, 0.00000510, -0.00000290, 0.00000228, 0.00000276, 0.00001403, 0.00000462, -0.00002162, 0.00002650, -0.00002000, 0.00001169, -0.00002694, 0.00001823, -0.00001181, 0.00000443, -0.00001369, 0.00001699, -0.00000935, 0.00000612, -0.00000787, 0.00000154, -0.00000377, -0.00000006, 0.00000107, },
		{ -0.00000792, 0.00000822, 0.00000102, -0.00000164, 0.00000083, -0.00000312, 0.00000243, -0.00000209, -0.00001683, 0.00001690, 0.00000550, -0.00000905, -0.00000326, 0.00000089, -0.00000310, 0.00000366, 0.00001214, -0.00001364, -0.00000682, 0.00000974, 0.00001226, -0.00000821, -0.00001046, 0.00001006, 0.00000013, -0.00000222, 0.00000661, -0.00000545, -0.00000119, 0.00000017, -0.00000613, 0.00000351, 0.00000533, -0.00000194, 0.00000590, -0.00000661, 0.00000258, -0.00000332, 0.00000647, -0.00000460, 0.00001960, -0.00002480, 0.00003916, -0.00004095, 0.00000481, -0.00000218, 0.00001116, -0.00000882, -0.00000121, -0.00000238, -0.00000612, 0.00000178, -0.00000145, -0.00000308, 0.00000180, -0.00000587, -0.00000237, -0.00000057, -0.00001783, 0.00002143, -0.00002813, 0.00001876, -0.00000465, -0.00000638, 0.00000760, -0.00001435, 0.00000283, -0.00001113, -0.00000059, -0.00001275, 0.00000017, -0.00002684, -0.00000169, -0.00004608, -0.00002704, -0.00013263, -0.00035656, 0.00061593, 0.00071458, -0.00032061, -0.00017989, -0.00001236, -0.00005412, -0.00000038, -0.00002427, -0.00000372, -0.00001417, -0.00000112, -0.00001446, 0.00000342, -0.00000998, 0.00000214, 0.00000156, -0.00001149, 0.00000323, -0.00002160, 0.00001899, -0.00002173, 0.00001911, 0.00000008, -0.00000481, 0.00000815, 0.00000451, -0.00000298, -0.00000500, 0.00000629, -0.00000790, 0.00001233, -0.00000014, 0.00002236, -0.00003521, 0.00003296, -0.00001948, 0.00000985, -0.00002355, 0.00001328, -0.00000755, 0.00000265, -0.00001272, 0.00001293, -0.00000353, 0.00000127, -0.00000299, -0.00000263, -0.00000053, -0.00000205, 0.00000185, -0.00000428, },
		{ 0.00000313, 0.00000583, -0.00000304, 0.00000260, -0.00000578, 0.00000364, -0.00000081, -0.00001517, 0.00000768, 0.00001523, -0.00001283, -0.00000144, -0.00000194, 0.00000116, -0.00000311, 0.00001610, -0.00001029, -0.00001054, 0.00000694, 0.00001431, -0.00000407, -0.00001360, 0.00000720, 0.00000342, -0.00000273, 0.00000594, -0.00000340, -0.00000294, 0.00000057, -0.00000501, 0.00000007, 0.00000827, -0.00000417, 0.00000749, -0.00000672, 0.00000291, -0.00000515, 0.00000965, -0.00001061, 0.00002419, -0.00002462, 0.00003157, -0.00002129, -0.00001698, 0.00001290, -0.00000147, 0.00000461, -0.00001353, 0.00001002, -0.00001622, 0.00000780, -0.00000726, 0.00000416, -0.00000664, 0.00000524, -0.00001391, 0.00001106, -0.00002890, 0.00002918, -0.00003040, 0.00001615, -0.00000183, -0.00000594, 0.00000368, -0.00000647, -0.00000657, -0.00000078, -0.00001122, -0.00000160, -0.00001239, -0.00000756, -0.00002523, -0.00001002, -0.00007025, -0.00002889, -0.00043505, 0.00035939, 0.00098284, -0.00027743, -0.00023074, -0.00001924, -0.00004986, -0.00001182, -0.00001579, -0.00001084, -0.00000925, -0.00000625, -0.00000930, -0.00000364, -0.00000226, -0.00000716, 0.00000987, -0.00001626, 0.00000713, -0.00002403, 0.00001819, -0.00001682, 0.00000959, 0.00000854, -0.00001128, 0.00001155, 0.00000012, 0.00000580, -0.00001462, 0.00001354, -0.00001541, 0.00001860, -0.00000925, 0.00003111, -0.00003627, 0.00002562, -0.00000713, -0.00000051, -0.00001423, 0.00000421, -0.00000010, -0.00000236, -0.00000770, 0.00000482, 0.00000504, -0.00000524, 0.00000287, -0.00000716, 0.00000271, -0.00000413, 0.00000322, -0.00000436, -0.00000703, },
		{ 0.00000926, -0.00000355, 0.00000474, -0.00000850, 0.00000443, -0.00000031, -0.00001142, -0.00000194, 0.00002189, -0.00001240, -0.00000271, -0.00000257, 0.00000301, -0.00000564, 0.00001409, -0.00000242, -0.00001561, 0.00000529, 0.00001350, 0.00000207, -0.00001641, 0.00000436, 0.00000478, -0.00000125, 0.00000428, -0.00000123, -0.00000400, 0.00000004, -0.00000326, -0.00000313, 0.00000997, -0.00000462, 0.00000701, -0.00000508, 0.00000206, -0.00000576, 0.00001074, -0.00001203, 0.00002109, -0.00001518, 0.00001507, 0.00000304, -0.00003696, 0.00002270, -0.00000843, 0.00001218, -0.00001956, 0.00001507, -0.00001784, 0.00000637, -0.00000749, 0.00000641, -0.00000990, 0.00001003, -0.00001754, 0.00001383, -0.00003012, 0.00002600, -0.00002187, 0.00000467, 0.00000754, -0.00001100, 0.00000548, -0.00000563, -0.00000718, -0.00000073, -0.00001031, -0.00000275, -0.00001211, -0.00000363, -0.00002896, -0.00000135, -0.00007466, -0.00000208, -0.00036476, -0.00001241, 0.00121496, -0.00014997, -0.00031520, -0.00001212, -0.00005872, -0.00001150, -0.00001913, -0.00000762, -0.00001274, -0.00000458, -0.00000993, -0.00000525, -0.00000037, -0.00000963, 0.00001011, -0.00001308, 0.00000391, -0.00001960, 0.00001062, -0.00000606, -0.00000344, 0.00001867, -0.00001875, 0.00001620, -0.00000604, 0.00001497, -0.00002186, 0.00001707, -0.00001856, 0.00001964, -0.00001125, 0.00002977, -0.00002600, 0.00000881, 0.00001023, -0.00001185, -0.00000587, -0.00000327, 0.00000588, -0.00000658, -0.00000262, -0.00000294, 0.00001186, -0.00000930, 0.00000614, -0.00000907, 0.00000367, -0.00000465, 0.00000387, -0.00000461, -0.00000511, -0.00000221, },
		{ -0.00000153, 0.00000530, -0.00000881, 0.00000245, 0.00000181, -0.00000891, -0.00000776, 0.00002191, -0.00000572, -0.00000815, 0.00000005, 0.00000113, -0.00000308, 0.00000680, 0.00000784, -0.00002009, 0.00000372, 0.00001107, 0.00000833, -0.00001743, 0.00000084, 0.00000521, 0.00000043, 0.00000362, -0.00000061, -0.00000324, -0.00000200, -0.00000069, -0.00000618, 0.00001080, -0.00000409, 0.00000577, -0.00000335, 0.00000163, -0.00000640, 0.00001067, -0.00001025, 0.00001355, -0.00000249, -0.00000145, 0.00002136, -0.00004506, 0.00001943, -0.00000385, 0.00000906, -0.00001499, 0.00000911, -0.00000849, -0.00000397, -0.00000055, 0.00000128, -0.00000487, 0.00000510, -0.00001017, 0.00000517, -0.00001932, 0.00001033, -0.00000181, -0.00001613, 0.00002459, -0.00002391, 0.00001650, -0.00001607, 0.00000555, -0.00001554, 0.00000673, -0.00002111, 0.00000705, -0.00002269, -0.00000386, -0.00003005, -0.00002970, -0.00005464, -0.00018555, -0.00039912, 0.00130473, 0.00011832, -0.00045467, 0.00002379, -0.00009385, 0.00001156, -0.00004312, 0.00001280, -0.00003043, 0.00000934, -0.00002174, 0.00000400, -0.00000954, -0.00000063, -0.00000119, 0.00000023, -0.00000791, -0.00000725, -0.00000420, 0.00001033, -0.00001962, 0.00003047, -0.00002747, 0.00002217, -0.00001335, 0.00002319, -0.00002572, 0.00001685, -0.00001799, 0.00001656, -0.00000808, 0.00002214, -0.00001094, -0.00000853, 0.00002288, -0.00001561, -0.00000502, -0.00000434, 0.00000659, -0.00000656, -0.00000114, -0.00000643, 0.00001326, -0.00000790, 0.00000444, -0.00000657, 0.00000105, -0.00000255, 0.00000263, -0.00000347, -0.00000439, -0.00000510, 0.00000905, },
		{ 0.00000356, -0.00000560, -0.00000312, 0.00000640, -0.00000919, -0.00000813, 0.00001501, 0.00000581, -0.00001583, 0.00000431, -0.00000302, 0.00000244, -0.00000230, 0.00001649, -0.00002107, 0.00000042, 0.00000901, 0.00001247, -0.00001522, -0.00000407, 0.00000588, 0.00000062, 0.00000529, -0.00000218, -0.00000059, -0.00000521, 0.00000226, -0.00000889, 0.00001104, -0.00000338, 0.00000497, -0.00000282, 0.00000266, -0.00000782, 0.00001037, -0.00000733, 0.00000570, 0.00000726, -0.00001059, 0.00002684, -0.00003729, 0.00000243, 0.00001089, -0.00000259, -0.00000237, -0.00000471, 0.00000762, -0.00001861, 0.00000971, -0.00000824, 0.00000576, -0.00000657, 0.00000446, -0.00001045, -0.00000147, -0.00001212, 0.00002337, -0.00003981, 0.00004364, -0.00003977, 0.00003213, -0.00003290, 0.00002590, -0.00003846, 0.00003213, -0.00004813, 0.00003602, -0.00005464, 0.00003754, -0.00007912, 0.00004006, -0.00014279, 0.00001423, -0.00067607, 0.00116159, 0.00054618, -0.00063566, 0.00007725, -0.00014766, 0.00005068, -0.00008127, 0.00004410, -0.00005691, 0.00003120, -0.00004131, 0.00002124, -0.00002700, 0.00001667, -0.00002010, 0.00001925, -0.00002405, 0.00000907, -0.00002233, 0.00002841, -0.00003552, 0.00004128, -0.00003526, 0.00002737, -0.00001951, 0.00002839, -0.00002545, 0.00001365, -0.00001540, 0.00001183, -0.00000344, 0.00001386, 0.00000138, -0.00001813, 0.00002386, -0.00000759, -0.00001339, 0.00000136, 0.00000201, -0.00000204, -0.00000395, -0.00000470, 0.00000873, -0.00000122, -0.00000159, -0.00000053, -0.00000425, 0.00000156, -0.00000040, -0.00000046, -0.00000595, -0.00000434, 0.00000471, 0.00000315, },
		{ 0.00000000, -0.00001068, 0.00001187, -0.00001111, -0.00000472, 0.00000452, 0.00001770, -0.00002153, 0.00000678, -0.00000625, 0.00000713, -0.00000874, 0.00001965, -0.00001645, -0.00000559, 0.00000856, 0.00001328, -0.00000958, -0.00001013, 0.00000709, -0.00000107, 0.00000894, -0.00000490, 0.00000255, -0.00000811, 0.00000435, -0.00001044, 0.00001034, -0.00000247, 0.00000466, -0.00000336, 0.00000457, -0.00000932, 0.00000958, -0.00000418, -0.00000010, 0.00001114, -0.00001031, 0.00002007, -0.00001817, -0.00002068, 0.00002718, -0.00001429, 0.00000993, -0.00001771, 0.00002135, -0.00002875, 0.00001563, -0.00001515, 0.00001477, -0.00001702, 0.00001742, -0.00002341, 0.00001334, -0.00003059, 0.00004239, -0.00005528, 0.00005419, -0.00004822, 0.00004144, -0.00004396, 0.00004010, -0.00005406, 0.00004893, -0.00006536, 0.00005464, -0.00007639, 0.00006732, -0.00011361, 0.00008918, -0.00020120, 0.00013795, -0.00074623, 0.00075881, 0.00108320, -0.00079421, 0.00010363, -0.00018763, 0.00007896, -0.00011032, 0.00006612, -0.00007509, 0.00004642, -0.00005582, 0.00003473, -0.00004158, 0.00003128, -0.00003569, 0.00003361, -0.00003525, 0.00002105, -0.00003599, 0.00004094, -0.00004492, 0.00004592, -0.00003753, 0.00002756, -0.00002064, 0.00002771, -0.00001996, 0.00000778, -0.00001173, 0.00000695, 0.00000024, 0.00000847, 0.00000693, -0.00001718, 0.00001327, 0.00000898, -0.00002600, 0.00000871, -0.00000346, 0.00000343, -0.00000816, -0.00000054, 0.00000152, 0.00000709, -0.00000830, 0.00000562, -0.00000923, 0.00000527, -0.00000348, 0.00000323, -0.00000887, -0.00000114, -0.00000186, 0.00000855, 0.00000076, },
		{ -0.00001691, 0.00001501, -0.00001169, -0.00000132, -0.00000442, 0.00002462, -0.00002109, 0.00000439, -0.00000572, 0.00000783, -0.00000945, 0.00001569, -0.00000654, -0.00001337, 0.00000929, 0.00001135, -0.00000198, -0.00001576, 0.00000787, -0.00000352, 0.00001261, -0.00000640, 0.00000386, -0.00000868, 0.00000400, -0.00000967, 0.00000783, -0.00000052, 0.00000377, -0.00000338, 0.00000535, -0.00000884, 0.00000677, -0.00000006, -0.00000424, 0.00001057, -0.00000444, 0.00000830, 0.00000167, -0.00003782, 0.00003369, -0.00001601, 0.00001261, -0.00002072, 0.00002368, -0.00002634, 0.00001040, -0.00001282, 0.00001480, -0.00001796, 0.00001959, -0.00002426, 0.00001544, -0.00003496, 0.00004495, -0.00005263, 0.00004670, -0.00003940, 0.00003345, -0.00003684, 0.00003423, -0.00004713, 0.00004075, -0.00005522, 0.00004366, -0.00006574, 0.00005896, -0.00010147, 0.00007837, -0.00017935, 0.00012658, -0.00058557, 0.00016193, 0.00160482, -0.00082569, 0.00003835, -0.00016897, 0.00006066, -0.00010060, 0.00005400, -0.00006426, 0.00003720, -0.00004931, 0.00002971, -0.00003924, 0.00002974, -0.00003527, 0.00003191, -0.00003166, 0.00001992, -0.00003699, 0.00004046, -0.00004140, 0.00003886, -0.00002926, 0.00001820, -0.00001285, 0.00001849, -0.00000813, -0.00000123, -0.00000634, 0.00000126, 0.00000343, 0.00000552, 0.00000725, -0.00000974, -0.00000166, 0.00002469, -0.00003341, 0.00000968, -0.00000343, 0.00000457, -0.00000906, 0.00000132, -0.00000327, 0.00001184, -0.00001091, 0.00000769, -0.00001031, 0.00000569, -0.00000423, 0.00000536, -0.00001078, 0.00000164, -0.00000721, 0.00001150, -0.00000059, 0.00000529, },
		{ 0.00001297, -0.00000825, -0.00000136, -0.00000770, 0.00002328, -0.00001292, -0.00000340, -0.00000095, 0.00000393, -0.00000447, 0.00000631, 0.00000554, -0.00002010, 0.00000932, 0.00000868, 0.00000498, -0.00001874, 0.00000667, -0.00000481, 0.00001376, -0.00000427, 0.00000147, -0.00000558, 0.00000031, -0.00000596, 0.00000292, 0.00000334, 0.00000095, -0.00000104, 0.00000285, -0.00000443, 0.00000043, 0.00000627, -0.00000864, 0.00000943, 0.00000045, 0.00000078, 0.00001148, -0.00003922, 0.00002321, -0.00000261, 0.00000152, -0.00000998, 0.00001140, -0.00000939, -0.00000710, 0.00000030, 0.00000333, -0.00000611, 0.00000741, -0.00000963, 0.00000162, -0.00002193, 0.00002794, -0.00002939, 0.00001906, -0.00001077, 0.00000479, -0.00000735, 0.00000358, -0.00001281, 0.00000281, -0.00001294, -0.00000203, -0.00001646, 0.00000487, -0.00003427, -0.00000065, -0.00007054, -0.00001243, -0.00025935, -0.00048005, 0.00194344, -0.00062281, -0.00016977, -0.00006132, -0.00002685, -0.00003446, -0.00000604, -0.00001354, -0.00000579, -0.00001319, -0.00000204, -0.00001208, 0.00000461, -0.00001222, 0.00000878, -0.00000907, 0.00000191, -0.00002164, 0.00002368, -0.00002235, 0.00001791, -0.00000843, -0.00000244, 0.00000503, 0.00000042, 0.00000972, -0.00001360, 0.00000202, -0.00000713, 0.00000852, 0.00000167, 0.00000758, -0.00000369, -0.00001085, 0.00002928, -0.00002727, -0.00000156, 0.00000611, -0.00000179, -0.00000357, -0.00000254, -0.00000188, 0.00000950, -0.00000651, 0.00000334, -0.00000566, 0.00000134, -0.00000121, 0.00000421, -0.00000950, 0.00000119, -0.00000823, 0.00000947, 0.00000154, 0.00000768, -0.00001870, },
		{ -0.00000045, -0.00000574, -0.00000450, 0.00001433, 0.00000065, -0.00001380, 0.00000564, -0.00000228, 0.00000319, -0.00000419, 0.00001520, -0.00002241, 0.00000652, 0.00000747, 0.00000893, -0.00001749, 0.00000253, -0.00000363, 0.00001089, 0.00000232, -0.00000464, 0.00000070, -0.00000593, -0.00000008, -0.00000385, 0.00000895, -0.00000406, 0.00000423, -0.00000346, 0.00000402, -0.00000917, 0.00001471, -0.00001433, 0.00001074, -0.00000044, 0.00000304, 0.00000673, -0.00002339, -0.00000235, 0.00002198, -0.00001881, 0.00000979, -0.00001015, 0.00001575, -0.00003029, 0.00001857, -0.00001490, 0.00001398, -0.00001422, 0.00001489, -0.00002207, 0.00000221, -0.00000168, 0.00000657, -0.00002037, 0.00002942, -0.00003636, 0.00003596, -0.00004246, 0.00003819, -0.00005265, 0.00004767, -0.00006722, 0.00005482, -0.00007584, 0.00006439, -0.00011694, 0.00008208, -0.00020799, 0.00010092, -0.00098238, 0.00194866, -0.00012949, -0.00051905, 0.00012313, -0.00017185, 0.00007679, -0.00010302, 0.00006672, -0.00007356, 0.00004501, -0.00005412, 0.00003422, -0.00003873, 0.00002797, -0.00002990, 0.00002660, -0.00002774, 0.00000555, -0.00000517, 0.00000797, -0.00001289, 0.00002120, -0.00003064, 0.00002924, -0.00002266, 0.00003026, -0.00002751, 0.00001308, -0.00001888, 0.00001698, -0.00000584, 0.00001251, -0.00000535, -0.00000765, 0.00001797, -0.00000595, -0.00002404, 0.00002309, -0.00001377, 0.00000706, -0.00001148, 0.00000540, 0.00000065, 0.00000381, -0.00000602, 0.00000321, -0.00000642, 0.00000469, 0.00000003, -0.00000468, -0.00000327, -0.00000418, 0.00000237, 0.00000708, 0.00000649, -0.00001505, 0.00000527, },
		{ -0.00001221, 0.00000240, 0.00000213, 0.00001430, -0.00002192, 0.00001004, -0.00000707, 0.00000927, -0.00001110, 0.00001858, -0.00001843, 0.00000020, 0.00000851, 0.00000915, -0.00001223, -0.00000386, -0.00000031, 0.00000458, 0.00001175, -0.00001214, 0.00000755, -0.00001216, 0.00000564, -0.00001046, 0.00001458, -0.00000966, 0.00001074, -0.00001146, 0.00001380, -0.00001905, 0.00002269, -0.00001980, 0.00001412, -0.00000684, 0.00001353, -0.00000817, 0.00000138, -0.00003176, 0.00004575, -0.00003672, 0.00002719, -0.00002928, 0.00003694, -0.00004746, 0.00003144, -0.00002982, 0.00003179, -0.00003388, 0.00003689, -0.00004266, 0.00002405, -0.00002958, 0.00003978, -0.00005548, 0.00006467, -0.00007270, 0.00007438, -0.00008325, 0.00008298, -0.00010038, 0.00009894, -0.00012164, 0.00011462, -0.00014471, 0.00014850, -0.00021382, 0.00020602, -0.00035932, 0.00035119, -0.00119385, 0.00155322, 0.00061339, -0.00093021, 0.00031720, -0.00032135, 0.00018887, -0.00019883, 0.00014358, -0.00013792, 0.00010084, -0.00010477, 0.00007982, -0.00008172, 0.00006763, -0.00006730, 0.00005986, -0.00005531, 0.00003229, -0.00003451, 0.00003850, -0.00004320, 0.00004991, -0.00005710, 0.00005096, -0.00004265, 0.00004677, -0.00003828, 0.00002362, -0.00003134, 0.00002684, -0.00001600, 0.00002201, -0.00001489, 0.00000662, -0.00000482, 0.00002302, -0.00004886, 0.00003912, -0.00002434, 0.00001705, -0.00002059, 0.00001390, -0.00000978, 0.00001490, -0.00001545, 0.00001179, -0.00001365, 0.00001030, -0.00000474, 0.00000165, -0.00000985, 0.00000271, -0.00000693, 0.00001339, 0.00000354, -0.00000776, -0.00000561, 0.00000913, },
		{ 0.00000827, -0.00000749, 0.00002229, -0.00002336, 0.00000905, -0.00000753, 0.00001070, -0.00001174, 0.00001464, -0.00000912, -0.00000802, 0.00001065, 0.00000701, -0.00000517, -0.00001016, 0.00000316, -0.00000264, 0.00002048, -0.00001713, 0.00001129, -0.00001504, 0.00000821, -0.00001411, 0.00001746, -0.00001293, 0.00001512, -0.00001711, 0.00002024, -0.00002446, 0.00002586, -0.00002129, 0.00001584, -0.00001365, 0.00002455, -0.00002222, 0.00002107, -0.00004994, 0.00005459, -0.00003958, 0.00003038, -0.00003420, 0.00004243, -0.00004781, 0.00002918, -0.00003157, 0.00003645, -0.00003958, 0.00004353, -0.00004695, 0.00002996, -0.00004112, 0.00005462, -0.00007001, 0.00007804, -0.00008603, 0.00008793, -0.00009682, 0.00009752, -0.00011434, 0.00011283, -0.00013493, 0.00012942, -0.00016328, 0.00017290, -0.00023757, 0.00023638, -0.00038478, 0.00039595, -0.00105831, 0.00081491, 0.00145557, -0.00125467, 0.00041434, -0.00039726, 0.00023954, -0.00024200, 0.00017387, -0.00016235, 0.00012255, -0.00012552, 0.00009867, -0.00010018, 0.00008421, -0.00008261, 0.00007205, -0.00006436, 0.00004360, -0.00005007, 0.00005558, -0.00006011, 0.00006546, -0.00007014, 0.00005937, -0.00004988, 0.00005124, -0.00003964, 0.00002798, -0.00003874, 0.00003244, -0.00002342, 0.00003062, -0.00002572, 0.00002285, -0.00002703, 0.00004593, -0.00006295, 0.00004327, -0.00002480, 0.00001912, -0.00002322, 0.00001709, -0.00001506, 0.00002019, -0.00001893, 0.00001474, -0.00001578, 0.00001173, -0.00000667, 0.00000610, -0.00001491, 0.00000826, -0.00001388, 0.00001647, 0.00000206, -0.00000044, -0.00001535, 0.00001621, -0.00001676, },
		{ -0.00001027, 0.00002142, -0.00001673, 0.00000231, -0.00000343, 0.00000727, -0.00000672, 0.00000576, 0.00000184, -0.00001475, 0.00001138, 0.00000509, 0.00000058, -0.00001354, 0.00000431, -0.00000768, 0.00002474, -0.00001612, 0.00000907, -0.00001224, 0.00000559, -0.00001276, 0.00001533, -0.00001116, 0.00001408, -0.00001652, 0.00001913, -0.00002132, 0.00002054, -0.00001525, 0.00001149, -0.00001439, 0.00002708, -0.00002419, 0.00002356, -0.00004609, 0.00004036, -0.00002135, 0.00001420, -0.00001993, 0.00002766, -0.00002785, 0.00000889, -0.00001647, 0.00002309, -0.00002559, 0.00002874, -0.00002894, 0.00001375, -0.00002937, 0.00004348, -0.00005608, 0.00006120, -0.00006702, 0.00006613, -0.00007173, 0.00006964, -0.00008186, 0.00007621, -0.00009332, 0.00008415, -0.00011426, 0.00011774, -0.00016629, 0.00014995, -0.00026160, 0.00022855, -0.00064055, -0.00008807, 0.00217129, -0.00132060, 0.00031176, -0.00033234, 0.00017865, -0.00019317, 0.00012603, -0.00012092, 0.00008766, -0.00009616, 0.00007222, -0.00007701, 0.00006211, -0.00006189, 0.00005114, -0.00004443, 0.00002963, -0.00004204, 0.00004955, -0.00005438, 0.00005891, -0.00006127, 0.00004683, -0.00003778, 0.00003831, -0.00002694, 0.00002104, -0.00003506, 0.00002733, -0.00002130, 0.00003063, -0.00002833, 0.00002925, -0.00003524, 0.00004990, -0.00005599, 0.00002829, -0.00001011, 0.00000908, -0.00001524, 0.00001056, -0.00001060, 0.00001540, -0.00001283, 0.00000905, -0.00001026, 0.00000676, -0.00000346, 0.00000596, -0.00001519, 0.00000872, -0.00001469, 0.00001329, 0.00000463, 0.00000366, -0.00002005, 0.00001722, -0.00001604, 0.00000922, },
		{ 0.00001273, -0.00000463, -0.00000720, 0.00000262, 0.00000154, 0.00000059, -0.00000352, 0.00000978, -0.00001665, 0.00000855, 0.00000554, 0.00000273, -0.00001234, 0.00000183, -0.00000864, 0.00002254, -0.00000802, 0.00000071, -0.00000404, -0.00000187, -0.00000655, 0.00000796, -0.00000372, 0.00000661, -0.00000843, 0.00000932, -0.00000898, 0.00000645, -0.00000121, -0.00000050, -0.00000570, 0.00001646, -0.00000932, 0.00000562, -0.00002020, 0.00000632, 0.00001293, -0.00001589, 0.00000810, -0.00000156, 0.00000579, -0.00002280, 0.00000998, -0.00000380, 0.00000376, -0.00000278, 0.00000606, -0.00001917, 0.00000057, 0.00001143, -0.00001906, 0.00001966, -0.00002110, 0.00001462, -0.00001434, 0.00000680, -0.00001189, -0.00000036, -0.00000892, -0.00000765, -0.00001272, 0.00000101, -0.00002330, -0.00002107, -0.00003731, -0.00006779, -0.00010308, -0.00091152, 0.00253257, -0.00100262, -0.00003508, -0.00011089, -0.00000039, -0.00005043, 0.00000087, -0.00001569, -0.00000187, -0.00001818, 0.00000170, -0.00001343, 0.00000283, -0.00000722, -0.00000020, 0.00000178, -0.00000775, -0.00001126, 0.00002066, -0.00002602, 0.00003007, -0.00003039, 0.00001374, -0.00000727, 0.00000918, -0.00000089, 0.00000197, -0.00001810, 0.00000870, -0.00000616, 0.00001740, -0.00001657, 0.00001847, -0.00002243, 0.00003031, -0.00002686, -0.00000412, 0.00001702, -0.00001074, 0.00000181, -0.00000483, 0.00000301, 0.00000139, 0.00000157, -0.00000383, 0.00000148, -0.00000346, 0.00000433, 0.00000101, -0.00000991, 0.00000300, -0.00000850, 0.00000380, 0.00001124, 0.00000359, -0.00001841, 0.00001139, -0.00000955, 0.00000438, -0.00000569, },
		{ 0.00000756, -0.00001473, 0.00000679, -0.00000288, 0.00000614, -0.00000893, 0.00001148, -0.00001246, 0.00000203, 0.00000858, 0.00000119, -0.00000726, -0.00000337, -0.00000606, 0.00001486, 0.00000493, -0.00001074, 0.00000635, -0.00001111, 0.00000183, -0.00000220, 0.00000706, -0.00000494, 0.00000457, -0.00000599, 0.00000862, -0.00001209, 0.00001677, -0.00001691, 0.00000989, -0.00000432, 0.00001748, -0.00002467, 0.00001600, -0.00003330, 0.00004850, -0.00004575, 0.00003623, -0.00003135, 0.00003883, -0.00005216, 0.00003527, -0.00003202, 0.00003568, -0.00003721, 0.00004328, -0.00005366, 0.00003353, -0.00002574, 0.00002427, -0.00002892, 0.00003308, -0.00004631, 0.00005282, -0.00006581, 0.00006753, -0.00008585, 0.00008408, -0.00010829, 0.00009914, -0.00012818, 0.00013218, -0.00020304, 0.00019279, -0.00035872, 0.00035745, -0.00143034, 0.00239479, -0.00028683, -0.00057206, 0.00020728, -0.00024532, 0.00014034, -0.00016101, 0.00011770, -0.00011510, 0.00008123, -0.00008838, 0.00006796, -0.00007244, 0.00006138, -0.00006313, 0.00005720, -0.00005375, 0.00002945, -0.00001942, 0.00001411, -0.00001083, 0.00001249, -0.00002986, 0.00003189, -0.00002709, 0.00003092, -0.00002365, 0.00000778, -0.00001946, 0.00001856, -0.00000648, 0.00000766, -0.00000735, 0.00000735, -0.00000556, 0.00001421, -0.00004262, 0.00004602, -0.00003151, 0.00002063, -0.00002267, 0.00001972, -0.00001568, 0.00001811, -0.00001814, 0.00001427, -0.00001452, 0.00001317, -0.00000599, -0.00000141, -0.00000655, 0.00000183, -0.00000858, 0.00001890, 0.00000130, -0.00001229, 0.00000128, 0.00000000, -0.00000371, 0.00000307, 0.00000089, },
		{ -0.00001637, 0.00000636, -0.00000360, 0.00000750, -0.00000864, 0.00000690, -0.00000403, -0.00000576, 0.00001216, -0.00000180, -0.00000117, -0.00000842, -0.00000256, 0.00000517, 0.00001802, -0.00002017, 0.00001405, -0.00001767, 0.00000819, -0.00001098, 0.00001677, -0.00001570, 0.00001691, -0.00002037, 0.00002426, -0.00002758, 0.00003136, -0.00003064, 0.00002483, -0.00002584, 0.00004389, -0.00005177, 0.00004451, -0.00005996, 0.00006805, -0.00005959, 0.00004947, -0.00004674, 0.00005631, -0.00006508, 0.00004608, -0.00004782, 0.00005510, -0.00005819, 0.00006545, -0.00007220, 0.00005165, -0.00004917, 0.00005368, -0.00006293, 0.00007218, -0.00009088, 0.00010111, -0.00011667, 0.00012193, -0.00014279, 0.00014497, -0.00017321, 0.00017161, -0.00021261, 0.00023278, -0.00031555, 0.00033049, -0.00051899, 0.00058715, -0.00151991, 0.00175909, 0.00069999, -0.00114376, 0.00049632, -0.00045789, 0.00029864, -0.00029195, 0.00022178, -0.00020276, 0.00015869, -0.00015894, 0.00013141, -0.00013067, 0.00011336, -0.00010972, 0.00009685, -0.00008637, 0.00006020, -0.00005207, 0.00004825, -0.00004674, 0.00005066, -0.00006754, 0.00006409, -0.00005625, 0.00005608, -0.00004515, 0.00003242, -0.00004681, 0.00004268, -0.00003115, 0.00003427, -0.00003677, 0.00004021, -0.00004122, 0.00004965, -0.00007074, 0.00006299, -0.00004199, 0.00003162, -0.00003412, 0.00003089, -0.00002725, 0.00002867, -0.00002652, 0.00002162, -0.00002076, 0.00001823, -0.00001072, 0.00000599, -0.00001528, 0.00001113, -0.00001847, 0.00002309, 0.00000035, -0.00000554, -0.00000855, 0.00000811, -0.00001072, 0.00001100, -0.00000829, 0.00001446, },
		{ 0.00000148, -0.00000088, 0.00000509, -0.00000407, -0.00000084, 0.00000438, -0.00001095, 0.00001319, -0.00000309, 0.00000236, -0.00001008, -0.00000126, -0.00000243, 0.00002625, -0.00002283, 0.00001497, -0.00001789, 0.00000898, -0.00001453, 0.00002099, -0.00002057, 0.00002272, -0.00002715, 0.00003079, -0.00003285, 0.00003551, -0.00003442, 0.00003067, -0.00003731, 0.00005660, -0.00006067, 0.00005011, -0.00006021, 0.00006075, -0.00004853, 0.00003967, -0.00003955, 0.00005020, -0.00005429, 0.00003539, -0.00004302, 0.00005238, -0.00005505, 0.00006146, -0.00006373, 0.00004358, -0.00004645, 0.00005543, -0.00006732, 0.00007947, -0.00010019, 0.00010951, -0.00012296, 0.00012686, -0.00014520, 0.00014608, -0.00017253, 0.00017152, -0.00021437, 0.00023509, -0.00030927, 0.00032002, -0.00048612, 0.00053207, -0.00119982, 0.00078277, 0.00171550, -0.00153713, 0.00060936, -0.00053350, 0.00034355, -0.00032676, 0.00024283, -0.00021950, 0.00017451, -0.00017426, 0.00014440, -0.00014219, 0.00012182, -0.00011588, 0.00009956, -0.00008685, 0.00006356, -0.00006041, 0.00005990, -0.00006147, 0.00006801, -0.00008360, 0.00007468, -0.00006504, 0.00006280, -0.00005156, 0.00004436, -0.00006112, 0.00005384, -0.00004406, 0.00004986, -0.00005479, 0.00005928, -0.00005896, 0.00006298, -0.00007499, 0.00005785, -0.00003452, 0.00002810, -0.00003262, 0.00002991, -0.00002679, 0.00002732, -0.00002387, 0.00001912, -0.00001838, 0.00001604, -0.00000969, 0.00000835, -0.00001873, 0.00001456, -0.00002124, 0.00002027, 0.00000364, -0.00000169, -0.00001401, 0.00001099, -0.00001310, 0.00001410, -0.00001081, 0.00001343, -0.00001133, },
		{ 0.00000268, 0.00000158, 0.00000125, -0.00000716, 0.00000833, -0.00001041, 0.00000960, -0.00000057, 0.00000121, -0.00000674, -0.00000376, -0.00000544, 0.00002686, -0.00001682, 0.00000804, -0.00001103, 0.00000334, -0.00001145, 0.00001770, -0.00001697, 0.00001894, -0.00002293, 0.00002485, -0.00002490, 0.00002638, -0.00002501, 0.00002294, -0.00003257, 0.00004849, -0.00004478, 0.00002841, -0.00003296, 0.00002845, -0.00001581, 0.00001016, -0.00001281, 0.00002364, -0.00002344, 0.00000647, -0.00001945, 0.00002820, -0.00002820, 0.00003205, -0.00002951, 0.00001053, -0.00001801, 0.00002917, -0.00004106, 0.00005299, -0.00007132, 0.00007470, -0.00008158, 0.00007985, -0.00009138, 0.00008639, -0.00010570, 0.00009842, -0.00013303, 0.00013987, -0.00018830, 0.00017195, -0.00028358, 0.00025176, -0.00062153, -0.00026968, 0.00247697, -0.00155907, 0.00044317, -0.00041189, 0.00023406, -0.00023577, 0.00015856, -0.00014760, 0.00011333, -0.00012058, 0.00009463, -0.00009618, 0.00007754, -0.00007400, 0.00005917, -0.00004979, 0.00003389, -0.00003811, 0.00004227, -0.00004784, 0.00005701, -0.00007068, 0.00005717, -0.00004786, 0.00004605, -0.00003748, 0.00003685, -0.00005460, 0.00004403, -0.00003685, 0.00004518, -0.00005101, 0.00005363, -0.00004895, 0.00004732, -0.00005214, 0.00003020, -0.00000983, 0.00001043, -0.00001780, 0.00001591, -0.00001345, 0.00001364, -0.00001025, 0.00000705, -0.00000762, 0.00000661, -0.00000232, 0.00000441, -0.00001511, 0.00001020, -0.00001550, 0.00001006, 0.00001137, -0.00000187, -0.00001359, 0.00000767, -0.00001009, 0.00001144, -0.00000635, 0.00000585, -0.00000237, -0.00000488, },
		{ -0.00000002, 0.00000391, -0.00000865, 0.00000562, -0.00000379, 0.00000196, 0.00000527, -0.00000400, 0.00000038, -0.00000897, -0.00000428, 0.00002064, -0.00000416, -0.00000393, 0.00000005, -0.00000626, -0.00000368, 0.00000846, -0.00000631, 0.00000703, -0.00000947, 0.00000879, -0.00000676, 0.00000732, -0.00000550, 0.00000402, -0.00001361, 0.00002252, -0.00000957, -0.00001165, 0.00000949, -0.00001469, 0.00002431, -0.00002562, 0.00002071, -0.00001064, 0.00001460, -0.00002854, 0.00001199, -0.00000681, 0.00001104, -0.00001041, 0.00001715, -0.00003400, 0.00002295, -0.00001193, 0.00000249, 0.00000636, -0.00001854, 0.00001255, -0.00001069, 0.00000160, -0.00000471, -0.00000795, -0.00000177, -0.00001537, -0.00000526, -0.00000930, -0.00000664, -0.00004451, -0.00000451, -0.00011456, -0.00000573, -0.00111732, 0.00275933, -0.00111732, -0.00000573, -0.00011456, -0.00000451, -0.00004451, -0.00000664, -0.00000930, -0.00000526, -0.00001537, -0.00000177, -0.00000795, -0.00000471, 0.00000160, -0.00001069, 0.00001255, -0.00001854, 0.00000636, 0.00000249, -0.00001193, 0.00002295, -0.00003400, 0.00001715, -0.00001041, 0.00001104, -0.00000681, 0.00001199, -0.00002854, 0.00001460, -0.00001064, 0.00002071, -0.00002562, 0.00002431, -0.00001469, 0.00000949, -0.00001165, -0.00000957, 0.00002252, -0.00001361, 0.00000402, -0.00000550, 0.00000732, -0.00000676, 0.00000879, -0.00000947, 0.00000703, -0.00000631, 0.00000846, -0.00000368, -0.00000626, 0.00000005, -0.00000393, -0.00000416, 0.00002064, -0.00000428, -0.00000897, 0.00000038, -0.00000400, 0.00000527, 0.00000196, -0.00000379, 0.00000562, -0.00000865, 0.00000391, },
		{ 0.00000268, -0.00000488, -0.00000237, 0.00000585, -0.00000635, 0.00001144, -0.00001009, 0.00000767, -0.00001359, -0.00000187, 0.00001137, 0.00001006, -0.00001550, 0.00001020, -0.00001511, 0.00000441, -0.00000232, 0.00000661, -0.00000762, 0.00000705, -0.00001025, 0.00001364, -0.00001345, 0.00001591, -0.00001780, 0.00001043, -0.00000983, 0.00003020, -0.00005214, 0.00004732, -0.00004895, 0.00005363, -0.00005101, 0.00004518, -0.00003685, 0.00004403, -0.00005460, 0.00003685, -0.00003748, 0.00004605, -0.00004786, 0.00005717, -0.00007068, 0.00005701, -0.00004784, 0.00004227, -0.00003811, 0.00003389, -0.00004979, 0.00005917, -0.00007400, 0.00007754, -0.00009618, 0.00009463, -0.00012058, 0.00011333, -0.00014760, 0.00015856, -0.00023577, 0.00023406, -0.00041189, 0.00044317, -0.00155907, 0.00247697, -0.00026968, -0.00062153, 0.00025176, -0.00028358, 0.00017195, -0.00018830, 0.00013987, -0.00013303, 0.00009842, -0.00010570, 0.00008639, -0.00009138, 0.00007985, -0.00008158, 0.00007470, -0.00007132, 0.00005299, -0.00004106, 0.00002917, -0.00001801, 0.00001053, -0.00002951, 0.00003205, -0.00002820, 0.00002820, -0.00001945, 0.00000647, -0.00002344, 0.00002364, -0.00001281, 0.00001016, -0.00001581, 0.00002845, -0.00003296, 0.00002841, -0.00004478, 0.00004849, -0.00003257, 0.00002294, -0.00002501, 0.00002638, -0.00002490, 0.00002485, -0.00002293, 0.00001894, -0.00001697, 0.00001770, -0.00001145, 0.00000334, -0.00001103, 0.00000804, -0.00001682, 0.00002686, -0.00000544, -0.00000376, -0.00000674, 0.00000121, -0.00000057, 0.00000960, -0.00001041, 0.00000833, -0.00000716, 0.00000125, 0.00000158, },
		{ 0.00000148, -0.00001133, 0.00001343, -0.00001081, 0.00001410, -0.00001310, 0.00001099, -0.00001401, -0.00000169, 0.00000364, 0.00002027, -0.00002124, 0.00001456, -0.00001873, 0.00000835, -0.00000969, 0.00001604, -0.00001838, 0.00001912, -0.00002387, 0.00002732, -0.00002679, 0.00002991, -0.00003262, 0.00002810, -0.00003452, 0.00005785, -0.00007499, 0.00006298, -0.00005896, 0.00005928, -0.00005479, 0.00004986, -0.00004406, 0.00005384, -0.00006112, 0.00004436, -0.00005156, 0.00006280, -0.00006504, 0.00007468, -0.00008360, 0.00006801, -0.00006147, 0.00005990, -0.00006041, 0.00006356, -0.00008685, 0.00009956, -0.00011588, 0.00012182, -0.00014219, 0.00014440, -0.00017426, 0.00017451, -0.00021950, 0.00024283, -0.00032676, 0.00034355, -0.00053350, 0.00060936, -0.00153713, 0.00171550, 0.00078277, -0.00119982, 0.00053207, -0.00048612, 0.00032002, -0.00030927, 0.00023509, -0.00021437, 0.00017152, -0.00017253, 0.00014608, -0.00014520, 0.00012686, -0.00012296, 0.00010951, -0.00010019, 0.00007947, -0.00006732, 0.00005543, -0.00004645, 0.00004358, -0.00006373, 0.00006146, -0.00005505, 0.00005238, -0.00004302, 0.00003539, -0.00005429, 0.00005020, -0.00003955, 0.00003967, -0.00004853, 0.00006075, -0.00006021, 0.00005011, -0.00006067, 0.00005660, -0.00003731, 0.00003067, -0.00003442, 0.00003551, -0.00003285, 0.00003079, -0.00002715, 0.00002272, -0.00002057, 0.00002099, -0.00001453, 0.00000898, -0.00001789, 0.00001497, -0.00002283, 0.00002625, -0.00000243, -0.00000126, -0.00001008, 0.00000236, -0.00000309, 0.00001319, -0.00001095, 0.00000438, -0.00000084, -0.00000407, 0.00000509, -0.00000088, },
		{ -0.00001637, 0.00001446, -0.00000829, 0.00001100, -0.00001072, 0.00000811, -0.00000855, -0.00000554, 0.00000035, 0.00002309, -0.00001847, 0.00001113, -0.00001528, 0.00000599, -0.00001072, 0.00001823, -0.00002076, 0.00002162, -0.00002652, 0.00002867, -0.00002725, 0.00003089, -0.00003412, 0.00003162, -0.00004199, 0.00006299, -0.00007074, 0.00004965, -0.00004122, 0.00004021, -0.00003677, 0.00003427, -0.00003115, 0.00004268, -0.00004681, 0.00003242, -0.00004515, 0.00005608, -0.00005625, 0.00006409, -0.00006754, 0.00005066, -0.00004674, 0.00004825, -0.00005207, 0.00006020, -0.00008637, 0.00009685, -0.00010972, 0.00011336, -0.00013067, 0.00013141, -0.00015894, 0.00015869, -0.00020276, 0.00022178, -0.00029195, 0.00029864, -0.00045789, 0.00049632, -0.00114376, 0.00069999, 0.00175909, -0.00151991, 0.00058715, -0.00051899, 0.00033049, -0.00031555, 0.00023278, -0.00021261, 0.00017161, -0.00017321, 0.00014497, -0.00014279, 0.00012193, -0.00011667, 0.00010111, -0.00009088, 0.00007218, -0.00006293, 0.00005368, -0.00004917, 0.00005165, -0.00007220, 0.00006545, -0.00005819, 0.00005510, -0.00004782, 0.00004608, -0.00006508, 0.00005631, -0.00004674, 0.00004947, -0.00005959, 0.00006805, -0.00005996, 0.00004451, -0.00005177, 0.00004389, -0.00002584, 0.00002483, -0.00003064, 0.00003136, -0.00002758, 0.00002426, -0.00002037, 0.00001691, -0.00001570, 0.00001677, -0.00001098, 0.00000819, -0.00001767, 0.00001405, -0.00002017, 0.00001802, 0.00000517, -0.00000256, -0.00000842, -0.00000117, -0.00000180, 0.00001216, -0.00000576, -0.00000403, 0.00000690, -0.00000864, 0.00000750, -0.00000360, 0.00000636, },
		{ 0.00000756, 0.00000089, 0.00000307, -0.00000371, 0.00000000, 0.00000128, -0.00001229, 0.00000130, 0.00001890, -0.00000858, 0.00000183, -0.00000655, -0.00000141, -0.00000599, 0.00001317, -0.00001452, 0.00001427, -0.00001814, 0.00001811, -0.00001568, 0.00001972, -0.00002267, 0.00002063, -0.00003151, 0.00004602, -0.00004262, 0.00001421, -0.00000556, 0.00000735, -0.00000735, 0.00000766, -0.00000648, 0.00001856, -0.00001946, 0.00000778, -0.00002365, 0.00003092, -0.00002709, 0.00003189, -0.00002986, 0.00001249, -0.00001083, 0.00001411, -0.00001942, 0.00002945, -0.00005375, 0.00005720, -0.00006313, 0.00006138, -0.00007244, 0.00006796, -0.00008838, 0.00008123, -0.00011510, 0.00011770, -0.00016101, 0.00014034, -0.00024532, 0.00020728, -0.00057206, -0.00028683, 0.00239479, -0.00143034, 0.00035745, -0.00035872, 0.00019279, -0.00020304, 0.00013218, -0.00012818, 0.00009914, -0.00010829, 0.00008408, -0.00008585, 0.00006753, -0.00006581, 0.00005282, -0.00004631, 0.00003308, -0.00002892, 0.00002427, -0.00002574, 0.00003353, -0.00005366, 0.00004328, -0.00003721, 0.00003568, -0.00003202, 0.00003527, -0.00005216, 0.00003883, -0.00003135, 0.00003623, -0.00004575, 0.00004850, -0.00003330, 0.00001600, -0.00002467, 0.00001748, -0.00000432, 0.00000989, -0.00001691, 0.00001677, -0.00001209, 0.00000862, -0.00000599, 0.00000457, -0.00000494, 0.00000706, -0.00000220, 0.00000183, -0.00001111, 0.00000635, -0.00001074, 0.00000493, 0.00001486, -0.00000606, -0.00000337, -0.00000726, 0.00000119, 0.00000858, 0.00000203, -0.00001246, 0.00001148, -0.00000893, 0.00000614, -0.00000288, 0.00000679, -0.00001473, },
		{ 0.00001273, -0.00000569, 0.00000438, -0.00000955, 0.00001139, -0.00001841, 0.00000359, 0.00001124, 0.00000380, -0.00000850, 0.00000300, -0.00000991, 0.00000101, 0.00000433, -0.00000346, 0.00000148, -0.00000383, 0.00000157, 0.00000139, 0.00000301, -0.00000483, 0.00000181, -0.00001074, 0.00001702, -0.00000412, -0.00002686, 0.00003031, -0.00002243, 0.00001847, -0.00001657, 0.00001740, -0.00000616, 0.00000870, -0.00001810, 0.00000197, -0.00000089, 0.00000918, -0.00000727, 0.00001374, -0.00003039, 0.00003007, -0.00002602, 0.00002066, -0.00001126, -0.00000775, 0.00000178, -0.00000020, -0.00000722, 0.00000283, -0.00001343, 0.00000170, -0.00001818, -0.00000187, -0.00001569, 0.00000087, -0.00005043, -0.00000039, -0.00011089, -0.00003508, -0.00100262, 0.00253257, -0.00091152, -0.00010308, -0.00006779, -0.00003731, -0.00002107, -0.00002330, 0.00000101, -0.00001272, -0.00000765, -0.00000892, -0.00000036, -0.00001189, 0.00000680, -0.00001434, 0.00001462, -0.00002110, 0.00001966, -0.00001906, 0.00001143, 0.00000057, -0.00001917, 0.00000606, -0.00000278, 0.00000376, -0.00000380, 0.00000998, -0.00002280, 0.00000579, -0.00000156, 0.00000810, -0.00001589, 0.00001293, 0.00000632, -0.00002020, 0.00000562, -0.00000932, 0.00001646, -0.00000570, -0.00000050, -0.00000121, 0.00000645, -0.00000898, 0.00000932, -0.00000843, 0.00000661, -0.00000372, 0.00000796, -0.00000655, -0.00000187, -0.00000404, 0.00000071, -0.00000802, 0.00002254, -0.00000864, 0.00000183, -0.00001234, 0.00000273, 0.00000554, 0.00000855, -0.00001665, 0.00000978, -0.00000352, 0.00000059, 0.00000154, 0.00000262, -0.00000720, -0.00000463, },
		{ -0.00001027, 0.00000922, -0.00001604, 0.00001722, -0.00002005, 0.00000366, 0.00000463, 0.00001329, -0.00001469, 0.00000872, -0.00001519, 0.00000596, -0.00000346, 0.00000676, -0.00001026, 0.00000905, -0.00001283, 0.00001540, -0.00001060, 0.00001056, -0.00001524, 0.00000908, -0.00001011, 0.00002829, -0.00005599, 0.00004990, -0.00003524, 0.00002925, -0.00002833, 0.00003063, -0.00002130, 0.00002733, -0.00003506, 0.00002104, -0.00002694, 0.00003831, -0.00003778, 0.00004683, -0.00006127, 0.00005891, -0.00005438, 0.00004955, -0.00004204, 0.00002963, -0.00004443, 0.00005114, -0.00006189, 0.00006211, -0.00007701, 0.00007222, -0.00009616, 0.00008766, -0.00012092, 0.00012603, -0.00019317, 0.00017865, -0.00033234, 0.00031176, -0.00132060, 0.00217129, -0.00008807, -0.00064055, 0.00022855, -0.00026160, 0.00014995, -0.00016629, 0.00011774, -0.00011426, 0.00008415, -0.00009332, 0.00007621, -0.00008186, 0.00006964, -0.00007173, 0.00006613, -0.00006702, 0.00006120, -0.00005608, 0.00004348, -0.00002937, 0.00001375, -0.00002894, 0.00002874, -0.00002559, 0.00002309, -0.00001647, 0.00000889, -0.00002785, 0.00002766, -0.00001993, 0.00001420, -0.00002135, 0.00004036, -0.00004609, 0.00002356, -0.00002419, 0.00002708, -0.00001439, 0.00001149, -0.00001525, 0.00002054, -0.00002132, 0.00001913, -0.00001652, 0.00001408, -0.00001116, 0.00001533, -0.00001276, 0.00000559, -0.00001224, 0.00000907, -0.00001612, 0.00002474, -0.00000768, 0.00000431, -0.00001354, 0.00000058, 0.00000509, 0.00001138, -0.00001475, 0.00000184, 0.00000576, -0.00000672, 0.00000727, -0.00000343, 0.00000231, -0.00001673, 0.00002142, },
		{ 0.00000827, -0.00001676, 0.00001621, -0.00001535, -0.00000044, 0.00000206, 0.00001647, -0.00001388, 0.00000826, -0.00001491, 0.00000610, -0.00000667, 0.00001173, -0.00001578, 0.00001474, -0.00001893, 0.00002019, -0.00001506, 0.00001709, -0.00002322, 0.00001912, -0.00002480, 0.00004327, -0.00006295, 0.00004593, -0.00002703, 0.00002285, -0.00002572, 0.00003062, -0.00002342, 0.00003244, -0.00003874, 0.00002798, -0.00003964, 0.00005124, -0.00004988, 0.00005937, -0.00007014, 0.00006546, -0.00006011, 0.00005558, -0.00005007, 0.00004360, -0.00006436, 0.00007205, -0.00008261, 0.00008421, -0.00010018, 0.00009867, -0.00012552, 0.00012255, -0.00016235, 0.00017387, -0.00024200, 0.00023954, -0.00039726, 0.00041434, -0.00125467, 0.00145557, 0.00081491, -0.00105831, 0.00039595, -0.00038478, 0.00023638, -0.00023757, 0.00017290, -0.00016328, 0.00012942, -0.00013493, 0.00011283, -0.00011434, 0.00009752, -0.00009682, 0.00008793, -0.00008603, 0.00007804, -0.00007001, 0.00005462, -0.00004112, 0.00002996, -0.00004695, 0.00004353, -0.00003958, 0.00003645, -0.00003157, 0.00002918, -0.00004781, 0.00004243, -0.00003420, 0.00003038, -0.00003958, 0.00005459, -0.00004994, 0.00002107, -0.00002222, 0.00002455, -0.00001365, 0.00001584, -0.00002129, 0.00002586, -0.00002446, 0.00002024, -0.00001711, 0.00001512, -0.00001293, 0.00001746, -0.00001411, 0.00000821, -0.00001504, 0.00001129, -0.00001713, 0.00002048, -0.00000264, 0.00000316, -0.00001016, -0.00000517, 0.00000701, 0.00001065, -0.00000802, -0.00000912, 0.00001464, -0.00001174, 0.00001070, -0.00000753, 0.00000905, -0.00002336, 0.00002229, -0.00000749, },
		{ -0.00001221, 0.00000913, -0.00000561, -0.00000776, 0.00000354, 0.00001339, -0.00000693, 0.00000271, -0.00000985, 0.00000165, -0.00000474, 0.00001030, -0.00001365, 0.00001179, -0.00001545, 0.00001490, -0.00000978, 0.00001390, -0.00002059, 0.00001705, -0.00002434, 0.00003912, -0.00004886, 0.00002302, -0.00000482, 0.00000662, -0.00001489, 0.00002201, -0.00001600, 0.00002684, -0.00003134, 0.00002362, -0.00003828, 0.00004677, -0.00004265, 0.00005096, -0.00005710, 0.00004991, -0.00004320, 0.00003850, -0.00003451, 0.00003229, -0.00005531, 0.00005986, -0.00006730, 0.00006763, -0.00008172, 0.00007982, -0.00010477, 0.00010084, -0.00013792, 0.00014358, -0.00019883, 0.00018887, -0.00032135, 0.00031720, -0.00093021, 0.00061339, 0.00155322, -0.00119385, 0.00035119, -0.00035932, 0.00020602, -0.00021382, 0.00014850, -0.00014471, 0.00011462, -0.00012164, 0.00009894, -0.00010038, 0.00008298, -0.00008325, 0.00007438, -0.00007270, 0.00006467, -0.00005548, 0.00003978, -0.00002958, 0.00002405, -0.00004266, 0.00003689, -0.00003388, 0.00003179, -0.00002982, 0.00003144, -0.00004746, 0.00003694, -0.00002928, 0.00002719, -0.00003672, 0.00004575, -0.00003176, 0.00000138, -0.00000817, 0.00001353, -0.00000684, 0.00001412, -0.00001980, 0.00002269, -0.00001905, 0.00001380, -0.00001146, 0.00001074, -0.00000966, 0.00001458, -0.00001046, 0.00000564, -0.00001216, 0.00000755, -0.00001214, 0.00001175, 0.00000458, -0.00000031, -0.00000386, -0.00001223, 0.00000915, 0.00000851, 0.00000020, -0.00001843, 0.00001858, -0.00001110, 0.00000927, -0.00000707, 0.00001004, -0.00002192, 0.00001430, 0.00000213, 0.00000240, },
		{ -0.00000045, 0.00000527, -0.00001505, 0.00000649, 0.00000708, 0.00000237, -0.00000418, -0.00000327, -0.00000468, 0.00000003, 0.00000469, -0.00000642, 0.00000321, -0.00000602, 0.00000381, 0.00000065, 0.00000540, -0.00001148, 0.00000706, -0.00001377, 0.00002309, -0.00002404, -0.00000595, 0.00001797, -0.00000765, -0.00000535, 0.00001251, -0.00000584, 0.00001698, -0.00001888, 0.00001308, -0.00002751, 0.00003026, -0.00002266, 0.00002924, -0.00003064, 0.00002120, -0.00001289, 0.00000797, -0.00000517, 0.00000555, -0.00002774, 0.00002660, -0.00002990, 0.00002797, -0.00003873, 0.00003422, -0.00005412, 0.00004501, -0.00007356, 0.00006672, -0.00010302, 0.00007679, -0.00017185, 0.00012313, -0.00051905, -0.00012949, 0.00194866, -0.00098238, 0.00010092, -0.00020799, 0.00008208, -0.00011694, 0.00006439, -0.00007584, 0.00005482, -0.00006722, 0.00004767, -0.00005265, 0.00003819, -0.00004246, 0.00003596, -0.00003636, 0.00002942, -0.00002037, 0.00000657, -0.00000168, 0.00000221, -0.00002207, 0.00001489, -0.00001422, 0.00001398, -0.00001490, 0.00001857, -0.00003029, 0.00001575, -0.00001015, 0.00000979, -0.00001881, 0.00002198, -0.00000235, -0.00002339, 0.00000673, 0.00000304, -0.00000044, 0.00001074, -0.00001433, 0.00001471, -0.00000917, 0.00000402, -0.00000346, 0.00000423, -0.00000406, 0.00000895, -0.00000385, -0.00000008, -0.00000593, 0.00000070, -0.00000464, 0.00000232, 0.00001089, -0.00000363, 0.00000253, -0.00001749, 0.00000893, 0.00000747, 0.00000652, -0.00002241, 0.00001520, -0.00000419, 0.00000319, -0.00000228, 0.00000564, -0.00001380, 0.00000065, 0.00001433, -0.00000450, -0.00000574, },
		{ 0.00001297, -0.00001870, 0.00000768, 0.00000154, 0.00000947, -0.00000823, 0.00000119, -0.00000950, 0.00000421, -0.00000121, 0.00000134, -0.00000566, 0.00000334, -0.00000651, 0.00000950, -0.00000188, -0.00000254, -0.00000357, -0.00000179, 0.00000611, -0.00000156, -0.00002727, 0.00002928, -0.00001085, -0.00000369, 0.00000758, 0.00000167, 0.00000852, -0.00000713, 0.00000202, -0.00001360, 0.00000972, 0.00000042, 0.00000503, -0.00000244, -0.00000843, 0.00001791, -0.00002235, 0.00002368, -0.00002164, 0.00000191, -0.00000907, 0.00000878, -0.00001222, 0.00000461, -0.00001208, -0.00000204, -0.00001319, -0.00000579, -0.00001354, -0.00000604, -0.00003446, -0.00002685, -0.00006132, -0.00016977, -0.00062281, 0.00194344, -0.00048005, -0.00025935, -0.00001243, -0.00007054, -0.00000065, -0.00003427, 0.00000487, -0.00001646, -0.00000203, -0.00001294, 0.00000281, -0.00001281, 0.00000358, -0.00000735, 0.00000479, -0.00001077, 0.00001906, -0.00002939, 0.00002794, -0.00002193, 0.00000162, -0.00000963, 0.00000741, -0.00000611, 0.00000333, 0.00000030, -0.00000710, -0.00000939, 0.00001140, -0.00000998, 0.00000152, -0.00000261, 0.00002321, -0.00003922, 0.00001148, 0.00000078, 0.00000045, 0.00000943, -0.00000864, 0.00000627, 0.00000043, -0.00000443, 0.00000285, -0.00000104, 0.00000095, 0.00000334, 0.00000292, -0.00000596, 0.00000031, -0.00000558, 0.00000147, -0.00000427, 0.00001376, -0.00000481, 0.00000667, -0.00001874, 0.00000498, 0.00000868, 0.00000932, -0.00002010, 0.00000554, 0.00000631, -0.00000447, 0.00000393, -0.00000095, -0.00000340, -0.00001292, 0.00002328, -0.00000770, -0.00000136, -0.00000825, },
		{ -0.00001691, 0.00000529, -0.00000059, 0.00001150, -0.00000721, 0.00000164, -0.00001078, 0.00000536, -0.00000423, 0.00000569, -0.00001031, 0.00000769, -0.00001091, 0.00001184, -0.00000327, 0.00000132, -0.00000906, 0.00000457, -0.00000343, 0.00000968, -0.00003341, 0.00002469, -0.00000166, -0.00000974, 0.00000725, 0.00000552, 0.00000343, 0.00000126, -0.00000634, -0.00000123, -0.00000813, 0.00001849, -0.00001285, 0.00001820, -0.00002926, 0.00003886, -0.00004140, 0.00004046, -0.00003699, 0.00001992, -0.00003166, 0.00003191, -0.00003527, 0.00002974, -0.00003924, 0.00002971, -0.00004931, 0.00003720, -0.00006426, 0.00005400, -0.00010060, 0.00006066, -0.00016897, 0.00003835, -0.00082569, 0.00160482, 0.00016193, -0.00058557, 0.00012658, -0.00017935, 0.00007837, -0.00010147, 0.00005896, -0.00006574, 0.00004366, -0.00005522, 0.00004075, -0.00004713, 0.00003423, -0.00003684, 0.00003345, -0.00003940, 0.00004670, -0.00005263, 0.00004495, -0.00003496, 0.00001544, -0.00002426, 0.00001959, -0.00001796, 0.00001480, -0.00001282, 0.00001040, -0.00002634, 0.00002368, -0.00002072, 0.00001261, -0.00001601, 0.00003369, -0.00003782, 0.00000167, 0.00000830, -0.00000444, 0.00001057, -0.00000424, -0.00000006, 0.00000677, -0.00000884, 0.00000535, -0.00000338, 0.00000377, -0.00000052, 0.00000783, -0.00000967, 0.00000400, -0.00000868, 0.00000386, -0.00000640, 0.00001261, -0.00000352, 0.00000787, -0.00001576, -0.00000198, 0.00001135, 0.00000929, -0.00001337, -0.00000654, 0.00001569, -0.00000945, 0.00000783, -0.00000572, 0.00000439, -0.00002109, 0.00002462, -0.00000442, -0.00000132, -0.00001169, 0.00001501, },
		{ 0.00000000, 0.00000076, 0.00000855, -0.00000186, -0.00000114, -0.00000887, 0.00000323, -0.00000348, 0.00000527, -0.00000923, 0.00000562, -0.00000830, 0.00000709, 0.00000152, -0.00000054, -0.00000816, 0.00000343, -0.00000346, 0.00000871, -0.00002600, 0.00000898, 0.00001327, -0.00001718, 0.00000693, 0.00000847, 0.00000024, 0.00000695, -0.00001173, 0.00000778, -0.00001996, 0.00002771, -0.00002064, 0.00002756, -0.00003753, 0.00004592, -0.00004492, 0.00004094, -0.00003599, 0.00002105, -0.00003525, 0.00003361, -0.00003569, 0.00003128, -0.00004158, 0.00003473, -0.00005582, 0.00004642, -0.00007509, 0.00006612, -0.00011032, 0.00007896, -0.00018763, 0.00010363, -0.00079421, 0.00108320, 0.00075881, -0.00074623, 0.00013795, -0.00020120, 0.00008918, -0.00011361, 0.00006732, -0.00007639, 0.00005464, -0.00006536, 0.00004893, -0.00005406, 0.00004010, -0.00004396, 0.00004144, -0.00004822, 0.00005419, -0.00005528, 0.00004239, -0.00003059, 0.00001334, -0.00002341, 0.00001742, -0.00001702, 0.00001477, -0.00001515, 0.00001563, -0.00002875, 0.00002135, -0.00001771, 0.00000993, -0.00001429, 0.00002718, -0.00002068, -0.00001817, 0.00002007, -0.00001031, 0.00001114, -0.00000010, -0.00000418, 0.00000958, -0.00000932, 0.00000457, -0.00000336, 0.00000466, -0.00000247, 0.00001034, -0.00001044, 0.00000435, -0.00000811, 0.00000255, -0.00000490, 0.00000894, -0.00000107, 0.00000709, -0.00001013, -0.00000958, 0.00001328, 0.00000856, -0.00000559, -0.00001645, 0.00001965, -0.00000874, 0.00000713, -0.00000625, 0.00000678, -0.00002153, 0.00001770, 0.00000452, -0.00000472, -0.00001111, 0.00001187, -0.00001068, },
		{ 0.00000356, 0.00000315, 0.00000471, -0.00000434, -0.00000595, -0.00000046, -0.00000040, 0.00000156, -0.00000425, -0.00000053, -0.00000159, -0.00000122, 0.00000873, -0.00000470, -0.00000395, -0.00000204, 0.00000201, 0.00000136, -0.00001339, -0.00000759, 0.00002386, -0.00001813, 0.00000138, 0.00001386, -0.00000344, 0.00001183, -0.00001540, 0.00001365, -0.00002545, 0.00002839, -0.00001951, 0.00002737, -0.00003526, 0.00004128, -0.00003552, 0.00002841, -0.00002233, 0.00000907, -0.00002405, 0.00001925, -0.00002010, 0.00001667, -0.00002700, 0.00002124, -0.00004131, 0.00003120, -0.00005691, 0.00004410, -0.00008127, 0.00005068, -0.00014766, 0.00007725, -0.00063566, 0.00054618, 0.00116159, -0.00067607, 0.00001423, -0.00014279, 0.00004006, -0.00007912, 0.00003754, -0.00005464, 0.00003602, -0.00004813, 0.00003213, -0.00003846, 0.00002590, -0.00003290, 0.00003213, -0.00003977, 0.00004364, -0.00003981, 0.00002337, -0.00001212, -0.00000147, -0.00001045, 0.00000446, -0.00000657, 0.00000576, -0.00000824, 0.00000971, -0.00001861, 0.00000762, -0.00000471, -0.00000237, -0.00000259, 0.00001089, 0.00000243, -0.00003729, 0.00002684, -0.00001059, 0.00000726, 0.00000570, -0.00000733, 0.00001037, -0.00000782, 0.00000266, -0.00000282, 0.00000497, -0.00000338, 0.00001104, -0.00000889, 0.00000226, -0.00000521, -0.00000059, -0.00000218, 0.00000529, 0.00000062, 0.00000588, -0.00000407, -0.00001522, 0.00001247, 0.00000901, 0.00000042, -0.00002107, 0.00001649, -0.00000230, 0.00000244, -0.00000302, 0.00000431, -0.00001583, 0.00000581, 0.00001501, -0.00000813, -0.00000919, 0.00000640, -0.00000312, -0.00000560, },
		{ -0.00000153, 0.00000905, -0.00000510, -0.00000439, -0.00000347, 0.00000263, -0.00000255, 0.00000105, -0.00000657, 0.00000444, -0.00000790, 0.00001326, -0.00000643, -0.00000114, -0.00000656, 0.00000659, -0.00000434, -0.00000502, -0.00001561, 0.00002288, -0.00000853, -0.00001094, 0.00002214, -0.00000808, 0.00001656, -0.00001799, 0.00001685, -0.00002572, 0.00002319, -0.00001335, 0.00002217, -0.00002747, 0.00003047, -0.00001962, 0.00001033, -0.00000420, -0.00000725, -0.00000791, 0.00000023, -0.00000119, -0.00000063, -0.00000954, 0.00000400, -0.00002174, 0.00000934, -0.00003043, 0.00001280, -0.00004312, 0.00001156, -0.00009385, 0.00002379, -0.00045467, 0.00011832, 0.00130473, -0.00039912, -0.00018555, -0.00005464, -0.00002970, -0.00003005, -0.00000386, -0.00002269, 0.00000705, -0.00002111, 0.00000673, -0.00001554, 0.00000555, -0.00001607, 0.00001650, -0.00002391, 0.00002459, -0.00001613, -0.00000181, 0.00001033, -0.00001932, 0.00000517, -0.00001017, 0.00000510, -0.00000487, 0.00000128, -0.00000055, -0.00000397, -0.00000849, 0.00000911, -0.00001499, 0.00000906, -0.00000385, 0.00001943, -0.00004506, 0.00002136, -0.00000145, -0.00000249, 0.00001355, -0.00001025, 0.00001067, -0.00000640, 0.00000163, -0.00000335, 0.00000577, -0.00000409, 0.00001080, -0.00000618, -0.00000069, -0.00000200, -0.00000324, -0.00000061, 0.00000362, 0.00000043, 0.00000521, 0.00000084, -0.00001743, 0.00000833, 0.00001107, 0.00000372, -0.00002009, 0.00000784, 0.00000680, -0.00000308, 0.00000113, 0.00000005, -0.00000815, -0.00000572, 0.00002191, -0.00000776, -0.00000891, 0.00000181, 0.00000245, -0.00000881, 0.00000530, },
		{ 0.00000926, -0.00000221, -0.00000511, -0.00000461, 0.00000387, -0.00000465, 0.00000367, -0.00000907, 0.00000614, -0.00000930, 0.00001186, -0.00000294, -0.00000262, -0.00000658, 0.00000588, -0.00000327, -0.00000587, -0.00001185, 0.00001023, 0.00000881, -0.00002600, 0.00002977, -0.00001125, 0.00001964, -0.00001856, 0.00001707, -0.00002186, 0.00001497, -0.00000604, 0.00001620, -0.00001875, 0.00001867, -0.00000344, -0.00000606, 0.00001062, -0.00001960, 0.00000391, -0.00001308, 0.00001011, -0.00000963, -0.00000037, -0.00000525, -0.00000993, -0.00000458, -0.00001274, -0.00000762, -0.00001913, -0.00001150, -0.00005872, -0.00001212, -0.00031520, -0.00014997, 0.00121496, -0.00001241, -0.00036476, -0.00000208, -0.00007466, -0.00000135, -0.00002896, -0.00000363, -0.00001211, -0.00000275, -0.00001031, -0.00000073, -0.00000718, -0.00000563, 0.00000548, -0.00001100, 0.00000754, 0.00000467, -0.00002187, 0.00002600, -0.00003012, 0.00001383, -0.00001754, 0.00001003, -0.00000990, 0.00000641, -0.00000749, 0.00000637, -0.00001784, 0.00001507, -0.00001956, 0.00001218, -0.00000843, 0.00002270, -0.00003696, 0.00000304, 0.00001507, -0.00001518, 0.00002109, -0.00001203, 0.00001074, -0.00000576, 0.00000206, -0.00000508, 0.00000701, -0.00000462, 0.00000997, -0.00000313, -0.00000326, 0.00000004, -0.00000400, -0.00000123, 0.00000428, -0.00000125, 0.00000478, 0.00000436, -0.00001641, 0.00000207, 0.00001350, 0.00000529, -0.00001561, -0.00000242, 0.00001409, -0.00000564, 0.00000301, -0.00000257, -0.00000271, -0.00001240, 0.00002189, -0.00000194, -0.00001142, -0.00000031, 0.00000443, -0.00000850, 0.00000474, -0.00000355, },
		{ 0.00000313, -0.00000703, -0.00000436, 0.00000322, -0.00000413, 0.00000271, -0.00000716, 0.00000287, -0.00000524, 0.00000504, 0.00000482, -0.00000770, -0.00000236, -0.00000010, 0.00000421, -0.00001423, -0.00000051, -0.00000713, 0.00002562, -0.00003627, 0.00003111, -0.00000925, 0.00001860, -0.00001541, 0.00001354, -0.00001462, 0.00000580, 0.00000012, 0.00001155, -0.00001128, 0.00000854, 0.00000959, -0.00001682, 0.00001819, -0.00002403, 0.00000713, -0.00001626, 0.00000987, -0.00000716, -0.00000226, -0.00000364, -0.00000930, -0.00000625, -0.00000925, -0.00001084, -0.00001579, -0.00001182, -0.00004986, -0.00001924, -0.00023074, -0.00027743, 0.00098284, 0.00035939, -0.00043505, -0.00002889, -0.00007025, -0.00001002, -0.00002523, -0.00000756, -0.00001239, -0.00000160, -0.00001122, -0.00000078, -0.00000657, -0.00000647, 0.00000368, -0.00000594, -0.00000183, 0.00001615, -0.00003040, 0.00002918, -0.00002890, 0.00001106, -0.00001391, 0.00000524, -0.00000664, 0.00000416, -0.00000726, 0.00000780, -0.00001622, 0.00001002, -0.00001353, 0.00000461, -0.00000147, 0.00001290, -0.00001698, -0.00002129, 0.00003157, -0.00002462, 0.00002419, -0.00001061, 0.00000965, -0.00000515, 0.00000291, -0.00000672, 0.00000749, -0.00000417, 0.00000827, 0.00000007, -0.00000501, 0.00000057, -0.00000294, -0.00000340, 0.00000594, -0.00000273, 0.00000342, 0.00000720, -0.00001360, -0.00000407, 0.00001431, 0.00000694, -0.00001054, -0.00001029, 0.00001610, -0.00000311, 0.00000116, -0.00000194, -0.00000144, -0.00001283, 0.00001523, 0.00000768, -0.00001517, -0.00000081, 0.00000364, -0.00000578, 0.00000260, -0.00000304, 0.00000583, },
		{ -0.00000792, -0.00000428, 0.00000185, -0.00000205, -0.00000053, -0.00000263, -0.00000299, 0.00000127, -0.00000353, 0.00001293, -0.00001272, 0.00000265, -0.00000755, 0.00001328, -0.00002355, 0.00000985, -0.00001948, 0.00003296, -0.00003521, 0.00002236, -0.00000014, 0.00001233, -0.00000790, 0.00000629, -0.00000500, -0.00000298, 0.00000451, 0.00000815, -0.00000481, 0.00000008, 0.00001911, -0.00002173, 0.00001899, -0.00002160, 0.00000323, -0.00001149, 0.00000156, 0.00000214, -0.00000998, 0.00000342, -0.00001446, -0.00000112, -0.00001417, -0.00000372, -0.00002427, -0.00000038, -0.00005412, -0.00001236, -0.00017989, -0.00032061, 0.00071458, 0.00061593, -0.00035656, -0.00013263, -0.00002704, -0.00004608, -0.00000169, -0.00002684, 0.00000017, -0.00001275, -0.00000059, -0.00001113, 0.00000283, -0.00001435, 0.00000760, -0.00000638, -0.00000465, 0.00001876, -0.00002813, 0.00002143, -0.00001783, -0.00000057, -0.00000237, -0.00000587, 0.00000180, -0.00000308, -0.00000145, 0.00000178, -0.00000612, -0.00000238, -0.00000121, -0.00000882, 0.00001116, -0.00000218, 0.00000481, -0.00004095, 0.00003916, -0.00002480, 0.00001960, -0.00000460, 0.00000647, -0.00000332, 0.00000258, -0.00000661, 0.00000590, -0.00000194, 0.00000533, 0.00000351, -0.00000613, 0.00000017, -0.00000119, -0.00000545, 0.00000661, -0.00000222, 0.00000013, 0.00001006, -0.00001046, -0.00000821, 0.00001226, 0.00000974, -0.00000682, -0.00001364, 0.00001214, 0.00000366, -0.00000310, 0.00000089, -0.00000326, -0.00000905, 0.00000550, 0.00001690, -0.00001683, -0.00000209, 0.00000243, -0.00000312, 0.00000083, -0.00000164, 0.00000102, 0.00000822, },
		{ -0.00000570, 0.00000107, -0.00000006, -0.00000377, 0.00000154, -0.00000787, 0.00000612, -0.00000935, 0.00001699, -0.00001369, 0.00000443, -0.00001181, 0.00001823, -0.00002694, 0.00001169, -0.00002000, 0.00002650, -0.00002162, 0.00000462, 0.00001403, 0.00000276, 0.00000228, -0.00000290, 0.00000510, -0.00001016, 0.00000730, 0.00000492, 0.00000185, -0.00000778, 0.00002622, -0.00002266, 0.00001604, -0.00001624, -0.00000313, -0.00000456, -0.00000765, 0.00001021, -0.00001520, 0.00000755, -0.00001683, 0.00000173, -0.00001733, 0.00000176, -0.00003008, 0.00000591, -0.00005222, -0.00001335, -0.00013319, -0.00033571, 0.00048800, 0.00071635, -0.00015377, -0.00026459, 0.00001568, -0.00008022, 0.00001833, -0.00004269, 0.00001021, -0.00002296, 0.00000939, -0.00002038, 0.00001053, -0.00001989, 0.00000930, -0.00000581, -0.00000635, 0.00001770, -0.00002091, 0.00000945, -0.00000399, -0.00001382, 0.00000968, -0.00001610, 0.00000899, -0.00000964, 0.00000473, -0.00000607, 0.00000553, -0.00001430, 0.00000945, -0.00001992, 0.00002061, -0.00001265, 0.00001804, -0.00004719, 0.00003256, -0.00001376, 0.00000756, 0.00000503, 0.00000163, 0.00000012, 0.00000023, -0.00000397, 0.00000189, 0.00000203, 0.00000139, 0.00000704, -0.00000684, -0.00000039, -0.00000001, -0.00000592, 0.00000494, 0.00000102, -0.00000492, 0.00001277, -0.00000761, -0.00000975, 0.00000763, 0.00001323, -0.00000455, -0.00001296, 0.00000446, 0.00001124, -0.00000653, 0.00000312, -0.00000524, -0.00000493, -0.00000251, 0.00002133, -0.00001350, -0.00000613, 0.00000289, -0.00000259, 0.00000103, -0.00000098, -0.00000277, 0.00001049, -0.00000592, },
		{ 0.00000124, 0.00000085, -0.00000534, 0.00000314, -0.00000924, 0.00000662, -0.00000982, 0.00001485, -0.00000896, 0.00000125, -0.00001056, 0.00001612, -0.00002131, 0.00000276, -0.00000862, 0.00000918, -0.00000088, -0.00001580, 0.00002754, -0.00000562, 0.00001150, -0.00001088, 0.00001308, -0.00001444, 0.00000868, 0.00000114, 0.00000927, -0.00001542, 0.00003163, -0.00002135, 0.00001228, -0.00001160, -0.00000767, -0.00000063, -0.00001185, 0.00001102, -0.00001234, 0.00000351, -0.00001118, -0.00000340, -0.00001212, -0.00000243, -0.00002373, -0.00000324, -0.00003330, -0.00003359, -0.00007520, -0.00035261, 0.00033085, 0.00068413, 0.00009909, -0.00035381, 0.00001332, -0.00008330, 0.00001327, -0.00003853, 0.00000435, -0.00002052, 0.00000787, -0.00001857, 0.00000763, -0.00001549, 0.00000252, 0.00000101, -0.00001178, 0.00001827, -0.00001507, 0.00000032, 0.00000537, -0.00002154, 0.00001527, -0.00001909, 0.00000944, -0.00001063, 0.00000640, -0.00001004, 0.00001192, -0.00001874, 0.00001191, -0.00002236, 0.00002053, -0.00001229, 0.00001765, -0.00003795, 0.00001337, 0.00000460, -0.00000751, 0.00001468, -0.00000284, 0.00000424, -0.00000354, 0.00000040, -0.00000343, 0.00000653, -0.00000249, 0.00000997, -0.00000700, -0.00000074, 0.00000001, -0.00000437, 0.00000106, 0.00000603, -0.00001018, 0.00001411, -0.00000456, -0.00000953, 0.00000218, 0.00001572, -0.00000237, -0.00001053, -0.00000312, 0.00001561, -0.00000598, 0.00000225, -0.00000474, -0.00000369, -0.00000550, 0.00001892, -0.00000479, -0.00001278, 0.00000521, -0.00000446, 0.00000318, -0.00000141, -0.00000446, 0.00000903, -0.00000090, -0.00000855, },
		{ 0.00000091, -0.00000501, 0.00000181, -0.00000689, 0.00000290, -0.00000552, 0.00000782, -0.00000030, -0.00000529, -0.00000509, 0.00000833, -0.00000888, -0.00001296, 0.00000823, -0.00001031, 0.00001762, -0.00003049, 0.00003389, -0.00000814, 0.00001617, -0.00001463, 0.00001674, -0.00001490, 0.00000865, -0.00000299, 0.00001665, -0.00002182, 0.00003473, -0.00001818, 0.00000885, -0.00000903, -0.00000894, -0.00000132, -0.00000954, 0.00000396, -0.00000186, -0.00000755, 0.00000111, -0.00001521, 0.00000050, -0.00001562, -0.00000612, -0.00002568, -0.00000118, -0.00006733, -0.00001172, -0.00036992, 0.00022639, 0.00058380, 0.00031412, -0.00034272, -0.00005670, -0.00004596, -0.00002197, -0.00001156, -0.00001929, -0.00000348, -0.00000730, -0.00000387, -0.00000708, -0.00000062, -0.00001272, 0.00001396, -0.00002136, 0.00002200, -0.00001334, -0.00000267, 0.00000709, -0.00002087, 0.00001189, -0.00001309, 0.00000216, -0.00000517, 0.00000222, -0.00000790, 0.00001026, -0.00001346, 0.00000488, -0.00001542, 0.00001067, -0.00000178, 0.00000621, -0.00001881, -0.00001046, 0.00002212, -0.00001885, 0.00001981, -0.00000438, 0.00000739, -0.00000707, 0.00000462, -0.00000802, 0.00000969, -0.00000486, 0.00001136, -0.00000619, -0.00000108, -0.00000083, -0.00000160, -0.00000359, 0.00001072, -0.00001337, 0.00001261, -0.00000046, -0.00000900, -0.00000196, 0.00001544, 0.00000116, -0.00000872, -0.00000747, 0.00001434, -0.00000045, -0.00000207, -0.00000125, -0.00000604, -0.00000334, 0.00001113, 0.00000659, -0.00001946, 0.00000753, -0.00000699, 0.00000554, -0.00000181, -0.00000464, 0.00000492, 0.00000535, -0.00001124, 0.00000147, },
		{ -0.00000377, -0.00000109, -0.00000276, -0.00000253, 0.00000041, -0.00000031, 0.00000822, -0.00001134, 0.00000082, -0.00000075, 0.00000462, -0.00002792, 0.00002139, -0.00002238, 0.00002555, -0.00003347, 0.00002953, -0.00000293, 0.00001501, -0.00001313, 0.00001555, -0.00001174, 0.00000739, -0.00000663, 0.00002223, -0.00002501, 0.00003397, -0.00001232, 0.00000516, -0.00000748, -0.00000847, -0.00000466, -0.00000362, -0.00000662, 0.00001062, -0.00001937, 0.00001348, -0.00002701, 0.00001352, -0.00003007, 0.00001347, -0.00004965, 0.00002969, -0.00009715, 0.00003803, -0.00036823, 0.00013795, 0.00048411, 0.00042878, -0.00021605, -0.00017882, 0.00001327, -0.00007166, 0.00002468, -0.00004938, 0.00001889, -0.00002822, 0.00001630, -0.00002640, 0.00001767, -0.00002975, 0.00002725, -0.00003056, 0.00002582, -0.00001371, -0.00000133, 0.00000335, -0.00001436, 0.00000254, -0.00000173, -0.00000890, 0.00000314, -0.00000494, -0.00000185, 0.00000281, -0.00000185, -0.00000719, -0.00000400, -0.00000346, 0.00001236, -0.00000809, 0.00000028, -0.00002873, 0.00003056, -0.00002106, 0.00001757, -0.00000173, 0.00000845, -0.00000881, 0.00000680, -0.00001002, 0.00001006, -0.00000473, 0.00001063, -0.00000408, -0.00000178, -0.00000173, 0.00000101, -0.00000709, 0.00001292, -0.00001279, 0.00000759, 0.00000494, -0.00000908, -0.00000355, 0.00001185, 0.00000631, -0.00000847, -0.00000764, 0.00000799, 0.00000814, -0.00000764, 0.00000329, -0.00001000, 0.00000110, 0.00000213, 0.00001606, -0.00002249, 0.00000714, -0.00000768, 0.00000579, -0.00000050, -0.00000475, 0.00000038, 0.00001018, -0.00001169, 0.00000039, 0.00000111, },
		{ -0.00000366, 0.00000058, -0.00000655, 0.00000436, -0.00000558, 0.00001272, -0.00001333, 0.00000346, -0.00000664, 0.00001353, -0.00003545, 0.00002410, -0.00002162, 0.00002010, -0.00002461, 0.00001621, 0.00000761, 0.00001007, -0.00000801, 0.00001108, -0.00000645, 0.00000552, -0.00000896, 0.00002433, -0.00002334, 0.00002819, -0.00000307, 0.00000013, -0.00000498, -0.00000898, -0.00000713, 0.00000119, -0.00001463, 0.00001819, -0.00002485, 0.00001884, -0.00003148, 0.00001912, -0.00003688, 0.00002431, -0.00006207, 0.00004421, -0.00010606, 0.00005635, -0.00033016, 0.00003645, 0.00042671, 0.00042956, -0.00000879, -0.00030355, 0.00005660, -0.00010806, 0.00004823, -0.00006827, 0.00003214, -0.00004224, 0.00003019, -0.00003926, 0.00002916, -0.00003942, 0.00003307, -0.00003303, 0.00002461, -0.00001163, -0.00000049, -0.00000048, -0.00000773, -0.00000673, 0.00000854, -0.00001741, 0.00000870, -0.00001016, 0.00000350, -0.00000519, 0.00000958, -0.00001712, 0.00000466, -0.00001437, 0.00002202, -0.00001649, 0.00001044, -0.00003433, 0.00002609, -0.00001326, 0.00000885, 0.00000389, 0.00000782, -0.00000831, 0.00000617, -0.00000884, 0.00000743, -0.00000222, 0.00000806, -0.00000088, -0.00000296, -0.00000206, 0.00000233, -0.00000811, 0.00001162, -0.00000825, -0.00000009, 0.00001070, -0.00000954, -0.00000285, 0.00000611, 0.00001175, -0.00000893, -0.00000508, -0.00000039, 0.00001585, -0.00001091, 0.00000584, -0.00001239, 0.00000393, -0.00000357, 0.00001964, -0.00001937, 0.00000245, -0.00000504, 0.00000262, 0.00000331, -0.00000578, -0.00000274, 0.00001161, -0.00000866, -0.00000270, 0.00000221, -0.00000285, },
		{ 0.00000147, -0.00000727, 0.00000439, -0.00000611, 0.00001172, -0.00001019, 0.00000159, -0.00000744, 0.00001537, -0.00003321, 0.00001534, -0.00000944, 0.00000541, -0.00000973, 0.00000016, 0.00001795, 0.00000554, -0.00000252, 0.00000613, -0.00000122, 0.00000392, -0.00000951, 0.00002219, -0.00001650, 0.00001769, 0.00000909, -0.00000637, -0.00000044, -0.00001222, -0.00000636, 0.00000160, -0.00001606, 0.00001686, -0.00002053, 0.00001409, -0.00002558, 0.00001384, -0.00003179, 0.00002100, -0.00005654, 0.00003586, -0.00008795, 0.00003901, -0.00025471, -0.00008315, 0.00041366, 0.00035085, 0.00020992, -0.00037013, 0.00004782, -0.00010822, 0.00004233, -0.00006327, 0.00002603, -0.00004027, 0.00002918, -0.00003776, 0.00002681, -0.00003577, 0.00002666, -0.00002501, 0.00001519, -0.00000378, -0.00000418, 0.00000024, -0.00000588, -0.00001079, 0.00001246, -0.00001862, 0.00000756, -0.00000994, 0.00000443, -0.00000910, 0.00001522, -0.00001931, 0.00000550, -0.00001712, 0.00002234, -0.00001449, 0.00000850, -0.00002686, 0.00001139, 0.00000016, -0.00000179, 0.00000908, 0.00000724, -0.00000631, 0.00000337, -0.00000536, 0.00000298, 0.00000143, 0.00000473, 0.00000259, -0.00000429, -0.00000170, 0.00000207, -0.00000661, 0.00000738, -0.00000131, -0.00000818, 0.00001493, -0.00000923, -0.00000137, 0.00000051, 0.00001528, -0.00000827, -0.00000242, -0.00000696, 0.00001891, -0.00000913, 0.00000431, -0.00001105, 0.00000261, -0.00000356, 0.00001616, -0.00001045, -0.00000566, 0.00000026, -0.00000317, 0.00000863, -0.00000734, -0.00000402, 0.00000941, -0.00000271, -0.00000708, 0.00000398, -0.00000269, -0.00000474, },
		{ -0.00000503, 0.00000111, -0.00000295, 0.00000683, -0.00000389, -0.00000304, -0.00000456, 0.00001167, -0.00002380, -0.00000024, 0.00000712, -0.00000995, 0.00000264, -0.00001128, 0.00002257, 0.00000518, 0.00000053, 0.00000300, 0.00000235, 0.00000309, -0.00000816, 0.00001620, -0.00000581, 0.00000450, 0.00002207, -0.00001299, 0.00000547, -0.00001774, -0.00000258, -0.00000243, -0.00001100, 0.00000763, -0.00000846, 0.00000190, -0.00001218, 0.00000056, -0.00001749, 0.00000618, -0.00003640, 0.00000982, -0.00005048, -0.00000223, -0.00015918, -0.00020236, 0.00041644, 0.00025262, 0.00036823, -0.00033600, -0.00002599, -0.00006898, 0.00000632, -0.00003497, 0.00000141, -0.00002262, 0.00001318, -0.00002193, 0.00001111, -0.00001986, 0.00000957, -0.00000824, -0.00000100, 0.00000921, -0.00001269, 0.00000633, -0.00000977, -0.00000856, 0.00000905, -0.00001217, -0.00000002, -0.00000451, 0.00000059, -0.00000767, 0.00001339, -0.00001282, -0.00000153, -0.00001206, 0.00001416, -0.00000398, -0.00000191, -0.00001239, -0.00000577, 0.00001171, -0.00000855, 0.00001023, 0.00000846, -0.00000398, -0.00000021, -0.00000132, -0.00000142, 0.00000455, 0.00000200, 0.00000538, -0.00000517, -0.00000106, 0.00000075, -0.00000361, 0.00000200, 0.00000551, -0.00001402, 0.00001578, -0.00000695, -0.00000075, -0.00000280, 0.00001522, -0.00000517, -0.00000171, -0.00000919, 0.00001576, -0.00000211, -0.00000096, -0.00000620, -0.00000264, 0.00000125, 0.00000794, 0.00000091, -0.00001398, 0.00000557, -0.00000905, 0.00001307, -0.00000784, -0.00000442, 0.00000508, 0.00000398, -0.00001083, 0.00000521, -0.00000276, -0.00000445, 0.00000001, },
		{ -0.00000299, 0.00000095, 0.00000153, 0.00000196, -0.00000705, -0.00000133, 0.00000637, -0.00001259, -0.00001543, 0.00001960, -0.00001779, 0.00000606, -0.00001386, 0.00001924, 0.00000981, 0.00000072, 0.00000210, 0.00000415, 0.00000271, -0.00000492, 0.00000761, 0.00000626, -0.00000835, 0.00003294, -0.00001740, 0.00001072, -0.00002320, 0.00000164, -0.00000809, -0.00000310, -0.00000452, 0.00000526, -0.00001112, 0.00000190, -0.00001372, -0.00000140, -0.00001181, -0.00001176, -0.00002095, -0.00000989, -0.00004659, -0.00006987, -0.00029296, 0.00039787, 0.00018978, 0.00042240, -0.00019813, -0.00014417, -0.00001094, -0.00004297, 0.00000238, -0.00002989, 0.00000101, -0.00000946, 0.00000031, -0.00001023, 0.00000079, -0.00001098, 0.00001070, -0.00001825, 0.00002270, -0.00002230, 0.00001437, -0.00001601, -0.00000333, 0.00000176, -0.00000205, -0.00000979, 0.00000235, -0.00000502, -0.00000327, 0.00000672, -0.00000145, -0.00001177, -0.00000411, 0.00000294, 0.00000860, -0.00001302, 0.00000019, -0.00001684, 0.00001499, -0.00000771, 0.00000594, 0.00001173, -0.00000183, -0.00000337, 0.00000173, -0.00000426, 0.00000594, 0.00000071, 0.00000687, -0.00000518, -0.00000063, -0.00000074, -0.00000058, -0.00000251, 0.00000995, -0.00001582, 0.00001257, -0.00000251, -0.00000170, -0.00000293, 0.00001147, 0.00000015, -0.00000324, -0.00000700, 0.00000781, 0.00000747, -0.00000719, -0.00000037, -0.00000915, 0.00000739, -0.00000050, 0.00001001, -0.00001866, 0.00000785, -0.00001218, 0.00001416, -0.00000566, -0.00000533, 0.00000079, 0.00000883, -0.00001193, 0.00000458, -0.00000209, -0.00000380, -0.00000239, -0.00000178, },
		{ 0.00000269, -0.00000108, 0.00000439, -0.00000779, -0.00000051, 0.00000302, -0.00000434, -0.00002453, 0.00002259, -0.00001459, -0.00000032, -0.00000873, 0.00001037, 0.00001663, 0.00000025, 0.00000184, 0.00000552, 0.00000175, 0.00000013, -0.00000187, 0.00001702, -0.00001790, 0.00003909, -0.00001753, 0.00001313, -0.00002587, 0.00000313, -0.00001168, 0.00000294, -0.00001370, 0.00001412, -0.00001844, 0.00001011, -0.00002220, 0.00000906, -0.00002436, 0.00000673, -0.00004329, 0.00001829, -0.00007543, -0.00000856, -0.00033397, 0.00033413, 0.00018864, 0.00037270, 0.00000445, -0.00025947, 0.00003131, -0.00008051, 0.00002899, -0.00005198, 0.00001753, -0.00002713, 0.00001802, -0.00002686, 0.00001649, -0.00002612, 0.00002391, -0.00002974, 0.00003083, -0.00002767, 0.00001906, -0.00001926, -0.00000038, -0.00000400, 0.00000597, -0.00001617, 0.00000584, -0.00000828, 0.00000015, -0.00000011, 0.00000890, -0.00001888, 0.00000057, -0.00000495, 0.00001637, -0.00001770, 0.00000414, -0.00001710, 0.00000827, 0.00000003, -0.00000181, 0.00001516, 0.00000084, -0.00000580, 0.00000318, -0.00000512, 0.00000550, 0.00000078, 0.00000721, -0.00000436, -0.00000065, -0.00000173, 0.00000140, -0.00000489, 0.00001101, -0.00001346, 0.00000624, 0.00000298, -0.00000365, -0.00000068, 0.00000566, 0.00000589, -0.00000557, -0.00000251, -0.00000136, 0.00001549, -0.00001076, 0.00000327, -0.00001356, 0.00001092, -0.00000491, 0.00001325, -0.00001741, 0.00000558, -0.00001117, 0.00001092, -0.00000045, -0.00000734, -0.00000198, 0.00001024, -0.00000950, 0.00000168, -0.00000019, -0.00000374, -0.00000411, 0.00000026, -0.00000531, },
		{ -0.00000005, 0.00000282, -0.00000500, -0.00000256, 0.00000270, -0.00000075, -0.00002601, 0.00001616, -0.00000301, -0.00001163, -0.00000159, 0.00000154, 0.00002094, 0.00000240, 0.00000000, 0.00000806, -0.00000062, 0.00000647, -0.00001043, 0.00002424, -0.00002226, 0.00003923, -0.00001254, 0.00001162, -0.00002410, -0.00000001, -0.00001093, 0.00000408, -0.00001628, 0.00001466, -0.00001710, 0.00000978, -0.00002206, 0.00001055, -0.00002717, 0.00001349, -0.00005040, 0.00002672, -0.00008069, 0.00001693, -0.00032236, 0.00022580, 0.00024039, 0.00025838, 0.00020763, -0.00031925, 0.00002786, -0.00008763, 0.00003116, -0.00005438, 0.00001838, -0.00003199, 0.00002363, -0.00003171, 0.00002085, -0.00003025, 0.00002663, -0.00003137, 0.00002994, -0.00002523, 0.00001659, -0.00001567, -0.00000339, -0.00000441, 0.00000793, -0.00001568, 0.00000326, -0.00000682, -0.00000008, -0.00000346, 0.00001375, -0.00001870, -0.00000154, -0.00000610, 0.00001606, -0.00001335, -0.00000158, -0.00000802, -0.00000445, 0.00000962, -0.00000836, 0.00001565, 0.00000527, -0.00000781, 0.00000338, -0.00000463, 0.00000409, 0.00000138, 0.00000706, -0.00000315, -0.00000100, -0.00000203, 0.00000201, -0.00000503, 0.00000921, -0.00000845, -0.00000097, 0.00000757, -0.00000520, 0.00000217, 0.00000029, 0.00000971, -0.00000669, 0.00000150, -0.00000801, 0.00001849, -0.00000928, 0.00000289, -0.00001389, 0.00000968, -0.00000352, 0.00001007, -0.00001090, -0.00000022, -0.00000694, 0.00000453, 0.00000637, -0.00000965, -0.00000316, 0.00000835, -0.00000439, -0.00000252, 0.00000241, -0.00000441, -0.00000440, -0.00000001, -0.00000469, 0.00000115, },
		{ -0.00000075, -0.00000088, -0.00000557, 0.00000386, -0.00000025, -0.00002244, 0.00000493, 0.00001016, -0.00002027, 0.00000085, -0.00000199, 0.00001912, 0.00000907, -0.00000434, 0.00001221, -0.00000417, 0.00001282, -0.00001623, 0.00002659, -0.00002096, 0.00003372, -0.00000323, 0.00000673, -0.00001813, -0.00000746, -0.00000610, 0.00000035, -0.00001245, 0.00000787, -0.00000898, 0.00000312, -0.00001546, 0.00000499, -0.00002178, 0.00000977, -0.00004387, 0.00001823, -0.00006658, 0.00001371, -0.00027273, 0.00009456, 0.00031172, 0.00013631, 0.00034989, -0.00029022, -0.00002928, -0.00006347, 0.00000932, -0.00003818, 0.00000452, -0.00002429, 0.00001688, -0.00002461, 0.00001401, -0.00002384, 0.00001959, -0.00002393, 0.00002075, -0.00001534, 0.00000704, -0.00000535, -0.00001212, 0.00000042, 0.00000389, -0.00000893, -0.00000425, -0.00000167, -0.00000361, -0.00000279, 0.00001234, -0.00001135, -0.00000939, -0.00000182, 0.00000947, -0.00000290, -0.00001236, 0.00000383, -0.00001575, 0.00001451, -0.00000909, 0.00001087, 0.00001203, -0.00000962, 0.00000296, -0.00000381, 0.00000283, 0.00000158, 0.00000707, -0.00000200, -0.00000142, -0.00000184, 0.00000161, -0.00000375, 0.00000608, -0.00000304, -0.00000664, 0.00000953, -0.00000512, 0.00000395, -0.00000262, 0.00001014, -0.00000541, 0.00000321, -0.00001006, 0.00001549, -0.00000305, -0.00000088, -0.00001065, 0.00000431, 0.00000225, 0.00000307, -0.00000246, -0.00000645, -0.00000217, -0.00000228, 0.00001218, -0.00001055, -0.00000383, 0.00000471, 0.00000124, -0.00000613, 0.00000448, -0.00000511, -0.00000365, -0.00000210, -0.00000193, -0.00000251, 0.00000309, },
		{ 0.00000152, -0.00000680, 0.00000393, 0.00000017, -0.00001809, -0.00000513, 0.00001807, -0.00002039, -0.00000499, 0.00000142, 0.00001116, 0.00001904, -0.00000966, 0.00001650, -0.00000720, 0.00001720, -0.00001774, 0.00002374, -0.00001484, 0.00002421, 0.00000840, 0.00000025, -0.00000978, -0.00001698, 0.00000034, -0.00000568, -0.00000542, -0.00000193, 0.00000084, -0.00000467, -0.00000752, -0.00000253, -0.00001342, 0.00000139, -0.00003090, 0.00000218, -0.00004468, -0.00000275, -0.00020821, -0.00002992, 0.00036467, 0.00005498, 0.00039698, -0.00017410, -0.00011987, -0.00002631, -0.00002236, -0.00001502, -0.00001461, -0.00001155, 0.00000419, -0.00001165, 0.00000185, -0.00001256, 0.00000812, -0.00001234, 0.00000769, -0.00000204, -0.00000581, 0.00000776, -0.00002256, 0.00000672, -0.00000243, -0.00000006, -0.00001243, 0.00000348, -0.00000753, -0.00000045, 0.00000738, -0.00000080, -0.00001828, 0.00000311, 0.00000182, 0.00000746, -0.00002101, 0.00001078, -0.00001891, 0.00001042, -0.00000232, 0.00000122, 0.00001984, -0.00001054, 0.00000212, -0.00000328, 0.00000229, 0.00000103, 0.00000742, -0.00000100, -0.00000181, -0.00000142, 0.00000081, -0.00000215, 0.00000318, 0.00000094, -0.00000932, 0.00000832, -0.00000325, 0.00000406, -0.00000248, 0.00000735, -0.00000218, 0.00000263, -0.00000789, 0.00000829, 0.00000514, -0.00000532, -0.00000627, -0.00000243, 0.00000885, -0.00000346, 0.00000371, -0.00000976, 0.00000040, -0.00000680, 0.00001467, -0.00000862, -0.00000514, 0.00000124, 0.00000515, -0.00000742, 0.00000485, -0.00000493, -0.00000290, -0.00000453, 0.00000095, -0.00000585, 0.00000556, -0.00000326, },
		{ -0.00000475, 0.00000137, 0.00000255, -0.00001600, -0.00001023, 0.00001757, -0.00001117, -0.00001732, 0.00000860, 0.00000092, 0.00002822, -0.00001253, 0.00001820, -0.00000735, 0.00001769, -0.00001416, 0.00001643, -0.00000565, 0.00001299, 0.00001995, -0.00000558, -0.00000146, -0.00002562, 0.00000525, -0.00001062, 0.00000071, -0.00000981, 0.00000722, -0.00000870, -0.00000289, -0.00000729, -0.00000726, -0.00000557, -0.00001915, -0.00001179, -0.00002638, -0.00001798, -0.00014899, -0.00012454, 0.00037461, 0.00003639, 0.00035271, -0.00000682, -0.00020376, -0.00000372, -0.00004475, 0.00000026, -0.00002733, -0.00000325, -0.00000603, -0.00000084, -0.00000803, -0.00000347, -0.00000132, -0.00000245, -0.00000377, 0.00000949, -0.00001675, 0.00001825, -0.00002934, 0.00000954, -0.00000617, 0.00000589, -0.00001652, 0.00000470, -0.00000860, 0.00000041, 0.00000284, 0.00000784, -0.00002290, 0.00000376, -0.00000181, 0.00001226, -0.00002209, 0.00000835, -0.00001176, -0.00000200, 0.00000927, -0.00000974, 0.00002566, -0.00000902, 0.00000048, -0.00000293, 0.00000223, 0.00000019, 0.00000760, 0.00000021, -0.00000233, -0.00000088, 0.00000006, -0.00000096, 0.00000140, 0.00000281, -0.00000906, 0.00000490, -0.00000060, 0.00000310, -0.00000025, 0.00000310, 0.00000121, 0.00000120, -0.00000365, 0.00000033, 0.00001153, -0.00000733, -0.00000351, -0.00000750, 0.00001275, -0.00000595, 0.00000486, -0.00000851, -0.00000040, -0.00000791, 0.00001313, -0.00000384, -0.00000734, -0.00000101, 0.00000619, -0.00000586, 0.00000331, -0.00000352, -0.00000294, -0.00000588, 0.00000207, -0.00000675, 0.00000515, -0.00000252, 0.00000034, },
		{ -0.00000309, 0.00000660, -0.00001633, -0.00001042, 0.00001040, 0.00000301, -0.00003001, 0.00001342, -0.00000602, 0.00003187, -0.00000962, 0.00001508, -0.00000304, 0.00001353, -0.00000609, 0.00000644, 0.00000435, 0.00000236, 0.00002933, -0.00000898, 0.00000483, -0.00003105, 0.00000634, -0.00001218, 0.00000322, -0.00001279, 0.00000759, -0.00000714, -0.00000311, -0.00000762, -0.00000538, -0.00000822, -0.00001263, -0.00001870, -0.00001715, -0.00002496, -0.00010484, -0.00018101, 0.00033817, 0.00007270, 0.00025210, 0.00015781, -0.00023975, -0.00001677, -0.00004567, -0.00000104, -0.00002718, -0.00000463, -0.00000877, 0.00000281, -0.00001097, -0.00000077, -0.00000498, 0.00000229, -0.00001026, 0.00001578, -0.00002207, 0.00002222, -0.00002887, 0.00000584, -0.00000447, 0.00000602, -0.00001418, 0.00000045, -0.00000556, -0.00000180, 0.00000126, 0.00001130, -0.00002047, -0.00000186, 0.00000040, 0.00000988, -0.00001491, -0.00000238, 0.00000216, -0.00001721, 0.00001967, -0.00001679, 0.00002608, -0.00000368, -0.00000243, -0.00000222, 0.00000185, -0.00000004, 0.00000689, 0.00000210, -0.00000320, -0.00000022, -0.00000050, -0.00000036, 0.00000071, 0.00000306, -0.00000718, 0.00000113, 0.00000125, 0.00000227, 0.00000241, -0.00000038, 0.00000285, 0.00000057, 0.00000026, -0.00000521, 0.00001346, -0.00000529, -0.00000361, -0.00000936, 0.00001232, -0.00000338, 0.00000111, -0.00000374, -0.00000342, -0.00000658, 0.00000886, 0.00000219, -0.00000943, -0.00000220, 0.00000472, -0.00000243, 0.00000084, -0.00000140, -0.00000380, -0.00000569, 0.00000078, -0.00000470, 0.00000160, 0.00000139, -0.00000390, -0.00000027, },
		{ 0.00001008, -0.00001682, -0.00000880, 0.00000152, 0.00001531, -0.00003608, 0.00001052, -0.00000577, 0.00002751, -0.00000002, 0.00000716, 0.00000518, 0.00000596, 0.00000434, -0.00000361, 0.00001288, -0.00000584, 0.00003517, -0.00000901, 0.00000814, -0.00003231, 0.00000308, -0.00001007, 0.00000180, -0.00001089, 0.00000289, -0.00000173, -0.00000623, -0.00000529, -0.00000637, -0.00000756, -0.00001067, -0.00001934, -0.00001549, -0.00002511, -0.00007469, -0.00020507, 0.00026950, 0.00013660, 0.00014232, 0.00027140, -0.00020529, -0.00006720, -0.00002784, -0.00001622, -0.00001678, -0.00001359, -0.00000527, -0.00000005, -0.00000756, -0.00000374, -0.00000363, 0.00000248, -0.00001199, 0.00001663, -0.00002129, 0.00001921, -0.00002120, -0.00000373, 0.00000192, 0.00000115, -0.00000680, -0.00000738, -0.00000011, -0.00000606, 0.00000244, 0.00000969, -0.00001214, -0.00001164, 0.00000613, 0.00000309, -0.00000345, -0.00001574, 0.00001561, -0.00002770, 0.00002271, -0.00001590, 0.00001943, 0.00000543, -0.00000635, -0.00000091, 0.00000055, 0.00000094, 0.00000489, 0.00000481, -0.00000441, 0.00000050, -0.00000089, -0.00000010, 0.00000051, 0.00000282, -0.00000531, -0.00000128, 0.00000119, 0.00000237, 0.00000421, -0.00000162, 0.00000186, 0.00000136, 0.00000257, -0.00000697, 0.00001065, 0.00000000, -0.00000567, -0.00000868, 0.00000847, 0.00000250, -0.00000476, 0.00000140, -0.00000597, -0.00000507, 0.00000428, 0.00000713, -0.00000990, -0.00000329, 0.00000216, 0.00000101, -0.00000102, 0.00000037, -0.00000480, -0.00000448, -0.00000212, -0.00000095, -0.00000338, 0.00000635, -0.00000862, 0.00000393, -0.00000695, },
		{ -0.00001483, -0.00000892, -0.00000413, 0.00002029, -0.00003167, -0.00000141, 0.00000136, 0.00001658, 0.00001374, -0.00000271, 0.00001450, -0.00000204, 0.00001393, -0.00001094, 0.00001804, -0.00001049, 0.00003696, -0.00000560, 0.00000854, -0.00002980, -0.00000343, -0.00000579, -0.00000185, -0.00000646, -0.00000359, 0.00000375, -0.00000859, -0.00000360, -0.00000710, -0.00000665, -0.00000995, -0.00001788, -0.00001602, -0.00002421, -0.00005187, -0.00020913, 0.00018928, 0.00019778, 0.00006179, 0.00030995, -0.00010708, -0.00013467, -0.00000784, -0.00003313, -0.00000579, -0.00002251, -0.00000142, -0.00000446, -0.00000251, -0.00000793, -0.00000133, 0.00000161, -0.00001182, 0.00001451, -0.00001675, 0.00001184, -0.00000968, -0.00001533, 0.00000922, -0.00000498, 0.00000148, -0.00001461, 0.00000417, -0.00000954, 0.00000405, 0.00000582, -0.00000198, -0.00002084, 0.00001067, -0.00000303, 0.00000624, -0.00002484, 0.00002166, -0.00002798, 0.00001550, -0.00000667, 0.00000724, 0.00001611, -0.00000982, 0.00000053, -0.00000154, 0.00000294, 0.00000201, 0.00000779, -0.00000547, 0.00000098, -0.00000107, 0.00000009, 0.00000020, 0.00000299, -0.00000450, -0.00000167, -0.00000071, 0.00000319, 0.00000493, -0.00000058, -0.00000101, 0.00000276, 0.00000361, -0.00000566, 0.00000513, 0.00000582, -0.00000729, -0.00000751, 0.00000364, 0.00000840, -0.00000901, 0.00000357, -0.00000560, -0.00000537, 0.00000142, 0.00000921, -0.00000781, -0.00000501, -0.00000006, 0.00000281, -0.00000109, 0.00000101, -0.00000519, -0.00000328, -0.00000509, 0.00000239, -0.00000728, 0.00000951, -0.00001083, 0.00000503, -0.00000771, 0.00001057, },
		{ -0.00001237, -0.00000449, 0.00001678, -0.00001821, -0.00001852, 0.00001081, 0.00000391, 0.00002660, -0.00001006, 0.00002110, -0.00000700, 0.00001964, -0.00001349, 0.00001892, -0.00001131, 0.00003490, 0.00000060, 0.00000684, -0.00002477, -0.00001127, -0.00000154, -0.00000539, -0.00000235, -0.00000839, 0.00000591, -0.00000744, -0.00000489, -0.00000538, -0.00000786, -0.00000758, -0.00001810, -0.00001405, -0.00002706, -0.00003079, -0.00020383, 0.00011374, 0.00023654, 0.00002703, 0.00027992, 0.00002247, -0.00018598, -0.00000654, -0.00003826, -0.00000429, -0.00002374, -0.00000324, -0.00000507, -0.00000096, -0.00000856, -0.00000230, 0.00000323, -0.00001278, 0.00001233, -0.00001163, 0.00000390, 0.00000119, -0.00002422, 0.00001305, -0.00000814, 0.00000634, -0.00001728, 0.00000415, -0.00000973, 0.00000359, 0.00000298, 0.00000561, -0.00002495, 0.00001002, -0.00000437, 0.00000973, -0.00002546, 0.00001737, -0.00001759, 0.00000025, 0.00000698, -0.00000598, 0.00002454, -0.00001068, 0.00000107, -0.00000363, 0.00000503, -0.00000068, 0.00001006, -0.00000563, 0.00000075, -0.00000089, 0.00000029, -0.00000043, 0.00000377, -0.00000473, -0.00000065, -0.00000319, 0.00000362, 0.00000530, 0.00000158, -0.00000393, 0.00000311, 0.00000463, -0.00000319, -0.00000015, 0.00000918, -0.00000626, -0.00000772, 0.00000015, 0.00001157, -0.00000907, 0.00000118, -0.00000174, -0.00000777, 0.00000073, 0.00000832, -0.00000363, -0.00000715, -0.00000138, 0.00000239, 0.00000068, 0.00000059, -0.00000477, -0.00000277, -0.00000686, 0.00000366, -0.00000828, 0.00000909, -0.00000893, 0.00000184, -0.00000457, 0.00000734, -0.00000937, },
		{ -0.00000116, 0.00000803, -0.00000146, -0.00003378, 0.00001627, -0.00000493, 0.00003359, -0.00001112, 0.00002219, -0.00000671, 0.00001999, -0.00001083, 0.00001593, -0.00000895, 0.00002983, 0.00000856, 0.00000414, -0.00001866, -0.00001860, 0.00000097, -0.00000724, -0.00000037, -0.00000975, 0.00000369, -0.00000260, -0.00000888, -0.00000147, -0.00001114, -0.00000324, -0.00002084, -0.00000857, -0.00003396, -0.00001081, -0.00019333, 0.00004923, 0.00024879, 0.00003199, 0.00021049, 0.00014046, -0.00019137, -0.00003618, -0.00002606, -0.00001578, -0.00001509, -0.00001235, -0.00000015, -0.00000482, -0.00000392, -0.00000785, 0.00000826, -0.00001571, 0.00001131, -0.00000779, -0.00000200, 0.00000825, -0.00002739, 0.00001102, -0.00000626, 0.00000578, -0.00001402, -0.00000077, -0.00000624, 0.00000031, 0.00000285, 0.00000833, -0.00002226, 0.00000330, -0.00000026, 0.00000654, -0.00001812, 0.00000513, -0.00000131, -0.00001664, 0.00001855, -0.00001472, 0.00002714, -0.00000733, -0.00000007, -0.00000479, 0.00000605, -0.00000203, 0.00001064, -0.00000425, -0.00000048, -0.00000022, 0.00000041, -0.00000117, 0.00000469, -0.00000514, 0.00000035, -0.00000459, 0.00000242, 0.00000625, 0.00000342, -0.00000508, 0.00000113, 0.00000654, -0.00000118, -0.00000301, 0.00000856, -0.00000208, -0.00000966, -0.00000129, 0.00001120, -0.00000488, -0.00000470, 0.00000385, -0.00001065, 0.00000093, 0.00000595, 0.00000092, -0.00000859, -0.00000225, 0.00000044, 0.00000317, 0.00000003, -0.00000399, -0.00000299, -0.00000704, 0.00000244, -0.00000622, 0.00000529, -0.00000359, -0.00000438, 0.00000090, 0.00000192, -0.00000186, -0.00001788, },
		{ -0.00000054, 0.00001165, -0.00004073, 0.00001329, -0.00000674, 0.00003242, -0.00000492, 0.00001760, -0.00000144, 0.00001579, -0.00000444, 0.00001073, -0.00000475, 0.00002296, 0.00001707, 0.00000154, -0.00001265, -0.00002428, 0.00000113, -0.00000721, -0.00000055, -0.00000812, -0.00000139, 0.00000358, -0.00001301, 0.00000223, -0.00001424, 0.00000099, -0.00002397, -0.00000228, -0.00004109, 0.00000417, -0.00017608, -0.00000618, 0.00024200, 0.00005755, 0.00013714, 0.00021250, -0.00013945, -0.00009163, -0.00000357, -0.00003440, -0.00000161, -0.00002458, 0.00000717, -0.00001161, 0.00000358, -0.00001551, 0.00001432, -0.00001867, 0.00001031, -0.00000493, -0.00000554, 0.00001099, -0.00002489, 0.00000394, -0.00000044, 0.00000109, -0.00000672, -0.00000823, -0.00000120, -0.00000431, 0.00000468, 0.00000691, -0.00001454, -0.00000682, 0.00000645, -0.00000006, -0.00000730, -0.00000893, 0.00001345, -0.00002786, 0.00002230, -0.00001536, 0.00002250, 0.00000007, -0.00000264, -0.00000469, 0.00000538, -0.00000148, 0.00000925, -0.00000133, -0.00000257, 0.00000080, 0.00000037, -0.00000163, 0.00000507, -0.00000473, 0.00000007, -0.00000381, -0.00000083, 0.00000801, 0.00000420, -0.00000374, -0.00000310, 0.00000906, 0.00000002, -0.00000308, 0.00000464, 0.00000372, -0.00001187, -0.00000167, 0.00000855, 0.00000138, -0.00001100, 0.00000803, -0.00001152, -0.00000001, 0.00000420, 0.00000383, -0.00000808, -0.00000346, -0.00000177, 0.00000479, 0.00000049, -0.00000353, -0.00000345, -0.00000616, -0.00000037, -0.00000251, 0.00000005, 0.00000271, -0.00001070, 0.00000565, -0.00000278, 0.00000480, -0.00002238, 0.00000209, },
		{ 0.00001669, -0.00003686, 0.00000188, -0.00000248, 0.00002463, 0.00000609, 0.00001003, 0.00000620, 0.00000975, 0.00000295, 0.00000558, -0.00000027, 0.00001561, 0.00002491, -0.00000002, -0.00000750, -0.00002795, -0.00000066, -0.00000621, -0.00000167, -0.00000545, -0.00000646, 0.00000779, -0.00001424, 0.00000300, -0.00001460, 0.00000246, -0.00002448, 0.00000094, -0.00004345, 0.00000946, -0.00014920, -0.00005692, 0.00022656, 0.00008443, 0.00008564, 0.00022606, -0.00004313, -0.00015095, 0.00001260, -0.00004835, 0.00000740, -0.00003260, 0.00001111, -0.00001632, 0.00000914, -0.00002072, 0.00001729, -0.00001833, 0.00000699, -0.00000139, -0.00000808, 0.00001109, -0.00001925, -0.00000491, 0.00000592, -0.00000424, 0.00000070, -0.00001433, 0.00000211, -0.00000765, 0.00000623, 0.00000406, -0.00000569, -0.00001581, 0.00001133, -0.00000533, 0.00000140, -0.00001848, 0.00002052, -0.00002890, 0.00001624, -0.00000814, 0.00001254, 0.00000915, -0.00000516, -0.00000385, 0.00000329, 0.00000059, 0.00000648, 0.00000241, -0.00000484, 0.00000173, 0.00000026, -0.00000161, 0.00000449, -0.00000301, -0.00000188, -0.00000098, -0.00000532, 0.00000980, 0.00000429, -0.00000062, -0.00000808, 0.00001072, 0.00000129, -0.00000158, -0.00000024, 0.00000837, -0.00001217, -0.00000278, 0.00000580, 0.00000672, -0.00001442, 0.00000811, -0.00000867, -0.00000329, 0.00000434, 0.00000406, -0.00000527, -0.00000530, -0.00000328, 0.00000450, 0.00000252, -0.00000371, -0.00000371, -0.00000505, -0.00000335, 0.00000084, -0.00000419, 0.00000721, -0.00001420, 0.00000697, -0.00000427, 0.00000808, -0.00002314, 0.00000190, -0.00000464, },
		{ -0.00002471, -0.00001347, 0.00000343, 0.00001452, 0.00001735, 0.00000355, 0.00001271, 0.00000502, 0.00000864, 0.00000234, 0.00000334, 0.00000884, 0.00003109, 0.00000008, -0.00000352, -0.00002980, -0.00000349, -0.00000548, -0.00000233, -0.00000364, -0.00000911, 0.00000784, -0.00001104, -0.00000021, -0.00001124, 0.00000003, -0.00002079, -0.00000112, -0.00003840, 0.00000343, -0.00011299, -0.00010425, 0.00020836, 0.00010249, 0.00006381, 0.00019255, 0.00006633, -0.00018524, 0.00000572, -0.00004736, 0.00000447, -0.00003083, 0.00000728, -0.00001497, 0.00000887, -0.00001988, 0.00001415, -0.00001252, -0.00000013, 0.00000405, -0.00001109, 0.00001071, -0.00001357, -0.00001192, 0.00000932, -0.00000679, 0.00000470, -0.00001592, 0.00000134, -0.00000786, 0.00000561, 0.00000246, 0.00000062, -0.00001999, 0.00001120, -0.00000605, 0.00000444, -0.00002024, 0.00001798, -0.00002021, 0.00000318, 0.00000279, 0.00000171, 0.00001628, -0.00000565, -0.00000327, 0.00000069, 0.00000307, 0.00000352, 0.00000582, -0.00000638, 0.00000203, 0.00000033, -0.00000118, 0.00000306, -0.00000029, -0.00000481, 0.00000259, -0.00000933, 0.00001020, 0.00000471, 0.00000271, -0.00001164, 0.00000979, 0.00000381, -0.00000022, -0.00000357, 0.00000959, -0.00000913, -0.00000571, 0.00000451, 0.00000911, -0.00001328, 0.00000348, -0.00000246, -0.00000834, 0.00000600, 0.00000220, -0.00000115, -0.00000716, -0.00000400, 0.00000231, 0.00000563, -0.00000412, -0.00000375, -0.00000425, -0.00000540, 0.00000239, -0.00000588, 0.00000834, -0.00001353, 0.00000400, -0.00000204, 0.00000741, -0.00001949, -0.00000260, -0.00000321, 0.00001365, },
		{ -0.00002622, 0.00000581, 0.00000653, 0.00002464, 0.00000145, 0.00001562, 0.00000356, 0.00001126, 0.00000162, 0.00000580, 0.00000319, 0.00003503, 0.00000213, -0.00000073, -0.00003022, -0.00000657, -0.00000574, -0.00000192, -0.00000350, -0.00000865, 0.00000380, -0.00000440, -0.00000601, -0.00000552, -0.00000508, -0.00001391, -0.00000758, -0.00002727, -0.00001084, -0.00007260, -0.00014435, 0.00018645, 0.00011255, 0.00006339, 0.00013846, 0.00015361, -0.00017282, -0.00003061, -0.00002995, -0.00001086, -0.00001940, -0.00000406, -0.00000748, 0.00000240, -0.00001277, 0.00000511, -0.00000196, -0.00001009, 0.00001075, -0.00001466, 0.00001086, -0.00000965, -0.00001514, 0.00000826, -0.00000526, 0.00000408, -0.00001243, -0.00000337, -0.00000517, 0.00000265, 0.00000308, 0.00000291, -0.00001841, 0.00000587, -0.00000220, 0.00000198, -0.00001531, 0.00000866, -0.00000672, -0.00001075, 0.00001154, -0.00000515, 0.00001845, -0.00000289, -0.00000356, -0.00000152, 0.00000483, 0.00000149, 0.00000795, -0.00000650, 0.00000131, 0.00000075, -0.00000058, 0.00000129, 0.00000255, -0.00000744, 0.00000519, -0.00001109, 0.00000807, 0.00000634, 0.00000485, -0.00001217, 0.00000546, 0.00000787, 0.00000010, -0.00000408, 0.00000687, -0.00000319, -0.00000991, 0.00000465, 0.00000850, -0.00000831, -0.00000401, 0.00000469, -0.00001298, 0.00000738, 0.00000010, 0.00000239, -0.00000778, -0.00000457, -0.00000082, 0.00000849, -0.00000378, -0.00000399, -0.00000380, -0.00000623, 0.00000188, -0.00000500, 0.00000643, -0.00000951, -0.00000187, 0.00000229, 0.00000437, -0.00001309, -0.00000959, 0.00000142, 0.00000635, -0.00001025, },
		{ 0.00000185, 0.00000273, 0.00002620, 0.00000448, 0.00001480, 0.00000524, 0.00001133, 0.00000263, 0.00000765, -0.00000128, 0.00003653, 0.00000604, 0.00000109, -0.00002954, -0.00000955, -0.00000687, -0.00000092, -0.00000441, -0.00000618, -0.00000218, 0.00000277, -0.00001137, -0.00000033, -0.00001009, -0.00000672, -0.00001518, -0.00001435, -0.00002700, -0.00003609, -0.00017053, 0.00015583, 0.00012162, 0.00006909, 0.00009114, 0.00019466, -0.00011023, -0.00008696, -0.00000553, -0.00003118, -0.00000452, -0.00001781, 0.00000214, -0.00000686, -0.00000275, -0.00000634, 0.00000978, -0.00001962, 0.00001626, -0.00001725, 0.00001082, -0.00000724, -0.00001512, 0.00000387, -0.00000103, 0.00000043, -0.00000606, -0.00000949, -0.00000180, -0.00000105, 0.00000492, 0.00000221, -0.00001302, -0.00000194, 0.00000339, -0.00000281, -0.00000789, -0.00000183, 0.00000500, -0.00001936, 0.00001353, -0.00000541, 0.00001500, 0.00000256, -0.00000434, -0.00000304, 0.00000533, 0.00000088, 0.00000848, -0.00000515, -0.00000031, 0.00000147, -0.00000001, -0.00000024, 0.00000460, -0.00000861, 0.00000558, -0.00000976, 0.00000334, 0.00000909, 0.00000540, -0.00000950, -0.00000144, 0.00001241, -0.00000014, -0.00000230, 0.00000173, 0.00000345, -0.00001348, 0.00000482, 0.00000648, -0.00000201, -0.00001108, 0.00000957, -0.00001469, 0.00000652, -0.00000034, 0.00000363, -0.00000623, -0.00000563, -0.00000368, 0.00000973, -0.00000187, -0.00000481, -0.00000344, -0.00000621, 0.00000010, -0.00000282, 0.00000319, -0.00000437, -0.00000794, 0.00000596, 0.00000160, -0.00000676, -0.00001590, 0.00000545, -0.00000014, 0.00000045, -0.00003136, },
		{ 0.00000186, 0.00002322, 0.00001069, 0.00001234, 0.00000813, 0.00001065, 0.00000378, 0.00000975, -0.00000465, 0.00003564, 0.00001145, 0.00000229, -0.00002798, -0.00001257, -0.00000814, -0.00000044, -0.00000502, -0.00000371, -0.00000725, 0.00000728, -0.00001339, 0.00000177, -0.00001242, -0.00000205, -0.00002042, -0.00000435, -0.00003863, -0.00001064, -0.00017759, 0.00011272, 0.00013576, 0.00006881, 0.00006683, 0.00018568, -0.00001453, -0.00014096, 0.00000995, -0.00004535, 0.00000511, -0.00002707, 0.00000830, -0.00001391, 0.00000531, -0.00001541, 0.00001816, -0.00002486, 0.00001768, -0.00001680, 0.00000900, -0.00000474, -0.00001402, -0.00000107, 0.00000296, -0.00000317, -0.00000029, -0.00001360, -0.00000049, -0.00000332, 0.00000610, 0.00000084, -0.00000717, -0.00000844, 0.00000704, -0.00000604, -0.00000253, -0.00000842, 0.00001019, -0.00001958, 0.00000793, -0.00000007, 0.00000815, 0.00000834, -0.00000430, -0.00000425, 0.00000474, 0.00000135, 0.00000787, -0.00000287, -0.00000227, 0.00000213, 0.00000052, -0.00000121, 0.00000537, -0.00000787, 0.00000360, -0.00000588, -0.00000276, 0.00001183, 0.00000510, -0.00000490, -0.00000875, 0.00001536, 0.00000045, 0.00000009, -0.00000328, 0.00000806, -0.00001432, 0.00000339, 0.00000504, 0.00000289, -0.00001479, 0.00000991, -0.00001217, 0.00000263, 0.00000162, 0.00000215, -0.00000265, -0.00000715, -0.00000565, 0.00000872, 0.00000168, -0.00000609, -0.00000303, -0.00000592, -0.00000184, -0.00000091, 0.00000056, -0.00000037, -0.00001177, 0.00000678, 0.00000108, -0.00000271, -0.00001902, 0.00000604, -0.00000243, 0.00000428, -0.00002788, -0.00000727, },
		{ 0.00001863, 0.00001681, 0.00001110, 0.00000997, 0.00001111, 0.00000382, 0.00001249, -0.00000678, 0.00003250, 0.00001788, 0.00000324, -0.00002561, -0.00001593, -0.00000870, -0.00000154, -0.00000420, -0.00000284, -0.00000938, 0.00000737, -0.00001098, 0.00000004, -0.00001126, -0.00000109, -0.00002149, 0.00000008, -0.00004229, 0.00000074, -0.00016528, 0.00005853, 0.00015512, 0.00005964, 0.00006613, 0.00014204, 0.00008419, -0.00016735, 0.00000291, -0.00004536, 0.00000366, -0.00002743, 0.00000740, -0.00001534, 0.00000793, -0.00001873, 0.00002019, -0.00002357, 0.00001356, -0.00001227, 0.00000438, -0.00000072, -0.00001396, -0.00000399, 0.00000422, -0.00000422, 0.00000222, -0.00001350, -0.00000269, -0.00000308, 0.00000538, 0.00000070, -0.00000351, -0.00001100, 0.00000663, -0.00000558, -0.00000157, -0.00000910, 0.00000814, -0.00001276, -0.00000203, 0.00000680, 0.00000195, 0.00001144, -0.00000206, -0.00000567, 0.00000367, 0.00000214, 0.00000695, -0.00000049, -0.00000387, 0.00000231, 0.00000109, -0.00000156, 0.00000492, -0.00000561, 0.00000015, -0.00000104, -0.00000824, 0.00001297, 0.00000520, -0.00000024, -0.00001401, 0.00001487, 0.00000303, 0.00000155, -0.00000601, 0.00000886, -0.00001155, -0.00000021, 0.00000520, 0.00000492, -0.00001404, 0.00000565, -0.00000624, -0.00000322, 0.00000502, -0.00000092, 0.00000157, -0.00000824, -0.00000689, 0.00000584, 0.00000599, -0.00000713, -0.00000281, -0.00000563, -0.00000319, -0.00000027, -0.00000043, 0.00000139, -0.00001246, 0.00000430, 0.00000302, -0.00000131, -0.00001851, 0.00000274, -0.00000029, 0.00000217, -0.00001903, -0.00001715, 0.00000055, },
		{ 0.00002021, 0.00001295, 0.00000961, 0.00001333, 0.00000269, 0.00001531, -0.00000708, 0.00002721, 0.00002476, 0.00000429, -0.00002249, -0.00001988, -0.00000810, -0.00000448, -0.00000180, -0.00000389, -0.00000833, 0.00000354, -0.00000538, -0.00000405, -0.00000787, -0.00000289, -0.00001900, -0.00000071, -0.00003848, 0.00000009, -0.00013888, 0.00000018, 0.00017359, 0.00004735, 0.00007771, 0.00008895, 0.00015603, -0.00015001, -0.00002960, -0.00003166, -0.00000813, -0.00001974, 0.00000011, -0.00001133, 0.00000494, -0.00001614, 0.00001598, -0.00001640, 0.00000496, -0.00000465, -0.00000244, 0.00000480, -0.00001556, -0.00000412, 0.00000221, -0.00000225, 0.00000108, -0.00000941, -0.00000756, -0.00000112, 0.00000312, 0.00000211, -0.00000271, -0.00000949, 0.00000263, -0.00000207, -0.00000418, -0.00000551, 0.00000194, -0.00000361, -0.00001096, 0.00001039, 0.00000000, 0.00001000, 0.00000268, -0.00000728, 0.00000253, 0.00000260, 0.00000634, 0.00000141, -0.00000467, 0.00000183, 0.00000178, -0.00000145, 0.00000373, -0.00000275, -0.00000337, 0.00000295, -0.00001134, 0.00001143, 0.00000654, 0.00000304, -0.00001571, 0.00001041, 0.00000753, 0.00000167, -0.00000585, 0.00000600, -0.00000610, -0.00000502, 0.00000649, 0.00000446, -0.00000993, -0.00000110, 0.00000051, -0.00000867, 0.00000781, -0.00000352, 0.00000449, -0.00000773, -0.00000803, 0.00000214, 0.00000967, -0.00000694, -0.00000317, -0.00000530, -0.00000386, -0.00000088, 0.00000000, 0.00000137, -0.00001082, -0.00000013, 0.00000590, -0.00000122, -0.00001580, -0.00000260, 0.00000384, -0.00000215, -0.00001008, -0.00002285, -0.00000418, 0.00001482, },
		{ 0.00001765, 0.00000782, 0.00001636, 0.00000161, 0.00001691, -0.00000482, 0.00002000, 0.00003147, 0.00000581, -0.00001874, -0.00002431, -0.00000657, -0.00000861, 0.00000124, -0.00000591, -0.00000546, -0.00000201, 0.00000063, -0.00000783, -0.00000460, -0.00000528, -0.00001509, -0.00000430, -0.00003052, -0.00000726, -0.00010679, -0.00005281, 0.00018248, 0.00004115, 0.00008700, 0.00004912, 0.00018348, -0.00008986, -0.00007680, -0.00001363, -0.00002331, -0.00000961, -0.00000911, -0.00000532, -0.00000079, -0.00001041, 0.00000843, -0.00000648, -0.00000501, 0.00000341, -0.00000941, 0.00001032, -0.00001767, -0.00000270, -0.00000146, 0.00000104, -0.00000186, -0.00000375, -0.00001248, 0.00000040, 0.00000089, 0.00000394, -0.00000353, -0.00000593, -0.00000237, 0.00000186, -0.00000745, -0.00000134, -0.00000372, 0.00000265, -0.00001434, 0.00000783, 0.00000341, 0.00000459, 0.00000849, -0.00000816, 0.00000118, 0.00000261, 0.00000611, 0.00000278, -0.00000471, 0.00000084, 0.00000244, -0.00000101, 0.00000231, -0.00000020, -0.00000579, 0.00000486, -0.00001134, 0.00000729, 0.00000892, 0.00000466, -0.00001391, 0.00000318, 0.00001243, 0.00000138, -0.00000377, 0.00000129, -0.00000027, -0.00000906, 0.00000744, 0.00000312, -0.00000490, -0.00000727, 0.00000514, -0.00001134, 0.00000810, -0.00000392, 0.00000474, -0.00000504, -0.00000945, -0.00000136, 0.00001153, -0.00000493, -0.00000432, -0.00000478, -0.00000416, -0.00000201, 0.00000076, 0.00000092, -0.00000859, -0.00000438, 0.00000766, -0.00000058, -0.00001300, -0.00000741, 0.00000683, -0.00000470, -0.00000536, -0.00002173, -0.00001268, 0.00001212, 0.00002028, },
		{ 0.00000673, 0.00001828, 0.00000234, 0.00001602, 0.00000032, 0.00001155, 0.00003718, 0.00000819, -0.00001462, -0.00002878, -0.00000493, -0.00001261, 0.00000342, -0.00000736, -0.00000276, -0.00000665, 0.00000430, -0.00000900, -0.00000335, -0.00000642, -0.00001184, -0.00000795, -0.00002222, -0.00001564, -0.00007686, -0.00009257, 0.00017543, 0.00004748, 0.00008439, 0.00003378, 0.00016721, -0.00000467, -0.00011846, -0.00000463, -0.00003297, -0.00000383, -0.00001499, -0.00000153, -0.00000545, -0.00000526, 0.00000131, 0.00000236, -0.00001283, 0.00000898, -0.00001410, 0.00001372, -0.00001826, -0.00000202, -0.00000421, 0.00000309, -0.00000394, 0.00000047, -0.00001460, -0.00000062, 0.00000028, 0.00000473, -0.00000407, -0.00000297, -0.00000546, 0.00000349, -0.00000853, 0.00000011, -0.00000535, 0.00000312, -0.00001092, -0.00000017, 0.00001005, -0.00000185, 0.00001271, -0.00000689, -0.00000084, 0.00000252, 0.00000584, 0.00000403, -0.00000438, -0.00000025, 0.00000282, -0.00000030, 0.00000103, 0.00000152, -0.00000664, 0.00000454, -0.00000880, 0.00000181, 0.00001122, 0.00000538, -0.00001000, -0.00000447, 0.00001554, 0.00000217, -0.00000146, -0.00000291, 0.00000357, -0.00001060, 0.00000673, 0.00000254, -0.00000122, -0.00001046, 0.00000586, -0.00001025, 0.00000537, -0.00000179, 0.00000238, -0.00000077, -0.00001079, -0.00000427, 0.00001121, -0.00000129, -0.00000592, -0.00000412, -0.00000447, -0.00000286, 0.00000077, 0.00000125, -0.00000720, -0.00000689, 0.00000700, 0.00000166, -0.00001140, -0.00001004, 0.00000682, -0.00000353, -0.00000605, -0.00001490, -0.00002222, 0.00000868, 0.00001831, 0.00002320, },
		{ 0.00001755, 0.00000591, 0.00001231, 0.00000764, 0.00000319, 0.00004086, 0.00001186, -0.00001050, -0.00003259, -0.00000424, -0.00001512, 0.00000344, -0.00000705, -0.00000170, -0.00000864, 0.00000433, -0.00000693, -0.00000444, -0.00000586, -0.00001013, -0.00001016, -0.00001598, -0.00002144, -0.00005367, -0.00011580, 0.00015152, 0.00006639, 0.00006919, 0.00004056, 0.00012317, 0.00007889, -0.00013421, -0.00001429, -0.00003194, -0.00000606, -0.00001483, -0.00000222, -0.00000678, -0.00000312, -0.00000287, 0.00000764, -0.00001637, 0.00001042, -0.00001513, 0.00001367, -0.00001587, -0.00000384, -0.00000415, 0.00000220, -0.00000343, 0.00000141, -0.00001264, -0.00000472, 0.00000159, 0.00000392, -0.00000322, -0.00000224, -0.00000520, 0.00000179, -0.00000644, -0.00000212, -0.00000248, -0.00000145, -0.00000322, -0.00000973, 0.00001566, -0.00000548, 0.00001276, -0.00000255, -0.00000374, 0.00000270, 0.00000504, 0.00000555, -0.00000407, -0.00000107, 0.00000271, 0.00000064, 0.00000007, 0.00000236, -0.00000617, 0.00000277, -0.00000518, -0.00000318, 0.00001204, 0.00000628, -0.00000576, -0.00001020, 0.00001522, 0.00000489, -0.00000006, -0.00000505, 0.00000430, -0.00000924, 0.00000415, 0.00000334, 0.00000016, -0.00001009, 0.00000296, -0.00000638, 0.00000080, 0.00000169, -0.00000114, 0.00000347, -0.00001107, -0.00000682, 0.00000919, 0.00000297, -0.00000716, -0.00000365, -0.00000488, -0.00000306, -0.00000039, 0.00000268, -0.00000698, -0.00000746, 0.00000419, 0.00000500, -0.00001076, -0.00001060, 0.00000410, 0.00000054, -0.00001005, -0.00000645, -0.00002848, 0.00000214, 0.00001607, 0.00002717, 0.00000847, },
		{ 0.00001185, 0.00000693, 0.00001538, -0.00000328, 0.00004145, 0.00001715, -0.00000670, -0.00003506, -0.00000537, -0.00001533, 0.00000088, -0.00000482, -0.00000255, -0.00000787, 0.00000126, -0.00000282, -0.00000659, -0.00000454, -0.00000939, -0.00001102, -0.00001216, -0.00002391, -0.00003790, -0.00012406, 0.00011517, 0.00009204, 0.00004807, 0.00005794, 0.00007364, 0.00013672, -0.00011328, -0.00004243, -0.00002230, -0.00001457, -0.00001011, -0.00000628, -0.00000532, -0.00000387, -0.00000403, 0.00000924, -0.00001573, 0.00000811, -0.00001285, 0.00001034, -0.00001053, -0.00000828, -0.00000126, -0.00000140, -0.00000060, -0.00000061, -0.00000753, -0.00001041, 0.00000351, 0.00000232, -0.00000124, -0.00000361, -0.00000226, -0.00000208, -0.00000248, -0.00000646, 0.00000253, -0.00000749, 0.00000410, -0.00001601, 0.00001627, -0.00000368, 0.00000781, 0.00000431, -0.00000689, 0.00000303, 0.00000363, 0.00000736, -0.00000380, -0.00000156, 0.00000218, 0.00000167, -0.00000054, 0.00000252, -0.00000501, 0.00000062, -0.00000199, -0.00000625, 0.00001063, 0.00000793, -0.00000237, -0.00001279, 0.00001132, 0.00000905, 0.00000053, -0.00000509, 0.00000249, -0.00000609, 0.00000075, 0.00000490, -0.00000021, -0.00000734, -0.00000159, -0.00000200, -0.00000344, 0.00000453, -0.00000383, 0.00000589, -0.00000937, -0.00000940, 0.00000633, 0.00000664, -0.00000720, -0.00000373, -0.00000523, -0.00000274, -0.00000229, 0.00000452, -0.00000709, -0.00000714, 0.00000071, 0.00000787, -0.00000978, -0.00001045, 0.00000052, 0.00000500, -0.00001369, -0.00000097, -0.00002814, -0.00000831, 0.00001422, 0.00002817, 0.00001379, 0.00001411, },
		{ 0.00000206, 0.00002125, -0.00000613, 0.00003833, 0.00002401, -0.00000332, -0.00003582, -0.00000862, -0.00001328, -0.00000367, -0.00000155, -0.00000444, -0.00000551, -0.00000301, 0.00000110, -0.00000787, -0.00000393, -0.00000849, -0.00001146, -0.00000978, -0.00002419, -0.00002762, -0.00012181, 0.00007353, 0.00011621, 0.00003016, 0.00007285, 0.00003706, 0.00015667, -0.00005957, -0.00007808, -0.00001256, -0.00002345, -0.00000541, -0.00001026, -0.00000361, -0.00000543, -0.00000399, 0.00000885, -0.00001278, 0.00000394, -0.00000895, 0.00000526, -0.00000370, -0.00001385, 0.00000276, -0.00000574, 0.00000265, -0.00000355, -0.00000186, -0.00001499, 0.00000395, 0.00000142, 0.00000065, -0.00000562, 0.00000126, -0.00000563, 0.00000084, -0.00001009, 0.00000622, -0.00001097, 0.00000692, -0.00001586, 0.00001043, 0.00000333, -0.00000046, 0.00001145, -0.00000877, 0.00000289, 0.00000206, 0.00000892, -0.00000317, -0.00000203, 0.00000152, 0.00000252, -0.00000072, 0.00000228, -0.00000375, -0.00000107, -0.00000012, -0.00000697, 0.00000731, 0.00000997, 0.00000005, -0.00001251, 0.00000526, 0.00001294, 0.00000140, -0.00000410, -0.00000018, -0.00000306, -0.00000195, 0.00000606, -0.00000098, -0.00000421, -0.00000535, 0.00000060, -0.00000551, 0.00000518, -0.00000431, 0.00000558, -0.00000557, -0.00001197, 0.00000328, 0.00000887, -0.00000571, -0.00000442, -0.00000539, -0.00000236, -0.00000409, 0.00000562, -0.00000633, -0.00000733, -0.00000178, 0.00000872, -0.00000722, -0.00001095, -0.00000208, 0.00000744, -0.00001411, -0.00000102, -0.00002096, -0.00002079, 0.00001169, 0.00002639, 0.00002166, 0.00000961, 0.00001829, },
		{ 0.00002348, -0.00000452, 0.00003177, 0.00003176, -0.00000003, -0.00003493, -0.00001368, -0.00000982, -0.00000885, 0.00000132, -0.00000604, -0.00000318, -0.00000643, 0.00000290, -0.00000695, -0.00000487, -0.00000675, -0.00001222, -0.00000784, -0.00002376, -0.00002046, -0.00011368, 0.00003310, 0.00013224, 0.00002203, 0.00007711, 0.00002158, 0.00014164, 0.00001022, -0.00010432, -0.00001270, -0.00002658, -0.00000516, -0.00001088, -0.00000418, -0.00000562, -0.00000480, 0.00000851, -0.00000965, 0.00000004, -0.00000534, 0.00000029, 0.00000262, -0.00001837, 0.00000550, -0.00000836, 0.00000402, -0.00000509, 0.00000173, -0.00001612, 0.00000140, 0.00000226, 0.00000137, -0.00000673, 0.00000324, -0.00000668, 0.00000151, -0.00001086, 0.00000623, -0.00000966, 0.00000387, -0.00000955, 0.00000012, 0.00001244, -0.00000842, 0.00001581, -0.00000776, 0.00000154, 0.00000103, 0.00000952, -0.00000166, -0.00000283, 0.00000110, 0.00000292, -0.00000044, 0.00000182, -0.00000268, -0.00000202, 0.00000041, -0.00000597, 0.00000330, 0.00001134, 0.00000216, -0.00001065, -0.00000074, 0.00001462, 0.00000372, -0.00000331, -0.00000208, -0.00000162, -0.00000296, 0.00000605, -0.00000107, -0.00000229, -0.00000673, 0.00000037, -0.00000491, 0.00000347, -0.00000259, 0.00000302, -0.00000071, -0.00001381, 0.00000014, 0.00000952, -0.00000303, -0.00000532, -0.00000541, -0.00000226, -0.00000507, 0.00000517, -0.00000399, -0.00000877, -0.00000248, 0.00000710, -0.00000296, -0.00001242, -0.00000303, 0.00000710, -0.00001073, -0.00000593, -0.00001002, -0.00003129, 0.00000642, 0.00002304, 0.00002988, 0.00000653, 0.00002287, -0.00000007, },
		{ 0.00000097, 0.00002310, 0.00003909, 0.00000387, -0.00003286, -0.00001970, -0.00000624, -0.00001316, 0.00000253, -0.00000636, -0.00000202, -0.00000782, 0.00000193, -0.00000392, -0.00000693, -0.00000453, -0.00001318, -0.00000608, -0.00002324, -0.00001509, -0.00010265, -0.00000231, 0.00013738, 0.00002509, 0.00006989, 0.00002448, 0.00010590, 0.00007429, -0.00010644, -0.00002800, -0.00002193, -0.00001062, -0.00000743, -0.00000756, -0.00000377, -0.00000723, 0.00000907, -0.00000740, -0.00000250, -0.00000307, -0.00000338, 0.00000697, -0.00002022, 0.00000527, -0.00000773, 0.00000234, -0.00000407, 0.00000203, -0.00001322, -0.00000395, 0.00000452, 0.00000090, -0.00000636, 0.00000278, -0.00000463, -0.00000072, -0.00000857, 0.00000254, -0.00000423, -0.00000315, -0.00000047, -0.00001032, 0.00001919, -0.00001224, 0.00001504, -0.00000315, -0.00000114, 0.00000091, 0.00000873, 0.00000093, -0.00000405, 0.00000103, 0.00000280, 0.00000025, 0.00000124, -0.00000183, -0.00000244, 0.00000022, -0.00000443, 0.00000006, 0.00001105, 0.00000466, -0.00000857, -0.00000488, 0.00001305, 0.00000767, -0.00000308, -0.00000259, -0.00000201, -0.00000249, 0.00000508, -0.00000028, -0.00000199, -0.00000570, -0.00000205, -0.00000270, 0.00000058, 0.00000002, -0.00000019, 0.00000348, -0.00001389, -0.00000332, 0.00000901, -0.00000003, -0.00000580, -0.00000548, -0.00000255, -0.00000498, 0.00000311, -0.00000032, -0.00001105, -0.00000192, 0.00000393, 0.00000181, -0.00001390, -0.00000312, 0.00000497, -0.00000528, -0.00001259, -0.00000008, -0.00003576, -0.00000289, 0.00001904, 0.00003632, 0.00000691, 0.00002403, 0.00000167, 0.00002164, },
		{ 0.00001444, 0.00004440, 0.00000919, -0.00003024, -0.00002569, -0.00000379, -0.00001554, 0.00000156, -0.00000526, -0.00000216, -0.00000724, -0.00000102, -0.00000018, -0.00000872, -0.00000290, -0.00001356, -0.00000498, -0.00002233, -0.00001150, -0.00008980, -0.00003173, 0.00013254, 0.00003636, 0.00005590, 0.00003652, 0.00006714, 0.00011505, -0.00007913, -0.00005510, -0.00001333, -0.00001888, -0.00000230, -0.00001188, -0.00000120, -0.00001027, 0.00000981, -0.00000563, -0.00000398, -0.00000191, -0.00000567, 0.00000903, -0.00001906, 0.00000194, -0.00000410, -0.00000177, -0.00000112, -0.00000024, -0.00000768, -0.00001019, 0.00000658, 0.00000026, -0.00000509, 0.00000053, -0.00000064, -0.00000434, -0.00000485, -0.00000282, 0.00000244, -0.00001029, 0.00000681, -0.00001643, 0.00002015, -0.00000991, 0.00000890, 0.00000403, -0.00000423, 0.00000138, 0.00000676, 0.00000417, -0.00000531, 0.00000107, 0.00000237, 0.00000114, 0.00000067, -0.00000111, -0.00000270, -0.00000001, -0.00000334, -0.00000151, 0.00000886, 0.00000763, -0.00000689, -0.00000657, 0.00000878, 0.00001209, -0.00000264, -0.00000234, -0.00000319, -0.00000178, 0.00000409, 0.00000079, -0.00000265, -0.00000348, -0.00000491, -0.00000071, -0.00000175, 0.00000185, -0.00000230, 0.00000552, -0.00001162, -0.00000716, 0.00000786, 0.00000247, -0.00000537, -0.00000572, -0.00000318, -0.00000408, 0.00000011, 0.00000360, -0.00001298, -0.00000143, 0.00000092, 0.00000533, -0.00001390, -0.00000378, 0.00000290, -0.00000033, -0.00001743, 0.00000493, -0.00003241, -0.00001531, 0.00001408, 0.00003983, 0.00001142, 0.00002183, 0.00000674, 0.00001692, 0.00000849, },
		{ 0.00004641, 0.00001639, -0.00002743, -0.00003091, -0.00000317, -0.00001573, -0.00000120, -0.00000343, -0.00000292, -0.00000569, -0.00000433, 0.00000247, -0.00000881, -0.00000284, -0.00001264, -0.00000513, -0.00002039, -0.00001017, -0.00007528, -0.00005563, 0.00012044, 0.00005104, 0.00004153, 0.00004797, 0.00003875, 0.00012558, -0.00002902, -0.00008288, -0.00000856, -0.00002478, 0.00000058, -0.00001420, -0.00000015, -0.00001206, 0.00000918, -0.00000315, -0.00000545, -0.00000098, -0.00000722, 0.00000935, -0.00001569, -0.00000328, 0.00000083, -0.00000633, 0.00000186, -0.00000306, -0.00000207, -0.00001469, 0.00000650, 0.00000072, -0.00000402, -0.00000211, 0.00000329, -0.00000713, -0.00000193, -0.00000721, 0.00000714, -0.00001396, 0.00000897, -0.00001603, 0.00001471, -0.00000235, -0.00000033, 0.00001118, -0.00000604, 0.00000159, 0.00000440, 0.00000717, -0.00000589, 0.00000075, 0.00000202, 0.00000190, 0.00000029, -0.00000051, -0.00000299, 0.00000007, -0.00000308, -0.00000149, 0.00000548, 0.00001031, -0.00000533, -0.00000653, 0.00000357, 0.00001503, -0.00000073, -0.00000245, -0.00000376, -0.00000210, 0.00000396, 0.00000149, -0.00000335, -0.00000154, -0.00000652, -0.00000042, -0.00000236, 0.00000188, -0.00000243, 0.00000500, -0.00000738, -0.00001085, 0.00000621, 0.00000409, -0.00000406, -0.00000599, -0.00000405, -0.00000285, -0.00000285, 0.00000649, -0.00001328, -0.00000231, -0.00000051, 0.00000632, -0.00001154, -0.00000589, 0.00000213, 0.00000241, -0.00001844, 0.00000384, -0.00002280, -0.00002756, 0.00000682, 0.00004027, 0.00001924, 0.00001779, 0.00001329, 0.00001155, 0.00001562, 0.00000792, },
		{ 0.00002523, -0.00002431, -0.00003514, -0.00000443, -0.00001428, -0.00000469, -0.00000197, -0.00000337, -0.00000432, -0.00000656, 0.00000283, -0.00000666, -0.00000443, -0.00001046, -0.00000653, -0.00001729, -0.00001114, -0.00005955, -0.00007438, 0.00010354, 0.00006523, 0.00003129, 0.00005315, 0.00002572, 0.00011070, 0.00002847, -0.00009751, -0.00001469, -0.00002446, -0.00000149, -0.00001262, -0.00000203, -0.00001137, 0.00000616, 0.00000084, -0.00000763, 0.00000040, -0.00000861, 0.00000871, -0.00001138, -0.00000860, 0.00000491, -0.00000919, 0.00000299, -0.00000451, 0.00000139, -0.00001562, 0.00000336, 0.00000277, -0.00000375, -0.00000398, 0.00000554, -0.00000753, -0.00000128, -0.00000896, 0.00000806, -0.00001273, 0.00000556, -0.00001016, 0.00000534, 0.00000700, -0.00000886, 0.00001525, -0.00000499, 0.00000078, 0.00000252, 0.00000900, -0.00000514, -0.00000031, 0.00000204, 0.00000228, 0.00000023, -0.00000008, -0.00000327, 0.00000034, -0.00000332, -0.00000070, 0.00000223, 0.00001163, -0.00000314, -0.00000603, -0.00000052, 0.00001490, 0.00000336, -0.00000351, -0.00000300, -0.00000384, 0.00000470, 0.00000175, -0.00000361, -0.00000074, -0.00000617, -0.00000205, -0.00000133, 0.00000034, -0.00000098, 0.00000279, -0.00000258, -0.00001333, 0.00000382, 0.00000499, -0.00000238, -0.00000588, -0.00000513, -0.00000174, -0.00000498, 0.00000752, -0.00001138, -0.00000495, -0.00000002, 0.00000479, -0.00000722, -0.00000909, 0.00000258, 0.00000291, -0.00001600, -0.00000155, -0.00001109, -0.00003545, -0.00000384, 0.00003778, 0.00002874, 0.00001411, 0.00001911, 0.00000776, 0.00002040, 0.00000487, 0.00004470, },
		{ -0.00002027, -0.00003859, -0.00000709, -0.00001216, -0.00000775, -0.00000173, -0.00000299, -0.00000371, -0.00000713, 0.00000092, -0.00000299, -0.00000670, -0.00000789, -0.00000842, -0.00001364, -0.00001359, -0.00004394, -0.00008751, 0.00008308, 0.00007713, 0.00002640, 0.00005150, 0.00002519, 0.00008269, 0.00007611, -0.00008924, -0.00003339, -0.00001826, -0.00000803, -0.00000778, -0.00000626, -0.00000856, 0.00000113, 0.00000587, -0.00001020, 0.00000203, -0.00000979, 0.00000750, -0.00000709, -0.00001267, 0.00000672, -0.00000917, 0.00000152, -0.00000385, 0.00000193, -0.00001287, -0.00000216, 0.00000563, -0.00000397, -0.00000486, 0.00000572, -0.00000545, -0.00000277, -0.00000816, 0.00000553, -0.00000774, -0.00000110, -0.00000232, -0.00000381, 0.00001402, -0.00001320, 0.00001426, -0.00000069, -0.00000106, 0.00000148, 0.00000922, -0.00000298, -0.00000202, 0.00000240, 0.00000228, 0.00000050, 0.00000013, -0.00000338, 0.00000047, -0.00000344, -0.00000017, 0.00000030, 0.00001085, 0.00000009, -0.00000603, -0.00000229, 0.00001142, 0.00000898, -0.00000494, -0.00000141, -0.00000620, 0.00000536, 0.00000222, -0.00000360, -0.00000099, -0.00000435, -0.00000462, 0.00000019, -0.00000158, 0.00000072, 0.00000051, 0.00000110, -0.00001362, 0.00000047, 0.00000552, -0.00000095, -0.00000506, -0.00000634, -0.00000108, -0.00000592, 0.00000660, -0.00000766, -0.00000870, 0.00000159, 0.00000195, -0.00000243, -0.00001199, 0.00000302, 0.00000255, -0.00001211, -0.00000784, -0.00000187, -0.00003616, -0.00001733, 0.00003194, 0.00003827, 0.00001269, 0.00002266, 0.00000679, 0.00002204, 0.00000540, 0.00003995, 0.00003469, },
		{ -0.00004168, -0.00001051, -0.00001030, -0.00000964, -0.00000287, -0.00000197, -0.00000369, -0.00000644, -0.00000217, 0.00000061, -0.00000814, -0.00000613, -0.00000977, -0.00001044, -0.00001619, -0.00003026, -0.00009410, 0.00005946, 0.00008652, 0.00002567, 0.00004593, 0.00003038, 0.00005482, 0.00010209, -0.00005724, -0.00005905, -0.00001108, -0.00001552, -0.00000251, -0.00001059, -0.00000541, -0.00000427, 0.00001043, -0.00001196, 0.00000296, -0.00001013, 0.00000557, -0.00000316, -0.00001502, 0.00000603, -0.00000661, -0.00000181, -0.00000184, 0.00000039, -0.00000800, -0.00000799, 0.00000760, -0.00000362, -0.00000536, 0.00000451, -0.00000207, -0.00000502, -0.00000620, 0.00000143, -0.00000173, -0.00000746, 0.00000352, -0.00000909, 0.00001598, -0.00001191, 0.00000844, 0.00000555, -0.00000293, 0.00000098, 0.00000813, 0.00000004, -0.00000381, 0.00000275, 0.00000213, 0.00000094, 0.00000016, -0.00000320, 0.00000018, -0.00000300, -0.00000053, 0.00000011, 0.00000816, 0.00000398, -0.00000657, -0.00000193, 0.00000593, 0.00001423, -0.00000524, -0.00000030, -0.00000766, 0.00000465, 0.00000367, -0.00000377, -0.00000169, -0.00000221, -0.00000667, 0.00000081, -0.00000259, 0.00000136, -0.00000047, 0.00000255, -0.00001147, -0.00000357, 0.00000577, -0.00000012, -0.00000356, -0.00000738, -0.00000110, -0.00000570, 0.00000428, -0.00000316, -0.00001231, 0.00000293, -0.00000056, 0.00000109, -0.00001310, 0.00000211, 0.00000285, -0.00000892, -0.00001211, 0.00000215, -0.00002980, -0.00003088, 0.00002176, 0.00004646, 0.00001457, 0.00002360, 0.00000861, 0.00002108, 0.00000845, 0.00003361, 0.00004335, -0.00001446, },
		{ -0.00001426, -0.00000916, -0.00001033, -0.00000483, -0.00000112, -0.00000353, -0.00000546, -0.00000496, 0.00000257, -0.00000764, -0.00000593, -0.00000995, -0.00000841, -0.00001780, -0.00001999, -0.00009368, 0.00003329, 0.00009351, 0.00002731, 0.00003987, 0.00003507, 0.00003581, 0.00010391, -0.00001005, -0.00008085, -0.00000972, -0.00001969, -0.00000008, -0.00001255, -0.00000391, -0.00000817, 0.00001283, -0.00001161, 0.00000220, -0.00000894, 0.00000270, 0.00000045, -0.00001589, 0.00000363, -0.00000288, -0.00000531, -0.00000003, -0.00000152, -0.00000328, -0.00001186, 0.00000713, -0.00000186, -0.00000615, 0.00000297, 0.00000110, -0.00000645, -0.00000470, -0.00000217, 0.00000266, -0.00001072, 0.00000500, -0.00000918, 0.00001283, -0.00000625, 0.00000025, 0.00001117, -0.00000330, 0.00000028, 0.00000650, 0.00000295, -0.00000493, 0.00000258, 0.00000215, 0.00000134, 0.00000012, -0.00000275, -0.00000055, -0.00000199, -0.00000168, 0.00000118, 0.00000462, 0.00000747, -0.00000688, -0.00000067, 0.00000064, 0.00001686, -0.00000295, -0.00000084, -0.00000713, 0.00000189, 0.00000626, -0.00000421, -0.00000234, -0.00000072, -0.00000717, -0.00000019, -0.00000218, 0.00000051, 0.00000020, 0.00000189, -0.00000772, -0.00000736, 0.00000543, 0.00000028, -0.00000185, -0.00000780, -0.00000193, -0.00000462, 0.00000137, 0.00000095, -0.00001458, 0.00000287, -0.00000154, 0.00000231, -0.00001179, -0.00000062, 0.00000442, -0.00000731, -0.00001339, 0.00000112, -0.00001949, -0.00004043, 0.00000655, 0.00005194, 0.00001995, 0.00002258, 0.00001224, 0.00001884, 0.00001249, 0.00002729, 0.00004990, -0.00000634, -0.00004456, },
		{ -0.00000868, -0.00001036, -0.00000665, -0.00000131, -0.00000261, -0.00000500, -0.00000639, 0.00000217, -0.00000510, -0.00000703, -0.00000913, -0.00000755, -0.00001803, -0.00001356, -0.00008702, 0.00000615, 0.00009748, 0.00003040, 0.00003521, 0.00003628, 0.00002774, 0.00008776, 0.00003833, -0.00008779, -0.00001880, -0.00001841, -0.00000194, -0.00001115, -0.00000486, -0.00000973, 0.00001236, -0.00000872, -0.00000048, -0.00000622, -0.00000097, 0.00000370, -0.00001576, 0.00000061, 0.00000047, -0.00000740, 0.00000028, -0.00000244, -0.00000037, -0.00001248, 0.00000387, 0.00000123, -0.00000729, 0.00000165, 0.00000313, -0.00000625, -0.00000445, -0.00000415, 0.00000421, -0.00001009, 0.00000232, -0.00000550, 0.00000700, 0.00000065, -0.00000690, 0.00001357, -0.00000104, -0.00000108, 0.00000502, 0.00000495, -0.00000484, 0.00000169, 0.00000250, 0.00000155, 0.00000013, -0.00000218, -0.00000147, -0.00000079, -0.00000303, 0.00000248, 0.00000162, 0.00000935, -0.00000606, 0.00000005, -0.00000245, 0.00001556, 0.00000220, -0.00000302, -0.00000473, -0.00000230, 0.00000910, -0.00000429, -0.00000301, -0.00000013, -0.00000612, -0.00000241, -0.00000090, -0.00000126, 0.00000173, 0.00000031, -0.00000398, -0.00000966, 0.00000398, 0.00000069, -0.00000059, -0.00000714, -0.00000354, -0.00000312, -0.00000139, 0.00000386, -0.00001497, 0.00000110, -0.00000081, 0.00000139, -0.00000860, -0.00000444, 0.00000659, -0.00000667, -0.00001266, -0.00000248, -0.00000968, -0.00004267, -0.00001275, 0.00005280, 0.00002855, 0.00002081, 0.00001638, 0.00001667, 0.00001623, 0.00002217, 0.00005359, 0.00000410, -0.00004693, -0.00001837, },
		{ -0.00001049, -0.00000748, -0.00000298, -0.00000084, -0.00000521, -0.00000634, -0.00000014, -0.00000155, -0.00000833, -0.00000818, -0.00000725, -0.00001728, -0.00001037, -0.00007605, -0.00001947, 0.00009708, 0.00003512, 0.00003194, 0.00003441, 0.00002740, 0.00006425, 0.00007459, -0.00007404, -0.00003748, -0.00001344, -0.00000677, -0.00000754, -0.00000736, -0.00000951, 0.00000959, -0.00000408, -0.00000426, -0.00000278, -0.00000469, 0.00000624, -0.00001488, -0.00000227, 0.00000251, -0.00000746, -0.00000116, -0.00000208, 0.00000038, -0.00001019, -0.00000107, 0.00000445, -0.00000805, 0.00000037, 0.00000404, -0.00000472, -0.00000509, -0.00000471, 0.00000333, -0.00000681, -0.00000229, -0.00000107, 0.00000183, 0.00000554, -0.00001032, 0.00001150, 0.00000367, -0.00000275, 0.00000384, 0.00000578, -0.00000363, 0.00000031, 0.00000299, 0.00000166, 0.00000020, -0.00000164, -0.00000224, 0.00000009, -0.00000388, 0.00000305, 0.00000017, 0.00000907, -0.00000374, -0.00000052, -0.00000261, 0.00001071, 0.00000886, -0.00000547, -0.00000184, -0.00000626, 0.00001065, -0.00000308, -0.00000409, -0.00000008, -0.00000430, -0.00000469, 0.00000011, -0.00000279, 0.00000286, -0.00000067, -0.00000173, -0.00000965, 0.00000127, 0.00000136, -0.00000024, -0.00000535, -0.00000559, -0.00000170, -0.00000347, 0.00000527, -0.00001361, -0.00000192, 0.00000095, -0.00000064, -0.00000490, -0.00000796, 0.00000802, -0.00000564, -0.00001175, -0.00000573, -0.00000385, -0.00003696, -0.00003283, 0.00004674, 0.00003960, 0.00001959, 0.00002001, 0.00001546, 0.00001901, 0.00001873, 0.00005439, 0.00001622, -0.00004804, -0.00002321, -0.00000845, },
		{ -0.00000710, -0.00000579, 0.00000114, -0.00000554, -0.00000550, -0.00000311, 0.00000142, -0.00000852, -0.00000801, -0.00000679, -0.00001620, -0.00000925, -0.00006319, -0.00004095, 0.00009102, 0.00004195, 0.00002909, 0.00003155, 0.00002971, 0.00004320, 0.00009123, -0.00004171, -0.00005915, -0.00000972, -0.00001129, -0.00000427, -0.00000947, -0.00000901, 0.00000599, 0.00000080, -0.00000774, 0.00000013, -0.00000744, 0.00000752, -0.00001319, -0.00000481, 0.00000322, -0.00000595, -0.00000349, -0.00000123, -0.00000018, -0.00000655, -0.00000569, 0.00000618, -0.00000749, -0.00000138, 0.00000445, -0.00000282, -0.00000564, -0.00000476, 0.00000141, -0.00000304, -0.00000608, 0.00000128, -0.00000033, 0.00000677, -0.00000941, 0.00000584, 0.00000915, -0.00000357, 0.00000256, 0.00000582, -0.00000191, -0.00000097, 0.00000324, 0.00000186, 0.00000028, -0.00000122, -0.00000266, 0.00000034, -0.00000387, 0.00000252, 0.00000041, 0.00000697, -0.00000048, -0.00000209, -0.00000057, 0.00000424, 0.00001456, -0.00000616, -0.00000026, -0.00000822, 0.00000965, -0.00000004, -0.00000576, -0.00000014, -0.00000267, -0.00000591, -0.00000009, -0.00000323, 0.00000267, -0.00000015, -0.00000159, -0.00000751, -0.00000210, 0.00000207, -0.00000065, -0.00000291, -0.00000740, -0.00000093, -0.00000457, 0.00000532, -0.00001109, -0.00000535, 0.00000276, -0.00000257, -0.00000201, -0.00000997, 0.00000767, -0.00000322, -0.00001191, -0.00000698, -0.00000290, -0.00002590, -0.00004876, 0.00003200, 0.00005151, 0.00002025, 0.00002252, 0.00001551, 0.00002076, 0.00001688, 0.00005272, 0.00002905, -0.00004702, -0.00002914, -0.00000813, -0.00001123, },
		{ -0.00000876, 0.00000227, -0.00000516, -0.00000484, -0.00000539, 0.00000254, -0.00000684, -0.00000895, -0.00000588, -0.00001521, -0.00000917, -0.00005051, -0.00005651, 0.00007898, 0.00005070, 0.00002606, 0.00002946, 0.00003081, 0.00003003, 0.00008860, 0.00000014, -0.00007426, -0.00001247, -0.00001259, -0.00000350, -0.00000961, -0.00000946, 0.00000286, 0.00000456, -0.00000973, 0.00000150, -0.00000852, 0.00000720, -0.00001063, -0.00000717, 0.00000316, -0.00000395, -0.00000541, -0.00000101, -0.00000079, -0.00000334, -0.00000822, 0.00000538, -0.00000520, -0.00000381, 0.00000484, -0.00000135, -0.00000536, -0.00000506, -0.00000033, -0.00000050, -0.00000733, 0.00000042, 0.00000076, 0.00000499, -0.00000574, -0.00000091, 0.00001295, -0.00000224, 0.00000068, 0.00000566, -0.00000047, -0.00000156, 0.00000297, 0.00000231, 0.00000031, -0.00000095, -0.00000267, -0.00000000, -0.00000313, 0.00000114, 0.00000171, 0.00000418, 0.00000260, -0.00000365, 0.00000207, -0.00000133, 0.00001696, -0.00000364, -0.00000102, -0.00000746, 0.00000606, 0.00000431, -0.00000753, -0.00000032, -0.00000168, -0.00000568, -0.00000159, -0.00000259, 0.00000118, 0.00000168, -0.00000291, -0.00000447, -0.00000493, 0.00000213, -0.00000126, -0.00000070, -0.00000820, -0.00000120, -0.00000464, 0.00000433, -0.00000805, -0.00000847, 0.00000393, -0.00000367, -0.00000056, -0.00001012, 0.00000548, 0.00000047, -0.00001298, -0.00000652, -0.00000508, -0.00001418, -0.00005596, 0.00000897, 0.00006130, 0.00002421, 0.00002364, 0.00001678, 0.00002168, 0.00001642, 0.00004913, 0.00004161, -0.00004318, -0.00003621, -0.00000774, -0.00001252, -0.00000595, },
		{ 0.00000166, -0.00000357, -0.00000494, -0.00000622, 0.00000146, -0.00000360, -0.00001048, -0.00000491, -0.00001422, -0.00000962, -0.00003923, -0.00006572, 0.00006205, 0.00006007, 0.00002328, 0.00002854, 0.00002944, 0.00002502, 0.00007317, 0.00003959, -0.00007480, -0.00002389, -0.00001019, -0.00000551, -0.00000769, -0.00001098, 0.00000059, 0.00000671, -0.00000989, 0.00000113, -0.00000790, 0.00000547, -0.00000738, -0.00000946, 0.00000292, -0.00000245, -0.00000599, -0.00000204, -0.00000072, -0.00000160, -0.00000802, 0.00000229, -0.00000185, -0.00000638, 0.00000507, -0.00000045, -0.00000428, -0.00000564, -0.00000146, 0.00000029, -0.00000605, -0.00000274, 0.00000334, 0.00000245, -0.00000198, -0.00000588, 0.00001313, 0.00000173, -0.00000179, 0.00000559, 0.00000027, -0.00000131, 0.00000222, 0.00000288, 0.00000038, -0.00000085, -0.00000237, -0.00000070, -0.00000209, -0.00000041, 0.00000305, 0.00000195, 0.00000442, -0.00000424, 0.00000380, -0.00000412, 0.00001505, 0.00000191, -0.00000355, -0.00000481, 0.00000124, 0.00000841, -0.00000829, -0.00000106, -0.00000117, -0.00000445, -0.00000361, -0.00000166, -0.00000076, 0.00000378, -0.00000426, -0.00000221, -0.00000596, 0.00000096, -0.00000146, 0.00000049, -0.00000749, -0.00000259, -0.00000390, 0.00000272, -0.00000497, -0.00001094, 0.00000428, -0.00000386, -0.00000037, -0.00000892, 0.00000227, 0.00000434, -0.00001381, -0.00000592, -0.00000749, -0.00000614, -0.00005255, -0.00001881, 0.00006475, 0.00003262, 0.00002344, 0.00001908, 0.00002197, 0.00001721, 0.00004409, 0.00005309, -0.00003618, -0.00004404, -0.00000772, -0.00001386, -0.00000494, -0.00001076, },
		{ -0.00000097, -0.00000567, -0.00000578, -0.00000113, 0.00000003, -0.00001147, -0.00000459, -0.00001283, -0.00001051, -0.00002970, -0.00006933, 0.00004234, 0.00006808, 0.00002185, 0.00002801, 0.00002653, 0.00002509, 0.00005368, 0.00006685, -0.00005811, -0.00004137, -0.00000671, -0.00000855, -0.00000510, -0.00001255, -0.00000140, 0.00000769, -0.00000871, -0.00000043, -0.00000622, 0.00000300, -0.00000393, -0.00001154, 0.00000273, -0.00000176, -0.00000517, -0.00000405, -0.00000016, -0.00000122, -0.00000584, -0.00000166, 0.00000110, -0.00000803, 0.00000457, 0.00000038, -0.00000305, -0.00000590, -0.00000233, -0.00000011, -0.00000353, -0.00000614, 0.00000499, 0.00000152, -0.00000031, -0.00000742, 0.00000935, 0.00000746, -0.00000396, 0.00000533, 0.00000050, -0.00000062, 0.00000148, 0.00000325, 0.00000069, -0.00000096, -0.00000189, -0.00000142, -0.00000125, -0.00000151, 0.00000366, 0.00000101, 0.00000457, -0.00000355, 0.00000396, -0.00000373, 0.00000970, 0.00000863, -0.00000593, -0.00000220, -0.00000271, 0.00001050, -0.00000703, -0.00000276, -0.00000073, -0.00000305, -0.00000507, -0.00000141, -0.00000218, 0.00000496, -0.00000433, -0.00000189, -0.00000471, -0.00000138, -0.00000113, 0.00000046, -0.00000545, -0.00000459, -0.00000287, 0.00000097, -0.00000220, -0.00001268, 0.00000394, -0.00000342, -0.00000089, -0.00000724, -0.00000088, 0.00000721, -0.00001316, -0.00000661, -0.00000801, -0.00000370, -0.00004059, -0.00004499, 0.00005759, 0.00004556, 0.00002266, 0.00002203, 0.00002182, 0.00001912, 0.00003806, 0.00006278, -0.00002600, -0.00005194, -0.00000875, -0.00001460, -0.00000489, -0.00001103, -0.00000089, },
		{ -0.00000624, -0.00000491, -0.00000393, 0.00000261, -0.00001090, -0.00000547, -0.00001081, -0.00001182, -0.00002189, -0.00006865, 0.00002221, 0.00007288, 0.00002278, 0.00002686, 0.00002372, 0.00002644, 0.00003720, 0.00007765, -0.00002804, -0.00005827, -0.00000644, -0.00001010, -0.00000370, -0.00001284, -0.00000394, 0.00000822, -0.00000702, -0.00000234, -0.00000433, 0.00000057, -0.00000085, -0.00001309, 0.00000239, -0.00000158, -0.00000369, -0.00000602, 0.00000006, -0.00000135, -0.00000318, -0.00000464, 0.00000222, -0.00000791, 0.00000285, 0.00000163, -0.00000239, -0.00000530, -0.00000336, -0.00000084, -0.00000133, -0.00000789, 0.00000408, 0.00000319, -0.00000117, -0.00000584, 0.00000319, 0.00001279, -0.00000442, 0.00000424, 0.00000079, -0.00000025, 0.00000135, 0.00000306, 0.00000139, -0.00000130, -0.00000136, -0.00000193, -0.00000086, -0.00000190, 0.00000334, 0.00000134, 0.00000349, -0.00000209, 0.00000294, -0.00000117, 0.00000319, 0.00001383, -0.00000601, -0.00000146, -0.00000421, 0.00000968, -0.00000365, -0.00000519, -0.00000029, -0.00000206, -0.00000534, -0.00000227, -0.00000262, 0.00000468, -0.00000267, -0.00000354, -0.00000188, -0.00000397, -0.00000082, -0.00000035, -0.00000287, -0.00000634, -0.00000224, -0.00000047, 0.00000004, -0.00001367, 0.00000303, -0.00000259, -0.00000170, -0.00000562, -0.00000337, 0.00000853, -0.00001074, -0.00000877, -0.00000647, -0.00000566, -0.00002530, -0.00006259, 0.00003769, 0.00006087, 0.00002302, 0.00002488, 0.00002163, 0.00002184, 0.00003178, 0.00006987, -0.00001272, -0.00005913, -0.00001135, -0.00001436, -0.00000614, -0.00000965, -0.00000468, 0.00000169, },
		},
		{ { 0.00000245, 0.00000196, -0.00001714, 0.00000070, -0.00000037, -0.00000294, -0.00000092, -0.00000559, 0.00000387, -0.00000174, -0.00000305, -0.00000407, -0.00000508, -0.00000715, -0.00001506, -0.00001484, -0.00006186, 0.00000433, 0.00007610, 0.00002580, 0.00002440, 0.00002309, 0.00002392, 0.00002752, 0.00007418, 0.00000439, -0.00006516, -0.00001431, -0.00001096, -0.00000615, -0.00000077, -0.00000391, -0.00000415, -0.00000437, -0.00000062, -0.00000455, -0.00000351, -0.00000235, 0.00000898, -0.00000180, -0.00000318, 0.00000033, 0.00000061, 0.00000165, 0.00000178, 0.00000174, -0.00000304, -0.00000143, 0.00000549, -0.00000408, 0.00001685, -0.00000373, -0.00000377, -0.00000240, -0.00000998, 0.00000326, -0.00001307, 0.00000042, -0.00000898, -0.00000237, -0.00000115, -0.00000026, 0.00000097, -0.00000175, -0.00000230, -0.00000175, 0.00000097, -0.00000026, -0.00000115, -0.00000237, -0.00000898, 0.00000042, -0.00001307, 0.00000326, -0.00000998, -0.00000240, -0.00000377, -0.00000373, 0.00001685, -0.00000408, 0.00000549, -0.00000143, -0.00000304, 0.00000174, 0.00000178, 0.00000165, 0.00000061, 0.00000033, -0.00000318, -0.00000180, 0.00000898, -0.00000235, -0.00000351, -0.00000455, -0.00000062, -0.00000437, -0.00000415, -0.00000391, -0.00000077, -0.00000615, -0.00001096, -0.00001431, -0.00006516, 0.00000439, 0.00007418, 0.00002752, 0.00002392, 0.00002309, 0.00002440, 0.00002580, 0.00007610, 0.00000433, -0.00006186, -0.00001484, -0.00001506, -0.00000715, -0.00000508, -0.00000407, -0.00000305, -0.00000174, 0.00000387, -0.00000559, -0.00000092, -0.00000294, -0.00000037, 0.00000070, -0.00001714, 0.00000196, },
		{ 0.00000385, -0.00001165, -0.00000866, 0.00000520, -0.00000589, 0.00000148, -0.00000675, 0.00000161, 0.00000109, -0.00000333, -0.00000319, -0.00000537, -0.00000499, -0.00001456, -0.00001057, -0.00004971, -0.00002668, 0.00007033, 0.00003791, 0.00002014, 0.00002302, 0.00002411, 0.00001923, 0.00005750, 0.00004644, -0.00005462, -0.00003470, -0.00000575, -0.00001168, 0.00000011, -0.00000370, -0.00000276, -0.00000565, -0.00000022, -0.00000447, -0.00000336, -0.00000426, 0.00000727, 0.00000224, -0.00000456, -0.00000028, 0.00000019, 0.00000203, 0.00000152, 0.00000196, -0.00000072, -0.00000471, 0.00000677, -0.00000399, 0.00000960, 0.00000830, -0.00001011, 0.00000359, -0.00001561, 0.00000615, -0.00001035, -0.00000470, -0.00000480, -0.00000651, 0.00000056, -0.00000201, 0.00000180, -0.00000092, -0.00000220, -0.00000215, -0.00000046, 0.00000121, -0.00000212, 0.00000002, -0.00000968, -0.00000028, -0.00000845, -0.00000447, -0.00000341, -0.00000776, 0.00000162, -0.00001092, 0.00001531, 0.00000243, 0.00000109, 0.00000264, -0.00000442, 0.00000075, 0.00000204, 0.00000162, 0.00000059, 0.00000055, -0.00000110, -0.00000475, 0.00000813, 0.00000111, -0.00000424, -0.00000390, -0.00000179, -0.00000251, -0.00000552, -0.00000354, -0.00000328, -0.00000094, -0.00001334, -0.00000629, -0.00005008, -0.00004272, 0.00006986, 0.00004696, 0.00002269, 0.00002414, 0.00002692, 0.00002087, 0.00006787, 0.00004021, -0.00006491, -0.00002523, -0.00001362, -0.00001054, -0.00000452, -0.00000505, -0.00000288, -0.00000327, 0.00000283, -0.00000137, -0.00000492, 0.00000079, -0.00000516, 0.00000545, -0.00001527, -0.00000463, 0.00000337, },
		{ -0.00000350, -0.00001571, 0.00000550, -0.00000400, -0.00000030, -0.00000392, -0.00000297, 0.00000360, -0.00000279, -0.00000272, -0.00000525, -0.00000395, -0.00001263, -0.00001006, -0.00003521, -0.00004624, 0.00005103, 0.00005245, 0.00001772, 0.00002248, 0.00002287, 0.00001901, 0.00003580, 0.00006643, -0.00002057, -0.00005406, -0.00000792, -0.00001163, -0.00000358, -0.00000152, -0.00000281, -0.00000493, -0.00000170, -0.00000274, -0.00000461, -0.00000410, 0.00000337, 0.00000650, -0.00000512, -0.00000060, -0.00000115, 0.00000287, 0.00000102, 0.00000227, 0.00000063, -0.00000479, 0.00000328, 0.00000181, -0.00000117, 0.00001748, -0.00001074, 0.00000426, -0.00001411, -0.00000006, -0.00000033, -0.00001243, -0.00000078, -0.00000942, 0.00000081, -0.00000250, 0.00000128, 0.00000031, -0.00000184, -0.00000232, -0.00000160, 0.00000127, -0.00000141, -0.00000007, -0.00000694, -0.00000464, -0.00000192, -0.00000982, -0.00000160, -0.00000899, 0.00000298, -0.00001128, 0.00000702, 0.00001027, -0.00000126, 0.00000387, -0.00000299, -0.00000154, 0.00000249, 0.00000150, 0.00000086, -0.00000015, 0.00000129, -0.00000606, 0.00000530, 0.00000466, -0.00000416, -0.00000373, -0.00000245, -0.00000149, -0.00000553, -0.00000411, -0.00000449, -0.00000011, -0.00000904, -0.00001001, -0.00002494, -0.00007060, 0.00003538, 0.00007185, 0.00002460, 0.00002601, 0.00002659, 0.00002319, 0.00005114, 0.00007096, -0.00005279, -0.00004190, -0.00001117, -0.00001452, -0.00000421, -0.00000583, -0.00000309, -0.00000364, 0.00000023, 0.00000238, -0.00000638, 0.00000107, -0.00000510, 0.00000330, -0.00000731, -0.00001264, 0.00000401, 0.00000290, },
		{ -0.00001548, -0.00000116, 0.00000173, -0.00000413, -0.00000010, -0.00000645, 0.00000348, -0.00000077, -0.00000261, -0.00000491, -0.00000351, -0.00001024, -0.00001085, -0.00002306, -0.00005293, 0.00002372, 0.00006271, 0.00002034, 0.00002053, 0.00002114, 0.00002148, 0.00002035, 0.00006310, 0.00002009, -0.00005618, -0.00002236, -0.00000659, -0.00000950, 0.00000041, -0.00000309, -0.00000330, -0.00000377, -0.00000072, -0.00000587, -0.00000347, -0.00000059, 0.00000900, -0.00000392, -0.00000123, -0.00000227, 0.00000279, 0.00000134, 0.00000208, 0.00000141, -0.00000297, -0.00000163, 0.00000712, -0.00000697, 0.00001661, -0.00000214, -0.00000190, -0.00000555, -0.00001168, 0.00000982, -0.00001559, -0.00000167, -0.00000853, -0.00000166, -0.00000124, -0.00000040, 0.00000149, -0.00000112, -0.00000236, -0.00000215, 0.00000011, 0.00000014, -0.00000116, -0.00000341, -0.00000838, 0.00000117, -0.00000913, -0.00000466, -0.00000784, 0.00000196, -0.00000777, -0.00000219, 0.00001362, 0.00000125, 0.00000207, -0.00000009, -0.00000335, 0.00000191, 0.00000196, 0.00000099, -0.00000083, 0.00000249, -0.00000485, 0.00000089, 0.00000758, -0.00000317, -0.00000371, -0.00000308, -0.00000078, -0.00000498, -0.00000477, -0.00000490, -0.00000197, -0.00000294, -0.00001437, -0.00000940, -0.00006617, -0.00001937, 0.00008502, 0.00003685, 0.00002669, 0.00002560, 0.00002829, 0.00003488, 0.00008724, -0.00002358, -0.00006132, -0.00001045, -0.00001772, -0.00000482, -0.00000633, -0.00000337, -0.00000388, -0.00000167, 0.00000306, -0.00000384, -0.00000221, -0.00000157, -0.00000189, 0.00000052, -0.00001630, 0.00000093, 0.00000265, 0.00000173, },
		{ -0.00000991, 0.00000561, -0.00000515, 0.00000068, -0.00000602, 0.00000007, 0.00000201, -0.00000224, -0.00000455, -0.00000345, -0.00000782, -0.00001164, -0.00001483, -0.00004968, -0.00000384, 0.00006335, 0.00002865, 0.00001765, 0.00001969, 0.00002279, 0.00001455, 0.00004611, 0.00004966, -0.00003621, -0.00004132, -0.00000506, -0.00001139, -0.00000185, -0.00000129, -0.00000292, -0.00000416, -0.00000059, -0.00000513, -0.00000433, -0.00000250, 0.00000816, -0.00000024, -0.00000302, -0.00000171, 0.00000048, 0.00000334, 0.00000063, 0.00000278, -0.00000172, -0.00000394, 0.00000705, -0.00000410, 0.00000639, 0.00001079, -0.00000873, 0.00000280, -0.00001938, 0.00001113, -0.00000916, -0.00000841, -0.00000466, -0.00000559, 0.00000048, -0.00000203, 0.00000191, -0.00000012, -0.00000208, -0.00000237, -0.00000128, 0.00000100, -0.00000145, -0.00000139, -0.00000897, -0.00000045, -0.00000431, -0.00000823, -0.00000801, 0.00000080, -0.00000340, -0.00000850, 0.00001137, 0.00000608, 0.00000102, 0.00000082, -0.00000228, -0.00000082, 0.00000349, 0.00000040, -0.00000045, 0.00000145, -0.00000114, -0.00000409, 0.00000930, -0.00000170, -0.00000302, -0.00000442, 0.00000038, -0.00000490, -0.00000436, -0.00000595, -0.00000322, -0.00000076, -0.00001149, -0.00000985, -0.00003941, -0.00006679, 0.00006718, 0.00006191, 0.00002641, 0.00002638, 0.00003130, 0.00002624, 0.00008539, 0.00001786, -0.00007595, -0.00001534, -0.00001856, -0.00000701, -0.00000648, -0.00000360, -0.00000447, -0.00000223, 0.00000110, 0.00000039, -0.00000544, 0.00000095, -0.00000469, 0.00000310, -0.00001321, -0.00000575, 0.00000354, 0.00000287, -0.00000937, },
		{ 0.00000304, -0.00000135, -0.00000208, -0.00000261, -0.00000432, 0.00000342, -0.00000083, -0.00000414, -0.00000390, -0.00000532, -0.00001220, -0.00000997, -0.00004128, -0.00002518, 0.00005278, 0.00004026, 0.00001548, 0.00001884, 0.00002139, 0.00001623, 0.00002693, 0.00006024, -0.00000352, -0.00005035, -0.00001440, -0.00000702, -0.00000799, 0.00000158, -0.00000346, -0.00000295, -0.00000213, -0.00000278, -0.00000626, -0.00000273, 0.00000491, 0.00000452, -0.00000505, -0.00000006, -0.00000293, 0.00000552, -0.00000077, 0.00000396, -0.00000108, -0.00000348, 0.00000279, 0.00000301, -0.00000449, 0.00001788, -0.00000779, 0.00000354, -0.00001681, 0.00000126, 0.00000393, -0.00001572, -0.00000227, -0.00000787, 0.00000060, -0.00000238, 0.00000121, 0.00000090, -0.00000133, -0.00000254, -0.00000212, 0.00000048, -0.00000054, -0.00000110, -0.00000698, -0.00000429, 0.00000011, -0.00000833, -0.00001051, -0.00000101, 0.00000133, -0.00001170, 0.00000644, 0.00000869, 0.00000370, -0.00000122, 0.00000092, -0.00000398, 0.00000442, 0.00000021, 0.00000024, -0.00000060, 0.00000270, -0.00000708, 0.00000834, 0.00000084, -0.00000238, -0.00000535, 0.00000071, -0.00000415, -0.00000409, -0.00000659, -0.00000398, -0.00000206, -0.00000435, -0.00001624, -0.00001473, -0.00008087, 0.00001508, 0.00008819, 0.00003158, 0.00002804, 0.00003125, 0.00002616, 0.00006992, 0.00006022, -0.00007632, -0.00002909, -0.00001615, -0.00001127, -0.00000593, -0.00000434, -0.00000472, -0.00000263, -0.00000101, 0.00000275, -0.00000530, -0.00000013, -0.00000342, 0.00000077, -0.00000664, -0.00001171, 0.00000279, 0.00000267, -0.00000286, -0.00001411, },
		{ 0.00000358, -0.00000447, -0.00000020, -0.00000611, 0.00000168, 0.00000160, -0.00000324, -0.00000491, -0.00000292, -0.00001225, -0.00000767, -0.00003147, -0.00003756, 0.00003386, 0.00005049, 0.00001617, 0.00001816, 0.00001824, 0.00002048, 0.00001379, 0.00005384, 0.00002754, -0.00004135, -0.00003123, -0.00000291, -0.00001187, 0.00000093, -0.00000200, -0.00000268, -0.00000281, -0.00000139, -0.00000679, -0.00000359, 0.00000195, 0.00000753, -0.00000499, 0.00000031, -0.00000450, 0.00000498, 0.00000032, 0.00000294, 0.00000080, -0.00000301, -0.00000084, 0.00000704, -0.00000756, 0.00001332, 0.00000249, -0.00000267, -0.00000708, -0.00001245, 0.00001387, -0.00001584, -0.00000543, -0.00000671, -0.00000157, -0.00000138, -0.00000020, 0.00000157, -0.00000028, -0.00000247, -0.00000245, -0.00000083, 0.00000051, -0.00000122, -0.00000451, -0.00000716, 0.00000113, -0.00000476, -0.00001277, -0.00000545, 0.00000634, -0.00001222, 0.00000126, 0.00000778, 0.00000822, -0.00000249, 0.00000208, -0.00000378, 0.00000208, 0.00000227, -0.00000048, -0.00000125, 0.00000366, -0.00000567, 0.00000360, 0.00000515, -0.00000311, -0.00000401, -0.00000166, -0.00000098, -0.00000583, -0.00000487, -0.00000634, -0.00000272, -0.00000068, -0.00001577, -0.00000871, -0.00005925, -0.00004851, 0.00009173, 0.00005076, 0.00002811, 0.00003076, 0.00003024, 0.00005072, 0.00009058, -0.00005576, -0.00005138, -0.00001168, -0.00001708, -0.00000466, -0.00000618, -0.00000374, -0.00000419, -0.00000134, 0.00000171, -0.00000183, -0.00000337, -0.00000069, -0.00000227, -0.00000185, -0.00001302, -0.00000126, 0.00000312, 0.00000054, -0.00001179, -0.00000432, },
		{ -0.00000263, -0.00000139, -0.00000412, -0.00000241, 0.00000353, -0.00000129, -0.00000604, -0.00000142, -0.00001101, -0.00000768, -0.00002201, -0.00004200, 0.00001229, 0.00005436, 0.00002137, 0.00001659, 0.00001591, 0.00002220, 0.00001029, 0.00003743, 0.00004742, -0.00001853, -0.00004316, -0.00000818, -0.00000847, -0.00000540, 0.00000220, -0.00000419, -0.00000151, -0.00000212, -0.00000493, -0.00000600, 0.00000057, 0.00000750, -0.00000197, -0.00000167, -0.00000288, 0.00000092, 0.00000407, -0.00000024, 0.00000409, -0.00000370, -0.00000151, 0.00000545, -0.00000244, 0.00000160, 0.00001404, -0.00000717, 0.00000029, -0.00001931, 0.00001234, -0.00000595, -0.00001289, -0.00000420, -0.00000426, -0.00000038, -0.00000137, 0.00000159, 0.00000070, -0.00000183, -0.00000269, -0.00000194, 0.00000059, -0.00000071, -0.00000295, -0.00000773, -0.00000094, -0.00000044, -0.00001168, -0.00001248, 0.00000929, -0.00000891, -0.00000435, 0.00000594, 0.00000970, 0.00000134, -0.00000159, 0.00000087, -0.00000323, 0.00000581, -0.00000223, -0.00000019, 0.00000144, -0.00000075, -0.00000300, 0.00000958, -0.00000439, -0.00000106, -0.00000569, 0.00000319, -0.00000826, -0.00000207, -0.00000936, -0.00000196, -0.00000241, -0.00000713, -0.00001675, -0.00002546, -0.00008782, 0.00005455, 0.00008190, 0.00002930, 0.00003109, 0.00003416, 0.00003635, 0.00010133, -0.00001544, -0.00007576, -0.00000960, -0.00002208, -0.00000399, -0.00000844, -0.00000206, -0.00000650, -0.00000042, -0.00000110, 0.00000195, -0.00000528, 0.00000013, -0.00000286, -0.00000088, -0.00001021, -0.00000635, 0.00000290, 0.00000178, -0.00000662, -0.00001025, 0.00000393, },
		{ -0.00000405, -0.00000124, -0.00000546, 0.00000277, 0.00000187, -0.00000640, -0.00000167, -0.00000801, -0.00000958, -0.00001370, -0.00004098, -0.00000665, 0.00004933, 0.00003058, 0.00001400, 0.00001611, 0.00001922, 0.00001475, 0.00001935, 0.00005337, 0.00000752, -0.00004041, -0.00002340, -0.00000160, -0.00001239, 0.00000465, -0.00000440, -0.00000066, -0.00000282, -0.00000284, -0.00000789, -0.00000079, 0.00000599, 0.00000210, -0.00000421, -0.00000009, -0.00000379, 0.00000718, -0.00000241, 0.00000598, -0.00000338, -0.00000138, 0.00000189, 0.00000399, -0.00000701, 0.00001665, -0.00000256, -0.00000054, -0.00001549, 0.00000054, 0.00000770, -0.00001774, -0.00000489, -0.00000494, -0.00000080, -0.00000163, 0.00000100, 0.00000126, -0.00000062, -0.00000285, -0.00000251, -0.00000036, 0.00000020, -0.00000210, -0.00000676, -0.00000384, 0.00000164, -0.00000683, -0.00001872, 0.00000660, -0.00000090, -0.00001059, 0.00000532, 0.00000619, 0.00000936, -0.00000689, 0.00000556, -0.00000678, 0.00000678, -0.00000192, 0.00000005, -0.00000089, 0.00000346, -0.00000691, 0.00001038, -0.00000310, 0.00000010, -0.00000734, 0.00000405, -0.00000720, -0.00000233, -0.00000910, -0.00000330, -0.00000439, 0.00000043, -0.00002189, -0.00000725, -0.00008276, -0.00001430, 0.00010423, 0.00004153, 0.00002996, 0.00003729, 0.00002959, 0.00009382, 0.00003452, -0.00009071, -0.00001628, -0.00002279, -0.00000640, -0.00000911, -0.00000175, -0.00000746, -0.00000059, -0.00000264, 0.00000290, -0.00000384, -0.00000166, -0.00000149, -0.00000166, -0.00000704, -0.00000881, 0.00000008, 0.00000296, -0.00000291, -0.00001053, -0.00000152, 0.00000219, },
		{ -0.00000129, -0.00000474, -0.00000114, 0.00000503, -0.00000499, -0.00000370, -0.00000402, -0.00001190, -0.00000770, -0.00003625, -0.00002053, 0.00003690, 0.00004034, 0.00001243, 0.00001757, 0.00001409, 0.00002087, 0.00000767, 0.00004644, 0.00002887, -0.00002436, -0.00003795, -0.00000168, -0.00001223, 0.00000022, -0.00000023, -0.00000274, -0.00000114, -0.00000289, -0.00000708, -0.00000385, 0.00000499, 0.00000475, -0.00000476, 0.00000087, -0.00000550, 0.00000623, -0.00000072, 0.00000407, -0.00000000, -0.00000304, 0.00000056, 0.00000575, -0.00000665, 0.00000833, 0.00000883, -0.00000541, -0.00000738, -0.00001159, 0.00001457, -0.00001344, -0.00001077, -0.00000352, -0.00000226, -0.00000142, 0.00000025, 0.00000129, 0.00000065, -0.00000250, -0.00000286, -0.00000151, 0.00000054, -0.00000115, -0.00000550, -0.00000573, 0.00000081, -0.00000089, -0.00002003, -0.00000282, 0.00000917, -0.00001477, 0.00000441, 0.00000121, 0.00001515, -0.00000688, 0.00000422, -0.00000391, 0.00000220, 0.00000231, -0.00000232, -0.00000072, 0.00000363, -0.00000527, 0.00000576, 0.00000221, -0.00000247, -0.00000428, -0.00000046, -0.00000113, -0.00000698, -0.00000438, -0.00000823, -0.00000283, 0.00000019, -0.00001487, -0.00001191, -0.00004653, -0.00007829, 0.00008959, 0.00007144, 0.00002712, 0.00003999, 0.00002922, 0.00007551, 0.00007993, -0.00008522, -0.00003551, -0.00001772, -0.00001318, -0.00000684, -0.00000441, -0.00000550, -0.00000341, -0.00000142, 0.00000047, -0.00000013, -0.00000396, -0.00000061, -0.00000149, -0.00000593, -0.00000801, -0.00000415, 0.00000353, -0.00000107, -0.00000693, -0.00000781, 0.00000433, -0.00000355, },
		{ -0.00000153, -0.00000550, 0.00000574, -0.00000121, -0.00000649, -0.00000079, -0.00001259, -0.00000528, -0.00002870, -0.00002952, 0.00002127, 0.00004584, 0.00001485, 0.00001718, 0.00001162, 0.00002195, 0.00000689, 0.00003038, 0.00004247, -0.00000402, -0.00003989, -0.00001412, -0.00000424, -0.00000923, 0.00000567, -0.00000568, 0.00000138, -0.00000395, -0.00000471, -0.00000738, 0.00000393, 0.00000587, -0.00000307, -0.00000058, -0.00000371, 0.00000160, 0.00000368, -0.00000011, 0.00000474, -0.00000535, 0.00000111, 0.00000344, -0.00000062, -0.00000301, 0.00001682, -0.00000506, -0.00000404, -0.00001515, 0.00000966, -0.00000111, -0.00001766, -0.00000351, -0.00000268, -0.00000189, -0.00000012, 0.00000095, 0.00000141, -0.00000136, -0.00000322, -0.00000227, -0.00000009, -0.00000003, -0.00000428, -0.00000638, -0.00000146, 0.00000283, -0.00001499, -0.00001534, 0.00001525, -0.00001266, 0.00000014, -0.00000031, 0.00001314, 0.00000163, -0.00000365, 0.00000381, -0.00000506, 0.00000749, -0.00000503, 0.00000065, 0.00000098, -0.00000032, -0.00000112, 0.00000838, -0.00000601, 0.00000054, -0.00000660, 0.00000542, -0.00001127, -0.00000000, -0.00001230, -0.00000094, -0.00000424, -0.00000256, -0.00002367, -0.00001300, -0.00010044, 0.00002904, 0.00010682, 0.00003144, 0.00003930, 0.00003436, 0.00005419, 0.00010972, -0.00005535, -0.00006395, -0.00001044, -0.00002191, -0.00000332, -0.00000888, -0.00000181, -0.00000757, 0.00000099, -0.00000291, 0.00000283, -0.00000431, -0.00000153, -0.00000020, -0.00000564, -0.00000696, -0.00000620, 0.00000127, 0.00000104, -0.00000402, -0.00000926, 0.00000016, 0.00000127, -0.00000421, },
		{ -0.00000639, 0.00000221, 0.00000420, -0.00000826, 0.00000012, -0.00001058, -0.00000637, -0.00001971, -0.00003431, 0.00000629, 0.00004426, 0.00002211, 0.00001366, 0.00001388, 0.00001649, 0.00001438, 0.00001278, 0.00004657, 0.00001362, -0.00002747, -0.00003153, 0.00000192, -0.00001472, 0.00000612, -0.00000433, 0.00000070, -0.00000265, -0.00000395, -0.00000842, 0.00000081, 0.00000699, -0.00000087, -0.00000244, -0.00000108, -0.00000297, 0.00000664, -0.00000241, 0.00000687, -0.00000486, 0.00000021, 0.00000161, 0.00000386, -0.00000809, 0.00001402, 0.00000399, -0.00000697, -0.00001091, -0.00000148, 0.00001016, -0.00001800, -0.00000850, -0.00000094, -0.00000322, -0.00000030, 0.00000062, 0.00000141, 0.00000022, -0.00000321, -0.00000281, -0.00000107, 0.00000050, -0.00000271, -0.00000649, -0.00000337, 0.00000294, -0.00000621, -0.00002451, 0.00001216, -0.00000315, -0.00000732, 0.00000233, 0.00000454, 0.00001362, -0.00001094, 0.00000883, -0.00000803, 0.00000796, -0.00000382, -0.00000029, -0.00000053, 0.00000297, -0.00000484, 0.00001036, -0.00000572, 0.00000183, -0.00000830, 0.00000613, -0.00000920, -0.00000168, -0.00001019, -0.00000390, -0.00000533, 0.00000302, -0.00002421, -0.00000532, -0.00007495, -0.00005074, 0.00011646, 0.00005590, 0.00003202, 0.00004413, 0.00003584, 0.00011920, -0.00000631, -0.00009108, -0.00000874, -0.00002729, -0.00000270, -0.00001161, 0.00000016, -0.00000978, 0.00000140, -0.00000408, 0.00000266, -0.00000211, -0.00000326, 0.00000029, -0.00000371, -0.00000823, -0.00000465, -0.00000354, 0.00000351, -0.00000335, -0.00000567, -0.00000715, 0.00000556, -0.00000544, -0.00000019, },
		{ -0.00000402, 0.00000848, -0.00000710, -0.00000177, -0.00000655, -0.00000918, -0.00001159, -0.00003493, -0.00000635, 0.00003632, 0.00003144, 0.00000957, 0.00001791, 0.00000951, 0.00002160, 0.00000290, 0.00003951, 0.00002742, -0.00000930, -0.00003990, -0.00000411, -0.00001004, -0.00000193, 0.00000234, -0.00000366, 0.00000132, -0.00000532, -0.00000660, -0.00000412, 0.00000803, 0.00000107, -0.00000329, 0.00000013, -0.00000466, 0.00000566, -0.00000092, 0.00000488, -0.00000069, -0.00000303, 0.00000218, 0.00000385, -0.00000524, 0.00000337, 0.00001492, -0.00000868, -0.00000732, -0.00000857, 0.00001166, -0.00000866, -0.00001702, 0.00000054, -0.00000362, -0.00000129, 0.00000079, 0.00000085, 0.00000146, -0.00000231, -0.00000349, -0.00000172, -0.00000004, -0.00000079, -0.00000627, -0.00000440, 0.00000064, 0.00000170, -0.00002552, -0.00000001, 0.00000918, -0.00001329, 0.00000494, -0.00000383, 0.00001998, -0.00000946, 0.00000519, -0.00000303, 0.00000202, 0.00000188, -0.00000405, 0.00000046, 0.00000246, -0.00000350, 0.00000685, -0.00000053, -0.00000171, -0.00000423, 0.00000032, -0.00000116, -0.00000833, -0.00000344, -0.00001041, -0.00000234, -0.00000031, -0.00001206, -0.00001760, -0.00003150, -0.00010324, 0.00007453, 0.00009936, 0.00002447, 0.00005284, 0.00002664, 0.00010854, 0.00005093, -0.00010425, -0.00001952, -0.00002542, -0.00000791, -0.00001012, -0.00000199, -0.00000800, -0.00000174, -0.00000209, -0.00000033, 0.00000099, -0.00000389, -0.00000121, -0.00000008, -0.00001055, -0.00000253, -0.00000691, 0.00000340, -0.00000243, -0.00000194, -0.00001062, 0.00000342, -0.00000130, -0.00000304, -0.00000298, },
		{ 0.00000810, -0.00000205, -0.00000545, -0.00000231, -0.00001135, -0.00000641, -0.00003149, -0.00001639, 0.00002501, 0.00003830, 0.00000906, 0.00001877, 0.00000746, 0.00002120, 0.00000512, 0.00002357, 0.00003725, 0.00000589, -0.00003239, -0.00002103, -0.00000047, -0.00001182, 0.00000789, -0.00000638, 0.00000356, -0.00000523, -0.00000519, -0.00000755, 0.00000658, 0.00000395, -0.00000372, 0.00000019, -0.00000400, 0.00000232, 0.00000212, 0.00000151, 0.00000392, -0.00000553, 0.00000257, 0.00000239, -0.00000019, -0.00000563, 0.00001777, -0.00000200, -0.00000987, -0.00000760, 0.00000397, 0.00000431, -0.00002181, -0.00000290, -0.00000103, -0.00000373, 0.00000140, 0.00000019, 0.00000188, -0.00000063, -0.00000398, -0.00000219, -0.00000107, 0.00000050, -0.00000511, -0.00000523, -0.00000175, 0.00000525, -0.00001824, -0.00001549, 0.00001671, -0.00001260, 0.00000337, -0.00000633, 0.00001599, 0.00000164, -0.00000460, 0.00000545, -0.00000498, 0.00000702, -0.00000659, 0.00000106, 0.00000078, -0.00000032, 0.00000179, 0.00000536, -0.00000579, 0.00000073, -0.00000595, 0.00000572, -0.00001245, 0.00000033, -0.00001323, -0.00000172, -0.00000446, -0.00000028, -0.00002716, -0.00000650, -0.00010068, -0.00000977, 0.00013402, 0.00003474, 0.00005060, 0.00003221, 0.00008164, 0.00010372, -0.00009429, -0.00004387, -0.00001761, -0.00001748, -0.00000578, -0.00000702, -0.00000373, -0.00000664, 0.00000095, -0.00000354, 0.00000261, -0.00000252, -0.00000368, 0.00000259, -0.00001020, -0.00000403, -0.00000538, -0.00000109, 0.00000068, -0.00000179, -0.00000709, -0.00000497, 0.00000548, -0.00000672, 0.00000076, -0.00000828, },
		{ 0.00000529, -0.00000861, 0.00000020, -0.00001145, -0.00000465, -0.00002522, -0.00002355, 0.00001310, 0.00003965, 0.00001400, 0.00001466, 0.00001176, 0.00001357, 0.00001460, 0.00000738, 0.00003968, 0.00001688, -0.00001521, -0.00003558, 0.00000139, -0.00001348, 0.00000493, -0.00000242, 0.00000037, -0.00000101, -0.00000664, -0.00000721, 0.00000150, 0.00000806, -0.00000414, 0.00000004, -0.00000293, -0.00000044, 0.00000345, 0.00000005, 0.00000560, -0.00000440, 0.00000027, 0.00000275, 0.00000190, -0.00000720, 0.00001030, 0.00001060, -0.00001412, -0.00000464, -0.00000358, 0.00001049, -0.00001630, -0.00001267, 0.00000344, -0.00000612, 0.00000132, 0.00000016, 0.00000144, 0.00000104, -0.00000351, -0.00000308, -0.00000157, 0.00000022, -0.00000271, -0.00000628, -0.00000291, 0.00000414, -0.00000698, -0.00002663, 0.00001429, -0.00000462, -0.00000194, -0.00000320, 0.00000503, 0.00001476, -0.00001150, 0.00000893, -0.00000606, 0.00000641, -0.00000427, -0.00000149, 0.00000093, 0.00000089, -0.00000056, 0.00000759, -0.00000587, 0.00000148, -0.00000694, 0.00000561, -0.00000886, -0.00000331, -0.00000873, -0.00000693, -0.00000353, 0.00000167, -0.00002132, -0.00001031, -0.00005821, -0.00009011, 0.00012045, 0.00007590, 0.00003515, 0.00005041, 0.00004882, 0.00013888, -0.00005757, -0.00007634, -0.00000991, -0.00002654, -0.00000273, -0.00001131, -0.00000053, -0.00000992, 0.00000196, -0.00000446, 0.00000163, -0.00000003, -0.00000489, 0.00000184, -0.00000550, -0.00000942, -0.00000055, -0.00000695, 0.00000361, -0.00000350, -0.00000067, -0.00001240, 0.00000748, -0.00000540, 0.00000002, -0.00000690, 0.00000209, },
		{ -0.00000831, -0.00000023, -0.00000927, -0.00000543, -0.00001804, -0.00002747, 0.00000230, 0.00003516, 0.00002254, 0.00000849, 0.00001786, 0.00000585, 0.00002124, 0.00000070, 0.00003138, 0.00002643, 0.00000066, -0.00003574, -0.00001026, -0.00000562, -0.00000510, 0.00000536, -0.00000515, 0.00000424, -0.00000835, -0.00000550, -0.00000416, 0.00001047, -0.00000277, -0.00000103, -0.00000168, -0.00000194, 0.00000287, 0.00000036, 0.00000478, -0.00000086, -0.00000308, 0.00000376, 0.00000177, -0.00000404, -0.00000011, 0.00001887, -0.00001088, -0.00000779, -0.00000333, 0.00000568, -0.00000253, -0.00002294, 0.00000453, -0.00000520, -0.00000104, 0.00000125, 0.00000045, 0.00000200, -0.00000185, -0.00000431, -0.00000147, -0.00000122, -0.00000013, -0.00000673, -0.00000340, 0.00000086, 0.00000251, -0.00002845, 0.00000259, 0.00000599, -0.00000680, 0.00000135, -0.00000556, 0.00002103, -0.00000885, 0.00000400, -0.00000047, 0.00000113, 0.00000099, -0.00000532, 0.00000187, 0.00000048, -0.00000045, 0.00000653, -0.00000242, -0.00000142, -0.00000353, 0.00000048, -0.00000106, -0.00000971, -0.00000235, -0.00001255, -0.00000151, -0.00000192, -0.00000816, -0.00002384, -0.00001799, -0.00011856, 0.00004462, 0.00013220, 0.00002352, 0.00006635, 0.00002714, 0.00014455, 0.00000358, -0.00010716, -0.00000941, -0.00003064, -0.00000413, -0.00001238, -0.00000059, -0.00000989, -0.00000005, -0.00000294, -0.00000085, 0.00000167, -0.00000348, -0.00000195, 0.00000098, -0.00001423, 0.00000190, -0.00000864, 0.00000224, -0.00000293, 0.00000226, -0.00001195, 0.00000065, 0.00000191, -0.00000484, -0.00000160, -0.00000580, 0.00001056, },
		{ -0.00000339, -0.00000559, -0.00000752, -0.00001138, -0.00002849, -0.00000631, 0.00002633, 0.00003093, 0.00000484, 0.00002008, 0.00000451, 0.00001901, 0.00000567, 0.00001591, 0.00003342, 0.00001100, -0.00002243, -0.00002679, 0.00000120, -0.00001198, 0.00000798, -0.00000559, 0.00000436, -0.00000548, -0.00000669, -0.00000609, 0.00000783, 0.00000227, -0.00000404, 0.00000052, -0.00000348, 0.00000254, -0.00000007, 0.00000431, 0.00000158, -0.00000389, 0.00000227, 0.00000293, -0.00000191, -0.00000515, 0.00001599, 0.00000208, -0.00001631, 0.00000177, -0.00000300, 0.00000876, -0.00002420, -0.00000280, 0.00000055, -0.00000551, 0.00000282, -0.00000045, 0.00000209, 0.00000022, -0.00000480, -0.00000182, -0.00000225, 0.00000080, -0.00000530, -0.00000447, -0.00000167, 0.00000651, -0.00002111, -0.00001273, 0.00001288, -0.00000766, 0.00000406, -0.00001097, 0.00001759, 0.00000151, -0.00000420, 0.00000532, -0.00000240, 0.00000358, -0.00000595, 0.00000027, 0.00000133, -0.00000096, 0.00000557, 0.00000072, -0.00000358, -0.00000098, -0.00000314, 0.00000331, -0.00001089, -0.00000197, -0.00001129, -0.00000512, -0.00000205, -0.00000184, -0.00002503, -0.00000850, -0.00008759, -0.00005859, 0.00015772, 0.00004334, 0.00006256, 0.00003058, 0.00011765, 0.00007798, -0.00012337, -0.00002178, -0.00002810, -0.00001026, -0.00001060, -0.00000332, -0.00000762, -0.00000354, -0.00000092, -0.00000271, 0.00000129, -0.00000022, -0.00000580, 0.00000470, -0.00001368, -0.00000183, -0.00000423, -0.00000360, 0.00000044, 0.00000023, -0.00000435, -0.00001025, 0.00000862, -0.00000766, 0.00000159, -0.00000951, 0.00000950, -0.00000309, },
		{ -0.00000196, -0.00000973, -0.00000597, -0.00002739, -0.00001220, 0.00001543, 0.00003576, 0.00000643, 0.00001656, 0.00000969, 0.00001079, 0.00001486, 0.00000343, 0.00003229, 0.00001935, -0.00000653, -0.00003370, -0.00000391, -0.00000849, 0.00000111, 0.00000123, -0.00000156, 0.00000183, -0.00001055, -0.00000438, 0.00000127, 0.00000890, -0.00000709, 0.00000253, -0.00000488, 0.00000297, -0.00000168, 0.00000470, 0.00000213, -0.00000180, -0.00000143, 0.00000533, -0.00000190, -0.00000429, 0.00000597, 0.00001590, -0.00002005, 0.00000142, -0.00000447, 0.00000815, -0.00001277, -0.00001663, 0.00000724, -0.00000871, 0.00000273, -0.00000024, 0.00000141, 0.00000169, -0.00000363, -0.00000332, -0.00000189, -0.00000057, -0.00000214, -0.00000614, -0.00000250, 0.00000516, -0.00000918, -0.00002463, 0.00001220, -0.00000443, 0.00000466, -0.00001076, 0.00000782, 0.00001222, -0.00000788, 0.00000520, -0.00000027, 0.00000146, -0.00000265, -0.00000382, 0.00000347, -0.00000251, 0.00000552, 0.00000224, -0.00000323, -0.00000148, -0.00000273, 0.00000196, -0.00000568, -0.00000752, -0.00000456, -0.00001243, 0.00000118, -0.00000392, -0.00001313, -0.00002144, -0.00003574, -0.00012650, 0.00011135, 0.00010224, 0.00004064, 0.00005487, 0.00007267, 0.00014338, -0.00010995, -0.00005031, -0.00001969, -0.00001966, -0.00000750, -0.00000723, -0.00000478, -0.00000682, 0.00000013, -0.00000315, -0.00000047, 0.00000246, -0.00000655, 0.00000319, -0.00000732, -0.00000962, 0.00000204, -0.00000889, 0.00000247, -0.00000253, 0.00000319, -0.00001577, 0.00000682, -0.00000329, -0.00000105, -0.00000660, 0.00000227, 0.00000484, -0.00000698, },
		{ -0.00001065, -0.00000274, -0.00002428, -0.00001622, 0.00000535, 0.00003483, 0.00001312, 0.00000962, 0.00001643, 0.00000414, 0.00001879, 0.00000172, 0.00002114, 0.00002782, 0.00000426, -0.00002600, -0.00001880, -0.00000025, -0.00000824, 0.00000795, -0.00000643, 0.00000678, -0.00001126, -0.00000420, -0.00000359, 0.00001160, -0.00000587, 0.00000123, -0.00000384, 0.00000191, -0.00000172, 0.00000316, 0.00000355, -0.00000031, -0.00000325, 0.00000510, -0.00000028, -0.00000329, -0.00000138, 0.00001948, -0.00001091, -0.00000921, 0.00000353, -0.00000204, 0.00000337, -0.00002692, 0.00000719, -0.00000627, -0.00000093, 0.00000159, 0.00000021, 0.00000223, -0.00000125, -0.00000507, -0.00000105, -0.00000270, 0.00000064, -0.00000675, -0.00000285, 0.00000157, 0.00000129, -0.00002823, 0.00000413, 0.00000049, 0.00000395, -0.00000619, -0.00000353, 0.00001772, -0.00000462, 0.00000037, 0.00000395, -0.00000075, -0.00000020, -0.00000591, 0.00000309, -0.00000184, 0.00000346, 0.00000475, -0.00000302, -0.00000199, -0.00000197, -0.00000009, -0.00000086, -0.00001083, -0.00000147, -0.00001421, -0.00000072, -0.00000414, -0.00000415, -0.00002874, -0.00000916, -0.00012134, 0.00000102, 0.00016459, 0.00002990, 0.00007568, 0.00003805, 0.00017362, -0.00005634, -0.00009273, -0.00000895, -0.00003016, -0.00000432, -0.00001171, -0.00000184, -0.00000973, 0.00000056, -0.00000331, -0.00000150, 0.00000227, -0.00000323, -0.00000215, 0.00000071, -0.00001554, 0.00000393, -0.00000857, -0.00000021, -0.00000218, 0.00000500, -0.00001141, -0.00000373, 0.00000542, -0.00000605, -0.00000123, -0.00000563, 0.00001016, -0.00000750, -0.00000031, },
		{ -0.00000275, -0.00001852, -0.00002016, -0.00000115, 0.00002753, 0.00002276, 0.00000312, 0.00001988, 0.00000364, 0.00001489, 0.00000859, 0.00000719, 0.00003145, 0.00001222, -0.00001218, -0.00002958, -0.00000021, -0.00000922, 0.00000569, -0.00000314, 0.00000362, -0.00000469, -0.00000896, -0.00000313, 0.00000751, 0.00000106, -0.00000405, 0.00000035, -0.00000197, 0.00000169, -0.00000191, 0.00000737, -0.00000153, -0.00000094, 0.00000056, 0.00000461, -0.00000535, -0.00000169, 0.00001154, 0.00000660, -0.00002193, 0.00001076, -0.00000906, 0.00001056, -0.00002372, -0.00000368, 0.00000199, -0.00000688, 0.00000378, -0.00000080, 0.00000208, 0.00000095, -0.00000535, -0.00000148, -0.00000338, 0.00000080, -0.00000482, -0.00000417, -0.00000111, 0.00000626, -0.00002290, -0.00000793, 0.00000455, 0.00000181, 0.00000174, -0.00001331, 0.00001720, 0.00000173, -0.00000282, 0.00000379, 0.00000219, -0.00000249, -0.00000309, -0.00000180, 0.00000259, -0.00000192, 0.00000944, -0.00000461, -0.00000005, -0.00000416, 0.00000147, -0.00000156, -0.00000674, -0.00000670, -0.00000681, -0.00001067, 0.00000256, -0.00000700, -0.00001745, -0.00001845, -0.00006407, -0.00011032, 0.00016942, 0.00006304, 0.00007107, 0.00003519, 0.00015687, 0.00003115, -0.00013510, -0.00000485, -0.00003729, -0.00000328, -0.00001563, 0.00000056, -0.00001214, 0.00000079, -0.00000407, -0.00000082, -0.00000067, 0.00000199, -0.00000720, 0.00000536, -0.00001527, -0.00000065, -0.00000314, -0.00000563, -0.00000009, 0.00000233, -0.00000182, -0.00001440, 0.00000942, -0.00000611, 0.00000016, -0.00000818, 0.00000880, -0.00000325, -0.00000148, -0.00000921, },
		{ -0.00001052, -0.00002455, -0.00000363, 0.00001595, 0.00003157, 0.00000044, 0.00001829, 0.00000804, 0.00000821, 0.00001482, 0.00000099, 0.00002426, 0.00002229, -0.00000269, -0.00002605, -0.00001288, -0.00000104, -0.00000414, 0.00000559, -0.00000411, 0.00000478, -0.00001451, -0.00000083, 0.00000069, 0.00000887, -0.00000881, 0.00000404, -0.00000580, 0.00000575, -0.00000710, 0.00001014, -0.00000246, 0.00000209, -0.00000411, 0.00000846, -0.00000668, 0.00000002, 0.00000174, 0.00001871, -0.00002304, 0.00000548, -0.00000318, 0.00000313, -0.00000796, -0.00001949, 0.00000942, -0.00001010, 0.00000339, -0.00000038, 0.00000136, 0.00000203, -0.00000350, -0.00000352, -0.00000216, -0.00000157, -0.00000125, -0.00000595, -0.00000211, 0.00000572, -0.00001225, -0.00001899, 0.00000595, -0.00000183, 0.00001086, -0.00001865, 0.00001164, 0.00000702, -0.00000105, -0.00000137, 0.00000832, -0.00000613, 0.00000062, -0.00000683, 0.00000638, -0.00000621, 0.00001209, -0.00000446, 0.00000132, -0.00000637, 0.00000361, -0.00000411, -0.00000045, -0.00001334, 0.00000113, -0.00001904, 0.00000733, -0.00001214, -0.00000176, -0.00003550, -0.00001291, -0.00015291, 0.00008474, 0.00013483, 0.00004972, 0.00005758, 0.00010963, 0.00012357, -0.00015145, -0.00002099, -0.00003419, -0.00000878, -0.00001567, -0.00000040, -0.00001183, -0.00000091, -0.00000393, -0.00000032, -0.00000336, 0.00000503, -0.00000800, 0.00000418, -0.00000924, -0.00000825, 0.00000213, -0.00000852, -0.00000036, 0.00000002, 0.00000500, -0.00001630, 0.00000309, 0.00000097, -0.00000352, -0.00000507, 0.00000228, 0.00000326, -0.00000381, -0.00000587, -0.00000579, },
		{ -0.00002719, -0.00000497, 0.00000500, 0.00003482, 0.00000389, 0.00001238, 0.00001356, 0.00000458, 0.00001422, 0.00000548, 0.00000923, 0.00003158, 0.00000247, -0.00001305, -0.00002724, 0.00000414, -0.00000995, 0.00000901, -0.00000650, 0.00000787, -0.00001307, -0.00000333, -0.00000198, 0.00001079, -0.00000734, 0.00000253, -0.00000523, 0.00000558, -0.00000691, 0.00000678, 0.00000149, 0.00000083, -0.00000337, 0.00000599, -0.00000224, -0.00000281, -0.00000054, 0.00001657, -0.00000847, -0.00001135, 0.00001072, -0.00000963, 0.00000725, -0.00002743, 0.00000733, -0.00000610, -0.00000127, 0.00000186, 0.00000014, 0.00000222, -0.00000072, -0.00000543, -0.00000092, -0.00000396, 0.00000113, -0.00000615, -0.00000281, 0.00000264, -0.00000183, -0.00002452, 0.00000350, -0.00000528, 0.00001644, -0.00001573, 0.00000108, 0.00001097, 0.00000238, -0.00000492, 0.00000961, -0.00000340, -0.00000161, -0.00000569, 0.00000369, -0.00000381, 0.00000750, 0.00000195, -0.00000238, -0.00000345, 0.00000035, -0.00000128, -0.00000075, -0.00001133, -0.00000131, -0.00001489, -0.00000049, -0.00000629, -0.00000104, -0.00003083, -0.00000692, -0.00011118, -0.00005200, 0.00018892, 0.00004998, 0.00007604, 0.00006715, 0.00018291, -0.00011758, -0.00006529, -0.00001798, -0.00002330, -0.00000915, -0.00000755, -0.00000626, -0.00000686, -0.00000059, -0.00000269, -0.00000257, 0.00000307, -0.00000347, -0.00000142, -0.00000139, -0.00001373, 0.00000286, -0.00000649, -0.00000382, -0.00000021, 0.00000608, -0.00000911, -0.00000902, 0.00000843, -0.00000620, -0.00000198, -0.00000337, 0.00000696, -0.00000423, -0.00000263, -0.00000959, -0.00000298, },
		{ -0.00000914, -0.00000045, 0.00002940, 0.00001360, 0.00000417, 0.00001755, 0.00000509, 0.00000911, 0.00001304, -0.00000184, 0.00003066, 0.00001125, -0.00000376, -0.00002831, -0.00000479, -0.00000394, 0.00000152, 0.00000059, 0.00000161, -0.00000328, -0.00001131, 0.00000066, 0.00000589, 0.00000028, -0.00000374, -0.00000025, 0.00000044, -0.00000054, -0.00000251, 0.00000949, -0.00000422, 0.00000219, -0.00000147, 0.00000615, -0.00000904, 0.00000343, 0.00000548, 0.00001047, -0.00002507, 0.00001710, -0.00001219, 0.00000846, -0.00001972, -0.00000582, 0.00000322, -0.00000757, 0.00000400, -0.00000078, 0.00000199, 0.00000128, -0.00000528, -0.00000154, -0.00000414, 0.00000039, -0.00000375, -0.00000427, -0.00000006, 0.00000425, -0.00002264, -0.00000286, -0.00000600, 0.00001390, -0.00000292, -0.00001308, 0.00001436, 0.00000286, -0.00000123, 0.00000191, 0.00000737, -0.00000974, 0.00000093, -0.00000443, 0.00000394, -0.00000242, 0.00001223, -0.00000915, 0.00000337, -0.00000751, 0.00000654, -0.00000754, -0.00000130, -0.00001246, -0.00000134, -0.00001663, 0.00000751, -0.00001367, -0.00000696, -0.00003274, -0.00003644, -0.00015518, 0.00015970, 0.00009944, 0.00007186, 0.00005349, 0.00018960, -0.00003250, -0.00012606, 0.00000130, -0.00004075, 0.00000015, -0.00001830, 0.00000243, -0.00001528, 0.00000467, -0.00000721, 0.00000113, -0.00000240, 0.00000332, -0.00000720, 0.00000396, -0.00001444, -0.00000072, -0.00000239, -0.00000675, -0.00000111, 0.00000465, -0.00000009, -0.00001650, 0.00000730, -0.00000199, -0.00000350, -0.00000407, 0.00000540, -0.00000167, -0.00000094, -0.00001137, 0.00000056, -0.00002482, },
		{ 0.00000027, 0.00001706, 0.00002574, -0.00000271, 0.00001863, 0.00000739, 0.00000573, 0.00001435, -0.00000014, 0.00001597, 0.00002580, -0.00000309, -0.00001466, -0.00002294, 0.00000648, -0.00000886, 0.00000907, -0.00000586, 0.00000633, -0.00001697, 0.00000215, 0.00000061, 0.00000729, -0.00000853, 0.00000377, -0.00000470, 0.00000631, -0.00001074, 0.00001435, -0.00000637, 0.00000566, -0.00000624, 0.00001053, -0.00001086, 0.00000454, -0.00000154, 0.00001823, -0.00002197, 0.00000644, 0.00000053, -0.00000394, -0.00000276, -0.00002041, 0.00000919, -0.00000959, 0.00000284, -0.00000009, 0.00000130, 0.00000201, -0.00000306, -0.00000366, -0.00000255, -0.00000242, -0.00000043, -0.00000552, -0.00000174, 0.00000543, -0.00001518, -0.00001109, -0.00000341, 0.00000342, 0.00001484, -0.00002444, 0.00001427, 0.00000136, 0.00000665, -0.00000832, 0.00001719, -0.00001424, 0.00000405, -0.00000926, 0.00000832, -0.00000858, 0.00001712, -0.00001040, 0.00000586, -0.00001135, 0.00001012, -0.00001061, 0.00000487, -0.00001861, 0.00000595, -0.00002417, 0.00001208, -0.00001969, 0.00000882, -0.00004752, 0.00000373, -0.00016264, 0.00003777, 0.00017229, 0.00006325, 0.00006103, 0.00015772, 0.00007396, -0.00017075, 0.00000171, -0.00004692, 0.00000160, -0.00002377, 0.00000628, -0.00001916, 0.00000583, -0.00000876, 0.00000290, -0.00000609, 0.00000688, -0.00000860, 0.00000430, -0.00001085, -0.00000546, -0.00000032, -0.00000591, -0.00000458, 0.00000390, 0.00000467, -0.00001394, -0.00000323, 0.00000660, -0.00000689, -0.00000227, 0.00000157, 0.00000135, -0.00000020, -0.00001012, -0.00000117, -0.00001686, -0.00001696, },
		{ 0.00000442, 0.00003338, -0.00000348, 0.00001532, 0.00001007, 0.00000642, 0.00000850, 0.00001043, -0.00000254, 0.00003595, -0.00000193, -0.00000043, -0.00003266, 0.00000559, -0.00000896, 0.00000764, -0.00000448, 0.00000656, -0.00001289, -0.00000345, 0.00000087, 0.00000774, -0.00000662, 0.00000216, -0.00000484, 0.00000751, -0.00001103, 0.00000995, -0.00000054, 0.00000197, -0.00000306, 0.00000607, -0.00000400, -0.00000218, 0.00000161, 0.00001092, -0.00000413, -0.00001341, 0.00001660, -0.00001517, 0.00000766, -0.00002354, 0.00000425, -0.00000418, -0.00000232, 0.00000212, 0.00000019, 0.00000210, -0.00000049, -0.00000507, -0.00000151, -0.00000445, 0.00000097, -0.00000483, -0.00000321, 0.00000377, -0.00000630, -0.00001758, -0.00000015, -0.00000886, 0.00002713, -0.00002399, 0.00000573, 0.00000305, 0.00001005, -0.00001005, 0.00001505, -0.00000606, -0.00000339, -0.00000453, 0.00000322, -0.00000467, 0.00001068, -0.00000100, -0.00000109, -0.00000527, 0.00000287, -0.00000261, -0.00000115, -0.00001071, -0.00000243, -0.00001395, -0.00000152, -0.00000754, 0.00000022, -0.00002923, -0.00001167, -0.00009009, -0.00010793, 0.00019712, 0.00008855, 0.00006561, 0.00011925, 0.00016024, -0.00016589, -0.00003383, -0.00003220, -0.00001254, -0.00001688, -0.00000132, -0.00001272, -0.00000196, -0.00000319, -0.00000127, -0.00000392, 0.00000391, -0.00000405, 0.00000002, -0.00000504, -0.00000908, -0.00000099, -0.00000292, -0.00000790, 0.00000244, 0.00000581, -0.00000564, -0.00001421, 0.00001008, -0.00000510, -0.00000348, 0.00000033, 0.00000161, 0.00000091, -0.00000677, -0.00000611, -0.00000680, -0.00002465, 0.00000258, },
		{ 0.00003112, 0.00000435, 0.00000727, 0.00001345, 0.00000820, 0.00000278, 0.00001744, -0.00000959, 0.00002956, 0.00001009, 0.00000127, -0.00002292, -0.00001162, 0.00000265, -0.00000347, 0.00000481, -0.00000107, -0.00000187, -0.00001289, 0.00000437, 0.00000353, -0.00000021, -0.00000314, -0.00000097, 0.00000318, -0.00000390, -0.00000143, 0.00000980, -0.00000540, 0.00000431, -0.00000257, 0.00000607, -0.00001107, 0.00000822, -0.00000052, 0.00001240, -0.00002432, 0.00001906, -0.00001113, 0.00000216, -0.00001233, -0.00000908, 0.00000412, -0.00000739, 0.00000333, -0.00000035, 0.00000191, 0.00000103, -0.00000436, -0.00000230, -0.00000426, -0.00000045, -0.00000225, -0.00000459, 0.00000129, 0.00000044, -0.00001952, 0.00000036, -0.00001558, 0.00002553, -0.00000823, -0.00001083, 0.00000925, 0.00000514, -0.00000014, 0.00000088, 0.00001131, -0.00001610, 0.00000448, -0.00000644, 0.00000446, -0.00000159, 0.00001271, -0.00001137, 0.00000508, -0.00000935, 0.00001019, -0.00001262, 0.00000341, -0.00001716, 0.00000286, -0.00002059, 0.00001013, -0.00001868, 0.00000233, -0.00004590, -0.00001267, -0.00018319, 0.00012104, 0.00015525, 0.00006315, 0.00009228, 0.00020343, -0.00010300, -0.00009842, -0.00000526, -0.00003610, -0.00000205, -0.00001681, 0.00000059, -0.00001524, 0.00000631, -0.00000886, 0.00000199, -0.00000302, 0.00000313, -0.00000543, 0.00000036, -0.00001109, -0.00000207, -0.00000202, -0.00000691, -0.00000246, 0.00000700, 0.00000071, -0.00001634, 0.00000256, 0.00000377, -0.00000823, 0.00000192, -0.00000026, 0.00000171, -0.00000236, -0.00001088, 0.00000041, -0.00002711, 0.00000078, -0.00000184, },
		{ 0.00001763, -0.00000226, 0.00001673, 0.00000809, 0.00000331, 0.00001340, -0.00000035, 0.00000814, 0.00002906, -0.00000577, -0.00000275, -0.00003085, 0.00001143, -0.00001110, 0.00001019, -0.00000556, 0.00000526, -0.00001659, 0.00000340, 0.00000167, 0.00000385, -0.00000597, 0.00000150, -0.00000125, 0.00000375, -0.00001101, 0.00001549, -0.00000783, 0.00000724, -0.00000625, 0.00000989, -0.00001270, 0.00000785, -0.00000318, 0.00001426, -0.00001666, 0.00000411, 0.00000606, -0.00001180, 0.00000176, -0.00001880, 0.00000618, -0.00000689, 0.00000089, 0.00000071, 0.00000121, 0.00000164, -0.00000234, -0.00000373, -0.00000314, -0.00000280, -0.00000003, -0.00000466, -0.00000137, 0.00000388, -0.00001675, -0.00000296, -0.00001394, 0.00001071, 0.00001533, -0.00002593, 0.00001338, -0.00000215, 0.00001232, -0.00001260, 0.00002313, -0.00001999, 0.00000554, -0.00000950, 0.00000778, -0.00000791, 0.00001851, -0.00001320, 0.00000800, -0.00001394, 0.00001417, -0.00001488, 0.00000767, -0.00002067, 0.00000709, -0.00002481, 0.00001210, -0.00002263, 0.00001390, -0.00005209, 0.00000822, -0.00015106, -0.00002945, 0.00021128, 0.00008197, 0.00007050, 0.00020928, -0.00000424, -0.00016094, 0.00000926, -0.00005118, 0.00000624, -0.00002750, 0.00000905, -0.00002331, 0.00001039, -0.00001199, 0.00000470, -0.00000724, 0.00000688, -0.00000747, 0.00000276, -0.00001125, -0.00000224, -0.00000432, -0.00000214, -0.00000901, 0.00000804, 0.00000298, -0.00000965, -0.00001065, 0.00001197, -0.00000995, 0.00000128, -0.00000004, -0.00000010, 0.00000259, -0.00001303, 0.00000202, -0.00002244, -0.00000676, -0.00000037, 0.00001983, },
		{ -0.00000718, 0.00001670, 0.00000725, 0.00000840, 0.00000317, 0.00001445, -0.00001174, 0.00003831, -0.00000547, 0.00000829, -0.00003268, 0.00000287, -0.00000476, 0.00000361, -0.00000014, 0.00000251, -0.00001029, -0.00000472, 0.00000475, 0.00000274, -0.00000376, -0.00000004, -0.00000213, 0.00000654, -0.00001251, 0.00001128, -0.00000153, 0.00000237, -0.00000179, 0.00000492, -0.00000529, -0.00000105, 0.00000399, 0.00000391, 0.00000096, -0.00001426, 0.00001963, -0.00001719, 0.00000402, -0.00001533, -0.00000189, -0.00000059, -0.00000402, 0.00000237, 0.00000032, 0.00000196, -0.00000065, -0.00000389, -0.00000302, -0.00000384, -0.00000006, -0.00000283, -0.00000382, 0.00000442, -0.00001113, -0.00000839, -0.00000694, -0.00000823, 0.00003254, -0.00002750, 0.00000749, -0.00000322, 0.00001574, -0.00001269, 0.00001831, -0.00000754, -0.00000592, -0.00000226, 0.00000132, -0.00000370, 0.00001191, -0.00000293, -0.00000020, -0.00000646, 0.00000461, -0.00000329, -0.00000262, -0.00000844, -0.00000537, -0.00001084, -0.00000448, -0.00000700, -0.00000120, -0.00002370, -0.00002247, -0.00006186, -0.00015944, 0.00018275, 0.00014636, 0.00004830, 0.00019237, 0.00009856, -0.00018819, -0.00000931, -0.00004465, -0.00000270, -0.00002381, 0.00000392, -0.00001861, 0.00000297, -0.00000591, 0.00000003, -0.00000480, 0.00000409, -0.00000425, 0.00000127, -0.00000904, -0.00000292, -0.00000631, 0.00000088, -0.00001129, 0.00000484, 0.00000497, -0.00000201, -0.00001810, 0.00000963, -0.00000277, -0.00000497, 0.00000420, -0.00000443, 0.00000646, -0.00001132, -0.00000149, -0.00001321, -0.00001655, 0.00000429, 0.00000630, 0.00002866, },
		{ 0.00001095, 0.00000871, 0.00001169, -0.00000254, 0.00001990, -0.00001426, 0.00002654, 0.00001043, 0.00000227, -0.00001435, -0.00001903, 0.00000886, -0.00000788, 0.00000852, -0.00000371, -0.00000104, -0.00001296, 0.00000712, 0.00000106, -0.00000061, -0.00000238, -0.00000133, 0.00000535, -0.00000752, 0.00000101, 0.00000812, -0.00000459, 0.00000475, -0.00000201, 0.00000340, -0.00000995, 0.00001081, -0.00000483, 0.00001145, -0.00001905, 0.00001597, -0.00000587, -0.00000743, -0.00000267, -0.00001278, 0.00000440, -0.00000624, 0.00000173, 0.00000047, 0.00000185, 0.00000025, -0.00000263, -0.00000379, -0.00000363, -0.00000164, -0.00000064, -0.00000480, 0.00000254, -0.00000472, -0.00001332, 0.00000005, -0.00002107, 0.00003333, -0.00001191, -0.00000781, 0.00000271, 0.00000822, 0.00000018, 0.00000148, 0.00001251, -0.00001964, 0.00000591, -0.00000673, 0.00000343, 0.00000114, 0.00001023, -0.00001030, 0.00000393, -0.00000837, 0.00001077, -0.00001489, 0.00000546, -0.00001883, 0.00000376, -0.00002045, 0.00000813, -0.00001907, 0.00000638, -0.00005253, 0.00000011, -0.00018719, 0.00005092, 0.00022735, 0.00004855, 0.00015429, 0.00018648, -0.00016667, -0.00005979, -0.00002198, -0.00002415, -0.00000972, -0.00001107, -0.00000526, -0.00001133, 0.00000472, -0.00000805, 0.00000100, -0.00000205, 0.00000125, -0.00000207, -0.00000501, -0.00000566, -0.00000451, -0.00000198, -0.00000638, -0.00000373, 0.00000890, 0.00000089, -0.00001436, -0.00000365, 0.00000945, -0.00001211, 0.00000793, -0.00000669, 0.00000588, -0.00000521, -0.00000773, -0.00000383, -0.00002347, 0.00000625, -0.00000234, 0.00003090, -0.00000330, },
		{ 0.00001259, 0.00001000, 0.00000113, 0.00001189, -0.00000000, 0.00000161, 0.00003083, -0.00000824, 0.00000637, -0.00003378, 0.00001183, -0.00000961, 0.00000826, -0.00000278, 0.00000117, -0.00001287, 0.00000247, 0.00000392, -0.00000110, -0.00000166, -0.00000211, 0.00000383, -0.00000166, -0.00000747, 0.00001275, -0.00000590, 0.00000588, -0.00000327, 0.00000560, -0.00001099, 0.00000888, -0.00000286, 0.00000738, -0.00000793, -0.00000064, 0.00001203, -0.00001875, 0.00000453, -0.00001450, 0.00000064, -0.00000228, -0.00000227, 0.00000192, 0.00000107, 0.00000104, -0.00000144, -0.00000371, -0.00000395, -0.00000253, -0.00000032, -0.00000326, -0.00000105, 0.00000088, -0.00001589, 0.00000321, -0.00002306, 0.00001850, 0.00001211, -0.00002197, 0.00000759, -0.00000166, 0.00001372, -0.00001172, 0.00002341, -0.00002086, 0.00000331, -0.00000637, 0.00000377, -0.00000317, 0.00001489, -0.00001115, 0.00000583, -0.00001208, 0.00001346, -0.00001459, 0.00000577, -0.00001739, 0.00000243, -0.00001878, 0.00000493, -0.00001792, 0.00000986, -0.00004539, -0.00000282, -0.00011755, -0.00011247, 0.00024562, 0.00010722, 0.00009317, 0.00025108, -0.00010156, -0.00012249, -0.00000225, -0.00004284, 0.00000144, -0.00002358, 0.00000489, -0.00002130, 0.00000999, -0.00001135, 0.00000332, -0.00000550, 0.00000408, -0.00000389, -0.00000113, -0.00000946, -0.00000002, -0.00000815, 0.00000107, -0.00001197, 0.00001088, 0.00000137, -0.00000509, -0.00001704, 0.00001499, -0.00001103, 0.00000451, -0.00000220, -0.00000082, 0.00000379, -0.00001337, 0.00000201, -0.00002458, 0.00000283, -0.00000358, 0.00002371, 0.00000777, 0.00000147, },
		{ 0.00000616, 0.00000934, -0.00000038, 0.00001570, -0.00001626, 0.00003632, -0.00000527, 0.00001074, -0.00002637, -0.00000391, 0.00000202, -0.00000230, 0.00000581, -0.00000374, -0.00000563, -0.00000673, 0.00000889, -0.00000329, 0.00000047, -0.00000344, 0.00000250, 0.00000244, -0.00001053, 0.00000988, -0.00000072, 0.00000149, 0.00000078, 0.00000221, -0.00000578, 0.00000069, 0.00000557, -0.00000286, 0.00000547, -0.00001296, 0.00001879, -0.00001514, -0.00000314, -0.00000403, -0.00000993, 0.00000395, -0.00000599, 0.00000242, 0.00000057, 0.00000179, -0.00000111, -0.00000210, -0.00000527, -0.00000218, -0.00000189, -0.00000042, -0.00000426, 0.00000395, -0.00001496, 0.00000132, -0.00001583, -0.00000281, 0.00003058, -0.00002392, 0.00000421, -0.00000554, 0.00001722, -0.00001087, 0.00001764, -0.00000672, -0.00000948, 0.00000108, -0.00000210, -0.00000052, 0.00001041, -0.00000279, -0.00000080, -0.00000594, 0.00000453, -0.00000252, -0.00000563, -0.00000422, -0.00001038, -0.00000531, -0.00000981, -0.00000402, -0.00000573, -0.00001467, -0.00003728, -0.00003132, -0.00019980, 0.00014278, 0.00021858, 0.00003538, 0.00027527, -0.00000004, -0.00017705, -0.00000079, -0.00004842, 0.00000098, -0.00002574, 0.00000461, -0.00002076, 0.00000532, -0.00000685, -0.00000015, -0.00000416, 0.00000271, -0.00000310, 0.00000111, -0.00001172, 0.00000271, -0.00001110, 0.00000326, -0.00001266, 0.00000588, 0.00000452, 0.00000069, -0.00001961, 0.00000656, 0.00000057, -0.00000565, 0.00000690, -0.00000946, 0.00001057, -0.00001438, 0.00000220, -0.00001978, -0.00000490, 0.00000036, 0.00001164, 0.00001990, -0.00000624, 0.00001531, },
		{ 0.00001413, -0.00000542, 0.00001892, -0.00001454, 0.00002050, 0.00001310, -0.00000031, -0.00000440, -0.00002500, 0.00001293, -0.00001037, 0.00001078, -0.00000570, -0.00000118, -0.00001111, 0.00000827, -0.00000102, -0.00000104, -0.00000158, -0.00000087, 0.00000597, -0.00001017, 0.00000384, 0.00000505, -0.00000211, 0.00000364, 0.00000008, -0.00000177, -0.00000527, 0.00001014, -0.00000651, 0.00000751, -0.00000979, 0.00000859, 0.00000222, -0.00001819, 0.00000726, -0.00001577, 0.00000365, -0.00000405, -0.00000071, 0.00000165, 0.00000170, -0.00000077, -0.00000049, -0.00000571, -0.00000248, -0.00000288, 0.00000069, -0.00000446, 0.00000306, -0.00001023, -0.00000474, -0.00000435, -0.00002045, 0.00003467, -0.00001190, -0.00000552, -0.00000390, 0.00001119, 0.00000014, 0.00000368, 0.00001040, -0.00001932, 0.00000434, -0.00000479, 0.00000076, 0.00000552, 0.00000505, -0.00000606, -0.00000016, -0.00000423, 0.00000759, -0.00001336, 0.00000382, -0.00001648, 0.00000046, -0.00001547, 0.00000073, -0.00001351, 0.00000290, -0.00004937, -0.00000191, -0.00016548, -0.00004573, 0.00030481, 0.00003928, 0.00023496, 0.00013147, -0.00020961, -0.00002115, -0.00004255, -0.00000876, -0.00002025, -0.00000299, -0.00001365, -0.00000442, 0.00000019, -0.00000480, -0.00000178, 0.00000026, -0.00000181, 0.00000205, -0.00001098, 0.00000075, -0.00000745, -0.00000223, -0.00000553, -0.00000438, 0.00000977, 0.00000101, -0.00001145, -0.00000979, 0.00001309, -0.00001301, 0.00001174, -0.00001181, 0.00000910, -0.00000801, -0.00000329, -0.00001096, -0.00001414, 0.00000563, 0.00000026, 0.00002717, -0.00000768, 0.00001349, 0.00000471, },
		{ -0.00000036, 0.00000961, 0.00000069, -0.00000303, 0.00003004, -0.00000842, 0.00001035, -0.00003028, 0.00000709, -0.00000444, 0.00000369, 0.00000187, -0.00000520, -0.00000642, -0.00000021, 0.00000668, -0.00000648, 0.00000319, -0.00000565, 0.00000889, -0.00000841, -0.00000112, 0.00000671, -0.00000097, 0.00000189, 0.00000229, -0.00000201, -0.00000566, 0.00000734, -0.00000092, -0.00000105, 0.00000247, -0.00000616, 0.00001659, -0.00002305, 0.00000478, -0.00000788, -0.00000654, 0.00000343, -0.00000606, 0.00000329, 0.00000086, 0.00000040, -0.00000053, -0.00000360, -0.00000488, -0.00000163, -0.00000133, -0.00000130, -0.00000088, -0.00000338, -0.00001210, 0.00000568, -0.00002826, 0.00002459, 0.00000632, -0.00001312, -0.00000277, 0.00000306, 0.00001019, -0.00000491, 0.00001695, -0.00001576, -0.00000318, 0.00000019, -0.00000348, 0.00000534, 0.00000641, -0.00000406, -0.00000110, -0.00000509, 0.00000709, -0.00000879, -0.00000155, -0.00000832, -0.00000828, -0.00000601, -0.00000959, -0.00000497, -0.00000417, -0.00002692, -0.00002838, -0.00006692, -0.00020160, 0.00026568, 0.00014214, 0.00013601, 0.00026648, -0.00020115, -0.00006418, -0.00003019, -0.00002238, -0.00001301, -0.00001153, -0.00000687, -0.00001211, 0.00000341, -0.00000576, -0.00000197, -0.00000046, -0.00000168, 0.00000215, -0.00000741, -0.00000496, -0.00000003, -0.00001010, 0.00000197, -0.00001187, 0.00001094, 0.00000129, -0.00000202, -0.00002037, 0.00001380, -0.00000859, 0.00000611, -0.00000405, -0.00000114, 0.00000341, -0.00001096, -0.00000177, -0.00002156, 0.00000873, -0.00000656, 0.00002693, -0.00000218, 0.00000712, 0.00000706, 0.00001243, },
		{ -0.00000149, 0.00001335, -0.00001519, 0.00002896, -0.00000005, 0.00000650, -0.00001471, -0.00001315, 0.00000962, -0.00000839, 0.00001178, -0.00001074, -0.00000008, -0.00000850, 0.00001208, -0.00000890, 0.00000468, -0.00000669, 0.00000773, -0.00000387, -0.00000544, 0.00000580, 0.00000190, -0.00000063, 0.00000448, -0.00000198, -0.00000520, 0.00000287, 0.00000573, -0.00000810, 0.00000835, -0.00000905, 0.00001396, -0.00000955, -0.00001220, 0.00000809, -0.00001792, 0.00000815, -0.00000754, 0.00000201, 0.00000102, 0.00000151, -0.00000158, -0.00000019, -0.00000768, 0.00000004, -0.00000407, 0.00000192, -0.00000412, 0.00000178, -0.00001643, 0.00000930, -0.00002458, 0.00000617, 0.00002143, -0.00001309, -0.00000445, -0.00000314, 0.00001361, -0.00000388, 0.00001232, -0.00000314, -0.00001396, 0.00000510, -0.00000663, 0.00000463, 0.00000609, -0.00000015, -0.00000348, -0.00000306, 0.00000200, 0.00000009, -0.00001015, 0.00000167, -0.00001703, 0.00000217, -0.00001721, 0.00000130, -0.00001303, -0.00000346, -0.00005314, -0.00000359, -0.00022386, 0.00007815, 0.00029498, 0.00004478, 0.00034726, -0.00012236, -0.00013412, -0.00001184, -0.00003969, -0.00000474, -0.00002001, -0.00000159, -0.00001691, 0.00000308, -0.00000453, -0.00000282, -0.00000128, -0.00000087, 0.00000011, -0.00000146, -0.00001152, 0.00000579, -0.00001341, 0.00000277, -0.00001098, 0.00000474, 0.00000524, 0.00000159, -0.00001806, 0.00000077, 0.00000465, -0.00000501, 0.00000746, -0.00001213, 0.00001179, -0.00001435, 0.00000291, -0.00002357, 0.00000674, -0.00000676, 0.00001931, 0.00000801, -0.00000110, 0.00001106, 0.00000707, 0.00000872, },
		{ 0.00001402, -0.00001021, 0.00001146, 0.00001777, -0.00000500, 0.00000467, -0.00002758, 0.00001343, -0.00000996, 0.00001093, -0.00000663, -0.00000236, -0.00000734, 0.00000758, -0.00000237, -0.00000161, -0.00000079, 0.00000060, 0.00000438, -0.00001072, 0.00000587, 0.00000170, 0.00000106, 0.00000184, 0.00000271, -0.00000820, 0.00000207, 0.00000641, -0.00000563, 0.00000144, 0.00000182, -0.00000112, 0.00001070, -0.00002727, 0.00001502, -0.00001671, 0.00000149, -0.00000091, -0.00000379, 0.00000310, 0.00000130, -0.00000157, 0.00000144, -0.00000745, -0.00000122, -0.00000380, 0.00000131, -0.00000317, 0.00000216, -0.00001469, 0.00000450, -0.00001179, -0.00001358, 0.00002859, -0.00000721, -0.00000509, -0.00000918, 0.00001288, 0.00000061, 0.00000664, 0.00000554, -0.00001529, -0.00000007, -0.00000106, -0.00000284, 0.00001050, -0.00000153, 0.00000008, -0.00000607, 0.00000211, 0.00000131, -0.00000842, -0.00000116, -0.00001063, -0.00000630, -0.00000673, -0.00001077, -0.00000306, -0.00000755, -0.00003671, -0.00001773, -0.00012310, -0.00015608, 0.00036904, 0.00005438, 0.00032073, 0.00003922, -0.00022188, 0.00000652, -0.00005909, 0.00000449, -0.00002948, 0.00000408, -0.00002172, 0.00000325, -0.00000573, -0.00000025, -0.00000540, 0.00000296, -0.00000492, 0.00000556, -0.00001595, 0.00000661, -0.00001001, -0.00000287, -0.00000466, -0.00000406, 0.00000920, 0.00000154, -0.00000848, -0.00001451, 0.00001326, -0.00000951, 0.00001163, -0.00001377, 0.00000959, -0.00000891, 0.00000025, -0.00001828, -0.00000158, -0.00000030, 0.00000645, 0.00001943, -0.00000791, 0.00001372, 0.00000253, 0.00001456, -0.00000513, },
		{ 0.00000168, -0.00000563, 0.00002631, -0.00000526, 0.00000847, -0.00002073, -0.00000174, 0.00000308, -0.00000206, 0.00000684, -0.00001215, 0.00000121, -0.00000343, 0.00000876, -0.00001086, 0.00000701, -0.00000746, 0.00001188, -0.00001420, 0.00000591, -0.00000073, 0.00000531, -0.00000317, 0.00000875, -0.00001120, 0.00000206, 0.00000389, 0.00000167, -0.00000918, 0.00001231, -0.00001046, 0.00001794, -0.00002332, 0.00000224, 0.00000015, -0.00001399, 0.00000899, -0.00000961, 0.00000441, 0.00000061, -0.00000008, 0.00000018, -0.00000337, -0.00000575, -0.00000033, -0.00000287, 0.00000108, -0.00000107, -0.00000828, -0.00000571, 0.00000368, -0.00002773, 0.00002675, 0.00000009, -0.00000158, -0.00001553, 0.00001030, 0.00000318, 0.00000638, 0.00000498, -0.00000573, -0.00001274, 0.00000868, -0.00001224, 0.00001565, -0.00000499, 0.00000631, -0.00001127, 0.00000569, -0.00000375, 0.00000133, -0.00001281, 0.00000466, -0.00002276, 0.00001082, -0.00002859, 0.00001340, -0.00002521, -0.00000062, -0.00006232, -0.00000965, -0.00028224, 0.00025854, 0.00019274, 0.00020306, 0.00023926, -0.00028191, -0.00000159, -0.00006552, 0.00000451, -0.00003316, 0.00000557, -0.00002371, 0.00000242, -0.00000811, 0.00000380, -0.00001023, 0.00000685, -0.00000927, 0.00000950, -0.00001506, 0.00000172, -0.00000261, -0.00000925, -0.00000033, -0.00000804, 0.00000758, 0.00000350, -0.00000148, -0.00001950, 0.00000757, -0.00000193, 0.00000515, -0.00000467, -0.00000179, 0.00000225, -0.00000677, -0.00000837, -0.00001345, 0.00000917, -0.00000679, 0.00002740, -0.00001005, 0.00001266, 0.00000162, 0.00001433, -0.00000064, 0.00000636, },
		{ -0.00000928, 0.00001704, 0.00000935, -0.00000265, -0.00000043, -0.00002210, 0.00001558, -0.00001246, 0.00001572, -0.00001649, 0.00000473, -0.00000879, 0.00001297, -0.00001251, 0.00000729, -0.00000814, 0.00001163, -0.00001048, 0.00000126, 0.00000014, 0.00000547, -0.00000325, 0.00000851, -0.00000704, -0.00000354, 0.00000516, 0.00000428, -0.00001102, 0.00000907, -0.00000280, 0.00000596, -0.00000191, -0.00002079, 0.00001827, -0.00002359, 0.00001054, -0.00000787, 0.00000084, 0.00000178, 0.00000100, -0.00000172, 0.00000121, -0.00000947, 0.00000208, -0.00000593, 0.00000362, -0.00000308, -0.00000234, -0.00001451, 0.00001337, -0.00003035, 0.00001580, 0.00000765, 0.00000263, -0.00001654, 0.00000271, 0.00000602, 0.00000715, 0.00000318, 0.00000254, -0.00001851, 0.00000876, -0.00001117, 0.00001063, -0.00000014, 0.00000442, -0.00000789, 0.00000190, -0.00000270, 0.00000407, -0.00001536, 0.00000805, -0.00002393, 0.00001009, -0.00002527, 0.00000774, -0.00002161, 0.00000756, -0.00006632, 0.00001621, -0.00022830, -0.00000679, 0.00036212, 0.00009782, 0.00038124, -0.00024524, -0.00007105, -0.00003857, -0.00001990, -0.00001913, -0.00000714, -0.00001429, -0.00000709, -0.00000396, 0.00000120, -0.00000794, 0.00000361, -0.00000636, 0.00000518, -0.00000649, -0.00000787, 0.00000520, -0.00001213, -0.00000116, -0.00000609, 0.00000130, 0.00000729, 0.00000050, -0.00001351, -0.00000725, 0.00000902, -0.00000292, 0.00000562, -0.00001192, 0.00000957, -0.00001062, -0.00000042, -0.00002227, 0.00001459, -0.00001339, 0.00002659, -0.00000451, 0.00000631, 0.00000509, 0.00000927, 0.00000692, -0.00000052, 0.00000790, },
		{ 0.00000066, 0.00002309, -0.00000969, 0.00001069, -0.00002547, 0.00000973, -0.00000634, 0.00000877, -0.00000637, -0.00000433, -0.00000212, 0.00000520, -0.00000288, -0.00000240, 0.00000011, 0.00000279, 0.00000062, -0.00000865, 0.00000615, -0.00000082, 0.00000378, 0.00000046, 0.00000444, -0.00001394, 0.00001010, 0.00000095, -0.00000322, -0.00000506, 0.00001334, -0.00001051, 0.00001678, -0.00003187, 0.00001842, -0.00001456, -0.00000223, 0.00000286, -0.00000710, 0.00000461, 0.00000055, -0.00000170, 0.00000249, -0.00000832, -0.00000042, -0.00000400, 0.00000093, -0.00000069, -0.00000069, -0.00001663, 0.00001207, -0.00001987, -0.00000236, 0.00001631, 0.00000162, -0.00000675, -0.00001223, 0.00001236, 0.00000252, 0.00000904, -0.00000060, -0.00000887, -0.00000609, 0.00000322, -0.00000602, 0.00001440, -0.00000751, 0.00000601, -0.00001171, 0.00000861, -0.00000619, -0.00000175, -0.00000777, -0.00000318, -0.00001428, 0.00000298, -0.00002311, 0.00000890, -0.00002155, -0.00001847, -0.00004159, -0.00007106, -0.00026117, 0.00039646, 0.00011820, 0.00038980, -0.00007962, -0.00020083, 0.00001570, -0.00006500, 0.00001045, -0.00003336, 0.00000672, -0.00002628, 0.00000882, -0.00001074, 0.00000381, -0.00000838, 0.00000469, -0.00000674, 0.00000702, -0.00001826, 0.00001028, -0.00001116, -0.00000414, -0.00000384, -0.00000277, 0.00000724, 0.00000263, -0.00000594, -0.00001708, 0.00000960, -0.00000154, 0.00000718, -0.00001175, 0.00000639, -0.00000661, 0.00000083, -0.00002264, 0.00001038, -0.00000871, 0.00001474, 0.00000876, -0.00000441, 0.00001145, 0.00000242, 0.00001285, -0.00000199, 0.00000596, -0.00000242, },
		{ 0.00002001, 0.00000093, 0.00000175, -0.00000725, -0.00001226, 0.00001070, -0.00000697, 0.00001020, -0.00001755, 0.00000800, -0.00000570, 0.00000891, -0.00001288, 0.00000852, -0.00000628, 0.00001112, -0.00001676, 0.00001112, -0.00000714, 0.00001063, -0.00000710, 0.00001369, -0.00001933, 0.00000999, -0.00000002, 0.00000356, -0.00001505, 0.00001941, -0.00001181, 0.00001490, -0.00001903, -0.00000268, 0.00000823, -0.00002006, 0.00001297, -0.00001194, 0.00000480, 0.00000044, -0.00000028, 0.00000055, -0.00000304, -0.00000628, 0.00000097, -0.00000454, 0.00000358, -0.00000179, -0.00001285, 0.00000207, -0.00000220, -0.00002101, 0.00002343, -0.00000413, 0.00000947, -0.00002720, 0.00001689, -0.00000419, 0.00001870, -0.00000909, 0.00000620, -0.00002247, 0.00001624, -0.00001950, 0.00002447, -0.00001579, 0.00001652, -0.00002134, 0.00001693, -0.00001578, 0.00001247, -0.00002439, 0.00001748, -0.00003641, 0.00002647, -0.00004633, 0.00003114, -0.00004657, 0.00002548, -0.00009401, 0.00003942, -0.00033664, 0.00020938, 0.00026799, 0.00029277, 0.00015822, -0.00032347, 0.00004668, -0.00009511, 0.00002832, -0.00005159, 0.00002151, -0.00004014, 0.00001761, -0.00002081, 0.00001418, -0.00001871, 0.00001389, -0.00001628, 0.00001585, -0.00002188, 0.00000882, -0.00000678, -0.00000604, -0.00000534, -0.00000121, 0.00000149, 0.00000754, -0.00000324, -0.00001479, -0.00000284, 0.00000819, 0.00000169, -0.00000363, -0.00000339, 0.00000133, -0.00000245, -0.00001558, -0.00000212, 0.00000429, -0.00000305, 0.00002370, -0.00001399, 0.00001610, -0.00000220, 0.00001470, 0.00000068, 0.00000209, 0.00000299, -0.00000645, },
		{ 0.00002032, -0.00001333, 0.00001270, -0.00002767, 0.00001752, -0.00001254, 0.00001579, -0.00001910, 0.00000725, -0.00000658, 0.00001062, -0.00001294, 0.00000709, -0.00000647, 0.00001238, -0.00001513, 0.00000741, -0.00000526, 0.00000854, -0.00000509, 0.00001155, -0.00001182, -0.00000118, 0.00000720, 0.00000150, -0.00001148, 0.00000780, 0.00000483, -0.00000360, 0.00000573, -0.00002642, 0.00002398, -0.00002496, 0.00000992, -0.00000639, -0.00000127, 0.00000286, 0.00000017, -0.00000125, 0.00000154, -0.00000986, 0.00000311, -0.00000677, 0.00000423, -0.00000110, -0.00000812, -0.00000899, 0.00001212, -0.00003046, 0.00002240, -0.00000651, 0.00001890, -0.00002824, 0.00000895, -0.00000268, 0.00001929, -0.00000729, 0.00000845, -0.00002153, 0.00001052, -0.00001402, 0.00001557, -0.00000635, 0.00000921, -0.00001250, 0.00000743, -0.00000810, 0.00000790, -0.00001942, 0.00001278, -0.00002868, 0.00001582, -0.00003135, 0.00001270, -0.00002860, 0.00001490, -0.00007241, 0.00002324, -0.00021128, -0.00010610, 0.00040702, 0.00021372, 0.00034954, -0.00034008, -0.00000734, -0.00007000, 0.00000394, -0.00003696, 0.00000852, -0.00002976, 0.00000574, -0.00001357, 0.00000850, -0.00001382, 0.00000891, -0.00001220, 0.00001066, -0.00001268, -0.00000165, 0.00000131, -0.00000761, -0.00000774, 0.00000093, -0.00000351, 0.00000996, -0.00000188, -0.00000699, -0.00001622, 0.00001299, 0.00000045, 0.00000185, -0.00000925, 0.00000452, -0.00000391, -0.00000735, -0.00001515, 0.00001599, -0.00001599, 0.00003024, -0.00001449, 0.00001320, -0.00000061, 0.00001137, 0.00000499, 0.00000132, 0.00000100, -0.00000072, 0.00000301, },
		{ -0.00001227, 0.00001217, -0.00001843, 0.00000222, 0.00000000, 0.00000461, -0.00000510, -0.00000663, 0.00000379, 0.00000163, -0.00000265, -0.00000341, 0.00000140, 0.00000498, -0.00000446, -0.00000433, 0.00000448, -0.00000187, 0.00000524, 0.00000036, 0.00000407, -0.00001702, 0.00001648, -0.00000435, -0.00000083, -0.00001007, 0.00002225, -0.00001700, 0.00001819, -0.00003009, 0.00001622, -0.00000899, -0.00000717, 0.00000667, -0.00001002, 0.00000581, -0.00000047, -0.00000097, 0.00000222, -0.00000781, -0.00000046, -0.00000329, -0.00000043, 0.00000274, -0.00000545, -0.00001507, 0.00001572, -0.00002547, 0.00000971, 0.00000093, 0.00001243, -0.00000962, -0.00001301, 0.00000939, 0.00000633, 0.00000970, -0.00000620, -0.00000199, -0.00001192, 0.00000648, -0.00000733, 0.00001560, -0.00001090, 0.00000955, -0.00001481, 0.00001287, -0.00001245, 0.00000426, -0.00001359, 0.00000324, -0.00002051, 0.00001016, -0.00003219, 0.00001778, -0.00003392, -0.00000090, -0.00006481, -0.00002353, -0.00034016, 0.00036382, 0.00025505, 0.00041601, -0.00020573, -0.00015253, 0.00000488, -0.00005735, 0.00000639, -0.00002957, 0.00000283, -0.00002510, 0.00001000, -0.00001278, 0.00000564, -0.00000933, 0.00000429, -0.00000622, 0.00000543, -0.00001676, 0.00001056, -0.00001011, -0.00000628, -0.00000294, -0.00000088, 0.00000433, 0.00000401, -0.00000387, -0.00001764, 0.00000308, 0.00000939, -0.00000044, -0.00000641, -0.00000010, -0.00000107, -0.00000244, -0.00002164, 0.00001786, -0.00001582, 0.00002231, -0.00000250, 0.00000123, 0.00000748, 0.00000382, 0.00000969, 0.00000272, -0.00000338, 0.00000663, -0.00000978, 0.00002704, },
		{ -0.00000736, 0.00000689, -0.00002142, 0.00001597, -0.00000917, 0.00001030, -0.00001960, 0.00001218, -0.00000582, 0.00000636, -0.00001186, 0.00000723, -0.00000186, 0.00000605, -0.00001479, 0.00001262, -0.00001041, 0.00001294, -0.00000789, 0.00001483, -0.00002375, 0.00001565, -0.00000276, 0.00000355, -0.00001723, 0.00002229, -0.00000938, 0.00000741, -0.00001075, -0.00000884, 0.00001477, -0.00002315, 0.00001410, -0.00001214, 0.00000396, 0.00000052, -0.00000022, 0.00000058, -0.00000272, -0.00000618, 0.00000175, -0.00000574, 0.00000572, -0.00000309, -0.00001603, 0.00000943, -0.00001019, -0.00000923, 0.00001438, -0.00000446, 0.00001682, -0.00003410, 0.00001929, -0.00000816, 0.00002767, -0.00002069, 0.00001581, -0.00002862, 0.00001942, -0.00002187, 0.00002809, -0.00002190, 0.00002233, -0.00002699, 0.00002415, -0.00002441, 0.00001997, -0.00003140, 0.00002482, -0.00004337, 0.00003441, -0.00005560, 0.00004036, -0.00005940, 0.00004103, -0.00011069, 0.00006407, -0.00034726, 0.00010439, 0.00037862, 0.00039682, 0.00002111, -0.00031160, 0.00006403, -0.00010558, 0.00003854, -0.00005964, 0.00002874, -0.00004924, 0.00002726, -0.00002932, 0.00002085, -0.00002351, 0.00001720, -0.00001954, 0.00001822, -0.00002494, 0.00001361, -0.00001036, -0.00000226, -0.00001112, 0.00000644, -0.00000525, 0.00001165, -0.00000571, -0.00000821, -0.00001488, 0.00001938, -0.00000295, -0.00000144, -0.00000592, 0.00000130, 0.00000041, -0.00002080, 0.00000934, -0.00000396, 0.00000417, 0.00001568, -0.00001312, 0.00001613, -0.00000331, 0.00001295, 0.00000359, -0.00000294, 0.00000457, -0.00000603, 0.00001224, 0.00000874, },
		{ 0.00002108, -0.00002744, 0.00001399, -0.00000774, 0.00001114, -0.00001749, 0.00000658, -0.00000156, 0.00000490, -0.00000993, 0.00000381, -0.00000130, 0.00000895, -0.00001601, 0.00001093, -0.00000854, 0.00000957, -0.00000482, 0.00001222, -0.00001496, 0.00000121, 0.00000870, -0.00000201, -0.00000993, 0.00000523, 0.00001236, -0.00001272, 0.00001128, -0.00002716, 0.00002356, -0.00002096, 0.00000575, -0.00000297, -0.00000420, 0.00000412, -0.00000089, -0.00000009, 0.00000053, -0.00000836, 0.00000255, -0.00000611, 0.00000352, 0.00000150, -0.00001463, -0.00000070, 0.00000546, -0.00002345, 0.00002275, -0.00001645, 0.00003073, -0.00003504, 0.00001170, -0.00000866, 0.00002859, -0.00001559, 0.00001194, -0.00002101, 0.00000857, -0.00001325, 0.00001712, -0.00000999, 0.00001171, -0.00001484, 0.00001102, -0.00001172, 0.00000919, -0.00001983, 0.00001309, -0.00002824, 0.00001603, -0.00003189, 0.00001250, -0.00003007, 0.00001426, -0.00006677, 0.00001314, -0.00017200, -0.00021421, 0.00042132, 0.00040345, 0.00023142, -0.00037965, 0.00003462, -0.00009110, 0.00002105, -0.00004972, 0.00001973, -0.00004159, 0.00001616, -0.00002135, 0.00001375, -0.00001740, 0.00001181, -0.00001571, 0.00001394, -0.00001743, 0.00000473, -0.00000386, -0.00000180, -0.00001477, 0.00000783, -0.00000776, 0.00001166, -0.00000395, -0.00000044, -0.00002416, 0.00001550, 0.00000483, -0.00000299, -0.00000518, -0.00000196, 0.00000397, -0.00001594, -0.00000350, 0.00001034, -0.00001250, 0.00002788, -0.00001913, 0.00001688, -0.00000398, 0.00001182, 0.00000419, 0.00000250, -0.00000516, 0.00000758, -0.00000979, 0.00002939, -0.00002164, },
		{ -0.00000744, -0.00000764, 0.00000775, -0.00000069, -0.00000327, -0.00000860, 0.00000942, -0.00000240, -0.00000197, -0.00000456, 0.00000329, 0.00000619, -0.00000943, 0.00000106, 0.00000145, -0.00000154, 0.00000534, 0.00000169, 0.00000119, -0.00001624, 0.00001935, -0.00000771, 0.00000013, -0.00001200, 0.00002665, -0.00001890, 0.00001390, -0.00002163, 0.00000864, -0.00000071, -0.00001241, 0.00000962, -0.00001179, 0.00000615, -0.00000141, 0.00000047, 0.00000066, -0.00000585, -0.00000140, -0.00000183, -0.00000233, 0.00000643, -0.00001139, -0.00000996, 0.00001409, -0.00002583, 0.00001871, -0.00001346, 0.00002209, -0.00001196, -0.00001234, 0.00000464, 0.00001167, 0.00000808, -0.00000978, 0.00000351, -0.00001586, 0.00000742, -0.00000583, 0.00001316, -0.00001050, 0.00000925, -0.00001380, 0.00001306, -0.00001534, 0.00000743, -0.00001639, 0.00000627, -0.00002244, 0.00001186, -0.00003453, 0.00001950, -0.00003966, 0.00000966, -0.00007868, 0.00000625, -0.00037582, 0.00025483, 0.00048217, 0.00037509, -0.00031511, -0.00009047, -0.00002097, -0.00003806, -0.00000671, -0.00001870, -0.00000729, -0.00001792, 0.00000601, -0.00001088, 0.00000434, -0.00000757, 0.00000136, -0.00000317, 0.00000076, -0.00001137, 0.00000717, -0.00000671, -0.00000925, -0.00000182, 0.00000106, 0.00000119, 0.00000514, -0.00000192, -0.00001694, -0.00000438, 0.00002055, -0.00000871, 0.00000018, -0.00000802, 0.00000619, -0.00000877, -0.00001479, 0.00001844, -0.00001834, 0.00002620, -0.00001154, 0.00000680, 0.00000323, 0.00000562, 0.00000629, 0.00000724, -0.00001166, 0.00001429, -0.00001745, 0.00002771, -0.00001120, 0.00000874, },
		{ -0.00002644, 0.00001705, -0.00000757, 0.00000645, -0.00001739, 0.00001278, -0.00000335, 0.00000118, -0.00000804, 0.00000369, 0.00000476, -0.00000236, -0.00000850, 0.00000989, -0.00000963, 0.00001132, -0.00000467, 0.00001104, -0.00002274, 0.00001726, -0.00000317, 0.00000106, -0.00001525, 0.00002066, -0.00000354, -0.00000328, -0.00000018, -0.00001451, 0.00001819, -0.00002204, 0.00001150, -0.00000962, 0.00000156, 0.00000114, -0.00000013, 0.00000046, -0.00000261, -0.00000512, 0.00000147, -0.00000578, 0.00000682, -0.00000469, -0.00001705, 0.00001453, -0.00001775, 0.00000508, 0.00000092, -0.00000028, 0.00001835, -0.00003358, 0.00001485, -0.00000570, 0.00002948, -0.00002554, 0.00001908, -0.00002781, 0.00001539, -0.00001667, 0.00002357, -0.00001997, 0.00002011, -0.00002437, 0.00002316, -0.00002518, 0.00001934, -0.00002924, 0.00002183, -0.00003842, 0.00002883, -0.00004988, 0.00003371, -0.00005530, 0.00003634, -0.00010096, 0.00005084, -0.00030117, -0.00006526, 0.00053424, 0.00050083, -0.00016336, -0.00024230, 0.00004059, -0.00008752, 0.00002710, -0.00005031, 0.00002090, -0.00004487, 0.00002550, -0.00002841, 0.00001928, -0.00002080, 0.00001349, -0.00001609, 0.00001388, -0.00002148, 0.00001321, -0.00001063, -0.00000055, -0.00001478, 0.00001182, -0.00000975, 0.00001331, -0.00000639, -0.00000280, -0.00002477, 0.00002810, -0.00000634, 0.00000041, -0.00000842, 0.00000192, 0.00000113, -0.00002207, 0.00001779, -0.00001264, 0.00001300, 0.00000457, -0.00000782, 0.00001245, -0.00000142, 0.00000911, 0.00000761, -0.00000812, 0.00000621, -0.00000503, 0.00000442, 0.00001609, -0.00001584, 0.00001821, },
		{ 0.00000516, 0.00000135, 0.00000230, -0.00001180, 0.00000291, 0.00000557, -0.00000323, -0.00000430, -0.00000166, 0.00000643, 0.00000169, -0.00001253, 0.00001066, -0.00000855, 0.00000766, -0.00000173, 0.00000963, -0.00001537, 0.00000293, 0.00000944, -0.00000549, -0.00000722, 0.00000246, 0.00001816, -0.00001945, 0.00001314, -0.00002215, 0.00001676, -0.00001179, -0.00000155, 0.00000199, -0.00000756, 0.00000533, -0.00000205, 0.00000166, -0.00000176, -0.00000491, 0.00000017, -0.00000377, 0.00000155, 0.00000414, -0.00002051, 0.00000858, -0.00000510, -0.00000975, 0.00001527, -0.00001880, 0.00003409, -0.00003330, 0.00000777, -0.00000852, 0.00003128, -0.00001826, 0.00001044, -0.00001522, 0.00000156, -0.00000733, 0.00001328, -0.00000862, 0.00000935, -0.00001226, 0.00000986, -0.00001077, 0.00000533, -0.00001401, 0.00000630, -0.00001974, 0.00000750, -0.00002337, 0.00000327, -0.00002186, 0.00000145, -0.00004534, -0.00001736, -0.00011046, -0.00032755, 0.00040449, 0.00066449, 0.00002001, -0.00034571, 0.00003591, -0.00008751, 0.00002035, -0.00004827, 0.00001855, -0.00004259, 0.00001783, -0.00002200, 0.00001262, -0.00001502, 0.00000906, -0.00001382, 0.00001198, -0.00001754, 0.00000797, -0.00000718, 0.00000227, -0.00001912, 0.00001166, -0.00000902, 0.00001038, -0.00000366, 0.00000369, -0.00002864, 0.00001525, 0.00001009, -0.00000805, -0.00000084, -0.00000822, 0.00001084, -0.00002343, 0.00000978, -0.00000071, -0.00000304, 0.00001886, -0.00001708, 0.00001566, -0.00000365, 0.00000959, 0.00000537, 0.00000182, -0.00000863, 0.00001285, -0.00001826, 0.00003339, -0.00002437, 0.00002232, -0.00002051, },
		{ 0.00001503, -0.00000590, -0.00000154, -0.00000956, 0.00001376, -0.00000603, -0.00000129, -0.00000556, 0.00000578, 0.00000566, -0.00001274, 0.00000591, -0.00000179, -0.00000056, 0.00000461, 0.00000379, -0.00000357, -0.00001159, 0.00001789, -0.00000817, -0.00000108, -0.00001021, 0.00002586, -0.00001592, 0.00000459, -0.00000802, -0.00000252, 0.00000858, -0.00001660, 0.00001070, -0.00001162, 0.00000507, -0.00000176, 0.00000208, -0.00000166, -0.00000287, -0.00000287, -0.00000015, -0.00000399, 0.00000920, -0.00001701, -0.00000249, 0.00000735, -0.00001964, 0.00002157, -0.00002303, 0.00002748, -0.00001183, -0.00001155, -0.00000057, 0.00001735, 0.00000456, -0.00001070, 0.00000645, -0.00001688, 0.00000554, -0.00000154, 0.00000731, -0.00000637, 0.00000504, -0.00000847, 0.00000862, -0.00001391, 0.00000664, -0.00001504, 0.00000480, -0.00001901, 0.00000694, -0.00002872, 0.00001212, -0.00003593, 0.00000923, -0.00007757, 0.00001014, -0.00035979, 0.00006636, 0.00080265, 0.00025194, -0.00038514, -0.00003185, -0.00005206, -0.00001328, -0.00002443, -0.00000425, -0.00002079, -0.00000679, -0.00000196, -0.00000564, 0.00000030, -0.00000356, -0.00000346, 0.00000155, -0.00000587, -0.00000321, 0.00000096, -0.00000159, -0.00001255, -0.00000056, 0.00000257, -0.00000149, 0.00000551, 0.00000027, -0.00001593, -0.00001057, 0.00002883, -0.00001461, 0.00000533, -0.00001471, 0.00001257, -0.00001574, -0.00000374, 0.00001202, -0.00001475, 0.00002451, -0.00001618, 0.00001030, 0.00000011, 0.00000673, 0.00000381, 0.00000996, -0.00001676, 0.00001829, -0.00002053, 0.00002398, -0.00000610, 0.00000133, 0.00000536, -0.00001755, },
		{ -0.00000229, -0.00000081, -0.00001125, 0.00000995, 0.00000119, -0.00000564, -0.00000255, -0.00000070, 0.00001157, -0.00001183, 0.00000032, 0.00000396, -0.00000545, 0.00000636, 0.00000186, 0.00000290, -0.00001628, 0.00001435, -0.00000108, -0.00000350, -0.00000983, 0.00001548, 0.00000409, -0.00001477, 0.00001011, -0.00001775, 0.00001731, -0.00001621, 0.00000502, -0.00000429, -0.00000254, 0.00000256, -0.00000042, 0.00000062, -0.00000304, -0.00000276, -0.00000038, -0.00000400, 0.00000612, -0.00000592, -0.00001584, 0.00001614, -0.00002245, 0.00001869, -0.00001440, 0.00000754, 0.00001363, -0.00002491, 0.00000292, 0.00000422, 0.00002219, -0.00002110, 0.00001362, -0.00001830, 0.00000306, -0.00000312, 0.00000994, -0.00000858, 0.00000814, -0.00001152, 0.00001155, -0.00001534, 0.00000781, -0.00001523, 0.00000582, -0.00001884, 0.00000676, -0.00002573, 0.00000705, -0.00002922, 0.00000568, -0.00005854, -0.00000676, -0.00019454, -0.00029912, 0.00073944, 0.00058728, -0.00037627, -0.00012356, -0.00002373, -0.00003863, -0.00000872, -0.00002082, -0.00000485, -0.00002391, 0.00000898, -0.00001481, 0.00000661, -0.00000826, 0.00000086, -0.00000433, 0.00000139, -0.00000988, 0.00000562, -0.00000531, -0.00000340, -0.00001352, 0.00001192, -0.00000919, 0.00001006, -0.00000286, -0.00000166, -0.00002868, 0.00003066, -0.00000571, -0.00000010, -0.00000921, 0.00000194, 0.00000032, -0.00001896, 0.00002115, -0.00001895, 0.00002100, -0.00000748, 0.00000046, 0.00000578, 0.00000297, 0.00000376, 0.00001192, -0.00001267, 0.00000761, -0.00000403, -0.00000200, 0.00002092, -0.00002086, 0.00002392, -0.00002552, 0.00001320, },
		{ -0.00000875, -0.00000347, -0.00000249, 0.00001316, -0.00001191, 0.00000218, -0.00000748, 0.00001461, -0.00000759, -0.00000558, 0.00000681, -0.00000533, 0.00000304, 0.00000387, 0.00000389, -0.00001267, 0.00000357, 0.00000923, -0.00000811, -0.00000436, 0.00000061, 0.00002080, -0.00002231, 0.00001063, -0.00001190, 0.00000483, 0.00000102, -0.00001067, 0.00000754, -0.00001077, 0.00000619, -0.00000312, 0.00000373, -0.00000494, 0.00000011, -0.00000379, 0.00000007, -0.00000140, 0.00000620, -0.00002431, 0.00001655, -0.00001683, 0.00000825, 0.00000076, -0.00001251, 0.00002728, -0.00002160, -0.00000410, -0.00000058, 0.00002517, -0.00001312, 0.00000246, -0.00000349, -0.00001082, 0.00000421, 0.00000305, -0.00000075, 0.00000040, -0.00000284, 0.00000176, -0.00000304, -0.00000557, -0.00000034, -0.00000916, -0.00000156, -0.00001166, -0.00000347, -0.00001774, -0.00000094, -0.00002631, -0.00000596, -0.00006935, -0.00002796, -0.00044477, 0.00036485, 0.00097853, -0.00027326, -0.00023483, -0.00001293, -0.00005049, -0.00000556, -0.00002624, -0.00000074, -0.00002735, 0.00000574, -0.00001122, 0.00000157, -0.00000381, -0.00000185, -0.00000416, 0.00000238, -0.00001028, 0.00000496, -0.00000545, 0.00000145, -0.00001773, 0.00000966, -0.00000510, 0.00000433, 0.00000087, 0.00000314, -0.00002744, 0.00001091, 0.00001636, -0.00001287, 0.00000305, -0.00001316, 0.00001501, -0.00002729, 0.00002119, -0.00001387, 0.00001018, 0.00000461, -0.00000883, 0.00000940, 0.00000057, 0.00000460, 0.00000862, -0.00000114, -0.00000840, 0.00001342, -0.00002043, 0.00003050, -0.00002000, 0.00001599, -0.00000793, -0.00000714, 0.00001261, },
		{ -0.00000058, -0.00000893, 0.00001592, -0.00000841, -0.00000114, -0.00000595, 0.00000848, 0.00000314, -0.00001329, 0.00000875, -0.00000398, 0.00000013, 0.00000395, 0.00000549, -0.00000875, -0.00000430, 0.00001265, -0.00000584, -0.00000419, -0.00000518, 0.00002051, -0.00000920, -0.00000752, 0.00000769, -0.00001429, 0.00001654, -0.00001824, 0.00000910, -0.00000904, 0.00000223, -0.00000105, 0.00000314, -0.00000388, 0.00000033, -0.00000419, 0.00000104, -0.00000450, 0.00000979, -0.00002046, 0.00000518, -0.00000264, -0.00000763, 0.00001715, -0.00002549, 0.00002662, -0.00000796, -0.00001180, -0.00000468, 0.00002171, 0.00000030, -0.00000925, 0.00000663, -0.00001496, 0.00000135, 0.00000443, -0.00000055, 0.00000008, -0.00000171, -0.00000020, 0.00000064, -0.00000876, 0.00000227, -0.00001000, -0.00000056, -0.00001114, -0.00000337, -0.00001615, -0.00000317, -0.00002329, -0.00000225, -0.00006110, -0.00001220, -0.00029598, -0.00018765, 0.00120065, 0.00004678, -0.00040091, 0.00000742, -0.00007713, 0.00000880, -0.00004039, 0.00000857, -0.00003319, 0.00000460, -0.00001113, 0.00000093, -0.00000493, 0.00000123, -0.00000861, 0.00000622, -0.00001243, 0.00000555, -0.00000621, 0.00000384, -0.00001526, 0.00000052, 0.00000343, -0.00000329, 0.00000494, 0.00000278, -0.00001520, -0.00001381, 0.00003180, -0.00001572, 0.00000673, -0.00001766, 0.00001530, -0.00002019, 0.00000816, 0.00000087, -0.00000584, 0.00001718, -0.00001561, 0.00001065, -0.00000108, 0.00000650, 0.00000287, 0.00001002, -0.00001757, 0.00001748, -0.00001837, 0.00001597, 0.00000212, -0.00000796, 0.00001713, -0.00002488, 0.00001975, -0.00000966, },
		{ -0.00000261, 0.00000480, 0.00000650, -0.00001252, 0.00000302, -0.00000417, 0.00001618, -0.00001944, 0.00000894, -0.00000294, 0.00000023, 0.00000002, 0.00000951, -0.00000737, -0.00000611, 0.00000800, 0.00000254, -0.00000875, -0.00000272, 0.00000875, 0.00001107, -0.00002406, 0.00001734, -0.00001693, 0.00001170, -0.00000606, -0.00000467, 0.00000331, -0.00000807, 0.00000497, -0.00000156, 0.00000159, -0.00000437, 0.00000109, -0.00000408, 0.00000007, 0.00000289, -0.00000581, -0.00001316, 0.00001413, -0.00002278, 0.00002867, -0.00002847, 0.00001694, 0.00000400, -0.00000955, -0.00001480, 0.00002017, 0.00000659, -0.00000751, -0.00000043, -0.00000082, -0.00001625, 0.00001714, -0.00001123, 0.00001097, -0.00001250, 0.00001071, -0.00001023, 0.00000503, -0.00001444, 0.00001016, -0.00002240, 0.00001422, -0.00003047, 0.00001556, -0.00003866, 0.00001828, -0.00005023, 0.00001496, -0.00010521, -0.00003572, -0.00058422, 0.00098990, 0.00063996, -0.00059196, 0.00002608, -0.00011894, 0.00003510, -0.00006496, 0.00002605, -0.00004641, 0.00001234, -0.00002175, 0.00001135, -0.00001709, 0.00001387, -0.00002023, 0.00001515, -0.00001855, 0.00000941, -0.00000939, 0.00000643, -0.00001211, -0.00000564, 0.00000491, -0.00000196, 0.00000062, 0.00000622, -0.00000670, -0.00002404, 0.00002448, 0.00000096, -0.00000472, -0.00000652, -0.00000040, -0.00000023, -0.00001278, 0.00001920, -0.00002127, 0.00002608, -0.00001818, 0.00000970, -0.00000249, 0.00000878, -0.00000216, 0.00001559, -0.00001583, 0.00000842, -0.00000344, -0.00000595, 0.00002173, -0.00002058, 0.00002261, -0.00001837, 0.00000506, 0.00000538, -0.00000984, },
		{ -0.00000769, 0.00001912, -0.00001888, 0.00000737, -0.00001138, 0.00002058, -0.00001615, 0.00000266, 0.00000092, -0.00000022, -0.00000289, 0.00001050, -0.00000373, -0.00000748, 0.00000319, 0.00000787, -0.00000918, -0.00000230, 0.00000054, 0.00001944, -0.00002064, 0.00000407, 0.00000183, -0.00000970, 0.00001499, -0.00001968, 0.00001251, -0.00001321, 0.00000653, -0.00000403, 0.00000588, -0.00000846, 0.00000602, -0.00000877, 0.00000500, -0.00000500, 0.00000732, -0.00002506, 0.00002116, -0.00002652, 0.00002670, -0.00001779, 0.00000075, 0.00001157, -0.00000147, -0.00002244, 0.00001424, 0.00001054, -0.00000015, -0.00001162, 0.00001323, -0.00002734, 0.00002027, -0.00001286, 0.00001339, -0.00001515, 0.00001364, -0.00001381, 0.00001208, -0.00002375, 0.00002104, -0.00003287, 0.00002575, -0.00004098, 0.00002766, -0.00005081, 0.00003314, -0.00006896, 0.00005031, -0.00014058, 0.00007182, -0.00056555, 0.00031794, 0.00131297, -0.00061847, -0.00006094, -0.00010842, 0.00001953, -0.00005722, 0.00001731, -0.00003921, 0.00000548, -0.00002170, 0.00001249, -0.00002053, 0.00001697, -0.00002145, 0.00001379, -0.00001549, 0.00000543, -0.00000601, 0.00000338, -0.00000633, -0.00000870, 0.00000028, 0.00000505, -0.00000721, 0.00001048, -0.00000336, -0.00001920, 0.00000170, 0.00002377, -0.00001741, 0.00000649, -0.00001658, 0.00001601, -0.00002620, 0.00002784, -0.00002543, 0.00002372, -0.00001186, 0.00000342, -0.00000051, 0.00000762, -0.00000227, 0.00001318, -0.00000585, -0.00000462, 0.00000907, -0.00001609, 0.00002081, -0.00000912, 0.00000378, 0.00000758, -0.00001980, 0.00002290, -0.00001915, 0.00000518, },
		{ 0.00001537, -0.00000895, -0.00000198, -0.00000517, 0.00001057, -0.00000084, -0.00001083, 0.00000887, -0.00000437, -0.00000015, 0.00000411, 0.00000569, -0.00001258, 0.00000363, 0.00000520, -0.00000183, -0.00000811, 0.00000162, 0.00001234, -0.00000095, -0.00001929, 0.00002169, -0.00002335, 0.00002093, -0.00001624, 0.00000453, -0.00000411, -0.00000227, 0.00000092, 0.00000306, -0.00000508, 0.00000284, -0.00000466, 0.00000114, -0.00000322, 0.00000729, -0.00002007, 0.00001046, -0.00001276, 0.00000744, 0.00000671, -0.00002085, 0.00001942, -0.00000039, -0.00001351, -0.00000660, 0.00002323, -0.00000322, -0.00000646, 0.00000473, -0.00001094, -0.00000383, 0.00001024, -0.00000820, 0.00000654, -0.00000857, 0.00000849, -0.00000852, -0.00000182, -0.00000392, -0.00000310, -0.00000769, -0.00000149, -0.00001578, -0.00000072, -0.00002220, -0.00000572, -0.00002065, -0.00003445, -0.00005285, -0.00020056, -0.00047642, 0.00164058, -0.00022187, -0.00035999, 0.00001757, -0.00008731, 0.00002088, -0.00004863, 0.00001472, -0.00003993, 0.00001218, -0.00001814, 0.00000618, -0.00000919, 0.00000487, -0.00001217, 0.00000883, -0.00001661, 0.00001238, -0.00001206, 0.00000782, -0.00001624, 0.00000084, 0.00000375, -0.00000420, 0.00000362, 0.00000523, -0.00001470, -0.00001354, 0.00002854, -0.00001114, 0.00000343, -0.00001561, 0.00001263, -0.00001942, 0.00001706, -0.00001121, 0.00000565, 0.00000603, -0.00001063, 0.00000800, -0.00000038, 0.00000501, 0.00000334, 0.00000755, -0.00001424, 0.00001215, -0.00001175, 0.00000507, 0.00001144, -0.00001639, 0.00002499, -0.00002732, 0.00002024, -0.00001085, -0.00000091, -0.00000638, },
		{ 0.00001101, -0.00001780, 0.00000716, -0.00000535, 0.00001674, -0.00002270, 0.00001462, -0.00000827, 0.00000498, -0.00000518, 0.00001542, -0.00001657, 0.00000478, 0.00000045, 0.00000591, -0.00001277, 0.00000378, 0.00000284, 0.00001492, -0.00002845, 0.00001928, -0.00001124, 0.00000190, 0.00000701, -0.00001603, 0.00001202, -0.00001435, 0.00000825, -0.00000385, 0.00000380, -0.00000681, 0.00000637, -0.00000954, 0.00000647, -0.00000324, -0.00000332, -0.00001046, 0.00000967, -0.00001875, 0.00003321, -0.00003862, 0.00002542, -0.00000794, 0.00000928, -0.00003466, 0.00003850, -0.00001401, 0.00001237, -0.00002029, 0.00002146, -0.00003893, 0.00004022, -0.00003597, 0.00003485, -0.00003801, 0.00003853, -0.00003865, 0.00003265, -0.00004397, 0.00004302, -0.00005833, 0.00005554, -0.00007692, 0.00006757, -0.00009678, 0.00008032, -0.00012342, 0.00010920, -0.00023027, 0.00015431, -0.00089426, 0.00126988, 0.00064863, -0.00078342, 0.00018160, -0.00022737, 0.00012080, -0.00013173, 0.00008228, -0.00009694, 0.00005805, -0.00006194, 0.00004620, -0.00004845, 0.00004243, -0.00004674, 0.00003936, -0.00004298, 0.00003370, -0.00002981, 0.00002341, -0.00002610, 0.00000861, -0.00000907, 0.00001163, -0.00001442, 0.00002032, -0.00001783, -0.00001055, 0.00000920, 0.00001399, -0.00001397, 0.00000064, -0.00000659, 0.00000164, -0.00000613, 0.00001354, -0.00001961, 0.00002724, -0.00002576, 0.00001798, -0.00001080, 0.00001478, -0.00000772, 0.00001795, -0.00001714, 0.00000849, -0.00000354, -0.00000689, 0.00001794, -0.00001454, 0.00001457, -0.00000624, -0.00000563, 0.00001338, -0.00001837, 0.00000644, -0.00000098, },
		{ -0.00002224, 0.00000952, -0.00001159, 0.00002201, -0.00002121, 0.00000953, -0.00000472, 0.00000467, -0.00000800, 0.00001591, -0.00001109, -0.00000126, 0.00000234, 0.00000513, -0.00000822, -0.00000169, 0.00000265, 0.00001402, -0.00001473, -0.00000521, 0.00001641, -0.00002367, 0.00002720, -0.00002651, 0.00001576, -0.00001443, 0.00000636, -0.00000486, 0.00000797, -0.00001178, 0.00001204, -0.00001407, 0.00001048, -0.00000903, 0.00000763, -0.00002267, 0.00002133, -0.00003150, 0.00004146, -0.00003600, 0.00001714, -0.00000916, 0.00002291, -0.00004329, 0.00003252, -0.00000977, 0.00001824, -0.00002944, 0.00003225, -0.00004524, 0.00003818, -0.00003204, 0.00003165, -0.00003528, 0.00003531, -0.00003524, 0.00003303, -0.00004729, 0.00004768, -0.00006198, 0.00005894, -0.00007705, 0.00006669, -0.00009286, 0.00007731, -0.00012257, 0.00011822, -0.00022444, 0.00018161, -0.00068840, 0.00028232, 0.00162637, -0.00097161, 0.00014670, -0.00023409, 0.00011212, -0.00012730, 0.00007663, -0.00009204, 0.00005202, -0.00006149, 0.00004670, -0.00005146, 0.00004516, -0.00004764, 0.00003804, -0.00003977, 0.00002820, -0.00002431, 0.00001918, -0.00002110, 0.00000782, -0.00001602, 0.00002066, -0.00002328, 0.00002444, -0.00001540, -0.00000418, -0.00001212, 0.00003219, -0.00002187, 0.00001002, -0.00001922, 0.00001471, -0.00002058, 0.00002845, -0.00003246, 0.00003410, -0.00002692, 0.00001652, -0.00001168, 0.00001562, -0.00000950, 0.00001773, -0.00001108, 0.00000148, 0.00000100, -0.00000670, 0.00000627, 0.00000580, -0.00001107, 0.00002235, -0.00002939, 0.00002908, -0.00002599, 0.00001177, -0.00001075, 0.00002157, },
		{ -0.00000396, -0.00000282, 0.00001112, -0.00000514, -0.00000612, 0.00000650, -0.00000298, -0.00000142, 0.00000525, 0.00000380, -0.00001369, 0.00001008, -0.00000240, 0.00000228, -0.00001131, 0.00000828, 0.00000362, 0.00000633, -0.00002757, 0.00003055, -0.00002697, 0.00002027, -0.00001036, -0.00000246, 0.00000239, -0.00000771, 0.00000381, 0.00000169, -0.00000472, 0.00000401, -0.00000387, -0.00000007, -0.00000011, 0.00000166, -0.00001503, 0.00001127, -0.00001964, 0.00002155, -0.00000647, -0.00001135, 0.00000782, 0.00000933, -0.00001603, -0.00000613, 0.00002112, -0.00000479, -0.00000353, 0.00000187, -0.00000613, -0.00000838, 0.00001395, -0.00001339, 0.00001062, -0.00001297, 0.00001474, -0.00001602, 0.00000434, -0.00000953, 0.00000311, -0.00001374, 0.00000651, -0.00002601, 0.00001243, -0.00003920, 0.00001071, -0.00003910, -0.00000667, -0.00009801, -0.00009857, -0.00075697, 0.00207138, -0.00051872, -0.00027394, -0.00000211, -0.00007910, 0.00001920, -0.00004579, 0.00001130, -0.00003815, 0.00001307, -0.00002034, 0.00000790, -0.00001067, 0.00000574, -0.00001265, 0.00000790, -0.00001666, 0.00001521, -0.00001462, 0.00000889, -0.00001458, -0.00000008, 0.00000379, -0.00000449, 0.00000208, 0.00000692, -0.00001372, -0.00001048, 0.00001993, -0.00000191, -0.00000365, -0.00000912, 0.00000464, -0.00001236, 0.00002000, -0.00002024, 0.00001601, -0.00000588, -0.00000333, 0.00000356, 0.00000137, 0.00000296, 0.00000446, 0.00000358, -0.00000811, 0.00000390, -0.00000266, -0.00000636, 0.00001936, -0.00002140, 0.00002686, -0.00002360, 0.00001568, -0.00000896, -0.00000263, -0.00000205, 0.00001205, -0.00000745, },
		{ 0.00000898, -0.00000377, 0.00001269, -0.00002036, 0.00001564, -0.00001015, 0.00000688, -0.00000714, 0.00001712, -0.00002161, 0.00001317, -0.00000569, 0.00000715, -0.00001378, 0.00000766, -0.00000026, 0.00001393, -0.00002631, 0.00001497, -0.00000112, -0.00001043, 0.00002077, -0.00002689, 0.00002016, -0.00002030, 0.00001195, -0.00000726, 0.00000734, -0.00001027, 0.00001263, -0.00001616, 0.00001461, -0.00001201, 0.00000224, -0.00000942, 0.00000489, -0.00001190, 0.00003225, -0.00004338, 0.00003089, -0.00001929, 0.00002765, -0.00005215, 0.00005446, -0.00003471, 0.00003379, -0.00004132, 0.00004374, -0.00005999, 0.00006095, -0.00005893, 0.00005765, -0.00006279, 0.00006614, -0.00006792, 0.00006181, -0.00007483, 0.00007681, -0.00009472, 0.00009689, -0.00012335, 0.00012003, -0.00015622, 0.00014493, -0.00020000, 0.00020688, -0.00035930, 0.00034465, -0.00119202, 0.00155239, 0.00061241, -0.00092800, 0.00031720, -0.00032814, 0.00020195, -0.00019574, 0.00013671, -0.00014649, 0.00010430, -0.00010379, 0.00008300, -0.00008153, 0.00007202, -0.00007364, 0.00006358, -0.00006722, 0.00005848, -0.00005179, 0.00004253, -0.00004285, 0.00002690, -0.00002771, 0.00002918, -0.00003252, 0.00003690, -0.00003262, 0.00000939, -0.00001286, 0.00003145, -0.00002661, 0.00001177, -0.00001696, 0.00000753, -0.00000191, 0.00000697, -0.00001548, 0.00002483, -0.00002945, 0.00002394, -0.00001779, 0.00001986, -0.00001220, 0.00001870, -0.00001662, 0.00000788, -0.00000443, -0.00000488, 0.00001004, -0.00000377, 0.00000165, 0.00000842, -0.00001645, 0.00001947, -0.00002415, 0.00001383, -0.00000578, 0.00001338, -0.00002025, },
		{ -0.00000751, 0.00001783, -0.00002093, 0.00001282, -0.00000796, 0.00000729, -0.00001021, 0.00001789, -0.00001585, 0.00000414, 0.00000185, 0.00000094, -0.00000518, -0.00000266, 0.00000666, 0.00000535, -0.00000577, -0.00001524, 0.00002892, -0.00003398, 0.00003506, -0.00002950, 0.00001652, -0.00001427, 0.00000590, -0.00000588, 0.00001001, -0.00001449, 0.00001741, -0.00001894, 0.00001597, -0.00001340, 0.00000776, -0.00001809, 0.00001739, -0.00003060, 0.00004934, -0.00004952, 0.00003192, -0.00002962, 0.00004590, -0.00006129, 0.00004927, -0.00003105, 0.00003784, -0.00004710, 0.00004977, -0.00006076, 0.00005422, -0.00005076, 0.00005034, -0.00005624, 0.00005833, -0.00005882, 0.00005620, -0.00007225, 0.00007513, -0.00009149, 0.00009244, -0.00011383, 0.00010730, -0.00013748, 0.00012488, -0.00017940, 0.00018835, -0.00030995, 0.00028997, -0.00080829, 0.00027421, 0.00187647, -0.00128286, 0.00034963, -0.00036426, 0.00020912, -0.00020194, 0.00014036, -0.00014932, 0.00010366, -0.00010637, 0.00008526, -0.00008580, 0.00007579, -0.00007580, 0.00006421, -0.00006624, 0.00005418, -0.00004677, 0.00003940, -0.00004045, 0.00002929, -0.00003657, 0.00003901, -0.00004120, 0.00004025, -0.00003073, 0.00001553, -0.00002893, 0.00004081, -0.00002629, 0.00001429, -0.00002227, 0.00001298, -0.00001247, 0.00002371, -0.00003375, 0.00003890, -0.00003745, 0.00002733, -0.00002150, 0.00002246, -0.00001546, 0.00002091, -0.00001541, 0.00000814, -0.00000864, 0.00000509, -0.00000996, 0.00002123, -0.00002475, 0.00003287, -0.00003318, 0.00002895, -0.00002728, 0.00001463, -0.00001048, 0.00001954, -0.00002107, 0.00000794, },
		{ 0.00000940, -0.00000834, -0.00000058, 0.00000272, -0.00000063, -0.00000306, 0.00000682, 0.00000013, -0.00001155, 0.00001355, -0.00000832, 0.00000496, -0.00001243, 0.00001300, -0.00000359, 0.00001059, -0.00003016, 0.00003210, -0.00002391, 0.00001442, -0.00000144, -0.00001049, 0.00000900, -0.00001279, 0.00000674, -0.00000056, -0.00000280, 0.00000375, -0.00000192, -0.00000219, 0.00000408, -0.00000606, -0.00000593, 0.00000687, -0.00002077, 0.00003074, -0.00001822, -0.00000070, -0.00000483, 0.00001853, -0.00001779, -0.00000410, 0.00001566, -0.00000398, -0.00000137, -0.00000088, -0.00000180, -0.00001098, 0.00001419, -0.00001465, 0.00001075, -0.00001316, 0.00001644, -0.00001952, 0.00000750, -0.00001241, 0.00000643, -0.00001630, 0.00000995, -0.00003042, 0.00001869, -0.00004876, 0.00001998, -0.00005046, 0.00001243, -0.00013234, -0.00001747, -0.00098259, 0.00243518, -0.00079782, -0.00016586, -0.00004314, -0.00005555, 0.00000487, -0.00003244, -0.00000126, -0.00002786, 0.00000670, -0.00001686, 0.00000521, -0.00000865, 0.00000329, -0.00000963, 0.00000312, -0.00001208, 0.00001318, -0.00001294, 0.00000639, -0.00001002, -0.00000240, 0.00000378, -0.00000450, 0.00000087, 0.00000710, -0.00001128, -0.00000627, 0.00000835, 0.00000932, -0.00001196, -0.00000053, -0.00000659, -0.00000016, 0.00001602, -0.00002344, 0.00002196, -0.00001533, 0.00000375, -0.00000101, 0.00000302, 0.00000127, 0.00000518, -0.00000043, -0.00000116, -0.00000494, 0.00000634, -0.00001578, 0.00002364, -0.00002129, 0.00002206, -0.00001397, 0.00000655, -0.00000436, -0.00000529, 0.00000343, 0.00000653, -0.00000425, -0.00000682, 0.00000109, },
		{ 0.00000497, -0.00001295, 0.00001189, -0.00000810, 0.00000523, -0.00000503, 0.00001345, -0.00002066, 0.00001672, -0.00000843, 0.00000502, -0.00001080, 0.00000778, 0.00000031, 0.00000778, -0.00001774, 0.00000514, 0.00001165, -0.00002280, 0.00003247, -0.00003481, 0.00002587, -0.00002459, 0.00001525, -0.00001130, 0.00001184, -0.00001420, 0.00001899, -0.00002280, 0.00002323, -0.00002233, 0.00001071, -0.00001129, 0.00000216, -0.00000486, 0.00002730, -0.00004285, 0.00003233, -0.00002772, 0.00004201, -0.00006326, 0.00006365, -0.00005054, 0.00005159, -0.00005841, 0.00006093, -0.00007446, 0.00007434, -0.00007495, 0.00007393, -0.00008105, 0.00008728, -0.00009153, 0.00008594, -0.00010018, 0.00010417, -0.00012349, 0.00012925, -0.00015954, 0.00016144, -0.00020428, 0.00019805, -0.00026359, 0.00028773, -0.00046612, 0.00050038, -0.00143511, 0.00180250, 0.00054074, -0.00101185, 0.00041194, -0.00040232, 0.00026280, -0.00024391, 0.00017807, -0.00018474, 0.00014141, -0.00013847, 0.00011389, -0.00010938, 0.00009638, -0.00009527, 0.00008252, -0.00008607, 0.00007858, -0.00007053, 0.00005958, -0.00005850, 0.00004540, -0.00004707, 0.00004674, -0.00004970, 0.00005200, -0.00004697, 0.00003124, -0.00003712, 0.00004943, -0.00003966, 0.00002471, -0.00003016, 0.00001753, -0.00000217, 0.00000237, -0.00001117, 0.00002021, -0.00002952, 0.00002694, -0.00002262, 0.00002330, -0.00001526, 0.00001803, -0.00001470, 0.00000692, -0.00000608, -0.00000058, -0.00000055, 0.00000955, -0.00001321, 0.00002259, -0.00002500, 0.00002190, -0.00002562, 0.00001810, -0.00000848, 0.00001293, -0.00001946, 0.00000850, -0.00000011, },
		{ -0.00001513, 0.00001165, -0.00000767, 0.00000646, -0.00000836, 0.00001510, -0.00001631, 0.00000712, 0.00000238, -0.00000447, -0.00000049, -0.00000481, 0.00001163, -0.00000498, 0.00000437, -0.00002382, 0.00003679, -0.00003831, 0.00003692, -0.00002777, 0.00001460, -0.00001289, 0.00000551, -0.00000737, 0.00001209, -0.00001634, 0.00002157, -0.00002273, 0.00002095, -0.00001805, 0.00000864, -0.00001306, 0.00001110, -0.00002453, 0.00004908, -0.00005537, 0.00004103, -0.00004483, 0.00006212, -0.00007135, 0.00005948, -0.00004815, 0.00005385, -0.00006030, 0.00006182, -0.00007021, 0.00006464, -0.00006497, 0.00006517, -0.00007341, 0.00007784, -0.00007965, 0.00007670, -0.00009350, 0.00009782, -0.00011532, 0.00011951, -0.00014388, 0.00014144, -0.00017617, 0.00016672, -0.00022905, 0.00024852, -0.00038307, 0.00038192, -0.00091531, 0.00030243, 0.00202903, -0.00150668, 0.00050892, -0.00047056, 0.00028934, -0.00026445, 0.00019443, -0.00019849, 0.00014914, -0.00014651, 0.00011970, -0.00011615, 0.00010223, -0.00009984, 0.00008656, -0.00008922, 0.00007785, -0.00006841, 0.00005955, -0.00006005, 0.00005130, -0.00005703, 0.00005600, -0.00005707, 0.00005419, -0.00004562, 0.00003623, -0.00004581, 0.00004807, -0.00003021, 0.00001954, -0.00002666, 0.00001284, -0.00000480, 0.00001595, -0.00002997, 0.00003740, -0.00004167, 0.00003361, -0.00002794, 0.00002653, -0.00001896, 0.00002177, -0.00001770, 0.00001361, -0.00001746, 0.00001625, -0.00002439, 0.00003353, -0.00003388, 0.00003678, -0.00003003, 0.00002206, -0.00002263, 0.00001342, -0.00000690, 0.00001339, -0.00001583, 0.00000331, -0.00000012, 0.00000873, },
		{ 0.00000421, -0.00000101, 0.00000149, -0.00000403, 0.00000779, -0.00000422, -0.00000671, 0.00001354, -0.00001146, 0.00000533, -0.00001071, 0.00001467, -0.00000796, 0.00001086, -0.00002639, 0.00002610, -0.00001473, 0.00000472, 0.00000866, -0.00001757, 0.00001392, -0.00001597, 0.00000851, -0.00000278, 0.00000004, 0.00000253, 0.00000056, -0.00000429, 0.00000793, -0.00001384, 0.00000521, -0.00000173, -0.00001549, 0.00003238, -0.00002487, 0.00000730, -0.00001487, 0.00002432, -0.00001692, -0.00000197, 0.00000808, -0.00000122, -0.00000028, -0.00000285, 0.00000129, -0.00001107, 0.00001069, -0.00001177, 0.00000673, -0.00000881, 0.00001295, -0.00001806, 0.00000662, -0.00001155, 0.00000579, -0.00001428, 0.00000760, -0.00002733, 0.00001570, -0.00004777, 0.00001843, -0.00005002, 0.00001581, -0.00014395, 0.00002022, -0.00111301, 0.00267866, -0.00101210, -0.00006450, -0.00009071, -0.00002502, -0.00001660, -0.00001294, -0.00001919, -0.00001206, -0.00000488, -0.00000896, -0.00000100, -0.00000386, -0.00000172, -0.00000401, -0.00000438, -0.00000396, 0.00000699, -0.00000751, 0.00000091, -0.00000322, -0.00000568, 0.00000364, -0.00000436, 0.00000026, 0.00000547, -0.00000666, -0.00000276, -0.00000320, 0.00001916, -0.00001822, 0.00000686, -0.00001753, 0.00001405, 0.00000650, -0.00002014, 0.00002169, -0.00002000, 0.00000847, -0.00000420, 0.00000358, 0.00000063, 0.00000472, -0.00000316, 0.00000465, -0.00001209, 0.00001289, -0.00002117, 0.00002301, -0.00001583, 0.00001166, -0.00000030, -0.00000536, 0.00000173, -0.00000791, 0.00000894, -0.00000006, -0.00000024, -0.00000979, 0.00000595, 0.00000528, -0.00000935, },
		{ 0.00000490, -0.00000320, 0.00000089, 0.00000033, 0.00000513, -0.00001378, 0.00001470, -0.00000708, -0.00000052, -0.00000412, 0.00000433, 0.00000404, -0.00000220, -0.00000461, -0.00000790, 0.00002423, -0.00003232, 0.00003951, -0.00003766, 0.00002752, -0.00002596, 0.00001714, -0.00001507, 0.00001647, -0.00001767, 0.00002423, -0.00002794, 0.00003051, -0.00003228, 0.00002090, -0.00001627, 0.00000317, -0.00000037, 0.00002088, -0.00003848, 0.00003002, -0.00003211, 0.00005023, -0.00006563, 0.00006340, -0.00005797, 0.00006170, -0.00006752, 0.00006925, -0.00007892, 0.00007713, -0.00008051, 0.00007990, -0.00008853, 0.00009710, -0.00010414, 0.00009948, -0.00011422, 0.00011895, -0.00013806, 0.00014532, -0.00017717, 0.00018227, -0.00023012, 0.00022740, -0.00029972, 0.00033378, -0.00052750, 0.00059020, -0.00158410, 0.00198356, 0.00045130, -0.00103194, 0.00045379, -0.00043734, 0.00029223, -0.00026679, 0.00019806, -0.00020381, 0.00016167, -0.00015862, 0.00013216, -0.00012598, 0.00011021, -0.00010691, 0.00009191, -0.00009531, 0.00008962, -0.00008170, 0.00007048, -0.00006904, 0.00005981, -0.00006273, 0.00005999, -0.00006179, 0.00006147, -0.00005630, 0.00004958, -0.00005794, 0.00006306, -0.00004917, 0.00003607, -0.00004331, 0.00002997, -0.00000727, 0.00000164, -0.00000881, 0.00001513, -0.00002689, 0.00002709, -0.00002496, 0.00002473, -0.00001688, 0.00001638, -0.00001209, 0.00000601, -0.00000830, 0.00000496, -0.00001178, 0.00002255, -0.00002665, 0.00003331, -0.00002946, 0.00001983, -0.00002231, 0.00001872, -0.00000880, 0.00000982, -0.00001584, 0.00000648, 0.00000409, -0.00000419, -0.00000261, },
		{ -0.00000418, 0.00000244, -0.00000273, 0.00000769, -0.00001205, 0.00000696, 0.00000410, -0.00001043, 0.00000493, -0.00000724, 0.00001616, -0.00001489, 0.00001356, -0.00002898, 0.00003838, -0.00003568, 0.00003240, -0.00002148, 0.00001043, -0.00001073, 0.00000550, -0.00000948, 0.00001427, -0.00001719, 0.00002410, -0.00002489, 0.00002487, -0.00002274, 0.00001104, -0.00000951, 0.00000502, -0.00001568, 0.00004168, -0.00005277, 0.00004231, -0.00005155, 0.00006805, -0.00007030, 0.00005979, -0.00005709, 0.00006234, -0.00006567, 0.00006551, -0.00007104, 0.00006674, -0.00007151, 0.00007251, -0.00008270, 0.00008925, -0.00009291, 0.00008971, -0.00010614, 0.00011060, -0.00012794, 0.00013400, -0.00016028, 0.00016133, -0.00020034, 0.00019346, -0.00026086, 0.00028629, -0.00042931, 0.00044119, -0.00099520, 0.00036545, 0.00206489, -0.00161111, 0.00059413, -0.00052929, 0.00033439, -0.00029996, 0.00022604, -0.00022789, 0.00017772, -0.00017231, 0.00014166, -0.00013524, 0.00011805, -0.00011392, 0.00009952, -0.00010311, 0.00009353, -0.00008375, 0.00007446, -0.00007484, 0.00006870, -0.00007249, 0.00006720, -0.00006682, 0.00006231, -0.00005584, 0.00005351, -0.00005915, 0.00005188, -0.00003259, 0.00002522, -0.00003243, 0.00001552, -0.00000027, 0.00000822, -0.00002325, 0.00003066, -0.00003948, 0.00003449, -0.00003002, 0.00002705, -0.00001957, 0.00002012, -0.00001748, 0.00001673, -0.00002358, 0.00002423, -0.00003415, 0.00004002, -0.00003649, 0.00003349, -0.00002077, 0.00000987, -0.00001338, 0.00000919, -0.00000127, 0.00000472, -0.00000820, -0.00000250, 0.00000826, -0.00000291, -0.00000546, 0.00000677, },
		{ 0.00000233, -0.00000344, 0.00000710, -0.00000779, -0.00000055, 0.00001059, -0.00001171, 0.00000335, -0.00000628, 0.00001309, -0.00000920, 0.00000744, -0.00001734, 0.00001426, -0.00000176, -0.00000633, 0.00001755, -0.00002163, 0.00001551, -0.00001593, 0.00000800, -0.00000390, 0.00000285, 0.00000121, 0.00000267, -0.00000525, 0.00000980, -0.00001919, 0.00001543, -0.00001188, -0.00000531, 0.00002614, -0.00002449, 0.00001000, -0.00001961, 0.00002466, -0.00001209, -0.00000125, 0.00000010, 0.00000237, 0.00000006, -0.00000393, 0.00000298, -0.00000898, 0.00000432, -0.00000590, -0.00000012, -0.00000127, 0.00000535, -0.00001232, 0.00000218, -0.00000742, 0.00000175, -0.00000844, 0.00000038, -0.00001771, 0.00000422, -0.00003665, 0.00000615, -0.00003734, 0.00000182, -0.00012849, 0.00000336, -0.00112376, 0.00276433, -0.00112376, 0.00000336, -0.00012849, 0.00000182, -0.00003734, 0.00000615, -0.00003665, 0.00000422, -0.00001771, 0.00000038, -0.00000844, 0.00000175, -0.00000742, 0.00000218, -0.00001232, 0.00000535, -0.00000127, -0.00000012, -0.00000590, 0.00000432, -0.00000898, 0.00000298, -0.00000393, 0.00000006, 0.00000237, 0.00000010, -0.00000125, -0.00001209, 0.00002466, -0.00001961, 0.00001000, -0.00002449, 0.00002614, -0.00000531, -0.00001188, 0.00001543, -0.00001919, 0.00000980, -0.00000525, 0.00000267, 0.00000121, 0.00000285, -0.00000390, 0.00000800, -0.00001593, 0.00001551, -0.00002163, 0.00001755, -0.00000633, -0.00000176, 0.00001426, -0.00001734, 0.00000744, -0.00000920, 0.00001309, -0.00000628, 0.00000335, -0.00001171, 0.00001059, -0.00000055, -0.00000779, 0.00000710, -0.00000344, },
		{ -0.00000418, 0.00000677, -0.00000546, -0.00000291, 0.00000826, -0.00000250, -0.00000820, 0.00000472, -0.00000127, 0.00000919, -0.00001338, 0.00000987, -0.00002077, 0.00003349, -0.00003649, 0.00004002, -0.00003415, 0.00002423, -0.00002358, 0.00001673, -0.00001748, 0.00002012, -0.00001957, 0.00002705, -0.00003002, 0.00003449, -0.00003948, 0.00003066, -0.00002325, 0.00000822, -0.00000027, 0.00001552, -0.00003243, 0.00002522, -0.00003259, 0.00005188, -0.00005915, 0.00005351, -0.00005584, 0.00006231, -0.00006682, 0.00006720, -0.00007249, 0.00006870, -0.00007484, 0.00007446, -0.00008375, 0.00009353, -0.00010311, 0.00009952, -0.00011392, 0.00011805, -0.00013524, 0.00014166, -0.00017231, 0.00017772, -0.00022789, 0.00022604, -0.00029996, 0.00033439, -0.00052929, 0.00059413, -0.00161111, 0.00206489, 0.00036545, -0.00099520, 0.00044119, -0.00042931, 0.00028629, -0.00026086, 0.00019346, -0.00020034, 0.00016133, -0.00016028, 0.00013400, -0.00012794, 0.00011060, -0.00010614, 0.00008971, -0.00009291, 0.00008925, -0.00008270, 0.00007251, -0.00007151, 0.00006674, -0.00007104, 0.00006551, -0.00006567, 0.00006234, -0.00005709, 0.00005979, -0.00007030, 0.00006805, -0.00005155, 0.00004231, -0.00005277, 0.00004168, -0.00001568, 0.00000502, -0.00000951, 0.00001104, -0.00002274, 0.00002487, -0.00002489, 0.00002410, -0.00001719, 0.00001427, -0.00000948, 0.00000550, -0.00001073, 0.00001043, -0.00002148, 0.00003240, -0.00003568, 0.00003838, -0.00002898, 0.00001356, -0.00001489, 0.00001616, -0.00000724, 0.00000493, -0.00001043, 0.00000410, 0.00000696, -0.00001205, 0.00000769, -0.00000273, 0.00000244, },
		{ 0.00000490, -0.00000261, -0.00000419, 0.00000409, 0.00000648, -0.00001584, 0.00000982, -0.00000880, 0.00001872, -0.00002231, 0.00001983, -0.00002946, 0.00003331, -0.00002665, 0.00002255, -0.00001178, 0.00000496, -0.00000830, 0.00000601, -0.00001209, 0.00001638, -0.00001688, 0.00002473, -0.00002496, 0.00002709, -0.00002689, 0.00001513, -0.00000881, 0.00000164, -0.00000727, 0.00002997, -0.00004331, 0.00003607, -0.00004917, 0.00006306, -0.00005794, 0.00004958, -0.00005630, 0.00006147, -0.00006179, 0.00005999, -0.00006273, 0.00005981, -0.00006904, 0.00007048, -0.00008170, 0.00008962, -0.00009531, 0.00009191, -0.00010691, 0.00011021, -0.00012598, 0.00013216, -0.00015862, 0.00016167, -0.00020381, 0.00019806, -0.00026679, 0.00029223, -0.00043734, 0.00045379, -0.00103194, 0.00045130, 0.00198356, -0.00158410, 0.00059020, -0.00052750, 0.00033378, -0.00029972, 0.00022740, -0.00023012, 0.00018227, -0.00017717, 0.00014532, -0.00013806, 0.00011895, -0.00011422, 0.00009948, -0.00010414, 0.00009710, -0.00008853, 0.00007990, -0.00008051, 0.00007713, -0.00007892, 0.00006925, -0.00006752, 0.00006170, -0.00005797, 0.00006340, -0.00006563, 0.00005023, -0.00003211, 0.00003002, -0.00003848, 0.00002088, -0.00000037, 0.00000317, -0.00001627, 0.00002090, -0.00003228, 0.00003051, -0.00002794, 0.00002423, -0.00001767, 0.00001647, -0.00001507, 0.00001714, -0.00002596, 0.00002752, -0.00003766, 0.00003951, -0.00003232, 0.00002423, -0.00000790, -0.00000461, -0.00000220, 0.00000404, 0.00000433, -0.00000412, -0.00000052, -0.00000708, 0.00001470, -0.00001378, 0.00000513, 0.00000033, 0.00000089, -0.00000320, },
		{ 0.00000421, -0.00000935, 0.00000528, 0.00000595, -0.00000979, -0.00000024, -0.00000006, 0.00000894, -0.00000791, 0.00000173, -0.00000536, -0.00000030, 0.00001166, -0.00001583, 0.00002301, -0.00002117, 0.00001289, -0.00001209, 0.00000465, -0.00000316, 0.00000472, 0.00000063, 0.00000358, -0.00000420, 0.00000847, -0.00002000, 0.00002169, -0.00002014, 0.00000650, 0.00001405, -0.00001753, 0.00000686, -0.00001822, 0.00001916, -0.00000320, -0.00000276, -0.00000666, 0.00000547, 0.00000026, -0.00000436, 0.00000364, -0.00000568, -0.00000322, 0.00000091, -0.00000751, 0.00000699, -0.00000396, -0.00000438, -0.00000401, -0.00000172, -0.00000386, -0.00000100, -0.00000896, -0.00000488, -0.00001206, -0.00001919, -0.00001294, -0.00001660, -0.00002502, -0.00009071, -0.00006450, -0.00101210, 0.00267866, -0.00111301, 0.00002022, -0.00014395, 0.00001581, -0.00005002, 0.00001843, -0.00004777, 0.00001570, -0.00002733, 0.00000760, -0.00001428, 0.00000579, -0.00001155, 0.00000662, -0.00001806, 0.00001295, -0.00000881, 0.00000673, -0.00001177, 0.00001069, -0.00001107, 0.00000129, -0.00000285, -0.00000028, -0.00000122, 0.00000808, -0.00000197, -0.00001692, 0.00002432, -0.00001487, 0.00000730, -0.00002487, 0.00003238, -0.00001549, -0.00000173, 0.00000521, -0.00001384, 0.00000793, -0.00000429, 0.00000056, 0.00000253, 0.00000004, -0.00000278, 0.00000851, -0.00001597, 0.00001392, -0.00001757, 0.00000866, 0.00000472, -0.00001473, 0.00002610, -0.00002639, 0.00001086, -0.00000796, 0.00001467, -0.00001071, 0.00000533, -0.00001146, 0.00001354, -0.00000671, -0.00000422, 0.00000779, -0.00000403, 0.00000149, -0.00000101, },
		{ -0.00001513, 0.00000873, -0.00000012, 0.00000331, -0.00001583, 0.00001339, -0.00000690, 0.00001342, -0.00002263, 0.00002206, -0.00003003, 0.00003678, -0.00003388, 0.00003353, -0.00002439, 0.00001625, -0.00001746, 0.00001361, -0.00001770, 0.00002177, -0.00001896, 0.00002653, -0.00002794, 0.00003361, -0.00004167, 0.00003740, -0.00002997, 0.00001595, -0.00000480, 0.00001284, -0.00002666, 0.00001954, -0.00003021, 0.00004807, -0.00004581, 0.00003623, -0.00004562, 0.00005419, -0.00005707, 0.00005600, -0.00005703, 0.00005130, -0.00006005, 0.00005955, -0.00006841, 0.00007785, -0.00008922, 0.00008656, -0.00009984, 0.00010223, -0.00011615, 0.00011970, -0.00014651, 0.00014914, -0.00019849, 0.00019443, -0.00026445, 0.00028934, -0.00047056, 0.00050892, -0.00150668, 0.00202903, 0.00030243, -0.00091531, 0.00038192, -0.00038307, 0.00024852, -0.00022905, 0.00016672, -0.00017617, 0.00014144, -0.00014388, 0.00011951, -0.00011532, 0.00009782, -0.00009350, 0.00007670, -0.00007965, 0.00007784, -0.00007341, 0.00006517, -0.00006497, 0.00006464, -0.00007021, 0.00006182, -0.00006030, 0.00005385, -0.00004815, 0.00005948, -0.00007135, 0.00006212, -0.00004483, 0.00004103, -0.00005537, 0.00004908, -0.00002453, 0.00001110, -0.00001306, 0.00000864, -0.00001805, 0.00002095, -0.00002273, 0.00002157, -0.00001634, 0.00001209, -0.00000737, 0.00000551, -0.00001289, 0.00001460, -0.00002777, 0.00003692, -0.00003831, 0.00003679, -0.00002382, 0.00000437, -0.00000498, 0.00001163, -0.00000481, -0.00000049, -0.00000447, 0.00000238, 0.00000712, -0.00001631, 0.00001510, -0.00000836, 0.00000646, -0.00000767, 0.00001165, },
		{ 0.00000497, -0.00000011, 0.00000850, -0.00001946, 0.00001293, -0.00000848, 0.00001810, -0.00002562, 0.00002190, -0.00002500, 0.00002259, -0.00001321, 0.00000955, -0.00000055, -0.00000058, -0.00000608, 0.00000692, -0.00001470, 0.00001803, -0.00001526, 0.00002330, -0.00002262, 0.00002694, -0.00002952, 0.00002021, -0.00001117, 0.00000237, -0.00000217, 0.00001753, -0.00003016, 0.00002471, -0.00003966, 0.00004943, -0.00003712, 0.00003124, -0.00004697, 0.00005200, -0.00004970, 0.00004674, -0.00004707, 0.00004540, -0.00005850, 0.00005958, -0.00007053, 0.00007858, -0.00008607, 0.00008252, -0.00009527, 0.00009638, -0.00010938, 0.00011389, -0.00013847, 0.00014141, -0.00018474, 0.00017807, -0.00024391, 0.00026280, -0.00040232, 0.00041194, -0.00101185, 0.00054074, 0.00180250, -0.00143511, 0.00050038, -0.00046612, 0.00028773, -0.00026359, 0.00019805, -0.00020428, 0.00016144, -0.00015954, 0.00012925, -0.00012349, 0.00010417, -0.00010018, 0.00008594, -0.00009153, 0.00008728, -0.00008105, 0.00007393, -0.00007495, 0.00007434, -0.00007446, 0.00006093, -0.00005841, 0.00005159, -0.00005054, 0.00006365, -0.00006326, 0.00004201, -0.00002772, 0.00003233, -0.00004285, 0.00002730, -0.00000486, 0.00000216, -0.00001129, 0.00001071, -0.00002233, 0.00002323, -0.00002280, 0.00001899, -0.00001420, 0.00001184, -0.00001130, 0.00001525, -0.00002459, 0.00002587, -0.00003481, 0.00003247, -0.00002280, 0.00001165, 0.00000514, -0.00001774, 0.00000778, 0.00000031, 0.00000778, -0.00001080, 0.00000502, -0.00000843, 0.00001672, -0.00002066, 0.00001345, -0.00000503, 0.00000523, -0.00000810, 0.00001189, -0.00001295, },
		{ 0.00000940, 0.00000109, -0.00000682, -0.00000425, 0.00000653, 0.00000343, -0.00000529, -0.00000436, 0.00000655, -0.00001397, 0.00002206, -0.00002129, 0.00002364, -0.00001578, 0.00000634, -0.00000494, -0.00000116, -0.00000043, 0.00000518, 0.00000127, 0.00000302, -0.00000101, 0.00000375, -0.00001533, 0.00002196, -0.00002344, 0.00001602, -0.00000016, -0.00000659, -0.00000053, -0.00001196, 0.00000932, 0.00000835, -0.00000627, -0.00001128, 0.00000710, 0.00000087, -0.00000450, 0.00000378, -0.00000240, -0.00001002, 0.00000639, -0.00001294, 0.00001318, -0.00001208, 0.00000312, -0.00000963, 0.00000329, -0.00000865, 0.00000521, -0.00001686, 0.00000670, -0.00002786, -0.00000126, -0.00003244, 0.00000487, -0.00005555, -0.00004314, -0.00016586, -0.00079782, 0.00243518, -0.00098259, -0.00001747, -0.00013234, 0.00001243, -0.00005046, 0.00001998, -0.00004876, 0.00001869, -0.00003042, 0.00000995, -0.00001630, 0.00000643, -0.00001241, 0.00000750, -0.00001952, 0.00001644, -0.00001316, 0.00001075, -0.00001465, 0.00001419, -0.00001098, -0.00000180, -0.00000088, -0.00000137, -0.00000398, 0.00001566, -0.00000410, -0.00001779, 0.00001853, -0.00000483, -0.00000070, -0.00001822, 0.00003074, -0.00002077, 0.00000687, -0.00000593, -0.00000606, 0.00000408, -0.00000219, -0.00000192, 0.00000375, -0.00000280, -0.00000056, 0.00000674, -0.00001279, 0.00000900, -0.00001049, -0.00000144, 0.00001442, -0.00002391, 0.00003210, -0.00003016, 0.00001059, -0.00000359, 0.00001300, -0.00001243, 0.00000496, -0.00000832, 0.00001355, -0.00001155, 0.00000013, 0.00000682, -0.00000306, -0.00000063, 0.00000272, -0.00000058, -0.00000834, },
		{ -0.00000751, 0.00000794, -0.00002107, 0.00001954, -0.00001048, 0.00001463, -0.00002728, 0.00002895, -0.00003318, 0.00003287, -0.00002475, 0.00002123, -0.00000996, 0.00000509, -0.00000864, 0.00000814, -0.00001541, 0.00002091, -0.00001546, 0.00002246, -0.00002150, 0.00002733, -0.00003745, 0.00003890, -0.00003375, 0.00002371, -0.00001247, 0.00001298, -0.00002227, 0.00001429, -0.00002629, 0.00004081, -0.00002893, 0.00001553, -0.00003073, 0.00004025, -0.00004120, 0.00003901, -0.00003657, 0.00002929, -0.00004045, 0.00003940, -0.00004677, 0.00005418, -0.00006624, 0.00006421, -0.00007580, 0.00007579, -0.00008580, 0.00008526, -0.00010637, 0.00010366, -0.00014932, 0.00014036, -0.00020194, 0.00020912, -0.00036426, 0.00034963, -0.00128286, 0.00187647, 0.00027421, -0.00080829, 0.00028997, -0.00030995, 0.00018835, -0.00017940, 0.00012488, -0.00013748, 0.00010730, -0.00011383, 0.00009244, -0.00009149, 0.00007513, -0.00007225, 0.00005620, -0.00005882, 0.00005833, -0.00005624, 0.00005034, -0.00005076, 0.00005422, -0.00006076, 0.00004977, -0.00004710, 0.00003784, -0.00003105, 0.00004927, -0.00006129, 0.00004590, -0.00002962, 0.00003192, -0.00004952, 0.00004934, -0.00003060, 0.00001739, -0.00001809, 0.00000776, -0.00001340, 0.00001597, -0.00001894, 0.00001741, -0.00001449, 0.00001001, -0.00000588, 0.00000590, -0.00001427, 0.00001652, -0.00002950, 0.00003506, -0.00003398, 0.00002892, -0.00001524, -0.00000577, 0.00000535, 0.00000666, -0.00000266, -0.00000518, 0.00000094, 0.00000185, 0.00000414, -0.00001585, 0.00001789, -0.00001021, 0.00000729, -0.00000796, 0.00001282, -0.00002093, 0.00001783, },
		{ 0.00000898, -0.00002025, 0.00001338, -0.00000578, 0.00001383, -0.00002415, 0.00001947, -0.00001645, 0.00000842, 0.00000165, -0.00000377, 0.00001004, -0.00000488, -0.00000443, 0.00000788, -0.00001662, 0.00001870, -0.00001220, 0.00001986, -0.00001779, 0.00002394, -0.00002945, 0.00002483, -0.00001548, 0.00000697, -0.00000191, 0.00000753, -0.00001696, 0.00001177, -0.00002661, 0.00003145, -0.00001286, 0.00000939, -0.00003262, 0.00003690, -0.00003252, 0.00002918, -0.00002771, 0.00002690, -0.00004285, 0.00004253, -0.00005179, 0.00005848, -0.00006722, 0.00006358, -0.00007364, 0.00007202, -0.00008153, 0.00008300, -0.00010379, 0.00010430, -0.00014649, 0.00013671, -0.00019574, 0.00020195, -0.00032814, 0.00031720, -0.00092800, 0.00061241, 0.00155239, -0.00119202, 0.00034465, -0.00035930, 0.00020688, -0.00020000, 0.00014493, -0.00015622, 0.00012003, -0.00012335, 0.00009689, -0.00009472, 0.00007681, -0.00007483, 0.00006181, -0.00006792, 0.00006614, -0.00006279, 0.00005765, -0.00005893, 0.00006095, -0.00005999, 0.00004374, -0.00004132, 0.00003379, -0.00003471, 0.00005446, -0.00005215, 0.00002765, -0.00001929, 0.00003089, -0.00004338, 0.00003225, -0.00001190, 0.00000489, -0.00000942, 0.00000224, -0.00001201, 0.00001461, -0.00001616, 0.00001263, -0.00001027, 0.00000734, -0.00000726, 0.00001195, -0.00002030, 0.00002016, -0.00002689, 0.00002077, -0.00001043, -0.00000112, 0.00001497, -0.00002631, 0.00001393, -0.00000026, 0.00000766, -0.00001378, 0.00000715, -0.00000569, 0.00001317, -0.00002161, 0.00001712, -0.00000714, 0.00000688, -0.00001015, 0.00001564, -0.00002036, 0.00001269, -0.00000377, },
		{ -0.00000396, -0.00000745, 0.00001205, -0.00000205, -0.00000263, -0.00000896, 0.00001568, -0.00002360, 0.00002686, -0.00002140, 0.00001936, -0.00000636, -0.00000266, 0.00000390, -0.00000811, 0.00000358, 0.00000446, 0.00000296, 0.00000137, 0.00000356, -0.00000333, -0.00000588, 0.00001601, -0.00002024, 0.00002000, -0.00001236, 0.00000464, -0.00000912, -0.00000365, -0.00000191, 0.00001993, -0.00001048, -0.00001372, 0.00000692, 0.00000208, -0.00000449, 0.00000379, -0.00000008, -0.00001458, 0.00000889, -0.00001462, 0.00001521, -0.00001666, 0.00000790, -0.00001265, 0.00000574, -0.00001067, 0.00000790, -0.00002034, 0.00001307, -0.00003815, 0.00001130, -0.00004579, 0.00001920, -0.00007910, -0.00000211, -0.00027394, -0.00051872, 0.00207138, -0.00075697, -0.00009857, -0.00009801, -0.00000667, -0.00003910, 0.00001071, -0.00003920, 0.00001243, -0.00002601, 0.00000651, -0.00001374, 0.00000311, -0.00000953, 0.00000434, -0.00001602, 0.00001474, -0.00001297, 0.00001062, -0.00001339, 0.00001395, -0.00000838, -0.00000613, 0.00000187, -0.00000353, -0.00000479, 0.00002112, -0.00000613, -0.00001603, 0.00000933, 0.00000782, -0.00001135, -0.00000647, 0.00002155, -0.00001964, 0.00001127, -0.00001503, 0.00000166, -0.00000011, -0.00000007, -0.00000387, 0.00000401, -0.00000472, 0.00000169, 0.00000381, -0.00000771, 0.00000239, -0.00000246, -0.00001036, 0.00002027, -0.00002697, 0.00003055, -0.00002757, 0.00000633, 0.00000362, 0.00000828, -0.00001131, 0.00000228, -0.00000240, 0.00001008, -0.00001369, 0.00000380, 0.00000525, -0.00000142, -0.00000298, 0.00000650, -0.00000612, -0.00000514, 0.00001112, -0.00000282, },
		{ -0.00002224, 0.00002157, -0.00001075, 0.00001177, -0.00002599, 0.00002908, -0.00002939, 0.00002235, -0.00001107, 0.00000580, 0.00000627, -0.00000670, 0.00000100, 0.00000148, -0.00001108, 0.00001773, -0.00000950, 0.00001562, -0.00001168, 0.00001652, -0.00002692, 0.00003410, -0.00003246, 0.00002845, -0.00002058, 0.00001471, -0.00001922, 0.00001002, -0.00002187, 0.00003219, -0.00001212, -0.00000418, -0.00001540, 0.00002444, -0.00002328, 0.00002066, -0.00001602, 0.00000782, -0.00002110, 0.00001918, -0.00002431, 0.00002820, -0.00003977, 0.00003804, -0.00004764, 0.00004516, -0.00005146, 0.00004670, -0.00006149, 0.00005202, -0.00009204, 0.00007663, -0.00012730, 0.00011212, -0.00023409, 0.00014670, -0.00097161, 0.00162637, 0.00028232, -0.00068840, 0.00018161, -0.00022444, 0.00011822, -0.00012257, 0.00007731, -0.00009286, 0.00006669, -0.00007705, 0.00005894, -0.00006198, 0.00004768, -0.00004729, 0.00003303, -0.00003524, 0.00003531, -0.00003528, 0.00003165, -0.00003204, 0.00003818, -0.00004524, 0.00003225, -0.00002944, 0.00001824, -0.00000977, 0.00003252, -0.00004329, 0.00002291, -0.00000916, 0.00001714, -0.00003600, 0.00004146, -0.00003150, 0.00002133, -0.00002267, 0.00000763, -0.00000903, 0.00001048, -0.00001407, 0.00001204, -0.00001178, 0.00000797, -0.00000486, 0.00000636, -0.00001443, 0.00001576, -0.00002651, 0.00002720, -0.00002367, 0.00001641, -0.00000521, -0.00001473, 0.00001402, 0.00000265, -0.00000169, -0.00000822, 0.00000513, 0.00000234, -0.00000126, -0.00001109, 0.00001591, -0.00000800, 0.00000467, -0.00000472, 0.00000953, -0.00002121, 0.00002201, -0.00001159, 0.00000952, },
		{ 0.00001101, -0.00000098, 0.00000644, -0.00001837, 0.00001338, -0.00000563, -0.00000624, 0.00001457, -0.00001454, 0.00001794, -0.00000689, -0.00000354, 0.00000849, -0.00001714, 0.00001795, -0.00000772, 0.00001478, -0.00001080, 0.00001798, -0.00002576, 0.00002724, -0.00001961, 0.00001354, -0.00000613, 0.00000164, -0.00000659, 0.00000064, -0.00001397, 0.00001399, 0.00000920, -0.00001055, -0.00001783, 0.00002032, -0.00001442, 0.00001163, -0.00000907, 0.00000861, -0.00002610, 0.00002341, -0.00002981, 0.00003370, -0.00004298, 0.00003936, -0.00004674, 0.00004243, -0.00004845, 0.00004620, -0.00006194, 0.00005805, -0.00009694, 0.00008228, -0.00013173, 0.00012080, -0.00022737, 0.00018160, -0.00078342, 0.00064863, 0.00126988, -0.00089426, 0.00015431, -0.00023027, 0.00010920, -0.00012342, 0.00008032, -0.00009678, 0.00006757, -0.00007692, 0.00005554, -0.00005833, 0.00004302, -0.00004397, 0.00003265, -0.00003865, 0.00003853, -0.00003801, 0.00003485, -0.00003597, 0.00004022, -0.00003893, 0.00002146, -0.00002029, 0.00001237, -0.00001401, 0.00003850, -0.00003466, 0.00000928, -0.00000794, 0.00002542, -0.00003862, 0.00003321, -0.00001875, 0.00000967, -0.00001046, -0.00000332, -0.00000324, 0.00000647, -0.00000954, 0.00000637, -0.00000681, 0.00000380, -0.00000385, 0.00000825, -0.00001435, 0.00001202, -0.00001603, 0.00000701, 0.00000190, -0.00001124, 0.00001928, -0.00002845, 0.00001492, 0.00000284, 0.00000378, -0.00001277, 0.00000591, 0.00000045, 0.00000478, -0.00001657, 0.00001542, -0.00000518, 0.00000498, -0.00000827, 0.00001462, -0.00002270, 0.00001674, -0.00000535, 0.00000716, -0.00001780, },
		{ 0.00001537, -0.00000638, -0.00000091, -0.00001085, 0.00002024, -0.00002732, 0.00002499, -0.00001639, 0.00001144, 0.00000507, -0.00001175, 0.00001215, -0.00001424, 0.00000755, 0.00000334, 0.00000501, -0.00000038, 0.00000800, -0.00001063, 0.00000603, 0.00000565, -0.00001121, 0.00001706, -0.00001942, 0.00001263, -0.00001561, 0.00000343, -0.00001114, 0.00002854, -0.00001354, -0.00001470, 0.00000523, 0.00000362, -0.00000420, 0.00000375, 0.00000084, -0.00001624, 0.00000782, -0.00001206, 0.00001238, -0.00001661, 0.00000883, -0.00001217, 0.00000487, -0.00000919, 0.00000618, -0.00001814, 0.00001218, -0.00003993, 0.00001472, -0.00004863, 0.00002088, -0.00008731, 0.00001757, -0.00035999, -0.00022187, 0.00164058, -0.00047642, -0.00020056, -0.00005285, -0.00003445, -0.00002065, -0.00000572, -0.00002220, -0.00000072, -0.00001578, -0.00000149, -0.00000769, -0.00000310, -0.00000392, -0.00000182, -0.00000852, 0.00000849, -0.00000857, 0.00000654, -0.00000820, 0.00001024, -0.00000383, -0.00001094, 0.00000473, -0.00000646, -0.00000322, 0.00002323, -0.00000660, -0.00001351, -0.00000039, 0.00001942, -0.00002085, 0.00000671, 0.00000744, -0.00001276, 0.00001046, -0.00002007, 0.00000729, -0.00000322, 0.00000114, -0.00000466, 0.00000284, -0.00000508, 0.00000306, 0.00000092, -0.00000227, -0.00000411, 0.00000453, -0.00001624, 0.00002093, -0.00002335, 0.00002169, -0.00001929, -0.00000095, 0.00001234, 0.00000162, -0.00000811, -0.00000183, 0.00000520, 0.00000363, -0.00001258, 0.00000569, 0.00000411, -0.00000015, -0.00000437, 0.00000887, -0.00001083, -0.00000084, 0.00001057, -0.00000517, -0.00000198, -0.00000895, },
		{ -0.00000769, 0.00000518, -0.00001915, 0.00002290, -0.00001980, 0.00000758, 0.00000378, -0.00000912, 0.00002081, -0.00001609, 0.00000907, -0.00000462, -0.00000585, 0.00001318, -0.00000227, 0.00000762, -0.00000051, 0.00000342, -0.00001186, 0.00002372, -0.00002543, 0.00002784, -0.00002620, 0.00001601, -0.00001658, 0.00000649, -0.00001741, 0.00002377, 0.00000170, -0.00001920, -0.00000336, 0.00001048, -0.00000721, 0.00000505, 0.00000028, -0.00000870, -0.00000633, 0.00000338, -0.00000601, 0.00000543, -0.00001549, 0.00001379, -0.00002145, 0.00001697, -0.00002053, 0.00001249, -0.00002170, 0.00000548, -0.00003921, 0.00001731, -0.00005722, 0.00001953, -0.00010842, -0.00006094, -0.00061847, 0.00131297, 0.00031794, -0.00056555, 0.00007182, -0.00014058, 0.00005031, -0.00006896, 0.00003314, -0.00005081, 0.00002766, -0.00004098, 0.00002575, -0.00003287, 0.00002104, -0.00002375, 0.00001208, -0.00001381, 0.00001364, -0.00001515, 0.00001339, -0.00001286, 0.00002027, -0.00002734, 0.00001323, -0.00001162, -0.00000015, 0.00001054, 0.00001424, -0.00002244, -0.00000147, 0.00001157, 0.00000075, -0.00001779, 0.00002670, -0.00002652, 0.00002116, -0.00002506, 0.00000732, -0.00000500, 0.00000500, -0.00000877, 0.00000602, -0.00000846, 0.00000588, -0.00000403, 0.00000653, -0.00001321, 0.00001251, -0.00001968, 0.00001499, -0.00000970, 0.00000183, 0.00000407, -0.00002064, 0.00001944, 0.00000054, -0.00000230, -0.00000918, 0.00000787, 0.00000319, -0.00000748, -0.00000373, 0.00001050, -0.00000289, -0.00000022, 0.00000092, 0.00000266, -0.00001615, 0.00002058, -0.00001138, 0.00000737, -0.00001888, 0.00001912, },
		{ -0.00000261, -0.00000984, 0.00000538, 0.00000506, -0.00001837, 0.00002261, -0.00002058, 0.00002173, -0.00000595, -0.00000344, 0.00000842, -0.00001583, 0.00001559, -0.00000216, 0.00000878, -0.00000249, 0.00000970, -0.00001818, 0.00002608, -0.00002127, 0.00001920, -0.00001278, -0.00000023, -0.00000040, -0.00000652, -0.00000472, 0.00000096, 0.00002448, -0.00002404, -0.00000670, 0.00000622, 0.00000062, -0.00000196, 0.00000491, -0.00000564, -0.00001211, 0.00000643, -0.00000939, 0.00000941, -0.00001855, 0.00001515, -0.00002023, 0.00001387, -0.00001709, 0.00001135, -0.00002175, 0.00001234, -0.00004641, 0.00002605, -0.00006496, 0.00003510, -0.00011894, 0.00002608, -0.00059196, 0.00063996, 0.00098990, -0.00058422, -0.00003572, -0.00010521, 0.00001496, -0.00005023, 0.00001828, -0.00003866, 0.00001556, -0.00003047, 0.00001422, -0.00002240, 0.00001016, -0.00001444, 0.00000503, -0.00001023, 0.00001071, -0.00001250, 0.00001097, -0.00001123, 0.00001714, -0.00001625, -0.00000082, -0.00000043, -0.00000751, 0.00000659, 0.00002017, -0.00001480, -0.00000955, 0.00000400, 0.00001694, -0.00002847, 0.00002867, -0.00002278, 0.00001413, -0.00001316, -0.00000581, 0.00000289, 0.00000007, -0.00000408, 0.00000109, -0.00000437, 0.00000159, -0.00000156, 0.00000497, -0.00000807, 0.00000331, -0.00000467, -0.00000606, 0.00001170, -0.00001693, 0.00001734, -0.00002406, 0.00001107, 0.00000875, -0.00000272, -0.00000875, 0.00000254, 0.00000800, -0.00000611, -0.00000737, 0.00000951, 0.00000002, 0.00000023, -0.00000294, 0.00000894, -0.00001944, 0.00001618, -0.00000417, 0.00000302, -0.00001252, 0.00000650, 0.00000480, },
		{ -0.00000058, -0.00000966, 0.00001975, -0.00002488, 0.00001713, -0.00000796, 0.00000212, 0.00001597, -0.00001837, 0.00001748, -0.00001757, 0.00001002, 0.00000287, 0.00000650, -0.00000108, 0.00001065, -0.00001561, 0.00001718, -0.00000584, 0.00000087, 0.00000816, -0.00002019, 0.00001530, -0.00001766, 0.00000673, -0.00001572, 0.00003180, -0.00001381, -0.00001520, 0.00000278, 0.00000494, -0.00000329, 0.00000343, 0.00000052, -0.00001526, 0.00000384, -0.00000621, 0.00000555, -0.00001243, 0.00000622, -0.00000861, 0.00000123, -0.00000493, 0.00000093, -0.00001113, 0.00000460, -0.00003319, 0.00000857, -0.00004039, 0.00000880, -0.00007713, 0.00000742, -0.00040091, 0.00004678, 0.00120065, -0.00018765, -0.00029598, -0.00001220, -0.00006110, -0.00000225, -0.00002329, -0.00000317, -0.00001615, -0.00000337, -0.00001114, -0.00000056, -0.00001000, 0.00000227, -0.00000876, 0.00000064, -0.00000020, -0.00000171, 0.00000008, -0.00000055, 0.00000443, 0.00000135, -0.00001496, 0.00000663, -0.00000925, 0.00000030, 0.00002171, -0.00000468, -0.00001180, -0.00000796, 0.00002662, -0.00002549, 0.00001715, -0.00000763, -0.00000264, 0.00000518, -0.00002046, 0.00000979, -0.00000450, 0.00000104, -0.00000419, 0.00000033, -0.00000388, 0.00000314, -0.00000105, 0.00000223, -0.00000904, 0.00000910, -0.00001824, 0.00001654, -0.00001429, 0.00000769, -0.00000752, -0.00000920, 0.00002051, -0.00000518, -0.00000419, -0.00000584, 0.00001265, -0.00000430, -0.00000875, 0.00000549, 0.00000395, 0.00000013, -0.00000398, 0.00000875, -0.00001329, 0.00000314, 0.00000848, -0.00000595, -0.00000114, -0.00000841, 0.00001592, -0.00000893, },
		{ -0.00000875, 0.00001261, -0.00000714, -0.00000793, 0.00001599, -0.00002000, 0.00003050, -0.00002043, 0.00001342, -0.00000840, -0.00000114, 0.00000862, 0.00000460, 0.00000057, 0.00000940, -0.00000883, 0.00000461, 0.00001018, -0.00001387, 0.00002119, -0.00002729, 0.00001501, -0.00001316, 0.00000305, -0.00001287, 0.00001636, 0.00001091, -0.00002744, 0.00000314, 0.00000087, 0.00000433, -0.00000510, 0.00000966, -0.00001773, 0.00000145, -0.00000545, 0.00000496, -0.00001028, 0.00000238, -0.00000416, -0.00000185, -0.00000381, 0.00000157, -0.00001122, 0.00000574, -0.00002735, -0.00000074, -0.00002624, -0.00000556, -0.00005049, -0.00001293, -0.00023483, -0.00027326, 0.00097853, 0.00036485, -0.00044477, -0.00002796, -0.00006935, -0.00000596, -0.00002631, -0.00000094, -0.00001774, -0.00000347, -0.00001166, -0.00000156, -0.00000916, -0.00000034, -0.00000557, -0.00000304, 0.00000176, -0.00000284, 0.00000040, -0.00000075, 0.00000305, 0.00000421, -0.00001082, -0.00000349, 0.00000246, -0.00001312, 0.00002517, -0.00000058, -0.00000410, -0.00002160, 0.00002728, -0.00001251, 0.00000076, 0.00000825, -0.00001683, 0.00001655, -0.00002431, 0.00000620, -0.00000140, 0.00000007, -0.00000379, 0.00000011, -0.00000494, 0.00000373, -0.00000312, 0.00000619, -0.00001077, 0.00000754, -0.00001067, 0.00000102, 0.00000483, -0.00001190, 0.00001063, -0.00002231, 0.00002080, 0.00000061, -0.00000436, -0.00000811, 0.00000923, 0.00000357, -0.00001267, 0.00000389, 0.00000387, 0.00000304, -0.00000533, 0.00000681, -0.00000558, -0.00000759, 0.00001461, -0.00000748, 0.00000218, -0.00001191, 0.00001316, -0.00000249, -0.00000347, },
		{ -0.00000229, 0.00001320, -0.00002552, 0.00002392, -0.00002086, 0.00002092, -0.00000200, -0.00000403, 0.00000761, -0.00001267, 0.00001192, 0.00000376, 0.00000297, 0.00000578, 0.00000046, -0.00000748, 0.00002100, -0.00001895, 0.00002115, -0.00001896, 0.00000032, 0.00000194, -0.00000921, -0.00000010, -0.00000571, 0.00003066, -0.00002868, -0.00000166, -0.00000286, 0.00001006, -0.00000919, 0.00001192, -0.00001352, -0.00000340, -0.00000531, 0.00000562, -0.00000988, 0.00000139, -0.00000433, 0.00000086, -0.00000826, 0.00000661, -0.00001481, 0.00000898, -0.00002391, -0.00000485, -0.00002082, -0.00000872, -0.00003863, -0.00002373, -0.00012356, -0.00037627, 0.00058728, 0.00073944, -0.00029912, -0.00019454, -0.00000676, -0.00005854, 0.00000568, -0.00002922, 0.00000705, -0.00002573, 0.00000676, -0.00001884, 0.00000582, -0.00001523, 0.00000781, -0.00001534, 0.00001155, -0.00001152, 0.00000814, -0.00000858, 0.00000994, -0.00000312, 0.00000306, -0.00001830, 0.00001362, -0.00002110, 0.00002219, 0.00000422, 0.00000292, -0.00002491, 0.00001363, 0.00000754, -0.00001440, 0.00001869, -0.00002245, 0.00001614, -0.00001584, -0.00000592, 0.00000612, -0.00000400, -0.00000038, -0.00000276, -0.00000304, 0.00000062, -0.00000042, 0.00000256, -0.00000254, -0.00000429, 0.00000502, -0.00001621, 0.00001731, -0.00001775, 0.00001011, -0.00001477, 0.00000409, 0.00001548, -0.00000983, -0.00000350, -0.00000108, 0.00001435, -0.00001628, 0.00000290, 0.00000186, 0.00000636, -0.00000545, 0.00000396, 0.00000032, -0.00001183, 0.00001157, -0.00000070, -0.00000255, -0.00000564, 0.00000119, 0.00000995, -0.00001125, -0.00000081, },
		{ 0.00001503, -0.00001755, 0.00000536, 0.00000133, -0.00000610, 0.00002398, -0.00002053, 0.00001829, -0.00001676, 0.00000996, 0.00000381, 0.00000673, 0.00000011, 0.00001030, -0.00001618, 0.00002451, -0.00001475, 0.00001202, -0.00000374, -0.00001574, 0.00001257, -0.00001471, 0.00000533, -0.00001461, 0.00002883, -0.00001057, -0.00001593, 0.00000027, 0.00000551, -0.00000149, 0.00000257, -0.00000056, -0.00001255, -0.00000159, 0.00000096, -0.00000321, -0.00000587, 0.00000155, -0.00000346, -0.00000356, 0.00000030, -0.00000564, -0.00000196, -0.00000679, -0.00002079, -0.00000425, -0.00002443, -0.00001328, -0.00005206, -0.00003185, -0.00038514, 0.00025194, 0.00080265, 0.00006636, -0.00035979, 0.00001014, -0.00007757, 0.00000923, -0.00003593, 0.00001212, -0.00002872, 0.00000694, -0.00001901, 0.00000480, -0.00001504, 0.00000664, -0.00001391, 0.00000862, -0.00000847, 0.00000504, -0.00000637, 0.00000731, -0.00000154, 0.00000554, -0.00001688, 0.00000645, -0.00001070, 0.00000456, 0.00001735, -0.00000057, -0.00001155, -0.00001183, 0.00002748, -0.00002303, 0.00002157, -0.00001964, 0.00000735, -0.00000249, -0.00001701, 0.00000920, -0.00000399, -0.00000015, -0.00000287, -0.00000287, -0.00000166, 0.00000208, -0.00000176, 0.00000507, -0.00001162, 0.00001070, -0.00001660, 0.00000858, -0.00000252, -0.00000802, 0.00000459, -0.00001592, 0.00002586, -0.00001021, -0.00000108, -0.00000817, 0.00001789, -0.00001159, -0.00000357, 0.00000379, 0.00000461, -0.00000056, -0.00000179, 0.00000591, -0.00001274, 0.00000566, 0.00000578, -0.00000556, -0.00000129, -0.00000603, 0.00001376, -0.00000956, -0.00000154, -0.00000590, },
		{ 0.00000516, -0.00002051, 0.00002232, -0.00002437, 0.00003339, -0.00001826, 0.00001285, -0.00000863, 0.00000182, 0.00000537, 0.00000959, -0.00000365, 0.00001566, -0.00001708, 0.00001886, -0.00000304, -0.00000071, 0.00000978, -0.00002343, 0.00001084, -0.00000822, -0.00000084, -0.00000805, 0.00001009, 0.00001525, -0.00002864, 0.00000369, -0.00000366, 0.00001038, -0.00000902, 0.00001166, -0.00001912, 0.00000227, -0.00000718, 0.00000797, -0.00001754, 0.00001198, -0.00001382, 0.00000906, -0.00001502, 0.00001262, -0.00002200, 0.00001783, -0.00004259, 0.00001855, -0.00004827, 0.00002035, -0.00008751, 0.00003591, -0.00034571, 0.00002001, 0.00066449, 0.00040449, -0.00032755, -0.00011046, -0.00001736, -0.00004534, 0.00000145, -0.00002186, 0.00000327, -0.00002337, 0.00000750, -0.00001974, 0.00000630, -0.00001401, 0.00000533, -0.00001077, 0.00000986, -0.00001226, 0.00000935, -0.00000862, 0.00001328, -0.00000733, 0.00000156, -0.00001522, 0.00001044, -0.00001826, 0.00003128, -0.00000852, 0.00000777, -0.00003330, 0.00003409, -0.00001880, 0.00001527, -0.00000975, -0.00000510, 0.00000858, -0.00002051, 0.00000414, 0.00000155, -0.00000377, 0.00000017, -0.00000491, -0.00000176, 0.00000166, -0.00000205, 0.00000533, -0.00000756, 0.00000199, -0.00000155, -0.00001179, 0.00001676, -0.00002215, 0.00001314, -0.00001945, 0.00001816, 0.00000246, -0.00000722, -0.00000549, 0.00000944, 0.00000293, -0.00001537, 0.00000963, -0.00000173, 0.00000766, -0.00000855, 0.00001066, -0.00001253, 0.00000169, 0.00000643, -0.00000166, -0.00000430, -0.00000323, 0.00000557, 0.00000291, -0.00001180, 0.00000230, 0.00000135, },
		{ -0.00002644, 0.00001821, -0.00001584, 0.00001609, 0.00000442, -0.00000503, 0.00000621, -0.00000812, 0.00000761, 0.00000911, -0.00000142, 0.00001245, -0.00000782, 0.00000457, 0.00001300, -0.00001264, 0.00001779, -0.00002207, 0.00000113, 0.00000192, -0.00000842, 0.00000041, -0.00000634, 0.00002810, -0.00002477, -0.00000280, -0.00000639, 0.00001331, -0.00000975, 0.00001182, -0.00001478, -0.00000055, -0.00001063, 0.00001321, -0.00002148, 0.00001388, -0.00001609, 0.00001349, -0.00002080, 0.00001928, -0.00002841, 0.00002550, -0.00004487, 0.00002090, -0.00005031, 0.00002710, -0.00008752, 0.00004059, -0.00024230, -0.00016336, 0.00050083, 0.00053424, -0.00006526, -0.00030117, 0.00005084, -0.00010096, 0.00003634, -0.00005530, 0.00003371, -0.00004988, 0.00002883, -0.00003842, 0.00002183, -0.00002924, 0.00001934, -0.00002518, 0.00002316, -0.00002437, 0.00002011, -0.00001997, 0.00002357, -0.00001667, 0.00001539, -0.00002781, 0.00001908, -0.00002554, 0.00002948, -0.00000570, 0.00001485, -0.00003358, 0.00001835, -0.00000028, 0.00000092, 0.00000508, -0.00001775, 0.00001453, -0.00001705, -0.00000469, 0.00000682, -0.00000578, 0.00000147, -0.00000512, -0.00000261, 0.00000046, -0.00000013, 0.00000114, 0.00000156, -0.00000962, 0.00001150, -0.00002204, 0.00001819, -0.00001451, -0.00000018, -0.00000328, -0.00000354, 0.00002066, -0.00001525, 0.00000106, -0.00000317, 0.00001726, -0.00002274, 0.00001104, -0.00000467, 0.00001132, -0.00000963, 0.00000989, -0.00000850, -0.00000236, 0.00000476, 0.00000369, -0.00000804, 0.00000118, -0.00000335, 0.00001278, -0.00001739, 0.00000645, -0.00000757, 0.00001705, },
		{ -0.00000744, 0.00000874, -0.00001120, 0.00002771, -0.00001745, 0.00001429, -0.00001166, 0.00000724, 0.00000629, 0.00000562, 0.00000323, 0.00000680, -0.00001154, 0.00002620, -0.00001834, 0.00001844, -0.00001479, -0.00000877, 0.00000619, -0.00000802, 0.00000018, -0.00000871, 0.00002055, -0.00000438, -0.00001694, -0.00000192, 0.00000514, 0.00000119, 0.00000106, -0.00000182, -0.00000925, -0.00000671, 0.00000717, -0.00001137, 0.00000076, -0.00000317, 0.00000136, -0.00000757, 0.00000434, -0.00001088, 0.00000601, -0.00001792, -0.00000729, -0.00001870, -0.00000671, -0.00003806, -0.00002097, -0.00009047, -0.00031511, 0.00037509, 0.00048217, 0.00025483, -0.00037582, 0.00000625, -0.00007868, 0.00000966, -0.00003966, 0.00001950, -0.00003453, 0.00001186, -0.00002244, 0.00000627, -0.00001639, 0.00000743, -0.00001534, 0.00001306, -0.00001380, 0.00000925, -0.00001050, 0.00001316, -0.00000583, 0.00000742, -0.00001586, 0.00000351, -0.00000978, 0.00000808, 0.00001167, 0.00000464, -0.00001234, -0.00001196, 0.00002209, -0.00001346, 0.00001871, -0.00002583, 0.00001409, -0.00000996, -0.00001139, 0.00000643, -0.00000233, -0.00000183, -0.00000140, -0.00000585, 0.00000066, 0.00000047, -0.00000141, 0.00000615, -0.00001179, 0.00000962, -0.00001241, -0.00000071, 0.00000864, -0.00002163, 0.00001390, -0.00001890, 0.00002665, -0.00001200, 0.00000013, -0.00000771, 0.00001935, -0.00001624, 0.00000119, 0.00000169, 0.00000534, -0.00000154, 0.00000145, 0.00000106, -0.00000943, 0.00000619, 0.00000329, -0.00000456, -0.00000197, -0.00000240, 0.00000942, -0.00000860, -0.00000327, -0.00000069, 0.00000775, -0.00000764, },
		{ 0.00002108, -0.00002164, 0.00002939, -0.00000979, 0.00000758, -0.00000516, 0.00000250, 0.00000419, 0.00001182, -0.00000398, 0.00001688, -0.00001913, 0.00002788, -0.00001250, 0.00001034, -0.00000350, -0.00001594, 0.00000397, -0.00000196, -0.00000518, -0.00000299, 0.00000483, 0.00001550, -0.00002416, -0.00000044, -0.00000395, 0.00001166, -0.00000776, 0.00000783, -0.00001477, -0.00000180, -0.00000386, 0.00000473, -0.00001743, 0.00001394, -0.00001571, 0.00001181, -0.00001740, 0.00001375, -0.00002135, 0.00001616, -0.00004159, 0.00001973, -0.00004972, 0.00002105, -0.00009110, 0.00003462, -0.00037965, 0.00023142, 0.00040345, 0.00042132, -0.00021421, -0.00017200, 0.00001314, -0.00006677, 0.00001426, -0.00003007, 0.00001250, -0.00003189, 0.00001603, -0.00002824, 0.00001309, -0.00001983, 0.00000919, -0.00001172, 0.00001102, -0.00001484, 0.00001171, -0.00000999, 0.00001712, -0.00001325, 0.00000857, -0.00002101, 0.00001194, -0.00001559, 0.00002859, -0.00000866, 0.00001170, -0.00003504, 0.00003073, -0.00001645, 0.00002275, -0.00002345, 0.00000546, -0.00000070, -0.00001463, 0.00000150, 0.00000352, -0.00000611, 0.00000255, -0.00000836, 0.00000053, -0.00000009, -0.00000089, 0.00000412, -0.00000420, -0.00000297, 0.00000575, -0.00002096, 0.00002356, -0.00002716, 0.00001128, -0.00001272, 0.00001236, 0.00000523, -0.00000993, -0.00000201, 0.00000870, 0.00000121, -0.00001496, 0.00001222, -0.00000482, 0.00000957, -0.00000854, 0.00001093, -0.00001601, 0.00000895, -0.00000130, 0.00000381, -0.00000993, 0.00000490, -0.00000156, 0.00000658, -0.00001749, 0.00001114, -0.00000774, 0.00001399, -0.00002744, },
		{ -0.00000736, 0.00000874, 0.00001224, -0.00000603, 0.00000457, -0.00000294, 0.00000359, 0.00001295, -0.00000331, 0.00001613, -0.00001312, 0.00001568, 0.00000417, -0.00000396, 0.00000934, -0.00002080, 0.00000041, 0.00000130, -0.00000592, -0.00000144, -0.00000295, 0.00001938, -0.00001488, -0.00000821, -0.00000571, 0.00001165, -0.00000525, 0.00000644, -0.00001112, -0.00000226, -0.00001036, 0.00001361, -0.00002494, 0.00001822, -0.00001954, 0.00001720, -0.00002351, 0.00002085, -0.00002932, 0.00002726, -0.00004924, 0.00002874, -0.00005964, 0.00003854, -0.00010558, 0.00006403, -0.00031160, 0.00002111, 0.00039682, 0.00037862, 0.00010439, -0.00034726, 0.00006407, -0.00011069, 0.00004103, -0.00005940, 0.00004036, -0.00005560, 0.00003441, -0.00004337, 0.00002482, -0.00003140, 0.00001997, -0.00002441, 0.00002415, -0.00002699, 0.00002233, -0.00002190, 0.00002809, -0.00002187, 0.00001942, -0.00002862, 0.00001581, -0.00002069, 0.00002767, -0.00000816, 0.00001929, -0.00003410, 0.00001682, -0.00000446, 0.00001438, -0.00000923, -0.00001019, 0.00000943, -0.00001603, -0.00000309, 0.00000572, -0.00000574, 0.00000175, -0.00000618, -0.00000272, 0.00000058, -0.00000022, 0.00000052, 0.00000396, -0.00001214, 0.00001410, -0.00002315, 0.00001477, -0.00000884, -0.00001075, 0.00000741, -0.00000938, 0.00002229, -0.00001723, 0.00000355, -0.00000276, 0.00001565, -0.00002375, 0.00001483, -0.00000789, 0.00001294, -0.00001041, 0.00001262, -0.00001479, 0.00000605, -0.00000186, 0.00000723, -0.00001186, 0.00000636, -0.00000582, 0.00001218, -0.00001960, 0.00001030, -0.00000917, 0.00001597, -0.00002142, 0.00000689, },
		{ -0.00001227, 0.00002704, -0.00000978, 0.00000663, -0.00000338, 0.00000272, 0.00000969, 0.00000382, 0.00000748, 0.00000123, -0.00000250, 0.00002231, -0.00001582, 0.00001786, -0.00002164, -0.00000244, -0.00000107, -0.00000010, -0.00000641, -0.00000044, 0.00000939, 0.00000308, -0.00001764, -0.00000387, 0.00000401, 0.00000433, -0.00000088, -0.00000294, -0.00000628, -0.00001011, 0.00001056, -0.00001676, 0.00000543, -0.00000622, 0.00000429, -0.00000933, 0.00000564, -0.00001278, 0.00001000, -0.00002510, 0.00000283, -0.00002957, 0.00000639, -0.00005735, 0.00000488, -0.00015253, -0.00020573, 0.00041601, 0.00025505, 0.00036382, -0.00034016, -0.00002353, -0.00006481, -0.00000090, -0.00003392, 0.00001778, -0.00003219, 0.00001016, -0.00002051, 0.00000324, -0.00001359, 0.00000426, -0.00001245, 0.00001287, -0.00001481, 0.00000955, -0.00001090, 0.00001560, -0.00000733, 0.00000648, -0.00001192, -0.00000199, -0.00000620, 0.00000970, 0.00000633, 0.00000939, -0.00001301, -0.00000962, 0.00001243, 0.00000093, 0.00000971, -0.00002547, 0.00001572, -0.00001507, -0.00000545, 0.00000274, -0.00000043, -0.00000329, -0.00000046, -0.00000781, 0.00000222, -0.00000097, -0.00000047, 0.00000581, -0.00001002, 0.00000667, -0.00000717, -0.00000899, 0.00001622, -0.00003009, 0.00001819, -0.00001700, 0.00002225, -0.00001007, -0.00000083, -0.00000435, 0.00001648, -0.00001702, 0.00000407, 0.00000036, 0.00000524, -0.00000187, 0.00000448, -0.00000433, -0.00000446, 0.00000498, 0.00000140, -0.00000341, -0.00000265, 0.00000163, 0.00000379, -0.00000663, -0.00000510, 0.00000461, 0.00000000, 0.00000222, -0.00001843, 0.00001217, },
		{ 0.00002032, 0.00000301, -0.00000072, 0.00000100, 0.00000132, 0.00000499, 0.00001137, -0.00000061, 0.00001320, -0.00001449, 0.00003024, -0.00001599, 0.00001599, -0.00001515, -0.00000735, -0.00000391, 0.00000452, -0.00000925, 0.00000185, 0.00000045, 0.00001299, -0.00001622, -0.00000699, -0.00000188, 0.00000996, -0.00000351, 0.00000093, -0.00000774, -0.00000761, 0.00000131, -0.00000165, -0.00001268, 0.00001066, -0.00001220, 0.00000891, -0.00001382, 0.00000850, -0.00001357, 0.00000574, -0.00002976, 0.00000852, -0.00003696, 0.00000394, -0.00007000, -0.00000734, -0.00034008, 0.00034954, 0.00021372, 0.00040702, -0.00010610, -0.00021128, 0.00002324, -0.00007241, 0.00001490, -0.00002860, 0.00001270, -0.00003135, 0.00001582, -0.00002868, 0.00001278, -0.00001942, 0.00000790, -0.00000810, 0.00000743, -0.00001250, 0.00000921, -0.00000635, 0.00001557, -0.00001402, 0.00001052, -0.00002153, 0.00000845, -0.00000729, 0.00001929, -0.00000268, 0.00000895, -0.00002824, 0.00001890, -0.00000651, 0.00002240, -0.00003046, 0.00001212, -0.00000899, -0.00000812, -0.00000110, 0.00000423, -0.00000677, 0.00000311, -0.00000986, 0.00000154, -0.00000125, 0.00000017, 0.00000286, -0.00000127, -0.00000639, 0.00000992, -0.00002496, 0.00002398, -0.00002642, 0.00000573, -0.00000360, 0.00000483, 0.00000780, -0.00001148, 0.00000150, 0.00000720, -0.00000118, -0.00001182, 0.00001155, -0.00000509, 0.00000854, -0.00000526, 0.00000741, -0.00001513, 0.00001238, -0.00000647, 0.00000709, -0.00001294, 0.00001062, -0.00000658, 0.00000725, -0.00001910, 0.00001579, -0.00001254, 0.00001752, -0.00002767, 0.00001270, -0.00001333, },
		{ 0.00002001, -0.00000645, 0.00000299, 0.00000209, 0.00000068, 0.00001470, -0.00000220, 0.00001610, -0.00001399, 0.00002370, -0.00000305, 0.00000429, -0.00000212, -0.00001558, -0.00000245, 0.00000133, -0.00000339, -0.00000363, 0.00000169, 0.00000819, -0.00000284, -0.00001479, -0.00000324, 0.00000754, 0.00000149, -0.00000121, -0.00000534, -0.00000604, -0.00000678, 0.00000882, -0.00002188, 0.00001585, -0.00001628, 0.00001389, -0.00001871, 0.00001418, -0.00002081, 0.00001761, -0.00004014, 0.00002151, -0.00005159, 0.00002832, -0.00009511, 0.00004668, -0.00032347, 0.00015822, 0.00029277, 0.00026799, 0.00020938, -0.00033664, 0.00003942, -0.00009401, 0.00002548, -0.00004657, 0.00003114, -0.00004633, 0.00002647, -0.00003641, 0.00001748, -0.00002439, 0.00001247, -0.00001578, 0.00001693, -0.00002134, 0.00001652, -0.00001579, 0.00002447, -0.00001950, 0.00001624, -0.00002247, 0.00000620, -0.00000909, 0.00001870, -0.00000419, 0.00001689, -0.00002720, 0.00000947, -0.00000413, 0.00002343, -0.00002101, -0.00000220, 0.00000207, -0.00001285, -0.00000179, 0.00000358, -0.00000454, 0.00000097, -0.00000628, -0.00000304, 0.00000055, -0.00000028, 0.00000044, 0.00000480, -0.00001194, 0.00001297, -0.00002006, 0.00000823, -0.00000268, -0.00001903, 0.00001490, -0.00001181, 0.00001941, -0.00001505, 0.00000356, -0.00000002, 0.00000999, -0.00001933, 0.00001369, -0.00000710, 0.00001063, -0.00000714, 0.00001112, -0.00001676, 0.00001112, -0.00000628, 0.00000852, -0.00001288, 0.00000891, -0.00000570, 0.00000800, -0.00001755, 0.00001020, -0.00000697, 0.00001070, -0.00001226, -0.00000725, 0.00000175, 0.00000093, },
		{ 0.00000066, -0.00000242, 0.00000596, -0.00000199, 0.00001285, 0.00000242, 0.00001145, -0.00000441, 0.00000876, 0.00001474, -0.00000871, 0.00001038, -0.00002264, 0.00000083, -0.00000661, 0.00000639, -0.00001175, 0.00000718, -0.00000154, 0.00000960, -0.00001708, -0.00000594, 0.00000263, 0.00000724, -0.00000277, -0.00000384, -0.00000414, -0.00001116, 0.00001028, -0.00001826, 0.00000702, -0.00000674, 0.00000469, -0.00000838, 0.00000381, -0.00001074, 0.00000882, -0.00002628, 0.00000672, -0.00003336, 0.00001045, -0.00006500, 0.00001570, -0.00020083, -0.00007962, 0.00038980, 0.00011820, 0.00039646, -0.00026117, -0.00007106, -0.00004159, -0.00001847, -0.00002155, 0.00000890, -0.00002311, 0.00000298, -0.00001428, -0.00000318, -0.00000777, -0.00000175, -0.00000619, 0.00000861, -0.00001171, 0.00000601, -0.00000751, 0.00001440, -0.00000602, 0.00000322, -0.00000609, -0.00000887, -0.00000060, 0.00000904, 0.00000252, 0.00001236, -0.00001223, -0.00000675, 0.00000162, 0.00001631, -0.00000236, -0.00001987, 0.00001207, -0.00001663, -0.00000069, -0.00000069, 0.00000093, -0.00000400, -0.00000042, -0.00000832, 0.00000249, -0.00000170, 0.00000055, 0.00000461, -0.00000710, 0.00000286, -0.00000223, -0.00001456, 0.00001842, -0.00003187, 0.00001678, -0.00001051, 0.00001334, -0.00000506, -0.00000322, 0.00000095, 0.00001010, -0.00001394, 0.00000444, 0.00000046, 0.00000378, -0.00000082, 0.00000615, -0.00000865, 0.00000062, 0.00000279, 0.00000011, -0.00000240, -0.00000288, 0.00000520, -0.00000212, -0.00000433, -0.00000637, 0.00000877, -0.00000634, 0.00000973, -0.00002547, 0.00001069, -0.00000969, 0.00002309, },
		{ -0.00000928, 0.00000790, -0.00000052, 0.00000692, 0.00000927, 0.00000509, 0.00000631, -0.00000451, 0.00002659, -0.00001339, 0.00001459, -0.00002227, -0.00000042, -0.00001062, 0.00000957, -0.00001192, 0.00000562, -0.00000292, 0.00000902, -0.00000725, -0.00001351, 0.00000050, 0.00000729, 0.00000130, -0.00000609, -0.00000116, -0.00001213, 0.00000520, -0.00000787, -0.00000649, 0.00000518, -0.00000636, 0.00000361, -0.00000794, 0.00000120, -0.00000396, -0.00000709, -0.00001429, -0.00000714, -0.00001913, -0.00001990, -0.00003857, -0.00007105, -0.00024524, 0.00038124, 0.00009782, 0.00036212, -0.00000679, -0.00022830, 0.00001621, -0.00006632, 0.00000756, -0.00002161, 0.00000774, -0.00002527, 0.00001009, -0.00002393, 0.00000805, -0.00001536, 0.00000407, -0.00000270, 0.00000190, -0.00000789, 0.00000442, -0.00000014, 0.00001063, -0.00001117, 0.00000876, -0.00001851, 0.00000254, 0.00000318, 0.00000715, 0.00000602, 0.00000271, -0.00001654, 0.00000263, 0.00000765, 0.00001580, -0.00003035, 0.00001337, -0.00001451, -0.00000234, -0.00000308, 0.00000362, -0.00000593, 0.00000208, -0.00000947, 0.00000121, -0.00000172, 0.00000100, 0.00000178, 0.00000084, -0.00000787, 0.00001054, -0.00002359, 0.00001827, -0.00002079, -0.00000191, 0.00000596, -0.00000280, 0.00000907, -0.00001102, 0.00000428, 0.00000516, -0.00000354, -0.00000704, 0.00000851, -0.00000325, 0.00000547, 0.00000014, 0.00000126, -0.00001048, 0.00001163, -0.00000814, 0.00000729, -0.00001251, 0.00001297, -0.00000879, 0.00000473, -0.00001649, 0.00001572, -0.00001246, 0.00001558, -0.00002210, -0.00000043, -0.00000265, 0.00000935, 0.00001704, },
		{ 0.00000168, 0.00000636, -0.00000064, 0.00001433, 0.00000162, 0.00001266, -0.00001005, 0.00002740, -0.00000679, 0.00000917, -0.00001345, -0.00000837, -0.00000677, 0.00000225, -0.00000179, -0.00000467, 0.00000515, -0.00000193, 0.00000757, -0.00001950, -0.00000148, 0.00000350, 0.00000758, -0.00000804, -0.00000033, -0.00000925, -0.00000261, 0.00000172, -0.00001506, 0.00000950, -0.00000927, 0.00000685, -0.00001023, 0.00000380, -0.00000811, 0.00000242, -0.00002371, 0.00000557, -0.00003316, 0.00000451, -0.00006552, -0.00000159, -0.00028191, 0.00023926, 0.00020306, 0.00019274, 0.00025854, -0.00028224, -0.00000965, -0.00006232, -0.00000062, -0.00002521, 0.00001340, -0.00002859, 0.00001082, -0.00002276, 0.00000466, -0.00001281, 0.00000133, -0.00000375, 0.00000569, -0.00001127, 0.00000631, -0.00000499, 0.00001565, -0.00001224, 0.00000868, -0.00001274, -0.00000573, 0.00000498, 0.00000638, 0.00000318, 0.00001030, -0.00001553, -0.00000158, 0.00000009, 0.00002675, -0.00002773, 0.00000368, -0.00000571, -0.00000828, -0.00000107, 0.00000108, -0.00000287, -0.00000033, -0.00000575, -0.00000337, 0.00000018, -0.00000008, 0.00000061, 0.00000441, -0.00000961, 0.00000899, -0.00001399, 0.00000015, 0.00000224, -0.00002332, 0.00001794, -0.00001046, 0.00001231, -0.00000918, 0.00000167, 0.00000389, 0.00000206, -0.00001120, 0.00000875, -0.00000317, 0.00000531, -0.00000073, 0.00000591, -0.00001420, 0.00001188, -0.00000746, 0.00000701, -0.00001086, 0.00000876, -0.00000343, 0.00000121, -0.00001215, 0.00000684, -0.00000206, 0.00000308, -0.00000174, -0.00002073, 0.00000847, -0.00000526, 0.00002631, -0.00000563, },
		{ 0.00001402, -0.00000513, 0.00001456, 0.00000253, 0.00001372, -0.00000791, 0.00001943, 0.00000645, -0.00000030, -0.00000158, -0.00001828, 0.00000025, -0.00000891, 0.00000959, -0.00001377, 0.00001163, -0.00000951, 0.00001326, -0.00001451, -0.00000848, 0.00000154, 0.00000920, -0.00000406, -0.00000466, -0.00000287, -0.00001001, 0.00000661, -0.00001595, 0.00000556, -0.00000492, 0.00000296, -0.00000540, -0.00000025, -0.00000573, 0.00000325, -0.00002172, 0.00000408, -0.00002948, 0.00000449, -0.00005909, 0.00000652, -0.00022188, 0.00003922, 0.00032073, 0.00005438, 0.00036904, -0.00015608, -0.00012310, -0.00001773, -0.00003671, -0.00000755, -0.00000306, -0.00001077, -0.00000673, -0.00000630, -0.00001063, -0.00000116, -0.00000842, 0.00000131, 0.00000211, -0.00000607, 0.00000008, -0.00000153, 0.00001050, -0.00000284, -0.00000106, -0.00000007, -0.00001529, 0.00000554, 0.00000664, 0.00000061, 0.00001288, -0.00000918, -0.00000509, -0.00000721, 0.00002859, -0.00001358, -0.00001179, 0.00000450, -0.00001469, 0.00000216, -0.00000317, 0.00000131, -0.00000380, -0.00000122, -0.00000745, 0.00000144, -0.00000157, 0.00000130, 0.00000310, -0.00000379, -0.00000091, 0.00000149, -0.00001671, 0.00001502, -0.00002727, 0.00001070, -0.00000112, 0.00000182, 0.00000144, -0.00000563, 0.00000641, 0.00000207, -0.00000820, 0.00000271, 0.00000184, 0.00000106, 0.00000170, 0.00000587, -0.00001072, 0.00000438, 0.00000060, -0.00000079, -0.00000161, -0.00000237, 0.00000758, -0.00000734, -0.00000236, -0.00000663, 0.00001093, -0.00000996, 0.00001343, -0.00002758, 0.00000467, -0.00000500, 0.00001777, 0.00001146, -0.00001021, },
		{ -0.00000149, 0.00000872, 0.00000707, 0.00001106, -0.00000110, 0.00000801, 0.00001931, -0.00000676, 0.00000674, -0.00002357, 0.00000291, -0.00001435, 0.00001179, -0.00001213, 0.00000746, -0.00000501, 0.00000465, 0.00000077, -0.00001806, 0.00000159, 0.00000524, 0.00000474, -0.00001098, 0.00000277, -0.00001341, 0.00000579, -0.00001152, -0.00000146, 0.00000011, -0.00000087, -0.00000128, -0.00000282, -0.00000453, 0.00000308, -0.00001691, -0.00000159, -0.00002001, -0.00000474, -0.00003969, -0.00001184, -0.00013412, -0.00012236, 0.00034726, 0.00004478, 0.00029498, 0.00007815, -0.00022386, -0.00000359, -0.00005314, -0.00000346, -0.00001303, 0.00000130, -0.00001721, 0.00000217, -0.00001703, 0.00000167, -0.00001015, 0.00000009, 0.00000200, -0.00000306, -0.00000348, -0.00000015, 0.00000609, 0.00000463, -0.00000663, 0.00000510, -0.00001396, -0.00000314, 0.00001232, -0.00000388, 0.00001361, -0.00000314, -0.00000445, -0.00001309, 0.00002143, 0.00000617, -0.00002458, 0.00000930, -0.00001643, 0.00000178, -0.00000412, 0.00000192, -0.00000407, 0.00000004, -0.00000768, -0.00000019, -0.00000158, 0.00000151, 0.00000102, 0.00000201, -0.00000754, 0.00000815, -0.00001792, 0.00000809, -0.00001220, -0.00000955, 0.00001396, -0.00000905, 0.00000835, -0.00000810, 0.00000573, 0.00000287, -0.00000520, -0.00000198, 0.00000448, -0.00000063, 0.00000190, 0.00000580, -0.00000544, -0.00000387, 0.00000773, -0.00000669, 0.00000468, -0.00000890, 0.00001208, -0.00000850, -0.00000008, -0.00001074, 0.00001178, -0.00000839, 0.00000962, -0.00001315, -0.00001471, 0.00000650, -0.00000005, 0.00002896, -0.00001519, 0.00001335, },
		{ -0.00000036, 0.00001243, 0.00000706, 0.00000712, -0.00000218, 0.00002693, -0.00000656, 0.00000873, -0.00002156, -0.00000177, -0.00001096, 0.00000341, -0.00000114, -0.00000405, 0.00000611, -0.00000859, 0.00001380, -0.00002037, -0.00000202, 0.00000129, 0.00001094, -0.00001187, 0.00000197, -0.00001010, -0.00000003, -0.00000496, -0.00000741, 0.00000215, -0.00000168, -0.00000046, -0.00000197, -0.00000576, 0.00000341, -0.00001211, -0.00000687, -0.00001153, -0.00001301, -0.00002238, -0.00003019, -0.00006418, -0.00020115, 0.00026648, 0.00013601, 0.00014214, 0.00026568, -0.00020160, -0.00006692, -0.00002838, -0.00002692, -0.00000417, -0.00000497, -0.00000959, -0.00000601, -0.00000828, -0.00000832, -0.00000155, -0.00000879, 0.00000709, -0.00000509, -0.00000110, -0.00000406, 0.00000641, 0.00000534, -0.00000348, 0.00000019, -0.00000318, -0.00001576, 0.00001695, -0.00000491, 0.00001019, 0.00000306, -0.00000277, -0.00001312, 0.00000632, 0.00002459, -0.00002826, 0.00000568, -0.00001210, -0.00000338, -0.00000088, -0.00000130, -0.00000133, -0.00000163, -0.00000488, -0.00000360, -0.00000053, 0.00000040, 0.00000086, 0.00000329, -0.00000606, 0.00000343, -0.00000654, -0.00000788, 0.00000478, -0.00002305, 0.00001659, -0.00000616, 0.00000247, -0.00000105, -0.00000092, 0.00000734, -0.00000566, -0.00000201, 0.00000229, 0.00000189, -0.00000097, 0.00000671, -0.00000112, -0.00000841, 0.00000889, -0.00000565, 0.00000319, -0.00000648, 0.00000668, -0.00000021, -0.00000642, -0.00000520, 0.00000187, 0.00000369, -0.00000444, 0.00000709, -0.00003028, 0.00001035, -0.00000842, 0.00003004, -0.00000303, 0.00000069, 0.00000961, },
		{ 0.00001413, 0.00000471, 0.00001349, -0.00000768, 0.00002717, 0.00000026, 0.00000563, -0.00001414, -0.00001096, -0.00000329, -0.00000801, 0.00000910, -0.00001181, 0.00001174, -0.00001301, 0.00001309, -0.00000979, -0.00001145, 0.00000101, 0.00000977, -0.00000438, -0.00000553, -0.00000223, -0.00000745, 0.00000075, -0.00001098, 0.00000205, -0.00000181, 0.00000026, -0.00000178, -0.00000480, 0.00000019, -0.00000442, -0.00001365, -0.00000299, -0.00002025, -0.00000876, -0.00004255, -0.00002115, -0.00020961, 0.00013147, 0.00023496, 0.00003928, 0.00030481, -0.00004573, -0.00016548, -0.00000191, -0.00004937, 0.00000290, -0.00001351, 0.00000073, -0.00001547, 0.00000046, -0.00001648, 0.00000382, -0.00001336, 0.00000759, -0.00000423, -0.00000016, -0.00000606, 0.00000505, 0.00000552, 0.00000076, -0.00000479, 0.00000434, -0.00001932, 0.00001040, 0.00000368, 0.00000014, 0.00001119, -0.00000390, -0.00000552, -0.00001190, 0.00003467, -0.00002045, -0.00000435, -0.00000474, -0.00001023, 0.00000306, -0.00000446, 0.00000069, -0.00000288, -0.00000248, -0.00000571, -0.00000049, -0.00000077, 0.00000170, 0.00000165, -0.00000071, -0.00000405, 0.00000365, -0.00001577, 0.00000726, -0.00001819, 0.00000222, 0.00000859, -0.00000979, 0.00000751, -0.00000651, 0.00001014, -0.00000527, -0.00000177, 0.00000008, 0.00000364, -0.00000211, 0.00000505, 0.00000384, -0.00001017, 0.00000597, -0.00000087, -0.00000158, -0.00000104, -0.00000102, 0.00000827, -0.00001111, -0.00000118, -0.00000570, 0.00001078, -0.00001037, 0.00001293, -0.00002500, -0.00000440, -0.00000031, 0.00001310, 0.00002050, -0.00001454, 0.00001892, -0.00000542, },
		{ 0.00000616, 0.00001531, -0.00000624, 0.00001990, 0.00001164, 0.00000036, -0.00000490, -0.00001978, 0.00000220, -0.00001438, 0.00001057, -0.00000946, 0.00000690, -0.00000565, 0.00000057, 0.00000656, -0.00001961, 0.00000069, 0.00000452, 0.00000588, -0.00001266, 0.00000326, -0.00001110, 0.00000271, -0.00001172, 0.00000111, -0.00000310, 0.00000271, -0.00000416, -0.00000015, -0.00000685, 0.00000532, -0.00002076, 0.00000461, -0.00002574, 0.00000098, -0.00004842, -0.00000079, -0.00017705, -0.00000004, 0.00027527, 0.00003538, 0.00021858, 0.00014278, -0.00019980, -0.00003132, -0.00003728, -0.00001467, -0.00000573, -0.00000402, -0.00000981, -0.00000531, -0.00001038, -0.00000422, -0.00000563, -0.00000252, 0.00000453, -0.00000594, -0.00000080, -0.00000279, 0.00001041, -0.00000052, -0.00000210, 0.00000108, -0.00000948, -0.00000672, 0.00001764, -0.00001087, 0.00001722, -0.00000554, 0.00000421, -0.00002392, 0.00003058, -0.00000281, -0.00001583, 0.00000132, -0.00001496, 0.00000395, -0.00000426, -0.00000042, -0.00000189, -0.00000218, -0.00000527, -0.00000210, -0.00000111, 0.00000179, 0.00000057, 0.00000242, -0.00000599, 0.00000395, -0.00000993, -0.00000403, -0.00000314, -0.00001514, 0.00001879, -0.00001296, 0.00000547, -0.00000286, 0.00000557, 0.00000069, -0.00000578, 0.00000221, 0.00000078, 0.00000149, -0.00000072, 0.00000988, -0.00001053, 0.00000244, 0.00000250, -0.00000344, 0.00000047, -0.00000329, 0.00000889, -0.00000673, -0.00000563, -0.00000374, 0.00000581, -0.00000230, 0.00000202, -0.00000391, -0.00002637, 0.00001074, -0.00000527, 0.00003632, -0.00001626, 0.00001570, -0.00000038, 0.00000934, },
		{ 0.00001259, 0.00000147, 0.00000777, 0.00002371, -0.00000358, 0.00000283, -0.00002458, 0.00000201, -0.00001337, 0.00000379, -0.00000082, -0.00000220, 0.00000451, -0.00001103, 0.00001499, -0.00001704, -0.00000509, 0.00000137, 0.00001088, -0.00001197, 0.00000107, -0.00000815, -0.00000002, -0.00000946, -0.00000113, -0.00000389, 0.00000408, -0.00000550, 0.00000332, -0.00001135, 0.00000999, -0.00002130, 0.00000489, -0.00002358, 0.00000144, -0.00004284, -0.00000225, -0.00012249, -0.00010156, 0.00025108, 0.00009317, 0.00010722, 0.00024562, -0.00011247, -0.00011755, -0.00000282, -0.00004539, 0.00000986, -0.00001792, 0.00000493, -0.00001878, 0.00000243, -0.00001739, 0.00000577, -0.00001459, 0.00001346, -0.00001208, 0.00000583, -0.00001115, 0.00001489, -0.00000317, 0.00000377, -0.00000637, 0.00000331, -0.00002086, 0.00002341, -0.00001172, 0.00001372, -0.00000166, 0.00000759, -0.00002197, 0.00001211, 0.00001850, -0.00002306, 0.00000321, -0.00001589, 0.00000088, -0.00000105, -0.00000326, -0.00000032, -0.00000253, -0.00000395, -0.00000371, -0.00000144, 0.00000104, 0.00000107, 0.00000192, -0.00000227, -0.00000228, 0.00000064, -0.00001450, 0.00000453, -0.00001875, 0.00001203, -0.00000064, -0.00000793, 0.00000738, -0.00000286, 0.00000888, -0.00001099, 0.00000560, -0.00000327, 0.00000588, -0.00000590, 0.00001275, -0.00000747, -0.00000166, 0.00000383, -0.00000211, -0.00000166, -0.00000110, 0.00000392, 0.00000247, -0.00001287, 0.00000117, -0.00000278, 0.00000826, -0.00000961, 0.00001183, -0.00003378, 0.00000637, -0.00000824, 0.00003083, 0.00000161, -0.00000000, 0.00001189, 0.00000113, 0.00001000, },
		{ 0.00001095, -0.00000330, 0.00003090, -0.00000234, 0.00000625, -0.00002347, -0.00000383, -0.00000773, -0.00000521, 0.00000588, -0.00000669, 0.00000793, -0.00001211, 0.00000945, -0.00000365, -0.00001436, 0.00000089, 0.00000890, -0.00000373, -0.00000638, -0.00000198, -0.00000451, -0.00000566, -0.00000501, -0.00000207, 0.00000125, -0.00000205, 0.00000100, -0.00000805, 0.00000472, -0.00001133, -0.00000526, -0.00001107, -0.00000972, -0.00002415, -0.00002198, -0.00005979, -0.00016667, 0.00018648, 0.00015429, 0.00004855, 0.00022735, 0.00005092, -0.00018719, 0.00000011, -0.00005253, 0.00000638, -0.00001907, 0.00000813, -0.00002045, 0.00000376, -0.00001883, 0.00000546, -0.00001489, 0.00001077, -0.00000837, 0.00000393, -0.00001030, 0.00001023, 0.00000114, 0.00000343, -0.00000673, 0.00000591, -0.00001964, 0.00001251, 0.00000148, 0.00000018, 0.00000822, 0.00000271, -0.00000781, -0.00001191, 0.00003333, -0.00002107, 0.00000005, -0.00001332, -0.00000472, 0.00000254, -0.00000480, -0.00000064, -0.00000164, -0.00000363, -0.00000379, -0.00000263, 0.00000025, 0.00000185, 0.00000047, 0.00000173, -0.00000624, 0.00000440, -0.00001278, -0.00000267, -0.00000743, -0.00000587, 0.00001597, -0.00001905, 0.00001145, -0.00000483, 0.00001081, -0.00000995, 0.00000340, -0.00000201, 0.00000475, -0.00000459, 0.00000812, 0.00000101, -0.00000752, 0.00000535, -0.00000133, -0.00000238, -0.00000061, 0.00000106, 0.00000712, -0.00001296, -0.00000104, -0.00000371, 0.00000852, -0.00000788, 0.00000886, -0.00001903, -0.00001435, 0.00000227, 0.00001043, 0.00002654, -0.00001426, 0.00001990, -0.00000254, 0.00001169, 0.00000871, },
		{ -0.00000718, 0.00002866, 0.00000630, 0.00000429, -0.00001655, -0.00001321, -0.00000149, -0.00001132, 0.00000646, -0.00000443, 0.00000420, -0.00000497, -0.00000277, 0.00000963, -0.00001810, -0.00000201, 0.00000497, 0.00000484, -0.00001129, 0.00000088, -0.00000631, -0.00000292, -0.00000904, 0.00000127, -0.00000425, 0.00000409, -0.00000480, 0.00000003, -0.00000591, 0.00000297, -0.00001861, 0.00000392, -0.00002381, -0.00000270, -0.00004465, -0.00000931, -0.00018819, 0.00009856, 0.00019237, 0.00004830, 0.00014636, 0.00018275, -0.00015944, -0.00006186, -0.00002247, -0.00002370, -0.00000120, -0.00000700, -0.00000448, -0.00001084, -0.00000537, -0.00000844, -0.00000262, -0.00000329, 0.00000461, -0.00000646, -0.00000020, -0.00000293, 0.00001191, -0.00000370, 0.00000132, -0.00000226, -0.00000592, -0.00000754, 0.00001831, -0.00001269, 0.00001574, -0.00000322, 0.00000749, -0.00002750, 0.00003254, -0.00000823, -0.00000694, -0.00000839, -0.00001113, 0.00000442, -0.00000382, -0.00000283, -0.00000006, -0.00000384, -0.00000302, -0.00000389, -0.00000065, 0.00000196, 0.00000032, 0.00000237, -0.00000402, -0.00000059, -0.00000189, -0.00001533, 0.00000402, -0.00001719, 0.00001963, -0.00001426, 0.00000096, 0.00000391, 0.00000399, -0.00000105, -0.00000529, 0.00000492, -0.00000179, 0.00000237, -0.00000153, 0.00001128, -0.00001251, 0.00000654, -0.00000213, -0.00000004, -0.00000376, 0.00000274, 0.00000475, -0.00000472, -0.00001029, 0.00000251, -0.00000014, 0.00000361, -0.00000476, 0.00000287, -0.00003268, 0.00000829, -0.00000547, 0.00003831, -0.00001174, 0.00001445, 0.00000317, 0.00000840, 0.00000725, 0.00001670, },
		{ 0.00001763, 0.00001983, -0.00000037, -0.00000676, -0.00002244, 0.00000202, -0.00001303, 0.00000259, -0.00000010, -0.00000004, 0.00000128, -0.00000995, 0.00001197, -0.00001065, -0.00000965, 0.00000298, 0.00000804, -0.00000901, -0.00000214, -0.00000432, -0.00000224, -0.00001125, 0.00000276, -0.00000747, 0.00000688, -0.00000724, 0.00000470, -0.00001199, 0.00001039, -0.00002331, 0.00000905, -0.00002750, 0.00000624, -0.00005118, 0.00000926, -0.00016094, -0.00000424, 0.00020928, 0.00007050, 0.00008197, 0.00021128, -0.00002945, -0.00015106, 0.00000822, -0.00005209, 0.00001390, -0.00002263, 0.00001210, -0.00002481, 0.00000709, -0.00002067, 0.00000767, -0.00001488, 0.00001417, -0.00001394, 0.00000800, -0.00001320, 0.00001851, -0.00000791, 0.00000778, -0.00000950, 0.00000554, -0.00001999, 0.00002313, -0.00001260, 0.00001232, -0.00000215, 0.00001338, -0.00002593, 0.00001533, 0.00001071, -0.00001394, -0.00000296, -0.00001675, 0.00000388, -0.00000137, -0.00000466, -0.00000003, -0.00000280, -0.00000314, -0.00000373, -0.00000234, 0.00000164, 0.00000121, 0.00000071, 0.00000089, -0.00000689, 0.00000618, -0.00001880, 0.00000176, -0.00001180, 0.00000606, 0.00000411, -0.00001666, 0.00001426, -0.00000318, 0.00000785, -0.00001270, 0.00000989, -0.00000625, 0.00000724, -0.00000783, 0.00001549, -0.00001101, 0.00000375, -0.00000125, 0.00000150, -0.00000597, 0.00000385, 0.00000167, 0.00000340, -0.00001659, 0.00000526, -0.00000556, 0.00001019, -0.00001110, 0.00001143, -0.00003085, -0.00000275, -0.00000577, 0.00002906, 0.00000814, -0.00000035, 0.00001340, 0.00000331, 0.00000809, 0.00001673, -0.00000226, },
		{ 0.00003112, -0.00000184, 0.00000078, -0.00002711, 0.00000041, -0.00001088, -0.00000236, 0.00000171, -0.00000026, 0.00000192, -0.00000823, 0.00000377, 0.00000256, -0.00001634, 0.00000071, 0.00000700, -0.00000246, -0.00000691, -0.00000202, -0.00000207, -0.00001109, 0.00000036, -0.00000543, 0.00000313, -0.00000302, 0.00000199, -0.00000886, 0.00000631, -0.00001524, 0.00000059, -0.00001681, -0.00000205, -0.00003610, -0.00000526, -0.00009842, -0.00010300, 0.00020343, 0.00009228, 0.00006315, 0.00015525, 0.00012104, -0.00018319, -0.00001267, -0.00004590, 0.00000233, -0.00001868, 0.00001013, -0.00002059, 0.00000286, -0.00001716, 0.00000341, -0.00001262, 0.00001019, -0.00000935, 0.00000508, -0.00001137, 0.00001271, -0.00000159, 0.00000446, -0.00000644, 0.00000448, -0.00001610, 0.00001131, 0.00000088, -0.00000014, 0.00000514, 0.00000925, -0.00001083, -0.00000823, 0.00002553, -0.00001558, 0.00000036, -0.00001952, 0.00000044, 0.00000129, -0.00000459, -0.00000225, -0.00000045, -0.00000426, -0.00000230, -0.00000436, 0.00000103, 0.00000191, -0.00000035, 0.00000333, -0.00000739, 0.00000412, -0.00000908, -0.00001233, 0.00000216, -0.00001113, 0.00001906, -0.00002432, 0.00001240, -0.00000052, 0.00000822, -0.00001107, 0.00000607, -0.00000257, 0.00000431, -0.00000540, 0.00000980, -0.00000143, -0.00000390, 0.00000318, -0.00000097, -0.00000314, -0.00000021, 0.00000353, 0.00000437, -0.00001289, -0.00000187, -0.00000107, 0.00000481, -0.00000347, 0.00000265, -0.00001162, -0.00002292, 0.00000127, 0.00001009, 0.00002956, -0.00000959, 0.00001744, 0.00000278, 0.00000820, 0.00001345, 0.00000727, 0.00000435, },
		{ 0.00000442, 0.00000258, -0.00002465, -0.00000680, -0.00000611, -0.00000677, 0.00000091, 0.00000161, 0.00000033, -0.00000348, -0.00000510, 0.00001008, -0.00001421, -0.00000564, 0.00000581, 0.00000244, -0.00000790, -0.00000292, -0.00000099, -0.00000908, -0.00000504, 0.00000002, -0.00000405, 0.00000391, -0.00000392, -0.00000127, -0.00000319, -0.00000196, -0.00001272, -0.00000132, -0.00001688, -0.00001254, -0.00003220, -0.00003383, -0.00016589, 0.00016024, 0.00011925, 0.00006561, 0.00008855, 0.00019712, -0.00010793, -0.00009009, -0.00001167, -0.00002923, 0.00000022, -0.00000754, -0.00000152, -0.00001395, -0.00000243, -0.00001071, -0.00000115, -0.00000261, 0.00000287, -0.00000527, -0.00000109, -0.00000100, 0.00001068, -0.00000467, 0.00000322, -0.00000453, -0.00000339, -0.00000606, 0.00001505, -0.00001005, 0.00001005, 0.00000305, 0.00000573, -0.00002399, 0.00002713, -0.00000886, -0.00000015, -0.00001758, -0.00000630, 0.00000377, -0.00000321, -0.00000483, 0.00000097, -0.00000445, -0.00000151, -0.00000507, -0.00000049, 0.00000210, 0.00000019, 0.00000212, -0.00000232, -0.00000418, 0.00000425, -0.00002354, 0.00000766, -0.00001517, 0.00001660, -0.00001341, -0.00000413, 0.00001092, 0.00000161, -0.00000218, -0.00000400, 0.00000607, -0.00000306, 0.00000197, -0.00000054, 0.00000995, -0.00001103, 0.00000751, -0.00000484, 0.00000216, -0.00000662, 0.00000774, 0.00000087, -0.00000345, -0.00001289, 0.00000656, -0.00000448, 0.00000764, -0.00000896, 0.00000559, -0.00003266, -0.00000043, -0.00000193, 0.00003595, -0.00000254, 0.00001043, 0.00000850, 0.00000642, 0.00001007, 0.00001532, -0.00000348, 0.00003338, },
		{ 0.00000027, -0.00001696, -0.00001686, -0.00000117, -0.00001012, -0.00000020, 0.00000135, 0.00000157, -0.00000227, -0.00000689, 0.00000660, -0.00000323, -0.00001394, 0.00000467, 0.00000390, -0.00000458, -0.00000591, -0.00000032, -0.00000546, -0.00001085, 0.00000430, -0.00000860, 0.00000688, -0.00000609, 0.00000290, -0.00000876, 0.00000583, -0.00001916, 0.00000628, -0.00002377, 0.00000160, -0.00004692, 0.00000171, -0.00017075, 0.00007396, 0.00015772, 0.00006103, 0.00006325, 0.00017229, 0.00003777, -0.00016264, 0.00000373, -0.00004752, 0.00000882, -0.00001969, 0.00001208, -0.00002417, 0.00000595, -0.00001861, 0.00000487, -0.00001061, 0.00001012, -0.00001135, 0.00000586, -0.00001040, 0.00001712, -0.00000858, 0.00000832, -0.00000926, 0.00000405, -0.00001424, 0.00001719, -0.00000832, 0.00000665, 0.00000136, 0.00001427, -0.00002444, 0.00001484, 0.00000342, -0.00000341, -0.00001109, -0.00001518, 0.00000543, -0.00000174, -0.00000552, -0.00000043, -0.00000242, -0.00000255, -0.00000366, -0.00000306, 0.00000201, 0.00000130, -0.00000009, 0.00000284, -0.00000959, 0.00000919, -0.00002041, -0.00000276, -0.00000394, 0.00000053, 0.00000644, -0.00002197, 0.00001823, -0.00000154, 0.00000454, -0.00001086, 0.00001053, -0.00000624, 0.00000566, -0.00000637, 0.00001435, -0.00001074, 0.00000631, -0.00000470, 0.00000377, -0.00000853, 0.00000729, 0.00000061, 0.00000215, -0.00001697, 0.00000633, -0.00000586, 0.00000907, -0.00000886, 0.00000648, -0.00002294, -0.00001466, -0.00000309, 0.00002580, 0.00001597, -0.00000014, 0.00001435, 0.00000573, 0.00000739, 0.00001863, -0.00000271, 0.00002574, 0.00001706, },
		{ -0.00000914, -0.00002482, 0.00000056, -0.00001137, -0.00000094, -0.00000167, 0.00000540, -0.00000407, -0.00000350, -0.00000199, 0.00000730, -0.00001650, -0.00000009, 0.00000465, -0.00000111, -0.00000675, -0.00000239, -0.00000072, -0.00001444, 0.00000396, -0.00000720, 0.00000332, -0.00000240, 0.00000113, -0.00000721, 0.00000467, -0.00001528, 0.00000243, -0.00001830, 0.00000015, -0.00004075, 0.00000130, -0.00012606, -0.00003250, 0.00018960, 0.00005349, 0.00007186, 0.00009944, 0.00015970, -0.00015518, -0.00003644, -0.00003274, -0.00000696, -0.00001367, 0.00000751, -0.00001663, -0.00000134, -0.00001246, -0.00000130, -0.00000754, 0.00000654, -0.00000751, 0.00000337, -0.00000915, 0.00001223, -0.00000242, 0.00000394, -0.00000443, 0.00000093, -0.00000974, 0.00000737, 0.00000191, -0.00000123, 0.00000286, 0.00001436, -0.00001308, -0.00000292, 0.00001390, -0.00000600, -0.00000286, -0.00002264, 0.00000425, -0.00000006, -0.00000427, -0.00000375, 0.00000039, -0.00000414, -0.00000154, -0.00000528, 0.00000128, 0.00000199, -0.00000078, 0.00000400, -0.00000757, 0.00000322, -0.00000582, -0.00001972, 0.00000846, -0.00001219, 0.00001710, -0.00002507, 0.00001047, 0.00000548, 0.00000343, -0.00000904, 0.00000615, -0.00000147, 0.00000219, -0.00000422, 0.00000949, -0.00000251, -0.00000054, 0.00000044, -0.00000025, -0.00000374, 0.00000028, 0.00000589, 0.00000066, -0.00001131, -0.00000328, 0.00000161, 0.00000059, 0.00000152, -0.00000394, -0.00000479, -0.00002831, -0.00000376, 0.00001125, 0.00003066, -0.00000184, 0.00001304, 0.00000911, 0.00000509, 0.00001755, 0.00000417, 0.00001360, 0.00002940, -0.00000045, },
		{ -0.00002719, -0.00000298, -0.00000959, -0.00000263, -0.00000423, 0.00000696, -0.00000337, -0.00000198, -0.00000620, 0.00000843, -0.00000902, -0.00000911, 0.00000608, -0.00000021, -0.00000382, -0.00000649, 0.00000286, -0.00001373, -0.00000139, -0.00000142, -0.00000347, 0.00000307, -0.00000257, -0.00000269, -0.00000059, -0.00000686, -0.00000626, -0.00000755, -0.00000915, -0.00002330, -0.00001798, -0.00006529, -0.00011758, 0.00018291, 0.00006715, 0.00007604, 0.00004998, 0.00018892, -0.00005200, -0.00011118, -0.00000692, -0.00003083, -0.00000104, -0.00000629, -0.00000049, -0.00001489, -0.00000131, -0.00001133, -0.00000075, -0.00000128, 0.00000035, -0.00000345, -0.00000238, 0.00000195, 0.00000750, -0.00000381, 0.00000369, -0.00000569, -0.00000161, -0.00000340, 0.00000961, -0.00000492, 0.00000238, 0.00001097, 0.00000108, -0.00001573, 0.00001644, -0.00000528, 0.00000350, -0.00002452, -0.00000183, 0.00000264, -0.00000281, -0.00000615, 0.00000113, -0.00000396, -0.00000092, -0.00000543, -0.00000072, 0.00000222, 0.00000014, 0.00000186, -0.00000127, -0.00000610, 0.00000733, -0.00002743, 0.00000725, -0.00000963, 0.00001072, -0.00001135, -0.00000847, 0.00001657, -0.00000054, -0.00000281, -0.00000224, 0.00000599, -0.00000337, 0.00000083, 0.00000149, 0.00000678, -0.00000691, 0.00000558, -0.00000523, 0.00000253, -0.00000734, 0.00001079, -0.00000198, -0.00000333, -0.00001307, 0.00000787, -0.00000650, 0.00000901, -0.00000995, 0.00000414, -0.00002724, -0.00001305, 0.00000247, 0.00003158, 0.00000923, 0.00000548, 0.00001422, 0.00000458, 0.00001356, 0.00001238, 0.00000389, 0.00003482, 0.00000500, -0.00000497, },
		{ -0.00001052, -0.00000579, -0.00000587, -0.00000381, 0.00000326, 0.00000228, -0.00000507, -0.00000352, 0.00000097, 0.00000309, -0.00001630, 0.00000500, 0.00000002, -0.00000036, -0.00000852, 0.00000213, -0.00000825, -0.00000924, 0.00000418, -0.00000800, 0.00000503, -0.00000336, -0.00000032, -0.00000393, -0.00000091, -0.00001183, -0.00000040, -0.00001567, -0.00000878, -0.00003419, -0.00002099, -0.00015145, 0.00012357, 0.00010963, 0.00005758, 0.00004972, 0.00013483, 0.00008474, -0.00015291, -0.00001291, -0.00003550, -0.00000176, -0.00001214, 0.00000733, -0.00001904, 0.00000113, -0.00001334, -0.00000045, -0.00000411, 0.00000361, -0.00000637, 0.00000132, -0.00000446, 0.00001209, -0.00000621, 0.00000638, -0.00000683, 0.00000062, -0.00000613, 0.00000832, -0.00000137, -0.00000105, 0.00000702, 0.00001164, -0.00001865, 0.00001086, -0.00000183, 0.00000595, -0.00001899, -0.00001225, 0.00000572, -0.00000211, -0.00000595, -0.00000125, -0.00000157, -0.00000216, -0.00000352, -0.00000350, 0.00000203, 0.00000136, -0.00000038, 0.00000339, -0.00001010, 0.00000942, -0.00001949, -0.00000796, 0.00000313, -0.00000318, 0.00000548, -0.00002304, 0.00001871, 0.00000174, 0.00000002, -0.00000668, 0.00000846, -0.00000411, 0.00000209, -0.00000246, 0.00001014, -0.00000710, 0.00000575, -0.00000580, 0.00000404, -0.00000881, 0.00000887, 0.00000069, -0.00000083, -0.00001451, 0.00000478, -0.00000411, 0.00000559, -0.00000414, -0.00000104, -0.00001288, -0.00002605, -0.00000269, 0.00002229, 0.00002426, 0.00000099, 0.00001482, 0.00000821, 0.00000804, 0.00001829, 0.00000044, 0.00003157, 0.00001595, -0.00000363, -0.00002455, },
		{ -0.00000275, -0.00000921, -0.00000148, -0.00000325, 0.00000880, -0.00000818, 0.00000016, -0.00000611, 0.00000942, -0.00001440, -0.00000182, 0.00000233, -0.00000009, -0.00000563, -0.00000314, -0.00000065, -0.00001527, 0.00000536, -0.00000720, 0.00000199, -0.00000067, -0.00000082, -0.00000407, 0.00000079, -0.00001214, 0.00000056, -0.00001563, -0.00000328, -0.00003729, -0.00000485, -0.00013510, 0.00003115, 0.00015687, 0.00003519, 0.00007107, 0.00006304, 0.00016942, -0.00011032, -0.00006407, -0.00001845, -0.00001745, -0.00000700, 0.00000256, -0.00001067, -0.00000681, -0.00000670, -0.00000674, -0.00000156, 0.00000147, -0.00000416, -0.00000005, -0.00000461, 0.00000944, -0.00000192, 0.00000259, -0.00000180, -0.00000309, -0.00000249, 0.00000219, 0.00000379, -0.00000282, 0.00000173, 0.00001720, -0.00001331, 0.00000174, 0.00000181, 0.00000455, -0.00000793, -0.00002290, 0.00000626, -0.00000111, -0.00000417, -0.00000482, 0.00000080, -0.00000338, -0.00000148, -0.00000535, 0.00000095, 0.00000208, -0.00000080, 0.00000378, -0.00000688, 0.00000199, -0.00000368, -0.00002372, 0.00001056, -0.00000906, 0.00001076, -0.00002193, 0.00000660, 0.00001154, -0.00000169, -0.00000535, 0.00000461, 0.00000056, -0.00000094, -0.00000153, 0.00000737, -0.00000191, 0.00000169, -0.00000197, 0.00000035, -0.00000405, 0.00000106, 0.00000751, -0.00000313, -0.00000896, -0.00000469, 0.00000362, -0.00000314, 0.00000569, -0.00000922, -0.00000021, -0.00002958, -0.00001218, 0.00001222, 0.00003145, 0.00000719, 0.00000859, 0.00001489, 0.00000364, 0.00001988, 0.00000312, 0.00002276, 0.00002753, -0.00000115, -0.00002016, -0.00001852, },
		{ -0.00001065, -0.00000031, -0.00000750, 0.00001016, -0.00000563, -0.00000123, -0.00000605, 0.00000542, -0.00000373, -0.00001141, 0.00000500, -0.00000218, -0.00000021, -0.00000857, 0.00000393, -0.00001554, 0.00000071, -0.00000215, -0.00000323, 0.00000227, -0.00000150, -0.00000331, 0.00000056, -0.00000973, -0.00000184, -0.00001171, -0.00000432, -0.00003016, -0.00000895, -0.00009273, -0.00005634, 0.00017362, 0.00003805, 0.00007568, 0.00002990, 0.00016459, 0.00000102, -0.00012134, -0.00000916, -0.00002874, -0.00000415, -0.00000414, -0.00000072, -0.00001421, -0.00000147, -0.00001083, -0.00000086, -0.00000009, -0.00000197, -0.00000199, -0.00000302, 0.00000475, 0.00000346, -0.00000184, 0.00000309, -0.00000591, -0.00000020, -0.00000075, 0.00000395, 0.00000037, -0.00000462, 0.00001772, -0.00000353, -0.00000619, 0.00000395, 0.00000049, 0.00000413, -0.00002823, 0.00000129, 0.00000157, -0.00000285, -0.00000675, 0.00000064, -0.00000270, -0.00000105, -0.00000507, -0.00000125, 0.00000223, 0.00000021, 0.00000159, -0.00000093, -0.00000627, 0.00000719, -0.00002692, 0.00000337, -0.00000204, 0.00000353, -0.00000921, -0.00001091, 0.00001948, -0.00000138, -0.00000329, -0.00000028, 0.00000510, -0.00000325, -0.00000031, 0.00000355, 0.00000316, -0.00000172, 0.00000191, -0.00000384, 0.00000123, -0.00000587, 0.00001160, -0.00000359, -0.00000420, -0.00001126, 0.00000678, -0.00000643, 0.00000795, -0.00000824, -0.00000025, -0.00001880, -0.00002600, 0.00000426, 0.00002782, 0.00002114, 0.00000172, 0.00001879, 0.00000414, 0.00001643, 0.00000962, 0.00001312, 0.00003483, 0.00000535, -0.00001622, -0.00002428, -0.00000274, },
		{ -0.00000196, -0.00000698, 0.00000484, 0.00000227, -0.00000660, -0.00000105, -0.00000329, 0.00000682, -0.00001577, 0.00000319, -0.00000253, 0.00000247, -0.00000889, 0.00000204, -0.00000962, -0.00000732, 0.00000319, -0.00000655, 0.00000246, -0.00000047, -0.00000315, 0.00000013, -0.00000682, -0.00000478, -0.00000723, -0.00000750, -0.00001966, -0.00001969, -0.00005031, -0.00010995, 0.00014338, 0.00007267, 0.00005487, 0.00004064, 0.00010224, 0.00011135, -0.00012650, -0.00003574, -0.00002144, -0.00001313, -0.00000392, 0.00000118, -0.00001243, -0.00000456, -0.00000752, -0.00000568, 0.00000196, -0.00000273, -0.00000148, -0.00000323, 0.00000224, 0.00000552, -0.00000251, 0.00000347, -0.00000382, -0.00000265, 0.00000146, -0.00000027, 0.00000520, -0.00000788, 0.00001222, 0.00000782, -0.00001076, 0.00000466, -0.00000443, 0.00001220, -0.00002463, -0.00000918, 0.00000516, -0.00000250, -0.00000614, -0.00000214, -0.00000057, -0.00000189, -0.00000332, -0.00000363, 0.00000169, 0.00000141, -0.00000024, 0.00000273, -0.00000871, 0.00000724, -0.00001663, -0.00001277, 0.00000815, -0.00000447, 0.00000142, -0.00002005, 0.00001590, 0.00000597, -0.00000429, -0.00000190, 0.00000533, -0.00000143, -0.00000180, 0.00000213, 0.00000470, -0.00000168, 0.00000297, -0.00000488, 0.00000253, -0.00000709, 0.00000890, 0.00000127, -0.00000438, -0.00001055, 0.00000183, -0.00000156, 0.00000123, 0.00000111, -0.00000849, -0.00000391, -0.00003370, -0.00000653, 0.00001935, 0.00003229, 0.00000343, 0.00001486, 0.00001079, 0.00000969, 0.00001656, 0.00000643, 0.00003576, 0.00001543, -0.00001220, -0.00002739, -0.00000597, -0.00000973, },
		{ -0.00000339, -0.00000309, 0.00000950, -0.00000951, 0.00000159, -0.00000766, 0.00000862, -0.00001025, -0.00000435, 0.00000023, 0.00000044, -0.00000360, -0.00000423, -0.00000183, -0.00001368, 0.00000470, -0.00000580, -0.00000022, 0.00000129, -0.00000271, -0.00000092, -0.00000354, -0.00000762, -0.00000332, -0.00001060, -0.00001026, -0.00002810, -0.00002178, -0.00012337, 0.00007798, 0.00011765, 0.00003058, 0.00006256, 0.00004334, 0.00015772, -0.00005859, -0.00008759, -0.00000850, -0.00002503, -0.00000184, -0.00000205, -0.00000512, -0.00001129, -0.00000197, -0.00001089, 0.00000331, -0.00000314, -0.00000098, -0.00000358, 0.00000072, 0.00000557, -0.00000096, 0.00000133, 0.00000027, -0.00000595, 0.00000358, -0.00000240, 0.00000532, -0.00000420, 0.00000151, 0.00001759, -0.00001097, 0.00000406, -0.00000766, 0.00001288, -0.00001273, -0.00002111, 0.00000651, -0.00000167, -0.00000447, -0.00000530, 0.00000080, -0.00000225, -0.00000182, -0.00000480, 0.00000022, 0.00000209, -0.00000045, 0.00000282, -0.00000551, 0.00000055, -0.00000280, -0.00002420, 0.00000876, -0.00000300, 0.00000177, -0.00001631, 0.00000208, 0.00001599, -0.00000515, -0.00000191, 0.00000293, 0.00000227, -0.00000389, 0.00000158, 0.00000431, -0.00000007, 0.00000254, -0.00000348, 0.00000052, -0.00000404, 0.00000227, 0.00000783, -0.00000609, -0.00000669, -0.00000548, 0.00000436, -0.00000559, 0.00000798, -0.00001198, 0.00000120, -0.00002679, -0.00002243, 0.00001100, 0.00003342, 0.00001591, 0.00000567, 0.00001901, 0.00000451, 0.00002008, 0.00000484, 0.00003093, 0.00002633, -0.00000631, -0.00002849, -0.00001138, -0.00000752, -0.00000559, },
		{ -0.00000831, 0.00001056, -0.00000580, -0.00000160, -0.00000484, 0.00000191, 0.00000065, -0.00001195, 0.00000226, -0.00000293, 0.00000224, -0.00000864, 0.00000190, -0.00001423, 0.00000098, -0.00000195, -0.00000348, 0.00000167, -0.00000085, -0.00000294, -0.00000005, -0.00000989, -0.00000059, -0.00001238, -0.00000413, -0.00003064, -0.00000941, -0.00010716, 0.00000358, 0.00014455, 0.00002714, 0.00006635, 0.00002352, 0.00013220, 0.00004462, -0.00011856, -0.00001799, -0.00002384, -0.00000816, -0.00000192, -0.00000151, -0.00001255, -0.00000235, -0.00000971, -0.00000106, 0.00000048, -0.00000353, -0.00000142, -0.00000242, 0.00000653, -0.00000045, 0.00000048, 0.00000187, -0.00000532, 0.00000099, 0.00000113, -0.00000047, 0.00000400, -0.00000885, 0.00002103, -0.00000556, 0.00000135, -0.00000680, 0.00000599, 0.00000259, -0.00002845, 0.00000251, 0.00000086, -0.00000340, -0.00000673, -0.00000013, -0.00000122, -0.00000147, -0.00000431, -0.00000185, 0.00000200, 0.00000045, 0.00000125, -0.00000104, -0.00000520, 0.00000453, -0.00002294, -0.00000253, 0.00000568, -0.00000333, -0.00000779, -0.00001088, 0.00001887, -0.00000011, -0.00000404, 0.00000177, 0.00000376, -0.00000308, -0.00000086, 0.00000478, 0.00000036, 0.00000287, -0.00000194, -0.00000168, -0.00000103, -0.00000277, 0.00001047, -0.00000416, -0.00000550, -0.00000835, 0.00000424, -0.00000515, 0.00000536, -0.00000510, -0.00000562, -0.00001026, -0.00003574, 0.00000066, 0.00002643, 0.00003138, 0.00000070, 0.00002124, 0.00000585, 0.00001786, 0.00000849, 0.00002254, 0.00003516, 0.00000230, -0.00002747, -0.00001804, -0.00000543, -0.00000927, -0.00000023, },
		{ 0.00000529, 0.00000209, -0.00000690, 0.00000002, -0.00000540, 0.00000748, -0.00001240, -0.00000067, -0.00000350, 0.00000361, -0.00000695, -0.00000055, -0.00000942, -0.00000550, 0.00000184, -0.00000489, -0.00000003, 0.00000163, -0.00000446, 0.00000196, -0.00000992, -0.00000053, -0.00001131, -0.00000273, -0.00002654, -0.00000991, -0.00007634, -0.00005757, 0.00013888, 0.00004882, 0.00005041, 0.00003515, 0.00007590, 0.00012045, -0.00009011, -0.00005821, -0.00001031, -0.00002132, 0.00000167, -0.00000353, -0.00000693, -0.00000873, -0.00000331, -0.00000886, 0.00000561, -0.00000694, 0.00000148, -0.00000587, 0.00000759, -0.00000056, 0.00000089, 0.00000093, -0.00000149, -0.00000427, 0.00000641, -0.00000606, 0.00000893, -0.00001150, 0.00001476, 0.00000503, -0.00000320, -0.00000194, -0.00000462, 0.00001429, -0.00002663, -0.00000698, 0.00000414, -0.00000291, -0.00000628, -0.00000271, 0.00000022, -0.00000157, -0.00000308, -0.00000351, 0.00000104, 0.00000144, 0.00000016, 0.00000132, -0.00000612, 0.00000344, -0.00001267, -0.00001630, 0.00001049, -0.00000358, -0.00000464, -0.00001412, 0.00001060, 0.00001030, -0.00000720, 0.00000190, 0.00000275, 0.00000027, -0.00000440, 0.00000560, 0.00000005, 0.00000345, -0.00000044, -0.00000293, 0.00000004, -0.00000414, 0.00000806, 0.00000150, -0.00000721, -0.00000664, -0.00000101, 0.00000037, -0.00000242, 0.00000493, -0.00001348, 0.00000139, -0.00003558, -0.00001521, 0.00001688, 0.00003968, 0.00000738, 0.00001460, 0.00001357, 0.00001176, 0.00001466, 0.00001400, 0.00003965, 0.00001310, -0.00002355, -0.00002522, -0.00000465, -0.00001145, 0.00000020, -0.00000861, },
		{ 0.00000810, -0.00000828, 0.00000076, -0.00000672, 0.00000548, -0.00000497, -0.00000709, -0.00000179, 0.00000068, -0.00000109, -0.00000538, -0.00000403, -0.00001020, 0.00000259, -0.00000368, -0.00000252, 0.00000261, -0.00000354, 0.00000095, -0.00000664, -0.00000373, -0.00000702, -0.00000578, -0.00001748, -0.00001761, -0.00004387, -0.00009429, 0.00010372, 0.00008164, 0.00003221, 0.00005060, 0.00003474, 0.00013402, -0.00000977, -0.00010068, -0.00000650, -0.00002716, -0.00000028, -0.00000446, -0.00000172, -0.00001323, 0.00000033, -0.00001245, 0.00000572, -0.00000595, 0.00000073, -0.00000579, 0.00000536, 0.00000179, -0.00000032, 0.00000078, 0.00000106, -0.00000659, 0.00000702, -0.00000498, 0.00000545, -0.00000460, 0.00000164, 0.00001599, -0.00000633, 0.00000337, -0.00001260, 0.00001671, -0.00001549, -0.00001824, 0.00000525, -0.00000175, -0.00000523, -0.00000511, 0.00000050, -0.00000107, -0.00000219, -0.00000398, -0.00000063, 0.00000188, 0.00000019, 0.00000140, -0.00000373, -0.00000103, -0.00000290, -0.00002181, 0.00000431, 0.00000397, -0.00000760, -0.00000987, -0.00000200, 0.00001777, -0.00000563, -0.00000019, 0.00000239, 0.00000257, -0.00000553, 0.00000392, 0.00000151, 0.00000212, 0.00000232, -0.00000400, 0.00000019, -0.00000372, 0.00000395, 0.00000658, -0.00000755, -0.00000519, -0.00000523, 0.00000356, -0.00000638, 0.00000789, -0.00001182, -0.00000047, -0.00002103, -0.00003239, 0.00000589, 0.00003725, 0.00002357, 0.00000512, 0.00002120, 0.00000746, 0.00001877, 0.00000906, 0.00003830, 0.00002501, -0.00001639, -0.00003149, -0.00000641, -0.00001135, -0.00000231, -0.00000545, -0.00000205, },
		{ -0.00000402, -0.00000298, -0.00000304, -0.00000130, 0.00000342, -0.00001062, -0.00000194, -0.00000243, 0.00000340, -0.00000691, -0.00000253, -0.00001055, -0.00000008, -0.00000121, -0.00000389, 0.00000099, -0.00000033, -0.00000209, -0.00000174, -0.00000800, -0.00000199, -0.00001012, -0.00000791, -0.00002542, -0.00001952, -0.00010425, 0.00005093, 0.00010854, 0.00002664, 0.00005284, 0.00002447, 0.00009936, 0.00007453, -0.00010324, -0.00003150, -0.00001760, -0.00001206, -0.00000031, -0.00000234, -0.00001041, -0.00000344, -0.00000833, -0.00000116, 0.00000032, -0.00000423, -0.00000171, -0.00000053, 0.00000685, -0.00000350, 0.00000246, 0.00000046, -0.00000405, 0.00000188, 0.00000202, -0.00000303, 0.00000519, -0.00000946, 0.00001998, -0.00000383, 0.00000494, -0.00001329, 0.00000918, -0.00000001, -0.00002552, 0.00000170, 0.00000064, -0.00000440, -0.00000627, -0.00000079, -0.00000004, -0.00000172, -0.00000349, -0.00000231, 0.00000146, 0.00000085, 0.00000079, -0.00000129, -0.00000362, 0.00000054, -0.00001702, -0.00000866, 0.00001166, -0.00000857, -0.00000732, -0.00000868, 0.00001492, 0.00000337, -0.00000524, 0.00000385, 0.00000218, -0.00000303, -0.00000069, 0.00000488, -0.00000092, 0.00000566, -0.00000466, 0.00000013, -0.00000329, 0.00000107, 0.00000803, -0.00000412, -0.00000660, -0.00000532, 0.00000132, -0.00000366, 0.00000234, -0.00000193, -0.00001004, -0.00000411, -0.00003990, -0.00000930, 0.00002742, 0.00003951, 0.00000290, 0.00002160, 0.00000951, 0.00001791, 0.00000957, 0.00003144, 0.00003632, -0.00000635, -0.00003493, -0.00001159, -0.00000918, -0.00000655, -0.00000177, -0.00000710, 0.00000848, },
		{ -0.00000639, -0.00000019, -0.00000544, 0.00000556, -0.00000715, -0.00000567, -0.00000335, 0.00000351, -0.00000354, -0.00000465, -0.00000823, -0.00000371, 0.00000029, -0.00000326, -0.00000211, 0.00000266, -0.00000408, 0.00000140, -0.00000978, 0.00000016, -0.00001161, -0.00000270, -0.00002729, -0.00000874, -0.00009108, -0.00000631, 0.00011920, 0.00003584, 0.00004413, 0.00003202, 0.00005590, 0.00011646, -0.00005074, -0.00007495, -0.00000532, -0.00002421, 0.00000302, -0.00000533, -0.00000390, -0.00001019, -0.00000168, -0.00000920, 0.00000613, -0.00000830, 0.00000183, -0.00000572, 0.00001036, -0.00000484, 0.00000297, -0.00000053, -0.00000029, -0.00000382, 0.00000796, -0.00000803, 0.00000883, -0.00001094, 0.00001362, 0.00000454, 0.00000233, -0.00000732, -0.00000315, 0.00001216, -0.00002451, -0.00000621, 0.00000294, -0.00000337, -0.00000649, -0.00000271, 0.00000050, -0.00000107, -0.00000281, -0.00000321, 0.00000022, 0.00000141, 0.00000062, -0.00000030, -0.00000322, -0.00000094, -0.00000850, -0.00001800, 0.00001016, -0.00000148, -0.00001091, -0.00000697, 0.00000399, 0.00001402, -0.00000809, 0.00000386, 0.00000161, 0.00000021, -0.00000486, 0.00000687, -0.00000241, 0.00000664, -0.00000297, -0.00000108, -0.00000244, -0.00000087, 0.00000699, 0.00000081, -0.00000842, -0.00000395, -0.00000265, 0.00000070, -0.00000433, 0.00000612, -0.00001472, 0.00000192, -0.00003153, -0.00002747, 0.00001362, 0.00004657, 0.00001278, 0.00001438, 0.00001649, 0.00001388, 0.00001366, 0.00002211, 0.00004426, 0.00000629, -0.00003431, -0.00001971, -0.00000637, -0.00001058, 0.00000012, -0.00000826, 0.00000420, 0.00000221, },
		{ -0.00000153, -0.00000421, 0.00000127, 0.00000016, -0.00000926, -0.00000402, 0.00000104, 0.00000127, -0.00000620, -0.00000696, -0.00000564, -0.00000020, -0.00000153, -0.00000431, 0.00000283, -0.00000291, 0.00000099, -0.00000757, -0.00000181, -0.00000888, -0.00000332, -0.00002191, -0.00001044, -0.00006395, -0.00005535, 0.00010972, 0.00005419, 0.00003436, 0.00003930, 0.00003144, 0.00010682, 0.00002904, -0.00010044, -0.00001300, -0.00002367, -0.00000256, -0.00000424, -0.00000094, -0.00001230, -0.00000000, -0.00001127, 0.00000542, -0.00000660, 0.00000054, -0.00000601, 0.00000838, -0.00000112, -0.00000032, 0.00000098, 0.00000065, -0.00000503, 0.00000749, -0.00000506, 0.00000381, -0.00000365, 0.00000163, 0.00001314, -0.00000031, 0.00000014, -0.00001266, 0.00001525, -0.00001534, -0.00001499, 0.00000283, -0.00000146, -0.00000638, -0.00000428, -0.00000003, -0.00000009, -0.00000227, -0.00000322, -0.00000136, 0.00000141, 0.00000095, -0.00000012, -0.00000189, -0.00000268, -0.00000351, -0.00001766, -0.00000111, 0.00000966, -0.00001515, -0.00000404, -0.00000506, 0.00001682, -0.00000301, -0.00000062, 0.00000344, 0.00000111, -0.00000535, 0.00000474, -0.00000011, 0.00000368, 0.00000160, -0.00000371, -0.00000058, -0.00000307, 0.00000587, 0.00000393, -0.00000738, -0.00000471, -0.00000395, 0.00000138, -0.00000568, 0.00000567, -0.00000923, -0.00000424, -0.00001412, -0.00003989, -0.00000402, 0.00004247, 0.00003038, 0.00000689, 0.00002195, 0.00001162, 0.00001718, 0.00001485, 0.00004584, 0.00002127, -0.00002952, -0.00002870, -0.00000528, -0.00001259, -0.00000079, -0.00000649, -0.00000121, 0.00000574, -0.00000550, },
		{ -0.00000129, -0.00000355, 0.00000433, -0.00000781, -0.00000693, -0.00000107, 0.00000353, -0.00000415, -0.00000801, -0.00000593, -0.00000149, -0.00000061, -0.00000396, -0.00000013, 0.00000047, -0.00000142, -0.00000341, -0.00000550, -0.00000441, -0.00000684, -0.00001318, -0.00001772, -0.00003551, -0.00008522, 0.00007993, 0.00007551, 0.00002922, 0.00003999, 0.00002712, 0.00007144, 0.00008959, -0.00007829, -0.00004653, -0.00001191, -0.00001487, 0.00000019, -0.00000283, -0.00000823, -0.00000438, -0.00000698, -0.00000113, -0.00000046, -0.00000428, -0.00000247, 0.00000221, 0.00000576, -0.00000527, 0.00000363, -0.00000072, -0.00000232, 0.00000231, 0.00000220, -0.00000391, 0.00000422, -0.00000688, 0.00001515, 0.00000121, 0.00000441, -0.00001477, 0.00000917, -0.00000282, -0.00002003, -0.00000089, 0.00000081, -0.00000573, -0.00000550, -0.00000115, 0.00000054, -0.00000151, -0.00000286, -0.00000250, 0.00000065, 0.00000129, 0.00000025, -0.00000142, -0.00000226, -0.00000352, -0.00001077, -0.00001344, 0.00001457, -0.00001159, -0.00000738, -0.00000541, 0.00000883, 0.00000833, -0.00000665, 0.00000575, 0.00000056, -0.00000304, -0.00000000, 0.00000407, -0.00000072, 0.00000623, -0.00000550, 0.00000087, -0.00000476, 0.00000475, 0.00000499, -0.00000385, -0.00000708, -0.00000289, -0.00000114, -0.00000274, -0.00000023, 0.00000022, -0.00001223, -0.00000168, -0.00003795, -0.00002436, 0.00002887, 0.00004644, 0.00000767, 0.00002087, 0.00001409, 0.00001757, 0.00001243, 0.00004034, 0.00003690, -0.00002053, -0.00003625, -0.00000770, -0.00001190, -0.00000402, -0.00000370, -0.00000499, 0.00000503, -0.00000114, -0.00000474, },
		{ -0.00000405, 0.00000219, -0.00000152, -0.00001053, -0.00000291, 0.00000296, 0.00000008, -0.00000881, -0.00000704, -0.00000166, -0.00000149, -0.00000166, -0.00000384, 0.00000290, -0.00000264, -0.00000059, -0.00000746, -0.00000175, -0.00000911, -0.00000640, -0.00002279, -0.00001628, -0.00009071, 0.00003452, 0.00009382, 0.00002959, 0.00003729, 0.00002996, 0.00004153, 0.00010423, -0.00001430, -0.00008276, -0.00000725, -0.00002189, 0.00000043, -0.00000439, -0.00000330, -0.00000910, -0.00000233, -0.00000720, 0.00000405, -0.00000734, 0.00000010, -0.00000310, 0.00001038, -0.00000691, 0.00000346, -0.00000089, 0.00000005, -0.00000192, 0.00000678, -0.00000678, 0.00000556, -0.00000689, 0.00000936, 0.00000619, 0.00000532, -0.00001059, -0.00000090, 0.00000660, -0.00001872, -0.00000683, 0.00000164, -0.00000384, -0.00000676, -0.00000210, 0.00000020, -0.00000036, -0.00000251, -0.00000285, -0.00000062, 0.00000126, 0.00000100, -0.00000163, -0.00000080, -0.00000494, -0.00000489, -0.00001774, 0.00000770, 0.00000054, -0.00001549, -0.00000054, -0.00000256, 0.00001665, -0.00000701, 0.00000399, 0.00000189, -0.00000138, -0.00000338, 0.00000598, -0.00000241, 0.00000718, -0.00000379, -0.00000009, -0.00000421, 0.00000210, 0.00000599, -0.00000079, -0.00000789, -0.00000284, -0.00000282, -0.00000066, -0.00000440, 0.00000465, -0.00001239, -0.00000160, -0.00002340, -0.00004041, 0.00000752, 0.00005337, 0.00001935, 0.00001475, 0.00001922, 0.00001611, 0.00001400, 0.00003058, 0.00004933, -0.00000665, -0.00004098, -0.00001370, -0.00000958, -0.00000801, -0.00000167, -0.00000640, 0.00000187, 0.00000277, -0.00000546, -0.00000124, },
		{ -0.00000263, 0.00000393, -0.00001025, -0.00000662, 0.00000178, 0.00000290, -0.00000635, -0.00001021, -0.00000088, -0.00000286, 0.00000013, -0.00000528, 0.00000195, -0.00000110, -0.00000042, -0.00000650, -0.00000206, -0.00000844, -0.00000399, -0.00002208, -0.00000960, -0.00007576, -0.00001544, 0.00010133, 0.00003635, 0.00003416, 0.00003109, 0.00002930, 0.00008190, 0.00005455, -0.00008782, -0.00002546, -0.00001675, -0.00000713, -0.00000241, -0.00000196, -0.00000936, -0.00000207, -0.00000826, 0.00000319, -0.00000569, -0.00000106, -0.00000439, 0.00000958, -0.00000300, -0.00000075, 0.00000144, -0.00000019, -0.00000223, 0.00000581, -0.00000323, 0.00000087, -0.00000159, 0.00000134, 0.00000970, 0.00000594, -0.00000435, -0.00000891, 0.00000929, -0.00001248, -0.00001168, -0.00000044, -0.00000094, -0.00000773, -0.00000295, -0.00000071, 0.00000059, -0.00000194, -0.00000269, -0.00000183, 0.00000070, 0.00000159, -0.00000137, -0.00000038, -0.00000426, -0.00000420, -0.00001289, -0.00000595, 0.00001234, -0.00001931, 0.00000029, -0.00000717, 0.00001404, 0.00000160, -0.00000244, 0.00000545, -0.00000151, -0.00000370, 0.00000409, -0.00000024, 0.00000407, 0.00000092, -0.00000288, -0.00000167, -0.00000197, 0.00000750, 0.00000057, -0.00000600, -0.00000493, -0.00000212, -0.00000151, -0.00000419, 0.00000220, -0.00000540, -0.00000847, -0.00000818, -0.00004316, -0.00001853, 0.00004742, 0.00003743, 0.00001029, 0.00002220, 0.00001591, 0.00001659, 0.00002137, 0.00005436, 0.00001229, -0.00004200, -0.00002201, -0.00000768, -0.00001101, -0.00000142, -0.00000604, -0.00000129, 0.00000353, -0.00000241, -0.00000412, -0.00000139, },
		{ 0.00000358, -0.00000432, -0.00001179, 0.00000054, 0.00000312, -0.00000126, -0.00001302, -0.00000185, -0.00000227, -0.00000069, -0.00000337, -0.00000183, 0.00000171, -0.00000134, -0.00000419, -0.00000374, -0.00000618, -0.00000466, -0.00001708, -0.00001168, -0.00005138, -0.00005576, 0.00009058, 0.00005072, 0.00003024, 0.00003076, 0.00002811, 0.00005076, 0.00009173, -0.00004851, -0.00005925, -0.00000871, -0.00001577, -0.00000068, -0.00000272, -0.00000634, -0.00000487, -0.00000583, -0.00000098, -0.00000166, -0.00000401, -0.00000311, 0.00000515, 0.00000360, -0.00000567, 0.00000366, -0.00000125, -0.00000048, 0.00000227, 0.00000208, -0.00000378, 0.00000208, -0.00000249, 0.00000822, 0.00000778, 0.00000126, -0.00001222, 0.00000634, -0.00000545, -0.00001277, -0.00000476, 0.00000113, -0.00000716, -0.00000451, -0.00000122, 0.00000051, -0.00000083, -0.00000245, -0.00000247, -0.00000028, 0.00000157, -0.00000020, -0.00000138, -0.00000157, -0.00000671, -0.00000543, -0.00001584, 0.00001387, -0.00001245, -0.00000708, -0.00000267, 0.00000249, 0.00001332, -0.00000756, 0.00000704, -0.00000084, -0.00000301, 0.00000080, 0.00000294, 0.00000032, 0.00000498, -0.00000450, 0.00000031, -0.00000499, 0.00000753, 0.00000195, -0.00000359, -0.00000679, -0.00000139, -0.00000281, -0.00000268, -0.00000200, 0.00000093, -0.00001187, -0.00000291, -0.00003123, -0.00004135, 0.00002754, 0.00005384, 0.00001379, 0.00002048, 0.00001824, 0.00001816, 0.00001617, 0.00005049, 0.00003386, -0.00003756, -0.00003147, -0.00000767, -0.00001225, -0.00000292, -0.00000491, -0.00000324, 0.00000160, 0.00000168, -0.00000611, -0.00000020, -0.00000447, },
		{ 0.00000304, -0.00001411, -0.00000286, 0.00000267, 0.00000279, -0.00001171, -0.00000664, 0.00000077, -0.00000342, -0.00000013, -0.00000530, 0.00000275, -0.00000101, -0.00000263, -0.00000472, -0.00000434, -0.00000593, -0.00001127, -0.00001615, -0.00002909, -0.00007632, 0.00006022, 0.00006992, 0.00002616, 0.00003125, 0.00002804, 0.00003158, 0.00008819, 0.00001508, -0.00008087, -0.00001473, -0.00001624, -0.00000435, -0.00000206, -0.00000398, -0.00000659, -0.00000409, -0.00000415, 0.00000071, -0.00000535, -0.00000238, 0.00000084, 0.00000834, -0.00000708, 0.00000270, -0.00000060, 0.00000024, 0.00000021, 0.00000442, -0.00000398, 0.00000092, -0.00000122, 0.00000370, 0.00000869, 0.00000644, -0.00001170, 0.00000133, -0.00000101, -0.00001051, -0.00000833, 0.00000011, -0.00000429, -0.00000698, -0.00000110, -0.00000054, 0.00000048, -0.00000212, -0.00000254, -0.00000133, 0.00000090, 0.00000121, -0.00000238, 0.00000060, -0.00000787, -0.00000227, -0.00001572, 0.00000393, 0.00000126, -0.00001681, 0.00000354, -0.00000779, 0.00001788, -0.00000449, 0.00000301, 0.00000279, -0.00000348, -0.00000108, 0.00000396, -0.00000077, 0.00000552, -0.00000293, -0.00000006, -0.00000505, 0.00000452, 0.00000491, -0.00000273, -0.00000626, -0.00000278, -0.00000213, -0.00000295, -0.00000346, 0.00000158, -0.00000799, -0.00000702, -0.00001440, -0.00005035, -0.00000352, 0.00006024, 0.00002693, 0.00001623, 0.00002139, 0.00001884, 0.00001548, 0.00004026, 0.00005278, -0.00002518, -0.00004128, -0.00000997, -0.00001220, -0.00000532, -0.00000390, -0.00000414, -0.00000083, 0.00000342, -0.00000432, -0.00000261, -0.00000208, -0.00000135, },
		{ -0.00000991, -0.00000937, 0.00000287, 0.00000354, -0.00000575, -0.00001321, 0.00000310, -0.00000469, 0.00000095, -0.00000544, 0.00000039, 0.00000110, -0.00000223, -0.00000447, -0.00000360, -0.00000648, -0.00000701, -0.00001856, -0.00001534, -0.00007595, 0.00001786, 0.00008539, 0.00002624, 0.00003130, 0.00002638, 0.00002641, 0.00006191, 0.00006718, -0.00006679, -0.00003941, -0.00000985, -0.00001149, -0.00000076, -0.00000322, -0.00000595, -0.00000436, -0.00000490, 0.00000038, -0.00000442, -0.00000302, -0.00000170, 0.00000930, -0.00000409, -0.00000114, 0.00000145, -0.00000045, 0.00000040, 0.00000349, -0.00000082, -0.00000228, 0.00000082, 0.00000102, 0.00000608, 0.00001137, -0.00000850, -0.00000340, 0.00000080, -0.00000801, -0.00000823, -0.00000431, -0.00000045, -0.00000897, -0.00000139, -0.00000145, 0.00000100, -0.00000128, -0.00000237, -0.00000208, -0.00000012, 0.00000191, -0.00000203, 0.00000048, -0.00000559, -0.00000466, -0.00000841, -0.00000916, 0.00001113, -0.00001938, 0.00000280, -0.00000873, 0.00001079, 0.00000639, -0.00000410, 0.00000705, -0.00000394, -0.00000172, 0.00000278, 0.00000063, 0.00000334, 0.00000048, -0.00000171, -0.00000302, -0.00000024, 0.00000816, -0.00000250, -0.00000433, -0.00000513, -0.00000059, -0.00000416, -0.00000292, -0.00000129, -0.00000185, -0.00001139, -0.00000506, -0.00004132, -0.00003621, 0.00004966, 0.00004611, 0.00001455, 0.00002279, 0.00001969, 0.00001765, 0.00002865, 0.00006335, -0.00000384, -0.00004968, -0.00001483, -0.00001164, -0.00000782, -0.00000345, -0.00000455, -0.00000224, 0.00000201, 0.00000007, -0.00000602, 0.00000068, -0.00000515, 0.00000561, },
		{ -0.00001548, 0.00000173, 0.00000265, 0.00000093, -0.00001630, 0.00000052, -0.00000189, -0.00000157, -0.00000221, -0.00000384, 0.00000306, -0.00000167, -0.00000388, -0.00000337, -0.00000633, -0.00000482, -0.00001772, -0.00001045, -0.00006132, -0.00002358, 0.00008724, 0.00003488, 0.00002829, 0.00002560, 0.00002669, 0.00003685, 0.00008502, -0.00001937, -0.00006617, -0.00000940, -0.00001437, -0.00000294, -0.00000197, -0.00000490, -0.00000477, -0.00000498, -0.00000078, -0.00000308, -0.00000371, -0.00000317, 0.00000758, 0.00000089, -0.00000485, 0.00000249, -0.00000083, 0.00000099, 0.00000196, 0.00000191, -0.00000335, -0.00000009, 0.00000207, 0.00000125, 0.00001362, -0.00000219, -0.00000777, 0.00000196, -0.00000784, -0.00000466, -0.00000913, 0.00000117, -0.00000838, -0.00000341, -0.00000116, 0.00000014, 0.00000011, -0.00000215, -0.00000236, -0.00000112, 0.00000149, -0.00000040, -0.00000124, -0.00000166, -0.00000853, -0.00000167, -0.00001559, 0.00000982, -0.00001168, -0.00000555, -0.00000190, -0.00000214, 0.00001661, -0.00000697, 0.00000712, -0.00000163, -0.00000297, 0.00000141, 0.00000208, 0.00000134, 0.00000279, -0.00000227, -0.00000123, -0.00000392, 0.00000900, -0.00000059, -0.00000347, -0.00000587, -0.00000072, -0.00000377, -0.00000330, -0.00000309, 0.00000041, -0.00000950, -0.00000659, -0.00002236, -0.00005618, 0.00002009, 0.00006310, 0.00002035, 0.00002148, 0.00002114, 0.00002053, 0.00002034, 0.00006271, 0.00002372, -0.00005293, -0.00002306, -0.00001085, -0.00001024, -0.00000351, -0.00000491, -0.00000261, -0.00000077, 0.00000348, -0.00000645, -0.00000010, -0.00000413, 0.00000173, -0.00000116, },
		{ -0.00000350, 0.00000290, 0.00000401, -0.00001264, -0.00000731, 0.00000330, -0.00000510, 0.00000107, -0.00000638, 0.00000238, 0.00000023, -0.00000364, -0.00000309, -0.00000583, -0.00000421, -0.00001452, -0.00001117, -0.00004190, -0.00005279, 0.00007096, 0.00005114, 0.00002319, 0.00002659, 0.00002601, 0.00002460, 0.00007185, 0.00003538, -0.00007060, -0.00002494, -0.00001001, -0.00000904, -0.00000011, -0.00000449, -0.00000411, -0.00000553, -0.00000149, -0.00000245, -0.00000373, -0.00000416, 0.00000466, 0.00000530, -0.00000606, 0.00000129, -0.00000015, 0.00000086, 0.00000150, 0.00000249, -0.00000154, -0.00000299, 0.00000387, -0.00000126, 0.00001027, 0.00000702, -0.00001128, 0.00000298, -0.00000899, -0.00000160, -0.00000982, -0.00000192, -0.00000464, -0.00000694, -0.00000007, -0.00000141, 0.00000127, -0.00000160, -0.00000232, -0.00000184, 0.00000031, 0.00000128, -0.00000250, 0.00000081, -0.00000942, -0.00000078, -0.00001243, -0.00000033, -0.00000006, -0.00001411, 0.00000426, -0.00001074, 0.00001748, -0.00000117, 0.00000181, 0.00000328, -0.00000479, 0.00000063, 0.00000227, 0.00000102, 0.00000287, -0.00000115, -0.00000060, -0.00000512, 0.00000650, 0.00000337, -0.00000410, -0.00000461, -0.00000274, -0.00000170, -0.00000493, -0.00000281, -0.00000152, -0.00000358, -0.00001163, -0.00000792, -0.00005406, -0.00002057, 0.00006643, 0.00003580, 0.00001901, 0.00002287, 0.00002248, 0.00001772, 0.00005245, 0.00005103, -0.00004624, -0.00003521, -0.00001006, -0.00001263, -0.00000395, -0.00000525, -0.00000272, -0.00000279, 0.00000360, -0.00000297, -0.00000392, -0.00000030, -0.00000400, 0.00000550, -0.00001571, },
		{ 0.00000385, 0.00000337, -0.00000463, -0.00001527, 0.00000545, -0.00000516, 0.00000079, -0.00000492, -0.00000137, 0.00000283, -0.00000327, -0.00000288, -0.00000505, -0.00000452, -0.00001054, -0.00001362, -0.00002523, -0.00006491, 0.00004021, 0.00006787, 0.00002087, 0.00002692, 0.00002414, 0.00002269, 0.00004696, 0.00006986, -0.00004272, -0.00005008, -0.00000629, -0.00001334, -0.00000094, -0.00000328, -0.00000354, -0.00000552, -0.00000251, -0.00000179, -0.00000390, -0.00000424, 0.00000111, 0.00000813, -0.00000475, -0.00000110, 0.00000055, 0.00000059, 0.00000162, 0.00000204, 0.00000075, -0.00000442, 0.00000264, 0.00000109, 0.00000243, 0.00001531, -0.00001092, 0.00000162, -0.00000776, -0.00000341, -0.00000447, -0.00000845, -0.00000028, -0.00000968, 0.00000002, -0.00000212, 0.00000121, -0.00000046, -0.00000215, -0.00000220, -0.00000092, 0.00000180, -0.00000201, 0.00000056, -0.00000651, -0.00000480, -0.00000470, -0.00001035, 0.00000615, -0.00001561, 0.00000359, -0.00001011, 0.00000830, 0.00000960, -0.00000399, 0.00000677, -0.00000471, -0.00000072, 0.00000196, 0.00000152, 0.00000203, 0.00000019, -0.00000028, -0.00000456, 0.00000224, 0.00000727, -0.00000426, -0.00000336, -0.00000447, -0.00000022, -0.00000565, -0.00000276, -0.00000370, 0.00000011, -0.00001168, -0.00000575, -0.00003470, -0.00005462, 0.00004644, 0.00005750, 0.00001923, 0.00002411, 0.00002302, 0.00002014, 0.00003791, 0.00007033, -0.00002668, -0.00004971, -0.00001057, -0.00001456, -0.00000499, -0.00000537, -0.00000319, -0.00000333, 0.00000109, 0.00000161, -0.00000675, 0.00000148, -0.00000589, 0.00000520, -0.00000866, -0.00001165, },
		},
		{ { -0.00000478, -0.00000386, 0.00001133, -0.00000195, -0.00000153, 0.00000027, 0.00000109, 0.00000133, -0.00000865, -0.00000306, 0.00000330, -0.00000069, -0.00000054, -0.00000121, 0.00000013, -0.00000180, -0.00000262, -0.00000362, 0.00000117, -0.00000539, -0.00000730, -0.00000980, -0.00000568, -0.00001258, -0.00006266, 0.00000229, 0.00008901, 0.00002433, 0.00002378, 0.00002073, 0.00001875, 0.00002820, 0.00006798, 0.00000916, -0.00006187, -0.00001606, -0.00001280, -0.00000786, -0.00000715, -0.00000654, -0.00000250, -0.00000241, -0.00001383, -0.00000168, -0.00000092, -0.00000109, -0.00000051, -0.00000442, -0.00000198, 0.00000217, -0.00002011, 0.00000232, 0.00000117, 0.00000180, -0.00000731, -0.00000036, 0.00001336, -0.00000277, -0.00000217, 0.00000003, -0.00000051, -0.00000104, -0.00000348, 0.00000098, 0.00000163, 0.00000098, -0.00000348, -0.00000104, -0.00000051, 0.00000003, -0.00000217, -0.00000277, 0.00001336, -0.00000036, -0.00000731, 0.00000180, 0.00000117, 0.00000232, -0.00002011, 0.00000217, -0.00000198, -0.00000442, -0.00000051, -0.00000109, -0.00000092, -0.00000168, -0.00001383, -0.00000241, -0.00000250, -0.00000654, -0.00000715, -0.00000786, -0.00001280, -0.00001606, -0.00006187, 0.00000916, 0.00006798, 0.00002820, 0.00001875, 0.00002073, 0.00002378, 0.00002433, 0.00008901, 0.00000229, -0.00006266, -0.00001258, -0.00000568, -0.00000980, -0.00000730, -0.00000539, 0.00000117, -0.00000362, -0.00000262, -0.00000180, 0.00000013, -0.00000121, -0.00000054, -0.00000069, 0.00000330, -0.00000306, -0.00000865, 0.00000133, 0.00000109, 0.00000027, -0.00000153, -0.00000195, 0.00001133, -0.00000386, },
		{ -0.00000561, 0.00000585, 0.00000669, -0.00000584, 0.00000205, -0.00000163, 0.00000432, -0.00000582, -0.00000673, 0.00000141, 0.00000171, -0.00000119, -0.00000059, -0.00000071, -0.00000035, -0.00000231, -0.00000362, -0.00000173, -0.00000042, -0.00000771, -0.00000763, -0.00000987, -0.00000277, -0.00004443, -0.00003986, 0.00006862, 0.00005558, 0.00001224, 0.00002773, 0.00001312, 0.00002275, 0.00004468, 0.00005696, -0.00004168, -0.00004092, -0.00000961, -0.00001181, -0.00000643, -0.00000642, -0.00000580, 0.00000039, -0.00001043, -0.00000831, 0.00000025, -0.00000117, -0.00000073, -0.00000157, -0.00000655, 0.00000542, -0.00001379, -0.00000928, 0.00000584, -0.00000116, -0.00000052, -0.00000890, 0.00001307, 0.00000321, -0.00000369, -0.00000020, -0.00000082, -0.00000032, -0.00000298, -0.00000120, 0.00000196, 0.00000145, -0.00000155, -0.00000311, 0.00000027, -0.00000063, 0.00000022, -0.00000562, 0.00000781, 0.00000908, -0.00000962, 0.00000040, -0.00000017, 0.00000600, -0.00001387, -0.00000981, 0.00000494, -0.00000706, -0.00000009, -0.00000179, -0.00000104, -0.00000016, -0.00000837, -0.00001090, 0.00000151, -0.00000647, -0.00000662, -0.00000747, -0.00001114, -0.00001060, -0.00003939, -0.00004999, 0.00006161, 0.00004839, 0.00002272, 0.00001754, 0.00002978, 0.00001387, 0.00007007, 0.00005968, -0.00006249, -0.00003293, -0.00000303, -0.00000997, -0.00000761, -0.00000787, -0.00000113, -0.00000030, -0.00000378, -0.00000225, -0.00000068, -0.00000053, -0.00000087, -0.00000110, 0.00000156, 0.00000170, -0.00000843, -0.00000348, 0.00000299, -0.00000101, 0.00000176, -0.00000543, 0.00000856, 0.00000294, -0.00000565, },
		{ -0.00000180, 0.00001202, -0.00000450, 0.00000001, -0.00000127, 0.00000287, -0.00000044, -0.00000844, -0.00000207, 0.00000279, -0.00000047, -0.00000046, -0.00000124, 0.00000037, -0.00000194, -0.00000215, -0.00000480, 0.00000109, -0.00000501, -0.00000677, -0.00001099, -0.00000259, -0.00002311, -0.00005503, 0.00002484, 0.00007710, 0.00001860, 0.00002254, 0.00001767, 0.00001574, 0.00002901, 0.00006009, 0.00001314, -0.00005811, -0.00001741, -0.00001196, -0.00000817, -0.00000613, -0.00000618, -0.00000272, -0.00000316, -0.00001298, -0.00000176, -0.00000064, -0.00000069, -0.00000051, -0.00000575, 0.00000020, -0.00000174, -0.00001874, 0.00000568, -0.00000149, 0.00000257, -0.00000945, 0.00000498, 0.00001067, -0.00000369, -0.00000062, -0.00000109, -0.00000048, -0.00000157, -0.00000297, 0.00000145, 0.00000151, 0.00000053, -0.00000387, -0.00000042, -0.00000055, 0.00000110, -0.00000428, -0.00000008, 0.00001369, -0.00000462, -0.00000515, 0.00000149, 0.00000347, -0.00000239, -0.00001901, 0.00000462, -0.00000412, -0.00000243, -0.00000114, -0.00000135, -0.00000118, -0.00000174, -0.00001433, -0.00000177, -0.00000251, -0.00000673, -0.00000823, -0.00000746, -0.00001370, -0.00001457, -0.00006563, 0.00000431, 0.00007666, 0.00002687, 0.00002280, 0.00002401, 0.00002450, 0.00003395, 0.00009630, -0.00002639, -0.00006050, -0.00000584, -0.00000830, -0.00000852, -0.00000801, -0.00000564, 0.00000147, -0.00000261, -0.00000308, -0.00000181, 0.00000000, -0.00000124, -0.00000057, -0.00000090, 0.00000377, -0.00000442, -0.00000782, 0.00000210, -0.00000022, 0.00000166, -0.00000284, 0.00000083, 0.00000945, -0.00000504, -0.00000486, },
		{ 0.00000864, 0.00000376, -0.00000515, 0.00000156, -0.00000130, 0.00000461, -0.00000762, -0.00000480, 0.00000082, 0.00000185, -0.00000112, -0.00000055, -0.00000058, -0.00000037, -0.00000206, -0.00000398, -0.00000262, -0.00000009, -0.00000684, -0.00000892, -0.00000739, -0.00000731, -0.00004993, -0.00001535, 0.00006862, 0.00004396, 0.00001157, 0.00002518, 0.00000977, 0.00002280, 0.00004098, 0.00005276, -0.00003459, -0.00004162, -0.00000914, -0.00001197, -0.00000582, -0.00000581, -0.00000543, -0.00000075, -0.00000991, -0.00000795, 0.00000028, -0.00000085, -0.00000027, -0.00000294, -0.00000528, 0.00000440, -0.00001674, -0.00000367, 0.00000348, -0.00000057, -0.00000255, -0.00000614, 0.00001495, -0.00000096, -0.00000102, -0.00000139, -0.00000071, -0.00000104, -0.00000264, -0.00000079, 0.00000229, 0.00000087, -0.00000187, -0.00000300, 0.00000068, -0.00000014, -0.00000030, -0.00000611, 0.00001159, 0.00000408, -0.00000886, 0.00000067, 0.00000188, 0.00000385, -0.00001650, -0.00000577, 0.00000333, -0.00000657, 0.00000109, -0.00000320, -0.00000051, -0.00000101, -0.00000799, -0.00001145, 0.00000257, -0.00000735, -0.00000646, -0.00000884, -0.00001006, -0.00001204, -0.00003715, -0.00005944, 0.00006675, 0.00005186, 0.00002326, 0.00002266, 0.00003133, 0.00001741, 0.00008616, 0.00003964, -0.00007775, -0.00001966, -0.00000654, -0.00000806, -0.00000886, -0.00000723, -0.00000220, 0.00000156, -0.00000439, -0.00000196, -0.00000124, -0.00000014, -0.00000130, -0.00000088, 0.00000140, 0.00000159, -0.00000951, -0.00000118, 0.00000106, 0.00000033, 0.00000087, -0.00000408, 0.00000916, 0.00000035, -0.00000548, -0.00000509, },
		{ 0.00001115, -0.00000624, 0.00000129, -0.00000253, 0.00000459, -0.00000278, -0.00000728, -0.00000155, 0.00000228, -0.00000015, -0.00000051, -0.00000113, 0.00000046, -0.00000194, -0.00000199, -0.00000574, 0.00000084, -0.00000399, -0.00000732, -0.00001057, -0.00000156, -0.00003384, -0.00003969, 0.00003906, 0.00006415, 0.00001541, 0.00002099, 0.00001474, 0.00001398, 0.00002870, 0.00005370, 0.00001560, -0.00005377, -0.00001907, -0.00001076, -0.00000876, -0.00000488, -0.00000594, -0.00000295, -0.00000411, -0.00001159, -0.00000230, 0.00000005, -0.00000062, -0.00000076, -0.00000642, 0.00000200, -0.00000661, -0.00001468, 0.00000659, -0.00000317, 0.00000286, -0.00001066, 0.00001031, 0.00000623, -0.00000273, -0.00000011, -0.00000191, -0.00000060, -0.00000192, -0.00000238, 0.00000187, 0.00000119, 0.00000012, -0.00000409, 0.00000019, -0.00000047, 0.00000170, -0.00000620, 0.00000366, 0.00001171, -0.00000731, -0.00000342, 0.00000216, 0.00000434, -0.00000711, -0.00001609, 0.00000545, -0.00000575, -0.00000000, -0.00000257, -0.00000108, -0.00000184, -0.00000156, -0.00001476, -0.00000105, -0.00000294, -0.00000649, -0.00000966, -0.00000665, -0.00001503, -0.00001251, -0.00006996, -0.00000068, 0.00008521, 0.00002602, 0.00002730, 0.00002787, 0.00002453, 0.00004877, 0.00009428, -0.00005612, -0.00004964, -0.00000457, -0.00000866, -0.00000842, -0.00000798, -0.00000632, 0.00000234, -0.00000210, -0.00000326, -0.00000213, 0.00000015, -0.00000149, -0.00000043, -0.00000118, 0.00000411, -0.00000589, -0.00000615, 0.00000175, -0.00000073, 0.00000258, -0.00000360, 0.00000331, 0.00000682, -0.00000545, -0.00000506, 0.00000107, },
		{ 0.00000029, -0.00000353, 0.00000037, -0.00000006, 0.00000367, -0.00000844, -0.00000304, 0.00000008, 0.00000200, -0.00000097, -0.00000066, -0.00000024, -0.00000063, -0.00000165, -0.00000467, -0.00000317, 0.00000028, -0.00000597, -0.00001052, -0.00000369, -0.00001611, -0.00004730, 0.00000608, 0.00006312, 0.00003494, 0.00001171, 0.00002182, 0.00000813, 0.00002190, 0.00003811, 0.00004839, -0.00002821, -0.00004185, -0.00000889, -0.00001190, -0.00000539, -0.00000502, -0.00000518, -0.00000210, -0.00000904, -0.00000769, 0.00000032, -0.00000043, -0.00000026, -0.00000403, -0.00000345, 0.00000165, -0.00001756, 0.00000147, -0.00000011, 0.00000130, -0.00000556, -0.00000134, 0.00001408, -0.00000359, 0.00000145, -0.00000289, -0.00000043, -0.00000179, -0.00000213, -0.00000033, 0.00000235, 0.00000033, -0.00000217, -0.00000266, 0.00000096, 0.00000047, -0.00000157, -0.00000490, 0.00001364, -0.00000084, -0.00000734, 0.00000101, 0.00000400, 0.00000000, -0.00001690, -0.00000242, 0.00000100, -0.00000494, 0.00000143, -0.00000440, 0.00000001, -0.00000199, -0.00000737, -0.00001192, 0.00000332, -0.00000811, -0.00000631, -0.00001014, -0.00000905, -0.00001339, -0.00003486, -0.00006931, 0.00007163, 0.00005561, 0.00002424, 0.00002872, 0.00003177, 0.00002482, 0.00010028, 0.00000932, -0.00008171, -0.00000924, -0.00001052, -0.00000563, -0.00001048, -0.00000631, -0.00000321, 0.00000342, -0.00000512, -0.00000172, -0.00000178, 0.00000023, -0.00000169, -0.00000069, 0.00000129, 0.00000103, -0.00000967, 0.00000054, -0.00000091, 0.00000205, -0.00000034, -0.00000207, 0.00000851, -0.00000160, -0.00000534, -0.00000382, 0.00001071, },
		{ -0.00000665, 0.00000185, -0.00000305, 0.00000565, -0.00000508, -0.00000549, -0.00000150, 0.00000180, 0.00000032, -0.00000079, -0.00000072, 0.00000023, -0.00000165, -0.00000231, -0.00000625, 0.00000034, -0.00000227, -0.00000924, -0.00000777, -0.00000461, -0.00004076, -0.00002084, 0.00004563, 0.00005192, 0.00001405, 0.00001887, 0.00001243, 0.00001293, 0.00002746, 0.00004882, 0.00001646, -0.00004872, -0.00002117, -0.00000906, -0.00000974, -0.00000330, -0.00000591, -0.00000321, -0.00000514, -0.00000978, -0.00000328, 0.00000120, -0.00000110, -0.00000095, -0.00000646, 0.00000290, -0.00001153, -0.00000848, 0.00000460, -0.00000287, 0.00000170, -0.00000990, 0.00001425, 0.00000129, -0.00000039, -0.00000068, -0.00000227, -0.00000093, -0.00000205, -0.00000169, 0.00000213, 0.00000078, -0.00000029, -0.00000407, 0.00000070, -0.00000016, 0.00000156, -0.00000716, 0.00000739, 0.00000809, -0.00000850, -0.00000202, 0.00000365, 0.00000333, -0.00001080, -0.00001216, 0.00000465, -0.00000637, 0.00000233, -0.00000456, -0.00000028, -0.00000299, -0.00000098, -0.00001524, -0.00000028, -0.00000375, -0.00000582, -0.00001139, -0.00000548, -0.00001676, -0.00000990, -0.00007494, -0.00000557, 0.00009309, 0.00002638, 0.00003185, 0.00003229, 0.00002452, 0.00006870, 0.00007884, -0.00008013, -0.00003414, -0.00000793, -0.00000655, -0.00000984, -0.00000692, -0.00000759, 0.00000382, -0.00000215, -0.00000313, -0.00000276, 0.00000055, -0.00000193, -0.00000012, -0.00000152, 0.00000420, -0.00000710, -0.00000405, 0.00000049, -0.00000037, 0.00000292, -0.00000369, 0.00000509, 0.00000389, -0.00000526, -0.00000517, 0.00000447, 0.00000862, },
		{ -0.00000145, -0.00000120, 0.00000174, 0.00000165, -0.00000810, -0.00000175, -0.00000071, 0.00000225, -0.00000089, -0.00000072, 0.00000008, -0.00000090, -0.00000138, -0.00000536, -0.00000370, 0.00000119, -0.00000602, -0.00001094, -0.00000101, -0.00002651, -0.00003722, 0.00002158, 0.00005521, 0.00002786, 0.00001255, 0.00001753, 0.00000854, 0.00001935, 0.00003680, 0.00004317, -0.00002193, -0.00004215, -0.00000836, -0.00001201, -0.00000479, -0.00000435, -0.00000486, -0.00000381, -0.00000752, -0.00000795, 0.00000083, -0.00000041, -0.00000033, -0.00000490, -0.00000130, -0.00000279, -0.00001546, 0.00000456, -0.00000337, 0.00000338, -0.00000867, 0.00000469, 0.00001059, -0.00000406, 0.00000295, -0.00000422, -0.00000021, -0.00000246, -0.00000150, 0.00000014, 0.00000216, -0.00000013, -0.00000242, -0.00000215, 0.00000117, 0.00000090, -0.00000307, -0.00000228, 0.00001356, -0.00000489, -0.00000558, 0.00000191, 0.00000520, -0.00000449, -0.00001528, -0.00000026, -0.00000152, -0.00000225, 0.00000058, -0.00000485, -0.00000000, -0.00000262, -0.00000687, -0.00001202, 0.00000341, -0.00000832, -0.00000658, -0.00001089, -0.00000866, -0.00001399, -0.00003334, -0.00007859, 0.00007511, 0.00006073, 0.00002497, 0.00003632, 0.00003024, 0.00003862, 0.00010679, -0.00002677, -0.00007372, -0.00000480, -0.00001241, -0.00000435, -0.00001130, -0.00000590, -0.00000361, 0.00000479, -0.00000557, -0.00000181, -0.00000201, 0.00000032, -0.00000183, -0.00000068, 0.00000126, 0.00000009, -0.00000886, 0.00000129, -0.00000243, 0.00000376, -0.00000158, 0.00000021, 0.00000679, -0.00000278, -0.00000535, -0.00000160, 0.00001142, -0.00000297, },
		{ 0.00000140, -0.00000250, 0.00000554, -0.00000671, -0.00000353, -0.00000180, 0.00000139, 0.00000089, -0.00000129, -0.00000002, -0.00000031, -0.00000111, -0.00000306, -0.00000641, -0.00000013, -0.00000040, -0.00001185, -0.00000315, -0.00001206, -0.00004108, -0.00000299, 0.00004691, 0.00004069, 0.00001445, 0.00001558, 0.00001161, 0.00001154, 0.00002613, 0.00004478, 0.00001625, -0.00004336, -0.00002337, -0.00000717, -0.00001086, -0.00000161, -0.00000591, -0.00000375, -0.00000583, -0.00000803, -0.00000428, 0.00000247, -0.00000197, -0.00000105, -0.00000582, 0.00000231, -0.00001520, -0.00000151, 0.00000031, -0.00000041, -0.00000130, -0.00000654, 0.00001557, -0.00000274, 0.00000238, -0.00000198, -0.00000224, -0.00000143, -0.00000199, -0.00000089, 0.00000210, 0.00000042, -0.00000076, -0.00000376, 0.00000104, 0.00000030, 0.00000064, -0.00000668, 0.00001011, 0.00000375, -0.00000861, -0.00000052, 0.00000524, 0.00000072, -0.00001288, -0.00000795, 0.00000240, -0.00000551, 0.00000380, -0.00000644, 0.00000060, -0.00000431, -0.00000022, -0.00001556, 0.00000023, -0.00000455, -0.00000512, -0.00001299, -0.00000445, -0.00001831, -0.00000743, -0.00007989, -0.00001094, 0.00010063, 0.00002793, 0.00003678, 0.00003642, 0.00002660, 0.00009099, 0.00004856, -0.00009226, -0.00001960, -0.00001310, -0.00000332, -0.00001199, -0.00000530, -0.00000907, 0.00000554, -0.00000248, -0.00000294, -0.00000342, 0.00000094, -0.00000234, 0.00000016, -0.00000179, 0.00000388, -0.00000766, -0.00000206, -0.00000129, 0.00000071, 0.00000272, -0.00000313, 0.00000599, 0.00000108, -0.00000474, -0.00000491, 0.00000787, 0.00000470, -0.00000554, },
		{ -0.00000256, 0.00000349, -0.00000096, -0.00000672, -0.00000105, -0.00000145, 0.00000264, -0.00000103, -0.00000053, 0.00000015, -0.00000094, -0.00000149, -0.00000577, -0.00000439, 0.00000285, -0.00000757, -0.00000901, -0.00000167, -0.00003482, -0.00002278, 0.00003087, 0.00004662, 0.00002248, 0.00001358, 0.00001286, 0.00001058, 0.00001519, 0.00003721, 0.00003684, -0.00001543, -0.00004279, -0.00000731, -0.00001254, -0.00000384, -0.00000395, -0.00000444, -0.00000579, -0.00000539, -0.00000882, 0.00000198, -0.00000109, -0.00000014, -0.00000556, 0.00000074, -0.00000835, -0.00001035, 0.00000446, -0.00000471, 0.00000430, -0.00001054, 0.00001053, 0.00000544, -0.00000248, 0.00000310, -0.00000501, -0.00000022, -0.00000293, -0.00000075, 0.00000052, 0.00000180, -0.00000053, -0.00000256, -0.00000157, 0.00000142, 0.00000083, -0.00000412, 0.00000098, 0.00001158, -0.00000780, -0.00000368, 0.00000337, 0.00000491, -0.00000865, -0.00001213, 0.00000037, -0.00000354, 0.00000095, -0.00000137, -0.00000438, -0.00000081, -0.00000262, -0.00000671, -0.00001165, 0.00000274, -0.00000785, -0.00000741, -0.00001093, -0.00000909, -0.00001360, -0.00003304, -0.00008661, 0.00007623, 0.00006824, 0.00002487, 0.00004582, 0.00002658, 0.00006045, 0.00009946, -0.00006109, -0.00005700, -0.00000675, -0.00001115, -0.00000511, -0.00001064, -0.00000645, -0.00000309, 0.00000537, -0.00000555, -0.00000234, -0.00000187, 0.00000007, -0.00000164, -0.00000090, 0.00000123, -0.00000099, -0.00000724, 0.00000099, -0.00000320, 0.00000518, -0.00000261, 0.00000242, 0.00000429, -0.00000315, -0.00000553, 0.00000158, 0.00001027, -0.00000518, 0.00000043, },
		{ -0.00000091, 0.00000406, -0.00000719, -0.00000185, -0.00000228, 0.00000111, 0.00000143, -0.00000182, 0.00000077, -0.00000098, -0.00000055, -0.00000397, -0.00000649, -0.00000005, 0.00000066, -0.00001373, 0.00000141, -0.00002218, -0.00003441, 0.00001071, 0.00004543, 0.00003023, 0.00001662, 0.00001068, 0.00001301, 0.00000877, 0.00002573, 0.00004062, 0.00001581, -0.00003835, -0.00002505, -0.00000558, -0.00001169, -0.00000020, -0.00000565, -0.00000484, -0.00000572, -0.00000690, -0.00000472, 0.00000325, -0.00000273, -0.00000125, -0.00000443, -0.00000023, -0.00001631, 0.00000436, -0.00000475, 0.00000339, -0.00000567, -0.00000078, 0.00001372, -0.00000473, 0.00000453, -0.00000342, -0.00000206, -0.00000196, -0.00000176, -0.00000002, 0.00000175, 0.00000020, -0.00000131, -0.00000312, 0.00000119, 0.00000075, -0.00000077, -0.00000470, 0.00001114, -0.00000047, -0.00000805, 0.00000148, 0.00000606, -0.00000275, -0.00001326, -0.00000410, -0.00000094, -0.00000295, 0.00000370, -0.00000739, 0.00000085, -0.00000518, 0.00000023, -0.00001527, -0.00000005, -0.00000476, -0.00000498, -0.00001378, -0.00000436, -0.00001878, -0.00000618, -0.00008356, -0.00001809, 0.00010895, 0.00002987, 0.00004310, 0.00003868, 0.00003425, 0.00010983, 0.00000627, -0.00008931, -0.00001094, -0.00001668, -0.00000108, -0.00001341, -0.00000413, -0.00000999, 0.00000683, -0.00000259, -0.00000309, -0.00000374, 0.00000099, -0.00000240, 0.00000016, -0.00000185, 0.00000305, -0.00000731, -0.00000071, -0.00000306, 0.00000221, 0.00000211, -0.00000205, 0.00000590, -0.00000129, -0.00000414, -0.00000390, 0.00001052, 0.00000006, -0.00000313, -0.00000003, },
		{ 0.00000449, -0.00000344, -0.00000471, -0.00000093, -0.00000207, 0.00000316, -0.00000143, -0.00000001, -0.00000013, -0.00000068, -0.00000205, -0.00000588, -0.00000506, 0.00000491, -0.00001029, -0.00000469, -0.00000679, -0.00003783, -0.00000803, 0.00003561, 0.00003770, 0.00001919, 0.00001375, 0.00000897, 0.00001306, 0.00001031, 0.00003866, 0.00002988, -0.00000900, -0.00004350, -0.00000594, -0.00001332, -0.00000270, -0.00000367, -0.00000419, -0.00000761, -0.00000309, -0.00000994, 0.00000349, -0.00000234, 0.00000030, -0.00000584, 0.00000199, -0.00001389, -0.00000311, 0.00000105, -0.00000323, 0.00000294, -0.00000988, 0.00001451, 0.00000015, 0.00000034, 0.00000205, -0.00000518, -0.00000052, -0.00000317, 0.00000012, 0.00000065, 0.00000142, -0.00000097, -0.00000245, -0.00000104, 0.00000173, 0.00000014, -0.00000413, 0.00000395, 0.00000833, -0.00000966, -0.00000138, 0.00000486, 0.00000317, -0.00001184, -0.00000798, -0.00000076, -0.00000424, 0.00000372, -0.00000371, -0.00000335, -0.00000216, -0.00000214, -0.00000672, -0.00001108, 0.00000160, -0.00000698, -0.00000844, -0.00001067, -0.00000989, -0.00001269, -0.00003356, -0.00009361, 0.00007496, 0.00007835, 0.00002415, 0.00005651, 0.00002236, 0.00008938, 0.00007391, -0.00008517, -0.00003765, -0.00001266, -0.00000764, -0.00000743, -0.00000874, -0.00000778, -0.00000187, 0.00000532, -0.00000521, -0.00000310, -0.00000155, -0.00000032, -0.00000129, -0.00000123, 0.00000105, -0.00000186, -0.00000525, -0.00000017, -0.00000311, 0.00000611, -0.00000330, 0.00000430, 0.00000133, -0.00000283, -0.00000573, 0.00000544, 0.00000710, -0.00000567, 0.00000142, -0.00000313, },
		{ 0.00000142, -0.00000636, -0.00000077, -0.00000279, 0.00000101, 0.00000176, -0.00000216, 0.00000140, -0.00000152, -0.00000024, -0.00000475, -0.00000672, 0.00000094, 0.00000010, -0.00001341, 0.00000345, -0.00003170, -0.00002287, 0.00001939, 0.00004270, 0.00002064, 0.00002004, 0.00000447, 0.00001658, 0.00000424, 0.00002681, 0.00003571, 0.00001578, -0.00003423, -0.00002576, -0.00000474, -0.00001188, 0.00000066, -0.00000500, -0.00000657, -0.00000461, -0.00000674, -0.00000418, 0.00000303, -0.00000292, -0.00000176, -0.00000238, -0.00000472, -0.00001399, 0.00000732, -0.00000859, 0.00000691, -0.00001006, 0.00000625, 0.00000910, -0.00000421, 0.00000530, -0.00000443, -0.00000201, -0.00000239, -0.00000142, 0.00000084, 0.00000111, 0.00000014, -0.00000191, -0.00000221, 0.00000119, 0.00000097, -0.00000215, -0.00000166, 0.00001033, -0.00000407, -0.00000695, 0.00000402, 0.00000552, -0.00000624, -0.00001211, -0.00000116, -0.00000467, 0.00000101, 0.00000180, -0.00000690, -0.00000008, -0.00000508, -0.00000009, -0.00001401, -0.00000151, -0.00000393, -0.00000584, -0.00001325, -0.00000576, -0.00001748, -0.00000709, -0.00008481, -0.00002839, 0.00011939, 0.00003121, 0.00005185, 0.00003768, 0.00005108, 0.00011729, -0.00004072, -0.00007255, -0.00001046, -0.00001641, -0.00000148, -0.00001285, -0.00000432, -0.00000968, 0.00000713, -0.00000205, -0.00000386, -0.00000347, 0.00000046, -0.00000190, -0.00000030, -0.00000159, 0.00000173, -0.00000595, -0.00000034, -0.00000436, 0.00000382, 0.00000123, -0.00000060, 0.00000484, -0.00000294, -0.00000370, -0.00000186, 0.00001160, -0.00000429, -0.00000007, -0.00000201, 0.00000131, },
		{ -0.00000507, -0.00000260, -0.00000121, -0.00000251, 0.00000369, -0.00000197, 0.00000074, -0.00000068, -0.00000022, -0.00000290, -0.00000588, -0.00000523, 0.00000645, -0.00001296, 0.00000064, -0.00001557, -0.00003420, 0.00000367, 0.00003798, 0.00002816, 0.00001849, 0.00001203, 0.00000711, 0.00001451, 0.00000605, 0.00003997, 0.00002324, -0.00000339, -0.00004364, -0.00000481, -0.00001386, -0.00000179, -0.00000312, -0.00000455, -0.00000863, -0.00000134, -0.00001061, 0.00000466, -0.00000360, 0.00000067, -0.00000541, 0.00000163, -0.00001784, 0.00000448, -0.00000456, 0.00000082, -0.00000092, -0.00000606, 0.00001534, -0.00000369, 0.00000324, 0.00000036, -0.00000489, -0.00000100, -0.00000322, 0.00000110, 0.00000044, 0.00000119, -0.00000155, -0.00000200, -0.00000067, 0.00000204, -0.00000110, -0.00000281, 0.00000583, 0.00000453, -0.00001070, 0.00000161, 0.00000558, 0.00000055, -0.00001379, -0.00000336, -0.00000363, -0.00000295, 0.00000498, -0.00000540, -0.00000264, -0.00000337, -0.00000176, -0.00000635, -0.00001096, 0.00000071, -0.00000643, -0.00000888, -0.00001102, -0.00001003, -0.00001247, -0.00003363, -0.00010082, 0.00007234, 0.00009019, 0.00002412, 0.00006640, 0.00002132, 0.00012060, 0.00003000, -0.00009255, -0.00002245, -0.00001849, -0.00000425, -0.00000971, -0.00000672, -0.00000904, -0.00000064, 0.00000519, -0.00000502, -0.00000365, -0.00000148, -0.00000046, -0.00000113, -0.00000140, 0.00000045, -0.00000211, -0.00000342, -0.00000179, -0.00000226, 0.00000648, -0.00000355, 0.00000558, -0.00000170, -0.00000207, -0.00000559, 0.00000936, 0.00000225, -0.00000421, 0.00000111, -0.00000251, 0.00000416, },
		{ -0.00000443, -0.00000041, -0.00000319, 0.00000109, 0.00000178, -0.00000215, 0.00000166, -0.00000174, -0.00000034, -0.00000527, -0.00000704, 0.00000271, -0.00000221, -0.00001018, 0.00000111, -0.00003710, -0.00001010, 0.00002426, 0.00003888, 0.00001274, 0.00002344, -0.00000191, 0.00002129, -0.00000142, 0.00002902, 0.00003016, 0.00001617, -0.00003101, -0.00002546, -0.00000466, -0.00001144, 0.00000102, -0.00000414, -0.00000860, -0.00000281, -0.00000736, -0.00000275, 0.00000177, -0.00000237, -0.00000251, -0.00000017, -0.00001050, -0.00000835, 0.00000636, -0.00000948, 0.00000841, -0.00001271, 0.00001274, 0.00000303, -0.00000159, 0.00000448, -0.00000468, -0.00000227, -0.00000260, -0.00000094, 0.00000155, 0.00000035, 0.00000014, -0.00000244, -0.00000114, 0.00000107, 0.00000077, -0.00000293, 0.00000170, 0.00000797, -0.00000692, -0.00000513, 0.00000661, 0.00000355, -0.00000914, -0.00000967, 0.00000029, -0.00000780, 0.00000545, -0.00000140, -0.00000508, -0.00000221, -0.00000392, -0.00000121, -0.00001185, -0.00000409, -0.00000209, -0.00000764, -0.00001152, -0.00000859, -0.00001445, -0.00001027, -0.00008330, -0.00004244, 0.00013262, 0.00003169, 0.00006309, 0.00003359, 0.00007877, 0.00010571, -0.00008222, -0.00004768, -0.00001694, -0.00001220, -0.00000474, -0.00001002, -0.00000611, -0.00000800, 0.00000631, -0.00000083, -0.00000522, -0.00000266, -0.00000057, -0.00000089, -0.00000120, -0.00000113, 0.00000016, -0.00000380, -0.00000103, -0.00000487, 0.00000522, 0.00000026, 0.00000104, 0.00000292, -0.00000375, -0.00000349, 0.00000130, 0.00001047, -0.00000728, 0.00000272, -0.00000382, 0.00000345, -0.00000177, },
		{ -0.00000092, -0.00000168, -0.00000272, 0.00000408, -0.00000246, 0.00000149, -0.00000128, 0.00000018, -0.00000378, -0.00000603, -0.00000441, 0.00000645, -0.00001392, 0.00000460, -0.00002534, -0.00002509, 0.00001094, 0.00003946, 0.00001802, 0.00002035, 0.00000805, 0.00000794, 0.00001391, 0.00000352, 0.00004004, 0.00001790, 0.00000060, -0.00004249, -0.00000456, -0.00001363, -0.00000156, -0.00000199, -0.00000584, -0.00000835, -0.00000077, -0.00001014, 0.00000472, -0.00000415, 0.00000049, -0.00000396, -0.00000096, -0.00001867, 0.00001017, -0.00001027, 0.00000606, -0.00000638, 0.00000045, 0.00001263, -0.00000500, 0.00000508, -0.00000124, -0.00000450, -0.00000152, -0.00000311, 0.00000217, -0.00000015, 0.00000120, -0.00000231, -0.00000113, -0.00000053, 0.00000224, -0.00000256, -0.00000027, 0.00000618, 0.00000077, -0.00001099, 0.00000534, 0.00000488, -0.00000222, -0.00001451, 0.00000117, -0.00000783, 0.00000049, 0.00000399, -0.00000546, -0.00000315, -0.00000358, -0.00000222, -0.00000494, -0.00001201, 0.00000084, -0.00000699, -0.00000786, -0.00001298, -0.00000834, -0.00001435, -0.00003163, -0.00011004, 0.00007011, 0.00010220, 0.00002668, 0.00007283, 0.00002867, 0.00014557, -0.00002675, -0.00008130, -0.00001627, -0.00002050, -0.00000353, -0.00001001, -0.00000600, -0.00000916, -0.00000030, 0.00000563, -0.00000550, -0.00000352, -0.00000209, 0.00000007, -0.00000152, -0.00000116, -0.00000073, -0.00000141, -0.00000228, -0.00000336, -0.00000090, 0.00000627, -0.00000330, 0.00000607, -0.00000442, -0.00000118, -0.00000468, 0.00001242, -0.00000339, -0.00000114, -0.00000043, -0.00000071, 0.00000233, -0.00000536, },
		{ -0.00000069, -0.00000341, 0.00000130, 0.00000151, -0.00000176, 0.00000151, -0.00000163, -0.00000085, -0.00000561, -0.00000708, 0.00000453, -0.00000541, -0.00000475, -0.00000581, -0.00003617, 0.00000031, 0.00002748, 0.00003329, 0.00000773, 0.00002519, -0.00000675, 0.00002544, -0.00000675, 0.00003120, 0.00002483, 0.00001636, -0.00002809, -0.00002467, -0.00000492, -0.00001076, 0.00000128, -0.00000360, -0.00001022, -0.00000105, -0.00000809, -0.00000110, -0.00000003, -0.00000142, -0.00000310, 0.00000131, -0.00001612, -0.00000066, 0.00000176, -0.00000676, 0.00000674, -0.00001214, 0.00001680, -0.00000269, 0.00000203, 0.00000246, -0.00000421, -0.00000284, -0.00000259, -0.00000031, 0.00000195, -0.00000033, 0.00000005, -0.00000272, -0.00000006, 0.00000084, 0.00000011, -0.00000268, 0.00000457, 0.00000461, -0.00000907, -0.00000229, 0.00000845, 0.00000064, -0.00001119, -0.00000622, -0.00000019, -0.00000922, 0.00000903, -0.00000477, -0.00000278, -0.00000491, -0.00000223, -0.00000261, -0.00000939, -0.00000715, 0.00000010, -0.00000963, -0.00000944, -0.00001186, -0.00001077, -0.00001467, -0.00007998, -0.00005948, 0.00014799, 0.00003242, 0.00007514, 0.00002931, 0.00011500, 0.00007084, -0.00010821, -0.00002289, -0.00002625, -0.00000626, -0.00000942, -0.00000588, -0.00000880, -0.00000555, 0.00000483, 0.00000066, -0.00000673, -0.00000175, -0.00000169, 0.00000027, -0.00000225, -0.00000073, -0.00000126, -0.00000129, -0.00000255, -0.00000447, 0.00000617, -0.00000062, 0.00000265, 0.00000037, -0.00000376, -0.00000337, 0.00000527, 0.00000696, -0.00000809, 0.00000437, -0.00000473, 0.00000469, -0.00000468, -0.00000192, },
		{ -0.00000213, -0.00000270, 0.00000418, -0.00000265, 0.00000197, -0.00000167, 0.00000029, -0.00000447, -0.00000641, -0.00000246, 0.00000435, -0.00001187, 0.00000477, -0.00003251, -0.00001366, 0.00001458, 0.00004010, 0.00000821, 0.00002373, 0.00000254, 0.00001099, 0.00001130, 0.00000296, 0.00003845, 0.00001431, 0.00000264, -0.00003972, -0.00000551, -0.00001238, -0.00000217, -0.00000028, -0.00000798, -0.00000676, -0.00000156, -0.00000828, 0.00000328, -0.00000352, -0.00000051, -0.00000153, -0.00000580, -0.00001556, 0.00001206, -0.00001374, 0.00001033, -0.00001158, 0.00000812, 0.00000712, -0.00000355, 0.00000518, -0.00000216, -0.00000432, -0.00000190, -0.00000287, 0.00000320, -0.00000100, 0.00000140, -0.00000319, 0.00000010, -0.00000066, 0.00000219, -0.00000376, 0.00000296, 0.00000501, -0.00000269, -0.00001032, 0.00000935, 0.00000261, -0.00000455, -0.00001404, 0.00000488, -0.00001246, 0.00000552, 0.00000075, -0.00000353, -0.00000535, -0.00000231, -0.00000393, -0.00000214, -0.00001459, 0.00000241, -0.00000904, -0.00000498, -0.00001701, -0.00000423, -0.00001917, -0.00002645, -0.00012265, 0.00006972, 0.00011315, 0.00003327, 0.00007398, 0.00004906, 0.00015384, -0.00008576, -0.00005516, -0.00002016, -0.00001700, -0.00000683, -0.00000720, -0.00000750, -0.00000746, -0.00000138, 0.00000703, -0.00000688, -0.00000250, -0.00000356, 0.00000142, -0.00000266, -0.00000038, -0.00000251, 0.00000029, -0.00000215, -0.00000440, 0.00000065, 0.00000554, -0.00000251, 0.00000561, -0.00000643, -0.00000047, -0.00000261, 0.00001362, -0.00000855, 0.00000269, -0.00000267, 0.00000178, -0.00000073, -0.00000419, 0.00000001, },
		{ -0.00000340, 0.00000151, 0.00000110, -0.00000114, 0.00000107, -0.00000132, -0.00000157, -0.00000601, -0.00000628, 0.00000528, -0.00000783, 0.00000084, -0.00001547, -0.00002899, 0.00000632, 0.00003078, 0.00002543, 0.00000642, 0.00002406, -0.00000864, 0.00002736, -0.00001018, 0.00003191, 0.00002092, 0.00001534, -0.00002456, -0.00002421, -0.00000481, -0.00001044, 0.00000196, -0.00000395, -0.00001064, -0.00000024, -0.00000798, -0.00000022, -0.00000145, -0.00000075, -0.00000288, 0.00000101, -0.00001974, 0.00000687, -0.00000479, -0.00000116, 0.00000197, -0.00000785, 0.00001713, -0.00000637, 0.00000526, -0.00000000, -0.00000332, -0.00000359, -0.00000242, 0.00000048, 0.00000191, -0.00000074, -0.00000028, -0.00000256, 0.00000082, 0.00000050, -0.00000085, -0.00000127, 0.00000635, 0.00000081, -0.00001052, 0.00000168, 0.00000874, -0.00000244, -0.00001241, -0.00000207, -0.00000269, -0.00000814, 0.00001044, -0.00000691, -0.00000121, -0.00000707, -0.00000096, -0.00000339, -0.00000762, -0.00000963, 0.00000158, -0.00001064, -0.00000835, -0.00001405, -0.00000820, -0.00001836, -0.00007692, -0.00007749, 0.00016350, 0.00003591, 0.00008466, 0.00003057, 0.00015233, 0.00001447, -0.00011233, -0.00000597, -0.00003320, -0.00000194, -0.00001304, -0.00000225, -0.00001101, -0.00000342, 0.00000355, 0.00000171, -0.00000771, -0.00000138, -0.00000231, 0.00000102, -0.00000302, -0.00000077, -0.00000203, 0.00000097, -0.00000446, -0.00000327, 0.00000647, -0.00000121, 0.00000395, -0.00000244, -0.00000321, -0.00000296, 0.00000931, 0.00000153, -0.00000648, 0.00000439, -0.00000424, 0.00000439, -0.00000650, 0.00000054, -0.00000137, },
		{ -0.00000246, 0.00000391, -0.00000243, 0.00000200, -0.00000171, -0.00000000, -0.00000496, -0.00000676, 0.00000014, 0.00000050, -0.00000668, -0.00000008, -0.00003403, -0.00000373, 0.00001683, 0.00003879, 0.00000046, 0.00002671, -0.00000279, 0.00001471, 0.00000783, 0.00000355, 0.00003573, 0.00001207, 0.00000311, -0.00003564, -0.00000741, -0.00001037, -0.00000331, 0.00000158, -0.00001035, -0.00000444, -0.00000322, -0.00000546, 0.00000060, -0.00000181, -0.00000210, 0.00000127, -0.00001201, -0.00000888, 0.00000936, -0.00001325, 0.00001162, -0.00001442, 0.00001479, 0.00000049, -0.00000008, 0.00000356, -0.00000216, -0.00000452, -0.00000207, -0.00000247, 0.00000400, -0.00000191, 0.00000165, -0.00000402, 0.00000156, -0.00000105, 0.00000181, -0.00000423, 0.00000620, 0.00000265, -0.00000577, -0.00000833, 0.00001275, -0.00000080, -0.00000624, -0.00001236, 0.00000698, -0.00001617, 0.00001080, -0.00000381, -0.00000012, -0.00000893, 0.00000021, -0.00000666, 0.00000172, -0.00001837, 0.00000507, -0.00001215, -0.00000077, -0.00002256, 0.00000176, -0.00002649, -0.00001834, -0.00013864, 0.00007131, 0.00012321, 0.00004347, 0.00007069, 0.00008388, 0.00013644, -0.00013402, -0.00002261, -0.00003092, -0.00000926, -0.00001344, -0.00000164, -0.00001101, -0.00000412, -0.00000375, 0.00000919, -0.00000891, -0.00000088, -0.00000559, 0.00000338, -0.00000435, 0.00000074, -0.00000459, 0.00000277, -0.00000306, -0.00000462, 0.00000208, 0.00000442, -0.00000123, 0.00000413, -0.00000750, -0.00000008, 0.00000075, 0.00001222, -0.00001193, 0.00000608, -0.00000471, 0.00000409, -0.00000425, -0.00000189, 0.00000011, -0.00000242, },
		{ 0.00000157, 0.00000077, -0.00000054, 0.00000057, -0.00000105, -0.00000228, -0.00000664, -0.00000425, 0.00000399, -0.00000774, 0.00000389, -0.00002452, -0.00001801, 0.00000815, 0.00003447, 0.00001582, 0.00000850, 0.00001990, -0.00000716, 0.00002629, -0.00001076, 0.00003016, 0.00001932, 0.00001238, -0.00001968, -0.00002476, -0.00000379, -0.00001092, 0.00000337, -0.00000548, -0.00000942, -0.00000098, -0.00000634, -0.00000092, -0.00000162, -0.00000105, -0.00000136, -0.00000183, -0.00001969, 0.00001168, -0.00001069, 0.00000528, -0.00000451, -0.00000063, 0.00001359, -0.00000703, 0.00000695, -0.00000206, -0.00000245, -0.00000430, -0.00000213, 0.00000138, 0.00000143, -0.00000080, -0.00000093, -0.00000187, 0.00000137, 0.00000000, -0.00000179, 0.00000105, 0.00000679, -0.00000310, -0.00001103, 0.00000641, 0.00000713, -0.00000505, -0.00001281, 0.00000216, -0.00000672, -0.00000448, 0.00000896, -0.00000677, -0.00000143, -0.00000771, -0.00000102, -0.00000269, -0.00000743, -0.00001060, 0.00000142, -0.00000966, -0.00000940, -0.00001378, -0.00000843, -0.00001933, -0.00007643, -0.00009405, 0.00017676, 0.00004504, 0.00008788, 0.00004447, 0.00017897, -0.00005458, -0.00009449, -0.00000133, -0.00003394, -0.00000207, -0.00001334, -0.00000092, -0.00001141, -0.00000270, 0.00000328, 0.00000167, -0.00000761, -0.00000205, -0.00000192, 0.00000090, -0.00000312, -0.00000153, -0.00000180, 0.00000244, -0.00000623, -0.00000149, 0.00000602, -0.00000126, 0.00000458, -0.00000504, -0.00000246, -0.00000175, 0.00001233, -0.00000471, -0.00000297, 0.00000288, -0.00000241, 0.00000240, -0.00000672, 0.00000235, -0.00000215, -0.00000317, },
		{ 0.00000333, -0.00000178, 0.00000157, -0.00000140, -0.00000063, -0.00000539, -0.00000655, 0.00000237, -0.00000374, 0.00000025, -0.00000923, -0.00002878, 0.00000185, 0.00001999, 0.00003422, -0.00000349, 0.00002728, -0.00000590, 0.00001702, 0.00000534, 0.00000374, 0.00003315, 0.00001011, 0.00000301, -0.00003113, -0.00000955, -0.00000824, -0.00000436, 0.00000283, -0.00001200, -0.00000242, -0.00000471, -0.00000270, -0.00000240, 0.00000028, -0.00000357, 0.00000328, -0.00001784, -0.00000033, 0.00000286, -0.00000848, 0.00000886, -0.00001331, 0.00001837, -0.00000521, 0.00000401, 0.00000085, -0.00000142, -0.00000502, -0.00000206, -0.00000185, 0.00000436, -0.00000260, 0.00000168, -0.00000450, 0.00000299, -0.00000165, 0.00000112, -0.00000369, 0.00000876, -0.00000044, -0.00000846, -0.00000471, 0.00001455, -0.00000448, -0.00000742, -0.00000945, 0.00000679, -0.00001761, 0.00001459, -0.00000806, 0.00000346, -0.00001278, 0.00000302, -0.00000945, 0.00000561, -0.00002229, 0.00000774, -0.00001512, 0.00000339, -0.00002806, 0.00000784, -0.00003447, -0.00000922, -0.00015623, 0.00007312, 0.00013460, 0.00005441, 0.00006762, 0.00012881, 0.00008973, -0.00016012, 0.00000600, -0.00004256, -0.00000091, -0.00002076, 0.00000482, -0.00001516, -0.00000025, -0.00000652, 0.00001135, -0.00001082, 0.00000060, -0.00000750, 0.00000529, -0.00000609, 0.00000176, -0.00000646, 0.00000548, -0.00000473, -0.00000394, 0.00000307, 0.00000313, 0.00000036, 0.00000178, -0.00000760, 0.00000009, 0.00000504, 0.00000805, -0.00001261, 0.00000799, -0.00000564, 0.00000528, -0.00000727, 0.00000088, -0.00000044, -0.00000251, -0.00000209, },
		{ 0.00000073, -0.00000023, 0.00000025, -0.00000103, -0.00000286, -0.00000741, -0.00000114, 0.00000042, -0.00000416, 0.00000224, -0.00002931, -0.00000711, 0.00000800, 0.00003727, 0.00000622, 0.00001234, 0.00001398, -0.00000330, 0.00002275, -0.00000871, 0.00002595, 0.00002010, 0.00000743, -0.00001334, -0.00002649, -0.00000175, -0.00001220, 0.00000533, -0.00000790, -0.00000676, -0.00000322, -0.00000315, -0.00000343, -0.00000017, -0.00000256, 0.00000139, -0.00000714, -0.00001526, 0.00001189, -0.00001333, 0.00000995, -0.00001033, 0.00000755, 0.00000730, -0.00000475, 0.00000657, -0.00000315, -0.00000194, -0.00000480, -0.00000174, 0.00000223, 0.00000067, -0.00000061, -0.00000178, -0.00000070, 0.00000148, -0.00000069, -0.00000231, 0.00000376, 0.00000600, -0.00000699, -0.00001012, 0.00001102, 0.00000394, -0.00000693, -0.00001225, 0.00000560, -0.00001108, 0.00000081, 0.00000492, -0.00000423, -0.00000374, -0.00000647, -0.00000273, -0.00000028, -0.00000910, -0.00000978, -0.00000059, -0.00000652, -0.00001286, -0.00001068, -0.00001209, -0.00001666, -0.00007979, -0.00010764, 0.00018632, 0.00006143, 0.00008286, 0.00007654, 0.00018181, -0.00012198, -0.00006144, -0.00000829, -0.00002778, -0.00000739, -0.00000951, -0.00000261, -0.00000941, -0.00000382, 0.00000431, 0.00000043, -0.00000636, -0.00000380, -0.00000046, -0.00000023, -0.00000246, -0.00000301, -0.00000049, 0.00000282, -0.00000740, 0.00000054, 0.00000481, -0.00000058, 0.00000421, -0.00000694, -0.00000191, 0.00000074, 0.00001318, -0.00001021, 0.00000132, 0.00000055, 0.00000017, -0.00000088, -0.00000532, 0.00000322, -0.00000281, -0.00000274, 0.00000134, },
		{ -0.00000089, 0.00000080, -0.00000087, -0.00000143, -0.00000598, -0.00000525, 0.00000299, -0.00000644, 0.00000634, -0.00001997, -0.00001816, 0.00000240, 0.00002506, 0.00002598, -0.00000282, 0.00002418, -0.00000538, 0.00001630, 0.00000547, 0.00000195, 0.00003208, 0.00000723, 0.00000351, -0.00002722, -0.00001105, -0.00000673, -0.00000469, 0.00000281, -0.00001206, -0.00000174, -0.00000495, -0.00000121, -0.00000451, 0.00000173, -0.00000401, 0.00000322, -0.00002117, 0.00000741, -0.00000515, -0.00000077, 0.00000254, -0.00000794, 0.00001766, -0.00000824, 0.00000717, -0.00000198, -0.00000039, -0.00000561, -0.00000196, -0.00000097, 0.00000415, -0.00000284, 0.00000128, -0.00000439, 0.00000409, -0.00000237, 0.00000030, -0.00000213, 0.00001020, -0.00000388, -0.00001053, 0.00000037, 0.00001404, -0.00000755, -0.00000828, -0.00000552, 0.00000417, -0.00001589, 0.00001538, -0.00001030, 0.00000561, -0.00001545, 0.00000479, -0.00001103, 0.00000818, -0.00002495, 0.00000901, -0.00001637, 0.00000570, -0.00003143, 0.00001162, -0.00004038, -0.00000205, -0.00017234, 0.00007200, 0.00015106, 0.00006127, 0.00007286, 0.00017308, 0.00001813, -0.00015831, 0.00002212, -0.00004888, 0.00000374, -0.00002545, 0.00000961, -0.00001801, 0.00000260, -0.00000847, 0.00001249, -0.00001167, 0.00000107, -0.00000843, 0.00000638, -0.00000721, 0.00000212, -0.00000751, 0.00000774, -0.00000662, -0.00000246, 0.00000340, 0.00000199, 0.00000192, -0.00000107, -0.00000695, 0.00000047, 0.00000942, 0.00000174, -0.00001039, 0.00000786, -0.00000492, 0.00000467, -0.00000887, 0.00000334, -0.00000130, -0.00000242, -0.00000169, 0.00000258, },
		{ -0.00000038, 0.00000027, -0.00000132, -0.00000335, -0.00000791, 0.00000228, -0.00000459, 0.00000243, -0.00000467, -0.00002745, 0.00000005, 0.00000874, 0.00003714, -0.00000091, 0.00001547, 0.00000852, 0.00000089, 0.00001847, -0.00000541, 0.00002035, 0.00002234, 0.00000129, -0.00000618, -0.00002889, 0.00000082, -0.00001375, 0.00000716, -0.00001033, -0.00000357, -0.00000611, 0.00000081, -0.00000710, 0.00000247, -0.00000483, 0.00000458, -0.00001378, -0.00000719, 0.00000704, -0.00001111, 0.00001069, -0.00001308, 0.00001417, 0.00000031, -0.00000062, 0.00000440, -0.00000317, -0.00000189, -0.00000508, -0.00000118, 0.00000279, -0.00000006, -0.00000043, -0.00000256, 0.00000074, 0.00000121, -0.00000157, -0.00000204, 0.00000621, 0.00000433, -0.00001084, -0.00000728, 0.00001427, 0.00000008, -0.00000828, -0.00001042, 0.00000718, -0.00001413, 0.00000597, -0.00000023, -0.00000029, -0.00000739, -0.00000396, -0.00000548, 0.00000316, -0.00001192, -0.00000789, -0.00000358, -0.00000221, -0.00001760, -0.00000592, -0.00001806, -0.00001134, -0.00008629, -0.00011884, 0.00019302, 0.00008393, 0.00007160, 0.00012733, 0.00015081, -0.00017179, -0.00002472, -0.00002139, -0.00001758, -0.00001602, -0.00000279, -0.00000650, -0.00000569, -0.00000625, 0.00000609, -0.00000143, -0.00000454, -0.00000608, 0.00000159, -0.00000196, -0.00000136, -0.00000483, 0.00000156, 0.00000216, -0.00000769, 0.00000247, 0.00000294, 0.00000090, 0.00000265, -0.00000776, -0.00000178, 0.00000466, 0.00001104, -0.00001346, 0.00000495, -0.00000149, 0.00000244, -0.00000451, -0.00000279, 0.00000315, -0.00000324, -0.00000213, 0.00000076, 0.00000113, },
		{ -0.00000006, -0.00000034, -0.00000225, -0.00000677, -0.00000264, 0.00000122, -0.00000592, 0.00000880, -0.00002846, -0.00000566, -0.00000039, 0.00003118, 0.00001525, 0.00000174, 0.00001771, -0.00000124, 0.00001216, 0.00000885, -0.00000257, 0.00003327, 0.00000272, 0.00000536, -0.00002463, -0.00001138, -0.00000623, -0.00000406, 0.00000135, -0.00001020, -0.00000290, -0.00000332, -0.00000180, -0.00000480, 0.00000173, -0.00000279, 0.00000029, -0.00002030, 0.00001160, -0.00001161, 0.00000723, -0.00000540, 0.00000046, 0.00001283, -0.00000780, 0.00000828, -0.00000402, 0.00000041, -0.00000604, -0.00000186, 0.00000010, 0.00000339, -0.00000256, 0.00000039, -0.00000353, 0.00000463, -0.00000316, -0.00000038, 0.00000016, 0.00001042, -0.00000750, -0.00001146, 0.00000619, 0.00001123, -0.00000951, -0.00000881, -0.00000120, -0.00000025, -0.00001114, 0.00001260, -0.00000951, 0.00000523, -0.00001583, 0.00000447, -0.00001042, 0.00000841, -0.00002529, 0.00000784, -0.00001481, 0.00000487, -0.00003113, 0.00001119, -0.00004191, 0.00000038, -0.00018391, 0.00006477, 0.00017627, 0.00005940, 0.00009549, 0.00020121, -0.00006554, -0.00013096, 0.00002227, -0.00004624, 0.00000179, -0.00002506, 0.00001069, -0.00001798, 0.00000318, -0.00000865, 0.00001188, -0.00001086, -0.00000003, -0.00000783, 0.00000607, -0.00000722, 0.00000144, -0.00000729, 0.00000895, -0.00000815, -0.00000049, 0.00000289, 0.00000139, 0.00000298, -0.00000387, -0.00000598, 0.00000165, 0.00001263, -0.00000536, -0.00000598, 0.00000589, -0.00000265, 0.00000213, -0.00000859, 0.00000489, -0.00000209, -0.00000218, -0.00000139, 0.00000189, 0.00000000, },
		{ 0.00000064, -0.00000189, -0.00000392, -0.00000756, 0.00000477, -0.00000929, 0.00001003, -0.00001521, -0.00001894, 0.00000152, 0.00001248, 0.00003244, -0.00000349, 0.00001560, 0.00000578, 0.00000309, 0.00001559, -0.00000282, 0.00001507, 0.00002458, -0.00000466, 0.00000065, -0.00003100, 0.00000308, -0.00001478, 0.00000796, -0.00001163, -0.00000111, -0.00000837, 0.00000418, -0.00001065, 0.00000524, -0.00000683, 0.00000678, -0.00001970, 0.00000234, -0.00000151, -0.00000419, 0.00000662, -0.00001118, 0.00001703, -0.00000510, 0.00000365, 0.00000134, -0.00000247, -0.00000212, -0.00000522, -0.00000031, 0.00000278, -0.00000040, -0.00000061, -0.00000288, 0.00000206, 0.00000067, -0.00000261, -0.00000080, 0.00000785, 0.00000214, -0.00001452, -0.00000230, 0.00001508, -0.00000330, -0.00000955, -0.00000705, 0.00000604, -0.00001425, 0.00000889, -0.00000441, 0.00000325, -0.00001079, -0.00000165, -0.00000783, 0.00000614, -0.00001435, -0.00000651, -0.00000581, 0.00000124, -0.00002135, -0.00000206, -0.00002355, -0.00000630, -0.00009299, -0.00013058, 0.00020028, 0.00010811, 0.00006115, 0.00018974, 0.00008355, -0.00019146, 0.00000316, -0.00003266, -0.00000856, -0.00002403, 0.00000392, -0.00001041, -0.00000199, -0.00000858, 0.00000741, -0.00000276, -0.00000326, -0.00000782, 0.00000324, -0.00000348, -0.00000054, -0.00000629, 0.00000367, 0.00000086, -0.00000710, 0.00000401, 0.00000068, 0.00000303, -0.00000005, -0.00000740, -0.00000197, 0.00000962, 0.00000578, -0.00001345, 0.00000670, -0.00000218, 0.00000332, -0.00000734, 0.00000003, 0.00000248, -0.00000345, -0.00000138, -0.00000014, 0.00000196, -0.00000103, },
		{ 0.00000003, -0.00000303, -0.00000750, 0.00000084, -0.00000281, -0.00000163, 0.00000586, -0.00003127, 0.00000445, -0.00000332, 0.00003595, 0.00000461, 0.00000793, 0.00000965, 0.00000501, 0.00000570, 0.00001471, -0.00000934, 0.00003642, -0.00000322, 0.00000849, -0.00002333, -0.00001058, -0.00000656, -0.00000284, -0.00000103, -0.00000690, -0.00000554, -0.00000007, -0.00000441, -0.00000310, 0.00000015, -0.00000008, -0.00000531, -0.00001472, 0.00001047, -0.00001377, 0.00001254, -0.00001213, 0.00000947, 0.00000544, -0.00000433, 0.00000701, -0.00000477, 0.00000067, -0.00000616, -0.00000171, 0.00000115, 0.00000232, -0.00000195, -0.00000083, -0.00000202, 0.00000451, -0.00000395, -0.00000062, 0.00000264, 0.00000958, -0.00001130, -0.00001056, 0.00001146, 0.00000687, -0.00001045, -0.00000865, 0.00000240, -0.00000499, -0.00000468, 0.00000698, -0.00000590, 0.00000228, -0.00001379, 0.00000197, -0.00000755, 0.00000623, -0.00002325, 0.00000428, -0.00001055, 0.00000098, -0.00002714, 0.00000628, -0.00003848, -0.00000294, -0.00018959, 0.00005007, 0.00021165, 0.00004714, 0.00014168, 0.00019724, -0.00014275, -0.00008832, 0.00000931, -0.00003527, -0.00000661, -0.00001933, 0.00000771, -0.00001477, 0.00000126, -0.00000698, 0.00000959, -0.00000850, -0.00000256, -0.00000578, 0.00000437, -0.00000608, -0.00000026, -0.00000577, 0.00000885, -0.00000886, 0.00000166, 0.00000150, 0.00000165, 0.00000303, -0.00000600, -0.00000516, 0.00000415, 0.00001339, -0.00001144, -0.00000081, 0.00000299, 0.00000040, -0.00000173, -0.00000652, 0.00000526, -0.00000260, -0.00000183, -0.00000132, 0.00000149, 0.00000060, -0.00000077, },
		{ -0.00000263, -0.00000466, -0.00000590, 0.00000515, -0.00001164, 0.00001572, -0.00002596, -0.00000633, -0.00000206, 0.00001936, 0.00002312, -0.00000087, 0.00001166, 0.00000699, 0.00000181, 0.00001573, -0.00000256, 0.00001161, 0.00002542, -0.00000911, 0.00000598, -0.00003188, 0.00000425, -0.00001463, 0.00000708, -0.00001093, -0.00000043, -0.00000881, 0.00000563, -0.00001267, 0.00000686, -0.00000742, 0.00000655, -0.00002255, 0.00001030, -0.00001089, 0.00000546, -0.00000128, -0.00000461, 0.00001511, -0.00000721, 0.00000639, -0.00000145, -0.00000167, -0.00000232, -0.00000540, 0.00000092, 0.00000204, -0.00000009, -0.00000144, -0.00000237, 0.00000286, 0.00000006, -0.00000369, 0.00000135, 0.00000835, -0.00000034, -0.00001758, 0.00000440, 0.00001288, -0.00000528, -0.00001110, -0.00000224, 0.00000200, -0.00001056, 0.00000796, -0.00000568, 0.00000449, -0.00001214, -0.00000120, -0.00000815, 0.00000696, -0.00001467, -0.00000741, -0.00000534, 0.00000162, -0.00002156, -0.00000208, -0.00002516, -0.00000540, -0.00009580, -0.00014719, 0.00021309, 0.00012756, 0.00006249, 0.00024885, -0.00001212, -0.00017645, 0.00001325, -0.00003505, -0.00000594, -0.00002725, 0.00000729, -0.00001178, -0.00000034, -0.00000918, 0.00000692, -0.00000234, -0.00000365, -0.00000792, 0.00000345, -0.00000387, -0.00000076, -0.00000660, 0.00000503, -0.00000040, -0.00000584, 0.00000490, -0.00000157, 0.00000542, -0.00000353, -0.00000608, -0.00000207, 0.00001465, -0.00000181, -0.00001008, 0.00000597, -0.00000092, 0.00000209, -0.00000841, 0.00000224, 0.00000166, -0.00000356, -0.00000056, -0.00000122, 0.00000306, -0.00000205, 0.00000126, },
		{ -0.00000385, -0.00000768, 0.00000418, -0.00000783, 0.00000537, -0.00000230, -0.00002679, 0.00000893, -0.00000318, 0.00003664, -0.00000293, 0.00001281, 0.00000266, 0.00001087, -0.00000085, 0.00002114, -0.00001681, 0.00004023, -0.00000945, 0.00001200, -0.00002261, -0.00000937, -0.00000701, -0.00000192, -0.00000320, -0.00000335, -0.00000851, 0.00000365, -0.00000803, -0.00000019, -0.00000227, 0.00000308, -0.00001212, -0.00000557, 0.00000398, -0.00001030, 0.00001291, -0.00001497, 0.00001626, -0.00000211, 0.00000068, 0.00000396, -0.00000432, 0.00000039, -0.00000601, -0.00000139, 0.00000186, 0.00000133, -0.00000138, -0.00000198, -0.00000019, 0.00000381, -0.00000473, -0.00000013, 0.00000475, 0.00000803, -0.00001525, -0.00000727, 0.00001469, 0.00000230, -0.00001099, -0.00000718, 0.00000394, -0.00000805, 0.00000125, 0.00000050, -0.00000099, -0.00000201, -0.00001044, -0.00000165, -0.00000357, 0.00000278, -0.00002005, -0.00000028, -0.00000518, -0.00000422, -0.00002136, -0.00000116, -0.00003195, -0.00001035, -0.00019092, 0.00002973, 0.00025471, 0.00002847, 0.00021041, 0.00015036, -0.00019448, -0.00004538, -0.00000868, -0.00002078, -0.00001810, -0.00001064, 0.00000232, -0.00000971, -0.00000205, -0.00000447, 0.00000664, -0.00000559, -0.00000559, -0.00000313, 0.00000196, -0.00000439, -0.00000241, -0.00000344, 0.00000762, -0.00000857, 0.00000364, -0.00000068, 0.00000294, 0.00000174, -0.00000696, -0.00000479, 0.00000815, 0.00001080, -0.00001478, 0.00000344, 0.00000051, 0.00000300, -0.00000579, -0.00000337, 0.00000462, -0.00000278, -0.00000136, -0.00000158, 0.00000152, 0.00000072, -0.00000115, 0.00000013, },
		{ -0.00000547, -0.00000298, 0.00000287, -0.00001017, 0.00001678, -0.00003293, 0.00000613, -0.00000776, 0.00002740, 0.00001106, 0.00000566, 0.00000438, 0.00001173, -0.00000289, 0.00001914, -0.00000515, 0.00001058, 0.00002422, -0.00001138, 0.00000920, -0.00003110, 0.00000406, -0.00001322, 0.00000450, -0.00000809, -0.00000187, -0.00000701, 0.00000450, -0.00001228, 0.00000651, -0.00000596, 0.00000308, -0.00002067, 0.00001382, -0.00001766, 0.00001462, -0.00001049, 0.00000493, 0.00000896, -0.00000549, 0.00000659, -0.00000304, -0.00000132, -0.00000221, -0.00000569, 0.00000245, 0.00000063, 0.00000091, -0.00000298, -0.00000087, 0.00000281, -0.00000042, -0.00000461, 0.00000409, 0.00000769, -0.00000307, -0.00001928, 0.00001171, 0.00000806, -0.00000555, -0.00001281, 0.00000325, -0.00000409, -0.00000342, 0.00000280, -0.00000311, 0.00000235, -0.00001035, -0.00000370, -0.00000541, 0.00000455, -0.00001178, -0.00001163, -0.00000104, -0.00000238, -0.00001665, -0.00000797, -0.00002040, -0.00001167, -0.00009123, -0.00017237, 0.00023564, 0.00013686, 0.00008692, 0.00028478, -0.00011851, -0.00013229, 0.00000346, -0.00002548, -0.00001237, -0.00002324, 0.00000519, -0.00000891, -0.00000211, -0.00000703, 0.00000386, 0.00000048, -0.00000630, -0.00000580, 0.00000160, -0.00000256, -0.00000251, -0.00000524, 0.00000497, -0.00000098, -0.00000427, 0.00000497, -0.00000331, 0.00000747, -0.00000714, -0.00000427, -0.00000141, 0.00001836, -0.00001011, -0.00000431, 0.00000311, 0.00000209, -0.00000127, -0.00000727, 0.00000321, 0.00000113, -0.00000365, 0.00000022, -0.00000228, 0.00000419, -0.00000323, 0.00000197, -0.00000348, },
		{ -0.00000673, 0.00000605, -0.00001184, 0.00001257, -0.00001325, -0.00001610, 0.00000687, 0.00000166, 0.00003161, -0.00000516, 0.00001400, -0.00000092, 0.00001391, -0.00000513, 0.00002590, -0.00002298, 0.00004294, -0.00001442, 0.00001456, -0.00002136, -0.00000870, -0.00000666, -0.00000230, -0.00000390, -0.00000098, -0.00001032, 0.00000626, -0.00001106, 0.00000253, -0.00000422, 0.00000502, -0.00001776, 0.00000448, -0.00000588, -0.00000188, 0.00000784, -0.00001247, 0.00001860, -0.00000737, 0.00000529, 0.00000030, -0.00000324, -0.00000012, -0.00000582, -0.00000069, 0.00000188, 0.00000090, -0.00000133, -0.00000254, 0.00000144, 0.00000280, -0.00000548, 0.00000120, 0.00000600, 0.00000604, -0.00001909, -0.00000148, 0.00001473, -0.00000108, -0.00001188, -0.00000389, 0.00000242, -0.00000762, 0.00000423, -0.00000430, 0.00000291, -0.00000552, -0.00000774, -0.00000446, -0.00000046, 0.00000012, -0.00001779, -0.00000350, -0.00000136, -0.00000771, -0.00001719, -0.00000739, -0.00002639, -0.00001767, -0.00019225, 0.00000875, 0.00029895, 0.00001373, 0.00029084, 0.00006005, -0.00020738, -0.00001673, -0.00002169, -0.00000969, -0.00002733, -0.00000308, -0.00000235, -0.00000528, -0.00000473, -0.00000288, 0.00000466, -0.00000367, -0.00000763, -0.00000124, 0.00000003, -0.00000314, -0.00000405, -0.00000120, 0.00000588, -0.00000742, 0.00000520, -0.00000338, 0.00000515, -0.00000094, -0.00000656, -0.00000483, 0.00001323, 0.00000482, -0.00001434, 0.00000537, -0.00000034, 0.00000390, -0.00000872, -0.00000016, 0.00000347, -0.00000277, -0.00000075, -0.00000217, 0.00000203, 0.00000028, -0.00000112, -0.00000004, -0.00000480, },
		{ 0.00000052, -0.00000150, -0.00000484, 0.00001197, -0.00003307, 0.00001414, -0.00001168, 0.00003330, -0.00000046, 0.00001331, -0.00000394, 0.00001798, -0.00000939, 0.00002455, -0.00000969, 0.00001136, 0.00002148, -0.00001179, 0.00001057, -0.00002894, 0.00000291, -0.00001116, 0.00000105, -0.00000391, -0.00000473, -0.00000354, 0.00000117, -0.00000961, 0.00000426, -0.00000278, -0.00000316, -0.00001384, 0.00001136, -0.00001907, 0.00002001, -0.00001773, 0.00001457, 0.00000055, -0.00000077, 0.00000419, -0.00000307, -0.00000169, -0.00000169, -0.00000601, 0.00000400, -0.00000112, 0.00000233, -0.00000497, 0.00000147, 0.00000182, -0.00000065, -0.00000515, 0.00000694, 0.00000609, -0.00000610, -0.00001880, 0.00001802, 0.00000182, -0.00000462, -0.00001405, 0.00000808, -0.00001043, 0.00000539, -0.00000536, 0.00000264, -0.00000283, -0.00000562, -0.00000892, 0.00000013, -0.00000084, -0.00000596, -0.00001877, 0.00000657, -0.00001023, -0.00000710, -0.00001945, -0.00000924, -0.00002560, -0.00007847, -0.00020694, 0.00026861, 0.00013526, 0.00014101, 0.00027840, -0.00021223, -0.00007344, -0.00002059, -0.00000638, -0.00002648, -0.00001264, -0.00000211, -0.00000197, -0.00000712, -0.00000240, -0.00000138, 0.00000525, -0.00001079, -0.00000178, -0.00000211, 0.00000034, -0.00000565, -0.00000229, 0.00000331, -0.00000045, -0.00000272, 0.00000414, -0.00000411, 0.00000851, -0.00001013, -0.00000253, 0.00000058, 0.00001933, -0.00001705, 0.00000216, -0.00000072, 0.00000589, -0.00000598, -0.00000422, 0.00000278, 0.00000108, -0.00000375, 0.00000083, -0.00000307, 0.00000504, -0.00000429, 0.00000260, -0.00000443, -0.00000598, },
		{ 0.00000541, -0.00001281, 0.00001688, -0.00002311, -0.00000271, 0.00000013, 0.00001040, 0.00002131, -0.00000176, 0.00001070, -0.00000011, 0.00001290, -0.00000570, 0.00002743, -0.00002635, 0.00004318, -0.00001683, 0.00001505, -0.00001868, -0.00000931, -0.00000494, -0.00000457, -0.00000234, -0.00000079, -0.00000984, 0.00000639, -0.00001198, 0.00000360, -0.00000439, 0.00000417, -0.00001978, 0.00001209, -0.00001571, 0.00000883, -0.00000117, -0.00000500, 0.00001572, -0.00000872, 0.00000774, -0.00000264, -0.00000227, -0.00000045, -0.00000579, 0.00000051, 0.00000102, 0.00000131, -0.00000216, -0.00000203, 0.00000230, 0.00000183, -0.00000619, 0.00000331, 0.00000616, 0.00000376, -0.00002221, 0.00000611, 0.00001128, -0.00000238, -0.00001355, 0.00000115, -0.00000233, -0.00000287, 0.00000259, -0.00000531, 0.00000368, -0.00000620, -0.00000765, -0.00000449, -0.00000024, 0.00000031, -0.00001858, -0.00000309, -0.00000170, -0.00000650, -0.00001814, -0.00000831, -0.00002648, -0.00001975, -0.00019908, -0.00000656, 0.00033597, 0.00001759, 0.00036277, -0.00006139, -0.00017869, -0.00001141, -0.00002199, -0.00000796, -0.00002937, -0.00000069, -0.00000311, -0.00000406, -0.00000472, -0.00000391, 0.00000518, -0.00000416, -0.00000732, -0.00000145, -0.00000021, -0.00000338, -0.00000418, -0.00000004, 0.00000445, -0.00000577, 0.00000613, -0.00000615, 0.00000784, -0.00000463, -0.00000499, -0.00000485, 0.00001832, -0.00000359, -0.00001015, 0.00000441, 0.00000106, 0.00000236, -0.00000948, 0.00000207, 0.00000241, -0.00000277, 0.00000001, -0.00000302, 0.00000289, -0.00000064, -0.00000073, -0.00000049, -0.00000585, -0.00000441, },
		{ -0.00000636, 0.00000274, 0.00000226, -0.00002566, 0.00001508, -0.00001065, 0.00003395, -0.00000802, 0.00001882, -0.00001038, 0.00002292, -0.00001508, 0.00002967, -0.00001426, 0.00001233, 0.00001858, -0.00001152, 0.00001107, -0.00002632, 0.00000172, -0.00000955, -0.00000191, 0.00000015, -0.00000759, 0.00000011, -0.00000295, -0.00000581, 0.00000117, 0.00000079, -0.00001037, -0.00000370, 0.00000344, -0.00001417, 0.00001949, -0.00002023, 0.00002123, -0.00000734, 0.00000507, 0.00000013, -0.00000184, -0.00000262, -0.00000093, -0.00000612, 0.00000514, -0.00000270, 0.00000362, -0.00000686, 0.00000415, 0.00000009, -0.00000064, -0.00000508, 0.00000936, 0.00000387, -0.00000939, -0.00001552, 0.00002164, -0.00000411, -0.00000347, -0.00001389, 0.00001064, -0.00001468, 0.00001316, -0.00001397, 0.00000947, -0.00000924, 0.00000037, -0.00001523, 0.00000680, -0.00000750, 0.00000098, -0.00002678, 0.00001516, -0.00001936, 0.00000427, -0.00003353, 0.00000526, -0.00004422, -0.00006046, -0.00024753, 0.00030738, 0.00012961, 0.00022152, 0.00021829, -0.00027093, -0.00001870, -0.00004778, 0.00001517, -0.00004310, 0.00000076, -0.00001184, 0.00000681, -0.00001353, 0.00000307, -0.00000722, 0.00001047, -0.00001569, 0.00000282, -0.00000657, 0.00000388, -0.00000933, 0.00000154, 0.00000046, 0.00000120, -0.00000146, 0.00000245, -0.00000367, 0.00000804, -0.00001185, -0.00000127, 0.00000412, 0.00001665, -0.00002075, 0.00000734, -0.00000395, 0.00000900, -0.00001065, -0.00000019, 0.00000131, 0.00000142, -0.00000377, 0.00000112, -0.00000340, 0.00000536, -0.00000501, 0.00000301, -0.00000555, -0.00000562, 0.00000339, },
		{ -0.00000964, 0.00001587, -0.00002789, 0.00000867, -0.00000739, 0.00002003, 0.00000834, 0.00000543, 0.00000411, 0.00000419, 0.00000835, -0.00000271, 0.00002557, -0.00002659, 0.00004058, -0.00001627, 0.00001310, -0.00001429, -0.00001130, -0.00000196, -0.00000846, 0.00000135, -0.00000285, -0.00000686, 0.00000361, -0.00001009, 0.00000231, -0.00000219, -0.00000016, -0.00001662, 0.00001440, -0.00002182, 0.00001813, -0.00001099, 0.00000524, 0.00000859, -0.00000597, 0.00000721, -0.00000398, -0.00000200, -0.00000031, -0.00000601, 0.00000210, -0.00000062, 0.00000258, -0.00000389, -0.00000028, 0.00000202, 0.00000121, -0.00000676, 0.00000591, 0.00000524, 0.00000113, -0.00002372, 0.00001407, 0.00000506, -0.00000158, -0.00001573, 0.00000705, -0.00000924, 0.00000553, -0.00000373, -0.00000185, 0.00000036, -0.00000304, -0.00001120, -0.00000076, -0.00000394, 0.00000437, -0.00002339, 0.00000198, -0.00000740, 0.00000088, -0.00002614, -0.00000147, -0.00003532, -0.00001295, -0.00021542, -0.00001171, 0.00035956, 0.00005418, 0.00040090, -0.00019036, -0.00011801, -0.00002949, -0.00000751, -0.00001772, -0.00002205, -0.00000547, 0.00000169, -0.00000735, -0.00000103, -0.00000828, 0.00000879, -0.00000759, -0.00000410, -0.00000436, 0.00000183, -0.00000564, -0.00000224, -0.00000067, 0.00000406, -0.00000404, 0.00000628, -0.00000842, 0.00001029, -0.00000860, -0.00000279, -0.00000416, 0.00002189, -0.00001253, -0.00000342, 0.00000108, 0.00000438, -0.00000149, -0.00000775, 0.00000271, 0.00000190, -0.00000293, 0.00000084, -0.00000392, 0.00000386, -0.00000184, -0.00000010, -0.00000123, -0.00000673, -0.00000100, 0.00000209, },
		{ 0.00000970, -0.00000930, -0.00001278, 0.00000912, -0.00000371, 0.00002795, -0.00000953, 0.00001986, -0.00001265, 0.00002421, -0.00001758, 0.00003216, -0.00001676, 0.00001165, 0.00001719, -0.00001202, 0.00001195, -0.00002433, 0.00000148, -0.00000944, -0.00000307, 0.00000258, -0.00000880, 0.00000216, -0.00000600, -0.00000260, -0.00000115, 0.00000283, -0.00001591, 0.00000666, -0.00000737, -0.00000429, 0.00001301, -0.00001674, 0.00002275, -0.00001218, 0.00000986, -0.00000415, -0.00000015, -0.00000366, -0.00000024, -0.00000574, 0.00000543, -0.00000352, 0.00000419, -0.00000794, 0.00000645, -0.00000192, -0.00000048, -0.00000427, 0.00001090, 0.00000129, -0.00001261, -0.00000950, 0.00002143, -0.00000817, -0.00000316, -0.00001159, 0.00000975, -0.00001485, 0.00001716, -0.00002007, 0.00001457, -0.00001425, 0.00000512, -0.00002014, 0.00001202, -0.00001279, 0.00000631, -0.00003265, 0.00002136, -0.00002594, 0.00001309, -0.00004533, 0.00001773, -0.00006179, -0.00004332, -0.00028705, 0.00034249, 0.00013475, 0.00031272, 0.00010650, -0.00028062, 0.00001512, -0.00006592, 0.00003038, -0.00005529, 0.00001151, -0.00001973, 0.00001401, -0.00001851, 0.00000693, -0.00001142, 0.00001406, -0.00001900, 0.00000614, -0.00001013, 0.00000665, -0.00001223, 0.00000503, -0.00000269, 0.00000360, -0.00000062, 0.00000014, -0.00000204, 0.00000593, -0.00001198, -0.00000053, 0.00000877, 0.00001033, -0.00002014, 0.00000973, -0.00000520, 0.00001007, -0.00001379, 0.00000355, -0.00000049, 0.00000186, -0.00000358, 0.00000099, -0.00000315, 0.00000501, -0.00000521, 0.00000308, -0.00000679, -0.00000397, 0.00000435, -0.00000955, },
		{ 0.00000889, -0.00002513, 0.00001393, -0.00001138, 0.00002661, -0.00000343, 0.00001313, -0.00000303, 0.00000950, 0.00000241, 0.00000207, 0.00002165, -0.00002474, 0.00003601, -0.00001343, 0.00000926, -0.00000878, -0.00001402, 0.00000142, -0.00001290, 0.00000609, -0.00000614, -0.00000236, -0.00000128, -0.00000587, -0.00000101, 0.00000179, -0.00000722, -0.00000845, 0.00001024, -0.00002160, 0.00002257, -0.00001803, 0.00001495, -0.00000033, -0.00000038, 0.00000397, -0.00000355, -0.00000258, 0.00000031, -0.00000632, 0.00000372, -0.00000261, 0.00000429, -0.00000615, 0.00000243, 0.00000057, 0.00000107, -0.00000704, 0.00000854, 0.00000351, -0.00000194, -0.00002264, 0.00002047, -0.00000225, 0.00000042, -0.00001748, 0.00001218, -0.00001611, 0.00001530, -0.00001295, 0.00000492, -0.00000621, 0.00000326, -0.00001769, 0.00000602, -0.00001086, 0.00001154, -0.00003130, 0.00001060, -0.00001727, 0.00001320, -0.00004007, 0.00001228, -0.00005244, 0.00000259, -0.00024105, -0.00000739, 0.00037034, 0.00013056, 0.00038223, -0.00029764, -0.00004480, -0.00006194, 0.00001701, -0.00003588, -0.00000730, -0.00001623, 0.00001114, -0.00001438, 0.00000555, -0.00001513, 0.00001461, -0.00001314, 0.00000131, -0.00000943, 0.00000576, -0.00000957, 0.00000152, -0.00000310, 0.00000502, -0.00000254, 0.00000551, -0.00000959, 0.00001166, -0.00001193, -0.00000062, -0.00000210, 0.00002243, -0.00001970, 0.00000380, -0.00000318, 0.00000842, -0.00000663, -0.00000403, 0.00000177, 0.00000204, -0.00000321, 0.00000157, -0.00000461, 0.00000460, -0.00000298, 0.00000059, -0.00000228, -0.00000697, 0.00000257, -0.00000298, -0.00000279, },
		{ -0.00001851, 0.00000130, -0.00000077, 0.00000739, 0.00001634, -0.00000511, 0.00001602, -0.00001018, 0.00002097, -0.00001575, 0.00003070, -0.00001587, 0.00000809, 0.00001850, -0.00001437, 0.00001409, -0.00002364, 0.00000267, -0.00001129, -0.00000176, 0.00000251, -0.00000737, 0.00000134, -0.00000647, -0.00000148, -0.00000133, 0.00000173, -0.00001724, 0.00001367, -0.00001720, 0.00000729, 0.00000269, -0.00000810, 0.00001868, -0.00001255, 0.00001187, -0.00000721, 0.00000112, -0.00000428, 0.00000007, -0.00000468, 0.00000464, -0.00000324, 0.00000360, -0.00000765, 0.00000766, -0.00000367, -0.00000031, -0.00000273, 0.00001134, -0.00000156, -0.00001503, -0.00000164, 0.00001730, -0.00000950, -0.00000414, -0.00000717, 0.00000529, -0.00001017, 0.00001575, -0.00002143, 0.00001562, -0.00001556, 0.00000636, -0.00002137, 0.00001345, -0.00001431, 0.00000757, -0.00003375, 0.00002215, -0.00002651, 0.00001528, -0.00005006, 0.00002261, -0.00007204, -0.00003394, -0.00031748, 0.00036278, 0.00017039, 0.00038767, -0.00003917, -0.00024094, 0.00001921, -0.00006675, 0.00003262, -0.00005740, 0.00001485, -0.00002196, 0.00001653, -0.00001954, 0.00000709, -0.00001211, 0.00001427, -0.00001906, 0.00000656, -0.00001127, 0.00000731, -0.00001308, 0.00000694, -0.00000506, 0.00000619, -0.00000029, -0.00000237, 0.00000031, 0.00000251, -0.00001067, 0.00000009, 0.00001337, 0.00000152, -0.00001540, 0.00000879, -0.00000385, 0.00000833, -0.00001431, 0.00000587, -0.00000188, 0.00000207, -0.00000306, 0.00000045, -0.00000240, 0.00000401, -0.00000484, 0.00000268, -0.00000795, -0.00000100, 0.00000261, -0.00000922, 0.00001295, },
		{ -0.00001501, 0.00001125, -0.00000897, 0.00002690, -0.00001041, 0.00001791, -0.00000765, 0.00001284, -0.00000208, 0.00000604, 0.00001790, -0.00002271, 0.00003118, -0.00000982, 0.00000482, -0.00000327, -0.00001636, 0.00000393, -0.00001635, 0.00001019, -0.00000897, 0.00000188, -0.00000654, -0.00000084, -0.00000495, 0.00000593, -0.00001486, 0.00000262, 0.00000068, -0.00001460, 0.00002025, -0.00001946, 0.00002088, -0.00000799, 0.00000580, -0.00000068, -0.00000193, -0.00000367, 0.00000110, -0.00000634, 0.00000480, -0.00000423, 0.00000571, -0.00000816, 0.00000539, -0.00000165, 0.00000130, -0.00000676, 0.00001063, 0.00000131, -0.00000543, -0.00001838, 0.00002346, -0.00000857, 0.00000219, -0.00001751, 0.00001461, -0.00002025, 0.00002323, -0.00002189, 0.00001222, -0.00001352, 0.00001035, -0.00002480, 0.00001346, -0.00001853, 0.00001925, -0.00003942, 0.00001953, -0.00002768, 0.00002645, -0.00005558, 0.00002832, -0.00007312, 0.00002194, -0.00027025, -0.00000134, 0.00037902, 0.00024079, 0.00029443, -0.00035690, 0.00001757, -0.00009401, 0.00004169, -0.00005503, 0.00000925, -0.00002869, 0.00002180, -0.00002228, 0.00001247, -0.00002207, 0.00002042, -0.00001868, 0.00000696, -0.00001497, 0.00001014, -0.00001387, 0.00000599, -0.00000660, 0.00000708, -0.00000135, 0.00000375, -0.00000924, 0.00001124, -0.00001376, 0.00000094, 0.00000161, 0.00001895, -0.00002304, 0.00000928, -0.00000653, 0.00001145, -0.00001139, 0.00000043, -0.00000014, 0.00000256, -0.00000339, 0.00000194, -0.00000479, 0.00000475, -0.00000374, 0.00000108, -0.00000358, -0.00000609, 0.00000504, -0.00000794, 0.00000565, -0.00000238, },
		{ 0.00001167, -0.00000996, 0.00001877, 0.00000245, 0.00000285, 0.00000901, -0.00000438, 0.00001427, -0.00001027, 0.00002558, -0.00001167, 0.00000165, 0.00002260, -0.00001872, 0.00001756, -0.00002416, 0.00000497, -0.00001460, 0.00000161, 0.00000016, -0.00000341, -0.00000248, -0.00000389, -0.00000303, 0.00000109, -0.00000302, -0.00001303, 0.00001461, -0.00002222, 0.00001646, -0.00000784, 0.00000296, 0.00001049, -0.00000865, 0.00001051, -0.00000826, 0.00000140, -0.00000422, -0.00000006, -0.00000308, 0.00000294, -0.00000192, 0.00000186, -0.00000584, 0.00000737, -0.00000474, -0.00000016, -0.00000070, 0.00001070, -0.00000475, -0.00001564, 0.00000632, 0.00001036, -0.00000840, -0.00000601, -0.00000169, -0.00000142, -0.00000168, 0.00000921, -0.00001766, 0.00001189, -0.00001234, 0.00000325, -0.00001808, 0.00001019, -0.00001116, 0.00000393, -0.00002922, 0.00001655, -0.00001982, 0.00000913, -0.00004542, 0.00001687, -0.00007128, -0.00003650, -0.00033401, 0.00036075, 0.00025462, 0.00041412, -0.00018784, -0.00016642, -0.00000387, -0.00004964, 0.00002066, -0.00004783, 0.00000909, -0.00001707, 0.00001317, -0.00001575, 0.00000293, -0.00000882, 0.00001068, -0.00001542, 0.00000353, -0.00000941, 0.00000530, -0.00001132, 0.00000655, -0.00000587, 0.00000845, -0.00000058, -0.00000444, 0.00000258, -0.00000139, -0.00000851, 0.00000129, 0.00001633, -0.00000765, -0.00000803, 0.00000524, -0.00000041, 0.00000408, -0.00001203, 0.00000622, -0.00000240, 0.00000190, -0.00000229, -0.00000032, -0.00000138, 0.00000256, -0.00000402, 0.00000173, -0.00000862, 0.00000272, -0.00000163, -0.00000469, 0.00001046, -0.00002162, },
		{ 0.00000185, 0.00000004, 0.00001989, -0.00001073, 0.00001764, -0.00000765, 0.00001204, -0.00000281, 0.00000689, 0.00001646, -0.00002245, 0.00002791, -0.00000706, 0.00000115, 0.00000111, -0.00001732, 0.00000452, -0.00001749, 0.00001214, -0.00000967, 0.00000399, -0.00001012, 0.00000306, -0.00000770, 0.00000812, -0.00002021, 0.00001310, -0.00001110, -0.00000281, 0.00001164, -0.00001435, 0.00002100, -0.00001183, 0.00001019, -0.00000501, -0.00000017, -0.00000460, 0.00000157, -0.00000563, 0.00000477, -0.00000477, 0.00000605, -0.00000900, 0.00000761, -0.00000391, 0.00000160, -0.00000568, 0.00001167, -0.00000102, -0.00000901, -0.00001103, 0.00002189, -0.00001212, 0.00000240, -0.00001481, 0.00001294, -0.00001943, 0.00002625, -0.00002711, 0.00001667, -0.00001834, 0.00001506, -0.00002934, 0.00001826, -0.00002356, 0.00002395, -0.00004391, 0.00002446, -0.00003375, 0.00003503, -0.00006636, 0.00003959, -0.00008958, 0.00003659, -0.00029296, -0.00000729, 0.00040642, 0.00036322, 0.00014243, -0.00035325, 0.00004973, -0.00011090, 0.00005553, -0.00006630, 0.00002047, -0.00003715, 0.00002902, -0.00002717, 0.00001638, -0.00002606, 0.00002338, -0.00002155, 0.00001034, -0.00001871, 0.00001296, -0.00001671, 0.00000948, -0.00000981, 0.00000947, -0.00000033, 0.00000105, -0.00000725, 0.00000870, -0.00001358, 0.00000169, 0.00000657, 0.00001152, -0.00002144, 0.00001132, -0.00000741, 0.00001193, -0.00001406, 0.00000410, -0.00000204, 0.00000295, -0.00000314, 0.00000169, -0.00000421, 0.00000405, -0.00000379, 0.00000108, -0.00000492, -0.00000386, 0.00000528, -0.00001064, 0.00001260, -0.00001440, -0.00000047, },
		{ -0.00001405, 0.00002604, -0.00000945, 0.00001067, 0.00000194, 0.00000190, 0.00000671, -0.00000338, 0.00001861, -0.00000566, -0.00000637, 0.00002843, -0.00002421, 0.00002159, -0.00002504, 0.00000731, -0.00001805, 0.00000567, -0.00000312, 0.00000176, -0.00000813, 0.00000087, -0.00000657, 0.00000524, -0.00001033, -0.00000390, 0.00000875, -0.00002012, 0.00001973, -0.00001467, 0.00001277, 0.00000107, -0.00000220, 0.00000644, -0.00000739, 0.00000071, -0.00000359, -0.00000034, -0.00000145, 0.00000091, -0.00000015, -0.00000045, -0.00000296, 0.00000569, -0.00000506, 0.00000014, 0.00000135, 0.00000929, -0.00000843, -0.00001339, 0.00001220, 0.00000272, -0.00000618, -0.00000743, 0.00000287, -0.00000778, 0.00000782, -0.00000004, -0.00001057, 0.00000484, -0.00000589, -0.00000295, -0.00001155, 0.00000355, -0.00000474, -0.00000297, -0.00002094, 0.00000663, -0.00000805, -0.00000318, -0.00003347, 0.00000229, -0.00006103, -0.00004927, -0.00033912, 0.00033830, 0.00039591, 0.00036371, -0.00030303, -0.00008274, -0.00004087, -0.00002229, -0.00000018, -0.00003025, -0.00000321, -0.00000705, 0.00000564, -0.00000872, -0.00000397, -0.00000304, 0.00000474, -0.00000935, -0.00000193, -0.00000537, 0.00000134, -0.00000749, 0.00000410, -0.00000502, 0.00001019, -0.00000172, -0.00000532, 0.00000370, -0.00000463, -0.00000633, 0.00000384, 0.00001600, -0.00001462, -0.00000041, 0.00000084, 0.00000364, -0.00000134, -0.00000775, 0.00000484, -0.00000210, 0.00000149, -0.00000153, -0.00000095, -0.00000054, 0.00000109, -0.00000300, 0.00000028, -0.00000832, 0.00000613, -0.00000711, 0.00000308, 0.00000220, -0.00001684, 0.00001465, },
		{ 0.00001299, 0.00000729, -0.00000503, 0.00001237, -0.00000291, 0.00000666, 0.00000094, 0.00000361, 0.00001843, -0.00002501, 0.00002727, -0.00000617, -0.00000094, 0.00000377, -0.00001657, 0.00000293, -0.00001589, 0.00001129, -0.00000746, 0.00000288, -0.00001063, 0.00000435, -0.00000789, 0.00000680, -0.00002078, 0.00001935, -0.00002090, 0.00000993, -0.00000052, -0.00000406, 0.00001525, -0.00001065, 0.00001113, -0.00000751, 0.00000071, -0.00000474, 0.00000131, -0.00000393, 0.00000332, -0.00000381, 0.00000477, -0.00000796, 0.00000819, -0.00000542, 0.00000159, -0.00000364, 0.00001133, -0.00000331, -0.00001196, -0.00000165, 0.00001583, -0.00001209, 0.00000046, -0.00000929, 0.00000703, -0.00001291, 0.00002269, -0.00002626, 0.00001570, -0.00001807, 0.00001477, -0.00002866, 0.00001767, -0.00002312, 0.00002273, -0.00004162, 0.00002182, -0.00003136, 0.00003408, -0.00006667, 0.00003934, -0.00009403, 0.00003767, -0.00029853, -0.00004066, 0.00047913, 0.00046259, -0.00004959, -0.00028901, 0.00004304, -0.00010377, 0.00005110, -0.00006319, 0.00002071, -0.00003697, 0.00002898, -0.00002590, 0.00001463, -0.00002476, 0.00002134, -0.00001966, 0.00000939, -0.00001870, 0.00001246, -0.00001644, 0.00001040, -0.00001131, 0.00001129, 0.00000069, -0.00000228, -0.00000399, 0.00000430, -0.00001141, 0.00000189, 0.00001169, 0.00000138, -0.00001522, 0.00000944, -0.00000518, 0.00000910, -0.00001352, 0.00000573, -0.00000302, 0.00000272, -0.00000223, 0.00000072, -0.00000285, 0.00000241, -0.00000297, 0.00000031, -0.00000586, -0.00000057, 0.00000286, -0.00000954, 0.00001516, -0.00002290, 0.00001372, -0.00001040, },
		{ 0.00002606, -0.00001570, 0.00001486, -0.00000205, 0.00000558, 0.00000131, 0.00000209, 0.00001231, -0.00000004, -0.00001397, 0.00003423, -0.00002942, 0.00002495, -0.00002510, 0.00000834, -0.00002002, 0.00000862, -0.00000547, 0.00000611, -0.00001354, 0.00000597, -0.00001041, 0.00000922, -0.00001774, 0.00000760, -0.00000224, -0.00001100, 0.00001556, -0.00001508, 0.00001796, -0.00000627, 0.00000417, 0.00000136, -0.00000552, -0.00000034, -0.00000293, -0.00000020, -0.00000054, -0.00000052, 0.00000109, -0.00000228, -0.00000003, 0.00000332, -0.00000496, 0.00000098, 0.00000274, 0.00000760, -0.00001268, -0.00000760, 0.00001395, -0.00000314, -0.00000478, -0.00000668, 0.00000420, -0.00001064, 0.00001453, -0.00000810, -0.00000374, -0.00000220, 0.00000058, -0.00000902, -0.00000511, -0.00000304, 0.00000134, -0.00000918, -0.00001336, -0.00000264, 0.00000322, -0.00001554, -0.00002082, -0.00001409, -0.00004889, -0.00006332, -0.00034459, 0.00031054, 0.00058606, 0.00022204, -0.00035303, -0.00001892, -0.00007247, 0.00000187, -0.00001958, -0.00001269, -0.00001579, 0.00000301, -0.00000179, -0.00000223, -0.00001015, 0.00000197, -0.00000052, -0.00000367, -0.00000729, -0.00000133, -0.00000261, -0.00000331, 0.00000089, -0.00000325, 0.00001170, -0.00000405, -0.00000431, 0.00000266, -0.00000606, -0.00000496, 0.00000816, 0.00001131, -0.00001712, 0.00000488, -0.00000222, 0.00000627, -0.00000593, -0.00000310, 0.00000269, -0.00000153, 0.00000130, -0.00000124, -0.00000097, -0.00000039, 0.00000013, -0.00000224, -0.00000144, -0.00000674, 0.00000802, -0.00001184, 0.00001166, -0.00000958, -0.00000519, 0.00000924, -0.00001059, },
		{ -0.00000701, 0.00000378, 0.00000427, 0.00000468, -0.00000170, 0.00000800, -0.00000305, 0.00002330, -0.00003006, 0.00002910, -0.00000713, -0.00000154, 0.00000501, -0.00001466, -0.00000008, -0.00001225, 0.00000822, -0.00000281, -0.00000121, -0.00000795, 0.00000274, -0.00000531, 0.00000174, -0.00001558, 0.00001892, -0.00002495, 0.00001923, -0.00001214, 0.00000821, 0.00000557, -0.00000508, 0.00000830, -0.00000752, 0.00000018, -0.00000383, 0.00000027, -0.00000140, 0.00000065, -0.00000144, 0.00000192, -0.00000492, 0.00000667, -0.00000564, 0.00000109, -0.00000077, 0.00000958, -0.00000558, -0.00001327, 0.00000786, 0.00000668, -0.00000901, -0.00000304, -0.00000214, -0.00000161, -0.00000203, 0.00001320, -0.00001922, 0.00000882, -0.00001203, 0.00000878, -0.00002199, 0.00001087, -0.00001637, 0.00001484, -0.00003182, 0.00001075, -0.00001939, 0.00002199, -0.00005426, 0.00002446, -0.00008249, 0.00002044, -0.00028105, -0.00011171, 0.00062139, 0.00049757, -0.00024206, -0.00018438, 0.00000305, -0.00007362, 0.00002822, -0.00004475, 0.00000868, -0.00002691, 0.00002067, -0.00001771, 0.00000673, -0.00001778, 0.00001396, -0.00001265, 0.00000363, -0.00001437, 0.00000809, -0.00001245, 0.00000797, -0.00001023, 0.00001189, 0.00000171, -0.00000564, -0.00000032, -0.00000111, -0.00000782, 0.00000222, 0.00001531, -0.00000916, -0.00000609, 0.00000455, -0.00000050, 0.00000341, -0.00000971, 0.00000485, -0.00000263, 0.00000169, -0.00000071, -0.00000080, -0.00000095, 0.00000005, -0.00000136, -0.00000137, -0.00000586, 0.00000295, -0.00000163, -0.00000444, 0.00001180, -0.00002456, 0.00002271, -0.00002044, 0.00002535, },
		{ -0.00001480, 0.00001357, -0.00000115, 0.00000470, 0.00000020, 0.00000387, 0.00000882, 0.00000318, -0.00001925, 0.00003832, -0.00003296, 0.00002655, -0.00002341, 0.00000712, -0.00001931, 0.00000905, -0.00000531, 0.00000781, -0.00001665, 0.00000937, -0.00001262, 0.00001085, -0.00002227, 0.00001772, -0.00001471, 0.00000251, 0.00000498, -0.00000851, 0.00001670, -0.00000896, 0.00000791, -0.00000271, -0.00000400, -0.00000084, -0.00000298, 0.00000104, -0.00000115, -0.00000040, 0.00000073, -0.00000240, 0.00000160, 0.00000138, -0.00000511, 0.00000289, 0.00000273, 0.00000619, -0.00001732, 0.00000169, 0.00001030, -0.00000516, -0.00000592, -0.00000235, 0.00000055, -0.00000744, 0.00001493, -0.00001092, -0.00000127, -0.00000520, 0.00000305, -0.00001086, -0.00000301, -0.00000517, 0.00000248, -0.00000968, -0.00001209, -0.00000491, 0.00000674, -0.00001970, -0.00001685, -0.00002170, -0.00004675, -0.00006438, -0.00036991, 0.00030544, 0.00079723, -0.00000388, -0.00032087, 0.00000194, -0.00008040, 0.00000886, -0.00002609, -0.00000459, -0.00002097, 0.00000676, -0.00000374, -0.00000093, -0.00001140, 0.00000234, -0.00000143, -0.00000186, -0.00000938, -0.00000016, -0.00000395, -0.00000118, -0.00000106, -0.00000191, 0.00001361, -0.00000790, -0.00000098, -0.00000119, -0.00000480, -0.00000493, 0.00001407, 0.00000221, -0.00001394, 0.00000594, -0.00000214, 0.00000568, -0.00000769, 0.00000007, 0.00000110, -0.00000152, 0.00000190, -0.00000190, 0.00000007, -0.00000143, 0.00000021, -0.00000227, -0.00000290, -0.00000397, 0.00000757, -0.00001386, 0.00001798, -0.00002103, 0.00000973, -0.00000240, -0.00000020, 0.00001830, },
		{ 0.00001175, -0.00000327, 0.00001196, -0.00001008, 0.00001573, -0.00001086, 0.00002923, -0.00003599, 0.00003209, -0.00000889, -0.00000158, 0.00000586, -0.00001286, -0.00000298, -0.00000821, 0.00000457, 0.00000260, -0.00000671, -0.00000331, -0.00000074, -0.00000115, -0.00000558, -0.00000572, 0.00001163, -0.00002135, 0.00002175, -0.00001920, 0.00001847, -0.00000473, 0.00000271, 0.00000278, -0.00000540, -0.00000153, -0.00000215, -0.00000112, 0.00000134, -0.00000248, 0.00000154, -0.00000171, -0.00000053, 0.00000335, -0.00000459, 0.00000028, 0.00000243, 0.00000678, -0.00000801, -0.00001186, 0.00001513, -0.00000307, -0.00000461, -0.00000644, 0.00000436, -0.00000999, 0.00000990, 0.00000080, -0.00000847, -0.00000188, -0.00000218, -0.00000102, -0.00001128, -0.00000016, -0.00000541, 0.00000266, -0.00001719, -0.00000578, -0.00000102, 0.00000208, -0.00003242, -0.00000192, -0.00005786, -0.00001194, -0.00024431, -0.00021859, 0.00084529, 0.00043201, -0.00038935, -0.00007240, -0.00005199, -0.00003167, -0.00000512, -0.00001676, -0.00001142, -0.00001029, 0.00000686, -0.00000514, -0.00000495, -0.00000739, 0.00000340, -0.00000244, -0.00000532, -0.00000706, 0.00000100, -0.00000569, 0.00000278, -0.00000674, 0.00001126, 0.00000236, -0.00000813, 0.00000250, -0.00000617, -0.00000385, 0.00000353, 0.00001572, -0.00001731, 0.00000320, -0.00000128, 0.00000484, -0.00000347, -0.00000380, 0.00000197, -0.00000111, 0.00000016, 0.00000104, -0.00000239, 0.00000094, -0.00000248, 0.00000065, -0.00000381, -0.00000449, 0.00000556, -0.00000666, 0.00000324, 0.00000313, -0.00001830, 0.00002342, -0.00002398, 0.00003251, -0.00001843, },
		{ 0.00000728, 0.00000436, -0.00000070, 0.00000371, 0.00000132, 0.00000897, 0.00000315, -0.00002133, 0.00003983, -0.00003419, 0.00002603, -0.00001989, 0.00000366, -0.00001579, 0.00000663, -0.00000215, 0.00000607, -0.00001632, 0.00000978, -0.00001199, 0.00000875, -0.00002163, 0.00002289, -0.00002427, 0.00001613, -0.00000876, 0.00000324, 0.00000928, -0.00000606, 0.00000743, -0.00000416, -0.00000401, 0.00000002, -0.00000430, 0.00000379, -0.00000374, 0.00000188, -0.00000195, 0.00000012, 0.00000073, 0.00000103, -0.00000625, 0.00000633, 0.00000070, 0.00000550, -0.00002188, 0.00001351, 0.00000132, -0.00000240, -0.00001044, 0.00000591, -0.00000844, 0.00000272, 0.00000718, -0.00000585, -0.00000613, -0.00000105, -0.00000168, -0.00000516, -0.00000872, 0.00000077, -0.00000509, -0.00000042, -0.00002169, 0.00000506, -0.00000365, -0.00000837, -0.00003023, -0.00001033, -0.00006657, -0.00003759, -0.00043676, 0.00035839, 0.00098489, -0.00028368, -0.00021074, -0.00002947, -0.00005441, -0.00001021, -0.00001177, -0.00001297, -0.00001281, -0.00000079, 0.00000399, -0.00000842, -0.00000451, -0.00000487, 0.00000482, -0.00000665, -0.00000558, -0.00000424, -0.00000041, -0.00000329, 0.00000027, -0.00000252, 0.00001660, -0.00001334, 0.00000461, -0.00000791, -0.00000056, -0.00000629, 0.00002069, -0.00001009, -0.00000536, 0.00000222, 0.00000182, 0.00000100, -0.00000538, 0.00000030, 0.00000122, -0.00000280, 0.00000377, -0.00000382, 0.00000238, -0.00000389, 0.00000168, -0.00000361, -0.00000339, -0.00000071, 0.00000481, -0.00001210, 0.00001952, -0.00002809, 0.00002295, -0.00001575, 0.00001373, 0.00000509, -0.00000789, },
		{ -0.00000717, 0.00001588, -0.00001542, 0.00002114, -0.00001714, 0.00003381, -0.00004062, 0.00003436, -0.00000994, -0.00000236, 0.00000757, -0.00001253, -0.00000412, -0.00000563, 0.00000231, 0.00000659, -0.00001137, 0.00000115, -0.00000412, 0.00000245, -0.00001244, 0.00000582, -0.00000026, -0.00001083, 0.00001641, -0.00001914, 0.00002336, -0.00001218, 0.00000977, -0.00000334, -0.00000246, -0.00000352, -0.00000051, -0.00000203, 0.00000325, -0.00000488, 0.00000385, -0.00000475, 0.00000381, -0.00000069, -0.00000293, -0.00000020, 0.00000507, 0.00000363, -0.00001086, -0.00000691, 0.00001787, -0.00001054, -0.00000132, -0.00000754, 0.00000743, -0.00001449, 0.00001846, -0.00000984, 0.00000155, -0.00001214, 0.00000731, -0.00001039, -0.00000092, -0.00001089, 0.00000497, -0.00000859, -0.00000358, -0.00002124, 0.00001642, -0.00001759, -0.00001007, -0.00003011, -0.00003076, -0.00004704, -0.00020409, -0.00034276, 0.00114247, 0.00024680, -0.00045207, 0.00001056, -0.00009709, 0.00000426, -0.00003501, 0.00000976, -0.00003082, 0.00000569, -0.00000631, 0.00000644, -0.00001550, 0.00000188, -0.00000610, 0.00000701, -0.00001390, 0.00000007, -0.00000599, 0.00000133, -0.00000312, -0.00000225, 0.00001021, 0.00000189, -0.00000864, 0.00000303, -0.00000930, -0.00000069, 0.00000650, 0.00001167, -0.00002057, 0.00000975, -0.00000549, 0.00000845, -0.00000914, 0.00000219, -0.00000152, 0.00000061, -0.00000112, 0.00000230, -0.00000333, 0.00000206, -0.00000443, 0.00000236, -0.00000651, -0.00000170, 0.00000623, -0.00001025, 0.00001088, -0.00000816, -0.00000590, 0.00001570, -0.00001921, 0.00003163, -0.00002346, 0.00001546, },
		{ 0.00001218, -0.00000860, 0.00001020, -0.00000440, 0.00001191, 0.00000051, -0.00002062, 0.00003904, -0.00003349, 0.00002398, -0.00001536, -0.00000097, -0.00001052, 0.00000234, 0.00000309, 0.00000158, -0.00001286, 0.00000729, -0.00000865, 0.00000310, -0.00001532, 0.00002112, -0.00002741, 0.00002544, -0.00002126, 0.00001657, -0.00000194, 0.00000144, 0.00000265, -0.00000243, -0.00000609, 0.00000257, -0.00000703, 0.00000798, -0.00000816, 0.00000622, -0.00000696, 0.00000549, -0.00000321, 0.00000301, -0.00000884, 0.00001145, -0.00000359, 0.00000575, -0.00002553, 0.00002601, -0.00001151, 0.00000456, -0.00001780, 0.00001700, -0.00002128, 0.00001845, -0.00000798, 0.00000714, -0.00001884, 0.00001098, -0.00001442, 0.00000897, -0.00002327, 0.00001585, -0.00002237, 0.00001964, -0.00004335, 0.00002879, -0.00003002, 0.00002130, -0.00006488, 0.00002523, -0.00011492, 0.00002590, -0.00056067, 0.00050241, 0.00109682, -0.00056852, -0.00004826, -0.00010514, 0.00000336, -0.00005495, 0.00002410, -0.00003915, 0.00001006, -0.00002085, 0.00002235, -0.00002542, 0.00001108, -0.00002015, 0.00001875, -0.00001872, 0.00000489, -0.00001438, 0.00000887, -0.00001066, 0.00000603, -0.00000604, 0.00002098, -0.00002001, 0.00001176, -0.00001676, 0.00000623, -0.00000857, 0.00002652, -0.00002329, 0.00000677, -0.00000532, 0.00000895, -0.00000720, 0.00000097, -0.00000288, 0.00000358, -0.00000565, 0.00000695, -0.00000693, 0.00000579, -0.00000763, 0.00000456, -0.00000649, -0.00000222, 0.00000195, 0.00000078, -0.00000693, 0.00001537, -0.00002803, 0.00002996, -0.00002564, 0.00002624, -0.00000925, 0.00000173, -0.00000143, },
		{ 0.00001480, -0.00001584, 0.00002219, -0.00001985, 0.00003515, -0.00004219, 0.00003432, -0.00000904, -0.00000484, 0.00001089, -0.00001446, -0.00000248, -0.00000574, 0.00000285, 0.00000750, -0.00001320, 0.00000345, -0.00000550, 0.00000333, -0.00001586, 0.00001515, -0.00001273, 0.00000342, 0.00000478, -0.00001181, 0.00002125, -0.00001423, 0.00001337, -0.00000775, -0.00000035, -0.00000456, 0.00000005, -0.00000149, 0.00000325, -0.00000528, 0.00000407, -0.00000555, 0.00000630, -0.00000383, -0.00000182, 0.00000062, 0.00000602, 0.00000110, -0.00001431, 0.00000177, 0.00001461, -0.00001326, -0.00000142, -0.00000438, 0.00000482, -0.00001203, 0.00001955, -0.00001391, 0.00000584, -0.00001695, 0.00001137, -0.00001408, 0.00000361, -0.00001557, 0.00000874, -0.00001235, 0.00000171, -0.00002741, 0.00002357, -0.00002629, 0.00000054, -0.00004611, -0.00001724, -0.00006551, -0.00018624, -0.00044945, 0.00148067, -0.00005157, -0.00040870, 0.00003581, -0.00010901, 0.00001605, -0.00004648, 0.00002228, -0.00003917, 0.00001234, -0.00001143, 0.00001055, -0.00001908, 0.00000465, -0.00000950, 0.00001094, -0.00001769, 0.00000302, -0.00000916, 0.00000520, -0.00000675, 0.00000098, 0.00001018, -0.00000064, -0.00000620, 0.00000008, -0.00000908, 0.00000060, 0.00001135, 0.00000286, -0.00001749, 0.00001133, -0.00000592, 0.00000819, -0.00001126, 0.00000595, -0.00000386, 0.00000131, -0.00000118, 0.00000223, -0.00000288, 0.00000163, -0.00000493, 0.00000288, -0.00000857, 0.00000196, 0.00000460, -0.00001083, 0.00001562, -0.00001817, 0.00000861, 0.00000246, -0.00000747, 0.00002271, -0.00002100, 0.00001336, -0.00000589, },
		{ -0.00001584, 0.00001676, -0.00001076, 0.00001550, -0.00000286, -0.00001868, 0.00003723, -0.00003202, 0.00002164, -0.00001130, -0.00000504, -0.00000540, -0.00000189, 0.00000841, -0.00000371, -0.00000794, 0.00000338, -0.00000420, -0.00000421, -0.00000499, 0.00001288, -0.00002283, 0.00002743, -0.00002854, 0.00002735, -0.00001339, 0.00001102, -0.00000501, 0.00000186, -0.00000985, 0.00000639, -0.00001064, 0.00001288, -0.00001356, 0.00001175, -0.00001337, 0.00001290, -0.00000982, 0.00000729, -0.00001276, 0.00001782, -0.00000985, 0.00000685, -0.00002733, 0.00003683, -0.00002555, 0.00001375, -0.00002618, 0.00002851, -0.00003482, 0.00003613, -0.00002710, 0.00002514, -0.00003685, 0.00002852, -0.00003286, 0.00002919, -0.00004425, 0.00003756, -0.00004657, 0.00004733, -0.00007361, 0.00006253, -0.00006847, 0.00006537, -0.00011708, 0.00008161, -0.00018850, 0.00012295, -0.00074267, 0.00075650, 0.00108637, -0.00080216, 0.00012575, -0.00020185, 0.00007833, -0.00011488, 0.00007384, -0.00007744, 0.00004317, -0.00004967, 0.00004803, -0.00004883, 0.00003249, -0.00004079, 0.00003792, -0.00003596, 0.00002025, -0.00002906, 0.00002262, -0.00002244, 0.00001584, -0.00001238, 0.00002638, -0.00002695, 0.00001910, -0.00002626, 0.00001447, -0.00001094, 0.00002988, -0.00003451, 0.00001952, -0.00001436, 0.00001725, -0.00001705, 0.00000995, -0.00000869, 0.00000774, -0.00000965, 0.00001097, -0.00001066, 0.00000969, -0.00001201, 0.00000834, -0.00001067, 0.00000100, 0.00000279, -0.00000276, -0.00000017, 0.00000678, -0.00002058, 0.00002840, -0.00002817, 0.00003279, -0.00002007, 0.00000996, -0.00000906, 0.00001902, },
		{ -0.00001144, 0.00001858, -0.00001849, 0.00003269, -0.00004006, 0.00003139, -0.00000580, -0.00000909, 0.00001569, -0.00001833, 0.00000172, -0.00000850, 0.00000628, 0.00000493, -0.00001145, 0.00000260, -0.00000393, 0.00000044, -0.00001392, 0.00001899, -0.00002144, 0.00001689, -0.00000948, 0.00000058, 0.00001279, -0.00001017, 0.00001198, -0.00000872, -0.00000052, -0.00000357, -0.00000132, 0.00000120, 0.00000062, -0.00000279, 0.00000118, -0.00000284, 0.00000531, -0.00000437, -0.00000259, 0.00000374, 0.00000415, 0.00000014, -0.00001836, 0.00001357, 0.00000511, -0.00000994, -0.00000622, 0.00000395, -0.00000432, -0.00000128, 0.00001084, -0.00000819, 0.00000072, -0.00001238, 0.00000587, -0.00000778, -0.00000228, -0.00000938, 0.00000085, -0.00000315, -0.00000744, -0.00001731, 0.00001229, -0.00001443, -0.00001205, -0.00003626, -0.00003358, -0.00004682, -0.00022005, -0.00049398, 0.00180711, -0.00042880, -0.00026277, -0.00000877, -0.00007449, -0.00000795, -0.00002903, 0.00001144, -0.00002838, 0.00000275, -0.00000255, 0.00000202, -0.00001106, -0.00000331, -0.00000278, 0.00000546, -0.00001300, -0.00000167, -0.00000527, 0.00000275, -0.00000515, 0.00000052, 0.00001268, -0.00000605, -0.00000025, -0.00000695, -0.00000465, -0.00000059, 0.00001767, -0.00000978, -0.00000815, 0.00000712, -0.00000155, 0.00000296, -0.00000832, 0.00000572, -0.00000351, -0.00000014, 0.00000078, 0.00000021, -0.00000052, -0.00000089, -0.00000334, 0.00000141, -0.00000890, 0.00000534, 0.00000127, -0.00000793, 0.00001552, -0.00002326, 0.00002033, -0.00001141, 0.00000727, 0.00000859, -0.00001264, 0.00000636, -0.00000014, 0.00000916, },
		{ 0.00002043, -0.00001503, 0.00001729, -0.00000475, -0.00001748, 0.00003598, -0.00003114, 0.00002031, -0.00000907, -0.00000687, -0.00000230, -0.00000407, 0.00001159, -0.00000740, -0.00000389, 0.00000020, -0.00000099, -0.00001020, 0.00000597, 0.00000101, -0.00001195, 0.00002160, -0.00002842, 0.00003235, -0.00002151, 0.00001948, -0.00001313, 0.00000709, -0.00001412, 0.00001045, -0.00001408, 0.00001727, -0.00001851, 0.00001690, -0.00001949, 0.00002062, -0.00001763, 0.00001287, -0.00001711, 0.00002434, -0.00001716, 0.00000839, -0.00002640, 0.00004358, -0.00003762, 0.00002235, -0.00003297, 0.00003733, -0.00004508, 0.00005088, -0.00004489, 0.00004298, -0.00005514, 0.00004657, -0.00005188, 0.00005019, -0.00006617, 0.00006010, -0.00007138, 0.00007566, -0.00010465, 0.00009753, -0.00010928, 0.00011308, -0.00017503, 0.00014587, -0.00027248, 0.00023559, -0.00096381, 0.00111602, 0.00092614, -0.00093519, 0.00026636, -0.00028831, 0.00014796, -0.00017234, 0.00012330, -0.00011647, 0.00007710, -0.00007925, 0.00007402, -0.00007233, 0.00005388, -0.00006135, 0.00005725, -0.00005376, 0.00003639, -0.00004454, 0.00003751, -0.00003580, 0.00002761, -0.00002016, 0.00003162, -0.00003269, 0.00002482, -0.00003443, 0.00002258, -0.00001244, 0.00002940, -0.00004108, 0.00002968, -0.00002204, 0.00002404, -0.00002580, 0.00001910, -0.00001537, 0.00001246, -0.00001374, 0.00001483, -0.00001407, 0.00001306, -0.00001603, 0.00001206, -0.00001536, 0.00000600, 0.00000107, -0.00000398, 0.00000546, -0.00000323, -0.00000814, 0.00001899, -0.00002222, 0.00003097, -0.00002412, 0.00001355, -0.00001269, 0.00002197, -0.00001943, },
		{ 0.00001189, -0.00001428, 0.00002742, -0.00003502, 0.00002617, -0.00000090, -0.00001429, 0.00002086, -0.00002280, 0.00000706, -0.00001253, 0.00001126, 0.00000006, -0.00000694, -0.00000080, -0.00000002, -0.00000554, -0.00000671, 0.00001596, -0.00002338, 0.00002535, -0.00002190, 0.00001423, 0.00000060, -0.00000123, 0.00000567, -0.00000568, -0.00000357, -0.00000008, -0.00000493, 0.00000616, -0.00000470, 0.00000268, -0.00000504, 0.00000382, -0.00000002, -0.00000118, -0.00000615, 0.00000979, -0.00000125, 0.00000149, -0.00002272, 0.00002706, -0.00000941, -0.00000087, -0.00001557, 0.00001678, -0.00001893, 0.00001670, -0.00000719, 0.00000770, -0.00001474, 0.00000278, -0.00001055, 0.00001007, -0.00002034, 0.00000956, -0.00002065, 0.00002104, -0.00003333, 0.00001182, -0.00002091, 0.00002252, -0.00005383, 0.00000736, -0.00008999, 0.00002298, -0.00032867, -0.00043299, 0.00205876, -0.00082839, -0.00004314, -0.00011475, 0.00000490, -0.00006821, 0.00001899, -0.00002509, 0.00000394, -0.00002530, 0.00002225, -0.00002075, 0.00000994, -0.00002323, 0.00001527, -0.00001075, 0.00000159, -0.00001544, 0.00000720, -0.00000765, 0.00000347, -0.00000530, 0.00001872, -0.00001455, 0.00000898, -0.00001779, 0.00000402, -0.00000430, 0.00002444, -0.00002425, 0.00000578, -0.00000204, 0.00000705, -0.00000680, -0.00000026, 0.00000091, 0.00000024, -0.00000423, 0.00000506, -0.00000392, 0.00000381, -0.00000551, 0.00000052, -0.00000249, -0.00000658, 0.00000699, -0.00000222, -0.00000258, 0.00001041, -0.00002148, 0.00002536, -0.00002092, 0.00001993, -0.00000615, -0.00000197, -0.00000274, 0.00000754, 0.00000134, -0.00000425, },
		{ -0.00001539, 0.00001558, -0.00000353, -0.00001847, 0.00003645, -0.00003172, 0.00002064, -0.00000927, -0.00000572, -0.00000219, -0.00000303, 0.00001116, -0.00000771, -0.00000259, -0.00000040, -0.00000104, -0.00001200, 0.00001367, -0.00001020, 0.00000156, 0.00001008, -0.00002118, 0.00003024, -0.00002390, 0.00002403, -0.00001914, 0.00001129, -0.00001734, 0.00001337, -0.00001607, 0.00001975, -0.00002143, 0.00001990, -0.00002330, 0.00002636, -0.00002448, 0.00001802, -0.00002042, 0.00002940, -0.00002408, 0.00000975, -0.00002221, 0.00004449, -0.00004484, 0.00002753, -0.00003563, 0.00004063, -0.00004846, 0.00005801, -0.00005583, 0.00005479, -0.00006773, 0.00005898, -0.00006508, 0.00006524, -0.00008195, 0.00007602, -0.00008872, 0.00009564, -0.00012641, 0.00012237, -0.00013937, 0.00014941, -0.00022148, 0.00019819, -0.00034306, 0.00033307, -0.00118529, 0.00154832, 0.00061790, -0.00093826, 0.00033717, -0.00033509, 0.00018914, -0.00020815, 0.00015633, -0.00014273, 0.00010043, -0.00009974, 0.00009168, -0.00008815, 0.00006814, -0.00007518, 0.00007048, -0.00006630, 0.00004795, -0.00005589, 0.00004894, -0.00004664, 0.00003799, -0.00002701, 0.00003493, -0.00003556, 0.00002706, -0.00003920, 0.00002887, -0.00001236, 0.00002444, -0.00004125, 0.00003468, -0.00002580, 0.00002679, -0.00003071, 0.00002570, -0.00002065, 0.00001600, -0.00001653, 0.00001731, -0.00001604, 0.00001480, -0.00001847, 0.00001445, -0.00001923, 0.00001173, -0.00000303, -0.00000180, 0.00000749, -0.00001091, 0.00000501, 0.00000521, -0.00000981, 0.00002133, -0.00002074, 0.00001134, -0.00001112, 0.00001977, -0.00001785, 0.00001941, },
		{ -0.00000958, 0.00002147, -0.00002891, 0.00002020, 0.00000427, -0.00001900, 0.00002481, -0.00002596, 0.00001147, -0.00001567, 0.00001548, -0.00000478, -0.00000179, -0.00000485, 0.00000421, -0.00001223, 0.00000350, 0.00000730, -0.00001801, 0.00002639, -0.00002879, 0.00002509, -0.00001168, 0.00000985, -0.00000390, 0.00000060, -0.00000909, 0.00000552, -0.00001032, 0.00001274, -0.00001192, 0.00001031, -0.00001370, 0.00001365, -0.00000932, 0.00000582, -0.00001261, 0.00001860, -0.00001018, 0.00000540, -0.00002687, 0.00004020, -0.00002649, 0.00001208, -0.00002782, 0.00003193, -0.00003610, 0.00003846, -0.00003115, 0.00003086, -0.00003806, 0.00002626, -0.00003570, 0.00003726, -0.00004839, 0.00003902, -0.00005328, 0.00005745, -0.00007291, 0.00005669, -0.00007264, 0.00008128, -0.00012198, 0.00008266, -0.00018523, 0.00014421, -0.00051928, -0.00023766, 0.00217719, -0.00118403, 0.00020311, -0.00025560, 0.00011274, -0.00015307, 0.00008911, -0.00008110, 0.00005300, -0.00006787, 0.00005956, -0.00005466, 0.00004103, -0.00005242, 0.00004222, -0.00003560, 0.00002435, -0.00003681, 0.00002704, -0.00002524, 0.00001886, -0.00001659, 0.00002822, -0.00002547, 0.00002034, -0.00003111, 0.00001600, -0.00000996, 0.00003031, -0.00003791, 0.00002143, -0.00001390, 0.00001791, -0.00001918, 0.00001143, -0.00000767, 0.00000694, -0.00001055, 0.00001121, -0.00000966, 0.00000950, -0.00001165, 0.00000614, -0.00000856, -0.00000124, 0.00000564, -0.00000388, 0.00000298, 0.00000219, -0.00001341, 0.00002223, -0.00002279, 0.00002626, -0.00001699, 0.00000697, -0.00001046, 0.00001385, -0.00000544, 0.00000267, 0.00000484, },
		{ 0.00001012, 0.00000113, -0.00002200, 0.00003882, -0.00003366, 0.00002226, -0.00001133, -0.00000212, -0.00000465, 0.00000089, 0.00000722, -0.00000435, -0.00000459, 0.00000213, -0.00000500, -0.00000816, 0.00001524, -0.00001668, 0.00001314, -0.00000318, -0.00000936, 0.00002197, -0.00002003, 0.00002316, -0.00002118, 0.00001276, -0.00001814, 0.00001397, -0.00001554, 0.00001924, -0.00002106, 0.00001932, -0.00002307, 0.00002798, -0.00002806, 0.00002076, -0.00002101, 0.00003121, -0.00002897, 0.00001023, -0.00001482, 0.00003884, -0.00004539, 0.00002732, -0.00003254, 0.00003688, -0.00004296, 0.00005455, -0.00005591, 0.00005599, -0.00006967, 0.00006055, -0.00006692, 0.00006839, -0.00008531, 0.00007865, -0.00009137, 0.00009929, -0.00012984, 0.00012655, -0.00014645, 0.00015975, -0.00023899, 0.00021763, -0.00037396, 0.00037912, -0.00135530, 0.00199620, 0.00019540, -0.00081040, 0.00032047, -0.00032385, 0.00018594, -0.00020814, 0.00016032, -0.00014521, 0.00010359, -0.00010282, 0.00009372, -0.00008973, 0.00006928, -0.00007665, 0.00007222, -0.00006843, 0.00005009, -0.00005850, 0.00005249, -0.00005077, 0.00004333, -0.00003024, 0.00003451, -0.00003414, 0.00002441, -0.00003889, 0.00003191, -0.00001037, 0.00001538, -0.00003470, 0.00003327, -0.00002422, 0.00002405, -0.00002990, 0.00002755, -0.00002253, 0.00001680, -0.00001675, 0.00001735, -0.00001563, 0.00001399, -0.00001832, 0.00001434, -0.00002077, 0.00001651, -0.00000825, 0.00000352, 0.00000468, -0.00001324, 0.00001435, -0.00000797, 0.00000473, 0.00000712, -0.00001190, 0.00000454, -0.00000538, 0.00001324, -0.00001160, 0.00001382, -0.00001171, },
		{ 0.00001709, -0.00002374, 0.00001514, 0.00000830, -0.00002195, 0.00002622, -0.00002625, 0.00001313, -0.00001596, 0.00001671, -0.00000711, 0.00000152, -0.00000717, 0.00000621, -0.00001642, 0.00001292, -0.00000357, -0.00000751, 0.00002023, -0.00002853, 0.00003025, -0.00002062, 0.00001983, -0.00001417, 0.00000832, -0.00001577, 0.00001207, -0.00001634, 0.00001965, -0.00001954, 0.00001843, -0.00002299, 0.00002474, -0.00002095, 0.00001546, -0.00002099, 0.00002905, -0.00002170, 0.00001151, -0.00003009, 0.00005080, -0.00004299, 0.00002604, -0.00004030, 0.00004631, -0.00005187, 0.00005908, -0.00005561, 0.00005587, -0.00006392, 0.00005275, -0.00006413, 0.00006815, -0.00008059, 0.00007286, -0.00009038, 0.00009871, -0.00011800, 0.00010814, -0.00013273, 0.00015058, -0.00020419, 0.00017637, -0.00030450, 0.00030000, -0.00077657, 0.00009508, 0.00212398, -0.00143544, 0.00042299, -0.00039474, 0.00022279, -0.00024204, 0.00016489, -0.00014334, 0.00010784, -0.00011567, 0.00010129, -0.00009244, 0.00007555, -0.00008468, 0.00007228, -0.00006380, 0.00005052, -0.00006148, 0.00005039, -0.00004671, 0.00003849, -0.00003164, 0.00003978, -0.00003721, 0.00003188, -0.00004476, 0.00002955, -0.00001659, 0.00003395, -0.00004812, 0.00003554, -0.00002546, 0.00002824, -0.00003128, 0.00002406, -0.00001795, 0.00001511, -0.00001788, 0.00001809, -0.00001590, 0.00001544, -0.00001811, 0.00001238, -0.00001577, 0.00000656, 0.00000083, -0.00000206, 0.00000602, -0.00000587, -0.00000202, 0.00001247, -0.00001670, 0.00002443, -0.00002110, 0.00001124, -0.00001412, 0.00001617, -0.00000840, 0.00000646, 0.00000012, -0.00000685, },
		{ 0.00000811, -0.00002720, 0.00004220, -0.00003595, 0.00002388, -0.00001363, 0.00000218, -0.00000794, 0.00000589, 0.00000148, 0.00000130, -0.00000874, 0.00000662, -0.00001159, 0.00000050, 0.00001012, -0.00001613, 0.00001904, -0.00001388, 0.00000319, 0.00001027, -0.00001130, 0.00001703, -0.00001864, 0.00001071, -0.00001578, 0.00001164, -0.00001207, 0.00001535, -0.00001697, 0.00001462, -0.00001800, 0.00002421, -0.00002671, 0.00001950, -0.00001758, 0.00002837, -0.00003031, 0.00000930, -0.00000491, 0.00002728, -0.00003899, 0.00002119, -0.00002359, 0.00002638, -0.00002894, 0.00004029, -0.00004406, 0.00004484, -0.00005885, 0.00004888, -0.00005472, 0.00005665, -0.00007297, 0.00006455, -0.00007566, 0.00008250, -0.00011022, 0.00010440, -0.00012345, 0.00013507, -0.00021593, 0.00018929, -0.00034513, 0.00034257, -0.00142117, 0.00238883, -0.00028145, -0.00057953, 0.00022180, -0.00025310, 0.00013509, -0.00016813, 0.00013060, -0.00011936, 0.00008239, -0.00008472, 0.00007681, -0.00007414, 0.00005463, -0.00006323, 0.00005991, -0.00005758, 0.00004027, -0.00004981, 0.00004549, -0.00004537, 0.00004091, -0.00002779, 0.00002917, -0.00002780, 0.00001637, -0.00003272, 0.00003085, -0.00000670, 0.00000360, -0.00002265, 0.00002589, -0.00001746, 0.00001589, -0.00002304, 0.00002377, -0.00001995, 0.00001397, -0.00001372, 0.00001441, -0.00001248, 0.00001030, -0.00001512, 0.00001100, -0.00001876, 0.00001848, -0.00001251, 0.00001024, -0.00000225, -0.00000921, 0.00001679, -0.00001602, 0.00001647, -0.00000706, -0.00000127, -0.00000395, 0.00000185, 0.00000496, -0.00000300, 0.00000555, -0.00000551, 0.00000226, },
		{ -0.00002091, 0.00001208, 0.00001044, -0.00002265, 0.00002473, -0.00002317, 0.00001136, -0.00001255, 0.00001382, -0.00000542, 0.00000130, -0.00000608, 0.00000414, -0.00001544, 0.00001776, -0.00001221, 0.00000403, 0.00000959, -0.00002192, 0.00002882, -0.00002408, 0.00002599, -0.00002246, 0.00001530, -0.00002183, 0.00001799, -0.00002150, 0.00002530, -0.00002578, 0.00002505, -0.00003061, 0.00003448, -0.00003230, 0.00002558, -0.00002942, 0.00003917, -0.00003400, 0.00001886, -0.00003166, 0.00005696, -0.00005584, 0.00003785, -0.00005013, 0.00005679, -0.00006239, 0.00007360, -0.00007463, 0.00007635, -0.00008573, 0.00007543, -0.00008871, 0.00009522, -0.00010902, 0.00010271, -0.00012290, 0.00013484, -0.00015744, 0.00015352, -0.00018672, 0.00021385, -0.00028144, 0.00026667, -0.00042199, 0.00045811, -0.00106232, 0.00053732, 0.00189204, -0.00154318, 0.00057164, -0.00049648, 0.00030723, -0.00031217, 0.00022700, -0.00019556, 0.00015465, -0.00015679, 0.00013698, -0.00012471, 0.00010494, -0.00011204, 0.00009796, -0.00008837, 0.00007363, -0.00008345, 0.00007165, -0.00006696, 0.00005806, -0.00004722, 0.00005095, -0.00004749, 0.00004119, -0.00005611, 0.00004243, -0.00002297, 0.00003446, -0.00005295, 0.00004528, -0.00003391, 0.00003527, -0.00004010, 0.00003458, -0.00002731, 0.00002266, -0.00002446, 0.00002414, -0.00002124, 0.00002023, -0.00002341, 0.00001767, -0.00002244, 0.00001532, -0.00000674, 0.00000378, 0.00000452, -0.00001029, 0.00000838, -0.00000002, -0.00000526, 0.00001555, -0.00001827, 0.00001005, -0.00001294, 0.00001363, -0.00000640, 0.00000566, -0.00000071, -0.00000757, 0.00001578, },
		{ -0.00003242, 0.00004513, -0.00003712, 0.00002388, -0.00001421, 0.00000499, -0.00000971, 0.00000944, -0.00000341, 0.00000668, -0.00001268, 0.00001062, -0.00001791, 0.00001104, 0.00000036, -0.00000895, 0.00001764, -0.00001891, 0.00001276, -0.00000135, -0.00000040, 0.00000732, -0.00001225, 0.00000543, -0.00001048, 0.00000662, -0.00000602, 0.00000861, -0.00000974, 0.00000641, -0.00000860, 0.00001513, -0.00002000, 0.00001370, -0.00000974, 0.00002032, -0.00002722, 0.00000673, 0.00000624, 0.00001163, -0.00002691, 0.00001016, -0.00001025, 0.00001126, -0.00000916, 0.00001799, -0.00002257, 0.00002320, -0.00003684, 0.00002541, -0.00002985, 0.00003127, -0.00004615, 0.00003504, -0.00004308, 0.00004690, -0.00006925, 0.00005753, -0.00007161, 0.00007581, -0.00015136, 0.00011041, -0.00025067, 0.00020828, -0.00134386, 0.00265771, -0.00073972, -0.00029518, 0.00006761, -0.00013837, 0.00004721, -0.00009555, 0.00007234, -0.00006892, 0.00003971, -0.00004783, 0.00004311, -0.00004335, 0.00002606, -0.00003660, 0.00003495, -0.00003483, 0.00001932, -0.00003039, 0.00002809, -0.00003007, 0.00002995, -0.00001901, 0.00001885, -0.00001702, 0.00000369, -0.00002105, 0.00002563, -0.00000202, -0.00000884, -0.00000751, 0.00001442, -0.00000714, 0.00000397, -0.00001142, 0.00001500, -0.00001316, 0.00000768, -0.00000765, 0.00000877, -0.00000693, 0.00000415, -0.00000923, 0.00000459, -0.00001275, 0.00001627, -0.00001355, 0.00001565, -0.00001077, -0.00000027, 0.00001185, -0.00001653, 0.00002171, -0.00001703, 0.00000733, -0.00001084, 0.00000746, -0.00000190, 0.00000486, -0.00000260, 0.00000071, -0.00000573, 0.00001544, },
		{ 0.00001102, 0.00001096, -0.00002170, 0.00002125, -0.00001764, 0.00000701, -0.00000625, 0.00000742, 0.00000006, -0.00000251, -0.00000150, -0.00000215, -0.00000840, 0.00001571, -0.00001499, 0.00001218, -0.00000141, -0.00001183, 0.00002203, -0.00002177, 0.00002694, -0.00002685, 0.00001968, -0.00002563, 0.00002181, -0.00002444, 0.00002829, -0.00002906, 0.00002840, -0.00003445, 0.00004028, -0.00004053, 0.00003357, -0.00003557, 0.00004654, -0.00004476, 0.00002606, -0.00003103, 0.00005752, -0.00006273, 0.00004496, -0.00005502, 0.00006118, -0.00006509, 0.00007847, -0.00008360, 0.00008700, -0.00009774, 0.00008818, -0.00010286, 0.00011138, -0.00012614, 0.00012057, -0.00014222, 0.00015631, -0.00018051, 0.00018050, -0.00022023, 0.00025407, -0.00033348, 0.00032932, -0.00050766, 0.00057833, -0.00132214, 0.00103307, 0.00150926, -0.00149799, 0.00062362, -0.00053637, 0.00034527, -0.00034525, 0.00025934, -0.00022366, 0.00018114, -0.00018045, 0.00015714, -0.00014293, 0.00012144, -0.00012725, 0.00011234, -0.00010272, 0.00008748, -0.00009682, 0.00008516, -0.00008060, 0.00007270, -0.00005947, 0.00005887, -0.00005399, 0.00004603, -0.00006267, 0.00005236, -0.00002804, 0.00003165, -0.00005164, 0.00004900, -0.00003741, 0.00003718, -0.00004344, 0.00004041, -0.00003330, 0.00002753, -0.00002856, 0.00002786, -0.00002436, 0.00002259, -0.00002618, 0.00002044, -0.00002664, 0.00002287, -0.00001516, 0.00001263, -0.00000195, -0.00000881, 0.00001382, -0.00001045, 0.00000707, 0.00000314, -0.00001076, 0.00000485, -0.00000818, 0.00000738, -0.00000031, 0.00000073, 0.00000217, -0.00001168, 0.00001759, -0.00002057, },
		{ 0.00004628, -0.00003602, 0.00002109, -0.00001164, 0.00000462, -0.00000809, 0.00000936, -0.00000493, 0.00000921, -0.00001388, 0.00001141, -0.00002055, 0.00001933, -0.00001002, 0.00000187, 0.00001009, -0.00001738, 0.00001702, -0.00000985, 0.00000959, -0.00000345, -0.00000379, -0.00000181, -0.00000331, -0.00000007, 0.00000145, 0.00000035, -0.00000085, -0.00000362, 0.00000329, 0.00000234, -0.00000899, 0.00000415, 0.00000161, 0.00000773, -0.00001973, 0.00000275, 0.00001698, -0.00000544, -0.00001167, -0.00000351, 0.00000489, -0.00000507, 0.00001204, -0.00000736, 0.00000356, -0.00000404, -0.00000854, -0.00000485, 0.00000254, -0.00000244, -0.00001037, -0.00000390, -0.00000021, -0.00000020, -0.00001513, -0.00000497, -0.00000093, -0.00000725, -0.00005660, -0.00000701, -0.00010316, -0.00001535, -0.00111031, 0.00275336, -0.00111031, -0.00001535, -0.00010316, -0.00000701, -0.00005660, -0.00000725, -0.00000093, -0.00000497, -0.00001513, -0.00000020, -0.00000021, -0.00000390, -0.00001037, -0.00000244, 0.00000254, -0.00000485, -0.00000854, -0.00000404, 0.00000356, -0.00000736, 0.00001204, -0.00000507, 0.00000489, -0.00000351, -0.00001167, -0.00000544, 0.00001698, 0.00000275, -0.00001973, 0.00000773, 0.00000161, 0.00000415, -0.00000899, 0.00000234, 0.00000329, -0.00000362, -0.00000085, 0.00000035, 0.00000145, -0.00000007, -0.00000331, -0.00000181, -0.00000379, -0.00000345, 0.00000959, -0.00000985, 0.00001702, -0.00001738, 0.00001009, 0.00000187, -0.00001002, 0.00001933, -0.00002055, 0.00001141, -0.00001388, 0.00000921, -0.00000493, 0.00000936, -0.00000809, 0.00000462, -0.00001164, 0.00002109, -0.00003602, },
		{ 0.00001102, -0.00002057, 0.00001759, -0.00001168, 0.00000217, 0.00000073, -0.00000031, 0.00000738, -0.00000818, 0.00000485, -0.00001076, 0.00000314, 0.00000707, -0.00001045, 0.00001382, -0.00000881, -0.00000195, 0.00001263, -0.00001516, 0.00002287, -0.00002664, 0.00002044, -0.00002618, 0.00002259, -0.00002436, 0.00002786, -0.00002856, 0.00002753, -0.00003330, 0.00004041, -0.00004344, 0.00003718, -0.00003741, 0.00004900, -0.00005164, 0.00003165, -0.00002804, 0.00005236, -0.00006267, 0.00004603, -0.00005399, 0.00005887, -0.00005947, 0.00007270, -0.00008060, 0.00008516, -0.00009682, 0.00008748, -0.00010272, 0.00011234, -0.00012725, 0.00012144, -0.00014293, 0.00015714, -0.00018045, 0.00018114, -0.00022366, 0.00025934, -0.00034525, 0.00034527, -0.00053637, 0.00062362, -0.00149799, 0.00150926, 0.00103307, -0.00132214, 0.00057833, -0.00050766, 0.00032932, -0.00033348, 0.00025407, -0.00022023, 0.00018050, -0.00018051, 0.00015631, -0.00014222, 0.00012057, -0.00012614, 0.00011138, -0.00010286, 0.00008818, -0.00009774, 0.00008700, -0.00008360, 0.00007847, -0.00006509, 0.00006118, -0.00005502, 0.00004496, -0.00006273, 0.00005752, -0.00003103, 0.00002606, -0.00004476, 0.00004654, -0.00003557, 0.00003357, -0.00004053, 0.00004028, -0.00003445, 0.00002840, -0.00002906, 0.00002829, -0.00002444, 0.00002181, -0.00002563, 0.00001968, -0.00002685, 0.00002694, -0.00002177, 0.00002203, -0.00001183, -0.00000141, 0.00001218, -0.00001499, 0.00001571, -0.00000840, -0.00000215, -0.00000150, -0.00000251, 0.00000006, 0.00000742, -0.00000625, 0.00000701, -0.00001764, 0.00002125, -0.00002170, 0.00001096, },
		{ -0.00003242, 0.00001544, -0.00000573, 0.00000071, -0.00000260, 0.00000486, -0.00000190, 0.00000746, -0.00001084, 0.00000733, -0.00001703, 0.00002171, -0.00001653, 0.00001185, -0.00000027, -0.00001077, 0.00001565, -0.00001355, 0.00001627, -0.00001275, 0.00000459, -0.00000923, 0.00000415, -0.00000693, 0.00000877, -0.00000765, 0.00000768, -0.00001316, 0.00001500, -0.00001142, 0.00000397, -0.00000714, 0.00001442, -0.00000751, -0.00000884, -0.00000202, 0.00002563, -0.00002105, 0.00000369, -0.00001702, 0.00001885, -0.00001901, 0.00002995, -0.00003007, 0.00002809, -0.00003039, 0.00001932, -0.00003483, 0.00003495, -0.00003660, 0.00002606, -0.00004335, 0.00004311, -0.00004783, 0.00003971, -0.00006892, 0.00007234, -0.00009555, 0.00004721, -0.00013837, 0.00006761, -0.00029518, -0.00073972, 0.00265771, -0.00134386, 0.00020828, -0.00025067, 0.00011041, -0.00015136, 0.00007581, -0.00007161, 0.00005753, -0.00006925, 0.00004690, -0.00004308, 0.00003504, -0.00004615, 0.00003127, -0.00002985, 0.00002541, -0.00003684, 0.00002320, -0.00002257, 0.00001799, -0.00000916, 0.00001126, -0.00001025, 0.00001016, -0.00002691, 0.00001163, 0.00000624, 0.00000673, -0.00002722, 0.00002032, -0.00000974, 0.00001370, -0.00002000, 0.00001513, -0.00000860, 0.00000641, -0.00000974, 0.00000861, -0.00000602, 0.00000662, -0.00001048, 0.00000543, -0.00001225, 0.00000732, -0.00000040, -0.00000135, 0.00001276, -0.00001891, 0.00001764, -0.00000895, 0.00000036, 0.00001104, -0.00001791, 0.00001062, -0.00001268, 0.00000668, -0.00000341, 0.00000944, -0.00000971, 0.00000499, -0.00001421, 0.00002388, -0.00003712, 0.00004513, },
		{ -0.00002091, 0.00001578, -0.00000757, -0.00000071, 0.00000566, -0.00000640, 0.00001363, -0.00001294, 0.00001005, -0.00001827, 0.00001555, -0.00000526, -0.00000002, 0.00000838, -0.00001029, 0.00000452, 0.00000378, -0.00000674, 0.00001532, -0.00002244, 0.00001767, -0.00002341, 0.00002023, -0.00002124, 0.00002414, -0.00002446, 0.00002266, -0.00002731, 0.00003458, -0.00004010, 0.00003527, -0.00003391, 0.00004528, -0.00005295, 0.00003446, -0.00002297, 0.00004243, -0.00005611, 0.00004119, -0.00004749, 0.00005095, -0.00004722, 0.00005806, -0.00006696, 0.00007165, -0.00008345, 0.00007363, -0.00008837, 0.00009796, -0.00011204, 0.00010494, -0.00012471, 0.00013698, -0.00015679, 0.00015465, -0.00019556, 0.00022700, -0.00031217, 0.00030723, -0.00049648, 0.00057164, -0.00154318, 0.00189204, 0.00053732, -0.00106232, 0.00045811, -0.00042199, 0.00026667, -0.00028144, 0.00021385, -0.00018672, 0.00015352, -0.00015744, 0.00013484, -0.00012290, 0.00010271, -0.00010902, 0.00009522, -0.00008871, 0.00007543, -0.00008573, 0.00007635, -0.00007463, 0.00007360, -0.00006239, 0.00005679, -0.00005013, 0.00003785, -0.00005584, 0.00005696, -0.00003166, 0.00001886, -0.00003400, 0.00003917, -0.00002942, 0.00002558, -0.00003230, 0.00003448, -0.00003061, 0.00002505, -0.00002578, 0.00002530, -0.00002150, 0.00001799, -0.00002183, 0.00001530, -0.00002246, 0.00002599, -0.00002408, 0.00002882, -0.00002192, 0.00000959, 0.00000403, -0.00001221, 0.00001776, -0.00001544, 0.00000414, -0.00000608, 0.00000130, -0.00000542, 0.00001382, -0.00001255, 0.00001136, -0.00002317, 0.00002473, -0.00002265, 0.00001044, 0.00001208, },
		{ 0.00000811, 0.00000226, -0.00000551, 0.00000555, -0.00000300, 0.00000496, 0.00000185, -0.00000395, -0.00000127, -0.00000706, 0.00001647, -0.00001602, 0.00001679, -0.00000921, -0.00000225, 0.00001024, -0.00001251, 0.00001848, -0.00001876, 0.00001100, -0.00001512, 0.00001030, -0.00001248, 0.00001441, -0.00001372, 0.00001397, -0.00001995, 0.00002377, -0.00002304, 0.00001589, -0.00001746, 0.00002589, -0.00002265, 0.00000360, -0.00000670, 0.00003085, -0.00003272, 0.00001637, -0.00002780, 0.00002917, -0.00002779, 0.00004091, -0.00004537, 0.00004549, -0.00004981, 0.00004027, -0.00005758, 0.00005991, -0.00006323, 0.00005463, -0.00007414, 0.00007681, -0.00008472, 0.00008239, -0.00011936, 0.00013060, -0.00016813, 0.00013509, -0.00025310, 0.00022180, -0.00057953, -0.00028145, 0.00238883, -0.00142117, 0.00034257, -0.00034513, 0.00018929, -0.00021593, 0.00013507, -0.00012345, 0.00010440, -0.00011022, 0.00008250, -0.00007566, 0.00006455, -0.00007297, 0.00005665, -0.00005472, 0.00004888, -0.00005885, 0.00004484, -0.00004406, 0.00004029, -0.00002894, 0.00002638, -0.00002359, 0.00002119, -0.00003899, 0.00002728, -0.00000491, 0.00000930, -0.00003031, 0.00002837, -0.00001758, 0.00001950, -0.00002671, 0.00002421, -0.00001800, 0.00001462, -0.00001697, 0.00001535, -0.00001207, 0.00001164, -0.00001578, 0.00001071, -0.00001864, 0.00001703, -0.00001130, 0.00001027, 0.00000319, -0.00001388, 0.00001904, -0.00001613, 0.00001012, 0.00000050, -0.00001159, 0.00000662, -0.00000874, 0.00000130, 0.00000148, 0.00000589, -0.00000794, 0.00000218, -0.00001363, 0.00002388, -0.00003595, 0.00004220, -0.00002720, },
		{ 0.00001709, -0.00000685, 0.00000012, 0.00000646, -0.00000840, 0.00001617, -0.00001412, 0.00001124, -0.00002110, 0.00002443, -0.00001670, 0.00001247, -0.00000202, -0.00000587, 0.00000602, -0.00000206, 0.00000083, 0.00000656, -0.00001577, 0.00001238, -0.00001811, 0.00001544, -0.00001590, 0.00001809, -0.00001788, 0.00001511, -0.00001795, 0.00002406, -0.00003128, 0.00002824, -0.00002546, 0.00003554, -0.00004812, 0.00003395, -0.00001659, 0.00002955, -0.00004476, 0.00003188, -0.00003721, 0.00003978, -0.00003164, 0.00003849, -0.00004671, 0.00005039, -0.00006148, 0.00005052, -0.00006380, 0.00007228, -0.00008468, 0.00007555, -0.00009244, 0.00010129, -0.00011567, 0.00010784, -0.00014334, 0.00016489, -0.00024204, 0.00022279, -0.00039474, 0.00042299, -0.00143544, 0.00212398, 0.00009508, -0.00077657, 0.00030000, -0.00030450, 0.00017637, -0.00020419, 0.00015058, -0.00013273, 0.00010814, -0.00011800, 0.00009871, -0.00009038, 0.00007286, -0.00008059, 0.00006815, -0.00006413, 0.00005275, -0.00006392, 0.00005587, -0.00005561, 0.00005908, -0.00005187, 0.00004631, -0.00004030, 0.00002604, -0.00004299, 0.00005080, -0.00003009, 0.00001151, -0.00002170, 0.00002905, -0.00002099, 0.00001546, -0.00002095, 0.00002474, -0.00002299, 0.00001843, -0.00001954, 0.00001965, -0.00001634, 0.00001207, -0.00001577, 0.00000832, -0.00001417, 0.00001983, -0.00002062, 0.00003025, -0.00002853, 0.00002023, -0.00000751, -0.00000357, 0.00001292, -0.00001642, 0.00000621, -0.00000717, 0.00000152, -0.00000711, 0.00001671, -0.00001596, 0.00001313, -0.00002625, 0.00002622, -0.00002195, 0.00000830, 0.00001514, -0.00002374, },
		{ 0.00001012, -0.00001171, 0.00001382, -0.00001160, 0.00001324, -0.00000538, 0.00000454, -0.00001190, 0.00000712, 0.00000473, -0.00000797, 0.00001435, -0.00001324, 0.00000468, 0.00000352, -0.00000825, 0.00001651, -0.00002077, 0.00001434, -0.00001832, 0.00001399, -0.00001563, 0.00001735, -0.00001675, 0.00001680, -0.00002253, 0.00002755, -0.00002990, 0.00002405, -0.00002422, 0.00003327, -0.00003470, 0.00001538, -0.00001037, 0.00003191, -0.00003889, 0.00002441, -0.00003414, 0.00003451, -0.00003024, 0.00004333, -0.00005077, 0.00005249, -0.00005850, 0.00005009, -0.00006843, 0.00007222, -0.00007665, 0.00006928, -0.00008973, 0.00009372, -0.00010282, 0.00010359, -0.00014521, 0.00016032, -0.00020814, 0.00018594, -0.00032385, 0.00032047, -0.00081040, 0.00019540, 0.00199620, -0.00135530, 0.00037912, -0.00037396, 0.00021763, -0.00023899, 0.00015975, -0.00014645, 0.00012655, -0.00012984, 0.00009929, -0.00009137, 0.00007865, -0.00008531, 0.00006839, -0.00006692, 0.00006055, -0.00006967, 0.00005599, -0.00005591, 0.00005455, -0.00004296, 0.00003688, -0.00003254, 0.00002732, -0.00004539, 0.00003884, -0.00001482, 0.00001023, -0.00002897, 0.00003121, -0.00002101, 0.00002076, -0.00002806, 0.00002798, -0.00002307, 0.00001932, -0.00002106, 0.00001924, -0.00001554, 0.00001397, -0.00001814, 0.00001276, -0.00002118, 0.00002316, -0.00002003, 0.00002197, -0.00000936, -0.00000318, 0.00001314, -0.00001668, 0.00001524, -0.00000816, -0.00000500, 0.00000213, -0.00000459, -0.00000435, 0.00000722, 0.00000089, -0.00000465, -0.00000212, -0.00001133, 0.00002226, -0.00003366, 0.00003882, -0.00002200, 0.00000113, },
		{ -0.00000958, 0.00000484, 0.00000267, -0.00000544, 0.00001385, -0.00001046, 0.00000697, -0.00001699, 0.00002626, -0.00002279, 0.00002223, -0.00001341, 0.00000219, 0.00000298, -0.00000388, 0.00000564, -0.00000124, -0.00000856, 0.00000614, -0.00001165, 0.00000950, -0.00000966, 0.00001121, -0.00001055, 0.00000694, -0.00000767, 0.00001143, -0.00001918, 0.00001791, -0.00001390, 0.00002143, -0.00003791, 0.00003031, -0.00000996, 0.00001600, -0.00003111, 0.00002034, -0.00002547, 0.00002822, -0.00001659, 0.00001886, -0.00002524, 0.00002704, -0.00003681, 0.00002435, -0.00003560, 0.00004222, -0.00005242, 0.00004103, -0.00005466, 0.00005956, -0.00006787, 0.00005300, -0.00008110, 0.00008911, -0.00015307, 0.00011274, -0.00025560, 0.00020311, -0.00118403, 0.00217719, -0.00023766, -0.00051928, 0.00014421, -0.00018523, 0.00008266, -0.00012198, 0.00008128, -0.00007264, 0.00005669, -0.00007291, 0.00005745, -0.00005328, 0.00003902, -0.00004839, 0.00003726, -0.00003570, 0.00002626, -0.00003806, 0.00003086, -0.00003115, 0.00003846, -0.00003610, 0.00003193, -0.00002782, 0.00001208, -0.00002649, 0.00004020, -0.00002687, 0.00000540, -0.00001018, 0.00001860, -0.00001261, 0.00000582, -0.00000932, 0.00001365, -0.00001370, 0.00001031, -0.00001192, 0.00001274, -0.00001032, 0.00000552, -0.00000909, 0.00000060, -0.00000390, 0.00000985, -0.00001168, 0.00002509, -0.00002879, 0.00002639, -0.00001801, 0.00000730, 0.00000350, -0.00001223, 0.00000421, -0.00000485, -0.00000179, -0.00000478, 0.00001548, -0.00001567, 0.00001147, -0.00002596, 0.00002481, -0.00001900, 0.00000427, 0.00002020, -0.00002891, 0.00002147, },
		{ -0.00001539, 0.00001941, -0.00001785, 0.00001977, -0.00001112, 0.00001134, -0.00002074, 0.00002133, -0.00000981, 0.00000521, 0.00000501, -0.00001091, 0.00000749, -0.00000180, -0.00000303, 0.00001173, -0.00001923, 0.00001445, -0.00001847, 0.00001480, -0.00001604, 0.00001731, -0.00001653, 0.00001600, -0.00002065, 0.00002570, -0.00003071, 0.00002679, -0.00002580, 0.00003468, -0.00004125, 0.00002444, -0.00001236, 0.00002887, -0.00003920, 0.00002706, -0.00003556, 0.00003493, -0.00002701, 0.00003799, -0.00004664, 0.00004894, -0.00005589, 0.00004795, -0.00006630, 0.00007048, -0.00007518, 0.00006814, -0.00008815, 0.00009168, -0.00009974, 0.00010043, -0.00014273, 0.00015633, -0.00020815, 0.00018914, -0.00033509, 0.00033717, -0.00093826, 0.00061790, 0.00154832, -0.00118529, 0.00033307, -0.00034306, 0.00019819, -0.00022148, 0.00014941, -0.00013937, 0.00012237, -0.00012641, 0.00009564, -0.00008872, 0.00007602, -0.00008195, 0.00006524, -0.00006508, 0.00005898, -0.00006773, 0.00005479, -0.00005583, 0.00005801, -0.00004846, 0.00004063, -0.00003563, 0.00002753, -0.00004484, 0.00004449, -0.00002221, 0.00000975, -0.00002408, 0.00002940, -0.00002042, 0.00001802, -0.00002448, 0.00002636, -0.00002330, 0.00001990, -0.00002143, 0.00001975, -0.00001607, 0.00001337, -0.00001734, 0.00001129, -0.00001914, 0.00002403, -0.00002390, 0.00003024, -0.00002118, 0.00001008, 0.00000156, -0.00001020, 0.00001367, -0.00001200, -0.00000104, -0.00000040, -0.00000259, -0.00000771, 0.00001116, -0.00000303, -0.00000219, -0.00000572, -0.00000927, 0.00002064, -0.00003172, 0.00003645, -0.00001847, -0.00000353, 0.00001558, },
		{ 0.00001189, -0.00000425, 0.00000134, 0.00000754, -0.00000274, -0.00000197, -0.00000615, 0.00001993, -0.00002092, 0.00002536, -0.00002148, 0.00001041, -0.00000258, -0.00000222, 0.00000699, -0.00000658, -0.00000249, 0.00000052, -0.00000551, 0.00000381, -0.00000392, 0.00000506, -0.00000423, 0.00000024, 0.00000091, -0.00000026, -0.00000680, 0.00000705, -0.00000204, 0.00000578, -0.00002425, 0.00002444, -0.00000430, 0.00000402, -0.00001779, 0.00000898, -0.00001455, 0.00001872, -0.00000530, 0.00000347, -0.00000765, 0.00000720, -0.00001544, 0.00000159, -0.00001075, 0.00001527, -0.00002323, 0.00000994, -0.00002075, 0.00002225, -0.00002530, 0.00000394, -0.00002509, 0.00001899, -0.00006821, 0.00000490, -0.00011475, -0.00004314, -0.00082839, 0.00205876, -0.00043299, -0.00032867, 0.00002298, -0.00008999, 0.00000736, -0.00005383, 0.00002252, -0.00002091, 0.00001182, -0.00003333, 0.00002104, -0.00002065, 0.00000956, -0.00002034, 0.00001007, -0.00001055, 0.00000278, -0.00001474, 0.00000770, -0.00000719, 0.00001670, -0.00001893, 0.00001678, -0.00001557, -0.00000087, -0.00000941, 0.00002706, -0.00002272, 0.00000149, -0.00000125, 0.00000979, -0.00000615, -0.00000118, -0.00000002, 0.00000382, -0.00000504, 0.00000268, -0.00000470, 0.00000616, -0.00000493, -0.00000008, -0.00000357, -0.00000568, 0.00000567, -0.00000123, 0.00000060, 0.00001423, -0.00002190, 0.00002535, -0.00002338, 0.00001596, -0.00000671, -0.00000554, -0.00000002, -0.00000080, -0.00000694, 0.00000006, 0.00001126, -0.00001253, 0.00000706, -0.00002280, 0.00002086, -0.00001429, -0.00000090, 0.00002617, -0.00003502, 0.00002742, -0.00001428, },
		{ 0.00002043, -0.00001943, 0.00002197, -0.00001269, 0.00001355, -0.00002412, 0.00003097, -0.00002222, 0.00001899, -0.00000814, -0.00000323, 0.00000546, -0.00000398, 0.00000107, 0.00000600, -0.00001536, 0.00001206, -0.00001603, 0.00001306, -0.00001407, 0.00001483, -0.00001374, 0.00001246, -0.00001537, 0.00001910, -0.00002580, 0.00002404, -0.00002204, 0.00002968, -0.00004108, 0.00002940, -0.00001244, 0.00002258, -0.00003443, 0.00002482, -0.00003269, 0.00003162, -0.00002016, 0.00002761, -0.00003580, 0.00003751, -0.00004454, 0.00003639, -0.00005376, 0.00005725, -0.00006135, 0.00005388, -0.00007233, 0.00007402, -0.00007925, 0.00007710, -0.00011647, 0.00012330, -0.00017234, 0.00014796, -0.00028831, 0.00026636, -0.00093519, 0.00092614, 0.00111602, -0.00096381, 0.00023559, -0.00027248, 0.00014587, -0.00017503, 0.00011308, -0.00010928, 0.00009753, -0.00010465, 0.00007566, -0.00007138, 0.00006010, -0.00006617, 0.00005019, -0.00005188, 0.00004657, -0.00005514, 0.00004298, -0.00004489, 0.00005088, -0.00004508, 0.00003733, -0.00003297, 0.00002235, -0.00003762, 0.00004358, -0.00002640, 0.00000839, -0.00001716, 0.00002434, -0.00001711, 0.00001287, -0.00001763, 0.00002062, -0.00001949, 0.00001690, -0.00001851, 0.00001727, -0.00001408, 0.00001045, -0.00001412, 0.00000709, -0.00001313, 0.00001948, -0.00002151, 0.00003235, -0.00002842, 0.00002160, -0.00001195, 0.00000101, 0.00000597, -0.00001020, -0.00000099, 0.00000020, -0.00000389, -0.00000740, 0.00001159, -0.00000407, -0.00000230, -0.00000687, -0.00000907, 0.00002031, -0.00003114, 0.00003598, -0.00001748, -0.00000475, 0.00001729, -0.00001503, },
		{ -0.00001144, 0.00000916, -0.00000014, 0.00000636, -0.00001264, 0.00000859, 0.00000727, -0.00001141, 0.00002033, -0.00002326, 0.00001552, -0.00000793, 0.00000127, 0.00000534, -0.00000890, 0.00000141, -0.00000334, -0.00000089, -0.00000052, 0.00000021, 0.00000078, -0.00000014, -0.00000351, 0.00000572, -0.00000832, 0.00000296, -0.00000155, 0.00000712, -0.00000815, -0.00000978, 0.00001767, -0.00000059, -0.00000465, -0.00000695, -0.00000025, -0.00000605, 0.00001268, 0.00000052, -0.00000515, 0.00000275, -0.00000527, -0.00000167, -0.00001300, 0.00000546, -0.00000278, -0.00000331, -0.00001106, 0.00000202, -0.00000255, 0.00000275, -0.00002838, 0.00001144, -0.00002903, -0.00000795, -0.00007449, -0.00000877, -0.00026277, -0.00042880, 0.00180711, -0.00049398, -0.00022005, -0.00004682, -0.00003358, -0.00003626, -0.00001205, -0.00001443, 0.00001229, -0.00001731, -0.00000744, -0.00000315, 0.00000085, -0.00000938, -0.00000228, -0.00000778, 0.00000587, -0.00001238, 0.00000072, -0.00000819, 0.00001084, -0.00000128, -0.00000432, 0.00000395, -0.00000622, -0.00000994, 0.00000511, 0.00001357, -0.00001836, 0.00000014, 0.00000415, 0.00000374, -0.00000259, -0.00000437, 0.00000531, -0.00000284, 0.00000118, -0.00000279, 0.00000062, 0.00000120, -0.00000132, -0.00000357, -0.00000052, -0.00000872, 0.00001198, -0.00001017, 0.00001279, 0.00000058, -0.00000948, 0.00001689, -0.00002144, 0.00001899, -0.00001392, 0.00000044, -0.00000393, 0.00000260, -0.00001145, 0.00000493, 0.00000628, -0.00000850, 0.00000172, -0.00001833, 0.00001569, -0.00000909, -0.00000580, 0.00003139, -0.00004006, 0.00003269, -0.00001849, 0.00001858, },
		{ -0.00001584, 0.00001902, -0.00000906, 0.00000996, -0.00002007, 0.00003279, -0.00002817, 0.00002840, -0.00002058, 0.00000678, -0.00000017, -0.00000276, 0.00000279, 0.00000100, -0.00001067, 0.00000834, -0.00001201, 0.00000969, -0.00001066, 0.00001097, -0.00000965, 0.00000774, -0.00000869, 0.00000995, -0.00001705, 0.00001725, -0.00001436, 0.00001952, -0.00003451, 0.00002988, -0.00001094, 0.00001447, -0.00002626, 0.00001910, -0.00002695, 0.00002638, -0.00001238, 0.00001584, -0.00002244, 0.00002262, -0.00002906, 0.00002025, -0.00003596, 0.00003792, -0.00004079, 0.00003249, -0.00004883, 0.00004803, -0.00004967, 0.00004317, -0.00007744, 0.00007384, -0.00011488, 0.00007833, -0.00020185, 0.00012575, -0.00080216, 0.00108637, 0.00075650, -0.00074267, 0.00012295, -0.00018850, 0.00008161, -0.00011708, 0.00006537, -0.00006847, 0.00006253, -0.00007361, 0.00004733, -0.00004657, 0.00003756, -0.00004425, 0.00002919, -0.00003286, 0.00002852, -0.00003685, 0.00002514, -0.00002710, 0.00003613, -0.00003482, 0.00002851, -0.00002618, 0.00001375, -0.00002555, 0.00003683, -0.00002733, 0.00000685, -0.00000985, 0.00001782, -0.00001276, 0.00000729, -0.00000982, 0.00001290, -0.00001337, 0.00001175, -0.00001356, 0.00001288, -0.00001064, 0.00000639, -0.00000985, 0.00000186, -0.00000501, 0.00001102, -0.00001339, 0.00002735, -0.00002854, 0.00002743, -0.00002283, 0.00001288, -0.00000499, -0.00000421, -0.00000420, 0.00000338, -0.00000794, -0.00000371, 0.00000841, -0.00000189, -0.00000540, -0.00000504, -0.00001130, 0.00002164, -0.00003202, 0.00003723, -0.00001868, -0.00000286, 0.00001550, -0.00001076, 0.00001676, },
		{ 0.00001480, -0.00000589, 0.00001336, -0.00002100, 0.00002271, -0.00000747, 0.00000246, 0.00000861, -0.00001817, 0.00001562, -0.00001083, 0.00000460, 0.00000196, -0.00000857, 0.00000288, -0.00000493, 0.00000163, -0.00000288, 0.00000223, -0.00000118, 0.00000131, -0.00000386, 0.00000595, -0.00001126, 0.00000819, -0.00000592, 0.00001133, -0.00001749, 0.00000286, 0.00001135, 0.00000060, -0.00000908, 0.00000008, -0.00000620, -0.00000064, 0.00001018, 0.00000098, -0.00000675, 0.00000520, -0.00000916, 0.00000302, -0.00001769, 0.00001094, -0.00000950, 0.00000465, -0.00001908, 0.00001055, -0.00001143, 0.00001234, -0.00003917, 0.00002228, -0.00004648, 0.00001605, -0.00010901, 0.00003581, -0.00040870, -0.00005157, 0.00148067, -0.00044945, -0.00018624, -0.00006551, -0.00001724, -0.00004611, 0.00000054, -0.00002629, 0.00002357, -0.00002741, 0.00000171, -0.00001235, 0.00000874, -0.00001557, 0.00000361, -0.00001408, 0.00001137, -0.00001695, 0.00000584, -0.00001391, 0.00001955, -0.00001203, 0.00000482, -0.00000438, -0.00000142, -0.00001326, 0.00001461, 0.00000177, -0.00001431, 0.00000110, 0.00000602, 0.00000062, -0.00000182, -0.00000383, 0.00000630, -0.00000555, 0.00000407, -0.00000528, 0.00000325, -0.00000149, 0.00000005, -0.00000456, -0.00000035, -0.00000775, 0.00001337, -0.00001423, 0.00002125, -0.00001181, 0.00000478, 0.00000342, -0.00001273, 0.00001515, -0.00001586, 0.00000333, -0.00000550, 0.00000345, -0.00001320, 0.00000750, 0.00000285, -0.00000574, -0.00000248, -0.00001446, 0.00001089, -0.00000484, -0.00000904, 0.00003432, -0.00004219, 0.00003515, -0.00001985, 0.00002219, -0.00001584, },
		{ 0.00001218, -0.00000143, 0.00000173, -0.00000925, 0.00002624, -0.00002564, 0.00002996, -0.00002803, 0.00001537, -0.00000693, 0.00000078, 0.00000195, -0.00000222, -0.00000649, 0.00000456, -0.00000763, 0.00000579, -0.00000693, 0.00000695, -0.00000565, 0.00000358, -0.00000288, 0.00000097, -0.00000720, 0.00000895, -0.00000532, 0.00000677, -0.00002329, 0.00002652, -0.00000857, 0.00000623, -0.00001676, 0.00001176, -0.00002001, 0.00002098, -0.00000604, 0.00000603, -0.00001066, 0.00000887, -0.00001438, 0.00000489, -0.00001872, 0.00001875, -0.00002015, 0.00001108, -0.00002542, 0.00002235, -0.00002085, 0.00001006, -0.00003915, 0.00002410, -0.00005495, 0.00000336, -0.00010514, -0.00004826, -0.00056852, 0.00109682, 0.00050241, -0.00056067, 0.00002590, -0.00011492, 0.00002523, -0.00006488, 0.00002130, -0.00003002, 0.00002879, -0.00004335, 0.00001964, -0.00002237, 0.00001585, -0.00002327, 0.00000897, -0.00001442, 0.00001098, -0.00001884, 0.00000714, -0.00000798, 0.00001845, -0.00002128, 0.00001700, -0.00001780, 0.00000456, -0.00001151, 0.00002601, -0.00002553, 0.00000575, -0.00000359, 0.00001145, -0.00000884, 0.00000301, -0.00000321, 0.00000549, -0.00000696, 0.00000622, -0.00000816, 0.00000798, -0.00000703, 0.00000257, -0.00000609, -0.00000243, 0.00000265, 0.00000144, -0.00000194, 0.00001657, -0.00002126, 0.00002544, -0.00002741, 0.00002112, -0.00001532, 0.00000310, -0.00000865, 0.00000729, -0.00001286, 0.00000158, 0.00000309, 0.00000234, -0.00001052, -0.00000097, -0.00001536, 0.00002398, -0.00003349, 0.00003904, -0.00002062, 0.00000051, 0.00001191, -0.00000440, 0.00001020, -0.00000860, },
		{ -0.00000717, 0.00001546, -0.00002346, 0.00003163, -0.00001921, 0.00001570, -0.00000590, -0.00000816, 0.00001088, -0.00001025, 0.00000623, -0.00000170, -0.00000651, 0.00000236, -0.00000443, 0.00000206, -0.00000333, 0.00000230, -0.00000112, 0.00000061, -0.00000152, 0.00000219, -0.00000914, 0.00000845, -0.00000549, 0.00000975, -0.00002057, 0.00001167, 0.00000650, -0.00000069, -0.00000930, 0.00000303, -0.00000864, 0.00000189, 0.00001021, -0.00000225, -0.00000312, 0.00000133, -0.00000599, 0.00000007, -0.00001390, 0.00000701, -0.00000610, 0.00000188, -0.00001550, 0.00000644, -0.00000631, 0.00000569, -0.00003082, 0.00000976, -0.00003501, 0.00000426, -0.00009709, 0.00001056, -0.00045207, 0.00024680, 0.00114247, -0.00034276, -0.00020409, -0.00004704, -0.00003076, -0.00003011, -0.00001007, -0.00001759, 0.00001642, -0.00002124, -0.00000358, -0.00000859, 0.00000497, -0.00001089, -0.00000092, -0.00001039, 0.00000731, -0.00001214, 0.00000155, -0.00000984, 0.00001846, -0.00001449, 0.00000743, -0.00000754, -0.00000132, -0.00001054, 0.00001787, -0.00000691, -0.00001086, 0.00000363, 0.00000507, -0.00000020, -0.00000293, -0.00000069, 0.00000381, -0.00000475, 0.00000385, -0.00000488, 0.00000325, -0.00000203, -0.00000051, -0.00000352, -0.00000246, -0.00000334, 0.00000977, -0.00001218, 0.00002336, -0.00001914, 0.00001641, -0.00001083, -0.00000026, 0.00000582, -0.00001244, 0.00000245, -0.00000412, 0.00000115, -0.00001137, 0.00000659, 0.00000231, -0.00000563, -0.00000412, -0.00001253, 0.00000757, -0.00000236, -0.00000994, 0.00003436, -0.00004062, 0.00003381, -0.00001714, 0.00002114, -0.00001542, 0.00001588, },
		{ 0.00000728, -0.00000789, 0.00000509, 0.00001373, -0.00001575, 0.00002295, -0.00002809, 0.00001952, -0.00001210, 0.00000481, -0.00000071, -0.00000339, -0.00000361, 0.00000168, -0.00000389, 0.00000238, -0.00000382, 0.00000377, -0.00000280, 0.00000122, 0.00000030, -0.00000538, 0.00000100, 0.00000182, 0.00000222, -0.00000536, -0.00001009, 0.00002069, -0.00000629, -0.00000056, -0.00000791, 0.00000461, -0.00001334, 0.00001660, -0.00000252, 0.00000027, -0.00000329, -0.00000041, -0.00000424, -0.00000558, -0.00000665, 0.00000482, -0.00000487, -0.00000451, -0.00000842, 0.00000399, -0.00000079, -0.00001281, -0.00001297, -0.00001177, -0.00001021, -0.00005441, -0.00002947, -0.00021074, -0.00028368, 0.00098489, 0.00035839, -0.00043676, -0.00003759, -0.00006657, -0.00001033, -0.00003023, -0.00000837, -0.00000365, 0.00000506, -0.00002169, -0.00000042, -0.00000509, 0.00000077, -0.00000872, -0.00000516, -0.00000168, -0.00000105, -0.00000613, -0.00000585, 0.00000718, 0.00000272, -0.00000844, 0.00000591, -0.00001044, -0.00000240, 0.00000132, 0.00001351, -0.00002188, 0.00000550, 0.00000070, 0.00000633, -0.00000625, 0.00000103, 0.00000073, 0.00000012, -0.00000195, 0.00000188, -0.00000374, 0.00000379, -0.00000430, 0.00000002, -0.00000401, -0.00000416, 0.00000743, -0.00000606, 0.00000928, 0.00000324, -0.00000876, 0.00001613, -0.00002427, 0.00002289, -0.00002163, 0.00000875, -0.00001199, 0.00000978, -0.00001632, 0.00000607, -0.00000215, 0.00000663, -0.00001579, 0.00000366, -0.00001989, 0.00002603, -0.00003419, 0.00003983, -0.00002133, 0.00000315, 0.00000897, 0.00000132, 0.00000371, -0.00000070, 0.00000436, },
		{ 0.00001175, -0.00001843, 0.00003251, -0.00002398, 0.00002342, -0.00001830, 0.00000313, 0.00000324, -0.00000666, 0.00000556, -0.00000449, -0.00000381, 0.00000065, -0.00000248, 0.00000094, -0.00000239, 0.00000104, 0.00000016, -0.00000111, 0.00000197, -0.00000380, -0.00000347, 0.00000484, -0.00000128, 0.00000320, -0.00001731, 0.00001572, 0.00000353, -0.00000385, -0.00000617, 0.00000250, -0.00000813, 0.00000236, 0.00001126, -0.00000674, 0.00000278, -0.00000569, 0.00000100, -0.00000706, -0.00000532, -0.00000244, 0.00000340, -0.00000739, -0.00000495, -0.00000514, 0.00000686, -0.00001029, -0.00001142, -0.00001676, -0.00000512, -0.00003167, -0.00005199, -0.00007240, -0.00038935, 0.00043201, 0.00084529, -0.00021859, -0.00024431, -0.00001194, -0.00005786, -0.00000192, -0.00003242, 0.00000208, -0.00000102, -0.00000578, -0.00001719, 0.00000266, -0.00000541, -0.00000016, -0.00001128, -0.00000102, -0.00000218, -0.00000188, -0.00000847, 0.00000080, 0.00000990, -0.00000999, 0.00000436, -0.00000644, -0.00000461, -0.00000307, 0.00001513, -0.00001186, -0.00000801, 0.00000678, 0.00000243, 0.00000028, -0.00000459, 0.00000335, -0.00000053, -0.00000171, 0.00000154, -0.00000248, 0.00000134, -0.00000112, -0.00000215, -0.00000153, -0.00000540, 0.00000278, 0.00000271, -0.00000473, 0.00001847, -0.00001920, 0.00002175, -0.00002135, 0.00001163, -0.00000572, -0.00000558, -0.00000115, -0.00000074, -0.00000331, -0.00000671, 0.00000260, 0.00000457, -0.00000821, -0.00000298, -0.00001286, 0.00000586, -0.00000158, -0.00000889, 0.00003209, -0.00003599, 0.00002923, -0.00001086, 0.00001573, -0.00001008, 0.00001196, -0.00000327, },
		{ -0.00001480, 0.00001830, -0.00000020, -0.00000240, 0.00000973, -0.00002103, 0.00001798, -0.00001386, 0.00000757, -0.00000397, -0.00000290, -0.00000227, 0.00000021, -0.00000143, 0.00000007, -0.00000190, 0.00000190, -0.00000152, 0.00000110, 0.00000007, -0.00000769, 0.00000568, -0.00000214, 0.00000594, -0.00001394, 0.00000221, 0.00001407, -0.00000493, -0.00000480, -0.00000119, -0.00000098, -0.00000790, 0.00001361, -0.00000191, -0.00000106, -0.00000118, -0.00000395, -0.00000016, -0.00000938, -0.00000186, -0.00000143, 0.00000234, -0.00001140, -0.00000093, -0.00000374, 0.00000676, -0.00002097, -0.00000459, -0.00002609, 0.00000886, -0.00008040, 0.00000194, -0.00032087, -0.00000388, 0.00079723, 0.00030544, -0.00036991, -0.00006438, -0.00004675, -0.00002170, -0.00001685, -0.00001970, 0.00000674, -0.00000491, -0.00001209, -0.00000968, 0.00000248, -0.00000517, -0.00000301, -0.00001086, 0.00000305, -0.00000520, -0.00000127, -0.00001092, 0.00001493, -0.00000744, 0.00000055, -0.00000235, -0.00000592, -0.00000516, 0.00001030, 0.00000169, -0.00001732, 0.00000619, 0.00000273, 0.00000289, -0.00000511, 0.00000138, 0.00000160, -0.00000240, 0.00000073, -0.00000040, -0.00000115, 0.00000104, -0.00000298, -0.00000084, -0.00000400, -0.00000271, 0.00000791, -0.00000896, 0.00001670, -0.00000851, 0.00000498, 0.00000251, -0.00001471, 0.00001772, -0.00002227, 0.00001085, -0.00001262, 0.00000937, -0.00001665, 0.00000781, -0.00000531, 0.00000905, -0.00001931, 0.00000712, -0.00002341, 0.00002655, -0.00003296, 0.00003832, -0.00001925, 0.00000318, 0.00000882, 0.00000387, 0.00000020, 0.00000470, -0.00000115, 0.00001357, },
		{ -0.00000701, 0.00002535, -0.00002044, 0.00002271, -0.00002456, 0.00001180, -0.00000444, -0.00000163, 0.00000295, -0.00000586, -0.00000137, -0.00000136, 0.00000005, -0.00000095, -0.00000080, -0.00000071, 0.00000169, -0.00000263, 0.00000485, -0.00000971, 0.00000341, -0.00000050, 0.00000455, -0.00000609, -0.00000916, 0.00001531, 0.00000222, -0.00000782, -0.00000111, -0.00000032, -0.00000564, 0.00000171, 0.00001189, -0.00001023, 0.00000797, -0.00001245, 0.00000809, -0.00001437, 0.00000363, -0.00001265, 0.00001396, -0.00001778, 0.00000673, -0.00001771, 0.00002067, -0.00002691, 0.00000868, -0.00004475, 0.00002822, -0.00007362, 0.00000305, -0.00018438, -0.00024206, 0.00049757, 0.00062139, -0.00011171, -0.00028105, 0.00002044, -0.00008249, 0.00002446, -0.00005426, 0.00002199, -0.00001939, 0.00001075, -0.00003182, 0.00001484, -0.00001637, 0.00001087, -0.00002199, 0.00000878, -0.00001203, 0.00000882, -0.00001922, 0.00001320, -0.00000203, -0.00000161, -0.00000214, -0.00000304, -0.00000901, 0.00000668, 0.00000786, -0.00001327, -0.00000558, 0.00000958, -0.00000077, 0.00000109, -0.00000564, 0.00000667, -0.00000492, 0.00000192, -0.00000144, 0.00000065, -0.00000140, 0.00000027, -0.00000383, 0.00000018, -0.00000752, 0.00000830, -0.00000508, 0.00000557, 0.00000821, -0.00001214, 0.00001923, -0.00002495, 0.00001892, -0.00001558, 0.00000174, -0.00000531, 0.00000274, -0.00000795, -0.00000121, -0.00000281, 0.00000822, -0.00001225, -0.00000008, -0.00001466, 0.00000501, -0.00000154, -0.00000713, 0.00002910, -0.00003006, 0.00002330, -0.00000305, 0.00000800, -0.00000170, 0.00000468, 0.00000427, 0.00000378, },
		{ 0.00002606, -0.00001059, 0.00000924, -0.00000519, -0.00000958, 0.00001166, -0.00001184, 0.00000802, -0.00000674, -0.00000144, -0.00000224, 0.00000013, -0.00000039, -0.00000097, -0.00000124, 0.00000130, -0.00000153, 0.00000269, -0.00000310, -0.00000593, 0.00000627, -0.00000222, 0.00000488, -0.00001712, 0.00001131, 0.00000816, -0.00000496, -0.00000606, 0.00000266, -0.00000431, -0.00000405, 0.00001170, -0.00000325, 0.00000089, -0.00000331, -0.00000261, -0.00000133, -0.00000729, -0.00000367, -0.00000052, 0.00000197, -0.00001015, -0.00000223, -0.00000179, 0.00000301, -0.00001579, -0.00001269, -0.00001958, 0.00000187, -0.00007247, -0.00001892, -0.00035303, 0.00022204, 0.00058606, 0.00031054, -0.00034459, -0.00006332, -0.00004889, -0.00001409, -0.00002082, -0.00001554, 0.00000322, -0.00000264, -0.00001336, -0.00000918, 0.00000134, -0.00000304, -0.00000511, -0.00000902, 0.00000058, -0.00000220, -0.00000374, -0.00000810, 0.00001453, -0.00001064, 0.00000420, -0.00000668, -0.00000478, -0.00000314, 0.00001395, -0.00000760, -0.00001268, 0.00000760, 0.00000274, 0.00000098, -0.00000496, 0.00000332, -0.00000003, -0.00000228, 0.00000109, -0.00000052, -0.00000054, -0.00000020, -0.00000293, -0.00000034, -0.00000552, 0.00000136, 0.00000417, -0.00000627, 0.00001796, -0.00001508, 0.00001556, -0.00001100, -0.00000224, 0.00000760, -0.00001774, 0.00000922, -0.00001041, 0.00000597, -0.00001354, 0.00000611, -0.00000547, 0.00000862, -0.00002002, 0.00000834, -0.00002510, 0.00002495, -0.00002942, 0.00003423, -0.00001397, -0.00000004, 0.00001231, 0.00000209, 0.00000131, 0.00000558, -0.00000205, 0.00001486, -0.00001570, },
		{ 0.00001299, -0.00001040, 0.00001372, -0.00002290, 0.00001516, -0.00000954, 0.00000286, -0.00000057, -0.00000586, 0.00000031, -0.00000297, 0.00000241, -0.00000285, 0.00000072, -0.00000223, 0.00000272, -0.00000302, 0.00000573, -0.00001352, 0.00000910, -0.00000518, 0.00000944, -0.00001522, 0.00000138, 0.00001169, 0.00000189, -0.00001141, 0.00000430, -0.00000399, -0.00000228, 0.00000069, 0.00001129, -0.00001131, 0.00001040, -0.00001644, 0.00001246, -0.00001870, 0.00000939, -0.00001966, 0.00002134, -0.00002476, 0.00001463, -0.00002590, 0.00002898, -0.00003697, 0.00002071, -0.00006319, 0.00005110, -0.00010377, 0.00004304, -0.00028901, -0.00004959, 0.00046259, 0.00047913, -0.00004066, -0.00029853, 0.00003767, -0.00009403, 0.00003934, -0.00006667, 0.00003408, -0.00003136, 0.00002182, -0.00004162, 0.00002273, -0.00002312, 0.00001767, -0.00002866, 0.00001477, -0.00001807, 0.00001570, -0.00002626, 0.00002269, -0.00001291, 0.00000703, -0.00000929, 0.00000046, -0.00001209, 0.00001583, -0.00000165, -0.00001196, -0.00000331, 0.00001133, -0.00000364, 0.00000159, -0.00000542, 0.00000819, -0.00000796, 0.00000477, -0.00000381, 0.00000332, -0.00000393, 0.00000131, -0.00000474, 0.00000071, -0.00000751, 0.00001113, -0.00001065, 0.00001525, -0.00000406, -0.00000052, 0.00000993, -0.00002090, 0.00001935, -0.00002078, 0.00000680, -0.00000789, 0.00000435, -0.00001063, 0.00000288, -0.00000746, 0.00001129, -0.00001589, 0.00000293, -0.00001657, 0.00000377, -0.00000094, -0.00000617, 0.00002727, -0.00002501, 0.00001843, 0.00000361, 0.00000094, 0.00000666, -0.00000291, 0.00001237, -0.00000503, 0.00000729, },
		{ -0.00001405, 0.00001465, -0.00001684, 0.00000220, 0.00000308, -0.00000711, 0.00000613, -0.00000832, 0.00000028, -0.00000300, 0.00000109, -0.00000054, -0.00000095, -0.00000153, 0.00000149, -0.00000210, 0.00000484, -0.00000775, -0.00000134, 0.00000364, 0.00000084, -0.00000041, -0.00001462, 0.00001600, 0.00000384, -0.00000633, -0.00000463, 0.00000370, -0.00000532, -0.00000172, 0.00001019, -0.00000502, 0.00000410, -0.00000749, 0.00000134, -0.00000537, -0.00000193, -0.00000935, 0.00000474, -0.00000304, -0.00000397, -0.00000872, 0.00000564, -0.00000705, -0.00000321, -0.00003025, -0.00000018, -0.00002229, -0.00004087, -0.00008274, -0.00030303, 0.00036371, 0.00039591, 0.00033830, -0.00033912, -0.00004927, -0.00006103, 0.00000229, -0.00003347, -0.00000318, -0.00000805, 0.00000663, -0.00002094, -0.00000297, -0.00000474, 0.00000355, -0.00001155, -0.00000295, -0.00000589, 0.00000484, -0.00001057, -0.00000004, 0.00000782, -0.00000778, 0.00000287, -0.00000743, -0.00000618, 0.00000272, 0.00001220, -0.00001339, -0.00000843, 0.00000929, 0.00000135, 0.00000014, -0.00000506, 0.00000569, -0.00000296, -0.00000045, -0.00000015, 0.00000091, -0.00000145, -0.00000034, -0.00000359, 0.00000071, -0.00000739, 0.00000644, -0.00000220, 0.00000107, 0.00001277, -0.00001467, 0.00001973, -0.00002012, 0.00000875, -0.00000390, -0.00001033, 0.00000524, -0.00000657, 0.00000087, -0.00000813, 0.00000176, -0.00000312, 0.00000567, -0.00001805, 0.00000731, -0.00002504, 0.00002159, -0.00002421, 0.00002843, -0.00000637, -0.00000566, 0.00001861, -0.00000338, 0.00000671, 0.00000190, 0.00000194, 0.00001067, -0.00000945, 0.00002604, },
		{ 0.00000185, -0.00000047, -0.00001440, 0.00001260, -0.00001064, 0.00000528, -0.00000386, -0.00000492, 0.00000108, -0.00000379, 0.00000405, -0.00000421, 0.00000169, -0.00000314, 0.00000295, -0.00000204, 0.00000410, -0.00001406, 0.00001193, -0.00000741, 0.00001132, -0.00002144, 0.00001152, 0.00000657, 0.00000169, -0.00001358, 0.00000870, -0.00000725, 0.00000105, -0.00000033, 0.00000947, -0.00000981, 0.00000948, -0.00001671, 0.00001296, -0.00001871, 0.00001034, -0.00002155, 0.00002338, -0.00002606, 0.00001638, -0.00002717, 0.00002902, -0.00003715, 0.00002047, -0.00006630, 0.00005553, -0.00011090, 0.00004973, -0.00035325, 0.00014243, 0.00036322, 0.00040642, -0.00000729, -0.00029296, 0.00003659, -0.00008958, 0.00003959, -0.00006636, 0.00003503, -0.00003375, 0.00002446, -0.00004391, 0.00002395, -0.00002356, 0.00001826, -0.00002934, 0.00001506, -0.00001834, 0.00001667, -0.00002711, 0.00002625, -0.00001943, 0.00001294, -0.00001481, 0.00000240, -0.00001212, 0.00002189, -0.00001103, -0.00000901, -0.00000102, 0.00001167, -0.00000568, 0.00000160, -0.00000391, 0.00000761, -0.00000900, 0.00000605, -0.00000477, 0.00000477, -0.00000563, 0.00000157, -0.00000460, -0.00000017, -0.00000501, 0.00001019, -0.00001183, 0.00002100, -0.00001435, 0.00001164, -0.00000281, -0.00001110, 0.00001310, -0.00002021, 0.00000812, -0.00000770, 0.00000306, -0.00001012, 0.00000399, -0.00000967, 0.00001214, -0.00001749, 0.00000452, -0.00001732, 0.00000111, 0.00000115, -0.00000706, 0.00002791, -0.00002245, 0.00001646, 0.00000689, -0.00000281, 0.00001204, -0.00000765, 0.00001764, -0.00001073, 0.00001989, 0.00000004, },
		{ 0.00001167, -0.00002162, 0.00001046, -0.00000469, -0.00000163, 0.00000272, -0.00000862, 0.00000173, -0.00000402, 0.00000256, -0.00000138, -0.00000032, -0.00000229, 0.00000190, -0.00000240, 0.00000622, -0.00001203, 0.00000408, -0.00000041, 0.00000524, -0.00000803, -0.00000765, 0.00001633, 0.00000129, -0.00000851, -0.00000139, 0.00000258, -0.00000444, -0.00000058, 0.00000845, -0.00000587, 0.00000655, -0.00001132, 0.00000530, -0.00000941, 0.00000353, -0.00001542, 0.00001068, -0.00000882, 0.00000293, -0.00001575, 0.00001317, -0.00001707, 0.00000909, -0.00004783, 0.00002066, -0.00004964, -0.00000387, -0.00016642, -0.00018784, 0.00041412, 0.00025462, 0.00036075, -0.00033401, -0.00003650, -0.00007128, 0.00001687, -0.00004542, 0.00000913, -0.00001982, 0.00001655, -0.00002922, 0.00000393, -0.00001116, 0.00001019, -0.00001808, 0.00000325, -0.00001234, 0.00001189, -0.00001766, 0.00000921, -0.00000168, -0.00000142, -0.00000169, -0.00000601, -0.00000840, 0.00001036, 0.00000632, -0.00001564, -0.00000475, 0.00001070, -0.00000070, -0.00000016, -0.00000474, 0.00000737, -0.00000584, 0.00000186, -0.00000192, 0.00000294, -0.00000308, -0.00000006, -0.00000422, 0.00000140, -0.00000826, 0.00001051, -0.00000865, 0.00001049, 0.00000296, -0.00000784, 0.00001646, -0.00002222, 0.00001461, -0.00001303, -0.00000302, 0.00000109, -0.00000303, -0.00000389, -0.00000248, -0.00000341, 0.00000016, 0.00000161, -0.00001460, 0.00000497, -0.00002416, 0.00001756, -0.00001872, 0.00002260, 0.00000165, -0.00001167, 0.00002558, -0.00001027, 0.00001427, -0.00000438, 0.00000901, 0.00000285, 0.00000245, 0.00001877, -0.00000996, },
		{ -0.00001501, -0.00000238, 0.00000565, -0.00000794, 0.00000504, -0.00000609, -0.00000358, 0.00000108, -0.00000374, 0.00000475, -0.00000479, 0.00000194, -0.00000339, 0.00000256, -0.00000014, 0.00000043, -0.00001139, 0.00001145, -0.00000653, 0.00000928, -0.00002304, 0.00001895, 0.00000161, 0.00000094, -0.00001376, 0.00001124, -0.00000924, 0.00000375, -0.00000135, 0.00000708, -0.00000660, 0.00000599, -0.00001387, 0.00001014, -0.00001497, 0.00000696, -0.00001868, 0.00002042, -0.00002207, 0.00001247, -0.00002228, 0.00002180, -0.00002869, 0.00000925, -0.00005503, 0.00004169, -0.00009401, 0.00001757, -0.00035690, 0.00029443, 0.00024079, 0.00037902, -0.00000134, -0.00027025, 0.00002194, -0.00007312, 0.00002832, -0.00005558, 0.00002645, -0.00002768, 0.00001953, -0.00003942, 0.00001925, -0.00001853, 0.00001346, -0.00002480, 0.00001035, -0.00001352, 0.00001222, -0.00002189, 0.00002323, -0.00002025, 0.00001461, -0.00001751, 0.00000219, -0.00000857, 0.00002346, -0.00001838, -0.00000543, 0.00000131, 0.00001063, -0.00000676, 0.00000130, -0.00000165, 0.00000539, -0.00000816, 0.00000571, -0.00000423, 0.00000480, -0.00000634, 0.00000110, -0.00000367, -0.00000193, -0.00000068, 0.00000580, -0.00000799, 0.00002088, -0.00001946, 0.00002025, -0.00001460, 0.00000068, 0.00000262, -0.00001486, 0.00000593, -0.00000495, -0.00000084, -0.00000654, 0.00000188, -0.00000897, 0.00001019, -0.00001635, 0.00000393, -0.00001636, -0.00000327, 0.00000482, -0.00000982, 0.00003118, -0.00002271, 0.00001790, 0.00000604, -0.00000208, 0.00001284, -0.00000765, 0.00001791, -0.00001041, 0.00002690, -0.00000897, 0.00001125, },
		{ -0.00001851, 0.00001295, -0.00000922, 0.00000261, -0.00000100, -0.00000795, 0.00000268, -0.00000484, 0.00000401, -0.00000240, 0.00000045, -0.00000306, 0.00000207, -0.00000188, 0.00000587, -0.00001431, 0.00000833, -0.00000385, 0.00000879, -0.00001540, 0.00000152, 0.00001337, 0.00000009, -0.00001067, 0.00000251, 0.00000031, -0.00000237, -0.00000029, 0.00000619, -0.00000506, 0.00000694, -0.00001308, 0.00000731, -0.00001127, 0.00000656, -0.00001906, 0.00001427, -0.00001211, 0.00000709, -0.00001954, 0.00001653, -0.00002196, 0.00001485, -0.00005740, 0.00003262, -0.00006675, 0.00001921, -0.00024094, -0.00003917, 0.00038767, 0.00017039, 0.00036278, -0.00031748, -0.00003394, -0.00007204, 0.00002261, -0.00005006, 0.00001528, -0.00002651, 0.00002215, -0.00003375, 0.00000757, -0.00001431, 0.00001345, -0.00002137, 0.00000636, -0.00001556, 0.00001562, -0.00002143, 0.00001575, -0.00001017, 0.00000529, -0.00000717, -0.00000414, -0.00000950, 0.00001730, -0.00000164, -0.00001503, -0.00000156, 0.00001134, -0.00000273, -0.00000031, -0.00000367, 0.00000766, -0.00000765, 0.00000360, -0.00000324, 0.00000464, -0.00000468, 0.00000007, -0.00000428, 0.00000112, -0.00000721, 0.00001187, -0.00001255, 0.00001868, -0.00000810, 0.00000269, 0.00000729, -0.00001720, 0.00001367, -0.00001724, 0.00000173, -0.00000133, -0.00000148, -0.00000647, 0.00000134, -0.00000737, 0.00000251, -0.00000176, -0.00001129, 0.00000267, -0.00002364, 0.00001409, -0.00001437, 0.00001850, 0.00000809, -0.00001587, 0.00003070, -0.00001575, 0.00002097, -0.00001018, 0.00001602, -0.00000511, 0.00001634, 0.00000739, -0.00000077, 0.00000130, },
		{ 0.00000889, -0.00000279, -0.00000298, 0.00000257, -0.00000697, -0.00000228, 0.00000059, -0.00000298, 0.00000460, -0.00000461, 0.00000157, -0.00000321, 0.00000204, 0.00000177, -0.00000403, -0.00000663, 0.00000842, -0.00000318, 0.00000380, -0.00001970, 0.00002243, -0.00000210, -0.00000062, -0.00001193, 0.00001166, -0.00000959, 0.00000551, -0.00000254, 0.00000502, -0.00000310, 0.00000152, -0.00000957, 0.00000576, -0.00000943, 0.00000131, -0.00001314, 0.00001461, -0.00001513, 0.00000555, -0.00001438, 0.00001114, -0.00001623, -0.00000730, -0.00003588, 0.00001701, -0.00006194, -0.00004480, -0.00029764, 0.00038223, 0.00013056, 0.00037034, -0.00000739, -0.00024105, 0.00000259, -0.00005244, 0.00001228, -0.00004007, 0.00001320, -0.00001727, 0.00001060, -0.00003130, 0.00001154, -0.00001086, 0.00000602, -0.00001769, 0.00000326, -0.00000621, 0.00000492, -0.00001295, 0.00001530, -0.00001611, 0.00001218, -0.00001748, 0.00000042, -0.00000225, 0.00002047, -0.00002264, -0.00000194, 0.00000351, 0.00000854, -0.00000704, 0.00000107, 0.00000057, 0.00000243, -0.00000615, 0.00000429, -0.00000261, 0.00000372, -0.00000632, 0.00000031, -0.00000258, -0.00000355, 0.00000397, -0.00000038, -0.00000033, 0.00001495, -0.00001803, 0.00002257, -0.00002160, 0.00001024, -0.00000845, -0.00000722, 0.00000179, -0.00000101, -0.00000587, -0.00000128, -0.00000236, -0.00000614, 0.00000609, -0.00001290, 0.00000142, -0.00001402, -0.00000878, 0.00000926, -0.00001343, 0.00003601, -0.00002474, 0.00002165, 0.00000207, 0.00000241, 0.00000950, -0.00000303, 0.00001313, -0.00000343, 0.00002661, -0.00001138, 0.00001393, -0.00002513, },
		{ 0.00000970, -0.00000955, 0.00000435, -0.00000397, -0.00000679, 0.00000308, -0.00000521, 0.00000501, -0.00000315, 0.00000099, -0.00000358, 0.00000186, -0.00000049, 0.00000355, -0.00001379, 0.00001007, -0.00000520, 0.00000973, -0.00002014, 0.00001033, 0.00000877, -0.00000053, -0.00001198, 0.00000593, -0.00000204, 0.00000014, -0.00000062, 0.00000360, -0.00000269, 0.00000503, -0.00001223, 0.00000665, -0.00001013, 0.00000614, -0.00001900, 0.00001406, -0.00001142, 0.00000693, -0.00001851, 0.00001401, -0.00001973, 0.00001151, -0.00005529, 0.00003038, -0.00006592, 0.00001512, -0.00028062, 0.00010650, 0.00031272, 0.00013475, 0.00034249, -0.00028705, -0.00004332, -0.00006179, 0.00001773, -0.00004533, 0.00001309, -0.00002594, 0.00002136, -0.00003265, 0.00000631, -0.00001279, 0.00001202, -0.00002014, 0.00000512, -0.00001425, 0.00001457, -0.00002007, 0.00001716, -0.00001485, 0.00000975, -0.00001159, -0.00000316, -0.00000817, 0.00002143, -0.00000950, -0.00001261, 0.00000129, 0.00001090, -0.00000427, -0.00000048, -0.00000192, 0.00000645, -0.00000794, 0.00000419, -0.00000352, 0.00000543, -0.00000574, -0.00000024, -0.00000366, -0.00000015, -0.00000415, 0.00000986, -0.00001218, 0.00002275, -0.00001674, 0.00001301, -0.00000429, -0.00000737, 0.00000666, -0.00001591, 0.00000283, -0.00000115, -0.00000260, -0.00000600, 0.00000216, -0.00000880, 0.00000258, -0.00000307, -0.00000944, 0.00000148, -0.00002433, 0.00001195, -0.00001202, 0.00001719, 0.00001165, -0.00001676, 0.00003216, -0.00001758, 0.00002421, -0.00001265, 0.00001986, -0.00000953, 0.00002795, -0.00000371, 0.00000912, -0.00001278, -0.00000930, },
		{ -0.00000964, 0.00000209, -0.00000100, -0.00000673, -0.00000123, -0.00000010, -0.00000184, 0.00000386, -0.00000392, 0.00000084, -0.00000293, 0.00000190, 0.00000271, -0.00000775, -0.00000149, 0.00000438, 0.00000108, -0.00000342, -0.00001253, 0.00002189, -0.00000416, -0.00000279, -0.00000860, 0.00001029, -0.00000842, 0.00000628, -0.00000404, 0.00000406, -0.00000067, -0.00000224, -0.00000564, 0.00000183, -0.00000436, -0.00000410, -0.00000759, 0.00000879, -0.00000828, -0.00000103, -0.00000735, 0.00000169, -0.00000547, -0.00002205, -0.00001772, -0.00000751, -0.00002949, -0.00011801, -0.00019036, 0.00040090, 0.00005418, 0.00035956, -0.00001171, -0.00021542, -0.00001295, -0.00003532, -0.00000147, -0.00002614, 0.00000088, -0.00000740, 0.00000198, -0.00002339, 0.00000437, -0.00000394, -0.00000076, -0.00001120, -0.00000304, 0.00000036, -0.00000185, -0.00000373, 0.00000553, -0.00000924, 0.00000705, -0.00001573, -0.00000158, 0.00000506, 0.00001407, -0.00002372, 0.00000113, 0.00000524, 0.00000591, -0.00000676, 0.00000121, 0.00000202, -0.00000028, -0.00000389, 0.00000258, -0.00000062, 0.00000210, -0.00000601, -0.00000031, -0.00000200, -0.00000398, 0.00000721, -0.00000597, 0.00000859, 0.00000524, -0.00001099, 0.00001813, -0.00002182, 0.00001440, -0.00001662, -0.00000016, -0.00000219, 0.00000231, -0.00001009, 0.00000361, -0.00000686, -0.00000285, 0.00000135, -0.00000846, -0.00000196, -0.00001130, -0.00001429, 0.00001310, -0.00001627, 0.00004058, -0.00002659, 0.00002557, -0.00000271, 0.00000835, 0.00000419, 0.00000411, 0.00000543, 0.00000834, 0.00002003, -0.00000739, 0.00000867, -0.00002789, 0.00001587, },
		{ -0.00000636, 0.00000339, -0.00000562, -0.00000555, 0.00000301, -0.00000501, 0.00000536, -0.00000340, 0.00000112, -0.00000377, 0.00000142, 0.00000131, -0.00000019, -0.00001065, 0.00000900, -0.00000395, 0.00000734, -0.00002075, 0.00001665, 0.00000412, -0.00000127, -0.00001185, 0.00000804, -0.00000367, 0.00000245, -0.00000146, 0.00000120, 0.00000046, 0.00000154, -0.00000933, 0.00000388, -0.00000657, 0.00000282, -0.00001569, 0.00001047, -0.00000722, 0.00000307, -0.00001353, 0.00000681, -0.00001184, 0.00000076, -0.00004310, 0.00001517, -0.00004778, -0.00001870, -0.00027093, 0.00021829, 0.00022152, 0.00012961, 0.00030738, -0.00024753, -0.00006046, -0.00004422, 0.00000526, -0.00003353, 0.00000427, -0.00001936, 0.00001516, -0.00002678, 0.00000098, -0.00000750, 0.00000680, -0.00001523, 0.00000037, -0.00000924, 0.00000947, -0.00001397, 0.00001316, -0.00001468, 0.00001064, -0.00001389, -0.00000347, -0.00000411, 0.00002164, -0.00001552, -0.00000939, 0.00000387, 0.00000936, -0.00000508, -0.00000064, 0.00000009, 0.00000415, -0.00000686, 0.00000362, -0.00000270, 0.00000514, -0.00000612, -0.00000093, -0.00000262, -0.00000184, 0.00000013, 0.00000507, -0.00000734, 0.00002123, -0.00002023, 0.00001949, -0.00001417, 0.00000344, -0.00000370, -0.00001037, 0.00000079, 0.00000117, -0.00000581, -0.00000295, 0.00000011, -0.00000759, 0.00000015, -0.00000191, -0.00000955, 0.00000172, -0.00002632, 0.00001107, -0.00001152, 0.00001858, 0.00001233, -0.00001426, 0.00002967, -0.00001508, 0.00002292, -0.00001038, 0.00001882, -0.00000802, 0.00003395, -0.00001065, 0.00001508, -0.00002566, 0.00000226, 0.00000274, },
		{ 0.00000541, -0.00000441, -0.00000585, -0.00000049, -0.00000073, -0.00000064, 0.00000289, -0.00000302, 0.00000001, -0.00000277, 0.00000241, 0.00000207, -0.00000948, 0.00000236, 0.00000106, 0.00000441, -0.00001015, -0.00000359, 0.00001832, -0.00000485, -0.00000499, -0.00000463, 0.00000784, -0.00000615, 0.00000613, -0.00000577, 0.00000445, -0.00000004, -0.00000418, -0.00000338, -0.00000021, -0.00000145, -0.00000732, -0.00000416, 0.00000518, -0.00000391, -0.00000472, -0.00000406, -0.00000311, -0.00000069, -0.00002937, -0.00000796, -0.00002199, -0.00001141, -0.00017869, -0.00006139, 0.00036277, 0.00001759, 0.00033597, -0.00000656, -0.00019908, -0.00001975, -0.00002648, -0.00000831, -0.00001814, -0.00000650, -0.00000170, -0.00000309, -0.00001858, 0.00000031, -0.00000024, -0.00000449, -0.00000765, -0.00000620, 0.00000368, -0.00000531, 0.00000259, -0.00000287, -0.00000233, 0.00000115, -0.00001355, -0.00000238, 0.00001128, 0.00000611, -0.00002221, 0.00000376, 0.00000616, 0.00000331, -0.00000619, 0.00000183, 0.00000230, -0.00000203, -0.00000216, 0.00000131, 0.00000102, 0.00000051, -0.00000579, -0.00000045, -0.00000227, -0.00000264, 0.00000774, -0.00000872, 0.00001572, -0.00000500, -0.00000117, 0.00000883, -0.00001571, 0.00001209, -0.00001978, 0.00000417, -0.00000439, 0.00000360, -0.00001198, 0.00000639, -0.00000984, -0.00000079, -0.00000234, -0.00000457, -0.00000494, -0.00000931, -0.00001868, 0.00001505, -0.00001683, 0.00004318, -0.00002635, 0.00002743, -0.00000570, 0.00001290, -0.00000011, 0.00001070, -0.00000176, 0.00002131, 0.00001040, 0.00000013, -0.00000271, -0.00002311, 0.00001688, -0.00001281, },
		{ 0.00000052, -0.00000598, -0.00000443, 0.00000260, -0.00000429, 0.00000504, -0.00000307, 0.00000083, -0.00000375, 0.00000108, 0.00000278, -0.00000422, -0.00000598, 0.00000589, -0.00000072, 0.00000216, -0.00001705, 0.00001933, 0.00000058, -0.00000253, -0.00001013, 0.00000851, -0.00000411, 0.00000414, -0.00000272, -0.00000045, 0.00000331, -0.00000229, -0.00000565, 0.00000034, -0.00000211, -0.00000178, -0.00001079, 0.00000525, -0.00000138, -0.00000240, -0.00000712, -0.00000197, -0.00000211, -0.00001264, -0.00002648, -0.00000638, -0.00002059, -0.00007344, -0.00021223, 0.00027840, 0.00014101, 0.00013526, 0.00026861, -0.00020694, -0.00007847, -0.00002560, -0.00000924, -0.00001945, -0.00000710, -0.00001023, 0.00000657, -0.00001877, -0.00000596, -0.00000084, 0.00000013, -0.00000892, -0.00000562, -0.00000283, 0.00000264, -0.00000536, 0.00000539, -0.00001043, 0.00000808, -0.00001405, -0.00000462, 0.00000182, 0.00001802, -0.00001880, -0.00000610, 0.00000609, 0.00000694, -0.00000515, -0.00000065, 0.00000182, 0.00000147, -0.00000497, 0.00000233, -0.00000112, 0.00000400, -0.00000601, -0.00000169, -0.00000169, -0.00000307, 0.00000419, -0.00000077, 0.00000055, 0.00001457, -0.00001773, 0.00002001, -0.00001907, 0.00001136, -0.00001384, -0.00000316, -0.00000278, 0.00000426, -0.00000961, 0.00000117, -0.00000354, -0.00000473, -0.00000391, 0.00000105, -0.00001116, 0.00000291, -0.00002894, 0.00001057, -0.00001179, 0.00002148, 0.00001136, -0.00000969, 0.00002455, -0.00000939, 0.00001798, -0.00000394, 0.00001331, -0.00000046, 0.00003330, -0.00001168, 0.00001414, -0.00003307, 0.00001197, -0.00000484, -0.00000150, },
		{ -0.00000673, -0.00000480, -0.00000004, -0.00000112, 0.00000028, 0.00000203, -0.00000217, -0.00000075, -0.00000277, 0.00000347, -0.00000016, -0.00000872, 0.00000390, -0.00000034, 0.00000537, -0.00001434, 0.00000482, 0.00001323, -0.00000483, -0.00000656, -0.00000094, 0.00000515, -0.00000338, 0.00000520, -0.00000742, 0.00000588, -0.00000120, -0.00000405, -0.00000314, 0.00000003, -0.00000124, -0.00000763, -0.00000367, 0.00000466, -0.00000288, -0.00000473, -0.00000528, -0.00000235, -0.00000308, -0.00002733, -0.00000969, -0.00002169, -0.00001673, -0.00020738, 0.00006005, 0.00029084, 0.00001373, 0.00029895, 0.00000875, -0.00019225, -0.00001767, -0.00002639, -0.00000739, -0.00001719, -0.00000771, -0.00000136, -0.00000350, -0.00001779, 0.00000012, -0.00000046, -0.00000446, -0.00000774, -0.00000552, 0.00000291, -0.00000430, 0.00000423, -0.00000762, 0.00000242, -0.00000389, -0.00001188, -0.00000108, 0.00001473, -0.00000148, -0.00001909, 0.00000604, 0.00000600, 0.00000120, -0.00000548, 0.00000280, 0.00000144, -0.00000254, -0.00000133, 0.00000090, 0.00000188, -0.00000069, -0.00000582, -0.00000012, -0.00000324, 0.00000030, 0.00000529, -0.00000737, 0.00001860, -0.00001247, 0.00000784, -0.00000188, -0.00000588, 0.00000448, -0.00001776, 0.00000502, -0.00000422, 0.00000253, -0.00001106, 0.00000626, -0.00001032, -0.00000098, -0.00000390, -0.00000230, -0.00000666, -0.00000870, -0.00002136, 0.00001456, -0.00001442, 0.00004294, -0.00002298, 0.00002590, -0.00000513, 0.00001391, -0.00000092, 0.00001400, -0.00000516, 0.00003161, 0.00000166, 0.00000687, -0.00001610, -0.00001325, 0.00001257, -0.00001184, 0.00000605, },
		{ -0.00000547, -0.00000348, 0.00000197, -0.00000323, 0.00000419, -0.00000228, 0.00000022, -0.00000365, 0.00000113, 0.00000321, -0.00000727, -0.00000127, 0.00000209, 0.00000311, -0.00000431, -0.00001011, 0.00001836, -0.00000141, -0.00000427, -0.00000714, 0.00000747, -0.00000331, 0.00000497, -0.00000427, -0.00000098, 0.00000497, -0.00000524, -0.00000251, -0.00000256, 0.00000160, -0.00000580, -0.00000630, 0.00000048, 0.00000386, -0.00000703, -0.00000211, -0.00000891, 0.00000519, -0.00002324, -0.00001237, -0.00002548, 0.00000346, -0.00013229, -0.00011851, 0.00028478, 0.00008692, 0.00013686, 0.00023564, -0.00017237, -0.00009123, -0.00001167, -0.00002040, -0.00000797, -0.00001665, -0.00000238, -0.00000104, -0.00001163, -0.00001178, 0.00000455, -0.00000541, -0.00000370, -0.00001035, 0.00000235, -0.00000311, 0.00000280, -0.00000342, -0.00000409, 0.00000325, -0.00001281, -0.00000555, 0.00000806, 0.00001171, -0.00001928, -0.00000307, 0.00000769, 0.00000409, -0.00000461, -0.00000042, 0.00000281, -0.00000087, -0.00000298, 0.00000091, 0.00000063, 0.00000245, -0.00000569, -0.00000221, -0.00000132, -0.00000304, 0.00000659, -0.00000549, 0.00000896, 0.00000493, -0.00001049, 0.00001462, -0.00001766, 0.00001382, -0.00002067, 0.00000308, -0.00000596, 0.00000651, -0.00001228, 0.00000450, -0.00000701, -0.00000187, -0.00000809, 0.00000450, -0.00001322, 0.00000406, -0.00003110, 0.00000920, -0.00001138, 0.00002422, 0.00001058, -0.00000515, 0.00001914, -0.00000289, 0.00001173, 0.00000438, 0.00000566, 0.00001106, 0.00002740, -0.00000776, 0.00000613, -0.00003293, 0.00001678, -0.00001017, 0.00000287, -0.00000298, },
		{ -0.00000385, 0.00000013, -0.00000115, 0.00000072, 0.00000152, -0.00000158, -0.00000136, -0.00000278, 0.00000462, -0.00000337, -0.00000579, 0.00000300, 0.00000051, 0.00000344, -0.00001478, 0.00001080, 0.00000815, -0.00000479, -0.00000696, 0.00000174, 0.00000294, -0.00000068, 0.00000364, -0.00000857, 0.00000762, -0.00000344, -0.00000241, -0.00000439, 0.00000196, -0.00000313, -0.00000559, -0.00000559, 0.00000664, -0.00000447, -0.00000205, -0.00000971, 0.00000232, -0.00001064, -0.00001810, -0.00002078, -0.00000868, -0.00004538, -0.00019448, 0.00015036, 0.00021041, 0.00002847, 0.00025471, 0.00002973, -0.00019092, -0.00001035, -0.00003195, -0.00000116, -0.00002136, -0.00000422, -0.00000518, -0.00000028, -0.00002005, 0.00000278, -0.00000357, -0.00000165, -0.00001044, -0.00000201, -0.00000099, 0.00000050, 0.00000125, -0.00000805, 0.00000394, -0.00000718, -0.00001099, 0.00000230, 0.00001469, -0.00000727, -0.00001525, 0.00000803, 0.00000475, -0.00000013, -0.00000473, 0.00000381, -0.00000019, -0.00000198, -0.00000138, 0.00000133, 0.00000186, -0.00000139, -0.00000601, 0.00000039, -0.00000432, 0.00000396, 0.00000068, -0.00000211, 0.00001626, -0.00001497, 0.00001291, -0.00001030, 0.00000398, -0.00000557, -0.00001212, 0.00000308, -0.00000227, -0.00000019, -0.00000803, 0.00000365, -0.00000851, -0.00000335, -0.00000320, -0.00000192, -0.00000701, -0.00000937, -0.00002261, 0.00001200, -0.00000945, 0.00004023, -0.00001681, 0.00002114, -0.00000085, 0.00001087, 0.00000266, 0.00001281, -0.00000293, 0.00003664, -0.00000318, 0.00000893, -0.00002679, -0.00000230, 0.00000537, -0.00000783, 0.00000418, -0.00000768, },
		{ -0.00000263, 0.00000126, -0.00000205, 0.00000306, -0.00000122, -0.00000056, -0.00000356, 0.00000166, 0.00000224, -0.00000841, 0.00000209, -0.00000092, 0.00000597, -0.00001008, -0.00000181, 0.00001465, -0.00000207, -0.00000608, -0.00000353, 0.00000542, -0.00000157, 0.00000490, -0.00000584, -0.00000040, 0.00000503, -0.00000660, -0.00000076, -0.00000387, 0.00000345, -0.00000792, -0.00000365, -0.00000234, 0.00000692, -0.00000918, -0.00000034, -0.00001178, 0.00000729, -0.00002725, -0.00000594, -0.00003505, 0.00001325, -0.00017645, -0.00001212, 0.00024885, 0.00006249, 0.00012756, 0.00021309, -0.00014719, -0.00009580, -0.00000540, -0.00002516, -0.00000208, -0.00002156, 0.00000162, -0.00000534, -0.00000741, -0.00001467, 0.00000696, -0.00000815, -0.00000120, -0.00001214, 0.00000449, -0.00000568, 0.00000796, -0.00001056, 0.00000200, -0.00000224, -0.00001110, -0.00000528, 0.00001288, 0.00000440, -0.00001758, -0.00000034, 0.00000835, 0.00000135, -0.00000369, 0.00000006, 0.00000286, -0.00000237, -0.00000144, -0.00000009, 0.00000204, 0.00000092, -0.00000540, -0.00000232, -0.00000167, -0.00000145, 0.00000639, -0.00000721, 0.00001511, -0.00000461, -0.00000128, 0.00000546, -0.00001089, 0.00001030, -0.00002255, 0.00000655, -0.00000742, 0.00000686, -0.00001267, 0.00000563, -0.00000881, -0.00000043, -0.00001093, 0.00000708, -0.00001463, 0.00000425, -0.00003188, 0.00000598, -0.00000911, 0.00002542, 0.00001161, -0.00000256, 0.00001573, 0.00000181, 0.00000699, 0.00001166, -0.00000087, 0.00002312, 0.00001936, -0.00000206, -0.00000633, -0.00002596, 0.00001572, -0.00001164, 0.00000515, -0.00000590, -0.00000466, },
		{ 0.00000003, -0.00000077, 0.00000060, 0.00000149, -0.00000132, -0.00000183, -0.00000260, 0.00000526, -0.00000652, -0.00000173, 0.00000040, 0.00000299, -0.00000081, -0.00001144, 0.00001339, 0.00000415, -0.00000516, -0.00000600, 0.00000303, 0.00000165, 0.00000150, 0.00000166, -0.00000886, 0.00000885, -0.00000577, -0.00000026, -0.00000608, 0.00000437, -0.00000578, -0.00000256, -0.00000850, 0.00000959, -0.00000698, 0.00000126, -0.00001477, 0.00000771, -0.00001933, -0.00000661, -0.00003527, 0.00000931, -0.00008832, -0.00014275, 0.00019724, 0.00014168, 0.00004714, 0.00021165, 0.00005007, -0.00018959, -0.00000294, -0.00003848, 0.00000628, -0.00002714, 0.00000098, -0.00001055, 0.00000428, -0.00002325, 0.00000623, -0.00000755, 0.00000197, -0.00001379, 0.00000228, -0.00000590, 0.00000698, -0.00000468, -0.00000499, 0.00000240, -0.00000865, -0.00001045, 0.00000687, 0.00001146, -0.00001056, -0.00001130, 0.00000958, 0.00000264, -0.00000062, -0.00000395, 0.00000451, -0.00000202, -0.00000083, -0.00000195, 0.00000232, 0.00000115, -0.00000171, -0.00000616, 0.00000067, -0.00000477, 0.00000701, -0.00000433, 0.00000544, 0.00000947, -0.00001213, 0.00001254, -0.00001377, 0.00001047, -0.00001472, -0.00000531, -0.00000008, 0.00000015, -0.00000310, -0.00000441, -0.00000007, -0.00000554, -0.00000690, -0.00000103, -0.00000284, -0.00000656, -0.00001058, -0.00002333, 0.00000849, -0.00000322, 0.00003642, -0.00000934, 0.00001471, 0.00000570, 0.00000501, 0.00000965, 0.00000793, 0.00000461, 0.00003595, -0.00000332, 0.00000445, -0.00003127, 0.00000586, -0.00000163, -0.00000281, 0.00000084, -0.00000750, -0.00000303, },
		{ 0.00000064, -0.00000103, 0.00000196, -0.00000014, -0.00000138, -0.00000345, 0.00000248, 0.00000003, -0.00000734, 0.00000332, -0.00000218, 0.00000670, -0.00001345, 0.00000578, 0.00000962, -0.00000197, -0.00000740, -0.00000005, 0.00000303, 0.00000068, 0.00000401, -0.00000710, 0.00000086, 0.00000367, -0.00000629, -0.00000054, -0.00000348, 0.00000324, -0.00000782, -0.00000326, -0.00000276, 0.00000741, -0.00000858, -0.00000199, -0.00001041, 0.00000392, -0.00002403, -0.00000856, -0.00003266, 0.00000316, -0.00019146, 0.00008355, 0.00018974, 0.00006115, 0.00010811, 0.00020028, -0.00013058, -0.00009299, -0.00000630, -0.00002355, -0.00000206, -0.00002135, 0.00000124, -0.00000581, -0.00000651, -0.00001435, 0.00000614, -0.00000783, -0.00000165, -0.00001079, 0.00000325, -0.00000441, 0.00000889, -0.00001425, 0.00000604, -0.00000705, -0.00000955, -0.00000330, 0.00001508, -0.00000230, -0.00001452, 0.00000214, 0.00000785, -0.00000080, -0.00000261, 0.00000067, 0.00000206, -0.00000288, -0.00000061, -0.00000040, 0.00000278, -0.00000031, -0.00000522, -0.00000212, -0.00000247, 0.00000134, 0.00000365, -0.00000510, 0.00001703, -0.00001118, 0.00000662, -0.00000419, -0.00000151, 0.00000234, -0.00001970, 0.00000678, -0.00000683, 0.00000524, -0.00001065, 0.00000418, -0.00000837, -0.00000111, -0.00001163, 0.00000796, -0.00001478, 0.00000308, -0.00003100, 0.00000065, -0.00000466, 0.00002458, 0.00001507, -0.00000282, 0.00001559, 0.00000309, 0.00000578, 0.00001560, -0.00000349, 0.00003244, 0.00001248, 0.00000152, -0.00001894, -0.00001521, 0.00001003, -0.00000929, 0.00000477, -0.00000756, -0.00000392, -0.00000189, },
		{ -0.00000006, 0.00000000, 0.00000189, -0.00000139, -0.00000218, -0.00000209, 0.00000489, -0.00000859, 0.00000213, -0.00000265, 0.00000589, -0.00000598, -0.00000536, 0.00001263, 0.00000165, -0.00000598, -0.00000387, 0.00000298, 0.00000139, 0.00000289, -0.00000049, -0.00000815, 0.00000895, -0.00000729, 0.00000144, -0.00000722, 0.00000607, -0.00000783, -0.00000003, -0.00001086, 0.00001188, -0.00000865, 0.00000318, -0.00001798, 0.00001069, -0.00002506, 0.00000179, -0.00004624, 0.00002227, -0.00013096, -0.00006554, 0.00020121, 0.00009549, 0.00005940, 0.00017627, 0.00006477, -0.00018391, 0.00000038, -0.00004191, 0.00001119, -0.00003113, 0.00000487, -0.00001481, 0.00000784, -0.00002529, 0.00000841, -0.00001042, 0.00000447, -0.00001583, 0.00000523, -0.00000951, 0.00001260, -0.00001114, -0.00000025, -0.00000120, -0.00000881, -0.00000951, 0.00001123, 0.00000619, -0.00001146, -0.00000750, 0.00001042, 0.00000016, -0.00000038, -0.00000316, 0.00000463, -0.00000353, 0.00000039, -0.00000256, 0.00000339, 0.00000010, -0.00000186, -0.00000604, 0.00000041, -0.00000402, 0.00000828, -0.00000780, 0.00001283, 0.00000046, -0.00000540, 0.00000723, -0.00001161, 0.00001160, -0.00002030, 0.00000029, -0.00000279, 0.00000173, -0.00000480, -0.00000180, -0.00000332, -0.00000290, -0.00001020, 0.00000135, -0.00000406, -0.00000623, -0.00001138, -0.00002463, 0.00000536, 0.00000272, 0.00003327, -0.00000257, 0.00000885, 0.00001216, -0.00000124, 0.00001771, 0.00000174, 0.00001525, 0.00003118, -0.00000039, -0.00000566, -0.00002846, 0.00000880, -0.00000592, 0.00000122, -0.00000264, -0.00000677, -0.00000225, -0.00000034, },
		{ -0.00000038, 0.00000113, 0.00000076, -0.00000213, -0.00000324, 0.00000315, -0.00000279, -0.00000451, 0.00000244, -0.00000149, 0.00000495, -0.00001346, 0.00001104, 0.00000466, -0.00000178, -0.00000776, 0.00000265, 0.00000090, 0.00000294, 0.00000247, -0.00000769, 0.00000216, 0.00000156, -0.00000483, -0.00000136, -0.00000196, 0.00000159, -0.00000608, -0.00000454, -0.00000143, 0.00000609, -0.00000625, -0.00000569, -0.00000650, -0.00000279, -0.00001602, -0.00001758, -0.00002139, -0.00002472, -0.00017179, 0.00015081, 0.00012733, 0.00007160, 0.00008393, 0.00019302, -0.00011884, -0.00008629, -0.00001134, -0.00001806, -0.00000592, -0.00001760, -0.00000221, -0.00000358, -0.00000789, -0.00001192, 0.00000316, -0.00000548, -0.00000396, -0.00000739, -0.00000029, -0.00000023, 0.00000597, -0.00001413, 0.00000718, -0.00001042, -0.00000828, 0.00000008, 0.00001427, -0.00000728, -0.00001084, 0.00000433, 0.00000621, -0.00000204, -0.00000157, 0.00000121, 0.00000074, -0.00000256, -0.00000043, -0.00000006, 0.00000279, -0.00000118, -0.00000508, -0.00000189, -0.00000317, 0.00000440, -0.00000062, 0.00000031, 0.00001417, -0.00001308, 0.00001069, -0.00001111, 0.00000704, -0.00000719, -0.00001378, 0.00000458, -0.00000483, 0.00000247, -0.00000710, 0.00000081, -0.00000611, -0.00000357, -0.00001033, 0.00000716, -0.00001375, 0.00000082, -0.00002889, -0.00000618, 0.00000129, 0.00002234, 0.00002035, -0.00000541, 0.00001847, 0.00000089, 0.00000852, 0.00001547, -0.00000091, 0.00003714, 0.00000874, 0.00000005, -0.00002745, -0.00000467, 0.00000243, -0.00000459, 0.00000228, -0.00000791, -0.00000335, -0.00000132, 0.00000027, },
		{ -0.00000089, 0.00000258, -0.00000169, -0.00000242, -0.00000130, 0.00000334, -0.00000887, 0.00000467, -0.00000492, 0.00000786, -0.00001039, 0.00000174, 0.00000942, 0.00000047, -0.00000695, -0.00000107, 0.00000192, 0.00000199, 0.00000340, -0.00000246, -0.00000662, 0.00000774, -0.00000751, 0.00000212, -0.00000721, 0.00000638, -0.00000843, 0.00000107, -0.00001167, 0.00001249, -0.00000847, 0.00000260, -0.00001801, 0.00000961, -0.00002545, 0.00000374, -0.00004888, 0.00002212, -0.00015831, 0.00001813, 0.00017308, 0.00007286, 0.00006127, 0.00015106, 0.00007200, -0.00017234, -0.00000205, -0.00004038, 0.00001162, -0.00003143, 0.00000570, -0.00001637, 0.00000901, -0.00002495, 0.00000818, -0.00001103, 0.00000479, -0.00001545, 0.00000561, -0.00001030, 0.00001538, -0.00001589, 0.00000417, -0.00000552, -0.00000828, -0.00000755, 0.00001404, 0.00000037, -0.00001053, -0.00000388, 0.00001020, -0.00000213, 0.00000030, -0.00000237, 0.00000409, -0.00000439, 0.00000128, -0.00000284, 0.00000415, -0.00000097, -0.00000196, -0.00000561, -0.00000039, -0.00000198, 0.00000717, -0.00000824, 0.00001766, -0.00000794, 0.00000254, -0.00000077, -0.00000515, 0.00000741, -0.00002117, 0.00000322, -0.00000401, 0.00000173, -0.00000451, -0.00000121, -0.00000495, -0.00000174, -0.00001206, 0.00000281, -0.00000469, -0.00000673, -0.00001105, -0.00002722, 0.00000351, 0.00000723, 0.00003208, 0.00000195, 0.00000547, 0.00001630, -0.00000538, 0.00002418, -0.00000282, 0.00002598, 0.00002506, 0.00000240, -0.00001816, -0.00001997, 0.00000634, -0.00000644, 0.00000299, -0.00000525, -0.00000598, -0.00000143, -0.00000087, 0.00000080, },
		{ 0.00000073, 0.00000134, -0.00000274, -0.00000281, 0.00000322, -0.00000532, -0.00000088, 0.00000017, 0.00000055, 0.00000132, -0.00001021, 0.00001318, 0.00000074, -0.00000191, -0.00000694, 0.00000421, -0.00000058, 0.00000481, 0.00000054, -0.00000740, 0.00000282, -0.00000049, -0.00000301, -0.00000246, -0.00000023, -0.00000046, -0.00000380, -0.00000636, 0.00000043, 0.00000431, -0.00000382, -0.00000941, -0.00000261, -0.00000951, -0.00000739, -0.00002778, -0.00000829, -0.00006144, -0.00012198, 0.00018181, 0.00007654, 0.00008286, 0.00006143, 0.00018632, -0.00010764, -0.00007979, -0.00001666, -0.00001209, -0.00001068, -0.00001286, -0.00000652, -0.00000059, -0.00000978, -0.00000910, -0.00000028, -0.00000273, -0.00000647, -0.00000374, -0.00000423, 0.00000492, 0.00000081, -0.00001108, 0.00000560, -0.00001225, -0.00000693, 0.00000394, 0.00001102, -0.00001012, -0.00000699, 0.00000600, 0.00000376, -0.00000231, -0.00000069, 0.00000148, -0.00000070, -0.00000178, -0.00000061, 0.00000067, 0.00000223, -0.00000174, -0.00000480, -0.00000194, -0.00000315, 0.00000657, -0.00000475, 0.00000730, 0.00000755, -0.00001033, 0.00000995, -0.00001333, 0.00001189, -0.00001526, -0.00000714, 0.00000139, -0.00000256, -0.00000017, -0.00000343, -0.00000315, -0.00000322, -0.00000676, -0.00000790, 0.00000533, -0.00001220, -0.00000175, -0.00002649, -0.00001334, 0.00000743, 0.00002010, 0.00002595, -0.00000871, 0.00002275, -0.00000330, 0.00001398, 0.00001234, 0.00000622, 0.00003727, 0.00000800, -0.00000711, -0.00002931, 0.00000224, -0.00000416, 0.00000042, -0.00000114, -0.00000741, -0.00000286, -0.00000103, 0.00000025, -0.00000023, },
		{ 0.00000333, -0.00000209, -0.00000251, -0.00000044, 0.00000088, -0.00000727, 0.00000528, -0.00000564, 0.00000799, -0.00001261, 0.00000805, 0.00000504, 0.00000009, -0.00000760, 0.00000178, 0.00000036, 0.00000313, 0.00000307, -0.00000394, -0.00000473, 0.00000548, -0.00000646, 0.00000176, -0.00000609, 0.00000529, -0.00000750, 0.00000060, -0.00001082, 0.00001135, -0.00000652, -0.00000025, -0.00001516, 0.00000482, -0.00002076, -0.00000091, -0.00004256, 0.00000600, -0.00016012, 0.00008973, 0.00012881, 0.00006762, 0.00005441, 0.00013460, 0.00007312, -0.00015623, -0.00000922, -0.00003447, 0.00000784, -0.00002806, 0.00000339, -0.00001512, 0.00000774, -0.00002229, 0.00000561, -0.00000945, 0.00000302, -0.00001278, 0.00000346, -0.00000806, 0.00001459, -0.00001761, 0.00000679, -0.00000945, -0.00000742, -0.00000448, 0.00001455, -0.00000471, -0.00000846, -0.00000044, 0.00000876, -0.00000369, 0.00000112, -0.00000165, 0.00000299, -0.00000450, 0.00000168, -0.00000260, 0.00000436, -0.00000185, -0.00000206, -0.00000502, -0.00000142, 0.00000085, 0.00000401, -0.00000521, 0.00001837, -0.00001331, 0.00000886, -0.00000848, 0.00000286, -0.00000033, -0.00001784, 0.00000328, -0.00000357, 0.00000028, -0.00000240, -0.00000270, -0.00000471, -0.00000242, -0.00001200, 0.00000283, -0.00000436, -0.00000824, -0.00000955, -0.00003113, 0.00000301, 0.00001011, 0.00003315, 0.00000374, 0.00000534, 0.00001702, -0.00000590, 0.00002728, -0.00000349, 0.00003422, 0.00001999, 0.00000185, -0.00002878, -0.00000923, 0.00000025, -0.00000374, 0.00000237, -0.00000655, -0.00000539, -0.00000063, -0.00000140, 0.00000157, -0.00000178, },
		{ 0.00000157, -0.00000317, -0.00000215, 0.00000235, -0.00000672, 0.00000240, -0.00000241, 0.00000288, -0.00000297, -0.00000471, 0.00001233, -0.00000175, -0.00000246, -0.00000504, 0.00000458, -0.00000126, 0.00000602, -0.00000149, -0.00000623, 0.00000244, -0.00000180, -0.00000153, -0.00000312, 0.00000090, -0.00000192, -0.00000205, -0.00000761, 0.00000167, 0.00000328, -0.00000270, -0.00001141, -0.00000092, -0.00001334, -0.00000207, -0.00003394, -0.00000133, -0.00009449, -0.00005458, 0.00017897, 0.00004447, 0.00008788, 0.00004504, 0.00017676, -0.00009405, -0.00007643, -0.00001933, -0.00000843, -0.00001378, -0.00000940, -0.00000966, 0.00000142, -0.00001060, -0.00000743, -0.00000269, -0.00000102, -0.00000771, -0.00000143, -0.00000677, 0.00000896, -0.00000448, -0.00000672, 0.00000216, -0.00001281, -0.00000505, 0.00000713, 0.00000641, -0.00001103, -0.00000310, 0.00000679, 0.00000105, -0.00000179, 0.00000000, 0.00000137, -0.00000187, -0.00000093, -0.00000080, 0.00000143, 0.00000138, -0.00000213, -0.00000430, -0.00000245, -0.00000206, 0.00000695, -0.00000703, 0.00001359, -0.00000063, -0.00000451, 0.00000528, -0.00001069, 0.00001168, -0.00001969, -0.00000183, -0.00000136, -0.00000105, -0.00000162, -0.00000092, -0.00000634, -0.00000098, -0.00000942, -0.00000548, 0.00000337, -0.00001092, -0.00000379, -0.00002476, -0.00001968, 0.00001238, 0.00001932, 0.00003016, -0.00001076, 0.00002629, -0.00000716, 0.00001990, 0.00000850, 0.00001582, 0.00003447, 0.00000815, -0.00001801, -0.00002452, 0.00000389, -0.00000774, 0.00000399, -0.00000425, -0.00000664, -0.00000228, -0.00000105, 0.00000057, -0.00000054, 0.00000077, },
		{ -0.00000246, -0.00000242, 0.00000011, -0.00000189, -0.00000425, 0.00000409, -0.00000471, 0.00000608, -0.00001193, 0.00001222, 0.00000075, -0.00000008, -0.00000750, 0.00000413, -0.00000123, 0.00000442, 0.00000208, -0.00000462, -0.00000306, 0.00000277, -0.00000459, 0.00000074, -0.00000435, 0.00000338, -0.00000559, -0.00000088, -0.00000891, 0.00000919, -0.00000375, -0.00000412, -0.00001101, -0.00000164, -0.00001344, -0.00000926, -0.00003092, -0.00002261, -0.00013402, 0.00013644, 0.00008388, 0.00007069, 0.00004347, 0.00012321, 0.00007131, -0.00013864, -0.00001834, -0.00002649, 0.00000176, -0.00002256, -0.00000077, -0.00001215, 0.00000507, -0.00001837, 0.00000172, -0.00000666, 0.00000021, -0.00000893, -0.00000012, -0.00000381, 0.00001080, -0.00001617, 0.00000698, -0.00001236, -0.00000624, -0.00000080, 0.00001275, -0.00000833, -0.00000577, 0.00000265, 0.00000620, -0.00000423, 0.00000181, -0.00000105, 0.00000156, -0.00000402, 0.00000165, -0.00000191, 0.00000400, -0.00000247, -0.00000207, -0.00000452, -0.00000216, 0.00000356, -0.00000008, 0.00000049, 0.00001479, -0.00001442, 0.00001162, -0.00001325, 0.00000936, -0.00000888, -0.00001201, 0.00000127, -0.00000210, -0.00000181, 0.00000060, -0.00000546, -0.00000322, -0.00000444, -0.00001035, 0.00000158, -0.00000331, -0.00001037, -0.00000741, -0.00003564, 0.00000311, 0.00001207, 0.00003573, 0.00000355, 0.00000783, 0.00001471, -0.00000279, 0.00002671, 0.00000046, 0.00003879, 0.00001683, -0.00000373, -0.00003403, -0.00000008, -0.00000668, 0.00000050, 0.00000014, -0.00000676, -0.00000496, -0.00000000, -0.00000171, 0.00000200, -0.00000243, 0.00000391, },
		{ -0.00000340, -0.00000137, 0.00000054, -0.00000650, 0.00000439, -0.00000424, 0.00000439, -0.00000648, 0.00000153, 0.00000931, -0.00000296, -0.00000321, -0.00000244, 0.00000395, -0.00000121, 0.00000647, -0.00000327, -0.00000446, 0.00000097, -0.00000203, -0.00000077, -0.00000302, 0.00000102, -0.00000231, -0.00000138, -0.00000771, 0.00000171, 0.00000355, -0.00000342, -0.00001101, -0.00000225, -0.00001304, -0.00000194, -0.00003320, -0.00000597, -0.00011233, 0.00001447, 0.00015233, 0.00003057, 0.00008466, 0.00003591, 0.00016350, -0.00007749, -0.00007692, -0.00001836, -0.00000820, -0.00001405, -0.00000835, -0.00001064, 0.00000158, -0.00000963, -0.00000762, -0.00000339, -0.00000096, -0.00000707, -0.00000121, -0.00000691, 0.00001044, -0.00000814, -0.00000269, -0.00000207, -0.00001241, -0.00000244, 0.00000874, 0.00000168, -0.00001052, 0.00000081, 0.00000635, -0.00000127, -0.00000085, 0.00000050, 0.00000082, -0.00000256, -0.00000028, -0.00000074, 0.00000191, 0.00000048, -0.00000242, -0.00000359, -0.00000332, -0.00000000, 0.00000526, -0.00000637, 0.00001713, -0.00000785, 0.00000197, -0.00000116, -0.00000479, 0.00000687, -0.00001974, 0.00000101, -0.00000288, -0.00000075, -0.00000145, -0.00000022, -0.00000798, -0.00000024, -0.00001064, -0.00000395, 0.00000196, -0.00001044, -0.00000481, -0.00002421, -0.00002456, 0.00001534, 0.00002092, 0.00003191, -0.00001018, 0.00002736, -0.00000864, 0.00002406, 0.00000642, 0.00002543, 0.00003078, 0.00000632, -0.00002899, -0.00001547, 0.00000084, -0.00000783, 0.00000528, -0.00000628, -0.00000601, -0.00000157, -0.00000132, 0.00000107, -0.00000114, 0.00000110, 0.00000151, },
		{ -0.00000213, 0.00000001, -0.00000419, -0.00000073, 0.00000178, -0.00000267, 0.00000269, -0.00000855, 0.00001362, -0.00000261, -0.00000047, -0.00000643, 0.00000561, -0.00000251, 0.00000554, 0.00000065, -0.00000440, -0.00000215, 0.00000029, -0.00000251, -0.00000038, -0.00000266, 0.00000142, -0.00000356, -0.00000250, -0.00000688, 0.00000703, -0.00000138, -0.00000746, -0.00000750, -0.00000720, -0.00000683, -0.00001700, -0.00002016, -0.00005516, -0.00008576, 0.00015384, 0.00004906, 0.00007398, 0.00003327, 0.00011315, 0.00006972, -0.00012265, -0.00002645, -0.00001917, -0.00000423, -0.00001701, -0.00000498, -0.00000904, 0.00000241, -0.00001459, -0.00000214, -0.00000393, -0.00000231, -0.00000535, -0.00000353, 0.00000075, 0.00000552, -0.00001246, 0.00000488, -0.00001404, -0.00000455, 0.00000261, 0.00000935, -0.00001032, -0.00000269, 0.00000501, 0.00000296, -0.00000376, 0.00000219, -0.00000066, 0.00000010, -0.00000319, 0.00000140, -0.00000100, 0.00000320, -0.00000287, -0.00000190, -0.00000432, -0.00000216, 0.00000518, -0.00000355, 0.00000712, 0.00000812, -0.00001158, 0.00001033, -0.00001374, 0.00001206, -0.00001556, -0.00000580, -0.00000153, -0.00000051, -0.00000352, 0.00000328, -0.00000828, -0.00000156, -0.00000676, -0.00000798, -0.00000028, -0.00000217, -0.00001238, -0.00000551, -0.00003972, 0.00000264, 0.00001431, 0.00003845, 0.00000296, 0.00001130, 0.00001099, 0.00000254, 0.00002373, 0.00000821, 0.00004010, 0.00001458, -0.00001366, -0.00003251, 0.00000477, -0.00001187, 0.00000435, -0.00000246, -0.00000641, -0.00000447, 0.00000029, -0.00000167, 0.00000197, -0.00000265, 0.00000418, -0.00000270, },
		{ -0.00000069, -0.00000192, -0.00000468, 0.00000469, -0.00000473, 0.00000437, -0.00000809, 0.00000696, 0.00000527, -0.00000337, -0.00000376, 0.00000037, 0.00000265, -0.00000062, 0.00000617, -0.00000447, -0.00000255, -0.00000129, -0.00000126, -0.00000073, -0.00000225, 0.00000027, -0.00000169, -0.00000175, -0.00000673, 0.00000066, 0.00000483, -0.00000555, -0.00000880, -0.00000588, -0.00000942, -0.00000626, -0.00002625, -0.00002289, -0.00010821, 0.00007084, 0.00011500, 0.00002931, 0.00007514, 0.00003242, 0.00014799, -0.00005948, -0.00007998, -0.00001467, -0.00001077, -0.00001186, -0.00000944, -0.00000963, 0.00000010, -0.00000715, -0.00000939, -0.00000261, -0.00000223, -0.00000491, -0.00000278, -0.00000477, 0.00000903, -0.00000922, -0.00000019, -0.00000622, -0.00001119, 0.00000064, 0.00000845, -0.00000229, -0.00000907, 0.00000461, 0.00000457, -0.00000268, 0.00000011, 0.00000084, -0.00000006, -0.00000272, 0.00000005, -0.00000033, 0.00000195, -0.00000031, -0.00000259, -0.00000284, -0.00000421, 0.00000246, 0.00000203, -0.00000269, 0.00001680, -0.00001214, 0.00000674, -0.00000676, 0.00000176, -0.00000066, -0.00001612, 0.00000131, -0.00000310, -0.00000142, -0.00000003, -0.00000110, -0.00000809, -0.00000105, -0.00001022, -0.00000360, 0.00000128, -0.00001076, -0.00000492, -0.00002467, -0.00002809, 0.00001636, 0.00002483, 0.00003120, -0.00000675, 0.00002544, -0.00000675, 0.00002519, 0.00000773, 0.00003329, 0.00002748, 0.00000031, -0.00003617, -0.00000581, -0.00000475, -0.00000541, 0.00000453, -0.00000708, -0.00000561, -0.00000085, -0.00000163, 0.00000151, -0.00000176, 0.00000151, 0.00000130, -0.00000341, },
		{ -0.00000092, -0.00000536, 0.00000233, -0.00000071, -0.00000043, -0.00000114, -0.00000339, 0.00001242, -0.00000468, -0.00000118, -0.00000442, 0.00000607, -0.00000330, 0.00000627, -0.00000090, -0.00000336, -0.00000228, -0.00000141, -0.00000073, -0.00000116, -0.00000152, 0.00000007, -0.00000209, -0.00000352, -0.00000550, 0.00000563, -0.00000030, -0.00000916, -0.00000600, -0.00001001, -0.00000353, -0.00002050, -0.00001627, -0.00008130, -0.00002675, 0.00014557, 0.00002867, 0.00007283, 0.00002668, 0.00010220, 0.00007011, -0.00011004, -0.00003163, -0.00001435, -0.00000834, -0.00001298, -0.00000786, -0.00000699, 0.00000084, -0.00001201, -0.00000494, -0.00000222, -0.00000358, -0.00000315, -0.00000546, 0.00000399, 0.00000049, -0.00000783, 0.00000117, -0.00001451, -0.00000222, 0.00000488, 0.00000534, -0.00001099, 0.00000077, 0.00000618, -0.00000027, -0.00000256, 0.00000224, -0.00000053, -0.00000113, -0.00000231, 0.00000120, -0.00000015, 0.00000217, -0.00000311, -0.00000152, -0.00000450, -0.00000124, 0.00000508, -0.00000500, 0.00001263, 0.00000045, -0.00000638, 0.00000606, -0.00001027, 0.00001017, -0.00001867, -0.00000096, -0.00000396, 0.00000049, -0.00000415, 0.00000472, -0.00001014, -0.00000077, -0.00000835, -0.00000584, -0.00000199, -0.00000156, -0.00001363, -0.00000456, -0.00004249, 0.00000060, 0.00001790, 0.00004004, 0.00000352, 0.00001391, 0.00000794, 0.00000805, 0.00002035, 0.00001802, 0.00003946, 0.00001094, -0.00002509, -0.00002534, 0.00000460, -0.00001392, 0.00000645, -0.00000441, -0.00000603, -0.00000378, 0.00000018, -0.00000128, 0.00000149, -0.00000246, 0.00000408, -0.00000272, -0.00000168, },
		{ -0.00000443, -0.00000177, 0.00000345, -0.00000382, 0.00000272, -0.00000728, 0.00001047, 0.00000130, -0.00000349, -0.00000375, 0.00000292, 0.00000104, 0.00000026, 0.00000522, -0.00000487, -0.00000103, -0.00000380, 0.00000016, -0.00000113, -0.00000120, -0.00000089, -0.00000057, -0.00000266, -0.00000522, -0.00000083, 0.00000631, -0.00000800, -0.00000611, -0.00001002, -0.00000474, -0.00001220, -0.00001694, -0.00004768, -0.00008222, 0.00010571, 0.00007877, 0.00003359, 0.00006309, 0.00003169, 0.00013262, -0.00004244, -0.00008330, -0.00001027, -0.00001445, -0.00000859, -0.00001152, -0.00000764, -0.00000209, -0.00000409, -0.00001185, -0.00000121, -0.00000392, -0.00000221, -0.00000508, -0.00000140, 0.00000545, -0.00000780, 0.00000029, -0.00000967, -0.00000914, 0.00000355, 0.00000661, -0.00000513, -0.00000692, 0.00000797, 0.00000170, -0.00000293, 0.00000077, 0.00000107, -0.00000114, -0.00000244, 0.00000014, 0.00000035, 0.00000155, -0.00000094, -0.00000260, -0.00000227, -0.00000468, 0.00000448, -0.00000159, 0.00000303, 0.00001274, -0.00001271, 0.00000841, -0.00000948, 0.00000636, -0.00000835, -0.00001050, -0.00000017, -0.00000251, -0.00000237, 0.00000177, -0.00000275, -0.00000736, -0.00000281, -0.00000860, -0.00000414, 0.00000102, -0.00001144, -0.00000466, -0.00002546, -0.00003101, 0.00001617, 0.00003016, 0.00002902, -0.00000142, 0.00002129, -0.00000191, 0.00002344, 0.00001274, 0.00003888, 0.00002426, -0.00001010, -0.00003710, 0.00000111, -0.00001018, -0.00000221, 0.00000271, -0.00000704, -0.00000527, -0.00000034, -0.00000174, 0.00000166, -0.00000215, 0.00000178, 0.00000109, -0.00000319, -0.00000041, },
		{ -0.00000507, 0.00000416, -0.00000251, 0.00000111, -0.00000421, 0.00000225, 0.00000936, -0.00000559, -0.00000207, -0.00000170, 0.00000558, -0.00000355, 0.00000648, -0.00000226, -0.00000179, -0.00000342, -0.00000211, 0.00000045, -0.00000140, -0.00000113, -0.00000046, -0.00000148, -0.00000365, -0.00000502, 0.00000519, -0.00000064, -0.00000904, -0.00000672, -0.00000971, -0.00000425, -0.00001849, -0.00002245, -0.00009255, 0.00003000, 0.00012060, 0.00002132, 0.00006640, 0.00002412, 0.00009019, 0.00007234, -0.00010082, -0.00003363, -0.00001247, -0.00001003, -0.00001102, -0.00000888, -0.00000643, 0.00000071, -0.00001096, -0.00000635, -0.00000176, -0.00000337, -0.00000264, -0.00000540, 0.00000498, -0.00000295, -0.00000363, -0.00000336, -0.00001379, 0.00000055, 0.00000558, 0.00000161, -0.00001070, 0.00000453, 0.00000583, -0.00000281, -0.00000110, 0.00000204, -0.00000067, -0.00000200, -0.00000155, 0.00000119, 0.00000044, 0.00000110, -0.00000322, -0.00000100, -0.00000489, 0.00000036, 0.00000324, -0.00000369, 0.00001534, -0.00000606, -0.00000092, 0.00000082, -0.00000456, 0.00000448, -0.00001784, 0.00000163, -0.00000541, 0.00000067, -0.00000360, 0.00000466, -0.00001061, -0.00000134, -0.00000863, -0.00000455, -0.00000312, -0.00000179, -0.00001386, -0.00000481, -0.00004364, -0.00000339, 0.00002324, 0.00003997, 0.00000605, 0.00001451, 0.00000711, 0.00001203, 0.00001849, 0.00002816, 0.00003798, 0.00000367, -0.00003420, -0.00001557, 0.00000064, -0.00001296, 0.00000645, -0.00000523, -0.00000588, -0.00000290, -0.00000022, -0.00000068, 0.00000074, -0.00000197, 0.00000369, -0.00000251, -0.00000121, -0.00000260, },
		{ 0.00000142, 0.00000131, -0.00000201, -0.00000007, -0.00000429, 0.00001160, -0.00000186, -0.00000370, -0.00000294, 0.00000484, -0.00000060, 0.00000123, 0.00000382, -0.00000436, -0.00000034, -0.00000595, 0.00000173, -0.00000159, -0.00000030, -0.00000190, 0.00000046, -0.00000347, -0.00000386, -0.00000205, 0.00000713, -0.00000968, -0.00000432, -0.00001285, -0.00000148, -0.00001641, -0.00001046, -0.00007255, -0.00004072, 0.00011729, 0.00005108, 0.00003768, 0.00005185, 0.00003121, 0.00011939, -0.00002839, -0.00008481, -0.00000709, -0.00001748, -0.00000576, -0.00001325, -0.00000584, -0.00000393, -0.00000151, -0.00001401, -0.00000009, -0.00000508, -0.00000008, -0.00000690, 0.00000180, 0.00000101, -0.00000467, -0.00000116, -0.00001211, -0.00000624, 0.00000552, 0.00000402, -0.00000695, -0.00000407, 0.00001033, -0.00000166, -0.00000215, 0.00000097, 0.00000119, -0.00000221, -0.00000191, 0.00000014, 0.00000111, 0.00000084, -0.00000142, -0.00000239, -0.00000201, -0.00000443, 0.00000530, -0.00000421, 0.00000910, 0.00000625, -0.00001006, 0.00000691, -0.00000859, 0.00000732, -0.00001399, -0.00000472, -0.00000238, -0.00000176, -0.00000292, 0.00000303, -0.00000418, -0.00000674, -0.00000461, -0.00000657, -0.00000500, 0.00000066, -0.00001188, -0.00000474, -0.00002576, -0.00003423, 0.00001578, 0.00003571, 0.00002681, 0.00000424, 0.00001658, 0.00000447, 0.00002004, 0.00002064, 0.00004270, 0.00001939, -0.00002287, -0.00003170, 0.00000345, -0.00001341, 0.00000010, 0.00000094, -0.00000672, -0.00000475, -0.00000024, -0.00000152, 0.00000140, -0.00000216, 0.00000176, 0.00000101, -0.00000279, -0.00000077, -0.00000636, },
		{ 0.00000449, -0.00000313, 0.00000142, -0.00000567, 0.00000710, 0.00000544, -0.00000573, -0.00000283, 0.00000133, 0.00000430, -0.00000330, 0.00000611, -0.00000311, -0.00000017, -0.00000525, -0.00000186, 0.00000105, -0.00000123, -0.00000129, -0.00000032, -0.00000155, -0.00000310, -0.00000521, 0.00000532, -0.00000187, -0.00000778, -0.00000874, -0.00000743, -0.00000764, -0.00001266, -0.00003765, -0.00008517, 0.00007391, 0.00008938, 0.00002236, 0.00005651, 0.00002415, 0.00007835, 0.00007496, -0.00009361, -0.00003356, -0.00001269, -0.00000989, -0.00001067, -0.00000844, -0.00000698, 0.00000160, -0.00001108, -0.00000672, -0.00000214, -0.00000216, -0.00000335, -0.00000371, 0.00000372, -0.00000424, -0.00000076, -0.00000798, -0.00001184, 0.00000317, 0.00000486, -0.00000138, -0.00000966, 0.00000833, 0.00000395, -0.00000413, 0.00000014, 0.00000173, -0.00000104, -0.00000245, -0.00000097, 0.00000142, 0.00000065, 0.00000012, -0.00000317, -0.00000052, -0.00000518, 0.00000205, 0.00000034, 0.00000015, 0.00001451, -0.00000988, 0.00000294, -0.00000323, 0.00000105, -0.00000311, -0.00001389, 0.00000199, -0.00000584, 0.00000030, -0.00000234, 0.00000349, -0.00000994, -0.00000309, -0.00000761, -0.00000419, -0.00000367, -0.00000270, -0.00001332, -0.00000594, -0.00004350, -0.00000900, 0.00002988, 0.00003866, 0.00001031, 0.00001306, 0.00000897, 0.00001375, 0.00001919, 0.00003770, 0.00003561, -0.00000803, -0.00003783, -0.00000679, -0.00000469, -0.00001029, 0.00000491, -0.00000506, -0.00000588, -0.00000205, -0.00000068, -0.00000013, -0.00000001, -0.00000143, 0.00000316, -0.00000207, -0.00000093, -0.00000471, -0.00000344, },
		{ -0.00000091, -0.00000003, -0.00000313, 0.00000006, 0.00001052, -0.00000390, -0.00000414, -0.00000129, 0.00000590, -0.00000205, 0.00000211, 0.00000221, -0.00000306, -0.00000071, -0.00000731, 0.00000305, -0.00000185, 0.00000016, -0.00000240, 0.00000099, -0.00000374, -0.00000309, -0.00000259, 0.00000683, -0.00000999, -0.00000413, -0.00001341, -0.00000108, -0.00001668, -0.00001094, -0.00008931, 0.00000627, 0.00010983, 0.00003425, 0.00003868, 0.00004310, 0.00002987, 0.00010895, -0.00001809, -0.00008356, -0.00000618, -0.00001878, -0.00000436, -0.00001378, -0.00000498, -0.00000476, -0.00000005, -0.00001527, 0.00000023, -0.00000518, 0.00000085, -0.00000739, 0.00000370, -0.00000295, -0.00000094, -0.00000410, -0.00001326, -0.00000275, 0.00000606, 0.00000148, -0.00000805, -0.00000047, 0.00001114, -0.00000470, -0.00000077, 0.00000075, 0.00000119, -0.00000312, -0.00000131, 0.00000020, 0.00000175, -0.00000002, -0.00000176, -0.00000196, -0.00000206, -0.00000342, 0.00000453, -0.00000473, 0.00001372, -0.00000078, -0.00000567, 0.00000339, -0.00000475, 0.00000436, -0.00001631, -0.00000023, -0.00000443, -0.00000125, -0.00000273, 0.00000325, -0.00000472, -0.00000690, -0.00000572, -0.00000484, -0.00000565, -0.00000020, -0.00001169, -0.00000558, -0.00002505, -0.00003835, 0.00001581, 0.00004062, 0.00002573, 0.00000877, 0.00001301, 0.00001068, 0.00001662, 0.00003023, 0.00004543, 0.00001071, -0.00003441, -0.00002218, 0.00000141, -0.00001373, 0.00000066, -0.00000005, -0.00000649, -0.00000397, -0.00000055, -0.00000098, 0.00000077, -0.00000182, 0.00000143, 0.00000111, -0.00000228, -0.00000185, -0.00000719, 0.00000406, },
		{ -0.00000256, 0.00000043, -0.00000518, 0.00001027, 0.00000158, -0.00000553, -0.00000315, 0.00000429, 0.00000242, -0.00000261, 0.00000518, -0.00000320, 0.00000099, -0.00000724, -0.00000099, 0.00000123, -0.00000090, -0.00000164, 0.00000007, -0.00000187, -0.00000234, -0.00000555, 0.00000537, -0.00000309, -0.00000645, -0.00001064, -0.00000511, -0.00001115, -0.00000675, -0.00005700, -0.00006109, 0.00009946, 0.00006045, 0.00002658, 0.00004582, 0.00002487, 0.00006824, 0.00007623, -0.00008661, -0.00003304, -0.00001360, -0.00000909, -0.00001093, -0.00000741, -0.00000785, 0.00000274, -0.00001165, -0.00000671, -0.00000262, -0.00000081, -0.00000438, -0.00000137, 0.00000095, -0.00000354, 0.00000037, -0.00001213, -0.00000865, 0.00000491, 0.00000337, -0.00000368, -0.00000780, 0.00001158, 0.00000098, -0.00000412, 0.00000083, 0.00000142, -0.00000157, -0.00000256, -0.00000053, 0.00000180, 0.00000052, -0.00000075, -0.00000293, -0.00000022, -0.00000501, 0.00000310, -0.00000248, 0.00000544, 0.00001053, -0.00001054, 0.00000430, -0.00000471, 0.00000446, -0.00001035, -0.00000835, 0.00000074, -0.00000556, -0.00000014, -0.00000109, 0.00000198, -0.00000882, -0.00000539, -0.00000579, -0.00000444, -0.00000395, -0.00000384, -0.00001254, -0.00000731, -0.00004279, -0.00001543, 0.00003684, 0.00003721, 0.00001519, 0.00001058, 0.00001286, 0.00001358, 0.00002248, 0.00004662, 0.00003087, -0.00002278, -0.00003482, -0.00000167, -0.00000901, -0.00000757, 0.00000285, -0.00000439, -0.00000577, -0.00000149, -0.00000094, 0.00000015, -0.00000053, -0.00000103, 0.00000264, -0.00000145, -0.00000105, -0.00000672, -0.00000096, 0.00000349, },
		{ 0.00000140, -0.00000554, 0.00000470, 0.00000787, -0.00000491, -0.00000474, 0.00000108, 0.00000599, -0.00000313, 0.00000272, 0.00000071, -0.00000129, -0.00000206, -0.00000766, 0.00000388, -0.00000179, 0.00000016, -0.00000234, 0.00000094, -0.00000342, -0.00000294, -0.00000248, 0.00000554, -0.00000907, -0.00000530, -0.00001199, -0.00000332, -0.00001310, -0.00001960, -0.00009226, 0.00004856, 0.00009099, 0.00002660, 0.00003642, 0.00003678, 0.00002793, 0.00010063, -0.00001094, -0.00007989, -0.00000743, -0.00001831, -0.00000445, -0.00001299, -0.00000512, -0.00000455, 0.00000023, -0.00001556, -0.00000022, -0.00000431, 0.00000060, -0.00000644, 0.00000380, -0.00000551, 0.00000240, -0.00000795, -0.00001288, 0.00000072, 0.00000524, -0.00000052, -0.00000861, 0.00000375, 0.00001011, -0.00000668, 0.00000064, 0.00000030, 0.00000104, -0.00000376, -0.00000076, 0.00000042, 0.00000210, -0.00000089, -0.00000199, -0.00000143, -0.00000224, -0.00000198, 0.00000238, -0.00000274, 0.00001557, -0.00000654, -0.00000130, -0.00000041, 0.00000031, -0.00000151, -0.00001520, 0.00000231, -0.00000582, -0.00000105, -0.00000197, 0.00000247, -0.00000428, -0.00000803, -0.00000583, -0.00000375, -0.00000591, -0.00000161, -0.00001086, -0.00000717, -0.00002337, -0.00004336, 0.00001625, 0.00004478, 0.00002613, 0.00001154, 0.00001161, 0.00001558, 0.00001445, 0.00004069, 0.00004691, -0.00000299, -0.00004108, -0.00001206, -0.00000315, -0.00001185, -0.00000040, -0.00000013, -0.00000641, -0.00000306, -0.00000111, -0.00000031, -0.00000002, -0.00000129, 0.00000089, 0.00000139, -0.00000180, -0.00000353, -0.00000671, 0.00000554, -0.00000250, },
		{ -0.00000145, -0.00000297, 0.00001142, -0.00000160, -0.00000535, -0.00000278, 0.00000679, 0.00000021, -0.00000158, 0.00000376, -0.00000243, 0.00000129, -0.00000886, 0.00000009, 0.00000126, -0.00000068, -0.00000183, 0.00000032, -0.00000201, -0.00000181, -0.00000557, 0.00000479, -0.00000361, -0.00000590, -0.00001130, -0.00000435, -0.00001241, -0.00000480, -0.00007372, -0.00002677, 0.00010679, 0.00003862, 0.00003024, 0.00003632, 0.00002497, 0.00006073, 0.00007511, -0.00007859, -0.00003334, -0.00001399, -0.00000866, -0.00001089, -0.00000658, -0.00000832, 0.00000341, -0.00001202, -0.00000687, -0.00000262, -0.00000000, -0.00000485, 0.00000058, -0.00000225, -0.00000152, -0.00000026, -0.00001528, -0.00000449, 0.00000520, 0.00000191, -0.00000558, -0.00000489, 0.00001356, -0.00000228, -0.00000307, 0.00000090, 0.00000117, -0.00000215, -0.00000242, -0.00000013, 0.00000216, 0.00000014, -0.00000150, -0.00000246, -0.00000021, -0.00000422, 0.00000295, -0.00000406, 0.00001059, 0.00000469, -0.00000867, 0.00000338, -0.00000337, 0.00000456, -0.00001546, -0.00000279, -0.00000130, -0.00000490, -0.00000033, -0.00000041, 0.00000083, -0.00000795, -0.00000752, -0.00000381, -0.00000486, -0.00000435, -0.00000479, -0.00001201, -0.00000836, -0.00004215, -0.00002193, 0.00004317, 0.00003680, 0.00001935, 0.00000854, 0.00001753, 0.00001255, 0.00002786, 0.00005521, 0.00002158, -0.00003722, -0.00002651, -0.00000101, -0.00001094, -0.00000602, 0.00000119, -0.00000370, -0.00000536, -0.00000138, -0.00000090, 0.00000008, -0.00000072, -0.00000089, 0.00000225, -0.00000071, -0.00000175, -0.00000810, 0.00000165, 0.00000174, -0.00000120, },
		{ -0.00000665, 0.00000862, 0.00000447, -0.00000517, -0.00000526, 0.00000389, 0.00000509, -0.00000369, 0.00000292, -0.00000037, 0.00000049, -0.00000405, -0.00000710, 0.00000420, -0.00000152, -0.00000012, -0.00000193, 0.00000055, -0.00000276, -0.00000313, -0.00000215, 0.00000382, -0.00000759, -0.00000692, -0.00000984, -0.00000655, -0.00000793, -0.00003414, -0.00008013, 0.00007884, 0.00006870, 0.00002452, 0.00003229, 0.00003185, 0.00002638, 0.00009309, -0.00000557, -0.00007494, -0.00000990, -0.00001676, -0.00000548, -0.00001139, -0.00000582, -0.00000375, -0.00000028, -0.00001524, -0.00000098, -0.00000299, -0.00000028, -0.00000456, 0.00000233, -0.00000637, 0.00000465, -0.00001216, -0.00001080, 0.00000333, 0.00000365, -0.00000202, -0.00000850, 0.00000809, 0.00000739, -0.00000716, 0.00000156, -0.00000016, 0.00000070, -0.00000407, -0.00000029, 0.00000078, 0.00000213, -0.00000169, -0.00000205, -0.00000093, -0.00000227, -0.00000068, -0.00000039, 0.00000129, 0.00001425, -0.00000990, 0.00000170, -0.00000287, 0.00000460, -0.00000848, -0.00001153, 0.00000290, -0.00000646, -0.00000095, -0.00000110, 0.00000120, -0.00000328, -0.00000978, -0.00000514, -0.00000321, -0.00000591, -0.00000330, -0.00000974, -0.00000906, -0.00002117, -0.00004872, 0.00001646, 0.00004882, 0.00002746, 0.00001293, 0.00001243, 0.00001887, 0.00001405, 0.00005192, 0.00004563, -0.00002084, -0.00004076, -0.00000461, -0.00000777, -0.00000924, -0.00000227, 0.00000034, -0.00000625, -0.00000231, -0.00000165, 0.00000023, -0.00000072, -0.00000079, 0.00000032, 0.00000180, -0.00000150, -0.00000549, -0.00000508, 0.00000565, -0.00000305, 0.00000185, },
		{ 0.00000029, 0.00001071, -0.00000382, -0.00000534, -0.00000160, 0.00000851, -0.00000207, -0.00000034, 0.00000205, -0.00000091, 0.00000054, -0.00000967, 0.00000103, 0.00000129, -0.00000069, -0.00000169, 0.00000023, -0.00000178, -0.00000172, -0.00000512, 0.00000342, -0.00000321, -0.00000631, -0.00001048, -0.00000563, -0.00001052, -0.00000924, -0.00008171, 0.00000932, 0.00010028, 0.00002482, 0.00003177, 0.00002872, 0.00002424, 0.00005561, 0.00007163, -0.00006931, -0.00003486, -0.00001339, -0.00000905, -0.00001014, -0.00000631, -0.00000811, 0.00000332, -0.00001192, -0.00000737, -0.00000199, 0.00000001, -0.00000440, 0.00000143, -0.00000494, 0.00000100, -0.00000242, -0.00001690, 0.00000000, 0.00000400, 0.00000101, -0.00000734, -0.00000084, 0.00001364, -0.00000490, -0.00000157, 0.00000047, 0.00000096, -0.00000266, -0.00000217, 0.00000033, 0.00000235, -0.00000033, -0.00000213, -0.00000179, -0.00000043, -0.00000289, 0.00000145, -0.00000359, 0.00001408, -0.00000134, -0.00000556, 0.00000130, -0.00000011, 0.00000147, -0.00001756, 0.00000165, -0.00000345, -0.00000403, -0.00000026, -0.00000043, 0.00000032, -0.00000769, -0.00000904, -0.00000210, -0.00000518, -0.00000502, -0.00000539, -0.00001190, -0.00000889, -0.00004185, -0.00002821, 0.00004839, 0.00003811, 0.00002190, 0.00000813, 0.00002182, 0.00001171, 0.00003494, 0.00006312, 0.00000608, -0.00004730, -0.00001611, -0.00000369, -0.00001052, -0.00000597, 0.00000028, -0.00000317, -0.00000467, -0.00000165, -0.00000063, -0.00000024, -0.00000066, -0.00000097, 0.00000200, 0.00000008, -0.00000304, -0.00000844, 0.00000367, -0.00000006, 0.00000037, -0.00000353, },
		{ 0.00001115, 0.00000107, -0.00000506, -0.00000545, 0.00000682, 0.00000331, -0.00000360, 0.00000258, -0.00000073, 0.00000175, -0.00000615, -0.00000589, 0.00000411, -0.00000118, -0.00000043, -0.00000149, 0.00000015, -0.00000213, -0.00000326, -0.00000210, 0.00000234, -0.00000632, -0.00000798, -0.00000842, -0.00000866, -0.00000457, -0.00004964, -0.00005612, 0.00009428, 0.00004877, 0.00002453, 0.00002787, 0.00002730, 0.00002602, 0.00008521, -0.00000068, -0.00006996, -0.00001251, -0.00001503, -0.00000665, -0.00000966, -0.00000649, -0.00000294, -0.00000105, -0.00001476, -0.00000156, -0.00000184, -0.00000108, -0.00000257, -0.00000000, -0.00000575, 0.00000545, -0.00001609, -0.00000711, 0.00000434, 0.00000216, -0.00000342, -0.00000731, 0.00001171, 0.00000366, -0.00000620, 0.00000170, -0.00000047, 0.00000019, -0.00000409, 0.00000012, 0.00000119, 0.00000187, -0.00000238, -0.00000192, -0.00000060, -0.00000191, -0.00000011, -0.00000273, 0.00000623, 0.00001031, -0.00001066, 0.00000286, -0.00000317, 0.00000659, -0.00001468, -0.00000661, 0.00000200, -0.00000642, -0.00000076, -0.00000062, 0.00000005, -0.00000230, -0.00001159, -0.00000411, -0.00000295, -0.00000594, -0.00000488, -0.00000876, -0.00001076, -0.00001907, -0.00005377, 0.00001560, 0.00005370, 0.00002870, 0.00001398, 0.00001474, 0.00002099, 0.00001541, 0.00006415, 0.00003906, -0.00003969, -0.00003384, -0.00000156, -0.00001057, -0.00000732, -0.00000399, 0.00000084, -0.00000574, -0.00000199, -0.00000194, 0.00000046, -0.00000113, -0.00000051, -0.00000015, 0.00000228, -0.00000155, -0.00000728, -0.00000278, 0.00000459, -0.00000253, 0.00000129, -0.00000624, },
		{ 0.00000864, -0.00000509, -0.00000548, 0.00000035, 0.00000916, -0.00000408, 0.00000087, 0.00000033, 0.00000106, -0.00000118, -0.00000951, 0.00000159, 0.00000140, -0.00000088, -0.00000130, -0.00000014, -0.00000124, -0.00000196, -0.00000439, 0.00000156, -0.00000220, -0.00000723, -0.00000886, -0.00000806, -0.00000654, -0.00001966, -0.00007775, 0.00003964, 0.00008616, 0.00001741, 0.00003133, 0.00002266, 0.00002326, 0.00005186, 0.00006675, -0.00005944, -0.00003715, -0.00001204, -0.00001006, -0.00000884, -0.00000646, -0.00000735, 0.00000257, -0.00001145, -0.00000799, -0.00000101, -0.00000051, -0.00000320, 0.00000109, -0.00000657, 0.00000333, -0.00000577, -0.00001650, 0.00000385, 0.00000188, 0.00000067, -0.00000886, 0.00000408, 0.00001159, -0.00000611, -0.00000030, -0.00000014, 0.00000068, -0.00000300, -0.00000187, 0.00000087, 0.00000229, -0.00000079, -0.00000264, -0.00000104, -0.00000071, -0.00000139, -0.00000102, -0.00000096, 0.00001495, -0.00000614, -0.00000255, -0.00000057, 0.00000348, -0.00000367, -0.00001674, 0.00000440, -0.00000528, -0.00000294, -0.00000027, -0.00000085, 0.00000028, -0.00000795, -0.00000991, -0.00000075, -0.00000543, -0.00000581, -0.00000582, -0.00001197, -0.00000914, -0.00004162, -0.00003459, 0.00005276, 0.00004098, 0.00002280, 0.00000977, 0.00002518, 0.00001157, 0.00004396, 0.00006862, -0.00001535, -0.00004993, -0.00000731, -0.00000739, -0.00000892, -0.00000684, -0.00000009, -0.00000262, -0.00000398, -0.00000206, -0.00000037, -0.00000058, -0.00000055, -0.00000112, 0.00000185, 0.00000082, -0.00000480, -0.00000762, 0.00000461, -0.00000130, 0.00000156, -0.00000515, 0.00000376, },
		{ -0.00000180, -0.00000486, -0.00000504, 0.00000945, 0.00000083, -0.00000284, 0.00000166, -0.00000022, 0.00000210, -0.00000782, -0.00000442, 0.00000377, -0.00000090, -0.00000057, -0.00000124, 0.00000000, -0.00000181, -0.00000308, -0.00000261, 0.00000147, -0.00000564, -0.00000801, -0.00000852, -0.00000830, -0.00000584, -0.00006050, -0.00002639, 0.00009630, 0.00003395, 0.00002450, 0.00002401, 0.00002280, 0.00002687, 0.00007666, 0.00000431, -0.00006563, -0.00001457, -0.00001370, -0.00000746, -0.00000823, -0.00000673, -0.00000251, -0.00000177, -0.00001433, -0.00000174, -0.00000118, -0.00000135, -0.00000114, -0.00000243, -0.00000412, 0.00000462, -0.00001901, -0.00000239, 0.00000347, 0.00000149, -0.00000515, -0.00000462, 0.00001369, -0.00000008, -0.00000428, 0.00000110, -0.00000055, -0.00000042, -0.00000387, 0.00000053, 0.00000151, 0.00000145, -0.00000297, -0.00000157, -0.00000048, -0.00000109, -0.00000062, -0.00000369, 0.00001067, 0.00000498, -0.00000945, 0.00000257, -0.00000149, 0.00000568, -0.00001874, -0.00000174, 0.00000020, -0.00000575, -0.00000051, -0.00000069, -0.00000064, -0.00000176, -0.00001298, -0.00000316, -0.00000272, -0.00000618, -0.00000613, -0.00000817, -0.00001196, -0.00001741, -0.00005811, 0.00001314, 0.00006009, 0.00002901, 0.00001574, 0.00001767, 0.00002254, 0.00001860, 0.00007710, 0.00002484, -0.00005503, -0.00002311, -0.00000259, -0.00001099, -0.00000677, -0.00000501, 0.00000109, -0.00000480, -0.00000215, -0.00000194, 0.00000037, -0.00000124, -0.00000046, -0.00000047, 0.00000279, -0.00000207, -0.00000844, -0.00000044, 0.00000287, -0.00000127, 0.00000001, -0.00000450, 0.00001202, },
		{ -0.00000561, -0.00000565, 0.00000294, 0.00000856, -0.00000543, 0.00000176, -0.00000101, 0.00000299, -0.00000348, -0.00000843, 0.00000170, 0.00000156, -0.00000110, -0.00000087, -0.00000053, -0.00000068, -0.00000225, -0.00000378, -0.00000030, -0.00000113, -0.00000787, -0.00000761, -0.00000997, -0.00000303, -0.00003293, -0.00006249, 0.00005968, 0.00007007, 0.00001387, 0.00002978, 0.00001754, 0.00002272, 0.00004839, 0.00006161, -0.00004999, -0.00003939, -0.00001060, -0.00001114, -0.00000747, -0.00000662, -0.00000647, 0.00000151, -0.00001090, -0.00000837, -0.00000016, -0.00000104, -0.00000179, -0.00000009, -0.00000706, 0.00000494, -0.00000981, -0.00001387, 0.00000600, -0.00000017, 0.00000040, -0.00000962, 0.00000908, 0.00000781, -0.00000562, 0.00000022, -0.00000063, 0.00000027, -0.00000311, -0.00000155, 0.00000145, 0.00000196, -0.00000120, -0.00000298, -0.00000032, -0.00000082, -0.00000020, -0.00000369, 0.00000321, 0.00001307, -0.00000890, -0.00000052, -0.00000116, 0.00000584, -0.00000928, -0.00001379, 0.00000542, -0.00000655, -0.00000157, -0.00000073, -0.00000117, 0.00000025, -0.00000831, -0.00001043, 0.00000039, -0.00000580, -0.00000642, -0.00000643, -0.00001181, -0.00000961, -0.00004092, -0.00004168, 0.00005696, 0.00004468, 0.00002275, 0.00001312, 0.00002773, 0.00001224, 0.00005558, 0.00006862, -0.00003986, -0.00004443, -0.00000277, -0.00000987, -0.00000763, -0.00000771, -0.00000042, -0.00000173, -0.00000362, -0.00000231, -0.00000035, -0.00000071, -0.00000059, -0.00000119, 0.00000171, 0.00000141, -0.00000673, -0.00000582, 0.00000432, -0.00000163, 0.00000205, -0.00000584, 0.00000669, 0.00000585, },
		},
		{ { -0.00000105, -0.00000290, 0.00000979, -0.00000428, -0.00000048, 0.00000202, -0.00000492, 0.00000425, -0.00001019, 0.00000448, -0.00000196, -0.00000429, -0.00000388, -0.00000137, -0.00000320, -0.00000083, -0.00000112, -0.00000088, -0.00000762, -0.00000045, 0.00000008, -0.00000073, -0.00000387, -0.00000482, -0.00000487, 0.00000046, -0.00001880, -0.00000291, -0.00000569, -0.00000779, -0.00001023, -0.00001630, -0.00005490, 0.00000447, 0.00007324, 0.00002799, 0.00002332, 0.00002042, 0.00003299, 0.00002830, 0.00007380, 0.00000172, -0.00005673, -0.00001583, -0.00001139, -0.00000640, -0.00001359, -0.00000288, -0.00000295, -0.00000348, 0.00000684, -0.00000204, -0.00000078, -0.00000111, -0.00000352, -0.00000395, 0.00000306, -0.00000362, 0.00000223, -0.00000127, 0.00000018, 0.00000073, 0.00000203, -0.00000130, -0.00000192, -0.00000130, 0.00000203, 0.00000073, 0.00000018, -0.00000127, 0.00000223, -0.00000362, 0.00000306, -0.00000395, -0.00000352, -0.00000111, -0.00000078, -0.00000204, 0.00000684, -0.00000348, -0.00000295, -0.00000288, -0.00001359, -0.00000640, -0.00001139, -0.00001583, -0.00005673, 0.00000172, 0.00007380, 0.00002830, 0.00003299, 0.00002042, 0.00002332, 0.00002799, 0.00007324, 0.00000447, -0.00005490, -0.00001630, -0.00001023, -0.00000779, -0.00000569, -0.00000291, -0.00001880, 0.00000046, -0.00000487, -0.00000482, -0.00000387, -0.00000073, 0.00000008, -0.00000045, -0.00000762, -0.00000088, -0.00000112, -0.00000083, -0.00000320, -0.00000137, -0.00000388, -0.00000429, -0.00000196, 0.00000448, -0.00001019, 0.00000425, -0.00000492, 0.00000202, -0.00000048, -0.00000428, 0.00000979, -0.00000290, },
		{ -0.00000198, 0.00000056, 0.00000904, -0.00000852, 0.00000584, -0.00000517, 0.00000154, -0.00000161, -0.00000662, 0.00000452, -0.00000450, -0.00000348, -0.00000347, -0.00000160, -0.00000285, -0.00000035, -0.00000055, -0.00000458, -0.00000553, 0.00000157, -0.00000098, -0.00000180, -0.00000419, -0.00000589, -0.00000036, -0.00000863, -0.00001525, 0.00000127, -0.00001040, -0.00000659, -0.00001135, -0.00003282, -0.00004422, 0.00005486, 0.00005000, 0.00002232, 0.00001906, 0.00002626, 0.00002665, 0.00004383, 0.00006171, -0.00003271, -0.00004361, -0.00001035, -0.00000962, -0.00000753, -0.00001265, 0.00000106, -0.00000699, 0.00000337, 0.00000295, -0.00000192, -0.00000082, -0.00000100, -0.00000579, 0.00000008, 0.00000051, -0.00000171, 0.00000179, -0.00000186, 0.00000087, 0.00000131, 0.00000103, -0.00000188, -0.00000209, 0.00000036, 0.00000158, 0.00000090, -0.00000125, 0.00000125, -0.00000068, -0.00000188, 0.00000222, -0.00000564, -0.00000187, -0.00000089, -0.00000214, 0.00000153, 0.00000535, -0.00000709, 0.00000081, -0.00000907, -0.00001116, -0.00000816, -0.00001158, -0.00002782, -0.00005845, 0.00004576, 0.00006270, 0.00002513, 0.00003318, 0.00001981, 0.00002398, 0.00005133, 0.00006292, -0.00004753, -0.00003507, -0.00001288, -0.00000729, -0.00000975, 0.00000044, -0.00001444, -0.00001021, 0.00000091, -0.00000681, -0.00000431, -0.00000239, -0.00000049, 0.00000133, -0.00000526, -0.00000465, -0.00000010, -0.00000115, -0.00000220, -0.00000208, -0.00000249, -0.00000335, -0.00000609, 0.00000378, -0.00000141, -0.00000726, 0.00000409, -0.00000621, 0.00000605, -0.00000672, 0.00000298, 0.00000608, -0.00000323, },
		{ -0.00000382, 0.00000654, 0.00000274, -0.00000599, 0.00000614, -0.00000840, 0.00000660, -0.00000835, -0.00000022, 0.00000048, -0.00000362, -0.00000443, -0.00000190, -0.00000287, -0.00000133, -0.00000037, -0.00000142, -0.00000752, -0.00000069, 0.00000024, -0.00000106, -0.00000325, -0.00000448, -0.00000469, -0.00000005, -0.00001736, -0.00000422, -0.00000412, -0.00000943, -0.00000760, -0.00001633, -0.00004928, 0.00000037, 0.00006857, 0.00002647, 0.00002324, 0.00001614, 0.00003011, 0.00002392, 0.00005935, 0.00003207, -0.00004970, -0.00002842, -0.00000848, -0.00000750, -0.00001050, -0.00000780, -0.00000014, -0.00000613, 0.00000724, -0.00000164, -0.00000010, -0.00000149, -0.00000177, -0.00000618, 0.00000320, -0.00000236, 0.00000141, -0.00000077, -0.00000056, 0.00000081, 0.00000198, -0.00000063, -0.00000185, -0.00000186, 0.00000176, 0.00000086, 0.00000066, -0.00000139, 0.00000260, -0.00000389, 0.00000154, -0.00000118, -0.00000482, -0.00000115, -0.00000111, -0.00000249, 0.00000572, -0.00000040, -0.00000543, 0.00000043, -0.00001377, -0.00000772, -0.00001219, -0.00001028, -0.00004805, -0.00003636, 0.00007927, 0.00004007, 0.00003290, 0.00002678, 0.00002351, 0.00002909, 0.00007852, 0.00000908, -0.00006127, -0.00001549, -0.00001372, -0.00000586, -0.00000721, -0.00000173, -0.00002011, 0.00000069, -0.00000455, -0.00000547, -0.00000433, -0.00000073, 0.00000018, -0.00000027, -0.00000783, -0.00000034, -0.00000139, -0.00000104, -0.00000304, -0.00000127, -0.00000318, -0.00000428, -0.00000475, 0.00000727, -0.00000835, -0.00000090, -0.00000057, -0.00000177, 0.00000373, -0.00000804, 0.00000860, 0.00000093, -0.00000184, },
		{ -0.00000380, 0.00001132, -0.00000570, 0.00000173, -0.00000064, -0.00000379, 0.00000464, -0.00000992, 0.00000317, -0.00000259, -0.00000289, -0.00000483, -0.00000100, -0.00000347, 0.00000004, -0.00000052, -0.00000459, -0.00000601, 0.00000201, -0.00000151, -0.00000133, -0.00000415, -0.00000496, -0.00000133, -0.00000730, -0.00001585, 0.00000181, -0.00001063, -0.00000617, -0.00001020, -0.00003009, -0.00004147, 0.00004761, 0.00004888, 0.00001997, 0.00001973, 0.00001936, 0.00002772, 0.00003106, 0.00006203, -0.00000175, -0.00005126, -0.00001560, -0.00000837, -0.00000635, -0.00001270, -0.00000190, -0.00000461, 0.00000005, 0.00000482, -0.00000239, 0.00000016, -0.00000124, -0.00000430, -0.00000314, 0.00000276, -0.00000229, 0.00000213, -0.00000225, 0.00000064, 0.00000118, 0.00000141, -0.00000147, -0.00000216, -0.00000039, 0.00000165, 0.00000104, -0.00000072, 0.00000082, 0.00000029, -0.00000362, 0.00000271, -0.00000389, -0.00000350, -0.00000046, -0.00000266, 0.00000053, 0.00000576, -0.00000501, -0.00000180, -0.00000377, -0.00001502, -0.00000622, -0.00001521, -0.00001273, -0.00006900, 0.00001198, 0.00008337, 0.00002663, 0.00003837, 0.00002264, 0.00002524, 0.00005232, 0.00007229, -0.00005142, -0.00003737, -0.00001402, -0.00000883, -0.00000850, -0.00000070, -0.00001343, -0.00001201, 0.00000227, -0.00000743, -0.00000475, -0.00000282, -0.00000034, 0.00000147, -0.00000521, -0.00000478, 0.00000055, -0.00000177, -0.00000213, -0.00000205, -0.00000209, -0.00000270, -0.00000702, 0.00000149, 0.00000384, -0.00001013, 0.00000278, -0.00000398, 0.00000347, -0.00000236, -0.00000384, 0.00000986, -0.00000278, -0.00000104, },
		{ 0.00000081, 0.00000988, -0.00000948, 0.00000746, -0.00000822, 0.00000450, -0.00000292, -0.00000579, 0.00000201, -0.00000265, -0.00000417, -0.00000330, -0.00000179, -0.00000252, 0.00000064, -0.00000185, -0.00000737, -0.00000126, 0.00000089, -0.00000180, -0.00000244, -0.00000454, -0.00000403, -0.00000079, -0.00001572, -0.00000574, -0.00000246, -0.00001088, -0.00000535, -0.00001654, -0.00004325, -0.00000414, 0.00006487, 0.00002461, 0.00002286, 0.00001387, 0.00002524, 0.00002336, 0.00004406, 0.00004865, -0.00002942, -0.00004201, -0.00000763, -0.00000829, -0.00000728, -0.00001144, 0.00000063, -0.00000674, 0.00000600, -0.00000053, -0.00000002, -0.00000123, -0.00000079, -0.00000672, 0.00000137, -0.00000002, 0.00000016, 0.00000011, -0.00000153, 0.00000101, 0.00000180, -0.00000006, -0.00000159, -0.00000228, 0.00000125, 0.00000096, 0.00000115, -0.00000144, 0.00000282, -0.00000363, -0.00000037, 0.00000083, -0.00000464, -0.00000232, -0.00000055, -0.00000343, 0.00000471, 0.00000162, -0.00000573, 0.00000039, -0.00000996, -0.00001216, -0.00000955, -0.00001246, -0.00002888, -0.00007015, 0.00006669, 0.00006097, 0.00003040, 0.00003454, 0.00002420, 0.00003011, 0.00008357, 0.00001537, -0.00006941, -0.00001344, -0.00001804, -0.00000384, -0.00000859, -0.00000061, -0.00002136, 0.00000067, -0.00000386, -0.00000613, -0.00000496, -0.00000069, 0.00000005, 0.00000028, -0.00000840, 0.00000029, -0.00000135, -0.00000142, -0.00000314, -0.00000086, -0.00000298, -0.00000339, -0.00000743, 0.00000792, -0.00000412, -0.00000618, 0.00000181, -0.00000295, 0.00000486, -0.00000754, 0.00000339, 0.00000641, -0.00000332, -0.00000261, },
		{ 0.00000877, 0.00000128, -0.00000477, 0.00000507, -0.00000879, 0.00000837, -0.00000944, -0.00000057, -0.00000052, -0.00000198, -0.00000572, -0.00000119, -0.00000324, -0.00000059, 0.00000023, -0.00000481, -0.00000639, 0.00000218, -0.00000157, -0.00000136, -0.00000386, -0.00000440, -0.00000174, -0.00000641, -0.00001601, 0.00000181, -0.00001038, -0.00000596, -0.00000932, -0.00002750, -0.00003819, 0.00003990, 0.00004899, 0.00001644, 0.00002148, 0.00001325, 0.00002725, 0.00002407, 0.00005349, 0.00002418, -0.00004581, -0.00002703, -0.00000535, -0.00000724, -0.00000991, -0.00000669, -0.00000157, -0.00000321, 0.00000627, -0.00000298, 0.00000141, -0.00000188, -0.00000222, -0.00000603, 0.00000370, -0.00000172, 0.00000176, -0.00000199, -0.00000010, 0.00000139, 0.00000143, -0.00000093, -0.00000208, -0.00000105, 0.00000144, 0.00000125, -0.00000018, 0.00000043, 0.00000117, -0.00000468, 0.00000193, -0.00000152, -0.00000472, -0.00000060, -0.00000258, -0.00000065, 0.00000535, -0.00000226, -0.00000435, 0.00000012, -0.00001589, -0.00000708, -0.00001628, -0.00000572, -0.00006051, -0.00003349, 0.00009613, 0.00003644, 0.00003959, 0.00002849, 0.00002589, 0.00005331, 0.00008230, -0.00005471, -0.00004116, -0.00001361, -0.00001187, -0.00000635, -0.00000233, -0.00001201, -0.00001420, 0.00000372, -0.00000771, -0.00000541, -0.00000320, -0.00000037, 0.00000166, -0.00000498, -0.00000525, 0.00000144, -0.00000212, -0.00000247, -0.00000194, -0.00000170, -0.00000227, -0.00000681, -0.00000183, 0.00000787, -0.00000970, -0.00000130, -0.00000065, 0.00000067, 0.00000161, -0.00000829, 0.00000965, 0.00000054, -0.00000165, -0.00000525, },
		{ 0.00001365, -0.00000821, 0.00000416, -0.00000360, -0.00000133, 0.00000405, -0.00001016, 0.00000177, -0.00000143, -0.00000311, -0.00000518, -0.00000058, -0.00000361, 0.00000112, -0.00000173, -0.00000734, -0.00000201, 0.00000170, -0.00000252, -0.00000177, -0.00000471, -0.00000323, -0.00000149, -0.00001405, -0.00000727, -0.00000094, -0.00001209, -0.00000319, -0.00001742, -0.00003664, -0.00000867, 0.00006126, 0.00002346, 0.00002127, 0.00001373, 0.00001924, 0.00002402, 0.00003201, 0.00005239, -0.00000396, -0.00004890, -0.00001252, -0.00000674, -0.00000607, -0.00001187, -0.00000175, -0.00000494, 0.00000311, 0.00000147, -0.00000097, 0.00000009, -0.00000102, -0.00000538, -0.00000194, 0.00000261, -0.00000091, 0.00000087, -0.00000226, 0.00000088, 0.00000188, 0.00000015, -0.00000106, -0.00000260, 0.00000074, 0.00000075, 0.00000183, -0.00000153, 0.00000301, -0.00000314, -0.00000204, 0.00000148, -0.00000293, -0.00000428, 0.00000045, -0.00000444, 0.00000385, 0.00000216, -0.00000371, -0.00000237, -0.00000384, -0.00001743, -0.00000593, -0.00001785, -0.00000948, -0.00008717, 0.00003182, 0.00008714, 0.00002943, 0.00004134, 0.00002671, 0.00003080, 0.00008824, 0.00002353, -0.00007905, -0.00001095, -0.00002218, -0.00000253, -0.00000941, 0.00000019, -0.00002232, 0.00000021, -0.00000286, -0.00000657, -0.00000595, -0.00000038, -0.00000055, 0.00000132, -0.00000927, 0.00000089, -0.00000114, -0.00000154, -0.00000387, 0.00000002, -0.00000325, -0.00000206, -0.00000929, 0.00000655, 0.00000095, -0.00000948, 0.00000094, -0.00000136, 0.00000343, -0.00000416, -0.00000332, 0.00001086, -0.00000353, -0.00000128, -0.00000435, },
		{ 0.00000945, -0.00001011, 0.00000830, -0.00000999, 0.00000722, -0.00000465, -0.00000603, 0.00000133, -0.00000120, -0.00000543, -0.00000266, -0.00000187, -0.00000218, 0.00000139, -0.00000506, -0.00000658, 0.00000185, -0.00000100, -0.00000189, -0.00000330, -0.00000431, -0.00000156, -0.00000601, -0.00001558, 0.00000118, -0.00000963, -0.00000601, -0.00000826, -0.00002599, -0.00003328, 0.00003093, 0.00005054, 0.00001202, 0.00002355, 0.00000853, 0.00002521, 0.00002083, 0.00004253, 0.00003957, -0.00002895, -0.00003951, -0.00000424, -0.00000787, -0.00000673, -0.00001059, -0.00000021, -0.00000469, 0.00000617, -0.00000273, 0.00000183, -0.00000177, -0.00000087, -0.00000722, 0.00000246, 0.00000030, 0.00000059, -0.00000106, -0.00000128, 0.00000177, 0.00000137, -0.00000054, -0.00000173, -0.00000167, 0.00000110, 0.00000122, 0.00000067, -0.00000016, 0.00000215, -0.00000533, 0.00000066, 0.00000025, -0.00000448, -0.00000203, -0.00000142, -0.00000233, 0.00000489, -0.00000042, -0.00000473, 0.00000044, -0.00001217, -0.00001196, -0.00001288, -0.00000826, -0.00003815, -0.00007691, 0.00008919, 0.00005827, 0.00003635, 0.00003729, 0.00002609, 0.00005474, 0.00009206, -0.00005613, -0.00004757, -0.00001109, -0.00001646, -0.00000345, -0.00000439, -0.00001013, -0.00001676, 0.00000517, -0.00000773, -0.00000599, -0.00000391, -0.00000013, 0.00000134, -0.00000408, -0.00000639, 0.00000268, -0.00000243, -0.00000264, -0.00000248, -0.00000063, -0.00000267, -0.00000533, -0.00000548, 0.00001003, -0.00000669, -0.00000598, 0.00000126, -0.00000025, 0.00000327, -0.00000891, 0.00000530, 0.00000593, -0.00000332, -0.00000415, 0.00000292, },
		{ -0.00000188, -0.00000274, 0.00000331, -0.00000766, 0.00000909, -0.00001096, -0.00000106, 0.00000028, -0.00000195, -0.00000643, -0.00000038, -0.00000353, 0.00000031, -0.00000070, -0.00000765, -0.00000256, 0.00000212, -0.00000264, -0.00000167, -0.00000462, -0.00000274, -0.00000173, -0.00001273, -0.00000833, -0.00000002, -0.00001278, -0.00000122, -0.00001861, -0.00003037, -0.00001180, 0.00005611, 0.00002443, 0.00001754, 0.00001582, 0.00001260, 0.00002482, 0.00002366, 0.00004832, 0.00001800, -0.00004455, -0.00002305, -0.00000359, -0.00000673, -0.00000946, -0.00000620, -0.00000208, -0.00000019, 0.00000352, -0.00000234, 0.00000179, -0.00000175, -0.00000318, -0.00000528, 0.00000415, -0.00000082, 0.00000082, -0.00000218, 0.00000003, 0.00000245, -0.00000005, -0.00000039, -0.00000279, 0.00000030, 0.00000024, 0.00000251, -0.00000140, 0.00000295, -0.00000235, -0.00000340, 0.00000107, -0.00000061, -0.00000584, 0.00000073, -0.00000455, 0.00000258, 0.00000186, -0.00000082, -0.00000538, 0.00000118, -0.00001981, -0.00000531, -0.00002046, 0.00000094, -0.00008142, -0.00001989, 0.00010768, 0.00003692, 0.00004358, 0.00003304, 0.00003019, 0.00009342, 0.00003249, -0.00008867, -0.00000982, -0.00002462, -0.00000291, -0.00000906, 0.00000022, -0.00002254, -0.00000100, -0.00000148, -0.00000680, -0.00000716, -0.00000000, -0.00000132, 0.00000233, -0.00000986, 0.00000102, -0.00000058, -0.00000149, -0.00000487, 0.00000074, -0.00000323, -0.00000128, -0.00000958, 0.00000348, 0.00000558, -0.00001001, -0.00000248, 0.00000128, 0.00000147, -0.00000046, -0.00000847, 0.00001180, -0.00000102, -0.00000141, -0.00000671, 0.00001230, },
		{ -0.00001079, 0.00000640, -0.00000589, 0.00000136, 0.00000272, -0.00001120, 0.00000193, -0.00000043, -0.00000411, -0.00000488, -0.00000016, -0.00000373, 0.00000203, -0.00000475, -0.00000687, 0.00000129, -0.00000022, -0.00000240, -0.00000287, -0.00000440, -0.00000110, -0.00000580, -0.00001480, 0.00000025, -0.00000876, -0.00000615, -0.00000682, -0.00002591, -0.00002677, 0.00002136, 0.00005230, 0.00000826, 0.00002440, 0.00000612, 0.00002147, 0.00001989, 0.00003250, 0.00004485, -0.00000746, -0.00004618, -0.00000845, -0.00000632, -0.00000524, -0.00001156, -0.00000189, -0.00000369, 0.00000418, -0.00000118, 0.00000083, -0.00000039, -0.00000084, -0.00000631, -0.00000070, 0.00000309, -0.00000073, -0.00000005, -0.00000230, 0.00000176, 0.00000169, -0.00000060, -0.00000103, -0.00000229, 0.00000087, 0.00000068, 0.00000193, -0.00000094, 0.00000316, -0.00000565, -0.00000060, 0.00000087, -0.00000270, -0.00000440, 0.00000026, -0.00000398, 0.00000439, -0.00000002, -0.00000259, -0.00000246, -0.00000535, -0.00001863, -0.00000754, -0.00001548, -0.00001365, -0.00010278, 0.00005543, 0.00008952, 0.00003248, 0.00004636, 0.00002752, 0.00005601, 0.00010172, -0.00005569, -0.00005619, -0.00000741, -0.00002141, -0.00000078, -0.00000626, -0.00000825, -0.00001927, 0.00000625, -0.00000741, -0.00000631, -0.00000513, 0.00000062, 0.00000027, -0.00000244, -0.00000806, 0.00000413, -0.00000278, -0.00000232, -0.00000392, 0.00000111, -0.00000368, -0.00000328, -0.00000843, 0.00000995, -0.00000226, -0.00000936, 0.00000051, 0.00000107, 0.00000273, -0.00000652, -0.00000117, 0.00001097, -0.00000421, -0.00000243, -0.00000308, 0.00001491, },
		{ -0.00000963, 0.00000841, -0.00001039, 0.00000908, -0.00000693, -0.00000636, 0.00000240, -0.00000132, -0.00000614, -0.00000189, -0.00000187, -0.00000179, 0.00000096, -0.00000815, -0.00000278, 0.00000189, -0.00000199, -0.00000215, -0.00000423, -0.00000274, -0.00000135, -0.00001199, -0.00000863, 0.00000007, -0.00001283, 0.00000035, -0.00001949, -0.00002561, -0.00001219, 0.00004844, 0.00002789, 0.00001188, 0.00001932, 0.00000618, 0.00002510, 0.00001820, 0.00004130, 0.00003174, -0.00003027, -0.00003470, -0.00000219, -0.00000717, -0.00000639, -0.00001020, -0.00000048, -0.00000211, 0.00000431, -0.00000292, 0.00000257, -0.00000174, -0.00000157, -0.00000706, 0.00000342, 0.00000098, -0.00000036, -0.00000113, -0.00000151, 0.00000330, -0.00000032, 0.00000010, -0.00000264, -0.00000018, -0.00000029, 0.00000274, -0.00000064, 0.00000231, -0.00000109, -0.00000462, 0.00000023, 0.00000118, -0.00000589, -0.00000053, -0.00000326, 0.00000065, 0.00000146, 0.00000126, -0.00000623, 0.00000240, -0.00001699, -0.00000971, -0.00001738, -0.00000040, -0.00005744, -0.00007322, 0.00010789, 0.00005906, 0.00003921, 0.00004407, 0.00002780, 0.00010000, 0.00004107, -0.00009675, -0.00001146, -0.00002456, -0.00000524, -0.00000752, -0.00000061, -0.00002186, -0.00000304, 0.00000022, -0.00000688, -0.00000825, -0.00000002, -0.00000166, 0.00000261, -0.00000952, 0.00000022, 0.00000063, -0.00000167, -0.00000540, 0.00000039, -0.00000209, -0.00000173, -0.00000811, -0.00000056, 0.00000883, -0.00000805, -0.00000672, 0.00000286, 0.00000086, 0.00000165, -0.00001030, 0.00000861, 0.00000392, -0.00000293, -0.00000613, 0.00000685, 0.00000686, },
		{ -0.00000035, 0.00000178, -0.00000607, 0.00000900, -0.00001327, -0.00000002, 0.00000096, -0.00000273, -0.00000645, 0.00000043, -0.00000387, 0.00000112, -0.00000334, -0.00000763, 0.00000102, 0.00000008, -0.00000221, -0.00000307, -0.00000423, -0.00000089, -0.00000526, -0.00001424, -0.00000032, -0.00000841, -0.00000591, -0.00000531, -0.00002669, -0.00001994, 0.00001309, 0.00005213, 0.00000711, 0.00002244, 0.00000682, 0.00001585, 0.00002065, 0.00002399, 0.00004389, 0.00001122, -0.00004286, -0.00001805, -0.00000320, -0.00000550, -0.00000968, -0.00000591, -0.00000126, 0.00000129, 0.00000098, -0.00000099, 0.00000159, -0.00000150, -0.00000420, -0.00000436, 0.00000514, -0.00000105, 0.00000014, -0.00000239, 0.00000082, 0.00000269, -0.00000115, -0.00000011, -0.00000283, 0.00000076, -0.00000025, 0.00000324, -0.00000142, 0.00000371, -0.00000535, -0.00000192, 0.00000063, -0.00000031, -0.00000643, 0.00000113, -0.00000448, 0.00000327, -0.00000047, 0.00000067, -0.00000601, 0.00000086, -0.00002275, -0.00000508, -0.00002008, 0.00000140, -0.00010207, -0.00000133, 0.00011869, 0.00003634, 0.00005083, 0.00003329, 0.00005528, 0.00011288, -0.00005497, -0.00006511, -0.00000477, -0.00002485, 0.00000034, -0.00000702, -0.00000712, -0.00002098, 0.00000641, -0.00000652, -0.00000647, -0.00000667, 0.00000160, -0.00000113, -0.00000070, -0.00000954, 0.00000514, -0.00000280, -0.00000177, -0.00000568, 0.00000262, -0.00000425, -0.00000192, -0.00000953, 0.00000760, 0.00000251, -0.00001057, -0.00000248, 0.00000325, 0.00000160, -0.00000327, -0.00000685, 0.00001307, -0.00000259, -0.00000196, -0.00000654, 0.00001516, -0.00000566, },
		{ 0.00000813, -0.00000711, 0.00000343, 0.00000076, -0.00001236, 0.00000393, -0.00000110, -0.00000445, -0.00000456, 0.00000048, -0.00000406, 0.00000218, -0.00000814, -0.00000305, 0.00000135, -0.00000106, -0.00000263, -0.00000396, -0.00000296, -0.00000063, -0.00001158, -0.00000839, -0.00000034, -0.00001264, 0.00000175, -0.00001983, -0.00002280, -0.00000987, 0.00003901, 0.00003239, 0.00000617, 0.00002219, 0.00000162, 0.00002379, 0.00001537, 0.00003363, 0.00003772, -0.00001209, -0.00004111, -0.00000587, -0.00000550, -0.00000477, -0.00001157, -0.00000176, -0.00000162, 0.00000311, -0.00000190, 0.00000161, -0.00000025, -0.00000136, -0.00000655, 0.00000037, 0.00000405, -0.00000209, 0.00000029, -0.00000304, 0.00000375, -0.00000009, 0.00000004, -0.00000203, -0.00000079, -0.00000058, 0.00000223, 0.00000085, 0.00000121, 0.00000041, -0.00000559, -0.00000072, 0.00000194, -0.00000433, -0.00000298, -0.00000124, -0.00000129, 0.00000095, 0.00000186, -0.00000428, -0.00000015, -0.00001012, -0.00001710, -0.00001100, -0.00000851, -0.00002834, -0.00010926, 0.00007689, 0.00009538, 0.00003128, 0.00005742, 0.00002535, 0.00010726, 0.00004959, -0.00010341, -0.00001543, -0.00002301, -0.00000831, -0.00000579, -0.00000161, -0.00002083, -0.00000534, 0.00000172, -0.00000665, -0.00000905, -0.00000069, -0.00000126, 0.00000184, -0.00000813, -0.00000144, 0.00000233, -0.00000215, -0.00000510, -0.00000126, 0.00000008, -0.00000311, -0.00000573, -0.00000421, 0.00000990, -0.00000429, -0.00001018, 0.00000223, 0.00000198, 0.00000200, -0.00000915, 0.00000282, 0.00000916, -0.00000416, -0.00000436, 0.00000037, 0.00001378, -0.00001231, },
		{ 0.00000864, -0.00001045, 0.00001036, -0.00001020, -0.00000503, 0.00000330, -0.00000226, -0.00000615, -0.00000112, -0.00000196, -0.00000151, -0.00000098, -0.00000880, 0.00000125, -0.00000045, -0.00000106, -0.00000396, -0.00000371, -0.00000121, -0.00000407, -0.00001432, 0.00000001, -0.00000906, -0.00000496, -0.00000409, -0.00002746, -0.00001435, 0.00000788, 0.00004858, 0.00000941, 0.00001752, 0.00001010, 0.00000909, 0.00002229, 0.00001696, 0.00004028, 0.00002318, -0.00003072, -0.00002888, -0.00000165, -0.00000559, -0.00000689, -0.00000989, 0.00000036, -0.00000069, 0.00000230, -0.00000225, 0.00000271, -0.00000150, -0.00000244, -0.00000668, 0.00000497, 0.00000055, -0.00000106, -0.00000123, -0.00000108, 0.00000412, -0.00000181, 0.00000063, -0.00000301, 0.00000057, -0.00000117, 0.00000395, -0.00000098, 0.00000332, -0.00000416, -0.00000348, 0.00000014, 0.00000154, -0.00000686, 0.00000017, -0.00000319, 0.00000119, -0.00000098, 0.00000317, -0.00000741, 0.00000313, -0.00002115, -0.00000859, -0.00001776, 0.00000230, -0.00007786, -0.00006497, 0.00012817, 0.00005694, 0.00004665, 0.00004571, 0.00005106, 0.00012718, -0.00005573, -0.00007233, -0.00000500, -0.00002558, -0.00000071, -0.00000629, -0.00000715, -0.00002150, 0.00000539, -0.00000502, -0.00000660, -0.00000809, 0.00000221, -0.00000206, 0.00000019, -0.00000993, 0.00000503, -0.00000197, -0.00000162, -0.00000675, 0.00000274, -0.00000331, -0.00000222, -0.00000835, 0.00000357, 0.00000656, -0.00000957, -0.00000631, 0.00000450, 0.00000145, -0.00000098, -0.00000976, 0.00001114, 0.00000168, -0.00000311, -0.00000675, 0.00001054, 0.00000313, -0.00000844, },
		{ 0.00000108, -0.00000508, 0.00000844, -0.00001600, 0.00000320, -0.00000035, -0.00000257, -0.00000674, 0.00000176, -0.00000483, 0.00000167, -0.00000685, -0.00000393, 0.00000119, -0.00000070, -0.00000228, -0.00000441, -0.00000291, -0.00000013, -0.00001091, -0.00000830, -0.00000040, -0.00001305, 0.00000362, -0.00002006, -0.00002125, -0.00000635, 0.00003012, 0.00003529, 0.00000301, 0.00002212, 0.00000055, 0.00001982, 0.00001545, 0.00002552, 0.00003926, 0.00000338, -0.00003830, -0.00001491, -0.00000215, -0.00000482, -0.00001005, -0.00000558, 0.00000062, 0.00000057, 0.00000029, -0.00000069, 0.00000215, -0.00000209, -0.00000445, -0.00000371, 0.00000685, -0.00000308, 0.00000098, -0.00000363, 0.00000311, 0.00000100, -0.00000063, -0.00000109, -0.00000140, -0.00000065, 0.00000114, 0.00000254, 0.00000036, 0.00000138, -0.00000572, -0.00000204, 0.00000204, -0.00000213, -0.00000519, -0.00000004, -0.00000187, -0.00000042, 0.00000162, -0.00000097, -0.00000383, -0.00000322, -0.00002256, -0.00000705, -0.00001464, -0.00000854, -0.00011505, 0.00001577, 0.00013465, 0.00002909, 0.00006710, 0.00002693, 0.00011251, 0.00006034, -0.00011085, -0.00001930, -0.00002259, -0.00000982, -0.00000559, -0.00000149, -0.00002055, -0.00000684, 0.00000220, -0.00000569, -0.00000976, -0.00000175, -0.00000044, 0.00000054, -0.00000645, -0.00000307, 0.00000375, -0.00000242, -0.00000441, -0.00000341, 0.00000212, -0.00000406, -0.00000406, -0.00000591, 0.00000826, 0.00000052, -0.00001208, -0.00000025, 0.00000366, 0.00000187, -0.00000679, -0.00000288, 0.00001204, -0.00000319, -0.00000354, -0.00000398, 0.00001562, -0.00000863, 0.00000187, },
		{ -0.00000834, 0.00000526, -0.00000244, -0.00001180, 0.00000582, -0.00000271, -0.00000395, -0.00000437, 0.00000116, -0.00000470, 0.00000118, -0.00000963, 0.00000158, -0.00000145, 0.00000053, -0.00000493, -0.00000326, -0.00000184, -0.00000239, -0.00001491, 0.00000110, -0.00001041, -0.00000367, -0.00000300, -0.00002785, -0.00001066, 0.00000597, 0.00004226, 0.00001377, 0.00001156, 0.00001370, 0.00000329, 0.00002301, 0.00001225, 0.00003517, 0.00002923, -0.00001541, -0.00003485, -0.00000507, -0.00000357, -0.00000533, -0.00001154, -0.00000070, -0.00000040, 0.00000169, -0.00000175, 0.00000181, 0.00000014, -0.00000212, -0.00000661, 0.00000211, 0.00000389, -0.00000325, 0.00000064, -0.00000323, 0.00000523, -0.00000191, 0.00000073, -0.00000261, 0.00000013, -0.00000170, 0.00000366, 0.00000056, 0.00000210, -0.00000239, -0.00000500, -0.00000043, 0.00000236, -0.00000552, -0.00000233, -0.00000080, -0.00000105, -0.00000160, 0.00000411, -0.00000571, 0.00000098, -0.00001441, -0.00001655, -0.00001042, -0.00000634, -0.00004432, -0.00011301, 0.00010263, 0.00009682, 0.00003531, 0.00006325, 0.00004457, 0.00014424, -0.00005788, -0.00007788, -0.00000777, -0.00002453, -0.00000277, -0.00000505, -0.00000763, -0.00002137, 0.00000380, -0.00000350, -0.00000649, -0.00000922, 0.00000210, -0.00000206, -0.00000023, -0.00000899, 0.00000375, -0.00000032, -0.00000208, -0.00000662, 0.00000109, -0.00000086, -0.00000389, -0.00000577, -0.00000059, 0.00000870, -0.00000662, -0.00000970, 0.00000387, 0.00000261, 0.00000004, -0.00000985, 0.00000626, 0.00000685, -0.00000448, -0.00000521, 0.00000394, 0.00001124, -0.00001310, 0.00001023, },
		{ -0.00001140, 0.00001165, -0.00001449, -0.00000097, 0.00000176, -0.00000187, -0.00000678, 0.00000053, -0.00000307, -0.00000101, -0.00000423, -0.00000551, 0.00000177, -0.00000142, -0.00000070, -0.00000575, -0.00000239, -0.00000025, -0.00000946, -0.00000899, 0.00000067, -0.00001476, 0.00000646, -0.00002073, -0.00001984, -0.00000359, 0.00002406, 0.00003453, 0.00000386, 0.00001831, 0.00000305, 0.00001343, 0.00001789, 0.00001716, 0.00003907, 0.00001290, -0.00002738, -0.00002541, -0.00000016, -0.00000478, -0.00000736, -0.00000982, 0.00000254, -0.00000158, 0.00000220, -0.00000280, 0.00000388, -0.00000234, -0.00000236, -0.00000686, 0.00000763, -0.00000211, 0.00000013, -0.00000276, 0.00000124, 0.00000272, -0.00000147, -0.00000033, -0.00000163, -0.00000083, 0.00000008, 0.00000355, 0.00000069, 0.00000105, -0.00000450, -0.00000405, 0.00000217, -0.00000060, -0.00000565, -0.00000103, -0.00000015, -0.00000316, 0.00000148, 0.00000158, -0.00000536, -0.00000039, -0.00002180, -0.00001001, -0.00001261, -0.00000546, -0.00009114, -0.00005936, 0.00015669, 0.00004456, 0.00006686, 0.00003657, 0.00011297, 0.00007609, -0.00012176, -0.00002033, -0.00002580, -0.00000799, -0.00000803, 0.00000057, -0.00002182, -0.00000677, 0.00000110, -0.00000375, -0.00001068, -0.00000263, 0.00000001, -0.00000027, -0.00000571, -0.00000353, 0.00000390, -0.00000170, -0.00000423, -0.00000472, 0.00000249, -0.00000314, -0.00000445, -0.00000480, 0.00000419, 0.00000539, -0.00001224, -0.00000340, 0.00000432, 0.00000259, -0.00000491, -0.00000635, 0.00001108, 0.00000059, -0.00000451, -0.00000488, 0.00001192, -0.00000001, -0.00000729, 0.00000979, },
		{ -0.00000410, 0.00000662, -0.00001720, 0.00000666, -0.00000296, -0.00000122, -0.00000748, 0.00000346, -0.00000646, 0.00000158, -0.00000913, 0.00000127, -0.00000208, 0.00000153, -0.00000500, -0.00000360, -0.00000224, -0.00000083, -0.00001538, 0.00000220, -0.00001150, -0.00000304, -0.00000127, -0.00002836, -0.00000819, 0.00000577, 0.00003570, 0.00001713, 0.00000772, 0.00001464, 0.00000091, 0.00002076, 0.00001121, 0.00002796, 0.00003271, -0.00000314, -0.00003195, -0.00001403, 0.00000037, -0.00000545, -0.00001025, -0.00000451, 0.00000193, -0.00000067, 0.00000050, -0.00000099, 0.00000309, -0.00000303, -0.00000451, -0.00000238, 0.00000756, -0.00000512, 0.00000205, -0.00000456, 0.00000520, -0.00000101, 0.00000012, -0.00000180, -0.00000041, -0.00000190, 0.00000258, 0.00000254, 0.00000098, -0.00000108, -0.00000560, -0.00000160, 0.00000269, -0.00000351, -0.00000474, 0.00000083, -0.00000179, -0.00000329, 0.00000421, -0.00000235, -0.00000287, -0.00000680, -0.00002350, -0.00000463, -0.00001425, -0.00001904, -0.00012759, 0.00003910, 0.00014614, 0.00002632, 0.00007938, 0.00004051, 0.00016083, -0.00005856, -0.00008448, -0.00001017, -0.00002477, -0.00000314, -0.00000537, -0.00000697, -0.00002201, 0.00000300, -0.00000307, -0.00000550, -0.00001036, 0.00000155, -0.00000147, -0.00000141, -0.00000754, 0.00000227, 0.00000126, -0.00000253, -0.00000584, -0.00000138, 0.00000172, -0.00000531, -0.00000376, -0.00000285, 0.00000787, -0.00000203, -0.00001210, 0.00000177, 0.00000396, 0.00000082, -0.00000857, 0.00000090, 0.00001023, -0.00000399, -0.00000423, -0.00000109, 0.00001464, -0.00001111, 0.00000483, 0.00000003, },
		{ 0.00000783, -0.00000737, -0.00000791, 0.00000488, -0.00000268, -0.00000434, -0.00000305, 0.00000055, -0.00000479, -0.00000141, -0.00000706, 0.00000273, -0.00000291, 0.00000161, -0.00000738, -0.00000181, -0.00000086, -0.00000727, -0.00001055, 0.00000295, -0.00001768, 0.00001007, -0.00002178, -0.00001805, -0.00000257, 0.00002148, 0.00003032, 0.00000766, 0.00001248, 0.00000693, 0.00000684, 0.00002048, 0.00001022, 0.00003726, 0.00001783, -0.00001389, -0.00003133, -0.00000307, -0.00000259, -0.00000570, -0.00001204, 0.00000204, -0.00000171, 0.00000236, -0.00000314, 0.00000342, -0.00000086, -0.00000166, -0.00000760, 0.00000549, 0.00000101, -0.00000214, -0.00000079, -0.00000119, 0.00000425, -0.00000173, -0.00000028, -0.00000118, -0.00000139, -0.00000045, 0.00000325, 0.00000255, -0.00000060, -0.00000220, -0.00000645, 0.00000245, -0.00000026, -0.00000403, -0.00000409, 0.00000326, -0.00000650, 0.00000144, 0.00000231, -0.00000330, -0.00000280, -0.00001457, -0.00001926, -0.00000328, -0.00001553, -0.00005208, -0.00012260, 0.00014117, 0.00008409, 0.00005585, 0.00005427, 0.00010876, 0.00009736, -0.00013678, -0.00001799, -0.00003274, -0.00000344, -0.00001223, 0.00000383, -0.00002416, -0.00000552, -0.00000110, -0.00000135, -0.00001161, -0.00000305, -0.00000044, 0.00000011, -0.00000660, -0.00000229, 0.00000252, 0.00000023, -0.00000504, -0.00000440, 0.00000055, -0.00000010, -0.00000681, -0.00000164, -0.00000075, 0.00000891, -0.00001056, -0.00000622, 0.00000320, 0.00000438, -0.00000389, -0.00000720, 0.00000690, 0.00000581, -0.00000619, -0.00000344, 0.00000524, 0.00000927, -0.00001414, 0.00001373, -0.00001112, },
		{ 0.00001205, -0.00001780, 0.00000398, -0.00000166, 0.00000023, -0.00000837, 0.00000304, -0.00000534, -0.00000058, -0.00000696, -0.00000001, -0.00000176, 0.00000122, -0.00000356, -0.00000503, -0.00000208, 0.00000008, -0.00001503, 0.00000240, -0.00001125, -0.00000408, 0.00000190, -0.00002977, -0.00000558, 0.00000494, 0.00003174, 0.00001677, 0.00000824, 0.00001126, 0.00000288, 0.00001497, 0.00001397, 0.00001861, 0.00003593, 0.00000323, -0.00002118, -0.00002490, 0.00000333, -0.00000566, -0.00000747, -0.00000924, 0.00000446, -0.00000324, 0.00000314, -0.00000418, 0.00000573, -0.00000378, -0.00000196, -0.00000650, 0.00000962, -0.00000520, 0.00000191, -0.00000431, 0.00000373, 0.00000069, -0.00000073, -0.00000115, -0.00000054, -0.00000224, 0.00000151, 0.00000378, 0.00000123, -0.00000141, -0.00000439, -0.00000397, 0.00000343, -0.00000236, -0.00000521, -0.00000013, 0.00000036, -0.00000691, 0.00000469, 0.00000018, -0.00000456, -0.00000336, -0.00002384, -0.00000659, -0.00001285, -0.00001308, -0.00010541, -0.00004760, 0.00018287, 0.00003428, 0.00008561, 0.00004474, 0.00017274, -0.00005376, -0.00009603, -0.00000842, -0.00002972, 0.00000073, -0.00000905, -0.00000373, -0.00002477, 0.00000426, -0.00000471, -0.00000307, -0.00001202, 0.00000133, -0.00000130, -0.00000207, -0.00000709, 0.00000206, 0.00000149, -0.00000189, -0.00000564, -0.00000296, 0.00000246, -0.00000456, -0.00000415, -0.00000183, 0.00000387, 0.00000349, -0.00001336, -0.00000066, 0.00000400, 0.00000257, -0.00000747, -0.00000270, 0.00001000, -0.00000059, -0.00000514, -0.00000257, 0.00001196, -0.00000294, -0.00000514, 0.00001054, -0.00001253, },
		{ 0.00000249, -0.00001470, 0.00000712, -0.00000386, -0.00000082, -0.00000708, 0.00000385, -0.00000747, -0.00000017, -0.00000741, 0.00000317, -0.00000421, 0.00000350, -0.00000817, -0.00000196, -0.00000139, -0.00000490, -0.00001236, 0.00000566, -0.00002075, 0.00001333, -0.00002237, -0.00001637, -0.00000269, 0.00002086, 0.00002524, 0.00001112, 0.00000818, 0.00000866, 0.00000325, 0.00002026, 0.00000719, 0.00003216, 0.00002189, -0.00000404, -0.00002845, -0.00001188, 0.00000188, -0.00000581, -0.00001127, -0.00000140, 0.00000043, 0.00000047, -0.00000123, 0.00000051, 0.00000234, -0.00000249, -0.00000592, 0.00000109, 0.00000513, -0.00000459, 0.00000107, -0.00000300, 0.00000464, -0.00000086, -0.00000108, -0.00000016, -0.00000219, -0.00000047, 0.00000187, 0.00000524, -0.00000249, -0.00000008, -0.00000806, 0.00000213, -0.00000041, -0.00000152, -0.00000743, 0.00000622, -0.00000842, 0.00000031, 0.00000203, 0.00000093, -0.00000779, -0.00000524, -0.00002898, 0.00000611, -0.00002716, -0.00001878, -0.00015038, 0.00007969, 0.00014050, 0.00004282, 0.00007366, 0.00010471, 0.00012084, -0.00015297, -0.00001523, -0.00004023, 0.00000046, -0.00001519, 0.00000594, -0.00002570, -0.00000476, -0.00000277, 0.00000014, -0.00001175, -0.00000341, -0.00000150, 0.00000139, -0.00000864, -0.00000016, 0.00000056, 0.00000247, -0.00000620, -0.00000300, -0.00000275, 0.00000358, -0.00000935, 0.00000138, -0.00000413, 0.00000947, -0.00000691, -0.00000843, 0.00000088, 0.00000605, -0.00000281, -0.00000662, 0.00000173, 0.00000985, -0.00000644, -0.00000207, -0.00000063, 0.00001449, -0.00001440, 0.00000995, -0.00000332, -0.00000169, },
		{ -0.00001215, -0.00000233, 0.00000138, -0.00000037, -0.00000624, -0.00000038, -0.00000166, -0.00000446, -0.00000411, -0.00000165, -0.00000072, -0.00000022, -0.00000101, -0.00000703, -0.00000170, 0.00000035, -0.00001368, 0.00000136, -0.00000929, -0.00000697, 0.00000654, -0.00003224, -0.00000203, 0.00000209, 0.00003159, 0.00001228, 0.00001260, 0.00000482, 0.00000740, 0.00000762, 0.00001834, 0.00000921, 0.00003815, 0.00000561, -0.00000809, -0.00003162, 0.00000154, -0.00000374, -0.00000544, -0.00001239, 0.00000499, -0.00000421, 0.00000446, -0.00000581, 0.00000622, -0.00000287, -0.00000059, -0.00000843, 0.00000871, -0.00000285, 0.00000005, -0.00000268, 0.00000143, 0.00000233, -0.00000094, -0.00000133, 0.00000019, -0.00000316, 0.00000117, 0.00000338, 0.00000354, -0.00000368, -0.00000139, -0.00000739, 0.00000481, -0.00000278, -0.00000308, -0.00000391, 0.00000509, -0.00001187, 0.00000578, 0.00000041, -0.00000196, -0.00000620, -0.00001610, -0.00001709, -0.00000130, -0.00002485, -0.00006038, -0.00012869, 0.00018145, 0.00007008, 0.00007809, 0.00005938, 0.00017856, -0.00004164, -0.00011437, -0.00000094, -0.00004033, 0.00000893, -0.00001577, 0.00000186, -0.00002960, 0.00000759, -0.00000825, 0.00000052, -0.00001421, 0.00000192, -0.00000233, -0.00000116, -0.00000873, 0.00000400, -0.00000028, 0.00000042, -0.00000684, -0.00000243, 0.00000026, -0.00000098, -0.00000726, 0.00000208, -0.00000195, 0.00000846, -0.00001320, -0.00000259, 0.00000211, 0.00000539, -0.00000699, -0.00000387, 0.00000630, 0.00000478, -0.00000727, -0.00000109, 0.00000520, 0.00000741, -0.00001429, 0.00001722, -0.00001468, 0.00001023, },
		{ -0.00001742, 0.00000612, -0.00000359, 0.00000160, -0.00000904, 0.00000443, -0.00000710, -0.00000173, -0.00000589, 0.00000251, -0.00000447, 0.00000399, -0.00000730, -0.00000333, -0.00000140, -0.00000302, -0.00001357, 0.00000767, -0.00002263, 0.00001492, -0.00002135, -0.00001585, -0.00000232, 0.00001950, 0.00002269, 0.00001081, 0.00000846, 0.00000567, 0.00000458, 0.00001562, 0.00000925, 0.00002300, 0.00002750, -0.00000062, -0.00001744, -0.00002335, 0.00000594, -0.00000630, -0.00000863, -0.00000632, 0.00000337, -0.00000230, 0.00000186, -0.00000351, 0.00000576, -0.00000358, -0.00000322, -0.00000363, 0.00000821, -0.00000558, 0.00000153, -0.00000325, 0.00000345, 0.00000099, -0.00000221, 0.00000074, -0.00000260, -0.00000060, 0.00000038, 0.00000728, -0.00000295, 0.00000021, -0.00000750, 0.00000015, 0.00000018, -0.00000002, -0.00000872, 0.00000631, -0.00000694, -0.00000327, 0.00000237, 0.00000437, -0.00001080, 0.00000016, -0.00003211, 0.00000738, -0.00002890, -0.00000686, -0.00013404, -0.00001579, 0.00019098, 0.00004456, 0.00008423, 0.00010832, 0.00014103, -0.00016520, -0.00001703, -0.00004340, -0.00000071, -0.00001348, 0.00000436, -0.00002432, -0.00000644, -0.00000206, -0.00000076, -0.00001013, -0.00000449, -0.00000221, 0.00000236, -0.00001030, 0.00000105, -0.00000019, 0.00000342, -0.00000634, -0.00000204, -0.00000531, 0.00000548, -0.00000966, 0.00000181, -0.00000391, 0.00000624, -0.00000152, -0.00001012, -0.00000139, 0.00000605, -0.00000050, -0.00000608, -0.00000211, 0.00001055, -0.00000377, -0.00000259, -0.00000299, 0.00001328, -0.00000738, -0.00000010, 0.00000841, -0.00001186, 0.00000992, },
		{ -0.00000977, 0.00000448, -0.00000203, -0.00000231, -0.00000501, 0.00000237, -0.00000771, -0.00000252, -0.00000234, 0.00000006, -0.00000176, 0.00000143, -0.00000838, -0.00000192, 0.00000049, -0.00001181, -0.00000042, -0.00000628, -0.00001076, 0.00001157, -0.00003496, 0.00000205, -0.00000235, 0.00003390, 0.00000616, 0.00001750, -0.00000094, 0.00001058, 0.00000243, 0.00002059, 0.00000330, 0.00003659, 0.00000840, 0.00000081, -0.00002949, -0.00000654, 0.00000083, -0.00000524, -0.00001257, 0.00000251, -0.00000282, 0.00000358, -0.00000487, 0.00000381, 0.00000010, -0.00000092, -0.00000766, 0.00000504, 0.00000110, -0.00000250, -0.00000077, -0.00000050, 0.00000285, 0.00000014, -0.00000259, 0.00000173, -0.00000456, 0.00000164, 0.00000145, 0.00000727, -0.00000680, 0.00000198, -0.00001045, 0.00000586, -0.00000399, 0.00000043, -0.00000867, 0.00001007, -0.00001591, 0.00000610, -0.00000087, 0.00000370, -0.00001296, -0.00000441, -0.00003034, 0.00001296, -0.00004190, -0.00001657, -0.00017429, 0.00012677, 0.00013054, 0.00006382, 0.00007892, 0.00018262, -0.00002505, -0.00013701, 0.00000963, -0.00005357, 0.00001809, -0.00002249, 0.00000740, -0.00003459, 0.00001125, -0.00001197, 0.00000376, -0.00001602, 0.00000294, -0.00000439, 0.00000119, -0.00001215, 0.00000744, -0.00000322, 0.00000362, -0.00000898, -0.00000015, -0.00000408, 0.00000404, -0.00001132, 0.00000657, -0.00000688, 0.00001085, -0.00001103, -0.00000404, -0.00000096, 0.00000796, -0.00000621, -0.00000355, 0.00000111, 0.00000965, -0.00000853, 0.00000094, -0.00000169, 0.00001471, -0.00001734, 0.00001586, -0.00000833, 0.00000140, -0.00000233, },
		{ 0.00000091, -0.00000104, 0.00000157, -0.00000755, 0.00000145, -0.00000352, -0.00000559, -0.00000324, 0.00000115, -0.00000372, 0.00000303, -0.00000495, -0.00000553, -0.00000109, -0.00000179, -0.00001374, 0.00000832, -0.00002256, 0.00001425, -0.00001834, -0.00001705, -0.00000026, 0.00001547, 0.00002456, 0.00000558, 0.00001360, -0.00000153, 0.00000972, 0.00000791, 0.00001463, 0.00001185, 0.00003351, -0.00000138, -0.00000388, -0.00003135, 0.00000562, -0.00000483, -0.00000635, -0.00001022, 0.00000485, -0.00000405, 0.00000427, -0.00000647, 0.00000728, -0.00000324, -0.00000130, -0.00000658, 0.00000853, -0.00000409, 0.00000010, -0.00000190, 0.00000119, 0.00000291, -0.00000266, 0.00000067, -0.00000196, -0.00000150, -0.00000018, 0.00000735, -0.00000083, -0.00000212, -0.00000441, -0.00000363, 0.00000202, -0.00000059, -0.00000681, 0.00000273, -0.00000183, -0.00000915, 0.00000400, 0.00000494, -0.00000879, -0.00000182, -0.00002551, -0.00000227, -0.00001735, -0.00001791, -0.00008542, -0.00011465, 0.00020487, 0.00007685, 0.00007866, 0.00012428, 0.00015463, -0.00017006, -0.00002680, -0.00003926, -0.00000915, -0.00000597, -0.00000147, -0.00001950, -0.00001123, 0.00000162, -0.00000439, -0.00000665, -0.00000662, -0.00000177, 0.00000188, -0.00001010, -0.00000021, 0.00000161, 0.00000197, -0.00000443, -0.00000283, -0.00000538, 0.00000388, -0.00000645, -0.00000123, 0.00000007, 0.00000018, 0.00000428, -0.00001100, -0.00000280, 0.00000373, 0.00000309, -0.00000605, -0.00000366, 0.00000758, 0.00000139, -0.00000483, -0.00000181, 0.00000691, 0.00000350, -0.00001114, 0.00001773, -0.00001651, 0.00001031, -0.00001360, },
		{ 0.00000491, -0.00000263, 0.00000093, -0.00000795, 0.00000390, -0.00000793, -0.00000381, -0.00000110, -0.00000027, -0.00000233, 0.00000252, -0.00000803, -0.00000343, 0.00000107, -0.00001014, -0.00000201, -0.00000348, -0.00001387, 0.00001542, -0.00003651, 0.00000535, -0.00000652, 0.00003569, 0.00000228, 0.00001883, -0.00000216, 0.00000894, 0.00000239, 0.00001793, 0.00000335, 0.00002934, 0.00001461, 0.00000238, -0.00001867, -0.00001823, 0.00000570, -0.00000576, -0.00001052, -0.00000197, -0.00000010, 0.00000123, -0.00000196, -0.00000049, 0.00000387, -0.00000197, -0.00000529, 0.00000030, 0.00000468, -0.00000404, -0.00000001, -0.00000095, 0.00000167, 0.00000241, -0.00000434, 0.00000330, -0.00000572, 0.00000216, -0.00000088, 0.00001065, -0.00000868, 0.00000358, -0.00001127, 0.00000506, -0.00000433, 0.00000301, -0.00001155, 0.00001222, -0.00001634, 0.00000370, -0.00000154, 0.00000893, -0.00001837, 0.00000414, -0.00003776, 0.00001966, -0.00004959, 0.00000508, -0.00016886, 0.00002605, 0.00019347, 0.00006098, 0.00009133, 0.00019388, -0.00001061, -0.00015801, 0.00001743, -0.00006365, 0.00002289, -0.00002494, 0.00000960, -0.00003696, 0.00001265, -0.00001344, 0.00000466, -0.00001611, 0.00000333, -0.00000633, 0.00000362, -0.00001562, 0.00001035, -0.00000529, 0.00000586, -0.00001043, 0.00000210, -0.00000815, 0.00000768, -0.00001339, 0.00000863, -0.00000820, 0.00000912, -0.00000655, -0.00000545, -0.00000369, 0.00000853, -0.00000393, -0.00000319, -0.00000320, 0.00001156, -0.00000700, 0.00000126, -0.00000537, 0.00001554, -0.00001229, 0.00000667, 0.00000382, -0.00000983, 0.00000653, -0.00001418, },
		{ 0.00000254, -0.00000028, -0.00000349, -0.00000358, 0.00000119, -0.00000942, -0.00000145, 0.00000043, -0.00000294, 0.00000162, -0.00000234, -0.00000735, -0.00000129, -0.00000077, -0.00001319, 0.00000787, -0.00002094, 0.00001203, -0.00001418, -0.00001934, 0.00000334, 0.00000885, 0.00002990, -0.00000246, 0.00002053, -0.00000929, 0.00001474, 0.00000110, 0.00001915, 0.00000305, 0.00003605, -0.00000111, 0.00000524, -0.00003059, -0.00000129, -0.00000042, -0.00000583, -0.00001135, 0.00000345, -0.00000359, 0.00000460, -0.00000676, 0.00000565, -0.00000066, -0.00000112, -0.00000675, 0.00000575, -0.00000051, -0.00000246, 0.00000007, -0.00000095, 0.00000372, -0.00000163, -0.00000073, -0.00000020, -0.00000324, 0.00000043, 0.00000534, 0.00000346, -0.00000611, -0.00000012, -0.00000778, 0.00000410, -0.00000247, -0.00000281, -0.00000279, 0.00000457, -0.00001499, 0.00000553, 0.00000321, -0.00000254, -0.00000938, -0.00001264, -0.00001762, 0.00000053, -0.00003942, -0.00003113, -0.00018045, 0.00015921, 0.00014187, 0.00006109, 0.00014886, 0.00016491, -0.00016955, -0.00004292, -0.00002967, -0.00002239, 0.00000437, -0.00000880, -0.00001339, -0.00001739, 0.00000672, -0.00000913, -0.00000275, -0.00000896, -0.00000041, -0.00000014, -0.00000784, -0.00000398, 0.00000566, -0.00000135, -0.00000095, -0.00000522, -0.00000296, -0.00000074, -0.00000081, -0.00000616, 0.00000560, -0.00000595, 0.00000819, -0.00001014, -0.00000360, 0.00000005, 0.00000651, -0.00000562, -0.00000366, 0.00000262, 0.00000676, -0.00000685, 0.00000069, -0.00000076, 0.00001266, -0.00001703, 0.00001915, -0.00001209, 0.00000265, -0.00000434, -0.00000669, },
		{ -0.00000065, 0.00000152, -0.00000720, 0.00000144, -0.00000438, -0.00000794, 0.00000168, -0.00000159, -0.00000169, 0.00000192, -0.00000589, -0.00000604, 0.00000205, -0.00000902, -0.00000274, -0.00000188, -0.00001516, 0.00001706, -0.00003606, 0.00000687, -0.00000870, 0.00003436, 0.00000334, 0.00001450, 0.00000244, 0.00000200, 0.00000736, 0.00001079, 0.00000844, 0.00001794, 0.00002311, -0.00000107, -0.00000457, -0.00002742, 0.00000677, -0.00000464, -0.00000836, -0.00000607, 0.00000182, -0.00000083, 0.00000107, -0.00000443, 0.00000620, -0.00000198, -0.00000318, -0.00000324, 0.00000588, -0.00000319, -0.00000120, 0.00000036, -0.00000083, 0.00000494, -0.00000544, 0.00000384, -0.00000571, 0.00000178, -0.00000220, 0.00001191, -0.00000760, 0.00000201, -0.00000884, 0.00000173, -0.00000282, 0.00000307, -0.00001073, 0.00000996, -0.00001216, -0.00000197, -0.00000025, 0.00001083, -0.00001821, 0.00000449, -0.00003410, 0.00001335, -0.00004106, -0.00000164, -0.00012032, -0.00008971, 0.00022361, 0.00009041, 0.00008545, 0.00022003, -0.00000383, -0.00017219, 0.00001742, -0.00006601, 0.00001968, -0.00002076, 0.00000686, -0.00003524, 0.00001030, -0.00001128, 0.00000223, -0.00001382, 0.00000234, -0.00000698, 0.00000456, -0.00001719, 0.00001064, -0.00000460, 0.00000549, -0.00000964, 0.00000246, -0.00000958, 0.00000751, -0.00001146, 0.00000664, -0.00000502, 0.00000359, -0.00000067, -0.00000672, -0.00000512, 0.00000632, -0.00000003, -0.00000332, -0.00000537, 0.00000969, -0.00000258, -0.00000056, -0.00000501, 0.00001020, -0.00000171, -0.00000552, 0.00001531, -0.00001677, 0.00000897, -0.00001215, 0.00000087, },
		{ -0.00000167, 0.00000051, -0.00000753, 0.00000384, -0.00001040, -0.00000252, 0.00000163, -0.00000321, 0.00000105, -0.00000095, -0.00000748, -0.00000283, 0.00000068, -0.00001264, 0.00000724, -0.00001907, 0.00000992, -0.00001057, -0.00002120, 0.00000708, 0.00000158, 0.00003556, -0.00000909, 0.00002460, -0.00001303, 0.00001530, -0.00000084, 0.00001884, 0.00000046, 0.00003185, 0.00000415, 0.00000609, -0.00002045, -0.00001279, 0.00000504, -0.00000639, -0.00000995, -0.00000034, -0.00000139, 0.00000299, -0.00000442, 0.00000138, 0.00000329, -0.00000208, -0.00000480, 0.00000119, 0.00000341, -0.00000451, 0.00000122, -0.00000173, 0.00000273, 0.00000084, -0.00000289, 0.00000184, -0.00000497, 0.00000138, 0.00000250, 0.00000793, -0.00000932, 0.00000275, -0.00000987, 0.00000437, -0.00000361, 0.00000059, -0.00000690, 0.00000855, -0.00001736, 0.00000435, 0.00000183, 0.00000399, -0.00001669, -0.00000153, -0.00002870, 0.00001243, -0.00005349, 0.00000188, -0.00018913, 0.00005436, 0.00021975, 0.00005001, 0.00016917, 0.00018188, -0.00017117, -0.00005886, -0.00002105, -0.00003403, 0.00001149, -0.00001268, -0.00000987, -0.00002163, 0.00001015, -0.00001207, -0.00000089, -0.00000979, 0.00000058, -0.00000243, -0.00000494, -0.00000845, 0.00000978, -0.00000435, 0.00000209, -0.00000745, 0.00000006, -0.00000583, 0.00000411, -0.00000970, 0.00000919, -0.00000879, 0.00000795, -0.00000664, -0.00000465, -0.00000316, 0.00000770, -0.00000342, -0.00000356, -0.00000192, 0.00000966, -0.00000638, 0.00000184, -0.00000574, 0.00001565, -0.00001438, 0.00001145, -0.00000010, -0.00000925, 0.00000561, -0.00001252, 0.00000427, },
		{ -0.00000044, -0.00000310, -0.00000394, 0.00000118, -0.00001274, 0.00000409, -0.00000261, -0.00000069, 0.00000051, -0.00000309, -0.00000856, 0.00000263, -0.00000811, -0.00000270, -0.00000154, -0.00001468, 0.00001685, -0.00003411, 0.00000686, -0.00000860, 0.00002944, 0.00000905, 0.00000600, 0.00001029, -0.00000702, 0.00001369, 0.00000295, 0.00001435, 0.00000718, 0.00002922, -0.00000359, 0.00000526, -0.00002819, 0.00000116, -0.00000041, -0.00000781, -0.00000775, 0.00000117, -0.00000101, 0.00000241, -0.00000588, 0.00000533, 0.00000030, -0.00000269, -0.00000410, 0.00000383, 0.00000009, -0.00000390, 0.00000263, -0.00000347, 0.00000643, -0.00000492, 0.00000278, -0.00000426, 0.00000022, -0.00000197, 0.00001052, -0.00000357, -0.00000212, -0.00000427, -0.00000287, -0.00000035, 0.00000112, -0.00000696, 0.00000463, -0.00000540, -0.00000875, 0.00000184, 0.00000947, -0.00001250, -0.00000261, -0.00002162, -0.00000205, -0.00002189, -0.00002532, -0.00005678, -0.00017769, 0.00018920, 0.00016102, 0.00006132, 0.00026016, -0.00000334, -0.00018017, 0.00000988, -0.00006126, 0.00000988, -0.00001212, 0.00000129, -0.00003102, 0.00000541, -0.00000661, -0.00000228, -0.00001033, 0.00000057, -0.00000628, 0.00000357, -0.00001620, 0.00000779, -0.00000097, 0.00000259, -0.00000666, 0.00000063, -0.00000783, 0.00000345, -0.00000613, 0.00000170, 0.00000081, -0.00000313, 0.00000421, -0.00000675, -0.00000561, 0.00000236, 0.00000397, -0.00000304, -0.00000592, 0.00000536, 0.00000278, -0.00000284, -0.00000252, 0.00000229, 0.00000878, -0.00001395, 0.00001956, -0.00001449, 0.00000265, -0.00000362, -0.00000668, 0.00000329, },
		{ 0.00000188, -0.00000776, 0.00000225, -0.00000714, -0.00000692, 0.00000476, -0.00000473, 0.00000188, -0.00000143, -0.00000553, -0.00000577, 0.00000267, -0.00001258, 0.00000732, -0.00001821, 0.00000942, -0.00000893, -0.00002139, 0.00000952, -0.00000412, 0.00003828, -0.00001073, 0.00002269, -0.00001037, 0.00000975, 0.00000316, 0.00001291, 0.00000440, 0.00002144, 0.00001382, 0.00000065, -0.00000595, -0.00002302, 0.00000747, -0.00000573, -0.00000795, -0.00000442, 0.00000069, 0.00000095, -0.00000106, -0.00000337, 0.00000642, -0.00000249, -0.00000249, -0.00000290, 0.00000550, -0.00000447, 0.00000051, -0.00000070, 0.00000022, 0.00000378, -0.00000450, 0.00000285, -0.00000549, 0.00000140, 0.00000063, 0.00001025, -0.00000936, 0.00000204, -0.00000813, 0.00000145, -0.00000235, 0.00000107, -0.00000691, 0.00000748, -0.00001422, -0.00000104, 0.00000305, 0.00000678, -0.00001806, 0.00000083, -0.00002771, 0.00000965, -0.00004884, 0.00000123, -0.00014254, -0.00008039, 0.00027170, 0.00007111, 0.00017008, 0.00021658, -0.00018311, -0.00006738, -0.00002026, -0.00003773, 0.00001014, -0.00000933, -0.00001177, -0.00002137, 0.00000940, -0.00001091, -0.00000286, -0.00000781, -0.00000008, -0.00000335, -0.00000343, -0.00001115, 0.00001133, -0.00000454, 0.00000247, -0.00000743, 0.00000127, -0.00000836, 0.00000518, -0.00000901, 0.00000833, -0.00000655, 0.00000306, -0.00000080, -0.00000625, -0.00000460, 0.00000560, 0.00000087, -0.00000404, -0.00000447, 0.00000876, -0.00000272, 0.00000054, -0.00000636, 0.00001165, -0.00000459, -0.00000138, 0.00001330, -0.00001846, 0.00001017, -0.00001239, 0.00000160, -0.00000043, },
		{ 0.00000175, -0.00000908, 0.00000536, -0.00001507, 0.00000379, -0.00000182, -0.00000062, -0.00000024, -0.00000128, -0.00000950, 0.00000184, -0.00000674, -0.00000257, -0.00000160, -0.00001366, 0.00001641, -0.00003234, 0.00000677, -0.00000751, 0.00002270, 0.00001632, -0.00000234, 0.00001648, -0.00001303, 0.00001642, -0.00000078, 0.00001610, 0.00000221, 0.00002825, -0.00000002, 0.00000604, -0.00001891, -0.00001002, 0.00000561, -0.00000866, -0.00000670, -0.00000215, 0.00000080, 0.00000152, -0.00000420, 0.00000118, 0.00000447, -0.00000371, -0.00000239, -0.00000067, 0.00000442, -0.00000664, 0.00000445, -0.00000496, 0.00000606, -0.00000269, 0.00000066, -0.00000221, -0.00000165, -0.00000103, 0.00000777, 0.00000134, -0.00000617, -0.00000051, -0.00000586, 0.00000062, -0.00000039, -0.00000327, 0.00000001, -0.00000037, -0.00001254, 0.00000153, 0.00000794, -0.00000552, -0.00001090, -0.00000899, -0.00001545, -0.00000610, -0.00004498, -0.00001159, -0.00020379, 0.00008132, 0.00025739, 0.00003625, 0.00030164, 0.00000114, -0.00018991, 0.00000168, -0.00005608, 0.00000027, -0.00000552, -0.00000170, -0.00002861, 0.00000166, -0.00000289, -0.00000560, -0.00000846, 0.00000002, -0.00000567, 0.00000192, -0.00001401, 0.00000352, 0.00000345, -0.00000064, -0.00000351, -0.00000166, -0.00000475, -0.00000195, -0.00000061, -0.00000265, 0.00000544, -0.00000717, 0.00000521, -0.00000403, -0.00000646, -0.00000117, 0.00000565, -0.00000078, -0.00000633, 0.00000101, 0.00000623, -0.00000307, -0.00000086, -0.00000371, 0.00001376, -0.00001382, 0.00001374, -0.00000305, -0.00000972, 0.00000731, -0.00001390, 0.00000641, -0.00000273, },
		{ -0.00000393, -0.00000314, -0.00000093, -0.00001279, 0.00000809, -0.00000638, 0.00000342, -0.00000310, -0.00000247, -0.00000903, 0.00000451, -0.00001283, 0.00000824, -0.00001875, 0.00001085, -0.00000929, -0.00002001, 0.00001048, -0.00000737, 0.00003683, -0.00000681, 0.00001541, -0.00000303, 0.00000060, 0.00001009, 0.00000459, 0.00001102, 0.00000957, 0.00002281, -0.00000464, 0.00000493, -0.00002548, 0.00000337, -0.00000196, -0.00000729, -0.00000653, 0.00000066, 0.00000023, 0.00000122, -0.00000609, 0.00000655, -0.00000078, -0.00000149, -0.00000462, 0.00000440, -0.00000181, -0.00000197, 0.00000155, -0.00000267, 0.00000576, -0.00000431, 0.00000198, -0.00000421, -0.00000015, 0.00000071, 0.00000937, -0.00000562, -0.00000225, -0.00000313, -0.00000383, 0.00000084, -0.00000134, -0.00000293, 0.00000204, -0.00000698, -0.00000902, 0.00000632, 0.00000509, -0.00001231, -0.00000640, -0.00001499, -0.00000610, -0.00002843, -0.00002438, -0.00006858, -0.00019655, 0.00025650, 0.00014129, 0.00014635, 0.00027208, -0.00020700, -0.00006689, -0.00002895, -0.00003235, 0.00000019, 0.00000035, -0.00001803, -0.00001730, 0.00000484, -0.00000599, -0.00000808, -0.00000369, -0.00000222, -0.00000242, -0.00000428, -0.00001083, 0.00000912, -0.00000112, -0.00000036, -0.00000458, -0.00000031, -0.00000706, 0.00000151, -0.00000390, 0.00000345, -0.00000048, -0.00000428, 0.00000516, -0.00000734, -0.00000457, 0.00000112, 0.00000572, -0.00000427, -0.00000528, 0.00000494, 0.00000261, -0.00000193, -0.00000400, 0.00000372, 0.00000694, -0.00001224, 0.00002062, -0.00001881, 0.00000565, -0.00000501, -0.00000552, 0.00000357, 0.00000006, },
		{ -0.00001074, 0.00000529, -0.00001289, -0.00000024, 0.00000124, -0.00000201, 0.00000055, -0.00000148, -0.00000815, -0.00000063, -0.00000462, -0.00000298, -0.00000101, -0.00001361, 0.00001753, -0.00003246, 0.00000818, -0.00000717, 0.00001678, 0.00002133, -0.00000616, 0.00001686, -0.00001252, 0.00001266, 0.00000196, 0.00001145, 0.00000487, 0.00001927, 0.00001008, -0.00000102, -0.00000393, -0.00002135, 0.00000957, -0.00000881, -0.00000449, -0.00000643, 0.00000316, -0.00000055, -0.00000054, -0.00000443, 0.00000857, -0.00000477, 0.00000039, -0.00000553, 0.00000767, -0.00000776, 0.00000468, -0.00000472, 0.00000405, 0.00000019, -0.00000099, -0.00000116, -0.00000229, -0.00000105, 0.00000589, 0.00000420, -0.00000698, -0.00000061, -0.00000454, -0.00000221, 0.00000101, -0.00000286, -0.00000017, -0.00000089, -0.00001002, -0.00000390, 0.00000968, -0.00000263, -0.00001303, -0.00000544, -0.00001613, -0.00000620, -0.00004398, -0.00000499, -0.00016106, -0.00007514, 0.00033892, 0.00004002, 0.00032583, 0.00002385, -0.00021247, 0.00000231, -0.00005934, -0.00000093, -0.00000801, 0.00000328, -0.00003221, 0.00000282, -0.00000373, -0.00000435, -0.00001096, 0.00000276, -0.00000703, 0.00000176, -0.00001313, 0.00000082, 0.00000543, -0.00000113, -0.00000296, -0.00000178, -0.00000331, -0.00000502, 0.00000117, -0.00000273, 0.00000544, -0.00000574, 0.00000086, 0.00000187, -0.00000852, -0.00000244, 0.00000346, 0.00000412, -0.00000754, -0.00000146, 0.00000591, 0.00000011, -0.00000183, -0.00000519, 0.00001115, -0.00000512, 0.00000067, 0.00001208, -0.00002135, 0.00001429, -0.00001602, 0.00000552, -0.00000264, 0.00000428, },
		{ -0.00000930, 0.00000479, -0.00001682, 0.00000906, -0.00000645, 0.00000429, -0.00000442, -0.00000007, -0.00001099, 0.00000511, -0.00001271, 0.00000942, -0.00001998, 0.00001314, -0.00001017, -0.00001860, 0.00001122, -0.00000918, 0.00003265, -0.00000009, 0.00000689, 0.00000407, -0.00000689, 0.00001464, -0.00000082, 0.00001462, 0.00000246, 0.00002503, -0.00000330, 0.00000625, -0.00001738, -0.00000715, 0.00000430, -0.00000828, -0.00000584, -0.00000216, 0.00000158, 0.00000123, -0.00000541, 0.00000293, 0.00000323, -0.00000222, -0.00000349, 0.00000034, 0.00000252, -0.00000502, 0.00000374, -0.00000464, 0.00000581, -0.00000212, -0.00000033, -0.00000191, -0.00000247, 0.00000199, 0.00000644, -0.00000009, -0.00000744, 0.00000201, -0.00000833, 0.00000319, -0.00000399, 0.00000185, -0.00000379, -0.00000029, -0.00001503, 0.00000802, 0.00000223, -0.00000378, -0.00001692, 0.00000059, -0.00002319, -0.00000737, -0.00005152, -0.00000692, -0.00024871, 0.00015352, 0.00025202, 0.00011231, 0.00033761, -0.00023366, -0.00006485, -0.00004065, -0.00002426, -0.00001173, 0.00000983, -0.00002310, -0.00001382, 0.00000024, -0.00000087, -0.00001310, -0.00000040, -0.00000372, -0.00000105, -0.00000634, -0.00000860, 0.00000462, 0.00000401, -0.00000437, -0.00000075, -0.00000314, -0.00000352, -0.00000458, 0.00000253, -0.00000195, 0.00000546, -0.00000991, 0.00000787, -0.00000592, -0.00000470, -0.00000321, 0.00000840, -0.00000244, -0.00000585, 0.00000060, 0.00000670, -0.00000292, -0.00000177, -0.00000334, 0.00001404, -0.00001502, 0.00001741, -0.00000873, -0.00000657, 0.00000649, -0.00001380, 0.00000788, -0.00000300, 0.00000160, },
		{ 0.00000067, -0.00000685, -0.00000658, 0.00000513, -0.00000402, 0.00000245, -0.00000328, -0.00000521, -0.00000390, -0.00000229, -0.00000392, 0.00000063, -0.00001524, 0.00002094, -0.00003497, 0.00001140, -0.00000824, 0.00001315, 0.00002203, -0.00000388, 0.00001095, -0.00000610, 0.00000389, 0.00000916, 0.00000270, 0.00001217, 0.00000650, 0.00002168, -0.00000930, 0.00000855, -0.00002591, 0.00000745, -0.00000607, -0.00000321, -0.00000935, 0.00000402, -0.00000201, 0.00000290, -0.00000877, 0.00001017, -0.00000414, 0.00000243, -0.00000864, 0.00000809, -0.00000637, 0.00000313, -0.00000318, 0.00000154, 0.00000208, -0.00000062, -0.00000238, -0.00000066, -0.00000313, 0.00000646, 0.00000322, -0.00000309, -0.00000544, 0.00000136, -0.00000877, 0.00000561, -0.00000655, 0.00000497, -0.00000740, -0.00000150, -0.00001387, 0.00001521, -0.00000591, -0.00000601, -0.00001408, -0.00000161, -0.00002387, -0.00002112, -0.00003317, -0.00007478, -0.00022768, 0.00035494, 0.00009836, 0.00032139, 0.00007274, -0.00025355, 0.00001661, -0.00007542, 0.00000987, -0.00002186, 0.00001713, -0.00004226, 0.00000957, -0.00001000, 0.00000222, -0.00001821, 0.00000896, -0.00001090, 0.00000425, -0.00001523, 0.00000176, 0.00000294, 0.00000278, -0.00000642, 0.00000166, -0.00000529, -0.00000360, -0.00000264, 0.00000264, 0.00000027, 0.00000082, -0.00000739, 0.00000924, -0.00001106, -0.00000145, -0.00000197, 0.00001035, -0.00000888, -0.00000196, 0.00000215, 0.00000573, -0.00000472, -0.00000279, 0.00000317, 0.00000738, -0.00001263, 0.00002284, -0.00002504, 0.00001239, -0.00001056, -0.00000042, 0.00000078, 0.00000315, -0.00000780, },
		{ 0.00000658, -0.00001633, 0.00000611, -0.00000408, 0.00000359, -0.00000418, 0.00000034, -0.00001058, 0.00000389, -0.00001174, 0.00001010, -0.00002073, 0.00001459, -0.00000953, -0.00001918, 0.00001354, -0.00001148, 0.00002859, 0.00000522, 0.00000212, 0.00000590, -0.00000813, 0.00001274, 0.00000038, 0.00001149, 0.00000365, 0.00001788, 0.00000650, -0.00000189, -0.00000207, -0.00001937, 0.00000958, -0.00000920, -0.00000344, -0.00000659, 0.00000408, -0.00000061, -0.00000186, -0.00000313, 0.00000798, -0.00000364, -0.00000053, -0.00000499, 0.00000659, -0.00000702, 0.00000471, -0.00000503, 0.00000414, 0.00000093, -0.00000235, -0.00000049, -0.00000364, 0.00000243, 0.00000412, 0.00000371, -0.00000959, 0.00000332, -0.00000834, 0.00000142, -0.00000345, 0.00000323, -0.00000513, 0.00000069, -0.00001429, 0.00000402, 0.00000299, 0.00000069, -0.00002154, 0.00000742, -0.00002771, -0.00000240, -0.00005743, 0.00001259, -0.00021811, -0.00002024, 0.00036330, 0.00010053, 0.00039202, -0.00024649, -0.00007472, -0.00004373, -0.00002422, -0.00001558, 0.00001029, -0.00002002, -0.00001651, 0.00000058, -0.00000029, -0.00001351, -0.00000168, -0.00000169, -0.00000179, -0.00000697, -0.00000742, 0.00000130, 0.00000710, -0.00000597, 0.00000075, -0.00000411, -0.00000118, -0.00000887, 0.00000568, -0.00000337, 0.00000693, -0.00001002, 0.00000483, -0.00000069, -0.00000655, -0.00000495, 0.00000674, 0.00000257, -0.00000746, -0.00000201, 0.00000715, -0.00000047, -0.00000215, -0.00000598, 0.00001327, -0.00000810, 0.00000498, 0.00000738, -0.00002024, 0.00001577, -0.00001821, 0.00000894, -0.00000425, 0.00000539, -0.00001251, },
		{ -0.00000023, -0.00001190, 0.00000720, -0.00000484, 0.00000405, -0.00000514, -0.00000246, -0.00000629, -0.00000088, -0.00000476, 0.00000287, -0.00001811, 0.00002585, -0.00003860, 0.00001510, -0.00000971, 0.00001122, 0.00001929, 0.00000234, 0.00000239, 0.00000160, -0.00000479, 0.00001551, -0.00000467, 0.00001804, -0.00000311, 0.00002757, -0.00001110, 0.00001144, -0.00001972, -0.00000168, -0.00000036, -0.00000367, -0.00000959, 0.00000222, -0.00000163, 0.00000436, -0.00000980, 0.00000789, -0.00000109, 0.00000278, -0.00000889, 0.00000533, -0.00000293, 0.00000068, -0.00000148, -0.00000013, 0.00000187, 0.00000223, -0.00000565, 0.00000266, -0.00000669, 0.00000900, -0.00000081, 0.00000384, -0.00001259, 0.00000889, -0.00001595, 0.00001064, -0.00001176, 0.00001240, -0.00001607, 0.00000841, -0.00002374, 0.00002089, -0.00001228, 0.00000624, -0.00002924, 0.00001984, -0.00004789, 0.00000861, -0.00007158, 0.00000992, -0.00031882, 0.00027075, 0.00021215, 0.00029788, 0.00014042, -0.00030665, 0.00003899, -0.00009933, 0.00002756, -0.00004142, 0.00003398, -0.00005373, 0.00001793, -0.00001834, 0.00001095, -0.00002706, 0.00001583, -0.00001535, 0.00000821, -0.00001955, 0.00000568, -0.00000305, 0.00000969, -0.00001229, 0.00000723, -0.00000965, 0.00000128, -0.00001027, 0.00001082, -0.00000690, 0.00000873, -0.00001534, 0.00001447, -0.00001180, -0.00000003, -0.00000794, 0.00001492, -0.00000839, -0.00000194, -0.00000279, 0.00001096, -0.00000695, 0.00000041, -0.00000528, 0.00001696, -0.00001892, 0.00002322, -0.00001755, 0.00000144, 0.00000055, -0.00000896, 0.00000566, -0.00000016, -0.00000217, -0.00000582, },
		{ -0.00001123, 0.00000008, -0.00000029, 0.00000184, -0.00000246, -0.00000121, -0.00000819, 0.00000148, -0.00001025, 0.00001008, -0.00002030, 0.00001409, -0.00000614, -0.00002271, 0.00001826, -0.00001542, 0.00002673, 0.00000620, 0.00000380, 0.00000076, -0.00000249, 0.00000458, 0.00000745, 0.00000264, 0.00001143, 0.00000448, 0.00002035, -0.00001316, 0.00001234, -0.00002621, 0.00000956, -0.00000770, -0.00000136, -0.00001042, 0.00000586, -0.00000270, 0.00000255, -0.00000896, 0.00001100, -0.00000402, 0.00000246, -0.00000940, 0.00000849, -0.00000689, 0.00000410, -0.00000422, 0.00000196, 0.00000294, -0.00000208, -0.00000174, -0.00000206, 0.00000029, 0.00000475, 0.00000306, -0.00000619, -0.00000124, -0.00000239, -0.00000556, 0.00000168, -0.00000082, 0.00000029, -0.00000606, -0.00000551, -0.00000668, 0.00000951, -0.00000290, -0.00001486, -0.00000071, -0.00001338, -0.00001988, -0.00003487, -0.00001377, -0.00012408, -0.00021085, 0.00041568, 0.00014604, 0.00041751, -0.00023245, -0.00010673, -0.00002951, -0.00003996, -0.00000472, -0.00000327, -0.00000555, -0.00002782, 0.00000832, -0.00000675, -0.00000703, -0.00000925, 0.00000516, -0.00000612, -0.00000413, -0.00000986, 0.00000215, 0.00000511, -0.00000246, -0.00000244, -0.00000088, -0.00000280, -0.00000812, 0.00000251, 0.00000156, 0.00000220, -0.00000379, -0.00000354, 0.00000742, -0.00000998, -0.00000352, 0.00000083, 0.00000983, -0.00000974, -0.00000232, 0.00000365, 0.00000519, -0.00000520, -0.00000391, 0.00000573, 0.00000473, -0.00001018, 0.00002132, -0.00002746, 0.00001690, -0.00001516, 0.00000472, -0.00000187, 0.00000534, -0.00001086, 0.00000329, },
		{ -0.00001295, 0.00000537, -0.00000322, 0.00000415, -0.00000556, -0.00000156, -0.00000642, -0.00000122, -0.00000486, 0.00000491, -0.00002099, 0.00003045, -0.00004114, 0.00001705, -0.00000961, 0.00000896, 0.00001605, 0.00000807, -0.00000337, 0.00000488, -0.00000790, 0.00001590, -0.00000581, 0.00001736, -0.00000433, 0.00002327, -0.00000292, 0.00000337, -0.00000485, -0.00001406, 0.00000545, -0.00000463, -0.00000766, -0.00000169, 0.00000035, 0.00000353, -0.00000727, 0.00000231, 0.00000340, 0.00000192, -0.00000674, 0.00000051, 0.00000099, -0.00000126, -0.00000068, -0.00000032, -0.00000027, 0.00000635, -0.00000906, 0.00000552, -0.00000954, 0.00001118, -0.00000486, 0.00000983, -0.00001746, 0.00001322, -0.00001909, 0.00001184, -0.00001412, 0.00001697, -0.00002092, 0.00001322, -0.00002727, 0.00002115, -0.00001557, 0.00001550, -0.00003998, 0.00003430, -0.00006123, 0.00002412, -0.00009087, 0.00005213, -0.00031488, 0.00009145, 0.00034493, 0.00028916, 0.00020434, -0.00035406, 0.00005467, -0.00011811, 0.00004006, -0.00005527, 0.00004363, -0.00005832, 0.00002116, -0.00002276, 0.00001614, -0.00003216, 0.00001877, -0.00001676, 0.00001060, -0.00002307, 0.00000934, -0.00000880, 0.00001549, -0.00001660, 0.00001127, -0.00001293, 0.00000587, -0.00001714, 0.00001671, -0.00001094, 0.00001289, -0.00001828, 0.00001408, -0.00000855, -0.00000045, -0.00001134, 0.00001500, -0.00000440, -0.00000301, -0.00000630, 0.00001292, -0.00000594, 0.00000130, -0.00000978, 0.00001885, -0.00001487, 0.00001277, -0.00000183, -0.00001319, 0.00001137, -0.00001509, 0.00000830, -0.00000243, 0.00000253, -0.00001016, 0.00000305, },
		{ -0.00000584, 0.00000233, 0.00000080, -0.00000058, -0.00000329, -0.00000549, -0.00000049, -0.00000934, 0.00000989, -0.00001890, 0.00001174, -0.00000042, -0.00002833, 0.00002437, -0.00002036, 0.00002700, 0.00000305, 0.00001055, -0.00000849, 0.00000609, -0.00000522, 0.00001545, -0.00000675, 0.00001978, -0.00000791, 0.00003028, -0.00001888, 0.00001771, -0.00002255, 0.00000230, -0.00000301, -0.00000086, -0.00001195, 0.00000538, -0.00000347, 0.00000563, -0.00001196, 0.00001035, -0.00000218, 0.00000413, -0.00001133, 0.00000732, -0.00000463, 0.00000251, -0.00000320, 0.00000069, 0.00000252, 0.00000130, -0.00000583, 0.00000212, -0.00000429, 0.00000838, -0.00000188, 0.00000182, -0.00000990, 0.00000709, -0.00001507, 0.00000905, -0.00000823, 0.00001008, -0.00001736, 0.00000729, -0.00001992, 0.00001875, -0.00001233, 0.00000045, -0.00001964, 0.00001327, -0.00005039, 0.00000294, -0.00006239, -0.00001547, -0.00034877, 0.00035821, 0.00026176, 0.00041721, -0.00019449, -0.00015827, -0.00000041, -0.00006914, 0.00001802, -0.00002707, 0.00001596, -0.00004391, 0.00002052, -0.00001787, 0.00000407, -0.00002071, 0.00001468, -0.00001268, 0.00000163, -0.00001584, 0.00000733, -0.00000184, 0.00000553, -0.00000951, 0.00000586, -0.00000811, -0.00000253, -0.00000607, 0.00001102, -0.00000629, 0.00000558, -0.00001339, 0.00001490, -0.00001270, -0.00000070, -0.00000667, 0.00001633, -0.00001071, -0.00000163, -0.00000182, 0.00001148, -0.00000858, 0.00000011, -0.00000389, 0.00001659, -0.00001999, 0.00002570, -0.00002335, 0.00000813, -0.00000527, -0.00000334, 0.00000300, 0.00000223, -0.00000558, -0.00000294, -0.00000443, },
		{ -0.00000015, 0.00000032, 0.00000288, -0.00000425, -0.00000286, -0.00000428, -0.00000300, -0.00000432, 0.00000633, -0.00002290, 0.00003322, -0.00004086, 0.00001572, -0.00000660, 0.00000472, 0.00001498, 0.00000957, -0.00000256, 0.00000076, -0.00000340, 0.00000909, 0.00000006, 0.00000958, 0.00000276, 0.00001039, 0.00001200, -0.00001008, 0.00001109, -0.00002272, 0.00000699, -0.00000380, -0.00000537, -0.00000582, 0.00000241, 0.00000155, -0.00000275, -0.00000418, 0.00000705, 0.00000137, -0.00000369, -0.00000434, 0.00000353, -0.00000156, -0.00000117, 0.00000058, -0.00000298, 0.00000950, -0.00001007, 0.00000548, -0.00000933, 0.00001047, -0.00000571, 0.00001107, -0.00001625, 0.00001089, -0.00001522, 0.00000665, -0.00001078, 0.00001509, -0.00001789, 0.00000895, -0.00002109, 0.00001293, -0.00001148, 0.00001525, -0.00003785, 0.00003173, -0.00005280, 0.00001309, -0.00007587, 0.00003754, -0.00022508, -0.00013122, 0.00042984, 0.00034003, 0.00023986, -0.00037778, 0.00004896, -0.00011910, 0.00003607, -0.00005347, 0.00003805, -0.00005012, 0.00001452, -0.00001878, 0.00001343, -0.00002951, 0.00001449, -0.00001248, 0.00000874, -0.00002272, 0.00000917, -0.00001026, 0.00001604, -0.00001555, 0.00001034, -0.00001179, 0.00000638, -0.00001888, 0.00001601, -0.00000814, 0.00001019, -0.00001402, 0.00000724, -0.00000122, -0.00000337, -0.00001079, 0.00000991, 0.00000276, -0.00000531, -0.00000725, 0.00001058, -0.00000115, -0.00000105, -0.00000894, 0.00001282, -0.00000291, -0.00000297, 0.00001423, -0.00002331, 0.00001524, -0.00001433, 0.00000580, -0.00000112, 0.00000346, -0.00000963, 0.00000100, -0.00000888, },
		{ 0.00000125, 0.00000206, -0.00000001, -0.00000418, -0.00000441, -0.00000038, -0.00001004, 0.00001026, -0.00001732, 0.00000867, 0.00000588, -0.00003379, 0.00002954, -0.00002425, 0.00002742, -0.00000145, 0.00001797, -0.00001623, 0.00001148, -0.00001057, 0.00001851, -0.00001084, 0.00002244, -0.00001256, 0.00002997, -0.00001370, 0.00001101, -0.00000895, -0.00000955, 0.00000272, -0.00000130, -0.00001105, 0.00000251, -0.00000248, 0.00000627, -0.00001103, 0.00000583, 0.00000157, 0.00000433, -0.00001052, 0.00000356, -0.00000132, 0.00000103, -0.00000291, 0.00000092, -0.00000020, 0.00000658, -0.00001079, 0.00000656, -0.00000893, 0.00001250, -0.00000790, 0.00001015, -0.00001762, 0.00001466, -0.00002168, 0.00001360, -0.00001385, 0.00001830, -0.00002632, 0.00001652, -0.00002823, 0.00002386, -0.00002030, 0.00001498, -0.00003706, 0.00003642, -0.00007421, 0.00003104, -0.00009745, 0.00005473, -0.00038211, 0.00017934, 0.00041864, 0.00042456, -0.00015572, -0.00021154, 0.00002857, -0.00009843, 0.00004008, -0.00004906, 0.00003349, -0.00005566, 0.00002973, -0.00002700, 0.00001348, -0.00003012, 0.00002167, -0.00001737, 0.00000691, -0.00002218, 0.00001352, -0.00000994, 0.00001383, -0.00001634, 0.00001230, -0.00001357, 0.00000411, -0.00001539, 0.00001968, -0.00001300, 0.00001245, -0.00001928, 0.00001743, -0.00001175, 0.00000060, -0.00001203, 0.00001856, -0.00000817, -0.00000189, -0.00000638, 0.00001513, -0.00000923, 0.00000244, -0.00001042, 0.00002147, -0.00001939, 0.00001814, -0.00000915, -0.00000646, 0.00000631, -0.00001058, 0.00000680, -0.00000079, -0.00000028, -0.00000799, -0.00000017, -0.00000799, },
		{ 0.00000333, 0.00000183, -0.00000233, -0.00000517, -0.00000139, -0.00000469, -0.00000415, 0.00000748, -0.00002375, 0.00003383, -0.00003765, 0.00001145, -0.00000119, -0.00000137, 0.00001667, 0.00000615, 0.00000448, -0.00000898, 0.00000573, -0.00000119, 0.00000879, -0.00000085, 0.00001278, -0.00000411, 0.00002545, -0.00001965, 0.00001915, -0.00002160, 0.00000145, 0.00000013, -0.00000426, -0.00000824, 0.00000281, 0.00000013, 0.00000136, -0.00000865, 0.00000756, 0.00000259, -0.00000134, -0.00000734, 0.00000338, 0.00000012, -0.00000256, 0.00000161, -0.00000468, 0.00000988, -0.00000735, 0.00000178, -0.00000560, 0.00000632, -0.00000251, 0.00000675, -0.00000885, 0.00000258, -0.00000576, -0.00000317, -0.00000316, 0.00000779, -0.00000821, -0.00000252, -0.00000795, -0.00000087, -0.00000174, 0.00000604, -0.00002334, 0.00001363, -0.00002602, -0.00001863, -0.00003576, -0.00001402, -0.00009916, -0.00031650, 0.00039743, 0.00047900, 0.00024123, -0.00037511, 0.00001967, -0.00010057, 0.00001430, -0.00003566, 0.00001824, -0.00003122, -0.00000005, -0.00000761, 0.00000363, -0.00001994, 0.00000410, -0.00000360, 0.00000307, -0.00001810, 0.00000424, -0.00000619, 0.00001035, -0.00000865, 0.00000419, -0.00000588, 0.00000204, -0.00001454, 0.00000843, 0.00000084, 0.00000202, -0.00000481, -0.00000296, 0.00000714, -0.00000688, -0.00000773, 0.00000199, 0.00001025, -0.00000700, -0.00000665, 0.00000543, 0.00000534, -0.00000477, -0.00000495, 0.00000296, 0.00001029, -0.00001550, 0.00002214, -0.00002256, 0.00000884, -0.00000591, -0.00000150, 0.00000356, 0.00000060, -0.00000467, -0.00000498, -0.00000203, -0.00000650, },
		{ 0.00000558, -0.00000101, -0.00000318, -0.00000571, 0.00000220, -0.00001235, 0.00001128, -0.00001614, 0.00000609, 0.00001092, -0.00003691, 0.00003173, -0.00002525, 0.00002587, -0.00000405, 0.00002159, -0.00001762, 0.00000939, -0.00000801, 0.00001387, -0.00000727, 0.00001707, -0.00000759, 0.00001909, 0.00000077, -0.00000347, 0.00000788, -0.00001956, 0.00000535, -0.00000064, -0.00000909, -0.00000138, -0.00000066, 0.00000495, -0.00000704, -0.00000075, 0.00000541, 0.00000408, -0.00000796, -0.00000123, 0.00000145, 0.00000065, -0.00000357, 0.00000213, -0.00000361, 0.00001108, -0.00001345, 0.00000812, -0.00001047, 0.00001368, -0.00001079, 0.00001379, -0.00001915, 0.00001522, -0.00002068, 0.00001100, -0.00001306, 0.00001949, -0.00002679, 0.00001583, -0.00002569, 0.00001922, -0.00001981, 0.00001921, -0.00004081, 0.00004145, -0.00007446, 0.00002968, -0.00009375, 0.00005804, -0.00030466, -0.00007674, 0.00054491, 0.00049456, -0.00014880, -0.00024305, 0.00003815, -0.00011111, 0.00004645, -0.00005582, 0.00003578, -0.00005428, 0.00002867, -0.00002745, 0.00001479, -0.00003155, 0.00002107, -0.00001599, 0.00000776, -0.00002466, 0.00001608, -0.00001405, 0.00001716, -0.00001801, 0.00001393, -0.00001478, 0.00000698, -0.00001994, 0.00002198, -0.00001287, 0.00001227, -0.00001755, 0.00001276, -0.00000582, -0.00000131, -0.00001288, 0.00001488, -0.00000164, -0.00000388, -0.00000824, 0.00001428, -0.00000571, 0.00000118, -0.00001126, 0.00001758, -0.00000915, 0.00000282, 0.00000806, -0.00001883, 0.00001254, -0.00001191, 0.00000599, -0.00000056, 0.00000161, -0.00000843, -0.00000122, -0.00000471, -0.00000396, },
		{ 0.00000283, -0.00000141, -0.00000663, 0.00000019, -0.00000450, -0.00000549, 0.00000912, -0.00002422, 0.00003320, -0.00003302, 0.00000630, 0.00000442, -0.00000737, 0.00001936, 0.00000028, 0.00001368, -0.00001882, 0.00001326, -0.00000854, 0.00001406, -0.00000748, 0.00001857, -0.00001224, 0.00002945, -0.00001816, 0.00001453, -0.00000964, -0.00000981, 0.00000587, -0.00000435, -0.00000820, 0.00000080, 0.00000039, 0.00000321, -0.00000914, 0.00000390, 0.00000595, -0.00000051, -0.00000752, 0.00000031, 0.00000326, -0.00000411, 0.00000203, -0.00000478, 0.00000755, -0.00000199, -0.00000365, -0.00000068, 0.00000114, 0.00000225, 0.00000005, 0.00000044, -0.00000648, 0.00000351, -0.00001171, 0.00000313, 0.00000072, 0.00000173, -0.00001373, 0.00000382, -0.00001185, 0.00000593, -0.00000407, -0.00000654, -0.00000689, 0.00000211, -0.00004941, 0.00000145, -0.00006097, -0.00000281, -0.00039072, 0.00021313, 0.00069034, 0.00023823, -0.00036849, -0.00001612, -0.00007685, -0.00001241, -0.00001404, -0.00000381, -0.00001279, -0.00001318, 0.00000302, -0.00000643, -0.00000991, -0.00000620, 0.00000443, -0.00000211, -0.00001269, -0.00000235, 0.00000029, 0.00000197, 0.00000011, -0.00000321, 0.00000115, -0.00000384, -0.00000764, -0.00000159, 0.00001078, -0.00000604, 0.00000348, -0.00001063, 0.00001163, -0.00000740, -0.00000555, -0.00000453, 0.00001400, -0.00000539, -0.00000672, 0.00000054, 0.00001000, -0.00000648, -0.00000178, -0.00000498, 0.00001768, -0.00001814, 0.00001760, -0.00001029, -0.00000527, 0.00000610, -0.00000962, 0.00000845, -0.00000323, 0.00000139, -0.00001088, 0.00000328, -0.00000976, 0.00000301, },
		{ -0.00000225, -0.00000116, -0.00000847, 0.00000601, -0.00001494, 0.00001212, -0.00001518, 0.00000438, 0.00001392, -0.00003700, 0.00003063, -0.00002323, 0.00002182, -0.00000340, 0.00001966, -0.00001177, 0.00000032, 0.00000078, 0.00000402, 0.00000105, 0.00000693, 0.00000274, 0.00000372, 0.00001669, -0.00001638, 0.00001859, -0.00002098, 0.00000145, 0.00000274, -0.00000770, -0.00000436, 0.00000034, 0.00000323, -0.00000225, -0.00000647, 0.00000696, 0.00000485, -0.00000519, -0.00000512, 0.00000224, 0.00000179, -0.00000480, 0.00000320, -0.00000576, 0.00001244, -0.00001168, 0.00000510, -0.00000748, 0.00001036, -0.00000849, 0.00001056, -0.00001284, 0.00000785, -0.00001187, 0.00000144, -0.00000572, 0.00001295, -0.00001805, 0.00000500, -0.00001286, 0.00000552, -0.00001024, 0.00001087, -0.00002799, 0.00002567, -0.00004957, -0.00000115, -0.00005331, 0.00000632, -0.00016245, -0.00032050, 0.00055081, 0.00067597, -0.00019258, -0.00024188, 0.00001972, -0.00009995, 0.00003102, -0.00004267, 0.00002002, -0.00003859, 0.00001650, -0.00001795, 0.00000654, -0.00002371, 0.00001212, -0.00000800, 0.00000312, -0.00002152, 0.00001268, -0.00001148, 0.00001301, -0.00001252, 0.00000905, -0.00001003, 0.00000391, -0.00001730, 0.00001604, -0.00000498, 0.00000494, -0.00000907, 0.00000281, 0.00000294, -0.00000525, -0.00000993, 0.00000696, 0.00000653, -0.00000617, -0.00000787, 0.00000974, 0.00000062, -0.00000253, -0.00000772, 0.00000806, 0.00000475, -0.00001184, 0.00001914, -0.00002142, 0.00000873, -0.00000523, -0.00000022, 0.00000362, -0.00000081, -0.00000384, -0.00000703, 0.00000232, -0.00001091, 0.00000923, },
		{ -0.00000215, -0.00000616, -0.00000071, -0.00000163, -0.00000868, 0.00001155, -0.00002500, 0.00003265, -0.00002901, 0.00000276, 0.00000777, -0.00001104, 0.00002056, -0.00000440, 0.00001992, -0.00002304, 0.00001374, -0.00000815, 0.00001162, -0.00000632, 0.00001599, -0.00001004, 0.00002142, -0.00000517, -0.00000018, 0.00000781, -0.00002119, 0.00000980, -0.00000412, -0.00000639, -0.00000299, 0.00000217, 0.00000235, -0.00000553, -0.00000301, 0.00001042, -0.00000093, -0.00000503, -0.00000485, 0.00000684, -0.00000531, 0.00000200, -0.00000417, 0.00000452, 0.00000277, -0.00000691, 0.00000138, -0.00000094, 0.00000411, -0.00000367, 0.00000525, -0.00000944, 0.00000564, -0.00001226, 0.00000170, 0.00000061, 0.00000423, -0.00001604, 0.00000505, -0.00001105, 0.00000274, -0.00000480, -0.00000065, -0.00001330, 0.00001152, -0.00005549, 0.00000706, -0.00006555, 0.00001680, -0.00032485, -0.00009486, 0.00090313, 0.00029474, -0.00039958, -0.00002840, -0.00007237, -0.00002292, -0.00000765, -0.00001096, -0.00000959, -0.00001289, 0.00000305, -0.00000753, -0.00000817, -0.00000833, 0.00000454, -0.00000085, -0.00001194, -0.00000508, 0.00000336, -0.00000277, 0.00000425, -0.00000572, 0.00000360, -0.00000576, -0.00000408, -0.00000742, 0.00001481, -0.00000751, 0.00000480, -0.00001054, 0.00000842, -0.00000222, -0.00000716, -0.00000610, 0.00001121, 0.00000100, -0.00000903, -0.00000151, 0.00001004, -0.00000377, -0.00000240, -0.00000712, 0.00001574, -0.00000968, 0.00000296, 0.00000773, -0.00001971, 0.00001473, -0.00001321, 0.00000964, -0.00000452, 0.00000471, -0.00001275, 0.00000357, -0.00000754, -0.00000172, 0.00000632, },
		{ 0.00000022, -0.00001081, 0.00000891, -0.00001588, 0.00001152, -0.00001371, 0.00000302, 0.00001551, -0.00003523, 0.00002797, -0.00002009, 0.00001688, -0.00000081, 0.00001407, -0.00000216, -0.00001070, 0.00000984, -0.00000474, 0.00000773, -0.00000134, 0.00001087, -0.00000734, 0.00002479, -0.00001893, 0.00001670, -0.00001115, -0.00000890, 0.00000841, -0.00000748, -0.00000514, -0.00000078, 0.00000272, 0.00000087, -0.00000860, 0.00000446, 0.00000766, -0.00000361, -0.00000649, 0.00000019, 0.00000445, -0.00000620, 0.00000353, -0.00000601, 0.00001043, -0.00000628, -0.00000081, -0.00000203, 0.00000467, -0.00000317, 0.00000322, -0.00000258, -0.00000256, -0.00000086, -0.00000921, 0.00000258, 0.00000429, -0.00000637, -0.00000862, 0.00000185, -0.00000865, 0.00000057, -0.00000205, -0.00000836, 0.00000184, -0.00001637, -0.00003955, -0.00000533, -0.00005499, -0.00003322, -0.00045335, 0.00037091, 0.00097579, -0.00026785, -0.00022595, -0.00001267, -0.00007680, 0.00000455, -0.00002009, -0.00000313, -0.00001883, 0.00000192, -0.00000562, -0.00000502, -0.00001256, 0.00000057, 0.00000148, -0.00000298, -0.00001581, 0.00000588, -0.00000469, 0.00000426, -0.00000324, 0.00000108, -0.00000240, -0.00000239, -0.00001031, 0.00000563, 0.00000599, -0.00000436, 0.00000053, -0.00000655, 0.00000943, -0.00000738, -0.00000678, -0.00000074, 0.00001196, -0.00000574, -0.00000760, 0.00000458, 0.00000617, -0.00000526, -0.00000382, -0.00000113, 0.00001453, -0.00001779, 0.00001793, -0.00001166, -0.00000428, 0.00000666, -0.00000874, 0.00000925, -0.00000526, 0.00000295, -0.00001393, 0.00000891, -0.00001558, 0.00001000, -0.00000174, },
		{ -0.00000432, -0.00000351, 0.00000298, -0.00001260, 0.00001411, -0.00002610, 0.00003282, -0.00002682, 0.00000208, 0.00000792, -0.00001156, 0.00001903, -0.00000569, 0.00002035, -0.00001935, 0.00000613, -0.00000016, 0.00000243, 0.00000128, 0.00000667, -0.00000010, 0.00000593, 0.00001255, -0.00001621, 0.00002153, -0.00002577, 0.00000825, -0.00000192, -0.00000420, -0.00000702, 0.00000429, -0.00000022, 0.00000048, -0.00001068, 0.00001394, -0.00000154, -0.00000096, -0.00001060, 0.00000975, -0.00000600, 0.00000227, -0.00000450, 0.00000345, 0.00000356, -0.00000465, -0.00000245, 0.00000297, -0.00000008, -0.00000061, 0.00000139, -0.00000241, -0.00000270, -0.00000213, -0.00000979, 0.00001018, -0.00000418, -0.00000567, -0.00000782, 0.00000450, -0.00001431, 0.00000845, -0.00001249, 0.00000276, -0.00000702, -0.00002741, -0.00002848, -0.00001783, -0.00004308, -0.00015301, -0.00043502, 0.00100867, 0.00048371, -0.00050588, 0.00000643, -0.00010543, -0.00000192, -0.00002956, 0.00000770, -0.00002995, 0.00000681, -0.00001256, 0.00000531, -0.00001959, 0.00000217, -0.00000692, 0.00000998, -0.00001919, -0.00000006, -0.00000140, 0.00000091, -0.00000084, 0.00000069, -0.00000216, -0.00000010, -0.00000790, -0.00000465, 0.00000903, 0.00000057, -0.00000295, -0.00000162, -0.00000232, 0.00000797, -0.00001263, -0.00000222, 0.00000239, 0.00001077, -0.00001297, -0.00000031, 0.00000528, 0.00000308, -0.00000675, -0.00000347, 0.00000622, 0.00000511, -0.00001402, 0.00002227, -0.00002624, 0.00001428, -0.00000909, 0.00000539, -0.00000172, 0.00000359, -0.00000930, -0.00000143, -0.00000079, -0.00000898, 0.00001051, -0.00000350, },
		{ -0.00001133, 0.00000937, -0.00001403, 0.00000886, -0.00001123, 0.00000122, 0.00001703, -0.00003371, 0.00002639, -0.00001850, 0.00001357, 0.00000101, 0.00000870, 0.00000571, -0.00001727, 0.00001276, -0.00000640, 0.00000721, -0.00000227, 0.00001093, -0.00000807, 0.00002013, -0.00000834, 0.00000272, 0.00000626, -0.00002131, 0.00001354, -0.00000766, -0.00000359, -0.00000428, 0.00000422, 0.00000080, -0.00000584, -0.00000230, 0.00001243, -0.00000394, -0.00000446, -0.00000490, 0.00000847, -0.00000791, 0.00000386, -0.00000577, 0.00000763, -0.00000115, -0.00000491, 0.00000087, 0.00000178, -0.00000039, -0.00000168, 0.00000385, -0.00000748, 0.00000348, -0.00001210, 0.00000319, 0.00000252, -0.00000190, -0.00001367, 0.00000625, -0.00001098, 0.00000027, -0.00000555, 0.00000074, -0.00000863, -0.00000098, -0.00005399, 0.00001119, -0.00007315, 0.00001212, -0.00041384, 0.00000430, 0.00133559, -0.00030841, -0.00024364, -0.00002316, -0.00007092, -0.00000762, -0.00001093, -0.00001290, -0.00001319, 0.00000004, -0.00000334, -0.00000817, -0.00000913, -0.00000329, 0.00000334, -0.00000279, -0.00001455, 0.00000252, -0.00000069, -0.00000162, 0.00000220, -0.00000271, 0.00000130, -0.00000540, -0.00000583, -0.00000157, 0.00001201, -0.00000785, 0.00000376, -0.00000852, 0.00000821, -0.00000353, -0.00000763, -0.00000343, 0.00001054, -0.00000001, -0.00000992, 0.00000223, 0.00000725, -0.00000357, -0.00000358, -0.00000475, 0.00001492, -0.00001169, 0.00000469, 0.00000639, -0.00002021, 0.00001735, -0.00001436, 0.00001236, -0.00000812, 0.00000793, -0.00001762, 0.00001106, -0.00001495, 0.00000628, 0.00000137, -0.00000002, },
		{ -0.00000636, 0.00000720, -0.00001528, 0.00001548, -0.00002682, 0.00003338, -0.00002614, 0.00000352, 0.00000611, -0.00001034, 0.00001586, -0.00000422, 0.00001601, -0.00001028, -0.00000542, 0.00001023, -0.00000779, 0.00000941, -0.00000314, 0.00001024, -0.00000805, 0.00002502, -0.00002364, 0.00002370, -0.00001953, 0.00000023, 0.00000240, -0.00000257, -0.00000974, 0.00000526, -0.00000276, 0.00000622, -0.00001604, 0.00001441, -0.00000110, 0.00000314, -0.00001509, 0.00001085, -0.00000605, 0.00000332, -0.00000661, 0.00000549, -0.00000056, 0.00000322, -0.00001143, 0.00001174, -0.00000918, 0.00000820, -0.00000965, 0.00001208, -0.00001793, 0.00001435, -0.00002664, 0.00002500, -0.00001921, 0.00001259, -0.00002900, 0.00002798, -0.00003817, 0.00002956, -0.00003620, 0.00003411, -0.00004375, 0.00002131, -0.00008670, 0.00005632, -0.00013843, 0.00003966, -0.00068416, 0.00090398, 0.00084623, -0.00068959, 0.00008280, -0.00017052, 0.00004545, -0.00007463, 0.00004646, -0.00006722, 0.00003898, -0.00003790, 0.00002746, -0.00004022, 0.00002145, -0.00002606, 0.00002687, -0.00003192, 0.00001112, -0.00001296, 0.00001215, -0.00001389, 0.00001422, -0.00001420, 0.00001143, -0.00001779, 0.00000532, -0.00000426, 0.00001490, -0.00001582, 0.00001150, -0.00001546, 0.00001855, -0.00001850, 0.00000393, -0.00000836, 0.00001972, -0.00001565, 0.00000208, -0.00000166, 0.00001087, -0.00001179, 0.00000234, -0.00000522, 0.00001830, -0.00002442, 0.00002574, -0.00002053, 0.00000392, 0.00000129, -0.00000222, 0.00000342, -0.00000035, -0.00000277, -0.00000856, 0.00000657, -0.00001478, 0.00001231, -0.00000339, -0.00000278, },
		{ 0.00000758, -0.00000995, 0.00000497, -0.00000820, -0.00000121, 0.00001942, -0.00003402, 0.00002765, -0.00002007, 0.00001338, 0.00000011, 0.00000660, 0.00000785, -0.00001563, 0.00000682, 0.00000066, -0.00000136, 0.00000470, 0.00000252, 0.00000111, 0.00000514, 0.00001039, -0.00001585, 0.00002286, -0.00002922, 0.00001470, -0.00000700, -0.00000041, -0.00000943, 0.00000749, -0.00000257, 0.00000123, -0.00001185, 0.00001806, -0.00000609, 0.00000096, -0.00001254, 0.00001364, -0.00001068, 0.00000589, -0.00000774, 0.00000794, -0.00000120, -0.00000202, -0.00000385, 0.00000674, -0.00000561, 0.00000220, -0.00000056, 0.00000014, -0.00000562, -0.00000092, -0.00000997, 0.00001425, -0.00001224, -0.00000180, -0.00000825, 0.00000689, -0.00001982, 0.00001114, -0.00001448, 0.00001083, -0.00002240, -0.00002318, -0.00002787, -0.00001943, -0.00005375, -0.00021059, -0.00046968, 0.00163690, -0.00021880, -0.00035401, 0.00002833, -0.00011358, 0.00002066, -0.00003768, 0.00001000, -0.00003753, 0.00002350, -0.00002199, 0.00000734, -0.00002314, 0.00000947, -0.00001020, 0.00001053, -0.00002440, 0.00000957, -0.00000683, 0.00000302, -0.00000365, 0.00000448, -0.00000520, 0.00000101, -0.00001037, 0.00000133, 0.00000658, 0.00000008, -0.00000409, 0.00000050, -0.00000289, 0.00000770, -0.00001440, 0.00000121, 0.00000107, 0.00001114, -0.00001549, 0.00000436, 0.00000225, 0.00000367, -0.00000846, -0.00000117, 0.00000575, 0.00000354, -0.00001418, 0.00002415, -0.00003073, 0.00002056, -0.00001317, 0.00001051, -0.00000723, 0.00000876, -0.00001612, 0.00000770, -0.00000927, -0.00000063, 0.00000566, -0.00000114, -0.00001025, },
		{ 0.00000916, -0.00001521, 0.00001475, -0.00002646, 0.00003346, -0.00002563, 0.00000498, 0.00000502, -0.00001017, 0.00001366, -0.00000274, 0.00001075, -0.00000143, -0.00001414, 0.00001594, -0.00001218, 0.00001151, -0.00000680, 0.00001366, -0.00001268, 0.00002507, -0.00001733, 0.00001263, -0.00000427, -0.00001122, 0.00000695, -0.00000145, -0.00001022, 0.00000400, -0.00000361, 0.00000918, -0.00001668, 0.00001053, 0.00000126, 0.00000563, -0.00001646, 0.00000899, -0.00000484, 0.00000437, -0.00000931, 0.00000889, -0.00000675, 0.00001245, -0.00002035, 0.00001976, -0.00001736, 0.00001650, -0.00002059, 0.00002537, -0.00003043, 0.00002691, -0.00003842, 0.00003477, -0.00003012, 0.00002680, -0.00004512, 0.00004483, -0.00005389, 0.00004338, -0.00005487, 0.00006022, -0.00007345, 0.00005964, -0.00012947, 0.00010858, -0.00020173, 0.00015580, -0.00073944, 0.00054315, 0.00135435, -0.00089639, 0.00015456, -0.00023148, 0.00008909, -0.00011644, 0.00008122, -0.00009918, 0.00006381, -0.00005626, 0.00004463, -0.00005704, 0.00003727, -0.00004141, 0.00003957, -0.00004136, 0.00002068, -0.00002389, 0.00002349, -0.00002701, 0.00002675, -0.00002473, 0.00002156, -0.00002683, 0.00001530, -0.00001701, 0.00002690, -0.00002526, 0.00002045, -0.00002314, 0.00002293, -0.00001953, 0.00000718, -0.00001525, 0.00002272, -0.00001358, 0.00000249, -0.00000673, 0.00001522, -0.00001333, 0.00000551, -0.00001231, 0.00002306, -0.00002288, 0.00001619, -0.00000472, -0.00001146, 0.00001233, -0.00000836, 0.00000710, -0.00000358, 0.00000282, -0.00001334, 0.00001014, -0.00001557, 0.00000956, -0.00000057, -0.00000282, -0.00000758, },
		{ -0.00000571, 0.00000179, -0.00000589, -0.00000366, 0.00002253, -0.00003624, 0.00003155, -0.00002411, 0.00001551, -0.00000312, 0.00000801, 0.00000418, -0.00000702, -0.00000516, 0.00001245, -0.00001341, 0.00001473, -0.00000919, 0.00001384, -0.00001188, 0.00002741, -0.00002874, 0.00003012, -0.00002782, 0.00001034, -0.00000523, 0.00000367, -0.00001502, 0.00001147, -0.00000804, 0.00001050, -0.00002153, 0.00002281, -0.00000937, 0.00000883, -0.00002146, 0.00001937, -0.00001504, 0.00001097, -0.00001388, 0.00001392, -0.00000906, 0.00000988, -0.00001748, 0.00002060, -0.00001999, 0.00001633, -0.00001711, 0.00002075, -0.00002769, 0.00002314, -0.00003537, 0.00003838, -0.00003670, 0.00002611, -0.00004009, 0.00004259, -0.00005736, 0.00004702, -0.00005408, 0.00005991, -0.00007891, 0.00004856, -0.00011462, 0.00009246, -0.00019935, 0.00007631, -0.00091246, 0.00173995, 0.00008381, -0.00059160, 0.00015852, -0.00021579, 0.00009752, -0.00010633, 0.00006943, -0.00009343, 0.00007246, -0.00006184, 0.00004241, -0.00005570, 0.00003966, -0.00003948, 0.00003730, -0.00004635, 0.00002878, -0.00002533, 0.00002056, -0.00002270, 0.00002388, -0.00002282, 0.00001789, -0.00002532, 0.00001567, -0.00001071, 0.00001870, -0.00002144, 0.00001806, -0.00002067, 0.00002327, -0.00002503, 0.00001136, -0.00001362, 0.00002453, -0.00002236, 0.00000995, -0.00000732, 0.00001439, -0.00001648, 0.00000715, -0.00000838, 0.00002058, -0.00002967, 0.00003318, -0.00003027, 0.00001415, -0.00000550, 0.00000494, -0.00000368, 0.00000646, -0.00001105, 0.00000150, -0.00000214, -0.00000679, 0.00000799, -0.00000103, -0.00000923, 0.00000533, },
		{ -0.00001255, 0.00001239, -0.00002521, 0.00003262, -0.00002409, 0.00000448, 0.00000695, -0.00001324, 0.00001457, -0.00000386, 0.00000842, 0.00000214, -0.00001486, 0.00001257, -0.00000727, 0.00000478, -0.00000170, 0.00000750, -0.00000586, 0.00001255, 0.00000007, -0.00000609, 0.00001299, -0.00002040, 0.00000891, -0.00000041, -0.00000817, 0.00000007, -0.00000173, 0.00000778, -0.00001160, 0.00000230, 0.00000583, 0.00000516, -0.00001316, 0.00000316, -0.00000120, 0.00000338, -0.00000954, 0.00000950, -0.00000947, 0.00001620, -0.00002184, 0.00001957, -0.00001699, 0.00001611, -0.00002210, 0.00002711, -0.00002944, 0.00002477, -0.00003450, 0.00002888, -0.00002558, 0.00002428, -0.00004229, 0.00004035, -0.00004648, 0.00003411, -0.00004999, 0.00005852, -0.00006926, 0.00005620, -0.00012035, 0.00009570, -0.00017866, 0.00013239, -0.00056633, -0.00002892, 0.00189990, -0.00102071, 0.00014643, -0.00023353, 0.00008527, -0.00011805, 0.00007989, -0.00009782, 0.00005756, -0.00004792, 0.00003960, -0.00005412, 0.00003466, -0.00003915, 0.00003579, -0.00003658, 0.00001825, -0.00002385, 0.00002439, -0.00002943, 0.00002778, -0.00002398, 0.00002117, -0.00002619, 0.00001613, -0.00001964, 0.00002730, -0.00002289, 0.00001771, -0.00001899, 0.00001628, -0.00001189, 0.00000343, -0.00001389, 0.00001652, -0.00000481, -0.00000131, -0.00000672, 0.00001291, -0.00000860, 0.00000289, -0.00001148, 0.00001703, -0.00001007, -0.00000184, 0.00001378, -0.00002396, 0.00001762, -0.00000873, 0.00000643, -0.00000349, 0.00000467, -0.00001315, 0.00000812, -0.00001086, 0.00000301, 0.00000394, -0.00000403, -0.00000704, 0.00000866, },
		{ 0.00000118, -0.00000553, -0.00000528, 0.00002550, -0.00003912, 0.00003609, -0.00002805, 0.00001729, -0.00000622, 0.00001046, -0.00000186, 0.00000331, -0.00001654, 0.00002181, -0.00002175, 0.00002083, -0.00001696, 0.00002199, -0.00002185, 0.00003386, -0.00002886, 0.00002444, -0.00001726, 0.00000241, -0.00000366, 0.00000833, -0.00001979, 0.00001477, -0.00001365, 0.00001902, -0.00002840, 0.00002490, -0.00001253, 0.00001726, -0.00002950, 0.00002426, -0.00002029, 0.00001849, -0.00002339, 0.00002439, -0.00002265, 0.00002729, -0.00003554, 0.00003843, -0.00003841, 0.00003525, -0.00003935, 0.00004648, -0.00005379, 0.00005038, -0.00006304, 0.00006500, -0.00006505, 0.00005871, -0.00007645, 0.00008199, -0.00009774, 0.00008740, -0.00010181, 0.00011889, -0.00014591, 0.00013041, -0.00021015, 0.00021152, -0.00034946, 0.00032827, -0.00117829, 0.00154200, 0.00062098, -0.00093307, 0.00033858, -0.00035240, 0.00020083, -0.00019664, 0.00014624, -0.00016258, 0.00013015, -0.00010872, 0.00008503, -0.00009590, 0.00007692, -0.00007467, 0.00006869, -0.00007300, 0.00005374, -0.00005024, 0.00004504, -0.00004850, 0.00004871, -0.00004502, 0.00003911, -0.00004491, 0.00003541, -0.00003289, 0.00004032, -0.00004043, 0.00003622, -0.00003743, 0.00003656, -0.00003423, 0.00002186, -0.00002768, 0.00003486, -0.00002686, 0.00001584, -0.00001753, 0.00002425, -0.00002342, 0.00001530, -0.00002085, 0.00003171, -0.00003493, 0.00002976, -0.00001923, 0.00000167, 0.00000386, 0.00000008, -0.00000109, 0.00000449, -0.00000643, -0.00000291, 0.00000167, -0.00000812, 0.00000561, 0.00000205, -0.00000989, 0.00000459, -0.00000350, },
		{ 0.00001025, -0.00002426, 0.00003129, -0.00002131, 0.00000142, 0.00001233, -0.00001961, 0.00001858, -0.00000808, 0.00001034, -0.00000107, -0.00000716, 0.00000118, 0.00000470, -0.00000789, 0.00000908, -0.00000484, 0.00000780, -0.00000582, 0.00001979, -0.00002273, 0.00002387, -0.00002244, 0.00000686, 0.00000032, -0.00000342, -0.00000636, 0.00000304, 0.00000184, -0.00000150, -0.00000895, 0.00001217, 0.00000099, -0.00000459, -0.00000695, 0.00000585, -0.00000188, -0.00000405, 0.00000320, -0.00000367, 0.00000899, -0.00001072, 0.00000644, -0.00000340, 0.00000200, -0.00000857, 0.00001148, -0.00000963, 0.00000320, -0.00001069, 0.00000326, -0.00000083, -0.00000059, -0.00001448, 0.00000865, -0.00001030, -0.00000449, -0.00001311, 0.00001825, -0.00001908, -0.00000169, -0.00004671, 0.00000542, -0.00005831, -0.00002855, -0.00021831, -0.00068028, 0.00232110, -0.00094202, -0.00001190, -0.00013186, 0.00000071, -0.00005284, 0.00002064, -0.00004543, 0.00000644, -0.00000206, 0.00000268, -0.00002202, 0.00000468, -0.00001124, 0.00000854, -0.00001122, -0.00000272, -0.00000587, 0.00000761, -0.00001390, 0.00001067, -0.00000619, 0.00000502, -0.00001070, 0.00000244, -0.00000683, 0.00001173, -0.00000558, 0.00000121, -0.00000205, -0.00000129, 0.00000408, -0.00000773, -0.00000370, 0.00000173, 0.00000941, -0.00000912, -0.00000090, 0.00000362, 0.00000216, -0.00000533, -0.00000335, 0.00000294, 0.00000816, -0.00002010, 0.00002651, -0.00002742, 0.00001417, -0.00000286, 0.00000204, -0.00000070, 0.00000334, -0.00000914, 0.00000271, -0.00000394, -0.00000354, 0.00000698, -0.00000424, -0.00000634, 0.00000735, -0.00000924, },
		{ -0.00000705, -0.00000592, 0.00002761, -0.00004116, 0.00003892, -0.00002912, 0.00001599, -0.00000650, 0.00001077, -0.00000577, 0.00000944, -0.00002096, 0.00002283, -0.00002126, 0.00001843, -0.00001627, 0.00002073, -0.00002024, 0.00002693, -0.00001630, 0.00000925, -0.00000296, -0.00000447, -0.00000437, 0.00001323, -0.00002251, 0.00001592, -0.00001722, 0.00002392, -0.00003001, 0.00002298, -0.00001408, 0.00002363, -0.00003379, 0.00002600, -0.00002414, 0.00002537, -0.00003217, 0.00003418, -0.00003529, 0.00004206, -0.00004908, 0.00005098, -0.00005131, 0.00004871, -0.00005572, 0.00006448, -0.00007026, 0.00006678, -0.00007883, 0.00007984, -0.00008210, 0.00007920, -0.00009888, 0.00010534, -0.00012040, 0.00011047, -0.00013259, 0.00015767, -0.00018758, 0.00017992, -0.00026392, 0.00027583, -0.00042291, 0.00044064, -0.00117470, 0.00102907, 0.00132269, -0.00128603, 0.00049304, -0.00046466, 0.00028218, -0.00026689, 0.00020353, -0.00021225, 0.00016830, -0.00013872, 0.00011419, -0.00012433, 0.00010307, -0.00009891, 0.00008956, -0.00009083, 0.00007183, -0.00006919, 0.00006405, -0.00006844, 0.00006658, -0.00006014, 0.00005378, -0.00005860, 0.00004976, -0.00004865, 0.00005380, -0.00005067, 0.00004541, -0.00004469, 0.00004069, -0.00003631, 0.00002693, -0.00003506, 0.00003723, -0.00002561, 0.00001859, -0.00002402, 0.00002877, -0.00002521, 0.00001879, -0.00002650, 0.00003259, -0.00002844, 0.00001643, -0.00000365, -0.00000967, 0.00000859, 0.00000048, -0.00000273, 0.00000584, -0.00000554, -0.00000228, -0.00000044, -0.00000328, -0.00000143, 0.00000761, -0.00001223, 0.00000593, -0.00000402, 0.00000347, },
		{ -0.00002503, 0.00003041, -0.00001806, -0.00000317, 0.00001950, -0.00002706, 0.00002341, -0.00001340, 0.00001469, -0.00000841, 0.00000460, -0.00001238, 0.00001716, -0.00001980, 0.00001889, -0.00001642, 0.00002030, -0.00002070, 0.00003222, -0.00002908, 0.00002353, -0.00001652, 0.00000229, -0.00000070, 0.00000411, -0.00001459, 0.00000992, -0.00000740, 0.00001135, -0.00002083, 0.00001912, -0.00000649, 0.00000836, -0.00002014, 0.00001595, -0.00001199, 0.00000814, -0.00001107, 0.00001166, -0.00000958, 0.00001221, -0.00001790, 0.00002131, -0.00002370, 0.00001777, -0.00001851, 0.00002477, -0.00003252, 0.00002704, -0.00003591, 0.00003823, -0.00004189, 0.00003163, -0.00004248, 0.00004582, -0.00006346, 0.00004730, -0.00005144, 0.00006540, -0.00009844, 0.00007054, -0.00013411, 0.00011963, -0.00026097, 0.00018298, -0.00123582, 0.00246054, -0.00057949, -0.00034338, 0.00007702, -0.00016324, 0.00007585, -0.00009177, 0.00005187, -0.00008229, 0.00007403, -0.00006024, 0.00003469, -0.00004857, 0.00003822, -0.00003802, 0.00003108, -0.00003962, 0.00002826, -0.00002539, 0.00001800, -0.00002235, 0.00002583, -0.00002401, 0.00001703, -0.00002342, 0.00001870, -0.00001599, 0.00002187, -0.00002380, 0.00002207, -0.00002405, 0.00002344, -0.00002254, 0.00001166, -0.00001719, 0.00002494, -0.00001834, 0.00000874, -0.00000996, 0.00001573, -0.00001599, 0.00000815, -0.00001329, 0.00002425, -0.00003115, 0.00002865, -0.00002101, 0.00000454, 0.00000503, -0.00000187, 0.00000117, 0.00000234, -0.00000542, -0.00000125, -0.00000000, -0.00000537, 0.00000514, -0.00000119, -0.00000751, 0.00000752, -0.00000777, 0.00001050, },
		{ -0.00000658, 0.00002901, -0.00004174, 0.00003883, -0.00002612, 0.00001066, -0.00000302, 0.00000745, -0.00000502, 0.00000835, -0.00001615, 0.00001456, -0.00001208, 0.00000831, -0.00000805, 0.00001119, -0.00000926, 0.00001108, 0.00000202, -0.00000701, 0.00000747, -0.00000586, -0.00000825, 0.00001725, -0.00002176, 0.00001361, -0.00001696, 0.00002316, -0.00002513, 0.00001650, -0.00001263, 0.00002526, -0.00003158, 0.00002216, -0.00002350, 0.00002720, -0.00003462, 0.00003655, -0.00003897, 0.00004531, -0.00004916, 0.00004948, -0.00004975, 0.00004714, -0.00005569, 0.00006356, -0.00006598, 0.00006140, -0.00007203, 0.00007198, -0.00007588, 0.00007421, -0.00009291, 0.00009749, -0.00010970, 0.00009913, -0.00012599, 0.00015189, -0.00017598, 0.00016570, -0.00024084, 0.00024547, -0.00037321, 0.00036892, -0.00090272, 0.00029445, 0.00203494, -0.00151243, 0.00052627, -0.00048443, 0.00028821, -0.00027300, 0.00020406, -0.00021089, 0.00016073, -0.00012999, 0.00011037, -0.00012264, 0.00010091, -0.00009647, 0.00008585, -0.00008702, 0.00007055, -0.00006965, 0.00006500, -0.00007008, 0.00006582, -0.00005760, 0.00005204, -0.00005684, 0.00004905, -0.00004844, 0.00005053, -0.00004490, 0.00003955, -0.00003769, 0.00003224, -0.00002844, 0.00002326, -0.00003241, 0.00002961, -0.00001750, 0.00001630, -0.00002400, 0.00002553, -0.00002002, 0.00001573, -0.00002374, 0.00002369, -0.00001407, 0.00000016, 0.00000853, -0.00001348, 0.00000511, 0.00000728, -0.00000854, 0.00001038, -0.00000831, 0.00000280, -0.00000684, 0.00000490, -0.00000961, 0.00001257, -0.00001419, 0.00000789, -0.00000583, 0.00000698, -0.00000882, },
		{ 0.00003044, -0.00001529, -0.00000765, 0.00002587, -0.00003251, 0.00002598, -0.00001685, 0.00001808, -0.00001519, 0.00001413, -0.00002109, 0.00002327, -0.00002467, 0.00002196, -0.00002139, 0.00002524, -0.00002565, 0.00003235, -0.00002304, 0.00001359, -0.00000676, -0.00000084, -0.00000532, 0.00001398, -0.00002315, 0.00001721, -0.00001749, 0.00002367, -0.00003042, 0.00002492, -0.00001555, 0.00002288, -0.00003337, 0.00002672, -0.00002478, 0.00002427, -0.00002973, 0.00003195, -0.00003347, 0.00003933, -0.00004491, 0.00004822, -0.00005169, 0.00004695, -0.00005159, 0.00006100, -0.00006868, 0.00006441, -0.00007412, 0.00007692, -0.00008409, 0.00007890, -0.00009410, 0.00010141, -0.00012135, 0.00010853, -0.00012452, 0.00015284, -0.00019627, 0.00018575, -0.00026712, 0.00028436, -0.00046698, 0.00049872, -0.00154362, 0.00223040, 0.00005995, -0.00079523, 0.00033963, -0.00036121, 0.00022877, -0.00022230, 0.00016220, -0.00017963, 0.00015490, -0.00012759, 0.00009711, -0.00010761, 0.00009270, -0.00008847, 0.00007645, -0.00008021, 0.00006723, -0.00006359, 0.00005516, -0.00005983, 0.00006067, -0.00005518, 0.00004690, -0.00005168, 0.00004687, -0.00004514, 0.00004864, -0.00004699, 0.00004362, -0.00004322, 0.00003890, -0.00003497, 0.00002617, -0.00003387, 0.00003649, -0.00002529, 0.00001866, -0.00002341, 0.00002747, -0.00002473, 0.00001817, -0.00002581, 0.00003265, -0.00003193, 0.00002124, -0.00000966, -0.00000454, 0.00000852, -0.00000025, -0.00000176, 0.00000531, -0.00000598, 0.00000036, -0.00000266, -0.00000009, -0.00000252, 0.00000542, -0.00001118, 0.00001032, -0.00000943, 0.00001405, -0.00002796, },
		{ 0.00003061, -0.00004138, 0.00003637, -0.00002010, 0.00000287, 0.00000263, 0.00000165, -0.00000040, 0.00000137, -0.00000501, 0.00000139, 0.00000052, -0.00000404, 0.00000227, -0.00000089, 0.00000415, -0.00000514, 0.00001674, -0.00001596, 0.00000886, -0.00000075, -0.00001353, 0.00001808, -0.00001619, 0.00000712, -0.00001202, 0.00001621, -0.00001423, 0.00000611, -0.00000753, 0.00002049, -0.00002157, 0.00001164, -0.00001631, 0.00002073, -0.00002674, 0.00002703, -0.00002890, 0.00003246, -0.00003209, 0.00003094, -0.00003086, 0.00002744, -0.00003591, 0.00004071, -0.00003880, 0.00003288, -0.00004169, 0.00004041, -0.00004469, 0.00004143, -0.00005628, 0.00005662, -0.00006412, 0.00005118, -0.00007811, 0.00009663, -0.00010675, 0.00008494, -0.00014064, 0.00012389, -0.00021090, 0.00014627, -0.00045727, -0.00048803, 0.00256776, -0.00148086, 0.00036751, -0.00037009, 0.00018987, -0.00019313, 0.00013115, -0.00014606, 0.00009852, -0.00007566, 0.00006687, -0.00008399, 0.00006406, -0.00006180, 0.00005283, -0.00005702, 0.00004491, -0.00004620, 0.00004244, -0.00004833, 0.00004227, -0.00003415, 0.00003104, -0.00003679, 0.00003039, -0.00002986, 0.00002938, -0.00002326, 0.00001961, -0.00001811, 0.00001342, -0.00001244, 0.00001172, -0.00002060, 0.00001389, -0.00000440, 0.00000949, -0.00001733, 0.00001506, -0.00000900, 0.00000733, -0.00001459, 0.00000928, 0.00000168, -0.00001191, 0.00001210, -0.00000824, -0.00000468, 0.00001661, -0.00001436, 0.00001438, -0.00001108, 0.00000808, -0.00001246, 0.00001081, -0.00001362, 0.00001291, -0.00001294, 0.00000806, -0.00000637, 0.00000896, -0.00000843, -0.00000883, },
		{ -0.00001316, -0.00001093, 0.00002964, -0.00003400, 0.00002450, -0.00001659, 0.00001806, -0.00001778, 0.00001700, -0.00002096, 0.00002014, -0.00002066, 0.00001702, -0.00001853, 0.00002145, -0.00002045, 0.00002222, -0.00000947, 0.00000096, 0.00000015, 0.00000144, -0.00001403, 0.00002439, -0.00002968, 0.00002248, -0.00002535, 0.00003194, -0.00003492, 0.00002757, -0.00002341, 0.00003483, -0.00004229, 0.00003414, -0.00003569, 0.00003847, -0.00004557, 0.00004869, -0.00005266, 0.00005940, -0.00006296, 0.00006564, -0.00006994, 0.00006607, -0.00007359, 0.00008388, -0.00008972, 0.00008570, -0.00009553, 0.00009899, -0.00010960, 0.00010786, -0.00012503, 0.00013395, -0.00015447, 0.00014469, -0.00017078, 0.00020756, -0.00025376, 0.00025114, -0.00033804, 0.00036942, -0.00056330, 0.00063061, -0.00153390, 0.00165456, 0.00086614, -0.00124349, 0.00055827, -0.00051818, 0.00034473, -0.00031901, 0.00024066, -0.00024670, 0.00020751, -0.00017127, 0.00014007, -0.00014890, 0.00013019, -0.00012252, 0.00010657, -0.00010776, 0.00009499, -0.00009116, 0.00008184, -0.00008621, 0.00008355, -0.00007483, 0.00006613, -0.00006997, 0.00006526, -0.00006335, 0.00006323, -0.00005819, 0.00005370, -0.00005123, 0.00004445, -0.00003990, 0.00003442, -0.00004296, 0.00004011, -0.00002719, 0.00002553, -0.00003253, 0.00003347, -0.00002820, 0.00002325, -0.00003131, 0.00003193, -0.00002453, 0.00001000, -0.00000083, -0.00000643, 0.00000328, 0.00000889, -0.00001026, 0.00001288, -0.00001153, 0.00000783, -0.00001122, 0.00001032, -0.00001309, 0.00001297, -0.00001566, 0.00001471, -0.00001327, 0.00001965, -0.00003177, 0.00003071, },
		{ -0.00004110, 0.00003316, -0.00001358, -0.00000434, 0.00000743, -0.00000377, 0.00000498, -0.00000717, 0.00000645, -0.00000973, 0.00000959, -0.00001208, 0.00000857, -0.00000925, 0.00001301, -0.00001435, 0.00002159, -0.00001414, 0.00000169, 0.00000732, -0.00001613, 0.00001319, -0.00000534, -0.00000304, -0.00000312, 0.00000443, 0.00000042, -0.00000619, 0.00000056, 0.00000968, -0.00000505, -0.00000404, -0.00000330, 0.00000612, -0.00000894, 0.00000659, -0.00000671, 0.00000641, -0.00000225, 0.00000064, -0.00000023, -0.00000459, -0.00000266, 0.00000334, 0.00000249, -0.00000911, 0.00000199, -0.00000457, 0.00000131, -0.00000849, -0.00000087, -0.00000394, 0.00000180, -0.00001832, -0.00000458, 0.00001002, -0.00000243, -0.00003409, 0.00000118, -0.00004365, 0.00000334, -0.00013543, 0.00001310, -0.00112628, 0.00276537, -0.00112628, 0.00001310, -0.00013543, 0.00000334, -0.00004365, 0.00000118, -0.00003409, -0.00000243, 0.00001002, -0.00000458, -0.00001832, 0.00000180, -0.00000394, -0.00000087, -0.00000849, 0.00000131, -0.00000457, 0.00000199, -0.00000911, 0.00000249, 0.00000334, -0.00000266, -0.00000459, -0.00000023, 0.00000064, -0.00000225, 0.00000641, -0.00000671, 0.00000659, -0.00000894, 0.00000612, -0.00000330, -0.00000404, -0.00000505, 0.00000968, 0.00000056, -0.00000619, 0.00000042, 0.00000443, -0.00000312, -0.00000304, -0.00000534, 0.00001319, -0.00001613, 0.00000732, 0.00000169, -0.00001414, 0.00002159, -0.00001435, 0.00001301, -0.00000925, 0.00000857, -0.00001208, 0.00000959, -0.00000973, 0.00000645, -0.00000717, 0.00000498, -0.00000377, 0.00000743, -0.00000434, -0.00001358, 0.00003316, },
		{ -0.00001316, 0.00003071, -0.00003177, 0.00001965, -0.00001327, 0.00001471, -0.00001566, 0.00001297, -0.00001309, 0.00001032, -0.00001122, 0.00000783, -0.00001153, 0.00001288, -0.00001026, 0.00000889, 0.00000328, -0.00000643, -0.00000083, 0.00001000, -0.00002453, 0.00003193, -0.00003131, 0.00002325, -0.00002820, 0.00003347, -0.00003253, 0.00002553, -0.00002719, 0.00004011, -0.00004296, 0.00003442, -0.00003990, 0.00004445, -0.00005123, 0.00005370, -0.00005819, 0.00006323, -0.00006335, 0.00006526, -0.00006997, 0.00006613, -0.00007483, 0.00008355, -0.00008621, 0.00008184, -0.00009116, 0.00009499, -0.00010776, 0.00010657, -0.00012252, 0.00013019, -0.00014890, 0.00014007, -0.00017127, 0.00020751, -0.00024670, 0.00024066, -0.00031901, 0.00034473, -0.00051818, 0.00055827, -0.00124349, 0.00086614, 0.00165456, -0.00153390, 0.00063061, -0.00056330, 0.00036942, -0.00033804, 0.00025114, -0.00025376, 0.00020756, -0.00017078, 0.00014469, -0.00015447, 0.00013395, -0.00012503, 0.00010786, -0.00010960, 0.00009899, -0.00009553, 0.00008570, -0.00008972, 0.00008388, -0.00007359, 0.00006607, -0.00006994, 0.00006564, -0.00006296, 0.00005940, -0.00005266, 0.00004869, -0.00004557, 0.00003847, -0.00003569, 0.00003414, -0.00004229, 0.00003483, -0.00002341, 0.00002757, -0.00003492, 0.00003194, -0.00002535, 0.00002248, -0.00002968, 0.00002439, -0.00001403, 0.00000144, 0.00000015, 0.00000096, -0.00000947, 0.00002222, -0.00002045, 0.00002145, -0.00001853, 0.00001702, -0.00002066, 0.00002014, -0.00002096, 0.00001700, -0.00001778, 0.00001806, -0.00001659, 0.00002450, -0.00003400, 0.00002964, -0.00001093, },
		{ 0.00003061, -0.00000883, -0.00000843, 0.00000896, -0.00000637, 0.00000806, -0.00001294, 0.00001291, -0.00001362, 0.00001081, -0.00001246, 0.00000808, -0.00001108, 0.00001438, -0.00001436, 0.00001661, -0.00000468, -0.00000824, 0.00001210, -0.00001191, 0.00000168, 0.00000928, -0.00001459, 0.00000733, -0.00000900, 0.00001506, -0.00001733, 0.00000949, -0.00000440, 0.00001389, -0.00002060, 0.00001172, -0.00001244, 0.00001342, -0.00001811, 0.00001961, -0.00002326, 0.00002938, -0.00002986, 0.00003039, -0.00003679, 0.00003104, -0.00003415, 0.00004227, -0.00004833, 0.00004244, -0.00004620, 0.00004491, -0.00005702, 0.00005283, -0.00006180, 0.00006406, -0.00008399, 0.00006687, -0.00007566, 0.00009852, -0.00014606, 0.00013115, -0.00019313, 0.00018987, -0.00037009, 0.00036751, -0.00148086, 0.00256776, -0.00048803, -0.00045727, 0.00014627, -0.00021090, 0.00012389, -0.00014064, 0.00008494, -0.00010675, 0.00009663, -0.00007811, 0.00005118, -0.00006412, 0.00005662, -0.00005628, 0.00004143, -0.00004469, 0.00004041, -0.00004169, 0.00003288, -0.00003880, 0.00004071, -0.00003591, 0.00002744, -0.00003086, 0.00003094, -0.00003209, 0.00003246, -0.00002890, 0.00002703, -0.00002674, 0.00002073, -0.00001631, 0.00001164, -0.00002157, 0.00002049, -0.00000753, 0.00000611, -0.00001423, 0.00001621, -0.00001202, 0.00000712, -0.00001619, 0.00001808, -0.00001353, -0.00000075, 0.00000886, -0.00001596, 0.00001674, -0.00000514, 0.00000415, -0.00000089, 0.00000227, -0.00000404, 0.00000052, 0.00000139, -0.00000501, 0.00000137, -0.00000040, 0.00000165, 0.00000263, 0.00000287, -0.00002010, 0.00003637, -0.00004138, },
		{ 0.00003044, -0.00002796, 0.00001405, -0.00000943, 0.00001032, -0.00001118, 0.00000542, -0.00000252, -0.00000009, -0.00000266, 0.00000036, -0.00000598, 0.00000531, -0.00000176, -0.00000025, 0.00000852, -0.00000454, -0.00000966, 0.00002124, -0.00003193, 0.00003265, -0.00002581, 0.00001817, -0.00002473, 0.00002747, -0.00002341, 0.00001866, -0.00002529, 0.00003649, -0.00003387, 0.00002617, -0.00003497, 0.00003890, -0.00004322, 0.00004362, -0.00004699, 0.00004864, -0.00004514, 0.00004687, -0.00005168, 0.00004690, -0.00005518, 0.00006067, -0.00005983, 0.00005516, -0.00006359, 0.00006723, -0.00008021, 0.00007645, -0.00008847, 0.00009270, -0.00010761, 0.00009711, -0.00012759, 0.00015490, -0.00017963, 0.00016220, -0.00022230, 0.00022877, -0.00036121, 0.00033963, -0.00079523, 0.00005995, 0.00223040, -0.00154362, 0.00049872, -0.00046698, 0.00028436, -0.00026712, 0.00018575, -0.00019627, 0.00015284, -0.00012452, 0.00010853, -0.00012135, 0.00010141, -0.00009410, 0.00007890, -0.00008409, 0.00007692, -0.00007412, 0.00006441, -0.00006868, 0.00006100, -0.00005159, 0.00004695, -0.00005169, 0.00004822, -0.00004491, 0.00003933, -0.00003347, 0.00003195, -0.00002973, 0.00002427, -0.00002478, 0.00002672, -0.00003337, 0.00002288, -0.00001555, 0.00002492, -0.00003042, 0.00002367, -0.00001749, 0.00001721, -0.00002315, 0.00001398, -0.00000532, -0.00000084, -0.00000676, 0.00001359, -0.00002304, 0.00003235, -0.00002565, 0.00002524, -0.00002139, 0.00002196, -0.00002467, 0.00002327, -0.00002109, 0.00001413, -0.00001519, 0.00001808, -0.00001685, 0.00002598, -0.00003251, 0.00002587, -0.00000765, -0.00001529, },
		{ -0.00000658, -0.00000882, 0.00000698, -0.00000583, 0.00000789, -0.00001419, 0.00001257, -0.00000961, 0.00000490, -0.00000684, 0.00000280, -0.00000831, 0.00001038, -0.00000854, 0.00000728, 0.00000511, -0.00001348, 0.00000853, 0.00000016, -0.00001407, 0.00002369, -0.00002374, 0.00001573, -0.00002002, 0.00002553, -0.00002400, 0.00001630, -0.00001750, 0.00002961, -0.00003241, 0.00002326, -0.00002844, 0.00003224, -0.00003769, 0.00003955, -0.00004490, 0.00005053, -0.00004844, 0.00004905, -0.00005684, 0.00005204, -0.00005760, 0.00006582, -0.00007008, 0.00006500, -0.00006965, 0.00007055, -0.00008702, 0.00008585, -0.00009647, 0.00010091, -0.00012264, 0.00011037, -0.00012999, 0.00016073, -0.00021089, 0.00020406, -0.00027300, 0.00028821, -0.00048443, 0.00052627, -0.00151243, 0.00203494, 0.00029445, -0.00090272, 0.00036892, -0.00037321, 0.00024547, -0.00024084, 0.00016570, -0.00017598, 0.00015189, -0.00012599, 0.00009913, -0.00010970, 0.00009749, -0.00009291, 0.00007421, -0.00007588, 0.00007198, -0.00007203, 0.00006140, -0.00006598, 0.00006356, -0.00005569, 0.00004714, -0.00004975, 0.00004948, -0.00004916, 0.00004531, -0.00003897, 0.00003655, -0.00003462, 0.00002720, -0.00002350, 0.00002216, -0.00003158, 0.00002526, -0.00001263, 0.00001650, -0.00002513, 0.00002316, -0.00001696, 0.00001361, -0.00002176, 0.00001725, -0.00000825, -0.00000586, 0.00000747, -0.00000701, 0.00000202, 0.00001108, -0.00000926, 0.00001119, -0.00000805, 0.00000831, -0.00001208, 0.00001456, -0.00001615, 0.00000835, -0.00000502, 0.00000745, -0.00000302, 0.00001066, -0.00002612, 0.00003883, -0.00004174, 0.00002901, },
		{ -0.00002503, 0.00001050, -0.00000777, 0.00000752, -0.00000751, -0.00000119, 0.00000514, -0.00000537, -0.00000000, -0.00000125, -0.00000542, 0.00000234, 0.00000117, -0.00000187, 0.00000503, 0.00000454, -0.00002101, 0.00002865, -0.00003115, 0.00002425, -0.00001329, 0.00000815, -0.00001599, 0.00001573, -0.00000996, 0.00000874, -0.00001834, 0.00002494, -0.00001719, 0.00001166, -0.00002254, 0.00002344, -0.00002405, 0.00002207, -0.00002380, 0.00002187, -0.00001599, 0.00001870, -0.00002342, 0.00001703, -0.00002401, 0.00002583, -0.00002235, 0.00001800, -0.00002539, 0.00002826, -0.00003962, 0.00003108, -0.00003802, 0.00003822, -0.00004857, 0.00003469, -0.00006024, 0.00007403, -0.00008229, 0.00005187, -0.00009177, 0.00007585, -0.00016324, 0.00007702, -0.00034338, -0.00057949, 0.00246054, -0.00123582, 0.00018298, -0.00026097, 0.00011963, -0.00013411, 0.00007054, -0.00009844, 0.00006540, -0.00005144, 0.00004730, -0.00006346, 0.00004582, -0.00004248, 0.00003163, -0.00004189, 0.00003823, -0.00003591, 0.00002704, -0.00003252, 0.00002477, -0.00001851, 0.00001777, -0.00002370, 0.00002131, -0.00001790, 0.00001221, -0.00000958, 0.00001166, -0.00001107, 0.00000814, -0.00001199, 0.00001595, -0.00002014, 0.00000836, -0.00000649, 0.00001912, -0.00002083, 0.00001135, -0.00000740, 0.00000992, -0.00001459, 0.00000411, -0.00000070, 0.00000229, -0.00001652, 0.00002353, -0.00002908, 0.00003222, -0.00002070, 0.00002030, -0.00001642, 0.00001889, -0.00001980, 0.00001716, -0.00001238, 0.00000460, -0.00000841, 0.00001469, -0.00001340, 0.00002341, -0.00002706, 0.00001950, -0.00000317, -0.00001806, 0.00003041, },
		{ -0.00000705, 0.00000347, -0.00000402, 0.00000593, -0.00001223, 0.00000761, -0.00000143, -0.00000328, -0.00000044, -0.00000228, -0.00000554, 0.00000584, -0.00000273, 0.00000048, 0.00000859, -0.00000967, -0.00000365, 0.00001643, -0.00002844, 0.00003259, -0.00002650, 0.00001879, -0.00002521, 0.00002877, -0.00002402, 0.00001859, -0.00002561, 0.00003723, -0.00003506, 0.00002693, -0.00003631, 0.00004069, -0.00004469, 0.00004541, -0.00005067, 0.00005380, -0.00004865, 0.00004976, -0.00005860, 0.00005378, -0.00006014, 0.00006658, -0.00006844, 0.00006405, -0.00006919, 0.00007183, -0.00009083, 0.00008956, -0.00009891, 0.00010307, -0.00012433, 0.00011419, -0.00013872, 0.00016830, -0.00021225, 0.00020353, -0.00026689, 0.00028218, -0.00046466, 0.00049304, -0.00128603, 0.00132269, 0.00102907, -0.00117470, 0.00044064, -0.00042291, 0.00027583, -0.00026392, 0.00017992, -0.00018758, 0.00015767, -0.00013259, 0.00011047, -0.00012040, 0.00010534, -0.00009888, 0.00007920, -0.00008210, 0.00007984, -0.00007883, 0.00006678, -0.00007026, 0.00006448, -0.00005572, 0.00004871, -0.00005131, 0.00005098, -0.00004908, 0.00004206, -0.00003529, 0.00003418, -0.00003217, 0.00002537, -0.00002414, 0.00002600, -0.00003379, 0.00002363, -0.00001408, 0.00002298, -0.00003001, 0.00002392, -0.00001722, 0.00001592, -0.00002251, 0.00001323, -0.00000437, -0.00000447, -0.00000296, 0.00000925, -0.00001630, 0.00002693, -0.00002024, 0.00002073, -0.00001627, 0.00001843, -0.00002126, 0.00002283, -0.00002096, 0.00000944, -0.00000577, 0.00001077, -0.00000650, 0.00001599, -0.00002912, 0.00003892, -0.00004116, 0.00002761, -0.00000592, },
		{ 0.00001025, -0.00000924, 0.00000735, -0.00000634, -0.00000424, 0.00000698, -0.00000354, -0.00000394, 0.00000271, -0.00000914, 0.00000334, -0.00000070, 0.00000204, -0.00000286, 0.00001417, -0.00002742, 0.00002651, -0.00002010, 0.00000816, 0.00000294, -0.00000335, -0.00000533, 0.00000216, 0.00000362, -0.00000090, -0.00000912, 0.00000941, 0.00000173, -0.00000370, -0.00000773, 0.00000408, -0.00000129, -0.00000205, 0.00000121, -0.00000558, 0.00001173, -0.00000683, 0.00000244, -0.00001070, 0.00000502, -0.00000619, 0.00001067, -0.00001390, 0.00000761, -0.00000587, -0.00000272, -0.00001122, 0.00000854, -0.00001124, 0.00000468, -0.00002202, 0.00000268, -0.00000206, 0.00000644, -0.00004543, 0.00002064, -0.00005284, 0.00000071, -0.00013186, -0.00001190, -0.00094202, 0.00232110, -0.00068028, -0.00021831, -0.00002855, -0.00005831, 0.00000542, -0.00004671, -0.00000169, -0.00001908, 0.00001825, -0.00001311, -0.00000449, -0.00001030, 0.00000865, -0.00001448, -0.00000059, -0.00000083, 0.00000326, -0.00001069, 0.00000320, -0.00000963, 0.00001148, -0.00000857, 0.00000200, -0.00000340, 0.00000644, -0.00001072, 0.00000899, -0.00000367, 0.00000320, -0.00000405, -0.00000188, 0.00000585, -0.00000695, -0.00000459, 0.00000099, 0.00001217, -0.00000895, -0.00000150, 0.00000184, 0.00000304, -0.00000636, -0.00000342, 0.00000032, 0.00000686, -0.00002244, 0.00002387, -0.00002273, 0.00001979, -0.00000582, 0.00000780, -0.00000484, 0.00000908, -0.00000789, 0.00000470, 0.00000118, -0.00000716, -0.00000107, 0.00001034, -0.00000808, 0.00001858, -0.00001961, 0.00001233, 0.00000142, -0.00002131, 0.00003129, -0.00002426, },
		{ 0.00000118, -0.00000350, 0.00000459, -0.00000989, 0.00000205, 0.00000561, -0.00000812, 0.00000167, -0.00000291, -0.00000643, 0.00000449, -0.00000109, 0.00000008, 0.00000386, 0.00000167, -0.00001923, 0.00002976, -0.00003493, 0.00003171, -0.00002085, 0.00001530, -0.00002342, 0.00002425, -0.00001753, 0.00001584, -0.00002686, 0.00003486, -0.00002768, 0.00002186, -0.00003423, 0.00003656, -0.00003743, 0.00003622, -0.00004043, 0.00004032, -0.00003289, 0.00003541, -0.00004491, 0.00003911, -0.00004502, 0.00004871, -0.00004850, 0.00004504, -0.00005024, 0.00005374, -0.00007300, 0.00006869, -0.00007467, 0.00007692, -0.00009590, 0.00008503, -0.00010872, 0.00013015, -0.00016258, 0.00014624, -0.00019664, 0.00020083, -0.00035240, 0.00033858, -0.00093307, 0.00062098, 0.00154200, -0.00117829, 0.00032827, -0.00034946, 0.00021152, -0.00021015, 0.00013041, -0.00014591, 0.00011889, -0.00010181, 0.00008740, -0.00009774, 0.00008199, -0.00007645, 0.00005871, -0.00006505, 0.00006500, -0.00006304, 0.00005038, -0.00005379, 0.00004648, -0.00003935, 0.00003525, -0.00003841, 0.00003843, -0.00003554, 0.00002729, -0.00002265, 0.00002439, -0.00002339, 0.00001849, -0.00002029, 0.00002426, -0.00002950, 0.00001726, -0.00001253, 0.00002490, -0.00002840, 0.00001902, -0.00001365, 0.00001477, -0.00001979, 0.00000833, -0.00000366, 0.00000241, -0.00001726, 0.00002444, -0.00002886, 0.00003386, -0.00002185, 0.00002199, -0.00001696, 0.00002083, -0.00002175, 0.00002181, -0.00001654, 0.00000331, -0.00000186, 0.00001046, -0.00000622, 0.00001729, -0.00002805, 0.00003609, -0.00003912, 0.00002550, -0.00000528, -0.00000553, },
		{ -0.00001255, 0.00000866, -0.00000704, -0.00000403, 0.00000394, 0.00000301, -0.00001086, 0.00000812, -0.00001315, 0.00000467, -0.00000349, 0.00000643, -0.00000873, 0.00001762, -0.00002396, 0.00001378, -0.00000184, -0.00001007, 0.00001703, -0.00001148, 0.00000289, -0.00000860, 0.00001291, -0.00000672, -0.00000131, -0.00000481, 0.00001652, -0.00001389, 0.00000343, -0.00001189, 0.00001628, -0.00001899, 0.00001771, -0.00002289, 0.00002730, -0.00001964, 0.00001613, -0.00002619, 0.00002117, -0.00002398, 0.00002778, -0.00002943, 0.00002439, -0.00002385, 0.00001825, -0.00003658, 0.00003579, -0.00003915, 0.00003466, -0.00005412, 0.00003960, -0.00004792, 0.00005756, -0.00009782, 0.00007989, -0.00011805, 0.00008527, -0.00023353, 0.00014643, -0.00102071, 0.00189990, -0.00002892, -0.00056633, 0.00013239, -0.00017866, 0.00009570, -0.00012035, 0.00005620, -0.00006926, 0.00005852, -0.00004999, 0.00003411, -0.00004648, 0.00004035, -0.00004229, 0.00002428, -0.00002558, 0.00002888, -0.00003450, 0.00002477, -0.00002944, 0.00002711, -0.00002210, 0.00001611, -0.00001699, 0.00001957, -0.00002184, 0.00001620, -0.00000947, 0.00000950, -0.00000954, 0.00000338, -0.00000120, 0.00000316, -0.00001316, 0.00000516, 0.00000583, 0.00000230, -0.00001160, 0.00000778, -0.00000173, 0.00000007, -0.00000817, -0.00000041, 0.00000891, -0.00002040, 0.00001299, -0.00000609, 0.00000007, 0.00001255, -0.00000586, 0.00000750, -0.00000170, 0.00000478, -0.00000727, 0.00001257, -0.00001486, 0.00000214, 0.00000842, -0.00000386, 0.00001457, -0.00001324, 0.00000695, 0.00000448, -0.00002409, 0.00003262, -0.00002521, 0.00001239, },
		{ -0.00000571, 0.00000533, -0.00000923, -0.00000103, 0.00000799, -0.00000679, -0.00000214, 0.00000150, -0.00001105, 0.00000646, -0.00000368, 0.00000494, -0.00000550, 0.00001415, -0.00003027, 0.00003318, -0.00002967, 0.00002058, -0.00000838, 0.00000715, -0.00001648, 0.00001439, -0.00000732, 0.00000995, -0.00002236, 0.00002453, -0.00001362, 0.00001136, -0.00002503, 0.00002327, -0.00002067, 0.00001806, -0.00002144, 0.00001870, -0.00001071, 0.00001567, -0.00002532, 0.00001789, -0.00002282, 0.00002388, -0.00002270, 0.00002056, -0.00002533, 0.00002878, -0.00004635, 0.00003730, -0.00003948, 0.00003966, -0.00005570, 0.00004241, -0.00006184, 0.00007246, -0.00009343, 0.00006943, -0.00010633, 0.00009752, -0.00021579, 0.00015852, -0.00059160, 0.00008381, 0.00173995, -0.00091246, 0.00007631, -0.00019935, 0.00009246, -0.00011462, 0.00004856, -0.00007891, 0.00005991, -0.00005408, 0.00004702, -0.00005736, 0.00004259, -0.00004009, 0.00002611, -0.00003670, 0.00003838, -0.00003537, 0.00002314, -0.00002769, 0.00002075, -0.00001711, 0.00001633, -0.00001999, 0.00002060, -0.00001748, 0.00000988, -0.00000906, 0.00001392, -0.00001388, 0.00001097, -0.00001504, 0.00001937, -0.00002146, 0.00000883, -0.00000937, 0.00002281, -0.00002153, 0.00001050, -0.00000804, 0.00001147, -0.00001502, 0.00000367, -0.00000523, 0.00001034, -0.00002782, 0.00003012, -0.00002874, 0.00002741, -0.00001188, 0.00001384, -0.00000919, 0.00001473, -0.00001341, 0.00001245, -0.00000516, -0.00000702, 0.00000418, 0.00000801, -0.00000312, 0.00001551, -0.00002411, 0.00003155, -0.00003624, 0.00002253, -0.00000366, -0.00000589, 0.00000179, },
		{ 0.00000916, -0.00000758, -0.00000282, -0.00000057, 0.00000956, -0.00001557, 0.00001014, -0.00001334, 0.00000282, -0.00000358, 0.00000710, -0.00000836, 0.00001233, -0.00001146, -0.00000472, 0.00001619, -0.00002288, 0.00002306, -0.00001231, 0.00000551, -0.00001333, 0.00001522, -0.00000673, 0.00000249, -0.00001358, 0.00002272, -0.00001525, 0.00000718, -0.00001953, 0.00002293, -0.00002314, 0.00002045, -0.00002526, 0.00002690, -0.00001701, 0.00001530, -0.00002683, 0.00002156, -0.00002473, 0.00002675, -0.00002701, 0.00002349, -0.00002389, 0.00002068, -0.00004136, 0.00003957, -0.00004141, 0.00003727, -0.00005704, 0.00004463, -0.00005626, 0.00006381, -0.00009918, 0.00008122, -0.00011644, 0.00008909, -0.00023148, 0.00015456, -0.00089639, 0.00135435, 0.00054315, -0.00073944, 0.00015580, -0.00020173, 0.00010858, -0.00012947, 0.00005964, -0.00007345, 0.00006022, -0.00005487, 0.00004338, -0.00005389, 0.00004483, -0.00004512, 0.00002680, -0.00003012, 0.00003477, -0.00003842, 0.00002691, -0.00003043, 0.00002537, -0.00002059, 0.00001650, -0.00001736, 0.00001976, -0.00002035, 0.00001245, -0.00000675, 0.00000889, -0.00000931, 0.00000437, -0.00000484, 0.00000899, -0.00001646, 0.00000563, 0.00000126, 0.00001053, -0.00001668, 0.00000918, -0.00000361, 0.00000400, -0.00001022, -0.00000145, 0.00000695, -0.00001122, -0.00000427, 0.00001263, -0.00001733, 0.00002507, -0.00001268, 0.00001366, -0.00000680, 0.00001151, -0.00001218, 0.00001594, -0.00001414, -0.00000143, 0.00001075, -0.00000274, 0.00001366, -0.00001017, 0.00000502, 0.00000498, -0.00002563, 0.00003346, -0.00002646, 0.00001475, -0.00001521, },
		{ 0.00000758, -0.00001025, -0.00000114, 0.00000566, -0.00000063, -0.00000927, 0.00000770, -0.00001612, 0.00000876, -0.00000723, 0.00001051, -0.00001317, 0.00002056, -0.00003073, 0.00002415, -0.00001418, 0.00000354, 0.00000575, -0.00000117, -0.00000846, 0.00000367, 0.00000225, 0.00000436, -0.00001549, 0.00001114, 0.00000107, 0.00000121, -0.00001440, 0.00000770, -0.00000289, 0.00000050, -0.00000409, 0.00000008, 0.00000658, 0.00000133, -0.00001037, 0.00000101, -0.00000520, 0.00000448, -0.00000365, 0.00000302, -0.00000683, 0.00000957, -0.00002440, 0.00001053, -0.00001020, 0.00000947, -0.00002314, 0.00000734, -0.00002199, 0.00002350, -0.00003753, 0.00001000, -0.00003768, 0.00002066, -0.00011358, 0.00002833, -0.00035401, -0.00021880, 0.00163690, -0.00046968, -0.00021059, -0.00005375, -0.00001943, -0.00002787, -0.00002318, -0.00002240, 0.00001083, -0.00001448, 0.00001114, -0.00001982, 0.00000689, -0.00000825, -0.00000180, -0.00001224, 0.00001425, -0.00000997, -0.00000092, -0.00000562, 0.00000014, -0.00000056, 0.00000220, -0.00000561, 0.00000674, -0.00000385, -0.00000202, -0.00000120, 0.00000794, -0.00000774, 0.00000589, -0.00001068, 0.00001364, -0.00001254, 0.00000096, -0.00000609, 0.00001806, -0.00001185, 0.00000123, -0.00000257, 0.00000749, -0.00000943, -0.00000041, -0.00000700, 0.00001470, -0.00002922, 0.00002286, -0.00001585, 0.00001039, 0.00000514, 0.00000111, 0.00000252, 0.00000470, -0.00000136, 0.00000066, 0.00000682, -0.00001563, 0.00000785, 0.00000660, 0.00000011, 0.00001338, -0.00002007, 0.00002765, -0.00003402, 0.00001942, -0.00000121, -0.00000820, 0.00000497, -0.00000995, },
		{ -0.00000636, -0.00000278, -0.00000339, 0.00001231, -0.00001478, 0.00000657, -0.00000856, -0.00000277, -0.00000035, 0.00000342, -0.00000222, 0.00000129, 0.00000392, -0.00002053, 0.00002574, -0.00002442, 0.00001830, -0.00000522, 0.00000234, -0.00001179, 0.00001087, -0.00000166, 0.00000208, -0.00001565, 0.00001972, -0.00000836, 0.00000393, -0.00001850, 0.00001855, -0.00001546, 0.00001150, -0.00001582, 0.00001490, -0.00000426, 0.00000532, -0.00001779, 0.00001143, -0.00001420, 0.00001422, -0.00001389, 0.00001215, -0.00001296, 0.00001112, -0.00003192, 0.00002687, -0.00002606, 0.00002145, -0.00004022, 0.00002746, -0.00003790, 0.00003898, -0.00006722, 0.00004646, -0.00007463, 0.00004545, -0.00017052, 0.00008280, -0.00068959, 0.00084623, 0.00090398, -0.00068416, 0.00003966, -0.00013843, 0.00005632, -0.00008670, 0.00002131, -0.00004375, 0.00003411, -0.00003620, 0.00002956, -0.00003817, 0.00002798, -0.00002900, 0.00001259, -0.00001921, 0.00002500, -0.00002664, 0.00001435, -0.00001793, 0.00001208, -0.00000965, 0.00000820, -0.00000918, 0.00001174, -0.00001143, 0.00000322, -0.00000056, 0.00000549, -0.00000661, 0.00000332, -0.00000605, 0.00001085, -0.00001509, 0.00000314, -0.00000110, 0.00001441, -0.00001604, 0.00000622, -0.00000276, 0.00000526, -0.00000974, -0.00000257, 0.00000240, 0.00000023, -0.00001953, 0.00002370, -0.00002364, 0.00002502, -0.00000805, 0.00001024, -0.00000314, 0.00000941, -0.00000779, 0.00001023, -0.00000542, -0.00001028, 0.00001601, -0.00000422, 0.00001586, -0.00001034, 0.00000611, 0.00000352, -0.00002614, 0.00003338, -0.00002682, 0.00001548, -0.00001528, 0.00000720, },
		{ -0.00001133, -0.00000002, 0.00000137, 0.00000628, -0.00001495, 0.00001106, -0.00001762, 0.00000793, -0.00000812, 0.00001236, -0.00001436, 0.00001735, -0.00002021, 0.00000639, 0.00000469, -0.00001169, 0.00001492, -0.00000475, -0.00000358, -0.00000357, 0.00000725, 0.00000223, -0.00000992, -0.00000001, 0.00001054, -0.00000343, -0.00000763, -0.00000353, 0.00000821, -0.00000852, 0.00000376, -0.00000785, 0.00001201, -0.00000157, -0.00000583, -0.00000540, 0.00000130, -0.00000271, 0.00000220, -0.00000162, -0.00000069, 0.00000252, -0.00001455, -0.00000279, 0.00000334, -0.00000329, -0.00000913, -0.00000817, -0.00000334, 0.00000004, -0.00001319, -0.00001290, -0.00001093, -0.00000762, -0.00007092, -0.00002316, -0.00024364, -0.00030841, 0.00133559, 0.00000430, -0.00041384, 0.00001212, -0.00007315, 0.00001119, -0.00005399, -0.00000098, -0.00000863, 0.00000074, -0.00000555, 0.00000027, -0.00001098, 0.00000625, -0.00001367, -0.00000190, 0.00000252, 0.00000319, -0.00001210, 0.00000348, -0.00000748, 0.00000385, -0.00000168, -0.00000039, 0.00000178, 0.00000087, -0.00000491, -0.00000115, 0.00000763, -0.00000577, 0.00000386, -0.00000791, 0.00000847, -0.00000490, -0.00000446, -0.00000394, 0.00001243, -0.00000230, -0.00000584, 0.00000080, 0.00000422, -0.00000428, -0.00000359, -0.00000766, 0.00001354, -0.00002131, 0.00000626, 0.00000272, -0.00000834, 0.00002013, -0.00000807, 0.00001093, -0.00000227, 0.00000721, -0.00000640, 0.00001276, -0.00001727, 0.00000571, 0.00000870, 0.00000101, 0.00001357, -0.00001850, 0.00002639, -0.00003371, 0.00001703, 0.00000122, -0.00001123, 0.00000886, -0.00001403, 0.00000937, },
		{ -0.00000432, -0.00000350, 0.00001051, -0.00000898, -0.00000079, -0.00000143, -0.00000930, 0.00000359, -0.00000172, 0.00000539, -0.00000909, 0.00001428, -0.00002624, 0.00002227, -0.00001402, 0.00000511, 0.00000622, -0.00000347, -0.00000675, 0.00000308, 0.00000528, -0.00000031, -0.00001297, 0.00001077, 0.00000239, -0.00000222, -0.00001263, 0.00000797, -0.00000232, -0.00000162, -0.00000295, 0.00000057, 0.00000903, -0.00000465, -0.00000790, -0.00000010, -0.00000216, 0.00000069, -0.00000084, 0.00000091, -0.00000140, -0.00000006, -0.00001919, 0.00000998, -0.00000692, 0.00000217, -0.00001959, 0.00000531, -0.00001256, 0.00000681, -0.00002995, 0.00000770, -0.00002956, -0.00000192, -0.00010543, 0.00000643, -0.00050588, 0.00048371, 0.00100867, -0.00043502, -0.00015301, -0.00004308, -0.00001783, -0.00002848, -0.00002741, -0.00000702, 0.00000276, -0.00001249, 0.00000845, -0.00001431, 0.00000450, -0.00000782, -0.00000567, -0.00000418, 0.00001018, -0.00000979, -0.00000213, -0.00000270, -0.00000241, 0.00000139, -0.00000061, -0.00000008, 0.00000297, -0.00000245, -0.00000465, 0.00000356, 0.00000345, -0.00000450, 0.00000227, -0.00000600, 0.00000975, -0.00001060, -0.00000096, -0.00000154, 0.00001394, -0.00001068, 0.00000048, -0.00000022, 0.00000429, -0.00000702, -0.00000420, -0.00000192, 0.00000825, -0.00002577, 0.00002153, -0.00001621, 0.00001255, 0.00000593, -0.00000010, 0.00000667, 0.00000128, 0.00000243, -0.00000016, 0.00000613, -0.00001935, 0.00002035, -0.00000569, 0.00001903, -0.00001156, 0.00000792, 0.00000208, -0.00002682, 0.00003282, -0.00002610, 0.00001411, -0.00001260, 0.00000298, -0.00000351, },
		{ 0.00000022, -0.00000174, 0.00001000, -0.00001558, 0.00000891, -0.00001393, 0.00000295, -0.00000526, 0.00000925, -0.00000874, 0.00000666, -0.00000428, -0.00001166, 0.00001793, -0.00001779, 0.00001453, -0.00000113, -0.00000382, -0.00000526, 0.00000617, 0.00000458, -0.00000760, -0.00000574, 0.00001196, -0.00000074, -0.00000678, -0.00000738, 0.00000943, -0.00000655, 0.00000053, -0.00000436, 0.00000599, 0.00000563, -0.00001031, -0.00000239, -0.00000240, 0.00000108, -0.00000324, 0.00000426, -0.00000469, 0.00000588, -0.00001581, -0.00000298, 0.00000148, 0.00000057, -0.00001256, -0.00000502, -0.00000562, 0.00000192, -0.00001883, -0.00000313, -0.00002009, 0.00000455, -0.00007680, -0.00001267, -0.00022595, -0.00026785, 0.00097579, 0.00037091, -0.00045335, -0.00003322, -0.00005499, -0.00000533, -0.00003955, -0.00001637, 0.00000184, -0.00000836, -0.00000205, 0.00000057, -0.00000865, 0.00000185, -0.00000862, -0.00000637, 0.00000429, 0.00000258, -0.00000921, -0.00000086, -0.00000256, -0.00000258, 0.00000322, -0.00000317, 0.00000467, -0.00000203, -0.00000081, -0.00000628, 0.00001043, -0.00000601, 0.00000353, -0.00000620, 0.00000445, 0.00000019, -0.00000649, -0.00000361, 0.00000766, 0.00000446, -0.00000860, 0.00000087, 0.00000272, -0.00000078, -0.00000514, -0.00000748, 0.00000841, -0.00000890, -0.00001115, 0.00001670, -0.00001893, 0.00002479, -0.00000734, 0.00001087, -0.00000134, 0.00000773, -0.00000474, 0.00000984, -0.00001070, -0.00000216, 0.00001407, -0.00000081, 0.00001688, -0.00002009, 0.00002797, -0.00003523, 0.00001551, 0.00000302, -0.00001371, 0.00001152, -0.00001588, 0.00000891, -0.00001081, },
		{ -0.00000215, 0.00000632, -0.00000172, -0.00000754, 0.00000357, -0.00001275, 0.00000471, -0.00000452, 0.00000964, -0.00001321, 0.00001473, -0.00001971, 0.00000773, 0.00000296, -0.00000968, 0.00001574, -0.00000712, -0.00000240, -0.00000377, 0.00001004, -0.00000151, -0.00000903, 0.00000100, 0.00001121, -0.00000610, -0.00000716, -0.00000222, 0.00000842, -0.00001054, 0.00000480, -0.00000751, 0.00001481, -0.00000742, -0.00000408, -0.00000576, 0.00000360, -0.00000572, 0.00000425, -0.00000277, 0.00000336, -0.00000508, -0.00001194, -0.00000085, 0.00000454, -0.00000833, -0.00000817, -0.00000753, 0.00000305, -0.00001289, -0.00000959, -0.00001096, -0.00000765, -0.00002292, -0.00007237, -0.00002840, -0.00039958, 0.00029474, 0.00090313, -0.00009486, -0.00032485, 0.00001680, -0.00006555, 0.00000706, -0.00005549, 0.00001152, -0.00001330, -0.00000065, -0.00000480, 0.00000274, -0.00001105, 0.00000505, -0.00001604, 0.00000423, 0.00000061, 0.00000170, -0.00001226, 0.00000564, -0.00000944, 0.00000525, -0.00000367, 0.00000411, -0.00000094, 0.00000138, -0.00000691, 0.00000277, 0.00000452, -0.00000417, 0.00000200, -0.00000531, 0.00000684, -0.00000485, -0.00000503, -0.00000093, 0.00001042, -0.00000301, -0.00000553, 0.00000235, 0.00000217, -0.00000299, -0.00000639, -0.00000412, 0.00000980, -0.00002119, 0.00000781, -0.00000018, -0.00000517, 0.00002142, -0.00001004, 0.00001599, -0.00000632, 0.00001162, -0.00000815, 0.00001374, -0.00002304, 0.00001992, -0.00000440, 0.00002056, -0.00001104, 0.00000777, 0.00000276, -0.00002901, 0.00003265, -0.00002500, 0.00001155, -0.00000868, -0.00000163, -0.00000071, -0.00000616, },
		{ -0.00000225, 0.00000923, -0.00001091, 0.00000232, -0.00000703, -0.00000384, -0.00000081, 0.00000362, -0.00000022, -0.00000523, 0.00000873, -0.00002142, 0.00001914, -0.00001184, 0.00000475, 0.00000806, -0.00000772, -0.00000253, 0.00000062, 0.00000974, -0.00000787, -0.00000617, 0.00000653, 0.00000696, -0.00000993, -0.00000525, 0.00000294, 0.00000281, -0.00000907, 0.00000494, -0.00000498, 0.00001604, -0.00001730, 0.00000391, -0.00001003, 0.00000905, -0.00001252, 0.00001301, -0.00001148, 0.00001268, -0.00002152, 0.00000312, -0.00000800, 0.00001212, -0.00002371, 0.00000654, -0.00001795, 0.00001650, -0.00003859, 0.00002002, -0.00004267, 0.00003102, -0.00009995, 0.00001972, -0.00024188, -0.00019258, 0.00067597, 0.00055081, -0.00032050, -0.00016245, 0.00000632, -0.00005331, -0.00000115, -0.00004957, 0.00002567, -0.00002799, 0.00001087, -0.00001024, 0.00000552, -0.00001286, 0.00000500, -0.00001805, 0.00001295, -0.00000572, 0.00000144, -0.00001187, 0.00000785, -0.00001284, 0.00001056, -0.00000849, 0.00001036, -0.00000748, 0.00000510, -0.00001168, 0.00001244, -0.00000576, 0.00000320, -0.00000480, 0.00000179, 0.00000224, -0.00000512, -0.00000519, 0.00000485, 0.00000696, -0.00000647, -0.00000225, 0.00000323, 0.00000034, -0.00000436, -0.00000770, 0.00000274, 0.00000145, -0.00002098, 0.00001859, -0.00001638, 0.00001669, 0.00000372, 0.00000274, 0.00000693, 0.00000105, 0.00000402, 0.00000078, 0.00000032, -0.00001177, 0.00001966, -0.00000340, 0.00002182, -0.00002323, 0.00003063, -0.00003700, 0.00001392, 0.00000438, -0.00001518, 0.00001212, -0.00001494, 0.00000601, -0.00000847, -0.00000116, },
		{ 0.00000283, 0.00000301, -0.00000976, 0.00000328, -0.00001088, 0.00000139, -0.00000323, 0.00000845, -0.00000962, 0.00000610, -0.00000527, -0.00001029, 0.00001760, -0.00001814, 0.00001768, -0.00000498, -0.00000178, -0.00000648, 0.00001000, 0.00000054, -0.00000672, -0.00000539, 0.00001400, -0.00000453, -0.00000555, -0.00000740, 0.00001163, -0.00001063, 0.00000348, -0.00000604, 0.00001078, -0.00000159, -0.00000764, -0.00000384, 0.00000115, -0.00000321, 0.00000011, 0.00000197, 0.00000029, -0.00000235, -0.00001269, -0.00000211, 0.00000443, -0.00000620, -0.00000991, -0.00000643, 0.00000302, -0.00001318, -0.00001279, -0.00000381, -0.00001404, -0.00001241, -0.00007685, -0.00001612, -0.00036849, 0.00023823, 0.00069034, 0.00021313, -0.00039072, -0.00000281, -0.00006097, 0.00000145, -0.00004941, 0.00000211, -0.00000689, -0.00000654, -0.00000407, 0.00000593, -0.00001185, 0.00000382, -0.00001373, 0.00000173, 0.00000072, 0.00000313, -0.00001171, 0.00000351, -0.00000648, 0.00000044, 0.00000005, 0.00000225, 0.00000114, -0.00000068, -0.00000365, -0.00000199, 0.00000755, -0.00000478, 0.00000203, -0.00000411, 0.00000326, 0.00000031, -0.00000752, -0.00000051, 0.00000595, 0.00000390, -0.00000914, 0.00000321, 0.00000039, 0.00000080, -0.00000820, -0.00000435, 0.00000587, -0.00000981, -0.00000964, 0.00001453, -0.00001816, 0.00002945, -0.00001224, 0.00001857, -0.00000748, 0.00001406, -0.00000854, 0.00001326, -0.00001882, 0.00001368, 0.00000028, 0.00001936, -0.00000737, 0.00000442, 0.00000630, -0.00003302, 0.00003320, -0.00002422, 0.00000912, -0.00000549, -0.00000450, 0.00000019, -0.00000663, -0.00000141, },
		{ 0.00000558, -0.00000396, -0.00000471, -0.00000122, -0.00000843, 0.00000161, -0.00000056, 0.00000599, -0.00001191, 0.00001254, -0.00001883, 0.00000806, 0.00000282, -0.00000915, 0.00001758, -0.00001126, 0.00000118, -0.00000571, 0.00001428, -0.00000824, -0.00000388, -0.00000164, 0.00001488, -0.00001288, -0.00000131, -0.00000582, 0.00001276, -0.00001755, 0.00001227, -0.00001287, 0.00002198, -0.00001994, 0.00000698, -0.00001478, 0.00001393, -0.00001801, 0.00001716, -0.00001405, 0.00001608, -0.00002466, 0.00000776, -0.00001599, 0.00002107, -0.00003155, 0.00001479, -0.00002745, 0.00002867, -0.00005428, 0.00003578, -0.00005582, 0.00004645, -0.00011111, 0.00003815, -0.00024305, -0.00014880, 0.00049456, 0.00054491, -0.00007674, -0.00030466, 0.00005804, -0.00009375, 0.00002968, -0.00007446, 0.00004145, -0.00004081, 0.00001921, -0.00001981, 0.00001922, -0.00002569, 0.00001583, -0.00002679, 0.00001949, -0.00001306, 0.00001100, -0.00002068, 0.00001522, -0.00001915, 0.00001379, -0.00001079, 0.00001368, -0.00001047, 0.00000812, -0.00001345, 0.00001108, -0.00000361, 0.00000213, -0.00000357, 0.00000065, 0.00000145, -0.00000123, -0.00000796, 0.00000408, 0.00000541, -0.00000075, -0.00000704, 0.00000495, -0.00000066, -0.00000138, -0.00000909, -0.00000064, 0.00000535, -0.00001956, 0.00000788, -0.00000347, 0.00000077, 0.00001909, -0.00000759, 0.00001707, -0.00000727, 0.00001387, -0.00000801, 0.00000939, -0.00001762, 0.00002159, -0.00000405, 0.00002587, -0.00002525, 0.00003173, -0.00003691, 0.00001092, 0.00000609, -0.00001614, 0.00001128, -0.00001235, 0.00000220, -0.00000571, -0.00000318, -0.00000101, },
		{ 0.00000333, -0.00000650, -0.00000203, -0.00000498, -0.00000467, 0.00000060, 0.00000356, -0.00000150, -0.00000591, 0.00000884, -0.00002256, 0.00002214, -0.00001550, 0.00001029, 0.00000296, -0.00000495, -0.00000477, 0.00000534, 0.00000543, -0.00000665, -0.00000700, 0.00001025, 0.00000199, -0.00000773, -0.00000688, 0.00000714, -0.00000296, -0.00000481, 0.00000202, 0.00000084, 0.00000843, -0.00001454, 0.00000204, -0.00000588, 0.00000419, -0.00000865, 0.00001035, -0.00000619, 0.00000424, -0.00001810, 0.00000307, -0.00000360, 0.00000410, -0.00001994, 0.00000363, -0.00000761, -0.00000005, -0.00003122, 0.00001824, -0.00003566, 0.00001430, -0.00010057, 0.00001967, -0.00037511, 0.00024123, 0.00047900, 0.00039743, -0.00031650, -0.00009916, -0.00001402, -0.00003576, -0.00001863, -0.00002602, 0.00001363, -0.00002334, 0.00000604, -0.00000174, -0.00000087, -0.00000795, -0.00000252, -0.00000821, 0.00000779, -0.00000316, -0.00000317, -0.00000576, 0.00000258, -0.00000885, 0.00000675, -0.00000251, 0.00000632, -0.00000560, 0.00000178, -0.00000735, 0.00000988, -0.00000468, 0.00000161, -0.00000256, 0.00000012, 0.00000338, -0.00000734, -0.00000134, 0.00000259, 0.00000756, -0.00000865, 0.00000136, 0.00000013, 0.00000281, -0.00000824, -0.00000426, 0.00000013, 0.00000145, -0.00002160, 0.00001915, -0.00001965, 0.00002545, -0.00000411, 0.00001278, -0.00000085, 0.00000879, -0.00000119, 0.00000573, -0.00000898, 0.00000448, 0.00000615, 0.00001667, -0.00000137, -0.00000119, 0.00001145, -0.00003765, 0.00003383, -0.00002375, 0.00000748, -0.00000415, -0.00000469, -0.00000139, -0.00000517, -0.00000233, 0.00000183, },
		{ 0.00000125, -0.00000799, -0.00000017, -0.00000799, -0.00000028, -0.00000079, 0.00000680, -0.00001058, 0.00000631, -0.00000646, -0.00000915, 0.00001814, -0.00001939, 0.00002147, -0.00001042, 0.00000244, -0.00000923, 0.00001513, -0.00000638, -0.00000189, -0.00000817, 0.00001856, -0.00001203, 0.00000060, -0.00001175, 0.00001743, -0.00001928, 0.00001245, -0.00001300, 0.00001968, -0.00001539, 0.00000411, -0.00001357, 0.00001230, -0.00001634, 0.00001383, -0.00000994, 0.00001352, -0.00002218, 0.00000691, -0.00001737, 0.00002167, -0.00003012, 0.00001348, -0.00002700, 0.00002973, -0.00005566, 0.00003349, -0.00004906, 0.00004008, -0.00009843, 0.00002857, -0.00021154, -0.00015572, 0.00042456, 0.00041864, 0.00017934, -0.00038211, 0.00005473, -0.00009745, 0.00003104, -0.00007421, 0.00003642, -0.00003706, 0.00001498, -0.00002030, 0.00002386, -0.00002823, 0.00001652, -0.00002632, 0.00001830, -0.00001385, 0.00001360, -0.00002168, 0.00001466, -0.00001762, 0.00001015, -0.00000790, 0.00001250, -0.00000893, 0.00000656, -0.00001079, 0.00000658, -0.00000020, 0.00000092, -0.00000291, 0.00000103, -0.00000132, 0.00000356, -0.00001052, 0.00000433, 0.00000157, 0.00000583, -0.00001103, 0.00000627, -0.00000248, 0.00000251, -0.00001105, -0.00000130, 0.00000272, -0.00000955, -0.00000895, 0.00001101, -0.00001370, 0.00002997, -0.00001256, 0.00002244, -0.00001084, 0.00001851, -0.00001057, 0.00001148, -0.00001623, 0.00001797, -0.00000145, 0.00002742, -0.00002425, 0.00002954, -0.00003379, 0.00000588, 0.00000867, -0.00001732, 0.00001026, -0.00001004, -0.00000038, -0.00000441, -0.00000418, -0.00000001, 0.00000206, },
		{ -0.00000015, -0.00000888, 0.00000100, -0.00000963, 0.00000346, -0.00000112, 0.00000580, -0.00001433, 0.00001524, -0.00002331, 0.00001423, -0.00000297, -0.00000291, 0.00001282, -0.00000894, -0.00000105, -0.00000115, 0.00001058, -0.00000725, -0.00000531, 0.00000276, 0.00000991, -0.00001079, -0.00000337, -0.00000122, 0.00000724, -0.00001402, 0.00001019, -0.00000814, 0.00001601, -0.00001888, 0.00000638, -0.00001179, 0.00001034, -0.00001555, 0.00001604, -0.00001026, 0.00000917, -0.00002272, 0.00000874, -0.00001248, 0.00001449, -0.00002951, 0.00001343, -0.00001878, 0.00001452, -0.00005012, 0.00003805, -0.00005347, 0.00003607, -0.00011910, 0.00004896, -0.00037778, 0.00023986, 0.00034003, 0.00042984, -0.00013122, -0.00022508, 0.00003754, -0.00007587, 0.00001309, -0.00005280, 0.00003173, -0.00003785, 0.00001525, -0.00001148, 0.00001293, -0.00002109, 0.00000895, -0.00001789, 0.00001509, -0.00001078, 0.00000665, -0.00001522, 0.00001089, -0.00001625, 0.00001107, -0.00000571, 0.00001047, -0.00000933, 0.00000548, -0.00001007, 0.00000950, -0.00000298, 0.00000058, -0.00000117, -0.00000156, 0.00000353, -0.00000434, -0.00000369, 0.00000137, 0.00000705, -0.00000418, -0.00000275, 0.00000155, 0.00000241, -0.00000582, -0.00000537, -0.00000380, 0.00000699, -0.00002272, 0.00001109, -0.00001008, 0.00001200, 0.00001039, 0.00000276, 0.00000958, 0.00000006, 0.00000909, -0.00000340, 0.00000076, -0.00000256, 0.00000957, 0.00001498, 0.00000472, -0.00000660, 0.00001572, -0.00004086, 0.00003322, -0.00002290, 0.00000633, -0.00000432, -0.00000300, -0.00000428, -0.00000286, -0.00000425, 0.00000288, 0.00000032, },
		{ -0.00000584, -0.00000443, -0.00000294, -0.00000558, 0.00000223, 0.00000300, -0.00000334, -0.00000527, 0.00000813, -0.00002335, 0.00002570, -0.00001999, 0.00001659, -0.00000389, 0.00000011, -0.00000858, 0.00001148, -0.00000182, -0.00000163, -0.00001071, 0.00001633, -0.00000667, -0.00000070, -0.00001270, 0.00001490, -0.00001339, 0.00000558, -0.00000629, 0.00001102, -0.00000607, -0.00000253, -0.00000811, 0.00000586, -0.00000951, 0.00000553, -0.00000184, 0.00000733, -0.00001584, 0.00000163, -0.00001268, 0.00001468, -0.00002071, 0.00000407, -0.00001787, 0.00002052, -0.00004391, 0.00001596, -0.00002707, 0.00001802, -0.00006914, -0.00000041, -0.00015827, -0.00019449, 0.00041721, 0.00026176, 0.00035821, -0.00034877, -0.00001547, -0.00006239, 0.00000294, -0.00005039, 0.00001327, -0.00001964, 0.00000045, -0.00001233, 0.00001875, -0.00001992, 0.00000729, -0.00001736, 0.00001008, -0.00000823, 0.00000905, -0.00001507, 0.00000709, -0.00000990, 0.00000182, -0.00000188, 0.00000838, -0.00000429, 0.00000212, -0.00000583, 0.00000130, 0.00000252, 0.00000069, -0.00000320, 0.00000251, -0.00000463, 0.00000732, -0.00001133, 0.00000413, -0.00000218, 0.00001035, -0.00001196, 0.00000563, -0.00000347, 0.00000538, -0.00001195, -0.00000086, -0.00000301, 0.00000230, -0.00002255, 0.00001771, -0.00001888, 0.00003028, -0.00000791, 0.00001978, -0.00000675, 0.00001545, -0.00000522, 0.00000609, -0.00000849, 0.00001055, 0.00000305, 0.00002700, -0.00002036, 0.00002437, -0.00002833, -0.00000042, 0.00001174, -0.00001890, 0.00000989, -0.00000934, -0.00000049, -0.00000549, -0.00000329, -0.00000058, 0.00000080, 0.00000233, },
		{ -0.00001295, 0.00000305, -0.00001016, 0.00000253, -0.00000243, 0.00000830, -0.00001509, 0.00001137, -0.00001319, -0.00000183, 0.00001277, -0.00001487, 0.00001885, -0.00000978, 0.00000130, -0.00000594, 0.00001292, -0.00000630, -0.00000301, -0.00000440, 0.00001500, -0.00001134, -0.00000045, -0.00000855, 0.00001408, -0.00001828, 0.00001289, -0.00001094, 0.00001671, -0.00001714, 0.00000587, -0.00001293, 0.00001127, -0.00001660, 0.00001549, -0.00000880, 0.00000934, -0.00002307, 0.00001060, -0.00001676, 0.00001877, -0.00003216, 0.00001614, -0.00002276, 0.00002116, -0.00005832, 0.00004363, -0.00005527, 0.00004006, -0.00011811, 0.00005467, -0.00035406, 0.00020434, 0.00028916, 0.00034493, 0.00009145, -0.00031488, 0.00005213, -0.00009087, 0.00002412, -0.00006123, 0.00003430, -0.00003998, 0.00001550, -0.00001557, 0.00002115, -0.00002727, 0.00001322, -0.00002092, 0.00001697, -0.00001412, 0.00001184, -0.00001909, 0.00001322, -0.00001746, 0.00000983, -0.00000486, 0.00001118, -0.00000954, 0.00000552, -0.00000906, 0.00000635, -0.00000027, -0.00000032, -0.00000068, -0.00000126, 0.00000099, 0.00000051, -0.00000674, 0.00000192, 0.00000340, 0.00000231, -0.00000727, 0.00000353, 0.00000035, -0.00000169, -0.00000766, -0.00000463, 0.00000545, -0.00001406, -0.00000485, 0.00000337, -0.00000292, 0.00002327, -0.00000433, 0.00001736, -0.00000581, 0.00001590, -0.00000790, 0.00000488, -0.00000337, 0.00000807, 0.00001605, 0.00000896, -0.00000961, 0.00001705, -0.00004114, 0.00003045, -0.00002099, 0.00000491, -0.00000486, -0.00000122, -0.00000642, -0.00000156, -0.00000556, 0.00000415, -0.00000322, 0.00000537, },
		{ -0.00001123, 0.00000329, -0.00001086, 0.00000534, -0.00000187, 0.00000472, -0.00001516, 0.00001690, -0.00002746, 0.00002132, -0.00001018, 0.00000473, 0.00000573, -0.00000391, -0.00000520, 0.00000519, 0.00000365, -0.00000232, -0.00000974, 0.00000983, 0.00000083, -0.00000352, -0.00000998, 0.00000742, -0.00000354, -0.00000379, 0.00000220, 0.00000156, 0.00000251, -0.00000812, -0.00000280, -0.00000088, -0.00000244, -0.00000246, 0.00000511, 0.00000215, -0.00000986, -0.00000413, -0.00000612, 0.00000516, -0.00000925, -0.00000703, -0.00000675, 0.00000832, -0.00002782, -0.00000555, -0.00000327, -0.00000472, -0.00003996, -0.00002951, -0.00010673, -0.00023245, 0.00041751, 0.00014604, 0.00041568, -0.00021085, -0.00012408, -0.00001377, -0.00003487, -0.00001988, -0.00001338, -0.00000071, -0.00001486, -0.00000290, 0.00000951, -0.00000668, -0.00000551, -0.00000606, 0.00000029, -0.00000082, 0.00000168, -0.00000556, -0.00000239, -0.00000124, -0.00000619, 0.00000306, 0.00000475, 0.00000029, -0.00000206, -0.00000174, -0.00000208, 0.00000294, 0.00000196, -0.00000422, 0.00000410, -0.00000689, 0.00000849, -0.00000940, 0.00000246, -0.00000402, 0.00001100, -0.00000896, 0.00000255, -0.00000270, 0.00000586, -0.00001042, -0.00000136, -0.00000770, 0.00000956, -0.00002621, 0.00001234, -0.00001316, 0.00002035, 0.00000448, 0.00001143, 0.00000264, 0.00000745, 0.00000458, -0.00000249, 0.00000076, 0.00000380, 0.00000620, 0.00002673, -0.00001542, 0.00001826, -0.00002271, -0.00000614, 0.00001409, -0.00002030, 0.00001008, -0.00001025, 0.00000148, -0.00000819, -0.00000121, -0.00000246, 0.00000184, -0.00000029, 0.00000008, },
		{ -0.00000023, -0.00000582, -0.00000217, -0.00000016, 0.00000566, -0.00000896, 0.00000055, 0.00000144, -0.00001755, 0.00002322, -0.00001892, 0.00001696, -0.00000528, 0.00000041, -0.00000695, 0.00001096, -0.00000279, -0.00000194, -0.00000839, 0.00001492, -0.00000794, -0.00000003, -0.00001180, 0.00001447, -0.00001534, 0.00000873, -0.00000690, 0.00001082, -0.00001027, 0.00000128, -0.00000965, 0.00000723, -0.00001229, 0.00000969, -0.00000305, 0.00000568, -0.00001955, 0.00000821, -0.00001535, 0.00001583, -0.00002706, 0.00001095, -0.00001834, 0.00001793, -0.00005373, 0.00003398, -0.00004142, 0.00002756, -0.00009933, 0.00003899, -0.00030665, 0.00014042, 0.00029788, 0.00021215, 0.00027075, -0.00031882, 0.00000992, -0.00007158, 0.00000861, -0.00004789, 0.00001984, -0.00002924, 0.00000624, -0.00001228, 0.00002089, -0.00002374, 0.00000841, -0.00001607, 0.00001240, -0.00001176, 0.00001064, -0.00001595, 0.00000889, -0.00001259, 0.00000384, -0.00000081, 0.00000900, -0.00000669, 0.00000266, -0.00000565, 0.00000223, 0.00000187, -0.00000013, -0.00000148, 0.00000068, -0.00000293, 0.00000533, -0.00000889, 0.00000278, -0.00000109, 0.00000789, -0.00000980, 0.00000436, -0.00000163, 0.00000222, -0.00000959, -0.00000367, -0.00000036, -0.00000168, -0.00001972, 0.00001144, -0.00001110, 0.00002757, -0.00000311, 0.00001804, -0.00000467, 0.00001551, -0.00000479, 0.00000160, 0.00000239, 0.00000234, 0.00001929, 0.00001122, -0.00000971, 0.00001510, -0.00003860, 0.00002585, -0.00001811, 0.00000287, -0.00000476, -0.00000088, -0.00000629, -0.00000246, -0.00000514, 0.00000405, -0.00000484, 0.00000720, -0.00001190, },
		{ 0.00000658, -0.00001251, 0.00000539, -0.00000425, 0.00000894, -0.00001821, 0.00001577, -0.00002024, 0.00000738, 0.00000498, -0.00000810, 0.00001327, -0.00000598, -0.00000215, -0.00000047, 0.00000715, -0.00000201, -0.00000746, 0.00000257, 0.00000674, -0.00000495, -0.00000655, -0.00000069, 0.00000483, -0.00001002, 0.00000693, -0.00000337, 0.00000568, -0.00000887, -0.00000118, -0.00000411, 0.00000075, -0.00000597, 0.00000710, 0.00000130, -0.00000742, -0.00000697, -0.00000179, -0.00000169, -0.00000168, -0.00001351, -0.00000029, 0.00000058, -0.00001651, -0.00002002, 0.00001029, -0.00001558, -0.00002422, -0.00004373, -0.00007472, -0.00024649, 0.00039202, 0.00010053, 0.00036330, -0.00002024, -0.00021811, 0.00001259, -0.00005743, -0.00000240, -0.00002771, 0.00000742, -0.00002154, 0.00000069, 0.00000299, 0.00000402, -0.00001429, 0.00000069, -0.00000513, 0.00000323, -0.00000345, 0.00000142, -0.00000834, 0.00000332, -0.00000959, 0.00000371, 0.00000412, 0.00000243, -0.00000364, -0.00000049, -0.00000235, 0.00000093, 0.00000414, -0.00000503, 0.00000471, -0.00000702, 0.00000659, -0.00000499, -0.00000053, -0.00000364, 0.00000798, -0.00000313, -0.00000186, -0.00000061, 0.00000408, -0.00000659, -0.00000344, -0.00000920, 0.00000958, -0.00001937, -0.00000207, -0.00000189, 0.00000650, 0.00001788, 0.00000365, 0.00001149, 0.00000038, 0.00001274, -0.00000813, 0.00000590, 0.00000212, 0.00000522, 0.00002859, -0.00001148, 0.00001354, -0.00001918, -0.00000953, 0.00001459, -0.00002073, 0.00001010, -0.00001174, 0.00000389, -0.00001058, 0.00000034, -0.00000418, 0.00000359, -0.00000408, 0.00000611, -0.00001633, },
		{ 0.00000067, -0.00000780, 0.00000315, 0.00000078, -0.00000042, -0.00001056, 0.00001239, -0.00002504, 0.00002284, -0.00001263, 0.00000738, 0.00000317, -0.00000279, -0.00000472, 0.00000573, 0.00000215, -0.00000196, -0.00000888, 0.00001035, -0.00000197, -0.00000145, -0.00001106, 0.00000924, -0.00000739, 0.00000082, 0.00000027, 0.00000264, -0.00000264, -0.00000360, -0.00000529, 0.00000166, -0.00000642, 0.00000278, 0.00000294, 0.00000176, -0.00001523, 0.00000425, -0.00001090, 0.00000896, -0.00001821, 0.00000222, -0.00001000, 0.00000957, -0.00004226, 0.00001713, -0.00002186, 0.00000987, -0.00007542, 0.00001661, -0.00025355, 0.00007274, 0.00032139, 0.00009836, 0.00035494, -0.00022768, -0.00007478, -0.00003317, -0.00002112, -0.00002387, -0.00000161, -0.00001408, -0.00000601, -0.00000591, 0.00001521, -0.00001387, -0.00000150, -0.00000740, 0.00000497, -0.00000655, 0.00000561, -0.00000877, 0.00000136, -0.00000544, -0.00000309, 0.00000322, 0.00000646, -0.00000313, -0.00000066, -0.00000238, -0.00000062, 0.00000208, 0.00000154, -0.00000318, 0.00000313, -0.00000637, 0.00000809, -0.00000864, 0.00000243, -0.00000414, 0.00001017, -0.00000877, 0.00000290, -0.00000201, 0.00000402, -0.00000935, -0.00000321, -0.00000607, 0.00000745, -0.00002591, 0.00000855, -0.00000930, 0.00002168, 0.00000650, 0.00001217, 0.00000270, 0.00000916, 0.00000389, -0.00000610, 0.00001095, -0.00000388, 0.00002203, 0.00001315, -0.00000824, 0.00001140, -0.00003497, 0.00002094, -0.00001524, 0.00000063, -0.00000392, -0.00000229, -0.00000390, -0.00000521, -0.00000328, 0.00000245, -0.00000402, 0.00000513, -0.00000658, -0.00000685, },
		{ -0.00000930, 0.00000160, -0.00000300, 0.00000788, -0.00001380, 0.00000649, -0.00000657, -0.00000873, 0.00001741, -0.00001502, 0.00001404, -0.00000334, -0.00000177, -0.00000292, 0.00000670, 0.00000060, -0.00000585, -0.00000244, 0.00000840, -0.00000321, -0.00000470, -0.00000592, 0.00000787, -0.00000991, 0.00000546, -0.00000195, 0.00000253, -0.00000458, -0.00000352, -0.00000314, -0.00000075, -0.00000437, 0.00000401, 0.00000462, -0.00000860, -0.00000634, -0.00000105, -0.00000372, -0.00000040, -0.00001310, -0.00000087, 0.00000024, -0.00001382, -0.00002310, 0.00000983, -0.00001173, -0.00002426, -0.00004065, -0.00006485, -0.00023366, 0.00033761, 0.00011231, 0.00025202, 0.00015352, -0.00024871, -0.00000692, -0.00005152, -0.00000737, -0.00002319, 0.00000059, -0.00001692, -0.00000378, 0.00000223, 0.00000802, -0.00001503, -0.00000029, -0.00000379, 0.00000185, -0.00000399, 0.00000319, -0.00000833, 0.00000201, -0.00000744, -0.00000009, 0.00000644, 0.00000199, -0.00000247, -0.00000191, -0.00000033, -0.00000212, 0.00000581, -0.00000464, 0.00000374, -0.00000502, 0.00000252, 0.00000034, -0.00000349, -0.00000222, 0.00000323, 0.00000293, -0.00000541, 0.00000123, 0.00000158, -0.00000216, -0.00000584, -0.00000828, 0.00000430, -0.00000715, -0.00001738, 0.00000625, -0.00000330, 0.00002503, 0.00000246, 0.00001462, -0.00000082, 0.00001464, -0.00000689, 0.00000407, 0.00000689, -0.00000009, 0.00003265, -0.00000918, 0.00001122, -0.00001860, -0.00001017, 0.00001314, -0.00001998, 0.00000942, -0.00001271, 0.00000511, -0.00001099, -0.00000007, -0.00000442, 0.00000429, -0.00000645, 0.00000906, -0.00001682, 0.00000479, },
		{ -0.00001074, 0.00000428, -0.00000264, 0.00000552, -0.00001602, 0.00001429, -0.00002135, 0.00001208, 0.00000067, -0.00000512, 0.00001115, -0.00000519, -0.00000183, 0.00000011, 0.00000591, -0.00000146, -0.00000754, 0.00000412, 0.00000346, -0.00000244, -0.00000852, 0.00000187, 0.00000086, -0.00000574, 0.00000544, -0.00000273, 0.00000117, -0.00000502, -0.00000331, -0.00000178, -0.00000296, -0.00000113, 0.00000543, 0.00000082, -0.00001313, 0.00000176, -0.00000703, 0.00000276, -0.00001096, -0.00000435, -0.00000373, 0.00000282, -0.00003221, 0.00000328, -0.00000801, -0.00000093, -0.00005934, 0.00000231, -0.00021247, 0.00002385, 0.00032583, 0.00004002, 0.00033892, -0.00007514, -0.00016106, -0.00000499, -0.00004398, -0.00000620, -0.00001613, -0.00000544, -0.00001303, -0.00000263, 0.00000968, -0.00000390, -0.00001002, -0.00000089, -0.00000017, -0.00000286, 0.00000101, -0.00000221, -0.00000454, -0.00000061, -0.00000698, 0.00000420, 0.00000589, -0.00000105, -0.00000229, -0.00000116, -0.00000099, 0.00000019, 0.00000405, -0.00000472, 0.00000468, -0.00000776, 0.00000767, -0.00000553, 0.00000039, -0.00000477, 0.00000857, -0.00000443, -0.00000054, -0.00000055, 0.00000316, -0.00000643, -0.00000449, -0.00000881, 0.00000957, -0.00002135, -0.00000393, -0.00000102, 0.00001008, 0.00001927, 0.00000487, 0.00001145, 0.00000196, 0.00001266, -0.00001252, 0.00001686, -0.00000616, 0.00002133, 0.00001678, -0.00000717, 0.00000818, -0.00003246, 0.00001753, -0.00001361, -0.00000101, -0.00000298, -0.00000462, -0.00000063, -0.00000815, -0.00000148, 0.00000055, -0.00000201, 0.00000124, -0.00000024, -0.00001289, 0.00000529, },
		{ -0.00000393, 0.00000006, 0.00000357, -0.00000552, -0.00000501, 0.00000565, -0.00001881, 0.00002062, -0.00001224, 0.00000694, 0.00000372, -0.00000400, -0.00000193, 0.00000261, 0.00000494, -0.00000528, -0.00000427, 0.00000572, 0.00000112, -0.00000457, -0.00000734, 0.00000516, -0.00000428, -0.00000048, 0.00000345, -0.00000390, 0.00000151, -0.00000706, -0.00000031, -0.00000458, -0.00000036, -0.00000112, 0.00000912, -0.00001083, -0.00000428, -0.00000242, -0.00000222, -0.00000369, -0.00000808, -0.00000599, 0.00000484, -0.00001730, -0.00001803, 0.00000035, 0.00000019, -0.00003235, -0.00002895, -0.00006689, -0.00020700, 0.00027208, 0.00014635, 0.00014129, 0.00025650, -0.00019655, -0.00006858, -0.00002438, -0.00002843, -0.00000610, -0.00001499, -0.00000640, -0.00001231, 0.00000509, 0.00000632, -0.00000902, -0.00000698, 0.00000204, -0.00000293, -0.00000134, 0.00000084, -0.00000383, -0.00000313, -0.00000225, -0.00000562, 0.00000937, 0.00000071, -0.00000015, -0.00000421, 0.00000198, -0.00000431, 0.00000576, -0.00000267, 0.00000155, -0.00000197, -0.00000181, 0.00000440, -0.00000462, -0.00000149, -0.00000078, 0.00000655, -0.00000609, 0.00000122, 0.00000023, 0.00000066, -0.00000653, -0.00000729, -0.00000196, 0.00000337, -0.00002548, 0.00000493, -0.00000464, 0.00002281, 0.00000957, 0.00001102, 0.00000459, 0.00001009, 0.00000060, -0.00000303, 0.00001541, -0.00000681, 0.00003683, -0.00000737, 0.00001048, -0.00002001, -0.00000929, 0.00001085, -0.00001875, 0.00000824, -0.00001283, 0.00000451, -0.00000903, -0.00000247, -0.00000310, 0.00000342, -0.00000638, 0.00000809, -0.00001279, -0.00000093, -0.00000314, },
		{ 0.00000175, -0.00000273, 0.00000641, -0.00001390, 0.00000731, -0.00000972, -0.00000305, 0.00001374, -0.00001382, 0.00001376, -0.00000371, -0.00000086, -0.00000307, 0.00000623, 0.00000101, -0.00000633, -0.00000078, 0.00000565, -0.00000117, -0.00000646, -0.00000403, 0.00000521, -0.00000717, 0.00000544, -0.00000265, -0.00000061, -0.00000195, -0.00000475, -0.00000166, -0.00000351, -0.00000064, 0.00000345, 0.00000352, -0.00001401, 0.00000192, -0.00000567, 0.00000002, -0.00000846, -0.00000560, -0.00000289, 0.00000166, -0.00002861, -0.00000170, -0.00000552, 0.00000027, -0.00005608, 0.00000168, -0.00018991, 0.00000114, 0.00030164, 0.00003625, 0.00025739, 0.00008132, -0.00020379, -0.00001159, -0.00004498, -0.00000610, -0.00001545, -0.00000899, -0.00001090, -0.00000552, 0.00000794, 0.00000153, -0.00001254, -0.00000037, 0.00000001, -0.00000327, -0.00000039, 0.00000062, -0.00000586, -0.00000051, -0.00000617, 0.00000134, 0.00000777, -0.00000103, -0.00000165, -0.00000221, 0.00000066, -0.00000269, 0.00000606, -0.00000496, 0.00000445, -0.00000664, 0.00000442, -0.00000067, -0.00000239, -0.00000371, 0.00000447, 0.00000118, -0.00000420, 0.00000152, 0.00000080, -0.00000215, -0.00000670, -0.00000866, 0.00000561, -0.00001002, -0.00001891, 0.00000604, -0.00000002, 0.00002825, 0.00000221, 0.00001610, -0.00000078, 0.00001642, -0.00001303, 0.00001648, -0.00000234, 0.00001632, 0.00002270, -0.00000751, 0.00000677, -0.00003234, 0.00001641, -0.00001366, -0.00000160, -0.00000257, -0.00000674, 0.00000184, -0.00000950, -0.00000128, -0.00000024, -0.00000062, -0.00000182, 0.00000379, -0.00001507, 0.00000536, -0.00000908, },
		{ 0.00000188, -0.00000043, 0.00000160, -0.00001239, 0.00001017, -0.00001846, 0.00001330, -0.00000138, -0.00000459, 0.00001165, -0.00000636, 0.00000054, -0.00000272, 0.00000876, -0.00000447, -0.00000404, 0.00000087, 0.00000560, -0.00000460, -0.00000625, -0.00000080, 0.00000306, -0.00000655, 0.00000833, -0.00000901, 0.00000518, -0.00000836, 0.00000127, -0.00000743, 0.00000247, -0.00000454, 0.00001133, -0.00001115, -0.00000343, -0.00000335, -0.00000008, -0.00000781, -0.00000286, -0.00001091, 0.00000940, -0.00002137, -0.00001177, -0.00000933, 0.00001014, -0.00003773, -0.00002026, -0.00006738, -0.00018311, 0.00021658, 0.00017008, 0.00007111, 0.00027170, -0.00008039, -0.00014254, 0.00000123, -0.00004884, 0.00000965, -0.00002771, 0.00000083, -0.00001806, 0.00000678, 0.00000305, -0.00000104, -0.00001422, 0.00000748, -0.00000691, 0.00000107, -0.00000235, 0.00000145, -0.00000813, 0.00000204, -0.00000936, 0.00001025, 0.00000063, 0.00000140, -0.00000549, 0.00000285, -0.00000450, 0.00000378, 0.00000022, -0.00000070, 0.00000051, -0.00000447, 0.00000550, -0.00000290, -0.00000249, -0.00000249, 0.00000642, -0.00000337, -0.00000106, 0.00000095, 0.00000069, -0.00000442, -0.00000795, -0.00000573, 0.00000747, -0.00002302, -0.00000595, 0.00000065, 0.00001382, 0.00002144, 0.00000440, 0.00001291, 0.00000316, 0.00000975, -0.00001037, 0.00002269, -0.00001073, 0.00003828, -0.00000412, 0.00000952, -0.00002139, -0.00000893, 0.00000942, -0.00001821, 0.00000732, -0.00001258, 0.00000267, -0.00000577, -0.00000553, -0.00000143, 0.00000188, -0.00000473, 0.00000476, -0.00000692, -0.00000714, 0.00000225, -0.00000776, },
		{ -0.00000044, 0.00000329, -0.00000668, -0.00000362, 0.00000265, -0.00001449, 0.00001956, -0.00001395, 0.00000878, 0.00000229, -0.00000252, -0.00000284, 0.00000278, 0.00000536, -0.00000592, -0.00000304, 0.00000397, 0.00000236, -0.00000561, -0.00000675, 0.00000421, -0.00000313, 0.00000081, 0.00000170, -0.00000613, 0.00000345, -0.00000783, 0.00000063, -0.00000666, 0.00000259, -0.00000097, 0.00000779, -0.00001620, 0.00000357, -0.00000628, 0.00000057, -0.00001033, -0.00000228, -0.00000661, 0.00000541, -0.00003102, 0.00000129, -0.00001212, 0.00000988, -0.00006126, 0.00000988, -0.00018017, -0.00000334, 0.00026016, 0.00006132, 0.00016102, 0.00018920, -0.00017769, -0.00005678, -0.00002532, -0.00002189, -0.00000205, -0.00002162, -0.00000261, -0.00001250, 0.00000947, 0.00000184, -0.00000875, -0.00000540, 0.00000463, -0.00000696, 0.00000112, -0.00000035, -0.00000287, -0.00000427, -0.00000212, -0.00000357, 0.00001052, -0.00000197, 0.00000022, -0.00000426, 0.00000278, -0.00000492, 0.00000643, -0.00000347, 0.00000263, -0.00000390, 0.00000009, 0.00000383, -0.00000410, -0.00000269, 0.00000030, 0.00000533, -0.00000588, 0.00000241, -0.00000101, 0.00000117, -0.00000775, -0.00000781, -0.00000041, 0.00000116, -0.00002819, 0.00000526, -0.00000359, 0.00002922, 0.00000718, 0.00001435, 0.00000295, 0.00001369, -0.00000702, 0.00001029, 0.00000600, 0.00000905, 0.00002944, -0.00000860, 0.00000686, -0.00003411, 0.00001685, -0.00001468, -0.00000154, -0.00000270, -0.00000811, 0.00000263, -0.00000856, -0.00000309, 0.00000051, -0.00000069, -0.00000261, 0.00000409, -0.00001274, 0.00000118, -0.00000394, -0.00000310, },
		{ -0.00000167, 0.00000427, -0.00001252, 0.00000561, -0.00000925, -0.00000010, 0.00001145, -0.00001438, 0.00001565, -0.00000574, 0.00000184, -0.00000638, 0.00000966, -0.00000192, -0.00000356, -0.00000342, 0.00000770, -0.00000316, -0.00000465, -0.00000664, 0.00000795, -0.00000879, 0.00000919, -0.00000970, 0.00000411, -0.00000583, 0.00000006, -0.00000745, 0.00000209, -0.00000435, 0.00000978, -0.00000845, -0.00000494, -0.00000243, 0.00000058, -0.00000979, -0.00000089, -0.00001207, 0.00001015, -0.00002163, -0.00000987, -0.00001268, 0.00001149, -0.00003403, -0.00002105, -0.00005886, -0.00017117, 0.00018188, 0.00016917, 0.00005001, 0.00021975, 0.00005436, -0.00018913, 0.00000188, -0.00005349, 0.00001243, -0.00002870, -0.00000153, -0.00001669, 0.00000399, 0.00000183, 0.00000435, -0.00001736, 0.00000855, -0.00000690, 0.00000059, -0.00000361, 0.00000437, -0.00000987, 0.00000275, -0.00000932, 0.00000793, 0.00000250, 0.00000138, -0.00000497, 0.00000184, -0.00000289, 0.00000084, 0.00000273, -0.00000173, 0.00000122, -0.00000451, 0.00000341, 0.00000119, -0.00000480, -0.00000208, 0.00000329, 0.00000138, -0.00000442, 0.00000299, -0.00000139, -0.00000034, -0.00000995, -0.00000639, 0.00000504, -0.00001279, -0.00002045, 0.00000609, 0.00000415, 0.00003185, 0.00000046, 0.00001884, -0.00000084, 0.00001530, -0.00001303, 0.00002460, -0.00000909, 0.00003556, 0.00000158, 0.00000708, -0.00002120, -0.00001057, 0.00000992, -0.00001907, 0.00000724, -0.00001264, 0.00000068, -0.00000283, -0.00000748, -0.00000095, 0.00000105, -0.00000321, 0.00000163, -0.00000252, -0.00001040, 0.00000384, -0.00000753, 0.00000051, },
		{ -0.00000065, 0.00000087, -0.00001215, 0.00000897, -0.00001677, 0.00001531, -0.00000552, -0.00000171, 0.00001020, -0.00000501, -0.00000056, -0.00000258, 0.00000969, -0.00000537, -0.00000332, -0.00000003, 0.00000632, -0.00000512, -0.00000672, -0.00000067, 0.00000359, -0.00000502, 0.00000664, -0.00001146, 0.00000751, -0.00000958, 0.00000246, -0.00000964, 0.00000549, -0.00000460, 0.00001064, -0.00001719, 0.00000456, -0.00000698, 0.00000234, -0.00001382, 0.00000223, -0.00001128, 0.00001030, -0.00003524, 0.00000686, -0.00002076, 0.00001968, -0.00006601, 0.00001742, -0.00017219, -0.00000383, 0.00022003, 0.00008545, 0.00009041, 0.00022361, -0.00008971, -0.00012032, -0.00000164, -0.00004106, 0.00001335, -0.00003410, 0.00000449, -0.00001821, 0.00001083, -0.00000025, -0.00000197, -0.00001216, 0.00000996, -0.00001073, 0.00000307, -0.00000282, 0.00000173, -0.00000884, 0.00000201, -0.00000760, 0.00001191, -0.00000220, 0.00000178, -0.00000571, 0.00000384, -0.00000544, 0.00000494, -0.00000083, 0.00000036, -0.00000120, -0.00000319, 0.00000588, -0.00000324, -0.00000318, -0.00000198, 0.00000620, -0.00000443, 0.00000107, -0.00000083, 0.00000182, -0.00000607, -0.00000836, -0.00000464, 0.00000677, -0.00002742, -0.00000457, -0.00000107, 0.00002311, 0.00001794, 0.00000844, 0.00001079, 0.00000736, 0.00000200, 0.00000244, 0.00001450, 0.00000334, 0.00003436, -0.00000870, 0.00000687, -0.00003606, 0.00001706, -0.00001516, -0.00000188, -0.00000274, -0.00000902, 0.00000205, -0.00000604, -0.00000589, 0.00000192, -0.00000169, -0.00000159, 0.00000168, -0.00000794, -0.00000438, 0.00000144, -0.00000720, 0.00000152, },
		{ 0.00000254, -0.00000669, -0.00000434, 0.00000265, -0.00001209, 0.00001915, -0.00001703, 0.00001266, -0.00000076, 0.00000069, -0.00000685, 0.00000676, 0.00000262, -0.00000366, -0.00000562, 0.00000651, 0.00000005, -0.00000360, -0.00001014, 0.00000819, -0.00000595, 0.00000560, -0.00000616, -0.00000081, -0.00000074, -0.00000296, -0.00000522, -0.00000095, -0.00000135, 0.00000566, -0.00000398, -0.00000784, -0.00000014, -0.00000041, -0.00000896, -0.00000275, -0.00000913, 0.00000672, -0.00001739, -0.00001339, -0.00000880, 0.00000437, -0.00002239, -0.00002967, -0.00004292, -0.00016955, 0.00016491, 0.00014886, 0.00006109, 0.00014187, 0.00015921, -0.00018045, -0.00003113, -0.00003942, 0.00000053, -0.00001762, -0.00001264, -0.00000938, -0.00000254, 0.00000321, 0.00000553, -0.00001499, 0.00000457, -0.00000279, -0.00000281, -0.00000247, 0.00000410, -0.00000778, -0.00000012, -0.00000611, 0.00000346, 0.00000534, 0.00000043, -0.00000324, -0.00000020, -0.00000073, -0.00000163, 0.00000372, -0.00000095, 0.00000007, -0.00000246, -0.00000051, 0.00000575, -0.00000675, -0.00000112, -0.00000066, 0.00000565, -0.00000676, 0.00000460, -0.00000359, 0.00000345, -0.00001135, -0.00000583, -0.00000042, -0.00000129, -0.00003059, 0.00000524, -0.00000111, 0.00003605, 0.00000305, 0.00001915, 0.00000110, 0.00001474, -0.00000929, 0.00002053, -0.00000246, 0.00002990, 0.00000885, 0.00000334, -0.00001934, -0.00001418, 0.00001203, -0.00002094, 0.00000787, -0.00001319, -0.00000077, -0.00000129, -0.00000735, -0.00000234, 0.00000162, -0.00000294, 0.00000043, -0.00000145, -0.00000942, 0.00000119, -0.00000358, -0.00000349, -0.00000028, },
		{ 0.00000491, -0.00001418, 0.00000653, -0.00000983, 0.00000382, 0.00000667, -0.00001229, 0.00001554, -0.00000537, 0.00000126, -0.00000700, 0.00001156, -0.00000320, -0.00000319, -0.00000393, 0.00000853, -0.00000369, -0.00000545, -0.00000655, 0.00000912, -0.00000820, 0.00000863, -0.00001339, 0.00000768, -0.00000815, 0.00000210, -0.00001043, 0.00000586, -0.00000529, 0.00001035, -0.00001562, 0.00000362, -0.00000633, 0.00000333, -0.00001611, 0.00000466, -0.00001344, 0.00001265, -0.00003696, 0.00000960, -0.00002494, 0.00002289, -0.00006365, 0.00001743, -0.00015801, -0.00001061, 0.00019388, 0.00009133, 0.00006098, 0.00019347, 0.00002605, -0.00016886, 0.00000508, -0.00004959, 0.00001966, -0.00003776, 0.00000414, -0.00001837, 0.00000893, -0.00000154, 0.00000370, -0.00001634, 0.00001222, -0.00001155, 0.00000301, -0.00000433, 0.00000506, -0.00001127, 0.00000358, -0.00000868, 0.00001065, -0.00000088, 0.00000216, -0.00000572, 0.00000330, -0.00000434, 0.00000241, 0.00000167, -0.00000095, -0.00000001, -0.00000404, 0.00000468, 0.00000030, -0.00000529, -0.00000197, 0.00000387, -0.00000049, -0.00000196, 0.00000123, -0.00000010, -0.00000197, -0.00001052, -0.00000576, 0.00000570, -0.00001823, -0.00001867, 0.00000238, 0.00001461, 0.00002934, 0.00000335, 0.00001793, 0.00000239, 0.00000894, -0.00000216, 0.00001883, 0.00000228, 0.00003569, -0.00000652, 0.00000535, -0.00003651, 0.00001542, -0.00001387, -0.00000348, -0.00000201, -0.00001014, 0.00000107, -0.00000343, -0.00000803, 0.00000252, -0.00000233, -0.00000027, -0.00000110, -0.00000381, -0.00000793, 0.00000390, -0.00000795, 0.00000093, -0.00000263, },
		{ 0.00000091, -0.00001360, 0.00001031, -0.00001651, 0.00001773, -0.00001114, 0.00000350, 0.00000691, -0.00000181, -0.00000483, 0.00000139, 0.00000758, -0.00000366, -0.00000605, 0.00000309, 0.00000373, -0.00000280, -0.00001100, 0.00000428, 0.00000018, 0.00000007, -0.00000123, -0.00000645, 0.00000388, -0.00000538, -0.00000283, -0.00000443, 0.00000197, 0.00000161, -0.00000021, -0.00001010, 0.00000188, -0.00000177, -0.00000662, -0.00000665, -0.00000439, 0.00000162, -0.00001123, -0.00001950, -0.00000147, -0.00000597, -0.00000915, -0.00003926, -0.00002680, -0.00017006, 0.00015463, 0.00012428, 0.00007866, 0.00007685, 0.00020487, -0.00011465, -0.00008542, -0.00001791, -0.00001735, -0.00000227, -0.00002551, -0.00000182, -0.00000879, 0.00000494, 0.00000400, -0.00000915, -0.00000183, 0.00000273, -0.00000681, -0.00000059, 0.00000202, -0.00000363, -0.00000441, -0.00000212, -0.00000083, 0.00000735, -0.00000018, -0.00000150, -0.00000196, 0.00000067, -0.00000266, 0.00000291, 0.00000119, -0.00000190, 0.00000010, -0.00000409, 0.00000853, -0.00000658, -0.00000130, -0.00000324, 0.00000728, -0.00000647, 0.00000427, -0.00000405, 0.00000485, -0.00001022, -0.00000635, -0.00000483, 0.00000562, -0.00003135, -0.00000388, -0.00000138, 0.00003351, 0.00001185, 0.00001463, 0.00000791, 0.00000972, -0.00000153, 0.00001360, 0.00000558, 0.00002456, 0.00001547, -0.00000026, -0.00001705, -0.00001834, 0.00001425, -0.00002256, 0.00000832, -0.00001374, -0.00000179, -0.00000109, -0.00000553, -0.00000495, 0.00000303, -0.00000372, 0.00000115, -0.00000324, -0.00000559, -0.00000352, 0.00000145, -0.00000755, 0.00000157, -0.00000104, },
		{ -0.00000977, -0.00000233, 0.00000140, -0.00000833, 0.00001586, -0.00001734, 0.00001471, -0.00000169, 0.00000094, -0.00000853, 0.00000965, 0.00000111, -0.00000355, -0.00000621, 0.00000796, -0.00000096, -0.00000404, -0.00001103, 0.00001085, -0.00000688, 0.00000657, -0.00001132, 0.00000404, -0.00000408, -0.00000015, -0.00000898, 0.00000362, -0.00000322, 0.00000744, -0.00001215, 0.00000119, -0.00000439, 0.00000294, -0.00001602, 0.00000376, -0.00001197, 0.00001125, -0.00003459, 0.00000740, -0.00002249, 0.00001809, -0.00005357, 0.00000963, -0.00013701, -0.00002505, 0.00018262, 0.00007892, 0.00006382, 0.00013054, 0.00012677, -0.00017429, -0.00001657, -0.00004190, 0.00001296, -0.00003034, -0.00000441, -0.00001296, 0.00000370, -0.00000087, 0.00000610, -0.00001591, 0.00001007, -0.00000867, 0.00000043, -0.00000399, 0.00000586, -0.00001045, 0.00000198, -0.00000680, 0.00000727, 0.00000145, 0.00000164, -0.00000456, 0.00000173, -0.00000259, 0.00000014, 0.00000285, -0.00000050, -0.00000077, -0.00000250, 0.00000110, 0.00000504, -0.00000766, -0.00000092, 0.00000010, 0.00000381, -0.00000487, 0.00000358, -0.00000282, 0.00000251, -0.00001257, -0.00000524, 0.00000083, -0.00000654, -0.00002949, 0.00000081, 0.00000840, 0.00003659, 0.00000330, 0.00002059, 0.00000243, 0.00001058, -0.00000094, 0.00001750, 0.00000616, 0.00003390, -0.00000235, 0.00000205, -0.00003496, 0.00001157, -0.00001076, -0.00000628, -0.00000042, -0.00001181, 0.00000049, -0.00000192, -0.00000838, 0.00000143, -0.00000176, 0.00000006, -0.00000234, -0.00000252, -0.00000771, 0.00000237, -0.00000501, -0.00000231, -0.00000203, 0.00000448, },
		{ -0.00001742, 0.00000992, -0.00001186, 0.00000841, -0.00000010, -0.00000738, 0.00001328, -0.00000299, -0.00000259, -0.00000377, 0.00001055, -0.00000211, -0.00000608, -0.00000050, 0.00000605, -0.00000139, -0.00001012, -0.00000152, 0.00000624, -0.00000391, 0.00000181, -0.00000966, 0.00000548, -0.00000531, -0.00000204, -0.00000634, 0.00000342, -0.00000019, 0.00000105, -0.00001030, 0.00000236, -0.00000221, -0.00000449, -0.00001013, -0.00000076, -0.00000206, -0.00000644, -0.00002432, 0.00000436, -0.00001348, -0.00000071, -0.00004340, -0.00001703, -0.00016520, 0.00014103, 0.00010832, 0.00008423, 0.00004456, 0.00019098, -0.00001579, -0.00013404, -0.00000686, -0.00002890, 0.00000738, -0.00003211, 0.00000016, -0.00001080, 0.00000437, 0.00000237, -0.00000327, -0.00000694, 0.00000631, -0.00000872, -0.00000002, 0.00000018, 0.00000015, -0.00000750, 0.00000021, -0.00000295, 0.00000728, 0.00000038, -0.00000060, -0.00000260, 0.00000074, -0.00000221, 0.00000099, 0.00000345, -0.00000325, 0.00000153, -0.00000558, 0.00000821, -0.00000363, -0.00000322, -0.00000358, 0.00000576, -0.00000351, 0.00000186, -0.00000230, 0.00000337, -0.00000632, -0.00000863, -0.00000630, 0.00000594, -0.00002335, -0.00001744, -0.00000062, 0.00002750, 0.00002300, 0.00000925, 0.00001562, 0.00000458, 0.00000567, 0.00000846, 0.00001081, 0.00002269, 0.00001950, -0.00000232, -0.00001585, -0.00002135, 0.00001492, -0.00002263, 0.00000767, -0.00001357, -0.00000302, -0.00000140, -0.00000333, -0.00000730, 0.00000399, -0.00000447, 0.00000251, -0.00000589, -0.00000173, -0.00000710, 0.00000443, -0.00000904, 0.00000160, -0.00000359, 0.00000612, },
		{ -0.00001215, 0.00001023, -0.00001468, 0.00001722, -0.00001429, 0.00000741, 0.00000520, -0.00000109, -0.00000727, 0.00000478, 0.00000630, -0.00000387, -0.00000699, 0.00000539, 0.00000211, -0.00000259, -0.00001320, 0.00000846, -0.00000195, 0.00000208, -0.00000726, -0.00000098, 0.00000026, -0.00000243, -0.00000684, 0.00000042, -0.00000028, 0.00000400, -0.00000873, -0.00000116, -0.00000233, 0.00000192, -0.00001421, 0.00000052, -0.00000825, 0.00000759, -0.00002960, 0.00000186, -0.00001577, 0.00000893, -0.00004033, -0.00000094, -0.00011437, -0.00004164, 0.00017856, 0.00005938, 0.00007809, 0.00007008, 0.00018145, -0.00012869, -0.00006038, -0.00002485, -0.00000130, -0.00001709, -0.00001610, -0.00000620, -0.00000196, 0.00000041, 0.00000578, -0.00001187, 0.00000509, -0.00000391, -0.00000308, -0.00000278, 0.00000481, -0.00000739, -0.00000139, -0.00000368, 0.00000354, 0.00000338, 0.00000117, -0.00000316, 0.00000019, -0.00000133, -0.00000094, 0.00000233, 0.00000143, -0.00000268, 0.00000005, -0.00000285, 0.00000871, -0.00000843, -0.00000059, -0.00000287, 0.00000622, -0.00000581, 0.00000446, -0.00000421, 0.00000499, -0.00001239, -0.00000544, -0.00000374, 0.00000154, -0.00003162, -0.00000809, 0.00000561, 0.00003815, 0.00000921, 0.00001834, 0.00000762, 0.00000740, 0.00000482, 0.00001260, 0.00001228, 0.00003159, 0.00000209, -0.00000203, -0.00003224, 0.00000654, -0.00000697, -0.00000929, 0.00000136, -0.00001368, 0.00000035, -0.00000170, -0.00000703, -0.00000101, -0.00000022, -0.00000072, -0.00000165, -0.00000411, -0.00000446, -0.00000166, -0.00000038, -0.00000624, -0.00000037, 0.00000138, -0.00000233, },
		{ 0.00000249, -0.00000169, -0.00000332, 0.00000995, -0.00001440, 0.00001449, -0.00000063, -0.00000207, -0.00000644, 0.00000985, 0.00000173, -0.00000662, -0.00000281, 0.00000605, 0.00000088, -0.00000843, -0.00000691, 0.00000947, -0.00000413, 0.00000138, -0.00000935, 0.00000358, -0.00000275, -0.00000300, -0.00000620, 0.00000247, 0.00000056, -0.00000016, -0.00000864, 0.00000139, -0.00000150, -0.00000341, -0.00001175, 0.00000014, -0.00000277, -0.00000476, -0.00002570, 0.00000594, -0.00001519, 0.00000046, -0.00004023, -0.00001523, -0.00015297, 0.00012084, 0.00010471, 0.00007366, 0.00004282, 0.00014050, 0.00007969, -0.00015038, -0.00001878, -0.00002716, 0.00000611, -0.00002898, -0.00000524, -0.00000779, 0.00000093, 0.00000203, 0.00000031, -0.00000842, 0.00000622, -0.00000743, -0.00000152, -0.00000041, 0.00000213, -0.00000806, -0.00000008, -0.00000249, 0.00000524, 0.00000187, -0.00000047, -0.00000219, -0.00000016, -0.00000108, -0.00000086, 0.00000464, -0.00000300, 0.00000107, -0.00000459, 0.00000513, 0.00000109, -0.00000592, -0.00000249, 0.00000234, 0.00000051, -0.00000123, 0.00000047, 0.00000043, -0.00000140, -0.00001127, -0.00000581, 0.00000188, -0.00001188, -0.00002845, -0.00000404, 0.00002189, 0.00003216, 0.00000719, 0.00002026, 0.00000325, 0.00000866, 0.00000818, 0.00001112, 0.00002524, 0.00002086, -0.00000269, -0.00001637, -0.00002237, 0.00001333, -0.00002075, 0.00000566, -0.00001236, -0.00000490, -0.00000139, -0.00000196, -0.00000817, 0.00000350, -0.00000421, 0.00000317, -0.00000741, -0.00000017, -0.00000747, 0.00000385, -0.00000708, -0.00000082, -0.00000386, 0.00000712, -0.00001470, },
		{ 0.00001205, -0.00001253, 0.00001054, -0.00000514, -0.00000294, 0.00001196, -0.00000257, -0.00000514, -0.00000059, 0.00001000, -0.00000270, -0.00000747, 0.00000257, 0.00000400, -0.00000066, -0.00001336, 0.00000349, 0.00000387, -0.00000183, -0.00000415, -0.00000456, 0.00000246, -0.00000296, -0.00000564, -0.00000189, 0.00000149, 0.00000206, -0.00000709, -0.00000207, -0.00000130, 0.00000133, -0.00001202, -0.00000307, -0.00000471, 0.00000426, -0.00002477, -0.00000373, -0.00000905, 0.00000073, -0.00002972, -0.00000842, -0.00009603, -0.00005376, 0.00017274, 0.00004474, 0.00008561, 0.00003428, 0.00018287, -0.00004760, -0.00010541, -0.00001308, -0.00001285, -0.00000659, -0.00002384, -0.00000336, -0.00000456, 0.00000018, 0.00000469, -0.00000691, 0.00000036, -0.00000013, -0.00000521, -0.00000236, 0.00000343, -0.00000397, -0.00000439, -0.00000141, 0.00000123, 0.00000378, 0.00000151, -0.00000224, -0.00000054, -0.00000115, -0.00000073, 0.00000069, 0.00000373, -0.00000431, 0.00000191, -0.00000520, 0.00000962, -0.00000650, -0.00000196, -0.00000378, 0.00000573, -0.00000418, 0.00000314, -0.00000324, 0.00000446, -0.00000924, -0.00000747, -0.00000566, 0.00000333, -0.00002490, -0.00002118, 0.00000323, 0.00003593, 0.00001861, 0.00001397, 0.00001497, 0.00000288, 0.00001126, 0.00000824, 0.00001677, 0.00003174, 0.00000494, -0.00000558, -0.00002977, 0.00000190, -0.00000408, -0.00001125, 0.00000240, -0.00001503, 0.00000008, -0.00000208, -0.00000503, -0.00000356, 0.00000122, -0.00000176, -0.00000001, -0.00000696, -0.00000058, -0.00000534, 0.00000304, -0.00000837, 0.00000023, -0.00000166, 0.00000398, -0.00001780, },
		{ 0.00000783, -0.00001112, 0.00001373, -0.00001414, 0.00000927, 0.00000524, -0.00000344, -0.00000619, 0.00000581, 0.00000690, -0.00000720, -0.00000389, 0.00000438, 0.00000320, -0.00000622, -0.00001056, 0.00000891, -0.00000075, -0.00000164, -0.00000681, -0.00000010, 0.00000055, -0.00000440, -0.00000504, 0.00000023, 0.00000252, -0.00000229, -0.00000660, 0.00000011, -0.00000044, -0.00000305, -0.00001161, -0.00000135, -0.00000110, -0.00000552, -0.00002416, 0.00000383, -0.00001223, -0.00000344, -0.00003274, -0.00001799, -0.00013678, 0.00009736, 0.00010876, 0.00005427, 0.00005585, 0.00008409, 0.00014117, -0.00012260, -0.00005208, -0.00001553, -0.00000328, -0.00001926, -0.00001457, -0.00000280, -0.00000330, 0.00000231, 0.00000144, -0.00000650, 0.00000326, -0.00000409, -0.00000403, -0.00000026, 0.00000245, -0.00000645, -0.00000220, -0.00000060, 0.00000255, 0.00000325, -0.00000045, -0.00000139, -0.00000118, -0.00000028, -0.00000173, 0.00000425, -0.00000119, -0.00000079, -0.00000214, 0.00000101, 0.00000549, -0.00000760, -0.00000166, -0.00000086, 0.00000342, -0.00000314, 0.00000236, -0.00000171, 0.00000204, -0.00001204, -0.00000570, -0.00000259, -0.00000307, -0.00003133, -0.00001389, 0.00001783, 0.00003726, 0.00001022, 0.00002048, 0.00000684, 0.00000693, 0.00001248, 0.00000766, 0.00003032, 0.00002148, -0.00000257, -0.00001805, -0.00002178, 0.00001007, -0.00001768, 0.00000295, -0.00001055, -0.00000727, -0.00000086, -0.00000181, -0.00000738, 0.00000161, -0.00000291, 0.00000273, -0.00000706, -0.00000141, -0.00000479, 0.00000055, -0.00000305, -0.00000434, -0.00000268, 0.00000488, -0.00000791, -0.00000737, },
		{ -0.00000410, 0.00000003, 0.00000483, -0.00001111, 0.00001464, -0.00000109, -0.00000423, -0.00000399, 0.00001023, 0.00000090, -0.00000857, 0.00000082, 0.00000396, 0.00000177, -0.00001210, -0.00000203, 0.00000787, -0.00000285, -0.00000376, -0.00000531, 0.00000172, -0.00000138, -0.00000584, -0.00000253, 0.00000126, 0.00000227, -0.00000754, -0.00000141, -0.00000147, 0.00000155, -0.00001036, -0.00000550, -0.00000307, 0.00000300, -0.00002201, -0.00000697, -0.00000537, -0.00000314, -0.00002477, -0.00001017, -0.00008448, -0.00005856, 0.00016083, 0.00004051, 0.00007938, 0.00002632, 0.00014614, 0.00003910, -0.00012759, -0.00001904, -0.00001425, -0.00000463, -0.00002350, -0.00000680, -0.00000287, -0.00000235, 0.00000421, -0.00000329, -0.00000179, 0.00000083, -0.00000474, -0.00000351, 0.00000269, -0.00000160, -0.00000560, -0.00000108, 0.00000098, 0.00000254, 0.00000258, -0.00000190, -0.00000041, -0.00000180, 0.00000012, -0.00000101, 0.00000520, -0.00000456, 0.00000205, -0.00000512, 0.00000756, -0.00000238, -0.00000451, -0.00000303, 0.00000309, -0.00000099, 0.00000050, -0.00000067, 0.00000193, -0.00000451, -0.00001025, -0.00000545, 0.00000037, -0.00001403, -0.00003195, -0.00000314, 0.00003271, 0.00002796, 0.00001121, 0.00002076, 0.00000091, 0.00001464, 0.00000772, 0.00001713, 0.00003570, 0.00000577, -0.00000819, -0.00002836, -0.00000127, -0.00000304, -0.00001150, 0.00000220, -0.00001538, -0.00000083, -0.00000224, -0.00000360, -0.00000500, 0.00000153, -0.00000208, 0.00000127, -0.00000913, 0.00000158, -0.00000646, 0.00000346, -0.00000748, -0.00000122, -0.00000296, 0.00000666, -0.00001720, 0.00000662, },
		{ -0.00001140, 0.00000979, -0.00000729, -0.00000001, 0.00001192, -0.00000488, -0.00000451, 0.00000059, 0.00001108, -0.00000635, -0.00000491, 0.00000259, 0.00000432, -0.00000340, -0.00001224, 0.00000539, 0.00000419, -0.00000480, -0.00000445, -0.00000314, 0.00000249, -0.00000472, -0.00000423, -0.00000170, 0.00000390, -0.00000353, -0.00000571, -0.00000027, 0.00000001, -0.00000263, -0.00001068, -0.00000375, 0.00000110, -0.00000677, -0.00002182, 0.00000057, -0.00000803, -0.00000799, -0.00002580, -0.00002033, -0.00012176, 0.00007609, 0.00011297, 0.00003657, 0.00006686, 0.00004456, 0.00015669, -0.00005936, -0.00009114, -0.00000546, -0.00001261, -0.00001001, -0.00002180, -0.00000039, -0.00000536, 0.00000158, 0.00000148, -0.00000316, -0.00000015, -0.00000103, -0.00000565, -0.00000060, 0.00000217, -0.00000405, -0.00000450, 0.00000105, 0.00000069, 0.00000355, 0.00000008, -0.00000083, -0.00000163, -0.00000033, -0.00000147, 0.00000272, 0.00000124, -0.00000276, 0.00000013, -0.00000211, 0.00000763, -0.00000686, -0.00000236, -0.00000234, 0.00000388, -0.00000280, 0.00000220, -0.00000158, 0.00000254, -0.00000982, -0.00000736, -0.00000478, -0.00000016, -0.00002541, -0.00002738, 0.00001290, 0.00003907, 0.00001716, 0.00001789, 0.00001343, 0.00000305, 0.00001831, 0.00000386, 0.00003453, 0.00002406, -0.00000359, -0.00001984, -0.00002073, 0.00000646, -0.00001476, 0.00000067, -0.00000899, -0.00000946, -0.00000025, -0.00000239, -0.00000575, -0.00000070, -0.00000142, 0.00000177, -0.00000551, -0.00000423, -0.00000101, -0.00000307, 0.00000053, -0.00000678, -0.00000187, 0.00000176, -0.00000097, -0.00001449, 0.00001165, },
		{ -0.00000834, 0.00001023, -0.00001310, 0.00001124, 0.00000394, -0.00000521, -0.00000448, 0.00000685, 0.00000626, -0.00000985, 0.00000004, 0.00000261, 0.00000387, -0.00000970, -0.00000662, 0.00000870, -0.00000059, -0.00000577, -0.00000389, -0.00000086, 0.00000109, -0.00000662, -0.00000208, -0.00000032, 0.00000375, -0.00000899, -0.00000023, -0.00000206, 0.00000210, -0.00000922, -0.00000649, -0.00000350, 0.00000380, -0.00002137, -0.00000763, -0.00000505, -0.00000277, -0.00002453, -0.00000777, -0.00007788, -0.00005788, 0.00014424, 0.00004457, 0.00006325, 0.00003531, 0.00009682, 0.00010263, -0.00011301, -0.00004432, -0.00000634, -0.00001042, -0.00001655, -0.00001441, 0.00000098, -0.00000571, 0.00000411, -0.00000160, -0.00000105, -0.00000080, -0.00000233, -0.00000552, 0.00000236, -0.00000043, -0.00000500, -0.00000239, 0.00000210, 0.00000056, 0.00000366, -0.00000170, 0.00000013, -0.00000261, 0.00000073, -0.00000191, 0.00000523, -0.00000323, 0.00000064, -0.00000325, 0.00000389, 0.00000211, -0.00000661, -0.00000212, 0.00000014, 0.00000181, -0.00000175, 0.00000169, -0.00000040, -0.00000070, -0.00001154, -0.00000533, -0.00000357, -0.00000507, -0.00003485, -0.00001541, 0.00002923, 0.00003517, 0.00001225, 0.00002301, 0.00000329, 0.00001370, 0.00001156, 0.00001377, 0.00004226, 0.00000597, -0.00001066, -0.00002785, -0.00000300, -0.00000367, -0.00001041, 0.00000110, -0.00001491, -0.00000239, -0.00000184, -0.00000326, -0.00000493, 0.00000053, -0.00000145, 0.00000158, -0.00000963, 0.00000118, -0.00000470, 0.00000116, -0.00000437, -0.00000395, -0.00000271, 0.00000582, -0.00001180, -0.00000244, 0.00000526, },
		{ 0.00000108, 0.00000187, -0.00000863, 0.00001562, -0.00000398, -0.00000354, -0.00000319, 0.00001204, -0.00000288, -0.00000679, 0.00000187, 0.00000366, -0.00000025, -0.00001208, 0.00000052, 0.00000826, -0.00000591, -0.00000406, -0.00000406, 0.00000212, -0.00000341, -0.00000441, -0.00000242, 0.00000375, -0.00000307, -0.00000645, 0.00000054, -0.00000044, -0.00000175, -0.00000976, -0.00000569, 0.00000220, -0.00000684, -0.00002055, -0.00000149, -0.00000559, -0.00000982, -0.00002259, -0.00001930, -0.00011085, 0.00006034, 0.00011251, 0.00002693, 0.00006710, 0.00002909, 0.00013465, 0.00001577, -0.00011505, -0.00000854, -0.00001464, -0.00000705, -0.00002256, -0.00000322, -0.00000383, -0.00000097, 0.00000162, -0.00000042, -0.00000187, -0.00000004, -0.00000519, -0.00000213, 0.00000204, -0.00000204, -0.00000572, 0.00000138, 0.00000036, 0.00000254, 0.00000114, -0.00000065, -0.00000140, -0.00000109, -0.00000063, 0.00000100, 0.00000311, -0.00000363, 0.00000098, -0.00000308, 0.00000685, -0.00000371, -0.00000445, -0.00000209, 0.00000215, -0.00000069, 0.00000029, 0.00000057, 0.00000062, -0.00000558, -0.00001005, -0.00000482, -0.00000215, -0.00001491, -0.00003830, 0.00000338, 0.00003926, 0.00002552, 0.00001545, 0.00001982, 0.00000055, 0.00002212, 0.00000301, 0.00003529, 0.00003012, -0.00000635, -0.00002125, -0.00002006, 0.00000362, -0.00001305, -0.00000040, -0.00000830, -0.00001091, -0.00000013, -0.00000291, -0.00000441, -0.00000228, -0.00000070, 0.00000119, -0.00000393, -0.00000685, 0.00000167, -0.00000483, 0.00000176, -0.00000674, -0.00000257, -0.00000035, 0.00000320, -0.00001600, 0.00000844, -0.00000508, },
		{ 0.00000864, -0.00000844, 0.00000313, 0.00001054, -0.00000675, -0.00000311, 0.00000168, 0.00001114, -0.00000976, -0.00000098, 0.00000145, 0.00000450, -0.00000631, -0.00000957, 0.00000656, 0.00000357, -0.00000835, -0.00000222, -0.00000331, 0.00000274, -0.00000675, -0.00000162, -0.00000197, 0.00000503, -0.00000993, 0.00000019, -0.00000206, 0.00000221, -0.00000809, -0.00000660, -0.00000502, 0.00000539, -0.00002150, -0.00000715, -0.00000629, -0.00000071, -0.00002558, -0.00000500, -0.00007233, -0.00005573, 0.00012718, 0.00005106, 0.00004571, 0.00004665, 0.00005694, 0.00012817, -0.00006497, -0.00007786, 0.00000230, -0.00001776, -0.00000859, -0.00002115, 0.00000313, -0.00000741, 0.00000317, -0.00000098, 0.00000119, -0.00000319, 0.00000017, -0.00000686, 0.00000154, 0.00000014, -0.00000348, -0.00000416, 0.00000332, -0.00000098, 0.00000395, -0.00000117, 0.00000057, -0.00000301, 0.00000063, -0.00000181, 0.00000412, -0.00000108, -0.00000123, -0.00000106, 0.00000055, 0.00000497, -0.00000668, -0.00000244, -0.00000150, 0.00000271, -0.00000225, 0.00000230, -0.00000069, 0.00000036, -0.00000989, -0.00000689, -0.00000559, -0.00000165, -0.00002888, -0.00003072, 0.00002318, 0.00004028, 0.00001696, 0.00002229, 0.00000909, 0.00001010, 0.00001752, 0.00000941, 0.00004858, 0.00000788, -0.00001435, -0.00002746, -0.00000409, -0.00000496, -0.00000906, 0.00000001, -0.00001432, -0.00000407, -0.00000121, -0.00000371, -0.00000396, -0.00000106, -0.00000045, 0.00000125, -0.00000880, -0.00000098, -0.00000151, -0.00000196, -0.00000112, -0.00000615, -0.00000226, 0.00000330, -0.00000503, -0.00001020, 0.00001036, -0.00001045, },
		{ 0.00000813, -0.00001231, 0.00001378, 0.00000037, -0.00000436, -0.00000416, 0.00000916, 0.00000282, -0.00000915, 0.00000200, 0.00000198, 0.00000223, -0.00001018, -0.00000429, 0.00000990, -0.00000421, -0.00000573, -0.00000311, 0.00000008, -0.00000126, -0.00000510, -0.00000215, 0.00000233, -0.00000144, -0.00000813, 0.00000184, -0.00000126, -0.00000069, -0.00000905, -0.00000665, 0.00000172, -0.00000534, -0.00002083, -0.00000161, -0.00000579, -0.00000831, -0.00002301, -0.00001543, -0.00010341, 0.00004959, 0.00010726, 0.00002535, 0.00005742, 0.00003128, 0.00009538, 0.00007689, -0.00010926, -0.00002834, -0.00000851, -0.00001100, -0.00001710, -0.00001012, -0.00000015, -0.00000428, 0.00000186, 0.00000095, -0.00000129, -0.00000124, -0.00000298, -0.00000433, 0.00000194, -0.00000072, -0.00000559, 0.00000041, 0.00000121, 0.00000085, 0.00000223, -0.00000058, -0.00000079, -0.00000203, 0.00000004, -0.00000009, 0.00000375, -0.00000304, 0.00000029, -0.00000209, 0.00000405, 0.00000037, -0.00000655, -0.00000136, -0.00000025, 0.00000161, -0.00000190, 0.00000311, -0.00000162, -0.00000176, -0.00001157, -0.00000477, -0.00000550, -0.00000587, -0.00004111, -0.00001209, 0.00003772, 0.00003363, 0.00001537, 0.00002379, 0.00000162, 0.00002219, 0.00000617, 0.00003239, 0.00003901, -0.00000987, -0.00002280, -0.00001983, 0.00000175, -0.00001264, -0.00000034, -0.00000839, -0.00001158, -0.00000063, -0.00000296, -0.00000396, -0.00000263, -0.00000106, 0.00000135, -0.00000305, -0.00000814, 0.00000218, -0.00000406, 0.00000048, -0.00000456, -0.00000445, -0.00000110, 0.00000393, -0.00001236, 0.00000076, 0.00000343, -0.00000711, },
		{ -0.00000035, -0.00000566, 0.00001516, -0.00000654, -0.00000196, -0.00000259, 0.00001307, -0.00000685, -0.00000327, 0.00000160, 0.00000325, -0.00000248, -0.00001057, 0.00000251, 0.00000760, -0.00000953, -0.00000192, -0.00000425, 0.00000262, -0.00000568, -0.00000177, -0.00000280, 0.00000514, -0.00000954, -0.00000070, -0.00000113, 0.00000160, -0.00000667, -0.00000647, -0.00000652, 0.00000641, -0.00002098, -0.00000712, -0.00000702, 0.00000034, -0.00002485, -0.00000477, -0.00006511, -0.00005497, 0.00011288, 0.00005528, 0.00003329, 0.00005083, 0.00003634, 0.00011869, -0.00000133, -0.00010207, 0.00000140, -0.00002008, -0.00000508, -0.00002275, 0.00000086, -0.00000601, 0.00000067, -0.00000047, 0.00000327, -0.00000448, 0.00000113, -0.00000643, -0.00000031, 0.00000063, -0.00000192, -0.00000535, 0.00000371, -0.00000142, 0.00000324, -0.00000025, 0.00000076, -0.00000283, -0.00000011, -0.00000115, 0.00000269, 0.00000082, -0.00000239, 0.00000014, -0.00000105, 0.00000514, -0.00000436, -0.00000420, -0.00000150, 0.00000159, -0.00000099, 0.00000098, 0.00000129, -0.00000126, -0.00000591, -0.00000968, -0.00000550, -0.00000320, -0.00001805, -0.00004286, 0.00001122, 0.00004389, 0.00002399, 0.00002065, 0.00001585, 0.00000682, 0.00002244, 0.00000711, 0.00005213, 0.00001309, -0.00001994, -0.00002669, -0.00000531, -0.00000591, -0.00000841, -0.00000032, -0.00001424, -0.00000526, -0.00000089, -0.00000423, -0.00000307, -0.00000221, 0.00000008, 0.00000102, -0.00000763, -0.00000334, 0.00000112, -0.00000387, 0.00000043, -0.00000645, -0.00000273, 0.00000096, -0.00000002, -0.00001327, 0.00000900, -0.00000607, 0.00000178, },
		{ -0.00000963, 0.00000686, 0.00000685, -0.00000613, -0.00000293, 0.00000392, 0.00000861, -0.00001030, 0.00000165, 0.00000086, 0.00000286, -0.00000672, -0.00000805, 0.00000883, -0.00000056, -0.00000811, -0.00000173, -0.00000209, 0.00000039, -0.00000540, -0.00000167, 0.00000063, 0.00000022, -0.00000952, 0.00000261, -0.00000166, -0.00000002, -0.00000825, -0.00000688, 0.00000022, -0.00000304, -0.00002186, -0.00000061, -0.00000752, -0.00000524, -0.00002456, -0.00001146, -0.00009675, 0.00004107, 0.00010000, 0.00002780, 0.00004407, 0.00003921, 0.00005906, 0.00010789, -0.00007322, -0.00005744, -0.00000040, -0.00001738, -0.00000971, -0.00001699, 0.00000240, -0.00000623, 0.00000126, 0.00000146, 0.00000065, -0.00000326, -0.00000053, -0.00000589, 0.00000118, 0.00000023, -0.00000462, -0.00000109, 0.00000231, -0.00000064, 0.00000274, -0.00000029, -0.00000018, -0.00000264, 0.00000010, -0.00000032, 0.00000330, -0.00000151, -0.00000113, -0.00000036, 0.00000098, 0.00000342, -0.00000706, -0.00000157, -0.00000174, 0.00000257, -0.00000292, 0.00000431, -0.00000211, -0.00000048, -0.00001020, -0.00000639, -0.00000717, -0.00000219, -0.00003470, -0.00003027, 0.00003174, 0.00004130, 0.00001820, 0.00002510, 0.00000618, 0.00001932, 0.00001188, 0.00002789, 0.00004844, -0.00001219, -0.00002561, -0.00001949, 0.00000035, -0.00001283, 0.00000007, -0.00000863, -0.00001199, -0.00000135, -0.00000274, -0.00000423, -0.00000215, -0.00000199, 0.00000189, -0.00000278, -0.00000815, 0.00000096, -0.00000179, -0.00000187, -0.00000189, -0.00000614, -0.00000132, 0.00000240, -0.00000636, -0.00000693, 0.00000908, -0.00001039, 0.00000841, },
		{ -0.00001079, 0.00001491, -0.00000308, -0.00000243, -0.00000421, 0.00001097, -0.00000117, -0.00000652, 0.00000273, 0.00000107, 0.00000051, -0.00000936, -0.00000226, 0.00000995, -0.00000843, -0.00000328, -0.00000368, 0.00000111, -0.00000392, -0.00000232, -0.00000278, 0.00000413, -0.00000806, -0.00000244, 0.00000027, 0.00000062, -0.00000513, -0.00000631, -0.00000741, 0.00000625, -0.00001927, -0.00000825, -0.00000626, -0.00000078, -0.00002141, -0.00000741, -0.00005619, -0.00005569, 0.00010172, 0.00005601, 0.00002752, 0.00004636, 0.00003248, 0.00008952, 0.00005543, -0.00010278, -0.00001365, -0.00001548, -0.00000754, -0.00001863, -0.00000535, -0.00000246, -0.00000259, -0.00000002, 0.00000439, -0.00000398, 0.00000026, -0.00000440, -0.00000270, 0.00000087, -0.00000060, -0.00000565, 0.00000316, -0.00000094, 0.00000193, 0.00000068, 0.00000087, -0.00000229, -0.00000103, -0.00000060, 0.00000169, 0.00000176, -0.00000230, -0.00000005, -0.00000073, 0.00000309, -0.00000070, -0.00000631, -0.00000084, -0.00000039, 0.00000083, -0.00000118, 0.00000418, -0.00000369, -0.00000189, -0.00001156, -0.00000524, -0.00000632, -0.00000845, -0.00004618, -0.00000746, 0.00004485, 0.00003250, 0.00001989, 0.00002147, 0.00000612, 0.00002440, 0.00000826, 0.00005230, 0.00002136, -0.00002677, -0.00002591, -0.00000682, -0.00000615, -0.00000876, 0.00000025, -0.00001480, -0.00000580, -0.00000110, -0.00000440, -0.00000287, -0.00000240, -0.00000022, 0.00000129, -0.00000687, -0.00000475, 0.00000203, -0.00000373, -0.00000016, -0.00000488, -0.00000411, -0.00000043, 0.00000193, -0.00001120, 0.00000272, 0.00000136, -0.00000589, 0.00000640, },
		{ -0.00000188, 0.00001230, -0.00000671, -0.00000141, -0.00000102, 0.00001180, -0.00000847, -0.00000046, 0.00000147, 0.00000128, -0.00000248, -0.00001001, 0.00000558, 0.00000348, -0.00000958, -0.00000128, -0.00000323, 0.00000074, -0.00000487, -0.00000149, -0.00000058, 0.00000102, -0.00000986, 0.00000233, -0.00000132, -0.00000000, -0.00000716, -0.00000680, -0.00000148, -0.00000100, -0.00002254, 0.00000022, -0.00000906, -0.00000291, -0.00002462, -0.00000982, -0.00008867, 0.00003249, 0.00009342, 0.00003019, 0.00003304, 0.00004358, 0.00003692, 0.00010768, -0.00001989, -0.00008142, 0.00000094, -0.00002046, -0.00000531, -0.00001981, 0.00000118, -0.00000538, -0.00000082, 0.00000186, 0.00000258, -0.00000455, 0.00000073, -0.00000584, -0.00000061, 0.00000107, -0.00000340, -0.00000235, 0.00000295, -0.00000140, 0.00000251, 0.00000024, 0.00000030, -0.00000279, -0.00000039, -0.00000005, 0.00000245, 0.00000003, -0.00000218, 0.00000082, -0.00000082, 0.00000415, -0.00000528, -0.00000318, -0.00000175, 0.00000179, -0.00000234, 0.00000352, -0.00000019, -0.00000208, -0.00000620, -0.00000946, -0.00000673, -0.00000359, -0.00002305, -0.00004455, 0.00001800, 0.00004832, 0.00002366, 0.00002482, 0.00001260, 0.00001582, 0.00001754, 0.00002443, 0.00005611, -0.00001180, -0.00003037, -0.00001861, -0.00000122, -0.00001278, -0.00000002, -0.00000833, -0.00001273, -0.00000173, -0.00000274, -0.00000462, -0.00000167, -0.00000264, 0.00000212, -0.00000256, -0.00000765, -0.00000070, 0.00000031, -0.00000353, -0.00000038, -0.00000643, -0.00000195, 0.00000028, -0.00000106, -0.00001096, 0.00000909, -0.00000766, 0.00000331, -0.00000274, },
		{ 0.00000945, 0.00000292, -0.00000415, -0.00000332, 0.00000593, 0.00000530, -0.00000891, 0.00000327, -0.00000025, 0.00000126, -0.00000598, -0.00000669, 0.00001003, -0.00000548, -0.00000533, -0.00000267, -0.00000063, -0.00000248, -0.00000264, -0.00000243, 0.00000268, -0.00000639, -0.00000408, 0.00000134, -0.00000013, -0.00000391, -0.00000599, -0.00000773, 0.00000517, -0.00001676, -0.00001013, -0.00000439, -0.00000345, -0.00001646, -0.00001109, -0.00004757, -0.00005613, 0.00009206, 0.00005474, 0.00002609, 0.00003729, 0.00003635, 0.00005827, 0.00008919, -0.00007691, -0.00003815, -0.00000826, -0.00001288, -0.00001196, -0.00001217, 0.00000044, -0.00000473, -0.00000042, 0.00000489, -0.00000233, -0.00000142, -0.00000203, -0.00000448, 0.00000025, 0.00000066, -0.00000533, 0.00000215, -0.00000016, 0.00000067, 0.00000122, 0.00000110, -0.00000167, -0.00000173, -0.00000054, 0.00000137, 0.00000177, -0.00000128, -0.00000106, 0.00000059, 0.00000030, 0.00000246, -0.00000722, -0.00000087, -0.00000177, 0.00000183, -0.00000273, 0.00000617, -0.00000469, -0.00000021, -0.00001059, -0.00000673, -0.00000787, -0.00000424, -0.00003951, -0.00002895, 0.00003957, 0.00004253, 0.00002083, 0.00002521, 0.00000853, 0.00002355, 0.00001202, 0.00005054, 0.00003093, -0.00003328, -0.00002599, -0.00000826, -0.00000601, -0.00000963, 0.00000118, -0.00001558, -0.00000601, -0.00000156, -0.00000431, -0.00000330, -0.00000189, -0.00000100, 0.00000185, -0.00000658, -0.00000506, 0.00000139, -0.00000218, -0.00000187, -0.00000266, -0.00000543, -0.00000120, 0.00000133, -0.00000603, -0.00000465, 0.00000722, -0.00000999, 0.00000830, -0.00001011, },
		{ 0.00001365, -0.00000435, -0.00000128, -0.00000353, 0.00001086, -0.00000332, -0.00000416, 0.00000343, -0.00000136, 0.00000094, -0.00000948, 0.00000095, 0.00000655, -0.00000929, -0.00000206, -0.00000325, 0.00000002, -0.00000387, -0.00000154, -0.00000114, 0.00000089, -0.00000927, 0.00000132, -0.00000055, -0.00000038, -0.00000595, -0.00000657, -0.00000286, 0.00000021, -0.00002232, 0.00000019, -0.00000941, -0.00000253, -0.00002218, -0.00001095, -0.00007905, 0.00002353, 0.00008824, 0.00003080, 0.00002671, 0.00004134, 0.00002943, 0.00008714, 0.00003182, -0.00008717, -0.00000948, -0.00001785, -0.00000593, -0.00001743, -0.00000384, -0.00000237, -0.00000371, 0.00000216, 0.00000385, -0.00000444, 0.00000045, -0.00000428, -0.00000293, 0.00000148, -0.00000204, -0.00000314, 0.00000301, -0.00000153, 0.00000183, 0.00000075, 0.00000074, -0.00000260, -0.00000106, 0.00000015, 0.00000188, 0.00000088, -0.00000226, 0.00000087, -0.00000091, 0.00000261, -0.00000194, -0.00000538, -0.00000102, 0.00000009, -0.00000097, 0.00000147, 0.00000311, -0.00000494, -0.00000175, -0.00001187, -0.00000607, -0.00000674, -0.00001252, -0.00004890, -0.00000396, 0.00005239, 0.00003201, 0.00002402, 0.00001924, 0.00001373, 0.00002127, 0.00002346, 0.00006126, -0.00000867, -0.00003664, -0.00001742, -0.00000319, -0.00001209, -0.00000094, -0.00000727, -0.00001405, -0.00000149, -0.00000323, -0.00000471, -0.00000177, -0.00000252, 0.00000170, -0.00000201, -0.00000734, -0.00000173, 0.00000112, -0.00000361, -0.00000058, -0.00000518, -0.00000311, -0.00000143, 0.00000177, -0.00001016, 0.00000405, -0.00000133, -0.00000360, 0.00000416, -0.00000821, },
		{ 0.00000877, -0.00000525, -0.00000165, 0.00000054, 0.00000965, -0.00000829, 0.00000161, 0.00000067, -0.00000065, -0.00000130, -0.00000970, 0.00000787, -0.00000183, -0.00000681, -0.00000227, -0.00000170, -0.00000194, -0.00000247, -0.00000212, 0.00000144, -0.00000525, -0.00000498, 0.00000166, -0.00000037, -0.00000320, -0.00000541, -0.00000771, 0.00000372, -0.00001420, -0.00001201, -0.00000233, -0.00000635, -0.00001187, -0.00001361, -0.00004116, -0.00005471, 0.00008230, 0.00005331, 0.00002589, 0.00002849, 0.00003959, 0.00003644, 0.00009613, -0.00003349, -0.00006051, -0.00000572, -0.00001628, -0.00000708, -0.00001589, 0.00000012, -0.00000435, -0.00000226, 0.00000535, -0.00000065, -0.00000258, -0.00000060, -0.00000472, -0.00000152, 0.00000193, -0.00000468, 0.00000117, 0.00000043, -0.00000018, 0.00000125, 0.00000144, -0.00000105, -0.00000208, -0.00000093, 0.00000143, 0.00000139, -0.00000010, -0.00000199, 0.00000176, -0.00000172, 0.00000370, -0.00000603, -0.00000222, -0.00000188, 0.00000141, -0.00000298, 0.00000627, -0.00000321, -0.00000157, -0.00000669, -0.00000991, -0.00000724, -0.00000535, -0.00002703, -0.00004581, 0.00002418, 0.00005349, 0.00002407, 0.00002725, 0.00001325, 0.00002148, 0.00001644, 0.00004899, 0.00003990, -0.00003819, -0.00002750, -0.00000932, -0.00000596, -0.00001038, 0.00000181, -0.00001601, -0.00000641, -0.00000174, -0.00000440, -0.00000386, -0.00000136, -0.00000157, 0.00000218, -0.00000639, -0.00000481, 0.00000023, -0.00000059, -0.00000324, -0.00000119, -0.00000572, -0.00000198, -0.00000052, -0.00000057, -0.00000944, 0.00000837, -0.00000879, 0.00000507, -0.00000477, 0.00000128, },
		{ 0.00000081, -0.00000261, -0.00000332, 0.00000641, 0.00000339, -0.00000754, 0.00000486, -0.00000295, 0.00000181, -0.00000618, -0.00000412, 0.00000792, -0.00000743, -0.00000339, -0.00000298, -0.00000086, -0.00000314, -0.00000142, -0.00000135, 0.00000029, -0.00000840, 0.00000028, 0.00000005, -0.00000069, -0.00000496, -0.00000613, -0.00000386, 0.00000067, -0.00002136, -0.00000061, -0.00000859, -0.00000384, -0.00001804, -0.00001344, -0.00006941, 0.00001537, 0.00008357, 0.00003011, 0.00002420, 0.00003454, 0.00003040, 0.00006097, 0.00006669, -0.00007015, -0.00002888, -0.00001246, -0.00000955, -0.00001216, -0.00000996, 0.00000039, -0.00000573, 0.00000162, 0.00000471, -0.00000343, -0.00000055, -0.00000232, -0.00000464, 0.00000083, -0.00000037, -0.00000363, 0.00000282, -0.00000144, 0.00000115, 0.00000096, 0.00000125, -0.00000228, -0.00000159, -0.00000006, 0.00000180, 0.00000101, -0.00000153, 0.00000011, 0.00000016, -0.00000002, 0.00000137, -0.00000672, -0.00000079, -0.00000123, -0.00000002, -0.00000053, 0.00000600, -0.00000674, 0.00000063, -0.00001144, -0.00000728, -0.00000829, -0.00000763, -0.00004201, -0.00002942, 0.00004865, 0.00004406, 0.00002336, 0.00002524, 0.00001387, 0.00002286, 0.00002461, 0.00006487, -0.00000414, -0.00004325, -0.00001654, -0.00000535, -0.00001088, -0.00000246, -0.00000574, -0.00001572, -0.00000079, -0.00000403, -0.00000454, -0.00000244, -0.00000180, 0.00000089, -0.00000126, -0.00000737, -0.00000185, 0.00000064, -0.00000252, -0.00000179, -0.00000330, -0.00000417, -0.00000265, 0.00000201, -0.00000579, -0.00000292, 0.00000450, -0.00000822, 0.00000746, -0.00000948, 0.00000988, },
		{ -0.00000380, -0.00000104, -0.00000278, 0.00000986, -0.00000384, -0.00000236, 0.00000347, -0.00000398, 0.00000278, -0.00001013, 0.00000384, 0.00000149, -0.00000702, -0.00000270, -0.00000209, -0.00000205, -0.00000213, -0.00000177, 0.00000055, -0.00000478, -0.00000521, 0.00000147, -0.00000034, -0.00000282, -0.00000475, -0.00000743, 0.00000227, -0.00001201, -0.00001343, -0.00000070, -0.00000850, -0.00000883, -0.00001402, -0.00003737, -0.00005142, 0.00007229, 0.00005232, 0.00002524, 0.00002264, 0.00003837, 0.00002663, 0.00008337, 0.00001198, -0.00006900, -0.00001273, -0.00001521, -0.00000622, -0.00001502, -0.00000377, -0.00000180, -0.00000501, 0.00000576, 0.00000053, -0.00000266, -0.00000046, -0.00000350, -0.00000389, 0.00000271, -0.00000362, 0.00000029, 0.00000082, -0.00000072, 0.00000104, 0.00000165, -0.00000039, -0.00000216, -0.00000147, 0.00000141, 0.00000118, 0.00000064, -0.00000225, 0.00000213, -0.00000229, 0.00000276, -0.00000314, -0.00000430, -0.00000124, 0.00000016, -0.00000239, 0.00000482, 0.00000005, -0.00000461, -0.00000190, -0.00001270, -0.00000635, -0.00000837, -0.00001560, -0.00005126, -0.00000175, 0.00006203, 0.00003106, 0.00002772, 0.00001936, 0.00001973, 0.00001997, 0.00004888, 0.00004761, -0.00004147, -0.00003009, -0.00001020, -0.00000617, -0.00001063, 0.00000181, -0.00001585, -0.00000730, -0.00000133, -0.00000496, -0.00000415, -0.00000133, -0.00000151, 0.00000201, -0.00000601, -0.00000459, -0.00000052, 0.00000004, -0.00000347, -0.00000100, -0.00000483, -0.00000289, -0.00000259, 0.00000317, -0.00000992, 0.00000464, -0.00000379, -0.00000064, 0.00000173, -0.00000570, 0.00001132, },
		{ -0.00000382, -0.00000184, 0.00000093, 0.00000860, -0.00000804, 0.00000373, -0.00000177, -0.00000057, -0.00000090, -0.00000835, 0.00000727, -0.00000475, -0.00000428, -0.00000318, -0.00000127, -0.00000304, -0.00000104, -0.00000139, -0.00000034, -0.00000783, -0.00000027, 0.00000018, -0.00000073, -0.00000433, -0.00000547, -0.00000455, 0.00000069, -0.00002011, -0.00000173, -0.00000721, -0.00000586, -0.00001372, -0.00001549, -0.00006127, 0.00000908, 0.00007852, 0.00002909, 0.00002351, 0.00002678, 0.00003290, 0.00004007, 0.00007927, -0.00003636, -0.00004805, -0.00001028, -0.00001219, -0.00000772, -0.00001377, 0.00000043, -0.00000543, -0.00000040, 0.00000572, -0.00000249, -0.00000111, -0.00000115, -0.00000482, -0.00000118, 0.00000154, -0.00000389, 0.00000260, -0.00000139, 0.00000066, 0.00000086, 0.00000176, -0.00000186, -0.00000185, -0.00000063, 0.00000198, 0.00000081, -0.00000056, -0.00000077, 0.00000141, -0.00000236, 0.00000320, -0.00000618, -0.00000177, -0.00000149, -0.00000010, -0.00000164, 0.00000724, -0.00000613, -0.00000014, -0.00000780, -0.00001050, -0.00000750, -0.00000848, -0.00002842, -0.00004970, 0.00003207, 0.00005935, 0.00002392, 0.00003011, 0.00001614, 0.00002324, 0.00002647, 0.00006857, 0.00000037, -0.00004928, -0.00001633, -0.00000760, -0.00000943, -0.00000412, -0.00000422, -0.00001736, -0.00000005, -0.00000469, -0.00000448, -0.00000325, -0.00000106, 0.00000024, -0.00000069, -0.00000752, -0.00000142, -0.00000037, -0.00000133, -0.00000287, -0.00000190, -0.00000443, -0.00000362, 0.00000048, -0.00000022, -0.00000835, 0.00000660, -0.00000840, 0.00000614, -0.00000599, 0.00000274, 0.00000654, },
		{ -0.00000198, -0.00000323, 0.00000608, 0.00000298, -0.00000672, 0.00000605, -0.00000621, 0.00000409, -0.00000726, -0.00000141, 0.00000378, -0.00000609, -0.00000335, -0.00000249, -0.00000208, -0.00000220, -0.00000115, -0.00000010, -0.00000465, -0.00000526, 0.00000133, -0.00000049, -0.00000239, -0.00000431, -0.00000681, 0.00000091, -0.00001021, -0.00001444, 0.00000044, -0.00000975, -0.00000729, -0.00001288, -0.00003507, -0.00004753, 0.00006292, 0.00005133, 0.00002398, 0.00001981, 0.00003318, 0.00002513, 0.00006270, 0.00004576, -0.00005845, -0.00002782, -0.00001158, -0.00000816, -0.00001116, -0.00000907, 0.00000081, -0.00000709, 0.00000535, 0.00000153, -0.00000214, -0.00000089, -0.00000187, -0.00000564, 0.00000222, -0.00000188, -0.00000068, 0.00000125, -0.00000125, 0.00000090, 0.00000158, 0.00000036, -0.00000209, -0.00000188, 0.00000103, 0.00000131, 0.00000087, -0.00000186, 0.00000179, -0.00000171, 0.00000051, 0.00000008, -0.00000579, -0.00000100, -0.00000082, -0.00000192, 0.00000295, 0.00000337, -0.00000699, 0.00000106, -0.00001265, -0.00000753, -0.00000962, -0.00001035, -0.00004361, -0.00003271, 0.00006171, 0.00004383, 0.00002665, 0.00002626, 0.00001906, 0.00002232, 0.00005000, 0.00005486, -0.00004422, -0.00003282, -0.00001135, -0.00000659, -0.00001040, 0.00000127, -0.00001525, -0.00000863, -0.00000036, -0.00000589, -0.00000419, -0.00000180, -0.00000098, 0.00000157, -0.00000553, -0.00000458, -0.00000055, -0.00000035, -0.00000285, -0.00000160, -0.00000347, -0.00000348, -0.00000450, 0.00000452, -0.00000662, -0.00000161, 0.00000154, -0.00000517, 0.00000584, -0.00000852, 0.00000904, 0.00000056, },
		},
		{ { -0.00000137, 0.00000110, -0.00001207, 0.00000422, -0.00000029, -0.00000365, -0.00000889, -0.00000355, 0.00001182, -0.00000376, -0.00000716, 0.00000200, 0.00000055, -0.00000255, -0.00000059, 0.00000130, -0.00000556, 0.00000118, 0.00000020, -0.00000258, -0.00000137, -0.00000199, 0.00001360, -0.00000528, -0.00000075, 0.00000032, 0.00000100, 0.00000130, 0.00000083, 0.00000056, -0.00000318, -0.00000161, -0.00000330, -0.00000513, 0.00000464, -0.00000659, -0.00000598, -0.00000752, -0.00001503, -0.00001442, -0.00006225, 0.00000676, 0.00007525, 0.00002559, 0.00002297, 0.00002146, 0.00003297, 0.00002660, 0.00007196, 0.00000540, -0.00006784, -0.00001466, -0.00001321, -0.00000884, -0.00000090, -0.00000340, -0.00000926, -0.00000076, -0.00000059, -0.00000051, -0.00000215, -0.00000356, -0.00000870, -0.00000204, -0.00000125, -0.00000204, -0.00000870, -0.00000356, -0.00000215, -0.00000051, -0.00000059, -0.00000076, -0.00000926, -0.00000340, -0.00000090, -0.00000884, -0.00001321, -0.00001466, -0.00006784, 0.00000540, 0.00007196, 0.00002660, 0.00003297, 0.00002146, 0.00002297, 0.00002559, 0.00007525, 0.00000676, -0.00006225, -0.00001442, -0.00001503, -0.00000752, -0.00000598, -0.00000659, 0.00000464, -0.00000513, -0.00000330, -0.00000161, -0.00000318, 0.00000056, 0.00000083, 0.00000130, 0.00000100, 0.00000032, -0.00000075, -0.00000528, 0.00001360, -0.00000199, -0.00000137, -0.00000258, 0.00000020, 0.00000118, -0.00000556, 0.00000130, -0.00000059, -0.00000255, 0.00000055, 0.00000200, -0.00000716, -0.00000376, 0.00001182, -0.00000355, -0.00000889, -0.00000365, -0.00000029, 0.00000422, -0.00001207, 0.00000110, },
		{ -0.00000021, -0.00000367, -0.00000980, 0.00000814, -0.00000576, -0.00000067, -0.00001300, 0.00000492, 0.00000804, -0.00000721, -0.00000445, 0.00000313, -0.00000092, -0.00000239, 0.00000050, 0.00000012, -0.00000562, 0.00000279, -0.00000184, -0.00000085, -0.00000385, 0.00000299, 0.00001092, -0.00000686, 0.00000103, -0.00000059, 0.00000213, 0.00000044, 0.00000146, -0.00000095, -0.00000321, -0.00000104, -0.00000452, -0.00000277, 0.00000436, -0.00000984, -0.00000276, -0.00001431, -0.00000938, -0.00003094, -0.00005358, 0.00004635, 0.00005803, 0.00001885, 0.00002139, 0.00002485, 0.00002911, 0.00003255, 0.00007000, -0.00002468, -0.00005608, -0.00000914, -0.00001408, -0.00000597, 0.00000033, -0.00000701, -0.00000678, 0.00000006, -0.00000061, -0.00000101, -0.00000247, -0.00000494, -0.00000770, -0.00000086, -0.00000134, -0.00000376, -0.00000878, -0.00000230, -0.00000164, -0.00000061, 0.00000009, -0.00000352, -0.00000895, -0.00000105, -0.00000284, -0.00001158, -0.00001098, -0.00002631, -0.00007043, 0.00003799, 0.00006391, 0.00002747, 0.00003162, 0.00002208, 0.00002200, 0.00004317, 0.00007434, -0.00003796, -0.00004888, -0.00000947, -0.00001589, -0.00000315, -0.00000910, -0.00000219, 0.00000278, -0.00000618, -0.00000241, -0.00000223, -0.00000247, 0.00000156, 0.00000021, 0.00000220, -0.00000048, 0.00000197, -0.00000384, -0.00000122, 0.00001334, -0.00000549, 0.00000102, -0.00000452, 0.00000279, -0.00000146, -0.00000429, 0.00000159, -0.00000164, -0.00000200, 0.00000151, 0.00000012, -0.00000929, 0.00000290, 0.00000950, -0.00000808, -0.00000627, -0.00000523, 0.00000543, -0.00000400, -0.00000788, 0.00000158, },
		{ 0.00000148, -0.00001040, -0.00000142, 0.00000455, -0.00000585, -0.00000138, -0.00001358, 0.00001258, 0.00000052, -0.00000664, -0.00000272, 0.00000405, -0.00000304, -0.00000096, 0.00000064, -0.00000102, -0.00000505, 0.00000359, -0.00000315, 0.00000015, -0.00000547, 0.00000813, 0.00000628, -0.00000617, 0.00000122, -0.00000016, 0.00000207, 0.00000051, 0.00000110, -0.00000201, -0.00000309, -0.00000034, -0.00000622, 0.00000127, 0.00000058, -0.00000909, -0.00000399, -0.00001720, -0.00000732, -0.00004948, -0.00002314, 0.00006581, 0.00003736, 0.00001831, 0.00001924, 0.00002933, 0.00002244, 0.00004414, 0.00005583, -0.00004552, -0.00004131, -0.00000785, -0.00001362, -0.00000275, -0.00000057, -0.00000955, -0.00000322, -0.00000066, 0.00000018, -0.00000238, -0.00000225, -0.00000649, -0.00000584, -0.00000043, -0.00000163, -0.00000574, -0.00000796, -0.00000110, -0.00000151, -0.00000043, 0.00000015, -0.00000678, -0.00000625, -0.00000096, -0.00000456, -0.00001400, -0.00000875, -0.00004347, -0.00005949, 0.00006512, 0.00005170, 0.00003200, 0.00002698, 0.00002566, 0.00002138, 0.00006859, 0.00004577, -0.00006690, -0.00002696, -0.00001390, -0.00001032, -0.00000408, -0.00000891, 0.00000054, 0.00000072, -0.00000686, -0.00000133, -0.00000319, -0.00000077, 0.00000132, 0.00000058, 0.00000198, -0.00000093, 0.00000288, -0.00000682, 0.00000442, 0.00001006, -0.00000661, 0.00000226, -0.00000577, 0.00000520, -0.00000472, -0.00000177, 0.00000042, -0.00000161, -0.00000176, 0.00000256, -0.00000348, -0.00000803, 0.00000872, 0.00000344, -0.00000825, -0.00000691, -0.00000227, 0.00000602, -0.00001025, -0.00000153, -0.00000037, },
		{ 0.00000014, -0.00001370, 0.00000747, -0.00000359, -0.00000048, -0.00000757, -0.00000719, 0.00001401, -0.00000556, -0.00000460, -0.00000121, 0.00000387, -0.00000451, 0.00000080, 0.00000011, -0.00000201, -0.00000396, 0.00000354, -0.00000353, -0.00000006, -0.00000525, 0.00001186, 0.00000132, -0.00000444, 0.00000080, 0.00000080, 0.00000143, 0.00000102, 0.00000003, -0.00000265, -0.00000242, -0.00000049, -0.00000687, 0.00000493, -0.00000483, -0.00000540, -0.00001006, -0.00001348, -0.00001569, -0.00005577, 0.00001570, 0.00006344, 0.00002319, 0.00001842, 0.00001970, 0.00003056, 0.00001812, 0.00005601, 0.00003243, -0.00005453, -0.00002837, -0.00000833, -0.00001223, 0.00000054, -0.00000402, -0.00000902, -0.00000102, -0.00000117, 0.00000059, -0.00000378, -0.00000187, -0.00000785, -0.00000351, -0.00000074, -0.00000210, -0.00000769, -0.00000619, -0.00000033, -0.00000163, 0.00000026, -0.00000138, -0.00000858, -0.00000302, -0.00000255, -0.00000596, -0.00001553, -0.00000903, -0.00006238, -0.00003471, 0.00008107, 0.00004146, 0.00003527, 0.00002365, 0.00002802, 0.00002870, 0.00008801, -0.00000468, -0.00006752, -0.00001309, -0.00001769, -0.00000405, -0.00000727, -0.00000650, 0.00000166, -0.00000147, -0.00000678, -0.00000068, -0.00000370, 0.00000099, 0.00000048, 0.00000160, 0.00000072, 0.00000008, 0.00000203, -0.00000820, 0.00001010, 0.00000446, -0.00000491, 0.00000136, -0.00000511, 0.00000603, -0.00000722, 0.00000082, -0.00000142, -0.00000074, -0.00000179, 0.00000319, -0.00000802, -0.00000240, 0.00000984, -0.00000179, -0.00000736, -0.00000826, 0.00000353, 0.00000042, -0.00001033, 0.00000216, -0.00000130, },
		{ -0.00000555, -0.00000958, 0.00001003, -0.00000920, 0.00000483, -0.00001522, 0.00000405, 0.00000805, -0.00000677, -0.00000383, 0.00000129, 0.00000162, -0.00000389, 0.00000135, -0.00000005, -0.00000339, -0.00000203, 0.00000243, -0.00000279, -0.00000169, -0.00000266, 0.00001304, -0.00000254, -0.00000290, 0.00000086, 0.00000110, 0.00000137, 0.00000083, -0.00000080, -0.00000341, -0.00000067, -0.00000257, -0.00000458, 0.00000512, -0.00000781, -0.00000347, -0.00001597, -0.00000790, -0.00003221, -0.00004208, 0.00004647, 0.00004796, 0.00001745, 0.00001696, 0.00002358, 0.00002629, 0.00002049, 0.00006147, 0.00000659, -0.00005393, -0.00001891, -0.00000936, -0.00000965, 0.00000236, -0.00000832, -0.00000575, -0.00000115, -0.00000030, -0.00000049, -0.00000408, -0.00000213, -0.00000840, -0.00000136, -0.00000153, -0.00000283, -0.00000926, -0.00000355, -0.00000059, -0.00000121, 0.00000055, -0.00000397, -0.00000810, -0.00000101, -0.00000460, -0.00000737, -0.00001581, -0.00001426, -0.00007729, -0.00000013, 0.00008446, 0.00003637, 0.00003444, 0.00002480, 0.00002674, 0.00004972, 0.00008291, -0.00005529, -0.00004554, -0.00001367, -0.00001371, -0.00000307, -0.00000808, -0.00000490, 0.00000283, -0.00000471, -0.00000494, -0.00000153, -0.00000254, 0.00000136, 0.00000050, 0.00000188, -0.00000031, 0.00000166, -0.00000042, -0.00000710, 0.00001413, -0.00000193, -0.00000102, -0.00000181, -0.00000190, 0.00000414, -0.00000750, 0.00000188, -0.00000236, -0.00000034, -0.00000109, 0.00000176, -0.00001051, 0.00000497, 0.00000602, -0.00000359, -0.00000848, -0.00000653, 0.00000722, -0.00000735, -0.00000507, 0.00000173, 0.00000005, },
		{ -0.00001241, 0.00000043, 0.00000400, -0.00000708, 0.00000381, -0.00001743, 0.00001335, -0.00000093, -0.00000416, -0.00000441, 0.00000423, -0.00000195, -0.00000136, 0.00000037, 0.00000045, -0.00000514, 0.00000055, 0.00000057, -0.00000145, -0.00000409, 0.00000181, 0.00001159, -0.00000490, -0.00000169, 0.00000136, 0.00000075, 0.00000196, -0.00000026, -0.00000115, -0.00000394, 0.00000129, -0.00000568, 0.00000014, 0.00000128, -0.00000647, -0.00000674, -0.00001629, -0.00000850, -0.00004612, -0.00001266, 0.00005956, 0.00003056, 0.00001675, 0.00001533, 0.00002837, 0.00001833, 0.00002991, 0.00005656, -0.00001489, -0.00004832, -0.00001209, -0.00001088, -0.00000539, 0.00000090, -0.00001045, -0.00000232, -0.00000224, 0.00000126, -0.00000279, -0.00000304, -0.00000341, -0.00000770, -0.00000001, -0.00000242, -0.00000403, -0.00000980, -0.00000074, -0.00000175, 0.00000017, -0.00000061, -0.00000614, -0.00000599, -0.00000100, -0.00000604, -0.00000926, -0.00001499, -0.00002569, -0.00008234, 0.00003723, 0.00007823, 0.00003542, 0.00003099, 0.00002881, 0.00002599, 0.00007939, 0.00004404, -0.00008120, -0.00002068, -0.00002098, -0.00000462, -0.00000698, -0.00000582, -0.00000473, 0.00000401, -0.00000837, -0.00000185, -0.00000351, 0.00000001, 0.00000020, 0.00000178, 0.00000080, -0.00000008, 0.00000261, -0.00000365, -0.00000329, 0.00001490, -0.00000681, 0.00000309, -0.00000579, 0.00000286, -0.00000007, -0.00000545, 0.00000099, -0.00000172, -0.00000102, 0.00000066, -0.00000244, -0.00000819, 0.00000953, 0.00000071, -0.00000321, -0.00001129, -0.00000069, 0.00000499, -0.00001161, 0.00000100, -0.00000047, 0.00000134, },
		{ -0.00001462, 0.00000946, -0.00000583, 0.00000120, -0.00000459, -0.00001066, 0.00001469, -0.00000670, -0.00000166, -0.00000419, 0.00000551, -0.00000446, 0.00000105, -0.00000079, 0.00000062, -0.00000624, 0.00000272, -0.00000097, -0.00000074, -0.00000576, 0.00000667, 0.00000845, -0.00000626, -0.00000011, 0.00000124, 0.00000089, 0.00000199, -0.00000120, -0.00000187, -0.00000298, 0.00000131, -0.00000689, 0.00000381, -0.00000320, -0.00000376, -0.00001308, -0.00001146, -0.00001846, -0.00004670, 0.00002020, 0.00005534, 0.00001870, 0.00001674, 0.00001611, 0.00002995, 0.00001174, 0.00004193, 0.00004251, -0.00002891, -0.00004116, -0.00000710, -0.00001239, -0.00000027, -0.00000385, -0.00000866, -0.00000129, -0.00000211, 0.00000171, -0.00000481, -0.00000150, -0.00000533, -0.00000583, 0.00000016, -0.00000304, -0.00000586, -0.00000857, 0.00000091, -0.00000262, 0.00000160, -0.00000308, -0.00000684, -0.00000348, -0.00000285, -0.00000615, -0.00001216, -0.00001367, -0.00004283, -0.00007333, 0.00006988, 0.00006741, 0.00003533, 0.00002912, 0.00003084, 0.00003541, 0.00009927, -0.00001721, -0.00007240, -0.00001077, -0.00002225, 0.00000093, -0.00001000, -0.00000407, -0.00000355, 0.00000293, -0.00000988, 0.00000005, -0.00000445, 0.00000189, -0.00000073, 0.00000291, -0.00000062, 0.00000104, 0.00000250, -0.00000673, 0.00000272, 0.00001154, -0.00000805, 0.00000489, -0.00000810, 0.00000678, -0.00000462, -0.00000258, -0.00000059, -0.00000051, -0.00000185, 0.00000193, -0.00000745, -0.00000142, 0.00000863, -0.00000223, -0.00000387, -0.00001255, 0.00000613, -0.00000251, -0.00000945, 0.00000364, -0.00000113, -0.00000094, },
		{ -0.00000876, 0.00001050, -0.00001112, 0.00000817, -0.00001445, 0.00000154, 0.00000796, -0.00000649, -0.00000186, -0.00000174, 0.00000385, -0.00000419, 0.00000149, -0.00000065, -0.00000067, -0.00000562, 0.00000351, -0.00000133, -0.00000161, -0.00000535, 0.00001042, 0.00000458, -0.00000682, 0.00000195, 0.00000000, 0.00000227, 0.00000054, -0.00000113, -0.00000332, -0.00000020, -0.00000159, -0.00000424, 0.00000325, -0.00000419, -0.00000471, -0.00001656, -0.00000801, -0.00003223, -0.00003121, 0.00004463, 0.00004075, 0.00001433, 0.00001513, 0.00002042, 0.00002555, 0.00001161, 0.00004992, 0.00002478, -0.00003677, -0.00003312, -0.00000448, -0.00001222, 0.00000310, -0.00000913, -0.00000433, -0.00000277, -0.00000008, 0.00000005, -0.00000522, -0.00000062, -0.00000705, -0.00000339, -0.00000075, -0.00000334, -0.00000804, -0.00000549, 0.00000044, -0.00000181, 0.00000138, -0.00000535, -0.00000626, -0.00000157, -0.00000576, -0.00000501, -0.00001622, -0.00001298, -0.00006311, -0.00004929, 0.00009242, 0.00005604, 0.00003426, 0.00003099, 0.00002913, 0.00006075, 0.00008771, -0.00007211, -0.00004174, -0.00001769, -0.00001274, -0.00000194, -0.00000796, -0.00000563, 0.00000036, -0.00000180, -0.00000752, -0.00000104, -0.00000264, 0.00000130, 0.00000031, 0.00000220, -0.00000073, 0.00000187, 0.00000150, -0.00000864, 0.00000939, 0.00000487, -0.00000516, 0.00000309, -0.00000701, 0.00000780, -0.00000752, -0.00000057, -0.00000137, -0.00000013, -0.00000148, 0.00000061, -0.00000955, 0.00000594, 0.00000362, -0.00000181, -0.00000726, -0.00000932, 0.00000904, -0.00001051, -0.00000279, 0.00000222, 0.00000040, -0.00000745, },
		{ 0.00000225, 0.00000259, -0.00000714, 0.00000697, -0.00001775, 0.00001112, -0.00000110, -0.00000274, -0.00000381, 0.00000173, 0.00000057, -0.00000193, 0.00000022, 0.00000064, -0.00000318, -0.00000343, 0.00000307, -0.00000088, -0.00000375, -0.00000272, 0.00001240, 0.00000030, -0.00000595, 0.00000334, -0.00000120, 0.00000388, -0.00000166, -0.00000068, -0.00000425, 0.00000239, -0.00000525, 0.00000047, -0.00000063, -0.00000164, -0.00001060, -0.00001389, -0.00001209, -0.00003988, -0.00000575, 0.00005481, 0.00002422, 0.00001518, 0.00001260, 0.00002636, 0.00001633, 0.00001927, 0.00004948, 0.00000891, -0.00004140, -0.00002358, -0.00000513, -0.00000891, 0.00000213, -0.00001117, -0.00000097, -0.00000436, 0.00000227, -0.00000288, -0.00000394, -0.00000098, -0.00000785, -0.00000118, -0.00000221, -0.00000373, -0.00000959, -0.00000174, -0.00000153, 0.00000065, -0.00000122, -0.00000602, -0.00000518, -0.00000098, -0.00000824, -0.00000399, -0.00002045, -0.00001497, -0.00008179, -0.00001304, 0.00010288, 0.00004551, 0.00003389, 0.00003362, 0.00003022, 0.00009328, 0.00003776, -0.00009333, -0.00001470, -0.00002802, 0.00000027, -0.00000920, -0.00000309, -0.00000863, 0.00000484, -0.00000790, -0.00000296, -0.00000388, 0.00000085, -0.00000103, 0.00000287, -0.00000020, 0.00000079, 0.00000183, -0.00000031, -0.00000821, 0.00001425, -0.00000243, -0.00000001, -0.00000129, -0.00000293, 0.00000568, -0.00000827, 0.00000030, -0.00000120, -0.00000061, -0.00000020, -0.00000338, -0.00000648, 0.00000943, -0.00000151, -0.00000030, -0.00001182, -0.00000194, 0.00000528, -0.00001364, 0.00000349, -0.00000033, 0.00000121, -0.00001410, },
		{ 0.00001046, -0.00000769, 0.00000273, -0.00000226, -0.00001148, 0.00001226, -0.00000610, 0.00000008, -0.00000472, 0.00000366, -0.00000162, -0.00000014, -0.00000081, 0.00000143, -0.00000534, -0.00000100, 0.00000256, -0.00000096, -0.00000568, 0.00000117, 0.00001265, -0.00000423, -0.00000302, 0.00000261, -0.00000060, 0.00000379, -0.00000284, -0.00000125, -0.00000289, 0.00000201, -0.00000592, 0.00000280, -0.00000320, -0.00000031, -0.00001683, -0.00000869, -0.00002279, -0.00003527, 0.00002025, 0.00005115, 0.00001232, 0.00001708, 0.00001181, 0.00002971, 0.00000735, 0.00003084, 0.00004139, -0.00000320, -0.00004382, -0.00001334, -0.00000814, -0.00000338, -0.00000296, -0.00000847, -0.00000108, -0.00000364, 0.00000273, -0.00000517, -0.00000200, -0.00000239, -0.00000741, 0.00000005, -0.00000351, -0.00000475, -0.00000930, 0.00000071, -0.00000298, 0.00000283, -0.00000491, -0.00000511, -0.00000400, -0.00000232, -0.00000852, -0.00000545, -0.00002273, -0.00002246, -0.00009266, 0.00002967, 0.00010192, 0.00003607, 0.00003705, 0.00003210, 0.00004543, 0.00010958, -0.00003218, -0.00007528, -0.00000880, -0.00002722, 0.00000573, -0.00001265, -0.00000114, -0.00000849, 0.00000555, -0.00001094, -0.00000015, -0.00000511, 0.00000292, -0.00000230, 0.00000447, -0.00000217, 0.00000255, 0.00000141, -0.00000300, -0.00000463, 0.00001506, -0.00000719, 0.00000437, -0.00000566, 0.00000206, 0.00000176, -0.00000765, 0.00000091, -0.00000120, -0.00000060, -0.00000009, -0.00000723, 0.00000022, 0.00000752, -0.00000359, -0.00000070, -0.00001406, 0.00000577, -0.00000348, -0.00000988, 0.00000549, -0.00000080, -0.00000214, -0.00001478, },
		{ 0.00000929, -0.00001141, 0.00000968, -0.00001219, -0.00000042, 0.00000606, -0.00000526, -0.00000011, -0.00000343, 0.00000316, -0.00000150, -0.00000029, -0.00000023, 0.00000047, -0.00000593, 0.00000067, 0.00000267, -0.00000244, -0.00000611, 0.00000531, 0.00001106, -0.00000795, 0.00000109, -0.00000004, 0.00000207, 0.00000137, -0.00000235, -0.00000284, 0.00000048, -0.00000164, -0.00000236, 0.00000036, -0.00000111, -0.00000469, -0.00001741, -0.00000829, -0.00003235, -0.00001998, 0.00003974, 0.00003781, 0.00000821, 0.00001664, 0.00001493, 0.00002679, 0.00000440, 0.00003978, 0.00003042, -0.00001342, -0.00004242, -0.00000523, -0.00001053, 0.00000099, -0.00000861, -0.00000336, -0.00000385, -0.00000057, 0.00000062, -0.00000555, -0.00000055, -0.00000422, -0.00000590, -0.00000001, -0.00000430, -0.00000639, -0.00000687, 0.00000066, -0.00000213, 0.00000246, -0.00000733, -0.00000384, -0.00000279, -0.00000544, -0.00000593, -0.00001100, -0.00002138, -0.00003745, -0.00009001, 0.00007204, 0.00009083, 0.00002988, 0.00004263, 0.00002744, 0.00007829, 0.00008676, -0.00008715, -0.00003828, -0.00002070, -0.00001351, 0.00000036, -0.00000895, -0.00000451, -0.00000359, 0.00000092, -0.00000886, -0.00000122, -0.00000285, 0.00000167, -0.00000065, 0.00000324, -0.00000178, 0.00000298, 0.00000120, -0.00000610, 0.00000145, 0.00001144, -0.00000780, 0.00000588, -0.00000793, 0.00000597, -0.00000262, -0.00000623, 0.00000162, -0.00000190, 0.00000059, -0.00000276, -0.00000739, 0.00000629, 0.00000240, -0.00000251, -0.00000402, -0.00001154, 0.00000906, -0.00001201, -0.00000193, 0.00000332, 0.00000080, -0.00000946, -0.00000705, },
		{ -0.00000035, -0.00000521, 0.00000734, -0.00001478, 0.00000714, -0.00000116, -0.00000180, -0.00000174, -0.00000152, 0.00000197, -0.00000055, -0.00000134, 0.00000121, -0.00000166, -0.00000522, 0.00000192, 0.00000274, -0.00000479, -0.00000487, 0.00000923, 0.00000714, -0.00000903, 0.00000399, -0.00000235, 0.00000483, -0.00000183, -0.00000148, -0.00000350, 0.00000286, -0.00000510, 0.00000207, -0.00000397, 0.00000328, -0.00001341, -0.00001186, -0.00001598, -0.00003306, -0.00000033, 0.00004941, 0.00002086, 0.00001094, 0.00001340, 0.00002119, 0.00001771, 0.00000971, 0.00004168, 0.00002091, -0.00002393, -0.00003556, -0.00000181, -0.00000973, 0.00000105, -0.00001059, -0.00000003, -0.00000607, 0.00000250, -0.00000263, -0.00000427, -0.00000011, -0.00000583, -0.00000386, -0.00000110, -0.00000478, -0.00000773, -0.00000353, -0.00000108, 0.00000066, -0.00000094, -0.00000713, -0.00000317, -0.00000210, -0.00000876, -0.00000218, -0.00001957, -0.00001722, -0.00005914, -0.00007052, 0.00010770, 0.00007116, 0.00003130, 0.00004419, 0.00003068, 0.00011370, 0.00002335, -0.00010019, -0.00001167, -0.00003236, 0.00000191, -0.00000875, -0.00000270, -0.00000936, 0.00000234, -0.00000558, -0.00000429, -0.00000416, 0.00000110, -0.00000139, 0.00000272, 0.00000001, 0.00000071, 0.00000197, 0.00000093, -0.00000824, 0.00000784, 0.00000525, -0.00000504, 0.00000450, -0.00000764, 0.00000806, -0.00000700, -0.00000355, 0.00000139, -0.00000211, 0.00000146, -0.00000695, -0.00000296, 0.00000824, -0.00000230, -0.00000049, -0.00000857, -0.00000494, 0.00000557, -0.00001505, 0.00000494, 0.00000043, 0.00000102, -0.00001561, 0.00000453, },
		{ -0.00001007, 0.00000538, -0.00000227, -0.00000856, 0.00000677, -0.00000401, 0.00000025, -0.00000241, -0.00000097, 0.00000229, -0.00000107, -0.00000114, 0.00000158, -0.00000333, -0.00000440, 0.00000373, 0.00000140, -0.00000654, -0.00000256, 0.00001251, 0.00000117, -0.00000650, 0.00000371, -0.00000203, 0.00000525, -0.00000366, -0.00000176, -0.00000153, 0.00000147, -0.00000441, 0.00000244, -0.00000465, 0.00000374, -0.00001970, -0.00000664, -0.00002690, -0.00002372, 0.00001819, 0.00004838, 0.00000698, 0.00001614, 0.00001008, 0.00002658, 0.00000710, 0.00002010, 0.00003722, 0.00001335, -0.00003393, -0.00002454, -0.00000271, -0.00000619, -0.00000296, -0.00000772, -0.00000077, -0.00000522, 0.00000312, -0.00000486, -0.00000251, -0.00000052, -0.00000683, -0.00000200, -0.00000267, -0.00000533, -0.00000773, -0.00000122, -0.00000228, 0.00000302, -0.00000543, -0.00000497, -0.00000306, -0.00000283, -0.00000991, -0.00000082, -0.00002749, -0.00001402, -0.00008328, -0.00003395, 0.00013040, 0.00004711, 0.00004215, 0.00003638, 0.00005514, 0.00012228, -0.00005310, -0.00007286, -0.00001018, -0.00002943, 0.00000722, -0.00001241, -0.00000066, -0.00001013, 0.00000407, -0.00000878, -0.00000179, -0.00000507, 0.00000312, -0.00000292, 0.00000472, -0.00000241, 0.00000317, 0.00000070, -0.00000020, -0.00000800, 0.00001213, -0.00000085, -0.00000087, 0.00000142, -0.00000534, 0.00000810, -0.00001092, 0.00000062, -0.00000084, -0.00000019, -0.00000059, -0.00000897, 0.00000325, 0.00000583, -0.00000441, 0.00000018, -0.00001150, 0.00000227, -0.00000287, -0.00001090, 0.00000698, -0.00000003, -0.00000346, -0.00001442, 0.00001109, },
		{ -0.00001097, 0.00001070, -0.00001057, 0.00000025, 0.00000128, -0.00000240, 0.00000014, -0.00000211, -0.00000133, 0.00000385, -0.00000329, 0.00000087, 0.00000014, -0.00000380, -0.00000378, 0.00000608, -0.00000172, -0.00000685, 0.00000053, 0.00001398, -0.00000495, -0.00000179, 0.00000085, 0.00000092, 0.00000290, -0.00000373, -0.00000264, 0.00000183, -0.00000275, 0.00000030, -0.00000218, 0.00000031, -0.00000321, -0.00001823, -0.00000894, -0.00003217, -0.00000932, 0.00003302, 0.00003758, 0.00000119, 0.00001881, 0.00001009, 0.00002667, 0.00000098, 0.00002957, 0.00003055, 0.00000508, -0.00003994, -0.00001337, -0.00000489, -0.00000310, -0.00000736, -0.00000284, -0.00000412, -0.00000178, 0.00000097, -0.00000503, -0.00000126, -0.00000141, -0.00000706, -0.00000080, -0.00000432, -0.00000584, -0.00000634, -0.00000080, -0.00000145, 0.00000274, -0.00000839, -0.00000253, -0.00000317, -0.00000513, -0.00000777, -0.00000462, -0.00003076, -0.00001669, -0.00010325, 0.00001699, 0.00013394, 0.00002784, 0.00005613, 0.00002488, 0.00009991, 0.00008202, -0.00010248, -0.00003313, -0.00002452, -0.00001391, 0.00000150, -0.00000883, -0.00000405, -0.00000603, 0.00000070, -0.00000723, -0.00000293, -0.00000256, 0.00000173, -0.00000135, 0.00000368, -0.00000213, 0.00000385, 0.00000014, -0.00000233, -0.00000522, 0.00001330, -0.00000523, 0.00000323, -0.00000245, -0.00000135, 0.00000538, -0.00001268, 0.00000461, -0.00000402, 0.00000324, -0.00000600, -0.00000608, 0.00000719, 0.00000153, -0.00000414, -0.00000116, -0.00001093, 0.00000593, -0.00001120, -0.00000257, 0.00000474, 0.00000138, -0.00001146, -0.00000469, 0.00000702, },
		{ -0.00000187, 0.00000587, -0.00001046, 0.00000384, -0.00000286, -0.00000017, 0.00000002, -0.00000291, -0.00000004, 0.00000394, -0.00000481, 0.00000285, -0.00000177, -0.00000399, -0.00000225, 0.00000738, -0.00000532, -0.00000603, 0.00000457, 0.00001224, -0.00000842, 0.00000201, -0.00000182, 0.00000403, -0.00000033, -0.00000353, -0.00000192, 0.00000295, -0.00000524, 0.00000415, -0.00000697, 0.00000625, -0.00001364, -0.00001128, -0.00001890, -0.00002701, 0.00000468, 0.00004256, 0.00002096, 0.00000413, 0.00001694, 0.00001418, 0.00002037, 0.00000229, 0.00003400, 0.00002490, -0.00000576, -0.00003931, -0.00000558, -0.00000569, -0.00000314, -0.00000835, -0.00000011, -0.00000655, 0.00000148, -0.00000213, -0.00000375, -0.00000071, -0.00000256, -0.00000660, -0.00000043, -0.00000592, -0.00000566, -0.00000475, -0.00000123, 0.00000064, -0.00000027, -0.00000866, -0.00000088, -0.00000379, -0.00000756, -0.00000394, -0.00001328, -0.00002791, -0.00002889, -0.00011128, 0.00007578, 0.00011497, 0.00002424, 0.00006050, 0.00002783, 0.00014033, 0.00000079, -0.00010183, -0.00001155, -0.00003381, 0.00000041, -0.00000616, -0.00000413, -0.00000782, -0.00000177, -0.00000346, -0.00000440, -0.00000492, 0.00000087, -0.00000092, 0.00000152, 0.00000098, 0.00000013, 0.00000288, 0.00000010, -0.00000460, -0.00000101, 0.00001187, -0.00000754, 0.00000667, -0.00000666, 0.00000416, -0.00000083, -0.00001009, 0.00000541, -0.00000502, 0.00000478, -0.00001154, 0.00000063, 0.00000696, -0.00000212, -0.00000283, -0.00000363, -0.00000717, 0.00000386, -0.00001442, 0.00000458, 0.00000197, 0.00000072, -0.00001687, 0.00000715, -0.00000443, },
		{ 0.00000921, -0.00000452, -0.00000298, 0.00000019, -0.00000214, -0.00000006, 0.00000118, -0.00000565, 0.00000423, 0.00000048, -0.00000325, 0.00000257, -0.00000238, -0.00000509, 0.00000130, 0.00000592, -0.00000768, -0.00000469, 0.00000920, 0.00000719, -0.00000787, 0.00000273, -0.00000193, 0.00000495, -0.00000254, -0.00000400, 0.00000128, -0.00000020, -0.00000255, 0.00000236, -0.00000609, 0.00000615, -0.00001959, -0.00000709, -0.00002874, -0.00001452, 0.00001663, 0.00004430, 0.00000532, 0.00001159, 0.00001264, 0.00001939, 0.00001105, 0.00000907, 0.00003366, 0.00001991, -0.00001759, -0.00003294, -0.00000162, -0.00000532, -0.00000561, -0.00000551, -0.00000117, -0.00000588, 0.00000226, -0.00000399, -0.00000237, -0.00000046, -0.00000395, -0.00000551, -0.00000103, -0.00000712, -0.00000456, -0.00000414, -0.00000080, 0.00000188, -0.00000410, -0.00000685, -0.00000026, -0.00000535, -0.00000818, -0.00000167, -0.00002362, -0.00002066, -0.00005173, -0.00009925, 0.00013041, 0.00007879, 0.00003984, 0.00004775, 0.00006172, 0.00013955, -0.00008204, -0.00006332, -0.00001671, -0.00002679, 0.00000347, -0.00000780, -0.00000402, -0.00000735, -0.00000166, -0.00000405, -0.00000426, -0.00000426, 0.00000188, -0.00000172, 0.00000281, -0.00000067, 0.00000224, 0.00000146, -0.00000008, -0.00000617, 0.00000354, 0.00000850, -0.00000753, 0.00000853, -0.00000995, 0.00000959, -0.00000926, -0.00000310, 0.00000162, -0.00000167, 0.00000114, -0.00001245, 0.00000691, 0.00000387, -0.00000432, -0.00000144, -0.00000584, -0.00000249, -0.00000256, -0.00001091, 0.00000701, 0.00000160, -0.00000497, -0.00001360, 0.00001157, -0.00001306, },
		{ 0.00001211, -0.00001064, 0.00000386, -0.00000560, 0.00000118, -0.00000083, 0.00000174, -0.00000779, 0.00000911, -0.00000505, 0.00000089, 0.00000001, -0.00000169, -0.00000658, 0.00000612, 0.00000174, -0.00000820, -0.00000268, 0.00001278, 0.00000094, -0.00000475, 0.00000110, 0.00000018, 0.00000363, -0.00000386, -0.00000375, 0.00000437, -0.00000492, 0.00000329, -0.00000408, 0.00000065, -0.00000184, -0.00001721, -0.00001150, -0.00002993, -0.00000153, 0.00002736, 0.00003661, -0.00000290, 0.00001767, 0.00000965, 0.00002173, 0.00000359, 0.00001703, 0.00003119, 0.00001323, -0.00002669, -0.00002426, 0.00000032, -0.00000598, -0.00000721, -0.00000192, -0.00000413, -0.00000300, 0.00000056, -0.00000383, -0.00000177, -0.00000014, -0.00000564, -0.00000382, -0.00000281, -0.00000705, -0.00000352, -0.00000411, 0.00000070, 0.00000114, -0.00000698, -0.00000396, -0.00000104, -0.00000709, -0.00000682, -0.00000312, -0.00003197, -0.00001285, -0.00008318, -0.00006056, 0.00016442, 0.00004170, 0.00006367, 0.00002824, 0.00012031, 0.00007762, -0.00012164, -0.00002329, -0.00003195, -0.00001083, -0.00000166, -0.00000480, -0.00000704, -0.00000424, -0.00000433, -0.00000184, -0.00000651, -0.00000109, 0.00000032, -0.00000030, 0.00000191, -0.00000029, 0.00000303, 0.00000028, -0.00000045, -0.00000705, 0.00000823, 0.00000312, -0.00000426, 0.00000699, -0.00000984, 0.00001154, -0.00001577, 0.00000467, -0.00000420, 0.00000414, -0.00000697, -0.00000687, 0.00000903, 0.00000049, -0.00000541, -0.00000027, -0.00000683, 0.00000018, -0.00000912, -0.00000355, 0.00000551, 0.00000249, -0.00001341, -0.00000205, 0.00000436, -0.00001055, },
		{ 0.00000406, -0.00000685, 0.00000301, -0.00000690, 0.00000250, 0.00000049, -0.00000129, -0.00000542, 0.00001011, -0.00000849, 0.00000436, -0.00000249, -0.00000153, -0.00000637, 0.00000969, -0.00000322, -0.00000762, 0.00000067, 0.00001340, -0.00000360, -0.00000195, -0.00000030, 0.00000220, 0.00000180, -0.00000556, -0.00000063, 0.00000336, -0.00000580, 0.00000609, -0.00000869, 0.00000676, -0.00001177, -0.00001123, -0.00002152, -0.00002109, 0.00000825, 0.00003590, 0.00002195, -0.00000188, 0.00001921, 0.00000964, 0.00001975, 0.00000053, 0.00002351, 0.00002786, 0.00000424, -0.00003089, -0.00001590, 0.00000130, -0.00000853, -0.00000562, -0.00000083, -0.00000585, -0.00000049, -0.00000175, -0.00000249, -0.00000185, 0.00000019, -0.00000744, -0.00000177, -0.00000572, -0.00000493, -0.00000409, -0.00000291, 0.00000158, -0.00000067, -0.00000859, -0.00000051, -0.00000393, -0.00000709, -0.00000553, -0.00000794, -0.00003592, -0.00000963, -0.00011647, 0.00000492, 0.00016433, 0.00002385, 0.00007547, 0.00002824, 0.00016755, -0.00002534, -0.00010209, -0.00001124, -0.00003506, -0.00000126, -0.00000458, -0.00000439, -0.00000715, -0.00000409, -0.00000441, -0.00000152, -0.00000711, 0.00000105, -0.00000080, 0.00000081, 0.00000095, 0.00000077, 0.00000302, -0.00000084, -0.00000071, -0.00000744, 0.00001285, -0.00000397, 0.00000251, 0.00000099, -0.00000457, 0.00000708, -0.00001588, 0.00000801, -0.00000694, 0.00000715, -0.00001435, 0.00000214, 0.00000680, -0.00000202, -0.00000549, 0.00000043, -0.00000648, -0.00000085, -0.00001154, 0.00000273, 0.00000377, 0.00000049, -0.00001765, 0.00000959, -0.00000867, 0.00000166, },
		{ -0.00000749, 0.00000287, -0.00000481, -0.00000233, 0.00000050, 0.00000403, -0.00000771, 0.00000226, 0.00000500, -0.00000687, 0.00000435, -0.00000278, -0.00000335, -0.00000293, 0.00000976, -0.00000690, -0.00000666, 0.00000515, 0.00001075, -0.00000516, -0.00000106, 0.00000007, 0.00000264, 0.00000033, -0.00000730, 0.00000472, -0.00000229, -0.00000062, 0.00000199, -0.00000619, 0.00000552, -0.00001582, -0.00001015, -0.00002795, -0.00000833, 0.00001631, 0.00003849, 0.00000698, 0.00000479, 0.00001720, 0.00001118, 0.00001526, 0.00000108, 0.00002858, 0.00002245, -0.00000471, -0.00003121, -0.00000802, 0.00000008, -0.00001077, -0.00000204, -0.00000255, -0.00000512, -0.00000002, -0.00000281, -0.00000140, -0.00000180, -0.00000012, -0.00000861, -0.00000026, -0.00000857, -0.00000154, -0.00000644, 0.00000031, 0.00000031, -0.00000205, -0.00000914, 0.00000267, -0.00000843, -0.00000455, -0.00000631, -0.00001469, -0.00003397, -0.00001751, -0.00013839, 0.00008501, 0.00013068, 0.00003425, 0.00006373, 0.00006823, 0.00015813, -0.00011606, -0.00004919, -0.00002647, -0.00002067, -0.00000394, -0.00000072, -0.00000937, -0.00000231, -0.00000887, 0.00000045, -0.00000566, -0.00000362, -0.00000021, 0.00000064, -0.00000036, 0.00000184, 0.00000092, 0.00000288, -0.00000176, -0.00000137, -0.00000633, 0.00001561, -0.00001036, 0.00001038, -0.00000745, 0.00000417, -0.00000285, -0.00000886, 0.00000473, -0.00000371, 0.00000335, -0.00001534, 0.00000907, 0.00000292, -0.00000413, -0.00000407, 0.00000029, -0.00000564, -0.00000465, -0.00000861, 0.00000501, 0.00000407, -0.00000664, -0.00001218, 0.00001155, -0.00001586, 0.00001307, },
		{ -0.00001170, 0.00000871, -0.00001205, 0.00000340, -0.00000148, 0.00000590, -0.00001286, 0.00001095, -0.00000342, -0.00000138, 0.00000125, -0.00000126, -0.00000624, 0.00000275, 0.00000639, -0.00000869, -0.00000503, 0.00000914, 0.00000667, -0.00000508, -0.00000098, 0.00000116, 0.00000234, -0.00000191, -0.00000636, 0.00000788, -0.00000771, 0.00000628, -0.00000569, 0.00000132, -0.00000254, -0.00001279, -0.00001697, -0.00002451, 0.00000177, 0.00002488, 0.00003244, -0.00000191, 0.00001178, 0.00001416, 0.00001227, 0.00001058, 0.00000382, 0.00003247, 0.00001396, -0.00001091, -0.00002969, -0.00000015, -0.00000473, -0.00000954, 0.00000010, -0.00000463, -0.00000332, -0.00000111, -0.00000216, -0.00000146, -0.00000085, -0.00000192, -0.00000809, -0.00000077, -0.00000914, 0.00000049, -0.00000817, 0.00000371, -0.00000221, -0.00000315, -0.00000785, 0.00000329, -0.00001155, -0.00000187, -0.00000850, -0.00002296, -0.00002520, -0.00004265, -0.00013177, 0.00015546, 0.00008261, 0.00006093, 0.00004149, 0.00013703, 0.00007476, -0.00014515, -0.00000856, -0.00004339, -0.00000335, -0.00000992, 0.00000350, -0.00001377, 0.00000173, -0.00001319, 0.00000575, -0.00001077, 0.00000124, -0.00000275, 0.00000286, -0.00000231, 0.00000372, 0.00000053, 0.00000211, -0.00000145, -0.00000385, -0.00000174, 0.00001349, -0.00001189, 0.00001454, -0.00001337, 0.00001107, -0.00001275, 0.00000065, -0.00000183, 0.00000283, -0.00000571, -0.00000879, 0.00001058, -0.00000015, -0.00000623, -0.00000108, -0.00000083, -0.00000570, -0.00000804, -0.00000307, 0.00000452, 0.00000446, -0.00001528, 0.00000077, 0.00000146, -0.00000999, 0.00001351, },
		{ -0.00000462, 0.00000444, -0.00001197, 0.00000492, 0.00000039, 0.00000211, -0.00001140, 0.00001445, -0.00000948, 0.00000371, -0.00000195, -0.00000033, -0.00000755, 0.00000776, 0.00000159, -0.00000931, -0.00000203, 0.00001091, 0.00000339, -0.00000540, 0.00000025, 0.00000105, 0.00000275, -0.00000595, -0.00000071, 0.00000465, -0.00000685, 0.00000769, -0.00000924, 0.00000593, -0.00000984, -0.00000966, -0.00002540, -0.00001379, 0.00000850, 0.00003210, 0.00002027, -0.00000313, 0.00001635, 0.00001101, 0.00001292, 0.00000594, 0.00000922, 0.00003305, 0.00000457, -0.00001455, -0.00002633, 0.00000600, -0.00001156, -0.00000444, -0.00000115, -0.00000479, -0.00000249, -0.00000209, -0.00000073, -0.00000247, 0.00000101, -0.00000536, -0.00000546, -0.00000394, -0.00000591, -0.00000141, -0.00000618, 0.00000392, -0.00000323, -0.00000547, -0.00000349, -0.00000097, -0.00000963, -0.00000312, -0.00000955, -0.00003262, -0.00001142, -0.00008541, -0.00008479, 0.00019125, 0.00004762, 0.00007897, 0.00004086, 0.00018802, -0.00004938, -0.00010527, -0.00000747, -0.00003903, 0.00000012, -0.00000721, -0.00000074, -0.00000997, -0.00000226, -0.00000981, 0.00000464, -0.00001084, 0.00000232, -0.00000225, 0.00000208, -0.00000161, 0.00000387, 0.00000117, -0.00000008, 0.00000056, -0.00000843, 0.00000681, 0.00000516, -0.00000598, 0.00001128, -0.00001246, 0.00001108, -0.00001648, 0.00000621, -0.00000558, 0.00000647, -0.00001372, 0.00000093, 0.00000774, -0.00000232, -0.00000732, 0.00000231, -0.00000273, -0.00000741, -0.00000798, 0.00000095, 0.00000466, 0.00000075, -0.00001790, 0.00001145, -0.00001233, 0.00000466, 0.00000259, },
		{ 0.00000707, -0.00000673, -0.00000494, 0.00000202, 0.00000540, -0.00000653, -0.00000268, 0.00001014, -0.00000955, 0.00000507, -0.00000287, -0.00000157, -0.00000559, 0.00000980, -0.00000274, -0.00000919, 0.00000194, 0.00001049, 0.00000124, -0.00000622, 0.00000258, -0.00000028, 0.00000326, -0.00000983, 0.00000719, -0.00000367, 0.00000086, 0.00000124, -0.00000473, 0.00000237, -0.00001001, -0.00001384, -0.00002618, -0.00000373, 0.00001572, 0.00003307, 0.00000860, 0.00000005, 0.00001885, 0.00000709, 0.00001443, 0.00000041, 0.00001851, 0.00002709, -0.00000140, -0.00001845, -0.00001917, 0.00000675, -0.00001577, 0.00000106, -0.00000421, -0.00000331, -0.00000314, -0.00000180, 0.00000016, -0.00000332, 0.00000262, -0.00000922, -0.00000198, -0.00000821, -0.00000027, -0.00000666, -0.00000016, -0.00000023, -0.00000140, -0.00000921, 0.00000293, -0.00000917, -0.00000258, -0.00000948, -0.00000896, -0.00004076, -0.00000068, -0.00013402, -0.00000235, 0.00018324, 0.00004150, 0.00007393, 0.00008428, 0.00016928, -0.00014761, -0.00003675, -0.00003452, -0.00001530, -0.00001068, 0.00000442, -0.00001262, 0.00000092, -0.00001347, 0.00000144, -0.00000399, -0.00000443, -0.00000173, 0.00000232, -0.00000258, 0.00000282, 0.00000129, 0.00000302, -0.00000280, 0.00000250, -0.00001242, 0.00001597, -0.00000619, 0.00000511, 0.00000174, -0.00000522, 0.00000401, -0.00001249, 0.00000516, -0.00000352, 0.00000324, -0.00001495, 0.00000794, 0.00000384, -0.00000474, -0.00000593, 0.00000442, -0.00000532, -0.00000969, -0.00000430, 0.00000176, 0.00000654, -0.00000811, -0.00001006, 0.00001051, -0.00001742, 0.00001569, -0.00000966, },
		{ 0.00001236, -0.00001552, 0.00000256, -0.00000053, 0.00000834, -0.00001404, 0.00000800, 0.00000128, -0.00000489, 0.00000304, -0.00000192, -0.00000405, -0.00000130, 0.00000888, -0.00000615, -0.00000779, 0.00000528, 0.00000957, -0.00000145, -0.00000546, 0.00000368, -0.00000062, 0.00000130, -0.00000923, 0.00001101, -0.00001000, 0.00000866, -0.00000722, 0.00000306, -0.00000582, -0.00000548, -0.00002417, -0.00001696, 0.00000137, 0.00002510, 0.00002571, 0.00000249, 0.00000402, 0.00001970, 0.00000304, 0.00001625, -0.00000471, 0.00002992, 0.00001466, -0.00000188, -0.00002409, -0.00000782, 0.00000036, -0.00001354, 0.00000269, -0.00000572, -0.00000239, -0.00000400, -0.00000035, -0.00000034, -0.00000270, 0.00000221, -0.00001126, -0.00000027, -0.00001015, 0.00000355, -0.00001095, 0.00000588, -0.00000562, 0.00000094, -0.00001146, 0.00000693, -0.00001573, 0.00000408, -0.00001661, -0.00001058, -0.00004099, -0.00000625, -0.00016546, 0.00009132, 0.00014589, 0.00005589, 0.00005963, 0.00015527, 0.00006787, -0.00016753, 0.00000606, -0.00005485, 0.00000530, -0.00001996, 0.00001251, -0.00002080, 0.00000824, -0.00002183, 0.00001177, -0.00001311, 0.00000282, -0.00000615, 0.00000660, -0.00000708, 0.00000774, -0.00000170, 0.00000402, -0.00000355, 0.00000112, -0.00001133, 0.00001990, -0.00001393, 0.00001487, -0.00000832, 0.00000306, -0.00000498, -0.00000514, 0.00000086, 0.00000135, -0.00000458, -0.00000919, 0.00000973, 0.00000075, -0.00000741, -0.00000201, 0.00000433, -0.00000886, -0.00000990, -0.00000013, 0.00000158, 0.00000699, -0.00001689, 0.00000379, -0.00000198, -0.00000845, 0.00001366, -0.00001212, },
		{ 0.00000620, -0.00001545, 0.00000498, 0.00000136, 0.00000480, -0.00001442, 0.00001378, -0.00000607, 0.00000024, 0.00000039, -0.00000124, -0.00000541, 0.00000284, 0.00000650, -0.00000902, -0.00000466, 0.00000684, 0.00000933, -0.00000572, -0.00000169, 0.00000173, 0.00000152, -0.00000417, -0.00000211, 0.00000639, -0.00000783, 0.00000867, -0.00000914, 0.00000523, -0.00000963, -0.00000546, -0.00003088, -0.00000488, 0.00000495, 0.00003217, 0.00001446, 0.00000135, 0.00000834, 0.00001722, 0.00000198, 0.00001490, -0.00000483, 0.00003730, 0.00000128, 0.00000001, -0.00002873, 0.00000371, -0.00000987, -0.00000584, -0.00000046, -0.00000420, -0.00000355, -0.00000359, 0.00000127, -0.00000209, -0.00000020, -0.00000123, -0.00000995, -0.00000206, -0.00000726, 0.00000182, -0.00000958, 0.00000665, -0.00000722, -0.00000020, -0.00000854, 0.00000377, -0.00001435, 0.00000308, -0.00001800, -0.00001914, -0.00002865, -0.00003801, -0.00015720, 0.00016525, 0.00010632, 0.00006814, 0.00006892, 0.00019961, -0.00007063, -0.00011102, -0.00000115, -0.00004510, 0.00000448, -0.00001380, 0.00000612, -0.00001549, 0.00000254, -0.00001762, 0.00001166, -0.00001432, 0.00000390, -0.00000500, 0.00000512, -0.00000617, 0.00000860, -0.00000193, 0.00000215, -0.00000090, -0.00000491, -0.00000301, 0.00001491, -0.00001292, 0.00001743, -0.00001321, 0.00000687, -0.00001001, -0.00000029, -0.00000121, 0.00000314, -0.00001064, -0.00000128, 0.00000811, -0.00000220, -0.00000848, 0.00000261, 0.00000213, -0.00001305, -0.00000622, 0.00000106, 0.00000364, 0.00000173, -0.00001759, 0.00001263, -0.00001535, 0.00000741, 0.00000090, -0.00000283, },
		{ -0.00000625, -0.00000829, 0.00000295, 0.00000631, -0.00000388, -0.00000724, 0.00001202, -0.00000833, 0.00000279, -0.00000097, -0.00000189, -0.00000455, 0.00000533, 0.00000364, -0.00001099, -0.00000075, 0.00000746, 0.00000861, -0.00000980, 0.00000336, -0.00000183, 0.00000421, -0.00000985, 0.00000753, -0.00000370, 0.00000163, 0.00000031, -0.00000248, -0.00000145, -0.00000486, -0.00001557, -0.00002543, 0.00000104, 0.00001303, 0.00003050, 0.00000679, 0.00000103, 0.00001429, 0.00000980, 0.00000663, 0.00000737, 0.00000412, 0.00003416, -0.00000555, -0.00000158, -0.00002739, 0.00000907, -0.00001701, 0.00000180, -0.00000501, -0.00000147, -0.00000599, -0.00000156, 0.00000195, -0.00000389, 0.00000296, -0.00000660, -0.00000619, -0.00000610, -0.00000091, -0.00000477, -0.00000228, 0.00000082, -0.00000327, -0.00000572, -0.00000070, -0.00000618, -0.00000451, -0.00000741, -0.00001197, -0.00003363, -0.00000828, -0.00009176, -0.00010278, 0.00020138, 0.00008221, 0.00006722, 0.00011971, 0.00016452, -0.00016991, -0.00003113, -0.00003704, -0.00001404, -0.00001313, 0.00000421, -0.00001095, -0.00000024, -0.00001331, -0.00000213, 0.00000080, -0.00000672, -0.00000193, 0.00000201, -0.00000231, 0.00000087, 0.00000438, 0.00000093, -0.00000166, 0.00000299, -0.00001242, 0.00000918, 0.00000333, -0.00000388, 0.00001214, -0.00001193, 0.00000536, -0.00000980, 0.00000039, 0.00000062, -0.00000059, -0.00001057, 0.00000385, 0.00000585, -0.00000597, -0.00000635, 0.00000579, -0.00000198, -0.00001595, -0.00000005, -0.00000081, 0.00000765, -0.00000883, -0.00000736, 0.00000828, -0.00001730, 0.00001666, -0.00001071, 0.00000937, },
		{ -0.00001516, -0.00000107, 0.00000163, 0.00000877, -0.00001144, 0.00000217, 0.00000584, -0.00000657, 0.00000306, -0.00000158, -0.00000273, -0.00000269, 0.00000669, -0.00000006, -0.00001083, 0.00000204, 0.00000908, 0.00000494, -0.00001025, 0.00000582, -0.00000336, 0.00000361, -0.00001018, 0.00001196, -0.00001047, 0.00000967, -0.00000834, 0.00000537, -0.00001020, 0.00000210, -0.00003005, -0.00001007, -0.00000016, 0.00002551, 0.00001945, 0.00000604, -0.00000043, 0.00002070, -0.00000033, 0.00001498, -0.00000409, 0.00002017, 0.00002002, -0.00000287, -0.00000931, -0.00001820, 0.00000510, -0.00001613, 0.00000412, -0.00000664, -0.00000072, -0.00000743, 0.00000118, 0.00000109, -0.00000414, 0.00000443, -0.00001111, -0.00000294, -0.00000861, 0.00000397, -0.00001079, 0.00000559, -0.00000698, 0.00000241, -0.00001142, 0.00000637, -0.00001585, 0.00000628, -0.00002061, -0.00000487, -0.00004527, 0.00000532, -0.00014713, -0.00001735, 0.00020398, 0.00006982, 0.00006958, 0.00018647, 0.00004642, -0.00017940, 0.00001263, -0.00005994, 0.00000956, -0.00002616, 0.00001669, -0.00002308, 0.00001042, -0.00002558, 0.00001256, -0.00001122, 0.00000201, -0.00000793, 0.00000843, -0.00000958, 0.00000895, -0.00000119, 0.00000375, -0.00000396, 0.00000371, -0.00001548, 0.00001819, -0.00000783, 0.00000727, 0.00000333, -0.00000802, 0.00000221, -0.00000797, 0.00000031, 0.00000255, -0.00000624, -0.00000588, 0.00000543, 0.00000339, -0.00000935, -0.00000172, 0.00000668, -0.00000804, -0.00001460, 0.00000434, -0.00000200, 0.00000890, -0.00001765, 0.00000691, -0.00000618, -0.00000527, 0.00001172, -0.00001082, 0.00001342, },
		{ -0.00001530, 0.00000155, 0.00000425, 0.00000512, -0.00001262, 0.00000775, 0.00000031, -0.00000403, 0.00000286, -0.00000276, -0.00000218, -0.00000159, 0.00000796, -0.00000490, -0.00000808, 0.00000303, 0.00001205, -0.00000216, -0.00000564, 0.00000359, -0.00000108, -0.00000152, -0.00000320, 0.00000700, -0.00000754, 0.00000839, -0.00000850, 0.00000507, -0.00001104, 0.00000011, -0.00003593, 0.00000355, -0.00000035, 0.00003428, 0.00000656, 0.00000836, -0.00000030, 0.00002281, -0.00000680, 0.00002025, -0.00001176, 0.00003458, 0.00000229, 0.00000477, -0.00001951, -0.00000547, -0.00000477, -0.00000850, 0.00000060, -0.00000419, -0.00000332, -0.00000621, 0.00000328, -0.00000106, -0.00000206, 0.00000251, -0.00001241, -0.00000253, -0.00000672, 0.00000314, -0.00001070, 0.00000752, -0.00001012, 0.00000395, -0.00001174, 0.00000617, -0.00001702, 0.00000816, -0.00002654, -0.00000575, -0.00004419, -0.00000354, -0.00017899, 0.00007361, 0.00018903, 0.00005628, 0.00010412, 0.00020970, -0.00009639, -0.00011221, 0.00000118, -0.00004787, 0.00000724, -0.00001966, 0.00001132, -0.00001904, 0.00000558, -0.00002286, 0.00001510, -0.00001444, 0.00000368, -0.00000707, 0.00000759, -0.00000997, 0.00001196, -0.00000355, 0.00000350, -0.00000249, -0.00000099, -0.00001067, 0.00001892, -0.00001246, 0.00001461, -0.00000436, -0.00000490, 0.00000075, -0.00000802, 0.00000309, -0.00000002, -0.00000805, -0.00000158, 0.00000591, -0.00000074, -0.00000969, 0.00000286, 0.00000551, -0.00001520, -0.00000778, 0.00000361, 0.00000093, 0.00000287, -0.00001644, 0.00001317, -0.00001800, 0.00001063, -0.00000203, -0.00000009, 0.00000621, },
		{ -0.00000980, 0.00000136, 0.00000832, -0.00000228, -0.00000798, 0.00000805, -0.00000235, -0.00000227, 0.00000265, -0.00000419, -0.00000061, -0.00000118, 0.00000874, -0.00000964, -0.00000419, 0.00000385, 0.00001380, -0.00000946, 0.00000119, -0.00000112, 0.00000251, -0.00000721, 0.00000615, -0.00000302, 0.00000224, -0.00000092, -0.00000021, -0.00000433, -0.00000210, -0.00001442, -0.00002580, 0.00000587, 0.00000839, 0.00003095, 0.00000113, 0.00000741, 0.00000508, 0.00001656, -0.00000423, 0.00001644, -0.00000843, 0.00003788, -0.00000891, 0.00000921, -0.00002532, 0.00000338, -0.00001298, -0.00000033, -0.00000452, -0.00000045, -0.00000755, -0.00000254, 0.00000375, -0.00000327, 0.00000128, -0.00000220, -0.00001073, -0.00000422, -0.00000170, -0.00000259, -0.00000429, 0.00000207, -0.00000637, -0.00000044, -0.00000590, -0.00000184, -0.00000830, -0.00000192, -0.00002125, -0.00001726, -0.00002934, -0.00003892, -0.00017322, 0.00015230, 0.00016466, 0.00004440, 0.00017497, 0.00014361, -0.00018357, -0.00003090, -0.00003596, -0.00001540, -0.00001224, -0.00000017, -0.00000602, -0.00000401, -0.00001052, -0.00000740, 0.00000564, -0.00000831, -0.00000190, 0.00000035, -0.00000032, -0.00000280, 0.00000855, -0.00000190, 0.00000066, 0.00000090, -0.00000834, -0.00000042, 0.00001223, -0.00000979, 0.00001649, -0.00000994, -0.00000213, -0.00000011, -0.00000866, 0.00000733, -0.00000657, -0.00000432, -0.00000054, 0.00000675, -0.00000662, -0.00000592, 0.00000547, 0.00000205, -0.00002044, 0.00000160, -0.00000093, 0.00000652, -0.00000860, -0.00000419, 0.00000511, -0.00001583, 0.00001661, -0.00001147, 0.00001097, -0.00000547, },
		{ -0.00000527, 0.00000309, 0.00000863, -0.00000745, -0.00000263, 0.00000623, -0.00000366, -0.00000030, 0.00000107, -0.00000385, -0.00000024, 0.00000038, 0.00000724, -0.00001211, -0.00000143, 0.00000678, 0.00001102, -0.00001230, 0.00000531, -0.00000390, 0.00000307, -0.00000771, 0.00000986, -0.00000870, 0.00000911, -0.00000867, 0.00000714, -0.00001344, 0.00000690, -0.00003181, -0.00000609, -0.00000068, 0.00002400, 0.00001619, 0.00000545, 0.00000166, 0.00001464, 0.00000380, 0.00000616, 0.00000449, 0.00000573, 0.00002760, -0.00000870, 0.00000585, -0.00002314, 0.00000410, -0.00001420, 0.00000282, -0.00000648, 0.00000071, -0.00001027, 0.00000189, 0.00000233, -0.00000399, 0.00000361, -0.00000725, -0.00000837, -0.00000494, 0.00000191, -0.00000782, 0.00000281, -0.00000570, 0.00000015, -0.00000662, 0.00000077, -0.00001088, 0.00000274, -0.00001714, -0.00001050, -0.00003261, -0.00001091, -0.00008995, -0.00013198, 0.00021469, 0.00012551, 0.00006089, 0.00023940, 0.00000285, -0.00017462, 0.00000646, -0.00005521, 0.00000629, -0.00002508, 0.00001290, -0.00001814, 0.00000614, -0.00002280, 0.00000756, -0.00000543, -0.00000108, -0.00000732, 0.00000703, -0.00000834, 0.00000620, 0.00000269, 0.00000075, -0.00000155, 0.00000209, -0.00001284, 0.00000914, 0.00000342, -0.00000337, 0.00001471, -0.00001475, 0.00000280, -0.00000411, -0.00000534, 0.00000738, -0.00001115, 0.00000073, -0.00000088, 0.00000617, -0.00001122, -0.00000023, 0.00000629, -0.00000448, -0.00001976, 0.00000778, -0.00000400, 0.00000867, -0.00001690, 0.00000976, -0.00001090, -0.00000064, 0.00000799, -0.00000817, 0.00001240, -0.00001235, },
		{ -0.00000424, 0.00000795, 0.00000351, -0.00000711, -0.00000062, 0.00000559, -0.00000551, 0.00000280, -0.00000254, -0.00000072, -0.00000247, 0.00000428, 0.00000256, -0.00001132, -0.00000035, 0.00001169, 0.00000340, -0.00000906, 0.00000437, -0.00000284, -0.00000037, -0.00000177, 0.00000484, -0.00000485, 0.00000622, -0.00000689, 0.00000465, -0.00001218, 0.00000397, -0.00003717, 0.00000832, -0.00000416, 0.00003510, 0.00000040, 0.00001285, -0.00000383, 0.00002192, -0.00000777, 0.00001668, -0.00000764, 0.00002240, 0.00001028, -0.00000055, -0.00000306, -0.00001561, -0.00000101, -0.00000931, 0.00000046, -0.00000418, -0.00000222, -0.00000931, 0.00000504, -0.00000019, -0.00000267, 0.00000334, -0.00001060, -0.00000691, -0.00000290, 0.00000092, -0.00000781, 0.00000464, -0.00000937, 0.00000356, -0.00000922, 0.00000228, -0.00001330, 0.00000650, -0.00002716, -0.00000374, -0.00004223, -0.00000201, -0.00014039, -0.00006370, 0.00025845, 0.00007075, 0.00013062, 0.00023106, -0.00013788, -0.00009902, -0.00000884, -0.00004057, 0.00000237, -0.00001865, 0.00000889, -0.00001526, 0.00000181, -0.00002086, 0.00001137, -0.00000894, -0.00000011, -0.00000619, 0.00000665, -0.00000982, 0.00001088, -0.00000112, 0.00000163, -0.00000132, -0.00000075, -0.00001151, 0.00001375, -0.00000314, 0.00000359, 0.00001029, -0.00001817, 0.00000941, -0.00001174, 0.00000354, -0.00000014, -0.00000841, 0.00000167, 0.00000082, 0.00000164, -0.00001090, 0.00000388, 0.00000613, -0.00001339, -0.00001181, 0.00000705, -0.00000170, 0.00000276, -0.00001379, 0.00001290, -0.00002033, 0.00001455, -0.00000647, 0.00000396, 0.00000389, -0.00001158, },
		{ -0.00000322, 0.00001171, -0.00000284, -0.00000391, -0.00000092, 0.00000576, -0.00000713, 0.00000549, -0.00000615, 0.00000318, -0.00000573, 0.00000910, -0.00000384, -0.00000864, 0.00000096, 0.00001529, -0.00000517, -0.00000283, 0.00000064, -0.00000047, -0.00000377, 0.00000506, -0.00000334, 0.00000394, -0.00000304, 0.00000206, -0.00000613, -0.00000119, -0.00001216, -0.00002490, 0.00000836, 0.00000435, 0.00003176, -0.00000528, 0.00001469, -0.00000298, 0.00002070, -0.00001084, 0.00001965, -0.00001158, 0.00003171, -0.00000417, 0.00000794, -0.00001174, -0.00000846, -0.00000609, -0.00000351, -0.00000351, -0.00000041, -0.00000742, -0.00000476, 0.00000553, -0.00000221, -0.00000050, 0.00000098, -0.00001238, -0.00000535, -0.00000012, -0.00000283, -0.00000349, 0.00000081, -0.00000742, 0.00000240, -0.00000771, -0.00000137, -0.00000846, 0.00000034, -0.00002760, -0.00000512, -0.00004284, -0.00000848, -0.00017989, 0.00002877, 0.00026902, 0.00002650, 0.00023662, 0.00011744, -0.00019837, -0.00002712, -0.00003923, -0.00001280, -0.00001368, -0.00000291, -0.00000377, -0.00000479, -0.00001060, -0.00000883, 0.00000574, -0.00000578, -0.00000411, -0.00000014, 0.00000041, -0.00000473, 0.00001019, -0.00000221, 0.00000128, -0.00000048, -0.00000496, -0.00000645, 0.00001404, -0.00000704, 0.00001081, 0.00000170, -0.00001608, 0.00001225, -0.00001739, 0.00001260, -0.00001116, 0.00000027, -0.00000209, 0.00000463, -0.00000589, -0.00000538, 0.00000502, 0.00000434, -0.00002098, -0.00000038, 0.00000144, 0.00000378, -0.00000819, -0.00000033, 0.00000125, -0.00001358, 0.00001640, -0.00001291, 0.00001303, -0.00000608, -0.00000801, },
		{ 0.00000200, 0.00000946, -0.00000462, -0.00000323, 0.00000058, 0.00000372, -0.00000588, 0.00000488, -0.00000633, 0.00000414, -0.00000652, 0.00001168, -0.00000915, -0.00000623, 0.00000465, 0.00001401, -0.00000979, 0.00000151, -0.00000188, -0.00000036, -0.00000239, 0.00000576, -0.00000581, 0.00000785, -0.00000849, 0.00000772, -0.00001423, 0.00000772, -0.00002914, -0.00000466, -0.00000059, 0.00002104, 0.00001553, 0.00000117, 0.00000870, 0.00000461, 0.00001106, -0.00000462, 0.00001382, -0.00000549, 0.00002975, -0.00001003, 0.00001169, -0.00001668, -0.00000487, -0.00000735, -0.00000104, -0.00000510, 0.00000112, -0.00001152, 0.00000134, 0.00000329, -0.00000236, 0.00000068, -0.00000182, -0.00001377, -0.00000172, -0.00000037, -0.00000436, -0.00000025, -0.00000371, -0.00000374, 0.00000012, -0.00000635, -0.00000450, -0.00000300, -0.00000997, -0.00002216, -0.00001174, -0.00003812, -0.00002886, -0.00020302, 0.00014424, 0.00022670, 0.00004038, 0.00030939, -0.00005892, -0.00015713, -0.00000841, -0.00004463, -0.00000138, -0.00001914, 0.00000379, -0.00000907, -0.00000149, -0.00001684, 0.00000064, 0.00000047, -0.00000372, -0.00000594, 0.00000368, -0.00000491, 0.00000156, 0.00000748, -0.00000270, 0.00000184, -0.00000190, -0.00000663, -0.00000175, 0.00001286, -0.00000950, 0.00001835, -0.00001167, -0.00000540, 0.00000581, -0.00001397, 0.00001338, -0.00001671, 0.00000751, -0.00000580, 0.00000646, -0.00001154, 0.00000144, 0.00000474, -0.00000091, -0.00002224, 0.00000774, -0.00000287, 0.00000570, -0.00001459, 0.00001216, -0.00001554, 0.00000455, 0.00000371, -0.00000555, 0.00001090, -0.00000949, -0.00000740, },
		{ 0.00001015, 0.00000190, -0.00000091, -0.00000689, 0.00000568, -0.00000180, -0.00000092, 0.00000021, -0.00000191, 0.00000037, -0.00000277, 0.00000998, -0.00001177, -0.00000448, 0.00001012, 0.00000781, -0.00000902, 0.00000206, -0.00000204, -0.00000231, 0.00000313, -0.00000038, 0.00000012, 0.00000238, -0.00000419, 0.00000295, -0.00001098, 0.00000351, -0.00003242, 0.00000778, -0.00000471, 0.00003270, -0.00000166, 0.00001194, 0.00000021, 0.00001321, -0.00000075, 0.00000527, 0.00000442, 0.00000548, 0.00002006, -0.00000822, 0.00001061, -0.00001860, -0.00000347, -0.00000579, -0.00000191, -0.00000353, -0.00000113, -0.00001182, 0.00000614, -0.00000003, -0.00000101, 0.00000047, -0.00000467, -0.00001419, 0.00000319, -0.00000416, -0.00000155, -0.00000160, -0.00000475, -0.00000198, -0.00000058, -0.00000787, -0.00000307, -0.00000289, -0.00001774, -0.00001586, -0.00002050, -0.00003079, -0.00006378, -0.00019703, 0.00026465, 0.00013629, 0.00013938, 0.00027018, -0.00020026, -0.00006762, -0.00003386, -0.00002144, -0.00001194, -0.00000851, -0.00000314, -0.00000287, -0.00000969, -0.00001115, 0.00000098, 0.00000103, -0.00000680, -0.00000204, 0.00000142, -0.00000484, 0.00000492, 0.00000525, -0.00000335, 0.00000303, -0.00000509, -0.00000542, 0.00000161, 0.00001056, -0.00000919, 0.00002261, -0.00002497, 0.00001011, -0.00000829, -0.00000101, 0.00000299, -0.00001136, 0.00000713, -0.00000483, 0.00000279, -0.00001086, 0.00000514, 0.00000476, -0.00000959, -0.00001540, 0.00000850, -0.00000194, -0.00000002, -0.00000914, 0.00001152, -0.00002193, 0.00001854, -0.00001142, 0.00000823, 0.00000052, -0.00000645, -0.00000959, },
		{ 0.00001481, -0.00000446, 0.00000289, -0.00001081, 0.00001119, -0.00000791, 0.00000477, -0.00000539, 0.00000413, -0.00000585, 0.00000406, 0.00000478, -0.00001228, -0.00000174, 0.00001395, 0.00000059, -0.00000577, 0.00000094, -0.00000230, -0.00000205, 0.00000643, -0.00000616, 0.00000761, -0.00000651, 0.00000474, -0.00000784, -0.00000001, -0.00001190, -0.00001959, 0.00000574, 0.00000392, 0.00002956, -0.00000837, 0.00001761, -0.00000412, 0.00001672, -0.00000818, 0.00001265, -0.00000269, 0.00001489, 0.00000871, -0.00000252, 0.00000659, -0.00001933, -0.00000151, -0.00000463, -0.00000336, -0.00000089, -0.00000599, -0.00000765, 0.00000723, -0.00000178, -0.00000028, 0.00000039, -0.00000901, -0.00001080, 0.00000499, -0.00000692, 0.00000172, -0.00000485, -0.00000373, -0.00000078, -0.00000217, -0.00000873, -0.00000055, -0.00000652, -0.00002265, -0.00000874, -0.00003334, -0.00001800, -0.00012007, -0.00013745, 0.00034605, 0.00004863, 0.00028489, 0.00010083, -0.00022653, -0.00000967, -0.00005513, 0.00000040, -0.00002348, 0.00000217, -0.00001011, 0.00000261, -0.00001828, -0.00000220, -0.00000197, 0.00000270, -0.00001020, 0.00000282, -0.00000302, -0.00000175, 0.00000637, 0.00000237, -0.00000200, 0.00000158, -0.00000605, -0.00000497, 0.00000639, 0.00000448, -0.00000241, 0.00001741, -0.00002907, 0.00002017, -0.00002031, 0.00001271, -0.00001163, 0.00000109, 0.00000023, -0.00000014, -0.00000464, -0.00000441, 0.00000493, 0.00000420, -0.00001785, -0.00000442, 0.00000419, 0.00000155, -0.00000910, 0.00000478, -0.00000330, -0.00001072, 0.00001618, -0.00001492, 0.00001544, -0.00000800, -0.00000376, -0.00000896, },
		{ 0.00001114, -0.00000361, 0.00000056, -0.00000910, 0.00001187, -0.00000995, 0.00000678, -0.00000733, 0.00000696, -0.00000994, 0.00001028, -0.00000149, -0.00001185, 0.00000326, 0.00001324, -0.00000362, -0.00000371, 0.00000080, -0.00000465, 0.00000326, 0.00000222, -0.00000392, 0.00000740, -0.00000862, 0.00000752, -0.00001306, 0.00000591, -0.00002461, -0.00000283, -0.00000253, 0.00001938, 0.00001441, -0.00000297, 0.00001526, -0.00000265, 0.00001400, -0.00000945, 0.00001562, -0.00000587, 0.00002027, -0.00000097, 0.00000475, -0.00000009, -0.00001823, 0.00000102, -0.00000516, -0.00000338, 0.00000035, -0.00001077, -0.00000057, 0.00000422, -0.00000049, -0.00000183, 0.00000164, -0.00001547, -0.00000206, 0.00000003, -0.00000350, -0.00000040, -0.00000438, -0.00000515, 0.00000352, -0.00000890, -0.00000309, -0.00000403, -0.00000755, -0.00002829, 0.00000086, -0.00005176, 0.00000028, -0.00019779, -0.00000698, 0.00034225, 0.00003900, 0.00037587, -0.00012271, -0.00014078, -0.00002003, -0.00003840, -0.00000507, -0.00001569, -0.00000327, -0.00000320, -0.00000564, -0.00001418, -0.00000198, 0.00000123, -0.00000214, -0.00000674, 0.00000157, -0.00000298, -0.00000080, 0.00000897, -0.00000286, 0.00000280, -0.00000435, -0.00000220, -0.00000785, 0.00001414, -0.00000605, 0.00001137, 0.00000058, -0.00001843, 0.00001629, -0.00002021, 0.00001610, -0.00001904, 0.00001080, -0.00000635, 0.00000277, -0.00000980, 0.00000261, 0.00000353, 0.00000051, -0.00002041, 0.00000385, 0.00000080, 0.00000115, -0.00001181, 0.00001442, -0.00001972, 0.00000946, -0.00000006, -0.00000394, 0.00001002, -0.00000800, -0.00000646, -0.00000160, },
		{ 0.00000176, 0.00000344, -0.00000824, -0.00000043, 0.00000596, -0.00000625, 0.00000357, -0.00000385, 0.00000412, -0.00000892, 0.00001305, -0.00000721, -0.00001012, 0.00000893, 0.00000886, -0.00000447, -0.00000325, 0.00000098, -0.00000665, 0.00001011, -0.00000680, 0.00000557, -0.00000175, -0.00000110, -0.00000008, -0.00000696, -0.00000135, -0.00002268, 0.00000354, -0.00000335, 0.00002827, -0.00000082, 0.00000717, 0.00000880, 0.00000139, 0.00000821, -0.00000667, 0.00001560, -0.00000671, 0.00002258, -0.00000842, 0.00001278, -0.00001012, -0.00001293, 0.00000146, -0.00000581, -0.00000227, -0.00000086, -0.00001272, 0.00000600, -0.00000036, 0.00000226, -0.00000434, 0.00000235, -0.00002074, 0.00000820, -0.00000902, 0.00000495, -0.00000808, 0.00000074, -0.00000925, 0.00000973, -0.00001896, 0.00000803, -0.00001423, -0.00000489, -0.00003310, 0.00000725, -0.00006630, 0.00000757, -0.00027026, 0.00017349, 0.00025093, 0.00014458, 0.00031600, -0.00027409, -0.00002677, -0.00006582, 0.00000230, -0.00002982, 0.00000589, -0.00001982, 0.00001287, -0.00002377, 0.00000107, -0.00001068, 0.00001051, -0.00001272, 0.00000283, -0.00000576, 0.00000229, -0.00000269, 0.00001202, -0.00000818, 0.00000775, -0.00001102, 0.00000308, -0.00001081, 0.00002075, -0.00001557, 0.00002522, -0.00002065, 0.00000187, 0.00000051, -0.00000750, 0.00000618, -0.00001364, 0.00001115, -0.00000743, 0.00000018, -0.00000824, 0.00000554, 0.00000322, -0.00000666, -0.00001554, 0.00000588, 0.00000127, -0.00000555, -0.00000288, 0.00000913, -0.00002226, 0.00002149, -0.00001533, 0.00001108, -0.00000205, -0.00000242, -0.00001192, 0.00000914, },
		{ -0.00000498, 0.00000900, -0.00001696, 0.00001016, -0.00000283, 0.00000011, -0.00000184, 0.00000221, -0.00000241, -0.00000344, 0.00001200, -0.00001209, -0.00000566, 0.00001182, 0.00000465, -0.00000493, -0.00000191, -0.00000154, -0.00000329, 0.00001075, -0.00001122, 0.00001266, -0.00001076, 0.00000778, -0.00001030, 0.00000315, -0.00001555, -0.00000874, -0.00000240, 0.00000773, 0.00002331, -0.00000659, 0.00001422, 0.00000280, 0.00000490, 0.00000207, -0.00000142, 0.00001292, -0.00000542, 0.00002187, -0.00001227, 0.00001891, -0.00002142, -0.00000350, -0.00000179, -0.00000478, -0.00000135, -0.00000420, -0.00001009, 0.00000838, -0.00000241, 0.00000262, -0.00000414, -0.00000157, -0.00001880, 0.00001189, -0.00001344, 0.00000999, -0.00001378, 0.00000406, -0.00000955, 0.00000978, -0.00002237, 0.00001390, -0.00002158, -0.00000664, -0.00002767, -0.00000329, -0.00005757, -0.00002714, -0.00028407, 0.00033703, 0.00013784, 0.00031212, 0.00009806, -0.00027069, 0.00002259, -0.00008380, 0.00002373, -0.00004156, 0.00001564, -0.00002535, 0.00001781, -0.00003284, 0.00001131, -0.00001556, 0.00001489, -0.00001862, 0.00000881, -0.00001067, 0.00000609, -0.00000232, 0.00001071, -0.00000813, 0.00000660, -0.00001141, 0.00000274, -0.00000701, 0.00001879, -0.00001534, 0.00002815, -0.00003301, 0.00001832, -0.00001511, 0.00000738, -0.00000860, -0.00000068, 0.00000426, -0.00000463, -0.00000533, -0.00000170, 0.00000444, 0.00000284, -0.00001353, -0.00000728, 0.00000435, 0.00000204, -0.00001249, 0.00001134, -0.00000851, -0.00000701, 0.00001523, -0.00001628, 0.00001677, -0.00000958, 0.00000040, -0.00001279, 0.00001515, },
		{ -0.00000292, 0.00000599, -0.00001802, 0.00001565, -0.00000868, 0.00000426, -0.00000482, 0.00000607, -0.00000827, 0.00000357, 0.00000807, -0.00001562, 0.00000141, 0.00001029, 0.00000316, -0.00000704, 0.00000142, -0.00000712, 0.00000649, 0.00000152, -0.00000450, 0.00000864, -0.00000966, 0.00000728, -0.00001147, 0.00000411, -0.00002165, 0.00000249, -0.00000886, 0.00002149, 0.00000984, -0.00000309, 0.00001628, -0.00000182, 0.00000791, -0.00000439, 0.00000690, 0.00000552, 0.00000056, 0.00001568, -0.00000909, 0.00001790, -0.00002805, 0.00000554, -0.00000657, -0.00000253, -0.00000120, -0.00000842, -0.00000333, 0.00000528, 0.00000028, -0.00000165, 0.00000060, -0.00001156, -0.00000706, 0.00000409, -0.00000657, 0.00000393, -0.00000963, -0.00000128, -0.00000004, -0.00000304, -0.00001044, 0.00000393, -0.00001570, -0.00002070, -0.00000583, -0.00003788, -0.00001618, -0.00012242, -0.00019339, 0.00040921, 0.00010084, 0.00041429, -0.00017076, -0.00013964, -0.00001732, -0.00004525, 0.00000211, -0.00002095, -0.00000189, -0.00000655, -0.00000118, -0.00001931, 0.00000350, -0.00000576, 0.00000520, -0.00001129, 0.00000302, -0.00000551, 0.00000219, 0.00000449, 0.00000228, -0.00000056, -0.00000279, -0.00000288, -0.00000600, 0.00000600, 0.00000570, -0.00000211, 0.00001492, -0.00002800, 0.00002003, -0.00001892, 0.00001237, -0.00001610, 0.00000930, -0.00000245, -0.00000362, -0.00000750, 0.00000396, 0.00000281, -0.00000033, -0.00001530, -0.00000161, 0.00000433, -0.00000249, -0.00001026, 0.00001715, -0.00002339, 0.00001384, -0.00000318, -0.00000307, 0.00000933, -0.00000741, -0.00000381, -0.00000545, 0.00001169, },
		{ 0.00000572, -0.00000528, -0.00000972, 0.00001320, -0.00000892, 0.00000414, -0.00000347, 0.00000524, -0.00001042, 0.00000962, 0.00000180, -0.00001587, 0.00000794, 0.00000676, 0.00000281, -0.00000895, 0.00000371, -0.00001074, 0.00001607, -0.00001133, 0.00000897, -0.00000427, 0.00000110, -0.00000341, -0.00000159, -0.00000807, -0.00001179, -0.00000036, -0.00000346, 0.00002509, -0.00000068, 0.00000309, 0.00001566, -0.00000602, 0.00001114, -0.00001094, 0.00001756, -0.00000685, 0.00001241, 0.00000292, 0.00000261, 0.00000637, -0.00002454, 0.00000846, -0.00000885, -0.00000114, -0.00000175, -0.00001140, 0.00000426, -0.00000042, 0.00000520, -0.00000782, 0.00000626, -0.00002235, 0.00000850, -0.00001019, 0.00000788, -0.00001076, 0.00000268, -0.00001285, 0.00001505, -0.00002344, 0.00001206, -0.00001875, 0.00000064, -0.00004112, 0.00002130, -0.00008046, 0.00003532, -0.00025269, -0.00000564, 0.00037074, 0.00018831, 0.00034001, -0.00033713, 0.00000481, -0.00008894, 0.00001742, -0.00004044, 0.00001501, -0.00003171, 0.00002279, -0.00003202, 0.00000799, -0.00001640, 0.00001351, -0.00001339, 0.00000462, -0.00001074, 0.00000684, -0.00000687, 0.00001427, -0.00000857, 0.00000888, -0.00001404, 0.00000813, -0.00001670, 0.00002131, -0.00001119, 0.00001648, -0.00000745, -0.00001000, 0.00000836, -0.00001027, 0.00000503, -0.00001165, 0.00001052, -0.00000476, -0.00000678, -0.00000331, 0.00000491, 0.00000261, -0.00000621, -0.00001137, -0.00000031, 0.00000650, -0.00001196, 0.00000335, 0.00000645, -0.00002123, 0.00002275, -0.00001730, 0.00001173, -0.00000302, -0.00000053, -0.00001076, 0.00000552, 0.00000255, },
		{ 0.00001224, -0.00001720, 0.00000268, 0.00000583, -0.00000552, 0.00000179, 0.00000008, 0.00000104, -0.00000871, 0.00001361, -0.00000636, -0.00001104, 0.00001008, 0.00000542, -0.00000001, -0.00000712, 0.00000082, -0.00000618, 0.00001598, -0.00001608, 0.00001680, -0.00001403, 0.00001013, -0.00001309, 0.00000796, -0.00002207, 0.00000518, -0.00001286, 0.00001299, 0.00001543, -0.00000230, 0.00000744, 0.00001318, -0.00000835, 0.00001223, -0.00001369, 0.00002496, -0.00001848, 0.00002522, -0.00001154, 0.00001781, -0.00001202, -0.00001170, 0.00000381, -0.00000691, -0.00000161, -0.00000316, -0.00001072, 0.00000807, -0.00000328, 0.00000665, -0.00000979, 0.00000583, -0.00002469, 0.00001626, -0.00001829, 0.00001701, -0.00002152, 0.00001140, -0.00001861, 0.00002156, -0.00003461, 0.00002669, -0.00003605, 0.00001008, -0.00004818, 0.00002758, -0.00009583, 0.00004656, -0.00034808, 0.00020785, 0.00028571, 0.00034058, 0.00009126, -0.00031578, 0.00005591, -0.00011285, 0.00004634, -0.00005898, 0.00002985, -0.00004191, 0.00003311, -0.00004685, 0.00002446, -0.00002789, 0.00002441, -0.00002452, 0.00001408, -0.00001885, 0.00001465, -0.00001121, 0.00001796, -0.00001275, 0.00001070, -0.00001684, 0.00001093, -0.00001842, 0.00002729, -0.00001957, 0.00002699, -0.00002446, 0.00000719, -0.00000484, 0.00000054, -0.00000583, -0.00000139, 0.00000618, -0.00000537, -0.00001000, 0.00000358, 0.00000279, 0.00000192, -0.00001065, -0.00000638, 0.00000046, 0.00000561, -0.00001779, 0.00001839, -0.00001384, -0.00000234, 0.00001277, -0.00001566, 0.00001560, -0.00000920, 0.00000227, -0.00001261, 0.00001158, -0.00000333, },
		{ 0.00000964, -0.00002176, 0.00001196, -0.00000129, -0.00000210, 0.00000036, 0.00000244, -0.00000353, -0.00000445, 0.00001468, -0.00001439, -0.00000237, 0.00000730, 0.00000746, -0.00000613, -0.00000137, -0.00000668, 0.00000645, 0.00000373, -0.00000695, 0.00001071, -0.00001095, 0.00000699, -0.00001035, 0.00000394, -0.00002178, 0.00001167, -0.00001890, 0.00002682, 0.00000186, 0.00000131, 0.00001106, 0.00000670, -0.00000514, 0.00000698, -0.00000718, 0.00002125, -0.00002004, 0.00002975, -0.00001840, 0.00002640, -0.00002724, 0.00000277, -0.00000393, -0.00000305, -0.00000290, -0.00000564, -0.00000563, 0.00000581, -0.00000015, 0.00000132, -0.00000428, -0.00000393, -0.00001404, 0.00000928, -0.00001137, 0.00001054, -0.00001741, 0.00000617, -0.00000879, 0.00000867, -0.00002316, 0.00001737, -0.00003113, -0.00000266, -0.00002719, -0.00000472, -0.00005973, -0.00002263, -0.00033900, 0.00035960, 0.00025520, 0.00041878, -0.00020128, -0.00015398, -0.00000156, -0.00006295, 0.00001758, -0.00003306, 0.00000682, -0.00001776, 0.00000989, -0.00002993, 0.00001436, -0.00001717, 0.00001497, -0.00001723, 0.00000686, -0.00001168, 0.00000942, -0.00000419, 0.00001046, -0.00000617, 0.00000131, -0.00000733, 0.00000115, -0.00000641, 0.00001846, -0.00001318, 0.00002188, -0.00002649, 0.00001284, -0.00000909, 0.00000314, -0.00000961, 0.00000522, 0.00000281, -0.00000913, -0.00000737, 0.00000690, 0.00000167, -0.00000192, -0.00000972, -0.00000519, 0.00000478, -0.00000318, -0.00001088, 0.00002038, -0.00002642, 0.00001781, -0.00000634, -0.00000166, 0.00000738, -0.00000599, -0.00000198, -0.00000660, 0.00000888, -0.00000054, },
		{ -0.00000122, -0.00001696, 0.00001479, -0.00000550, -0.00000002, 0.00000065, 0.00000235, -0.00000675, 0.00000148, 0.00001140, -0.00001846, 0.00000546, 0.00000363, 0.00000936, -0.00001158, 0.00000338, -0.00001190, 0.00001828, -0.00001183, 0.00000894, -0.00000432, 0.00000167, -0.00000589, 0.00000303, -0.00001337, -0.00000369, -0.00000051, -0.00000742, 0.00002539, -0.00000400, 0.00000336, 0.00001607, -0.00000436, 0.00000430, -0.00000483, 0.00000879, 0.00000427, -0.00000700, 0.00002045, -0.00001170, 0.00002097, -0.00003029, 0.00001026, -0.00000860, -0.00000091, -0.00000361, -0.00000818, 0.00000127, 0.00000021, 0.00000615, -0.00000753, 0.00000458, -0.00001735, 0.00000321, -0.00000674, 0.00000586, -0.00000793, -0.00000108, -0.00000977, 0.00001159, -0.00001749, 0.00000531, -0.00001216, -0.00000691, -0.00003200, 0.00001154, -0.00006142, 0.00001067, -0.00016056, -0.00021205, 0.00041805, 0.00032402, 0.00031314, -0.00036913, 0.00001130, -0.00009104, 0.00001458, -0.00003647, 0.00001220, -0.00003210, 0.00002068, -0.00002924, 0.00000518, -0.00001258, 0.00000768, -0.00000739, 0.00000169, -0.00001096, 0.00000573, -0.00000460, 0.00000962, -0.00000289, 0.00000490, -0.00001198, 0.00000693, -0.00001380, 0.00001218, 0.00000128, 0.00000202, 0.00000636, -0.00001680, 0.00000834, -0.00000505, -0.00000256, -0.00000450, 0.00000522, 0.00000193, -0.00001557, 0.00000150, 0.00000458, 0.00000245, -0.00000761, -0.00000477, -0.00000700, 0.00001049, -0.00001634, 0.00000752, 0.00000442, -0.00001920, 0.00002244, -0.00001764, 0.00001097, -0.00000331, 0.00000023, -0.00000802, 0.00000195, 0.00000206, 0.00000772, },
		{ -0.00001308, -0.00000715, 0.00001273, -0.00000748, 0.00000195, 0.00000084, 0.00000113, -0.00000857, 0.00000846, 0.00000328, -0.00001574, 0.00000788, 0.00000384, 0.00000649, -0.00001155, 0.00000192, -0.00000774, 0.00001878, -0.00001753, 0.00001746, -0.00001437, 0.00001033, -0.00001476, 0.00001225, -0.00002798, 0.00001734, -0.00002058, 0.00001540, 0.00000992, 0.00000018, 0.00000262, 0.00001993, -0.00001494, 0.00001407, -0.00001596, 0.00002538, -0.00001756, 0.00001444, 0.00000185, 0.00000479, 0.00000374, -0.00002059, 0.00000823, -0.00000784, -0.00000178, -0.00000376, -0.00000851, 0.00000537, -0.00000288, 0.00000896, -0.00001237, 0.00000817, -0.00002366, 0.00001389, -0.00001729, 0.00001809, -0.00002294, 0.00001251, -0.00002099, 0.00002477, -0.00003619, 0.00002799, -0.00003890, 0.00001403, -0.00005275, 0.00003601, -0.00010115, 0.00006051, -0.00030156, -0.00002202, 0.00043428, 0.00042438, 0.00004384, -0.00033383, 0.00006702, -0.00012286, 0.00005213, -0.00006244, 0.00003301, -0.00004833, 0.00003757, -0.00005035, 0.00002812, -0.00003069, 0.00002446, -0.00002263, 0.00001389, -0.00002235, 0.00001812, -0.00001426, 0.00001850, -0.00001104, 0.00000943, -0.00001735, 0.00001340, -0.00002107, 0.00002434, -0.00001167, 0.00001378, -0.00000730, -0.00000671, 0.00000361, -0.00000214, -0.00000738, 0.00000199, 0.00000363, -0.00000125, -0.00001804, 0.00001012, 0.00000062, 0.00000181, -0.00001001, -0.00000182, -0.00000601, 0.00000999, -0.00002245, 0.00002376, -0.00001804, 0.00000280, 0.00000873, -0.00001279, 0.00001190, -0.00000682, 0.00000164, -0.00000885, 0.00000551, -0.00000072, 0.00001307, },
		{ -0.00001914, 0.00000212, 0.00000869, -0.00000835, 0.00000471, -0.00000058, 0.00000016, -0.00000891, 0.00001455, -0.00000717, -0.00000769, 0.00000489, 0.00000837, -0.00000142, -0.00000609, -0.00000489, 0.00000497, 0.00000638, -0.00000849, 0.00001118, -0.00001062, 0.00000555, -0.00000911, 0.00000461, -0.00002325, 0.00002135, -0.00002839, 0.00003143, -0.00000597, 0.00000669, 0.00000386, 0.00001676, -0.00001723, 0.00001564, -0.00001660, 0.00003011, -0.00003008, 0.00003089, -0.00001413, 0.00001971, -0.00001475, -0.00000596, 0.00000121, -0.00000426, -0.00000405, -0.00000416, -0.00000556, 0.00000429, -0.00000015, 0.00000444, -0.00000894, 0.00000206, -0.00001725, 0.00001025, -0.00001272, 0.00001393, -0.00002187, 0.00001089, -0.00001532, 0.00001723, -0.00003101, 0.00002523, -0.00004162, 0.00001119, -0.00004418, 0.00002133, -0.00009072, 0.00003984, -0.00037404, 0.00015991, 0.00048602, 0.00042103, -0.00023977, -0.00016279, 0.00000857, -0.00007496, 0.00002705, -0.00004009, 0.00001248, -0.00002670, 0.00001754, -0.00003651, 0.00002147, -0.00002427, 0.00001949, -0.00001856, 0.00000815, -0.00001647, 0.00001532, -0.00001098, 0.00001556, -0.00000859, 0.00000305, -0.00001023, 0.00000675, -0.00001455, 0.00002321, -0.00001356, 0.00001546, -0.00001205, -0.00000317, 0.00000488, -0.00000672, -0.00000393, 0.00000256, 0.00000553, -0.00001057, -0.00001091, 0.00001166, -0.00000023, -0.00000295, -0.00000591, -0.00000490, 0.00000136, -0.00000119, -0.00001259, 0.00002284, -0.00002804, 0.00002125, -0.00001001, 0.00000123, 0.00000323, -0.00000269, -0.00000255, -0.00000376, 0.00000240, 0.00000387, 0.00001005, },
		{ -0.00001827, 0.00000859, 0.00000361, -0.00000726, 0.00000668, -0.00000255, -0.00000099, -0.00000651, 0.00001646, -0.00001491, 0.00000030, 0.00000135, 0.00001248, -0.00000932, -0.00000074, -0.00000962, 0.00001650, -0.00000879, 0.00000659, -0.00000308, 0.00000144, -0.00000753, 0.00000595, -0.00001577, 0.00000036, 0.00000303, -0.00001394, 0.00002783, -0.00000985, 0.00000756, 0.00001053, 0.00000443, -0.00000850, 0.00000632, -0.00000379, 0.00001753, -0.00002466, 0.00003250, -0.00001806, 0.00002346, -0.00002437, 0.00000454, -0.00000445, -0.00000178, -0.00000572, -0.00000450, -0.00000140, 0.00000066, 0.00000542, -0.00000405, -0.00000112, -0.00000865, -0.00000404, -0.00000205, 0.00000188, -0.00000258, -0.00000759, -0.00000283, 0.00000263, -0.00000550, -0.00000685, -0.00000014, -0.00002166, -0.00001249, -0.00001275, -0.00002396, -0.00003843, -0.00005277, -0.00035543, 0.00031543, 0.00058272, 0.00023246, -0.00037203, -0.00000427, -0.00007636, -0.00000136, -0.00002272, 0.00000124, -0.00002388, 0.00000967, -0.00001909, -0.00000348, -0.00000342, -0.00000237, 0.00000093, -0.00000283, -0.00000835, 0.00000065, 0.00000197, 0.00000096, 0.00000545, -0.00000112, -0.00000735, 0.00000202, -0.00000613, -0.00000017, 0.00001341, -0.00000843, 0.00001128, -0.00001183, -0.00000229, 0.00000760, -0.00001421, 0.00000483, -0.00000137, 0.00000857, -0.00002197, 0.00000315, 0.00000615, 0.00000163, -0.00000906, 0.00000119, -0.00001069, 0.00001042, -0.00001686, 0.00000885, 0.00000318, -0.00001630, 0.00002103, -0.00001732, 0.00001031, -0.00000454, 0.00000167, -0.00000622, 0.00000117, -0.00000161, 0.00001119, 0.00000073, },
		{ -0.00001382, 0.00001293, -0.00000292, -0.00000262, 0.00000519, -0.00000263, -0.00000366, -0.00000043, 0.00001213, -0.00001612, 0.00000338, 0.00000222, 0.00001105, -0.00001190, -0.00000076, -0.00000563, 0.00001686, -0.00001392, 0.00001374, -0.00001130, 0.00000789, -0.00001417, 0.00001380, -0.00002991, 0.00002374, -0.00002214, 0.00001264, 0.00000885, -0.00000169, 0.00000318, 0.00001883, -0.00001108, 0.00000506, -0.00000698, 0.00001433, -0.00000520, -0.00000511, 0.00002035, -0.00001023, 0.00001554, -0.00002277, 0.00000748, -0.00000608, -0.00000192, -0.00000629, -0.00000349, 0.00000036, -0.00000040, 0.00000755, -0.00000926, 0.00000306, -0.00001427, 0.00000413, -0.00000960, 0.00001182, -0.00001618, 0.00000502, -0.00001346, 0.00001561, -0.00002343, 0.00001427, -0.00002591, -0.00000043, -0.00003394, 0.00001316, -0.00006442, 0.00001167, -0.00017198, -0.00027552, 0.00048998, 0.00062704, -0.00007767, -0.00030289, 0.00003928, -0.00010129, 0.00003164, -0.00004460, 0.00001839, -0.00003786, 0.00002512, -0.00003841, 0.00001815, -0.00002073, 0.00001300, -0.00001168, 0.00000653, -0.00001845, 0.00001336, -0.00000859, 0.00001026, -0.00000181, 0.00000169, -0.00001125, 0.00000804, -0.00001375, 0.00001116, 0.00000427, -0.00000440, 0.00000921, -0.00001456, 0.00000399, 0.00000250, -0.00001410, 0.00000902, -0.00000205, 0.00000510, -0.00002574, 0.00001458, 0.00000001, 0.00000130, -0.00001019, 0.00000370, -0.00001132, 0.00001149, -0.00002335, 0.00002529, -0.00001989, 0.00000757, 0.00000392, -0.00000877, 0.00000726, -0.00000414, 0.00000027, -0.00000411, 0.00000016, 0.00000026, 0.00001369, -0.00000833, },
		{ -0.00000873, 0.00001531, -0.00000967, 0.00000474, -0.00000004, -0.00000015, -0.00000745, 0.00000770, 0.00000329, -0.00001165, 0.00000168, 0.00000745, 0.00000403, -0.00000922, -0.00000487, 0.00000520, 0.00000608, -0.00000607, 0.00000767, -0.00000695, 0.00000161, -0.00000619, 0.00000368, -0.00002229, 0.00002644, -0.00003197, 0.00003064, -0.00000928, 0.00000828, 0.00000059, 0.00002084, -0.00002013, 0.00001357, -0.00001371, 0.00002493, -0.00002400, 0.00001628, 0.00000377, 0.00000156, 0.00000296, -0.00001488, 0.00000537, -0.00000506, -0.00000386, -0.00000568, -0.00000136, -0.00000094, 0.00000273, 0.00000381, -0.00000815, 0.00000054, -0.00001118, 0.00000216, -0.00000568, 0.00000846, -0.00001667, 0.00000569, -0.00001103, 0.00001254, -0.00002299, 0.00001581, -0.00003385, 0.00000457, -0.00003604, 0.00001355, -0.00007419, 0.00002284, -0.00027425, -0.00015897, 0.00071963, 0.00048893, -0.00033216, -0.00013201, -0.00001465, -0.00005844, 0.00001170, -0.00002648, 0.00000123, -0.00002006, 0.00000930, -0.00002782, 0.00001463, -0.00001792, 0.00001124, -0.00000927, 0.00000124, -0.00001365, 0.00001309, -0.00000901, 0.00001134, -0.00000253, -0.00000247, -0.00000621, 0.00000444, -0.00001191, 0.00001500, -0.00000092, -0.00000295, 0.00000989, -0.00002044, 0.00001578, -0.00001179, -0.00000257, 0.00000361, 0.00000433, -0.00000799, -0.00001635, 0.00001600, -0.00000141, -0.00000385, -0.00000377, -0.00000184, -0.00000349, 0.00000045, -0.00001241, 0.00002217, -0.00002684, 0.00002332, -0.00001374, 0.00000516, -0.00000231, 0.00000162, -0.00000478, 0.00000164, -0.00000539, 0.00000830, 0.00000853, -0.00001277, },
		{ -0.00000274, 0.00001377, -0.00001288, 0.00001058, -0.00000566, 0.00000248, -0.00000967, 0.00001394, -0.00000528, -0.00000631, -0.00000014, 0.00001209, -0.00000354, -0.00000636, -0.00000621, 0.00001325, -0.00000529, 0.00000509, -0.00000326, 0.00000244, -0.00000963, 0.00000814, -0.00001658, 0.00000274, 0.00000634, -0.00001821, 0.00002785, -0.00001398, 0.00001127, 0.00000442, 0.00001295, -0.00001821, 0.00001258, -0.00000955, 0.00002146, -0.00002917, 0.00002922, -0.00000883, 0.00001023, -0.00000748, -0.00000643, 0.00000182, -0.00000341, -0.00000674, -0.00000283, -0.00000111, -0.00000141, 0.00000568, -0.00000129, -0.00000534, -0.00000301, -0.00000624, -0.00000264, 0.00000237, -0.00000164, -0.00000935, -0.00000029, -0.00000212, 0.00000095, -0.00001140, 0.00000320, -0.00002774, -0.00000178, -0.00002677, -0.00000153, -0.00006361, -0.00000069, -0.00035924, 0.00003035, 0.00093971, 0.00014084, -0.00038086, -0.00001363, -0.00006932, -0.00000918, -0.00001735, -0.00000261, -0.00002048, 0.00000270, -0.00001409, -0.00000633, 0.00000010, -0.00000633, 0.00000260, -0.00000176, -0.00000914, -0.00000198, 0.00000582, -0.00000418, 0.00000912, -0.00000262, -0.00000616, -0.00000019, -0.00000153, -0.00000654, 0.00001596, -0.00000694, 0.00000239, 0.00000520, -0.00001991, 0.00002221, -0.00002443, 0.00001166, -0.00000520, 0.00001145, -0.00002328, 0.00000080, 0.00000930, 0.00000025, -0.00000969, 0.00000485, -0.00001012, 0.00000634, -0.00001455, 0.00000889, 0.00000126, -0.00001182, 0.00001845, -0.00001682, 0.00001054, -0.00000730, 0.00000438, -0.00000634, 0.00000376, -0.00000800, 0.00001577, -0.00000011, -0.00001357, },
		{ 0.00000535, 0.00000664, -0.00000930, 0.00001048, -0.00000738, 0.00000211, -0.00000768, 0.00001510, -0.00000974, -0.00000406, 0.00000198, 0.00001164, -0.00000718, -0.00000709, -0.00000020, 0.00001089, -0.00000668, 0.00000755, -0.00000664, 0.00000410, -0.00001176, 0.00001219, -0.00002710, 0.00002399, -0.00001841, 0.00000671, 0.00001017, -0.00000622, 0.00000774, 0.00001162, -0.00000030, -0.00000926, 0.00000620, 0.00000032, 0.00000806, -0.00002128, 0.00003151, -0.00001533, 0.00001446, -0.00001402, 0.00000050, -0.00000186, -0.00000188, -0.00000994, 0.00000264, -0.00000511, 0.00000295, 0.00000328, -0.00000188, -0.00000686, -0.00000061, -0.00000827, 0.00000038, 0.00000219, -0.00000528, -0.00000695, -0.00000026, -0.00000090, -0.00000282, -0.00000677, -0.00000531, -0.00002441, -0.00000245, -0.00002634, -0.00000769, -0.00006166, -0.00003241, -0.00044696, 0.00036484, 0.00098190, -0.00027286, -0.00022974, -0.00001978, -0.00005633, -0.00000685, -0.00001310, -0.00000773, -0.00001575, 0.00000115, -0.00001677, 0.00000031, -0.00000386, -0.00000392, 0.00000283, -0.00000392, -0.00001006, 0.00000315, 0.00000246, -0.00000258, 0.00001038, -0.00000836, -0.00000215, -0.00000134, -0.00000185, -0.00000452, 0.00001882, -0.00001729, 0.00001584, -0.00001076, -0.00000495, 0.00001250, -0.00002242, 0.00001576, -0.00000665, 0.00000900, -0.00002860, 0.00001388, 0.00000257, -0.00000085, -0.00000924, 0.00000710, -0.00001215, 0.00000769, -0.00001918, 0.00002263, -0.00001942, 0.00001164, -0.00000081, -0.00000502, 0.00000360, -0.00000308, 0.00000016, -0.00000101, -0.00000185, -0.00000206, 0.00001566, -0.00000592, -0.00001405, },
		{ 0.00001404, -0.00000395, -0.00000004, 0.00000400, -0.00000405, -0.00000184, -0.00000172, 0.00001159, -0.00001045, -0.00000423, 0.00000706, 0.00000679, -0.00000774, -0.00000876, 0.00000903, 0.00000111, 0.00000130, -0.00000005, 0.00000000, -0.00000491, -0.00000076, -0.00000076, -0.00001602, 0.00002365, -0.00002697, 0.00002276, -0.00000653, 0.00000407, 0.00000410, 0.00001552, -0.00001128, -0.00000018, 0.00000096, 0.00000777, -0.00000629, -0.00000677, 0.00002586, -0.00001698, 0.00001571, -0.00001756, 0.00000574, -0.00000509, -0.00000150, -0.00001103, 0.00000733, -0.00001023, 0.00000976, -0.00000292, 0.00000083, -0.00001080, 0.00000535, -0.00001542, 0.00001047, -0.00000726, 0.00000006, -0.00001190, 0.00000739, -0.00000911, 0.00000424, -0.00001401, -0.00000397, -0.00002776, 0.00000383, -0.00003518, -0.00000436, -0.00007012, -0.00007933, -0.00050231, 0.00084472, 0.00069813, -0.00051777, -0.00003742, -0.00008865, -0.00000071, -0.00003780, 0.00001501, -0.00003358, 0.00000871, -0.00002049, 0.00000056, -0.00000975, 0.00000459, -0.00001138, 0.00001169, -0.00001553, -0.00000051, -0.00000028, 0.00000433, -0.00000384, 0.00001275, -0.00001592, 0.00000593, -0.00000714, 0.00000167, -0.00000372, 0.00001939, -0.00002515, 0.00002953, -0.00003052, 0.00001841, -0.00001021, -0.00000506, 0.00000665, 0.00000182, -0.00000521, -0.00001909, 0.00001595, 0.00000074, -0.00000645, -0.00000115, 0.00000063, -0.00000559, -0.00000212, -0.00000731, 0.00001645, -0.00002176, 0.00002308, -0.00001632, 0.00000844, -0.00000699, 0.00000460, -0.00000623, 0.00000637, -0.00001080, 0.00000983, 0.00000769, -0.00000686, -0.00001495, },
		{ 0.00001872, -0.00001219, 0.00000921, -0.00000444, 0.00000133, -0.00000669, 0.00000520, 0.00000653, -0.00001070, -0.00000286, 0.00001040, 0.00000215, -0.00000958, -0.00000509, 0.00001199, -0.00000488, 0.00000750, -0.00000716, 0.00000637, -0.00001364, 0.00001142, -0.00001861, 0.00000747, 0.00000436, -0.00001508, 0.00002093, -0.00001207, 0.00000963, 0.00000386, 0.00001359, -0.00001680, 0.00000655, -0.00000179, 0.00001016, -0.00001658, 0.00000967, 0.00001382, -0.00001299, 0.00001299, -0.00001699, 0.00000743, -0.00000572, -0.00000480, -0.00000585, 0.00000478, -0.00000863, 0.00001054, -0.00000437, -0.00000172, -0.00000731, 0.00000300, -0.00001425, 0.00001330, -0.00001191, 0.00000134, -0.00001095, 0.00000775, -0.00001015, 0.00000458, -0.00001590, -0.00000935, -0.00001917, -0.00000636, -0.00002451, -0.00002651, -0.00004816, -0.00020290, -0.00039463, 0.00130454, 0.00012193, -0.00045306, 0.00002511, -0.00010324, 0.00001814, -0.00004244, 0.00001898, -0.00003827, 0.00001528, -0.00002851, 0.00000931, -0.00001341, 0.00000592, -0.00001080, 0.00001187, -0.00002002, 0.00000518, -0.00000086, 0.00000187, 0.00000123, 0.00000618, -0.00001377, 0.00000613, -0.00000655, -0.00000066, 0.00000463, 0.00000781, -0.00001765, 0.00002775, -0.00003657, 0.00003189, -0.00002861, 0.00001352, -0.00000511, 0.00001048, -0.00002094, -0.00000260, 0.00001084, 0.00000057, -0.00001079, 0.00000731, -0.00000745, 0.00000164, -0.00001316, 0.00001106, -0.00000377, -0.00000443, 0.00001392, -0.00001547, 0.00001077, -0.00001021, 0.00000685, -0.00000692, 0.00000745, -0.00001401, 0.00001884, -0.00000171, -0.00000640, -0.00001329, },
		{ 0.00001568, -0.00001294, 0.00001246, -0.00000922, 0.00000445, -0.00000915, 0.00001023, 0.00000216, -0.00001225, 0.00000217, 0.00000911, 0.00000028, -0.00001394, 0.00000543, 0.00000425, 0.00000034, 0.00000269, -0.00000362, 0.00000153, -0.00000946, 0.00000921, -0.00002193, 0.00002160, -0.00001420, 0.00000304, 0.00000877, -0.00000855, 0.00001116, 0.00000470, 0.00000892, -0.00001887, 0.00001297, -0.00000563, 0.00001079, -0.00002315, 0.00002643, -0.00000454, -0.00000104, 0.00000340, -0.00000974, 0.00000309, -0.00000189, -0.00001254, 0.00000650, -0.00000747, 0.00000378, 0.00000010, 0.00000435, -0.00001449, 0.00000898, -0.00001481, 0.00000462, -0.00000273, 0.00000099, -0.00001381, 0.00000793, -0.00001225, 0.00001095, -0.00001881, 0.00000519, -0.00003778, 0.00001730, -0.00005131, 0.00002772, -0.00010159, 0.00003522, -0.00045183, 0.00002595, 0.00145656, -0.00048754, -0.00016003, -0.00007331, -0.00001925, -0.00003659, 0.00000930, -0.00002616, 0.00000093, -0.00001804, -0.00000091, -0.00001119, 0.00000673, -0.00001413, 0.00001033, -0.00000874, -0.00000500, -0.00000446, 0.00001024, -0.00001106, 0.00001690, -0.00001303, 0.00000281, -0.00000718, 0.00000581, -0.00001291, 0.00002265, -0.00001739, 0.00000873, 0.00000361, -0.00001854, 0.00002186, -0.00002687, 0.00001787, -0.00000663, 0.00000753, -0.00002497, 0.00000824, 0.00000717, -0.00000386, -0.00000696, 0.00000754, -0.00000830, -0.00000023, -0.00001201, 0.00001814, -0.00001855, 0.00001585, -0.00000549, -0.00000201, 0.00000153, -0.00000393, 0.00000152, -0.00000001, -0.00000058, -0.00000659, 0.00001810, -0.00000587, -0.00000880, -0.00000619, },
		{ 0.00000615, -0.00000615, 0.00000820, -0.00000804, 0.00000342, -0.00000811, 0.00001302, -0.00000238, -0.00001280, 0.00000802, 0.00000557, -0.00000154, -0.00001563, 0.00001534, -0.00000698, 0.00001131, -0.00000929, 0.00000776, -0.00001213, 0.00000614, -0.00000831, -0.00000485, 0.00001405, -0.00001568, 0.00001100, -0.00000081, -0.00000329, 0.00001196, 0.00000393, 0.00000417, -0.00001838, 0.00001874, -0.00001134, 0.00001129, -0.00002527, 0.00003938, -0.00002459, 0.00001609, -0.00001140, 0.00000243, -0.00000521, 0.00000329, -0.00001924, 0.00001845, -0.00002135, 0.00001925, -0.00001422, 0.00001564, -0.00002815, 0.00002704, -0.00003658, 0.00003021, -0.00002806, 0.00002300, -0.00003629, 0.00003392, -0.00004055, 0.00004172, -0.00005382, 0.00003682, -0.00007335, 0.00006023, -0.00010430, 0.00008875, -0.00018965, 0.00012257, -0.00074765, 0.00076060, 0.00108692, -0.00079483, 0.00011028, -0.00020170, 0.00008618, -0.00010978, 0.00007384, -0.00008313, 0.00005193, -0.00006407, 0.00004065, -0.00004394, 0.00003567, -0.00004090, 0.00003764, -0.00003703, 0.00001893, -0.00002081, 0.00002442, -0.00002457, 0.00003156, -0.00003214, 0.00002154, -0.00002301, 0.00001960, -0.00002423, 0.00003730, -0.00004078, 0.00003836, -0.00002922, 0.00001306, -0.00000528, -0.00000690, 0.00000709, 0.00000266, -0.00000709, -0.00001483, 0.00000887, 0.00000751, -0.00001185, 0.00000378, -0.00000032, -0.00000210, -0.00001060, 0.00000319, 0.00000612, -0.00001344, 0.00002049, -0.00001693, 0.00000956, -0.00000880, 0.00000432, -0.00000496, 0.00000797, -0.00001154, 0.00000714, 0.00000883, -0.00000402, -0.00001402, 0.00000495, },
		{ -0.00000413, 0.00000305, 0.00000015, -0.00000303, -0.00000043, -0.00000465, 0.00001461, -0.00000904, -0.00000864, 0.00000979, 0.00000442, -0.00000767, -0.00000812, 0.00001432, -0.00000889, 0.00001435, -0.00001461, 0.00001292, -0.00001955, 0.00001661, -0.00002368, 0.00001700, -0.00000534, -0.00000297, 0.00000641, -0.00000343, 0.00000137, 0.00001097, 0.00000392, -0.00000309, -0.00001115, 0.00001724, -0.00001242, 0.00000676, -0.00001716, 0.00003907, -0.00003460, 0.00002751, -0.00002199, 0.00001092, -0.00000951, 0.00000180, -0.00001501, 0.00001715, -0.00002189, 0.00002180, -0.00001611, 0.00001273, -0.00002390, 0.00002483, -0.00003734, 0.00003555, -0.00003506, 0.00002666, -0.00003760, 0.00003580, -0.00004257, 0.00004427, -0.00006035, 0.00003718, -0.00006959, 0.00005437, -0.00009843, 0.00007555, -0.00018537, 0.00006200, -0.00084712, 0.00154717, 0.00026573, -0.00063639, 0.00014051, -0.00019783, 0.00009456, -0.00010815, 0.00007394, -0.00008452, 0.00005435, -0.00006813, 0.00004780, -0.00004741, 0.00003696, -0.00004083, 0.00003923, -0.00004292, 0.00002500, -0.00002110, 0.00002156, -0.00001955, 0.00002578, -0.00003067, 0.00002189, -0.00002202, 0.00001683, -0.00001724, 0.00002970, -0.00004015, 0.00004630, -0.00004496, 0.00003314, -0.00002672, 0.00001256, -0.00000432, 0.00000971, -0.00002000, -0.00000159, 0.00000635, 0.00000564, -0.00001475, 0.00001135, -0.00000666, 0.00000094, -0.00001669, 0.00001825, -0.00001354, 0.00000661, 0.00000663, -0.00001204, 0.00000912, -0.00001073, 0.00000643, -0.00000522, 0.00000882, -0.00001592, 0.00001750, -0.00000136, -0.00000110, -0.00001764, 0.00001470, },
		{ -0.00000941, 0.00000868, -0.00000634, 0.00000193, -0.00000473, 0.00000023, 0.00001455, -0.00001684, 0.00000025, 0.00000628, 0.00000637, -0.00001707, 0.00000755, 0.00000068, 0.00000335, 0.00000237, -0.00000468, 0.00000216, -0.00000919, 0.00000740, -0.00001807, 0.00002132, -0.00001491, 0.00000675, 0.00000029, -0.00000446, 0.00000852, 0.00000431, 0.00000896, -0.00001613, 0.00000660, 0.00000224, -0.00000163, -0.00000832, 0.00000589, 0.00001812, -0.00002391, 0.00002190, -0.00001831, 0.00000717, -0.00000319, -0.00001092, 0.00000424, -0.00000294, -0.00000180, 0.00000305, 0.00000282, -0.00001205, 0.00000634, -0.00000789, -0.00000393, 0.00000457, -0.00000593, -0.00000592, 0.00000014, -0.00000564, 0.00000341, -0.00000640, -0.00001095, -0.00002154, 0.00000232, -0.00003340, 0.00000698, -0.00006532, -0.00001594, -0.00027142, -0.00047465, 0.00194545, -0.00062254, -0.00016368, -0.00007032, -0.00003010, -0.00002538, -0.00000364, -0.00001492, -0.00000746, -0.00001407, -0.00000753, -0.00000167, -0.00000202, -0.00000606, 0.00000105, 0.00000043, -0.00001145, -0.00000066, 0.00000570, -0.00000648, 0.00001012, -0.00000588, -0.00000241, -0.00000307, 0.00000237, -0.00000788, 0.00001196, -0.00000414, -0.00000860, 0.00002162, -0.00002924, 0.00002489, -0.00002443, 0.00001458, -0.00000260, 0.00000244, -0.00001814, 0.00000220, 0.00000944, -0.00000449, -0.00000571, 0.00000733, -0.00000318, -0.00000771, -0.00000654, 0.00001594, -0.00002016, 0.00002179, -0.00001112, 0.00000122, -0.00000024, -0.00000479, 0.00000230, 0.00000094, 0.00000114, -0.00000979, 0.00001807, -0.00000570, -0.00000223, -0.00001533, 0.00001806, },
		{ -0.00000826, 0.00000841, -0.00000839, 0.00000443, -0.00000813, 0.00000679, 0.00001014, -0.00002095, 0.00000863, 0.00000206, 0.00000626, -0.00002168, 0.00002042, -0.00001422, 0.00001953, -0.00001580, 0.00001269, -0.00001705, 0.00001193, -0.00001622, 0.00000755, 0.00000261, -0.00000424, 0.00000175, 0.00000202, -0.00000825, 0.00001817, -0.00000654, 0.00001671, -0.00003025, 0.00002850, -0.00002115, 0.00001744, -0.00002896, 0.00003654, -0.00001696, 0.00000430, 0.00000036, -0.00000151, -0.00000660, 0.00000937, -0.00002703, 0.00002755, -0.00002960, 0.00002658, -0.00002459, 0.00002914, -0.00004270, 0.00004389, -0.00005121, 0.00004384, -0.00004395, 0.00004120, -0.00005543, 0.00005510, -0.00006612, 0.00007128, -0.00008354, 0.00006667, -0.00010725, 0.00010295, -0.00015564, 0.00015536, -0.00026757, 0.00023786, -0.00080060, 0.00043087, 0.00161879, -0.00110077, 0.00025646, -0.00030913, 0.00016449, -0.00017219, 0.00012284, -0.00012434, 0.00008893, -0.00010183, 0.00007402, -0.00007114, 0.00005994, -0.00006324, 0.00005633, -0.00005290, 0.00003512, -0.00003864, 0.00004094, -0.00004042, 0.00004412, -0.00004244, 0.00003247, -0.00003463, 0.00003226, -0.00003640, 0.00004367, -0.00004308, 0.00003377, -0.00001889, 0.00000517, -0.00000302, -0.00000311, 0.00000143, 0.00000935, -0.00001522, -0.00000349, -0.00000353, 0.00001658, -0.00001832, 0.00001026, -0.00000445, 0.00000593, -0.00002252, 0.00001579, -0.00000537, -0.00000465, 0.00001705, -0.00001611, 0.00000853, -0.00000769, 0.00000108, -0.00000140, 0.00000669, -0.00000845, 0.00000203, 0.00001051, -0.00000362, -0.00000789, -0.00000711, 0.00001447, },
		{ -0.00000361, 0.00000418, -0.00000662, 0.00000424, -0.00001005, 0.00001484, -0.00000043, -0.00001680, 0.00001052, 0.00000268, -0.00000124, -0.00001373, 0.00001857, -0.00001526, 0.00002285, -0.00002258, 0.00001937, -0.00002528, 0.00002243, -0.00003091, 0.00003005, -0.00002040, 0.00001442, -0.00001127, 0.00000778, -0.00000990, 0.00002224, -0.00001163, 0.00001632, -0.00003188, 0.00003761, -0.00003516, 0.00002829, -0.00003795, 0.00005441, -0.00004444, 0.00003047, -0.00002045, 0.00001343, -0.00001608, 0.00001332, -0.00002906, 0.00003363, -0.00003863, 0.00003739, -0.00003432, 0.00003478, -0.00004802, 0.00005303, -0.00006521, 0.00006285, -0.00006501, 0.00006019, -0.00007335, 0.00007445, -0.00008725, 0.00009654, -0.00011732, 0.00009791, -0.00013728, 0.00013599, -0.00019717, 0.00020506, -0.00035250, 0.00033289, -0.00118798, 0.00155278, 0.00061842, -0.00093126, 0.00032094, -0.00033935, 0.00020588, -0.00020084, 0.00015006, -0.00014891, 0.00011072, -0.00012365, 0.00009967, -0.00009281, 0.00007806, -0.00007911, 0.00007298, -0.00007210, 0.00005274, -0.00004999, 0.00004924, -0.00004688, 0.00004972, -0.00005109, 0.00004181, -0.00004220, 0.00003860, -0.00004049, 0.00004867, -0.00005542, 0.00005383, -0.00004292, 0.00002819, -0.00002356, 0.00001433, -0.00000810, 0.00001418, -0.00002536, 0.00000775, -0.00000623, 0.00001628, -0.00002243, 0.00001865, -0.00001029, 0.00000655, -0.00002619, 0.00003023, -0.00002690, 0.00002003, -0.00000293, -0.00000604, 0.00000434, -0.00000718, 0.00000160, 0.00000026, 0.00000594, -0.00001208, 0.00001110, 0.00000171, 0.00000030, -0.00001424, 0.00000309, 0.00000749, },
		{ 0.00000071, -0.00000066, -0.00000308, 0.00000175, -0.00000908, 0.00002138, -0.00001421, -0.00000544, 0.00000544, 0.00000800, -0.00001415, 0.00000467, 0.00000151, 0.00000129, 0.00000685, -0.00000953, 0.00000569, -0.00001118, 0.00000797, -0.00001768, 0.00002569, -0.00002264, 0.00001874, -0.00001519, 0.00000607, -0.00000088, 0.00001197, -0.00000173, -0.00000100, -0.00001103, 0.00002038, -0.00002394, 0.00001560, -0.00002011, 0.00004132, -0.00004298, 0.00003345, -0.00002195, 0.00001064, -0.00000802, -0.00000235, -0.00000640, 0.00001011, -0.00001550, 0.00001478, -0.00000998, 0.00000394, -0.00001128, 0.00001409, -0.00002637, 0.00002578, -0.00002861, 0.00001984, -0.00002735, 0.00002333, -0.00002998, 0.00003435, -0.00005731, 0.00002870, -0.00005298, 0.00003360, -0.00007508, 0.00004823, -0.00017184, 0.00003351, -0.00100993, 0.00230186, -0.00057368, -0.00029820, 0.00002391, -0.00011257, 0.00004352, -0.00006420, 0.00003575, -0.00004983, 0.00002138, -0.00004345, 0.00003405, -0.00003423, 0.00002348, -0.00002726, 0.00002561, -0.00003221, 0.00001815, -0.00001464, 0.00001377, -0.00001135, 0.00001368, -0.00001847, 0.00001200, -0.00001282, 0.00000953, -0.00000950, 0.00001611, -0.00002773, 0.00003491, -0.00003182, 0.00002050, -0.00001749, 0.00001026, 0.00000026, -0.00000045, -0.00001465, 0.00000224, 0.00000433, 0.00000085, -0.00000856, 0.00001003, -0.00000147, -0.00000962, -0.00000712, 0.00001910, -0.00002599, 0.00003030, -0.00001863, 0.00000613, -0.00000387, -0.00000289, -0.00000031, 0.00000469, -0.00000015, -0.00000812, 0.00001293, -0.00000288, 0.00000148, -0.00001723, 0.00001108, 0.00000114, },
		{ 0.00000302, -0.00000418, 0.00000079, -0.00000288, -0.00000319, 0.00002129, -0.00002398, 0.00000607, -0.00000056, 0.00001130, -0.00002267, 0.00002060, -0.00001688, 0.00002223, -0.00001646, 0.00001237, -0.00001753, 0.00001431, -0.00002129, 0.00001664, -0.00000503, 0.00000134, 0.00000039, -0.00000246, -0.00000651, 0.00001772, -0.00000949, 0.00001882, -0.00002856, 0.00002420, -0.00001675, 0.00000865, -0.00001670, 0.00001945, 0.00000068, -0.00001215, 0.00001086, -0.00000264, -0.00000910, 0.00001368, -0.00002962, 0.00002932, -0.00002982, 0.00002630, -0.00002736, 0.00003308, -0.00004514, 0.00004630, -0.00005028, 0.00004267, -0.00004551, 0.00004389, -0.00005725, 0.00005733, -0.00007021, 0.00007551, -0.00008462, 0.00006559, -0.00010653, 0.00010452, -0.00015587, 0.00015527, -0.00025543, 0.00021483, -0.00064126, -0.00008731, 0.00217778, -0.00132444, 0.00032072, -0.00035177, 0.00018788, -0.00018785, 0.00013167, -0.00013163, 0.00009679, -0.00011243, 0.00008143, -0.00007449, 0.00006246, -0.00006519, 0.00005644, -0.00005315, 0.00003816, -0.00004335, 0.00004343, -0.00004164, 0.00004200, -0.00003968, 0.00003201, -0.00003498, 0.00003254, -0.00003427, 0.00003521, -0.00003218, 0.00001988, -0.00000640, 0.00000123, -0.00000743, 0.00000722, -0.00000889, 0.00001910, -0.00002569, 0.00000955, -0.00001491, 0.00002222, -0.00002148, 0.00001478, -0.00000816, 0.00001381, -0.00003216, 0.00002481, -0.00001308, 0.00000096, 0.00001507, -0.00001556, 0.00000705, -0.00000575, -0.00000238, 0.00000164, 0.00000534, -0.00000503, -0.00000150, 0.00000950, -0.00000282, -0.00000119, -0.00001559, 0.00001525, -0.00000301, },
		{ 0.00000378, -0.00000611, 0.00000408, -0.00000864, 0.00000766, 0.00001164, -0.00002358, 0.00001034, -0.00000074, 0.00000600, -0.00001779, 0.00002083, -0.00001982, 0.00002846, -0.00002680, 0.00002215, -0.00002785, 0.00002615, -0.00003710, 0.00004221, -0.00003488, 0.00002995, -0.00002493, 0.00001512, -0.00001807, 0.00003109, -0.00002494, 0.00003132, -0.00004460, 0.00004824, -0.00004710, 0.00003735, -0.00004347, 0.00005351, -0.00003997, 0.00002356, -0.00001615, 0.00001764, -0.00002572, 0.00002739, -0.00004347, 0.00004859, -0.00005306, 0.00005142, -0.00005203, 0.00005572, -0.00006930, 0.00007617, -0.00008643, 0.00008407, -0.00009013, 0.00008904, -0.00010391, 0.00010770, -0.00012568, 0.00014012, -0.00016336, 0.00014715, -0.00019277, 0.00020285, -0.00027605, 0.00030442, -0.00047177, 0.00050491, -0.00131224, 0.00127885, 0.00114375, -0.00127122, 0.00050654, -0.00047821, 0.00031014, -0.00028494, 0.00021723, -0.00020572, 0.00016127, -0.00017315, 0.00014445, -0.00013106, 0.00011237, -0.00011035, 0.00009976, -0.00009601, 0.00007750, -0.00007624, 0.00007329, -0.00006925, 0.00006745, -0.00006574, 0.00005715, -0.00005799, 0.00005472, -0.00005515, 0.00005628, -0.00005795, 0.00004938, -0.00003423, 0.00002321, -0.00002517, 0.00002319, -0.00001901, 0.00002525, -0.00003728, 0.00002394, -0.00002368, 0.00002898, -0.00003116, 0.00002752, -0.00001704, 0.00001632, -0.00003812, 0.00004279, -0.00003963, 0.00003235, -0.00001256, 0.00000127, -0.00000278, -0.00000038, -0.00000640, 0.00000815, 0.00000005, -0.00000439, 0.00000237, 0.00000534, -0.00000080, -0.00000607, -0.00001033, 0.00001605, -0.00000545, },
		{ 0.00000349, -0.00000633, 0.00000538, -0.00001253, 0.00001936, -0.00000443, -0.00001351, 0.00000624, 0.00000529, -0.00000654, -0.00000107, 0.00000465, -0.00000339, 0.00001308, -0.00001504, 0.00000951, -0.00001354, 0.00000978, -0.00001988, 0.00003319, -0.00003447, 0.00003452, -0.00003250, 0.00001828, -0.00001285, 0.00002351, -0.00001783, 0.00001899, -0.00003054, 0.00003841, -0.00004477, 0.00003614, -0.00003886, 0.00005380, -0.00005008, 0.00003476, -0.00002153, 0.00001578, -0.00001820, 0.00001354, -0.00002406, 0.00002922, -0.00003471, 0.00003309, -0.00003152, 0.00002973, -0.00003855, 0.00004449, -0.00005578, 0.00005454, -0.00006079, 0.00005670, -0.00006709, 0.00006635, -0.00007916, 0.00009111, -0.00011981, 0.00009751, -0.00013083, 0.00012744, -0.00018850, 0.00019625, -0.00036496, 0.00034921, -0.00142704, 0.00239825, -0.00028607, -0.00056846, 0.00019941, -0.00025335, 0.00015588, -0.00015867, 0.00011361, -0.00011582, 0.00007893, -0.00009909, 0.00008622, -0.00008059, 0.00006568, -0.00006629, 0.00005962, -0.00006170, 0.00004713, -0.00004555, 0.00004364, -0.00004079, 0.00003900, -0.00003952, 0.00003264, -0.00003383, 0.00003184, -0.00003275, 0.00003400, -0.00004040, 0.00003865, -0.00002662, 0.00001322, -0.00001265, 0.00001167, -0.00000410, 0.00000438, -0.00001958, 0.00001205, -0.00000979, 0.00001266, -0.00001621, 0.00001722, -0.00000550, -0.00000403, -0.00001433, 0.00002708, -0.00003473, 0.00004003, -0.00002745, 0.00001307, -0.00001043, 0.00000311, -0.00000740, 0.00001222, -0.00000572, -0.00000076, 0.00000292, 0.00000265, 0.00000104, -0.00001146, -0.00000257, 0.00001381, -0.00000648, },
		{ 0.00000179, -0.00000432, 0.00000307, -0.00001073, 0.00002521, -0.00001881, -0.00000088, -0.00000037, 0.00001103, -0.00001693, 0.00001504, -0.00001416, 0.00001886, -0.00001144, 0.00000737, -0.00001394, 0.00001293, -0.00002205, 0.00002001, -0.00000618, -0.00000098, 0.00000749, -0.00001380, 0.00000161, 0.00001021, -0.00000381, 0.00000964, -0.00001422, 0.00000868, -0.00000248, -0.00000914, 0.00000336, -0.00000174, 0.00001751, -0.00002359, 0.00001441, -0.00000020, -0.00001027, 0.00001225, -0.00002294, 0.00002047, -0.00001935, 0.00001564, -0.00001876, 0.00002298, -0.00003120, 0.00003090, -0.00003146, 0.00002459, -0.00002974, 0.00002709, -0.00003718, 0.00003543, -0.00004697, 0.00004847, -0.00005174, 0.00002836, -0.00006498, 0.00005756, -0.00009721, 0.00008208, -0.00015167, 0.00006877, -0.00033531, -0.00065758, 0.00260554, -0.00135089, 0.00023732, -0.00028917, 0.00012858, -0.00013694, 0.00008551, -0.00009297, 0.00006455, -0.00008587, 0.00005463, -0.00004760, 0.00003800, -0.00004250, 0.00003458, -0.00003452, 0.00002396, -0.00003028, 0.00002767, -0.00002514, 0.00002349, -0.00002295, 0.00001909, -0.00002276, 0.00001972, -0.00001900, 0.00001576, -0.00001425, 0.00000431, 0.00000162, 0.00000401, -0.00001683, 0.00001908, -0.00001720, 0.00002475, -0.00003093, 0.00001624, -0.00001772, 0.00001869, -0.00001714, 0.00001357, -0.00000716, 0.00001624, -0.00003401, 0.00002569, -0.00001369, 0.00000126, 0.00001593, -0.00001680, 0.00000720, -0.00000577, -0.00000284, 0.00000102, 0.00000699, -0.00000485, -0.00000005, 0.00000348, 0.00000067, 0.00000217, -0.00001635, 0.00000694, 0.00000823, -0.00000535, },
		{ -0.00000151, -0.00000030, -0.00000271, -0.00000198, 0.00002084, -0.00002423, 0.00000640, -0.00000258, 0.00000974, -0.00001632, 0.00001781, -0.00001924, 0.00002791, -0.00002465, 0.00001925, -0.00002540, 0.00002552, -0.00003894, 0.00004871, -0.00004286, 0.00003742, -0.00002920, 0.00001425, -0.00001893, 0.00003253, -0.00002972, 0.00003476, -0.00004285, 0.00004408, -0.00004438, 0.00003154, -0.00003383, 0.00003963, -0.00002718, 0.00001416, -0.00001467, 0.00002505, -0.00003616, 0.00003896, -0.00005176, 0.00005530, -0.00005858, 0.00005693, -0.00006096, 0.00006571, -0.00007706, 0.00008348, -0.00009076, 0.00008911, -0.00009923, 0.00010042, -0.00011497, 0.00011953, -0.00013970, 0.00015494, -0.00017710, 0.00016169, -0.00020935, 0.00022320, -0.00029701, 0.00032822, -0.00048903, 0.00052075, -0.00119659, 0.00078341, 0.00172089, -0.00153849, 0.00061050, -0.00054995, 0.00035684, -0.00031997, 0.00024233, -0.00022688, 0.00018075, -0.00019332, 0.00016122, -0.00014384, 0.00012364, -0.00012016, 0.00010707, -0.00010339, 0.00008797, -0.00008812, 0.00008219, -0.00007616, 0.00007005, -0.00006706, 0.00006075, -0.00006207, 0.00005809, -0.00005565, 0.00004991, -0.00004861, 0.00003755, -0.00002550, 0.00002347, -0.00003367, 0.00003793, -0.00003346, 0.00003809, -0.00004985, 0.00003972, -0.00003810, 0.00003683, -0.00003554, 0.00003344, -0.00002226, 0.00002455, -0.00004599, 0.00004972, -0.00004643, 0.00003948, -0.00001932, 0.00000739, -0.00000964, 0.00000658, -0.00001390, 0.00001477, -0.00000522, 0.00000288, -0.00000418, 0.00000597, -0.00000121, 0.00000185, -0.00001901, 0.00001635, 0.00000000, -0.00000175, },
		{ -0.00000508, 0.00000374, -0.00000896, 0.00001030, 0.00000798, -0.00001958, 0.00000596, 0.00000113, 0.00000156, -0.00000523, 0.00000607, -0.00000652, 0.00001664, -0.00001720, 0.00001033, -0.00001353, 0.00001026, -0.00002147, 0.00003926, -0.00004422, 0.00004674, -0.00004436, 0.00002569, -0.00002122, 0.00003220, -0.00003105, 0.00003372, -0.00004132, 0.00004608, -0.00005421, 0.00004395, -0.00004281, 0.00005119, -0.00004477, 0.00002911, -0.00002133, 0.00002489, -0.00003314, 0.00003323, -0.00004313, 0.00004786, -0.00005280, 0.00005112, -0.00005373, 0.00005543, -0.00006407, 0.00007105, -0.00008004, 0.00007896, -0.00008985, 0.00009052, -0.00010333, 0.00010569, -0.00012375, 0.00014066, -0.00017289, 0.00015666, -0.00019859, 0.00020862, -0.00028395, 0.00031550, -0.00051046, 0.00056217, -0.00161224, 0.00218577, 0.00020679, -0.00090786, 0.00039476, -0.00040101, 0.00026815, -0.00025019, 0.00018752, -0.00017872, 0.00013445, -0.00015168, 0.00013369, -0.00012202, 0.00010325, -0.00010046, 0.00008889, -0.00008801, 0.00007479, -0.00007506, 0.00007087, -0.00006628, 0.00005935, -0.00005623, 0.00005003, -0.00005153, 0.00004937, -0.00004851, 0.00004226, -0.00004268, 0.00003373, -0.00001891, 0.00000985, -0.00001581, 0.00002255, -0.00001730, 0.00001725, -0.00003199, 0.00002888, -0.00002850, 0.00002652, -0.00002528, 0.00002646, -0.00001312, 0.00000604, -0.00002397, 0.00003524, -0.00004203, 0.00004745, -0.00003522, 0.00002046, -0.00001864, 0.00001173, -0.00001692, 0.00002136, -0.00001347, 0.00000946, -0.00000844, 0.00000809, -0.00000143, -0.00000187, -0.00001643, 0.00002183, -0.00000785, 0.00000270, },
		{ -0.00000665, 0.00000509, -0.00001139, 0.00001980, -0.00000666, -0.00000998, 0.00000159, 0.00000634, -0.00000666, 0.00000659, -0.00000876, 0.00001236, -0.00000416, 0.00000101, -0.00000897, 0.00000966, -0.00001967, 0.00001865, -0.00000254, -0.00000801, 0.00001829, -0.00002660, 0.00001120, -0.00000001, 0.00000603, -0.00000505, 0.00000474, -0.00000854, 0.00001150, -0.00002511, 0.00001967, -0.00001576, 0.00002431, -0.00002455, 0.00001052, 0.00000281, -0.00000647, 0.00000252, -0.00000637, 0.00000225, -0.00000049, -0.00000308, -0.00000081, 0.00000139, -0.00000481, 0.00000310, -0.00000129, -0.00000382, -0.00000256, -0.00000338, -0.00000167, -0.00000334, -0.00000471, 0.00000033, 0.00000271, -0.00003092, 0.00000179, -0.00001964, -0.00000295, -0.00003256, -0.00000163, -0.00012874, -0.00000298, -0.00111849, 0.00276597, -0.00111849, -0.00000298, -0.00012874, -0.00000163, -0.00003256, -0.00000295, -0.00001964, 0.00000179, -0.00003092, 0.00000271, 0.00000033, -0.00000471, -0.00000334, -0.00000167, -0.00000338, -0.00000256, -0.00000382, -0.00000129, 0.00000310, -0.00000481, 0.00000139, -0.00000081, -0.00000308, -0.00000049, 0.00000225, -0.00000637, 0.00000252, -0.00000647, 0.00000281, 0.00001052, -0.00002455, 0.00002431, -0.00001576, 0.00001967, -0.00002511, 0.00001150, -0.00000854, 0.00000474, -0.00000505, 0.00000603, -0.00000001, 0.00001120, -0.00002660, 0.00001829, -0.00000801, -0.00000254, 0.00001865, -0.00001967, 0.00000966, -0.00000897, 0.00000101, -0.00000416, 0.00001236, -0.00000876, 0.00000659, -0.00000666, 0.00000634, 0.00000159, -0.00000998, -0.00000666, 0.00001980, -0.00001139, 0.00000509, },
		{ -0.00000508, 0.00000270, -0.00000785, 0.00002183, -0.00001643, -0.00000187, -0.00000143, 0.00000809, -0.00000844, 0.00000946, -0.00001347, 0.00002136, -0.00001692, 0.00001173, -0.00001864, 0.00002046, -0.00003522, 0.00004745, -0.00004203, 0.00003524, -0.00002397, 0.00000604, -0.00001312, 0.00002646, -0.00002528, 0.00002652, -0.00002850, 0.00002888, -0.00003199, 0.00001725, -0.00001730, 0.00002255, -0.00001581, 0.00000985, -0.00001891, 0.00003373, -0.00004268, 0.00004226, -0.00004851, 0.00004937, -0.00005153, 0.00005003, -0.00005623, 0.00005935, -0.00006628, 0.00007087, -0.00007506, 0.00007479, -0.00008801, 0.00008889, -0.00010046, 0.00010325, -0.00012202, 0.00013369, -0.00015168, 0.00013445, -0.00017872, 0.00018752, -0.00025019, 0.00026815, -0.00040101, 0.00039476, -0.00090786, 0.00020679, 0.00218577, -0.00161224, 0.00056217, -0.00051046, 0.00031550, -0.00028395, 0.00020862, -0.00019859, 0.00015666, -0.00017289, 0.00014066, -0.00012375, 0.00010569, -0.00010333, 0.00009052, -0.00008985, 0.00007896, -0.00008004, 0.00007105, -0.00006407, 0.00005543, -0.00005373, 0.00005112, -0.00005280, 0.00004786, -0.00004313, 0.00003323, -0.00003314, 0.00002489, -0.00002133, 0.00002911, -0.00004477, 0.00005119, -0.00004281, 0.00004395, -0.00005421, 0.00004608, -0.00004132, 0.00003372, -0.00003105, 0.00003220, -0.00002122, 0.00002569, -0.00004436, 0.00004674, -0.00004422, 0.00003926, -0.00002147, 0.00001026, -0.00001353, 0.00001033, -0.00001720, 0.00001664, -0.00000652, 0.00000607, -0.00000523, 0.00000156, 0.00000113, 0.00000596, -0.00001958, 0.00000798, 0.00001030, -0.00000896, 0.00000374, },
		{ -0.00000151, -0.00000175, 0.00000000, 0.00001635, -0.00001901, 0.00000185, -0.00000121, 0.00000597, -0.00000418, 0.00000288, -0.00000522, 0.00001477, -0.00001390, 0.00000658, -0.00000964, 0.00000739, -0.00001932, 0.00003948, -0.00004643, 0.00004972, -0.00004599, 0.00002455, -0.00002226, 0.00003344, -0.00003554, 0.00003683, -0.00003810, 0.00003972, -0.00004985, 0.00003809, -0.00003346, 0.00003793, -0.00003367, 0.00002347, -0.00002550, 0.00003755, -0.00004861, 0.00004991, -0.00005565, 0.00005809, -0.00006207, 0.00006075, -0.00006706, 0.00007005, -0.00007616, 0.00008219, -0.00008812, 0.00008797, -0.00010339, 0.00010707, -0.00012016, 0.00012364, -0.00014384, 0.00016122, -0.00019332, 0.00018075, -0.00022688, 0.00024233, -0.00031997, 0.00035684, -0.00054995, 0.00061050, -0.00153849, 0.00172089, 0.00078341, -0.00119659, 0.00052075, -0.00048903, 0.00032822, -0.00029701, 0.00022320, -0.00020935, 0.00016169, -0.00017710, 0.00015494, -0.00013970, 0.00011953, -0.00011497, 0.00010042, -0.00009923, 0.00008911, -0.00009076, 0.00008348, -0.00007706, 0.00006571, -0.00006096, 0.00005693, -0.00005858, 0.00005530, -0.00005176, 0.00003896, -0.00003616, 0.00002505, -0.00001467, 0.00001416, -0.00002718, 0.00003963, -0.00003383, 0.00003154, -0.00004438, 0.00004408, -0.00004285, 0.00003476, -0.00002972, 0.00003253, -0.00001893, 0.00001425, -0.00002920, 0.00003742, -0.00004286, 0.00004871, -0.00003894, 0.00002552, -0.00002540, 0.00001925, -0.00002465, 0.00002791, -0.00001924, 0.00001781, -0.00001632, 0.00000974, -0.00000258, 0.00000640, -0.00002423, 0.00002084, -0.00000198, -0.00000271, -0.00000030, },
		{ 0.00000179, -0.00000535, 0.00000823, 0.00000694, -0.00001635, 0.00000217, 0.00000067, 0.00000348, -0.00000005, -0.00000485, 0.00000699, 0.00000102, -0.00000284, -0.00000577, 0.00000720, -0.00001680, 0.00001593, 0.00000126, -0.00001369, 0.00002569, -0.00003401, 0.00001624, -0.00000716, 0.00001357, -0.00001714, 0.00001869, -0.00001772, 0.00001624, -0.00003093, 0.00002475, -0.00001720, 0.00001908, -0.00001683, 0.00000401, 0.00000162, 0.00000431, -0.00001425, 0.00001576, -0.00001900, 0.00001972, -0.00002276, 0.00001909, -0.00002295, 0.00002349, -0.00002514, 0.00002767, -0.00003028, 0.00002396, -0.00003452, 0.00003458, -0.00004250, 0.00003800, -0.00004760, 0.00005463, -0.00008587, 0.00006455, -0.00009297, 0.00008551, -0.00013694, 0.00012858, -0.00028917, 0.00023732, -0.00135089, 0.00260554, -0.00065758, -0.00033531, 0.00006877, -0.00015167, 0.00008208, -0.00009721, 0.00005756, -0.00006498, 0.00002836, -0.00005174, 0.00004847, -0.00004697, 0.00003543, -0.00003718, 0.00002709, -0.00002974, 0.00002459, -0.00003146, 0.00003090, -0.00003120, 0.00002298, -0.00001876, 0.00001564, -0.00001935, 0.00002047, -0.00002294, 0.00001225, -0.00001027, -0.00000020, 0.00001441, -0.00002359, 0.00001751, -0.00000174, 0.00000336, -0.00000914, -0.00000248, 0.00000868, -0.00001422, 0.00000964, -0.00000381, 0.00001021, 0.00000161, -0.00001380, 0.00000749, -0.00000098, -0.00000618, 0.00002001, -0.00002205, 0.00001293, -0.00001394, 0.00000737, -0.00001144, 0.00001886, -0.00001416, 0.00001504, -0.00001693, 0.00001103, -0.00000037, -0.00000088, -0.00001881, 0.00002521, -0.00001073, 0.00000307, -0.00000432, },
		{ 0.00000349, -0.00000648, 0.00001381, -0.00000257, -0.00001146, 0.00000104, 0.00000265, 0.00000292, -0.00000076, -0.00000572, 0.00001222, -0.00000740, 0.00000311, -0.00001043, 0.00001307, -0.00002745, 0.00004003, -0.00003473, 0.00002708, -0.00001433, -0.00000403, -0.00000550, 0.00001722, -0.00001621, 0.00001266, -0.00000979, 0.00001205, -0.00001958, 0.00000438, -0.00000410, 0.00001167, -0.00001265, 0.00001322, -0.00002662, 0.00003865, -0.00004040, 0.00003400, -0.00003275, 0.00003184, -0.00003383, 0.00003264, -0.00003952, 0.00003900, -0.00004079, 0.00004364, -0.00004555, 0.00004713, -0.00006170, 0.00005962, -0.00006629, 0.00006568, -0.00008059, 0.00008622, -0.00009909, 0.00007893, -0.00011582, 0.00011361, -0.00015867, 0.00015588, -0.00025335, 0.00019941, -0.00056846, -0.00028607, 0.00239825, -0.00142704, 0.00034921, -0.00036496, 0.00019625, -0.00018850, 0.00012744, -0.00013083, 0.00009751, -0.00011981, 0.00009111, -0.00007916, 0.00006635, -0.00006709, 0.00005670, -0.00006079, 0.00005454, -0.00005578, 0.00004449, -0.00003855, 0.00002973, -0.00003152, 0.00003309, -0.00003471, 0.00002922, -0.00002406, 0.00001354, -0.00001820, 0.00001578, -0.00002153, 0.00003476, -0.00005008, 0.00005380, -0.00003886, 0.00003614, -0.00004477, 0.00003841, -0.00003054, 0.00001899, -0.00001783, 0.00002351, -0.00001285, 0.00001828, -0.00003250, 0.00003452, -0.00003447, 0.00003319, -0.00001988, 0.00000978, -0.00001354, 0.00000951, -0.00001504, 0.00001308, -0.00000339, 0.00000465, -0.00000107, -0.00000654, 0.00000529, 0.00000624, -0.00001351, -0.00000443, 0.00001936, -0.00001253, 0.00000538, -0.00000633, },
		{ 0.00000378, -0.00000545, 0.00001605, -0.00001033, -0.00000607, -0.00000080, 0.00000534, 0.00000237, -0.00000439, 0.00000005, 0.00000815, -0.00000640, -0.00000038, -0.00000278, 0.00000127, -0.00001256, 0.00003235, -0.00003963, 0.00004279, -0.00003812, 0.00001632, -0.00001704, 0.00002752, -0.00003116, 0.00002898, -0.00002368, 0.00002394, -0.00003728, 0.00002525, -0.00001901, 0.00002319, -0.00002517, 0.00002321, -0.00003423, 0.00004938, -0.00005795, 0.00005628, -0.00005515, 0.00005472, -0.00005799, 0.00005715, -0.00006574, 0.00006745, -0.00006925, 0.00007329, -0.00007624, 0.00007750, -0.00009601, 0.00009976, -0.00011035, 0.00011237, -0.00013106, 0.00014445, -0.00017315, 0.00016127, -0.00020572, 0.00021723, -0.00028494, 0.00031014, -0.00047821, 0.00050654, -0.00127122, 0.00114375, 0.00127885, -0.00131224, 0.00050491, -0.00047177, 0.00030442, -0.00027605, 0.00020285, -0.00019277, 0.00014715, -0.00016336, 0.00014012, -0.00012568, 0.00010770, -0.00010391, 0.00008904, -0.00009013, 0.00008407, -0.00008643, 0.00007617, -0.00006930, 0.00005572, -0.00005203, 0.00005142, -0.00005306, 0.00004859, -0.00004347, 0.00002739, -0.00002572, 0.00001764, -0.00001615, 0.00002356, -0.00003997, 0.00005351, -0.00004347, 0.00003735, -0.00004710, 0.00004824, -0.00004460, 0.00003132, -0.00002494, 0.00003109, -0.00001807, 0.00001512, -0.00002493, 0.00002995, -0.00003488, 0.00004221, -0.00003710, 0.00002615, -0.00002785, 0.00002215, -0.00002680, 0.00002846, -0.00001982, 0.00002083, -0.00001779, 0.00000600, -0.00000074, 0.00001034, -0.00002358, 0.00001164, 0.00000766, -0.00000864, 0.00000408, -0.00000611, },
		{ 0.00000302, -0.00000301, 0.00001525, -0.00001559, -0.00000119, -0.00000282, 0.00000950, -0.00000150, -0.00000503, 0.00000534, 0.00000164, -0.00000238, -0.00000575, 0.00000705, -0.00001556, 0.00001507, 0.00000096, -0.00001308, 0.00002481, -0.00003216, 0.00001381, -0.00000816, 0.00001478, -0.00002148, 0.00002222, -0.00001491, 0.00000955, -0.00002569, 0.00001910, -0.00000889, 0.00000722, -0.00000743, 0.00000123, -0.00000640, 0.00001988, -0.00003218, 0.00003521, -0.00003427, 0.00003254, -0.00003498, 0.00003201, -0.00003968, 0.00004200, -0.00004164, 0.00004343, -0.00004335, 0.00003816, -0.00005315, 0.00005644, -0.00006519, 0.00006246, -0.00007449, 0.00008143, -0.00011243, 0.00009679, -0.00013163, 0.00013167, -0.00018785, 0.00018788, -0.00035177, 0.00032072, -0.00132444, 0.00217778, -0.00008731, -0.00064126, 0.00021483, -0.00025543, 0.00015527, -0.00015587, 0.00010452, -0.00010653, 0.00006559, -0.00008462, 0.00007551, -0.00007021, 0.00005733, -0.00005725, 0.00004389, -0.00004551, 0.00004267, -0.00005028, 0.00004630, -0.00004514, 0.00003308, -0.00002736, 0.00002630, -0.00002982, 0.00002932, -0.00002962, 0.00001368, -0.00000910, -0.00000264, 0.00001086, -0.00001215, 0.00000068, 0.00001945, -0.00001670, 0.00000865, -0.00001675, 0.00002420, -0.00002856, 0.00001882, -0.00000949, 0.00001772, -0.00000651, -0.00000246, 0.00000039, 0.00000134, -0.00000503, 0.00001664, -0.00002129, 0.00001431, -0.00001753, 0.00001237, -0.00001646, 0.00002223, -0.00001688, 0.00002060, -0.00002267, 0.00001130, -0.00000056, 0.00000607, -0.00002398, 0.00002129, -0.00000319, -0.00000288, 0.00000079, -0.00000418, },
		{ 0.00000071, 0.00000114, 0.00001108, -0.00001723, 0.00000148, -0.00000288, 0.00001293, -0.00000812, -0.00000015, 0.00000469, -0.00000031, -0.00000289, -0.00000387, 0.00000613, -0.00001863, 0.00003030, -0.00002599, 0.00001910, -0.00000712, -0.00000962, -0.00000147, 0.00001003, -0.00000856, 0.00000085, 0.00000433, 0.00000224, -0.00001465, -0.00000045, 0.00000026, 0.00001026, -0.00001749, 0.00002050, -0.00003182, 0.00003491, -0.00002773, 0.00001611, -0.00000950, 0.00000953, -0.00001282, 0.00001200, -0.00001847, 0.00001368, -0.00001135, 0.00001377, -0.00001464, 0.00001815, -0.00003221, 0.00002561, -0.00002726, 0.00002348, -0.00003423, 0.00003405, -0.00004345, 0.00002138, -0.00004983, 0.00003575, -0.00006420, 0.00004352, -0.00011257, 0.00002391, -0.00029820, -0.00057368, 0.00230186, -0.00100993, 0.00003351, -0.00017184, 0.00004823, -0.00007508, 0.00003360, -0.00005298, 0.00002870, -0.00005731, 0.00003435, -0.00002998, 0.00002333, -0.00002735, 0.00001984, -0.00002861, 0.00002578, -0.00002637, 0.00001409, -0.00001128, 0.00000394, -0.00000998, 0.00001478, -0.00001550, 0.00001011, -0.00000640, -0.00000235, -0.00000802, 0.00001064, -0.00002195, 0.00003345, -0.00004298, 0.00004132, -0.00002011, 0.00001560, -0.00002394, 0.00002038, -0.00001103, -0.00000100, -0.00000173, 0.00001197, -0.00000088, 0.00000607, -0.00001519, 0.00001874, -0.00002264, 0.00002569, -0.00001768, 0.00000797, -0.00001118, 0.00000569, -0.00000953, 0.00000685, 0.00000129, 0.00000151, 0.00000467, -0.00001415, 0.00000800, 0.00000544, -0.00000544, -0.00001421, 0.00002138, -0.00000908, 0.00000175, -0.00000308, -0.00000066, },
		{ -0.00000361, 0.00000749, 0.00000309, -0.00001424, 0.00000030, 0.00000171, 0.00001110, -0.00001208, 0.00000594, 0.00000026, 0.00000160, -0.00000718, 0.00000434, -0.00000604, -0.00000293, 0.00002003, -0.00002690, 0.00003023, -0.00002619, 0.00000655, -0.00001029, 0.00001865, -0.00002243, 0.00001628, -0.00000623, 0.00000775, -0.00002536, 0.00001418, -0.00000810, 0.00001433, -0.00002356, 0.00002819, -0.00004292, 0.00005383, -0.00005542, 0.00004867, -0.00004049, 0.00003860, -0.00004220, 0.00004181, -0.00005109, 0.00004972, -0.00004688, 0.00004924, -0.00004999, 0.00005274, -0.00007210, 0.00007298, -0.00007911, 0.00007806, -0.00009281, 0.00009967, -0.00012365, 0.00011072, -0.00014891, 0.00015006, -0.00020084, 0.00020588, -0.00033935, 0.00032094, -0.00093126, 0.00061842, 0.00155278, -0.00118798, 0.00033289, -0.00035250, 0.00020506, -0.00019717, 0.00013599, -0.00013728, 0.00009791, -0.00011732, 0.00009654, -0.00008725, 0.00007445, -0.00007335, 0.00006019, -0.00006501, 0.00006285, -0.00006521, 0.00005303, -0.00004802, 0.00003478, -0.00003432, 0.00003739, -0.00003863, 0.00003363, -0.00002906, 0.00001332, -0.00001608, 0.00001343, -0.00002045, 0.00003047, -0.00004444, 0.00005441, -0.00003795, 0.00002829, -0.00003516, 0.00003761, -0.00003188, 0.00001632, -0.00001163, 0.00002224, -0.00000990, 0.00000778, -0.00001127, 0.00001442, -0.00002040, 0.00003005, -0.00003091, 0.00002243, -0.00002528, 0.00001937, -0.00002258, 0.00002285, -0.00001526, 0.00001857, -0.00001373, -0.00000124, 0.00000268, 0.00001052, -0.00001680, -0.00000043, 0.00001484, -0.00001005, 0.00000424, -0.00000662, 0.00000418, },
		{ -0.00000826, 0.00001447, -0.00000711, -0.00000789, -0.00000362, 0.00001051, 0.00000203, -0.00000845, 0.00000669, -0.00000140, 0.00000108, -0.00000769, 0.00000853, -0.00001611, 0.00001705, -0.00000465, -0.00000537, 0.00001579, -0.00002252, 0.00000593, -0.00000445, 0.00001026, -0.00001832, 0.00001658, -0.00000353, -0.00000349, -0.00001522, 0.00000935, 0.00000143, -0.00000311, -0.00000302, 0.00000517, -0.00001889, 0.00003377, -0.00004308, 0.00004367, -0.00003640, 0.00003226, -0.00003463, 0.00003247, -0.00004244, 0.00004412, -0.00004042, 0.00004094, -0.00003864, 0.00003512, -0.00005290, 0.00005633, -0.00006324, 0.00005994, -0.00007114, 0.00007402, -0.00010183, 0.00008893, -0.00012434, 0.00012284, -0.00017219, 0.00016449, -0.00030913, 0.00025646, -0.00110077, 0.00161879, 0.00043087, -0.00080060, 0.00023786, -0.00026757, 0.00015536, -0.00015564, 0.00010295, -0.00010725, 0.00006667, -0.00008354, 0.00007128, -0.00006612, 0.00005510, -0.00005543, 0.00004120, -0.00004395, 0.00004384, -0.00005121, 0.00004389, -0.00004270, 0.00002914, -0.00002459, 0.00002658, -0.00002960, 0.00002755, -0.00002703, 0.00000937, -0.00000660, -0.00000151, 0.00000036, 0.00000430, -0.00001696, 0.00003654, -0.00002896, 0.00001744, -0.00002115, 0.00002850, -0.00003025, 0.00001671, -0.00000654, 0.00001817, -0.00000825, 0.00000202, 0.00000175, -0.00000424, 0.00000261, 0.00000755, -0.00001622, 0.00001193, -0.00001705, 0.00001269, -0.00001580, 0.00001953, -0.00001422, 0.00002042, -0.00002168, 0.00000626, 0.00000206, 0.00000863, -0.00002095, 0.00001014, 0.00000679, -0.00000813, 0.00000443, -0.00000839, 0.00000841, },
		{ -0.00000941, 0.00001806, -0.00001533, -0.00000223, -0.00000570, 0.00001807, -0.00000979, 0.00000114, 0.00000094, 0.00000230, -0.00000479, -0.00000024, 0.00000122, -0.00001112, 0.00002179, -0.00002016, 0.00001594, -0.00000654, -0.00000771, -0.00000318, 0.00000733, -0.00000571, -0.00000449, 0.00000944, 0.00000220, -0.00001814, 0.00000244, -0.00000260, 0.00001458, -0.00002443, 0.00002489, -0.00002924, 0.00002162, -0.00000860, -0.00000414, 0.00001196, -0.00000788, 0.00000237, -0.00000307, -0.00000241, -0.00000588, 0.00001012, -0.00000648, 0.00000570, -0.00000066, -0.00001145, 0.00000043, 0.00000105, -0.00000606, -0.00000202, -0.00000167, -0.00000753, -0.00001407, -0.00000746, -0.00001492, -0.00000364, -0.00002538, -0.00003010, -0.00007032, -0.00016368, -0.00062254, 0.00194545, -0.00047465, -0.00027142, -0.00001594, -0.00006532, 0.00000698, -0.00003340, 0.00000232, -0.00002154, -0.00001095, -0.00000640, 0.00000341, -0.00000564, 0.00000014, -0.00000592, -0.00000593, 0.00000457, -0.00000393, -0.00000789, 0.00000634, -0.00001205, 0.00000282, 0.00000305, -0.00000180, -0.00000294, 0.00000424, -0.00001092, -0.00000319, 0.00000717, -0.00001831, 0.00002190, -0.00002391, 0.00001812, 0.00000589, -0.00000832, -0.00000163, 0.00000224, 0.00000660, -0.00001613, 0.00000896, 0.00000431, 0.00000852, -0.00000446, 0.00000029, 0.00000675, -0.00001491, 0.00002132, -0.00001807, 0.00000740, -0.00000919, 0.00000216, -0.00000468, 0.00000237, 0.00000335, 0.00000068, 0.00000755, -0.00001707, 0.00000637, 0.00000628, 0.00000025, -0.00001684, 0.00001455, 0.00000023, -0.00000473, 0.00000193, -0.00000634, 0.00000868, },
		{ -0.00000413, 0.00001470, -0.00001764, -0.00000110, -0.00000136, 0.00001750, -0.00001592, 0.00000882, -0.00000522, 0.00000643, -0.00001073, 0.00000912, -0.00001204, 0.00000663, 0.00000661, -0.00001354, 0.00001825, -0.00001669, 0.00000094, -0.00000666, 0.00001135, -0.00001475, 0.00000564, 0.00000635, -0.00000159, -0.00002000, 0.00000971, -0.00000432, 0.00001256, -0.00002672, 0.00003314, -0.00004496, 0.00004630, -0.00004015, 0.00002970, -0.00001724, 0.00001683, -0.00002202, 0.00002189, -0.00003067, 0.00002578, -0.00001955, 0.00002156, -0.00002110, 0.00002500, -0.00004292, 0.00003923, -0.00004083, 0.00003696, -0.00004741, 0.00004780, -0.00006813, 0.00005435, -0.00008452, 0.00007394, -0.00010815, 0.00009456, -0.00019783, 0.00014051, -0.00063639, 0.00026573, 0.00154717, -0.00084712, 0.00006200, -0.00018537, 0.00007555, -0.00009843, 0.00005437, -0.00006959, 0.00003718, -0.00006035, 0.00004427, -0.00004257, 0.00003580, -0.00003760, 0.00002666, -0.00003506, 0.00003555, -0.00003734, 0.00002483, -0.00002390, 0.00001273, -0.00001611, 0.00002180, -0.00002189, 0.00001715, -0.00001501, 0.00000180, -0.00000951, 0.00001092, -0.00002199, 0.00002751, -0.00003460, 0.00003907, -0.00001716, 0.00000676, -0.00001242, 0.00001724, -0.00001115, -0.00000309, 0.00000392, 0.00001097, 0.00000137, -0.00000343, 0.00000641, -0.00000297, -0.00000534, 0.00001700, -0.00002368, 0.00001661, -0.00001955, 0.00001292, -0.00001461, 0.00001435, -0.00000889, 0.00001432, -0.00000812, -0.00000767, 0.00000442, 0.00000979, -0.00000864, -0.00000904, 0.00001461, -0.00000465, -0.00000043, -0.00000303, 0.00000015, 0.00000305, },
		{ 0.00000615, 0.00000495, -0.00001402, -0.00000402, 0.00000883, 0.00000714, -0.00001154, 0.00000797, -0.00000496, 0.00000432, -0.00000880, 0.00000956, -0.00001693, 0.00002049, -0.00001344, 0.00000612, 0.00000319, -0.00001060, -0.00000210, -0.00000032, 0.00000378, -0.00001185, 0.00000751, 0.00000887, -0.00001483, -0.00000709, 0.00000266, 0.00000709, -0.00000690, -0.00000528, 0.00001306, -0.00002922, 0.00003836, -0.00004078, 0.00003730, -0.00002423, 0.00001960, -0.00002301, 0.00002154, -0.00003214, 0.00003156, -0.00002457, 0.00002442, -0.00002081, 0.00001893, -0.00003703, 0.00003764, -0.00004090, 0.00003567, -0.00004394, 0.00004065, -0.00006407, 0.00005193, -0.00008313, 0.00007384, -0.00010978, 0.00008618, -0.00020170, 0.00011028, -0.00079483, 0.00108692, 0.00076060, -0.00074765, 0.00012257, -0.00018965, 0.00008875, -0.00010430, 0.00006023, -0.00007335, 0.00003682, -0.00005382, 0.00004172, -0.00004055, 0.00003392, -0.00003629, 0.00002300, -0.00002806, 0.00003021, -0.00003658, 0.00002704, -0.00002815, 0.00001564, -0.00001422, 0.00001925, -0.00002135, 0.00001845, -0.00001924, 0.00000329, -0.00000521, 0.00000243, -0.00001140, 0.00001609, -0.00002459, 0.00003938, -0.00002527, 0.00001129, -0.00001134, 0.00001874, -0.00001838, 0.00000417, 0.00000393, 0.00001196, -0.00000329, -0.00000081, 0.00001100, -0.00001568, 0.00001405, -0.00000485, -0.00000831, 0.00000614, -0.00001213, 0.00000776, -0.00000929, 0.00001131, -0.00000698, 0.00001534, -0.00001563, -0.00000154, 0.00000557, 0.00000802, -0.00001280, -0.00000238, 0.00001302, -0.00000811, 0.00000342, -0.00000804, 0.00000820, -0.00000615, },
		{ 0.00001568, -0.00000619, -0.00000880, -0.00000587, 0.00001810, -0.00000659, -0.00000058, -0.00000001, 0.00000152, -0.00000393, 0.00000153, -0.00000201, -0.00000549, 0.00001585, -0.00001855, 0.00001814, -0.00001201, -0.00000023, -0.00000830, 0.00000754, -0.00000696, -0.00000386, 0.00000717, 0.00000824, -0.00002497, 0.00000753, -0.00000663, 0.00001787, -0.00002687, 0.00002186, -0.00001854, 0.00000361, 0.00000873, -0.00001739, 0.00002265, -0.00001291, 0.00000581, -0.00000718, 0.00000281, -0.00001303, 0.00001690, -0.00001106, 0.00001024, -0.00000446, -0.00000500, -0.00000874, 0.00001033, -0.00001413, 0.00000673, -0.00001119, -0.00000091, -0.00001804, 0.00000093, -0.00002616, 0.00000930, -0.00003659, -0.00001925, -0.00007331, -0.00016003, -0.00048754, 0.00145656, 0.00002595, -0.00045183, 0.00003522, -0.00010159, 0.00002772, -0.00005131, 0.00001730, -0.00003778, 0.00000519, -0.00001881, 0.00001095, -0.00001225, 0.00000793, -0.00001381, 0.00000099, -0.00000273, 0.00000462, -0.00001481, 0.00000898, -0.00001449, 0.00000435, 0.00000010, 0.00000378, -0.00000747, 0.00000650, -0.00001254, -0.00000189, 0.00000309, -0.00000974, 0.00000340, -0.00000104, -0.00000454, 0.00002643, -0.00002315, 0.00001079, -0.00000563, 0.00001297, -0.00001887, 0.00000892, 0.00000470, 0.00001116, -0.00000855, 0.00000877, 0.00000304, -0.00001420, 0.00002160, -0.00002193, 0.00000921, -0.00000946, 0.00000153, -0.00000362, 0.00000269, 0.00000034, 0.00000425, 0.00000543, -0.00001394, 0.00000028, 0.00000911, 0.00000217, -0.00001225, 0.00000216, 0.00001023, -0.00000915, 0.00000445, -0.00000922, 0.00001246, -0.00001294, },
		{ 0.00001872, -0.00001329, -0.00000640, -0.00000171, 0.00001884, -0.00001401, 0.00000745, -0.00000692, 0.00000685, -0.00001021, 0.00001077, -0.00001547, 0.00001392, -0.00000443, -0.00000377, 0.00001106, -0.00001316, 0.00000164, -0.00000745, 0.00000731, -0.00001079, 0.00000057, 0.00001084, -0.00000260, -0.00002094, 0.00001048, -0.00000511, 0.00001352, -0.00002861, 0.00003189, -0.00003657, 0.00002775, -0.00001765, 0.00000781, 0.00000463, -0.00000066, -0.00000655, 0.00000613, -0.00001377, 0.00000618, 0.00000123, 0.00000187, -0.00000086, 0.00000518, -0.00002002, 0.00001187, -0.00001080, 0.00000592, -0.00001341, 0.00000931, -0.00002851, 0.00001528, -0.00003827, 0.00001898, -0.00004244, 0.00001814, -0.00010324, 0.00002511, -0.00045306, 0.00012193, 0.00130454, -0.00039463, -0.00020290, -0.00004816, -0.00002651, -0.00002451, -0.00000636, -0.00001917, -0.00000935, -0.00001590, 0.00000458, -0.00001015, 0.00000775, -0.00001095, 0.00000134, -0.00001191, 0.00001330, -0.00001425, 0.00000300, -0.00000731, -0.00000172, -0.00000437, 0.00001054, -0.00000863, 0.00000478, -0.00000585, -0.00000480, -0.00000572, 0.00000743, -0.00001699, 0.00001299, -0.00001299, 0.00001382, 0.00000967, -0.00001658, 0.00001016, -0.00000179, 0.00000655, -0.00001680, 0.00001359, 0.00000386, 0.00000963, -0.00001207, 0.00002093, -0.00001508, 0.00000436, 0.00000747, -0.00001861, 0.00001142, -0.00001364, 0.00000637, -0.00000716, 0.00000750, -0.00000488, 0.00001199, -0.00000509, -0.00000958, 0.00000215, 0.00001040, -0.00000286, -0.00001070, 0.00000653, 0.00000520, -0.00000669, 0.00000133, -0.00000444, 0.00000921, -0.00001219, },
		{ 0.00001404, -0.00001495, -0.00000686, 0.00000769, 0.00000983, -0.00001080, 0.00000637, -0.00000623, 0.00000460, -0.00000699, 0.00000844, -0.00001632, 0.00002308, -0.00002176, 0.00001645, -0.00000731, -0.00000212, -0.00000559, 0.00000063, -0.00000115, -0.00000645, 0.00000074, 0.00001595, -0.00001909, -0.00000521, 0.00000182, 0.00000665, -0.00000506, -0.00001021, 0.00001841, -0.00003052, 0.00002953, -0.00002515, 0.00001939, -0.00000372, 0.00000167, -0.00000714, 0.00000593, -0.00001592, 0.00001275, -0.00000384, 0.00000433, -0.00000028, -0.00000051, -0.00001553, 0.00001169, -0.00001138, 0.00000459, -0.00000975, 0.00000056, -0.00002049, 0.00000871, -0.00003358, 0.00001501, -0.00003780, -0.00000071, -0.00008865, -0.00003742, -0.00051777, 0.00069813, 0.00084472, -0.00050231, -0.00007933, -0.00007012, -0.00000436, -0.00003518, 0.00000383, -0.00002776, -0.00000397, -0.00001401, 0.00000424, -0.00000911, 0.00000739, -0.00001190, 0.00000006, -0.00000726, 0.00001047, -0.00001542, 0.00000535, -0.00001080, 0.00000083, -0.00000292, 0.00000976, -0.00001023, 0.00000733, -0.00001103, -0.00000150, -0.00000509, 0.00000574, -0.00001756, 0.00001571, -0.00001698, 0.00002586, -0.00000677, -0.00000629, 0.00000777, 0.00000096, -0.00000018, -0.00001128, 0.00001552, 0.00000410, 0.00000407, -0.00000653, 0.00002276, -0.00002697, 0.00002365, -0.00001602, -0.00000076, -0.00000076, -0.00000491, 0.00000000, -0.00000005, 0.00000130, 0.00000111, 0.00000903, -0.00000876, -0.00000774, 0.00000679, 0.00000706, -0.00000423, -0.00001045, 0.00001159, -0.00000172, -0.00000184, -0.00000405, 0.00000400, -0.00000004, -0.00000395, },
		{ 0.00000535, -0.00001405, -0.00000592, 0.00001566, -0.00000206, -0.00000185, -0.00000101, 0.00000016, -0.00000308, 0.00000360, -0.00000502, -0.00000081, 0.00001164, -0.00001942, 0.00002263, -0.00001918, 0.00000769, -0.00001215, 0.00000710, -0.00000924, -0.00000085, 0.00000257, 0.00001388, -0.00002860, 0.00000900, -0.00000665, 0.00001576, -0.00002242, 0.00001250, -0.00000495, -0.00001076, 0.00001584, -0.00001729, 0.00001882, -0.00000452, -0.00000185, -0.00000134, -0.00000215, -0.00000836, 0.00001038, -0.00000258, 0.00000246, 0.00000315, -0.00001006, -0.00000392, 0.00000283, -0.00000392, -0.00000386, 0.00000031, -0.00001677, 0.00000115, -0.00001575, -0.00000773, -0.00001310, -0.00000685, -0.00005633, -0.00001978, -0.00022974, -0.00027286, 0.00098190, 0.00036484, -0.00044696, -0.00003241, -0.00006166, -0.00000769, -0.00002634, -0.00000245, -0.00002441, -0.00000531, -0.00000677, -0.00000282, -0.00000090, -0.00000026, -0.00000695, -0.00000528, 0.00000219, 0.00000038, -0.00000827, -0.00000061, -0.00000686, -0.00000188, 0.00000328, 0.00000295, -0.00000511, 0.00000264, -0.00000994, -0.00000188, -0.00000186, 0.00000050, -0.00001402, 0.00001446, -0.00001533, 0.00003151, -0.00002128, 0.00000806, 0.00000032, 0.00000620, -0.00000926, -0.00000030, 0.00001162, 0.00000774, -0.00000622, 0.00001017, 0.00000671, -0.00001841, 0.00002399, -0.00002710, 0.00001219, -0.00001176, 0.00000410, -0.00000664, 0.00000755, -0.00000668, 0.00001089, -0.00000020, -0.00000709, -0.00000718, 0.00001164, 0.00000198, -0.00000406, -0.00000974, 0.00001510, -0.00000768, 0.00000211, -0.00000738, 0.00001048, -0.00000930, 0.00000664, },
		{ -0.00000274, -0.00001357, -0.00000011, 0.00001577, -0.00000800, 0.00000376, -0.00000634, 0.00000438, -0.00000730, 0.00001054, -0.00001682, 0.00001845, -0.00001182, 0.00000126, 0.00000889, -0.00001455, 0.00000634, -0.00001012, 0.00000485, -0.00000969, 0.00000025, 0.00000930, 0.00000080, -0.00002328, 0.00001145, -0.00000520, 0.00001166, -0.00002443, 0.00002221, -0.00001991, 0.00000520, 0.00000239, -0.00000694, 0.00001596, -0.00000654, -0.00000153, -0.00000019, -0.00000616, -0.00000262, 0.00000912, -0.00000418, 0.00000582, -0.00000198, -0.00000914, -0.00000176, 0.00000260, -0.00000633, 0.00000010, -0.00000633, -0.00001409, 0.00000270, -0.00002048, -0.00000261, -0.00001735, -0.00000918, -0.00006932, -0.00001363, -0.00038086, 0.00014084, 0.00093971, 0.00003035, -0.00035924, -0.00000069, -0.00006361, -0.00000153, -0.00002677, -0.00000178, -0.00002774, 0.00000320, -0.00001140, 0.00000095, -0.00000212, -0.00000029, -0.00000935, -0.00000164, 0.00000237, -0.00000264, -0.00000624, -0.00000301, -0.00000534, -0.00000129, 0.00000568, -0.00000141, -0.00000111, -0.00000283, -0.00000674, -0.00000341, 0.00000182, -0.00000643, -0.00000748, 0.00001023, -0.00000883, 0.00002922, -0.00002917, 0.00002146, -0.00000955, 0.00001258, -0.00001821, 0.00001295, 0.00000442, 0.00001127, -0.00001398, 0.00002785, -0.00001821, 0.00000634, 0.00000274, -0.00001658, 0.00000814, -0.00000963, 0.00000244, -0.00000326, 0.00000509, -0.00000529, 0.00001325, -0.00000621, -0.00000636, -0.00000354, 0.00001209, -0.00000014, -0.00000631, -0.00000528, 0.00001394, -0.00000967, 0.00000248, -0.00000566, 0.00001058, -0.00001288, 0.00001377, },
		{ -0.00000873, -0.00001277, 0.00000853, 0.00000830, -0.00000539, 0.00000164, -0.00000478, 0.00000162, -0.00000231, 0.00000516, -0.00001374, 0.00002332, -0.00002684, 0.00002217, -0.00001241, 0.00000045, -0.00000349, -0.00000184, -0.00000377, -0.00000385, -0.00000141, 0.00001600, -0.00001635, -0.00000799, 0.00000433, 0.00000361, -0.00000257, -0.00001179, 0.00001578, -0.00002044, 0.00000989, -0.00000295, -0.00000092, 0.00001500, -0.00001191, 0.00000444, -0.00000621, -0.00000247, -0.00000253, 0.00001134, -0.00000901, 0.00001309, -0.00001365, 0.00000124, -0.00000927, 0.00001124, -0.00001792, 0.00001463, -0.00002782, 0.00000930, -0.00002006, 0.00000123, -0.00002648, 0.00001170, -0.00005844, -0.00001465, -0.00013201, -0.00033216, 0.00048893, 0.00071963, -0.00015897, -0.00027425, 0.00002284, -0.00007419, 0.00001355, -0.00003604, 0.00000457, -0.00003385, 0.00001581, -0.00002299, 0.00001254, -0.00001103, 0.00000569, -0.00001667, 0.00000846, -0.00000568, 0.00000216, -0.00001118, 0.00000054, -0.00000815, 0.00000381, 0.00000273, -0.00000094, -0.00000136, -0.00000568, -0.00000386, -0.00000506, 0.00000537, -0.00001488, 0.00000296, 0.00000156, 0.00000377, 0.00001628, -0.00002400, 0.00002493, -0.00001371, 0.00001357, -0.00002013, 0.00002084, 0.00000059, 0.00000828, -0.00000928, 0.00003064, -0.00003197, 0.00002644, -0.00002229, 0.00000368, -0.00000619, 0.00000161, -0.00000695, 0.00000767, -0.00000607, 0.00000608, 0.00000520, -0.00000487, -0.00000922, 0.00000403, 0.00000745, 0.00000168, -0.00001165, 0.00000329, 0.00000770, -0.00000745, -0.00000015, -0.00000004, 0.00000474, -0.00000967, 0.00001531, },
		{ -0.00001382, -0.00000833, 0.00001369, 0.00000026, 0.00000016, -0.00000411, 0.00000027, -0.00000414, 0.00000726, -0.00000877, 0.00000392, 0.00000757, -0.00001989, 0.00002529, -0.00002335, 0.00001149, -0.00001132, 0.00000370, -0.00001019, 0.00000130, 0.00000001, 0.00001458, -0.00002574, 0.00000510, -0.00000205, 0.00000902, -0.00001410, 0.00000250, 0.00000399, -0.00001456, 0.00000921, -0.00000440, 0.00000427, 0.00001116, -0.00001375, 0.00000804, -0.00001125, 0.00000169, -0.00000181, 0.00001026, -0.00000859, 0.00001336, -0.00001845, 0.00000653, -0.00001168, 0.00001300, -0.00002073, 0.00001815, -0.00003841, 0.00002512, -0.00003786, 0.00001839, -0.00004460, 0.00003164, -0.00010129, 0.00003928, -0.00030289, -0.00007767, 0.00062704, 0.00048998, -0.00027552, -0.00017198, 0.00001167, -0.00006442, 0.00001316, -0.00003394, -0.00000043, -0.00002591, 0.00001427, -0.00002343, 0.00001561, -0.00001346, 0.00000502, -0.00001618, 0.00001182, -0.00000960, 0.00000413, -0.00001427, 0.00000306, -0.00000926, 0.00000755, -0.00000040, 0.00000036, -0.00000349, -0.00000629, -0.00000192, -0.00000608, 0.00000748, -0.00002277, 0.00001554, -0.00001023, 0.00002035, -0.00000511, -0.00000520, 0.00001433, -0.00000698, 0.00000506, -0.00001108, 0.00001883, 0.00000318, -0.00000169, 0.00000885, 0.00001264, -0.00002214, 0.00002374, -0.00002991, 0.00001380, -0.00001417, 0.00000789, -0.00001130, 0.00001374, -0.00001392, 0.00001686, -0.00000563, -0.00000076, -0.00001190, 0.00001105, 0.00000222, 0.00000338, -0.00001612, 0.00001213, -0.00000043, -0.00000366, -0.00000263, 0.00000519, -0.00000262, -0.00000292, 0.00001293, },
		{ -0.00001827, 0.00000073, 0.00001119, -0.00000161, 0.00000117, -0.00000622, 0.00000167, -0.00000454, 0.00001031, -0.00001732, 0.00002103, -0.00001630, 0.00000318, 0.00000885, -0.00001686, 0.00001042, -0.00001069, 0.00000119, -0.00000906, 0.00000163, 0.00000615, 0.00000315, -0.00002197, 0.00000857, -0.00000137, 0.00000483, -0.00001421, 0.00000760, -0.00000229, -0.00001183, 0.00001128, -0.00000843, 0.00001341, -0.00000017, -0.00000613, 0.00000202, -0.00000735, -0.00000112, 0.00000545, 0.00000096, 0.00000197, 0.00000065, -0.00000835, -0.00000283, 0.00000093, -0.00000237, -0.00000342, -0.00000348, -0.00001909, 0.00000967, -0.00002388, 0.00000124, -0.00002272, -0.00000136, -0.00007636, -0.00000427, -0.00037203, 0.00023246, 0.00058272, 0.00031543, -0.00035543, -0.00005277, -0.00003843, -0.00002396, -0.00001275, -0.00001249, -0.00002166, -0.00000014, -0.00000685, -0.00000550, 0.00000263, -0.00000283, -0.00000759, -0.00000258, 0.00000188, -0.00000205, -0.00000404, -0.00000865, -0.00000112, -0.00000405, 0.00000542, 0.00000066, -0.00000140, -0.00000450, -0.00000572, -0.00000178, -0.00000445, 0.00000454, -0.00002437, 0.00002346, -0.00001806, 0.00003250, -0.00002466, 0.00001753, -0.00000379, 0.00000632, -0.00000850, 0.00000443, 0.00001053, 0.00000756, -0.00000985, 0.00002783, -0.00001394, 0.00000303, 0.00000036, -0.00001577, 0.00000595, -0.00000753, 0.00000144, -0.00000308, 0.00000659, -0.00000879, 0.00001650, -0.00000962, -0.00000074, -0.00000932, 0.00001248, 0.00000135, 0.00000030, -0.00001491, 0.00001646, -0.00000651, -0.00000099, -0.00000255, 0.00000668, -0.00000726, 0.00000361, 0.00000859, },
		{ -0.00001914, 0.00001005, 0.00000387, 0.00000240, -0.00000376, -0.00000255, -0.00000269, 0.00000323, 0.00000123, -0.00001001, 0.00002125, -0.00002804, 0.00002284, -0.00001259, -0.00000119, 0.00000136, -0.00000490, -0.00000591, -0.00000295, -0.00000023, 0.00001166, -0.00001091, -0.00001057, 0.00000553, 0.00000256, -0.00000393, -0.00000672, 0.00000488, -0.00000317, -0.00001205, 0.00001546, -0.00001356, 0.00002321, -0.00001455, 0.00000675, -0.00001023, 0.00000305, -0.00000859, 0.00001556, -0.00001098, 0.00001532, -0.00001647, 0.00000815, -0.00001856, 0.00001949, -0.00002427, 0.00002147, -0.00003651, 0.00001754, -0.00002670, 0.00001248, -0.00004009, 0.00002705, -0.00007496, 0.00000857, -0.00016279, -0.00023977, 0.00042103, 0.00048602, 0.00015991, -0.00037404, 0.00003984, -0.00009072, 0.00002133, -0.00004418, 0.00001119, -0.00004162, 0.00002523, -0.00003101, 0.00001723, -0.00001532, 0.00001089, -0.00002187, 0.00001393, -0.00001272, 0.00001025, -0.00001725, 0.00000206, -0.00000894, 0.00000444, -0.00000015, 0.00000429, -0.00000556, -0.00000416, -0.00000405, -0.00000426, 0.00000121, -0.00000596, -0.00001475, 0.00001971, -0.00001413, 0.00003089, -0.00003008, 0.00003011, -0.00001660, 0.00001564, -0.00001723, 0.00001676, 0.00000386, 0.00000669, -0.00000597, 0.00003143, -0.00002839, 0.00002135, -0.00002325, 0.00000461, -0.00000911, 0.00000555, -0.00001062, 0.00001118, -0.00000849, 0.00000638, 0.00000497, -0.00000489, -0.00000609, -0.00000142, 0.00000837, 0.00000489, -0.00000769, -0.00000717, 0.00001455, -0.00000891, 0.00000016, -0.00000058, 0.00000471, -0.00000835, 0.00000869, 0.00000212, },
		{ -0.00001308, 0.00001307, -0.00000072, 0.00000551, -0.00000885, 0.00000164, -0.00000682, 0.00001190, -0.00001279, 0.00000873, 0.00000280, -0.00001804, 0.00002376, -0.00002245, 0.00000999, -0.00000601, -0.00000182, -0.00001001, 0.00000181, 0.00000062, 0.00001012, -0.00001804, -0.00000125, 0.00000363, 0.00000199, -0.00000738, -0.00000214, 0.00000361, -0.00000671, -0.00000730, 0.00001378, -0.00001167, 0.00002434, -0.00002107, 0.00001340, -0.00001735, 0.00000943, -0.00001104, 0.00001850, -0.00001426, 0.00001812, -0.00002235, 0.00001389, -0.00002263, 0.00002446, -0.00003069, 0.00002812, -0.00005035, 0.00003757, -0.00004833, 0.00003301, -0.00006244, 0.00005213, -0.00012286, 0.00006702, -0.00033383, 0.00004384, 0.00042438, 0.00043428, -0.00002202, -0.00030156, 0.00006051, -0.00010115, 0.00003601, -0.00005275, 0.00001403, -0.00003890, 0.00002799, -0.00003619, 0.00002477, -0.00002099, 0.00001251, -0.00002294, 0.00001809, -0.00001729, 0.00001389, -0.00002366, 0.00000817, -0.00001237, 0.00000896, -0.00000288, 0.00000537, -0.00000851, -0.00000376, -0.00000178, -0.00000784, 0.00000823, -0.00002059, 0.00000374, 0.00000479, 0.00000185, 0.00001444, -0.00001756, 0.00002538, -0.00001596, 0.00001407, -0.00001494, 0.00001993, 0.00000262, 0.00000018, 0.00000992, 0.00001540, -0.00002058, 0.00001734, -0.00002798, 0.00001225, -0.00001476, 0.00001033, -0.00001437, 0.00001746, -0.00001753, 0.00001878, -0.00000774, 0.00000192, -0.00001155, 0.00000649, 0.00000384, 0.00000788, -0.00001574, 0.00000328, 0.00000846, -0.00000857, 0.00000113, 0.00000084, 0.00000195, -0.00000748, 0.00001273, -0.00000715, },
		{ -0.00000122, 0.00000772, 0.00000206, 0.00000195, -0.00000802, 0.00000023, -0.00000331, 0.00001097, -0.00001764, 0.00002244, -0.00001920, 0.00000442, 0.00000752, -0.00001634, 0.00001049, -0.00000700, -0.00000477, -0.00000761, 0.00000245, 0.00000458, 0.00000150, -0.00001557, 0.00000193, 0.00000522, -0.00000450, -0.00000256, -0.00000505, 0.00000834, -0.00001680, 0.00000636, 0.00000202, 0.00000128, 0.00001218, -0.00001380, 0.00000693, -0.00001198, 0.00000490, -0.00000289, 0.00000962, -0.00000460, 0.00000573, -0.00001096, 0.00000169, -0.00000739, 0.00000768, -0.00001258, 0.00000518, -0.00002924, 0.00002068, -0.00003210, 0.00001220, -0.00003647, 0.00001458, -0.00009104, 0.00001130, -0.00036913, 0.00031314, 0.00032402, 0.00041805, -0.00021205, -0.00016056, 0.00001067, -0.00006142, 0.00001154, -0.00003200, -0.00000691, -0.00001216, 0.00000531, -0.00001749, 0.00001159, -0.00000977, -0.00000108, -0.00000793, 0.00000586, -0.00000674, 0.00000321, -0.00001735, 0.00000458, -0.00000753, 0.00000615, 0.00000021, 0.00000127, -0.00000818, -0.00000361, -0.00000091, -0.00000860, 0.00001026, -0.00003029, 0.00002097, -0.00001170, 0.00002045, -0.00000700, 0.00000427, 0.00000879, -0.00000483, 0.00000430, -0.00000436, 0.00001607, 0.00000336, -0.00000400, 0.00002539, -0.00000742, -0.00000051, -0.00000369, -0.00001337, 0.00000303, -0.00000589, 0.00000167, -0.00000432, 0.00000894, -0.00001183, 0.00001828, -0.00001190, 0.00000338, -0.00001158, 0.00000936, 0.00000363, 0.00000546, -0.00001846, 0.00001140, 0.00000148, -0.00000675, 0.00000235, 0.00000065, -0.00000002, -0.00000550, 0.00001479, -0.00001696, },
		{ 0.00000964, -0.00000054, 0.00000888, -0.00000660, -0.00000198, -0.00000599, 0.00000738, -0.00000166, -0.00000634, 0.00001781, -0.00002642, 0.00002038, -0.00001088, -0.00000318, 0.00000478, -0.00000519, -0.00000972, -0.00000192, 0.00000167, 0.00000690, -0.00000737, -0.00000913, 0.00000281, 0.00000522, -0.00000961, 0.00000314, -0.00000909, 0.00001284, -0.00002649, 0.00002188, -0.00001318, 0.00001846, -0.00000641, 0.00000115, -0.00000733, 0.00000131, -0.00000617, 0.00001046, -0.00000419, 0.00000942, -0.00001168, 0.00000686, -0.00001723, 0.00001497, -0.00001717, 0.00001436, -0.00002993, 0.00000989, -0.00001776, 0.00000682, -0.00003306, 0.00001758, -0.00006295, -0.00000156, -0.00015398, -0.00020128, 0.00041878, 0.00025520, 0.00035960, -0.00033900, -0.00002263, -0.00005973, -0.00000472, -0.00002719, -0.00000266, -0.00003113, 0.00001737, -0.00002316, 0.00000867, -0.00000879, 0.00000617, -0.00001741, 0.00001054, -0.00001137, 0.00000928, -0.00001404, -0.00000393, -0.00000428, 0.00000132, -0.00000015, 0.00000581, -0.00000563, -0.00000564, -0.00000290, -0.00000305, -0.00000393, 0.00000277, -0.00002724, 0.00002640, -0.00001840, 0.00002975, -0.00002004, 0.00002125, -0.00000718, 0.00000698, -0.00000514, 0.00000670, 0.00001106, 0.00000131, 0.00000186, 0.00002682, -0.00001890, 0.00001167, -0.00002178, 0.00000394, -0.00001035, 0.00000699, -0.00001095, 0.00001071, -0.00000695, 0.00000373, 0.00000645, -0.00000668, -0.00000137, -0.00000613, 0.00000746, 0.00000730, -0.00000237, -0.00001439, 0.00001468, -0.00000445, -0.00000353, 0.00000244, 0.00000036, -0.00000210, -0.00000129, 0.00001196, -0.00002176, },
		{ 0.00001224, -0.00000333, 0.00001158, -0.00001261, 0.00000227, -0.00000920, 0.00001560, -0.00001566, 0.00001277, -0.00000234, -0.00001384, 0.00001839, -0.00001779, 0.00000561, 0.00000046, -0.00000638, -0.00001065, 0.00000192, 0.00000279, 0.00000358, -0.00001000, -0.00000537, 0.00000618, -0.00000139, -0.00000583, 0.00000054, -0.00000484, 0.00000719, -0.00002446, 0.00002699, -0.00001957, 0.00002729, -0.00001842, 0.00001093, -0.00001684, 0.00001070, -0.00001275, 0.00001796, -0.00001121, 0.00001465, -0.00001885, 0.00001408, -0.00002452, 0.00002441, -0.00002789, 0.00002446, -0.00004685, 0.00003311, -0.00004191, 0.00002985, -0.00005898, 0.00004634, -0.00011285, 0.00005591, -0.00031578, 0.00009126, 0.00034058, 0.00028571, 0.00020785, -0.00034808, 0.00004656, -0.00009583, 0.00002758, -0.00004818, 0.00001008, -0.00003605, 0.00002669, -0.00003461, 0.00002156, -0.00001861, 0.00001140, -0.00002152, 0.00001701, -0.00001829, 0.00001626, -0.00002469, 0.00000583, -0.00000979, 0.00000665, -0.00000328, 0.00000807, -0.00001072, -0.00000316, -0.00000161, -0.00000691, 0.00000381, -0.00001170, -0.00001202, 0.00001781, -0.00001154, 0.00002522, -0.00001848, 0.00002496, -0.00001369, 0.00001223, -0.00000835, 0.00001318, 0.00000744, -0.00000230, 0.00001543, 0.00001299, -0.00001286, 0.00000518, -0.00002207, 0.00000796, -0.00001309, 0.00001013, -0.00001403, 0.00001680, -0.00001608, 0.00001598, -0.00000618, 0.00000082, -0.00000712, -0.00000001, 0.00000542, 0.00001008, -0.00001104, -0.00000636, 0.00001361, -0.00000871, 0.00000104, 0.00000008, 0.00000179, -0.00000552, 0.00000583, 0.00000268, -0.00001720, },
		{ 0.00000572, 0.00000255, 0.00000552, -0.00001076, -0.00000053, -0.00000302, 0.00001173, -0.00001730, 0.00002275, -0.00002123, 0.00000645, 0.00000335, -0.00001196, 0.00000650, -0.00000031, -0.00001137, -0.00000621, 0.00000261, 0.00000491, -0.00000331, -0.00000678, -0.00000476, 0.00001052, -0.00001165, 0.00000503, -0.00001027, 0.00000836, -0.00001000, -0.00000745, 0.00001648, -0.00001119, 0.00002131, -0.00001670, 0.00000813, -0.00001404, 0.00000888, -0.00000857, 0.00001427, -0.00000687, 0.00000684, -0.00001074, 0.00000462, -0.00001339, 0.00001351, -0.00001640, 0.00000799, -0.00003202, 0.00002279, -0.00003171, 0.00001501, -0.00004044, 0.00001742, -0.00008894, 0.00000481, -0.00033713, 0.00034001, 0.00018831, 0.00037074, -0.00000564, -0.00025269, 0.00003532, -0.00008046, 0.00002130, -0.00004112, 0.00000064, -0.00001875, 0.00001206, -0.00002344, 0.00001505, -0.00001285, 0.00000268, -0.00001076, 0.00000788, -0.00001019, 0.00000850, -0.00002235, 0.00000626, -0.00000782, 0.00000520, -0.00000042, 0.00000426, -0.00001140, -0.00000175, -0.00000114, -0.00000885, 0.00000846, -0.00002454, 0.00000637, 0.00000261, 0.00000292, 0.00001241, -0.00000685, 0.00001756, -0.00001094, 0.00001114, -0.00000602, 0.00001566, 0.00000309, -0.00000068, 0.00002509, -0.00000346, -0.00000036, -0.00001179, -0.00000807, -0.00000159, -0.00000341, 0.00000110, -0.00000427, 0.00000897, -0.00001133, 0.00001607, -0.00001074, 0.00000371, -0.00000895, 0.00000281, 0.00000676, 0.00000794, -0.00001587, 0.00000180, 0.00000962, -0.00001042, 0.00000524, -0.00000347, 0.00000414, -0.00000892, 0.00001320, -0.00000972, -0.00000528, },
		{ -0.00000292, 0.00001169, -0.00000545, -0.00000381, -0.00000741, 0.00000933, -0.00000307, -0.00000318, 0.00001384, -0.00002339, 0.00001715, -0.00001026, -0.00000249, 0.00000433, -0.00000161, -0.00001530, -0.00000033, 0.00000281, 0.00000396, -0.00000750, -0.00000362, -0.00000245, 0.00000930, -0.00001610, 0.00001237, -0.00001892, 0.00002003, -0.00002800, 0.00001492, -0.00000211, 0.00000570, 0.00000600, -0.00000600, -0.00000288, -0.00000279, -0.00000056, 0.00000228, 0.00000449, 0.00000219, -0.00000551, 0.00000302, -0.00001129, 0.00000520, -0.00000576, 0.00000350, -0.00001931, -0.00000118, -0.00000655, -0.00000189, -0.00002095, 0.00000211, -0.00004525, -0.00001732, -0.00013964, -0.00017076, 0.00041429, 0.00010084, 0.00040921, -0.00019339, -0.00012242, -0.00001618, -0.00003788, -0.00000583, -0.00002070, -0.00001570, 0.00000393, -0.00001044, -0.00000304, -0.00000004, -0.00000128, -0.00000963, 0.00000393, -0.00000657, 0.00000409, -0.00000706, -0.00001156, 0.00000060, -0.00000165, 0.00000028, 0.00000528, -0.00000333, -0.00000842, -0.00000120, -0.00000253, -0.00000657, 0.00000554, -0.00002805, 0.00001790, -0.00000909, 0.00001568, 0.00000056, 0.00000552, 0.00000690, -0.00000439, 0.00000791, -0.00000182, 0.00001628, -0.00000309, 0.00000984, 0.00002149, -0.00000886, 0.00000249, -0.00002165, 0.00000411, -0.00001147, 0.00000728, -0.00000966, 0.00000864, -0.00000450, 0.00000152, 0.00000649, -0.00000712, 0.00000142, -0.00000704, 0.00000316, 0.00001029, 0.00000141, -0.00001562, 0.00000807, 0.00000357, -0.00000827, 0.00000607, -0.00000482, 0.00000426, -0.00000868, 0.00001565, -0.00001802, 0.00000599, },
		{ -0.00000498, 0.00001515, -0.00001279, 0.00000040, -0.00000958, 0.00001677, -0.00001628, 0.00001523, -0.00000701, -0.00000851, 0.00001134, -0.00001249, 0.00000204, 0.00000435, -0.00000728, -0.00001353, 0.00000284, 0.00000444, -0.00000170, -0.00000533, -0.00000463, 0.00000426, -0.00000068, -0.00000860, 0.00000738, -0.00001511, 0.00001832, -0.00003301, 0.00002815, -0.00001534, 0.00001879, -0.00000701, 0.00000274, -0.00001141, 0.00000660, -0.00000813, 0.00001071, -0.00000232, 0.00000609, -0.00001067, 0.00000881, -0.00001862, 0.00001489, -0.00001556, 0.00001131, -0.00003284, 0.00001781, -0.00002535, 0.00001564, -0.00004156, 0.00002373, -0.00008380, 0.00002259, -0.00027069, 0.00009806, 0.00031212, 0.00013784, 0.00033703, -0.00028407, -0.00002714, -0.00005757, -0.00000329, -0.00002767, -0.00000664, -0.00002158, 0.00001390, -0.00002237, 0.00000978, -0.00000955, 0.00000406, -0.00001378, 0.00000999, -0.00001344, 0.00001189, -0.00001880, -0.00000157, -0.00000414, 0.00000262, -0.00000241, 0.00000838, -0.00001009, -0.00000420, -0.00000135, -0.00000478, -0.00000179, -0.00000350, -0.00002142, 0.00001891, -0.00001227, 0.00002187, -0.00000542, 0.00001292, -0.00000142, 0.00000207, 0.00000490, 0.00000280, 0.00001422, -0.00000659, 0.00002331, 0.00000773, -0.00000240, -0.00000874, -0.00001555, 0.00000315, -0.00001030, 0.00000778, -0.00001076, 0.00001266, -0.00001122, 0.00001075, -0.00000329, -0.00000154, -0.00000191, -0.00000493, 0.00000465, 0.00001182, -0.00000566, -0.00001209, 0.00001200, -0.00000344, -0.00000241, 0.00000221, -0.00000184, 0.00000011, -0.00000283, 0.00001016, -0.00001696, 0.00000900, },
		{ 0.00000176, 0.00000914, -0.00001192, -0.00000242, -0.00000205, 0.00001108, -0.00001533, 0.00002149, -0.00002226, 0.00000913, -0.00000288, -0.00000555, 0.00000127, 0.00000588, -0.00001554, -0.00000666, 0.00000322, 0.00000554, -0.00000824, 0.00000018, -0.00000743, 0.00001115, -0.00001364, 0.00000618, -0.00000750, 0.00000051, 0.00000187, -0.00002065, 0.00002522, -0.00001557, 0.00002075, -0.00001081, 0.00000308, -0.00001102, 0.00000775, -0.00000818, 0.00001202, -0.00000269, 0.00000229, -0.00000576, 0.00000283, -0.00001272, 0.00001051, -0.00001068, 0.00000107, -0.00002377, 0.00001287, -0.00001982, 0.00000589, -0.00002982, 0.00000230, -0.00006582, -0.00002677, -0.00027409, 0.00031600, 0.00014458, 0.00025093, 0.00017349, -0.00027026, 0.00000757, -0.00006630, 0.00000725, -0.00003310, -0.00000489, -0.00001423, 0.00000803, -0.00001896, 0.00000973, -0.00000925, 0.00000074, -0.00000808, 0.00000495, -0.00000902, 0.00000820, -0.00002074, 0.00000235, -0.00000434, 0.00000226, -0.00000036, 0.00000600, -0.00001272, -0.00000086, -0.00000227, -0.00000581, 0.00000146, -0.00001293, -0.00001012, 0.00001278, -0.00000842, 0.00002258, -0.00000671, 0.00001560, -0.00000667, 0.00000821, 0.00000139, 0.00000880, 0.00000717, -0.00000082, 0.00002827, -0.00000335, 0.00000354, -0.00002268, -0.00000135, -0.00000696, -0.00000008, -0.00000110, -0.00000175, 0.00000557, -0.00000680, 0.00001011, -0.00000665, 0.00000098, -0.00000325, -0.00000447, 0.00000886, 0.00000893, -0.00001012, -0.00000721, 0.00001305, -0.00000892, 0.00000412, -0.00000385, 0.00000357, -0.00000625, 0.00000596, -0.00000043, -0.00000824, 0.00000344, },
		{ 0.00001114, -0.00000160, -0.00000646, -0.00000800, 0.00001002, -0.00000394, -0.00000006, 0.00000946, -0.00001972, 0.00001442, -0.00001181, 0.00000115, 0.00000080, 0.00000385, -0.00002041, 0.00000051, 0.00000353, 0.00000261, -0.00000980, 0.00000277, -0.00000635, 0.00001080, -0.00001904, 0.00001610, -0.00002021, 0.00001629, -0.00001843, 0.00000058, 0.00001137, -0.00000605, 0.00001414, -0.00000785, -0.00000220, -0.00000435, 0.00000280, -0.00000286, 0.00000897, -0.00000080, -0.00000298, 0.00000157, -0.00000674, -0.00000214, 0.00000123, -0.00000198, -0.00001418, -0.00000564, -0.00000320, -0.00000327, -0.00001569, -0.00000507, -0.00003840, -0.00002003, -0.00014078, -0.00012271, 0.00037587, 0.00003900, 0.00034225, -0.00000698, -0.00019779, 0.00000028, -0.00005176, 0.00000086, -0.00002829, -0.00000755, -0.00000403, -0.00000309, -0.00000890, 0.00000352, -0.00000515, -0.00000438, -0.00000040, -0.00000350, 0.00000003, -0.00000206, -0.00001547, 0.00000164, -0.00000183, -0.00000049, 0.00000422, -0.00000057, -0.00001077, 0.00000035, -0.00000338, -0.00000516, 0.00000102, -0.00001823, -0.00000009, 0.00000475, -0.00000097, 0.00002027, -0.00000587, 0.00001562, -0.00000945, 0.00001400, -0.00000265, 0.00001526, -0.00000297, 0.00001441, 0.00001938, -0.00000253, -0.00000283, -0.00002461, 0.00000591, -0.00001306, 0.00000752, -0.00000862, 0.00000740, -0.00000392, 0.00000222, 0.00000326, -0.00000465, 0.00000080, -0.00000371, -0.00000362, 0.00001324, 0.00000326, -0.00001185, -0.00000149, 0.00001028, -0.00000994, 0.00000696, -0.00000733, 0.00000678, -0.00000995, 0.00001187, -0.00000910, 0.00000056, -0.00000361, },
		{ 0.00001481, -0.00000896, -0.00000376, -0.00000800, 0.00001544, -0.00001492, 0.00001618, -0.00001072, -0.00000330, 0.00000478, -0.00000910, 0.00000155, 0.00000419, -0.00000442, -0.00001785, 0.00000420, 0.00000493, -0.00000441, -0.00000464, -0.00000014, 0.00000023, 0.00000109, -0.00001163, 0.00001271, -0.00002031, 0.00002017, -0.00002907, 0.00001741, -0.00000241, 0.00000448, 0.00000639, -0.00000497, -0.00000605, 0.00000158, -0.00000200, 0.00000237, 0.00000637, -0.00000175, -0.00000302, 0.00000282, -0.00001020, 0.00000270, -0.00000197, -0.00000220, -0.00001828, 0.00000261, -0.00001011, 0.00000217, -0.00002348, 0.00000040, -0.00005513, -0.00000967, -0.00022653, 0.00010083, 0.00028489, 0.00004863, 0.00034605, -0.00013745, -0.00012007, -0.00001800, -0.00003334, -0.00000874, -0.00002265, -0.00000652, -0.00000055, -0.00000873, -0.00000217, -0.00000078, -0.00000373, -0.00000485, 0.00000172, -0.00000692, 0.00000499, -0.00001080, -0.00000901, 0.00000039, -0.00000028, -0.00000178, 0.00000723, -0.00000765, -0.00000599, -0.00000089, -0.00000336, -0.00000463, -0.00000151, -0.00001933, 0.00000659, -0.00000252, 0.00000871, 0.00001489, -0.00000269, 0.00001265, -0.00000818, 0.00001672, -0.00000412, 0.00001761, -0.00000837, 0.00002956, 0.00000392, 0.00000574, -0.00001959, -0.00001190, -0.00000001, -0.00000784, 0.00000474, -0.00000651, 0.00000761, -0.00000616, 0.00000643, -0.00000205, -0.00000230, 0.00000094, -0.00000577, 0.00000059, 0.00001395, -0.00000174, -0.00001228, 0.00000478, 0.00000406, -0.00000585, 0.00000413, -0.00000539, 0.00000477, -0.00000791, 0.00001119, -0.00001081, 0.00000289, -0.00000446, },
		{ 0.00001015, -0.00000959, -0.00000645, 0.00000052, 0.00000823, -0.00001142, 0.00001854, -0.00002193, 0.00001152, -0.00000914, -0.00000002, -0.00000194, 0.00000850, -0.00001540, -0.00000959, 0.00000476, 0.00000514, -0.00001086, 0.00000279, -0.00000483, 0.00000713, -0.00001136, 0.00000299, -0.00000101, -0.00000829, 0.00001011, -0.00002497, 0.00002261, -0.00000919, 0.00001056, 0.00000161, -0.00000542, -0.00000509, 0.00000303, -0.00000335, 0.00000525, 0.00000492, -0.00000484, 0.00000142, -0.00000204, -0.00000680, 0.00000103, 0.00000098, -0.00001115, -0.00000969, -0.00000287, -0.00000314, -0.00000851, -0.00001194, -0.00002144, -0.00003386, -0.00006762, -0.00020026, 0.00027018, 0.00013938, 0.00013629, 0.00026465, -0.00019703, -0.00006378, -0.00003079, -0.00002050, -0.00001586, -0.00001774, -0.00000289, -0.00000307, -0.00000787, -0.00000058, -0.00000198, -0.00000475, -0.00000160, -0.00000155, -0.00000416, 0.00000319, -0.00001419, -0.00000467, 0.00000047, -0.00000101, -0.00000003, 0.00000614, -0.00001182, -0.00000113, -0.00000353, -0.00000191, -0.00000579, -0.00000347, -0.00001860, 0.00001061, -0.00000822, 0.00002006, 0.00000548, 0.00000442, 0.00000527, -0.00000075, 0.00001321, 0.00000021, 0.00001194, -0.00000166, 0.00003270, -0.00000471, 0.00000778, -0.00003242, 0.00000351, -0.00001098, 0.00000295, -0.00000419, 0.00000238, 0.00000012, -0.00000038, 0.00000313, -0.00000231, -0.00000204, 0.00000206, -0.00000902, 0.00000781, 0.00001012, -0.00000448, -0.00001177, 0.00000998, -0.00000277, 0.00000037, -0.00000191, 0.00000021, -0.00000092, -0.00000180, 0.00000568, -0.00000689, -0.00000091, 0.00000190, },
		{ 0.00000200, -0.00000740, -0.00000949, 0.00001090, -0.00000555, 0.00000371, 0.00000455, -0.00001554, 0.00001216, -0.00001459, 0.00000570, -0.00000287, 0.00000774, -0.00002224, -0.00000091, 0.00000474, 0.00000144, -0.00001154, 0.00000646, -0.00000580, 0.00000751, -0.00001671, 0.00001338, -0.00001397, 0.00000581, -0.00000540, -0.00001167, 0.00001835, -0.00000950, 0.00001286, -0.00000175, -0.00000663, -0.00000190, 0.00000184, -0.00000270, 0.00000748, 0.00000156, -0.00000491, 0.00000368, -0.00000594, -0.00000372, 0.00000047, 0.00000064, -0.00001684, -0.00000149, -0.00000907, 0.00000379, -0.00001914, -0.00000138, -0.00004463, -0.00000841, -0.00015713, -0.00005892, 0.00030939, 0.00004038, 0.00022670, 0.00014424, -0.00020302, -0.00002886, -0.00003812, -0.00001174, -0.00002216, -0.00000997, -0.00000300, -0.00000450, -0.00000635, 0.00000012, -0.00000374, -0.00000371, -0.00000025, -0.00000436, -0.00000037, -0.00000172, -0.00001377, -0.00000182, 0.00000068, -0.00000236, 0.00000329, 0.00000134, -0.00001152, 0.00000112, -0.00000510, -0.00000104, -0.00000735, -0.00000487, -0.00001668, 0.00001169, -0.00001003, 0.00002975, -0.00000549, 0.00001382, -0.00000462, 0.00001106, 0.00000461, 0.00000870, 0.00000117, 0.00001553, 0.00002104, -0.00000059, -0.00000466, -0.00002914, 0.00000772, -0.00001423, 0.00000772, -0.00000849, 0.00000785, -0.00000581, 0.00000576, -0.00000239, -0.00000036, -0.00000188, 0.00000151, -0.00000979, 0.00001401, 0.00000465, -0.00000623, -0.00000915, 0.00001168, -0.00000652, 0.00000414, -0.00000633, 0.00000488, -0.00000588, 0.00000372, 0.00000058, -0.00000323, -0.00000462, 0.00000946, },
		{ -0.00000322, -0.00000801, -0.00000608, 0.00001303, -0.00001291, 0.00001640, -0.00001358, 0.00000125, -0.00000033, -0.00000819, 0.00000378, 0.00000144, -0.00000038, -0.00002098, 0.00000434, 0.00000502, -0.00000538, -0.00000589, 0.00000463, -0.00000209, 0.00000027, -0.00001116, 0.00001260, -0.00001739, 0.00001225, -0.00001608, 0.00000170, 0.00001081, -0.00000704, 0.00001404, -0.00000645, -0.00000496, -0.00000048, 0.00000128, -0.00000221, 0.00001019, -0.00000473, 0.00000041, -0.00000014, -0.00000411, -0.00000578, 0.00000574, -0.00000883, -0.00001060, -0.00000479, -0.00000377, -0.00000291, -0.00001368, -0.00001280, -0.00003923, -0.00002712, -0.00019837, 0.00011744, 0.00023662, 0.00002650, 0.00026902, 0.00002877, -0.00017989, -0.00000848, -0.00004284, -0.00000512, -0.00002760, 0.00000034, -0.00000846, -0.00000137, -0.00000771, 0.00000240, -0.00000742, 0.00000081, -0.00000349, -0.00000283, -0.00000012, -0.00000535, -0.00001238, 0.00000098, -0.00000050, -0.00000221, 0.00000553, -0.00000476, -0.00000742, -0.00000041, -0.00000351, -0.00000351, -0.00000609, -0.00000846, -0.00001174, 0.00000794, -0.00000417, 0.00003171, -0.00001158, 0.00001965, -0.00001084, 0.00002070, -0.00000298, 0.00001469, -0.00000528, 0.00003176, 0.00000435, 0.00000836, -0.00002490, -0.00001216, -0.00000119, -0.00000613, 0.00000206, -0.00000304, 0.00000394, -0.00000334, 0.00000506, -0.00000377, -0.00000047, 0.00000064, -0.00000283, -0.00000517, 0.00001529, 0.00000096, -0.00000864, -0.00000384, 0.00000910, -0.00000573, 0.00000318, -0.00000615, 0.00000549, -0.00000713, 0.00000576, -0.00000092, -0.00000391, -0.00000284, 0.00001171, },
		{ -0.00000424, -0.00001158, 0.00000389, 0.00000396, -0.00000647, 0.00001455, -0.00002033, 0.00001290, -0.00001379, 0.00000276, -0.00000170, 0.00000705, -0.00001181, -0.00001339, 0.00000613, 0.00000388, -0.00001090, 0.00000164, 0.00000082, 0.00000167, -0.00000841, -0.00000014, 0.00000354, -0.00001174, 0.00000941, -0.00001817, 0.00001029, 0.00000359, -0.00000314, 0.00001375, -0.00001151, -0.00000075, -0.00000132, 0.00000163, -0.00000112, 0.00001088, -0.00000982, 0.00000665, -0.00000619, -0.00000011, -0.00000894, 0.00001137, -0.00002086, 0.00000181, -0.00001526, 0.00000889, -0.00001865, 0.00000237, -0.00004057, -0.00000884, -0.00009902, -0.00013788, 0.00023106, 0.00013062, 0.00007075, 0.00025845, -0.00006370, -0.00014039, -0.00000201, -0.00004223, -0.00000374, -0.00002716, 0.00000650, -0.00001330, 0.00000228, -0.00000922, 0.00000356, -0.00000937, 0.00000464, -0.00000781, 0.00000092, -0.00000290, -0.00000691, -0.00001060, 0.00000334, -0.00000267, -0.00000019, 0.00000504, -0.00000931, -0.00000222, -0.00000418, 0.00000046, -0.00000931, -0.00000101, -0.00001561, -0.00000306, -0.00000055, 0.00001028, 0.00002240, -0.00000764, 0.00001668, -0.00000777, 0.00002192, -0.00000383, 0.00001285, 0.00000040, 0.00003510, -0.00000416, 0.00000832, -0.00003717, 0.00000397, -0.00001218, 0.00000465, -0.00000689, 0.00000622, -0.00000485, 0.00000484, -0.00000177, -0.00000037, -0.00000284, 0.00000437, -0.00000906, 0.00000340, 0.00001169, -0.00000035, -0.00001132, 0.00000256, 0.00000428, -0.00000247, -0.00000072, -0.00000254, 0.00000280, -0.00000551, 0.00000559, -0.00000062, -0.00000711, 0.00000351, 0.00000795, },
		{ -0.00000527, -0.00001235, 0.00001240, -0.00000817, 0.00000799, -0.00000064, -0.00001090, 0.00000976, -0.00001690, 0.00000867, -0.00000400, 0.00000778, -0.00001976, -0.00000448, 0.00000629, -0.00000023, -0.00001122, 0.00000617, -0.00000088, 0.00000073, -0.00001115, 0.00000738, -0.00000534, -0.00000411, 0.00000280, -0.00001475, 0.00001471, -0.00000337, 0.00000342, 0.00000914, -0.00001284, 0.00000209, -0.00000155, 0.00000075, 0.00000269, 0.00000620, -0.00000834, 0.00000703, -0.00000732, -0.00000108, -0.00000543, 0.00000756, -0.00002280, 0.00000614, -0.00001814, 0.00001290, -0.00002508, 0.00000629, -0.00005521, 0.00000646, -0.00017462, 0.00000285, 0.00023940, 0.00006089, 0.00012551, 0.00021469, -0.00013198, -0.00008995, -0.00001091, -0.00003261, -0.00001050, -0.00001714, 0.00000274, -0.00001088, 0.00000077, -0.00000662, 0.00000015, -0.00000570, 0.00000281, -0.00000782, 0.00000191, -0.00000494, -0.00000837, -0.00000725, 0.00000361, -0.00000399, 0.00000233, 0.00000189, -0.00001027, 0.00000071, -0.00000648, 0.00000282, -0.00001420, 0.00000410, -0.00002314, 0.00000585, -0.00000870, 0.00002760, 0.00000573, 0.00000449, 0.00000616, 0.00000380, 0.00001464, 0.00000166, 0.00000545, 0.00001619, 0.00002400, -0.00000068, -0.00000609, -0.00003181, 0.00000690, -0.00001344, 0.00000714, -0.00000867, 0.00000911, -0.00000870, 0.00000986, -0.00000771, 0.00000307, -0.00000390, 0.00000531, -0.00001230, 0.00001102, 0.00000678, -0.00000143, -0.00001211, 0.00000724, 0.00000038, -0.00000024, -0.00000385, 0.00000107, -0.00000030, -0.00000366, 0.00000623, -0.00000263, -0.00000745, 0.00000863, 0.00000309, },
		{ -0.00000980, -0.00000547, 0.00001097, -0.00001147, 0.00001661, -0.00001583, 0.00000511, -0.00000419, -0.00000860, 0.00000652, -0.00000093, 0.00000160, -0.00002044, 0.00000205, 0.00000547, -0.00000592, -0.00000662, 0.00000675, -0.00000054, -0.00000432, -0.00000657, 0.00000733, -0.00000866, -0.00000011, -0.00000213, -0.00000994, 0.00001649, -0.00000979, 0.00001223, -0.00000042, -0.00000834, 0.00000090, 0.00000066, -0.00000190, 0.00000855, -0.00000280, -0.00000032, 0.00000035, -0.00000190, -0.00000831, 0.00000564, -0.00000740, -0.00001052, -0.00000401, -0.00000602, -0.00000017, -0.00001224, -0.00001540, -0.00003596, -0.00003090, -0.00018357, 0.00014361, 0.00017497, 0.00004440, 0.00016466, 0.00015230, -0.00017322, -0.00003892, -0.00002934, -0.00001726, -0.00002125, -0.00000192, -0.00000830, -0.00000184, -0.00000590, -0.00000044, -0.00000637, 0.00000207, -0.00000429, -0.00000259, -0.00000170, -0.00000422, -0.00001073, -0.00000220, 0.00000128, -0.00000327, 0.00000375, -0.00000254, -0.00000755, -0.00000045, -0.00000452, -0.00000033, -0.00001298, 0.00000338, -0.00002532, 0.00000921, -0.00000891, 0.00003788, -0.00000843, 0.00001644, -0.00000423, 0.00001656, 0.00000508, 0.00000741, 0.00000113, 0.00003095, 0.00000839, 0.00000587, -0.00002580, -0.00001442, -0.00000210, -0.00000433, -0.00000021, -0.00000092, 0.00000224, -0.00000302, 0.00000615, -0.00000721, 0.00000251, -0.00000112, 0.00000119, -0.00000946, 0.00001380, 0.00000385, -0.00000419, -0.00000964, 0.00000874, -0.00000118, -0.00000061, -0.00000419, 0.00000265, -0.00000227, -0.00000235, 0.00000805, -0.00000798, -0.00000228, 0.00000832, 0.00000136, },
		{ -0.00001530, 0.00000621, -0.00000009, -0.00000203, 0.00001063, -0.00001800, 0.00001317, -0.00001644, 0.00000287, 0.00000093, 0.00000361, -0.00000778, -0.00001520, 0.00000551, 0.00000286, -0.00000969, -0.00000074, 0.00000591, -0.00000158, -0.00000805, -0.00000002, 0.00000309, -0.00000802, 0.00000075, -0.00000490, -0.00000436, 0.00001461, -0.00001246, 0.00001892, -0.00001067, -0.00000099, -0.00000249, 0.00000350, -0.00000355, 0.00001196, -0.00000997, 0.00000759, -0.00000707, 0.00000368, -0.00001444, 0.00001510, -0.00002286, 0.00000558, -0.00001904, 0.00001132, -0.00001966, 0.00000724, -0.00004787, 0.00000118, -0.00011221, -0.00009639, 0.00020970, 0.00010412, 0.00005628, 0.00018903, 0.00007361, -0.00017899, -0.00000354, -0.00004419, -0.00000575, -0.00002654, 0.00000816, -0.00001702, 0.00000617, -0.00001174, 0.00000395, -0.00001012, 0.00000752, -0.00001070, 0.00000314, -0.00000672, -0.00000253, -0.00001241, 0.00000251, -0.00000206, -0.00000106, 0.00000328, -0.00000621, -0.00000332, -0.00000419, 0.00000060, -0.00000850, -0.00000477, -0.00000547, -0.00001951, 0.00000477, 0.00000229, 0.00003458, -0.00001176, 0.00002025, -0.00000680, 0.00002281, -0.00000030, 0.00000836, 0.00000656, 0.00003428, -0.00000035, 0.00000355, -0.00003593, 0.00000011, -0.00001104, 0.00000507, -0.00000850, 0.00000839, -0.00000754, 0.00000700, -0.00000320, -0.00000152, -0.00000108, 0.00000359, -0.00000564, -0.00000216, 0.00001205, 0.00000303, -0.00000808, -0.00000490, 0.00000796, -0.00000159, -0.00000218, -0.00000276, 0.00000286, -0.00000403, 0.00000031, 0.00000775, -0.00001262, 0.00000512, 0.00000425, 0.00000155, },
		{ -0.00001516, 0.00001342, -0.00001082, 0.00001172, -0.00000527, -0.00000618, 0.00000691, -0.00001765, 0.00000890, -0.00000200, 0.00000434, -0.00001460, -0.00000804, 0.00000668, -0.00000172, -0.00000935, 0.00000339, 0.00000543, -0.00000588, -0.00000624, 0.00000255, 0.00000031, -0.00000797, 0.00000221, -0.00000802, 0.00000333, 0.00000727, -0.00000783, 0.00001819, -0.00001548, 0.00000371, -0.00000396, 0.00000375, -0.00000119, 0.00000895, -0.00000958, 0.00000843, -0.00000793, 0.00000201, -0.00001122, 0.00001256, -0.00002558, 0.00001042, -0.00002308, 0.00001669, -0.00002616, 0.00000956, -0.00005994, 0.00001263, -0.00017940, 0.00004642, 0.00018647, 0.00006958, 0.00006982, 0.00020398, -0.00001735, -0.00014713, 0.00000532, -0.00004527, -0.00000487, -0.00002061, 0.00000628, -0.00001585, 0.00000637, -0.00001142, 0.00000241, -0.00000698, 0.00000559, -0.00001079, 0.00000397, -0.00000861, -0.00000294, -0.00001111, 0.00000443, -0.00000414, 0.00000109, 0.00000118, -0.00000743, -0.00000072, -0.00000664, 0.00000412, -0.00001613, 0.00000510, -0.00001820, -0.00000931, -0.00000287, 0.00002002, 0.00002017, -0.00000409, 0.00001498, -0.00000033, 0.00002070, -0.00000043, 0.00000604, 0.00001945, 0.00002551, -0.00000016, -0.00001007, -0.00003005, 0.00000210, -0.00001020, 0.00000537, -0.00000834, 0.00000967, -0.00001047, 0.00001196, -0.00001018, 0.00000361, -0.00000336, 0.00000582, -0.00001025, 0.00000494, 0.00000908, 0.00000204, -0.00001083, -0.00000006, 0.00000669, -0.00000269, -0.00000273, -0.00000158, 0.00000306, -0.00000657, 0.00000584, 0.00000217, -0.00001144, 0.00000877, 0.00000163, -0.00000107, },
		{ -0.00000625, 0.00000937, -0.00001071, 0.00001666, -0.00001730, 0.00000828, -0.00000736, -0.00000883, 0.00000765, -0.00000081, -0.00000005, -0.00001595, -0.00000198, 0.00000579, -0.00000635, -0.00000597, 0.00000585, 0.00000385, -0.00001057, -0.00000059, 0.00000062, 0.00000039, -0.00000980, 0.00000536, -0.00001193, 0.00001214, -0.00000388, 0.00000333, 0.00000918, -0.00001242, 0.00000299, -0.00000166, 0.00000093, 0.00000438, 0.00000087, -0.00000231, 0.00000201, -0.00000193, -0.00000672, 0.00000080, -0.00000213, -0.00001331, -0.00000024, -0.00001095, 0.00000421, -0.00001313, -0.00001404, -0.00003704, -0.00003113, -0.00016991, 0.00016452, 0.00011971, 0.00006722, 0.00008221, 0.00020138, -0.00010278, -0.00009176, -0.00000828, -0.00003363, -0.00001197, -0.00000741, -0.00000451, -0.00000618, -0.00000070, -0.00000572, -0.00000327, 0.00000082, -0.00000228, -0.00000477, -0.00000091, -0.00000610, -0.00000619, -0.00000660, 0.00000296, -0.00000389, 0.00000195, -0.00000156, -0.00000599, -0.00000147, -0.00000501, 0.00000180, -0.00001701, 0.00000907, -0.00002739, -0.00000158, -0.00000555, 0.00003416, 0.00000412, 0.00000737, 0.00000663, 0.00000980, 0.00001429, 0.00000103, 0.00000679, 0.00003050, 0.00001303, 0.00000104, -0.00002543, -0.00001557, -0.00000486, -0.00000145, -0.00000248, 0.00000031, 0.00000163, -0.00000370, 0.00000753, -0.00000985, 0.00000421, -0.00000183, 0.00000336, -0.00000980, 0.00000861, 0.00000746, -0.00000075, -0.00001099, 0.00000364, 0.00000533, -0.00000455, -0.00000189, -0.00000097, 0.00000279, -0.00000833, 0.00001202, -0.00000724, -0.00000388, 0.00000631, 0.00000295, -0.00000829, },
		{ 0.00000620, -0.00000283, 0.00000090, 0.00000741, -0.00001535, 0.00001263, -0.00001759, 0.00000173, 0.00000364, 0.00000106, -0.00000622, -0.00001305, 0.00000213, 0.00000261, -0.00000848, -0.00000220, 0.00000811, -0.00000128, -0.00001064, 0.00000314, -0.00000121, -0.00000029, -0.00001001, 0.00000687, -0.00001321, 0.00001743, -0.00001292, 0.00001491, -0.00000301, -0.00000491, -0.00000090, 0.00000215, -0.00000193, 0.00000860, -0.00000617, 0.00000512, -0.00000500, 0.00000390, -0.00001432, 0.00001166, -0.00001762, 0.00000254, -0.00001549, 0.00000612, -0.00001380, 0.00000448, -0.00004510, -0.00000115, -0.00011102, -0.00007063, 0.00019961, 0.00006892, 0.00006814, 0.00010632, 0.00016525, -0.00015720, -0.00003801, -0.00002865, -0.00001914, -0.00001800, 0.00000308, -0.00001435, 0.00000377, -0.00000854, -0.00000020, -0.00000722, 0.00000665, -0.00000958, 0.00000182, -0.00000726, -0.00000206, -0.00000995, -0.00000123, -0.00000020, -0.00000209, 0.00000127, -0.00000359, -0.00000355, -0.00000420, -0.00000046, -0.00000584, -0.00000987, 0.00000371, -0.00002873, 0.00000001, 0.00000128, 0.00003730, -0.00000483, 0.00001490, 0.00000198, 0.00001722, 0.00000834, 0.00000135, 0.00001446, 0.00003217, 0.00000495, -0.00000488, -0.00003088, -0.00000546, -0.00000963, 0.00000523, -0.00000914, 0.00000867, -0.00000783, 0.00000639, -0.00000211, -0.00000417, 0.00000152, 0.00000173, -0.00000169, -0.00000572, 0.00000933, 0.00000684, -0.00000466, -0.00000902, 0.00000650, 0.00000284, -0.00000541, -0.00000124, 0.00000039, 0.00000024, -0.00000607, 0.00001378, -0.00001442, 0.00000480, 0.00000136, 0.00000498, -0.00001545, },
		{ 0.00001236, -0.00001212, 0.00001366, -0.00000845, -0.00000198, 0.00000379, -0.00001689, 0.00000699, 0.00000158, -0.00000013, -0.00000990, -0.00000886, 0.00000433, -0.00000201, -0.00000741, 0.00000075, 0.00000973, -0.00000919, -0.00000458, 0.00000135, 0.00000086, -0.00000514, -0.00000498, 0.00000306, -0.00000832, 0.00001487, -0.00001393, 0.00001990, -0.00001133, 0.00000112, -0.00000355, 0.00000402, -0.00000170, 0.00000774, -0.00000708, 0.00000660, -0.00000615, 0.00000282, -0.00001311, 0.00001177, -0.00002183, 0.00000824, -0.00002080, 0.00001251, -0.00001996, 0.00000530, -0.00005485, 0.00000606, -0.00016753, 0.00006787, 0.00015527, 0.00005963, 0.00005589, 0.00014589, 0.00009132, -0.00016546, -0.00000625, -0.00004099, -0.00001058, -0.00001661, 0.00000408, -0.00001573, 0.00000693, -0.00001146, 0.00000094, -0.00000562, 0.00000588, -0.00001095, 0.00000355, -0.00001015, -0.00000027, -0.00001126, 0.00000221, -0.00000270, -0.00000034, -0.00000035, -0.00000400, -0.00000239, -0.00000572, 0.00000269, -0.00001354, 0.00000036, -0.00000782, -0.00002409, -0.00000188, 0.00001466, 0.00002992, -0.00000471, 0.00001625, 0.00000304, 0.00001970, 0.00000402, 0.00000249, 0.00002571, 0.00002510, 0.00000137, -0.00001696, -0.00002417, -0.00000548, -0.00000582, 0.00000306, -0.00000722, 0.00000866, -0.00001000, 0.00001101, -0.00000923, 0.00000130, -0.00000062, 0.00000368, -0.00000546, -0.00000145, 0.00000957, 0.00000528, -0.00000779, -0.00000615, 0.00000888, -0.00000130, -0.00000405, -0.00000192, 0.00000304, -0.00000489, 0.00000128, 0.00000800, -0.00001404, 0.00000834, -0.00000053, 0.00000256, -0.00001552, },
		{ 0.00000707, -0.00000966, 0.00001569, -0.00001742, 0.00001051, -0.00001006, -0.00000811, 0.00000654, 0.00000176, -0.00000430, -0.00000969, -0.00000532, 0.00000442, -0.00000593, -0.00000474, 0.00000384, 0.00000794, -0.00001495, 0.00000324, -0.00000352, 0.00000516, -0.00001249, 0.00000401, -0.00000522, 0.00000174, 0.00000511, -0.00000619, 0.00001597, -0.00001242, 0.00000250, -0.00000280, 0.00000302, 0.00000129, 0.00000282, -0.00000258, 0.00000232, -0.00000173, -0.00000443, -0.00000399, 0.00000144, -0.00001347, 0.00000092, -0.00001262, 0.00000442, -0.00001068, -0.00001530, -0.00003452, -0.00003675, -0.00014761, 0.00016928, 0.00008428, 0.00007393, 0.00004150, 0.00018324, -0.00000235, -0.00013402, -0.00000068, -0.00004076, -0.00000896, -0.00000948, -0.00000258, -0.00000917, 0.00000293, -0.00000921, -0.00000140, -0.00000023, -0.00000016, -0.00000666, -0.00000027, -0.00000821, -0.00000198, -0.00000922, 0.00000262, -0.00000332, 0.00000016, -0.00000180, -0.00000314, -0.00000331, -0.00000421, 0.00000106, -0.00001577, 0.00000675, -0.00001917, -0.00001845, -0.00000140, 0.00002709, 0.00001851, 0.00000041, 0.00001443, 0.00000709, 0.00001885, 0.00000005, 0.00000860, 0.00003307, 0.00001572, -0.00000373, -0.00002618, -0.00001384, -0.00001001, 0.00000237, -0.00000473, 0.00000124, 0.00000086, -0.00000367, 0.00000719, -0.00000983, 0.00000326, -0.00000028, 0.00000258, -0.00000622, 0.00000124, 0.00001049, 0.00000194, -0.00000919, -0.00000274, 0.00000980, -0.00000559, -0.00000157, -0.00000287, 0.00000507, -0.00000955, 0.00001014, -0.00000268, -0.00000653, 0.00000540, 0.00000202, -0.00000494, -0.00000673, },
		{ -0.00000462, 0.00000259, 0.00000466, -0.00001233, 0.00001145, -0.00001790, 0.00000075, 0.00000466, 0.00000095, -0.00000798, -0.00000741, -0.00000273, 0.00000231, -0.00000732, -0.00000232, 0.00000774, 0.00000093, -0.00001372, 0.00000647, -0.00000558, 0.00000621, -0.00001648, 0.00001108, -0.00001246, 0.00001128, -0.00000598, 0.00000516, 0.00000681, -0.00000843, 0.00000056, -0.00000008, 0.00000117, 0.00000387, -0.00000161, 0.00000208, -0.00000225, 0.00000232, -0.00001084, 0.00000464, -0.00000981, -0.00000226, -0.00000997, -0.00000074, -0.00000721, 0.00000012, -0.00003903, -0.00000747, -0.00010527, -0.00004938, 0.00018802, 0.00004086, 0.00007897, 0.00004762, 0.00019125, -0.00008479, -0.00008541, -0.00001142, -0.00003262, -0.00000955, -0.00000312, -0.00000963, -0.00000097, -0.00000349, -0.00000547, -0.00000323, 0.00000392, -0.00000618, -0.00000141, -0.00000591, -0.00000394, -0.00000546, -0.00000536, 0.00000101, -0.00000247, -0.00000073, -0.00000209, -0.00000249, -0.00000479, -0.00000115, -0.00000444, -0.00001156, 0.00000600, -0.00002633, -0.00001455, 0.00000457, 0.00003305, 0.00000922, 0.00000594, 0.00001292, 0.00001101, 0.00001635, -0.00000313, 0.00002027, 0.00003210, 0.00000850, -0.00001379, -0.00002540, -0.00000966, -0.00000984, 0.00000593, -0.00000924, 0.00000769, -0.00000685, 0.00000465, -0.00000071, -0.00000595, 0.00000275, 0.00000105, 0.00000025, -0.00000540, 0.00000339, 0.00001091, -0.00000203, -0.00000931, 0.00000159, 0.00000776, -0.00000755, -0.00000033, -0.00000195, 0.00000371, -0.00000948, 0.00001445, -0.00001140, 0.00000211, 0.00000039, 0.00000492, -0.00001197, 0.00000444, },
		{ -0.00001170, 0.00001351, -0.00000999, 0.00000146, 0.00000077, -0.00001528, 0.00000446, 0.00000452, -0.00000307, -0.00000804, -0.00000570, -0.00000083, -0.00000108, -0.00000623, -0.00000015, 0.00001058, -0.00000879, -0.00000571, 0.00000283, -0.00000183, 0.00000065, -0.00001275, 0.00001107, -0.00001337, 0.00001454, -0.00001189, 0.00001349, -0.00000174, -0.00000385, -0.00000145, 0.00000211, 0.00000053, 0.00000372, -0.00000231, 0.00000286, -0.00000275, 0.00000124, -0.00001077, 0.00000575, -0.00001319, 0.00000173, -0.00001377, 0.00000350, -0.00000992, -0.00000335, -0.00004339, -0.00000856, -0.00014515, 0.00007476, 0.00013703, 0.00004149, 0.00006093, 0.00008261, 0.00015546, -0.00013177, -0.00004265, -0.00002520, -0.00002296, -0.00000850, -0.00000187, -0.00001155, 0.00000329, -0.00000785, -0.00000315, -0.00000221, 0.00000371, -0.00000817, 0.00000049, -0.00000914, -0.00000077, -0.00000809, -0.00000192, -0.00000085, -0.00000146, -0.00000216, -0.00000111, -0.00000332, -0.00000463, 0.00000010, -0.00000954, -0.00000473, -0.00000015, -0.00002969, -0.00001091, 0.00001396, 0.00003247, 0.00000382, 0.00001058, 0.00001227, 0.00001416, 0.00001178, -0.00000191, 0.00003244, 0.00002488, 0.00000177, -0.00002451, -0.00001697, -0.00001279, -0.00000254, 0.00000132, -0.00000569, 0.00000628, -0.00000771, 0.00000788, -0.00000636, -0.00000191, 0.00000234, 0.00000116, -0.00000098, -0.00000508, 0.00000667, 0.00000914, -0.00000503, -0.00000869, 0.00000639, 0.00000275, -0.00000624, -0.00000126, 0.00000125, -0.00000138, -0.00000342, 0.00001095, -0.00001286, 0.00000590, -0.00000148, 0.00000340, -0.00001205, 0.00000871, },
		{ -0.00000749, 0.00001307, -0.00001586, 0.00001155, -0.00001218, -0.00000664, 0.00000407, 0.00000501, -0.00000861, -0.00000465, -0.00000564, 0.00000029, -0.00000407, -0.00000413, 0.00000292, 0.00000907, -0.00001534, 0.00000335, -0.00000371, 0.00000473, -0.00000886, -0.00000285, 0.00000417, -0.00000745, 0.00001038, -0.00001036, 0.00001561, -0.00000633, -0.00000137, -0.00000176, 0.00000288, 0.00000092, 0.00000184, -0.00000036, 0.00000064, -0.00000021, -0.00000362, -0.00000566, 0.00000045, -0.00000887, -0.00000231, -0.00000937, -0.00000072, -0.00000394, -0.00002067, -0.00002647, -0.00004919, -0.00011606, 0.00015813, 0.00006823, 0.00006373, 0.00003425, 0.00013068, 0.00008501, -0.00013839, -0.00001751, -0.00003397, -0.00001469, -0.00000631, -0.00000455, -0.00000843, 0.00000267, -0.00000914, -0.00000205, 0.00000031, 0.00000031, -0.00000644, -0.00000154, -0.00000857, -0.00000026, -0.00000861, -0.00000012, -0.00000180, -0.00000140, -0.00000281, -0.00000002, -0.00000512, -0.00000255, -0.00000204, -0.00001077, 0.00000008, -0.00000802, -0.00003121, -0.00000471, 0.00002245, 0.00002858, 0.00000108, 0.00001526, 0.00001118, 0.00001720, 0.00000479, 0.00000698, 0.00003849, 0.00001631, -0.00000833, -0.00002795, -0.00001015, -0.00001582, 0.00000552, -0.00000619, 0.00000199, -0.00000062, -0.00000229, 0.00000472, -0.00000730, 0.00000033, 0.00000264, 0.00000007, -0.00000106, -0.00000516, 0.00001075, 0.00000515, -0.00000666, -0.00000690, 0.00000976, -0.00000293, -0.00000335, -0.00000278, 0.00000435, -0.00000687, 0.00000500, 0.00000226, -0.00000771, 0.00000403, 0.00000050, -0.00000233, -0.00000481, 0.00000287, },
		{ 0.00000406, 0.00000166, -0.00000867, 0.00000959, -0.00001765, 0.00000049, 0.00000377, 0.00000273, -0.00001154, -0.00000085, -0.00000648, 0.00000043, -0.00000549, -0.00000202, 0.00000680, 0.00000214, -0.00001435, 0.00000715, -0.00000694, 0.00000801, -0.00001588, 0.00000708, -0.00000457, 0.00000099, 0.00000251, -0.00000397, 0.00001285, -0.00000744, -0.00000071, -0.00000084, 0.00000302, 0.00000077, 0.00000095, 0.00000081, -0.00000080, 0.00000105, -0.00000711, -0.00000152, -0.00000441, -0.00000409, -0.00000715, -0.00000439, -0.00000458, -0.00000126, -0.00003506, -0.00001124, -0.00010209, -0.00002534, 0.00016755, 0.00002824, 0.00007547, 0.00002385, 0.00016433, 0.00000492, -0.00011647, -0.00000963, -0.00003592, -0.00000794, -0.00000553, -0.00000709, -0.00000393, -0.00000051, -0.00000859, -0.00000067, 0.00000158, -0.00000291, -0.00000409, -0.00000493, -0.00000572, -0.00000177, -0.00000744, 0.00000019, -0.00000185, -0.00000249, -0.00000175, -0.00000049, -0.00000585, -0.00000083, -0.00000562, -0.00000853, 0.00000130, -0.00001590, -0.00003089, 0.00000424, 0.00002786, 0.00002351, 0.00000053, 0.00001975, 0.00000964, 0.00001921, -0.00000188, 0.00002195, 0.00003590, 0.00000825, -0.00002109, -0.00002152, -0.00001123, -0.00001177, 0.00000676, -0.00000869, 0.00000609, -0.00000580, 0.00000336, -0.00000063, -0.00000556, 0.00000180, 0.00000220, -0.00000030, -0.00000195, -0.00000360, 0.00001340, 0.00000067, -0.00000762, -0.00000322, 0.00000969, -0.00000637, -0.00000153, -0.00000249, 0.00000436, -0.00000849, 0.00001011, -0.00000542, -0.00000129, 0.00000049, 0.00000250, -0.00000690, 0.00000301, -0.00000685, },
		{ 0.00001211, -0.00001055, 0.00000436, -0.00000205, -0.00001341, 0.00000249, 0.00000551, -0.00000355, -0.00000912, 0.00000018, -0.00000683, -0.00000027, -0.00000541, 0.00000049, 0.00000903, -0.00000687, -0.00000697, 0.00000414, -0.00000420, 0.00000467, -0.00001577, 0.00001154, -0.00000984, 0.00000699, -0.00000426, 0.00000312, 0.00000823, -0.00000705, -0.00000045, 0.00000028, 0.00000303, -0.00000029, 0.00000191, -0.00000030, 0.00000032, -0.00000109, -0.00000651, -0.00000184, -0.00000433, -0.00000424, -0.00000704, -0.00000480, -0.00000166, -0.00001083, -0.00003195, -0.00002329, -0.00012164, 0.00007762, 0.00012031, 0.00002824, 0.00006367, 0.00004170, 0.00016442, -0.00006056, -0.00008318, -0.00001285, -0.00003197, -0.00000312, -0.00000682, -0.00000709, -0.00000104, -0.00000396, -0.00000698, 0.00000114, 0.00000070, -0.00000411, -0.00000352, -0.00000705, -0.00000281, -0.00000382, -0.00000564, -0.00000014, -0.00000177, -0.00000383, 0.00000056, -0.00000300, -0.00000413, -0.00000192, -0.00000721, -0.00000598, 0.00000032, -0.00002426, -0.00002669, 0.00001323, 0.00003119, 0.00001703, 0.00000359, 0.00002173, 0.00000965, 0.00001767, -0.00000290, 0.00003661, 0.00002736, -0.00000153, -0.00002993, -0.00001150, -0.00001721, -0.00000184, 0.00000065, -0.00000408, 0.00000329, -0.00000492, 0.00000437, -0.00000375, -0.00000386, 0.00000363, 0.00000018, 0.00000110, -0.00000475, 0.00000094, 0.00001278, -0.00000268, -0.00000820, 0.00000174, 0.00000612, -0.00000658, -0.00000169, 0.00000001, 0.00000089, -0.00000505, 0.00000911, -0.00000779, 0.00000174, -0.00000083, 0.00000118, -0.00000560, 0.00000386, -0.00001064, },
		{ 0.00000921, -0.00001306, 0.00001157, -0.00001360, -0.00000497, 0.00000160, 0.00000701, -0.00001091, -0.00000256, -0.00000249, -0.00000584, -0.00000144, -0.00000432, 0.00000387, 0.00000691, -0.00001245, 0.00000114, -0.00000167, 0.00000162, -0.00000310, -0.00000926, 0.00000959, -0.00000995, 0.00000853, -0.00000753, 0.00000850, 0.00000354, -0.00000617, -0.00000008, 0.00000146, 0.00000224, -0.00000067, 0.00000281, -0.00000172, 0.00000188, -0.00000426, -0.00000426, -0.00000405, -0.00000166, -0.00000735, -0.00000402, -0.00000780, 0.00000347, -0.00002679, -0.00001671, -0.00006332, -0.00008204, 0.00013955, 0.00006172, 0.00004775, 0.00003984, 0.00007879, 0.00013041, -0.00009925, -0.00005173, -0.00002066, -0.00002362, -0.00000167, -0.00000818, -0.00000535, -0.00000026, -0.00000685, -0.00000410, 0.00000188, -0.00000080, -0.00000414, -0.00000456, -0.00000712, -0.00000103, -0.00000551, -0.00000395, -0.00000046, -0.00000237, -0.00000399, 0.00000226, -0.00000588, -0.00000117, -0.00000551, -0.00000561, -0.00000532, -0.00000162, -0.00003294, -0.00001759, 0.00001991, 0.00003366, 0.00000907, 0.00001105, 0.00001939, 0.00001264, 0.00001159, 0.00000532, 0.00004430, 0.00001663, -0.00001452, -0.00002874, -0.00000709, -0.00001959, 0.00000615, -0.00000609, 0.00000236, -0.00000255, -0.00000020, 0.00000128, -0.00000400, -0.00000254, 0.00000495, -0.00000193, 0.00000273, -0.00000787, 0.00000719, 0.00000920, -0.00000469, -0.00000768, 0.00000592, 0.00000130, -0.00000509, -0.00000238, 0.00000257, -0.00000325, 0.00000048, 0.00000423, -0.00000565, 0.00000118, -0.00000006, -0.00000214, 0.00000019, -0.00000298, -0.00000452, },
		{ -0.00000187, -0.00000443, 0.00000715, -0.00001687, 0.00000072, 0.00000197, 0.00000458, -0.00001442, 0.00000386, -0.00000717, -0.00000363, -0.00000283, -0.00000212, 0.00000696, 0.00000063, -0.00001154, 0.00000478, -0.00000502, 0.00000541, -0.00001009, -0.00000083, 0.00000416, -0.00000666, 0.00000667, -0.00000754, 0.00001187, -0.00000101, -0.00000460, 0.00000010, 0.00000288, 0.00000013, 0.00000098, 0.00000152, -0.00000092, 0.00000087, -0.00000492, -0.00000440, -0.00000346, -0.00000177, -0.00000782, -0.00000413, -0.00000616, 0.00000041, -0.00003381, -0.00001155, -0.00010183, 0.00000079, 0.00014033, 0.00002783, 0.00006050, 0.00002424, 0.00011497, 0.00007578, -0.00011128, -0.00002889, -0.00002791, -0.00001328, -0.00000394, -0.00000756, -0.00000379, -0.00000088, -0.00000866, -0.00000027, 0.00000064, -0.00000123, -0.00000475, -0.00000566, -0.00000592, -0.00000043, -0.00000660, -0.00000256, -0.00000071, -0.00000375, -0.00000213, 0.00000148, -0.00000655, -0.00000011, -0.00000835, -0.00000314, -0.00000569, -0.00000558, -0.00003931, -0.00000576, 0.00002490, 0.00003400, 0.00000229, 0.00002037, 0.00001418, 0.00001694, 0.00000413, 0.00002096, 0.00004256, 0.00000468, -0.00002701, -0.00001890, -0.00001128, -0.00001364, 0.00000625, -0.00000697, 0.00000415, -0.00000524, 0.00000295, -0.00000192, -0.00000353, -0.00000033, 0.00000403, -0.00000182, 0.00000201, -0.00000842, 0.00001224, 0.00000457, -0.00000603, -0.00000532, 0.00000738, -0.00000225, -0.00000399, -0.00000177, 0.00000285, -0.00000481, 0.00000394, -0.00000004, -0.00000291, 0.00000002, -0.00000017, -0.00000286, 0.00000384, -0.00001046, 0.00000587, },
		{ -0.00001097, 0.00000702, -0.00000469, -0.00001146, 0.00000138, 0.00000474, -0.00000257, -0.00001120, 0.00000593, -0.00001093, -0.00000116, -0.00000414, 0.00000153, 0.00000719, -0.00000608, -0.00000600, 0.00000324, -0.00000402, 0.00000461, -0.00001268, 0.00000538, -0.00000135, -0.00000245, 0.00000323, -0.00000523, 0.00001330, -0.00000522, -0.00000233, 0.00000014, 0.00000385, -0.00000213, 0.00000368, -0.00000135, 0.00000173, -0.00000256, -0.00000293, -0.00000723, 0.00000070, -0.00000603, -0.00000405, -0.00000883, 0.00000150, -0.00001391, -0.00002452, -0.00003313, -0.00010248, 0.00008202, 0.00009991, 0.00002488, 0.00005613, 0.00002784, 0.00013394, 0.00001699, -0.00010325, -0.00001669, -0.00003076, -0.00000462, -0.00000777, -0.00000513, -0.00000317, -0.00000253, -0.00000839, 0.00000274, -0.00000145, -0.00000080, -0.00000634, -0.00000584, -0.00000432, -0.00000080, -0.00000706, -0.00000141, -0.00000126, -0.00000503, 0.00000097, -0.00000178, -0.00000412, -0.00000284, -0.00000736, -0.00000310, -0.00000489, -0.00001337, -0.00003994, 0.00000508, 0.00003055, 0.00002957, 0.00000098, 0.00002667, 0.00001009, 0.00001881, 0.00000119, 0.00003758, 0.00003302, -0.00000932, -0.00003217, -0.00000894, -0.00001823, -0.00000321, 0.00000031, -0.00000218, 0.00000030, -0.00000275, 0.00000183, -0.00000264, -0.00000373, 0.00000290, 0.00000092, 0.00000085, -0.00000179, -0.00000495, 0.00001398, 0.00000053, -0.00000685, -0.00000172, 0.00000608, -0.00000378, -0.00000380, 0.00000014, 0.00000087, -0.00000329, 0.00000385, -0.00000133, -0.00000211, 0.00000014, -0.00000240, 0.00000128, 0.00000025, -0.00001057, 0.00001070, },
		{ -0.00001007, 0.00001109, -0.00001442, -0.00000346, -0.00000003, 0.00000698, -0.00001090, -0.00000287, 0.00000227, -0.00001150, 0.00000018, -0.00000441, 0.00000583, 0.00000325, -0.00000897, -0.00000059, -0.00000019, -0.00000084, 0.00000062, -0.00001092, 0.00000810, -0.00000534, 0.00000142, -0.00000087, -0.00000085, 0.00001213, -0.00000800, -0.00000020, 0.00000070, 0.00000317, -0.00000241, 0.00000472, -0.00000292, 0.00000312, -0.00000507, -0.00000179, -0.00000878, 0.00000407, -0.00001013, -0.00000066, -0.00001241, 0.00000722, -0.00002943, -0.00001018, -0.00007286, -0.00005310, 0.00012228, 0.00005514, 0.00003638, 0.00004215, 0.00004711, 0.00013040, -0.00003395, -0.00008328, -0.00001402, -0.00002749, -0.00000082, -0.00000991, -0.00000283, -0.00000306, -0.00000497, -0.00000543, 0.00000302, -0.00000228, -0.00000122, -0.00000773, -0.00000533, -0.00000267, -0.00000200, -0.00000683, -0.00000052, -0.00000251, -0.00000486, 0.00000312, -0.00000522, -0.00000077, -0.00000772, -0.00000296, -0.00000619, -0.00000271, -0.00002454, -0.00003393, 0.00001335, 0.00003722, 0.00002010, 0.00000710, 0.00002658, 0.00001008, 0.00001614, 0.00000698, 0.00004838, 0.00001819, -0.00002372, -0.00002690, -0.00000664, -0.00001970, 0.00000374, -0.00000465, 0.00000244, -0.00000441, 0.00000147, -0.00000153, -0.00000176, -0.00000366, 0.00000525, -0.00000203, 0.00000371, -0.00000650, 0.00000117, 0.00001251, -0.00000256, -0.00000654, 0.00000140, 0.00000373, -0.00000440, -0.00000333, 0.00000158, -0.00000114, -0.00000107, 0.00000229, -0.00000097, -0.00000241, 0.00000025, -0.00000401, 0.00000677, -0.00000856, -0.00000227, 0.00000538, },
		{ -0.00000035, 0.00000453, -0.00001561, 0.00000102, 0.00000043, 0.00000494, -0.00001505, 0.00000557, -0.00000494, -0.00000857, -0.00000049, -0.00000230, 0.00000824, -0.00000296, -0.00000695, 0.00000146, -0.00000211, 0.00000139, -0.00000355, -0.00000700, 0.00000806, -0.00000764, 0.00000450, -0.00000504, 0.00000525, 0.00000784, -0.00000824, 0.00000093, 0.00000197, 0.00000071, 0.00000001, 0.00000272, -0.00000139, 0.00000110, -0.00000416, -0.00000429, -0.00000558, 0.00000234, -0.00000936, -0.00000270, -0.00000875, 0.00000191, -0.00003236, -0.00001167, -0.00010019, 0.00002335, 0.00011370, 0.00003068, 0.00004419, 0.00003130, 0.00007116, 0.00010770, -0.00007052, -0.00005914, -0.00001722, -0.00001957, -0.00000218, -0.00000876, -0.00000210, -0.00000317, -0.00000713, -0.00000094, 0.00000066, -0.00000108, -0.00000353, -0.00000773, -0.00000478, -0.00000110, -0.00000386, -0.00000583, -0.00000011, -0.00000427, -0.00000263, 0.00000250, -0.00000607, -0.00000003, -0.00001059, 0.00000105, -0.00000973, -0.00000181, -0.00003556, -0.00002393, 0.00002091, 0.00004168, 0.00000971, 0.00001771, 0.00002119, 0.00001340, 0.00001094, 0.00002086, 0.00004941, -0.00000033, -0.00003306, -0.00001598, -0.00001186, -0.00001341, 0.00000328, -0.00000397, 0.00000207, -0.00000510, 0.00000286, -0.00000350, -0.00000148, -0.00000183, 0.00000483, -0.00000235, 0.00000399, -0.00000903, 0.00000714, 0.00000923, -0.00000487, -0.00000479, 0.00000274, 0.00000192, -0.00000522, -0.00000166, 0.00000121, -0.00000134, -0.00000055, 0.00000197, -0.00000152, -0.00000174, -0.00000180, -0.00000116, 0.00000714, -0.00001478, 0.00000734, -0.00000521, },
		{ 0.00000929, -0.00000705, -0.00000946, 0.00000080, 0.00000332, -0.00000193, -0.00001201, 0.00000906, -0.00001154, -0.00000402, -0.00000251, 0.00000240, 0.00000629, -0.00000739, -0.00000276, 0.00000059, -0.00000190, 0.00000162, -0.00000623, -0.00000262, 0.00000597, -0.00000793, 0.00000588, -0.00000780, 0.00001144, 0.00000145, -0.00000610, 0.00000120, 0.00000298, -0.00000178, 0.00000324, -0.00000065, 0.00000167, -0.00000285, -0.00000122, -0.00000886, 0.00000092, -0.00000359, -0.00000451, -0.00000895, 0.00000036, -0.00001351, -0.00002070, -0.00003828, -0.00008715, 0.00008676, 0.00007829, 0.00002744, 0.00004263, 0.00002988, 0.00009083, 0.00007204, -0.00009001, -0.00003745, -0.00002138, -0.00001100, -0.00000593, -0.00000544, -0.00000279, -0.00000384, -0.00000733, 0.00000246, -0.00000213, 0.00000066, -0.00000687, -0.00000639, -0.00000430, -0.00000001, -0.00000590, -0.00000422, -0.00000055, -0.00000555, 0.00000062, -0.00000057, -0.00000385, -0.00000336, -0.00000861, 0.00000099, -0.00001053, -0.00000523, -0.00004242, -0.00001342, 0.00003042, 0.00003978, 0.00000440, 0.00002679, 0.00001493, 0.00001664, 0.00000821, 0.00003781, 0.00003974, -0.00001998, -0.00003235, -0.00000829, -0.00001741, -0.00000469, -0.00000111, 0.00000036, -0.00000236, -0.00000164, 0.00000048, -0.00000284, -0.00000235, 0.00000137, 0.00000207, -0.00000004, 0.00000109, -0.00000795, 0.00001106, 0.00000531, -0.00000611, -0.00000244, 0.00000267, 0.00000067, -0.00000593, 0.00000047, -0.00000023, -0.00000029, -0.00000150, 0.00000316, -0.00000343, -0.00000011, -0.00000526, 0.00000606, -0.00000042, -0.00001219, 0.00000968, -0.00001141, },
		{ 0.00001046, -0.00001478, -0.00000214, -0.00000080, 0.00000549, -0.00000988, -0.00000348, 0.00000577, -0.00001406, -0.00000070, -0.00000359, 0.00000752, 0.00000022, -0.00000723, -0.00000009, -0.00000060, -0.00000120, 0.00000091, -0.00000765, 0.00000176, 0.00000206, -0.00000566, 0.00000437, -0.00000719, 0.00001506, -0.00000463, -0.00000300, 0.00000141, 0.00000255, -0.00000217, 0.00000447, -0.00000230, 0.00000292, -0.00000511, -0.00000015, -0.00001094, 0.00000555, -0.00000849, -0.00000114, -0.00001265, 0.00000573, -0.00002722, -0.00000880, -0.00007528, -0.00003218, 0.00010958, 0.00004543, 0.00003210, 0.00003705, 0.00003607, 0.00010192, 0.00002967, -0.00009266, -0.00002246, -0.00002273, -0.00000545, -0.00000852, -0.00000232, -0.00000400, -0.00000511, -0.00000491, 0.00000283, -0.00000298, 0.00000071, -0.00000930, -0.00000475, -0.00000351, 0.00000005, -0.00000741, -0.00000239, -0.00000200, -0.00000517, 0.00000273, -0.00000364, -0.00000108, -0.00000847, -0.00000296, -0.00000338, -0.00000814, -0.00001334, -0.00004382, -0.00000320, 0.00004139, 0.00003084, 0.00000735, 0.00002971, 0.00001181, 0.00001708, 0.00001232, 0.00005115, 0.00002025, -0.00003527, -0.00002279, -0.00000869, -0.00001683, -0.00000031, -0.00000320, 0.00000280, -0.00000592, 0.00000201, -0.00000289, -0.00000125, -0.00000284, 0.00000379, -0.00000060, 0.00000261, -0.00000302, -0.00000423, 0.00001265, 0.00000117, -0.00000568, -0.00000096, 0.00000256, -0.00000100, -0.00000534, 0.00000143, -0.00000081, -0.00000014, -0.00000162, 0.00000366, -0.00000472, 0.00000008, -0.00000610, 0.00001226, -0.00001148, -0.00000226, 0.00000273, -0.00000769, },
		{ 0.00000225, -0.00001410, 0.00000121, -0.00000033, 0.00000349, -0.00001364, 0.00000528, -0.00000194, -0.00001182, -0.00000030, -0.00000151, 0.00000943, -0.00000648, -0.00000338, -0.00000020, -0.00000061, -0.00000120, 0.00000030, -0.00000827, 0.00000568, -0.00000293, -0.00000129, -0.00000001, -0.00000243, 0.00001425, -0.00000821, -0.00000031, 0.00000183, 0.00000079, -0.00000020, 0.00000287, -0.00000103, 0.00000085, -0.00000388, -0.00000296, -0.00000790, 0.00000484, -0.00000863, -0.00000309, -0.00000920, 0.00000027, -0.00002802, -0.00001470, -0.00009333, 0.00003776, 0.00009328, 0.00003022, 0.00003362, 0.00003389, 0.00004551, 0.00010288, -0.00001304, -0.00008179, -0.00001497, -0.00002045, -0.00000399, -0.00000824, -0.00000098, -0.00000518, -0.00000602, -0.00000122, 0.00000065, -0.00000153, -0.00000174, -0.00000959, -0.00000373, -0.00000221, -0.00000118, -0.00000785, -0.00000098, -0.00000394, -0.00000288, 0.00000227, -0.00000436, -0.00000097, -0.00001117, 0.00000213, -0.00000891, -0.00000513, -0.00002358, -0.00004140, 0.00000891, 0.00004948, 0.00001927, 0.00001633, 0.00002636, 0.00001260, 0.00001518, 0.00002422, 0.00005481, -0.00000575, -0.00003988, -0.00001209, -0.00001389, -0.00001060, -0.00000164, -0.00000063, 0.00000047, -0.00000525, 0.00000239, -0.00000425, -0.00000068, -0.00000166, 0.00000388, -0.00000120, 0.00000334, -0.00000595, 0.00000030, 0.00001240, -0.00000272, -0.00000375, -0.00000088, 0.00000307, -0.00000343, -0.00000318, 0.00000064, 0.00000022, -0.00000193, 0.00000057, 0.00000173, -0.00000381, -0.00000274, -0.00000110, 0.00001112, -0.00001775, 0.00000697, -0.00000714, 0.00000259, },
		{ -0.00000876, -0.00000745, 0.00000040, 0.00000222, -0.00000279, -0.00001051, 0.00000904, -0.00000932, -0.00000726, -0.00000181, 0.00000362, 0.00000594, -0.00000955, 0.00000061, -0.00000148, -0.00000013, -0.00000137, -0.00000057, -0.00000752, 0.00000780, -0.00000701, 0.00000309, -0.00000516, 0.00000487, 0.00000939, -0.00000864, 0.00000150, 0.00000187, -0.00000073, 0.00000220, 0.00000031, 0.00000130, -0.00000264, -0.00000104, -0.00000752, -0.00000180, 0.00000036, -0.00000563, -0.00000796, -0.00000194, -0.00001274, -0.00001769, -0.00004174, -0.00007211, 0.00008771, 0.00006075, 0.00002913, 0.00003099, 0.00003426, 0.00005604, 0.00009242, -0.00004929, -0.00006311, -0.00001298, -0.00001622, -0.00000501, -0.00000576, -0.00000157, -0.00000626, -0.00000535, 0.00000138, -0.00000181, 0.00000044, -0.00000549, -0.00000804, -0.00000334, -0.00000075, -0.00000339, -0.00000705, -0.00000062, -0.00000522, 0.00000005, -0.00000008, -0.00000277, -0.00000433, -0.00000913, 0.00000310, -0.00001222, -0.00000448, -0.00003312, -0.00003677, 0.00002478, 0.00004992, 0.00001161, 0.00002555, 0.00002042, 0.00001513, 0.00001433, 0.00004075, 0.00004463, -0.00003121, -0.00003223, -0.00000801, -0.00001656, -0.00000471, -0.00000419, 0.00000325, -0.00000424, -0.00000159, -0.00000020, -0.00000332, -0.00000113, 0.00000054, 0.00000227, 0.00000000, 0.00000195, -0.00000682, 0.00000458, 0.00001042, -0.00000535, -0.00000161, -0.00000133, 0.00000351, -0.00000562, -0.00000067, -0.00000065, 0.00000149, -0.00000419, 0.00000385, -0.00000174, -0.00000186, -0.00000649, 0.00000796, 0.00000154, -0.00001445, 0.00000817, -0.00001112, 0.00001050, },
		{ -0.00001462, -0.00000094, -0.00000113, 0.00000364, -0.00000945, -0.00000251, 0.00000613, -0.00001255, -0.00000387, -0.00000223, 0.00000863, -0.00000142, -0.00000745, 0.00000193, -0.00000185, -0.00000051, -0.00000059, -0.00000258, -0.00000462, 0.00000678, -0.00000810, 0.00000489, -0.00000805, 0.00001154, 0.00000272, -0.00000673, 0.00000250, 0.00000104, -0.00000062, 0.00000291, -0.00000073, 0.00000189, -0.00000445, 0.00000005, -0.00000988, 0.00000293, -0.00000355, -0.00000407, -0.00001000, 0.00000093, -0.00002225, -0.00001077, -0.00007240, -0.00001721, 0.00009927, 0.00003541, 0.00003084, 0.00002912, 0.00003533, 0.00006741, 0.00006988, -0.00007333, -0.00004283, -0.00001367, -0.00001216, -0.00000615, -0.00000285, -0.00000348, -0.00000684, -0.00000308, 0.00000160, -0.00000262, 0.00000091, -0.00000857, -0.00000586, -0.00000304, 0.00000016, -0.00000583, -0.00000533, -0.00000150, -0.00000481, 0.00000171, -0.00000211, -0.00000129, -0.00000866, -0.00000385, -0.00000027, -0.00001239, -0.00000710, -0.00004116, -0.00002891, 0.00004251, 0.00004193, 0.00001174, 0.00002995, 0.00001611, 0.00001674, 0.00001870, 0.00005534, 0.00002020, -0.00004670, -0.00001846, -0.00001146, -0.00001308, -0.00000376, -0.00000320, 0.00000381, -0.00000689, 0.00000131, -0.00000298, -0.00000187, -0.00000120, 0.00000199, 0.00000089, 0.00000124, -0.00000011, -0.00000626, 0.00000845, 0.00000667, -0.00000576, -0.00000074, -0.00000097, 0.00000272, -0.00000624, 0.00000062, -0.00000079, 0.00000105, -0.00000446, 0.00000551, -0.00000419, -0.00000166, -0.00000670, 0.00001469, -0.00001066, -0.00000459, 0.00000120, -0.00000583, 0.00000946, },
		{ -0.00001241, 0.00000134, -0.00000047, 0.00000100, -0.00001161, 0.00000499, -0.00000069, -0.00001129, -0.00000321, 0.00000071, 0.00000953, -0.00000819, -0.00000244, 0.00000066, -0.00000102, -0.00000172, 0.00000099, -0.00000545, -0.00000007, 0.00000286, -0.00000579, 0.00000309, -0.00000681, 0.00001490, -0.00000329, -0.00000365, 0.00000261, -0.00000008, 0.00000080, 0.00000178, 0.00000020, 0.00000001, -0.00000351, -0.00000185, -0.00000837, 0.00000401, -0.00000473, -0.00000582, -0.00000698, -0.00000462, -0.00002098, -0.00002068, -0.00008120, 0.00004404, 0.00007939, 0.00002599, 0.00002881, 0.00003099, 0.00003542, 0.00007823, 0.00003723, -0.00008234, -0.00002569, -0.00001499, -0.00000926, -0.00000604, -0.00000100, -0.00000599, -0.00000614, -0.00000061, 0.00000017, -0.00000175, -0.00000074, -0.00000980, -0.00000403, -0.00000242, -0.00000001, -0.00000770, -0.00000341, -0.00000304, -0.00000279, 0.00000126, -0.00000224, -0.00000232, -0.00001045, 0.00000090, -0.00000539, -0.00001088, -0.00001209, -0.00004832, -0.00001489, 0.00005656, 0.00002991, 0.00001833, 0.00002837, 0.00001533, 0.00001675, 0.00003056, 0.00005956, -0.00001266, -0.00004612, -0.00000850, -0.00001629, -0.00000674, -0.00000647, 0.00000128, 0.00000014, -0.00000568, 0.00000129, -0.00000394, -0.00000115, -0.00000026, 0.00000196, 0.00000075, 0.00000136, -0.00000169, -0.00000490, 0.00001159, 0.00000181, -0.00000409, -0.00000145, 0.00000057, 0.00000055, -0.00000514, 0.00000045, 0.00000037, -0.00000136, -0.00000195, 0.00000423, -0.00000441, -0.00000416, -0.00000093, 0.00001335, -0.00001743, 0.00000381, -0.00000708, 0.00000400, 0.00000043, },
		{ -0.00000555, 0.00000005, 0.00000173, -0.00000507, -0.00000735, 0.00000722, -0.00000653, -0.00000848, -0.00000359, 0.00000602, 0.00000497, -0.00001051, 0.00000176, -0.00000109, -0.00000034, -0.00000236, 0.00000188, -0.00000750, 0.00000414, -0.00000190, -0.00000181, -0.00000102, -0.00000193, 0.00001413, -0.00000710, -0.00000042, 0.00000166, -0.00000031, 0.00000188, 0.00000050, 0.00000136, -0.00000254, -0.00000153, -0.00000494, -0.00000471, 0.00000283, -0.00000490, -0.00000808, -0.00000307, -0.00001371, -0.00001367, -0.00004554, -0.00005529, 0.00008291, 0.00004972, 0.00002674, 0.00002480, 0.00003444, 0.00003637, 0.00008446, -0.00000013, -0.00007729, -0.00001426, -0.00001581, -0.00000737, -0.00000460, -0.00000101, -0.00000810, -0.00000397, 0.00000055, -0.00000121, -0.00000059, -0.00000355, -0.00000926, -0.00000283, -0.00000153, -0.00000136, -0.00000840, -0.00000213, -0.00000408, -0.00000049, -0.00000030, -0.00000115, -0.00000575, -0.00000832, 0.00000236, -0.00000965, -0.00000936, -0.00001891, -0.00005393, 0.00000659, 0.00006147, 0.00002049, 0.00002629, 0.00002358, 0.00001696, 0.00001745, 0.00004796, 0.00004647, -0.00004208, -0.00003221, -0.00000790, -0.00001597, -0.00000347, -0.00000781, 0.00000512, -0.00000458, -0.00000257, -0.00000067, -0.00000341, -0.00000080, 0.00000083, 0.00000137, 0.00000110, 0.00000086, -0.00000290, -0.00000254, 0.00001304, -0.00000266, -0.00000169, -0.00000279, 0.00000243, -0.00000203, -0.00000339, -0.00000005, 0.00000135, -0.00000389, 0.00000162, 0.00000129, -0.00000383, -0.00000677, 0.00000805, 0.00000405, -0.00001522, 0.00000483, -0.00000920, 0.00001003, -0.00000958, },
		{ 0.00000014, -0.00000130, 0.00000216, -0.00001033, 0.00000042, 0.00000353, -0.00000826, -0.00000736, -0.00000179, 0.00000984, -0.00000240, -0.00000802, 0.00000319, -0.00000179, -0.00000074, -0.00000142, 0.00000082, -0.00000722, 0.00000603, -0.00000511, 0.00000136, -0.00000491, 0.00000446, 0.00001010, -0.00000820, 0.00000203, 0.00000008, 0.00000072, 0.00000160, 0.00000048, 0.00000099, -0.00000370, -0.00000068, -0.00000678, -0.00000147, 0.00000166, -0.00000650, -0.00000727, -0.00000405, -0.00001769, -0.00001309, -0.00006752, -0.00000468, 0.00008801, 0.00002870, 0.00002802, 0.00002365, 0.00003527, 0.00004146, 0.00008107, -0.00003471, -0.00006238, -0.00000903, -0.00001553, -0.00000596, -0.00000255, -0.00000302, -0.00000858, -0.00000138, 0.00000026, -0.00000163, -0.00000033, -0.00000619, -0.00000769, -0.00000210, -0.00000074, -0.00000351, -0.00000785, -0.00000187, -0.00000378, 0.00000059, -0.00000117, -0.00000102, -0.00000902, -0.00000402, 0.00000054, -0.00001223, -0.00000833, -0.00002837, -0.00005453, 0.00003243, 0.00005601, 0.00001812, 0.00003056, 0.00001970, 0.00001842, 0.00002319, 0.00006344, 0.00001570, -0.00005577, -0.00001569, -0.00001348, -0.00001006, -0.00000540, -0.00000483, 0.00000493, -0.00000687, -0.00000049, -0.00000242, -0.00000265, 0.00000003, 0.00000102, 0.00000143, 0.00000080, 0.00000080, -0.00000444, 0.00000132, 0.00001186, -0.00000525, -0.00000006, -0.00000353, 0.00000354, -0.00000396, -0.00000201, 0.00000011, 0.00000080, -0.00000451, 0.00000387, -0.00000121, -0.00000460, -0.00000556, 0.00001401, -0.00000719, -0.00000757, -0.00000048, -0.00000359, 0.00000747, -0.00001370, },
		{ 0.00000148, -0.00000037, -0.00000153, -0.00001025, 0.00000602, -0.00000227, -0.00000691, -0.00000825, 0.00000344, 0.00000872, -0.00000803, -0.00000348, 0.00000256, -0.00000176, -0.00000161, 0.00000042, -0.00000177, -0.00000472, 0.00000520, -0.00000577, 0.00000226, -0.00000661, 0.00001006, 0.00000442, -0.00000682, 0.00000288, -0.00000093, 0.00000198, 0.00000058, 0.00000132, -0.00000077, -0.00000319, -0.00000133, -0.00000686, 0.00000072, 0.00000054, -0.00000891, -0.00000408, -0.00001032, -0.00001390, -0.00002696, -0.00006690, 0.00004577, 0.00006859, 0.00002138, 0.00002566, 0.00002698, 0.00003200, 0.00005170, 0.00006512, -0.00005949, -0.00004347, -0.00000875, -0.00001400, -0.00000456, -0.00000096, -0.00000625, -0.00000678, 0.00000015, -0.00000043, -0.00000151, -0.00000110, -0.00000796, -0.00000574, -0.00000163, -0.00000043, -0.00000584, -0.00000649, -0.00000225, -0.00000238, 0.00000018, -0.00000066, -0.00000322, -0.00000955, -0.00000057, -0.00000275, -0.00001362, -0.00000785, -0.00004131, -0.00004552, 0.00005583, 0.00004414, 0.00002244, 0.00002933, 0.00001924, 0.00001831, 0.00003736, 0.00006581, -0.00002314, -0.00004948, -0.00000732, -0.00001720, -0.00000399, -0.00000909, 0.00000058, 0.00000127, -0.00000622, -0.00000034, -0.00000309, -0.00000201, 0.00000110, 0.00000051, 0.00000207, -0.00000016, 0.00000122, -0.00000617, 0.00000628, 0.00000813, -0.00000547, 0.00000015, -0.00000315, 0.00000359, -0.00000505, -0.00000102, 0.00000064, -0.00000096, -0.00000304, 0.00000405, -0.00000272, -0.00000664, 0.00000052, 0.00001258, -0.00001358, -0.00000138, -0.00000585, 0.00000455, -0.00000142, -0.00001040, },
		{ -0.00000021, 0.00000158, -0.00000788, -0.00000400, 0.00000543, -0.00000523, -0.00000627, -0.00000808, 0.00000950, 0.00000290, -0.00000929, 0.00000012, 0.00000151, -0.00000200, -0.00000164, 0.00000159, -0.00000429, -0.00000146, 0.00000279, -0.00000452, 0.00000102, -0.00000549, 0.00001334, -0.00000122, -0.00000384, 0.00000197, -0.00000048, 0.00000220, 0.00000021, 0.00000156, -0.00000247, -0.00000223, -0.00000241, -0.00000618, 0.00000278, -0.00000219, -0.00000910, -0.00000315, -0.00001589, -0.00000947, -0.00004888, -0.00003796, 0.00007434, 0.00004317, 0.00002200, 0.00002208, 0.00003162, 0.00002747, 0.00006391, 0.00003799, -0.00007043, -0.00002631, -0.00001098, -0.00001158, -0.00000284, -0.00000105, -0.00000895, -0.00000352, 0.00000009, -0.00000061, -0.00000164, -0.00000230, -0.00000878, -0.00000376, -0.00000134, -0.00000086, -0.00000770, -0.00000494, -0.00000247, -0.00000101, -0.00000061, 0.00000006, -0.00000678, -0.00000701, 0.00000033, -0.00000597, -0.00001408, -0.00000914, -0.00005608, -0.00002468, 0.00007000, 0.00003255, 0.00002911, 0.00002485, 0.00002139, 0.00001885, 0.00005803, 0.00004635, -0.00005358, -0.00003094, -0.00000938, -0.00001431, -0.00000276, -0.00000984, 0.00000436, -0.00000277, -0.00000452, -0.00000104, -0.00000321, -0.00000095, 0.00000146, 0.00000044, 0.00000213, -0.00000059, 0.00000103, -0.00000686, 0.00001092, 0.00000299, -0.00000385, -0.00000085, -0.00000184, 0.00000279, -0.00000562, 0.00000012, 0.00000050, -0.00000239, -0.00000092, 0.00000313, -0.00000445, -0.00000721, 0.00000804, 0.00000492, -0.00001300, -0.00000067, -0.00000576, 0.00000814, -0.00000980, -0.00000367, },
		},
		{ { -0.00000248, -0.00000239, 0.00000150, -0.00000032, 0.00000066, 0.00000012, 0.00000975, -0.00000118, -0.00000179, 0.00000174, -0.00000029, 0.00000138, 0.00000162, 0.00000156, -0.00000392, -0.00000241, 0.00000655, 0.00000032, -0.00000094, 0.00000121, -0.00000170, -0.00000238, -0.00001084, 0.00000127, 0.00000639, -0.00000467, -0.00000168, -0.00000454, -0.00000207, -0.00000128, -0.00000002, -0.00000136, 0.00000008, 0.00000156, -0.00001615, 0.00000172, 0.00000056, -0.00000077, -0.00000822, -0.00000050, -0.00001559, 0.00000025, -0.00000220, -0.00000409, -0.00000715, -0.00001209, 0.00000272, -0.00001924, -0.00006293, 0.00000531, 0.00007698, 0.00002598, 0.00002457, 0.00002336, 0.00002482, 0.00002611, 0.00007279, 0.00000377, -0.00006593, -0.00001677, -0.00001513, -0.00001038, -0.00000289, -0.00000589, -0.00000581, -0.00000589, -0.00000289, -0.00001038, -0.00001513, -0.00001677, -0.00006593, 0.00000377, 0.00007279, 0.00002611, 0.00002482, 0.00002336, 0.00002457, 0.00002598, 0.00007698, 0.00000531, -0.00006293, -0.00001924, 0.00000272, -0.00001209, -0.00000715, -0.00000409, -0.00000220, 0.00000025, -0.00001559, -0.00000050, -0.00000822, -0.00000077, 0.00000056, 0.00000172, -0.00001615, 0.00000156, 0.00000008, -0.00000136, -0.00000002, -0.00000128, -0.00000207, -0.00000454, -0.00000168, -0.00000467, 0.00000639, 0.00000127, -0.00001084, -0.00000238, -0.00000170, 0.00000121, -0.00000094, 0.00000032, 0.00000655, -0.00000241, -0.00000392, 0.00000156, 0.00000162, 0.00000138, -0.00000029, 0.00000174, -0.00000179, -0.00000118, 0.00000975, 0.00000012, 0.00000066, -0.00000032, 0.00000150, -0.00000239, },
		{ -0.00000260, -0.00000231, 0.00000164, 0.00000005, -0.00000014, 0.00000246, 0.00000901, -0.00000341, -0.00000029, 0.00000135, 0.00000037, 0.00000064, 0.00000243, 0.00000076, -0.00000464, -0.00000078, 0.00000660, -0.00000179, 0.00000074, -0.00000010, 0.00000009, -0.00000626, -0.00000898, 0.00000415, 0.00000418, -0.00000470, -0.00000177, -0.00000477, -0.00000136, -0.00000185, 0.00000121, -0.00000264, 0.00000216, -0.00000208, -0.00001465, 0.00000325, 0.00000026, -0.00000236, -0.00000757, -0.00000073, -0.00001619, 0.00000241, -0.00000368, -0.00000323, -0.00000916, -0.00001103, 0.00000428, -0.00003023, -0.00005884, 0.00003191, 0.00006692, 0.00001999, 0.00002455, 0.00002200, 0.00002280, 0.00002991, 0.00007011, -0.00000960, -0.00006132, -0.00001315, -0.00001611, -0.00000781, -0.00000328, -0.00000621, -0.00000590, -0.00000513, -0.00000339, -0.00001245, -0.00001414, -0.00002137, -0.00006919, 0.00001859, 0.00007300, 0.00002333, 0.00002678, 0.00002520, 0.00002358, 0.00003671, 0.00007964, -0.00002425, -0.00005753, -0.00001219, 0.00000139, -0.00001303, -0.00000500, -0.00000527, -0.00000052, -0.00000231, -0.00001429, -0.00000118, -0.00000778, 0.00000015, 0.00000111, -0.00000078, -0.00001607, 0.00000430, -0.00000188, -0.00000008, -0.00000135, -0.00000050, -0.00000307, -0.00000391, -0.00000191, -0.00000421, 0.00000851, -0.00000325, -0.00001001, 0.00000002, -0.00000334, 0.00000265, -0.00000278, 0.00000322, 0.00000518, -0.00000323, -0.00000322, 0.00000214, 0.00000092, 0.00000215, -0.00000124, 0.00000237, -0.00000350, 0.00000217, 0.00000874, -0.00000125, 0.00000126, -0.00000072, 0.00000144, -0.00000236, },
		{ -0.00000293, -0.00000208, 0.00000190, 0.00000014, -0.00000063, 0.00000515, 0.00000676, -0.00000428, 0.00000065, 0.00000150, 0.00000037, 0.00000030, 0.00000313, -0.00000024, -0.00000516, 0.00000142, 0.00000525, -0.00000261, 0.00000157, -0.00000044, 0.00000076, -0.00001017, -0.00000522, 0.00000517, 0.00000240, -0.00000461, -0.00000205, -0.00000456, -0.00000113, -0.00000191, 0.00000193, -0.00000346, 0.00000373, -0.00000586, -0.00001203, 0.00000394, 0.00000000, -0.00000422, -0.00000606, -0.00000185, -0.00001601, 0.00000418, -0.00000517, -0.00000231, -0.00001160, -0.00000880, 0.00000406, -0.00004246, -0.00004543, 0.00005137, 0.00005351, 0.00001778, 0.00002342, 0.00002116, 0.00002088, 0.00003429, 0.00006517, -0.00002123, -0.00005560, -0.00001070, -0.00001666, -0.00000516, -0.00000436, -0.00000608, -0.00000612, -0.00000411, -0.00000475, -0.00001380, -0.00001348, -0.00002672, -0.00007086, 0.00003454, 0.00007059, 0.00002195, 0.00002851, 0.00002731, 0.00002233, 0.00005201, 0.00007133, -0.00005093, -0.00004534, -0.00000973, 0.00000139, -0.00001451, -0.00000233, -0.00000704, 0.00000152, -0.00000526, -0.00001246, -0.00000256, -0.00000630, 0.00000028, 0.00000192, -0.00000412, -0.00001426, 0.00000563, -0.00000316, 0.00000075, -0.00000241, 0.00000014, -0.00000408, -0.00000302, -0.00000250, -0.00000304, 0.00000978, -0.00000857, -0.00000654, 0.00000004, -0.00000363, 0.00000320, -0.00000385, 0.00000599, 0.00000305, -0.00000333, -0.00000259, 0.00000244, 0.00000056, 0.00000257, -0.00000203, 0.00000280, -0.00000481, 0.00000596, 0.00000620, -0.00000137, 0.00000129, -0.00000087, 0.00000135, -0.00000226, },
		{ -0.00000342, -0.00000167, 0.00000227, -0.00000018, -0.00000035, 0.00000739, 0.00000370, -0.00000399, 0.00000105, 0.00000215, -0.00000041, 0.00000064, 0.00000345, -0.00000137, -0.00000521, 0.00000365, 0.00000287, -0.00000216, 0.00000146, 0.00000044, -0.00000051, -0.00001265, -0.00000077, 0.00000464, 0.00000127, -0.00000464, -0.00000233, -0.00000409, -0.00000134, -0.00000138, 0.00000199, -0.00000354, 0.00000431, -0.00000907, -0.00000889, 0.00000403, -0.00000043, -0.00000594, -0.00000407, -0.00000378, -0.00001496, 0.00000539, -0.00000657, -0.00000135, -0.00001457, -0.00000486, 0.00000020, -0.00005221, -0.00002539, 0.00006195, 0.00004009, 0.00001796, 0.00002140, 0.00002081, 0.00001920, 0.00003888, 0.00005819, -0.00003083, -0.00004916, -0.00000945, -0.00001643, -0.00000293, -0.00000574, -0.00000568, -0.00000627, -0.00000317, -0.00000664, -0.00001450, -0.00001327, -0.00003277, -0.00007058, 0.00005105, 0.00006573, 0.00002220, 0.00003001, 0.00002921, 0.00002234, 0.00006987, 0.00005043, -0.00006887, -0.00003104, -0.00001034, 0.00000239, -0.00001629, 0.00000063, -0.00000921, 0.00000379, -0.00000840, -0.00001044, -0.00000410, -0.00000424, -0.00000013, 0.00000267, -0.00000774, -0.00001109, 0.00000553, -0.00000357, 0.00000094, -0.00000300, 0.00000046, -0.00000487, -0.00000204, -0.00000341, -0.00000093, 0.00000938, -0.00001330, -0.00000142, -0.00000230, -0.00000203, 0.00000213, -0.00000331, 0.00000765, 0.00000100, -0.00000311, -0.00000193, 0.00000241, 0.00000064, 0.00000238, -0.00000228, 0.00000262, -0.00000507, 0.00000926, 0.00000282, -0.00000041, 0.00000064, -0.00000061, 0.00000115, -0.00000213, },
		{ -0.00000400, -0.00000105, 0.00000264, -0.00000089, 0.00000092, 0.00000849, 0.00000067, -0.00000306, 0.00000123, 0.00000295, -0.00000176, 0.00000169, 0.00000324, -0.00000251, -0.00000459, 0.00000530, 0.00000014, -0.00000091, 0.00000080, 0.00000214, -0.00000378, -0.00001279, 0.00000314, 0.00000322, 0.00000074, -0.00000494, -0.00000234, -0.00000365, -0.00000174, -0.00000046, 0.00000151, -0.00000298, 0.00000380, -0.00001129, -0.00000568, 0.00000365, -0.00000112, -0.00000723, -0.00000194, -0.00000635, -0.00001287, 0.00000565, -0.00000739, -0.00000089, -0.00001743, 0.00000024, -0.00000785, -0.00005623, -0.00000278, 0.00006408, 0.00002890, 0.00001914, 0.00001893, 0.00002079, 0.00001785, 0.00004332, 0.00004944, -0.00003811, -0.00004252, -0.00000921, -0.00001527, -0.00000154, -0.00000700, -0.00000525, -0.00000612, -0.00000276, -0.00000853, -0.00001494, -0.00001327, -0.00003974, -0.00006770, 0.00006721, 0.00005897, 0.00002394, 0.00003147, 0.00003023, 0.00002577, 0.00008618, 0.00001851, -0.00007421, -0.00001957, -0.00001125, 0.00000297, -0.00001735, 0.00000298, -0.00001107, 0.00000576, -0.00001128, -0.00000876, -0.00000516, -0.00000231, -0.00000056, 0.00000281, -0.00001083, -0.00000738, 0.00000444, -0.00000331, 0.00000063, -0.00000323, 0.00000050, -0.00000542, -0.00000106, -0.00000453, 0.00000211, 0.00000665, -0.00001590, 0.00000363, -0.00000596, 0.00000111, -0.00000050, -0.00000089, 0.00000757, -0.00000025, -0.00000302, -0.00000107, 0.00000200, 0.00000116, 0.00000155, -0.00000186, 0.00000165, -0.00000387, 0.00001120, -0.00000050, 0.00000114, -0.00000055, 0.00000007, 0.00000082, -0.00000203, },
		{ -0.00000455, -0.00000022, 0.00000285, -0.00000172, 0.00000310, 0.00000804, -0.00000157, -0.00000209, 0.00000165, 0.00000336, -0.00000317, 0.00000321, 0.00000248, -0.00000352, -0.00000326, 0.00000592, -0.00000224, 0.00000052, 0.00000025, 0.00000374, -0.00000833, -0.00001039, 0.00000558, 0.00000163, 0.00000055, -0.00000547, -0.00000194, -0.00000354, -0.00000195, 0.00000048, 0.00000083, -0.00000211, 0.00000244, -0.00001237, -0.00000271, 0.00000288, -0.00000196, -0.00000794, 0.00000003, -0.00000936, -0.00000971, 0.00000459, -0.00000702, -0.00000181, -0.00001889, 0.00000479, -0.00001885, -0.00005297, 0.00001845, 0.00005960, 0.00002103, 0.00002017, 0.00001656, 0.00002081, 0.00001707, 0.00004717, 0.00003936, -0.00004295, -0.00003619, -0.00000962, -0.00001325, -0.00000122, -0.00000774, -0.00000509, -0.00000544, -0.00000328, -0.00000986, -0.00001566, -0.00001310, -0.00004800, -0.00006138, 0.00008180, 0.00005124, 0.00002677, 0.00003317, 0.00002988, 0.00003476, 0.00009534, -0.00001928, -0.00006660, -0.00001426, -0.00000994, 0.00000155, -0.00001647, 0.00000366, -0.00001173, 0.00000676, -0.00001342, -0.00000780, -0.00000517, -0.00000115, -0.00000049, 0.00000183, -0.00001258, -0.00000407, 0.00000307, -0.00000283, 0.00000016, -0.00000337, 0.00000049, -0.00000582, -0.00000012, -0.00000567, 0.00000577, 0.00000149, -0.00001518, 0.00000675, -0.00000922, 0.00000455, -0.00000385, 0.00000288, 0.00000576, -0.00000036, -0.00000336, 0.00000010, 0.00000124, 0.00000197, 0.00000025, -0.00000089, 0.00000006, -0.00000118, 0.00001122, -0.00000288, 0.00000259, -0.00000187, 0.00000100, 0.00000038, -0.00000203, },
		{ -0.00000497, 0.00000081, 0.00000273, -0.00000228, 0.00000571, 0.00000611, -0.00000261, -0.00000154, 0.00000268, 0.00000283, -0.00000404, 0.00000477, 0.00000128, -0.00000425, -0.00000138, 0.00000531, -0.00000375, 0.00000166, 0.00000037, 0.00000425, -0.00001293, -0.00000600, 0.00000616, 0.00000050, 0.00000033, -0.00000604, -0.00000121, -0.00000390, -0.00000171, 0.00000113, 0.00000031, -0.00000136, 0.00000064, -0.00001244, -0.00000016, 0.00000173, -0.00000281, -0.00000805, 0.00000150, -0.00001239, -0.00000574, 0.00000218, -0.00000518, -0.00000473, -0.00001774, 0.00000663, -0.00003007, -0.00004299, 0.00003547, 0.00005082, 0.00001675, 0.00002016, 0.00001491, 0.00002043, 0.00001720, 0.00004984, 0.00002863, -0.00004548, -0.00003056, -0.00001027, -0.00001067, -0.00000200, -0.00000773, -0.00000535, -0.00000418, -0.00000492, -0.00001021, -0.00001709, -0.00001246, -0.00005782, -0.00005088, 0.00009359, 0.00004358, 0.00003027, 0.00003512, 0.00002841, 0.00005024, 0.00009182, -0.00005522, -0.00004965, -0.00001548, -0.00000541, -0.00000253, -0.00001309, 0.00000211, -0.00001064, 0.00000637, -0.00001463, -0.00000759, -0.00000404, -0.00000105, 0.00000025, -0.00000035, -0.00001258, -0.00000185, 0.00000201, -0.00000255, -0.00000018, -0.00000365, 0.00000063, -0.00000618, 0.00000069, -0.00000648, 0.00000931, -0.00000543, -0.00001088, 0.00000660, -0.00001041, 0.00000675, -0.00000651, 0.00000682, 0.00000294, 0.00000046, -0.00000411, 0.00000150, 0.00000027, 0.00000285, -0.00000120, 0.00000025, -0.00000167, 0.00000251, 0.00000927, -0.00000378, 0.00000335, -0.00000287, 0.00000195, -0.00000010, -0.00000220, },
		{ -0.00000516, 0.00000198, 0.00000215, -0.00000216, 0.00000807, 0.00000320, -0.00000249, -0.00000148, 0.00000437, 0.00000110, -0.00000383, 0.00000586, -0.00000008, -0.00000459, 0.00000073, 0.00000358, -0.00000421, 0.00000231, 0.00000132, 0.00000295, -0.00001628, -0.00000065, 0.00000500, 0.00000019, -0.00000027, -0.00000629, -0.00000042, -0.00000458, -0.00000101, 0.00000136, 0.00000012, -0.00000104, -0.00000119, -0.00001165, 0.00000179, 0.00000035, -0.00000360, -0.00000752, 0.00000211, -0.00001478, -0.00000171, -0.00000095, -0.00000231, -0.00000941, -0.00001366, 0.00000420, -0.00003838, -0.00002850, 0.00004696, 0.00003998, 0.00001569, 0.00001862, 0.00001451, 0.00001923, 0.00001859, 0.00005072, 0.00001808, -0.00004612, -0.00002581, -0.00001075, -0.00000795, -0.00000360, -0.00000699, -0.00000597, -0.00000255, -0.00000749, -0.00000961, -0.00001923, -0.00001145, -0.00006903, -0.00003591, 0.00010167, 0.00003683, 0.00003423, 0.00003681, 0.00002735, 0.00007073, 0.00007233, -0.00008121, -0.00002966, -0.00002079, 0.00000128, -0.00000853, -0.00000776, -0.00000136, -0.00000795, 0.00000472, -0.00001515, -0.00000768, -0.00000225, -0.00000168, 0.00000132, -0.00000326, -0.00001107, -0.00000087, 0.00000150, -0.00000256, -0.00000037, -0.00000406, 0.00000091, -0.00000645, 0.00000122, -0.00000650, 0.00001167, -0.00001265, -0.00000388, 0.00000299, -0.00000857, 0.00000647, -0.00000712, 0.00000951, 0.00000025, 0.00000155, -0.00000490, 0.00000285, -0.00000070, 0.00000352, -0.00000243, 0.00000113, -0.00000285, 0.00000636, 0.00000588, -0.00000315, 0.00000308, -0.00000320, 0.00000266, -0.00000054, -0.00000256, },
		{ -0.00000506, 0.00000318, 0.00000112, -0.00000107, 0.00000946, 0.00000013, -0.00000166, -0.00000161, 0.00000634, -0.00000167, -0.00000234, 0.00000607, -0.00000130, -0.00000451, 0.00000269, 0.00000109, -0.00000371, 0.00000259, 0.00000284, -0.00000025, -0.00001739, 0.00000444, 0.00000266, 0.00000073, -0.00000148, -0.00000589, 0.00000001, -0.00000523, -0.00000010, 0.00000134, 0.00000015, -0.00000118, -0.00000278, -0.00001019, 0.00000295, -0.00000100, -0.00000438, -0.00000631, 0.00000151, -0.00001582, 0.00000137, -0.00000366, 0.00000038, -0.00001460, -0.00000772, -0.00000251, -0.00004151, -0.00001234, 0.00005280, 0.00002899, 0.00001703, 0.00001555, 0.00001551, 0.00001700, 0.00002141, 0.00004936, 0.00000853, -0.00004546, -0.00002189, -0.00001079, -0.00000561, -0.00000552, -0.00000584, -0.00000666, -0.00000106, -0.00001040, -0.00000857, -0.00002161, -0.00001073, -0.00008079, -0.00001686, 0.00010575, 0.00003140, 0.00003881, 0.00003723, 0.00002950, 0.00009174, 0.00003764, -0.00009151, -0.00001337, -0.00002610, 0.00000766, -0.00001451, -0.00000203, -0.00000557, -0.00000449, 0.00000240, -0.00001557, -0.00000725, -0.00000070, -0.00000229, 0.00000194, -0.00000594, -0.00000889, -0.00000069, 0.00000128, -0.00000263, -0.00000071, -0.00000427, 0.00000109, -0.00000643, 0.00000123, -0.00000533, 0.00001175, -0.00001823, 0.00000387, -0.00000295, -0.00000400, 0.00000346, -0.00000499, 0.00000989, -0.00000120, 0.00000209, -0.00000518, 0.00000375, -0.00000137, 0.00000376, -0.00000314, 0.00000140, -0.00000289, 0.00000937, 0.00000196, -0.00000141, 0.00000185, -0.00000273, 0.00000297, -0.00000089, -0.00000310, },
		{ -0.00000465, 0.00000427, -0.00000021, 0.00000104, 0.00000934, -0.00000229, -0.00000076, -0.00000137, 0.00000789, -0.00000486, 0.00000020, 0.00000528, -0.00000214, -0.00000403, 0.00000416, -0.00000173, -0.00000246, 0.00000279, 0.00000429, -0.00000488, -0.00001586, 0.00000818, -0.00000005, 0.00000184, -0.00000323, -0.00000471, -0.00000027, -0.00000543, 0.00000064, 0.00000135, 0.00000012, -0.00000160, -0.00000404, -0.00000821, 0.00000317, -0.00000204, -0.00000520, -0.00000446, -0.00000041, -0.00001500, 0.00000259, -0.00000476, 0.00000138, -0.00001840, -0.00000211, -0.00001174, -0.00003888, 0.00000308, 0.00005355, 0.00001943, 0.00001950, 0.00001156, 0.00001760, 0.00001395, 0.00002541, 0.00004571, 0.00000055, -0.00004406, -0.00001867, -0.00001024, -0.00000412, -0.00000715, -0.00000478, -0.00000696, -0.00000035, -0.00001284, -0.00000791, -0.00002336, -0.00001142, -0.00009167, 0.00000519, 0.00010613, 0.00002738, 0.00004429, 0.00003535, 0.00003810, 0.00010630, -0.00000660, -0.00008492, -0.00000541, -0.00002776, 0.00001112, -0.00001836, 0.00000231, -0.00000905, -0.00000134, 0.00000013, -0.00001643, -0.00000562, -0.00000027, -0.00000210, 0.00000134, -0.00000737, -0.00000709, -0.00000056, 0.00000088, -0.00000233, -0.00000162, -0.00000381, 0.00000075, -0.00000585, 0.00000056, -0.00000282, 0.00000887, -0.00002040, 0.00000995, -0.00000910, 0.00000182, -0.00000139, -0.00000049, 0.00000773, -0.00000080, 0.00000149, -0.00000452, 0.00000386, -0.00000153, 0.00000345, -0.00000325, 0.00000099, -0.00000151, 0.00001072, -0.00000147, 0.00000069, 0.00000006, -0.00000158, 0.00000285, -0.00000118, -0.00000375, },
		{ -0.00000394, 0.00000511, -0.00000152, 0.00000387, 0.00000757, -0.00000347, -0.00000036, -0.00000025, 0.00000821, -0.00000755, 0.00000317, 0.00000371, -0.00000252, -0.00000322, 0.00000486, -0.00000441, -0.00000075, 0.00000314, 0.00000495, -0.00000999, -0.00001191, 0.00000989, -0.00000230, 0.00000298, -0.00000518, -0.00000291, -0.00000131, -0.00000497, 0.00000095, 0.00000168, -0.00000024, -0.00000207, -0.00000490, -0.00000597, 0.00000250, -0.00000264, -0.00000602, -0.00000230, -0.00000331, -0.00001245, 0.00000169, -0.00000363, -0.00000033, -0.00001919, 0.00000069, -0.00002058, -0.00003167, 0.00001634, 0.00004997, 0.00001260, 0.00002164, 0.00000769, 0.00001994, 0.00001077, 0.00002989, 0.00004022, -0.00000570, -0.00004232, -0.00001592, -0.00000912, -0.00000375, -0.00000798, -0.00000425, -0.00000650, -0.00000095, -0.00001406, -0.00000841, -0.00002361, -0.00001470, -0.00010004, 0.00002882, 0.00010336, 0.00002482, 0.00005054, 0.00003109, 0.00005523, 0.00010689, -0.00005102, -0.00006538, -0.00000665, -0.00002418, 0.00001037, -0.00001889, 0.00000414, -0.00001081, 0.00000076, -0.00000172, -0.00001781, -0.00000264, -0.00000130, -0.00000078, -0.00000079, -0.00000696, -0.00000647, 0.00000015, -0.00000012, -0.00000139, -0.00000338, -0.00000236, -0.00000038, -0.00000457, -0.00000069, 0.00000064, 0.00000319, -0.00001824, 0.00001239, -0.00001311, 0.00000675, -0.00000628, 0.00000502, 0.00000378, 0.00000128, -0.00000031, -0.00000282, 0.00000304, -0.00000111, 0.00000266, -0.00000290, 0.00000017, 0.00000112, 0.00001006, -0.00000358, 0.00000235, -0.00000172, -0.00000005, 0.00000238, -0.00000144, -0.00000442, },
		{ -0.00000294, 0.00000551, -0.00000242, 0.00000684, 0.00000449, -0.00000327, -0.00000068, 0.00000196, 0.00000677, -0.00000886, 0.00000573, 0.00000185, -0.00000255, -0.00000213, 0.00000463, -0.00000657, 0.00000124, 0.00000369, 0.00000427, -0.00001443, -0.00000633, 0.00000938, -0.00000346, 0.00000356, -0.00000678, -0.00000098, -0.00000283, -0.00000395, 0.00000084, 0.00000233, -0.00000107, -0.00000241, -0.00000529, -0.00000387, 0.00000125, -0.00000290, -0.00000659, -0.00000037, -0.00000640, -0.00000899, -0.00000067, -0.00000070, -0.00000462, -0.00001651, -0.00000088, -0.00002611, -0.00002209, 0.00002697, 0.00004292, 0.00000930, 0.00002212, 0.00000511, 0.00002152, 0.00000834, 0.00003391, 0.00003369, -0.00001041, -0.00004048, -0.00001339, -0.00000771, -0.00000444, -0.00000781, -0.00000443, -0.00000523, -0.00000297, -0.00001377, -0.00001044, -0.00002197, -0.00002132, -0.00010454, 0.00005275, 0.00009787, 0.00002422, 0.00005650, 0.00002617, 0.00008003, 0.00008833, -0.00008521, -0.00004056, -0.00001415, -0.00001651, 0.00000608, -0.00001645, 0.00000358, -0.00001086, 0.00000183, -0.00000342, -0.00001910, 0.00000103, -0.00000334, 0.00000124, -0.00000397, -0.00000493, -0.00000716, 0.00000160, -0.00000168, 0.00000007, -0.00000582, -0.00000001, -0.00000226, -0.00000277, -0.00000208, 0.00000411, -0.00000413, -0.00001218, 0.00001041, -0.00001339, 0.00000886, -0.00000921, 0.00000964, -0.00000050, 0.00000416, -0.00000281, -0.00000042, 0.00000146, -0.00000024, 0.00000160, -0.00000239, -0.00000055, 0.00000436, 0.00000762, -0.00000403, 0.00000300, -0.00000293, 0.00000146, 0.00000169, -0.00000171, -0.00000504, },
		{ -0.00000166, 0.00000538, -0.00000249, 0.00000918, 0.00000083, -0.00000204, -0.00000149, 0.00000496, 0.00000354, -0.00000826, 0.00000714, 0.00000028, -0.00000246, -0.00000078, 0.00000338, -0.00000786, 0.00000334, 0.00000428, 0.00000209, -0.00001722, -0.00000023, 0.00000702, -0.00000326, 0.00000315, -0.00000747, 0.00000048, -0.00000430, -0.00000275, 0.00000057, 0.00000310, -0.00000223, -0.00000259, -0.00000510, -0.00000237, -0.00000007, -0.00000313, -0.00000657, 0.00000064, -0.00000861, -0.00000595, -0.00000307, 0.00000257, -0.00001005, -0.00001153, -0.00000658, -0.00002669, -0.00001230, 0.00003490, 0.00003353, 0.00000957, 0.00002021, 0.00000465, 0.00002153, 0.00000743, 0.00003657, 0.00002702, -0.00001400, -0.00003858, -0.00001083, -0.00000646, -0.00000575, -0.00000689, -0.00000512, -0.00000351, -0.00000600, -0.00001229, -0.00001370, -0.00001875, -0.00003125, -0.00010449, 0.00007614, 0.00008974, 0.00002659, 0.00006018, 0.00002437, 0.00010757, 0.00005037, -0.00010128, -0.00001905, -0.00002279, -0.00000787, 0.00000051, -0.00001275, 0.00000189, -0.00001008, 0.00000256, -0.00000573, -0.00001918, 0.00000412, -0.00000531, 0.00000293, -0.00000702, -0.00000226, -0.00000855, 0.00000333, -0.00000333, 0.00000144, -0.00000829, 0.00000270, -0.00000443, -0.00000087, -0.00000298, 0.00000629, -0.00001109, -0.00000404, 0.00000485, -0.00000978, 0.00000729, -0.00000884, 0.00001175, -0.00000351, 0.00000661, -0.00000512, 0.00000203, -0.00000042, 0.00000081, 0.00000054, -0.00000207, -0.00000056, 0.00000725, 0.00000417, -0.00000304, 0.00000249, -0.00000325, 0.00000266, 0.00000090, -0.00000199, -0.00000558, },
		{ -0.00000013, 0.00000467, -0.00000149, 0.00001019, -0.00000248, -0.00000048, -0.00000215, 0.00000797, -0.00000086, -0.00000580, 0.00000705, -0.00000053, -0.00000250, 0.00000076, 0.00000113, -0.00000803, 0.00000538, 0.00000461, -0.00000138, -0.00001770, 0.00000519, 0.00000358, -0.00000191, 0.00000169, -0.00000698, 0.00000097, -0.00000518, -0.00000185, 0.00000056, 0.00000361, -0.00000348, -0.00000262, -0.00000431, -0.00000181, -0.00000098, -0.00000363, -0.00000574, 0.00000032, -0.00000915, -0.00000456, -0.00000396, 0.00000432, -0.00001443, -0.00000672, -0.00001432, -0.00002253, -0.00000357, 0.00004000, 0.00002331, 0.00001257, 0.00001605, 0.00000642, 0.00001970, 0.00000832, 0.00003736, 0.00002092, -0.00001688, -0.00003660, -0.00000804, -0.00000592, -0.00000698, -0.00000575, -0.00000580, -0.00000203, -0.00000921, -0.00001043, -0.00001733, -0.00001493, -0.00004375, -0.00009983, 0.00009855, 0.00007886, 0.00003303, 0.00005929, 0.00003060, 0.00012922, -0.00000100, -0.00009683, -0.00000704, -0.00002777, -0.00000162, -0.00000371, -0.00000987, 0.00000074, -0.00000973, 0.00000382, -0.00000945, -0.00001700, 0.00000533, -0.00000597, 0.00000307, -0.00000857, -0.00000032, -0.00000962, 0.00000457, -0.00000432, 0.00000192, -0.00000991, 0.00000495, -0.00000627, 0.00000059, -0.00000283, 0.00000606, -0.00001556, 0.00000356, -0.00000209, -0.00000369, 0.00000267, -0.00000509, 0.00001059, -0.00000417, 0.00000760, -0.00000639, 0.00000379, -0.00000209, 0.00000173, -0.00000032, -0.00000212, 0.00000056, 0.00000890, 0.00000074, -0.00000129, 0.00000112, -0.00000266, 0.00000336, 0.00000008, -0.00000224, -0.00000605, },
		{ 0.00000158, 0.00000352, 0.00000054, 0.00000946, -0.00000463, 0.00000068, -0.00000197, 0.00000993, -0.00000533, -0.00000220, 0.00000562, -0.00000048, -0.00000276, 0.00000236, -0.00000194, -0.00000690, 0.00000713, 0.00000436, -0.00000557, -0.00001571, 0.00000892, 0.00000005, 0.00000000, -0.00000045, -0.00000546, 0.00000038, -0.00000519, -0.00000157, 0.00000109, 0.00000355, -0.00000457, -0.00000247, -0.00000314, -0.00000222, -0.00000127, -0.00000446, -0.00000422, -0.00000121, -0.00000791, -0.00000531, -0.00000245, 0.00000323, -0.00001583, -0.00000474, -0.00002096, -0.00001545, 0.00000392, 0.00004187, 0.00001409, 0.00001673, 0.00001062, 0.00000978, 0.00001643, 0.00001073, 0.00003629, 0.00001575, -0.00001939, -0.00003444, -0.00000500, -0.00000644, -0.00000750, -0.00000501, -0.00000595, -0.00000148, -0.00001175, -0.00000906, -0.00002036, -0.00001171, -0.00005772, -0.00009075, 0.00011948, 0.00006553, 0.00004393, 0.00005265, 0.00004880, 0.00013496, -0.00005454, -0.00007595, -0.00000607, -0.00002683, 0.00000044, -0.00000507, -0.00000908, 0.00000116, -0.00001055, 0.00000598, -0.00001467, -0.00001225, 0.00000402, -0.00000467, 0.00000101, -0.00000769, -0.00000021, -0.00000946, 0.00000468, -0.00000410, 0.00000091, -0.00000996, 0.00000604, -0.00000723, 0.00000133, -0.00000155, 0.00000301, -0.00001603, 0.00000813, -0.00000764, 0.00000245, -0.00000311, 0.00000078, 0.00000666, -0.00000232, 0.00000671, -0.00000615, 0.00000435, -0.00000313, 0.00000231, -0.00000090, -0.00000245, 0.00000283, 0.00000872, -0.00000169, 0.00000034, -0.00000055, -0.00000142, 0.00000356, -0.00000072, -0.00000242, -0.00000646, },
		{ 0.00000334, 0.00000222, 0.00000323, 0.00000703, -0.00000520, 0.00000098, -0.00000046, 0.00000995, -0.00000860, 0.00000146, 0.00000353, 0.00000017, -0.00000310, 0.00000369, -0.00000545, -0.00000446, 0.00000832, 0.00000331, -0.00000968, -0.00001162, 0.00001038, -0.00000259, 0.00000167, -0.00000256, -0.00000351, -0.00000093, -0.00000448, -0.00000186, 0.00000218, 0.00000283, -0.00000540, -0.00000196, -0.00000207, -0.00000321, -0.00000113, -0.00000530, -0.00000262, -0.00000315, -0.00000571, -0.00000755, 0.00000106, -0.00000067, -0.00001373, -0.00000713, -0.00002370, -0.00000791, 0.00001067, 0.00004010, 0.00000756, 0.00002022, 0.00000535, 0.00001362, 0.00001255, 0.00001395, 0.00003379, 0.00001155, -0.00002173, -0.00003188, -0.00000198, -0.00000802, -0.00000700, -0.00000498, -0.00000533, -0.00000215, -0.00001316, -0.00000869, -0.00002222, -0.00000993, -0.00007232, -0.00007720, 0.00013790, 0.00005111, 0.00005805, 0.00004164, 0.00007943, 0.00011694, -0.00009706, -0.00004782, -0.00001288, -0.00002115, -0.00000104, -0.00000401, -0.00001006, 0.00000289, -0.00001222, 0.00000844, -0.00002053, -0.00000571, 0.00000064, -0.00000174, -0.00000290, -0.00000444, -0.00000222, -0.00000777, 0.00000352, -0.00000269, -0.00000159, -0.00000827, 0.00000573, -0.00000719, 0.00000149, 0.00000023, -0.00000213, -0.00001240, 0.00000832, -0.00000965, 0.00000616, -0.00000762, 0.00000666, 0.00000148, 0.00000124, 0.00000432, -0.00000454, 0.00000363, -0.00000336, 0.00000250, -0.00000135, -0.00000269, 0.00000578, 0.00000672, -0.00000257, 0.00000116, -0.00000186, 0.00000007, 0.00000335, -0.00000147, -0.00000254, -0.00000677, },
		{ 0.00000491, 0.00000116, 0.00000591, 0.00000347, -0.00000434, 0.00000045, 0.00000230, 0.00000768, -0.00000972, 0.00000405, 0.00000166, 0.00000088, -0.00000321, 0.00000431, -0.00000880, -0.00000093, 0.00000870, 0.00000143, -0.00001293, -0.00000621, 0.00000951, -0.00000375, 0.00000238, -0.00000387, -0.00000191, -0.00000229, -0.00000356, -0.00000235, 0.00000352, 0.00000164, -0.00000602, -0.00000084, -0.00000171, -0.00000409, -0.00000104, -0.00000564, -0.00000174, -0.00000433, -0.00000402, -0.00000969, 0.00000494, -0.00000581, -0.00000944, -0.00001342, -0.00002130, -0.00000178, 0.00001706, 0.00003487, 0.00000467, 0.00002166, 0.00000150, 0.00001686, 0.00000890, 0.00001726, 0.00003041, 0.00000819, -0.00002406, -0.00002861, 0.00000044, -0.00001011, -0.00000574, -0.00000547, -0.00000426, -0.00000372, -0.00001365, -0.00000909, -0.00002311, -0.00000970, -0.00008729, -0.00005852, 0.00015194, 0.00003820, 0.00007221, 0.00003084, 0.00011757, 0.00007321, -0.00011789, -0.00002316, -0.00002121, -0.00001447, -0.00000331, -0.00000267, -0.00001100, 0.00000443, -0.00001343, 0.00000980, -0.00002529, 0.00000085, -0.00000335, 0.00000149, -0.00000730, 0.00000007, -0.00000566, -0.00000502, 0.00000166, -0.00000075, -0.00000489, -0.00000535, 0.00000436, -0.00000649, 0.00000166, 0.00000132, -0.00000759, -0.00000614, 0.00000454, -0.00000745, 0.00000598, -0.00000893, 0.00001041, -0.00000303, 0.00000507, 0.00000137, -0.00000223, 0.00000200, -0.00000294, 0.00000242, -0.00000190, -0.00000230, 0.00000858, 0.00000353, -0.00000195, 0.00000084, -0.00000238, 0.00000144, 0.00000283, -0.00000209, -0.00000267, -0.00000687, },
		{ 0.00000608, 0.00000076, 0.00000781, -0.00000033, -0.00000264, -0.00000036, 0.00000560, 0.00000354, -0.00000845, 0.00000488, 0.00000077, 0.00000110, -0.00000274, 0.00000376, -0.00001126, 0.00000317, 0.00000811, -0.00000114, -0.00001463, -0.00000057, 0.00000685, -0.00000336, 0.00000184, -0.00000390, -0.00000134, -0.00000301, -0.00000303, -0.00000251, 0.00000468, 0.00000026, -0.00000649, 0.00000097, -0.00000253, -0.00000424, -0.00000150, -0.00000505, -0.00000219, -0.00000375, -0.00000430, -0.00000989, 0.00000706, -0.00000972, -0.00000569, -0.00002111, -0.00001468, 0.00000244, 0.00002271, 0.00002725, 0.00000521, 0.00002060, -0.00000031, 0.00001890, 0.00000595, 0.00002017, 0.00002659, 0.00000554, -0.00002651, -0.00002426, 0.00000148, -0.00001180, -0.00000445, -0.00000585, -0.00000349, -0.00000538, -0.00001400, -0.00000948, -0.00002386, -0.00001047, -0.00010288, -0.00003371, 0.00015929, 0.00003010, 0.00008213, 0.00002723, 0.00015291, 0.00000976, -0.00011286, -0.00001007, -0.00002501, -0.00001096, -0.00000316, -0.00000371, -0.00000963, 0.00000385, -0.00001257, 0.00000850, -0.00002711, 0.00000542, -0.00000612, 0.00000330, -0.00001036, 0.00000411, -0.00000923, -0.00000223, 0.00000003, 0.00000068, -0.00000788, -0.00000219, 0.00000266, -0.00000573, 0.00000250, 0.00000044, -0.00001112, 0.00000015, -0.00000110, -0.00000220, 0.00000214, -0.00000643, 0.00001077, -0.00000532, 0.00000781, -0.00000102, -0.00000013, 0.00000012, -0.00000219, 0.00000226, -0.00000272, -0.00000080, 0.00001025, 0.00000016, -0.00000042, -0.00000044, -0.00000198, 0.00000246, 0.00000206, -0.00000248, -0.00000298, -0.00000659, },
		{ 0.00000670, 0.00000126, 0.00000827, -0.00000339, -0.00000092, -0.00000066, 0.00000830, -0.00000132, -0.00000537, 0.00000393, 0.00000117, 0.00000051, -0.00000154, 0.00000179, -0.00001214, 0.00000717, 0.00000658, -0.00000411, -0.00001438, 0.00000416, 0.00000339, -0.00000191, 0.00000030, -0.00000271, -0.00000202, -0.00000268, -0.00000333, -0.00000195, 0.00000528, -0.00000110, -0.00000670, 0.00000324, -0.00000456, -0.00000338, -0.00000266, -0.00000355, -0.00000394, -0.00000125, -0.00000711, -0.00000711, 0.00000584, -0.00001024, -0.00000540, -0.00002671, -0.00000635, 0.00000558, 0.00002651, 0.00001914, 0.00000789, 0.00001764, -0.00000031, 0.00001975, 0.00000369, 0.00002265, 0.00002250, 0.00000355, -0.00002910, -0.00001868, 0.00000052, -0.00001218, -0.00000396, -0.00000540, -0.00000382, -0.00000623, -0.00001504, -0.00000900, -0.00002541, -0.00001164, -0.00011919, -0.00000216, 0.00015821, 0.00002950, 0.00008421, 0.00003766, 0.00017214, -0.00006010, -0.00008609, -0.00001101, -0.00002131, -0.00001287, 0.00000119, -0.00000865, -0.00000456, -0.00000007, -0.00000874, 0.00000378, -0.00002495, 0.00000666, -0.00000640, 0.00000247, -0.00001062, 0.00000609, -0.00001162, -0.00000036, -0.00000055, 0.00000071, -0.00000956, 0.00000028, 0.00000133, -0.00000535, 0.00000428, -0.00000309, -0.00001095, 0.00000373, -0.00000566, 0.00000353, -0.00000353, -0.00000110, 0.00000785, -0.00000485, 0.00000869, -0.00000208, 0.00000098, -0.00000140, -0.00000145, 0.00000214, -0.00000381, 0.00000195, 0.00001007, -0.00000231, 0.00000112, -0.00000209, -0.00000087, 0.00000308, 0.00000107, -0.00000251, -0.00000361, -0.00000575, },
		{ 0.00000678, 0.00000260, 0.00000703, -0.00000505, 0.00000014, 0.00000024, 0.00000924, -0.00000541, -0.00000171, 0.00000190, 0.00000258, -0.00000077, 0.00000021, -0.00000151, -0.00001099, 0.00001032, 0.00000429, -0.00000708, -0.00001213, 0.00000708, 0.00000024, -0.00000027, -0.00000148, -0.00000090, -0.00000359, -0.00000147, -0.00000443, -0.00000063, 0.00000524, -0.00000241, -0.00000633, 0.00000530, -0.00000727, -0.00000180, -0.00000415, -0.00000178, -0.00000610, 0.00000221, -0.00001171, -0.00000177, 0.00000131, -0.00000680, -0.00001017, -0.00002730, 0.00000081, 0.00000904, 0.00002717, 0.00001267, 0.00001086, 0.00001400, 0.00000071, 0.00001983, 0.00000180, 0.00002488, 0.00001816, 0.00000208, -0.00003165, -0.00001214, -0.00000251, -0.00001080, -0.00000470, -0.00000380, -0.00000559, -0.00000587, -0.00001709, -0.00000738, -0.00002805, -0.00001326, -0.00013540, 0.00003543, 0.00014862, 0.00003696, 0.00007780, 0.00006542, 0.00016345, -0.00011945, -0.00004865, -0.00002212, -0.00001160, -0.00001917, 0.00000893, -0.00001679, 0.00000374, -0.00000697, -0.00000242, -0.00000372, -0.00001924, 0.00000464, -0.00000415, -0.00000103, -0.00000774, 0.00000533, -0.00001219, 0.00000022, 0.00000012, -0.00000087, -0.00000951, 0.00000162, 0.00000064, -0.00000528, 0.00000650, -0.00000876, -0.00000666, 0.00000294, -0.00000663, 0.00000692, -0.00000835, 0.00000487, 0.00000304, -0.00000220, 0.00000782, -0.00000172, 0.00000081, -0.00000217, -0.00000093, 0.00000200, -0.00000484, 0.00000560, 0.00000788, -0.00000316, 0.00000180, -0.00000340, 0.00000053, 0.00000336, -0.00000009, -0.00000215, -0.00000463, -0.00000432, },
		{ 0.00000654, 0.00000435, 0.00000434, -0.00000513, 0.00000033, 0.00000249, 0.00000778, -0.00000746, 0.00000112, -0.00000009, 0.00000427, -0.00000223, 0.00000191, -0.00000556, -0.00000784, 0.00001204, 0.00000153, -0.00000957, -0.00000830, 0.00000775, -0.00000169, 0.00000066, -0.00000259, 0.00000059, -0.00000523, 0.00000003, -0.00000590, 0.00000116, 0.00000475, -0.00000382, -0.00000493, 0.00000627, -0.00000970, -0.00000025, -0.00000521, -0.00000077, -0.00000723, 0.00000482, -0.00001611, 0.00000416, -0.00000468, -0.00000101, -0.00001920, -0.00002199, 0.00000501, 0.00001367, 0.00002410, 0.00000929, 0.00001253, 0.00001088, 0.00000198, 0.00001957, 0.00000004, 0.00002701, 0.00001366, 0.00000090, -0.00003361, -0.00000545, -0.00000695, -0.00000797, -0.00000643, -0.00000147, -0.00000832, -0.00000479, -0.00001951, -0.00000541, -0.00003098, -0.00001660, -0.00014920, 0.00007645, 0.00013280, 0.00005007, 0.00006660, 0.00010726, 0.00012139, -0.00015332, -0.00001444, -0.00003531, -0.00000091, -0.00002597, 0.00001678, -0.00002530, 0.00001287, -0.00001486, 0.00000450, -0.00001198, -0.00001187, 0.00000084, -0.00000070, -0.00000588, -0.00000273, 0.00000234, -0.00001124, -0.00000004, 0.00000143, -0.00000343, -0.00000818, 0.00000209, 0.00000023, -0.00000487, 0.00000788, -0.00001479, 0.00000038, -0.00000202, -0.00000302, 0.00000607, -0.00000996, 0.00000907, -0.00000162, 0.00000124, 0.00000610, -0.00000054, -0.00000037, -0.00000221, -0.00000061, 0.00000164, -0.00000526, 0.00000934, 0.00000417, -0.00000226, 0.00000119, -0.00000382, 0.00000179, 0.00000339, -0.00000131, -0.00000152, -0.00000594, -0.00000239, },
		{ 0.00000634, 0.00000583, 0.00000095, -0.00000404, 0.00000003, 0.00000555, 0.00000408, -0.00000687, 0.00000211, -0.00000100, 0.00000536, -0.00000320, 0.00000281, -0.00000938, -0.00000324, 0.00001208, -0.00000141, -0.00001108, -0.00000367, 0.00000637, -0.00000205, 0.00000034, -0.00000234, 0.00000086, -0.00000603, 0.00000100, -0.00000707, 0.00000290, 0.00000412, -0.00000547, -0.00000221, 0.00000539, -0.00001086, 0.00000043, -0.00000513, -0.00000141, -0.00000603, 0.00000470, -0.00001795, 0.00000798, -0.00000911, 0.00000385, -0.00002941, -0.00001235, 0.00000635, 0.00001899, 0.00001801, 0.00000918, 0.00001232, 0.00000882, 0.00000317, 0.00001895, -0.00000145, 0.00002885, 0.00000930, -0.00000048, -0.00003423, 0.00000013, -0.00001148, -0.00000476, -0.00000819, 0.00000050, -0.00001081, -0.00000424, -0.00002086, -0.00000479, -0.00003238, -0.00002392, -0.00015710, 0.00011675, 0.00011495, 0.00006398, 0.00005831, 0.00015259, 0.00005032, -0.00015407, 0.00000509, -0.00004218, 0.00000479, -0.00002845, 0.00002053, -0.00003038, 0.00001953, -0.00002098, 0.00000950, -0.00001839, -0.00000540, -0.00000251, 0.00000189, -0.00000995, 0.00000248, -0.00000148, -0.00000981, -0.00000015, 0.00000223, -0.00000576, -0.00000660, 0.00000251, -0.00000064, -0.00000319, 0.00000683, -0.00001866, 0.00000738, -0.00000894, 0.00000405, 0.00000095, -0.00000735, 0.00000985, -0.00000432, 0.00000392, 0.00000471, 0.00000048, -0.00000188, -0.00000184, -0.00000032, 0.00000083, -0.00000450, 0.00001213, -0.00000004, -0.00000016, -0.00000063, -0.00000315, 0.00000267, 0.00000320, -0.00000241, -0.00000087, -0.00000727, -0.00000023, },
		{ 0.00000654, 0.00000633, -0.00000217, -0.00000255, -0.00000000, 0.00000830, -0.00000077, -0.00000405, 0.00000100, -0.00000026, 0.00000527, -0.00000317, 0.00000223, -0.00001192, 0.00000187, 0.00001058, -0.00000431, -0.00001119, 0.00000069, 0.00000371, -0.00000110, -0.00000114, -0.00000062, -0.00000046, -0.00000542, 0.00000083, -0.00000740, 0.00000418, 0.00000353, -0.00000723, 0.00000164, 0.00000246, -0.00001026, -0.00000015, -0.00000369, -0.00000386, -0.00000209, 0.00000089, -0.00001567, 0.00000755, -0.00000922, 0.00000418, -0.00003656, -0.00000181, 0.00000666, 0.00002327, 0.00001096, 0.00001109, 0.00001075, 0.00000751, 0.00000458, 0.00001748, -0.00000204, 0.00002986, 0.00000562, -0.00000255, -0.00003276, 0.00000333, -0.00001458, -0.00000253, -0.00000874, 0.00000081, -0.00001163, -0.00000564, -0.00001955, -0.00000740, -0.00003029, -0.00003763, -0.00015542, 0.00015184, 0.00009967, 0.00007341, 0.00006198, 0.00018564, -0.00003558, -0.00012435, 0.00000522, -0.00003787, 0.00000171, -0.00002340, 0.00001721, -0.00002915, 0.00002108, -0.00002311, 0.00001069, -0.00002098, -0.00000189, -0.00000351, 0.00000182, -0.00001128, 0.00000591, -0.00000457, -0.00000900, 0.00000084, 0.00000147, -0.00000669, -0.00000585, 0.00000369, -0.00000257, 0.00000030, 0.00000227, -0.00001813, 0.00001114, -0.00001450, 0.00001174, -0.00000654, -0.00000133, 0.00000703, -0.00000424, 0.00000491, 0.00000456, 0.00000044, -0.00000294, -0.00000149, 0.00000011, -0.00000050, -0.00000223, 0.00001302, -0.00000352, 0.00000217, -0.00000306, -0.00000158, 0.00000310, 0.00000275, -0.00000313, -0.00000054, -0.00000828, 0.00000184, },
		{ 0.00000730, 0.00000537, -0.00000419, -0.00000144, 0.00000102, 0.00000942, -0.00000515, -0.00000028, -0.00000156, 0.00000194, 0.00000397, -0.00000214, -0.00000002, -0.00001234, 0.00000641, 0.00000795, -0.00000693, -0.00000968, 0.00000378, 0.00000089, 0.00000034, -0.00000309, 0.00000200, -0.00000308, -0.00000344, -0.00000060, -0.00000671, 0.00000488, 0.00000289, -0.00000860, 0.00000584, -0.00000195, -0.00000813, -0.00000170, -0.00000147, -0.00000726, 0.00000366, -0.00000594, -0.00000945, 0.00000256, -0.00000403, -0.00000210, -0.00003716, 0.00000596, 0.00000837, 0.00002427, 0.00000554, 0.00001299, 0.00000915, 0.00000609, 0.00000685, 0.00001442, -0.00000094, 0.00002936, 0.00000309, -0.00000559, -0.00002896, 0.00000345, -0.00001521, -0.00000227, -0.00000727, -0.00000137, -0.00000996, -0.00000976, -0.00001480, -0.00001420, -0.00002372, -0.00005892, -0.00014178, 0.00017862, 0.00008982, 0.00007563, 0.00008382, 0.00019052, -0.00011597, -0.00007632, -0.00001058, -0.00002325, -0.00000967, -0.00001102, 0.00000674, -0.00002115, 0.00001683, -0.00002075, 0.00000781, -0.00001939, -0.00000192, -0.00000165, -0.00000141, -0.00000906, 0.00000647, -0.00000602, -0.00000931, 0.00000321, -0.00000115, -0.00000573, -0.00000642, 0.00000591, -0.00000553, 0.00000525, -0.00000561, -0.00001234, 0.00000947, -0.00001554, 0.00001662, -0.00001334, 0.00000578, 0.00000194, -0.00000185, 0.00000421, 0.00000579, -0.00000107, -0.00000304, -0.00000148, 0.00000070, -0.00000214, 0.00000141, 0.00001158, -0.00000528, 0.00000368, -0.00000526, 0.00000040, 0.00000321, 0.00000200, -0.00000332, -0.00000076, -0.00000865, 0.00000361, },
		{ 0.00000840, 0.00000299, -0.00000472, -0.00000109, 0.00000345, 0.00000799, -0.00000746, 0.00000280, -0.00000430, 0.00000477, 0.00000202, -0.00000062, -0.00000353, -0.00001040, 0.00000948, 0.00000471, -0.00000905, -0.00000668, 0.00000494, -0.00000104, 0.00000126, -0.00000448, 0.00000445, -0.00000610, -0.00000072, -0.00000289, -0.00000526, 0.00000521, 0.00000188, -0.00000885, 0.00000919, -0.00000665, -0.00000538, -0.00000335, 0.00000032, -0.00000991, 0.00000906, -0.00001352, -0.00000146, -0.00000504, 0.00000484, -0.00001432, -0.00003009, 0.00000880, 0.00001308, 0.00002055, 0.00000380, 0.00001301, 0.00000880, 0.00000376, 0.00001043, 0.00000933, 0.00000240, 0.00002690, 0.00000191, -0.00000936, -0.00002331, 0.00000074, -0.00001327, -0.00000404, -0.00000394, -0.00000577, -0.00000611, -0.00001606, -0.00000732, -0.00002447, -0.00001358, -0.00008684, -0.00011621, 0.00019664, 0.00008498, 0.00007276, 0.00012300, 0.00015744, -0.00017096, -0.00002714, -0.00003267, -0.00000432, -0.00002475, 0.00000489, -0.00000780, -0.00000868, 0.00000848, -0.00001540, 0.00000248, -0.00001516, -0.00000425, 0.00000193, -0.00000664, -0.00000407, 0.00000445, -0.00000604, -0.00001027, 0.00000622, -0.00000489, -0.00000344, -0.00000785, 0.00000859, -0.00000863, 0.00001020, -0.00001503, -0.00000245, 0.00000222, -0.00001059, 0.00001617, -0.00001646, 0.00001117, -0.00000317, 0.00000139, 0.00000279, 0.00000777, -0.00000375, -0.00000213, -0.00000187, 0.00000129, -0.00000357, 0.00000573, 0.00000805, -0.00000493, 0.00000366, -0.00000647, 0.00000224, 0.00000316, 0.00000097, -0.00000297, -0.00000162, -0.00000829, 0.00000507, },
		{ 0.00000933, -0.00000028, -0.00000396, -0.00000129, 0.00000688, 0.00000397, -0.00000681, 0.00000382, -0.00000586, 0.00000710, 0.00000029, 0.00000044, -0.00000728, -0.00000656, 0.00001068, 0.00000132, -0.00001032, -0.00000277, 0.00000409, -0.00000149, 0.00000095, -0.00000443, 0.00000564, -0.00000841, 0.00000179, -0.00000530, -0.00000355, 0.00000553, 0.00000018, -0.00000736, 0.00001054, -0.00001026, -0.00000319, -0.00000403, 0.00000043, -0.00001003, 0.00001164, -0.00001886, 0.00000502, -0.00001168, 0.00001358, -0.00002896, -0.00001732, 0.00000705, 0.00002047, 0.00001240, 0.00000628, 0.00001038, 0.00001020, 0.00000039, 0.00001512, 0.00000251, 0.00000786, 0.00002257, 0.00000180, -0.00001313, -0.00001709, -0.00000355, -0.00000977, -0.00000687, 0.00000010, -0.00001102, -0.00000160, -0.00002270, 0.00000066, -0.00003577, -0.00000275, -0.00011812, -0.00008128, 0.00020817, 0.00008147, 0.00007212, 0.00016973, 0.00008770, -0.00018806, 0.00000747, -0.00004974, 0.00001083, -0.00003693, 0.00001863, -0.00002145, 0.00000415, -0.00000077, -0.00000981, -0.00000255, -0.00001096, -0.00000646, 0.00000492, -0.00001152, 0.00000161, 0.00000133, -0.00000575, -0.00001065, 0.00000842, -0.00000819, -0.00000122, -0.00000893, 0.00001052, -0.00001042, 0.00001304, -0.00002303, 0.00000846, -0.00000836, -0.00000055, 0.00000989, -0.00001433, 0.00001267, -0.00000616, 0.00000371, 0.00000201, 0.00000929, -0.00000675, -0.00000062, -0.00000244, 0.00000157, -0.00000414, 0.00000966, 0.00000339, -0.00000283, 0.00000202, -0.00000631, 0.00000354, 0.00000307, -0.00000020, -0.00000227, -0.00000294, -0.00000731, 0.00000643, },
		{ 0.00000938, -0.00000350, -0.00000260, -0.00000131, 0.00001019, -0.00000164, -0.00000342, 0.00000227, -0.00000539, 0.00000806, -0.00000044, 0.00000012, -0.00000997, -0.00000186, 0.00001012, -0.00000194, -0.00001035, 0.00000113, 0.00000182, -0.00000057, -0.00000066, -0.00000265, 0.00000495, -0.00000917, 0.00000325, -0.00000717, -0.00000202, 0.00000601, -0.00000217, -0.00000404, 0.00000928, -0.00001184, -0.00000236, -0.00000317, -0.00000169, -0.00000671, 0.00000978, -0.00001959, 0.00000703, -0.00001375, 0.00001771, -0.00004091, -0.00000312, 0.00000337, 0.00002824, 0.00000202, 0.00001171, 0.00000577, 0.00001278, -0.00000330, 0.00001987, -0.00000496, 0.00001459, 0.00001697, 0.00000215, -0.00001592, -0.00001184, -0.00000763, -0.00000634, -0.00000920, 0.00000308, -0.00001518, 0.00000149, -0.00002724, 0.00000619, -0.00004474, 0.00000486, -0.00014824, -0.00004089, 0.00021668, 0.00007438, 0.00008372, 0.00020684, -0.00000506, -0.00016675, 0.00001904, -0.00005391, 0.00001595, -0.00004079, 0.00002524, -0.00002961, 0.00001337, -0.00000775, -0.00000653, -0.00000482, -0.00000927, -0.00000625, 0.00000504, -0.00001362, 0.00000564, -0.00000108, -0.00000650, -0.00000918, 0.00000835, -0.00000944, -0.00000061, -0.00000833, 0.00001048, -0.00000963, 0.00001187, -0.00002656, 0.00001653, -0.00001838, 0.00001143, -0.00000037, -0.00000746, 0.00000974, -0.00000589, 0.00000392, 0.00000302, 0.00000917, -0.00000902, 0.00000085, -0.00000289, 0.00000132, -0.00000330, 0.00001206, -0.00000111, 0.00000008, -0.00000076, -0.00000484, 0.00000418, 0.00000292, -0.00000127, -0.00000155, -0.00000436, -0.00000598, 0.00000801, },
		{ 0.00000797, -0.00000574, -0.00000150, -0.00000027, 0.00001190, -0.00000706, 0.00000144, -0.00000131, -0.00000291, 0.00000744, 0.00000006, -0.00000204, -0.00001052, 0.00000241, 0.00000831, -0.00000496, -0.00000884, 0.00000397, -0.00000080, 0.00000098, -0.00000296, 0.00000041, 0.00000251, -0.00000820, 0.00000330, -0.00000816, -0.00000082, 0.00000650, -0.00000467, 0.00000043, 0.00000574, -0.00001134, -0.00000287, -0.00000111, -0.00000542, -0.00000055, 0.00000362, -0.00001524, 0.00000347, -0.00000938, 0.00001414, -0.00004567, 0.00000777, 0.00000143, 0.00003310, -0.00000730, 0.00001755, 0.00000091, 0.00001515, -0.00000600, 0.00002321, -0.00001151, 0.00002129, 0.00001099, 0.00000242, -0.00001706, -0.00000878, -0.00000997, -0.00000445, -0.00000968, 0.00000359, -0.00001672, 0.00000154, -0.00002784, 0.00000697, -0.00004867, 0.00000586, -0.00017317, 0.00000167, 0.00022440, 0.00006092, 0.00011549, 0.00021520, -0.00009826, -0.00011866, 0.00000897, -0.00004401, 0.00000964, -0.00003475, 0.00002292, -0.00003024, 0.00001698, -0.00001098, -0.00000651, -0.00000348, -0.00001106, -0.00000267, 0.00000132, -0.00001163, 0.00000653, -0.00000162, -0.00000897, -0.00000537, 0.00000543, -0.00000783, -0.00000246, -0.00000533, 0.00000799, -0.00000597, 0.00000608, -0.00002384, 0.00001858, -0.00002380, 0.00002119, -0.00001109, 0.00000176, 0.00000359, -0.00000268, 0.00000192, 0.00000613, 0.00000676, -0.00000974, 0.00000171, -0.00000299, 0.00000051, -0.00000092, 0.00001221, -0.00000421, 0.00000260, -0.00000378, -0.00000257, 0.00000425, 0.00000265, -0.00000203, -0.00000116, -0.00000553, -0.00000457, 0.00001001, },
		{ 0.00000496, -0.00000640, -0.00000118, 0.00000236, 0.00001082, -0.00001038, 0.00000586, -0.00000551, 0.00000071, 0.00000577, 0.00000130, -0.00000572, -0.00000849, 0.00000518, 0.00000593, -0.00000761, -0.00000575, 0.00000493, -0.00000263, 0.00000212, -0.00000488, 0.00000366, -0.00000086, -0.00000599, 0.00000220, -0.00000840, 0.00000028, 0.00000654, -0.00000639, 0.00000476, 0.00000111, -0.00000958, -0.00000392, 0.00000097, -0.00000906, 0.00000632, -0.00000470, -0.00000765, -0.00000416, 0.00000041, 0.00000278, -0.00004127, 0.00001223, 0.00000399, 0.00003235, -0.00001248, 0.00002121, -0.00000228, 0.00001592, -0.00000655, 0.00002385, -0.00001571, 0.00002675, 0.00000532, 0.00000245, -0.00001663, -0.00000813, -0.00001002, -0.00000462, -0.00000794, 0.00000131, -0.00001528, -0.00000166, -0.00002429, 0.00000248, -0.00004673, -0.00000108, -0.00019088, 0.00004546, 0.00023021, 0.00004369, 0.00016792, 0.00018119, -0.00016769, -0.00006310, -0.00001286, -0.00002561, -0.00000409, -0.00002183, 0.00001380, -0.00002463, 0.00001577, -0.00001129, -0.00000855, 0.00000016, -0.00001519, 0.00000319, -0.00000525, -0.00000613, 0.00000436, -0.00000041, -0.00001266, 0.00000002, 0.00000035, -0.00000386, -0.00000642, -0.00000037, 0.00000369, -0.00000051, -0.00000310, -0.00001530, 0.00001358, -0.00002213, 0.00002520, -0.00001851, 0.00000999, -0.00000328, 0.00000188, -0.00000126, 0.00001066, 0.00000225, -0.00000866, 0.00000168, -0.00000272, -0.00000051, 0.00000258, 0.00001001, -0.00000519, 0.00000375, -0.00000608, -0.00000013, 0.00000399, 0.00000220, -0.00000237, -0.00000126, -0.00000621, -0.00000317, 0.00001233, },
		{ 0.00000078, -0.00000553, -0.00000160, 0.00000635, 0.00000662, -0.00001032, 0.00000808, -0.00000867, 0.00000412, 0.00000410, 0.00000220, -0.00000981, -0.00000433, 0.00000594, 0.00000349, -0.00000964, -0.00000149, 0.00000374, -0.00000290, 0.00000198, -0.00000538, 0.00000586, -0.00000395, -0.00000351, 0.00000059, -0.00000835, 0.00000167, 0.00000560, -0.00000643, 0.00000749, -0.00000303, -0.00000788, -0.00000429, 0.00000159, -0.00001049, 0.00001105, -0.00001190, -0.00000029, -0.00001217, 0.00001182, -0.00001309, -0.00002932, 0.00001014, 0.00001153, 0.00002526, -0.00001204, 0.00002112, -0.00000269, 0.00001438, -0.00000452, 0.00002129, -0.00001679, 0.00003026, 0.00000023, 0.00000267, -0.00001558, -0.00000897, -0.00000847, -0.00000617, -0.00000487, -0.00000274, -0.00001203, -0.00000668, -0.00001826, -0.00000562, -0.00004049, -0.00001485, -0.00020185, 0.00009227, 0.00022921, 0.00003171, 0.00023080, 0.00010337, -0.00019611, -0.00001953, -0.00003316, -0.00000793, -0.00001794, -0.00000807, 0.00000285, -0.00001668, 0.00001269, -0.00001128, -0.00000985, 0.00000319, -0.00001888, 0.00000867, -0.00001205, 0.00000069, 0.00000074, 0.00000122, -0.00001604, 0.00000522, -0.00000500, 0.00000078, -0.00001098, 0.00000511, -0.00000077, 0.00000450, -0.00001279, -0.00000369, 0.00000324, -0.00001351, 0.00002201, -0.00002011, 0.00001437, -0.00000822, 0.00000567, -0.00000393, 0.00001517, -0.00000340, -0.00000611, 0.00000084, -0.00000228, -0.00000117, 0.00000628, 0.00000614, -0.00000410, 0.00000313, -0.00000704, 0.00000191, 0.00000363, 0.00000162, -0.00000240, -0.00000177, -0.00000644, -0.00000163, 0.00001449, },
		{ -0.00000364, -0.00000377, -0.00000207, 0.00001063, 0.00000014, -0.00000677, 0.00000718, -0.00000956, 0.00000617, 0.00000345, 0.00000158, -0.00001278, 0.00000075, 0.00000490, 0.00000121, -0.00001062, 0.00000304, 0.00000090, -0.00000160, 0.00000032, -0.00000397, 0.00000619, -0.00000576, -0.00000166, -0.00000085, -0.00000840, 0.00000357, 0.00000353, -0.00000438, 0.00000767, -0.00000545, -0.00000730, -0.00000315, -0.00000015, -0.00000833, 0.00001142, -0.00001497, 0.00000327, -0.00001638, 0.00001988, -0.00002805, -0.00001431, 0.00000434, 0.00002185, 0.00001350, -0.00000661, 0.00001739, -0.00000042, 0.00001085, -0.00000050, 0.00001604, -0.00001489, 0.00003171, -0.00000437, 0.00000379, -0.00001532, -0.00000966, -0.00000686, -0.00000757, -0.00000218, -0.00000665, -0.00000898, -0.00001117, -0.00001256, -0.00001432, -0.00003303, -0.00003269, -0.00020794, 0.00014508, 0.00021481, 0.00003808, 0.00028411, -0.00000471, -0.00017972, -0.00000015, -0.00004142, 0.00000092, -0.00002507, 0.00000053, -0.00000467, -0.00001069, 0.00001104, -0.00001369, -0.00000763, 0.00000265, -0.00001922, 0.00001082, -0.00001605, 0.00000598, -0.00000213, 0.00000148, -0.00001730, 0.00000817, -0.00000842, 0.00000395, -0.00001421, 0.00000929, -0.00000354, 0.00000663, -0.00001948, 0.00000683, -0.00000853, -0.00000081, 0.00001290, -0.00001565, 0.00001362, -0.00000952, 0.00000700, -0.00000448, 0.00001801, -0.00000880, -0.00000290, -0.00000043, -0.00000198, -0.00000086, 0.00000905, 0.00000182, -0.00000169, 0.00000101, -0.00000651, 0.00000324, 0.00000325, 0.00000105, -0.00000239, -0.00000240, -0.00000646, 0.00000042, 0.00001572, },
		{ -0.00000720, -0.00000204, -0.00000161, 0.00001358, -0.00000678, -0.00000091, 0.00000347, -0.00000792, 0.00000636, 0.00000430, -0.00000131, -0.00001319, 0.00000519, 0.00000277, -0.00000098, -0.00001005, 0.00000667, -0.00000252, 0.00000059, -0.00000231, -0.00000101, 0.00000460, -0.00000593, -0.00000091, -0.00000177, -0.00000862, 0.00000577, 0.00000079, -0.00000073, 0.00000536, -0.00000579, -0.00000807, -0.00000057, -0.00000394, -0.00000276, 0.00000699, -0.00001274, 0.00000118, -0.00001409, 0.00002061, -0.00003681, -0.00000168, -0.00000075, 0.00003114, 0.00000054, 0.00000143, 0.00001152, 0.00000334, 0.00000650, 0.00000422, 0.00000935, -0.00001075, 0.00003131, -0.00000854, 0.00000623, -0.00001694, -0.00000871, -0.00000669, -0.00000738, -0.00000144, -0.00000873, -0.00000790, -0.00001312, -0.00000966, -0.00002085, -0.00002726, -0.00005219, -0.00021014, 0.00020517, 0.00018283, 0.00007449, 0.00030367, -0.00011728, -0.00012975, -0.00000565, -0.00003452, -0.00000205, -0.00002256, 0.00000108, -0.00000594, -0.00000907, 0.00001250, -0.00001960, -0.00000081, -0.00000281, -0.00001470, 0.00000800, -0.00001528, 0.00000771, -0.00000262, -0.00000076, -0.00001540, 0.00000768, -0.00000848, 0.00000411, -0.00001468, 0.00001096, -0.00000358, 0.00000445, -0.00002059, 0.00001240, -0.00001713, 0.00001152, 0.00000132, -0.00000717, 0.00000849, -0.00000703, 0.00000533, -0.00000210, 0.00001803, -0.00001268, 0.00000008, -0.00000172, -0.00000198, 0.00000076, 0.00000994, -0.00000162, 0.00000084, -0.00000180, -0.00000486, 0.00000387, 0.00000287, 0.00000068, -0.00000265, -0.00000277, -0.00000658, 0.00000330, 0.00001525, },
		{ -0.00000905, -0.00000111, 0.00000056, 0.00001367, -0.00001189, 0.00000518, -0.00000163, -0.00000454, 0.00000516, 0.00000624, -0.00000631, -0.00001037, 0.00000760, 0.00000049, -0.00000322, -0.00000761, 0.00000832, -0.00000519, 0.00000247, -0.00000475, 0.00000243, 0.00000185, -0.00000485, -0.00000109, -0.00000236, -0.00000863, 0.00000762, -0.00000160, 0.00000324, 0.00000164, -0.00000478, -0.00000944, 0.00000229, -0.00000813, 0.00000428, -0.00000045, -0.00000655, -0.00000565, -0.00000558, 0.00001308, -0.00003658, 0.00000461, -0.00000136, 0.00003569, -0.00000993, 0.00000913, 0.00000557, 0.00000707, 0.00000260, 0.00000829, 0.00000253, -0.00000522, 0.00002920, -0.00001189, 0.00000955, -0.00002046, -0.00000570, -0.00000844, -0.00000520, -0.00000306, -0.00000856, -0.00000909, -0.00001221, -0.00001017, -0.00002436, -0.00002408, -0.00007314, -0.00020637, 0.00026916, 0.00013585, 0.00014435, 0.00027005, -0.00020442, -0.00006813, -0.00002588, -0.00001768, -0.00001328, -0.00001294, -0.00000471, -0.00000202, -0.00001107, 0.00001602, -0.00002749, 0.00000907, -0.00001199, -0.00000625, 0.00000087, -0.00000993, 0.00000563, -0.00000061, -0.00000525, -0.00001081, 0.00000410, -0.00000524, 0.00000113, -0.00001226, 0.00001021, -0.00000134, -0.00000154, -0.00001570, 0.00001109, -0.00001920, 0.00001921, -0.00000853, 0.00000189, 0.00000135, -0.00000214, 0.00000143, 0.00000284, 0.00001493, -0.00001428, 0.00000217, -0.00000277, -0.00000216, 0.00000350, 0.00000860, -0.00000322, 0.00000237, -0.00000429, -0.00000273, 0.00000400, 0.00000245, 0.00000057, -0.00000335, -0.00000265, -0.00000692, 0.00000703, 0.00001269, },
		{ -0.00000895, -0.00000117, 0.00000449, 0.00001019, -0.00001345, 0.00000932, -0.00000620, -0.00000090, 0.00000373, 0.00000806, -0.00001223, -0.00000466, 0.00000730, -0.00000123, -0.00000547, -0.00000344, 0.00000744, -0.00000607, 0.00000296, -0.00000577, 0.00000496, -0.00000078, -0.00000350, -0.00000152, -0.00000312, -0.00000786, 0.00000832, -0.00000245, 0.00000585, -0.00000176, -0.00000386, -0.00001005, 0.00000376, -0.00001041, 0.00000974, -0.00000752, 0.00000015, -0.00001371, 0.00000569, 0.00000002, -0.00002836, 0.00000380, 0.00000387, 0.00003368, -0.00001547, 0.00001432, 0.00000107, 0.00000978, -0.00000017, 0.00001100, -0.00000361, 0.00000121, 0.00002512, -0.00001334, 0.00001228, -0.00002453, -0.00000165, -0.00001116, -0.00000210, -0.00000586, -0.00000740, -0.00001099, -0.00001027, -0.00001231, -0.00002658, -0.00002176, -0.00009823, -0.00019076, 0.00032803, 0.00008540, 0.00023777, 0.00017746, -0.00024256, -0.00001894, -0.00004522, -0.00000142, -0.00002460, -0.00000282, -0.00001144, 0.00000280, -0.00001316, 0.00001824, -0.00003371, 0.00001829, -0.00002147, 0.00000301, -0.00000772, -0.00000228, 0.00000139, 0.00000245, -0.00001030, -0.00000538, -0.00000068, -0.00000037, -0.00000362, -0.00000820, 0.00000841, 0.00000122, -0.00000886, -0.00000698, 0.00000388, -0.00001405, 0.00001995, -0.00001341, 0.00000821, -0.00000490, 0.00000292, -0.00000296, 0.00000902, 0.00000943, -0.00001357, 0.00000312, -0.00000354, -0.00000212, 0.00000668, 0.00000542, -0.00000273, 0.00000224, -0.00000565, -0.00000076, 0.00000388, 0.00000202, 0.00000062, -0.00000437, -0.00000209, -0.00000730, 0.00001122, 0.00000824, },
		{ -0.00000737, -0.00000168, 0.00000925, 0.00000368, -0.00001089, 0.00001012, -0.00000865, 0.00000155, 0.00000337, 0.00000824, -0.00001725, 0.00000254, 0.00000452, -0.00000214, -0.00000736, 0.00000175, 0.00000423, -0.00000486, 0.00000155, -0.00000462, 0.00000551, -0.00000212, -0.00000288, -0.00000146, -0.00000456, -0.00000589, 0.00000743, -0.00000102, 0.00000584, -0.00000332, -0.00000435, -0.00000874, 0.00000255, -0.00000892, 0.00001091, -0.00001080, 0.00000344, -0.00001849, 0.00001457, -0.00001320, -0.00001652, -0.00000141, 0.00001324, 0.00002598, -0.00001588, 0.00001640, -0.00000160, 0.00001143, -0.00000202, 0.00001253, -0.00000897, 0.00000850, 0.00001855, -0.00001156, 0.00001234, -0.00002690, 0.00000133, -0.00001284, -0.00000022, -0.00000759, -0.00000764, -0.00001086, -0.00001049, -0.00001268, -0.00003095, -0.00001683, -0.00013176, -0.00015538, 0.00036921, 0.00005013, 0.00033107, 0.00003873, -0.00022341, 0.00000098, -0.00005005, 0.00000402, -0.00002759, 0.00000059, -0.00001292, 0.00000332, -0.00001108, 0.00001531, -0.00003422, 0.00002260, -0.00002714, 0.00000917, -0.00001399, 0.00000429, -0.00000235, 0.00000433, -0.00001366, -0.00000159, -0.00000408, 0.00000364, -0.00000793, -0.00000445, 0.00000754, 0.00000156, -0.00001397, 0.00000152, -0.00000563, -0.00000400, 0.00001427, -0.00001229, 0.00000986, -0.00000808, 0.00000609, -0.00000594, 0.00001461, 0.00000287, -0.00001116, 0.00000306, -0.00000417, -0.00000135, 0.00000926, 0.00000143, -0.00000069, 0.00000049, -0.00000556, 0.00000063, 0.00000368, 0.00000172, 0.00000052, -0.00000535, -0.00000145, -0.00000724, 0.00001501, 0.00000276, },
		{ -0.00000527, -0.00000164, 0.00001321, -0.00000409, -0.00000514, 0.00000751, -0.00000832, 0.00000202, 0.00000488, 0.00000564, -0.00001947, 0.00000930, 0.00000021, -0.00000248, -0.00000828, 0.00000673, -0.00000027, -0.00000213, -0.00000138, -0.00000146, 0.00000383, -0.00000165, -0.00000355, -0.00000056, -0.00000678, -0.00000283, 0.00000519, 0.00000246, 0.00000298, -0.00000244, -0.00000673, -0.00000532, -0.00000138, -0.00000339, 0.00000682, -0.00000850, 0.00000084, -0.00001667, 0.00001654, -0.00002092, -0.00000680, -0.00000630, 0.00002285, 0.00001564, -0.00001289, 0.00001625, -0.00000310, 0.00001271, -0.00000383, 0.00001366, -0.00001389, 0.00001669, 0.00000925, -0.00000563, 0.00000805, -0.00002548, 0.00000119, -0.00001151, -0.00000158, -0.00000616, -0.00001135, -0.00000638, -0.00001566, -0.00000824, -0.00004051, -0.00000661, -0.00017667, -0.00009402, 0.00038159, 0.00004954, 0.00039248, -0.00011630, -0.00015749, -0.00001160, -0.00003509, -0.00000595, -0.00001806, -0.00000660, -0.00000552, -0.00000356, -0.00000241, 0.00000521, -0.00002680, 0.00001942, -0.00002629, 0.00000947, -0.00001506, 0.00000700, -0.00000334, 0.00000330, -0.00001369, -0.00000120, -0.00000412, 0.00000474, -0.00000994, -0.00000257, 0.00000896, -0.00000213, -0.00001391, 0.00000566, -0.00001272, 0.00000659, 0.00000527, -0.00000669, 0.00000701, -0.00000751, 0.00000635, -0.00000625, 0.00001805, -0.00000325, -0.00000796, 0.00000234, -0.00000478, 0.00000052, 0.00001025, -0.00000204, 0.00000181, -0.00000219, -0.00000424, 0.00000134, 0.00000342, 0.00000172, -0.00000010, -0.00000578, -0.00000121, -0.00000617, 0.00001738, -0.00000251, },
		{ -0.00000359, -0.00000007, 0.00001457, -0.00001069, 0.00000172, 0.00000274, -0.00000581, 0.00000079, 0.00000804, 0.00000008, -0.00001772, 0.00001374, -0.00000429, -0.00000271, -0.00000759, 0.00001018, -0.00000459, 0.00000088, -0.00000465, 0.00000266, 0.00000065, 0.00000026, -0.00000531, 0.00000098, -0.00000931, 0.00000061, 0.00000264, 0.00000675, -0.00000171, 0.00000021, -0.00001030, -0.00000084, -0.00000653, 0.00000443, -0.00000107, -0.00000157, -0.00000719, -0.00000797, 0.00001004, -0.00001984, -0.00000363, -0.00000649, 0.00002856, 0.00000639, -0.00000908, 0.00001540, -0.00000439, 0.00001437, -0.00000636, 0.00001506, -0.00001852, 0.00002532, -0.00000209, 0.00000403, -0.00000082, -0.00001957, -0.00000287, -0.00000653, -0.00000674, -0.00000110, -0.00001880, 0.00000266, -0.00002620, 0.00000149, -0.00005545, 0.00000814, -0.00023124, -0.00000659, 0.00036166, 0.00009522, 0.00039252, -0.00024928, -0.00007058, -0.00004569, -0.00000572, -0.00002771, 0.00000142, -0.00002260, 0.00000957, -0.00001694, 0.00001176, -0.00001073, -0.00001253, 0.00000926, -0.00001905, 0.00000380, -0.00001042, 0.00000506, -0.00000104, -0.00000078, -0.00001045, -0.00000424, -0.00000048, 0.00000237, -0.00000913, -0.00000278, 0.00001247, -0.00000969, -0.00000781, 0.00000303, -0.00001359, 0.00001316, -0.00000281, 0.00000021, 0.00000169, -0.00000418, 0.00000404, -0.00000375, 0.00001857, -0.00000777, -0.00000484, 0.00000131, -0.00000533, 0.00000346, 0.00000910, -0.00000386, 0.00000355, -0.00000478, -0.00000231, 0.00000154, 0.00000309, 0.00000212, -0.00000147, -0.00000530, -0.00000174, -0.00000370, 0.00001756, -0.00000643, },
		{ -0.00000284, 0.00000338, 0.00001223, -0.00001395, 0.00000722, -0.00000216, -0.00000262, -0.00000095, 0.00001163, -0.00000741, -0.00001200, 0.00001473, -0.00000783, -0.00000320, -0.00000499, 0.00001110, -0.00000734, 0.00000279, -0.00000679, 0.00000615, -0.00000263, 0.00000251, -0.00000735, 0.00000240, -0.00001132, 0.00000338, 0.00000120, 0.00000997, -0.00000629, 0.00000298, -0.00001347, 0.00000277, -0.00001033, 0.00001130, -0.00000925, 0.00000651, -0.00001719, 0.00000434, -0.00000249, -0.00001071, -0.00000804, -0.00000012, 0.00002799, 0.00000084, -0.00000643, 0.00001495, -0.00000592, 0.00001647, -0.00000959, 0.00001665, -0.00002229, 0.00003305, -0.00001355, 0.00001535, -0.00001251, -0.00001050, -0.00000960, 0.00000070, -0.00001411, 0.00000573, -0.00002766, 0.00001346, -0.00003913, 0.00001328, -0.00007187, 0.00002182, -0.00028736, 0.00009848, 0.00031774, 0.00018317, 0.00031582, -0.00032608, 0.00000584, -0.00008153, 0.00002482, -0.00005099, 0.00002250, -0.00004047, 0.00002660, -0.00003193, 0.00002686, -0.00002792, 0.00000429, -0.00000418, -0.00000860, -0.00000507, -0.00000227, 0.00000005, 0.00000312, -0.00000620, -0.00000581, -0.00000883, 0.00000520, -0.00000216, -0.00000666, -0.00000368, 0.00001607, -0.00001866, 0.00000231, -0.00000570, -0.00000712, 0.00001296, -0.00000644, 0.00000497, -0.00000333, -0.00000011, 0.00000061, 0.00000063, 0.00001636, -0.00001022, -0.00000236, 0.00000016, -0.00000555, 0.00000695, 0.00000595, -0.00000353, 0.00000365, -0.00000633, -0.00000049, 0.00000153, 0.00000271, 0.00000280, -0.00000354, -0.00000384, -0.00000304, 0.00000006, 0.00001533, -0.00000836, },
		{ -0.00000278, 0.00000810, 0.00000632, -0.00001281, 0.00000952, -0.00000527, -0.00000042, -0.00000163, 0.00001388, -0.00001485, -0.00000360, 0.00001217, -0.00000983, -0.00000393, -0.00000082, 0.00000926, -0.00000776, 0.00000268, -0.00000668, 0.00000760, -0.00000452, 0.00000382, -0.00000862, 0.00000289, -0.00001202, 0.00000463, 0.00000201, 0.00001040, -0.00000879, 0.00000408, -0.00001456, 0.00000368, -0.00001036, 0.00001392, -0.00001372, 0.00001124, -0.00002415, 0.00001472, -0.00001542, 0.00000189, -0.00001715, 0.00001119, 0.00002168, -0.00000061, -0.00000539, 0.00001481, -0.00000708, 0.00001799, -0.00001241, 0.00001739, -0.00002389, 0.00003787, -0.00002231, 0.00002500, -0.00002370, -0.00000132, -0.00001608, 0.00000707, -0.00002033, 0.00001068, -0.00003363, 0.00002098, -0.00004887, 0.00002101, -0.00008271, 0.00002504, -0.00033185, 0.00020543, 0.00026958, 0.00029090, 0.00016927, -0.00032894, 0.00004648, -0.00009959, 0.00004191, -0.00006377, 0.00003490, -0.00005126, 0.00003784, -0.00004188, 0.00003685, -0.00004043, 0.00001789, -0.00001565, 0.00000031, -0.00001272, 0.00000550, -0.00000488, 0.00000642, -0.00001025, -0.00000267, -0.00001195, 0.00001004, -0.00000636, -0.00000477, -0.00000294, 0.00001675, -0.00002509, 0.00001216, -0.00001684, 0.00000447, 0.00000625, -0.00000429, 0.00000546, -0.00000576, 0.00000265, -0.00000217, 0.00000531, 0.00001241, -0.00001080, -0.00000069, -0.00000107, -0.00000500, 0.00001007, 0.00000166, -0.00000134, 0.00000191, -0.00000632, 0.00000067, 0.00000153, 0.00000238, 0.00000344, -0.00000588, -0.00000176, -0.00000469, 0.00000446, 0.00001121, -0.00000841, },
		{ -0.00000252, 0.00001252, -0.00000160, -0.00000779, 0.00000811, -0.00000557, -0.00000032, -0.00000013, 0.00001321, -0.00001998, 0.00000537, 0.00000699, -0.00001037, -0.00000447, 0.00000399, 0.00000533, -0.00000604, 0.00000055, -0.00000408, 0.00000639, -0.00000414, 0.00000327, -0.00000841, 0.00000198, -0.00001112, 0.00000420, 0.00000530, 0.00000735, -0.00000810, 0.00000248, -0.00001278, 0.00000122, -0.00000567, 0.00001055, -0.00001187, 0.00000932, -0.00002386, 0.00001782, -0.00002237, 0.00001146, -0.00002548, 0.00002313, 0.00001274, 0.00000030, -0.00000486, 0.00001388, -0.00000653, 0.00001729, -0.00001310, 0.00001578, -0.00002184, 0.00003784, -0.00002574, 0.00002964, -0.00003078, 0.00000448, -0.00001895, 0.00000915, -0.00002183, 0.00000997, -0.00003244, 0.00002011, -0.00004963, 0.00001828, -0.00008072, 0.00000832, -0.00035081, 0.00029663, 0.00024222, 0.00038200, -0.00001710, -0.00026326, 0.00004161, -0.00008925, 0.00003664, -0.00005819, 0.00003148, -0.00004838, 0.00003737, -0.00004170, 0.00003714, -0.00004367, 0.00002364, -0.00002076, 0.00000357, -0.00001516, 0.00000920, -0.00000670, 0.00000639, -0.00001061, -0.00000346, -0.00001097, 0.00001136, -0.00000783, -0.00000546, 0.00000134, 0.00001203, -0.00002526, 0.00001688, -0.00002512, 0.00001651, -0.00000384, 0.00000228, 0.00000168, -0.00000477, 0.00000296, -0.00000296, 0.00000877, 0.00000805, -0.00001015, 0.00000029, -0.00000241, -0.00000332, 0.00001182, -0.00000251, 0.00000174, -0.00000115, -0.00000481, 0.00000098, 0.00000162, 0.00000231, 0.00000358, -0.00000786, 0.00000029, -0.00000599, 0.00000846, 0.00000631, -0.00000735, },
		{ -0.00000101, 0.00001475, -0.00000908, -0.00000086, 0.00000396, -0.00000340, -0.00000227, 0.00000365, 0.00000898, -0.00002110, 0.00001263, 0.00000071, -0.00000998, -0.00000413, 0.00000811, 0.00000065, -0.00000320, -0.00000266, 0.00000019, 0.00000297, -0.00000159, 0.00000079, -0.00000667, -0.00000016, -0.00000906, 0.00000287, 0.00001005, 0.00000156, -0.00000455, -0.00000152, -0.00000878, -0.00000354, 0.00000249, 0.00000206, -0.00000397, 0.00000039, -0.00001515, 0.00001115, -0.00001916, 0.00001263, -0.00002760, 0.00003070, 0.00000529, 0.00000082, -0.00000310, 0.00001086, -0.00000304, 0.00001299, -0.00001027, 0.00001077, -0.00001540, 0.00003217, -0.00002259, 0.00002743, -0.00003148, 0.00000470, -0.00001624, 0.00000504, -0.00001675, 0.00000186, -0.00002215, 0.00000839, -0.00003846, 0.00000185, -0.00006240, -0.00003330, -0.00033554, 0.00036035, 0.00025557, 0.00041754, -0.00019821, -0.00015582, -0.00000003, -0.00005327, 0.00001024, -0.00003462, 0.00001197, -0.00003104, 0.00002413, -0.00003044, 0.00002703, -0.00003676, 0.00002032, -0.00001818, -0.00000015, -0.00001089, 0.00000721, -0.00000415, 0.00000230, -0.00000677, -0.00000871, -0.00000505, 0.00000810, -0.00000563, -0.00000929, 0.00000936, 0.00000146, -0.00001757, 0.00001327, -0.00002596, 0.00002379, -0.00001267, 0.00000994, -0.00000439, -0.00000123, 0.00000100, -0.00000145, 0.00001012, 0.00000439, -0.00000902, 0.00000081, -0.00000378, -0.00000045, 0.00001147, -0.00000528, 0.00000442, -0.00000449, -0.00000238, 0.00000059, 0.00000178, 0.00000265, 0.00000280, -0.00000885, 0.00000167, -0.00000624, 0.00001102, 0.00000189, -0.00000616, },
		{ 0.00000233, 0.00001334, -0.00001360, 0.00000536, -0.00000086, -0.00000021, -0.00000511, 0.00000863, 0.00000190, -0.00001770, 0.00001655, -0.00000517, -0.00000927, -0.00000238, 0.00001034, -0.00000325, -0.00000069, -0.00000547, 0.00000459, -0.00000132, 0.00000212, -0.00000287, -0.00000405, -0.00000275, -0.00000683, 0.00000206, 0.00001435, -0.00000501, 0.00000023, -0.00000636, -0.00000441, -0.00000808, 0.00001092, -0.00000814, 0.00000669, -0.00001236, -0.00000085, -0.00000338, -0.00000594, 0.00000366, -0.00002072, 0.00003070, 0.00000239, -0.00000118, 0.00000099, 0.00000526, 0.00000353, 0.00000497, -0.00000376, 0.00000255, -0.00000531, 0.00002190, -0.00001388, 0.00001899, -0.00002608, -0.00000039, -0.00000853, -0.00000437, -0.00000627, -0.00001200, -0.00000479, -0.00001217, -0.00001736, -0.00002598, -0.00003078, -0.00009679, -0.00028707, 0.00039753, 0.00031378, 0.00037066, -0.00032841, -0.00004493, -0.00005590, -0.00000614, -0.00002634, -0.00000173, -0.00001659, -0.00000485, 0.00000264, -0.00001208, 0.00001030, -0.00002310, 0.00001073, -0.00001033, -0.00000864, -0.00000166, 0.00000078, 0.00000153, -0.00000430, -0.00000056, -0.00001655, 0.00000419, 0.00000148, -0.00000095, -0.00001473, 0.00001902, -0.00001261, -0.00000366, 0.00000145, -0.00001756, 0.00002286, -0.00001613, 0.00001492, -0.00000983, 0.00000283, -0.00000188, 0.00000161, 0.00000936, 0.00000202, -0.00000798, 0.00000103, -0.00000495, 0.00000325, 0.00000892, -0.00000588, 0.00000558, -0.00000701, 0.00000012, -0.00000014, 0.00000194, 0.00000341, 0.00000100, -0.00000853, 0.00000202, -0.00000515, 0.00001154, -0.00000112, -0.00000554, },
		{ 0.00000714, 0.00000804, -0.00001364, 0.00000863, -0.00000418, 0.00000212, -0.00000710, 0.00001291, -0.00000617, -0.00001071, 0.00001658, -0.00000967, -0.00000855, 0.00000078, 0.00001007, -0.00000529, 0.00000032, -0.00000650, 0.00000746, -0.00000477, 0.00000548, -0.00000651, -0.00000152, -0.00000488, -0.00000547, 0.00000317, 0.00001612, -0.00001001, 0.00000407, -0.00001000, -0.00000187, -0.00000958, 0.00001583, -0.00001551, 0.00001513, -0.00002351, 0.00001319, -0.00001999, 0.00001254, -0.00001259, -0.00000608, 0.00002316, 0.00000468, -0.00000602, 0.00000705, -0.00000194, 0.00001177, -0.00000525, 0.00000498, -0.00000720, 0.00000620, 0.00000983, -0.00000272, 0.00000740, -0.00001741, -0.00000795, 0.00000099, -0.00001552, 0.00000554, -0.00002677, 0.00001393, -0.00003528, 0.00000682, -0.00005746, 0.00000469, -0.00017109, -0.00021658, 0.00042362, 0.00039908, 0.00023887, -0.00037738, 0.00003392, -0.00009990, 0.00003297, -0.00005759, 0.00002739, -0.00004296, 0.00002062, -0.00001905, 0.00000632, -0.00000644, -0.00000884, 0.00000039, -0.00000225, -0.00001716, 0.00000827, -0.00000654, 0.00000720, -0.00001019, 0.00000458, -0.00002336, 0.00001332, -0.00000548, 0.00000344, -0.00001877, 0.00002657, -0.00002563, 0.00001183, -0.00001485, -0.00000190, 0.00001353, -0.00001241, 0.00001469, -0.00001211, 0.00000528, -0.00000394, 0.00000478, 0.00000731, 0.00000091, -0.00000726, 0.00000099, -0.00000551, 0.00000699, 0.00000478, -0.00000428, 0.00000472, -0.00000790, 0.00000196, -0.00000083, 0.00000211, 0.00000435, -0.00000158, -0.00000701, 0.00000146, -0.00000298, 0.00001011, -0.00000249, -0.00000549, },
		{ 0.00001206, 0.00000009, -0.00000930, 0.00000797, -0.00000463, 0.00000218, -0.00000675, 0.00001458, -0.00001279, -0.00000215, 0.00001333, -0.00001254, -0.00000765, 0.00000471, 0.00000759, -0.00000526, -0.00000055, -0.00000515, 0.00000778, -0.00000612, 0.00000724, -0.00000906, 0.00000014, -0.00000599, -0.00000539, 0.00000678, 0.00001418, -0.00001174, 0.00000531, -0.00001103, -0.00000247, -0.00000633, 0.00001460, -0.00001650, 0.00001698, -0.00002777, 0.00002058, -0.00003131, 0.00002889, -0.00002987, 0.00001159, 0.00001131, 0.00001018, -0.00001214, 0.00001330, -0.00000857, 0.00001910, -0.00001495, 0.00001341, -0.00001593, 0.00001620, -0.00000059, 0.00000689, -0.00000317, -0.00000955, -0.00001391, 0.00000800, -0.00002367, 0.00001353, -0.00003646, 0.00002691, -0.00005290, 0.00002517, -0.00008246, 0.00003164, -0.00024128, -0.00014047, 0.00046351, 0.00047412, 0.00004848, -0.00034112, 0.00006106, -0.00011372, 0.00004918, -0.00007071, 0.00004140, -0.00005692, 0.00003628, -0.00003315, 0.00001798, -0.00001690, 0.00000006, -0.00000521, 0.00000096, -0.00002077, 0.00001427, -0.00001082, 0.00000952, -0.00001217, 0.00000528, -0.00002550, 0.00001865, -0.00000946, 0.00000456, -0.00001836, 0.00002828, -0.00003267, 0.00002303, -0.00002958, 0.00001591, -0.00000098, -0.00000269, 0.00000901, -0.00001021, 0.00000485, -0.00000381, 0.00000659, 0.00000514, 0.00000053, -0.00000680, 0.00000067, -0.00000505, 0.00000976, 0.00000018, -0.00000122, 0.00000207, -0.00000697, 0.00000272, -0.00000125, 0.00000239, 0.00000505, -0.00000433, -0.00000481, 0.00000050, -0.00000051, 0.00000752, -0.00000270, -0.00000532, },
		{ 0.00001514, -0.00000809, -0.00000231, 0.00000407, -0.00000224, -0.00000030, -0.00000358, 0.00001256, -0.00001590, 0.00000560, 0.00000821, -0.00001414, -0.00000604, 0.00000828, 0.00000398, -0.00000385, -0.00000271, -0.00000185, 0.00000560, -0.00000502, 0.00000687, -0.00001008, 0.00000078, -0.00000624, -0.00000607, 0.00001219, 0.00000880, -0.00000994, 0.00000367, -0.00000949, -0.00000577, 0.00000108, 0.00000728, -0.00001034, 0.00001072, -0.00002265, 0.00001745, -0.00003173, 0.00003622, -0.00004127, 0.00002619, -0.00000001, 0.00001550, -0.00001702, 0.00001748, -0.00001232, 0.00002296, -0.00002145, 0.00001908, -0.00002147, 0.00002249, -0.00000683, 0.00001197, -0.00000938, -0.00000576, -0.00001513, 0.00000923, -0.00002541, 0.00001408, -0.00003695, 0.00002909, -0.00005904, 0.00003093, -0.00009329, 0.00004046, -0.00029543, -0.00007146, 0.00054006, 0.00049343, -0.00015175, -0.00024338, 0.00003908, -0.00009436, 0.00003876, -0.00006169, 0.00003609, -0.00005409, 0.00003787, -0.00003593, 0.00001984, -0.00001827, 0.00000083, -0.00000349, -0.00000314, -0.00001685, 0.00001367, -0.00000985, 0.00000682, -0.00000884, 0.00000009, -0.00002115, 0.00001813, -0.00000859, 0.00000091, -0.00001218, 0.00002234, -0.00003067, 0.00002528, -0.00003689, 0.00002965, -0.00001540, 0.00000938, -0.00000009, -0.00000495, 0.00000167, -0.00000112, 0.00000616, 0.00000394, 0.00000017, -0.00000638, 0.00000004, -0.00000335, 0.00001071, -0.00000354, 0.00000204, -0.00000145, -0.00000465, 0.00000241, -0.00000133, 0.00000294, 0.00000503, -0.00000656, -0.00000264, -0.00000015, 0.00000127, 0.00000492, -0.00000263, -0.00000399, },
		{ 0.00001467, -0.00001377, 0.00000463, -0.00000108, 0.00000165, -0.00000428, 0.00000155, 0.00000714, -0.00001457, 0.00001060, 0.00000281, -0.00001506, -0.00000317, 0.00001026, 0.00000069, -0.00000237, -0.00000484, 0.00000203, 0.00000203, -0.00000222, 0.00000480, -0.00000989, 0.00000088, -0.00000646, -0.00000614, 0.00001754, 0.00000176, -0.00000597, 0.00000035, -0.00000691, -0.00000954, 0.00000980, -0.00000320, 0.00000032, -0.00000131, -0.00000994, 0.00000446, -0.00002005, 0.00003125, -0.00004234, 0.00003304, -0.00000666, 0.00001765, -0.00001869, 0.00001819, -0.00001202, 0.00002206, -0.00002331, 0.00002084, -0.00002313, 0.00002469, -0.00000861, 0.00001199, -0.00001050, -0.00000673, -0.00001122, 0.00000444, -0.00002062, 0.00000734, -0.00002843, 0.00002026, -0.00005302, 0.00002318, -0.00008891, 0.00002934, -0.00033112, -0.00000936, 0.00066040, 0.00042085, -0.00030628, -0.00012647, -0.00001000, -0.00005447, 0.00001059, -0.00003709, 0.00001640, -0.00003798, 0.00002786, -0.00002948, 0.00001407, -0.00001266, -0.00000464, 0.00000381, -0.00001294, -0.00000661, 0.00000719, -0.00000437, 0.00000027, -0.00000166, -0.00000953, -0.00001142, 0.00001243, -0.00000355, -0.00000649, -0.00000164, 0.00001013, -0.00002006, 0.00001738, -0.00003357, 0.00003457, -0.00002460, 0.00001940, -0.00000932, 0.00000150, -0.00000291, 0.00000338, 0.00000346, 0.00000431, -0.00000074, -0.00000573, -0.00000080, -0.00000056, 0.00000944, -0.00000533, 0.00000422, -0.00000464, -0.00000179, 0.00000140, -0.00000117, 0.00000382, 0.00000402, -0.00000771, -0.00000105, -0.00000002, 0.00000164, 0.00000329, -0.00000305, -0.00000074, },
		{ 0.00000999, -0.00001496, 0.00000891, -0.00000508, 0.00000499, -0.00000790, 0.00000678, -0.00000001, -0.00000936, 0.00001197, -0.00000173, -0.00001569, 0.00000110, 0.00000990, -0.00000110, -0.00000206, -0.00000551, 0.00000481, -0.00000125, 0.00000088, 0.00000209, -0.00000929, 0.00000126, -0.00000770, -0.00000403, 0.00002052, -0.00000436, -0.00000209, -0.00000255, -0.00000555, -0.00001078, 0.00001580, -0.00001211, 0.00001041, -0.00001376, 0.00000479, -0.00001308, -0.00000026, 0.00001580, -0.00003297, 0.00003090, -0.00000702, 0.00001554, -0.00001682, 0.00001571, -0.00000835, 0.00001712, -0.00002110, 0.00001939, -0.00002214, 0.00002459, -0.00000814, 0.00000929, -0.00000874, -0.00001007, -0.00000494, -0.00000320, -0.00001288, -0.00000242, -0.00001585, 0.00000564, -0.00004026, 0.00000779, -0.00007611, 0.00000580, -0.00035758, 0.00006336, 0.00080595, 0.00024638, -0.00037263, -0.00003485, -0.00005562, -0.00001551, -0.00001850, -0.00001070, -0.00000618, -0.00001810, 0.00001400, -0.00002047, 0.00000695, -0.00000607, -0.00001078, 0.00001138, -0.00002334, 0.00000539, -0.00000137, 0.00000221, -0.00000661, 0.00000555, -0.00001961, 0.00000000, 0.00000474, 0.00000266, -0.00001436, 0.00000940, -0.00000399, -0.00000499, 0.00000227, -0.00002058, 0.00002934, -0.00002569, 0.00002400, -0.00001561, 0.00000668, -0.00000696, 0.00000819, -0.00000069, 0.00000623, -0.00000253, -0.00000472, -0.00000162, 0.00000280, 0.00000620, -0.00000477, 0.00000444, -0.00000643, 0.00000066, 0.00000019, -0.00000084, 0.00000492, 0.00000210, -0.00000767, -0.00000020, 0.00000088, 0.00000045, 0.00000304, -0.00000412, 0.00000432, },
		{ 0.00000203, -0.00001130, 0.00000911, -0.00000618, 0.00000605, -0.00000938, 0.00000996, -0.00000643, -0.00000212, 0.00001011, -0.00000500, -0.00001593, 0.00000624, 0.00000725, -0.00000093, -0.00000349, -0.00000391, 0.00000526, -0.00000280, 0.00000288, -0.00000014, -0.00000905, 0.00000249, -0.00001050, 0.00000115, 0.00001946, -0.00000739, -0.00000038, -0.00000325, -0.00000717, -0.00000725, 0.00001577, -0.00001511, 0.00001482, -0.00002065, 0.00001459, -0.00002743, 0.00002005, -0.00000403, -0.00001734, 0.00002223, -0.00000246, 0.00001030, -0.00001285, 0.00001193, -0.00000363, 0.00001053, -0.00001695, 0.00001679, -0.00002083, 0.00002508, -0.00000896, 0.00000782, -0.00000815, -0.00001159, -0.00000094, -0.00000861, -0.00000776, -0.00000878, -0.00000670, -0.00000646, -0.00002973, -0.00000537, -0.00006633, -0.00001705, -0.00039141, 0.00017784, 0.00093168, -0.00000452, -0.00033757, 0.00000216, -0.00007279, 0.00000301, -0.00003226, 0.00000347, -0.00001935, -0.00000501, 0.00000513, -0.00001643, 0.00000541, -0.00000516, -0.00001124, 0.00001310, -0.00002831, 0.00001348, -0.00000716, 0.00000565, -0.00000967, 0.00000845, -0.00002551, 0.00000854, -0.00000081, 0.00000622, -0.00001874, 0.00001631, -0.00001441, 0.00000829, -0.00001396, -0.00000273, 0.00001653, -0.00001906, 0.00002218, -0.00001734, 0.00000890, -0.00000881, 0.00001166, -0.00000499, 0.00000912, -0.00000510, -0.00000335, -0.00000209, 0.00000591, 0.00000186, -0.00000226, 0.00000257, -0.00000628, 0.00000203, -0.00000080, -0.00000038, 0.00000594, -0.00000033, -0.00000674, 0.00000023, 0.00000199, -0.00000169, 0.00000384, -0.00000525, 0.00000999, },
		{ -0.00000691, -0.00000423, 0.00000551, -0.00000400, 0.00000416, -0.00000781, 0.00000962, -0.00000981, 0.00000473, 0.00000633, -0.00000732, -0.00001520, 0.00001112, 0.00000329, 0.00000065, -0.00000625, -0.00000027, 0.00000318, -0.00000203, 0.00000307, -0.00000133, -0.00000935, 0.00000435, -0.00001435, 0.00000892, 0.00001421, -0.00000659, -0.00000160, -0.00000134, -0.00001176, 0.00000111, 0.00000891, -0.00001044, 0.00001092, -0.00001832, 0.00001439, -0.00003186, 0.00003303, -0.00002065, -0.00000175, 0.00001177, 0.00000378, 0.00000437, -0.00000901, 0.00000919, -0.00000046, 0.00000500, -0.00001321, 0.00001499, -0.00002114, 0.00002836, -0.00001386, 0.00001083, -0.00001214, -0.00000778, -0.00000299, -0.00000772, -0.00000962, -0.00000678, -0.00000696, -0.00000915, -0.00002895, -0.00000806, -0.00006903, -0.00002863, -0.00044612, 0.00036615, 0.00097720, -0.00027439, -0.00022315, -0.00001866, -0.00005335, -0.00000672, -0.00002333, -0.00000154, -0.00001651, -0.00000477, 0.00000629, -0.00002139, 0.00001306, -0.00001348, -0.00000251, 0.00000542, -0.00002402, 0.00001383, -0.00000687, 0.00000329, -0.00000650, 0.00000446, -0.00002425, 0.00001102, -0.00000131, 0.00000455, -0.00001707, 0.00001600, -0.00001686, 0.00001412, -0.00002470, 0.00001341, 0.00000138, -0.00000794, 0.00001543, -0.00001482, 0.00000783, -0.00000774, 0.00001257, -0.00000808, 0.00001205, -0.00000807, -0.00000174, -0.00000193, 0.00000787, -0.00000232, 0.00000111, -0.00000071, -0.00000438, 0.00000213, -0.00000136, 0.00000030, 0.00000643, -0.00000259, -0.00000551, 0.00000082, 0.00000238, -0.00000370, 0.00000482, -0.00000537, 0.00001430, },
		{ -0.00001393, 0.00000356, -0.00000006, 0.00000030, 0.00000012, -0.00000369, 0.00000565, -0.00000887, 0.00000903, 0.00000228, -0.00000937, -0.00001280, 0.00001435, -0.00000047, 0.00000234, -0.00000901, 0.00000408, -0.00000040, 0.00000046, 0.00000175, -0.00000175, -0.00000962, 0.00000583, -0.00001769, 0.00001734, 0.00000645, -0.00000306, -0.00000475, 0.00000182, -0.00001728, 0.00001168, -0.00000230, -0.00000009, 0.00000016, -0.00000752, 0.00000368, -0.00002387, 0.00003408, -0.00002847, 0.00000838, 0.00000396, 0.00000856, -0.00000020, -0.00000677, 0.00000871, -0.00000019, 0.00000194, -0.00001090, 0.00001438, -0.00002293, 0.00003412, -0.00002274, 0.00001863, -0.00002108, 0.00000163, -0.00001136, -0.00000031, -0.00001846, 0.00000362, -0.00001711, -0.00000157, -0.00003882, 0.00000034, -0.00008460, -0.00002994, -0.00051942, 0.00064346, 0.00088733, -0.00049015, -0.00007936, -0.00007416, -0.00000939, -0.00003666, 0.00000263, -0.00002169, -0.00000037, -0.00001547, 0.00001565, -0.00003316, 0.00002762, -0.00002900, 0.00001368, -0.00001029, -0.00001122, 0.00000662, -0.00000077, -0.00000416, 0.00000190, -0.00000551, -0.00001624, 0.00000740, 0.00000323, -0.00000206, -0.00000986, 0.00000877, -0.00001064, 0.00001013, -0.00002569, 0.00002227, -0.00001058, 0.00000321, 0.00000676, -0.00000979, 0.00000442, -0.00000405, 0.00001046, -0.00000891, 0.00001398, -0.00001079, -0.00000014, -0.00000094, 0.00000798, -0.00000506, 0.00000394, -0.00000417, -0.00000154, 0.00000125, -0.00000153, 0.00000128, 0.00000605, -0.00000408, -0.00000461, 0.00000212, 0.00000125, -0.00000448, 0.00000505, -0.00000348, 0.00001527, },
		{ -0.00001660, 0.00000916, -0.00000503, 0.00000464, -0.00000428, 0.00000125, -0.00000055, -0.00000383, 0.00000965, -0.00000071, -0.00001169, -0.00000831, 0.00001485, -0.00000256, 0.00000265, -0.00001013, 0.00000728, -0.00000369, 0.00000318, 0.00000006, -0.00000236, -0.00000876, 0.00000543, -0.00001837, 0.00002358, -0.00000091, 0.00000067, -0.00000752, 0.00000362, -0.00002030, 0.00002003, -0.00001286, 0.00001084, -0.00001223, 0.00000654, -0.00001284, -0.00000665, 0.00002405, -0.00002637, 0.00001106, 0.00000080, 0.00001066, -0.00000310, -0.00000565, 0.00000957, -0.00000184, 0.00000047, -0.00000881, 0.00001300, -0.00002342, 0.00003889, -0.00003193, 0.00002762, -0.00003128, 0.00001261, -0.00002168, 0.00000884, -0.00002875, 0.00001612, -0.00003045, 0.00000916, -0.00005132, 0.00001013, -0.00010084, -0.00003856, -0.00058444, 0.00099198, 0.00063556, -0.00058766, 0.00003434, -0.00012576, 0.00003282, -0.00006661, 0.00002929, -0.00004347, 0.00001793, -0.00002785, 0.00002498, -0.00004379, 0.00004134, -0.00004445, 0.00003053, -0.00002779, 0.00000469, -0.00000372, 0.00000712, -0.00001256, 0.00001117, -0.00001708, -0.00000542, 0.00000104, 0.00000965, -0.00001025, -0.00000085, -0.00000136, 0.00000066, -0.00000144, -0.00001708, 0.00002173, -0.00001586, 0.00001081, -0.00000084, -0.00000445, 0.00000021, 0.00000119, 0.00000563, -0.00000690, 0.00001400, -0.00001255, 0.00000118, 0.00000084, 0.00000600, -0.00000547, 0.00000496, -0.00000646, 0.00000112, 0.00000002, -0.00000152, 0.00000261, 0.00000466, -0.00000446, -0.00000432, 0.00000425, -0.00000171, -0.00000339, 0.00000410, 0.00000065, 0.00001181, },
		{ -0.00001400, 0.00001062, -0.00000727, 0.00000698, -0.00000700, 0.00000476, -0.00000660, 0.00000365, 0.00000676, -0.00000210, -0.00001434, -0.00000196, 0.00001227, -0.00000211, 0.00000061, -0.00000838, 0.00000768, -0.00000486, 0.00000441, -0.00000061, -0.00000421, -0.00000579, 0.00000198, -0.00001469, 0.00002511, -0.00000493, 0.00000187, -0.00000742, 0.00000157, -0.00001765, 0.00002184, -0.00001745, 0.00001633, -0.00001950, 0.00001629, -0.00002705, 0.00001215, 0.00000878, -0.00001807, 0.00000838, 0.00000106, 0.00001131, -0.00000607, -0.00000332, 0.00000889, -0.00000237, -0.00000228, -0.00000398, 0.00000728, -0.00001811, 0.00003717, -0.00003527, 0.00003144, -0.00003616, 0.00001810, -0.00002643, 0.00001156, -0.00003127, 0.00002015, -0.00003527, 0.00001020, -0.00005189, 0.00000393, -0.00009599, -0.00008574, -0.00059035, 0.00135443, 0.00024106, -0.00053555, 0.00007176, -0.00013693, 0.00004554, -0.00007393, 0.00003750, -0.00005035, 0.00002417, -0.00002984, 0.00002368, -0.00004325, 0.00004444, -0.00005042, 0.00003909, -0.00003856, 0.00001597, -0.00001054, 0.00001095, -0.00001619, 0.00001543, -0.00002422, 0.00000245, -0.00000287, 0.00001311, -0.00001509, 0.00000454, -0.00000826, 0.00001048, -0.00001448, -0.00000339, 0.00001389, -0.00001440, 0.00001356, -0.00000571, -0.00000033, -0.00000339, 0.00000666, -0.00000105, -0.00000211, 0.00001157, -0.00001272, 0.00000195, 0.00000314, 0.00000221, -0.00000335, 0.00000350, -0.00000657, 0.00000255, -0.00000085, -0.00000158, 0.00000418, 0.00000247, -0.00000385, -0.00000450, 0.00000676, -0.00000606, -0.00000058, 0.00000231, 0.00000617, 0.00000436, },
		{ -0.00000713, 0.00000775, -0.00000597, 0.00000629, -0.00000680, 0.00000508, -0.00000999, 0.00001109, 0.00000164, -0.00000219, -0.00001677, 0.00000527, 0.00000727, 0.00000066, -0.00000371, -0.00000358, 0.00000463, -0.00000289, 0.00000305, 0.00000063, -0.00000779, -0.00000053, -0.00000458, -0.00000633, 0.00002085, -0.00000396, -0.00000105, -0.00000323, -0.00000529, -0.00000819, 0.00001501, -0.00001293, 0.00001235, -0.00001662, 0.00001537, -0.00003107, 0.00002379, -0.00000359, -0.00001007, 0.00000512, 0.00000135, 0.00001317, -0.00001170, 0.00000311, 0.00000335, 0.00000167, -0.00000947, 0.00000650, -0.00000589, -0.00000322, 0.00002415, -0.00002710, 0.00002406, -0.00002944, 0.00001145, -0.00001853, 0.00000032, -0.00001757, 0.00000591, -0.00002044, -0.00001082, -0.00002651, -0.00003510, -0.00004866, -0.00020426, -0.00047530, 0.00164218, -0.00022831, -0.00034908, 0.00001818, -0.00008990, 0.00001342, -0.00004507, 0.00001509, -0.00003127, 0.00000853, -0.00001307, 0.00000457, -0.00002477, 0.00003014, -0.00004015, 0.00003265, -0.00003590, 0.00001628, -0.00000838, 0.00000608, -0.00001074, 0.00001024, -0.00002219, 0.00000260, 0.00000000, 0.00000960, -0.00001263, 0.00000195, -0.00000654, 0.00001225, -0.00002168, 0.00000845, 0.00000398, -0.00000919, 0.00001269, -0.00000800, 0.00000221, -0.00000569, 0.00001124, -0.00000838, 0.00000486, 0.00000663, -0.00001094, 0.00000205, 0.00000547, -0.00000254, 0.00000067, -0.00000027, -0.00000423, 0.00000218, -0.00000088, -0.00000180, 0.00000569, -0.00000001, -0.00000272, -0.00000459, 0.00000874, -0.00001072, 0.00000312, 0.00000070, 0.00001123, -0.00000504, },
		{ 0.00000142, 0.00000213, -0.00000196, 0.00000302, -0.00000383, 0.00000175, -0.00000911, 0.00001610, -0.00000388, -0.00000183, -0.00001805, 0.00001188, 0.00000135, 0.00000457, -0.00000914, 0.00000315, -0.00000109, 0.00000187, -0.00000082, 0.00000364, -0.00001259, 0.00000606, -0.00001286, 0.00000516, 0.00001190, 0.00000149, -0.00000761, 0.00000417, -0.00001543, 0.00000618, 0.00000112, -0.00000019, -0.00000082, -0.00000308, 0.00000193, -0.00002111, 0.00002254, -0.00000664, -0.00000823, 0.00000590, -0.00000151, 0.00001832, -0.00002147, 0.00001499, -0.00000857, 0.00001189, -0.00002226, 0.00002317, -0.00002672, 0.00002159, -0.00000118, -0.00000564, 0.00000338, -0.00000899, -0.00000955, 0.00000422, -0.00002695, 0.00001463, -0.00002960, 0.00001773, -0.00005805, 0.00002940, -0.00011251, 0.00004864, -0.00041048, -0.00018912, 0.00175818, -0.00067061, -0.00008796, -0.00010344, 0.00000547, -0.00005783, 0.00001648, -0.00003589, 0.00001281, -0.00002859, 0.00002194, -0.00003143, 0.00001077, -0.00000122, -0.00001349, 0.00001059, -0.00001863, 0.00000413, 0.00000389, -0.00000803, 0.00000401, -0.00000473, -0.00001019, -0.00000610, 0.00001067, -0.00000157, -0.00000234, -0.00000935, 0.00000550, 0.00000263, -0.00001799, 0.00001234, -0.00000228, -0.00000455, 0.00001091, -0.00000915, 0.00000383, -0.00000672, 0.00001417, -0.00001507, 0.00001283, -0.00000020, -0.00000735, 0.00000155, 0.00000719, -0.00000705, 0.00000533, -0.00000530, -0.00000004, 0.00000013, -0.00000001, -0.00000214, 0.00000677, -0.00000214, -0.00000173, -0.00000393, 0.00000914, -0.00001431, 0.00000657, 0.00000044, 0.00001367, -0.00001341, },
		{ 0.00000850, -0.00000364, 0.00000271, -0.00000114, 0.00000038, -0.00000415, -0.00000383, 0.00001718, -0.00000815, -0.00000197, -0.00001731, 0.00001631, -0.00000366, 0.00000784, -0.00001377, 0.00000970, -0.00000752, 0.00000778, -0.00000585, 0.00000719, -0.00001712, 0.00001200, -0.00002028, 0.00001667, 0.00000131, 0.00000877, -0.00001530, 0.00001190, -0.00002513, 0.00002083, -0.00001480, 0.00001575, -0.00001819, 0.00001627, -0.00001985, 0.00000028, 0.00000865, 0.00000113, -0.00001458, 0.00001241, -0.00000822, 0.00002631, -0.00003402, 0.00003044, -0.00002491, 0.00002632, -0.00003820, 0.00004267, -0.00005098, 0.00005155, -0.00003412, 0.00002479, -0.00002635, 0.00002066, -0.00003999, 0.00003632, -0.00006386, 0.00005802, -0.00007853, 0.00007081, -0.00012192, 0.00010439, -0.00021435, 0.00017831, -0.00068737, 0.00028124, 0.00162833, -0.00097848, 0.00016222, -0.00024103, 0.00011469, -0.00014213, 0.00008991, -0.00009829, 0.00006764, -0.00007515, 0.00006438, -0.00007403, 0.00005372, -0.00004092, 0.00002188, -0.00002035, 0.00000754, -0.00001574, 0.00002187, -0.00002674, 0.00002324, -0.00002473, 0.00000757, -0.00002001, 0.00002567, -0.00001683, 0.00001191, -0.00002532, 0.00002432, -0.00001630, -0.00000328, 0.00000597, -0.00000144, -0.00000372, 0.00001064, -0.00001069, 0.00000540, -0.00000681, 0.00001497, -0.00001975, 0.00002020, -0.00000769, -0.00000259, 0.00000078, 0.00000764, -0.00001003, 0.00000900, -0.00000997, 0.00000469, -0.00000283, 0.00000140, -0.00000237, 0.00000708, -0.00000336, -0.00000142, -0.00000208, 0.00000724, -0.00001574, 0.00000887, 0.00000224, 0.00001193, -0.00001791, },
		{ 0.00001164, -0.00000705, 0.00000580, -0.00000405, 0.00000360, -0.00001035, 0.00000450, 0.00001408, -0.00001016, -0.00000317, -0.00001409, 0.00001749, -0.00000622, 0.00000882, -0.00001564, 0.00001372, -0.00001214, 0.00001246, -0.00000998, 0.00000944, -0.00001945, 0.00001504, -0.00002395, 0.00002455, -0.00000696, 0.00001417, -0.00002057, 0.00001623, -0.00002987, 0.00003001, -0.00002608, 0.00002763, -0.00003184, 0.00003280, -0.00004094, 0.00002470, -0.00001138, 0.00001530, -0.00002610, 0.00002217, -0.00001599, 0.00003350, -0.00004479, 0.00004420, -0.00004014, 0.00003936, -0.00005127, 0.00005803, -0.00007047, 0.00007732, -0.00006473, 0.00005417, -0.00005491, 0.00004869, -0.00006825, 0.00006510, -0.00009609, 0.00009634, -0.00012277, 0.00011865, -0.00017932, 0.00017090, -0.00030645, 0.00029497, -0.00097672, 0.00089368, 0.00123102, -0.00107302, 0.00031875, -0.00033362, 0.00019247, -0.00020308, 0.00014474, -0.00014601, 0.00011072, -0.00011166, 0.00009685, -0.00010694, 0.00008866, -0.00007468, 0.00005282, -0.00004810, 0.00003184, -0.00003393, 0.00003700, -0.00004172, 0.00003855, -0.00004144, 0.00002323, -0.00003198, 0.00003823, -0.00002940, 0.00002304, -0.00003826, 0.00004198, -0.00003732, 0.00001707, -0.00000789, 0.00000587, -0.00000716, 0.00001255, -0.00001306, 0.00000721, -0.00000616, 0.00001327, -0.00002116, 0.00002516, -0.00001413, 0.00000217, 0.00000028, 0.00000629, -0.00001043, 0.00001021, -0.00001258, 0.00000832, -0.00000551, 0.00000271, -0.00000218, 0.00000641, -0.00000341, -0.00000197, 0.00000091, 0.00000298, -0.00001457, 0.00000977, 0.00000580, 0.00000584, -0.00001699, },
		{ 0.00001003, -0.00000681, 0.00000592, -0.00000421, 0.00000386, -0.00001430, 0.00001351, 0.00000780, -0.00000981, -0.00000542, -0.00000868, 0.00001524, -0.00000560, 0.00000661, -0.00001357, 0.00001351, -0.00001288, 0.00001375, -0.00001123, 0.00000873, -0.00001807, 0.00001357, -0.00002189, 0.00002600, -0.00000961, 0.00001444, -0.00002040, 0.00001427, -0.00002635, 0.00002935, -0.00002721, 0.00002904, -0.00003430, 0.00003768, -0.00005110, 0.00004124, -0.00002738, 0.00002738, -0.00003588, 0.00002947, -0.00001956, 0.00003435, -0.00004757, 0.00004931, -0.00004691, 0.00004370, -0.00005398, 0.00006103, -0.00007577, 0.00008807, -0.00008114, 0.00007018, -0.00006958, 0.00006169, -0.00008004, 0.00007509, -0.00010634, 0.00010980, -0.00013988, 0.00013580, -0.00020081, 0.00019334, -0.00034351, 0.00033532, -0.00118641, 0.00154905, 0.00061337, -0.00093093, 0.00033045, -0.00033545, 0.00020162, -0.00020923, 0.00015367, -0.00015485, 0.00012064, -0.00011929, 0.00010253, -0.00011445, 0.00010056, -0.00008827, 0.00006602, -0.00006020, 0.00004282, -0.00004035, 0.00004030, -0.00004447, 0.00004153, -0.00004637, 0.00002852, -0.00003432, 0.00004114, -0.00003228, 0.00002374, -0.00003998, 0.00004924, -0.00005076, 0.00003417, -0.00002239, 0.00001524, -0.00001227, 0.00001506, -0.00001521, 0.00000857, -0.00000447, 0.00000879, -0.00001835, 0.00002612, -0.00001777, 0.00000562, 0.00000064, 0.00000293, -0.00000780, 0.00000810, -0.00001190, 0.00000945, -0.00000673, 0.00000317, -0.00000132, 0.00000479, -0.00000235, -0.00000318, 0.00000445, -0.00000296, -0.00001113, 0.00000979, 0.00000977, -0.00000310, -0.00001102, },
		{ 0.00000482, -0.00000335, 0.00000317, -0.00000146, 0.00000042, -0.00001415, 0.00002073, 0.00000004, -0.00000772, -0.00000815, -0.00000203, 0.00001029, -0.00000219, 0.00000150, -0.00000768, 0.00000874, -0.00000894, 0.00001058, -0.00000858, 0.00000436, -0.00001263, 0.00000749, -0.00001393, 0.00002033, -0.00000538, 0.00000824, -0.00001381, 0.00000549, -0.00001420, 0.00001794, -0.00001645, 0.00001746, -0.00002213, 0.00002602, -0.00004344, 0.00004116, -0.00003004, 0.00002873, -0.00003656, 0.00002829, -0.00001388, 0.00002412, -0.00003735, 0.00004026, -0.00003937, 0.00003362, -0.00004082, 0.00004590, -0.00006035, 0.00007603, -0.00007444, 0.00006337, -0.00006063, 0.00004945, -0.00006459, 0.00005482, -0.00008184, 0.00008356, -0.00011257, 0.00010231, -0.00016280, 0.00014243, -0.00028653, 0.00023919, -0.00121365, 0.00210941, -0.00011337, -0.00059479, 0.00019570, -0.00023444, 0.00012936, -0.00014819, 0.00010489, -0.00011353, 0.00008681, -0.00008856, 0.00007302, -0.00008861, 0.00008148, -0.00007389, 0.00005394, -0.00004931, 0.00003355, -0.00002895, 0.00002666, -0.00003041, 0.00002767, -0.00003477, 0.00001864, -0.00002255, 0.00003035, -0.00002169, 0.00001012, -0.00002564, 0.00003968, -0.00004866, 0.00003935, -0.00002964, 0.00002056, -0.00001474, 0.00001522, -0.00001503, 0.00000811, -0.00000123, 0.00000153, -0.00001107, 0.00002219, -0.00001735, 0.00000667, 0.00000219, -0.00000215, -0.00000250, 0.00000277, -0.00000760, 0.00000738, -0.00000571, 0.00000230, 0.00000029, 0.00000248, -0.00000055, -0.00000455, 0.00000760, -0.00000938, -0.00000633, 0.00000988, 0.00001216, -0.00001215, -0.00000223, },
		{ -0.00000147, 0.00000142, -0.00000088, 0.00000282, -0.00000584, -0.00000947, 0.00002440, -0.00000740, -0.00000483, -0.00001053, 0.00000452, 0.00000413, 0.00000271, -0.00000513, 0.00000056, 0.00000067, -0.00000118, 0.00000348, -0.00000246, -0.00000297, -0.00000424, -0.00000163, -0.00000204, 0.00000934, 0.00000436, -0.00000313, -0.00000239, -0.00000774, 0.00000344, -0.00000089, 0.00000307, -0.00000413, 0.00000203, -0.00000044, -0.00001793, 0.00002208, -0.00001532, 0.00001485, -0.00002408, 0.00001564, 0.00000287, 0.00000175, -0.00001331, 0.00001604, -0.00001636, 0.00000825, -0.00001151, 0.00001279, -0.00002429, 0.00004060, -0.00004322, 0.00003200, -0.00002640, 0.00001047, -0.00002055, 0.00000321, -0.00002149, 0.00001582, -0.00003806, 0.00001437, -0.00006022, 0.00001057, -0.00012238, -0.00002423, -0.00097840, 0.00243313, -0.00080000, -0.00016186, -0.00003717, -0.00005763, -0.00000598, -0.00003336, 0.00000854, -0.00002975, 0.00001518, -0.00002457, 0.00001309, -0.00003367, 0.00003483, -0.00003420, 0.00001872, -0.00001700, 0.00000515, -0.00000097, -0.00000216, -0.00000165, -0.00000092, -0.00000837, -0.00000512, 0.00000208, 0.00000731, 0.00000054, -0.00001577, 0.00000331, 0.00001312, -0.00002860, 0.00002818, -0.00002437, 0.00001707, -0.00001089, 0.00001041, -0.00001068, 0.00000465, 0.00000377, -0.00000786, -0.00000010, 0.00001370, -0.00001266, 0.00000495, 0.00000479, -0.00000809, 0.00000429, -0.00000464, -0.00000052, 0.00000241, -0.00000244, 0.00000002, 0.00000251, -0.00000010, 0.00000145, -0.00000550, 0.00000940, -0.00001494, -0.00000121, 0.00001085, 0.00001126, -0.00001830, 0.00000619, },
		{ -0.00000606, 0.00000508, -0.00000397, 0.00000633, -0.00001275, -0.00000137, 0.00002390, -0.00001318, -0.00000204, -0.00001181, 0.00000965, -0.00000156, 0.00000734, -0.00001127, 0.00000877, -0.00000825, 0.00000811, -0.00000558, 0.00000536, -0.00001137, 0.00000485, -0.00001092, 0.00001023, -0.00000316, 0.00001599, -0.00001615, 0.00001009, -0.00002087, 0.00002084, -0.00002056, 0.00002424, -0.00002807, 0.00002993, -0.00003328, 0.00001788, -0.00001039, 0.00001329, -0.00001205, -0.00000020, -0.00000623, 0.00002740, -0.00002848, 0.00001963, -0.00001824, 0.00001703, -0.00002690, 0.00002755, -0.00003091, 0.00002427, -0.00000995, 0.00000450, -0.00001578, 0.00002429, -0.00004563, 0.00004139, -0.00006755, 0.00006092, -0.00007833, 0.00006730, -0.00010993, 0.00008639, -0.00017874, 0.00012258, -0.00043498, -0.00045620, 0.00241670, -0.00129983, 0.00024540, -0.00028713, 0.00013839, -0.00016071, 0.00009981, -0.00010566, 0.00007131, -0.00007270, 0.00005372, -0.00005987, 0.00003440, -0.00002498, 0.00001781, -0.00002780, 0.00002606, -0.00003291, 0.00003478, -0.00003744, 0.00003303, -0.00003555, 0.00002464, -0.00003522, 0.00003241, -0.00002086, 0.00002697, -0.00004592, 0.00003881, -0.00002343, 0.00000455, 0.00000281, -0.00000663, 0.00000392, 0.00000015, 0.00000006, -0.00000186, -0.00000183, 0.00000992, -0.00001780, 0.00001253, 0.00000238, -0.00000480, 0.00000106, 0.00000770, -0.00001354, 0.00001079, -0.00001217, 0.00000755, -0.00000414, 0.00000230, -0.00000319, 0.00000492, -0.00000241, 0.00000316, -0.00000566, 0.00000923, -0.00001864, 0.00000352, 0.00001281, 0.00000639, -0.00001943, 0.00001131, },
		{ -0.00000712, 0.00000583, -0.00000428, 0.00000686, -0.00001766, 0.00000799, 0.00001979, -0.00001665, 0.00000008, -0.00001160, 0.00001243, -0.00000536, 0.00001009, -0.00001494, 0.00001448, -0.00001516, 0.00001599, -0.00001382, 0.00001237, -0.00001837, 0.00001198, -0.00001727, 0.00001898, -0.00001272, 0.00002501, -0.00002643, 0.00001922, -0.00002893, 0.00003186, -0.00003373, 0.00003874, -0.00004490, 0.00005075, -0.00006033, 0.00005126, -0.00004443, 0.00004564, -0.00004342, 0.00002758, -0.00003005, 0.00005181, -0.00005769, 0.00005176, -0.00005240, 0.00005046, -0.00006107, 0.00006465, -0.00007221, 0.00007102, -0.00006041, 0.00005312, -0.00006381, 0.00007423, -0.00010021, 0.00010083, -0.00013464, 0.00013948, -0.00016980, 0.00017093, -0.00023344, 0.00023340, -0.00037274, 0.00038201, -0.00091531, 0.00030187, 0.00202967, -0.00151076, 0.00051716, -0.00046921, 0.00028847, -0.00028185, 0.00020660, -0.00019906, 0.00015593, -0.00014732, 0.00012008, -0.00012187, 0.00009336, -0.00007753, 0.00006341, -0.00006829, 0.00006390, -0.00006617, 0.00006502, -0.00006652, 0.00006120, -0.00006392, 0.00005237, -0.00006051, 0.00005789, -0.00004389, 0.00004712, -0.00006903, 0.00006874, -0.00005785, 0.00003992, -0.00002833, 0.00001781, -0.00001503, 0.00001555, -0.00001349, 0.00000948, -0.00000977, 0.00001551, -0.00002591, 0.00002392, -0.00000891, 0.00000396, -0.00000352, 0.00000970, -0.00001694, 0.00001504, -0.00001757, 0.00001429, -0.00001031, 0.00000709, -0.00000644, 0.00000695, -0.00000395, 0.00000418, -0.00000494, 0.00000702, -0.00002003, 0.00000770, 0.00001500, -0.00000160, -0.00001522, 0.00001164, },
		{ -0.00000452, 0.00000336, -0.00000128, 0.00000337, -0.00001854, 0.00001623, 0.00001335, -0.00001779, 0.00000133, -0.00000993, 0.00001259, -0.00000652, 0.00001003, -0.00001496, 0.00001593, -0.00001778, 0.00001988, -0.00001864, 0.00001620, -0.00002184, 0.00001512, -0.00001848, 0.00002137, -0.00001585, 0.00002774, -0.00003043, 0.00002173, -0.00002858, 0.00003237, -0.00003520, 0.00004040, -0.00004728, 0.00005544, -0.00007048, 0.00006926, -0.00006651, 0.00006890, -0.00006768, 0.00004896, -0.00004620, 0.00006638, -0.00007550, 0.00007185, -0.00007462, 0.00007190, -0.00008200, 0.00008681, -0.00009692, 0.00010023, -0.00009377, 0.00008489, -0.00009359, 0.00010370, -0.00013203, 0.00013467, -0.00017229, 0.00018482, -0.00022501, 0.00023442, -0.00031150, 0.00032717, -0.00050405, 0.00056530, -0.00134261, 0.00117128, 0.00132995, -0.00140485, 0.00058962, -0.00052385, 0.00034291, -0.00032694, 0.00024995, -0.00023866, 0.00019457, -0.00018240, 0.00015105, -0.00015139, 0.00012312, -0.00010417, 0.00008535, -0.00008657, 0.00008141, -0.00008090, 0.00007734, -0.00007778, 0.00007161, -0.00007486, 0.00006388, -0.00007061, 0.00006879, -0.00005254, 0.00005172, -0.00007493, 0.00008152, -0.00007740, 0.00006462, -0.00005367, 0.00003963, -0.00003257, 0.00002967, -0.00002565, 0.00001961, -0.00001634, 0.00001821, -0.00002957, 0.00003094, -0.00001695, 0.00001085, -0.00000686, 0.00000950, -0.00001703, 0.00001551, -0.00001897, 0.00001758, -0.00001405, 0.00001032, -0.00000864, 0.00000795, -0.00000430, 0.00000430, -0.00000357, 0.00000318, -0.00001920, 0.00001157, 0.00001598, -0.00001047, -0.00000724, 0.00000769, },
		{ 0.00000015, -0.00000101, 0.00000383, -0.00000350, -0.00001478, 0.00002160, 0.00000603, -0.00001694, 0.00000179, -0.00000721, 0.00001049, -0.00000512, 0.00000720, -0.00001128, 0.00001271, -0.00001520, 0.00001841, -0.00001851, 0.00001552, -0.00002077, 0.00001362, -0.00001416, 0.00001670, -0.00001132, 0.00002267, -0.00002677, 0.00001682, -0.00001955, 0.00002200, -0.00002414, 0.00002790, -0.00003318, 0.00004071, -0.00005831, 0.00006394, -0.00006687, 0.00007279, -0.00007495, 0.00005509, -0.00004688, 0.00006365, -0.00007417, 0.00007158, -0.00007604, 0.00007237, -0.00008077, 0.00008489, -0.00009521, 0.00010085, -0.00009776, 0.00008680, -0.00009160, 0.00009837, -0.00012569, 0.00012622, -0.00016190, 0.00017541, -0.00021872, 0.00022812, -0.00030869, 0.00032356, -0.00051426, 0.00058599, -0.00158079, 0.00198021, 0.00045321, -0.00103543, 0.00046242, -0.00043633, 0.00028534, -0.00027990, 0.00021448, -0.00020972, 0.00017321, -0.00016506, 0.00013508, -0.00013780, 0.00011359, -0.00009539, 0.00007479, -0.00007421, 0.00007047, -0.00006970, 0.00006529, -0.00006552, 0.00005890, -0.00006296, 0.00005364, -0.00006023, 0.00006031, -0.00004260, 0.00003685, -0.00005904, 0.00007093, -0.00007390, 0.00006887, -0.00006314, 0.00004979, -0.00004118, 0.00003639, -0.00003137, 0.00002445, -0.00001858, 0.00001600, -0.00002691, 0.00003137, -0.00001921, 0.00001349, -0.00000732, 0.00000627, -0.00001328, 0.00001164, -0.00001553, 0.00001618, -0.00001397, 0.00001069, -0.00000888, 0.00000746, -0.00000322, 0.00000345, -0.00000193, -0.00000158, -0.00001656, 0.00001545, 0.00001431, -0.00001743, 0.00000160, 0.00000160, },
		{ 0.00000443, -0.00000495, 0.00000866, -0.00001155, -0.00000739, 0.00002347, -0.00000106, -0.00001454, 0.00000158, -0.00000401, 0.00000692, -0.00000192, 0.00000250, -0.00000500, 0.00000587, -0.00000814, 0.00001191, -0.00001349, 0.00001044, -0.00001556, 0.00000841, -0.00000582, 0.00000669, -0.00000055, 0.00001101, -0.00001677, 0.00000645, -0.00000482, 0.00000441, -0.00000448, 0.00000543, -0.00000700, 0.00001066, -0.00002666, 0.00003582, -0.00004366, 0.00005395, -0.00006139, 0.00004263, -0.00002968, 0.00004193, -0.00005223, 0.00004924, -0.00005462, 0.00004986, -0.00005577, 0.00005765, -0.00006593, 0.00007128, -0.00007007, 0.00005627, -0.00005532, 0.00005570, -0.00007852, 0.00007281, -0.00010052, 0.00010718, -0.00014513, 0.00014408, -0.00021414, 0.00020683, -0.00037804, 0.00039400, -0.00152138, 0.00255341, -0.00042177, -0.00051931, 0.00019462, -0.00023961, 0.00013767, -0.00015657, 0.00011218, -0.00012134, 0.00009876, -0.00010089, 0.00007719, -0.00008564, 0.00006868, -0.00005465, 0.00003496, -0.00003405, 0.00003341, -0.00003473, 0.00003132, -0.00003252, 0.00002606, -0.00003096, 0.00002398, -0.00003153, 0.00003476, -0.00001694, 0.00000599, -0.00002473, 0.00003910, -0.00004727, 0.00005008, -0.00005233, 0.00004342, -0.00003640, 0.00003192, -0.00002745, 0.00002136, -0.00001478, 0.00000823, -0.00001772, 0.00002477, -0.00001486, 0.00001082, -0.00000426, 0.00000009, -0.00000626, 0.00000413, -0.00000782, 0.00001022, -0.00000976, 0.00000774, -0.00000678, 0.00000535, -0.00000071, 0.00000167, -0.00000038, -0.00000648, -0.00001252, 0.00001926, 0.00000928, -0.00002020, 0.00000823, -0.00000379, },
		{ 0.00000615, -0.00000622, 0.00001066, -0.00001792, 0.00000144, 0.00002231, -0.00000732, -0.00001094, 0.00000083, -0.00000091, 0.00000288, 0.00000194, -0.00000264, 0.00000201, -0.00000246, 0.00000126, 0.00000224, -0.00000511, 0.00000242, -0.00000786, 0.00000162, 0.00000359, -0.00000517, 0.00001292, -0.00000387, -0.00000376, -0.00000549, 0.00001053, -0.00001419, 0.00001669, -0.00001891, 0.00002201, -0.00002449, 0.00001413, -0.00000601, -0.00000391, 0.00001752, -0.00003097, 0.00001538, 0.00000088, 0.00000660, -0.00001564, 0.00001101, -0.00001647, 0.00001070, -0.00001397, 0.00001297, -0.00001778, 0.00002066, -0.00001991, 0.00000288, 0.00000484, -0.00001294, -0.00000294, -0.00001170, -0.00000370, -0.00000289, -0.00002236, 0.00000176, -0.00004907, -0.00000017, -0.00011836, -0.00000020, -0.00112098, 0.00276051, -0.00112098, -0.00000020, -0.00011836, -0.00000017, -0.00004907, 0.00000176, -0.00002236, -0.00000289, -0.00000370, -0.00001170, -0.00000294, -0.00001294, 0.00000484, 0.00000288, -0.00001991, 0.00002066, -0.00001778, 0.00001297, -0.00001397, 0.00001070, -0.00001647, 0.00001101, -0.00001564, 0.00000660, 0.00000088, 0.00001538, -0.00003097, 0.00001752, -0.00000391, -0.00000601, 0.00001413, -0.00002449, 0.00002201, -0.00001891, 0.00001669, -0.00001419, 0.00001053, -0.00000549, -0.00000376, -0.00000387, 0.00001292, -0.00000517, 0.00000359, 0.00000162, -0.00000786, 0.00000242, -0.00000511, 0.00000224, 0.00000126, -0.00000246, 0.00000201, -0.00000264, 0.00000194, 0.00000288, -0.00000091, 0.00000083, -0.00001094, -0.00000732, 0.00002231, 0.00000144, -0.00001792, 0.00001066, -0.00000622, },
		{ 0.00000443, -0.00000379, 0.00000823, -0.00002020, 0.00000928, 0.00001926, -0.00001252, -0.00000648, -0.00000038, 0.00000167, -0.00000071, 0.00000535, -0.00000678, 0.00000774, -0.00000976, 0.00001022, -0.00000782, 0.00000413, -0.00000626, 0.00000009, -0.00000426, 0.00001082, -0.00001486, 0.00002477, -0.00001772, 0.00000823, -0.00001478, 0.00002136, -0.00002745, 0.00003192, -0.00003640, 0.00004342, -0.00005233, 0.00005008, -0.00004727, 0.00003910, -0.00002473, 0.00000599, -0.00001694, 0.00003476, -0.00003153, 0.00002398, -0.00003096, 0.00002606, -0.00003252, 0.00003132, -0.00003473, 0.00003341, -0.00003405, 0.00003496, -0.00005465, 0.00006868, -0.00008564, 0.00007719, -0.00010089, 0.00009876, -0.00012134, 0.00011218, -0.00015657, 0.00013767, -0.00023961, 0.00019462, -0.00051931, -0.00042177, 0.00255341, -0.00152138, 0.00039400, -0.00037804, 0.00020683, -0.00021414, 0.00014408, -0.00014513, 0.00010718, -0.00010052, 0.00007281, -0.00007852, 0.00005570, -0.00005532, 0.00005627, -0.00007007, 0.00007128, -0.00006593, 0.00005765, -0.00005577, 0.00004986, -0.00005462, 0.00004924, -0.00005223, 0.00004193, -0.00002968, 0.00004263, -0.00006139, 0.00005395, -0.00004366, 0.00003582, -0.00002666, 0.00001066, -0.00000700, 0.00000543, -0.00000448, 0.00000441, -0.00000482, 0.00000645, -0.00001677, 0.00001101, -0.00000055, 0.00000669, -0.00000582, 0.00000841, -0.00001556, 0.00001044, -0.00001349, 0.00001191, -0.00000814, 0.00000587, -0.00000500, 0.00000250, -0.00000192, 0.00000692, -0.00000401, 0.00000158, -0.00001454, -0.00000106, 0.00002347, -0.00000739, -0.00001155, 0.00000866, -0.00000495, },
		{ 0.00000015, 0.00000160, 0.00000160, -0.00001743, 0.00001431, 0.00001545, -0.00001656, -0.00000158, -0.00000193, 0.00000345, -0.00000322, 0.00000746, -0.00000888, 0.00001069, -0.00001397, 0.00001618, -0.00001553, 0.00001164, -0.00001328, 0.00000627, -0.00000732, 0.00001349, -0.00001921, 0.00003137, -0.00002691, 0.00001600, -0.00001858, 0.00002445, -0.00003137, 0.00003639, -0.00004118, 0.00004979, -0.00006314, 0.00006887, -0.00007390, 0.00007093, -0.00005904, 0.00003685, -0.00004260, 0.00006031, -0.00006023, 0.00005364, -0.00006296, 0.00005890, -0.00006552, 0.00006529, -0.00006970, 0.00007047, -0.00007421, 0.00007479, -0.00009539, 0.00011359, -0.00013780, 0.00013508, -0.00016506, 0.00017321, -0.00020972, 0.00021448, -0.00027990, 0.00028534, -0.00043633, 0.00046242, -0.00103543, 0.00045321, 0.00198021, -0.00158079, 0.00058599, -0.00051426, 0.00032356, -0.00030869, 0.00022812, -0.00021872, 0.00017541, -0.00016190, 0.00012622, -0.00012569, 0.00009837, -0.00009160, 0.00008680, -0.00009776, 0.00010085, -0.00009521, 0.00008489, -0.00008077, 0.00007237, -0.00007604, 0.00007158, -0.00007417, 0.00006365, -0.00004688, 0.00005509, -0.00007495, 0.00007279, -0.00006687, 0.00006394, -0.00005831, 0.00004071, -0.00003318, 0.00002790, -0.00002414, 0.00002200, -0.00001955, 0.00001682, -0.00002677, 0.00002267, -0.00001132, 0.00001670, -0.00001416, 0.00001362, -0.00002077, 0.00001552, -0.00001851, 0.00001841, -0.00001520, 0.00001271, -0.00001128, 0.00000720, -0.00000512, 0.00001049, -0.00000721, 0.00000179, -0.00001694, 0.00000603, 0.00002160, -0.00001478, -0.00000350, 0.00000383, -0.00000101, },
		{ -0.00000452, 0.00000769, -0.00000724, -0.00001047, 0.00001598, 0.00001157, -0.00001920, 0.00000318, -0.00000357, 0.00000430, -0.00000430, 0.00000795, -0.00000864, 0.00001032, -0.00001405, 0.00001758, -0.00001897, 0.00001551, -0.00001703, 0.00000950, -0.00000686, 0.00001085, -0.00001695, 0.00003094, -0.00002957, 0.00001821, -0.00001634, 0.00001961, -0.00002565, 0.00002967, -0.00003257, 0.00003963, -0.00005367, 0.00006462, -0.00007740, 0.00008152, -0.00007493, 0.00005172, -0.00005254, 0.00006879, -0.00007061, 0.00006388, -0.00007486, 0.00007161, -0.00007778, 0.00007734, -0.00008090, 0.00008141, -0.00008657, 0.00008535, -0.00010417, 0.00012312, -0.00015139, 0.00015105, -0.00018240, 0.00019457, -0.00023866, 0.00024995, -0.00032694, 0.00034291, -0.00052385, 0.00058962, -0.00140485, 0.00132995, 0.00117128, -0.00134261, 0.00056530, -0.00050405, 0.00032717, -0.00031150, 0.00023442, -0.00022501, 0.00018482, -0.00017229, 0.00013467, -0.00013203, 0.00010370, -0.00009359, 0.00008489, -0.00009377, 0.00010023, -0.00009692, 0.00008681, -0.00008200, 0.00007190, -0.00007462, 0.00007185, -0.00007550, 0.00006638, -0.00004620, 0.00004896, -0.00006768, 0.00006890, -0.00006651, 0.00006926, -0.00007048, 0.00005544, -0.00004728, 0.00004040, -0.00003520, 0.00003237, -0.00002858, 0.00002173, -0.00003043, 0.00002774, -0.00001585, 0.00002137, -0.00001848, 0.00001512, -0.00002184, 0.00001620, -0.00001864, 0.00001988, -0.00001778, 0.00001593, -0.00001496, 0.00001003, -0.00000652, 0.00001259, -0.00000993, 0.00000133, -0.00001779, 0.00001335, 0.00001623, -0.00001854, 0.00000337, -0.00000128, 0.00000336, },
		{ -0.00000712, 0.00001164, -0.00001522, -0.00000160, 0.00001500, 0.00000770, -0.00002003, 0.00000702, -0.00000494, 0.00000418, -0.00000395, 0.00000695, -0.00000644, 0.00000709, -0.00001031, 0.00001429, -0.00001757, 0.00001504, -0.00001694, 0.00000970, -0.00000352, 0.00000396, -0.00000891, 0.00002392, -0.00002591, 0.00001551, -0.00000977, 0.00000948, -0.00001349, 0.00001555, -0.00001503, 0.00001781, -0.00002833, 0.00003992, -0.00005785, 0.00006874, -0.00006903, 0.00004712, -0.00004389, 0.00005789, -0.00006051, 0.00005237, -0.00006392, 0.00006120, -0.00006652, 0.00006502, -0.00006617, 0.00006390, -0.00006829, 0.00006341, -0.00007753, 0.00009336, -0.00012187, 0.00012008, -0.00014732, 0.00015593, -0.00019906, 0.00020660, -0.00028185, 0.00028847, -0.00046921, 0.00051716, -0.00151076, 0.00202967, 0.00030187, -0.00091531, 0.00038201, -0.00037274, 0.00023340, -0.00023344, 0.00017093, -0.00016980, 0.00013948, -0.00013464, 0.00010083, -0.00010021, 0.00007423, -0.00006381, 0.00005312, -0.00006041, 0.00007102, -0.00007221, 0.00006465, -0.00006107, 0.00005046, -0.00005240, 0.00005176, -0.00005769, 0.00005181, -0.00003005, 0.00002758, -0.00004342, 0.00004564, -0.00004443, 0.00005126, -0.00006033, 0.00005075, -0.00004490, 0.00003874, -0.00003373, 0.00003186, -0.00002893, 0.00001922, -0.00002643, 0.00002501, -0.00001272, 0.00001898, -0.00001727, 0.00001198, -0.00001837, 0.00001237, -0.00001382, 0.00001599, -0.00001516, 0.00001448, -0.00001494, 0.00001009, -0.00000536, 0.00001243, -0.00001160, 0.00000008, -0.00001665, 0.00001979, 0.00000799, -0.00001766, 0.00000686, -0.00000428, 0.00000583, },
		{ -0.00000606, 0.00001131, -0.00001943, 0.00000639, 0.00001281, 0.00000352, -0.00001864, 0.00000923, -0.00000566, 0.00000316, -0.00000241, 0.00000492, -0.00000319, 0.00000230, -0.00000414, 0.00000755, -0.00001217, 0.00001079, -0.00001354, 0.00000770, 0.00000106, -0.00000480, 0.00000238, 0.00001253, -0.00001780, 0.00000992, -0.00000183, -0.00000186, 0.00000006, 0.00000015, 0.00000392, -0.00000663, 0.00000281, 0.00000455, -0.00002343, 0.00003881, -0.00004592, 0.00002697, -0.00002086, 0.00003241, -0.00003522, 0.00002464, -0.00003555, 0.00003303, -0.00003744, 0.00003478, -0.00003291, 0.00002606, -0.00002780, 0.00001781, -0.00002498, 0.00003440, -0.00005987, 0.00005372, -0.00007270, 0.00007131, -0.00010566, 0.00009981, -0.00016071, 0.00013839, -0.00028713, 0.00024540, -0.00129983, 0.00241670, -0.00045620, -0.00043498, 0.00012258, -0.00017874, 0.00008639, -0.00010993, 0.00006730, -0.00007833, 0.00006092, -0.00006755, 0.00004139, -0.00004563, 0.00002429, -0.00001578, 0.00000450, -0.00000995, 0.00002427, -0.00003091, 0.00002755, -0.00002690, 0.00001703, -0.00001824, 0.00001963, -0.00002848, 0.00002740, -0.00000623, -0.00000020, -0.00001205, 0.00001329, -0.00001039, 0.00001788, -0.00003328, 0.00002993, -0.00002807, 0.00002424, -0.00002056, 0.00002084, -0.00002087, 0.00001009, -0.00001615, 0.00001599, -0.00000316, 0.00001023, -0.00001092, 0.00000485, -0.00001137, 0.00000536, -0.00000558, 0.00000811, -0.00000825, 0.00000877, -0.00001127, 0.00000734, -0.00000156, 0.00000965, -0.00001181, -0.00000204, -0.00001318, 0.00002390, -0.00000137, -0.00001275, 0.00000633, -0.00000397, 0.00000508, },
		{ -0.00000147, 0.00000619, -0.00001830, 0.00001126, 0.00001085, -0.00000121, -0.00001494, 0.00000940, -0.00000550, 0.00000145, -0.00000010, 0.00000251, 0.00000002, -0.00000244, 0.00000241, -0.00000052, -0.00000464, 0.00000429, -0.00000809, 0.00000479, 0.00000495, -0.00001266, 0.00001370, -0.00000010, -0.00000786, 0.00000377, 0.00000465, -0.00001068, 0.00001041, -0.00001089, 0.00001707, -0.00002437, 0.00002818, -0.00002860, 0.00001312, 0.00000331, -0.00001577, 0.00000054, 0.00000731, 0.00000208, -0.00000512, -0.00000837, -0.00000092, -0.00000165, -0.00000216, -0.00000097, 0.00000515, -0.00001700, 0.00001872, -0.00003420, 0.00003483, -0.00003367, 0.00001309, -0.00002457, 0.00001518, -0.00002975, 0.00000854, -0.00003336, -0.00000598, -0.00005763, -0.00003717, -0.00016186, -0.00080000, 0.00243313, -0.00097840, -0.00002423, -0.00012238, 0.00001057, -0.00006022, 0.00001437, -0.00003806, 0.00001582, -0.00002149, 0.00000321, -0.00002055, 0.00001047, -0.00002640, 0.00003200, -0.00004322, 0.00004060, -0.00002429, 0.00001279, -0.00001151, 0.00000825, -0.00001636, 0.00001604, -0.00001331, 0.00000175, 0.00000287, 0.00001564, -0.00002408, 0.00001485, -0.00001532, 0.00002208, -0.00001793, -0.00000044, 0.00000203, -0.00000413, 0.00000307, -0.00000089, 0.00000344, -0.00000774, -0.00000239, -0.00000313, 0.00000436, 0.00000934, -0.00000204, -0.00000163, -0.00000424, -0.00000297, -0.00000246, 0.00000348, -0.00000118, 0.00000067, 0.00000056, -0.00000513, 0.00000271, 0.00000413, 0.00000452, -0.00001053, -0.00000483, -0.00000740, 0.00002440, -0.00000947, -0.00000584, 0.00000282, -0.00000088, 0.00000142, },
		{ 0.00000482, -0.00000223, -0.00001215, 0.00001216, 0.00000988, -0.00000633, -0.00000938, 0.00000760, -0.00000455, -0.00000055, 0.00000248, 0.00000029, 0.00000230, -0.00000571, 0.00000738, -0.00000760, 0.00000277, -0.00000250, -0.00000215, 0.00000219, 0.00000667, -0.00001735, 0.00002219, -0.00001107, 0.00000153, -0.00000123, 0.00000811, -0.00001503, 0.00001522, -0.00001474, 0.00002056, -0.00002964, 0.00003935, -0.00004866, 0.00003968, -0.00002564, 0.00001012, -0.00002169, 0.00003035, -0.00002255, 0.00001864, -0.00003477, 0.00002767, -0.00003041, 0.00002666, -0.00002895, 0.00003355, -0.00004931, 0.00005394, -0.00007389, 0.00008148, -0.00008861, 0.00007302, -0.00008856, 0.00008681, -0.00011353, 0.00010489, -0.00014819, 0.00012936, -0.00023444, 0.00019570, -0.00059479, -0.00011337, 0.00210941, -0.00121365, 0.00023919, -0.00028653, 0.00014243, -0.00016280, 0.00010231, -0.00011257, 0.00008356, -0.00008184, 0.00005482, -0.00006459, 0.00004945, -0.00006063, 0.00006337, -0.00007444, 0.00007603, -0.00006035, 0.00004590, -0.00004082, 0.00003362, -0.00003937, 0.00004026, -0.00003735, 0.00002412, -0.00001388, 0.00002829, -0.00003656, 0.00002873, -0.00003004, 0.00004116, -0.00004344, 0.00002602, -0.00002213, 0.00001746, -0.00001645, 0.00001794, -0.00001420, 0.00000549, -0.00001381, 0.00000824, -0.00000538, 0.00002033, -0.00001393, 0.00000749, -0.00001263, 0.00000436, -0.00000858, 0.00001058, -0.00000894, 0.00000874, -0.00000768, 0.00000150, -0.00000219, 0.00001029, -0.00000203, -0.00000815, -0.00000772, 0.00000004, 0.00002073, -0.00001415, 0.00000042, -0.00000146, 0.00000317, -0.00000335, },
		{ 0.00001003, -0.00001102, -0.00000310, 0.00000977, 0.00000979, -0.00001113, -0.00000296, 0.00000445, -0.00000318, -0.00000235, 0.00000479, -0.00000132, 0.00000317, -0.00000673, 0.00000945, -0.00001190, 0.00000810, -0.00000780, 0.00000293, 0.00000064, 0.00000562, -0.00001777, 0.00002612, -0.00001835, 0.00000879, -0.00000447, 0.00000857, -0.00001521, 0.00001506, -0.00001227, 0.00001524, -0.00002239, 0.00003417, -0.00005076, 0.00004924, -0.00003998, 0.00002374, -0.00003228, 0.00004114, -0.00003432, 0.00002852, -0.00004637, 0.00004153, -0.00004447, 0.00004030, -0.00004035, 0.00004282, -0.00006020, 0.00006602, -0.00008827, 0.00010056, -0.00011445, 0.00010253, -0.00011929, 0.00012064, -0.00015485, 0.00015367, -0.00020923, 0.00020162, -0.00033545, 0.00033045, -0.00093093, 0.00061337, 0.00154905, -0.00118641, 0.00033532, -0.00034351, 0.00019334, -0.00020081, 0.00013580, -0.00013988, 0.00010980, -0.00010634, 0.00007509, -0.00008004, 0.00006169, -0.00006958, 0.00007018, -0.00008114, 0.00008807, -0.00007577, 0.00006103, -0.00005398, 0.00004370, -0.00004691, 0.00004931, -0.00004757, 0.00003435, -0.00001956, 0.00002947, -0.00003588, 0.00002738, -0.00002738, 0.00004124, -0.00005110, 0.00003768, -0.00003430, 0.00002904, -0.00002721, 0.00002935, -0.00002635, 0.00001427, -0.00002040, 0.00001444, -0.00000961, 0.00002600, -0.00002189, 0.00001357, -0.00001807, 0.00000873, -0.00001123, 0.00001375, -0.00001288, 0.00001351, -0.00001357, 0.00000661, -0.00000560, 0.00001524, -0.00000868, -0.00000542, -0.00000981, 0.00000780, 0.00001351, -0.00001430, 0.00000386, -0.00000421, 0.00000592, -0.00000681, },
		{ 0.00001164, -0.00001699, 0.00000584, 0.00000580, 0.00000977, -0.00001457, 0.00000298, 0.00000091, -0.00000197, -0.00000341, 0.00000641, -0.00000218, 0.00000271, -0.00000551, 0.00000832, -0.00001258, 0.00001021, -0.00001043, 0.00000629, 0.00000028, 0.00000217, -0.00001413, 0.00002516, -0.00002116, 0.00001327, -0.00000616, 0.00000721, -0.00001306, 0.00001255, -0.00000716, 0.00000587, -0.00000789, 0.00001707, -0.00003732, 0.00004198, -0.00003826, 0.00002304, -0.00002940, 0.00003823, -0.00003198, 0.00002323, -0.00004144, 0.00003855, -0.00004172, 0.00003700, -0.00003393, 0.00003184, -0.00004810, 0.00005282, -0.00007468, 0.00008866, -0.00010694, 0.00009685, -0.00011166, 0.00011072, -0.00014601, 0.00014474, -0.00020308, 0.00019247, -0.00033362, 0.00031875, -0.00107302, 0.00123102, 0.00089368, -0.00097672, 0.00029497, -0.00030645, 0.00017090, -0.00017932, 0.00011865, -0.00012277, 0.00009634, -0.00009609, 0.00006510, -0.00006825, 0.00004869, -0.00005491, 0.00005417, -0.00006473, 0.00007732, -0.00007047, 0.00005803, -0.00005127, 0.00003936, -0.00004014, 0.00004420, -0.00004479, 0.00003350, -0.00001599, 0.00002217, -0.00002610, 0.00001530, -0.00001138, 0.00002470, -0.00004094, 0.00003280, -0.00003184, 0.00002763, -0.00002608, 0.00003001, -0.00002987, 0.00001623, -0.00002057, 0.00001417, -0.00000696, 0.00002455, -0.00002395, 0.00001504, -0.00001945, 0.00000944, -0.00000998, 0.00001246, -0.00001214, 0.00001372, -0.00001564, 0.00000882, -0.00000622, 0.00001749, -0.00001409, -0.00000317, -0.00001016, 0.00001408, 0.00000450, -0.00001035, 0.00000360, -0.00000405, 0.00000580, -0.00000705, },
		{ 0.00000850, -0.00001791, 0.00001193, 0.00000224, 0.00000887, -0.00001574, 0.00000724, -0.00000208, -0.00000142, -0.00000336, 0.00000708, -0.00000237, 0.00000140, -0.00000283, 0.00000469, -0.00000997, 0.00000900, -0.00001003, 0.00000764, 0.00000078, -0.00000259, -0.00000769, 0.00002020, -0.00001975, 0.00001497, -0.00000681, 0.00000540, -0.00001069, 0.00001064, -0.00000372, -0.00000144, 0.00000597, -0.00000328, -0.00001630, 0.00002432, -0.00002532, 0.00001191, -0.00001683, 0.00002567, -0.00002001, 0.00000757, -0.00002473, 0.00002324, -0.00002674, 0.00002187, -0.00001574, 0.00000754, -0.00002035, 0.00002188, -0.00004092, 0.00005372, -0.00007403, 0.00006438, -0.00007515, 0.00006764, -0.00009829, 0.00008991, -0.00014213, 0.00011469, -0.00024103, 0.00016222, -0.00097848, 0.00162833, 0.00028124, -0.00068737, 0.00017831, -0.00021435, 0.00010439, -0.00012192, 0.00007081, -0.00007853, 0.00005802, -0.00006386, 0.00003632, -0.00003999, 0.00002066, -0.00002635, 0.00002479, -0.00003412, 0.00005155, -0.00005098, 0.00004267, -0.00003820, 0.00002632, -0.00002491, 0.00003044, -0.00003402, 0.00002631, -0.00000822, 0.00001241, -0.00001458, 0.00000113, 0.00000865, 0.00000028, -0.00001985, 0.00001627, -0.00001819, 0.00001575, -0.00001480, 0.00002083, -0.00002513, 0.00001190, -0.00001530, 0.00000877, 0.00000131, 0.00001667, -0.00002028, 0.00001200, -0.00001712, 0.00000719, -0.00000585, 0.00000778, -0.00000752, 0.00000970, -0.00001377, 0.00000784, -0.00000366, 0.00001631, -0.00001731, -0.00000197, -0.00000815, 0.00001718, -0.00000383, -0.00000415, 0.00000038, -0.00000114, 0.00000271, -0.00000364, },
		{ 0.00000142, -0.00001341, 0.00001367, 0.00000044, 0.00000657, -0.00001431, 0.00000914, -0.00000393, -0.00000173, -0.00000214, 0.00000677, -0.00000214, -0.00000001, 0.00000013, -0.00000004, -0.00000530, 0.00000533, -0.00000705, 0.00000719, 0.00000155, -0.00000735, -0.00000020, 0.00001283, -0.00001507, 0.00001417, -0.00000672, 0.00000383, -0.00000915, 0.00001091, -0.00000455, -0.00000228, 0.00001234, -0.00001799, 0.00000263, 0.00000550, -0.00000935, -0.00000234, -0.00000157, 0.00001067, -0.00000610, -0.00001019, -0.00000473, 0.00000401, -0.00000803, 0.00000389, 0.00000413, -0.00001863, 0.00001059, -0.00001349, -0.00000122, 0.00001077, -0.00003143, 0.00002194, -0.00002859, 0.00001281, -0.00003589, 0.00001648, -0.00005783, 0.00000547, -0.00010344, -0.00008796, -0.00067061, 0.00175818, -0.00018912, -0.00041048, 0.00004864, -0.00011251, 0.00002940, -0.00005805, 0.00001773, -0.00002960, 0.00001463, -0.00002695, 0.00000422, -0.00000955, -0.00000899, 0.00000338, -0.00000564, -0.00000118, 0.00002159, -0.00002672, 0.00002317, -0.00002226, 0.00001189, -0.00000857, 0.00001499, -0.00002147, 0.00001832, -0.00000151, 0.00000590, -0.00000823, -0.00000664, 0.00002254, -0.00002111, 0.00000193, -0.00000308, -0.00000082, -0.00000019, 0.00000112, 0.00000618, -0.00001543, 0.00000417, -0.00000761, 0.00000149, 0.00001190, 0.00000516, -0.00001286, 0.00000606, -0.00001259, 0.00000364, -0.00000082, 0.00000187, -0.00000109, 0.00000315, -0.00000914, 0.00000457, 0.00000135, 0.00001188, -0.00001805, -0.00000183, -0.00000388, 0.00001610, -0.00000911, 0.00000175, -0.00000383, 0.00000302, -0.00000196, 0.00000213, },
		{ -0.00000713, -0.00000504, 0.00001123, 0.00000070, 0.00000312, -0.00001072, 0.00000874, -0.00000459, -0.00000272, -0.00000001, 0.00000569, -0.00000180, -0.00000088, 0.00000218, -0.00000423, -0.00000027, 0.00000067, -0.00000254, 0.00000547, 0.00000205, -0.00001094, 0.00000663, 0.00000486, -0.00000838, 0.00001124, -0.00000569, 0.00000221, -0.00000800, 0.00001269, -0.00000919, 0.00000398, 0.00000845, -0.00002168, 0.00001225, -0.00000654, 0.00000195, -0.00001263, 0.00000960, 0.00000000, 0.00000260, -0.00002219, 0.00001024, -0.00001074, 0.00000608, -0.00000838, 0.00001628, -0.00003590, 0.00003265, -0.00004015, 0.00003014, -0.00002477, 0.00000457, -0.00001307, 0.00000853, -0.00003127, 0.00001509, -0.00004507, 0.00001342, -0.00008990, 0.00001818, -0.00034908, -0.00022831, 0.00164218, -0.00047530, -0.00020426, -0.00004866, -0.00003510, -0.00002651, -0.00001082, -0.00002044, 0.00000591, -0.00001757, 0.00000032, -0.00001853, 0.00001145, -0.00002944, 0.00002406, -0.00002710, 0.00002415, -0.00000322, -0.00000589, 0.00000650, -0.00000947, 0.00000167, 0.00000335, 0.00000311, -0.00001170, 0.00001317, 0.00000135, 0.00000512, -0.00001007, -0.00000359, 0.00002379, -0.00003107, 0.00001537, -0.00001662, 0.00001235, -0.00001293, 0.00001501, -0.00000819, -0.00000529, -0.00000323, -0.00000105, -0.00000396, 0.00002085, -0.00000633, -0.00000458, -0.00000053, -0.00000779, 0.00000063, 0.00000305, -0.00000289, 0.00000463, -0.00000358, -0.00000371, 0.00000066, 0.00000727, 0.00000527, -0.00001677, -0.00000219, 0.00000164, 0.00001109, -0.00000999, 0.00000508, -0.00000680, 0.00000629, -0.00000597, 0.00000775, },
		{ -0.00001400, 0.00000436, 0.00000617, 0.00000231, -0.00000058, -0.00000606, 0.00000676, -0.00000450, -0.00000385, 0.00000247, 0.00000418, -0.00000158, -0.00000085, 0.00000255, -0.00000657, 0.00000350, -0.00000335, 0.00000221, 0.00000314, 0.00000195, -0.00001272, 0.00001157, -0.00000211, -0.00000105, 0.00000666, -0.00000339, -0.00000033, -0.00000571, 0.00001356, -0.00001440, 0.00001389, -0.00000339, -0.00001448, 0.00001048, -0.00000826, 0.00000454, -0.00001509, 0.00001311, -0.00000287, 0.00000245, -0.00002422, 0.00001543, -0.00001619, 0.00001095, -0.00001054, 0.00001597, -0.00003856, 0.00003909, -0.00005042, 0.00004444, -0.00004325, 0.00002368, -0.00002984, 0.00002417, -0.00005035, 0.00003750, -0.00007393, 0.00004554, -0.00013693, 0.00007176, -0.00053555, 0.00024106, 0.00135443, -0.00059035, -0.00008574, -0.00009599, 0.00000393, -0.00005189, 0.00001020, -0.00003527, 0.00002015, -0.00003127, 0.00001156, -0.00002643, 0.00001810, -0.00003616, 0.00003144, -0.00003527, 0.00003717, -0.00001811, 0.00000728, -0.00000398, -0.00000228, -0.00000237, 0.00000889, -0.00000332, -0.00000607, 0.00001131, 0.00000106, 0.00000838, -0.00001807, 0.00000878, 0.00001215, -0.00002705, 0.00001629, -0.00001950, 0.00001633, -0.00001745, 0.00002184, -0.00001765, 0.00000157, -0.00000742, 0.00000187, -0.00000493, 0.00002511, -0.00001469, 0.00000198, -0.00000579, -0.00000421, -0.00000061, 0.00000441, -0.00000486, 0.00000768, -0.00000838, 0.00000061, -0.00000211, 0.00001227, -0.00000196, -0.00001434, -0.00000210, 0.00000676, 0.00000365, -0.00000660, 0.00000476, -0.00000700, 0.00000698, -0.00000727, 0.00001062, },
		{ -0.00001660, 0.00001181, 0.00000065, 0.00000410, -0.00000339, -0.00000171, 0.00000425, -0.00000432, -0.00000446, 0.00000466, 0.00000261, -0.00000152, 0.00000002, 0.00000112, -0.00000646, 0.00000496, -0.00000547, 0.00000600, 0.00000084, 0.00000118, -0.00001255, 0.00001400, -0.00000690, 0.00000563, 0.00000119, 0.00000021, -0.00000445, -0.00000084, 0.00001081, -0.00001586, 0.00002173, -0.00001708, -0.00000144, 0.00000066, -0.00000136, -0.00000085, -0.00001025, 0.00000965, 0.00000104, -0.00000542, -0.00001708, 0.00001117, -0.00001256, 0.00000712, -0.00000372, 0.00000469, -0.00002779, 0.00003053, -0.00004445, 0.00004134, -0.00004379, 0.00002498, -0.00002785, 0.00001793, -0.00004347, 0.00002929, -0.00006661, 0.00003282, -0.00012576, 0.00003434, -0.00058766, 0.00063556, 0.00099198, -0.00058444, -0.00003856, -0.00010084, 0.00001013, -0.00005132, 0.00000916, -0.00003045, 0.00001612, -0.00002875, 0.00000884, -0.00002168, 0.00001261, -0.00003128, 0.00002762, -0.00003193, 0.00003889, -0.00002342, 0.00001300, -0.00000881, 0.00000047, -0.00000184, 0.00000957, -0.00000565, -0.00000310, 0.00001066, 0.00000080, 0.00001106, -0.00002637, 0.00002405, -0.00000665, -0.00001284, 0.00000654, -0.00001223, 0.00001084, -0.00001286, 0.00002003, -0.00002030, 0.00000362, -0.00000752, 0.00000067, -0.00000091, 0.00002358, -0.00001837, 0.00000543, -0.00000876, -0.00000236, 0.00000006, 0.00000318, -0.00000369, 0.00000728, -0.00001013, 0.00000265, -0.00000256, 0.00001485, -0.00000831, -0.00001169, -0.00000071, 0.00000965, -0.00000383, -0.00000055, 0.00000125, -0.00000428, 0.00000464, -0.00000503, 0.00000916, },
		{ -0.00001393, 0.00001527, -0.00000348, 0.00000505, -0.00000448, 0.00000125, 0.00000212, -0.00000461, -0.00000408, 0.00000605, 0.00000128, -0.00000153, 0.00000125, -0.00000154, -0.00000417, 0.00000394, -0.00000506, 0.00000798, -0.00000094, -0.00000014, -0.00001079, 0.00001398, -0.00000891, 0.00001046, -0.00000405, 0.00000442, -0.00000979, 0.00000676, 0.00000321, -0.00001058, 0.00002227, -0.00002569, 0.00001013, -0.00001064, 0.00000877, -0.00000986, -0.00000206, 0.00000323, 0.00000740, -0.00001624, -0.00000551, 0.00000190, -0.00000416, -0.00000077, 0.00000662, -0.00001122, -0.00001029, 0.00001368, -0.00002900, 0.00002762, -0.00003316, 0.00001565, -0.00001547, -0.00000037, -0.00002169, 0.00000263, -0.00003666, -0.00000939, -0.00007416, -0.00007936, -0.00049015, 0.00088733, 0.00064346, -0.00051942, -0.00002994, -0.00008460, 0.00000034, -0.00003882, -0.00000157, -0.00001711, 0.00000362, -0.00001846, -0.00000031, -0.00001136, 0.00000163, -0.00002108, 0.00001863, -0.00002274, 0.00003412, -0.00002293, 0.00001438, -0.00001090, 0.00000194, -0.00000019, 0.00000871, -0.00000677, -0.00000020, 0.00000856, 0.00000396, 0.00000838, -0.00002847, 0.00003408, -0.00002387, 0.00000368, -0.00000752, 0.00000016, -0.00000009, -0.00000230, 0.00001168, -0.00001728, 0.00000182, -0.00000475, -0.00000306, 0.00000645, 0.00001734, -0.00001769, 0.00000583, -0.00000962, -0.00000175, 0.00000175, 0.00000046, -0.00000040, 0.00000408, -0.00000901, 0.00000234, -0.00000047, 0.00001435, -0.00001280, -0.00000937, 0.00000228, 0.00000903, -0.00000887, 0.00000565, -0.00000369, 0.00000012, 0.00000030, -0.00000006, 0.00000356, },
		{ -0.00000691, 0.00001430, -0.00000537, 0.00000482, -0.00000370, 0.00000238, 0.00000082, -0.00000551, -0.00000259, 0.00000643, 0.00000030, -0.00000136, 0.00000213, -0.00000438, -0.00000071, 0.00000111, -0.00000232, 0.00000787, -0.00000193, -0.00000174, -0.00000807, 0.00001205, -0.00000808, 0.00001257, -0.00000774, 0.00000783, -0.00001482, 0.00001543, -0.00000794, 0.00000138, 0.00001341, -0.00002470, 0.00001412, -0.00001686, 0.00001600, -0.00001707, 0.00000455, -0.00000131, 0.00001102, -0.00002425, 0.00000446, -0.00000650, 0.00000329, -0.00000687, 0.00001383, -0.00002402, 0.00000542, -0.00000251, -0.00001348, 0.00001306, -0.00002139, 0.00000629, -0.00000477, -0.00001651, -0.00000154, -0.00002333, -0.00000672, -0.00005335, -0.00001866, -0.00022315, -0.00027439, 0.00097720, 0.00036615, -0.00044612, -0.00002863, -0.00006903, -0.00000806, -0.00002895, -0.00000915, -0.00000696, -0.00000678, -0.00000962, -0.00000772, -0.00000299, -0.00000778, -0.00001214, 0.00001083, -0.00001386, 0.00002836, -0.00002114, 0.00001499, -0.00001321, 0.00000500, -0.00000046, 0.00000919, -0.00000901, 0.00000437, 0.00000378, 0.00001177, -0.00000175, -0.00002065, 0.00003303, -0.00003186, 0.00001439, -0.00001832, 0.00001092, -0.00001044, 0.00000891, 0.00000111, -0.00001176, -0.00000134, -0.00000160, -0.00000659, 0.00001421, 0.00000892, -0.00001435, 0.00000435, -0.00000935, -0.00000133, 0.00000307, -0.00000203, 0.00000318, -0.00000027, -0.00000625, 0.00000065, 0.00000329, 0.00001112, -0.00001520, -0.00000732, 0.00000633, 0.00000473, -0.00000981, 0.00000962, -0.00000781, 0.00000416, -0.00000400, 0.00000551, -0.00000423, },
		{ 0.00000203, 0.00000999, -0.00000525, 0.00000384, -0.00000169, 0.00000199, 0.00000023, -0.00000674, -0.00000033, 0.00000594, -0.00000038, -0.00000080, 0.00000203, -0.00000628, 0.00000257, -0.00000226, 0.00000186, 0.00000591, -0.00000209, -0.00000335, -0.00000510, 0.00000912, -0.00000499, 0.00001166, -0.00000881, 0.00000890, -0.00001734, 0.00002218, -0.00001906, 0.00001653, -0.00000273, -0.00001396, 0.00000829, -0.00001441, 0.00001631, -0.00001874, 0.00000622, -0.00000081, 0.00000854, -0.00002551, 0.00000845, -0.00000967, 0.00000565, -0.00000716, 0.00001348, -0.00002831, 0.00001310, -0.00001124, -0.00000516, 0.00000541, -0.00001643, 0.00000513, -0.00000501, -0.00001935, 0.00000347, -0.00003226, 0.00000301, -0.00007279, 0.00000216, -0.00033757, -0.00000452, 0.00093168, 0.00017784, -0.00039141, -0.00001705, -0.00006633, -0.00000537, -0.00002973, -0.00000646, -0.00000670, -0.00000878, -0.00000776, -0.00000861, -0.00000094, -0.00001159, -0.00000815, 0.00000782, -0.00000896, 0.00002508, -0.00002083, 0.00001679, -0.00001695, 0.00001053, -0.00000363, 0.00001193, -0.00001285, 0.00001030, -0.00000246, 0.00002223, -0.00001734, -0.00000403, 0.00002005, -0.00002743, 0.00001459, -0.00002065, 0.00001482, -0.00001511, 0.00001577, -0.00000725, -0.00000717, -0.00000325, -0.00000038, -0.00000739, 0.00001946, 0.00000115, -0.00001050, 0.00000249, -0.00000905, -0.00000014, 0.00000288, -0.00000280, 0.00000526, -0.00000391, -0.00000349, -0.00000093, 0.00000725, 0.00000624, -0.00001593, -0.00000500, 0.00001011, -0.00000212, -0.00000643, 0.00000996, -0.00000938, 0.00000605, -0.00000618, 0.00000911, -0.00001130, },
		{ 0.00000999, 0.00000432, -0.00000412, 0.00000304, 0.00000045, 0.00000088, -0.00000020, -0.00000767, 0.00000210, 0.00000492, -0.00000084, 0.00000019, 0.00000066, -0.00000643, 0.00000444, -0.00000477, 0.00000620, 0.00000280, -0.00000162, -0.00000472, -0.00000253, 0.00000623, -0.00000069, 0.00000819, -0.00000696, 0.00000668, -0.00001561, 0.00002400, -0.00002569, 0.00002934, -0.00002058, 0.00000227, -0.00000499, -0.00000399, 0.00000940, -0.00001436, 0.00000266, 0.00000474, 0.00000000, -0.00001961, 0.00000555, -0.00000661, 0.00000221, -0.00000137, 0.00000539, -0.00002334, 0.00001138, -0.00001078, -0.00000607, 0.00000695, -0.00002047, 0.00001400, -0.00001810, -0.00000618, -0.00001070, -0.00001850, -0.00001551, -0.00005562, -0.00003485, -0.00037263, 0.00024638, 0.00080595, 0.00006336, -0.00035758, 0.00000580, -0.00007611, 0.00000779, -0.00004026, 0.00000564, -0.00001585, -0.00000242, -0.00001288, -0.00000320, -0.00000494, -0.00001007, -0.00000874, 0.00000929, -0.00000814, 0.00002459, -0.00002214, 0.00001939, -0.00002110, 0.00001712, -0.00000835, 0.00001571, -0.00001682, 0.00001554, -0.00000702, 0.00003090, -0.00003297, 0.00001580, -0.00000026, -0.00001308, 0.00000479, -0.00001376, 0.00001041, -0.00001211, 0.00001580, -0.00001078, -0.00000555, -0.00000255, -0.00000209, -0.00000436, 0.00002052, -0.00000403, -0.00000770, 0.00000126, -0.00000929, 0.00000209, 0.00000088, -0.00000125, 0.00000481, -0.00000551, -0.00000206, -0.00000110, 0.00000990, 0.00000110, -0.00001569, -0.00000173, 0.00001197, -0.00000936, -0.00000001, 0.00000678, -0.00000790, 0.00000499, -0.00000508, 0.00000891, -0.00001496, },
		{ 0.00001467, -0.00000074, -0.00000305, 0.00000329, 0.00000164, -0.00000002, -0.00000105, -0.00000771, 0.00000402, 0.00000382, -0.00000117, 0.00000140, -0.00000179, -0.00000464, 0.00000422, -0.00000533, 0.00000944, -0.00000056, -0.00000080, -0.00000573, -0.00000074, 0.00000431, 0.00000346, 0.00000338, -0.00000291, 0.00000150, -0.00000932, 0.00001940, -0.00002460, 0.00003457, -0.00003357, 0.00001738, -0.00002006, 0.00001013, -0.00000164, -0.00000649, -0.00000355, 0.00001243, -0.00001142, -0.00000953, -0.00000166, 0.00000027, -0.00000437, 0.00000719, -0.00000661, -0.00001294, 0.00000381, -0.00000464, -0.00001266, 0.00001407, -0.00002948, 0.00002786, -0.00003798, 0.00001640, -0.00003709, 0.00001059, -0.00005447, -0.00001000, -0.00012647, -0.00030628, 0.00042085, 0.00066040, -0.00000936, -0.00033112, 0.00002934, -0.00008891, 0.00002318, -0.00005302, 0.00002026, -0.00002843, 0.00000734, -0.00002062, 0.00000444, -0.00001122, -0.00000673, -0.00001050, 0.00001199, -0.00000861, 0.00002469, -0.00002313, 0.00002084, -0.00002331, 0.00002206, -0.00001202, 0.00001819, -0.00001869, 0.00001765, -0.00000666, 0.00003304, -0.00004234, 0.00003125, -0.00002005, 0.00000446, -0.00000994, -0.00000131, 0.00000032, -0.00000320, 0.00000980, -0.00000954, -0.00000691, 0.00000035, -0.00000597, 0.00000176, 0.00001754, -0.00000614, -0.00000646, 0.00000088, -0.00000989, 0.00000480, -0.00000222, 0.00000203, 0.00000203, -0.00000484, -0.00000237, 0.00000069, 0.00001026, -0.00000317, -0.00001506, 0.00000281, 0.00001060, -0.00001457, 0.00000714, 0.00000155, -0.00000428, 0.00000165, -0.00000108, 0.00000463, -0.00001377, },
		{ 0.00001514, -0.00000399, -0.00000263, 0.00000492, 0.00000127, -0.00000015, -0.00000264, -0.00000656, 0.00000503, 0.00000294, -0.00000133, 0.00000241, -0.00000465, -0.00000145, 0.00000204, -0.00000354, 0.00001071, -0.00000335, 0.00000004, -0.00000638, 0.00000017, 0.00000394, 0.00000616, -0.00000112, 0.00000167, -0.00000495, -0.00000009, 0.00000938, -0.00001540, 0.00002965, -0.00003689, 0.00002528, -0.00003067, 0.00002234, -0.00001218, 0.00000091, -0.00000859, 0.00001813, -0.00002115, 0.00000009, -0.00000884, 0.00000682, -0.00000985, 0.00001367, -0.00001685, -0.00000314, -0.00000349, 0.00000083, -0.00001827, 0.00001984, -0.00003593, 0.00003787, -0.00005409, 0.00003609, -0.00006169, 0.00003876, -0.00009436, 0.00003908, -0.00024338, -0.00015175, 0.00049343, 0.00054006, -0.00007146, -0.00029543, 0.00004046, -0.00009329, 0.00003093, -0.00005904, 0.00002909, -0.00003695, 0.00001408, -0.00002541, 0.00000923, -0.00001513, -0.00000576, -0.00000938, 0.00001197, -0.00000683, 0.00002249, -0.00002147, 0.00001908, -0.00002145, 0.00002296, -0.00001232, 0.00001748, -0.00001702, 0.00001550, -0.00000001, 0.00002619, -0.00004127, 0.00003622, -0.00003173, 0.00001745, -0.00002265, 0.00001072, -0.00001034, 0.00000728, 0.00000108, -0.00000577, -0.00000949, 0.00000367, -0.00000994, 0.00000880, 0.00001219, -0.00000607, -0.00000624, 0.00000078, -0.00001008, 0.00000687, -0.00000502, 0.00000560, -0.00000185, -0.00000271, -0.00000385, 0.00000398, 0.00000828, -0.00000604, -0.00001414, 0.00000821, 0.00000560, -0.00001590, 0.00001256, -0.00000358, -0.00000030, -0.00000224, 0.00000407, -0.00000231, -0.00000809, },
		{ 0.00001206, -0.00000532, -0.00000270, 0.00000752, -0.00000051, 0.00000050, -0.00000481, -0.00000433, 0.00000505, 0.00000239, -0.00000125, 0.00000272, -0.00000697, 0.00000207, -0.00000122, 0.00000018, 0.00000976, -0.00000505, 0.00000067, -0.00000680, 0.00000053, 0.00000514, 0.00000659, -0.00000381, 0.00000485, -0.00001021, 0.00000901, -0.00000269, -0.00000098, 0.00001591, -0.00002958, 0.00002303, -0.00003267, 0.00002828, -0.00001836, 0.00000456, -0.00000946, 0.00001865, -0.00002550, 0.00000528, -0.00001217, 0.00000952, -0.00001082, 0.00001427, -0.00002077, 0.00000096, -0.00000521, 0.00000006, -0.00001690, 0.00001798, -0.00003315, 0.00003628, -0.00005692, 0.00004140, -0.00007071, 0.00004918, -0.00011372, 0.00006106, -0.00034112, 0.00004848, 0.00047412, 0.00046351, -0.00014047, -0.00024128, 0.00003164, -0.00008246, 0.00002517, -0.00005290, 0.00002691, -0.00003646, 0.00001353, -0.00002367, 0.00000800, -0.00001391, -0.00000955, -0.00000317, 0.00000689, -0.00000059, 0.00001620, -0.00001593, 0.00001341, -0.00001495, 0.00001910, -0.00000857, 0.00001330, -0.00001214, 0.00001018, 0.00001131, 0.00001159, -0.00002987, 0.00002889, -0.00003131, 0.00002058, -0.00002777, 0.00001698, -0.00001650, 0.00001460, -0.00000633, -0.00000247, -0.00001103, 0.00000531, -0.00001174, 0.00001418, 0.00000678, -0.00000539, -0.00000599, 0.00000014, -0.00000906, 0.00000724, -0.00000612, 0.00000778, -0.00000515, -0.00000055, -0.00000526, 0.00000759, 0.00000471, -0.00000765, -0.00001254, 0.00001333, -0.00000215, -0.00001279, 0.00001458, -0.00000675, 0.00000218, -0.00000463, 0.00000797, -0.00000930, 0.00000009, },
		{ 0.00000714, -0.00000549, -0.00000249, 0.00001011, -0.00000298, 0.00000146, -0.00000701, -0.00000158, 0.00000435, 0.00000211, -0.00000083, 0.00000196, -0.00000790, 0.00000472, -0.00000428, 0.00000478, 0.00000699, -0.00000551, 0.00000099, -0.00000726, 0.00000091, 0.00000731, 0.00000478, -0.00000394, 0.00000528, -0.00001211, 0.00001469, -0.00001241, 0.00001353, -0.00000190, -0.00001485, 0.00001183, -0.00002563, 0.00002657, -0.00001877, 0.00000344, -0.00000548, 0.00001332, -0.00002336, 0.00000458, -0.00001019, 0.00000720, -0.00000654, 0.00000827, -0.00001716, -0.00000225, 0.00000039, -0.00000884, -0.00000644, 0.00000632, -0.00001905, 0.00002062, -0.00004296, 0.00002739, -0.00005759, 0.00003297, -0.00009990, 0.00003392, -0.00037738, 0.00023887, 0.00039908, 0.00042362, -0.00021658, -0.00017109, 0.00000469, -0.00005746, 0.00000682, -0.00003528, 0.00001393, -0.00002677, 0.00000554, -0.00001552, 0.00000099, -0.00000795, -0.00001741, 0.00000740, -0.00000272, 0.00000983, 0.00000620, -0.00000720, 0.00000498, -0.00000525, 0.00001177, -0.00000194, 0.00000705, -0.00000602, 0.00000468, 0.00002316, -0.00000608, -0.00001259, 0.00001254, -0.00001999, 0.00001319, -0.00002351, 0.00001513, -0.00001551, 0.00001583, -0.00000958, -0.00000187, -0.00001000, 0.00000407, -0.00001001, 0.00001612, 0.00000317, -0.00000547, -0.00000488, -0.00000152, -0.00000651, 0.00000548, -0.00000477, 0.00000746, -0.00000650, 0.00000032, -0.00000529, 0.00001007, 0.00000078, -0.00000855, -0.00000967, 0.00001658, -0.00001071, -0.00000617, 0.00001291, -0.00000710, 0.00000212, -0.00000418, 0.00000863, -0.00001364, 0.00000804, },
		{ 0.00000233, -0.00000554, -0.00000112, 0.00001154, -0.00000515, 0.00000202, -0.00000853, 0.00000100, 0.00000341, 0.00000194, -0.00000014, 0.00000012, -0.00000701, 0.00000558, -0.00000588, 0.00000892, 0.00000325, -0.00000495, 0.00000103, -0.00000798, 0.00000202, 0.00000936, 0.00000161, -0.00000188, 0.00000283, -0.00000983, 0.00001492, -0.00001613, 0.00002286, -0.00001756, 0.00000145, -0.00000366, -0.00001261, 0.00001902, -0.00001473, -0.00000095, 0.00000148, 0.00000419, -0.00001655, -0.00000056, -0.00000430, 0.00000153, 0.00000078, -0.00000166, -0.00000864, -0.00001033, 0.00001073, -0.00002310, 0.00001030, -0.00001208, 0.00000264, -0.00000485, -0.00001659, -0.00000173, -0.00002634, -0.00000614, -0.00005590, -0.00004493, -0.00032841, 0.00037066, 0.00031378, 0.00039753, -0.00028707, -0.00009679, -0.00003078, -0.00002598, -0.00001736, -0.00001217, -0.00000479, -0.00001200, -0.00000627, -0.00000437, -0.00000853, -0.00000039, -0.00002608, 0.00001899, -0.00001388, 0.00002190, -0.00000531, 0.00000255, -0.00000376, 0.00000497, 0.00000353, 0.00000526, 0.00000099, -0.00000118, 0.00000239, 0.00003070, -0.00002072, 0.00000366, -0.00000594, -0.00000338, -0.00000085, -0.00001236, 0.00000669, -0.00000814, 0.00001092, -0.00000808, -0.00000441, -0.00000636, 0.00000023, -0.00000501, 0.00001435, 0.00000206, -0.00000683, -0.00000275, -0.00000405, -0.00000287, 0.00000212, -0.00000132, 0.00000459, -0.00000547, -0.00000069, -0.00000325, 0.00001034, -0.00000238, -0.00000927, -0.00000517, 0.00001655, -0.00001770, 0.00000190, 0.00000863, -0.00000511, -0.00000021, -0.00000086, 0.00000536, -0.00001360, 0.00001334, },
		{ -0.00000101, -0.00000616, 0.00000189, 0.00001102, -0.00000624, 0.00000167, -0.00000885, 0.00000280, 0.00000265, 0.00000178, 0.00000059, -0.00000238, -0.00000449, 0.00000442, -0.00000528, 0.00001147, -0.00000045, -0.00000378, 0.00000081, -0.00000902, 0.00000439, 0.00001012, -0.00000145, 0.00000100, -0.00000123, -0.00000439, 0.00000994, -0.00001267, 0.00002379, -0.00002596, 0.00001327, -0.00001757, 0.00000146, 0.00000936, -0.00000929, -0.00000563, 0.00000810, -0.00000505, -0.00000871, -0.00000677, 0.00000230, -0.00000415, 0.00000721, -0.00001089, -0.00000015, -0.00001818, 0.00002032, -0.00003676, 0.00002703, -0.00003044, 0.00002413, -0.00003104, 0.00001197, -0.00003462, 0.00001024, -0.00005327, -0.00000003, -0.00015582, -0.00019821, 0.00041754, 0.00025557, 0.00036035, -0.00033554, -0.00003330, -0.00006240, 0.00000185, -0.00003846, 0.00000839, -0.00002215, 0.00000186, -0.00001675, 0.00000504, -0.00001624, 0.00000470, -0.00003148, 0.00002743, -0.00002259, 0.00003217, -0.00001540, 0.00001077, -0.00001027, 0.00001299, -0.00000304, 0.00001086, -0.00000310, 0.00000082, 0.00000529, 0.00003070, -0.00002760, 0.00001263, -0.00001916, 0.00001115, -0.00001515, 0.00000039, -0.00000397, 0.00000206, 0.00000249, -0.00000354, -0.00000878, -0.00000152, -0.00000455, 0.00000156, 0.00001005, 0.00000287, -0.00000906, -0.00000016, -0.00000667, 0.00000079, -0.00000159, 0.00000297, 0.00000019, -0.00000266, -0.00000320, 0.00000065, 0.00000811, -0.00000413, -0.00000998, 0.00000071, 0.00001263, -0.00002110, 0.00000898, 0.00000365, -0.00000227, -0.00000340, 0.00000396, -0.00000086, -0.00000908, 0.00001475, },
		{ -0.00000252, -0.00000735, 0.00000631, 0.00000846, -0.00000599, 0.00000029, -0.00000786, 0.00000358, 0.00000231, 0.00000162, 0.00000098, -0.00000481, -0.00000115, 0.00000174, -0.00000251, 0.00001182, -0.00000332, -0.00000241, 0.00000029, -0.00001015, 0.00000805, 0.00000877, -0.00000296, 0.00000296, -0.00000477, 0.00000168, 0.00000228, -0.00000384, 0.00001651, -0.00002512, 0.00001688, -0.00002526, 0.00001203, 0.00000134, -0.00000546, -0.00000783, 0.00001136, -0.00001097, -0.00000346, -0.00001061, 0.00000639, -0.00000670, 0.00000920, -0.00001516, 0.00000357, -0.00002076, 0.00002364, -0.00004367, 0.00003714, -0.00004170, 0.00003737, -0.00004838, 0.00003148, -0.00005819, 0.00003664, -0.00008925, 0.00004161, -0.00026326, -0.00001710, 0.00038200, 0.00024222, 0.00029663, -0.00035081, 0.00000832, -0.00008072, 0.00001828, -0.00004963, 0.00002011, -0.00003244, 0.00000997, -0.00002183, 0.00000915, -0.00001895, 0.00000448, -0.00003078, 0.00002964, -0.00002574, 0.00003784, -0.00002184, 0.00001578, -0.00001310, 0.00001729, -0.00000653, 0.00001388, -0.00000486, 0.00000030, 0.00001274, 0.00002313, -0.00002548, 0.00001146, -0.00002237, 0.00001782, -0.00002386, 0.00000932, -0.00001187, 0.00001055, -0.00000567, 0.00000122, -0.00001278, 0.00000248, -0.00000810, 0.00000735, 0.00000530, 0.00000420, -0.00001112, 0.00000198, -0.00000841, 0.00000327, -0.00000414, 0.00000639, -0.00000408, 0.00000055, -0.00000604, 0.00000533, 0.00000399, -0.00000447, -0.00001037, 0.00000699, 0.00000537, -0.00001998, 0.00001321, -0.00000013, -0.00000032, -0.00000557, 0.00000811, -0.00000779, -0.00000160, 0.00001252, },
		{ -0.00000278, -0.00000841, 0.00001121, 0.00000446, -0.00000469, -0.00000176, -0.00000588, 0.00000344, 0.00000238, 0.00000153, 0.00000067, -0.00000632, 0.00000191, -0.00000134, 0.00000166, 0.00001007, -0.00000500, -0.00000107, -0.00000069, -0.00001080, 0.00001241, 0.00000531, -0.00000217, 0.00000265, -0.00000576, 0.00000546, -0.00000429, 0.00000625, 0.00000447, -0.00001684, 0.00001216, -0.00002509, 0.00001675, -0.00000294, -0.00000477, -0.00000636, 0.00001004, -0.00001195, -0.00000267, -0.00001025, 0.00000642, -0.00000488, 0.00000550, -0.00001272, 0.00000031, -0.00001565, 0.00001789, -0.00004043, 0.00003685, -0.00004188, 0.00003784, -0.00005126, 0.00003490, -0.00006377, 0.00004191, -0.00009959, 0.00004648, -0.00032894, 0.00016927, 0.00029090, 0.00026958, 0.00020543, -0.00033185, 0.00002504, -0.00008271, 0.00002101, -0.00004887, 0.00002098, -0.00003363, 0.00001068, -0.00002033, 0.00000707, -0.00001608, -0.00000132, -0.00002370, 0.00002500, -0.00002231, 0.00003787, -0.00002389, 0.00001739, -0.00001241, 0.00001799, -0.00000708, 0.00001481, -0.00000539, -0.00000061, 0.00002168, 0.00001119, -0.00001715, 0.00000189, -0.00001542, 0.00001472, -0.00002415, 0.00001124, -0.00001372, 0.00001392, -0.00001036, 0.00000368, -0.00001456, 0.00000408, -0.00000879, 0.00001040, 0.00000201, 0.00000463, -0.00001202, 0.00000289, -0.00000862, 0.00000382, -0.00000452, 0.00000760, -0.00000668, 0.00000268, -0.00000776, 0.00000926, -0.00000082, -0.00000393, -0.00000983, 0.00001217, -0.00000360, -0.00001485, 0.00001388, -0.00000163, -0.00000042, -0.00000527, 0.00000952, -0.00001281, 0.00000632, 0.00000810, },
		{ -0.00000284, -0.00000836, 0.00001533, 0.00000006, -0.00000304, -0.00000384, -0.00000354, 0.00000280, 0.00000271, 0.00000153, -0.00000049, -0.00000633, 0.00000365, -0.00000353, 0.00000595, 0.00000695, -0.00000555, 0.00000016, -0.00000236, -0.00001022, 0.00001636, 0.00000063, 0.00000061, -0.00000011, -0.00000333, 0.00000497, -0.00000644, 0.00001296, -0.00000712, -0.00000570, 0.00000231, -0.00001866, 0.00001607, -0.00000368, -0.00000666, -0.00000216, 0.00000520, -0.00000883, -0.00000581, -0.00000620, 0.00000312, 0.00000005, -0.00000227, -0.00000507, -0.00000860, -0.00000418, 0.00000429, -0.00002792, 0.00002686, -0.00003193, 0.00002660, -0.00004047, 0.00002250, -0.00005099, 0.00002482, -0.00008153, 0.00000584, -0.00032608, 0.00031582, 0.00018317, 0.00031774, 0.00009848, -0.00028736, 0.00002182, -0.00007187, 0.00001328, -0.00003913, 0.00001346, -0.00002766, 0.00000573, -0.00001411, 0.00000070, -0.00000960, -0.00001050, -0.00001251, 0.00001535, -0.00001355, 0.00003305, -0.00002229, 0.00001665, -0.00000959, 0.00001647, -0.00000592, 0.00001495, -0.00000643, 0.00000084, 0.00002799, -0.00000012, -0.00000804, -0.00001071, -0.00000249, 0.00000434, -0.00001719, 0.00000651, -0.00000925, 0.00001130, -0.00001033, 0.00000277, -0.00001347, 0.00000298, -0.00000629, 0.00000997, 0.00000120, 0.00000338, -0.00001132, 0.00000240, -0.00000735, 0.00000251, -0.00000263, 0.00000615, -0.00000679, 0.00000279, -0.00000734, 0.00001110, -0.00000499, -0.00000320, -0.00000783, 0.00001473, -0.00001200, -0.00000741, 0.00001163, -0.00000095, -0.00000262, -0.00000216, 0.00000722, -0.00001395, 0.00001223, 0.00000338, },
		{ -0.00000359, -0.00000643, 0.00001756, -0.00000370, -0.00000174, -0.00000530, -0.00000147, 0.00000212, 0.00000309, 0.00000154, -0.00000231, -0.00000478, 0.00000355, -0.00000386, 0.00000910, 0.00000346, -0.00000533, 0.00000131, -0.00000484, -0.00000777, 0.00001857, -0.00000375, 0.00000404, -0.00000418, 0.00000169, 0.00000021, -0.00000281, 0.00001316, -0.00001359, 0.00000303, -0.00000781, -0.00000969, 0.00001247, -0.00000278, -0.00000913, 0.00000237, -0.00000048, -0.00000424, -0.00001045, -0.00000078, -0.00000104, 0.00000506, -0.00001042, 0.00000380, -0.00001905, 0.00000926, -0.00001253, -0.00001073, 0.00001176, -0.00001694, 0.00000957, -0.00002260, 0.00000142, -0.00002771, -0.00000572, -0.00004569, -0.00007058, -0.00024928, 0.00039252, 0.00009522, 0.00036166, -0.00000659, -0.00023124, 0.00000814, -0.00005545, 0.00000149, -0.00002620, 0.00000266, -0.00001880, -0.00000110, -0.00000674, -0.00000653, -0.00000287, -0.00001957, -0.00000082, 0.00000403, -0.00000209, 0.00002532, -0.00001852, 0.00001506, -0.00000636, 0.00001437, -0.00000439, 0.00001540, -0.00000908, 0.00000639, 0.00002856, -0.00000649, -0.00000363, -0.00001984, 0.00001004, -0.00000797, -0.00000719, -0.00000157, -0.00000107, 0.00000443, -0.00000653, -0.00000084, -0.00001030, 0.00000021, -0.00000171, 0.00000675, 0.00000264, 0.00000061, -0.00000931, 0.00000098, -0.00000531, 0.00000026, 0.00000065, 0.00000266, -0.00000465, 0.00000088, -0.00000459, 0.00001018, -0.00000759, -0.00000271, -0.00000429, 0.00001374, -0.00001772, 0.00000008, 0.00000804, 0.00000079, -0.00000581, 0.00000274, 0.00000172, -0.00001069, 0.00001457, -0.00000007, },
		{ -0.00000527, -0.00000251, 0.00001738, -0.00000617, -0.00000121, -0.00000578, -0.00000010, 0.00000172, 0.00000342, 0.00000134, -0.00000424, -0.00000219, 0.00000181, -0.00000204, 0.00001025, 0.00000052, -0.00000478, 0.00000234, -0.00000796, -0.00000325, 0.00001805, -0.00000625, 0.00000635, -0.00000751, 0.00000701, -0.00000669, 0.00000527, 0.00000659, -0.00001272, 0.00000566, -0.00001391, -0.00000213, 0.00000896, -0.00000257, -0.00000994, 0.00000474, -0.00000412, -0.00000120, -0.00001369, 0.00000330, -0.00000334, 0.00000700, -0.00001506, 0.00000947, -0.00002629, 0.00001942, -0.00002680, 0.00000521, -0.00000241, -0.00000356, -0.00000552, -0.00000660, -0.00001806, -0.00000595, -0.00003509, -0.00001160, -0.00015749, -0.00011630, 0.00039248, 0.00004954, 0.00038159, -0.00009402, -0.00017667, -0.00000661, -0.00004051, -0.00000824, -0.00001566, -0.00000638, -0.00001135, -0.00000616, -0.00000158, -0.00001151, 0.00000119, -0.00002548, 0.00000805, -0.00000563, 0.00000925, 0.00001669, -0.00001389, 0.00001366, -0.00000383, 0.00001271, -0.00000310, 0.00001625, -0.00001289, 0.00001564, 0.00002285, -0.00000630, -0.00000680, -0.00002092, 0.00001654, -0.00001667, 0.00000084, -0.00000850, 0.00000682, -0.00000339, -0.00000138, -0.00000532, -0.00000673, -0.00000244, 0.00000298, 0.00000246, 0.00000519, -0.00000283, -0.00000678, -0.00000056, -0.00000355, -0.00000165, 0.00000383, -0.00000146, -0.00000138, -0.00000213, -0.00000027, 0.00000673, -0.00000828, -0.00000248, 0.00000021, 0.00000930, -0.00001947, 0.00000564, 0.00000488, 0.00000202, -0.00000832, 0.00000751, -0.00000514, -0.00000409, 0.00001321, -0.00000164, },
		{ -0.00000737, 0.00000276, 0.00001501, -0.00000724, -0.00000145, -0.00000535, 0.00000052, 0.00000172, 0.00000368, 0.00000063, -0.00000556, 0.00000049, -0.00000069, 0.00000143, 0.00000926, -0.00000135, -0.00000417, 0.00000306, -0.00001116, 0.00000287, 0.00001461, -0.00000594, 0.00000609, -0.00000808, 0.00000986, -0.00001229, 0.00001427, -0.00000400, -0.00000563, 0.00000152, -0.00001397, 0.00000156, 0.00000754, -0.00000445, -0.00000793, 0.00000364, -0.00000408, -0.00000159, -0.00001366, 0.00000433, -0.00000235, 0.00000429, -0.00001399, 0.00000917, -0.00002714, 0.00002260, -0.00003422, 0.00001531, -0.00001108, 0.00000332, -0.00001292, 0.00000059, -0.00002759, 0.00000402, -0.00005005, 0.00000098, -0.00022341, 0.00003873, 0.00033107, 0.00005013, 0.00036921, -0.00015538, -0.00013176, -0.00001683, -0.00003095, -0.00001268, -0.00001049, -0.00001086, -0.00000764, -0.00000759, -0.00000022, -0.00001284, 0.00000133, -0.00002690, 0.00001234, -0.00001156, 0.00001855, 0.00000850, -0.00000897, 0.00001253, -0.00000202, 0.00001143, -0.00000160, 0.00001640, -0.00001588, 0.00002598, 0.00001324, -0.00000141, -0.00001652, -0.00001320, 0.00001457, -0.00001849, 0.00000344, -0.00001080, 0.00001091, -0.00000892, 0.00000255, -0.00000874, -0.00000435, -0.00000332, 0.00000584, -0.00000102, 0.00000743, -0.00000589, -0.00000456, -0.00000146, -0.00000288, -0.00000212, 0.00000551, -0.00000462, 0.00000155, -0.00000486, 0.00000423, 0.00000175, -0.00000736, -0.00000214, 0.00000452, 0.00000254, -0.00001725, 0.00000824, 0.00000337, 0.00000155, -0.00000865, 0.00001012, -0.00001089, 0.00000368, 0.00000925, -0.00000168, },
		{ -0.00000895, 0.00000824, 0.00001122, -0.00000730, -0.00000209, -0.00000437, 0.00000062, 0.00000202, 0.00000388, -0.00000076, -0.00000565, 0.00000224, -0.00000273, 0.00000542, 0.00000668, -0.00000212, -0.00000354, 0.00000312, -0.00001357, 0.00000943, 0.00000902, -0.00000296, 0.00000292, -0.00000490, 0.00000821, -0.00001341, 0.00001995, -0.00001405, 0.00000388, -0.00000698, -0.00000886, 0.00000122, 0.00000841, -0.00000820, -0.00000362, -0.00000037, -0.00000068, -0.00000538, -0.00001030, 0.00000245, 0.00000139, -0.00000228, -0.00000772, 0.00000301, -0.00002147, 0.00001829, -0.00003371, 0.00001824, -0.00001316, 0.00000280, -0.00001144, -0.00000282, -0.00002460, -0.00000142, -0.00004522, -0.00001894, -0.00024256, 0.00017746, 0.00023777, 0.00008540, 0.00032803, -0.00019076, -0.00009823, -0.00002176, -0.00002658, -0.00001231, -0.00001027, -0.00001099, -0.00000740, -0.00000586, -0.00000210, -0.00001116, -0.00000165, -0.00002453, 0.00001228, -0.00001334, 0.00002512, 0.00000121, -0.00000361, 0.00001100, -0.00000017, 0.00000978, 0.00000107, 0.00001432, -0.00001547, 0.00003368, 0.00000387, 0.00000380, -0.00002836, 0.00000002, 0.00000569, -0.00001371, 0.00000015, -0.00000752, 0.00000974, -0.00001041, 0.00000376, -0.00001005, -0.00000386, -0.00000176, 0.00000585, -0.00000245, 0.00000832, -0.00000786, -0.00000312, -0.00000152, -0.00000350, -0.00000078, 0.00000496, -0.00000577, 0.00000296, -0.00000607, 0.00000744, -0.00000344, -0.00000547, -0.00000123, 0.00000730, -0.00000466, -0.00001223, 0.00000806, 0.00000373, -0.00000090, -0.00000620, 0.00000932, -0.00001345, 0.00001019, 0.00000449, -0.00000117, },
		{ -0.00000905, 0.00001269, 0.00000703, -0.00000692, -0.00000265, -0.00000335, 0.00000057, 0.00000245, 0.00000400, -0.00000273, -0.00000429, 0.00000237, -0.00000322, 0.00000860, 0.00000350, -0.00000216, -0.00000277, 0.00000217, -0.00001428, 0.00001493, 0.00000284, 0.00000143, -0.00000214, 0.00000135, 0.00000189, -0.00000853, 0.00001921, -0.00001920, 0.00001109, -0.00001570, -0.00000154, -0.00000134, 0.00001021, -0.00001226, 0.00000113, -0.00000524, 0.00000410, -0.00001081, -0.00000525, -0.00000061, 0.00000563, -0.00000993, 0.00000087, -0.00000625, -0.00001199, 0.00000907, -0.00002749, 0.00001602, -0.00001107, -0.00000202, -0.00000471, -0.00001294, -0.00001328, -0.00001768, -0.00002588, -0.00006813, -0.00020442, 0.00027005, 0.00014435, 0.00013585, 0.00026916, -0.00020637, -0.00007314, -0.00002408, -0.00002436, -0.00001017, -0.00001221, -0.00000909, -0.00000856, -0.00000306, -0.00000520, -0.00000844, -0.00000570, -0.00002046, 0.00000955, -0.00001189, 0.00002920, -0.00000522, 0.00000253, 0.00000829, 0.00000260, 0.00000707, 0.00000557, 0.00000913, -0.00000993, 0.00003569, -0.00000136, 0.00000461, -0.00003658, 0.00001308, -0.00000558, -0.00000565, -0.00000655, -0.00000045, 0.00000428, -0.00000813, 0.00000229, -0.00000944, -0.00000478, 0.00000164, 0.00000324, -0.00000160, 0.00000762, -0.00000863, -0.00000236, -0.00000109, -0.00000485, 0.00000185, 0.00000243, -0.00000475, 0.00000247, -0.00000519, 0.00000832, -0.00000761, -0.00000322, 0.00000049, 0.00000760, -0.00001037, -0.00000631, 0.00000624, 0.00000516, -0.00000454, -0.00000163, 0.00000518, -0.00001189, 0.00001367, 0.00000056, -0.00000111, },
		{ -0.00000720, 0.00001525, 0.00000330, -0.00000658, -0.00000277, -0.00000265, 0.00000068, 0.00000287, 0.00000387, -0.00000486, -0.00000180, 0.00000084, -0.00000162, 0.00000994, 0.00000076, -0.00000198, -0.00000172, 0.00000008, -0.00001268, 0.00001803, -0.00000210, 0.00000533, -0.00000703, 0.00000849, -0.00000717, 0.00000132, 0.00001152, -0.00001713, 0.00001240, -0.00002059, 0.00000445, -0.00000358, 0.00001096, -0.00001468, 0.00000411, -0.00000848, 0.00000768, -0.00001540, -0.00000076, -0.00000262, 0.00000771, -0.00001528, 0.00000800, -0.00001470, -0.00000281, -0.00000081, -0.00001960, 0.00001250, -0.00000907, -0.00000594, 0.00000108, -0.00002256, -0.00000205, -0.00003452, -0.00000565, -0.00012975, -0.00011728, 0.00030367, 0.00007449, 0.00018283, 0.00020517, -0.00021014, -0.00005219, -0.00002726, -0.00002085, -0.00000966, -0.00001312, -0.00000790, -0.00000873, -0.00000144, -0.00000738, -0.00000669, -0.00000871, -0.00001694, 0.00000623, -0.00000854, 0.00003131, -0.00001075, 0.00000935, 0.00000422, 0.00000650, 0.00000334, 0.00001152, 0.00000143, 0.00000054, 0.00003114, -0.00000075, -0.00000168, -0.00003681, 0.00002061, -0.00001409, 0.00000118, -0.00001274, 0.00000699, -0.00000276, -0.00000394, -0.00000057, -0.00000807, -0.00000579, 0.00000536, -0.00000073, 0.00000079, 0.00000577, -0.00000862, -0.00000177, -0.00000091, -0.00000593, 0.00000460, -0.00000101, -0.00000231, 0.00000059, -0.00000252, 0.00000667, -0.00001005, -0.00000098, 0.00000277, 0.00000519, -0.00001319, -0.00000131, 0.00000430, 0.00000636, -0.00000792, 0.00000347, -0.00000091, -0.00000678, 0.00001358, -0.00000161, -0.00000204, },
		{ -0.00000364, 0.00001572, 0.00000042, -0.00000646, -0.00000240, -0.00000239, 0.00000105, 0.00000325, 0.00000324, -0.00000651, 0.00000101, -0.00000169, 0.00000182, 0.00000905, -0.00000086, -0.00000198, -0.00000043, -0.00000290, -0.00000880, 0.00001801, -0.00000448, 0.00000700, -0.00000952, 0.00001362, -0.00001565, 0.00001290, -0.00000081, -0.00000853, 0.00000683, -0.00001948, 0.00000663, -0.00000354, 0.00000929, -0.00001421, 0.00000395, -0.00000842, 0.00000817, -0.00001730, 0.00000148, -0.00000213, 0.00000598, -0.00001605, 0.00001082, -0.00001922, 0.00000265, -0.00000763, -0.00001369, 0.00001104, -0.00001069, -0.00000467, 0.00000053, -0.00002507, 0.00000092, -0.00004142, -0.00000015, -0.00017972, -0.00000471, 0.00028411, 0.00003808, 0.00021481, 0.00014508, -0.00020794, -0.00003269, -0.00003303, -0.00001432, -0.00001256, -0.00001117, -0.00000898, -0.00000665, -0.00000218, -0.00000757, -0.00000686, -0.00000966, -0.00001532, 0.00000379, -0.00000437, 0.00003171, -0.00001489, 0.00001604, -0.00000050, 0.00001085, -0.00000042, 0.00001739, -0.00000661, 0.00001350, 0.00002185, 0.00000434, -0.00001431, -0.00002805, 0.00001988, -0.00001638, 0.00000327, -0.00001497, 0.00001142, -0.00000833, -0.00000015, -0.00000315, -0.00000730, -0.00000545, 0.00000767, -0.00000438, 0.00000353, 0.00000357, -0.00000840, -0.00000085, -0.00000166, -0.00000576, 0.00000619, -0.00000397, 0.00000032, -0.00000160, 0.00000090, 0.00000304, -0.00001062, 0.00000121, 0.00000490, 0.00000075, -0.00001278, 0.00000158, 0.00000345, 0.00000617, -0.00000956, 0.00000718, -0.00000677, 0.00000014, 0.00001063, -0.00000207, -0.00000377, },
		{ 0.00000078, 0.00001449, -0.00000163, -0.00000644, -0.00000177, -0.00000240, 0.00000162, 0.00000363, 0.00000191, -0.00000704, 0.00000313, -0.00000410, 0.00000614, 0.00000628, -0.00000117, -0.00000228, 0.00000084, -0.00000611, -0.00000340, 0.00001517, -0.00000393, 0.00000567, -0.00000822, 0.00001437, -0.00002011, 0.00002201, -0.00001351, 0.00000324, -0.00000369, -0.00001279, 0.00000450, -0.00000077, 0.00000511, -0.00001098, 0.00000078, -0.00000500, 0.00000522, -0.00001604, 0.00000122, 0.00000074, 0.00000069, -0.00001205, 0.00000867, -0.00001888, 0.00000319, -0.00000985, -0.00001128, 0.00001269, -0.00001668, 0.00000285, -0.00000807, -0.00001794, -0.00000793, -0.00003316, -0.00001953, -0.00019611, 0.00010337, 0.00023080, 0.00003171, 0.00022921, 0.00009227, -0.00020185, -0.00001485, -0.00004049, -0.00000562, -0.00001826, -0.00000668, -0.00001203, -0.00000274, -0.00000487, -0.00000617, -0.00000847, -0.00000897, -0.00001558, 0.00000267, 0.00000023, 0.00003026, -0.00001679, 0.00002129, -0.00000452, 0.00001438, -0.00000269, 0.00002112, -0.00001204, 0.00002526, 0.00001153, 0.00001014, -0.00002932, -0.00001309, 0.00001182, -0.00001217, -0.00000029, -0.00001190, 0.00001105, -0.00001049, 0.00000159, -0.00000429, -0.00000788, -0.00000303, 0.00000749, -0.00000643, 0.00000560, 0.00000167, -0.00000835, 0.00000059, -0.00000351, -0.00000395, 0.00000586, -0.00000538, 0.00000198, -0.00000290, 0.00000374, -0.00000149, -0.00000964, 0.00000349, 0.00000594, -0.00000433, -0.00000981, 0.00000220, 0.00000410, 0.00000412, -0.00000867, 0.00000808, -0.00001032, 0.00000662, 0.00000635, -0.00000160, -0.00000553, },
		{ 0.00000496, 0.00001233, -0.00000317, -0.00000621, -0.00000126, -0.00000237, 0.00000220, 0.00000399, -0.00000013, -0.00000608, 0.00000375, -0.00000519, 0.00001001, 0.00000258, -0.00000051, -0.00000272, 0.00000168, -0.00000866, 0.00000225, 0.00001066, -0.00000126, 0.00000188, -0.00000328, 0.00000999, -0.00001851, 0.00002520, -0.00002213, 0.00001358, -0.00001530, -0.00000310, -0.00000051, 0.00000369, -0.00000037, -0.00000642, -0.00000386, 0.00000035, 0.00000002, -0.00001266, -0.00000041, 0.00000436, -0.00000613, -0.00000525, 0.00000319, -0.00001519, 0.00000016, -0.00000855, -0.00001129, 0.00001577, -0.00002463, 0.00001380, -0.00002183, -0.00000409, -0.00002561, -0.00001286, -0.00006310, -0.00016769, 0.00018119, 0.00016792, 0.00004369, 0.00023021, 0.00004546, -0.00019088, -0.00000108, -0.00004673, 0.00000248, -0.00002429, -0.00000166, -0.00001528, 0.00000131, -0.00000794, -0.00000462, -0.00001002, -0.00000813, -0.00001663, 0.00000245, 0.00000532, 0.00002675, -0.00001571, 0.00002385, -0.00000655, 0.00001592, -0.00000228, 0.00002121, -0.00001248, 0.00003235, 0.00000399, 0.00001223, -0.00004127, 0.00000278, 0.00000041, -0.00000416, -0.00000765, -0.00000470, 0.00000632, -0.00000906, 0.00000097, -0.00000392, -0.00000958, 0.00000111, 0.00000476, -0.00000639, 0.00000654, 0.00000028, -0.00000840, 0.00000220, -0.00000599, -0.00000086, 0.00000366, -0.00000488, 0.00000212, -0.00000263, 0.00000493, -0.00000575, -0.00000761, 0.00000593, 0.00000518, -0.00000849, -0.00000572, 0.00000130, 0.00000577, 0.00000071, -0.00000551, 0.00000586, -0.00001038, 0.00001082, 0.00000236, -0.00000118, -0.00000640, },
		{ 0.00000797, 0.00001001, -0.00000457, -0.00000553, -0.00000116, -0.00000203, 0.00000265, 0.00000425, -0.00000257, -0.00000378, 0.00000260, -0.00000421, 0.00001221, -0.00000092, 0.00000051, -0.00000299, 0.00000171, -0.00000974, 0.00000676, 0.00000613, 0.00000192, -0.00000268, 0.00000359, 0.00000176, -0.00001109, 0.00002119, -0.00002380, 0.00001858, -0.00002384, 0.00000608, -0.00000597, 0.00000799, -0.00000533, -0.00000246, -0.00000783, 0.00000543, -0.00000537, -0.00000897, -0.00000162, 0.00000653, -0.00001163, 0.00000132, -0.00000267, -0.00001106, -0.00000348, -0.00000651, -0.00001098, 0.00001698, -0.00003024, 0.00002292, -0.00003475, 0.00000964, -0.00004401, 0.00000897, -0.00011866, -0.00009826, 0.00021520, 0.00011549, 0.00006092, 0.00022440, 0.00000167, -0.00017317, 0.00000586, -0.00004867, 0.00000697, -0.00002784, 0.00000154, -0.00001672, 0.00000359, -0.00000968, -0.00000445, -0.00000997, -0.00000878, -0.00001706, 0.00000242, 0.00001099, 0.00002129, -0.00001151, 0.00002321, -0.00000600, 0.00001515, 0.00000091, 0.00001755, -0.00000730, 0.00003310, 0.00000143, 0.00000777, -0.00004567, 0.00001414, -0.00000938, 0.00000347, -0.00001524, 0.00000362, -0.00000055, -0.00000542, -0.00000111, -0.00000287, -0.00001134, 0.00000574, 0.00000043, -0.00000467, 0.00000650, -0.00000082, -0.00000816, 0.00000330, -0.00000820, 0.00000251, 0.00000041, -0.00000296, 0.00000098, -0.00000080, 0.00000397, -0.00000884, -0.00000496, 0.00000831, 0.00000241, -0.00001052, -0.00000204, 0.00000006, 0.00000744, -0.00000291, -0.00000131, 0.00000144, -0.00000706, 0.00001190, -0.00000027, -0.00000150, -0.00000574, },
		{ 0.00000938, 0.00000801, -0.00000598, -0.00000436, -0.00000155, -0.00000127, 0.00000292, 0.00000418, -0.00000484, -0.00000076, 0.00000008, -0.00000111, 0.00001206, -0.00000330, 0.00000132, -0.00000289, 0.00000085, -0.00000902, 0.00000917, 0.00000302, 0.00000392, -0.00000589, 0.00000974, -0.00000746, -0.00000037, 0.00001143, -0.00001838, 0.00001653, -0.00002656, 0.00001187, -0.00000963, 0.00001048, -0.00000833, -0.00000061, -0.00000944, 0.00000835, -0.00000918, -0.00000650, -0.00000108, 0.00000564, -0.00001362, 0.00000504, -0.00000625, -0.00000927, -0.00000482, -0.00000653, -0.00000775, 0.00001337, -0.00002961, 0.00002524, -0.00004079, 0.00001595, -0.00005391, 0.00001904, -0.00016675, -0.00000506, 0.00020684, 0.00008372, 0.00007438, 0.00021668, -0.00004089, -0.00014824, 0.00000486, -0.00004474, 0.00000619, -0.00002724, 0.00000149, -0.00001518, 0.00000308, -0.00000920, -0.00000634, -0.00000763, -0.00001184, -0.00001592, 0.00000215, 0.00001697, 0.00001459, -0.00000496, 0.00001987, -0.00000330, 0.00001278, 0.00000577, 0.00001171, 0.00000202, 0.00002824, 0.00000337, -0.00000312, -0.00004091, 0.00001771, -0.00001375, 0.00000703, -0.00001959, 0.00000978, -0.00000671, -0.00000169, -0.00000317, -0.00000236, -0.00001184, 0.00000928, -0.00000404, -0.00000217, 0.00000601, -0.00000202, -0.00000717, 0.00000325, -0.00000917, 0.00000495, -0.00000265, -0.00000066, -0.00000057, 0.00000182, 0.00000113, -0.00001035, -0.00000194, 0.00001012, -0.00000186, -0.00000997, 0.00000012, -0.00000044, 0.00000806, -0.00000539, 0.00000227, -0.00000342, -0.00000164, 0.00001019, -0.00000131, -0.00000260, -0.00000350, },
		{ 0.00000933, 0.00000643, -0.00000731, -0.00000294, -0.00000227, -0.00000020, 0.00000307, 0.00000354, -0.00000631, 0.00000202, -0.00000283, 0.00000339, 0.00000966, -0.00000414, 0.00000157, -0.00000244, -0.00000062, -0.00000675, 0.00000929, 0.00000201, 0.00000371, -0.00000616, 0.00001267, -0.00001433, 0.00000989, -0.00000055, -0.00000836, 0.00000846, -0.00002303, 0.00001304, -0.00001042, 0.00001052, -0.00000893, -0.00000122, -0.00000819, 0.00000842, -0.00001065, -0.00000575, 0.00000133, 0.00000161, -0.00001152, 0.00000492, -0.00000646, -0.00001096, -0.00000255, -0.00000981, -0.00000077, 0.00000415, -0.00002145, 0.00001863, -0.00003693, 0.00001083, -0.00004974, 0.00000747, -0.00018806, 0.00008770, 0.00016973, 0.00007212, 0.00008147, 0.00020817, -0.00008128, -0.00011812, -0.00000275, -0.00003577, 0.00000066, -0.00002270, -0.00000160, -0.00001102, 0.00000010, -0.00000687, -0.00000977, -0.00000355, -0.00001709, -0.00001313, 0.00000180, 0.00002257, 0.00000786, 0.00000251, 0.00001512, 0.00000039, 0.00001020, 0.00001038, 0.00000628, 0.00001240, 0.00002047, 0.00000705, -0.00001732, -0.00002896, 0.00001358, -0.00001168, 0.00000502, -0.00001886, 0.00001164, -0.00001003, 0.00000043, -0.00000403, -0.00000319, -0.00001026, 0.00001054, -0.00000736, 0.00000018, 0.00000553, -0.00000355, -0.00000530, 0.00000179, -0.00000841, 0.00000564, -0.00000443, 0.00000095, -0.00000149, 0.00000409, -0.00000277, -0.00001032, 0.00000132, 0.00001068, -0.00000656, -0.00000728, 0.00000044, 0.00000029, 0.00000710, -0.00000586, 0.00000382, -0.00000681, 0.00000397, 0.00000688, -0.00000129, -0.00000396, -0.00000028, },
		{ 0.00000840, 0.00000507, -0.00000829, -0.00000162, -0.00000297, 0.00000097, 0.00000316, 0.00000224, -0.00000647, 0.00000366, -0.00000493, 0.00000805, 0.00000573, -0.00000357, 0.00000129, -0.00000187, -0.00000213, -0.00000375, 0.00000777, 0.00000279, 0.00000139, -0.00000317, 0.00001117, -0.00001646, 0.00001617, -0.00001059, 0.00000222, -0.00000245, -0.00001503, 0.00001020, -0.00000863, 0.00000859, -0.00000785, -0.00000344, -0.00000489, 0.00000622, -0.00001027, -0.00000604, 0.00000445, -0.00000407, -0.00000664, 0.00000193, -0.00000425, -0.00001516, 0.00000248, -0.00001540, 0.00000848, -0.00000868, -0.00000780, 0.00000489, -0.00002475, -0.00000432, -0.00003267, -0.00002714, -0.00017096, 0.00015744, 0.00012300, 0.00007276, 0.00008498, 0.00019664, -0.00011621, -0.00008684, -0.00001358, -0.00002447, -0.00000732, -0.00001606, -0.00000611, -0.00000577, -0.00000394, -0.00000404, -0.00001327, 0.00000074, -0.00002331, -0.00000936, 0.00000191, 0.00002690, 0.00000240, 0.00000933, 0.00001043, 0.00000376, 0.00000880, 0.00001301, 0.00000380, 0.00002055, 0.00001308, 0.00000880, -0.00003009, -0.00001432, 0.00000484, -0.00000504, -0.00000146, -0.00001352, 0.00000906, -0.00000991, 0.00000032, -0.00000335, -0.00000538, -0.00000665, 0.00000919, -0.00000885, 0.00000188, 0.00000521, -0.00000526, -0.00000289, -0.00000072, -0.00000610, 0.00000445, -0.00000448, 0.00000126, -0.00000104, 0.00000494, -0.00000668, -0.00000905, 0.00000471, 0.00000948, -0.00001040, -0.00000353, -0.00000062, 0.00000202, 0.00000477, -0.00000430, 0.00000280, -0.00000746, 0.00000799, 0.00000345, -0.00000109, -0.00000472, 0.00000299, },
		{ 0.00000730, 0.00000361, -0.00000865, -0.00000076, -0.00000332, 0.00000200, 0.00000321, 0.00000040, -0.00000526, 0.00000368, -0.00000528, 0.00001158, 0.00000141, -0.00000214, 0.00000070, -0.00000148, -0.00000304, -0.00000107, 0.00000579, 0.00000421, -0.00000185, 0.00000194, 0.00000578, -0.00001334, 0.00001662, -0.00001554, 0.00000947, -0.00001234, -0.00000561, 0.00000525, -0.00000553, 0.00000591, -0.00000642, -0.00000573, -0.00000115, 0.00000321, -0.00000931, -0.00000602, 0.00000647, -0.00000906, -0.00000141, -0.00000165, -0.00000192, -0.00001939, 0.00000781, -0.00002075, 0.00001683, -0.00002115, 0.00000674, -0.00001102, -0.00000967, -0.00002325, -0.00001058, -0.00007632, -0.00011597, 0.00019052, 0.00008382, 0.00007563, 0.00008982, 0.00017862, -0.00014178, -0.00005892, -0.00002372, -0.00001420, -0.00001480, -0.00000976, -0.00000996, -0.00000137, -0.00000727, -0.00000227, -0.00001521, 0.00000345, -0.00002896, -0.00000559, 0.00000309, 0.00002936, -0.00000094, 0.00001442, 0.00000685, 0.00000609, 0.00000915, 0.00001299, 0.00000554, 0.00002427, 0.00000837, 0.00000596, -0.00003716, -0.00000210, -0.00000403, 0.00000256, -0.00000945, -0.00000594, 0.00000366, -0.00000726, -0.00000147, -0.00000170, -0.00000813, -0.00000195, 0.00000584, -0.00000860, 0.00000289, 0.00000488, -0.00000671, -0.00000060, -0.00000344, -0.00000308, 0.00000200, -0.00000309, 0.00000034, 0.00000089, 0.00000378, -0.00000968, -0.00000693, 0.00000795, 0.00000641, -0.00001234, -0.00000002, -0.00000214, 0.00000397, 0.00000194, -0.00000156, -0.00000028, -0.00000515, 0.00000942, 0.00000102, -0.00000144, -0.00000419, 0.00000537, },
		{ 0.00000654, 0.00000184, -0.00000828, -0.00000054, -0.00000313, 0.00000275, 0.00000310, -0.00000158, -0.00000306, 0.00000217, -0.00000352, 0.00001302, -0.00000223, -0.00000050, 0.00000011, -0.00000149, -0.00000294, 0.00000044, 0.00000456, 0.00000491, -0.00000424, 0.00000703, -0.00000133, -0.00000654, 0.00001174, -0.00001450, 0.00001114, -0.00001813, 0.00000227, 0.00000030, -0.00000257, 0.00000369, -0.00000585, -0.00000669, 0.00000147, 0.00000084, -0.00000900, -0.00000457, 0.00000591, -0.00001128, 0.00000182, -0.00000351, -0.00000189, -0.00002098, 0.00001069, -0.00002311, 0.00002108, -0.00002915, 0.00001721, -0.00002340, 0.00000171, -0.00003787, 0.00000522, -0.00012435, -0.00003558, 0.00018564, 0.00006198, 0.00007341, 0.00009967, 0.00015184, -0.00015542, -0.00003763, -0.00003029, -0.00000740, -0.00001955, -0.00000564, -0.00001163, 0.00000081, -0.00000874, -0.00000253, -0.00001458, 0.00000333, -0.00003276, -0.00000255, 0.00000562, 0.00002986, -0.00000204, 0.00001748, 0.00000458, 0.00000751, 0.00001075, 0.00001109, 0.00001096, 0.00002327, 0.00000666, -0.00000181, -0.00003656, 0.00000418, -0.00000922, 0.00000755, -0.00001567, 0.00000089, -0.00000209, -0.00000386, -0.00000369, -0.00000015, -0.00001026, 0.00000246, 0.00000164, -0.00000723, 0.00000353, 0.00000418, -0.00000740, 0.00000083, -0.00000542, -0.00000046, -0.00000062, -0.00000114, -0.00000110, 0.00000371, 0.00000069, -0.00001119, -0.00000431, 0.00001058, 0.00000187, -0.00001192, 0.00000223, -0.00000317, 0.00000527, -0.00000026, 0.00000100, -0.00000405, -0.00000077, 0.00000830, -0.00000000, -0.00000255, -0.00000217, 0.00000633, },
		{ 0.00000634, -0.00000023, -0.00000727, -0.00000087, -0.00000241, 0.00000320, 0.00000267, -0.00000315, -0.00000063, -0.00000016, -0.00000004, 0.00001213, -0.00000450, 0.00000083, -0.00000032, -0.00000184, -0.00000188, 0.00000048, 0.00000471, 0.00000392, -0.00000432, 0.00000985, -0.00000735, 0.00000095, 0.00000405, -0.00000894, 0.00000738, -0.00001866, 0.00000683, -0.00000319, -0.00000064, 0.00000251, -0.00000660, -0.00000576, 0.00000223, -0.00000015, -0.00000981, -0.00000148, 0.00000248, -0.00000995, 0.00000189, -0.00000251, -0.00000540, -0.00001839, 0.00000950, -0.00002098, 0.00001953, -0.00003038, 0.00002053, -0.00002845, 0.00000479, -0.00004218, 0.00000509, -0.00015407, 0.00005032, 0.00015259, 0.00005831, 0.00006398, 0.00011495, 0.00011675, -0.00015710, -0.00002392, -0.00003238, -0.00000479, -0.00002086, -0.00000424, -0.00001081, 0.00000050, -0.00000819, -0.00000476, -0.00001148, 0.00000013, -0.00003423, -0.00000048, 0.00000930, 0.00002885, -0.00000145, 0.00001895, 0.00000317, 0.00000882, 0.00001232, 0.00000918, 0.00001801, 0.00001899, 0.00000635, -0.00001235, -0.00002941, 0.00000385, -0.00000911, 0.00000798, -0.00001795, 0.00000470, -0.00000603, -0.00000141, -0.00000513, 0.00000043, -0.00001086, 0.00000539, -0.00000221, -0.00000547, 0.00000412, 0.00000290, -0.00000707, 0.00000100, -0.00000603, 0.00000086, -0.00000234, 0.00000034, -0.00000205, 0.00000637, -0.00000367, -0.00001108, -0.00000141, 0.00001208, -0.00000324, -0.00000938, 0.00000281, -0.00000320, 0.00000536, -0.00000100, 0.00000211, -0.00000687, 0.00000408, 0.00000555, 0.00000003, -0.00000404, 0.00000095, 0.00000583, },
		{ 0.00000654, -0.00000239, -0.00000594, -0.00000152, -0.00000131, 0.00000339, 0.00000179, -0.00000382, 0.00000119, -0.00000226, 0.00000417, 0.00000934, -0.00000526, 0.00000164, -0.00000061, -0.00000221, -0.00000037, -0.00000054, 0.00000610, 0.00000124, -0.00000162, 0.00000907, -0.00000996, 0.00000607, -0.00000302, -0.00000202, 0.00000038, -0.00001479, 0.00000788, -0.00000487, 0.00000023, 0.00000209, -0.00000818, -0.00000343, 0.00000143, -0.00000004, -0.00001124, 0.00000234, -0.00000273, -0.00000588, -0.00000070, 0.00000084, -0.00001187, -0.00001198, 0.00000450, -0.00001486, 0.00001287, -0.00002530, 0.00001678, -0.00002597, -0.00000091, -0.00003531, -0.00001444, -0.00015332, 0.00012139, 0.00010726, 0.00006660, 0.00005007, 0.00013280, 0.00007645, -0.00014920, -0.00001660, -0.00003098, -0.00000541, -0.00001951, -0.00000479, -0.00000832, -0.00000147, -0.00000643, -0.00000797, -0.00000695, -0.00000545, -0.00003361, 0.00000090, 0.00001366, 0.00002701, 0.00000004, 0.00001957, 0.00000198, 0.00001088, 0.00001253, 0.00000929, 0.00002410, 0.00001367, 0.00000501, -0.00002199, -0.00001920, -0.00000101, -0.00000468, 0.00000416, -0.00001611, 0.00000482, -0.00000723, -0.00000077, -0.00000521, -0.00000025, -0.00000970, 0.00000627, -0.00000493, -0.00000382, 0.00000475, 0.00000116, -0.00000590, 0.00000003, -0.00000523, 0.00000059, -0.00000259, 0.00000066, -0.00000169, 0.00000775, -0.00000830, -0.00000957, 0.00000153, 0.00001204, -0.00000784, -0.00000556, 0.00000191, -0.00000223, 0.00000427, -0.00000009, 0.00000112, -0.00000746, 0.00000778, 0.00000249, 0.00000033, -0.00000513, 0.00000434, 0.00000435, },
		{ 0.00000678, -0.00000432, -0.00000463, -0.00000215, -0.00000009, 0.00000336, 0.00000053, -0.00000340, 0.00000180, -0.00000316, 0.00000788, 0.00000560, -0.00000484, 0.00000200, -0.00000093, -0.00000217, 0.00000081, -0.00000172, 0.00000782, -0.00000220, 0.00000304, 0.00000487, -0.00000835, 0.00000692, -0.00000663, 0.00000294, -0.00000666, -0.00000876, 0.00000650, -0.00000528, 0.00000064, 0.00000162, -0.00000951, -0.00000087, 0.00000012, 0.00000022, -0.00001219, 0.00000533, -0.00000774, -0.00000103, -0.00000415, 0.00000464, -0.00001924, -0.00000372, -0.00000242, -0.00000697, 0.00000374, -0.00001679, 0.00000893, -0.00001917, -0.00001160, -0.00002212, -0.00004865, -0.00011945, 0.00016345, 0.00006542, 0.00007780, 0.00003696, 0.00014862, 0.00003543, -0.00013540, -0.00001326, -0.00002805, -0.00000738, -0.00001709, -0.00000587, -0.00000559, -0.00000380, -0.00000470, -0.00001080, -0.00000251, -0.00001214, -0.00003165, 0.00000208, 0.00001816, 0.00002488, 0.00000180, 0.00001983, 0.00000071, 0.00001400, 0.00001086, 0.00001267, 0.00002717, 0.00000904, 0.00000081, -0.00002730, -0.00001017, -0.00000680, 0.00000131, -0.00000177, -0.00001171, 0.00000221, -0.00000610, -0.00000178, -0.00000415, -0.00000180, -0.00000727, 0.00000530, -0.00000633, -0.00000241, 0.00000524, -0.00000063, -0.00000443, -0.00000147, -0.00000359, -0.00000090, -0.00000148, -0.00000027, 0.00000024, 0.00000708, -0.00001213, -0.00000708, 0.00000429, 0.00001032, -0.00001099, -0.00000151, 0.00000021, -0.00000077, 0.00000258, 0.00000190, -0.00000171, -0.00000541, 0.00000924, 0.00000024, 0.00000014, -0.00000505, 0.00000703, 0.00000260, },
		{ 0.00000670, -0.00000575, -0.00000361, -0.00000251, 0.00000107, 0.00000308, -0.00000087, -0.00000209, 0.00000112, -0.00000231, 0.00001007, 0.00000195, -0.00000381, 0.00000214, -0.00000145, -0.00000140, 0.00000098, -0.00000208, 0.00000869, -0.00000485, 0.00000785, -0.00000110, -0.00000353, 0.00000353, -0.00000566, 0.00000373, -0.00001095, -0.00000309, 0.00000428, -0.00000535, 0.00000133, 0.00000028, -0.00000956, 0.00000071, -0.00000055, -0.00000036, -0.00001162, 0.00000609, -0.00001062, 0.00000247, -0.00000640, 0.00000666, -0.00002495, 0.00000378, -0.00000874, -0.00000007, -0.00000456, -0.00000865, 0.00000119, -0.00001287, -0.00002131, -0.00001101, -0.00008609, -0.00006010, 0.00017214, 0.00003766, 0.00008421, 0.00002950, 0.00015821, -0.00000216, -0.00011919, -0.00001164, -0.00002541, -0.00000900, -0.00001504, -0.00000623, -0.00000382, -0.00000540, -0.00000396, -0.00001218, 0.00000052, -0.00001868, -0.00002910, 0.00000355, 0.00002250, 0.00002265, 0.00000369, 0.00001975, -0.00000031, 0.00001764, 0.00000789, 0.00001914, 0.00002651, 0.00000558, -0.00000635, -0.00002671, -0.00000540, -0.00001024, 0.00000584, -0.00000711, -0.00000711, -0.00000125, -0.00000394, -0.00000355, -0.00000266, -0.00000338, -0.00000456, 0.00000324, -0.00000670, -0.00000110, 0.00000528, -0.00000195, -0.00000333, -0.00000268, -0.00000202, -0.00000271, 0.00000030, -0.00000191, 0.00000339, 0.00000416, -0.00001438, -0.00000411, 0.00000658, 0.00000717, -0.00001214, 0.00000179, -0.00000154, 0.00000051, 0.00000117, 0.00000393, -0.00000537, -0.00000132, 0.00000830, -0.00000066, -0.00000092, -0.00000339, 0.00000827, 0.00000126, },
		{ 0.00000608, -0.00000659, -0.00000298, -0.00000248, 0.00000206, 0.00000246, -0.00000198, -0.00000044, -0.00000042, 0.00000016, 0.00001025, -0.00000080, -0.00000272, 0.00000226, -0.00000219, 0.00000012, -0.00000013, -0.00000102, 0.00000781, -0.00000532, 0.00001077, -0.00000643, 0.00000214, -0.00000220, -0.00000110, 0.00000015, -0.00001112, 0.00000044, 0.00000250, -0.00000573, 0.00000266, -0.00000219, -0.00000788, 0.00000068, 0.00000003, -0.00000223, -0.00000923, 0.00000411, -0.00001036, 0.00000330, -0.00000612, 0.00000542, -0.00002711, 0.00000850, -0.00001257, 0.00000385, -0.00000963, -0.00000371, -0.00000316, -0.00001096, -0.00002501, -0.00001007, -0.00011286, 0.00000976, 0.00015291, 0.00002723, 0.00008213, 0.00003010, 0.00015929, -0.00003371, -0.00010288, -0.00001047, -0.00002386, -0.00000948, -0.00001400, -0.00000538, -0.00000349, -0.00000585, -0.00000445, -0.00001180, 0.00000148, -0.00002426, -0.00002651, 0.00000554, 0.00002659, 0.00002017, 0.00000595, 0.00001890, -0.00000031, 0.00002060, 0.00000521, 0.00002725, 0.00002271, 0.00000244, -0.00001468, -0.00002111, -0.00000569, -0.00000972, 0.00000706, -0.00000989, -0.00000430, -0.00000375, -0.00000219, -0.00000505, -0.00000150, -0.00000424, -0.00000253, 0.00000097, -0.00000649, 0.00000026, 0.00000468, -0.00000251, -0.00000303, -0.00000301, -0.00000134, -0.00000390, 0.00000184, -0.00000336, 0.00000685, -0.00000057, -0.00001463, -0.00000114, 0.00000811, 0.00000317, -0.00001126, 0.00000376, -0.00000274, 0.00000110, 0.00000077, 0.00000488, -0.00000845, 0.00000354, 0.00000560, -0.00000036, -0.00000264, -0.00000033, 0.00000781, 0.00000076, },
		{ 0.00000491, -0.00000687, -0.00000267, -0.00000209, 0.00000283, 0.00000144, -0.00000238, 0.00000084, -0.00000195, 0.00000353, 0.00000858, -0.00000230, -0.00000190, 0.00000242, -0.00000294, 0.00000200, -0.00000223, 0.00000137, 0.00000507, -0.00000303, 0.00001041, -0.00000893, 0.00000598, -0.00000745, 0.00000454, -0.00000614, -0.00000759, 0.00000132, 0.00000166, -0.00000649, 0.00000436, -0.00000535, -0.00000489, -0.00000075, 0.00000166, -0.00000502, -0.00000566, 0.00000007, -0.00000730, 0.00000149, -0.00000335, 0.00000085, -0.00002529, 0.00000980, -0.00001343, 0.00000443, -0.00001100, -0.00000267, -0.00000331, -0.00001447, -0.00002121, -0.00002316, -0.00011789, 0.00007321, 0.00011757, 0.00003084, 0.00007221, 0.00003820, 0.00015194, -0.00005852, -0.00008729, -0.00000970, -0.00002311, -0.00000909, -0.00001365, -0.00000372, -0.00000426, -0.00000547, -0.00000574, -0.00001011, 0.00000044, -0.00002861, -0.00002406, 0.00000819, 0.00003041, 0.00001726, 0.00000890, 0.00001686, 0.00000150, 0.00002166, 0.00000467, 0.00003487, 0.00001706, -0.00000178, -0.00002130, -0.00001342, -0.00000944, -0.00000581, 0.00000494, -0.00000969, -0.00000402, -0.00000433, -0.00000174, -0.00000564, -0.00000104, -0.00000409, -0.00000171, -0.00000084, -0.00000602, 0.00000164, 0.00000352, -0.00000235, -0.00000356, -0.00000229, -0.00000191, -0.00000387, 0.00000238, -0.00000375, 0.00000951, -0.00000621, -0.00001293, 0.00000143, 0.00000870, -0.00000093, -0.00000880, 0.00000431, -0.00000321, 0.00000088, 0.00000166, 0.00000405, -0.00000972, 0.00000768, 0.00000230, 0.00000045, -0.00000434, 0.00000347, 0.00000591, 0.00000116, },
		{ 0.00000334, -0.00000677, -0.00000254, -0.00000147, 0.00000335, 0.00000007, -0.00000186, 0.00000116, -0.00000257, 0.00000672, 0.00000578, -0.00000269, -0.00000135, 0.00000250, -0.00000336, 0.00000363, -0.00000454, 0.00000432, 0.00000124, 0.00000148, 0.00000666, -0.00000762, 0.00000616, -0.00000965, 0.00000832, -0.00001240, -0.00000213, 0.00000023, 0.00000149, -0.00000719, 0.00000573, -0.00000827, -0.00000159, -0.00000269, 0.00000352, -0.00000777, -0.00000222, -0.00000444, -0.00000290, -0.00000174, 0.00000064, -0.00000571, -0.00002053, 0.00000844, -0.00001222, 0.00000289, -0.00001006, -0.00000401, -0.00000104, -0.00002115, -0.00001288, -0.00004782, -0.00009706, 0.00011694, 0.00007943, 0.00004164, 0.00005805, 0.00005111, 0.00013790, -0.00007720, -0.00007232, -0.00000993, -0.00002222, -0.00000869, -0.00001316, -0.00000215, -0.00000533, -0.00000498, -0.00000700, -0.00000802, -0.00000198, -0.00003188, -0.00002173, 0.00001155, 0.00003379, 0.00001395, 0.00001255, 0.00001362, 0.00000535, 0.00002022, 0.00000756, 0.00004010, 0.00001067, -0.00000791, -0.00002370, -0.00000713, -0.00001373, -0.00000067, 0.00000106, -0.00000755, -0.00000571, -0.00000315, -0.00000262, -0.00000530, -0.00000113, -0.00000321, -0.00000207, -0.00000196, -0.00000540, 0.00000283, 0.00000218, -0.00000186, -0.00000448, -0.00000093, -0.00000351, -0.00000256, 0.00000167, -0.00000259, 0.00001038, -0.00001162, -0.00000968, 0.00000331, 0.00000832, -0.00000446, -0.00000545, 0.00000369, -0.00000310, 0.00000017, 0.00000353, 0.00000146, -0.00000860, 0.00000995, -0.00000046, 0.00000098, -0.00000520, 0.00000703, 0.00000323, 0.00000222, },
		{ 0.00000158, -0.00000646, -0.00000242, -0.00000072, 0.00000356, -0.00000142, -0.00000055, 0.00000034, -0.00000169, 0.00000872, 0.00000283, -0.00000245, -0.00000090, 0.00000231, -0.00000313, 0.00000435, -0.00000615, 0.00000671, -0.00000232, 0.00000666, 0.00000078, -0.00000311, 0.00000245, -0.00000764, 0.00000813, -0.00001603, 0.00000301, -0.00000155, 0.00000133, -0.00000723, 0.00000604, -0.00000996, 0.00000091, -0.00000410, 0.00000468, -0.00000946, -0.00000021, -0.00000769, 0.00000101, -0.00000467, 0.00000402, -0.00001225, -0.00001467, 0.00000598, -0.00001055, 0.00000116, -0.00000908, -0.00000507, 0.00000044, -0.00002683, -0.00000607, -0.00007595, -0.00005454, 0.00013496, 0.00004880, 0.00005265, 0.00004393, 0.00006553, 0.00011948, -0.00009075, -0.00005772, -0.00001171, -0.00002036, -0.00000906, -0.00001175, -0.00000148, -0.00000595, -0.00000501, -0.00000750, -0.00000644, -0.00000500, -0.00003444, -0.00001939, 0.00001575, 0.00003629, 0.00001073, 0.00001643, 0.00000978, 0.00001062, 0.00001673, 0.00001409, 0.00004187, 0.00000392, -0.00001545, -0.00002096, -0.00000474, -0.00001583, 0.00000323, -0.00000245, -0.00000531, -0.00000791, -0.00000121, -0.00000422, -0.00000446, -0.00000127, -0.00000222, -0.00000314, -0.00000247, -0.00000457, 0.00000355, 0.00000109, -0.00000157, -0.00000519, 0.00000038, -0.00000546, -0.00000045, 0.00000000, 0.00000005, 0.00000892, -0.00001571, -0.00000557, 0.00000436, 0.00000713, -0.00000690, -0.00000194, 0.00000236, -0.00000276, -0.00000048, 0.00000562, -0.00000220, -0.00000533, 0.00000993, -0.00000197, 0.00000068, -0.00000463, 0.00000946, 0.00000054, 0.00000352, },
		{ -0.00000013, -0.00000605, -0.00000224, 0.00000008, 0.00000336, -0.00000266, 0.00000112, -0.00000129, 0.00000074, 0.00000890, 0.00000056, -0.00000212, -0.00000032, 0.00000173, -0.00000209, 0.00000379, -0.00000639, 0.00000760, -0.00000417, 0.00001059, -0.00000509, 0.00000267, -0.00000369, -0.00000209, 0.00000356, -0.00001556, 0.00000606, -0.00000283, 0.00000059, -0.00000627, 0.00000495, -0.00000991, 0.00000192, -0.00000432, 0.00000457, -0.00000962, -0.00000032, -0.00000857, 0.00000307, -0.00000597, 0.00000533, -0.00001700, -0.00000945, 0.00000382, -0.00000973, 0.00000074, -0.00000987, -0.00000371, -0.00000162, -0.00002777, -0.00000704, -0.00009683, -0.00000100, 0.00012922, 0.00003060, 0.00005929, 0.00003303, 0.00007886, 0.00009855, -0.00009983, -0.00004375, -0.00001493, -0.00001733, -0.00001043, -0.00000921, -0.00000203, -0.00000580, -0.00000575, -0.00000698, -0.00000592, -0.00000804, -0.00003660, -0.00001688, 0.00002092, 0.00003736, 0.00000832, 0.00001970, 0.00000642, 0.00001605, 0.00001257, 0.00002331, 0.00004000, -0.00000357, -0.00002253, -0.00001432, -0.00000672, -0.00001443, 0.00000432, -0.00000396, -0.00000456, -0.00000915, 0.00000032, -0.00000574, -0.00000363, -0.00000098, -0.00000181, -0.00000431, -0.00000262, -0.00000348, 0.00000361, 0.00000056, -0.00000185, -0.00000518, 0.00000097, -0.00000698, 0.00000169, -0.00000191, 0.00000358, 0.00000519, -0.00001770, -0.00000138, 0.00000461, 0.00000538, -0.00000803, 0.00000113, 0.00000076, -0.00000250, -0.00000053, 0.00000705, -0.00000580, -0.00000086, 0.00000797, -0.00000215, -0.00000048, -0.00000248, 0.00001019, -0.00000149, 0.00000467, },
		{ -0.00000166, -0.00000558, -0.00000199, 0.00000090, 0.00000266, -0.00000325, 0.00000249, -0.00000304, 0.00000417, 0.00000725, -0.00000056, -0.00000207, 0.00000054, 0.00000081, -0.00000042, 0.00000203, -0.00000512, 0.00000661, -0.00000351, 0.00001175, -0.00000884, 0.00000729, -0.00000978, 0.00000485, -0.00000404, -0.00001109, 0.00000629, -0.00000298, -0.00000087, -0.00000443, 0.00000270, -0.00000829, 0.00000144, -0.00000333, 0.00000333, -0.00000855, -0.00000226, -0.00000702, 0.00000293, -0.00000531, 0.00000412, -0.00001918, -0.00000573, 0.00000256, -0.00001008, 0.00000189, -0.00001275, 0.00000051, -0.00000787, -0.00002279, -0.00001905, -0.00010128, 0.00005037, 0.00010757, 0.00002437, 0.00006018, 0.00002659, 0.00008974, 0.00007614, -0.00010449, -0.00003125, -0.00001875, -0.00001370, -0.00001229, -0.00000600, -0.00000351, -0.00000512, -0.00000689, -0.00000575, -0.00000646, -0.00001083, -0.00003858, -0.00001400, 0.00002702, 0.00003657, 0.00000743, 0.00002153, 0.00000465, 0.00002021, 0.00000957, 0.00003353, 0.00003490, -0.00001230, -0.00002669, -0.00000658, -0.00001153, -0.00001005, 0.00000257, -0.00000307, -0.00000595, -0.00000861, 0.00000064, -0.00000657, -0.00000313, -0.00000007, -0.00000237, -0.00000510, -0.00000259, -0.00000223, 0.00000310, 0.00000057, -0.00000275, -0.00000430, 0.00000048, -0.00000747, 0.00000315, -0.00000326, 0.00000702, -0.00000023, -0.00001722, 0.00000209, 0.00000428, 0.00000334, -0.00000786, 0.00000338, -0.00000078, -0.00000246, 0.00000028, 0.00000714, -0.00000826, 0.00000354, 0.00000496, -0.00000149, -0.00000204, 0.00000083, 0.00000918, -0.00000249, 0.00000538, },
		{ -0.00000294, -0.00000504, -0.00000171, 0.00000169, 0.00000146, -0.00000293, 0.00000300, -0.00000403, 0.00000762, 0.00000436, -0.00000055, -0.00000239, 0.00000160, -0.00000024, 0.00000146, -0.00000042, -0.00000281, 0.00000416, -0.00000050, 0.00000964, -0.00000921, 0.00000886, -0.00001339, 0.00001041, -0.00001218, -0.00000413, 0.00000411, -0.00000208, -0.00000277, -0.00000226, -0.00000001, -0.00000582, 0.00000007, -0.00000168, 0.00000160, -0.00000716, -0.00000493, -0.00000397, 0.00000124, -0.00000334, 0.00000103, -0.00001910, -0.00000342, 0.00000183, -0.00001086, 0.00000358, -0.00001645, 0.00000608, -0.00001651, -0.00001415, -0.00004056, -0.00008521, 0.00008833, 0.00008003, 0.00002617, 0.00005650, 0.00002422, 0.00009787, 0.00005275, -0.00010454, -0.00002132, -0.00002197, -0.00001044, -0.00001377, -0.00000297, -0.00000523, -0.00000443, -0.00000781, -0.00000444, -0.00000771, -0.00001339, -0.00004048, -0.00001041, 0.00003369, 0.00003391, 0.00000834, 0.00002152, 0.00000511, 0.00002212, 0.00000930, 0.00004292, 0.00002697, -0.00002209, -0.00002611, -0.00000088, -0.00001651, -0.00000462, -0.00000070, -0.00000067, -0.00000899, -0.00000640, -0.00000037, -0.00000659, -0.00000290, 0.00000125, -0.00000387, -0.00000529, -0.00000241, -0.00000107, 0.00000233, 0.00000084, -0.00000395, -0.00000283, -0.00000098, -0.00000678, 0.00000356, -0.00000346, 0.00000938, -0.00000633, -0.00001443, 0.00000427, 0.00000369, 0.00000124, -0.00000657, 0.00000463, -0.00000213, -0.00000255, 0.00000185, 0.00000573, -0.00000886, 0.00000677, 0.00000196, -0.00000068, -0.00000327, 0.00000449, 0.00000684, -0.00000242, 0.00000551, },
		{ -0.00000394, -0.00000442, -0.00000144, 0.00000238, -0.00000005, -0.00000172, 0.00000235, -0.00000358, 0.00001006, 0.00000112, 0.00000017, -0.00000290, 0.00000266, -0.00000111, 0.00000304, -0.00000282, -0.00000031, 0.00000128, 0.00000378, 0.00000502, -0.00000628, 0.00000675, -0.00001311, 0.00001239, -0.00001824, 0.00000319, 0.00000064, -0.00000069, -0.00000457, -0.00000038, -0.00000236, -0.00000338, -0.00000139, -0.00000012, 0.00000015, -0.00000647, -0.00000696, -0.00000079, -0.00000078, -0.00000130, -0.00000264, -0.00001781, -0.00000172, 0.00000076, -0.00001081, 0.00000414, -0.00001889, 0.00001037, -0.00002418, -0.00000665, -0.00006538, -0.00005102, 0.00010689, 0.00005523, 0.00003109, 0.00005054, 0.00002482, 0.00010336, 0.00002882, -0.00010004, -0.00001470, -0.00002361, -0.00000841, -0.00001406, -0.00000095, -0.00000650, -0.00000425, -0.00000798, -0.00000375, -0.00000912, -0.00001592, -0.00004232, -0.00000570, 0.00004022, 0.00002989, 0.00001077, 0.00001994, 0.00000769, 0.00002164, 0.00001260, 0.00004997, 0.00001634, -0.00003167, -0.00002058, 0.00000069, -0.00001919, -0.00000033, -0.00000363, 0.00000169, -0.00001245, -0.00000331, -0.00000230, -0.00000602, -0.00000264, 0.00000250, -0.00000597, -0.00000490, -0.00000207, -0.00000024, 0.00000168, 0.00000095, -0.00000497, -0.00000131, -0.00000291, -0.00000518, 0.00000298, -0.00000230, 0.00000989, -0.00001191, -0.00000999, 0.00000495, 0.00000314, -0.00000075, -0.00000441, 0.00000486, -0.00000322, -0.00000252, 0.00000371, 0.00000317, -0.00000755, 0.00000821, -0.00000025, -0.00000036, -0.00000347, 0.00000757, 0.00000387, -0.00000152, 0.00000511, },
		{ -0.00000465, -0.00000375, -0.00000118, 0.00000285, -0.00000158, 0.00000006, 0.00000069, -0.00000147, 0.00001072, -0.00000151, 0.00000099, -0.00000325, 0.00000345, -0.00000153, 0.00000386, -0.00000452, 0.00000149, -0.00000080, 0.00000773, -0.00000049, -0.00000139, 0.00000182, -0.00000910, 0.00000995, -0.00002040, 0.00000887, -0.00000282, 0.00000056, -0.00000585, 0.00000075, -0.00000381, -0.00000162, -0.00000233, 0.00000088, -0.00000056, -0.00000709, -0.00000737, 0.00000134, -0.00000210, -0.00000027, -0.00000562, -0.00001643, 0.00000013, -0.00000134, -0.00000905, 0.00000231, -0.00001836, 0.00001112, -0.00002776, -0.00000541, -0.00008492, -0.00000660, 0.00010630, 0.00003810, 0.00003535, 0.00004429, 0.00002738, 0.00010613, 0.00000519, -0.00009167, -0.00001142, -0.00002336, -0.00000791, -0.00001284, -0.00000035, -0.00000696, -0.00000478, -0.00000715, -0.00000412, -0.00001024, -0.00001867, -0.00004406, 0.00000055, 0.00004571, 0.00002541, 0.00001395, 0.00001760, 0.00001156, 0.00001950, 0.00001943, 0.00005355, 0.00000308, -0.00003888, -0.00001174, -0.00000211, -0.00001840, 0.00000138, -0.00000476, 0.00000259, -0.00001500, -0.00000041, -0.00000446, -0.00000520, -0.00000204, 0.00000317, -0.00000821, -0.00000404, -0.00000160, 0.00000012, 0.00000135, 0.00000064, -0.00000543, -0.00000027, -0.00000471, -0.00000323, 0.00000184, -0.00000005, 0.00000818, -0.00001586, -0.00000488, 0.00000429, 0.00000279, -0.00000246, -0.00000173, 0.00000416, -0.00000403, -0.00000214, 0.00000528, 0.00000020, -0.00000486, 0.00000789, -0.00000137, -0.00000076, -0.00000229, 0.00000934, 0.00000104, -0.00000021, 0.00000427, },
		{ -0.00000506, -0.00000310, -0.00000089, 0.00000297, -0.00000273, 0.00000185, -0.00000141, 0.00000196, 0.00000937, -0.00000289, 0.00000140, -0.00000314, 0.00000376, -0.00000137, 0.00000375, -0.00000518, 0.00000209, -0.00000120, 0.00000989, -0.00000499, 0.00000346, -0.00000400, -0.00000295, 0.00000387, -0.00001823, 0.00001175, -0.00000533, 0.00000123, -0.00000643, 0.00000109, -0.00000427, -0.00000071, -0.00000263, 0.00000128, -0.00000069, -0.00000889, -0.00000594, 0.00000194, -0.00000229, -0.00000070, -0.00000725, -0.00001557, 0.00000240, -0.00000449, -0.00000557, -0.00000203, -0.00001451, 0.00000766, -0.00002610, -0.00001337, -0.00009151, 0.00003764, 0.00009174, 0.00002950, 0.00003723, 0.00003881, 0.00003140, 0.00010575, -0.00001686, -0.00008079, -0.00001073, -0.00002161, -0.00000857, -0.00001040, -0.00000106, -0.00000666, -0.00000584, -0.00000552, -0.00000561, -0.00001079, -0.00002189, -0.00004546, 0.00000853, 0.00004936, 0.00002141, 0.00001700, 0.00001551, 0.00001555, 0.00001703, 0.00002899, 0.00005280, -0.00001234, -0.00004151, -0.00000251, -0.00000772, -0.00001460, 0.00000038, -0.00000366, 0.00000137, -0.00001582, 0.00000151, -0.00000631, -0.00000438, -0.00000100, 0.00000295, -0.00001019, -0.00000278, -0.00000118, 0.00000015, 0.00000134, -0.00000010, -0.00000523, 0.00000001, -0.00000589, -0.00000148, 0.00000073, 0.00000266, 0.00000444, -0.00001739, -0.00000025, 0.00000284, 0.00000259, -0.00000371, 0.00000109, 0.00000269, -0.00000451, -0.00000130, 0.00000607, -0.00000234, -0.00000167, 0.00000634, -0.00000161, -0.00000166, 0.00000013, 0.00000946, -0.00000107, 0.00000112, 0.00000318, },
		{ -0.00000516, -0.00000256, -0.00000054, 0.00000266, -0.00000320, 0.00000308, -0.00000315, 0.00000588, 0.00000636, -0.00000285, 0.00000113, -0.00000243, 0.00000352, -0.00000070, 0.00000285, -0.00000490, 0.00000155, 0.00000025, 0.00000951, -0.00000712, 0.00000647, -0.00000857, 0.00000299, -0.00000388, -0.00001265, 0.00001167, -0.00000650, 0.00000122, -0.00000645, 0.00000091, -0.00000406, -0.00000037, -0.00000256, 0.00000150, -0.00000087, -0.00001107, -0.00000326, 0.00000132, -0.00000168, -0.00000225, -0.00000768, -0.00001515, 0.00000472, -0.00000795, -0.00000136, -0.00000776, -0.00000853, 0.00000128, -0.00002079, -0.00002966, -0.00008121, 0.00007233, 0.00007073, 0.00002735, 0.00003681, 0.00003423, 0.00003683, 0.00010167, -0.00003591, -0.00006903, -0.00001145, -0.00001923, -0.00000961, -0.00000749, -0.00000255, -0.00000597, -0.00000699, -0.00000360, -0.00000795, -0.00001075, -0.00002581, -0.00004612, 0.00001808, 0.00005072, 0.00001859, 0.00001923, 0.00001451, 0.00001862, 0.00001569, 0.00003998, 0.00004696, -0.00002850, -0.00003838, 0.00000420, -0.00001366, -0.00000941, -0.00000231, -0.00000095, -0.00000171, -0.00001478, 0.00000211, -0.00000752, -0.00000360, 0.00000035, 0.00000179, -0.00001165, -0.00000119, -0.00000104, 0.00000012, 0.00000136, -0.00000101, -0.00000458, -0.00000042, -0.00000629, -0.00000027, 0.00000019, 0.00000500, -0.00000065, -0.00001628, 0.00000295, 0.00000132, 0.00000231, -0.00000421, 0.00000358, 0.00000073, -0.00000459, -0.00000008, 0.00000586, -0.00000383, 0.00000110, 0.00000437, -0.00000148, -0.00000249, 0.00000320, 0.00000807, -0.00000216, 0.00000215, 0.00000198, },
		{ -0.00000497, -0.00000220, -0.00000010, 0.00000195, -0.00000287, 0.00000335, -0.00000378, 0.00000927, 0.00000251, -0.00000167, 0.00000025, -0.00000120, 0.00000285, 0.00000027, 0.00000150, -0.00000411, 0.00000046, 0.00000294, 0.00000682, -0.00000651, 0.00000675, -0.00001041, 0.00000660, -0.00001088, -0.00000543, 0.00000931, -0.00000648, 0.00000069, -0.00000618, 0.00000063, -0.00000365, -0.00000018, -0.00000255, 0.00000201, -0.00000185, -0.00001258, -0.00000035, 0.00000025, -0.00000105, -0.00000404, -0.00000759, -0.00001463, 0.00000637, -0.00001064, 0.00000211, -0.00001309, -0.00000253, -0.00000541, -0.00001548, -0.00004965, -0.00005522, 0.00009182, 0.00005024, 0.00002841, 0.00003512, 0.00003027, 0.00004358, 0.00009359, -0.00005088, -0.00005782, -0.00001246, -0.00001709, -0.00001021, -0.00000492, -0.00000418, -0.00000535, -0.00000773, -0.00000200, -0.00001067, -0.00001027, -0.00003056, -0.00004548, 0.00002863, 0.00004984, 0.00001720, 0.00002043, 0.00001491, 0.00002016, 0.00001675, 0.00005082, 0.00003547, -0.00004299, -0.00003007, 0.00000663, -0.00001774, -0.00000473, -0.00000518, 0.00000218, -0.00000574, -0.00001239, 0.00000150, -0.00000805, -0.00000281, 0.00000173, -0.00000016, -0.00001244, 0.00000064, -0.00000136, 0.00000031, 0.00000113, -0.00000171, -0.00000390, -0.00000121, -0.00000604, 0.00000033, 0.00000050, 0.00000616, -0.00000600, -0.00001293, 0.00000425, 0.00000037, 0.00000166, -0.00000375, 0.00000531, -0.00000138, -0.00000425, 0.00000128, 0.00000477, -0.00000404, 0.00000283, 0.00000268, -0.00000154, -0.00000261, 0.00000611, 0.00000571, -0.00000228, 0.00000273, 0.00000081, },
		{ -0.00000455, -0.00000203, 0.00000038, 0.00000100, -0.00000187, 0.00000259, -0.00000288, 0.00001122, -0.00000118, 0.00000006, -0.00000089, 0.00000025, 0.00000197, 0.00000124, 0.00000010, -0.00000336, -0.00000036, 0.00000576, 0.00000288, -0.00000385, 0.00000455, -0.00000922, 0.00000675, -0.00001518, 0.00000149, 0.00000577, -0.00000567, -0.00000012, -0.00000582, 0.00000049, -0.00000337, 0.00000016, -0.00000283, 0.00000307, -0.00000407, -0.00001258, 0.00000183, -0.00000049, -0.00000115, -0.00000517, -0.00000780, -0.00001342, 0.00000676, -0.00001173, 0.00000366, -0.00001647, 0.00000155, -0.00000994, -0.00001426, -0.00006660, -0.00001928, 0.00009534, 0.00003476, 0.00002988, 0.00003317, 0.00002677, 0.00005124, 0.00008180, -0.00006138, -0.00004800, -0.00001310, -0.00001566, -0.00000986, -0.00000328, -0.00000544, -0.00000509, -0.00000774, -0.00000122, -0.00001325, -0.00000962, -0.00003619, -0.00004295, 0.00003936, 0.00004717, 0.00001707, 0.00002081, 0.00001656, 0.00002017, 0.00002103, 0.00005960, 0.00001845, -0.00005297, -0.00001885, 0.00000479, -0.00001889, -0.00000181, -0.00000702, 0.00000459, -0.00000971, -0.00000936, 0.00000003, -0.00000794, -0.00000196, 0.00000288, -0.00000271, -0.00001237, 0.00000244, -0.00000211, 0.00000083, 0.00000048, -0.00000195, -0.00000354, -0.00000194, -0.00000547, 0.00000055, 0.00000163, 0.00000558, -0.00001039, -0.00000833, 0.00000374, 0.00000025, 0.00000052, -0.00000224, 0.00000592, -0.00000326, -0.00000352, 0.00000248, 0.00000321, -0.00000317, 0.00000336, 0.00000165, -0.00000209, -0.00000157, 0.00000804, 0.00000310, -0.00000172, 0.00000285, -0.00000022, },
		{ -0.00000400, -0.00000203, 0.00000082, 0.00000007, -0.00000055, 0.00000114, -0.00000050, 0.00001120, -0.00000387, 0.00000165, -0.00000186, 0.00000155, 0.00000116, 0.00000200, -0.00000107, -0.00000302, -0.00000025, 0.00000757, -0.00000089, -0.00000050, 0.00000111, -0.00000596, 0.00000363, -0.00001590, 0.00000665, 0.00000211, -0.00000453, -0.00000106, -0.00000542, 0.00000050, -0.00000323, 0.00000063, -0.00000331, 0.00000444, -0.00000738, -0.00001083, 0.00000281, -0.00000056, -0.00000231, -0.00000516, -0.00000876, -0.00001128, 0.00000576, -0.00001107, 0.00000298, -0.00001735, 0.00000297, -0.00001125, -0.00001957, -0.00007421, 0.00001851, 0.00008618, 0.00002577, 0.00003023, 0.00003147, 0.00002394, 0.00005897, 0.00006721, -0.00006770, -0.00003974, -0.00001327, -0.00001494, -0.00000853, -0.00000276, -0.00000612, -0.00000525, -0.00000700, -0.00000154, -0.00001527, -0.00000921, -0.00004252, -0.00003811, 0.00004944, 0.00004332, 0.00001785, 0.00002079, 0.00001893, 0.00001914, 0.00002890, 0.00006408, -0.00000278, -0.00005623, -0.00000785, 0.00000024, -0.00001743, -0.00000089, -0.00000739, 0.00000565, -0.00001287, -0.00000635, -0.00000194, -0.00000723, -0.00000112, 0.00000365, -0.00000568, -0.00001129, 0.00000380, -0.00000298, 0.00000151, -0.00000046, -0.00000174, -0.00000365, -0.00000234, -0.00000494, 0.00000074, 0.00000322, 0.00000314, -0.00001279, -0.00000378, 0.00000214, 0.00000080, -0.00000091, 0.00000014, 0.00000530, -0.00000459, -0.00000251, 0.00000324, 0.00000169, -0.00000176, 0.00000295, 0.00000123, -0.00000306, 0.00000067, 0.00000849, 0.00000092, -0.00000089, 0.00000264, -0.00000105, },
		{ -0.00000342, -0.00000213, 0.00000115, -0.00000061, 0.00000064, -0.00000041, 0.00000282, 0.00000926, -0.00000507, 0.00000262, -0.00000228, 0.00000238, 0.00000064, 0.00000241, -0.00000193, -0.00000311, 0.00000100, 0.00000765, -0.00000331, 0.00000213, -0.00000203, -0.00000230, -0.00000142, -0.00001330, 0.00000938, -0.00000093, -0.00000341, -0.00000204, -0.00000487, 0.00000046, -0.00000300, 0.00000094, -0.00000357, 0.00000553, -0.00001109, -0.00000774, 0.00000267, -0.00000013, -0.00000424, -0.00000410, -0.00001044, -0.00000840, 0.00000379, -0.00000921, 0.00000063, -0.00001629, 0.00000239, -0.00001034, -0.00003104, -0.00006887, 0.00005043, 0.00006987, 0.00002234, 0.00002921, 0.00003001, 0.00002220, 0.00006573, 0.00005105, -0.00007058, -0.00003277, -0.00001327, -0.00001450, -0.00000664, -0.00000317, -0.00000627, -0.00000568, -0.00000574, -0.00000293, -0.00001643, -0.00000945, -0.00004916, -0.00003083, 0.00005819, 0.00003888, 0.00001920, 0.00002081, 0.00002140, 0.00001796, 0.00004009, 0.00006195, -0.00002539, -0.00005221, 0.00000020, -0.00000486, -0.00001457, -0.00000135, -0.00000657, 0.00000539, -0.00001496, -0.00000378, -0.00000407, -0.00000594, -0.00000043, 0.00000403, -0.00000889, -0.00000907, 0.00000431, -0.00000354, 0.00000199, -0.00000138, -0.00000134, -0.00000409, -0.00000233, -0.00000464, 0.00000127, 0.00000464, -0.00000077, -0.00001265, -0.00000051, 0.00000044, 0.00000146, -0.00000216, 0.00000287, 0.00000365, -0.00000521, -0.00000137, 0.00000345, 0.00000064, -0.00000041, 0.00000215, 0.00000105, -0.00000399, 0.00000370, 0.00000739, -0.00000035, -0.00000018, 0.00000227, -0.00000167, },
		{ -0.00000293, -0.00000226, 0.00000135, -0.00000087, 0.00000129, -0.00000137, 0.00000620, 0.00000596, -0.00000481, 0.00000280, -0.00000203, 0.00000257, 0.00000056, 0.00000244, -0.00000259, -0.00000333, 0.00000305, 0.00000599, -0.00000385, 0.00000320, -0.00000363, 0.00000004, -0.00000654, -0.00000857, 0.00000978, -0.00000304, -0.00000250, -0.00000302, -0.00000408, 0.00000014, -0.00000241, 0.00000075, -0.00000316, 0.00000563, -0.00001426, -0.00000412, 0.00000192, 0.00000028, -0.00000630, -0.00000256, -0.00001246, -0.00000526, 0.00000152, -0.00000704, -0.00000233, -0.00001451, 0.00000139, -0.00000973, -0.00004534, -0.00005093, 0.00007133, 0.00005201, 0.00002233, 0.00002731, 0.00002851, 0.00002195, 0.00007059, 0.00003454, -0.00007086, -0.00002672, -0.00001348, -0.00001380, -0.00000475, -0.00000411, -0.00000612, -0.00000608, -0.00000436, -0.00000516, -0.00001666, -0.00001070, -0.00005560, -0.00002123, 0.00006517, 0.00003429, 0.00002088, 0.00002116, 0.00002342, 0.00001778, 0.00005351, 0.00005137, -0.00004543, -0.00004246, 0.00000406, -0.00000880, -0.00001160, -0.00000231, -0.00000517, 0.00000418, -0.00001601, -0.00000185, -0.00000606, -0.00000422, 0.00000000, 0.00000394, -0.00001203, -0.00000586, 0.00000373, -0.00000346, 0.00000193, -0.00000191, -0.00000113, -0.00000456, -0.00000205, -0.00000461, 0.00000240, 0.00000517, -0.00000522, -0.00001017, 0.00000076, -0.00000044, 0.00000157, -0.00000261, 0.00000525, 0.00000142, -0.00000516, -0.00000024, 0.00000313, 0.00000030, 0.00000037, 0.00000150, 0.00000065, -0.00000428, 0.00000676, 0.00000515, -0.00000063, 0.00000014, 0.00000190, -0.00000208, },
		{ -0.00000260, -0.00000236, 0.00000144, -0.00000072, 0.00000126, -0.00000125, 0.00000874, 0.00000217, -0.00000350, 0.00000237, -0.00000124, 0.00000215, 0.00000092, 0.00000214, -0.00000322, -0.00000323, 0.00000518, 0.00000322, -0.00000278, 0.00000265, -0.00000334, 0.00000002, -0.00001001, -0.00000325, 0.00000851, -0.00000421, -0.00000191, -0.00000391, -0.00000307, -0.00000050, -0.00000135, -0.00000008, -0.00000188, 0.00000430, -0.00001607, -0.00000078, 0.00000111, 0.00000015, -0.00000778, -0.00000118, -0.00001429, -0.00000231, -0.00000052, -0.00000527, -0.00000500, -0.00001303, 0.00000139, -0.00001219, -0.00005753, -0.00002425, 0.00007964, 0.00003671, 0.00002358, 0.00002520, 0.00002678, 0.00002333, 0.00007300, 0.00001859, -0.00006919, -0.00002137, -0.00001414, -0.00001245, -0.00000339, -0.00000513, -0.00000590, -0.00000621, -0.00000328, -0.00000781, -0.00001611, -0.00001315, -0.00006132, -0.00000960, 0.00007011, 0.00002991, 0.00002280, 0.00002200, 0.00002455, 0.00001999, 0.00006692, 0.00003191, -0.00005884, -0.00003023, 0.00000428, -0.00001103, -0.00000916, -0.00000323, -0.00000368, 0.00000241, -0.00001619, -0.00000073, -0.00000757, -0.00000236, 0.00000026, 0.00000325, -0.00001465, -0.00000208, 0.00000216, -0.00000264, 0.00000121, -0.00000185, -0.00000136, -0.00000477, -0.00000177, -0.00000470, 0.00000418, 0.00000415, -0.00000898, -0.00000626, 0.00000009, -0.00000010, 0.00000074, -0.00000179, 0.00000660, -0.00000078, -0.00000464, 0.00000076, 0.00000243, 0.00000064, 0.00000037, 0.00000135, -0.00000029, -0.00000341, 0.00000901, 0.00000246, -0.00000014, 0.00000005, 0.00000164, -0.00000231, },
		},
		{ { 0.00000268, 0.00000227, -0.00000342, -0.00000069, -0.00000109, 0.00000006, 0.00000026, -0.00000037, -0.00000002, -0.00000267, 0.00000421, -0.00000192, -0.00000164, -0.00000047, 0.00000334, 0.00000019, -0.00000140, -0.00000215, 0.00000023, -0.00000016, 0.00000127, 0.00000277, -0.00000581, 0.00000228, -0.00000623, 0.00000118, 0.00000101, 0.00000126, -0.00000056, -0.00000107, -0.00000321, 0.00000001, -0.00000157, -0.00000289, 0.00001107, -0.00000192, -0.00000148, -0.00000096, -0.00000209, -0.00000262, 0.00001094, -0.00000570, 0.00000180, -0.00000166, -0.00000064, 0.00000127, -0.00002021, 0.00000252, -0.00000321, -0.00000824, -0.00000267, -0.00000600, -0.00000658, -0.00000956, -0.00002372, -0.00001232, -0.00006514, -0.00000387, 0.00008161, 0.00001641, 0.00001206, 0.00000535, -0.00003970, 0.00003509, 0.00013328, 0.00003509, -0.00003970, 0.00000535, 0.00001206, 0.00001641, 0.00008161, -0.00000387, -0.00006514, -0.00001232, -0.00002372, -0.00000956, -0.00000658, -0.00000600, -0.00000267, -0.00000824, -0.00000321, 0.00000252, -0.00002021, 0.00000127, -0.00000064, -0.00000166, 0.00000180, -0.00000570, 0.00001094, -0.00000262, -0.00000209, -0.00000096, -0.00000148, -0.00000192, 0.00001107, -0.00000289, -0.00000157, 0.00000001, -0.00000321, -0.00000107, -0.00000056, 0.00000126, 0.00000101, 0.00000118, -0.00000623, 0.00000228, -0.00000581, 0.00000277, 0.00000127, -0.00000016, 0.00000023, -0.00000215, -0.00000140, 0.00000019, 0.00000334, -0.00000047, -0.00000164, -0.00000192, 0.00000421, -0.00000267, -0.00000002, -0.00000037, 0.00000026, 0.00000006, -0.00000109, -0.00000069, -0.00000342, 0.00000227, },
		{ 0.00000273, 0.00000171, -0.00000390, -0.00000010, -0.00000139, 0.00000049, -0.00000022, 0.00000011, -0.00000058, -0.00000212, 0.00000418, -0.00000231, -0.00000156, -0.00000014, 0.00000344, -0.00000047, -0.00000082, -0.00000271, 0.00000068, -0.00000051, 0.00000218, 0.00000157, -0.00000521, 0.00000192, -0.00000619, 0.00000178, 0.00000043, 0.00000184, -0.00000111, -0.00000058, -0.00000412, 0.00000086, -0.00000238, -0.00000137, 0.00001036, -0.00000230, -0.00000145, -0.00000045, -0.00000302, -0.00000138, 0.00001108, -0.00000747, 0.00000294, -0.00000283, 0.00000157, -0.00000138, -0.00001937, 0.00000332, -0.00000340, -0.00000875, -0.00000258, -0.00000587, -0.00000619, -0.00001081, -0.00002216, -0.00001663, -0.00006290, 0.00000867, 0.00007500, 0.00001377, 0.00001075, 0.00000515, -0.00004389, 0.00003290, 0.00013343, 0.00003713, -0.00003545, 0.00000569, 0.00001321, 0.00002006, 0.00008708, -0.00001801, -0.00006483, -0.00000922, -0.00002483, -0.00000844, -0.00000694, -0.00000623, -0.00000262, -0.00000777, -0.00000293, 0.00000131, -0.00002040, 0.00000356, -0.00000295, -0.00000018, 0.00000032, -0.00000340, 0.00001012, -0.00000340, -0.00000137, -0.00000138, -0.00000160, -0.00000129, 0.00001140, -0.00000420, -0.00000082, -0.00000082, -0.00000220, -0.00000173, 0.00000015, 0.00000053, 0.00000178, 0.00000032, -0.00000592, 0.00000228, -0.00000608, 0.00000378, 0.00000038, 0.00000020, -0.00000021, -0.00000160, -0.00000205, 0.00000103, 0.00000299, -0.00000065, -0.00000179, -0.00000141, 0.00000402, -0.00000302, 0.00000043, -0.00000080, 0.00000072, -0.00000040, -0.00000076, -0.00000133, -0.00000273, 0.00000262, },
		{ 0.00000287, 0.00000095, -0.00000414, 0.00000039, -0.00000164, 0.00000085, -0.00000066, 0.00000060, -0.00000119, -0.00000140, 0.00000393, -0.00000259, -0.00000154, 0.00000035, 0.00000329, -0.00000090, -0.00000038, -0.00000319, 0.00000108, -0.00000077, 0.00000301, 0.00000029, -0.00000439, 0.00000127, -0.00000583, 0.00000210, 0.00000004, 0.00000224, -0.00000150, -0.00000031, -0.00000484, 0.00000163, -0.00000314, 0.00000027, 0.00000934, -0.00000243, -0.00000151, 0.00000013, -0.00000409, 0.00000025, 0.00001051, -0.00000860, 0.00000362, -0.00000349, 0.00000341, -0.00000414, -0.00001807, 0.00000382, -0.00000359, -0.00000921, -0.00000250, -0.00000569, -0.00000596, -0.00001192, -0.00002054, -0.00002171, -0.00005827, 0.00001924, 0.00006776, 0.00001197, 0.00000931, 0.00000505, -0.00004800, 0.00003054, 0.00013388, 0.00003906, -0.00003117, 0.00000621, 0.00001418, 0.00002493, 0.00009082, -0.00003316, -0.00006202, -0.00000760, -0.00002521, -0.00000769, -0.00000710, -0.00000668, -0.00000232, -0.00000746, -0.00000250, -0.00000036, -0.00001982, 0.00000522, -0.00000507, 0.00000138, -0.00000130, -0.00000072, 0.00000867, -0.00000370, -0.00000088, -0.00000167, -0.00000182, -0.00000042, 0.00001132, -0.00000523, -0.00000021, -0.00000154, -0.00000119, -0.00000249, 0.00000096, -0.00000031, 0.00000268, -0.00000077, -0.00000526, 0.00000191, -0.00000597, 0.00000451, -0.00000040, 0.00000050, -0.00000056, -0.00000113, -0.00000267, 0.00000197, 0.00000245, -0.00000066, -0.00000200, -0.00000081, 0.00000363, -0.00000316, 0.00000073, -0.00000114, 0.00000112, -0.00000083, -0.00000047, -0.00000196, -0.00000189, 0.00000278, },
		{ 0.00000307, 0.00000000, -0.00000411, 0.00000073, -0.00000179, 0.00000111, -0.00000102, 0.00000103, -0.00000181, -0.00000057, 0.00000350, -0.00000275, -0.00000156, 0.00000096, 0.00000289, -0.00000105, -0.00000016, -0.00000353, 0.00000137, -0.00000088, 0.00000366, -0.00000095, -0.00000344, 0.00000042, -0.00000519, 0.00000218, -0.00000016, 0.00000248, -0.00000172, -0.00000029, -0.00000532, 0.00000226, -0.00000379, 0.00000192, 0.00000808, -0.00000235, -0.00000162, 0.00000071, -0.00000520, 0.00000217, 0.00000931, -0.00000909, 0.00000376, -0.00000352, 0.00000466, -0.00000673, -0.00001650, 0.00000409, -0.00000387, -0.00000948, -0.00000254, -0.00000533, -0.00000608, -0.00001261, -0.00001925, -0.00002706, -0.00005162, 0.00002767, 0.00006030, 0.00001084, 0.00000775, 0.00000505, -0.00005202, 0.00002800, 0.00013462, 0.00004090, -0.00002690, 0.00000693, 0.00001496, 0.00003114, 0.00009221, -0.00004859, -0.00005699, -0.00000756, -0.00002466, -0.00000749, -0.00000692, -0.00000745, -0.00000169, -0.00000739, -0.00000187, -0.00000250, -0.00001836, 0.00000604, -0.00000672, 0.00000276, -0.00000284, 0.00000215, 0.00000672, -0.00000356, -0.00000065, -0.00000185, -0.00000209, 0.00000065, 0.00001083, -0.00000591, 0.00000020, -0.00000206, -0.00000029, -0.00000326, 0.00000180, -0.00000120, 0.00000366, -0.00000203, -0.00000432, 0.00000117, -0.00000545, 0.00000489, -0.00000099, 0.00000068, -0.00000078, -0.00000082, -0.00000317, 0.00000292, 0.00000178, -0.00000056, -0.00000224, -0.00000014, 0.00000305, -0.00000307, 0.00000086, -0.00000134, 0.00000141, -0.00000120, -0.00000024, -0.00000251, -0.00000097, 0.00000281, },
		{ 0.00000328, -0.00000108, -0.00000382, 0.00000092, -0.00000183, 0.00000124, -0.00000126, 0.00000136, -0.00000236, 0.00000031, 0.00000291, -0.00000281, -0.00000158, 0.00000163, 0.00000228, -0.00000095, -0.00000018, -0.00000368, 0.00000151, -0.00000078, 0.00000406, -0.00000204, -0.00000248, -0.00000056, -0.00000434, 0.00000206, -0.00000020, 0.00000259, -0.00000180, -0.00000050, -0.00000551, 0.00000266, -0.00000423, 0.00000350, 0.00000665, -0.00000210, -0.00000177, 0.00000124, -0.00000624, 0.00000424, 0.00000758, -0.00000899, 0.00000339, -0.00000288, 0.00000517, -0.00000894, -0.00001483, 0.00000424, -0.00000428, -0.00000951, -0.00000278, -0.00000469, -0.00000666, -0.00001268, -0.00001865, -0.00003211, -0.00004344, 0.00003396, 0.00005292, 0.00001024, 0.00000608, 0.00000515, -0.00005594, 0.00002527, 0.00013564, 0.00004270, -0.00002266, 0.00000788, 0.00001560, 0.00003877, 0.00009065, -0.00006341, -0.00005023, -0.00000900, -0.00002316, -0.00000791, -0.00000635, -0.00000857, -0.00000070, -0.00000757, -0.00000108, -0.00000507, -0.00001603, 0.00000588, -0.00000766, 0.00000371, -0.00000406, 0.00000498, 0.00000442, -0.00000304, -0.00000063, -0.00000192, -0.00000240, 0.00000188, 0.00000995, -0.00000624, 0.00000036, -0.00000231, 0.00000041, -0.00000394, 0.00000259, -0.00000207, 0.00000463, -0.00000336, -0.00000315, 0.00000013, -0.00000453, 0.00000489, -0.00000134, 0.00000070, -0.00000080, -0.00000073, -0.00000345, 0.00000377, 0.00000106, -0.00000037, -0.00000250, 0.00000055, 0.00000233, -0.00000278, 0.00000080, -0.00000138, 0.00000158, -0.00000148, -0.00000012, -0.00000292, -0.00000004, 0.00000275, },
		{ 0.00000343, -0.00000224, -0.00000330, 0.00000093, -0.00000174, 0.00000123, -0.00000135, 0.00000155, -0.00000279, 0.00000118, 0.00000222, -0.00000281, -0.00000157, 0.00000231, 0.00000154, -0.00000061, -0.00000044, -0.00000362, 0.00000147, -0.00000045, 0.00000416, -0.00000289, -0.00000162, -0.00000154, -0.00000337, 0.00000178, -0.00000013, 0.00000260, -0.00000180, -0.00000090, -0.00000541, 0.00000282, -0.00000441, 0.00000490, 0.00000515, -0.00000174, -0.00000189, 0.00000164, -0.00000710, 0.00000628, 0.00000549, -0.00000840, 0.00000259, -0.00000161, 0.00000488, -0.00001062, -0.00001318, 0.00000430, -0.00000485, -0.00000926, -0.00000328, -0.00000376, -0.00000777, -0.00001199, -0.00001903, -0.00003634, -0.00003430, 0.00003825, 0.00004580, 0.00001008, 0.00000430, 0.00000536, -0.00005979, 0.00002236, 0.00013691, 0.00004448, -0.00001846, 0.00000904, 0.00001622, 0.00004776, 0.00008564, -0.00007672, -0.00004241, -0.00001164, -0.00002079, -0.00000891, -0.00000542, -0.00001000, 0.00000061, -0.00000797, -0.00000019, -0.00000794, -0.00001292, 0.00000470, -0.00000775, 0.00000405, -0.00000476, 0.00000753, 0.00000198, -0.00000225, -0.00000080, -0.00000193, -0.00000267, 0.00000319, 0.00000873, -0.00000620, 0.00000025, -0.00000225, 0.00000082, -0.00000442, 0.00000324, -0.00000282, 0.00000550, -0.00000466, -0.00000188, -0.00000114, -0.00000326, 0.00000452, -0.00000143, 0.00000052, -0.00000064, -0.00000089, -0.00000346, 0.00000443, 0.00000036, -0.00000016, -0.00000273, 0.00000120, 0.00000152, -0.00000232, 0.00000058, -0.00000127, 0.00000159, -0.00000165, -0.00000013, -0.00000315, 0.00000082, 0.00000267, },
		{ 0.00000346, -0.00000339, -0.00000260, 0.00000079, -0.00000154, 0.00000108, -0.00000130, 0.00000157, -0.00000306, 0.00000197, 0.00000146, -0.00000276, -0.00000148, 0.00000293, 0.00000073, -0.00000011, -0.00000091, -0.00000336, 0.00000127, 0.00000010, 0.00000392, -0.00000345, -0.00000094, -0.00000245, -0.00000235, 0.00000140, 0.00000003, 0.00000256, -0.00000178, -0.00000144, -0.00000505, 0.00000271, -0.00000428, 0.00000605, 0.00000367, -0.00000132, -0.00000197, 0.00000188, -0.00000767, 0.00000812, 0.00000322, -0.00000749, 0.00000151, 0.00000015, 0.00000381, -0.00001169, -0.00001164, 0.00000429, -0.00000554, -0.00000875, -0.00000399, -0.00000257, -0.00000935, -0.00001052, -0.00002055, -0.00003930, -0.00002481, 0.00004082, 0.00003907, 0.00001027, 0.00000239, 0.00000569, -0.00006356, 0.00001926, 0.00013845, 0.00004623, -0.00001428, 0.00001033, 0.00001700, 0.00005788, 0.00007684, -0.00008762, -0.00003430, -0.00001505, -0.00001779, -0.00001036, -0.00000423, -0.00001162, 0.00000214, -0.00000850, 0.00000068, -0.00001093, -0.00000920, 0.00000257, -0.00000692, 0.00000363, -0.00000476, 0.00000960, -0.00000041, -0.00000131, -0.00000108, -0.00000191, -0.00000284, 0.00000448, 0.00000726, -0.00000584, -0.00000012, -0.00000187, 0.00000088, -0.00000462, 0.00000366, -0.00000336, 0.00000616, -0.00000581, -0.00000063, -0.00000252, -0.00000173, 0.00000382, -0.00000128, 0.00000018, -0.00000029, -0.00000129, -0.00000314, 0.00000483, -0.00000023, 0.00000003, -0.00000288, 0.00000177, 0.00000070, -0.00000174, 0.00000022, -0.00000101, 0.00000145, -0.00000169, -0.00000028, -0.00000316, 0.00000155, 0.00000263, },
		{ 0.00000331, -0.00000446, -0.00000178, 0.00000052, -0.00000125, 0.00000082, -0.00000110, 0.00000142, -0.00000313, 0.00000264, 0.00000068, -0.00000269, -0.00000129, 0.00000341, -0.00000005, 0.00000048, -0.00000154, -0.00000293, 0.00000093, 0.00000083, 0.00000337, -0.00000367, -0.00000049, -0.00000321, -0.00000135, 0.00000098, 0.00000022, 0.00000251, -0.00000178, -0.00000204, -0.00000447, 0.00000234, -0.00000383, 0.00000688, 0.00000229, -0.00000090, -0.00000197, 0.00000190, -0.00000787, 0.00000961, 0.00000098, -0.00000642, 0.00000034, 0.00000222, 0.00000208, -0.00001215, -0.00001025, 0.00000420, -0.00000629, -0.00000807, -0.00000483, -0.00000123, -0.00001128, -0.00000839, -0.00002320, -0.00004069, -0.00001550, 0.00004196, 0.00003276, 0.00001075, 0.00000034, 0.00000614, -0.00006725, 0.00001594, 0.00014029, 0.00004793, -0.00001004, 0.00001166, 0.00001825, 0.00006872, 0.00006414, -0.00009533, -0.00002670, -0.00001871, -0.00001449, -0.00001204, -0.00000297, -0.00001327, 0.00000373, -0.00000902, 0.00000135, -0.00001378, -0.00000514, -0.00000032, -0.00000522, 0.00000242, -0.00000396, 0.00001102, -0.00000258, -0.00000034, -0.00000140, -0.00000192, -0.00000287, 0.00000566, 0.00000561, -0.00000521, -0.00000070, -0.00000119, 0.00000056, -0.00000450, 0.00000379, -0.00000363, 0.00000651, -0.00000669, 0.00000047, -0.00000386, -0.00000007, 0.00000287, -0.00000092, -0.00000031, 0.00000019, -0.00000191, -0.00000251, 0.00000493, -0.00000066, 0.00000013, -0.00000292, 0.00000221, -0.00000009, -0.00000110, -0.00000023, -0.00000064, 0.00000118, -0.00000162, -0.00000054, -0.00000295, 0.00000212, 0.00000267, },
		{ 0.00000293, -0.00000535, -0.00000092, 0.00000018, -0.00000091, 0.00000047, -0.00000078, 0.00000111, -0.00000298, 0.00000315, -0.00000008, -0.00000261, -0.00000098, 0.00000371, -0.00000071, 0.00000105, -0.00000225, -0.00000240, 0.00000052, 0.00000168, 0.00000255, -0.00000358, -0.00000032, -0.00000377, -0.00000042, 0.00000055, 0.00000042, 0.00000247, -0.00000187, -0.00000263, -0.00000373, 0.00000175, -0.00000306, 0.00000737, 0.00000106, -0.00000050, -0.00000187, 0.00000170, -0.00000764, 0.00001060, -0.00000106, -0.00000539, -0.00000072, 0.00000436, -0.00000014, -0.00001204, -0.00000901, 0.00000395, -0.00000698, -0.00000733, -0.00000568, 0.00000008, -0.00001333, -0.00000583, -0.00002685, -0.00004039, -0.00000681, 0.00004199, 0.00002690, 0.00001147, -0.00000182, 0.00000672, -0.00007084, 0.00001236, 0.00014248, 0.00004952, -0.00000566, 0.00001287, 0.00002035, 0.00007966, 0.00004772, -0.00009928, -0.00002036, -0.00002207, -0.00001127, -0.00001367, -0.00000185, -0.00001474, 0.00000519, -0.00000935, 0.00000162, -0.00001622, -0.00000108, -0.00000371, -0.00000280, 0.00000048, -0.00000234, 0.00001167, -0.00000438, 0.00000054, -0.00000170, -0.00000199, -0.00000269, 0.00000663, 0.00000391, -0.00000439, -0.00000143, -0.00000025, -0.00000011, -0.00000401, 0.00000357, -0.00000355, 0.00000650, -0.00000720, 0.00000128, -0.00000502, 0.00000158, 0.00000177, -0.00000043, -0.00000087, 0.00000074, -0.00000267, -0.00000159, 0.00000471, -0.00000088, 0.00000012, -0.00000282, 0.00000246, -0.00000077, -0.00000045, -0.00000073, -0.00000020, 0.00000080, -0.00000147, -0.00000091, -0.00000253, 0.00000250, 0.00000283, },
		{ 0.00000229, -0.00000600, -0.00000009, -0.00000019, -0.00000054, 0.00000007, -0.00000037, 0.00000065, -0.00000263, 0.00000346, -0.00000079, -0.00000252, -0.00000053, 0.00000378, -0.00000118, 0.00000153, -0.00000296, -0.00000182, 0.00000010, 0.00000258, 0.00000153, -0.00000320, -0.00000040, -0.00000411, 0.00000039, 0.00000015, 0.00000062, 0.00000246, -0.00000209, -0.00000313, -0.00000290, 0.00000100, -0.00000201, 0.00000748, 0.00000003, -0.00000016, -0.00000169, 0.00000128, -0.00000696, 0.00001102, -0.00000276, -0.00000455, -0.00000146, 0.00000632, -0.00000264, -0.00001147, -0.00000787, 0.00000350, -0.00000750, -0.00000667, -0.00000638, 0.00000116, -0.00001525, -0.00000314, -0.00003120, -0.00003844, 0.00000095, 0.00004117, 0.00002149, 0.00001234, -0.00000407, 0.00000737, -0.00007428, 0.00000846, 0.00014507, 0.00005092, -0.00000099, 0.00001377, 0.00002373, 0.00008990, 0.00002807, -0.00009919, -0.00001590, -0.00002462, -0.00000852, -0.00001497, -0.00000110, -0.00001582, 0.00000634, -0.00000934, 0.00000131, -0.00001797, 0.00000262, -0.00000726, 0.00000009, -0.00000204, 0.00000003, 0.00001153, -0.00000570, 0.00000123, -0.00000192, -0.00000216, -0.00000226, 0.00000730, 0.00000226, -0.00000347, -0.00000222, 0.00000083, -0.00000109, -0.00000317, 0.00000302, -0.00000313, 0.00000607, -0.00000727, 0.00000172, -0.00000587, 0.00000310, 0.00000065, 0.00000011, -0.00000143, 0.00000128, -0.00000349, -0.00000046, 0.00000420, -0.00000088, -0.00000002, -0.00000257, 0.00000251, -0.00000129, 0.00000013, -0.00000122, 0.00000027, 0.00000035, -0.00000125, -0.00000132, -0.00000194, 0.00000269, 0.00000311, },
		{ 0.00000140, -0.00000636, 0.00000065, -0.00000055, -0.00000020, -0.00000033, 0.00000008, 0.00000010, -0.00000208, 0.00000356, -0.00000143, -0.00000243, 0.00000002, 0.00000361, -0.00000142, 0.00000182, -0.00000358, -0.00000127, -0.00000027, 0.00000343, 0.00000040, -0.00000260, -0.00000073, -0.00000421, 0.00000106, -0.00000021, 0.00000082, 0.00000248, -0.00000246, -0.00000347, -0.00000207, 0.00000014, -0.00000073, 0.00000724, -0.00000078, 0.00000011, -0.00000142, 0.00000069, -0.00000588, 0.00001082, -0.00000401, -0.00000401, -0.00000175, 0.00000789, -0.00000519, -0.00001056, -0.00000678, 0.00000275, -0.00000772, -0.00000622, -0.00000677, 0.00000181, -0.00001675, -0.00000071, -0.00003583, -0.00003506, 0.00000760, 0.00003974, 0.00001657, 0.00001329, -0.00000635, 0.00000804, -0.00007752, 0.00000415, 0.00014817, 0.00005204, 0.00000410, 0.00001417, 0.00002887, 0.00009847, 0.00000598, -0.00009506, -0.00001370, -0.00002596, -0.00000655, -0.00001569, -0.00000090, -0.00001633, 0.00000699, -0.00000883, 0.00000027, -0.00001878, 0.00000562, -0.00001059, 0.00000315, -0.00000492, 0.00000300, 0.00001062, -0.00000650, 0.00000168, -0.00000202, -0.00000244, -0.00000155, 0.00000760, 0.00000076, -0.00000254, -0.00000297, 0.00000196, -0.00000228, -0.00000205, 0.00000215, -0.00000237, 0.00000524, -0.00000687, 0.00000170, -0.00000632, 0.00000435, -0.00000040, 0.00000061, -0.00000192, 0.00000171, -0.00000424, 0.00000080, 0.00000343, -0.00000067, -0.00000029, -0.00000217, 0.00000235, -0.00000162, 0.00000060, -0.00000165, 0.00000072, -0.00000012, -0.00000102, -0.00000174, -0.00000121, 0.00000272, 0.00000349, },
		{ 0.00000026, -0.00000640, 0.00000123, -0.00000084, 0.00000008, -0.00000069, 0.00000052, -0.00000051, -0.00000138, 0.00000345, -0.00000199, -0.00000232, 0.00000066, 0.00000322, -0.00000138, 0.00000187, -0.00000404, -0.00000081, -0.00000050, 0.00000416, -0.00000074, -0.00000186, -0.00000124, -0.00000411, 0.00000160, -0.00000054, 0.00000101, 0.00000251, -0.00000297, -0.00000361, -0.00000130, -0.00000073, 0.00000071, 0.00000669, -0.00000135, 0.00000031, -0.00000112, -0.00000001, -0.00000445, 0.00001003, -0.00000478, -0.00000383, -0.00000149, 0.00000887, -0.00000757, -0.00000946, -0.00000570, 0.00000167, -0.00000756, -0.00000611, -0.00000673, 0.00000186, -0.00001756, 0.00000110, -0.00004029, -0.00003058, 0.00001313, 0.00003782, 0.00001217, 0.00001419, -0.00000858, 0.00000866, -0.00008048, -0.00000064, 0.00015184, 0.00005278, 0.00000974, 0.00001392, 0.00003621, 0.00010431, -0.00001746, -0.00008725, -0.00001393, -0.00002583, -0.00000560, -0.00001568, -0.00000140, -0.00001616, 0.00000705, -0.00000775, -0.00000160, -0.00001847, 0.00000760, -0.00001333, 0.00000603, -0.00000786, 0.00000636, 0.00000905, -0.00000677, 0.00000186, -0.00000199, -0.00000282, -0.00000059, 0.00000749, -0.00000049, -0.00000170, -0.00000358, 0.00000301, -0.00000356, -0.00000073, 0.00000104, -0.00000133, 0.00000403, -0.00000601, 0.00000120, -0.00000628, 0.00000523, -0.00000126, 0.00000100, -0.00000228, 0.00000194, -0.00000483, 0.00000207, 0.00000249, -0.00000028, -0.00000069, -0.00000163, 0.00000197, -0.00000173, 0.00000092, -0.00000197, 0.00000109, -0.00000057, -0.00000081, -0.00000210, -0.00000043, 0.00000265, 0.00000395, },
		{ -0.00000106, -0.00000612, 0.00000163, -0.00000104, 0.00000026, -0.00000096, 0.00000091, -0.00000111, -0.00000056, 0.00000316, -0.00000247, -0.00000216, 0.00000133, 0.00000266, -0.00000109, 0.00000164, -0.00000429, -0.00000050, -0.00000056, 0.00000467, -0.00000179, -0.00000107, -0.00000186, -0.00000383, 0.00000202, -0.00000083, 0.00000123, 0.00000250, -0.00000361, -0.00000352, -0.00000067, -0.00000155, 0.00000223, 0.00000586, -0.00000171, 0.00000045, -0.00000083, -0.00000073, -0.00000279, 0.00000872, -0.00000508, -0.00000399, -0.00000065, 0.00000914, -0.00000962, -0.00000826, -0.00000460, 0.00000024, -0.00000697, -0.00000638, -0.00000618, 0.00000119, -0.00001750, 0.00000195, -0.00004408, -0.00002545, 0.00001762, 0.00003551, 0.00000835, 0.00001493, -0.00001065, 0.00000915, -0.00008309, -0.00000598, 0.00015615, 0.00005311, 0.00001599, 0.00001296, 0.00004608, 0.00010636, -0.00004097, -0.00007645, -0.00001647, -0.00002418, -0.00000574, -0.00001486, -0.00000262, -0.00001525, 0.00000646, -0.00000608, -0.00000429, -0.00001696, 0.00000835, -0.00001516, 0.00000840, -0.00001058, 0.00000985, 0.00000699, -0.00000657, 0.00000177, -0.00000184, -0.00000328, 0.00000062, 0.00000695, -0.00000142, -0.00000104, -0.00000395, 0.00000383, -0.00000479, 0.00000067, -0.00000021, -0.00000011, 0.00000255, -0.00000475, 0.00000024, -0.00000574, 0.00000568, -0.00000186, 0.00000122, -0.00000245, 0.00000192, -0.00000515, 0.00000323, 0.00000147, 0.00000023, -0.00000117, -0.00000099, 0.00000141, -0.00000164, 0.00000107, -0.00000216, 0.00000134, -0.00000096, -0.00000066, -0.00000236, 0.00000034, 0.00000252, 0.00000441, },
		{ -0.00000249, -0.00000555, 0.00000182, -0.00000113, 0.00000032, -0.00000111, 0.00000119, -0.00000165, 0.00000031, 0.00000270, -0.00000288, -0.00000193, 0.00000198, 0.00000197, -0.00000057, 0.00000114, -0.00000431, -0.00000035, -0.00000040, 0.00000492, -0.00000267, -0.00000032, -0.00000254, -0.00000343, 0.00000233, -0.00000109, 0.00000149, 0.00000243, -0.00000433, -0.00000320, -0.00000023, -0.00000224, 0.00000375, 0.00000483, -0.00000188, 0.00000055, -0.00000059, -0.00000137, -0.00000103, 0.00000700, -0.00000499, -0.00000442, 0.00000071, 0.00000867, -0.00001122, -0.00000705, -0.00000350, -0.00000149, -0.00000597, -0.00000705, -0.00000515, -0.00000021, -0.00001646, 0.00000161, -0.00004678, -0.00002013, 0.00002125, 0.00003284, 0.00000517, 0.00001540, -0.00001248, 0.00000943, -0.00008530, -0.00001193, 0.00016113, 0.00005302, 0.00002286, 0.00001133, 0.00005865, 0.00010365, -0.00006314, -0.00006357, -0.00002093, -0.00002117, -0.00000691, -0.00001330, -0.00000450, -0.00001369, 0.00000530, -0.00000390, -0.00000770, -0.00001430, 0.00000774, -0.00001584, 0.00000998, -0.00001277, 0.00001318, 0.00000462, -0.00000599, 0.00000145, -0.00000158, -0.00000377, 0.00000201, 0.00000598, -0.00000200, -0.00000059, -0.00000401, 0.00000431, -0.00000582, 0.00000201, -0.00000147, 0.00000117, 0.00000091, -0.00000320, -0.00000110, -0.00000473, 0.00000568, -0.00000216, 0.00000123, -0.00000240, 0.00000161, -0.00000512, 0.00000417, 0.00000047, 0.00000079, -0.00000168, -0.00000031, 0.00000071, -0.00000134, 0.00000105, -0.00000218, 0.00000146, -0.00000127, -0.00000059, -0.00000246, 0.00000103, 0.00000240, 0.00000481, },
		{ -0.00000395, -0.00000473, 0.00000182, -0.00000109, 0.00000026, -0.00000114, 0.00000134, -0.00000207, 0.00000118, 0.00000211, -0.00000322, -0.00000161, 0.00000253, 0.00000124, 0.00000010, 0.00000040, -0.00000410, -0.00000038, -0.00000004, 0.00000488, -0.00000333, 0.00000034, -0.00000320, -0.00000293, 0.00000254, -0.00000134, 0.00000180, 0.00000224, -0.00000508, -0.00000265, 0.00000000, -0.00000275, 0.00000517, 0.00000368, -0.00000190, 0.00000065, -0.00000046, -0.00000184, 0.00000067, 0.00000504, -0.00000463, -0.00000499, 0.00000246, 0.00000750, -0.00001231, -0.00000586, -0.00000245, -0.00000345, -0.00000462, -0.00000803, -0.00000370, -0.00000226, -0.00001448, -0.00000001, -0.00004806, -0.00001506, 0.00002422, 0.00002982, 0.00000268, 0.00001548, -0.00001399, 0.00000944, -0.00008707, -0.00001848, 0.00016673, 0.00005259, 0.00003026, 0.00000923, 0.00007385, 0.00009541, -0.00008255, -0.00004975, -0.00002672, -0.00001713, -0.00000888, -0.00001116, -0.00000687, -0.00001161, 0.00000370, -0.00000140, -0.00001163, -0.00001064, 0.00000583, -0.00001525, 0.00001054, -0.00001417, 0.00001610, 0.00000217, -0.00000516, 0.00000095, -0.00000126, -0.00000424, 0.00000352, 0.00000463, -0.00000220, -0.00000039, -0.00000372, 0.00000436, -0.00000651, 0.00000314, -0.00000260, 0.00000236, -0.00000074, -0.00000150, -0.00000270, -0.00000333, 0.00000525, -0.00000216, 0.00000102, -0.00000215, 0.00000100, -0.00000469, 0.00000480, -0.00000041, 0.00000133, -0.00000217, 0.00000037, -0.00000008, -0.00000089, 0.00000086, -0.00000205, 0.00000142, -0.00000146, -0.00000063, -0.00000237, 0.00000158, 0.00000236, 0.00000507, },
		{ -0.00000534, -0.00000373, 0.00000164, -0.00000095, 0.00000008, -0.00000102, 0.00000134, -0.00000234, 0.00000199, 0.00000143, -0.00000350, -0.00000116, 0.00000292, 0.00000056, 0.00000084, -0.00000052, -0.00000368, -0.00000055, 0.00000051, 0.00000455, -0.00000372, 0.00000083, -0.00000379, -0.00000239, 0.00000267, -0.00000156, 0.00000218, 0.00000190, -0.00000580, -0.00000191, 0.00000000, -0.00000302, 0.00000641, 0.00000247, -0.00000181, 0.00000076, -0.00000048, -0.00000204, 0.00000217, 0.00000298, -0.00000415, -0.00000552, 0.00000440, 0.00000574, -0.00001291, -0.00000469, -0.00000153, -0.00000548, -0.00000307, -0.00000919, -0.00000202, -0.00000476, -0.00001172, -0.00000287, -0.00004775, -0.00001060, 0.00002675, 0.00002645, 0.00000090, 0.00001512, -0.00001513, 0.00000917, -0.00008843, -0.00002559, 0.00017287, 0.00005197, 0.00003799, 0.00000705, 0.00009129, 0.00008117, -0.00009789, -0.00003619, -0.00003307, -0.00001254, -0.00001132, -0.00000875, -0.00000946, -0.00000928, 0.00000190, 0.00000117, -0.00001576, -0.00000627, 0.00000277, -0.00001340, 0.00000999, -0.00001461, 0.00001837, -0.00000015, -0.00000420, 0.00000035, -0.00000093, -0.00000463, 0.00000504, 0.00000297, -0.00000205, -0.00000043, -0.00000307, 0.00000391, -0.00000676, 0.00000394, -0.00000346, 0.00000331, -0.00000224, 0.00000018, -0.00000442, -0.00000165, 0.00000447, -0.00000189, 0.00000064, -0.00000173, 0.00000012, -0.00000385, 0.00000506, -0.00000109, 0.00000178, -0.00000258, 0.00000099, -0.00000090, -0.00000033, 0.00000055, -0.00000179, 0.00000122, -0.00000154, -0.00000076, -0.00000209, 0.00000196, 0.00000245, 0.00000511, },
		{ -0.00000656, -0.00000264, 0.00000133, -0.00000073, -0.00000019, -0.00000078, 0.00000118, -0.00000242, 0.00000269, 0.00000069, -0.00000372, -0.00000058, 0.00000311, 0.00000002, 0.00000156, -0.00000154, -0.00000313, -0.00000081, 0.00000120, 0.00000396, -0.00000384, 0.00000114, -0.00000428, -0.00000182, 0.00000272, -0.00000174, 0.00000260, 0.00000139, -0.00000643, -0.00000103, -0.00000022, -0.00000302, 0.00000742, 0.00000130, -0.00000166, 0.00000091, -0.00000067, -0.00000190, 0.00000334, 0.00000101, -0.00000369, -0.00000585, 0.00000632, 0.00000357, -0.00001309, -0.00000348, -0.00000087, -0.00000743, -0.00000150, -0.00001034, -0.00000034, -0.00000744, -0.00000846, -0.00000675, -0.00004584, -0.00000699, 0.00002898, 0.00002277, -0.00000019, 0.00001431, -0.00001591, 0.00000863, -0.00008943, -0.00003315, 0.00017937, 0.00005141, 0.00004573, 0.00000537, 0.00011028, 0.00006087, -0.00010811, -0.00002406, -0.00003916, -0.00000800, -0.00001378, -0.00000640, -0.00001194, -0.00000699, 0.00000017, 0.00000352, -0.00001973, -0.00000160, -0.00000114, -0.00001044, 0.00000833, -0.00001401, 0.00001982, -0.00000216, -0.00000325, -0.00000027, -0.00000064, -0.00000486, 0.00000649, 0.00000108, -0.00000159, -0.00000068, -0.00000210, 0.00000295, -0.00000649, 0.00000430, -0.00000394, 0.00000391, -0.00000344, 0.00000169, -0.00000607, 0.00000016, 0.00000343, -0.00000141, 0.00000013, -0.00000122, -0.00000095, -0.00000264, 0.00000492, -0.00000152, 0.00000209, -0.00000287, 0.00000148, -0.00000167, 0.00000028, 0.00000016, -0.00000141, 0.00000089, -0.00000152, -0.00000097, -0.00000161, 0.00000213, 0.00000270, 0.00000487, },
		{ -0.00000752, -0.00000153, 0.00000093, -0.00000045, -0.00000053, -0.00000045, 0.00000088, -0.00000231, 0.00000325, -0.00000008, -0.00000389, 0.00000012, 0.00000306, -0.00000031, 0.00000215, -0.00000257, -0.00000249, -0.00000112, 0.00000196, 0.00000317, -0.00000373, 0.00000126, -0.00000464, -0.00000125, 0.00000270, -0.00000187, 0.00000305, 0.00000070, -0.00000692, -0.00000005, -0.00000064, -0.00000274, 0.00000814, 0.00000021, -0.00000148, 0.00000111, -0.00000105, -0.00000140, 0.00000406, -0.00000076, -0.00000339, -0.00000579, 0.00000797, 0.00000121, -0.00001294, -0.00000218, -0.00000058, -0.00000909, -0.00000011, -0.00001124, 0.00000107, -0.00000995, -0.00000510, -0.00001127, -0.00004250, -0.00000435, 0.00003102, 0.00001886, -0.00000067, 0.00001307, -0.00001636, 0.00000790, -0.00009018, -0.00004103, 0.00018603, 0.00005122, 0.00005308, 0.00000491, 0.00012977, 0.00003488, -0.00011251, -0.00001438, -0.00004419, -0.00000408, -0.00001583, -0.00000450, -0.00001397, -0.00000506, -0.00000117, 0.00000531, -0.00002314, 0.00000295, -0.00000549, -0.00000665, 0.00000571, -0.00001237, 0.00002035, -0.00000371, -0.00000241, -0.00000086, -0.00000044, -0.00000487, 0.00000775, -0.00000091, -0.00000090, -0.00000107, -0.00000087, 0.00000151, -0.00000568, 0.00000417, -0.00000397, 0.00000404, -0.00000420, 0.00000285, -0.00000750, 0.00000193, 0.00000227, -0.00000082, -0.00000042, -0.00000071, -0.00000211, -0.00000112, 0.00000440, -0.00000166, 0.00000221, -0.00000301, 0.00000181, -0.00000235, 0.00000089, -0.00000027, -0.00000097, 0.00000045, -0.00000142, -0.00000124, -0.00000098, 0.00000212, 0.00000313, 0.00000432, },
		{ -0.00000816, -0.00000049, 0.00000049, -0.00000016, -0.00000090, -0.00000004, 0.00000045, -0.00000201, 0.00000364, -0.00000086, -0.00000398, 0.00000092, 0.00000276, -0.00000038, 0.00000252, -0.00000352, -0.00000186, -0.00000139, 0.00000270, 0.00000226, -0.00000343, 0.00000121, -0.00000487, -0.00000069, 0.00000258, -0.00000191, 0.00000349, -0.00000017, -0.00000721, 0.00000095, -0.00000123, -0.00000219, 0.00000852, -0.00000072, -0.00000131, 0.00000136, -0.00000159, -0.00000053, 0.00000429, -0.00000220, -0.00000335, -0.00000522, 0.00000913, -0.00000113, -0.00001261, -0.00000071, -0.00000079, -0.00001030, 0.00000089, -0.00001167, 0.00000194, -0.00001195, -0.00000208, -0.00001599, -0.00003804, -0.00000262, 0.00003286, 0.00001486, -0.00000066, 0.00001151, -0.00001659, 0.00000707, -0.00009082, -0.00004903, 0.00019259, 0.00005174, 0.00005961, 0.00000652, 0.00014841, 0.00000408, -0.00011082, -0.00000791, -0.00004748, -0.00000132, -0.00001704, -0.00000339, -0.00001523, -0.00000379, -0.00000186, 0.00000624, -0.00002559, 0.00000687, -0.00000980, -0.00000240, 0.00000237, -0.00000983, 0.00001996, -0.00000471, -0.00000178, -0.00000136, -0.00000034, -0.00000463, 0.00000874, -0.00000291, -0.00000006, -0.00000152, 0.00000050, -0.00000031, -0.00000435, 0.00000353, -0.00000350, 0.00000365, -0.00000444, 0.00000355, -0.00000854, 0.00000350, 0.00000112, -0.00000022, -0.00000094, -0.00000030, -0.00000323, 0.00000059, 0.00000354, -0.00000151, 0.00000213, -0.00000296, 0.00000195, -0.00000287, 0.00000143, -0.00000069, -0.00000051, -0.00000006, -0.00000126, -0.00000150, -0.00000025, 0.00000193, 0.00000372, 0.00000345, },
		{ -0.00000843, 0.00000040, 0.00000008, 0.00000011, -0.00000125, 0.00000038, -0.00000005, -0.00000152, 0.00000383, -0.00000162, -0.00000398, 0.00000178, 0.00000223, -0.00000015, 0.00000262, -0.00000429, -0.00000130, -0.00000155, 0.00000334, 0.00000132, -0.00000300, 0.00000102, -0.00000497, -0.00000013, 0.00000235, -0.00000182, 0.00000385, -0.00000117, -0.00000729, 0.00000192, -0.00000194, -0.00000139, 0.00000858, -0.00000147, -0.00000116, 0.00000163, -0.00000226, 0.00000068, 0.00000401, -0.00000328, -0.00000361, -0.00000404, 0.00000962, -0.00000323, -0.00001221, 0.00000098, -0.00000157, -0.00001090, 0.00000135, -0.00001147, 0.00000202, -0.00001309, 0.00000017, -0.00002035, -0.00003291, -0.00000165, 0.00003444, 0.00001092, -0.00000036, 0.00000975, -0.00001671, 0.00000628, -0.00009150, -0.00005697, 0.00019878, 0.00005334, 0.00006486, 0.00001107, 0.00016463, -0.00003022, -0.00010326, -0.00000508, -0.00004855, -0.00000012, -0.00001709, -0.00000334, -0.00001546, -0.00000343, -0.00000167, 0.00000607, -0.00002676, 0.00000974, -0.00001360, 0.00000188, -0.00000136, -0.00000658, 0.00001872, -0.00000514, -0.00000140, -0.00000174, -0.00000037, -0.00000411, 0.00000937, -0.00000480, 0.00000082, -0.00000191, 0.00000188, -0.00000240, -0.00000260, 0.00000243, -0.00000255, 0.00000272, -0.00000409, 0.00000369, -0.00000907, 0.00000472, 0.00000011, 0.00000031, -0.00000133, -0.00000009, -0.00000418, 0.00000237, 0.00000242, -0.00000111, 0.00000185, -0.00000273, 0.00000189, -0.00000322, 0.00000187, -0.00000104, -0.00000008, -0.00000060, -0.00000109, -0.00000172, 0.00000053, 0.00000164, 0.00000442, 0.00000225, },
		{ -0.00000831, 0.00000110, -0.00000028, 0.00000031, -0.00000154, 0.00000078, -0.00000059, -0.00000090, 0.00000383, -0.00000235, -0.00000387, 0.00000265, 0.00000151, 0.00000036, 0.00000240, -0.00000484, -0.00000087, -0.00000156, 0.00000380, 0.00000043, -0.00000251, 0.00000075, -0.00000498, 0.00000043, 0.00000201, -0.00000157, 0.00000408, -0.00000225, -0.00000715, 0.00000280, -0.00000270, -0.00000040, 0.00000830, -0.00000200, -0.00000106, 0.00000192, -0.00000300, 0.00000214, 0.00000326, -0.00000399, -0.00000415, -0.00000226, 0.00000935, -0.00000495, -0.00001183, 0.00000289, -0.00000295, -0.00001083, 0.00000115, -0.00001050, 0.00000115, -0.00001311, 0.00000124, -0.00002385, -0.00002761, -0.00000116, 0.00003561, 0.00000727, 0.00000003, 0.00000797, -0.00001688, 0.00000567, -0.00009236, -0.00006469, 0.00020440, 0.00005631, 0.00006848, 0.00001939, 0.00017672, -0.00006628, -0.00009054, -0.00000591, -0.00004719, -0.00000071, -0.00001581, -0.00000449, -0.00001454, -0.00000410, -0.00000048, 0.00000468, -0.00002643, 0.00001119, -0.00001642, 0.00000575, -0.00000512, -0.00000291, 0.00001677, -0.00000502, -0.00000129, -0.00000199, -0.00000051, -0.00000331, 0.00000959, -0.00000647, 0.00000166, -0.00000215, 0.00000310, -0.00000458, -0.00000055, 0.00000096, -0.00000120, 0.00000130, -0.00000319, 0.00000325, -0.00000903, 0.00000550, -0.00000065, 0.00000066, -0.00000152, -0.00000017, -0.00000483, 0.00000407, 0.00000117, -0.00000052, 0.00000141, -0.00000234, 0.00000162, -0.00000337, 0.00000217, -0.00000129, 0.00000028, -0.00000112, -0.00000093, -0.00000185, 0.00000129, 0.00000129, 0.00000518, 0.00000079, },
		{ -0.00000782, 0.00000158, -0.00000054, 0.00000042, -0.00000175, 0.00000112, -0.00000111, -0.00000017, 0.00000365, -0.00000303, -0.00000363, 0.00000347, 0.00000067, 0.00000111, 0.00000185, -0.00000511, -0.00000061, -0.00000136, 0.00000404, -0.00000033, -0.00000204, 0.00000043, -0.00000491, 0.00000100, 0.00000154, -0.00000112, 0.00000411, -0.00000336, -0.00000678, 0.00000356, -0.00000345, 0.00000073, 0.00000772, -0.00000232, -0.00000099, 0.00000217, -0.00000372, 0.00000376, 0.00000213, -0.00000439, -0.00000491, 0.00000006, 0.00000828, -0.00000618, -0.00001152, 0.00000501, -0.00000489, -0.00001008, 0.00000028, -0.00000874, -0.00000073, -0.00001188, 0.00000086, -0.00002601, -0.00002265, -0.00000085, 0.00003621, 0.00000413, 0.00000030, 0.00000632, -0.00001722, 0.00000533, -0.00009351, -0.00007207, 0.00020929, 0.00006086, 0.00007026, 0.00003212, 0.00018299, -0.00010210, -0.00007380, -0.00001007, -0.00004350, -0.00000306, -0.00001319, -0.00000684, -0.00001246, -0.00000581, 0.00000169, 0.00000207, -0.00002451, 0.00001098, -0.00001789, 0.00000878, -0.00000850, 0.00000087, 0.00001432, -0.00000443, -0.00000143, -0.00000212, -0.00000075, -0.00000224, 0.00000936, -0.00000784, 0.00000238, -0.00000214, 0.00000403, -0.00000667, 0.00000163, -0.00000074, 0.00000044, -0.00000050, -0.00000180, 0.00000226, -0.00000839, 0.00000576, -0.00000109, 0.00000079, -0.00000149, -0.00000059, -0.00000507, 0.00000555, -0.00000011, 0.00000018, 0.00000085, -0.00000183, 0.00000118, -0.00000332, 0.00000230, -0.00000140, 0.00000051, -0.00000159, -0.00000082, -0.00000184, 0.00000194, 0.00000098, 0.00000590, -0.00000086, },
		{ -0.00000701, 0.00000182, -0.00000069, 0.00000043, -0.00000185, 0.00000136, -0.00000157, 0.00000062, 0.00000329, -0.00000367, -0.00000324, 0.00000416, -0.00000020, 0.00000204, 0.00000101, -0.00000511, -0.00000054, -0.00000096, 0.00000400, -0.00000091, -0.00000164, 0.00000012, -0.00000478, 0.00000155, 0.00000096, -0.00000046, 0.00000389, -0.00000442, -0.00000621, 0.00000414, -0.00000414, 0.00000192, 0.00000689, -0.00000243, -0.00000095, 0.00000236, -0.00000433, 0.00000540, 0.00000073, -0.00000458, -0.00000574, 0.00000277, 0.00000647, -0.00000688, -0.00001127, 0.00000722, -0.00000728, -0.00000874, -0.00000119, -0.00000628, -0.00000355, -0.00000939, -0.00000110, -0.00002651, -0.00001849, -0.00000042, 0.00003608, 0.00000167, 0.00000026, 0.00000495, -0.00001783, 0.00000533, -0.00009499, -0.00007909, 0.00021343, 0.00006704, 0.00007021, 0.00004962, 0.00018190, -0.00013554, -0.00005454, -0.00001681, -0.00003783, -0.00000695, -0.00000942, -0.00001020, -0.00000938, -0.00000842, 0.00000469, -0.00000160, -0.00002108, 0.00000904, -0.00001779, 0.00001064, -0.00001115, 0.00000442, 0.00001158, -0.00000348, -0.00000178, -0.00000216, -0.00000103, -0.00000096, 0.00000868, -0.00000885, 0.00000289, -0.00000180, 0.00000451, -0.00000848, 0.00000375, -0.00000249, 0.00000219, -0.00000252, -0.00000005, 0.00000082, -0.00000719, 0.00000551, -0.00000119, 0.00000068, -0.00000121, -0.00000137, -0.00000483, 0.00000669, -0.00000130, 0.00000092, 0.00000023, -0.00000124, 0.00000061, -0.00000308, 0.00000228, -0.00000136, 0.00000061, -0.00000197, -0.00000078, -0.00000165, 0.00000244, 0.00000077, 0.00000649, -0.00000261, },
		{ -0.00000593, 0.00000183, -0.00000073, 0.00000034, -0.00000183, 0.00000147, -0.00000192, 0.00000142, 0.00000279, -0.00000425, -0.00000268, 0.00000466, -0.00000102, 0.00000305, -0.00000007, -0.00000485, -0.00000064, -0.00000036, 0.00000370, -0.00000127, -0.00000135, -0.00000015, -0.00000459, 0.00000208, 0.00000028, 0.00000040, 0.00000338, -0.00000537, -0.00000547, 0.00000453, -0.00000469, 0.00000311, 0.00000587, -0.00000237, -0.00000092, 0.00000243, -0.00000476, 0.00000693, -0.00000079, -0.00000469, -0.00000647, 0.00000566, 0.00000406, -0.00000711, -0.00001103, 0.00000939, -0.00000992, -0.00000697, -0.00000309, -0.00000332, -0.00000709, -0.00000581, -0.00000460, -0.00002518, -0.00001544, 0.00000039, 0.00003509, 0.00000005, -0.00000023, 0.00000395, -0.00001875, 0.00000566, -0.00009675, -0.00008585, 0.00021692, 0.00007472, 0.00006861, 0.00007184, 0.00017225, -0.00016446, -0.00003449, -0.00002515, -0.00003083, -0.00001192, -0.00000487, -0.00001424, -0.00000560, -0.00001165, 0.00000825, -0.00000604, -0.00001640, 0.00000548, -0.00001602, 0.00001111, -0.00001280, 0.00000745, 0.00000881, -0.00000230, -0.00000228, -0.00000215, -0.00000130, 0.00000047, 0.00000757, -0.00000947, 0.00000318, -0.00000108, 0.00000444, -0.00000982, 0.00000562, -0.00000412, 0.00000389, -0.00000459, 0.00000189, -0.00000094, -0.00000555, 0.00000478, -0.00000094, 0.00000034, -0.00000073, -0.00000249, -0.00000409, 0.00000740, -0.00000228, 0.00000159, -0.00000038, -0.00000064, -0.00000006, -0.00000271, 0.00000211, -0.00000118, 0.00000055, -0.00000226, -0.00000081, -0.00000128, 0.00000273, 0.00000073, 0.00000687, -0.00000434, },
		{ -0.00000467, 0.00000165, -0.00000066, 0.00000015, -0.00000169, 0.00000144, -0.00000212, 0.00000217, 0.00000216, -0.00000476, -0.00000194, 0.00000491, -0.00000168, 0.00000404, -0.00000129, -0.00000440, -0.00000086, 0.00000039, 0.00000315, -0.00000140, -0.00000119, -0.00000038, -0.00000435, 0.00000255, -0.00000044, 0.00000143, 0.00000257, -0.00000614, -0.00000460, 0.00000471, -0.00000506, 0.00000421, 0.00000473, -0.00000216, -0.00000089, 0.00000236, -0.00000492, 0.00000821, -0.00000228, -0.00000486, -0.00000692, 0.00000847, 0.00000125, -0.00000697, -0.00001070, 0.00001133, -0.00001259, -0.00000501, -0.00000516, -0.00000013, -0.00001105, -0.00000146, -0.00000939, -0.00002207, -0.00001369, 0.00000173, 0.00003321, -0.00000068, -0.00000123, 0.00000333, -0.00001996, 0.00000623, -0.00009865, -0.00009254, 0.00022003, 0.00008357, 0.00006608, 0.00009829, 0.00015332, -0.00018698, -0.00001543, -0.00003389, -0.00002331, -0.00001737, -0.00000004, -0.00001851, -0.00000154, -0.00001512, 0.00001195, -0.00001081, -0.00001087, 0.00000057, -0.00001268, 0.00001013, -0.00001325, 0.00000971, 0.00000621, -0.00000102, -0.00000285, -0.00000213, -0.00000152, 0.00000196, 0.00000608, -0.00000968, 0.00000323, 0.00000001, 0.00000375, -0.00001057, 0.00000705, -0.00000543, 0.00000533, -0.00000650, 0.00000382, -0.00000283, -0.00000360, 0.00000368, -0.00000042, -0.00000018, -0.00000010, -0.00000388, -0.00000286, 0.00000763, -0.00000298, 0.00000213, -0.00000093, -0.00000007, -0.00000075, -0.00000223, 0.00000182, -0.00000088, 0.00000034, -0.00000244, -0.00000090, -0.00000073, 0.00000279, 0.00000090, 0.00000695, -0.00000595, },
		{ -0.00000334, 0.00000131, -0.00000052, -0.00000011, -0.00000144, 0.00000128, -0.00000216, 0.00000284, 0.00000143, -0.00000518, -0.00000105, 0.00000488, -0.00000210, 0.00000489, -0.00000254, -0.00000382, -0.00000115, 0.00000122, 0.00000242, -0.00000135, -0.00000115, -0.00000055, -0.00000404, 0.00000292, -0.00000115, 0.00000257, 0.00000146, -0.00000670, -0.00000365, 0.00000468, -0.00000520, 0.00000514, 0.00000355, -0.00000184, -0.00000084, 0.00000213, -0.00000476, 0.00000912, -0.00000360, -0.00000523, -0.00000689, 0.00001094, -0.00000169, -0.00000661, -0.00001015, 0.00001282, -0.00001500, -0.00000316, -0.00000711, 0.00000291, -0.00001499, 0.00000320, -0.00001506, -0.00001741, -0.00001323, 0.00000362, 0.00003048, -0.00000055, -0.00000271, 0.00000304, -0.00002134, 0.00000686, -0.00010044, -0.00009951, 0.00022318, 0.00009307, 0.00006353, 0.00012791, 0.00012496, -0.00020156, 0.00000092, -0.00004179, -0.00001616, -0.00002260, 0.00000448, -0.00002248, 0.00000232, -0.00001836, 0.00001534, -0.00001540, -0.00000503, -0.00000524, -0.00000804, 0.00000778, -0.00001246, 0.00001106, 0.00000397, 0.00000023, -0.00000344, -0.00000214, -0.00000161, 0.00000344, 0.00000429, -0.00000953, 0.00000305, 0.00000146, 0.00000242, -0.00001061, 0.00000791, -0.00000627, 0.00000633, -0.00000806, 0.00000552, -0.00000463, -0.00000153, 0.00000233, 0.00000029, -0.00000079, 0.00000056, -0.00000543, -0.00000121, 0.00000738, -0.00000335, 0.00000248, -0.00000136, 0.00000041, -0.00000141, -0.00000169, 0.00000145, -0.00000047, 0.00000000, -0.00000253, -0.00000104, -0.00000002, 0.00000262, 0.00000132, 0.00000668, -0.00000732, },
		{ -0.00000202, 0.00000089, -0.00000033, -0.00000041, -0.00000112, 0.00000099, -0.00000202, 0.00000339, 0.00000063, -0.00000550, -0.00000002, 0.00000455, -0.00000222, 0.00000550, -0.00000373, -0.00000320, -0.00000144, 0.00000205, 0.00000157, -0.00000115, -0.00000122, -0.00000069, -0.00000363, 0.00000315, -0.00000176, 0.00000374, 0.00000011, -0.00000701, -0.00000265, 0.00000442, -0.00000508, 0.00000583, 0.00000241, -0.00000147, -0.00000074, 0.00000172, -0.00000425, 0.00000960, -0.00000464, -0.00000589, -0.00000622, 0.00001280, -0.00000452, -0.00000620, -0.00000923, 0.00001366, -0.00001688, -0.00000171, -0.00000861, 0.00000544, -0.00001845, 0.00000763, -0.00002103, -0.00001164, -0.00001388, 0.00000598, 0.00002707, 0.00000030, -0.00000457, 0.00000293, -0.00002271, 0.00000732, -0.00010181, -0.00010716, 0.00022687, 0.00010256, 0.00006213, 0.00015915, 0.00008775, -0.00020723, 0.00001315, -0.00004769, -0.00001029, -0.00002689, 0.00000811, -0.00002562, 0.00000547, -0.00002091, 0.00001792, -0.00001925, 0.00000053, -0.00001138, -0.00000252, 0.00000429, -0.00001050, 0.00001144, 0.00000220, 0.00000136, -0.00000400, -0.00000220, -0.00000155, 0.00000482, 0.00000229, -0.00000904, 0.00000270, 0.00000317, 0.00000050, -0.00000991, 0.00000811, -0.00000651, 0.00000677, -0.00000909, 0.00000680, -0.00000615, 0.00000046, 0.00000090, 0.00000107, -0.00000140, 0.00000115, -0.00000698, 0.00000074, 0.00000669, -0.00000336, 0.00000259, -0.00000163, 0.00000076, -0.00000200, -0.00000116, 0.00000105, -0.00000001, -0.00000044, -0.00000255, -0.00000118, 0.00000079, 0.00000224, 0.00000197, 0.00000600, -0.00000835, },
		{ -0.00000080, 0.00000042, -0.00000013, -0.00000072, -0.00000076, 0.00000060, -0.00000171, 0.00000379, -0.00000023, -0.00000569, 0.00000110, 0.00000394, -0.00000199, 0.00000578, -0.00000475, -0.00000262, -0.00000164, 0.00000280, 0.00000070, -0.00000086, -0.00000134, -0.00000082, -0.00000310, 0.00000316, -0.00000219, 0.00000484, -0.00000143, -0.00000706, -0.00000164, 0.00000398, -0.00000466, 0.00000622, 0.00000138, -0.00000110, -0.00000060, 0.00000115, -0.00000338, 0.00000958, -0.00000532, -0.00000691, -0.00000482, 0.00001385, -0.00000698, -0.00000590, -0.00000783, 0.00001367, -0.00001799, -0.00000093, -0.00000936, 0.00000708, -0.00002097, 0.00001126, -0.00002667, -0.00000533, -0.00001530, 0.00000855, 0.00002323, 0.00000164, -0.00000661, 0.00000280, -0.00002383, 0.00000731, -0.00010240, -0.00011597, 0.00023171, 0.00011127, 0.00006328, 0.00018997, 0.00004296, -0.00020365, 0.00002020, -0.00005063, -0.00000647, -0.00002963, 0.00001031, -0.00002743, 0.00000746, -0.00002232, 0.00001925, -0.00002184, 0.00000517, -0.00001724, 0.00000336, 0.00000002, -0.00000754, 0.00001087, 0.00000097, 0.00000232, -0.00000450, -0.00000233, -0.00000129, 0.00000599, 0.00000018, -0.00000830, 0.00000226, 0.00000505, -0.00000193, -0.00000850, 0.00000760, -0.00000610, 0.00000654, -0.00000947, 0.00000751, -0.00000720, 0.00000219, -0.00000047, 0.00000181, -0.00000190, 0.00000152, -0.00000839, 0.00000285, 0.00000564, -0.00000305, 0.00000246, -0.00000172, 0.00000095, -0.00000247, -0.00000067, 0.00000066, 0.00000048, -0.00000096, -0.00000251, -0.00000128, 0.00000166, 0.00000171, 0.00000282, 0.00000492, -0.00000899, },
		{ 0.00000024, -0.00000003, 0.00000004, -0.00000100, -0.00000039, 0.00000014, -0.00000123, 0.00000402, -0.00000113, -0.00000572, 0.00000224, 0.00000309, -0.00000141, 0.00000567, -0.00000551, -0.00000217, -0.00000169, 0.00000337, -0.00000011, -0.00000056, -0.00000146, -0.00000096, -0.00000243, 0.00000293, -0.00000236, 0.00000576, -0.00000306, -0.00000688, -0.00000066, 0.00000335, -0.00000396, 0.00000628, 0.00000051, -0.00000074, -0.00000041, 0.00000044, -0.00000219, 0.00000908, -0.00000564, -0.00000828, -0.00000266, 0.00001394, -0.00000888, -0.00000585, -0.00000587, 0.00001275, -0.00001813, -0.00000105, -0.00000912, 0.00000753, -0.00002214, 0.00001354, -0.00003132, 0.00000089, -0.00001704, 0.00001103, 0.00001930, 0.00000319, -0.00000859, 0.00000241, -0.00002445, 0.00000652, -0.00010184, -0.00012640, 0.00023827, 0.00011848, 0.00006842, 0.00021797, -0.00000747, -0.00019122, 0.00002155, -0.00005000, -0.00000524, -0.00003034, 0.00001068, -0.00002754, 0.00000793, -0.00002227, 0.00001897, -0.00002271, 0.00000834, -0.00002218, 0.00000902, -0.00000459, -0.00000390, 0.00000948, 0.00000025, 0.00000306, -0.00000494, -0.00000253, -0.00000083, 0.00000691, -0.00000195, -0.00000739, 0.00000181, 0.00000693, -0.00000470, -0.00000648, 0.00000644, -0.00000505, 0.00000562, -0.00000914, 0.00000754, -0.00000764, 0.00000349, -0.00000161, 0.00000239, -0.00000219, 0.00000158, -0.00000948, 0.00000494, 0.00000435, -0.00000246, 0.00000210, -0.00000164, 0.00000097, -0.00000279, -0.00000025, 0.00000031, 0.00000093, -0.00000152, -0.00000245, -0.00000131, 0.00000250, 0.00000110, 0.00000381, 0.00000348, -0.00000919, },
		{ 0.00000104, -0.00000042, 0.00000016, -0.00000122, -0.00000006, -0.00000033, -0.00000062, 0.00000408, -0.00000204, -0.00000557, 0.00000333, 0.00000206, -0.00000051, 0.00000517, -0.00000595, -0.00000190, -0.00000154, 0.00000371, -0.00000079, -0.00000032, -0.00000156, -0.00000115, -0.00000161, 0.00000241, -0.00000220, 0.00000639, -0.00000469, -0.00000650, 0.00000028, 0.00000259, -0.00000298, 0.00000597, -0.00000013, -0.00000045, -0.00000018, -0.00000035, -0.00000075, 0.00000814, -0.00000561, -0.00000993, 0.00000022, 0.00001300, -0.00001009, -0.00000612, -0.00000337, 0.00001085, -0.00001722, -0.00000216, -0.00000776, 0.00000662, -0.00002167, 0.00001405, -0.00003437, 0.00000634, -0.00001857, 0.00001302, 0.00001564, 0.00000463, -0.00001023, 0.00000153, -0.00002432, 0.00000469, -0.00009983, -0.00013883, 0.00024702, 0.00012358, 0.00007893, 0.00024053, -0.00006099, -0.00017101, 0.00001724, -0.00004557, -0.00000688, -0.00002880, 0.00000903, -0.00002577, 0.00000669, -0.00002059, 0.00001689, -0.00002159, 0.00000960, -0.00002566, 0.00001390, -0.00000904, 0.00000008, 0.00000748, 0.00000000, 0.00000362, -0.00000531, -0.00000275, -0.00000016, 0.00000750, -0.00000397, -0.00000637, 0.00000144, 0.00000866, -0.00000762, -0.00000400, 0.00000473, -0.00000342, 0.00000406, -0.00000809, 0.00000685, -0.00000737, 0.00000424, -0.00000242, 0.00000271, -0.00000220, 0.00000122, -0.00001011, 0.00000684, 0.00000296, -0.00000168, 0.00000157, -0.00000141, 0.00000083, -0.00000296, 0.00000005, 0.00000005, 0.00000132, -0.00000207, -0.00000239, -0.00000122, 0.00000324, 0.00000049, 0.00000485, 0.00000174, -0.00000894, },
		{ 0.00000159, -0.00000072, 0.00000020, -0.00000136, 0.00000021, -0.00000079, 0.00000009, 0.00000395, -0.00000294, -0.00000522, 0.00000430, 0.00000095, 0.00000063, 0.00000427, -0.00000606, -0.00000183, -0.00000116, 0.00000377, -0.00000125, -0.00000018, -0.00000158, -0.00000139, -0.00000062, 0.00000159, -0.00000166, 0.00000664, -0.00000620, -0.00000594, 0.00000113, 0.00000172, -0.00000176, 0.00000530, -0.00000052, -0.00000023, 0.00000007, -0.00000116, 0.00000084, 0.00000684, -0.00000533, -0.00001175, 0.00000368, 0.00001110, -0.00001055, -0.00000672, -0.00000038, 0.00000805, -0.00001528, -0.00000428, -0.00000528, 0.00000431, -0.00001945, 0.00001251, -0.00003536, 0.00001045, -0.00001942, 0.00001416, 0.00001261, 0.00000563, -0.00001130, -0.00000006, -0.00002327, 0.00000164, -0.00009617, -0.00015352, 0.00025827, 0.00012620, 0.00009592, 0.00025504, -0.00011466, -0.00014472, 0.00000787, -0.00003752, -0.00001130, -0.00002503, 0.00000538, -0.00002214, 0.00000376, -0.00001726, 0.00001300, -0.00001836, 0.00000868, -0.00002722, 0.00001747, -0.00001286, 0.00000400, 0.00000510, 0.00000010, 0.00000402, -0.00000565, -0.00000298, 0.00000068, 0.00000773, -0.00000582, -0.00000535, 0.00000126, 0.00001007, -0.00001046, -0.00000127, 0.00000263, -0.00000135, 0.00000197, -0.00000642, 0.00000548, -0.00000640, 0.00000439, -0.00000281, 0.00000272, -0.00000191, 0.00000040, -0.00001018, 0.00000838, 0.00000161, -0.00000082, 0.00000091, -0.00000107, 0.00000056, -0.00000298, 0.00000024, -0.00000011, 0.00000161, -0.00000259, -0.00000235, -0.00000097, 0.00000382, -0.00000002, 0.00000583, -0.00000020, -0.00000827, },
		{ 0.00000186, -0.00000092, 0.00000016, -0.00000140, 0.00000039, -0.00000119, 0.00000087, 0.00000364, -0.00000381, -0.00000464, 0.00000508, -0.00000014, 0.00000194, 0.00000304, -0.00000583, -0.00000198, -0.00000055, 0.00000353, -0.00000149, -0.00000019, -0.00000150, -0.00000168, 0.00000049, 0.00000049, -0.00000073, 0.00000643, -0.00000750, -0.00000528, 0.00000190, 0.00000080, -0.00000034, 0.00000430, -0.00000065, -0.00000010, 0.00000030, -0.00000192, 0.00000247, 0.00000532, -0.00000491, -0.00001355, 0.00000748, 0.00000835, -0.00001032, -0.00000758, 0.00000294, 0.00000453, -0.00001246, -0.00000728, -0.00000181, 0.00000072, -0.00001555, 0.00000887, -0.00003401, 0.00001276, -0.00001917, 0.00001413, 0.00001051, 0.00000596, -0.00001162, -0.00000247, -0.00002121, -0.00000270, -0.00009082, -0.00017052, 0.00027209, 0.00012635, 0.00012009, 0.00025911, -0.00016530, -0.00011458, -0.00000545, -0.00002649, -0.00001811, -0.00001935, 0.00000003, -0.00001689, -0.00000065, -0.00001251, 0.00000749, -0.00001314, 0.00000553, -0.00002664, 0.00001938, -0.00001563, 0.00000750, 0.00000261, 0.00000043, 0.00000432, -0.00000599, -0.00000317, 0.00000165, 0.00000758, -0.00000741, -0.00000439, 0.00000133, 0.00001101, -0.00001301, 0.00000149, 0.00000035, 0.00000096, -0.00000047, -0.00000427, 0.00000353, -0.00000477, 0.00000392, -0.00000278, 0.00000242, -0.00000130, -0.00000088, -0.00000960, 0.00000945, 0.00000044, 0.00000004, 0.00000020, -0.00000065, 0.00000020, -0.00000286, 0.00000031, -0.00000016, 0.00000178, -0.00000305, -0.00000234, -0.00000055, 0.00000417, -0.00000036, 0.00000664, -0.00000224, -0.00000724, },
		{ 0.00000189, -0.00000101, 0.00000003, -0.00000134, 0.00000046, -0.00000150, 0.00000167, 0.00000317, -0.00000463, -0.00000384, 0.00000559, -0.00000110, 0.00000328, 0.00000157, -0.00000531, -0.00000231, 0.00000025, 0.00000302, -0.00000149, -0.00000035, -0.00000133, -0.00000200, 0.00000169, -0.00000084, 0.00000057, 0.00000573, -0.00000851, -0.00000455, 0.00000254, -0.00000010, 0.00000119, 0.00000301, -0.00000054, -0.00000005, 0.00000050, -0.00000255, 0.00000401, 0.00000370, -0.00000449, -0.00001512, 0.00001136, 0.00000499, -0.00000952, -0.00000856, 0.00000636, 0.00000056, -0.00000899, -0.00001090, 0.00000237, -0.00000383, -0.00001023, 0.00000328, -0.00003029, 0.00001302, -0.00001754, 0.00001272, 0.00000954, 0.00000545, -0.00001109, -0.00000574, -0.00001817, -0.00000823, -0.00008392, -0.00018965, 0.00028820, 0.00012448, 0.00015154, 0.00025081, -0.00020978, -0.00008310, -0.00002121, -0.00001346, -0.00002659, -0.00001233, -0.00000653, -0.00001048, -0.00000614, -0.00000674, 0.00000078, -0.00000628, 0.00000032, -0.00002387, 0.00001938, -0.00001705, 0.00001025, 0.00000026, 0.00000085, 0.00000457, -0.00000636, -0.00000328, 0.00000271, 0.00000705, -0.00000867, -0.00000355, 0.00000172, 0.00001135, -0.00001505, 0.00000402, -0.00000187, 0.00000329, -0.00000305, -0.00000183, 0.00000117, -0.00000260, 0.00000291, -0.00000234, 0.00000183, -0.00000046, -0.00000258, -0.00000838, 0.00000995, -0.00000043, 0.00000079, -0.00000049, -0.00000023, -0.00000023, -0.00000264, 0.00000028, -0.00000010, 0.00000181, -0.00000343, -0.00000237, 0.00000005, 0.00000426, -0.00000046, 0.00000718, -0.00000423, -0.00000594, },
		{ 0.00000170, -0.00000102, -0.00000015, -0.00000119, 0.00000042, -0.00000169, 0.00000246, 0.00000254, -0.00000536, -0.00000283, 0.00000579, -0.00000184, 0.00000453, -0.00000004, -0.00000458, -0.00000276, 0.00000118, 0.00000229, -0.00000128, -0.00000064, -0.00000109, -0.00000230, 0.00000290, -0.00000230, 0.00000217, 0.00000455, -0.00000917, -0.00000379, 0.00000306, -0.00000094, 0.00000277, 0.00000151, -0.00000021, -0.00000006, 0.00000063, -0.00000298, 0.00000534, 0.00000214, -0.00000422, -0.00001623, 0.00001500, 0.00000131, -0.00000833, -0.00000947, 0.00000960, -0.00000354, -0.00000523, -0.00001481, 0.00000687, -0.00000892, -0.00000397, -0.00000384, -0.00002441, 0.00001122, -0.00001443, 0.00000986, 0.00000977, 0.00000407, -0.00000975, -0.00000975, -0.00001435, -0.00001469, -0.00007585, -0.00021047, 0.00030601, 0.00012155, 0.00018965, 0.00022888, -0.00024527, -0.00005292, -0.00003763, 0.00000035, -0.00003581, -0.00000475, -0.00001362, -0.00000352, -0.00001212, -0.00000051, -0.00000658, 0.00000167, -0.00000652, -0.00001911, 0.00001748, -0.00001695, 0.00001204, -0.00000174, 0.00000124, 0.00000483, -0.00000678, -0.00000327, 0.00000380, 0.00000616, -0.00000958, -0.00000286, 0.00000243, 0.00001100, -0.00001639, 0.00000610, -0.00000381, 0.00000541, -0.00000553, 0.00000065, -0.00000136, -0.00000008, 0.00000148, -0.00000158, 0.00000103, 0.00000054, -0.00000457, -0.00000657, 0.00000985, -0.00000093, 0.00000134, -0.00000111, 0.00000016, -0.00000067, -0.00000235, 0.00000017, 0.00000005, 0.00000172, -0.00000374, -0.00000242, 0.00000080, 0.00000409, -0.00000026, 0.00000735, -0.00000606, -0.00000445, },
		{ 0.00000136, -0.00000095, -0.00000038, -0.00000097, 0.00000026, -0.00000174, 0.00000319, 0.00000176, -0.00000598, -0.00000164, 0.00000565, -0.00000226, 0.00000557, -0.00000165, -0.00000373, -0.00000325, 0.00000215, 0.00000141, -0.00000092, -0.00000102, -0.00000082, -0.00000253, 0.00000403, -0.00000379, 0.00000396, 0.00000294, -0.00000945, -0.00000304, 0.00000341, -0.00000164, 0.00000429, -0.00000013, 0.00000029, -0.00000011, 0.00000066, -0.00000316, 0.00000635, 0.00000076, -0.00000424, -0.00001668, 0.00001807, -0.00000237, -0.00000701, -0.00001008, 0.00001234, -0.00000739, -0.00000155, -0.00001857, 0.00001120, -0.00001399, 0.00000264, -0.00001190, -0.00001684, 0.00000758, -0.00000995, 0.00000567, 0.00001110, 0.00000191, -0.00000778, -0.00001426, -0.00001006, -0.00002165, -0.00006713, -0.00023229, 0.00032460, 0.00011905, 0.00023298, 0.00019294, -0.00026948, -0.00002651, -0.00005289, 0.00001356, -0.00004470, 0.00000252, -0.00002044, 0.00000326, -0.00001792, 0.00000551, -0.00001386, 0.00001003, -0.00001440, -0.00001276, 0.00001384, -0.00001532, 0.00001272, -0.00000326, 0.00000151, 0.00000511, -0.00000726, -0.00000311, 0.00000486, 0.00000496, -0.00001012, -0.00000233, 0.00000346, 0.00000992, -0.00001689, 0.00000754, -0.00000525, 0.00000708, -0.00000765, 0.00000292, -0.00000384, 0.00000258, -0.00000021, -0.00000062, 0.00000014, 0.00000155, -0.00000671, -0.00000428, 0.00000917, -0.00000102, 0.00000166, -0.00000160, 0.00000048, -0.00000108, -0.00000202, -0.00000000, 0.00000028, 0.00000151, -0.00000397, -0.00000246, 0.00000167, 0.00000366, 0.00000024, 0.00000710, -0.00000759, -0.00000290, },
		{ 0.00000091, -0.00000086, -0.00000063, -0.00000070, 0.00000002, -0.00000165, 0.00000383, 0.00000087, -0.00000645, -0.00000032, 0.00000518, -0.00000230, 0.00000628, -0.00000314, -0.00000286, -0.00000371, 0.00000307, 0.00000048, -0.00000047, -0.00000144, -0.00000057, -0.00000263, 0.00000497, -0.00000518, 0.00000582, 0.00000098, -0.00000937, -0.00000232, 0.00000359, -0.00000213, 0.00000566, -0.00000180, 0.00000090, -0.00000017, 0.00000058, -0.00000303, 0.00000697, -0.00000034, -0.00000464, -0.00001631, 0.00002030, -0.00000573, -0.00000579, -0.00001018, 0.00001429, -0.00001060, 0.00000163, -0.00002177, 0.00001488, -0.00001844, 0.00000888, -0.00002014, -0.00000822, 0.00000253, -0.00000440, 0.00000041, 0.00001332, -0.00000078, -0.00000543, -0.00001894, -0.00000574, -0.00002857, -0.00005848, -0.00025426, 0.00034280, 0.00011892, 0.00027927, 0.00014361, -0.00028086, -0.00000599, -0.00006526, 0.00002487, -0.00005220, 0.00000857, -0.00002618, 0.00000912, -0.00002284, 0.00001062, -0.00002033, 0.00001799, -0.00002257, -0.00000540, 0.00000881, -0.00001233, 0.00001231, -0.00000422, 0.00000162, 0.00000544, -0.00000779, -0.00000279, 0.00000583, 0.00000349, -0.00001029, -0.00000195, 0.00000474, 0.00000816, -0.00001650, 0.00000820, -0.00000603, 0.00000812, -0.00000920, 0.00000474, -0.00000600, 0.00000514, -0.00000196, 0.00000039, -0.00000072, 0.00000242, -0.00000881, -0.00000167, 0.00000799, -0.00000071, 0.00000170, -0.00000194, 0.00000068, -0.00000142, -0.00000169, -0.00000019, 0.00000055, 0.00000119, -0.00000413, -0.00000246, 0.00000260, 0.00000301, 0.00000103, 0.00000639, -0.00000873, -0.00000139, },
		{ 0.00000042, -0.00000076, -0.00000085, -0.00000041, -0.00000030, -0.00000141, 0.00000436, -0.00000014, -0.00000674, 0.00000106, 0.00000439, -0.00000195, 0.00000658, -0.00000441, -0.00000206, -0.00000405, 0.00000385, -0.00000040, -0.00000003, -0.00000183, -0.00000040, -0.00000251, 0.00000562, -0.00000632, 0.00000757, -0.00000119, -0.00000896, -0.00000163, 0.00000357, -0.00000237, 0.00000679, -0.00000341, 0.00000154, -0.00000022, 0.00000037, -0.00000260, 0.00000713, -0.00000109, -0.00000546, -0.00001502, 0.00002145, -0.00000848, -0.00000489, -0.00000958, 0.00001520, -0.00001284, 0.00000394, -0.00002400, 0.00001743, -0.00002169, 0.00001405, -0.00002776, 0.00000063, -0.00000330, 0.00000176, -0.00000551, 0.00001607, -0.00000369, -0.00000308, -0.00002335, -0.00000190, -0.00003484, -0.00005067, -0.00027540, 0.00035926, 0.00012349, 0.00032552, 0.00008258, -0.00027884, 0.00000709, -0.00007335, 0.00003310, -0.00005736, 0.00001261, -0.00003012, 0.00001337, -0.00002624, 0.00001417, -0.00002528, 0.00002476, -0.00003022, 0.00000225, 0.00000291, -0.00000826, 0.00001092, -0.00000463, 0.00000156, 0.00000578, -0.00000834, -0.00000233, 0.00000667, 0.00000182, -0.00001012, -0.00000168, 0.00000617, 0.00000580, -0.00001523, 0.00000802, -0.00000605, 0.00000840, -0.00001002, 0.00000592, -0.00000763, 0.00000737, -0.00000359, 0.00000130, -0.00000142, 0.00000300, -0.00001066, 0.00000107, 0.00000643, -0.00000005, 0.00000149, -0.00000211, 0.00000077, -0.00000168, -0.00000138, -0.00000038, 0.00000084, 0.00000081, -0.00000424, -0.00000240, 0.00000352, 0.00000223, 0.00000206, 0.00000521, -0.00000941, -0.00000000, },
		{ -0.00000005, -0.00000068, -0.00000104, -0.00000014, -0.00000065, -0.00000104, 0.00000476, -0.00000122, -0.00000680, 0.00000244, 0.00000336, -0.00000120, 0.00000643, -0.00000535, -0.00000142, -0.00000418, 0.00000439, -0.00000113, 0.00000033, -0.00000211, -0.00000035, -0.00000214, 0.00000590, -0.00000708, 0.00000905, -0.00000344, -0.00000828, -0.00000097, 0.00000334, -0.00000229, 0.00000758, -0.00000485, 0.00000216, -0.00000022, 0.00000004, -0.00000189, 0.00000686, -0.00000148, -0.00000669, -0.00001279, 0.00002137, -0.00001041, -0.00000449, -0.00000816, 0.00001490, -0.00001387, 0.00000511, -0.00002494, 0.00001849, -0.00002325, 0.00001749, -0.00003394, 0.00000886, -0.00000920, 0.00000798, -0.00001160, 0.00001892, -0.00000642, -0.00000112, -0.00002706, 0.00000095, -0.00003984, -0.00004447, -0.00029474, 0.00037265, 0.00013525, 0.00036809, 0.00001254, -0.00026382, 0.00001191, -0.00007622, 0.00003745, -0.00005947, 0.00001401, -0.00003169, 0.00001548, -0.00002762, 0.00001564, -0.00002809, 0.00002962, -0.00003656, 0.00000946, -0.00000328, -0.00000351, 0.00000875, -0.00000457, 0.00000136, 0.00000610, -0.00000885, -0.00000173, 0.00000734, 0.00000002, -0.00000965, -0.00000148, 0.00000764, 0.00000298, -0.00001315, 0.00000702, -0.00000528, 0.00000786, -0.00001001, 0.00000631, -0.00000856, 0.00000909, -0.00000492, 0.00000198, -0.00000183, 0.00000314, -0.00001208, 0.00000372, 0.00000465, 0.00000089, 0.00000104, -0.00000212, 0.00000074, -0.00000184, -0.00000112, -0.00000053, 0.00000111, 0.00000037, -0.00000432, -0.00000223, 0.00000436, 0.00000138, 0.00000325, 0.00000363, -0.00000959, 0.00000117, },
		{ -0.00000047, -0.00000066, -0.00000115, 0.00000008, -0.00000101, -0.00000055, 0.00000499, -0.00000236, -0.00000661, 0.00000372, 0.00000217, -0.00000012, 0.00000580, -0.00000591, -0.00000100, -0.00000408, 0.00000463, -0.00000165, 0.00000054, -0.00000225, -0.00000048, -0.00000146, 0.00000574, -0.00000736, 0.00001010, -0.00000560, -0.00000741, -0.00000033, 0.00000290, -0.00000188, 0.00000799, -0.00000605, 0.00000270, -0.00000017, -0.00000038, -0.00000093, 0.00000618, -0.00000156, -0.00000823, -0.00000973, 0.00002006, -0.00001138, -0.00000466, -0.00000591, 0.00001333, -0.00001356, 0.00000498, -0.00002443, 0.00001780, -0.00002277, 0.00001871, -0.00003798, 0.00001568, -0.00001444, 0.00001363, -0.00001732, 0.00002142, -0.00000858, 0.00000006, -0.00002967, 0.00000237, -0.00004305, -0.00004054, -0.00031147, 0.00038176, 0.00015668, 0.00040291, -0.00006289, -0.00023725, 0.00000848, -0.00007350, 0.00003749, -0.00005816, 0.00001247, -0.00003060, 0.00001516, -0.00002667, 0.00001473, -0.00002836, 0.00003201, -0.00004089, 0.00001551, -0.00000913, 0.00000146, 0.00000607, -0.00000414, 0.00000108, 0.00000634, -0.00000927, -0.00000105, 0.00000779, -0.00000184, -0.00000893, -0.00000126, 0.00000898, -0.00000008, -0.00001044, 0.00000529, -0.00000379, 0.00000653, -0.00000915, 0.00000587, -0.00000868, 0.00001016, -0.00000582, 0.00000233, -0.00000189, 0.00000275, -0.00001288, 0.00000607, 0.00000283, 0.00000199, 0.00000043, -0.00000201, 0.00000061, -0.00000190, -0.00000092, -0.00000064, 0.00000133, -0.00000009, -0.00000438, -0.00000192, 0.00000505, 0.00000058, 0.00000449, 0.00000171, -0.00000927, 0.00000207, },
		{ -0.00000081, -0.00000070, -0.00000118, 0.00000024, -0.00000135, 0.00000003, 0.00000505, -0.00000353, -0.00000615, 0.00000482, 0.00000092, 0.00000122, 0.00000475, -0.00000608, -0.00000082, -0.00000371, 0.00000455, -0.00000191, 0.00000056, -0.00000221, -0.00000078, -0.00000050, 0.00000512, -0.00000708, 0.00001061, -0.00000754, -0.00000643, 0.00000029, 0.00000227, -0.00000115, 0.00000798, -0.00000695, 0.00000310, -0.00000006, -0.00000085, 0.00000019, 0.00000517, -0.00000142, -0.00000993, -0.00000601, 0.00001757, -0.00001140, -0.00000539, -0.00000291, 0.00001056, -0.00001194, 0.00000353, -0.00002245, 0.00001535, -0.00002013, 0.00001741, -0.00003939, 0.00002039, -0.00001836, 0.00001817, -0.00002217, 0.00002316, -0.00000984, 0.00000016, -0.00003090, 0.00000207, -0.00004414, -0.00003930, -0.00032505, 0.00038577, 0.00018995, 0.00042582, -0.00013942, -0.00020153, -0.00000240, -0.00006546, 0.00003331, -0.00005347, 0.00000799, -0.00002683, 0.00001239, -0.00002337, 0.00001136, -0.00002594, 0.00003161, -0.00004273, 0.00001979, -0.00001406, 0.00000618, 0.00000320, -0.00000351, 0.00000080, 0.00000646, -0.00000955, -0.00000031, 0.00000800, -0.00000366, -0.00000804, -0.00000097, 0.00001004, -0.00000316, -0.00000730, 0.00000300, -0.00000170, 0.00000452, -0.00000753, 0.00000461, -0.00000796, 0.00001050, -0.00000620, 0.00000228, -0.00000155, 0.00000177, -0.00001296, 0.00000795, 0.00000111, 0.00000314, -0.00000029, -0.00000181, 0.00000039, -0.00000187, -0.00000078, -0.00000069, 0.00000148, -0.00000055, -0.00000442, -0.00000145, 0.00000551, -0.00000009, 0.00000566, -0.00000043, -0.00000850, 0.00000269, },
		{ -0.00000105, -0.00000080, -0.00000113, 0.00000031, -0.00000164, 0.00000067, 0.00000494, -0.00000469, -0.00000541, 0.00000566, -0.00000026, 0.00000270, 0.00000333, -0.00000588, -0.00000087, -0.00000309, 0.00000415, -0.00000189, 0.00000037, -0.00000198, -0.00000124, 0.00000073, 0.00000407, -0.00000623, 0.00001047, -0.00000911, -0.00000542, 0.00000089, 0.00000148, -0.00000011, 0.00000754, -0.00000751, 0.00000336, 0.00000010, -0.00000132, 0.00000136, 0.00000397, -0.00000119, -0.00001157, -0.00000191, 0.00001414, -0.00001055, -0.00000655, 0.00000064, 0.00000682, -0.00000916, 0.00000092, -0.00001916, 0.00001130, -0.00001545, 0.00001357, -0.00003792, 0.00002256, -0.00002045, 0.00002115, -0.00002579, 0.00002386, -0.00000999, -0.00000096, -0.00003063, -0.00000005, -0.00004304, -0.00004085, -0.00033536, 0.00038440, 0.00023671, 0.00043281, -0.00021231, -0.00015981, -0.00001914, -0.00005297, 0.00002551, -0.00004583, 0.00000095, -0.00002071, 0.00000744, -0.00001799, 0.00000579, -0.00002098, 0.00002838, -0.00004179, 0.00002187, -0.00001761, 0.00001024, 0.00000041, -0.00000283, 0.00000057, 0.00000642, -0.00000965, 0.00000046, 0.00000793, -0.00000537, -0.00000705, -0.00000052, 0.00001065, -0.00000598, -0.00000402, 0.00000040, 0.00000075, 0.00000202, -0.00000530, 0.00000266, -0.00000646, 0.00001013, -0.00000606, 0.00000183, -0.00000083, 0.00000023, -0.00001226, 0.00000921, -0.00000033, 0.00000422, -0.00000104, -0.00000157, 0.00000013, -0.00000178, -0.00000070, -0.00000067, 0.00000155, -0.00000099, -0.00000445, -0.00000082, 0.00000571, -0.00000053, 0.00000665, -0.00000266, -0.00000733, 0.00000301, },
		{ -0.00000119, -0.00000096, -0.00000100, 0.00000029, -0.00000184, 0.00000135, 0.00000464, -0.00000579, -0.00000440, 0.00000617, -0.00000126, 0.00000418, 0.00000166, -0.00000537, -0.00000112, -0.00000227, 0.00000347, -0.00000162, -0.00000001, -0.00000160, -0.00000181, 0.00000214, 0.00000266, -0.00000485, 0.00000967, -0.00001025, -0.00000443, 0.00000146, 0.00000058, 0.00000117, 0.00000672, -0.00000774, 0.00000346, 0.00000028, -0.00000170, 0.00000248, 0.00000270, -0.00000104, -0.00001290, 0.00000223, 0.00001004, -0.00000905, -0.00000795, 0.00000444, 0.00000244, -0.00000556, -0.00000252, -0.00001493, 0.00000602, -0.00000910, 0.00000747, -0.00003364, 0.00002201, -0.00002044, 0.00002232, -0.00002795, 0.00002337, -0.00000898, -0.00000326, -0.00002897, -0.00000379, -0.00003999, -0.00004490, -0.00034281, 0.00037806, 0.00029775, 0.00042044, -0.00027676, -0.00011582, -0.00003953, -0.00003747, 0.00001510, -0.00003607, -0.00000793, -0.00001287, 0.00000088, -0.00001106, -0.00000148, -0.00001391, 0.00002260, -0.00003812, 0.00002157, -0.00001947, 0.00001333, -0.00000205, -0.00000220, 0.00000044, 0.00000620, -0.00000954, 0.00000123, 0.00000757, -0.00000686, -0.00000604, 0.00000016, 0.00001066, -0.00000831, -0.00000088, -0.00000225, 0.00000332, -0.00000071, -0.00000269, 0.00000022, -0.00000435, 0.00000913, -0.00000545, 0.00000105, 0.00000016, -0.00000178, -0.00001079, 0.00000979, -0.00000139, 0.00000513, -0.00000175, -0.00000134, -0.00000015, -0.00000163, -0.00000067, -0.00000059, 0.00000152, -0.00000139, -0.00000447, -0.00000003, 0.00000561, -0.00000066, 0.00000732, -0.00000484, -0.00000586, 0.00000306, },
		{ -0.00000126, -0.00000116, -0.00000080, 0.00000018, -0.00000195, 0.00000203, 0.00000417, -0.00000681, -0.00000314, 0.00000631, -0.00000196, 0.00000551, -0.00000011, -0.00000463, -0.00000148, -0.00000135, 0.00000258, -0.00000116, -0.00000052, -0.00000113, -0.00000240, 0.00000361, 0.00000101, -0.00000304, 0.00000823, -0.00001090, -0.00000351, 0.00000196, -0.00000033, 0.00000259, 0.00000560, -0.00000768, 0.00000344, 0.00000043, -0.00000193, 0.00000340, 0.00000152, -0.00000113, -0.00001366, 0.00000603, 0.00000567, -0.00000718, -0.00000928, 0.00000809, -0.00000214, -0.00000157, -0.00000634, -0.00001025, 0.00000010, -0.00000168, -0.00000034, -0.00002695, 0.00001888, -0.00001833, 0.00002167, -0.00002865, 0.00002176, -0.00000699, -0.00000649, -0.00002630, -0.00000868, -0.00003561, -0.00005071, -0.00034837, 0.00036799, 0.00037285, 0.00038610, -0.00032825, -0.00007348, -0.00006095, -0.00002079, 0.00000351, -0.00002535, -0.00001764, -0.00000422, -0.00000645, -0.00000336, -0.00000967, -0.00000546, 0.00001483, -0.00003205, 0.00001896, -0.00001952, 0.00001525, -0.00000402, -0.00000171, 0.00000041, 0.00000584, -0.00000924, 0.00000199, 0.00000686, -0.00000804, -0.00000509, 0.00000110, 0.00000996, -0.00000990, 0.00000183, -0.00000465, 0.00000571, -0.00000339, 0.00000001, -0.00000246, -0.00000182, 0.00000766, -0.00000448, 0.00000005, 0.00000131, -0.00000413, -0.00000864, 0.00000967, -0.00000201, 0.00000581, -0.00000237, -0.00000113, -0.00000042, -0.00000144, -0.00000068, -0.00000046, 0.00000139, -0.00000174, -0.00000447, 0.00000088, 0.00000521, -0.00000046, 0.00000759, -0.00000682, -0.00000422, 0.00000288, },
		{ -0.00000127, -0.00000138, -0.00000056, -0.00000001, -0.00000195, 0.00000269, 0.00000352, -0.00000769, -0.00000168, 0.00000604, -0.00000229, 0.00000657, -0.00000186, -0.00000380, -0.00000187, -0.00000042, 0.00000161, -0.00000060, -0.00000107, -0.00000066, -0.00000290, 0.00000501, -0.00000069, -0.00000096, 0.00000625, -0.00001107, -0.00000265, 0.00000233, -0.00000116, 0.00000403, 0.00000428, -0.00000741, 0.00000335, 0.00000051, -0.00000191, 0.00000400, 0.00000059, -0.00000164, -0.00001359, 0.00000910, 0.00000144, -0.00000529, -0.00001020, 0.00001118, -0.00000642, 0.00000228, -0.00001000, -0.00000574, -0.00000577, 0.00000603, -0.00000906, -0.00001853, 0.00001363, -0.00001440, 0.00001942, -0.00002812, 0.00001932, -0.00000439, -0.00001016, -0.00002322, -0.00001396, -0.00003083, -0.00005709, -0.00035361, 0.00035632, 0.00046058, 0.00032839, -0.00036292, -0.00003665, -0.00008060, -0.00000501, -0.00000764, -0.00001501, -0.00002700, 0.00000417, -0.00001357, 0.00000414, -0.00001783, 0.00000345, 0.00000591, -0.00002417, 0.00001435, -0.00001788, 0.00001597, -0.00000545, -0.00000136, 0.00000042, 0.00000540, -0.00000879, 0.00000276, 0.00000578, -0.00000882, -0.00000428, 0.00000235, 0.00000845, -0.00001055, 0.00000385, -0.00000652, 0.00000764, -0.00000570, 0.00000250, -0.00000507, 0.00000084, 0.00000593, -0.00000332, -0.00000103, 0.00000243, -0.00000660, -0.00000597, 0.00000891, -0.00000218, 0.00000621, -0.00000287, -0.00000097, -0.00000067, -0.00000123, -0.00000072, -0.00000027, 0.00000116, -0.00000204, -0.00000442, 0.00000188, 0.00000453, 0.00000011, 0.00000738, -0.00000847, -0.00000251, 0.00000254, },
		{ -0.00000125, -0.00000158, -0.00000030, -0.00000027, -0.00000183, 0.00000329, 0.00000271, -0.00000838, -0.00000008, 0.00000539, -0.00000217, 0.00000722, -0.00000343, -0.00000298, -0.00000216, 0.00000040, 0.00000066, -0.00000005, -0.00000157, -0.00000029, -0.00000320, 0.00000616, -0.00000226, 0.00000119, 0.00000389, -0.00001083, -0.00000182, 0.00000252, -0.00000179, 0.00000533, 0.00000290, -0.00000702, 0.00000325, 0.00000044, -0.00000157, 0.00000417, 0.00000005, -0.00000269, -0.00001248, 0.00001107, -0.00000222, -0.00000375, -0.00001038, 0.00001326, -0.00000986, 0.00000544, -0.00001290, -0.00000207, -0.00001079, 0.00001311, -0.00001770, -0.00000931, 0.00000700, -0.00000921, 0.00001605, -0.00002679, 0.00001652, -0.00000174, -0.00001360, -0.00002054, -0.00001865, -0.00002691, -0.00006251, -0.00036064, 0.00034603, 0.00055822, 0.00024728, -0.00037786, -0.00000876, -0.00009577, 0.00000779, -0.00001667, -0.00000649, -0.00003472, 0.00001114, -0.00001942, 0.00001042, -0.00002492, 0.00001174, -0.00000317, -0.00001533, 0.00000832, -0.00001485, 0.00001565, -0.00000638, -0.00000109, 0.00000037, 0.00000499, -0.00000828, 0.00000358, 0.00000429, -0.00000911, -0.00000369, 0.00000392, 0.00000613, -0.00001014, 0.00000495, -0.00000763, 0.00000882, -0.00000736, 0.00000446, -0.00000730, 0.00000336, 0.00000418, -0.00000216, -0.00000200, 0.00000331, -0.00000893, -0.00000302, 0.00000765, -0.00000193, 0.00000635, -0.00000325, -0.00000086, -0.00000090, -0.00000099, -0.00000079, -0.00000004, 0.00000083, -0.00000227, -0.00000430, 0.00000291, 0.00000363, 0.00000103, 0.00000665, -0.00000968, -0.00000086, 0.00000209, },
		{ -0.00000124, -0.00000174, -0.00000005, -0.00000056, -0.00000159, 0.00000379, 0.00000176, -0.00000886, 0.00000159, 0.00000439, -0.00000158, 0.00000737, -0.00000468, -0.00000230, -0.00000227, 0.00000099, -0.00000014, 0.00000037, -0.00000190, -0.00000013, -0.00000315, 0.00000689, -0.00000348, 0.00000316, 0.00000134, -0.00001027, -0.00000099, 0.00000243, -0.00000210, 0.00000635, 0.00000164, -0.00000663, 0.00000323, 0.00000017, -0.00000085, 0.00000381, -0.00000002, -0.00000437, -0.00001018, 0.00001164, -0.00000493, -0.00000287, -0.00000950, 0.00001395, -0.00001200, 0.00000739, -0.00001448, 0.00000012, -0.00001419, 0.00001863, -0.00002521, -0.00000037, -0.00000007, -0.00000353, 0.00001223, -0.00002530, 0.00001399, 0.00000024, -0.00001598, -0.00001922, -0.00002157, -0.00002527, -0.00006515, -0.00037195, 0.00034084, 0.00066177, 0.00014429, -0.00037142, 0.00000747, -0.00010412, 0.00001569, -0.00002199, -0.00000112, -0.00003964, 0.00001561, -0.00002296, 0.00001447, -0.00002988, 0.00001831, -0.00001129, -0.00000649, 0.00000160, -0.00001092, 0.00001453, -0.00000698, -0.00000077, 0.00000014, 0.00000476, -0.00000783, 0.00000450, 0.00000234, -0.00000884, -0.00000337, 0.00000577, 0.00000300, -0.00000860, 0.00000500, -0.00000778, 0.00000907, -0.00000812, 0.00000563, -0.00000887, 0.00000543, 0.00000266, -0.00000122, -0.00000268, 0.00000375, -0.00001088, -0.00000003, 0.00000605, -0.00000137, 0.00000627, -0.00000353, -0.00000078, -0.00000112, -0.00000072, -0.00000090, 0.00000023, 0.00000039, -0.00000243, -0.00000411, 0.00000391, 0.00000257, 0.00000222, 0.00000540, -0.00001037, 0.00000064, 0.00000159, },
		{ -0.00000126, -0.00000185, 0.00000017, -0.00000088, -0.00000123, 0.00000417, 0.00000070, -0.00000907, 0.00000324, 0.00000311, -0.00000053, 0.00000697, -0.00000553, -0.00000185, -0.00000210, 0.00000124, -0.00000066, 0.00000055, -0.00000196, -0.00000028, -0.00000267, 0.00000707, -0.00000415, 0.00000472, -0.00000117, -0.00000954, -0.00000007, 0.00000199, -0.00000196, 0.00000692, 0.00000064, -0.00000639, 0.00000339, -0.00000038, 0.00000031, 0.00000286, 0.00000045, -0.00000670, -0.00000664, 0.00001060, -0.00000639, -0.00000295, -0.00000733, 0.00001294, -0.00001243, 0.00000769, -0.00001428, 0.00000028, -0.00001531, 0.00002173, -0.00003053, 0.00000713, -0.00000651, 0.00000174, 0.00000876, -0.00002438, 0.00001248, 0.00000078, -0.00001645, -0.00002028, -0.00002152, -0.00002742, -0.00006309, -0.00039028, 0.00034502, 0.00076613, 0.00002249, -0.00034340, 0.00001029, -0.00010388, 0.00001721, -0.00002233, -0.00000001, -0.00004073, 0.00001666, -0.00002333, 0.00001545, -0.00003181, 0.00002216, -0.00001738, 0.00000131, -0.00000495, -0.00000670, 0.00001303, -0.00000747, -0.00000021, -0.00000047, 0.00000489, -0.00000757, 0.00000558, -0.00000009, -0.00000794, -0.00000337, 0.00000788, -0.00000083, -0.00000595, 0.00000393, -0.00000687, 0.00000825, -0.00000782, 0.00000580, -0.00000953, 0.00000681, 0.00000160, -0.00000067, -0.00000290, 0.00000355, -0.00001219, 0.00000271, 0.00000434, -0.00000061, 0.00000602, -0.00000376, -0.00000070, -0.00000136, -0.00000041, -0.00000105, 0.00000054, -0.00000015, -0.00000252, -0.00000381, 0.00000481, 0.00000145, 0.00000362, 0.00000369, -0.00001051, 0.00000189, 0.00000111, },
		{ -0.00000133, -0.00000188, 0.00000034, -0.00000121, -0.00000077, 0.00000440, -0.00000043, -0.00000900, 0.00000479, 0.00000163, 0.00000091, 0.00000601, -0.00000588, -0.00000172, -0.00000158, 0.00000109, -0.00000083, 0.00000041, -0.00000167, -0.00000081, -0.00000166, 0.00000658, -0.00000410, 0.00000564, -0.00000340, -0.00000878, 0.00000101, 0.00000113, -0.00000128, 0.00000692, 0.00000007, -0.00000642, 0.00000382, -0.00000124, 0.00000194, 0.00000130, 0.00000148, -0.00000966, -0.00000189, 0.00000789, -0.00000641, -0.00000414, -0.00000375, 0.00001007, -0.00001091, 0.00000604, -0.00001196, -0.00000196, -0.00001362, 0.00002173, -0.00003276, 0.00001214, -0.00001128, 0.00000565, 0.00000649, -0.00002477, 0.00001269, -0.00000088, -0.00001417, -0.00002466, -0.00001738, -0.00003476, -0.00005452, -0.00041831, 0.00036314, 0.00086525, -0.00011358, -0.00029505, -0.00000093, -0.00009408, 0.00001140, -0.00001686, -0.00000390, -0.00003734, 0.00001365, -0.00001994, 0.00001273, -0.00003004, 0.00002246, -0.00002050, 0.00000710, -0.00001049, -0.00000285, 0.00001157, -0.00000813, 0.00000077, -0.00000162, 0.00000554, -0.00000765, 0.00000689, -0.00000302, -0.00000636, -0.00000372, 0.00001015, -0.00000524, -0.00000229, 0.00000178, -0.00000491, 0.00000631, -0.00000639, 0.00000486, -0.00000911, 0.00000730, 0.00000119, -0.00000067, -0.00000255, 0.00000259, -0.00001264, 0.00000496, 0.00000274, 0.00000017, 0.00000572, -0.00000398, -0.00000057, -0.00000165, -0.00000002, -0.00000126, 0.00000089, -0.00000078, -0.00000251, -0.00000342, 0.00000555, 0.00000036, 0.00000512, 0.00000160, -0.00001010, 0.00000284, 0.00000069, },
		{ -0.00000144, -0.00000184, 0.00000045, -0.00000151, -0.00000020, 0.00000445, -0.00000159, -0.00000865, 0.00000616, 0.00000006, 0.00000268, 0.00000452, -0.00000573, -0.00000194, -0.00000070, 0.00000048, -0.00000057, -0.00000012, -0.00000099, -0.00000175, -0.00000010, 0.00000534, -0.00000321, 0.00000572, -0.00000514, -0.00000816, 0.00000232, -0.00000020, 0.00000001, 0.00000622, 0.00000005, -0.00000683, 0.00000458, -0.00000246, 0.00000401, -0.00000085, 0.00000303, -0.00001311, 0.00000391, 0.00000354, -0.00000491, -0.00000650, 0.00000124, 0.00000532, -0.00000736, 0.00000234, -0.00000740, -0.00000678, -0.00000888, 0.00001819, -0.00003124, 0.00001377, -0.00001343, 0.00000735, 0.00000618, -0.00002717, 0.00001527, -0.00000538, -0.00000846, -0.00003310, -0.00000825, -0.00004842, -0.00003791, -0.00045843, 0.00039967, 0.00095254, -0.00025810, -0.00022911, -0.00002561, -0.00007464, -0.00000199, -0.00000528, -0.00001306, -0.00002919, 0.00000631, -0.00001251, 0.00000606, -0.00002420, 0.00001872, -0.00001993, 0.00001005, -0.00001421, -0.00000001, 0.00001062, -0.00000925, 0.00000236, -0.00000347, 0.00000687, -0.00000817, 0.00000847, -0.00000644, -0.00000410, -0.00000443, 0.00001251, -0.00001002, 0.00000220, -0.00000134, -0.00000197, 0.00000333, -0.00000383, 0.00000277, -0.00000755, 0.00000679, 0.00000155, -0.00000131, -0.00000156, 0.00000078, -0.00001208, 0.00000647, 0.00000145, 0.00000084, 0.00000546, -0.00000426, -0.00000035, -0.00000202, 0.00000045, -0.00000155, 0.00000129, -0.00000152, -0.00000241, -0.00000292, 0.00000609, -0.00000060, 0.00000658, -0.00000076, -0.00000918, 0.00000346, 0.00000035, },
		{ -0.00000161, -0.00000174, 0.00000050, -0.00000179, 0.00000046, 0.00000429, -0.00000271, -0.00000800, 0.00000728, -0.00000148, 0.00000463, 0.00000257, -0.00000510, -0.00000250, 0.00000055, -0.00000057, 0.00000011, -0.00000105, 0.00000010, -0.00000310, 0.00000199, 0.00000338, -0.00000144, 0.00000485, -0.00000622, -0.00000779, 0.00000390, -0.00000201, 0.00000195, 0.00000478, 0.00000068, -0.00000771, 0.00000573, -0.00000402, 0.00000649, -0.00000353, 0.00000499, -0.00001687, 0.00001051, -0.00000225, -0.00000199, -0.00000994, 0.00000747, -0.00000115, -0.00000189, -0.00000334, -0.00000071, -0.00001406, -0.00000113, 0.00001100, -0.00002561, 0.00001144, -0.00001226, 0.00000615, 0.00000845, -0.00003209, 0.00002067, -0.00001313, 0.00000107, -0.00004606, 0.00000638, -0.00006910, -0.00001227, -0.00051243, 0.00045864, 0.00102125, -0.00040433, -0.00014961, -0.00006200, -0.00004640, -0.00002247, 0.00001208, -0.00002723, -0.00001650, -0.00000517, -0.00000122, -0.00000445, -0.00001433, 0.00001077, -0.00001531, 0.00000961, -0.00001552, 0.00000128, 0.00001057, -0.00001107, 0.00000472, -0.00000613, 0.00000898, -0.00000925, 0.00001034, -0.00001031, -0.00000118, -0.00000548, 0.00001482, -0.00001495, 0.00000728, -0.00000521, 0.00000175, -0.00000054, -0.00000030, -0.00000037, -0.00000488, 0.00000527, 0.00000270, -0.00000261, 0.00000007, -0.00000187, -0.00001043, 0.00000710, 0.00000066, 0.00000125, 0.00000533, -0.00000467, 0.00000001, -0.00000250, 0.00000104, -0.00000192, 0.00000173, -0.00000234, -0.00000218, -0.00000234, 0.00000638, -0.00000135, 0.00000788, -0.00000325, -0.00000781, 0.00000376, 0.00000010, },
		{ -0.00000181, -0.00000158, 0.00000048, -0.00000202, 0.00000119, 0.00000390, -0.00000376, -0.00000710, 0.00000807, -0.00000287, 0.00000664, 0.00000029, -0.00000404, -0.00000337, 0.00000210, -0.00000201, 0.00000119, -0.00000234, 0.00000154, -0.00000479, 0.00000454, 0.00000075, 0.00000118, 0.00000298, -0.00000652, -0.00000777, 0.00000577, -0.00000428, 0.00000449, 0.00000259, 0.00000199, -0.00000909, 0.00000727, -0.00000591, 0.00000930, -0.00000662, 0.00000724, -0.00002071, 0.00001756, -0.00000917, 0.00000215, -0.00001424, 0.00001464, -0.00000900, 0.00000520, -0.00001069, 0.00000780, -0.00002345, 0.00000929, 0.00000042, -0.00001594, 0.00000493, -0.00000739, 0.00000166, 0.00001364, -0.00003980, 0.00002906, -0.00002426, 0.00001451, -0.00006358, 0.00002656, -0.00009690, 0.00002267, -0.00058124, 0.00054326, 0.00106490, -0.00054502, -0.00006160, -0.00010730, -0.00001107, -0.00004884, 0.00003429, -0.00004564, 0.00000007, -0.00002020, 0.00001340, -0.00001828, -0.00000087, -0.00000112, -0.00000664, 0.00000553, -0.00001403, 0.00000064, 0.00001172, -0.00001375, 0.00000792, -0.00000965, 0.00001192, -0.00001092, 0.00001250, -0.00001455, 0.00000235, -0.00000684, 0.00001697, -0.00001978, 0.00001265, -0.00000957, 0.00000601, -0.00000505, 0.00000400, -0.00000437, -0.00000124, 0.00000280, 0.00000461, -0.00000452, 0.00000223, -0.00000526, -0.00000773, 0.00000676, 0.00000049, 0.00000129, 0.00000540, -0.00000524, 0.00000054, -0.00000311, 0.00000176, -0.00000240, 0.00000221, -0.00000326, -0.00000182, -0.00000170, 0.00000642, -0.00000181, 0.00000889, -0.00000572, -0.00000611, 0.00000375, -0.00000004, },
		{ -0.00000202, -0.00000139, 0.00000042, -0.00000220, 0.00000197, 0.00000328, -0.00000465, -0.00000598, 0.00000851, -0.00000400, 0.00000856, -0.00000218, -0.00000266, -0.00000445, 0.00000386, -0.00000376, 0.00000258, -0.00000391, 0.00000327, -0.00000672, 0.00000739, -0.00000238, 0.00000454, 0.00000014, -0.00000599, -0.00000814, 0.00000790, -0.00000692, 0.00000756, -0.00000028, 0.00000395, -0.00001097, 0.00000917, -0.00000805, 0.00001232, -0.00000999, 0.00000961, -0.00002436, 0.00002466, -0.00001680, 0.00000717, -0.00001907, 0.00002232, -0.00001774, 0.00001343, -0.00001921, 0.00001756, -0.00003433, 0.00002172, -0.00001295, -0.00000268, -0.00000555, 0.00000117, -0.00000618, 0.00002178, -0.00005023, 0.00004029, -0.00003849, 0.00003150, -0.00008520, 0.00005175, -0.00013124, 0.00006629, -0.00066466, 0.00065553, 0.00107783, -0.00067281, 0.00002919, -0.00015790, 0.00002890, -0.00007927, 0.00005988, -0.00006702, 0.00001940, -0.00003777, 0.00003040, -0.00003452, 0.00001535, -0.00001626, 0.00000565, -0.00000207, -0.00000966, -0.00000209, 0.00001419, -0.00001735, 0.00001194, -0.00001398, 0.00001566, -0.00001317, 0.00001487, -0.00001903, 0.00000636, -0.00000844, 0.00001881, -0.00002424, 0.00001799, -0.00001409, 0.00001049, -0.00000988, 0.00000874, -0.00000895, 0.00000313, -0.00000043, 0.00000713, -0.00000689, 0.00000478, -0.00000922, -0.00000408, 0.00000545, 0.00000100, 0.00000090, 0.00000571, -0.00000598, 0.00000123, -0.00000386, 0.00000260, -0.00000297, 0.00000273, -0.00000424, -0.00000132, -0.00000103, 0.00000620, -0.00000195, 0.00000952, -0.00000801, -0.00000420, 0.00000349, -0.00000012, },
		{ -0.00000222, -0.00000119, 0.00000032, -0.00000233, 0.00000279, 0.00000241, -0.00000535, -0.00000469, 0.00000858, -0.00000479, 0.00001024, -0.00000468, -0.00000108, -0.00000565, 0.00000571, -0.00000569, 0.00000417, -0.00000566, 0.00000514, -0.00000873, 0.00001036, -0.00000581, 0.00000844, -0.00000353, -0.00000467, -0.00000888, 0.00001020, -0.00000982, 0.00001100, -0.00000371, 0.00000648, -0.00001325, 0.00001136, -0.00001034, 0.00001542, -0.00001347, 0.00001190, -0.00002751, 0.00003135, -0.00002465, 0.00001263, -0.00002399, 0.00002998, -0.00002678, 0.00002217, -0.00002825, 0.00002786, -0.00004589, 0.00003523, -0.00002815, 0.00001333, -0.00001939, 0.00001301, -0.00001703, 0.00003254, -0.00006299, 0.00005384, -0.00005521, 0.00005127, -0.00011000, 0.00008084, -0.00017080, 0.00011706, -0.00076122, 0.00079595, 0.00105555, -0.00078084, 0.00011678, -0.00020968, 0.00007051, -0.00011143, 0.00008695, -0.00008973, 0.00004002, -0.00005654, 0.00004851, -0.00005197, 0.00003316, -0.00003362, 0.00002078, -0.00001270, -0.00000263, -0.00000685, 0.00001796, -0.00002178, 0.00001664, -0.00001896, 0.00002006, -0.00001589, 0.00001734, -0.00002358, 0.00001072, -0.00001019, 0.00002023, -0.00002808, 0.00002296, -0.00001843, 0.00001483, -0.00001466, 0.00001357, -0.00001375, 0.00000792, -0.00000417, 0.00001007, -0.00000955, 0.00000749, -0.00001349, 0.00000029, 0.00000328, 0.00000216, 0.00000007, 0.00000628, -0.00000689, 0.00000207, -0.00000471, 0.00000354, -0.00000362, 0.00000326, -0.00000528, -0.00000066, -0.00000037, 0.00000576, -0.00000174, 0.00000968, -0.00000999, -0.00000220, 0.00000303, -0.00000014, },
		{ -0.00000237, -0.00000102, 0.00000021, -0.00000240, 0.00000362, 0.00000131, -0.00000579, -0.00000331, 0.00000829, -0.00000517, 0.00001155, -0.00000705, 0.00000055, -0.00000684, 0.00000751, -0.00000764, 0.00000581, -0.00000742, 0.00000701, -0.00001065, 0.00001323, -0.00000928, 0.00001259, -0.00000780, -0.00000268, -0.00000991, 0.00001256, -0.00001278, 0.00001462, -0.00000750, 0.00000941, -0.00001583, 0.00001372, -0.00001264, 0.00001842, -0.00001687, 0.00001392, -0.00002989, 0.00003719, -0.00003219, 0.00001807, -0.00002853, 0.00003705, -0.00003544, 0.00003071, -0.00003708, 0.00003787, -0.00005715, 0.00004874, -0.00004400, 0.00003091, -0.00003557, 0.00002732, -0.00003020, 0.00004525, -0.00007733, 0.00006884, -0.00007339, 0.00007262, -0.00013656, 0.00011211, -0.00021353, 0.00017243, -0.00086809, 0.00096328, 0.00099521, -0.00086311, 0.00019535, -0.00025829, 0.00011052, -0.00014272, 0.00011334, -0.00011189, 0.00006025, -0.00007497, 0.00006628, -0.00006917, 0.00005113, -0.00005185, 0.00003760, -0.00002554, 0.00000652, -0.00001330, 0.00002279, -0.00002681, 0.00002177, -0.00002434, 0.00002491, -0.00001892, 0.00001973, -0.00002798, 0.00001523, -0.00001199, 0.00002110, -0.00003106, 0.00002725, -0.00002225, 0.00001869, -0.00001902, 0.00001810, -0.00001839, 0.00001277, -0.00000811, 0.00001315, -0.00001224, 0.00001009, -0.00001774, 0.00000511, 0.00000044, 0.00000388, -0.00000115, 0.00000706, -0.00000792, 0.00000302, -0.00000564, 0.00000456, -0.00000433, 0.00000378, -0.00000634, 0.00000015, 0.00000022, 0.00000514, -0.00000120, 0.00000933, -0.00001153, -0.00000024, 0.00000245, -0.00000014, },
		{ -0.00000247, -0.00000089, 0.00000011, -0.00000240, 0.00000441, 0.00000001, -0.00000593, -0.00000191, 0.00000768, -0.00000511, 0.00001238, -0.00000913, 0.00000208, -0.00000787, 0.00000911, -0.00000945, 0.00000732, -0.00000905, 0.00000870, -0.00001228, 0.00001576, -0.00001248, 0.00001668, -0.00001240, -0.00000021, -0.00001110, 0.00001481, -0.00001559, 0.00001816, -0.00001140, 0.00001254, -0.00001853, 0.00001608, -0.00001481, 0.00002115, -0.00002000, 0.00001549, -0.00003123, 0.00004175, -0.00003889, 0.00002297, -0.00003220, 0.00004294, -0.00004305, 0.00003833, -0.00004490, 0.00004671, -0.00006709, 0.00006103, -0.00005912, 0.00004863, -0.00005277, 0.00004294, -0.00004466, 0.00005890, -0.00009221, 0.00008413, -0.00009170, 0.00009402, -0.00016306, 0.00014339, -0.00025674, 0.00022901, -0.00098114, 0.00115447, 0.00089589, -0.00091499, 0.00025969, -0.00029960, 0.00014566, -0.00017048, 0.00013680, -0.00013155, 0.00007832, -0.00009143, 0.00008216, -0.00008461, 0.00006772, -0.00006944, 0.00005475, -0.00003949, 0.00001701, -0.00002089, 0.00002828, -0.00003208, 0.00002697, -0.00002976, 0.00002990, -0.00002203, 0.00002184, -0.00003198, 0.00001966, -0.00001370, 0.00002132, -0.00003300, 0.00003058, -0.00002525, 0.00002173, -0.00002261, 0.00002195, -0.00002247, 0.00001729, -0.00001191, 0.00001610, -0.00001470, 0.00001229, -0.00002164, 0.00001002, -0.00000281, 0.00000600, -0.00000266, 0.00000797, -0.00000900, 0.00000401, -0.00000658, 0.00000559, -0.00000506, 0.00000425, -0.00000737, 0.00000111, 0.00000072, 0.00000440, -0.00000037, 0.00000846, -0.00001255, 0.00000155, 0.00000182, -0.00000015, },
		{ -0.00000249, -0.00000083, 0.00000002, -0.00000233, 0.00000512, -0.00000147, -0.00000574, -0.00000057, 0.00000680, -0.00000461, 0.00001265, -0.00001079, 0.00000339, -0.00000864, 0.00001036, -0.00001096, 0.00000857, -0.00001036, 0.00001004, -0.00001343, 0.00001770, -0.00001513, 0.00002037, -0.00001699, 0.00000249, -0.00001227, 0.00001676, -0.00001803, 0.00002134, -0.00001514, 0.00001562, -0.00002113, 0.00001827, -0.00001665, 0.00002341, -0.00002267, 0.00001645, -0.00003136, 0.00004469, -0.00004429, 0.00002689, -0.00003456, 0.00004713, -0.00004896, 0.00004433, -0.00005097, 0.00005356, -0.00007475, 0.00007093, -0.00007212, 0.00006496, -0.00006948, 0.00005852, -0.00005913, 0.00007228, -0.00010635, 0.00009833, -0.00010860, 0.00011367, -0.00018741, 0.00017213, -0.00029729, 0.00028269, -0.00109502, 0.00136462, 0.00075874, -0.00093350, 0.00030558, -0.00032995, 0.00017296, -0.00019225, 0.00015523, -0.00014683, 0.00009256, -0.00010437, 0.00009469, -0.00009682, 0.00008142, -0.00008483, 0.00007073, -0.00005329, 0.00002786, -0.00002891, 0.00003389, -0.00003714, 0.00003180, -0.00003481, 0.00003466, -0.00002494, 0.00002342, -0.00003530, 0.00002377, -0.00001520, 0.00002081, -0.00003376, 0.00003274, -0.00002717, 0.00002369, -0.00002513, 0.00002478, -0.00002564, 0.00002112, -0.00001525, 0.00001863, -0.00001668, 0.00001382, -0.00002485, 0.00001466, -0.00000619, 0.00000831, -0.00000431, 0.00000891, -0.00001003, 0.00000494, -0.00000744, 0.00000656, -0.00000575, 0.00000462, -0.00000832, 0.00000217, 0.00000108, 0.00000361, 0.00000068, 0.00000710, -0.00001299, 0.00000309, 0.00000120, -0.00000019, },
		{ -0.00000242, -0.00000083, -0.00000005, -0.00000219, 0.00000572, -0.00000305, -0.00000523, 0.00000064, 0.00000573, -0.00000370, 0.00001234, -0.00001193, 0.00000435, -0.00000903, 0.00001116, -0.00001204, 0.00000939, -0.00001122, 0.00001089, -0.00001395, 0.00001886, -0.00001696, 0.00002331, -0.00002122, 0.00000515, -0.00001322, 0.00001822, -0.00001985, 0.00002390, -0.00001840, 0.00001837, -0.00002339, 0.00002007, -0.00001799, 0.00002502, -0.00002472, 0.00001669, -0.00003014, 0.00004575, -0.00004800, 0.00002944, -0.00003526, 0.00004920, -0.00005266, 0.00004813, -0.00005467, 0.00005771, -0.00007927, 0.00007736, -0.00008170, 0.00007838, -0.00008413, 0.00007255, -0.00007222, 0.00008403, -0.00011838, 0.00010997, -0.00012245, 0.00012971, -0.00020740, 0.00019564, -0.00033179, 0.00032890, -0.00120347, 0.00158720, 0.00058702, -0.00091758, 0.00033009, -0.00034653, 0.00019000, -0.00020595, 0.00016685, -0.00015617, 0.00010153, -0.00011248, 0.00010258, -0.00010450, 0.00009084, -0.00009654, 0.00008408, -0.00006561, 0.00003799, -0.00003653, 0.00003899, -0.00004147, 0.00003580, -0.00003904, 0.00003880, -0.00002735, 0.00002421, -0.00003766, 0.00002729, -0.00001636, 0.00001954, -0.00003326, 0.00003359, -0.00002785, 0.00002436, -0.00002633, 0.00002633, -0.00002760, 0.00002393, -0.00001780, 0.00002047, -0.00001795, 0.00001443, -0.00002703, 0.00001866, -0.00000935, 0.00001057, -0.00000594, 0.00000974, -0.00001089, 0.00000571, -0.00000814, 0.00000739, -0.00000633, 0.00000483, -0.00000913, 0.00000331, 0.00000128, 0.00000285, 0.00000185, 0.00000532, -0.00001284, 0.00000431, 0.00000064, -0.00000028, },
		{ -0.00000227, -0.00000091, -0.00000009, -0.00000195, 0.00000615, -0.00000468, -0.00000441, 0.00000167, 0.00000456, -0.00000244, 0.00001144, -0.00001250, 0.00000490, -0.00000898, 0.00001142, -0.00001257, 0.00000968, -0.00001152, 0.00001113, -0.00001371, 0.00001906, -0.00001772, 0.00002520, -0.00002474, 0.00000749, -0.00001375, 0.00001898, -0.00002083, 0.00002557, -0.00002091, 0.00002051, -0.00002507, 0.00002128, -0.00001867, 0.00002583, -0.00002601, 0.00001614, -0.00002755, 0.00004479, -0.00004974, 0.00003033, -0.00003405, 0.00004886, -0.00005378, 0.00004931, -0.00005555, 0.00005865, -0.00008004, 0.00007952, -0.00008675, 0.00008755, -0.00009524, 0.00008357, -0.00008253, 0.00009282, -0.00012696, 0.00011760, -0.00013166, 0.00014034, -0.00022090, 0.00021131, -0.00035687, 0.00036296, -0.00129968, 0.00181430, 0.00038602, -0.00086814, 0.00033176, -0.00034753, 0.00019512, -0.00021010, 0.00017036, -0.00015841, 0.00010418, -0.00011476, 0.00010491, -0.00010670, 0.00009491, -0.00010337, 0.00009347, -0.00007520, 0.00004633, -0.00004290, 0.00004293, -0.00004453, 0.00003848, -0.00004200, 0.00004191, -0.00002893, 0.00002399, -0.00003882, 0.00002999, -0.00001705, 0.00001750, -0.00003149, 0.00003306, -0.00002718, 0.00002365, -0.00002610, 0.00002643, -0.00002812, 0.00002548, -0.00001934, 0.00002142, -0.00001833, 0.00001395, -0.00002795, 0.00002169, -0.00001200, 0.00001253, -0.00000737, 0.00001034, -0.00001147, 0.00000622, -0.00000857, 0.00000800, -0.00000675, 0.00000484, -0.00000973, 0.00000447, 0.00000132, 0.00000218, 0.00000304, 0.00000322, -0.00001212, 0.00000516, 0.00000017, -0.00000043, },
		{ -0.00000205, -0.00000104, -0.00000011, -0.00000162, 0.00000636, -0.00000626, -0.00000331, 0.00000245, 0.00000337, -0.00000093, 0.00001001, -0.00001247, 0.00000498, -0.00000845, 0.00001109, -0.00001251, 0.00000937, -0.00001118, 0.00001071, -0.00001265, 0.00001823, -0.00001728, 0.00002580, -0.00002725, 0.00000924, -0.00001367, 0.00001890, -0.00002080, 0.00002612, -0.00002241, 0.00002179, -0.00002595, 0.00002173, -0.00001852, 0.00002571, -0.00002643, 0.00001480, -0.00002363, 0.00004180, -0.00004939, 0.00002943, -0.00003084, 0.00004600, -0.00005213, 0.00004764, -0.00005337, 0.00005614, -0.00007672, 0.00007690, -0.00008653, 0.00009140, -0.00010157, 0.00009032, -0.00008883, 0.00009744, -0.00013089, 0.00011998, -0.00013486, 0.00014400, -0.00022608, 0.00021682, -0.00036947, 0.00038042, -0.00137667, 0.00203708, 0.00016278, -0.00078803, 0.00031076, -0.00033233, 0.00018755, -0.00020395, 0.00016508, -0.00015293, 0.00009994, -0.00011070, 0.00010118, -0.00010288, 0.00009297, -0.00010442, 0.00009787, -0.00008099, 0.00005191, -0.00004724, 0.00004511, -0.00004584, 0.00003943, -0.00004328, 0.00004363, -0.00002942, 0.00002254, -0.00003855, 0.00003164, -0.00001716, 0.00001472, -0.00002851, 0.00003117, -0.00002520, 0.00002155, -0.00002440, 0.00002502, -0.00002710, 0.00002561, -0.00001970, 0.00002135, -0.00001774, 0.00001230, -0.00002743, 0.00002350, -0.00001387, 0.00001397, -0.00000841, 0.00001057, -0.00001166, 0.00000637, -0.00000866, 0.00000830, -0.00000694, 0.00000459, -0.00001006, 0.00000558, 0.00000119, 0.00000166, 0.00000415, 0.00000091, -0.00001090, 0.00000563, -0.00000017, -0.00000063, },
		{ -0.00000180, -0.00000121, -0.00000012, -0.00000117, 0.00000629, -0.00000771, -0.00000199, 0.00000298, 0.00000224, 0.00000073, 0.00000812, -0.00001186, 0.00000461, -0.00000744, 0.00001017, -0.00001183, 0.00000845, -0.00001019, 0.00000962, -0.00001078, 0.00001635, -0.00001558, 0.00002497, -0.00002854, 0.00001020, -0.00001285, 0.00001787, -0.00001963, 0.00002540, -0.00002270, 0.00002202, -0.00002586, 0.00002125, -0.00001744, 0.00002458, -0.00002595, 0.00001270, -0.00001854, 0.00003688, -0.00004695, 0.00002674, -0.00002570, 0.00004069, -0.00004773, 0.00004313, -0.00004815, 0.00005020, -0.00006931, 0.00006936, -0.00008066, 0.00008927, -0.00010222, 0.00009180, -0.00009014, 0.00009698, -0.00012926, 0.00011617, -0.00013105, 0.00013953, -0.00022157, 0.00021043, -0.00036716, 0.00037745, -0.00142780, 0.00224623, -0.00007432, -0.00068183, 0.00026874, -0.00030152, 0.00016749, -0.00018753, 0.00015102, -0.00013970, 0.00008877, -0.00010028, 0.00009137, -0.00009298, 0.00008481, -0.00009930, 0.00009662, -0.00008219, 0.00005397, -0.00004889, 0.00004503, -0.00004503, 0.00003830, -0.00004259, 0.00004367, -0.00002861, 0.00001973, -0.00003671, 0.00003207, -0.00001663, 0.00001127, -0.00002445, 0.00002804, -0.00002198, 0.00001816, -0.00002133, 0.00002215, -0.00002455, 0.00002429, -0.00001882, 0.00002020, -0.00001617, 0.00000947, -0.00002541, 0.00002393, -0.00001473, 0.00001469, -0.00000892, 0.00001033, -0.00001138, 0.00000608, -0.00000833, 0.00000824, -0.00000684, 0.00000403, -0.00001004, 0.00000658, 0.00000094, 0.00000132, 0.00000508, -0.00000146, -0.00000927, 0.00000574, -0.00000040, -0.00000086, },
		{ -0.00000152, -0.00000140, -0.00000013, -0.00000061, 0.00000592, -0.00000893, -0.00000054, 0.00000324, 0.00000125, 0.00000243, 0.00000589, -0.00001075, 0.00000380, -0.00000601, 0.00000872, -0.00001057, 0.00000694, -0.00000859, 0.00000791, -0.00000819, 0.00001352, -0.00001266, 0.00002265, -0.00002845, 0.00001021, -0.00001121, 0.00001586, -0.00001729, 0.00002334, -0.00002166, 0.00002106, -0.00002467, 0.00001976, -0.00001537, 0.00002239, -0.00002454, 0.00000994, -0.00001251, 0.00003029, -0.00004259, 0.00002240, -0.00001886, 0.00003320, -0.00004084, 0.00003600, -0.00004015, 0.00004116, -0.00005816, 0.00005720, -0.00006925, 0.00008098, -0.00009676, 0.00008745, -0.00008586, 0.00009086, -0.00012155, 0.00010566, -0.00011968, 0.00012634, -0.00020662, 0.00019111, -0.00034838, 0.00035120, -0.00144726, 0.00243255, -0.00031601, -0.00055553, 0.00020878, -0.00025684, 0.00013608, -0.00016170, 0.00012884, -0.00011932, 0.00007122, -0.00008398, 0.00007596, -0.00007745, 0.00007076, -0.00008808, 0.00008951, -0.00007840, 0.00005205, -0.00004745, 0.00004238, -0.00004185, 0.00003493, -0.00003973, 0.00004184, -0.00002637, 0.00001552, -0.00003322, 0.00003117, -0.00001541, 0.00000727, -0.00001950, 0.00002385, -0.00001772, 0.00001368, -0.00001705, 0.00001799, -0.00002059, 0.00002157, -0.00001673, 0.00001802, -0.00001369, 0.00000558, -0.00002195, 0.00002290, -0.00001448, 0.00001457, -0.00000882, 0.00000955, -0.00001057, 0.00000532, -0.00000754, 0.00000776, -0.00000644, 0.00000315, -0.00000964, 0.00000738, 0.00000059, 0.00000119, 0.00000574, -0.00000376, -0.00000733, 0.00000552, -0.00000051, -0.00000111, },
		{ -0.00000127, -0.00000157, -0.00000015, 0.00000007, 0.00000521, -0.00000983, 0.00000097, 0.00000326, 0.00000046, 0.00000406, 0.00000344, -0.00000921, 0.00000265, -0.00000423, 0.00000681, -0.00000882, 0.00000493, -0.00000648, 0.00000569, -0.00000501, 0.00000988, -0.00000866, 0.00001892, -0.00002696, 0.00000923, -0.00000874, 0.00001289, -0.00001383, 0.00001995, -0.00001928, 0.00001885, -0.00002232, 0.00001723, -0.00001232, 0.00001916, -0.00002225, 0.00000666, -0.00000582, 0.00002237, -0.00003658, 0.00001670, -0.00001068, 0.00002396, -0.00003187, 0.00002671, -0.00002987, 0.00002958, -0.00004393, 0.00004110, -0.00005287, 0.00006687, -0.00008524, 0.00007719, -0.00007587, 0.00007901, -0.00010773, 0.00008847, -0.00010079, 0.00010447, -0.00018126, 0.00015873, -0.00031266, 0.00030013, -0.00143052, 0.00258754, -0.00055271, -0.00041612, 0.00013506, -0.00020102, 0.00009532, -0.00012798, 0.00009985, -0.00009293, 0.00004832, -0.00006280, 0.00005592, -0.00005720, 0.00005162, -0.00007134, 0.00007682, -0.00006963, 0.00004601, -0.00004273, 0.00003703, -0.00003627, 0.00002928, -0.00003466, 0.00003809, -0.00002269, 0.00000995, -0.00002812, 0.00002889, -0.00001352, 0.00000285, -0.00001387, 0.00001884, -0.00001267, 0.00000838, -0.00001186, 0.00001281, -0.00001546, 0.00001764, -0.00001357, 0.00001494, -0.00001048, 0.00000084, -0.00001721, 0.00002048, -0.00001308, 0.00001356, -0.00000805, 0.00000822, -0.00000924, 0.00000408, -0.00000629, 0.00000686, -0.00000570, 0.00000193, -0.00000882, 0.00000793, 0.00000021, 0.00000126, 0.00000606, -0.00000586, -0.00000521, 0.00000503, -0.00000054, -0.00000135, },
		{ -0.00000106, -0.00000171, -0.00000019, 0.00000085, 0.00000414, -0.00001034, 0.00000242, 0.00000306, -0.00000010, 0.00000551, 0.00000090, -0.00000739, 0.00000124, -0.00000222, 0.00000459, -0.00000670, 0.00000256, -0.00000398, 0.00000313, -0.00000145, 0.00000569, -0.00000383, 0.00001396, -0.00002414, 0.00000727, -0.00000550, 0.00000910, -0.00000939, 0.00001537, -0.00001562, 0.00001545, -0.00001886, 0.00001369, -0.00000834, 0.00001499, -0.00001916, 0.00000304, 0.00000117, 0.00001353, -0.00002931, 0.00001000, -0.00000162, 0.00001350, -0.00002140, 0.00001585, -0.00001799, 0.00001628, -0.00002757, 0.00002205, -0.00003251, 0.00004773, -0.00006824, 0.00006141, -0.00006052, 0.00006179, -0.00008825, 0.00006513, -0.00007501, 0.00007465, -0.00014631, 0.00011415, -0.00026073, 0.00022422, -0.00137473, 0.00270390, -0.00077514, -0.00027111, 0.00005255, -0.00013752, 0.00004784, -0.00008853, 0.00006587, -0.00006214, 0.00002152, -0.00003811, 0.00003258, -0.00003356, 0.00002864, -0.00005014, 0.00005929, -0.00005629, 0.00003606, -0.00003488, 0.00002913, -0.00002843, 0.00002153, -0.00002754, 0.00003251, -0.00001766, 0.00000318, -0.00002154, 0.00002528, -0.00001098, -0.00000182, -0.00000784, 0.00001329, -0.00000713, 0.00000257, -0.00000609, 0.00000694, -0.00000947, 0.00001275, -0.00000955, 0.00001116, -0.00000677, -0.00000447, -0.00001147, 0.00001684, -0.00001061, 0.00001169, -0.00000664, 0.00000637, -0.00000741, 0.00000239, -0.00000460, 0.00000556, -0.00000466, 0.00000040, -0.00000756, 0.00000816, -0.00000014, 0.00000150, 0.00000602, -0.00000765, -0.00000304, 0.00000433, -0.00000050, -0.00000156, },
		{ -0.00000092, -0.00000178, -0.00000025, 0.00000171, 0.00000274, -0.00001039, 0.00000374, 0.00000272, -0.00000042, 0.00000669, -0.00000158, -0.00000539, -0.00000031, -0.00000011, 0.00000218, -0.00000436, -0.00000004, -0.00000126, 0.00000041, 0.00000225, 0.00000122, 0.00000151, 0.00000807, -0.00002018, 0.00000446, -0.00000166, 0.00000468, -0.00000419, 0.00000981, -0.00001087, 0.00001099, -0.00001441, 0.00000927, -0.00000358, 0.00000999, -0.00001541, -0.00000072, 0.00000811, 0.00000426, -0.00002122, 0.00000275, 0.00000777, 0.00000247, -0.00001012, 0.00000415, -0.00000531, 0.00000220, -0.00001019, 0.00000136, -0.00000949, 0.00002481, -0.00004680, 0.00004102, -0.00004062, 0.00004006, -0.00006407, 0.00003672, -0.00004357, 0.00003827, -0.00010336, 0.00005920, -0.00019459, 0.00012510, -0.00127899, 0.00277607, -0.00097480, -0.00012803, -0.00003338, -0.00007030, -0.00000327, -0.00004587, 0.00002909, -0.00002888, -0.00000740, -0.00001155, 0.00000755, -0.00000814, 0.00000339, -0.00002591, 0.00003806, -0.00003922, 0.00002277, -0.00002429, 0.00001904, -0.00001871, 0.00001202, -0.00001866, 0.00002534, -0.00001152, -0.00000452, -0.00001371, 0.00002047, -0.00000791, -0.00000655, -0.00000166, 0.00000750, -0.00000142, -0.00000340, -0.00000011, 0.00000077, -0.00000299, 0.00000724, -0.00000496, 0.00000695, -0.00000284, -0.00000999, -0.00000507, 0.00001223, -0.00000723, 0.00000906, -0.00000466, 0.00000408, -0.00000518, 0.00000034, -0.00000255, 0.00000390, -0.00000336, -0.00000138, -0.00000590, 0.00000804, -0.00000040, 0.00000187, 0.00000559, -0.00000904, -0.00000092, 0.00000350, -0.00000042, -0.00000171, },
		{ -0.00000088, -0.00000178, -0.00000033, 0.00000261, 0.00000103, -0.00000996, 0.00000482, 0.00000230, -0.00000051, 0.00000755, -0.00000388, -0.00000334, -0.00000188, 0.00000198, -0.00000025, -0.00000196, -0.00000267, 0.00000148, -0.00000225, 0.00000584, -0.00000319, 0.00000699, 0.00000159, -0.00001535, 0.00000100, 0.00000258, -0.00000011, 0.00000147, 0.00000357, -0.00000529, 0.00000570, -0.00000916, 0.00000415, 0.00000179, 0.00000436, -0.00001114, -0.00000444, 0.00001466, -0.00000498, -0.00001280, -0.00000458, 0.00001694, -0.00000846, 0.00000126, -0.00000762, 0.00000731, -0.00001165, 0.00000699, -0.00001955, 0.00001459, -0.00000032, -0.00002236, 0.00001731, -0.00001742, 0.00001509, -0.00003656, 0.00000476, -0.00000816, -0.00000271, -0.00005473, -0.00000343, -0.00011741, 0.00000608, -0.00114455, 0.00280052, -0.00114455, 0.00000608, -0.00011741, -0.00000343, -0.00005473, -0.00000271, -0.00000816, 0.00000476, -0.00003656, 0.00001509, -0.00001742, 0.00001731, -0.00002236, -0.00000032, 0.00001459, -0.00001955, 0.00000699, -0.00001165, 0.00000731, -0.00000762, 0.00000126, -0.00000846, 0.00001694, -0.00000458, -0.00001280, -0.00000498, 0.00001466, -0.00000444, -0.00001114, 0.00000436, 0.00000179, 0.00000415, -0.00000916, 0.00000570, -0.00000529, 0.00000357, 0.00000147, -0.00000011, 0.00000258, 0.00000100, -0.00001535, 0.00000159, 0.00000699, -0.00000319, 0.00000584, -0.00000225, 0.00000148, -0.00000267, -0.00000196, -0.00000025, 0.00000198, -0.00000188, -0.00000334, -0.00000388, 0.00000755, -0.00000051, 0.00000230, 0.00000482, -0.00000996, 0.00000103, 0.00000261, -0.00000033, -0.00000178, },
		{ -0.00000092, -0.00000171, -0.00000042, 0.00000350, -0.00000092, -0.00000904, 0.00000559, 0.00000187, -0.00000040, 0.00000804, -0.00000590, -0.00000138, -0.00000336, 0.00000390, -0.00000255, 0.00000034, -0.00000518, 0.00000408, -0.00000466, 0.00000906, -0.00000723, 0.00001223, -0.00000507, -0.00000999, -0.00000284, 0.00000695, -0.00000496, 0.00000724, -0.00000299, 0.00000077, -0.00000011, -0.00000340, -0.00000142, 0.00000750, -0.00000166, -0.00000655, -0.00000791, 0.00002047, -0.00001371, -0.00000452, -0.00001152, 0.00002534, -0.00001866, 0.00001202, -0.00001871, 0.00001904, -0.00002429, 0.00002277, -0.00003922, 0.00003806, -0.00002591, 0.00000339, -0.00000814, 0.00000755, -0.00001155, -0.00000740, -0.00002888, 0.00002909, -0.00004587, -0.00000327, -0.00007030, -0.00003338, -0.00012803, -0.00097480, 0.00277607, -0.00127899, 0.00012510, -0.00019459, 0.00005920, -0.00010336, 0.00003827, -0.00004357, 0.00003672, -0.00006407, 0.00004006, -0.00004062, 0.00004102, -0.00004680, 0.00002481, -0.00000949, 0.00000136, -0.00001019, 0.00000220, -0.00000531, 0.00000415, -0.00001012, 0.00000247, 0.00000777, 0.00000275, -0.00002122, 0.00000426, 0.00000811, -0.00000072, -0.00001541, 0.00000999, -0.00000358, 0.00000927, -0.00001441, 0.00001099, -0.00001087, 0.00000981, -0.00000419, 0.00000468, -0.00000166, 0.00000446, -0.00002018, 0.00000807, 0.00000151, 0.00000122, 0.00000225, 0.00000041, -0.00000126, -0.00000004, -0.00000436, 0.00000218, -0.00000011, -0.00000031, -0.00000539, -0.00000158, 0.00000669, -0.00000042, 0.00000272, 0.00000374, -0.00001039, 0.00000274, 0.00000171, -0.00000025, -0.00000178, },
		{ -0.00000106, -0.00000156, -0.00000050, 0.00000433, -0.00000304, -0.00000765, 0.00000602, 0.00000150, -0.00000014, 0.00000816, -0.00000756, 0.00000040, -0.00000466, 0.00000556, -0.00000460, 0.00000239, -0.00000741, 0.00000637, -0.00000664, 0.00001169, -0.00001061, 0.00001684, -0.00001147, -0.00000447, -0.00000677, 0.00001116, -0.00000955, 0.00001275, -0.00000947, 0.00000694, -0.00000609, 0.00000257, -0.00000713, 0.00001329, -0.00000784, -0.00000182, -0.00001098, 0.00002528, -0.00002154, 0.00000318, -0.00001766, 0.00003251, -0.00002754, 0.00002153, -0.00002843, 0.00002913, -0.00003488, 0.00003606, -0.00005629, 0.00005929, -0.00005014, 0.00002864, -0.00003356, 0.00003258, -0.00003811, 0.00002152, -0.00006214, 0.00006587, -0.00008853, 0.00004784, -0.00013752, 0.00005255, -0.00027111, -0.00077514, 0.00270390, -0.00137473, 0.00022422, -0.00026073, 0.00011415, -0.00014631, 0.00007465, -0.00007501, 0.00006513, -0.00008825, 0.00006179, -0.00006052, 0.00006141, -0.00006824, 0.00004773, -0.00003251, 0.00002205, -0.00002757, 0.00001628, -0.00001799, 0.00001585, -0.00002140, 0.00001350, -0.00000162, 0.00001000, -0.00002931, 0.00001353, 0.00000117, 0.00000304, -0.00001916, 0.00001499, -0.00000834, 0.00001369, -0.00001886, 0.00001545, -0.00001562, 0.00001537, -0.00000939, 0.00000910, -0.00000550, 0.00000727, -0.00002414, 0.00001396, -0.00000383, 0.00000569, -0.00000145, 0.00000313, -0.00000398, 0.00000256, -0.00000670, 0.00000459, -0.00000222, 0.00000124, -0.00000739, 0.00000090, 0.00000551, -0.00000010, 0.00000306, 0.00000242, -0.00001034, 0.00000414, 0.00000085, -0.00000019, -0.00000171, },
		{ -0.00000127, -0.00000135, -0.00000054, 0.00000503, -0.00000521, -0.00000586, 0.00000606, 0.00000126, 0.00000021, 0.00000793, -0.00000882, 0.00000193, -0.00000570, 0.00000686, -0.00000629, 0.00000408, -0.00000924, 0.00000822, -0.00000805, 0.00001356, -0.00001308, 0.00002048, -0.00001721, 0.00000084, -0.00001048, 0.00001494, -0.00001357, 0.00001764, -0.00001546, 0.00001281, -0.00001186, 0.00000838, -0.00001267, 0.00001884, -0.00001387, 0.00000285, -0.00001352, 0.00002889, -0.00002812, 0.00000995, -0.00002269, 0.00003809, -0.00003466, 0.00002928, -0.00003627, 0.00003703, -0.00004273, 0.00004601, -0.00006963, 0.00007682, -0.00007134, 0.00005162, -0.00005720, 0.00005592, -0.00006280, 0.00004832, -0.00009293, 0.00009985, -0.00012798, 0.00009532, -0.00020102, 0.00013506, -0.00041612, -0.00055271, 0.00258754, -0.00143052, 0.00030013, -0.00031266, 0.00015873, -0.00018126, 0.00010447, -0.00010079, 0.00008847, -0.00010773, 0.00007901, -0.00007587, 0.00007719, -0.00008524, 0.00006687, -0.00005287, 0.00004110, -0.00004393, 0.00002958, -0.00002987, 0.00002671, -0.00003187, 0.00002396, -0.00001068, 0.00001670, -0.00003658, 0.00002237, -0.00000582, 0.00000666, -0.00002225, 0.00001916, -0.00001232, 0.00001723, -0.00002232, 0.00001885, -0.00001928, 0.00001995, -0.00001383, 0.00001289, -0.00000874, 0.00000923, -0.00002696, 0.00001892, -0.00000866, 0.00000988, -0.00000501, 0.00000569, -0.00000648, 0.00000493, -0.00000882, 0.00000681, -0.00000423, 0.00000265, -0.00000921, 0.00000344, 0.00000406, 0.00000046, 0.00000326, 0.00000097, -0.00000983, 0.00000521, 0.00000007, -0.00000015, -0.00000157, },
		{ -0.00000152, -0.00000111, -0.00000051, 0.00000552, -0.00000733, -0.00000376, 0.00000574, 0.00000119, 0.00000059, 0.00000738, -0.00000964, 0.00000315, -0.00000644, 0.00000776, -0.00000754, 0.00000532, -0.00001057, 0.00000955, -0.00000882, 0.00001457, -0.00001448, 0.00002290, -0.00002195, 0.00000558, -0.00001369, 0.00001802, -0.00001673, 0.00002157, -0.00002059, 0.00001799, -0.00001705, 0.00001368, -0.00001772, 0.00002385, -0.00001950, 0.00000727, -0.00001541, 0.00003117, -0.00003322, 0.00001552, -0.00002637, 0.00004184, -0.00003973, 0.00003493, -0.00004185, 0.00004238, -0.00004745, 0.00005205, -0.00007840, 0.00008951, -0.00008808, 0.00007076, -0.00007745, 0.00007596, -0.00008398, 0.00007122, -0.00011932, 0.00012884, -0.00016170, 0.00013608, -0.00025684, 0.00020878, -0.00055553, -0.00031601, 0.00243255, -0.00144726, 0.00035120, -0.00034838, 0.00019111, -0.00020662, 0.00012634, -0.00011968, 0.00010566, -0.00012155, 0.00009086, -0.00008586, 0.00008745, -0.00009676, 0.00008098, -0.00006925, 0.00005720, -0.00005816, 0.00004116, -0.00004015, 0.00003600, -0.00004084, 0.00003320, -0.00001886, 0.00002240, -0.00004259, 0.00003029, -0.00001251, 0.00000994, -0.00002454, 0.00002239, -0.00001537, 0.00001976, -0.00002467, 0.00002106, -0.00002166, 0.00002334, -0.00001729, 0.00001586, -0.00001121, 0.00001021, -0.00002845, 0.00002265, -0.00001266, 0.00001352, -0.00000819, 0.00000791, -0.00000859, 0.00000694, -0.00001057, 0.00000872, -0.00000601, 0.00000380, -0.00001075, 0.00000589, 0.00000243, 0.00000125, 0.00000324, -0.00000054, -0.00000893, 0.00000592, -0.00000061, -0.00000013, -0.00000140, },
		{ -0.00000180, -0.00000086, -0.00000040, 0.00000574, -0.00000927, -0.00000146, 0.00000508, 0.00000132, 0.00000094, 0.00000658, -0.00001004, 0.00000403, -0.00000684, 0.00000824, -0.00000833, 0.00000608, -0.00001138, 0.00001033, -0.00000892, 0.00001469, -0.00001473, 0.00002393, -0.00002541, 0.00000947, -0.00001617, 0.00002020, -0.00001882, 0.00002429, -0.00002455, 0.00002215, -0.00002133, 0.00001816, -0.00002198, 0.00002804, -0.00002445, 0.00001127, -0.00001663, 0.00003207, -0.00003671, 0.00001973, -0.00002861, 0.00004367, -0.00004259, 0.00003830, -0.00004503, 0.00004503, -0.00004889, 0.00005397, -0.00008219, 0.00009662, -0.00009930, 0.00008481, -0.00009298, 0.00009137, -0.00010028, 0.00008877, -0.00013970, 0.00015102, -0.00018753, 0.00016749, -0.00030152, 0.00026874, -0.00068183, -0.00007432, 0.00224623, -0.00142780, 0.00037745, -0.00036716, 0.00021043, -0.00022157, 0.00013953, -0.00013105, 0.00011617, -0.00012926, 0.00009698, -0.00009014, 0.00009180, -0.00010222, 0.00008927, -0.00008066, 0.00006936, -0.00006931, 0.00005020, -0.00004815, 0.00004313, -0.00004773, 0.00004069, -0.00002570, 0.00002674, -0.00004695, 0.00003688, -0.00001854, 0.00001270, -0.00002595, 0.00002458, -0.00001744, 0.00002125, -0.00002586, 0.00002202, -0.00002270, 0.00002540, -0.00001963, 0.00001787, -0.00001285, 0.00001020, -0.00002854, 0.00002497, -0.00001558, 0.00001635, -0.00001078, 0.00000962, -0.00001019, 0.00000845, -0.00001183, 0.00001017, -0.00000744, 0.00000461, -0.00001186, 0.00000812, 0.00000073, 0.00000224, 0.00000298, -0.00000199, -0.00000771, 0.00000629, -0.00000117, -0.00000012, -0.00000121, },
		{ -0.00000205, -0.00000063, -0.00000017, 0.00000563, -0.00001090, 0.00000091, 0.00000415, 0.00000166, 0.00000119, 0.00000558, -0.00001006, 0.00000459, -0.00000694, 0.00000830, -0.00000866, 0.00000637, -0.00001166, 0.00001057, -0.00000841, 0.00001397, -0.00001387, 0.00002350, -0.00002743, 0.00001230, -0.00001774, 0.00002135, -0.00001970, 0.00002561, -0.00002710, 0.00002502, -0.00002440, 0.00002155, -0.00002520, 0.00003117, -0.00002851, 0.00001472, -0.00001716, 0.00003164, -0.00003855, 0.00002254, -0.00002942, 0.00004363, -0.00004328, 0.00003943, -0.00004584, 0.00004511, -0.00004724, 0.00005191, -0.00008099, 0.00009787, -0.00010442, 0.00009297, -0.00010288, 0.00010118, -0.00011070, 0.00009994, -0.00015293, 0.00016508, -0.00020395, 0.00018755, -0.00033233, 0.00031076, -0.00078803, 0.00016278, 0.00203708, -0.00137667, 0.00038042, -0.00036947, 0.00021682, -0.00022608, 0.00014400, -0.00013486, 0.00011998, -0.00013089, 0.00009744, -0.00008883, 0.00009032, -0.00010157, 0.00009140, -0.00008653, 0.00007690, -0.00007672, 0.00005614, -0.00005337, 0.00004764, -0.00005213, 0.00004600, -0.00003084, 0.00002943, -0.00004939, 0.00004180, -0.00002363, 0.00001480, -0.00002643, 0.00002571, -0.00001852, 0.00002173, -0.00002595, 0.00002179, -0.00002241, 0.00002612, -0.00002080, 0.00001890, -0.00001367, 0.00000924, -0.00002725, 0.00002580, -0.00001728, 0.00001823, -0.00001265, 0.00001071, -0.00001118, 0.00000937, -0.00001251, 0.00001109, -0.00000845, 0.00000498, -0.00001247, 0.00001001, -0.00000093, 0.00000337, 0.00000245, -0.00000331, -0.00000626, 0.00000636, -0.00000162, -0.00000011, -0.00000104, },
		{ -0.00000227, -0.00000043, 0.00000017, 0.00000516, -0.00001212, 0.00000322, 0.00000304, 0.00000218, 0.00000132, 0.00000447, -0.00000973, 0.00000484, -0.00000675, 0.00000800, -0.00000857, 0.00000622, -0.00001147, 0.00001034, -0.00000737, 0.00001253, -0.00001200, 0.00002169, -0.00002795, 0.00001395, -0.00001833, 0.00002142, -0.00001934, 0.00002548, -0.00002812, 0.00002643, -0.00002610, 0.00002365, -0.00002718, 0.00003306, -0.00003149, 0.00001750, -0.00001705, 0.00002999, -0.00003882, 0.00002399, -0.00002893, 0.00004191, -0.00004200, 0.00003848, -0.00004453, 0.00004293, -0.00004290, 0.00004633, -0.00007520, 0.00009347, -0.00010337, 0.00009491, -0.00010670, 0.00010491, -0.00011476, 0.00010418, -0.00015841, 0.00017036, -0.00021010, 0.00019512, -0.00034753, 0.00033176, -0.00086814, 0.00038602, 0.00181430, -0.00129968, 0.00036296, -0.00035687, 0.00021131, -0.00022090, 0.00014034, -0.00013166, 0.00011760, -0.00012696, 0.00009282, -0.00008253, 0.00008357, -0.00009524, 0.00008755, -0.00008675, 0.00007952, -0.00008004, 0.00005865, -0.00005555, 0.00004931, -0.00005378, 0.00004886, -0.00003405, 0.00003033, -0.00004974, 0.00004479, -0.00002755, 0.00001614, -0.00002601, 0.00002583, -0.00001867, 0.00002128, -0.00002507, 0.00002051, -0.00002091, 0.00002557, -0.00002083, 0.00001898, -0.00001375, 0.00000749, -0.00002474, 0.00002520, -0.00001772, 0.00001906, -0.00001371, 0.00001113, -0.00001152, 0.00000968, -0.00001257, 0.00001142, -0.00000898, 0.00000490, -0.00001250, 0.00001144, -0.00000244, 0.00000456, 0.00000167, -0.00000441, -0.00000468, 0.00000615, -0.00000195, -0.00000009, -0.00000091, },
		{ -0.00000242, -0.00000028, 0.00000064, 0.00000431, -0.00001284, 0.00000532, 0.00000185, 0.00000285, 0.00000128, 0.00000331, -0.00000913, 0.00000483, -0.00000633, 0.00000739, -0.00000814, 0.00000571, -0.00001089, 0.00000974, -0.00000594, 0.00001057, -0.00000935, 0.00001866, -0.00002703, 0.00001443, -0.00001795, 0.00002047, -0.00001780, 0.00002393, -0.00002760, 0.00002633, -0.00002633, 0.00002436, -0.00002785, 0.00003359, -0.00003326, 0.00001954, -0.00001636, 0.00002729, -0.00003766, 0.00002421, -0.00002735, 0.00003880, -0.00003904, 0.00003580, -0.00004147, 0.00003899, -0.00003653, 0.00003799, -0.00006561, 0.00008408, -0.00009654, 0.00009084, -0.00010450, 0.00010258, -0.00011248, 0.00010153, -0.00015617, 0.00016685, -0.00020595, 0.00019000, -0.00034653, 0.00033009, -0.00091758, 0.00058702, 0.00158720, -0.00120347, 0.00032890, -0.00033179, 0.00019564, -0.00020740, 0.00012971, -0.00012245, 0.00010997, -0.00011838, 0.00008403, -0.00007222, 0.00007255, -0.00008413, 0.00007838, -0.00008170, 0.00007736, -0.00007927, 0.00005771, -0.00005467, 0.00004813, -0.00005266, 0.00004920, -0.00003526, 0.00002944, -0.00004800, 0.00004575, -0.00003014, 0.00001669, -0.00002472, 0.00002502, -0.00001799, 0.00002007, -0.00002339, 0.00001837, -0.00001840, 0.00002390, -0.00001985, 0.00001822, -0.00001322, 0.00000515, -0.00002122, 0.00002331, -0.00001696, 0.00001886, -0.00001395, 0.00001089, -0.00001122, 0.00000939, -0.00001204, 0.00001116, -0.00000903, 0.00000435, -0.00001193, 0.00001234, -0.00000370, 0.00000573, 0.00000064, -0.00000523, -0.00000305, 0.00000572, -0.00000219, -0.00000005, -0.00000083, },
		{ -0.00000249, -0.00000019, 0.00000120, 0.00000309, -0.00001299, 0.00000710, 0.00000068, 0.00000361, 0.00000108, 0.00000217, -0.00000832, 0.00000462, -0.00000575, 0.00000656, -0.00000744, 0.00000494, -0.00001003, 0.00000891, -0.00000431, 0.00000831, -0.00000619, 0.00001466, -0.00002485, 0.00001382, -0.00001668, 0.00001863, -0.00001525, 0.00002112, -0.00002564, 0.00002478, -0.00002513, 0.00002369, -0.00002717, 0.00003274, -0.00003376, 0.00002081, -0.00001520, 0.00002377, -0.00003530, 0.00002342, -0.00002494, 0.00003466, -0.00003481, 0.00003180, -0.00003714, 0.00003389, -0.00002891, 0.00002786, -0.00005329, 0.00007073, -0.00008483, 0.00008142, -0.00009682, 0.00009469, -0.00010437, 0.00009256, -0.00014683, 0.00015523, -0.00019225, 0.00017296, -0.00032995, 0.00030558, -0.00093350, 0.00075874, 0.00136462, -0.00109502, 0.00028269, -0.00029729, 0.00017213, -0.00018741, 0.00011367, -0.00010860, 0.00009833, -0.00010635, 0.00007228, -0.00005913, 0.00005852, -0.00006948, 0.00006496, -0.00007212, 0.00007093, -0.00007475, 0.00005356, -0.00005097, 0.00004433, -0.00004896, 0.00004713, -0.00003456, 0.00002689, -0.00004429, 0.00004469, -0.00003136, 0.00001645, -0.00002267, 0.00002341, -0.00001665, 0.00001827, -0.00002113, 0.00001562, -0.00001514, 0.00002134, -0.00001803, 0.00001676, -0.00001227, 0.00000249, -0.00001699, 0.00002037, -0.00001513, 0.00001770, -0.00001343, 0.00001004, -0.00001036, 0.00000857, -0.00001096, 0.00001036, -0.00000864, 0.00000339, -0.00001079, 0.00001265, -0.00000461, 0.00000680, -0.00000057, -0.00000574, -0.00000147, 0.00000512, -0.00000233, 0.00000002, -0.00000083, },
		{ -0.00000247, -0.00000015, 0.00000182, 0.00000155, -0.00001255, 0.00000846, -0.00000037, 0.00000440, 0.00000072, 0.00000111, -0.00000737, 0.00000425, -0.00000506, 0.00000559, -0.00000658, 0.00000401, -0.00000900, 0.00000797, -0.00000266, 0.00000600, -0.00000281, 0.00001002, -0.00002164, 0.00001229, -0.00001470, 0.00001610, -0.00001191, 0.00001729, -0.00002247, 0.00002195, -0.00002261, 0.00002173, -0.00002525, 0.00003058, -0.00003300, 0.00002132, -0.00001370, 0.00001966, -0.00003198, 0.00002184, -0.00002203, 0.00002990, -0.00002976, 0.00002697, -0.00003208, 0.00002828, -0.00002089, 0.00001701, -0.00003949, 0.00005475, -0.00006944, 0.00006772, -0.00008461, 0.00008216, -0.00009143, 0.00007832, -0.00013155, 0.00013680, -0.00017048, 0.00014566, -0.00029960, 0.00025969, -0.00091499, 0.00089589, 0.00115447, -0.00098114, 0.00022901, -0.00025674, 0.00014339, -0.00016306, 0.00009402, -0.00009170, 0.00008413, -0.00009221, 0.00005890, -0.00004466, 0.00004294, -0.00005277, 0.00004863, -0.00005912, 0.00006103, -0.00006709, 0.00004671, -0.00004490, 0.00003833, -0.00004305, 0.00004294, -0.00003220, 0.00002297, -0.00003889, 0.00004175, -0.00003123, 0.00001549, -0.00002000, 0.00002115, -0.00001481, 0.00001608, -0.00001853, 0.00001254, -0.00001140, 0.00001816, -0.00001559, 0.00001481, -0.00001110, -0.00000021, -0.00001240, 0.00001668, -0.00001248, 0.00001576, -0.00001228, 0.00000870, -0.00000905, 0.00000732, -0.00000945, 0.00000911, -0.00000787, 0.00000208, -0.00000913, 0.00001238, -0.00000511, 0.00000768, -0.00000191, -0.00000593, 0.00000001, 0.00000441, -0.00000240, 0.00000011, -0.00000089, },
		{ -0.00000237, -0.00000014, 0.00000245, -0.00000024, -0.00001153, 0.00000933, -0.00000120, 0.00000514, 0.00000022, 0.00000015, -0.00000634, 0.00000378, -0.00000433, 0.00000456, -0.00000564, 0.00000302, -0.00000792, 0.00000706, -0.00000115, 0.00000388, 0.00000044, 0.00000511, -0.00001774, 0.00001009, -0.00001224, 0.00001315, -0.00000811, 0.00001277, -0.00001839, 0.00001810, -0.00001902, 0.00001869, -0.00002225, 0.00002725, -0.00003106, 0.00002110, -0.00001199, 0.00001523, -0.00002798, 0.00001973, -0.00001892, 0.00002491, -0.00002434, 0.00002177, -0.00002681, 0.00002279, -0.00001330, 0.00000652, -0.00002554, 0.00003760, -0.00005185, 0.00005113, -0.00006917, 0.00006628, -0.00007497, 0.00006025, -0.00011189, 0.00011334, -0.00014272, 0.00011052, -0.00025829, 0.00019535, -0.00086311, 0.00099521, 0.00096328, -0.00086809, 0.00017243, -0.00021353, 0.00011211, -0.00013656, 0.00007262, -0.00007339, 0.00006884, -0.00007733, 0.00004525, -0.00003020, 0.00002732, -0.00003557, 0.00003091, -0.00004400, 0.00004874, -0.00005715, 0.00003787, -0.00003708, 0.00003071, -0.00003544, 0.00003705, -0.00002853, 0.00001807, -0.00003219, 0.00003719, -0.00002989, 0.00001392, -0.00001687, 0.00001842, -0.00001264, 0.00001372, -0.00001583, 0.00000941, -0.00000750, 0.00001462, -0.00001278, 0.00001256, -0.00000991, -0.00000268, -0.00000780, 0.00001259, -0.00000928, 0.00001323, -0.00001065, 0.00000701, -0.00000742, 0.00000581, -0.00000764, 0.00000751, -0.00000684, 0.00000055, -0.00000705, 0.00001155, -0.00000517, 0.00000829, -0.00000331, -0.00000579, 0.00000131, 0.00000362, -0.00000240, 0.00000021, -0.00000102, },
		{ -0.00000222, -0.00000014, 0.00000303, -0.00000220, -0.00000999, 0.00000968, -0.00000174, 0.00000576, -0.00000037, -0.00000066, -0.00000528, 0.00000326, -0.00000362, 0.00000354, -0.00000471, 0.00000207, -0.00000689, 0.00000628, 0.00000007, 0.00000216, 0.00000328, 0.00000029, -0.00001349, 0.00000749, -0.00000955, 0.00001007, -0.00000417, 0.00000792, -0.00001375, 0.00001357, -0.00001466, 0.00001483, -0.00001843, 0.00002296, -0.00002808, 0.00002023, -0.00001019, 0.00001072, -0.00002358, 0.00001734, -0.00001589, 0.00002006, -0.00001896, 0.00001664, -0.00002178, 0.00001796, -0.00000685, -0.00000263, -0.00001270, 0.00002078, -0.00003362, 0.00003316, -0.00005197, 0.00004851, -0.00005654, 0.00004002, -0.00008973, 0.00008695, -0.00011143, 0.00007051, -0.00020968, 0.00011678, -0.00078084, 0.00105555, 0.00079595, -0.00076122, 0.00011706, -0.00017080, 0.00008084, -0.00011000, 0.00005127, -0.00005521, 0.00005384, -0.00006299, 0.00003254, -0.00001703, 0.00001301, -0.00001939, 0.00001333, -0.00002815, 0.00003523, -0.00004589, 0.00002786, -0.00002825, 0.00002217, -0.00002678, 0.00002998, -0.00002399, 0.00001263, -0.00002465, 0.00003135, -0.00002751, 0.00001190, -0.00001347, 0.00001542, -0.00001034, 0.00001136, -0.00001325, 0.00000648, -0.00000371, 0.00001100, -0.00000982, 0.00001020, -0.00000888, -0.00000467, -0.00000353, 0.00000844, -0.00000581, 0.00001036, -0.00000873, 0.00000514, -0.00000566, 0.00000417, -0.00000569, 0.00000571, -0.00000565, -0.00000108, -0.00000468, 0.00001024, -0.00000479, 0.00000858, -0.00000469, -0.00000535, 0.00000241, 0.00000279, -0.00000233, 0.00000032, -0.00000119, },
		{ -0.00000202, -0.00000012, 0.00000349, -0.00000420, -0.00000801, 0.00000952, -0.00000195, 0.00000620, -0.00000103, -0.00000132, -0.00000424, 0.00000273, -0.00000297, 0.00000260, -0.00000386, 0.00000123, -0.00000598, 0.00000571, 0.00000090, 0.00000100, 0.00000545, -0.00000408, -0.00000922, 0.00000478, -0.00000689, 0.00000713, -0.00000043, 0.00000313, -0.00000895, 0.00000874, -0.00000988, 0.00001049, -0.00001409, 0.00001799, -0.00002424, 0.00001881, -0.00000844, 0.00000636, -0.00001903, 0.00001487, -0.00001317, 0.00001566, -0.00001398, 0.00001194, -0.00001735, 0.00001419, -0.00000209, -0.00000966, -0.00000207, 0.00000565, -0.00001626, 0.00001535, -0.00003452, 0.00003040, -0.00003777, 0.00001940, -0.00006702, 0.00005988, -0.00007927, 0.00002890, -0.00015790, 0.00002919, -0.00067281, 0.00107783, 0.00065553, -0.00066466, 0.00006629, -0.00013124, 0.00005175, -0.00008520, 0.00003150, -0.00003849, 0.00004029, -0.00005023, 0.00002178, -0.00000618, 0.00000117, -0.00000555, -0.00000268, -0.00001295, 0.00002172, -0.00003433, 0.00001756, -0.00001921, 0.00001343, -0.00001774, 0.00002232, -0.00001907, 0.00000717, -0.00001680, 0.00002466, -0.00002436, 0.00000961, -0.00000999, 0.00001232, -0.00000805, 0.00000917, -0.00001097, 0.00000395, -0.00000028, 0.00000756, -0.00000692, 0.00000790, -0.00000814, -0.00000599, 0.00000014, 0.00000454, -0.00000238, 0.00000739, -0.00000672, 0.00000327, -0.00000391, 0.00000258, -0.00000376, 0.00000386, -0.00000445, -0.00000266, -0.00000218, 0.00000856, -0.00000400, 0.00000851, -0.00000598, -0.00000465, 0.00000328, 0.00000197, -0.00000220, 0.00000042, -0.00000139, },
		{ -0.00000181, -0.00000004, 0.00000375, -0.00000611, -0.00000572, 0.00000889, -0.00000181, 0.00000642, -0.00000170, -0.00000182, -0.00000326, 0.00000221, -0.00000240, 0.00000176, -0.00000311, 0.00000054, -0.00000524, 0.00000540, 0.00000129, 0.00000049, 0.00000676, -0.00000773, -0.00000526, 0.00000223, -0.00000452, 0.00000461, 0.00000280, -0.00000124, -0.00000437, 0.00000400, -0.00000505, 0.00000601, -0.00000957, 0.00001265, -0.00001978, 0.00001697, -0.00000684, 0.00000235, -0.00001455, 0.00001250, -0.00001092, 0.00001192, -0.00000965, 0.00000792, -0.00001375, 0.00001172, 0.00000064, -0.00001403, 0.00000553, -0.00000664, -0.00000112, -0.00000087, -0.00001828, 0.00001340, -0.00002020, 0.00000007, -0.00004564, 0.00003429, -0.00004884, -0.00001107, -0.00010730, -0.00006160, -0.00054502, 0.00106490, 0.00054326, -0.00058124, 0.00002267, -0.00009690, 0.00002656, -0.00006358, 0.00001451, -0.00002426, 0.00002906, -0.00003980, 0.00001364, 0.00000166, -0.00000739, 0.00000493, -0.00001594, 0.00000042, 0.00000929, -0.00002345, 0.00000780, -0.00001069, 0.00000520, -0.00000900, 0.00001464, -0.00001424, 0.00000215, -0.00000917, 0.00001756, -0.00002071, 0.00000724, -0.00000662, 0.00000930, -0.00000591, 0.00000727, -0.00000909, 0.00000199, 0.00000259, 0.00000449, -0.00000428, 0.00000577, -0.00000777, -0.00000652, 0.00000298, 0.00000118, 0.00000075, 0.00000454, -0.00000479, 0.00000154, -0.00000234, 0.00000119, -0.00000201, 0.00000210, -0.00000337, -0.00000404, 0.00000029, 0.00000664, -0.00000287, 0.00000807, -0.00000710, -0.00000376, 0.00000390, 0.00000119, -0.00000202, 0.00000048, -0.00000158, },
		{ -0.00000161, 0.00000010, 0.00000376, -0.00000781, -0.00000325, 0.00000788, -0.00000135, 0.00000638, -0.00000234, -0.00000218, -0.00000234, 0.00000173, -0.00000192, 0.00000104, -0.00000250, 0.00000001, -0.00000467, 0.00000533, 0.00000125, 0.00000066, 0.00000710, -0.00001043, -0.00000187, 0.00000007, -0.00000261, 0.00000270, 0.00000527, -0.00000488, -0.00000037, -0.00000030, -0.00000054, 0.00000175, -0.00000521, 0.00000728, -0.00001495, 0.00001482, -0.00000548, -0.00000118, -0.00001031, 0.00001034, -0.00000925, 0.00000898, -0.00000613, 0.00000472, -0.00001107, 0.00001057, 0.00000128, -0.00001552, 0.00000961, -0.00001531, 0.00001077, -0.00001433, -0.00000445, -0.00000122, -0.00000517, -0.00001650, -0.00002723, 0.00001208, -0.00002247, -0.00004640, -0.00006200, -0.00014961, -0.00040433, 0.00102125, 0.00045864, -0.00051243, -0.00001227, -0.00006910, 0.00000638, -0.00004606, 0.00000107, -0.00001313, 0.00002067, -0.00003209, 0.00000845, 0.00000615, -0.00001226, 0.00001144, -0.00002561, 0.00001100, -0.00000113, -0.00001406, -0.00000071, -0.00000334, -0.00000189, -0.00000115, 0.00000747, -0.00000994, -0.00000199, -0.00000225, 0.00001051, -0.00001687, 0.00000499, -0.00000353, 0.00000649, -0.00000402, 0.00000573, -0.00000771, 0.00000068, 0.00000478, 0.00000195, -0.00000201, 0.00000390, -0.00000779, -0.00000622, 0.00000485, -0.00000144, 0.00000338, 0.00000199, -0.00000310, 0.00000010, -0.00000105, 0.00000011, -0.00000057, 0.00000055, -0.00000250, -0.00000510, 0.00000257, 0.00000463, -0.00000148, 0.00000728, -0.00000800, -0.00000271, 0.00000429, 0.00000046, -0.00000179, 0.00000050, -0.00000174, },
		{ -0.00000144, 0.00000035, 0.00000346, -0.00000918, -0.00000076, 0.00000658, -0.00000060, 0.00000609, -0.00000292, -0.00000241, -0.00000152, 0.00000129, -0.00000155, 0.00000045, -0.00000202, -0.00000035, -0.00000426, 0.00000546, 0.00000084, 0.00000145, 0.00000647, -0.00001208, 0.00000078, -0.00000156, -0.00000131, 0.00000155, 0.00000679, -0.00000755, 0.00000277, -0.00000383, 0.00000333, -0.00000197, -0.00000134, 0.00000220, -0.00001002, 0.00001251, -0.00000443, -0.00000410, -0.00000644, 0.00000847, -0.00000817, 0.00000687, -0.00000347, 0.00000236, -0.00000925, 0.00001062, -0.00000001, -0.00001421, 0.00001005, -0.00001993, 0.00001872, -0.00002420, 0.00000606, -0.00001251, 0.00000631, -0.00002919, -0.00001306, -0.00000528, -0.00000199, -0.00007464, -0.00002561, -0.00022911, -0.00025810, 0.00095254, 0.00039967, -0.00045843, -0.00003791, -0.00004842, -0.00000825, -0.00003310, -0.00000846, -0.00000538, 0.00001527, -0.00002717, 0.00000618, 0.00000735, -0.00001343, 0.00001377, -0.00003124, 0.00001819, -0.00000888, -0.00000678, -0.00000740, 0.00000234, -0.00000736, 0.00000532, 0.00000124, -0.00000650, -0.00000491, 0.00000354, 0.00000391, -0.00001311, 0.00000303, -0.00000085, 0.00000401, -0.00000246, 0.00000458, -0.00000683, 0.00000005, 0.00000622, 0.00000001, -0.00000020, 0.00000232, -0.00000816, -0.00000514, 0.00000572, -0.00000321, 0.00000534, -0.00000010, -0.00000175, -0.00000099, -0.00000012, -0.00000057, 0.00000048, -0.00000070, -0.00000194, -0.00000573, 0.00000452, 0.00000268, 0.00000006, 0.00000616, -0.00000865, -0.00000159, 0.00000445, -0.00000020, -0.00000151, 0.00000045, -0.00000184, },
		{ -0.00000133, 0.00000069, 0.00000284, -0.00001010, 0.00000160, 0.00000512, 0.00000036, 0.00000555, -0.00000342, -0.00000251, -0.00000078, 0.00000089, -0.00000126, -0.00000002, -0.00000165, -0.00000057, -0.00000398, 0.00000572, 0.00000017, 0.00000274, 0.00000496, -0.00001264, 0.00000259, -0.00000255, -0.00000067, 0.00000119, 0.00000730, -0.00000911, 0.00000486, -0.00000639, 0.00000631, -0.00000491, 0.00000178, -0.00000229, -0.00000524, 0.00001015, -0.00000372, -0.00000636, -0.00000302, 0.00000689, -0.00000765, 0.00000554, -0.00000162, 0.00000077, -0.00000813, 0.00001157, -0.00000285, -0.00001049, 0.00000710, -0.00002050, 0.00002246, -0.00003004, 0.00001273, -0.00001994, 0.00001365, -0.00003734, -0.00000390, -0.00001686, 0.00001140, -0.00009408, -0.00000093, -0.00029505, -0.00011358, 0.00086525, 0.00036314, -0.00041831, -0.00005452, -0.00003476, -0.00001738, -0.00002466, -0.00001417, -0.00000088, 0.00001269, -0.00002477, 0.00000649, 0.00000565, -0.00001128, 0.00001214, -0.00003276, 0.00002173, -0.00001362, -0.00000196, -0.00001196, 0.00000604, -0.00001091, 0.00001007, -0.00000375, -0.00000414, -0.00000641, 0.00000789, -0.00000189, -0.00000966, 0.00000148, 0.00000130, 0.00000194, -0.00000124, 0.00000382, -0.00000642, 0.00000007, 0.00000692, -0.00000128, 0.00000113, 0.00000101, -0.00000878, -0.00000340, 0.00000564, -0.00000410, 0.00000658, -0.00000166, -0.00000081, -0.00000167, 0.00000041, -0.00000083, 0.00000109, -0.00000158, -0.00000172, -0.00000588, 0.00000601, 0.00000091, 0.00000163, 0.00000479, -0.00000900, -0.00000043, 0.00000440, -0.00000077, -0.00000121, 0.00000034, -0.00000188, },
		{ -0.00000126, 0.00000111, 0.00000189, -0.00001051, 0.00000369, 0.00000362, 0.00000145, 0.00000481, -0.00000381, -0.00000252, -0.00000015, 0.00000054, -0.00000105, -0.00000041, -0.00000136, -0.00000070, -0.00000376, 0.00000602, -0.00000061, 0.00000434, 0.00000271, -0.00001219, 0.00000355, -0.00000290, -0.00000067, 0.00000160, 0.00000681, -0.00000953, 0.00000580, -0.00000782, 0.00000825, -0.00000687, 0.00000393, -0.00000595, -0.00000083, 0.00000788, -0.00000337, -0.00000794, -0.00000009, 0.00000558, -0.00000757, 0.00000489, -0.00000047, -0.00000021, -0.00000747, 0.00001303, -0.00000670, -0.00000495, 0.00000131, -0.00001738, 0.00002216, -0.00003181, 0.00001545, -0.00002333, 0.00001666, -0.00004073, -0.00000001, -0.00002233, 0.00001721, -0.00010388, 0.00001029, -0.00034340, 0.00002249, 0.00076613, 0.00034502, -0.00039028, -0.00006309, -0.00002742, -0.00002152, -0.00002028, -0.00001645, 0.00000078, 0.00001248, -0.00002438, 0.00000876, 0.00000174, -0.00000651, 0.00000713, -0.00003053, 0.00002173, -0.00001531, 0.00000028, -0.00001428, 0.00000769, -0.00001243, 0.00001294, -0.00000733, -0.00000295, -0.00000639, 0.00001060, -0.00000664, -0.00000670, 0.00000045, 0.00000286, 0.00000031, -0.00000038, 0.00000339, -0.00000639, 0.00000064, 0.00000692, -0.00000196, 0.00000199, -0.00000007, -0.00000954, -0.00000117, 0.00000472, -0.00000415, 0.00000707, -0.00000267, -0.00000028, -0.00000196, 0.00000055, -0.00000066, 0.00000124, -0.00000210, -0.00000185, -0.00000553, 0.00000697, -0.00000053, 0.00000311, 0.00000324, -0.00000907, 0.00000070, 0.00000417, -0.00000123, -0.00000088, 0.00000017, -0.00000185, },
		{ -0.00000124, 0.00000159, 0.00000064, -0.00001037, 0.00000540, 0.00000222, 0.00000257, 0.00000391, -0.00000411, -0.00000243, 0.00000039, 0.00000023, -0.00000090, -0.00000072, -0.00000112, -0.00000078, -0.00000353, 0.00000627, -0.00000137, 0.00000605, -0.00000003, -0.00001088, 0.00000375, -0.00000268, -0.00000122, 0.00000266, 0.00000543, -0.00000887, 0.00000563, -0.00000812, 0.00000907, -0.00000778, 0.00000500, -0.00000860, 0.00000300, 0.00000577, -0.00000337, -0.00000884, 0.00000234, 0.00000450, -0.00000783, 0.00000476, 0.00000014, -0.00000077, -0.00000698, 0.00001453, -0.00001092, 0.00000160, -0.00000649, -0.00001129, 0.00001831, -0.00002988, 0.00001447, -0.00002296, 0.00001561, -0.00003964, -0.00000112, -0.00002199, 0.00001569, -0.00010412, 0.00000747, -0.00037142, 0.00014429, 0.00066177, 0.00034084, -0.00037195, -0.00006515, -0.00002527, -0.00002157, -0.00001922, -0.00001598, 0.00000024, 0.00001399, -0.00002530, 0.00001223, -0.00000353, -0.00000007, -0.00000037, -0.00002521, 0.00001863, -0.00001419, 0.00000012, -0.00001448, 0.00000739, -0.00001200, 0.00001395, -0.00000950, -0.00000287, -0.00000493, 0.00001164, -0.00001018, -0.00000437, -0.00000002, 0.00000381, -0.00000085, 0.00000017, 0.00000323, -0.00000663, 0.00000164, 0.00000635, -0.00000210, 0.00000243, -0.00000099, -0.00001027, 0.00000134, 0.00000316, -0.00000348, 0.00000689, -0.00000315, -0.00000013, -0.00000190, 0.00000037, -0.00000014, 0.00000099, -0.00000227, -0.00000230, -0.00000468, 0.00000737, -0.00000158, 0.00000439, 0.00000159, -0.00000886, 0.00000176, 0.00000379, -0.00000159, -0.00000056, -0.00000005, -0.00000174, },
		{ -0.00000125, 0.00000209, -0.00000086, -0.00000968, 0.00000665, 0.00000103, 0.00000363, 0.00000291, -0.00000430, -0.00000227, 0.00000083, -0.00000004, -0.00000079, -0.00000099, -0.00000090, -0.00000086, -0.00000325, 0.00000635, -0.00000193, 0.00000765, -0.00000302, -0.00000893, 0.00000331, -0.00000200, -0.00000216, 0.00000418, 0.00000336, -0.00000730, 0.00000446, -0.00000736, 0.00000882, -0.00000763, 0.00000495, -0.00001014, 0.00000613, 0.00000392, -0.00000369, -0.00000911, 0.00000429, 0.00000358, -0.00000828, 0.00000499, 0.00000037, -0.00000109, -0.00000638, 0.00001565, -0.00001485, 0.00000832, -0.00001533, -0.00000317, 0.00001174, -0.00002492, 0.00001042, -0.00001942, 0.00001114, -0.00003472, -0.00000649, -0.00001667, 0.00000779, -0.00009577, -0.00000876, -0.00037786, 0.00024728, 0.00055822, 0.00034603, -0.00036064, -0.00006251, -0.00002691, -0.00001865, -0.00002054, -0.00001360, -0.00000174, 0.00001652, -0.00002679, 0.00001605, -0.00000921, 0.00000700, -0.00000931, -0.00001770, 0.00001311, -0.00001079, -0.00000207, -0.00001290, 0.00000544, -0.00000986, 0.00001326, -0.00001038, -0.00000375, -0.00000222, 0.00001107, -0.00001248, -0.00000269, 0.00000005, 0.00000417, -0.00000157, 0.00000044, 0.00000325, -0.00000702, 0.00000290, 0.00000533, -0.00000179, 0.00000252, -0.00000182, -0.00001083, 0.00000389, 0.00000119, -0.00000226, 0.00000616, -0.00000320, -0.00000029, -0.00000157, -0.00000005, 0.00000066, 0.00000040, -0.00000216, -0.00000298, -0.00000343, 0.00000722, -0.00000217, 0.00000539, -0.00000008, -0.00000838, 0.00000271, 0.00000329, -0.00000183, -0.00000027, -0.00000030, -0.00000158, },
		{ -0.00000127, 0.00000254, -0.00000251, -0.00000847, 0.00000738, 0.00000011, 0.00000453, 0.00000188, -0.00000442, -0.00000204, 0.00000116, -0.00000027, -0.00000072, -0.00000123, -0.00000067, -0.00000097, -0.00000287, 0.00000621, -0.00000218, 0.00000891, -0.00000597, -0.00000660, 0.00000243, -0.00000103, -0.00000332, 0.00000593, 0.00000084, -0.00000507, 0.00000250, -0.00000570, 0.00000764, -0.00000652, 0.00000385, -0.00001055, 0.00000845, 0.00000235, -0.00000428, -0.00000882, 0.00000578, 0.00000276, -0.00000879, 0.00000540, 0.00000042, -0.00000136, -0.00000545, 0.00001597, -0.00001788, 0.00001435, -0.00002417, 0.00000591, 0.00000345, -0.00001783, 0.00000414, -0.00001357, 0.00000417, -0.00002700, -0.00001501, -0.00000764, -0.00000501, -0.00008060, -0.00003665, -0.00036292, 0.00032839, 0.00046058, 0.00035632, -0.00035361, -0.00005709, -0.00003083, -0.00001396, -0.00002322, -0.00001016, -0.00000439, 0.00001932, -0.00002812, 0.00001942, -0.00001440, 0.00001363, -0.00001853, -0.00000906, 0.00000603, -0.00000577, -0.00000574, -0.00001000, 0.00000228, -0.00000642, 0.00001118, -0.00001020, -0.00000529, 0.00000144, 0.00000910, -0.00001359, -0.00000164, 0.00000059, 0.00000400, -0.00000191, 0.00000051, 0.00000335, -0.00000741, 0.00000428, 0.00000403, -0.00000116, 0.00000233, -0.00000265, -0.00001107, 0.00000625, -0.00000096, -0.00000069, 0.00000501, -0.00000290, -0.00000066, -0.00000107, -0.00000060, 0.00000161, -0.00000042, -0.00000187, -0.00000380, -0.00000186, 0.00000657, -0.00000229, 0.00000604, -0.00000168, -0.00000769, 0.00000352, 0.00000269, -0.00000195, -0.00000001, -0.00000056, -0.00000138, },
		{ -0.00000126, 0.00000288, -0.00000422, -0.00000682, 0.00000759, -0.00000046, 0.00000521, 0.00000088, -0.00000447, -0.00000174, 0.00000139, -0.00000046, -0.00000068, -0.00000144, -0.00000042, -0.00000113, -0.00000237, 0.00000581, -0.00000201, 0.00000967, -0.00000864, -0.00000413, 0.00000131, 0.00000005, -0.00000448, 0.00000766, -0.00000182, -0.00000246, 0.00000001, -0.00000339, 0.00000571, -0.00000465, 0.00000183, -0.00000990, 0.00000996, 0.00000110, -0.00000509, -0.00000804, 0.00000686, 0.00000199, -0.00000924, 0.00000584, 0.00000041, -0.00000171, -0.00000402, 0.00001525, -0.00001952, 0.00001896, -0.00003205, 0.00001483, -0.00000546, -0.00000967, -0.00000336, -0.00000645, -0.00000422, -0.00001764, -0.00002535, 0.00000351, -0.00002079, -0.00006095, -0.00007348, -0.00032825, 0.00038610, 0.00037285, 0.00036799, -0.00034837, -0.00005071, -0.00003561, -0.00000868, -0.00002630, -0.00000649, -0.00000699, 0.00002176, -0.00002865, 0.00002167, -0.00001833, 0.00001888, -0.00002695, -0.00000034, -0.00000168, 0.00000010, -0.00001025, -0.00000634, -0.00000157, -0.00000214, 0.00000809, -0.00000928, -0.00000718, 0.00000567, 0.00000603, -0.00001366, -0.00000113, 0.00000152, 0.00000340, -0.00000193, 0.00000043, 0.00000344, -0.00000768, 0.00000560, 0.00000259, -0.00000033, 0.00000196, -0.00000351, -0.00001090, 0.00000823, -0.00000304, 0.00000101, 0.00000361, -0.00000240, -0.00000113, -0.00000052, -0.00000116, 0.00000258, -0.00000135, -0.00000148, -0.00000463, -0.00000011, 0.00000551, -0.00000196, 0.00000631, -0.00000314, -0.00000681, 0.00000417, 0.00000203, -0.00000195, 0.00000018, -0.00000080, -0.00000116, },
		{ -0.00000119, 0.00000306, -0.00000586, -0.00000484, 0.00000732, -0.00000066, 0.00000561, -0.00000003, -0.00000447, -0.00000139, 0.00000152, -0.00000059, -0.00000067, -0.00000163, -0.00000015, -0.00000134, -0.00000175, 0.00000513, -0.00000139, 0.00000979, -0.00001079, -0.00000178, 0.00000016, 0.00000105, -0.00000545, 0.00000913, -0.00000435, 0.00000022, -0.00000269, -0.00000071, 0.00000332, -0.00000225, -0.00000088, -0.00000831, 0.00001066, 0.00000016, -0.00000604, -0.00000686, 0.00000757, 0.00000123, -0.00000954, 0.00000620, 0.00000044, -0.00000220, -0.00000205, 0.00001333, -0.00001947, 0.00002157, -0.00003812, 0.00002260, -0.00001391, -0.00000148, -0.00001106, 0.00000088, -0.00001287, -0.00000793, -0.00003607, 0.00001510, -0.00003747, -0.00003953, -0.00011582, -0.00027676, 0.00042044, 0.00029775, 0.00037806, -0.00034281, -0.00004490, -0.00003999, -0.00000379, -0.00002897, -0.00000326, -0.00000898, 0.00002337, -0.00002795, 0.00002232, -0.00002044, 0.00002201, -0.00003364, 0.00000747, -0.00000910, 0.00000602, -0.00001493, -0.00000252, -0.00000556, 0.00000244, 0.00000444, -0.00000795, -0.00000905, 0.00001004, 0.00000223, -0.00001290, -0.00000104, 0.00000270, 0.00000248, -0.00000170, 0.00000028, 0.00000346, -0.00000774, 0.00000672, 0.00000117, 0.00000058, 0.00000146, -0.00000443, -0.00001025, 0.00000967, -0.00000485, 0.00000266, 0.00000214, -0.00000181, -0.00000160, -0.00000001, -0.00000162, 0.00000347, -0.00000227, -0.00000112, -0.00000537, 0.00000166, 0.00000418, -0.00000126, 0.00000617, -0.00000440, -0.00000579, 0.00000464, 0.00000135, -0.00000184, 0.00000029, -0.00000100, -0.00000096, },
		{ -0.00000105, 0.00000301, -0.00000733, -0.00000266, 0.00000665, -0.00000053, 0.00000571, -0.00000082, -0.00000445, -0.00000099, 0.00000155, -0.00000067, -0.00000070, -0.00000178, 0.00000013, -0.00000157, -0.00000104, 0.00000422, -0.00000033, 0.00000921, -0.00001226, 0.00000023, -0.00000083, 0.00000183, -0.00000606, 0.00001013, -0.00000646, 0.00000266, -0.00000530, 0.00000202, 0.00000075, 0.00000040, -0.00000402, -0.00000598, 0.00001065, -0.00000052, -0.00000705, -0.00000537, 0.00000793, 0.00000046, -0.00000965, 0.00000642, 0.00000057, -0.00000283, 0.00000041, 0.00001024, -0.00001761, 0.00002187, -0.00004179, 0.00002838, -0.00002098, 0.00000579, -0.00001799, 0.00000744, -0.00002071, 0.00000095, -0.00004583, 0.00002551, -0.00005297, -0.00001914, -0.00015981, -0.00021231, 0.00043281, 0.00023671, 0.00038440, -0.00033536, -0.00004085, -0.00004304, -0.00000005, -0.00003063, -0.00000096, -0.00000999, 0.00002386, -0.00002579, 0.00002115, -0.00002045, 0.00002256, -0.00003792, 0.00001357, -0.00001545, 0.00001130, -0.00001916, 0.00000092, -0.00000916, 0.00000682, 0.00000064, -0.00000655, -0.00001055, 0.00001414, -0.00000191, -0.00001157, -0.00000119, 0.00000397, 0.00000136, -0.00000132, 0.00000010, 0.00000336, -0.00000751, 0.00000754, -0.00000011, 0.00000148, 0.00000089, -0.00000542, -0.00000911, 0.00001047, -0.00000623, 0.00000407, 0.00000073, -0.00000124, -0.00000198, 0.00000037, -0.00000189, 0.00000415, -0.00000309, -0.00000087, -0.00000588, 0.00000333, 0.00000270, -0.00000026, 0.00000566, -0.00000541, -0.00000469, 0.00000494, 0.00000067, -0.00000164, 0.00000031, -0.00000113, -0.00000080, },
		{ -0.00000081, 0.00000269, -0.00000850, -0.00000043, 0.00000566, -0.00000009, 0.00000551, -0.00000145, -0.00000442, -0.00000055, 0.00000148, -0.00000069, -0.00000078, -0.00000187, 0.00000039, -0.00000181, -0.00000029, 0.00000314, 0.00000111, 0.00000795, -0.00001296, 0.00000177, -0.00000155, 0.00000228, -0.00000620, 0.00001050, -0.00000796, 0.00000461, -0.00000753, 0.00000452, -0.00000170, 0.00000300, -0.00000730, -0.00000316, 0.00001004, -0.00000097, -0.00000804, -0.00000366, 0.00000800, -0.00000031, -0.00000955, 0.00000646, 0.00000080, -0.00000351, 0.00000320, 0.00000618, -0.00001406, 0.00001979, -0.00004273, 0.00003161, -0.00002594, 0.00001136, -0.00002337, 0.00001239, -0.00002683, 0.00000799, -0.00005347, 0.00003331, -0.00006546, -0.00000240, -0.00020153, -0.00013942, 0.00042582, 0.00018995, 0.00038577, -0.00032505, -0.00003930, -0.00004414, 0.00000207, -0.00003090, 0.00000016, -0.00000984, 0.00002316, -0.00002217, 0.00001817, -0.00001836, 0.00002039, -0.00003939, 0.00001741, -0.00002013, 0.00001535, -0.00002245, 0.00000353, -0.00001194, 0.00001056, -0.00000291, -0.00000539, -0.00001140, 0.00001757, -0.00000601, -0.00000993, -0.00000142, 0.00000517, 0.00000019, -0.00000085, -0.00000006, 0.00000310, -0.00000695, 0.00000798, -0.00000115, 0.00000227, 0.00000029, -0.00000643, -0.00000754, 0.00001061, -0.00000708, 0.00000512, -0.00000050, -0.00000078, -0.00000221, 0.00000056, -0.00000191, 0.00000455, -0.00000371, -0.00000082, -0.00000608, 0.00000475, 0.00000122, 0.00000092, 0.00000482, -0.00000615, -0.00000353, 0.00000505, 0.00000003, -0.00000135, 0.00000024, -0.00000118, -0.00000070, },
		{ -0.00000047, 0.00000207, -0.00000927, 0.00000171, 0.00000449, 0.00000058, 0.00000505, -0.00000192, -0.00000438, -0.00000009, 0.00000133, -0.00000064, -0.00000092, -0.00000190, 0.00000061, -0.00000201, 0.00000043, 0.00000199, 0.00000283, 0.00000607, -0.00001288, 0.00000275, -0.00000189, 0.00000233, -0.00000582, 0.00001016, -0.00000868, 0.00000587, -0.00000915, 0.00000653, -0.00000379, 0.00000529, -0.00001044, -0.00000008, 0.00000898, -0.00000126, -0.00000893, -0.00000184, 0.00000779, -0.00000105, -0.00000927, 0.00000634, 0.00000108, -0.00000414, 0.00000607, 0.00000146, -0.00000913, 0.00001551, -0.00004089, 0.00003201, -0.00002836, 0.00001473, -0.00002667, 0.00001516, -0.00003060, 0.00001247, -0.00005816, 0.00003749, -0.00007350, 0.00000848, -0.00023725, -0.00006289, 0.00040291, 0.00015668, 0.00038176, -0.00031147, -0.00004054, -0.00004305, 0.00000237, -0.00002967, 0.00000006, -0.00000858, 0.00002142, -0.00001732, 0.00001363, -0.00001444, 0.00001568, -0.00003798, 0.00001871, -0.00002277, 0.00001780, -0.00002443, 0.00000498, -0.00001356, 0.00001333, -0.00000591, -0.00000466, -0.00001138, 0.00002006, -0.00000973, -0.00000823, -0.00000156, 0.00000618, -0.00000093, -0.00000038, -0.00000017, 0.00000270, -0.00000605, 0.00000799, -0.00000188, 0.00000290, -0.00000033, -0.00000741, -0.00000560, 0.00001010, -0.00000736, 0.00000574, -0.00000146, -0.00000048, -0.00000225, 0.00000054, -0.00000165, 0.00000463, -0.00000408, -0.00000100, -0.00000591, 0.00000580, -0.00000012, 0.00000217, 0.00000372, -0.00000661, -0.00000236, 0.00000499, -0.00000055, -0.00000101, 0.00000008, -0.00000115, -0.00000066, },
		{ -0.00000005, 0.00000117, -0.00000959, 0.00000363, 0.00000325, 0.00000138, 0.00000436, -0.00000223, -0.00000432, 0.00000037, 0.00000111, -0.00000053, -0.00000112, -0.00000184, 0.00000074, -0.00000212, 0.00000104, 0.00000089, 0.00000465, 0.00000372, -0.00001208, 0.00000314, -0.00000183, 0.00000198, -0.00000492, 0.00000909, -0.00000856, 0.00000631, -0.00001001, 0.00000786, -0.00000528, 0.00000702, -0.00001315, 0.00000298, 0.00000764, -0.00000148, -0.00000965, 0.00000002, 0.00000734, -0.00000173, -0.00000885, 0.00000610, 0.00000136, -0.00000457, 0.00000875, -0.00000351, -0.00000328, 0.00000946, -0.00003656, 0.00002962, -0.00002809, 0.00001564, -0.00002762, 0.00001548, -0.00003169, 0.00001401, -0.00005947, 0.00003745, -0.00007622, 0.00001191, -0.00026382, 0.00001254, 0.00036809, 0.00013525, 0.00037265, -0.00029474, -0.00004447, -0.00003984, 0.00000095, -0.00002706, -0.00000112, -0.00000642, 0.00001892, -0.00001160, 0.00000798, -0.00000920, 0.00000886, -0.00003394, 0.00001749, -0.00002325, 0.00001849, -0.00002494, 0.00000511, -0.00001387, 0.00001490, -0.00000816, -0.00000449, -0.00001041, 0.00002137, -0.00001279, -0.00000669, -0.00000148, 0.00000686, -0.00000189, 0.00000004, -0.00000022, 0.00000216, -0.00000485, 0.00000758, -0.00000229, 0.00000334, -0.00000097, -0.00000828, -0.00000344, 0.00000905, -0.00000708, 0.00000590, -0.00000214, -0.00000035, -0.00000211, 0.00000033, -0.00000113, 0.00000439, -0.00000418, -0.00000142, -0.00000535, 0.00000643, -0.00000120, 0.00000336, 0.00000244, -0.00000680, -0.00000122, 0.00000476, -0.00000104, -0.00000065, -0.00000014, -0.00000104, -0.00000068, },
		{ 0.00000042, -0.00000000, -0.00000941, 0.00000521, 0.00000206, 0.00000223, 0.00000352, -0.00000240, -0.00000424, 0.00000081, 0.00000084, -0.00000038, -0.00000138, -0.00000168, 0.00000077, -0.00000211, 0.00000149, -0.00000005, 0.00000643, 0.00000107, -0.00001066, 0.00000300, -0.00000142, 0.00000130, -0.00000359, 0.00000737, -0.00000763, 0.00000592, -0.00001002, 0.00000840, -0.00000605, 0.00000802, -0.00001523, 0.00000580, 0.00000617, -0.00000168, -0.00001012, 0.00000182, 0.00000667, -0.00000233, -0.00000834, 0.00000578, 0.00000156, -0.00000463, 0.00001092, -0.00000826, 0.00000291, 0.00000225, -0.00003022, 0.00002476, -0.00002528, 0.00001417, -0.00002624, 0.00001337, -0.00003012, 0.00001261, -0.00005736, 0.00003310, -0.00007335, 0.00000709, -0.00027884, 0.00008258, 0.00032552, 0.00012349, 0.00035926, -0.00027540, -0.00005067, -0.00003484, -0.00000190, -0.00002335, -0.00000308, -0.00000369, 0.00001607, -0.00000551, 0.00000176, -0.00000330, 0.00000063, -0.00002776, 0.00001405, -0.00002169, 0.00001743, -0.00002400, 0.00000394, -0.00001284, 0.00001520, -0.00000958, -0.00000489, -0.00000848, 0.00002145, -0.00001502, -0.00000546, -0.00000109, 0.00000713, -0.00000260, 0.00000037, -0.00000022, 0.00000154, -0.00000341, 0.00000679, -0.00000237, 0.00000357, -0.00000163, -0.00000896, -0.00000119, 0.00000757, -0.00000632, 0.00000562, -0.00000251, -0.00000040, -0.00000183, -0.00000003, -0.00000040, 0.00000385, -0.00000405, -0.00000206, -0.00000441, 0.00000658, -0.00000195, 0.00000439, 0.00000106, -0.00000674, -0.00000014, 0.00000436, -0.00000141, -0.00000030, -0.00000041, -0.00000085, -0.00000076, },
		{ 0.00000091, -0.00000139, -0.00000873, 0.00000639, 0.00000103, 0.00000301, 0.00000260, -0.00000246, -0.00000413, 0.00000119, 0.00000055, -0.00000019, -0.00000169, -0.00000142, 0.00000068, -0.00000194, 0.00000170, -0.00000071, 0.00000799, -0.00000167, -0.00000881, 0.00000242, -0.00000072, 0.00000039, -0.00000196, 0.00000514, -0.00000600, 0.00000474, -0.00000920, 0.00000812, -0.00000603, 0.00000820, -0.00001650, 0.00000816, 0.00000474, -0.00000195, -0.00001029, 0.00000349, 0.00000583, -0.00000279, -0.00000779, 0.00000544, 0.00000162, -0.00000422, 0.00001231, -0.00001233, 0.00000881, -0.00000540, -0.00002257, 0.00001799, -0.00002033, 0.00001062, -0.00002284, 0.00000912, -0.00002618, 0.00000857, -0.00005220, 0.00002487, -0.00006526, -0.00000599, -0.00028086, 0.00014361, 0.00027927, 0.00011892, 0.00034280, -0.00025426, -0.00005848, -0.00002857, -0.00000574, -0.00001894, -0.00000543, -0.00000078, 0.00001332, 0.00000041, -0.00000440, 0.00000253, -0.00000822, -0.00002014, 0.00000888, -0.00001844, 0.00001488, -0.00002177, 0.00000163, -0.00001060, 0.00001429, -0.00001018, -0.00000579, -0.00000573, 0.00002030, -0.00001631, -0.00000464, -0.00000034, 0.00000697, -0.00000303, 0.00000058, -0.00000017, 0.00000090, -0.00000180, 0.00000566, -0.00000213, 0.00000359, -0.00000232, -0.00000937, 0.00000098, 0.00000582, -0.00000518, 0.00000497, -0.00000263, -0.00000057, -0.00000144, -0.00000047, 0.00000048, 0.00000307, -0.00000371, -0.00000286, -0.00000314, 0.00000628, -0.00000230, 0.00000518, -0.00000032, -0.00000645, 0.00000087, 0.00000383, -0.00000165, 0.00000002, -0.00000070, -0.00000063, -0.00000086, },
		{ 0.00000136, -0.00000290, -0.00000759, 0.00000710, 0.00000024, 0.00000366, 0.00000167, -0.00000246, -0.00000397, 0.00000151, 0.00000028, -0.00000000, -0.00000202, -0.00000108, 0.00000048, -0.00000160, 0.00000166, -0.00000102, 0.00000917, -0.00000428, -0.00000671, 0.00000155, 0.00000014, -0.00000062, -0.00000021, 0.00000258, -0.00000384, 0.00000292, -0.00000765, 0.00000708, -0.00000525, 0.00000754, -0.00001689, 0.00000992, 0.00000346, -0.00000233, -0.00001012, 0.00000496, 0.00000486, -0.00000311, -0.00000726, 0.00000511, 0.00000151, -0.00000326, 0.00001272, -0.00001532, 0.00001384, -0.00001276, -0.00001440, 0.00001003, -0.00001386, 0.00000551, -0.00001792, 0.00000326, -0.00002044, 0.00000252, -0.00004470, 0.00001356, -0.00005289, -0.00002651, -0.00026948, 0.00019294, 0.00023298, 0.00011905, 0.00032460, -0.00023229, -0.00006713, -0.00002165, -0.00001006, -0.00001426, -0.00000778, 0.00000191, 0.00001110, 0.00000567, -0.00000995, 0.00000758, -0.00001684, -0.00001190, 0.00000264, -0.00001399, 0.00001120, -0.00001857, -0.00000155, -0.00000739, 0.00001234, -0.00001008, -0.00000701, -0.00000237, 0.00001807, -0.00001668, -0.00000424, 0.00000076, 0.00000635, -0.00000316, 0.00000066, -0.00000011, 0.00000029, -0.00000013, 0.00000429, -0.00000164, 0.00000341, -0.00000304, -0.00000945, 0.00000294, 0.00000396, -0.00000379, 0.00000403, -0.00000253, -0.00000082, -0.00000102, -0.00000092, 0.00000141, 0.00000215, -0.00000325, -0.00000373, -0.00000165, 0.00000557, -0.00000226, 0.00000565, -0.00000164, -0.00000598, 0.00000176, 0.00000319, -0.00000174, 0.00000026, -0.00000097, -0.00000038, -0.00000095, },
		{ 0.00000170, -0.00000445, -0.00000606, 0.00000735, -0.00000026, 0.00000409, 0.00000080, -0.00000242, -0.00000374, 0.00000172, 0.00000005, 0.00000017, -0.00000235, -0.00000067, 0.00000016, -0.00000111, 0.00000134, -0.00000093, 0.00000985, -0.00000657, -0.00000457, 0.00000054, 0.00000103, -0.00000158, 0.00000148, -0.00000008, -0.00000136, 0.00000065, -0.00000553, 0.00000541, -0.00000381, 0.00000610, -0.00001639, 0.00001100, 0.00000243, -0.00000286, -0.00000958, 0.00000616, 0.00000380, -0.00000327, -0.00000678, 0.00000483, 0.00000124, -0.00000174, 0.00001204, -0.00001695, 0.00001748, -0.00001911, -0.00000652, 0.00000167, -0.00000658, -0.00000051, -0.00001212, -0.00000352, -0.00001362, -0.00000475, -0.00003581, 0.00000035, -0.00003763, -0.00005292, -0.00024527, 0.00022888, 0.00018965, 0.00012155, 0.00030601, -0.00021047, -0.00007585, -0.00001469, -0.00001435, -0.00000975, -0.00000975, 0.00000407, 0.00000977, 0.00000986, -0.00001443, 0.00001122, -0.00002441, -0.00000384, -0.00000397, -0.00000892, 0.00000687, -0.00001481, -0.00000523, -0.00000354, 0.00000960, -0.00000947, -0.00000833, 0.00000131, 0.00001500, -0.00001623, -0.00000422, 0.00000214, 0.00000534, -0.00000298, 0.00000063, -0.00000006, -0.00000021, 0.00000151, 0.00000277, -0.00000094, 0.00000306, -0.00000379, -0.00000917, 0.00000455, 0.00000217, -0.00000230, 0.00000290, -0.00000230, -0.00000109, -0.00000064, -0.00000128, 0.00000229, 0.00000118, -0.00000276, -0.00000458, -0.00000004, 0.00000453, -0.00000184, 0.00000579, -0.00000283, -0.00000536, 0.00000254, 0.00000246, -0.00000169, 0.00000042, -0.00000119, -0.00000015, -0.00000102, },
		{ 0.00000189, -0.00000594, -0.00000423, 0.00000718, -0.00000046, 0.00000426, 0.00000005, -0.00000237, -0.00000343, 0.00000181, -0.00000010, 0.00000028, -0.00000264, -0.00000023, -0.00000023, -0.00000049, 0.00000079, -0.00000043, 0.00000995, -0.00000838, -0.00000258, -0.00000046, 0.00000183, -0.00000234, 0.00000291, -0.00000260, 0.00000117, -0.00000183, -0.00000305, 0.00000329, -0.00000187, 0.00000402, -0.00001505, 0.00001135, 0.00000172, -0.00000355, -0.00000867, 0.00000705, 0.00000271, -0.00000328, -0.00000636, 0.00000457, 0.00000085, 0.00000026, 0.00001025, -0.00001705, 0.00001938, -0.00002387, 0.00000032, -0.00000628, 0.00000078, -0.00000674, -0.00000614, -0.00001048, -0.00000653, -0.00001233, -0.00002659, -0.00001346, -0.00002121, -0.00008310, -0.00020978, 0.00025081, 0.00015154, 0.00012448, 0.00028820, -0.00018965, -0.00008392, -0.00000823, -0.00001817, -0.00000574, -0.00001109, 0.00000545, 0.00000954, 0.00001272, -0.00001754, 0.00001302, -0.00003029, 0.00000328, -0.00001023, -0.00000383, 0.00000237, -0.00001090, -0.00000899, 0.00000056, 0.00000636, -0.00000856, -0.00000952, 0.00000499, 0.00001136, -0.00001512, -0.00000449, 0.00000370, 0.00000401, -0.00000255, 0.00000050, -0.00000005, -0.00000054, 0.00000301, 0.00000119, -0.00000010, 0.00000254, -0.00000455, -0.00000851, 0.00000573, 0.00000057, -0.00000084, 0.00000169, -0.00000200, -0.00000133, -0.00000035, -0.00000149, 0.00000302, 0.00000025, -0.00000231, -0.00000531, 0.00000157, 0.00000328, -0.00000110, 0.00000559, -0.00000384, -0.00000463, 0.00000317, 0.00000167, -0.00000150, 0.00000046, -0.00000134, 0.00000003, -0.00000101, },
		{ 0.00000186, -0.00000724, -0.00000224, 0.00000664, -0.00000036, 0.00000417, -0.00000055, -0.00000234, -0.00000305, 0.00000178, -0.00000016, 0.00000031, -0.00000286, 0.00000020, -0.00000065, 0.00000020, 0.00000004, 0.00000044, 0.00000945, -0.00000960, -0.00000088, -0.00000130, 0.00000242, -0.00000278, 0.00000392, -0.00000477, 0.00000353, -0.00000427, -0.00000047, 0.00000096, 0.00000035, 0.00000149, -0.00001301, 0.00001101, 0.00000133, -0.00000439, -0.00000741, 0.00000758, 0.00000165, -0.00000317, -0.00000599, 0.00000432, 0.00000043, 0.00000261, 0.00000750, -0.00001563, 0.00001938, -0.00002664, 0.00000553, -0.00001314, 0.00000749, -0.00001251, -0.00000065, -0.00001689, 0.00000003, -0.00001935, -0.00001811, -0.00002649, -0.00000545, -0.00011458, -0.00016530, 0.00025911, 0.00012009, 0.00012635, 0.00027209, -0.00017052, -0.00009082, -0.00000270, -0.00002121, -0.00000247, -0.00001162, 0.00000596, 0.00001051, 0.00001413, -0.00001917, 0.00001276, -0.00003401, 0.00000887, -0.00001555, 0.00000072, -0.00000181, -0.00000728, -0.00001246, 0.00000453, 0.00000294, -0.00000758, -0.00001032, 0.00000835, 0.00000748, -0.00001355, -0.00000491, 0.00000532, 0.00000247, -0.00000192, 0.00000030, -0.00000010, -0.00000065, 0.00000430, -0.00000034, 0.00000080, 0.00000190, -0.00000528, -0.00000750, 0.00000643, -0.00000073, 0.00000049, 0.00000049, -0.00000168, -0.00000150, -0.00000019, -0.00000149, 0.00000353, -0.00000055, -0.00000198, -0.00000583, 0.00000304, 0.00000194, -0.00000014, 0.00000508, -0.00000464, -0.00000381, 0.00000364, 0.00000087, -0.00000119, 0.00000039, -0.00000140, 0.00000016, -0.00000092, },
		{ 0.00000159, -0.00000827, -0.00000020, 0.00000583, -0.00000002, 0.00000382, -0.00000097, -0.00000235, -0.00000259, 0.00000161, -0.00000011, 0.00000024, -0.00000298, 0.00000056, -0.00000107, 0.00000091, -0.00000082, 0.00000161, 0.00000838, -0.00001018, 0.00000040, -0.00000191, 0.00000272, -0.00000281, 0.00000439, -0.00000640, 0.00000548, -0.00000642, 0.00000197, -0.00000135, 0.00000263, -0.00000127, -0.00001046, 0.00001007, 0.00000126, -0.00000535, -0.00000582, 0.00000773, 0.00000068, -0.00000298, -0.00000565, 0.00000402, 0.00000010, 0.00000510, 0.00000400, -0.00001286, 0.00001747, -0.00002722, 0.00000868, -0.00001836, 0.00001300, -0.00001726, 0.00000376, -0.00002214, 0.00000538, -0.00002503, -0.00001130, -0.00003752, 0.00000787, -0.00014472, -0.00011466, 0.00025504, 0.00009592, 0.00012620, 0.00025827, -0.00015352, -0.00009617, 0.00000164, -0.00002327, -0.00000006, -0.00001130, 0.00000563, 0.00001261, 0.00001416, -0.00001942, 0.00001045, -0.00003536, 0.00001251, -0.00001945, 0.00000431, -0.00000528, -0.00000428, -0.00001528, 0.00000805, -0.00000038, -0.00000672, -0.00001055, 0.00001110, 0.00000368, -0.00001175, -0.00000533, 0.00000684, 0.00000084, -0.00000116, 0.00000007, -0.00000023, -0.00000052, 0.00000530, -0.00000176, 0.00000172, 0.00000113, -0.00000594, -0.00000620, 0.00000664, -0.00000166, 0.00000159, -0.00000062, -0.00000139, -0.00000158, -0.00000018, -0.00000125, 0.00000377, -0.00000116, -0.00000183, -0.00000606, 0.00000427, 0.00000063, 0.00000095, 0.00000430, -0.00000522, -0.00000294, 0.00000395, 0.00000009, -0.00000079, 0.00000021, -0.00000136, 0.00000020, -0.00000072, },
		{ 0.00000104, -0.00000894, 0.00000174, 0.00000485, 0.00000049, 0.00000324, -0.00000122, -0.00000239, -0.00000207, 0.00000132, 0.00000005, 0.00000005, -0.00000296, 0.00000083, -0.00000141, 0.00000157, -0.00000168, 0.00000296, 0.00000684, -0.00001011, 0.00000122, -0.00000220, 0.00000271, -0.00000242, 0.00000424, -0.00000737, 0.00000685, -0.00000809, 0.00000406, -0.00000342, 0.00000473, -0.00000400, -0.00000762, 0.00000866, 0.00000144, -0.00000637, -0.00000397, 0.00000750, -0.00000016, -0.00000275, -0.00000531, 0.00000362, 0.00000000, 0.00000748, 0.00000008, -0.00000904, 0.00001390, -0.00002566, 0.00000960, -0.00002159, 0.00001689, -0.00002059, 0.00000669, -0.00002577, 0.00000903, -0.00002880, -0.00000688, -0.00004557, 0.00001724, -0.00017101, -0.00006099, 0.00024053, 0.00007893, 0.00012358, 0.00024702, -0.00013883, -0.00009983, 0.00000469, -0.00002432, 0.00000153, -0.00001023, 0.00000463, 0.00001564, 0.00001302, -0.00001857, 0.00000634, -0.00003437, 0.00001405, -0.00002167, 0.00000662, -0.00000776, -0.00000216, -0.00001722, 0.00001085, -0.00000337, -0.00000612, -0.00001009, 0.00001300, 0.00000022, -0.00000993, -0.00000561, 0.00000814, -0.00000075, -0.00000035, -0.00000018, -0.00000045, -0.00000013, 0.00000597, -0.00000298, 0.00000259, 0.00000028, -0.00000650, -0.00000469, 0.00000639, -0.00000220, 0.00000241, -0.00000161, -0.00000115, -0.00000156, -0.00000032, -0.00000079, 0.00000371, -0.00000154, -0.00000190, -0.00000595, 0.00000517, -0.00000051, 0.00000206, 0.00000333, -0.00000557, -0.00000204, 0.00000408, -0.00000062, -0.00000033, -0.00000006, -0.00000122, 0.00000016, -0.00000042, },
		{ 0.00000024, -0.00000919, 0.00000348, 0.00000381, 0.00000110, 0.00000250, -0.00000131, -0.00000245, -0.00000152, 0.00000093, 0.00000031, -0.00000025, -0.00000279, 0.00000097, -0.00000164, 0.00000210, -0.00000246, 0.00000435, 0.00000494, -0.00000948, 0.00000158, -0.00000219, 0.00000239, -0.00000161, 0.00000349, -0.00000764, 0.00000754, -0.00000914, 0.00000562, -0.00000505, 0.00000644, -0.00000648, -0.00000470, 0.00000693, 0.00000181, -0.00000739, -0.00000195, 0.00000691, -0.00000083, -0.00000253, -0.00000494, 0.00000306, 0.00000025, 0.00000948, -0.00000390, -0.00000459, 0.00000902, -0.00002218, 0.00000834, -0.00002271, 0.00001897, -0.00002227, 0.00000793, -0.00002754, 0.00001068, -0.00003034, -0.00000524, -0.00005000, 0.00002155, -0.00019122, -0.00000747, 0.00021797, 0.00006842, 0.00011848, 0.00023827, -0.00012640, -0.00010184, 0.00000652, -0.00002445, 0.00000241, -0.00000859, 0.00000319, 0.00001930, 0.00001103, -0.00001704, 0.00000089, -0.00003132, 0.00001354, -0.00002214, 0.00000753, -0.00000912, -0.00000105, -0.00001813, 0.00001275, -0.00000587, -0.00000585, -0.00000888, 0.00001394, -0.00000266, -0.00000828, -0.00000564, 0.00000908, -0.00000219, 0.00000044, -0.00000041, -0.00000074, 0.00000051, 0.00000628, -0.00000396, 0.00000335, -0.00000066, -0.00000688, -0.00000306, 0.00000576, -0.00000236, 0.00000293, -0.00000243, -0.00000096, -0.00000146, -0.00000056, -0.00000011, 0.00000337, -0.00000169, -0.00000217, -0.00000551, 0.00000567, -0.00000141, 0.00000309, 0.00000224, -0.00000572, -0.00000113, 0.00000402, -0.00000123, 0.00000014, -0.00000039, -0.00000100, 0.00000004, -0.00000003, },
		{ -0.00000080, -0.00000899, 0.00000492, 0.00000282, 0.00000171, 0.00000166, -0.00000128, -0.00000251, -0.00000096, 0.00000048, 0.00000066, -0.00000067, -0.00000247, 0.00000095, -0.00000172, 0.00000246, -0.00000305, 0.00000564, 0.00000285, -0.00000839, 0.00000152, -0.00000190, 0.00000181, -0.00000047, 0.00000219, -0.00000720, 0.00000751, -0.00000947, 0.00000654, -0.00000610, 0.00000760, -0.00000850, -0.00000193, 0.00000505, 0.00000226, -0.00000830, 0.00000018, 0.00000599, -0.00000129, -0.00000233, -0.00000450, 0.00000232, 0.00000097, 0.00001087, -0.00000754, 0.00000002, 0.00000336, -0.00001724, 0.00000517, -0.00002184, 0.00001925, -0.00002232, 0.00000746, -0.00002743, 0.00001031, -0.00002963, -0.00000647, -0.00005063, 0.00002020, -0.00020365, 0.00004296, 0.00018997, 0.00006328, 0.00011127, 0.00023171, -0.00011597, -0.00010240, 0.00000731, -0.00002383, 0.00000280, -0.00000661, 0.00000164, 0.00002323, 0.00000855, -0.00001530, -0.00000533, -0.00002667, 0.00001126, -0.00002097, 0.00000708, -0.00000936, -0.00000093, -0.00001799, 0.00001367, -0.00000783, -0.00000590, -0.00000698, 0.00001385, -0.00000482, -0.00000691, -0.00000532, 0.00000958, -0.00000338, 0.00000115, -0.00000060, -0.00000110, 0.00000138, 0.00000622, -0.00000466, 0.00000398, -0.00000164, -0.00000706, -0.00000143, 0.00000484, -0.00000219, 0.00000316, -0.00000310, -0.00000082, -0.00000134, -0.00000086, 0.00000070, 0.00000280, -0.00000164, -0.00000262, -0.00000475, 0.00000578, -0.00000199, 0.00000394, 0.00000110, -0.00000569, -0.00000023, 0.00000379, -0.00000171, 0.00000060, -0.00000076, -0.00000072, -0.00000013, 0.00000042, },
		{ -0.00000202, -0.00000835, 0.00000600, 0.00000197, 0.00000224, 0.00000079, -0.00000118, -0.00000255, -0.00000044, -0.00000001, 0.00000105, -0.00000116, -0.00000200, 0.00000076, -0.00000163, 0.00000259, -0.00000336, 0.00000669, 0.00000074, -0.00000698, 0.00000115, -0.00000140, 0.00000107, 0.00000090, 0.00000046, -0.00000615, 0.00000680, -0.00000909, 0.00000677, -0.00000651, 0.00000811, -0.00000991, 0.00000050, 0.00000317, 0.00000270, -0.00000904, 0.00000229, 0.00000482, -0.00000155, -0.00000220, -0.00000400, 0.00000136, 0.00000220, 0.00001144, -0.00001050, 0.00000429, -0.00000252, -0.00001138, 0.00000053, -0.00001925, 0.00001792, -0.00002091, 0.00000547, -0.00002562, 0.00000811, -0.00002689, -0.00001029, -0.00004769, 0.00001315, -0.00020723, 0.00008775, 0.00015915, 0.00006213, 0.00010256, 0.00022687, -0.00010716, -0.00010181, 0.00000732, -0.00002271, 0.00000293, -0.00000457, 0.00000030, 0.00002707, 0.00000598, -0.00001388, -0.00001164, -0.00002103, 0.00000763, -0.00001845, 0.00000544, -0.00000861, -0.00000171, -0.00001688, 0.00001366, -0.00000923, -0.00000620, -0.00000452, 0.00001280, -0.00000622, -0.00000589, -0.00000464, 0.00000960, -0.00000425, 0.00000172, -0.00000074, -0.00000147, 0.00000241, 0.00000583, -0.00000508, 0.00000442, -0.00000265, -0.00000701, 0.00000011, 0.00000374, -0.00000176, 0.00000315, -0.00000363, -0.00000069, -0.00000122, -0.00000115, 0.00000157, 0.00000205, -0.00000144, -0.00000320, -0.00000373, 0.00000550, -0.00000222, 0.00000455, -0.00000002, -0.00000550, 0.00000063, 0.00000339, -0.00000202, 0.00000099, -0.00000112, -0.00000041, -0.00000033, 0.00000089, },
		{ -0.00000334, -0.00000732, 0.00000668, 0.00000132, 0.00000262, -0.00000002, -0.00000104, -0.00000253, 0.00000000, -0.00000047, 0.00000145, -0.00000169, -0.00000141, 0.00000041, -0.00000136, 0.00000248, -0.00000335, 0.00000738, -0.00000121, -0.00000543, 0.00000056, -0.00000079, 0.00000029, 0.00000233, -0.00000153, -0.00000463, 0.00000552, -0.00000806, 0.00000633, -0.00000627, 0.00000791, -0.00001061, 0.00000242, 0.00000146, 0.00000305, -0.00000953, 0.00000429, 0.00000344, -0.00000161, -0.00000214, -0.00000344, 0.00000023, 0.00000397, 0.00001106, -0.00001246, 0.00000778, -0.00000804, -0.00000524, -0.00000503, -0.00001540, 0.00001534, -0.00001836, 0.00000232, -0.00002248, 0.00000448, -0.00002260, -0.00001616, -0.00004179, 0.00000092, -0.00020156, 0.00012496, 0.00012791, 0.00006353, 0.00009307, 0.00022318, -0.00009951, -0.00010044, 0.00000686, -0.00002134, 0.00000304, -0.00000271, -0.00000055, 0.00003048, 0.00000362, -0.00001323, -0.00001741, -0.00001506, 0.00000320, -0.00001499, 0.00000291, -0.00000711, -0.00000316, -0.00001500, 0.00001282, -0.00001015, -0.00000661, -0.00000169, 0.00001094, -0.00000689, -0.00000523, -0.00000360, 0.00000912, -0.00000476, 0.00000213, -0.00000084, -0.00000184, 0.00000355, 0.00000514, -0.00000520, 0.00000468, -0.00000365, -0.00000670, 0.00000146, 0.00000257, -0.00000115, 0.00000292, -0.00000404, -0.00000055, -0.00000115, -0.00000135, 0.00000242, 0.00000122, -0.00000115, -0.00000382, -0.00000254, 0.00000489, -0.00000210, 0.00000488, -0.00000105, -0.00000518, 0.00000143, 0.00000284, -0.00000216, 0.00000128, -0.00000144, -0.00000011, -0.00000052, 0.00000131, },
		{ -0.00000467, -0.00000595, 0.00000695, 0.00000090, 0.00000279, -0.00000073, -0.00000090, -0.00000244, 0.00000034, -0.00000088, 0.00000182, -0.00000223, -0.00000075, -0.00000007, -0.00000093, 0.00000213, -0.00000298, 0.00000763, -0.00000286, -0.00000388, -0.00000010, -0.00000018, -0.00000042, 0.00000368, -0.00000360, -0.00000283, 0.00000382, -0.00000650, 0.00000533, -0.00000543, 0.00000705, -0.00001057, 0.00000375, 0.00000001, 0.00000323, -0.00000968, 0.00000608, 0.00000196, -0.00000152, -0.00000213, -0.00000285, -0.00000102, 0.00000621, 0.00000971, -0.00001325, 0.00001013, -0.00001268, 0.00000057, -0.00001087, -0.00001081, 0.00001195, -0.00001512, -0.00000154, -0.00001851, -0.00000004, -0.00001737, -0.00002331, -0.00003389, -0.00001543, -0.00018698, 0.00015332, 0.00009829, 0.00006608, 0.00008357, 0.00022003, -0.00009254, -0.00009865, 0.00000623, -0.00001996, 0.00000333, -0.00000123, -0.00000068, 0.00003321, 0.00000173, -0.00001369, -0.00002207, -0.00000939, -0.00000146, -0.00001105, -0.00000013, -0.00000516, -0.00000501, -0.00001259, 0.00001133, -0.00001070, -0.00000697, 0.00000125, 0.00000847, -0.00000692, -0.00000486, -0.00000228, 0.00000821, -0.00000492, 0.00000236, -0.00000089, -0.00000216, 0.00000473, 0.00000421, -0.00000506, 0.00000471, -0.00000460, -0.00000614, 0.00000257, 0.00000143, -0.00000044, 0.00000255, -0.00000435, -0.00000038, -0.00000119, -0.00000140, 0.00000315, 0.00000039, -0.00000086, -0.00000440, -0.00000129, 0.00000404, -0.00000168, 0.00000491, -0.00000194, -0.00000476, 0.00000216, 0.00000217, -0.00000212, 0.00000144, -0.00000169, 0.00000015, -0.00000066, 0.00000165, },
		{ -0.00000593, -0.00000434, 0.00000687, 0.00000073, 0.00000273, -0.00000128, -0.00000081, -0.00000226, 0.00000055, -0.00000118, 0.00000211, -0.00000271, -0.00000006, -0.00000064, -0.00000038, 0.00000159, -0.00000228, 0.00000740, -0.00000409, -0.00000249, -0.00000073, 0.00000034, -0.00000094, 0.00000478, -0.00000555, -0.00000094, 0.00000189, -0.00000459, 0.00000389, -0.00000412, 0.00000562, -0.00000982, 0.00000444, -0.00000108, 0.00000318, -0.00000947, 0.00000757, 0.00000047, -0.00000130, -0.00000215, -0.00000228, -0.00000230, 0.00000881, 0.00000745, -0.00001280, 0.00001111, -0.00001602, 0.00000548, -0.00001640, -0.00000604, 0.00000825, -0.00001165, -0.00000560, -0.00001424, -0.00000487, -0.00001192, -0.00003083, -0.00002515, -0.00003449, -0.00016446, 0.00017225, 0.00007184, 0.00006861, 0.00007472, 0.00021692, -0.00008585, -0.00009675, 0.00000566, -0.00001875, 0.00000395, -0.00000023, 0.00000005, 0.00003509, 0.00000039, -0.00001544, -0.00002518, -0.00000460, -0.00000581, -0.00000709, -0.00000332, -0.00000309, -0.00000697, -0.00000992, 0.00000939, -0.00001103, -0.00000711, 0.00000406, 0.00000566, -0.00000647, -0.00000469, -0.00000079, 0.00000693, -0.00000476, 0.00000243, -0.00000092, -0.00000237, 0.00000587, 0.00000311, -0.00000469, 0.00000453, -0.00000547, -0.00000537, 0.00000338, 0.00000040, 0.00000028, 0.00000208, -0.00000459, -0.00000015, -0.00000135, -0.00000127, 0.00000370, -0.00000036, -0.00000064, -0.00000485, -0.00000007, 0.00000305, -0.00000102, 0.00000466, -0.00000268, -0.00000425, 0.00000279, 0.00000142, -0.00000192, 0.00000147, -0.00000183, 0.00000034, -0.00000073, 0.00000183, },
		{ -0.00000701, -0.00000261, 0.00000649, 0.00000077, 0.00000244, -0.00000165, -0.00000078, -0.00000197, 0.00000061, -0.00000136, 0.00000228, -0.00000308, 0.00000061, -0.00000124, 0.00000023, 0.00000092, -0.00000130, 0.00000669, -0.00000483, -0.00000137, -0.00000121, 0.00000068, -0.00000119, 0.00000551, -0.00000719, 0.00000082, -0.00000005, -0.00000252, 0.00000219, -0.00000249, 0.00000375, -0.00000848, 0.00000451, -0.00000180, 0.00000289, -0.00000885, 0.00000868, -0.00000096, -0.00000103, -0.00000216, -0.00000178, -0.00000348, 0.00001158, 0.00000442, -0.00001115, 0.00001064, -0.00001779, 0.00000904, -0.00002108, -0.00000160, 0.00000469, -0.00000842, -0.00000938, -0.00001020, -0.00000942, -0.00000695, -0.00003783, -0.00001681, -0.00005454, -0.00013554, 0.00018190, 0.00004962, 0.00007021, 0.00006704, 0.00021343, -0.00007909, -0.00009499, 0.00000533, -0.00001783, 0.00000495, 0.00000026, 0.00000167, 0.00003608, -0.00000042, -0.00001849, -0.00002651, -0.00000110, -0.00000939, -0.00000355, -0.00000628, -0.00000119, -0.00000874, -0.00000728, 0.00000722, -0.00001127, -0.00000688, 0.00000647, 0.00000277, -0.00000574, -0.00000458, 0.00000073, 0.00000540, -0.00000433, 0.00000236, -0.00000095, -0.00000243, 0.00000689, 0.00000192, -0.00000414, 0.00000414, -0.00000621, -0.00000442, 0.00000389, -0.00000046, 0.00000096, 0.00000155, -0.00000478, 0.00000012, -0.00000164, -0.00000091, 0.00000400, -0.00000096, -0.00000054, -0.00000511, 0.00000101, 0.00000204, -0.00000020, 0.00000416, -0.00000324, -0.00000367, 0.00000329, 0.00000062, -0.00000157, 0.00000136, -0.00000185, 0.00000043, -0.00000069, 0.00000182, },
		{ -0.00000782, -0.00000086, 0.00000590, 0.00000098, 0.00000194, -0.00000184, -0.00000082, -0.00000159, 0.00000051, -0.00000140, 0.00000230, -0.00000332, 0.00000118, -0.00000183, 0.00000085, 0.00000018, -0.00000011, 0.00000555, -0.00000507, -0.00000059, -0.00000149, 0.00000079, -0.00000109, 0.00000576, -0.00000839, 0.00000226, -0.00000180, -0.00000050, 0.00000044, -0.00000074, 0.00000163, -0.00000667, 0.00000403, -0.00000214, 0.00000238, -0.00000784, 0.00000936, -0.00000224, -0.00000075, -0.00000212, -0.00000143, -0.00000443, 0.00001432, 0.00000087, -0.00000850, 0.00000878, -0.00001789, 0.00001098, -0.00002451, 0.00000207, 0.00000169, -0.00000581, -0.00001246, -0.00000684, -0.00001319, -0.00000306, -0.00004350, -0.00001007, -0.00007380, -0.00010210, 0.00018299, 0.00003212, 0.00007026, 0.00006086, 0.00020929, -0.00007207, -0.00009351, 0.00000533, -0.00001722, 0.00000632, 0.00000030, 0.00000413, 0.00003621, -0.00000085, -0.00002265, -0.00002601, 0.00000086, -0.00001188, -0.00000073, -0.00000874, 0.00000028, -0.00001008, -0.00000489, 0.00000501, -0.00001152, -0.00000618, 0.00000828, 0.00000006, -0.00000491, -0.00000439, 0.00000213, 0.00000376, -0.00000372, 0.00000217, -0.00000099, -0.00000232, 0.00000772, 0.00000073, -0.00000345, 0.00000356, -0.00000678, -0.00000336, 0.00000411, -0.00000112, 0.00000154, 0.00000100, -0.00000491, 0.00000043, -0.00000204, -0.00000033, 0.00000404, -0.00000136, -0.00000061, -0.00000511, 0.00000185, 0.00000111, 0.00000067, 0.00000347, -0.00000363, -0.00000303, 0.00000365, -0.00000017, -0.00000111, 0.00000112, -0.00000175, 0.00000042, -0.00000054, 0.00000158, },
		{ -0.00000831, 0.00000079, 0.00000518, 0.00000129, 0.00000129, -0.00000185, -0.00000093, -0.00000112, 0.00000028, -0.00000129, 0.00000217, -0.00000337, 0.00000162, -0.00000234, 0.00000141, -0.00000052, 0.00000117, 0.00000407, -0.00000483, -0.00000017, -0.00000152, 0.00000066, -0.00000065, 0.00000550, -0.00000903, 0.00000325, -0.00000319, 0.00000130, -0.00000120, 0.00000096, -0.00000055, -0.00000458, 0.00000310, -0.00000215, 0.00000166, -0.00000647, 0.00000959, -0.00000331, -0.00000051, -0.00000199, -0.00000129, -0.00000502, 0.00001677, -0.00000291, -0.00000512, 0.00000575, -0.00001642, 0.00001119, -0.00002643, 0.00000468, -0.00000048, -0.00000410, -0.00001454, -0.00000449, -0.00001581, -0.00000071, -0.00004719, -0.00000591, -0.00009054, -0.00006628, 0.00017672, 0.00001939, 0.00006848, 0.00005631, 0.00020440, -0.00006469, -0.00009236, 0.00000567, -0.00001688, 0.00000797, 0.00000003, 0.00000727, 0.00003561, -0.00000116, -0.00002761, -0.00002385, 0.00000124, -0.00001311, 0.00000115, -0.00001050, 0.00000115, -0.00001083, -0.00000295, 0.00000289, -0.00001183, -0.00000495, 0.00000935, -0.00000226, -0.00000415, -0.00000399, 0.00000326, 0.00000214, -0.00000300, 0.00000192, -0.00000106, -0.00000200, 0.00000830, -0.00000040, -0.00000270, 0.00000280, -0.00000715, -0.00000225, 0.00000408, -0.00000157, 0.00000201, 0.00000043, -0.00000498, 0.00000075, -0.00000251, 0.00000043, 0.00000380, -0.00000156, -0.00000087, -0.00000484, 0.00000240, 0.00000036, 0.00000151, 0.00000265, -0.00000387, -0.00000235, 0.00000383, -0.00000090, -0.00000059, 0.00000078, -0.00000154, 0.00000031, -0.00000028, 0.00000110, },
		{ -0.00000843, 0.00000225, 0.00000442, 0.00000164, 0.00000053, -0.00000172, -0.00000109, -0.00000060, -0.00000008, -0.00000104, 0.00000187, -0.00000322, 0.00000189, -0.00000273, 0.00000185, -0.00000111, 0.00000242, 0.00000237, -0.00000418, -0.00000009, -0.00000133, 0.00000031, 0.00000011, 0.00000472, -0.00000907, 0.00000369, -0.00000409, 0.00000272, -0.00000255, 0.00000243, -0.00000260, -0.00000240, 0.00000188, -0.00000191, 0.00000082, -0.00000480, 0.00000937, -0.00000411, -0.00000037, -0.00000174, -0.00000140, -0.00000514, 0.00001872, -0.00000658, -0.00000136, 0.00000188, -0.00001360, 0.00000974, -0.00002676, 0.00000607, -0.00000167, -0.00000343, -0.00001546, -0.00000334, -0.00001709, -0.00000012, -0.00004855, -0.00000508, -0.00010326, -0.00003022, 0.00016463, 0.00001107, 0.00006486, 0.00005334, 0.00019878, -0.00005697, -0.00009150, 0.00000628, -0.00001671, 0.00000975, -0.00000036, 0.00001092, 0.00003444, -0.00000165, -0.00003291, -0.00002035, 0.00000017, -0.00001309, 0.00000202, -0.00001147, 0.00000135, -0.00001090, -0.00000157, 0.00000098, -0.00001221, -0.00000323, 0.00000962, -0.00000404, -0.00000361, -0.00000328, 0.00000401, 0.00000068, -0.00000226, 0.00000163, -0.00000116, -0.00000147, 0.00000858, -0.00000139, -0.00000194, 0.00000192, -0.00000729, -0.00000117, 0.00000385, -0.00000182, 0.00000235, -0.00000013, -0.00000497, 0.00000102, -0.00000300, 0.00000132, 0.00000334, -0.00000155, -0.00000130, -0.00000429, 0.00000262, -0.00000015, 0.00000223, 0.00000178, -0.00000398, -0.00000162, 0.00000383, -0.00000152, -0.00000005, 0.00000038, -0.00000125, 0.00000011, 0.00000008, 0.00000040, },
		{ -0.00000816, 0.00000345, 0.00000372, 0.00000193, -0.00000025, -0.00000150, -0.00000126, -0.00000006, -0.00000051, -0.00000069, 0.00000143, -0.00000287, 0.00000195, -0.00000296, 0.00000213, -0.00000151, 0.00000354, 0.00000059, -0.00000323, -0.00000030, -0.00000094, -0.00000022, 0.00000112, 0.00000350, -0.00000854, 0.00000355, -0.00000444, 0.00000365, -0.00000350, 0.00000353, -0.00000435, -0.00000031, 0.00000050, -0.00000152, -0.00000006, -0.00000291, 0.00000874, -0.00000463, -0.00000034, -0.00000136, -0.00000178, -0.00000471, 0.00001996, -0.00000983, 0.00000237, -0.00000240, -0.00000980, 0.00000687, -0.00002559, 0.00000624, -0.00000186, -0.00000379, -0.00001523, -0.00000339, -0.00001704, -0.00000132, -0.00004748, -0.00000791, -0.00011082, 0.00000408, 0.00014841, 0.00000652, 0.00005961, 0.00005174, 0.00019259, -0.00004903, -0.00009082, 0.00000707, -0.00001659, 0.00001151, -0.00000066, 0.00001486, 0.00003286, -0.00000262, -0.00003804, -0.00001599, -0.00000208, -0.00001195, 0.00000194, -0.00001167, 0.00000089, -0.00001030, -0.00000079, -0.00000071, -0.00001261, -0.00000113, 0.00000913, -0.00000522, -0.00000335, -0.00000220, 0.00000429, -0.00000053, -0.00000159, 0.00000136, -0.00000131, -0.00000072, 0.00000852, -0.00000219, -0.00000123, 0.00000095, -0.00000721, -0.00000017, 0.00000349, -0.00000191, 0.00000258, -0.00000069, -0.00000487, 0.00000121, -0.00000343, 0.00000226, 0.00000270, -0.00000139, -0.00000186, -0.00000352, 0.00000252, -0.00000038, 0.00000276, 0.00000092, -0.00000398, -0.00000086, 0.00000364, -0.00000201, 0.00000045, -0.00000004, -0.00000090, -0.00000016, 0.00000049, -0.00000049, },
		{ -0.00000752, 0.00000432, 0.00000313, 0.00000212, -0.00000098, -0.00000124, -0.00000142, 0.00000045, -0.00000097, -0.00000027, 0.00000089, -0.00000235, 0.00000181, -0.00000301, 0.00000221, -0.00000166, 0.00000440, -0.00000112, -0.00000211, -0.00000071, -0.00000042, -0.00000082, 0.00000227, 0.00000193, -0.00000750, 0.00000285, -0.00000420, 0.00000404, -0.00000397, 0.00000417, -0.00000568, 0.00000151, -0.00000087, -0.00000107, -0.00000090, -0.00000091, 0.00000775, -0.00000487, -0.00000044, -0.00000086, -0.00000241, -0.00000371, 0.00002035, -0.00001237, 0.00000571, -0.00000665, -0.00000549, 0.00000295, -0.00002314, 0.00000531, -0.00000117, -0.00000506, -0.00001397, -0.00000450, -0.00001583, -0.00000408, -0.00004419, -0.00001438, -0.00011251, 0.00003488, 0.00012977, 0.00000491, 0.00005308, 0.00005122, 0.00018603, -0.00004103, -0.00009018, 0.00000790, -0.00001636, 0.00001307, -0.00000067, 0.00001886, 0.00003102, -0.00000435, -0.00004250, -0.00001127, -0.00000510, -0.00000995, 0.00000107, -0.00001124, -0.00000011, -0.00000909, -0.00000058, -0.00000218, -0.00001294, 0.00000121, 0.00000797, -0.00000579, -0.00000339, -0.00000076, 0.00000406, -0.00000140, -0.00000105, 0.00000111, -0.00000148, 0.00000021, 0.00000814, -0.00000274, -0.00000064, -0.00000005, -0.00000692, 0.00000070, 0.00000305, -0.00000187, 0.00000270, -0.00000125, -0.00000464, 0.00000126, -0.00000373, 0.00000317, 0.00000196, -0.00000112, -0.00000249, -0.00000257, 0.00000215, -0.00000031, 0.00000306, 0.00000012, -0.00000389, -0.00000008, 0.00000325, -0.00000231, 0.00000088, -0.00000045, -0.00000053, -0.00000045, 0.00000093, -0.00000153, },
		{ -0.00000656, 0.00000487, 0.00000270, 0.00000213, -0.00000161, -0.00000097, -0.00000152, 0.00000089, -0.00000141, 0.00000016, 0.00000028, -0.00000167, 0.00000148, -0.00000287, 0.00000209, -0.00000152, 0.00000492, -0.00000264, -0.00000095, -0.00000122, 0.00000013, -0.00000141, 0.00000343, 0.00000016, -0.00000607, 0.00000169, -0.00000344, 0.00000391, -0.00000394, 0.00000430, -0.00000649, 0.00000295, -0.00000210, -0.00000068, -0.00000159, 0.00000108, 0.00000649, -0.00000486, -0.00000064, -0.00000027, -0.00000325, -0.00000216, 0.00001982, -0.00001401, 0.00000833, -0.00001044, -0.00000114, -0.00000160, -0.00001973, 0.00000352, 0.00000017, -0.00000699, -0.00001194, -0.00000640, -0.00001378, -0.00000800, -0.00003916, -0.00002406, -0.00010811, 0.00006087, 0.00011028, 0.00000537, 0.00004573, 0.00005141, 0.00017937, -0.00003315, -0.00008943, 0.00000863, -0.00001591, 0.00001431, -0.00000019, 0.00002277, 0.00002898, -0.00000699, -0.00004584, -0.00000675, -0.00000846, -0.00000744, -0.00000034, -0.00001034, -0.00000150, -0.00000743, -0.00000087, -0.00000348, -0.00001309, 0.00000357, 0.00000632, -0.00000585, -0.00000369, 0.00000101, 0.00000334, -0.00000190, -0.00000067, 0.00000091, -0.00000166, 0.00000130, 0.00000742, -0.00000302, -0.00000022, -0.00000103, -0.00000643, 0.00000139, 0.00000260, -0.00000174, 0.00000272, -0.00000182, -0.00000428, 0.00000114, -0.00000384, 0.00000396, 0.00000120, -0.00000081, -0.00000313, -0.00000154, 0.00000156, 0.00000002, 0.00000311, -0.00000058, -0.00000372, 0.00000069, 0.00000269, -0.00000242, 0.00000118, -0.00000078, -0.00000019, -0.00000073, 0.00000133, -0.00000264, },
		{ -0.00000534, 0.00000511, 0.00000245, 0.00000196, -0.00000209, -0.00000076, -0.00000154, 0.00000122, -0.00000179, 0.00000055, -0.00000033, -0.00000090, 0.00000099, -0.00000258, 0.00000178, -0.00000109, 0.00000506, -0.00000385, 0.00000012, -0.00000173, 0.00000064, -0.00000189, 0.00000447, -0.00000165, -0.00000442, 0.00000018, -0.00000224, 0.00000331, -0.00000346, 0.00000394, -0.00000676, 0.00000391, -0.00000307, -0.00000043, -0.00000205, 0.00000297, 0.00000504, -0.00000463, -0.00000093, 0.00000035, -0.00000420, -0.00000015, 0.00001837, -0.00001461, 0.00000999, -0.00001340, 0.00000277, -0.00000627, -0.00001576, 0.00000117, 0.00000190, -0.00000928, -0.00000946, -0.00000875, -0.00001132, -0.00001254, -0.00003307, -0.00003619, -0.00009789, 0.00008117, 0.00009129, 0.00000705, 0.00003799, 0.00005197, 0.00017287, -0.00002559, -0.00008843, 0.00000917, -0.00001513, 0.00001512, 0.00000090, 0.00002645, 0.00002675, -0.00001060, -0.00004775, -0.00000287, -0.00001172, -0.00000476, -0.00000202, -0.00000919, -0.00000307, -0.00000548, -0.00000153, -0.00000469, -0.00001291, 0.00000574, 0.00000440, -0.00000552, -0.00000415, 0.00000298, 0.00000217, -0.00000204, -0.00000048, 0.00000076, -0.00000181, 0.00000247, 0.00000641, -0.00000302, 0.00000000, -0.00000191, -0.00000580, 0.00000190, 0.00000218, -0.00000156, 0.00000267, -0.00000239, -0.00000379, 0.00000083, -0.00000372, 0.00000455, 0.00000051, -0.00000055, -0.00000368, -0.00000052, 0.00000084, 0.00000056, 0.00000292, -0.00000116, -0.00000350, 0.00000143, 0.00000199, -0.00000234, 0.00000134, -0.00000102, 0.00000008, -0.00000095, 0.00000164, -0.00000373, },
		{ -0.00000395, 0.00000507, 0.00000236, 0.00000158, -0.00000237, -0.00000063, -0.00000146, 0.00000142, -0.00000205, 0.00000086, -0.00000089, -0.00000008, 0.00000037, -0.00000217, 0.00000133, -0.00000041, 0.00000480, -0.00000469, 0.00000100, -0.00000215, 0.00000102, -0.00000216, 0.00000525, -0.00000333, -0.00000270, -0.00000150, -0.00000074, 0.00000236, -0.00000260, 0.00000314, -0.00000651, 0.00000436, -0.00000372, -0.00000039, -0.00000220, 0.00000463, 0.00000352, -0.00000424, -0.00000126, 0.00000095, -0.00000516, 0.00000217, 0.00001610, -0.00001417, 0.00001054, -0.00001525, 0.00000583, -0.00001064, -0.00001163, -0.00000140, 0.00000370, -0.00001161, -0.00000687, -0.00001116, -0.00000888, -0.00001713, -0.00002672, -0.00004975, -0.00008255, 0.00009541, 0.00007385, 0.00000923, 0.00003026, 0.00005259, 0.00016673, -0.00001848, -0.00008707, 0.00000944, -0.00001399, 0.00001548, 0.00000268, 0.00002982, 0.00002422, -0.00001506, -0.00004806, -0.00000001, -0.00001448, -0.00000226, -0.00000370, -0.00000803, -0.00000462, -0.00000345, -0.00000245, -0.00000586, -0.00001231, 0.00000750, 0.00000246, -0.00000499, -0.00000463, 0.00000504, 0.00000067, -0.00000184, -0.00000046, 0.00000065, -0.00000190, 0.00000368, 0.00000517, -0.00000275, 0.00000000, -0.00000265, -0.00000508, 0.00000224, 0.00000180, -0.00000134, 0.00000254, -0.00000293, -0.00000320, 0.00000034, -0.00000333, 0.00000488, -0.00000004, -0.00000038, -0.00000410, 0.00000040, 0.00000010, 0.00000124, 0.00000253, -0.00000161, -0.00000322, 0.00000211, 0.00000118, -0.00000207, 0.00000134, -0.00000114, 0.00000026, -0.00000109, 0.00000182, -0.00000473, },
		{ -0.00000249, 0.00000481, 0.00000240, 0.00000103, -0.00000246, -0.00000059, -0.00000127, 0.00000146, -0.00000218, 0.00000105, -0.00000134, 0.00000071, -0.00000031, -0.00000168, 0.00000079, 0.00000047, 0.00000417, -0.00000512, 0.00000161, -0.00000240, 0.00000123, -0.00000216, 0.00000568, -0.00000473, -0.00000110, -0.00000320, 0.00000091, 0.00000117, -0.00000147, 0.00000201, -0.00000582, 0.00000431, -0.00000401, -0.00000059, -0.00000200, 0.00000598, 0.00000201, -0.00000377, -0.00000158, 0.00000145, -0.00000599, 0.00000462, 0.00001318, -0.00001277, 0.00000998, -0.00001584, 0.00000774, -0.00001430, -0.00000770, -0.00000390, 0.00000530, -0.00001369, -0.00000450, -0.00001330, -0.00000691, -0.00002117, -0.00002093, -0.00006357, -0.00006314, 0.00010365, 0.00005865, 0.00001133, 0.00002286, 0.00005302, 0.00016113, -0.00001193, -0.00008530, 0.00000943, -0.00001248, 0.00001540, 0.00000517, 0.00003284, 0.00002125, -0.00002013, -0.00004678, 0.00000161, -0.00001646, -0.00000021, -0.00000515, -0.00000705, -0.00000597, -0.00000149, -0.00000350, -0.00000705, -0.00001122, 0.00000867, 0.00000071, -0.00000442, -0.00000499, 0.00000700, -0.00000103, -0.00000137, -0.00000059, 0.00000055, -0.00000188, 0.00000483, 0.00000375, -0.00000224, -0.00000023, -0.00000320, -0.00000433, 0.00000243, 0.00000149, -0.00000109, 0.00000233, -0.00000343, -0.00000254, -0.00000032, -0.00000267, 0.00000492, -0.00000040, -0.00000035, -0.00000431, 0.00000114, -0.00000057, 0.00000197, 0.00000198, -0.00000193, -0.00000288, 0.00000270, 0.00000031, -0.00000165, 0.00000119, -0.00000111, 0.00000032, -0.00000113, 0.00000182, -0.00000555, },
		{ -0.00000106, 0.00000441, 0.00000252, 0.00000034, -0.00000236, -0.00000066, -0.00000096, 0.00000134, -0.00000216, 0.00000107, -0.00000164, 0.00000141, -0.00000099, -0.00000117, 0.00000023, 0.00000147, 0.00000323, -0.00000515, 0.00000192, -0.00000245, 0.00000122, -0.00000186, 0.00000568, -0.00000574, 0.00000024, -0.00000475, 0.00000255, -0.00000011, -0.00000021, 0.00000067, -0.00000479, 0.00000383, -0.00000395, -0.00000104, -0.00000142, 0.00000695, 0.00000062, -0.00000328, -0.00000184, 0.00000177, -0.00000657, 0.00000699, 0.00000985, -0.00001058, 0.00000840, -0.00001516, 0.00000835, -0.00001696, -0.00000429, -0.00000608, 0.00000646, -0.00001525, -0.00000262, -0.00001486, -0.00000574, -0.00002418, -0.00001647, -0.00007645, -0.00004097, 0.00010636, 0.00004608, 0.00001296, 0.00001599, 0.00005311, 0.00015615, -0.00000598, -0.00008309, 0.00000915, -0.00001065, 0.00001493, 0.00000835, 0.00003551, 0.00001762, -0.00002545, -0.00004408, 0.00000195, -0.00001750, 0.00000119, -0.00000618, -0.00000638, -0.00000697, 0.00000024, -0.00000460, -0.00000826, -0.00000962, 0.00000914, -0.00000065, -0.00000399, -0.00000508, 0.00000872, -0.00000279, -0.00000073, -0.00000083, 0.00000045, -0.00000171, 0.00000586, 0.00000223, -0.00000155, -0.00000067, -0.00000352, -0.00000361, 0.00000250, 0.00000123, -0.00000083, 0.00000202, -0.00000383, -0.00000186, -0.00000107, -0.00000179, 0.00000467, -0.00000056, -0.00000050, -0.00000429, 0.00000164, -0.00000109, 0.00000266, 0.00000133, -0.00000216, -0.00000247, 0.00000316, -0.00000056, -0.00000111, 0.00000091, -0.00000096, 0.00000026, -0.00000104, 0.00000163, -0.00000612, },
		{ 0.00000026, 0.00000395, 0.00000265, -0.00000043, -0.00000210, -0.00000081, -0.00000057, 0.00000109, -0.00000197, 0.00000092, -0.00000173, 0.00000197, -0.00000163, -0.00000069, -0.00000028, 0.00000249, 0.00000207, -0.00000483, 0.00000194, -0.00000228, 0.00000100, -0.00000126, 0.00000523, -0.00000628, 0.00000120, -0.00000601, 0.00000403, -0.00000133, 0.00000104, -0.00000073, -0.00000356, 0.00000301, -0.00000358, -0.00000170, -0.00000049, 0.00000749, -0.00000059, -0.00000282, -0.00000199, 0.00000186, -0.00000677, 0.00000905, 0.00000636, -0.00000786, 0.00000603, -0.00001333, 0.00000760, -0.00001847, -0.00000160, -0.00000775, 0.00000705, -0.00001616, -0.00000140, -0.00001568, -0.00000560, -0.00002583, -0.00001393, -0.00008725, -0.00001746, 0.00010431, 0.00003621, 0.00001392, 0.00000974, 0.00005278, 0.00015184, -0.00000064, -0.00008048, 0.00000866, -0.00000858, 0.00001419, 0.00001217, 0.00003782, 0.00001313, -0.00003058, -0.00004029, 0.00000110, -0.00001756, 0.00000186, -0.00000673, -0.00000611, -0.00000756, 0.00000167, -0.00000570, -0.00000946, -0.00000757, 0.00000887, -0.00000149, -0.00000383, -0.00000478, 0.00001003, -0.00000445, -0.00000001, -0.00000112, 0.00000031, -0.00000135, 0.00000669, 0.00000071, -0.00000073, -0.00000130, -0.00000361, -0.00000297, 0.00000251, 0.00000101, -0.00000054, 0.00000160, -0.00000411, -0.00000124, -0.00000186, -0.00000074, 0.00000416, -0.00000050, -0.00000081, -0.00000404, 0.00000187, -0.00000138, 0.00000322, 0.00000066, -0.00000232, -0.00000199, 0.00000345, -0.00000138, -0.00000051, 0.00000052, -0.00000069, 0.00000008, -0.00000084, 0.00000123, -0.00000640, },
		{ 0.00000140, 0.00000349, 0.00000272, -0.00000121, -0.00000174, -0.00000102, -0.00000012, 0.00000072, -0.00000165, 0.00000060, -0.00000162, 0.00000235, -0.00000217, -0.00000029, -0.00000067, 0.00000343, 0.00000080, -0.00000424, 0.00000171, -0.00000192, 0.00000061, -0.00000040, 0.00000435, -0.00000632, 0.00000170, -0.00000687, 0.00000524, -0.00000237, 0.00000215, -0.00000205, -0.00000228, 0.00000196, -0.00000297, -0.00000254, 0.00000076, 0.00000760, -0.00000155, -0.00000244, -0.00000202, 0.00000168, -0.00000650, 0.00001062, 0.00000300, -0.00000492, 0.00000315, -0.00001059, 0.00000562, -0.00001878, 0.00000027, -0.00000883, 0.00000699, -0.00001633, -0.00000090, -0.00001569, -0.00000655, -0.00002596, -0.00001370, -0.00009506, 0.00000598, 0.00009847, 0.00002887, 0.00001417, 0.00000410, 0.00005204, 0.00014817, 0.00000415, -0.00007752, 0.00000804, -0.00000635, 0.00001329, 0.00001657, 0.00003974, 0.00000760, -0.00003506, -0.00003583, -0.00000071, -0.00001675, 0.00000181, -0.00000677, -0.00000622, -0.00000772, 0.00000275, -0.00000678, -0.00001056, -0.00000519, 0.00000789, -0.00000175, -0.00000401, -0.00000401, 0.00001082, -0.00000588, 0.00000069, -0.00000142, 0.00000011, -0.00000078, 0.00000724, -0.00000073, 0.00000014, -0.00000207, -0.00000347, -0.00000246, 0.00000248, 0.00000082, -0.00000021, 0.00000106, -0.00000421, -0.00000073, -0.00000260, 0.00000040, 0.00000343, -0.00000027, -0.00000127, -0.00000358, 0.00000182, -0.00000142, 0.00000361, 0.00000002, -0.00000243, -0.00000143, 0.00000356, -0.00000208, 0.00000010, 0.00000008, -0.00000033, -0.00000020, -0.00000055, 0.00000065, -0.00000636, },
		{ 0.00000229, 0.00000311, 0.00000269, -0.00000194, -0.00000132, -0.00000125, 0.00000035, 0.00000027, -0.00000122, 0.00000013, -0.00000129, 0.00000251, -0.00000257, -0.00000002, -0.00000088, 0.00000420, -0.00000046, -0.00000349, 0.00000128, -0.00000143, 0.00000011, 0.00000065, 0.00000310, -0.00000587, 0.00000172, -0.00000727, 0.00000607, -0.00000313, 0.00000302, -0.00000317, -0.00000109, 0.00000083, -0.00000222, -0.00000347, 0.00000226, 0.00000730, -0.00000226, -0.00000216, -0.00000192, 0.00000123, -0.00000570, 0.00001153, 0.00000003, -0.00000204, 0.00000009, -0.00000726, 0.00000262, -0.00001797, 0.00000131, -0.00000934, 0.00000634, -0.00001582, -0.00000110, -0.00001497, -0.00000852, -0.00002462, -0.00001590, -0.00009919, 0.00002807, 0.00008990, 0.00002373, 0.00001377, -0.00000099, 0.00005092, 0.00014507, 0.00000846, -0.00007428, 0.00000737, -0.00000407, 0.00001234, 0.00002149, 0.00004117, 0.00000095, -0.00003844, -0.00003120, -0.00000314, -0.00001525, 0.00000116, -0.00000638, -0.00000667, -0.00000750, 0.00000350, -0.00000787, -0.00001147, -0.00000264, 0.00000632, -0.00000146, -0.00000455, -0.00000276, 0.00001102, -0.00000696, 0.00000128, -0.00000169, -0.00000016, 0.00000003, 0.00000748, -0.00000201, 0.00000100, -0.00000290, -0.00000313, -0.00000209, 0.00000246, 0.00000062, 0.00000015, 0.00000039, -0.00000411, -0.00000040, -0.00000320, 0.00000153, 0.00000258, 0.00000010, -0.00000182, -0.00000296, 0.00000153, -0.00000118, 0.00000378, -0.00000053, -0.00000252, -0.00000079, 0.00000346, -0.00000263, 0.00000065, -0.00000037, 0.00000007, -0.00000054, -0.00000019, -0.00000009, -0.00000600, },
		{ 0.00000293, 0.00000283, 0.00000250, -0.00000253, -0.00000091, -0.00000147, 0.00000080, -0.00000020, -0.00000073, -0.00000045, -0.00000077, 0.00000246, -0.00000282, 0.00000012, -0.00000088, 0.00000471, -0.00000159, -0.00000267, 0.00000074, -0.00000087, -0.00000043, 0.00000177, 0.00000158, -0.00000502, 0.00000128, -0.00000720, 0.00000650, -0.00000355, 0.00000357, -0.00000401, -0.00000011, -0.00000025, -0.00000143, -0.00000439, 0.00000391, 0.00000663, -0.00000269, -0.00000199, -0.00000170, 0.00000054, -0.00000438, 0.00001167, -0.00000234, 0.00000048, -0.00000280, -0.00000371, -0.00000108, -0.00001622, 0.00000162, -0.00000935, 0.00000519, -0.00001474, -0.00000185, -0.00001367, -0.00001127, -0.00002207, -0.00002036, -0.00009928, 0.00004772, 0.00007966, 0.00002035, 0.00001287, -0.00000566, 0.00004952, 0.00014248, 0.00001236, -0.00007084, 0.00000672, -0.00000182, 0.00001147, 0.00002690, 0.00004199, -0.00000681, -0.00004039, -0.00002685, -0.00000583, -0.00001333, 0.00000008, -0.00000568, -0.00000733, -0.00000698, 0.00000395, -0.00000901, -0.00001204, -0.00000014, 0.00000436, -0.00000072, -0.00000539, -0.00000106, 0.00001060, -0.00000764, 0.00000170, -0.00000187, -0.00000050, 0.00000106, 0.00000737, -0.00000306, 0.00000175, -0.00000373, -0.00000263, -0.00000187, 0.00000247, 0.00000042, 0.00000055, -0.00000042, -0.00000377, -0.00000032, -0.00000358, 0.00000255, 0.00000168, 0.00000052, -0.00000240, -0.00000225, 0.00000105, -0.00000071, 0.00000371, -0.00000098, -0.00000261, -0.00000008, 0.00000315, -0.00000298, 0.00000111, -0.00000078, 0.00000047, -0.00000091, 0.00000018, -0.00000092, -0.00000535, },
		{ 0.00000331, 0.00000267, 0.00000212, -0.00000295, -0.00000054, -0.00000162, 0.00000118, -0.00000064, -0.00000023, -0.00000110, -0.00000009, 0.00000221, -0.00000292, 0.00000013, -0.00000066, 0.00000493, -0.00000251, -0.00000191, 0.00000019, -0.00000031, -0.00000092, 0.00000287, -0.00000007, -0.00000386, 0.00000047, -0.00000669, 0.00000651, -0.00000363, 0.00000379, -0.00000450, 0.00000056, -0.00000119, -0.00000070, -0.00000521, 0.00000561, 0.00000566, -0.00000287, -0.00000192, -0.00000140, -0.00000034, -0.00000258, 0.00001102, -0.00000396, 0.00000242, -0.00000522, -0.00000032, -0.00000514, -0.00001378, 0.00000135, -0.00000902, 0.00000373, -0.00001327, -0.00000297, -0.00001204, -0.00001449, -0.00001871, -0.00002670, -0.00009533, 0.00006414, 0.00006872, 0.00001825, 0.00001166, -0.00001004, 0.00004793, 0.00014029, 0.00001594, -0.00006725, 0.00000614, 0.00000034, 0.00001075, 0.00003276, 0.00004196, -0.00001550, -0.00004069, -0.00002320, -0.00000839, -0.00001128, -0.00000123, -0.00000483, -0.00000807, -0.00000629, 0.00000420, -0.00001025, -0.00001215, 0.00000208, 0.00000222, 0.00000034, -0.00000642, 0.00000098, 0.00000961, -0.00000787, 0.00000190, -0.00000197, -0.00000090, 0.00000229, 0.00000688, -0.00000383, 0.00000234, -0.00000447, -0.00000204, -0.00000178, 0.00000251, 0.00000022, 0.00000098, -0.00000135, -0.00000321, -0.00000049, -0.00000367, 0.00000337, 0.00000083, 0.00000093, -0.00000293, -0.00000154, 0.00000048, -0.00000005, 0.00000341, -0.00000129, -0.00000269, 0.00000068, 0.00000264, -0.00000313, 0.00000142, -0.00000110, 0.00000082, -0.00000125, 0.00000052, -0.00000178, -0.00000446, },
		{ 0.00000346, 0.00000263, 0.00000155, -0.00000316, -0.00000028, -0.00000169, 0.00000145, -0.00000101, 0.00000022, -0.00000174, 0.00000070, 0.00000177, -0.00000288, 0.00000003, -0.00000023, 0.00000483, -0.00000314, -0.00000129, -0.00000029, 0.00000018, -0.00000128, 0.00000382, -0.00000173, -0.00000252, -0.00000063, -0.00000581, 0.00000616, -0.00000336, 0.00000366, -0.00000462, 0.00000088, -0.00000187, -0.00000012, -0.00000584, 0.00000726, 0.00000448, -0.00000284, -0.00000191, -0.00000108, -0.00000131, -0.00000041, 0.00000960, -0.00000476, 0.00000363, -0.00000692, 0.00000257, -0.00000920, -0.00001093, 0.00000068, -0.00000850, 0.00000214, -0.00001162, -0.00000423, -0.00001036, -0.00001779, -0.00001505, -0.00003430, -0.00008762, 0.00007684, 0.00005788, 0.00001700, 0.00001033, -0.00001428, 0.00004623, 0.00013845, 0.00001926, -0.00006356, 0.00000569, 0.00000239, 0.00001027, 0.00003907, 0.00004082, -0.00002481, -0.00003930, -0.00002055, -0.00001052, -0.00000935, -0.00000257, -0.00000399, -0.00000875, -0.00000554, 0.00000429, -0.00001164, -0.00001169, 0.00000381, 0.00000015, 0.00000151, -0.00000749, 0.00000322, 0.00000812, -0.00000767, 0.00000188, -0.00000197, -0.00000132, 0.00000367, 0.00000605, -0.00000428, 0.00000271, -0.00000505, -0.00000144, -0.00000178, 0.00000256, 0.00000003, 0.00000140, -0.00000235, -0.00000245, -0.00000094, -0.00000345, 0.00000392, 0.00000010, 0.00000127, -0.00000336, -0.00000091, -0.00000011, 0.00000073, 0.00000293, -0.00000148, -0.00000276, 0.00000146, 0.00000197, -0.00000306, 0.00000157, -0.00000130, 0.00000108, -0.00000154, 0.00000079, -0.00000260, -0.00000339, },
		{ 0.00000343, 0.00000267, 0.00000082, -0.00000315, -0.00000013, -0.00000165, 0.00000159, -0.00000127, 0.00000058, -0.00000232, 0.00000152, 0.00000120, -0.00000273, -0.00000016, 0.00000036, 0.00000443, -0.00000346, -0.00000089, -0.00000064, 0.00000052, -0.00000143, 0.00000452, -0.00000326, -0.00000114, -0.00000188, -0.00000466, 0.00000550, -0.00000282, 0.00000324, -0.00000442, 0.00000082, -0.00000225, 0.00000025, -0.00000620, 0.00000873, 0.00000319, -0.00000267, -0.00000193, -0.00000080, -0.00000225, 0.00000198, 0.00000753, -0.00000476, 0.00000405, -0.00000775, 0.00000470, -0.00001292, -0.00000794, -0.00000019, -0.00000797, 0.00000061, -0.00001000, -0.00000542, -0.00000891, -0.00002079, -0.00001164, -0.00004241, -0.00007672, 0.00008564, 0.00004776, 0.00001622, 0.00000904, -0.00001846, 0.00004448, 0.00013691, 0.00002236, -0.00005979, 0.00000536, 0.00000430, 0.00001008, 0.00004580, 0.00003825, -0.00003430, -0.00003634, -0.00001903, -0.00001199, -0.00000777, -0.00000376, -0.00000328, -0.00000926, -0.00000485, 0.00000430, -0.00001318, -0.00001062, 0.00000488, -0.00000161, 0.00000259, -0.00000840, 0.00000549, 0.00000628, -0.00000710, 0.00000164, -0.00000189, -0.00000174, 0.00000515, 0.00000490, -0.00000441, 0.00000282, -0.00000541, -0.00000090, -0.00000180, 0.00000260, -0.00000013, 0.00000178, -0.00000337, -0.00000154, -0.00000162, -0.00000289, 0.00000416, -0.00000045, 0.00000147, -0.00000362, -0.00000044, -0.00000061, 0.00000154, 0.00000231, -0.00000157, -0.00000281, 0.00000222, 0.00000118, -0.00000279, 0.00000155, -0.00000135, 0.00000123, -0.00000174, 0.00000093, -0.00000330, -0.00000224, },
		{ 0.00000328, 0.00000275, -0.00000004, -0.00000292, -0.00000012, -0.00000148, 0.00000158, -0.00000138, 0.00000080, -0.00000278, 0.00000233, 0.00000055, -0.00000250, -0.00000037, 0.00000106, 0.00000377, -0.00000345, -0.00000073, -0.00000080, 0.00000070, -0.00000134, 0.00000489, -0.00000453, 0.00000013, -0.00000315, -0.00000336, 0.00000463, -0.00000207, 0.00000259, -0.00000394, 0.00000041, -0.00000231, 0.00000036, -0.00000624, 0.00000995, 0.00000188, -0.00000240, -0.00000192, -0.00000063, -0.00000304, 0.00000442, 0.00000498, -0.00000406, 0.00000371, -0.00000766, 0.00000588, -0.00001603, -0.00000507, -0.00000108, -0.00000757, -0.00000070, -0.00000857, -0.00000635, -0.00000791, -0.00002316, -0.00000900, -0.00005023, -0.00006341, 0.00009065, 0.00003877, 0.00001560, 0.00000788, -0.00002266, 0.00004270, 0.00013564, 0.00002527, -0.00005594, 0.00000515, 0.00000608, 0.00001024, 0.00005292, 0.00003396, -0.00004344, -0.00003211, -0.00001865, -0.00001268, -0.00000666, -0.00000469, -0.00000278, -0.00000951, -0.00000428, 0.00000424, -0.00001483, -0.00000894, 0.00000517, -0.00000288, 0.00000339, -0.00000899, 0.00000758, 0.00000424, -0.00000624, 0.00000124, -0.00000177, -0.00000210, 0.00000665, 0.00000350, -0.00000423, 0.00000266, -0.00000551, -0.00000050, -0.00000180, 0.00000259, -0.00000020, 0.00000206, -0.00000434, -0.00000056, -0.00000248, -0.00000204, 0.00000406, -0.00000078, 0.00000151, -0.00000368, -0.00000018, -0.00000095, 0.00000228, 0.00000163, -0.00000158, -0.00000281, 0.00000291, 0.00000031, -0.00000236, 0.00000136, -0.00000126, 0.00000124, -0.00000183, 0.00000092, -0.00000382, -0.00000108, },
		{ 0.00000307, 0.00000281, -0.00000097, -0.00000251, -0.00000024, -0.00000120, 0.00000141, -0.00000134, 0.00000086, -0.00000307, 0.00000305, -0.00000014, -0.00000224, -0.00000056, 0.00000178, 0.00000292, -0.00000317, -0.00000082, -0.00000078, 0.00000068, -0.00000099, 0.00000489, -0.00000545, 0.00000117, -0.00000432, -0.00000203, 0.00000366, -0.00000120, 0.00000180, -0.00000326, -0.00000029, -0.00000206, 0.00000020, -0.00000591, 0.00001083, 0.00000065, -0.00000209, -0.00000185, -0.00000065, -0.00000356, 0.00000672, 0.00000215, -0.00000284, 0.00000276, -0.00000672, 0.00000604, -0.00001836, -0.00000250, -0.00000187, -0.00000739, -0.00000169, -0.00000745, -0.00000692, -0.00000749, -0.00002466, -0.00000756, -0.00005699, -0.00004859, 0.00009221, 0.00003114, 0.00001496, 0.00000693, -0.00002690, 0.00004090, 0.00013462, 0.00002800, -0.00005202, 0.00000505, 0.00000775, 0.00001084, 0.00006030, 0.00002767, -0.00005162, -0.00002706, -0.00001925, -0.00001261, -0.00000608, -0.00000533, -0.00000254, -0.00000948, -0.00000387, 0.00000409, -0.00001650, -0.00000673, 0.00000466, -0.00000352, 0.00000376, -0.00000909, 0.00000931, 0.00000217, -0.00000520, 0.00000071, -0.00000162, -0.00000235, 0.00000808, 0.00000192, -0.00000379, 0.00000226, -0.00000532, -0.00000029, -0.00000172, 0.00000248, -0.00000016, 0.00000218, -0.00000519, 0.00000042, -0.00000344, -0.00000095, 0.00000366, -0.00000088, 0.00000137, -0.00000353, -0.00000016, -0.00000105, 0.00000289, 0.00000096, -0.00000156, -0.00000275, 0.00000350, -0.00000057, -0.00000181, 0.00000103, -0.00000102, 0.00000111, -0.00000179, 0.00000073, -0.00000411, 0.00000000, },
		{ 0.00000287, 0.00000278, -0.00000189, -0.00000196, -0.00000047, -0.00000083, 0.00000112, -0.00000114, 0.00000073, -0.00000316, 0.00000363, -0.00000081, -0.00000200, -0.00000066, 0.00000245, 0.00000197, -0.00000267, -0.00000113, -0.00000056, 0.00000050, -0.00000040, 0.00000451, -0.00000597, 0.00000191, -0.00000526, -0.00000077, 0.00000268, -0.00000031, 0.00000096, -0.00000249, -0.00000119, -0.00000154, -0.00000021, -0.00000523, 0.00001132, -0.00000042, -0.00000182, -0.00000167, -0.00000088, -0.00000370, 0.00000867, -0.00000072, -0.00000130, 0.00000138, -0.00000507, 0.00000522, -0.00001982, -0.00000036, -0.00000250, -0.00000746, -0.00000232, -0.00000668, -0.00000710, -0.00000769, -0.00002521, -0.00000760, -0.00006202, -0.00003316, 0.00009082, 0.00002493, 0.00001418, 0.00000621, -0.00003117, 0.00003906, 0.00013388, 0.00003054, -0.00004800, 0.00000505, 0.00000931, 0.00001197, 0.00006776, 0.00001924, -0.00005827, -0.00002171, -0.00002054, -0.00001192, -0.00000596, -0.00000569, -0.00000250, -0.00000921, -0.00000359, 0.00000382, -0.00001807, -0.00000414, 0.00000341, -0.00000349, 0.00000362, -0.00000860, 0.00001051, 0.00000025, -0.00000409, 0.00000013, -0.00000151, -0.00000243, 0.00000934, 0.00000027, -0.00000314, 0.00000163, -0.00000484, -0.00000031, -0.00000150, 0.00000224, 0.00000004, 0.00000210, -0.00000583, 0.00000127, -0.00000439, 0.00000029, 0.00000301, -0.00000077, 0.00000108, -0.00000319, -0.00000038, -0.00000090, 0.00000329, 0.00000035, -0.00000154, -0.00000259, 0.00000393, -0.00000140, -0.00000119, 0.00000060, -0.00000066, 0.00000085, -0.00000164, 0.00000039, -0.00000414, 0.00000095, },
		{ 0.00000273, 0.00000262, -0.00000273, -0.00000133, -0.00000076, -0.00000040, 0.00000072, -0.00000080, 0.00000043, -0.00000302, 0.00000402, -0.00000141, -0.00000179, -0.00000065, 0.00000299, 0.00000103, -0.00000205, -0.00000160, -0.00000021, 0.00000020, 0.00000038, 0.00000378, -0.00000608, 0.00000228, -0.00000592, 0.00000032, 0.00000178, 0.00000053, 0.00000015, -0.00000173, -0.00000220, -0.00000082, -0.00000082, -0.00000420, 0.00001140, -0.00000129, -0.00000160, -0.00000138, -0.00000137, -0.00000340, 0.00001012, -0.00000340, 0.00000032, -0.00000018, -0.00000295, 0.00000356, -0.00002040, 0.00000131, -0.00000293, -0.00000777, -0.00000262, -0.00000623, -0.00000694, -0.00000844, -0.00002483, -0.00000922, -0.00006483, -0.00001801, 0.00008708, 0.00002006, 0.00001321, 0.00000569, -0.00003545, 0.00003713, 0.00013343, 0.00003290, -0.00004389, 0.00000515, 0.00001075, 0.00001377, 0.00007500, 0.00000867, -0.00006290, -0.00001663, -0.00002216, -0.00001081, -0.00000619, -0.00000587, -0.00000258, -0.00000875, -0.00000340, 0.00000332, -0.00001937, -0.00000138, 0.00000157, -0.00000283, 0.00000294, -0.00000747, 0.00001108, -0.00000138, -0.00000302, -0.00000045, -0.00000145, -0.00000230, 0.00001036, -0.00000137, -0.00000238, 0.00000086, -0.00000412, -0.00000058, -0.00000111, 0.00000184, 0.00000043, 0.00000178, -0.00000619, 0.00000192, -0.00000521, 0.00000157, 0.00000218, -0.00000051, 0.00000068, -0.00000271, -0.00000082, -0.00000047, 0.00000344, -0.00000014, -0.00000156, -0.00000231, 0.00000418, -0.00000212, -0.00000058, 0.00000011, -0.00000022, 0.00000049, -0.00000139, -0.00000010, -0.00000390, 0.00000171, },
		},
		{ { 0.00000194, 0.00000096, 0.00000094, -0.00000213, -0.00000248, -0.00000244, 0.00000602, -0.00000116, -0.00000333, -0.00000185, -0.00000179, 0.00000131, 0.00000336, 0.00000266, -0.00000917, 0.00000021, 0.00000147, 0.00000209, -0.00000154, 0.00000063, 0.00000049, -0.00000001, -0.00000182, -0.00000133, 0.00000647, -0.00000005, -0.00000718, 0.00000138, 0.00000187, 0.00000107, 0.00000171, -0.00000098, -0.00000148, -0.00000322, 0.00001011, -0.00000461, -0.00000217, -0.00000062, 0.00000449, 0.00000067, -0.00001360, 0.00000185, -0.00000366, -0.00000539, -0.00000516, -0.00000405, 0.00000518, -0.00000641, -0.00000554, -0.00000397, -0.00001204, -0.00000951, -0.00001468, -0.00001787, -0.00005882, -0.00000125, 0.00007226, 0.00002992, 0.00000817, 0.00001995, 0.00002012, 0.00002234, 0.00006207, -0.00001279, -0.00011173, -0.00001279, 0.00006207, 0.00002234, 0.00002012, 0.00001995, 0.00000817, 0.00002992, 0.00007226, -0.00000125, -0.00005882, -0.00001787, -0.00001468, -0.00000951, -0.00001204, -0.00000397, -0.00000554, -0.00000641, 0.00000518, -0.00000405, -0.00000516, -0.00000539, -0.00000366, 0.00000185, -0.00001360, 0.00000067, 0.00000449, -0.00000062, -0.00000217, -0.00000461, 0.00001011, -0.00000322, -0.00000148, -0.00000098, 0.00000171, 0.00000107, 0.00000187, 0.00000138, -0.00000718, -0.00000005, 0.00000647, -0.00000133, -0.00000182, -0.00000001, 0.00000049, 0.00000063, -0.00000154, 0.00000209, 0.00000147, 0.00000021, -0.00000917, 0.00000266, 0.00000336, 0.00000131, -0.00000179, -0.00000185, -0.00000333, -0.00000116, 0.00000602, -0.00000244, -0.00000248, -0.00000213, 0.00000094, 0.00000096, },
		{ 0.00000182, 0.00000106, 0.00000079, -0.00000167, -0.00000235, -0.00000303, 0.00000514, 0.00000070, -0.00000410, -0.00000142, -0.00000258, 0.00000170, 0.00000230, 0.00000426, -0.00000873, -0.00000125, 0.00000178, 0.00000176, -0.00000106, 0.00000001, 0.00000127, -0.00000028, -0.00000110, -0.00000252, 0.00000639, 0.00000115, -0.00000723, 0.00000022, 0.00000255, 0.00000048, 0.00000253, -0.00000165, -0.00000010, -0.00000507, 0.00000971, -0.00000188, -0.00000406, 0.00000058, 0.00000229, 0.00000329, -0.00001360, -0.00000017, -0.00000256, -0.00000522, -0.00000514, -0.00000471, 0.00000489, -0.00000502, -0.00000655, -0.00000327, -0.00001288, -0.00000774, -0.00001664, -0.00001369, -0.00005685, -0.00001930, 0.00007592, 0.00003611, 0.00000887, 0.00002129, 0.00002117, 0.00002242, 0.00006616, -0.00001083, -0.00011181, -0.00001467, 0.00005814, 0.00002211, 0.00001918, 0.00001875, 0.00000727, 0.00002532, 0.00006644, 0.00001471, -0.00005672, -0.00002394, -0.00001215, -0.00001144, -0.00001121, -0.00000469, -0.00000452, -0.00000762, 0.00000513, -0.00000321, -0.00000519, -0.00000544, -0.00000485, 0.00000344, -0.00001271, -0.00000243, 0.00000647, -0.00000139, -0.00000075, -0.00000642, 0.00000930, -0.00000064, -0.00000306, -0.00000025, 0.00000087, 0.00000165, 0.00000124, 0.00000230, -0.00000681, -0.00000122, 0.00000624, -0.00000005, -0.00000241, 0.00000019, -0.00000014, 0.00000096, -0.00000172, 0.00000216, 0.00000136, 0.00000126, -0.00000886, 0.00000056, 0.00000451, 0.00000086, -0.00000094, -0.00000231, -0.00000251, -0.00000281, 0.00000637, -0.00000147, -0.00000270, -0.00000250, 0.00000087, 0.00000111, },
		{ 0.00000151, 0.00000133, 0.00000050, -0.00000118, -0.00000227, -0.00000330, 0.00000388, 0.00000254, -0.00000461, -0.00000116, -0.00000319, 0.00000189, 0.00000151, 0.00000520, -0.00000752, -0.00000305, 0.00000224, 0.00000121, -0.00000035, -0.00000080, 0.00000205, -0.00000050, -0.00000037, -0.00000353, 0.00000599, 0.00000235, -0.00000698, -0.00000107, 0.00000314, -0.00000000, 0.00000319, -0.00000212, 0.00000090, -0.00000601, 0.00000815, 0.00000151, -0.00000608, 0.00000183, 0.00000033, 0.00000503, -0.00001257, -0.00000262, -0.00000154, -0.00000501, -0.00000502, -0.00000532, 0.00000446, -0.00000373, -0.00000723, -0.00000290, -0.00001336, -0.00000661, -0.00001746, -0.00001190, -0.00005108, -0.00003821, 0.00007639, 0.00004408, 0.00000944, 0.00002271, 0.00002237, 0.00002238, 0.00007039, -0.00000875, -0.00011206, -0.00001646, 0.00005435, 0.00002179, 0.00001830, 0.00001771, 0.00000612, 0.00002207, 0.00005940, 0.00002776, -0.00005085, -0.00003104, -0.00000982, -0.00001297, -0.00001080, -0.00000509, -0.00000377, -0.00000838, 0.00000455, -0.00000211, -0.00000533, -0.00000528, -0.00000615, 0.00000461, -0.00001115, -0.00000561, 0.00000783, -0.00000147, 0.00000003, -0.00000723, 0.00000739, 0.00000242, -0.00000458, 0.00000033, 0.00000014, 0.00000209, 0.00000075, 0.00000291, -0.00000613, -0.00000232, 0.00000576, 0.00000121, -0.00000280, 0.00000024, -0.00000053, 0.00000093, -0.00000161, 0.00000200, 0.00000143, 0.00000191, -0.00000793, -0.00000180, 0.00000551, 0.00000056, -0.00000022, -0.00000264, -0.00000181, -0.00000405, 0.00000610, -0.00000015, -0.00000297, -0.00000272, 0.00000054, 0.00000150, },
		{ 0.00000109, 0.00000169, 0.00000019, -0.00000075, -0.00000219, -0.00000334, 0.00000243, 0.00000415, -0.00000470, -0.00000121, -0.00000349, 0.00000176, 0.00000113, 0.00000542, -0.00000564, -0.00000505, 0.00000274, 0.00000053, 0.00000046, -0.00000162, 0.00000266, -0.00000055, 0.00000028, -0.00000425, 0.00000524, 0.00000352, -0.00000648, -0.00000237, 0.00000352, -0.00000027, 0.00000360, -0.00000228, 0.00000139, -0.00000605, 0.00000565, 0.00000516, -0.00000782, 0.00000273, -0.00000098, 0.00000562, -0.00001049, -0.00000539, -0.00000062, -0.00000484, -0.00000471, -0.00000599, 0.00000406, -0.00000278, -0.00000738, -0.00000306, -0.00001322, -0.00000641, -0.00001684, -0.00001253, -0.00004230, -0.00005641, 0.00007266, 0.00005384, 0.00001004, 0.00002420, 0.00002371, 0.00002224, 0.00007470, -0.00000647, -0.00011254, -0.00001811, 0.00005061, 0.00002146, 0.00001738, 0.00001693, 0.00000468, 0.00001993, 0.00005190, 0.00003756, -0.00004191, -0.00003809, -0.00000847, -0.00001359, -0.00001115, -0.00000494, -0.00000351, -0.00000853, 0.00000338, -0.00000071, -0.00000560, -0.00000488, -0.00000755, 0.00000540, -0.00000915, -0.00000850, 0.00000829, -0.00000074, 0.00000015, -0.00000710, 0.00000465, 0.00000558, -0.00000574, 0.00000060, -0.00000034, 0.00000230, 0.00000051, 0.00000318, -0.00000519, -0.00000329, 0.00000506, 0.00000236, -0.00000292, 0.00000009, -0.00000064, 0.00000059, -0.00000130, 0.00000171, 0.00000161, 0.00000224, -0.00000655, -0.00000414, 0.00000613, 0.00000054, 0.00000025, -0.00000271, -0.00000138, -0.00000478, 0.00000520, 0.00000143, -0.00000326, -0.00000281, -0.00000004, 0.00000208, },
		{ 0.00000069, 0.00000198, -0.00000001, -0.00000046, -0.00000206, -0.00000323, 0.00000098, 0.00000532, -0.00000427, -0.00000161, -0.00000344, 0.00000128, 0.00000117, 0.00000499, -0.00000327, -0.00000701, 0.00000310, -0.00000012, 0.00000120, -0.00000228, 0.00000296, -0.00000036, 0.00000075, -0.00000462, 0.00000417, 0.00000462, -0.00000577, -0.00000356, 0.00000360, -0.00000027, 0.00000372, -0.00000210, 0.00000138, -0.00000530, 0.00000255, 0.00000861, -0.00000886, 0.00000291, -0.00000133, 0.00000494, -0.00000747, -0.00000829, 0.00000012, -0.00000472, -0.00000420, -0.00000675, 0.00000382, -0.00000230, -0.00000690, -0.00000382, -0.00001239, -0.00000721, -0.00001484, -0.00001518, -0.00003172, -0.00007224, 0.00006392, 0.00006516, 0.00001093, 0.00002566, 0.00002520, 0.00002204, 0.00007901, -0.00000389, -0.00011335, -0.00001955, 0.00004685, 0.00002122, 0.00001633, 0.00001648, 0.00000290, 0.00001871, 0.00004446, 0.00004412, -0.00003093, -0.00004397, -0.00000874, -0.00001300, -0.00001241, -0.00000414, -0.00000382, -0.00000804, 0.00000167, 0.00000089, -0.00000597, -0.00000427, -0.00000897, 0.00000582, -0.00000691, -0.00001082, 0.00000773, 0.00000072, -0.00000027, -0.00000626, 0.00000146, 0.00000845, -0.00000628, 0.00000041, -0.00000049, 0.00000223, 0.00000053, 0.00000312, -0.00000407, -0.00000407, 0.00000418, 0.00000333, -0.00000276, -0.00000024, -0.00000052, 0.00000001, -0.00000091, 0.00000140, 0.00000180, 0.00000236, -0.00000493, -0.00000620, 0.00000618, 0.00000092, 0.00000040, -0.00000245, -0.00000126, -0.00000498, 0.00000376, 0.00000314, -0.00000345, -0.00000281, -0.00000080, 0.00000271, },
		{ 0.00000045, 0.00000209, -0.00000002, -0.00000037, -0.00000185, -0.00000307, -0.00000030, 0.00000593, -0.00000329, -0.00000233, -0.00000309, 0.00000049, 0.00000155, 0.00000406, -0.00000069, -0.00000868, 0.00000315, -0.00000058, 0.00000170, -0.00000259, 0.00000283, 0.00000014, 0.00000103, -0.00000462, 0.00000281, 0.00000560, -0.00000489, -0.00000456, 0.00000333, -0.00000000, 0.00000358, -0.00000162, 0.00000095, -0.00000399, -0.00000073, 0.00001138, -0.00000885, 0.00000212, -0.00000058, 0.00000308, -0.00000382, -0.00001105, 0.00000056, -0.00000463, -0.00000351, -0.00000757, 0.00000375, -0.00000230, -0.00000585, -0.00000506, -0.00001101, -0.00000880, -0.00001186, -0.00001907, -0.00002081, -0.00008410, 0.00004980, 0.00007746, 0.00001252, 0.00002698, 0.00002689, 0.00002181, 0.00008327, -0.00000091, -0.00011455, -0.00002072, 0.00004295, 0.00002119, 0.00001502, 0.00001644, 0.00000074, 0.00001821, 0.00003742, 0.00004780, -0.00001906, -0.00004775, -0.00001100, -0.00001114, -0.00001450, -0.00000279, -0.00000460, -0.00000703, -0.00000042, 0.00000252, -0.00000631, -0.00000355, -0.00001032, 0.00000585, -0.00000458, -0.00001242, 0.00000621, 0.00000268, -0.00000094, -0.00000503, -0.00000177, 0.00001066, -0.00000602, -0.00000029, -0.00000031, 0.00000187, 0.00000079, 0.00000279, -0.00000288, -0.00000460, 0.00000315, 0.00000410, -0.00000233, -0.00000071, -0.00000022, -0.00000065, -0.00000060, 0.00000119, 0.00000191, 0.00000239, -0.00000329, -0.00000776, 0.00000555, 0.00000171, 0.00000023, -0.00000189, -0.00000146, -0.00000470, 0.00000195, 0.00000477, -0.00000344, -0.00000279, -0.00000162, 0.00000327, },
		{ 0.00000049, 0.00000194, 0.00000022, -0.00000048, -0.00000155, -0.00000293, -0.00000130, 0.00000594, -0.00000183, -0.00000326, -0.00000254, -0.00000048, 0.00000214, 0.00000286, 0.00000179, -0.00000980, 0.00000273, -0.00000072, 0.00000182, -0.00000245, 0.00000223, 0.00000091, 0.00000112, -0.00000429, 0.00000126, 0.00000639, -0.00000388, -0.00000530, 0.00000272, 0.00000047, 0.00000323, -0.00000093, 0.00000027, -0.00000242, -0.00000377, 0.00001310, -0.00000759, 0.00000031, 0.00000117, 0.00000034, 0.00000003, -0.00001330, 0.00000051, -0.00000446, -0.00000279, -0.00000836, 0.00000377, -0.00000265, -0.00000446, -0.00000651, -0.00000939, -0.00001077, -0.00000849, -0.00002320, -0.00001106, -0.00009077, 0.00003045, 0.00008976, 0.00001538, 0.00002797, 0.00002887, 0.00002154, 0.00008744, 0.00000250, -0.00011619, -0.00002156, 0.00003886, 0.00002141, 0.00001341, 0.00001685, -0.00000177, 0.00001829, 0.00003098, 0.00004908, -0.00000737, -0.00004879, -0.00001521, -0.00000828, -0.00001713, -0.00000118, -0.00000561, -0.00000575, -0.00000258, 0.00000394, -0.00000646, -0.00000289, -0.00001143, 0.00000546, -0.00000225, -0.00001326, 0.00000391, 0.00000481, -0.00000156, -0.00000375, -0.00000467, 0.00001194, -0.00000487, -0.00000147, 0.00000015, 0.00000130, 0.00000121, 0.00000232, -0.00000176, -0.00000481, 0.00000198, 0.00000462, -0.00000168, -0.00000124, 0.00000013, -0.00000126, -0.00000050, 0.00000119, 0.00000187, 0.00000241, -0.00000178, -0.00000868, 0.00000425, 0.00000283, -0.00000018, -0.00000111, -0.00000186, -0.00000410, -0.00000000, 0.00000610, -0.00000310, -0.00000283, -0.00000240, 0.00000362, },
		{ 0.00000088, 0.00000150, 0.00000069, -0.00000075, -0.00000119, -0.00000282, -0.00000199, 0.00000540, -0.00000005, -0.00000423, -0.00000196, -0.00000147, 0.00000274, 0.00000166, 0.00000388, -0.00001015, 0.00000173, -0.00000047, 0.00000150, -0.00000183, 0.00000121, 0.00000187, 0.00000110, -0.00000370, -0.00000035, 0.00000691, -0.00000275, -0.00000578, 0.00000183, 0.00000105, 0.00000281, -0.00000019, -0.00000048, -0.00000086, -0.00000620, 0.00001351, -0.00000506, -0.00000236, 0.00000365, -0.00000283, 0.00000356, -0.00001468, -0.00000018, -0.00000408, -0.00000217, -0.00000894, 0.00000372, -0.00000314, -0.00000303, -0.00000781, -0.00000792, -0.00001261, -0.00000543, -0.00002660, -0.00000373, -0.00009159, 0.00000672, 0.00010069, 0.00002018, 0.00002841, 0.00003126, 0.00002118, 0.00009155, 0.00000635, -0.00011824, -0.00002211, 0.00003459, 0.00002191, 0.00001148, 0.00001766, -0.00000458, 0.00001876, 0.00002526, 0.00004851, 0.00000329, -0.00004689, -0.00002096, -0.00000498, -0.00001979, 0.00000030, -0.00000651, -0.00000454, -0.00000448, 0.00000486, -0.00000623, -0.00000244, -0.00001217, 0.00000460, 0.00000001, -0.00001340, 0.00000114, 0.00000673, -0.00000181, -0.00000269, -0.00000695, 0.00001214, -0.00000289, -0.00000298, 0.00000077, 0.00000063, 0.00000169, 0.00000182, -0.00000081, -0.00000466, 0.00000074, 0.00000491, -0.00000088, -0.00000175, 0.00000043, -0.00000168, -0.00000071, 0.00000143, 0.00000168, 0.00000246, -0.00000050, -0.00000892, 0.00000238, 0.00000411, -0.00000068, -0.00000023, -0.00000234, -0.00000336, -0.00000186, 0.00000694, -0.00000235, -0.00000298, -0.00000303, 0.00000363, },
		{ 0.00000161, 0.00000082, 0.00000132, -0.00000110, -0.00000085, -0.00000274, -0.00000238, 0.00000442, 0.00000183, -0.00000502, -0.00000152, -0.00000229, 0.00000315, 0.00000070, 0.00000534, -0.00000964, 0.00000015, 0.00000018, 0.00000075, -0.00000081, -0.00000010, 0.00000284, 0.00000109, -0.00000297, -0.00000189, 0.00000709, -0.00000152, -0.00000598, 0.00000075, 0.00000162, 0.00000244, 0.00000047, -0.00000112, 0.00000042, -0.00000776, 0.00001255, -0.00000145, -0.00000556, 0.00000640, -0.00000586, 0.00000624, -0.00001485, -0.00000166, -0.00000341, -0.00000181, -0.00000916, 0.00000346, -0.00000355, -0.00000188, -0.00000861, -0.00000700, -0.00001385, -0.00000333, -0.00002843, 0.00000036, -0.00008661, -0.00001990, 0.00010861, 0.00002759, 0.00002808, 0.00003419, 0.00002066, 0.00009565, 0.00001061, -0.00012064, -0.00002244, 0.00003017, 0.00002261, 0.00000932, 0.00001878, -0.00000755, 0.00001940, 0.00002035, 0.00004655, 0.00001236, -0.00004226, -0.00002753, -0.00000195, -0.00002192, 0.00000121, -0.00000693, -0.00000375, -0.00000581, 0.00000506, -0.00000547, -0.00000233, -0.00001242, 0.00000327, 0.00000218, -0.00001296, -0.00000179, 0.00000808, -0.00000146, -0.00000206, -0.00000846, 0.00001128, -0.00000028, -0.00000462, 0.00000135, -0.00000000, 0.00000209, 0.00000142, -0.00000015, -0.00000415, -0.00000055, 0.00000496, -0.00000000, -0.00000214, 0.00000059, -0.00000182, -0.00000126, 0.00000188, 0.00000138, 0.00000252, 0.00000053, -0.00000853, 0.00000013, 0.00000533, -0.00000109, 0.00000056, -0.00000273, -0.00000266, -0.00000340, 0.00000716, -0.00000117, -0.00000325, -0.00000345, 0.00000327, },
		{ 0.00000260, 0.00000002, 0.00000196, -0.00000138, -0.00000061, -0.00000264, -0.00000254, 0.00000318, 0.00000359, -0.00000545, -0.00000138, -0.00000280, 0.00000323, 0.00000013, 0.00000604, -0.00000826, -0.00000190, 0.00000111, -0.00000031, 0.00000047, -0.00000148, 0.00000365, 0.00000120, -0.00000222, -0.00000322, 0.00000686, -0.00000020, -0.00000595, -0.00000041, 0.00000204, 0.00000223, 0.00000095, -0.00000152, 0.00000125, -0.00000836, 0.00001034, 0.00000285, -0.00000880, 0.00000888, -0.00000817, 0.00000760, -0.00001364, -0.00000394, -0.00000240, -0.00000180, -0.00000891, 0.00000285, -0.00000366, -0.00000129, -0.00000862, -0.00000690, -0.00001414, -0.00000258, -0.00002826, 0.00000096, -0.00007661, -0.00004736, 0.00011177, 0.00003814, 0.00002692, 0.00003774, 0.00001991, 0.00009979, 0.00001520, -0.00012329, -0.00002266, 0.00002573, 0.00002340, 0.00000707, 0.00002003, -0.00001048, 0.00001996, 0.00001637, 0.00004356, 0.00001962, -0.00003548, -0.00003395, 0.00000005, -0.00002300, 0.00000117, -0.00000658, -0.00000362, -0.00000635, 0.00000441, -0.00000414, -0.00000262, -0.00001215, 0.00000150, 0.00000420, -0.00001206, -0.00000455, 0.00000861, -0.00000041, -0.00000188, -0.00000918, 0.00000952, 0.00000268, -0.00000610, 0.00000173, -0.00000046, 0.00000230, 0.00000121, 0.00000018, -0.00000330, -0.00000179, 0.00000477, 0.00000088, -0.00000236, 0.00000056, -0.00000165, -0.00000209, 0.00000244, 0.00000106, 0.00000256, 0.00000133, -0.00000763, -0.00000226, 0.00000627, -0.00000124, 0.00000114, -0.00000288, -0.00000219, -0.00000447, 0.00000668, 0.00000040, -0.00000362, -0.00000365, 0.00000254, },
		{ 0.00000372, -0.00000075, 0.00000248, -0.00000148, -0.00000054, -0.00000249, -0.00000254, 0.00000187, 0.00000500, -0.00000535, -0.00000163, -0.00000292, 0.00000289, 0.00000005, 0.00000597, -0.00000616, -0.00000422, 0.00000213, -0.00000149, 0.00000177, -0.00000270, 0.00000409, 0.00000157, -0.00000156, -0.00000421, 0.00000618, 0.00000118, -0.00000572, -0.00000154, 0.00000221, 0.00000227, 0.00000115, -0.00000159, 0.00000157, -0.00000802, 0.00000717, 0.00000735, -0.00001153, 0.00001056, -0.00000929, 0.00000738, -0.00001103, -0.00000691, -0.00000112, -0.00000217, -0.00000819, 0.00000186, -0.00000337, -0.00000139, -0.00000774, -0.00000771, -0.00001337, -0.00000331, -0.00002606, -0.00000155, -0.00006306, -0.00007325, 0.00010856, 0.00005205, 0.00002502, 0.00004190, 0.00001895, 0.00010402, 0.00002006, -0.00012608, -0.00002290, 0.00002140, 0.00002413, 0.00000490, 0.00002120, -0.00001313, 0.00002019, 0.00001341, 0.00003977, 0.00002513, -0.00002737, -0.00003926, 0.00000037, -0.00002266, -0.00000002, -0.00000535, -0.00000427, -0.00000603, 0.00000293, -0.00000231, -0.00000324, -0.00001142, -0.00000057, 0.00000599, -0.00001081, -0.00000692, 0.00000821, 0.00000127, -0.00000207, -0.00000927, 0.00000714, 0.00000562, -0.00000714, 0.00000172, -0.00000062, 0.00000225, 0.00000126, 0.00000016, -0.00000219, -0.00000292, 0.00000433, 0.00000173, -0.00000238, 0.00000032, -0.00000122, -0.00000311, 0.00000297, 0.00000085, 0.00000248, 0.00000197, -0.00000640, -0.00000451, 0.00000669, -0.00000100, 0.00000139, -0.00000271, -0.00000202, -0.00000498, 0.00000555, 0.00000224, -0.00000398, -0.00000367, 0.00000152, },
		{ 0.00000478, -0.00000129, 0.00000272, -0.00000129, -0.00000070, -0.00000225, -0.00000249, 0.00000065, 0.00000590, -0.00000467, -0.00000229, -0.00000266, 0.00000216, 0.00000042, 0.00000526, -0.00000357, -0.00000653, 0.00000299, -0.00000255, 0.00000284, -0.00000351, 0.00000398, 0.00000228, -0.00000108, -0.00000479, 0.00000506, 0.00000258, -0.00000533, -0.00000253, 0.00000204, 0.00000258, 0.00000108, -0.00000134, 0.00000141, -0.00000692, 0.00000347, 0.00001148, -0.00001324, 0.00001102, -0.00000893, 0.00000553, -0.00000724, -0.00001027, 0.00000025, -0.00000279, -0.00000708, 0.00000059, -0.00000272, -0.00000215, -0.00000606, -0.00000929, -0.00001172, -0.00000529, -0.00002225, -0.00000627, -0.00004788, -0.00009512, 0.00009776, 0.00006907, 0.00002278, 0.00004650, 0.00001787, 0.00010831, 0.00002517, -0.00012892, -0.00002328, 0.00001727, 0.00002464, 0.00000299, 0.00002208, -0.00001529, 0.00001986, 0.00001158, 0.00003535, 0.00002913, -0.00001884, -0.00004262, -0.00000135, -0.00002082, -0.00000230, -0.00000330, -0.00000561, -0.00000496, 0.00000082, -0.00000019, -0.00000404, -0.00001040, -0.00000274, 0.00000740, -0.00000927, -0.00000876, 0.00000693, 0.00000338, -0.00000239, -0.00000897, 0.00000451, 0.00000818, -0.00000752, 0.00000121, -0.00000043, 0.00000190, 0.00000157, -0.00000016, -0.00000093, -0.00000383, 0.00000365, 0.00000250, -0.00000219, -0.00000009, -0.00000059, -0.00000416, 0.00000331, 0.00000087, 0.00000222, 0.00000251, -0.00000501, -0.00000640, 0.00000647, -0.00000031, 0.00000129, -0.00000220, -0.00000219, -0.00000496, 0.00000391, 0.00000415, -0.00000422, -0.00000359, 0.00000034, },
		{ 0.00000560, -0.00000147, 0.00000257, -0.00000076, -0.00000108, -0.00000193, -0.00000243, -0.00000035, 0.00000622, -0.00000342, -0.00000327, -0.00000209, 0.00000114, 0.00000110, 0.00000411, -0.00000084, -0.00000848, 0.00000346, -0.00000326, 0.00000345, -0.00000372, 0.00000324, 0.00000333, -0.00000081, -0.00000493, 0.00000355, 0.00000393, -0.00000479, -0.00000331, 0.00000152, 0.00000315, 0.00000077, -0.00000084, 0.00000089, -0.00000533, -0.00000030, 0.00001472, -0.00001354, 0.00001004, -0.00000705, 0.00000227, -0.00000270, -0.00001360, 0.00000141, -0.00000347, -0.00000579, -0.00000077, -0.00000188, -0.00000340, -0.00000386, -0.00001129, -0.00000957, -0.00000799, -0.00001761, -0.00001192, -0.00003319, -0.00011083, 0.00007889, 0.00008832, 0.00002093, 0.00005117, 0.00001690, 0.00011252, 0.00003059, -0.00013181, -0.00002383, 0.00001344, 0.00002484, 0.00000147, 0.00002251, -0.00001678, 0.00001880, 0.00001088, 0.00003040, 0.00003194, -0.00001069, -0.00004351, -0.00000512, -0.00001773, -0.00000537, -0.00000075, -0.00000737, -0.00000345, -0.00000160, 0.00000189, -0.00000474, -0.00000932, -0.00000477, 0.00000829, -0.00000747, -0.00001006, 0.00000493, 0.00000558, -0.00000256, -0.00000855, 0.00000196, 0.00001004, -0.00000708, 0.00000018, 0.00000008, 0.00000129, 0.00000209, -0.00000070, 0.00000035, -0.00000443, 0.00000273, 0.00000317, -0.00000181, -0.00000061, 0.00000011, -0.00000508, 0.00000331, 0.00000119, 0.00000176, 0.00000301, -0.00000360, -0.00000776, 0.00000555, 0.00000077, 0.00000090, -0.00000140, -0.00000263, -0.00000453, 0.00000195, 0.00000590, -0.00000418, -0.00000351, -0.00000084, },
		{ 0.00000602, -0.00000116, 0.00000199, 0.00000009, -0.00000165, -0.00000155, -0.00000239, -0.00000107, 0.00000595, -0.00000173, -0.00000441, -0.00000139, -0.00000000, 0.00000189, 0.00000281, 0.00000170, -0.00000979, 0.00000330, -0.00000345, 0.00000345, -0.00000322, 0.00000188, 0.00000467, -0.00000070, -0.00000468, 0.00000178, 0.00000512, -0.00000411, -0.00000386, 0.00000068, 0.00000389, 0.00000033, -0.00000020, 0.00000018, -0.00000355, -0.00000368, 0.00001666, -0.00001225, 0.00000765, -0.00000391, -0.00000192, 0.00000199, -0.00001638, 0.00000201, -0.00000393, -0.00000463, -0.00000194, -0.00000110, -0.00000479, -0.00000156, -0.00001322, -0.00000750, -0.00001071, -0.00001310, -0.00001707, -0.00002097, -0.00011882, 0.00005231, 0.00010830, 0.00002048, 0.00005543, 0.00001637, 0.00011646, 0.00003645, -0.00013480, -0.00002454, 0.00000987, 0.00002471, 0.00000038, 0.00002242, -0.00001755, 0.00001698, 0.00001124, 0.00002508, 0.00003384, -0.00000351, -0.00004177, -0.00001057, -0.00001395, -0.00000870, 0.00000184, -0.00000912, -0.00000190, -0.00000389, 0.00000356, -0.00000507, -0.00000845, -0.00000637, 0.00000850, -0.00000541, -0.00001086, 0.00000250, 0.00000751, -0.00000229, -0.00000829, -0.00000018, 0.00001100, -0.00000579, -0.00000130, 0.00000085, 0.00000050, 0.00000271, -0.00000132, 0.00000151, -0.00000464, 0.00000159, 0.00000372, -0.00000129, -0.00000116, 0.00000077, -0.00000573, 0.00000290, 0.00000184, 0.00000111, 0.00000346, -0.00000227, -0.00000851, 0.00000399, 0.00000213, 0.00000034, -0.00000044, -0.00000320, -0.00000385, -0.00000005, 0.00000726, -0.00000374, -0.00000354, -0.00000189, },
		{ 0.00000593, -0.00000035, 0.00000103, 0.00000118, -0.00000227, -0.00000121, -0.00000236, -0.00000151, 0.00000520, 0.00000020, -0.00000550, -0.00000076, -0.00000105, 0.00000256, 0.00000165, 0.00000371, -0.00001022, 0.00000240, -0.00000303, 0.00000278, -0.00000205, 0.00000000, 0.00000613, -0.00000067, -0.00000412, -0.00000013, 0.00000605, -0.00000329, -0.00000416, -0.00000042, 0.00000469, -0.00000013, 0.00000043, -0.00000054, -0.00000187, -0.00000633, 0.00001708, -0.00000941, 0.00000413, 0.00000004, -0.00000637, 0.00000612, -0.00001807, 0.00000173, -0.00000389, -0.00000388, -0.00000262, -0.00000070, -0.00000597, 0.00000036, -0.00001452, -0.00000608, -0.00001272, -0.00000969, -0.00002046, -0.00001274, -0.00011847, 0.00001938, 0.00012688, 0.00002266, 0.00005867, 0.00001670, 0.00011985, 0.00004295, -0.00013803, -0.00002530, 0.00000650, 0.00002429, -0.00000031, 0.00002185, -0.00001765, 0.00001449, 0.00001245, 0.00001962, 0.00003499, 0.00000239, -0.00003768, -0.00001693, -0.00001025, -0.00001161, 0.00000392, -0.00001040, -0.00000076, -0.00000561, 0.00000445, -0.00000475, -0.00000805, -0.00000733, 0.00000793, -0.00000307, -0.00001127, -0.00000004, 0.00000881, -0.00000135, -0.00000837, -0.00000175, 0.00001098, -0.00000375, -0.00000307, 0.00000170, -0.00000032, 0.00000332, -0.00000189, 0.00000242, -0.00000442, 0.00000027, 0.00000411, -0.00000066, -0.00000165, 0.00000127, -0.00000599, 0.00000203, 0.00000276, 0.00000034, 0.00000384, -0.00000106, -0.00000865, 0.00000193, 0.00000354, -0.00000021, 0.00000050, -0.00000374, -0.00000314, -0.00000186, 0.00000802, -0.00000282, -0.00000372, -0.00000268, },
		{ 0.00000532, 0.00000091, -0.00000019, 0.00000234, -0.00000281, -0.00000100, -0.00000230, -0.00000171, 0.00000411, 0.00000213, -0.00000630, -0.00000039, -0.00000182, 0.00000292, 0.00000086, 0.00000498, -0.00000966, 0.00000075, -0.00000202, 0.00000151, -0.00000034, -0.00000216, 0.00000751, -0.00000057, -0.00000338, -0.00000198, 0.00000660, -0.00000230, -0.00000425, -0.00000167, 0.00000541, -0.00000046, 0.00000092, -0.00000109, -0.00000053, -0.00000800, 0.00001597, -0.00000532, -0.00000002, 0.00000413, -0.00001034, 0.00000903, -0.00001825, 0.00000031, -0.00000314, -0.00000379, -0.00000256, -0.00000091, -0.00000660, 0.00000148, -0.00001475, -0.00000583, -0.00001342, -0.00000810, -0.00002118, -0.00000932, -0.00011014, -0.00001757, 0.00014152, 0.00002872, 0.00006037, 0.00001828, 0.00012242, 0.00005032, -0.00014166, -0.00002596, 0.00000318, 0.00002372, -0.00000073, 0.00002094, -0.00001725, 0.00001154, 0.00001421, 0.00001433, 0.00003542, 0.00000704, -0.00003188, -0.00002325, -0.00000750, -0.00001345, 0.00000499, -0.00001079, -0.00000041, -0.00000642, 0.00000431, -0.00000361, -0.00000826, -0.00000751, 0.00000655, -0.00000052, -0.00001139, -0.00000240, 0.00000922, 0.00000036, -0.00000883, -0.00000267, 0.00001005, -0.00000117, -0.00000489, 0.00000246, -0.00000104, 0.00000379, -0.00000230, 0.00000300, -0.00000378, -0.00000115, 0.00000432, 0.00000003, -0.00000203, 0.00000155, -0.00000584, 0.00000077, 0.00000383, -0.00000040, 0.00000407, 0.00000004, -0.00000828, -0.00000039, 0.00000477, -0.00000054, 0.00000126, -0.00000409, -0.00000258, -0.00000326, 0.00000807, -0.00000140, -0.00000407, -0.00000316, },
		{ 0.00000424, 0.00000246, -0.00000147, 0.00000338, -0.00000311, -0.00000101, -0.00000216, -0.00000178, 0.00000288, 0.00000382, -0.00000661, -0.00000044, -0.00000215, 0.00000281, 0.00000061, 0.00000539, -0.00000814, -0.00000150, -0.00000059, -0.00000018, 0.00000166, -0.00000431, 0.00000855, -0.00000025, -0.00000262, -0.00000360, 0.00000668, -0.00000115, -0.00000416, -0.00000294, 0.00000593, -0.00000058, 0.00000118, -0.00000136, 0.00000033, -0.00000864, 0.00001351, -0.00000047, -0.00000413, 0.00000765, -0.00001313, 0.00001018, -0.00001664, -0.00000233, -0.00000162, -0.00000448, -0.00000163, -0.00000188, -0.00000647, 0.00000152, -0.00001365, -0.00000698, -0.00001251, -0.00000871, -0.00001888, -0.00001072, -0.00009524, -0.00005550, 0.00014956, 0.00003968, 0.00006017, 0.00002137, 0.00012396, 0.00005872, -0.00014584, -0.00002637, -0.00000027, 0.00002316, -0.00000107, 0.00001989, -0.00001661, 0.00000844, 0.00001610, 0.00000959, 0.00003501, 0.00001071, -0.00002522, -0.00002848, -0.00000647, -0.00001370, 0.00000468, -0.00001001, -0.00000111, -0.00000612, 0.00000303, -0.00000163, -0.00000912, -0.00000690, 0.00000445, 0.00000214, -0.00001128, -0.00000436, 0.00000861, 0.00000278, -0.00000959, -0.00000305, 0.00000843, 0.00000163, -0.00000647, 0.00000293, -0.00000151, 0.00000401, -0.00000247, 0.00000320, -0.00000277, -0.00000258, 0.00000431, 0.00000076, -0.00000227, 0.00000158, -0.00000530, -0.00000075, 0.00000488, -0.00000096, 0.00000405, 0.00000108, -0.00000752, -0.00000272, 0.00000557, -0.00000050, 0.00000171, -0.00000410, -0.00000233, -0.00000411, 0.00000737, 0.00000043, -0.00000453, -0.00000335, },
		{ 0.00000282, 0.00000409, -0.00000259, 0.00000408, -0.00000301, -0.00000132, -0.00000189, -0.00000179, 0.00000169, 0.00000505, -0.00000629, -0.00000098, -0.00000201, 0.00000219, 0.00000091, 0.00000498, -0.00000584, -0.00000411, 0.00000101, -0.00000200, 0.00000363, -0.00000612, 0.00000901, 0.00000042, -0.00000197, -0.00000486, 0.00000623, 0.00000015, -0.00000393, -0.00000414, 0.00000613, -0.00000040, 0.00000115, -0.00000131, 0.00000067, -0.00000833, 0.00001010, 0.00000451, -0.00000753, 0.00000995, -0.00001420, 0.00000928, -0.00001325, -0.00000604, 0.00000058, -0.00000590, 0.00000011, -0.00000357, -0.00000556, 0.00000045, -0.00001120, -0.00000951, -0.00001003, -0.00001143, -0.00001383, -0.00001609, -0.00007604, -0.00009095, 0.00014857, 0.00005606, 0.00005810, 0.00002602, 0.00012446, 0.00006822, -0.00015067, -0.00002643, -0.00000396, 0.00002275, -0.00000149, 0.00001893, -0.00001603, 0.00000553, 0.00001772, 0.00000577, 0.00003358, 0.00001381, -0.00001862, -0.00003172, -0.00000771, -0.00001211, 0.00000291, -0.00000804, -0.00000286, -0.00000472, 0.00000071, 0.00000104, -0.00001049, -0.00000564, 0.00000185, 0.00000472, -0.00001094, -0.00000581, 0.00000702, 0.00000569, -0.00001043, -0.00000308, 0.00000641, 0.00000429, -0.00000756, 0.00000296, -0.00000164, 0.00000393, -0.00000235, 0.00000304, -0.00000148, -0.00000390, 0.00000404, 0.00000151, -0.00000235, 0.00000139, -0.00000448, -0.00000234, 0.00000568, -0.00000118, 0.00000371, 0.00000212, -0.00000653, -0.00000482, 0.00000575, 0.00000002, 0.00000176, -0.00000373, -0.00000244, -0.00000437, 0.00000599, 0.00000252, -0.00000498, -0.00000331, },
		{ 0.00000125, 0.00000556, -0.00000331, 0.00000429, -0.00000243, -0.00000194, -0.00000150, -0.00000181, 0.00000069, 0.00000570, -0.00000532, -0.00000198, -0.00000143, 0.00000113, 0.00000170, 0.00000393, -0.00000306, -0.00000672, 0.00000245, -0.00000363, 0.00000523, -0.00000730, 0.00000871, 0.00000151, -0.00000150, -0.00000564, 0.00000523, 0.00000156, -0.00000359, -0.00000516, 0.00000597, 0.00000007, 0.00000085, -0.00000095, 0.00000057, -0.00000728, 0.00000619, 0.00000899, -0.00000962, 0.00001055, -0.00001327, 0.00000635, -0.00000838, -0.00001044, 0.00000315, -0.00000781, 0.00000242, -0.00000575, -0.00000406, -0.00000154, -0.00000771, -0.00001304, -0.00000640, -0.00001570, -0.00000688, -0.00002395, -0.00005534, -0.00012046, 0.00013679, 0.00007764, 0.00005468, 0.00003196, 0.00012409, 0.00007874, -0.00015608, -0.00002613, -0.00000795, 0.00002259, -0.00000215, 0.00001825, -0.00001573, 0.00000313, 0.00001869, 0.00000317, 0.00003097, 0.00001676, -0.00001290, -0.00003235, -0.00001137, -0.00000880, -0.00000012, -0.00000509, -0.00000545, -0.00000248, -0.00000233, 0.00000406, -0.00001209, -0.00000401, -0.00000092, 0.00000695, -0.00001030, -0.00000676, 0.00000465, 0.00000875, -0.00001105, -0.00000303, 0.00000432, 0.00000649, -0.00000793, 0.00000246, -0.00000140, 0.00000353, -0.00000197, 0.00000259, -0.00000005, -0.00000497, 0.00000347, 0.00000227, -0.00000229, 0.00000104, -0.00000351, -0.00000380, 0.00000606, -0.00000094, 0.00000299, 0.00000318, -0.00000544, -0.00000647, 0.00000521, 0.00000103, 0.00000145, -0.00000300, -0.00000289, -0.00000411, 0.00000412, 0.00000464, -0.00000526, -0.00000317, },
		{ -0.00000027, 0.00000664, -0.00000345, 0.00000391, -0.00000136, -0.00000283, -0.00000101, -0.00000187, -0.00000005, 0.00000574, -0.00000376, -0.00000331, -0.00000056, -0.00000021, 0.00000276, 0.00000251, -0.00000019, -0.00000895, 0.00000341, -0.00000478, 0.00000618, -0.00000762, 0.00000756, 0.00000302, -0.00000125, -0.00000591, 0.00000375, 0.00000298, -0.00000314, -0.00000595, 0.00000543, 0.00000081, 0.00000033, -0.00000039, 0.00000015, -0.00000580, 0.00000231, 0.00001238, -0.00000996, 0.00000925, -0.00001041, 0.00000178, -0.00000262, -0.00001496, 0.00000562, -0.00000982, 0.00000491, -0.00000803, -0.00000235, -0.00000402, -0.00000370, -0.00001694, -0.00000235, -0.00002062, 0.00000071, -0.00003237, -0.00003608, -0.00014110, 0.00011347, 0.00010327, 0.00005098, 0.00003857, 0.00012330, 0.00008999, -0.00016189, -0.00002559, -0.00001218, 0.00002263, -0.00000307, 0.00001794, -0.00001589, 0.00000142, 0.00001880, 0.00000197, 0.00002714, 0.00001986, -0.00000860, -0.00003017, -0.00001716, -0.00000427, -0.00000388, -0.00000165, -0.00000842, 0.00000016, -0.00000560, 0.00000699, -0.00001355, -0.00000237, -0.00000348, 0.00000856, -0.00000926, -0.00000730, 0.00000180, 0.00001153, -0.00001112, -0.00000316, 0.00000250, 0.00000794, -0.00000748, 0.00000142, -0.00000082, 0.00000288, -0.00000140, 0.00000196, 0.00000136, -0.00000569, 0.00000257, 0.00000303, -0.00000211, 0.00000060, -0.00000253, -0.00000493, 0.00000587, -0.00000020, 0.00000193, 0.00000424, -0.00000436, -0.00000757, 0.00000395, 0.00000244, 0.00000086, -0.00000200, -0.00000358, -0.00000348, 0.00000199, 0.00000653, -0.00000521, -0.00000304, },
		{ -0.00000158, 0.00000716, -0.00000293, 0.00000294, 0.00000011, -0.00000387, -0.00000052, -0.00000196, -0.00000050, 0.00000521, -0.00000179, -0.00000476, 0.00000038, -0.00000160, 0.00000384, 0.00000105, 0.00000237, -0.00001044, 0.00000361, -0.00000521, 0.00000628, -0.00000697, 0.00000561, 0.00000484, -0.00000117, -0.00000572, 0.00000190, 0.00000431, -0.00000257, -0.00000649, 0.00000456, 0.00000171, -0.00000029, 0.00000026, -0.00000040, -0.00000419, -0.00000108, 0.00001427, -0.00000839, 0.00000617, -0.00000602, -0.00000377, 0.00000325, -0.00001888, 0.00000746, -0.00001144, 0.00000709, -0.00000991, -0.00000090, -0.00000643, 0.00000013, -0.00002044, 0.00000125, -0.00002511, 0.00000754, -0.00003938, -0.00002091, -0.00015085, 0.00007913, 0.00013081, 0.00004858, 0.00004497, 0.00012273, 0.00010153, -0.00016776, -0.00002507, -0.00001644, 0.00002276, -0.00000419, 0.00001799, -0.00001652, 0.00000045, 0.00001799, 0.00000213, 0.00002221, 0.00002324, -0.00000592, -0.00002546, -0.00002436, 0.00000067, -0.00000762, 0.00000164, -0.00001122, 0.00000264, -0.00000856, 0.00000932, -0.00001445, -0.00000111, -0.00000544, 0.00000926, -0.00000771, -0.00000761, -0.00000114, 0.00001362, -0.00001037, -0.00000369, 0.00000116, 0.00000851, -0.00000620, -0.00000006, -0.00000002, 0.00000208, -0.00000076, 0.00000128, 0.00000261, -0.00000597, 0.00000137, 0.00000377, -0.00000185, 0.00000016, -0.00000170, -0.00000559, 0.00000506, 0.00000099, 0.00000063, 0.00000525, -0.00000331, -0.00000808, 0.00000207, 0.00000405, 0.00000017, -0.00000091, -0.00000434, -0.00000268, -0.00000010, 0.00000794, -0.00000469, -0.00000305, },
		{ -0.00000254, 0.00000700, -0.00000175, 0.00000148, 0.00000183, -0.00000488, -0.00000015, -0.00000202, -0.00000071, 0.00000426, 0.00000034, -0.00000606, 0.00000115, -0.00000278, 0.00000467, -0.00000012, 0.00000429, -0.00001092, 0.00000290, -0.00000481, 0.00000547, -0.00000542, 0.00000303, 0.00000676, -0.00000115, -0.00000515, -0.00000016, 0.00000541, -0.00000185, -0.00000680, 0.00000343, 0.00000265, -0.00000088, 0.00000087, -0.00000090, -0.00000274, -0.00000363, 0.00001448, -0.00000505, 0.00000174, -0.00000078, -0.00000944, 0.00000833, -0.00002148, 0.00000812, -0.00001222, 0.00000847, -0.00001092, -0.00000021, -0.00000821, 0.00000310, -0.00002275, 0.00000354, -0.00002814, 0.00001228, -0.00004331, -0.00001177, -0.00014900, 0.00003572, 0.00015723, 0.00004940, 0.00005011, 0.00012313, 0.00011281, -0.00017326, -0.00002493, -0.00002045, 0.00002273, -0.00000531, 0.00001824, -0.00001751, 0.00000010, 0.00001640, 0.00000341, 0.00001650, 0.00002672, -0.00000467, -0.00001892, -0.00003193, 0.00000501, -0.00001052, 0.00000408, -0.00001322, 0.00000439, -0.00001066, 0.00001059, -0.00001444, -0.00000058, -0.00000649, 0.00000888, -0.00000560, -0.00000782, -0.00000384, 0.00001466, -0.00000863, -0.00000470, 0.00000041, 0.00000816, -0.00000423, -0.00000180, 0.00000084, 0.00000128, -0.00000018, 0.00000070, 0.00000358, -0.00000577, -0.00000009, 0.00000443, -0.00000152, -0.00000022, -0.00000109, -0.00000570, 0.00000367, 0.00000251, -0.00000076, 0.00000608, -0.00000228, -0.00000806, -0.00000021, 0.00000562, -0.00000040, 0.00000008, -0.00000497, -0.00000195, -0.00000190, 0.00000865, -0.00000362, -0.00000326, },
		{ -0.00000310, 0.00000619, -0.00000002, -0.00000026, 0.00000354, -0.00000566, -0.00000002, -0.00000198, -0.00000077, 0.00000308, 0.00000236, -0.00000695, 0.00000152, -0.00000355, 0.00000503, -0.00000074, 0.00000531, -0.00001029, 0.00000128, -0.00000362, 0.00000387, -0.00000316, 0.00000012, 0.00000853, -0.00000104, -0.00000434, -0.00000221, 0.00000614, -0.00000096, -0.00000689, 0.00000215, 0.00000351, -0.00000134, 0.00000131, -0.00000121, -0.00000165, -0.00000516, 0.00001304, -0.00000033, -0.00000336, 0.00000444, -0.00001426, 0.00001179, -0.00002219, 0.00000722, -0.00001180, 0.00000869, -0.00001069, -0.00000066, -0.00000887, 0.00000464, -0.00002328, 0.00000388, -0.00002894, 0.00001403, -0.00004307, -0.00000955, -0.00013624, -0.00001350, 0.00017888, 0.00005543, 0.00005302, 0.00012524, 0.00012334, -0.00017795, -0.00002554, -0.00002387, 0.00002224, -0.00000620, 0.00001848, -0.00001863, 0.00000012, 0.00001432, 0.00000541, 0.00001051, 0.00002996, -0.00000435, -0.00001153, -0.00003868, 0.00000774, -0.00001181, 0.00000508, -0.00001394, 0.00000495, -0.00001148, 0.00001050, -0.00001329, -0.00000096, -0.00000644, 0.00000737, -0.00000294, -0.00000804, -0.00000601, 0.00001447, -0.00000590, -0.00000614, 0.00000022, 0.00000704, -0.00000182, -0.00000354, 0.00000157, 0.00000062, 0.00000023, 0.00000030, 0.00000418, -0.00000510, -0.00000171, 0.00000495, -0.00000113, -0.00000050, -0.00000078, -0.00000529, 0.00000183, 0.00000414, -0.00000201, 0.00000659, -0.00000121, -0.00000763, -0.00000266, 0.00000687, -0.00000064, 0.00000080, -0.00000531, -0.00000147, -0.00000320, 0.00000857, -0.00000200, -0.00000367, },
		{ -0.00000330, 0.00000481, 0.00000203, -0.00000201, 0.00000499, -0.00000600, -0.00000025, -0.00000179, -0.00000079, 0.00000188, 0.00000401, -0.00000722, 0.00000134, -0.00000376, 0.00000478, -0.00000070, 0.00000537, -0.00000858, -0.00000111, -0.00000185, 0.00000171, -0.00000052, -0.00000275, 0.00000985, -0.00000067, -0.00000347, -0.00000407, 0.00000638, 0.00000013, -0.00000682, 0.00000084, 0.00000415, -0.00000157, 0.00000151, -0.00000125, -0.00000102, -0.00000565, 0.00001024, 0.00000515, -0.00000831, 0.00000874, -0.00001739, 0.00001298, -0.00002064, 0.00000460, -0.00001005, 0.00000757, -0.00000905, -0.00000244, -0.00000817, 0.00000445, -0.00002173, 0.00000198, -0.00002719, 0.00001241, -0.00003838, -0.00001404, -0.00011465, -0.00006425, 0.00019191, 0.00006834, 0.00005302, 0.00012954, 0.00013275, -0.00018152, -0.00002722, -0.00002642, 0.00002102, -0.00000659, 0.00001843, -0.00001960, 0.00000016, 0.00001217, 0.00000758, 0.00000484, 0.00003244, -0.00000427, -0.00000442, -0.00004345, 0.00000802, -0.00001093, 0.00000428, -0.00001309, 0.00000409, -0.00001082, 0.00000893, -0.00001099, -0.00000229, -0.00000530, 0.00000483, 0.00000010, -0.00000830, -0.00000749, 0.00001301, -0.00000239, -0.00000780, 0.00000042, 0.00000537, 0.00000073, -0.00000500, 0.00000198, 0.00000023, 0.00000036, 0.00000017, 0.00000440, -0.00000403, -0.00000337, 0.00000526, -0.00000065, -0.00000068, -0.00000074, -0.00000443, -0.00000024, 0.00000561, -0.00000289, 0.00000666, -0.00000002, -0.00000695, -0.00000499, 0.00000757, -0.00000038, 0.00000110, -0.00000523, -0.00000137, -0.00000388, 0.00000768, 0.00000006, -0.00000419, },
		{ -0.00000324, 0.00000305, 0.00000413, -0.00000349, 0.00000592, -0.00000575, -0.00000090, -0.00000139, -0.00000085, 0.00000084, 0.00000510, -0.00000674, 0.00000056, -0.00000338, 0.00000391, 0.00000002, 0.00000453, -0.00000602, -0.00000395, 0.00000019, -0.00000065, 0.00000208, -0.00000516, 0.00001044, 0.00000010, -0.00000268, -0.00000555, 0.00000603, 0.00000140, -0.00000662, -0.00000039, 0.00000450, -0.00000152, 0.00000144, -0.00000101, -0.00000086, -0.00000524, 0.00000653, 0.00001066, -0.00001228, 0.00001133, -0.00001822, 0.00001160, -0.00001684, 0.00000040, -0.00000709, 0.00000522, -0.00000610, -0.00000543, -0.00000617, 0.00000257, -0.00001821, -0.00000198, -0.00002308, 0.00000773, -0.00002982, -0.00002388, -0.00008746, -0.00011173, 0.00019282, 0.00008911, 0.00005000, 0.00013616, 0.00014099, -0.00018388, -0.00003010, -0.00002793, 0.00001890, -0.00000631, 0.00001788, -0.00002013, -0.00000011, 0.00001038, 0.00000940, 0.00000009, 0.00003363, -0.00000370, 0.00000138, -0.00004535, 0.00000539, -0.00000775, 0.00000165, -0.00001070, 0.00000183, -0.00000878, 0.00000607, -0.00000775, -0.00000438, -0.00000326, 0.00000155, 0.00000326, -0.00000850, -0.00000829, 0.00001046, 0.00000157, -0.00000936, 0.00000076, 0.00000348, 0.00000307, -0.00000594, 0.00000193, 0.00000019, 0.00000020, 0.00000031, 0.00000427, -0.00000270, -0.00000491, 0.00000526, -0.00000008, -0.00000076, -0.00000092, -0.00000329, -0.00000227, 0.00000665, -0.00000322, 0.00000618, 0.00000132, -0.00000615, -0.00000696, 0.00000752, 0.00000044, 0.00000095, -0.00000471, -0.00000169, -0.00000393, 0.00000610, 0.00000237, -0.00000470, },
		{ -0.00000305, 0.00000113, 0.00000598, -0.00000440, 0.00000614, -0.00000482, -0.00000197, -0.00000081, -0.00000101, 0.00000008, 0.00000552, -0.00000553, -0.00000077, -0.00000250, 0.00000253, 0.00000126, 0.00000303, -0.00000297, -0.00000683, 0.00000210, -0.00000281, 0.00000424, -0.00000677, 0.00001009, 0.00000134, -0.00000205, -0.00000654, 0.00000509, 0.00000283, -0.00000632, -0.00000146, 0.00000449, -0.00000119, 0.00000114, -0.00000053, -0.00000106, -0.00000419, 0.00000245, 0.00001545, -0.00001458, 0.00001169, -0.00001648, 0.00000774, -0.00001117, -0.00000491, -0.00000335, 0.00000201, -0.00000223, -0.00000927, -0.00000322, -0.00000061, -0.00001321, -0.00000741, -0.00001732, 0.00000089, -0.00001871, -0.00003688, -0.00005857, -0.00015119, 0.00017901, 0.00011762, 0.00004462, 0.00014470, 0.00014839, -0.00018520, -0.00003410, -0.00002842, 0.00001585, -0.00000529, 0.00001671, -0.00002005, -0.00000097, 0.00000928, 0.00001041, -0.00000327, 0.00003314, -0.00000207, 0.00000510, -0.00004394, -0.00000014, -0.00000252, -0.00000245, -0.00000712, -0.00000148, -0.00000570, 0.00000231, -0.00000399, -0.00000686, -0.00000070, -0.00000204, 0.00000616, -0.00000847, -0.00000853, 0.00000715, 0.00000550, -0.00001045, 0.00000094, 0.00000170, 0.00000488, -0.00000616, 0.00000135, 0.00000051, -0.00000026, 0.00000067, 0.00000390, -0.00000126, -0.00000618, 0.00000490, 0.00000061, -0.00000079, -0.00000124, -0.00000204, -0.00000401, 0.00000704, -0.00000286, 0.00000513, 0.00000281, -0.00000534, -0.00000841, 0.00000668, 0.00000180, 0.00000040, -0.00000379, -0.00000238, -0.00000347, 0.00000403, 0.00000467, -0.00000501, },
		{ -0.00000290, -0.00000069, 0.00000732, -0.00000457, 0.00000558, -0.00000325, -0.00000336, -0.00000012, -0.00000126, -0.00000035, 0.00000528, -0.00000369, -0.00000245, -0.00000132, 0.00000086, 0.00000277, 0.00000120, 0.00000014, -0.00000932, 0.00000348, -0.00000438, 0.00000559, -0.00000729, 0.00000873, 0.00000303, -0.00000164, -0.00000697, 0.00000358, 0.00000432, -0.00000592, -0.00000231, 0.00000412, -0.00000064, 0.00000069, 0.00000010, -0.00000148, -0.00000280, -0.00000147, 0.00001893, -0.00001475, 0.00000964, -0.00001234, 0.00000191, -0.00000436, -0.00001058, 0.00000052, -0.00000147, 0.00000196, -0.00001334, 0.00000007, -0.00000440, -0.00000757, -0.00001333, -0.00001102, -0.00000674, -0.00000688, -0.00005033, -0.00003201, -0.00017859, 0.00014930, 0.00015241, 0.00003839, 0.00015422, 0.00015567, -0.00018599, -0.00003889, -0.00002814, 0.00001208, -0.00000367, 0.00001498, -0.00001932, -0.00000253, 0.00000908, 0.00001037, -0.00000499, 0.00003078, 0.00000093, 0.00000642, -0.00003929, -0.00000803, 0.00000403, -0.00000732, -0.00000298, -0.00000525, -0.00000217, -0.00000172, -0.00000029, -0.00000924, 0.00000192, -0.00000542, 0.00000840, -0.00000798, -0.00000844, 0.00000351, 0.00000889, -0.00001068, 0.00000070, 0.00000029, 0.00000593, -0.00000557, 0.00000025, 0.00000114, -0.00000093, 0.00000117, 0.00000340, 0.00000012, -0.00000704, 0.00000412, 0.00000141, -0.00000079, -0.00000160, -0.00000089, -0.00000523, 0.00000665, -0.00000180, 0.00000358, 0.00000438, -0.00000456, -0.00000926, 0.00000507, 0.00000356, -0.00000041, -0.00000262, -0.00000327, -0.00000267, 0.00000176, 0.00000667, -0.00000494, },
		{ -0.00000290, -0.00000223, 0.00000794, -0.00000389, 0.00000426, -0.00000118, -0.00000488, 0.00000054, -0.00000154, -0.00000049, 0.00000450, -0.00000147, -0.00000421, -0.00000011, -0.00000080, 0.00000423, -0.00000056, 0.00000286, -0.00001099, 0.00000401, -0.00000508, 0.00000589, -0.00000662, 0.00000640, 0.00000505, -0.00000140, -0.00000686, 0.00000163, 0.00000576, -0.00000540, -0.00000292, 0.00000343, 0.00000005, 0.00000017, 0.00000075, -0.00000195, -0.00000138, -0.00000477, 0.00002068, -0.00001266, 0.00000540, -0.00000637, -0.00000501, 0.00000259, -0.00001569, 0.00000376, -0.00000451, 0.00000575, -0.00001690, 0.00000295, -0.00000798, -0.00000227, -0.00001864, -0.00000543, -0.00001362, 0.00000367, -0.00006153, -0.00001133, -0.00019122, 0.00010431, 0.00019060, 0.00003361, 0.00016331, 0.00016393, -0.00018704, -0.00004389, -0.00002755, 0.00000794, -0.00000175, 0.00001288, -0.00001808, -0.00000474, 0.00000975, 0.00000922, -0.00000506, 0.00002667, 0.00000524, 0.00000547, -0.00003203, -0.00001728, 0.00001086, -0.00001202, 0.00000090, -0.00000875, 0.00000110, -0.00000535, 0.00000274, -0.00001099, 0.00000407, -0.00000811, 0.00000959, -0.00000682, -0.00000827, 0.00000001, 0.00001125, -0.00000979, -0.00000013, -0.00000055, 0.00000612, -0.00000421, -0.00000125, 0.00000194, -0.00000169, 0.00000168, 0.00000291, 0.00000129, -0.00000739, 0.00000294, 0.00000230, -0.00000077, -0.00000193, 0.00000004, -0.00000579, 0.00000543, -0.00000014, 0.00000171, 0.00000590, -0.00000380, -0.00000954, 0.00000287, 0.00000548, -0.00000125, -0.00000141, -0.00000418, -0.00000177, -0.00000041, 0.00000812, -0.00000434, },
		{ -0.00000314, -0.00000335, 0.00000775, -0.00000241, 0.00000235, 0.00000114, -0.00000630, 0.00000102, -0.00000177, -0.00000043, 0.00000336, 0.00000085, -0.00000574, 0.00000083, -0.00000217, 0.00000532, -0.00000191, 0.00000480, -0.00001155, 0.00000347, -0.00000476, 0.00000508, -0.00000484, 0.00000333, 0.00000716, -0.00000121, -0.00000631, -0.00000060, 0.00000700, -0.00000473, -0.00000331, 0.00000251, 0.00000076, -0.00000029, 0.00000131, -0.00000231, -0.00000020, -0.00000712, 0.00002056, -0.00000853, -0.00000045, 0.00000051, -0.00001192, 0.00000857, -0.00001932, 0.00000560, -0.00000640, 0.00000841, -0.00001924, 0.00000469, -0.00001048, 0.00000170, -0.00002223, -0.00000177, -0.00001831, 0.00001117, -0.00006824, 0.00000093, -0.00018808, 0.00004661, 0.00022801, 0.00003315, 0.00017030, 0.00017444, -0.00018924, -0.00004842, -0.00002723, 0.00000394, 0.00000005, 0.00001078, -0.00001660, -0.00000735, 0.00001108, 0.00000720, -0.00000379, 0.00002121, 0.00001047, 0.00000283, -0.00002321, -0.00002656, 0.00001673, -0.00001554, 0.00000368, -0.00001126, 0.00000344, -0.00000791, 0.00000453, -0.00001165, 0.00000532, -0.00000967, 0.00000946, -0.00000485, -0.00000824, -0.00000294, 0.00001223, -0.00000765, -0.00000158, -0.00000078, 0.00000546, -0.00000223, -0.00000294, 0.00000276, -0.00000240, 0.00000208, 0.00000257, 0.00000214, -0.00000720, 0.00000141, 0.00000319, -0.00000074, -0.00000217, 0.00000064, -0.00000564, 0.00000350, 0.00000192, -0.00000023, 0.00000717, -0.00000297, -0.00000934, 0.00000031, 0.00000728, -0.00000189, -0.00000037, -0.00000489, -0.00000102, -0.00000220, 0.00000879, -0.00000313, },
		{ -0.00000359, -0.00000401, 0.00000678, -0.00000029, 0.00000013, 0.00000340, -0.00000735, 0.00000114, -0.00000184, -0.00000030, 0.00000209, 0.00000296, -0.00000674, 0.00000125, -0.00000300, 0.00000580, -0.00000256, 0.00000572, -0.00001086, 0.00000186, -0.00000348, 0.00000328, -0.00000217, -0.00000013, 0.00000907, -0.00000091, -0.00000545, -0.00000289, 0.00000787, -0.00000387, -0.00000350, 0.00000144, 0.00000137, -0.00000060, 0.00000169, -0.00000247, 0.00000060, -0.00000839, 0.00001869, -0.00000288, -0.00000703, 0.00000723, -0.00001765, 0.00001255, -0.00002070, 0.00000548, -0.00000664, 0.00000943, -0.00001983, 0.00000472, -0.00001127, 0.00000359, -0.00002329, -0.00000091, -0.00001978, 0.00001438, -0.00006906, 0.00000363, -0.00017018, -0.00001941, 0.00025952, 0.00004001, 0.00017358, 0.00018840, -0.00019347, -0.00005182, -0.00002775, 0.00000061, 0.00000126, 0.00000906, -0.00001524, -0.00001002, 0.00001272, 0.00000470, -0.00000168, 0.00001508, 0.00001591, -0.00000057, -0.00001416, -0.00003446, 0.00002047, -0.00001700, 0.00000468, -0.00001218, 0.00000430, -0.00000893, 0.00000470, -0.00001095, 0.00000542, -0.00000986, 0.00000790, -0.00000208, -0.00000849, -0.00000507, 0.00001168, -0.00000435, -0.00000350, -0.00000052, 0.00000414, 0.00000011, -0.00000459, 0.00000339, -0.00000293, 0.00000225, 0.00000245, 0.00000259, -0.00000651, -0.00000034, 0.00000401, -0.00000065, -0.00000231, 0.00000089, -0.00000485, 0.00000106, 0.00000405, -0.00000196, 0.00000799, -0.00000197, -0.00000883, -0.00000230, 0.00000864, -0.00000210, 0.00000030, -0.00000522, -0.00000060, -0.00000341, 0.00000860, -0.00000133, },
		{ -0.00000416, -0.00000425, 0.00000517, 0.00000219, -0.00000207, 0.00000527, -0.00000778, 0.00000076, -0.00000166, -0.00000022, 0.00000092, 0.00000458, -0.00000700, 0.00000099, -0.00000313, 0.00000553, -0.00000239, 0.00000555, -0.00000898, -0.00000065, -0.00000145, 0.00000078, 0.00000098, -0.00000352, 0.00001043, -0.00000035, -0.00000447, -0.00000503, 0.00000825, -0.00000279, -0.00000354, 0.00000034, 0.00000178, -0.00000070, 0.00000185, -0.00000237, 0.00000096, -0.00000863, 0.00001544, 0.00000353, -0.00001335, 0.00001267, -0.00002118, 0.00001378, -0.00001935, 0.00000312, -0.00000497, 0.00000855, -0.00001842, 0.00000280, -0.00001003, 0.00000303, -0.00002148, -0.00000319, -0.00001766, 0.00001292, -0.00006371, -0.00000282, -0.00014036, -0.00008792, 0.00027966, 0.00005685, 0.00017195, 0.00020673, -0.00020034, -0.00005360, -0.00002955, -0.00000164, 0.00000148, 0.00000809, -0.00001437, -0.00001239, 0.00001422, 0.00000223, 0.00000062, 0.00000905, 0.00002074, -0.00000369, -0.00000625, -0.00003970, 0.00002119, -0.00001582, 0.00000351, -0.00001120, 0.00000341, -0.00000819, 0.00000314, -0.00000885, 0.00000432, -0.00000869, 0.00000504, 0.00000133, -0.00000900, -0.00000628, 0.00000964, -0.00000017, -0.00000560, 0.00000002, 0.00000241, 0.00000250, -0.00000591, 0.00000367, -0.00000315, 0.00000213, 0.00000261, 0.00000267, -0.00000541, -0.00000216, 0.00000462, -0.00000045, -0.00000237, 0.00000082, -0.00000356, -0.00000159, 0.00000594, -0.00000317, 0.00000818, -0.00000071, -0.00000818, -0.00000469, 0.00000933, -0.00000172, 0.00000049, -0.00000506, -0.00000064, -0.00000395, 0.00000755, 0.00000092, },
		{ -0.00000470, -0.00000420, 0.00000314, 0.00000469, -0.00000389, 0.00000645, -0.00000742, -0.00000016, -0.00000121, -0.00000030, 0.00000002, 0.00000552, -0.00000640, 0.00000001, -0.00000257, 0.00000452, -0.00000144, 0.00000440, -0.00000618, -0.00000373, 0.00000095, -0.00000200, 0.00000412, -0.00000639, 0.00001094, 0.00000064, -0.00000350, -0.00000681, 0.00000801, -0.00000150, -0.00000346, -0.00000069, 0.00000190, -0.00000055, 0.00000179, -0.00000203, 0.00000091, -0.00000803, 0.00001133, 0.00000987, -0.00001844, 0.00001593, -0.00002182, 0.00001191, -0.00001528, -0.00000134, -0.00000156, 0.00000591, -0.00001517, -0.00000094, -0.00000686, 0.00000014, -0.00001699, -0.00000834, -0.00001227, 0.00000727, -0.00005308, -0.00001651, -0.00010302, -0.00015233, 0.00028338, 0.00008536, 0.00016505, 0.00022974, -0.00021001, -0.00005367, -0.00003280, -0.00000260, 0.00000050, 0.00000809, -0.00001422, -0.00001413, 0.00001521, 0.00000026, 0.00000249, 0.00000388, 0.00002416, -0.00000556, -0.00000059, -0.00004141, 0.00001845, -0.00001189, 0.00000019, -0.00000839, 0.00000087, -0.00000581, 0.00000007, -0.00000560, 0.00000224, -0.00000638, 0.00000120, 0.00000505, -0.00000961, -0.00000663, 0.00000641, 0.00000442, -0.00000746, 0.00000051, 0.00000061, 0.00000460, -0.00000668, 0.00000346, -0.00000301, 0.00000170, 0.00000303, 0.00000242, -0.00000406, -0.00000386, 0.00000490, -0.00000009, -0.00000239, 0.00000052, -0.00000197, -0.00000412, 0.00000724, -0.00000363, 0.00000763, 0.00000085, -0.00000751, -0.00000663, 0.00000918, -0.00000068, 0.00000016, -0.00000439, -0.00000115, -0.00000384, 0.00000580, 0.00000340, },
		{ -0.00000501, -0.00000401, 0.00000095, 0.00000688, -0.00000502, 0.00000674, -0.00000621, -0.00000159, -0.00000051, -0.00000056, -0.00000049, 0.00000572, -0.00000498, -0.00000159, -0.00000145, 0.00000295, 0.00000009, 0.00000255, -0.00000285, -0.00000690, 0.00000326, -0.00000458, 0.00000676, -0.00000829, 0.00001037, 0.00000212, -0.00000268, -0.00000807, 0.00000712, -0.00000000, -0.00000329, -0.00000156, 0.00000169, -0.00000016, 0.00000156, -0.00000152, 0.00000057, -0.00000686, 0.00000694, 0.00001533, -0.00002151, 0.00001641, -0.00001933, 0.00000709, -0.00000893, -0.00000733, 0.00000305, 0.00000201, -0.00001059, -0.00000593, -0.00000231, -0.00000449, -0.00001055, -0.00001547, -0.00000466, -0.00000123, -0.00003906, -0.00003435, -0.00006349, -0.00020607, 0.00026676, 0.00012575, 0.00015362, 0.00025695, -0.00022200, -0.00005240, -0.00003727, -0.00000240, -0.00000160, 0.00000902, -0.00001484, -0.00001511, 0.00001544, -0.00000085, 0.00000344, 0.00000016, 0.00002557, -0.00000546, 0.00000209, -0.00003925, 0.00001240, -0.00000563, -0.00000480, -0.00000421, -0.00000291, -0.00000224, -0.00000402, -0.00000171, -0.00000037, -0.00000338, -0.00000308, 0.00000861, -0.00001006, -0.00000637, 0.00000244, 0.00000883, -0.00000862, 0.00000064, -0.00000093, 0.00000611, -0.00000671, 0.00000271, -0.00000251, 0.00000098, 0.00000365, 0.00000196, -0.00000261, -0.00000527, 0.00000476, 0.00000047, -0.00000241, 0.00000010, -0.00000033, -0.00000622, 0.00000770, -0.00000321, 0.00000633, 0.00000269, -0.00000691, -0.00000797, 0.00000816, 0.00000094, -0.00000058, -0.00000332, -0.00000202, -0.00000321, 0.00000358, 0.00000584, },
		{ -0.00000488, -0.00000387, -0.00000112, 0.00000844, -0.00000524, 0.00000606, -0.00000421, -0.00000340, 0.00000032, -0.00000096, -0.00000063, 0.00000522, -0.00000291, -0.00000357, -0.00000001, 0.00000108, 0.00000188, 0.00000042, 0.00000050, -0.00000965, 0.00000500, -0.00000649, 0.00000845, -0.00000890, 0.00000860, 0.00000410, -0.00000206, -0.00000872, 0.00000558, 0.00000161, -0.00000305, -0.00000220, 0.00000112, 0.00000042, 0.00000120, -0.00000091, 0.00000007, -0.00000545, 0.00000282, 0.00001929, -0.00002211, 0.00001401, -0.00001399, 0.00000000, -0.00000122, -0.00001392, 0.00000802, -0.00000231, -0.00000555, -0.00001132, 0.00000272, -0.00000985, -0.00000334, -0.00002321, 0.00000359, -0.00001064, -0.00002424, -0.00005263, -0.00002728, -0.00024350, 0.00022784, 0.00017633, 0.00013968, 0.00028707, -0.00023524, -0.00005065, -0.00004232, -0.00000153, -0.00000444, 0.00001059, -0.00001605, -0.00001542, 0.00001489, -0.00000099, 0.00000326, -0.00000182, 0.00002473, -0.00000312, 0.00000165, -0.00003355, 0.00000380, 0.00000205, -0.00001058, 0.00000055, -0.00000719, 0.00000179, -0.00000841, 0.00000215, -0.00000291, -0.00000029, -0.00000720, 0.00001147, -0.00000998, -0.00000585, -0.00000171, 0.00001244, -0.00000867, 0.00000014, -0.00000196, 0.00000683, -0.00000593, 0.00000146, -0.00000172, 0.00000008, 0.00000437, 0.00000142, -0.00000126, -0.00000623, 0.00000414, 0.00000123, -0.00000247, -0.00000034, 0.00000118, -0.00000765, 0.00000718, -0.00000190, 0.00000439, 0.00000467, -0.00000637, -0.00000867, 0.00000634, 0.00000296, -0.00000157, -0.00000201, -0.00000309, -0.00000228, 0.00000121, 0.00000792, },
		{ -0.00000418, -0.00000390, -0.00000286, 0.00000917, -0.00000447, 0.00000448, -0.00000163, -0.00000532, 0.00000109, -0.00000140, -0.00000048, 0.00000419, -0.00000047, -0.00000559, 0.00000139, -0.00000073, 0.00000354, -0.00000155, 0.00000337, -0.00001151, 0.00000577, -0.00000739, 0.00000891, -0.00000809, 0.00000571, 0.00000643, -0.00000163, -0.00000874, 0.00000348, 0.00000326, -0.00000274, -0.00000258, 0.00000023, 0.00000113, 0.00000080, -0.00000029, -0.00000044, -0.00000404, -0.00000060, 0.00002138, -0.00002018, 0.00000905, -0.00000657, -0.00000824, 0.00000662, -0.00001994, 0.00001229, -0.00000605, -0.00000107, -0.00001606, 0.00000714, -0.00001473, 0.00000321, -0.00002997, 0.00001070, -0.00001879, -0.00001138, -0.00006761, 0.00000073, -0.00026072, 0.00016705, 0.00023337, 0.00012644, 0.00031807, -0.00024809, -0.00004974, -0.00004697, -0.00000078, -0.00000735, 0.00001224, -0.00001740, -0.00001538, 0.00001378, -0.00000029, 0.00000200, -0.00000213, 0.00002178, 0.00000124, -0.00000149, -0.00002525, -0.00000606, 0.00000986, -0.00001603, 0.00000492, -0.00001107, 0.00000545, -0.00001226, 0.00000521, -0.00000472, 0.00000226, -0.00001050, 0.00001312, -0.00000906, -0.00000543, -0.00000548, 0.00001472, -0.00000730, -0.00000117, -0.00000233, 0.00000666, -0.00000439, -0.00000019, -0.00000076, -0.00000090, 0.00000506, 0.00000093, -0.00000015, -0.00000666, 0.00000306, 0.00000214, -0.00000256, -0.00000074, 0.00000237, -0.00000827, 0.00000568, 0.00000014, 0.00000206, 0.00000660, -0.00000581, -0.00000882, 0.00000395, 0.00000510, -0.00000254, -0.00000070, -0.00000411, -0.00000132, -0.00000099, 0.00000937, },
		{ -0.00000283, -0.00000417, -0.00000413, 0.00000899, -0.00000278, 0.00000221, 0.00000121, -0.00000707, 0.00000160, -0.00000174, -0.00000018, 0.00000288, 0.00000199, -0.00000728, 0.00000242, -0.00000213, 0.00000470, -0.00000295, 0.00000533, -0.00001212, 0.00000532, -0.00000707, 0.00000803, -0.00000590, 0.00000192, 0.00000889, -0.00000130, -0.00000820, 0.00000097, 0.00000479, -0.00000235, -0.00000267, -0.00000096, 0.00000189, 0.00000042, 0.00000028, -0.00000088, -0.00000284, -0.00000308, 0.00002153, -0.00001604, 0.00000232, 0.00000175, -0.00001627, 0.00001321, -0.00002419, 0.00001483, -0.00000824, 0.00000186, -0.00001911, 0.00000988, -0.00001792, 0.00000776, -0.00003424, 0.00001495, -0.00002365, -0.00000297, -0.00007619, 0.00001700, -0.00025613, 0.00008754, 0.00029122, 0.00011805, 0.00034748, -0.00025858, -0.00005119, -0.00005002, -0.00000113, -0.00000948, 0.00001323, -0.00001829, -0.00001549, 0.00001253, 0.00000086, 0.00000002, -0.00000116, 0.00001728, 0.00000696, -0.00000635, -0.00001574, -0.00001560, 0.00001635, -0.00001994, 0.00000788, -0.00001368, 0.00000789, -0.00001478, 0.00000679, -0.00000525, 0.00000372, -0.00001245, 0.00001319, -0.00000706, -0.00000544, -0.00000839, 0.00001527, -0.00000440, -0.00000326, -0.00000202, 0.00000567, -0.00000225, -0.00000204, 0.00000021, -0.00000183, 0.00000563, 0.00000059, 0.00000063, -0.00000653, 0.00000159, 0.00000312, -0.00000264, -0.00000104, 0.00000319, -0.00000806, 0.00000335, 0.00000261, -0.00000034, 0.00000821, -0.00000511, -0.00000856, 0.00000127, 0.00000703, -0.00000323, 0.00000036, -0.00000487, -0.00000057, -0.00000275, 0.00000998, },
		{ -0.00000086, -0.00000466, -0.00000489, 0.00000794, -0.00000039, -0.00000039, 0.00000394, -0.00000831, 0.00000163, -0.00000184, 0.00000010, 0.00000156, 0.00000410, -0.00000832, 0.00000278, -0.00000283, 0.00000509, -0.00000349, 0.00000610, -0.00001132, 0.00000360, -0.00000557, 0.00000592, -0.00000261, -0.00000238, 0.00001117, -0.00000094, -0.00000720, -0.00000176, 0.00000608, -0.00000185, -0.00000248, -0.00000238, 0.00000264, 0.00000010, 0.00000077, -0.00000122, -0.00000193, -0.00000457, 0.00001998, -0.00001039, -0.00000510, 0.00000965, -0.00002266, 0.00001731, -0.00002568, 0.00001481, -0.00000809, 0.00000245, -0.00001969, 0.00001011, -0.00001849, 0.00000923, -0.00003487, 0.00001511, -0.00002380, -0.00000069, -0.00007643, 0.00001987, -0.00023083, -0.00000498, 0.00034281, 0.00011899, 0.00037282, -0.00026480, -0.00005651, -0.00005031, -0.00000357, -0.00000998, 0.00001280, -0.00001806, -0.00001634, 0.00001169, 0.00000193, -0.00000212, 0.00000041, 0.00001208, 0.00001303, -0.00001165, -0.00000666, -0.00002316, 0.00002017, -0.00002127, 0.00000863, -0.00001430, 0.00000847, -0.00001540, 0.00000643, -0.00000412, 0.00000374, -0.00001273, 0.00001149, -0.00000392, -0.00000604, -0.00001012, 0.00001393, -0.00000009, -0.00000596, -0.00000119, 0.00000402, 0.00000026, -0.00000385, 0.00000102, -0.00000259, 0.00000599, 0.00000048, 0.00000105, -0.00000592, -0.00000011, 0.00000401, -0.00000264, -0.00000129, 0.00000362, -0.00000712, 0.00000048, 0.00000516, -0.00000243, 0.00000924, -0.00000411, -0.00000809, -0.00000138, 0.00000842, -0.00000339, 0.00000099, -0.00000516, -0.00000025, -0.00000387, 0.00000967, },
		{ 0.00000156, -0.00000526, -0.00000520, 0.00000618, 0.00000239, -0.00000294, 0.00000615, -0.00000875, 0.00000103, -0.00000159, 0.00000020, 0.00000049, 0.00000559, -0.00000845, 0.00000229, -0.00000268, 0.00000455, -0.00000303, 0.00000562, -0.00000918, 0.00000078, -0.00000311, 0.00000290, 0.00000135, -0.00000670, 0.00001291, -0.00000037, -0.00000590, -0.00000448, 0.00000697, -0.00000123, -0.00000202, -0.00000393, 0.00000330, -0.00000014, 0.00000120, -0.00000146, -0.00000129, -0.00000520, 0.00001720, -0.00000411, -0.00001200, 0.00001578, -0.00002619, 0.00001796, -0.00002378, 0.00001181, -0.00000523, 0.00000032, -0.00001741, 0.00000744, -0.00001598, 0.00000711, -0.00003134, 0.00001066, -0.00001865, -0.00000514, -0.00006789, 0.00000979, -0.00018846, -0.00010273, 0.00038038, 0.00013336, 0.00039218, -0.00026527, -0.00006685, -0.00004699, -0.00000882, -0.00000815, 0.00001034, -0.00001616, -0.00001844, 0.00001177, 0.00000235, -0.00000378, 0.00000182, 0.00000714, 0.00001832, -0.00001600, 0.00000039, -0.00002733, 0.00002029, -0.00001939, 0.00000671, -0.00001256, 0.00000687, -0.00001384, 0.00000397, -0.00000125, 0.00000224, -0.00001128, 0.00000810, 0.00000022, -0.00000724, -0.00001059, 0.00001081, 0.00000528, -0.00000889, -0.00000009, 0.00000201, 0.00000282, -0.00000539, 0.00000152, -0.00000309, 0.00000610, 0.00000060, 0.00000114, -0.00000494, -0.00000186, 0.00000467, -0.00000246, -0.00000152, 0.00000375, -0.00000566, -0.00000256, 0.00000734, -0.00000386, 0.00000946, -0.00000269, -0.00000759, -0.00000368, 0.00000903, -0.00000285, 0.00000105, -0.00000490, -0.00000043, -0.00000426, 0.00000848, },
		{ 0.00000419, -0.00000580, -0.00000520, 0.00000398, 0.00000517, -0.00000501, 0.00000750, -0.00000820, -0.00000026, -0.00000096, 0.00000002, -0.00000017, 0.00000626, -0.00000757, 0.00000092, -0.00000167, 0.00000312, -0.00000163, 0.00000403, -0.00000596, -0.00000276, -0.00000011, -0.00000057, 0.00000545, -0.00001051, 0.00001377, 0.00000053, -0.00000447, -0.00000697, 0.00000734, -0.00000048, -0.00000134, -0.00000555, 0.00000381, -0.00000029, 0.00000161, -0.00000167, -0.00000084, -0.00000523, 0.00001379, 0.00000184, -0.00001727, 0.00001909, -0.00002608, 0.00001477, -0.00001841, 0.00000591, 0.00000022, -0.00000440, -0.00001244, 0.00000202, -0.00001050, 0.00000153, -0.00002389, 0.00000194, -0.00000864, -0.00001566, -0.00005174, -0.00001071, -0.00013487, -0.00019669, 0.00039650, 0.00016409, 0.00040480, -0.00025935, -0.00008266, -0.00003973, -0.00001721, -0.00000369, 0.00000555, -0.00001229, -0.00002211, 0.00001316, 0.00000166, -0.00000441, 0.00000235, 0.00000339, 0.00002179, -0.00001820, 0.00000416, -0.00002718, 0.00001629, -0.00001418, 0.00000216, -0.00000855, 0.00000317, -0.00001025, -0.00000038, 0.00000311, -0.00000057, -0.00000834, 0.00000339, 0.00000499, -0.00000883, -0.00000992, 0.00000627, 0.00001116, -0.00001158, 0.00000091, -0.00000004, 0.00000512, -0.00000643, 0.00000160, -0.00000331, 0.00000595, 0.00000090, 0.00000099, -0.00000378, -0.00000343, 0.00000492, -0.00000201, -0.00000180, 0.00000368, -0.00000393, -0.00000540, 0.00000879, -0.00000435, 0.00000874, -0.00000081, -0.00000720, -0.00000543, 0.00000869, -0.00000159, 0.00000054, -0.00000409, -0.00000113, -0.00000397, 0.00000657, },
		{ 0.00000674, -0.00000607, -0.00000507, 0.00000163, 0.00000757, -0.00000627, 0.00000778, -0.00000662, -0.00000218, -0.00000001, -0.00000045, -0.00000033, 0.00000606, -0.00000575, -0.00000119, 0.00000000, 0.00000101, 0.00000044, 0.00000166, -0.00000213, -0.00000652, 0.00000290, -0.00000393, 0.00000908, -0.00001331, 0.00001349, 0.00000186, -0.00000304, -0.00000905, 0.00000712, 0.00000038, -0.00000046, -0.00000712, 0.00000411, -0.00000038, 0.00000203, -0.00000192, -0.00000044, -0.00000494, 0.00001038, 0.00000659, -0.00002006, 0.00001896, -0.00002208, 0.00000795, -0.00001012, -0.00000221, 0.00000760, -0.00001106, -0.00000545, -0.00000544, -0.00000281, -0.00000667, -0.00001352, -0.00000986, 0.00000481, -0.00003045, -0.00003054, -0.00003751, -0.00007724, -0.00027766, 0.00038515, 0.00021208, 0.00041150, -0.00024757, -0.00010352, -0.00002893, -0.00002843, 0.00000322, -0.00000146, -0.00000654, -0.00002734, 0.00001593, -0.00000034, -0.00000367, 0.00000149, 0.00000147, 0.00002271, -0.00001745, 0.00000396, -0.00002249, 0.00000843, -0.00000617, -0.00000446, -0.00000278, -0.00000211, -0.00000515, -0.00000605, 0.00000839, -0.00000418, -0.00000443, -0.00000207, 0.00000984, -0.00001045, -0.00000846, 0.00000087, 0.00001686, -0.00001351, 0.00000148, -0.00000180, 0.00000688, -0.00000681, 0.00000123, -0.00000326, 0.00000562, 0.00000132, 0.00000073, -0.00000261, -0.00000463, 0.00000466, -0.00000124, -0.00000219, 0.00000353, -0.00000220, -0.00000767, 0.00000921, -0.00000374, 0.00000708, 0.00000148, -0.00000697, -0.00000649, 0.00000740, 0.00000032, -0.00000042, -0.00000283, -0.00000221, -0.00000317, 0.00000422, },
		{ 0.00000888, -0.00000587, -0.00000500, -0.00000058, 0.00000925, -0.00000648, 0.00000691, -0.00000409, -0.00000454, 0.00000112, -0.00000113, -0.00000003, 0.00000511, -0.00000321, -0.00000373, 0.00000204, -0.00000144, 0.00000281, -0.00000102, 0.00000175, -0.00000990, 0.00000539, -0.00000664, 0.00001175, -0.00001473, 0.00001195, 0.00000364, -0.00000173, -0.00001055, 0.00000627, 0.00000134, 0.00000056, -0.00000854, 0.00000416, -0.00000041, 0.00000249, -0.00000224, -0.00000000, -0.00000462, 0.00000748, 0.00000954, -0.00001992, 0.00001531, -0.00001463, -0.00000164, -0.00000002, -0.00001141, 0.00001579, -0.00001856, 0.00000241, -0.00001375, 0.00000586, -0.00001611, -0.00000185, -0.00002286, 0.00001954, -0.00004681, -0.00000783, -0.00006551, -0.00002318, -0.00033752, 0.00034278, 0.00027561, 0.00041503, -0.00023170, -0.00012804, -0.00001577, -0.00004157, 0.00001182, -0.00001004, 0.00000056, -0.00003373, 0.00001985, -0.00000354, -0.00000155, -0.00000090, 0.00000166, 0.00002080, -0.00001352, -0.00000017, -0.00001378, -0.00000234, 0.00000354, -0.00001212, 0.00000381, -0.00000812, 0.00000063, -0.00001219, 0.00001378, -0.00000789, -0.00000024, -0.00000753, 0.00001412, -0.00001164, -0.00000667, -0.00000472, 0.00002167, -0.00001419, 0.00000129, -0.00000301, 0.00000789, -0.00000646, 0.00000044, -0.00000300, 0.00000517, 0.00000175, 0.00000048, -0.00000162, -0.00000531, 0.00000380, -0.00000015, -0.00000272, 0.00000343, -0.00000070, -0.00000911, 0.00000845, -0.00000206, 0.00000462, 0.00000401, -0.00000685, -0.00000688, 0.00000528, 0.00000263, -0.00000162, -0.00000135, -0.00000347, -0.00000211, 0.00000176, },
		{ 0.00001033, -0.00000505, -0.00000510, -0.00000242, 0.00001001, -0.00000557, 0.00000500, -0.00000090, -0.00000703, 0.00000219, -0.00000188, 0.00000058, 0.00000362, -0.00000030, -0.00000632, 0.00000403, -0.00000380, 0.00000502, -0.00000348, 0.00000510, -0.00001236, 0.00000686, -0.00000826, 0.00001307, -0.00001454, 0.00000916, 0.00000576, -0.00000056, -0.00001141, 0.00000480, 0.00000233, 0.00000166, -0.00000973, 0.00000392, -0.00000037, 0.00000301, -0.00000266, 0.00000053, -0.00000446, 0.00000544, 0.00001040, -0.00001690, 0.00000868, -0.00000474, -0.00001261, 0.00001037, -0.00002023, 0.00002343, -0.00002553, 0.00000974, -0.00002146, 0.00001397, -0.00002507, 0.00000917, -0.00003485, 0.00003301, -0.00006170, 0.00001252, -0.00008949, 0.00002039, -0.00037030, 0.00026909, 0.00034997, 0.00042002, -0.00021472, -0.00015395, -0.00000212, -0.00005511, 0.00002087, -0.00001914, 0.00000811, -0.00004054, 0.00002429, -0.00000747, 0.00000159, -0.00000459, 0.00000377, 0.00001632, -0.00000681, -0.00000751, -0.00000229, -0.00001448, 0.00001344, -0.00001946, 0.00001004, -0.00001379, 0.00000607, -0.00001782, 0.00001841, -0.00001091, 0.00000347, -0.00001221, 0.00001718, -0.00001195, -0.00000504, -0.00000980, 0.00002495, -0.00001327, 0.00000019, -0.00000353, 0.00000807, -0.00000542, -0.00000068, -0.00000263, 0.00000471, 0.00000209, 0.00000036, -0.00000093, -0.00000538, 0.00000239, 0.00000121, -0.00000334, 0.00000341, 0.00000042, -0.00000958, 0.00000656, 0.00000049, 0.00000167, 0.00000652, -0.00000672, -0.00000670, 0.00000258, 0.00000504, -0.00000277, 0.00000010, -0.00000465, -0.00000105, -0.00000049, },
		{ 0.00001089, -0.00000355, -0.00000544, -0.00000376, 0.00000978, -0.00000365, 0.00000232, 0.00000258, -0.00000928, 0.00000292, -0.00000249, 0.00000130, 0.00000191, 0.00000257, -0.00000851, 0.00000555, -0.00000566, 0.00000662, -0.00000526, 0.00000744, -0.00001348, 0.00000698, -0.00000853, 0.00001288, -0.00001277, 0.00000535, 0.00000803, 0.00000048, -0.00001159, 0.00000282, 0.00000329, 0.00000281, -0.00001059, 0.00000335, -0.00000027, 0.00000354, -0.00000312, 0.00000115, -0.00000452, 0.00000437, 0.00000925, -0.00001152, 0.00000009, 0.00000612, -0.00002327, 0.00001937, -0.00002711, 0.00002909, -0.00003059, 0.00001514, -0.00002710, 0.00001997, -0.00003185, 0.00001760, -0.00004368, 0.00004280, -0.00007223, 0.00002697, -0.00010500, 0.00004826, -0.00037312, 0.00016753, 0.00042759, 0.00043267, -0.00020040, -0.00017846, 0.00000976, -0.00006724, 0.00002885, -0.00002743, 0.00001494, -0.00004677, 0.00002843, -0.00001141, 0.00000513, -0.00000899, 0.00000720, 0.00001003, 0.00000172, -0.00001673, 0.00001023, -0.00002611, 0.00002181, -0.00002508, 0.00001474, -0.00001804, 0.00001017, -0.00002199, 0.00002142, -0.00001252, 0.00000599, -0.00001544, 0.00001850, -0.00001100, -0.00000401, -0.00001381, 0.00002624, -0.00001060, -0.00000183, -0.00000335, 0.00000746, -0.00000381, -0.00000197, -0.00000230, 0.00000432, 0.00000228, 0.00000042, -0.00000060, -0.00000487, 0.00000054, 0.00000268, -0.00000399, 0.00000349, 0.00000110, -0.00000912, 0.00000372, 0.00000356, -0.00000137, 0.00000868, -0.00000639, -0.00000616, -0.00000035, 0.00000718, -0.00000357, 0.00000126, -0.00000551, -0.00000028, -0.00000223, },
		{ 0.00001050, -0.00000144, -0.00000600, -0.00000458, 0.00000861, -0.00000097, -0.00000072, 0.00000589, -0.00001089, 0.00000306, -0.00000278, 0.00000190, 0.00000030, 0.00000499, -0.00000991, 0.00000626, -0.00000667, 0.00000727, -0.00000600, 0.00000845, -0.00001305, 0.00000567, -0.00000743, 0.00001123, -0.00000966, 0.00000086, 0.00001017, 0.00000147, -0.00001116, 0.00000045, 0.00000412, 0.00000395, -0.00001106, 0.00000245, -0.00000009, 0.00000404, -0.00000354, 0.00000175, -0.00000475, 0.00000417, 0.00000651, -0.00000467, -0.00000911, 0.00001629, -0.00003187, 0.00002537, -0.00003072, 0.00003161, -0.00003263, 0.00001746, -0.00002951, 0.00002263, -0.00003505, 0.00002189, -0.00004765, 0.00004703, -0.00007624, 0.00003299, -0.00010910, 0.00005770, -0.00034668, 0.00004529, 0.00049848, 0.00046004, -0.00019268, -0.00019871, 0.00001758, -0.00007612, 0.00003420, -0.00003353, 0.00001984, -0.00005136, 0.00003134, -0.00001452, 0.00000824, -0.00001330, 0.00001107, 0.00000301, 0.00001073, -0.00002616, 0.00002181, -0.00003529, 0.00002707, -0.00002774, 0.00001689, -0.00002001, 0.00001211, -0.00002399, 0.00002223, -0.00001222, 0.00000684, -0.00001675, 0.00001778, -0.00000864, -0.00000385, -0.00001635, 0.00002536, -0.00000629, -0.00000458, -0.00000261, 0.00000623, -0.00000185, -0.00000324, -0.00000210, 0.00000407, 0.00000228, 0.00000069, -0.00000061, -0.00000388, -0.00000156, 0.00000409, -0.00000453, 0.00000360, 0.00000138, -0.00000786, 0.00000030, 0.00000669, -0.00000405, 0.00001016, -0.00000566, -0.00000549, -0.00000317, 0.00000870, -0.00000376, 0.00000193, -0.00000586, 0.00000003, -0.00000329, },
		{ 0.00000922, 0.00000112, -0.00000663, -0.00000494, 0.00000671, 0.00000211, -0.00000366, 0.00000857, -0.00001154, 0.00000243, -0.00000261, 0.00000219, -0.00000092, 0.00000662, -0.00001024, 0.00000594, -0.00000664, 0.00000681, -0.00000555, 0.00000802, -0.00001109, 0.00000305, -0.00000515, 0.00000844, -0.00000563, -0.00000378, 0.00001183, 0.00000253, -0.00001023, -0.00000212, 0.00000473, 0.00000506, -0.00001109, 0.00000121, 0.00000019, 0.00000442, -0.00000381, 0.00000222, -0.00000501, 0.00000457, 0.00000278, 0.00000255, -0.00001746, 0.00002417, -0.00003693, 0.00002721, -0.00003019, 0.00003030, -0.00003100, 0.00001607, -0.00002804, 0.00002127, -0.00003392, 0.00002119, -0.00004591, 0.00004481, -0.00007274, 0.00002952, -0.00010088, 0.00004888, -0.00029536, -0.00008727, 0.00055122, 0.00050907, -0.00019492, -0.00021244, 0.00001954, -0.00008031, 0.00003568, -0.00003634, 0.00002182, -0.00005345, 0.00003220, -0.00001601, 0.00001013, -0.00001667, 0.00001440, -0.00000352, 0.00001876, -0.00003408, 0.00003061, -0.00004042, 0.00002806, -0.00002668, 0.00001593, -0.00001921, 0.00001147, -0.00002343, 0.00002057, -0.00000984, 0.00000585, -0.00001601, 0.00001503, -0.00000494, -0.00000460, -0.00001730, 0.00002240, -0.00000071, -0.00000768, -0.00000158, 0.00000464, 0.00000019, -0.00000427, -0.00000216, 0.00000401, 0.00000209, 0.00000113, -0.00000089, -0.00000258, -0.00000367, 0.00000522, -0.00000482, 0.00000365, 0.00000137, -0.00000608, -0.00000327, 0.00000940, -0.00000595, 0.00001068, -0.00000438, -0.00000490, -0.00000555, 0.00000933, -0.00000319, 0.00000198, -0.00000561, -0.00000022, -0.00000360, },
		{ 0.00000723, 0.00000386, -0.00000716, -0.00000499, 0.00000437, 0.00000515, -0.00000605, 0.00001023, -0.00001100, 0.00000095, -0.00000193, 0.00000206, -0.00000157, 0.00000729, -0.00000940, 0.00000457, -0.00000559, 0.00000528, -0.00000399, 0.00000633, -0.00000790, -0.00000049, -0.00000213, 0.00000499, -0.00000128, -0.00000802, 0.00001269, 0.00000378, -0.00000893, -0.00000466, 0.00000498, 0.00000613, -0.00001069, -0.00000030, 0.00000053, 0.00000465, -0.00000383, 0.00000243, -0.00000515, 0.00000525, -0.00000121, 0.00000906, -0.00002369, 0.00002853, -0.00003752, 0.00002437, -0.00002536, 0.00002517, -0.00002575, 0.00001102, -0.00002279, 0.00001598, -0.00002853, 0.00001562, -0.00003864, 0.00003646, -0.00006215, 0.00001727, -0.00008165, 0.00002486, -0.00022675, -0.00021757, 0.00057425, 0.00058542, -0.00020897, -0.00021855, 0.00001479, -0.00007918, 0.00003274, -0.00003533, 0.00002040, -0.00005258, 0.00003057, -0.00001539, 0.00001025, -0.00001844, 0.00001636, -0.00000851, 0.00002456, -0.00003905, 0.00003522, -0.00004051, 0.00002428, -0.00002173, 0.00001186, -0.00001568, 0.00000831, -0.00002043, 0.00001663, -0.00000561, 0.00000321, -0.00001342, 0.00001058, -0.00000027, -0.00000605, -0.00001682, 0.00001775, 0.00000554, -0.00001061, -0.00000063, 0.00000299, 0.00000203, -0.00000489, -0.00000253, 0.00000413, 0.00000176, 0.00000166, -0.00000130, -0.00000117, -0.00000554, 0.00000586, -0.00000474, 0.00000355, 0.00000122, -0.00000406, -0.00000656, 0.00001124, -0.00000674, 0.00001008, -0.00000250, -0.00000453, -0.00000727, 0.00000893, -0.00000183, 0.00000141, -0.00000476, -0.00000101, -0.00000322, },
		{ 0.00000482, 0.00000645, -0.00000737, -0.00000494, 0.00000189, 0.00000774, -0.00000751, 0.00001064, -0.00000924, -0.00000129, -0.00000083, 0.00000153, -0.00000159, 0.00000696, -0.00000748, 0.00000232, -0.00000373, 0.00000293, -0.00000161, 0.00000374, -0.00000397, -0.00000438, 0.00000106, 0.00000149, 0.00000275, -0.00001131, 0.00001245, 0.00000534, -0.00000745, -0.00000694, 0.00000476, 0.00000717, -0.00000992, -0.00000196, 0.00000087, 0.00000471, -0.00000357, 0.00000232, -0.00000504, 0.00000589, -0.00000483, 0.00001396, -0.00002687, 0.00002871, -0.00003343, 0.00001712, -0.00001682, 0.00001695, -0.00001766, 0.00000313, -0.00001460, 0.00000766, -0.00001982, 0.00000622, -0.00002713, 0.00002349, -0.00004632, -0.00000139, -0.00005482, -0.00000883, -0.00015072, -0.00033222, 0.00055737, 0.00069224, -0.00023441, -0.00021770, 0.00000386, -0.00007319, 0.00002572, -0.00003075, 0.00001576, -0.00004891, 0.00002652, -0.00001264, 0.00000842, -0.00001831, 0.00001644, -0.00001124, 0.00002729, -0.00004021, 0.00003500, -0.00003538, 0.00001607, -0.00001352, 0.00000532, -0.00001001, 0.00000319, -0.00001556, 0.00001104, -0.00000014, -0.00000054, -0.00000955, 0.00000507, 0.00000477, -0.00000777, -0.00001532, 0.00001208, 0.00001169, -0.00001280, -0.00000014, 0.00000163, 0.00000340, -0.00000496, -0.00000324, 0.00000440, 0.00000134, 0.00000219, -0.00000173, 0.00000014, -0.00000695, 0.00000587, -0.00000421, 0.00000323, 0.00000106, -0.00000211, -0.00000916, 0.00001190, -0.00000626, 0.00000837, -0.00000008, -0.00000441, -0.00000823, 0.00000751, 0.00000021, 0.00000038, -0.00000346, -0.00000218, -0.00000235, },
		{ 0.00000235, 0.00000859, -0.00000707, -0.00000494, -0.00000043, 0.00000956, -0.00000781, 0.00000975, -0.00000640, -0.00000404, 0.00000050, 0.00000070, -0.00000107, 0.00000580, -0.00000476, -0.00000045, -0.00000144, 0.00000016, 0.00000111, 0.00000081, 0.00000007, -0.00000797, 0.00000378, -0.00000142, 0.00000585, -0.00001316, 0.00001090, 0.00000725, -0.00000597, -0.00000873, 0.00000395, 0.00000824, -0.00000890, -0.00000360, 0.00000110, 0.00000467, -0.00000307, 0.00000193, -0.00000466, 0.00000631, -0.00000764, 0.00001672, -0.00002667, 0.00002479, -0.00002529, 0.00000655, -0.00000595, 0.00000709, -0.00000816, -0.00000615, -0.00000500, -0.00000209, -0.00000951, -0.00000506, -0.00001360, 0.00000849, -0.00002825, -0.00002271, -0.00002531, -0.00004505, -0.00007819, -0.00041863, 0.00049323, 0.00082914, -0.00026796, -0.00021258, -0.00001115, -0.00006409, 0.00001605, -0.00002378, 0.00000892, -0.00004329, 0.00002076, -0.00000829, 0.00000502, -0.00001648, 0.00001469, -0.00001157, 0.00002676, -0.00003746, 0.00003015, -0.00002580, 0.00000459, -0.00000333, -0.00000248, -0.00000329, -0.00000288, -0.00000981, 0.00000476, 0.00000565, -0.00000457, -0.00000521, -0.00000063, 0.00000938, -0.00000918, -0.00001340, 0.00000617, 0.00001694, -0.00001367, -0.00000046, 0.00000083, 0.00000408, -0.00000439, -0.00000430, 0.00000477, 0.00000091, 0.00000265, -0.00000205, 0.00000120, -0.00000776, 0.00000521, -0.00000325, 0.00000269, 0.00000099, -0.00000045, -0.00001081, 0.00001123, -0.00000455, 0.00000573, 0.00000268, -0.00000446, -0.00000847, 0.00000524, 0.00000268, -0.00000088, -0.00000194, -0.00000352, -0.00000125, },
		{ 0.00000014, 0.00001000, -0.00000611, -0.00000513, -0.00000237, 0.00001040, -0.00000691, 0.00000772, -0.00000282, -0.00000693, 0.00000176, -0.00000020, -0.00000024, 0.00000412, -0.00000168, -0.00000324, 0.00000078, -0.00000248, 0.00000363, -0.00000184, 0.00000352, -0.00001061, 0.00000542, -0.00000318, 0.00000751, -0.00001327, 0.00000800, 0.00000953, -0.00000463, -0.00000980, 0.00000245, 0.00000940, -0.00000781, -0.00000500, 0.00000102, 0.00000468, -0.00000250, 0.00000141, -0.00000414, 0.00000650, -0.00000950, 0.00001725, -0.00002336, 0.00001758, -0.00001445, -0.00000555, 0.00000530, -0.00000249, 0.00000088, -0.00001491, 0.00000403, -0.00001117, 0.00000016, -0.00001576, -0.00000088, -0.00000533, -0.00001165, -0.00004220, 0.00000117, -0.00007607, -0.00001966, -0.00046659, 0.00037868, 0.00099149, -0.00030322, -0.00020798, -0.00002665, -0.00005479, 0.00000611, -0.00001642, 0.00000162, -0.00003725, 0.00001457, -0.00000343, 0.00000095, -0.00001370, 0.00001168, -0.00000998, 0.00002349, -0.00003149, 0.00002180, -0.00001336, -0.00000827, 0.00000701, -0.00000992, 0.00000306, -0.00000859, -0.00000442, -0.00000100, 0.00001067, -0.00000790, -0.00000135, -0.00000554, 0.00001274, -0.00000964, -0.00001174, 0.00000087, 0.00002049, -0.00001277, -0.00000185, 0.00000080, 0.00000395, -0.00000314, -0.00000564, 0.00000521, 0.00000049, 0.00000299, -0.00000221, 0.00000194, -0.00000794, 0.00000394, -0.00000196, 0.00000200, 0.00000103, 0.00000078, -0.00001141, 0.00000932, -0.00000187, 0.00000252, 0.00000543, -0.00000450, -0.00000816, 0.00000240, 0.00000520, -0.00000205, -0.00000048, -0.00000475, -0.00000019, },
		{ -0.00000154, 0.00001049, -0.00000448, -0.00000553, -0.00000381, 0.00001021, -0.00000494, 0.00000487, 0.00000105, -0.00000950, 0.00000259, -0.00000089, 0.00000061, 0.00000231, 0.00000126, -0.00000553, 0.00000241, -0.00000448, 0.00000537, -0.00000362, 0.00000579, -0.00001174, 0.00000555, -0.00000339, 0.00000743, -0.00001152, 0.00000383, 0.00001211, -0.00000357, -0.00000998, 0.00000018, 0.00001075, -0.00000687, -0.00000589, 0.00000043, 0.00000495, -0.00000209, 0.00000105, -0.00000375, 0.00000666, -0.00001058, 0.00001595, -0.00001780, 0.00000852, -0.00000288, -0.00001697, 0.00001472, -0.00000968, 0.00000740, -0.00002111, 0.00001036, -0.00001736, 0.00000682, -0.00002322, 0.00000807, -0.00001465, -0.00000032, -0.00005535, 0.00001908, -0.00009484, 0.00001623, -0.00046970, 0.00021563, 0.00117016, -0.00033092, -0.00021038, -0.00003784, -0.00004913, -0.00000099, -0.00001131, -0.00000384, -0.00003280, 0.00000973, 0.00000041, -0.00000245, -0.00001115, 0.00000848, -0.00000749, 0.00001863, -0.00002374, 0.00001176, -0.00000028, -0.00002020, 0.00001539, -0.00001524, 0.00000752, -0.00001257, -0.00000068, -0.00000503, 0.00001381, -0.00000957, 0.00000109, -0.00000876, 0.00001409, -0.00000855, -0.00001095, -0.00000305, 0.00002176, -0.00000984, -0.00000440, 0.00000162, 0.00000294, -0.00000126, -0.00000722, 0.00000568, 0.00000009, 0.00000324, -0.00000224, 0.00000237, -0.00000759, 0.00000224, -0.00000056, 0.00000132, 0.00000112, 0.00000158, -0.00001103, 0.00000642, 0.00000137, -0.00000078, 0.00000779, -0.00000427, -0.00000754, -0.00000061, 0.00000739, -0.00000281, 0.00000065, -0.00000562, 0.00000054, },
		{ -0.00000250, 0.00001002, -0.00000226, -0.00000609, -0.00000476, 0.00000910, -0.00000222, 0.00000168, 0.00000464, -0.00001126, 0.00000266, -0.00000110, 0.00000118, 0.00000080, 0.00000354, -0.00000680, 0.00000297, -0.00000538, 0.00000590, -0.00000408, 0.00000645, -0.00001102, 0.00000392, -0.00000190, 0.00000553, -0.00000802, -0.00000135, 0.00001483, -0.00000289, -0.00000914, -0.00000289, 0.00001236, -0.00000632, -0.00000598, -0.00000094, 0.00000574, -0.00000218, 0.00000121, -0.00000386, 0.00000716, -0.00001132, 0.00001358, -0.00001130, -0.00000056, 0.00000724, -0.00002541, 0.00002012, -0.00001251, 0.00000959, -0.00002292, 0.00001211, -0.00001870, 0.00000835, -0.00002512, 0.00001065, -0.00001659, 0.00000252, -0.00005842, 0.00002397, -0.00009604, 0.00002387, -0.00042631, 0.00001145, 0.00135186, -0.00033961, -0.00022728, -0.00003940, -0.00005126, -0.00000185, -0.00001132, -0.00000497, -0.00003214, 0.00000816, 0.00000152, -0.00000362, -0.00001029, 0.00000647, -0.00000553, 0.00001380, -0.00001609, 0.00000227, 0.00001095, -0.00002883, 0.00001982, -0.00001682, 0.00000873, -0.00001360, 0.00000029, -0.00000630, 0.00001419, -0.00000880, 0.00000136, -0.00000957, 0.00001289, -0.00000553, -0.00001152, -0.00000502, 0.00002037, -0.00000484, -0.00000803, 0.00000326, 0.00000113, 0.00000116, -0.00000897, 0.00000618, -0.00000034, 0.00000347, -0.00000223, 0.00000263, -0.00000690, 0.00000042, 0.00000066, 0.00000087, 0.00000113, 0.00000207, -0.00000990, 0.00000298, 0.00000462, -0.00000362, 0.00000934, -0.00000352, -0.00000689, -0.00000341, 0.00000886, -0.00000289, 0.00000123, -0.00000593, 0.00000076, },
		{ -0.00000272, 0.00000868, 0.00000035, -0.00000665, -0.00000531, 0.00000730, 0.00000082, -0.00000130, 0.00000743, -0.00001182, 0.00000172, -0.00000062, 0.00000118, -0.00000006, 0.00000477, -0.00000670, 0.00000219, -0.00000490, 0.00000494, -0.00000296, 0.00000529, -0.00000839, 0.00000055, 0.00000121, 0.00000200, -0.00000310, -0.00000716, 0.00001751, -0.00000263, -0.00000722, -0.00000675, 0.00001432, -0.00000636, -0.00000503, -0.00000327, 0.00000729, -0.00000308, 0.00000228, -0.00000490, 0.00000845, -0.00001232, 0.00001115, -0.00000537, -0.00000772, 0.00001379, -0.00002887, 0.00001982, -0.00000957, 0.00000617, -0.00001908, 0.00000801, -0.00001385, 0.00000334, -0.00001988, 0.00000511, -0.00000929, -0.00000516, -0.00004910, 0.00001327, -0.00007697, 0.00000155, -0.00033996, -0.00022139, 0.00152015, -0.00031682, -0.00026611, -0.00002631, -0.00006504, 0.00000658, -0.00001903, 0.00000051, -0.00003729, 0.00001163, -0.00000172, -0.00000106, -0.00001257, 0.00000709, -0.00000562, 0.00001073, -0.00001057, -0.00000436, 0.00001798, -0.00003211, 0.00001873, -0.00001352, 0.00000580, -0.00001088, -0.00000225, -0.00000417, 0.00001129, -0.00000515, -0.00000098, -0.00000758, 0.00000891, -0.00000046, -0.00001369, -0.00000476, 0.00001630, 0.00000198, -0.00001248, 0.00000555, -0.00000137, 0.00000399, -0.00001080, 0.00000674, -0.00000091, 0.00000381, -0.00000234, 0.00000289, -0.00000615, -0.00000121, 0.00000140, 0.00000088, 0.00000087, 0.00000242, -0.00000837, -0.00000049, 0.00000730, -0.00000551, 0.00000973, -0.00000205, -0.00000648, -0.00000565, 0.00000933, -0.00000212, 0.00000113, -0.00000558, 0.00000039, },
		{ -0.00000226, 0.00000668, 0.00000306, -0.00000701, -0.00000565, 0.00000514, 0.00000372, -0.00000356, 0.00000896, -0.00001091, -0.00000035, 0.00000063, 0.00000047, -0.00000002, 0.00000468, -0.00000506, -0.00000005, -0.00000299, 0.00000248, -0.00000028, 0.00000242, -0.00000406, -0.00000424, 0.00000559, -0.00000275, 0.00000275, -0.00001312, 0.00001992, -0.00000280, -0.00000424, -0.00001131, 0.00001662, -0.00000716, -0.00000289, -0.00000669, 0.00000977, -0.00000506, 0.00000458, -0.00000724, 0.00001095, -0.00001419, 0.00000966, -0.00000144, -0.00001130, 0.00001513, -0.00002602, 0.00001287, -0.00000027, -0.00000329, -0.00000921, -0.00000234, -0.00000242, -0.00000866, -0.00000701, -0.00000905, 0.00000773, -0.00002379, -0.00002698, -0.00001322, -0.00003802, -0.00004833, -0.00021916, -0.00046627, 0.00165685, -0.00025056, -0.00033303, 0.00000520, -0.00009319, 0.00002642, -0.00003621, 0.00001414, -0.00004960, 0.00002136, -0.00001043, 0.00000632, -0.00001908, 0.00001152, -0.00000907, 0.00001099, -0.00000895, -0.00000623, 0.00001902, -0.00002868, 0.00001133, -0.00000494, -0.00000154, -0.00000418, -0.00000848, 0.00000151, 0.00000507, 0.00000135, -0.00000595, -0.00000280, 0.00000230, 0.00000647, -0.00001737, -0.00000231, 0.00000985, 0.00001009, -0.00001732, 0.00000823, -0.00000434, 0.00000706, -0.00001267, 0.00000742, -0.00000172, 0.00000442, -0.00000274, 0.00000339, -0.00000563, -0.00000228, 0.00000135, 0.00000157, 0.00000018, 0.00000285, -0.00000681, -0.00000346, 0.00000888, -0.00000605, 0.00000876, 0.00000025, -0.00000645, -0.00000708, 0.00000864, -0.00000046, 0.00000037, -0.00000461, -0.00000053, },
		{ -0.00000134, 0.00000432, 0.00000554, -0.00000694, -0.00000596, 0.00000295, 0.00000602, -0.00000469, 0.00000897, -0.00000847, -0.00000345, 0.00000260, -0.00000097, 0.00000099, 0.00000326, -0.00000196, -0.00000356, 0.00000015, -0.00000125, 0.00000367, -0.00000181, 0.00000150, -0.00000992, 0.00001070, -0.00000815, 0.00000890, -0.00001869, 0.00002181, -0.00000336, -0.00000030, -0.00001637, 0.00001921, -0.00000878, 0.00000048, -0.00001120, 0.00001323, -0.00000825, 0.00000827, -0.00001106, 0.00001494, -0.00001741, 0.00000993, -0.00000062, -0.00001014, 0.00001040, -0.00001646, -0.00000061, 0.00001497, -0.00001824, 0.00000610, -0.00001832, 0.00001493, -0.00002692, 0.00001273, -0.00003097, 0.00003337, -0.00005203, 0.00000624, -0.00005318, 0.00001728, -0.00011944, -0.00007647, -0.00070397, 0.00174398, -0.00013102, -0.00043168, 0.00005672, -0.00013665, 0.00005827, -0.00006336, 0.00003637, -0.00006946, 0.00003770, -0.00002497, 0.00001895, -0.00003034, 0.00002036, -0.00001668, 0.00001559, -0.00001241, -0.00000217, 0.00001318, -0.00001821, -0.00000223, 0.00000850, -0.00001281, 0.00000603, -0.00001797, 0.00001028, -0.00000399, 0.00001023, -0.00001310, 0.00000431, -0.00000637, 0.00001472, -0.00002222, 0.00000198, 0.00000160, 0.00001880, -0.00002202, 0.00001098, -0.00000755, 0.00001021, -0.00001453, 0.00000827, -0.00000288, 0.00000543, -0.00000360, 0.00000432, -0.00000560, -0.00000252, 0.00000030, 0.00000306, -0.00000105, 0.00000356, -0.00000557, -0.00000548, 0.00000901, -0.00000503, 0.00000641, 0.00000330, -0.00000686, -0.00000760, 0.00000681, 0.00000196, -0.00000091, -0.00000316, -0.00000185, },
		{ -0.00000022, 0.00000196, 0.00000748, -0.00000626, -0.00000643, 0.00000103, 0.00000739, -0.00000446, 0.00000741, -0.00000467, -0.00000733, 0.00000505, -0.00000299, 0.00000283, 0.00000067, 0.00000229, -0.00000796, 0.00000410, -0.00000580, 0.00000840, -0.00000681, 0.00000763, -0.00001580, 0.00001587, -0.00001351, 0.00001468, -0.00002338, 0.00002298, -0.00000425, 0.00000440, -0.00002168, 0.00002195, -0.00001115, 0.00000494, -0.00001660, 0.00001755, -0.00001257, 0.00001330, -0.00001634, 0.00002042, -0.00002217, 0.00001241, -0.00000346, -0.00000390, -0.00000032, -0.00000087, -0.00001946, 0.00003468, -0.00003712, 0.00002528, -0.00003829, 0.00003645, -0.00004960, 0.00003737, -0.00005838, 0.00006504, -0.00008683, 0.00004694, -0.00010203, 0.00008272, -0.00020242, 0.00007286, -0.00091465, 0.00176580, 0.00004779, -0.00056201, 0.00012693, -0.00019402, 0.00010082, -0.00009934, 0.00006622, -0.00009599, 0.00005988, -0.00004472, 0.00003634, -0.00004603, 0.00003346, -0.00002847, 0.00002475, -0.00002123, 0.00000799, 0.00000069, -0.00000146, -0.00002076, 0.00002546, -0.00002678, 0.00001867, -0.00002971, 0.00002114, -0.00001490, 0.00002055, -0.00002161, 0.00001294, -0.00001624, 0.00002349, -0.00002761, 0.00000747, -0.00000759, 0.00002725, -0.00002600, 0.00001346, -0.00001074, 0.00001326, -0.00001630, 0.00000933, -0.00000446, 0.00000693, -0.00000502, 0.00000581, -0.00000625, -0.00000177, -0.00000180, 0.00000536, -0.00000284, 0.00000466, -0.00000491, -0.00000625, 0.00000751, -0.00000248, 0.00000286, 0.00000689, -0.00000761, -0.00000726, 0.00000403, 0.00000486, -0.00000244, -0.00000147, -0.00000332, },
		{ 0.00000081, -0.00000009, 0.00000862, -0.00000487, -0.00000713, -0.00000043, 0.00000765, -0.00000284, 0.00000448, 0.00000010, -0.00001153, 0.00000762, -0.00000529, 0.00000523, -0.00000269, 0.00000718, -0.00001267, 0.00000828, -0.00001054, 0.00001324, -0.00001187, 0.00001355, -0.00002113, 0.00002038, -0.00001818, 0.00001949, -0.00002675, 0.00002327, -0.00000535, 0.00000957, -0.00002688, 0.00002460, -0.00001407, 0.00001019, -0.00002255, 0.00002240, -0.00001773, 0.00001936, -0.00002274, 0.00002710, -0.00002826, 0.00001703, -0.00000982, 0.00000691, -0.00001590, 0.00001902, -0.00004150, 0.00005650, -0.00005754, 0.00004592, -0.00005977, 0.00005957, -0.00007396, 0.00006390, -0.00008797, 0.00009894, -0.00012384, 0.00009003, -0.00015353, 0.00015024, -0.00028606, 0.00021265, -0.00107994, 0.00171075, 0.00028667, -0.00071961, 0.00021122, -0.00026137, 0.00015071, -0.00014130, 0.00010122, -0.00012703, 0.00008598, -0.00006799, 0.00005705, -0.00006491, 0.00004978, -0.00004363, 0.00003780, -0.00003472, 0.00002333, -0.00001709, 0.00001972, -0.00004213, 0.00004390, -0.00004163, 0.00003213, -0.00004222, 0.00003271, -0.00002630, 0.00003108, -0.00003036, 0.00002199, -0.00002619, 0.00003179, -0.00003274, 0.00001336, -0.00001676, 0.00003457, -0.00002872, 0.00001534, -0.00001366, 0.00001602, -0.00001793, 0.00001059, -0.00000647, 0.00000890, -0.00000699, 0.00000786, -0.00000760, -0.00000003, -0.00000487, 0.00000833, -0.00000508, 0.00000614, -0.00000495, -0.00000570, 0.00000450, 0.00000131, -0.00000149, 0.00001066, -0.00000846, -0.00000627, 0.00000060, 0.00000787, -0.00000392, 0.00000018, -0.00000468, },
		{ 0.00000149, -0.00000157, 0.00000880, -0.00000276, -0.00000807, -0.00000135, 0.00000679, -0.00000005, 0.00000060, 0.00000527, -0.00001550, 0.00000985, -0.00000751, 0.00000778, -0.00000631, 0.00001207, -0.00001701, 0.00001201, -0.00001481, 0.00001749, -0.00001624, 0.00001849, -0.00002519, 0.00002362, -0.00002157, 0.00002280, -0.00002850, 0.00002257, -0.00000654, 0.00001485, -0.00003154, 0.00002683, -0.00001720, 0.00001575, -0.00002848, 0.00002726, -0.00002320, 0.00002586, -0.00002962, 0.00003432, -0.00003510, 0.00002320, -0.00001886, 0.00002093, -0.00003437, 0.00004066, -0.00006387, 0.00007748, -0.00007662, 0.00006516, -0.00007982, 0.00008118, -0.00009670, 0.00008878, -0.00011577, 0.00013059, -0.00015799, 0.00012965, -0.00020064, 0.00021106, -0.00035883, 0.00032773, -0.00118507, 0.00157317, 0.00058037, -0.00089540, 0.00030172, -0.00033235, 0.00020268, -0.00018483, 0.00013759, -0.00015922, 0.00011302, -0.00009218, 0.00007877, -0.00008492, 0.00006748, -0.00006047, 0.00005316, -0.00005121, 0.00004189, -0.00003781, 0.00004260, -0.00006358, 0.00006128, -0.00005518, 0.00004451, -0.00005375, 0.00004333, -0.00003666, 0.00004042, -0.00003810, 0.00003023, -0.00003502, 0.00003860, -0.00003680, 0.00001877, -0.00002494, 0.00003995, -0.00002973, 0.00001640, -0.00001610, 0.00001833, -0.00001932, 0.00001199, -0.00000881, 0.00001122, -0.00000937, 0.00001034, -0.00000952, 0.00000250, -0.00000862, 0.00001165, -0.00000753, 0.00000786, -0.00000561, -0.00000395, 0.00000029, 0.00000585, -0.00000608, 0.00001413, -0.00000913, -0.00000491, -0.00000304, 0.00001055, -0.00000499, 0.00000150, -0.00000567, },
		{ 0.00000166, -0.00000234, 0.00000804, -0.00000009, -0.00000913, -0.00000179, 0.00000500, 0.00000353, -0.00000366, 0.00001018, -0.00001862, 0.00001130, -0.00000924, 0.00001002, -0.00000962, 0.00001630, -0.00002033, 0.00001466, -0.00001796, 0.00002050, -0.00001926, 0.00002182, -0.00002743, 0.00002510, -0.00002325, 0.00002430, -0.00002846, 0.00002089, -0.00000765, 0.00001982, -0.00003519, 0.00002826, -0.00002011, 0.00002101, -0.00003373, 0.00003148, -0.00002829, 0.00003201, -0.00003612, 0.00004118, -0.00004176, 0.00002989, -0.00002916, 0.00003616, -0.00005310, 0.00006100, -0.00008335, 0.00009452, -0.00009138, 0.00008006, -0.00009543, 0.00009812, -0.00011445, 0.00010835, -0.00013775, 0.00015547, -0.00018422, 0.00016003, -0.00023652, 0.00025688, -0.00041037, 0.00040573, -0.00122049, 0.00135427, 0.00091734, -0.00107601, 0.00038788, -0.00039879, 0.00025011, -0.00022444, 0.00017059, -0.00018842, 0.00013732, -0.00011402, 0.00009857, -0.00010340, 0.00008410, -0.00007669, 0.00006856, -0.00006829, 0.00006099, -0.00005842, 0.00006397, -0.00008206, 0.00007498, -0.00006526, 0.00005392, -0.00006258, 0.00005140, -0.00004450, 0.00004723, -0.00004363, 0.00003651, -0.00004162, 0.00004299, -0.00003901, 0.00002289, -0.00003124, 0.00004277, -0.00002878, 0.00001647, -0.00001788, 0.00002002, -0.00002034, 0.00001338, -0.00001126, 0.00001365, -0.00001190, 0.00001298, -0.00001172, 0.00000547, -0.00001263, 0.00001491, -0.00000984, 0.00000957, -0.00000669, -0.00000135, -0.00000457, 0.00001049, -0.00001028, 0.00001681, -0.00000927, -0.00000354, -0.00000646, 0.00001252, -0.00000537, 0.00000228, -0.00000610, },
		{ 0.00000125, -0.00000239, 0.00000645, 0.00000290, -0.00001011, -0.00000191, 0.00000258, 0.00000738, -0.00000763, 0.00001414, -0.00002037, 0.00001156, -0.00001011, 0.00001152, -0.00001208, 0.00001927, -0.00002206, 0.00001573, -0.00001951, 0.00002178, -0.00002045, 0.00002310, -0.00002750, 0.00002458, -0.00002304, 0.00002384, -0.00002664, 0.00001827, -0.00000855, 0.00002404, -0.00003737, 0.00002851, -0.00002230, 0.00002532, -0.00003758, 0.00003434, -0.00003221, 0.00003690, -0.00004124, 0.00004660, -0.00004709, 0.00003575, -0.00003897, 0.00005024, -0.00006923, 0.00007693, -0.00009693, 0.00010479, -0.00009921, 0.00008808, -0.00010398, 0.00010766, -0.00012431, 0.00011946, -0.00015040, 0.00016971, -0.00019822, 0.00017631, -0.00025553, 0.00028109, -0.00043287, 0.00043846, -0.00118300, 0.00106256, 0.00128028, -0.00124473, 0.00045746, -0.00045158, 0.00028579, -0.00025421, 0.00019519, -0.00021019, 0.00015498, -0.00013002, 0.00011330, -0.00011746, 0.00009691, -0.00008969, 0.00008140, -0.00008317, 0.00007759, -0.00007567, 0.00008059, -0.00009469, 0.00008267, -0.00007001, 0.00005877, -0.00006726, 0.00005557, -0.00004857, 0.00005044, -0.00004597, 0.00003987, -0.00004510, 0.00004426, -0.00003881, 0.00002509, -0.00003501, 0.00004264, -0.00002577, 0.00001551, -0.00001886, 0.00002094, -0.00002084, 0.00001457, -0.00001354, 0.00001584, -0.00001420, 0.00001540, -0.00001381, 0.00000842, -0.00001635, 0.00001759, -0.00001160, 0.00001094, -0.00000786, 0.00000163, -0.00000943, 0.00001454, -0.00001347, 0.00001825, -0.00000861, -0.00000246, -0.00000927, 0.00001343, -0.00000487, 0.00000238, -0.00000588, },
		{ 0.00000032, -0.00000184, 0.00000430, 0.00000587, -0.00001074, -0.00000195, -0.00000007, 0.00001092, -0.00001070, 0.00001660, -0.00002035, 0.00001040, -0.00000989, 0.00001197, -0.00001329, 0.00002056, -0.00002185, 0.00001492, -0.00001918, 0.00002109, -0.00001959, 0.00002216, -0.00002534, 0.00002207, -0.00002095, 0.00002152, -0.00002323, 0.00001485, -0.00000905, 0.00002710, -0.00003767, 0.00002722, -0.00002330, 0.00002801, -0.00003934, 0.00003516, -0.00003419, 0.00003965, -0.00004399, 0.00004952, -0.00004992, 0.00003938, -0.00004640, 0.00006082, -0.00008009, 0.00008579, -0.00010217, 0.00010623, -0.00009831, 0.00008748, -0.00010370, 0.00010791, -0.00012426, 0.00011991, -0.00015129, 0.00017063, -0.00019709, 0.00017528, -0.00025406, 0.00027964, -0.00042208, 0.00042289, -0.00107619, 0.00071328, 0.00164737, -0.00138307, 0.00049797, -0.00048184, 0.00030287, -0.00026861, 0.00020670, -0.00022045, 0.00016240, -0.00013700, 0.00012006, -0.00012438, 0.00010334, -0.00009695, 0.00008911, -0.00009312, 0.00008875, -0.00008657, 0.00008971, -0.00009922, 0.00008267, -0.00006819, 0.00005804, -0.00006685, 0.00005498, -0.00004810, 0.00004937, -0.00004453, 0.00003974, -0.00004494, 0.00004207, -0.00003594, 0.00002498, -0.00003587, 0.00003947, -0.00002085, 0.00001361, -0.00001899, 0.00002095, -0.00002065, 0.00001528, -0.00001529, 0.00001737, -0.00001585, 0.00001716, -0.00001531, 0.00001082, -0.00001924, 0.00001922, -0.00001242, 0.00001165, -0.00000873, 0.00000445, -0.00001362, 0.00001733, -0.00001512, 0.00001814, -0.00000698, -0.00000189, -0.00001117, 0.00001312, -0.00000345, 0.00000181, -0.00000505, },
		{ -0.00000096, -0.00000091, 0.00000192, 0.00000845, -0.00001075, -0.00000216, -0.00000254, 0.00001362, -0.00001234, 0.00001716, -0.00001838, 0.00000777, -0.00000852, 0.00001121, -0.00001303, 0.00001996, -0.00001957, 0.00001220, -0.00001696, 0.00001844, -0.00001673, 0.00001911, -0.00002116, 0.00001782, -0.00001725, 0.00001759, -0.00001850, 0.00001081, -0.00000904, 0.00002861, -0.00003578, 0.00002415, -0.00002272, 0.00002856, -0.00003848, 0.00003341, -0.00003361, 0.00003952, -0.00004357, 0.00004907, -0.00004927, 0.00003954, -0.00004985, 0.00006593, -0.00008365, 0.00008573, -0.00009769, 0.00009788, -0.00008797, 0.00007763, -0.00009392, 0.00009815, -0.00011352, 0.00010884, -0.00013945, 0.00015719, -0.00017977, 0.00015585, -0.00023100, 0.00025165, -0.00037780, 0.00036132, -0.00091002, 0.00032720, 0.00199416, -0.00147269, 0.00049823, -0.00048222, 0.00029593, -0.00026340, 0.00020158, -0.00021616, 0.00015692, -0.00013259, 0.00011667, -0.00012210, 0.00010139, -0.00009649, 0.00008962, -0.00009590, 0.00009214, -0.00008887, 0.00008943, -0.00009438, 0.00007431, -0.00005946, 0.00005144, -0.00006105, 0.00004934, -0.00004287, 0.00004390, -0.00003921, 0.00003595, -0.00004104, 0.00003643, -0.00003043, 0.00002249, -0.00003377, 0.00003347, -0.00001436, 0.00001094, -0.00001822, 0.00001994, -0.00001960, 0.00001525, -0.00001614, 0.00001785, -0.00001644, 0.00001783, -0.00001580, 0.00001218, -0.00002079, 0.00001936, -0.00001196, 0.00001139, -0.00000896, 0.00000662, -0.00001655, 0.00001836, -0.00001491, 0.00001636, -0.00000434, -0.00000195, -0.00001202, 0.00001156, -0.00000122, 0.00000071, -0.00000375, },
		{ -0.00000236, 0.00000012, -0.00000033, 0.00001030, -0.00000994, -0.00000274, -0.00000448, 0.00001508, -0.00001224, 0.00001570, -0.00001456, 0.00000386, -0.00000612, 0.00000927, -0.00001129, 0.00001752, -0.00001541, 0.00000781, -0.00001313, 0.00001410, -0.00001219, 0.00001435, -0.00001541, 0.00001228, -0.00001234, 0.00001247, -0.00001286, 0.00000637, -0.00000839, 0.00002833, -0.00003156, 0.00001923, -0.00002033, 0.00002661, -0.00003468, 0.00002880, -0.00003010, 0.00003607, -0.00003948, 0.00004470, -0.00004452, 0.00003543, -0.00004821, 0.00006432, -0.00007876, 0.00007609, -0.00008335, 0.00008006, -0.00006877, 0.00005915, -0.00007521, 0.00007896, -0.00009267, 0.00008682, -0.00011550, 0.00013014, -0.00014722, 0.00011927, -0.00018795, 0.00019945, -0.00030383, 0.00026104, -0.00069971, -0.00007133, 0.00229591, -0.00149744, 0.00045004, -0.00044806, 0.00026190, -0.00023632, 0.00017802, -0.00019580, 0.00013726, -0.00011567, 0.00010208, -0.00010959, 0.00009000, -0.00008718, 0.00008170, -0.00009016, 0.00008641, -0.00008142, 0.00007911, -0.00008012, 0.00005799, -0.00004440, 0.00003952, -0.00005031, 0.00003906, -0.00003328, 0.00003443, -0.00003040, 0.00002884, -0.00003372, 0.00002779, -0.00002268, 0.00001784, -0.00002896, 0.00002512, -0.00000679, 0.00000773, -0.00001657, 0.00001785, -0.00001755, 0.00001422, -0.00001577, 0.00001694, -0.00001563, 0.00001708, -0.00001491, 0.00001211, -0.00002064, 0.00001777, -0.00001006, 0.00000999, -0.00000829, 0.00000772, -0.00001779, 0.00001736, -0.00001278, 0.00001303, -0.00000087, -0.00000258, -0.00001182, 0.00000892, 0.00000157, -0.00000068, -0.00000220, },
		{ -0.00000360, 0.00000097, -0.00000213, 0.00001116, -0.00000820, -0.00000380, -0.00000567, 0.00001507, -0.00001033, 0.00001237, -0.00000922, -0.00000094, -0.00000299, 0.00000639, -0.00000828, 0.00001354, -0.00000978, 0.00000224, -0.00000814, 0.00000859, -0.00000649, 0.00000844, -0.00000871, 0.00000605, -0.00000677, 0.00000664, -0.00000674, 0.00000176, -0.00000705, 0.00002611, -0.00002507, 0.00001257, -0.00001612, 0.00002210, -0.00002795, 0.00002138, -0.00002366, 0.00002923, -0.00003167, 0.00003637, -0.00003558, 0.00002680, -0.00004107, 0.00005563, -0.00006546, 0.00005746, -0.00006038, 0.00005440, -0.00004249, 0.00003382, -0.00004936, 0.00005215, -0.00006360, 0.00005588, -0.00008163, 0.00009195, -0.00010231, 0.00006888, -0.00012898, 0.00012828, -0.00020742, 0.00013326, -0.00046405, -0.00045672, 0.00252998, -0.00144546, 0.00034953, -0.00037838, 0.00020070, -0.00018767, 0.00013641, -0.00015976, 0.00010384, -0.00008664, 0.00007661, -0.00008708, 0.00006932, -0.00006908, 0.00006527, -0.00007570, 0.00007143, -0.00006441, 0.00005944, -0.00005767, 0.00003522, -0.00002451, 0.00002355, -0.00003572, 0.00002514, -0.00002029, 0.00002189, -0.00001891, 0.00001913, -0.00002374, 0.00001691, -0.00001336, 0.00001152, -0.00002198, 0.00001510, 0.00000125, 0.00000426, -0.00001410, 0.00001465, -0.00001442, 0.00001204, -0.00001395, 0.00001444, -0.00001324, 0.00001474, -0.00001248, 0.00001044, -0.00001867, 0.00001441, -0.00000675, 0.00000746, -0.00000662, 0.00000756, -0.00001717, 0.00001434, -0.00000893, 0.00000847, 0.00000313, -0.00000360, -0.00001076, 0.00000548, 0.00000457, -0.00000205, -0.00000069, },
		{ -0.00000446, 0.00000139, -0.00000324, 0.00001092, -0.00000555, -0.00000534, -0.00000603, 0.00001359, -0.00000681, 0.00000758, -0.00000293, -0.00000607, 0.00000042, 0.00000291, -0.00000440, 0.00000851, -0.00000329, -0.00000388, -0.00000265, 0.00000253, -0.00000030, 0.00000208, -0.00000175, -0.00000023, -0.00000108, 0.00000061, -0.00000057, -0.00000279, -0.00000501, 0.00002200, -0.00001658, 0.00000449, -0.00001029, 0.00001524, -0.00001864, 0.00001156, -0.00001465, 0.00001937, -0.00002058, 0.00002455, -0.00002291, 0.00001408, -0.00002887, 0.00004053, -0.00004492, 0.00003168, -0.00003116, 0.00002361, -0.00001192, 0.00000438, -0.00001910, 0.00002054, -0.00002927, 0.00001915, -0.00004125, 0.00004645, -0.00004944, 0.00000974, -0.00006005, 0.00004554, -0.00009824, -0.00000835, -0.00022327, -0.00080464, 0.00267826, -0.00131076, 0.00019816, -0.00027637, 0.00011557, -0.00012046, 0.00007941, -0.00011046, 0.00005886, -0.00004750, 0.00004205, -0.00005614, 0.00004073, -0.00004340, 0.00004140, -0.00005356, 0.00004837, -0.00003934, 0.00003241, -0.00002937, 0.00000845, -0.00000198, 0.00000540, -0.00001889, 0.00000903, -0.00000526, 0.00000757, -0.00000590, 0.00000785, -0.00001208, 0.00000481, -0.00000331, 0.00000422, -0.00001354, 0.00000425, 0.00000912, 0.00000082, -0.00001089, 0.00001040, -0.00001024, 0.00000867, -0.00001063, 0.00001032, -0.00000930, 0.00001086, -0.00000855, 0.00000720, -0.00001495, 0.00000946, -0.00000224, 0.00000397, -0.00000400, 0.00000612, -0.00001475, 0.00000957, -0.00000381, 0.00000322, 0.00000720, -0.00000473, -0.00000911, 0.00000164, 0.00000738, -0.00000310, 0.00000052, },
		{ -0.00000476, 0.00000125, -0.00000355, 0.00000960, -0.00000217, -0.00000719, -0.00000564, 0.00001084, -0.00000210, 0.00000195, 0.00000358, -0.00001088, 0.00000362, -0.00000070, -0.00000014, 0.00000304, 0.00000336, -0.00000984, 0.00000268, -0.00000339, 0.00000569, -0.00000404, 0.00000479, -0.00000595, 0.00000420, -0.00000514, 0.00000525, -0.00000707, -0.00000234, 0.00001621, -0.00000658, -0.00000444, -0.00000330, 0.00000652, -0.00000741, 0.00000007, -0.00000381, 0.00000729, -0.00000708, 0.00001022, -0.00000755, -0.00000167, -0.00001279, 0.00002057, -0.00001932, 0.00000160, 0.00000100, -0.00000887, 0.00001954, -0.00002588, 0.00001228, -0.00001249, 0.00000675, -0.00001952, 0.00000143, -0.00000166, 0.00000612, -0.00005210, 0.00001180, -0.00004023, 0.00001293, -0.00014937, 0.00000307, -0.00109438, 0.00272904, -0.00109438, 0.00000307, -0.00014937, 0.00001293, -0.00004023, 0.00001180, -0.00005210, 0.00000612, -0.00000166, 0.00000143, -0.00001952, 0.00000675, -0.00001249, 0.00001228, -0.00002588, 0.00001954, -0.00000887, 0.00000100, 0.00000160, -0.00001932, 0.00002057, -0.00001279, -0.00000167, -0.00000755, 0.00001022, -0.00000708, 0.00000729, -0.00000381, 0.00000007, -0.00000741, 0.00000652, -0.00000330, -0.00000444, -0.00000658, 0.00001621, -0.00000234, -0.00000707, 0.00000525, -0.00000514, 0.00000420, -0.00000595, 0.00000479, -0.00000404, 0.00000569, -0.00000339, 0.00000268, -0.00000984, 0.00000336, 0.00000304, -0.00000014, -0.00000070, 0.00000362, -0.00001088, 0.00000358, 0.00000195, -0.00000210, 0.00001084, -0.00000564, -0.00000719, -0.00000217, 0.00000960, -0.00000355, 0.00000125, },
		{ -0.00000446, 0.00000052, -0.00000310, 0.00000738, 0.00000164, -0.00000911, -0.00000473, 0.00000720, 0.00000322, -0.00000381, 0.00000957, -0.00001475, 0.00000612, -0.00000400, 0.00000397, -0.00000224, 0.00000946, -0.00001495, 0.00000720, -0.00000855, 0.00001086, -0.00000930, 0.00001032, -0.00001063, 0.00000867, -0.00001024, 0.00001040, -0.00001089, 0.00000082, 0.00000912, 0.00000425, -0.00001354, 0.00000422, -0.00000331, 0.00000481, -0.00001208, 0.00000785, -0.00000590, 0.00000757, -0.00000526, 0.00000903, -0.00001889, 0.00000540, -0.00000198, 0.00000845, -0.00002937, 0.00003241, -0.00003934, 0.00004837, -0.00005356, 0.00004140, -0.00004340, 0.00004073, -0.00005614, 0.00004205, -0.00004750, 0.00005886, -0.00011046, 0.00007941, -0.00012046, 0.00011557, -0.00027637, 0.00019816, -0.00131076, 0.00267826, -0.00080464, -0.00022327, -0.00000835, -0.00009824, 0.00004554, -0.00006005, 0.00000974, -0.00004944, 0.00004645, -0.00004125, 0.00001915, -0.00002927, 0.00002054, -0.00001910, 0.00000438, -0.00001192, 0.00002361, -0.00003116, 0.00003168, -0.00004492, 0.00004053, -0.00002887, 0.00001408, -0.00002291, 0.00002455, -0.00002058, 0.00001937, -0.00001465, 0.00001156, -0.00001864, 0.00001524, -0.00001029, 0.00000449, -0.00001658, 0.00002200, -0.00000501, -0.00000279, -0.00000057, 0.00000061, -0.00000108, -0.00000023, -0.00000175, 0.00000208, -0.00000030, 0.00000253, -0.00000265, -0.00000388, -0.00000329, 0.00000851, -0.00000440, 0.00000291, 0.00000042, -0.00000607, -0.00000293, 0.00000758, -0.00000681, 0.00001359, -0.00000603, -0.00000534, -0.00000555, 0.00001092, -0.00000324, 0.00000139, },
		{ -0.00000360, -0.00000069, -0.00000205, 0.00000457, 0.00000548, -0.00001076, -0.00000360, 0.00000313, 0.00000847, -0.00000893, 0.00001434, -0.00001717, 0.00000756, -0.00000662, 0.00000746, -0.00000675, 0.00001441, -0.00001867, 0.00001044, -0.00001248, 0.00001474, -0.00001324, 0.00001444, -0.00001395, 0.00001204, -0.00001442, 0.00001465, -0.00001410, 0.00000426, 0.00000125, 0.00001510, -0.00002198, 0.00001152, -0.00001336, 0.00001691, -0.00002374, 0.00001913, -0.00001891, 0.00002189, -0.00002029, 0.00002514, -0.00003572, 0.00002355, -0.00002451, 0.00003522, -0.00005767, 0.00005944, -0.00006441, 0.00007143, -0.00007570, 0.00006527, -0.00006908, 0.00006932, -0.00008708, 0.00007661, -0.00008664, 0.00010384, -0.00015976, 0.00013641, -0.00018767, 0.00020070, -0.00037838, 0.00034953, -0.00144546, 0.00252998, -0.00045672, -0.00046405, 0.00013326, -0.00020742, 0.00012828, -0.00012898, 0.00006888, -0.00010231, 0.00009195, -0.00008163, 0.00005588, -0.00006360, 0.00005215, -0.00004936, 0.00003382, -0.00004249, 0.00005440, -0.00006038, 0.00005746, -0.00006546, 0.00005563, -0.00004107, 0.00002680, -0.00003558, 0.00003637, -0.00003167, 0.00002923, -0.00002366, 0.00002138, -0.00002795, 0.00002210, -0.00001612, 0.00001257, -0.00002507, 0.00002611, -0.00000705, 0.00000176, -0.00000674, 0.00000664, -0.00000677, 0.00000605, -0.00000871, 0.00000844, -0.00000649, 0.00000859, -0.00000814, 0.00000224, -0.00000978, 0.00001354, -0.00000828, 0.00000639, -0.00000299, -0.00000094, -0.00000922, 0.00001237, -0.00001033, 0.00001507, -0.00000567, -0.00000380, -0.00000820, 0.00001116, -0.00000213, 0.00000097, },
		{ -0.00000236, -0.00000220, -0.00000068, 0.00000157, 0.00000892, -0.00001182, -0.00000258, -0.00000087, 0.00001303, -0.00001278, 0.00001736, -0.00001779, 0.00000772, -0.00000829, 0.00000999, -0.00001006, 0.00001777, -0.00002064, 0.00001211, -0.00001491, 0.00001708, -0.00001563, 0.00001694, -0.00001577, 0.00001422, -0.00001755, 0.00001785, -0.00001657, 0.00000773, -0.00000679, 0.00002512, -0.00002896, 0.00001784, -0.00002268, 0.00002779, -0.00003372, 0.00002884, -0.00003040, 0.00003443, -0.00003328, 0.00003906, -0.00005031, 0.00003952, -0.00004440, 0.00005799, -0.00008012, 0.00007911, -0.00008142, 0.00008641, -0.00009016, 0.00008170, -0.00008718, 0.00009000, -0.00010959, 0.00010208, -0.00011567, 0.00013726, -0.00019580, 0.00017802, -0.00023632, 0.00026190, -0.00044806, 0.00045004, -0.00149744, 0.00229591, -0.00007133, -0.00069971, 0.00026104, -0.00030383, 0.00019945, -0.00018795, 0.00011927, -0.00014722, 0.00013014, -0.00011550, 0.00008682, -0.00009267, 0.00007896, -0.00007521, 0.00005915, -0.00006877, 0.00008006, -0.00008335, 0.00007609, -0.00007876, 0.00006432, -0.00004821, 0.00003543, -0.00004452, 0.00004470, -0.00003948, 0.00003607, -0.00003010, 0.00002880, -0.00003468, 0.00002661, -0.00002033, 0.00001923, -0.00003156, 0.00002833, -0.00000839, 0.00000637, -0.00001286, 0.00001247, -0.00001234, 0.00001228, -0.00001541, 0.00001435, -0.00001219, 0.00001410, -0.00001313, 0.00000781, -0.00001541, 0.00001752, -0.00001129, 0.00000927, -0.00000612, 0.00000386, -0.00001456, 0.00001570, -0.00001224, 0.00001508, -0.00000448, -0.00000274, -0.00000994, 0.00001030, -0.00000033, 0.00000012, },
		{ -0.00000096, -0.00000375, 0.00000071, -0.00000122, 0.00001156, -0.00001202, -0.00000195, -0.00000434, 0.00001636, -0.00001491, 0.00001836, -0.00001655, 0.00000662, -0.00000896, 0.00001139, -0.00001196, 0.00001936, -0.00002079, 0.00001218, -0.00001580, 0.00001783, -0.00001644, 0.00001785, -0.00001614, 0.00001525, -0.00001960, 0.00001994, -0.00001822, 0.00001094, -0.00001436, 0.00003347, -0.00003377, 0.00002249, -0.00003043, 0.00003643, -0.00004104, 0.00003595, -0.00003921, 0.00004390, -0.00004287, 0.00004934, -0.00006105, 0.00005144, -0.00005946, 0.00007431, -0.00009438, 0.00008943, -0.00008887, 0.00009214, -0.00009590, 0.00008962, -0.00009649, 0.00010139, -0.00012210, 0.00011667, -0.00013259, 0.00015692, -0.00021616, 0.00020158, -0.00026340, 0.00029593, -0.00048222, 0.00049823, -0.00147269, 0.00199416, 0.00032720, -0.00091002, 0.00036132, -0.00037780, 0.00025165, -0.00023100, 0.00015585, -0.00017977, 0.00015719, -0.00013945, 0.00010884, -0.00011352, 0.00009815, -0.00009392, 0.00007763, -0.00008797, 0.00009788, -0.00009769, 0.00008573, -0.00008365, 0.00006593, -0.00004985, 0.00003954, -0.00004927, 0.00004907, -0.00004357, 0.00003952, -0.00003361, 0.00003341, -0.00003848, 0.00002856, -0.00002272, 0.00002415, -0.00003578, 0.00002861, -0.00000904, 0.00001081, -0.00001850, 0.00001759, -0.00001725, 0.00001782, -0.00002116, 0.00001911, -0.00001673, 0.00001844, -0.00001696, 0.00001220, -0.00001957, 0.00001996, -0.00001303, 0.00001121, -0.00000852, 0.00000777, -0.00001838, 0.00001716, -0.00001234, 0.00001362, -0.00000254, -0.00000216, -0.00001075, 0.00000845, 0.00000192, -0.00000091, },
		{ 0.00000032, -0.00000505, 0.00000181, -0.00000345, 0.00001312, -0.00001117, -0.00000189, -0.00000698, 0.00001814, -0.00001512, 0.00001733, -0.00001362, 0.00000445, -0.00000873, 0.00001165, -0.00001242, 0.00001922, -0.00001924, 0.00001082, -0.00001531, 0.00001716, -0.00001585, 0.00001737, -0.00001529, 0.00001528, -0.00002065, 0.00002095, -0.00001899, 0.00001361, -0.00002085, 0.00003947, -0.00003587, 0.00002498, -0.00003594, 0.00004207, -0.00004494, 0.00003974, -0.00004453, 0.00004937, -0.00004810, 0.00005498, -0.00006685, 0.00005804, -0.00006819, 0.00008267, -0.00009922, 0.00008971, -0.00008657, 0.00008875, -0.00009312, 0.00008911, -0.00009695, 0.00010334, -0.00012438, 0.00012006, -0.00013700, 0.00016240, -0.00022045, 0.00020670, -0.00026861, 0.00030287, -0.00048184, 0.00049797, -0.00138307, 0.00164737, 0.00071328, -0.00107619, 0.00042289, -0.00042208, 0.00027964, -0.00025406, 0.00017528, -0.00019709, 0.00017063, -0.00015129, 0.00011991, -0.00012426, 0.00010791, -0.00010370, 0.00008748, -0.00009831, 0.00010623, -0.00010217, 0.00008579, -0.00008009, 0.00006082, -0.00004640, 0.00003938, -0.00004992, 0.00004952, -0.00004399, 0.00003965, -0.00003419, 0.00003516, -0.00003934, 0.00002801, -0.00002330, 0.00002722, -0.00003767, 0.00002710, -0.00000905, 0.00001485, -0.00002323, 0.00002152, -0.00002095, 0.00002207, -0.00002534, 0.00002216, -0.00001959, 0.00002109, -0.00001918, 0.00001492, -0.00002185, 0.00002056, -0.00001329, 0.00001197, -0.00000989, 0.00001040, -0.00002035, 0.00001660, -0.00001070, 0.00001092, -0.00000007, -0.00000195, -0.00001074, 0.00000587, 0.00000430, -0.00000184, },
		{ 0.00000125, -0.00000588, 0.00000238, -0.00000487, 0.00001343, -0.00000927, -0.00000246, -0.00000861, 0.00001825, -0.00001347, 0.00001454, -0.00000943, 0.00000163, -0.00000786, 0.00001094, -0.00001160, 0.00001759, -0.00001635, 0.00000842, -0.00001381, 0.00001540, -0.00001420, 0.00001584, -0.00001354, 0.00001457, -0.00002084, 0.00002094, -0.00001886, 0.00001551, -0.00002577, 0.00004264, -0.00003501, 0.00002509, -0.00003881, 0.00004426, -0.00004510, 0.00003987, -0.00004597, 0.00005044, -0.00004857, 0.00005557, -0.00006726, 0.00005877, -0.00007001, 0.00008267, -0.00009469, 0.00008059, -0.00007567, 0.00007759, -0.00008317, 0.00008140, -0.00008969, 0.00009691, -0.00011746, 0.00011330, -0.00013002, 0.00015498, -0.00021019, 0.00019519, -0.00025421, 0.00028579, -0.00045158, 0.00045746, -0.00124473, 0.00128028, 0.00106256, -0.00118300, 0.00043846, -0.00043287, 0.00028109, -0.00025553, 0.00017631, -0.00019822, 0.00016971, -0.00015040, 0.00011946, -0.00012431, 0.00010766, -0.00010398, 0.00008808, -0.00009921, 0.00010479, -0.00009693, 0.00007693, -0.00006923, 0.00005024, -0.00003897, 0.00003575, -0.00004709, 0.00004660, -0.00004124, 0.00003690, -0.00003221, 0.00003434, -0.00003758, 0.00002532, -0.00002230, 0.00002851, -0.00003737, 0.00002404, -0.00000855, 0.00001827, -0.00002664, 0.00002384, -0.00002304, 0.00002458, -0.00002750, 0.00002310, -0.00002045, 0.00002178, -0.00001951, 0.00001573, -0.00002206, 0.00001927, -0.00001208, 0.00001152, -0.00001011, 0.00001156, -0.00002037, 0.00001414, -0.00000763, 0.00000738, 0.00000258, -0.00000191, -0.00001011, 0.00000290, 0.00000645, -0.00000239, },
		{ 0.00000166, -0.00000610, 0.00000228, -0.00000537, 0.00001252, -0.00000646, -0.00000354, -0.00000927, 0.00001681, -0.00001028, 0.00001049, -0.00000457, -0.00000135, -0.00000669, 0.00000957, -0.00000984, 0.00001491, -0.00001263, 0.00000547, -0.00001172, 0.00001298, -0.00001190, 0.00001365, -0.00001126, 0.00001338, -0.00002034, 0.00002002, -0.00001788, 0.00001647, -0.00002878, 0.00004277, -0.00003124, 0.00002289, -0.00003901, 0.00004299, -0.00004162, 0.00003651, -0.00004363, 0.00004723, -0.00004450, 0.00005140, -0.00006258, 0.00005392, -0.00006526, 0.00007498, -0.00008206, 0.00006397, -0.00005842, 0.00006099, -0.00006829, 0.00006856, -0.00007669, 0.00008410, -0.00010340, 0.00009857, -0.00011402, 0.00013732, -0.00018842, 0.00017059, -0.00022444, 0.00025011, -0.00039879, 0.00038788, -0.00107601, 0.00091734, 0.00135427, -0.00122049, 0.00040573, -0.00041037, 0.00025688, -0.00023652, 0.00016003, -0.00018422, 0.00015547, -0.00013775, 0.00010835, -0.00011445, 0.00009812, -0.00009543, 0.00008006, -0.00009138, 0.00009452, -0.00008335, 0.00006100, -0.00005310, 0.00003616, -0.00002916, 0.00002989, -0.00004176, 0.00004118, -0.00003612, 0.00003201, -0.00002829, 0.00003148, -0.00003373, 0.00002101, -0.00002011, 0.00002826, -0.00003519, 0.00001982, -0.00000765, 0.00002089, -0.00002846, 0.00002430, -0.00002325, 0.00002510, -0.00002743, 0.00002182, -0.00001926, 0.00002050, -0.00001796, 0.00001466, -0.00002033, 0.00001630, -0.00000962, 0.00001002, -0.00000924, 0.00001130, -0.00001862, 0.00001018, -0.00000366, 0.00000353, 0.00000500, -0.00000179, -0.00000913, -0.00000009, 0.00000804, -0.00000234, },
		{ 0.00000149, -0.00000567, 0.00000150, -0.00000499, 0.00001055, -0.00000304, -0.00000491, -0.00000913, 0.00001413, -0.00000608, 0.00000585, 0.00000029, -0.00000395, -0.00000561, 0.00000786, -0.00000753, 0.00001165, -0.00000862, 0.00000250, -0.00000952, 0.00001034, -0.00000937, 0.00001122, -0.00000881, 0.00001199, -0.00001932, 0.00001833, -0.00001610, 0.00001640, -0.00002973, 0.00003995, -0.00002494, 0.00001877, -0.00003680, 0.00003860, -0.00003502, 0.00003023, -0.00003810, 0.00004042, -0.00003666, 0.00004333, -0.00005375, 0.00004451, -0.00005518, 0.00006128, -0.00006358, 0.00004260, -0.00003781, 0.00004189, -0.00005121, 0.00005316, -0.00006047, 0.00006748, -0.00008492, 0.00007877, -0.00009218, 0.00011302, -0.00015922, 0.00013759, -0.00018483, 0.00020268, -0.00033235, 0.00030172, -0.00089540, 0.00058037, 0.00157317, -0.00118507, 0.00032773, -0.00035883, 0.00021106, -0.00020064, 0.00012965, -0.00015799, 0.00013059, -0.00011577, 0.00008878, -0.00009670, 0.00008118, -0.00007982, 0.00006516, -0.00007662, 0.00007748, -0.00006387, 0.00004066, -0.00003437, 0.00002093, -0.00001886, 0.00002320, -0.00003510, 0.00003432, -0.00002962, 0.00002586, -0.00002320, 0.00002726, -0.00002848, 0.00001575, -0.00001720, 0.00002683, -0.00003154, 0.00001485, -0.00000654, 0.00002257, -0.00002850, 0.00002280, -0.00002157, 0.00002362, -0.00002519, 0.00001849, -0.00001624, 0.00001749, -0.00001481, 0.00001201, -0.00001701, 0.00001207, -0.00000631, 0.00000778, -0.00000751, 0.00000985, -0.00001550, 0.00000527, 0.00000060, -0.00000005, 0.00000679, -0.00000135, -0.00000807, -0.00000276, 0.00000880, -0.00000157, },
		{ 0.00000081, -0.00000468, 0.00000018, -0.00000392, 0.00000787, 0.00000060, -0.00000627, -0.00000846, 0.00001066, -0.00000149, 0.00000131, 0.00000450, -0.00000570, -0.00000495, 0.00000614, -0.00000508, 0.00000833, -0.00000487, -0.00000003, -0.00000760, 0.00000786, -0.00000699, 0.00000890, -0.00000647, 0.00001059, -0.00001793, 0.00001602, -0.00001366, 0.00001534, -0.00002872, 0.00003457, -0.00001676, 0.00001336, -0.00003274, 0.00003179, -0.00002619, 0.00002199, -0.00003036, 0.00003108, -0.00002630, 0.00003271, -0.00004222, 0.00003213, -0.00004163, 0.00004390, -0.00004213, 0.00001972, -0.00001709, 0.00002333, -0.00003472, 0.00003780, -0.00004363, 0.00004978, -0.00006491, 0.00005705, -0.00006799, 0.00008598, -0.00012703, 0.00010122, -0.00014130, 0.00015071, -0.00026137, 0.00021122, -0.00071961, 0.00028667, 0.00171075, -0.00107994, 0.00021265, -0.00028606, 0.00015024, -0.00015353, 0.00009003, -0.00012384, 0.00009894, -0.00008797, 0.00006390, -0.00007396, 0.00005957, -0.00005977, 0.00004592, -0.00005754, 0.00005650, -0.00004150, 0.00001902, -0.00001590, 0.00000691, -0.00000982, 0.00001703, -0.00002826, 0.00002710, -0.00002274, 0.00001936, -0.00001773, 0.00002240, -0.00002255, 0.00001019, -0.00001407, 0.00002460, -0.00002688, 0.00000957, -0.00000535, 0.00002327, -0.00002675, 0.00001949, -0.00001818, 0.00002038, -0.00002113, 0.00001355, -0.00001187, 0.00001324, -0.00001054, 0.00000828, -0.00001267, 0.00000718, -0.00000269, 0.00000523, -0.00000529, 0.00000762, -0.00001153, 0.00000010, 0.00000448, -0.00000284, 0.00000765, -0.00000043, -0.00000713, -0.00000487, 0.00000862, -0.00000009, },
		{ -0.00000022, -0.00000332, -0.00000147, -0.00000244, 0.00000486, 0.00000403, -0.00000726, -0.00000761, 0.00000689, 0.00000286, -0.00000248, 0.00000751, -0.00000625, -0.00000491, 0.00000466, -0.00000284, 0.00000536, -0.00000180, -0.00000177, -0.00000625, 0.00000581, -0.00000502, 0.00000693, -0.00000446, 0.00000933, -0.00001630, 0.00001326, -0.00001074, 0.00001346, -0.00002600, 0.00002725, -0.00000759, 0.00000747, -0.00002761, 0.00002349, -0.00001624, 0.00001294, -0.00002161, 0.00002055, -0.00001490, 0.00002114, -0.00002971, 0.00001867, -0.00002678, 0.00002546, -0.00002076, -0.00000146, 0.00000069, 0.00000799, -0.00002123, 0.00002475, -0.00002847, 0.00003346, -0.00004603, 0.00003634, -0.00004472, 0.00005988, -0.00009599, 0.00006622, -0.00009934, 0.00010082, -0.00019402, 0.00012693, -0.00056201, 0.00004779, 0.00176580, -0.00091465, 0.00007286, -0.00020242, 0.00008272, -0.00010203, 0.00004694, -0.00008683, 0.00006504, -0.00005838, 0.00003737, -0.00004960, 0.00003645, -0.00003829, 0.00002528, -0.00003712, 0.00003468, -0.00001946, -0.00000087, -0.00000032, -0.00000390, -0.00000346, 0.00001241, -0.00002217, 0.00002042, -0.00001634, 0.00001330, -0.00001257, 0.00001755, -0.00001660, 0.00000494, -0.00001115, 0.00002195, -0.00002168, 0.00000440, -0.00000425, 0.00002298, -0.00002338, 0.00001468, -0.00001351, 0.00001587, -0.00001580, 0.00000763, -0.00000681, 0.00000840, -0.00000580, 0.00000410, -0.00000796, 0.00000229, 0.00000067, 0.00000283, -0.00000299, 0.00000505, -0.00000733, -0.00000467, 0.00000741, -0.00000446, 0.00000739, 0.00000103, -0.00000643, -0.00000626, 0.00000748, 0.00000196, },
		{ -0.00000134, -0.00000185, -0.00000316, -0.00000091, 0.00000196, 0.00000681, -0.00000760, -0.00000686, 0.00000330, 0.00000641, -0.00000503, 0.00000901, -0.00000548, -0.00000557, 0.00000356, -0.00000105, 0.00000306, 0.00000030, -0.00000252, -0.00000560, 0.00000432, -0.00000360, 0.00000543, -0.00000288, 0.00000827, -0.00001453, 0.00001021, -0.00000755, 0.00001098, -0.00002202, 0.00001880, 0.00000160, 0.00000198, -0.00002222, 0.00001472, -0.00000637, 0.00000431, -0.00001310, 0.00001023, -0.00000399, 0.00001028, -0.00001797, 0.00000603, -0.00001281, 0.00000850, -0.00000223, -0.00001821, 0.00001318, -0.00000217, -0.00001241, 0.00001559, -0.00001668, 0.00002036, -0.00003034, 0.00001895, -0.00002497, 0.00003770, -0.00006946, 0.00003637, -0.00006336, 0.00005827, -0.00013665, 0.00005672, -0.00043168, -0.00013102, 0.00174398, -0.00070397, -0.00007647, -0.00011944, 0.00001728, -0.00005318, 0.00000624, -0.00005203, 0.00003337, -0.00003097, 0.00001273, -0.00002692, 0.00001493, -0.00001832, 0.00000610, -0.00001824, 0.00001497, -0.00000061, -0.00001646, 0.00001040, -0.00001014, -0.00000062, 0.00000993, -0.00001741, 0.00001494, -0.00001106, 0.00000827, -0.00000825, 0.00001323, -0.00001120, 0.00000048, -0.00000878, 0.00001921, -0.00001637, -0.00000030, -0.00000336, 0.00002181, -0.00001869, 0.00000890, -0.00000815, 0.00001070, -0.00000992, 0.00000150, -0.00000181, 0.00000367, -0.00000125, 0.00000015, -0.00000356, -0.00000196, 0.00000326, 0.00000099, -0.00000097, 0.00000260, -0.00000345, -0.00000847, 0.00000897, -0.00000469, 0.00000602, 0.00000295, -0.00000596, -0.00000694, 0.00000554, 0.00000432, },
		{ -0.00000226, -0.00000053, -0.00000461, 0.00000037, -0.00000046, 0.00000864, -0.00000708, -0.00000645, 0.00000025, 0.00000876, -0.00000605, 0.00000888, -0.00000346, -0.00000681, 0.00000285, 0.00000018, 0.00000157, 0.00000135, -0.00000228, -0.00000563, 0.00000339, -0.00000274, 0.00000442, -0.00000172, 0.00000742, -0.00001267, 0.00000706, -0.00000434, 0.00000823, -0.00001732, 0.00001009, 0.00000985, -0.00000231, -0.00001737, 0.00000647, 0.00000230, -0.00000280, -0.00000595, 0.00000135, 0.00000507, 0.00000151, -0.00000848, -0.00000418, -0.00000154, -0.00000494, 0.00001133, -0.00002868, 0.00001902, -0.00000623, -0.00000895, 0.00001099, -0.00000907, 0.00001152, -0.00001908, 0.00000632, -0.00001043, 0.00002136, -0.00004960, 0.00001414, -0.00003621, 0.00002642, -0.00009319, 0.00000520, -0.00033303, -0.00025056, 0.00165685, -0.00046627, -0.00021916, -0.00004833, -0.00003802, -0.00001322, -0.00002698, -0.00002379, 0.00000773, -0.00000905, -0.00000701, -0.00000866, -0.00000242, -0.00000234, -0.00000921, -0.00000329, -0.00000027, 0.00001287, -0.00002602, 0.00001513, -0.00001130, -0.00000144, 0.00000966, -0.00001419, 0.00001095, -0.00000724, 0.00000458, -0.00000506, 0.00000977, -0.00000669, -0.00000289, -0.00000716, 0.00001662, -0.00001131, -0.00000424, -0.00000280, 0.00001992, -0.00001312, 0.00000275, -0.00000275, 0.00000559, -0.00000424, -0.00000406, 0.00000242, -0.00000028, 0.00000248, -0.00000299, -0.00000005, -0.00000506, 0.00000468, -0.00000002, 0.00000047, 0.00000063, -0.00000035, -0.00001091, 0.00000896, -0.00000356, 0.00000372, 0.00000514, -0.00000565, -0.00000701, 0.00000306, 0.00000668, },
		{ -0.00000272, 0.00000039, -0.00000558, 0.00000113, -0.00000212, 0.00000933, -0.00000565, -0.00000648, -0.00000205, 0.00000973, -0.00000551, 0.00000730, -0.00000049, -0.00000837, 0.00000242, 0.00000087, 0.00000088, 0.00000140, -0.00000121, -0.00000615, 0.00000289, -0.00000234, 0.00000381, -0.00000091, 0.00000674, -0.00001080, 0.00000399, -0.00000137, 0.00000555, -0.00001248, 0.00000198, 0.00001630, -0.00000476, -0.00001369, -0.00000046, 0.00000891, -0.00000758, -0.00000098, -0.00000515, 0.00001129, -0.00000417, -0.00000225, -0.00001088, 0.00000580, -0.00001352, 0.00001873, -0.00003211, 0.00001798, -0.00000436, -0.00001057, 0.00001073, -0.00000562, 0.00000709, -0.00001257, -0.00000106, -0.00000172, 0.00001163, -0.00003729, 0.00000051, -0.00001903, 0.00000658, -0.00006504, -0.00002631, -0.00026611, -0.00031682, 0.00152015, -0.00022139, -0.00033996, 0.00000155, -0.00007697, 0.00001327, -0.00004910, -0.00000516, -0.00000929, 0.00000511, -0.00001988, 0.00000334, -0.00001385, 0.00000801, -0.00001908, 0.00000617, -0.00000957, 0.00001982, -0.00002887, 0.00001379, -0.00000772, -0.00000537, 0.00001115, -0.00001232, 0.00000845, -0.00000490, 0.00000228, -0.00000308, 0.00000729, -0.00000327, -0.00000503, -0.00000636, 0.00001432, -0.00000675, -0.00000722, -0.00000263, 0.00001751, -0.00000716, -0.00000310, 0.00000200, 0.00000121, 0.00000055, -0.00000839, 0.00000529, -0.00000296, 0.00000494, -0.00000490, 0.00000219, -0.00000670, 0.00000477, -0.00000006, 0.00000118, -0.00000062, 0.00000172, -0.00001182, 0.00000743, -0.00000130, 0.00000082, 0.00000730, -0.00000531, -0.00000665, 0.00000035, 0.00000868, },
		{ -0.00000250, 0.00000076, -0.00000593, 0.00000123, -0.00000289, 0.00000886, -0.00000341, -0.00000689, -0.00000352, 0.00000934, -0.00000362, 0.00000462, 0.00000298, -0.00000990, 0.00000207, 0.00000113, 0.00000087, 0.00000066, 0.00000042, -0.00000690, 0.00000263, -0.00000223, 0.00000347, -0.00000034, 0.00000618, -0.00000897, 0.00000116, 0.00000113, 0.00000326, -0.00000803, -0.00000484, 0.00002037, -0.00000502, -0.00001152, -0.00000553, 0.00001289, -0.00000957, 0.00000136, -0.00000880, 0.00001419, -0.00000630, 0.00000029, -0.00001360, 0.00000873, -0.00001682, 0.00001982, -0.00002883, 0.00001095, 0.00000227, -0.00001609, 0.00001380, -0.00000553, 0.00000647, -0.00001029, -0.00000362, 0.00000152, 0.00000816, -0.00003214, -0.00000497, -0.00001132, -0.00000185, -0.00005126, -0.00003940, -0.00022728, -0.00033961, 0.00135186, 0.00001145, -0.00042631, 0.00002387, -0.00009604, 0.00002397, -0.00005842, 0.00000252, -0.00001659, 0.00001065, -0.00002512, 0.00000835, -0.00001870, 0.00001211, -0.00002292, 0.00000959, -0.00001251, 0.00002012, -0.00002541, 0.00000724, -0.00000056, -0.00001130, 0.00001358, -0.00001132, 0.00000716, -0.00000386, 0.00000121, -0.00000218, 0.00000574, -0.00000094, -0.00000598, -0.00000632, 0.00001236, -0.00000289, -0.00000914, -0.00000289, 0.00001483, -0.00000135, -0.00000802, 0.00000553, -0.00000190, 0.00000392, -0.00001102, 0.00000645, -0.00000408, 0.00000590, -0.00000538, 0.00000297, -0.00000680, 0.00000354, 0.00000080, 0.00000118, -0.00000110, 0.00000266, -0.00001126, 0.00000464, 0.00000168, -0.00000222, 0.00000910, -0.00000476, -0.00000609, -0.00000226, 0.00001002, },
		{ -0.00000154, 0.00000054, -0.00000562, 0.00000065, -0.00000281, 0.00000739, -0.00000061, -0.00000754, -0.00000427, 0.00000779, -0.00000078, 0.00000137, 0.00000642, -0.00001103, 0.00000158, 0.00000112, 0.00000132, -0.00000056, 0.00000224, -0.00000759, 0.00000237, -0.00000224, 0.00000324, 0.00000009, 0.00000568, -0.00000722, -0.00000126, 0.00000294, 0.00000162, -0.00000440, -0.00000984, 0.00002176, -0.00000305, -0.00001095, -0.00000855, 0.00001409, -0.00000876, 0.00000109, -0.00000957, 0.00001381, -0.00000503, -0.00000068, -0.00001257, 0.00000752, -0.00001524, 0.00001539, -0.00002020, -0.00000028, 0.00001176, -0.00002374, 0.00001863, -0.00000749, 0.00000848, -0.00001115, -0.00000245, 0.00000041, 0.00000973, -0.00003280, -0.00000384, -0.00001131, -0.00000099, -0.00004913, -0.00003784, -0.00021038, -0.00033092, 0.00117016, 0.00021563, -0.00046970, 0.00001623, -0.00009484, 0.00001908, -0.00005535, -0.00000032, -0.00001465, 0.00000807, -0.00002322, 0.00000682, -0.00001736, 0.00001036, -0.00002111, 0.00000740, -0.00000968, 0.00001472, -0.00001697, -0.00000288, 0.00000852, -0.00001780, 0.00001595, -0.00001058, 0.00000666, -0.00000375, 0.00000105, -0.00000209, 0.00000495, 0.00000043, -0.00000589, -0.00000687, 0.00001075, 0.00000018, -0.00000998, -0.00000357, 0.00001211, 0.00000383, -0.00001152, 0.00000743, -0.00000339, 0.00000555, -0.00001174, 0.00000579, -0.00000362, 0.00000537, -0.00000448, 0.00000241, -0.00000553, 0.00000126, 0.00000231, 0.00000061, -0.00000089, 0.00000259, -0.00000950, 0.00000105, 0.00000487, -0.00000494, 0.00001021, -0.00000381, -0.00000553, -0.00000448, 0.00001049, },
		{ 0.00000014, -0.00000019, -0.00000475, -0.00000048, -0.00000205, 0.00000520, 0.00000240, -0.00000816, -0.00000450, 0.00000543, 0.00000252, -0.00000187, 0.00000932, -0.00001141, 0.00000078, 0.00000103, 0.00000200, -0.00000196, 0.00000394, -0.00000794, 0.00000194, -0.00000221, 0.00000299, 0.00000049, 0.00000521, -0.00000564, -0.00000314, 0.00000395, 0.00000080, -0.00000185, -0.00001277, 0.00002049, 0.00000087, -0.00001174, -0.00000964, 0.00001274, -0.00000554, -0.00000135, -0.00000790, 0.00001067, -0.00000100, -0.00000442, -0.00000859, 0.00000306, -0.00000992, 0.00000701, -0.00000827, -0.00001336, 0.00002180, -0.00003149, 0.00002349, -0.00000998, 0.00001168, -0.00001370, 0.00000095, -0.00000343, 0.00001457, -0.00003725, 0.00000162, -0.00001642, 0.00000611, -0.00005479, -0.00002665, -0.00020798, -0.00030322, 0.00099149, 0.00037868, -0.00046659, -0.00001966, -0.00007607, 0.00000117, -0.00004220, -0.00001165, -0.00000533, -0.00000088, -0.00001576, 0.00000016, -0.00001117, 0.00000403, -0.00001491, 0.00000088, -0.00000249, 0.00000530, -0.00000555, -0.00001445, 0.00001758, -0.00002336, 0.00001725, -0.00000950, 0.00000650, -0.00000414, 0.00000141, -0.00000250, 0.00000468, 0.00000102, -0.00000500, -0.00000781, 0.00000940, 0.00000245, -0.00000980, -0.00000463, 0.00000953, 0.00000800, -0.00001327, 0.00000751, -0.00000318, 0.00000542, -0.00001061, 0.00000352, -0.00000184, 0.00000363, -0.00000248, 0.00000078, -0.00000324, -0.00000168, 0.00000412, -0.00000024, -0.00000020, 0.00000176, -0.00000693, -0.00000282, 0.00000772, -0.00000691, 0.00001040, -0.00000237, -0.00000513, -0.00000611, 0.00001000, },
		{ 0.00000235, -0.00000125, -0.00000352, -0.00000194, -0.00000088, 0.00000268, 0.00000524, -0.00000847, -0.00000446, 0.00000268, 0.00000573, -0.00000455, 0.00001123, -0.00001081, -0.00000045, 0.00000099, 0.00000269, -0.00000325, 0.00000521, -0.00000776, 0.00000120, -0.00000205, 0.00000265, 0.00000091, 0.00000477, -0.00000430, -0.00000439, 0.00000408, 0.00000083, -0.00000046, -0.00001367, 0.00001694, 0.00000617, -0.00001340, -0.00000918, 0.00000938, -0.00000063, -0.00000521, -0.00000457, 0.00000565, 0.00000476, -0.00000981, -0.00000288, -0.00000329, -0.00000248, -0.00000333, 0.00000459, -0.00002580, 0.00003015, -0.00003746, 0.00002676, -0.00001157, 0.00001469, -0.00001648, 0.00000502, -0.00000829, 0.00002076, -0.00004329, 0.00000892, -0.00002378, 0.00001605, -0.00006409, -0.00001115, -0.00021258, -0.00026796, 0.00082914, 0.00049323, -0.00041863, -0.00007819, -0.00004505, -0.00002531, -0.00002271, -0.00002825, 0.00000849, -0.00001360, -0.00000506, -0.00000951, -0.00000209, -0.00000500, -0.00000615, -0.00000816, 0.00000709, -0.00000595, 0.00000655, -0.00002529, 0.00002479, -0.00002667, 0.00001672, -0.00000764, 0.00000631, -0.00000466, 0.00000193, -0.00000307, 0.00000467, 0.00000110, -0.00000360, -0.00000890, 0.00000824, 0.00000395, -0.00000873, -0.00000597, 0.00000725, 0.00001090, -0.00001316, 0.00000585, -0.00000142, 0.00000378, -0.00000797, 0.00000007, 0.00000081, 0.00000111, 0.00000016, -0.00000144, -0.00000045, -0.00000476, 0.00000580, -0.00000107, 0.00000070, 0.00000050, -0.00000404, -0.00000640, 0.00000975, -0.00000781, 0.00000956, -0.00000043, -0.00000494, -0.00000707, 0.00000859, },
		{ 0.00000482, -0.00000235, -0.00000218, -0.00000346, 0.00000038, 0.00000021, 0.00000751, -0.00000823, -0.00000441, -0.00000008, 0.00000837, -0.00000626, 0.00001190, -0.00000916, -0.00000211, 0.00000106, 0.00000323, -0.00000421, 0.00000587, -0.00000695, 0.00000014, -0.00000173, 0.00000219, 0.00000134, 0.00000440, -0.00000324, -0.00000496, 0.00000340, 0.00000163, -0.00000014, -0.00001280, 0.00001169, 0.00001208, -0.00001532, -0.00000777, 0.00000477, 0.00000507, -0.00000955, -0.00000054, -0.00000014, 0.00001104, -0.00001556, 0.00000319, -0.00001001, 0.00000532, -0.00001352, 0.00001607, -0.00003538, 0.00003500, -0.00004021, 0.00002729, -0.00001124, 0.00001644, -0.00001831, 0.00000842, -0.00001264, 0.00002652, -0.00004891, 0.00001576, -0.00003075, 0.00002572, -0.00007319, 0.00000386, -0.00021770, -0.00023441, 0.00069224, 0.00055737, -0.00033222, -0.00015072, -0.00000883, -0.00005482, -0.00000139, -0.00004632, 0.00002349, -0.00002713, 0.00000622, -0.00001982, 0.00000766, -0.00001460, 0.00000313, -0.00001766, 0.00001695, -0.00001682, 0.00001712, -0.00003343, 0.00002871, -0.00002687, 0.00001396, -0.00000483, 0.00000589, -0.00000504, 0.00000232, -0.00000357, 0.00000471, 0.00000087, -0.00000196, -0.00000992, 0.00000717, 0.00000476, -0.00000694, -0.00000745, 0.00000534, 0.00001245, -0.00001131, 0.00000275, 0.00000149, 0.00000106, -0.00000438, -0.00000397, 0.00000374, -0.00000161, 0.00000293, -0.00000373, 0.00000232, -0.00000748, 0.00000696, -0.00000159, 0.00000153, -0.00000083, -0.00000129, -0.00000924, 0.00001064, -0.00000751, 0.00000774, 0.00000189, -0.00000494, -0.00000737, 0.00000645, },
		{ 0.00000723, -0.00000322, -0.00000101, -0.00000476, 0.00000141, -0.00000183, 0.00000893, -0.00000727, -0.00000453, -0.00000250, 0.00001008, -0.00000674, 0.00001124, -0.00000656, -0.00000406, 0.00000122, 0.00000355, -0.00000474, 0.00000586, -0.00000554, -0.00000117, -0.00000130, 0.00000166, 0.00000176, 0.00000413, -0.00000253, -0.00000489, 0.00000203, 0.00000299, -0.00000063, -0.00001061, 0.00000554, 0.00001775, -0.00001682, -0.00000605, -0.00000027, 0.00001058, -0.00001342, 0.00000321, -0.00000561, 0.00001663, -0.00002043, 0.00000831, -0.00001568, 0.00001186, -0.00002173, 0.00002428, -0.00004051, 0.00003522, -0.00003905, 0.00002456, -0.00000851, 0.00001636, -0.00001844, 0.00001025, -0.00001539, 0.00003057, -0.00005258, 0.00002040, -0.00003533, 0.00003274, -0.00007918, 0.00001479, -0.00021855, -0.00020897, 0.00058542, 0.00057425, -0.00021757, -0.00022675, 0.00002486, -0.00008165, 0.00001727, -0.00006215, 0.00003646, -0.00003864, 0.00001562, -0.00002853, 0.00001598, -0.00002279, 0.00001102, -0.00002575, 0.00002517, -0.00002536, 0.00002437, -0.00003752, 0.00002853, -0.00002369, 0.00000906, -0.00000121, 0.00000525, -0.00000515, 0.00000243, -0.00000383, 0.00000465, 0.00000053, -0.00000030, -0.00001069, 0.00000613, 0.00000498, -0.00000466, -0.00000893, 0.00000378, 0.00001269, -0.00000802, -0.00000128, 0.00000499, -0.00000213, -0.00000049, -0.00000790, 0.00000633, -0.00000399, 0.00000528, -0.00000559, 0.00000457, -0.00000940, 0.00000729, -0.00000157, 0.00000206, -0.00000193, 0.00000095, -0.00001100, 0.00001023, -0.00000605, 0.00000515, 0.00000437, -0.00000499, -0.00000716, 0.00000386, },
		{ 0.00000922, -0.00000360, -0.00000022, -0.00000561, 0.00000198, -0.00000319, 0.00000933, -0.00000555, -0.00000490, -0.00000438, 0.00001068, -0.00000595, 0.00000940, -0.00000327, -0.00000608, 0.00000137, 0.00000365, -0.00000482, 0.00000522, -0.00000367, -0.00000258, -0.00000089, 0.00000113, 0.00000209, 0.00000401, -0.00000216, -0.00000427, 0.00000019, 0.00000464, -0.00000158, -0.00000768, -0.00000071, 0.00002240, -0.00001730, -0.00000460, -0.00000494, 0.00001503, -0.00001601, 0.00000585, -0.00000984, 0.00002057, -0.00002343, 0.00001147, -0.00001921, 0.00001593, -0.00002668, 0.00002806, -0.00004042, 0.00003061, -0.00003408, 0.00001876, -0.00000352, 0.00001440, -0.00001667, 0.00001013, -0.00001601, 0.00003220, -0.00005345, 0.00002182, -0.00003634, 0.00003568, -0.00008031, 0.00001954, -0.00021244, -0.00019492, 0.00050907, 0.00055122, -0.00008727, -0.00029536, 0.00004888, -0.00010088, 0.00002952, -0.00007274, 0.00004481, -0.00004591, 0.00002119, -0.00003392, 0.00002127, -0.00002804, 0.00001607, -0.00003100, 0.00003030, -0.00003019, 0.00002721, -0.00003693, 0.00002417, -0.00001746, 0.00000255, 0.00000278, 0.00000457, -0.00000501, 0.00000222, -0.00000381, 0.00000442, 0.00000019, 0.00000121, -0.00001109, 0.00000506, 0.00000473, -0.00000212, -0.00001023, 0.00000253, 0.00001183, -0.00000378, -0.00000563, 0.00000844, -0.00000515, 0.00000305, -0.00001109, 0.00000802, -0.00000555, 0.00000681, -0.00000664, 0.00000594, -0.00001024, 0.00000662, -0.00000092, 0.00000219, -0.00000261, 0.00000243, -0.00001154, 0.00000857, -0.00000366, 0.00000211, 0.00000671, -0.00000494, -0.00000663, 0.00000112, },
		{ 0.00001050, -0.00000329, 0.00000003, -0.00000586, 0.00000193, -0.00000376, 0.00000870, -0.00000317, -0.00000549, -0.00000566, 0.00001016, -0.00000405, 0.00000669, 0.00000030, -0.00000786, 0.00000138, 0.00000360, -0.00000453, 0.00000409, -0.00000156, -0.00000388, -0.00000061, 0.00000069, 0.00000228, 0.00000407, -0.00000210, -0.00000324, -0.00000185, 0.00000623, -0.00000261, -0.00000458, -0.00000629, 0.00002536, -0.00001635, -0.00000385, -0.00000864, 0.00001778, -0.00001675, 0.00000684, -0.00001222, 0.00002223, -0.00002399, 0.00001211, -0.00002001, 0.00001689, -0.00002774, 0.00002707, -0.00003529, 0.00002181, -0.00002616, 0.00001073, 0.00000301, 0.00001107, -0.00001330, 0.00000824, -0.00001452, 0.00003134, -0.00005136, 0.00001984, -0.00003353, 0.00003420, -0.00007612, 0.00001758, -0.00019871, -0.00019268, 0.00046004, 0.00049848, 0.00004529, -0.00034668, 0.00005770, -0.00010910, 0.00003299, -0.00007624, 0.00004703, -0.00004765, 0.00002189, -0.00003505, 0.00002263, -0.00002951, 0.00001746, -0.00003263, 0.00003161, -0.00003072, 0.00002537, -0.00003187, 0.00001629, -0.00000911, -0.00000467, 0.00000651, 0.00000417, -0.00000475, 0.00000175, -0.00000354, 0.00000404, -0.00000009, 0.00000245, -0.00001106, 0.00000395, 0.00000412, 0.00000045, -0.00001116, 0.00000147, 0.00001017, 0.00000086, -0.00000966, 0.00001123, -0.00000743, 0.00000567, -0.00001305, 0.00000845, -0.00000600, 0.00000727, -0.00000667, 0.00000626, -0.00000991, 0.00000499, 0.00000030, 0.00000190, -0.00000278, 0.00000306, -0.00001089, 0.00000589, -0.00000072, -0.00000097, 0.00000861, -0.00000458, -0.00000600, -0.00000144, },
		{ 0.00001089, -0.00000223, -0.00000028, -0.00000551, 0.00000126, -0.00000357, 0.00000718, -0.00000035, -0.00000616, -0.00000639, 0.00000868, -0.00000137, 0.00000356, 0.00000372, -0.00000912, 0.00000110, 0.00000349, -0.00000399, 0.00000268, 0.00000054, -0.00000487, -0.00000060, 0.00000042, 0.00000228, 0.00000432, -0.00000230, -0.00000197, -0.00000381, 0.00000746, -0.00000335, -0.00000183, -0.00001060, 0.00002624, -0.00001381, -0.00000401, -0.00001100, 0.00001850, -0.00001544, 0.00000599, -0.00001252, 0.00002142, -0.00002199, 0.00001017, -0.00001804, 0.00001474, -0.00002508, 0.00002181, -0.00002611, 0.00001023, -0.00001673, 0.00000172, 0.00001003, 0.00000720, -0.00000899, 0.00000513, -0.00001141, 0.00002843, -0.00004677, 0.00001494, -0.00002743, 0.00002885, -0.00006724, 0.00000976, -0.00017846, -0.00020040, 0.00043267, 0.00042759, 0.00016753, -0.00037312, 0.00004826, -0.00010500, 0.00002697, -0.00007223, 0.00004280, -0.00004368, 0.00001760, -0.00003185, 0.00001997, -0.00002710, 0.00001514, -0.00003059, 0.00002909, -0.00002711, 0.00001937, -0.00002327, 0.00000612, 0.00000009, -0.00001152, 0.00000925, 0.00000437, -0.00000452, 0.00000115, -0.00000312, 0.00000354, -0.00000027, 0.00000335, -0.00001059, 0.00000281, 0.00000329, 0.00000282, -0.00001159, 0.00000048, 0.00000803, 0.00000535, -0.00001277, 0.00001288, -0.00000853, 0.00000698, -0.00001348, 0.00000744, -0.00000526, 0.00000662, -0.00000566, 0.00000555, -0.00000851, 0.00000257, 0.00000191, 0.00000130, -0.00000249, 0.00000292, -0.00000928, 0.00000258, 0.00000232, -0.00000365, 0.00000978, -0.00000376, -0.00000544, -0.00000355, },
		{ 0.00001033, -0.00000049, -0.00000105, -0.00000465, 0.00000010, -0.00000277, 0.00000504, 0.00000258, -0.00000670, -0.00000672, 0.00000652, 0.00000167, 0.00000049, 0.00000656, -0.00000958, 0.00000042, 0.00000341, -0.00000334, 0.00000121, 0.00000239, -0.00000538, -0.00000093, 0.00000036, 0.00000209, 0.00000471, -0.00000263, -0.00000068, -0.00000542, 0.00000807, -0.00000353, 0.00000019, -0.00001327, 0.00002495, -0.00000980, -0.00000504, -0.00001195, 0.00001718, -0.00001221, 0.00000347, -0.00001091, 0.00001841, -0.00001782, 0.00000607, -0.00001379, 0.00001004, -0.00001946, 0.00001344, -0.00001448, -0.00000229, -0.00000751, -0.00000681, 0.00001632, 0.00000377, -0.00000459, 0.00000159, -0.00000747, 0.00002429, -0.00004054, 0.00000811, -0.00001914, 0.00002087, -0.00005511, -0.00000212, -0.00015395, -0.00021472, 0.00042002, 0.00034997, 0.00026909, -0.00037030, 0.00002039, -0.00008949, 0.00001252, -0.00006170, 0.00003301, -0.00003485, 0.00000917, -0.00002507, 0.00001397, -0.00002146, 0.00000974, -0.00002553, 0.00002343, -0.00002023, 0.00001037, -0.00001261, -0.00000474, 0.00000868, -0.00001690, 0.00001040, 0.00000544, -0.00000446, 0.00000053, -0.00000266, 0.00000301, -0.00000037, 0.00000392, -0.00000973, 0.00000166, 0.00000233, 0.00000480, -0.00001141, -0.00000056, 0.00000576, 0.00000916, -0.00001454, 0.00001307, -0.00000826, 0.00000686, -0.00001236, 0.00000510, -0.00000348, 0.00000502, -0.00000380, 0.00000403, -0.00000632, -0.00000030, 0.00000362, 0.00000058, -0.00000188, 0.00000219, -0.00000703, -0.00000090, 0.00000500, -0.00000557, 0.00001001, -0.00000242, -0.00000510, -0.00000505, },
		{ 0.00000888, 0.00000176, -0.00000211, -0.00000347, -0.00000135, -0.00000162, 0.00000263, 0.00000528, -0.00000688, -0.00000685, 0.00000401, 0.00000462, -0.00000206, 0.00000845, -0.00000911, -0.00000070, 0.00000343, -0.00000272, -0.00000015, 0.00000380, -0.00000531, -0.00000162, 0.00000048, 0.00000175, 0.00000517, -0.00000300, 0.00000044, -0.00000646, 0.00000789, -0.00000301, 0.00000129, -0.00001419, 0.00002167, -0.00000472, -0.00000667, -0.00001164, 0.00001412, -0.00000753, -0.00000024, -0.00000789, 0.00001378, -0.00001219, 0.00000063, -0.00000812, 0.00000381, -0.00001212, 0.00000354, -0.00000234, -0.00001378, -0.00000017, -0.00001352, 0.00002080, 0.00000166, -0.00000090, -0.00000155, -0.00000354, 0.00001985, -0.00003373, 0.00000056, -0.00001004, 0.00001182, -0.00004157, -0.00001577, -0.00012804, -0.00023170, 0.00041503, 0.00027561, 0.00034278, -0.00033752, -0.00002318, -0.00006551, -0.00000783, -0.00004681, 0.00001954, -0.00002286, -0.00000185, -0.00001611, 0.00000586, -0.00001375, 0.00000241, -0.00001856, 0.00001579, -0.00001141, -0.00000002, -0.00000164, -0.00001463, 0.00001531, -0.00001992, 0.00000954, 0.00000748, -0.00000462, -0.00000000, -0.00000224, 0.00000249, -0.00000041, 0.00000416, -0.00000854, 0.00000056, 0.00000134, 0.00000627, -0.00001055, -0.00000173, 0.00000364, 0.00001195, -0.00001473, 0.00001175, -0.00000664, 0.00000539, -0.00000990, 0.00000175, -0.00000102, 0.00000281, -0.00000144, 0.00000204, -0.00000373, -0.00000321, 0.00000511, -0.00000003, -0.00000113, 0.00000112, -0.00000454, -0.00000409, 0.00000691, -0.00000648, 0.00000925, -0.00000058, -0.00000500, -0.00000587, },
		{ 0.00000674, 0.00000422, -0.00000317, -0.00000221, -0.00000283, -0.00000042, 0.00000032, 0.00000740, -0.00000649, -0.00000697, 0.00000148, 0.00000708, -0.00000374, 0.00000921, -0.00000767, -0.00000220, 0.00000353, -0.00000219, -0.00000124, 0.00000466, -0.00000463, -0.00000261, 0.00000073, 0.00000132, 0.00000562, -0.00000326, 0.00000123, -0.00000681, 0.00000688, -0.00000180, 0.00000148, -0.00001351, 0.00001686, 0.00000087, -0.00000846, -0.00001045, 0.00000984, -0.00000207, -0.00000443, -0.00000418, 0.00000839, -0.00000605, -0.00000515, -0.00000211, -0.00000278, -0.00000446, -0.00000617, 0.00000843, -0.00002249, 0.00000396, -0.00001745, 0.00002271, 0.00000147, 0.00000149, -0.00000367, -0.00000034, 0.00001593, -0.00002734, -0.00000654, -0.00000146, 0.00000322, -0.00002843, -0.00002893, -0.00010352, -0.00024757, 0.00041150, 0.00021208, 0.00038515, -0.00027766, -0.00007724, -0.00003751, -0.00003054, -0.00003045, 0.00000481, -0.00000986, -0.00001352, -0.00000667, -0.00000281, -0.00000544, -0.00000545, -0.00001106, 0.00000760, -0.00000221, -0.00001012, 0.00000795, -0.00002208, 0.00001896, -0.00002006, 0.00000659, 0.00001038, -0.00000494, -0.00000044, -0.00000192, 0.00000203, -0.00000038, 0.00000411, -0.00000712, -0.00000046, 0.00000038, 0.00000712, -0.00000905, -0.00000304, 0.00000186, 0.00001349, -0.00001331, 0.00000908, -0.00000393, 0.00000290, -0.00000652, -0.00000213, 0.00000166, 0.00000044, 0.00000101, 0.00000000, -0.00000119, -0.00000575, 0.00000606, -0.00000033, -0.00000045, -0.00000001, -0.00000218, -0.00000662, 0.00000778, -0.00000627, 0.00000757, 0.00000163, -0.00000507, -0.00000607, },
		{ 0.00000419, 0.00000657, -0.00000397, -0.00000113, -0.00000409, 0.00000054, -0.00000159, 0.00000869, -0.00000543, -0.00000720, -0.00000081, 0.00000874, -0.00000435, 0.00000879, -0.00000540, -0.00000393, 0.00000368, -0.00000180, -0.00000201, 0.00000492, -0.00000343, -0.00000378, 0.00000099, 0.00000090, 0.00000595, -0.00000331, 0.00000160, -0.00000643, 0.00000512, -0.00000004, 0.00000091, -0.00001158, 0.00001116, 0.00000627, -0.00000992, -0.00000883, 0.00000499, 0.00000339, -0.00000834, -0.00000057, 0.00000311, -0.00000038, -0.00001025, 0.00000317, -0.00000855, 0.00000216, -0.00001418, 0.00001629, -0.00002718, 0.00000416, -0.00001820, 0.00002179, 0.00000339, 0.00000235, -0.00000441, 0.00000166, 0.00001316, -0.00002211, -0.00001229, 0.00000555, -0.00000369, -0.00001721, -0.00003973, -0.00008266, -0.00025935, 0.00040480, 0.00016409, 0.00039650, -0.00019669, -0.00013487, -0.00001071, -0.00005174, -0.00001566, -0.00000864, 0.00000194, -0.00002389, 0.00000153, -0.00001050, 0.00000202, -0.00001244, -0.00000440, 0.00000022, 0.00000591, -0.00001841, 0.00001477, -0.00002608, 0.00001909, -0.00001727, 0.00000184, 0.00001379, -0.00000523, -0.00000084, -0.00000167, 0.00000161, -0.00000029, 0.00000381, -0.00000555, -0.00000134, -0.00000048, 0.00000734, -0.00000697, -0.00000447, 0.00000053, 0.00001377, -0.00001051, 0.00000545, -0.00000057, -0.00000011, -0.00000276, -0.00000596, 0.00000403, -0.00000163, 0.00000312, -0.00000167, 0.00000092, -0.00000757, 0.00000626, -0.00000017, 0.00000002, -0.00000096, -0.00000026, -0.00000820, 0.00000750, -0.00000501, 0.00000517, 0.00000398, -0.00000520, -0.00000580, },
		{ 0.00000156, 0.00000848, -0.00000426, -0.00000043, -0.00000490, 0.00000105, -0.00000285, 0.00000903, -0.00000368, -0.00000759, -0.00000269, 0.00000946, -0.00000386, 0.00000734, -0.00000256, -0.00000566, 0.00000375, -0.00000152, -0.00000246, 0.00000467, -0.00000186, -0.00000494, 0.00000114, 0.00000060, 0.00000610, -0.00000309, 0.00000152, -0.00000539, 0.00000282, 0.00000201, -0.00000009, -0.00000889, 0.00000528, 0.00001081, -0.00001059, -0.00000724, 0.00000022, 0.00000810, -0.00001128, 0.00000224, -0.00000125, 0.00000397, -0.00001384, 0.00000687, -0.00001256, 0.00000671, -0.00001939, 0.00002029, -0.00002733, 0.00000039, -0.00001600, 0.00001832, 0.00000714, 0.00000182, -0.00000378, 0.00000235, 0.00001177, -0.00001844, -0.00001616, 0.00001034, -0.00000815, -0.00000882, -0.00004699, -0.00006685, -0.00026527, 0.00039218, 0.00013336, 0.00038038, -0.00010273, -0.00018846, 0.00000979, -0.00006789, -0.00000514, -0.00001865, 0.00001066, -0.00003134, 0.00000711, -0.00001598, 0.00000744, -0.00001741, 0.00000032, -0.00000523, 0.00001181, -0.00002378, 0.00001796, -0.00002619, 0.00001578, -0.00001200, -0.00000411, 0.00001720, -0.00000520, -0.00000129, -0.00000146, 0.00000120, -0.00000014, 0.00000330, -0.00000393, -0.00000202, -0.00000123, 0.00000697, -0.00000448, -0.00000590, -0.00000037, 0.00001291, -0.00000670, 0.00000135, 0.00000290, -0.00000311, 0.00000078, -0.00000918, 0.00000562, -0.00000303, 0.00000455, -0.00000268, 0.00000229, -0.00000845, 0.00000559, 0.00000049, 0.00000020, -0.00000159, 0.00000103, -0.00000875, 0.00000615, -0.00000294, 0.00000239, 0.00000618, -0.00000520, -0.00000526, },
		{ -0.00000086, 0.00000967, -0.00000387, -0.00000025, -0.00000516, 0.00000099, -0.00000339, 0.00000842, -0.00000138, -0.00000809, -0.00000411, 0.00000924, -0.00000243, 0.00000516, 0.00000048, -0.00000712, 0.00000362, -0.00000129, -0.00000264, 0.00000401, -0.00000011, -0.00000592, 0.00000105, 0.00000048, 0.00000599, -0.00000259, 0.00000102, -0.00000385, 0.00000026, 0.00000402, -0.00000119, -0.00000596, -0.00000009, 0.00001393, -0.00001012, -0.00000604, -0.00000392, 0.00001149, -0.00001273, 0.00000374, -0.00000412, 0.00000643, -0.00001540, 0.00000847, -0.00001430, 0.00000863, -0.00002127, 0.00002017, -0.00002316, -0.00000666, -0.00001165, 0.00001303, 0.00001208, 0.00000041, -0.00000212, 0.00000193, 0.00001169, -0.00001634, -0.00001806, 0.00001280, -0.00000998, -0.00000357, -0.00005031, -0.00005651, -0.00026480, 0.00037282, 0.00011899, 0.00034281, -0.00000498, -0.00023083, 0.00001987, -0.00007643, -0.00000069, -0.00002380, 0.00001511, -0.00003487, 0.00000923, -0.00001849, 0.00001011, -0.00001969, 0.00000245, -0.00000809, 0.00001481, -0.00002568, 0.00001731, -0.00002266, 0.00000965, -0.00000510, -0.00001039, 0.00001998, -0.00000457, -0.00000193, -0.00000122, 0.00000077, 0.00000010, 0.00000264, -0.00000238, -0.00000248, -0.00000185, 0.00000608, -0.00000176, -0.00000720, -0.00000094, 0.00001117, -0.00000238, -0.00000261, 0.00000592, -0.00000557, 0.00000360, -0.00001132, 0.00000610, -0.00000349, 0.00000509, -0.00000283, 0.00000278, -0.00000832, 0.00000410, 0.00000156, 0.00000010, -0.00000184, 0.00000163, -0.00000831, 0.00000394, -0.00000039, -0.00000039, 0.00000794, -0.00000489, -0.00000466, },
		{ -0.00000283, 0.00000998, -0.00000275, -0.00000057, -0.00000487, 0.00000036, -0.00000323, 0.00000703, 0.00000127, -0.00000856, -0.00000511, 0.00000821, -0.00000034, 0.00000261, 0.00000335, -0.00000806, 0.00000319, -0.00000104, -0.00000264, 0.00000312, 0.00000159, -0.00000653, 0.00000063, 0.00000059, 0.00000563, -0.00000183, 0.00000021, -0.00000204, -0.00000225, 0.00000567, -0.00000202, -0.00000326, -0.00000440, 0.00001527, -0.00000839, -0.00000544, -0.00000706, 0.00001319, -0.00001245, 0.00000372, -0.00000525, 0.00000679, -0.00001478, 0.00000789, -0.00001368, 0.00000788, -0.00001994, 0.00001635, -0.00001560, -0.00001574, -0.00000635, 0.00000696, 0.00001728, -0.00000116, 0.00000002, 0.00000086, 0.00001253, -0.00001549, -0.00001829, 0.00001323, -0.00000948, -0.00000113, -0.00005002, -0.00005119, -0.00025858, 0.00034748, 0.00011805, 0.00029122, 0.00008754, -0.00025613, 0.00001700, -0.00007619, -0.00000297, -0.00002365, 0.00001495, -0.00003424, 0.00000776, -0.00001792, 0.00000988, -0.00001911, 0.00000186, -0.00000824, 0.00001483, -0.00002419, 0.00001321, -0.00001627, 0.00000175, 0.00000232, -0.00001604, 0.00002153, -0.00000308, -0.00000284, -0.00000088, 0.00000028, 0.00000042, 0.00000189, -0.00000096, -0.00000267, -0.00000235, 0.00000479, 0.00000097, -0.00000820, -0.00000130, 0.00000889, 0.00000192, -0.00000590, 0.00000803, -0.00000707, 0.00000532, -0.00001212, 0.00000533, -0.00000295, 0.00000470, -0.00000213, 0.00000242, -0.00000728, 0.00000199, 0.00000288, -0.00000018, -0.00000174, 0.00000160, -0.00000707, 0.00000121, 0.00000221, -0.00000278, 0.00000899, -0.00000413, -0.00000417, },
		{ -0.00000418, 0.00000937, -0.00000099, -0.00000132, -0.00000411, -0.00000070, -0.00000254, 0.00000510, 0.00000395, -0.00000882, -0.00000581, 0.00000660, 0.00000206, 0.00000014, 0.00000568, -0.00000827, 0.00000237, -0.00000074, -0.00000256, 0.00000214, 0.00000306, -0.00000666, -0.00000015, 0.00000093, 0.00000506, -0.00000090, -0.00000076, -0.00000019, -0.00000439, 0.00000666, -0.00000233, -0.00000117, -0.00000730, 0.00001472, -0.00000548, -0.00000543, -0.00000906, 0.00001312, -0.00001050, 0.00000226, -0.00000472, 0.00000521, -0.00001226, 0.00000545, -0.00001107, 0.00000492, -0.00001603, 0.00000986, -0.00000606, -0.00002525, -0.00000149, 0.00000124, 0.00002178, -0.00000213, 0.00000200, -0.00000029, 0.00001378, -0.00001538, -0.00001740, 0.00001224, -0.00000735, -0.00000078, -0.00004697, -0.00004974, -0.00024809, 0.00031807, 0.00012644, 0.00023337, 0.00016705, -0.00026072, 0.00000073, -0.00006761, -0.00001138, -0.00001879, 0.00001070, -0.00002997, 0.00000321, -0.00001473, 0.00000714, -0.00001606, -0.00000107, -0.00000605, 0.00001229, -0.00001994, 0.00000662, -0.00000824, -0.00000657, 0.00000905, -0.00002018, 0.00002138, -0.00000060, -0.00000404, -0.00000044, -0.00000029, 0.00000080, 0.00000113, 0.00000023, -0.00000258, -0.00000274, 0.00000326, 0.00000348, -0.00000874, -0.00000163, 0.00000643, 0.00000571, -0.00000809, 0.00000891, -0.00000739, 0.00000577, -0.00001151, 0.00000337, -0.00000155, 0.00000354, -0.00000073, 0.00000139, -0.00000559, -0.00000047, 0.00000419, -0.00000048, -0.00000140, 0.00000109, -0.00000532, -0.00000163, 0.00000448, -0.00000447, 0.00000917, -0.00000286, -0.00000390, },
		{ -0.00000488, 0.00000792, 0.00000121, -0.00000228, -0.00000309, -0.00000201, -0.00000157, 0.00000296, 0.00000634, -0.00000867, -0.00000637, 0.00000467, 0.00000439, -0.00000190, 0.00000718, -0.00000765, 0.00000118, -0.00000034, -0.00000247, 0.00000123, 0.00000414, -0.00000623, -0.00000126, 0.00000142, 0.00000437, 0.00000008, -0.00000172, 0.00000146, -0.00000593, 0.00000683, -0.00000196, 0.00000014, -0.00000867, 0.00001244, -0.00000171, -0.00000585, -0.00000998, 0.00001147, -0.00000720, -0.00000029, -0.00000291, 0.00000215, -0.00000841, 0.00000179, -0.00000719, 0.00000055, -0.00001058, 0.00000205, 0.00000380, -0.00003355, 0.00000165, -0.00000312, 0.00002473, -0.00000182, 0.00000326, -0.00000099, 0.00001489, -0.00001542, -0.00001605, 0.00001059, -0.00000444, -0.00000153, -0.00004232, -0.00005065, -0.00023524, 0.00028707, 0.00013968, 0.00017633, 0.00022784, -0.00024350, -0.00002728, -0.00005263, -0.00002424, -0.00001064, 0.00000359, -0.00002321, -0.00000334, -0.00000985, 0.00000272, -0.00001132, -0.00000555, -0.00000231, 0.00000802, -0.00001392, -0.00000122, 0.00000000, -0.00001399, 0.00001401, -0.00002211, 0.00001929, 0.00000282, -0.00000545, 0.00000007, -0.00000091, 0.00000120, 0.00000042, 0.00000112, -0.00000220, -0.00000305, 0.00000161, 0.00000558, -0.00000872, -0.00000206, 0.00000410, 0.00000860, -0.00000890, 0.00000845, -0.00000649, 0.00000500, -0.00000965, 0.00000050, 0.00000042, 0.00000188, 0.00000108, -0.00000001, -0.00000357, -0.00000291, 0.00000522, -0.00000063, -0.00000096, 0.00000032, -0.00000340, -0.00000421, 0.00000606, -0.00000524, 0.00000844, -0.00000112, -0.00000387, },
		{ -0.00000501, 0.00000584, 0.00000358, -0.00000321, -0.00000202, -0.00000332, -0.00000058, 0.00000094, 0.00000816, -0.00000797, -0.00000691, 0.00000269, 0.00000633, -0.00000321, 0.00000770, -0.00000622, -0.00000033, 0.00000010, -0.00000241, 0.00000047, 0.00000476, -0.00000527, -0.00000261, 0.00000196, 0.00000365, 0.00000098, -0.00000251, 0.00000271, -0.00000671, 0.00000611, -0.00000093, 0.00000064, -0.00000862, 0.00000883, 0.00000244, -0.00000637, -0.00001006, 0.00000861, -0.00000308, -0.00000338, -0.00000037, -0.00000171, -0.00000402, -0.00000224, -0.00000291, -0.00000421, -0.00000480, -0.00000563, 0.00001240, -0.00003925, 0.00000209, -0.00000546, 0.00002557, 0.00000016, 0.00000344, -0.00000085, 0.00001544, -0.00001511, -0.00001484, 0.00000902, -0.00000160, -0.00000240, -0.00003727, -0.00005240, -0.00022200, 0.00025695, 0.00015362, 0.00012575, 0.00026676, -0.00020607, -0.00006349, -0.00003435, -0.00003906, -0.00000123, -0.00000466, -0.00001547, -0.00001055, -0.00000449, -0.00000231, -0.00000593, -0.00001059, 0.00000201, 0.00000305, -0.00000733, -0.00000893, 0.00000709, -0.00001933, 0.00001641, -0.00002151, 0.00001533, 0.00000694, -0.00000686, 0.00000057, -0.00000152, 0.00000156, -0.00000016, 0.00000169, -0.00000156, -0.00000329, -0.00000000, 0.00000712, -0.00000807, -0.00000268, 0.00000212, 0.00001037, -0.00000829, 0.00000676, -0.00000458, 0.00000326, -0.00000690, -0.00000285, 0.00000255, 0.00000009, 0.00000295, -0.00000145, -0.00000159, -0.00000498, 0.00000572, -0.00000049, -0.00000056, -0.00000051, -0.00000159, -0.00000621, 0.00000674, -0.00000502, 0.00000688, 0.00000095, -0.00000401, },
		{ -0.00000470, 0.00000340, 0.00000580, -0.00000384, -0.00000115, -0.00000439, 0.00000016, -0.00000068, 0.00000918, -0.00000663, -0.00000751, 0.00000085, 0.00000763, -0.00000363, 0.00000724, -0.00000412, -0.00000197, 0.00000052, -0.00000239, -0.00000009, 0.00000490, -0.00000386, -0.00000406, 0.00000242, 0.00000303, 0.00000170, -0.00000301, 0.00000346, -0.00000668, 0.00000460, 0.00000061, 0.00000051, -0.00000746, 0.00000442, 0.00000641, -0.00000663, -0.00000961, 0.00000505, 0.00000120, -0.00000638, 0.00000224, -0.00000560, 0.00000007, -0.00000581, 0.00000087, -0.00000839, 0.00000019, -0.00001189, 0.00001845, -0.00004141, -0.00000059, -0.00000556, 0.00002416, 0.00000388, 0.00000249, 0.00000026, 0.00001521, -0.00001413, -0.00001422, 0.00000809, 0.00000050, -0.00000260, -0.00003280, -0.00005367, -0.00021001, 0.00022974, 0.00016505, 0.00008536, 0.00028338, -0.00015233, -0.00010302, -0.00001651, -0.00005308, 0.00000727, -0.00001227, -0.00000834, -0.00001699, 0.00000014, -0.00000686, -0.00000094, -0.00001517, 0.00000591, -0.00000156, -0.00000134, -0.00001528, 0.00001191, -0.00002182, 0.00001593, -0.00001844, 0.00000987, 0.00001133, -0.00000803, 0.00000091, -0.00000203, 0.00000179, -0.00000055, 0.00000190, -0.00000069, -0.00000346, -0.00000150, 0.00000801, -0.00000681, -0.00000350, 0.00000064, 0.00001094, -0.00000639, 0.00000412, -0.00000200, 0.00000095, -0.00000373, -0.00000618, 0.00000440, -0.00000144, 0.00000452, -0.00000257, 0.00000001, -0.00000640, 0.00000552, 0.00000002, -0.00000030, -0.00000121, -0.00000016, -0.00000742, 0.00000645, -0.00000389, 0.00000469, 0.00000314, -0.00000420, },
		{ -0.00000416, 0.00000092, 0.00000755, -0.00000395, -0.00000064, -0.00000506, 0.00000049, -0.00000172, 0.00000933, -0.00000469, -0.00000818, -0.00000071, 0.00000818, -0.00000317, 0.00000594, -0.00000159, -0.00000356, 0.00000082, -0.00000237, -0.00000045, 0.00000462, -0.00000216, -0.00000541, 0.00000267, 0.00000261, 0.00000213, -0.00000315, 0.00000367, -0.00000591, 0.00000250, 0.00000241, 0.00000002, -0.00000560, -0.00000017, 0.00000964, -0.00000628, -0.00000900, 0.00000133, 0.00000504, -0.00000869, 0.00000432, -0.00000885, 0.00000314, -0.00000819, 0.00000341, -0.00001120, 0.00000351, -0.00001582, 0.00002119, -0.00003970, -0.00000625, -0.00000369, 0.00002074, 0.00000905, 0.00000062, 0.00000223, 0.00001422, -0.00001239, -0.00001437, 0.00000809, 0.00000148, -0.00000164, -0.00002955, -0.00005360, -0.00020034, 0.00020673, 0.00017195, 0.00005685, 0.00027966, -0.00008792, -0.00014036, -0.00000282, -0.00006371, 0.00001292, -0.00001766, -0.00000319, -0.00002148, 0.00000303, -0.00001003, 0.00000280, -0.00001842, 0.00000855, -0.00000497, 0.00000312, -0.00001935, 0.00001378, -0.00002118, 0.00001267, -0.00001335, 0.00000353, 0.00001544, -0.00000863, 0.00000096, -0.00000237, 0.00000185, -0.00000070, 0.00000178, 0.00000034, -0.00000354, -0.00000279, 0.00000825, -0.00000503, -0.00000447, -0.00000035, 0.00001043, -0.00000352, 0.00000098, 0.00000078, -0.00000145, -0.00000065, -0.00000898, 0.00000555, -0.00000239, 0.00000553, -0.00000313, 0.00000099, -0.00000700, 0.00000458, 0.00000092, -0.00000022, -0.00000166, 0.00000076, -0.00000778, 0.00000527, -0.00000207, 0.00000219, 0.00000517, -0.00000425, },
		{ -0.00000359, -0.00000133, 0.00000860, -0.00000341, -0.00000060, -0.00000522, 0.00000030, -0.00000210, 0.00000864, -0.00000230, -0.00000883, -0.00000197, 0.00000799, -0.00000196, 0.00000405, 0.00000106, -0.00000485, 0.00000089, -0.00000231, -0.00000065, 0.00000401, -0.00000034, -0.00000651, 0.00000259, 0.00000245, 0.00000225, -0.00000293, 0.00000339, -0.00000459, 0.00000011, 0.00000414, -0.00000052, -0.00000350, -0.00000435, 0.00001168, -0.00000507, -0.00000849, -0.00000208, 0.00000790, -0.00000986, 0.00000542, -0.00001095, 0.00000470, -0.00000893, 0.00000430, -0.00001218, 0.00000468, -0.00001700, 0.00002047, -0.00003446, -0.00001416, -0.00000057, 0.00001591, 0.00001508, -0.00000168, 0.00000470, 0.00001272, -0.00001002, -0.00001524, 0.00000906, 0.00000126, 0.00000061, -0.00002775, -0.00005182, -0.00019347, 0.00018840, 0.00017358, 0.00004001, 0.00025952, -0.00001941, -0.00017018, 0.00000363, -0.00006906, 0.00001438, -0.00001978, -0.00000091, -0.00002329, 0.00000359, -0.00001127, 0.00000472, -0.00001983, 0.00000943, -0.00000664, 0.00000548, -0.00002070, 0.00001255, -0.00001765, 0.00000723, -0.00000703, -0.00000288, 0.00001869, -0.00000839, 0.00000060, -0.00000247, 0.00000169, -0.00000060, 0.00000137, 0.00000144, -0.00000350, -0.00000387, 0.00000787, -0.00000289, -0.00000545, -0.00000091, 0.00000907, -0.00000013, -0.00000217, 0.00000328, -0.00000348, 0.00000186, -0.00001086, 0.00000572, -0.00000256, 0.00000580, -0.00000300, 0.00000125, -0.00000674, 0.00000296, 0.00000209, -0.00000030, -0.00000184, 0.00000114, -0.00000735, 0.00000340, 0.00000013, -0.00000029, 0.00000678, -0.00000401, },
		{ -0.00000314, -0.00000313, 0.00000879, -0.00000220, -0.00000102, -0.00000489, -0.00000037, -0.00000189, 0.00000728, 0.00000031, -0.00000934, -0.00000297, 0.00000717, -0.00000023, 0.00000192, 0.00000350, -0.00000564, 0.00000064, -0.00000217, -0.00000074, 0.00000319, 0.00000141, -0.00000720, 0.00000214, 0.00000257, 0.00000208, -0.00000240, 0.00000276, -0.00000294, -0.00000223, 0.00000546, -0.00000078, -0.00000158, -0.00000765, 0.00001223, -0.00000294, -0.00000824, -0.00000485, 0.00000946, -0.00000967, 0.00000532, -0.00001165, 0.00000453, -0.00000791, 0.00000344, -0.00001126, 0.00000368, -0.00001554, 0.00001673, -0.00002656, -0.00002321, 0.00000283, 0.00001047, 0.00002121, -0.00000379, 0.00000720, 0.00001108, -0.00000735, -0.00001660, 0.00001078, 0.00000005, 0.00000394, -0.00002723, -0.00004842, -0.00018924, 0.00017444, 0.00017030, 0.00003315, 0.00022801, 0.00004661, -0.00018808, 0.00000093, -0.00006824, 0.00001117, -0.00001831, -0.00000177, -0.00002223, 0.00000170, -0.00001048, 0.00000469, -0.00001924, 0.00000841, -0.00000640, 0.00000560, -0.00001932, 0.00000857, -0.00001192, 0.00000051, -0.00000045, -0.00000853, 0.00002056, -0.00000712, -0.00000020, -0.00000231, 0.00000131, -0.00000029, 0.00000076, 0.00000251, -0.00000331, -0.00000473, 0.00000700, -0.00000060, -0.00000631, -0.00000121, 0.00000716, 0.00000333, -0.00000484, 0.00000508, -0.00000476, 0.00000347, -0.00001155, 0.00000480, -0.00000191, 0.00000532, -0.00000217, 0.00000083, -0.00000574, 0.00000085, 0.00000336, -0.00000043, -0.00000177, 0.00000102, -0.00000630, 0.00000114, 0.00000235, -0.00000241, 0.00000775, -0.00000335, },
		{ -0.00000290, -0.00000434, 0.00000812, -0.00000041, -0.00000177, -0.00000418, -0.00000141, -0.00000125, 0.00000548, 0.00000287, -0.00000954, -0.00000380, 0.00000590, 0.00000171, -0.00000014, 0.00000543, -0.00000579, 0.00000004, -0.00000193, -0.00000077, 0.00000230, 0.00000294, -0.00000739, 0.00000129, 0.00000291, 0.00000168, -0.00000169, 0.00000194, -0.00000125, -0.00000421, 0.00000612, -0.00000055, -0.00000013, -0.00000979, 0.00001125, 0.00000001, -0.00000827, -0.00000682, 0.00000959, -0.00000811, 0.00000407, -0.00001099, 0.00000274, -0.00000535, 0.00000110, -0.00000875, 0.00000090, -0.00001202, 0.00001086, -0.00001728, -0.00003203, 0.00000547, 0.00000524, 0.00002667, -0.00000506, 0.00000922, 0.00000975, -0.00000474, -0.00001808, 0.00001288, -0.00000175, 0.00000794, -0.00002755, -0.00004389, -0.00018704, 0.00016393, 0.00016331, 0.00003361, 0.00019060, 0.00010431, -0.00019122, -0.00001133, -0.00006153, 0.00000367, -0.00001362, -0.00000543, -0.00001864, -0.00000227, -0.00000798, 0.00000295, -0.00001690, 0.00000575, -0.00000451, 0.00000376, -0.00001569, 0.00000259, -0.00000501, -0.00000637, 0.00000540, -0.00001266, 0.00002068, -0.00000477, -0.00000138, -0.00000195, 0.00000075, 0.00000017, 0.00000005, 0.00000343, -0.00000292, -0.00000540, 0.00000576, 0.00000163, -0.00000686, -0.00000140, 0.00000505, 0.00000640, -0.00000662, 0.00000589, -0.00000508, 0.00000401, -0.00001099, 0.00000286, -0.00000056, 0.00000423, -0.00000080, -0.00000011, -0.00000421, -0.00000147, 0.00000450, -0.00000049, -0.00000154, 0.00000054, -0.00000488, -0.00000118, 0.00000426, -0.00000389, 0.00000794, -0.00000223, },
		{ -0.00000290, -0.00000494, 0.00000667, 0.00000176, -0.00000267, -0.00000327, -0.00000262, -0.00000041, 0.00000356, 0.00000507, -0.00000926, -0.00000456, 0.00000438, 0.00000358, -0.00000180, 0.00000665, -0.00000523, -0.00000089, -0.00000160, -0.00000079, 0.00000141, 0.00000412, -0.00000704, 0.00000012, 0.00000340, 0.00000117, -0.00000093, 0.00000114, 0.00000025, -0.00000557, 0.00000593, 0.00000029, 0.00000070, -0.00001068, 0.00000889, 0.00000351, -0.00000844, -0.00000798, 0.00000840, -0.00000542, 0.00000192, -0.00000924, -0.00000029, -0.00000172, -0.00000217, -0.00000525, -0.00000298, -0.00000732, 0.00000403, -0.00000803, -0.00003929, 0.00000642, 0.00000093, 0.00003078, -0.00000499, 0.00001037, 0.00000908, -0.00000253, -0.00001932, 0.00001498, -0.00000367, 0.00001208, -0.00002814, -0.00003889, -0.00018599, 0.00015567, 0.00015422, 0.00003839, 0.00015241, 0.00014930, -0.00017859, -0.00003201, -0.00005033, -0.00000688, -0.00000674, -0.00001102, -0.00001333, -0.00000757, -0.00000440, 0.00000007, -0.00001334, 0.00000196, -0.00000147, 0.00000052, -0.00001058, -0.00000436, 0.00000191, -0.00001234, 0.00000964, -0.00001475, 0.00001893, -0.00000147, -0.00000280, -0.00000148, 0.00000010, 0.00000069, -0.00000064, 0.00000412, -0.00000231, -0.00000592, 0.00000432, 0.00000358, -0.00000697, -0.00000164, 0.00000303, 0.00000873, -0.00000729, 0.00000559, -0.00000438, 0.00000348, -0.00000932, 0.00000014, 0.00000120, 0.00000277, 0.00000086, -0.00000132, -0.00000245, -0.00000369, 0.00000528, -0.00000035, -0.00000126, -0.00000012, -0.00000336, -0.00000325, 0.00000558, -0.00000457, 0.00000732, -0.00000069, },
		{ -0.00000305, -0.00000501, 0.00000467, 0.00000403, -0.00000347, -0.00000238, -0.00000379, 0.00000040, 0.00000180, 0.00000668, -0.00000841, -0.00000534, 0.00000281, 0.00000513, -0.00000286, 0.00000704, -0.00000401, -0.00000204, -0.00000124, -0.00000079, 0.00000061, 0.00000490, -0.00000618, -0.00000126, 0.00000390, 0.00000067, -0.00000026, 0.00000051, 0.00000135, -0.00000616, 0.00000488, 0.00000170, 0.00000094, -0.00001045, 0.00000550, 0.00000715, -0.00000853, -0.00000847, 0.00000616, -0.00000204, -0.00000070, -0.00000686, -0.00000399, 0.00000231, -0.00000570, -0.00000148, -0.00000712, -0.00000245, -0.00000252, -0.00000014, -0.00004394, 0.00000510, -0.00000207, 0.00003314, -0.00000327, 0.00001041, 0.00000928, -0.00000097, -0.00002005, 0.00001671, -0.00000529, 0.00001585, -0.00002842, -0.00003410, -0.00018520, 0.00014839, 0.00014470, 0.00004462, 0.00011762, 0.00017901, -0.00015119, -0.00005857, -0.00003688, -0.00001871, 0.00000089, -0.00001732, -0.00000741, -0.00001321, -0.00000061, -0.00000322, -0.00000927, -0.00000223, 0.00000201, -0.00000335, -0.00000491, -0.00001117, 0.00000774, -0.00001648, 0.00001169, -0.00001458, 0.00001545, 0.00000245, -0.00000419, -0.00000106, -0.00000053, 0.00000114, -0.00000119, 0.00000449, -0.00000146, -0.00000632, 0.00000283, 0.00000509, -0.00000654, -0.00000205, 0.00000134, 0.00001009, -0.00000677, 0.00000424, -0.00000281, 0.00000210, -0.00000683, -0.00000297, 0.00000303, 0.00000126, 0.00000253, -0.00000250, -0.00000077, -0.00000553, 0.00000552, 0.00000008, -0.00000101, -0.00000081, -0.00000197, -0.00000482, 0.00000614, -0.00000440, 0.00000598, 0.00000113, },
		{ -0.00000324, -0.00000470, 0.00000237, 0.00000610, -0.00000393, -0.00000169, -0.00000471, 0.00000095, 0.00000044, 0.00000752, -0.00000696, -0.00000615, 0.00000132, 0.00000618, -0.00000322, 0.00000665, -0.00000227, -0.00000329, -0.00000092, -0.00000076, -0.00000008, 0.00000526, -0.00000491, -0.00000270, 0.00000427, 0.00000031, 0.00000020, 0.00000019, 0.00000193, -0.00000594, 0.00000307, 0.00000348, 0.00000076, -0.00000936, 0.00000157, 0.00001046, -0.00000829, -0.00000850, 0.00000326, 0.00000155, -0.00000326, -0.00000438, -0.00000775, 0.00000607, -0.00000878, 0.00000183, -0.00001070, 0.00000165, -0.00000775, 0.00000539, -0.00004535, 0.00000138, -0.00000370, 0.00003363, 0.00000009, 0.00000940, 0.00001038, -0.00000011, -0.00002013, 0.00001788, -0.00000631, 0.00001890, -0.00002793, -0.00003010, -0.00018388, 0.00014099, 0.00013616, 0.00005000, 0.00008911, 0.00019282, -0.00011173, -0.00008746, -0.00002388, -0.00002982, 0.00000773, -0.00002308, -0.00000198, -0.00001821, 0.00000257, -0.00000617, -0.00000543, -0.00000610, 0.00000522, -0.00000709, 0.00000040, -0.00001684, 0.00001160, -0.00001822, 0.00001133, -0.00001228, 0.00001066, 0.00000653, -0.00000524, -0.00000086, -0.00000101, 0.00000144, -0.00000152, 0.00000450, -0.00000039, -0.00000662, 0.00000140, 0.00000603, -0.00000555, -0.00000268, 0.00000010, 0.00001044, -0.00000516, 0.00000208, -0.00000065, 0.00000019, -0.00000395, -0.00000602, 0.00000453, 0.00000002, 0.00000391, -0.00000338, 0.00000056, -0.00000674, 0.00000510, 0.00000084, -0.00000085, -0.00000139, -0.00000090, -0.00000575, 0.00000592, -0.00000349, 0.00000413, 0.00000305, },
		{ -0.00000330, -0.00000419, 0.00000006, 0.00000768, -0.00000388, -0.00000137, -0.00000523, 0.00000110, -0.00000038, 0.00000757, -0.00000499, -0.00000695, -0.00000002, 0.00000666, -0.00000289, 0.00000561, -0.00000024, -0.00000443, -0.00000074, -0.00000068, -0.00000065, 0.00000526, -0.00000337, -0.00000403, 0.00000440, 0.00000017, 0.00000036, 0.00000023, 0.00000198, -0.00000500, 0.00000073, 0.00000537, 0.00000042, -0.00000780, -0.00000239, 0.00001301, -0.00000749, -0.00000830, 0.00000010, 0.00000483, -0.00000530, -0.00000229, -0.00001099, 0.00000893, -0.00001082, 0.00000409, -0.00001309, 0.00000428, -0.00001093, 0.00000802, -0.00004345, -0.00000442, -0.00000427, 0.00003244, 0.00000484, 0.00000758, 0.00001217, 0.00000016, -0.00001960, 0.00001843, -0.00000659, 0.00002102, -0.00002642, -0.00002722, -0.00018152, 0.00013275, 0.00012954, 0.00005302, 0.00006834, 0.00019191, -0.00006425, -0.00011465, -0.00001404, -0.00003838, 0.00001241, -0.00002719, 0.00000198, -0.00002173, 0.00000445, -0.00000817, -0.00000244, -0.00000905, 0.00000757, -0.00001005, 0.00000460, -0.00002064, 0.00001298, -0.00001739, 0.00000874, -0.00000831, 0.00000515, 0.00001024, -0.00000565, -0.00000102, -0.00000125, 0.00000151, -0.00000157, 0.00000415, 0.00000084, -0.00000682, 0.00000013, 0.00000638, -0.00000407, -0.00000347, -0.00000067, 0.00000985, -0.00000275, -0.00000052, 0.00000171, -0.00000185, -0.00000111, -0.00000858, 0.00000537, -0.00000070, 0.00000478, -0.00000376, 0.00000134, -0.00000722, 0.00000401, 0.00000188, -0.00000079, -0.00000179, -0.00000025, -0.00000600, 0.00000499, -0.00000201, 0.00000203, 0.00000481, },
		{ -0.00000310, -0.00000367, -0.00000200, 0.00000857, -0.00000320, -0.00000147, -0.00000531, 0.00000080, -0.00000064, 0.00000687, -0.00000266, -0.00000763, -0.00000121, 0.00000659, -0.00000201, 0.00000414, 0.00000183, -0.00000529, -0.00000078, -0.00000050, -0.00000113, 0.00000495, -0.00000171, -0.00000510, 0.00000418, 0.00000030, 0.00000023, 0.00000062, 0.00000157, -0.00000354, -0.00000182, 0.00000704, 0.00000022, -0.00000614, -0.00000590, 0.00001447, -0.00000601, -0.00000804, -0.00000294, 0.00000737, -0.00000644, -0.00000096, -0.00001329, 0.00001050, -0.00001148, 0.00000495, -0.00001394, 0.00000508, -0.00001181, 0.00000774, -0.00003868, -0.00001153, -0.00000435, 0.00002996, 0.00001051, 0.00000541, 0.00001432, 0.00000012, -0.00001863, 0.00001848, -0.00000620, 0.00002224, -0.00002387, -0.00002554, -0.00017795, 0.00012334, 0.00012524, 0.00005302, 0.00005543, 0.00017888, -0.00001350, -0.00013624, -0.00000955, -0.00004307, 0.00001403, -0.00002894, 0.00000388, -0.00002328, 0.00000464, -0.00000887, -0.00000066, -0.00001069, 0.00000869, -0.00001180, 0.00000722, -0.00002219, 0.00001179, -0.00001426, 0.00000444, -0.00000336, -0.00000033, 0.00001304, -0.00000516, -0.00000165, -0.00000121, 0.00000131, -0.00000134, 0.00000351, 0.00000215, -0.00000689, -0.00000096, 0.00000614, -0.00000221, -0.00000434, -0.00000104, 0.00000853, 0.00000012, -0.00000316, 0.00000387, -0.00000362, 0.00000128, -0.00001029, 0.00000531, -0.00000074, 0.00000503, -0.00000355, 0.00000152, -0.00000695, 0.00000236, 0.00000308, -0.00000077, -0.00000198, -0.00000002, -0.00000566, 0.00000354, -0.00000026, -0.00000002, 0.00000619, },
		{ -0.00000254, -0.00000326, -0.00000362, 0.00000865, -0.00000190, -0.00000195, -0.00000497, 0.00000008, -0.00000040, 0.00000562, -0.00000021, -0.00000806, -0.00000228, 0.00000608, -0.00000076, 0.00000251, 0.00000367, -0.00000570, -0.00000109, -0.00000022, -0.00000152, 0.00000443, -0.00000009, -0.00000577, 0.00000358, 0.00000070, -0.00000018, 0.00000128, 0.00000084, -0.00000180, -0.00000423, 0.00000816, 0.00000041, -0.00000470, -0.00000863, 0.00001466, -0.00000384, -0.00000782, -0.00000560, 0.00000888, -0.00000649, -0.00000058, -0.00001444, 0.00001059, -0.00001066, 0.00000439, -0.00001322, 0.00000408, -0.00001052, 0.00000501, -0.00003193, -0.00001892, -0.00000467, 0.00002672, 0.00001650, 0.00000341, 0.00001640, 0.00000010, -0.00001751, 0.00001824, -0.00000531, 0.00002273, -0.00002045, -0.00002493, -0.00017326, 0.00011281, 0.00012313, 0.00005011, 0.00004940, 0.00015723, 0.00003572, -0.00014900, -0.00001177, -0.00004331, 0.00001228, -0.00002814, 0.00000354, -0.00002275, 0.00000310, -0.00000821, -0.00000021, -0.00001092, 0.00000847, -0.00001222, 0.00000812, -0.00002148, 0.00000833, -0.00000944, -0.00000078, 0.00000174, -0.00000505, 0.00001448, -0.00000363, -0.00000274, -0.00000090, 0.00000087, -0.00000088, 0.00000265, 0.00000343, -0.00000680, -0.00000185, 0.00000541, -0.00000016, -0.00000515, -0.00000115, 0.00000676, 0.00000303, -0.00000542, 0.00000547, -0.00000481, 0.00000290, -0.00001092, 0.00000429, -0.00000012, 0.00000467, -0.00000278, 0.00000115, -0.00000606, 0.00000034, 0.00000426, -0.00000071, -0.00000202, -0.00000015, -0.00000488, 0.00000183, 0.00000148, -0.00000175, 0.00000700, },
		{ -0.00000158, -0.00000305, -0.00000469, 0.00000794, -0.00000010, -0.00000268, -0.00000434, -0.00000091, 0.00000017, 0.00000405, 0.00000207, -0.00000808, -0.00000331, 0.00000525, 0.00000063, 0.00000099, 0.00000506, -0.00000559, -0.00000170, 0.00000016, -0.00000185, 0.00000377, 0.00000137, -0.00000597, 0.00000261, 0.00000128, -0.00000076, 0.00000208, -0.00000002, -0.00000006, -0.00000620, 0.00000851, 0.00000116, -0.00000369, -0.00001037, 0.00001362, -0.00000114, -0.00000761, -0.00000771, 0.00000926, -0.00000544, -0.00000111, -0.00001445, 0.00000932, -0.00000856, 0.00000264, -0.00001122, 0.00000164, -0.00000762, 0.00000067, -0.00002436, -0.00002546, -0.00000592, 0.00002324, 0.00002221, 0.00000213, 0.00001799, 0.00000045, -0.00001652, 0.00001799, -0.00000419, 0.00002276, -0.00001644, -0.00002507, -0.00016776, 0.00010153, 0.00012273, 0.00004497, 0.00004858, 0.00013081, 0.00007913, -0.00015085, -0.00002091, -0.00003938, 0.00000754, -0.00002511, 0.00000125, -0.00002044, 0.00000013, -0.00000643, -0.00000090, -0.00000991, 0.00000709, -0.00001144, 0.00000746, -0.00001888, 0.00000325, -0.00000377, -0.00000602, 0.00000617, -0.00000839, 0.00001427, -0.00000108, -0.00000419, -0.00000040, 0.00000026, -0.00000029, 0.00000171, 0.00000456, -0.00000649, -0.00000257, 0.00000431, 0.00000190, -0.00000572, -0.00000117, 0.00000484, 0.00000561, -0.00000697, 0.00000628, -0.00000521, 0.00000361, -0.00001044, 0.00000237, 0.00000105, 0.00000384, -0.00000160, 0.00000038, -0.00000476, -0.00000179, 0.00000521, -0.00000050, -0.00000196, -0.00000052, -0.00000387, 0.00000011, 0.00000294, -0.00000293, 0.00000716, },
		{ -0.00000027, -0.00000304, -0.00000521, 0.00000653, 0.00000199, -0.00000348, -0.00000358, -0.00000200, 0.00000086, 0.00000244, 0.00000395, -0.00000757, -0.00000436, 0.00000424, 0.00000193, -0.00000020, 0.00000587, -0.00000493, -0.00000253, 0.00000060, -0.00000211, 0.00000303, 0.00000257, -0.00000569, 0.00000136, 0.00000196, -0.00000140, 0.00000288, -0.00000082, 0.00000142, -0.00000748, 0.00000794, 0.00000250, -0.00000316, -0.00001112, 0.00001153, 0.00000180, -0.00000730, -0.00000926, 0.00000856, -0.00000348, -0.00000237, -0.00001355, 0.00000699, -0.00000560, 0.00000016, -0.00000842, -0.00000165, -0.00000388, -0.00000427, -0.00001716, -0.00003017, -0.00000860, 0.00001986, 0.00002714, 0.00000197, 0.00001880, 0.00000142, -0.00001589, 0.00001794, -0.00000307, 0.00002263, -0.00001218, -0.00002559, -0.00016189, 0.00008999, 0.00012330, 0.00003857, 0.00005098, 0.00010327, 0.00011347, -0.00014110, -0.00003608, -0.00003237, 0.00000071, -0.00002062, -0.00000235, -0.00001694, -0.00000370, -0.00000402, -0.00000235, -0.00000803, 0.00000491, -0.00000982, 0.00000562, -0.00001496, -0.00000262, 0.00000178, -0.00001041, 0.00000925, -0.00000996, 0.00001238, 0.00000231, -0.00000580, 0.00000015, -0.00000039, 0.00000033, 0.00000081, 0.00000543, -0.00000595, -0.00000314, 0.00000298, 0.00000375, -0.00000591, -0.00000125, 0.00000302, 0.00000756, -0.00000762, 0.00000618, -0.00000478, 0.00000341, -0.00000895, -0.00000019, 0.00000251, 0.00000276, -0.00000021, -0.00000056, -0.00000331, -0.00000376, 0.00000574, -0.00000005, -0.00000187, -0.00000101, -0.00000283, -0.00000136, 0.00000391, -0.00000345, 0.00000664, },
		{ 0.00000125, -0.00000317, -0.00000526, 0.00000464, 0.00000412, -0.00000411, -0.00000289, -0.00000300, 0.00000145, 0.00000103, 0.00000521, -0.00000647, -0.00000544, 0.00000318, 0.00000299, -0.00000094, 0.00000606, -0.00000380, -0.00000351, 0.00000104, -0.00000229, 0.00000227, 0.00000347, -0.00000497, -0.00000005, 0.00000259, -0.00000197, 0.00000353, -0.00000140, 0.00000246, -0.00000793, 0.00000649, 0.00000432, -0.00000303, -0.00001105, 0.00000875, 0.00000465, -0.00000676, -0.00001030, 0.00000695, -0.00000092, -0.00000401, -0.00001209, 0.00000406, -0.00000233, -0.00000248, -0.00000545, -0.00000509, -0.00000012, -0.00000880, -0.00001137, -0.00003235, -0.00001290, 0.00001676, 0.00003097, 0.00000317, 0.00001869, 0.00000313, -0.00001573, 0.00001825, -0.00000215, 0.00002259, -0.00000795, -0.00002613, -0.00015608, 0.00007874, 0.00012409, 0.00003196, 0.00005468, 0.00007764, 0.00013679, -0.00012046, -0.00005534, -0.00002395, -0.00000688, -0.00001570, -0.00000640, -0.00001304, -0.00000771, -0.00000154, -0.00000406, -0.00000575, 0.00000242, -0.00000781, 0.00000315, -0.00001044, -0.00000838, 0.00000635, -0.00001327, 0.00001055, -0.00000962, 0.00000899, 0.00000619, -0.00000728, 0.00000057, -0.00000095, 0.00000085, 0.00000007, 0.00000597, -0.00000516, -0.00000359, 0.00000156, 0.00000523, -0.00000564, -0.00000150, 0.00000151, 0.00000871, -0.00000730, 0.00000523, -0.00000363, 0.00000245, -0.00000672, -0.00000306, 0.00000393, 0.00000170, 0.00000113, -0.00000143, -0.00000198, -0.00000532, 0.00000570, 0.00000069, -0.00000181, -0.00000150, -0.00000194, -0.00000243, 0.00000429, -0.00000331, 0.00000556, },
		{ 0.00000282, -0.00000331, -0.00000498, 0.00000252, 0.00000599, -0.00000437, -0.00000244, -0.00000373, 0.00000176, 0.00000002, 0.00000575, -0.00000482, -0.00000653, 0.00000212, 0.00000371, -0.00000118, 0.00000568, -0.00000234, -0.00000448, 0.00000139, -0.00000235, 0.00000151, 0.00000404, -0.00000390, -0.00000148, 0.00000304, -0.00000235, 0.00000393, -0.00000164, 0.00000296, -0.00000756, 0.00000429, 0.00000641, -0.00000308, -0.00001043, 0.00000569, 0.00000702, -0.00000581, -0.00001094, 0.00000472, 0.00000185, -0.00000564, -0.00001049, 0.00000104, 0.00000071, -0.00000472, -0.00000286, -0.00000804, 0.00000291, -0.00001211, -0.00000771, -0.00003172, -0.00001862, 0.00001381, 0.00003358, 0.00000577, 0.00001772, 0.00000553, -0.00001603, 0.00001893, -0.00000149, 0.00002275, -0.00000396, -0.00002643, -0.00015067, 0.00006822, 0.00012446, 0.00002602, 0.00005810, 0.00005606, 0.00014857, -0.00009095, -0.00007604, -0.00001609, -0.00001383, -0.00001143, -0.00001003, -0.00000951, -0.00001120, 0.00000045, -0.00000556, -0.00000357, 0.00000011, -0.00000590, 0.00000058, -0.00000604, -0.00001325, 0.00000928, -0.00001420, 0.00000995, -0.00000753, 0.00000451, 0.00001010, -0.00000833, 0.00000067, -0.00000131, 0.00000115, -0.00000040, 0.00000613, -0.00000414, -0.00000393, 0.00000015, 0.00000623, -0.00000486, -0.00000197, 0.00000042, 0.00000901, -0.00000612, 0.00000363, -0.00000200, 0.00000101, -0.00000411, -0.00000584, 0.00000498, 0.00000091, 0.00000219, -0.00000201, -0.00000098, -0.00000629, 0.00000505, 0.00000169, -0.00000179, -0.00000189, -0.00000132, -0.00000301, 0.00000408, -0.00000259, 0.00000409, },
		{ 0.00000424, -0.00000335, -0.00000453, 0.00000043, 0.00000737, -0.00000411, -0.00000233, -0.00000410, 0.00000171, -0.00000050, 0.00000557, -0.00000272, -0.00000752, 0.00000108, 0.00000405, -0.00000096, 0.00000488, -0.00000075, -0.00000530, 0.00000158, -0.00000227, 0.00000076, 0.00000431, -0.00000258, -0.00000277, 0.00000320, -0.00000247, 0.00000401, -0.00000151, 0.00000293, -0.00000647, 0.00000163, 0.00000843, -0.00000305, -0.00000959, 0.00000278, 0.00000861, -0.00000436, -0.00001128, 0.00000214, 0.00000445, -0.00000690, -0.00000912, -0.00000163, 0.00000303, -0.00000612, -0.00000111, -0.00001001, 0.00000468, -0.00001370, -0.00000647, -0.00002848, -0.00002522, 0.00001071, 0.00003501, 0.00000959, 0.00001610, 0.00000844, -0.00001661, 0.00001989, -0.00000107, 0.00002316, -0.00000027, -0.00002637, -0.00014584, 0.00005872, 0.00012396, 0.00002137, 0.00006017, 0.00003968, 0.00014956, -0.00005550, -0.00009524, -0.00001072, -0.00001888, -0.00000871, -0.00001251, -0.00000698, -0.00001365, 0.00000152, -0.00000647, -0.00000188, -0.00000163, -0.00000448, -0.00000162, -0.00000233, -0.00001664, 0.00001018, -0.00001313, 0.00000765, -0.00000413, -0.00000047, 0.00001351, -0.00000864, 0.00000033, -0.00000136, 0.00000118, -0.00000058, 0.00000593, -0.00000294, -0.00000416, -0.00000115, 0.00000668, -0.00000360, -0.00000262, -0.00000025, 0.00000855, -0.00000431, 0.00000166, -0.00000018, -0.00000059, -0.00000150, -0.00000814, 0.00000539, 0.00000061, 0.00000281, -0.00000215, -0.00000044, -0.00000661, 0.00000382, 0.00000288, -0.00000178, -0.00000216, -0.00000101, -0.00000311, 0.00000338, -0.00000147, 0.00000246, },
		{ 0.00000532, -0.00000316, -0.00000407, -0.00000140, 0.00000807, -0.00000326, -0.00000258, -0.00000409, 0.00000126, -0.00000054, 0.00000477, -0.00000039, -0.00000828, 0.00000004, 0.00000407, -0.00000040, 0.00000383, 0.00000077, -0.00000584, 0.00000155, -0.00000203, 0.00000003, 0.00000432, -0.00000115, -0.00000378, 0.00000300, -0.00000230, 0.00000379, -0.00000104, 0.00000246, -0.00000489, -0.00000117, 0.00001005, -0.00000267, -0.00000883, 0.00000036, 0.00000922, -0.00000240, -0.00001139, -0.00000052, 0.00000655, -0.00000751, -0.00000826, -0.00000361, 0.00000431, -0.00000642, -0.00000041, -0.00001079, 0.00000499, -0.00001345, -0.00000750, -0.00002325, -0.00003188, 0.00000704, 0.00003542, 0.00001433, 0.00001421, 0.00001154, -0.00001725, 0.00002094, -0.00000073, 0.00002372, 0.00000318, -0.00002596, -0.00014166, 0.00005032, 0.00012242, 0.00001828, 0.00006037, 0.00002872, 0.00014152, -0.00001757, -0.00011014, -0.00000932, -0.00002118, -0.00000810, -0.00001342, -0.00000583, -0.00001475, 0.00000148, -0.00000660, -0.00000091, -0.00000256, -0.00000379, -0.00000314, 0.00000031, -0.00001825, 0.00000903, -0.00001034, 0.00000413, -0.00000002, -0.00000532, 0.00001597, -0.00000800, -0.00000053, -0.00000109, 0.00000092, -0.00000046, 0.00000541, -0.00000167, -0.00000425, -0.00000230, 0.00000660, -0.00000198, -0.00000338, -0.00000057, 0.00000751, -0.00000216, -0.00000034, 0.00000151, -0.00000202, 0.00000075, -0.00000966, 0.00000498, 0.00000086, 0.00000292, -0.00000182, -0.00000039, -0.00000630, 0.00000213, 0.00000411, -0.00000171, -0.00000230, -0.00000100, -0.00000281, 0.00000234, -0.00000019, 0.00000091, },
		{ 0.00000593, -0.00000268, -0.00000372, -0.00000282, 0.00000802, -0.00000186, -0.00000314, -0.00000374, 0.00000050, -0.00000021, 0.00000354, 0.00000193, -0.00000865, -0.00000106, 0.00000384, 0.00000034, 0.00000276, 0.00000203, -0.00000599, 0.00000127, -0.00000165, -0.00000066, 0.00000411, 0.00000027, -0.00000442, 0.00000242, -0.00000189, 0.00000332, -0.00000032, 0.00000170, -0.00000307, -0.00000375, 0.00001098, -0.00000175, -0.00000837, -0.00000135, 0.00000881, -0.00000004, -0.00001127, -0.00000307, 0.00000793, -0.00000733, -0.00000805, -0.00000475, 0.00000445, -0.00000561, -0.00000076, -0.00001040, 0.00000392, -0.00001161, -0.00001025, -0.00001693, -0.00003768, 0.00000239, 0.00003499, 0.00001962, 0.00001245, 0.00001449, -0.00001765, 0.00002185, -0.00000031, 0.00002429, 0.00000650, -0.00002530, -0.00013803, 0.00004295, 0.00011985, 0.00001670, 0.00005867, 0.00002266, 0.00012688, 0.00001938, -0.00011847, -0.00001274, -0.00002046, -0.00000969, -0.00001272, -0.00000608, -0.00001452, 0.00000036, -0.00000597, -0.00000070, -0.00000262, -0.00000388, -0.00000389, 0.00000173, -0.00001807, 0.00000612, -0.00000637, 0.00000004, 0.00000413, -0.00000941, 0.00001708, -0.00000633, -0.00000187, -0.00000054, 0.00000043, -0.00000013, 0.00000469, -0.00000042, -0.00000416, -0.00000329, 0.00000605, -0.00000013, -0.00000412, -0.00000067, 0.00000613, 0.00000000, -0.00000205, 0.00000278, -0.00000303, 0.00000240, -0.00001022, 0.00000371, 0.00000165, 0.00000256, -0.00000105, -0.00000076, -0.00000550, 0.00000020, 0.00000520, -0.00000151, -0.00000236, -0.00000121, -0.00000227, 0.00000118, 0.00000103, -0.00000035, },
		{ 0.00000602, -0.00000189, -0.00000354, -0.00000374, 0.00000726, -0.00000005, -0.00000385, -0.00000320, -0.00000044, 0.00000034, 0.00000213, 0.00000399, -0.00000851, -0.00000227, 0.00000346, 0.00000111, 0.00000184, 0.00000290, -0.00000573, 0.00000077, -0.00000116, -0.00000129, 0.00000372, 0.00000159, -0.00000464, 0.00000151, -0.00000132, 0.00000271, 0.00000050, 0.00000085, -0.00000130, -0.00000579, 0.00001100, -0.00000018, -0.00000829, -0.00000229, 0.00000751, 0.00000250, -0.00001086, -0.00000541, 0.00000850, -0.00000637, -0.00000845, -0.00000507, 0.00000356, -0.00000389, -0.00000190, -0.00000912, 0.00000184, -0.00000870, -0.00001395, -0.00001057, -0.00004177, -0.00000351, 0.00003384, 0.00002508, 0.00001124, 0.00001698, -0.00001755, 0.00002242, 0.00000038, 0.00002471, 0.00000987, -0.00002454, -0.00013480, 0.00003645, 0.00011646, 0.00001637, 0.00005543, 0.00002048, 0.00010830, 0.00005231, -0.00011882, -0.00002097, -0.00001707, -0.00001310, -0.00001071, -0.00000750, -0.00001322, -0.00000156, -0.00000479, -0.00000110, -0.00000194, -0.00000463, -0.00000393, 0.00000201, -0.00001638, 0.00000199, -0.00000192, -0.00000391, 0.00000765, -0.00001225, 0.00001666, -0.00000368, -0.00000355, 0.00000018, -0.00000020, 0.00000033, 0.00000389, 0.00000068, -0.00000386, -0.00000411, 0.00000512, 0.00000178, -0.00000468, -0.00000070, 0.00000467, 0.00000188, -0.00000322, 0.00000345, -0.00000345, 0.00000330, -0.00000979, 0.00000170, 0.00000281, 0.00000189, -0.00000000, -0.00000139, -0.00000441, -0.00000173, 0.00000595, -0.00000107, -0.00000239, -0.00000155, -0.00000165, 0.00000009, 0.00000199, -0.00000116, },
		{ 0.00000560, -0.00000084, -0.00000351, -0.00000418, 0.00000590, 0.00000195, -0.00000453, -0.00000263, -0.00000140, 0.00000090, 0.00000077, 0.00000555, -0.00000776, -0.00000360, 0.00000301, 0.00000176, 0.00000119, 0.00000331, -0.00000508, 0.00000011, -0.00000061, -0.00000181, 0.00000317, 0.00000273, -0.00000443, 0.00000035, -0.00000070, 0.00000209, 0.00000129, 0.00000008, 0.00000018, -0.00000708, 0.00001004, 0.00000196, -0.00000855, -0.00000256, 0.00000558, 0.00000493, -0.00001006, -0.00000747, 0.00000829, -0.00000477, -0.00000932, -0.00000474, 0.00000189, -0.00000160, -0.00000345, -0.00000737, -0.00000075, -0.00000537, -0.00001773, -0.00000512, -0.00004351, -0.00001069, 0.00003194, 0.00003040, 0.00001088, 0.00001880, -0.00001678, 0.00002251, 0.00000147, 0.00002484, 0.00001344, -0.00002383, -0.00013181, 0.00003059, 0.00011252, 0.00001690, 0.00005117, 0.00002093, 0.00008832, 0.00007889, -0.00011083, -0.00003319, -0.00001192, -0.00001761, -0.00000799, -0.00000957, -0.00001129, -0.00000386, -0.00000340, -0.00000188, -0.00000077, -0.00000579, -0.00000347, 0.00000141, -0.00001360, -0.00000270, 0.00000227, -0.00000705, 0.00001004, -0.00001354, 0.00001472, -0.00000030, -0.00000533, 0.00000089, -0.00000084, 0.00000077, 0.00000315, 0.00000152, -0.00000331, -0.00000479, 0.00000393, 0.00000355, -0.00000493, -0.00000081, 0.00000333, 0.00000324, -0.00000372, 0.00000345, -0.00000326, 0.00000346, -0.00000848, -0.00000084, 0.00000411, 0.00000110, 0.00000114, -0.00000209, -0.00000327, -0.00000342, 0.00000622, -0.00000035, -0.00000243, -0.00000193, -0.00000108, -0.00000076, 0.00000257, -0.00000147, },
		{ 0.00000478, 0.00000034, -0.00000359, -0.00000422, 0.00000415, 0.00000391, -0.00000496, -0.00000219, -0.00000220, 0.00000129, -0.00000031, 0.00000647, -0.00000640, -0.00000501, 0.00000251, 0.00000222, 0.00000087, 0.00000331, -0.00000416, -0.00000059, -0.00000009, -0.00000219, 0.00000250, 0.00000365, -0.00000383, -0.00000093, -0.00000016, 0.00000157, 0.00000190, -0.00000043, 0.00000121, -0.00000752, 0.00000818, 0.00000451, -0.00000897, -0.00000239, 0.00000338, 0.00000693, -0.00000876, -0.00000927, 0.00000740, -0.00000274, -0.00001040, -0.00000404, -0.00000019, 0.00000082, -0.00000496, -0.00000561, -0.00000330, -0.00000230, -0.00002082, -0.00000135, -0.00004262, -0.00001884, 0.00002913, 0.00003535, 0.00001158, 0.00001986, -0.00001529, 0.00002208, 0.00000299, 0.00002464, 0.00001727, -0.00002328, -0.00012892, 0.00002517, 0.00010831, 0.00001787, 0.00004650, 0.00002278, 0.00006907, 0.00009776, -0.00009512, -0.00004788, -0.00000627, -0.00002225, -0.00000529, -0.00001172, -0.00000929, -0.00000606, -0.00000215, -0.00000272, 0.00000059, -0.00000708, -0.00000279, 0.00000025, -0.00001027, -0.00000724, 0.00000553, -0.00000893, 0.00001102, -0.00001324, 0.00001148, 0.00000347, -0.00000692, 0.00000141, -0.00000134, 0.00000108, 0.00000258, 0.00000204, -0.00000253, -0.00000533, 0.00000258, 0.00000506, -0.00000479, -0.00000108, 0.00000228, 0.00000398, -0.00000351, 0.00000284, -0.00000255, 0.00000299, -0.00000653, -0.00000357, 0.00000526, 0.00000042, 0.00000216, -0.00000266, -0.00000229, -0.00000467, 0.00000590, 0.00000065, -0.00000249, -0.00000225, -0.00000070, -0.00000129, 0.00000272, -0.00000129, },
		{ 0.00000372, 0.00000152, -0.00000367, -0.00000398, 0.00000224, 0.00000555, -0.00000498, -0.00000202, -0.00000271, 0.00000139, -0.00000100, 0.00000669, -0.00000451, -0.00000640, 0.00000197, 0.00000248, 0.00000085, 0.00000297, -0.00000311, -0.00000122, 0.00000032, -0.00000238, 0.00000173, 0.00000433, -0.00000292, -0.00000219, 0.00000016, 0.00000126, 0.00000225, -0.00000062, 0.00000172, -0.00000714, 0.00000562, 0.00000714, -0.00000927, -0.00000207, 0.00000127, 0.00000821, -0.00000692, -0.00001081, 0.00000599, -0.00000057, -0.00001142, -0.00000324, -0.00000231, 0.00000293, -0.00000603, -0.00000427, -0.00000535, -0.00000002, -0.00002266, 0.00000037, -0.00003926, -0.00002737, 0.00002513, 0.00003977, 0.00001341, 0.00002019, -0.00001313, 0.00002120, 0.00000490, 0.00002413, 0.00002140, -0.00002290, -0.00012608, 0.00002006, 0.00010402, 0.00001895, 0.00004190, 0.00002502, 0.00005205, 0.00010856, -0.00007325, -0.00006306, -0.00000155, -0.00002606, -0.00000331, -0.00001337, -0.00000771, -0.00000774, -0.00000139, -0.00000337, 0.00000186, -0.00000819, -0.00000217, -0.00000112, -0.00000691, -0.00001103, 0.00000738, -0.00000929, 0.00001056, -0.00001153, 0.00000735, 0.00000717, -0.00000802, 0.00000157, -0.00000159, 0.00000115, 0.00000227, 0.00000221, -0.00000154, -0.00000572, 0.00000118, 0.00000618, -0.00000421, -0.00000156, 0.00000157, 0.00000409, -0.00000270, 0.00000177, -0.00000149, 0.00000213, -0.00000422, -0.00000616, 0.00000597, 0.00000005, 0.00000289, -0.00000292, -0.00000163, -0.00000535, 0.00000500, 0.00000187, -0.00000254, -0.00000249, -0.00000054, -0.00000148, 0.00000248, -0.00000075, },
		{ 0.00000260, 0.00000254, -0.00000365, -0.00000362, 0.00000040, 0.00000668, -0.00000447, -0.00000219, -0.00000288, 0.00000114, -0.00000124, 0.00000627, -0.00000226, -0.00000763, 0.00000133, 0.00000256, 0.00000106, 0.00000244, -0.00000209, -0.00000165, 0.00000056, -0.00000236, 0.00000088, 0.00000477, -0.00000179, -0.00000330, 0.00000018, 0.00000121, 0.00000230, -0.00000046, 0.00000173, -0.00000610, 0.00000268, 0.00000952, -0.00000918, -0.00000188, -0.00000041, 0.00000861, -0.00000455, -0.00001206, 0.00000420, 0.00000150, -0.00001215, -0.00000262, -0.00000414, 0.00000441, -0.00000635, -0.00000362, -0.00000658, 0.00000117, -0.00002300, 0.00000005, -0.00003395, -0.00003548, 0.00001962, 0.00004356, 0.00001637, 0.00001996, -0.00001048, 0.00002003, 0.00000707, 0.00002340, 0.00002573, -0.00002266, -0.00012329, 0.00001520, 0.00009979, 0.00001991, 0.00003774, 0.00002692, 0.00003814, 0.00011177, -0.00004736, -0.00007661, 0.00000096, -0.00002826, -0.00000258, -0.00001414, -0.00000690, -0.00000862, -0.00000129, -0.00000366, 0.00000285, -0.00000891, -0.00000180, -0.00000240, -0.00000394, -0.00001364, 0.00000760, -0.00000817, 0.00000888, -0.00000880, 0.00000285, 0.00001034, -0.00000836, 0.00000125, -0.00000152, 0.00000095, 0.00000223, 0.00000204, -0.00000041, -0.00000595, -0.00000020, 0.00000686, -0.00000322, -0.00000222, 0.00000120, 0.00000365, -0.00000148, 0.00000047, -0.00000031, 0.00000111, -0.00000190, -0.00000826, 0.00000604, 0.00000013, 0.00000323, -0.00000280, -0.00000138, -0.00000545, 0.00000359, 0.00000318, -0.00000254, -0.00000264, -0.00000061, -0.00000138, 0.00000196, 0.00000002, },
		{ 0.00000161, 0.00000327, -0.00000345, -0.00000325, -0.00000117, 0.00000716, -0.00000340, -0.00000266, -0.00000273, 0.00000056, -0.00000109, 0.00000533, 0.00000013, -0.00000853, 0.00000053, 0.00000252, 0.00000138, 0.00000188, -0.00000126, -0.00000182, 0.00000059, -0.00000214, -0.00000000, 0.00000496, -0.00000055, -0.00000415, -0.00000015, 0.00000142, 0.00000209, -0.00000000, 0.00000135, -0.00000462, -0.00000028, 0.00001128, -0.00000846, -0.00000206, -0.00000146, 0.00000808, -0.00000179, -0.00001296, 0.00000218, 0.00000327, -0.00001242, -0.00000233, -0.00000547, 0.00000506, -0.00000581, -0.00000375, -0.00000693, 0.00000121, -0.00002192, -0.00000195, -0.00002753, -0.00004226, 0.00001236, 0.00004655, 0.00002035, 0.00001940, -0.00000755, 0.00001878, 0.00000932, 0.00002261, 0.00003017, -0.00002244, -0.00012064, 0.00001061, 0.00009565, 0.00002066, 0.00003419, 0.00002808, 0.00002759, 0.00010861, -0.00001990, -0.00008661, 0.00000036, -0.00002843, -0.00000333, -0.00001385, -0.00000700, -0.00000861, -0.00000188, -0.00000355, 0.00000346, -0.00000916, -0.00000181, -0.00000341, -0.00000166, -0.00001485, 0.00000624, -0.00000586, 0.00000640, -0.00000556, -0.00000145, 0.00001255, -0.00000776, 0.00000042, -0.00000112, 0.00000047, 0.00000244, 0.00000162, 0.00000075, -0.00000598, -0.00000152, 0.00000709, -0.00000189, -0.00000297, 0.00000109, 0.00000284, -0.00000010, -0.00000081, 0.00000075, 0.00000018, 0.00000015, -0.00000964, 0.00000534, 0.00000070, 0.00000315, -0.00000229, -0.00000152, -0.00000502, 0.00000183, 0.00000442, -0.00000238, -0.00000274, -0.00000085, -0.00000110, 0.00000132, 0.00000082, },
		{ 0.00000088, 0.00000363, -0.00000303, -0.00000298, -0.00000235, 0.00000694, -0.00000186, -0.00000336, -0.00000234, -0.00000023, -0.00000068, 0.00000411, 0.00000238, -0.00000892, -0.00000050, 0.00000246, 0.00000168, 0.00000143, -0.00000071, -0.00000168, 0.00000043, -0.00000175, -0.00000088, 0.00000491, 0.00000074, -0.00000466, -0.00000081, 0.00000182, 0.00000169, 0.00000063, 0.00000077, -0.00000298, -0.00000289, 0.00001214, -0.00000695, -0.00000269, -0.00000181, 0.00000673, 0.00000114, -0.00001340, 0.00000001, 0.00000460, -0.00001217, -0.00000244, -0.00000623, 0.00000486, -0.00000448, -0.00000454, -0.00000651, 0.00000030, -0.00001979, -0.00000498, -0.00002096, -0.00004689, 0.00000329, 0.00004851, 0.00002526, 0.00001876, -0.00000458, 0.00001766, 0.00001148, 0.00002191, 0.00003459, -0.00002211, -0.00011824, 0.00000635, 0.00009155, 0.00002118, 0.00003126, 0.00002841, 0.00002018, 0.00010069, 0.00000672, -0.00009159, -0.00000373, -0.00002660, -0.00000543, -0.00001261, -0.00000792, -0.00000781, -0.00000303, -0.00000314, 0.00000372, -0.00000894, -0.00000217, -0.00000408, -0.00000018, -0.00001468, 0.00000356, -0.00000283, 0.00000365, -0.00000236, -0.00000506, 0.00001351, -0.00000620, -0.00000086, -0.00000048, -0.00000019, 0.00000281, 0.00000105, 0.00000183, -0.00000578, -0.00000275, 0.00000691, -0.00000035, -0.00000370, 0.00000110, 0.00000187, 0.00000121, -0.00000183, 0.00000150, -0.00000047, 0.00000173, -0.00001015, 0.00000388, 0.00000166, 0.00000274, -0.00000147, -0.00000196, -0.00000423, -0.00000005, 0.00000540, -0.00000199, -0.00000282, -0.00000119, -0.00000075, 0.00000069, 0.00000150, },
		{ 0.00000049, 0.00000362, -0.00000240, -0.00000283, -0.00000310, 0.00000610, -0.00000000, -0.00000410, -0.00000186, -0.00000111, -0.00000018, 0.00000283, 0.00000425, -0.00000868, -0.00000178, 0.00000241, 0.00000187, 0.00000119, -0.00000050, -0.00000126, 0.00000013, -0.00000124, -0.00000168, 0.00000462, 0.00000198, -0.00000481, -0.00000176, 0.00000232, 0.00000121, 0.00000130, 0.00000015, -0.00000147, -0.00000487, 0.00001194, -0.00000467, -0.00000375, -0.00000156, 0.00000481, 0.00000391, -0.00001326, -0.00000225, 0.00000546, -0.00001143, -0.00000289, -0.00000646, 0.00000394, -0.00000258, -0.00000575, -0.00000561, -0.00000118, -0.00001713, -0.00000828, -0.00001521, -0.00004879, -0.00000737, 0.00004908, 0.00003098, 0.00001829, -0.00000177, 0.00001685, 0.00001341, 0.00002141, 0.00003886, -0.00002156, -0.00011619, 0.00000250, 0.00008744, 0.00002154, 0.00002887, 0.00002797, 0.00001538, 0.00008976, 0.00003045, -0.00009077, -0.00001106, -0.00002320, -0.00000849, -0.00001077, -0.00000939, -0.00000651, -0.00000446, -0.00000265, 0.00000377, -0.00000836, -0.00000279, -0.00000446, 0.00000051, -0.00001330, 0.00000003, 0.00000034, 0.00000117, 0.00000031, -0.00000759, 0.00001310, -0.00000377, -0.00000242, 0.00000027, -0.00000093, 0.00000323, 0.00000047, 0.00000272, -0.00000530, -0.00000388, 0.00000639, 0.00000126, -0.00000429, 0.00000112, 0.00000091, 0.00000223, -0.00000245, 0.00000182, -0.00000072, 0.00000273, -0.00000980, 0.00000179, 0.00000286, 0.00000214, -0.00000048, -0.00000254, -0.00000326, -0.00000183, 0.00000594, -0.00000130, -0.00000293, -0.00000155, -0.00000048, 0.00000022, 0.00000194, },
		{ 0.00000045, 0.00000327, -0.00000162, -0.00000279, -0.00000344, 0.00000477, 0.00000195, -0.00000470, -0.00000146, -0.00000189, 0.00000023, 0.00000171, 0.00000555, -0.00000776, -0.00000329, 0.00000239, 0.00000191, 0.00000119, -0.00000060, -0.00000065, -0.00000022, -0.00000071, -0.00000233, 0.00000410, 0.00000315, -0.00000460, -0.00000288, 0.00000279, 0.00000079, 0.00000187, -0.00000031, -0.00000029, -0.00000602, 0.00001066, -0.00000177, -0.00000503, -0.00000094, 0.00000268, 0.00000621, -0.00001242, -0.00000458, 0.00000585, -0.00001032, -0.00000355, -0.00000631, 0.00000252, -0.00000042, -0.00000703, -0.00000460, -0.00000279, -0.00001450, -0.00001114, -0.00001100, -0.00004775, -0.00001906, 0.00004780, 0.00003742, 0.00001821, 0.00000074, 0.00001644, 0.00001502, 0.00002119, 0.00004295, -0.00002072, -0.00011455, -0.00000091, 0.00008327, 0.00002181, 0.00002689, 0.00002698, 0.00001252, 0.00007746, 0.00004980, -0.00008410, -0.00002081, -0.00001907, -0.00001186, -0.00000880, -0.00001101, -0.00000506, -0.00000585, -0.00000230, 0.00000375, -0.00000757, -0.00000351, -0.00000463, 0.00000056, -0.00001105, -0.00000382, 0.00000308, -0.00000058, 0.00000212, -0.00000885, 0.00001138, -0.00000073, -0.00000399, 0.00000095, -0.00000162, 0.00000358, -0.00000000, 0.00000333, -0.00000456, -0.00000489, 0.00000560, 0.00000281, -0.00000462, 0.00000103, 0.00000014, 0.00000283, -0.00000259, 0.00000170, -0.00000058, 0.00000315, -0.00000868, -0.00000069, 0.00000406, 0.00000155, 0.00000049, -0.00000309, -0.00000233, -0.00000329, 0.00000593, -0.00000030, -0.00000307, -0.00000185, -0.00000037, -0.00000002, 0.00000209, },
		{ 0.00000069, 0.00000271, -0.00000080, -0.00000281, -0.00000345, 0.00000314, 0.00000376, -0.00000498, -0.00000126, -0.00000245, 0.00000040, 0.00000092, 0.00000618, -0.00000620, -0.00000493, 0.00000236, 0.00000180, 0.00000140, -0.00000091, 0.00000001, -0.00000052, -0.00000024, -0.00000276, 0.00000333, 0.00000418, -0.00000407, -0.00000407, 0.00000312, 0.00000053, 0.00000223, -0.00000049, 0.00000041, -0.00000628, 0.00000845, 0.00000146, -0.00000626, -0.00000027, 0.00000072, 0.00000773, -0.00001082, -0.00000691, 0.00000582, -0.00000897, -0.00000427, -0.00000597, 0.00000089, 0.00000167, -0.00000804, -0.00000382, -0.00000414, -0.00001241, -0.00001300, -0.00000874, -0.00004397, -0.00003093, 0.00004412, 0.00004446, 0.00001871, 0.00000290, 0.00001648, 0.00001633, 0.00002122, 0.00004685, -0.00001955, -0.00011335, -0.00000389, 0.00007901, 0.00002204, 0.00002520, 0.00002566, 0.00001093, 0.00006516, 0.00006392, -0.00007224, -0.00003172, -0.00001518, -0.00001484, -0.00000721, -0.00001239, -0.00000382, -0.00000690, -0.00000230, 0.00000382, -0.00000675, -0.00000420, -0.00000472, 0.00000012, -0.00000829, -0.00000747, 0.00000494, -0.00000133, 0.00000291, -0.00000886, 0.00000861, 0.00000255, -0.00000530, 0.00000138, -0.00000210, 0.00000372, -0.00000027, 0.00000360, -0.00000356, -0.00000577, 0.00000462, 0.00000417, -0.00000462, 0.00000075, -0.00000036, 0.00000296, -0.00000228, 0.00000120, -0.00000012, 0.00000310, -0.00000701, -0.00000327, 0.00000499, 0.00000117, 0.00000128, -0.00000344, -0.00000161, -0.00000427, 0.00000532, 0.00000098, -0.00000323, -0.00000206, -0.00000046, -0.00000001, 0.00000198, },
		{ 0.00000109, 0.00000208, -0.00000004, -0.00000281, -0.00000326, 0.00000143, 0.00000520, -0.00000478, -0.00000138, -0.00000271, 0.00000025, 0.00000054, 0.00000613, -0.00000414, -0.00000655, 0.00000224, 0.00000161, 0.00000171, -0.00000130, 0.00000059, -0.00000064, 0.00000009, -0.00000292, 0.00000236, 0.00000506, -0.00000329, -0.00000519, 0.00000318, 0.00000051, 0.00000230, -0.00000034, 0.00000060, -0.00000574, 0.00000558, 0.00000465, -0.00000710, 0.00000015, -0.00000074, 0.00000829, -0.00000850, -0.00000915, 0.00000540, -0.00000755, -0.00000488, -0.00000560, -0.00000071, 0.00000338, -0.00000853, -0.00000351, -0.00000494, -0.00001115, -0.00001359, -0.00000847, -0.00003809, -0.00004191, 0.00003756, 0.00005190, 0.00001993, 0.00000468, 0.00001693, 0.00001738, 0.00002146, 0.00005061, -0.00001811, -0.00011254, -0.00000647, 0.00007470, 0.00002224, 0.00002371, 0.00002420, 0.00001004, 0.00005384, 0.00007266, -0.00005641, -0.00004230, -0.00001253, -0.00001684, -0.00000641, -0.00001322, -0.00000306, -0.00000738, -0.00000278, 0.00000406, -0.00000599, -0.00000471, -0.00000484, -0.00000062, -0.00000539, -0.00001049, 0.00000562, -0.00000098, 0.00000273, -0.00000782, 0.00000516, 0.00000565, -0.00000605, 0.00000139, -0.00000228, 0.00000360, -0.00000027, 0.00000352, -0.00000237, -0.00000648, 0.00000352, 0.00000524, -0.00000425, 0.00000028, -0.00000055, 0.00000266, -0.00000162, 0.00000046, 0.00000053, 0.00000274, -0.00000505, -0.00000564, 0.00000542, 0.00000113, 0.00000176, -0.00000349, -0.00000121, -0.00000470, 0.00000415, 0.00000243, -0.00000334, -0.00000219, -0.00000075, 0.00000019, 0.00000169, },
		{ 0.00000151, 0.00000150, 0.00000054, -0.00000272, -0.00000297, -0.00000015, 0.00000610, -0.00000405, -0.00000181, -0.00000264, -0.00000022, 0.00000056, 0.00000551, -0.00000180, -0.00000793, 0.00000191, 0.00000143, 0.00000200, -0.00000161, 0.00000093, -0.00000053, 0.00000024, -0.00000280, 0.00000121, 0.00000576, -0.00000232, -0.00000613, 0.00000291, 0.00000075, 0.00000209, 0.00000014, 0.00000033, -0.00000458, 0.00000242, 0.00000739, -0.00000723, 0.00000003, -0.00000147, 0.00000783, -0.00000561, -0.00001115, 0.00000461, -0.00000615, -0.00000528, -0.00000533, -0.00000211, 0.00000455, -0.00000838, -0.00000377, -0.00000509, -0.00001080, -0.00001297, -0.00000982, -0.00003104, -0.00005085, 0.00002776, 0.00005940, 0.00002207, 0.00000612, 0.00001771, 0.00001830, 0.00002179, 0.00005435, -0.00001646, -0.00011206, -0.00000875, 0.00007039, 0.00002238, 0.00002237, 0.00002271, 0.00000944, 0.00004408, 0.00007639, -0.00003821, -0.00005108, -0.00001190, -0.00001746, -0.00000661, -0.00001336, -0.00000290, -0.00000723, -0.00000373, 0.00000446, -0.00000532, -0.00000502, -0.00000501, -0.00000154, -0.00000262, -0.00001257, 0.00000503, 0.00000033, 0.00000183, -0.00000608, 0.00000151, 0.00000815, -0.00000601, 0.00000090, -0.00000212, 0.00000319, -0.00000000, 0.00000314, -0.00000107, -0.00000698, 0.00000235, 0.00000599, -0.00000353, -0.00000037, -0.00000050, 0.00000205, -0.00000080, -0.00000035, 0.00000121, 0.00000224, -0.00000305, -0.00000752, 0.00000520, 0.00000151, 0.00000189, -0.00000319, -0.00000116, -0.00000461, 0.00000254, 0.00000388, -0.00000330, -0.00000227, -0.00000118, 0.00000050, 0.00000133, },
		{ 0.00000182, 0.00000111, 0.00000087, -0.00000250, -0.00000270, -0.00000147, 0.00000637, -0.00000281, -0.00000251, -0.00000231, -0.00000094, 0.00000086, 0.00000451, 0.00000056, -0.00000886, 0.00000126, 0.00000136, 0.00000216, -0.00000172, 0.00000096, -0.00000014, 0.00000019, -0.00000241, -0.00000005, 0.00000624, -0.00000122, -0.00000681, 0.00000230, 0.00000124, 0.00000165, 0.00000087, -0.00000025, -0.00000306, -0.00000064, 0.00000930, -0.00000642, -0.00000075, -0.00000139, 0.00000647, -0.00000243, -0.00001271, 0.00000344, -0.00000485, -0.00000544, -0.00000519, -0.00000321, 0.00000513, -0.00000762, -0.00000452, -0.00000469, -0.00001121, -0.00001144, -0.00001215, -0.00002394, -0.00005672, 0.00001471, 0.00006644, 0.00002532, 0.00000727, 0.00001875, 0.00001918, 0.00002211, 0.00005814, -0.00001467, -0.00011181, -0.00001083, 0.00006616, 0.00002242, 0.00002117, 0.00002129, 0.00000887, 0.00003611, 0.00007592, -0.00001930, -0.00005685, -0.00001369, -0.00001664, -0.00000774, -0.00001288, -0.00000327, -0.00000655, -0.00000502, 0.00000489, -0.00000471, -0.00000514, -0.00000522, -0.00000256, -0.00000017, -0.00001360, 0.00000329, 0.00000229, 0.00000058, -0.00000406, -0.00000188, 0.00000971, -0.00000507, -0.00000010, -0.00000165, 0.00000253, 0.00000048, 0.00000255, 0.00000022, -0.00000723, 0.00000115, 0.00000639, -0.00000252, -0.00000110, -0.00000028, 0.00000127, 0.00000001, -0.00000106, 0.00000176, 0.00000178, -0.00000125, -0.00000873, 0.00000426, 0.00000230, 0.00000170, -0.00000258, -0.00000142, -0.00000410, 0.00000070, 0.00000514, -0.00000303, -0.00000235, -0.00000167, 0.00000079, 0.00000106, },
		},
		{ { -0.00000211, -0.00000095, -0.00000171, 0.00000240, 0.00000228, 0.00000213, -0.00000076, 0.00000164, 0.00000193, -0.00000203, 0.00001537, -0.00000271, -0.00000309, -0.00000318, 0.00000660, 0.00000069, -0.00000735, -0.00000028, -0.00000144, -0.00000456, -0.00000386, -0.00000243, 0.00000329, 0.00000012, -0.00000607, 0.00000097, -0.00000542, -0.00000135, -0.00000112, -0.00000081, -0.00000693, -0.00000251, -0.00000131, 0.00000157, -0.00001384, 0.00000147, -0.00000132, -0.00000280, -0.00000407, -0.00000528, 0.00000863, -0.00000728, -0.00001134, -0.00000441, -0.00000880, -0.00001265, -0.00006699, 0.00000698, 0.00007375, 0.00002708, 0.00002658, 0.00002206, 0.00002468, 0.00002631, 0.00007347, 0.00000350, -0.00006494, -0.00001574, -0.00001328, -0.00000835, -0.00000743, -0.00000843, 0.00000317, -0.00000579, -0.00000449, -0.00000579, 0.00000317, -0.00000843, -0.00000743, -0.00000835, -0.00001328, -0.00001574, -0.00006494, 0.00000350, 0.00007347, 0.00002631, 0.00002468, 0.00002206, 0.00002658, 0.00002708, 0.00007375, 0.00000698, -0.00006699, -0.00001265, -0.00000880, -0.00000441, -0.00001134, -0.00000728, 0.00000863, -0.00000528, -0.00000407, -0.00000280, -0.00000132, 0.00000147, -0.00001384, 0.00000157, -0.00000131, -0.00000251, -0.00000693, -0.00000081, -0.00000112, -0.00000135, -0.00000542, 0.00000097, -0.00000607, 0.00000012, 0.00000329, -0.00000243, -0.00000386, -0.00000456, -0.00000144, -0.00000028, -0.00000735, 0.00000069, 0.00000660, -0.00000318, -0.00000309, -0.00000271, 0.00001537, -0.00000203, 0.00000193, 0.00000164, -0.00000076, 0.00000213, 0.00000228, 0.00000240, -0.00000171, -0.00000095, },
		{ -0.00000193, -0.00000136, -0.00000130, 0.00000139, 0.00000306, 0.00000136, -0.00000025, 0.00000080, 0.00000424, -0.00000442, 0.00001280, 0.00000319, -0.00000584, -0.00000116, 0.00000319, 0.00000408, -0.00000789, -0.00000133, -0.00000084, -0.00000336, -0.00000530, -0.00000239, 0.00000163, 0.00000285, -0.00000792, 0.00000222, -0.00000464, -0.00000318, -0.00000135, 0.00000006, -0.00000602, -0.00000369, -0.00000159, 0.00000239, -0.00001310, -0.00000086, 0.00000057, -0.00000433, -0.00000234, -0.00000702, 0.00000822, -0.00000394, -0.00001285, -0.00000370, -0.00001010, -0.00000644, -0.00005827, -0.00002804, 0.00007765, 0.00003720, 0.00002644, 0.00002372, 0.00002686, 0.00002278, 0.00007288, 0.00002221, -0.00006865, -0.00002189, -0.00001182, -0.00001080, -0.00000547, -0.00001090, 0.00000261, -0.00000393, -0.00000491, -0.00000624, 0.00000169, -0.00000468, -0.00000965, -0.00000614, -0.00001436, -0.00001135, -0.00005913, -0.00001272, 0.00006987, 0.00003178, 0.00002184, 0.00002163, 0.00002488, 0.00002302, 0.00006087, 0.00003769, -0.00006259, -0.00002530, -0.00000566, -0.00000647, -0.00000877, -0.00001074, 0.00000853, -0.00000337, -0.00000562, -0.00000171, -0.00000260, 0.00000303, -0.00001368, 0.00000004, -0.00000048, -0.00000224, -0.00000634, -0.00000287, -0.00000035, -0.00000082, -0.00000397, -0.00000188, -0.00000354, -0.00000267, 0.00000468, -0.00000262, -0.00000227, -0.00000535, -0.00000206, -0.00000029, -0.00000519, -0.00000365, 0.00000968, -0.00000421, -0.00000146, -0.00000601, 0.00001383, 0.00000319, -0.00000117, 0.00000294, -0.00000208, 0.00000350, 0.00000134, 0.00000334, -0.00000190, -0.00000086, },
		{ -0.00000142, -0.00000188, -0.00000096, 0.00000069, 0.00000325, 0.00000149, -0.00000062, 0.00000050, 0.00000533, -0.00000396, 0.00000746, 0.00000936, -0.00000742, -0.00000031, 0.00000152, 0.00000499, -0.00000614, -0.00000353, -0.00000011, -0.00000233, -0.00000591, -0.00000305, 0.00000038, 0.00000457, -0.00000803, 0.00000110, -0.00000167, -0.00000584, -0.00000122, -0.00000017, -0.00000405, -0.00000518, -0.00000177, 0.00000272, -0.00001155, -0.00000393, 0.00000299, -0.00000622, -0.00000043, -0.00000872, 0.00000768, -0.00000140, -0.00001255, -0.00000510, -0.00000855, -0.00000671, -0.00004137, -0.00005811, 0.00006685, 0.00005368, 0.00002509, 0.00002653, 0.00002801, 0.00002185, 0.00006808, 0.00004235, -0.00006933, -0.00002970, -0.00001051, -0.00001283, -0.00000459, -0.00001147, 0.00000003, -0.00000113, -0.00000578, -0.00000566, -0.00000097, -0.00000088, -0.00001103, -0.00000489, -0.00001459, -0.00000875, -0.00005193, -0.00002600, 0.00006278, 0.00003806, 0.00001912, 0.00002182, 0.00002204, 0.00002266, 0.00004556, 0.00005759, -0.00004499, -0.00004059, -0.00000370, -0.00000801, -0.00000661, -0.00001317, 0.00000721, -0.00000088, -0.00000725, -0.00000089, -0.00000343, 0.00000394, -0.00001269, -0.00000220, 0.00000096, -0.00000288, -0.00000460, -0.00000545, 0.00000046, -0.00000113, -0.00000138, -0.00000472, -0.00000189, -0.00000433, 0.00000495, -0.00000218, -0.00000141, -0.00000507, -0.00000305, -0.00000073, -0.00000276, -0.00000703, 0.00001050, -0.00000271, -0.00000220, -0.00000593, 0.00000863, 0.00000971, -0.00000373, 0.00000396, -0.00000355, 0.00000457, 0.00000102, 0.00000367, -0.00000151, -0.00000124, },
		{ -0.00000069, -0.00000237, -0.00000084, 0.00000039, 0.00000289, 0.00000223, -0.00000131, 0.00000028, 0.00000559, -0.00000179, 0.00000174, 0.00001313, -0.00000588, -0.00000198, 0.00000245, 0.00000334, -0.00000274, -0.00000605, 0.00000029, -0.00000146, -0.00000564, -0.00000428, -0.00000045, 0.00000507, -0.00000638, -0.00000185, 0.00000229, -0.00000802, -0.00000160, -0.00000072, -0.00000205, -0.00000599, -0.00000245, 0.00000292, -0.00000947, -0.00000738, 0.00000547, -0.00000790, 0.00000112, -0.00000989, 0.00000671, 0.00000059, -0.00001099, -0.00000769, -0.00000551, -0.00001019, -0.00002389, -0.00007480, 0.00003959, 0.00007317, 0.00002502, 0.00002944, 0.00002854, 0.00002328, 0.00006007, 0.00006206, -0.00006572, -0.00003923, -0.00000962, -0.00001421, -0.00000490, -0.00001048, -0.00000360, 0.00000146, -0.00000622, -0.00000494, -0.00000357, 0.00000166, -0.00001075, -0.00000493, -0.00001386, -0.00000772, -0.00004404, -0.00003618, 0.00005310, 0.00004409, 0.00001721, 0.00002197, 0.00001908, 0.00002344, 0.00003262, 0.00006475, -0.00001881, -0.00005255, -0.00000614, -0.00000749, -0.00000593, -0.00001381, 0.00000434, 0.00000227, -0.00000893, -0.00000037, -0.00000386, 0.00000428, -0.00001095, -0.00000498, 0.00000272, -0.00000391, -0.00000259, -0.00000750, 0.00000049, -0.00000146, 0.00000096, -0.00000596, -0.00000221, -0.00000426, 0.00000377, -0.00000076, -0.00000166, -0.00000361, -0.00000445, -0.00000117, -0.00000106, -0.00000829, 0.00000841, 0.00000124, -0.00000481, -0.00000355, 0.00000191, 0.00001505, -0.00000412, 0.00000380, -0.00000423, 0.00000442, 0.00000195, 0.00000305, -0.00000041, -0.00000209, },
		{ 0.00000000, -0.00000267, -0.00000095, 0.00000032, 0.00000231, 0.00000297, -0.00000156, -0.00000047, 0.00000566, 0.00000068, -0.00000224, 0.00001296, -0.00000091, -0.00000574, 0.00000499, 0.00000076, 0.00000042, -0.00000733, -0.00000050, -0.00000033, -0.00000507, -0.00000537, -0.00000136, 0.00000483, -0.00000381, -0.00000520, 0.00000538, -0.00000822, -0.00000334, -0.00000079, -0.00000101, -0.00000542, -0.00000396, 0.00000316, -0.00000718, -0.00001063, 0.00000719, -0.00000852, 0.00000150, -0.00000973, 0.00000455, 0.00000286, -0.00000946, -0.00000962, -0.00000350, -0.00001259, -0.00001247, -0.00007426, 0.00000014, 0.00008860, 0.00003003, 0.00003081, 0.00002946, 0.00002606, 0.00005070, 0.00007899, -0.00005650, -0.00005051, -0.00000917, -0.00001520, -0.00000588, -0.00000895, -0.00000684, 0.00000254, -0.00000538, -0.00000484, -0.00000516, 0.00000227, -0.00000876, -0.00000595, -0.00001252, -0.00000774, -0.00003621, -0.00004318, 0.00004162, 0.00004913, 0.00001642, 0.00002168, 0.00001678, 0.00002364, 0.00002421, 0.00006130, 0.00000897, -0.00005596, -0.00001418, -0.00000508, -0.00000653, -0.00001299, 0.00000046, 0.00000548, -0.00001012, -0.00000056, -0.00000367, 0.00000393, -0.00000848, -0.00000805, 0.00000433, -0.00000465, -0.00000121, -0.00000820, -0.00000071, -0.00000126, 0.00000213, -0.00000493, -0.00000448, -0.00000291, 0.00000168, 0.00000118, -0.00000259, -0.00000159, -0.00000568, -0.00000168, -0.00000032, -0.00000753, 0.00000431, 0.00000606, -0.00000746, -0.00000102, -0.00000364, 0.00001696, -0.00000132, 0.00000208, -0.00000365, 0.00000260, 0.00000419, 0.00000171, 0.00000114, -0.00000312, },
		{ 0.00000035, -0.00000266, -0.00000122, 0.00000028, 0.00000187, 0.00000325, -0.00000097, -0.00000191, 0.00000582, 0.00000256, -0.00000357, 0.00000915, 0.00000584, -0.00000941, 0.00000677, -0.00000033, 0.00000128, -0.00000608, -0.00000304, 0.00000141, -0.00000476, -0.00000565, -0.00000282, 0.00000445, -0.00000138, -0.00000749, 0.00000618, -0.00000570, -0.00000655, -0.00000018, -0.00000123, -0.00000350, -0.00000595, 0.00000322, -0.00000482, -0.00001319, 0.00000751, -0.00000745, 0.00000012, -0.00000774, 0.00000079, 0.00000599, -0.00000901, -0.00000947, -0.00000427, -0.00001130, -0.00000972, -0.00005915, -0.00004171, 0.00009123, 0.00004320, 0.00002972, 0.00003155, 0.00002909, 0.00004196, 0.00009102, -0.00004095, -0.00006318, -0.00000924, -0.00001620, -0.00000679, -0.00000801, -0.00000851, 0.00000143, -0.00000311, -0.00000550, -0.00000554, 0.00000114, -0.00000579, -0.00000710, -0.00001122, -0.00000813, -0.00002913, -0.00004702, 0.00002905, 0.00005272, 0.00001688, 0.00002076, 0.00001551, 0.00002252, 0.00002025, 0.00005151, 0.00003200, -0.00004876, -0.00002591, -0.00000290, -0.00000698, -0.00001191, -0.00000322, 0.00000767, -0.00000997, -0.00000201, -0.00000257, 0.00000275, -0.00000535, -0.00001110, 0.00000532, -0.00000458, -0.00000093, -0.00000741, -0.00000291, -0.00000066, 0.00000207, -0.00000211, -0.00000751, -0.00000159, -0.00000015, 0.00000267, -0.00000323, -0.00000009, -0.00000591, -0.00000267, -0.00000014, -0.00000576, -0.00000004, 0.00000965, -0.00000822, -0.00000026, -0.00000616, 0.00001456, 0.00000424, -0.00000057, -0.00000208, -0.00000048, 0.00000697, 0.00000042, 0.00000252, -0.00000387, },
		{ 0.00000009, -0.00000224, -0.00000164, 0.00000021, 0.00000166, 0.00000299, 0.00000031, -0.00000363, 0.00000578, 0.00000384, -0.00000275, 0.00000367, 0.00001151, -0.00001032, 0.00000554, 0.00000183, -0.00000107, -0.00000229, -0.00000681, 0.00000333, -0.00000471, -0.00000509, -0.00000472, 0.00000404, 0.00000037, -0.00000806, 0.00000444, -0.00000108, -0.00001020, 0.00000037, -0.00000209, -0.00000117, -0.00000746, 0.00000251, -0.00000227, -0.00001488, 0.00000624, -0.00000469, -0.00000278, -0.00000426, -0.00000409, 0.00000959, -0.00000951, -0.00000736, -0.00000754, -0.00000677, -0.00001344, -0.00003748, -0.00007404, 0.00007459, 0.00006425, 0.00002740, 0.00003441, 0.00003194, 0.00003512, 0.00009708, -0.00001947, -0.00007604, -0.00001037, -0.00001727, -0.00000725, -0.00000818, -0.00000833, -0.00000155, -0.00000014, -0.00000634, -0.00000521, -0.00000084, -0.00000297, -0.00000748, -0.00001049, -0.00000845, -0.00002321, -0.00004804, 0.00001622, 0.00005439, 0.00001873, 0.00001901, 0.00001546, 0.00002001, 0.00001959, 0.00003960, 0.00004674, -0.00003283, -0.00003696, -0.00000385, -0.00000573, -0.00001175, -0.00000564, 0.00000802, -0.00000797, -0.00000490, -0.00000064, 0.00000094, -0.00000193, -0.00001361, 0.00000527, -0.00000348, -0.00000171, -0.00000559, -0.00000535, -0.00000024, 0.00000136, 0.00000127, -0.00000966, -0.00000174, -0.00000067, 0.00000286, -0.00000279, 0.00000011, -0.00000469, -0.00000430, -0.00000008, -0.00000408, -0.00000308, 0.00001065, -0.00000626, -0.00000183, -0.00000547, 0.00000886, 0.00001071, -0.00000261, -0.00000052, -0.00000374, 0.00000907, 0.00000017, 0.00000306, -0.00000387, },
		{ -0.00000079, -0.00000146, -0.00000218, 0.00000013, 0.00000155, 0.00000250, 0.00000170, -0.00000484, 0.00000495, 0.00000502, -0.00000108, -0.00000104, 0.00001357, -0.00000690, 0.00000065, 0.00000700, -0.00000550, 0.00000232, -0.00001009, 0.00000421, -0.00000415, -0.00000445, -0.00000625, 0.00000312, 0.00000164, -0.00000729, 0.00000123, 0.00000386, -0.00001248, -0.00000037, -0.00000244, 0.00000027, -0.00000741, 0.00000047, 0.00000061, -0.00001577, 0.00000370, -0.00000097, -0.00000623, -0.00000048, -0.00000873, 0.00001235, -0.00000973, -0.00000487, -0.00001115, -0.00000194, -0.00001841, -0.00001880, -0.00008779, 0.00003833, 0.00008776, 0.00002774, 0.00003628, 0.00003521, 0.00003040, 0.00009748, 0.00000615, -0.00008702, -0.00001356, -0.00001803, -0.00000755, -0.00000913, -0.00000703, -0.00000510, 0.00000218, -0.00000638, -0.00000500, -0.00000261, -0.00000131, -0.00000664, -0.00001035, -0.00000868, -0.00001837, -0.00004692, 0.00000410, 0.00005359, 0.00002218, 0.00001623, 0.00001667, 0.00001638, 0.00002081, 0.00002855, 0.00005280, -0.00001275, -0.00004267, -0.00000968, -0.00000248, -0.00001266, -0.00000668, 0.00000659, -0.00000445, -0.00000860, 0.00000139, -0.00000081, 0.00000109, -0.00001497, 0.00000385, -0.00000140, -0.00000312, -0.00000354, -0.00000714, -0.00000060, 0.00000068, 0.00000398, -0.00000967, -0.00000398, 0.00000031, 0.00000173, -0.00000126, -0.00000090, -0.00000241, -0.00000612, -0.00000013, -0.00000300, -0.00000429, 0.00000910, -0.00000230, -0.00000472, -0.00000302, 0.00000220, 0.00001556, -0.00000245, 0.00000005, -0.00000606, 0.00000936, 0.00000163, 0.00000248, -0.00000302, },
		{ -0.00000198, -0.00000055, -0.00000275, 0.00000012, 0.00000134, 0.00000216, 0.00000259, -0.00000493, 0.00000296, 0.00000650, 0.00000028, -0.00000330, 0.00001118, 0.00000025, -0.00000625, 0.00001283, -0.00000918, 0.00000500, -0.00001072, 0.00000266, -0.00000217, -0.00000470, -0.00000645, 0.00000110, 0.00000296, -0.00000615, -0.00000186, 0.00000713, -0.00001186, -0.00000328, -0.00000152, -0.00000003, -0.00000531, -0.00000288, 0.00000363, -0.00001589, 0.00000045, 0.00000270, -0.00000895, 0.00000219, -0.00001161, 0.00001283, -0.00000818, -0.00000391, -0.00001255, -0.00000008, -0.00001969, -0.00000972, -0.00008085, -0.00001005, 0.00010391, 0.00003582, 0.00003507, 0.00003987, 0.00002731, 0.00009351, 0.00003329, -0.00009368, -0.00001999, -0.00001779, -0.00000840, -0.00000995, -0.00000592, -0.00000764, 0.00000257, -0.00000495, -0.00000546, -0.00000353, -0.00000112, -0.00000483, -0.00001032, -0.00000916, -0.00001425, -0.00004456, -0.00000634, 0.00004990, 0.00002729, 0.00001249, 0.00001884, 0.00001224, 0.00002258, 0.00001995, 0.00005194, 0.00000655, -0.00004043, -0.00001949, 0.00000112, -0.00001339, -0.00000731, 0.00000442, -0.00000062, -0.00001179, 0.00000231, -0.00000154, 0.00000287, -0.00001459, 0.00000094, 0.00000136, -0.00000462, -0.00000193, -0.00000780, -0.00000185, 0.00000028, 0.00000543, -0.00000736, -0.00000773, 0.00000189, 0.00000019, 0.00000051, -0.00000218, -0.00000019, -0.00000717, -0.00000072, -0.00000234, -0.00000421, 0.00000626, 0.00000189, -0.00000713, -0.00000084, -0.00000295, 0.00001686, 0.00000065, -0.00000067, -0.00000688, 0.00000747, 0.00000463, 0.00000118, -0.00000168, },
		{ -0.00000300, 0.00000018, -0.00000320, 0.00000016, 0.00000095, 0.00000213, 0.00000275, -0.00000381, 0.00000004, 0.00000814, 0.00000098, -0.00000293, 0.00000555, 0.00000845, -0.00001191, 0.00001598, -0.00000909, 0.00000352, -0.00000746, -0.00000173, 0.00000143, -0.00000620, -0.00000502, -0.00000207, 0.00000451, -0.00000536, -0.00000362, 0.00000759, -0.00000800, -0.00000800, 0.00000039, -0.00000184, -0.00000181, -0.00000661, 0.00000603, -0.00001502, -0.00000316, 0.00000557, -0.00001013, 0.00000295, -0.00001196, 0.00001042, -0.00000427, -0.00000541, -0.00001059, -0.00000251, -0.00001552, -0.00001108, -0.00005905, -0.00005724, 0.00010209, 0.00005482, 0.00003038, 0.00004593, 0.00002567, 0.00008652, 0.00005946, -0.00009410, -0.00003026, -0.00001619, -0.00001044, -0.00000977, -0.00000613, -0.00000814, 0.00000061, -0.00000217, -0.00000644, -0.00000368, -0.00000197, -0.00000287, -0.00000964, -0.00001030, -0.00001050, -0.00004168, -0.00001446, 0.00004335, 0.00003361, 0.00000845, 0.00002108, 0.00000861, 0.00002360, 0.00001457, 0.00004646, 0.00002176, -0.00003088, -0.00002980, 0.00000214, -0.00001212, -0.00000892, 0.00000284, 0.00000211, -0.00001311, 0.00000109, -0.00000056, 0.00000293, -0.00001231, -0.00000316, 0.00000428, -0.00000570, -0.00000110, -0.00000739, -0.00000356, -0.00000012, 0.00000577, -0.00000357, -0.00001147, 0.00000255, -0.00000047, 0.00000136, -0.00000259, 0.00000081, -0.00000667, -0.00000221, -0.00000169, -0.00000377, 0.00000367, 0.00000465, -0.00000766, -0.00000030, -0.00000524, 0.00001423, 0.00000593, -0.00000193, -0.00000657, 0.00000399, 0.00000816, 0.00000011, -0.00000052, },
		{ -0.00000344, 0.00000048, -0.00000338, 0.00000013, 0.00000050, 0.00000228, 0.00000241, -0.00000201, -0.00000297, 0.00000922, 0.00000149, -0.00000106, -0.00000069, 0.00001426, -0.00001320, 0.00001402, -0.00000381, -0.00000232, -0.00000110, -0.00000775, 0.00000553, -0.00000816, -0.00000277, -0.00000545, 0.00000571, -0.00000486, -0.00000397, 0.00000563, -0.00000217, -0.00001287, 0.00000192, -0.00000386, 0.00000151, -0.00000918, 0.00000672, -0.00001268, -0.00000709, 0.00000749, -0.00000980, 0.00000203, -0.00001020, 0.00000587, 0.00000113, -0.00000856, -0.00000626, -0.00000778, -0.00000803, -0.00001826, -0.00003339, -0.00008924, 0.00007611, 0.00008269, 0.00002519, 0.00005150, 0.00002640, 0.00007713, 0.00008308, -0.00008751, -0.00004393, -0.00001359, -0.00001364, -0.00000842, -0.00000788, -0.00000670, -0.00000299, 0.00000092, -0.00000712, -0.00000371, -0.00000299, -0.00000172, -0.00000775, -0.00001216, -0.00000709, -0.00003859, -0.00002026, 0.00003469, 0.00003995, 0.00000540, 0.00002204, 0.00000679, 0.00002266, 0.00001269, 0.00003827, 0.00003194, -0.00001733, -0.00003616, -0.00000187, -0.00000784, -0.00001212, 0.00000254, 0.00000302, -0.00001199, -0.00000243, 0.00000195, 0.00000159, -0.00000871, -0.00000766, 0.00000660, -0.00000592, -0.00000108, -0.00000634, -0.00000507, -0.00000095, 0.00000552, 0.00000047, -0.00001362, 0.00000109, 0.00000051, 0.00000072, -0.00000158, 0.00000019, -0.00000462, -0.00000435, -0.00000099, -0.00000360, 0.00000222, 0.00000536, -0.00000619, -0.00000141, -0.00000494, 0.00000899, 0.00001142, -0.00000229, -0.00000603, 0.00000009, 0.00001085, 0.00000030, -0.00000017, },
		{ -0.00000332, 0.00000035, -0.00000327, -0.00000008, 0.00000023, 0.00000228, 0.00000204, -0.00000030, -0.00000514, 0.00000900, 0.00000252, 0.00000078, -0.00000499, 0.00001525, -0.00000886, 0.00000700, 0.00000534, -0.00001016, 0.00000556, -0.00001273, 0.00000806, -0.00000897, -0.00000128, -0.00000753, 0.00000554, -0.00000398, -0.00000376, 0.00000277, 0.00000335, -0.00001562, 0.00000139, -0.00000451, 0.00000299, -0.00000919, 0.00000491, -0.00000861, -0.00001138, 0.00000871, -0.00000861, 0.00000039, -0.00000763, 0.00000084, 0.00000616, -0.00001137, -0.00000203, -0.00001262, -0.00000149, -0.00002446, -0.00001469, -0.00009751, 0.00002847, 0.00011070, 0.00002572, 0.00005315, 0.00003129, 0.00006523, 0.00010354, -0.00007438, -0.00005955, -0.00001113, -0.00001729, -0.00000653, -0.00001045, -0.00000443, -0.00000665, 0.00000283, -0.00000656, -0.00000431, -0.00000337, -0.00000197, -0.00000469, -0.00001428, -0.00000443, -0.00003513, -0.00002431, 0.00002523, 0.00004470, 0.00000488, 0.00002040, 0.00000776, 0.00001911, 0.00001411, 0.00002874, 0.00003777, -0.00000384, -0.00003545, -0.00001109, -0.00000155, -0.00001600, 0.00000291, 0.00000258, -0.00000909, -0.00000722, 0.00000479, -0.00000002, -0.00000495, -0.00001138, 0.00000752, -0.00000499, -0.00000174, -0.00000513, -0.00000588, -0.00000238, 0.00000499, 0.00000382, -0.00001333, -0.00000258, 0.00000278, -0.00000098, 0.00000034, -0.00000133, -0.00000205, -0.00000617, -0.00000074, -0.00000361, 0.00000175, 0.00000470, -0.00000384, -0.00000300, -0.00000351, 0.00000337, 0.00001490, -0.00000051, -0.00000602, -0.00000314, 0.00001163, 0.00000223, -0.00000069, },
		{ -0.00000308, 0.00000007, -0.00000298, -0.00000051, 0.00000029, 0.00000190, 0.00000202, 0.00000075, -0.00000588, 0.00000718, 0.00000440, 0.00000159, -0.00000604, 0.00001119, -0.00000033, -0.00000235, 0.00001471, -0.00001603, 0.00000897, -0.00001396, 0.00000714, -0.00000721, -0.00000194, -0.00000713, 0.00000329, -0.00000212, -0.00000402, 0.00000071, 0.00000650, -0.00001469, -0.00000207, -0.00000306, 0.00000186, -0.00000634, 0.00000082, -0.00000328, -0.00001569, 0.00000935, -0.00000722, -0.00000098, -0.00000545, -0.00000315, 0.00000918, -0.00001207, -0.00000015, -0.00001420, 0.00000058, -0.00002478, -0.00000856, -0.00008288, -0.00002902, 0.00012559, 0.00003875, 0.00004797, 0.00004153, 0.00005104, 0.00012044, -0.00005563, -0.00007527, -0.00001017, -0.00002039, -0.00000513, -0.00001264, -0.00000284, -0.00000880, 0.00000247, -0.00000433, -0.00000569, -0.00000292, -0.00000343, -0.00000120, -0.00001573, -0.00000317, -0.00003091, -0.00002743, 0.00001639, 0.00004641, 0.00000792, 0.00001562, 0.00001155, 0.00001329, 0.00001778, 0.00001923, 0.00004027, 0.00000682, -0.00002756, -0.00002280, 0.00000384, -0.00001845, 0.00000240, 0.00000213, -0.00000589, -0.00001154, 0.00000632, -0.00000052, -0.00000231, -0.00001328, 0.00000649, -0.00000285, -0.00000285, -0.00000405, -0.00000599, -0.00000407, 0.00000409, 0.00000621, -0.00001085, -0.00000739, 0.00000499, -0.00000243, 0.00000188, -0.00000236, -0.00000042, -0.00000652, -0.00000154, -0.00000335, 0.00000149, 0.00000396, -0.00000209, -0.00000376, -0.00000245, -0.00000072, 0.00001503, 0.00000358, -0.00000652, -0.00000532, 0.00001032, 0.00000548, -0.00000148, },
		{ -0.00000334, -0.00000000, -0.00000269, -0.00000111, 0.00000068, 0.00000114, 0.00000237, 0.00000108, -0.00000531, 0.00000418, 0.00000676, 0.00000138, -0.00000423, 0.00000403, 0.00000890, -0.00000991, 0.00002015, -0.00001643, 0.00000681, -0.00001029, 0.00000244, -0.00000282, -0.00000485, -0.00000434, -0.00000064, 0.00000052, -0.00000510, 0.00000025, 0.00000657, -0.00001019, -0.00000768, -0.00000024, -0.00000112, -0.00000178, -0.00000411, 0.00000193, -0.00001906, 0.00000902, -0.00000567, -0.00000192, -0.00000399, -0.00000564, 0.00000981, -0.00001027, -0.00000120, -0.00001188, -0.00000230, -0.00001888, -0.00001333, -0.00005510, -0.00007913, 0.00011505, 0.00006714, 0.00003652, 0.00005590, 0.00003636, 0.00013255, -0.00003173, -0.00008980, -0.00001150, -0.00002233, -0.00000498, -0.00001356, -0.00000290, -0.00000872, -0.00000018, -0.00000102, -0.00000724, -0.00000216, -0.00000525, 0.00000156, -0.00001554, -0.00000379, -0.00002569, -0.00003023, 0.00000919, 0.00004440, 0.00001444, 0.00000849, 0.00001692, 0.00000674, 0.00002183, 0.00001141, 0.00003983, 0.00001408, -0.00001531, -0.00003241, 0.00000493, -0.00001743, -0.00000033, 0.00000289, -0.00000379, -0.00001390, 0.00000533, 0.00000092, -0.00000144, -0.00001298, 0.00000360, 0.00000011, -0.00000408, -0.00000318, -0.00000573, -0.00000537, 0.00000247, 0.00000786, -0.00000716, -0.00001162, 0.00000552, -0.00000230, 0.00000185, -0.00000175, -0.00000071, -0.00000491, -0.00000348, -0.00000265, 0.00000079, 0.00000409, -0.00000178, -0.00000319, -0.00000234, -0.00000264, 0.00001209, 0.00000878, -0.00000657, -0.00000689, 0.00000763, 0.00000886, -0.00000151, },
		{ -0.00000442, 0.00000023, -0.00000244, -0.00000182, 0.00000125, 0.00000025, 0.00000280, 0.00000103, -0.00000405, 0.00000093, 0.00000873, 0.00000091, -0.00000114, -0.00000315, 0.00001504, -0.00001224, 0.00001919, -0.00001032, -0.00000047, -0.00000316, -0.00000423, 0.00000254, -0.00000858, -0.00000072, -0.00000463, 0.00000278, -0.00000636, 0.00000090, 0.00000452, -0.00000395, -0.00001322, 0.00000203, -0.00000407, 0.00000234, -0.00000774, 0.00000527, -0.00002023, 0.00000697, -0.00000338, -0.00000307, -0.00000250, -0.00000741, 0.00000907, -0.00000723, -0.00000377, -0.00000756, -0.00000743, -0.00001062, -0.00002193, -0.00002800, -0.00010644, 0.00007429, 0.00010590, 0.00002448, 0.00006989, 0.00002509, 0.00013738, -0.00000231, -0.00010265, -0.00001509, -0.00002324, -0.00000607, -0.00001318, -0.00000452, -0.00000693, -0.00000392, 0.00000193, -0.00000782, -0.00000202, -0.00000636, 0.00000253, -0.00001316, -0.00000624, -0.00001970, -0.00003286, 0.00000387, 0.00003909, 0.00002310, 0.00000097, 0.00002164, 0.00000167, 0.00002403, 0.00000691, 0.00003632, 0.00001904, -0.00000289, -0.00003576, -0.00000008, -0.00001259, -0.00000529, 0.00000497, -0.00000312, -0.00001390, 0.00000181, 0.00000393, -0.00000192, -0.00001106, -0.00000032, 0.00000311, -0.00000498, -0.00000256, -0.00000548, -0.00000580, -0.00000003, 0.00000901, -0.00000332, -0.00001389, 0.00000348, -0.00000019, 0.00000002, 0.00000058, -0.00000270, -0.00000205, -0.00000570, -0.00000198, -0.00000028, 0.00000508, -0.00000249, -0.00000201, -0.00000259, -0.00000308, 0.00000767, 0.00001306, -0.00000488, -0.00000856, 0.00000467, 0.00001106, 0.00000006, },
		{ -0.00000597, 0.00000042, -0.00000202, -0.00000267, 0.00000182, -0.00000044, 0.00000292, 0.00000110, -0.00000282, -0.00000165, 0.00000952, 0.00000103, 0.00000154, -0.00000775, 0.00001581, -0.00000842, 0.00001244, 0.00000012, -0.00000955, 0.00000387, -0.00000966, 0.00000623, -0.00001086, 0.00000151, -0.00000668, 0.00000323, -0.00000673, 0.00000137, 0.00000226, 0.00000140, -0.00001613, 0.00000172, -0.00000510, 0.00000402, -0.00000836, 0.00000549, -0.00001838, 0.00000261, 0.00000029, -0.00000535, 0.00000004, -0.00000965, 0.00000851, -0.00000480, -0.00000562, -0.00000418, -0.00001088, -0.00000516, -0.00002658, -0.00001270, -0.00010432, 0.00001022, 0.00014164, 0.00002158, 0.00007711, 0.00002203, 0.00013225, 0.00003311, -0.00011368, -0.00002045, -0.00002375, -0.00000784, -0.00001222, -0.00000675, -0.00000487, -0.00000694, 0.00000290, -0.00000643, -0.00000318, -0.00000604, 0.00000133, -0.00000885, -0.00000982, -0.00001368, -0.00003493, -0.00000003, 0.00003176, 0.00003177, -0.00000452, 0.00002348, -0.00000007, 0.00002287, 0.00000653, 0.00002988, 0.00002304, 0.00000642, -0.00003129, -0.00001002, -0.00000593, -0.00001073, 0.00000709, -0.00000303, -0.00001242, -0.00000296, 0.00000709, -0.00000249, -0.00000877, -0.00000399, 0.00000517, -0.00000507, -0.00000226, -0.00000541, -0.00000532, -0.00000303, 0.00000951, 0.00000014, -0.00001381, -0.00000071, 0.00000302, -0.00000259, 0.00000347, -0.00000491, 0.00000037, -0.00000673, -0.00000229, -0.00000107, 0.00000605, -0.00000295, -0.00000162, -0.00000207, -0.00000331, 0.00000372, 0.00001462, -0.00000074, -0.00001064, 0.00000216, 0.00001135, 0.00000330, },
		{ -0.00000697, -0.00000012, -0.00000106, -0.00000375, 0.00000229, -0.00000072, 0.00000252, 0.00000153, -0.00000202, -0.00000316, 0.00000892, 0.00000206, 0.00000289, -0.00000877, 0.00001146, -0.00000046, 0.00000333, 0.00001043, -0.00001586, 0.00000692, -0.00001097, 0.00000622, -0.00001009, 0.00000084, -0.00000563, 0.00000126, -0.00000562, 0.00000064, 0.00000142, 0.00000395, -0.00001499, -0.00000187, -0.00000356, 0.00000265, -0.00000575, 0.00000276, -0.00001386, -0.00000371, 0.00000526, -0.00000895, 0.00000394, -0.00001278, 0.00000885, -0.00000399, -0.00000543, -0.00000361, -0.00001026, -0.00000541, -0.00002345, -0.00001256, -0.00007808, -0.00005957, 0.00015667, 0.00003706, 0.00007285, 0.00003017, 0.00011621, 0.00007353, -0.00012181, -0.00002761, -0.00002419, -0.00000978, -0.00001146, -0.00000848, -0.00000393, -0.00000786, 0.00000111, -0.00000301, -0.00000550, -0.00000444, -0.00000155, -0.00000367, -0.00001328, -0.00000862, -0.00003582, -0.00000332, 0.00002401, 0.00003833, -0.00000612, 0.00002125, 0.00000206, 0.00001829, 0.00000961, 0.00002166, 0.00002639, 0.00001169, -0.00002080, -0.00002096, -0.00000102, -0.00001411, 0.00000744, -0.00000208, -0.00001095, -0.00000722, 0.00000872, -0.00000178, -0.00000734, -0.00000633, 0.00000562, -0.00000409, -0.00000236, -0.00000540, -0.00000442, -0.00000571, 0.00000887, 0.00000328, -0.00001198, -0.00000557, 0.00000558, -0.00000431, 0.00000518, -0.00000551, 0.00000060, -0.00000534, -0.00000421, -0.00000098, 0.00000606, -0.00000195, -0.00000306, -0.00000018, -0.00000410, 0.00000140, 0.00001295, 0.00000526, -0.00001250, 0.00000005, 0.00000997, 0.00000731, },
		{ -0.00000625, -0.00000199, 0.00000062, -0.00000501, 0.00000252, -0.00000054, 0.00000168, 0.00000218, -0.00000156, -0.00000379, 0.00000736, 0.00000363, 0.00000303, -0.00000688, 0.00000431, 0.00000781, -0.00000368, 0.00001627, -0.00001601, 0.00000410, -0.00000749, 0.00000253, -0.00000646, -0.00000248, -0.00000208, -0.00000226, -0.00000361, -0.00000124, 0.00000231, 0.00000351, -0.00001041, -0.00000753, -0.00000061, -0.00000060, -0.00000140, -0.00000126, -0.00000829, -0.00001054, 0.00001034, -0.00001285, 0.00000810, -0.00001573, 0.00000923, -0.00000403, -0.00000387, -0.00000532, -0.00000628, -0.00001011, -0.00001457, -0.00002230, -0.00004243, -0.00011327, 0.00013672, 0.00007364, 0.00005794, 0.00004807, 0.00009204, 0.00011518, -0.00012406, -0.00003789, -0.00002390, -0.00001215, -0.00001102, -0.00000938, -0.00000453, -0.00000659, -0.00000281, 0.00000126, -0.00000786, -0.00000255, -0.00000482, 0.00000088, -0.00001533, -0.00000536, -0.00003506, -0.00000670, 0.00001715, 0.00004145, -0.00000328, 0.00001538, 0.00000693, 0.00001185, 0.00001411, 0.00001379, 0.00002816, 0.00001422, -0.00000831, -0.00002814, -0.00000097, -0.00001370, 0.00000500, 0.00000052, -0.00001045, -0.00000978, 0.00000787, 0.00000070, -0.00000714, -0.00000709, 0.00000452, -0.00000229, -0.00000275, -0.00000523, -0.00000374, -0.00000721, 0.00000664, 0.00000633, -0.00000941, -0.00000937, 0.00000589, -0.00000383, 0.00000453, -0.00000344, -0.00000200, -0.00000159, -0.00000733, -0.00000021, 0.00000490, 0.00000075, -0.00000609, 0.00000249, -0.00000508, 0.00000053, 0.00000906, 0.00001132, -0.00001279, -0.00000236, 0.00000794, 0.00001064, },
		{ -0.00000318, -0.00000517, 0.00000277, -0.00000617, 0.00000236, 0.00000007, 0.00000065, 0.00000271, -0.00000106, -0.00000407, 0.00000556, 0.00000504, 0.00000270, -0.00000374, -0.00000255, 0.00001276, -0.00000548, 0.00001566, -0.00000973, -0.00000322, -0.00000145, -0.00000248, -0.00000212, -0.00000644, 0.00000179, -0.00000520, -0.00000225, -0.00000322, 0.00000392, 0.00000158, -0.00000472, -0.00001264, 0.00000141, -0.00000344, 0.00000220, -0.00000416, -0.00000384, -0.00001588, 0.00001366, -0.00001513, 0.00001042, -0.00001637, 0.00000764, -0.00000287, -0.00000313, -0.00000678, -0.00000222, -0.00001483, -0.00000606, -0.00003194, -0.00001429, -0.00013421, 0.00007889, 0.00012317, 0.00004056, 0.00006919, 0.00006639, 0.00015152, -0.00011580, -0.00005367, -0.00002144, -0.00001598, -0.00001016, -0.00001012, -0.00000585, -0.00000444, -0.00000692, 0.00000433, -0.00000864, -0.00000170, -0.00000705, 0.00000344, -0.00001512, -0.00000424, -0.00003258, -0.00001049, 0.00001186, 0.00004086, 0.00000319, 0.00000764, 0.00001231, 0.00000591, 0.00001754, 0.00000847, 0.00002717, 0.00001607, 0.00000214, -0.00002848, -0.00000645, -0.00001005, 0.00000054, 0.00000410, -0.00001060, -0.00001077, 0.00000500, 0.00000419, -0.00000746, -0.00000698, 0.00000268, -0.00000039, -0.00000306, -0.00000488, -0.00000365, -0.00000716, 0.00000297, 0.00000919, -0.00000682, -0.00001108, 0.00000347, -0.00000115, 0.00000169, 0.00000080, -0.00000638, 0.00000296, -0.00001009, 0.00000016, 0.00000334, 0.00000416, -0.00000924, 0.00000431, -0.00000505, -0.00000006, 0.00000489, 0.00001522, -0.00001020, -0.00000575, 0.00000628, 0.00001205, },
		{ 0.00000181, -0.00000880, 0.00000455, -0.00000664, 0.00000153, 0.00000104, -0.00000030, 0.00000282, -0.00000024, -0.00000438, 0.00000403, 0.00000585, 0.00000252, -0.00000084, -0.00000689, 0.00001271, -0.00000185, 0.00001005, -0.00000017, -0.00001092, 0.00000312, -0.00000535, 0.00000011, -0.00000854, 0.00000349, -0.00000546, -0.00000297, -0.00000407, 0.00000472, 0.00000028, -0.00000062, -0.00001460, 0.00000046, -0.00000394, 0.00000309, -0.00000421, -0.00000202, -0.00001826, 0.00001372, -0.00001410, 0.00000898, -0.00001284, 0.00000236, 0.00000131, -0.00000527, -0.00000545, -0.00000153, -0.00001499, -0.00000383, -0.00003297, -0.00000463, -0.00011845, -0.00000467, 0.00016721, 0.00003378, 0.00008439, 0.00004748, 0.00017543, -0.00009257, -0.00007685, -0.00001563, -0.00002222, -0.00000795, -0.00001184, -0.00000642, -0.00000334, -0.00000900, 0.00000431, -0.00000664, -0.00000276, -0.00000735, 0.00000342, -0.00001260, -0.00000493, -0.00002878, -0.00001462, 0.00000819, 0.00003718, 0.00001155, 0.00000032, 0.00001602, 0.00000234, 0.00001828, 0.00000673, 0.00002320, 0.00001831, 0.00000868, -0.00002222, -0.00001490, -0.00000605, -0.00000354, 0.00000681, -0.00001005, -0.00001140, 0.00000166, 0.00000700, -0.00000689, -0.00000721, 0.00000125, 0.00000076, -0.00000286, -0.00000448, -0.00000412, -0.00000592, -0.00000130, 0.00001121, -0.00000427, -0.00001080, -0.00000077, 0.00000238, -0.00000179, 0.00000537, -0.00001025, 0.00000586, -0.00001046, -0.00000122, 0.00000254, 0.00000674, -0.00001060, 0.00000357, -0.00000291, -0.00000146, 0.00000217, 0.00001554, -0.00000447, -0.00001000, 0.00000538, 0.00001122, },
		{ 0.00000729, -0.00001133, 0.00000486, -0.00000579, -0.00000020, 0.00000231, -0.00000101, 0.00000244, 0.00000084, -0.00000471, 0.00000279, 0.00000611, 0.00000261, 0.00000119, -0.00000816, 0.00000849, 0.00000459, 0.00000341, 0.00000783, -0.00001434, 0.00000264, -0.00000372, -0.00000134, -0.00000745, 0.00000186, -0.00000237, -0.00000593, -0.00000353, 0.00000394, 0.00000089, 0.00000040, -0.00001248, -0.00000376, -0.00000186, 0.00000104, -0.00000146, -0.00000270, -0.00001767, 0.00001031, -0.00000941, 0.00000341, -0.00000501, -0.00000649, 0.00000843, -0.00001041, -0.00000079, -0.00000532, -0.00000911, -0.00000961, -0.00002331, -0.00001363, -0.00007680, -0.00008986, 0.00018348, 0.00004912, 0.00008700, 0.00004116, 0.00018248, -0.00005280, -0.00010678, -0.00000726, -0.00003052, -0.00000429, -0.00001509, -0.00000528, -0.00000459, -0.00000783, 0.00000063, -0.00000201, -0.00000546, -0.00000591, 0.00000124, -0.00000861, -0.00000657, -0.00002431, -0.00001874, 0.00000581, 0.00003147, 0.00002000, -0.00000482, 0.00001691, 0.00000161, 0.00001636, 0.00000782, 0.00001765, 0.00002028, 0.00001212, -0.00001268, -0.00002173, -0.00000536, -0.00000470, 0.00000683, -0.00000741, -0.00001300, -0.00000058, 0.00000766, -0.00000439, -0.00000859, 0.00000092, 0.00000076, -0.00000201, -0.00000417, -0.00000478, -0.00000433, -0.00000493, 0.00001153, -0.00000136, -0.00000945, -0.00000504, 0.00000474, -0.00000392, 0.00000810, -0.00001134, 0.00000514, -0.00000727, -0.00000490, 0.00000312, 0.00000745, -0.00000905, -0.00000027, 0.00000129, -0.00000377, 0.00000139, 0.00001243, 0.00000318, -0.00001391, 0.00000466, 0.00000892, },
		{ 0.00001143, -0.00001134, 0.00000295, -0.00000336, -0.00000275, 0.00000373, -0.00000145, 0.00000178, 0.00000184, -0.00000467, 0.00000141, 0.00000634, 0.00000260, 0.00000253, -0.00000728, 0.00000268, 0.00001000, 0.00000000, 0.00001039, -0.00001096, -0.00000361, 0.00000194, -0.00000551, -0.00000418, -0.00000207, 0.00000263, -0.00000949, -0.00000271, 0.00000211, 0.00000311, -0.00000112, -0.00000757, -0.00000942, 0.00000108, -0.00000225, 0.00000221, -0.00000412, -0.00001557, 0.00000480, -0.00000245, -0.00000465, 0.00000496, -0.00001640, 0.00001598, -0.00001614, 0.00000494, -0.00001133, 0.00000011, -0.00001974, -0.00000813, -0.00003166, -0.00002960, -0.00015001, 0.00015603, 0.00008895, 0.00007772, 0.00004735, 0.00017359, 0.00000019, -0.00013888, 0.00000009, -0.00003848, -0.00000071, -0.00001900, -0.00000288, -0.00000786, -0.00000405, -0.00000538, 0.00000354, -0.00000833, -0.00000389, -0.00000180, -0.00000448, -0.00000809, -0.00001988, -0.00002249, 0.00000429, 0.00002476, 0.00002721, -0.00000708, 0.00001531, 0.00000269, 0.00001333, 0.00000961, 0.00001295, 0.00002020, 0.00001482, -0.00000418, -0.00002285, -0.00001008, -0.00000215, 0.00000384, -0.00000260, -0.00001580, -0.00000123, 0.00000590, -0.00000014, -0.00001082, 0.00000137, 0.00000000, -0.00000088, -0.00000386, -0.00000530, -0.00000317, -0.00000694, 0.00000966, 0.00000214, -0.00000803, -0.00000773, 0.00000449, -0.00000352, 0.00000781, -0.00000867, 0.00000051, -0.00000110, -0.00000993, 0.00000446, 0.00000649, -0.00000502, -0.00000609, 0.00000600, -0.00000585, 0.00000167, 0.00000753, 0.00001042, -0.00001571, 0.00000304, 0.00000654, },
		{ 0.00001298, -0.00000824, -0.00000103, 0.00000016, -0.00000561, 0.00000493, -0.00000156, 0.00000110, 0.00000232, -0.00000386, -0.00000049, 0.00000695, 0.00000214, 0.00000368, -0.00000567, -0.00000206, 0.00001144, 0.00000195, 0.00000680, -0.00000203, -0.00001276, 0.00000814, -0.00000910, -0.00000157, -0.00000559, 0.00000662, -0.00001100, -0.00000351, 0.00000070, 0.00000538, -0.00000309, -0.00000269, -0.00001350, 0.00000221, -0.00000422, 0.00000422, -0.00000400, -0.00001396, -0.00000073, 0.00000438, -0.00001227, 0.00001356, -0.00002358, 0.00002019, -0.00001873, 0.00000793, -0.00001534, 0.00000740, -0.00002743, 0.00000366, -0.00004536, 0.00000291, -0.00016735, 0.00008420, 0.00014205, 0.00006613, 0.00005964, 0.00015512, 0.00005853, -0.00016528, 0.00000074, -0.00004229, 0.00000008, -0.00002149, -0.00000109, -0.00001126, 0.00000004, -0.00001097, 0.00000738, -0.00000938, -0.00000284, -0.00000419, -0.00000154, -0.00000870, -0.00001593, -0.00002561, 0.00000324, 0.00001788, 0.00003250, -0.00000678, 0.00001249, 0.00000382, 0.00001111, 0.00000996, 0.00001110, 0.00001680, 0.00001862, 0.00000055, -0.00001716, -0.00001903, 0.00000217, -0.00000029, 0.00000274, -0.00001851, -0.00000131, 0.00000301, 0.00000430, -0.00001246, 0.00000139, -0.00000043, -0.00000027, -0.00000319, -0.00000563, -0.00000281, -0.00000713, 0.00000599, 0.00000584, -0.00000689, -0.00000824, 0.00000157, -0.00000092, 0.00000502, -0.00000322, -0.00000624, 0.00000565, -0.00001404, 0.00000493, 0.00000520, -0.00000021, -0.00001154, 0.00000886, -0.00000601, 0.00000155, 0.00000303, 0.00001488, -0.00001401, -0.00000024, 0.00000521, },
		{ 0.00001183, -0.00000275, -0.00000588, 0.00000360, -0.00000786, 0.00000537, -0.00000120, 0.00000052, 0.00000213, -0.00000226, -0.00000287, 0.00000787, 0.00000135, 0.00000474, -0.00000425, -0.00000430, 0.00000834, 0.00000815, -0.00000007, 0.00000793, -0.00001958, 0.00001019, -0.00000842, -0.00000253, -0.00000604, 0.00000704, -0.00000844, -0.00000717, 0.00000083, 0.00000610, -0.00000333, -0.00000049, -0.00001361, -0.00000030, -0.00000318, 0.00000296, -0.00000107, -0.00001402, -0.00000474, 0.00000899, -0.00001680, 0.00001768, -0.00002486, 0.00001816, -0.00001541, 0.00000530, -0.00001391, 0.00000830, -0.00002707, 0.00000511, -0.00004535, 0.00000995, -0.00014096, -0.00001453, 0.00018568, 0.00006683, 0.00006881, 0.00013576, 0.00011272, -0.00017759, -0.00001064, -0.00003863, -0.00000435, -0.00002041, -0.00000205, -0.00001242, 0.00000177, -0.00001338, 0.00000728, -0.00000724, -0.00000371, -0.00000502, -0.00000044, -0.00000814, -0.00001257, -0.00002798, 0.00000229, 0.00001145, 0.00003564, -0.00000465, 0.00000975, 0.00000378, 0.00001065, 0.00000813, 0.00001234, 0.00001069, 0.00002322, 0.00000186, -0.00000727, -0.00002789, 0.00000428, -0.00000243, 0.00000603, -0.00001903, -0.00000271, 0.00000108, 0.00000678, -0.00001177, -0.00000038, 0.00000056, -0.00000092, -0.00000184, -0.00000592, -0.00000303, -0.00000609, 0.00000167, 0.00000872, -0.00000565, -0.00000715, -0.00000265, 0.00000215, 0.00000162, 0.00000264, -0.00001217, 0.00000991, -0.00001479, 0.00000290, 0.00000504, 0.00000339, -0.00001432, 0.00000806, -0.00000328, 0.00000010, 0.00000045, 0.00001537, -0.00000875, -0.00000490, 0.00000511, },
		{ 0.00000910, 0.00000335, -0.00000976, 0.00000558, -0.00000860, 0.00000461, -0.00000024, -0.00000000, 0.00000147, -0.00000030, -0.00000515, 0.00000848, 0.00000088, 0.00000533, -0.00000304, -0.00000434, 0.00000256, 0.00001500, -0.00000541, 0.00001353, -0.00001936, 0.00000499, -0.00000183, -0.00000789, -0.00000281, 0.00000338, -0.00000194, -0.00001302, 0.00000220, 0.00000492, -0.00000105, -0.00000180, -0.00000949, -0.00000607, 0.00000043, -0.00000103, 0.00000387, -0.00001512, -0.00000724, 0.00001081, -0.00001726, 0.00001625, -0.00001962, 0.00000978, -0.00000634, -0.00000275, -0.00000686, 0.00000214, -0.00001781, -0.00000452, -0.00003118, -0.00000553, -0.00008696, -0.00011023, 0.00019467, 0.00009114, 0.00006909, 0.00012162, 0.00015583, -0.00017053, -0.00003609, -0.00002699, -0.00001434, -0.00001518, -0.00000671, -0.00001009, -0.00000032, -0.00001137, 0.00000277, -0.00000218, -0.00000617, -0.00000441, -0.00000091, -0.00000687, -0.00000955, -0.00002954, 0.00000109, 0.00000604, 0.00003653, -0.00000128, 0.00000765, 0.00000263, 0.00001132, 0.00000524, 0.00001480, 0.00000448, 0.00002619, 0.00000273, 0.00000185, -0.00003136, 0.00000045, -0.00000014, 0.00000545, -0.00001590, -0.00000677, 0.00000159, 0.00000596, -0.00000794, -0.00000437, 0.00000319, -0.00000282, 0.00000010, -0.00000622, -0.00000344, -0.00000481, -0.00000187, 0.00000973, -0.00000368, -0.00000563, -0.00000623, 0.00000363, -0.00000034, 0.00000652, -0.00001469, 0.00000957, -0.00001108, -0.00000201, 0.00000649, 0.00000482, -0.00001348, 0.00000345, 0.00000173, -0.00000230, -0.00000014, 0.00001241, -0.00000144, -0.00000950, 0.00000541, },
		{ 0.00000635, 0.00000808, -0.00001108, 0.00000520, -0.00000744, 0.00000255, 0.00000130, -0.00000057, 0.00000076, 0.00000131, -0.00000650, 0.00000795, 0.00000149, 0.00000484, -0.00000152, -0.00000356, -0.00000289, 0.00001845, -0.00000515, 0.00001154, -0.00001075, -0.00000672, 0.00000866, -0.00001531, 0.00000197, -0.00000220, 0.00000586, -0.00001841, 0.00000291, 0.00000308, 0.00000265, -0.00000517, -0.00000338, -0.00001243, 0.00000408, -0.00000526, 0.00000826, -0.00001514, -0.00000966, 0.00001086, -0.00001466, 0.00001075, -0.00001009, -0.00000196, 0.00000511, -0.00001278, 0.00000240, -0.00000748, -0.00000406, -0.00001940, -0.00001086, -0.00002994, -0.00003061, -0.00017281, 0.00015361, 0.00013846, 0.00006339, 0.00011255, 0.00018645, -0.00014435, -0.00007260, -0.00001084, -0.00002726, -0.00000758, -0.00001391, -0.00000508, -0.00000552, -0.00000601, -0.00000440, 0.00000381, -0.00000864, -0.00000350, -0.00000192, -0.00000574, -0.00000656, -0.00003022, -0.00000073, 0.00000213, 0.00003503, 0.00000319, 0.00000580, 0.00000162, 0.00001126, 0.00000356, 0.00001562, 0.00000145, 0.00002464, 0.00000653, 0.00000581, -0.00002623, -0.00001026, 0.00000635, 0.00000141, -0.00000959, -0.00001309, 0.00000436, 0.00000229, -0.00000187, -0.00000951, 0.00000643, -0.00000500, 0.00000188, -0.00000623, -0.00000380, -0.00000399, -0.00000378, 0.00000848, -0.00000082, -0.00000457, -0.00000778, 0.00000239, 0.00000010, 0.00000738, -0.00001298, 0.00000469, -0.00000401, -0.00000831, 0.00000850, 0.00000465, -0.00000991, -0.00000319, 0.00000688, -0.00000408, 0.00000010, 0.00000788, 0.00000547, -0.00001216, 0.00000485, },
		{ 0.00000472, 0.00001021, -0.00000932, 0.00000259, -0.00000481, -0.00000029, 0.00000306, -0.00000117, 0.00000033, 0.00000203, -0.00000638, 0.00000582, 0.00000352, 0.00000307, 0.00000069, -0.00000327, -0.00000565, 0.00001628, 0.00000171, 0.00000279, 0.00000318, -0.00002021, 0.00001798, -0.00002025, 0.00000443, -0.00000605, 0.00001120, -0.00001999, 0.00000062, 0.00000246, 0.00000561, -0.00000786, 0.00000134, -0.00001592, 0.00000469, -0.00000679, 0.00000932, -0.00001192, -0.00001357, 0.00001071, -0.00001110, 0.00000405, -0.00000013, -0.00001253, 0.00001415, -0.00001988, 0.00000887, -0.00001498, 0.00000728, -0.00003083, 0.00000447, -0.00004736, 0.00000572, -0.00018524, 0.00006633, 0.00019255, 0.00006381, 0.00010249, 0.00020837, -0.00010425, -0.00011299, 0.00000343, -0.00003840, -0.00000112, -0.00002079, 0.00000003, -0.00001123, -0.00000021, -0.00001104, 0.00000785, -0.00000911, -0.00000364, -0.00000233, -0.00000548, -0.00000349, -0.00002980, -0.00000352, 0.00000008, 0.00003109, 0.00000884, 0.00000334, 0.00000234, 0.00000864, 0.00000501, 0.00001271, 0.00000355, 0.00001735, 0.00001452, 0.00000342, -0.00001347, -0.00002471, 0.00001364, -0.00000321, -0.00000260, -0.00001949, 0.00000740, -0.00000204, 0.00000400, -0.00001353, 0.00000834, -0.00000588, 0.00000238, -0.00000540, -0.00000426, -0.00000376, -0.00000412, 0.00000562, 0.00000231, -0.00000400, -0.00000716, -0.00000115, 0.00000220, 0.00000600, -0.00000833, -0.00000246, 0.00000348, -0.00001328, 0.00000911, 0.00000451, -0.00000570, -0.00000912, 0.00000959, -0.00000356, -0.00000022, 0.00000381, 0.00000979, -0.00001163, 0.00000272, },
		{ 0.00000429, 0.00000980, -0.00000532, -0.00000098, -0.00000188, -0.00000301, 0.00000449, -0.00000161, 0.00000026, 0.00000174, -0.00000484, 0.00000241, 0.00000649, 0.00000059, 0.00000329, -0.00000385, -0.00000516, 0.00000915, 0.00001254, -0.00000815, 0.00001624, -0.00002890, 0.00002052, -0.00001848, 0.00000140, -0.00000534, 0.00001133, -0.00001581, -0.00000570, 0.00000406, 0.00000623, -0.00000765, 0.00000211, -0.00001433, 0.00000070, -0.00000424, 0.00000591, -0.00000492, -0.00001925, 0.00001109, -0.00000808, -0.00000139, 0.00000698, -0.00001834, 0.00001729, -0.00002072, 0.00000914, -0.00001632, 0.00001111, -0.00003260, 0.00000740, -0.00004835, 0.00001260, -0.00015095, -0.00004313, 0.00022606, 0.00008564, 0.00008443, 0.00022656, -0.00005692, -0.00014920, 0.00000947, -0.00004344, 0.00000094, -0.00002448, 0.00000247, -0.00001460, 0.00000300, -0.00001424, 0.00000780, -0.00000646, -0.00000545, -0.00000167, -0.00000621, -0.00000066, -0.00002795, -0.00000750, -0.00000002, 0.00002491, 0.00001561, -0.00000027, 0.00000558, 0.00000295, 0.00000975, 0.00000620, 0.00001002, 0.00000609, 0.00002463, -0.00000248, 0.00000187, -0.00003686, 0.00001668, -0.00000464, 0.00000190, -0.00002315, 0.00000808, -0.00000428, 0.00000697, -0.00001420, 0.00000721, -0.00000419, 0.00000084, -0.00000335, -0.00000505, -0.00000371, -0.00000372, 0.00000252, 0.00000450, -0.00000328, -0.00000530, -0.00000527, 0.00000406, 0.00000434, -0.00000329, -0.00000867, 0.00000811, -0.00001442, 0.00000672, 0.00000580, -0.00000278, -0.00001216, 0.00000837, -0.00000024, -0.00000157, 0.00000130, 0.00001072, -0.00000808, -0.00000062, },
		{ 0.00000421, 0.00000801, -0.00000083, -0.00000380, 0.00000007, -0.00000473, 0.00000507, -0.00000163, 0.00000037, 0.00000080, -0.00000257, -0.00000133, 0.00000926, -0.00000148, 0.00000538, -0.00000469, -0.00000264, 0.00000007, 0.00002250, -0.00001536, 0.00002230, -0.00002786, 0.00001345, -0.00000893, -0.00000730, -0.00000006, 0.00000645, -0.00000682, -0.00001454, 0.00000690, 0.00000468, -0.00000431, -0.00000120, -0.00000824, -0.00000672, 0.00000108, -0.00000044, 0.00000394, -0.00002490, 0.00001099, -0.00000554, -0.00000493, 0.00001031, -0.00001867, 0.00001432, -0.00001552, 0.00000358, -0.00001161, 0.00000717, -0.00002458, -0.00000161, -0.00003440, -0.00000357, -0.00009163, -0.00013945, 0.00021250, 0.00013714, 0.00005755, 0.00024200, -0.00000618, -0.00017608, 0.00000418, -0.00004108, -0.00000228, -0.00002397, 0.00000099, -0.00001424, 0.00000223, -0.00001301, 0.00000358, -0.00000138, -0.00000812, -0.00000055, -0.00000721, 0.00000113, -0.00002428, -0.00001265, 0.00000154, 0.00001707, 0.00002296, -0.00000475, 0.00001073, -0.00000444, 0.00001579, -0.00000144, 0.00001760, -0.00000492, 0.00003242, -0.00000674, 0.00001329, -0.00004074, 0.00001165, -0.00000054, 0.00000208, -0.00002238, 0.00000480, -0.00000278, 0.00000565, -0.00001070, 0.00000271, 0.00000005, -0.00000251, -0.00000037, -0.00000616, -0.00000345, -0.00000353, 0.00000049, 0.00000479, -0.00000177, -0.00000346, -0.00000808, 0.00000383, 0.00000420, -0.00000001, -0.00001152, 0.00000803, -0.00001100, 0.00000138, 0.00000855, -0.00000167, -0.00001187, 0.00000372, 0.00000464, -0.00000307, 0.00000003, 0.00000906, -0.00000310, -0.00000374, },
		{ 0.00000342, 0.00000625, 0.00000243, -0.00000458, 0.00000035, -0.00000514, 0.00000469, -0.00000117, 0.00000041, -0.00000022, -0.00000048, -0.00000425, 0.00001064, -0.00000203, 0.00000605, -0.00000479, -0.00000007, -0.00000733, 0.00002714, -0.00001472, 0.00001855, -0.00001664, -0.00000131, 0.00000513, -0.00001812, 0.00000654, -0.00000026, 0.00000330, -0.00002226, 0.00000832, 0.00000285, 0.00000031, -0.00000625, -0.00000077, -0.00001402, 0.00000578, -0.00000626, 0.00001101, -0.00002740, 0.00000824, -0.00000201, -0.00000780, 0.00001131, -0.00001571, 0.00000826, -0.00000785, -0.00000392, -0.00000482, -0.00000016, -0.00001235, -0.00001509, -0.00001578, -0.00002606, -0.00003618, -0.00019137, 0.00014046, 0.00021050, 0.00003199, 0.00024880, 0.00004924, -0.00019333, -0.00001081, -0.00003396, -0.00000857, -0.00002083, -0.00000324, -0.00001113, -0.00000146, -0.00000887, -0.00000259, 0.00000370, -0.00000974, -0.00000037, -0.00000724, 0.00000097, -0.00001859, -0.00001866, 0.00000414, 0.00000856, 0.00002983, -0.00000895, 0.00001593, -0.00001083, 0.00001999, -0.00000671, 0.00002219, -0.00001112, 0.00003358, -0.00000493, 0.00001627, -0.00003378, -0.00000147, 0.00000803, -0.00000116, -0.00001788, -0.00000186, 0.00000192, 0.00000089, -0.00000438, -0.00000359, 0.00000529, -0.00000622, 0.00000244, -0.00000704, -0.00000299, -0.00000399, 0.00000003, 0.00000317, 0.00000044, -0.00000225, -0.00000859, 0.00000092, 0.00000595, 0.00000093, -0.00001065, 0.00000385, -0.00000470, -0.00000488, 0.00001120, -0.00000128, -0.00000966, -0.00000207, 0.00000856, -0.00000301, -0.00000118, 0.00000654, 0.00000114, -0.00000508, },
		{ 0.00000158, 0.00000530, 0.00000362, -0.00000319, -0.00000065, -0.00000473, 0.00000377, -0.00000043, 0.00000029, -0.00000088, 0.00000075, -0.00000562, 0.00001006, -0.00000068, 0.00000503, -0.00000363, 0.00000107, -0.00001068, 0.00002454, -0.00000598, 0.00000697, 0.00000025, -0.00001760, 0.00001737, -0.00002546, 0.00000973, -0.00000438, 0.00001002, -0.00002496, 0.00000560, 0.00000298, 0.00000358, -0.00000974, 0.00000415, -0.00001728, 0.00000634, -0.00000815, 0.00001305, -0.00002422, 0.00000118, 0.00000390, -0.00001163, 0.00001233, -0.00001278, 0.00000323, -0.00000230, -0.00000856, -0.00000096, -0.00000507, -0.00000324, -0.00002374, -0.00000429, -0.00003826, -0.00000654, -0.00018598, 0.00002247, 0.00027992, 0.00002703, 0.00023654, 0.00011374, -0.00020382, -0.00003078, -0.00002705, -0.00001404, -0.00001810, -0.00000758, -0.00000786, -0.00000538, -0.00000488, -0.00000744, 0.00000591, -0.00000839, -0.00000235, -0.00000539, -0.00000154, -0.00001126, -0.00002477, 0.00000684, 0.00000060, 0.00003490, -0.00001131, 0.00001892, -0.00001349, 0.00001964, -0.00000700, 0.00002110, -0.00001007, 0.00002659, 0.00000390, 0.00001081, -0.00001852, -0.00001821, 0.00001677, -0.00000449, -0.00001237, -0.00000937, 0.00000733, -0.00000457, 0.00000184, -0.00000893, 0.00000909, -0.00000828, 0.00000365, -0.00000686, -0.00000277, -0.00000478, 0.00000058, 0.00000068, 0.00000239, -0.00000138, -0.00000716, -0.00000363, 0.00000832, 0.00000073, -0.00000777, -0.00000174, 0.00000118, -0.00000907, 0.00001157, 0.00000015, -0.00000772, -0.00000626, 0.00000918, -0.00000015, -0.00000319, 0.00000463, 0.00000312, -0.00000392, },
		{ -0.00000058, 0.00000493, 0.00000320, -0.00000070, -0.00000166, -0.00000450, 0.00000299, 0.00000020, 0.00000009, -0.00000107, 0.00000098, -0.00000547, 0.00000780, 0.00000201, 0.00000295, -0.00000154, 0.00000053, -0.00000982, 0.00001611, 0.00000724, -0.00000667, 0.00001550, -0.00002798, 0.00002166, -0.00002484, 0.00000624, -0.00000303, 0.00001067, -0.00002084, -0.00000198, 0.00000581, 0.00000404, -0.00000955, 0.00000417, -0.00001461, 0.00000148, -0.00000498, 0.00000922, -0.00001533, -0.00000969, 0.00001184, -0.00001676, 0.00001451, -0.00001182, 0.00000161, -0.00000133, -0.00000793, -0.00000251, -0.00000446, -0.00000142, -0.00002251, -0.00000579, -0.00003313, -0.00000784, -0.00013467, -0.00010708, 0.00030995, 0.00006179, 0.00019778, 0.00018928, -0.00020912, -0.00005187, -0.00002421, -0.00001602, -0.00001788, -0.00000994, -0.00000665, -0.00000710, -0.00000360, -0.00000859, 0.00000375, -0.00000359, -0.00000645, -0.00000185, -0.00000579, -0.00000343, -0.00002980, 0.00000854, -0.00000560, 0.00003696, -0.00001049, 0.00001804, -0.00001094, 0.00001393, -0.00000204, 0.00001450, -0.00000271, 0.00001373, 0.00001658, 0.00000136, -0.00000141, -0.00003168, 0.00002029, -0.00000413, -0.00000892, -0.00001483, 0.00001057, -0.00000771, 0.00000502, -0.00001083, 0.00000951, -0.00000728, 0.00000239, -0.00000510, -0.00000328, -0.00000520, 0.00000101, -0.00000109, 0.00000280, -0.00000006, -0.00000501, -0.00000781, 0.00000921, 0.00000142, -0.00000536, -0.00000560, 0.00000357, -0.00000901, 0.00000840, 0.00000365, -0.00000750, -0.00000728, 0.00000582, 0.00000514, -0.00000566, 0.00000362, 0.00000277, -0.00000101, },
		{ -0.00000162, 0.00000421, 0.00000238, 0.00000119, -0.00000128, -0.00000531, 0.00000283, 0.00000051, -0.00000009, -0.00000089, 0.00000050, -0.00000441, 0.00000481, 0.00000489, 0.00000094, 0.00000055, -0.00000091, -0.00000635, 0.00000543, 0.00001943, -0.00001590, 0.00002271, -0.00002770, 0.00001561, -0.00001574, -0.00000346, 0.00000308, 0.00000612, -0.00001164, -0.00001214, 0.00000969, 0.00000244, -0.00000606, -0.00000011, -0.00000738, -0.00000680, 0.00000114, 0.00000192, -0.00000374, -0.00002120, 0.00001920, -0.00002129, 0.00001663, -0.00001199, 0.00000248, -0.00000363, -0.00000374, -0.00000756, -0.00000005, -0.00000527, -0.00001359, -0.00001678, -0.00001622, -0.00002783, -0.00006720, -0.00020529, 0.00027141, 0.00014232, 0.00013661, 0.00026950, -0.00020507, -0.00007468, -0.00002510, -0.00001549, -0.00001934, -0.00001067, -0.00000755, -0.00000636, -0.00000528, -0.00000622, -0.00000172, 0.00000289, -0.00001089, 0.00000180, -0.00001007, 0.00000308, -0.00003231, 0.00000814, -0.00000901, 0.00003517, -0.00000585, 0.00001288, -0.00000362, 0.00000434, 0.00000596, 0.00000518, 0.00000716, -0.00000002, 0.00002751, -0.00000577, 0.00001051, -0.00003608, 0.00001531, 0.00000151, -0.00000881, -0.00001682, 0.00001007, -0.00000695, 0.00000393, -0.00000862, 0.00000635, -0.00000338, -0.00000095, -0.00000212, -0.00000449, -0.00000480, 0.00000037, -0.00000102, 0.00000101, 0.00000216, -0.00000329, -0.00000990, 0.00000713, 0.00000428, -0.00000507, -0.00000597, 0.00000140, -0.00000476, 0.00000250, 0.00000847, -0.00000867, -0.00000566, 0.00000001, 0.00001066, -0.00000696, 0.00000258, 0.00000137, 0.00000187, },
		{ -0.00000038, 0.00000241, 0.00000227, 0.00000125, 0.00000113, -0.00000718, 0.00000307, 0.00000071, -0.00000035, -0.00000050, -0.00000021, -0.00000320, 0.00000211, 0.00000689, -0.00000004, 0.00000185, -0.00000222, -0.00000243, -0.00000368, 0.00002608, -0.00001679, 0.00001967, -0.00001722, 0.00000216, -0.00000238, -0.00001491, 0.00000987, 0.00000039, -0.00000186, -0.00002047, 0.00001130, 0.00000126, -0.00000181, -0.00000556, 0.00000045, -0.00001418, 0.00000601, -0.00000447, 0.00000584, -0.00002887, 0.00002222, -0.00002207, 0.00001577, -0.00001026, 0.00000229, -0.00000498, -0.00000077, -0.00001097, 0.00000281, -0.00000877, -0.00000463, -0.00002718, -0.00000104, -0.00004567, -0.00001677, -0.00023975, 0.00015781, 0.00025210, 0.00007270, 0.00033817, -0.00018101, -0.00010484, -0.00002496, -0.00001715, -0.00001870, -0.00001263, -0.00000822, -0.00000538, -0.00000762, -0.00000311, -0.00000714, 0.00000759, -0.00001279, 0.00000322, -0.00001218, 0.00000634, -0.00003105, 0.00000483, -0.00000898, 0.00002933, 0.00000236, 0.00000435, 0.00000644, -0.00000609, 0.00001353, -0.00000304, 0.00001508, -0.00000962, 0.00003187, -0.00000602, 0.00001342, -0.00003001, 0.00000301, 0.00001039, -0.00001043, -0.00001633, 0.00000660, -0.00000309, -0.00000028, -0.00000390, 0.00000139, 0.00000160, -0.00000470, 0.00000078, -0.00000569, -0.00000380, -0.00000141, 0.00000084, -0.00000243, 0.00000471, -0.00000220, -0.00000943, 0.00000219, 0.00000886, -0.00000658, -0.00000342, -0.00000374, 0.00000111, -0.00000337, 0.00001233, -0.00000936, -0.00000361, -0.00000528, 0.00001346, -0.00000521, 0.00000027, 0.00000057, 0.00000285, },
		{ 0.00000311, -0.00000025, 0.00000310, -0.00000060, 0.00000490, -0.00000906, 0.00000281, 0.00000140, -0.00000096, 0.00000006, -0.00000088, -0.00000233, 0.00000021, 0.00000760, 0.00000020, 0.00000223, -0.00000293, 0.00000048, -0.00000902, 0.00002566, -0.00000974, 0.00000927, -0.00000200, -0.00001177, 0.00000835, -0.00002209, 0.00001226, -0.00000181, 0.00000375, -0.00002290, 0.00000783, 0.00000283, 0.00000041, -0.00000860, 0.00000470, -0.00001652, 0.00000588, -0.00000617, 0.00000954, -0.00002934, 0.00001825, -0.00001675, 0.00000949, -0.00000377, -0.00000245, -0.00000132, -0.00000347, -0.00000803, -0.00000085, -0.00000603, -0.00000325, -0.00002733, 0.00000026, -0.00004475, -0.00000372, -0.00020376, -0.00000681, 0.00035272, 0.00003639, 0.00037461, -0.00012454, -0.00014899, -0.00001797, -0.00002638, -0.00001179, -0.00001915, -0.00000557, -0.00000726, -0.00000728, -0.00000289, -0.00000870, 0.00000722, -0.00000981, 0.00000071, -0.00001062, 0.00000525, -0.00002562, -0.00000146, -0.00000558, 0.00001995, 0.00001299, -0.00000565, 0.00001643, -0.00001417, 0.00001769, -0.00000735, 0.00001820, -0.00001253, 0.00002821, 0.00000092, 0.00000859, -0.00001732, -0.00001118, 0.00001757, -0.00001023, -0.00001601, 0.00000255, 0.00000136, -0.00000475, 0.00000034, -0.00000252, 0.00000515, -0.00000675, 0.00000206, -0.00000588, -0.00000295, -0.00000352, 0.00000331, -0.00000587, 0.00000618, -0.00000101, -0.00000734, -0.00000384, 0.00001313, -0.00000791, -0.00000040, -0.00000850, 0.00000486, -0.00000595, 0.00001275, -0.00000749, -0.00000350, -0.00000733, 0.00001153, 0.00000034, -0.00000364, 0.00000120, 0.00000122, },
		{ 0.00000736, -0.00000248, 0.00000406, -0.00000325, 0.00000832, -0.00000932, 0.00000095, 0.00000318, -0.00000215, 0.00000082, -0.00000141, -0.00000181, -0.00000100, 0.00000742, 0.00000103, 0.00000229, -0.00000328, 0.00000212, -0.00001054, 0.00001984, 0.00000122, -0.00000232, 0.00001042, -0.00001892, 0.00001078, -0.00002101, 0.00000746, 0.00000182, 0.00000310, -0.00001828, -0.00000080, 0.00000738, -0.00000045, -0.00000754, 0.00000347, -0.00001243, -0.00000006, -0.00000243, 0.00000672, -0.00002256, 0.00000776, -0.00000582, -0.00000204, 0.00000769, -0.00001234, 0.00000812, -0.00001257, 0.00000185, -0.00001165, 0.00000419, -0.00001155, -0.00001461, -0.00001502, -0.00002236, -0.00002631, -0.00011987, -0.00017409, 0.00039699, 0.00005498, 0.00036467, -0.00002992, -0.00020821, -0.00000275, -0.00004468, 0.00000218, -0.00003090, 0.00000140, -0.00001342, -0.00000253, -0.00000752, -0.00000466, 0.00000085, -0.00000192, -0.00000542, -0.00000568, 0.00000034, -0.00001698, -0.00000978, 0.00000025, 0.00000840, 0.00002421, -0.00001484, 0.00002374, -0.00001774, 0.00001720, -0.00000720, 0.00001650, -0.00000966, 0.00001904, 0.00001116, 0.00000142, -0.00000499, -0.00002040, 0.00001807, -0.00000513, -0.00001810, 0.00000016, 0.00000392, -0.00000680, 0.00000152, -0.00000326, 0.00000556, -0.00000585, 0.00000095, -0.00000453, -0.00000290, -0.00000493, 0.00000485, -0.00000742, 0.00000515, 0.00000124, -0.00000514, -0.00000862, 0.00001467, -0.00000680, 0.00000040, -0.00000976, 0.00000371, -0.00000346, 0.00000885, -0.00000243, -0.00000627, -0.00000532, 0.00000514, 0.00000829, -0.00000789, 0.00000263, -0.00000218, },
		{ 0.00001014, -0.00000261, 0.00000396, -0.00000512, 0.00000953, -0.00000664, -0.00000303, 0.00000608, -0.00000375, 0.00000161, -0.00000184, -0.00000142, -0.00000200, 0.00000707, 0.00000158, 0.00000283, -0.00000381, 0.00000296, -0.00000962, 0.00001203, 0.00001087, -0.00000909, 0.00001451, -0.00001575, 0.00000383, -0.00001236, -0.00000291, 0.00000947, -0.00000183, -0.00000939, -0.00001135, 0.00001234, -0.00000279, -0.00000361, -0.00000167, -0.00000426, -0.00000893, 0.00000388, 0.00000041, -0.00001212, -0.00000536, 0.00000704, -0.00001535, 0.00002075, -0.00002393, 0.00001959, -0.00002384, 0.00001401, -0.00002461, 0.00001688, -0.00002429, 0.00000452, -0.00003817, 0.00000932, -0.00006347, -0.00002928, -0.00029022, 0.00034989, 0.00013631, 0.00031173, 0.00009457, -0.00027273, 0.00001372, -0.00006657, 0.00001823, -0.00004387, 0.00000978, -0.00002178, 0.00000500, -0.00001545, 0.00000312, -0.00000898, 0.00000788, -0.00001245, 0.00000035, -0.00000610, -0.00000746, -0.00001813, 0.00000673, -0.00000323, 0.00003372, -0.00002096, 0.00002659, -0.00001623, 0.00001282, -0.00000417, 0.00001221, -0.00000434, 0.00000907, 0.00001912, -0.00000199, 0.00000084, -0.00002027, 0.00001015, 0.00000493, -0.00002244, -0.00000025, 0.00000386, -0.00000557, -0.00000089, -0.00000076, 0.00000309, -0.00000251, -0.00000193, -0.00000210, -0.00000365, -0.00000511, 0.00000448, -0.00000613, 0.00000124, 0.00000471, -0.00000383, -0.00001055, 0.00001218, -0.00000228, -0.00000216, -0.00000645, -0.00000246, 0.00000307, 0.00000225, 0.00000431, -0.00001065, -0.00000088, -0.00000305, 0.00001549, -0.00001006, 0.00000321, -0.00000540, },
		{ 0.00000972, 0.00000029, 0.00000218, -0.00000520, 0.00000757, -0.00000096, -0.00000845, 0.00000922, -0.00000502, 0.00000201, -0.00000203, -0.00000099, -0.00000315, 0.00000706, 0.00000138, 0.00000409, -0.00000463, 0.00000338, -0.00000782, 0.00000527, 0.00001565, -0.00000836, 0.00000962, -0.00000445, -0.00000802, -0.00000158, -0.00001335, 0.00001605, -0.00000610, -0.00000154, -0.00001870, 0.00001375, -0.00000346, -0.00000009, -0.00000682, 0.00000326, -0.00001569, 0.00000793, -0.00000442, -0.00000339, -0.00001567, 0.00001659, -0.00002523, 0.00002994, -0.00003137, 0.00002663, -0.00003025, 0.00002085, -0.00003171, 0.00002363, -0.00003199, 0.00001838, -0.00005438, 0.00003117, -0.00008763, 0.00002786, -0.00031925, 0.00020764, 0.00025838, 0.00024040, 0.00022580, -0.00032236, 0.00001693, -0.00008069, 0.00002672, -0.00005040, 0.00001349, -0.00002717, 0.00001055, -0.00002205, 0.00000978, -0.00001710, 0.00001466, -0.00001628, 0.00000409, -0.00001093, -0.00000001, -0.00002410, 0.00001162, -0.00001254, 0.00003923, -0.00002226, 0.00002424, -0.00001043, 0.00000647, -0.00000063, 0.00000806, -0.00000000, 0.00000240, 0.00002094, 0.00000153, -0.00000159, -0.00001163, -0.00000301, 0.00001616, -0.00002601, -0.00000075, 0.00000270, -0.00000256, -0.00000501, 0.00000282, -0.00000006, 0.00000115, -0.00000469, -0.00000001, -0.00000440, -0.00000441, 0.00000241, -0.00000252, -0.00000439, 0.00000835, -0.00000316, -0.00000965, 0.00000637, 0.00000453, -0.00000694, -0.00000022, -0.00001089, 0.00001007, -0.00000352, 0.00000968, -0.00001389, 0.00000289, -0.00000928, 0.00001850, -0.00000801, 0.00000150, -0.00000668, },
		{ 0.00000589, 0.00000566, -0.00000068, -0.00000365, 0.00000298, 0.00000625, -0.00001346, 0.00001101, -0.00000488, 0.00000140, -0.00000173, -0.00000064, -0.00000436, 0.00000721, 0.00000079, 0.00000550, -0.00000512, 0.00000318, -0.00000580, 0.00000083, 0.00001516, -0.00000181, 0.00000002, 0.00000827, -0.00001710, 0.00000414, -0.00001771, 0.00001636, -0.00000495, 0.00000057, -0.00001889, 0.00000889, -0.00000011, 0.00000015, -0.00000828, 0.00000584, -0.00001618, 0.00000597, -0.00000400, -0.00000038, -0.00001926, 0.00001906, -0.00002768, 0.00003082, -0.00002974, 0.00002391, -0.00002612, 0.00001649, -0.00002686, 0.00001802, -0.00002713, 0.00001753, -0.00005198, 0.00002899, -0.00008051, 0.00003131, -0.00025946, 0.00000445, 0.00037270, 0.00018864, 0.00033414, -0.00033397, -0.00000855, -0.00007543, 0.00001830, -0.00004329, 0.00000673, -0.00002436, 0.00000906, -0.00002220, 0.00001011, -0.00001843, 0.00001412, -0.00001370, 0.00000294, -0.00001168, 0.00000313, -0.00002587, 0.00001313, -0.00001753, 0.00003909, -0.00001790, 0.00001702, -0.00000187, 0.00000013, 0.00000175, 0.00000552, 0.00000184, 0.00000024, 0.00001663, 0.00001036, -0.00000873, -0.00000033, -0.00001459, 0.00002259, -0.00002453, -0.00000434, 0.00000302, -0.00000051, -0.00000779, 0.00000438, -0.00000108, 0.00000269, -0.00000532, 0.00000026, -0.00000411, -0.00000374, -0.00000019, 0.00000168, -0.00000950, 0.00001024, -0.00000198, -0.00000734, -0.00000045, 0.00001092, -0.00001117, 0.00000558, -0.00001741, 0.00001326, -0.00000491, 0.00001092, -0.00001356, 0.00000328, -0.00001076, 0.00001549, -0.00000136, -0.00000251, -0.00000557, },
		{ 0.00000015, 0.00001147, -0.00000293, -0.00000170, -0.00000251, 0.00001257, -0.00001582, 0.00000995, -0.00000251, -0.00000058, -0.00000073, -0.00000063, -0.00000518, 0.00000688, 0.00000071, 0.00000594, -0.00000426, 0.00000173, -0.00000337, -0.00000183, 0.00001173, 0.00000593, -0.00000771, 0.00001498, -0.00001685, 0.00000018, -0.00001303, 0.00000859, 0.00000294, -0.00000412, -0.00001177, -0.00000145, 0.00000671, -0.00000327, -0.00000502, 0.00000235, -0.00000979, -0.00000205, 0.00000175, -0.00000334, -0.00001602, 0.00001437, -0.00002231, 0.00002270, -0.00001825, 0.00001070, -0.00001098, 0.00000079, -0.00001023, 0.00000031, -0.00000946, 0.00000101, -0.00002989, 0.00000238, -0.00004297, -0.00001094, -0.00014417, -0.00019813, 0.00042241, 0.00018978, 0.00039787, -0.00029296, -0.00006987, -0.00004659, -0.00000989, -0.00002095, -0.00001176, -0.00001181, -0.00000140, -0.00001372, 0.00000191, -0.00001112, 0.00000526, -0.00000452, -0.00000310, -0.00000809, 0.00000164, -0.00002320, 0.00001073, -0.00001740, 0.00003294, -0.00000835, 0.00000626, 0.00000761, -0.00000492, 0.00000270, 0.00000415, 0.00000210, 0.00000071, 0.00000981, 0.00001923, -0.00001386, 0.00000606, -0.00001779, 0.00001960, -0.00001543, -0.00001259, 0.00000637, -0.00000133, -0.00000705, 0.00000196, 0.00000153, 0.00000095, -0.00000300, -0.00000178, -0.00000239, -0.00000380, -0.00000209, 0.00000457, -0.00001193, 0.00000883, 0.00000079, -0.00000533, -0.00000566, 0.00001417, -0.00001218, 0.00000785, -0.00001866, 0.00001001, -0.00000049, 0.00000739, -0.00000915, -0.00000037, -0.00000719, 0.00000748, 0.00000782, -0.00000700, -0.00000324, },
		{ -0.00000516, 0.00001523, -0.00000279, -0.00000074, -0.00000695, 0.00001579, -0.00001402, 0.00000551, 0.00000200, -0.00000361, 0.00000075, -0.00000106, -0.00000517, 0.00000538, 0.00000200, 0.00000455, -0.00000142, -0.00000132, -0.00000021, -0.00000399, 0.00000846, 0.00001023, -0.00000856, 0.00001171, -0.00000578, -0.00001239, -0.00000191, -0.00000398, 0.00001416, -0.00001206, -0.00000154, -0.00001282, 0.00001339, -0.00000767, 0.00000059, -0.00000451, -0.00000002, -0.00001217, 0.00000904, -0.00000856, -0.00000977, 0.00000633, -0.00001269, 0.00000920, -0.00000100, -0.00000824, 0.00000957, -0.00001986, 0.00001111, -0.00002193, 0.00001318, -0.00002262, 0.00000141, -0.00003497, 0.00000632, -0.00006898, -0.00002599, -0.00033600, 0.00036823, 0.00025263, 0.00041644, -0.00020236, -0.00015918, -0.00000222, -0.00005048, 0.00000982, -0.00003639, 0.00000619, -0.00001749, 0.00000056, -0.00001217, 0.00000190, -0.00000846, 0.00000763, -0.00001100, -0.00000243, -0.00000258, -0.00001774, 0.00000548, -0.00001299, 0.00002207, 0.00000450, -0.00000581, 0.00001620, -0.00000816, 0.00000309, 0.00000235, 0.00000300, 0.00000053, 0.00000518, 0.00002257, -0.00001129, 0.00000263, -0.00000995, 0.00000712, -0.00000025, -0.00002380, 0.00001167, -0.00000457, -0.00000304, -0.00000389, 0.00000683, -0.00000295, 0.00000111, -0.00000503, 0.00000001, -0.00000445, -0.00000276, 0.00000520, -0.00001083, 0.00000397, 0.00000508, -0.00000442, -0.00000784, 0.00001307, -0.00000905, 0.00000557, -0.00001398, 0.00000091, 0.00000795, 0.00000126, -0.00000264, -0.00000620, -0.00000096, -0.00000211, 0.00001576, -0.00000918, -0.00000171, },
		{ -0.00000826, 0.00001528, 0.00000051, -0.00000137, -0.00000922, 0.00001493, -0.00000818, -0.00000131, 0.00000738, -0.00000661, 0.00000207, -0.00000170, -0.00000429, 0.00000260, 0.00000473, 0.00000143, 0.00000298, -0.00000536, 0.00000337, -0.00000631, 0.00000724, 0.00000907, -0.00000179, 0.00000016, 0.00001139, -0.00002686, 0.00000850, -0.00001450, 0.00002234, -0.00001712, 0.00000550, -0.00001931, 0.00001521, -0.00000911, 0.00000443, -0.00000994, 0.00000756, -0.00001863, 0.00001246, -0.00001080, -0.00000588, 0.00000024, -0.00000418, -0.00000378, 0.00001519, -0.00002501, 0.00002666, -0.00003577, 0.00002681, -0.00003776, 0.00002918, -0.00004027, 0.00002603, -0.00006327, 0.00004233, -0.00010822, 0.00004783, -0.00037013, 0.00020992, 0.00035085, 0.00041367, -0.00008315, -0.00025471, 0.00003901, -0.00008794, 0.00003587, -0.00005654, 0.00002100, -0.00003179, 0.00001385, -0.00002557, 0.00001409, -0.00002053, 0.00001687, -0.00001606, 0.00000160, -0.00000635, -0.00001222, -0.00000044, -0.00000637, 0.00000909, 0.00001769, -0.00001650, 0.00002219, -0.00000951, 0.00000391, -0.00000122, 0.00000613, -0.00000253, 0.00000554, 0.00001795, 0.00000016, -0.00000974, 0.00000541, -0.00000944, 0.00001534, -0.00003322, 0.00001537, -0.00000744, 0.00000159, -0.00001019, 0.00001171, -0.00000612, 0.00000439, -0.00000728, 0.00000147, -0.00000474, -0.00000269, 0.00000398, -0.00000708, -0.00000272, 0.00000941, -0.00000402, -0.00000734, 0.00000863, -0.00000317, 0.00000026, -0.00000566, -0.00001045, 0.00001616, -0.00000356, 0.00000261, -0.00001104, 0.00000431, -0.00000912, 0.00001892, -0.00000696, -0.00000242, },
		{ -0.00000892, 0.00001175, 0.00000611, -0.00000285, -0.00000954, 0.00001070, -0.00000009, -0.00000825, 0.00001162, -0.00000811, 0.00000233, -0.00000206, -0.00000296, -0.00000088, 0.00000806, -0.00000222, 0.00000743, -0.00000884, 0.00000617, -0.00000831, 0.00000782, 0.00000389, 0.00000885, -0.00001326, 0.00002609, -0.00003434, 0.00001044, -0.00001649, 0.00002202, -0.00001437, 0.00000466, -0.00001712, 0.00000958, -0.00000519, 0.00000350, -0.00001016, 0.00000869, -0.00001741, 0.00000853, -0.00000674, -0.00000773, -0.00000048, -0.00000049, -0.00001163, 0.00002461, -0.00003303, 0.00003307, -0.00003942, 0.00002916, -0.00003926, 0.00003019, -0.00004224, 0.00003214, -0.00006827, 0.00004823, -0.00010806, 0.00005660, -0.00030355, -0.00000879, 0.00042956, 0.00042671, 0.00003645, -0.00033016, 0.00005635, -0.00010605, 0.00004421, -0.00006206, 0.00002431, -0.00003688, 0.00001912, -0.00003148, 0.00001884, -0.00002485, 0.00001819, -0.00001463, 0.00000119, -0.00000713, -0.00000898, -0.00000498, 0.00000013, -0.00000307, 0.00002819, -0.00002334, 0.00002433, -0.00000896, 0.00000552, -0.00000646, 0.00001108, -0.00000801, 0.00001007, 0.00000760, 0.00001620, -0.00002461, 0.00002010, -0.00002162, 0.00002410, -0.00003546, 0.00001353, -0.00000664, 0.00000346, -0.00001333, 0.00001271, -0.00000558, 0.00000436, -0.00000655, 0.00000058, -0.00000366, -0.00000285, 0.00000221, -0.00000270, -0.00000866, 0.00001161, -0.00000274, -0.00000578, 0.00000331, 0.00000262, -0.00000504, 0.00000245, -0.00001937, 0.00001964, -0.00000357, 0.00000394, -0.00001239, 0.00000585, -0.00001090, 0.00001585, -0.00000039, -0.00000508, },
		{ -0.00000847, 0.00000632, 0.00001185, -0.00000354, -0.00000907, 0.00000495, 0.00000759, -0.00001279, 0.00001292, -0.00000708, 0.00000101, -0.00000173, -0.00000178, -0.00000407, 0.00001064, -0.00000473, 0.00001006, -0.00001002, 0.00000680, -0.00000881, 0.00000844, -0.00000173, 0.00001756, -0.00002107, 0.00003056, -0.00002873, 0.00000027, -0.00000810, 0.00001236, -0.00000347, -0.00000401, -0.00000719, -0.00000185, 0.00000281, -0.00000185, -0.00000494, 0.00000314, -0.00000890, -0.00000173, 0.00000253, -0.00001437, 0.00000335, -0.00000133, -0.00001371, 0.00002582, -0.00003056, 0.00002725, -0.00002975, 0.00001767, -0.00002640, 0.00001630, -0.00002822, 0.00001889, -0.00004938, 0.00002468, -0.00007166, 0.00001327, -0.00017882, -0.00021604, 0.00042879, 0.00048411, 0.00013795, -0.00036823, 0.00003803, -0.00009715, 0.00002970, -0.00004965, 0.00001347, -0.00003007, 0.00001353, -0.00002700, 0.00001348, -0.00001936, 0.00001062, -0.00000662, -0.00000362, -0.00000466, -0.00000847, -0.00000748, 0.00000516, -0.00001232, 0.00003397, -0.00002502, 0.00002222, -0.00000663, 0.00000738, -0.00001174, 0.00001555, -0.00001313, 0.00001501, -0.00000294, 0.00002953, -0.00003347, 0.00002554, -0.00002239, 0.00002139, -0.00002792, 0.00000462, -0.00000075, 0.00000082, -0.00001135, 0.00000822, -0.00000031, 0.00000041, -0.00000253, -0.00000276, -0.00000109, -0.00000378, 0.00000111, 0.00000039, -0.00001169, 0.00001018, 0.00000038, -0.00000475, -0.00000049, 0.00000579, -0.00000768, 0.00000714, -0.00002249, 0.00001606, 0.00000213, 0.00000110, -0.00000999, 0.00000330, -0.00000764, 0.00000814, 0.00000799, -0.00000763, },
		{ -0.00000872, 0.00000117, 0.00001544, -0.00000196, -0.00000900, -0.00000045, 0.00001261, -0.00001337, 0.00001072, -0.00000359, -0.00000160, -0.00000083, -0.00000108, -0.00000619, 0.00001136, -0.00000486, 0.00000969, -0.00000802, 0.00000462, -0.00000707, 0.00000738, -0.00000438, 0.00001981, -0.00001885, 0.00002212, -0.00001046, -0.00001881, 0.00000621, -0.00000178, 0.00001067, -0.00001542, 0.00000488, -0.00001347, 0.00001026, -0.00000790, 0.00000221, -0.00000517, 0.00000216, -0.00001309, 0.00001189, -0.00002087, 0.00000709, -0.00000267, -0.00001334, 0.00002200, -0.00002136, 0.00001396, -0.00001272, -0.00000062, -0.00000708, -0.00000387, -0.00000730, -0.00000348, -0.00001929, -0.00001156, -0.00002196, -0.00004596, -0.00005670, -0.00034272, 0.00031412, 0.00058380, 0.00022639, -0.00036992, -0.00001172, -0.00006733, -0.00000117, -0.00002568, -0.00000612, -0.00001562, 0.00000051, -0.00001521, 0.00000111, -0.00000755, -0.00000186, 0.00000396, -0.00000954, -0.00000132, -0.00000894, -0.00000903, 0.00000885, -0.00001818, 0.00003473, -0.00002182, 0.00001664, -0.00000299, 0.00000864, -0.00001490, 0.00001674, -0.00001463, 0.00001616, -0.00000814, 0.00003389, -0.00003049, 0.00001762, -0.00001032, 0.00000822, -0.00001296, -0.00000888, 0.00000832, -0.00000509, -0.00000530, -0.00000031, 0.00000782, -0.00000552, 0.00000290, -0.00000689, 0.00000181, -0.00000501, 0.00000090, 0.00000147, -0.00001125, 0.00000535, 0.00000492, -0.00000463, -0.00000181, 0.00000554, -0.00000699, 0.00000753, -0.00001946, 0.00000660, 0.00001113, -0.00000334, -0.00000604, -0.00000125, -0.00000207, -0.00000044, 0.00001435, -0.00000747, },
		{ -0.00001053, -0.00000237, 0.00001572, 0.00000218, -0.00000953, -0.00000455, 0.00001411, -0.00001018, 0.00000603, 0.00000106, -0.00000437, 0.00000001, -0.00000074, -0.00000700, 0.00000997, -0.00000249, 0.00000653, -0.00000343, 0.00000040, -0.00000354, 0.00000424, -0.00000284, 0.00001468, -0.00000751, 0.00000460, 0.00001336, -0.00003796, 0.00001765, -0.00001229, 0.00002052, -0.00002237, 0.00001191, -0.00001874, 0.00001192, -0.00001004, 0.00000640, -0.00001063, 0.00000944, -0.00001909, 0.00001526, -0.00002154, 0.00000537, 0.00000032, -0.00001507, 0.00001827, -0.00001179, 0.00000101, 0.00000252, -0.00001549, 0.00000763, -0.00001857, 0.00000787, -0.00002052, 0.00000435, -0.00003853, 0.00001327, -0.00008330, 0.00001332, -0.00035381, 0.00009909, 0.00068413, 0.00033085, -0.00035261, -0.00007520, -0.00003359, -0.00003330, -0.00000323, -0.00002373, -0.00000243, -0.00001212, -0.00000339, -0.00001118, 0.00000351, -0.00001234, 0.00001102, -0.00001185, -0.00000062, -0.00000766, -0.00001160, 0.00001228, -0.00002135, 0.00003163, -0.00001542, 0.00000927, 0.00000114, 0.00000867, -0.00001444, 0.00001308, -0.00001088, 0.00001150, -0.00000562, 0.00002754, -0.00001580, -0.00000089, 0.00000918, -0.00000862, 0.00000275, -0.00002131, 0.00001612, -0.00001057, 0.00000125, -0.00000896, 0.00001485, -0.00000982, 0.00000662, -0.00000924, 0.00000314, -0.00000534, 0.00000085, 0.00000124, -0.00000855, -0.00000090, 0.00000903, -0.00000446, -0.00000141, 0.00000318, -0.00000446, 0.00000521, -0.00001278, -0.00000478, 0.00001892, -0.00000549, -0.00000369, -0.00000473, 0.00000225, -0.00000598, 0.00001561, -0.00000312, },
		{ -0.00001295, -0.00000455, 0.00001324, 0.00000763, -0.00000975, -0.00000760, 0.00001277, -0.00000492, 0.00000102, 0.00000494, -0.00000591, -0.00000001, -0.00000039, -0.00000684, 0.00000704, 0.00000139, 0.00000203, 0.00000189, -0.00000397, 0.00000023, 0.00000012, 0.00000163, 0.00000503, 0.00000755, -0.00001376, 0.00003255, -0.00004720, 0.00001804, -0.00001265, 0.00002061, -0.00001992, 0.00000945, -0.00001430, 0.00000552, -0.00000608, 0.00000473, -0.00000964, 0.00000898, -0.00001610, 0.00000968, -0.00001382, -0.00000399, 0.00000945, -0.00002091, 0.00001770, -0.00000635, -0.00000582, 0.00000930, -0.00001989, 0.00001053, -0.00002038, 0.00000939, -0.00002296, 0.00001021, -0.00004269, 0.00001833, -0.00008022, 0.00001568, -0.00026459, -0.00015377, 0.00071636, 0.00048801, -0.00033571, -0.00013318, -0.00001335, -0.00005222, 0.00000591, -0.00003008, 0.00000176, -0.00001733, 0.00000173, -0.00001682, 0.00000755, -0.00001519, 0.00001021, -0.00000765, -0.00000456, -0.00000313, -0.00001624, 0.00001604, -0.00002266, 0.00002622, -0.00000778, 0.00000185, 0.00000491, 0.00000730, -0.00001016, 0.00000510, -0.00000291, 0.00000227, 0.00000275, 0.00001402, 0.00000462, -0.00002162, 0.00002649, -0.00002001, 0.00001169, -0.00002694, 0.00001822, -0.00001182, 0.00000443, -0.00001369, 0.00001698, -0.00000936, 0.00000611, -0.00000788, 0.00000154, -0.00000377, -0.00000006, 0.00000107, -0.00000570, -0.00000592, 0.00001049, -0.00000277, -0.00000098, 0.00000103, -0.00000259, 0.00000289, -0.00000613, -0.00001350, 0.00002133, -0.00000251, -0.00000492, -0.00000524, 0.00000312, -0.00000653, 0.00001125, 0.00000447, },
		{ -0.00001363, -0.00000682, 0.00000974, 0.00001226, -0.00000820, -0.00001046, 0.00001006, 0.00000013, -0.00000221, 0.00000661, -0.00000545, -0.00000119, 0.00000017, -0.00000613, 0.00000351, 0.00000533, -0.00000194, 0.00000590, -0.00000661, 0.00000258, -0.00000332, 0.00000647, -0.00000460, 0.00001960, -0.00002480, 0.00003916, -0.00004095, 0.00000481, -0.00000218, 0.00001116, -0.00000882, -0.00000121, -0.00000238, -0.00000612, 0.00000177, -0.00000146, -0.00000308, 0.00000180, -0.00000587, -0.00000237, -0.00000057, -0.00001783, 0.00002143, -0.00002813, 0.00001876, -0.00000465, -0.00000638, 0.00000760, -0.00001435, 0.00000283, -0.00001113, -0.00000059, -0.00001275, 0.00000017, -0.00002684, -0.00000169, -0.00004608, -0.00002704, -0.00013263, -0.00035656, 0.00061594, 0.00071458, -0.00032060, -0.00017989, -0.00001236, -0.00005412, -0.00000038, -0.00002427, -0.00000372, -0.00001416, -0.00000112, -0.00001446, 0.00000342, -0.00000998, 0.00000214, 0.00000156, -0.00001149, 0.00000323, -0.00002160, 0.00001899, -0.00002173, 0.00001911, 0.00000007, -0.00000482, 0.00000815, 0.00000451, -0.00000298, -0.00000500, 0.00000629, -0.00000790, 0.00001233, -0.00000014, 0.00002235, -0.00003521, 0.00003295, -0.00001948, 0.00000985, -0.00002356, 0.00001327, -0.00000755, 0.00000264, -0.00001272, 0.00001292, -0.00000353, 0.00000127, -0.00000299, -0.00000263, -0.00000053, -0.00000205, 0.00000185, -0.00000428, -0.00000792, 0.00000822, 0.00000103, -0.00000164, 0.00000083, -0.00000312, 0.00000243, -0.00000209, -0.00001683, 0.00001690, 0.00000550, -0.00000905, -0.00000326, 0.00000089, -0.00000310, 0.00000366, 0.00001214, },
		{ -0.00001028, -0.00001053, 0.00000694, 0.00001432, -0.00000407, -0.00001359, 0.00000720, 0.00000342, -0.00000273, 0.00000594, -0.00000340, -0.00000294, 0.00000057, -0.00000501, 0.00000007, 0.00000827, -0.00000417, 0.00000749, -0.00000672, 0.00000291, -0.00000515, 0.00000964, -0.00001061, 0.00002419, -0.00002463, 0.00003157, -0.00002129, -0.00001698, 0.00001290, -0.00000147, 0.00000461, -0.00001354, 0.00001001, -0.00001622, 0.00000780, -0.00000726, 0.00000415, -0.00000665, 0.00000524, -0.00001392, 0.00001105, -0.00002890, 0.00002918, -0.00003040, 0.00001615, -0.00000183, -0.00000594, 0.00000368, -0.00000647, -0.00000657, -0.00000078, -0.00001122, -0.00000160, -0.00001239, -0.00000756, -0.00002523, -0.00001002, -0.00007025, -0.00002889, -0.00043505, 0.00035939, 0.00098284, -0.00027743, -0.00023074, -0.00001923, -0.00004986, -0.00001181, -0.00001578, -0.00001084, -0.00000925, -0.00000625, -0.00000930, -0.00000364, -0.00000225, -0.00000716, 0.00000987, -0.00001626, 0.00000713, -0.00002403, 0.00001819, -0.00001682, 0.00000959, 0.00000854, -0.00001128, 0.00001155, 0.00000012, 0.00000580, -0.00001462, 0.00001354, -0.00001541, 0.00001859, -0.00000925, 0.00003111, -0.00003627, 0.00002562, -0.00000713, -0.00000051, -0.00001424, 0.00000420, -0.00000011, -0.00000236, -0.00000770, 0.00000482, 0.00000504, -0.00000524, 0.00000287, -0.00000716, 0.00000271, -0.00000413, 0.00000322, -0.00000436, -0.00000703, 0.00000313, 0.00000583, -0.00000304, 0.00000260, -0.00000578, 0.00000364, -0.00000081, -0.00001517, 0.00000768, 0.00001523, -0.00001283, -0.00000143, -0.00000193, 0.00000117, -0.00000311, 0.00001611, },
		{ -0.00000242, -0.00001561, 0.00000529, 0.00001350, 0.00000208, -0.00001641, 0.00000436, 0.00000478, -0.00000125, 0.00000428, -0.00000123, -0.00000400, 0.00000004, -0.00000326, -0.00000313, 0.00000997, -0.00000462, 0.00000701, -0.00000508, 0.00000206, -0.00000576, 0.00001074, -0.00001203, 0.00002109, -0.00001519, 0.00001507, 0.00000304, -0.00003696, 0.00002270, -0.00000843, 0.00001217, -0.00001956, 0.00001507, -0.00001784, 0.00000637, -0.00000750, 0.00000641, -0.00000990, 0.00001003, -0.00001754, 0.00001382, -0.00003013, 0.00002599, -0.00002188, 0.00000467, 0.00000754, -0.00001101, 0.00000548, -0.00000563, -0.00000718, -0.00000073, -0.00001031, -0.00000275, -0.00001211, -0.00000363, -0.00002896, -0.00000135, -0.00007466, -0.00000208, -0.00036475, -0.00001240, 0.00121496, -0.00014997, -0.00031519, -0.00001212, -0.00005871, -0.00001150, -0.00001913, -0.00000762, -0.00001274, -0.00000458, -0.00000993, -0.00000525, -0.00000036, -0.00000963, 0.00001011, -0.00001308, 0.00000391, -0.00001960, 0.00001062, -0.00000606, -0.00000344, 0.00001867, -0.00001875, 0.00001619, -0.00000604, 0.00001497, -0.00002186, 0.00001707, -0.00001856, 0.00001964, -0.00001125, 0.00002976, -0.00002601, 0.00000880, 0.00001023, -0.00001185, -0.00000588, -0.00000328, 0.00000588, -0.00000658, -0.00000263, -0.00000295, 0.00001186, -0.00000930, 0.00000613, -0.00000907, 0.00000367, -0.00000465, 0.00000387, -0.00000462, -0.00000511, -0.00000221, 0.00000926, -0.00000355, 0.00000474, -0.00000849, 0.00000443, -0.00000031, -0.00001142, -0.00000194, 0.00002190, -0.00001239, -0.00000270, -0.00000256, 0.00000301, -0.00000564, 0.00001410, },
		{ 0.00000784, -0.00002009, 0.00000372, 0.00001108, 0.00000833, -0.00001743, 0.00000084, 0.00000521, 0.00000043, 0.00000363, -0.00000061, -0.00000324, -0.00000200, -0.00000069, -0.00000618, 0.00001081, -0.00000409, 0.00000577, -0.00000335, 0.00000163, -0.00000640, 0.00001066, -0.00001025, 0.00001355, -0.00000249, -0.00000145, 0.00002136, -0.00004506, 0.00001942, -0.00000386, 0.00000906, -0.00001499, 0.00000910, -0.00000850, -0.00000398, -0.00000055, 0.00000128, -0.00000487, 0.00000510, -0.00001017, 0.00000517, -0.00001932, 0.00001033, -0.00000181, -0.00001613, 0.00002459, -0.00002391, 0.00001650, -0.00001607, 0.00000555, -0.00001554, 0.00000673, -0.00002111, 0.00000705, -0.00002269, -0.00000385, -0.00003005, -0.00002970, -0.00005464, -0.00018554, -0.00039912, 0.00130473, 0.00011832, -0.00045467, 0.00002379, -0.00009384, 0.00001156, -0.00004312, 0.00001280, -0.00003043, 0.00000934, -0.00002174, 0.00000400, -0.00000954, -0.00000063, -0.00000119, 0.00000023, -0.00000791, -0.00000725, -0.00000420, 0.00001033, -0.00001962, 0.00003047, -0.00002747, 0.00002217, -0.00001335, 0.00002318, -0.00002572, 0.00001685, -0.00001799, 0.00001656, -0.00000808, 0.00002214, -0.00001094, -0.00000853, 0.00002288, -0.00001561, -0.00000502, -0.00000435, 0.00000659, -0.00000656, -0.00000114, -0.00000643, 0.00001326, -0.00000790, 0.00000444, -0.00000658, 0.00000105, -0.00000255, 0.00000263, -0.00000348, -0.00000439, -0.00000510, 0.00000905, -0.00000153, 0.00000530, -0.00000881, 0.00000245, 0.00000181, -0.00000890, -0.00000775, 0.00002191, -0.00000571, -0.00000815, 0.00000006, 0.00000113, -0.00000308, 0.00000680, },
		{ 0.00001650, -0.00002107, 0.00000043, 0.00000901, 0.00001247, -0.00001522, -0.00000407, 0.00000589, 0.00000063, 0.00000529, -0.00000218, -0.00000059, -0.00000521, 0.00000226, -0.00000889, 0.00001104, -0.00000338, 0.00000497, -0.00000282, 0.00000266, -0.00000782, 0.00001037, -0.00000733, 0.00000570, 0.00000726, -0.00001059, 0.00002684, -0.00003730, 0.00000243, 0.00001088, -0.00000260, -0.00000237, -0.00000472, 0.00000762, -0.00001862, 0.00000971, -0.00000824, 0.00000576, -0.00000657, 0.00000446, -0.00001045, -0.00000147, -0.00001212, 0.00002337, -0.00003981, 0.00004364, -0.00003977, 0.00003213, -0.00003291, 0.00002590, -0.00003845, 0.00003213, -0.00004813, 0.00003602, -0.00005464, 0.00003754, -0.00007912, 0.00004007, -0.00014279, 0.00001423, -0.00067607, 0.00116159, 0.00054618, -0.00063566, 0.00007726, -0.00014766, 0.00005068, -0.00008126, 0.00004410, -0.00005691, 0.00003120, -0.00004131, 0.00002124, -0.00002700, 0.00001667, -0.00002010, 0.00001925, -0.00002405, 0.00000907, -0.00002233, 0.00002841, -0.00003552, 0.00004128, -0.00003526, 0.00002737, -0.00001951, 0.00002839, -0.00002545, 0.00001365, -0.00001540, 0.00001183, -0.00000345, 0.00001385, 0.00000137, -0.00001813, 0.00002386, -0.00000759, -0.00001339, 0.00000136, 0.00000200, -0.00000204, -0.00000396, -0.00000470, 0.00000873, -0.00000122, -0.00000159, -0.00000054, -0.00000426, 0.00000156, -0.00000040, -0.00000046, -0.00000595, -0.00000434, 0.00000471, 0.00000315, 0.00000356, -0.00000560, -0.00000312, 0.00000640, -0.00000919, -0.00000812, 0.00001501, 0.00000581, -0.00001582, 0.00000431, -0.00000302, 0.00000244, -0.00000230, },
		{ 0.00001966, -0.00001645, -0.00000558, 0.00000856, 0.00001329, -0.00000958, -0.00001012, 0.00000709, -0.00000107, 0.00000894, -0.00000490, 0.00000255, -0.00000811, 0.00000435, -0.00001044, 0.00001034, -0.00000247, 0.00000466, -0.00000336, 0.00000457, -0.00000932, 0.00000958, -0.00000418, -0.00000010, 0.00001114, -0.00001031, 0.00002007, -0.00001817, -0.00002068, 0.00002717, -0.00001430, 0.00000993, -0.00001771, 0.00002135, -0.00002875, 0.00001563, -0.00001515, 0.00001477, -0.00001703, 0.00001742, -0.00002341, 0.00001334, -0.00003059, 0.00004239, -0.00005528, 0.00005419, -0.00004822, 0.00004144, -0.00004396, 0.00004010, -0.00005406, 0.00004893, -0.00006536, 0.00005464, -0.00007639, 0.00006732, -0.00011361, 0.00008918, -0.00020120, 0.00013795, -0.00074622, 0.00075881, 0.00108320, -0.00079421, 0.00010364, -0.00018763, 0.00007896, -0.00011032, 0.00006612, -0.00007509, 0.00004642, -0.00005582, 0.00003473, -0.00004158, 0.00003128, -0.00003569, 0.00003362, -0.00003525, 0.00002106, -0.00003599, 0.00004094, -0.00004492, 0.00004592, -0.00003753, 0.00002756, -0.00002064, 0.00002771, -0.00001996, 0.00000778, -0.00001173, 0.00000695, 0.00000024, 0.00000847, 0.00000693, -0.00001719, 0.00001326, 0.00000897, -0.00002600, 0.00000871, -0.00000347, 0.00000342, -0.00000816, -0.00000054, 0.00000151, 0.00000709, -0.00000830, 0.00000562, -0.00000923, 0.00000527, -0.00000348, 0.00000323, -0.00000887, -0.00000114, -0.00000186, 0.00000855, 0.00000076, 0.00000001, -0.00001068, 0.00001187, -0.00001111, -0.00000472, 0.00000453, 0.00001770, -0.00002153, 0.00000678, -0.00000625, 0.00000713, -0.00000874, },
		{ 0.00001570, -0.00000654, -0.00001337, 0.00000930, 0.00001135, -0.00000198, -0.00001576, 0.00000788, -0.00000352, 0.00001261, -0.00000640, 0.00000387, -0.00000868, 0.00000400, -0.00000967, 0.00000783, -0.00000052, 0.00000377, -0.00000338, 0.00000535, -0.00000884, 0.00000677, -0.00000007, -0.00000425, 0.00001056, -0.00000444, 0.00000830, 0.00000166, -0.00003782, 0.00003369, -0.00001602, 0.00001261, -0.00002073, 0.00002367, -0.00002634, 0.00001040, -0.00001282, 0.00001479, -0.00001796, 0.00001958, -0.00002426, 0.00001544, -0.00003497, 0.00004495, -0.00005263, 0.00004670, -0.00003940, 0.00003345, -0.00003684, 0.00003423, -0.00004713, 0.00004076, -0.00005522, 0.00004366, -0.00006574, 0.00005896, -0.00010147, 0.00007837, -0.00017935, 0.00012658, -0.00058557, 0.00016193, 0.00160482, -0.00082569, 0.00003836, -0.00016897, 0.00006067, -0.00010060, 0.00005401, -0.00006426, 0.00003720, -0.00004931, 0.00002971, -0.00003924, 0.00002974, -0.00003527, 0.00003191, -0.00003166, 0.00001992, -0.00003699, 0.00004046, -0.00004140, 0.00003886, -0.00002926, 0.00001820, -0.00001286, 0.00001849, -0.00000813, -0.00000124, -0.00000634, 0.00000126, 0.00000343, 0.00000552, 0.00000725, -0.00000974, -0.00000167, 0.00002469, -0.00003341, 0.00000968, -0.00000343, 0.00000457, -0.00000907, 0.00000131, -0.00000327, 0.00001184, -0.00001092, 0.00000769, -0.00001031, 0.00000568, -0.00000423, 0.00000536, -0.00001078, 0.00000164, -0.00000721, 0.00001150, -0.00000059, 0.00000530, -0.00001691, 0.00001501, -0.00001169, -0.00000132, -0.00000441, 0.00002462, -0.00002109, 0.00000439, -0.00000571, 0.00000783, -0.00000945, },
		{ 0.00000631, 0.00000555, -0.00002009, 0.00000932, 0.00000869, 0.00000498, -0.00001874, 0.00000667, -0.00000481, 0.00001376, -0.00000427, 0.00000148, -0.00000557, 0.00000032, -0.00000596, 0.00000292, 0.00000334, 0.00000095, -0.00000104, 0.00000285, -0.00000443, 0.00000043, 0.00000627, -0.00000864, 0.00000943, 0.00000044, 0.00000078, 0.00001148, -0.00003923, 0.00002320, -0.00000262, 0.00000152, -0.00000998, 0.00001140, -0.00000940, -0.00000710, 0.00000030, 0.00000332, -0.00000611, 0.00000741, -0.00000963, 0.00000162, -0.00002193, 0.00002794, -0.00002939, 0.00001906, -0.00001077, 0.00000479, -0.00000735, 0.00000358, -0.00001281, 0.00000281, -0.00001294, -0.00000203, -0.00001645, 0.00000487, -0.00003427, -0.00000065, -0.00007053, -0.00001243, -0.00025934, -0.00048004, 0.00194344, -0.00062281, -0.00016977, -0.00006131, -0.00002685, -0.00003446, -0.00000604, -0.00001353, -0.00000579, -0.00001319, -0.00000204, -0.00001208, 0.00000461, -0.00001221, 0.00000878, -0.00000907, 0.00000191, -0.00002164, 0.00002368, -0.00002235, 0.00001791, -0.00000843, -0.00000244, 0.00000503, 0.00000041, 0.00000972, -0.00001360, 0.00000201, -0.00000714, 0.00000851, 0.00000167, 0.00000757, -0.00000369, -0.00001085, 0.00002928, -0.00002727, -0.00000157, 0.00000611, -0.00000179, -0.00000357, -0.00000254, -0.00000188, 0.00000949, -0.00000651, 0.00000334, -0.00000567, 0.00000134, -0.00000121, 0.00000421, -0.00000950, 0.00000119, -0.00000823, 0.00000947, 0.00000154, 0.00000768, -0.00001870, 0.00001298, -0.00000825, -0.00000136, -0.00000770, 0.00002328, -0.00001292, -0.00000340, -0.00000094, 0.00000394, -0.00000446, },
		{ -0.00000419, 0.00001520, -0.00002241, 0.00000652, 0.00000747, 0.00000893, -0.00001749, 0.00000253, -0.00000363, 0.00001089, 0.00000232, -0.00000464, 0.00000070, -0.00000592, -0.00000008, -0.00000385, 0.00000896, -0.00000406, 0.00000423, -0.00000346, 0.00000402, -0.00000917, 0.00001471, -0.00001434, 0.00001074, -0.00000044, 0.00000304, 0.00000673, -0.00002340, -0.00000235, 0.00002198, -0.00001882, 0.00000978, -0.00001015, 0.00001575, -0.00003029, 0.00001857, -0.00001490, 0.00001397, -0.00001422, 0.00001489, -0.00002207, 0.00000221, -0.00000169, 0.00000656, -0.00002037, 0.00002942, -0.00003636, 0.00003596, -0.00004246, 0.00003819, -0.00005265, 0.00004767, -0.00006721, 0.00005482, -0.00007584, 0.00006439, -0.00011694, 0.00008208, -0.00020798, 0.00010093, -0.00098238, 0.00194866, -0.00012949, -0.00051905, 0.00012313, -0.00017184, 0.00007679, -0.00010302, 0.00006672, -0.00007356, 0.00004501, -0.00005412, 0.00003423, -0.00003873, 0.00002798, -0.00002990, 0.00002660, -0.00002774, 0.00000555, -0.00000517, 0.00000797, -0.00001289, 0.00002120, -0.00003065, 0.00002924, -0.00002266, 0.00003026, -0.00002751, 0.00001308, -0.00001888, 0.00001698, -0.00000585, 0.00001250, -0.00000535, -0.00000765, 0.00001797, -0.00000596, -0.00002404, 0.00002309, -0.00001377, 0.00000706, -0.00001148, 0.00000540, 0.00000065, 0.00000380, -0.00000602, 0.00000321, -0.00000642, 0.00000469, 0.00000003, -0.00000468, -0.00000327, -0.00000418, 0.00000237, 0.00000708, 0.00000650, -0.00001505, 0.00000527, -0.00000044, -0.00000574, -0.00000450, 0.00001433, 0.00000066, -0.00001380, 0.00000565, -0.00000228, 0.00000319, },
		{ -0.00001110, 0.00001858, -0.00001842, 0.00000020, 0.00000852, 0.00000916, -0.00001223, -0.00000386, -0.00000031, 0.00000458, 0.00001175, -0.00001214, 0.00000755, -0.00001215, 0.00000564, -0.00001046, 0.00001458, -0.00000966, 0.00001074, -0.00001146, 0.00001379, -0.00001905, 0.00002269, -0.00001980, 0.00001412, -0.00000684, 0.00001352, -0.00000817, 0.00000138, -0.00003176, 0.00004574, -0.00003673, 0.00002719, -0.00002928, 0.00003693, -0.00004746, 0.00003143, -0.00002982, 0.00003179, -0.00003388, 0.00003689, -0.00004266, 0.00002405, -0.00002958, 0.00003978, -0.00005548, 0.00006467, -0.00007270, 0.00007438, -0.00008325, 0.00008298, -0.00010038, 0.00009894, -0.00012164, 0.00011462, -0.00014471, 0.00014850, -0.00021382, 0.00020602, -0.00035932, 0.00035120, -0.00119384, 0.00155322, 0.00061339, -0.00093020, 0.00031720, -0.00032135, 0.00018887, -0.00019883, 0.00014358, -0.00013792, 0.00010085, -0.00010477, 0.00007982, -0.00008172, 0.00006763, -0.00006730, 0.00005986, -0.00005531, 0.00003229, -0.00003451, 0.00003850, -0.00004320, 0.00004991, -0.00005710, 0.00005096, -0.00004265, 0.00004677, -0.00003829, 0.00002362, -0.00003134, 0.00002684, -0.00001600, 0.00002200, -0.00001490, 0.00000661, -0.00000482, 0.00002301, -0.00004886, 0.00003912, -0.00002434, 0.00001705, -0.00002059, 0.00001390, -0.00000978, 0.00001490, -0.00001545, 0.00001179, -0.00001365, 0.00001030, -0.00000474, 0.00000165, -0.00000985, 0.00000271, -0.00000693, 0.00001339, 0.00000354, -0.00000776, -0.00000561, 0.00000913, -0.00001221, 0.00000240, 0.00000213, 0.00001431, -0.00002192, 0.00001005, -0.00000707, 0.00000927, },
		{ -0.00001173, 0.00001464, -0.00000911, -0.00000802, 0.00001066, 0.00000701, -0.00000517, -0.00001016, 0.00000316, -0.00000264, 0.00002048, -0.00001713, 0.00001129, -0.00001503, 0.00000821, -0.00001411, 0.00001746, -0.00001293, 0.00001511, -0.00001711, 0.00002024, -0.00002446, 0.00002586, -0.00002129, 0.00001584, -0.00001365, 0.00002455, -0.00002222, 0.00002107, -0.00004994, 0.00005458, -0.00003959, 0.00003038, -0.00003420, 0.00004243, -0.00004782, 0.00002917, -0.00003157, 0.00003645, -0.00003958, 0.00004353, -0.00004695, 0.00002996, -0.00004112, 0.00005462, -0.00007001, 0.00007804, -0.00008603, 0.00008793, -0.00009682, 0.00009752, -0.00011434, 0.00011283, -0.00013493, 0.00012942, -0.00016328, 0.00017290, -0.00023757, 0.00023638, -0.00038477, 0.00039595, -0.00105831, 0.00081491, 0.00145557, -0.00125467, 0.00041434, -0.00039726, 0.00023954, -0.00024199, 0.00017387, -0.00016235, 0.00012255, -0.00012552, 0.00009867, -0.00010017, 0.00008422, -0.00008261, 0.00007205, -0.00006436, 0.00004360, -0.00005007, 0.00005558, -0.00006011, 0.00006546, -0.00007014, 0.00005937, -0.00004988, 0.00005124, -0.00003964, 0.00002798, -0.00003874, 0.00003243, -0.00002342, 0.00003062, -0.00002572, 0.00002285, -0.00002703, 0.00004593, -0.00006295, 0.00004326, -0.00002480, 0.00001912, -0.00002322, 0.00001709, -0.00001507, 0.00002019, -0.00001893, 0.00001474, -0.00001578, 0.00001173, -0.00000667, 0.00000610, -0.00001491, 0.00000826, -0.00001388, 0.00001647, 0.00000206, -0.00000044, -0.00001535, 0.00001621, -0.00001675, 0.00000827, -0.00000748, 0.00002229, -0.00002336, 0.00000905, -0.00000753, 0.00001070, },
		{ -0.00000671, 0.00000577, 0.00000185, -0.00001475, 0.00001138, 0.00000509, 0.00000058, -0.00001354, 0.00000432, -0.00000767, 0.00002474, -0.00001612, 0.00000907, -0.00001224, 0.00000559, -0.00001276, 0.00001533, -0.00001116, 0.00001408, -0.00001652, 0.00001913, -0.00002132, 0.00002053, -0.00001525, 0.00001149, -0.00001439, 0.00002708, -0.00002420, 0.00002355, -0.00004609, 0.00004035, -0.00002135, 0.00001420, -0.00001994, 0.00002765, -0.00002785, 0.00000888, -0.00001647, 0.00002308, -0.00002560, 0.00002874, -0.00002894, 0.00001374, -0.00002937, 0.00004348, -0.00005608, 0.00006120, -0.00006702, 0.00006613, -0.00007173, 0.00006964, -0.00008186, 0.00007622, -0.00009332, 0.00008415, -0.00011426, 0.00011774, -0.00016628, 0.00014995, -0.00026160, 0.00022855, -0.00064055, -0.00008806, 0.00217129, -0.00132060, 0.00031176, -0.00033233, 0.00017865, -0.00019317, 0.00012603, -0.00012092, 0.00008766, -0.00009616, 0.00007222, -0.00007701, 0.00006211, -0.00006189, 0.00005114, -0.00004443, 0.00002963, -0.00004204, 0.00004955, -0.00005438, 0.00005891, -0.00006127, 0.00004683, -0.00003779, 0.00003831, -0.00002694, 0.00002104, -0.00003506, 0.00002732, -0.00002130, 0.00003063, -0.00002833, 0.00002925, -0.00003524, 0.00004989, -0.00005599, 0.00002829, -0.00001011, 0.00000908, -0.00001525, 0.00001056, -0.00001061, 0.00001539, -0.00001283, 0.00000905, -0.00001026, 0.00000676, -0.00000346, 0.00000596, -0.00001519, 0.00000872, -0.00001469, 0.00001329, 0.00000463, 0.00000366, -0.00002005, 0.00001723, -0.00001604, 0.00000922, -0.00001027, 0.00002142, -0.00001673, 0.00000231, -0.00000343, 0.00000727, },
		{ 0.00000059, -0.00000352, 0.00000979, -0.00001665, 0.00000855, 0.00000554, 0.00000273, -0.00001234, 0.00000184, -0.00000864, 0.00002254, -0.00000802, 0.00000071, -0.00000404, -0.00000186, -0.00000654, 0.00000796, -0.00000372, 0.00000661, -0.00000843, 0.00000932, -0.00000899, 0.00000645, -0.00000121, -0.00000051, -0.00000570, 0.00001646, -0.00000933, 0.00000562, -0.00002020, 0.00000632, 0.00001292, -0.00001590, 0.00000810, -0.00000156, 0.00000578, -0.00002280, 0.00000998, -0.00000380, 0.00000376, -0.00000279, 0.00000606, -0.00001917, 0.00000057, 0.00001143, -0.00001906, 0.00001966, -0.00002110, 0.00001462, -0.00001434, 0.00000681, -0.00001189, -0.00000036, -0.00000892, -0.00000765, -0.00001272, 0.00000102, -0.00002329, -0.00002107, -0.00003731, -0.00006779, -0.00010308, -0.00091152, 0.00253257, -0.00100261, -0.00003508, -0.00011088, -0.00000039, -0.00005043, 0.00000087, -0.00001569, -0.00000186, -0.00001817, 0.00000170, -0.00001343, 0.00000283, -0.00000722, -0.00000020, 0.00000178, -0.00000775, -0.00001126, 0.00002066, -0.00002602, 0.00003007, -0.00003040, 0.00001374, -0.00000727, 0.00000917, -0.00000089, 0.00000197, -0.00001810, 0.00000869, -0.00000616, 0.00001739, -0.00001657, 0.00001847, -0.00002243, 0.00003030, -0.00002686, -0.00000413, 0.00001702, -0.00001074, 0.00000181, -0.00000483, 0.00000301, 0.00000138, 0.00000156, -0.00000383, 0.00000148, -0.00000347, 0.00000433, 0.00000101, -0.00000991, 0.00000300, -0.00000850, 0.00000380, 0.00001124, 0.00000360, -0.00001841, 0.00001139, -0.00000955, 0.00000439, -0.00000568, 0.00001273, -0.00000463, -0.00000719, 0.00000263, 0.00000154, },
		{ 0.00000614, -0.00000893, 0.00001149, -0.00001246, 0.00000203, 0.00000858, 0.00000120, -0.00000726, -0.00000337, -0.00000606, 0.00001486, 0.00000494, -0.00001074, 0.00000635, -0.00001111, 0.00000183, -0.00000220, 0.00000706, -0.00000494, 0.00000457, -0.00000599, 0.00000862, -0.00001209, 0.00001677, -0.00001691, 0.00000989, -0.00000432, 0.00001748, -0.00002467, 0.00001600, -0.00003331, 0.00004850, -0.00004576, 0.00003623, -0.00003135, 0.00003883, -0.00005216, 0.00003526, -0.00003203, 0.00003568, -0.00003722, 0.00004328, -0.00005366, 0.00003353, -0.00002574, 0.00002427, -0.00002892, 0.00003308, -0.00004631, 0.00005282, -0.00006581, 0.00006753, -0.00008585, 0.00008408, -0.00010829, 0.00009914, -0.00012818, 0.00013218, -0.00020304, 0.00019279, -0.00035872, 0.00035745, -0.00143034, 0.00239480, -0.00028683, -0.00057206, 0.00020729, -0.00024532, 0.00014034, -0.00016100, 0.00011770, -0.00011510, 0.00008123, -0.00008838, 0.00006797, -0.00007244, 0.00006138, -0.00006313, 0.00005720, -0.00005375, 0.00002945, -0.00001942, 0.00001411, -0.00001083, 0.00001249, -0.00002986, 0.00003189, -0.00002709, 0.00003091, -0.00002365, 0.00000778, -0.00001947, 0.00001856, -0.00000648, 0.00000766, -0.00000735, 0.00000735, -0.00000556, 0.00001420, -0.00004262, 0.00004601, -0.00003152, 0.00002063, -0.00002267, 0.00001972, -0.00001568, 0.00001811, -0.00001814, 0.00001427, -0.00001452, 0.00001317, -0.00000599, -0.00000141, -0.00000655, 0.00000183, -0.00000858, 0.00001890, 0.00000130, -0.00001229, 0.00000128, 0.00000001, -0.00000370, 0.00000307, 0.00000089, 0.00000757, -0.00001472, 0.00000680, -0.00000288, },
		{ 0.00000750, -0.00000864, 0.00000690, -0.00000403, -0.00000576, 0.00001216, -0.00000180, -0.00000117, -0.00000842, -0.00000256, 0.00000517, 0.00001802, -0.00002017, 0.00001405, -0.00001767, 0.00000819, -0.00001098, 0.00001677, -0.00001570, 0.00001691, -0.00002037, 0.00002426, -0.00002758, 0.00003136, -0.00003064, 0.00002483, -0.00002584, 0.00004389, -0.00005178, 0.00004450, -0.00005997, 0.00006805, -0.00005960, 0.00004947, -0.00004674, 0.00005630, -0.00006508, 0.00004607, -0.00004782, 0.00005510, -0.00005820, 0.00006545, -0.00007220, 0.00005165, -0.00004917, 0.00005367, -0.00006293, 0.00007218, -0.00009088, 0.00010111, -0.00011667, 0.00012193, -0.00014278, 0.00014497, -0.00017321, 0.00017161, -0.00021261, 0.00023278, -0.00031555, 0.00033050, -0.00051899, 0.00058715, -0.00151991, 0.00175909, 0.00069999, -0.00114375, 0.00049633, -0.00045789, 0.00029865, -0.00029195, 0.00022178, -0.00020275, 0.00015869, -0.00015893, 0.00013141, -0.00013066, 0.00011336, -0.00010972, 0.00009685, -0.00008637, 0.00006020, -0.00005208, 0.00004825, -0.00004674, 0.00005066, -0.00006755, 0.00006409, -0.00005625, 0.00005607, -0.00004515, 0.00003242, -0.00004681, 0.00004268, -0.00003115, 0.00003426, -0.00003677, 0.00004021, -0.00004122, 0.00004965, -0.00007074, 0.00006298, -0.00004199, 0.00003161, -0.00003412, 0.00003089, -0.00002725, 0.00002867, -0.00002652, 0.00002162, -0.00002076, 0.00001823, -0.00001072, 0.00000599, -0.00001528, 0.00001113, -0.00001847, 0.00002310, 0.00000035, -0.00000554, -0.00000855, 0.00000811, -0.00001072, 0.00001100, -0.00000829, 0.00001447, -0.00001637, 0.00000637, -0.00000360, },
		{ 0.00000509, -0.00000407, -0.00000083, 0.00000439, -0.00001095, 0.00001320, -0.00000309, 0.00000236, -0.00001008, -0.00000125, -0.00000243, 0.00002625, -0.00002283, 0.00001497, -0.00001788, 0.00000898, -0.00001453, 0.00002099, -0.00002057, 0.00002272, -0.00002715, 0.00003079, -0.00003286, 0.00003551, -0.00003442, 0.00003067, -0.00003731, 0.00005659, -0.00006067, 0.00005010, -0.00006022, 0.00006075, -0.00004853, 0.00003966, -0.00003955, 0.00005020, -0.00005429, 0.00003539, -0.00004302, 0.00005238, -0.00005506, 0.00006146, -0.00006373, 0.00004358, -0.00004645, 0.00005543, -0.00006732, 0.00007947, -0.00010019, 0.00010951, -0.00012295, 0.00012687, -0.00014520, 0.00014608, -0.00017253, 0.00017152, -0.00021437, 0.00023509, -0.00030927, 0.00032002, -0.00048612, 0.00053207, -0.00119982, 0.00078277, 0.00171550, -0.00153713, 0.00060936, -0.00053350, 0.00034356, -0.00032676, 0.00024283, -0.00021950, 0.00017451, -0.00017425, 0.00014440, -0.00014218, 0.00012182, -0.00011588, 0.00009956, -0.00008685, 0.00006356, -0.00006041, 0.00005990, -0.00006147, 0.00006801, -0.00008360, 0.00007468, -0.00006504, 0.00006280, -0.00005156, 0.00004436, -0.00006112, 0.00005384, -0.00004406, 0.00004986, -0.00005480, 0.00005928, -0.00005896, 0.00006298, -0.00007499, 0.00005785, -0.00003452, 0.00002809, -0.00003262, 0.00002991, -0.00002679, 0.00002732, -0.00002387, 0.00001912, -0.00001838, 0.00001604, -0.00000969, 0.00000835, -0.00001873, 0.00001456, -0.00002124, 0.00002027, 0.00000364, -0.00000169, -0.00001401, 0.00001100, -0.00001310, 0.00001411, -0.00001081, 0.00001344, -0.00001132, 0.00000148, -0.00000087, },
		{ 0.00000158, 0.00000125, -0.00000716, 0.00000833, -0.00001040, 0.00000961, -0.00000057, 0.00000121, -0.00000674, -0.00000375, -0.00000544, 0.00002686, -0.00001681, 0.00000805, -0.00001103, 0.00000334, -0.00001145, 0.00001770, -0.00001697, 0.00001894, -0.00002293, 0.00002485, -0.00002490, 0.00002638, -0.00002502, 0.00002293, -0.00003257, 0.00004849, -0.00004478, 0.00002840, -0.00003297, 0.00002845, -0.00001581, 0.00001015, -0.00001281, 0.00002364, -0.00002344, 0.00000647, -0.00001945, 0.00002820, -0.00002820, 0.00003205, -0.00002951, 0.00001053, -0.00001801, 0.00002917, -0.00004106, 0.00005299, -0.00007132, 0.00007470, -0.00008158, 0.00007985, -0.00009138, 0.00008639, -0.00010570, 0.00009842, -0.00013303, 0.00013987, -0.00018829, 0.00017196, -0.00028358, 0.00025177, -0.00062153, -0.00026967, 0.00247698, -0.00155907, 0.00044317, -0.00041189, 0.00023406, -0.00023577, 0.00015856, -0.00014760, 0.00011334, -0.00012057, 0.00009463, -0.00009618, 0.00007754, -0.00007400, 0.00005917, -0.00004979, 0.00003389, -0.00003811, 0.00004226, -0.00004784, 0.00005701, -0.00007068, 0.00005717, -0.00004786, 0.00004605, -0.00003749, 0.00003684, -0.00005460, 0.00004403, -0.00003686, 0.00004518, -0.00005101, 0.00005363, -0.00004896, 0.00004732, -0.00005214, 0.00003020, -0.00000983, 0.00001042, -0.00001780, 0.00001591, -0.00001345, 0.00001364, -0.00001025, 0.00000705, -0.00000762, 0.00000661, -0.00000232, 0.00000441, -0.00001511, 0.00001020, -0.00001550, 0.00001006, 0.00001137, -0.00000187, -0.00001359, 0.00000767, -0.00001008, 0.00001144, -0.00000635, 0.00000585, -0.00000236, -0.00000488, 0.00000269, },
		{ -0.00000002, 0.00000391, -0.00000865, 0.00000562, -0.00000379, 0.00000197, 0.00000528, -0.00000400, 0.00000039, -0.00000897, -0.00000428, 0.00002065, -0.00000416, -0.00000393, 0.00000005, -0.00000626, -0.00000368, 0.00000846, -0.00000631, 0.00000703, -0.00000947, 0.00000879, -0.00000676, 0.00000732, -0.00000550, 0.00000401, -0.00001361, 0.00002252, -0.00000957, -0.00001166, 0.00000948, -0.00001469, 0.00002431, -0.00002562, 0.00002070, -0.00001064, 0.00001460, -0.00002854, 0.00001199, -0.00000681, 0.00001104, -0.00001041, 0.00001714, -0.00003400, 0.00002295, -0.00001193, 0.00000249, 0.00000636, -0.00001854, 0.00001255, -0.00001069, 0.00000160, -0.00000471, -0.00000795, -0.00000177, -0.00001537, -0.00000526, -0.00000929, -0.00000663, -0.00004451, -0.00000450, -0.00011456, -0.00000573, -0.00111732, 0.00275934, -0.00111732, -0.00000573, -0.00011456, -0.00000450, -0.00004451, -0.00000663, -0.00000929, -0.00000526, -0.00001537, -0.00000177, -0.00000795, -0.00000471, 0.00000160, -0.00001069, 0.00001255, -0.00001854, 0.00000636, 0.00000249, -0.00001193, 0.00002295, -0.00003400, 0.00001714, -0.00001041, 0.00001104, -0.00000681, 0.00001199, -0.00002854, 0.00001460, -0.00001064, 0.00002070, -0.00002562, 0.00002431, -0.00001469, 0.00000948, -0.00001166, -0.00000957, 0.00002252, -0.00001361, 0.00000401, -0.00000550, 0.00000732, -0.00000676, 0.00000879, -0.00000947, 0.00000703, -0.00000631, 0.00000846, -0.00000368, -0.00000626, 0.00000005, -0.00000393, -0.00000416, 0.00002065, -0.00000428, -0.00000897, 0.00000039, -0.00000400, 0.00000528, 0.00000197, -0.00000379, 0.00000562, -0.00000865, 0.00000391, },
		{ 0.00000158, 0.00000269, -0.00000488, -0.00000236, 0.00000585, -0.00000635, 0.00001144, -0.00001008, 0.00000767, -0.00001359, -0.00000187, 0.00001137, 0.00001006, -0.00001550, 0.00001020, -0.00001511, 0.00000441, -0.00000232, 0.00000661, -0.00000762, 0.00000705, -0.00001025, 0.00001364, -0.00001345, 0.00001591, -0.00001780, 0.00001042, -0.00000983, 0.00003020, -0.00005214, 0.00004732, -0.00004896, 0.00005363, -0.00005101, 0.00004518, -0.00003686, 0.00004403, -0.00005460, 0.00003684, -0.00003749, 0.00004605, -0.00004786, 0.00005717, -0.00007068, 0.00005701, -0.00004784, 0.00004226, -0.00003811, 0.00003389, -0.00004979, 0.00005917, -0.00007400, 0.00007754, -0.00009618, 0.00009463, -0.00012057, 0.00011334, -0.00014760, 0.00015856, -0.00023577, 0.00023406, -0.00041189, 0.00044317, -0.00155907, 0.00247698, -0.00026967, -0.00062153, 0.00025177, -0.00028358, 0.00017196, -0.00018829, 0.00013987, -0.00013303, 0.00009842, -0.00010570, 0.00008639, -0.00009138, 0.00007985, -0.00008158, 0.00007470, -0.00007132, 0.00005299, -0.00004106, 0.00002917, -0.00001801, 0.00001053, -0.00002951, 0.00003205, -0.00002820, 0.00002820, -0.00001945, 0.00000647, -0.00002344, 0.00002364, -0.00001281, 0.00001015, -0.00001581, 0.00002845, -0.00003297, 0.00002840, -0.00004478, 0.00004849, -0.00003257, 0.00002293, -0.00002502, 0.00002638, -0.00002490, 0.00002485, -0.00002293, 0.00001894, -0.00001697, 0.00001770, -0.00001145, 0.00000334, -0.00001103, 0.00000805, -0.00001681, 0.00002686, -0.00000544, -0.00000375, -0.00000674, 0.00000121, -0.00000057, 0.00000961, -0.00001040, 0.00000833, -0.00000716, 0.00000125, },
		{ 0.00000509, -0.00000087, 0.00000148, -0.00001132, 0.00001344, -0.00001081, 0.00001411, -0.00001310, 0.00001100, -0.00001401, -0.00000169, 0.00000364, 0.00002027, -0.00002124, 0.00001456, -0.00001873, 0.00000835, -0.00000969, 0.00001604, -0.00001838, 0.00001912, -0.00002387, 0.00002732, -0.00002679, 0.00002991, -0.00003262, 0.00002809, -0.00003452, 0.00005785, -0.00007499, 0.00006298, -0.00005896, 0.00005928, -0.00005480, 0.00004986, -0.00004406, 0.00005384, -0.00006112, 0.00004436, -0.00005156, 0.00006280, -0.00006504, 0.00007468, -0.00008360, 0.00006801, -0.00006147, 0.00005990, -0.00006041, 0.00006356, -0.00008685, 0.00009956, -0.00011588, 0.00012182, -0.00014218, 0.00014440, -0.00017425, 0.00017451, -0.00021950, 0.00024283, -0.00032676, 0.00034356, -0.00053350, 0.00060936, -0.00153713, 0.00171550, 0.00078277, -0.00119982, 0.00053207, -0.00048612, 0.00032002, -0.00030927, 0.00023509, -0.00021437, 0.00017152, -0.00017253, 0.00014608, -0.00014520, 0.00012687, -0.00012295, 0.00010951, -0.00010019, 0.00007947, -0.00006732, 0.00005543, -0.00004645, 0.00004358, -0.00006373, 0.00006146, -0.00005506, 0.00005238, -0.00004302, 0.00003539, -0.00005429, 0.00005020, -0.00003955, 0.00003966, -0.00004853, 0.00006075, -0.00006022, 0.00005010, -0.00006067, 0.00005659, -0.00003731, 0.00003067, -0.00003442, 0.00003551, -0.00003286, 0.00003079, -0.00002715, 0.00002272, -0.00002057, 0.00002099, -0.00001453, 0.00000898, -0.00001788, 0.00001497, -0.00002283, 0.00002625, -0.00000243, -0.00000125, -0.00001008, 0.00000236, -0.00000309, 0.00001320, -0.00001095, 0.00000439, -0.00000083, -0.00000407, },
		{ 0.00000750, -0.00000360, 0.00000637, -0.00001637, 0.00001447, -0.00000829, 0.00001100, -0.00001072, 0.00000811, -0.00000855, -0.00000554, 0.00000035, 0.00002310, -0.00001847, 0.00001113, -0.00001528, 0.00000599, -0.00001072, 0.00001823, -0.00002076, 0.00002162, -0.00002652, 0.00002867, -0.00002725, 0.00003089, -0.00003412, 0.00003161, -0.00004199, 0.00006298, -0.00007074, 0.00004965, -0.00004122, 0.00004021, -0.00003677, 0.00003426, -0.00003115, 0.00004268, -0.00004681, 0.00003242, -0.00004515, 0.00005607, -0.00005625, 0.00006409, -0.00006755, 0.00005066, -0.00004674, 0.00004825, -0.00005208, 0.00006020, -0.00008637, 0.00009685, -0.00010972, 0.00011336, -0.00013066, 0.00013141, -0.00015893, 0.00015869, -0.00020275, 0.00022178, -0.00029195, 0.00029865, -0.00045789, 0.00049633, -0.00114375, 0.00069999, 0.00175909, -0.00151991, 0.00058715, -0.00051899, 0.00033050, -0.00031555, 0.00023278, -0.00021261, 0.00017161, -0.00017321, 0.00014497, -0.00014278, 0.00012193, -0.00011667, 0.00010111, -0.00009088, 0.00007218, -0.00006293, 0.00005367, -0.00004917, 0.00005165, -0.00007220, 0.00006545, -0.00005820, 0.00005510, -0.00004782, 0.00004607, -0.00006508, 0.00005630, -0.00004674, 0.00004947, -0.00005960, 0.00006805, -0.00005997, 0.00004450, -0.00005178, 0.00004389, -0.00002584, 0.00002483, -0.00003064, 0.00003136, -0.00002758, 0.00002426, -0.00002037, 0.00001691, -0.00001570, 0.00001677, -0.00001098, 0.00000819, -0.00001767, 0.00001405, -0.00002017, 0.00001802, 0.00000517, -0.00000256, -0.00000842, -0.00000117, -0.00000180, 0.00001216, -0.00000576, -0.00000403, 0.00000690, -0.00000864, },
		{ 0.00000614, -0.00000288, 0.00000680, -0.00001472, 0.00000757, 0.00000089, 0.00000307, -0.00000370, 0.00000001, 0.00000128, -0.00001229, 0.00000130, 0.00001890, -0.00000858, 0.00000183, -0.00000655, -0.00000141, -0.00000599, 0.00001317, -0.00001452, 0.00001427, -0.00001814, 0.00001811, -0.00001568, 0.00001972, -0.00002267, 0.00002063, -0.00003152, 0.00004601, -0.00004262, 0.00001420, -0.00000556, 0.00000735, -0.00000735, 0.00000766, -0.00000648, 0.00001856, -0.00001947, 0.00000778, -0.00002365, 0.00003091, -0.00002709, 0.00003189, -0.00002986, 0.00001249, -0.00001083, 0.00001411, -0.00001942, 0.00002945, -0.00005375, 0.00005720, -0.00006313, 0.00006138, -0.00007244, 0.00006797, -0.00008838, 0.00008123, -0.00011510, 0.00011770, -0.00016100, 0.00014034, -0.00024532, 0.00020729, -0.00057206, -0.00028683, 0.00239480, -0.00143034, 0.00035745, -0.00035872, 0.00019279, -0.00020304, 0.00013218, -0.00012818, 0.00009914, -0.00010829, 0.00008408, -0.00008585, 0.00006753, -0.00006581, 0.00005282, -0.00004631, 0.00003308, -0.00002892, 0.00002427, -0.00002574, 0.00003353, -0.00005366, 0.00004328, -0.00003722, 0.00003568, -0.00003203, 0.00003526, -0.00005216, 0.00003883, -0.00003135, 0.00003623, -0.00004576, 0.00004850, -0.00003331, 0.00001600, -0.00002467, 0.00001748, -0.00000432, 0.00000989, -0.00001691, 0.00001677, -0.00001209, 0.00000862, -0.00000599, 0.00000457, -0.00000494, 0.00000706, -0.00000220, 0.00000183, -0.00001111, 0.00000635, -0.00001074, 0.00000494, 0.00001486, -0.00000606, -0.00000337, -0.00000726, 0.00000120, 0.00000858, 0.00000203, -0.00001246, 0.00001149, -0.00000893, },
		{ 0.00000059, 0.00000154, 0.00000263, -0.00000719, -0.00000463, 0.00001273, -0.00000568, 0.00000439, -0.00000955, 0.00001139, -0.00001841, 0.00000360, 0.00001124, 0.00000380, -0.00000850, 0.00000300, -0.00000991, 0.00000101, 0.00000433, -0.00000347, 0.00000148, -0.00000383, 0.00000156, 0.00000138, 0.00000301, -0.00000483, 0.00000181, -0.00001074, 0.00001702, -0.00000413, -0.00002686, 0.00003030, -0.00002243, 0.00001847, -0.00001657, 0.00001739, -0.00000616, 0.00000869, -0.00001810, 0.00000197, -0.00000089, 0.00000917, -0.00000727, 0.00001374, -0.00003040, 0.00003007, -0.00002602, 0.00002066, -0.00001126, -0.00000775, 0.00000178, -0.00000020, -0.00000722, 0.00000283, -0.00001343, 0.00000170, -0.00001817, -0.00000186, -0.00001569, 0.00000087, -0.00005043, -0.00000039, -0.00011088, -0.00003508, -0.00100261, 0.00253257, -0.00091152, -0.00010308, -0.00006779, -0.00003731, -0.00002107, -0.00002329, 0.00000102, -0.00001272, -0.00000765, -0.00000892, -0.00000036, -0.00001189, 0.00000681, -0.00001434, 0.00001462, -0.00002110, 0.00001966, -0.00001906, 0.00001143, 0.00000057, -0.00001917, 0.00000606, -0.00000279, 0.00000376, -0.00000380, 0.00000998, -0.00002280, 0.00000578, -0.00000156, 0.00000810, -0.00001590, 0.00001292, 0.00000632, -0.00002020, 0.00000562, -0.00000933, 0.00001646, -0.00000570, -0.00000051, -0.00000121, 0.00000645, -0.00000899, 0.00000932, -0.00000843, 0.00000661, -0.00000372, 0.00000796, -0.00000654, -0.00000186, -0.00000404, 0.00000071, -0.00000802, 0.00002254, -0.00000864, 0.00000184, -0.00001234, 0.00000273, 0.00000554, 0.00000855, -0.00001665, 0.00000979, -0.00000352, },
		{ -0.00000671, 0.00000727, -0.00000343, 0.00000231, -0.00001673, 0.00002142, -0.00001027, 0.00000922, -0.00001604, 0.00001723, -0.00002005, 0.00000366, 0.00000463, 0.00001329, -0.00001469, 0.00000872, -0.00001519, 0.00000596, -0.00000346, 0.00000676, -0.00001026, 0.00000905, -0.00001283, 0.00001539, -0.00001061, 0.00001056, -0.00001525, 0.00000908, -0.00001011, 0.00002829, -0.00005599, 0.00004989, -0.00003524, 0.00002925, -0.00002833, 0.00003063, -0.00002130, 0.00002732, -0.00003506, 0.00002104, -0.00002694, 0.00003831, -0.00003779, 0.00004683, -0.00006127, 0.00005891, -0.00005438, 0.00004955, -0.00004204, 0.00002963, -0.00004443, 0.00005114, -0.00006189, 0.00006211, -0.00007701, 0.00007222, -0.00009616, 0.00008766, -0.00012092, 0.00012603, -0.00019317, 0.00017865, -0.00033233, 0.00031176, -0.00132060, 0.00217129, -0.00008806, -0.00064055, 0.00022855, -0.00026160, 0.00014995, -0.00016628, 0.00011774, -0.00011426, 0.00008415, -0.00009332, 0.00007622, -0.00008186, 0.00006964, -0.00007173, 0.00006613, -0.00006702, 0.00006120, -0.00005608, 0.00004348, -0.00002937, 0.00001374, -0.00002894, 0.00002874, -0.00002560, 0.00002308, -0.00001647, 0.00000888, -0.00002785, 0.00002765, -0.00001994, 0.00001420, -0.00002135, 0.00004035, -0.00004609, 0.00002355, -0.00002420, 0.00002708, -0.00001439, 0.00001149, -0.00001525, 0.00002053, -0.00002132, 0.00001913, -0.00001652, 0.00001408, -0.00001116, 0.00001533, -0.00001276, 0.00000559, -0.00001224, 0.00000907, -0.00001612, 0.00002474, -0.00000767, 0.00000432, -0.00001354, 0.00000058, 0.00000509, 0.00001138, -0.00001475, 0.00000185, 0.00000577, },
		{ -0.00001173, 0.00001070, -0.00000753, 0.00000905, -0.00002336, 0.00002229, -0.00000748, 0.00000827, -0.00001675, 0.00001621, -0.00001535, -0.00000044, 0.00000206, 0.00001647, -0.00001388, 0.00000826, -0.00001491, 0.00000610, -0.00000667, 0.00001173, -0.00001578, 0.00001474, -0.00001893, 0.00002019, -0.00001507, 0.00001709, -0.00002322, 0.00001912, -0.00002480, 0.00004326, -0.00006295, 0.00004593, -0.00002703, 0.00002285, -0.00002572, 0.00003062, -0.00002342, 0.00003243, -0.00003874, 0.00002798, -0.00003964, 0.00005124, -0.00004988, 0.00005937, -0.00007014, 0.00006546, -0.00006011, 0.00005558, -0.00005007, 0.00004360, -0.00006436, 0.00007205, -0.00008261, 0.00008422, -0.00010017, 0.00009867, -0.00012552, 0.00012255, -0.00016235, 0.00017387, -0.00024199, 0.00023954, -0.00039726, 0.00041434, -0.00125467, 0.00145557, 0.00081491, -0.00105831, 0.00039595, -0.00038477, 0.00023638, -0.00023757, 0.00017290, -0.00016328, 0.00012942, -0.00013493, 0.00011283, -0.00011434, 0.00009752, -0.00009682, 0.00008793, -0.00008603, 0.00007804, -0.00007001, 0.00005462, -0.00004112, 0.00002996, -0.00004695, 0.00004353, -0.00003958, 0.00003645, -0.00003157, 0.00002917, -0.00004782, 0.00004243, -0.00003420, 0.00003038, -0.00003959, 0.00005458, -0.00004994, 0.00002107, -0.00002222, 0.00002455, -0.00001365, 0.00001584, -0.00002129, 0.00002586, -0.00002446, 0.00002024, -0.00001711, 0.00001511, -0.00001293, 0.00001746, -0.00001411, 0.00000821, -0.00001503, 0.00001129, -0.00001713, 0.00002048, -0.00000264, 0.00000316, -0.00001016, -0.00000517, 0.00000701, 0.00001066, -0.00000802, -0.00000911, 0.00001464, },
		{ -0.00001110, 0.00000927, -0.00000707, 0.00001005, -0.00002192, 0.00001431, 0.00000213, 0.00000240, -0.00001221, 0.00000913, -0.00000561, -0.00000776, 0.00000354, 0.00001339, -0.00000693, 0.00000271, -0.00000985, 0.00000165, -0.00000474, 0.00001030, -0.00001365, 0.00001179, -0.00001545, 0.00001490, -0.00000978, 0.00001390, -0.00002059, 0.00001705, -0.00002434, 0.00003912, -0.00004886, 0.00002301, -0.00000482, 0.00000661, -0.00001490, 0.00002200, -0.00001600, 0.00002684, -0.00003134, 0.00002362, -0.00003829, 0.00004677, -0.00004265, 0.00005096, -0.00005710, 0.00004991, -0.00004320, 0.00003850, -0.00003451, 0.00003229, -0.00005531, 0.00005986, -0.00006730, 0.00006763, -0.00008172, 0.00007982, -0.00010477, 0.00010085, -0.00013792, 0.00014358, -0.00019883, 0.00018887, -0.00032135, 0.00031720, -0.00093020, 0.00061339, 0.00155322, -0.00119384, 0.00035120, -0.00035932, 0.00020602, -0.00021382, 0.00014850, -0.00014471, 0.00011462, -0.00012164, 0.00009894, -0.00010038, 0.00008298, -0.00008325, 0.00007438, -0.00007270, 0.00006467, -0.00005548, 0.00003978, -0.00002958, 0.00002405, -0.00004266, 0.00003689, -0.00003388, 0.00003179, -0.00002982, 0.00003143, -0.00004746, 0.00003693, -0.00002928, 0.00002719, -0.00003673, 0.00004574, -0.00003176, 0.00000138, -0.00000817, 0.00001352, -0.00000684, 0.00001412, -0.00001980, 0.00002269, -0.00001905, 0.00001379, -0.00001146, 0.00001074, -0.00000966, 0.00001458, -0.00001046, 0.00000564, -0.00001215, 0.00000755, -0.00001214, 0.00001175, 0.00000458, -0.00000031, -0.00000386, -0.00001223, 0.00000916, 0.00000852, 0.00000020, -0.00001842, 0.00001858, },
		{ -0.00000419, 0.00000319, -0.00000228, 0.00000565, -0.00001380, 0.00000066, 0.00001433, -0.00000450, -0.00000574, -0.00000044, 0.00000527, -0.00001505, 0.00000650, 0.00000708, 0.00000237, -0.00000418, -0.00000327, -0.00000468, 0.00000003, 0.00000469, -0.00000642, 0.00000321, -0.00000602, 0.00000380, 0.00000065, 0.00000540, -0.00001148, 0.00000706, -0.00001377, 0.00002309, -0.00002404, -0.00000596, 0.00001797, -0.00000765, -0.00000535, 0.00001250, -0.00000585, 0.00001698, -0.00001888, 0.00001308, -0.00002751, 0.00003026, -0.00002266, 0.00002924, -0.00003065, 0.00002120, -0.00001289, 0.00000797, -0.00000517, 0.00000555, -0.00002774, 0.00002660, -0.00002990, 0.00002798, -0.00003873, 0.00003423, -0.00005412, 0.00004501, -0.00007356, 0.00006672, -0.00010302, 0.00007679, -0.00017184, 0.00012313, -0.00051905, -0.00012949, 0.00194866, -0.00098238, 0.00010093, -0.00020798, 0.00008208, -0.00011694, 0.00006439, -0.00007584, 0.00005482, -0.00006721, 0.00004767, -0.00005265, 0.00003819, -0.00004246, 0.00003596, -0.00003636, 0.00002942, -0.00002037, 0.00000656, -0.00000169, 0.00000221, -0.00002207, 0.00001489, -0.00001422, 0.00001397, -0.00001490, 0.00001857, -0.00003029, 0.00001575, -0.00001015, 0.00000978, -0.00001882, 0.00002198, -0.00000235, -0.00002340, 0.00000673, 0.00000304, -0.00000044, 0.00001074, -0.00001434, 0.00001471, -0.00000917, 0.00000402, -0.00000346, 0.00000423, -0.00000406, 0.00000896, -0.00000385, -0.00000008, -0.00000592, 0.00000070, -0.00000464, 0.00000232, 0.00001089, -0.00000363, 0.00000253, -0.00001749, 0.00000893, 0.00000747, 0.00000652, -0.00002241, 0.00001520, },
		{ 0.00000631, -0.00000446, 0.00000394, -0.00000094, -0.00000340, -0.00001292, 0.00002328, -0.00000770, -0.00000136, -0.00000825, 0.00001298, -0.00001870, 0.00000768, 0.00000154, 0.00000947, -0.00000823, 0.00000119, -0.00000950, 0.00000421, -0.00000121, 0.00000134, -0.00000567, 0.00000334, -0.00000651, 0.00000949, -0.00000188, -0.00000254, -0.00000357, -0.00000179, 0.00000611, -0.00000157, -0.00002727, 0.00002928, -0.00001085, -0.00000369, 0.00000757, 0.00000167, 0.00000851, -0.00000714, 0.00000201, -0.00001360, 0.00000972, 0.00000041, 0.00000503, -0.00000244, -0.00000843, 0.00001791, -0.00002235, 0.00002368, -0.00002164, 0.00000191, -0.00000907, 0.00000878, -0.00001221, 0.00000461, -0.00001208, -0.00000204, -0.00001319, -0.00000579, -0.00001353, -0.00000604, -0.00003446, -0.00002685, -0.00006131, -0.00016977, -0.00062281, 0.00194344, -0.00048004, -0.00025934, -0.00001243, -0.00007053, -0.00000065, -0.00003427, 0.00000487, -0.00001645, -0.00000203, -0.00001294, 0.00000281, -0.00001281, 0.00000358, -0.00000735, 0.00000479, -0.00001077, 0.00001906, -0.00002939, 0.00002794, -0.00002193, 0.00000162, -0.00000963, 0.00000741, -0.00000611, 0.00000332, 0.00000030, -0.00000710, -0.00000940, 0.00001140, -0.00000998, 0.00000152, -0.00000262, 0.00002320, -0.00003923, 0.00001148, 0.00000078, 0.00000044, 0.00000943, -0.00000864, 0.00000627, 0.00000043, -0.00000443, 0.00000285, -0.00000104, 0.00000095, 0.00000334, 0.00000292, -0.00000596, 0.00000032, -0.00000557, 0.00000148, -0.00000427, 0.00001376, -0.00000481, 0.00000667, -0.00001874, 0.00000498, 0.00000869, 0.00000932, -0.00002009, 0.00000555, },
		{ 0.00001570, -0.00000945, 0.00000783, -0.00000571, 0.00000439, -0.00002109, 0.00002462, -0.00000441, -0.00000132, -0.00001169, 0.00001501, -0.00001691, 0.00000530, -0.00000059, 0.00001150, -0.00000721, 0.00000164, -0.00001078, 0.00000536, -0.00000423, 0.00000568, -0.00001031, 0.00000769, -0.00001092, 0.00001184, -0.00000327, 0.00000131, -0.00000907, 0.00000457, -0.00000343, 0.00000968, -0.00003341, 0.00002469, -0.00000167, -0.00000974, 0.00000725, 0.00000552, 0.00000343, 0.00000126, -0.00000634, -0.00000124, -0.00000813, 0.00001849, -0.00001286, 0.00001820, -0.00002926, 0.00003886, -0.00004140, 0.00004046, -0.00003699, 0.00001992, -0.00003166, 0.00003191, -0.00003527, 0.00002974, -0.00003924, 0.00002971, -0.00004931, 0.00003720, -0.00006426, 0.00005401, -0.00010060, 0.00006067, -0.00016897, 0.00003836, -0.00082569, 0.00160482, 0.00016193, -0.00058557, 0.00012658, -0.00017935, 0.00007837, -0.00010147, 0.00005896, -0.00006574, 0.00004366, -0.00005522, 0.00004076, -0.00004713, 0.00003423, -0.00003684, 0.00003345, -0.00003940, 0.00004670, -0.00005263, 0.00004495, -0.00003497, 0.00001544, -0.00002426, 0.00001958, -0.00001796, 0.00001479, -0.00001282, 0.00001040, -0.00002634, 0.00002367, -0.00002073, 0.00001261, -0.00001602, 0.00003369, -0.00003782, 0.00000166, 0.00000830, -0.00000444, 0.00001056, -0.00000425, -0.00000007, 0.00000677, -0.00000884, 0.00000535, -0.00000338, 0.00000377, -0.00000052, 0.00000783, -0.00000967, 0.00000400, -0.00000868, 0.00000387, -0.00000640, 0.00001261, -0.00000352, 0.00000788, -0.00001576, -0.00000198, 0.00001135, 0.00000930, -0.00001337, -0.00000654, },
		{ 0.00001966, -0.00000874, 0.00000713, -0.00000625, 0.00000678, -0.00002153, 0.00001770, 0.00000453, -0.00000472, -0.00001111, 0.00001187, -0.00001068, 0.00000001, 0.00000076, 0.00000855, -0.00000186, -0.00000114, -0.00000887, 0.00000323, -0.00000348, 0.00000527, -0.00000923, 0.00000562, -0.00000830, 0.00000709, 0.00000151, -0.00000054, -0.00000816, 0.00000342, -0.00000347, 0.00000871, -0.00002600, 0.00000897, 0.00001326, -0.00001719, 0.00000693, 0.00000847, 0.00000024, 0.00000695, -0.00001173, 0.00000778, -0.00001996, 0.00002771, -0.00002064, 0.00002756, -0.00003753, 0.00004592, -0.00004492, 0.00004094, -0.00003599, 0.00002106, -0.00003525, 0.00003362, -0.00003569, 0.00003128, -0.00004158, 0.00003473, -0.00005582, 0.00004642, -0.00007509, 0.00006612, -0.00011032, 0.00007896, -0.00018763, 0.00010364, -0.00079421, 0.00108320, 0.00075881, -0.00074622, 0.00013795, -0.00020120, 0.00008918, -0.00011361, 0.00006732, -0.00007639, 0.00005464, -0.00006536, 0.00004893, -0.00005406, 0.00004010, -0.00004396, 0.00004144, -0.00004822, 0.00005419, -0.00005528, 0.00004239, -0.00003059, 0.00001334, -0.00002341, 0.00001742, -0.00001703, 0.00001477, -0.00001515, 0.00001563, -0.00002875, 0.00002135, -0.00001771, 0.00000993, -0.00001430, 0.00002717, -0.00002068, -0.00001817, 0.00002007, -0.00001031, 0.00001114, -0.00000010, -0.00000418, 0.00000958, -0.00000932, 0.00000457, -0.00000336, 0.00000466, -0.00000247, 0.00001034, -0.00001044, 0.00000435, -0.00000811, 0.00000255, -0.00000490, 0.00000894, -0.00000107, 0.00000709, -0.00001012, -0.00000958, 0.00001329, 0.00000856, -0.00000558, -0.00001645, },
		{ 0.00001650, -0.00000230, 0.00000244, -0.00000302, 0.00000431, -0.00001582, 0.00000581, 0.00001501, -0.00000812, -0.00000919, 0.00000640, -0.00000312, -0.00000560, 0.00000356, 0.00000315, 0.00000471, -0.00000434, -0.00000595, -0.00000046, -0.00000040, 0.00000156, -0.00000426, -0.00000054, -0.00000159, -0.00000122, 0.00000873, -0.00000470, -0.00000396, -0.00000204, 0.00000200, 0.00000136, -0.00001339, -0.00000759, 0.00002386, -0.00001813, 0.00000137, 0.00001385, -0.00000345, 0.00001183, -0.00001540, 0.00001365, -0.00002545, 0.00002839, -0.00001951, 0.00002737, -0.00003526, 0.00004128, -0.00003552, 0.00002841, -0.00002233, 0.00000907, -0.00002405, 0.00001925, -0.00002010, 0.00001667, -0.00002700, 0.00002124, -0.00004131, 0.00003120, -0.00005691, 0.00004410, -0.00008126, 0.00005068, -0.00014766, 0.00007726, -0.00063566, 0.00054618, 0.00116159, -0.00067607, 0.00001423, -0.00014279, 0.00004007, -0.00007912, 0.00003754, -0.00005464, 0.00003602, -0.00004813, 0.00003213, -0.00003845, 0.00002590, -0.00003291, 0.00003213, -0.00003977, 0.00004364, -0.00003981, 0.00002337, -0.00001212, -0.00000147, -0.00001045, 0.00000446, -0.00000657, 0.00000576, -0.00000824, 0.00000971, -0.00001862, 0.00000762, -0.00000472, -0.00000237, -0.00000260, 0.00001088, 0.00000243, -0.00003730, 0.00002684, -0.00001059, 0.00000726, 0.00000570, -0.00000733, 0.00001037, -0.00000782, 0.00000266, -0.00000282, 0.00000497, -0.00000338, 0.00001104, -0.00000889, 0.00000226, -0.00000521, -0.00000059, -0.00000218, 0.00000529, 0.00000063, 0.00000589, -0.00000407, -0.00001522, 0.00001247, 0.00000901, 0.00000043, -0.00002107, },
		{ 0.00000784, 0.00000680, -0.00000308, 0.00000113, 0.00000006, -0.00000815, -0.00000571, 0.00002191, -0.00000775, -0.00000890, 0.00000181, 0.00000245, -0.00000881, 0.00000530, -0.00000153, 0.00000905, -0.00000510, -0.00000439, -0.00000348, 0.00000263, -0.00000255, 0.00000105, -0.00000658, 0.00000444, -0.00000790, 0.00001326, -0.00000643, -0.00000114, -0.00000656, 0.00000659, -0.00000435, -0.00000502, -0.00001561, 0.00002288, -0.00000853, -0.00001094, 0.00002214, -0.00000808, 0.00001656, -0.00001799, 0.00001685, -0.00002572, 0.00002318, -0.00001335, 0.00002217, -0.00002747, 0.00003047, -0.00001962, 0.00001033, -0.00000420, -0.00000725, -0.00000791, 0.00000023, -0.00000119, -0.00000063, -0.00000954, 0.00000400, -0.00002174, 0.00000934, -0.00003043, 0.00001280, -0.00004312, 0.00001156, -0.00009384, 0.00002379, -0.00045467, 0.00011832, 0.00130473, -0.00039912, -0.00018554, -0.00005464, -0.00002970, -0.00003005, -0.00000385, -0.00002269, 0.00000705, -0.00002111, 0.00000673, -0.00001554, 0.00000555, -0.00001607, 0.00001650, -0.00002391, 0.00002459, -0.00001613, -0.00000181, 0.00001033, -0.00001932, 0.00000517, -0.00001017, 0.00000510, -0.00000487, 0.00000128, -0.00000055, -0.00000398, -0.00000850, 0.00000910, -0.00001499, 0.00000906, -0.00000386, 0.00001942, -0.00004506, 0.00002136, -0.00000145, -0.00000249, 0.00001355, -0.00001025, 0.00001066, -0.00000640, 0.00000163, -0.00000335, 0.00000577, -0.00000409, 0.00001081, -0.00000618, -0.00000069, -0.00000200, -0.00000324, -0.00000061, 0.00000363, 0.00000043, 0.00000521, 0.00000084, -0.00001743, 0.00000833, 0.00001108, 0.00000372, -0.00002009, },
		{ -0.00000242, 0.00001410, -0.00000564, 0.00000301, -0.00000256, -0.00000270, -0.00001239, 0.00002190, -0.00000194, -0.00001142, -0.00000031, 0.00000443, -0.00000849, 0.00000474, -0.00000355, 0.00000926, -0.00000221, -0.00000511, -0.00000462, 0.00000387, -0.00000465, 0.00000367, -0.00000907, 0.00000613, -0.00000930, 0.00001186, -0.00000295, -0.00000263, -0.00000658, 0.00000588, -0.00000328, -0.00000588, -0.00001185, 0.00001023, 0.00000880, -0.00002601, 0.00002976, -0.00001125, 0.00001964, -0.00001856, 0.00001707, -0.00002186, 0.00001497, -0.00000604, 0.00001619, -0.00001875, 0.00001867, -0.00000344, -0.00000606, 0.00001062, -0.00001960, 0.00000391, -0.00001308, 0.00001011, -0.00000963, -0.00000036, -0.00000525, -0.00000993, -0.00000458, -0.00001274, -0.00000762, -0.00001913, -0.00001150, -0.00005871, -0.00001212, -0.00031519, -0.00014997, 0.00121496, -0.00001240, -0.00036475, -0.00000208, -0.00007466, -0.00000135, -0.00002896, -0.00000363, -0.00001211, -0.00000275, -0.00001031, -0.00000073, -0.00000718, -0.00000563, 0.00000548, -0.00001101, 0.00000754, 0.00000467, -0.00002188, 0.00002599, -0.00003013, 0.00001382, -0.00001754, 0.00001003, -0.00000990, 0.00000641, -0.00000750, 0.00000637, -0.00001784, 0.00001507, -0.00001956, 0.00001217, -0.00000843, 0.00002270, -0.00003696, 0.00000304, 0.00001507, -0.00001519, 0.00002109, -0.00001203, 0.00001074, -0.00000576, 0.00000206, -0.00000508, 0.00000701, -0.00000462, 0.00000997, -0.00000313, -0.00000326, 0.00000004, -0.00000400, -0.00000123, 0.00000428, -0.00000125, 0.00000478, 0.00000436, -0.00001641, 0.00000208, 0.00001350, 0.00000529, -0.00001561, },
		{ -0.00001028, 0.00001611, -0.00000311, 0.00000117, -0.00000193, -0.00000143, -0.00001283, 0.00001523, 0.00000768, -0.00001517, -0.00000081, 0.00000364, -0.00000578, 0.00000260, -0.00000304, 0.00000583, 0.00000313, -0.00000703, -0.00000436, 0.00000322, -0.00000413, 0.00000271, -0.00000716, 0.00000287, -0.00000524, 0.00000504, 0.00000482, -0.00000770, -0.00000236, -0.00000011, 0.00000420, -0.00001424, -0.00000051, -0.00000713, 0.00002562, -0.00003627, 0.00003111, -0.00000925, 0.00001859, -0.00001541, 0.00001354, -0.00001462, 0.00000580, 0.00000012, 0.00001155, -0.00001128, 0.00000854, 0.00000959, -0.00001682, 0.00001819, -0.00002403, 0.00000713, -0.00001626, 0.00000987, -0.00000716, -0.00000225, -0.00000364, -0.00000930, -0.00000625, -0.00000925, -0.00001084, -0.00001578, -0.00001181, -0.00004986, -0.00001923, -0.00023074, -0.00027743, 0.00098284, 0.00035939, -0.00043505, -0.00002889, -0.00007025, -0.00001002, -0.00002523, -0.00000756, -0.00001239, -0.00000160, -0.00001122, -0.00000078, -0.00000657, -0.00000647, 0.00000368, -0.00000594, -0.00000183, 0.00001615, -0.00003040, 0.00002918, -0.00002890, 0.00001105, -0.00001392, 0.00000524, -0.00000665, 0.00000415, -0.00000726, 0.00000780, -0.00001622, 0.00001001, -0.00001354, 0.00000461, -0.00000147, 0.00001290, -0.00001698, -0.00002129, 0.00003157, -0.00002463, 0.00002419, -0.00001061, 0.00000964, -0.00000515, 0.00000291, -0.00000672, 0.00000749, -0.00000417, 0.00000827, 0.00000007, -0.00000501, 0.00000057, -0.00000294, -0.00000340, 0.00000594, -0.00000273, 0.00000342, 0.00000720, -0.00001359, -0.00000407, 0.00001432, 0.00000694, -0.00001053, },
		{ -0.00001363, 0.00001214, 0.00000366, -0.00000310, 0.00000089, -0.00000326, -0.00000905, 0.00000550, 0.00001690, -0.00001683, -0.00000209, 0.00000243, -0.00000312, 0.00000083, -0.00000164, 0.00000103, 0.00000822, -0.00000792, -0.00000428, 0.00000185, -0.00000205, -0.00000053, -0.00000263, -0.00000299, 0.00000127, -0.00000353, 0.00001292, -0.00001272, 0.00000264, -0.00000755, 0.00001327, -0.00002356, 0.00000985, -0.00001948, 0.00003295, -0.00003521, 0.00002235, -0.00000014, 0.00001233, -0.00000790, 0.00000629, -0.00000500, -0.00000298, 0.00000451, 0.00000815, -0.00000482, 0.00000007, 0.00001911, -0.00002173, 0.00001899, -0.00002160, 0.00000323, -0.00001149, 0.00000156, 0.00000214, -0.00000998, 0.00000342, -0.00001446, -0.00000112, -0.00001416, -0.00000372, -0.00002427, -0.00000038, -0.00005412, -0.00001236, -0.00017989, -0.00032060, 0.00071458, 0.00061594, -0.00035656, -0.00013263, -0.00002704, -0.00004608, -0.00000169, -0.00002684, 0.00000017, -0.00001275, -0.00000059, -0.00001113, 0.00000283, -0.00001435, 0.00000760, -0.00000638, -0.00000465, 0.00001876, -0.00002813, 0.00002143, -0.00001783, -0.00000057, -0.00000237, -0.00000587, 0.00000180, -0.00000308, -0.00000146, 0.00000177, -0.00000612, -0.00000238, -0.00000121, -0.00000882, 0.00001116, -0.00000218, 0.00000481, -0.00004095, 0.00003916, -0.00002480, 0.00001960, -0.00000460, 0.00000647, -0.00000332, 0.00000258, -0.00000661, 0.00000590, -0.00000194, 0.00000533, 0.00000351, -0.00000613, 0.00000017, -0.00000119, -0.00000545, 0.00000661, -0.00000221, 0.00000013, 0.00001006, -0.00001046, -0.00000820, 0.00001226, 0.00000974, -0.00000682, },
		{ -0.00001295, 0.00000447, 0.00001125, -0.00000653, 0.00000312, -0.00000524, -0.00000492, -0.00000251, 0.00002133, -0.00001350, -0.00000613, 0.00000289, -0.00000259, 0.00000103, -0.00000098, -0.00000277, 0.00001049, -0.00000592, -0.00000570, 0.00000107, -0.00000006, -0.00000377, 0.00000154, -0.00000788, 0.00000611, -0.00000936, 0.00001698, -0.00001369, 0.00000443, -0.00001182, 0.00001822, -0.00002694, 0.00001169, -0.00002001, 0.00002649, -0.00002162, 0.00000462, 0.00001402, 0.00000275, 0.00000227, -0.00000291, 0.00000510, -0.00001016, 0.00000730, 0.00000491, 0.00000185, -0.00000778, 0.00002622, -0.00002266, 0.00001604, -0.00001624, -0.00000313, -0.00000456, -0.00000765, 0.00001021, -0.00001519, 0.00000755, -0.00001682, 0.00000173, -0.00001733, 0.00000176, -0.00003008, 0.00000591, -0.00005222, -0.00001335, -0.00013318, -0.00033571, 0.00048801, 0.00071636, -0.00015377, -0.00026459, 0.00001568, -0.00008022, 0.00001833, -0.00004269, 0.00001021, -0.00002296, 0.00000939, -0.00002038, 0.00001053, -0.00001989, 0.00000930, -0.00000582, -0.00000635, 0.00001770, -0.00002091, 0.00000945, -0.00000399, -0.00001382, 0.00000968, -0.00001610, 0.00000898, -0.00000964, 0.00000473, -0.00000608, 0.00000552, -0.00001430, 0.00000945, -0.00001992, 0.00002061, -0.00001265, 0.00001804, -0.00004720, 0.00003255, -0.00001376, 0.00000755, 0.00000503, 0.00000163, 0.00000012, 0.00000023, -0.00000397, 0.00000189, 0.00000203, 0.00000139, 0.00000704, -0.00000684, -0.00000039, -0.00000001, -0.00000591, 0.00000494, 0.00000102, -0.00000492, 0.00001277, -0.00000760, -0.00000975, 0.00000763, 0.00001324, -0.00000455, },
		{ -0.00001053, -0.00000312, 0.00001561, -0.00000598, 0.00000225, -0.00000473, -0.00000369, -0.00000549, 0.00001892, -0.00000478, -0.00001278, 0.00000521, -0.00000446, 0.00000318, -0.00000141, -0.00000446, 0.00000903, -0.00000090, -0.00000855, 0.00000124, 0.00000085, -0.00000534, 0.00000314, -0.00000924, 0.00000662, -0.00000982, 0.00001485, -0.00000896, 0.00000125, -0.00001057, 0.00001612, -0.00002131, 0.00000275, -0.00000862, 0.00000918, -0.00000089, -0.00001580, 0.00002754, -0.00000562, 0.00001150, -0.00001088, 0.00001308, -0.00001444, 0.00000867, 0.00000114, 0.00000927, -0.00001542, 0.00003163, -0.00002135, 0.00001228, -0.00001160, -0.00000766, -0.00000062, -0.00001185, 0.00001102, -0.00001234, 0.00000351, -0.00001118, -0.00000339, -0.00001212, -0.00000243, -0.00002373, -0.00000323, -0.00003330, -0.00003359, -0.00007520, -0.00035261, 0.00033085, 0.00068413, 0.00009909, -0.00035381, 0.00001332, -0.00008330, 0.00001327, -0.00003853, 0.00000435, -0.00002052, 0.00000787, -0.00001857, 0.00000763, -0.00001549, 0.00000252, 0.00000101, -0.00001179, 0.00001827, -0.00001507, 0.00000032, 0.00000537, -0.00002154, 0.00001526, -0.00001909, 0.00000944, -0.00001063, 0.00000640, -0.00001004, 0.00001192, -0.00001874, 0.00001191, -0.00002237, 0.00002052, -0.00001229, 0.00001765, -0.00003796, 0.00001336, 0.00000460, -0.00000751, 0.00001468, -0.00000284, 0.00000424, -0.00000354, 0.00000040, -0.00000343, 0.00000653, -0.00000249, 0.00000997, -0.00000700, -0.00000074, 0.00000001, -0.00000437, 0.00000106, 0.00000603, -0.00001018, 0.00001411, -0.00000455, -0.00000953, 0.00000218, 0.00001572, -0.00000237, },
		{ -0.00000872, -0.00000747, 0.00001435, -0.00000044, -0.00000207, -0.00000125, -0.00000604, -0.00000334, 0.00001113, 0.00000660, -0.00001946, 0.00000753, -0.00000699, 0.00000554, -0.00000181, -0.00000463, 0.00000492, 0.00000535, -0.00001125, 0.00000147, 0.00000090, -0.00000501, 0.00000181, -0.00000689, 0.00000290, -0.00000552, 0.00000782, -0.00000031, -0.00000530, -0.00000509, 0.00000832, -0.00000888, -0.00001296, 0.00000822, -0.00001032, 0.00001762, -0.00003049, 0.00003389, -0.00000814, 0.00001616, -0.00001463, 0.00001674, -0.00001490, 0.00000864, -0.00000299, 0.00001664, -0.00002182, 0.00003473, -0.00001818, 0.00000885, -0.00000903, -0.00000894, -0.00000132, -0.00000954, 0.00000396, -0.00000186, -0.00000755, 0.00000111, -0.00001521, 0.00000051, -0.00001562, -0.00000612, -0.00002568, -0.00000117, -0.00006733, -0.00001172, -0.00036992, 0.00022639, 0.00058380, 0.00031412, -0.00034272, -0.00005670, -0.00004596, -0.00002196, -0.00001156, -0.00001929, -0.00000348, -0.00000730, -0.00000387, -0.00000708, -0.00000062, -0.00001272, 0.00001396, -0.00002136, 0.00002200, -0.00001334, -0.00000267, 0.00000709, -0.00002087, 0.00001189, -0.00001309, 0.00000216, -0.00000517, 0.00000221, -0.00000790, 0.00001026, -0.00001347, 0.00000488, -0.00001542, 0.00001067, -0.00000178, 0.00000621, -0.00001881, -0.00001046, 0.00002212, -0.00001885, 0.00001981, -0.00000438, 0.00000738, -0.00000707, 0.00000462, -0.00000802, 0.00000969, -0.00000486, 0.00001136, -0.00000619, -0.00000108, -0.00000083, -0.00000160, -0.00000359, 0.00001072, -0.00001337, 0.00001261, -0.00000045, -0.00000900, -0.00000196, 0.00001544, 0.00000117, },
		{ -0.00000847, -0.00000763, 0.00000799, 0.00000814, -0.00000764, 0.00000330, -0.00000999, 0.00000110, 0.00000213, 0.00001606, -0.00002249, 0.00000714, -0.00000768, 0.00000579, -0.00000049, -0.00000475, 0.00000038, 0.00001018, -0.00001169, 0.00000039, 0.00000111, -0.00000378, -0.00000109, -0.00000276, -0.00000253, 0.00000041, -0.00000031, 0.00000822, -0.00001135, 0.00000082, -0.00000075, 0.00000462, -0.00002792, 0.00002139, -0.00002239, 0.00002554, -0.00003347, 0.00002953, -0.00000294, 0.00001501, -0.00001313, 0.00001555, -0.00001174, 0.00000738, -0.00000663, 0.00002222, -0.00002502, 0.00003397, -0.00001232, 0.00000516, -0.00000748, -0.00000847, -0.00000466, -0.00000362, -0.00000662, 0.00001062, -0.00001936, 0.00001348, -0.00002700, 0.00001353, -0.00003007, 0.00001347, -0.00004965, 0.00002970, -0.00009715, 0.00003803, -0.00036823, 0.00013795, 0.00048411, 0.00042879, -0.00021604, -0.00017882, 0.00001327, -0.00007166, 0.00002468, -0.00004938, 0.00001889, -0.00002822, 0.00001630, -0.00002640, 0.00001767, -0.00002975, 0.00002725, -0.00003056, 0.00002582, -0.00001371, -0.00000133, 0.00000335, -0.00001437, 0.00000253, -0.00000173, -0.00000890, 0.00000314, -0.00000494, -0.00000185, 0.00000281, -0.00000185, -0.00000719, -0.00000401, -0.00000347, 0.00001236, -0.00000810, 0.00000027, -0.00002873, 0.00003056, -0.00002107, 0.00001756, -0.00000173, 0.00000844, -0.00000881, 0.00000680, -0.00001002, 0.00001006, -0.00000473, 0.00001064, -0.00000407, -0.00000178, -0.00000173, 0.00000101, -0.00000708, 0.00001292, -0.00001279, 0.00000759, 0.00000495, -0.00000907, -0.00000354, 0.00001185, 0.00000632, },
		{ -0.00000892, -0.00000508, -0.00000039, 0.00001585, -0.00001090, 0.00000585, -0.00001239, 0.00000394, -0.00000357, 0.00001964, -0.00001937, 0.00000245, -0.00000504, 0.00000262, 0.00000331, -0.00000578, -0.00000274, 0.00001161, -0.00000866, -0.00000270, 0.00000221, -0.00000285, -0.00000366, 0.00000058, -0.00000655, 0.00000436, -0.00000558, 0.00001271, -0.00001333, 0.00000346, -0.00000664, 0.00001353, -0.00003546, 0.00002410, -0.00002162, 0.00002010, -0.00002461, 0.00001620, 0.00000760, 0.00001007, -0.00000801, 0.00001108, -0.00000646, 0.00000552, -0.00000896, 0.00002433, -0.00002334, 0.00002819, -0.00000307, 0.00000013, -0.00000498, -0.00000898, -0.00000713, 0.00000119, -0.00001463, 0.00001819, -0.00002485, 0.00001884, -0.00003148, 0.00001912, -0.00003688, 0.00002431, -0.00006206, 0.00004421, -0.00010605, 0.00005635, -0.00033016, 0.00003645, 0.00042671, 0.00042956, -0.00000879, -0.00030355, 0.00005660, -0.00010806, 0.00004823, -0.00006827, 0.00003214, -0.00004224, 0.00003019, -0.00003926, 0.00002916, -0.00003942, 0.00003307, -0.00003303, 0.00002461, -0.00001163, -0.00000049, -0.00000048, -0.00000773, -0.00000674, 0.00000853, -0.00001741, 0.00000869, -0.00001016, 0.00000350, -0.00000519, 0.00000958, -0.00001712, 0.00000466, -0.00001437, 0.00002202, -0.00001649, 0.00001044, -0.00003434, 0.00002609, -0.00001326, 0.00000885, 0.00000389, 0.00000782, -0.00000831, 0.00000617, -0.00000884, 0.00000743, -0.00000222, 0.00000806, -0.00000088, -0.00000296, -0.00000206, 0.00000233, -0.00000811, 0.00001162, -0.00000825, -0.00000009, 0.00001070, -0.00000954, -0.00000285, 0.00000611, 0.00001175, },
		{ -0.00000826, -0.00000242, -0.00000696, 0.00001892, -0.00000912, 0.00000431, -0.00001104, 0.00000261, -0.00000356, 0.00001616, -0.00001045, -0.00000566, 0.00000026, -0.00000317, 0.00000863, -0.00000734, -0.00000402, 0.00000941, -0.00000272, -0.00000708, 0.00000398, -0.00000269, -0.00000474, 0.00000147, -0.00000728, 0.00000439, -0.00000612, 0.00001171, -0.00001019, 0.00000159, -0.00000744, 0.00001537, -0.00003322, 0.00001534, -0.00000944, 0.00000541, -0.00000974, 0.00000016, 0.00001795, 0.00000554, -0.00000253, 0.00000613, -0.00000122, 0.00000391, -0.00000951, 0.00002219, -0.00001650, 0.00001769, 0.00000909, -0.00000637, -0.00000044, -0.00001222, -0.00000635, 0.00000160, -0.00001606, 0.00001687, -0.00002053, 0.00001409, -0.00002557, 0.00001385, -0.00003179, 0.00002100, -0.00005654, 0.00003587, -0.00008794, 0.00003901, -0.00025471, -0.00008315, 0.00041367, 0.00035085, 0.00020992, -0.00037013, 0.00004783, -0.00010822, 0.00004233, -0.00006327, 0.00002603, -0.00004027, 0.00002918, -0.00003776, 0.00002681, -0.00003577, 0.00002666, -0.00002501, 0.00001519, -0.00000378, -0.00000418, 0.00000024, -0.00000588, -0.00001080, 0.00001246, -0.00001863, 0.00000756, -0.00000994, 0.00000443, -0.00000911, 0.00001521, -0.00001931, 0.00000550, -0.00001712, 0.00002234, -0.00001450, 0.00000850, -0.00002686, 0.00001139, 0.00000016, -0.00000179, 0.00000907, 0.00000724, -0.00000631, 0.00000337, -0.00000536, 0.00000298, 0.00000143, 0.00000473, 0.00000260, -0.00000429, -0.00000170, 0.00000207, -0.00000661, 0.00000738, -0.00000131, -0.00000818, 0.00001493, -0.00000922, -0.00000137, 0.00000051, 0.00001528, },
		{ -0.00000516, -0.00000171, -0.00000918, 0.00001576, -0.00000211, -0.00000096, -0.00000620, -0.00000264, 0.00000126, 0.00000795, 0.00000091, -0.00001398, 0.00000557, -0.00000905, 0.00001307, -0.00000784, -0.00000442, 0.00000508, 0.00000397, -0.00001083, 0.00000520, -0.00000276, -0.00000445, 0.00000001, -0.00000503, 0.00000111, -0.00000295, 0.00000683, -0.00000389, -0.00000304, -0.00000457, 0.00001167, -0.00002380, -0.00000025, 0.00000712, -0.00000995, 0.00000263, -0.00001129, 0.00002257, 0.00000518, 0.00000053, 0.00000300, 0.00000235, 0.00000309, -0.00000816, 0.00001620, -0.00000581, 0.00000450, 0.00002207, -0.00001299, 0.00000548, -0.00001774, -0.00000258, -0.00000243, -0.00001100, 0.00000763, -0.00000846, 0.00000190, -0.00001217, 0.00000056, -0.00001749, 0.00000619, -0.00003639, 0.00000982, -0.00005048, -0.00000222, -0.00015918, -0.00020236, 0.00041644, 0.00025263, 0.00036823, -0.00033600, -0.00002599, -0.00006898, 0.00000632, -0.00003497, 0.00000141, -0.00002262, 0.00001318, -0.00002193, 0.00001111, -0.00001986, 0.00000957, -0.00000824, -0.00000100, 0.00000920, -0.00001269, 0.00000633, -0.00000977, -0.00000856, 0.00000904, -0.00001217, -0.00000002, -0.00000451, 0.00000059, -0.00000767, 0.00001339, -0.00001282, -0.00000154, -0.00001206, 0.00001416, -0.00000398, -0.00000191, -0.00001239, -0.00000578, 0.00001171, -0.00000856, 0.00001023, 0.00000846, -0.00000399, -0.00000021, -0.00000132, -0.00000142, 0.00000455, 0.00000200, 0.00000538, -0.00000517, -0.00000106, 0.00000075, -0.00000361, 0.00000200, 0.00000551, -0.00001402, 0.00001579, -0.00000695, -0.00000074, -0.00000279, 0.00001523, },
		{ 0.00000015, -0.00000324, -0.00000700, 0.00000782, 0.00000748, -0.00000719, -0.00000037, -0.00000915, 0.00000739, -0.00000049, 0.00001001, -0.00001866, 0.00000785, -0.00001218, 0.00001417, -0.00000566, -0.00000533, 0.00000079, 0.00000883, -0.00001193, 0.00000457, -0.00000209, -0.00000380, -0.00000239, -0.00000178, -0.00000300, 0.00000095, 0.00000153, 0.00000196, -0.00000705, -0.00000133, 0.00000637, -0.00001259, -0.00001543, 0.00001960, -0.00001779, 0.00000606, -0.00001386, 0.00001923, 0.00000981, 0.00000071, 0.00000210, 0.00000415, 0.00000270, -0.00000492, 0.00000761, 0.00000626, -0.00000835, 0.00003294, -0.00001740, 0.00001073, -0.00002320, 0.00000164, -0.00000809, -0.00000310, -0.00000452, 0.00000526, -0.00001112, 0.00000191, -0.00001372, -0.00000140, -0.00001181, -0.00001176, -0.00002095, -0.00000989, -0.00004659, -0.00006987, -0.00029296, 0.00039787, 0.00018978, 0.00042241, -0.00019813, -0.00014417, -0.00001094, -0.00004297, 0.00000238, -0.00002989, 0.00000101, -0.00000946, 0.00000031, -0.00001023, 0.00000079, -0.00001098, 0.00001070, -0.00001825, 0.00002270, -0.00002231, 0.00001437, -0.00001602, -0.00000334, 0.00000175, -0.00000205, -0.00000979, 0.00000235, -0.00000502, -0.00000327, 0.00000671, -0.00000145, -0.00001177, -0.00000412, 0.00000294, 0.00000859, -0.00001303, 0.00000018, -0.00001685, 0.00001498, -0.00000771, 0.00000593, 0.00001173, -0.00000183, -0.00000337, 0.00000173, -0.00000426, 0.00000594, 0.00000071, 0.00000688, -0.00000518, -0.00000063, -0.00000073, -0.00000058, -0.00000251, 0.00000995, -0.00001582, 0.00001257, -0.00000251, -0.00000170, -0.00000293, 0.00001147, },
		{ 0.00000589, -0.00000557, -0.00000251, -0.00000136, 0.00001549, -0.00001076, 0.00000328, -0.00001356, 0.00001092, -0.00000491, 0.00001326, -0.00001741, 0.00000558, -0.00001117, 0.00001092, -0.00000045, -0.00000734, -0.00000198, 0.00001024, -0.00000950, 0.00000168, -0.00000019, -0.00000374, -0.00000411, 0.00000026, -0.00000532, 0.00000269, -0.00000108, 0.00000438, -0.00000779, -0.00000051, 0.00000302, -0.00000434, -0.00002453, 0.00002259, -0.00001459, -0.00000033, -0.00000873, 0.00001036, 0.00001663, 0.00000024, 0.00000184, 0.00000552, 0.00000175, 0.00000013, -0.00000187, 0.00001702, -0.00001790, 0.00003909, -0.00001753, 0.00001313, -0.00002587, 0.00000313, -0.00001168, 0.00000294, -0.00001370, 0.00001412, -0.00001843, 0.00001011, -0.00002220, 0.00000906, -0.00002436, 0.00000673, -0.00004329, 0.00001830, -0.00007543, -0.00000855, -0.00033397, 0.00033414, 0.00018864, 0.00037270, 0.00000445, -0.00025946, 0.00003131, -0.00008051, 0.00002899, -0.00005198, 0.00001753, -0.00002713, 0.00001802, -0.00002686, 0.00001649, -0.00002612, 0.00002391, -0.00002974, 0.00003082, -0.00002768, 0.00001906, -0.00001926, -0.00000038, -0.00000400, 0.00000597, -0.00001618, 0.00000584, -0.00000828, 0.00000015, -0.00000011, 0.00000889, -0.00001889, 0.00000057, -0.00000495, 0.00001636, -0.00001771, 0.00000414, -0.00001710, 0.00000827, 0.00000002, -0.00000181, 0.00001516, 0.00000083, -0.00000580, 0.00000318, -0.00000512, 0.00000550, 0.00000079, 0.00000721, -0.00000436, -0.00000064, -0.00000173, 0.00000140, -0.00000488, 0.00001101, -0.00001346, 0.00000625, 0.00000298, -0.00000365, -0.00000068, 0.00000566, },
		{ 0.00000972, -0.00000668, 0.00000150, -0.00000801, 0.00001850, -0.00000928, 0.00000289, -0.00001389, 0.00000968, -0.00000352, 0.00001007, -0.00001089, -0.00000022, -0.00000694, 0.00000453, 0.00000637, -0.00000965, -0.00000316, 0.00000835, -0.00000439, -0.00000252, 0.00000241, -0.00000441, -0.00000440, -0.00000001, -0.00000469, 0.00000115, -0.00000006, 0.00000282, -0.00000501, -0.00000256, 0.00000270, -0.00000075, -0.00002601, 0.00001616, -0.00000301, -0.00001163, -0.00000159, 0.00000153, 0.00002094, 0.00000240, -0.00000000, 0.00000806, -0.00000063, 0.00000647, -0.00001043, 0.00002424, -0.00002226, 0.00003923, -0.00001254, 0.00001162, -0.00002410, -0.00000001, -0.00001093, 0.00000409, -0.00001628, 0.00001466, -0.00001710, 0.00000978, -0.00002205, 0.00001055, -0.00002717, 0.00001349, -0.00005040, 0.00002672, -0.00008069, 0.00001693, -0.00032236, 0.00022580, 0.00024040, 0.00025838, 0.00020764, -0.00031925, 0.00002786, -0.00008763, 0.00003117, -0.00005438, 0.00001838, -0.00003199, 0.00002363, -0.00003171, 0.00002085, -0.00003025, 0.00002663, -0.00003137, 0.00002994, -0.00002523, 0.00001659, -0.00001567, -0.00000339, -0.00000442, 0.00000793, -0.00001569, 0.00000326, -0.00000682, -0.00000009, -0.00000346, 0.00001375, -0.00001870, -0.00000154, -0.00000610, 0.00001605, -0.00001335, -0.00000158, -0.00000802, -0.00000445, 0.00000962, -0.00000836, 0.00001565, 0.00000527, -0.00000782, 0.00000338, -0.00000463, 0.00000409, 0.00000138, 0.00000706, -0.00000315, -0.00000099, -0.00000203, 0.00000201, -0.00000502, 0.00000922, -0.00000845, -0.00000096, 0.00000757, -0.00000520, 0.00000218, 0.00000029, },
		{ 0.00001014, -0.00000540, 0.00000321, -0.00001006, 0.00001549, -0.00000305, -0.00000088, -0.00001065, 0.00000431, 0.00000225, 0.00000307, -0.00000246, -0.00000645, -0.00000216, -0.00000228, 0.00001218, -0.00001055, -0.00000383, 0.00000471, 0.00000124, -0.00000613, 0.00000448, -0.00000511, -0.00000365, -0.00000210, -0.00000193, -0.00000251, 0.00000309, -0.00000076, -0.00000089, -0.00000557, 0.00000386, -0.00000025, -0.00002244, 0.00000493, 0.00001015, -0.00002027, 0.00000084, -0.00000199, 0.00001912, 0.00000907, -0.00000434, 0.00001221, -0.00000417, 0.00001282, -0.00001623, 0.00002659, -0.00002096, 0.00003372, -0.00000323, 0.00000673, -0.00001813, -0.00000746, -0.00000610, 0.00000035, -0.00001245, 0.00000788, -0.00000898, 0.00000312, -0.00001545, 0.00000500, -0.00002178, 0.00000978, -0.00004387, 0.00001823, -0.00006657, 0.00001372, -0.00027273, 0.00009457, 0.00031173, 0.00013631, 0.00034989, -0.00029022, -0.00002928, -0.00006347, 0.00000932, -0.00003817, 0.00000452, -0.00002429, 0.00001688, -0.00002461, 0.00001401, -0.00002384, 0.00001959, -0.00002393, 0.00002075, -0.00001535, 0.00000704, -0.00000536, -0.00001212, 0.00000041, 0.00000388, -0.00000893, -0.00000426, -0.00000167, -0.00000361, -0.00000279, 0.00001234, -0.00001135, -0.00000939, -0.00000183, 0.00000947, -0.00000291, -0.00001236, 0.00000383, -0.00001575, 0.00001451, -0.00000909, 0.00001087, 0.00001203, -0.00000962, 0.00000296, -0.00000381, 0.00000283, 0.00000158, 0.00000707, -0.00000200, -0.00000142, -0.00000184, 0.00000161, -0.00000375, 0.00000608, -0.00000303, -0.00000664, 0.00000953, -0.00000512, 0.00000396, -0.00000261, },
		{ 0.00000736, -0.00000218, 0.00000263, -0.00000789, 0.00000829, 0.00000514, -0.00000532, -0.00000627, -0.00000243, 0.00000885, -0.00000346, 0.00000371, -0.00000976, 0.00000040, -0.00000680, 0.00001467, -0.00000862, -0.00000514, 0.00000124, 0.00000515, -0.00000742, 0.00000485, -0.00000493, -0.00000290, -0.00000453, 0.00000095, -0.00000585, 0.00000556, -0.00000326, 0.00000152, -0.00000680, 0.00000392, 0.00000016, -0.00001810, -0.00000513, 0.00001807, -0.00002040, -0.00000499, 0.00000142, 0.00001116, 0.00001904, -0.00000966, 0.00001650, -0.00000720, 0.00001720, -0.00001774, 0.00002374, -0.00001484, 0.00002421, 0.00000840, 0.00000025, -0.00000978, -0.00001698, 0.00000034, -0.00000568, -0.00000542, -0.00000192, 0.00000085, -0.00000466, -0.00000752, -0.00000253, -0.00001342, 0.00000140, -0.00003090, 0.00000218, -0.00004468, -0.00000275, -0.00020821, -0.00002992, 0.00036467, 0.00005498, 0.00039699, -0.00017409, -0.00011987, -0.00002631, -0.00002236, -0.00001502, -0.00001461, -0.00001155, 0.00000419, -0.00001165, 0.00000185, -0.00001257, 0.00000812, -0.00001234, 0.00000769, -0.00000204, -0.00000582, 0.00000776, -0.00002256, 0.00000672, -0.00000243, -0.00000006, -0.00001243, 0.00000347, -0.00000754, -0.00000045, 0.00000738, -0.00000080, -0.00001828, 0.00000310, 0.00000182, 0.00000746, -0.00002101, 0.00001078, -0.00001892, 0.00001042, -0.00000232, 0.00000122, 0.00001984, -0.00001054, 0.00000212, -0.00000328, 0.00000229, 0.00000103, 0.00000742, -0.00000100, -0.00000181, -0.00000141, 0.00000082, -0.00000215, 0.00000318, 0.00000095, -0.00000932, 0.00000832, -0.00000325, 0.00000406, -0.00000248, },
		{ 0.00000311, 0.00000122, 0.00000120, -0.00000364, 0.00000034, 0.00001153, -0.00000733, -0.00000350, -0.00000749, 0.00001275, -0.00000595, 0.00000486, -0.00000850, -0.00000040, -0.00000791, 0.00001313, -0.00000384, -0.00000734, -0.00000101, 0.00000618, -0.00000587, 0.00000331, -0.00000352, -0.00000295, -0.00000588, 0.00000206, -0.00000675, 0.00000515, -0.00000252, 0.00000034, -0.00000475, 0.00000136, 0.00000255, -0.00001601, -0.00001023, 0.00001757, -0.00001118, -0.00001732, 0.00000859, 0.00000092, 0.00002821, -0.00001253, 0.00001820, -0.00000735, 0.00001769, -0.00001417, 0.00001643, -0.00000565, 0.00001299, 0.00001995, -0.00000558, -0.00000146, -0.00002562, 0.00000525, -0.00001062, 0.00000071, -0.00000981, 0.00000722, -0.00000870, -0.00000289, -0.00000728, -0.00000726, -0.00000557, -0.00001915, -0.00001179, -0.00002638, -0.00001797, -0.00014899, -0.00012454, 0.00037461, 0.00003639, 0.00035272, -0.00000681, -0.00020376, -0.00000372, -0.00004475, 0.00000026, -0.00002733, -0.00000325, -0.00000603, -0.00000085, -0.00000803, -0.00000347, -0.00000132, -0.00000245, -0.00000377, 0.00000949, -0.00001675, 0.00001825, -0.00002934, 0.00000954, -0.00000617, 0.00000588, -0.00001652, 0.00000470, -0.00000860, 0.00000041, 0.00000283, 0.00000783, -0.00002290, 0.00000375, -0.00000181, 0.00001226, -0.00002209, 0.00000835, -0.00001177, -0.00000200, 0.00000927, -0.00000974, 0.00002566, -0.00000902, 0.00000048, -0.00000293, 0.00000223, 0.00000020, 0.00000760, 0.00000021, -0.00000233, -0.00000088, 0.00000006, -0.00000096, 0.00000140, 0.00000281, -0.00000906, 0.00000490, -0.00000060, 0.00000310, -0.00000025, },
		{ -0.00000038, 0.00000285, 0.00000057, 0.00000027, -0.00000521, 0.00001346, -0.00000528, -0.00000361, -0.00000936, 0.00001233, -0.00000337, 0.00000111, -0.00000374, -0.00000342, -0.00000658, 0.00000886, 0.00000219, -0.00000943, -0.00000220, 0.00000471, -0.00000243, 0.00000084, -0.00000141, -0.00000380, -0.00000569, 0.00000078, -0.00000470, 0.00000160, 0.00000139, -0.00000390, -0.00000028, -0.00000309, 0.00000660, -0.00001633, -0.00001043, 0.00001039, 0.00000301, -0.00003001, 0.00001342, -0.00000602, 0.00003187, -0.00000962, 0.00001508, -0.00000304, 0.00001353, -0.00000609, 0.00000644, 0.00000435, 0.00000236, 0.00002933, -0.00000898, 0.00000483, -0.00003105, 0.00000634, -0.00001218, 0.00000322, -0.00001279, 0.00000759, -0.00000714, -0.00000311, -0.00000762, -0.00000538, -0.00000822, -0.00001263, -0.00001870, -0.00001715, -0.00002496, -0.00010484, -0.00018101, 0.00033817, 0.00007270, 0.00025210, 0.00015781, -0.00023975, -0.00001677, -0.00004567, -0.00000104, -0.00002718, -0.00000463, -0.00000877, 0.00000281, -0.00001097, -0.00000077, -0.00000498, 0.00000229, -0.00001026, 0.00001577, -0.00002207, 0.00002222, -0.00002887, 0.00000584, -0.00000447, 0.00000601, -0.00001418, 0.00000045, -0.00000556, -0.00000181, 0.00000126, 0.00001130, -0.00002047, -0.00000186, 0.00000039, 0.00000987, -0.00001491, -0.00000238, 0.00000216, -0.00001722, 0.00001967, -0.00001679, 0.00002608, -0.00000368, -0.00000243, -0.00000222, 0.00000185, -0.00000004, 0.00000689, 0.00000211, -0.00000320, -0.00000021, -0.00000050, -0.00000035, 0.00000071, 0.00000307, -0.00000718, 0.00000113, 0.00000125, 0.00000227, 0.00000241, },
		{ -0.00000162, 0.00000187, 0.00000137, 0.00000258, -0.00000696, 0.00001066, 0.00000001, -0.00000566, -0.00000867, 0.00000847, 0.00000250, -0.00000476, 0.00000140, -0.00000597, -0.00000507, 0.00000428, 0.00000713, -0.00000990, -0.00000329, 0.00000216, 0.00000101, -0.00000102, 0.00000037, -0.00000480, -0.00000449, -0.00000212, -0.00000095, -0.00000338, 0.00000635, -0.00000862, 0.00000393, -0.00000695, 0.00001007, -0.00001682, -0.00000881, 0.00000151, 0.00001531, -0.00003608, 0.00001051, -0.00000577, 0.00002751, -0.00000002, 0.00000716, 0.00000518, 0.00000596, 0.00000434, -0.00000362, 0.00001288, -0.00000585, 0.00003517, -0.00000901, 0.00000814, -0.00003231, 0.00000308, -0.00001007, 0.00000180, -0.00001089, 0.00000289, -0.00000172, -0.00000622, -0.00000528, -0.00000636, -0.00000755, -0.00001067, -0.00001934, -0.00001549, -0.00002510, -0.00007468, -0.00020507, 0.00026950, 0.00013661, 0.00014232, 0.00027141, -0.00020529, -0.00006720, -0.00002783, -0.00001622, -0.00001678, -0.00001359, -0.00000527, -0.00000005, -0.00000756, -0.00000374, -0.00000363, 0.00000248, -0.00001199, 0.00001663, -0.00002129, 0.00001920, -0.00002120, -0.00000374, 0.00000192, 0.00000114, -0.00000680, -0.00000738, -0.00000011, -0.00000606, 0.00000244, 0.00000969, -0.00001214, -0.00001164, 0.00000612, 0.00000308, -0.00000346, -0.00001574, 0.00001561, -0.00002770, 0.00002271, -0.00001590, 0.00001943, 0.00000543, -0.00000635, -0.00000091, 0.00000055, 0.00000094, 0.00000489, 0.00000481, -0.00000441, 0.00000050, -0.00000089, -0.00000009, 0.00000051, 0.00000283, -0.00000531, -0.00000128, 0.00000119, 0.00000238, 0.00000421, },
		{ -0.00000058, -0.00000101, 0.00000277, 0.00000362, -0.00000566, 0.00000514, 0.00000582, -0.00000728, -0.00000750, 0.00000365, 0.00000840, -0.00000901, 0.00000357, -0.00000560, -0.00000536, 0.00000142, 0.00000921, -0.00000781, -0.00000501, -0.00000006, 0.00000280, -0.00000109, 0.00000101, -0.00000520, -0.00000328, -0.00000510, 0.00000239, -0.00000728, 0.00000951, -0.00001083, 0.00000502, -0.00000771, 0.00001057, -0.00001483, -0.00000892, -0.00000413, 0.00002029, -0.00003168, -0.00000141, 0.00000136, 0.00001658, 0.00001373, -0.00000271, 0.00001450, -0.00000204, 0.00001393, -0.00001094, 0.00001804, -0.00001049, 0.00003696, -0.00000560, 0.00000854, -0.00002980, -0.00000343, -0.00000579, -0.00000185, -0.00000645, -0.00000359, 0.00000375, -0.00000859, -0.00000360, -0.00000710, -0.00000665, -0.00000994, -0.00001788, -0.00001602, -0.00002421, -0.00005187, -0.00020912, 0.00018928, 0.00019778, 0.00006179, 0.00030995, -0.00010708, -0.00013467, -0.00000784, -0.00003313, -0.00000579, -0.00002251, -0.00000142, -0.00000446, -0.00000251, -0.00000793, -0.00000133, 0.00000161, -0.00001182, 0.00001451, -0.00001676, 0.00001184, -0.00000969, -0.00001533, 0.00000922, -0.00000498, 0.00000148, -0.00001461, 0.00000417, -0.00000955, 0.00000404, 0.00000581, -0.00000198, -0.00002084, 0.00001067, -0.00000303, 0.00000624, -0.00002484, 0.00002166, -0.00002798, 0.00001550, -0.00000667, 0.00000724, 0.00001611, -0.00000982, 0.00000053, -0.00000154, 0.00000295, 0.00000201, 0.00000780, -0.00000547, 0.00000098, -0.00000107, 0.00000009, 0.00000020, 0.00000299, -0.00000450, -0.00000166, -0.00000070, 0.00000320, 0.00000493, },
		{ 0.00000158, -0.00000392, 0.00000312, 0.00000463, -0.00000319, -0.00000015, 0.00000918, -0.00000626, -0.00000772, 0.00000015, 0.00001157, -0.00000907, 0.00000118, -0.00000174, -0.00000777, 0.00000073, 0.00000832, -0.00000363, -0.00000716, -0.00000138, 0.00000239, 0.00000068, 0.00000058, -0.00000478, -0.00000277, -0.00000686, 0.00000365, -0.00000828, 0.00000909, -0.00000893, 0.00000184, -0.00000457, 0.00000733, -0.00000937, -0.00001237, -0.00000449, 0.00001677, -0.00001821, -0.00001852, 0.00001081, 0.00000390, 0.00002659, -0.00001007, 0.00002110, -0.00000700, 0.00001964, -0.00001349, 0.00001892, -0.00001131, 0.00003490, 0.00000060, 0.00000684, -0.00002477, -0.00001126, -0.00000154, -0.00000539, -0.00000235, -0.00000839, 0.00000591, -0.00000744, -0.00000488, -0.00000538, -0.00000786, -0.00000758, -0.00001810, -0.00001404, -0.00002705, -0.00003078, -0.00020382, 0.00011374, 0.00023654, 0.00002703, 0.00027992, 0.00002247, -0.00018598, -0.00000654, -0.00003826, -0.00000429, -0.00002374, -0.00000324, -0.00000507, -0.00000096, -0.00000856, -0.00000230, 0.00000323, -0.00001278, 0.00001233, -0.00001163, 0.00000390, 0.00000118, -0.00002422, 0.00001305, -0.00000815, 0.00000634, -0.00001728, 0.00000415, -0.00000974, 0.00000358, 0.00000298, 0.00000560, -0.00002496, 0.00001002, -0.00000438, 0.00000973, -0.00002546, 0.00001737, -0.00001760, 0.00000025, 0.00000697, -0.00000598, 0.00002454, -0.00001068, 0.00000107, -0.00000363, 0.00000503, -0.00000068, 0.00001006, -0.00000562, 0.00000075, -0.00000088, 0.00000029, -0.00000043, 0.00000377, -0.00000473, -0.00000065, -0.00000319, 0.00000362, 0.00000530, },
		{ 0.00000342, -0.00000508, 0.00000114, 0.00000654, -0.00000118, -0.00000301, 0.00000856, -0.00000207, -0.00000966, -0.00000128, 0.00001120, -0.00000488, -0.00000470, 0.00000385, -0.00001065, 0.00000093, 0.00000595, 0.00000092, -0.00000859, -0.00000225, 0.00000044, 0.00000317, 0.00000003, -0.00000399, -0.00000299, -0.00000704, 0.00000244, -0.00000622, 0.00000529, -0.00000359, -0.00000438, 0.00000089, 0.00000192, -0.00000186, -0.00001788, -0.00000116, 0.00000803, -0.00000147, -0.00003378, 0.00001627, -0.00000493, 0.00003358, -0.00001112, 0.00002219, -0.00000671, 0.00001999, -0.00001083, 0.00001593, -0.00000895, 0.00002983, 0.00000856, 0.00000414, -0.00001866, -0.00001859, 0.00000097, -0.00000724, -0.00000037, -0.00000974, 0.00000370, -0.00000259, -0.00000887, -0.00000146, -0.00001113, -0.00000324, -0.00002083, -0.00000857, -0.00003396, -0.00001081, -0.00019333, 0.00004924, 0.00024880, 0.00003199, 0.00021050, 0.00014046, -0.00019137, -0.00003618, -0.00002606, -0.00001578, -0.00001509, -0.00001235, -0.00000016, -0.00000482, -0.00000392, -0.00000785, 0.00000826, -0.00001571, 0.00001131, -0.00000780, -0.00000201, 0.00000824, -0.00002740, 0.00001101, -0.00000626, 0.00000578, -0.00001402, -0.00000077, -0.00000625, 0.00000031, 0.00000285, 0.00000832, -0.00002226, 0.00000330, -0.00000026, 0.00000654, -0.00001812, 0.00000513, -0.00000131, -0.00001664, 0.00001855, -0.00001472, 0.00002714, -0.00000733, -0.00000007, -0.00000479, 0.00000605, -0.00000203, 0.00001064, -0.00000425, -0.00000048, -0.00000022, 0.00000041, -0.00000117, 0.00000469, -0.00000514, 0.00000035, -0.00000458, 0.00000243, 0.00000625, },
		{ 0.00000421, -0.00000374, -0.00000310, 0.00000906, 0.00000003, -0.00000307, 0.00000464, 0.00000372, -0.00001187, -0.00000167, 0.00000855, 0.00000138, -0.00001100, 0.00000803, -0.00001152, -0.00000001, 0.00000420, 0.00000383, -0.00000808, -0.00000346, -0.00000177, 0.00000479, 0.00000049, -0.00000353, -0.00000345, -0.00000616, -0.00000037, -0.00000251, 0.00000005, 0.00000271, -0.00001070, 0.00000565, -0.00000278, 0.00000480, -0.00002238, 0.00000208, -0.00000054, 0.00001165, -0.00004074, 0.00001329, -0.00000674, 0.00003242, -0.00000492, 0.00001760, -0.00000144, 0.00001579, -0.00000444, 0.00001073, -0.00000475, 0.00002296, 0.00001707, 0.00000154, -0.00001265, -0.00002428, 0.00000113, -0.00000721, -0.00000055, -0.00000812, -0.00000138, 0.00000358, -0.00001301, 0.00000223, -0.00001424, 0.00000099, -0.00002397, -0.00000228, -0.00004108, 0.00000418, -0.00017608, -0.00000618, 0.00024200, 0.00005755, 0.00013714, 0.00021250, -0.00013945, -0.00009163, -0.00000357, -0.00003440, -0.00000161, -0.00002458, 0.00000717, -0.00001161, 0.00000358, -0.00001552, 0.00001432, -0.00001867, 0.00001031, -0.00000493, -0.00000554, 0.00001099, -0.00002490, 0.00000394, -0.00000044, 0.00000108, -0.00000672, -0.00000824, -0.00000120, -0.00000431, 0.00000468, 0.00000690, -0.00001454, -0.00000682, 0.00000645, -0.00000006, -0.00000730, -0.00000893, 0.00001345, -0.00002786, 0.00002230, -0.00001536, 0.00002250, 0.00000007, -0.00000264, -0.00000469, 0.00000538, -0.00000148, 0.00000926, -0.00000133, -0.00000257, 0.00000080, 0.00000037, -0.00000163, 0.00000507, -0.00000473, 0.00000007, -0.00000380, -0.00000083, 0.00000801, },
		{ 0.00000429, -0.00000062, -0.00000808, 0.00001072, 0.00000130, -0.00000157, -0.00000024, 0.00000837, -0.00001216, -0.00000278, 0.00000580, 0.00000672, -0.00001442, 0.00000811, -0.00000867, -0.00000329, 0.00000434, 0.00000406, -0.00000527, -0.00000530, -0.00000328, 0.00000450, 0.00000252, -0.00000372, -0.00000371, -0.00000505, -0.00000335, 0.00000084, -0.00000419, 0.00000721, -0.00001420, 0.00000697, -0.00000428, 0.00000808, -0.00002315, 0.00000190, -0.00000464, 0.00001668, -0.00003686, 0.00000187, -0.00000248, 0.00002463, 0.00000609, 0.00001002, 0.00000620, 0.00000975, 0.00000295, 0.00000558, -0.00000027, 0.00001561, 0.00002491, -0.00000002, -0.00000750, -0.00002795, -0.00000066, -0.00000621, -0.00000167, -0.00000545, -0.00000646, 0.00000780, -0.00001424, 0.00000300, -0.00001460, 0.00000247, -0.00002448, 0.00000094, -0.00004344, 0.00000947, -0.00014920, -0.00005692, 0.00022656, 0.00008443, 0.00008564, 0.00022606, -0.00004313, -0.00015095, 0.00001260, -0.00004835, 0.00000740, -0.00003260, 0.00001111, -0.00001632, 0.00000914, -0.00002072, 0.00001729, -0.00001834, 0.00000698, -0.00000139, -0.00000808, 0.00001109, -0.00001925, -0.00000492, 0.00000591, -0.00000424, 0.00000070, -0.00001433, 0.00000211, -0.00000765, 0.00000623, 0.00000406, -0.00000570, -0.00001581, 0.00001133, -0.00000534, 0.00000140, -0.00001848, 0.00002052, -0.00002890, 0.00001624, -0.00000815, 0.00001254, 0.00000915, -0.00000516, -0.00000385, 0.00000329, 0.00000059, 0.00000649, 0.00000241, -0.00000484, 0.00000174, 0.00000026, -0.00000161, 0.00000449, -0.00000301, -0.00000188, -0.00000098, -0.00000532, 0.00000980, },
		{ 0.00000472, 0.00000272, -0.00001163, 0.00000979, 0.00000381, -0.00000022, -0.00000356, 0.00000959, -0.00000912, -0.00000570, 0.00000451, 0.00000911, -0.00001328, 0.00000348, -0.00000246, -0.00000833, 0.00000600, 0.00000220, -0.00000115, -0.00000716, -0.00000400, 0.00000231, 0.00000562, -0.00000412, -0.00000376, -0.00000426, -0.00000540, 0.00000238, -0.00000588, 0.00000834, -0.00001353, 0.00000400, -0.00000204, 0.00000740, -0.00001949, -0.00000260, -0.00000321, 0.00001364, -0.00002471, -0.00001347, 0.00000342, 0.00001452, 0.00001735, 0.00000355, 0.00001271, 0.00000501, 0.00000864, 0.00000234, 0.00000334, 0.00000884, 0.00003109, 0.00000008, -0.00000352, -0.00002980, -0.00000349, -0.00000548, -0.00000233, -0.00000364, -0.00000911, 0.00000785, -0.00001104, -0.00000021, -0.00001123, 0.00000003, -0.00002079, -0.00000112, -0.00003840, 0.00000343, -0.00011299, -0.00010425, 0.00020837, 0.00010249, 0.00006381, 0.00019255, 0.00006633, -0.00018524, 0.00000572, -0.00004736, 0.00000447, -0.00003083, 0.00000728, -0.00001498, 0.00000887, -0.00001988, 0.00001415, -0.00001253, -0.00000013, 0.00000405, -0.00001110, 0.00001071, -0.00001357, -0.00001192, 0.00000932, -0.00000679, 0.00000469, -0.00001592, 0.00000134, -0.00000786, 0.00000561, 0.00000246, 0.00000062, -0.00001999, 0.00001120, -0.00000605, 0.00000443, -0.00002025, 0.00001798, -0.00002021, 0.00000318, 0.00000279, 0.00000171, 0.00001628, -0.00000565, -0.00000327, 0.00000069, 0.00000307, 0.00000352, 0.00000582, -0.00000638, 0.00000203, 0.00000033, -0.00000117, 0.00000306, -0.00000029, -0.00000481, 0.00000259, -0.00000932, 0.00001021, },
		{ 0.00000635, 0.00000485, -0.00001216, 0.00000547, 0.00000788, 0.00000010, -0.00000408, 0.00000688, -0.00000319, -0.00000991, 0.00000465, 0.00000850, -0.00000831, -0.00000401, 0.00000469, -0.00001298, 0.00000738, 0.00000010, 0.00000239, -0.00000778, -0.00000457, -0.00000082, 0.00000848, -0.00000378, -0.00000399, -0.00000380, -0.00000623, 0.00000188, -0.00000500, 0.00000643, -0.00000951, -0.00000187, 0.00000229, 0.00000436, -0.00001309, -0.00000959, 0.00000141, 0.00000635, -0.00001026, -0.00002623, 0.00000581, 0.00000653, 0.00002464, 0.00000145, 0.00001562, 0.00000356, 0.00001126, 0.00000162, 0.00000580, 0.00000319, 0.00003503, 0.00000213, -0.00000073, -0.00003022, -0.00000656, -0.00000574, -0.00000192, -0.00000350, -0.00000864, 0.00000381, -0.00000440, -0.00000601, -0.00000552, -0.00000508, -0.00001391, -0.00000758, -0.00002726, -0.00001084, -0.00007260, -0.00014435, 0.00018645, 0.00011255, 0.00006339, 0.00013846, 0.00015361, -0.00017281, -0.00003061, -0.00002994, -0.00001086, -0.00001940, -0.00000406, -0.00000748, 0.00000240, -0.00001278, 0.00000511, -0.00000196, -0.00001009, 0.00001075, -0.00001466, 0.00001086, -0.00000966, -0.00001514, 0.00000826, -0.00000526, 0.00000408, -0.00001243, -0.00000338, -0.00000517, 0.00000265, 0.00000308, 0.00000291, -0.00001841, 0.00000586, -0.00000220, 0.00000197, -0.00001531, 0.00000866, -0.00000672, -0.00001075, 0.00001154, -0.00000515, 0.00001845, -0.00000289, -0.00000356, -0.00000152, 0.00000484, 0.00000149, 0.00000795, -0.00000650, 0.00000131, 0.00000076, -0.00000057, 0.00000130, 0.00000255, -0.00000744, 0.00000520, -0.00001108, 0.00000808, },
		{ 0.00000910, 0.00000541, -0.00000950, -0.00000144, 0.00001241, -0.00000014, -0.00000230, 0.00000173, 0.00000345, -0.00001348, 0.00000482, 0.00000649, -0.00000201, -0.00001108, 0.00000957, -0.00001469, 0.00000652, -0.00000034, 0.00000363, -0.00000623, -0.00000563, -0.00000368, 0.00000973, -0.00000187, -0.00000481, -0.00000344, -0.00000622, 0.00000010, -0.00000282, 0.00000319, -0.00000437, -0.00000794, 0.00000596, 0.00000159, -0.00000677, -0.00001590, 0.00000545, -0.00000014, 0.00000045, -0.00003136, 0.00000185, 0.00000273, 0.00002619, 0.00000448, 0.00001480, 0.00000524, 0.00001132, 0.00000263, 0.00000765, -0.00000128, 0.00003653, 0.00000604, 0.00000109, -0.00002954, -0.00000955, -0.00000687, -0.00000091, -0.00000441, -0.00000617, -0.00000218, 0.00000277, -0.00001137, -0.00000032, -0.00001009, -0.00000671, -0.00001518, -0.00001434, -0.00002699, -0.00003609, -0.00017053, 0.00015583, 0.00012162, 0.00006909, 0.00009114, 0.00019467, -0.00011023, -0.00008696, -0.00000553, -0.00003118, -0.00000452, -0.00001781, 0.00000214, -0.00000686, -0.00000275, -0.00000634, 0.00000978, -0.00001962, 0.00001625, -0.00001726, 0.00001081, -0.00000724, -0.00001512, 0.00000387, -0.00000103, 0.00000043, -0.00000607, -0.00000949, -0.00000180, -0.00000105, 0.00000492, 0.00000220, -0.00001302, -0.00000194, 0.00000338, -0.00000281, -0.00000789, -0.00000183, 0.00000499, -0.00001936, 0.00001353, -0.00000541, 0.00001500, 0.00000256, -0.00000434, -0.00000304, 0.00000533, 0.00000088, 0.00000848, -0.00000515, -0.00000030, 0.00000147, -0.00000000, -0.00000024, 0.00000461, -0.00000860, 0.00000558, -0.00000976, 0.00000335, },
		{ 0.00001183, 0.00000511, -0.00000490, -0.00000875, 0.00001537, 0.00000045, 0.00000010, -0.00000328, 0.00000806, -0.00001432, 0.00000339, 0.00000504, 0.00000290, -0.00001479, 0.00000991, -0.00001217, 0.00000264, 0.00000162, 0.00000215, -0.00000265, -0.00000715, -0.00000565, 0.00000872, 0.00000167, -0.00000609, -0.00000303, -0.00000592, -0.00000184, -0.00000092, 0.00000056, -0.00000038, -0.00001177, 0.00000678, 0.00000108, -0.00000271, -0.00001903, 0.00000603, -0.00000243, 0.00000428, -0.00002789, -0.00000727, 0.00000186, 0.00002322, 0.00001069, 0.00001234, 0.00000813, 0.00001065, 0.00000378, 0.00000975, -0.00000465, 0.00003564, 0.00001145, 0.00000229, -0.00002798, -0.00001257, -0.00000814, -0.00000044, -0.00000502, -0.00000371, -0.00000724, 0.00000728, -0.00001338, 0.00000177, -0.00001242, -0.00000205, -0.00002041, -0.00000435, -0.00003863, -0.00001064, -0.00017759, 0.00011272, 0.00013576, 0.00006881, 0.00006683, 0.00018568, -0.00001453, -0.00014096, 0.00000995, -0.00004535, 0.00000511, -0.00002707, 0.00000830, -0.00001391, 0.00000530, -0.00001541, 0.00001816, -0.00002486, 0.00001768, -0.00001680, 0.00000899, -0.00000474, -0.00001402, -0.00000107, 0.00000296, -0.00000318, -0.00000030, -0.00001361, -0.00000049, -0.00000333, 0.00000610, 0.00000083, -0.00000717, -0.00000844, 0.00000704, -0.00000604, -0.00000253, -0.00000842, 0.00001019, -0.00001958, 0.00000793, -0.00000007, 0.00000815, 0.00000834, -0.00000430, -0.00000425, 0.00000474, 0.00000135, 0.00000787, -0.00000287, -0.00000226, 0.00000213, 0.00000052, -0.00000120, 0.00000537, -0.00000786, 0.00000360, -0.00000588, -0.00000275, },
		{ 0.00001298, 0.00000521, -0.00000024, -0.00001401, 0.00001488, 0.00000303, 0.00000155, -0.00000601, 0.00000886, -0.00001154, -0.00000021, 0.00000520, 0.00000493, -0.00001404, 0.00000565, -0.00000624, -0.00000322, 0.00000502, -0.00000092, 0.00000157, -0.00000824, -0.00000689, 0.00000584, 0.00000599, -0.00000713, -0.00000281, -0.00000563, -0.00000319, -0.00000027, -0.00000043, 0.00000139, -0.00001246, 0.00000430, 0.00000301, -0.00000131, -0.00001851, 0.00000274, -0.00000029, 0.00000217, -0.00001903, -0.00001716, 0.00000055, 0.00001862, 0.00001680, 0.00001110, 0.00000996, 0.00001111, 0.00000382, 0.00001249, -0.00000678, 0.00003250, 0.00001788, 0.00000324, -0.00002561, -0.00001593, -0.00000870, -0.00000154, -0.00000419, -0.00000284, -0.00000938, 0.00000738, -0.00001097, 0.00000004, -0.00001126, -0.00000109, -0.00002149, 0.00000008, -0.00004229, 0.00000074, -0.00016528, 0.00005853, 0.00015512, 0.00005964, 0.00006613, 0.00014205, 0.00008420, -0.00016735, 0.00000291, -0.00004536, 0.00000366, -0.00002743, 0.00000740, -0.00001534, 0.00000793, -0.00001873, 0.00002019, -0.00002358, 0.00001356, -0.00001227, 0.00000438, -0.00000073, -0.00001396, -0.00000400, 0.00000422, -0.00000422, 0.00000221, -0.00001350, -0.00000269, -0.00000309, 0.00000538, 0.00000070, -0.00000351, -0.00001100, 0.00000662, -0.00000559, -0.00000157, -0.00000910, 0.00000814, -0.00001276, -0.00000203, 0.00000680, 0.00000195, 0.00001144, -0.00000206, -0.00000567, 0.00000368, 0.00000214, 0.00000695, -0.00000049, -0.00000386, 0.00000232, 0.00000110, -0.00000156, 0.00000493, -0.00000561, 0.00000016, -0.00000103, -0.00000824, },
		{ 0.00001143, 0.00000654, 0.00000304, -0.00001571, 0.00001042, 0.00000753, 0.00000167, -0.00000585, 0.00000600, -0.00000609, -0.00000502, 0.00000649, 0.00000446, -0.00000993, -0.00000110, 0.00000051, -0.00000867, 0.00000781, -0.00000352, 0.00000449, -0.00000773, -0.00000803, 0.00000214, 0.00000966, -0.00000694, -0.00000317, -0.00000530, -0.00000386, -0.00000088, 0.00000000, 0.00000137, -0.00001082, -0.00000014, 0.00000590, -0.00000123, -0.00001580, -0.00000260, 0.00000384, -0.00000215, -0.00001008, -0.00002285, -0.00000418, 0.00001482, 0.00002020, 0.00001295, 0.00000961, 0.00001333, 0.00000269, 0.00001531, -0.00000708, 0.00002721, 0.00002476, 0.00000429, -0.00002249, -0.00001988, -0.00000809, -0.00000448, -0.00000180, -0.00000389, -0.00000833, 0.00000354, -0.00000538, -0.00000405, -0.00000786, -0.00000288, -0.00001900, -0.00000071, -0.00003848, 0.00000009, -0.00013888, 0.00000019, 0.00017359, 0.00004735, 0.00007772, 0.00008895, 0.00015603, -0.00015001, -0.00002960, -0.00003166, -0.00000813, -0.00001974, 0.00000011, -0.00001133, 0.00000494, -0.00001614, 0.00001598, -0.00001640, 0.00000496, -0.00000465, -0.00000245, 0.00000480, -0.00001557, -0.00000412, 0.00000221, -0.00000225, 0.00000108, -0.00000942, -0.00000757, -0.00000112, 0.00000311, 0.00000211, -0.00000271, -0.00000949, 0.00000263, -0.00000207, -0.00000418, -0.00000551, 0.00000194, -0.00000361, -0.00001096, 0.00001039, 0.00000000, 0.00001000, 0.00000268, -0.00000728, 0.00000253, 0.00000260, 0.00000634, 0.00000141, -0.00000467, 0.00000184, 0.00000178, -0.00000145, 0.00000373, -0.00000275, -0.00000336, 0.00000295, -0.00001134, },
		{ 0.00000729, 0.00000892, 0.00000466, -0.00001391, 0.00000318, 0.00001243, 0.00000139, -0.00000377, 0.00000129, -0.00000027, -0.00000905, 0.00000745, 0.00000312, -0.00000490, -0.00000727, 0.00000514, -0.00001134, 0.00000810, -0.00000392, 0.00000474, -0.00000504, -0.00000945, -0.00000136, 0.00001153, -0.00000493, -0.00000433, -0.00000478, -0.00000417, -0.00000201, 0.00000076, 0.00000092, -0.00000859, -0.00000439, 0.00000766, -0.00000058, -0.00001300, -0.00000741, 0.00000683, -0.00000470, -0.00000536, -0.00002173, -0.00001268, 0.00001212, 0.00002028, 0.00001765, 0.00000782, 0.00001636, 0.00000161, 0.00001691, -0.00000482, 0.00002000, 0.00003147, 0.00000581, -0.00001874, -0.00002431, -0.00000657, -0.00000861, 0.00000124, -0.00000591, -0.00000546, -0.00000201, 0.00000063, -0.00000783, -0.00000459, -0.00000528, -0.00001509, -0.00000429, -0.00003052, -0.00000726, -0.00010678, -0.00005280, 0.00018248, 0.00004116, 0.00008700, 0.00004912, 0.00018348, -0.00008986, -0.00007680, -0.00001363, -0.00002331, -0.00000961, -0.00000911, -0.00000532, -0.00000079, -0.00001041, 0.00000843, -0.00000649, -0.00000501, 0.00000341, -0.00000941, 0.00001031, -0.00001767, -0.00000270, -0.00000146, 0.00000104, -0.00000186, -0.00000376, -0.00001248, 0.00000040, 0.00000089, 0.00000394, -0.00000353, -0.00000593, -0.00000237, 0.00000186, -0.00000745, -0.00000134, -0.00000372, 0.00000264, -0.00001434, 0.00000783, 0.00000341, 0.00000459, 0.00000849, -0.00000816, 0.00000119, 0.00000261, 0.00000611, 0.00000279, -0.00000471, 0.00000084, 0.00000244, -0.00000101, 0.00000231, -0.00000020, -0.00000579, 0.00000486, -0.00001133, },
		{ 0.00000181, 0.00001122, 0.00000538, -0.00001000, -0.00000447, 0.00001554, 0.00000217, -0.00000146, -0.00000291, 0.00000357, -0.00001060, 0.00000674, 0.00000254, -0.00000122, -0.00001046, 0.00000586, -0.00001025, 0.00000537, -0.00000179, 0.00000238, -0.00000077, -0.00001080, -0.00000427, 0.00001121, -0.00000130, -0.00000592, -0.00000412, -0.00000448, -0.00000286, 0.00000076, 0.00000125, -0.00000721, -0.00000689, 0.00000700, 0.00000166, -0.00001140, -0.00001005, 0.00000681, -0.00000354, -0.00000605, -0.00001490, -0.00002222, 0.00000868, 0.00001831, 0.00002320, 0.00000673, 0.00001828, 0.00000234, 0.00001602, 0.00000032, 0.00001155, 0.00003718, 0.00000819, -0.00001462, -0.00002878, -0.00000493, -0.00001260, 0.00000342, -0.00000735, -0.00000276, -0.00000664, 0.00000431, -0.00000900, -0.00000334, -0.00000642, -0.00001184, -0.00000795, -0.00002222, -0.00001563, -0.00007685, -0.00009257, 0.00017543, 0.00004748, 0.00008439, 0.00003378, 0.00016721, -0.00000467, -0.00011845, -0.00000463, -0.00003297, -0.00000383, -0.00001499, -0.00000153, -0.00000545, -0.00000527, 0.00000131, 0.00000236, -0.00001284, 0.00000898, -0.00001410, 0.00001372, -0.00001826, -0.00000202, -0.00000421, 0.00000309, -0.00000394, 0.00000046, -0.00001460, -0.00000062, 0.00000028, 0.00000472, -0.00000407, -0.00000297, -0.00000546, 0.00000349, -0.00000854, 0.00000011, -0.00000535, 0.00000312, -0.00001092, -0.00000017, 0.00001005, -0.00000185, 0.00001271, -0.00000689, -0.00000084, 0.00000252, 0.00000585, 0.00000403, -0.00000438, -0.00000024, 0.00000282, -0.00000030, 0.00000104, 0.00000153, -0.00000664, 0.00000455, -0.00000880, },
		{ -0.00000318, 0.00001205, 0.00000628, -0.00000575, -0.00001020, 0.00001522, 0.00000489, -0.00000006, -0.00000505, 0.00000431, -0.00000924, 0.00000416, 0.00000334, 0.00000016, -0.00001009, 0.00000296, -0.00000638, 0.00000080, 0.00000169, -0.00000115, 0.00000347, -0.00001108, -0.00000682, 0.00000919, 0.00000297, -0.00000716, -0.00000365, -0.00000488, -0.00000306, -0.00000039, 0.00000268, -0.00000698, -0.00000746, 0.00000419, 0.00000500, -0.00001077, -0.00001060, 0.00000410, 0.00000054, -0.00001005, -0.00000645, -0.00002848, 0.00000214, 0.00001607, 0.00002717, 0.00000847, 0.00001754, 0.00000591, 0.00001231, 0.00000764, 0.00000319, 0.00004086, 0.00001186, -0.00001049, -0.00003258, -0.00000424, -0.00001512, 0.00000344, -0.00000705, -0.00000170, -0.00000864, 0.00000433, -0.00000692, -0.00000444, -0.00000585, -0.00001012, -0.00001016, -0.00001598, -0.00002144, -0.00005367, -0.00011580, 0.00015152, 0.00006639, 0.00006919, 0.00004056, 0.00012317, 0.00007889, -0.00013421, -0.00001429, -0.00003194, -0.00000606, -0.00001483, -0.00000222, -0.00000678, -0.00000313, -0.00000287, 0.00000764, -0.00001637, 0.00001042, -0.00001513, 0.00001366, -0.00001588, -0.00000384, -0.00000416, 0.00000220, -0.00000344, 0.00000141, -0.00001264, -0.00000472, 0.00000158, 0.00000392, -0.00000322, -0.00000225, -0.00000520, 0.00000179, -0.00000644, -0.00000212, -0.00000248, -0.00000145, -0.00000322, -0.00000973, 0.00001566, -0.00000548, 0.00001276, -0.00000255, -0.00000374, 0.00000270, 0.00000504, 0.00000556, -0.00000407, -0.00000106, 0.00000271, 0.00000065, 0.00000007, 0.00000236, -0.00000617, 0.00000277, -0.00000517, },
		{ -0.00000625, 0.00001064, 0.00000794, -0.00000236, -0.00001279, 0.00001132, 0.00000906, 0.00000053, -0.00000508, 0.00000249, -0.00000609, 0.00000075, 0.00000490, -0.00000021, -0.00000733, -0.00000159, -0.00000200, -0.00000344, 0.00000453, -0.00000383, 0.00000589, -0.00000937, -0.00000941, 0.00000633, 0.00000664, -0.00000721, -0.00000374, -0.00000523, -0.00000275, -0.00000229, 0.00000452, -0.00000709, -0.00000714, 0.00000070, 0.00000787, -0.00000978, -0.00001045, 0.00000052, 0.00000500, -0.00001370, -0.00000097, -0.00002814, -0.00000831, 0.00001422, 0.00002816, 0.00001379, 0.00001411, 0.00001185, 0.00000693, 0.00001538, -0.00000328, 0.00004145, 0.00001715, -0.00000670, -0.00003506, -0.00000536, -0.00001533, 0.00000088, -0.00000482, -0.00000255, -0.00000786, 0.00000126, -0.00000281, -0.00000659, -0.00000453, -0.00000938, -0.00001102, -0.00001215, -0.00002390, -0.00003789, -0.00012406, 0.00011518, 0.00009204, 0.00004807, 0.00005794, 0.00007364, 0.00013672, -0.00011327, -0.00004243, -0.00002230, -0.00001457, -0.00001011, -0.00000628, -0.00000532, -0.00000387, -0.00000403, 0.00000923, -0.00001573, 0.00000810, -0.00001285, 0.00001034, -0.00001054, -0.00000829, -0.00000126, -0.00000140, -0.00000060, -0.00000061, -0.00000753, -0.00001041, 0.00000351, 0.00000231, -0.00000124, -0.00000361, -0.00000226, -0.00000208, -0.00000248, -0.00000646, 0.00000253, -0.00000749, 0.00000410, -0.00001601, 0.00001627, -0.00000368, 0.00000781, 0.00000431, -0.00000688, 0.00000303, 0.00000363, 0.00000736, -0.00000379, -0.00000156, 0.00000218, 0.00000168, -0.00000054, 0.00000252, -0.00000501, 0.00000062, -0.00000199, },
		{ -0.00000697, 0.00000731, 0.00000997, 0.00000005, -0.00001250, 0.00000526, 0.00001295, 0.00000140, -0.00000410, -0.00000018, -0.00000306, -0.00000195, 0.00000606, -0.00000098, -0.00000421, -0.00000534, 0.00000060, -0.00000551, 0.00000518, -0.00000431, 0.00000558, -0.00000557, -0.00001198, 0.00000328, 0.00000887, -0.00000571, -0.00000442, -0.00000540, -0.00000236, -0.00000409, 0.00000562, -0.00000633, -0.00000734, -0.00000178, 0.00000872, -0.00000722, -0.00001095, -0.00000208, 0.00000744, -0.00001411, -0.00000102, -0.00002096, -0.00002080, 0.00001169, 0.00002639, 0.00002166, 0.00000961, 0.00001829, 0.00000206, 0.00002125, -0.00000612, 0.00003833, 0.00002401, -0.00000332, -0.00003582, -0.00000862, -0.00001328, -0.00000367, -0.00000155, -0.00000444, -0.00000550, -0.00000301, 0.00000111, -0.00000786, -0.00000393, -0.00000848, -0.00001146, -0.00000978, -0.00002419, -0.00002761, -0.00012181, 0.00007353, 0.00011621, 0.00003017, 0.00007285, 0.00003706, 0.00015667, -0.00005957, -0.00007808, -0.00001256, -0.00002345, -0.00000541, -0.00001026, -0.00000361, -0.00000543, -0.00000399, 0.00000885, -0.00001278, 0.00000394, -0.00000895, 0.00000526, -0.00000371, -0.00001386, 0.00000276, -0.00000575, 0.00000265, -0.00000356, -0.00000187, -0.00001499, 0.00000395, 0.00000142, 0.00000064, -0.00000562, 0.00000126, -0.00000563, 0.00000084, -0.00001009, 0.00000622, -0.00001097, 0.00000692, -0.00001586, 0.00001043, 0.00000333, -0.00000046, 0.00001146, -0.00000877, 0.00000289, 0.00000206, 0.00000892, -0.00000316, -0.00000202, 0.00000153, 0.00000252, -0.00000072, 0.00000229, -0.00000375, -0.00000106, -0.00000012, },
		{ -0.00000597, 0.00000330, 0.00001135, 0.00000216, -0.00001064, -0.00000074, 0.00001462, 0.00000372, -0.00000331, -0.00000207, -0.00000162, -0.00000295, 0.00000605, -0.00000107, -0.00000229, -0.00000673, 0.00000037, -0.00000491, 0.00000347, -0.00000259, 0.00000302, -0.00000071, -0.00001381, 0.00000014, 0.00000951, -0.00000303, -0.00000532, -0.00000541, -0.00000226, -0.00000507, 0.00000517, -0.00000399, -0.00000877, -0.00000249, 0.00000709, -0.00000296, -0.00001242, -0.00000303, 0.00000709, -0.00001073, -0.00000593, -0.00001002, -0.00003129, 0.00000642, 0.00002304, 0.00002988, 0.00000653, 0.00002287, -0.00000007, 0.00002348, -0.00000452, 0.00003177, 0.00003176, -0.00000003, -0.00003493, -0.00001368, -0.00000982, -0.00000885, 0.00000133, -0.00000604, -0.00000318, -0.00000643, 0.00000290, -0.00000694, -0.00000487, -0.00000675, -0.00001222, -0.00000784, -0.00002375, -0.00002045, -0.00011368, 0.00003311, 0.00013225, 0.00002203, 0.00007711, 0.00002158, 0.00014164, 0.00001022, -0.00010432, -0.00001270, -0.00002658, -0.00000516, -0.00001088, -0.00000418, -0.00000562, -0.00000480, 0.00000851, -0.00000965, 0.00000004, -0.00000535, 0.00000029, 0.00000261, -0.00001838, 0.00000549, -0.00000836, 0.00000402, -0.00000510, 0.00000172, -0.00001613, 0.00000140, 0.00000226, 0.00000137, -0.00000673, 0.00000323, -0.00000668, 0.00000151, -0.00001086, 0.00000623, -0.00000966, 0.00000387, -0.00000955, 0.00000012, 0.00001244, -0.00000842, 0.00001581, -0.00000775, 0.00000154, 0.00000103, 0.00000952, -0.00000165, -0.00000282, 0.00000110, 0.00000292, -0.00000044, 0.00000182, -0.00000267, -0.00000202, 0.00000042, },
		{ -0.00000442, 0.00000006, 0.00001106, 0.00000467, -0.00000856, -0.00000488, 0.00001306, 0.00000767, -0.00000308, -0.00000259, -0.00000201, -0.00000249, 0.00000508, -0.00000028, -0.00000198, -0.00000570, -0.00000205, -0.00000270, 0.00000058, 0.00000002, -0.00000019, 0.00000348, -0.00001389, -0.00000332, 0.00000901, -0.00000003, -0.00000580, -0.00000548, -0.00000256, -0.00000498, 0.00000311, -0.00000032, -0.00001106, -0.00000192, 0.00000393, 0.00000181, -0.00001390, -0.00000312, 0.00000497, -0.00000529, -0.00001259, -0.00000008, -0.00003576, -0.00000289, 0.00001904, 0.00003632, 0.00000691, 0.00002403, 0.00000167, 0.00002164, 0.00000097, 0.00002310, 0.00003909, 0.00000387, -0.00003286, -0.00001970, -0.00000624, -0.00001316, 0.00000253, -0.00000636, -0.00000202, -0.00000782, 0.00000193, -0.00000392, -0.00000693, -0.00000452, -0.00001318, -0.00000607, -0.00002324, -0.00001509, -0.00010265, -0.00000231, 0.00013738, 0.00002509, 0.00006989, 0.00002448, 0.00010590, 0.00007429, -0.00010644, -0.00002800, -0.00002193, -0.00001062, -0.00000743, -0.00000756, -0.00000377, -0.00000723, 0.00000907, -0.00000741, -0.00000250, -0.00000307, -0.00000338, 0.00000697, -0.00002023, 0.00000527, -0.00000774, 0.00000234, -0.00000407, 0.00000203, -0.00001322, -0.00000395, 0.00000452, 0.00000090, -0.00000636, 0.00000278, -0.00000463, -0.00000072, -0.00000858, 0.00000254, -0.00000423, -0.00000316, -0.00000047, -0.00001032, 0.00001919, -0.00001224, 0.00001504, -0.00000315, -0.00000114, 0.00000091, 0.00000873, 0.00000093, -0.00000405, 0.00000103, 0.00000280, 0.00000025, 0.00000125, -0.00000182, -0.00000244, 0.00000023, },
		{ -0.00000334, -0.00000151, 0.00000886, 0.00000763, -0.00000689, -0.00000657, 0.00000878, 0.00001209, -0.00000264, -0.00000234, -0.00000319, -0.00000178, 0.00000409, 0.00000079, -0.00000265, -0.00000348, -0.00000491, -0.00000071, -0.00000175, 0.00000185, -0.00000230, 0.00000552, -0.00001162, -0.00000716, 0.00000786, 0.00000247, -0.00000537, -0.00000573, -0.00000318, -0.00000408, 0.00000011, 0.00000360, -0.00001298, -0.00000144, 0.00000092, 0.00000533, -0.00001390, -0.00000379, 0.00000289, -0.00000033, -0.00001743, 0.00000493, -0.00003241, -0.00001531, 0.00001408, 0.00003983, 0.00001141, 0.00002183, 0.00000674, 0.00001692, 0.00000849, 0.00001444, 0.00004440, 0.00000919, -0.00003023, -0.00002569, -0.00000379, -0.00001554, 0.00000156, -0.00000525, -0.00000216, -0.00000724, -0.00000102, -0.00000018, -0.00000872, -0.00000290, -0.00001356, -0.00000498, -0.00002233, -0.00001150, -0.00008980, -0.00003173, 0.00013255, 0.00003636, 0.00005590, 0.00003652, 0.00006714, 0.00011505, -0.00007913, -0.00005510, -0.00001333, -0.00001888, -0.00000230, -0.00001188, -0.00000120, -0.00001027, 0.00000981, -0.00000564, -0.00000399, -0.00000192, -0.00000567, 0.00000902, -0.00001906, 0.00000193, -0.00000411, -0.00000178, -0.00000112, -0.00000024, -0.00000768, -0.00001019, 0.00000657, 0.00000025, -0.00000510, 0.00000052, -0.00000064, -0.00000434, -0.00000485, -0.00000282, 0.00000244, -0.00001029, 0.00000681, -0.00001643, 0.00002015, -0.00000991, 0.00000890, 0.00000403, -0.00000423, 0.00000138, 0.00000676, 0.00000418, -0.00000531, 0.00000108, 0.00000237, 0.00000114, 0.00000068, -0.00000111, -0.00000269, -0.00000000, },
		{ -0.00000308, -0.00000148, 0.00000548, 0.00001032, -0.00000532, -0.00000652, 0.00000358, 0.00001503, -0.00000072, -0.00000245, -0.00000376, -0.00000209, 0.00000396, 0.00000149, -0.00000335, -0.00000154, -0.00000652, -0.00000042, -0.00000236, 0.00000188, -0.00000243, 0.00000499, -0.00000739, -0.00001085, 0.00000621, 0.00000409, -0.00000407, -0.00000599, -0.00000405, -0.00000285, -0.00000285, 0.00000649, -0.00001328, -0.00000231, -0.00000052, 0.00000632, -0.00001154, -0.00000589, 0.00000213, 0.00000240, -0.00001845, 0.00000384, -0.00002280, -0.00002756, 0.00000682, 0.00004027, 0.00001923, 0.00001778, 0.00001329, 0.00001155, 0.00001562, 0.00000792, 0.00004641, 0.00001639, -0.00002743, -0.00003091, -0.00000317, -0.00001573, -0.00000120, -0.00000343, -0.00000292, -0.00000569, -0.00000433, 0.00000247, -0.00000880, -0.00000284, -0.00001264, -0.00000513, -0.00002039, -0.00001017, -0.00007527, -0.00005563, 0.00012044, 0.00005104, 0.00004153, 0.00004797, 0.00003875, 0.00012559, -0.00002902, -0.00008288, -0.00000856, -0.00002478, 0.00000058, -0.00001420, -0.00000015, -0.00001207, 0.00000918, -0.00000315, -0.00000545, -0.00000098, -0.00000722, 0.00000935, -0.00001569, -0.00000328, 0.00000082, -0.00000634, 0.00000186, -0.00000306, -0.00000207, -0.00001469, 0.00000650, 0.00000071, -0.00000402, -0.00000212, 0.00000329, -0.00000713, -0.00000194, -0.00000721, 0.00000714, -0.00001396, 0.00000897, -0.00001603, 0.00001471, -0.00000235, -0.00000033, 0.00001119, -0.00000604, 0.00000159, 0.00000440, 0.00000718, -0.00000588, 0.00000075, 0.00000202, 0.00000190, 0.00000029, -0.00000051, -0.00000298, 0.00000007, },
		{ -0.00000332, -0.00000069, 0.00000223, 0.00001163, -0.00000314, -0.00000602, -0.00000051, 0.00001490, 0.00000337, -0.00000351, -0.00000300, -0.00000384, 0.00000470, 0.00000175, -0.00000361, -0.00000074, -0.00000617, -0.00000205, -0.00000133, 0.00000034, -0.00000098, 0.00000278, -0.00000258, -0.00001333, 0.00000382, 0.00000499, -0.00000238, -0.00000588, -0.00000513, -0.00000174, -0.00000499, 0.00000752, -0.00001138, -0.00000495, -0.00000002, 0.00000479, -0.00000722, -0.00000909, 0.00000258, 0.00000291, -0.00001600, -0.00000155, -0.00001109, -0.00003545, -0.00000384, 0.00003777, 0.00002874, 0.00001411, 0.00001911, 0.00000776, 0.00002040, 0.00000488, 0.00004470, 0.00002523, -0.00002431, -0.00003513, -0.00000443, -0.00001428, -0.00000469, -0.00000197, -0.00000337, -0.00000431, -0.00000656, 0.00000283, -0.00000665, -0.00000443, -0.00001045, -0.00000653, -0.00001729, -0.00001113, -0.00005955, -0.00007438, 0.00010354, 0.00006523, 0.00003129, 0.00005315, 0.00002572, 0.00011070, 0.00002847, -0.00009751, -0.00001469, -0.00002446, -0.00000149, -0.00001262, -0.00000203, -0.00001137, 0.00000616, 0.00000084, -0.00000763, 0.00000039, -0.00000861, 0.00000871, -0.00001138, -0.00000861, 0.00000491, -0.00000919, 0.00000299, -0.00000451, 0.00000139, -0.00001562, 0.00000335, 0.00000277, -0.00000376, -0.00000398, 0.00000554, -0.00000753, -0.00000128, -0.00000897, 0.00000806, -0.00001273, 0.00000556, -0.00001016, 0.00000534, 0.00000700, -0.00000886, 0.00001525, -0.00000499, 0.00000078, 0.00000252, 0.00000900, -0.00000514, -0.00000030, 0.00000204, 0.00000228, 0.00000023, -0.00000008, -0.00000327, 0.00000035, },
		{ -0.00000344, -0.00000017, 0.00000030, 0.00001085, 0.00000009, -0.00000603, -0.00000229, 0.00001142, 0.00000899, -0.00000494, -0.00000141, -0.00000619, 0.00000536, 0.00000222, -0.00000360, -0.00000099, -0.00000435, -0.00000462, 0.00000019, -0.00000158, 0.00000072, 0.00000051, 0.00000109, -0.00001362, 0.00000047, 0.00000552, -0.00000095, -0.00000507, -0.00000634, -0.00000108, -0.00000592, 0.00000660, -0.00000766, -0.00000871, 0.00000159, 0.00000195, -0.00000243, -0.00001199, 0.00000302, 0.00000254, -0.00001212, -0.00000784, -0.00000187, -0.00003616, -0.00001733, 0.00003194, 0.00003827, 0.00001269, 0.00002266, 0.00000679, 0.00002204, 0.00000540, 0.00003995, 0.00003469, -0.00002026, -0.00003859, -0.00000709, -0.00001216, -0.00000775, -0.00000172, -0.00000299, -0.00000371, -0.00000712, 0.00000092, -0.00000299, -0.00000670, -0.00000788, -0.00000842, -0.00001364, -0.00001359, -0.00004393, -0.00008751, 0.00008308, 0.00007713, 0.00002640, 0.00005150, 0.00002519, 0.00008269, 0.00007611, -0.00008924, -0.00003339, -0.00001826, -0.00000803, -0.00000778, -0.00000626, -0.00000856, 0.00000113, 0.00000587, -0.00001020, 0.00000203, -0.00000980, 0.00000749, -0.00000709, -0.00001268, 0.00000672, -0.00000918, 0.00000151, -0.00000386, 0.00000192, -0.00001287, -0.00000217, 0.00000563, -0.00000397, -0.00000486, 0.00000571, -0.00000545, -0.00000277, -0.00000816, 0.00000553, -0.00000775, -0.00000110, -0.00000232, -0.00000381, 0.00001402, -0.00001320, 0.00001426, -0.00000069, -0.00000106, 0.00000149, 0.00000922, -0.00000297, -0.00000201, 0.00000241, 0.00000228, 0.00000050, 0.00000013, -0.00000338, 0.00000048, },
		{ -0.00000300, -0.00000052, 0.00000011, 0.00000816, 0.00000399, -0.00000657, -0.00000193, 0.00000593, 0.00001423, -0.00000524, -0.00000030, -0.00000766, 0.00000465, 0.00000367, -0.00000377, -0.00000169, -0.00000221, -0.00000667, 0.00000081, -0.00000259, 0.00000136, -0.00000047, 0.00000255, -0.00001147, -0.00000357, 0.00000577, -0.00000012, -0.00000356, -0.00000739, -0.00000110, -0.00000570, 0.00000428, -0.00000316, -0.00001231, 0.00000293, -0.00000056, 0.00000109, -0.00001311, 0.00000211, 0.00000284, -0.00000892, -0.00001212, 0.00000214, -0.00002980, -0.00003088, 0.00002176, 0.00004646, 0.00001457, 0.00002360, 0.00000861, 0.00002108, 0.00000845, 0.00003361, 0.00004335, -0.00001446, -0.00004168, -0.00001050, -0.00001030, -0.00000964, -0.00000287, -0.00000197, -0.00000368, -0.00000644, -0.00000217, 0.00000061, -0.00000814, -0.00000613, -0.00000977, -0.00001044, -0.00001619, -0.00003026, -0.00009410, 0.00005946, 0.00008652, 0.00002567, 0.00004593, 0.00003038, 0.00005482, 0.00010209, -0.00005724, -0.00005905, -0.00001108, -0.00001552, -0.00000251, -0.00001059, -0.00000541, -0.00000427, 0.00001042, -0.00001196, 0.00000295, -0.00001013, 0.00000557, -0.00000316, -0.00001502, 0.00000603, -0.00000661, -0.00000181, -0.00000184, 0.00000039, -0.00000800, -0.00000800, 0.00000759, -0.00000362, -0.00000536, 0.00000451, -0.00000207, -0.00000502, -0.00000620, 0.00000143, -0.00000173, -0.00000746, 0.00000352, -0.00000909, 0.00001598, -0.00001191, 0.00000845, 0.00000555, -0.00000293, 0.00000098, 0.00000814, 0.00000004, -0.00000381, 0.00000275, 0.00000213, 0.00000095, 0.00000016, -0.00000320, 0.00000018, },
		{ -0.00000198, -0.00000168, 0.00000118, 0.00000463, 0.00000747, -0.00000688, -0.00000067, 0.00000065, 0.00001686, -0.00000295, -0.00000084, -0.00000713, 0.00000189, 0.00000626, -0.00000421, -0.00000234, -0.00000072, -0.00000717, -0.00000019, -0.00000218, 0.00000051, 0.00000019, 0.00000189, -0.00000773, -0.00000736, 0.00000543, 0.00000028, -0.00000185, -0.00000780, -0.00000193, -0.00000462, 0.00000136, 0.00000094, -0.00001459, 0.00000287, -0.00000154, 0.00000231, -0.00001179, -0.00000062, 0.00000442, -0.00000731, -0.00001339, 0.00000112, -0.00001949, -0.00004043, 0.00000655, 0.00005194, 0.00001995, 0.00002258, 0.00001224, 0.00001884, 0.00001249, 0.00002729, 0.00004990, -0.00000634, -0.00004456, -0.00001425, -0.00000916, -0.00001032, -0.00000483, -0.00000112, -0.00000353, -0.00000546, -0.00000495, 0.00000257, -0.00000764, -0.00000592, -0.00000995, -0.00000840, -0.00001779, -0.00001999, -0.00009368, 0.00003329, 0.00009351, 0.00002731, 0.00003987, 0.00003507, 0.00003582, 0.00010391, -0.00001005, -0.00008085, -0.00000972, -0.00001969, -0.00000008, -0.00001255, -0.00000391, -0.00000818, 0.00001283, -0.00001161, 0.00000219, -0.00000895, 0.00000270, 0.00000045, -0.00001589, 0.00000363, -0.00000288, -0.00000531, -0.00000003, -0.00000152, -0.00000328, -0.00001186, 0.00000713, -0.00000186, -0.00000615, 0.00000296, 0.00000110, -0.00000645, -0.00000470, -0.00000217, 0.00000266, -0.00001072, 0.00000500, -0.00000918, 0.00001283, -0.00000625, 0.00000025, 0.00001118, -0.00000330, 0.00000028, 0.00000650, 0.00000296, -0.00000493, 0.00000259, 0.00000216, 0.00000134, 0.00000012, -0.00000275, -0.00000055, },
		{ -0.00000079, -0.00000302, 0.00000248, 0.00000163, 0.00000936, -0.00000606, 0.00000005, -0.00000245, 0.00001556, 0.00000220, -0.00000302, -0.00000472, -0.00000230, 0.00000910, -0.00000429, -0.00000300, -0.00000013, -0.00000612, -0.00000241, -0.00000090, -0.00000126, 0.00000173, 0.00000031, -0.00000398, -0.00000967, 0.00000398, 0.00000068, -0.00000060, -0.00000714, -0.00000354, -0.00000312, -0.00000140, 0.00000385, -0.00001497, 0.00000109, -0.00000081, 0.00000139, -0.00000860, -0.00000445, 0.00000659, -0.00000668, -0.00001266, -0.00000248, -0.00000968, -0.00004267, -0.00001275, 0.00005280, 0.00002855, 0.00002081, 0.00001638, 0.00001667, 0.00001623, 0.00002218, 0.00005359, 0.00000410, -0.00004692, -0.00001837, -0.00000868, -0.00001035, -0.00000664, -0.00000131, -0.00000261, -0.00000500, -0.00000638, 0.00000218, -0.00000510, -0.00000703, -0.00000913, -0.00000755, -0.00001803, -0.00001356, -0.00008702, 0.00000615, 0.00009748, 0.00003040, 0.00003521, 0.00003628, 0.00002774, 0.00008776, 0.00003833, -0.00008779, -0.00001880, -0.00001841, -0.00000194, -0.00001115, -0.00000487, -0.00000973, 0.00001235, -0.00000873, -0.00000048, -0.00000623, -0.00000097, 0.00000370, -0.00001577, 0.00000061, 0.00000047, -0.00000741, 0.00000027, -0.00000244, -0.00000037, -0.00001248, 0.00000386, 0.00000123, -0.00000729, 0.00000164, 0.00000312, -0.00000625, -0.00000445, -0.00000415, 0.00000421, -0.00001009, 0.00000232, -0.00000550, 0.00000700, 0.00000065, -0.00000690, 0.00001357, -0.00000104, -0.00000108, 0.00000502, 0.00000495, -0.00000484, 0.00000170, 0.00000250, 0.00000155, 0.00000013, -0.00000218, -0.00000146, },
		{ 0.00000009, -0.00000387, 0.00000306, 0.00000017, 0.00000907, -0.00000374, -0.00000052, -0.00000261, 0.00001071, 0.00000886, -0.00000547, -0.00000183, -0.00000626, 0.00001065, -0.00000308, -0.00000408, -0.00000008, -0.00000430, -0.00000469, 0.00000011, -0.00000279, 0.00000286, -0.00000067, -0.00000174, -0.00000966, 0.00000127, 0.00000136, -0.00000024, -0.00000535, -0.00000559, -0.00000171, -0.00000348, 0.00000527, -0.00001361, -0.00000193, 0.00000094, -0.00000064, -0.00000490, -0.00000797, 0.00000802, -0.00000564, -0.00001175, -0.00000573, -0.00000385, -0.00003696, -0.00003283, 0.00004674, 0.00003960, 0.00001959, 0.00002001, 0.00001546, 0.00001901, 0.00001873, 0.00005439, 0.00001622, -0.00004804, -0.00002321, -0.00000845, -0.00001049, -0.00000748, -0.00000297, -0.00000084, -0.00000521, -0.00000634, -0.00000014, -0.00000155, -0.00000833, -0.00000818, -0.00000725, -0.00001727, -0.00001037, -0.00007604, -0.00001947, 0.00009708, 0.00003512, 0.00003194, 0.00003441, 0.00002740, 0.00006425, 0.00007459, -0.00007404, -0.00003748, -0.00001344, -0.00000677, -0.00000754, -0.00000736, -0.00000951, 0.00000959, -0.00000409, -0.00000426, -0.00000278, -0.00000469, 0.00000624, -0.00001488, -0.00000227, 0.00000251, -0.00000746, -0.00000117, -0.00000209, 0.00000037, -0.00001020, -0.00000108, 0.00000444, -0.00000806, 0.00000037, 0.00000404, -0.00000472, -0.00000509, -0.00000471, 0.00000333, -0.00000681, -0.00000229, -0.00000107, 0.00000183, 0.00000554, -0.00001032, 0.00001151, 0.00000367, -0.00000275, 0.00000384, 0.00000578, -0.00000363, 0.00000031, 0.00000299, 0.00000166, 0.00000021, -0.00000164, -0.00000224, },
		{ 0.00000035, -0.00000387, 0.00000252, 0.00000042, 0.00000697, -0.00000048, -0.00000208, -0.00000057, 0.00000424, 0.00001456, -0.00000616, -0.00000026, -0.00000822, 0.00000965, -0.00000004, -0.00000576, -0.00000014, -0.00000267, -0.00000591, -0.00000009, -0.00000323, 0.00000267, -0.00000015, -0.00000159, -0.00000751, -0.00000211, 0.00000207, -0.00000066, -0.00000291, -0.00000741, -0.00000093, -0.00000458, 0.00000532, -0.00001110, -0.00000535, 0.00000275, -0.00000257, -0.00000201, -0.00000997, 0.00000767, -0.00000322, -0.00001191, -0.00000698, -0.00000290, -0.00002591, -0.00004876, 0.00003200, 0.00005151, 0.00002025, 0.00002252, 0.00001551, 0.00002076, 0.00001688, 0.00005272, 0.00002905, -0.00004702, -0.00002913, -0.00000813, -0.00001122, -0.00000710, -0.00000579, 0.00000114, -0.00000554, -0.00000550, -0.00000311, 0.00000143, -0.00000851, -0.00000801, -0.00000679, -0.00001620, -0.00000924, -0.00006318, -0.00004095, 0.00009102, 0.00004196, 0.00002909, 0.00003155, 0.00002972, 0.00004320, 0.00009123, -0.00004171, -0.00005915, -0.00000972, -0.00001130, -0.00000427, -0.00000947, -0.00000901, 0.00000599, 0.00000079, -0.00000774, 0.00000012, -0.00000745, 0.00000751, -0.00001319, -0.00000482, 0.00000322, -0.00000595, -0.00000350, -0.00000123, -0.00000018, -0.00000655, -0.00000570, 0.00000618, -0.00000749, -0.00000138, 0.00000445, -0.00000282, -0.00000565, -0.00000476, 0.00000141, -0.00000304, -0.00000608, 0.00000128, -0.00000033, 0.00000677, -0.00000941, 0.00000584, 0.00000915, -0.00000357, 0.00000256, 0.00000582, -0.00000191, -0.00000097, 0.00000325, 0.00000187, 0.00000028, -0.00000122, -0.00000266, },
		{ 0.00000000, -0.00000312, 0.00000114, 0.00000171, 0.00000419, 0.00000260, -0.00000365, 0.00000208, -0.00000132, 0.00001696, -0.00000364, -0.00000102, -0.00000746, 0.00000606, 0.00000431, -0.00000753, -0.00000032, -0.00000168, -0.00000568, -0.00000159, -0.00000259, 0.00000118, 0.00000168, -0.00000291, -0.00000448, -0.00000493, 0.00000213, -0.00000126, -0.00000071, -0.00000820, -0.00000121, -0.00000465, 0.00000433, -0.00000805, -0.00000848, 0.00000393, -0.00000367, -0.00000056, -0.00001012, 0.00000548, 0.00000046, -0.00001299, -0.00000653, -0.00000508, -0.00001418, -0.00005596, 0.00000897, 0.00006130, 0.00002421, 0.00002364, 0.00001678, 0.00002168, 0.00001642, 0.00004913, 0.00004162, -0.00004318, -0.00003621, -0.00000774, -0.00001252, -0.00000595, -0.00000876, 0.00000227, -0.00000516, -0.00000484, -0.00000538, 0.00000254, -0.00000684, -0.00000895, -0.00000588, -0.00001520, -0.00000917, -0.00005051, -0.00005650, 0.00007899, 0.00005070, 0.00002606, 0.00002946, 0.00003081, 0.00003003, 0.00008860, 0.00000014, -0.00007426, -0.00001247, -0.00001259, -0.00000350, -0.00000962, -0.00000946, 0.00000286, 0.00000455, -0.00000973, 0.00000150, -0.00000852, 0.00000719, -0.00001063, -0.00000718, 0.00000316, -0.00000396, -0.00000542, -0.00000101, -0.00000079, -0.00000334, -0.00000822, 0.00000538, -0.00000520, -0.00000381, 0.00000483, -0.00000136, -0.00000537, -0.00000507, -0.00000033, -0.00000050, -0.00000733, 0.00000042, 0.00000076, 0.00000499, -0.00000574, -0.00000091, 0.00001296, -0.00000224, 0.00000068, 0.00000566, -0.00000047, -0.00000156, 0.00000297, 0.00000231, 0.00000032, -0.00000095, -0.00000267, },
		{ -0.00000069, -0.00000209, -0.00000041, 0.00000305, 0.00000195, 0.00000442, -0.00000423, 0.00000380, -0.00000412, 0.00001505, 0.00000191, -0.00000355, -0.00000481, 0.00000124, 0.00000841, -0.00000829, -0.00000106, -0.00000117, -0.00000445, -0.00000361, -0.00000166, -0.00000076, 0.00000377, -0.00000426, -0.00000221, -0.00000596, 0.00000096, -0.00000146, 0.00000049, -0.00000750, -0.00000259, -0.00000391, 0.00000272, -0.00000498, -0.00001095, 0.00000428, -0.00000386, -0.00000037, -0.00000893, 0.00000227, 0.00000434, -0.00001381, -0.00000593, -0.00000749, -0.00000614, -0.00005255, -0.00001881, 0.00006475, 0.00003262, 0.00002344, 0.00001908, 0.00002197, 0.00001721, 0.00004409, 0.00005310, -0.00003618, -0.00004404, -0.00000772, -0.00001386, -0.00000493, -0.00001075, 0.00000166, -0.00000357, -0.00000494, -0.00000622, 0.00000146, -0.00000360, -0.00001048, -0.00000490, -0.00001421, -0.00000962, -0.00003923, -0.00006572, 0.00006206, 0.00006007, 0.00002328, 0.00002854, 0.00002944, 0.00002502, 0.00007317, 0.00003959, -0.00007480, -0.00002389, -0.00001019, -0.00000551, -0.00000769, -0.00001099, 0.00000059, 0.00000671, -0.00000989, 0.00000112, -0.00000790, 0.00000547, -0.00000738, -0.00000947, 0.00000292, -0.00000245, -0.00000599, -0.00000205, -0.00000072, -0.00000160, -0.00000802, 0.00000229, -0.00000185, -0.00000638, 0.00000507, -0.00000045, -0.00000428, -0.00000564, -0.00000146, 0.00000029, -0.00000605, -0.00000274, 0.00000334, 0.00000245, -0.00000198, -0.00000588, 0.00001313, 0.00000174, -0.00000179, 0.00000559, 0.00000028, -0.00000131, 0.00000223, 0.00000289, 0.00000039, -0.00000084, -0.00000237, },
		{ -0.00000142, -0.00000124, -0.00000151, 0.00000367, 0.00000102, 0.00000457, -0.00000355, 0.00000396, -0.00000373, 0.00000971, 0.00000863, -0.00000593, -0.00000220, -0.00000271, 0.00001050, -0.00000703, -0.00000276, -0.00000073, -0.00000305, -0.00000507, -0.00000141, -0.00000218, 0.00000495, -0.00000433, -0.00000189, -0.00000472, -0.00000138, -0.00000113, 0.00000046, -0.00000545, -0.00000460, -0.00000288, 0.00000096, -0.00000220, -0.00001269, 0.00000394, -0.00000343, -0.00000089, -0.00000725, -0.00000088, 0.00000721, -0.00001317, -0.00000661, -0.00000801, -0.00000370, -0.00004059, -0.00004499, 0.00005759, 0.00004556, 0.00002266, 0.00002204, 0.00002182, 0.00001912, 0.00003806, 0.00006278, -0.00002600, -0.00005193, -0.00000875, -0.00001459, -0.00000489, -0.00001103, -0.00000088, -0.00000097, -0.00000566, -0.00000578, -0.00000113, 0.00000003, -0.00001147, -0.00000459, -0.00001283, -0.00001051, -0.00002970, -0.00006933, 0.00004235, 0.00006808, 0.00002185, 0.00002801, 0.00002653, 0.00002509, 0.00005368, 0.00006685, -0.00005811, -0.00004137, -0.00000671, -0.00000855, -0.00000510, -0.00001255, -0.00000140, 0.00000768, -0.00000872, -0.00000043, -0.00000622, 0.00000299, -0.00000393, -0.00001155, 0.00000272, -0.00000177, -0.00000518, -0.00000405, -0.00000017, -0.00000122, -0.00000584, -0.00000167, 0.00000110, -0.00000803, 0.00000457, 0.00000038, -0.00000305, -0.00000591, -0.00000233, -0.00000011, -0.00000353, -0.00000614, 0.00000499, 0.00000152, -0.00000031, -0.00000742, 0.00000936, 0.00000746, -0.00000396, 0.00000533, 0.00000050, -0.00000062, 0.00000149, 0.00000325, 0.00000069, -0.00000096, -0.00000188, },
		{ -0.00000193, -0.00000086, -0.00000190, 0.00000334, 0.00000134, 0.00000350, -0.00000208, 0.00000294, -0.00000117, 0.00000319, 0.00001383, -0.00000601, -0.00000146, -0.00000421, 0.00000968, -0.00000365, -0.00000519, -0.00000029, -0.00000206, -0.00000535, -0.00000227, -0.00000262, 0.00000468, -0.00000267, -0.00000354, -0.00000188, -0.00000397, -0.00000082, -0.00000035, -0.00000287, -0.00000634, -0.00000224, -0.00000048, 0.00000004, -0.00001368, 0.00000303, -0.00000260, -0.00000171, -0.00000562, -0.00000337, 0.00000853, -0.00001074, -0.00000877, -0.00000647, -0.00000566, -0.00002530, -0.00006259, 0.00003769, 0.00006087, 0.00002302, 0.00002488, 0.00002163, 0.00002184, 0.00003178, 0.00006987, -0.00001272, -0.00005913, -0.00001135, -0.00001436, -0.00000614, -0.00000965, -0.00000468, 0.00000169, -0.00000624, -0.00000491, -0.00000393, 0.00000261, -0.00001090, -0.00000547, -0.00001080, -0.00001182, -0.00002189, -0.00006865, 0.00002221, 0.00007288, 0.00002278, 0.00002686, 0.00002372, 0.00002644, 0.00003720, 0.00007765, -0.00002804, -0.00005827, -0.00000644, -0.00001010, -0.00000370, -0.00001285, -0.00000394, 0.00000822, -0.00000702, -0.00000234, -0.00000433, 0.00000057, -0.00000086, -0.00001310, 0.00000239, -0.00000159, -0.00000369, -0.00000602, 0.00000006, -0.00000135, -0.00000318, -0.00000464, 0.00000222, -0.00000792, 0.00000285, 0.00000163, -0.00000239, -0.00000530, -0.00000336, -0.00000084, -0.00000133, -0.00000789, 0.00000408, 0.00000319, -0.00000116, -0.00000584, 0.00000319, 0.00001280, -0.00000442, 0.00000424, 0.00000080, -0.00000025, 0.00000136, 0.00000306, 0.00000139, -0.00000130, -0.00000136, },
		},
		{ { -0.00000230, -0.00000175, 0.00000097, -0.00000026, -0.00000115, -0.00000237, -0.00000898, 0.00000042, -0.00001307, 0.00000326, -0.00000998, -0.00000240, -0.00000377, -0.00000373, 0.00001685, -0.00000408, 0.00000549, -0.00000143, -0.00000304, 0.00000174, 0.00000178, 0.00000165, 0.00000061, 0.00000033, -0.00000318, -0.00000180, 0.00000898, -0.00000235, -0.00000351, -0.00000455, -0.00000062, -0.00000437, -0.00000415, -0.00000391, -0.00000077, -0.00000615, -0.00001096, -0.00001431, -0.00006516, 0.00000439, 0.00007418, 0.00002752, 0.00002392, 0.00002309, 0.00002440, 0.00002580, 0.00007610, 0.00000433, -0.00006186, -0.00001484, -0.00001506, -0.00000715, -0.00000508, -0.00000407, -0.00000305, -0.00000174, 0.00000387, -0.00000559, -0.00000092, -0.00000294, -0.00000037, 0.00000070, -0.00001714, 0.00000196, 0.00000245, 0.00000196, -0.00001714, 0.00000070, -0.00000037, -0.00000294, -0.00000092, -0.00000559, 0.00000387, -0.00000174, -0.00000305, -0.00000407, -0.00000508, -0.00000715, -0.00001506, -0.00001484, -0.00006186, 0.00000433, 0.00007610, 0.00002580, 0.00002440, 0.00002309, 0.00002392, 0.00002752, 0.00007418, 0.00000439, -0.00006516, -0.00001431, -0.00001096, -0.00000615, -0.00000077, -0.00000391, -0.00000415, -0.00000437, -0.00000062, -0.00000455, -0.00000351, -0.00000235, 0.00000898, -0.00000180, -0.00000318, 0.00000033, 0.00000061, 0.00000165, 0.00000178, 0.00000174, -0.00000304, -0.00000143, 0.00000549, -0.00000408, 0.00001685, -0.00000373, -0.00000377, -0.00000240, -0.00000998, 0.00000326, -0.00001307, 0.00000042, -0.00000898, -0.00000237, -0.00000115, -0.00000026, 0.00000097, -0.00000175, },
		{ -0.00000220, -0.00000214, -0.00000046, 0.00000121, -0.00000212, 0.00000002, -0.00000968, -0.00000028, -0.00000845, -0.00000447, -0.00000341, -0.00000776, 0.00000162, -0.00001092, 0.00001531, 0.00000243, 0.00000109, 0.00000264, -0.00000442, 0.00000075, 0.00000204, 0.00000162, 0.00000059, 0.00000055, -0.00000110, -0.00000475, 0.00000813, 0.00000111, -0.00000424, -0.00000390, -0.00000179, -0.00000251, -0.00000552, -0.00000354, -0.00000328, -0.00000094, -0.00001334, -0.00000629, -0.00005008, -0.00004272, 0.00006986, 0.00004696, 0.00002269, 0.00002414, 0.00002692, 0.00002087, 0.00006787, 0.00004021, -0.00006491, -0.00002523, -0.00001362, -0.00001055, -0.00000452, -0.00000505, -0.00000288, -0.00000327, 0.00000283, -0.00000137, -0.00000492, 0.00000079, -0.00000516, 0.00000545, -0.00001527, -0.00000463, 0.00000337, 0.00000385, -0.00001165, -0.00000866, 0.00000520, -0.00000589, 0.00000148, -0.00000675, 0.00000161, 0.00000109, -0.00000333, -0.00000319, -0.00000537, -0.00000499, -0.00001456, -0.00001057, -0.00004971, -0.00002668, 0.00007033, 0.00003791, 0.00002014, 0.00002302, 0.00002411, 0.00001923, 0.00005750, 0.00004644, -0.00005462, -0.00003470, -0.00000575, -0.00001168, 0.00000011, -0.00000370, -0.00000276, -0.00000565, -0.00000022, -0.00000447, -0.00000336, -0.00000426, 0.00000727, 0.00000224, -0.00000456, -0.00000028, 0.00000019, 0.00000203, 0.00000152, 0.00000196, -0.00000072, -0.00000471, 0.00000677, -0.00000399, 0.00000960, 0.00000830, -0.00001011, 0.00000359, -0.00001561, 0.00000615, -0.00001035, -0.00000470, -0.00000480, -0.00000650, 0.00000056, -0.00000201, 0.00000180, -0.00000092, },
		{ -0.00000184, -0.00000232, -0.00000160, 0.00000127, -0.00000141, -0.00000007, -0.00000694, -0.00000464, -0.00000192, -0.00000982, -0.00000160, -0.00000899, 0.00000298, -0.00001128, 0.00000702, 0.00001027, -0.00000126, 0.00000387, -0.00000299, -0.00000154, 0.00000250, 0.00000150, 0.00000086, -0.00000015, 0.00000129, -0.00000606, 0.00000530, 0.00000466, -0.00000416, -0.00000373, -0.00000245, -0.00000149, -0.00000553, -0.00000411, -0.00000449, -0.00000011, -0.00000904, -0.00001001, -0.00002494, -0.00007060, 0.00003538, 0.00007185, 0.00002460, 0.00002601, 0.00002659, 0.00002319, 0.00005114, 0.00007096, -0.00005279, -0.00004190, -0.00001117, -0.00001452, -0.00000421, -0.00000583, -0.00000309, -0.00000364, 0.00000023, 0.00000238, -0.00000639, 0.00000107, -0.00000510, 0.00000330, -0.00000731, -0.00001264, 0.00000401, 0.00000290, -0.00000350, -0.00001571, 0.00000549, -0.00000400, -0.00000030, -0.00000392, -0.00000297, 0.00000360, -0.00000279, -0.00000272, -0.00000525, -0.00000395, -0.00001263, -0.00001006, -0.00003521, -0.00004625, 0.00005103, 0.00005245, 0.00001772, 0.00002248, 0.00002287, 0.00001901, 0.00003580, 0.00006643, -0.00002057, -0.00005406, -0.00000792, -0.00001163, -0.00000358, -0.00000152, -0.00000281, -0.00000493, -0.00000170, -0.00000274, -0.00000461, -0.00000410, 0.00000337, 0.00000650, -0.00000512, -0.00000060, -0.00000115, 0.00000287, 0.00000102, 0.00000227, 0.00000063, -0.00000479, 0.00000328, 0.00000181, -0.00000117, 0.00001748, -0.00001074, 0.00000426, -0.00001411, -0.00000006, -0.00000033, -0.00001243, -0.00000078, -0.00000942, 0.00000081, -0.00000250, 0.00000128, 0.00000032, },
		{ -0.00000112, -0.00000236, -0.00000215, 0.00000011, 0.00000014, -0.00000116, -0.00000341, -0.00000838, 0.00000117, -0.00000913, -0.00000466, -0.00000784, 0.00000196, -0.00000777, -0.00000219, 0.00001362, 0.00000125, 0.00000207, -0.00000009, -0.00000335, 0.00000191, 0.00000196, 0.00000099, -0.00000083, 0.00000249, -0.00000485, 0.00000089, 0.00000758, -0.00000317, -0.00000371, -0.00000308, -0.00000078, -0.00000498, -0.00000477, -0.00000490, -0.00000197, -0.00000294, -0.00001437, -0.00000940, -0.00006617, -0.00001938, 0.00008502, 0.00003685, 0.00002669, 0.00002560, 0.00002829, 0.00003488, 0.00008724, -0.00002358, -0.00006132, -0.00001045, -0.00001772, -0.00000482, -0.00000633, -0.00000337, -0.00000388, -0.00000167, 0.00000306, -0.00000384, -0.00000221, -0.00000158, -0.00000189, 0.00000052, -0.00001630, 0.00000093, 0.00000265, 0.00000173, -0.00001548, -0.00000116, 0.00000173, -0.00000413, -0.00000010, -0.00000645, 0.00000348, -0.00000077, -0.00000261, -0.00000491, -0.00000351, -0.00001024, -0.00001085, -0.00002306, -0.00005293, 0.00002372, 0.00006271, 0.00002034, 0.00002053, 0.00002114, 0.00002148, 0.00002035, 0.00006310, 0.00002009, -0.00005618, -0.00002236, -0.00000659, -0.00000950, 0.00000041, -0.00000309, -0.00000330, -0.00000377, -0.00000072, -0.00000587, -0.00000347, -0.00000059, 0.00000900, -0.00000392, -0.00000123, -0.00000227, 0.00000279, 0.00000134, 0.00000208, 0.00000141, -0.00000297, -0.00000163, 0.00000712, -0.00000697, 0.00001661, -0.00000214, -0.00000190, -0.00000555, -0.00001168, 0.00000982, -0.00001559, -0.00000167, -0.00000853, -0.00000166, -0.00000124, -0.00000040, 0.00000149, },
		{ -0.00000012, -0.00000208, -0.00000237, -0.00000128, 0.00000100, -0.00000145, -0.00000139, -0.00000897, -0.00000045, -0.00000431, -0.00000823, -0.00000801, 0.00000080, -0.00000340, -0.00000850, 0.00001137, 0.00000608, 0.00000102, 0.00000082, -0.00000228, -0.00000082, 0.00000349, 0.00000040, -0.00000045, 0.00000145, -0.00000114, -0.00000409, 0.00000930, -0.00000170, -0.00000302, -0.00000442, 0.00000038, -0.00000490, -0.00000436, -0.00000595, -0.00000322, -0.00000076, -0.00001149, -0.00000985, -0.00003941, -0.00006679, 0.00006718, 0.00006191, 0.00002641, 0.00002638, 0.00003130, 0.00002624, 0.00008539, 0.00001786, -0.00007595, -0.00001534, -0.00001856, -0.00000701, -0.00000648, -0.00000360, -0.00000447, -0.00000223, 0.00000110, 0.00000039, -0.00000544, 0.00000095, -0.00000469, 0.00000310, -0.00001321, -0.00000575, 0.00000354, 0.00000287, -0.00000937, -0.00000991, 0.00000561, -0.00000515, 0.00000068, -0.00000602, 0.00000007, 0.00000201, -0.00000224, -0.00000456, -0.00000345, -0.00000782, -0.00001164, -0.00001483, -0.00004968, -0.00000384, 0.00006335, 0.00002865, 0.00001765, 0.00001969, 0.00002279, 0.00001455, 0.00004611, 0.00004966, -0.00003622, -0.00004132, -0.00000506, -0.00001139, -0.00000185, -0.00000129, -0.00000292, -0.00000416, -0.00000059, -0.00000513, -0.00000433, -0.00000250, 0.00000816, -0.00000024, -0.00000302, -0.00000171, 0.00000048, 0.00000334, 0.00000063, 0.00000278, -0.00000172, -0.00000394, 0.00000705, -0.00000410, 0.00000640, 0.00001079, -0.00000873, 0.00000280, -0.00001938, 0.00001113, -0.00000916, -0.00000841, -0.00000466, -0.00000559, 0.00000048, -0.00000203, 0.00000191, },
		{ 0.00000090, -0.00000133, -0.00000254, -0.00000212, 0.00000048, -0.00000054, -0.00000110, -0.00000697, -0.00000429, 0.00000011, -0.00000833, -0.00001051, -0.00000101, 0.00000133, -0.00001170, 0.00000644, 0.00000869, 0.00000370, -0.00000122, 0.00000092, -0.00000398, 0.00000442, 0.00000021, 0.00000024, -0.00000060, 0.00000270, -0.00000708, 0.00000834, 0.00000084, -0.00000238, -0.00000535, 0.00000071, -0.00000415, -0.00000409, -0.00000659, -0.00000398, -0.00000206, -0.00000435, -0.00001624, -0.00001473, -0.00008087, 0.00001508, 0.00008819, 0.00003158, 0.00002804, 0.00003125, 0.00002616, 0.00006992, 0.00006022, -0.00007632, -0.00002909, -0.00001615, -0.00001127, -0.00000593, -0.00000434, -0.00000472, -0.00000263, -0.00000101, 0.00000275, -0.00000530, -0.00000013, -0.00000342, 0.00000077, -0.00000664, -0.00001171, 0.00000279, 0.00000267, -0.00000286, -0.00001411, 0.00000304, -0.00000135, -0.00000208, -0.00000261, -0.00000432, 0.00000342, -0.00000083, -0.00000414, -0.00000390, -0.00000532, -0.00001220, -0.00000997, -0.00004128, -0.00002518, 0.00005278, 0.00004026, 0.00001548, 0.00001884, 0.00002139, 0.00001623, 0.00002693, 0.00006023, -0.00000352, -0.00005035, -0.00001440, -0.00000702, -0.00000799, 0.00000158, -0.00000346, -0.00000295, -0.00000213, -0.00000278, -0.00000626, -0.00000273, 0.00000491, 0.00000452, -0.00000505, -0.00000006, -0.00000293, 0.00000552, -0.00000077, 0.00000396, -0.00000108, -0.00000348, 0.00000279, 0.00000301, -0.00000449, 0.00001788, -0.00000779, 0.00000354, -0.00001681, 0.00000126, 0.00000393, -0.00001572, -0.00000227, -0.00000786, 0.00000060, -0.00000238, 0.00000121, },
		{ 0.00000157, -0.00000028, -0.00000247, -0.00000245, -0.00000083, 0.00000051, -0.00000122, -0.00000451, -0.00000716, 0.00000113, -0.00000475, -0.00001277, -0.00000545, 0.00000634, -0.00001222, 0.00000126, 0.00000778, 0.00000822, -0.00000249, 0.00000208, -0.00000378, 0.00000208, 0.00000227, -0.00000048, -0.00000125, 0.00000366, -0.00000567, 0.00000360, 0.00000515, -0.00000311, -0.00000401, -0.00000166, -0.00000098, -0.00000583, -0.00000487, -0.00000634, -0.00000272, -0.00000068, -0.00001577, -0.00000871, -0.00005925, -0.00004851, 0.00009173, 0.00005076, 0.00002811, 0.00003076, 0.00003024, 0.00005072, 0.00009058, -0.00005576, -0.00005138, -0.00001168, -0.00001708, -0.00000466, -0.00000618, -0.00000374, -0.00000419, -0.00000134, 0.00000171, -0.00000183, -0.00000337, -0.00000069, -0.00000227, -0.00000185, -0.00001302, -0.00000126, 0.00000312, 0.00000054, -0.00001179, -0.00000432, 0.00000358, -0.00000448, -0.00000020, -0.00000611, 0.00000168, 0.00000160, -0.00000324, -0.00000491, -0.00000292, -0.00001225, -0.00000767, -0.00003147, -0.00003756, 0.00003386, 0.00005049, 0.00001617, 0.00001816, 0.00001824, 0.00002048, 0.00001379, 0.00005384, 0.00002754, -0.00004135, -0.00003123, -0.00000291, -0.00001187, 0.00000093, -0.00000200, -0.00000268, -0.00000281, -0.00000139, -0.00000679, -0.00000359, 0.00000195, 0.00000753, -0.00000499, 0.00000031, -0.00000450, 0.00000498, 0.00000032, 0.00000294, 0.00000080, -0.00000301, -0.00000084, 0.00000704, -0.00000756, 0.00001332, 0.00000249, -0.00000267, -0.00000708, -0.00001245, 0.00001387, -0.00001584, -0.00000543, -0.00000670, -0.00000157, -0.00000138, -0.00000020, },
		{ 0.00000159, 0.00000070, -0.00000183, -0.00000269, -0.00000194, 0.00000059, -0.00000071, -0.00000295, -0.00000773, -0.00000094, -0.00000044, -0.00001168, -0.00001248, 0.00000929, -0.00000891, -0.00000435, 0.00000594, 0.00000970, 0.00000134, -0.00000159, 0.00000087, -0.00000323, 0.00000581, -0.00000223, -0.00000019, 0.00000144, -0.00000075, -0.00000300, 0.00000958, -0.00000439, -0.00000106, -0.00000569, 0.00000319, -0.00000826, -0.00000207, -0.00000936, -0.00000196, -0.00000241, -0.00000713, -0.00001675, -0.00002546, -0.00008782, 0.00005455, 0.00008190, 0.00002930, 0.00003109, 0.00003416, 0.00003635, 0.00010133, -0.00001544, -0.00007576, -0.00000960, -0.00002208, -0.00000399, -0.00000844, -0.00000206, -0.00000650, -0.00000042, -0.00000110, 0.00000195, -0.00000528, 0.00000013, -0.00000286, -0.00000088, -0.00001021, -0.00000635, 0.00000290, 0.00000178, -0.00000662, -0.00001025, 0.00000393, -0.00000263, -0.00000139, -0.00000412, -0.00000241, 0.00000352, -0.00000129, -0.00000604, -0.00000142, -0.00001101, -0.00000768, -0.00002201, -0.00004200, 0.00001229, 0.00005435, 0.00002137, 0.00001659, 0.00001591, 0.00002220, 0.00001029, 0.00003743, 0.00004742, -0.00001853, -0.00004316, -0.00000818, -0.00000847, -0.00000540, 0.00000220, -0.00000419, -0.00000151, -0.00000212, -0.00000493, -0.00000600, 0.00000057, 0.00000750, -0.00000197, -0.00000167, -0.00000288, 0.00000092, 0.00000407, -0.00000024, 0.00000409, -0.00000370, -0.00000151, 0.00000545, -0.00000244, 0.00000160, 0.00001404, -0.00000717, 0.00000029, -0.00001931, 0.00001234, -0.00000595, -0.00001289, -0.00000420, -0.00000426, -0.00000038, -0.00000137, },
		{ 0.00000100, 0.00000126, -0.00000062, -0.00000285, -0.00000251, -0.00000036, 0.00000020, -0.00000210, -0.00000676, -0.00000384, 0.00000164, -0.00000683, -0.00001872, 0.00000660, -0.00000090, -0.00001059, 0.00000532, 0.00000619, 0.00000936, -0.00000689, 0.00000556, -0.00000678, 0.00000678, -0.00000192, 0.00000005, -0.00000089, 0.00000346, -0.00000691, 0.00001038, -0.00000309, 0.00000010, -0.00000734, 0.00000405, -0.00000720, -0.00000233, -0.00000910, -0.00000330, -0.00000439, 0.00000043, -0.00002189, -0.00000726, -0.00008276, -0.00001430, 0.00010423, 0.00004153, 0.00002995, 0.00003729, 0.00002959, 0.00009382, 0.00003452, -0.00009071, -0.00001628, -0.00002279, -0.00000640, -0.00000911, -0.00000175, -0.00000746, -0.00000059, -0.00000264, 0.00000290, -0.00000384, -0.00000166, -0.00000149, -0.00000166, -0.00000704, -0.00000881, 0.00000008, 0.00000296, -0.00000291, -0.00001053, -0.00000152, 0.00000219, -0.00000405, -0.00000124, -0.00000546, 0.00000277, 0.00000187, -0.00000640, -0.00000167, -0.00000801, -0.00000958, -0.00001370, -0.00004098, -0.00000665, 0.00004933, 0.00003058, 0.00001400, 0.00001611, 0.00001922, 0.00001475, 0.00001935, 0.00005337, 0.00000752, -0.00004041, -0.00002340, -0.00000160, -0.00001239, 0.00000465, -0.00000440, -0.00000066, -0.00000282, -0.00000284, -0.00000789, -0.00000079, 0.00000599, 0.00000210, -0.00000421, -0.00000008, -0.00000379, 0.00000718, -0.00000241, 0.00000598, -0.00000338, -0.00000138, 0.00000189, 0.00000399, -0.00000701, 0.00001665, -0.00000256, -0.00000054, -0.00001549, 0.00000054, 0.00000770, -0.00001774, -0.00000489, -0.00000494, -0.00000080, -0.00000163, },
		{ 0.00000025, 0.00000129, 0.00000065, -0.00000250, -0.00000286, -0.00000151, 0.00000054, -0.00000115, -0.00000550, -0.00000573, 0.00000081, -0.00000089, -0.00002003, -0.00000282, 0.00000917, -0.00001477, 0.00000441, 0.00000121, 0.00001515, -0.00000688, 0.00000422, -0.00000391, 0.00000220, 0.00000231, -0.00000232, -0.00000072, 0.00000363, -0.00000527, 0.00000576, 0.00000221, -0.00000247, -0.00000428, -0.00000046, -0.00000113, -0.00000698, -0.00000438, -0.00000823, -0.00000283, 0.00000019, -0.00001487, -0.00001191, -0.00004653, -0.00007829, 0.00008959, 0.00007144, 0.00002712, 0.00003999, 0.00002922, 0.00007551, 0.00007993, -0.00008522, -0.00003551, -0.00001772, -0.00001318, -0.00000684, -0.00000442, -0.00000550, -0.00000341, -0.00000142, 0.00000047, -0.00000013, -0.00000396, -0.00000061, -0.00000149, -0.00000593, -0.00000801, -0.00000415, 0.00000353, -0.00000107, -0.00000694, -0.00000781, 0.00000433, -0.00000355, -0.00000129, -0.00000474, -0.00000115, 0.00000503, -0.00000499, -0.00000370, -0.00000402, -0.00001190, -0.00000770, -0.00003625, -0.00002053, 0.00003690, 0.00004034, 0.00001243, 0.00001757, 0.00001409, 0.00002087, 0.00000767, 0.00004643, 0.00002887, -0.00002436, -0.00003795, -0.00000168, -0.00001223, 0.00000022, -0.00000023, -0.00000274, -0.00000114, -0.00000289, -0.00000708, -0.00000385, 0.00000499, 0.00000475, -0.00000476, 0.00000087, -0.00000550, 0.00000623, -0.00000072, 0.00000407, -0.00000000, -0.00000304, 0.00000056, 0.00000575, -0.00000665, 0.00000833, 0.00000883, -0.00000541, -0.00000738, -0.00001159, 0.00001457, -0.00001344, -0.00001077, -0.00000352, -0.00000226, -0.00000142, },
		{ -0.00000012, 0.00000095, 0.00000141, -0.00000135, -0.00000322, -0.00000227, -0.00000009, -0.00000003, -0.00000428, -0.00000638, -0.00000146, 0.00000283, -0.00001499, -0.00001534, 0.00001525, -0.00001266, 0.00000014, -0.00000031, 0.00001314, 0.00000163, -0.00000365, 0.00000381, -0.00000506, 0.00000749, -0.00000503, 0.00000065, 0.00000098, -0.00000032, -0.00000112, 0.00000838, -0.00000601, 0.00000054, -0.00000660, 0.00000542, -0.00001127, -0.00000000, -0.00001230, -0.00000094, -0.00000424, -0.00000256, -0.00002367, -0.00001300, -0.00010044, 0.00002904, 0.00010682, 0.00003144, 0.00003930, 0.00003436, 0.00005419, 0.00010972, -0.00005535, -0.00006395, -0.00001045, -0.00002192, -0.00000332, -0.00000888, -0.00000181, -0.00000758, 0.00000099, -0.00000291, 0.00000283, -0.00000431, -0.00000153, -0.00000020, -0.00000564, -0.00000696, -0.00000620, 0.00000127, 0.00000104, -0.00000402, -0.00000926, 0.00000016, 0.00000127, -0.00000421, -0.00000153, -0.00000550, 0.00000574, -0.00000121, -0.00000649, -0.00000079, -0.00001259, -0.00000528, -0.00002870, -0.00002952, 0.00002127, 0.00004584, 0.00001485, 0.00001718, 0.00001162, 0.00002195, 0.00000689, 0.00003038, 0.00004247, -0.00000402, -0.00003989, -0.00001412, -0.00000424, -0.00000923, 0.00000567, -0.00000568, 0.00000138, -0.00000395, -0.00000471, -0.00000737, 0.00000393, 0.00000587, -0.00000307, -0.00000058, -0.00000371, 0.00000160, 0.00000368, -0.00000010, 0.00000474, -0.00000535, 0.00000111, 0.00000344, -0.00000062, -0.00000301, 0.00001682, -0.00000506, -0.00000404, -0.00001515, 0.00000966, -0.00000111, -0.00001766, -0.00000351, -0.00000268, -0.00000189, },
		{ -0.00000030, 0.00000062, 0.00000141, 0.00000022, -0.00000321, -0.00000281, -0.00000107, 0.00000050, -0.00000271, -0.00000649, -0.00000337, 0.00000294, -0.00000621, -0.00002451, 0.00001216, -0.00000314, -0.00000732, 0.00000233, 0.00000454, 0.00001362, -0.00001094, 0.00000883, -0.00000803, 0.00000796, -0.00000382, -0.00000029, -0.00000053, 0.00000297, -0.00000484, 0.00001036, -0.00000572, 0.00000183, -0.00000830, 0.00000613, -0.00000920, -0.00000168, -0.00001019, -0.00000390, -0.00000533, 0.00000302, -0.00002421, -0.00000532, -0.00007495, -0.00005074, 0.00011646, 0.00005590, 0.00003202, 0.00004413, 0.00003584, 0.00011920, -0.00000631, -0.00009108, -0.00000874, -0.00002730, -0.00000270, -0.00001162, 0.00000016, -0.00000978, 0.00000140, -0.00000408, 0.00000266, -0.00000211, -0.00000326, 0.00000029, -0.00000371, -0.00000823, -0.00000465, -0.00000354, 0.00000351, -0.00000335, -0.00000567, -0.00000715, 0.00000556, -0.00000544, -0.00000019, -0.00000639, 0.00000221, 0.00000420, -0.00000826, 0.00000012, -0.00001058, -0.00000637, -0.00001971, -0.00003431, 0.00000629, 0.00004426, 0.00002211, 0.00001366, 0.00001388, 0.00001649, 0.00001438, 0.00001278, 0.00004657, 0.00001362, -0.00002747, -0.00003153, 0.00000192, -0.00001472, 0.00000612, -0.00000433, 0.00000070, -0.00000265, -0.00000395, -0.00000842, 0.00000081, 0.00000699, -0.00000087, -0.00000244, -0.00000108, -0.00000297, 0.00000664, -0.00000241, 0.00000687, -0.00000486, 0.00000021, 0.00000161, 0.00000386, -0.00000809, 0.00001402, 0.00000400, -0.00000697, -0.00001091, -0.00000148, 0.00001016, -0.00001800, -0.00000850, -0.00000094, -0.00000322, },
		{ -0.00000129, 0.00000079, 0.00000085, 0.00000146, -0.00000231, -0.00000349, -0.00000172, -0.00000004, -0.00000079, -0.00000627, -0.00000440, 0.00000064, 0.00000170, -0.00002551, -0.00000001, 0.00000918, -0.00001329, 0.00000494, -0.00000383, 0.00001999, -0.00000946, 0.00000519, -0.00000303, 0.00000202, 0.00000188, -0.00000405, 0.00000046, 0.00000246, -0.00000350, 0.00000685, -0.00000053, -0.00000171, -0.00000423, 0.00000032, -0.00000116, -0.00000833, -0.00000344, -0.00001041, -0.00000234, -0.00000031, -0.00001206, -0.00001760, -0.00003150, -0.00010324, 0.00007453, 0.00009936, 0.00002447, 0.00005284, 0.00002664, 0.00010854, 0.00005093, -0.00010425, -0.00001952, -0.00002542, -0.00000791, -0.00001012, -0.00000199, -0.00000800, -0.00000174, -0.00000209, -0.00000033, 0.00000099, -0.00000389, -0.00000121, -0.00000008, -0.00001056, -0.00000253, -0.00000692, 0.00000340, -0.00000243, -0.00000194, -0.00001062, 0.00000342, -0.00000130, -0.00000304, -0.00000298, -0.00000402, 0.00000848, -0.00000710, -0.00000177, -0.00000655, -0.00000918, -0.00001159, -0.00003493, -0.00000635, 0.00003632, 0.00003144, 0.00000957, 0.00001791, 0.00000951, 0.00002160, 0.00000290, 0.00003951, 0.00002742, -0.00000930, -0.00003990, -0.00000411, -0.00001004, -0.00000193, 0.00000234, -0.00000366, 0.00000132, -0.00000531, -0.00000660, -0.00000412, 0.00000803, 0.00000107, -0.00000329, 0.00000013, -0.00000466, 0.00000566, -0.00000092, 0.00000488, -0.00000069, -0.00000303, 0.00000218, 0.00000385, -0.00000524, 0.00000337, 0.00001492, -0.00000868, -0.00000732, -0.00000857, 0.00001166, -0.00000866, -0.00001702, 0.00000054, -0.00000362, },
		{ -0.00000373, 0.00000140, 0.00000020, 0.00000188, -0.00000063, -0.00000398, -0.00000219, -0.00000107, 0.00000050, -0.00000511, -0.00000523, -0.00000175, 0.00000525, -0.00001824, -0.00001549, 0.00001671, -0.00001260, 0.00000337, -0.00000633, 0.00001599, 0.00000164, -0.00000460, 0.00000545, -0.00000498, 0.00000702, -0.00000659, 0.00000106, 0.00000078, -0.00000032, 0.00000179, 0.00000536, -0.00000579, 0.00000073, -0.00000595, 0.00000572, -0.00001245, 0.00000033, -0.00001323, -0.00000172, -0.00000446, -0.00000028, -0.00002716, -0.00000650, -0.00010068, -0.00000977, 0.00013402, 0.00003474, 0.00005060, 0.00003221, 0.00008164, 0.00010372, -0.00009429, -0.00004387, -0.00001761, -0.00001748, -0.00000578, -0.00000702, -0.00000373, -0.00000664, 0.00000095, -0.00000354, 0.00000261, -0.00000252, -0.00000368, 0.00000259, -0.00001020, -0.00000403, -0.00000538, -0.00000109, 0.00000068, -0.00000179, -0.00000709, -0.00000497, 0.00000548, -0.00000672, 0.00000076, -0.00000828, 0.00000810, -0.00000205, -0.00000545, -0.00000231, -0.00001135, -0.00000641, -0.00003149, -0.00001639, 0.00002501, 0.00003830, 0.00000906, 0.00001877, 0.00000746, 0.00002120, 0.00000512, 0.00002357, 0.00003725, 0.00000589, -0.00003239, -0.00002103, -0.00000047, -0.00001182, 0.00000789, -0.00000638, 0.00000356, -0.00000523, -0.00000519, -0.00000755, 0.00000658, 0.00000396, -0.00000372, 0.00000019, -0.00000400, 0.00000232, 0.00000212, 0.00000151, 0.00000392, -0.00000553, 0.00000257, 0.00000239, -0.00000019, -0.00000563, 0.00001777, -0.00000200, -0.00000987, -0.00000760, 0.00000397, 0.00000431, -0.00002181, -0.00000290, -0.00000103, },
		{ -0.00000612, 0.00000132, 0.00000016, 0.00000144, 0.00000104, -0.00000351, -0.00000308, -0.00000157, 0.00000022, -0.00000271, -0.00000628, -0.00000291, 0.00000414, -0.00000698, -0.00002663, 0.00001429, -0.00000462, -0.00000194, -0.00000320, 0.00000503, 0.00001476, -0.00001149, 0.00000893, -0.00000606, 0.00000641, -0.00000427, -0.00000149, 0.00000093, 0.00000089, -0.00000056, 0.00000759, -0.00000587, 0.00000148, -0.00000694, 0.00000561, -0.00000886, -0.00000331, -0.00000873, -0.00000693, -0.00000353, 0.00000167, -0.00002132, -0.00001031, -0.00005821, -0.00009011, 0.00012045, 0.00007590, 0.00003515, 0.00005041, 0.00004882, 0.00013888, -0.00005757, -0.00007634, -0.00000991, -0.00002654, -0.00000273, -0.00001131, -0.00000053, -0.00000992, 0.00000196, -0.00000446, 0.00000163, -0.00000003, -0.00000489, 0.00000184, -0.00000550, -0.00000942, -0.00000055, -0.00000695, 0.00000361, -0.00000350, -0.00000067, -0.00001240, 0.00000748, -0.00000540, 0.00000002, -0.00000690, 0.00000209, 0.00000529, -0.00000861, 0.00000020, -0.00001145, -0.00000465, -0.00002522, -0.00002355, 0.00001310, 0.00003965, 0.00001400, 0.00001466, 0.00001176, 0.00001357, 0.00001460, 0.00000738, 0.00003968, 0.00001688, -0.00001521, -0.00003558, 0.00000139, -0.00001348, 0.00000493, -0.00000242, 0.00000037, -0.00000101, -0.00000664, -0.00000721, 0.00000150, 0.00000806, -0.00000414, 0.00000004, -0.00000293, -0.00000044, 0.00000345, 0.00000005, 0.00000560, -0.00000440, 0.00000027, 0.00000275, 0.00000190, -0.00000720, 0.00001030, 0.00001060, -0.00001412, -0.00000464, -0.00000358, 0.00001049, -0.00001630, -0.00001267, 0.00000344, },
		{ -0.00000520, -0.00000104, 0.00000125, 0.00000045, 0.00000200, -0.00000185, -0.00000431, -0.00000147, -0.00000122, -0.00000013, -0.00000673, -0.00000340, 0.00000086, 0.00000251, -0.00002845, 0.00000259, 0.00000599, -0.00000680, 0.00000135, -0.00000556, 0.00002103, -0.00000885, 0.00000400, -0.00000047, 0.00000113, 0.00000099, -0.00000532, 0.00000187, 0.00000048, -0.00000045, 0.00000653, -0.00000242, -0.00000142, -0.00000353, 0.00000048, -0.00000106, -0.00000971, -0.00000235, -0.00001255, -0.00000151, -0.00000192, -0.00000816, -0.00002384, -0.00001799, -0.00011856, 0.00004462, 0.00013220, 0.00002352, 0.00006635, 0.00002714, 0.00014455, 0.00000358, -0.00010716, -0.00000941, -0.00003064, -0.00000413, -0.00001238, -0.00000059, -0.00000989, -0.00000005, -0.00000294, -0.00000085, 0.00000167, -0.00000348, -0.00000195, 0.00000098, -0.00001423, 0.00000190, -0.00000864, 0.00000224, -0.00000293, 0.00000226, -0.00001195, 0.00000065, 0.00000191, -0.00000484, -0.00000160, -0.00000580, 0.00001056, -0.00000831, -0.00000023, -0.00000928, -0.00000543, -0.00001804, -0.00002747, 0.00000230, 0.00003516, 0.00002254, 0.00000849, 0.00001786, 0.00000585, 0.00002124, 0.00000070, 0.00003138, 0.00002643, 0.00000066, -0.00003574, -0.00001026, -0.00000562, -0.00000510, 0.00000536, -0.00000515, 0.00000424, -0.00000835, -0.00000550, -0.00000416, 0.00001047, -0.00000277, -0.00000103, -0.00000168, -0.00000194, 0.00000287, 0.00000036, 0.00000478, -0.00000086, -0.00000308, 0.00000376, 0.00000177, -0.00000404, -0.00000011, 0.00001887, -0.00001088, -0.00000779, -0.00000333, 0.00000568, -0.00000253, -0.00002294, 0.00000453, },
		{ 0.00000055, -0.00000551, 0.00000282, -0.00000044, 0.00000209, 0.00000022, -0.00000480, -0.00000182, -0.00000225, 0.00000080, -0.00000530, -0.00000447, -0.00000167, 0.00000651, -0.00002111, -0.00001273, 0.00001288, -0.00000766, 0.00000406, -0.00001097, 0.00001759, 0.00000151, -0.00000420, 0.00000532, -0.00000240, 0.00000358, -0.00000595, 0.00000027, 0.00000133, -0.00000096, 0.00000557, 0.00000072, -0.00000358, -0.00000098, -0.00000314, 0.00000331, -0.00001089, -0.00000197, -0.00001129, -0.00000512, -0.00000205, -0.00000184, -0.00002503, -0.00000850, -0.00008759, -0.00005859, 0.00015772, 0.00004334, 0.00006256, 0.00003058, 0.00011765, 0.00007798, -0.00012337, -0.00002178, -0.00002810, -0.00001026, -0.00001060, -0.00000332, -0.00000762, -0.00000354, -0.00000092, -0.00000271, 0.00000128, -0.00000022, -0.00000580, 0.00000470, -0.00001368, -0.00000183, -0.00000423, -0.00000360, 0.00000044, 0.00000023, -0.00000435, -0.00001025, 0.00000862, -0.00000766, 0.00000159, -0.00000951, 0.00000950, -0.00000309, -0.00000339, -0.00000559, -0.00000752, -0.00001138, -0.00002849, -0.00000631, 0.00002633, 0.00003093, 0.00000484, 0.00002008, 0.00000451, 0.00001901, 0.00000567, 0.00001591, 0.00003342, 0.00001100, -0.00002243, -0.00002679, 0.00000120, -0.00001198, 0.00000798, -0.00000559, 0.00000437, -0.00000548, -0.00000669, -0.00000609, 0.00000783, 0.00000227, -0.00000404, 0.00000052, -0.00000348, 0.00000254, -0.00000007, 0.00000431, 0.00000158, -0.00000389, 0.00000227, 0.00000293, -0.00000191, -0.00000515, 0.00001599, 0.00000208, -0.00001631, 0.00000177, -0.00000300, 0.00000876, -0.00002420, -0.00000280, },
		{ 0.00000724, -0.00000871, 0.00000273, -0.00000024, 0.00000141, 0.00000169, -0.00000363, -0.00000332, -0.00000189, -0.00000057, -0.00000214, -0.00000614, -0.00000250, 0.00000516, -0.00000918, -0.00002463, 0.00001220, -0.00000443, 0.00000466, -0.00001076, 0.00000782, 0.00001222, -0.00000788, 0.00000520, -0.00000027, 0.00000146, -0.00000265, -0.00000382, 0.00000347, -0.00000251, 0.00000552, 0.00000224, -0.00000323, -0.00000148, -0.00000273, 0.00000196, -0.00000568, -0.00000752, -0.00000456, -0.00001243, 0.00000118, -0.00000392, -0.00001313, -0.00002144, -0.00003574, -0.00012650, 0.00011135, 0.00010224, 0.00004064, 0.00005487, 0.00007267, 0.00014338, -0.00010995, -0.00005031, -0.00001969, -0.00001966, -0.00000750, -0.00000723, -0.00000478, -0.00000682, 0.00000013, -0.00000315, -0.00000047, 0.00000246, -0.00000655, 0.00000319, -0.00000732, -0.00000962, 0.00000204, -0.00000889, 0.00000247, -0.00000253, 0.00000319, -0.00001577, 0.00000682, -0.00000329, -0.00000105, -0.00000660, 0.00000227, 0.00000484, -0.00000698, -0.00000196, -0.00000973, -0.00000597, -0.00002739, -0.00001220, 0.00001543, 0.00003576, 0.00000643, 0.00001656, 0.00000969, 0.00001079, 0.00001486, 0.00000343, 0.00003229, 0.00001935, -0.00000653, -0.00003370, -0.00000391, -0.00000849, 0.00000111, 0.00000123, -0.00000156, 0.00000183, -0.00001055, -0.00000438, 0.00000127, 0.00000890, -0.00000709, 0.00000253, -0.00000488, 0.00000297, -0.00000168, 0.00000470, 0.00000213, -0.00000180, -0.00000143, 0.00000533, -0.00000190, -0.00000429, 0.00000597, 0.00001590, -0.00002004, 0.00000142, -0.00000447, 0.00000815, -0.00001277, -0.00001663, },
		{ 0.00000719, -0.00000627, -0.00000093, 0.00000159, 0.00000021, 0.00000223, -0.00000125, -0.00000507, -0.00000105, -0.00000270, 0.00000064, -0.00000674, -0.00000285, 0.00000157, 0.00000129, -0.00002823, 0.00000413, 0.00000049, 0.00000395, -0.00000619, -0.00000353, 0.00001772, -0.00000462, 0.00000037, 0.00000395, -0.00000075, -0.00000020, -0.00000591, 0.00000309, -0.00000184, 0.00000346, 0.00000475, -0.00000302, -0.00000199, -0.00000197, -0.00000009, -0.00000086, -0.00001083, -0.00000147, -0.00001421, -0.00000072, -0.00000414, -0.00000415, -0.00002874, -0.00000916, -0.00012134, 0.00000102, 0.00016459, 0.00002990, 0.00007568, 0.00003805, 0.00017362, -0.00005635, -0.00009273, -0.00000895, -0.00003016, -0.00000432, -0.00001171, -0.00000184, -0.00000973, 0.00000056, -0.00000331, -0.00000150, 0.00000227, -0.00000323, -0.00000215, 0.00000071, -0.00001554, 0.00000393, -0.00000857, -0.00000021, -0.00000218, 0.00000500, -0.00001141, -0.00000373, 0.00000542, -0.00000605, -0.00000123, -0.00000563, 0.00001016, -0.00000750, -0.00000031, -0.00001065, -0.00000274, -0.00002428, -0.00001622, 0.00000535, 0.00003483, 0.00001312, 0.00000962, 0.00001643, 0.00000414, 0.00001879, 0.00000172, 0.00002114, 0.00002782, 0.00000426, -0.00002600, -0.00001880, -0.00000025, -0.00000824, 0.00000795, -0.00000643, 0.00000678, -0.00001126, -0.00000420, -0.00000359, 0.00001160, -0.00000587, 0.00000123, -0.00000384, 0.00000191, -0.00000172, 0.00000316, 0.00000355, -0.00000030, -0.00000325, 0.00000510, -0.00000028, -0.00000329, -0.00000138, 0.00001948, -0.00001091, -0.00000921, 0.00000353, -0.00000204, 0.00000337, -0.00002692, },
		{ -0.00000368, 0.00000199, -0.00000688, 0.00000378, -0.00000080, 0.00000208, 0.00000095, -0.00000535, -0.00000148, -0.00000338, 0.00000080, -0.00000482, -0.00000417, -0.00000111, 0.00000626, -0.00002290, -0.00000793, 0.00000455, 0.00000181, 0.00000174, -0.00001331, 0.00001720, 0.00000173, -0.00000282, 0.00000379, 0.00000219, -0.00000249, -0.00000309, -0.00000180, 0.00000259, -0.00000192, 0.00000944, -0.00000461, -0.00000005, -0.00000416, 0.00000147, -0.00000156, -0.00000674, -0.00000670, -0.00000681, -0.00001067, 0.00000256, -0.00000700, -0.00001745, -0.00001845, -0.00006407, -0.00011032, 0.00016942, 0.00006304, 0.00007107, 0.00003518, 0.00015687, 0.00003115, -0.00013510, -0.00000485, -0.00003729, -0.00000328, -0.00001563, 0.00000056, -0.00001214, 0.00000079, -0.00000407, -0.00000082, -0.00000067, 0.00000199, -0.00000720, 0.00000536, -0.00001527, -0.00000065, -0.00000314, -0.00000563, -0.00000009, 0.00000233, -0.00000182, -0.00001440, 0.00000942, -0.00000611, 0.00000016, -0.00000818, 0.00000880, -0.00000325, -0.00000148, -0.00000921, -0.00000275, -0.00001852, -0.00002016, -0.00000115, 0.00002753, 0.00002276, 0.00000312, 0.00001988, 0.00000364, 0.00001489, 0.00000859, 0.00000719, 0.00003145, 0.00001222, -0.00001218, -0.00002958, -0.00000021, -0.00000922, 0.00000569, -0.00000314, 0.00000362, -0.00000469, -0.00000896, -0.00000313, 0.00000751, 0.00000106, -0.00000405, 0.00000035, -0.00000197, 0.00000169, -0.00000191, 0.00000737, -0.00000153, -0.00000094, 0.00000056, 0.00000461, -0.00000535, -0.00000169, 0.00001154, 0.00000660, -0.00002193, 0.00001076, -0.00000906, 0.00001056, -0.00002372, },
		{ -0.00001949, 0.00000942, -0.00001010, 0.00000339, -0.00000038, 0.00000136, 0.00000203, -0.00000350, -0.00000352, -0.00000216, -0.00000157, -0.00000125, -0.00000595, -0.00000211, 0.00000572, -0.00001225, -0.00001899, 0.00000595, -0.00000183, 0.00001086, -0.00001865, 0.00001164, 0.00000702, -0.00000105, -0.00000137, 0.00000832, -0.00000613, 0.00000062, -0.00000683, 0.00000638, -0.00000621, 0.00001209, -0.00000446, 0.00000132, -0.00000637, 0.00000361, -0.00000411, -0.00000045, -0.00001334, 0.00000113, -0.00001904, 0.00000733, -0.00001214, -0.00000176, -0.00003550, -0.00001291, -0.00015291, 0.00008474, 0.00013483, 0.00004972, 0.00005758, 0.00010963, 0.00012357, -0.00015145, -0.00002099, -0.00003419, -0.00000878, -0.00001567, -0.00000041, -0.00001183, -0.00000091, -0.00000393, -0.00000032, -0.00000336, 0.00000503, -0.00000800, 0.00000418, -0.00000924, -0.00000825, 0.00000213, -0.00000852, -0.00000036, 0.00000002, 0.00000500, -0.00001630, 0.00000309, 0.00000097, -0.00000352, -0.00000507, 0.00000228, 0.00000326, -0.00000381, -0.00000587, -0.00000579, -0.00001052, -0.00002455, -0.00000363, 0.00001595, 0.00003157, 0.00000044, 0.00001829, 0.00000804, 0.00000821, 0.00001482, 0.00000099, 0.00002426, 0.00002229, -0.00000269, -0.00002605, -0.00001288, -0.00000104, -0.00000414, 0.00000559, -0.00000411, 0.00000478, -0.00001451, -0.00000083, 0.00000069, 0.00000887, -0.00000881, 0.00000404, -0.00000580, 0.00000575, -0.00000710, 0.00001014, -0.00000246, 0.00000209, -0.00000411, 0.00000846, -0.00000668, 0.00000002, 0.00000175, 0.00001872, -0.00002304, 0.00000548, -0.00000318, 0.00000313, -0.00000796, },
		{ -0.00002743, 0.00000733, -0.00000609, -0.00000127, 0.00000186, 0.00000014, 0.00000222, -0.00000072, -0.00000543, -0.00000092, -0.00000396, 0.00000113, -0.00000615, -0.00000281, 0.00000264, -0.00000183, -0.00002452, 0.00000350, -0.00000528, 0.00001644, -0.00001573, 0.00000108, 0.00001097, 0.00000238, -0.00000492, 0.00000961, -0.00000340, -0.00000161, -0.00000569, 0.00000369, -0.00000381, 0.00000750, 0.00000195, -0.00000238, -0.00000345, 0.00000035, -0.00000128, -0.00000075, -0.00001133, -0.00000131, -0.00001489, -0.00000049, -0.00000629, -0.00000104, -0.00003083, -0.00000692, -0.00011118, -0.00005200, 0.00018892, 0.00004998, 0.00007604, 0.00006715, 0.00018291, -0.00011758, -0.00006529, -0.00001798, -0.00002331, -0.00000915, -0.00000755, -0.00000626, -0.00000686, -0.00000059, -0.00000269, -0.00000257, 0.00000307, -0.00000347, -0.00000142, -0.00000139, -0.00001374, 0.00000286, -0.00000649, -0.00000382, -0.00000021, 0.00000608, -0.00000911, -0.00000902, 0.00000843, -0.00000620, -0.00000198, -0.00000337, 0.00000696, -0.00000423, -0.00000263, -0.00000959, -0.00000298, -0.00002719, -0.00000497, 0.00000500, 0.00003482, 0.00000389, 0.00001238, 0.00001356, 0.00000458, 0.00001422, 0.00000548, 0.00000923, 0.00003158, 0.00000247, -0.00001305, -0.00002724, 0.00000414, -0.00000995, 0.00000901, -0.00000650, 0.00000787, -0.00001307, -0.00000333, -0.00000198, 0.00001079, -0.00000734, 0.00000253, -0.00000523, 0.00000558, -0.00000691, 0.00000678, 0.00000149, 0.00000083, -0.00000337, 0.00000599, -0.00000224, -0.00000281, -0.00000054, 0.00001657, -0.00000847, -0.00001135, 0.00001072, -0.00000963, 0.00000725, },
		{ -0.00001972, -0.00000582, 0.00000322, -0.00000757, 0.00000400, -0.00000078, 0.00000199, 0.00000128, -0.00000528, -0.00000154, -0.00000414, 0.00000039, -0.00000375, -0.00000427, -0.00000006, 0.00000425, -0.00002264, -0.00000286, -0.00000600, 0.00001390, -0.00000292, -0.00001308, 0.00001436, 0.00000286, -0.00000123, 0.00000191, 0.00000737, -0.00000974, 0.00000093, -0.00000443, 0.00000394, -0.00000242, 0.00001223, -0.00000915, 0.00000337, -0.00000751, 0.00000654, -0.00000754, -0.00000130, -0.00001246, -0.00000134, -0.00001663, 0.00000751, -0.00001367, -0.00000696, -0.00003274, -0.00003644, -0.00015518, 0.00015970, 0.00009944, 0.00007186, 0.00005349, 0.00018960, -0.00003250, -0.00012606, 0.00000130, -0.00004075, 0.00000015, -0.00001830, 0.00000243, -0.00001528, 0.00000467, -0.00000721, 0.00000113, -0.00000240, 0.00000332, -0.00000720, 0.00000396, -0.00001444, -0.00000072, -0.00000239, -0.00000675, -0.00000111, 0.00000465, -0.00000009, -0.00001650, 0.00000730, -0.00000199, -0.00000350, -0.00000407, 0.00000540, -0.00000167, -0.00000094, -0.00001137, 0.00000056, -0.00002482, -0.00000914, -0.00000045, 0.00002940, 0.00001360, 0.00000417, 0.00001755, 0.00000509, 0.00000911, 0.00001304, -0.00000184, 0.00003066, 0.00001125, -0.00000376, -0.00002831, -0.00000479, -0.00000394, 0.00000152, 0.00000059, 0.00000161, -0.00000328, -0.00001131, 0.00000066, 0.00000589, 0.00000028, -0.00000374, -0.00000025, 0.00000044, -0.00000054, -0.00000251, 0.00000949, -0.00000422, 0.00000219, -0.00000147, 0.00000615, -0.00000904, 0.00000343, 0.00000548, 0.00001047, -0.00002507, 0.00001710, -0.00001219, 0.00000846, },
		{ -0.00000276, -0.00002041, 0.00000919, -0.00000959, 0.00000284, -0.00000009, 0.00000130, 0.00000201, -0.00000306, -0.00000366, -0.00000255, -0.00000242, -0.00000043, -0.00000552, -0.00000174, 0.00000543, -0.00001518, -0.00001109, -0.00000341, 0.00000342, 0.00001484, -0.00002444, 0.00001427, 0.00000136, 0.00000665, -0.00000832, 0.00001719, -0.00001424, 0.00000405, -0.00000926, 0.00000832, -0.00000858, 0.00001712, -0.00001040, 0.00000586, -0.00001135, 0.00001012, -0.00001061, 0.00000487, -0.00001861, 0.00000595, -0.00002417, 0.00001208, -0.00001969, 0.00000882, -0.00004752, 0.00000373, -0.00016264, 0.00003777, 0.00017229, 0.00006325, 0.00006103, 0.00015772, 0.00007396, -0.00017075, 0.00000171, -0.00004692, 0.00000160, -0.00002377, 0.00000628, -0.00001916, 0.00000583, -0.00000876, 0.00000290, -0.00000609, 0.00000688, -0.00000860, 0.00000430, -0.00001085, -0.00000546, -0.00000032, -0.00000591, -0.00000458, 0.00000390, 0.00000467, -0.00001394, -0.00000323, 0.00000660, -0.00000689, -0.00000227, 0.00000157, 0.00000135, -0.00000020, -0.00001012, -0.00000117, -0.00001686, -0.00001696, 0.00000027, 0.00001706, 0.00002574, -0.00000271, 0.00001863, 0.00000739, 0.00000573, 0.00001435, -0.00000014, 0.00001597, 0.00002580, -0.00000309, -0.00001466, -0.00002294, 0.00000648, -0.00000886, 0.00000907, -0.00000586, 0.00000633, -0.00001697, 0.00000215, 0.00000061, 0.00000729, -0.00000853, 0.00000377, -0.00000470, 0.00000631, -0.00001074, 0.00001435, -0.00000637, 0.00000566, -0.00000624, 0.00001053, -0.00001086, 0.00000454, -0.00000154, 0.00001823, -0.00002197, 0.00000644, 0.00000053, -0.00000394, },
		{ 0.00000766, -0.00002354, 0.00000425, -0.00000418, -0.00000232, 0.00000212, 0.00000019, 0.00000210, -0.00000049, -0.00000507, -0.00000151, -0.00000445, 0.00000097, -0.00000483, -0.00000321, 0.00000377, -0.00000630, -0.00001758, -0.00000015, -0.00000886, 0.00002713, -0.00002399, 0.00000573, 0.00000305, 0.00001005, -0.00001004, 0.00001505, -0.00000606, -0.00000339, -0.00000453, 0.00000322, -0.00000467, 0.00001068, -0.00000100, -0.00000109, -0.00000527, 0.00000287, -0.00000261, -0.00000115, -0.00001071, -0.00000243, -0.00001395, -0.00000152, -0.00000754, 0.00000022, -0.00002923, -0.00001167, -0.00009009, -0.00010793, 0.00019712, 0.00008855, 0.00006561, 0.00011925, 0.00016024, -0.00016589, -0.00003383, -0.00003220, -0.00001254, -0.00001688, -0.00000132, -0.00001272, -0.00000196, -0.00000319, -0.00000127, -0.00000392, 0.00000391, -0.00000405, 0.00000002, -0.00000504, -0.00000908, -0.00000099, -0.00000292, -0.00000790, 0.00000244, 0.00000581, -0.00000565, -0.00001421, 0.00001008, -0.00000510, -0.00000348, 0.00000033, 0.00000161, 0.00000091, -0.00000677, -0.00000611, -0.00000680, -0.00002465, 0.00000258, 0.00000442, 0.00003338, -0.00000348, 0.00001532, 0.00001007, 0.00000642, 0.00000850, 0.00001043, -0.00000254, 0.00003595, -0.00000193, -0.00000043, -0.00003266, 0.00000559, -0.00000896, 0.00000764, -0.00000448, 0.00000656, -0.00001289, -0.00000345, 0.00000087, 0.00000774, -0.00000662, 0.00000216, -0.00000483, 0.00000751, -0.00001103, 0.00000995, -0.00000054, 0.00000197, -0.00000306, 0.00000607, -0.00000400, -0.00000218, 0.00000161, 0.00001092, -0.00000413, -0.00001341, 0.00001660, -0.00001517, },
		{ 0.00000216, -0.00001233, -0.00000908, 0.00000412, -0.00000739, 0.00000333, -0.00000035, 0.00000191, 0.00000103, -0.00000436, -0.00000230, -0.00000426, -0.00000045, -0.00000225, -0.00000459, 0.00000129, 0.00000044, -0.00001952, 0.00000036, -0.00001558, 0.00002553, -0.00000823, -0.00001083, 0.00000925, 0.00000514, -0.00000014, 0.00000088, 0.00001131, -0.00001610, 0.00000448, -0.00000644, 0.00000446, -0.00000159, 0.00001271, -0.00001137, 0.00000508, -0.00000935, 0.00001019, -0.00001262, 0.00000341, -0.00001716, 0.00000286, -0.00002059, 0.00001013, -0.00001868, 0.00000233, -0.00004590, -0.00001267, -0.00018319, 0.00012104, 0.00015525, 0.00006315, 0.00009228, 0.00020343, -0.00010300, -0.00009842, -0.00000526, -0.00003610, -0.00000205, -0.00001681, 0.00000059, -0.00001524, 0.00000631, -0.00000886, 0.00000199, -0.00000302, 0.00000313, -0.00000543, 0.00000036, -0.00001109, -0.00000207, -0.00000202, -0.00000691, -0.00000246, 0.00000700, 0.00000071, -0.00001634, 0.00000256, 0.00000377, -0.00000823, 0.00000192, -0.00000026, 0.00000171, -0.00000236, -0.00001088, 0.00000041, -0.00002711, 0.00000078, -0.00000184, 0.00003112, 0.00000435, 0.00000727, 0.00001345, 0.00000820, 0.00000278, 0.00001744, -0.00000959, 0.00002956, 0.00001009, 0.00000127, -0.00002292, -0.00001162, 0.00000265, -0.00000347, 0.00000481, -0.00000107, -0.00000187, -0.00001289, 0.00000437, 0.00000354, -0.00000021, -0.00000314, -0.00000097, 0.00000318, -0.00000390, -0.00000143, 0.00000980, -0.00000540, 0.00000431, -0.00000257, 0.00000607, -0.00001107, 0.00000822, -0.00000052, 0.00001240, -0.00002432, 0.00001906, -0.00001113, },
		{ -0.00001180, 0.00000176, -0.00001880, 0.00000618, -0.00000689, 0.00000089, 0.00000071, 0.00000121, 0.00000164, -0.00000234, -0.00000373, -0.00000314, -0.00000280, -0.00000003, -0.00000466, -0.00000137, 0.00000388, -0.00001675, -0.00000296, -0.00001394, 0.00001071, 0.00001533, -0.00002593, 0.00001338, -0.00000215, 0.00001232, -0.00001260, 0.00002313, -0.00001999, 0.00000554, -0.00000950, 0.00000778, -0.00000791, 0.00001851, -0.00001320, 0.00000800, -0.00001394, 0.00001417, -0.00001488, 0.00000767, -0.00002067, 0.00000709, -0.00002481, 0.00001210, -0.00002263, 0.00001390, -0.00005209, 0.00000822, -0.00015106, -0.00002945, 0.00021128, 0.00008197, 0.00007050, 0.00020928, -0.00000424, -0.00016094, 0.00000926, -0.00005118, 0.00000623, -0.00002750, 0.00000905, -0.00002331, 0.00001039, -0.00001199, 0.00000470, -0.00000724, 0.00000688, -0.00000747, 0.00000276, -0.00001125, -0.00000224, -0.00000432, -0.00000214, -0.00000901, 0.00000804, 0.00000298, -0.00000965, -0.00001065, 0.00001197, -0.00000995, 0.00000128, -0.00000004, -0.00000010, 0.00000259, -0.00001303, 0.00000202, -0.00002244, -0.00000676, -0.00000037, 0.00001983, 0.00001763, -0.00000226, 0.00001673, 0.00000809, 0.00000331, 0.00001340, -0.00000035, 0.00000814, 0.00002906, -0.00000577, -0.00000275, -0.00003085, 0.00001143, -0.00001110, 0.00001019, -0.00000556, 0.00000526, -0.00001659, 0.00000340, 0.00000167, 0.00000385, -0.00000597, 0.00000150, -0.00000125, 0.00000375, -0.00001101, 0.00001549, -0.00000783, 0.00000724, -0.00000625, 0.00000989, -0.00001270, 0.00000785, -0.00000318, 0.00001426, -0.00001666, 0.00000411, 0.00000606, },
		{ -0.00001719, 0.00000402, -0.00001533, -0.00000189, -0.00000059, -0.00000402, 0.00000237, 0.00000032, 0.00000196, -0.00000065, -0.00000389, -0.00000302, -0.00000384, -0.00000006, -0.00000283, -0.00000382, 0.00000442, -0.00001113, -0.00000839, -0.00000694, -0.00000823, 0.00003254, -0.00002750, 0.00000749, -0.00000322, 0.00001574, -0.00001269, 0.00001831, -0.00000754, -0.00000592, -0.00000226, 0.00000132, -0.00000370, 0.00001191, -0.00000293, -0.00000020, -0.00000646, 0.00000461, -0.00000329, -0.00000262, -0.00000844, -0.00000537, -0.00001084, -0.00000448, -0.00000700, -0.00000120, -0.00002370, -0.00002247, -0.00006186, -0.00015944, 0.00018275, 0.00014636, 0.00004830, 0.00019237, 0.00009856, -0.00018819, -0.00000931, -0.00004465, -0.00000270, -0.00002381, 0.00000392, -0.00001861, 0.00000297, -0.00000591, 0.00000003, -0.00000480, 0.00000409, -0.00000425, 0.00000127, -0.00000904, -0.00000292, -0.00000631, 0.00000088, -0.00001129, 0.00000484, 0.00000497, -0.00000201, -0.00001810, 0.00000963, -0.00000277, -0.00000497, 0.00000420, -0.00000443, 0.00000646, -0.00001132, -0.00000149, -0.00001321, -0.00001655, 0.00000429, 0.00000630, 0.00002866, -0.00000718, 0.00001670, 0.00000725, 0.00000840, 0.00000317, 0.00001445, -0.00001174, 0.00003831, -0.00000547, 0.00000829, -0.00003268, 0.00000287, -0.00000476, 0.00000361, -0.00000014, 0.00000251, -0.00001029, -0.00000472, 0.00000475, 0.00000274, -0.00000376, -0.00000004, -0.00000213, 0.00000654, -0.00001251, 0.00001128, -0.00000153, 0.00000237, -0.00000179, 0.00000492, -0.00000529, -0.00000105, 0.00000399, 0.00000391, 0.00000096, -0.00001426, 0.00001963, },
		{ -0.00000587, -0.00000743, -0.00000267, -0.00001278, 0.00000440, -0.00000624, 0.00000173, 0.00000047, 0.00000185, 0.00000025, -0.00000263, -0.00000379, -0.00000363, -0.00000164, -0.00000064, -0.00000480, 0.00000254, -0.00000472, -0.00001332, 0.00000005, -0.00002107, 0.00003333, -0.00001190, -0.00000781, 0.00000271, 0.00000822, 0.00000018, 0.00000148, 0.00001251, -0.00001964, 0.00000591, -0.00000673, 0.00000343, 0.00000114, 0.00001023, -0.00001030, 0.00000393, -0.00000837, 0.00001077, -0.00001489, 0.00000546, -0.00001883, 0.00000376, -0.00002045, 0.00000813, -0.00001907, 0.00000638, -0.00005253, 0.00000011, -0.00018719, 0.00005092, 0.00022734, 0.00004855, 0.00015429, 0.00018648, -0.00016667, -0.00005979, -0.00002198, -0.00002415, -0.00000972, -0.00001107, -0.00000526, -0.00001133, 0.00000472, -0.00000805, 0.00000100, -0.00000205, 0.00000125, -0.00000207, -0.00000501, -0.00000566, -0.00000451, -0.00000198, -0.00000638, -0.00000373, 0.00000890, 0.00000089, -0.00001436, -0.00000365, 0.00000945, -0.00001211, 0.00000793, -0.00000669, 0.00000588, -0.00000521, -0.00000773, -0.00000383, -0.00002347, 0.00000625, -0.00000234, 0.00003090, -0.00000330, 0.00001095, 0.00000871, 0.00001169, -0.00000254, 0.00001990, -0.00001426, 0.00002654, 0.00001043, 0.00000227, -0.00001435, -0.00001903, 0.00000887, -0.00000788, 0.00000852, -0.00000371, -0.00000104, -0.00001296, 0.00000712, 0.00000106, -0.00000061, -0.00000238, -0.00000133, 0.00000535, -0.00000752, 0.00000101, 0.00000812, -0.00000459, 0.00000475, -0.00000201, 0.00000340, -0.00000995, 0.00001081, -0.00000483, 0.00001145, -0.00001905, 0.00001598, },
		{ 0.00001203, -0.00001875, 0.00000453, -0.00001450, 0.00000064, -0.00000228, -0.00000227, 0.00000193, 0.00000107, 0.00000104, -0.00000144, -0.00000371, -0.00000395, -0.00000253, -0.00000032, -0.00000326, -0.00000105, 0.00000089, -0.00001589, 0.00000321, -0.00002306, 0.00001850, 0.00001211, -0.00002197, 0.00000759, -0.00000166, 0.00001372, -0.00001172, 0.00002341, -0.00002086, 0.00000331, -0.00000637, 0.00000377, -0.00000317, 0.00001489, -0.00001115, 0.00000583, -0.00001208, 0.00001346, -0.00001459, 0.00000577, -0.00001739, 0.00000243, -0.00001878, 0.00000493, -0.00001792, 0.00000986, -0.00004539, -0.00000282, -0.00011755, -0.00011247, 0.00024562, 0.00010722, 0.00009317, 0.00025108, -0.00010156, -0.00012249, -0.00000225, -0.00004284, 0.00000144, -0.00002358, 0.00000489, -0.00002130, 0.00000999, -0.00001135, 0.00000332, -0.00000550, 0.00000408, -0.00000389, -0.00000113, -0.00000946, -0.00000002, -0.00000815, 0.00000107, -0.00001197, 0.00001088, 0.00000137, -0.00000509, -0.00001704, 0.00001499, -0.00001103, 0.00000451, -0.00000220, -0.00000082, 0.00000379, -0.00001337, 0.00000201, -0.00002458, 0.00000283, -0.00000358, 0.00002371, 0.00000777, 0.00000147, 0.00001259, 0.00001000, 0.00000113, 0.00001189, -0.00000000, 0.00000161, 0.00003083, -0.00000824, 0.00000637, -0.00003378, 0.00001183, -0.00000961, 0.00000826, -0.00000278, 0.00000117, -0.00001287, 0.00000247, 0.00000392, -0.00000110, -0.00000165, -0.00000211, 0.00000383, -0.00000166, -0.00000747, 0.00001275, -0.00000590, 0.00000588, -0.00000327, 0.00000560, -0.00001099, 0.00000888, -0.00000286, 0.00000738, -0.00000793, -0.00000064, },
		{ 0.00001879, -0.00001514, -0.00000314, -0.00000403, -0.00000993, 0.00000395, -0.00000599, 0.00000242, 0.00000057, 0.00000179, -0.00000111, -0.00000210, -0.00000527, -0.00000218, -0.00000189, -0.00000042, -0.00000426, 0.00000395, -0.00001496, 0.00000132, -0.00001583, -0.00000281, 0.00003058, -0.00002392, 0.00000421, -0.00000554, 0.00001722, -0.00001087, 0.00001764, -0.00000672, -0.00000948, 0.00000108, -0.00000210, -0.00000052, 0.00001041, -0.00000279, -0.00000080, -0.00000594, 0.00000453, -0.00000252, -0.00000563, -0.00000422, -0.00001038, -0.00000531, -0.00000981, -0.00000402, -0.00000573, -0.00001467, -0.00003728, -0.00003132, -0.00019980, 0.00014278, 0.00021858, 0.00003538, 0.00027527, -0.00000004, -0.00017705, -0.00000079, -0.00004842, 0.00000098, -0.00002574, 0.00000461, -0.00002076, 0.00000531, -0.00000685, -0.00000015, -0.00000416, 0.00000271, -0.00000310, 0.00000111, -0.00001172, 0.00000271, -0.00001110, 0.00000326, -0.00001266, 0.00000588, 0.00000452, 0.00000068, -0.00001961, 0.00000656, 0.00000057, -0.00000565, 0.00000690, -0.00000946, 0.00001057, -0.00001438, 0.00000220, -0.00001978, -0.00000490, 0.00000036, 0.00001164, 0.00001990, -0.00000624, 0.00001531, 0.00000616, 0.00000934, -0.00000038, 0.00001570, -0.00001626, 0.00003632, -0.00000527, 0.00001074, -0.00002637, -0.00000391, 0.00000202, -0.00000230, 0.00000581, -0.00000374, -0.00000563, -0.00000673, 0.00000889, -0.00000329, 0.00000047, -0.00000344, 0.00000250, 0.00000244, -0.00001053, 0.00000988, -0.00000072, 0.00000150, 0.00000078, 0.00000221, -0.00000578, 0.00000069, 0.00000557, -0.00000286, 0.00000547, -0.00001296, },
		{ 0.00000859, 0.00000222, -0.00001819, 0.00000726, -0.00001577, 0.00000365, -0.00000405, -0.00000071, 0.00000165, 0.00000170, -0.00000077, -0.00000049, -0.00000571, -0.00000248, -0.00000288, 0.00000069, -0.00000446, 0.00000306, -0.00001023, -0.00000474, -0.00000435, -0.00002045, 0.00003467, -0.00001190, -0.00000552, -0.00000390, 0.00001119, 0.00000014, 0.00000368, 0.00001040, -0.00001932, 0.00000434, -0.00000479, 0.00000076, 0.00000552, 0.00000505, -0.00000606, -0.00000016, -0.00000423, 0.00000759, -0.00001336, 0.00000382, -0.00001648, 0.00000046, -0.00001547, 0.00000073, -0.00001351, 0.00000290, -0.00004937, -0.00000191, -0.00016548, -0.00004573, 0.00030481, 0.00003928, 0.00023496, 0.00013147, -0.00020961, -0.00002115, -0.00004255, -0.00000876, -0.00002025, -0.00000299, -0.00001365, -0.00000442, 0.00000019, -0.00000480, -0.00000178, 0.00000026, -0.00000181, 0.00000205, -0.00001098, 0.00000075, -0.00000745, -0.00000223, -0.00000553, -0.00000438, 0.00000977, 0.00000101, -0.00001146, -0.00000979, 0.00001309, -0.00001301, 0.00001174, -0.00001181, 0.00000910, -0.00000801, -0.00000329, -0.00001096, -0.00001414, 0.00000563, 0.00000026, 0.00002717, -0.00000768, 0.00001349, 0.00000471, 0.00001413, -0.00000542, 0.00001892, -0.00001454, 0.00002050, 0.00001310, -0.00000031, -0.00000440, -0.00002500, 0.00001293, -0.00001037, 0.00001078, -0.00000570, -0.00000118, -0.00001110, 0.00000827, -0.00000102, -0.00000104, -0.00000158, -0.00000087, 0.00000597, -0.00001017, 0.00000384, 0.00000505, -0.00000211, 0.00000364, 0.00000008, -0.00000177, -0.00000527, 0.00001014, -0.00000651, 0.00000751, -0.00000979, },
		{ -0.00000616, 0.00001659, -0.00002305, 0.00000478, -0.00000788, -0.00000654, 0.00000343, -0.00000606, 0.00000330, 0.00000086, 0.00000040, -0.00000053, -0.00000360, -0.00000488, -0.00000163, -0.00000133, -0.00000130, -0.00000088, -0.00000338, -0.00001210, 0.00000568, -0.00002826, 0.00002459, 0.00000632, -0.00001312, -0.00000277, 0.00000306, 0.00001019, -0.00000491, 0.00001695, -0.00001576, -0.00000318, 0.00000019, -0.00000348, 0.00000534, 0.00000641, -0.00000406, -0.00000110, -0.00000509, 0.00000709, -0.00000879, -0.00000155, -0.00000832, -0.00000828, -0.00000601, -0.00000959, -0.00000497, -0.00000417, -0.00002692, -0.00002838, -0.00006692, -0.00020160, 0.00026568, 0.00014214, 0.00013601, 0.00026648, -0.00020115, -0.00006418, -0.00003019, -0.00002238, -0.00001301, -0.00001153, -0.00000687, -0.00001211, 0.00000341, -0.00000576, -0.00000197, -0.00000046, -0.00000168, 0.00000215, -0.00000741, -0.00000496, -0.00000003, -0.00001010, 0.00000197, -0.00001187, 0.00001094, 0.00000129, -0.00000202, -0.00002037, 0.00001380, -0.00000859, 0.00000611, -0.00000405, -0.00000114, 0.00000341, -0.00001096, -0.00000177, -0.00002156, 0.00000873, -0.00000656, 0.00002693, -0.00000218, 0.00000712, 0.00000706, 0.00001243, -0.00000036, 0.00000961, 0.00000069, -0.00000303, 0.00003004, -0.00000842, 0.00001035, -0.00003028, 0.00000709, -0.00000444, 0.00000369, 0.00000187, -0.00000520, -0.00000642, -0.00000021, 0.00000668, -0.00000648, 0.00000319, -0.00000565, 0.00000889, -0.00000841, -0.00000112, 0.00000671, -0.00000097, 0.00000189, 0.00000229, -0.00000201, -0.00000566, 0.00000734, -0.00000092, -0.00000105, 0.00000247, },
		{ -0.00000905, 0.00001396, -0.00000955, -0.00001220, 0.00000809, -0.00001792, 0.00000815, -0.00000754, 0.00000201, 0.00000102, 0.00000152, -0.00000158, -0.00000019, -0.00000768, 0.00000004, -0.00000407, 0.00000192, -0.00000412, 0.00000178, -0.00001643, 0.00000930, -0.00002458, 0.00000617, 0.00002143, -0.00001309, -0.00000445, -0.00000314, 0.00001361, -0.00000388, 0.00001232, -0.00000314, -0.00001396, 0.00000510, -0.00000663, 0.00000463, 0.00000609, -0.00000015, -0.00000348, -0.00000306, 0.00000200, 0.00000009, -0.00001015, 0.00000167, -0.00001703, 0.00000217, -0.00001721, 0.00000130, -0.00001303, -0.00000346, -0.00005314, -0.00000359, -0.00022386, 0.00007815, 0.00029498, 0.00004478, 0.00034726, -0.00012236, -0.00013412, -0.00001184, -0.00003969, -0.00000474, -0.00002001, -0.00000159, -0.00001691, 0.00000308, -0.00000453, -0.00000283, -0.00000128, -0.00000087, 0.00000011, -0.00000146, -0.00001152, 0.00000579, -0.00001341, 0.00000277, -0.00001098, 0.00000474, 0.00000524, 0.00000159, -0.00001806, 0.00000077, 0.00000465, -0.00000501, 0.00000746, -0.00001213, 0.00001179, -0.00001435, 0.00000291, -0.00002357, 0.00000674, -0.00000676, 0.00001931, 0.00000801, -0.00000110, 0.00001106, 0.00000707, 0.00000872, -0.00000149, 0.00001335, -0.00001519, 0.00002896, -0.00000005, 0.00000650, -0.00001471, -0.00001315, 0.00000962, -0.00000839, 0.00001178, -0.00001074, -0.00000008, -0.00000850, 0.00001208, -0.00000890, 0.00000468, -0.00000669, 0.00000773, -0.00000387, -0.00000544, 0.00000580, 0.00000190, -0.00000063, 0.00000448, -0.00000198, -0.00000520, 0.00000287, 0.00000573, -0.00000810, 0.00000835, },
		{ 0.00000182, -0.00000112, 0.00001070, -0.00002727, 0.00001502, -0.00001671, 0.00000149, -0.00000091, -0.00000379, 0.00000310, 0.00000130, -0.00000157, 0.00000144, -0.00000745, -0.00000122, -0.00000380, 0.00000131, -0.00000316, 0.00000216, -0.00001469, 0.00000450, -0.00001179, -0.00001358, 0.00002859, -0.00000721, -0.00000509, -0.00000918, 0.00001288, 0.00000061, 0.00000664, 0.00000554, -0.00001529, -0.00000007, -0.00000106, -0.00000284, 0.00001050, -0.00000153, 0.00000008, -0.00000607, 0.00000211, 0.00000131, -0.00000842, -0.00000116, -0.00001063, -0.00000630, -0.00000673, -0.00001077, -0.00000306, -0.00000755, -0.00003671, -0.00001773, -0.00012310, -0.00015608, 0.00036904, 0.00005438, 0.00032073, 0.00003922, -0.00022188, 0.00000652, -0.00005909, 0.00000449, -0.00002949, 0.00000408, -0.00002172, 0.00000325, -0.00000573, -0.00000025, -0.00000540, 0.00000296, -0.00000492, 0.00000556, -0.00001595, 0.00000661, -0.00001001, -0.00000287, -0.00000466, -0.00000406, 0.00000920, 0.00000154, -0.00000848, -0.00001451, 0.00001325, -0.00000951, 0.00001163, -0.00001378, 0.00000959, -0.00000891, 0.00000025, -0.00001829, -0.00000158, -0.00000030, 0.00000645, 0.00001943, -0.00000791, 0.00001372, 0.00000253, 0.00001456, -0.00000513, 0.00001402, -0.00001021, 0.00001146, 0.00001778, -0.00000500, 0.00000467, -0.00002758, 0.00001343, -0.00000996, 0.00001093, -0.00000663, -0.00000236, -0.00000734, 0.00000758, -0.00000237, -0.00000161, -0.00000079, 0.00000060, 0.00000438, -0.00001072, 0.00000587, 0.00000170, 0.00000106, 0.00000184, 0.00000271, -0.00000820, 0.00000207, 0.00000642, -0.00000563, 0.00000144, },
		{ 0.00001231, -0.00001046, 0.00001794, -0.00002332, 0.00000224, 0.00000015, -0.00001399, 0.00000899, -0.00000961, 0.00000442, 0.00000061, -0.00000008, 0.00000018, -0.00000337, -0.00000575, -0.00000033, -0.00000287, 0.00000108, -0.00000107, -0.00000828, -0.00000571, 0.00000368, -0.00002773, 0.00002675, 0.00000009, -0.00000158, -0.00001553, 0.00001030, 0.00000318, 0.00000638, 0.00000498, -0.00000573, -0.00001274, 0.00000868, -0.00001224, 0.00001565, -0.00000499, 0.00000631, -0.00001127, 0.00000569, -0.00000375, 0.00000133, -0.00001281, 0.00000466, -0.00002276, 0.00001082, -0.00002859, 0.00001340, -0.00002521, -0.00000062, -0.00006232, -0.00000965, -0.00028224, 0.00025854, 0.00019274, 0.00020306, 0.00023926, -0.00028191, -0.00000159, -0.00006552, 0.00000451, -0.00003316, 0.00000557, -0.00002371, 0.00000242, -0.00000811, 0.00000379, -0.00001023, 0.00000685, -0.00000927, 0.00000950, -0.00001506, 0.00000172, -0.00000261, -0.00000925, -0.00000033, -0.00000804, 0.00000758, 0.00000350, -0.00000148, -0.00001950, 0.00000757, -0.00000193, 0.00000515, -0.00000467, -0.00000179, 0.00000225, -0.00000677, -0.00000837, -0.00001345, 0.00000917, -0.00000679, 0.00002740, -0.00001005, 0.00001266, 0.00000162, 0.00001433, -0.00000064, 0.00000636, 0.00000168, -0.00000563, 0.00002631, -0.00000526, 0.00000847, -0.00002073, -0.00000174, 0.00000308, -0.00000206, 0.00000684, -0.00001215, 0.00000121, -0.00000343, 0.00000876, -0.00001085, 0.00000701, -0.00000746, 0.00001188, -0.00001420, 0.00000591, -0.00000073, 0.00000531, -0.00000317, 0.00000875, -0.00001120, 0.00000207, 0.00000389, 0.00000167, -0.00000917, },
		{ 0.00000907, -0.00000280, 0.00000596, -0.00000191, -0.00002079, 0.00001828, -0.00002359, 0.00001054, -0.00000787, 0.00000084, 0.00000178, 0.00000100, -0.00000172, 0.00000121, -0.00000947, 0.00000208, -0.00000593, 0.00000362, -0.00000308, -0.00000234, -0.00001451, 0.00001337, -0.00003035, 0.00001580, 0.00000765, 0.00000263, -0.00001654, 0.00000271, 0.00000602, 0.00000715, 0.00000318, 0.00000254, -0.00001851, 0.00000876, -0.00001117, 0.00001063, -0.00000014, 0.00000442, -0.00000789, 0.00000190, -0.00000270, 0.00000407, -0.00001536, 0.00000804, -0.00002393, 0.00001009, -0.00002527, 0.00000774, -0.00002161, 0.00000756, -0.00006632, 0.00001621, -0.00022830, -0.00000679, 0.00036212, 0.00009782, 0.00038124, -0.00024524, -0.00007105, -0.00003857, -0.00001990, -0.00001913, -0.00000714, -0.00001429, -0.00000709, -0.00000396, 0.00000120, -0.00000794, 0.00000361, -0.00000636, 0.00000518, -0.00000649, -0.00000787, 0.00000520, -0.00001213, -0.00000116, -0.00000609, 0.00000130, 0.00000729, 0.00000050, -0.00001351, -0.00000725, 0.00000902, -0.00000292, 0.00000562, -0.00001192, 0.00000957, -0.00001062, -0.00000042, -0.00002227, 0.00001459, -0.00001339, 0.00002659, -0.00000451, 0.00000631, 0.00000509, 0.00000927, 0.00000692, -0.00000052, 0.00000790, -0.00000928, 0.00001704, 0.00000935, -0.00000265, -0.00000043, -0.00002210, 0.00001558, -0.00001246, 0.00001572, -0.00001649, 0.00000473, -0.00000879, 0.00001297, -0.00001251, 0.00000729, -0.00000814, 0.00001163, -0.00001048, 0.00000126, 0.00000014, 0.00000548, -0.00000325, 0.00000851, -0.00000704, -0.00000354, 0.00000517, 0.00000428, -0.00001102, },
		{ -0.00000506, 0.00001334, -0.00001051, 0.00001678, -0.00003187, 0.00001842, -0.00001456, -0.00000223, 0.00000286, -0.00000710, 0.00000461, 0.00000055, -0.00000170, 0.00000249, -0.00000832, -0.00000042, -0.00000400, 0.00000093, -0.00000069, -0.00000069, -0.00001663, 0.00001207, -0.00001987, -0.00000236, 0.00001631, 0.00000162, -0.00000675, -0.00001223, 0.00001236, 0.00000252, 0.00000904, -0.00000060, -0.00000887, -0.00000609, 0.00000322, -0.00000602, 0.00001440, -0.00000751, 0.00000601, -0.00001171, 0.00000861, -0.00000619, -0.00000175, -0.00000777, -0.00000318, -0.00001428, 0.00000298, -0.00002311, 0.00000890, -0.00002155, -0.00001847, -0.00004159, -0.00007106, -0.00026117, 0.00039646, 0.00011820, 0.00038980, -0.00007962, -0.00020083, 0.00001570, -0.00006500, 0.00001045, -0.00003336, 0.00000672, -0.00002628, 0.00000882, -0.00001074, 0.00000381, -0.00000838, 0.00000469, -0.00000674, 0.00000702, -0.00001826, 0.00001028, -0.00001116, -0.00000414, -0.00000384, -0.00000277, 0.00000724, 0.00000263, -0.00000594, -0.00001708, 0.00000960, -0.00000154, 0.00000718, -0.00001175, 0.00000639, -0.00000661, 0.00000083, -0.00002264, 0.00001038, -0.00000871, 0.00001474, 0.00000876, -0.00000441, 0.00001145, 0.00000242, 0.00001285, -0.00000199, 0.00000596, -0.00000242, 0.00000066, 0.00002309, -0.00000969, 0.00001069, -0.00002547, 0.00000973, -0.00000634, 0.00000877, -0.00000637, -0.00000433, -0.00000212, 0.00000520, -0.00000288, -0.00000240, 0.00000011, 0.00000279, 0.00000062, -0.00000865, 0.00000615, -0.00000082, 0.00000378, 0.00000046, 0.00000444, -0.00001394, 0.00001010, 0.00000095, -0.00000322, },
		{ -0.00001505, 0.00001941, -0.00001181, 0.00001490, -0.00001903, -0.00000268, 0.00000823, -0.00002006, 0.00001297, -0.00001194, 0.00000480, 0.00000044, -0.00000028, 0.00000055, -0.00000304, -0.00000628, 0.00000097, -0.00000454, 0.00000358, -0.00000179, -0.00001285, 0.00000207, -0.00000220, -0.00002101, 0.00002343, -0.00000413, 0.00000947, -0.00002720, 0.00001689, -0.00000419, 0.00001870, -0.00000909, 0.00000620, -0.00002247, 0.00001624, -0.00001950, 0.00002447, -0.00001579, 0.00001652, -0.00002134, 0.00001693, -0.00001578, 0.00001246, -0.00002439, 0.00001748, -0.00003641, 0.00002647, -0.00004633, 0.00003114, -0.00004657, 0.00002548, -0.00009401, 0.00003942, -0.00033664, 0.00020938, 0.00026799, 0.00029277, 0.00015822, -0.00032347, 0.00004668, -0.00009511, 0.00002832, -0.00005159, 0.00002151, -0.00004014, 0.00001761, -0.00002081, 0.00001418, -0.00001871, 0.00001389, -0.00001628, 0.00001585, -0.00002188, 0.00000882, -0.00000678, -0.00000604, -0.00000534, -0.00000121, 0.00000149, 0.00000754, -0.00000324, -0.00001479, -0.00000284, 0.00000819, 0.00000169, -0.00000363, -0.00000339, 0.00000133, -0.00000245, -0.00001558, -0.00000212, 0.00000429, -0.00000305, 0.00002370, -0.00001399, 0.00001610, -0.00000220, 0.00001470, 0.00000068, 0.00000209, 0.00000299, -0.00000645, 0.00002001, 0.00000093, 0.00000175, -0.00000725, -0.00001225, 0.00001070, -0.00000697, 0.00001020, -0.00001755, 0.00000800, -0.00000570, 0.00000891, -0.00001288, 0.00000852, -0.00000628, 0.00001112, -0.00001676, 0.00001112, -0.00000714, 0.00001063, -0.00000710, 0.00001369, -0.00001933, 0.00000999, -0.00000002, 0.00000356, },
		{ -0.00001148, 0.00000780, 0.00000483, -0.00000360, 0.00000573, -0.00002642, 0.00002398, -0.00002496, 0.00000992, -0.00000639, -0.00000127, 0.00000286, 0.00000017, -0.00000125, 0.00000154, -0.00000986, 0.00000311, -0.00000677, 0.00000423, -0.00000110, -0.00000812, -0.00000898, 0.00001212, -0.00003046, 0.00002240, -0.00000650, 0.00001890, -0.00002824, 0.00000895, -0.00000268, 0.00001929, -0.00000729, 0.00000845, -0.00002153, 0.00001052, -0.00001402, 0.00001557, -0.00000635, 0.00000921, -0.00001250, 0.00000743, -0.00000810, 0.00000790, -0.00001942, 0.00001278, -0.00002868, 0.00001582, -0.00003135, 0.00001270, -0.00002860, 0.00001490, -0.00007241, 0.00002324, -0.00021128, -0.00010610, 0.00040702, 0.00021372, 0.00034954, -0.00034008, -0.00000734, -0.00007000, 0.00000394, -0.00003696, 0.00000852, -0.00002976, 0.00000574, -0.00001357, 0.00000850, -0.00001382, 0.00000891, -0.00001220, 0.00001066, -0.00001268, -0.00000165, 0.00000131, -0.00000761, -0.00000774, 0.00000093, -0.00000351, 0.00000996, -0.00000188, -0.00000699, -0.00001622, 0.00001299, 0.00000045, 0.00000185, -0.00000925, 0.00000452, -0.00000391, -0.00000735, -0.00001515, 0.00001599, -0.00001599, 0.00003024, -0.00001449, 0.00001320, -0.00000061, 0.00001137, 0.00000499, 0.00000132, 0.00000100, -0.00000072, 0.00000301, 0.00002032, -0.00001333, 0.00001270, -0.00002767, 0.00001752, -0.00001254, 0.00001579, -0.00001910, 0.00000725, -0.00000658, 0.00001062, -0.00001294, 0.00000709, -0.00000647, 0.00001238, -0.00001513, 0.00000741, -0.00000526, 0.00000854, -0.00000509, 0.00001155, -0.00001182, -0.00000118, 0.00000720, 0.00000150, },
		{ -0.00000083, -0.00001007, 0.00002225, -0.00001700, 0.00001819, -0.00003009, 0.00001622, -0.00000899, -0.00000717, 0.00000667, -0.00001002, 0.00000581, -0.00000047, -0.00000097, 0.00000222, -0.00000781, -0.00000046, -0.00000329, -0.00000043, 0.00000274, -0.00000545, -0.00001507, 0.00001572, -0.00002547, 0.00000971, 0.00000093, 0.00001243, -0.00000962, -0.00001301, 0.00000939, 0.00000633, 0.00000970, -0.00000620, -0.00000199, -0.00001192, 0.00000648, -0.00000733, 0.00001560, -0.00001090, 0.00000955, -0.00001481, 0.00001287, -0.00001245, 0.00000426, -0.00001359, 0.00000324, -0.00002051, 0.00001016, -0.00003219, 0.00001778, -0.00003392, -0.00000090, -0.00006481, -0.00002353, -0.00034016, 0.00036382, 0.00025505, 0.00041601, -0.00020573, -0.00015253, 0.00000488, -0.00005735, 0.00000639, -0.00002957, 0.00000283, -0.00002510, 0.00001000, -0.00001279, 0.00000564, -0.00000933, 0.00000429, -0.00000622, 0.00000543, -0.00001676, 0.00001056, -0.00001011, -0.00000628, -0.00000294, -0.00000088, 0.00000433, 0.00000401, -0.00000387, -0.00001764, 0.00000308, 0.00000939, -0.00000044, -0.00000641, -0.00000010, -0.00000107, -0.00000244, -0.00002164, 0.00001786, -0.00001582, 0.00002231, -0.00000250, 0.00000123, 0.00000748, 0.00000382, 0.00000969, 0.00000272, -0.00000338, 0.00000663, -0.00000978, 0.00002704, -0.00001227, 0.00001217, -0.00001843, 0.00000222, 0.00000000, 0.00000461, -0.00000510, -0.00000663, 0.00000379, 0.00000163, -0.00000265, -0.00000341, 0.00000140, 0.00000498, -0.00000446, -0.00000433, 0.00000448, -0.00000187, 0.00000524, 0.00000036, 0.00000407, -0.00001702, 0.00001648, -0.00000435, },
		{ 0.00000355, -0.00001723, 0.00002229, -0.00000938, 0.00000741, -0.00001075, -0.00000884, 0.00001477, -0.00002315, 0.00001410, -0.00001214, 0.00000396, 0.00000052, -0.00000022, 0.00000058, -0.00000272, -0.00000618, 0.00000175, -0.00000574, 0.00000572, -0.00000309, -0.00001603, 0.00000943, -0.00001019, -0.00000923, 0.00001438, -0.00000446, 0.00001682, -0.00003410, 0.00001929, -0.00000816, 0.00002767, -0.00002069, 0.00001581, -0.00002862, 0.00001942, -0.00002187, 0.00002809, -0.00002190, 0.00002233, -0.00002699, 0.00002415, -0.00002441, 0.00001997, -0.00003140, 0.00002482, -0.00004337, 0.00003441, -0.00005560, 0.00004036, -0.00005940, 0.00004103, -0.00011069, 0.00006407, -0.00034726, 0.00010439, 0.00037862, 0.00039682, 0.00002111, -0.00031160, 0.00006403, -0.00010558, 0.00003854, -0.00005964, 0.00002874, -0.00004924, 0.00002726, -0.00002932, 0.00002085, -0.00002351, 0.00001720, -0.00001954, 0.00001822, -0.00002494, 0.00001361, -0.00001036, -0.00000226, -0.00001112, 0.00000644, -0.00000525, 0.00001165, -0.00000571, -0.00000821, -0.00001488, 0.00001938, -0.00000295, -0.00000144, -0.00000592, 0.00000130, 0.00000041, -0.00002080, 0.00000934, -0.00000396, 0.00000417, 0.00001568, -0.00001312, 0.00001613, -0.00000331, 0.00001295, 0.00000359, -0.00000294, 0.00000457, -0.00000603, 0.00001224, 0.00000874, -0.00000736, 0.00000689, -0.00002142, 0.00001597, -0.00000917, 0.00001030, -0.00001960, 0.00001218, -0.00000582, 0.00000636, -0.00001186, 0.00000723, -0.00000186, 0.00000605, -0.00001479, 0.00001262, -0.00001041, 0.00001294, -0.00000789, 0.00001483, -0.00002375, 0.00001565, -0.00000276, },
		{ -0.00000201, -0.00000993, 0.00000524, 0.00001236, -0.00001272, 0.00001128, -0.00002716, 0.00002356, -0.00002096, 0.00000575, -0.00000297, -0.00000420, 0.00000412, -0.00000089, -0.00000009, 0.00000053, -0.00000836, 0.00000255, -0.00000611, 0.00000352, 0.00000150, -0.00001463, -0.00000070, 0.00000546, -0.00002345, 0.00002275, -0.00001645, 0.00003073, -0.00003504, 0.00001170, -0.00000866, 0.00002859, -0.00001559, 0.00001194, -0.00002101, 0.00000857, -0.00001325, 0.00001712, -0.00000999, 0.00001171, -0.00001484, 0.00001102, -0.00001172, 0.00000919, -0.00001983, 0.00001309, -0.00002824, 0.00001603, -0.00003189, 0.00001250, -0.00003007, 0.00001426, -0.00006677, 0.00001314, -0.00017200, -0.00021421, 0.00042132, 0.00040345, 0.00023142, -0.00037965, 0.00003462, -0.00009110, 0.00002105, -0.00004972, 0.00001973, -0.00004159, 0.00001616, -0.00002135, 0.00001375, -0.00001740, 0.00001181, -0.00001571, 0.00001394, -0.00001743, 0.00000473, -0.00000386, -0.00000180, -0.00001477, 0.00000783, -0.00000776, 0.00001166, -0.00000395, -0.00000045, -0.00002416, 0.00001550, 0.00000483, -0.00000299, -0.00000518, -0.00000196, 0.00000397, -0.00001594, -0.00000351, 0.00001034, -0.00001250, 0.00002788, -0.00001913, 0.00001688, -0.00000398, 0.00001182, 0.00000419, 0.00000250, -0.00000516, 0.00000758, -0.00000979, 0.00002939, -0.00002164, 0.00002108, -0.00002744, 0.00001399, -0.00000774, 0.00001114, -0.00001749, 0.00000658, -0.00000156, 0.00000490, -0.00000993, 0.00000381, -0.00000130, 0.00000895, -0.00001601, 0.00001093, -0.00000854, 0.00000957, -0.00000482, 0.00001222, -0.00001496, 0.00000121, 0.00000870, },
		{ -0.00000771, 0.00000013, -0.00001200, 0.00002665, -0.00001890, 0.00001390, -0.00002163, 0.00000864, -0.00000071, -0.00001241, 0.00000962, -0.00001179, 0.00000615, -0.00000141, 0.00000047, 0.00000066, -0.00000585, -0.00000140, -0.00000183, -0.00000233, 0.00000643, -0.00001139, -0.00000996, 0.00001409, -0.00002583, 0.00001871, -0.00001346, 0.00002209, -0.00001196, -0.00001234, 0.00000464, 0.00001167, 0.00000808, -0.00000978, 0.00000351, -0.00001586, 0.00000742, -0.00000583, 0.00001316, -0.00001050, 0.00000925, -0.00001380, 0.00001306, -0.00001534, 0.00000743, -0.00001639, 0.00000627, -0.00002244, 0.00001186, -0.00003453, 0.00001950, -0.00003966, 0.00000966, -0.00007868, 0.00000625, -0.00037582, 0.00025482, 0.00048217, 0.00037509, -0.00031511, -0.00009047, -0.00002097, -0.00003806, -0.00000671, -0.00001871, -0.00000729, -0.00001792, 0.00000601, -0.00001088, 0.00000434, -0.00000757, 0.00000136, -0.00000317, 0.00000076, -0.00001137, 0.00000717, -0.00000671, -0.00000925, -0.00000182, 0.00000106, 0.00000119, 0.00000514, -0.00000192, -0.00001694, -0.00000438, 0.00002055, -0.00000871, 0.00000018, -0.00000802, 0.00000619, -0.00000877, -0.00001479, 0.00001844, -0.00001834, 0.00002620, -0.00001154, 0.00000680, 0.00000323, 0.00000562, 0.00000629, 0.00000724, -0.00001166, 0.00001429, -0.00001745, 0.00002771, -0.00001120, 0.00000874, -0.00000744, -0.00000764, 0.00000775, -0.00000069, -0.00000327, -0.00000860, 0.00000942, -0.00000240, -0.00000197, -0.00000456, 0.00000329, 0.00000619, -0.00000943, 0.00000107, 0.00000145, -0.00000154, 0.00000534, 0.00000169, 0.00000119, -0.00001624, 0.00001935, },
		{ -0.00000317, 0.00000106, -0.00001525, 0.00002066, -0.00000354, -0.00000328, -0.00000018, -0.00001451, 0.00001819, -0.00002204, 0.00001150, -0.00000962, 0.00000156, 0.00000114, -0.00000013, 0.00000046, -0.00000261, -0.00000512, 0.00000147, -0.00000578, 0.00000682, -0.00000469, -0.00001705, 0.00001453, -0.00001775, 0.00000508, 0.00000092, -0.00000028, 0.00001835, -0.00003358, 0.00001485, -0.00000570, 0.00002948, -0.00002554, 0.00001908, -0.00002781, 0.00001539, -0.00001667, 0.00002357, -0.00001997, 0.00002011, -0.00002437, 0.00002316, -0.00002518, 0.00001934, -0.00002924, 0.00002183, -0.00003842, 0.00002883, -0.00004988, 0.00003371, -0.00005530, 0.00003634, -0.00010096, 0.00005084, -0.00030117, -0.00006526, 0.00053424, 0.00050083, -0.00016336, -0.00024230, 0.00004059, -0.00008752, 0.00002710, -0.00005031, 0.00002090, -0.00004488, 0.00002550, -0.00002841, 0.00001928, -0.00002080, 0.00001349, -0.00001609, 0.00001388, -0.00002148, 0.00001321, -0.00001063, -0.00000055, -0.00001478, 0.00001182, -0.00000975, 0.00001331, -0.00000639, -0.00000280, -0.00002477, 0.00002810, -0.00000634, 0.00000041, -0.00000842, 0.00000192, 0.00000113, -0.00002207, 0.00001779, -0.00001264, 0.00001300, 0.00000457, -0.00000782, 0.00001245, -0.00000142, 0.00000911, 0.00000761, -0.00000812, 0.00000621, -0.00000503, 0.00000442, 0.00001609, -0.00001584, 0.00001821, -0.00002644, 0.00001705, -0.00000757, 0.00000645, -0.00001739, 0.00001278, -0.00000335, 0.00000118, -0.00000804, 0.00000369, 0.00000477, -0.00000236, -0.00000850, 0.00000989, -0.00000963, 0.00001132, -0.00000466, 0.00001104, -0.00002274, 0.00001726, },
		{ 0.00000944, -0.00000549, -0.00000721, 0.00000246, 0.00001816, -0.00001945, 0.00001314, -0.00002215, 0.00001676, -0.00001179, -0.00000155, 0.00000199, -0.00000756, 0.00000533, -0.00000205, 0.00000166, -0.00000176, -0.00000491, 0.00000017, -0.00000377, 0.00000155, 0.00000414, -0.00002051, 0.00000858, -0.00000510, -0.00000975, 0.00001527, -0.00001880, 0.00003409, -0.00003330, 0.00000777, -0.00000852, 0.00003128, -0.00001826, 0.00001044, -0.00001522, 0.00000156, -0.00000733, 0.00001328, -0.00000862, 0.00000935, -0.00001226, 0.00000986, -0.00001077, 0.00000533, -0.00001401, 0.00000630, -0.00001974, 0.00000750, -0.00002337, 0.00000327, -0.00002186, 0.00000145, -0.00004534, -0.00001736, -0.00011046, -0.00032755, 0.00040449, 0.00066449, 0.00002001, -0.00034571, 0.00003591, -0.00008751, 0.00002035, -0.00004827, 0.00001855, -0.00004259, 0.00001783, -0.00002200, 0.00001262, -0.00001502, 0.00000906, -0.00001382, 0.00001198, -0.00001754, 0.00000797, -0.00000718, 0.00000227, -0.00001912, 0.00001166, -0.00000902, 0.00001038, -0.00000366, 0.00000369, -0.00002864, 0.00001525, 0.00001009, -0.00000805, -0.00000084, -0.00000822, 0.00001084, -0.00002343, 0.00000978, -0.00000071, -0.00000304, 0.00001886, -0.00001708, 0.00001566, -0.00000365, 0.00000959, 0.00000537, 0.00000182, -0.00000863, 0.00001285, -0.00001825, 0.00003339, -0.00002437, 0.00002232, -0.00002051, 0.00000516, 0.00000136, 0.00000230, -0.00001180, 0.00000291, 0.00000557, -0.00000323, -0.00000430, -0.00000166, 0.00000643, 0.00000169, -0.00001253, 0.00001066, -0.00000855, 0.00000766, -0.00000172, 0.00000963, -0.00001537, 0.00000293, },
		{ 0.00001789, -0.00000817, -0.00000107, -0.00001021, 0.00002586, -0.00001592, 0.00000459, -0.00000802, -0.00000252, 0.00000858, -0.00001660, 0.00001070, -0.00001162, 0.00000507, -0.00000176, 0.00000208, -0.00000166, -0.00000287, -0.00000287, -0.00000015, -0.00000399, 0.00000920, -0.00001701, -0.00000249, 0.00000735, -0.00001964, 0.00002157, -0.00002303, 0.00002748, -0.00001183, -0.00001155, -0.00000057, 0.00001735, 0.00000456, -0.00001070, 0.00000645, -0.00001688, 0.00000554, -0.00000154, 0.00000731, -0.00000637, 0.00000504, -0.00000847, 0.00000862, -0.00001391, 0.00000664, -0.00001504, 0.00000480, -0.00001901, 0.00000694, -0.00002872, 0.00001211, -0.00003593, 0.00000923, -0.00007757, 0.00001014, -0.00035980, 0.00006636, 0.00080265, 0.00025194, -0.00038514, -0.00003185, -0.00005206, -0.00001328, -0.00002443, -0.00000425, -0.00002079, -0.00000679, -0.00000196, -0.00000564, 0.00000030, -0.00000356, -0.00000346, 0.00000155, -0.00000587, -0.00000321, 0.00000096, -0.00000159, -0.00001255, -0.00000056, 0.00000257, -0.00000149, 0.00000551, 0.00000027, -0.00001593, -0.00001057, 0.00002883, -0.00001461, 0.00000533, -0.00001471, 0.00001257, -0.00001574, -0.00000374, 0.00001202, -0.00001475, 0.00002451, -0.00001618, 0.00001030, 0.00000011, 0.00000673, 0.00000381, 0.00000996, -0.00001676, 0.00001829, -0.00002053, 0.00002398, -0.00000610, 0.00000133, 0.00000536, -0.00001755, 0.00001503, -0.00000590, -0.00000154, -0.00000956, 0.00001376, -0.00000603, -0.00000129, -0.00000556, 0.00000578, 0.00000566, -0.00001274, 0.00000591, -0.00000179, -0.00000056, 0.00000461, 0.00000379, -0.00000357, -0.00001159, },
		{ 0.00001435, -0.00000108, -0.00000350, -0.00000983, 0.00001548, 0.00000409, -0.00001476, 0.00001011, -0.00001775, 0.00001731, -0.00001621, 0.00000502, -0.00000429, -0.00000254, 0.00000256, -0.00000042, 0.00000062, -0.00000304, -0.00000276, -0.00000038, -0.00000400, 0.00000612, -0.00000592, -0.00001584, 0.00001614, -0.00002245, 0.00001869, -0.00001440, 0.00000754, 0.00001363, -0.00002491, 0.00000292, 0.00000422, 0.00002219, -0.00002110, 0.00001362, -0.00001830, 0.00000306, -0.00000312, 0.00000994, -0.00000858, 0.00000814, -0.00001152, 0.00001155, -0.00001534, 0.00000781, -0.00001523, 0.00000582, -0.00001884, 0.00000676, -0.00002573, 0.00000705, -0.00002922, 0.00000568, -0.00005854, -0.00000676, -0.00019454, -0.00029912, 0.00073944, 0.00058728, -0.00037627, -0.00012356, -0.00002373, -0.00003863, -0.00000872, -0.00002082, -0.00000485, -0.00002391, 0.00000898, -0.00001481, 0.00000661, -0.00000826, 0.00000086, -0.00000433, 0.00000139, -0.00000988, 0.00000562, -0.00000531, -0.00000340, -0.00001352, 0.00001192, -0.00000919, 0.00001006, -0.00000286, -0.00000166, -0.00002868, 0.00003066, -0.00000571, -0.00000010, -0.00000921, 0.00000194, 0.00000032, -0.00001896, 0.00002115, -0.00001895, 0.00002100, -0.00000748, 0.00000046, 0.00000578, 0.00000297, 0.00000376, 0.00001192, -0.00001267, 0.00000761, -0.00000403, -0.00000200, 0.00002092, -0.00002086, 0.00002392, -0.00002552, 0.00001320, -0.00000229, -0.00000081, -0.00001125, 0.00000995, 0.00000119, -0.00000564, -0.00000255, -0.00000070, 0.00001157, -0.00001183, 0.00000032, 0.00000396, -0.00000545, 0.00000636, 0.00000186, 0.00000290, -0.00001628, },
		{ 0.00000357, 0.00000924, -0.00000811, -0.00000436, 0.00000061, 0.00002080, -0.00002231, 0.00001063, -0.00001190, 0.00000483, 0.00000102, -0.00001067, 0.00000754, -0.00001077, 0.00000619, -0.00000312, 0.00000373, -0.00000494, 0.00000011, -0.00000378, 0.00000007, -0.00000140, 0.00000620, -0.00002431, 0.00001655, -0.00001683, 0.00000825, 0.00000076, -0.00001251, 0.00002728, -0.00002160, -0.00000410, -0.00000058, 0.00002517, -0.00001312, 0.00000246, -0.00000349, -0.00001082, 0.00000421, 0.00000305, -0.00000075, 0.00000040, -0.00000285, 0.00000176, -0.00000304, -0.00000557, -0.00000034, -0.00000916, -0.00000156, -0.00001166, -0.00000347, -0.00001774, -0.00000094, -0.00002631, -0.00000596, -0.00006935, -0.00002796, -0.00044477, 0.00036485, 0.00097853, -0.00027326, -0.00023483, -0.00001293, -0.00005050, -0.00000556, -0.00002624, -0.00000074, -0.00002736, 0.00000574, -0.00001122, 0.00000157, -0.00000381, -0.00000185, -0.00000416, 0.00000238, -0.00001028, 0.00000496, -0.00000545, 0.00000145, -0.00001773, 0.00000966, -0.00000510, 0.00000432, 0.00000087, 0.00000314, -0.00002744, 0.00001091, 0.00001636, -0.00001287, 0.00000305, -0.00001316, 0.00001501, -0.00002729, 0.00002119, -0.00001387, 0.00001018, 0.00000461, -0.00000883, 0.00000940, 0.00000057, 0.00000460, 0.00000862, -0.00000114, -0.00000840, 0.00001342, -0.00002043, 0.00003050, -0.00002000, 0.00001600, -0.00000793, -0.00000714, 0.00001261, -0.00000875, -0.00000347, -0.00000249, 0.00001316, -0.00001191, 0.00000218, -0.00000748, 0.00001461, -0.00000759, -0.00000558, 0.00000681, -0.00000533, 0.00000304, 0.00000387, 0.00000389, -0.00001267, },
		{ -0.00000429, 0.00001265, -0.00000584, -0.00000419, -0.00000518, 0.00002051, -0.00000920, -0.00000752, 0.00000769, -0.00001429, 0.00001654, -0.00001824, 0.00000910, -0.00000904, 0.00000223, -0.00000105, 0.00000314, -0.00000388, 0.00000033, -0.00000419, 0.00000105, -0.00000450, 0.00000979, -0.00002046, 0.00000518, -0.00000264, -0.00000763, 0.00001715, -0.00002549, 0.00002662, -0.00000796, -0.00001180, -0.00000468, 0.00002171, 0.00000030, -0.00000925, 0.00000663, -0.00001496, 0.00000135, 0.00000443, -0.00000055, 0.00000008, -0.00000171, -0.00000020, 0.00000064, -0.00000876, 0.00000227, -0.00001000, -0.00000056, -0.00001114, -0.00000337, -0.00001615, -0.00000317, -0.00002329, -0.00000225, -0.00006110, -0.00001220, -0.00029598, -0.00018765, 0.00120065, 0.00004678, -0.00040091, 0.00000742, -0.00007713, 0.00000880, -0.00004039, 0.00000857, -0.00003319, 0.00000460, -0.00001113, 0.00000093, -0.00000493, 0.00000123, -0.00000861, 0.00000622, -0.00001243, 0.00000555, -0.00000621, 0.00000384, -0.00001526, 0.00000052, 0.00000343, -0.00000329, 0.00000494, 0.00000278, -0.00001520, -0.00001381, 0.00003180, -0.00001572, 0.00000673, -0.00001766, 0.00001530, -0.00002019, 0.00000816, 0.00000087, -0.00000584, 0.00001718, -0.00001561, 0.00001065, -0.00000108, 0.00000650, 0.00000287, 0.00001002, -0.00001757, 0.00001748, -0.00001837, 0.00001597, 0.00000212, -0.00000796, 0.00001713, -0.00002488, 0.00001975, -0.00000966, -0.00000058, -0.00000893, 0.00001592, -0.00000841, -0.00000114, -0.00000594, 0.00000848, 0.00000314, -0.00001328, 0.00000876, -0.00000398, 0.00000013, 0.00000395, 0.00000549, -0.00000875, },
		{ -0.00000611, 0.00000800, 0.00000254, -0.00000875, -0.00000272, 0.00000875, 0.00001107, -0.00002406, 0.00001734, -0.00001693, 0.00001170, -0.00000606, -0.00000467, 0.00000331, -0.00000807, 0.00000497, -0.00000156, 0.00000159, -0.00000437, 0.00000109, -0.00000408, 0.00000007, 0.00000289, -0.00000581, -0.00001316, 0.00001413, -0.00002278, 0.00002867, -0.00002847, 0.00001694, 0.00000400, -0.00000955, -0.00001480, 0.00002017, 0.00000659, -0.00000751, -0.00000043, -0.00000082, -0.00001625, 0.00001714, -0.00001123, 0.00001097, -0.00001250, 0.00001071, -0.00001023, 0.00000503, -0.00001444, 0.00001016, -0.00002240, 0.00001422, -0.00003047, 0.00001556, -0.00003866, 0.00001828, -0.00005023, 0.00001496, -0.00010521, -0.00003572, -0.00058422, 0.00098989, 0.00063996, -0.00059196, 0.00002608, -0.00011894, 0.00003510, -0.00006496, 0.00002605, -0.00004641, 0.00001234, -0.00002175, 0.00001135, -0.00001709, 0.00001387, -0.00002024, 0.00001514, -0.00001855, 0.00000941, -0.00000939, 0.00000643, -0.00001211, -0.00000564, 0.00000491, -0.00000196, 0.00000062, 0.00000622, -0.00000670, -0.00002404, 0.00002448, 0.00000096, -0.00000472, -0.00000652, -0.00000040, -0.00000023, -0.00001278, 0.00001920, -0.00002127, 0.00002608, -0.00001818, 0.00000970, -0.00000249, 0.00000878, -0.00000216, 0.00001559, -0.00001583, 0.00000842, -0.00000344, -0.00000595, 0.00002173, -0.00002058, 0.00002261, -0.00001837, 0.00000506, 0.00000538, -0.00000984, -0.00000261, 0.00000481, 0.00000650, -0.00001252, 0.00000302, -0.00000417, 0.00001618, -0.00001944, 0.00000894, -0.00000294, 0.00000023, 0.00000002, 0.00000951, -0.00000737, },
		{ -0.00000748, 0.00000319, 0.00000787, -0.00000918, -0.00000230, 0.00000054, 0.00001944, -0.00002064, 0.00000407, 0.00000183, -0.00000970, 0.00001499, -0.00001968, 0.00001251, -0.00001321, 0.00000653, -0.00000403, 0.00000588, -0.00000846, 0.00000602, -0.00000877, 0.00000500, -0.00000500, 0.00000732, -0.00002506, 0.00002116, -0.00002651, 0.00002670, -0.00001779, 0.00000075, 0.00001157, -0.00000147, -0.00002244, 0.00001424, 0.00001054, -0.00000015, -0.00001162, 0.00001323, -0.00002734, 0.00002027, -0.00001286, 0.00001338, -0.00001515, 0.00001364, -0.00001381, 0.00001208, -0.00002375, 0.00002104, -0.00003287, 0.00002575, -0.00004098, 0.00002766, -0.00005081, 0.00003314, -0.00006896, 0.00005031, -0.00014058, 0.00007182, -0.00056555, 0.00031794, 0.00131297, -0.00061847, -0.00006094, -0.00010842, 0.00001953, -0.00005722, 0.00001731, -0.00003921, 0.00000548, -0.00002170, 0.00001249, -0.00002053, 0.00001697, -0.00002145, 0.00001379, -0.00001549, 0.00000543, -0.00000601, 0.00000338, -0.00000633, -0.00000870, 0.00000028, 0.00000505, -0.00000721, 0.00001048, -0.00000336, -0.00001920, 0.00000170, 0.00002377, -0.00001741, 0.00000649, -0.00001658, 0.00001601, -0.00002620, 0.00002784, -0.00002543, 0.00002372, -0.00001186, 0.00000342, -0.00000051, 0.00000762, -0.00000227, 0.00001318, -0.00000585, -0.00000462, 0.00000907, -0.00001609, 0.00002081, -0.00000912, 0.00000379, 0.00000758, -0.00001980, 0.00002290, -0.00001915, 0.00000518, -0.00000769, 0.00001912, -0.00001888, 0.00000738, -0.00001138, 0.00002058, -0.00001615, 0.00000267, 0.00000092, -0.00000022, -0.00000289, 0.00001050, -0.00000373, },
		{ -0.00001258, 0.00000363, 0.00000520, -0.00000182, -0.00000811, 0.00000162, 0.00001234, -0.00000095, -0.00001929, 0.00002169, -0.00002335, 0.00002093, -0.00001624, 0.00000453, -0.00000411, -0.00000227, 0.00000092, 0.00000306, -0.00000508, 0.00000284, -0.00000466, 0.00000114, -0.00000322, 0.00000729, -0.00002007, 0.00001046, -0.00001276, 0.00000744, 0.00000671, -0.00002085, 0.00001942, -0.00000039, -0.00001351, -0.00000660, 0.00002323, -0.00000322, -0.00000646, 0.00000473, -0.00001094, -0.00000383, 0.00001024, -0.00000820, 0.00000654, -0.00000857, 0.00000849, -0.00000852, -0.00000182, -0.00000392, -0.00000310, -0.00000769, -0.00000149, -0.00001578, -0.00000072, -0.00002220, -0.00000572, -0.00002065, -0.00003445, -0.00005285, -0.00020056, -0.00047642, 0.00164058, -0.00022187, -0.00035999, 0.00001757, -0.00008731, 0.00002088, -0.00004863, 0.00001472, -0.00003993, 0.00001218, -0.00001814, 0.00000618, -0.00000919, 0.00000487, -0.00001217, 0.00000883, -0.00001661, 0.00001238, -0.00001206, 0.00000782, -0.00001624, 0.00000084, 0.00000375, -0.00000420, 0.00000362, 0.00000523, -0.00001470, -0.00001354, 0.00002854, -0.00001114, 0.00000343, -0.00001561, 0.00001263, -0.00001942, 0.00001706, -0.00001121, 0.00000565, 0.00000603, -0.00001063, 0.00000800, -0.00000038, 0.00000501, 0.00000334, 0.00000755, -0.00001424, 0.00001215, -0.00001175, 0.00000507, 0.00001144, -0.00001639, 0.00002499, -0.00002732, 0.00002024, -0.00001085, -0.00000091, -0.00000638, 0.00001537, -0.00000895, -0.00000198, -0.00000517, 0.00001057, -0.00000084, -0.00001083, 0.00000887, -0.00000437, -0.00000015, 0.00000411, 0.00000569, },
		{ -0.00001657, 0.00000478, 0.00000045, 0.00000591, -0.00001277, 0.00000378, 0.00000284, 0.00001492, -0.00002845, 0.00001928, -0.00001124, 0.00000190, 0.00000701, -0.00001603, 0.00001202, -0.00001435, 0.00000825, -0.00000385, 0.00000380, -0.00000681, 0.00000637, -0.00000954, 0.00000647, -0.00000324, -0.00000332, -0.00001046, 0.00000967, -0.00001875, 0.00003321, -0.00003862, 0.00002542, -0.00000794, 0.00000928, -0.00003466, 0.00003850, -0.00001401, 0.00001237, -0.00002029, 0.00002146, -0.00003893, 0.00004022, -0.00003597, 0.00003485, -0.00003801, 0.00003853, -0.00003866, 0.00003265, -0.00004397, 0.00004302, -0.00005833, 0.00005554, -0.00007693, 0.00006757, -0.00009678, 0.00008032, -0.00012342, 0.00010920, -0.00023027, 0.00015431, -0.00089427, 0.00126988, 0.00064863, -0.00078342, 0.00018160, -0.00022737, 0.00012080, -0.00013173, 0.00008228, -0.00009694, 0.00005805, -0.00006194, 0.00004620, -0.00004845, 0.00004243, -0.00004674, 0.00003936, -0.00004298, 0.00003370, -0.00002981, 0.00002341, -0.00002610, 0.00000861, -0.00000907, 0.00001163, -0.00001442, 0.00002032, -0.00001783, -0.00001055, 0.00000920, 0.00001399, -0.00001397, 0.00000064, -0.00000659, 0.00000164, -0.00000613, 0.00001354, -0.00001961, 0.00002724, -0.00002576, 0.00001798, -0.00001080, 0.00001478, -0.00000772, 0.00001795, -0.00001714, 0.00000849, -0.00000354, -0.00000689, 0.00001794, -0.00001454, 0.00001457, -0.00000624, -0.00000563, 0.00001338, -0.00001836, 0.00000644, -0.00000098, 0.00001101, -0.00001780, 0.00000716, -0.00000535, 0.00001674, -0.00002270, 0.00001462, -0.00000827, 0.00000498, -0.00000518, 0.00001542, },
		{ -0.00001109, -0.00000126, 0.00000234, 0.00000513, -0.00000822, -0.00000169, 0.00000265, 0.00001402, -0.00001473, -0.00000521, 0.00001641, -0.00002367, 0.00002720, -0.00002651, 0.00001576, -0.00001443, 0.00000636, -0.00000486, 0.00000797, -0.00001178, 0.00001204, -0.00001407, 0.00001048, -0.00000903, 0.00000763, -0.00002267, 0.00002133, -0.00003150, 0.00004146, -0.00003600, 0.00001714, -0.00000916, 0.00002291, -0.00004329, 0.00003252, -0.00000977, 0.00001824, -0.00002944, 0.00003225, -0.00004524, 0.00003818, -0.00003204, 0.00003165, -0.00003528, 0.00003531, -0.00003524, 0.00003303, -0.00004730, 0.00004768, -0.00006198, 0.00005894, -0.00007705, 0.00006669, -0.00009286, 0.00007731, -0.00012257, 0.00011822, -0.00022444, 0.00018161, -0.00068840, 0.00028232, 0.00162637, -0.00097161, 0.00014670, -0.00023409, 0.00011212, -0.00012730, 0.00007663, -0.00009204, 0.00005202, -0.00006149, 0.00004670, -0.00005146, 0.00004516, -0.00004764, 0.00003804, -0.00003977, 0.00002820, -0.00002431, 0.00001918, -0.00002110, 0.00000782, -0.00001602, 0.00002066, -0.00002328, 0.00002444, -0.00001540, -0.00000418, -0.00001212, 0.00003219, -0.00002187, 0.00001002, -0.00001922, 0.00001471, -0.00002058, 0.00002845, -0.00003246, 0.00003410, -0.00002692, 0.00001652, -0.00001168, 0.00001562, -0.00000950, 0.00001773, -0.00001108, 0.00000148, 0.00000100, -0.00000670, 0.00000627, 0.00000581, -0.00001107, 0.00002235, -0.00002939, 0.00002908, -0.00002599, 0.00001177, -0.00001075, 0.00002157, -0.00002224, 0.00000952, -0.00001159, 0.00002201, -0.00002121, 0.00000953, -0.00000472, 0.00000467, -0.00000800, 0.00001591, },
		{ 0.00000380, -0.00001369, 0.00001008, -0.00000240, 0.00000228, -0.00001131, 0.00000828, 0.00000362, 0.00000633, -0.00002757, 0.00003055, -0.00002697, 0.00002027, -0.00001036, -0.00000246, 0.00000239, -0.00000771, 0.00000381, 0.00000169, -0.00000472, 0.00000401, -0.00000387, -0.00000007, -0.00000011, 0.00000166, -0.00001503, 0.00001127, -0.00001964, 0.00002155, -0.00000647, -0.00001135, 0.00000782, 0.00000933, -0.00001603, -0.00000613, 0.00002112, -0.00000480, -0.00000353, 0.00000187, -0.00000613, -0.00000838, 0.00001395, -0.00001339, 0.00001062, -0.00001297, 0.00001474, -0.00001602, 0.00000434, -0.00000953, 0.00000311, -0.00001374, 0.00000651, -0.00002601, 0.00001243, -0.00003921, 0.00001071, -0.00003910, -0.00000667, -0.00009801, -0.00009857, -0.00075697, 0.00207138, -0.00051872, -0.00027394, -0.00000211, -0.00007910, 0.00001920, -0.00004579, 0.00001130, -0.00003815, 0.00001307, -0.00002034, 0.00000790, -0.00001067, 0.00000574, -0.00001265, 0.00000790, -0.00001666, 0.00001521, -0.00001462, 0.00000889, -0.00001458, -0.00000008, 0.00000379, -0.00000450, 0.00000208, 0.00000692, -0.00001372, -0.00001048, 0.00001993, -0.00000191, -0.00000365, -0.00000912, 0.00000464, -0.00001236, 0.00002000, -0.00002024, 0.00001601, -0.00000588, -0.00000333, 0.00000356, 0.00000137, 0.00000296, 0.00000446, 0.00000358, -0.00000811, 0.00000390, -0.00000266, -0.00000636, 0.00001936, -0.00002140, 0.00002686, -0.00002360, 0.00001568, -0.00000896, -0.00000263, -0.00000205, 0.00001205, -0.00000745, -0.00000396, -0.00000282, 0.00001112, -0.00000514, -0.00000612, 0.00000650, -0.00000298, -0.00000142, 0.00000525, },
		{ 0.00001712, -0.00002161, 0.00001317, -0.00000569, 0.00000715, -0.00001378, 0.00000766, -0.00000026, 0.00001393, -0.00002631, 0.00001497, -0.00000112, -0.00001043, 0.00002077, -0.00002689, 0.00002016, -0.00002030, 0.00001195, -0.00000726, 0.00000734, -0.00001026, 0.00001263, -0.00001616, 0.00001461, -0.00001201, 0.00000224, -0.00000942, 0.00000489, -0.00001190, 0.00003225, -0.00004338, 0.00003089, -0.00001929, 0.00002765, -0.00005215, 0.00005446, -0.00003471, 0.00003379, -0.00004132, 0.00004374, -0.00005999, 0.00006095, -0.00005893, 0.00005765, -0.00006279, 0.00006614, -0.00006792, 0.00006181, -0.00007483, 0.00007681, -0.00009472, 0.00009689, -0.00012335, 0.00012003, -0.00015622, 0.00014493, -0.00020000, 0.00020688, -0.00035930, 0.00034465, -0.00119202, 0.00155239, 0.00061241, -0.00092801, 0.00031720, -0.00032814, 0.00020195, -0.00019574, 0.00013671, -0.00014649, 0.00010430, -0.00010379, 0.00008300, -0.00008153, 0.00007202, -0.00007364, 0.00006357, -0.00006722, 0.00005848, -0.00005179, 0.00004253, -0.00004285, 0.00002690, -0.00002771, 0.00002918, -0.00003252, 0.00003690, -0.00003262, 0.00000939, -0.00001286, 0.00003145, -0.00002661, 0.00001177, -0.00001696, 0.00000753, -0.00000191, 0.00000697, -0.00001548, 0.00002483, -0.00002945, 0.00002394, -0.00001779, 0.00001986, -0.00001220, 0.00001870, -0.00001662, 0.00000788, -0.00000443, -0.00000488, 0.00001004, -0.00000376, 0.00000165, 0.00000842, -0.00001645, 0.00001947, -0.00002414, 0.00001383, -0.00000578, 0.00001338, -0.00002025, 0.00000898, -0.00000377, 0.00001269, -0.00002036, 0.00001564, -0.00001015, 0.00000688, -0.00000713, },
		{ 0.00001789, -0.00001585, 0.00000414, 0.00000185, 0.00000094, -0.00000518, -0.00000266, 0.00000666, 0.00000535, -0.00000577, -0.00001524, 0.00002892, -0.00003398, 0.00003506, -0.00002950, 0.00001652, -0.00001427, 0.00000590, -0.00000588, 0.00001001, -0.00001449, 0.00001741, -0.00001894, 0.00001597, -0.00001340, 0.00000776, -0.00001809, 0.00001739, -0.00003060, 0.00004934, -0.00004952, 0.00003192, -0.00002962, 0.00004590, -0.00006129, 0.00004927, -0.00003105, 0.00003784, -0.00004710, 0.00004977, -0.00006076, 0.00005422, -0.00005076, 0.00005034, -0.00005624, 0.00005833, -0.00005882, 0.00005620, -0.00007225, 0.00007513, -0.00009149, 0.00009244, -0.00011383, 0.00010730, -0.00013748, 0.00012488, -0.00017940, 0.00018835, -0.00030995, 0.00028997, -0.00080829, 0.00027421, 0.00187647, -0.00128286, 0.00034963, -0.00036426, 0.00020912, -0.00020194, 0.00014036, -0.00014932, 0.00010366, -0.00010637, 0.00008526, -0.00008580, 0.00007579, -0.00007580, 0.00006421, -0.00006624, 0.00005418, -0.00004677, 0.00003940, -0.00004045, 0.00002929, -0.00003657, 0.00003901, -0.00004120, 0.00004025, -0.00003073, 0.00001553, -0.00002893, 0.00004081, -0.00002629, 0.00001429, -0.00002227, 0.00001298, -0.00001247, 0.00002371, -0.00003375, 0.00003890, -0.00003745, 0.00002733, -0.00002150, 0.00002246, -0.00001546, 0.00002091, -0.00001541, 0.00000814, -0.00000864, 0.00000509, -0.00000996, 0.00002123, -0.00002475, 0.00003287, -0.00003318, 0.00002895, -0.00002728, 0.00001463, -0.00001048, 0.00001954, -0.00002107, 0.00000794, -0.00000751, 0.00001783, -0.00002093, 0.00001282, -0.00000796, 0.00000729, -0.00001021, },
		{ 0.00000682, 0.00000013, -0.00001155, 0.00001355, -0.00000832, 0.00000496, -0.00001243, 0.00001300, -0.00000359, 0.00001059, -0.00003016, 0.00003210, -0.00002391, 0.00001442, -0.00000144, -0.00001049, 0.00000900, -0.00001279, 0.00000674, -0.00000056, -0.00000280, 0.00000375, -0.00000192, -0.00000219, 0.00000408, -0.00000606, -0.00000593, 0.00000687, -0.00002077, 0.00003074, -0.00001822, -0.00000070, -0.00000483, 0.00001853, -0.00001779, -0.00000410, 0.00001566, -0.00000398, -0.00000137, -0.00000088, -0.00000180, -0.00001098, 0.00001419, -0.00001465, 0.00001075, -0.00001316, 0.00001644, -0.00001952, 0.00000750, -0.00001241, 0.00000643, -0.00001630, 0.00000995, -0.00003042, 0.00001869, -0.00004876, 0.00001998, -0.00005046, 0.00001243, -0.00013234, -0.00001747, -0.00098259, 0.00243518, -0.00079782, -0.00016586, -0.00004314, -0.00005555, 0.00000487, -0.00003244, -0.00000126, -0.00002786, 0.00000670, -0.00001686, 0.00000521, -0.00000865, 0.00000329, -0.00000963, 0.00000312, -0.00001208, 0.00001318, -0.00001294, 0.00000639, -0.00001002, -0.00000240, 0.00000378, -0.00000450, 0.00000087, 0.00000710, -0.00001128, -0.00000627, 0.00000835, 0.00000932, -0.00001196, -0.00000053, -0.00000659, -0.00000016, 0.00001602, -0.00002344, 0.00002196, -0.00001533, 0.00000375, -0.00000101, 0.00000302, 0.00000127, 0.00000518, -0.00000043, -0.00000116, -0.00000494, 0.00000634, -0.00001578, 0.00002364, -0.00002129, 0.00002206, -0.00001397, 0.00000655, -0.00000436, -0.00000529, 0.00000343, 0.00000653, -0.00000425, -0.00000682, 0.00000109, 0.00000940, -0.00000834, -0.00000058, 0.00000272, -0.00000063, -0.00000306, },
		{ -0.00000503, 0.00001345, -0.00002066, 0.00001672, -0.00000843, 0.00000502, -0.00001080, 0.00000778, 0.00000031, 0.00000778, -0.00001774, 0.00000514, 0.00001165, -0.00002280, 0.00003247, -0.00003481, 0.00002587, -0.00002459, 0.00001525, -0.00001130, 0.00001184, -0.00001420, 0.00001899, -0.00002280, 0.00002323, -0.00002233, 0.00001071, -0.00001129, 0.00000216, -0.00000486, 0.00002730, -0.00004285, 0.00003233, -0.00002772, 0.00004201, -0.00006326, 0.00006365, -0.00005054, 0.00005159, -0.00005841, 0.00006093, -0.00007446, 0.00007434, -0.00007495, 0.00007393, -0.00008105, 0.00008728, -0.00009153, 0.00008594, -0.00010018, 0.00010417, -0.00012349, 0.00012925, -0.00015954, 0.00016144, -0.00020428, 0.00019805, -0.00026359, 0.00028773, -0.00046613, 0.00050038, -0.00143511, 0.00180250, 0.00054074, -0.00101185, 0.00041194, -0.00040232, 0.00026280, -0.00024391, 0.00017807, -0.00018474, 0.00014141, -0.00013847, 0.00011389, -0.00010938, 0.00009638, -0.00009527, 0.00008252, -0.00008607, 0.00007858, -0.00007053, 0.00005958, -0.00005850, 0.00004540, -0.00004707, 0.00004674, -0.00004970, 0.00005200, -0.00004697, 0.00003124, -0.00003712, 0.00004943, -0.00003966, 0.00002471, -0.00003016, 0.00001753, -0.00000217, 0.00000237, -0.00001117, 0.00002021, -0.00002952, 0.00002694, -0.00002262, 0.00002330, -0.00001526, 0.00001803, -0.00001470, 0.00000692, -0.00000608, -0.00000058, -0.00000055, 0.00000955, -0.00001321, 0.00002259, -0.00002500, 0.00002190, -0.00002562, 0.00001810, -0.00000848, 0.00001293, -0.00001946, 0.00000850, -0.00000011, 0.00000497, -0.00001295, 0.00001189, -0.00000810, 0.00000523, },
		{ -0.00000836, 0.00001510, -0.00001631, 0.00000712, 0.00000238, -0.00000447, -0.00000049, -0.00000481, 0.00001163, -0.00000498, 0.00000437, -0.00002382, 0.00003679, -0.00003831, 0.00003692, -0.00002777, 0.00001460, -0.00001289, 0.00000551, -0.00000737, 0.00001209, -0.00001634, 0.00002157, -0.00002273, 0.00002095, -0.00001805, 0.00000864, -0.00001306, 0.00001110, -0.00002453, 0.00004908, -0.00005537, 0.00004103, -0.00004483, 0.00006212, -0.00007135, 0.00005948, -0.00004815, 0.00005385, -0.00006030, 0.00006181, -0.00007021, 0.00006464, -0.00006497, 0.00006517, -0.00007341, 0.00007784, -0.00007965, 0.00007670, -0.00009350, 0.00009782, -0.00011532, 0.00011951, -0.00014389, 0.00014144, -0.00017617, 0.00016672, -0.00022905, 0.00024852, -0.00038307, 0.00038192, -0.00091531, 0.00030243, 0.00202903, -0.00150668, 0.00050892, -0.00047056, 0.00028934, -0.00026445, 0.00019443, -0.00019849, 0.00014914, -0.00014651, 0.00011970, -0.00011615, 0.00010223, -0.00009984, 0.00008656, -0.00008922, 0.00007785, -0.00006841, 0.00005955, -0.00006005, 0.00005130, -0.00005703, 0.00005600, -0.00005707, 0.00005419, -0.00004562, 0.00003623, -0.00004581, 0.00004807, -0.00003021, 0.00001954, -0.00002666, 0.00001284, -0.00000480, 0.00001595, -0.00002997, 0.00003740, -0.00004167, 0.00003361, -0.00002794, 0.00002653, -0.00001896, 0.00002177, -0.00001770, 0.00001361, -0.00001746, 0.00001625, -0.00002439, 0.00003353, -0.00003388, 0.00003678, -0.00003003, 0.00002206, -0.00002263, 0.00001342, -0.00000690, 0.00001339, -0.00001583, 0.00000331, -0.00000012, 0.00000873, -0.00001513, 0.00001165, -0.00000767, 0.00000646, },
		{ -0.00000403, 0.00000779, -0.00000422, -0.00000671, 0.00001354, -0.00001146, 0.00000533, -0.00001070, 0.00001467, -0.00000796, 0.00001086, -0.00002639, 0.00002610, -0.00001473, 0.00000472, 0.00000866, -0.00001757, 0.00001392, -0.00001597, 0.00000851, -0.00000278, 0.00000004, 0.00000253, 0.00000056, -0.00000429, 0.00000793, -0.00001384, 0.00000521, -0.00000173, -0.00001549, 0.00003238, -0.00002487, 0.00000730, -0.00001487, 0.00002432, -0.00001692, -0.00000197, 0.00000808, -0.00000122, -0.00000028, -0.00000285, 0.00000129, -0.00001107, 0.00001069, -0.00001177, 0.00000673, -0.00000881, 0.00001295, -0.00001806, 0.00000661, -0.00001155, 0.00000579, -0.00001428, 0.00000760, -0.00002733, 0.00001570, -0.00004777, 0.00001843, -0.00005002, 0.00001581, -0.00014395, 0.00002022, -0.00111301, 0.00267866, -0.00101210, -0.00006450, -0.00009071, -0.00002502, -0.00001660, -0.00001294, -0.00001919, -0.00001206, -0.00000488, -0.00000896, -0.00000100, -0.00000386, -0.00000172, -0.00000401, -0.00000438, -0.00000396, 0.00000699, -0.00000751, 0.00000091, -0.00000322, -0.00000568, 0.00000364, -0.00000436, 0.00000026, 0.00000547, -0.00000666, -0.00000276, -0.00000320, 0.00001916, -0.00001822, 0.00000686, -0.00001753, 0.00001405, 0.00000650, -0.00002014, 0.00002169, -0.00002000, 0.00000847, -0.00000420, 0.00000358, 0.00000063, 0.00000472, -0.00000316, 0.00000465, -0.00001209, 0.00001289, -0.00002117, 0.00002301, -0.00001583, 0.00001166, -0.00000030, -0.00000536, 0.00000173, -0.00000791, 0.00000894, -0.00000006, -0.00000024, -0.00000979, 0.00000595, 0.00000528, -0.00000935, 0.00000421, -0.00000101, 0.00000149, },
		{ 0.00000089, 0.00000033, 0.00000513, -0.00001378, 0.00001470, -0.00000708, -0.00000052, -0.00000412, 0.00000433, 0.00000404, -0.00000220, -0.00000461, -0.00000790, 0.00002423, -0.00003232, 0.00003951, -0.00003766, 0.00002752, -0.00002596, 0.00001714, -0.00001507, 0.00001647, -0.00001767, 0.00002423, -0.00002794, 0.00003051, -0.00003228, 0.00002090, -0.00001627, 0.00000317, -0.00000037, 0.00002088, -0.00003848, 0.00003002, -0.00003211, 0.00005023, -0.00006563, 0.00006340, -0.00005797, 0.00006170, -0.00006752, 0.00006925, -0.00007892, 0.00007713, -0.00008051, 0.00007990, -0.00008853, 0.00009710, -0.00010414, 0.00009948, -0.00011422, 0.00011895, -0.00013806, 0.00014532, -0.00017717, 0.00018227, -0.00023012, 0.00022740, -0.00029972, 0.00033378, -0.00052750, 0.00059020, -0.00158410, 0.00198356, 0.00045130, -0.00103194, 0.00045379, -0.00043734, 0.00029223, -0.00026679, 0.00019806, -0.00020381, 0.00016167, -0.00015862, 0.00013216, -0.00012599, 0.00011021, -0.00010691, 0.00009191, -0.00009531, 0.00008962, -0.00008170, 0.00007048, -0.00006904, 0.00005981, -0.00006273, 0.00005999, -0.00006179, 0.00006147, -0.00005630, 0.00004958, -0.00005794, 0.00006306, -0.00004917, 0.00003607, -0.00004331, 0.00002997, -0.00000727, 0.00000164, -0.00000881, 0.00001513, -0.00002689, 0.00002709, -0.00002496, 0.00002473, -0.00001688, 0.00001638, -0.00001209, 0.00000601, -0.00000830, 0.00000496, -0.00001178, 0.00002255, -0.00002665, 0.00003331, -0.00002946, 0.00001983, -0.00002231, 0.00001872, -0.00000880, 0.00000982, -0.00001584, 0.00000648, 0.00000409, -0.00000419, -0.00000261, 0.00000491, -0.00000320, },
		{ 0.00000244, -0.00000273, 0.00000769, -0.00001205, 0.00000696, 0.00000410, -0.00001043, 0.00000493, -0.00000724, 0.00001616, -0.00001489, 0.00001356, -0.00002898, 0.00003838, -0.00003568, 0.00003240, -0.00002148, 0.00001043, -0.00001073, 0.00000550, -0.00000948, 0.00001427, -0.00001719, 0.00002410, -0.00002489, 0.00002487, -0.00002274, 0.00001104, -0.00000951, 0.00000502, -0.00001568, 0.00004168, -0.00005277, 0.00004231, -0.00005155, 0.00006805, -0.00007030, 0.00005979, -0.00005709, 0.00006234, -0.00006567, 0.00006551, -0.00007104, 0.00006674, -0.00007151, 0.00007251, -0.00008270, 0.00008925, -0.00009291, 0.00008971, -0.00010614, 0.00011060, -0.00012794, 0.00013400, -0.00016028, 0.00016133, -0.00020035, 0.00019346, -0.00026086, 0.00028629, -0.00042931, 0.00044119, -0.00099520, 0.00036545, 0.00206489, -0.00161111, 0.00059413, -0.00052929, 0.00033439, -0.00029996, 0.00022604, -0.00022789, 0.00017772, -0.00017231, 0.00014166, -0.00013524, 0.00011805, -0.00011393, 0.00009952, -0.00010311, 0.00009353, -0.00008375, 0.00007446, -0.00007484, 0.00006870, -0.00007249, 0.00006720, -0.00006682, 0.00006231, -0.00005584, 0.00005351, -0.00005915, 0.00005188, -0.00003259, 0.00002522, -0.00003243, 0.00001552, -0.00000027, 0.00000822, -0.00002325, 0.00003066, -0.00003948, 0.00003449, -0.00003002, 0.00002705, -0.00001957, 0.00002012, -0.00001748, 0.00001673, -0.00002358, 0.00002423, -0.00003415, 0.00004002, -0.00003649, 0.00003349, -0.00002077, 0.00000987, -0.00001338, 0.00000919, -0.00000127, 0.00000472, -0.00000820, -0.00000250, 0.00000826, -0.00000291, -0.00000546, 0.00000677, -0.00000418, },
		{ 0.00000233, -0.00000344, 0.00000710, -0.00000779, -0.00000055, 0.00001059, -0.00001171, 0.00000335, -0.00000627, 0.00001309, -0.00000920, 0.00000744, -0.00001733, 0.00001426, -0.00000176, -0.00000633, 0.00001755, -0.00002162, 0.00001551, -0.00001593, 0.00000800, -0.00000390, 0.00000285, 0.00000121, 0.00000267, -0.00000525, 0.00000980, -0.00001919, 0.00001543, -0.00001188, -0.00000531, 0.00002614, -0.00002449, 0.00001000, -0.00001961, 0.00002466, -0.00001209, -0.00000125, 0.00000010, 0.00000237, 0.00000006, -0.00000393, 0.00000298, -0.00000898, 0.00000432, -0.00000590, -0.00000012, -0.00000127, 0.00000535, -0.00001232, 0.00000218, -0.00000742, 0.00000175, -0.00000844, 0.00000038, -0.00001771, 0.00000422, -0.00003665, 0.00000615, -0.00003734, 0.00000182, -0.00012849, 0.00000336, -0.00112376, 0.00276433, -0.00112376, 0.00000336, -0.00012849, 0.00000182, -0.00003734, 0.00000615, -0.00003665, 0.00000422, -0.00001771, 0.00000038, -0.00000844, 0.00000175, -0.00000742, 0.00000218, -0.00001232, 0.00000535, -0.00000127, -0.00000012, -0.00000590, 0.00000432, -0.00000898, 0.00000298, -0.00000393, 0.00000006, 0.00000237, 0.00000010, -0.00000125, -0.00001209, 0.00002466, -0.00001961, 0.00001000, -0.00002449, 0.00002614, -0.00000531, -0.00001188, 0.00001543, -0.00001919, 0.00000980, -0.00000525, 0.00000267, 0.00000121, 0.00000285, -0.00000390, 0.00000800, -0.00001593, 0.00001551, -0.00002162, 0.00001755, -0.00000633, -0.00000176, 0.00001426, -0.00001733, 0.00000744, -0.00000920, 0.00001309, -0.00000627, 0.00000335, -0.00001171, 0.00001059, -0.00000055, -0.00000779, 0.00000710, -0.00000344, },
		{ 0.00000244, -0.00000418, 0.00000677, -0.00000546, -0.00000291, 0.00000826, -0.00000250, -0.00000820, 0.00000472, -0.00000127, 0.00000919, -0.00001338, 0.00000987, -0.00002077, 0.00003349, -0.00003649, 0.00004002, -0.00003415, 0.00002423, -0.00002358, 0.00001673, -0.00001748, 0.00002012, -0.00001957, 0.00002705, -0.00003002, 0.00003449, -0.00003948, 0.00003066, -0.00002325, 0.00000822, -0.00000027, 0.00001552, -0.00003243, 0.00002522, -0.00003259, 0.00005188, -0.00005915, 0.00005351, -0.00005584, 0.00006231, -0.00006682, 0.00006720, -0.00007249, 0.00006870, -0.00007484, 0.00007446, -0.00008375, 0.00009353, -0.00010311, 0.00009952, -0.00011393, 0.00011805, -0.00013524, 0.00014166, -0.00017231, 0.00017772, -0.00022789, 0.00022604, -0.00029996, 0.00033439, -0.00052929, 0.00059413, -0.00161111, 0.00206489, 0.00036545, -0.00099520, 0.00044119, -0.00042931, 0.00028629, -0.00026086, 0.00019346, -0.00020035, 0.00016133, -0.00016028, 0.00013400, -0.00012794, 0.00011060, -0.00010614, 0.00008971, -0.00009291, 0.00008925, -0.00008270, 0.00007251, -0.00007151, 0.00006674, -0.00007104, 0.00006551, -0.00006567, 0.00006234, -0.00005709, 0.00005979, -0.00007030, 0.00006805, -0.00005155, 0.00004231, -0.00005277, 0.00004168, -0.00001568, 0.00000502, -0.00000951, 0.00001104, -0.00002274, 0.00002487, -0.00002489, 0.00002410, -0.00001719, 0.00001427, -0.00000948, 0.00000550, -0.00001073, 0.00001043, -0.00002148, 0.00003240, -0.00003568, 0.00003838, -0.00002898, 0.00001356, -0.00001489, 0.00001616, -0.00000724, 0.00000493, -0.00001043, 0.00000410, 0.00000696, -0.00001205, 0.00000769, -0.00000273, },
		{ 0.00000089, -0.00000320, 0.00000491, -0.00000261, -0.00000419, 0.00000409, 0.00000648, -0.00001584, 0.00000982, -0.00000880, 0.00001872, -0.00002231, 0.00001983, -0.00002946, 0.00003331, -0.00002665, 0.00002255, -0.00001178, 0.00000496, -0.00000830, 0.00000601, -0.00001209, 0.00001638, -0.00001688, 0.00002473, -0.00002496, 0.00002709, -0.00002689, 0.00001513, -0.00000881, 0.00000164, -0.00000727, 0.00002997, -0.00004331, 0.00003607, -0.00004917, 0.00006306, -0.00005794, 0.00004958, -0.00005630, 0.00006147, -0.00006179, 0.00005999, -0.00006273, 0.00005981, -0.00006904, 0.00007048, -0.00008170, 0.00008962, -0.00009531, 0.00009191, -0.00010691, 0.00011021, -0.00012599, 0.00013216, -0.00015862, 0.00016167, -0.00020381, 0.00019806, -0.00026679, 0.00029223, -0.00043734, 0.00045379, -0.00103194, 0.00045130, 0.00198356, -0.00158410, 0.00059020, -0.00052750, 0.00033378, -0.00029972, 0.00022740, -0.00023012, 0.00018227, -0.00017717, 0.00014532, -0.00013806, 0.00011895, -0.00011422, 0.00009948, -0.00010414, 0.00009710, -0.00008853, 0.00007990, -0.00008051, 0.00007713, -0.00007892, 0.00006925, -0.00006752, 0.00006170, -0.00005797, 0.00006340, -0.00006563, 0.00005023, -0.00003211, 0.00003002, -0.00003848, 0.00002088, -0.00000037, 0.00000317, -0.00001627, 0.00002090, -0.00003228, 0.00003051, -0.00002794, 0.00002423, -0.00001767, 0.00001647, -0.00001507, 0.00001714, -0.00002596, 0.00002752, -0.00003766, 0.00003951, -0.00003232, 0.00002423, -0.00000790, -0.00000461, -0.00000220, 0.00000404, 0.00000433, -0.00000412, -0.00000052, -0.00000708, 0.00001470, -0.00001378, 0.00000513, 0.00000033, },
		{ -0.00000403, 0.00000149, -0.00000101, 0.00000421, -0.00000935, 0.00000528, 0.00000595, -0.00000979, -0.00000024, -0.00000006, 0.00000894, -0.00000791, 0.00000173, -0.00000536, -0.00000030, 0.00001166, -0.00001583, 0.00002301, -0.00002117, 0.00001289, -0.00001209, 0.00000465, -0.00000316, 0.00000472, 0.00000063, 0.00000358, -0.00000420, 0.00000847, -0.00002000, 0.00002169, -0.00002014, 0.00000650, 0.00001405, -0.00001753, 0.00000686, -0.00001822, 0.00001916, -0.00000320, -0.00000276, -0.00000666, 0.00000547, 0.00000026, -0.00000436, 0.00000364, -0.00000568, -0.00000322, 0.00000091, -0.00000751, 0.00000699, -0.00000396, -0.00000438, -0.00000401, -0.00000172, -0.00000386, -0.00000100, -0.00000896, -0.00000488, -0.00001206, -0.00001919, -0.00001294, -0.00001660, -0.00002502, -0.00009071, -0.00006450, -0.00101210, 0.00267866, -0.00111301, 0.00002022, -0.00014395, 0.00001581, -0.00005002, 0.00001843, -0.00004777, 0.00001570, -0.00002733, 0.00000760, -0.00001428, 0.00000579, -0.00001155, 0.00000661, -0.00001806, 0.00001295, -0.00000881, 0.00000673, -0.00001177, 0.00001069, -0.00001107, 0.00000129, -0.00000285, -0.00000028, -0.00000122, 0.00000808, -0.00000197, -0.00001692, 0.00002432, -0.00001487, 0.00000730, -0.00002487, 0.00003238, -0.00001549, -0.00000173, 0.00000521, -0.00001384, 0.00000793, -0.00000429, 0.00000056, 0.00000253, 0.00000004, -0.00000278, 0.00000851, -0.00001597, 0.00001392, -0.00001757, 0.00000866, 0.00000472, -0.00001473, 0.00002610, -0.00002639, 0.00001086, -0.00000796, 0.00001467, -0.00001070, 0.00000533, -0.00001146, 0.00001354, -0.00000671, -0.00000422, 0.00000779, },
		{ -0.00000836, 0.00000646, -0.00000767, 0.00001165, -0.00001513, 0.00000873, -0.00000012, 0.00000331, -0.00001583, 0.00001339, -0.00000690, 0.00001342, -0.00002263, 0.00002206, -0.00003003, 0.00003678, -0.00003388, 0.00003353, -0.00002439, 0.00001625, -0.00001746, 0.00001361, -0.00001770, 0.00002177, -0.00001896, 0.00002653, -0.00002794, 0.00003361, -0.00004167, 0.00003740, -0.00002997, 0.00001595, -0.00000480, 0.00001284, -0.00002666, 0.00001954, -0.00003021, 0.00004807, -0.00004581, 0.00003623, -0.00004562, 0.00005419, -0.00005707, 0.00005600, -0.00005703, 0.00005130, -0.00006005, 0.00005955, -0.00006841, 0.00007785, -0.00008922, 0.00008656, -0.00009984, 0.00010223, -0.00011615, 0.00011970, -0.00014651, 0.00014914, -0.00019849, 0.00019443, -0.00026445, 0.00028934, -0.00047056, 0.00050892, -0.00150668, 0.00202903, 0.00030243, -0.00091531, 0.00038192, -0.00038307, 0.00024852, -0.00022905, 0.00016672, -0.00017617, 0.00014144, -0.00014389, 0.00011951, -0.00011532, 0.00009782, -0.00009350, 0.00007670, -0.00007965, 0.00007784, -0.00007341, 0.00006517, -0.00006497, 0.00006464, -0.00007021, 0.00006181, -0.00006030, 0.00005385, -0.00004815, 0.00005948, -0.00007135, 0.00006212, -0.00004483, 0.00004103, -0.00005537, 0.00004908, -0.00002453, 0.00001110, -0.00001306, 0.00000864, -0.00001805, 0.00002095, -0.00002273, 0.00002157, -0.00001634, 0.00001209, -0.00000737, 0.00000551, -0.00001289, 0.00001460, -0.00002777, 0.00003692, -0.00003831, 0.00003679, -0.00002382, 0.00000437, -0.00000498, 0.00001163, -0.00000481, -0.00000049, -0.00000447, 0.00000238, 0.00000712, -0.00001631, 0.00001510, },
		{ -0.00000503, 0.00000523, -0.00000810, 0.00001189, -0.00001295, 0.00000497, -0.00000011, 0.00000850, -0.00001946, 0.00001293, -0.00000848, 0.00001810, -0.00002562, 0.00002190, -0.00002500, 0.00002259, -0.00001321, 0.00000955, -0.00000055, -0.00000058, -0.00000608, 0.00000692, -0.00001470, 0.00001803, -0.00001526, 0.00002330, -0.00002262, 0.00002694, -0.00002952, 0.00002021, -0.00001117, 0.00000237, -0.00000217, 0.00001753, -0.00003016, 0.00002471, -0.00003966, 0.00004943, -0.00003712, 0.00003124, -0.00004697, 0.00005200, -0.00004970, 0.00004674, -0.00004707, 0.00004540, -0.00005850, 0.00005958, -0.00007053, 0.00007858, -0.00008607, 0.00008252, -0.00009527, 0.00009638, -0.00010938, 0.00011389, -0.00013847, 0.00014141, -0.00018474, 0.00017807, -0.00024391, 0.00026280, -0.00040232, 0.00041194, -0.00101185, 0.00054074, 0.00180250, -0.00143511, 0.00050038, -0.00046613, 0.00028773, -0.00026359, 0.00019805, -0.00020428, 0.00016144, -0.00015954, 0.00012925, -0.00012349, 0.00010417, -0.00010018, 0.00008594, -0.00009153, 0.00008728, -0.00008105, 0.00007393, -0.00007495, 0.00007434, -0.00007446, 0.00006093, -0.00005841, 0.00005159, -0.00005054, 0.00006365, -0.00006326, 0.00004201, -0.00002772, 0.00003233, -0.00004285, 0.00002730, -0.00000486, 0.00000216, -0.00001129, 0.00001071, -0.00002233, 0.00002323, -0.00002280, 0.00001899, -0.00001420, 0.00001184, -0.00001130, 0.00001525, -0.00002459, 0.00002587, -0.00003481, 0.00003247, -0.00002280, 0.00001165, 0.00000514, -0.00001774, 0.00000778, 0.00000031, 0.00000778, -0.00001080, 0.00000502, -0.00000843, 0.00001672, -0.00002066, 0.00001345, },
		{ 0.00000682, -0.00000306, -0.00000063, 0.00000272, -0.00000058, -0.00000834, 0.00000940, 0.00000109, -0.00000682, -0.00000425, 0.00000653, 0.00000343, -0.00000529, -0.00000436, 0.00000655, -0.00001397, 0.00002206, -0.00002129, 0.00002364, -0.00001578, 0.00000634, -0.00000494, -0.00000116, -0.00000043, 0.00000518, 0.00000127, 0.00000302, -0.00000101, 0.00000375, -0.00001533, 0.00002196, -0.00002344, 0.00001602, -0.00000016, -0.00000659, -0.00000053, -0.00001196, 0.00000932, 0.00000835, -0.00000627, -0.00001128, 0.00000710, 0.00000087, -0.00000450, 0.00000378, -0.00000240, -0.00001002, 0.00000639, -0.00001294, 0.00001318, -0.00001208, 0.00000312, -0.00000963, 0.00000329, -0.00000865, 0.00000521, -0.00001686, 0.00000670, -0.00002786, -0.00000126, -0.00003244, 0.00000487, -0.00005555, -0.00004314, -0.00016586, -0.00079782, 0.00243518, -0.00098259, -0.00001747, -0.00013234, 0.00001243, -0.00005046, 0.00001998, -0.00004876, 0.00001869, -0.00003042, 0.00000995, -0.00001630, 0.00000643, -0.00001241, 0.00000750, -0.00001952, 0.00001644, -0.00001316, 0.00001075, -0.00001465, 0.00001419, -0.00001098, -0.00000180, -0.00000088, -0.00000137, -0.00000398, 0.00001566, -0.00000410, -0.00001779, 0.00001853, -0.00000483, -0.00000070, -0.00001822, 0.00003074, -0.00002077, 0.00000687, -0.00000593, -0.00000606, 0.00000408, -0.00000219, -0.00000192, 0.00000375, -0.00000280, -0.00000056, 0.00000674, -0.00001279, 0.00000900, -0.00001049, -0.00000144, 0.00001442, -0.00002391, 0.00003210, -0.00003016, 0.00001059, -0.00000359, 0.00001300, -0.00001243, 0.00000496, -0.00000832, 0.00001355, -0.00001155, 0.00000013, },
		{ 0.00001789, -0.00001021, 0.00000729, -0.00000796, 0.00001282, -0.00002093, 0.00001783, -0.00000751, 0.00000794, -0.00002107, 0.00001954, -0.00001048, 0.00001463, -0.00002728, 0.00002895, -0.00003318, 0.00003287, -0.00002475, 0.00002123, -0.00000996, 0.00000509, -0.00000864, 0.00000814, -0.00001541, 0.00002091, -0.00001546, 0.00002246, -0.00002150, 0.00002733, -0.00003745, 0.00003890, -0.00003375, 0.00002371, -0.00001247, 0.00001298, -0.00002227, 0.00001429, -0.00002629, 0.00004081, -0.00002893, 0.00001553, -0.00003073, 0.00004025, -0.00004120, 0.00003901, -0.00003657, 0.00002929, -0.00004045, 0.00003940, -0.00004677, 0.00005418, -0.00006624, 0.00006421, -0.00007580, 0.00007579, -0.00008580, 0.00008526, -0.00010637, 0.00010366, -0.00014932, 0.00014036, -0.00020194, 0.00020912, -0.00036426, 0.00034963, -0.00128286, 0.00187647, 0.00027421, -0.00080829, 0.00028997, -0.00030995, 0.00018835, -0.00017940, 0.00012488, -0.00013748, 0.00010730, -0.00011383, 0.00009244, -0.00009149, 0.00007513, -0.00007225, 0.00005620, -0.00005882, 0.00005833, -0.00005624, 0.00005034, -0.00005076, 0.00005422, -0.00006076, 0.00004977, -0.00004710, 0.00003784, -0.00003105, 0.00004927, -0.00006129, 0.00004590, -0.00002962, 0.00003192, -0.00004952, 0.00004934, -0.00003060, 0.00001739, -0.00001809, 0.00000776, -0.00001340, 0.00001597, -0.00001894, 0.00001741, -0.00001449, 0.00001001, -0.00000588, 0.00000590, -0.00001427, 0.00001652, -0.00002950, 0.00003506, -0.00003398, 0.00002892, -0.00001524, -0.00000577, 0.00000535, 0.00000666, -0.00000266, -0.00000518, 0.00000094, 0.00000185, 0.00000414, -0.00001585, },
		{ 0.00001712, -0.00000713, 0.00000688, -0.00001015, 0.00001564, -0.00002036, 0.00001269, -0.00000377, 0.00000898, -0.00002025, 0.00001338, -0.00000578, 0.00001383, -0.00002414, 0.00001947, -0.00001645, 0.00000842, 0.00000165, -0.00000376, 0.00001004, -0.00000488, -0.00000443, 0.00000788, -0.00001662, 0.00001870, -0.00001220, 0.00001986, -0.00001779, 0.00002394, -0.00002945, 0.00002483, -0.00001548, 0.00000697, -0.00000191, 0.00000753, -0.00001696, 0.00001177, -0.00002661, 0.00003145, -0.00001286, 0.00000939, -0.00003262, 0.00003690, -0.00003252, 0.00002918, -0.00002771, 0.00002690, -0.00004285, 0.00004253, -0.00005179, 0.00005848, -0.00006722, 0.00006357, -0.00007364, 0.00007202, -0.00008153, 0.00008300, -0.00010379, 0.00010430, -0.00014649, 0.00013671, -0.00019574, 0.00020195, -0.00032814, 0.00031720, -0.00092801, 0.00061241, 0.00155239, -0.00119202, 0.00034465, -0.00035930, 0.00020688, -0.00020000, 0.00014493, -0.00015622, 0.00012003, -0.00012335, 0.00009689, -0.00009472, 0.00007681, -0.00007483, 0.00006181, -0.00006792, 0.00006614, -0.00006279, 0.00005765, -0.00005893, 0.00006095, -0.00005999, 0.00004374, -0.00004132, 0.00003379, -0.00003471, 0.00005446, -0.00005215, 0.00002765, -0.00001929, 0.00003089, -0.00004338, 0.00003225, -0.00001190, 0.00000489, -0.00000942, 0.00000224, -0.00001201, 0.00001461, -0.00001616, 0.00001263, -0.00001026, 0.00000734, -0.00000726, 0.00001195, -0.00002030, 0.00002016, -0.00002689, 0.00002077, -0.00001043, -0.00000112, 0.00001497, -0.00002631, 0.00001393, -0.00000026, 0.00000766, -0.00001378, 0.00000715, -0.00000569, 0.00001317, -0.00002161, },
		{ 0.00000380, 0.00000525, -0.00000142, -0.00000298, 0.00000650, -0.00000612, -0.00000514, 0.00001112, -0.00000282, -0.00000396, -0.00000745, 0.00001205, -0.00000205, -0.00000263, -0.00000896, 0.00001568, -0.00002360, 0.00002686, -0.00002140, 0.00001936, -0.00000636, -0.00000266, 0.00000390, -0.00000811, 0.00000358, 0.00000446, 0.00000296, 0.00000137, 0.00000356, -0.00000333, -0.00000588, 0.00001601, -0.00002024, 0.00002000, -0.00001236, 0.00000464, -0.00000912, -0.00000365, -0.00000191, 0.00001993, -0.00001048, -0.00001372, 0.00000692, 0.00000208, -0.00000450, 0.00000379, -0.00000008, -0.00001458, 0.00000889, -0.00001462, 0.00001521, -0.00001666, 0.00000790, -0.00001265, 0.00000574, -0.00001067, 0.00000790, -0.00002034, 0.00001307, -0.00003815, 0.00001130, -0.00004579, 0.00001920, -0.00007910, -0.00000211, -0.00027394, -0.00051872, 0.00207138, -0.00075697, -0.00009857, -0.00009801, -0.00000667, -0.00003910, 0.00001071, -0.00003921, 0.00001243, -0.00002601, 0.00000651, -0.00001374, 0.00000311, -0.00000953, 0.00000434, -0.00001602, 0.00001474, -0.00001297, 0.00001062, -0.00001339, 0.00001395, -0.00000838, -0.00000613, 0.00000187, -0.00000353, -0.00000480, 0.00002112, -0.00000613, -0.00001603, 0.00000933, 0.00000782, -0.00001135, -0.00000647, 0.00002155, -0.00001964, 0.00001127, -0.00001503, 0.00000166, -0.00000011, -0.00000007, -0.00000387, 0.00000401, -0.00000472, 0.00000169, 0.00000381, -0.00000771, 0.00000239, -0.00000246, -0.00001036, 0.00002027, -0.00002697, 0.00003055, -0.00002757, 0.00000633, 0.00000362, 0.00000828, -0.00001131, 0.00000228, -0.00000240, 0.00001008, -0.00001369, },
		{ -0.00001109, 0.00001591, -0.00000800, 0.00000467, -0.00000472, 0.00000953, -0.00002121, 0.00002201, -0.00001159, 0.00000952, -0.00002224, 0.00002157, -0.00001075, 0.00001177, -0.00002599, 0.00002908, -0.00002939, 0.00002235, -0.00001107, 0.00000581, 0.00000627, -0.00000670, 0.00000100, 0.00000148, -0.00001108, 0.00001773, -0.00000950, 0.00001562, -0.00001168, 0.00001652, -0.00002692, 0.00003410, -0.00003246, 0.00002845, -0.00002058, 0.00001471, -0.00001922, 0.00001002, -0.00002187, 0.00003219, -0.00001212, -0.00000418, -0.00001540, 0.00002444, -0.00002328, 0.00002066, -0.00001602, 0.00000782, -0.00002110, 0.00001918, -0.00002431, 0.00002820, -0.00003977, 0.00003804, -0.00004764, 0.00004516, -0.00005146, 0.00004670, -0.00006149, 0.00005202, -0.00009204, 0.00007663, -0.00012730, 0.00011212, -0.00023409, 0.00014670, -0.00097161, 0.00162637, 0.00028232, -0.00068840, 0.00018161, -0.00022444, 0.00011822, -0.00012257, 0.00007731, -0.00009286, 0.00006669, -0.00007705, 0.00005894, -0.00006198, 0.00004768, -0.00004730, 0.00003303, -0.00003524, 0.00003531, -0.00003528, 0.00003165, -0.00003204, 0.00003818, -0.00004524, 0.00003225, -0.00002944, 0.00001824, -0.00000977, 0.00003252, -0.00004329, 0.00002291, -0.00000916, 0.00001714, -0.00003600, 0.00004146, -0.00003150, 0.00002133, -0.00002267, 0.00000763, -0.00000903, 0.00001048, -0.00001407, 0.00001204, -0.00001178, 0.00000797, -0.00000486, 0.00000636, -0.00001443, 0.00001576, -0.00002651, 0.00002720, -0.00002367, 0.00001641, -0.00000521, -0.00001473, 0.00001402, 0.00000265, -0.00000169, -0.00000822, 0.00000513, 0.00000234, -0.00000126, },
		{ -0.00001657, 0.00001542, -0.00000518, 0.00000498, -0.00000827, 0.00001462, -0.00002270, 0.00001674, -0.00000535, 0.00000716, -0.00001780, 0.00001101, -0.00000098, 0.00000644, -0.00001836, 0.00001338, -0.00000563, -0.00000624, 0.00001457, -0.00001454, 0.00001794, -0.00000689, -0.00000354, 0.00000849, -0.00001714, 0.00001795, -0.00000772, 0.00001478, -0.00001080, 0.00001798, -0.00002576, 0.00002724, -0.00001961, 0.00001354, -0.00000613, 0.00000164, -0.00000659, 0.00000064, -0.00001397, 0.00001399, 0.00000920, -0.00001055, -0.00001783, 0.00002032, -0.00001442, 0.00001163, -0.00000907, 0.00000861, -0.00002610, 0.00002341, -0.00002981, 0.00003370, -0.00004298, 0.00003936, -0.00004674, 0.00004243, -0.00004845, 0.00004620, -0.00006194, 0.00005805, -0.00009694, 0.00008228, -0.00013173, 0.00012080, -0.00022737, 0.00018160, -0.00078342, 0.00064863, 0.00126988, -0.00089427, 0.00015431, -0.00023027, 0.00010920, -0.00012342, 0.00008032, -0.00009678, 0.00006757, -0.00007693, 0.00005554, -0.00005833, 0.00004302, -0.00004397, 0.00003265, -0.00003866, 0.00003853, -0.00003801, 0.00003485, -0.00003597, 0.00004022, -0.00003893, 0.00002146, -0.00002029, 0.00001237, -0.00001401, 0.00003850, -0.00003466, 0.00000928, -0.00000794, 0.00002542, -0.00003862, 0.00003321, -0.00001875, 0.00000967, -0.00001046, -0.00000332, -0.00000324, 0.00000647, -0.00000954, 0.00000637, -0.00000681, 0.00000380, -0.00000385, 0.00000825, -0.00001435, 0.00001202, -0.00001603, 0.00000701, 0.00000190, -0.00001124, 0.00001928, -0.00002845, 0.00001492, 0.00000284, 0.00000378, -0.00001277, 0.00000591, 0.00000045, 0.00000478, },
		{ -0.00001258, 0.00000569, 0.00000411, -0.00000015, -0.00000437, 0.00000887, -0.00001083, -0.00000084, 0.00001057, -0.00000517, -0.00000198, -0.00000895, 0.00001537, -0.00000638, -0.00000091, -0.00001085, 0.00002024, -0.00002732, 0.00002499, -0.00001639, 0.00001144, 0.00000507, -0.00001175, 0.00001215, -0.00001424, 0.00000755, 0.00000334, 0.00000501, -0.00000038, 0.00000800, -0.00001063, 0.00000603, 0.00000565, -0.00001121, 0.00001706, -0.00001942, 0.00001263, -0.00001561, 0.00000343, -0.00001114, 0.00002854, -0.00001354, -0.00001470, 0.00000523, 0.00000362, -0.00000420, 0.00000375, 0.00000084, -0.00001624, 0.00000782, -0.00001206, 0.00001238, -0.00001661, 0.00000883, -0.00001217, 0.00000487, -0.00000919, 0.00000618, -0.00001814, 0.00001218, -0.00003993, 0.00001472, -0.00004863, 0.00002088, -0.00008731, 0.00001757, -0.00035999, -0.00022187, 0.00164058, -0.00047642, -0.00020056, -0.00005285, -0.00003445, -0.00002065, -0.00000572, -0.00002220, -0.00000072, -0.00001578, -0.00000149, -0.00000769, -0.00000310, -0.00000392, -0.00000182, -0.00000852, 0.00000849, -0.00000857, 0.00000654, -0.00000820, 0.00001024, -0.00000383, -0.00001094, 0.00000473, -0.00000646, -0.00000322, 0.00002323, -0.00000660, -0.00001351, -0.00000039, 0.00001942, -0.00002085, 0.00000671, 0.00000744, -0.00001276, 0.00001046, -0.00002007, 0.00000729, -0.00000322, 0.00000114, -0.00000466, 0.00000284, -0.00000508, 0.00000306, 0.00000092, -0.00000227, -0.00000411, 0.00000453, -0.00001624, 0.00002093, -0.00002335, 0.00002169, -0.00001929, -0.00000095, 0.00001234, 0.00000162, -0.00000811, -0.00000182, 0.00000520, 0.00000363, },
		{ -0.00000748, -0.00000373, 0.00001050, -0.00000289, -0.00000022, 0.00000092, 0.00000267, -0.00001615, 0.00002058, -0.00001138, 0.00000738, -0.00001888, 0.00001912, -0.00000769, 0.00000518, -0.00001915, 0.00002290, -0.00001980, 0.00000758, 0.00000379, -0.00000912, 0.00002081, -0.00001609, 0.00000907, -0.00000462, -0.00000585, 0.00001318, -0.00000227, 0.00000762, -0.00000051, 0.00000342, -0.00001186, 0.00002372, -0.00002543, 0.00002784, -0.00002620, 0.00001601, -0.00001658, 0.00000649, -0.00001741, 0.00002377, 0.00000170, -0.00001920, -0.00000336, 0.00001048, -0.00000721, 0.00000505, 0.00000028, -0.00000870, -0.00000633, 0.00000338, -0.00000601, 0.00000543, -0.00001549, 0.00001379, -0.00002145, 0.00001697, -0.00002053, 0.00001249, -0.00002170, 0.00000548, -0.00003921, 0.00001731, -0.00005722, 0.00001953, -0.00010842, -0.00006094, -0.00061847, 0.00131297, 0.00031794, -0.00056555, 0.00007182, -0.00014058, 0.00005031, -0.00006896, 0.00003314, -0.00005081, 0.00002766, -0.00004098, 0.00002575, -0.00003287, 0.00002104, -0.00002375, 0.00001208, -0.00001381, 0.00001364, -0.00001515, 0.00001338, -0.00001286, 0.00002027, -0.00002734, 0.00001323, -0.00001162, -0.00000015, 0.00001054, 0.00001424, -0.00002244, -0.00000147, 0.00001157, 0.00000075, -0.00001779, 0.00002670, -0.00002651, 0.00002116, -0.00002506, 0.00000732, -0.00000500, 0.00000500, -0.00000877, 0.00000602, -0.00000846, 0.00000588, -0.00000403, 0.00000653, -0.00001321, 0.00001251, -0.00001968, 0.00001499, -0.00000970, 0.00000183, 0.00000407, -0.00002064, 0.00001944, 0.00000054, -0.00000230, -0.00000918, 0.00000787, 0.00000319, },
		{ -0.00000611, -0.00000737, 0.00000951, 0.00000002, 0.00000023, -0.00000294, 0.00000894, -0.00001944, 0.00001618, -0.00000417, 0.00000302, -0.00001252, 0.00000650, 0.00000481, -0.00000261, -0.00000984, 0.00000538, 0.00000506, -0.00001837, 0.00002261, -0.00002058, 0.00002173, -0.00000595, -0.00000344, 0.00000842, -0.00001583, 0.00001559, -0.00000216, 0.00000878, -0.00000249, 0.00000970, -0.00001818, 0.00002608, -0.00002127, 0.00001920, -0.00001278, -0.00000023, -0.00000040, -0.00000652, -0.00000472, 0.00000096, 0.00002448, -0.00002404, -0.00000670, 0.00000622, 0.00000062, -0.00000196, 0.00000491, -0.00000564, -0.00001211, 0.00000643, -0.00000939, 0.00000941, -0.00001855, 0.00001514, -0.00002024, 0.00001387, -0.00001709, 0.00001135, -0.00002175, 0.00001234, -0.00004641, 0.00002605, -0.00006496, 0.00003510, -0.00011894, 0.00002608, -0.00059196, 0.00063996, 0.00098989, -0.00058422, -0.00003572, -0.00010521, 0.00001496, -0.00005023, 0.00001828, -0.00003866, 0.00001556, -0.00003047, 0.00001422, -0.00002240, 0.00001016, -0.00001444, 0.00000503, -0.00001023, 0.00001071, -0.00001250, 0.00001097, -0.00001123, 0.00001714, -0.00001625, -0.00000082, -0.00000043, -0.00000751, 0.00000659, 0.00002017, -0.00001480, -0.00000955, 0.00000400, 0.00001694, -0.00002847, 0.00002867, -0.00002278, 0.00001413, -0.00001316, -0.00000581, 0.00000289, 0.00000007, -0.00000408, 0.00000109, -0.00000437, 0.00000159, -0.00000156, 0.00000497, -0.00000807, 0.00000331, -0.00000467, -0.00000606, 0.00001170, -0.00001693, 0.00001734, -0.00002406, 0.00001107, 0.00000875, -0.00000272, -0.00000875, 0.00000254, 0.00000800, },
		{ -0.00000429, -0.00000875, 0.00000549, 0.00000395, 0.00000013, -0.00000398, 0.00000876, -0.00001328, 0.00000314, 0.00000848, -0.00000594, -0.00000114, -0.00000841, 0.00001592, -0.00000893, -0.00000058, -0.00000966, 0.00001975, -0.00002488, 0.00001713, -0.00000796, 0.00000212, 0.00001597, -0.00001837, 0.00001748, -0.00001757, 0.00001002, 0.00000287, 0.00000650, -0.00000108, 0.00001065, -0.00001561, 0.00001718, -0.00000584, 0.00000087, 0.00000816, -0.00002019, 0.00001530, -0.00001766, 0.00000673, -0.00001572, 0.00003180, -0.00001381, -0.00001520, 0.00000278, 0.00000494, -0.00000329, 0.00000343, 0.00000052, -0.00001526, 0.00000384, -0.00000621, 0.00000555, -0.00001243, 0.00000622, -0.00000861, 0.00000123, -0.00000493, 0.00000093, -0.00001113, 0.00000460, -0.00003319, 0.00000857, -0.00004039, 0.00000880, -0.00007713, 0.00000742, -0.00040091, 0.00004678, 0.00120065, -0.00018765, -0.00029598, -0.00001220, -0.00006110, -0.00000225, -0.00002329, -0.00000317, -0.00001615, -0.00000337, -0.00001114, -0.00000056, -0.00001000, 0.00000227, -0.00000876, 0.00000064, -0.00000020, -0.00000171, 0.00000008, -0.00000055, 0.00000443, 0.00000135, -0.00001496, 0.00000663, -0.00000925, 0.00000030, 0.00002171, -0.00000468, -0.00001180, -0.00000796, 0.00002662, -0.00002549, 0.00001715, -0.00000763, -0.00000264, 0.00000518, -0.00002046, 0.00000979, -0.00000450, 0.00000105, -0.00000419, 0.00000033, -0.00000388, 0.00000314, -0.00000105, 0.00000223, -0.00000904, 0.00000910, -0.00001824, 0.00001654, -0.00001429, 0.00000769, -0.00000752, -0.00000920, 0.00002051, -0.00000518, -0.00000419, -0.00000584, 0.00001265, },
		{ 0.00000357, -0.00001267, 0.00000389, 0.00000387, 0.00000304, -0.00000533, 0.00000681, -0.00000558, -0.00000759, 0.00001461, -0.00000748, 0.00000218, -0.00001191, 0.00001316, -0.00000249, -0.00000347, -0.00000875, 0.00001261, -0.00000714, -0.00000793, 0.00001600, -0.00002000, 0.00003050, -0.00002043, 0.00001342, -0.00000840, -0.00000114, 0.00000862, 0.00000460, 0.00000057, 0.00000940, -0.00000883, 0.00000461, 0.00001018, -0.00001387, 0.00002119, -0.00002729, 0.00001501, -0.00001316, 0.00000305, -0.00001287, 0.00001636, 0.00001091, -0.00002744, 0.00000314, 0.00000087, 0.00000432, -0.00000510, 0.00000966, -0.00001773, 0.00000145, -0.00000545, 0.00000496, -0.00001028, 0.00000238, -0.00000416, -0.00000185, -0.00000381, 0.00000157, -0.00001122, 0.00000574, -0.00002736, -0.00000074, -0.00002624, -0.00000556, -0.00005050, -0.00001293, -0.00023483, -0.00027326, 0.00097853, 0.00036485, -0.00044477, -0.00002796, -0.00006935, -0.00000596, -0.00002631, -0.00000094, -0.00001774, -0.00000347, -0.00001166, -0.00000156, -0.00000916, -0.00000034, -0.00000557, -0.00000304, 0.00000176, -0.00000285, 0.00000040, -0.00000075, 0.00000305, 0.00000421, -0.00001082, -0.00000349, 0.00000246, -0.00001312, 0.00002517, -0.00000058, -0.00000410, -0.00002160, 0.00002728, -0.00001251, 0.00000076, 0.00000825, -0.00001683, 0.00001655, -0.00002431, 0.00000620, -0.00000140, 0.00000007, -0.00000378, 0.00000011, -0.00000494, 0.00000373, -0.00000312, 0.00000619, -0.00001077, 0.00000754, -0.00001067, 0.00000102, 0.00000483, -0.00001190, 0.00001063, -0.00002231, 0.00002080, 0.00000061, -0.00000436, -0.00000811, 0.00000924, },
		{ 0.00001435, -0.00001628, 0.00000290, 0.00000186, 0.00000636, -0.00000545, 0.00000396, 0.00000032, -0.00001183, 0.00001157, -0.00000070, -0.00000255, -0.00000564, 0.00000119, 0.00000995, -0.00001125, -0.00000081, -0.00000229, 0.00001320, -0.00002552, 0.00002392, -0.00002086, 0.00002092, -0.00000200, -0.00000403, 0.00000761, -0.00001267, 0.00001192, 0.00000376, 0.00000297, 0.00000578, 0.00000046, -0.00000748, 0.00002100, -0.00001895, 0.00002115, -0.00001896, 0.00000032, 0.00000194, -0.00000921, -0.00000010, -0.00000571, 0.00003066, -0.00002868, -0.00000166, -0.00000286, 0.00001006, -0.00000919, 0.00001192, -0.00001352, -0.00000340, -0.00000531, 0.00000562, -0.00000988, 0.00000139, -0.00000433, 0.00000086, -0.00000826, 0.00000661, -0.00001481, 0.00000898, -0.00002391, -0.00000485, -0.00002082, -0.00000872, -0.00003863, -0.00002373, -0.00012356, -0.00037627, 0.00058728, 0.00073944, -0.00029912, -0.00019454, -0.00000676, -0.00005854, 0.00000568, -0.00002922, 0.00000705, -0.00002573, 0.00000676, -0.00001884, 0.00000582, -0.00001523, 0.00000781, -0.00001534, 0.00001155, -0.00001152, 0.00000814, -0.00000858, 0.00000994, -0.00000312, 0.00000306, -0.00001830, 0.00001362, -0.00002110, 0.00002219, 0.00000422, 0.00000292, -0.00002491, 0.00001363, 0.00000754, -0.00001440, 0.00001869, -0.00002245, 0.00001614, -0.00001584, -0.00000592, 0.00000612, -0.00000400, -0.00000038, -0.00000276, -0.00000304, 0.00000062, -0.00000042, 0.00000256, -0.00000254, -0.00000429, 0.00000502, -0.00001621, 0.00001731, -0.00001775, 0.00001011, -0.00001476, 0.00000409, 0.00001548, -0.00000983, -0.00000350, -0.00000108, },
		{ 0.00001789, -0.00001159, -0.00000357, 0.00000379, 0.00000461, -0.00000056, -0.00000179, 0.00000591, -0.00001274, 0.00000566, 0.00000578, -0.00000556, -0.00000129, -0.00000603, 0.00001376, -0.00000956, -0.00000154, -0.00000590, 0.00001503, -0.00001755, 0.00000536, 0.00000133, -0.00000610, 0.00002398, -0.00002053, 0.00001829, -0.00001676, 0.00000996, 0.00000381, 0.00000673, 0.00000011, 0.00001030, -0.00001618, 0.00002451, -0.00001475, 0.00001202, -0.00000374, -0.00001574, 0.00001257, -0.00001471, 0.00000533, -0.00001461, 0.00002883, -0.00001057, -0.00001593, 0.00000027, 0.00000551, -0.00000149, 0.00000257, -0.00000056, -0.00001255, -0.00000159, 0.00000096, -0.00000321, -0.00000587, 0.00000155, -0.00000346, -0.00000356, 0.00000030, -0.00000564, -0.00000196, -0.00000679, -0.00002079, -0.00000425, -0.00002443, -0.00001328, -0.00005206, -0.00003185, -0.00038514, 0.00025194, 0.00080265, 0.00006636, -0.00035980, 0.00001014, -0.00007757, 0.00000923, -0.00003593, 0.00001211, -0.00002872, 0.00000694, -0.00001901, 0.00000480, -0.00001504, 0.00000664, -0.00001391, 0.00000862, -0.00000847, 0.00000504, -0.00000637, 0.00000731, -0.00000154, 0.00000554, -0.00001688, 0.00000645, -0.00001070, 0.00000456, 0.00001735, -0.00000057, -0.00001155, -0.00001183, 0.00002748, -0.00002303, 0.00002157, -0.00001964, 0.00000735, -0.00000249, -0.00001701, 0.00000920, -0.00000399, -0.00000015, -0.00000287, -0.00000287, -0.00000166, 0.00000208, -0.00000176, 0.00000507, -0.00001162, 0.00001070, -0.00001660, 0.00000858, -0.00000252, -0.00000802, 0.00000459, -0.00001592, 0.00002586, -0.00001021, -0.00000107, -0.00000817, },
		{ 0.00000944, 0.00000293, -0.00001537, 0.00000963, -0.00000172, 0.00000766, -0.00000855, 0.00001066, -0.00001253, 0.00000169, 0.00000643, -0.00000166, -0.00000430, -0.00000323, 0.00000557, 0.00000291, -0.00001180, 0.00000230, 0.00000136, 0.00000516, -0.00002051, 0.00002232, -0.00002437, 0.00003339, -0.00001825, 0.00001285, -0.00000863, 0.00000182, 0.00000537, 0.00000959, -0.00000365, 0.00001566, -0.00001708, 0.00001886, -0.00000304, -0.00000071, 0.00000978, -0.00002343, 0.00001084, -0.00000822, -0.00000084, -0.00000805, 0.00001009, 0.00001525, -0.00002864, 0.00000369, -0.00000366, 0.00001038, -0.00000902, 0.00001166, -0.00001912, 0.00000227, -0.00000718, 0.00000797, -0.00001754, 0.00001198, -0.00001382, 0.00000906, -0.00001502, 0.00001262, -0.00002200, 0.00001783, -0.00004259, 0.00001855, -0.00004827, 0.00002035, -0.00008751, 0.00003591, -0.00034571, 0.00002001, 0.00066449, 0.00040449, -0.00032755, -0.00011046, -0.00001736, -0.00004534, 0.00000145, -0.00002186, 0.00000327, -0.00002337, 0.00000750, -0.00001974, 0.00000630, -0.00001401, 0.00000533, -0.00001077, 0.00000986, -0.00001226, 0.00000935, -0.00000862, 0.00001328, -0.00000733, 0.00000156, -0.00001522, 0.00001044, -0.00001826, 0.00003128, -0.00000852, 0.00000777, -0.00003330, 0.00003409, -0.00001880, 0.00001527, -0.00000975, -0.00000510, 0.00000858, -0.00002051, 0.00000414, 0.00000155, -0.00000377, 0.00000017, -0.00000491, -0.00000176, 0.00000166, -0.00000205, 0.00000533, -0.00000756, 0.00000199, -0.00000155, -0.00001179, 0.00001676, -0.00002215, 0.00001314, -0.00001945, 0.00001816, 0.00000246, -0.00000721, -0.00000549, },
		{ -0.00000317, 0.00001726, -0.00002274, 0.00001104, -0.00000466, 0.00001132, -0.00000963, 0.00000989, -0.00000850, -0.00000236, 0.00000477, 0.00000369, -0.00000804, 0.00000118, -0.00000335, 0.00001278, -0.00001739, 0.00000645, -0.00000757, 0.00001705, -0.00002644, 0.00001821, -0.00001584, 0.00001609, 0.00000442, -0.00000503, 0.00000621, -0.00000812, 0.00000761, 0.00000911, -0.00000142, 0.00001245, -0.00000782, 0.00000457, 0.00001300, -0.00001264, 0.00001779, -0.00002207, 0.00000113, 0.00000192, -0.00000842, 0.00000041, -0.00000634, 0.00002810, -0.00002477, -0.00000280, -0.00000639, 0.00001331, -0.00000975, 0.00001182, -0.00001478, -0.00000055, -0.00001063, 0.00001321, -0.00002148, 0.00001388, -0.00001609, 0.00001349, -0.00002080, 0.00001928, -0.00002841, 0.00002550, -0.00004488, 0.00002090, -0.00005031, 0.00002710, -0.00008752, 0.00004059, -0.00024230, -0.00016336, 0.00050083, 0.00053424, -0.00006526, -0.00030117, 0.00005084, -0.00010096, 0.00003634, -0.00005530, 0.00003371, -0.00004988, 0.00002883, -0.00003842, 0.00002183, -0.00002924, 0.00001934, -0.00002518, 0.00002316, -0.00002437, 0.00002011, -0.00001997, 0.00002357, -0.00001667, 0.00001539, -0.00002781, 0.00001908, -0.00002554, 0.00002948, -0.00000570, 0.00001485, -0.00003358, 0.00001835, -0.00000028, 0.00000092, 0.00000508, -0.00001775, 0.00001453, -0.00001705, -0.00000469, 0.00000682, -0.00000578, 0.00000147, -0.00000512, -0.00000261, 0.00000046, -0.00000013, 0.00000114, 0.00000156, -0.00000962, 0.00001150, -0.00002204, 0.00001819, -0.00001451, -0.00000018, -0.00000328, -0.00000354, 0.00002066, -0.00001525, 0.00000106, },
		{ -0.00000771, 0.00001935, -0.00001624, 0.00000119, 0.00000169, 0.00000534, -0.00000154, 0.00000145, 0.00000107, -0.00000943, 0.00000619, 0.00000329, -0.00000456, -0.00000197, -0.00000240, 0.00000942, -0.00000860, -0.00000327, -0.00000069, 0.00000775, -0.00000764, -0.00000744, 0.00000874, -0.00001120, 0.00002771, -0.00001745, 0.00001429, -0.00001166, 0.00000724, 0.00000629, 0.00000562, 0.00000323, 0.00000680, -0.00001154, 0.00002620, -0.00001834, 0.00001844, -0.00001479, -0.00000877, 0.00000619, -0.00000802, 0.00000018, -0.00000871, 0.00002055, -0.00000438, -0.00001694, -0.00000192, 0.00000514, 0.00000119, 0.00000106, -0.00000182, -0.00000925, -0.00000671, 0.00000717, -0.00001137, 0.00000076, -0.00000317, 0.00000136, -0.00000757, 0.00000434, -0.00001088, 0.00000601, -0.00001792, -0.00000729, -0.00001871, -0.00000671, -0.00003806, -0.00002097, -0.00009047, -0.00031511, 0.00037509, 0.00048217, 0.00025482, -0.00037582, 0.00000625, -0.00007868, 0.00000966, -0.00003966, 0.00001950, -0.00003453, 0.00001186, -0.00002244, 0.00000627, -0.00001639, 0.00000743, -0.00001534, 0.00001306, -0.00001380, 0.00000925, -0.00001050, 0.00001316, -0.00000583, 0.00000742, -0.00001586, 0.00000351, -0.00000978, 0.00000808, 0.00001167, 0.00000464, -0.00001234, -0.00001196, 0.00002209, -0.00001346, 0.00001871, -0.00002583, 0.00001409, -0.00000996, -0.00001139, 0.00000643, -0.00000233, -0.00000183, -0.00000140, -0.00000585, 0.00000066, 0.00000047, -0.00000141, 0.00000615, -0.00001179, 0.00000962, -0.00001241, -0.00000071, 0.00000864, -0.00002163, 0.00001390, -0.00001890, 0.00002665, -0.00001200, 0.00000013, },
		{ -0.00000201, 0.00000870, 0.00000121, -0.00001496, 0.00001222, -0.00000482, 0.00000957, -0.00000854, 0.00001093, -0.00001601, 0.00000895, -0.00000130, 0.00000381, -0.00000993, 0.00000490, -0.00000156, 0.00000658, -0.00001749, 0.00001114, -0.00000774, 0.00001399, -0.00002744, 0.00002108, -0.00002164, 0.00002939, -0.00000979, 0.00000758, -0.00000516, 0.00000250, 0.00000419, 0.00001182, -0.00000398, 0.00001688, -0.00001913, 0.00002788, -0.00001250, 0.00001034, -0.00000351, -0.00001594, 0.00000397, -0.00000196, -0.00000518, -0.00000299, 0.00000483, 0.00001550, -0.00002416, -0.00000045, -0.00000395, 0.00001166, -0.00000776, 0.00000783, -0.00001477, -0.00000180, -0.00000386, 0.00000473, -0.00001743, 0.00001394, -0.00001571, 0.00001181, -0.00001740, 0.00001375, -0.00002135, 0.00001616, -0.00004159, 0.00001973, -0.00004972, 0.00002105, -0.00009110, 0.00003462, -0.00037965, 0.00023142, 0.00040345, 0.00042132, -0.00021421, -0.00017200, 0.00001314, -0.00006677, 0.00001426, -0.00003007, 0.00001250, -0.00003189, 0.00001603, -0.00002824, 0.00001309, -0.00001983, 0.00000919, -0.00001172, 0.00001102, -0.00001484, 0.00001171, -0.00000999, 0.00001712, -0.00001325, 0.00000857, -0.00002101, 0.00001194, -0.00001559, 0.00002859, -0.00000866, 0.00001170, -0.00003504, 0.00003073, -0.00001645, 0.00002275, -0.00002345, 0.00000546, -0.00000070, -0.00001463, 0.00000150, 0.00000352, -0.00000611, 0.00000255, -0.00000836, 0.00000053, -0.00000009, -0.00000089, 0.00000412, -0.00000420, -0.00000297, 0.00000575, -0.00002096, 0.00002356, -0.00002716, 0.00001128, -0.00001272, 0.00001236, 0.00000524, -0.00000993, },
		{ 0.00000355, -0.00000276, 0.00001565, -0.00002375, 0.00001483, -0.00000789, 0.00001294, -0.00001041, 0.00001262, -0.00001479, 0.00000605, -0.00000186, 0.00000723, -0.00001186, 0.00000636, -0.00000582, 0.00001218, -0.00001960, 0.00001030, -0.00000917, 0.00001597, -0.00002142, 0.00000689, -0.00000736, 0.00000874, 0.00001224, -0.00000603, 0.00000457, -0.00000294, 0.00000359, 0.00001295, -0.00000331, 0.00001613, -0.00001312, 0.00001568, 0.00000417, -0.00000396, 0.00000934, -0.00002080, 0.00000041, 0.00000130, -0.00000592, -0.00000144, -0.00000295, 0.00001938, -0.00001488, -0.00000821, -0.00000571, 0.00001165, -0.00000525, 0.00000644, -0.00001112, -0.00000226, -0.00001036, 0.00001361, -0.00002494, 0.00001822, -0.00001954, 0.00001720, -0.00002351, 0.00002085, -0.00002932, 0.00002726, -0.00004924, 0.00002874, -0.00005964, 0.00003854, -0.00010558, 0.00006403, -0.00031160, 0.00002111, 0.00039682, 0.00037862, 0.00010439, -0.00034726, 0.00006407, -0.00011069, 0.00004103, -0.00005940, 0.00004036, -0.00005560, 0.00003441, -0.00004337, 0.00002482, -0.00003140, 0.00001997, -0.00002441, 0.00002415, -0.00002699, 0.00002233, -0.00002190, 0.00002809, -0.00002187, 0.00001942, -0.00002862, 0.00001581, -0.00002069, 0.00002767, -0.00000816, 0.00001929, -0.00003410, 0.00001682, -0.00000446, 0.00001438, -0.00000923, -0.00001019, 0.00000943, -0.00001603, -0.00000309, 0.00000572, -0.00000574, 0.00000175, -0.00000618, -0.00000272, 0.00000058, -0.00000022, 0.00000052, 0.00000396, -0.00001214, 0.00001410, -0.00002315, 0.00001477, -0.00000884, -0.00001075, 0.00000741, -0.00000938, 0.00002229, -0.00001723, },
		{ -0.00000083, -0.00000435, 0.00001648, -0.00001702, 0.00000407, 0.00000036, 0.00000524, -0.00000187, 0.00000448, -0.00000433, -0.00000446, 0.00000498, 0.00000140, -0.00000341, -0.00000265, 0.00000163, 0.00000379, -0.00000663, -0.00000510, 0.00000461, 0.00000000, 0.00000222, -0.00001843, 0.00001217, -0.00001227, 0.00002704, -0.00000978, 0.00000663, -0.00000338, 0.00000272, 0.00000969, 0.00000382, 0.00000748, 0.00000123, -0.00000250, 0.00002231, -0.00001582, 0.00001786, -0.00002164, -0.00000244, -0.00000107, -0.00000010, -0.00000641, -0.00000044, 0.00000939, 0.00000308, -0.00001764, -0.00000387, 0.00000401, 0.00000433, -0.00000088, -0.00000294, -0.00000628, -0.00001011, 0.00001056, -0.00001676, 0.00000543, -0.00000622, 0.00000429, -0.00000933, 0.00000564, -0.00001279, 0.00001000, -0.00002510, 0.00000283, -0.00002957, 0.00000639, -0.00005735, 0.00000488, -0.00015253, -0.00020573, 0.00041601, 0.00025505, 0.00036382, -0.00034016, -0.00002353, -0.00006481, -0.00000090, -0.00003392, 0.00001778, -0.00003219, 0.00001016, -0.00002051, 0.00000324, -0.00001359, 0.00000426, -0.00001245, 0.00001287, -0.00001481, 0.00000955, -0.00001090, 0.00001560, -0.00000733, 0.00000648, -0.00001192, -0.00000199, -0.00000620, 0.00000970, 0.00000633, 0.00000939, -0.00001301, -0.00000962, 0.00001243, 0.00000093, 0.00000971, -0.00002547, 0.00001572, -0.00001507, -0.00000545, 0.00000274, -0.00000043, -0.00000329, -0.00000046, -0.00000781, 0.00000222, -0.00000097, -0.00000047, 0.00000581, -0.00001002, 0.00000667, -0.00000717, -0.00000899, 0.00001622, -0.00003009, 0.00001819, -0.00001700, 0.00002225, -0.00001007, },
		{ -0.00001148, 0.00000150, 0.00000720, -0.00000118, -0.00001182, 0.00001155, -0.00000509, 0.00000854, -0.00000526, 0.00000741, -0.00001513, 0.00001238, -0.00000647, 0.00000709, -0.00001294, 0.00001062, -0.00000658, 0.00000725, -0.00001910, 0.00001579, -0.00001254, 0.00001752, -0.00002767, 0.00001270, -0.00001333, 0.00002032, 0.00000301, -0.00000072, 0.00000100, 0.00000132, 0.00000499, 0.00001137, -0.00000061, 0.00001320, -0.00001449, 0.00003024, -0.00001599, 0.00001599, -0.00001515, -0.00000735, -0.00000391, 0.00000452, -0.00000925, 0.00000185, 0.00000045, 0.00001299, -0.00001622, -0.00000699, -0.00000188, 0.00000996, -0.00000351, 0.00000093, -0.00000774, -0.00000761, 0.00000131, -0.00000165, -0.00001268, 0.00001066, -0.00001220, 0.00000891, -0.00001382, 0.00000850, -0.00001357, 0.00000574, -0.00002976, 0.00000852, -0.00003696, 0.00000394, -0.00007000, -0.00000734, -0.00034008, 0.00034954, 0.00021372, 0.00040702, -0.00010610, -0.00021128, 0.00002324, -0.00007241, 0.00001490, -0.00002860, 0.00001270, -0.00003135, 0.00001582, -0.00002868, 0.00001278, -0.00001942, 0.00000790, -0.00000810, 0.00000743, -0.00001250, 0.00000921, -0.00000635, 0.00001557, -0.00001402, 0.00001052, -0.00002153, 0.00000845, -0.00000729, 0.00001929, -0.00000268, 0.00000895, -0.00002824, 0.00001890, -0.00000650, 0.00002240, -0.00003046, 0.00001212, -0.00000898, -0.00000812, -0.00000110, 0.00000423, -0.00000677, 0.00000311, -0.00000986, 0.00000154, -0.00000125, 0.00000017, 0.00000286, -0.00000127, -0.00000639, 0.00000992, -0.00002496, 0.00002398, -0.00002642, 0.00000573, -0.00000360, 0.00000483, 0.00000780, },
		{ -0.00001505, 0.00000356, -0.00000002, 0.00000999, -0.00001933, 0.00001369, -0.00000710, 0.00001063, -0.00000714, 0.00001112, -0.00001676, 0.00001112, -0.00000628, 0.00000852, -0.00001288, 0.00000891, -0.00000570, 0.00000800, -0.00001755, 0.00001020, -0.00000697, 0.00001070, -0.00001225, -0.00000725, 0.00000175, 0.00000093, 0.00002001, -0.00000645, 0.00000299, 0.00000209, 0.00000068, 0.00001470, -0.00000220, 0.00001610, -0.00001399, 0.00002370, -0.00000305, 0.00000429, -0.00000212, -0.00001558, -0.00000245, 0.00000133, -0.00000339, -0.00000363, 0.00000169, 0.00000819, -0.00000284, -0.00001479, -0.00000324, 0.00000754, 0.00000149, -0.00000121, -0.00000534, -0.00000604, -0.00000678, 0.00000882, -0.00002188, 0.00001585, -0.00001628, 0.00001389, -0.00001871, 0.00001418, -0.00002081, 0.00001761, -0.00004014, 0.00002151, -0.00005159, 0.00002832, -0.00009511, 0.00004668, -0.00032347, 0.00015822, 0.00029277, 0.00026799, 0.00020938, -0.00033664, 0.00003942, -0.00009401, 0.00002548, -0.00004657, 0.00003114, -0.00004633, 0.00002647, -0.00003641, 0.00001748, -0.00002439, 0.00001246, -0.00001578, 0.00001693, -0.00002134, 0.00001652, -0.00001579, 0.00002447, -0.00001950, 0.00001624, -0.00002247, 0.00000620, -0.00000909, 0.00001870, -0.00000419, 0.00001689, -0.00002720, 0.00000947, -0.00000413, 0.00002343, -0.00002101, -0.00000220, 0.00000207, -0.00001285, -0.00000179, 0.00000358, -0.00000454, 0.00000097, -0.00000628, -0.00000304, 0.00000055, -0.00000028, 0.00000044, 0.00000480, -0.00001194, 0.00001297, -0.00002006, 0.00000823, -0.00000268, -0.00001903, 0.00001490, -0.00001181, 0.00001941, },
		{ -0.00000506, -0.00000322, 0.00000095, 0.00001010, -0.00001394, 0.00000444, 0.00000046, 0.00000378, -0.00000082, 0.00000615, -0.00000865, 0.00000062, 0.00000279, 0.00000011, -0.00000240, -0.00000288, 0.00000520, -0.00000212, -0.00000433, -0.00000637, 0.00000877, -0.00000634, 0.00000973, -0.00002547, 0.00001069, -0.00000969, 0.00002309, 0.00000066, -0.00000242, 0.00000596, -0.00000199, 0.00001285, 0.00000242, 0.00001145, -0.00000441, 0.00000876, 0.00001474, -0.00000871, 0.00001038, -0.00002264, 0.00000083, -0.00000661, 0.00000639, -0.00001175, 0.00000718, -0.00000154, 0.00000960, -0.00001708, -0.00000594, 0.00000263, 0.00000724, -0.00000277, -0.00000384, -0.00000414, -0.00001116, 0.00001028, -0.00001826, 0.00000702, -0.00000674, 0.00000469, -0.00000838, 0.00000381, -0.00001074, 0.00000882, -0.00002628, 0.00000672, -0.00003336, 0.00001045, -0.00006500, 0.00001570, -0.00020083, -0.00007962, 0.00038980, 0.00011820, 0.00039646, -0.00026117, -0.00007106, -0.00004159, -0.00001847, -0.00002155, 0.00000890, -0.00002311, 0.00000298, -0.00001428, -0.00000318, -0.00000777, -0.00000175, -0.00000619, 0.00000861, -0.00001171, 0.00000601, -0.00000751, 0.00001440, -0.00000602, 0.00000322, -0.00000609, -0.00000887, -0.00000060, 0.00000904, 0.00000252, 0.00001236, -0.00001223, -0.00000675, 0.00000162, 0.00001631, -0.00000236, -0.00001987, 0.00001207, -0.00001663, -0.00000069, -0.00000069, 0.00000093, -0.00000400, -0.00000042, -0.00000832, 0.00000249, -0.00000170, 0.00000055, 0.00000461, -0.00000710, 0.00000286, -0.00000223, -0.00001456, 0.00001842, -0.00003187, 0.00001678, -0.00001051, 0.00001334, },
		{ 0.00000907, -0.00001102, 0.00000428, 0.00000517, -0.00000354, -0.00000704, 0.00000851, -0.00000325, 0.00000548, 0.00000014, 0.00000126, -0.00001048, 0.00001163, -0.00000814, 0.00000729, -0.00001251, 0.00001297, -0.00000879, 0.00000473, -0.00001649, 0.00001572, -0.00001246, 0.00001558, -0.00002210, -0.00000043, -0.00000265, 0.00000935, 0.00001704, -0.00000928, 0.00000790, -0.00000052, 0.00000692, 0.00000927, 0.00000509, 0.00000631, -0.00000451, 0.00002659, -0.00001339, 0.00001459, -0.00002227, -0.00000042, -0.00001062, 0.00000957, -0.00001192, 0.00000562, -0.00000292, 0.00000902, -0.00000725, -0.00001351, 0.00000050, 0.00000729, 0.00000130, -0.00000609, -0.00000116, -0.00001213, 0.00000520, -0.00000787, -0.00000649, 0.00000518, -0.00000636, 0.00000361, -0.00000794, 0.00000120, -0.00000396, -0.00000709, -0.00001429, -0.00000714, -0.00001913, -0.00001990, -0.00003857, -0.00007105, -0.00024524, 0.00038124, 0.00009782, 0.00036212, -0.00000679, -0.00022830, 0.00001621, -0.00006632, 0.00000756, -0.00002161, 0.00000774, -0.00002527, 0.00001009, -0.00002393, 0.00000804, -0.00001536, 0.00000407, -0.00000270, 0.00000190, -0.00000789, 0.00000442, -0.00000014, 0.00001063, -0.00001117, 0.00000876, -0.00001851, 0.00000254, 0.00000318, 0.00000715, 0.00000602, 0.00000271, -0.00001654, 0.00000263, 0.00000765, 0.00001580, -0.00003035, 0.00001337, -0.00001451, -0.00000234, -0.00000308, 0.00000362, -0.00000593, 0.00000208, -0.00000947, 0.00000121, -0.00000172, 0.00000100, 0.00000178, 0.00000084, -0.00000787, 0.00001054, -0.00002359, 0.00001828, -0.00002079, -0.00000191, 0.00000596, -0.00000280, },
		{ 0.00001231, -0.00000917, 0.00000167, 0.00000389, 0.00000207, -0.00001120, 0.00000875, -0.00000317, 0.00000531, -0.00000073, 0.00000591, -0.00001420, 0.00001188, -0.00000746, 0.00000701, -0.00001085, 0.00000876, -0.00000343, 0.00000121, -0.00001215, 0.00000684, -0.00000206, 0.00000308, -0.00000174, -0.00002073, 0.00000847, -0.00000526, 0.00002631, -0.00000563, 0.00000168, 0.00000636, -0.00000064, 0.00001433, 0.00000162, 0.00001266, -0.00001005, 0.00002740, -0.00000679, 0.00000917, -0.00001345, -0.00000837, -0.00000677, 0.00000225, -0.00000179, -0.00000467, 0.00000515, -0.00000193, 0.00000757, -0.00001950, -0.00000148, 0.00000350, 0.00000758, -0.00000804, -0.00000033, -0.00000925, -0.00000261, 0.00000172, -0.00001506, 0.00000950, -0.00000927, 0.00000685, -0.00001023, 0.00000379, -0.00000811, 0.00000242, -0.00002371, 0.00000557, -0.00003316, 0.00000451, -0.00006552, -0.00000159, -0.00028191, 0.00023926, 0.00020306, 0.00019274, 0.00025854, -0.00028224, -0.00000965, -0.00006232, -0.00000062, -0.00002521, 0.00001340, -0.00002859, 0.00001082, -0.00002276, 0.00000466, -0.00001281, 0.00000133, -0.00000375, 0.00000569, -0.00001127, 0.00000631, -0.00000499, 0.00001565, -0.00001224, 0.00000868, -0.00001274, -0.00000573, 0.00000498, 0.00000638, 0.00000318, 0.00001030, -0.00001553, -0.00000158, 0.00000009, 0.00002675, -0.00002773, 0.00000368, -0.00000571, -0.00000828, -0.00000107, 0.00000108, -0.00000287, -0.00000033, -0.00000575, -0.00000337, 0.00000018, -0.00000008, 0.00000061, 0.00000442, -0.00000961, 0.00000899, -0.00001399, 0.00000015, 0.00000224, -0.00002332, 0.00001794, -0.00001046, },
		{ 0.00000182, 0.00000144, -0.00000563, 0.00000642, 0.00000207, -0.00000820, 0.00000271, 0.00000184, 0.00000106, 0.00000170, 0.00000587, -0.00001072, 0.00000438, 0.00000060, -0.00000079, -0.00000161, -0.00000237, 0.00000758, -0.00000734, -0.00000236, -0.00000663, 0.00001093, -0.00000996, 0.00001343, -0.00002758, 0.00000467, -0.00000500, 0.00001778, 0.00001146, -0.00001021, 0.00001402, -0.00000513, 0.00001456, 0.00000253, 0.00001372, -0.00000791, 0.00001943, 0.00000645, -0.00000030, -0.00000158, -0.00001829, 0.00000025, -0.00000891, 0.00000959, -0.00001378, 0.00001163, -0.00000951, 0.00001325, -0.00001451, -0.00000848, 0.00000154, 0.00000920, -0.00000406, -0.00000466, -0.00000287, -0.00001001, 0.00000661, -0.00001595, 0.00000556, -0.00000492, 0.00000296, -0.00000540, -0.00000025, -0.00000573, 0.00000325, -0.00002172, 0.00000408, -0.00002949, 0.00000449, -0.00005909, 0.00000652, -0.00022188, 0.00003922, 0.00032073, 0.00005438, 0.00036904, -0.00015608, -0.00012310, -0.00001773, -0.00003671, -0.00000755, -0.00000306, -0.00001077, -0.00000673, -0.00000630, -0.00001063, -0.00000116, -0.00000842, 0.00000131, 0.00000211, -0.00000607, 0.00000008, -0.00000153, 0.00001050, -0.00000284, -0.00000106, -0.00000007, -0.00001529, 0.00000554, 0.00000664, 0.00000061, 0.00001288, -0.00000918, -0.00000509, -0.00000721, 0.00002859, -0.00001358, -0.00001179, 0.00000450, -0.00001469, 0.00000216, -0.00000316, 0.00000131, -0.00000380, -0.00000122, -0.00000745, 0.00000144, -0.00000157, 0.00000130, 0.00000310, -0.00000379, -0.00000091, 0.00000149, -0.00001671, 0.00001502, -0.00002727, 0.00001070, -0.00000112, },
		{ -0.00000905, 0.00000835, -0.00000810, 0.00000573, 0.00000287, -0.00000520, -0.00000198, 0.00000448, -0.00000063, 0.00000190, 0.00000580, -0.00000544, -0.00000387, 0.00000773, -0.00000669, 0.00000468, -0.00000890, 0.00001208, -0.00000850, -0.00000008, -0.00001074, 0.00001178, -0.00000839, 0.00000962, -0.00001315, -0.00001471, 0.00000650, -0.00000005, 0.00002896, -0.00001519, 0.00001335, -0.00000149, 0.00000872, 0.00000707, 0.00001106, -0.00000110, 0.00000801, 0.00001931, -0.00000676, 0.00000674, -0.00002357, 0.00000291, -0.00001435, 0.00001179, -0.00001213, 0.00000746, -0.00000501, 0.00000465, 0.00000077, -0.00001806, 0.00000159, 0.00000524, 0.00000474, -0.00001098, 0.00000277, -0.00001341, 0.00000579, -0.00001152, -0.00000146, 0.00000011, -0.00000087, -0.00000128, -0.00000283, -0.00000453, 0.00000308, -0.00001691, -0.00000159, -0.00002001, -0.00000474, -0.00003969, -0.00001184, -0.00013412, -0.00012236, 0.00034726, 0.00004478, 0.00029498, 0.00007815, -0.00022386, -0.00000359, -0.00005314, -0.00000346, -0.00001303, 0.00000130, -0.00001721, 0.00000217, -0.00001703, 0.00000167, -0.00001015, 0.00000009, 0.00000200, -0.00000306, -0.00000348, -0.00000015, 0.00000609, 0.00000463, -0.00000663, 0.00000510, -0.00001396, -0.00000314, 0.00001232, -0.00000388, 0.00001361, -0.00000314, -0.00000445, -0.00001309, 0.00002143, 0.00000617, -0.00002458, 0.00000930, -0.00001643, 0.00000178, -0.00000412, 0.00000192, -0.00000407, 0.00000004, -0.00000768, -0.00000019, -0.00000158, 0.00000152, 0.00000102, 0.00000201, -0.00000754, 0.00000815, -0.00001792, 0.00000809, -0.00001220, -0.00000955, 0.00001396, },
		{ -0.00000616, 0.00000247, -0.00000105, -0.00000092, 0.00000734, -0.00000566, -0.00000201, 0.00000229, 0.00000189, -0.00000097, 0.00000671, -0.00000112, -0.00000841, 0.00000889, -0.00000565, 0.00000319, -0.00000648, 0.00000668, -0.00000021, -0.00000642, -0.00000520, 0.00000187, 0.00000369, -0.00000444, 0.00000709, -0.00003028, 0.00001035, -0.00000842, 0.00003004, -0.00000303, 0.00000069, 0.00000961, -0.00000036, 0.00001243, 0.00000706, 0.00000712, -0.00000218, 0.00002693, -0.00000656, 0.00000873, -0.00002156, -0.00000177, -0.00001096, 0.00000341, -0.00000114, -0.00000405, 0.00000611, -0.00000859, 0.00001380, -0.00002037, -0.00000202, 0.00000129, 0.00001094, -0.00001187, 0.00000197, -0.00001010, -0.00000003, -0.00000496, -0.00000741, 0.00000215, -0.00000168, -0.00000046, -0.00000197, -0.00000576, 0.00000341, -0.00001211, -0.00000687, -0.00001153, -0.00001301, -0.00002238, -0.00003019, -0.00006418, -0.00020115, 0.00026648, 0.00013601, 0.00014214, 0.00026568, -0.00020160, -0.00006692, -0.00002838, -0.00002692, -0.00000417, -0.00000497, -0.00000959, -0.00000601, -0.00000828, -0.00000832, -0.00000155, -0.00000879, 0.00000709, -0.00000509, -0.00000110, -0.00000406, 0.00000641, 0.00000534, -0.00000348, 0.00000019, -0.00000318, -0.00001576, 0.00001695, -0.00000491, 0.00001019, 0.00000306, -0.00000277, -0.00001312, 0.00000632, 0.00002459, -0.00002826, 0.00000568, -0.00001210, -0.00000338, -0.00000088, -0.00000130, -0.00000133, -0.00000163, -0.00000488, -0.00000360, -0.00000053, 0.00000040, 0.00000086, 0.00000330, -0.00000606, 0.00000343, -0.00000654, -0.00000788, 0.00000478, -0.00002305, 0.00001659, },
		{ 0.00000859, -0.00000979, 0.00000751, -0.00000651, 0.00001014, -0.00000527, -0.00000177, 0.00000008, 0.00000364, -0.00000211, 0.00000505, 0.00000384, -0.00001017, 0.00000597, -0.00000087, -0.00000158, -0.00000104, -0.00000102, 0.00000827, -0.00001110, -0.00000118, -0.00000570, 0.00001078, -0.00001037, 0.00001293, -0.00002500, -0.00000440, -0.00000031, 0.00001310, 0.00002050, -0.00001454, 0.00001892, -0.00000542, 0.00001413, 0.00000471, 0.00001349, -0.00000768, 0.00002717, 0.00000026, 0.00000563, -0.00001414, -0.00001096, -0.00000329, -0.00000801, 0.00000910, -0.00001181, 0.00001174, -0.00001301, 0.00001309, -0.00000979, -0.00001146, 0.00000101, 0.00000977, -0.00000438, -0.00000553, -0.00000223, -0.00000745, 0.00000075, -0.00001098, 0.00000205, -0.00000181, 0.00000026, -0.00000178, -0.00000480, 0.00000019, -0.00000442, -0.00001365, -0.00000299, -0.00002025, -0.00000876, -0.00004255, -0.00002115, -0.00020961, 0.00013147, 0.00023496, 0.00003928, 0.00030481, -0.00004573, -0.00016548, -0.00000191, -0.00004937, 0.00000290, -0.00001351, 0.00000073, -0.00001547, 0.00000046, -0.00001648, 0.00000382, -0.00001336, 0.00000759, -0.00000423, -0.00000016, -0.00000606, 0.00000505, 0.00000552, 0.00000076, -0.00000479, 0.00000434, -0.00001932, 0.00001040, 0.00000368, 0.00000014, 0.00001119, -0.00000390, -0.00000552, -0.00001190, 0.00003467, -0.00002045, -0.00000435, -0.00000474, -0.00001023, 0.00000306, -0.00000446, 0.00000069, -0.00000288, -0.00000248, -0.00000571, -0.00000049, -0.00000077, 0.00000170, 0.00000165, -0.00000071, -0.00000405, 0.00000365, -0.00001577, 0.00000726, -0.00001819, 0.00000222, },
		{ 0.00001879, -0.00001296, 0.00000547, -0.00000286, 0.00000557, 0.00000069, -0.00000578, 0.00000221, 0.00000078, 0.00000150, -0.00000072, 0.00000988, -0.00001053, 0.00000244, 0.00000250, -0.00000344, 0.00000047, -0.00000329, 0.00000889, -0.00000673, -0.00000563, -0.00000374, 0.00000581, -0.00000230, 0.00000202, -0.00000391, -0.00002637, 0.00001074, -0.00000527, 0.00003632, -0.00001626, 0.00001570, -0.00000038, 0.00000934, 0.00000616, 0.00001531, -0.00000624, 0.00001990, 0.00001164, 0.00000036, -0.00000490, -0.00001978, 0.00000220, -0.00001438, 0.00001057, -0.00000946, 0.00000690, -0.00000565, 0.00000057, 0.00000656, -0.00001961, 0.00000068, 0.00000452, 0.00000588, -0.00001266, 0.00000326, -0.00001110, 0.00000271, -0.00001172, 0.00000111, -0.00000310, 0.00000271, -0.00000416, -0.00000015, -0.00000685, 0.00000531, -0.00002076, 0.00000461, -0.00002574, 0.00000098, -0.00004842, -0.00000079, -0.00017705, -0.00000004, 0.00027527, 0.00003538, 0.00021858, 0.00014278, -0.00019980, -0.00003132, -0.00003728, -0.00001467, -0.00000573, -0.00000402, -0.00000981, -0.00000531, -0.00001038, -0.00000422, -0.00000563, -0.00000252, 0.00000453, -0.00000594, -0.00000080, -0.00000279, 0.00001041, -0.00000052, -0.00000210, 0.00000108, -0.00000948, -0.00000672, 0.00001764, -0.00001087, 0.00001722, -0.00000554, 0.00000421, -0.00002392, 0.00003058, -0.00000281, -0.00001583, 0.00000132, -0.00001496, 0.00000395, -0.00000426, -0.00000042, -0.00000189, -0.00000218, -0.00000527, -0.00000210, -0.00000111, 0.00000179, 0.00000057, 0.00000242, -0.00000599, 0.00000395, -0.00000993, -0.00000403, -0.00000314, -0.00001514, },
		{ 0.00001203, -0.00000064, -0.00000793, 0.00000738, -0.00000286, 0.00000888, -0.00001099, 0.00000560, -0.00000327, 0.00000588, -0.00000590, 0.00001275, -0.00000747, -0.00000166, 0.00000383, -0.00000211, -0.00000165, -0.00000110, 0.00000392, 0.00000247, -0.00001287, 0.00000117, -0.00000278, 0.00000826, -0.00000961, 0.00001183, -0.00003378, 0.00000637, -0.00000824, 0.00003083, 0.00000161, -0.00000000, 0.00001189, 0.00000113, 0.00001000, 0.00001259, 0.00000147, 0.00000777, 0.00002371, -0.00000358, 0.00000283, -0.00002458, 0.00000201, -0.00001337, 0.00000379, -0.00000082, -0.00000220, 0.00000451, -0.00001103, 0.00001499, -0.00001704, -0.00000509, 0.00000137, 0.00001088, -0.00001197, 0.00000107, -0.00000815, -0.00000002, -0.00000946, -0.00000113, -0.00000389, 0.00000408, -0.00000550, 0.00000332, -0.00001135, 0.00000999, -0.00002130, 0.00000489, -0.00002358, 0.00000144, -0.00004284, -0.00000225, -0.00012249, -0.00010156, 0.00025108, 0.00009317, 0.00010722, 0.00024562, -0.00011247, -0.00011755, -0.00000282, -0.00004539, 0.00000986, -0.00001792, 0.00000493, -0.00001878, 0.00000243, -0.00001739, 0.00000577, -0.00001459, 0.00001346, -0.00001208, 0.00000583, -0.00001115, 0.00001489, -0.00000317, 0.00000377, -0.00000637, 0.00000331, -0.00002086, 0.00002341, -0.00001172, 0.00001372, -0.00000166, 0.00000759, -0.00002197, 0.00001211, 0.00001850, -0.00002306, 0.00000321, -0.00001589, 0.00000089, -0.00000105, -0.00000326, -0.00000032, -0.00000253, -0.00000395, -0.00000371, -0.00000144, 0.00000104, 0.00000107, 0.00000193, -0.00000227, -0.00000228, 0.00000064, -0.00001450, 0.00000453, -0.00001875, },
		{ -0.00000587, 0.00001598, -0.00001905, 0.00001145, -0.00000483, 0.00001081, -0.00000995, 0.00000340, -0.00000201, 0.00000475, -0.00000459, 0.00000812, 0.00000101, -0.00000752, 0.00000535, -0.00000133, -0.00000238, -0.00000061, 0.00000106, 0.00000712, -0.00001296, -0.00000104, -0.00000371, 0.00000852, -0.00000788, 0.00000887, -0.00001903, -0.00001435, 0.00000227, 0.00001043, 0.00002654, -0.00001426, 0.00001990, -0.00000254, 0.00001169, 0.00000871, 0.00001095, -0.00000330, 0.00003090, -0.00000234, 0.00000625, -0.00002347, -0.00000383, -0.00000773, -0.00000521, 0.00000588, -0.00000669, 0.00000793, -0.00001211, 0.00000945, -0.00000365, -0.00001436, 0.00000089, 0.00000890, -0.00000373, -0.00000638, -0.00000198, -0.00000451, -0.00000566, -0.00000501, -0.00000207, 0.00000125, -0.00000205, 0.00000100, -0.00000805, 0.00000472, -0.00001133, -0.00000526, -0.00001107, -0.00000972, -0.00002415, -0.00002198, -0.00005979, -0.00016667, 0.00018648, 0.00015429, 0.00004855, 0.00022734, 0.00005092, -0.00018719, 0.00000011, -0.00005253, 0.00000638, -0.00001907, 0.00000813, -0.00002045, 0.00000376, -0.00001883, 0.00000546, -0.00001489, 0.00001077, -0.00000837, 0.00000393, -0.00001030, 0.00001023, 0.00000114, 0.00000343, -0.00000673, 0.00000591, -0.00001964, 0.00001251, 0.00000148, 0.00000018, 0.00000822, 0.00000271, -0.00000781, -0.00001190, 0.00003333, -0.00002107, 0.00000005, -0.00001332, -0.00000472, 0.00000254, -0.00000480, -0.00000064, -0.00000164, -0.00000363, -0.00000379, -0.00000263, 0.00000025, 0.00000185, 0.00000047, 0.00000173, -0.00000624, 0.00000440, -0.00001278, -0.00000267, -0.00000743, },
		{ -0.00001719, 0.00001963, -0.00001426, 0.00000096, 0.00000391, 0.00000399, -0.00000105, -0.00000529, 0.00000492, -0.00000179, 0.00000237, -0.00000153, 0.00001128, -0.00001251, 0.00000654, -0.00000213, -0.00000004, -0.00000376, 0.00000274, 0.00000475, -0.00000472, -0.00001029, 0.00000251, -0.00000014, 0.00000361, -0.00000476, 0.00000287, -0.00003268, 0.00000829, -0.00000547, 0.00003831, -0.00001174, 0.00001445, 0.00000317, 0.00000840, 0.00000725, 0.00001670, -0.00000718, 0.00002866, 0.00000630, 0.00000429, -0.00001655, -0.00001321, -0.00000149, -0.00001132, 0.00000646, -0.00000443, 0.00000420, -0.00000497, -0.00000277, 0.00000963, -0.00001810, -0.00000201, 0.00000497, 0.00000484, -0.00001129, 0.00000088, -0.00000631, -0.00000292, -0.00000904, 0.00000127, -0.00000425, 0.00000409, -0.00000480, 0.00000003, -0.00000591, 0.00000297, -0.00001861, 0.00000392, -0.00002381, -0.00000270, -0.00004465, -0.00000931, -0.00018819, 0.00009856, 0.00019237, 0.00004830, 0.00014636, 0.00018275, -0.00015944, -0.00006186, -0.00002247, -0.00002370, -0.00000120, -0.00000700, -0.00000448, -0.00001084, -0.00000537, -0.00000844, -0.00000262, -0.00000329, 0.00000461, -0.00000646, -0.00000020, -0.00000293, 0.00001191, -0.00000370, 0.00000132, -0.00000226, -0.00000592, -0.00000754, 0.00001831, -0.00001269, 0.00001574, -0.00000322, 0.00000749, -0.00002750, 0.00003254, -0.00000823, -0.00000694, -0.00000839, -0.00001113, 0.00000442, -0.00000382, -0.00000283, -0.00000006, -0.00000384, -0.00000302, -0.00000389, -0.00000065, 0.00000196, 0.00000032, 0.00000237, -0.00000402, -0.00000059, -0.00000189, -0.00001533, 0.00000402, },
		{ -0.00001180, 0.00000606, 0.00000411, -0.00001666, 0.00001426, -0.00000318, 0.00000785, -0.00001270, 0.00000989, -0.00000625, 0.00000724, -0.00000783, 0.00001549, -0.00001101, 0.00000375, -0.00000125, 0.00000150, -0.00000597, 0.00000385, 0.00000167, 0.00000340, -0.00001659, 0.00000526, -0.00000556, 0.00001019, -0.00001110, 0.00001143, -0.00003085, -0.00000275, -0.00000577, 0.00002906, 0.00000814, -0.00000035, 0.00001340, 0.00000331, 0.00000809, 0.00001673, -0.00000226, 0.00001763, 0.00001983, -0.00000037, -0.00000676, -0.00002244, 0.00000202, -0.00001303, 0.00000259, -0.00000010, -0.00000004, 0.00000128, -0.00000995, 0.00001197, -0.00001065, -0.00000965, 0.00000298, 0.00000804, -0.00000901, -0.00000214, -0.00000432, -0.00000224, -0.00001125, 0.00000276, -0.00000747, 0.00000688, -0.00000724, 0.00000470, -0.00001199, 0.00001039, -0.00002331, 0.00000905, -0.00002750, 0.00000623, -0.00005118, 0.00000926, -0.00016094, -0.00000424, 0.00020928, 0.00007050, 0.00008197, 0.00021128, -0.00002945, -0.00015106, 0.00000822, -0.00005209, 0.00001390, -0.00002263, 0.00001210, -0.00002481, 0.00000709, -0.00002067, 0.00000767, -0.00001488, 0.00001417, -0.00001394, 0.00000800, -0.00001320, 0.00001851, -0.00000791, 0.00000778, -0.00000950, 0.00000554, -0.00001999, 0.00002313, -0.00001260, 0.00001232, -0.00000215, 0.00001338, -0.00002593, 0.00001533, 0.00001071, -0.00001394, -0.00000296, -0.00001675, 0.00000388, -0.00000137, -0.00000466, -0.00000003, -0.00000280, -0.00000314, -0.00000373, -0.00000234, 0.00000164, 0.00000121, 0.00000071, 0.00000089, -0.00000689, 0.00000618, -0.00001880, 0.00000176, },
		{ 0.00000216, -0.00001113, 0.00001906, -0.00002432, 0.00001240, -0.00000052, 0.00000822, -0.00001107, 0.00000607, -0.00000257, 0.00000431, -0.00000540, 0.00000980, -0.00000143, -0.00000390, 0.00000318, -0.00000097, -0.00000314, -0.00000021, 0.00000354, 0.00000437, -0.00001289, -0.00000187, -0.00000107, 0.00000481, -0.00000347, 0.00000265, -0.00001162, -0.00002292, 0.00000127, 0.00001009, 0.00002956, -0.00000959, 0.00001744, 0.00000278, 0.00000820, 0.00001345, 0.00000727, 0.00000435, 0.00003112, -0.00000184, 0.00000078, -0.00002711, 0.00000041, -0.00001088, -0.00000236, 0.00000171, -0.00000026, 0.00000192, -0.00000823, 0.00000377, 0.00000256, -0.00001634, 0.00000071, 0.00000700, -0.00000246, -0.00000691, -0.00000202, -0.00000207, -0.00001109, 0.00000036, -0.00000543, 0.00000313, -0.00000302, 0.00000199, -0.00000886, 0.00000631, -0.00001524, 0.00000059, -0.00001681, -0.00000205, -0.00003610, -0.00000526, -0.00009842, -0.00010300, 0.00020343, 0.00009228, 0.00006315, 0.00015525, 0.00012104, -0.00018319, -0.00001267, -0.00004590, 0.00000233, -0.00001868, 0.00001013, -0.00002059, 0.00000286, -0.00001716, 0.00000341, -0.00001262, 0.00001019, -0.00000935, 0.00000508, -0.00001137, 0.00001271, -0.00000159, 0.00000446, -0.00000644, 0.00000448, -0.00001610, 0.00001131, 0.00000088, -0.00000014, 0.00000514, 0.00000925, -0.00001083, -0.00000823, 0.00002553, -0.00001558, 0.00000036, -0.00001952, 0.00000044, 0.00000129, -0.00000459, -0.00000225, -0.00000045, -0.00000426, -0.00000230, -0.00000436, 0.00000103, 0.00000191, -0.00000035, 0.00000333, -0.00000739, 0.00000412, -0.00000908, -0.00001233, },
		{ 0.00000766, -0.00001517, 0.00001660, -0.00001341, -0.00000413, 0.00001092, 0.00000161, -0.00000218, -0.00000400, 0.00000607, -0.00000306, 0.00000197, -0.00000054, 0.00000995, -0.00001103, 0.00000751, -0.00000483, 0.00000216, -0.00000662, 0.00000774, 0.00000087, -0.00000345, -0.00001289, 0.00000656, -0.00000448, 0.00000764, -0.00000896, 0.00000559, -0.00003266, -0.00000043, -0.00000193, 0.00003595, -0.00000254, 0.00001043, 0.00000850, 0.00000642, 0.00001007, 0.00001532, -0.00000348, 0.00003338, 0.00000442, 0.00000258, -0.00002465, -0.00000680, -0.00000611, -0.00000677, 0.00000091, 0.00000161, 0.00000033, -0.00000348, -0.00000510, 0.00001008, -0.00001421, -0.00000565, 0.00000581, 0.00000244, -0.00000790, -0.00000292, -0.00000099, -0.00000908, -0.00000504, 0.00000002, -0.00000405, 0.00000391, -0.00000392, -0.00000127, -0.00000319, -0.00000196, -0.00001272, -0.00000132, -0.00001688, -0.00001254, -0.00003220, -0.00003383, -0.00016589, 0.00016024, 0.00011925, 0.00006561, 0.00008855, 0.00019712, -0.00010793, -0.00009009, -0.00001167, -0.00002923, 0.00000022, -0.00000754, -0.00000152, -0.00001395, -0.00000243, -0.00001071, -0.00000115, -0.00000261, 0.00000287, -0.00000527, -0.00000109, -0.00000100, 0.00001068, -0.00000467, 0.00000322, -0.00000453, -0.00000339, -0.00000606, 0.00001505, -0.00001004, 0.00001005, 0.00000305, 0.00000573, -0.00002399, 0.00002713, -0.00000886, -0.00000015, -0.00001758, -0.00000630, 0.00000377, -0.00000321, -0.00000483, 0.00000097, -0.00000445, -0.00000151, -0.00000507, -0.00000049, 0.00000210, 0.00000019, 0.00000212, -0.00000232, -0.00000418, 0.00000425, -0.00002354, },
		{ -0.00000276, -0.00000394, 0.00000053, 0.00000644, -0.00002197, 0.00001823, -0.00000154, 0.00000454, -0.00001086, 0.00001053, -0.00000624, 0.00000566, -0.00000637, 0.00001435, -0.00001074, 0.00000631, -0.00000470, 0.00000377, -0.00000853, 0.00000729, 0.00000061, 0.00000215, -0.00001697, 0.00000633, -0.00000586, 0.00000907, -0.00000886, 0.00000648, -0.00002294, -0.00001466, -0.00000309, 0.00002580, 0.00001597, -0.00000014, 0.00001435, 0.00000573, 0.00000739, 0.00001863, -0.00000271, 0.00002574, 0.00001706, 0.00000027, -0.00001696, -0.00001686, -0.00000117, -0.00001012, -0.00000020, 0.00000135, 0.00000157, -0.00000227, -0.00000689, 0.00000660, -0.00000323, -0.00001394, 0.00000467, 0.00000390, -0.00000458, -0.00000591, -0.00000032, -0.00000546, -0.00001085, 0.00000430, -0.00000860, 0.00000688, -0.00000609, 0.00000290, -0.00000876, 0.00000583, -0.00001916, 0.00000628, -0.00002377, 0.00000160, -0.00004692, 0.00000171, -0.00017075, 0.00007396, 0.00015772, 0.00006103, 0.00006325, 0.00017229, 0.00003777, -0.00016264, 0.00000373, -0.00004752, 0.00000882, -0.00001969, 0.00001208, -0.00002417, 0.00000595, -0.00001861, 0.00000487, -0.00001061, 0.00001012, -0.00001135, 0.00000586, -0.00001040, 0.00001712, -0.00000858, 0.00000832, -0.00000926, 0.00000405, -0.00001424, 0.00001719, -0.00000832, 0.00000665, 0.00000136, 0.00001427, -0.00002444, 0.00001484, 0.00000342, -0.00000341, -0.00001109, -0.00001518, 0.00000543, -0.00000174, -0.00000552, -0.00000043, -0.00000242, -0.00000255, -0.00000366, -0.00000306, 0.00000201, 0.00000130, -0.00000009, 0.00000284, -0.00000959, 0.00000919, -0.00002041, },
		{ -0.00001972, 0.00000846, -0.00001219, 0.00001710, -0.00002507, 0.00001047, 0.00000548, 0.00000343, -0.00000904, 0.00000615, -0.00000147, 0.00000219, -0.00000422, 0.00000949, -0.00000251, -0.00000054, 0.00000044, -0.00000025, -0.00000374, 0.00000028, 0.00000589, 0.00000066, -0.00001131, -0.00000328, 0.00000161, 0.00000059, 0.00000152, -0.00000394, -0.00000479, -0.00002831, -0.00000376, 0.00001125, 0.00003066, -0.00000184, 0.00001304, 0.00000911, 0.00000509, 0.00001755, 0.00000417, 0.00001360, 0.00002940, -0.00000045, -0.00000914, -0.00002482, 0.00000056, -0.00001137, -0.00000094, -0.00000167, 0.00000540, -0.00000407, -0.00000350, -0.00000199, 0.00000730, -0.00001650, -0.00000009, 0.00000465, -0.00000111, -0.00000675, -0.00000239, -0.00000072, -0.00001444, 0.00000396, -0.00000720, 0.00000332, -0.00000240, 0.00000113, -0.00000721, 0.00000467, -0.00001528, 0.00000243, -0.00001830, 0.00000015, -0.00004075, 0.00000130, -0.00012606, -0.00003250, 0.00018960, 0.00005349, 0.00007186, 0.00009944, 0.00015970, -0.00015518, -0.00003644, -0.00003274, -0.00000696, -0.00001367, 0.00000751, -0.00001663, -0.00000134, -0.00001246, -0.00000130, -0.00000754, 0.00000654, -0.00000751, 0.00000337, -0.00000915, 0.00001223, -0.00000242, 0.00000394, -0.00000443, 0.00000093, -0.00000974, 0.00000737, 0.00000191, -0.00000123, 0.00000286, 0.00001436, -0.00001308, -0.00000292, 0.00001390, -0.00000600, -0.00000286, -0.00002264, 0.00000425, -0.00000006, -0.00000427, -0.00000375, 0.00000039, -0.00000414, -0.00000154, -0.00000528, 0.00000128, 0.00000199, -0.00000078, 0.00000400, -0.00000757, 0.00000322, -0.00000582, },
		{ -0.00002743, 0.00000725, -0.00000963, 0.00001072, -0.00001135, -0.00000847, 0.00001657, -0.00000054, -0.00000281, -0.00000224, 0.00000599, -0.00000337, 0.00000083, 0.00000149, 0.00000678, -0.00000691, 0.00000558, -0.00000523, 0.00000253, -0.00000734, 0.00001079, -0.00000198, -0.00000333, -0.00001307, 0.00000787, -0.00000650, 0.00000901, -0.00000995, 0.00000414, -0.00002724, -0.00001305, 0.00000247, 0.00003158, 0.00000923, 0.00000548, 0.00001422, 0.00000458, 0.00001356, 0.00001238, 0.00000389, 0.00003482, 0.00000500, -0.00000497, -0.00002719, -0.00000298, -0.00000959, -0.00000263, -0.00000423, 0.00000696, -0.00000337, -0.00000198, -0.00000620, 0.00000843, -0.00000902, -0.00000911, 0.00000608, -0.00000021, -0.00000382, -0.00000649, 0.00000286, -0.00001374, -0.00000139, -0.00000142, -0.00000347, 0.00000307, -0.00000257, -0.00000269, -0.00000059, -0.00000686, -0.00000626, -0.00000755, -0.00000915, -0.00002331, -0.00001798, -0.00006529, -0.00011758, 0.00018291, 0.00006715, 0.00007604, 0.00004998, 0.00018892, -0.00005200, -0.00011118, -0.00000692, -0.00003083, -0.00000104, -0.00000629, -0.00000049, -0.00001489, -0.00000131, -0.00001133, -0.00000075, -0.00000128, 0.00000035, -0.00000345, -0.00000238, 0.00000195, 0.00000750, -0.00000381, 0.00000369, -0.00000569, -0.00000161, -0.00000340, 0.00000961, -0.00000492, 0.00000238, 0.00001097, 0.00000108, -0.00001573, 0.00001644, -0.00000528, 0.00000350, -0.00002452, -0.00000183, 0.00000264, -0.00000281, -0.00000615, 0.00000113, -0.00000396, -0.00000092, -0.00000543, -0.00000072, 0.00000222, 0.00000014, 0.00000186, -0.00000127, -0.00000609, 0.00000733, },
		{ -0.00001949, -0.00000796, 0.00000313, -0.00000318, 0.00000548, -0.00002304, 0.00001872, 0.00000175, 0.00000002, -0.00000668, 0.00000846, -0.00000411, 0.00000209, -0.00000246, 0.00001014, -0.00000710, 0.00000575, -0.00000580, 0.00000404, -0.00000881, 0.00000887, 0.00000069, -0.00000083, -0.00001451, 0.00000478, -0.00000411, 0.00000559, -0.00000414, -0.00000104, -0.00001288, -0.00002605, -0.00000269, 0.00002229, 0.00002426, 0.00000099, 0.00001482, 0.00000821, 0.00000804, 0.00001829, 0.00000044, 0.00003157, 0.00001595, -0.00000363, -0.00002455, -0.00001052, -0.00000579, -0.00000587, -0.00000381, 0.00000326, 0.00000228, -0.00000507, -0.00000352, 0.00000097, 0.00000309, -0.00001630, 0.00000500, 0.00000002, -0.00000036, -0.00000852, 0.00000213, -0.00000825, -0.00000924, 0.00000418, -0.00000800, 0.00000503, -0.00000336, -0.00000032, -0.00000393, -0.00000091, -0.00001183, -0.00000041, -0.00001567, -0.00000878, -0.00003419, -0.00002099, -0.00015145, 0.00012357, 0.00010963, 0.00005758, 0.00004972, 0.00013483, 0.00008474, -0.00015291, -0.00001291, -0.00003550, -0.00000176, -0.00001214, 0.00000733, -0.00001904, 0.00000113, -0.00001334, -0.00000045, -0.00000411, 0.00000361, -0.00000637, 0.00000132, -0.00000446, 0.00001209, -0.00000621, 0.00000638, -0.00000683, 0.00000062, -0.00000613, 0.00000832, -0.00000137, -0.00000105, 0.00000702, 0.00001164, -0.00001865, 0.00001086, -0.00000183, 0.00000595, -0.00001899, -0.00001225, 0.00000572, -0.00000211, -0.00000595, -0.00000125, -0.00000157, -0.00000216, -0.00000352, -0.00000350, 0.00000203, 0.00000136, -0.00000038, 0.00000339, -0.00001010, 0.00000942, },
		{ -0.00000368, -0.00002372, 0.00001056, -0.00000906, 0.00001076, -0.00002193, 0.00000660, 0.00001154, -0.00000169, -0.00000535, 0.00000461, 0.00000056, -0.00000094, -0.00000153, 0.00000737, -0.00000191, 0.00000169, -0.00000197, 0.00000035, -0.00000405, 0.00000106, 0.00000751, -0.00000313, -0.00000896, -0.00000469, 0.00000362, -0.00000314, 0.00000569, -0.00000922, -0.00000021, -0.00002958, -0.00001218, 0.00001222, 0.00003145, 0.00000719, 0.00000859, 0.00001489, 0.00000364, 0.00001988, 0.00000312, 0.00002276, 0.00002753, -0.00000115, -0.00002016, -0.00001852, -0.00000275, -0.00000921, -0.00000148, -0.00000325, 0.00000880, -0.00000818, 0.00000016, -0.00000611, 0.00000942, -0.00001440, -0.00000182, 0.00000233, -0.00000009, -0.00000563, -0.00000314, -0.00000065, -0.00001527, 0.00000536, -0.00000720, 0.00000199, -0.00000067, -0.00000082, -0.00000407, 0.00000079, -0.00001214, 0.00000056, -0.00001563, -0.00000328, -0.00003729, -0.00000485, -0.00013510, 0.00003115, 0.00015687, 0.00003518, 0.00007107, 0.00006304, 0.00016942, -0.00011032, -0.00006407, -0.00001845, -0.00001745, -0.00000700, 0.00000256, -0.00001067, -0.00000681, -0.00000670, -0.00000674, -0.00000156, 0.00000147, -0.00000416, -0.00000005, -0.00000461, 0.00000944, -0.00000192, 0.00000259, -0.00000180, -0.00000309, -0.00000249, 0.00000219, 0.00000379, -0.00000282, 0.00000173, 0.00001720, -0.00001331, 0.00000174, 0.00000181, 0.00000455, -0.00000793, -0.00002290, 0.00000626, -0.00000111, -0.00000417, -0.00000482, 0.00000080, -0.00000338, -0.00000148, -0.00000535, 0.00000095, 0.00000208, -0.00000080, 0.00000378, -0.00000688, 0.00000199, },
		{ 0.00000719, -0.00002692, 0.00000337, -0.00000204, 0.00000353, -0.00000921, -0.00001091, 0.00001948, -0.00000138, -0.00000329, -0.00000028, 0.00000510, -0.00000325, -0.00000030, 0.00000355, 0.00000316, -0.00000172, 0.00000191, -0.00000384, 0.00000123, -0.00000587, 0.00001160, -0.00000359, -0.00000420, -0.00001126, 0.00000678, -0.00000643, 0.00000795, -0.00000824, -0.00000025, -0.00001880, -0.00002600, 0.00000426, 0.00002782, 0.00002114, 0.00000172, 0.00001879, 0.00000414, 0.00001643, 0.00000962, 0.00001312, 0.00003483, 0.00000535, -0.00001622, -0.00002428, -0.00000274, -0.00001065, -0.00000031, -0.00000750, 0.00001016, -0.00000563, -0.00000123, -0.00000605, 0.00000542, -0.00000373, -0.00001141, 0.00000500, -0.00000218, -0.00000021, -0.00000857, 0.00000393, -0.00001554, 0.00000071, -0.00000215, -0.00000323, 0.00000227, -0.00000150, -0.00000331, 0.00000056, -0.00000973, -0.00000184, -0.00001171, -0.00000432, -0.00003016, -0.00000895, -0.00009273, -0.00005635, 0.00017362, 0.00003805, 0.00007568, 0.00002990, 0.00016459, 0.00000102, -0.00012134, -0.00000916, -0.00002874, -0.00000415, -0.00000414, -0.00000072, -0.00001421, -0.00000147, -0.00001083, -0.00000086, -0.00000009, -0.00000197, -0.00000199, -0.00000302, 0.00000475, 0.00000346, -0.00000184, 0.00000309, -0.00000591, -0.00000020, -0.00000075, 0.00000395, 0.00000037, -0.00000462, 0.00001772, -0.00000353, -0.00000619, 0.00000395, 0.00000049, 0.00000413, -0.00002823, 0.00000129, 0.00000157, -0.00000285, -0.00000674, 0.00000064, -0.00000270, -0.00000105, -0.00000507, -0.00000125, 0.00000223, 0.00000021, 0.00000159, -0.00000093, -0.00000627, },
		{ 0.00000724, -0.00001663, -0.00001277, 0.00000815, -0.00000447, 0.00000142, -0.00002004, 0.00001590, 0.00000597, -0.00000429, -0.00000190, 0.00000533, -0.00000143, -0.00000180, 0.00000213, 0.00000470, -0.00000168, 0.00000297, -0.00000488, 0.00000253, -0.00000709, 0.00000890, 0.00000127, -0.00000438, -0.00001055, 0.00000183, -0.00000156, 0.00000123, 0.00000111, -0.00000849, -0.00000391, -0.00003370, -0.00000653, 0.00001935, 0.00003229, 0.00000343, 0.00001486, 0.00001079, 0.00000969, 0.00001656, 0.00000643, 0.00003576, 0.00001543, -0.00001220, -0.00002739, -0.00000597, -0.00000973, -0.00000196, -0.00000698, 0.00000484, 0.00000227, -0.00000660, -0.00000105, -0.00000329, 0.00000682, -0.00001577, 0.00000319, -0.00000253, 0.00000247, -0.00000889, 0.00000204, -0.00000962, -0.00000732, 0.00000319, -0.00000655, 0.00000246, -0.00000047, -0.00000315, 0.00000013, -0.00000682, -0.00000478, -0.00000723, -0.00000750, -0.00001966, -0.00001969, -0.00005031, -0.00010995, 0.00014338, 0.00007267, 0.00005487, 0.00004064, 0.00010224, 0.00011135, -0.00012650, -0.00003574, -0.00002144, -0.00001313, -0.00000392, 0.00000118, -0.00001243, -0.00000456, -0.00000752, -0.00000568, 0.00000196, -0.00000273, -0.00000148, -0.00000323, 0.00000224, 0.00000552, -0.00000251, 0.00000347, -0.00000382, -0.00000265, 0.00000146, -0.00000027, 0.00000520, -0.00000788, 0.00001222, 0.00000782, -0.00001076, 0.00000466, -0.00000443, 0.00001220, -0.00002463, -0.00000918, 0.00000516, -0.00000250, -0.00000614, -0.00000214, -0.00000057, -0.00000189, -0.00000332, -0.00000363, 0.00000169, 0.00000141, -0.00000024, 0.00000273, -0.00000871, },
		{ 0.00000055, -0.00000280, -0.00002420, 0.00000876, -0.00000300, 0.00000177, -0.00001631, 0.00000208, 0.00001599, -0.00000515, -0.00000191, 0.00000293, 0.00000227, -0.00000389, 0.00000158, 0.00000431, -0.00000007, 0.00000254, -0.00000348, 0.00000052, -0.00000404, 0.00000227, 0.00000783, -0.00000609, -0.00000669, -0.00000548, 0.00000437, -0.00000559, 0.00000798, -0.00001198, 0.00000120, -0.00002679, -0.00002243, 0.00001100, 0.00003342, 0.00001591, 0.00000567, 0.00001901, 0.00000451, 0.00002008, 0.00000484, 0.00003093, 0.00002633, -0.00000631, -0.00002849, -0.00001138, -0.00000752, -0.00000559, -0.00000339, -0.00000309, 0.00000950, -0.00000951, 0.00000159, -0.00000766, 0.00000862, -0.00001025, -0.00000435, 0.00000023, 0.00000044, -0.00000360, -0.00000423, -0.00000183, -0.00001368, 0.00000470, -0.00000580, -0.00000022, 0.00000128, -0.00000271, -0.00000092, -0.00000354, -0.00000762, -0.00000332, -0.00001060, -0.00001026, -0.00002810, -0.00002178, -0.00012337, 0.00007798, 0.00011765, 0.00003058, 0.00006256, 0.00004334, 0.00015772, -0.00005859, -0.00008759, -0.00000850, -0.00002503, -0.00000184, -0.00000205, -0.00000512, -0.00001129, -0.00000197, -0.00001089, 0.00000331, -0.00000314, -0.00000098, -0.00000358, 0.00000072, 0.00000557, -0.00000096, 0.00000133, 0.00000027, -0.00000595, 0.00000358, -0.00000240, 0.00000532, -0.00000420, 0.00000151, 0.00001759, -0.00001097, 0.00000406, -0.00000766, 0.00001288, -0.00001273, -0.00002111, 0.00000651, -0.00000167, -0.00000447, -0.00000530, 0.00000080, -0.00000225, -0.00000182, -0.00000480, 0.00000022, 0.00000209, -0.00000044, 0.00000282, -0.00000551, },
		{ -0.00000520, 0.00000453, -0.00002294, -0.00000253, 0.00000568, -0.00000333, -0.00000779, -0.00001088, 0.00001887, -0.00000011, -0.00000404, 0.00000177, 0.00000376, -0.00000308, -0.00000086, 0.00000478, 0.00000036, 0.00000287, -0.00000194, -0.00000168, -0.00000103, -0.00000277, 0.00001047, -0.00000416, -0.00000550, -0.00000835, 0.00000424, -0.00000515, 0.00000536, -0.00000510, -0.00000562, -0.00001026, -0.00003574, 0.00000066, 0.00002643, 0.00003138, 0.00000070, 0.00002124, 0.00000585, 0.00001786, 0.00000849, 0.00002254, 0.00003516, 0.00000230, -0.00002747, -0.00001804, -0.00000543, -0.00000928, -0.00000023, -0.00000831, 0.00001056, -0.00000580, -0.00000160, -0.00000484, 0.00000191, 0.00000065, -0.00001195, 0.00000226, -0.00000293, 0.00000224, -0.00000864, 0.00000190, -0.00001423, 0.00000098, -0.00000195, -0.00000348, 0.00000167, -0.00000085, -0.00000294, -0.00000005, -0.00000989, -0.00000059, -0.00001238, -0.00000413, -0.00003064, -0.00000941, -0.00010716, 0.00000358, 0.00014455, 0.00002714, 0.00006635, 0.00002352, 0.00013220, 0.00004462, -0.00011856, -0.00001799, -0.00002384, -0.00000816, -0.00000192, -0.00000151, -0.00001255, -0.00000235, -0.00000971, -0.00000106, 0.00000048, -0.00000353, -0.00000142, -0.00000242, 0.00000653, -0.00000045, 0.00000048, 0.00000187, -0.00000532, 0.00000099, 0.00000113, -0.00000047, 0.00000400, -0.00000885, 0.00002103, -0.00000556, 0.00000135, -0.00000680, 0.00000599, 0.00000259, -0.00002845, 0.00000251, 0.00000086, -0.00000340, -0.00000673, -0.00000013, -0.00000122, -0.00000147, -0.00000431, -0.00000185, 0.00000200, 0.00000045, 0.00000125, -0.00000104, },
		{ -0.00000612, 0.00000344, -0.00001267, -0.00001630, 0.00001049, -0.00000358, -0.00000464, -0.00001412, 0.00001060, 0.00001030, -0.00000720, 0.00000190, 0.00000275, 0.00000027, -0.00000440, 0.00000560, 0.00000005, 0.00000345, -0.00000044, -0.00000293, 0.00000004, -0.00000414, 0.00000806, 0.00000150, -0.00000721, -0.00000664, -0.00000101, 0.00000037, -0.00000242, 0.00000493, -0.00001348, 0.00000139, -0.00003558, -0.00001521, 0.00001688, 0.00003968, 0.00000738, 0.00001460, 0.00001357, 0.00001176, 0.00001466, 0.00001400, 0.00003965, 0.00001310, -0.00002355, -0.00002522, -0.00000465, -0.00001145, 0.00000020, -0.00000861, 0.00000529, 0.00000209, -0.00000690, 0.00000002, -0.00000540, 0.00000748, -0.00001240, -0.00000067, -0.00000350, 0.00000361, -0.00000695, -0.00000055, -0.00000942, -0.00000550, 0.00000184, -0.00000489, -0.00000003, 0.00000163, -0.00000446, 0.00000196, -0.00000992, -0.00000053, -0.00001131, -0.00000273, -0.00002654, -0.00000991, -0.00007634, -0.00005757, 0.00013888, 0.00004882, 0.00005041, 0.00003515, 0.00007590, 0.00012045, -0.00009011, -0.00005821, -0.00001031, -0.00002132, 0.00000167, -0.00000353, -0.00000693, -0.00000873, -0.00000331, -0.00000886, 0.00000561, -0.00000694, 0.00000148, -0.00000587, 0.00000759, -0.00000056, 0.00000089, 0.00000093, -0.00000149, -0.00000427, 0.00000641, -0.00000606, 0.00000893, -0.00001149, 0.00001476, 0.00000503, -0.00000320, -0.00000194, -0.00000462, 0.00001429, -0.00002663, -0.00000698, 0.00000414, -0.00000291, -0.00000628, -0.00000271, 0.00000022, -0.00000157, -0.00000308, -0.00000351, 0.00000104, 0.00000144, 0.00000016, 0.00000132, },
		{ -0.00000373, -0.00000103, -0.00000290, -0.00002181, 0.00000431, 0.00000397, -0.00000760, -0.00000987, -0.00000200, 0.00001777, -0.00000563, -0.00000019, 0.00000239, 0.00000257, -0.00000553, 0.00000392, 0.00000151, 0.00000212, 0.00000232, -0.00000400, 0.00000019, -0.00000372, 0.00000396, 0.00000658, -0.00000755, -0.00000519, -0.00000523, 0.00000356, -0.00000638, 0.00000789, -0.00001182, -0.00000047, -0.00002103, -0.00003239, 0.00000589, 0.00003725, 0.00002357, 0.00000512, 0.00002120, 0.00000746, 0.00001877, 0.00000906, 0.00003830, 0.00002501, -0.00001639, -0.00003149, -0.00000641, -0.00001135, -0.00000231, -0.00000545, -0.00000205, 0.00000810, -0.00000828, 0.00000076, -0.00000672, 0.00000548, -0.00000497, -0.00000709, -0.00000179, 0.00000068, -0.00000109, -0.00000538, -0.00000403, -0.00001020, 0.00000259, -0.00000368, -0.00000252, 0.00000261, -0.00000354, 0.00000095, -0.00000664, -0.00000373, -0.00000702, -0.00000578, -0.00001748, -0.00001761, -0.00004387, -0.00009429, 0.00010372, 0.00008164, 0.00003221, 0.00005060, 0.00003474, 0.00013402, -0.00000977, -0.00010068, -0.00000650, -0.00002716, -0.00000028, -0.00000446, -0.00000172, -0.00001323, 0.00000033, -0.00001245, 0.00000572, -0.00000595, 0.00000073, -0.00000579, 0.00000536, 0.00000179, -0.00000032, 0.00000078, 0.00000106, -0.00000659, 0.00000702, -0.00000498, 0.00000545, -0.00000460, 0.00000164, 0.00001599, -0.00000633, 0.00000337, -0.00001260, 0.00001671, -0.00001549, -0.00001824, 0.00000525, -0.00000175, -0.00000523, -0.00000511, 0.00000050, -0.00000107, -0.00000219, -0.00000398, -0.00000063, 0.00000188, 0.00000020, 0.00000140, },
		{ -0.00000129, -0.00000362, 0.00000054, -0.00001702, -0.00000866, 0.00001166, -0.00000857, -0.00000732, -0.00000868, 0.00001492, 0.00000337, -0.00000524, 0.00000385, 0.00000218, -0.00000303, -0.00000069, 0.00000488, -0.00000092, 0.00000566, -0.00000466, 0.00000013, -0.00000329, 0.00000107, 0.00000803, -0.00000412, -0.00000660, -0.00000531, 0.00000132, -0.00000366, 0.00000234, -0.00000193, -0.00001004, -0.00000411, -0.00003990, -0.00000930, 0.00002742, 0.00003951, 0.00000290, 0.00002160, 0.00000951, 0.00001791, 0.00000957, 0.00003144, 0.00003632, -0.00000635, -0.00003493, -0.00001159, -0.00000918, -0.00000655, -0.00000177, -0.00000710, 0.00000848, -0.00000402, -0.00000298, -0.00000304, -0.00000130, 0.00000342, -0.00001062, -0.00000194, -0.00000243, 0.00000340, -0.00000692, -0.00000253, -0.00001056, -0.00000008, -0.00000121, -0.00000389, 0.00000099, -0.00000033, -0.00000209, -0.00000174, -0.00000800, -0.00000199, -0.00001012, -0.00000791, -0.00002542, -0.00001952, -0.00010425, 0.00005093, 0.00010854, 0.00002664, 0.00005284, 0.00002447, 0.00009936, 0.00007453, -0.00010324, -0.00003150, -0.00001760, -0.00001206, -0.00000031, -0.00000234, -0.00001041, -0.00000344, -0.00000833, -0.00000116, 0.00000032, -0.00000423, -0.00000171, -0.00000053, 0.00000685, -0.00000350, 0.00000246, 0.00000046, -0.00000405, 0.00000188, 0.00000202, -0.00000303, 0.00000519, -0.00000946, 0.00001999, -0.00000383, 0.00000494, -0.00001329, 0.00000918, -0.00000001, -0.00002551, 0.00000170, 0.00000064, -0.00000440, -0.00000627, -0.00000079, -0.00000004, -0.00000172, -0.00000349, -0.00000231, 0.00000146, 0.00000085, 0.00000079, },
		{ -0.00000030, -0.00000322, -0.00000094, -0.00000850, -0.00001800, 0.00001016, -0.00000148, -0.00001091, -0.00000697, 0.00000400, 0.00001402, -0.00000809, 0.00000386, 0.00000161, 0.00000021, -0.00000486, 0.00000687, -0.00000241, 0.00000664, -0.00000297, -0.00000108, -0.00000244, -0.00000087, 0.00000699, 0.00000081, -0.00000842, -0.00000395, -0.00000265, 0.00000070, -0.00000433, 0.00000612, -0.00001472, 0.00000192, -0.00003153, -0.00002747, 0.00001362, 0.00004657, 0.00001278, 0.00001438, 0.00001649, 0.00001388, 0.00001366, 0.00002211, 0.00004426, 0.00000629, -0.00003431, -0.00001971, -0.00000637, -0.00001058, 0.00000012, -0.00000826, 0.00000420, 0.00000221, -0.00000639, -0.00000019, -0.00000544, 0.00000556, -0.00000715, -0.00000567, -0.00000335, 0.00000351, -0.00000354, -0.00000465, -0.00000823, -0.00000371, 0.00000029, -0.00000326, -0.00000211, 0.00000266, -0.00000408, 0.00000140, -0.00000978, 0.00000016, -0.00001162, -0.00000270, -0.00002730, -0.00000874, -0.00009108, -0.00000631, 0.00011920, 0.00003584, 0.00004413, 0.00003202, 0.00005590, 0.00011646, -0.00005074, -0.00007495, -0.00000532, -0.00002421, 0.00000302, -0.00000533, -0.00000390, -0.00001019, -0.00000168, -0.00000920, 0.00000613, -0.00000830, 0.00000183, -0.00000572, 0.00001036, -0.00000484, 0.00000297, -0.00000053, -0.00000029, -0.00000382, 0.00000796, -0.00000803, 0.00000883, -0.00001094, 0.00001362, 0.00000454, 0.00000233, -0.00000732, -0.00000314, 0.00001216, -0.00002451, -0.00000621, 0.00000294, -0.00000337, -0.00000649, -0.00000271, 0.00000050, -0.00000107, -0.00000281, -0.00000321, 0.00000022, 0.00000141, 0.00000062, },
		{ -0.00000012, -0.00000189, -0.00000268, -0.00000351, -0.00001766, -0.00000111, 0.00000966, -0.00001515, -0.00000404, -0.00000506, 0.00001682, -0.00000301, -0.00000062, 0.00000344, 0.00000111, -0.00000535, 0.00000474, -0.00000010, 0.00000368, 0.00000160, -0.00000371, -0.00000058, -0.00000307, 0.00000587, 0.00000393, -0.00000737, -0.00000471, -0.00000395, 0.00000138, -0.00000568, 0.00000567, -0.00000923, -0.00000424, -0.00001412, -0.00003989, -0.00000402, 0.00004247, 0.00003038, 0.00000689, 0.00002195, 0.00001162, 0.00001718, 0.00001485, 0.00004584, 0.00002127, -0.00002952, -0.00002870, -0.00000528, -0.00001259, -0.00000079, -0.00000649, -0.00000121, 0.00000574, -0.00000550, -0.00000153, -0.00000421, 0.00000127, 0.00000016, -0.00000926, -0.00000402, 0.00000104, 0.00000127, -0.00000620, -0.00000696, -0.00000564, -0.00000020, -0.00000153, -0.00000431, 0.00000283, -0.00000291, 0.00000099, -0.00000758, -0.00000181, -0.00000888, -0.00000332, -0.00002192, -0.00001045, -0.00006395, -0.00005535, 0.00010972, 0.00005419, 0.00003436, 0.00003930, 0.00003144, 0.00010682, 0.00002904, -0.00010044, -0.00001300, -0.00002367, -0.00000256, -0.00000424, -0.00000094, -0.00001230, -0.00000000, -0.00001127, 0.00000542, -0.00000660, 0.00000054, -0.00000601, 0.00000838, -0.00000112, -0.00000032, 0.00000098, 0.00000065, -0.00000503, 0.00000749, -0.00000506, 0.00000381, -0.00000365, 0.00000163, 0.00001314, -0.00000031, 0.00000014, -0.00001266, 0.00001525, -0.00001534, -0.00001499, 0.00000283, -0.00000146, -0.00000638, -0.00000428, -0.00000003, -0.00000009, -0.00000227, -0.00000322, -0.00000135, 0.00000141, 0.00000095, },
		{ 0.00000025, -0.00000142, -0.00000226, -0.00000352, -0.00001077, -0.00001344, 0.00001457, -0.00001159, -0.00000738, -0.00000541, 0.00000883, 0.00000833, -0.00000665, 0.00000575, 0.00000056, -0.00000304, -0.00000000, 0.00000407, -0.00000072, 0.00000623, -0.00000550, 0.00000087, -0.00000476, 0.00000475, 0.00000499, -0.00000385, -0.00000708, -0.00000289, -0.00000114, -0.00000274, -0.00000023, 0.00000022, -0.00001223, -0.00000168, -0.00003795, -0.00002436, 0.00002887, 0.00004643, 0.00000767, 0.00002087, 0.00001409, 0.00001757, 0.00001243, 0.00004034, 0.00003690, -0.00002053, -0.00003625, -0.00000770, -0.00001190, -0.00000402, -0.00000370, -0.00000499, 0.00000503, -0.00000115, -0.00000474, -0.00000129, -0.00000355, 0.00000433, -0.00000781, -0.00000694, -0.00000107, 0.00000353, -0.00000415, -0.00000801, -0.00000593, -0.00000149, -0.00000061, -0.00000396, -0.00000013, 0.00000047, -0.00000142, -0.00000341, -0.00000550, -0.00000442, -0.00000684, -0.00001318, -0.00001772, -0.00003551, -0.00008522, 0.00007993, 0.00007551, 0.00002922, 0.00003999, 0.00002712, 0.00007144, 0.00008959, -0.00007829, -0.00004653, -0.00001191, -0.00001487, 0.00000019, -0.00000283, -0.00000823, -0.00000438, -0.00000698, -0.00000113, -0.00000046, -0.00000428, -0.00000247, 0.00000221, 0.00000576, -0.00000527, 0.00000363, -0.00000072, -0.00000232, 0.00000231, 0.00000220, -0.00000391, 0.00000422, -0.00000688, 0.00001515, 0.00000121, 0.00000441, -0.00001477, 0.00000917, -0.00000282, -0.00002003, -0.00000089, 0.00000081, -0.00000573, -0.00000550, -0.00000115, 0.00000054, -0.00000151, -0.00000286, -0.00000250, 0.00000065, 0.00000129, },
		{ 0.00000100, -0.00000163, -0.00000080, -0.00000494, -0.00000489, -0.00001774, 0.00000770, 0.00000054, -0.00001549, -0.00000054, -0.00000256, 0.00001665, -0.00000701, 0.00000399, 0.00000189, -0.00000138, -0.00000338, 0.00000598, -0.00000241, 0.00000718, -0.00000379, -0.00000008, -0.00000421, 0.00000210, 0.00000599, -0.00000079, -0.00000789, -0.00000284, -0.00000282, -0.00000066, -0.00000440, 0.00000465, -0.00001239, -0.00000160, -0.00002340, -0.00004041, 0.00000752, 0.00005337, 0.00001935, 0.00001475, 0.00001922, 0.00001611, 0.00001400, 0.00003058, 0.00004933, -0.00000665, -0.00004098, -0.00001370, -0.00000958, -0.00000801, -0.00000167, -0.00000640, 0.00000187, 0.00000277, -0.00000546, -0.00000124, -0.00000405, 0.00000219, -0.00000152, -0.00001053, -0.00000291, 0.00000296, 0.00000008, -0.00000881, -0.00000704, -0.00000166, -0.00000149, -0.00000166, -0.00000384, 0.00000290, -0.00000264, -0.00000059, -0.00000746, -0.00000175, -0.00000911, -0.00000640, -0.00002279, -0.00001628, -0.00009071, 0.00003452, 0.00009382, 0.00002959, 0.00003729, 0.00002995, 0.00004153, 0.00010423, -0.00001430, -0.00008276, -0.00000726, -0.00002189, 0.00000043, -0.00000439, -0.00000330, -0.00000910, -0.00000233, -0.00000720, 0.00000405, -0.00000734, 0.00000010, -0.00000309, 0.00001038, -0.00000691, 0.00000346, -0.00000089, 0.00000005, -0.00000192, 0.00000678, -0.00000678, 0.00000556, -0.00000689, 0.00000936, 0.00000619, 0.00000532, -0.00001059, -0.00000090, 0.00000660, -0.00001872, -0.00000683, 0.00000164, -0.00000384, -0.00000676, -0.00000210, 0.00000020, -0.00000036, -0.00000251, -0.00000285, -0.00000062, 0.00000126, },
		{ 0.00000159, -0.00000137, -0.00000038, -0.00000426, -0.00000420, -0.00001289, -0.00000595, 0.00001234, -0.00001931, 0.00000029, -0.00000717, 0.00001404, 0.00000160, -0.00000244, 0.00000545, -0.00000151, -0.00000370, 0.00000409, -0.00000024, 0.00000407, 0.00000092, -0.00000288, -0.00000167, -0.00000197, 0.00000750, 0.00000057, -0.00000600, -0.00000493, -0.00000212, -0.00000151, -0.00000419, 0.00000220, -0.00000540, -0.00000847, -0.00000818, -0.00004316, -0.00001853, 0.00004742, 0.00003743, 0.00001029, 0.00002220, 0.00001591, 0.00001659, 0.00002137, 0.00005435, 0.00001229, -0.00004200, -0.00002201, -0.00000768, -0.00001101, -0.00000142, -0.00000604, -0.00000129, 0.00000352, -0.00000241, -0.00000412, -0.00000139, -0.00000263, 0.00000393, -0.00001025, -0.00000662, 0.00000178, 0.00000290, -0.00000635, -0.00001021, -0.00000088, -0.00000286, 0.00000013, -0.00000528, 0.00000195, -0.00000110, -0.00000042, -0.00000650, -0.00000206, -0.00000844, -0.00000399, -0.00002208, -0.00000960, -0.00007576, -0.00001544, 0.00010133, 0.00003635, 0.00003416, 0.00003109, 0.00002930, 0.00008190, 0.00005455, -0.00008782, -0.00002546, -0.00001675, -0.00000713, -0.00000241, -0.00000196, -0.00000936, -0.00000207, -0.00000826, 0.00000319, -0.00000569, -0.00000106, -0.00000439, 0.00000958, -0.00000300, -0.00000075, 0.00000144, -0.00000019, -0.00000223, 0.00000581, -0.00000323, 0.00000087, -0.00000159, 0.00000134, 0.00000970, 0.00000594, -0.00000435, -0.00000891, 0.00000929, -0.00001248, -0.00001168, -0.00000044, -0.00000094, -0.00000773, -0.00000295, -0.00000071, 0.00000059, -0.00000194, -0.00000269, -0.00000183, 0.00000070, },
		{ 0.00000157, -0.00000020, -0.00000138, -0.00000157, -0.00000670, -0.00000543, -0.00001584, 0.00001387, -0.00001245, -0.00000708, -0.00000267, 0.00000249, 0.00001332, -0.00000756, 0.00000704, -0.00000084, -0.00000301, 0.00000080, 0.00000294, 0.00000032, 0.00000498, -0.00000450, 0.00000031, -0.00000499, 0.00000753, 0.00000195, -0.00000359, -0.00000679, -0.00000139, -0.00000281, -0.00000268, -0.00000200, 0.00000093, -0.00001187, -0.00000291, -0.00003123, -0.00004135, 0.00002754, 0.00005384, 0.00001379, 0.00002048, 0.00001824, 0.00001816, 0.00001617, 0.00005049, 0.00003386, -0.00003756, -0.00003147, -0.00000767, -0.00001225, -0.00000292, -0.00000491, -0.00000324, 0.00000160, 0.00000168, -0.00000611, -0.00000020, -0.00000448, 0.00000358, -0.00000432, -0.00001179, 0.00000054, 0.00000312, -0.00000126, -0.00001302, -0.00000185, -0.00000227, -0.00000069, -0.00000337, -0.00000183, 0.00000171, -0.00000134, -0.00000419, -0.00000374, -0.00000618, -0.00000466, -0.00001708, -0.00001168, -0.00005138, -0.00005576, 0.00009058, 0.00005072, 0.00003024, 0.00003076, 0.00002811, 0.00005076, 0.00009173, -0.00004851, -0.00005925, -0.00000871, -0.00001577, -0.00000068, -0.00000272, -0.00000634, -0.00000487, -0.00000583, -0.00000098, -0.00000166, -0.00000401, -0.00000311, 0.00000515, 0.00000360, -0.00000567, 0.00000366, -0.00000125, -0.00000048, 0.00000227, 0.00000208, -0.00000378, 0.00000208, -0.00000249, 0.00000822, 0.00000778, 0.00000126, -0.00001222, 0.00000634, -0.00000545, -0.00001277, -0.00000475, 0.00000113, -0.00000716, -0.00000451, -0.00000122, 0.00000051, -0.00000083, -0.00000245, -0.00000247, -0.00000028, },
		{ 0.00000090, 0.00000121, -0.00000238, 0.00000060, -0.00000786, -0.00000227, -0.00001572, 0.00000393, 0.00000126, -0.00001681, 0.00000354, -0.00000779, 0.00001788, -0.00000449, 0.00000301, 0.00000279, -0.00000348, -0.00000108, 0.00000396, -0.00000077, 0.00000552, -0.00000293, -0.00000006, -0.00000505, 0.00000452, 0.00000491, -0.00000273, -0.00000626, -0.00000278, -0.00000213, -0.00000295, -0.00000346, 0.00000158, -0.00000799, -0.00000702, -0.00001440, -0.00005035, -0.00000352, 0.00006023, 0.00002693, 0.00001623, 0.00002139, 0.00001884, 0.00001548, 0.00004026, 0.00005278, -0.00002518, -0.00004128, -0.00000997, -0.00001220, -0.00000532, -0.00000390, -0.00000414, -0.00000083, 0.00000342, -0.00000432, -0.00000261, -0.00000208, -0.00000135, 0.00000304, -0.00001411, -0.00000286, 0.00000267, 0.00000279, -0.00001171, -0.00000664, 0.00000077, -0.00000342, -0.00000013, -0.00000530, 0.00000275, -0.00000101, -0.00000263, -0.00000472, -0.00000434, -0.00000593, -0.00001127, -0.00001615, -0.00002909, -0.00007632, 0.00006022, 0.00006992, 0.00002616, 0.00003125, 0.00002804, 0.00003158, 0.00008819, 0.00001508, -0.00008087, -0.00001473, -0.00001624, -0.00000435, -0.00000206, -0.00000398, -0.00000659, -0.00000409, -0.00000415, 0.00000071, -0.00000535, -0.00000238, 0.00000084, 0.00000834, -0.00000708, 0.00000270, -0.00000060, 0.00000024, 0.00000021, 0.00000442, -0.00000398, 0.00000092, -0.00000122, 0.00000370, 0.00000869, 0.00000644, -0.00001170, 0.00000133, -0.00000101, -0.00001051, -0.00000833, 0.00000011, -0.00000429, -0.00000697, -0.00000110, -0.00000054, 0.00000048, -0.00000212, -0.00000254, -0.00000133, },
		{ -0.00000012, 0.00000191, -0.00000203, 0.00000048, -0.00000559, -0.00000466, -0.00000841, -0.00000916, 0.00001113, -0.00001938, 0.00000280, -0.00000873, 0.00001079, 0.00000640, -0.00000410, 0.00000705, -0.00000394, -0.00000172, 0.00000278, 0.00000063, 0.00000334, 0.00000048, -0.00000171, -0.00000302, -0.00000024, 0.00000816, -0.00000250, -0.00000433, -0.00000513, -0.00000059, -0.00000416, -0.00000292, -0.00000129, -0.00000185, -0.00001139, -0.00000506, -0.00004132, -0.00003622, 0.00004966, 0.00004611, 0.00001455, 0.00002279, 0.00001969, 0.00001765, 0.00002865, 0.00006335, -0.00000384, -0.00004968, -0.00001483, -0.00001164, -0.00000782, -0.00000345, -0.00000456, -0.00000224, 0.00000201, 0.00000007, -0.00000602, 0.00000068, -0.00000515, 0.00000561, -0.00000991, -0.00000937, 0.00000287, 0.00000354, -0.00000575, -0.00001321, 0.00000310, -0.00000469, 0.00000095, -0.00000544, 0.00000039, 0.00000110, -0.00000223, -0.00000447, -0.00000360, -0.00000648, -0.00000701, -0.00001856, -0.00001534, -0.00007595, 0.00001786, 0.00008539, 0.00002624, 0.00003130, 0.00002638, 0.00002641, 0.00006191, 0.00006718, -0.00006679, -0.00003941, -0.00000985, -0.00001149, -0.00000076, -0.00000322, -0.00000595, -0.00000436, -0.00000490, 0.00000038, -0.00000442, -0.00000302, -0.00000170, 0.00000930, -0.00000409, -0.00000114, 0.00000145, -0.00000045, 0.00000040, 0.00000349, -0.00000082, -0.00000228, 0.00000082, 0.00000102, 0.00000608, 0.00001137, -0.00000850, -0.00000340, 0.00000080, -0.00000801, -0.00000823, -0.00000431, -0.00000045, -0.00000897, -0.00000139, -0.00000145, 0.00000100, -0.00000128, -0.00000237, -0.00000208, },
		{ -0.00000112, 0.00000149, -0.00000040, -0.00000124, -0.00000166, -0.00000853, -0.00000167, -0.00001559, 0.00000982, -0.00001168, -0.00000555, -0.00000190, -0.00000214, 0.00001661, -0.00000697, 0.00000712, -0.00000163, -0.00000297, 0.00000141, 0.00000208, 0.00000134, 0.00000279, -0.00000227, -0.00000123, -0.00000392, 0.00000900, -0.00000059, -0.00000347, -0.00000587, -0.00000072, -0.00000377, -0.00000330, -0.00000309, 0.00000041, -0.00000950, -0.00000659, -0.00002236, -0.00005618, 0.00002009, 0.00006310, 0.00002035, 0.00002148, 0.00002114, 0.00002053, 0.00002034, 0.00006271, 0.00002372, -0.00005293, -0.00002306, -0.00001085, -0.00001024, -0.00000351, -0.00000491, -0.00000261, -0.00000077, 0.00000348, -0.00000645, -0.00000010, -0.00000413, 0.00000173, -0.00000116, -0.00001548, 0.00000173, 0.00000265, 0.00000093, -0.00001630, 0.00000052, -0.00000189, -0.00000158, -0.00000221, -0.00000384, 0.00000306, -0.00000167, -0.00000388, -0.00000337, -0.00000633, -0.00000482, -0.00001772, -0.00001045, -0.00006132, -0.00002358, 0.00008724, 0.00003488, 0.00002829, 0.00002560, 0.00002669, 0.00003685, 0.00008502, -0.00001938, -0.00006617, -0.00000940, -0.00001437, -0.00000294, -0.00000197, -0.00000490, -0.00000477, -0.00000498, -0.00000078, -0.00000308, -0.00000371, -0.00000317, 0.00000758, 0.00000089, -0.00000485, 0.00000249, -0.00000083, 0.00000099, 0.00000196, 0.00000191, -0.00000335, -0.00000009, 0.00000207, 0.00000125, 0.00001362, -0.00000219, -0.00000777, 0.00000196, -0.00000784, -0.00000466, -0.00000913, 0.00000117, -0.00000838, -0.00000341, -0.00000116, 0.00000014, 0.00000011, -0.00000215, -0.00000236, },
		{ -0.00000184, 0.00000032, 0.00000128, -0.00000250, 0.00000081, -0.00000942, -0.00000078, -0.00001243, -0.00000033, -0.00000006, -0.00001411, 0.00000426, -0.00001074, 0.00001748, -0.00000117, 0.00000181, 0.00000328, -0.00000479, 0.00000063, 0.00000227, 0.00000102, 0.00000287, -0.00000115, -0.00000060, -0.00000512, 0.00000650, 0.00000337, -0.00000410, -0.00000461, -0.00000274, -0.00000170, -0.00000493, -0.00000281, -0.00000152, -0.00000358, -0.00001163, -0.00000792, -0.00005406, -0.00002057, 0.00006643, 0.00003580, 0.00001901, 0.00002287, 0.00002248, 0.00001772, 0.00005245, 0.00005103, -0.00004625, -0.00003521, -0.00001006, -0.00001263, -0.00000395, -0.00000525, -0.00000272, -0.00000279, 0.00000360, -0.00000297, -0.00000392, -0.00000030, -0.00000400, 0.00000549, -0.00001571, -0.00000350, 0.00000290, 0.00000401, -0.00001264, -0.00000731, 0.00000330, -0.00000510, 0.00000107, -0.00000639, 0.00000238, 0.00000023, -0.00000364, -0.00000309, -0.00000583, -0.00000421, -0.00001452, -0.00001117, -0.00004190, -0.00005279, 0.00007096, 0.00005114, 0.00002319, 0.00002659, 0.00002601, 0.00002460, 0.00007185, 0.00003538, -0.00007060, -0.00002494, -0.00001001, -0.00000904, -0.00000011, -0.00000449, -0.00000411, -0.00000553, -0.00000149, -0.00000245, -0.00000373, -0.00000416, 0.00000466, 0.00000530, -0.00000606, 0.00000129, -0.00000015, 0.00000086, 0.00000150, 0.00000250, -0.00000154, -0.00000299, 0.00000387, -0.00000126, 0.00001027, 0.00000702, -0.00001128, 0.00000298, -0.00000899, -0.00000160, -0.00000982, -0.00000192, -0.00000464, -0.00000694, -0.00000007, -0.00000141, 0.00000127, -0.00000160, -0.00000232, },
		{ -0.00000220, -0.00000092, 0.00000180, -0.00000201, 0.00000056, -0.00000650, -0.00000480, -0.00000470, -0.00001035, 0.00000615, -0.00001561, 0.00000359, -0.00001011, 0.00000830, 0.00000960, -0.00000399, 0.00000677, -0.00000471, -0.00000072, 0.00000196, 0.00000152, 0.00000203, 0.00000019, -0.00000028, -0.00000456, 0.00000224, 0.00000727, -0.00000426, -0.00000336, -0.00000447, -0.00000022, -0.00000565, -0.00000276, -0.00000370, 0.00000011, -0.00001168, -0.00000575, -0.00003470, -0.00005462, 0.00004644, 0.00005750, 0.00001923, 0.00002411, 0.00002302, 0.00002014, 0.00003791, 0.00007033, -0.00002668, -0.00004971, -0.00001057, -0.00001456, -0.00000499, -0.00000537, -0.00000319, -0.00000333, 0.00000109, 0.00000161, -0.00000675, 0.00000148, -0.00000589, 0.00000520, -0.00000866, -0.00001165, 0.00000385, 0.00000337, -0.00000463, -0.00001527, 0.00000545, -0.00000516, 0.00000079, -0.00000492, -0.00000137, 0.00000283, -0.00000327, -0.00000288, -0.00000505, -0.00000452, -0.00001055, -0.00001362, -0.00002523, -0.00006491, 0.00004021, 0.00006787, 0.00002087, 0.00002692, 0.00002414, 0.00002269, 0.00004696, 0.00006986, -0.00004272, -0.00005008, -0.00000629, -0.00001334, -0.00000094, -0.00000328, -0.00000354, -0.00000552, -0.00000251, -0.00000179, -0.00000390, -0.00000424, 0.00000111, 0.00000813, -0.00000475, -0.00000110, 0.00000055, 0.00000059, 0.00000162, 0.00000204, 0.00000075, -0.00000442, 0.00000264, 0.00000109, 0.00000243, 0.00001531, -0.00001092, 0.00000162, -0.00000776, -0.00000341, -0.00000447, -0.00000845, -0.00000028, -0.00000968, 0.00000002, -0.00000212, 0.00000121, -0.00000046, -0.00000214, },
		},
		{ { 0.00000163, 0.00000098, -0.00000348, -0.00000104, -0.00000051, 0.00000003, -0.00000217, -0.00000277, 0.00001336, -0.00000036, -0.00000731, 0.00000179, 0.00000117, 0.00000232, -0.00002011, 0.00000217, -0.00000198, -0.00000442, -0.00000051, -0.00000109, -0.00000092, -0.00000168, -0.00001383, -0.00000241, -0.00000251, -0.00000654, -0.00000715, -0.00000786, -0.00001280, -0.00001606, -0.00006187, 0.00000916, 0.00006798, 0.00002820, 0.00001875, 0.00002073, 0.00002378, 0.00002433, 0.00008901, 0.00000230, -0.00006266, -0.00001257, -0.00000568, -0.00000980, -0.00000730, -0.00000538, 0.00000117, -0.00000362, -0.00000262, -0.00000180, 0.00000013, -0.00000120, -0.00000054, -0.00000068, 0.00000330, -0.00000306, -0.00000865, 0.00000133, 0.00000109, 0.00000027, -0.00000153, -0.00000195, 0.00001133, -0.00000386, -0.00000478, -0.00000386, 0.00001133, -0.00000195, -0.00000153, 0.00000027, 0.00000109, 0.00000133, -0.00000865, -0.00000306, 0.00000330, -0.00000068, -0.00000054, -0.00000120, 0.00000013, -0.00000180, -0.00000262, -0.00000362, 0.00000117, -0.00000538, -0.00000730, -0.00000980, -0.00000568, -0.00001257, -0.00006266, 0.00000230, 0.00008901, 0.00002433, 0.00002378, 0.00002073, 0.00001875, 0.00002820, 0.00006798, 0.00000916, -0.00006187, -0.00001606, -0.00001280, -0.00000786, -0.00000715, -0.00000654, -0.00000251, -0.00000241, -0.00001383, -0.00000168, -0.00000092, -0.00000109, -0.00000051, -0.00000442, -0.00000198, 0.00000217, -0.00002011, 0.00000232, 0.00000117, 0.00000179, -0.00000731, -0.00000036, 0.00001336, -0.00000277, -0.00000217, 0.00000003, -0.00000051, -0.00000104, -0.00000348, 0.00000098, },
		{ 0.00000196, 0.00000145, -0.00000155, -0.00000311, 0.00000027, -0.00000063, 0.00000022, -0.00000563, 0.00000781, 0.00000908, -0.00000962, 0.00000040, -0.00000017, 0.00000600, -0.00001387, -0.00000981, 0.00000494, -0.00000706, -0.00000009, -0.00000179, -0.00000104, -0.00000016, -0.00000837, -0.00001090, 0.00000151, -0.00000647, -0.00000662, -0.00000747, -0.00001114, -0.00001060, -0.00003939, -0.00004999, 0.00006161, 0.00004839, 0.00002272, 0.00001754, 0.00002978, 0.00001387, 0.00007008, 0.00005968, -0.00006249, -0.00003293, -0.00000303, -0.00000997, -0.00000761, -0.00000787, -0.00000113, -0.00000030, -0.00000378, -0.00000225, -0.00000068, -0.00000053, -0.00000086, -0.00000110, 0.00000156, 0.00000170, -0.00000843, -0.00000348, 0.00000299, -0.00000101, 0.00000176, -0.00000543, 0.00000856, 0.00000294, -0.00000565, -0.00000561, 0.00000585, 0.00000669, -0.00000584, 0.00000205, -0.00000163, 0.00000432, -0.00000582, -0.00000673, 0.00000141, 0.00000171, -0.00000119, -0.00000059, -0.00000071, -0.00000035, -0.00000231, -0.00000362, -0.00000173, -0.00000042, -0.00000771, -0.00000763, -0.00000987, -0.00000277, -0.00004443, -0.00003986, 0.00006862, 0.00005558, 0.00001224, 0.00002773, 0.00001312, 0.00002275, 0.00004468, 0.00005696, -0.00004169, -0.00004092, -0.00000961, -0.00001181, -0.00000644, -0.00000642, -0.00000580, 0.00000039, -0.00001043, -0.00000831, 0.00000025, -0.00000117, -0.00000073, -0.00000157, -0.00000655, 0.00000542, -0.00001379, -0.00000928, 0.00000584, -0.00000116, -0.00000052, -0.00000890, 0.00001307, 0.00000321, -0.00000369, -0.00000020, -0.00000083, -0.00000032, -0.00000298, -0.00000120, },
		{ 0.00000145, 0.00000151, 0.00000053, -0.00000387, -0.00000042, -0.00000056, 0.00000110, -0.00000428, -0.00000008, 0.00001368, -0.00000463, -0.00000515, 0.00000149, 0.00000347, -0.00000239, -0.00001901, 0.00000462, -0.00000412, -0.00000244, -0.00000114, -0.00000135, -0.00000119, -0.00000174, -0.00001433, -0.00000177, -0.00000251, -0.00000673, -0.00000823, -0.00000746, -0.00001370, -0.00001457, -0.00006563, 0.00000431, 0.00007666, 0.00002687, 0.00002280, 0.00002401, 0.00002450, 0.00003395, 0.00009630, -0.00002639, -0.00006050, -0.00000584, -0.00000830, -0.00000852, -0.00000801, -0.00000564, 0.00000147, -0.00000261, -0.00000308, -0.00000180, 0.00000000, -0.00000123, -0.00000057, -0.00000090, 0.00000377, -0.00000442, -0.00000782, 0.00000210, -0.00000022, 0.00000166, -0.00000284, 0.00000083, 0.00000945, -0.00000504, -0.00000486, -0.00000180, 0.00001202, -0.00000450, 0.00000001, -0.00000127, 0.00000287, -0.00000044, -0.00000844, -0.00000207, 0.00000279, -0.00000047, -0.00000046, -0.00000124, 0.00000037, -0.00000194, -0.00000215, -0.00000480, 0.00000109, -0.00000501, -0.00000677, -0.00001099, -0.00000259, -0.00002311, -0.00005503, 0.00002484, 0.00007710, 0.00001860, 0.00002254, 0.00001767, 0.00001574, 0.00002901, 0.00006009, 0.00001314, -0.00005811, -0.00001741, -0.00001196, -0.00000817, -0.00000613, -0.00000618, -0.00000272, -0.00000316, -0.00001298, -0.00000176, -0.00000064, -0.00000069, -0.00000051, -0.00000575, 0.00000020, -0.00000174, -0.00001874, 0.00000568, -0.00000149, 0.00000257, -0.00000945, 0.00000498, 0.00001067, -0.00000370, -0.00000062, -0.00000109, -0.00000048, -0.00000157, -0.00000297, },
		{ -0.00000079, 0.00000229, 0.00000087, -0.00000187, -0.00000300, 0.00000068, -0.00000014, -0.00000030, -0.00000611, 0.00001159, 0.00000408, -0.00000886, 0.00000067, 0.00000188, 0.00000384, -0.00001650, -0.00000577, 0.00000333, -0.00000657, 0.00000109, -0.00000320, -0.00000051, -0.00000101, -0.00000799, -0.00001145, 0.00000257, -0.00000735, -0.00000646, -0.00000884, -0.00001006, -0.00001204, -0.00003715, -0.00005944, 0.00006675, 0.00005186, 0.00002326, 0.00002266, 0.00003133, 0.00001741, 0.00008617, 0.00003964, -0.00007775, -0.00001966, -0.00000654, -0.00000806, -0.00000886, -0.00000723, -0.00000219, 0.00000156, -0.00000439, -0.00000196, -0.00000124, -0.00000014, -0.00000130, -0.00000088, 0.00000140, 0.00000159, -0.00000951, -0.00000118, 0.00000106, 0.00000033, 0.00000087, -0.00000408, 0.00000916, 0.00000035, -0.00000548, -0.00000509, 0.00000864, 0.00000376, -0.00000515, 0.00000156, -0.00000130, 0.00000461, -0.00000762, -0.00000480, 0.00000083, 0.00000185, -0.00000112, -0.00000055, -0.00000058, -0.00000037, -0.00000206, -0.00000398, -0.00000262, -0.00000009, -0.00000684, -0.00000892, -0.00000739, -0.00000731, -0.00004993, -0.00001535, 0.00006862, 0.00004396, 0.00001157, 0.00002518, 0.00000977, 0.00002280, 0.00004098, 0.00005276, -0.00003459, -0.00004162, -0.00000914, -0.00001197, -0.00000582, -0.00000581, -0.00000543, -0.00000075, -0.00000991, -0.00000795, 0.00000028, -0.00000085, -0.00000027, -0.00000294, -0.00000528, 0.00000440, -0.00001674, -0.00000367, 0.00000348, -0.00000057, -0.00000255, -0.00000614, 0.00001495, -0.00000096, -0.00000102, -0.00000139, -0.00000071, -0.00000104, -0.00000264, },
		{ -0.00000238, 0.00000187, 0.00000119, 0.00000012, -0.00000409, 0.00000019, -0.00000047, 0.00000170, -0.00000620, 0.00000366, 0.00001170, -0.00000731, -0.00000342, 0.00000216, 0.00000434, -0.00000712, -0.00001609, 0.00000545, -0.00000575, -0.00000000, -0.00000257, -0.00000108, -0.00000184, -0.00000156, -0.00001476, -0.00000105, -0.00000294, -0.00000649, -0.00000966, -0.00000665, -0.00001503, -0.00001251, -0.00006996, -0.00000068, 0.00008521, 0.00002602, 0.00002730, 0.00002787, 0.00002453, 0.00004877, 0.00009428, -0.00005612, -0.00004964, -0.00000457, -0.00000866, -0.00000842, -0.00000798, -0.00000632, 0.00000234, -0.00000210, -0.00000326, -0.00000213, 0.00000015, -0.00000149, -0.00000043, -0.00000118, 0.00000411, -0.00000589, -0.00000615, 0.00000176, -0.00000073, 0.00000258, -0.00000360, 0.00000331, 0.00000682, -0.00000545, -0.00000506, 0.00000107, 0.00001115, -0.00000623, 0.00000129, -0.00000253, 0.00000459, -0.00000278, -0.00000728, -0.00000155, 0.00000229, -0.00000015, -0.00000050, -0.00000113, 0.00000046, -0.00000194, -0.00000199, -0.00000574, 0.00000085, -0.00000399, -0.00000732, -0.00001057, -0.00000156, -0.00003384, -0.00003969, 0.00003906, 0.00006415, 0.00001541, 0.00002099, 0.00001474, 0.00001398, 0.00002870, 0.00005370, 0.00001560, -0.00005377, -0.00001907, -0.00001076, -0.00000876, -0.00000488, -0.00000594, -0.00000295, -0.00000412, -0.00001159, -0.00000230, 0.00000005, -0.00000062, -0.00000076, -0.00000642, 0.00000200, -0.00000662, -0.00001468, 0.00000659, -0.00000317, 0.00000286, -0.00001066, 0.00001031, 0.00000623, -0.00000273, -0.00000011, -0.00000191, -0.00000060, -0.00000192, },
		{ -0.00000213, -0.00000033, 0.00000235, 0.00000033, -0.00000217, -0.00000266, 0.00000095, 0.00000047, -0.00000157, -0.00000490, 0.00001363, -0.00000084, -0.00000734, 0.00000100, 0.00000400, 0.00000000, -0.00001690, -0.00000242, 0.00000100, -0.00000494, 0.00000143, -0.00000440, 0.00000001, -0.00000199, -0.00000737, -0.00001192, 0.00000332, -0.00000811, -0.00000631, -0.00001014, -0.00000905, -0.00001339, -0.00003486, -0.00006931, 0.00007163, 0.00005561, 0.00002424, 0.00002872, 0.00003177, 0.00002482, 0.00010028, 0.00000932, -0.00008171, -0.00000924, -0.00001052, -0.00000563, -0.00001048, -0.00000631, -0.00000321, 0.00000342, -0.00000512, -0.00000172, -0.00000178, 0.00000023, -0.00000169, -0.00000069, 0.00000129, 0.00000103, -0.00000967, 0.00000054, -0.00000090, 0.00000205, -0.00000034, -0.00000207, 0.00000851, -0.00000160, -0.00000533, -0.00000382, 0.00001071, 0.00000029, -0.00000353, 0.00000037, -0.00000006, 0.00000367, -0.00000844, -0.00000304, 0.00000008, 0.00000200, -0.00000097, -0.00000066, -0.00000024, -0.00000063, -0.00000165, -0.00000467, -0.00000317, 0.00000028, -0.00000596, -0.00001052, -0.00000369, -0.00001611, -0.00004730, 0.00000608, 0.00006312, 0.00003494, 0.00001171, 0.00002182, 0.00000813, 0.00002190, 0.00003811, 0.00004839, -0.00002821, -0.00004185, -0.00000889, -0.00001190, -0.00000539, -0.00000502, -0.00000518, -0.00000210, -0.00000904, -0.00000769, 0.00000032, -0.00000043, -0.00000026, -0.00000403, -0.00000345, 0.00000165, -0.00001756, 0.00000147, -0.00000011, 0.00000129, -0.00000556, -0.00000134, 0.00001408, -0.00000359, 0.00000145, -0.00000289, -0.00000043, -0.00000180, },
		{ -0.00000205, -0.00000169, 0.00000212, 0.00000078, -0.00000029, -0.00000407, 0.00000070, -0.00000017, 0.00000156, -0.00000716, 0.00000739, 0.00000808, -0.00000850, -0.00000203, 0.00000365, 0.00000333, -0.00001080, -0.00001216, 0.00000465, -0.00000637, 0.00000233, -0.00000456, -0.00000028, -0.00000299, -0.00000098, -0.00001524, -0.00000028, -0.00000375, -0.00000582, -0.00001140, -0.00000548, -0.00001676, -0.00000990, -0.00007494, -0.00000557, 0.00009309, 0.00002638, 0.00003185, 0.00003229, 0.00002452, 0.00006870, 0.00007884, -0.00008013, -0.00003414, -0.00000793, -0.00000655, -0.00000984, -0.00000691, -0.00000759, 0.00000383, -0.00000215, -0.00000313, -0.00000276, 0.00000055, -0.00000193, -0.00000012, -0.00000152, 0.00000420, -0.00000710, -0.00000404, 0.00000049, -0.00000037, 0.00000292, -0.00000369, 0.00000510, 0.00000389, -0.00000526, -0.00000517, 0.00000447, 0.00000862, -0.00000665, 0.00000186, -0.00000304, 0.00000565, -0.00000508, -0.00000549, -0.00000150, 0.00000180, 0.00000032, -0.00000079, -0.00000072, 0.00000023, -0.00000165, -0.00000231, -0.00000625, 0.00000034, -0.00000226, -0.00000924, -0.00000777, -0.00000460, -0.00004076, -0.00002084, 0.00004563, 0.00005192, 0.00001405, 0.00001887, 0.00001243, 0.00001293, 0.00002746, 0.00004882, 0.00001646, -0.00004872, -0.00002117, -0.00000906, -0.00000974, -0.00000330, -0.00000591, -0.00000321, -0.00000514, -0.00000978, -0.00000328, 0.00000120, -0.00000110, -0.00000095, -0.00000646, 0.00000290, -0.00001153, -0.00000848, 0.00000460, -0.00000287, 0.00000170, -0.00000990, 0.00001425, 0.00000129, -0.00000039, -0.00000069, -0.00000227, -0.00000093, },
		{ -0.00000246, -0.00000150, 0.00000014, 0.00000216, -0.00000013, -0.00000243, -0.00000215, 0.00000117, 0.00000090, -0.00000307, -0.00000228, 0.00001356, -0.00000489, -0.00000558, 0.00000191, 0.00000520, -0.00000449, -0.00001528, -0.00000026, -0.00000153, -0.00000225, 0.00000057, -0.00000485, -0.00000000, -0.00000262, -0.00000688, -0.00001202, 0.00000341, -0.00000832, -0.00000658, -0.00001089, -0.00000866, -0.00001399, -0.00003334, -0.00007859, 0.00007511, 0.00006073, 0.00002497, 0.00003632, 0.00003024, 0.00003862, 0.00010679, -0.00002677, -0.00007372, -0.00000480, -0.00001241, -0.00000435, -0.00001130, -0.00000590, -0.00000361, 0.00000479, -0.00000556, -0.00000181, -0.00000201, 0.00000032, -0.00000182, -0.00000068, 0.00000126, 0.00000009, -0.00000886, 0.00000129, -0.00000243, 0.00000376, -0.00000157, 0.00000021, 0.00000679, -0.00000278, -0.00000535, -0.00000160, 0.00001142, -0.00000297, -0.00000145, -0.00000120, 0.00000174, 0.00000165, -0.00000810, -0.00000175, -0.00000071, 0.00000225, -0.00000089, -0.00000072, 0.00000008, -0.00000090, -0.00000138, -0.00000535, -0.00000370, 0.00000119, -0.00000602, -0.00001094, -0.00000101, -0.00002651, -0.00003722, 0.00002158, 0.00005521, 0.00002786, 0.00001255, 0.00001753, 0.00000854, 0.00001935, 0.00003680, 0.00004317, -0.00002193, -0.00004216, -0.00000836, -0.00001201, -0.00000479, -0.00000435, -0.00000486, -0.00000381, -0.00000752, -0.00000795, 0.00000083, -0.00000041, -0.00000033, -0.00000490, -0.00000130, -0.00000279, -0.00001546, 0.00000456, -0.00000337, 0.00000338, -0.00000867, 0.00000469, 0.00001059, -0.00000406, 0.00000295, -0.00000422, -0.00000021, },
		{ -0.00000143, -0.00000199, -0.00000089, 0.00000210, 0.00000042, -0.00000076, -0.00000376, 0.00000104, 0.00000030, 0.00000064, -0.00000668, 0.00001011, 0.00000375, -0.00000861, -0.00000052, 0.00000524, 0.00000072, -0.00001288, -0.00000795, 0.00000240, -0.00000551, 0.00000380, -0.00000644, 0.00000060, -0.00000431, -0.00000022, -0.00001556, 0.00000023, -0.00000455, -0.00000512, -0.00001299, -0.00000445, -0.00001831, -0.00000743, -0.00007989, -0.00001094, 0.00010063, 0.00002793, 0.00003678, 0.00003642, 0.00002660, 0.00009099, 0.00004856, -0.00009226, -0.00001960, -0.00001310, -0.00000332, -0.00001199, -0.00000530, -0.00000907, 0.00000554, -0.00000248, -0.00000294, -0.00000342, 0.00000094, -0.00000234, 0.00000016, -0.00000179, 0.00000389, -0.00000766, -0.00000206, -0.00000129, 0.00000071, 0.00000272, -0.00000313, 0.00000599, 0.00000108, -0.00000474, -0.00000491, 0.00000787, 0.00000470, -0.00000554, 0.00000140, -0.00000250, 0.00000554, -0.00000671, -0.00000353, -0.00000180, 0.00000139, 0.00000089, -0.00000129, -0.00000002, -0.00000031, -0.00000111, -0.00000306, -0.00000641, -0.00000012, -0.00000040, -0.00001184, -0.00000315, -0.00001206, -0.00004108, -0.00000299, 0.00004691, 0.00004069, 0.00001445, 0.00001558, 0.00001161, 0.00001154, 0.00002613, 0.00004478, 0.00001625, -0.00004336, -0.00002337, -0.00000717, -0.00001086, -0.00000161, -0.00000591, -0.00000375, -0.00000583, -0.00000803, -0.00000428, 0.00000246, -0.00000197, -0.00000105, -0.00000582, 0.00000231, -0.00001520, -0.00000151, 0.00000031, -0.00000041, -0.00000130, -0.00000654, 0.00001556, -0.00000275, 0.00000238, -0.00000199, -0.00000224, },
		{ -0.00000022, -0.00000293, -0.00000075, 0.00000052, 0.00000179, -0.00000053, -0.00000256, -0.00000157, 0.00000142, 0.00000083, -0.00000412, 0.00000098, 0.00001158, -0.00000780, -0.00000368, 0.00000337, 0.00000491, -0.00000865, -0.00001213, 0.00000037, -0.00000354, 0.00000095, -0.00000137, -0.00000438, -0.00000081, -0.00000262, -0.00000671, -0.00001165, 0.00000274, -0.00000785, -0.00000741, -0.00001093, -0.00000909, -0.00001360, -0.00003304, -0.00008661, 0.00007623, 0.00006824, 0.00002487, 0.00004582, 0.00002658, 0.00006045, 0.00009946, -0.00006109, -0.00005700, -0.00000675, -0.00001115, -0.00000511, -0.00001064, -0.00000645, -0.00000309, 0.00000537, -0.00000554, -0.00000234, -0.00000187, 0.00000007, -0.00000164, -0.00000090, 0.00000123, -0.00000099, -0.00000724, 0.00000099, -0.00000320, 0.00000518, -0.00000261, 0.00000242, 0.00000429, -0.00000315, -0.00000553, 0.00000158, 0.00001027, -0.00000518, 0.00000043, -0.00000256, 0.00000349, -0.00000096, -0.00000672, -0.00000105, -0.00000145, 0.00000264, -0.00000103, -0.00000053, 0.00000015, -0.00000094, -0.00000149, -0.00000577, -0.00000439, 0.00000285, -0.00000757, -0.00000901, -0.00000167, -0.00003482, -0.00002278, 0.00003087, 0.00004662, 0.00002248, 0.00001358, 0.00001286, 0.00001058, 0.00001519, 0.00003721, 0.00003684, -0.00001543, -0.00004279, -0.00000731, -0.00001254, -0.00000384, -0.00000395, -0.00000444, -0.00000580, -0.00000539, -0.00000882, 0.00000198, -0.00000109, -0.00000014, -0.00000556, 0.00000074, -0.00000835, -0.00001036, 0.00000445, -0.00000471, 0.00000430, -0.00001054, 0.00001053, 0.00000544, -0.00000249, 0.00000310, -0.00000501, },
		{ -0.00000206, -0.00000196, -0.00000177, -0.00000002, 0.00000175, 0.00000020, -0.00000131, -0.00000312, 0.00000119, 0.00000075, -0.00000077, -0.00000470, 0.00001114, -0.00000047, -0.00000805, 0.00000148, 0.00000606, -0.00000275, -0.00001326, -0.00000410, -0.00000094, -0.00000295, 0.00000370, -0.00000739, 0.00000085, -0.00000519, 0.00000023, -0.00001527, -0.00000005, -0.00000476, -0.00000498, -0.00001378, -0.00000436, -0.00001878, -0.00000618, -0.00008356, -0.00001809, 0.00010895, 0.00002987, 0.00004310, 0.00003868, 0.00003425, 0.00010983, 0.00000627, -0.00008931, -0.00001094, -0.00001668, -0.00000108, -0.00001341, -0.00000413, -0.00000999, 0.00000683, -0.00000259, -0.00000308, -0.00000374, 0.00000099, -0.00000240, 0.00000016, -0.00000184, 0.00000305, -0.00000731, -0.00000071, -0.00000306, 0.00000221, 0.00000211, -0.00000205, 0.00000590, -0.00000129, -0.00000414, -0.00000390, 0.00001052, 0.00000006, -0.00000313, -0.00000003, -0.00000091, 0.00000406, -0.00000719, -0.00000185, -0.00000228, 0.00000111, 0.00000143, -0.00000182, 0.00000077, -0.00000098, -0.00000055, -0.00000397, -0.00000649, -0.00000005, 0.00000066, -0.00001373, 0.00000141, -0.00002218, -0.00003441, 0.00001071, 0.00004543, 0.00003023, 0.00001662, 0.00001068, 0.00001301, 0.00000877, 0.00002573, 0.00004062, 0.00001581, -0.00003835, -0.00002505, -0.00000558, -0.00001169, -0.00000020, -0.00000566, -0.00000484, -0.00000572, -0.00000690, -0.00000472, 0.00000324, -0.00000273, -0.00000126, -0.00000443, -0.00000023, -0.00001631, 0.00000436, -0.00000476, 0.00000339, -0.00000567, -0.00000078, 0.00001372, -0.00000473, 0.00000452, -0.00000342, },
		{ -0.00000518, -0.00000052, -0.00000318, 0.00000012, 0.00000065, 0.00000142, -0.00000097, -0.00000245, -0.00000104, 0.00000173, 0.00000014, -0.00000413, 0.00000395, 0.00000833, -0.00000966, -0.00000138, 0.00000486, 0.00000317, -0.00001184, -0.00000799, -0.00000076, -0.00000424, 0.00000372, -0.00000371, -0.00000335, -0.00000216, -0.00000214, -0.00000672, -0.00001109, 0.00000160, -0.00000698, -0.00000844, -0.00001067, -0.00000989, -0.00001269, -0.00003356, -0.00009361, 0.00007496, 0.00007835, 0.00002415, 0.00005651, 0.00002236, 0.00008938, 0.00007391, -0.00008517, -0.00003765, -0.00001266, -0.00000764, -0.00000743, -0.00000874, -0.00000778, -0.00000187, 0.00000532, -0.00000521, -0.00000310, -0.00000155, -0.00000032, -0.00000129, -0.00000123, 0.00000105, -0.00000186, -0.00000525, -0.00000017, -0.00000311, 0.00000611, -0.00000329, 0.00000430, 0.00000133, -0.00000283, -0.00000573, 0.00000544, 0.00000710, -0.00000567, 0.00000142, -0.00000313, 0.00000449, -0.00000343, -0.00000471, -0.00000093, -0.00000207, 0.00000316, -0.00000143, -0.00000001, -0.00000013, -0.00000068, -0.00000205, -0.00000588, -0.00000506, 0.00000491, -0.00001029, -0.00000469, -0.00000679, -0.00003783, -0.00000803, 0.00003561, 0.00003770, 0.00001919, 0.00001375, 0.00000897, 0.00001306, 0.00001031, 0.00003866, 0.00002988, -0.00000900, -0.00004350, -0.00000594, -0.00001332, -0.00000270, -0.00000367, -0.00000419, -0.00000761, -0.00000309, -0.00000994, 0.00000349, -0.00000234, 0.00000030, -0.00000584, 0.00000199, -0.00001390, -0.00000311, 0.00000105, -0.00000324, 0.00000294, -0.00000988, 0.00001451, 0.00000015, 0.00000034, 0.00000205, },
		{ -0.00000443, -0.00000201, -0.00000239, -0.00000142, 0.00000084, 0.00000111, 0.00000014, -0.00000191, -0.00000221, 0.00000119, 0.00000097, -0.00000215, -0.00000166, 0.00001033, -0.00000407, -0.00000695, 0.00000402, 0.00000552, -0.00000624, -0.00001211, -0.00000116, -0.00000467, 0.00000100, 0.00000180, -0.00000690, -0.00000008, -0.00000509, -0.00000009, -0.00001401, -0.00000151, -0.00000393, -0.00000584, -0.00001325, -0.00000576, -0.00001748, -0.00000709, -0.00008481, -0.00002839, 0.00011939, 0.00003121, 0.00005185, 0.00003768, 0.00005108, 0.00011729, -0.00004072, -0.00007255, -0.00001046, -0.00001640, -0.00000148, -0.00001285, -0.00000432, -0.00000968, 0.00000713, -0.00000205, -0.00000386, -0.00000347, 0.00000047, -0.00000190, -0.00000030, -0.00000159, 0.00000173, -0.00000595, -0.00000034, -0.00000436, 0.00000382, 0.00000123, -0.00000060, 0.00000484, -0.00000294, -0.00000370, -0.00000186, 0.00001160, -0.00000429, -0.00000007, -0.00000201, 0.00000131, 0.00000142, -0.00000636, -0.00000077, -0.00000279, 0.00000101, 0.00000176, -0.00000216, 0.00000140, -0.00000152, -0.00000024, -0.00000475, -0.00000672, 0.00000094, 0.00000010, -0.00001341, 0.00000345, -0.00003170, -0.00002287, 0.00001939, 0.00004270, 0.00002064, 0.00002004, 0.00000447, 0.00001658, 0.00000424, 0.00002681, 0.00003571, 0.00001577, -0.00003423, -0.00002576, -0.00000474, -0.00001188, 0.00000066, -0.00000500, -0.00000657, -0.00000462, -0.00000674, -0.00000418, 0.00000303, -0.00000292, -0.00000176, -0.00000239, -0.00000473, -0.00001400, 0.00000732, -0.00000859, 0.00000691, -0.00001006, 0.00000625, 0.00000910, -0.00000421, 0.00000530, },
		{ 0.00000036, -0.00000489, -0.00000100, -0.00000322, 0.00000110, 0.00000044, 0.00000119, -0.00000155, -0.00000200, -0.00000067, 0.00000204, -0.00000110, -0.00000281, 0.00000583, 0.00000453, -0.00001070, 0.00000161, 0.00000558, 0.00000055, -0.00001379, -0.00000336, -0.00000363, -0.00000295, 0.00000498, -0.00000540, -0.00000264, -0.00000337, -0.00000176, -0.00000635, -0.00001096, 0.00000071, -0.00000643, -0.00000888, -0.00001102, -0.00001003, -0.00001247, -0.00003363, -0.00010082, 0.00007234, 0.00009019, 0.00002412, 0.00006640, 0.00002132, 0.00012061, 0.00003000, -0.00009255, -0.00002245, -0.00001849, -0.00000425, -0.00000971, -0.00000672, -0.00000904, -0.00000064, 0.00000519, -0.00000502, -0.00000365, -0.00000148, -0.00000046, -0.00000113, -0.00000140, 0.00000045, -0.00000210, -0.00000342, -0.00000179, -0.00000226, 0.00000648, -0.00000355, 0.00000558, -0.00000170, -0.00000207, -0.00000559, 0.00000936, 0.00000225, -0.00000421, 0.00000111, -0.00000251, 0.00000417, -0.00000507, -0.00000260, -0.00000121, -0.00000251, 0.00000369, -0.00000197, 0.00000074, -0.00000068, -0.00000022, -0.00000290, -0.00000588, -0.00000523, 0.00000645, -0.00001296, 0.00000064, -0.00001557, -0.00003420, 0.00000368, 0.00003798, 0.00002816, 0.00001849, 0.00001203, 0.00000711, 0.00001451, 0.00000605, 0.00003997, 0.00002324, -0.00000339, -0.00004364, -0.00000481, -0.00001386, -0.00000179, -0.00000312, -0.00000455, -0.00000863, -0.00000134, -0.00001061, 0.00000466, -0.00000360, 0.00000067, -0.00000541, 0.00000163, -0.00001784, 0.00000448, -0.00000456, 0.00000082, -0.00000093, -0.00000607, 0.00001534, -0.00000370, 0.00000324, },
		{ 0.00000448, -0.00000468, -0.00000227, -0.00000260, -0.00000094, 0.00000155, 0.00000035, 0.00000014, -0.00000244, -0.00000114, 0.00000107, 0.00000077, -0.00000294, 0.00000170, 0.00000797, -0.00000692, -0.00000513, 0.00000661, 0.00000355, -0.00000914, -0.00000967, 0.00000029, -0.00000780, 0.00000545, -0.00000140, -0.00000509, -0.00000221, -0.00000392, -0.00000121, -0.00001185, -0.00000409, -0.00000209, -0.00000764, -0.00001152, -0.00000859, -0.00001445, -0.00001027, -0.00008330, -0.00004244, 0.00013262, 0.00003169, 0.00006309, 0.00003359, 0.00007877, 0.00010571, -0.00008222, -0.00004768, -0.00001694, -0.00001220, -0.00000474, -0.00001002, -0.00000611, -0.00000800, 0.00000631, -0.00000083, -0.00000522, -0.00000266, -0.00000057, -0.00000089, -0.00000120, -0.00000113, 0.00000016, -0.00000380, -0.00000103, -0.00000487, 0.00000523, 0.00000026, 0.00000104, 0.00000292, -0.00000375, -0.00000349, 0.00000130, 0.00001047, -0.00000728, 0.00000272, -0.00000382, 0.00000345, -0.00000177, -0.00000443, -0.00000041, -0.00000319, 0.00000110, 0.00000178, -0.00000215, 0.00000167, -0.00000174, -0.00000034, -0.00000527, -0.00000704, 0.00000271, -0.00000221, -0.00001018, 0.00000111, -0.00003710, -0.00001010, 0.00002426, 0.00003888, 0.00001274, 0.00002344, -0.00000191, 0.00002129, -0.00000142, 0.00002902, 0.00003016, 0.00001617, -0.00003101, -0.00002546, -0.00000466, -0.00001144, 0.00000102, -0.00000414, -0.00000860, -0.00000281, -0.00000737, -0.00000275, 0.00000177, -0.00000237, -0.00000251, -0.00000017, -0.00001050, -0.00000836, 0.00000636, -0.00000948, 0.00000841, -0.00001271, 0.00001274, 0.00000303, -0.00000159, },
		{ 0.00000507, -0.00000124, -0.00000450, -0.00000152, -0.00000311, 0.00000217, -0.00000015, 0.00000120, -0.00000231, -0.00000114, -0.00000053, 0.00000224, -0.00000256, -0.00000028, 0.00000618, 0.00000077, -0.00001099, 0.00000534, 0.00000488, -0.00000222, -0.00001451, 0.00000117, -0.00000784, 0.00000049, 0.00000399, -0.00000546, -0.00000315, -0.00000358, -0.00000222, -0.00000494, -0.00001201, 0.00000084, -0.00000699, -0.00000786, -0.00001298, -0.00000834, -0.00001435, -0.00003163, -0.00011004, 0.00007011, 0.00010220, 0.00002668, 0.00007283, 0.00002867, 0.00014557, -0.00002675, -0.00008130, -0.00001627, -0.00002050, -0.00000353, -0.00001001, -0.00000600, -0.00000916, -0.00000030, 0.00000563, -0.00000550, -0.00000352, -0.00000209, 0.00000007, -0.00000152, -0.00000116, -0.00000073, -0.00000141, -0.00000227, -0.00000336, -0.00000090, 0.00000628, -0.00000330, 0.00000608, -0.00000441, -0.00000118, -0.00000468, 0.00001242, -0.00000339, -0.00000113, -0.00000043, -0.00000071, 0.00000233, -0.00000535, -0.00000092, -0.00000168, -0.00000272, 0.00000408, -0.00000245, 0.00000149, -0.00000128, 0.00000018, -0.00000378, -0.00000603, -0.00000441, 0.00000645, -0.00001392, 0.00000460, -0.00002534, -0.00002509, 0.00001094, 0.00003946, 0.00001802, 0.00002035, 0.00000805, 0.00000794, 0.00001391, 0.00000352, 0.00004004, 0.00001790, 0.00000060, -0.00004249, -0.00000456, -0.00001363, -0.00000157, -0.00000199, -0.00000584, -0.00000835, -0.00000077, -0.00001014, 0.00000472, -0.00000416, 0.00000049, -0.00000396, -0.00000096, -0.00001867, 0.00001017, -0.00001027, 0.00000605, -0.00000638, 0.00000044, 0.00001263, -0.00000500, },
		{ 0.00000203, 0.00000246, -0.00000421, -0.00000284, -0.00000260, -0.00000032, 0.00000195, -0.00000033, 0.00000005, -0.00000272, -0.00000006, 0.00000084, 0.00000011, -0.00000268, 0.00000457, 0.00000461, -0.00000907, -0.00000229, 0.00000845, 0.00000064, -0.00001119, -0.00000622, -0.00000019, -0.00000922, 0.00000903, -0.00000477, -0.00000278, -0.00000491, -0.00000223, -0.00000261, -0.00000939, -0.00000715, 0.00000010, -0.00000963, -0.00000944, -0.00001186, -0.00001077, -0.00001467, -0.00007998, -0.00005948, 0.00014799, 0.00003242, 0.00007514, 0.00002931, 0.00011500, 0.00007084, -0.00010821, -0.00002289, -0.00002625, -0.00000626, -0.00000942, -0.00000588, -0.00000880, -0.00000555, 0.00000483, 0.00000066, -0.00000673, -0.00000175, -0.00000169, 0.00000027, -0.00000225, -0.00000073, -0.00000126, -0.00000129, -0.00000255, -0.00000447, 0.00000617, -0.00000062, 0.00000265, 0.00000037, -0.00000376, -0.00000337, 0.00000527, 0.00000696, -0.00000808, 0.00000437, -0.00000473, 0.00000469, -0.00000468, -0.00000192, -0.00000069, -0.00000341, 0.00000130, 0.00000151, -0.00000176, 0.00000151, -0.00000163, -0.00000085, -0.00000561, -0.00000708, 0.00000453, -0.00000541, -0.00000475, -0.00000581, -0.00003617, 0.00000031, 0.00002748, 0.00003329, 0.00000773, 0.00002519, -0.00000675, 0.00002544, -0.00000675, 0.00003120, 0.00002483, 0.00001636, -0.00002809, -0.00002467, -0.00000492, -0.00001076, 0.00000128, -0.00000360, -0.00001022, -0.00000105, -0.00000809, -0.00000110, -0.00000003, -0.00000142, -0.00000310, 0.00000131, -0.00001612, -0.00000066, 0.00000175, -0.00000676, 0.00000674, -0.00001214, 0.00001680, -0.00000269, },
		{ -0.00000355, 0.00000518, -0.00000216, -0.00000433, -0.00000190, -0.00000287, 0.00000320, -0.00000100, 0.00000140, -0.00000319, 0.00000010, -0.00000066, 0.00000219, -0.00000376, 0.00000296, 0.00000501, -0.00000269, -0.00001032, 0.00000934, 0.00000261, -0.00000455, -0.00001405, 0.00000488, -0.00001246, 0.00000552, 0.00000075, -0.00000353, -0.00000535, -0.00000231, -0.00000393, -0.00000214, -0.00001459, 0.00000241, -0.00000904, -0.00000498, -0.00001701, -0.00000423, -0.00001917, -0.00002645, -0.00012264, 0.00006972, 0.00011315, 0.00003327, 0.00007398, 0.00004906, 0.00015384, -0.00008576, -0.00005516, -0.00002015, -0.00001700, -0.00000683, -0.00000720, -0.00000749, -0.00000746, -0.00000138, 0.00000703, -0.00000688, -0.00000250, -0.00000356, 0.00000143, -0.00000265, -0.00000038, -0.00000251, 0.00000029, -0.00000215, -0.00000440, 0.00000065, 0.00000555, -0.00000251, 0.00000561, -0.00000643, -0.00000047, -0.00000261, 0.00001362, -0.00000855, 0.00000269, -0.00000267, 0.00000178, -0.00000073, -0.00000419, 0.00000001, -0.00000213, -0.00000270, 0.00000418, -0.00000265, 0.00000197, -0.00000167, 0.00000029, -0.00000447, -0.00000641, -0.00000246, 0.00000435, -0.00001187, 0.00000477, -0.00003251, -0.00001366, 0.00001458, 0.00004010, 0.00000821, 0.00002373, 0.00000253, 0.00001099, 0.00001130, 0.00000296, 0.00003845, 0.00001431, 0.00000264, -0.00003972, -0.00000551, -0.00001238, -0.00000217, -0.00000028, -0.00000798, -0.00000676, -0.00000157, -0.00000828, 0.00000328, -0.00000352, -0.00000051, -0.00000153, -0.00000580, -0.00001556, 0.00001206, -0.00001374, 0.00001033, -0.00001158, 0.00000812, 0.00000712, },
		{ -0.00000637, 0.00000526, -0.00000000, -0.00000332, -0.00000359, -0.00000242, 0.00000047, 0.00000191, -0.00000074, -0.00000028, -0.00000256, 0.00000082, 0.00000050, -0.00000085, -0.00000128, 0.00000635, 0.00000081, -0.00001052, 0.00000168, 0.00000874, -0.00000244, -0.00001241, -0.00000207, -0.00000269, -0.00000814, 0.00001044, -0.00000691, -0.00000121, -0.00000708, -0.00000096, -0.00000339, -0.00000762, -0.00000963, 0.00000158, -0.00001064, -0.00000835, -0.00001405, -0.00000820, -0.00001836, -0.00007692, -0.00007749, 0.00016350, 0.00003591, 0.00008466, 0.00003057, 0.00015233, 0.00001447, -0.00011232, -0.00000597, -0.00003320, -0.00000194, -0.00001304, -0.00000225, -0.00001101, -0.00000342, 0.00000355, 0.00000171, -0.00000771, -0.00000138, -0.00000231, 0.00000102, -0.00000302, -0.00000077, -0.00000203, 0.00000097, -0.00000446, -0.00000327, 0.00000647, -0.00000121, 0.00000395, -0.00000244, -0.00000321, -0.00000296, 0.00000931, 0.00000153, -0.00000648, 0.00000439, -0.00000424, 0.00000439, -0.00000650, 0.00000054, -0.00000137, -0.00000340, 0.00000151, 0.00000110, -0.00000114, 0.00000107, -0.00000132, -0.00000157, -0.00000601, -0.00000628, 0.00000528, -0.00000783, 0.00000084, -0.00001547, -0.00002899, 0.00000632, 0.00003078, 0.00002543, 0.00000642, 0.00002405, -0.00000864, 0.00002736, -0.00001018, 0.00003191, 0.00002092, 0.00001534, -0.00002456, -0.00002421, -0.00000482, -0.00001044, 0.00000196, -0.00000395, -0.00001064, -0.00000024, -0.00000798, -0.00000022, -0.00000145, -0.00000075, -0.00000288, 0.00000101, -0.00001974, 0.00000687, -0.00000480, -0.00000116, 0.00000197, -0.00000785, 0.00001713, },
		{ 0.00000048, -0.00000008, 0.00000356, -0.00000216, -0.00000452, -0.00000207, -0.00000247, 0.00000400, -0.00000191, 0.00000164, -0.00000402, 0.00000156, -0.00000105, 0.00000181, -0.00000423, 0.00000620, 0.00000265, -0.00000577, -0.00000833, 0.00001275, -0.00000080, -0.00000624, -0.00001236, 0.00000698, -0.00001617, 0.00001080, -0.00000381, -0.00000012, -0.00000893, 0.00000021, -0.00000666, 0.00000172, -0.00001837, 0.00000507, -0.00001215, -0.00000077, -0.00002256, 0.00000176, -0.00002649, -0.00001834, -0.00013864, 0.00007131, 0.00012321, 0.00004347, 0.00007069, 0.00008388, 0.00013644, -0.00013402, -0.00002261, -0.00003092, -0.00000926, -0.00001344, -0.00000164, -0.00001100, -0.00000412, -0.00000375, 0.00000919, -0.00000891, -0.00000088, -0.00000559, 0.00000338, -0.00000435, 0.00000074, -0.00000459, 0.00000277, -0.00000306, -0.00000462, 0.00000208, 0.00000442, -0.00000123, 0.00000413, -0.00000750, -0.00000008, 0.00000075, 0.00001222, -0.00001193, 0.00000608, -0.00000471, 0.00000409, -0.00000425, -0.00000189, 0.00000011, -0.00000242, -0.00000246, 0.00000391, -0.00000243, 0.00000201, -0.00000171, -0.00000000, -0.00000496, -0.00000676, 0.00000014, 0.00000050, -0.00000668, -0.00000008, -0.00003403, -0.00000373, 0.00001683, 0.00003879, 0.00000046, 0.00002671, -0.00000279, 0.00001471, 0.00000783, 0.00000355, 0.00003573, 0.00001207, 0.00000311, -0.00003564, -0.00000741, -0.00001037, -0.00000331, 0.00000158, -0.00001035, -0.00000444, -0.00000322, -0.00000546, 0.00000060, -0.00000181, -0.00000210, 0.00000127, -0.00001201, -0.00000888, 0.00000936, -0.00001325, 0.00001162, -0.00001442, 0.00001479, },
		{ 0.00001359, -0.00000703, 0.00000695, -0.00000207, -0.00000246, -0.00000430, -0.00000214, 0.00000138, 0.00000143, -0.00000081, -0.00000093, -0.00000188, 0.00000136, 0.00000000, -0.00000179, 0.00000105, 0.00000679, -0.00000310, -0.00001103, 0.00000641, 0.00000713, -0.00000506, -0.00001281, 0.00000216, -0.00000672, -0.00000448, 0.00000896, -0.00000677, -0.00000143, -0.00000771, -0.00000102, -0.00000269, -0.00000743, -0.00001060, 0.00000142, -0.00000966, -0.00000940, -0.00001378, -0.00000843, -0.00001933, -0.00007643, -0.00009405, 0.00017676, 0.00004504, 0.00008788, 0.00004447, 0.00017897, -0.00005458, -0.00009449, -0.00000133, -0.00003394, -0.00000207, -0.00001334, -0.00000092, -0.00001140, -0.00000269, 0.00000328, 0.00000167, -0.00000761, -0.00000205, -0.00000192, 0.00000090, -0.00000312, -0.00000152, -0.00000180, 0.00000244, -0.00000623, -0.00000149, 0.00000602, -0.00000126, 0.00000458, -0.00000504, -0.00000246, -0.00000175, 0.00001233, -0.00000471, -0.00000297, 0.00000288, -0.00000241, 0.00000240, -0.00000672, 0.00000236, -0.00000215, -0.00000317, 0.00000157, 0.00000077, -0.00000054, 0.00000057, -0.00000105, -0.00000228, -0.00000664, -0.00000425, 0.00000399, -0.00000774, 0.00000389, -0.00002452, -0.00001801, 0.00000815, 0.00003446, 0.00001582, 0.00000850, 0.00001990, -0.00000716, 0.00002629, -0.00001076, 0.00003016, 0.00001932, 0.00001238, -0.00001968, -0.00002476, -0.00000379, -0.00001092, 0.00000337, -0.00000548, -0.00000942, -0.00000099, -0.00000634, -0.00000092, -0.00000162, -0.00000105, -0.00000137, -0.00000183, -0.00001969, 0.00001168, -0.00001069, 0.00000528, -0.00000451, -0.00000063, },
		{ 0.00001837, -0.00000522, 0.00000401, 0.00000085, -0.00000142, -0.00000502, -0.00000206, -0.00000185, 0.00000436, -0.00000260, 0.00000168, -0.00000450, 0.00000299, -0.00000165, 0.00000112, -0.00000369, 0.00000876, -0.00000044, -0.00000847, -0.00000472, 0.00001454, -0.00000448, -0.00000742, -0.00000945, 0.00000679, -0.00001761, 0.00001458, -0.00000806, 0.00000346, -0.00001278, 0.00000302, -0.00000945, 0.00000561, -0.00002229, 0.00000774, -0.00001512, 0.00000339, -0.00002806, 0.00000784, -0.00003447, -0.00000922, -0.00015623, 0.00007312, 0.00013460, 0.00005441, 0.00006762, 0.00012881, 0.00008973, -0.00016012, 0.00000600, -0.00004256, -0.00000090, -0.00002076, 0.00000482, -0.00001516, -0.00000025, -0.00000652, 0.00001135, -0.00001081, 0.00000060, -0.00000749, 0.00000529, -0.00000609, 0.00000176, -0.00000646, 0.00000548, -0.00000472, -0.00000394, 0.00000308, 0.00000313, 0.00000036, 0.00000179, -0.00000760, 0.00000009, 0.00000504, 0.00000805, -0.00001261, 0.00000799, -0.00000564, 0.00000528, -0.00000726, 0.00000088, -0.00000044, -0.00000251, -0.00000209, 0.00000333, -0.00000178, 0.00000157, -0.00000140, -0.00000063, -0.00000539, -0.00000655, 0.00000237, -0.00000374, 0.00000025, -0.00000923, -0.00002878, 0.00000185, 0.00001999, 0.00003422, -0.00000349, 0.00002728, -0.00000590, 0.00001702, 0.00000534, 0.00000373, 0.00003315, 0.00001011, 0.00000301, -0.00003113, -0.00000955, -0.00000824, -0.00000436, 0.00000283, -0.00001200, -0.00000242, -0.00000471, -0.00000270, -0.00000240, 0.00000028, -0.00000357, 0.00000328, -0.00001784, -0.00000034, 0.00000286, -0.00000848, 0.00000886, -0.00001331, },
		{ 0.00000755, 0.00000730, -0.00000475, 0.00000657, -0.00000316, -0.00000195, -0.00000480, -0.00000175, 0.00000223, 0.00000067, -0.00000062, -0.00000178, -0.00000070, 0.00000148, -0.00000069, -0.00000231, 0.00000376, 0.00000600, -0.00000699, -0.00001012, 0.00001102, 0.00000394, -0.00000693, -0.00001225, 0.00000560, -0.00001108, 0.00000081, 0.00000492, -0.00000423, -0.00000374, -0.00000647, -0.00000273, -0.00000028, -0.00000910, -0.00000978, -0.00000059, -0.00000652, -0.00001286, -0.00001068, -0.00001209, -0.00001666, -0.00007979, -0.00010764, 0.00018633, 0.00006143, 0.00008286, 0.00007654, 0.00018181, -0.00012198, -0.00006144, -0.00000829, -0.00002778, -0.00000739, -0.00000951, -0.00000261, -0.00000941, -0.00000382, 0.00000431, 0.00000043, -0.00000636, -0.00000380, -0.00000045, -0.00000022, -0.00000245, -0.00000301, -0.00000049, 0.00000282, -0.00000740, 0.00000054, 0.00000481, -0.00000058, 0.00000422, -0.00000693, -0.00000191, 0.00000074, 0.00001319, -0.00001021, 0.00000132, 0.00000055, 0.00000017, -0.00000088, -0.00000532, 0.00000322, -0.00000281, -0.00000274, 0.00000134, 0.00000073, -0.00000023, 0.00000025, -0.00000103, -0.00000286, -0.00000741, -0.00000114, 0.00000042, -0.00000416, 0.00000224, -0.00002931, -0.00000711, 0.00000800, 0.00003727, 0.00000622, 0.00001234, 0.00001398, -0.00000330, 0.00002275, -0.00000871, 0.00002595, 0.00002010, 0.00000743, -0.00001334, -0.00002649, -0.00000175, -0.00001220, 0.00000533, -0.00000790, -0.00000676, -0.00000322, -0.00000315, -0.00000343, -0.00000017, -0.00000256, 0.00000139, -0.00000714, -0.00001526, 0.00001189, -0.00001333, 0.00000995, -0.00001034, },
		{ -0.00000794, 0.00001766, -0.00000824, 0.00000717, -0.00000198, -0.00000039, -0.00000562, -0.00000196, -0.00000097, 0.00000415, -0.00000284, 0.00000128, -0.00000439, 0.00000409, -0.00000237, 0.00000030, -0.00000213, 0.00001020, -0.00000388, -0.00001053, 0.00000037, 0.00001404, -0.00000755, -0.00000828, -0.00000552, 0.00000417, -0.00001589, 0.00001538, -0.00001030, 0.00000561, -0.00001545, 0.00000479, -0.00001103, 0.00000818, -0.00002495, 0.00000901, -0.00001637, 0.00000570, -0.00003142, 0.00001162, -0.00004038, -0.00000205, -0.00017234, 0.00007200, 0.00015106, 0.00006128, 0.00007286, 0.00017308, 0.00001813, -0.00015831, 0.00002213, -0.00004888, 0.00000374, -0.00002545, 0.00000961, -0.00001801, 0.00000261, -0.00000847, 0.00001249, -0.00001167, 0.00000107, -0.00000843, 0.00000638, -0.00000721, 0.00000213, -0.00000751, 0.00000774, -0.00000662, -0.00000246, 0.00000340, 0.00000199, 0.00000193, -0.00000107, -0.00000695, 0.00000048, 0.00000942, 0.00000174, -0.00001039, 0.00000786, -0.00000492, 0.00000467, -0.00000887, 0.00000334, -0.00000130, -0.00000242, -0.00000168, 0.00000258, -0.00000089, 0.00000080, -0.00000087, -0.00000143, -0.00000598, -0.00000525, 0.00000299, -0.00000644, 0.00000634, -0.00001997, -0.00001816, 0.00000240, 0.00002506, 0.00002598, -0.00000282, 0.00002418, -0.00000538, 0.00001630, 0.00000547, 0.00000195, 0.00003208, 0.00000723, 0.00000351, -0.00002722, -0.00001105, -0.00000673, -0.00000469, 0.00000281, -0.00001206, -0.00000174, -0.00000496, -0.00000121, -0.00000451, 0.00000173, -0.00000401, 0.00000322, -0.00002117, 0.00000741, -0.00000515, -0.00000077, 0.00000254, },
		{ -0.00001308, 0.00001417, 0.00000030, -0.00000062, 0.00000440, -0.00000317, -0.00000189, -0.00000508, -0.00000118, 0.00000279, -0.00000006, -0.00000044, -0.00000256, 0.00000074, 0.00000121, -0.00000158, -0.00000204, 0.00000621, 0.00000433, -0.00001084, -0.00000728, 0.00001427, 0.00000008, -0.00000828, -0.00001042, 0.00000718, -0.00001413, 0.00000597, -0.00000023, -0.00000029, -0.00000739, -0.00000396, -0.00000548, 0.00000316, -0.00001192, -0.00000789, -0.00000358, -0.00000221, -0.00001760, -0.00000592, -0.00001806, -0.00001134, -0.00008628, -0.00011884, 0.00019302, 0.00008393, 0.00007160, 0.00012733, 0.00015081, -0.00017178, -0.00002471, -0.00002139, -0.00001757, -0.00001601, -0.00000279, -0.00000650, -0.00000569, -0.00000625, 0.00000609, -0.00000143, -0.00000454, -0.00000607, 0.00000159, -0.00000195, -0.00000136, -0.00000483, 0.00000156, 0.00000216, -0.00000769, 0.00000247, 0.00000294, 0.00000090, 0.00000265, -0.00000776, -0.00000178, 0.00000466, 0.00001104, -0.00001346, 0.00000495, -0.00000149, 0.00000244, -0.00000451, -0.00000279, 0.00000315, -0.00000324, -0.00000212, 0.00000076, 0.00000113, -0.00000038, 0.00000027, -0.00000132, -0.00000335, -0.00000791, 0.00000228, -0.00000459, 0.00000243, -0.00000467, -0.00002745, 0.00000005, 0.00000874, 0.00003714, -0.00000091, 0.00001547, 0.00000852, 0.00000089, 0.00001847, -0.00000541, 0.00002035, 0.00002234, 0.00000129, -0.00000618, -0.00002889, 0.00000082, -0.00001375, 0.00000716, -0.00001033, -0.00000357, -0.00000611, 0.00000080, -0.00000710, 0.00000247, -0.00000483, 0.00000458, -0.00001378, -0.00000719, 0.00000704, -0.00001111, 0.00001068, },
		{ -0.00000540, 0.00000046, 0.00001283, -0.00000780, 0.00000828, -0.00000402, 0.00000041, -0.00000604, -0.00000186, 0.00000009, 0.00000339, -0.00000256, 0.00000039, -0.00000354, 0.00000463, -0.00000316, -0.00000038, 0.00000016, 0.00001041, -0.00000750, -0.00001146, 0.00000619, 0.00001123, -0.00000951, -0.00000881, -0.00000120, -0.00000025, -0.00001114, 0.00001260, -0.00000951, 0.00000523, -0.00001583, 0.00000447, -0.00001042, 0.00000841, -0.00002529, 0.00000784, -0.00001481, 0.00000487, -0.00003113, 0.00001119, -0.00004191, 0.00000038, -0.00018391, 0.00006477, 0.00017627, 0.00005940, 0.00009549, 0.00020121, -0.00006554, -0.00013096, 0.00002227, -0.00004624, 0.00000179, -0.00002505, 0.00001070, -0.00001798, 0.00000318, -0.00000865, 0.00001188, -0.00001086, -0.00000003, -0.00000783, 0.00000607, -0.00000722, 0.00000144, -0.00000729, 0.00000896, -0.00000815, -0.00000049, 0.00000289, 0.00000139, 0.00000298, -0.00000387, -0.00000598, 0.00000165, 0.00001263, -0.00000536, -0.00000598, 0.00000589, -0.00000265, 0.00000213, -0.00000859, 0.00000489, -0.00000209, -0.00000218, -0.00000139, 0.00000189, 0.00000000, -0.00000006, -0.00000034, -0.00000225, -0.00000677, -0.00000264, 0.00000122, -0.00000592, 0.00000880, -0.00002846, -0.00000566, -0.00000039, 0.00003118, 0.00001525, 0.00000174, 0.00001771, -0.00000124, 0.00001216, 0.00000885, -0.00000257, 0.00003327, 0.00000272, 0.00000536, -0.00002463, -0.00001138, -0.00000623, -0.00000406, 0.00000135, -0.00001020, -0.00000290, -0.00000332, -0.00000180, -0.00000480, 0.00000172, -0.00000279, 0.00000029, -0.00002030, 0.00001160, -0.00001161, 0.00000723, },
		{ 0.00000662, -0.00001118, 0.00001703, -0.00000510, 0.00000365, 0.00000133, -0.00000247, -0.00000212, -0.00000523, -0.00000031, 0.00000278, -0.00000040, -0.00000062, -0.00000288, 0.00000206, 0.00000067, -0.00000262, -0.00000080, 0.00000785, 0.00000214, -0.00001452, -0.00000231, 0.00001507, -0.00000330, -0.00000955, -0.00000705, 0.00000604, -0.00001425, 0.00000889, -0.00000441, 0.00000325, -0.00001079, -0.00000165, -0.00000783, 0.00000614, -0.00001435, -0.00000651, -0.00000581, 0.00000124, -0.00002135, -0.00000206, -0.00002355, -0.00000630, -0.00009299, -0.00013058, 0.00020028, 0.00010811, 0.00006116, 0.00018974, 0.00008355, -0.00019146, 0.00000317, -0.00003266, -0.00000856, -0.00002403, 0.00000392, -0.00001041, -0.00000198, -0.00000858, 0.00000741, -0.00000276, -0.00000326, -0.00000781, 0.00000324, -0.00000348, -0.00000054, -0.00000629, 0.00000367, 0.00000086, -0.00000710, 0.00000401, 0.00000068, 0.00000303, -0.00000005, -0.00000740, -0.00000197, 0.00000962, 0.00000578, -0.00001345, 0.00000670, -0.00000218, 0.00000332, -0.00000734, 0.00000003, 0.00000248, -0.00000345, -0.00000138, -0.00000014, 0.00000196, -0.00000103, 0.00000064, -0.00000189, -0.00000392, -0.00000756, 0.00000477, -0.00000929, 0.00001003, -0.00001521, -0.00001894, 0.00000152, 0.00001248, 0.00003244, -0.00000349, 0.00001560, 0.00000578, 0.00000309, 0.00001559, -0.00000282, 0.00001507, 0.00002458, -0.00000466, 0.00000065, -0.00003100, 0.00000308, -0.00001478, 0.00000796, -0.00001163, -0.00000111, -0.00000837, 0.00000418, -0.00001065, 0.00000524, -0.00000683, 0.00000678, -0.00001970, 0.00000234, -0.00000152, -0.00000419, },
		{ 0.00001254, -0.00001213, 0.00000947, 0.00000543, -0.00000434, 0.00000701, -0.00000477, 0.00000067, -0.00000616, -0.00000171, 0.00000115, 0.00000231, -0.00000195, -0.00000083, -0.00000202, 0.00000451, -0.00000395, -0.00000062, 0.00000264, 0.00000958, -0.00001130, -0.00001056, 0.00001146, 0.00000687, -0.00001045, -0.00000865, 0.00000240, -0.00000499, -0.00000468, 0.00000698, -0.00000590, 0.00000228, -0.00001379, 0.00000197, -0.00000755, 0.00000623, -0.00002325, 0.00000428, -0.00001055, 0.00000098, -0.00002714, 0.00000628, -0.00003847, -0.00000294, -0.00018959, 0.00005007, 0.00021165, 0.00004714, 0.00014168, 0.00019724, -0.00014275, -0.00008832, 0.00000931, -0.00003527, -0.00000661, -0.00001933, 0.00000771, -0.00001477, 0.00000127, -0.00000698, 0.00000959, -0.00000850, -0.00000256, -0.00000578, 0.00000437, -0.00000608, -0.00000026, -0.00000577, 0.00000885, -0.00000886, 0.00000166, 0.00000150, 0.00000165, 0.00000303, -0.00000600, -0.00000516, 0.00000415, 0.00001339, -0.00001144, -0.00000080, 0.00000299, 0.00000040, -0.00000173, -0.00000652, 0.00000526, -0.00000260, -0.00000183, -0.00000132, 0.00000149, 0.00000060, -0.00000077, 0.00000003, -0.00000303, -0.00000750, 0.00000084, -0.00000281, -0.00000163, 0.00000586, -0.00003127, 0.00000445, -0.00000332, 0.00003595, 0.00000461, 0.00000793, 0.00000965, 0.00000501, 0.00000570, 0.00001471, -0.00000934, 0.00003642, -0.00000322, 0.00000849, -0.00002333, -0.00001059, -0.00000656, -0.00000284, -0.00000103, -0.00000690, -0.00000554, -0.00000007, -0.00000441, -0.00000310, 0.00000015, -0.00000008, -0.00000531, -0.00001472, 0.00001046, -0.00001378, },
		{ 0.00000546, -0.00000128, -0.00000461, 0.00001511, -0.00000721, 0.00000639, -0.00000146, -0.00000167, -0.00000232, -0.00000540, 0.00000092, 0.00000204, -0.00000009, -0.00000144, -0.00000237, 0.00000286, 0.00000006, -0.00000369, 0.00000134, 0.00000835, -0.00000034, -0.00001758, 0.00000440, 0.00001288, -0.00000528, -0.00001110, -0.00000224, 0.00000200, -0.00001056, 0.00000796, -0.00000568, 0.00000449, -0.00001214, -0.00000120, -0.00000815, 0.00000696, -0.00001467, -0.00000741, -0.00000534, 0.00000162, -0.00002156, -0.00000208, -0.00002516, -0.00000540, -0.00009580, -0.00014719, 0.00021309, 0.00012756, 0.00006249, 0.00024885, -0.00001211, -0.00017645, 0.00001325, -0.00003505, -0.00000594, -0.00002725, 0.00000729, -0.00001178, -0.00000034, -0.00000918, 0.00000692, -0.00000233, -0.00000365, -0.00000792, 0.00000345, -0.00000387, -0.00000076, -0.00000660, 0.00000503, -0.00000040, -0.00000584, 0.00000491, -0.00000157, 0.00000543, -0.00000353, -0.00000608, -0.00000207, 0.00001465, -0.00000181, -0.00001008, 0.00000597, -0.00000092, 0.00000209, -0.00000841, 0.00000224, 0.00000166, -0.00000356, -0.00000056, -0.00000122, 0.00000306, -0.00000205, 0.00000126, -0.00000263, -0.00000466, -0.00000590, 0.00000515, -0.00001164, 0.00001572, -0.00002597, -0.00000633, -0.00000206, 0.00001936, 0.00002312, -0.00000087, 0.00001166, 0.00000699, 0.00000181, 0.00001573, -0.00000256, 0.00001161, 0.00002542, -0.00000911, 0.00000597, -0.00003188, 0.00000425, -0.00001463, 0.00000707, -0.00001093, -0.00000043, -0.00000881, 0.00000563, -0.00001267, 0.00000686, -0.00000742, 0.00000655, -0.00002255, 0.00001030, -0.00001089, },
		{ -0.00001030, 0.00001291, -0.00001497, 0.00001626, -0.00000212, 0.00000068, 0.00000395, -0.00000432, 0.00000039, -0.00000602, -0.00000139, 0.00000185, 0.00000133, -0.00000138, -0.00000199, -0.00000019, 0.00000381, -0.00000473, -0.00000013, 0.00000475, 0.00000803, -0.00001525, -0.00000727, 0.00001469, 0.00000230, -0.00001099, -0.00000718, 0.00000394, -0.00000805, 0.00000125, 0.00000050, -0.00000099, -0.00000201, -0.00001044, -0.00000165, -0.00000357, 0.00000278, -0.00002005, -0.00000028, -0.00000518, -0.00000422, -0.00002136, -0.00000116, -0.00003195, -0.00001035, -0.00019092, 0.00002973, 0.00025471, 0.00002847, 0.00021041, 0.00015036, -0.00019448, -0.00004537, -0.00000868, -0.00002078, -0.00001810, -0.00001064, 0.00000232, -0.00000970, -0.00000205, -0.00000447, 0.00000664, -0.00000559, -0.00000559, -0.00000313, 0.00000196, -0.00000439, -0.00000241, -0.00000344, 0.00000762, -0.00000857, 0.00000364, -0.00000068, 0.00000294, 0.00000174, -0.00000696, -0.00000479, 0.00000815, 0.00001080, -0.00001478, 0.00000344, 0.00000051, 0.00000300, -0.00000579, -0.00000337, 0.00000462, -0.00000278, -0.00000136, -0.00000158, 0.00000152, 0.00000072, -0.00000114, 0.00000013, -0.00000385, -0.00000768, 0.00000418, -0.00000783, 0.00000537, -0.00000230, -0.00002679, 0.00000893, -0.00000318, 0.00003664, -0.00000293, 0.00001281, 0.00000266, 0.00001087, -0.00000085, 0.00002114, -0.00001681, 0.00004023, -0.00000946, 0.00001200, -0.00002261, -0.00000937, -0.00000701, -0.00000192, -0.00000320, -0.00000335, -0.00000851, 0.00000365, -0.00000803, -0.00000019, -0.00000227, 0.00000308, -0.00001212, -0.00000557, 0.00000398, },
		{ -0.00001766, 0.00001462, -0.00001049, 0.00000493, 0.00000895, -0.00000549, 0.00000658, -0.00000304, -0.00000132, -0.00000221, -0.00000569, 0.00000245, 0.00000063, 0.00000091, -0.00000298, -0.00000087, 0.00000281, -0.00000042, -0.00000461, 0.00000409, 0.00000769, -0.00000307, -0.00001928, 0.00001171, 0.00000806, -0.00000555, -0.00001281, 0.00000325, -0.00000409, -0.00000342, 0.00000280, -0.00000311, 0.00000235, -0.00001035, -0.00000370, -0.00000541, 0.00000455, -0.00001178, -0.00001163, -0.00000104, -0.00000238, -0.00001665, -0.00000797, -0.00002040, -0.00001167, -0.00009123, -0.00017237, 0.00023564, 0.00013686, 0.00008692, 0.00028478, -0.00011850, -0.00013229, 0.00000346, -0.00002548, -0.00001237, -0.00002324, 0.00000519, -0.00000891, -0.00000211, -0.00000703, 0.00000386, 0.00000048, -0.00000630, -0.00000580, 0.00000160, -0.00000256, -0.00000251, -0.00000524, 0.00000497, -0.00000098, -0.00000427, 0.00000497, -0.00000331, 0.00000747, -0.00000714, -0.00000427, -0.00000141, 0.00001836, -0.00001011, -0.00000431, 0.00000312, 0.00000209, -0.00000127, -0.00000727, 0.00000321, 0.00000113, -0.00000365, 0.00000022, -0.00000228, 0.00000419, -0.00000323, 0.00000197, -0.00000348, -0.00000547, -0.00000298, 0.00000287, -0.00001017, 0.00001678, -0.00003293, 0.00000613, -0.00000776, 0.00002740, 0.00001106, 0.00000566, 0.00000438, 0.00001173, -0.00000289, 0.00001913, -0.00000515, 0.00001058, 0.00002422, -0.00001138, 0.00000920, -0.00003110, 0.00000406, -0.00001322, 0.00000450, -0.00000809, -0.00000187, -0.00000701, 0.00000450, -0.00001228, 0.00000651, -0.00000597, 0.00000308, -0.00002067, 0.00001382, },
		{ -0.00000588, -0.00000188, 0.00000784, -0.00001247, 0.00001860, -0.00000737, 0.00000528, 0.00000030, -0.00000324, -0.00000012, -0.00000582, -0.00000069, 0.00000188, 0.00000089, -0.00000133, -0.00000254, 0.00000144, 0.00000280, -0.00000548, 0.00000120, 0.00000600, 0.00000604, -0.00001909, -0.00000148, 0.00001473, -0.00000108, -0.00001188, -0.00000389, 0.00000242, -0.00000762, 0.00000423, -0.00000430, 0.00000291, -0.00000552, -0.00000774, -0.00000446, -0.00000046, 0.00000012, -0.00001779, -0.00000350, -0.00000136, -0.00000771, -0.00001719, -0.00000739, -0.00002639, -0.00001767, -0.00019225, 0.00000875, 0.00029895, 0.00001373, 0.00029084, 0.00006005, -0.00020738, -0.00001673, -0.00002169, -0.00000969, -0.00002733, -0.00000308, -0.00000235, -0.00000528, -0.00000473, -0.00000288, 0.00000467, -0.00000367, -0.00000763, -0.00000124, 0.00000003, -0.00000314, -0.00000405, -0.00000120, 0.00000588, -0.00000742, 0.00000520, -0.00000338, 0.00000515, -0.00000094, -0.00000656, -0.00000483, 0.00001323, 0.00000482, -0.00001434, 0.00000537, -0.00000034, 0.00000390, -0.00000872, -0.00000016, 0.00000347, -0.00000277, -0.00000075, -0.00000217, 0.00000203, 0.00000028, -0.00000112, -0.00000004, -0.00000480, -0.00000673, 0.00000605, -0.00001184, 0.00001257, -0.00001325, -0.00001610, 0.00000687, 0.00000166, 0.00003161, -0.00000516, 0.00001400, -0.00000092, 0.00001391, -0.00000513, 0.00002590, -0.00002298, 0.00004294, -0.00001442, 0.00001456, -0.00002136, -0.00000870, -0.00000666, -0.00000231, -0.00000390, -0.00000098, -0.00001032, 0.00000625, -0.00001106, 0.00000252, -0.00000422, 0.00000502, -0.00001777, 0.00000448, },
		{ 0.00001136, -0.00001907, 0.00002001, -0.00001773, 0.00001457, 0.00000055, -0.00000077, 0.00000419, -0.00000307, -0.00000169, -0.00000169, -0.00000601, 0.00000400, -0.00000113, 0.00000233, -0.00000497, 0.00000146, 0.00000181, -0.00000065, -0.00000515, 0.00000694, 0.00000609, -0.00000610, -0.00001880, 0.00001802, 0.00000182, -0.00000462, -0.00001405, 0.00000808, -0.00001043, 0.00000539, -0.00000536, 0.00000264, -0.00000283, -0.00000562, -0.00000892, 0.00000013, -0.00000084, -0.00000596, -0.00001877, 0.00000657, -0.00001023, -0.00000710, -0.00001945, -0.00000924, -0.00002560, -0.00007847, -0.00020694, 0.00026861, 0.00013526, 0.00014101, 0.00027840, -0.00021223, -0.00007344, -0.00002059, -0.00000638, -0.00002648, -0.00001264, -0.00000211, -0.00000197, -0.00000712, -0.00000239, -0.00000138, 0.00000525, -0.00001079, -0.00000178, -0.00000211, 0.00000034, -0.00000565, -0.00000229, 0.00000331, -0.00000045, -0.00000272, 0.00000414, -0.00000410, 0.00000851, -0.00001013, -0.00000253, 0.00000059, 0.00001933, -0.00001705, 0.00000216, -0.00000072, 0.00000589, -0.00000598, -0.00000422, 0.00000278, 0.00000108, -0.00000375, 0.00000083, -0.00000307, 0.00000504, -0.00000429, 0.00000260, -0.00000443, -0.00000598, 0.00000052, -0.00000150, -0.00000484, 0.00001197, -0.00003307, 0.00001414, -0.00001168, 0.00003330, -0.00000046, 0.00001331, -0.00000394, 0.00001798, -0.00000939, 0.00002455, -0.00000969, 0.00001136, 0.00002148, -0.00001179, 0.00001057, -0.00002894, 0.00000291, -0.00001117, 0.00000105, -0.00000391, -0.00000474, -0.00000354, 0.00000117, -0.00000961, 0.00000426, -0.00000278, -0.00000316, -0.00001384, },
		{ 0.00001208, -0.00001571, 0.00000883, -0.00000117, -0.00000500, 0.00001571, -0.00000872, 0.00000774, -0.00000264, -0.00000227, -0.00000046, -0.00000579, 0.00000051, 0.00000102, 0.00000131, -0.00000216, -0.00000203, 0.00000230, 0.00000183, -0.00000619, 0.00000331, 0.00000616, 0.00000376, -0.00002221, 0.00000611, 0.00001128, -0.00000238, -0.00001355, 0.00000115, -0.00000233, -0.00000287, 0.00000259, -0.00000531, 0.00000368, -0.00000620, -0.00000765, -0.00000449, -0.00000024, 0.00000031, -0.00001858, -0.00000309, -0.00000170, -0.00000650, -0.00001814, -0.00000831, -0.00002648, -0.00001975, -0.00019908, -0.00000656, 0.00033597, 0.00001760, 0.00036278, -0.00006139, -0.00017869, -0.00001141, -0.00002199, -0.00000796, -0.00002937, -0.00000069, -0.00000311, -0.00000405, -0.00000472, -0.00000391, 0.00000518, -0.00000416, -0.00000732, -0.00000145, -0.00000021, -0.00000338, -0.00000418, -0.00000004, 0.00000445, -0.00000577, 0.00000613, -0.00000615, 0.00000784, -0.00000463, -0.00000499, -0.00000485, 0.00001832, -0.00000358, -0.00001015, 0.00000441, 0.00000106, 0.00000236, -0.00000948, 0.00000207, 0.00000241, -0.00000277, 0.00000001, -0.00000302, 0.00000289, -0.00000064, -0.00000073, -0.00000049, -0.00000585, -0.00000441, 0.00000541, -0.00001281, 0.00001688, -0.00002311, -0.00000271, 0.00000013, 0.00001040, 0.00002131, -0.00000176, 0.00001070, -0.00000011, 0.00001290, -0.00000570, 0.00002743, -0.00002635, 0.00004318, -0.00001683, 0.00001505, -0.00001868, -0.00000931, -0.00000494, -0.00000457, -0.00000234, -0.00000079, -0.00000984, 0.00000639, -0.00001198, 0.00000360, -0.00000439, 0.00000417, -0.00001978, },
		{ -0.00000370, 0.00000344, -0.00001417, 0.00001949, -0.00002023, 0.00002123, -0.00000734, 0.00000507, 0.00000013, -0.00000184, -0.00000262, -0.00000093, -0.00000613, 0.00000514, -0.00000270, 0.00000362, -0.00000686, 0.00000415, 0.00000009, -0.00000064, -0.00000508, 0.00000936, 0.00000387, -0.00000939, -0.00001552, 0.00002164, -0.00000411, -0.00000347, -0.00001389, 0.00001064, -0.00001468, 0.00001316, -0.00001397, 0.00000947, -0.00000924, 0.00000037, -0.00001523, 0.00000680, -0.00000750, 0.00000098, -0.00002677, 0.00001516, -0.00001936, 0.00000427, -0.00003353, 0.00000526, -0.00004422, -0.00006046, -0.00024753, 0.00030738, 0.00012961, 0.00022152, 0.00021830, -0.00027093, -0.00001870, -0.00004778, 0.00001517, -0.00004310, 0.00000077, -0.00001183, 0.00000681, -0.00001353, 0.00000307, -0.00000722, 0.00001047, -0.00001568, 0.00000282, -0.00000657, 0.00000388, -0.00000933, 0.00000154, 0.00000046, 0.00000120, -0.00000146, 0.00000245, -0.00000366, 0.00000804, -0.00001184, -0.00000127, 0.00000412, 0.00001665, -0.00002075, 0.00000734, -0.00000395, 0.00000901, -0.00001065, -0.00000019, 0.00000131, 0.00000142, -0.00000377, 0.00000112, -0.00000340, 0.00000536, -0.00000501, 0.00000301, -0.00000555, -0.00000562, 0.00000339, -0.00000636, 0.00000274, 0.00000226, -0.00002566, 0.00001508, -0.00001065, 0.00003395, -0.00000802, 0.00001882, -0.00001038, 0.00002292, -0.00001508, 0.00002967, -0.00001426, 0.00001233, 0.00001857, -0.00001152, 0.00001107, -0.00002632, 0.00000172, -0.00000955, -0.00000191, 0.00000015, -0.00000759, 0.00000011, -0.00000295, -0.00000581, 0.00000117, 0.00000079, -0.00001037, },
		{ -0.00001662, 0.00001440, -0.00002182, 0.00001813, -0.00001099, 0.00000524, 0.00000859, -0.00000597, 0.00000721, -0.00000398, -0.00000200, -0.00000031, -0.00000601, 0.00000210, -0.00000062, 0.00000258, -0.00000389, -0.00000028, 0.00000202, 0.00000120, -0.00000676, 0.00000591, 0.00000524, 0.00000113, -0.00002372, 0.00001407, 0.00000506, -0.00000159, -0.00001573, 0.00000705, -0.00000924, 0.00000553, -0.00000373, -0.00000185, 0.00000036, -0.00000304, -0.00001120, -0.00000076, -0.00000394, 0.00000437, -0.00002339, 0.00000198, -0.00000740, 0.00000088, -0.00002614, -0.00000147, -0.00003532, -0.00001295, -0.00021542, -0.00001171, 0.00035956, 0.00005418, 0.00040090, -0.00019035, -0.00011800, -0.00002949, -0.00000751, -0.00001772, -0.00002205, -0.00000547, 0.00000169, -0.00000735, -0.00000102, -0.00000828, 0.00000879, -0.00000759, -0.00000410, -0.00000436, 0.00000183, -0.00000564, -0.00000224, -0.00000066, 0.00000406, -0.00000404, 0.00000628, -0.00000842, 0.00001029, -0.00000860, -0.00000279, -0.00000415, 0.00002189, -0.00001253, -0.00000342, 0.00000108, 0.00000438, -0.00000149, -0.00000775, 0.00000271, 0.00000190, -0.00000293, 0.00000084, -0.00000392, 0.00000386, -0.00000184, -0.00000010, -0.00000123, -0.00000673, -0.00000100, 0.00000209, -0.00000964, 0.00001587, -0.00002789, 0.00000867, -0.00000739, 0.00002003, 0.00000834, 0.00000543, 0.00000411, 0.00000419, 0.00000835, -0.00000272, 0.00002557, -0.00002659, 0.00004058, -0.00001627, 0.00001310, -0.00001429, -0.00001130, -0.00000196, -0.00000846, 0.00000135, -0.00000285, -0.00000686, 0.00000361, -0.00001009, 0.00000230, -0.00000219, -0.00000017, },
		{ -0.00001591, 0.00000666, -0.00000738, -0.00000429, 0.00001301, -0.00001674, 0.00002275, -0.00001218, 0.00000986, -0.00000415, -0.00000015, -0.00000366, -0.00000024, -0.00000574, 0.00000543, -0.00000352, 0.00000419, -0.00000794, 0.00000645, -0.00000192, -0.00000048, -0.00000427, 0.00001090, 0.00000129, -0.00001261, -0.00000950, 0.00002143, -0.00000817, -0.00000316, -0.00001159, 0.00000975, -0.00001485, 0.00001716, -0.00002007, 0.00001457, -0.00001425, 0.00000512, -0.00002014, 0.00001202, -0.00001279, 0.00000631, -0.00003265, 0.00002136, -0.00002594, 0.00001309, -0.00004533, 0.00001774, -0.00006179, -0.00004332, -0.00028705, 0.00034249, 0.00013475, 0.00031272, 0.00010650, -0.00028061, 0.00001512, -0.00006592, 0.00003038, -0.00005529, 0.00001151, -0.00001973, 0.00001402, -0.00001851, 0.00000693, -0.00001142, 0.00001406, -0.00001900, 0.00000614, -0.00001013, 0.00000665, -0.00001223, 0.00000503, -0.00000268, 0.00000360, -0.00000062, 0.00000014, -0.00000204, 0.00000593, -0.00001198, -0.00000053, 0.00000877, 0.00001033, -0.00002014, 0.00000973, -0.00000520, 0.00001007, -0.00001379, 0.00000355, -0.00000049, 0.00000186, -0.00000357, 0.00000099, -0.00000315, 0.00000501, -0.00000521, 0.00000308, -0.00000679, -0.00000397, 0.00000435, -0.00000955, 0.00000970, -0.00000930, -0.00001278, 0.00000912, -0.00000371, 0.00002795, -0.00000953, 0.00001986, -0.00001266, 0.00002421, -0.00001758, 0.00003216, -0.00001676, 0.00001165, 0.00001719, -0.00001202, 0.00001195, -0.00002433, 0.00000148, -0.00000944, -0.00000307, 0.00000258, -0.00000880, 0.00000216, -0.00000600, -0.00000260, -0.00000115, 0.00000283, },
		{ -0.00000723, -0.00000845, 0.00001024, -0.00002160, 0.00002256, -0.00001803, 0.00001494, -0.00000033, -0.00000038, 0.00000397, -0.00000355, -0.00000258, 0.00000031, -0.00000632, 0.00000372, -0.00000261, 0.00000429, -0.00000615, 0.00000243, 0.00000057, 0.00000107, -0.00000704, 0.00000854, 0.00000351, -0.00000194, -0.00002264, 0.00002047, -0.00000225, 0.00000042, -0.00001748, 0.00001218, -0.00001611, 0.00001530, -0.00001295, 0.00000492, -0.00000621, 0.00000326, -0.00001769, 0.00000602, -0.00001086, 0.00001155, -0.00003130, 0.00001060, -0.00001727, 0.00001320, -0.00004007, 0.00001228, -0.00005244, 0.00000259, -0.00024105, -0.00000739, 0.00037034, 0.00013056, 0.00038223, -0.00029764, -0.00004480, -0.00006194, 0.00001701, -0.00003588, -0.00000730, -0.00001623, 0.00001114, -0.00001438, 0.00000555, -0.00001512, 0.00001461, -0.00001314, 0.00000131, -0.00000943, 0.00000576, -0.00000957, 0.00000152, -0.00000310, 0.00000503, -0.00000254, 0.00000551, -0.00000959, 0.00001166, -0.00001192, -0.00000062, -0.00000209, 0.00002243, -0.00001970, 0.00000380, -0.00000318, 0.00000842, -0.00000663, -0.00000403, 0.00000177, 0.00000204, -0.00000321, 0.00000157, -0.00000461, 0.00000460, -0.00000298, 0.00000059, -0.00000228, -0.00000697, 0.00000257, -0.00000298, -0.00000279, 0.00000889, -0.00002513, 0.00001393, -0.00001138, 0.00002661, -0.00000343, 0.00001313, -0.00000303, 0.00000950, 0.00000241, 0.00000207, 0.00002165, -0.00002474, 0.00003601, -0.00001343, 0.00000926, -0.00000878, -0.00001402, 0.00000142, -0.00001290, 0.00000609, -0.00000614, -0.00000236, -0.00000129, -0.00000587, -0.00000101, 0.00000179, },
		{ 0.00000173, -0.00001724, 0.00001367, -0.00001720, 0.00000729, 0.00000269, -0.00000810, 0.00001868, -0.00001255, 0.00001187, -0.00000721, 0.00000112, -0.00000429, 0.00000007, -0.00000468, 0.00000464, -0.00000324, 0.00000360, -0.00000765, 0.00000766, -0.00000367, -0.00000031, -0.00000273, 0.00001134, -0.00000156, -0.00001503, -0.00000164, 0.00001730, -0.00000950, -0.00000414, -0.00000717, 0.00000529, -0.00001017, 0.00001575, -0.00002143, 0.00001562, -0.00001556, 0.00000636, -0.00002137, 0.00001345, -0.00001431, 0.00000757, -0.00003375, 0.00002215, -0.00002651, 0.00001528, -0.00005006, 0.00002261, -0.00007204, -0.00003394, -0.00031748, 0.00036278, 0.00017039, 0.00038767, -0.00003917, -0.00024093, 0.00001921, -0.00006675, 0.00003262, -0.00005740, 0.00001485, -0.00002196, 0.00001653, -0.00001954, 0.00000709, -0.00001211, 0.00001427, -0.00001906, 0.00000656, -0.00001127, 0.00000731, -0.00001308, 0.00000694, -0.00000505, 0.00000619, -0.00000029, -0.00000237, 0.00000032, 0.00000251, -0.00001067, 0.00000009, 0.00001337, 0.00000152, -0.00001540, 0.00000879, -0.00000385, 0.00000833, -0.00001431, 0.00000587, -0.00000188, 0.00000207, -0.00000306, 0.00000045, -0.00000240, 0.00000401, -0.00000484, 0.00000268, -0.00000795, -0.00000100, 0.00000261, -0.00000922, 0.00001295, -0.00001851, 0.00000130, -0.00000077, 0.00000738, 0.00001634, -0.00000511, 0.00001602, -0.00001018, 0.00002096, -0.00001575, 0.00003070, -0.00001587, 0.00000809, 0.00001850, -0.00001437, 0.00001409, -0.00002364, 0.00000267, -0.00001129, -0.00000176, 0.00000251, -0.00000737, 0.00000134, -0.00000647, -0.00000148, -0.00000133, },
		{ 0.00000592, -0.00001486, 0.00000261, 0.00000068, -0.00001460, 0.00002025, -0.00001946, 0.00002087, -0.00000799, 0.00000580, -0.00000068, -0.00000193, -0.00000367, 0.00000110, -0.00000634, 0.00000480, -0.00000423, 0.00000571, -0.00000816, 0.00000539, -0.00000165, 0.00000130, -0.00000676, 0.00001063, 0.00000131, -0.00000543, -0.00001838, 0.00002346, -0.00000857, 0.00000219, -0.00001751, 0.00001461, -0.00002025, 0.00002323, -0.00002189, 0.00001222, -0.00001352, 0.00001035, -0.00002480, 0.00001346, -0.00001853, 0.00001925, -0.00003942, 0.00001953, -0.00002768, 0.00002645, -0.00005558, 0.00002832, -0.00007312, 0.00002194, -0.00027025, -0.00000134, 0.00037902, 0.00024080, 0.00029443, -0.00035690, 0.00001757, -0.00009401, 0.00004169, -0.00005502, 0.00000926, -0.00002869, 0.00002180, -0.00002228, 0.00001247, -0.00002206, 0.00002042, -0.00001868, 0.00000696, -0.00001497, 0.00001014, -0.00001387, 0.00000599, -0.00000660, 0.00000708, -0.00000135, 0.00000375, -0.00000923, 0.00001124, -0.00001376, 0.00000094, 0.00000161, 0.00001896, -0.00002304, 0.00000928, -0.00000653, 0.00001145, -0.00001139, 0.00000043, -0.00000014, 0.00000256, -0.00000339, 0.00000194, -0.00000479, 0.00000475, -0.00000374, 0.00000108, -0.00000358, -0.00000609, 0.00000504, -0.00000794, 0.00000564, -0.00000238, -0.00001501, 0.00001125, -0.00000897, 0.00002690, -0.00001041, 0.00001791, -0.00000765, 0.00001284, -0.00000208, 0.00000604, 0.00001789, -0.00002271, 0.00003118, -0.00000982, 0.00000482, -0.00000327, -0.00001636, 0.00000393, -0.00001635, 0.00001019, -0.00000897, 0.00000188, -0.00000654, -0.00000084, -0.00000495, },
		{ 0.00000109, -0.00000302, -0.00001303, 0.00001461, -0.00002222, 0.00001645, -0.00000784, 0.00000296, 0.00001049, -0.00000865, 0.00001051, -0.00000826, 0.00000140, -0.00000422, -0.00000006, -0.00000309, 0.00000294, -0.00000192, 0.00000186, -0.00000584, 0.00000737, -0.00000474, -0.00000016, -0.00000070, 0.00001070, -0.00000475, -0.00001564, 0.00000632, 0.00001036, -0.00000840, -0.00000601, -0.00000169, -0.00000142, -0.00000168, 0.00000921, -0.00001766, 0.00001189, -0.00001234, 0.00000325, -0.00001808, 0.00001019, -0.00001116, 0.00000393, -0.00002922, 0.00001656, -0.00001981, 0.00000913, -0.00004542, 0.00001687, -0.00007128, -0.00003650, -0.00033401, 0.00036075, 0.00025462, 0.00041412, -0.00018784, -0.00016642, -0.00000387, -0.00004964, 0.00002066, -0.00004783, 0.00000909, -0.00001707, 0.00001317, -0.00001575, 0.00000293, -0.00000882, 0.00001069, -0.00001542, 0.00000353, -0.00000941, 0.00000530, -0.00001132, 0.00000655, -0.00000587, 0.00000845, -0.00000058, -0.00000444, 0.00000258, -0.00000139, -0.00000851, 0.00000129, 0.00001633, -0.00000765, -0.00000803, 0.00000524, -0.00000040, 0.00000408, -0.00001203, 0.00000622, -0.00000240, 0.00000190, -0.00000228, -0.00000032, -0.00000138, 0.00000256, -0.00000402, 0.00000173, -0.00000862, 0.00000272, -0.00000163, -0.00000469, 0.00001046, -0.00002162, 0.00001167, -0.00000996, 0.00001877, 0.00000244, 0.00000285, 0.00000901, -0.00000438, 0.00001427, -0.00001027, 0.00002558, -0.00001167, 0.00000165, 0.00002260, -0.00001872, 0.00001756, -0.00002416, 0.00000497, -0.00001460, 0.00000161, 0.00000016, -0.00000341, -0.00000248, -0.00000389, -0.00000304, },
		{ -0.00000770, 0.00000812, -0.00002021, 0.00001310, -0.00001110, -0.00000281, 0.00001164, -0.00001435, 0.00002100, -0.00001183, 0.00001019, -0.00000501, -0.00000017, -0.00000460, 0.00000157, -0.00000563, 0.00000477, -0.00000478, 0.00000605, -0.00000900, 0.00000761, -0.00000391, 0.00000160, -0.00000568, 0.00001167, -0.00000102, -0.00000901, -0.00001103, 0.00002189, -0.00001212, 0.00000240, -0.00001481, 0.00001294, -0.00001943, 0.00002625, -0.00002711, 0.00001667, -0.00001834, 0.00001506, -0.00002934, 0.00001826, -0.00002356, 0.00002395, -0.00004391, 0.00002446, -0.00003375, 0.00003503, -0.00006636, 0.00003960, -0.00008958, 0.00003659, -0.00029296, -0.00000729, 0.00040642, 0.00036322, 0.00014243, -0.00035325, 0.00004973, -0.00011090, 0.00005553, -0.00006630, 0.00002047, -0.00003715, 0.00002902, -0.00002717, 0.00001638, -0.00002606, 0.00002338, -0.00002155, 0.00001034, -0.00001870, 0.00001296, -0.00001670, 0.00000948, -0.00000980, 0.00000947, -0.00000033, 0.00000105, -0.00000725, 0.00000870, -0.00001358, 0.00000169, 0.00000657, 0.00001152, -0.00002144, 0.00001132, -0.00000741, 0.00001193, -0.00001406, 0.00000410, -0.00000204, 0.00000295, -0.00000314, 0.00000169, -0.00000421, 0.00000405, -0.00000379, 0.00000108, -0.00000492, -0.00000386, 0.00000528, -0.00001064, 0.00001260, -0.00001440, -0.00000047, 0.00000185, 0.00000004, 0.00001989, -0.00001073, 0.00001764, -0.00000765, 0.00001203, -0.00000281, 0.00000689, 0.00001646, -0.00002245, 0.00002791, -0.00000706, 0.00000115, 0.00000111, -0.00001732, 0.00000452, -0.00001749, 0.00001214, -0.00000967, 0.00000399, -0.00001012, 0.00000306, },
		{ -0.00000657, 0.00000524, -0.00001033, -0.00000390, 0.00000875, -0.00002012, 0.00001972, -0.00001467, 0.00001277, 0.00000107, -0.00000220, 0.00000644, -0.00000739, 0.00000071, -0.00000359, -0.00000034, -0.00000145, 0.00000091, -0.00000015, -0.00000045, -0.00000296, 0.00000569, -0.00000506, 0.00000014, 0.00000135, 0.00000929, -0.00000843, -0.00001339, 0.00001220, 0.00000272, -0.00000618, -0.00000743, 0.00000287, -0.00000778, 0.00000782, -0.00000004, -0.00001057, 0.00000484, -0.00000589, -0.00000295, -0.00001155, 0.00000355, -0.00000474, -0.00000297, -0.00002094, 0.00000663, -0.00000805, -0.00000318, -0.00003347, 0.00000229, -0.00006103, -0.00004927, -0.00033911, 0.00033830, 0.00039591, 0.00036371, -0.00030303, -0.00008274, -0.00004087, -0.00002229, -0.00000018, -0.00003025, -0.00000321, -0.00000705, 0.00000564, -0.00000872, -0.00000397, -0.00000304, 0.00000474, -0.00000935, -0.00000193, -0.00000537, 0.00000135, -0.00000749, 0.00000410, -0.00000502, 0.00001019, -0.00000172, -0.00000532, 0.00000370, -0.00000463, -0.00000633, 0.00000384, 0.00001600, -0.00001462, -0.00000041, 0.00000084, 0.00000364, -0.00000134, -0.00000775, 0.00000484, -0.00000210, 0.00000149, -0.00000153, -0.00000095, -0.00000054, 0.00000109, -0.00000300, 0.00000028, -0.00000832, 0.00000613, -0.00000711, 0.00000308, 0.00000219, -0.00001684, 0.00001465, -0.00001405, 0.00002604, -0.00000945, 0.00001067, 0.00000194, 0.00000190, 0.00000671, -0.00000338, 0.00001861, -0.00000566, -0.00000637, 0.00002843, -0.00002421, 0.00002159, -0.00002504, 0.00000731, -0.00001805, 0.00000567, -0.00000313, 0.00000176, -0.00000813, 0.00000087, },
		{ 0.00000435, -0.00000789, 0.00000680, -0.00002078, 0.00001935, -0.00002090, 0.00000993, -0.00000052, -0.00000406, 0.00001525, -0.00001065, 0.00001113, -0.00000751, 0.00000071, -0.00000474, 0.00000131, -0.00000393, 0.00000332, -0.00000381, 0.00000477, -0.00000796, 0.00000819, -0.00000542, 0.00000159, -0.00000364, 0.00001133, -0.00000331, -0.00001196, -0.00000165, 0.00001583, -0.00001209, 0.00000046, -0.00000929, 0.00000703, -0.00001291, 0.00002269, -0.00002626, 0.00001570, -0.00001806, 0.00001477, -0.00002866, 0.00001767, -0.00002312, 0.00002273, -0.00004162, 0.00002182, -0.00003136, 0.00003408, -0.00006667, 0.00003934, -0.00009403, 0.00003767, -0.00029853, -0.00004066, 0.00047913, 0.00046259, -0.00004959, -0.00028901, 0.00004304, -0.00010377, 0.00005110, -0.00006319, 0.00002071, -0.00003697, 0.00002898, -0.00002590, 0.00001464, -0.00002476, 0.00002134, -0.00001966, 0.00000939, -0.00001870, 0.00001246, -0.00001644, 0.00001040, -0.00001131, 0.00001129, 0.00000069, -0.00000228, -0.00000399, 0.00000430, -0.00001141, 0.00000189, 0.00001169, 0.00000138, -0.00001522, 0.00000944, -0.00000518, 0.00000910, -0.00001352, 0.00000573, -0.00000302, 0.00000272, -0.00000223, 0.00000072, -0.00000285, 0.00000241, -0.00000297, 0.00000031, -0.00000586, -0.00000057, 0.00000286, -0.00000954, 0.00001516, -0.00002290, 0.00001372, -0.00001040, 0.00001299, 0.00000729, -0.00000503, 0.00001237, -0.00000291, 0.00000666, 0.00000094, 0.00000361, 0.00001843, -0.00002501, 0.00002727, -0.00000618, -0.00000094, 0.00000377, -0.00001657, 0.00000293, -0.00001589, 0.00001129, -0.00000746, 0.00000288, -0.00001064, },
		{ 0.00000597, -0.00001041, 0.00000922, -0.00001774, 0.00000760, -0.00000225, -0.00001100, 0.00001556, -0.00001508, 0.00001796, -0.00000627, 0.00000417, 0.00000136, -0.00000553, -0.00000034, -0.00000293, -0.00000020, -0.00000054, -0.00000052, 0.00000109, -0.00000228, -0.00000003, 0.00000332, -0.00000496, 0.00000098, 0.00000274, 0.00000760, -0.00001268, -0.00000760, 0.00001395, -0.00000314, -0.00000478, -0.00000668, 0.00000420, -0.00001064, 0.00001453, -0.00000810, -0.00000374, -0.00000220, 0.00000058, -0.00000902, -0.00000511, -0.00000304, 0.00000134, -0.00000918, -0.00001336, -0.00000264, 0.00000322, -0.00001554, -0.00002081, -0.00001409, -0.00004889, -0.00006332, -0.00034459, 0.00031054, 0.00058606, 0.00022204, -0.00035303, -0.00001892, -0.00007247, 0.00000187, -0.00001958, -0.00001269, -0.00001579, 0.00000301, -0.00000179, -0.00000223, -0.00001015, 0.00000197, -0.00000052, -0.00000367, -0.00000729, -0.00000133, -0.00000261, -0.00000331, 0.00000089, -0.00000324, 0.00001170, -0.00000405, -0.00000431, 0.00000267, -0.00000606, -0.00000496, 0.00000816, 0.00001131, -0.00001712, 0.00000488, -0.00000222, 0.00000627, -0.00000593, -0.00000310, 0.00000269, -0.00000153, 0.00000130, -0.00000124, -0.00000097, -0.00000039, 0.00000013, -0.00000224, -0.00000144, -0.00000674, 0.00000802, -0.00001184, 0.00001166, -0.00000958, -0.00000519, 0.00000924, -0.00001059, 0.00002606, -0.00001570, 0.00001486, -0.00000205, 0.00000557, 0.00000131, 0.00000208, 0.00001231, -0.00000005, -0.00001397, 0.00003423, -0.00002942, 0.00002495, -0.00002510, 0.00000834, -0.00002002, 0.00000862, -0.00000547, 0.00000611, -0.00001354, },
		{ -0.00000795, 0.00000274, -0.00000531, 0.00000174, -0.00001558, 0.00001892, -0.00002495, 0.00001923, -0.00001214, 0.00000821, 0.00000557, -0.00000508, 0.00000830, -0.00000752, 0.00000018, -0.00000383, 0.00000027, -0.00000140, 0.00000065, -0.00000144, 0.00000192, -0.00000492, 0.00000667, -0.00000564, 0.00000109, -0.00000077, 0.00000958, -0.00000558, -0.00001327, 0.00000786, 0.00000668, -0.00000901, -0.00000304, -0.00000214, -0.00000161, -0.00000203, 0.00001320, -0.00001922, 0.00000882, -0.00001203, 0.00000878, -0.00002199, 0.00001087, -0.00001636, 0.00001484, -0.00003181, 0.00001075, -0.00001938, 0.00002200, -0.00005426, 0.00002446, -0.00008249, 0.00002044, -0.00028105, -0.00011171, 0.00062139, 0.00049757, -0.00024206, -0.00018438, 0.00000305, -0.00007362, 0.00002822, -0.00004475, 0.00000868, -0.00002691, 0.00002067, -0.00001771, 0.00000673, -0.00001778, 0.00001396, -0.00001265, 0.00000363, -0.00001436, 0.00000809, -0.00001245, 0.00000797, -0.00001023, 0.00001189, 0.00000171, -0.00000564, -0.00000032, -0.00000111, -0.00000782, 0.00000222, 0.00001531, -0.00000916, -0.00000609, 0.00000455, -0.00000050, 0.00000341, -0.00000971, 0.00000485, -0.00000263, 0.00000169, -0.00000071, -0.00000080, -0.00000095, 0.00000005, -0.00000136, -0.00000137, -0.00000586, 0.00000295, -0.00000163, -0.00000444, 0.00001180, -0.00002456, 0.00002271, -0.00002044, 0.00002535, -0.00000701, 0.00000378, 0.00000427, 0.00000468, -0.00000170, 0.00000800, -0.00000305, 0.00002330, -0.00003006, 0.00002910, -0.00000713, -0.00000154, 0.00000501, -0.00001466, -0.00000008, -0.00001225, 0.00000822, -0.00000281, -0.00000121, },
		{ -0.00001665, 0.00000936, -0.00001262, 0.00001085, -0.00002227, 0.00001772, -0.00001471, 0.00000251, 0.00000498, -0.00000851, 0.00001669, -0.00000896, 0.00000791, -0.00000271, -0.00000400, -0.00000084, -0.00000298, 0.00000104, -0.00000115, -0.00000040, 0.00000073, -0.00000240, 0.00000160, 0.00000138, -0.00000511, 0.00000288, 0.00000273, 0.00000619, -0.00001732, 0.00000169, 0.00001030, -0.00000516, -0.00000592, -0.00000235, 0.00000055, -0.00000744, 0.00001493, -0.00001092, -0.00000127, -0.00000520, 0.00000305, -0.00001086, -0.00000301, -0.00000517, 0.00000248, -0.00000968, -0.00001209, -0.00000491, 0.00000674, -0.00001969, -0.00001685, -0.00002169, -0.00004675, -0.00006438, -0.00036991, 0.00030544, 0.00079723, -0.00000388, -0.00032087, 0.00000194, -0.00008040, 0.00000886, -0.00002609, -0.00000459, -0.00002097, 0.00000676, -0.00000374, -0.00000093, -0.00001140, 0.00000234, -0.00000143, -0.00000186, -0.00000938, -0.00000016, -0.00000395, -0.00000117, -0.00000105, -0.00000191, 0.00001361, -0.00000790, -0.00000098, -0.00000119, -0.00000480, -0.00000493, 0.00001407, 0.00000221, -0.00001394, 0.00000595, -0.00000214, 0.00000568, -0.00000769, 0.00000007, 0.00000110, -0.00000152, 0.00000190, -0.00000190, 0.00000007, -0.00000143, 0.00000021, -0.00000227, -0.00000290, -0.00000397, 0.00000757, -0.00001386, 0.00001798, -0.00002103, 0.00000973, -0.00000240, -0.00000020, 0.00001830, -0.00001480, 0.00001357, -0.00000115, 0.00000470, 0.00000019, 0.00000387, 0.00000882, 0.00000317, -0.00001925, 0.00003832, -0.00003296, 0.00002655, -0.00002341, 0.00000712, -0.00001931, 0.00000905, -0.00000531, 0.00000781, },
		{ -0.00000671, -0.00000332, -0.00000074, -0.00000115, -0.00000558, -0.00000572, 0.00001163, -0.00002135, 0.00002175, -0.00001920, 0.00001847, -0.00000473, 0.00000271, 0.00000278, -0.00000540, -0.00000153, -0.00000215, -0.00000112, 0.00000133, -0.00000248, 0.00000153, -0.00000172, -0.00000053, 0.00000335, -0.00000459, 0.00000028, 0.00000243, 0.00000678, -0.00000801, -0.00001186, 0.00001513, -0.00000307, -0.00000461, -0.00000644, 0.00000436, -0.00000999, 0.00000990, 0.00000080, -0.00000847, -0.00000188, -0.00000218, -0.00000102, -0.00001128, -0.00000016, -0.00000541, 0.00000266, -0.00001719, -0.00000578, -0.00000102, 0.00000208, -0.00003242, -0.00000192, -0.00005785, -0.00001193, -0.00024431, -0.00021859, 0.00084529, 0.00043201, -0.00038935, -0.00007240, -0.00005199, -0.00003167, -0.00000512, -0.00001676, -0.00001142, -0.00001029, 0.00000687, -0.00000514, -0.00000495, -0.00000739, 0.00000340, -0.00000244, -0.00000532, -0.00000706, 0.00000100, -0.00000569, 0.00000278, -0.00000674, 0.00001126, 0.00000236, -0.00000813, 0.00000250, -0.00000617, -0.00000385, 0.00000353, 0.00001572, -0.00001731, 0.00000320, -0.00000128, 0.00000484, -0.00000347, -0.00000380, 0.00000197, -0.00000111, 0.00000016, 0.00000104, -0.00000239, 0.00000094, -0.00000249, 0.00000065, -0.00000381, -0.00000449, 0.00000556, -0.00000666, 0.00000324, 0.00000313, -0.00001830, 0.00002342, -0.00002398, 0.00003251, -0.00001843, 0.00001175, -0.00000327, 0.00001196, -0.00001008, 0.00001573, -0.00001087, 0.00002923, -0.00003599, 0.00003209, -0.00000889, -0.00000158, 0.00000586, -0.00001286, -0.00000299, -0.00000821, 0.00000457, 0.00000260, },
		{ 0.00000607, -0.00001632, 0.00000978, -0.00001199, 0.00000875, -0.00002164, 0.00002289, -0.00002427, 0.00001613, -0.00000876, 0.00000324, 0.00000928, -0.00000606, 0.00000743, -0.00000416, -0.00000401, 0.00000002, -0.00000431, 0.00000379, -0.00000374, 0.00000188, -0.00000195, 0.00000012, 0.00000073, 0.00000103, -0.00000625, 0.00000633, 0.00000070, 0.00000550, -0.00002188, 0.00001351, 0.00000132, -0.00000240, -0.00001044, 0.00000591, -0.00000844, 0.00000272, 0.00000718, -0.00000585, -0.00000613, -0.00000105, -0.00000168, -0.00000516, -0.00000872, 0.00000077, -0.00000509, -0.00000042, -0.00002169, 0.00000506, -0.00000365, -0.00000836, -0.00003023, -0.00001033, -0.00006657, -0.00003759, -0.00043675, 0.00035840, 0.00098489, -0.00028368, -0.00021073, -0.00002947, -0.00005440, -0.00001021, -0.00001177, -0.00001297, -0.00001281, -0.00000079, 0.00000399, -0.00000842, -0.00000451, -0.00000487, 0.00000482, -0.00000665, -0.00000558, -0.00000424, -0.00000041, -0.00000328, 0.00000027, -0.00000251, 0.00001660, -0.00001334, 0.00000462, -0.00000791, -0.00000056, -0.00000629, 0.00002069, -0.00001009, -0.00000536, 0.00000222, 0.00000182, 0.00000100, -0.00000538, 0.00000030, 0.00000122, -0.00000280, 0.00000377, -0.00000382, 0.00000238, -0.00000389, 0.00000168, -0.00000361, -0.00000339, -0.00000071, 0.00000481, -0.00001210, 0.00001952, -0.00002809, 0.00002295, -0.00001575, 0.00001373, 0.00000509, -0.00000789, 0.00000728, 0.00000436, -0.00000070, 0.00000371, 0.00000132, 0.00000897, 0.00000314, -0.00002133, 0.00003983, -0.00003419, 0.00002603, -0.00001989, 0.00000366, -0.00001579, 0.00000663, -0.00000215, },
		{ 0.00000659, -0.00001137, 0.00000115, -0.00000412, 0.00000244, -0.00001244, 0.00000582, -0.00000027, -0.00001083, 0.00001641, -0.00001915, 0.00002336, -0.00001218, 0.00000977, -0.00000334, -0.00000246, -0.00000352, -0.00000051, -0.00000203, 0.00000324, -0.00000488, 0.00000385, -0.00000475, 0.00000381, -0.00000069, -0.00000293, -0.00000020, 0.00000507, 0.00000363, -0.00001086, -0.00000691, 0.00001787, -0.00001054, -0.00000132, -0.00000754, 0.00000743, -0.00001449, 0.00001846, -0.00000984, 0.00000155, -0.00001214, 0.00000731, -0.00001039, -0.00000092, -0.00001089, 0.00000497, -0.00000859, -0.00000358, -0.00002124, 0.00001642, -0.00001759, -0.00001007, -0.00003011, -0.00003076, -0.00004704, -0.00020409, -0.00034276, 0.00114247, 0.00024680, -0.00045207, 0.00001056, -0.00009708, 0.00000426, -0.00003501, 0.00000976, -0.00003082, 0.00000569, -0.00000631, 0.00000644, -0.00001550, 0.00000188, -0.00000610, 0.00000701, -0.00001390, 0.00000008, -0.00000599, 0.00000134, -0.00000312, -0.00000225, 0.00001022, 0.00000189, -0.00000864, 0.00000303, -0.00000930, -0.00000069, 0.00000650, 0.00001167, -0.00002057, 0.00000975, -0.00000549, 0.00000845, -0.00000914, 0.00000219, -0.00000152, 0.00000061, -0.00000112, 0.00000230, -0.00000333, 0.00000206, -0.00000443, 0.00000236, -0.00000651, -0.00000170, 0.00000623, -0.00001025, 0.00001088, -0.00000816, -0.00000590, 0.00001570, -0.00001921, 0.00003163, -0.00002346, 0.00001546, -0.00000717, 0.00001588, -0.00001542, 0.00002114, -0.00001714, 0.00003381, -0.00004063, 0.00003436, -0.00000995, -0.00000236, 0.00000756, -0.00001253, -0.00000412, -0.00000563, 0.00000231, },
		{ 0.00000309, 0.00000158, -0.00001286, 0.00000729, -0.00000866, 0.00000310, -0.00001532, 0.00002112, -0.00002741, 0.00002544, -0.00002126, 0.00001656, -0.00000194, 0.00000144, 0.00000265, -0.00000243, -0.00000609, 0.00000257, -0.00000703, 0.00000798, -0.00000816, 0.00000622, -0.00000696, 0.00000549, -0.00000321, 0.00000301, -0.00000884, 0.00001145, -0.00000359, 0.00000575, -0.00002553, 0.00002601, -0.00001151, 0.00000456, -0.00001780, 0.00001700, -0.00002128, 0.00001845, -0.00000798, 0.00000714, -0.00001884, 0.00001098, -0.00001442, 0.00000897, -0.00002327, 0.00001585, -0.00002237, 0.00001964, -0.00004335, 0.00002879, -0.00003002, 0.00002130, -0.00006488, 0.00002523, -0.00011492, 0.00002590, -0.00056067, 0.00050241, 0.00109682, -0.00056852, -0.00004826, -0.00010514, 0.00000336, -0.00005494, 0.00002410, -0.00003915, 0.00001007, -0.00002085, 0.00002235, -0.00002541, 0.00001108, -0.00002014, 0.00001875, -0.00001872, 0.00000489, -0.00001438, 0.00000887, -0.00001066, 0.00000603, -0.00000604, 0.00002098, -0.00002001, 0.00001176, -0.00001676, 0.00000623, -0.00000857, 0.00002652, -0.00002329, 0.00000677, -0.00000532, 0.00000895, -0.00000720, 0.00000097, -0.00000288, 0.00000358, -0.00000565, 0.00000695, -0.00000693, 0.00000579, -0.00000763, 0.00000456, -0.00000649, -0.00000222, 0.00000195, 0.00000078, -0.00000693, 0.00001537, -0.00002803, 0.00002996, -0.00002564, 0.00002624, -0.00000925, 0.00000173, -0.00000143, 0.00001218, -0.00000860, 0.00001020, -0.00000440, 0.00001191, 0.00000051, -0.00002062, 0.00003904, -0.00003349, 0.00002398, -0.00001536, -0.00000097, -0.00001052, 0.00000234, },
		{ 0.00000285, 0.00000750, -0.00001320, 0.00000344, -0.00000550, 0.00000333, -0.00001586, 0.00001515, -0.00001273, 0.00000342, 0.00000477, -0.00001181, 0.00002125, -0.00001423, 0.00001337, -0.00000775, -0.00000036, -0.00000456, 0.00000005, -0.00000150, 0.00000325, -0.00000528, 0.00000407, -0.00000555, 0.00000630, -0.00000383, -0.00000183, 0.00000062, 0.00000602, 0.00000110, -0.00001431, 0.00000177, 0.00001461, -0.00001326, -0.00000142, -0.00000438, 0.00000482, -0.00001203, 0.00001955, -0.00001391, 0.00000584, -0.00001695, 0.00001137, -0.00001408, 0.00000361, -0.00001557, 0.00000875, -0.00001235, 0.00000171, -0.00002741, 0.00002358, -0.00002629, 0.00000054, -0.00004611, -0.00001724, -0.00006551, -0.00018624, -0.00044945, 0.00148067, -0.00005157, -0.00040870, 0.00003581, -0.00010901, 0.00001605, -0.00004648, 0.00002228, -0.00003917, 0.00001235, -0.00001143, 0.00001055, -0.00001908, 0.00000465, -0.00000950, 0.00001094, -0.00001769, 0.00000302, -0.00000916, 0.00000520, -0.00000675, 0.00000098, 0.00001018, -0.00000064, -0.00000620, 0.00000009, -0.00000908, 0.00000060, 0.00001135, 0.00000286, -0.00001749, 0.00001133, -0.00000592, 0.00000819, -0.00001126, 0.00000595, -0.00000386, 0.00000131, -0.00000118, 0.00000223, -0.00000288, 0.00000163, -0.00000493, 0.00000288, -0.00000857, 0.00000196, 0.00000460, -0.00001083, 0.00001562, -0.00001817, 0.00000860, 0.00000246, -0.00000747, 0.00002270, -0.00002100, 0.00001336, -0.00000589, 0.00001480, -0.00001584, 0.00002219, -0.00001985, 0.00003515, -0.00004219, 0.00003432, -0.00000904, -0.00000484, 0.00001089, -0.00001446, -0.00000248, -0.00000575, },
		{ -0.00000190, 0.00000841, -0.00000371, -0.00000794, 0.00000338, -0.00000420, -0.00000421, -0.00000499, 0.00001288, -0.00002283, 0.00002743, -0.00002854, 0.00002735, -0.00001339, 0.00001102, -0.00000501, 0.00000186, -0.00000985, 0.00000639, -0.00001064, 0.00001288, -0.00001356, 0.00001174, -0.00001337, 0.00001290, -0.00000982, 0.00000729, -0.00001276, 0.00001782, -0.00000985, 0.00000685, -0.00002733, 0.00003683, -0.00002555, 0.00001375, -0.00002618, 0.00002851, -0.00003482, 0.00003613, -0.00002710, 0.00002515, -0.00003685, 0.00002852, -0.00003286, 0.00002919, -0.00004425, 0.00003756, -0.00004657, 0.00004733, -0.00007361, 0.00006253, -0.00006847, 0.00006537, -0.00011708, 0.00008161, -0.00018850, 0.00012295, -0.00074267, 0.00075650, 0.00108637, -0.00080216, 0.00012575, -0.00020184, 0.00007833, -0.00011488, 0.00007385, -0.00007744, 0.00004317, -0.00004967, 0.00004803, -0.00004883, 0.00003249, -0.00004079, 0.00003792, -0.00003596, 0.00002025, -0.00002906, 0.00002262, -0.00002243, 0.00001584, -0.00001238, 0.00002638, -0.00002695, 0.00001910, -0.00002626, 0.00001447, -0.00001094, 0.00002989, -0.00003451, 0.00001952, -0.00001436, 0.00001725, -0.00001705, 0.00000995, -0.00000869, 0.00000774, -0.00000965, 0.00001097, -0.00001066, 0.00000969, -0.00001202, 0.00000834, -0.00001067, 0.00000100, 0.00000279, -0.00000276, -0.00000017, 0.00000678, -0.00002058, 0.00002840, -0.00002817, 0.00003279, -0.00002007, 0.00000995, -0.00000907, 0.00001902, -0.00001584, 0.00001676, -0.00001076, 0.00001550, -0.00000286, -0.00001868, 0.00003723, -0.00003202, 0.00002164, -0.00001130, -0.00000504, -0.00000540, },
		{ -0.00000850, 0.00000628, 0.00000492, -0.00001145, 0.00000260, -0.00000393, 0.00000044, -0.00001392, 0.00001898, -0.00002144, 0.00001689, -0.00000948, 0.00000058, 0.00001279, -0.00001017, 0.00001197, -0.00000872, -0.00000052, -0.00000357, -0.00000132, 0.00000120, 0.00000062, -0.00000279, 0.00000118, -0.00000284, 0.00000531, -0.00000437, -0.00000259, 0.00000374, 0.00000415, 0.00000014, -0.00001836, 0.00001357, 0.00000511, -0.00000994, -0.00000622, 0.00000395, -0.00000432, -0.00000128, 0.00001084, -0.00000819, 0.00000072, -0.00001238, 0.00000588, -0.00000778, -0.00000228, -0.00000938, 0.00000085, -0.00000315, -0.00000743, -0.00001731, 0.00001229, -0.00001443, -0.00001204, -0.00003626, -0.00003358, -0.00004682, -0.00022005, -0.00049398, 0.00180712, -0.00042880, -0.00026277, -0.00000877, -0.00007449, -0.00000795, -0.00002903, 0.00001144, -0.00002838, 0.00000275, -0.00000254, 0.00000202, -0.00001106, -0.00000330, -0.00000278, 0.00000546, -0.00001300, -0.00000167, -0.00000526, 0.00000275, -0.00000514, 0.00000052, 0.00001268, -0.00000605, -0.00000025, -0.00000695, -0.00000465, -0.00000059, 0.00001767, -0.00000978, -0.00000815, 0.00000712, -0.00000155, 0.00000296, -0.00000832, 0.00000572, -0.00000351, -0.00000014, 0.00000078, 0.00000021, -0.00000052, -0.00000089, -0.00000334, 0.00000140, -0.00000890, 0.00000534, 0.00000127, -0.00000793, 0.00001552, -0.00002326, 0.00002033, -0.00001141, 0.00000727, 0.00000859, -0.00001264, 0.00000636, -0.00000014, 0.00000916, -0.00001145, 0.00001858, -0.00001849, 0.00003269, -0.00004006, 0.00003139, -0.00000580, -0.00000910, 0.00001569, -0.00001834, 0.00000172, },
		{ -0.00000230, -0.00000407, 0.00001158, -0.00000740, -0.00000389, 0.00000020, -0.00000099, -0.00001020, 0.00000597, 0.00000101, -0.00001195, 0.00002160, -0.00002842, 0.00003235, -0.00002151, 0.00001948, -0.00001313, 0.00000709, -0.00001412, 0.00001045, -0.00001408, 0.00001727, -0.00001851, 0.00001690, -0.00001949, 0.00002062, -0.00001763, 0.00001287, -0.00001711, 0.00002434, -0.00001716, 0.00000839, -0.00002640, 0.00004358, -0.00003762, 0.00002235, -0.00003297, 0.00003733, -0.00004508, 0.00005088, -0.00004489, 0.00004298, -0.00005514, 0.00004657, -0.00005188, 0.00005019, -0.00006617, 0.00006010, -0.00007138, 0.00007566, -0.00010465, 0.00009753, -0.00010928, 0.00011308, -0.00017503, 0.00014587, -0.00027248, 0.00023559, -0.00096381, 0.00111602, 0.00092614, -0.00093519, 0.00026636, -0.00028831, 0.00014797, -0.00017234, 0.00012330, -0.00011647, 0.00007710, -0.00007925, 0.00007402, -0.00007233, 0.00005388, -0.00006135, 0.00005725, -0.00005376, 0.00003639, -0.00004454, 0.00003751, -0.00003580, 0.00002761, -0.00002016, 0.00003162, -0.00003269, 0.00002482, -0.00003443, 0.00002258, -0.00001244, 0.00002940, -0.00004108, 0.00002968, -0.00002204, 0.00002404, -0.00002580, 0.00001910, -0.00001537, 0.00001246, -0.00001374, 0.00001483, -0.00001407, 0.00001306, -0.00001603, 0.00001206, -0.00001536, 0.00000600, 0.00000107, -0.00000398, 0.00000546, -0.00000323, -0.00000814, 0.00001898, -0.00002222, 0.00003097, -0.00002412, 0.00001355, -0.00001269, 0.00002197, -0.00001943, 0.00002043, -0.00001503, 0.00001729, -0.00000475, -0.00001748, 0.00003598, -0.00003114, 0.00002031, -0.00000907, -0.00000688, },
		{ 0.00000706, -0.00001253, 0.00001126, 0.00000006, -0.00000694, -0.00000080, -0.00000002, -0.00000554, -0.00000671, 0.00001596, -0.00002338, 0.00002535, -0.00002190, 0.00001423, 0.00000060, -0.00000123, 0.00000567, -0.00000568, -0.00000357, -0.00000008, -0.00000493, 0.00000615, -0.00000470, 0.00000268, -0.00000504, 0.00000382, -0.00000002, -0.00000118, -0.00000615, 0.00000979, -0.00000125, 0.00000149, -0.00002272, 0.00002706, -0.00000941, -0.00000087, -0.00001557, 0.00001678, -0.00001893, 0.00001670, -0.00000718, 0.00000770, -0.00001474, 0.00000278, -0.00001055, 0.00001007, -0.00002034, 0.00000956, -0.00002065, 0.00002104, -0.00003333, 0.00001182, -0.00002090, 0.00002252, -0.00005383, 0.00000736, -0.00008999, 0.00002298, -0.00032867, -0.00043299, 0.00205876, -0.00082839, -0.00004313, -0.00011475, 0.00000490, -0.00006821, 0.00001899, -0.00002509, 0.00000394, -0.00002529, 0.00002225, -0.00002075, 0.00000994, -0.00002323, 0.00001527, -0.00001075, 0.00000159, -0.00001543, 0.00000720, -0.00000765, 0.00000347, -0.00000530, 0.00001872, -0.00001455, 0.00000898, -0.00001779, 0.00000402, -0.00000430, 0.00002444, -0.00002425, 0.00000578, -0.00000204, 0.00000705, -0.00000680, -0.00000026, 0.00000091, 0.00000024, -0.00000423, 0.00000506, -0.00000392, 0.00000381, -0.00000551, 0.00000052, -0.00000249, -0.00000658, 0.00000699, -0.00000222, -0.00000258, 0.00001041, -0.00002148, 0.00002536, -0.00002092, 0.00001993, -0.00000615, -0.00000197, -0.00000274, 0.00000754, 0.00000134, -0.00000425, 0.00001189, -0.00001428, 0.00002742, -0.00003502, 0.00002617, -0.00000090, -0.00001429, 0.00002086, -0.00002280, },
		{ -0.00000572, -0.00000219, -0.00000303, 0.00001116, -0.00000771, -0.00000260, -0.00000040, -0.00000104, -0.00001200, 0.00001367, -0.00001020, 0.00000156, 0.00001008, -0.00002118, 0.00003024, -0.00002390, 0.00002403, -0.00001914, 0.00001129, -0.00001734, 0.00001337, -0.00001607, 0.00001975, -0.00002143, 0.00001990, -0.00002330, 0.00002636, -0.00002448, 0.00001802, -0.00002043, 0.00002940, -0.00002408, 0.00000975, -0.00002221, 0.00004449, -0.00004484, 0.00002753, -0.00003563, 0.00004063, -0.00004846, 0.00005801, -0.00005583, 0.00005479, -0.00006773, 0.00005898, -0.00006508, 0.00006524, -0.00008195, 0.00007602, -0.00008872, 0.00009564, -0.00012641, 0.00012237, -0.00013937, 0.00014941, -0.00022148, 0.00019819, -0.00034306, 0.00033307, -0.00118529, 0.00154833, 0.00061791, -0.00093826, 0.00033717, -0.00033508, 0.00018914, -0.00020815, 0.00015633, -0.00014273, 0.00010043, -0.00009974, 0.00009169, -0.00008815, 0.00006814, -0.00007518, 0.00007048, -0.00006630, 0.00004795, -0.00005589, 0.00004894, -0.00004664, 0.00003799, -0.00002701, 0.00003493, -0.00003556, 0.00002706, -0.00003920, 0.00002887, -0.00001236, 0.00002444, -0.00004125, 0.00003468, -0.00002580, 0.00002679, -0.00003071, 0.00002570, -0.00002065, 0.00001600, -0.00001653, 0.00001731, -0.00001604, 0.00001480, -0.00001847, 0.00001445, -0.00001923, 0.00001173, -0.00000303, -0.00000180, 0.00000749, -0.00001091, 0.00000501, 0.00000521, -0.00000981, 0.00002133, -0.00002074, 0.00001134, -0.00001112, 0.00001977, -0.00001785, 0.00001941, -0.00001540, 0.00001557, -0.00000353, -0.00001847, 0.00003645, -0.00003172, 0.00002063, -0.00000927, },
		{ -0.00002596, 0.00001147, -0.00001567, 0.00001548, -0.00000478, -0.00000179, -0.00000485, 0.00000420, -0.00001223, 0.00000350, 0.00000730, -0.00001801, 0.00002639, -0.00002880, 0.00002509, -0.00001169, 0.00000985, -0.00000390, 0.00000060, -0.00000909, 0.00000552, -0.00001032, 0.00001274, -0.00001192, 0.00001031, -0.00001370, 0.00001365, -0.00000932, 0.00000582, -0.00001261, 0.00001860, -0.00001018, 0.00000540, -0.00002687, 0.00004020, -0.00002649, 0.00001208, -0.00002782, 0.00003193, -0.00003610, 0.00003846, -0.00003115, 0.00003086, -0.00003806, 0.00002626, -0.00003570, 0.00003726, -0.00004839, 0.00003902, -0.00005328, 0.00005745, -0.00007291, 0.00005669, -0.00007264, 0.00008128, -0.00012198, 0.00008267, -0.00018523, 0.00014422, -0.00051928, -0.00023766, 0.00217719, -0.00118403, 0.00020311, -0.00025560, 0.00011274, -0.00015307, 0.00008911, -0.00008110, 0.00005300, -0.00006787, 0.00005956, -0.00005466, 0.00004103, -0.00005242, 0.00004222, -0.00003560, 0.00002435, -0.00003681, 0.00002704, -0.00002524, 0.00001886, -0.00001659, 0.00002822, -0.00002547, 0.00002034, -0.00003111, 0.00001600, -0.00000996, 0.00003031, -0.00003791, 0.00002143, -0.00001390, 0.00001791, -0.00001918, 0.00001143, -0.00000767, 0.00000694, -0.00001055, 0.00001121, -0.00000966, 0.00000950, -0.00001165, 0.00000614, -0.00000856, -0.00000124, 0.00000564, -0.00000388, 0.00000298, 0.00000219, -0.00001341, 0.00002223, -0.00002279, 0.00002626, -0.00001699, 0.00000697, -0.00001046, 0.00001385, -0.00000544, 0.00000267, 0.00000484, -0.00000959, 0.00002147, -0.00002891, 0.00002020, 0.00000427, -0.00001900, 0.00002481, },
		{ -0.00001133, -0.00000212, -0.00000466, 0.00000089, 0.00000722, -0.00000435, -0.00000459, 0.00000213, -0.00000500, -0.00000817, 0.00001524, -0.00001668, 0.00001314, -0.00000318, -0.00000936, 0.00002197, -0.00002003, 0.00002316, -0.00002118, 0.00001276, -0.00001814, 0.00001397, -0.00001554, 0.00001924, -0.00002106, 0.00001932, -0.00002307, 0.00002798, -0.00002806, 0.00002076, -0.00002101, 0.00003121, -0.00002897, 0.00001023, -0.00001482, 0.00003885, -0.00004539, 0.00002732, -0.00003254, 0.00003688, -0.00004296, 0.00005455, -0.00005591, 0.00005599, -0.00006967, 0.00006055, -0.00006692, 0.00006839, -0.00008530, 0.00007865, -0.00009137, 0.00009929, -0.00012984, 0.00012655, -0.00014645, 0.00015975, -0.00023899, 0.00021763, -0.00037396, 0.00037913, -0.00135530, 0.00199621, 0.00019541, -0.00081040, 0.00032047, -0.00032385, 0.00018594, -0.00020814, 0.00016032, -0.00014521, 0.00010360, -0.00010282, 0.00009372, -0.00008973, 0.00006928, -0.00007665, 0.00007222, -0.00006843, 0.00005009, -0.00005850, 0.00005249, -0.00005077, 0.00004333, -0.00003024, 0.00003451, -0.00003414, 0.00002441, -0.00003889, 0.00003191, -0.00001037, 0.00001538, -0.00003470, 0.00003327, -0.00002422, 0.00002405, -0.00002990, 0.00002755, -0.00002253, 0.00001680, -0.00001675, 0.00001735, -0.00001563, 0.00001399, -0.00001832, 0.00001434, -0.00002077, 0.00001650, -0.00000825, 0.00000352, 0.00000468, -0.00001324, 0.00001435, -0.00000797, 0.00000473, 0.00000712, -0.00001190, 0.00000453, -0.00000538, 0.00001324, -0.00001160, 0.00001382, -0.00001171, 0.00001012, 0.00000112, -0.00002200, 0.00003882, -0.00003367, 0.00002225, },
		{ 0.00002622, -0.00002625, 0.00001313, -0.00001596, 0.00001671, -0.00000711, 0.00000152, -0.00000717, 0.00000621, -0.00001643, 0.00001292, -0.00000357, -0.00000751, 0.00002023, -0.00002853, 0.00003025, -0.00002062, 0.00001983, -0.00001417, 0.00000832, -0.00001577, 0.00001207, -0.00001634, 0.00001965, -0.00001954, 0.00001843, -0.00002299, 0.00002474, -0.00002095, 0.00001546, -0.00002099, 0.00002905, -0.00002170, 0.00001151, -0.00003009, 0.00005080, -0.00004299, 0.00002604, -0.00004030, 0.00004631, -0.00005187, 0.00005909, -0.00005561, 0.00005587, -0.00006392, 0.00005275, -0.00006413, 0.00006815, -0.00008059, 0.00007286, -0.00009038, 0.00009871, -0.00011800, 0.00010814, -0.00013273, 0.00015058, -0.00020419, 0.00017637, -0.00030449, 0.00030000, -0.00077656, 0.00009508, 0.00212398, -0.00143544, 0.00042299, -0.00039474, 0.00022279, -0.00024204, 0.00016489, -0.00014334, 0.00010784, -0.00011567, 0.00010129, -0.00009244, 0.00007555, -0.00008468, 0.00007228, -0.00006380, 0.00005052, -0.00006148, 0.00005039, -0.00004671, 0.00003849, -0.00003164, 0.00003978, -0.00003721, 0.00003188, -0.00004476, 0.00002955, -0.00001659, 0.00003395, -0.00004812, 0.00003554, -0.00002546, 0.00002824, -0.00003128, 0.00002406, -0.00001795, 0.00001511, -0.00001788, 0.00001809, -0.00001590, 0.00001544, -0.00001811, 0.00001238, -0.00001577, 0.00000656, 0.00000083, -0.00000206, 0.00000602, -0.00000587, -0.00000202, 0.00001247, -0.00001670, 0.00002443, -0.00002110, 0.00001124, -0.00001412, 0.00001617, -0.00000840, 0.00000646, 0.00000012, -0.00000685, 0.00001709, -0.00002374, 0.00001514, 0.00000830, -0.00002195, },
		{ 0.00002388, -0.00001363, 0.00000218, -0.00000794, 0.00000589, 0.00000148, 0.00000130, -0.00000874, 0.00000662, -0.00001159, 0.00000050, 0.00001012, -0.00001614, 0.00001903, -0.00001389, 0.00000319, 0.00001027, -0.00001130, 0.00001703, -0.00001864, 0.00001071, -0.00001578, 0.00001164, -0.00001207, 0.00001535, -0.00001698, 0.00001462, -0.00001800, 0.00002421, -0.00002671, 0.00001950, -0.00001758, 0.00002837, -0.00003031, 0.00000930, -0.00000491, 0.00002728, -0.00003898, 0.00002119, -0.00002359, 0.00002638, -0.00002894, 0.00004029, -0.00004406, 0.00004484, -0.00005885, 0.00004888, -0.00005472, 0.00005666, -0.00007297, 0.00006455, -0.00007566, 0.00008250, -0.00011022, 0.00010441, -0.00012345, 0.00013508, -0.00021593, 0.00018929, -0.00034513, 0.00034257, -0.00142117, 0.00238883, -0.00028145, -0.00057953, 0.00022180, -0.00025310, 0.00013510, -0.00016813, 0.00013060, -0.00011936, 0.00008239, -0.00008472, 0.00007681, -0.00007414, 0.00005463, -0.00006323, 0.00005991, -0.00005758, 0.00004027, -0.00004981, 0.00004549, -0.00004536, 0.00004091, -0.00002779, 0.00002917, -0.00002780, 0.00001637, -0.00003272, 0.00003085, -0.00000670, 0.00000360, -0.00002265, 0.00002589, -0.00001746, 0.00001589, -0.00002304, 0.00002377, -0.00001995, 0.00001397, -0.00001372, 0.00001441, -0.00001248, 0.00001030, -0.00001513, 0.00001100, -0.00001876, 0.00001848, -0.00001251, 0.00001024, -0.00000225, -0.00000921, 0.00001679, -0.00001602, 0.00001647, -0.00000706, -0.00000127, -0.00000395, 0.00000185, 0.00000496, -0.00000300, 0.00000555, -0.00000552, 0.00000226, 0.00000811, -0.00002720, 0.00004220, -0.00003595, },
		{ -0.00002265, 0.00002473, -0.00002317, 0.00001136, -0.00001255, 0.00001382, -0.00000542, 0.00000130, -0.00000608, 0.00000414, -0.00001544, 0.00001775, -0.00001222, 0.00000403, 0.00000959, -0.00002192, 0.00002882, -0.00002408, 0.00002599, -0.00002246, 0.00001530, -0.00002183, 0.00001799, -0.00002150, 0.00002530, -0.00002578, 0.00002505, -0.00003061, 0.00003448, -0.00003230, 0.00002558, -0.00002942, 0.00003917, -0.00003400, 0.00001886, -0.00003166, 0.00005696, -0.00005584, 0.00003785, -0.00005013, 0.00005679, -0.00006239, 0.00007360, -0.00007463, 0.00007635, -0.00008573, 0.00007543, -0.00008871, 0.00009522, -0.00010902, 0.00010271, -0.00012290, 0.00013484, -0.00015744, 0.00015352, -0.00018671, 0.00021385, -0.00028144, 0.00026667, -0.00042199, 0.00045811, -0.00106232, 0.00053732, 0.00189204, -0.00154318, 0.00057164, -0.00049648, 0.00030723, -0.00031216, 0.00022700, -0.00019556, 0.00015465, -0.00015679, 0.00013698, -0.00012471, 0.00010495, -0.00011204, 0.00009796, -0.00008837, 0.00007363, -0.00008345, 0.00007165, -0.00006696, 0.00005806, -0.00004722, 0.00005095, -0.00004749, 0.00004120, -0.00005611, 0.00004243, -0.00002297, 0.00003446, -0.00005295, 0.00004528, -0.00003391, 0.00003527, -0.00004010, 0.00003458, -0.00002731, 0.00002266, -0.00002446, 0.00002414, -0.00002124, 0.00002023, -0.00002341, 0.00001767, -0.00002244, 0.00001532, -0.00000674, 0.00000378, 0.00000452, -0.00001029, 0.00000838, -0.00000002, -0.00000526, 0.00001555, -0.00001827, 0.00001005, -0.00001294, 0.00001362, -0.00000640, 0.00000566, -0.00000071, -0.00000757, 0.00001578, -0.00002091, 0.00001208, 0.00001044, },
		{ -0.00003712, 0.00002388, -0.00001421, 0.00000498, -0.00000971, 0.00000944, -0.00000341, 0.00000668, -0.00001268, 0.00001062, -0.00001791, 0.00001104, 0.00000036, -0.00000895, 0.00001764, -0.00001891, 0.00001276, -0.00000135, -0.00000040, 0.00000732, -0.00001225, 0.00000543, -0.00001048, 0.00000662, -0.00000602, 0.00000861, -0.00000974, 0.00000641, -0.00000860, 0.00001513, -0.00002000, 0.00001370, -0.00000974, 0.00002032, -0.00002722, 0.00000673, 0.00000624, 0.00001163, -0.00002691, 0.00001016, -0.00001024, 0.00001126, -0.00000916, 0.00001799, -0.00002257, 0.00002320, -0.00003684, 0.00002541, -0.00002985, 0.00003127, -0.00004615, 0.00003505, -0.00004308, 0.00004690, -0.00006925, 0.00005753, -0.00007161, 0.00007581, -0.00015136, 0.00011041, -0.00025067, 0.00020828, -0.00134386, 0.00265771, -0.00073972, -0.00029518, 0.00006761, -0.00013836, 0.00004721, -0.00009555, 0.00007235, -0.00006892, 0.00003971, -0.00004783, 0.00004311, -0.00004335, 0.00002606, -0.00003660, 0.00003495, -0.00003482, 0.00001932, -0.00003039, 0.00002809, -0.00003007, 0.00002995, -0.00001901, 0.00001885, -0.00001702, 0.00000370, -0.00002105, 0.00002563, -0.00000202, -0.00000884, -0.00000751, 0.00001442, -0.00000714, 0.00000397, -0.00001142, 0.00001500, -0.00001316, 0.00000768, -0.00000765, 0.00000877, -0.00000693, 0.00000415, -0.00000923, 0.00000459, -0.00001276, 0.00001627, -0.00001355, 0.00001564, -0.00001077, -0.00000027, 0.00001185, -0.00001653, 0.00002171, -0.00001703, 0.00000733, -0.00001084, 0.00000746, -0.00000190, 0.00000486, -0.00000260, 0.00000071, -0.00000573, 0.00001544, -0.00003242, 0.00004513, },
		{ 0.00001096, -0.00002170, 0.00002124, -0.00001764, 0.00000701, -0.00000626, 0.00000741, 0.00000005, -0.00000251, -0.00000150, -0.00000215, -0.00000840, 0.00001571, -0.00001499, 0.00001218, -0.00000141, -0.00001183, 0.00002203, -0.00002177, 0.00002694, -0.00002685, 0.00001968, -0.00002563, 0.00002181, -0.00002444, 0.00002829, -0.00002906, 0.00002840, -0.00003445, 0.00004028, -0.00004053, 0.00003357, -0.00003557, 0.00004654, -0.00004476, 0.00002606, -0.00003103, 0.00005752, -0.00006273, 0.00004496, -0.00005502, 0.00006118, -0.00006509, 0.00007847, -0.00008360, 0.00008700, -0.00009774, 0.00008818, -0.00010286, 0.00011139, -0.00012614, 0.00012058, -0.00014222, 0.00015631, -0.00018051, 0.00018050, -0.00022022, 0.00025407, -0.00033348, 0.00032932, -0.00050766, 0.00057834, -0.00132214, 0.00103307, 0.00150926, -0.00149799, 0.00062362, -0.00053637, 0.00034527, -0.00034525, 0.00025934, -0.00022366, 0.00018114, -0.00018045, 0.00015714, -0.00014293, 0.00012144, -0.00012725, 0.00011234, -0.00010272, 0.00008749, -0.00009682, 0.00008516, -0.00008060, 0.00007270, -0.00005947, 0.00005887, -0.00005399, 0.00004603, -0.00006267, 0.00005236, -0.00002804, 0.00003165, -0.00005164, 0.00004900, -0.00003741, 0.00003718, -0.00004344, 0.00004041, -0.00003330, 0.00002753, -0.00002856, 0.00002786, -0.00002436, 0.00002258, -0.00002618, 0.00002044, -0.00002664, 0.00002287, -0.00001516, 0.00001263, -0.00000195, -0.00000881, 0.00001382, -0.00001045, 0.00000707, 0.00000314, -0.00001076, 0.00000485, -0.00000819, 0.00000738, -0.00000031, 0.00000073, 0.00000217, -0.00001168, 0.00001759, -0.00002057, 0.00001102, },
		{ 0.00004628, -0.00003602, 0.00002109, -0.00001164, 0.00000462, -0.00000809, 0.00000936, -0.00000493, 0.00000921, -0.00001388, 0.00001141, -0.00002055, 0.00001933, -0.00001002, 0.00000187, 0.00001009, -0.00001738, 0.00001702, -0.00000985, 0.00000959, -0.00000345, -0.00000379, -0.00000181, -0.00000331, -0.00000007, 0.00000145, 0.00000035, -0.00000085, -0.00000362, 0.00000329, 0.00000234, -0.00000899, 0.00000415, 0.00000161, 0.00000773, -0.00001973, 0.00000275, 0.00001698, -0.00000544, -0.00001167, -0.00000351, 0.00000489, -0.00000507, 0.00001204, -0.00000736, 0.00000356, -0.00000404, -0.00000854, -0.00000485, 0.00000254, -0.00000244, -0.00001037, -0.00000390, -0.00000021, -0.00000020, -0.00001513, -0.00000496, -0.00000093, -0.00000725, -0.00005659, -0.00000701, -0.00010316, -0.00001535, -0.00111031, 0.00275336, -0.00111031, -0.00001535, -0.00010316, -0.00000701, -0.00005659, -0.00000725, -0.00000093, -0.00000496, -0.00001513, -0.00000020, -0.00000021, -0.00000390, -0.00001037, -0.00000244, 0.00000254, -0.00000485, -0.00000854, -0.00000404, 0.00000356, -0.00000736, 0.00001204, -0.00000507, 0.00000489, -0.00000351, -0.00001167, -0.00000544, 0.00001698, 0.00000275, -0.00001973, 0.00000773, 0.00000161, 0.00000415, -0.00000899, 0.00000234, 0.00000329, -0.00000362, -0.00000085, 0.00000035, 0.00000145, -0.00000007, -0.00000331, -0.00000181, -0.00000379, -0.00000345, 0.00000959, -0.00000985, 0.00001702, -0.00001738, 0.00001009, 0.00000187, -0.00001002, 0.00001933, -0.00002055, 0.00001141, -0.00001388, 0.00000921, -0.00000493, 0.00000936, -0.00000809, 0.00000462, -0.00001164, 0.00002109, -0.00003602, },
		{ 0.00001096, 0.00001102, -0.00002057, 0.00001759, -0.00001168, 0.00000217, 0.00000073, -0.00000031, 0.00000738, -0.00000819, 0.00000485, -0.00001076, 0.00000314, 0.00000707, -0.00001045, 0.00001382, -0.00000881, -0.00000195, 0.00001263, -0.00001516, 0.00002287, -0.00002664, 0.00002044, -0.00002618, 0.00002258, -0.00002436, 0.00002786, -0.00002856, 0.00002753, -0.00003330, 0.00004041, -0.00004344, 0.00003718, -0.00003741, 0.00004900, -0.00005164, 0.00003165, -0.00002804, 0.00005236, -0.00006267, 0.00004603, -0.00005399, 0.00005887, -0.00005947, 0.00007270, -0.00008060, 0.00008516, -0.00009682, 0.00008749, -0.00010272, 0.00011234, -0.00012725, 0.00012144, -0.00014293, 0.00015714, -0.00018045, 0.00018114, -0.00022366, 0.00025934, -0.00034525, 0.00034527, -0.00053637, 0.00062362, -0.00149799, 0.00150926, 0.00103307, -0.00132214, 0.00057834, -0.00050766, 0.00032932, -0.00033348, 0.00025407, -0.00022022, 0.00018050, -0.00018051, 0.00015631, -0.00014222, 0.00012058, -0.00012614, 0.00011139, -0.00010286, 0.00008818, -0.00009774, 0.00008700, -0.00008360, 0.00007847, -0.00006509, 0.00006118, -0.00005502, 0.00004496, -0.00006273, 0.00005752, -0.00003103, 0.00002606, -0.00004476, 0.00004654, -0.00003557, 0.00003357, -0.00004053, 0.00004028, -0.00003445, 0.00002840, -0.00002906, 0.00002829, -0.00002444, 0.00002181, -0.00002563, 0.00001968, -0.00002685, 0.00002694, -0.00002177, 0.00002203, -0.00001183, -0.00000141, 0.00001218, -0.00001499, 0.00001571, -0.00000840, -0.00000215, -0.00000150, -0.00000251, 0.00000005, 0.00000741, -0.00000626, 0.00000701, -0.00001764, 0.00002124, -0.00002170, },
		{ -0.00003712, 0.00004513, -0.00003242, 0.00001544, -0.00000573, 0.00000071, -0.00000260, 0.00000486, -0.00000190, 0.00000746, -0.00001084, 0.00000733, -0.00001703, 0.00002171, -0.00001653, 0.00001185, -0.00000027, -0.00001077, 0.00001564, -0.00001355, 0.00001627, -0.00001276, 0.00000459, -0.00000923, 0.00000415, -0.00000693, 0.00000877, -0.00000765, 0.00000768, -0.00001316, 0.00001500, -0.00001142, 0.00000397, -0.00000714, 0.00001442, -0.00000751, -0.00000884, -0.00000202, 0.00002563, -0.00002105, 0.00000370, -0.00001702, 0.00001885, -0.00001901, 0.00002995, -0.00003007, 0.00002809, -0.00003039, 0.00001932, -0.00003482, 0.00003495, -0.00003660, 0.00002606, -0.00004335, 0.00004311, -0.00004783, 0.00003971, -0.00006892, 0.00007235, -0.00009555, 0.00004721, -0.00013836, 0.00006761, -0.00029518, -0.00073972, 0.00265771, -0.00134386, 0.00020828, -0.00025067, 0.00011041, -0.00015136, 0.00007581, -0.00007161, 0.00005753, -0.00006925, 0.00004690, -0.00004308, 0.00003505, -0.00004615, 0.00003127, -0.00002985, 0.00002541, -0.00003684, 0.00002320, -0.00002257, 0.00001799, -0.00000916, 0.00001126, -0.00001024, 0.00001016, -0.00002691, 0.00001163, 0.00000624, 0.00000673, -0.00002722, 0.00002032, -0.00000974, 0.00001370, -0.00002000, 0.00001513, -0.00000860, 0.00000641, -0.00000974, 0.00000861, -0.00000602, 0.00000662, -0.00001048, 0.00000543, -0.00001225, 0.00000732, -0.00000040, -0.00000135, 0.00001276, -0.00001891, 0.00001764, -0.00000895, 0.00000036, 0.00001104, -0.00001791, 0.00001062, -0.00001268, 0.00000668, -0.00000341, 0.00000944, -0.00000971, 0.00000498, -0.00001421, 0.00002388, },
		{ -0.00002265, 0.00001044, 0.00001208, -0.00002091, 0.00001578, -0.00000757, -0.00000071, 0.00000566, -0.00000640, 0.00001362, -0.00001294, 0.00001005, -0.00001827, 0.00001555, -0.00000526, -0.00000002, 0.00000838, -0.00001029, 0.00000452, 0.00000378, -0.00000674, 0.00001532, -0.00002244, 0.00001767, -0.00002341, 0.00002023, -0.00002124, 0.00002414, -0.00002446, 0.00002266, -0.00002731, 0.00003458, -0.00004010, 0.00003527, -0.00003391, 0.00004528, -0.00005295, 0.00003446, -0.00002297, 0.00004243, -0.00005611, 0.00004120, -0.00004749, 0.00005095, -0.00004722, 0.00005806, -0.00006696, 0.00007165, -0.00008345, 0.00007363, -0.00008837, 0.00009796, -0.00011204, 0.00010495, -0.00012471, 0.00013698, -0.00015679, 0.00015465, -0.00019556, 0.00022700, -0.00031216, 0.00030723, -0.00049648, 0.00057164, -0.00154318, 0.00189204, 0.00053732, -0.00106232, 0.00045811, -0.00042199, 0.00026667, -0.00028144, 0.00021385, -0.00018671, 0.00015352, -0.00015744, 0.00013484, -0.00012290, 0.00010271, -0.00010902, 0.00009522, -0.00008871, 0.00007543, -0.00008573, 0.00007635, -0.00007463, 0.00007360, -0.00006239, 0.00005679, -0.00005013, 0.00003785, -0.00005584, 0.00005696, -0.00003166, 0.00001886, -0.00003400, 0.00003917, -0.00002942, 0.00002558, -0.00003230, 0.00003448, -0.00003061, 0.00002505, -0.00002578, 0.00002530, -0.00002150, 0.00001799, -0.00002183, 0.00001530, -0.00002246, 0.00002599, -0.00002408, 0.00002882, -0.00002192, 0.00000959, 0.00000403, -0.00001222, 0.00001775, -0.00001544, 0.00000414, -0.00000608, 0.00000130, -0.00000542, 0.00001382, -0.00001255, 0.00001136, -0.00002317, 0.00002473, },
		{ 0.00002388, -0.00003595, 0.00004220, -0.00002720, 0.00000811, 0.00000226, -0.00000552, 0.00000555, -0.00000300, 0.00000496, 0.00000185, -0.00000395, -0.00000127, -0.00000706, 0.00001647, -0.00001602, 0.00001679, -0.00000921, -0.00000225, 0.00001024, -0.00001251, 0.00001848, -0.00001876, 0.00001100, -0.00001513, 0.00001030, -0.00001248, 0.00001441, -0.00001372, 0.00001397, -0.00001995, 0.00002377, -0.00002304, 0.00001589, -0.00001746, 0.00002589, -0.00002265, 0.00000360, -0.00000670, 0.00003085, -0.00003272, 0.00001637, -0.00002780, 0.00002917, -0.00002779, 0.00004091, -0.00004536, 0.00004549, -0.00004981, 0.00004027, -0.00005758, 0.00005991, -0.00006323, 0.00005463, -0.00007414, 0.00007681, -0.00008472, 0.00008239, -0.00011936, 0.00013060, -0.00016813, 0.00013510, -0.00025310, 0.00022180, -0.00057953, -0.00028145, 0.00238883, -0.00142117, 0.00034257, -0.00034513, 0.00018929, -0.00021593, 0.00013508, -0.00012345, 0.00010441, -0.00011022, 0.00008250, -0.00007566, 0.00006455, -0.00007297, 0.00005666, -0.00005472, 0.00004888, -0.00005885, 0.00004484, -0.00004406, 0.00004029, -0.00002894, 0.00002638, -0.00002359, 0.00002119, -0.00003898, 0.00002728, -0.00000491, 0.00000930, -0.00003031, 0.00002837, -0.00001758, 0.00001950, -0.00002671, 0.00002421, -0.00001800, 0.00001462, -0.00001698, 0.00001535, -0.00001207, 0.00001164, -0.00001578, 0.00001071, -0.00001864, 0.00001703, -0.00001130, 0.00001027, 0.00000319, -0.00001389, 0.00001903, -0.00001614, 0.00001012, 0.00000050, -0.00001159, 0.00000662, -0.00000874, 0.00000130, 0.00000148, 0.00000589, -0.00000794, 0.00000218, -0.00001363, },
		{ 0.00002622, -0.00002195, 0.00000830, 0.00001514, -0.00002374, 0.00001709, -0.00000685, 0.00000012, 0.00000646, -0.00000840, 0.00001617, -0.00001412, 0.00001124, -0.00002110, 0.00002443, -0.00001670, 0.00001247, -0.00000202, -0.00000587, 0.00000602, -0.00000206, 0.00000083, 0.00000656, -0.00001577, 0.00001238, -0.00001811, 0.00001544, -0.00001590, 0.00001809, -0.00001788, 0.00001511, -0.00001795, 0.00002406, -0.00003128, 0.00002824, -0.00002546, 0.00003554, -0.00004812, 0.00003395, -0.00001659, 0.00002955, -0.00004476, 0.00003188, -0.00003721, 0.00003978, -0.00003164, 0.00003849, -0.00004671, 0.00005039, -0.00006148, 0.00005052, -0.00006380, 0.00007228, -0.00008468, 0.00007555, -0.00009244, 0.00010129, -0.00011567, 0.00010784, -0.00014334, 0.00016489, -0.00024204, 0.00022279, -0.00039474, 0.00042299, -0.00143544, 0.00212398, 0.00009508, -0.00077656, 0.00030000, -0.00030449, 0.00017637, -0.00020419, 0.00015058, -0.00013273, 0.00010814, -0.00011800, 0.00009871, -0.00009038, 0.00007286, -0.00008059, 0.00006815, -0.00006413, 0.00005275, -0.00006392, 0.00005587, -0.00005561, 0.00005909, -0.00005187, 0.00004631, -0.00004030, 0.00002604, -0.00004299, 0.00005080, -0.00003009, 0.00001151, -0.00002170, 0.00002905, -0.00002099, 0.00001546, -0.00002095, 0.00002474, -0.00002299, 0.00001843, -0.00001954, 0.00001965, -0.00001634, 0.00001207, -0.00001577, 0.00000832, -0.00001417, 0.00001983, -0.00002062, 0.00003025, -0.00002853, 0.00002023, -0.00000751, -0.00000357, 0.00001292, -0.00001643, 0.00000621, -0.00000717, 0.00000152, -0.00000711, 0.00001671, -0.00001596, 0.00001313, -0.00002625, },
		{ -0.00001133, 0.00002225, -0.00003367, 0.00003882, -0.00002200, 0.00000112, 0.00001012, -0.00001171, 0.00001382, -0.00001160, 0.00001324, -0.00000538, 0.00000453, -0.00001190, 0.00000712, 0.00000473, -0.00000797, 0.00001435, -0.00001324, 0.00000468, 0.00000352, -0.00000825, 0.00001650, -0.00002077, 0.00001434, -0.00001832, 0.00001399, -0.00001563, 0.00001735, -0.00001675, 0.00001680, -0.00002253, 0.00002755, -0.00002990, 0.00002405, -0.00002422, 0.00003327, -0.00003470, 0.00001538, -0.00001037, 0.00003191, -0.00003889, 0.00002441, -0.00003414, 0.00003451, -0.00003024, 0.00004333, -0.00005077, 0.00005249, -0.00005850, 0.00005009, -0.00006843, 0.00007222, -0.00007665, 0.00006928, -0.00008973, 0.00009372, -0.00010282, 0.00010360, -0.00014521, 0.00016032, -0.00020814, 0.00018594, -0.00032385, 0.00032047, -0.00081040, 0.00019541, 0.00199621, -0.00135530, 0.00037913, -0.00037396, 0.00021763, -0.00023899, 0.00015975, -0.00014645, 0.00012655, -0.00012984, 0.00009929, -0.00009137, 0.00007865, -0.00008530, 0.00006839, -0.00006692, 0.00006055, -0.00006967, 0.00005599, -0.00005591, 0.00005455, -0.00004296, 0.00003688, -0.00003254, 0.00002732, -0.00004539, 0.00003885, -0.00001482, 0.00001023, -0.00002897, 0.00003121, -0.00002101, 0.00002076, -0.00002806, 0.00002798, -0.00002307, 0.00001932, -0.00002106, 0.00001924, -0.00001554, 0.00001397, -0.00001814, 0.00001276, -0.00002118, 0.00002316, -0.00002003, 0.00002197, -0.00000936, -0.00000318, 0.00001314, -0.00001668, 0.00001524, -0.00000817, -0.00000500, 0.00000213, -0.00000459, -0.00000435, 0.00000722, 0.00000089, -0.00000466, -0.00000212, },
		{ -0.00002596, 0.00002481, -0.00001900, 0.00000427, 0.00002020, -0.00002891, 0.00002147, -0.00000959, 0.00000484, 0.00000267, -0.00000544, 0.00001385, -0.00001046, 0.00000697, -0.00001699, 0.00002626, -0.00002279, 0.00002223, -0.00001341, 0.00000219, 0.00000298, -0.00000388, 0.00000564, -0.00000124, -0.00000856, 0.00000614, -0.00001165, 0.00000950, -0.00000966, 0.00001121, -0.00001055, 0.00000694, -0.00000767, 0.00001143, -0.00001918, 0.00001791, -0.00001390, 0.00002143, -0.00003791, 0.00003031, -0.00000996, 0.00001600, -0.00003111, 0.00002034, -0.00002547, 0.00002822, -0.00001659, 0.00001886, -0.00002524, 0.00002704, -0.00003681, 0.00002435, -0.00003560, 0.00004222, -0.00005242, 0.00004103, -0.00005466, 0.00005956, -0.00006787, 0.00005300, -0.00008110, 0.00008911, -0.00015307, 0.00011274, -0.00025560, 0.00020311, -0.00118403, 0.00217719, -0.00023766, -0.00051928, 0.00014422, -0.00018523, 0.00008267, -0.00012198, 0.00008128, -0.00007264, 0.00005669, -0.00007291, 0.00005745, -0.00005328, 0.00003902, -0.00004839, 0.00003726, -0.00003570, 0.00002626, -0.00003806, 0.00003086, -0.00003115, 0.00003846, -0.00003610, 0.00003193, -0.00002782, 0.00001208, -0.00002649, 0.00004020, -0.00002687, 0.00000540, -0.00001018, 0.00001860, -0.00001261, 0.00000582, -0.00000932, 0.00001365, -0.00001370, 0.00001031, -0.00001192, 0.00001274, -0.00001032, 0.00000552, -0.00000909, 0.00000060, -0.00000390, 0.00000985, -0.00001169, 0.00002509, -0.00002880, 0.00002639, -0.00001801, 0.00000730, 0.00000350, -0.00001223, 0.00000420, -0.00000485, -0.00000179, -0.00000478, 0.00001548, -0.00001567, 0.00001147, },
		{ -0.00000572, -0.00000927, 0.00002063, -0.00003172, 0.00003645, -0.00001847, -0.00000353, 0.00001557, -0.00001540, 0.00001941, -0.00001785, 0.00001977, -0.00001112, 0.00001134, -0.00002074, 0.00002133, -0.00000981, 0.00000521, 0.00000501, -0.00001091, 0.00000749, -0.00000180, -0.00000303, 0.00001173, -0.00001923, 0.00001445, -0.00001847, 0.00001480, -0.00001604, 0.00001731, -0.00001653, 0.00001600, -0.00002065, 0.00002570, -0.00003071, 0.00002679, -0.00002580, 0.00003468, -0.00004125, 0.00002444, -0.00001236, 0.00002887, -0.00003920, 0.00002706, -0.00003556, 0.00003493, -0.00002701, 0.00003799, -0.00004664, 0.00004894, -0.00005589, 0.00004795, -0.00006630, 0.00007048, -0.00007518, 0.00006814, -0.00008815, 0.00009169, -0.00009974, 0.00010043, -0.00014273, 0.00015633, -0.00020815, 0.00018914, -0.00033508, 0.00033717, -0.00093826, 0.00061791, 0.00154833, -0.00118529, 0.00033307, -0.00034306, 0.00019819, -0.00022148, 0.00014941, -0.00013937, 0.00012237, -0.00012641, 0.00009564, -0.00008872, 0.00007602, -0.00008195, 0.00006524, -0.00006508, 0.00005898, -0.00006773, 0.00005479, -0.00005583, 0.00005801, -0.00004846, 0.00004063, -0.00003563, 0.00002753, -0.00004484, 0.00004449, -0.00002221, 0.00000975, -0.00002408, 0.00002940, -0.00002043, 0.00001802, -0.00002448, 0.00002636, -0.00002330, 0.00001990, -0.00002143, 0.00001975, -0.00001607, 0.00001337, -0.00001734, 0.00001129, -0.00001914, 0.00002403, -0.00002390, 0.00003024, -0.00002118, 0.00001008, 0.00000156, -0.00001020, 0.00001367, -0.00001200, -0.00000104, -0.00000040, -0.00000260, -0.00000771, 0.00001116, -0.00000303, -0.00000219, },
		{ 0.00000706, -0.00002280, 0.00002086, -0.00001429, -0.00000090, 0.00002617, -0.00003502, 0.00002742, -0.00001428, 0.00001189, -0.00000425, 0.00000134, 0.00000754, -0.00000274, -0.00000197, -0.00000615, 0.00001993, -0.00002092, 0.00002536, -0.00002148, 0.00001041, -0.00000258, -0.00000222, 0.00000699, -0.00000658, -0.00000249, 0.00000052, -0.00000551, 0.00000381, -0.00000392, 0.00000506, -0.00000423, 0.00000024, 0.00000091, -0.00000026, -0.00000680, 0.00000705, -0.00000204, 0.00000578, -0.00002425, 0.00002444, -0.00000430, 0.00000402, -0.00001779, 0.00000898, -0.00001455, 0.00001872, -0.00000530, 0.00000347, -0.00000765, 0.00000720, -0.00001543, 0.00000159, -0.00001075, 0.00001527, -0.00002323, 0.00000994, -0.00002075, 0.00002225, -0.00002529, 0.00000394, -0.00002509, 0.00001899, -0.00006821, 0.00000490, -0.00011475, -0.00004313, -0.00082839, 0.00205876, -0.00043299, -0.00032867, 0.00002298, -0.00008999, 0.00000736, -0.00005383, 0.00002252, -0.00002090, 0.00001182, -0.00003333, 0.00002104, -0.00002065, 0.00000956, -0.00002034, 0.00001007, -0.00001055, 0.00000278, -0.00001474, 0.00000770, -0.00000718, 0.00001670, -0.00001893, 0.00001678, -0.00001557, -0.00000087, -0.00000941, 0.00002706, -0.00002272, 0.00000149, -0.00000125, 0.00000979, -0.00000615, -0.00000118, -0.00000002, 0.00000382, -0.00000504, 0.00000268, -0.00000470, 0.00000615, -0.00000493, -0.00000008, -0.00000357, -0.00000568, 0.00000567, -0.00000123, 0.00000060, 0.00001423, -0.00002190, 0.00002535, -0.00002338, 0.00001596, -0.00000671, -0.00000554, -0.00000002, -0.00000080, -0.00000694, 0.00000006, 0.00001126, -0.00001253, },
		{ -0.00000230, -0.00000688, -0.00000907, 0.00002031, -0.00003114, 0.00003598, -0.00001748, -0.00000475, 0.00001729, -0.00001503, 0.00002043, -0.00001943, 0.00002197, -0.00001269, 0.00001355, -0.00002412, 0.00003097, -0.00002222, 0.00001898, -0.00000814, -0.00000323, 0.00000546, -0.00000398, 0.00000107, 0.00000600, -0.00001536, 0.00001206, -0.00001603, 0.00001306, -0.00001407, 0.00001483, -0.00001374, 0.00001246, -0.00001537, 0.00001910, -0.00002580, 0.00002404, -0.00002204, 0.00002968, -0.00004108, 0.00002940, -0.00001244, 0.00002258, -0.00003443, 0.00002482, -0.00003269, 0.00003162, -0.00002016, 0.00002761, -0.00003580, 0.00003751, -0.00004454, 0.00003639, -0.00005376, 0.00005725, -0.00006135, 0.00005388, -0.00007233, 0.00007402, -0.00007925, 0.00007710, -0.00011647, 0.00012330, -0.00017234, 0.00014797, -0.00028831, 0.00026636, -0.00093519, 0.00092614, 0.00111602, -0.00096381, 0.00023559, -0.00027248, 0.00014587, -0.00017503, 0.00011308, -0.00010928, 0.00009753, -0.00010465, 0.00007566, -0.00007138, 0.00006010, -0.00006617, 0.00005019, -0.00005188, 0.00004657, -0.00005514, 0.00004298, -0.00004489, 0.00005088, -0.00004508, 0.00003733, -0.00003297, 0.00002235, -0.00003762, 0.00004358, -0.00002640, 0.00000839, -0.00001716, 0.00002434, -0.00001711, 0.00001287, -0.00001763, 0.00002062, -0.00001949, 0.00001690, -0.00001851, 0.00001727, -0.00001408, 0.00001045, -0.00001412, 0.00000709, -0.00001313, 0.00001948, -0.00002151, 0.00003235, -0.00002842, 0.00002160, -0.00001195, 0.00000101, 0.00000597, -0.00001020, -0.00000099, 0.00000020, -0.00000389, -0.00000740, 0.00001158, -0.00000407, },
		{ -0.00000850, 0.00000172, -0.00001834, 0.00001569, -0.00000910, -0.00000580, 0.00003139, -0.00004006, 0.00003269, -0.00001849, 0.00001858, -0.00001145, 0.00000916, -0.00000014, 0.00000636, -0.00001264, 0.00000859, 0.00000727, -0.00001141, 0.00002033, -0.00002326, 0.00001552, -0.00000793, 0.00000127, 0.00000534, -0.00000890, 0.00000140, -0.00000334, -0.00000089, -0.00000052, 0.00000021, 0.00000078, -0.00000014, -0.00000351, 0.00000572, -0.00000832, 0.00000296, -0.00000155, 0.00000712, -0.00000815, -0.00000978, 0.00001767, -0.00000059, -0.00000465, -0.00000695, -0.00000025, -0.00000605, 0.00001268, 0.00000052, -0.00000514, 0.00000275, -0.00000526, -0.00000167, -0.00001300, 0.00000546, -0.00000278, -0.00000330, -0.00001106, 0.00000202, -0.00000254, 0.00000275, -0.00002838, 0.00001144, -0.00002903, -0.00000795, -0.00007449, -0.00000877, -0.00026277, -0.00042880, 0.00180712, -0.00049398, -0.00022005, -0.00004682, -0.00003358, -0.00003626, -0.00001204, -0.00001443, 0.00001229, -0.00001731, -0.00000743, -0.00000315, 0.00000085, -0.00000938, -0.00000228, -0.00000778, 0.00000588, -0.00001238, 0.00000072, -0.00000819, 0.00001084, -0.00000128, -0.00000432, 0.00000395, -0.00000622, -0.00000994, 0.00000511, 0.00001357, -0.00001836, 0.00000014, 0.00000415, 0.00000374, -0.00000259, -0.00000437, 0.00000531, -0.00000284, 0.00000118, -0.00000279, 0.00000062, 0.00000120, -0.00000132, -0.00000357, -0.00000052, -0.00000872, 0.00001197, -0.00001017, 0.00001279, 0.00000058, -0.00000948, 0.00001689, -0.00002144, 0.00001898, -0.00001392, 0.00000044, -0.00000393, 0.00000260, -0.00001145, 0.00000492, 0.00000628, },
		{ -0.00000190, -0.00000540, -0.00000504, -0.00001130, 0.00002164, -0.00003202, 0.00003723, -0.00001868, -0.00000286, 0.00001550, -0.00001076, 0.00001676, -0.00001584, 0.00001902, -0.00000907, 0.00000995, -0.00002007, 0.00003279, -0.00002817, 0.00002840, -0.00002058, 0.00000678, -0.00000017, -0.00000276, 0.00000279, 0.00000100, -0.00001067, 0.00000834, -0.00001202, 0.00000969, -0.00001066, 0.00001097, -0.00000965, 0.00000774, -0.00000869, 0.00000995, -0.00001705, 0.00001725, -0.00001436, 0.00001952, -0.00003451, 0.00002989, -0.00001094, 0.00001447, -0.00002626, 0.00001910, -0.00002695, 0.00002638, -0.00001238, 0.00001584, -0.00002243, 0.00002262, -0.00002906, 0.00002025, -0.00003596, 0.00003792, -0.00004079, 0.00003249, -0.00004883, 0.00004803, -0.00004967, 0.00004317, -0.00007744, 0.00007385, -0.00011488, 0.00007833, -0.00020184, 0.00012575, -0.00080216, 0.00108637, 0.00075650, -0.00074267, 0.00012295, -0.00018850, 0.00008161, -0.00011708, 0.00006537, -0.00006847, 0.00006253, -0.00007361, 0.00004733, -0.00004657, 0.00003756, -0.00004425, 0.00002919, -0.00003286, 0.00002852, -0.00003685, 0.00002515, -0.00002710, 0.00003613, -0.00003482, 0.00002851, -0.00002618, 0.00001375, -0.00002555, 0.00003683, -0.00002733, 0.00000685, -0.00000985, 0.00001782, -0.00001276, 0.00000729, -0.00000982, 0.00001290, -0.00001337, 0.00001174, -0.00001356, 0.00001288, -0.00001064, 0.00000639, -0.00000985, 0.00000186, -0.00000501, 0.00001102, -0.00001339, 0.00002735, -0.00002854, 0.00002743, -0.00002283, 0.00001288, -0.00000499, -0.00000421, -0.00000420, 0.00000338, -0.00000794, -0.00000371, 0.00000841, },
		{ 0.00000285, -0.00000575, -0.00000248, -0.00001446, 0.00001089, -0.00000484, -0.00000904, 0.00003432, -0.00004219, 0.00003515, -0.00001985, 0.00002219, -0.00001584, 0.00001480, -0.00000589, 0.00001336, -0.00002100, 0.00002270, -0.00000747, 0.00000246, 0.00000860, -0.00001817, 0.00001562, -0.00001083, 0.00000460, 0.00000196, -0.00000857, 0.00000288, -0.00000493, 0.00000163, -0.00000288, 0.00000223, -0.00000118, 0.00000131, -0.00000386, 0.00000595, -0.00001126, 0.00000819, -0.00000592, 0.00001133, -0.00001749, 0.00000286, 0.00001135, 0.00000060, -0.00000908, 0.00000009, -0.00000620, -0.00000064, 0.00001018, 0.00000098, -0.00000675, 0.00000520, -0.00000916, 0.00000302, -0.00001769, 0.00001094, -0.00000950, 0.00000465, -0.00001908, 0.00001055, -0.00001143, 0.00001235, -0.00003917, 0.00002228, -0.00004648, 0.00001605, -0.00010901, 0.00003581, -0.00040870, -0.00005157, 0.00148067, -0.00044945, -0.00018624, -0.00006551, -0.00001724, -0.00004611, 0.00000054, -0.00002629, 0.00002358, -0.00002741, 0.00000171, -0.00001235, 0.00000875, -0.00001557, 0.00000361, -0.00001408, 0.00001137, -0.00001695, 0.00000584, -0.00001391, 0.00001955, -0.00001203, 0.00000482, -0.00000438, -0.00000142, -0.00001326, 0.00001461, 0.00000177, -0.00001431, 0.00000110, 0.00000602, 0.00000062, -0.00000183, -0.00000383, 0.00000630, -0.00000555, 0.00000407, -0.00000528, 0.00000325, -0.00000150, 0.00000005, -0.00000456, -0.00000036, -0.00000775, 0.00001337, -0.00001423, 0.00002125, -0.00001181, 0.00000477, 0.00000342, -0.00001273, 0.00001515, -0.00001586, 0.00000333, -0.00000550, 0.00000344, -0.00001320, 0.00000750, },
		{ 0.00000309, 0.00000234, -0.00001052, -0.00000097, -0.00001536, 0.00002398, -0.00003349, 0.00003904, -0.00002062, 0.00000051, 0.00001191, -0.00000440, 0.00001020, -0.00000860, 0.00001218, -0.00000143, 0.00000173, -0.00000925, 0.00002624, -0.00002564, 0.00002996, -0.00002803, 0.00001537, -0.00000693, 0.00000078, 0.00000195, -0.00000222, -0.00000649, 0.00000456, -0.00000763, 0.00000579, -0.00000693, 0.00000695, -0.00000565, 0.00000358, -0.00000288, 0.00000097, -0.00000720, 0.00000895, -0.00000532, 0.00000677, -0.00002329, 0.00002652, -0.00000857, 0.00000623, -0.00001676, 0.00001176, -0.00002001, 0.00002098, -0.00000604, 0.00000603, -0.00001066, 0.00000887, -0.00001438, 0.00000489, -0.00001872, 0.00001875, -0.00002014, 0.00001108, -0.00002541, 0.00002235, -0.00002085, 0.00001007, -0.00003915, 0.00002410, -0.00005494, 0.00000336, -0.00010514, -0.00004826, -0.00056852, 0.00109682, 0.00050241, -0.00056067, 0.00002590, -0.00011492, 0.00002523, -0.00006488, 0.00002130, -0.00003002, 0.00002879, -0.00004335, 0.00001964, -0.00002237, 0.00001585, -0.00002327, 0.00000897, -0.00001442, 0.00001098, -0.00001884, 0.00000714, -0.00000798, 0.00001845, -0.00002128, 0.00001700, -0.00001780, 0.00000456, -0.00001151, 0.00002601, -0.00002553, 0.00000575, -0.00000359, 0.00001145, -0.00000884, 0.00000301, -0.00000321, 0.00000549, -0.00000696, 0.00000622, -0.00000816, 0.00000798, -0.00000703, 0.00000257, -0.00000609, -0.00000243, 0.00000265, 0.00000144, -0.00000194, 0.00001656, -0.00002126, 0.00002544, -0.00002741, 0.00002112, -0.00001532, 0.00000310, -0.00000866, 0.00000729, -0.00001286, 0.00000158, },
		{ 0.00000659, 0.00000231, -0.00000563, -0.00000412, -0.00001253, 0.00000756, -0.00000236, -0.00000995, 0.00003436, -0.00004063, 0.00003381, -0.00001714, 0.00002114, -0.00001542, 0.00001588, -0.00000717, 0.00001546, -0.00002346, 0.00003163, -0.00001921, 0.00001570, -0.00000590, -0.00000816, 0.00001088, -0.00001025, 0.00000623, -0.00000170, -0.00000651, 0.00000236, -0.00000443, 0.00000206, -0.00000333, 0.00000230, -0.00000112, 0.00000061, -0.00000152, 0.00000219, -0.00000914, 0.00000845, -0.00000549, 0.00000975, -0.00002057, 0.00001167, 0.00000650, -0.00000069, -0.00000930, 0.00000303, -0.00000864, 0.00000189, 0.00001022, -0.00000225, -0.00000312, 0.00000134, -0.00000599, 0.00000008, -0.00001390, 0.00000701, -0.00000610, 0.00000188, -0.00001550, 0.00000644, -0.00000631, 0.00000569, -0.00003082, 0.00000976, -0.00003501, 0.00000426, -0.00009708, 0.00001056, -0.00045207, 0.00024680, 0.00114247, -0.00034276, -0.00020409, -0.00004704, -0.00003076, -0.00003011, -0.00001007, -0.00001759, 0.00001642, -0.00002124, -0.00000358, -0.00000859, 0.00000497, -0.00001089, -0.00000092, -0.00001039, 0.00000731, -0.00001214, 0.00000155, -0.00000984, 0.00001846, -0.00001449, 0.00000743, -0.00000754, -0.00000132, -0.00001054, 0.00001787, -0.00000691, -0.00001086, 0.00000363, 0.00000507, -0.00000020, -0.00000293, -0.00000069, 0.00000381, -0.00000475, 0.00000385, -0.00000488, 0.00000324, -0.00000203, -0.00000051, -0.00000352, -0.00000246, -0.00000334, 0.00000977, -0.00001218, 0.00002336, -0.00001915, 0.00001641, -0.00001083, -0.00000027, 0.00000582, -0.00001244, 0.00000244, -0.00000412, 0.00000115, -0.00001137, },
		{ 0.00000607, -0.00000215, 0.00000663, -0.00001579, 0.00000366, -0.00001989, 0.00002603, -0.00003419, 0.00003983, -0.00002133, 0.00000314, 0.00000897, 0.00000132, 0.00000371, -0.00000070, 0.00000436, 0.00000728, -0.00000789, 0.00000509, 0.00001373, -0.00001575, 0.00002295, -0.00002809, 0.00001952, -0.00001210, 0.00000481, -0.00000071, -0.00000339, -0.00000361, 0.00000168, -0.00000389, 0.00000238, -0.00000382, 0.00000377, -0.00000280, 0.00000122, 0.00000030, -0.00000538, 0.00000100, 0.00000182, 0.00000222, -0.00000536, -0.00001009, 0.00002069, -0.00000629, -0.00000056, -0.00000791, 0.00000462, -0.00001334, 0.00001660, -0.00000251, 0.00000027, -0.00000328, -0.00000041, -0.00000424, -0.00000558, -0.00000665, 0.00000482, -0.00000487, -0.00000451, -0.00000842, 0.00000399, -0.00000079, -0.00001281, -0.00001297, -0.00001177, -0.00001021, -0.00005440, -0.00002947, -0.00021073, -0.00028368, 0.00098489, 0.00035840, -0.00043675, -0.00003759, -0.00006657, -0.00001033, -0.00003023, -0.00000836, -0.00000365, 0.00000506, -0.00002169, -0.00000042, -0.00000509, 0.00000077, -0.00000872, -0.00000516, -0.00000168, -0.00000105, -0.00000613, -0.00000585, 0.00000718, 0.00000272, -0.00000844, 0.00000591, -0.00001044, -0.00000240, 0.00000132, 0.00001351, -0.00002188, 0.00000550, 0.00000070, 0.00000633, -0.00000625, 0.00000103, 0.00000073, 0.00000012, -0.00000195, 0.00000188, -0.00000374, 0.00000379, -0.00000431, 0.00000002, -0.00000401, -0.00000416, 0.00000743, -0.00000606, 0.00000928, 0.00000324, -0.00000876, 0.00001613, -0.00002427, 0.00002289, -0.00002164, 0.00000875, -0.00001199, 0.00000978, -0.00001632, },
		{ -0.00000671, 0.00000260, 0.00000457, -0.00000821, -0.00000299, -0.00001286, 0.00000586, -0.00000158, -0.00000889, 0.00003209, -0.00003599, 0.00002923, -0.00001087, 0.00001573, -0.00001008, 0.00001196, -0.00000327, 0.00001175, -0.00001843, 0.00003251, -0.00002398, 0.00002342, -0.00001830, 0.00000313, 0.00000324, -0.00000666, 0.00000556, -0.00000449, -0.00000381, 0.00000065, -0.00000249, 0.00000094, -0.00000239, 0.00000104, 0.00000016, -0.00000111, 0.00000197, -0.00000380, -0.00000347, 0.00000484, -0.00000128, 0.00000320, -0.00001731, 0.00001572, 0.00000353, -0.00000385, -0.00000617, 0.00000250, -0.00000813, 0.00000236, 0.00001126, -0.00000674, 0.00000278, -0.00000569, 0.00000100, -0.00000706, -0.00000532, -0.00000244, 0.00000340, -0.00000739, -0.00000495, -0.00000514, 0.00000687, -0.00001029, -0.00001142, -0.00001676, -0.00000512, -0.00003167, -0.00005199, -0.00007240, -0.00038935, 0.00043201, 0.00084529, -0.00021859, -0.00024431, -0.00001193, -0.00005785, -0.00000192, -0.00003242, 0.00000208, -0.00000102, -0.00000578, -0.00001719, 0.00000266, -0.00000541, -0.00000016, -0.00001128, -0.00000102, -0.00000218, -0.00000188, -0.00000847, 0.00000080, 0.00000990, -0.00000999, 0.00000436, -0.00000644, -0.00000461, -0.00000307, 0.00001513, -0.00001186, -0.00000801, 0.00000678, 0.00000243, 0.00000028, -0.00000459, 0.00000335, -0.00000053, -0.00000172, 0.00000153, -0.00000248, 0.00000133, -0.00000112, -0.00000215, -0.00000153, -0.00000540, 0.00000278, 0.00000271, -0.00000473, 0.00001847, -0.00001920, 0.00002175, -0.00002135, 0.00001163, -0.00000572, -0.00000558, -0.00000115, -0.00000074, -0.00000332, },
		{ -0.00001665, 0.00000781, -0.00000531, 0.00000905, -0.00001931, 0.00000712, -0.00002341, 0.00002655, -0.00003296, 0.00003832, -0.00001925, 0.00000317, 0.00000882, 0.00000387, 0.00000019, 0.00000470, -0.00000115, 0.00001357, -0.00001480, 0.00001830, -0.00000020, -0.00000240, 0.00000973, -0.00002103, 0.00001798, -0.00001386, 0.00000757, -0.00000397, -0.00000290, -0.00000227, 0.00000021, -0.00000143, 0.00000007, -0.00000190, 0.00000190, -0.00000152, 0.00000110, 0.00000007, -0.00000769, 0.00000568, -0.00000214, 0.00000595, -0.00001394, 0.00000221, 0.00001407, -0.00000493, -0.00000480, -0.00000119, -0.00000098, -0.00000790, 0.00001361, -0.00000191, -0.00000105, -0.00000117, -0.00000395, -0.00000016, -0.00000938, -0.00000186, -0.00000143, 0.00000234, -0.00001140, -0.00000093, -0.00000374, 0.00000676, -0.00002097, -0.00000459, -0.00002609, 0.00000886, -0.00008040, 0.00000194, -0.00032087, -0.00000388, 0.00079723, 0.00030544, -0.00036991, -0.00006438, -0.00004675, -0.00002169, -0.00001685, -0.00001969, 0.00000674, -0.00000491, -0.00001209, -0.00000968, 0.00000248, -0.00000517, -0.00000301, -0.00001086, 0.00000305, -0.00000520, -0.00000127, -0.00001092, 0.00001493, -0.00000744, 0.00000055, -0.00000235, -0.00000592, -0.00000516, 0.00001030, 0.00000169, -0.00001732, 0.00000619, 0.00000273, 0.00000288, -0.00000511, 0.00000138, 0.00000160, -0.00000240, 0.00000073, -0.00000040, -0.00000115, 0.00000104, -0.00000298, -0.00000084, -0.00000400, -0.00000271, 0.00000791, -0.00000896, 0.00001669, -0.00000851, 0.00000498, 0.00000251, -0.00001471, 0.00001772, -0.00002227, 0.00001085, -0.00001262, 0.00000936, },
		{ -0.00000795, -0.00000121, -0.00000281, 0.00000822, -0.00001225, -0.00000008, -0.00001466, 0.00000501, -0.00000154, -0.00000713, 0.00002910, -0.00003006, 0.00002330, -0.00000305, 0.00000800, -0.00000170, 0.00000468, 0.00000427, 0.00000378, -0.00000701, 0.00002535, -0.00002044, 0.00002271, -0.00002456, 0.00001180, -0.00000444, -0.00000163, 0.00000295, -0.00000586, -0.00000137, -0.00000136, 0.00000005, -0.00000095, -0.00000080, -0.00000071, 0.00000169, -0.00000263, 0.00000485, -0.00000971, 0.00000341, -0.00000050, 0.00000455, -0.00000609, -0.00000916, 0.00001531, 0.00000222, -0.00000782, -0.00000111, -0.00000032, -0.00000564, 0.00000171, 0.00001189, -0.00001023, 0.00000797, -0.00001245, 0.00000809, -0.00001436, 0.00000363, -0.00001265, 0.00001396, -0.00001778, 0.00000673, -0.00001771, 0.00002067, -0.00002691, 0.00000868, -0.00004475, 0.00002822, -0.00007362, 0.00000305, -0.00018438, -0.00024206, 0.00049757, 0.00062139, -0.00011171, -0.00028105, 0.00002044, -0.00008249, 0.00002446, -0.00005426, 0.00002200, -0.00001938, 0.00001075, -0.00003181, 0.00001484, -0.00001636, 0.00001087, -0.00002199, 0.00000878, -0.00001203, 0.00000882, -0.00001922, 0.00001320, -0.00000203, -0.00000161, -0.00000214, -0.00000304, -0.00000901, 0.00000668, 0.00000786, -0.00001327, -0.00000558, 0.00000958, -0.00000077, 0.00000109, -0.00000564, 0.00000667, -0.00000492, 0.00000192, -0.00000144, 0.00000065, -0.00000140, 0.00000027, -0.00000383, 0.00000018, -0.00000752, 0.00000830, -0.00000508, 0.00000557, 0.00000821, -0.00001214, 0.00001923, -0.00002495, 0.00001892, -0.00001558, 0.00000174, -0.00000531, 0.00000274, },
		{ 0.00000597, -0.00001354, 0.00000611, -0.00000547, 0.00000862, -0.00002002, 0.00000834, -0.00002510, 0.00002495, -0.00002942, 0.00003423, -0.00001397, -0.00000005, 0.00001231, 0.00000208, 0.00000131, 0.00000557, -0.00000205, 0.00001486, -0.00001570, 0.00002606, -0.00001059, 0.00000924, -0.00000519, -0.00000958, 0.00001166, -0.00001184, 0.00000802, -0.00000674, -0.00000144, -0.00000224, 0.00000013, -0.00000039, -0.00000097, -0.00000124, 0.00000130, -0.00000153, 0.00000269, -0.00000310, -0.00000593, 0.00000627, -0.00000222, 0.00000488, -0.00001712, 0.00001131, 0.00000816, -0.00000496, -0.00000606, 0.00000267, -0.00000431, -0.00000405, 0.00001170, -0.00000324, 0.00000089, -0.00000331, -0.00000261, -0.00000133, -0.00000729, -0.00000367, -0.00000052, 0.00000197, -0.00001015, -0.00000223, -0.00000179, 0.00000301, -0.00001579, -0.00001269, -0.00001958, 0.00000187, -0.00007247, -0.00001892, -0.00035303, 0.00022204, 0.00058606, 0.00031054, -0.00034459, -0.00006332, -0.00004889, -0.00001409, -0.00002081, -0.00001554, 0.00000322, -0.00000264, -0.00001336, -0.00000918, 0.00000134, -0.00000304, -0.00000511, -0.00000902, 0.00000058, -0.00000220, -0.00000374, -0.00000810, 0.00001453, -0.00001064, 0.00000420, -0.00000668, -0.00000478, -0.00000314, 0.00001395, -0.00000760, -0.00001268, 0.00000760, 0.00000274, 0.00000098, -0.00000496, 0.00000332, -0.00000003, -0.00000228, 0.00000109, -0.00000052, -0.00000054, -0.00000020, -0.00000293, -0.00000034, -0.00000553, 0.00000136, 0.00000417, -0.00000627, 0.00001796, -0.00001508, 0.00001556, -0.00001100, -0.00000225, 0.00000760, -0.00001774, 0.00000922, -0.00001041, },
		{ 0.00000435, -0.00001064, 0.00000288, -0.00000746, 0.00001129, -0.00001589, 0.00000293, -0.00001657, 0.00000377, -0.00000094, -0.00000618, 0.00002727, -0.00002501, 0.00001843, 0.00000361, 0.00000094, 0.00000666, -0.00000291, 0.00001237, -0.00000503, 0.00000729, 0.00001299, -0.00001040, 0.00001372, -0.00002290, 0.00001516, -0.00000954, 0.00000286, -0.00000057, -0.00000586, 0.00000031, -0.00000297, 0.00000241, -0.00000285, 0.00000072, -0.00000223, 0.00000272, -0.00000302, 0.00000573, -0.00001352, 0.00000910, -0.00000518, 0.00000944, -0.00001522, 0.00000138, 0.00001169, 0.00000189, -0.00001141, 0.00000430, -0.00000399, -0.00000228, 0.00000069, 0.00001129, -0.00001131, 0.00001040, -0.00001644, 0.00001246, -0.00001870, 0.00000939, -0.00001966, 0.00002134, -0.00002476, 0.00001464, -0.00002590, 0.00002898, -0.00003697, 0.00002071, -0.00006319, 0.00005110, -0.00010377, 0.00004304, -0.00028901, -0.00004959, 0.00046259, 0.00047913, -0.00004066, -0.00029853, 0.00003767, -0.00009403, 0.00003934, -0.00006667, 0.00003408, -0.00003136, 0.00002182, -0.00004162, 0.00002273, -0.00002312, 0.00001767, -0.00002866, 0.00001477, -0.00001806, 0.00001570, -0.00002626, 0.00002269, -0.00001291, 0.00000703, -0.00000929, 0.00000046, -0.00001209, 0.00001583, -0.00000165, -0.00001196, -0.00000331, 0.00001133, -0.00000364, 0.00000159, -0.00000542, 0.00000819, -0.00000796, 0.00000477, -0.00000381, 0.00000332, -0.00000393, 0.00000131, -0.00000474, 0.00000071, -0.00000751, 0.00001113, -0.00001065, 0.00001525, -0.00000406, -0.00000052, 0.00000993, -0.00002090, 0.00001935, -0.00002078, 0.00000680, -0.00000789, },
		{ -0.00000657, 0.00000087, -0.00000813, 0.00000176, -0.00000313, 0.00000567, -0.00001805, 0.00000731, -0.00002504, 0.00002159, -0.00002421, 0.00002843, -0.00000637, -0.00000566, 0.00001861, -0.00000338, 0.00000671, 0.00000190, 0.00000194, 0.00001067, -0.00000945, 0.00002604, -0.00001405, 0.00001465, -0.00001684, 0.00000219, 0.00000308, -0.00000711, 0.00000613, -0.00000832, 0.00000028, -0.00000300, 0.00000109, -0.00000054, -0.00000095, -0.00000153, 0.00000149, -0.00000210, 0.00000484, -0.00000775, -0.00000134, 0.00000364, 0.00000084, -0.00000041, -0.00001462, 0.00001600, 0.00000384, -0.00000633, -0.00000463, 0.00000370, -0.00000532, -0.00000172, 0.00001019, -0.00000502, 0.00000410, -0.00000749, 0.00000135, -0.00000537, -0.00000193, -0.00000935, 0.00000474, -0.00000304, -0.00000397, -0.00000872, 0.00000564, -0.00000705, -0.00000321, -0.00003025, -0.00000018, -0.00002229, -0.00004087, -0.00008274, -0.00030303, 0.00036371, 0.00039591, 0.00033830, -0.00033911, -0.00004927, -0.00006103, 0.00000229, -0.00003347, -0.00000318, -0.00000805, 0.00000663, -0.00002094, -0.00000297, -0.00000474, 0.00000355, -0.00001155, -0.00000295, -0.00000589, 0.00000484, -0.00001057, -0.00000004, 0.00000782, -0.00000778, 0.00000287, -0.00000743, -0.00000618, 0.00000272, 0.00001220, -0.00001339, -0.00000843, 0.00000929, 0.00000135, 0.00000014, -0.00000506, 0.00000569, -0.00000296, -0.00000045, -0.00000015, 0.00000091, -0.00000145, -0.00000034, -0.00000359, 0.00000071, -0.00000739, 0.00000644, -0.00000220, 0.00000107, 0.00001277, -0.00001467, 0.00001972, -0.00002012, 0.00000875, -0.00000390, -0.00001033, 0.00000524, },
		{ -0.00000770, 0.00000306, -0.00001012, 0.00000399, -0.00000967, 0.00001214, -0.00001749, 0.00000452, -0.00001732, 0.00000111, 0.00000115, -0.00000706, 0.00002791, -0.00002245, 0.00001646, 0.00000689, -0.00000281, 0.00001203, -0.00000765, 0.00001764, -0.00001073, 0.00001989, 0.00000004, 0.00000185, -0.00000047, -0.00001440, 0.00001260, -0.00001064, 0.00000528, -0.00000386, -0.00000492, 0.00000108, -0.00000379, 0.00000405, -0.00000421, 0.00000169, -0.00000314, 0.00000295, -0.00000204, 0.00000410, -0.00001406, 0.00001193, -0.00000741, 0.00001132, -0.00002144, 0.00001152, 0.00000657, 0.00000169, -0.00001358, 0.00000870, -0.00000725, 0.00000105, -0.00000033, 0.00000947, -0.00000980, 0.00000948, -0.00001670, 0.00001296, -0.00001870, 0.00001034, -0.00002155, 0.00002338, -0.00002606, 0.00001638, -0.00002717, 0.00002902, -0.00003715, 0.00002047, -0.00006630, 0.00005553, -0.00011090, 0.00004973, -0.00035325, 0.00014243, 0.00036322, 0.00040642, -0.00000729, -0.00029296, 0.00003659, -0.00008958, 0.00003960, -0.00006636, 0.00003503, -0.00003375, 0.00002446, -0.00004391, 0.00002395, -0.00002356, 0.00001826, -0.00002934, 0.00001506, -0.00001834, 0.00001667, -0.00002711, 0.00002625, -0.00001943, 0.00001294, -0.00001481, 0.00000240, -0.00001212, 0.00002189, -0.00001103, -0.00000901, -0.00000102, 0.00001167, -0.00000568, 0.00000160, -0.00000391, 0.00000761, -0.00000900, 0.00000605, -0.00000478, 0.00000477, -0.00000563, 0.00000157, -0.00000460, -0.00000017, -0.00000501, 0.00001019, -0.00001183, 0.00002100, -0.00001435, 0.00001164, -0.00000281, -0.00001110, 0.00001310, -0.00002021, 0.00000812, },
		{ 0.00000109, -0.00000304, -0.00000389, -0.00000248, -0.00000341, 0.00000016, 0.00000161, -0.00001460, 0.00000497, -0.00002416, 0.00001756, -0.00001872, 0.00002260, 0.00000165, -0.00001167, 0.00002558, -0.00001027, 0.00001427, -0.00000438, 0.00000901, 0.00000285, 0.00000244, 0.00001877, -0.00000996, 0.00001167, -0.00002162, 0.00001046, -0.00000469, -0.00000163, 0.00000272, -0.00000862, 0.00000173, -0.00000402, 0.00000256, -0.00000138, -0.00000032, -0.00000228, 0.00000190, -0.00000240, 0.00000622, -0.00001203, 0.00000408, -0.00000040, 0.00000524, -0.00000803, -0.00000765, 0.00001633, 0.00000129, -0.00000851, -0.00000139, 0.00000258, -0.00000444, -0.00000058, 0.00000845, -0.00000587, 0.00000655, -0.00001132, 0.00000530, -0.00000941, 0.00000353, -0.00001542, 0.00001069, -0.00000882, 0.00000293, -0.00001575, 0.00001317, -0.00001707, 0.00000909, -0.00004783, 0.00002066, -0.00004964, -0.00000387, -0.00016642, -0.00018784, 0.00041412, 0.00025462, 0.00036075, -0.00033401, -0.00003650, -0.00007128, 0.00001687, -0.00004542, 0.00000913, -0.00001981, 0.00001656, -0.00002922, 0.00000393, -0.00001116, 0.00001019, -0.00001808, 0.00000325, -0.00001234, 0.00001189, -0.00001766, 0.00000921, -0.00000168, -0.00000142, -0.00000169, -0.00000601, -0.00000840, 0.00001036, 0.00000632, -0.00001564, -0.00000475, 0.00001070, -0.00000070, -0.00000016, -0.00000474, 0.00000737, -0.00000584, 0.00000186, -0.00000192, 0.00000294, -0.00000309, -0.00000006, -0.00000422, 0.00000140, -0.00000826, 0.00001051, -0.00000865, 0.00001049, 0.00000296, -0.00000784, 0.00001645, -0.00002222, 0.00001461, -0.00001303, -0.00000302, },
		{ 0.00000592, -0.00000495, -0.00000084, -0.00000654, 0.00000188, -0.00000897, 0.00001019, -0.00001635, 0.00000393, -0.00001636, -0.00000327, 0.00000482, -0.00000982, 0.00003118, -0.00002271, 0.00001789, 0.00000604, -0.00000208, 0.00001284, -0.00000765, 0.00001791, -0.00001041, 0.00002690, -0.00000897, 0.00001125, -0.00001501, -0.00000238, 0.00000564, -0.00000794, 0.00000504, -0.00000609, -0.00000358, 0.00000108, -0.00000374, 0.00000475, -0.00000479, 0.00000194, -0.00000339, 0.00000256, -0.00000014, 0.00000043, -0.00001139, 0.00001145, -0.00000653, 0.00000928, -0.00002304, 0.00001896, 0.00000161, 0.00000094, -0.00001376, 0.00001124, -0.00000923, 0.00000375, -0.00000135, 0.00000708, -0.00000660, 0.00000599, -0.00001387, 0.00001014, -0.00001497, 0.00000696, -0.00001868, 0.00002042, -0.00002206, 0.00001247, -0.00002228, 0.00002180, -0.00002869, 0.00000926, -0.00005502, 0.00004169, -0.00009401, 0.00001757, -0.00035690, 0.00029443, 0.00024080, 0.00037902, -0.00000134, -0.00027025, 0.00002194, -0.00007312, 0.00002832, -0.00005558, 0.00002645, -0.00002768, 0.00001953, -0.00003942, 0.00001925, -0.00001853, 0.00001346, -0.00002480, 0.00001035, -0.00001352, 0.00001222, -0.00002189, 0.00002323, -0.00002025, 0.00001461, -0.00001751, 0.00000219, -0.00000857, 0.00002346, -0.00001838, -0.00000543, 0.00000131, 0.00001063, -0.00000676, 0.00000130, -0.00000165, 0.00000539, -0.00000816, 0.00000571, -0.00000423, 0.00000480, -0.00000634, 0.00000110, -0.00000367, -0.00000193, -0.00000068, 0.00000580, -0.00000799, 0.00002087, -0.00001946, 0.00002025, -0.00001460, 0.00000068, 0.00000261, -0.00001486, },
		{ 0.00000173, -0.00000133, -0.00000148, -0.00000647, 0.00000134, -0.00000737, 0.00000251, -0.00000176, -0.00001129, 0.00000267, -0.00002364, 0.00001409, -0.00001437, 0.00001850, 0.00000809, -0.00001587, 0.00003070, -0.00001575, 0.00002096, -0.00001018, 0.00001602, -0.00000511, 0.00001634, 0.00000738, -0.00000077, 0.00000130, -0.00001851, 0.00001295, -0.00000922, 0.00000261, -0.00000100, -0.00000795, 0.00000268, -0.00000484, 0.00000401, -0.00000240, 0.00000045, -0.00000306, 0.00000207, -0.00000188, 0.00000587, -0.00001431, 0.00000833, -0.00000385, 0.00000879, -0.00001540, 0.00000152, 0.00001337, 0.00000009, -0.00001067, 0.00000251, 0.00000032, -0.00000237, -0.00000029, 0.00000619, -0.00000505, 0.00000694, -0.00001308, 0.00000731, -0.00001127, 0.00000656, -0.00001906, 0.00001427, -0.00001211, 0.00000709, -0.00001954, 0.00001653, -0.00002196, 0.00001485, -0.00005740, 0.00003262, -0.00006675, 0.00001921, -0.00024093, -0.00003917, 0.00038767, 0.00017039, 0.00036278, -0.00031748, -0.00003394, -0.00007204, 0.00002261, -0.00005006, 0.00001528, -0.00002651, 0.00002215, -0.00003375, 0.00000757, -0.00001431, 0.00001345, -0.00002137, 0.00000636, -0.00001556, 0.00001562, -0.00002143, 0.00001575, -0.00001017, 0.00000529, -0.00000717, -0.00000414, -0.00000950, 0.00001730, -0.00000164, -0.00001503, -0.00000156, 0.00001134, -0.00000273, -0.00000031, -0.00000367, 0.00000766, -0.00000765, 0.00000360, -0.00000324, 0.00000464, -0.00000468, 0.00000007, -0.00000429, 0.00000112, -0.00000721, 0.00001187, -0.00001255, 0.00001868, -0.00000810, 0.00000269, 0.00000729, -0.00001720, 0.00001367, -0.00001724, },
		{ -0.00000723, 0.00000179, -0.00000101, -0.00000587, -0.00000129, -0.00000236, -0.00000614, 0.00000609, -0.00001290, 0.00000142, -0.00001402, -0.00000878, 0.00000926, -0.00001343, 0.00003601, -0.00002474, 0.00002165, 0.00000207, 0.00000241, 0.00000950, -0.00000303, 0.00001313, -0.00000343, 0.00002661, -0.00001138, 0.00001393, -0.00002513, 0.00000889, -0.00000279, -0.00000298, 0.00000257, -0.00000697, -0.00000228, 0.00000059, -0.00000298, 0.00000460, -0.00000461, 0.00000157, -0.00000321, 0.00000204, 0.00000177, -0.00000403, -0.00000663, 0.00000842, -0.00000318, 0.00000380, -0.00001970, 0.00002243, -0.00000209, -0.00000062, -0.00001192, 0.00001166, -0.00000959, 0.00000551, -0.00000254, 0.00000503, -0.00000310, 0.00000152, -0.00000957, 0.00000576, -0.00000943, 0.00000131, -0.00001314, 0.00001461, -0.00001512, 0.00000555, -0.00001438, 0.00001114, -0.00001623, -0.00000730, -0.00003588, 0.00001701, -0.00006194, -0.00004480, -0.00029764, 0.00038223, 0.00013056, 0.00037034, -0.00000739, -0.00024105, 0.00000259, -0.00005244, 0.00001228, -0.00004007, 0.00001320, -0.00001727, 0.00001060, -0.00003130, 0.00001155, -0.00001086, 0.00000602, -0.00001769, 0.00000326, -0.00000621, 0.00000492, -0.00001295, 0.00001530, -0.00001611, 0.00001218, -0.00001748, 0.00000042, -0.00000225, 0.00002047, -0.00002264, -0.00000194, 0.00000351, 0.00000854, -0.00000704, 0.00000107, 0.00000057, 0.00000243, -0.00000615, 0.00000429, -0.00000261, 0.00000372, -0.00000632, 0.00000031, -0.00000258, -0.00000355, 0.00000397, -0.00000038, -0.00000033, 0.00001494, -0.00001803, 0.00002256, -0.00002160, 0.00001024, -0.00000845, },
		{ -0.00001591, 0.00000283, -0.00000115, -0.00000260, -0.00000600, 0.00000216, -0.00000880, 0.00000258, -0.00000307, -0.00000944, 0.00000148, -0.00002433, 0.00001195, -0.00001202, 0.00001719, 0.00001165, -0.00001676, 0.00003216, -0.00001758, 0.00002421, -0.00001266, 0.00001986, -0.00000953, 0.00002795, -0.00000371, 0.00000912, -0.00001278, -0.00000930, 0.00000970, -0.00000955, 0.00000435, -0.00000397, -0.00000679, 0.00000308, -0.00000521, 0.00000501, -0.00000315, 0.00000099, -0.00000357, 0.00000186, -0.00000049, 0.00000355, -0.00001379, 0.00001007, -0.00000520, 0.00000973, -0.00002014, 0.00001033, 0.00000877, -0.00000053, -0.00001198, 0.00000593, -0.00000204, 0.00000014, -0.00000062, 0.00000360, -0.00000268, 0.00000503, -0.00001223, 0.00000665, -0.00001013, 0.00000614, -0.00001900, 0.00001406, -0.00001142, 0.00000693, -0.00001851, 0.00001402, -0.00001973, 0.00001151, -0.00005529, 0.00003038, -0.00006592, 0.00001512, -0.00028061, 0.00010650, 0.00031272, 0.00013475, 0.00034249, -0.00028705, -0.00004332, -0.00006179, 0.00001774, -0.00004533, 0.00001309, -0.00002594, 0.00002136, -0.00003265, 0.00000631, -0.00001279, 0.00001202, -0.00002014, 0.00000512, -0.00001425, 0.00001457, -0.00002007, 0.00001716, -0.00001485, 0.00000975, -0.00001159, -0.00000316, -0.00000817, 0.00002143, -0.00000950, -0.00001261, 0.00000129, 0.00001090, -0.00000427, -0.00000048, -0.00000192, 0.00000645, -0.00000794, 0.00000419, -0.00000352, 0.00000543, -0.00000574, -0.00000024, -0.00000366, -0.00000015, -0.00000415, 0.00000986, -0.00001218, 0.00002275, -0.00001674, 0.00001301, -0.00000429, -0.00000738, 0.00000666, },
		{ -0.00001662, -0.00000017, -0.00000219, 0.00000230, -0.00001009, 0.00000361, -0.00000686, -0.00000285, 0.00000135, -0.00000846, -0.00000196, -0.00001130, -0.00001429, 0.00001310, -0.00001627, 0.00004058, -0.00002659, 0.00002557, -0.00000272, 0.00000835, 0.00000419, 0.00000411, 0.00000543, 0.00000834, 0.00002003, -0.00000739, 0.00000867, -0.00002789, 0.00001587, -0.00000964, 0.00000209, -0.00000100, -0.00000673, -0.00000123, -0.00000010, -0.00000184, 0.00000386, -0.00000392, 0.00000084, -0.00000293, 0.00000190, 0.00000271, -0.00000775, -0.00000149, 0.00000438, 0.00000108, -0.00000342, -0.00001253, 0.00002189, -0.00000415, -0.00000279, -0.00000860, 0.00001029, -0.00000842, 0.00000628, -0.00000404, 0.00000406, -0.00000066, -0.00000224, -0.00000564, 0.00000183, -0.00000436, -0.00000410, -0.00000759, 0.00000879, -0.00000828, -0.00000102, -0.00000735, 0.00000169, -0.00000547, -0.00002205, -0.00001772, -0.00000751, -0.00002949, -0.00011800, -0.00019035, 0.00040090, 0.00005418, 0.00035956, -0.00001171, -0.00021542, -0.00001295, -0.00003532, -0.00000147, -0.00002614, 0.00000088, -0.00000740, 0.00000198, -0.00002339, 0.00000437, -0.00000394, -0.00000076, -0.00001120, -0.00000304, 0.00000036, -0.00000185, -0.00000373, 0.00000553, -0.00000924, 0.00000705, -0.00001573, -0.00000159, 0.00000506, 0.00001407, -0.00002372, 0.00000113, 0.00000524, 0.00000591, -0.00000676, 0.00000120, 0.00000202, -0.00000028, -0.00000389, 0.00000258, -0.00000062, 0.00000210, -0.00000601, -0.00000031, -0.00000200, -0.00000398, 0.00000721, -0.00000597, 0.00000859, 0.00000524, -0.00001099, 0.00001813, -0.00002182, 0.00001440, },
		{ -0.00000370, -0.00001037, 0.00000079, 0.00000117, -0.00000581, -0.00000295, 0.00000011, -0.00000759, 0.00000015, -0.00000191, -0.00000955, 0.00000172, -0.00002632, 0.00001107, -0.00001152, 0.00001857, 0.00001233, -0.00001426, 0.00002967, -0.00001508, 0.00002292, -0.00001038, 0.00001882, -0.00000802, 0.00003395, -0.00001065, 0.00001508, -0.00002566, 0.00000226, 0.00000274, -0.00000636, 0.00000339, -0.00000562, -0.00000555, 0.00000301, -0.00000501, 0.00000536, -0.00000340, 0.00000112, -0.00000377, 0.00000142, 0.00000131, -0.00000019, -0.00001065, 0.00000901, -0.00000395, 0.00000734, -0.00002075, 0.00001665, 0.00000412, -0.00000127, -0.00001184, 0.00000804, -0.00000366, 0.00000245, -0.00000146, 0.00000120, 0.00000046, 0.00000154, -0.00000933, 0.00000388, -0.00000657, 0.00000282, -0.00001568, 0.00001047, -0.00000722, 0.00000307, -0.00001353, 0.00000681, -0.00001183, 0.00000077, -0.00004310, 0.00001517, -0.00004778, -0.00001870, -0.00027093, 0.00021830, 0.00022152, 0.00012961, 0.00030738, -0.00024753, -0.00006046, -0.00004422, 0.00000526, -0.00003353, 0.00000427, -0.00001936, 0.00001516, -0.00002677, 0.00000098, -0.00000750, 0.00000680, -0.00001523, 0.00000037, -0.00000924, 0.00000947, -0.00001397, 0.00001316, -0.00001468, 0.00001064, -0.00001389, -0.00000347, -0.00000411, 0.00002164, -0.00001552, -0.00000939, 0.00000387, 0.00000936, -0.00000508, -0.00000064, 0.00000009, 0.00000415, -0.00000686, 0.00000362, -0.00000270, 0.00000514, -0.00000613, -0.00000093, -0.00000262, -0.00000184, 0.00000013, 0.00000507, -0.00000734, 0.00002123, -0.00002023, 0.00001949, -0.00001417, 0.00000344, },
		{ 0.00001208, -0.00001978, 0.00000417, -0.00000439, 0.00000360, -0.00001198, 0.00000639, -0.00000984, -0.00000079, -0.00000234, -0.00000457, -0.00000494, -0.00000931, -0.00001868, 0.00001505, -0.00001683, 0.00004318, -0.00002635, 0.00002743, -0.00000570, 0.00001290, -0.00000011, 0.00001070, -0.00000176, 0.00002131, 0.00001040, 0.00000013, -0.00000271, -0.00002311, 0.00001688, -0.00001281, 0.00000541, -0.00000441, -0.00000585, -0.00000049, -0.00000073, -0.00000064, 0.00000289, -0.00000302, 0.00000001, -0.00000277, 0.00000241, 0.00000207, -0.00000948, 0.00000236, 0.00000106, 0.00000441, -0.00001015, -0.00000358, 0.00001832, -0.00000485, -0.00000499, -0.00000463, 0.00000784, -0.00000615, 0.00000613, -0.00000577, 0.00000445, -0.00000004, -0.00000418, -0.00000338, -0.00000021, -0.00000145, -0.00000732, -0.00000416, 0.00000518, -0.00000391, -0.00000472, -0.00000405, -0.00000311, -0.00000069, -0.00002937, -0.00000796, -0.00002199, -0.00001141, -0.00017869, -0.00006139, 0.00036278, 0.00001760, 0.00033597, -0.00000656, -0.00019908, -0.00001975, -0.00002648, -0.00000831, -0.00001814, -0.00000650, -0.00000170, -0.00000309, -0.00001858, 0.00000031, -0.00000024, -0.00000449, -0.00000765, -0.00000620, 0.00000368, -0.00000531, 0.00000259, -0.00000287, -0.00000233, 0.00000115, -0.00001355, -0.00000238, 0.00001128, 0.00000611, -0.00002221, 0.00000376, 0.00000616, 0.00000331, -0.00000619, 0.00000183, 0.00000230, -0.00000203, -0.00000216, 0.00000131, 0.00000102, 0.00000051, -0.00000579, -0.00000046, -0.00000227, -0.00000264, 0.00000774, -0.00000872, 0.00001571, -0.00000500, -0.00000117, 0.00000883, -0.00001571, },
		{ 0.00001136, -0.00001384, -0.00000316, -0.00000278, 0.00000426, -0.00000961, 0.00000117, -0.00000354, -0.00000474, -0.00000391, 0.00000105, -0.00001117, 0.00000291, -0.00002894, 0.00001057, -0.00001179, 0.00002148, 0.00001136, -0.00000969, 0.00002455, -0.00000939, 0.00001798, -0.00000394, 0.00001331, -0.00000046, 0.00003330, -0.00001168, 0.00001414, -0.00003307, 0.00001197, -0.00000484, -0.00000150, 0.00000052, -0.00000598, -0.00000443, 0.00000260, -0.00000429, 0.00000504, -0.00000307, 0.00000083, -0.00000375, 0.00000108, 0.00000278, -0.00000422, -0.00000598, 0.00000589, -0.00000072, 0.00000216, -0.00001705, 0.00001933, 0.00000059, -0.00000253, -0.00001013, 0.00000851, -0.00000410, 0.00000414, -0.00000272, -0.00000045, 0.00000331, -0.00000229, -0.00000565, 0.00000034, -0.00000211, -0.00000178, -0.00001079, 0.00000525, -0.00000138, -0.00000239, -0.00000712, -0.00000197, -0.00000211, -0.00001264, -0.00002648, -0.00000638, -0.00002059, -0.00007344, -0.00021223, 0.00027840, 0.00014101, 0.00013526, 0.00026861, -0.00020694, -0.00007847, -0.00002560, -0.00000924, -0.00001945, -0.00000710, -0.00001023, 0.00000657, -0.00001877, -0.00000596, -0.00000084, 0.00000013, -0.00000892, -0.00000562, -0.00000283, 0.00000264, -0.00000536, 0.00000539, -0.00001043, 0.00000808, -0.00001405, -0.00000462, 0.00000182, 0.00001802, -0.00001880, -0.00000610, 0.00000609, 0.00000694, -0.00000515, -0.00000065, 0.00000181, 0.00000146, -0.00000497, 0.00000233, -0.00000113, 0.00000400, -0.00000601, -0.00000169, -0.00000169, -0.00000307, 0.00000419, -0.00000077, 0.00000055, 0.00001457, -0.00001773, 0.00002001, -0.00001907, },
		{ -0.00000588, 0.00000448, -0.00001777, 0.00000502, -0.00000422, 0.00000252, -0.00001106, 0.00000625, -0.00001032, -0.00000098, -0.00000390, -0.00000231, -0.00000666, -0.00000870, -0.00002136, 0.00001456, -0.00001442, 0.00004294, -0.00002298, 0.00002590, -0.00000513, 0.00001391, -0.00000092, 0.00001400, -0.00000516, 0.00003161, 0.00000166, 0.00000687, -0.00001610, -0.00001325, 0.00001257, -0.00001184, 0.00000605, -0.00000673, -0.00000480, -0.00000004, -0.00000112, 0.00000028, 0.00000203, -0.00000217, -0.00000075, -0.00000277, 0.00000347, -0.00000016, -0.00000872, 0.00000390, -0.00000034, 0.00000537, -0.00001434, 0.00000482, 0.00001323, -0.00000483, -0.00000656, -0.00000094, 0.00000515, -0.00000338, 0.00000520, -0.00000742, 0.00000588, -0.00000120, -0.00000405, -0.00000314, 0.00000003, -0.00000124, -0.00000763, -0.00000367, 0.00000467, -0.00000288, -0.00000473, -0.00000528, -0.00000235, -0.00000308, -0.00002733, -0.00000969, -0.00002169, -0.00001673, -0.00020738, 0.00006005, 0.00029084, 0.00001373, 0.00029895, 0.00000875, -0.00019225, -0.00001767, -0.00002639, -0.00000739, -0.00001719, -0.00000771, -0.00000136, -0.00000350, -0.00001779, 0.00000012, -0.00000046, -0.00000446, -0.00000774, -0.00000552, 0.00000291, -0.00000430, 0.00000423, -0.00000762, 0.00000242, -0.00000389, -0.00001188, -0.00000108, 0.00001473, -0.00000148, -0.00001909, 0.00000604, 0.00000600, 0.00000120, -0.00000548, 0.00000280, 0.00000144, -0.00000254, -0.00000133, 0.00000089, 0.00000188, -0.00000069, -0.00000582, -0.00000012, -0.00000324, 0.00000030, 0.00000528, -0.00000737, 0.00001860, -0.00001247, 0.00000784, -0.00000188, },
		{ -0.00001766, 0.00001382, -0.00002067, 0.00000308, -0.00000597, 0.00000651, -0.00001228, 0.00000450, -0.00000701, -0.00000187, -0.00000809, 0.00000450, -0.00001322, 0.00000406, -0.00003110, 0.00000920, -0.00001138, 0.00002422, 0.00001058, -0.00000515, 0.00001913, -0.00000289, 0.00001173, 0.00000438, 0.00000566, 0.00001106, 0.00002740, -0.00000776, 0.00000613, -0.00003293, 0.00001678, -0.00001017, 0.00000287, -0.00000298, -0.00000547, -0.00000348, 0.00000197, -0.00000323, 0.00000419, -0.00000228, 0.00000022, -0.00000365, 0.00000113, 0.00000321, -0.00000727, -0.00000127, 0.00000209, 0.00000312, -0.00000431, -0.00001011, 0.00001836, -0.00000141, -0.00000427, -0.00000714, 0.00000747, -0.00000331, 0.00000497, -0.00000427, -0.00000098, 0.00000497, -0.00000524, -0.00000251, -0.00000256, 0.00000160, -0.00000580, -0.00000630, 0.00000048, 0.00000386, -0.00000703, -0.00000211, -0.00000891, 0.00000519, -0.00002324, -0.00001237, -0.00002548, 0.00000346, -0.00013229, -0.00011850, 0.00028478, 0.00008692, 0.00013686, 0.00023564, -0.00017237, -0.00009123, -0.00001167, -0.00002040, -0.00000797, -0.00001665, -0.00000238, -0.00000104, -0.00001163, -0.00001178, 0.00000455, -0.00000541, -0.00000370, -0.00001035, 0.00000235, -0.00000311, 0.00000280, -0.00000342, -0.00000409, 0.00000325, -0.00001281, -0.00000555, 0.00000806, 0.00001171, -0.00001928, -0.00000307, 0.00000769, 0.00000409, -0.00000461, -0.00000042, 0.00000281, -0.00000087, -0.00000298, 0.00000091, 0.00000063, 0.00000245, -0.00000569, -0.00000221, -0.00000132, -0.00000304, 0.00000658, -0.00000549, 0.00000895, 0.00000493, -0.00001049, 0.00001462, },
		{ -0.00001030, 0.00000398, -0.00000557, -0.00001212, 0.00000308, -0.00000227, -0.00000019, -0.00000803, 0.00000365, -0.00000851, -0.00000335, -0.00000320, -0.00000192, -0.00000701, -0.00000937, -0.00002261, 0.00001200, -0.00000946, 0.00004023, -0.00001681, 0.00002114, -0.00000085, 0.00001087, 0.00000266, 0.00001281, -0.00000293, 0.00003664, -0.00000318, 0.00000893, -0.00002679, -0.00000230, 0.00000537, -0.00000783, 0.00000418, -0.00000768, -0.00000385, 0.00000013, -0.00000114, 0.00000072, 0.00000152, -0.00000158, -0.00000136, -0.00000278, 0.00000462, -0.00000337, -0.00000579, 0.00000300, 0.00000051, 0.00000344, -0.00001478, 0.00001080, 0.00000815, -0.00000479, -0.00000696, 0.00000174, 0.00000294, -0.00000068, 0.00000364, -0.00000857, 0.00000762, -0.00000344, -0.00000241, -0.00000439, 0.00000196, -0.00000313, -0.00000559, -0.00000559, 0.00000664, -0.00000447, -0.00000205, -0.00000970, 0.00000232, -0.00001064, -0.00001810, -0.00002078, -0.00000868, -0.00004537, -0.00019448, 0.00015036, 0.00021041, 0.00002847, 0.00025471, 0.00002973, -0.00019092, -0.00001035, -0.00003195, -0.00000116, -0.00002136, -0.00000422, -0.00000518, -0.00000028, -0.00002005, 0.00000278, -0.00000357, -0.00000165, -0.00001044, -0.00000201, -0.00000099, 0.00000050, 0.00000125, -0.00000805, 0.00000394, -0.00000718, -0.00001099, 0.00000230, 0.00001469, -0.00000727, -0.00001525, 0.00000803, 0.00000475, -0.00000013, -0.00000473, 0.00000381, -0.00000019, -0.00000199, -0.00000138, 0.00000133, 0.00000185, -0.00000139, -0.00000602, 0.00000039, -0.00000432, 0.00000395, 0.00000068, -0.00000212, 0.00001626, -0.00001497, 0.00001291, },
		{ 0.00000546, -0.00001089, 0.00001030, -0.00002255, 0.00000655, -0.00000742, 0.00000686, -0.00001267, 0.00000563, -0.00000881, -0.00000043, -0.00001093, 0.00000707, -0.00001463, 0.00000425, -0.00003188, 0.00000597, -0.00000911, 0.00002542, 0.00001161, -0.00000256, 0.00001573, 0.00000181, 0.00000699, 0.00001166, -0.00000087, 0.00002312, 0.00001936, -0.00000206, -0.00000633, -0.00002597, 0.00001572, -0.00001164, 0.00000515, -0.00000590, -0.00000466, -0.00000263, 0.00000126, -0.00000205, 0.00000306, -0.00000122, -0.00000056, -0.00000356, 0.00000166, 0.00000224, -0.00000841, 0.00000209, -0.00000092, 0.00000597, -0.00001008, -0.00000181, 0.00001465, -0.00000207, -0.00000608, -0.00000353, 0.00000543, -0.00000157, 0.00000491, -0.00000584, -0.00000040, 0.00000503, -0.00000660, -0.00000076, -0.00000387, 0.00000345, -0.00000792, -0.00000365, -0.00000233, 0.00000692, -0.00000918, -0.00000034, -0.00001178, 0.00000729, -0.00002725, -0.00000594, -0.00003505, 0.00001325, -0.00017645, -0.00001211, 0.00024885, 0.00006249, 0.00012756, 0.00021309, -0.00014719, -0.00009580, -0.00000540, -0.00002516, -0.00000208, -0.00002156, 0.00000162, -0.00000534, -0.00000741, -0.00001467, 0.00000696, -0.00000815, -0.00000120, -0.00001214, 0.00000449, -0.00000568, 0.00000796, -0.00001056, 0.00000200, -0.00000224, -0.00001110, -0.00000528, 0.00001288, 0.00000440, -0.00001758, -0.00000034, 0.00000835, 0.00000134, -0.00000369, 0.00000006, 0.00000286, -0.00000237, -0.00000144, -0.00000009, 0.00000204, 0.00000092, -0.00000540, -0.00000232, -0.00000167, -0.00000146, 0.00000639, -0.00000721, 0.00001511, -0.00000461, -0.00000128, },
		{ 0.00001254, -0.00001378, 0.00001046, -0.00001472, -0.00000531, -0.00000008, 0.00000015, -0.00000310, -0.00000441, -0.00000007, -0.00000554, -0.00000690, -0.00000103, -0.00000284, -0.00000656, -0.00001059, -0.00002333, 0.00000849, -0.00000322, 0.00003642, -0.00000934, 0.00001471, 0.00000570, 0.00000501, 0.00000965, 0.00000793, 0.00000461, 0.00003595, -0.00000332, 0.00000445, -0.00003127, 0.00000586, -0.00000163, -0.00000281, 0.00000084, -0.00000750, -0.00000303, 0.00000003, -0.00000077, 0.00000060, 0.00000149, -0.00000132, -0.00000183, -0.00000260, 0.00000526, -0.00000652, -0.00000173, 0.00000040, 0.00000299, -0.00000080, -0.00001144, 0.00001339, 0.00000415, -0.00000516, -0.00000600, 0.00000303, 0.00000165, 0.00000150, 0.00000166, -0.00000886, 0.00000885, -0.00000577, -0.00000026, -0.00000608, 0.00000437, -0.00000578, -0.00000256, -0.00000850, 0.00000959, -0.00000698, 0.00000127, -0.00001477, 0.00000771, -0.00001933, -0.00000661, -0.00003527, 0.00000931, -0.00008832, -0.00014275, 0.00019724, 0.00014168, 0.00004714, 0.00021165, 0.00005007, -0.00018959, -0.00000294, -0.00003847, 0.00000628, -0.00002714, 0.00000098, -0.00001055, 0.00000428, -0.00002325, 0.00000623, -0.00000755, 0.00000197, -0.00001379, 0.00000228, -0.00000590, 0.00000698, -0.00000468, -0.00000499, 0.00000240, -0.00000865, -0.00001045, 0.00000687, 0.00001146, -0.00001056, -0.00001130, 0.00000958, 0.00000264, -0.00000062, -0.00000395, 0.00000451, -0.00000202, -0.00000083, -0.00000195, 0.00000231, 0.00000115, -0.00000171, -0.00000616, 0.00000067, -0.00000477, 0.00000701, -0.00000434, 0.00000543, 0.00000947, -0.00001213, },
		{ 0.00000662, -0.00000419, -0.00000152, 0.00000234, -0.00001970, 0.00000678, -0.00000683, 0.00000524, -0.00001065, 0.00000418, -0.00000837, -0.00000111, -0.00001163, 0.00000796, -0.00001478, 0.00000308, -0.00003100, 0.00000065, -0.00000466, 0.00002458, 0.00001507, -0.00000282, 0.00001559, 0.00000309, 0.00000578, 0.00001560, -0.00000349, 0.00003244, 0.00001248, 0.00000152, -0.00001894, -0.00001521, 0.00001003, -0.00000929, 0.00000477, -0.00000756, -0.00000392, -0.00000189, 0.00000064, -0.00000103, 0.00000196, -0.00000014, -0.00000138, -0.00000345, 0.00000248, 0.00000003, -0.00000734, 0.00000332, -0.00000218, 0.00000670, -0.00001345, 0.00000578, 0.00000962, -0.00000197, -0.00000740, -0.00000005, 0.00000303, 0.00000068, 0.00000401, -0.00000710, 0.00000086, 0.00000367, -0.00000629, -0.00000054, -0.00000348, 0.00000324, -0.00000781, -0.00000326, -0.00000276, 0.00000741, -0.00000858, -0.00000198, -0.00001041, 0.00000392, -0.00002403, -0.00000856, -0.00003266, 0.00000317, -0.00019146, 0.00008355, 0.00018974, 0.00006116, 0.00010811, 0.00020028, -0.00013058, -0.00009299, -0.00000630, -0.00002355, -0.00000206, -0.00002135, 0.00000124, -0.00000581, -0.00000651, -0.00001435, 0.00000614, -0.00000783, -0.00000165, -0.00001079, 0.00000325, -0.00000441, 0.00000889, -0.00001425, 0.00000604, -0.00000705, -0.00000955, -0.00000330, 0.00001507, -0.00000231, -0.00001452, 0.00000214, 0.00000785, -0.00000080, -0.00000262, 0.00000067, 0.00000206, -0.00000288, -0.00000062, -0.00000040, 0.00000278, -0.00000031, -0.00000523, -0.00000212, -0.00000247, 0.00000133, 0.00000365, -0.00000510, 0.00001703, -0.00001118, },
		{ -0.00000540, 0.00000723, -0.00001161, 0.00001160, -0.00002030, 0.00000029, -0.00000279, 0.00000172, -0.00000480, -0.00000180, -0.00000332, -0.00000290, -0.00001020, 0.00000135, -0.00000406, -0.00000623, -0.00001138, -0.00002463, 0.00000536, 0.00000272, 0.00003327, -0.00000257, 0.00000885, 0.00001216, -0.00000124, 0.00001771, 0.00000174, 0.00001525, 0.00003118, -0.00000039, -0.00000566, -0.00002846, 0.00000880, -0.00000592, 0.00000122, -0.00000264, -0.00000677, -0.00000225, -0.00000034, -0.00000006, 0.00000000, 0.00000189, -0.00000139, -0.00000218, -0.00000209, 0.00000489, -0.00000859, 0.00000213, -0.00000265, 0.00000589, -0.00000598, -0.00000536, 0.00001263, 0.00000165, -0.00000598, -0.00000387, 0.00000298, 0.00000139, 0.00000289, -0.00000049, -0.00000815, 0.00000896, -0.00000729, 0.00000144, -0.00000722, 0.00000607, -0.00000783, -0.00000003, -0.00001086, 0.00001188, -0.00000865, 0.00000318, -0.00001798, 0.00001070, -0.00002505, 0.00000179, -0.00004624, 0.00002227, -0.00013096, -0.00006554, 0.00020121, 0.00009549, 0.00005940, 0.00017627, 0.00006477, -0.00018391, 0.00000038, -0.00004191, 0.00001119, -0.00003113, 0.00000487, -0.00001481, 0.00000784, -0.00002529, 0.00000841, -0.00001042, 0.00000447, -0.00001583, 0.00000523, -0.00000951, 0.00001260, -0.00001114, -0.00000025, -0.00000120, -0.00000881, -0.00000951, 0.00001123, 0.00000619, -0.00001146, -0.00000750, 0.00001041, 0.00000016, -0.00000038, -0.00000316, 0.00000463, -0.00000354, 0.00000039, -0.00000256, 0.00000339, 0.00000009, -0.00000186, -0.00000604, 0.00000041, -0.00000402, 0.00000828, -0.00000780, 0.00001283, 0.00000046, },
		{ -0.00001308, 0.00001068, -0.00001111, 0.00000704, -0.00000719, -0.00001378, 0.00000458, -0.00000483, 0.00000247, -0.00000710, 0.00000080, -0.00000611, -0.00000357, -0.00001033, 0.00000716, -0.00001375, 0.00000082, -0.00002889, -0.00000618, 0.00000129, 0.00002234, 0.00002035, -0.00000541, 0.00001847, 0.00000089, 0.00000852, 0.00001547, -0.00000091, 0.00003714, 0.00000874, 0.00000005, -0.00002745, -0.00000467, 0.00000243, -0.00000459, 0.00000228, -0.00000791, -0.00000335, -0.00000132, 0.00000027, -0.00000038, 0.00000113, 0.00000076, -0.00000212, -0.00000324, 0.00000315, -0.00000279, -0.00000451, 0.00000244, -0.00000149, 0.00000495, -0.00001346, 0.00001104, 0.00000466, -0.00000178, -0.00000776, 0.00000265, 0.00000090, 0.00000294, 0.00000247, -0.00000769, 0.00000216, 0.00000156, -0.00000483, -0.00000136, -0.00000195, 0.00000159, -0.00000607, -0.00000454, -0.00000143, 0.00000609, -0.00000625, -0.00000569, -0.00000650, -0.00000279, -0.00001601, -0.00001757, -0.00002139, -0.00002471, -0.00017178, 0.00015081, 0.00012733, 0.00007160, 0.00008393, 0.00019302, -0.00011884, -0.00008628, -0.00001134, -0.00001806, -0.00000592, -0.00001760, -0.00000221, -0.00000358, -0.00000789, -0.00001192, 0.00000316, -0.00000548, -0.00000396, -0.00000739, -0.00000029, -0.00000023, 0.00000597, -0.00001413, 0.00000718, -0.00001042, -0.00000828, 0.00000008, 0.00001427, -0.00000728, -0.00001084, 0.00000433, 0.00000621, -0.00000204, -0.00000158, 0.00000121, 0.00000074, -0.00000256, -0.00000044, -0.00000006, 0.00000279, -0.00000118, -0.00000508, -0.00000189, -0.00000317, 0.00000440, -0.00000062, 0.00000030, 0.00001417, },
		{ -0.00000794, 0.00000254, -0.00000077, -0.00000515, 0.00000741, -0.00002117, 0.00000322, -0.00000401, 0.00000173, -0.00000451, -0.00000121, -0.00000496, -0.00000174, -0.00001206, 0.00000281, -0.00000469, -0.00000673, -0.00001105, -0.00002722, 0.00000351, 0.00000723, 0.00003208, 0.00000195, 0.00000547, 0.00001630, -0.00000538, 0.00002418, -0.00000282, 0.00002598, 0.00002506, 0.00000240, -0.00001816, -0.00001997, 0.00000634, -0.00000644, 0.00000299, -0.00000525, -0.00000598, -0.00000143, -0.00000087, 0.00000080, -0.00000089, 0.00000258, -0.00000168, -0.00000242, -0.00000130, 0.00000334, -0.00000887, 0.00000467, -0.00000492, 0.00000786, -0.00001039, 0.00000174, 0.00000942, 0.00000048, -0.00000695, -0.00000107, 0.00000193, 0.00000199, 0.00000340, -0.00000246, -0.00000662, 0.00000774, -0.00000751, 0.00000213, -0.00000721, 0.00000638, -0.00000843, 0.00000107, -0.00001167, 0.00001249, -0.00000847, 0.00000261, -0.00001801, 0.00000961, -0.00002545, 0.00000374, -0.00004888, 0.00002213, -0.00015831, 0.00001813, 0.00017308, 0.00007286, 0.00006128, 0.00015106, 0.00007200, -0.00017234, -0.00000205, -0.00004038, 0.00001162, -0.00003142, 0.00000570, -0.00001637, 0.00000901, -0.00002495, 0.00000818, -0.00001103, 0.00000479, -0.00001545, 0.00000561, -0.00001030, 0.00001538, -0.00001589, 0.00000417, -0.00000552, -0.00000828, -0.00000755, 0.00001404, 0.00000037, -0.00001053, -0.00000388, 0.00001020, -0.00000213, 0.00000030, -0.00000237, 0.00000409, -0.00000439, 0.00000128, -0.00000284, 0.00000415, -0.00000097, -0.00000196, -0.00000562, -0.00000039, -0.00000198, 0.00000717, -0.00000824, 0.00001766, },
		{ 0.00000755, -0.00001034, 0.00000995, -0.00001333, 0.00001189, -0.00001526, -0.00000714, 0.00000139, -0.00000256, -0.00000017, -0.00000343, -0.00000315, -0.00000322, -0.00000676, -0.00000790, 0.00000533, -0.00001220, -0.00000175, -0.00002649, -0.00001334, 0.00000743, 0.00002010, 0.00002595, -0.00000871, 0.00002275, -0.00000330, 0.00001398, 0.00001234, 0.00000622, 0.00003727, 0.00000800, -0.00000711, -0.00002931, 0.00000224, -0.00000416, 0.00000042, -0.00000114, -0.00000741, -0.00000286, -0.00000103, 0.00000025, -0.00000023, 0.00000073, 0.00000134, -0.00000274, -0.00000281, 0.00000322, -0.00000532, -0.00000088, 0.00000017, 0.00000055, 0.00000132, -0.00001021, 0.00001319, 0.00000074, -0.00000191, -0.00000693, 0.00000422, -0.00000058, 0.00000481, 0.00000054, -0.00000740, 0.00000282, -0.00000049, -0.00000301, -0.00000245, -0.00000022, -0.00000045, -0.00000380, -0.00000636, 0.00000043, 0.00000431, -0.00000382, -0.00000941, -0.00000261, -0.00000951, -0.00000739, -0.00002778, -0.00000829, -0.00006144, -0.00012198, 0.00018181, 0.00007654, 0.00008286, 0.00006143, 0.00018633, -0.00010764, -0.00007979, -0.00001666, -0.00001209, -0.00001068, -0.00001286, -0.00000652, -0.00000059, -0.00000978, -0.00000910, -0.00000028, -0.00000273, -0.00000647, -0.00000374, -0.00000423, 0.00000492, 0.00000081, -0.00001108, 0.00000560, -0.00001225, -0.00000693, 0.00000394, 0.00001102, -0.00001012, -0.00000699, 0.00000600, 0.00000376, -0.00000231, -0.00000069, 0.00000148, -0.00000070, -0.00000178, -0.00000062, 0.00000067, 0.00000223, -0.00000175, -0.00000480, -0.00000195, -0.00000316, 0.00000657, -0.00000475, 0.00000730, },
		{ 0.00001837, -0.00001331, 0.00000886, -0.00000848, 0.00000286, -0.00000034, -0.00001784, 0.00000328, -0.00000357, 0.00000028, -0.00000240, -0.00000270, -0.00000471, -0.00000242, -0.00001200, 0.00000283, -0.00000436, -0.00000824, -0.00000955, -0.00003113, 0.00000301, 0.00001011, 0.00003315, 0.00000373, 0.00000534, 0.00001702, -0.00000590, 0.00002728, -0.00000349, 0.00003422, 0.00001999, 0.00000185, -0.00002878, -0.00000923, 0.00000025, -0.00000374, 0.00000237, -0.00000655, -0.00000539, -0.00000063, -0.00000140, 0.00000157, -0.00000178, 0.00000333, -0.00000209, -0.00000251, -0.00000044, 0.00000088, -0.00000726, 0.00000528, -0.00000564, 0.00000799, -0.00001261, 0.00000805, 0.00000504, 0.00000009, -0.00000760, 0.00000179, 0.00000036, 0.00000313, 0.00000308, -0.00000394, -0.00000472, 0.00000548, -0.00000646, 0.00000176, -0.00000609, 0.00000529, -0.00000749, 0.00000060, -0.00001081, 0.00001135, -0.00000652, -0.00000025, -0.00001516, 0.00000482, -0.00002076, -0.00000090, -0.00004256, 0.00000600, -0.00016012, 0.00008973, 0.00012881, 0.00006762, 0.00005441, 0.00013460, 0.00007312, -0.00015623, -0.00000922, -0.00003447, 0.00000784, -0.00002806, 0.00000339, -0.00001512, 0.00000774, -0.00002229, 0.00000561, -0.00000945, 0.00000302, -0.00001278, 0.00000346, -0.00000806, 0.00001458, -0.00001761, 0.00000679, -0.00000945, -0.00000742, -0.00000448, 0.00001454, -0.00000472, -0.00000847, -0.00000044, 0.00000876, -0.00000369, 0.00000112, -0.00000165, 0.00000299, -0.00000450, 0.00000168, -0.00000260, 0.00000436, -0.00000185, -0.00000206, -0.00000502, -0.00000142, 0.00000085, 0.00000401, -0.00000522, },
		{ 0.00001359, -0.00000063, -0.00000451, 0.00000528, -0.00001069, 0.00001168, -0.00001969, -0.00000183, -0.00000137, -0.00000105, -0.00000162, -0.00000092, -0.00000634, -0.00000099, -0.00000942, -0.00000548, 0.00000337, -0.00001092, -0.00000379, -0.00002476, -0.00001968, 0.00001238, 0.00001932, 0.00003016, -0.00001076, 0.00002629, -0.00000716, 0.00001990, 0.00000850, 0.00001582, 0.00003446, 0.00000815, -0.00001801, -0.00002452, 0.00000389, -0.00000774, 0.00000399, -0.00000425, -0.00000664, -0.00000228, -0.00000105, 0.00000057, -0.00000054, 0.00000077, 0.00000157, -0.00000317, -0.00000215, 0.00000236, -0.00000672, 0.00000240, -0.00000241, 0.00000288, -0.00000297, -0.00000471, 0.00001233, -0.00000175, -0.00000246, -0.00000504, 0.00000458, -0.00000126, 0.00000602, -0.00000149, -0.00000623, 0.00000244, -0.00000180, -0.00000152, -0.00000312, 0.00000090, -0.00000192, -0.00000205, -0.00000761, 0.00000167, 0.00000328, -0.00000269, -0.00001140, -0.00000092, -0.00001334, -0.00000207, -0.00003394, -0.00000133, -0.00009449, -0.00005458, 0.00017897, 0.00004447, 0.00008788, 0.00004504, 0.00017676, -0.00009405, -0.00007643, -0.00001933, -0.00000843, -0.00001378, -0.00000940, -0.00000966, 0.00000142, -0.00001060, -0.00000743, -0.00000269, -0.00000102, -0.00000771, -0.00000143, -0.00000677, 0.00000896, -0.00000448, -0.00000672, 0.00000216, -0.00001281, -0.00000506, 0.00000713, 0.00000641, -0.00001103, -0.00000310, 0.00000679, 0.00000105, -0.00000179, 0.00000000, 0.00000136, -0.00000188, -0.00000093, -0.00000081, 0.00000143, 0.00000138, -0.00000214, -0.00000430, -0.00000246, -0.00000207, 0.00000695, -0.00000703, },
		{ 0.00000048, 0.00001479, -0.00001442, 0.00001162, -0.00001325, 0.00000936, -0.00000888, -0.00001201, 0.00000127, -0.00000210, -0.00000181, 0.00000060, -0.00000546, -0.00000322, -0.00000444, -0.00001035, 0.00000158, -0.00000331, -0.00001037, -0.00000741, -0.00003564, 0.00000311, 0.00001207, 0.00003573, 0.00000355, 0.00000783, 0.00001471, -0.00000279, 0.00002671, 0.00000046, 0.00003879, 0.00001683, -0.00000373, -0.00003403, -0.00000008, -0.00000668, 0.00000050, 0.00000014, -0.00000676, -0.00000496, -0.00000000, -0.00000171, 0.00000201, -0.00000243, 0.00000391, -0.00000246, -0.00000242, 0.00000011, -0.00000189, -0.00000425, 0.00000409, -0.00000471, 0.00000608, -0.00001193, 0.00001222, 0.00000075, -0.00000008, -0.00000750, 0.00000413, -0.00000123, 0.00000442, 0.00000208, -0.00000462, -0.00000306, 0.00000277, -0.00000459, 0.00000074, -0.00000435, 0.00000338, -0.00000559, -0.00000088, -0.00000891, 0.00000919, -0.00000375, -0.00000412, -0.00001100, -0.00000164, -0.00001344, -0.00000926, -0.00003092, -0.00002261, -0.00013402, 0.00013644, 0.00008388, 0.00007069, 0.00004347, 0.00012321, 0.00007131, -0.00013864, -0.00001834, -0.00002649, 0.00000176, -0.00002256, -0.00000077, -0.00001215, 0.00000507, -0.00001837, 0.00000172, -0.00000666, 0.00000021, -0.00000893, -0.00000012, -0.00000381, 0.00001080, -0.00001617, 0.00000698, -0.00001236, -0.00000624, -0.00000080, 0.00001275, -0.00000833, -0.00000577, 0.00000265, 0.00000620, -0.00000423, 0.00000181, -0.00000105, 0.00000156, -0.00000402, 0.00000164, -0.00000191, 0.00000400, -0.00000247, -0.00000207, -0.00000452, -0.00000216, 0.00000356, -0.00000008, },
		{ -0.00000637, 0.00001713, -0.00000785, 0.00000197, -0.00000116, -0.00000480, 0.00000687, -0.00001974, 0.00000101, -0.00000288, -0.00000075, -0.00000145, -0.00000022, -0.00000798, -0.00000024, -0.00001064, -0.00000395, 0.00000196, -0.00001044, -0.00000482, -0.00002421, -0.00002456, 0.00001534, 0.00002092, 0.00003191, -0.00001018, 0.00002736, -0.00000864, 0.00002405, 0.00000642, 0.00002543, 0.00003078, 0.00000632, -0.00002899, -0.00001547, 0.00000084, -0.00000783, 0.00000528, -0.00000628, -0.00000601, -0.00000157, -0.00000132, 0.00000107, -0.00000114, 0.00000110, 0.00000151, -0.00000340, -0.00000137, 0.00000054, -0.00000650, 0.00000439, -0.00000424, 0.00000439, -0.00000648, 0.00000153, 0.00000931, -0.00000296, -0.00000321, -0.00000244, 0.00000395, -0.00000121, 0.00000647, -0.00000327, -0.00000446, 0.00000097, -0.00000203, -0.00000077, -0.00000302, 0.00000102, -0.00000231, -0.00000138, -0.00000771, 0.00000171, 0.00000355, -0.00000342, -0.00001101, -0.00000225, -0.00001304, -0.00000194, -0.00003320, -0.00000597, -0.00011232, 0.00001447, 0.00015233, 0.00003057, 0.00008466, 0.00003591, 0.00016350, -0.00007749, -0.00007692, -0.00001836, -0.00000820, -0.00001405, -0.00000835, -0.00001064, 0.00000158, -0.00000963, -0.00000762, -0.00000339, -0.00000096, -0.00000708, -0.00000121, -0.00000691, 0.00001044, -0.00000814, -0.00000269, -0.00000207, -0.00001241, -0.00000244, 0.00000874, 0.00000168, -0.00001052, 0.00000081, 0.00000635, -0.00000128, -0.00000085, 0.00000050, 0.00000082, -0.00000256, -0.00000028, -0.00000074, 0.00000191, 0.00000047, -0.00000242, -0.00000359, -0.00000332, -0.00000000, 0.00000526, },
		{ -0.00000355, 0.00000712, 0.00000812, -0.00001158, 0.00001033, -0.00001374, 0.00001206, -0.00001556, -0.00000580, -0.00000153, -0.00000051, -0.00000352, 0.00000328, -0.00000828, -0.00000157, -0.00000676, -0.00000798, -0.00000028, -0.00000217, -0.00001238, -0.00000551, -0.00003972, 0.00000264, 0.00001431, 0.00003845, 0.00000296, 0.00001130, 0.00001099, 0.00000253, 0.00002373, 0.00000821, 0.00004010, 0.00001458, -0.00001366, -0.00003251, 0.00000477, -0.00001187, 0.00000435, -0.00000246, -0.00000641, -0.00000447, 0.00000029, -0.00000167, 0.00000197, -0.00000265, 0.00000418, -0.00000270, -0.00000213, 0.00000001, -0.00000419, -0.00000073, 0.00000178, -0.00000267, 0.00000269, -0.00000855, 0.00001362, -0.00000261, -0.00000047, -0.00000643, 0.00000561, -0.00000251, 0.00000555, 0.00000065, -0.00000440, -0.00000215, 0.00000029, -0.00000251, -0.00000038, -0.00000265, 0.00000143, -0.00000356, -0.00000250, -0.00000688, 0.00000703, -0.00000138, -0.00000746, -0.00000749, -0.00000720, -0.00000683, -0.00001700, -0.00002015, -0.00005516, -0.00008576, 0.00015384, 0.00004906, 0.00007398, 0.00003327, 0.00011315, 0.00006972, -0.00012264, -0.00002645, -0.00001917, -0.00000423, -0.00001701, -0.00000498, -0.00000904, 0.00000241, -0.00001459, -0.00000214, -0.00000393, -0.00000231, -0.00000535, -0.00000353, 0.00000075, 0.00000552, -0.00001246, 0.00000488, -0.00001405, -0.00000455, 0.00000261, 0.00000934, -0.00001032, -0.00000269, 0.00000501, 0.00000296, -0.00000376, 0.00000219, -0.00000066, 0.00000010, -0.00000319, 0.00000140, -0.00000100, 0.00000320, -0.00000287, -0.00000190, -0.00000433, -0.00000216, 0.00000518, },
		{ 0.00000203, -0.00000269, 0.00001680, -0.00001214, 0.00000674, -0.00000676, 0.00000175, -0.00000066, -0.00001612, 0.00000131, -0.00000310, -0.00000142, -0.00000003, -0.00000110, -0.00000809, -0.00000105, -0.00001022, -0.00000360, 0.00000128, -0.00001076, -0.00000492, -0.00002467, -0.00002809, 0.00001636, 0.00002483, 0.00003120, -0.00000675, 0.00002544, -0.00000675, 0.00002519, 0.00000773, 0.00003329, 0.00002748, 0.00000031, -0.00003617, -0.00000581, -0.00000475, -0.00000541, 0.00000453, -0.00000708, -0.00000561, -0.00000085, -0.00000163, 0.00000151, -0.00000176, 0.00000151, 0.00000130, -0.00000341, -0.00000069, -0.00000192, -0.00000468, 0.00000469, -0.00000473, 0.00000437, -0.00000808, 0.00000696, 0.00000527, -0.00000337, -0.00000376, 0.00000037, 0.00000265, -0.00000062, 0.00000617, -0.00000447, -0.00000255, -0.00000129, -0.00000126, -0.00000073, -0.00000225, 0.00000027, -0.00000169, -0.00000175, -0.00000673, 0.00000066, 0.00000483, -0.00000555, -0.00000880, -0.00000588, -0.00000942, -0.00000626, -0.00002625, -0.00002289, -0.00010821, 0.00007084, 0.00011500, 0.00002931, 0.00007514, 0.00003242, 0.00014799, -0.00005948, -0.00007998, -0.00001467, -0.00001077, -0.00001186, -0.00000944, -0.00000963, 0.00000010, -0.00000715, -0.00000939, -0.00000261, -0.00000223, -0.00000491, -0.00000278, -0.00000477, 0.00000903, -0.00000922, -0.00000019, -0.00000622, -0.00001119, 0.00000064, 0.00000845, -0.00000229, -0.00000907, 0.00000461, 0.00000457, -0.00000268, 0.00000011, 0.00000084, -0.00000006, -0.00000272, 0.00000005, -0.00000033, 0.00000195, -0.00000032, -0.00000260, -0.00000284, -0.00000421, 0.00000246, },
		{ 0.00000507, -0.00000500, 0.00001263, 0.00000044, -0.00000638, 0.00000605, -0.00001027, 0.00001017, -0.00001867, -0.00000096, -0.00000396, 0.00000049, -0.00000416, 0.00000472, -0.00001014, -0.00000077, -0.00000835, -0.00000584, -0.00000199, -0.00000157, -0.00001363, -0.00000456, -0.00004249, 0.00000060, 0.00001790, 0.00004004, 0.00000352, 0.00001391, 0.00000794, 0.00000805, 0.00002035, 0.00001802, 0.00003946, 0.00001094, -0.00002509, -0.00002534, 0.00000460, -0.00001392, 0.00000645, -0.00000441, -0.00000603, -0.00000378, 0.00000018, -0.00000128, 0.00000149, -0.00000245, 0.00000408, -0.00000272, -0.00000168, -0.00000092, -0.00000535, 0.00000233, -0.00000071, -0.00000043, -0.00000113, -0.00000339, 0.00001242, -0.00000468, -0.00000118, -0.00000441, 0.00000608, -0.00000330, 0.00000628, -0.00000090, -0.00000336, -0.00000227, -0.00000141, -0.00000073, -0.00000116, -0.00000152, 0.00000007, -0.00000209, -0.00000352, -0.00000550, 0.00000563, -0.00000030, -0.00000916, -0.00000600, -0.00001001, -0.00000353, -0.00002050, -0.00001627, -0.00008130, -0.00002675, 0.00014557, 0.00002867, 0.00007283, 0.00002668, 0.00010220, 0.00007011, -0.00011004, -0.00003163, -0.00001435, -0.00000834, -0.00001298, -0.00000786, -0.00000699, 0.00000084, -0.00001201, -0.00000494, -0.00000222, -0.00000358, -0.00000315, -0.00000546, 0.00000399, 0.00000049, -0.00000784, 0.00000117, -0.00001451, -0.00000222, 0.00000488, 0.00000534, -0.00001099, 0.00000077, 0.00000618, -0.00000028, -0.00000256, 0.00000224, -0.00000053, -0.00000114, -0.00000231, 0.00000120, -0.00000015, 0.00000217, -0.00000311, -0.00000152, -0.00000450, -0.00000124, },
		{ 0.00000448, -0.00000159, 0.00000303, 0.00001274, -0.00001271, 0.00000841, -0.00000948, 0.00000636, -0.00000836, -0.00001050, -0.00000017, -0.00000251, -0.00000237, 0.00000177, -0.00000275, -0.00000737, -0.00000281, -0.00000860, -0.00000414, 0.00000102, -0.00001144, -0.00000466, -0.00002546, -0.00003101, 0.00001617, 0.00003016, 0.00002902, -0.00000142, 0.00002129, -0.00000191, 0.00002344, 0.00001274, 0.00003888, 0.00002426, -0.00001010, -0.00003710, 0.00000111, -0.00001018, -0.00000221, 0.00000271, -0.00000704, -0.00000527, -0.00000034, -0.00000174, 0.00000167, -0.00000215, 0.00000178, 0.00000110, -0.00000319, -0.00000041, -0.00000443, -0.00000177, 0.00000345, -0.00000382, 0.00000272, -0.00000728, 0.00001047, 0.00000130, -0.00000349, -0.00000375, 0.00000292, 0.00000104, 0.00000026, 0.00000523, -0.00000487, -0.00000103, -0.00000380, 0.00000016, -0.00000113, -0.00000120, -0.00000089, -0.00000057, -0.00000266, -0.00000522, -0.00000083, 0.00000631, -0.00000800, -0.00000611, -0.00001002, -0.00000474, -0.00001220, -0.00001694, -0.00004768, -0.00008222, 0.00010571, 0.00007877, 0.00003359, 0.00006309, 0.00003169, 0.00013262, -0.00004244, -0.00008330, -0.00001027, -0.00001445, -0.00000859, -0.00001152, -0.00000764, -0.00000209, -0.00000409, -0.00001185, -0.00000121, -0.00000392, -0.00000221, -0.00000509, -0.00000140, 0.00000545, -0.00000780, 0.00000029, -0.00000967, -0.00000914, 0.00000355, 0.00000661, -0.00000513, -0.00000692, 0.00000797, 0.00000170, -0.00000294, 0.00000077, 0.00000107, -0.00000114, -0.00000244, 0.00000014, 0.00000035, 0.00000155, -0.00000094, -0.00000260, -0.00000227, -0.00000468, },
		{ 0.00000036, 0.00000324, -0.00000370, 0.00001534, -0.00000607, -0.00000093, 0.00000082, -0.00000456, 0.00000448, -0.00001784, 0.00000163, -0.00000541, 0.00000067, -0.00000360, 0.00000466, -0.00001061, -0.00000134, -0.00000863, -0.00000455, -0.00000312, -0.00000179, -0.00001386, -0.00000481, -0.00004364, -0.00000339, 0.00002324, 0.00003997, 0.00000605, 0.00001451, 0.00000711, 0.00001203, 0.00001849, 0.00002816, 0.00003798, 0.00000368, -0.00003420, -0.00001557, 0.00000064, -0.00001296, 0.00000645, -0.00000523, -0.00000588, -0.00000290, -0.00000022, -0.00000068, 0.00000074, -0.00000197, 0.00000369, -0.00000251, -0.00000121, -0.00000260, -0.00000507, 0.00000417, -0.00000251, 0.00000111, -0.00000421, 0.00000225, 0.00000936, -0.00000559, -0.00000207, -0.00000170, 0.00000558, -0.00000355, 0.00000648, -0.00000226, -0.00000179, -0.00000342, -0.00000210, 0.00000045, -0.00000140, -0.00000113, -0.00000046, -0.00000148, -0.00000365, -0.00000502, 0.00000519, -0.00000064, -0.00000904, -0.00000672, -0.00000971, -0.00000425, -0.00001849, -0.00002245, -0.00009255, 0.00003000, 0.00012061, 0.00002132, 0.00006640, 0.00002412, 0.00009019, 0.00007234, -0.00010082, -0.00003363, -0.00001247, -0.00001003, -0.00001102, -0.00000888, -0.00000643, 0.00000071, -0.00001096, -0.00000635, -0.00000176, -0.00000337, -0.00000264, -0.00000540, 0.00000498, -0.00000295, -0.00000363, -0.00000336, -0.00001379, 0.00000055, 0.00000558, 0.00000161, -0.00001070, 0.00000453, 0.00000583, -0.00000281, -0.00000110, 0.00000204, -0.00000067, -0.00000200, -0.00000155, 0.00000119, 0.00000044, 0.00000110, -0.00000322, -0.00000100, -0.00000489, },
		{ -0.00000443, 0.00000530, -0.00000421, 0.00000910, 0.00000625, -0.00001006, 0.00000691, -0.00000859, 0.00000732, -0.00001400, -0.00000473, -0.00000239, -0.00000176, -0.00000292, 0.00000303, -0.00000418, -0.00000674, -0.00000462, -0.00000657, -0.00000500, 0.00000066, -0.00001188, -0.00000474, -0.00002576, -0.00003423, 0.00001577, 0.00003571, 0.00002681, 0.00000424, 0.00001658, 0.00000447, 0.00002004, 0.00002064, 0.00004270, 0.00001939, -0.00002287, -0.00003170, 0.00000345, -0.00001341, 0.00000010, 0.00000094, -0.00000672, -0.00000475, -0.00000024, -0.00000152, 0.00000140, -0.00000216, 0.00000176, 0.00000101, -0.00000279, -0.00000077, -0.00000636, 0.00000142, 0.00000131, -0.00000201, -0.00000007, -0.00000429, 0.00001160, -0.00000186, -0.00000370, -0.00000294, 0.00000484, -0.00000060, 0.00000123, 0.00000382, -0.00000436, -0.00000034, -0.00000595, 0.00000173, -0.00000159, -0.00000030, -0.00000190, 0.00000047, -0.00000347, -0.00000386, -0.00000205, 0.00000713, -0.00000968, -0.00000432, -0.00001285, -0.00000148, -0.00001640, -0.00001046, -0.00007255, -0.00004072, 0.00011729, 0.00005108, 0.00003768, 0.00005185, 0.00003121, 0.00011939, -0.00002839, -0.00008481, -0.00000709, -0.00001748, -0.00000576, -0.00001325, -0.00000584, -0.00000393, -0.00000151, -0.00001401, -0.00000009, -0.00000509, -0.00000008, -0.00000690, 0.00000180, 0.00000100, -0.00000467, -0.00000116, -0.00001211, -0.00000624, 0.00000552, 0.00000402, -0.00000695, -0.00000407, 0.00001033, -0.00000166, -0.00000215, 0.00000097, 0.00000119, -0.00000221, -0.00000191, 0.00000014, 0.00000111, 0.00000084, -0.00000142, -0.00000239, -0.00000201, },
		{ -0.00000518, 0.00000205, 0.00000034, 0.00000015, 0.00001451, -0.00000988, 0.00000294, -0.00000324, 0.00000105, -0.00000311, -0.00001390, 0.00000199, -0.00000584, 0.00000030, -0.00000234, 0.00000349, -0.00000994, -0.00000309, -0.00000761, -0.00000419, -0.00000367, -0.00000270, -0.00001332, -0.00000594, -0.00004350, -0.00000900, 0.00002988, 0.00003866, 0.00001031, 0.00001306, 0.00000897, 0.00001375, 0.00001919, 0.00003770, 0.00003561, -0.00000803, -0.00003783, -0.00000679, -0.00000469, -0.00001029, 0.00000491, -0.00000506, -0.00000588, -0.00000205, -0.00000068, -0.00000013, -0.00000001, -0.00000143, 0.00000316, -0.00000207, -0.00000093, -0.00000471, -0.00000343, 0.00000449, -0.00000313, 0.00000142, -0.00000567, 0.00000710, 0.00000544, -0.00000573, -0.00000283, 0.00000133, 0.00000430, -0.00000329, 0.00000611, -0.00000311, -0.00000017, -0.00000525, -0.00000186, 0.00000105, -0.00000123, -0.00000129, -0.00000032, -0.00000155, -0.00000310, -0.00000521, 0.00000532, -0.00000187, -0.00000778, -0.00000874, -0.00000743, -0.00000764, -0.00001266, -0.00003765, -0.00008517, 0.00007391, 0.00008938, 0.00002236, 0.00005651, 0.00002415, 0.00007835, 0.00007496, -0.00009361, -0.00003356, -0.00001269, -0.00000989, -0.00001067, -0.00000844, -0.00000698, 0.00000160, -0.00001109, -0.00000672, -0.00000214, -0.00000216, -0.00000335, -0.00000371, 0.00000372, -0.00000424, -0.00000076, -0.00000799, -0.00001184, 0.00000317, 0.00000486, -0.00000138, -0.00000966, 0.00000833, 0.00000395, -0.00000413, 0.00000014, 0.00000173, -0.00000104, -0.00000245, -0.00000097, 0.00000142, 0.00000065, 0.00000012, -0.00000318, -0.00000052, },
		{ -0.00000206, -0.00000342, 0.00000452, -0.00000473, 0.00001372, -0.00000078, -0.00000567, 0.00000339, -0.00000476, 0.00000436, -0.00001631, -0.00000023, -0.00000443, -0.00000126, -0.00000273, 0.00000324, -0.00000472, -0.00000690, -0.00000572, -0.00000484, -0.00000566, -0.00000020, -0.00001169, -0.00000558, -0.00002505, -0.00003835, 0.00001581, 0.00004062, 0.00002573, 0.00000877, 0.00001301, 0.00001068, 0.00001662, 0.00003023, 0.00004543, 0.00001071, -0.00003441, -0.00002218, 0.00000141, -0.00001373, 0.00000066, -0.00000005, -0.00000649, -0.00000397, -0.00000055, -0.00000098, 0.00000077, -0.00000182, 0.00000143, 0.00000111, -0.00000228, -0.00000185, -0.00000719, 0.00000406, -0.00000091, -0.00000003, -0.00000313, 0.00000006, 0.00001052, -0.00000390, -0.00000414, -0.00000129, 0.00000590, -0.00000205, 0.00000211, 0.00000221, -0.00000306, -0.00000071, -0.00000731, 0.00000305, -0.00000184, 0.00000016, -0.00000240, 0.00000099, -0.00000374, -0.00000308, -0.00000259, 0.00000683, -0.00000999, -0.00000413, -0.00001341, -0.00000108, -0.00001668, -0.00001094, -0.00008931, 0.00000627, 0.00010983, 0.00003425, 0.00003868, 0.00004310, 0.00002987, 0.00010895, -0.00001809, -0.00008356, -0.00000618, -0.00001878, -0.00000436, -0.00001378, -0.00000498, -0.00000476, -0.00000005, -0.00001527, 0.00000023, -0.00000519, 0.00000085, -0.00000739, 0.00000370, -0.00000295, -0.00000094, -0.00000410, -0.00001326, -0.00000275, 0.00000606, 0.00000148, -0.00000805, -0.00000047, 0.00001114, -0.00000470, -0.00000077, 0.00000075, 0.00000119, -0.00000312, -0.00000131, 0.00000020, 0.00000175, -0.00000002, -0.00000177, -0.00000196, },
		{ -0.00000022, -0.00000501, 0.00000310, -0.00000249, 0.00000544, 0.00001053, -0.00001054, 0.00000430, -0.00000471, 0.00000445, -0.00001036, -0.00000835, 0.00000074, -0.00000556, -0.00000014, -0.00000109, 0.00000198, -0.00000882, -0.00000539, -0.00000580, -0.00000444, -0.00000395, -0.00000384, -0.00001254, -0.00000731, -0.00004279, -0.00001543, 0.00003684, 0.00003721, 0.00001519, 0.00001058, 0.00001286, 0.00001358, 0.00002248, 0.00004662, 0.00003087, -0.00002278, -0.00003482, -0.00000167, -0.00000901, -0.00000757, 0.00000285, -0.00000439, -0.00000577, -0.00000149, -0.00000094, 0.00000015, -0.00000053, -0.00000103, 0.00000264, -0.00000145, -0.00000105, -0.00000672, -0.00000096, 0.00000349, -0.00000256, 0.00000043, -0.00000518, 0.00001027, 0.00000158, -0.00000553, -0.00000315, 0.00000429, 0.00000242, -0.00000261, 0.00000518, -0.00000320, 0.00000099, -0.00000724, -0.00000099, 0.00000123, -0.00000090, -0.00000164, 0.00000007, -0.00000187, -0.00000234, -0.00000554, 0.00000537, -0.00000309, -0.00000645, -0.00001064, -0.00000511, -0.00001115, -0.00000675, -0.00005700, -0.00006109, 0.00009946, 0.00006045, 0.00002658, 0.00004582, 0.00002487, 0.00006824, 0.00007623, -0.00008661, -0.00003304, -0.00001360, -0.00000909, -0.00001093, -0.00000741, -0.00000785, 0.00000274, -0.00001165, -0.00000671, -0.00000262, -0.00000081, -0.00000438, -0.00000137, 0.00000095, -0.00000354, 0.00000037, -0.00001213, -0.00000865, 0.00000491, 0.00000337, -0.00000368, -0.00000780, 0.00001158, 0.00000098, -0.00000412, 0.00000083, 0.00000142, -0.00000157, -0.00000256, -0.00000053, 0.00000179, 0.00000052, -0.00000075, -0.00000293, },
		{ -0.00000143, -0.00000224, -0.00000199, 0.00000238, -0.00000275, 0.00001556, -0.00000654, -0.00000130, -0.00000041, 0.00000031, -0.00000151, -0.00001520, 0.00000231, -0.00000582, -0.00000105, -0.00000197, 0.00000246, -0.00000428, -0.00000803, -0.00000583, -0.00000375, -0.00000591, -0.00000161, -0.00001086, -0.00000717, -0.00002337, -0.00004336, 0.00001625, 0.00004478, 0.00002613, 0.00001154, 0.00001161, 0.00001558, 0.00001445, 0.00004069, 0.00004691, -0.00000299, -0.00004108, -0.00001206, -0.00000315, -0.00001184, -0.00000040, -0.00000012, -0.00000641, -0.00000306, -0.00000111, -0.00000031, -0.00000002, -0.00000129, 0.00000089, 0.00000139, -0.00000180, -0.00000353, -0.00000671, 0.00000554, -0.00000250, 0.00000140, -0.00000554, 0.00000470, 0.00000787, -0.00000491, -0.00000474, 0.00000108, 0.00000599, -0.00000313, 0.00000272, 0.00000071, -0.00000129, -0.00000206, -0.00000766, 0.00000389, -0.00000179, 0.00000016, -0.00000234, 0.00000094, -0.00000342, -0.00000294, -0.00000248, 0.00000554, -0.00000907, -0.00000530, -0.00001199, -0.00000332, -0.00001310, -0.00001960, -0.00009226, 0.00004856, 0.00009099, 0.00002660, 0.00003642, 0.00003678, 0.00002793, 0.00010063, -0.00001094, -0.00007989, -0.00000743, -0.00001831, -0.00000445, -0.00001299, -0.00000512, -0.00000455, 0.00000023, -0.00001556, -0.00000022, -0.00000431, 0.00000060, -0.00000644, 0.00000380, -0.00000551, 0.00000240, -0.00000795, -0.00001288, 0.00000072, 0.00000524, -0.00000052, -0.00000861, 0.00000375, 0.00001011, -0.00000668, 0.00000064, 0.00000030, 0.00000104, -0.00000376, -0.00000076, 0.00000042, 0.00000210, -0.00000089, -0.00000199, },
		{ -0.00000246, -0.00000021, -0.00000422, 0.00000295, -0.00000406, 0.00001059, 0.00000469, -0.00000867, 0.00000338, -0.00000337, 0.00000456, -0.00001546, -0.00000279, -0.00000130, -0.00000490, -0.00000033, -0.00000041, 0.00000083, -0.00000795, -0.00000752, -0.00000381, -0.00000486, -0.00000435, -0.00000479, -0.00001201, -0.00000836, -0.00004216, -0.00002193, 0.00004317, 0.00003680, 0.00001935, 0.00000854, 0.00001753, 0.00001255, 0.00002786, 0.00005521, 0.00002158, -0.00003722, -0.00002651, -0.00000101, -0.00001094, -0.00000602, 0.00000119, -0.00000370, -0.00000535, -0.00000138, -0.00000090, 0.00000008, -0.00000072, -0.00000089, 0.00000225, -0.00000071, -0.00000175, -0.00000810, 0.00000165, 0.00000174, -0.00000120, -0.00000145, -0.00000297, 0.00001142, -0.00000160, -0.00000535, -0.00000278, 0.00000679, 0.00000021, -0.00000157, 0.00000376, -0.00000243, 0.00000129, -0.00000886, 0.00000009, 0.00000126, -0.00000068, -0.00000182, 0.00000032, -0.00000201, -0.00000181, -0.00000556, 0.00000479, -0.00000361, -0.00000590, -0.00001130, -0.00000435, -0.00001241, -0.00000480, -0.00007372, -0.00002677, 0.00010679, 0.00003862, 0.00003024, 0.00003632, 0.00002497, 0.00006073, 0.00007511, -0.00007859, -0.00003334, -0.00001399, -0.00000866, -0.00001089, -0.00000658, -0.00000832, 0.00000341, -0.00001202, -0.00000688, -0.00000262, -0.00000000, -0.00000485, 0.00000057, -0.00000225, -0.00000153, -0.00000026, -0.00001528, -0.00000449, 0.00000520, 0.00000191, -0.00000558, -0.00000489, 0.00001356, -0.00000228, -0.00000307, 0.00000090, 0.00000117, -0.00000215, -0.00000243, -0.00000013, 0.00000216, 0.00000014, -0.00000150, },
		{ -0.00000205, -0.00000093, -0.00000227, -0.00000069, -0.00000039, 0.00000129, 0.00001425, -0.00000990, 0.00000170, -0.00000287, 0.00000460, -0.00000848, -0.00001153, 0.00000290, -0.00000646, -0.00000095, -0.00000110, 0.00000120, -0.00000328, -0.00000978, -0.00000514, -0.00000321, -0.00000591, -0.00000330, -0.00000974, -0.00000906, -0.00002117, -0.00004872, 0.00001646, 0.00004882, 0.00002746, 0.00001293, 0.00001243, 0.00001887, 0.00001405, 0.00005192, 0.00004563, -0.00002084, -0.00004076, -0.00000460, -0.00000777, -0.00000924, -0.00000226, 0.00000034, -0.00000625, -0.00000231, -0.00000165, 0.00000023, -0.00000072, -0.00000079, 0.00000032, 0.00000180, -0.00000150, -0.00000549, -0.00000508, 0.00000565, -0.00000304, 0.00000186, -0.00000665, 0.00000862, 0.00000447, -0.00000517, -0.00000526, 0.00000389, 0.00000510, -0.00000369, 0.00000292, -0.00000037, 0.00000049, -0.00000404, -0.00000710, 0.00000420, -0.00000152, -0.00000012, -0.00000193, 0.00000055, -0.00000276, -0.00000313, -0.00000215, 0.00000383, -0.00000759, -0.00000691, -0.00000984, -0.00000655, -0.00000793, -0.00003414, -0.00008013, 0.00007884, 0.00006870, 0.00002452, 0.00003229, 0.00003185, 0.00002638, 0.00009309, -0.00000557, -0.00007494, -0.00000990, -0.00001676, -0.00000548, -0.00001140, -0.00000582, -0.00000375, -0.00000028, -0.00001524, -0.00000098, -0.00000299, -0.00000028, -0.00000456, 0.00000233, -0.00000637, 0.00000465, -0.00001216, -0.00001080, 0.00000333, 0.00000365, -0.00000203, -0.00000850, 0.00000808, 0.00000739, -0.00000716, 0.00000156, -0.00000017, 0.00000070, -0.00000407, -0.00000029, 0.00000078, 0.00000212, -0.00000169, },
		{ -0.00000213, -0.00000180, -0.00000043, -0.00000289, 0.00000145, -0.00000359, 0.00001408, -0.00000134, -0.00000556, 0.00000129, -0.00000011, 0.00000147, -0.00001756, 0.00000165, -0.00000345, -0.00000403, -0.00000026, -0.00000043, 0.00000032, -0.00000769, -0.00000904, -0.00000210, -0.00000518, -0.00000502, -0.00000539, -0.00001190, -0.00000889, -0.00004185, -0.00002821, 0.00004839, 0.00003811, 0.00002190, 0.00000813, 0.00002182, 0.00001171, 0.00003494, 0.00006312, 0.00000608, -0.00004730, -0.00001611, -0.00000369, -0.00001052, -0.00000596, 0.00000028, -0.00000317, -0.00000467, -0.00000165, -0.00000063, -0.00000024, -0.00000066, -0.00000097, 0.00000200, 0.00000008, -0.00000304, -0.00000844, 0.00000367, -0.00000006, 0.00000037, -0.00000353, 0.00000029, 0.00001071, -0.00000382, -0.00000533, -0.00000160, 0.00000851, -0.00000207, -0.00000034, 0.00000205, -0.00000090, 0.00000054, -0.00000967, 0.00000103, 0.00000129, -0.00000069, -0.00000169, 0.00000023, -0.00000178, -0.00000172, -0.00000512, 0.00000342, -0.00000321, -0.00000631, -0.00001048, -0.00000563, -0.00001052, -0.00000924, -0.00008171, 0.00000932, 0.00010028, 0.00002482, 0.00003177, 0.00002872, 0.00002424, 0.00005561, 0.00007163, -0.00006931, -0.00003486, -0.00001339, -0.00000905, -0.00001014, -0.00000631, -0.00000811, 0.00000332, -0.00001192, -0.00000737, -0.00000199, 0.00000001, -0.00000440, 0.00000143, -0.00000494, 0.00000100, -0.00000242, -0.00001690, 0.00000000, 0.00000400, 0.00000100, -0.00000734, -0.00000084, 0.00001363, -0.00000490, -0.00000157, 0.00000047, 0.00000095, -0.00000266, -0.00000217, 0.00000033, 0.00000235, -0.00000033, },
		{ -0.00000238, -0.00000192, -0.00000060, -0.00000191, -0.00000011, -0.00000273, 0.00000623, 0.00001031, -0.00001066, 0.00000286, -0.00000317, 0.00000659, -0.00001468, -0.00000662, 0.00000200, -0.00000642, -0.00000076, -0.00000062, 0.00000005, -0.00000230, -0.00001159, -0.00000412, -0.00000295, -0.00000594, -0.00000488, -0.00000876, -0.00001076, -0.00001907, -0.00005377, 0.00001560, 0.00005370, 0.00002870, 0.00001398, 0.00001474, 0.00002099, 0.00001541, 0.00006415, 0.00003906, -0.00003969, -0.00003384, -0.00000156, -0.00001057, -0.00000732, -0.00000399, 0.00000085, -0.00000574, -0.00000199, -0.00000194, 0.00000046, -0.00000113, -0.00000050, -0.00000015, 0.00000229, -0.00000155, -0.00000728, -0.00000278, 0.00000459, -0.00000253, 0.00000129, -0.00000623, 0.00001115, 0.00000107, -0.00000506, -0.00000545, 0.00000682, 0.00000331, -0.00000360, 0.00000258, -0.00000073, 0.00000176, -0.00000615, -0.00000589, 0.00000411, -0.00000118, -0.00000043, -0.00000149, 0.00000015, -0.00000213, -0.00000326, -0.00000210, 0.00000234, -0.00000632, -0.00000798, -0.00000842, -0.00000866, -0.00000457, -0.00004964, -0.00005612, 0.00009428, 0.00004877, 0.00002453, 0.00002787, 0.00002730, 0.00002602, 0.00008521, -0.00000068, -0.00006996, -0.00001251, -0.00001503, -0.00000665, -0.00000966, -0.00000649, -0.00000294, -0.00000105, -0.00001476, -0.00000156, -0.00000184, -0.00000108, -0.00000257, -0.00000000, -0.00000575, 0.00000545, -0.00001609, -0.00000712, 0.00000434, 0.00000216, -0.00000342, -0.00000731, 0.00001170, 0.00000366, -0.00000620, 0.00000170, -0.00000047, 0.00000019, -0.00000409, 0.00000012, 0.00000119, 0.00000187, },
		{ -0.00000079, -0.00000264, -0.00000104, -0.00000071, -0.00000139, -0.00000102, -0.00000096, 0.00001495, -0.00000614, -0.00000255, -0.00000057, 0.00000348, -0.00000367, -0.00001674, 0.00000440, -0.00000528, -0.00000294, -0.00000027, -0.00000085, 0.00000028, -0.00000795, -0.00000991, -0.00000075, -0.00000543, -0.00000581, -0.00000582, -0.00001197, -0.00000914, -0.00004162, -0.00003459, 0.00005276, 0.00004098, 0.00002280, 0.00000977, 0.00002518, 0.00001157, 0.00004396, 0.00006862, -0.00001535, -0.00004993, -0.00000731, -0.00000739, -0.00000892, -0.00000684, -0.00000009, -0.00000262, -0.00000398, -0.00000206, -0.00000037, -0.00000058, -0.00000055, -0.00000112, 0.00000185, 0.00000083, -0.00000480, -0.00000762, 0.00000461, -0.00000130, 0.00000156, -0.00000515, 0.00000376, 0.00000864, -0.00000509, -0.00000548, 0.00000035, 0.00000916, -0.00000408, 0.00000087, 0.00000033, 0.00000106, -0.00000118, -0.00000951, 0.00000159, 0.00000140, -0.00000088, -0.00000130, -0.00000014, -0.00000124, -0.00000196, -0.00000439, 0.00000156, -0.00000219, -0.00000723, -0.00000886, -0.00000806, -0.00000654, -0.00001966, -0.00007775, 0.00003964, 0.00008617, 0.00001741, 0.00003133, 0.00002266, 0.00002326, 0.00005186, 0.00006675, -0.00005944, -0.00003715, -0.00001204, -0.00001006, -0.00000884, -0.00000646, -0.00000735, 0.00000257, -0.00001145, -0.00000799, -0.00000101, -0.00000051, -0.00000320, 0.00000109, -0.00000657, 0.00000333, -0.00000577, -0.00001650, 0.00000384, 0.00000188, 0.00000067, -0.00000886, 0.00000408, 0.00001159, -0.00000611, -0.00000030, -0.00000014, 0.00000068, -0.00000300, -0.00000187, 0.00000087, 0.00000229, },
		{ 0.00000145, -0.00000297, -0.00000157, -0.00000048, -0.00000109, -0.00000062, -0.00000370, 0.00001067, 0.00000498, -0.00000945, 0.00000257, -0.00000149, 0.00000568, -0.00001874, -0.00000174, 0.00000020, -0.00000575, -0.00000051, -0.00000069, -0.00000064, -0.00000176, -0.00001298, -0.00000316, -0.00000272, -0.00000618, -0.00000613, -0.00000817, -0.00001196, -0.00001741, -0.00005811, 0.00001314, 0.00006009, 0.00002901, 0.00001574, 0.00001767, 0.00002254, 0.00001860, 0.00007710, 0.00002484, -0.00005503, -0.00002311, -0.00000259, -0.00001099, -0.00000677, -0.00000501, 0.00000109, -0.00000480, -0.00000215, -0.00000194, 0.00000037, -0.00000124, -0.00000046, -0.00000047, 0.00000279, -0.00000207, -0.00000844, -0.00000044, 0.00000287, -0.00000127, 0.00000001, -0.00000450, 0.00001202, -0.00000180, -0.00000486, -0.00000504, 0.00000945, 0.00000083, -0.00000284, 0.00000166, -0.00000022, 0.00000210, -0.00000782, -0.00000442, 0.00000377, -0.00000090, -0.00000057, -0.00000123, 0.00000000, -0.00000180, -0.00000308, -0.00000261, 0.00000147, -0.00000564, -0.00000801, -0.00000852, -0.00000830, -0.00000584, -0.00006050, -0.00002639, 0.00009630, 0.00003395, 0.00002450, 0.00002401, 0.00002280, 0.00002687, 0.00007666, 0.00000431, -0.00006563, -0.00001457, -0.00001370, -0.00000746, -0.00000823, -0.00000673, -0.00000251, -0.00000177, -0.00001433, -0.00000174, -0.00000119, -0.00000135, -0.00000114, -0.00000244, -0.00000412, 0.00000462, -0.00001901, -0.00000239, 0.00000347, 0.00000149, -0.00000515, -0.00000463, 0.00001368, -0.00000008, -0.00000428, 0.00000110, -0.00000056, -0.00000042, -0.00000387, 0.00000053, 0.00000151, },
		{ 0.00000196, -0.00000120, -0.00000298, -0.00000032, -0.00000083, -0.00000020, -0.00000369, 0.00000321, 0.00001307, -0.00000890, -0.00000052, -0.00000116, 0.00000584, -0.00000928, -0.00001379, 0.00000542, -0.00000655, -0.00000157, -0.00000073, -0.00000117, 0.00000025, -0.00000831, -0.00001043, 0.00000039, -0.00000580, -0.00000642, -0.00000644, -0.00001181, -0.00000961, -0.00004092, -0.00004169, 0.00005696, 0.00004468, 0.00002275, 0.00001312, 0.00002773, 0.00001224, 0.00005558, 0.00006862, -0.00003986, -0.00004443, -0.00000277, -0.00000987, -0.00000763, -0.00000771, -0.00000042, -0.00000173, -0.00000362, -0.00000231, -0.00000035, -0.00000071, -0.00000059, -0.00000119, 0.00000171, 0.00000141, -0.00000673, -0.00000582, 0.00000432, -0.00000163, 0.00000205, -0.00000584, 0.00000669, 0.00000585, -0.00000561, -0.00000565, 0.00000294, 0.00000856, -0.00000543, 0.00000176, -0.00000101, 0.00000299, -0.00000348, -0.00000843, 0.00000170, 0.00000156, -0.00000110, -0.00000086, -0.00000053, -0.00000068, -0.00000225, -0.00000378, -0.00000030, -0.00000113, -0.00000787, -0.00000761, -0.00000997, -0.00000303, -0.00003293, -0.00006249, 0.00005968, 0.00007008, 0.00001387, 0.00002978, 0.00001754, 0.00002272, 0.00004839, 0.00006161, -0.00004999, -0.00003939, -0.00001060, -0.00001114, -0.00000747, -0.00000662, -0.00000647, 0.00000151, -0.00001090, -0.00000837, -0.00000016, -0.00000104, -0.00000179, -0.00000009, -0.00000706, 0.00000494, -0.00000981, -0.00001387, 0.00000600, -0.00000017, 0.00000040, -0.00000962, 0.00000908, 0.00000781, -0.00000563, 0.00000022, -0.00000063, 0.00000027, -0.00000311, -0.00000155, 0.00000145, },
		},
		{ { -0.00000192, -0.00000130, 0.00000203, 0.00000073, 0.00000018, -0.00000127, 0.00000223, -0.00000362, 0.00000306, -0.00000395, -0.00000352, -0.00000111, -0.00000078, -0.00000204, 0.00000684, -0.00000348, -0.00000295, -0.00000288, -0.00001359, -0.00000640, -0.00001139, -0.00001583, -0.00005673, 0.00000172, 0.00007380, 0.00002830, 0.00003299, 0.00002042, 0.00002332, 0.00002799, 0.00007324, 0.00000447, -0.00005490, -0.00001630, -0.00001023, -0.00000779, -0.00000569, -0.00000291, -0.00001880, 0.00000046, -0.00000487, -0.00000482, -0.00000387, -0.00000073, 0.00000008, -0.00000045, -0.00000762, -0.00000088, -0.00000112, -0.00000083, -0.00000320, -0.00000137, -0.00000388, -0.00000429, -0.00000196, 0.00000448, -0.00001019, 0.00000425, -0.00000492, 0.00000202, -0.00000048, -0.00000428, 0.00000979, -0.00000290, -0.00000105, -0.00000290, 0.00000979, -0.00000428, -0.00000048, 0.00000202, -0.00000492, 0.00000425, -0.00001019, 0.00000448, -0.00000196, -0.00000429, -0.00000388, -0.00000137, -0.00000320, -0.00000083, -0.00000112, -0.00000088, -0.00000762, -0.00000045, 0.00000008, -0.00000073, -0.00000387, -0.00000482, -0.00000487, 0.00000046, -0.00001880, -0.00000291, -0.00000569, -0.00000779, -0.00001023, -0.00001630, -0.00005490, 0.00000447, 0.00007324, 0.00002799, 0.00002332, 0.00002042, 0.00003299, 0.00002830, 0.00007380, 0.00000172, -0.00005673, -0.00001583, -0.00001139, -0.00000640, -0.00001359, -0.00000288, -0.00000295, -0.00000348, 0.00000684, -0.00000204, -0.00000078, -0.00000111, -0.00000352, -0.00000395, 0.00000306, -0.00000362, 0.00000223, -0.00000127, 0.00000018, 0.00000073, 0.00000203, -0.00000130, },
		{ -0.00000188, -0.00000209, 0.00000036, 0.00000158, 0.00000090, -0.00000125, 0.00000125, -0.00000068, -0.00000188, 0.00000222, -0.00000564, -0.00000187, -0.00000089, -0.00000214, 0.00000153, 0.00000535, -0.00000709, 0.00000081, -0.00000907, -0.00001116, -0.00000816, -0.00001158, -0.00002782, -0.00005845, 0.00004576, 0.00006270, 0.00002513, 0.00003318, 0.00001981, 0.00002398, 0.00005133, 0.00006292, -0.00004753, -0.00003507, -0.00001288, -0.00000729, -0.00000975, 0.00000044, -0.00001444, -0.00001021, 0.00000091, -0.00000681, -0.00000431, -0.00000239, -0.00000049, 0.00000133, -0.00000526, -0.00000465, -0.00000010, -0.00000115, -0.00000220, -0.00000208, -0.00000249, -0.00000335, -0.00000609, 0.00000378, -0.00000141, -0.00000726, 0.00000409, -0.00000621, 0.00000605, -0.00000672, 0.00000298, 0.00000608, -0.00000323, -0.00000198, 0.00000056, 0.00000904, -0.00000852, 0.00000584, -0.00000517, 0.00000154, -0.00000161, -0.00000662, 0.00000452, -0.00000450, -0.00000348, -0.00000347, -0.00000160, -0.00000285, -0.00000035, -0.00000055, -0.00000458, -0.00000553, 0.00000157, -0.00000098, -0.00000180, -0.00000419, -0.00000589, -0.00000036, -0.00000863, -0.00001525, 0.00000127, -0.00001040, -0.00000659, -0.00001135, -0.00003282, -0.00004422, 0.00005486, 0.00005000, 0.00002232, 0.00001906, 0.00002626, 0.00002665, 0.00004383, 0.00006171, -0.00003271, -0.00004361, -0.00001035, -0.00000962, -0.00000753, -0.00001265, 0.00000106, -0.00000699, 0.00000337, 0.00000295, -0.00000192, -0.00000082, -0.00000100, -0.00000579, 0.00000008, 0.00000051, -0.00000171, 0.00000179, -0.00000186, 0.00000087, 0.00000131, 0.00000103, },
		{ -0.00000063, -0.00000185, -0.00000186, 0.00000176, 0.00000086, 0.00000066, -0.00000139, 0.00000260, -0.00000389, 0.00000154, -0.00000118, -0.00000482, -0.00000115, -0.00000111, -0.00000249, 0.00000572, -0.00000040, -0.00000543, 0.00000043, -0.00001377, -0.00000772, -0.00001219, -0.00001028, -0.00004805, -0.00003636, 0.00007927, 0.00004007, 0.00003290, 0.00002678, 0.00002351, 0.00002909, 0.00007852, 0.00000908, -0.00006127, -0.00001549, -0.00001372, -0.00000586, -0.00000721, -0.00000173, -0.00002011, 0.00000069, -0.00000455, -0.00000547, -0.00000433, -0.00000073, 0.00000018, -0.00000027, -0.00000783, -0.00000034, -0.00000139, -0.00000104, -0.00000304, -0.00000127, -0.00000318, -0.00000428, -0.00000475, 0.00000727, -0.00000835, -0.00000090, -0.00000057, -0.00000177, 0.00000373, -0.00000804, 0.00000860, 0.00000093, -0.00000184, -0.00000382, 0.00000654, 0.00000274, -0.00000599, 0.00000614, -0.00000840, 0.00000660, -0.00000835, -0.00000022, 0.00000048, -0.00000362, -0.00000443, -0.00000190, -0.00000287, -0.00000133, -0.00000037, -0.00000142, -0.00000752, -0.00000069, 0.00000024, -0.00000106, -0.00000325, -0.00000448, -0.00000469, -0.00000005, -0.00001736, -0.00000422, -0.00000412, -0.00000943, -0.00000760, -0.00001633, -0.00004928, 0.00000037, 0.00006857, 0.00002647, 0.00002324, 0.00001614, 0.00003011, 0.00002392, 0.00005935, 0.00003207, -0.00004970, -0.00002842, -0.00000848, -0.00000750, -0.00001050, -0.00000780, -0.00000014, -0.00000613, 0.00000724, -0.00000164, -0.00000010, -0.00000149, -0.00000177, -0.00000618, 0.00000320, -0.00000236, 0.00000141, -0.00000077, -0.00000056, 0.00000081, 0.00000198, },
		{ 0.00000141, -0.00000147, -0.00000216, -0.00000039, 0.00000165, 0.00000104, -0.00000072, 0.00000082, 0.00000029, -0.00000362, 0.00000271, -0.00000389, -0.00000350, -0.00000046, -0.00000266, 0.00000053, 0.00000576, -0.00000501, -0.00000180, -0.00000377, -0.00001502, -0.00000622, -0.00001521, -0.00001273, -0.00006900, 0.00001198, 0.00008337, 0.00002663, 0.00003837, 0.00002264, 0.00002524, 0.00005232, 0.00007229, -0.00005142, -0.00003737, -0.00001402, -0.00000883, -0.00000850, -0.00000070, -0.00001343, -0.00001201, 0.00000227, -0.00000743, -0.00000475, -0.00000282, -0.00000034, 0.00000147, -0.00000521, -0.00000478, 0.00000055, -0.00000177, -0.00000213, -0.00000205, -0.00000209, -0.00000270, -0.00000702, 0.00000149, 0.00000384, -0.00001013, 0.00000278, -0.00000398, 0.00000347, -0.00000236, -0.00000384, 0.00000986, -0.00000278, -0.00000104, -0.00000380, 0.00001132, -0.00000570, 0.00000173, -0.00000064, -0.00000379, 0.00000464, -0.00000992, 0.00000317, -0.00000259, -0.00000289, -0.00000483, -0.00000100, -0.00000347, 0.00000004, -0.00000052, -0.00000459, -0.00000601, 0.00000201, -0.00000151, -0.00000133, -0.00000415, -0.00000496, -0.00000133, -0.00000730, -0.00001585, 0.00000181, -0.00001063, -0.00000617, -0.00001020, -0.00003009, -0.00004147, 0.00004761, 0.00004888, 0.00001997, 0.00001973, 0.00001936, 0.00002772, 0.00003106, 0.00006203, -0.00000175, -0.00005126, -0.00001560, -0.00000837, -0.00000635, -0.00001270, -0.00000190, -0.00000461, 0.00000005, 0.00000482, -0.00000239, 0.00000016, -0.00000124, -0.00000430, -0.00000314, 0.00000276, -0.00000229, 0.00000213, -0.00000225, 0.00000064, 0.00000118, },
		{ 0.00000180, -0.00000006, -0.00000159, -0.00000228, 0.00000125, 0.00000096, 0.00000115, -0.00000144, 0.00000282, -0.00000363, -0.00000037, 0.00000083, -0.00000464, -0.00000232, -0.00000055, -0.00000343, 0.00000471, 0.00000162, -0.00000573, 0.00000039, -0.00000996, -0.00001216, -0.00000955, -0.00001246, -0.00002888, -0.00007015, 0.00006669, 0.00006097, 0.00003040, 0.00003454, 0.00002420, 0.00003011, 0.00008357, 0.00001537, -0.00006941, -0.00001344, -0.00001804, -0.00000384, -0.00000859, -0.00000061, -0.00002136, 0.00000067, -0.00000386, -0.00000613, -0.00000496, -0.00000069, 0.00000005, 0.00000028, -0.00000840, 0.00000029, -0.00000135, -0.00000142, -0.00000314, -0.00000086, -0.00000298, -0.00000339, -0.00000743, 0.00000792, -0.00000412, -0.00000618, 0.00000181, -0.00000295, 0.00000486, -0.00000754, 0.00000339, 0.00000641, -0.00000332, -0.00000261, 0.00000081, 0.00000988, -0.00000948, 0.00000746, -0.00000822, 0.00000450, -0.00000292, -0.00000579, 0.00000201, -0.00000265, -0.00000417, -0.00000330, -0.00000179, -0.00000252, 0.00000064, -0.00000185, -0.00000737, -0.00000126, 0.00000089, -0.00000180, -0.00000244, -0.00000454, -0.00000403, -0.00000079, -0.00001572, -0.00000574, -0.00000246, -0.00001088, -0.00000535, -0.00001654, -0.00004325, -0.00000414, 0.00006487, 0.00002461, 0.00002286, 0.00001387, 0.00002524, 0.00002336, 0.00004406, 0.00004865, -0.00002942, -0.00004201, -0.00000763, -0.00000829, -0.00000728, -0.00001144, 0.00000063, -0.00000674, 0.00000600, -0.00000053, -0.00000002, -0.00000123, -0.00000079, -0.00000672, 0.00000137, -0.00000002, 0.00000016, 0.00000011, -0.00000153, 0.00000101, },
		{ 0.00000139, 0.00000143, -0.00000093, -0.00000208, -0.00000105, 0.00000144, 0.00000125, -0.00000018, 0.00000043, 0.00000117, -0.00000468, 0.00000193, -0.00000152, -0.00000472, -0.00000060, -0.00000258, -0.00000065, 0.00000535, -0.00000226, -0.00000435, 0.00000012, -0.00001589, -0.00000708, -0.00001628, -0.00000572, -0.00006051, -0.00003349, 0.00009613, 0.00003644, 0.00003959, 0.00002849, 0.00002589, 0.00005331, 0.00008230, -0.00005471, -0.00004116, -0.00001361, -0.00001187, -0.00000635, -0.00000233, -0.00001201, -0.00001420, 0.00000372, -0.00000771, -0.00000541, -0.00000320, -0.00000037, 0.00000166, -0.00000498, -0.00000525, 0.00000144, -0.00000212, -0.00000247, -0.00000194, -0.00000170, -0.00000227, -0.00000681, -0.00000183, 0.00000787, -0.00000970, -0.00000130, -0.00000065, 0.00000067, 0.00000161, -0.00000829, 0.00000965, 0.00000054, -0.00000165, -0.00000525, 0.00000877, 0.00000128, -0.00000477, 0.00000507, -0.00000879, 0.00000837, -0.00000944, -0.00000057, -0.00000052, -0.00000198, -0.00000572, -0.00000119, -0.00000324, -0.00000059, 0.00000023, -0.00000481, -0.00000639, 0.00000218, -0.00000157, -0.00000136, -0.00000386, -0.00000440, -0.00000174, -0.00000641, -0.00001601, 0.00000181, -0.00001038, -0.00000596, -0.00000932, -0.00002750, -0.00003819, 0.00003990, 0.00004899, 0.00001644, 0.00002148, 0.00001325, 0.00002725, 0.00002407, 0.00005349, 0.00002418, -0.00004581, -0.00002703, -0.00000535, -0.00000724, -0.00000991, -0.00000669, -0.00000157, -0.00000321, 0.00000627, -0.00000298, 0.00000141, -0.00000188, -0.00000222, -0.00000603, 0.00000370, -0.00000172, 0.00000176, -0.00000199, -0.00000010, },
		{ 0.00000088, 0.00000188, 0.00000015, -0.00000106, -0.00000260, 0.00000074, 0.00000075, 0.00000183, -0.00000153, 0.00000301, -0.00000314, -0.00000204, 0.00000148, -0.00000293, -0.00000428, 0.00000045, -0.00000444, 0.00000385, 0.00000216, -0.00000371, -0.00000237, -0.00000384, -0.00001743, -0.00000593, -0.00001785, -0.00000948, -0.00008717, 0.00003182, 0.00008714, 0.00002943, 0.00004134, 0.00002671, 0.00003080, 0.00008824, 0.00002353, -0.00007905, -0.00001095, -0.00002218, -0.00000253, -0.00000941, 0.00000019, -0.00002232, 0.00000021, -0.00000286, -0.00000657, -0.00000595, -0.00000038, -0.00000055, 0.00000132, -0.00000927, 0.00000089, -0.00000114, -0.00000154, -0.00000387, 0.00000002, -0.00000325, -0.00000206, -0.00000929, 0.00000655, 0.00000095, -0.00000948, 0.00000094, -0.00000136, 0.00000343, -0.00000416, -0.00000332, 0.00001086, -0.00000353, -0.00000128, -0.00000435, 0.00001365, -0.00000821, 0.00000416, -0.00000360, -0.00000133, 0.00000405, -0.00001016, 0.00000177, -0.00000143, -0.00000311, -0.00000518, -0.00000058, -0.00000361, 0.00000112, -0.00000173, -0.00000734, -0.00000201, 0.00000170, -0.00000252, -0.00000177, -0.00000471, -0.00000323, -0.00000149, -0.00001405, -0.00000727, -0.00000094, -0.00001209, -0.00000319, -0.00001742, -0.00003664, -0.00000867, 0.00006126, 0.00002346, 0.00002127, 0.00001373, 0.00001924, 0.00002402, 0.00003201, 0.00005239, -0.00000396, -0.00004890, -0.00001252, -0.00000674, -0.00000607, -0.00001187, -0.00000175, -0.00000494, 0.00000311, 0.00000147, -0.00000097, 0.00000009, -0.00000102, -0.00000538, -0.00000194, 0.00000261, -0.00000091, 0.00000087, -0.00000226, },
		{ -0.00000128, 0.00000177, 0.00000137, -0.00000054, -0.00000173, -0.00000167, 0.00000110, 0.00000122, 0.00000067, -0.00000016, 0.00000215, -0.00000533, 0.00000066, 0.00000025, -0.00000448, -0.00000203, -0.00000142, -0.00000233, 0.00000489, -0.00000042, -0.00000473, 0.00000044, -0.00001217, -0.00001196, -0.00001288, -0.00000826, -0.00003815, -0.00007691, 0.00008919, 0.00005827, 0.00003635, 0.00003729, 0.00002609, 0.00005474, 0.00009206, -0.00005613, -0.00004757, -0.00001109, -0.00001646, -0.00000345, -0.00000439, -0.00001013, -0.00001676, 0.00000517, -0.00000773, -0.00000599, -0.00000391, -0.00000013, 0.00000134, -0.00000408, -0.00000639, 0.00000268, -0.00000243, -0.00000264, -0.00000248, -0.00000063, -0.00000267, -0.00000533, -0.00000548, 0.00001003, -0.00000669, -0.00000598, 0.00000126, -0.00000025, 0.00000327, -0.00000891, 0.00000530, 0.00000593, -0.00000332, -0.00000415, 0.00000292, 0.00000945, -0.00001011, 0.00000830, -0.00000999, 0.00000722, -0.00000465, -0.00000603, 0.00000133, -0.00000120, -0.00000543, -0.00000266, -0.00000187, -0.00000218, 0.00000139, -0.00000506, -0.00000658, 0.00000185, -0.00000100, -0.00000189, -0.00000330, -0.00000431, -0.00000156, -0.00000601, -0.00001558, 0.00000118, -0.00000963, -0.00000601, -0.00000826, -0.00002599, -0.00003328, 0.00003093, 0.00005054, 0.00001202, 0.00002355, 0.00000853, 0.00002521, 0.00002083, 0.00004253, 0.00003957, -0.00002895, -0.00003951, -0.00000424, -0.00000787, -0.00000673, -0.00001059, -0.00000021, -0.00000469, 0.00000617, -0.00000273, 0.00000183, -0.00000177, -0.00000087, -0.00000722, 0.00000246, 0.00000030, 0.00000059, -0.00000106, },
		{ -0.00000218, 0.00000003, 0.00000245, -0.00000005, -0.00000039, -0.00000279, 0.00000030, 0.00000024, 0.00000251, -0.00000140, 0.00000295, -0.00000235, -0.00000340, 0.00000107, -0.00000061, -0.00000584, 0.00000073, -0.00000455, 0.00000258, 0.00000186, -0.00000082, -0.00000538, 0.00000118, -0.00001981, -0.00000531, -0.00002046, 0.00000094, -0.00008142, -0.00001989, 0.00010768, 0.00003692, 0.00004358, 0.00003304, 0.00003019, 0.00009342, 0.00003249, -0.00008867, -0.00000982, -0.00002462, -0.00000291, -0.00000906, 0.00000022, -0.00002254, -0.00000100, -0.00000148, -0.00000680, -0.00000716, -0.00000000, -0.00000132, 0.00000233, -0.00000986, 0.00000102, -0.00000058, -0.00000149, -0.00000487, 0.00000074, -0.00000323, -0.00000128, -0.00000958, 0.00000348, 0.00000558, -0.00001001, -0.00000248, 0.00000128, 0.00000147, -0.00000046, -0.00000847, 0.00001180, -0.00000102, -0.00000141, -0.00000671, 0.00001230, -0.00000188, -0.00000274, 0.00000331, -0.00000766, 0.00000909, -0.00001096, -0.00000106, 0.00000028, -0.00000195, -0.00000643, -0.00000038, -0.00000353, 0.00000031, -0.00000070, -0.00000765, -0.00000256, 0.00000212, -0.00000264, -0.00000167, -0.00000462, -0.00000274, -0.00000173, -0.00001273, -0.00000833, -0.00000002, -0.00001278, -0.00000122, -0.00001861, -0.00003037, -0.00001180, 0.00005611, 0.00002443, 0.00001754, 0.00001582, 0.00001260, 0.00002482, 0.00002366, 0.00004832, 0.00001800, -0.00004455, -0.00002305, -0.00000359, -0.00000673, -0.00000946, -0.00000620, -0.00000208, -0.00000019, 0.00000352, -0.00000234, 0.00000179, -0.00000175, -0.00000318, -0.00000528, 0.00000415, -0.00000082, 0.00000082, },
		{ -0.00000005, -0.00000230, 0.00000176, 0.00000169, -0.00000060, -0.00000103, -0.00000229, 0.00000087, 0.00000068, 0.00000193, -0.00000094, 0.00000316, -0.00000565, -0.00000060, 0.00000087, -0.00000270, -0.00000440, 0.00000026, -0.00000398, 0.00000439, -0.00000002, -0.00000259, -0.00000246, -0.00000535, -0.00001863, -0.00000754, -0.00001548, -0.00001365, -0.00010278, 0.00005543, 0.00008952, 0.00003248, 0.00004636, 0.00002752, 0.00005601, 0.00010172, -0.00005569, -0.00005619, -0.00000741, -0.00002141, -0.00000078, -0.00000626, -0.00000825, -0.00001927, 0.00000625, -0.00000741, -0.00000631, -0.00000513, 0.00000062, 0.00000027, -0.00000244, -0.00000806, 0.00000413, -0.00000278, -0.00000232, -0.00000392, 0.00000111, -0.00000368, -0.00000328, -0.00000843, 0.00000995, -0.00000226, -0.00000936, 0.00000051, 0.00000107, 0.00000273, -0.00000652, -0.00000117, 0.00001097, -0.00000421, -0.00000243, -0.00000308, 0.00001491, -0.00001079, 0.00000640, -0.00000589, 0.00000136, 0.00000272, -0.00001120, 0.00000193, -0.00000043, -0.00000411, -0.00000488, -0.00000016, -0.00000373, 0.00000203, -0.00000475, -0.00000687, 0.00000129, -0.00000022, -0.00000240, -0.00000287, -0.00000440, -0.00000110, -0.00000580, -0.00001480, 0.00000025, -0.00000876, -0.00000615, -0.00000682, -0.00002591, -0.00002677, 0.00002136, 0.00005230, 0.00000826, 0.00002440, 0.00000612, 0.00002147, 0.00001989, 0.00003250, 0.00004485, -0.00000746, -0.00004618, -0.00000845, -0.00000632, -0.00000524, -0.00001156, -0.00000189, -0.00000369, 0.00000418, -0.00000118, 0.00000083, -0.00000039, -0.00000084, -0.00000631, -0.00000070, 0.00000309, -0.00000073, },
		{ -0.00000036, -0.00000113, -0.00000151, 0.00000330, -0.00000032, 0.00000010, -0.00000264, -0.00000018, -0.00000029, 0.00000274, -0.00000064, 0.00000231, -0.00000109, -0.00000462, 0.00000023, 0.00000118, -0.00000589, -0.00000053, -0.00000326, 0.00000065, 0.00000146, 0.00000126, -0.00000623, 0.00000240, -0.00001699, -0.00000971, -0.00001738, -0.00000040, -0.00005744, -0.00007322, 0.00010789, 0.00005906, 0.00003921, 0.00004407, 0.00002780, 0.00010000, 0.00004107, -0.00009675, -0.00001146, -0.00002456, -0.00000524, -0.00000752, -0.00000061, -0.00002186, -0.00000304, 0.00000022, -0.00000688, -0.00000825, -0.00000002, -0.00000166, 0.00000261, -0.00000952, 0.00000022, 0.00000063, -0.00000167, -0.00000540, 0.00000039, -0.00000209, -0.00000173, -0.00000811, -0.00000056, 0.00000883, -0.00000805, -0.00000672, 0.00000286, 0.00000086, 0.00000165, -0.00001030, 0.00000861, 0.00000392, -0.00000293, -0.00000613, 0.00000685, 0.00000686, -0.00000963, 0.00000841, -0.00001039, 0.00000908, -0.00000693, -0.00000636, 0.00000240, -0.00000132, -0.00000614, -0.00000189, -0.00000187, -0.00000179, 0.00000096, -0.00000815, -0.00000278, 0.00000189, -0.00000199, -0.00000215, -0.00000423, -0.00000274, -0.00000135, -0.00001199, -0.00000863, 0.00000007, -0.00001283, 0.00000035, -0.00001949, -0.00002561, -0.00001219, 0.00004844, 0.00002789, 0.00001188, 0.00001932, 0.00000618, 0.00002510, 0.00001820, 0.00004130, 0.00003174, -0.00003027, -0.00003470, -0.00000219, -0.00000717, -0.00000639, -0.00001020, -0.00000048, -0.00000211, 0.00000431, -0.00000292, 0.00000257, -0.00000174, -0.00000157, -0.00000706, 0.00000342, 0.00000098, },
		{ -0.00000105, 0.00000014, -0.00000239, 0.00000082, 0.00000269, -0.00000115, -0.00000011, -0.00000283, 0.00000076, -0.00000025, 0.00000324, -0.00000142, 0.00000371, -0.00000535, -0.00000192, 0.00000063, -0.00000031, -0.00000643, 0.00000113, -0.00000448, 0.00000327, -0.00000047, 0.00000067, -0.00000601, 0.00000086, -0.00002275, -0.00000508, -0.00002008, 0.00000140, -0.00010207, -0.00000133, 0.00011869, 0.00003634, 0.00005083, 0.00003329, 0.00005528, 0.00011288, -0.00005497, -0.00006511, -0.00000477, -0.00002485, 0.00000034, -0.00000702, -0.00000712, -0.00002098, 0.00000641, -0.00000652, -0.00000647, -0.00000667, 0.00000160, -0.00000113, -0.00000070, -0.00000954, 0.00000514, -0.00000280, -0.00000177, -0.00000568, 0.00000262, -0.00000425, -0.00000192, -0.00000953, 0.00000760, 0.00000251, -0.00001057, -0.00000248, 0.00000325, 0.00000160, -0.00000327, -0.00000685, 0.00001307, -0.00000259, -0.00000196, -0.00000654, 0.00001516, -0.00000566, -0.00000035, 0.00000178, -0.00000607, 0.00000900, -0.00001327, -0.00000002, 0.00000096, -0.00000273, -0.00000645, 0.00000043, -0.00000387, 0.00000112, -0.00000334, -0.00000763, 0.00000102, 0.00000008, -0.00000221, -0.00000307, -0.00000423, -0.00000089, -0.00000526, -0.00001424, -0.00000032, -0.00000841, -0.00000591, -0.00000531, -0.00002669, -0.00001994, 0.00001309, 0.00005213, 0.00000711, 0.00002244, 0.00000682, 0.00001585, 0.00002065, 0.00002399, 0.00004389, 0.00001122, -0.00004286, -0.00001805, -0.00000320, -0.00000550, -0.00000968, -0.00000591, -0.00000126, 0.00000129, 0.00000098, -0.00000099, 0.00000159, -0.00000150, -0.00000420, -0.00000436, 0.00000514, },
		{ 0.00000405, -0.00000209, 0.00000029, -0.00000304, 0.00000375, -0.00000009, 0.00000004, -0.00000203, -0.00000079, -0.00000058, 0.00000223, 0.00000085, 0.00000121, 0.00000041, -0.00000559, -0.00000072, 0.00000194, -0.00000433, -0.00000298, -0.00000124, -0.00000129, 0.00000095, 0.00000186, -0.00000428, -0.00000015, -0.00001012, -0.00001710, -0.00001100, -0.00000851, -0.00002834, -0.00010926, 0.00007689, 0.00009538, 0.00003128, 0.00005742, 0.00002535, 0.00010726, 0.00004959, -0.00010341, -0.00001543, -0.00002301, -0.00000831, -0.00000579, -0.00000161, -0.00002083, -0.00000534, 0.00000172, -0.00000665, -0.00000905, -0.00000069, -0.00000126, 0.00000184, -0.00000813, -0.00000144, 0.00000233, -0.00000215, -0.00000510, -0.00000126, 0.00000008, -0.00000311, -0.00000573, -0.00000421, 0.00000990, -0.00000429, -0.00001018, 0.00000223, 0.00000198, 0.00000200, -0.00000915, 0.00000282, 0.00000916, -0.00000416, -0.00000436, 0.00000037, 0.00001378, -0.00001231, 0.00000813, -0.00000711, 0.00000343, 0.00000076, -0.00001236, 0.00000393, -0.00000110, -0.00000445, -0.00000456, 0.00000048, -0.00000406, 0.00000218, -0.00000814, -0.00000305, 0.00000135, -0.00000106, -0.00000263, -0.00000396, -0.00000296, -0.00000063, -0.00001158, -0.00000839, -0.00000034, -0.00001264, 0.00000175, -0.00001983, -0.00002280, -0.00000987, 0.00003901, 0.00003239, 0.00000617, 0.00002219, 0.00000162, 0.00002379, 0.00001537, 0.00003363, 0.00003772, -0.00001209, -0.00004111, -0.00000587, -0.00000550, -0.00000477, -0.00001157, -0.00000176, -0.00000162, 0.00000311, -0.00000190, 0.00000161, -0.00000025, -0.00000136, -0.00000655, 0.00000037, },
		{ 0.00000497, 0.00000055, -0.00000106, -0.00000123, -0.00000108, 0.00000412, -0.00000181, 0.00000063, -0.00000301, 0.00000057, -0.00000117, 0.00000395, -0.00000098, 0.00000332, -0.00000416, -0.00000348, 0.00000014, 0.00000154, -0.00000686, 0.00000017, -0.00000319, 0.00000119, -0.00000098, 0.00000317, -0.00000741, 0.00000313, -0.00002115, -0.00000859, -0.00001776, 0.00000230, -0.00007786, -0.00006497, 0.00012817, 0.00005694, 0.00004665, 0.00004571, 0.00005106, 0.00012718, -0.00005573, -0.00007233, -0.00000500, -0.00002558, -0.00000071, -0.00000629, -0.00000715, -0.00002150, 0.00000539, -0.00000502, -0.00000660, -0.00000809, 0.00000221, -0.00000206, 0.00000019, -0.00000993, 0.00000503, -0.00000197, -0.00000162, -0.00000675, 0.00000274, -0.00000331, -0.00000222, -0.00000835, 0.00000357, 0.00000656, -0.00000957, -0.00000631, 0.00000450, 0.00000145, -0.00000098, -0.00000976, 0.00001114, 0.00000168, -0.00000311, -0.00000675, 0.00001054, 0.00000313, -0.00000844, 0.00000864, -0.00001045, 0.00001036, -0.00001020, -0.00000503, 0.00000330, -0.00000226, -0.00000615, -0.00000112, -0.00000196, -0.00000151, -0.00000098, -0.00000880, 0.00000125, -0.00000045, -0.00000106, -0.00000396, -0.00000371, -0.00000121, -0.00000407, -0.00001432, 0.00000001, -0.00000906, -0.00000496, -0.00000409, -0.00002746, -0.00001435, 0.00000788, 0.00004858, 0.00000941, 0.00001752, 0.00001010, 0.00000909, 0.00002229, 0.00001696, 0.00004028, 0.00002318, -0.00003072, -0.00002888, -0.00000165, -0.00000559, -0.00000689, -0.00000989, 0.00000036, -0.00000069, 0.00000230, -0.00000225, 0.00000271, -0.00000150, -0.00000244, -0.00000668, },
		{ -0.00000371, 0.00000685, -0.00000308, 0.00000098, -0.00000363, 0.00000311, 0.00000100, -0.00000063, -0.00000109, -0.00000140, -0.00000065, 0.00000114, 0.00000254, 0.00000036, 0.00000138, -0.00000572, -0.00000204, 0.00000204, -0.00000213, -0.00000519, -0.00000004, -0.00000187, -0.00000042, 0.00000162, -0.00000097, -0.00000383, -0.00000322, -0.00002256, -0.00000705, -0.00001464, -0.00000854, -0.00011505, 0.00001577, 0.00013465, 0.00002909, 0.00006710, 0.00002693, 0.00011251, 0.00006034, -0.00011085, -0.00001930, -0.00002259, -0.00000982, -0.00000559, -0.00000149, -0.00002055, -0.00000684, 0.00000220, -0.00000569, -0.00000976, -0.00000175, -0.00000044, 0.00000054, -0.00000645, -0.00000307, 0.00000375, -0.00000242, -0.00000441, -0.00000341, 0.00000212, -0.00000406, -0.00000406, -0.00000591, 0.00000826, 0.00000052, -0.00001208, -0.00000025, 0.00000366, 0.00000187, -0.00000679, -0.00000288, 0.00001204, -0.00000319, -0.00000354, -0.00000398, 0.00001562, -0.00000863, 0.00000187, 0.00000108, -0.00000508, 0.00000844, -0.00001600, 0.00000320, -0.00000035, -0.00000257, -0.00000674, 0.00000176, -0.00000483, 0.00000167, -0.00000685, -0.00000393, 0.00000119, -0.00000070, -0.00000228, -0.00000441, -0.00000291, -0.00000013, -0.00001091, -0.00000830, -0.00000040, -0.00001305, 0.00000362, -0.00002006, -0.00002125, -0.00000635, 0.00003012, 0.00003529, 0.00000301, 0.00002212, 0.00000055, 0.00001982, 0.00001545, 0.00002552, 0.00003926, 0.00000338, -0.00003830, -0.00001491, -0.00000215, -0.00000482, -0.00001005, -0.00000558, 0.00000062, 0.00000057, 0.00000029, -0.00000069, 0.00000215, -0.00000209, -0.00000445, },
		{ -0.00000661, 0.00000211, 0.00000389, -0.00000325, 0.00000064, -0.00000323, 0.00000523, -0.00000191, 0.00000073, -0.00000261, 0.00000013, -0.00000170, 0.00000366, 0.00000056, 0.00000210, -0.00000239, -0.00000500, -0.00000043, 0.00000236, -0.00000552, -0.00000233, -0.00000080, -0.00000105, -0.00000160, 0.00000411, -0.00000571, 0.00000098, -0.00001441, -0.00001655, -0.00001042, -0.00000634, -0.00004432, -0.00011301, 0.00010263, 0.00009682, 0.00003531, 0.00006325, 0.00004457, 0.00014424, -0.00005788, -0.00007788, -0.00000777, -0.00002453, -0.00000277, -0.00000505, -0.00000763, -0.00002137, 0.00000380, -0.00000350, -0.00000649, -0.00000922, 0.00000210, -0.00000206, -0.00000023, -0.00000899, 0.00000375, -0.00000032, -0.00000208, -0.00000662, 0.00000109, -0.00000086, -0.00000389, -0.00000577, -0.00000059, 0.00000870, -0.00000662, -0.00000970, 0.00000387, 0.00000261, 0.00000004, -0.00000985, 0.00000626, 0.00000685, -0.00000448, -0.00000521, 0.00000394, 0.00001124, -0.00001310, 0.00001023, -0.00000834, 0.00000526, -0.00000244, -0.00001180, 0.00000582, -0.00000271, -0.00000395, -0.00000437, 0.00000116, -0.00000470, 0.00000118, -0.00000963, 0.00000158, -0.00000145, 0.00000053, -0.00000493, -0.00000326, -0.00000184, -0.00000239, -0.00001491, 0.00000110, -0.00001041, -0.00000367, -0.00000300, -0.00002785, -0.00001066, 0.00000597, 0.00004226, 0.00001377, 0.00001156, 0.00001370, 0.00000329, 0.00002301, 0.00001225, 0.00003517, 0.00002923, -0.00001541, -0.00003485, -0.00000507, -0.00000357, -0.00000533, -0.00001154, -0.00000070, -0.00000040, 0.00000169, -0.00000175, 0.00000181, 0.00000014, -0.00000212, },
		{ -0.00000236, -0.00000686, 0.00000763, -0.00000211, 0.00000013, -0.00000276, 0.00000124, 0.00000272, -0.00000147, -0.00000033, -0.00000163, -0.00000083, 0.00000008, 0.00000355, 0.00000069, 0.00000105, -0.00000450, -0.00000405, 0.00000217, -0.00000060, -0.00000565, -0.00000103, -0.00000015, -0.00000316, 0.00000148, 0.00000158, -0.00000536, -0.00000039, -0.00002180, -0.00001001, -0.00001261, -0.00000546, -0.00009114, -0.00005936, 0.00015669, 0.00004456, 0.00006686, 0.00003657, 0.00011297, 0.00007609, -0.00012176, -0.00002033, -0.00002580, -0.00000799, -0.00000803, 0.00000057, -0.00002182, -0.00000677, 0.00000110, -0.00000375, -0.00001068, -0.00000263, 0.00000001, -0.00000027, -0.00000571, -0.00000353, 0.00000390, -0.00000170, -0.00000423, -0.00000472, 0.00000249, -0.00000314, -0.00000445, -0.00000480, 0.00000419, 0.00000539, -0.00001224, -0.00000340, 0.00000432, 0.00000259, -0.00000491, -0.00000635, 0.00001108, 0.00000059, -0.00000451, -0.00000488, 0.00001192, -0.00000001, -0.00000729, 0.00000979, -0.00001140, 0.00001165, -0.00001449, -0.00000097, 0.00000176, -0.00000187, -0.00000678, 0.00000053, -0.00000307, -0.00000101, -0.00000423, -0.00000551, 0.00000177, -0.00000142, -0.00000070, -0.00000575, -0.00000239, -0.00000025, -0.00000946, -0.00000899, 0.00000067, -0.00001476, 0.00000646, -0.00002073, -0.00001984, -0.00000359, 0.00002406, 0.00003453, 0.00000386, 0.00001831, 0.00000305, 0.00001343, 0.00001789, 0.00001716, 0.00003907, 0.00001290, -0.00002738, -0.00002541, -0.00000016, -0.00000478, -0.00000736, -0.00000982, 0.00000254, -0.00000158, 0.00000220, -0.00000280, 0.00000388, -0.00000234, },
		{ -0.00000303, -0.00000451, -0.00000238, 0.00000756, -0.00000512, 0.00000205, -0.00000456, 0.00000520, -0.00000101, 0.00000012, -0.00000180, -0.00000041, -0.00000190, 0.00000258, 0.00000254, 0.00000098, -0.00000108, -0.00000560, -0.00000160, 0.00000269, -0.00000351, -0.00000474, 0.00000083, -0.00000179, -0.00000329, 0.00000421, -0.00000235, -0.00000287, -0.00000680, -0.00002350, -0.00000463, -0.00001425, -0.00001904, -0.00012759, 0.00003910, 0.00014614, 0.00002632, 0.00007938, 0.00004051, 0.00016083, -0.00005856, -0.00008448, -0.00001017, -0.00002477, -0.00000314, -0.00000537, -0.00000697, -0.00002201, 0.00000300, -0.00000307, -0.00000550, -0.00001036, 0.00000155, -0.00000147, -0.00000141, -0.00000754, 0.00000227, 0.00000126, -0.00000253, -0.00000584, -0.00000138, 0.00000172, -0.00000531, -0.00000376, -0.00000285, 0.00000787, -0.00000203, -0.00001210, 0.00000177, 0.00000396, 0.00000082, -0.00000857, 0.00000090, 0.00001023, -0.00000399, -0.00000423, -0.00000109, 0.00001464, -0.00001111, 0.00000483, 0.00000003, -0.00000410, 0.00000662, -0.00001720, 0.00000666, -0.00000296, -0.00000122, -0.00000748, 0.00000346, -0.00000646, 0.00000158, -0.00000913, 0.00000127, -0.00000208, 0.00000153, -0.00000500, -0.00000360, -0.00000224, -0.00000083, -0.00001538, 0.00000220, -0.00001150, -0.00000304, -0.00000127, -0.00002836, -0.00000819, 0.00000577, 0.00003570, 0.00001713, 0.00000772, 0.00001464, 0.00000091, 0.00002076, 0.00001121, 0.00002796, 0.00003271, -0.00000314, -0.00003195, -0.00001403, 0.00000037, -0.00000545, -0.00001025, -0.00000451, 0.00000193, -0.00000067, 0.00000050, -0.00000099, 0.00000309, },
		{ -0.00000086, -0.00000166, -0.00000760, 0.00000549, 0.00000101, -0.00000214, -0.00000079, -0.00000119, 0.00000425, -0.00000173, -0.00000028, -0.00000118, -0.00000139, -0.00000045, 0.00000325, 0.00000255, -0.00000060, -0.00000220, -0.00000645, 0.00000245, -0.00000026, -0.00000403, -0.00000409, 0.00000326, -0.00000650, 0.00000144, 0.00000231, -0.00000330, -0.00000280, -0.00001457, -0.00001926, -0.00000328, -0.00001553, -0.00005208, -0.00012260, 0.00014117, 0.00008409, 0.00005585, 0.00005427, 0.00010876, 0.00009736, -0.00013678, -0.00001799, -0.00003274, -0.00000344, -0.00001223, 0.00000383, -0.00002416, -0.00000552, -0.00000110, -0.00000135, -0.00001161, -0.00000305, -0.00000044, 0.00000011, -0.00000660, -0.00000229, 0.00000252, 0.00000023, -0.00000504, -0.00000440, 0.00000055, -0.00000010, -0.00000681, -0.00000164, -0.00000075, 0.00000891, -0.00001056, -0.00000622, 0.00000320, 0.00000438, -0.00000389, -0.00000720, 0.00000690, 0.00000581, -0.00000619, -0.00000344, 0.00000524, 0.00000927, -0.00001414, 0.00001373, -0.00001112, 0.00000783, -0.00000737, -0.00000791, 0.00000488, -0.00000268, -0.00000434, -0.00000305, 0.00000055, -0.00000479, -0.00000141, -0.00000706, 0.00000273, -0.00000291, 0.00000161, -0.00000738, -0.00000181, -0.00000086, -0.00000727, -0.00001055, 0.00000295, -0.00001768, 0.00001007, -0.00002178, -0.00001805, -0.00000257, 0.00002148, 0.00003032, 0.00000766, 0.00001248, 0.00000693, 0.00000684, 0.00002048, 0.00001022, 0.00003726, 0.00001783, -0.00001389, -0.00003133, -0.00000307, -0.00000259, -0.00000570, -0.00001204, 0.00000204, -0.00000171, 0.00000236, -0.00000314, 0.00000342, },
		{ 0.00000573, -0.00000378, -0.00000196, -0.00000650, 0.00000962, -0.00000520, 0.00000191, -0.00000431, 0.00000373, 0.00000069, -0.00000073, -0.00000115, -0.00000054, -0.00000224, 0.00000151, 0.00000378, 0.00000123, -0.00000141, -0.00000439, -0.00000397, 0.00000343, -0.00000236, -0.00000521, -0.00000013, 0.00000036, -0.00000691, 0.00000469, 0.00000018, -0.00000456, -0.00000336, -0.00002384, -0.00000659, -0.00001285, -0.00001308, -0.00010541, -0.00004760, 0.00018287, 0.00003428, 0.00008561, 0.00004474, 0.00017274, -0.00005376, -0.00009603, -0.00000842, -0.00002972, 0.00000073, -0.00000905, -0.00000373, -0.00002477, 0.00000426, -0.00000471, -0.00000307, -0.00001202, 0.00000133, -0.00000130, -0.00000207, -0.00000709, 0.00000206, 0.00000149, -0.00000189, -0.00000564, -0.00000296, 0.00000246, -0.00000456, -0.00000415, -0.00000183, 0.00000387, 0.00000349, -0.00001336, -0.00000066, 0.00000400, 0.00000257, -0.00000747, -0.00000270, 0.00001000, -0.00000059, -0.00000514, -0.00000257, 0.00001196, -0.00000294, -0.00000514, 0.00001054, -0.00001253, 0.00001205, -0.00001780, 0.00000398, -0.00000166, 0.00000023, -0.00000837, 0.00000304, -0.00000534, -0.00000058, -0.00000696, -0.00000001, -0.00000176, 0.00000122, -0.00000356, -0.00000503, -0.00000208, 0.00000008, -0.00001503, 0.00000240, -0.00001125, -0.00000408, 0.00000190, -0.00002977, -0.00000558, 0.00000494, 0.00003174, 0.00001677, 0.00000824, 0.00001126, 0.00000288, 0.00001497, 0.00001397, 0.00001861, 0.00003593, 0.00000323, -0.00002118, -0.00002490, 0.00000333, -0.00000566, -0.00000747, -0.00000924, 0.00000446, -0.00000324, 0.00000314, -0.00000418, },
		{ 0.00000051, 0.00000234, -0.00000249, -0.00000592, 0.00000109, 0.00000513, -0.00000459, 0.00000107, -0.00000300, 0.00000464, -0.00000086, -0.00000108, -0.00000016, -0.00000219, -0.00000047, 0.00000187, 0.00000524, -0.00000249, -0.00000008, -0.00000806, 0.00000213, -0.00000041, -0.00000152, -0.00000743, 0.00000622, -0.00000842, 0.00000031, 0.00000203, 0.00000093, -0.00000779, -0.00000524, -0.00002898, 0.00000611, -0.00002716, -0.00001878, -0.00015038, 0.00007969, 0.00014050, 0.00004282, 0.00007366, 0.00010471, 0.00012084, -0.00015297, -0.00001523, -0.00004023, 0.00000046, -0.00001519, 0.00000594, -0.00002570, -0.00000476, -0.00000277, 0.00000014, -0.00001175, -0.00000341, -0.00000150, 0.00000139, -0.00000864, -0.00000016, 0.00000056, 0.00000247, -0.00000620, -0.00000300, -0.00000275, 0.00000358, -0.00000935, 0.00000138, -0.00000413, 0.00000947, -0.00000691, -0.00000843, 0.00000088, 0.00000605, -0.00000281, -0.00000662, 0.00000173, 0.00000985, -0.00000644, -0.00000207, -0.00000063, 0.00001449, -0.00001440, 0.00000995, -0.00000332, -0.00000169, 0.00000249, -0.00001470, 0.00000712, -0.00000386, -0.00000082, -0.00000708, 0.00000385, -0.00000747, -0.00000017, -0.00000741, 0.00000317, -0.00000421, 0.00000350, -0.00000817, -0.00000196, -0.00000139, -0.00000490, -0.00001236, 0.00000566, -0.00002075, 0.00001333, -0.00002237, -0.00001637, -0.00000269, 0.00002086, 0.00002524, 0.00001112, 0.00000818, 0.00000866, 0.00000325, 0.00002026, 0.00000719, 0.00003216, 0.00002189, -0.00000404, -0.00002845, -0.00001188, 0.00000188, -0.00000581, -0.00001127, -0.00000140, 0.00000043, 0.00000047, -0.00000123, },
		{ -0.00000581, 0.00000622, -0.00000287, -0.00000059, -0.00000843, 0.00000871, -0.00000285, 0.00000005, -0.00000268, 0.00000143, 0.00000233, -0.00000094, -0.00000133, 0.00000019, -0.00000316, 0.00000117, 0.00000338, 0.00000354, -0.00000368, -0.00000139, -0.00000739, 0.00000481, -0.00000278, -0.00000308, -0.00000391, 0.00000509, -0.00001187, 0.00000578, 0.00000041, -0.00000196, -0.00000620, -0.00001610, -0.00001709, -0.00000130, -0.00002485, -0.00006038, -0.00012869, 0.00018145, 0.00007008, 0.00007809, 0.00005938, 0.00017856, -0.00004164, -0.00011437, -0.00000094, -0.00004033, 0.00000893, -0.00001577, 0.00000186, -0.00002960, 0.00000759, -0.00000825, 0.00000052, -0.00001421, 0.00000192, -0.00000233, -0.00000116, -0.00000873, 0.00000400, -0.00000028, 0.00000042, -0.00000684, -0.00000243, 0.00000026, -0.00000098, -0.00000726, 0.00000208, -0.00000195, 0.00000846, -0.00001320, -0.00000259, 0.00000211, 0.00000539, -0.00000699, -0.00000387, 0.00000630, 0.00000478, -0.00000727, -0.00000109, 0.00000520, 0.00000741, -0.00001429, 0.00001722, -0.00001468, 0.00001023, -0.00001215, -0.00000233, 0.00000138, -0.00000037, -0.00000624, -0.00000038, -0.00000166, -0.00000446, -0.00000411, -0.00000165, -0.00000072, -0.00000022, -0.00000101, -0.00000703, -0.00000170, 0.00000035, -0.00001368, 0.00000136, -0.00000929, -0.00000697, 0.00000654, -0.00003224, -0.00000203, 0.00000209, 0.00003159, 0.00001228, 0.00001260, 0.00000482, 0.00000740, 0.00000762, 0.00001834, 0.00000921, 0.00003815, 0.00000561, -0.00000809, -0.00003162, 0.00000154, -0.00000374, -0.00000544, -0.00001239, 0.00000499, -0.00000421, 0.00000446, },
		{ 0.00000186, -0.00000351, 0.00000576, -0.00000358, -0.00000322, -0.00000363, 0.00000821, -0.00000558, 0.00000153, -0.00000325, 0.00000345, 0.00000099, -0.00000221, 0.00000074, -0.00000260, -0.00000060, 0.00000038, 0.00000728, -0.00000295, 0.00000021, -0.00000750, 0.00000015, 0.00000018, -0.00000002, -0.00000872, 0.00000631, -0.00000694, -0.00000327, 0.00000237, 0.00000437, -0.00001080, 0.00000016, -0.00003211, 0.00000738, -0.00002890, -0.00000686, -0.00013404, -0.00001579, 0.00019098, 0.00004456, 0.00008423, 0.00010832, 0.00014103, -0.00016520, -0.00001703, -0.00004340, -0.00000071, -0.00001348, 0.00000436, -0.00002432, -0.00000644, -0.00000206, -0.00000076, -0.00001013, -0.00000449, -0.00000221, 0.00000236, -0.00001030, 0.00000105, -0.00000019, 0.00000342, -0.00000634, -0.00000204, -0.00000531, 0.00000548, -0.00000966, 0.00000181, -0.00000391, 0.00000624, -0.00000152, -0.00001012, -0.00000139, 0.00000605, -0.00000050, -0.00000608, -0.00000211, 0.00001055, -0.00000377, -0.00000259, -0.00000299, 0.00001328, -0.00000738, -0.00000010, 0.00000841, -0.00001186, 0.00000992, -0.00001742, 0.00000612, -0.00000359, 0.00000160, -0.00000904, 0.00000443, -0.00000710, -0.00000173, -0.00000589, 0.00000251, -0.00000447, 0.00000399, -0.00000730, -0.00000333, -0.00000140, -0.00000302, -0.00001357, 0.00000767, -0.00002263, 0.00001492, -0.00002135, -0.00001585, -0.00000232, 0.00001950, 0.00002269, 0.00001081, 0.00000846, 0.00000567, 0.00000458, 0.00001562, 0.00000925, 0.00002300, 0.00002750, -0.00000062, -0.00001744, -0.00002335, 0.00000594, -0.00000630, -0.00000863, -0.00000632, 0.00000337, -0.00000230, },
		{ 0.00000358, -0.00000487, 0.00000381, 0.00000010, -0.00000092, -0.00000766, 0.00000504, 0.00000110, -0.00000250, -0.00000077, -0.00000050, 0.00000285, 0.00000014, -0.00000259, 0.00000173, -0.00000456, 0.00000164, 0.00000145, 0.00000727, -0.00000680, 0.00000198, -0.00001045, 0.00000586, -0.00000399, 0.00000043, -0.00000867, 0.00001007, -0.00001591, 0.00000610, -0.00000087, 0.00000370, -0.00001296, -0.00000441, -0.00003034, 0.00001296, -0.00004190, -0.00001657, -0.00017429, 0.00012677, 0.00013054, 0.00006382, 0.00007892, 0.00018262, -0.00002505, -0.00013701, 0.00000963, -0.00005357, 0.00001809, -0.00002249, 0.00000740, -0.00003459, 0.00001125, -0.00001197, 0.00000376, -0.00001602, 0.00000294, -0.00000439, 0.00000119, -0.00001215, 0.00000744, -0.00000322, 0.00000362, -0.00000898, -0.00000015, -0.00000408, 0.00000404, -0.00001132, 0.00000657, -0.00000688, 0.00001085, -0.00001103, -0.00000404, -0.00000096, 0.00000796, -0.00000621, -0.00000355, 0.00000111, 0.00000965, -0.00000853, 0.00000094, -0.00000169, 0.00001471, -0.00001734, 0.00001586, -0.00000833, 0.00000140, -0.00000233, -0.00000977, 0.00000448, -0.00000203, -0.00000231, -0.00000501, 0.00000237, -0.00000771, -0.00000252, -0.00000234, 0.00000006, -0.00000176, 0.00000143, -0.00000838, -0.00000192, 0.00000049, -0.00001181, -0.00000042, -0.00000628, -0.00001076, 0.00001157, -0.00003496, 0.00000205, -0.00000235, 0.00003390, 0.00000616, 0.00001750, -0.00000094, 0.00001058, 0.00000243, 0.00002059, 0.00000330, 0.00003659, 0.00000840, 0.00000081, -0.00002949, -0.00000654, 0.00000083, -0.00000524, -0.00001257, 0.00000251, -0.00000282, },
		{ -0.00000405, 0.00000427, -0.00000647, 0.00000728, -0.00000324, -0.00000130, -0.00000658, 0.00000853, -0.00000409, 0.00000010, -0.00000190, 0.00000119, 0.00000291, -0.00000266, 0.00000067, -0.00000196, -0.00000150, -0.00000018, 0.00000735, -0.00000083, -0.00000212, -0.00000441, -0.00000363, 0.00000202, -0.00000059, -0.00000681, 0.00000273, -0.00000183, -0.00000915, 0.00000400, 0.00000494, -0.00000879, -0.00000182, -0.00002551, -0.00000227, -0.00001735, -0.00001791, -0.00008542, -0.00011465, 0.00020487, 0.00007685, 0.00007866, 0.00012428, 0.00015463, -0.00017006, -0.00002680, -0.00003926, -0.00000915, -0.00000597, -0.00000147, -0.00001950, -0.00001123, 0.00000162, -0.00000439, -0.00000665, -0.00000662, -0.00000177, 0.00000188, -0.00001010, -0.00000021, 0.00000161, 0.00000197, -0.00000443, -0.00000283, -0.00000538, 0.00000388, -0.00000645, -0.00000123, 0.00000007, 0.00000018, 0.00000428, -0.00001100, -0.00000280, 0.00000373, 0.00000309, -0.00000605, -0.00000366, 0.00000758, 0.00000139, -0.00000483, -0.00000181, 0.00000691, 0.00000350, -0.00001114, 0.00001773, -0.00001651, 0.00001031, -0.00001360, 0.00000091, -0.00000104, 0.00000157, -0.00000755, 0.00000145, -0.00000352, -0.00000559, -0.00000324, 0.00000115, -0.00000372, 0.00000303, -0.00000495, -0.00000553, -0.00000109, -0.00000179, -0.00001374, 0.00000832, -0.00002256, 0.00001425, -0.00001834, -0.00001705, -0.00000026, 0.00001547, 0.00002456, 0.00000558, 0.00001360, -0.00000153, 0.00000972, 0.00000791, 0.00001463, 0.00001185, 0.00003351, -0.00000138, -0.00000388, -0.00003135, 0.00000562, -0.00000483, -0.00000635, -0.00001022, 0.00000485, },
		{ -0.00000010, 0.00000123, -0.00000196, -0.00000049, 0.00000387, -0.00000197, -0.00000529, 0.00000030, 0.00000468, -0.00000404, -0.00000001, -0.00000095, 0.00000167, 0.00000241, -0.00000434, 0.00000330, -0.00000572, 0.00000216, -0.00000088, 0.00001065, -0.00000868, 0.00000358, -0.00001127, 0.00000506, -0.00000433, 0.00000301, -0.00001155, 0.00001222, -0.00001634, 0.00000370, -0.00000154, 0.00000893, -0.00001837, 0.00000414, -0.00003776, 0.00001966, -0.00004959, 0.00000508, -0.00016886, 0.00002605, 0.00019347, 0.00006098, 0.00009133, 0.00019388, -0.00001061, -0.00015801, 0.00001743, -0.00006365, 0.00002289, -0.00002494, 0.00000960, -0.00003696, 0.00001265, -0.00001344, 0.00000466, -0.00001611, 0.00000333, -0.00000633, 0.00000362, -0.00001562, 0.00001035, -0.00000529, 0.00000586, -0.00001043, 0.00000210, -0.00000815, 0.00000768, -0.00001339, 0.00000863, -0.00000820, 0.00000912, -0.00000655, -0.00000545, -0.00000369, 0.00000853, -0.00000393, -0.00000319, -0.00000320, 0.00001156, -0.00000700, 0.00000126, -0.00000537, 0.00001554, -0.00001229, 0.00000667, 0.00000382, -0.00000983, 0.00000653, -0.00001418, 0.00000491, -0.00000263, 0.00000093, -0.00000795, 0.00000390, -0.00000793, -0.00000381, -0.00000110, -0.00000027, -0.00000233, 0.00000252, -0.00000803, -0.00000343, 0.00000107, -0.00001014, -0.00000201, -0.00000348, -0.00001387, 0.00001542, -0.00003651, 0.00000535, -0.00000652, 0.00003569, 0.00000228, 0.00001883, -0.00000216, 0.00000894, 0.00000239, 0.00001793, 0.00000335, 0.00002934, 0.00001461, 0.00000238, -0.00001867, -0.00001823, 0.00000570, -0.00000576, -0.00001052, -0.00000197, },
		{ 0.00000345, -0.00000359, 0.00000460, -0.00000676, 0.00000565, -0.00000066, -0.00000112, -0.00000675, 0.00000575, -0.00000051, -0.00000246, 0.00000007, -0.00000095, 0.00000372, -0.00000163, -0.00000073, -0.00000020, -0.00000324, 0.00000043, 0.00000534, 0.00000346, -0.00000611, -0.00000012, -0.00000778, 0.00000410, -0.00000247, -0.00000281, -0.00000279, 0.00000457, -0.00001499, 0.00000553, 0.00000321, -0.00000254, -0.00000938, -0.00001264, -0.00001762, 0.00000053, -0.00003942, -0.00003113, -0.00018045, 0.00015921, 0.00014187, 0.00006109, 0.00014886, 0.00016491, -0.00016955, -0.00004292, -0.00002967, -0.00002239, 0.00000437, -0.00000880, -0.00001339, -0.00001739, 0.00000672, -0.00000913, -0.00000275, -0.00000896, -0.00000041, -0.00000014, -0.00000784, -0.00000398, 0.00000566, -0.00000135, -0.00000095, -0.00000522, -0.00000296, -0.00000074, -0.00000081, -0.00000616, 0.00000560, -0.00000595, 0.00000819, -0.00001014, -0.00000360, 0.00000005, 0.00000651, -0.00000562, -0.00000366, 0.00000262, 0.00000676, -0.00000685, 0.00000069, -0.00000076, 0.00001266, -0.00001703, 0.00001915, -0.00001209, 0.00000265, -0.00000434, -0.00000669, 0.00000254, -0.00000028, -0.00000349, -0.00000358, 0.00000119, -0.00000942, -0.00000145, 0.00000043, -0.00000294, 0.00000162, -0.00000234, -0.00000735, -0.00000129, -0.00000077, -0.00001319, 0.00000787, -0.00002094, 0.00001203, -0.00001418, -0.00001934, 0.00000334, 0.00000885, 0.00002990, -0.00000246, 0.00002053, -0.00000929, 0.00001474, 0.00000110, 0.00001915, 0.00000305, 0.00003605, -0.00000111, 0.00000524, -0.00003059, -0.00000129, -0.00000042, -0.00000583, -0.00001135, },
		{ -0.00000607, 0.00000182, -0.00000083, 0.00000107, -0.00000443, 0.00000620, -0.00000198, -0.00000318, -0.00000324, 0.00000588, -0.00000319, -0.00000120, 0.00000036, -0.00000083, 0.00000494, -0.00000544, 0.00000384, -0.00000571, 0.00000178, -0.00000220, 0.00001191, -0.00000760, 0.00000201, -0.00000884, 0.00000173, -0.00000282, 0.00000307, -0.00001073, 0.00000996, -0.00001216, -0.00000197, -0.00000025, 0.00001083, -0.00001821, 0.00000449, -0.00003410, 0.00001335, -0.00004106, -0.00000164, -0.00012032, -0.00008971, 0.00022361, 0.00009041, 0.00008545, 0.00022003, -0.00000383, -0.00017219, 0.00001742, -0.00006601, 0.00001968, -0.00002076, 0.00000686, -0.00003524, 0.00001030, -0.00001128, 0.00000223, -0.00001382, 0.00000234, -0.00000698, 0.00000456, -0.00001719, 0.00001064, -0.00000460, 0.00000549, -0.00000964, 0.00000246, -0.00000958, 0.00000751, -0.00001146, 0.00000664, -0.00000502, 0.00000359, -0.00000067, -0.00000672, -0.00000512, 0.00000632, -0.00000003, -0.00000332, -0.00000537, 0.00000969, -0.00000258, -0.00000056, -0.00000501, 0.00001020, -0.00000171, -0.00000552, 0.00001531, -0.00001677, 0.00000897, -0.00001215, 0.00000087, -0.00000065, 0.00000152, -0.00000720, 0.00000144, -0.00000438, -0.00000794, 0.00000168, -0.00000159, -0.00000169, 0.00000192, -0.00000589, -0.00000604, 0.00000205, -0.00000902, -0.00000274, -0.00000188, -0.00001516, 0.00001706, -0.00003606, 0.00000687, -0.00000870, 0.00003436, 0.00000334, 0.00001450, 0.00000244, 0.00000200, 0.00000736, 0.00001079, 0.00000844, 0.00001794, 0.00002311, -0.00000107, -0.00000457, -0.00002742, 0.00000677, -0.00000464, -0.00000836, },
		{ -0.00000995, -0.00000034, -0.00000139, 0.00000299, -0.00000442, 0.00000138, 0.00000329, -0.00000208, -0.00000480, 0.00000119, 0.00000341, -0.00000451, 0.00000122, -0.00000173, 0.00000273, 0.00000084, -0.00000289, 0.00000184, -0.00000497, 0.00000138, 0.00000250, 0.00000793, -0.00000932, 0.00000275, -0.00000987, 0.00000437, -0.00000361, 0.00000059, -0.00000690, 0.00000855, -0.00001736, 0.00000435, 0.00000183, 0.00000399, -0.00001669, -0.00000153, -0.00002870, 0.00001243, -0.00005349, 0.00000188, -0.00018913, 0.00005436, 0.00021975, 0.00005001, 0.00016917, 0.00018188, -0.00017117, -0.00005886, -0.00002105, -0.00003403, 0.00001149, -0.00001268, -0.00000987, -0.00002163, 0.00001015, -0.00001207, -0.00000089, -0.00000979, 0.00000058, -0.00000243, -0.00000494, -0.00000845, 0.00000978, -0.00000435, 0.00000209, -0.00000745, 0.00000006, -0.00000583, 0.00000411, -0.00000970, 0.00000919, -0.00000879, 0.00000795, -0.00000664, -0.00000465, -0.00000316, 0.00000770, -0.00000342, -0.00000356, -0.00000192, 0.00000966, -0.00000638, 0.00000184, -0.00000574, 0.00001565, -0.00001438, 0.00001145, -0.00000010, -0.00000925, 0.00000561, -0.00001252, 0.00000427, -0.00000167, 0.00000051, -0.00000753, 0.00000384, -0.00001040, -0.00000252, 0.00000163, -0.00000321, 0.00000105, -0.00000095, -0.00000748, -0.00000283, 0.00000068, -0.00001264, 0.00000724, -0.00001907, 0.00000992, -0.00001057, -0.00002120, 0.00000708, 0.00000158, 0.00003556, -0.00000909, 0.00002460, -0.00001303, 0.00001530, -0.00000084, 0.00001884, 0.00000046, 0.00003185, 0.00000415, 0.00000609, -0.00002045, -0.00001279, 0.00000504, -0.00000639, },
		{ -0.00000781, -0.00000775, 0.00000117, -0.00000101, 0.00000241, -0.00000588, 0.00000533, 0.00000030, -0.00000269, -0.00000410, 0.00000383, 0.00000009, -0.00000390, 0.00000263, -0.00000347, 0.00000643, -0.00000492, 0.00000278, -0.00000426, 0.00000022, -0.00000197, 0.00001052, -0.00000357, -0.00000212, -0.00000427, -0.00000287, -0.00000035, 0.00000112, -0.00000696, 0.00000463, -0.00000540, -0.00000875, 0.00000184, 0.00000947, -0.00001250, -0.00000261, -0.00002162, -0.00000205, -0.00002189, -0.00002532, -0.00005678, -0.00017769, 0.00018920, 0.00016102, 0.00006132, 0.00026016, -0.00000334, -0.00018017, 0.00000988, -0.00006126, 0.00000988, -0.00001212, 0.00000129, -0.00003102, 0.00000541, -0.00000661, -0.00000228, -0.00001033, 0.00000057, -0.00000628, 0.00000357, -0.00001620, 0.00000779, -0.00000097, 0.00000259, -0.00000666, 0.00000063, -0.00000783, 0.00000345, -0.00000613, 0.00000170, 0.00000081, -0.00000313, 0.00000421, -0.00000675, -0.00000561, 0.00000236, 0.00000397, -0.00000304, -0.00000592, 0.00000536, 0.00000278, -0.00000284, -0.00000252, 0.00000229, 0.00000878, -0.00001395, 0.00001956, -0.00001449, 0.00000265, -0.00000362, -0.00000668, 0.00000329, -0.00000044, -0.00000310, -0.00000394, 0.00000118, -0.00001274, 0.00000409, -0.00000261, -0.00000069, 0.00000051, -0.00000309, -0.00000856, 0.00000263, -0.00000811, -0.00000270, -0.00000154, -0.00001468, 0.00001685, -0.00003411, 0.00000686, -0.00000860, 0.00002944, 0.00000905, 0.00000600, 0.00001029, -0.00000702, 0.00001369, 0.00000295, 0.00001435, 0.00000718, 0.00002922, -0.00000359, 0.00000526, -0.00002819, 0.00000116, -0.00000041, },
		{ -0.00000573, -0.00000795, -0.00000442, 0.00000069, 0.00000095, -0.00000106, -0.00000337, 0.00000642, -0.00000249, -0.00000249, -0.00000290, 0.00000550, -0.00000447, 0.00000051, -0.00000070, 0.00000022, 0.00000378, -0.00000450, 0.00000285, -0.00000549, 0.00000140, 0.00000063, 0.00001025, -0.00000936, 0.00000204, -0.00000813, 0.00000145, -0.00000235, 0.00000107, -0.00000691, 0.00000748, -0.00001422, -0.00000104, 0.00000305, 0.00000678, -0.00001806, 0.00000083, -0.00002771, 0.00000965, -0.00004884, 0.00000123, -0.00014254, -0.00008039, 0.00027170, 0.00007111, 0.00017008, 0.00021658, -0.00018311, -0.00006738, -0.00002026, -0.00003773, 0.00001014, -0.00000933, -0.00001177, -0.00002137, 0.00000940, -0.00001091, -0.00000286, -0.00000781, -0.00000008, -0.00000335, -0.00000343, -0.00001115, 0.00001133, -0.00000454, 0.00000247, -0.00000743, 0.00000127, -0.00000836, 0.00000518, -0.00000901, 0.00000833, -0.00000655, 0.00000306, -0.00000080, -0.00000625, -0.00000460, 0.00000560, 0.00000087, -0.00000404, -0.00000447, 0.00000876, -0.00000272, 0.00000054, -0.00000636, 0.00001165, -0.00000459, -0.00000138, 0.00001330, -0.00001846, 0.00001017, -0.00001239, 0.00000160, -0.00000043, 0.00000188, -0.00000776, 0.00000225, -0.00000714, -0.00000692, 0.00000476, -0.00000473, 0.00000188, -0.00000143, -0.00000553, -0.00000577, 0.00000267, -0.00001258, 0.00000732, -0.00001821, 0.00000942, -0.00000893, -0.00002139, 0.00000952, -0.00000412, 0.00003828, -0.00001073, 0.00002269, -0.00001037, 0.00000975, 0.00000316, 0.00001291, 0.00000440, 0.00002144, 0.00001382, 0.00000065, -0.00000595, -0.00002302, 0.00000747, },
		{ 0.00000561, -0.00000866, -0.00000670, -0.00000215, 0.00000080, 0.00000152, -0.00000420, 0.00000118, 0.00000447, -0.00000371, -0.00000239, -0.00000067, 0.00000442, -0.00000664, 0.00000445, -0.00000496, 0.00000606, -0.00000269, 0.00000066, -0.00000221, -0.00000165, -0.00000103, 0.00000777, 0.00000134, -0.00000617, -0.00000051, -0.00000586, 0.00000062, -0.00000039, -0.00000327, 0.00000001, -0.00000037, -0.00001254, 0.00000153, 0.00000794, -0.00000552, -0.00001090, -0.00000899, -0.00001545, -0.00000610, -0.00004498, -0.00001159, -0.00020379, 0.00008132, 0.00025739, 0.00003625, 0.00030164, 0.00000114, -0.00018991, 0.00000168, -0.00005608, 0.00000027, -0.00000552, -0.00000170, -0.00002861, 0.00000166, -0.00000289, -0.00000560, -0.00000846, 0.00000002, -0.00000567, 0.00000192, -0.00001401, 0.00000352, 0.00000345, -0.00000064, -0.00000351, -0.00000166, -0.00000475, -0.00000195, -0.00000061, -0.00000265, 0.00000544, -0.00000717, 0.00000521, -0.00000403, -0.00000646, -0.00000117, 0.00000565, -0.00000078, -0.00000633, 0.00000101, 0.00000623, -0.00000307, -0.00000086, -0.00000371, 0.00001376, -0.00001382, 0.00001374, -0.00000305, -0.00000972, 0.00000731, -0.00001390, 0.00000641, -0.00000273, 0.00000175, -0.00000908, 0.00000536, -0.00001507, 0.00000379, -0.00000182, -0.00000062, -0.00000024, -0.00000128, -0.00000950, 0.00000184, -0.00000674, -0.00000257, -0.00000160, -0.00001366, 0.00001641, -0.00003234, 0.00000677, -0.00000751, 0.00002270, 0.00001632, -0.00000234, 0.00001648, -0.00001303, 0.00001642, -0.00000078, 0.00001610, 0.00000221, 0.00002825, -0.00000002, 0.00000604, -0.00001891, -0.00001002, },
		{ 0.00000337, -0.00000196, -0.00000729, -0.00000653, 0.00000066, 0.00000023, 0.00000122, -0.00000609, 0.00000655, -0.00000078, -0.00000149, -0.00000462, 0.00000440, -0.00000181, -0.00000197, 0.00000155, -0.00000267, 0.00000576, -0.00000431, 0.00000198, -0.00000421, -0.00000015, 0.00000071, 0.00000937, -0.00000562, -0.00000225, -0.00000313, -0.00000383, 0.00000084, -0.00000134, -0.00000293, 0.00000204, -0.00000698, -0.00000902, 0.00000632, 0.00000509, -0.00001231, -0.00000640, -0.00001499, -0.00000610, -0.00002843, -0.00002438, -0.00006858, -0.00019655, 0.00025650, 0.00014129, 0.00014635, 0.00027208, -0.00020700, -0.00006689, -0.00002895, -0.00003235, 0.00000019, 0.00000035, -0.00001803, -0.00001730, 0.00000484, -0.00000599, -0.00000808, -0.00000369, -0.00000222, -0.00000242, -0.00000428, -0.00001083, 0.00000912, -0.00000112, -0.00000036, -0.00000458, -0.00000031, -0.00000706, 0.00000151, -0.00000390, 0.00000345, -0.00000048, -0.00000428, 0.00000516, -0.00000734, -0.00000457, 0.00000112, 0.00000572, -0.00000427, -0.00000528, 0.00000494, 0.00000261, -0.00000193, -0.00000400, 0.00000372, 0.00000694, -0.00001224, 0.00002062, -0.00001881, 0.00000565, -0.00000501, -0.00000552, 0.00000357, 0.00000006, -0.00000393, -0.00000314, -0.00000093, -0.00001279, 0.00000809, -0.00000638, 0.00000342, -0.00000310, -0.00000247, -0.00000903, 0.00000451, -0.00001283, 0.00000824, -0.00001875, 0.00001085, -0.00000929, -0.00002001, 0.00001048, -0.00000737, 0.00003683, -0.00000681, 0.00001541, -0.00000303, 0.00000060, 0.00001009, 0.00000459, 0.00001102, 0.00000957, 0.00002281, -0.00000464, 0.00000493, -0.00002548, },
		{ -0.00002135, 0.00000957, -0.00000881, -0.00000449, -0.00000643, 0.00000316, -0.00000055, -0.00000054, -0.00000443, 0.00000857, -0.00000477, 0.00000039, -0.00000553, 0.00000767, -0.00000776, 0.00000468, -0.00000472, 0.00000405, 0.00000019, -0.00000099, -0.00000116, -0.00000229, -0.00000105, 0.00000589, 0.00000420, -0.00000698, -0.00000061, -0.00000454, -0.00000221, 0.00000101, -0.00000286, -0.00000017, -0.00000089, -0.00001002, -0.00000390, 0.00000968, -0.00000263, -0.00001303, -0.00000544, -0.00001613, -0.00000620, -0.00004398, -0.00000499, -0.00016106, -0.00007514, 0.00033892, 0.00004002, 0.00032583, 0.00002385, -0.00021247, 0.00000231, -0.00005934, -0.00000093, -0.00000801, 0.00000328, -0.00003221, 0.00000282, -0.00000373, -0.00000435, -0.00001096, 0.00000276, -0.00000703, 0.00000176, -0.00001313, 0.00000082, 0.00000543, -0.00000113, -0.00000296, -0.00000178, -0.00000331, -0.00000502, 0.00000117, -0.00000273, 0.00000544, -0.00000574, 0.00000086, 0.00000187, -0.00000852, -0.00000244, 0.00000346, 0.00000412, -0.00000754, -0.00000146, 0.00000591, 0.00000011, -0.00000183, -0.00000519, 0.00001115, -0.00000512, 0.00000067, 0.00001208, -0.00002135, 0.00001429, -0.00001602, 0.00000552, -0.00000264, 0.00000428, -0.00001074, 0.00000529, -0.00001289, -0.00000024, 0.00000124, -0.00000201, 0.00000055, -0.00000148, -0.00000815, -0.00000063, -0.00000462, -0.00000298, -0.00000101, -0.00001361, 0.00001753, -0.00003246, 0.00000818, -0.00000717, 0.00001678, 0.00002133, -0.00000616, 0.00001686, -0.00001252, 0.00001266, 0.00000196, 0.00001145, 0.00000487, 0.00001927, 0.00001008, -0.00000102, -0.00000393, },
		{ -0.00001738, -0.00000715, 0.00000430, -0.00000828, -0.00000584, -0.00000216, 0.00000158, 0.00000123, -0.00000541, 0.00000293, 0.00000323, -0.00000222, -0.00000349, 0.00000034, 0.00000252, -0.00000502, 0.00000374, -0.00000464, 0.00000581, -0.00000212, -0.00000033, -0.00000191, -0.00000247, 0.00000199, 0.00000644, -0.00000009, -0.00000744, 0.00000201, -0.00000833, 0.00000319, -0.00000399, 0.00000185, -0.00000379, -0.00000029, -0.00001503, 0.00000802, 0.00000223, -0.00000378, -0.00001692, 0.00000059, -0.00002319, -0.00000737, -0.00005152, -0.00000692, -0.00024871, 0.00015352, 0.00025202, 0.00011231, 0.00033761, -0.00023366, -0.00006485, -0.00004065, -0.00002426, -0.00001173, 0.00000983, -0.00002310, -0.00001382, 0.00000024, -0.00000087, -0.00001310, -0.00000040, -0.00000372, -0.00000105, -0.00000634, -0.00000860, 0.00000462, 0.00000401, -0.00000437, -0.00000075, -0.00000314, -0.00000352, -0.00000458, 0.00000253, -0.00000195, 0.00000546, -0.00000991, 0.00000787, -0.00000592, -0.00000470, -0.00000321, 0.00000840, -0.00000244, -0.00000585, 0.00000060, 0.00000670, -0.00000292, -0.00000177, -0.00000334, 0.00001404, -0.00001502, 0.00001741, -0.00000873, -0.00000657, 0.00000649, -0.00001380, 0.00000788, -0.00000300, 0.00000160, -0.00000930, 0.00000479, -0.00001682, 0.00000906, -0.00000645, 0.00000429, -0.00000442, -0.00000007, -0.00001099, 0.00000511, -0.00001271, 0.00000942, -0.00001998, 0.00001314, -0.00001017, -0.00001860, 0.00001122, -0.00000918, 0.00003265, -0.00000009, 0.00000689, 0.00000407, -0.00000689, 0.00001464, -0.00000082, 0.00001462, 0.00000246, 0.00002503, -0.00000330, 0.00000625, },
		{ 0.00000855, -0.00002591, 0.00000745, -0.00000607, -0.00000321, -0.00000935, 0.00000402, -0.00000201, 0.00000290, -0.00000877, 0.00001017, -0.00000414, 0.00000243, -0.00000864, 0.00000809, -0.00000637, 0.00000313, -0.00000318, 0.00000154, 0.00000208, -0.00000062, -0.00000238, -0.00000066, -0.00000313, 0.00000646, 0.00000322, -0.00000309, -0.00000544, 0.00000136, -0.00000877, 0.00000561, -0.00000655, 0.00000497, -0.00000740, -0.00000150, -0.00001387, 0.00001521, -0.00000591, -0.00000601, -0.00001408, -0.00000161, -0.00002387, -0.00002112, -0.00003317, -0.00007478, -0.00022768, 0.00035494, 0.00009836, 0.00032139, 0.00007274, -0.00025355, 0.00001661, -0.00007542, 0.00000987, -0.00002186, 0.00001713, -0.00004226, 0.00000957, -0.00001000, 0.00000222, -0.00001821, 0.00000896, -0.00001090, 0.00000425, -0.00001523, 0.00000176, 0.00000294, 0.00000278, -0.00000642, 0.00000166, -0.00000529, -0.00000360, -0.00000264, 0.00000264, 0.00000027, 0.00000082, -0.00000739, 0.00000924, -0.00001106, -0.00000145, -0.00000197, 0.00001035, -0.00000888, -0.00000196, 0.00000215, 0.00000573, -0.00000472, -0.00000279, 0.00000317, 0.00000738, -0.00001263, 0.00002284, -0.00002504, 0.00001239, -0.00001056, -0.00000042, 0.00000078, 0.00000315, -0.00000780, 0.00000067, -0.00000685, -0.00000658, 0.00000513, -0.00000402, 0.00000245, -0.00000328, -0.00000521, -0.00000390, -0.00000229, -0.00000392, 0.00000063, -0.00001524, 0.00002094, -0.00003497, 0.00001140, -0.00000824, 0.00001315, 0.00002203, -0.00000388, 0.00001095, -0.00000610, 0.00000389, 0.00000916, 0.00000270, 0.00001217, 0.00000650, 0.00002168, -0.00000930, },
		{ -0.00000189, -0.00000207, -0.00001937, 0.00000958, -0.00000920, -0.00000344, -0.00000659, 0.00000408, -0.00000061, -0.00000186, -0.00000313, 0.00000798, -0.00000364, -0.00000053, -0.00000499, 0.00000659, -0.00000702, 0.00000471, -0.00000503, 0.00000414, 0.00000093, -0.00000235, -0.00000049, -0.00000364, 0.00000243, 0.00000412, 0.00000371, -0.00000959, 0.00000332, -0.00000834, 0.00000142, -0.00000345, 0.00000323, -0.00000513, 0.00000069, -0.00001429, 0.00000402, 0.00000299, 0.00000069, -0.00002154, 0.00000742, -0.00002771, -0.00000240, -0.00005743, 0.00001259, -0.00021811, -0.00002024, 0.00036330, 0.00010053, 0.00039202, -0.00024649, -0.00007472, -0.00004373, -0.00002422, -0.00001558, 0.00001029, -0.00002002, -0.00001651, 0.00000058, -0.00000029, -0.00001351, -0.00000168, -0.00000169, -0.00000179, -0.00000697, -0.00000742, 0.00000130, 0.00000710, -0.00000597, 0.00000075, -0.00000411, -0.00000118, -0.00000887, 0.00000568, -0.00000337, 0.00000693, -0.00001002, 0.00000483, -0.00000069, -0.00000655, -0.00000495, 0.00000674, 0.00000257, -0.00000746, -0.00000201, 0.00000715, -0.00000047, -0.00000215, -0.00000598, 0.00001327, -0.00000810, 0.00000498, 0.00000738, -0.00002024, 0.00001577, -0.00001821, 0.00000894, -0.00000425, 0.00000539, -0.00001251, 0.00000658, -0.00001633, 0.00000611, -0.00000408, 0.00000359, -0.00000418, 0.00000034, -0.00001058, 0.00000389, -0.00001174, 0.00001010, -0.00002073, 0.00001459, -0.00000953, -0.00001918, 0.00001354, -0.00001148, 0.00002859, 0.00000522, 0.00000212, 0.00000590, -0.00000813, 0.00001274, 0.00000038, 0.00001149, 0.00000365, 0.00001788, 0.00000650, },
		{ -0.00001110, 0.00001144, -0.00001972, -0.00000168, -0.00000036, -0.00000367, -0.00000959, 0.00000222, -0.00000163, 0.00000436, -0.00000980, 0.00000789, -0.00000109, 0.00000278, -0.00000889, 0.00000533, -0.00000293, 0.00000068, -0.00000148, -0.00000013, 0.00000187, 0.00000223, -0.00000565, 0.00000266, -0.00000669, 0.00000900, -0.00000081, 0.00000384, -0.00001259, 0.00000889, -0.00001595, 0.00001064, -0.00001176, 0.00001240, -0.00001607, 0.00000841, -0.00002374, 0.00002089, -0.00001228, 0.00000624, -0.00002924, 0.00001984, -0.00004789, 0.00000861, -0.00007158, 0.00000992, -0.00031882, 0.00027075, 0.00021215, 0.00029788, 0.00014042, -0.00030665, 0.00003899, -0.00009933, 0.00002756, -0.00004142, 0.00003398, -0.00005373, 0.00001793, -0.00001834, 0.00001095, -0.00002706, 0.00001583, -0.00001535, 0.00000821, -0.00001955, 0.00000568, -0.00000305, 0.00000969, -0.00001229, 0.00000723, -0.00000965, 0.00000128, -0.00001027, 0.00001082, -0.00000690, 0.00000873, -0.00001534, 0.00001447, -0.00001180, -0.00000003, -0.00000794, 0.00001492, -0.00000839, -0.00000194, -0.00000279, 0.00001096, -0.00000695, 0.00000041, -0.00000528, 0.00001696, -0.00001892, 0.00002322, -0.00001755, 0.00000144, 0.00000055, -0.00000896, 0.00000566, -0.00000016, -0.00000217, -0.00000582, -0.00000023, -0.00001190, 0.00000720, -0.00000484, 0.00000405, -0.00000514, -0.00000246, -0.00000629, -0.00000088, -0.00000476, 0.00000287, -0.00001811, 0.00002585, -0.00003860, 0.00001510, -0.00000971, 0.00001122, 0.00001929, 0.00000234, 0.00000239, 0.00000160, -0.00000479, 0.00001551, -0.00000467, 0.00001804, -0.00000311, 0.00002757, },
		{ 0.00002035, -0.00001316, 0.00001234, -0.00002621, 0.00000956, -0.00000770, -0.00000136, -0.00001042, 0.00000586, -0.00000270, 0.00000255, -0.00000896, 0.00001100, -0.00000402, 0.00000246, -0.00000940, 0.00000849, -0.00000689, 0.00000410, -0.00000422, 0.00000196, 0.00000294, -0.00000208, -0.00000174, -0.00000206, 0.00000029, 0.00000475, 0.00000306, -0.00000619, -0.00000124, -0.00000239, -0.00000556, 0.00000168, -0.00000082, 0.00000029, -0.00000606, -0.00000551, -0.00000668, 0.00000951, -0.00000290, -0.00001486, -0.00000071, -0.00001338, -0.00001988, -0.00003487, -0.00001377, -0.00012408, -0.00021085, 0.00041568, 0.00014604, 0.00041751, -0.00023245, -0.00010673, -0.00002951, -0.00003996, -0.00000472, -0.00000327, -0.00000555, -0.00002782, 0.00000832, -0.00000675, -0.00000703, -0.00000925, 0.00000516, -0.00000612, -0.00000413, -0.00000986, 0.00000215, 0.00000511, -0.00000246, -0.00000244, -0.00000088, -0.00000280, -0.00000812, 0.00000251, 0.00000156, 0.00000220, -0.00000379, -0.00000354, 0.00000742, -0.00000998, -0.00000352, 0.00000083, 0.00000983, -0.00000974, -0.00000232, 0.00000365, 0.00000519, -0.00000520, -0.00000391, 0.00000573, 0.00000473, -0.00001018, 0.00002132, -0.00002746, 0.00001690, -0.00001516, 0.00000472, -0.00000187, 0.00000534, -0.00001086, 0.00000329, -0.00001123, 0.00000008, -0.00000029, 0.00000184, -0.00000246, -0.00000121, -0.00000819, 0.00000148, -0.00001025, 0.00001008, -0.00002030, 0.00001409, -0.00000614, -0.00002271, 0.00001826, -0.00001542, 0.00002673, 0.00000620, 0.00000380, 0.00000076, -0.00000249, 0.00000458, 0.00000745, 0.00000264, 0.00001143, 0.00000448, },
		{ 0.00002327, -0.00000292, 0.00000337, -0.00000485, -0.00001406, 0.00000545, -0.00000463, -0.00000766, -0.00000169, 0.00000035, 0.00000353, -0.00000727, 0.00000231, 0.00000340, 0.00000192, -0.00000674, 0.00000051, 0.00000099, -0.00000126, -0.00000068, -0.00000032, -0.00000027, 0.00000635, -0.00000906, 0.00000552, -0.00000954, 0.00001118, -0.00000486, 0.00000983, -0.00001746, 0.00001322, -0.00001909, 0.00001184, -0.00001412, 0.00001697, -0.00002092, 0.00001322, -0.00002727, 0.00002115, -0.00001557, 0.00001550, -0.00003998, 0.00003430, -0.00006123, 0.00002412, -0.00009087, 0.00005213, -0.00031488, 0.00009145, 0.00034493, 0.00028916, 0.00020434, -0.00035406, 0.00005467, -0.00011811, 0.00004006, -0.00005527, 0.00004363, -0.00005832, 0.00002116, -0.00002276, 0.00001614, -0.00003216, 0.00001877, -0.00001676, 0.00001060, -0.00002307, 0.00000934, -0.00000880, 0.00001549, -0.00001660, 0.00001127, -0.00001293, 0.00000587, -0.00001714, 0.00001671, -0.00001094, 0.00001289, -0.00001828, 0.00001408, -0.00000855, -0.00000045, -0.00001134, 0.00001500, -0.00000440, -0.00000301, -0.00000630, 0.00001292, -0.00000594, 0.00000130, -0.00000978, 0.00001885, -0.00001487, 0.00001277, -0.00000183, -0.00001319, 0.00001137, -0.00001509, 0.00000830, -0.00000243, 0.00000253, -0.00001016, 0.00000305, -0.00001295, 0.00000537, -0.00000322, 0.00000415, -0.00000556, -0.00000156, -0.00000642, -0.00000122, -0.00000486, 0.00000491, -0.00002099, 0.00003045, -0.00004114, 0.00001705, -0.00000961, 0.00000896, 0.00001605, 0.00000807, -0.00000337, 0.00000488, -0.00000790, 0.00001590, -0.00000581, 0.00001736, -0.00000433, },
		{ -0.00000791, 0.00003028, -0.00001888, 0.00001771, -0.00002255, 0.00000230, -0.00000301, -0.00000086, -0.00001195, 0.00000538, -0.00000347, 0.00000563, -0.00001196, 0.00001035, -0.00000218, 0.00000413, -0.00001133, 0.00000732, -0.00000463, 0.00000251, -0.00000320, 0.00000069, 0.00000252, 0.00000130, -0.00000583, 0.00000212, -0.00000429, 0.00000838, -0.00000188, 0.00000182, -0.00000990, 0.00000709, -0.00001507, 0.00000905, -0.00000823, 0.00001008, -0.00001736, 0.00000729, -0.00001992, 0.00001875, -0.00001233, 0.00000045, -0.00001964, 0.00001327, -0.00005039, 0.00000294, -0.00006239, -0.00001547, -0.00034877, 0.00035821, 0.00026176, 0.00041721, -0.00019449, -0.00015827, -0.00000041, -0.00006914, 0.00001802, -0.00002707, 0.00001596, -0.00004391, 0.00002052, -0.00001787, 0.00000407, -0.00002071, 0.00001468, -0.00001268, 0.00000163, -0.00001584, 0.00000733, -0.00000184, 0.00000553, -0.00000951, 0.00000586, -0.00000811, -0.00000253, -0.00000607, 0.00001102, -0.00000629, 0.00000558, -0.00001339, 0.00001490, -0.00001270, -0.00000070, -0.00000667, 0.00001633, -0.00001071, -0.00000163, -0.00000182, 0.00001148, -0.00000858, 0.00000011, -0.00000389, 0.00001659, -0.00001999, 0.00002570, -0.00002335, 0.00000813, -0.00000527, -0.00000334, 0.00000300, 0.00000223, -0.00000558, -0.00000294, -0.00000443, -0.00000584, 0.00000233, 0.00000080, -0.00000058, -0.00000329, -0.00000549, -0.00000049, -0.00000934, 0.00000989, -0.00001890, 0.00001174, -0.00000042, -0.00002833, 0.00002437, -0.00002036, 0.00002700, 0.00000305, 0.00001055, -0.00000849, 0.00000609, -0.00000522, 0.00001545, -0.00000675, 0.00001978, },
		{ 0.00000276, 0.00001039, 0.00001200, -0.00001008, 0.00001109, -0.00002272, 0.00000699, -0.00000380, -0.00000537, -0.00000582, 0.00000241, 0.00000155, -0.00000275, -0.00000418, 0.00000705, 0.00000137, -0.00000369, -0.00000434, 0.00000353, -0.00000156, -0.00000117, 0.00000058, -0.00000298, 0.00000950, -0.00001007, 0.00000548, -0.00000933, 0.00001047, -0.00000571, 0.00001107, -0.00001625, 0.00001089, -0.00001522, 0.00000665, -0.00001078, 0.00001509, -0.00001789, 0.00000895, -0.00002109, 0.00001293, -0.00001148, 0.00001525, -0.00003785, 0.00003173, -0.00005280, 0.00001309, -0.00007587, 0.00003754, -0.00022508, -0.00013122, 0.00042984, 0.00034003, 0.00023986, -0.00037778, 0.00004896, -0.00011910, 0.00003607, -0.00005347, 0.00003805, -0.00005012, 0.00001452, -0.00001878, 0.00001343, -0.00002951, 0.00001449, -0.00001248, 0.00000874, -0.00002272, 0.00000917, -0.00001026, 0.00001604, -0.00001555, 0.00001034, -0.00001179, 0.00000638, -0.00001888, 0.00001601, -0.00000814, 0.00001019, -0.00001402, 0.00000724, -0.00000122, -0.00000337, -0.00001079, 0.00000991, 0.00000276, -0.00000531, -0.00000725, 0.00001058, -0.00000115, -0.00000105, -0.00000894, 0.00001282, -0.00000291, -0.00000297, 0.00001423, -0.00002331, 0.00001524, -0.00001433, 0.00000580, -0.00000112, 0.00000346, -0.00000963, 0.00000100, -0.00000888, -0.00000015, 0.00000032, 0.00000288, -0.00000425, -0.00000286, -0.00000428, -0.00000300, -0.00000432, 0.00000633, -0.00002290, 0.00003322, -0.00004086, 0.00001572, -0.00000660, 0.00000472, 0.00001498, 0.00000957, -0.00000256, 0.00000076, -0.00000340, 0.00000909, 0.00000006, 0.00000958, },
		{ 0.00002244, -0.00001256, 0.00002997, -0.00001370, 0.00001101, -0.00000895, -0.00000955, 0.00000272, -0.00000130, -0.00001105, 0.00000251, -0.00000248, 0.00000627, -0.00001103, 0.00000583, 0.00000157, 0.00000433, -0.00001052, 0.00000356, -0.00000132, 0.00000103, -0.00000291, 0.00000092, -0.00000020, 0.00000658, -0.00001079, 0.00000656, -0.00000893, 0.00001250, -0.00000790, 0.00001015, -0.00001762, 0.00001466, -0.00002168, 0.00001360, -0.00001385, 0.00001830, -0.00002632, 0.00001652, -0.00002823, 0.00002386, -0.00002030, 0.00001498, -0.00003706, 0.00003642, -0.00007421, 0.00003104, -0.00009745, 0.00005473, -0.00038211, 0.00017934, 0.00041864, 0.00042456, -0.00015572, -0.00021154, 0.00002857, -0.00009843, 0.00004008, -0.00004906, 0.00003349, -0.00005566, 0.00002973, -0.00002700, 0.00001348, -0.00003012, 0.00002167, -0.00001737, 0.00000691, -0.00002218, 0.00001352, -0.00000994, 0.00001383, -0.00001634, 0.00001230, -0.00001357, 0.00000411, -0.00001539, 0.00001968, -0.00001300, 0.00001245, -0.00001928, 0.00001743, -0.00001175, 0.00000060, -0.00001203, 0.00001856, -0.00000817, -0.00000189, -0.00000638, 0.00001513, -0.00000923, 0.00000244, -0.00001042, 0.00002147, -0.00001939, 0.00001814, -0.00000915, -0.00000646, 0.00000631, -0.00001058, 0.00000680, -0.00000079, -0.00000028, -0.00000799, -0.00000017, -0.00000799, 0.00000125, 0.00000206, -0.00000001, -0.00000418, -0.00000441, -0.00000038, -0.00001004, 0.00001026, -0.00001732, 0.00000867, 0.00000588, -0.00003379, 0.00002954, -0.00002425, 0.00002742, -0.00000145, 0.00001797, -0.00001623, 0.00001148, -0.00001057, 0.00001851, -0.00001084, },
		{ -0.00000085, 0.00001278, -0.00000411, 0.00002545, -0.00001965, 0.00001915, -0.00002160, 0.00000145, 0.00000013, -0.00000426, -0.00000824, 0.00000281, 0.00000013, 0.00000136, -0.00000865, 0.00000756, 0.00000259, -0.00000134, -0.00000734, 0.00000338, 0.00000012, -0.00000256, 0.00000161, -0.00000468, 0.00000988, -0.00000735, 0.00000178, -0.00000560, 0.00000632, -0.00000251, 0.00000675, -0.00000885, 0.00000258, -0.00000576, -0.00000317, -0.00000316, 0.00000779, -0.00000821, -0.00000252, -0.00000795, -0.00000087, -0.00000174, 0.00000604, -0.00002334, 0.00001363, -0.00002602, -0.00001863, -0.00003576, -0.00001402, -0.00009916, -0.00031650, 0.00039743, 0.00047900, 0.00024123, -0.00037511, 0.00001967, -0.00010057, 0.00001430, -0.00003566, 0.00001824, -0.00003122, -0.00000005, -0.00000761, 0.00000363, -0.00001994, 0.00000410, -0.00000360, 0.00000307, -0.00001810, 0.00000424, -0.00000619, 0.00001035, -0.00000865, 0.00000419, -0.00000588, 0.00000204, -0.00001454, 0.00000843, 0.00000084, 0.00000202, -0.00000481, -0.00000296, 0.00000714, -0.00000688, -0.00000773, 0.00000199, 0.00001025, -0.00000700, -0.00000665, 0.00000543, 0.00000534, -0.00000477, -0.00000495, 0.00000296, 0.00001029, -0.00001550, 0.00002214, -0.00002256, 0.00000884, -0.00000591, -0.00000150, 0.00000356, 0.00000060, -0.00000467, -0.00000498, -0.00000203, -0.00000650, 0.00000333, 0.00000183, -0.00000233, -0.00000517, -0.00000139, -0.00000469, -0.00000415, 0.00000748, -0.00002375, 0.00003383, -0.00003765, 0.00001145, -0.00000119, -0.00000137, 0.00001667, 0.00000615, 0.00000448, -0.00000898, 0.00000573, -0.00000119, 0.00000879, },
		{ -0.00000727, 0.00001707, -0.00000759, 0.00001909, 0.00000077, -0.00000347, 0.00000788, -0.00001956, 0.00000535, -0.00000064, -0.00000909, -0.00000138, -0.00000066, 0.00000495, -0.00000704, -0.00000075, 0.00000541, 0.00000408, -0.00000796, -0.00000123, 0.00000145, 0.00000065, -0.00000357, 0.00000213, -0.00000361, 0.00001108, -0.00001345, 0.00000812, -0.00001047, 0.00001368, -0.00001079, 0.00001379, -0.00001915, 0.00001522, -0.00002068, 0.00001100, -0.00001306, 0.00001949, -0.00002679, 0.00001583, -0.00002569, 0.00001922, -0.00001981, 0.00001921, -0.00004081, 0.00004145, -0.00007446, 0.00002968, -0.00009375, 0.00005804, -0.00030466, -0.00007674, 0.00054491, 0.00049456, -0.00014880, -0.00024305, 0.00003815, -0.00011111, 0.00004645, -0.00005582, 0.00003578, -0.00005428, 0.00002867, -0.00002745, 0.00001479, -0.00003155, 0.00002107, -0.00001599, 0.00000776, -0.00002466, 0.00001608, -0.00001405, 0.00001716, -0.00001801, 0.00001393, -0.00001478, 0.00000698, -0.00001994, 0.00002198, -0.00001287, 0.00001227, -0.00001755, 0.00001276, -0.00000582, -0.00000131, -0.00001288, 0.00001488, -0.00000164, -0.00000388, -0.00000824, 0.00001428, -0.00000571, 0.00000118, -0.00001126, 0.00001758, -0.00000915, 0.00000282, 0.00000806, -0.00001883, 0.00001254, -0.00001191, 0.00000599, -0.00000056, 0.00000161, -0.00000843, -0.00000122, -0.00000471, -0.00000396, 0.00000558, -0.00000101, -0.00000318, -0.00000571, 0.00000220, -0.00001235, 0.00001128, -0.00001614, 0.00000609, 0.00001092, -0.00003691, 0.00003173, -0.00002525, 0.00002587, -0.00000405, 0.00002159, -0.00001762, 0.00000939, -0.00000801, 0.00001387, },
		{ 0.00001406, -0.00000748, 0.00001857, -0.00001224, 0.00002945, -0.00001816, 0.00001453, -0.00000964, -0.00000981, 0.00000587, -0.00000435, -0.00000820, 0.00000080, 0.00000039, 0.00000321, -0.00000914, 0.00000390, 0.00000595, -0.00000051, -0.00000752, 0.00000031, 0.00000326, -0.00000411, 0.00000203, -0.00000478, 0.00000755, -0.00000199, -0.00000365, -0.00000068, 0.00000114, 0.00000225, 0.00000005, 0.00000044, -0.00000648, 0.00000351, -0.00001171, 0.00000313, 0.00000072, 0.00000173, -0.00001373, 0.00000382, -0.00001185, 0.00000593, -0.00000407, -0.00000654, -0.00000689, 0.00000211, -0.00004941, 0.00000145, -0.00006097, -0.00000281, -0.00039072, 0.00021313, 0.00069034, 0.00023823, -0.00036849, -0.00001612, -0.00007685, -0.00001241, -0.00001404, -0.00000381, -0.00001279, -0.00001318, 0.00000302, -0.00000643, -0.00000991, -0.00000620, 0.00000443, -0.00000211, -0.00001269, -0.00000235, 0.00000029, 0.00000197, 0.00000011, -0.00000321, 0.00000115, -0.00000384, -0.00000764, -0.00000159, 0.00001078, -0.00000604, 0.00000348, -0.00001063, 0.00001163, -0.00000740, -0.00000555, -0.00000453, 0.00001400, -0.00000539, -0.00000672, 0.00000054, 0.00001000, -0.00000648, -0.00000178, -0.00000498, 0.00001768, -0.00001814, 0.00001760, -0.00001029, -0.00000527, 0.00000610, -0.00000962, 0.00000845, -0.00000323, 0.00000139, -0.00001088, 0.00000328, -0.00000976, 0.00000301, 0.00000283, -0.00000141, -0.00000663, 0.00000019, -0.00000450, -0.00000549, 0.00000912, -0.00002422, 0.00003320, -0.00003302, 0.00000630, 0.00000442, -0.00000737, 0.00001936, 0.00000028, 0.00001368, -0.00001882, 0.00001326, -0.00000854, },
		{ 0.00000402, 0.00000105, 0.00000693, 0.00000274, 0.00000372, 0.00001669, -0.00001638, 0.00001859, -0.00002098, 0.00000145, 0.00000274, -0.00000770, -0.00000436, 0.00000034, 0.00000323, -0.00000225, -0.00000647, 0.00000696, 0.00000485, -0.00000519, -0.00000512, 0.00000224, 0.00000179, -0.00000480, 0.00000320, -0.00000576, 0.00001244, -0.00001168, 0.00000510, -0.00000748, 0.00001036, -0.00000849, 0.00001056, -0.00001284, 0.00000785, -0.00001187, 0.00000144, -0.00000572, 0.00001295, -0.00001805, 0.00000500, -0.00001286, 0.00000552, -0.00001024, 0.00001087, -0.00002799, 0.00002567, -0.00004957, -0.00000115, -0.00005331, 0.00000632, -0.00016245, -0.00032050, 0.00055081, 0.00067597, -0.00019258, -0.00024188, 0.00001972, -0.00009995, 0.00003102, -0.00004267, 0.00002002, -0.00003859, 0.00001650, -0.00001795, 0.00000654, -0.00002371, 0.00001212, -0.00000800, 0.00000312, -0.00002152, 0.00001268, -0.00001148, 0.00001301, -0.00001252, 0.00000905, -0.00001003, 0.00000391, -0.00001730, 0.00001604, -0.00000498, 0.00000494, -0.00000907, 0.00000281, 0.00000294, -0.00000525, -0.00000993, 0.00000696, 0.00000653, -0.00000617, -0.00000787, 0.00000974, 0.00000062, -0.00000253, -0.00000772, 0.00000806, 0.00000475, -0.00001184, 0.00001914, -0.00002142, 0.00000873, -0.00000523, -0.00000022, 0.00000362, -0.00000081, -0.00000384, -0.00000703, 0.00000232, -0.00001091, 0.00000923, -0.00000225, -0.00000116, -0.00000847, 0.00000601, -0.00001494, 0.00001212, -0.00001518, 0.00000438, 0.00001392, -0.00003700, 0.00003063, -0.00002323, 0.00002182, -0.00000340, 0.00001966, -0.00001177, 0.00000032, 0.00000078, },
		{ -0.00000815, 0.00001162, -0.00000632, 0.00001599, -0.00001004, 0.00002142, -0.00000517, -0.00000018, 0.00000781, -0.00002119, 0.00000980, -0.00000412, -0.00000639, -0.00000299, 0.00000217, 0.00000235, -0.00000553, -0.00000301, 0.00001042, -0.00000093, -0.00000503, -0.00000485, 0.00000684, -0.00000531, 0.00000200, -0.00000417, 0.00000452, 0.00000277, -0.00000691, 0.00000138, -0.00000094, 0.00000411, -0.00000367, 0.00000525, -0.00000944, 0.00000564, -0.00001226, 0.00000170, 0.00000061, 0.00000423, -0.00001604, 0.00000505, -0.00001105, 0.00000274, -0.00000480, -0.00000065, -0.00001330, 0.00001152, -0.00005549, 0.00000706, -0.00006555, 0.00001680, -0.00032485, -0.00009486, 0.00090313, 0.00029474, -0.00039958, -0.00002840, -0.00007237, -0.00002292, -0.00000765, -0.00001096, -0.00000959, -0.00001289, 0.00000305, -0.00000753, -0.00000817, -0.00000833, 0.00000454, -0.00000085, -0.00001194, -0.00000508, 0.00000336, -0.00000277, 0.00000425, -0.00000572, 0.00000360, -0.00000576, -0.00000408, -0.00000742, 0.00001481, -0.00000751, 0.00000480, -0.00001054, 0.00000842, -0.00000222, -0.00000716, -0.00000610, 0.00001121, 0.00000100, -0.00000903, -0.00000151, 0.00001004, -0.00000377, -0.00000240, -0.00000712, 0.00001574, -0.00000968, 0.00000296, 0.00000773, -0.00001971, 0.00001473, -0.00001321, 0.00000964, -0.00000452, 0.00000471, -0.00001275, 0.00000357, -0.00000754, -0.00000172, 0.00000632, -0.00000215, -0.00000616, -0.00000071, -0.00000163, -0.00000868, 0.00001155, -0.00002500, 0.00003265, -0.00002901, 0.00000276, 0.00000777, -0.00001104, 0.00002056, -0.00000440, 0.00001992, -0.00002304, 0.00001374, },
		{ 0.00000984, -0.00000474, 0.00000773, -0.00000134, 0.00001087, -0.00000734, 0.00002479, -0.00001893, 0.00001670, -0.00001115, -0.00000890, 0.00000841, -0.00000748, -0.00000514, -0.00000078, 0.00000272, 0.00000087, -0.00000860, 0.00000446, 0.00000766, -0.00000361, -0.00000649, 0.00000019, 0.00000445, -0.00000620, 0.00000353, -0.00000601, 0.00001043, -0.00000628, -0.00000081, -0.00000203, 0.00000467, -0.00000317, 0.00000322, -0.00000258, -0.00000256, -0.00000086, -0.00000921, 0.00000258, 0.00000429, -0.00000637, -0.00000862, 0.00000185, -0.00000865, 0.00000057, -0.00000205, -0.00000836, 0.00000184, -0.00001637, -0.00003955, -0.00000533, -0.00005499, -0.00003322, -0.00045335, 0.00037091, 0.00097579, -0.00026785, -0.00022595, -0.00001267, -0.00007680, 0.00000455, -0.00002009, -0.00000313, -0.00001883, 0.00000192, -0.00000562, -0.00000502, -0.00001256, 0.00000057, 0.00000148, -0.00000298, -0.00001581, 0.00000588, -0.00000469, 0.00000426, -0.00000324, 0.00000108, -0.00000240, -0.00000239, -0.00001031, 0.00000563, 0.00000599, -0.00000436, 0.00000053, -0.00000655, 0.00000943, -0.00000738, -0.00000678, -0.00000074, 0.00001196, -0.00000574, -0.00000760, 0.00000458, 0.00000617, -0.00000526, -0.00000382, -0.00000113, 0.00001453, -0.00001779, 0.00001793, -0.00001166, -0.00000428, 0.00000666, -0.00000874, 0.00000925, -0.00000526, 0.00000295, -0.00001393, 0.00000891, -0.00001558, 0.00001000, -0.00000174, 0.00000022, -0.00001081, 0.00000891, -0.00001588, 0.00001152, -0.00001371, 0.00000302, 0.00001551, -0.00003523, 0.00002797, -0.00002009, 0.00001688, -0.00000081, 0.00001407, -0.00000216, -0.00001070, },
		{ 0.00000613, -0.00000016, 0.00000243, 0.00000128, 0.00000667, -0.00000010, 0.00000593, 0.00001255, -0.00001621, 0.00002153, -0.00002577, 0.00000825, -0.00000192, -0.00000420, -0.00000702, 0.00000429, -0.00000022, 0.00000048, -0.00001068, 0.00001394, -0.00000154, -0.00000096, -0.00001060, 0.00000975, -0.00000600, 0.00000227, -0.00000450, 0.00000345, 0.00000356, -0.00000465, -0.00000245, 0.00000297, -0.00000008, -0.00000061, 0.00000139, -0.00000241, -0.00000270, -0.00000213, -0.00000979, 0.00001018, -0.00000418, -0.00000567, -0.00000782, 0.00000450, -0.00001431, 0.00000845, -0.00001249, 0.00000276, -0.00000702, -0.00002741, -0.00002848, -0.00001783, -0.00004308, -0.00015301, -0.00043502, 0.00100867, 0.00048371, -0.00050588, 0.00000643, -0.00010543, -0.00000192, -0.00002956, 0.00000770, -0.00002995, 0.00000681, -0.00001256, 0.00000531, -0.00001959, 0.00000217, -0.00000692, 0.00000998, -0.00001919, -0.00000006, -0.00000140, 0.00000091, -0.00000084, 0.00000069, -0.00000216, -0.00000010, -0.00000790, -0.00000465, 0.00000903, 0.00000057, -0.00000295, -0.00000162, -0.00000232, 0.00000797, -0.00001263, -0.00000222, 0.00000239, 0.00001077, -0.00001297, -0.00000031, 0.00000528, 0.00000308, -0.00000675, -0.00000347, 0.00000622, 0.00000511, -0.00001402, 0.00002227, -0.00002624, 0.00001428, -0.00000909, 0.00000539, -0.00000172, 0.00000359, -0.00000930, -0.00000143, -0.00000079, -0.00000898, 0.00001051, -0.00000350, -0.00000432, -0.00000351, 0.00000298, -0.00001260, 0.00001411, -0.00002610, 0.00003282, -0.00002682, 0.00000208, 0.00000792, -0.00001156, 0.00001903, -0.00000569, 0.00002035, -0.00001935, },
		{ -0.00001727, 0.00001276, -0.00000640, 0.00000721, -0.00000227, 0.00001093, -0.00000807, 0.00002013, -0.00000834, 0.00000272, 0.00000626, -0.00002131, 0.00001354, -0.00000766, -0.00000359, -0.00000428, 0.00000422, 0.00000080, -0.00000584, -0.00000230, 0.00001243, -0.00000394, -0.00000446, -0.00000490, 0.00000847, -0.00000791, 0.00000386, -0.00000577, 0.00000763, -0.00000115, -0.00000491, 0.00000087, 0.00000178, -0.00000039, -0.00000168, 0.00000385, -0.00000748, 0.00000348, -0.00001210, 0.00000319, 0.00000252, -0.00000190, -0.00001367, 0.00000625, -0.00001098, 0.00000027, -0.00000555, 0.00000074, -0.00000863, -0.00000098, -0.00005399, 0.00001119, -0.00007315, 0.00001212, -0.00041384, 0.00000430, 0.00133559, -0.00030841, -0.00024364, -0.00002316, -0.00007092, -0.00000762, -0.00001093, -0.00001290, -0.00001319, 0.00000004, -0.00000334, -0.00000817, -0.00000913, -0.00000329, 0.00000334, -0.00000279, -0.00001455, 0.00000252, -0.00000069, -0.00000162, 0.00000220, -0.00000271, 0.00000130, -0.00000540, -0.00000583, -0.00000157, 0.00001201, -0.00000785, 0.00000376, -0.00000852, 0.00000821, -0.00000353, -0.00000763, -0.00000343, 0.00001054, -0.00000001, -0.00000992, 0.00000223, 0.00000725, -0.00000357, -0.00000358, -0.00000475, 0.00001492, -0.00001169, 0.00000469, 0.00000639, -0.00002021, 0.00001735, -0.00001436, 0.00001236, -0.00000812, 0.00000793, -0.00001762, 0.00001106, -0.00001495, 0.00000628, 0.00000137, -0.00000002, -0.00001133, 0.00000937, -0.00001403, 0.00000886, -0.00001123, 0.00000122, 0.00001703, -0.00003371, 0.00002639, -0.00001850, 0.00001357, 0.00000101, 0.00000870, 0.00000571, },
		{ -0.00001028, -0.00000542, 0.00001023, -0.00000779, 0.00000941, -0.00000314, 0.00001024, -0.00000805, 0.00002502, -0.00002364, 0.00002370, -0.00001953, 0.00000023, 0.00000240, -0.00000257, -0.00000974, 0.00000526, -0.00000276, 0.00000622, -0.00001604, 0.00001441, -0.00000110, 0.00000314, -0.00001509, 0.00001085, -0.00000605, 0.00000332, -0.00000661, 0.00000549, -0.00000056, 0.00000322, -0.00001143, 0.00001174, -0.00000918, 0.00000820, -0.00000965, 0.00001208, -0.00001793, 0.00001435, -0.00002664, 0.00002500, -0.00001921, 0.00001259, -0.00002900, 0.00002798, -0.00003817, 0.00002956, -0.00003620, 0.00003411, -0.00004375, 0.00002131, -0.00008670, 0.00005632, -0.00013843, 0.00003966, -0.00068416, 0.00090398, 0.00084623, -0.00068959, 0.00008280, -0.00017052, 0.00004545, -0.00007463, 0.00004646, -0.00006722, 0.00003898, -0.00003790, 0.00002746, -0.00004022, 0.00002145, -0.00002606, 0.00002687, -0.00003192, 0.00001112, -0.00001296, 0.00001215, -0.00001389, 0.00001422, -0.00001420, 0.00001143, -0.00001779, 0.00000532, -0.00000426, 0.00001490, -0.00001582, 0.00001150, -0.00001546, 0.00001855, -0.00001850, 0.00000393, -0.00000836, 0.00001972, -0.00001565, 0.00000208, -0.00000166, 0.00001087, -0.00001179, 0.00000234, -0.00000522, 0.00001830, -0.00002442, 0.00002574, -0.00002053, 0.00000392, 0.00000129, -0.00000222, 0.00000342, -0.00000035, -0.00000277, -0.00000856, 0.00000657, -0.00001478, 0.00001231, -0.00000339, -0.00000278, -0.00000636, 0.00000720, -0.00001528, 0.00001548, -0.00002682, 0.00003338, -0.00002614, 0.00000352, 0.00000611, -0.00001034, 0.00001586, -0.00000422, 0.00001601, },
		{ 0.00000785, -0.00001563, 0.00000682, 0.00000066, -0.00000136, 0.00000470, 0.00000252, 0.00000111, 0.00000514, 0.00001039, -0.00001585, 0.00002286, -0.00002922, 0.00001470, -0.00000700, -0.00000041, -0.00000943, 0.00000749, -0.00000257, 0.00000123, -0.00001185, 0.00001806, -0.00000609, 0.00000096, -0.00001254, 0.00001364, -0.00001068, 0.00000589, -0.00000774, 0.00000794, -0.00000120, -0.00000202, -0.00000385, 0.00000674, -0.00000561, 0.00000220, -0.00000056, 0.00000014, -0.00000562, -0.00000092, -0.00000997, 0.00001425, -0.00001224, -0.00000180, -0.00000825, 0.00000689, -0.00001982, 0.00001114, -0.00001448, 0.00001083, -0.00002240, -0.00002318, -0.00002787, -0.00001943, -0.00005375, -0.00021059, -0.00046968, 0.00163690, -0.00021880, -0.00035401, 0.00002833, -0.00011358, 0.00002066, -0.00003768, 0.00001000, -0.00003753, 0.00002350, -0.00002199, 0.00000734, -0.00002314, 0.00000947, -0.00001020, 0.00001053, -0.00002440, 0.00000957, -0.00000683, 0.00000302, -0.00000365, 0.00000448, -0.00000520, 0.00000101, -0.00001037, 0.00000133, 0.00000658, 0.00000008, -0.00000409, 0.00000050, -0.00000289, 0.00000770, -0.00001440, 0.00000121, 0.00000107, 0.00001114, -0.00001549, 0.00000436, 0.00000225, 0.00000367, -0.00000846, -0.00000117, 0.00000575, 0.00000354, -0.00001418, 0.00002415, -0.00003073, 0.00002056, -0.00001317, 0.00001051, -0.00000723, 0.00000876, -0.00001612, 0.00000770, -0.00000927, -0.00000063, 0.00000566, -0.00000114, -0.00001025, 0.00000758, -0.00000995, 0.00000497, -0.00000820, -0.00000121, 0.00001942, -0.00003402, 0.00002765, -0.00002007, 0.00001338, 0.00000011, 0.00000660, },
		{ 0.00001075, -0.00000143, -0.00001414, 0.00001594, -0.00001218, 0.00001151, -0.00000680, 0.00001366, -0.00001268, 0.00002507, -0.00001733, 0.00001263, -0.00000427, -0.00001122, 0.00000695, -0.00000145, -0.00001022, 0.00000400, -0.00000361, 0.00000918, -0.00001668, 0.00001053, 0.00000126, 0.00000563, -0.00001646, 0.00000899, -0.00000484, 0.00000437, -0.00000931, 0.00000889, -0.00000675, 0.00001245, -0.00002035, 0.00001976, -0.00001736, 0.00001650, -0.00002059, 0.00002537, -0.00003043, 0.00002691, -0.00003842, 0.00003477, -0.00003012, 0.00002680, -0.00004512, 0.00004483, -0.00005389, 0.00004338, -0.00005487, 0.00006022, -0.00007345, 0.00005964, -0.00012947, 0.00010858, -0.00020173, 0.00015580, -0.00073944, 0.00054315, 0.00135435, -0.00089639, 0.00015456, -0.00023148, 0.00008909, -0.00011644, 0.00008122, -0.00009918, 0.00006381, -0.00005626, 0.00004463, -0.00005704, 0.00003727, -0.00004141, 0.00003957, -0.00004136, 0.00002068, -0.00002389, 0.00002349, -0.00002701, 0.00002675, -0.00002473, 0.00002156, -0.00002683, 0.00001530, -0.00001701, 0.00002690, -0.00002526, 0.00002045, -0.00002314, 0.00002293, -0.00001953, 0.00000718, -0.00001525, 0.00002272, -0.00001358, 0.00000249, -0.00000673, 0.00001522, -0.00001333, 0.00000551, -0.00001231, 0.00002306, -0.00002288, 0.00001619, -0.00000472, -0.00001146, 0.00001233, -0.00000836, 0.00000710, -0.00000358, 0.00000282, -0.00001334, 0.00001014, -0.00001557, 0.00000956, -0.00000057, -0.00000282, -0.00000758, 0.00000916, -0.00001521, 0.00001475, -0.00002646, 0.00003346, -0.00002563, 0.00000498, 0.00000502, -0.00001017, 0.00001366, -0.00000274, },
		{ 0.00000801, 0.00000418, -0.00000702, -0.00000516, 0.00001245, -0.00001341, 0.00001473, -0.00000919, 0.00001384, -0.00001188, 0.00002741, -0.00002874, 0.00003012, -0.00002782, 0.00001034, -0.00000523, 0.00000367, -0.00001502, 0.00001147, -0.00000804, 0.00001050, -0.00002153, 0.00002281, -0.00000937, 0.00000883, -0.00002146, 0.00001937, -0.00001504, 0.00001097, -0.00001388, 0.00001392, -0.00000906, 0.00000988, -0.00001748, 0.00002060, -0.00001999, 0.00001633, -0.00001711, 0.00002075, -0.00002769, 0.00002314, -0.00003537, 0.00003838, -0.00003670, 0.00002611, -0.00004009, 0.00004259, -0.00005736, 0.00004702, -0.00005408, 0.00005991, -0.00007891, 0.00004856, -0.00011462, 0.00009246, -0.00019935, 0.00007631, -0.00091246, 0.00173995, 0.00008381, -0.00059160, 0.00015852, -0.00021579, 0.00009752, -0.00010633, 0.00006943, -0.00009343, 0.00007246, -0.00006184, 0.00004241, -0.00005570, 0.00003966, -0.00003948, 0.00003730, -0.00004635, 0.00002878, -0.00002533, 0.00002056, -0.00002270, 0.00002388, -0.00002282, 0.00001789, -0.00002532, 0.00001567, -0.00001071, 0.00001870, -0.00002144, 0.00001806, -0.00002067, 0.00002327, -0.00002503, 0.00001136, -0.00001362, 0.00002453, -0.00002236, 0.00000995, -0.00000732, 0.00001439, -0.00001648, 0.00000715, -0.00000838, 0.00002058, -0.00002967, 0.00003318, -0.00003027, 0.00001415, -0.00000550, 0.00000494, -0.00000368, 0.00000646, -0.00001105, 0.00000150, -0.00000214, -0.00000679, 0.00000799, -0.00000103, -0.00000923, 0.00000533, -0.00000571, 0.00000179, -0.00000589, -0.00000366, 0.00002253, -0.00003624, 0.00003155, -0.00002411, 0.00001551, -0.00000312, },
		{ -0.00000386, 0.00000842, 0.00000214, -0.00001486, 0.00001257, -0.00000727, 0.00000478, -0.00000170, 0.00000750, -0.00000586, 0.00001255, 0.00000007, -0.00000609, 0.00001299, -0.00002040, 0.00000891, -0.00000041, -0.00000817, 0.00000007, -0.00000173, 0.00000778, -0.00001160, 0.00000230, 0.00000583, 0.00000516, -0.00001316, 0.00000316, -0.00000120, 0.00000338, -0.00000954, 0.00000950, -0.00000947, 0.00001620, -0.00002184, 0.00001957, -0.00001699, 0.00001611, -0.00002210, 0.00002711, -0.00002944, 0.00002477, -0.00003450, 0.00002888, -0.00002558, 0.00002428, -0.00004229, 0.00004035, -0.00004648, 0.00003411, -0.00004999, 0.00005852, -0.00006926, 0.00005620, -0.00012035, 0.00009570, -0.00017866, 0.00013239, -0.00056633, -0.00002892, 0.00189990, -0.00102071, 0.00014643, -0.00023353, 0.00008527, -0.00011805, 0.00007989, -0.00009782, 0.00005756, -0.00004792, 0.00003960, -0.00005412, 0.00003466, -0.00003915, 0.00003579, -0.00003658, 0.00001825, -0.00002385, 0.00002439, -0.00002943, 0.00002778, -0.00002398, 0.00002117, -0.00002619, 0.00001613, -0.00001964, 0.00002730, -0.00002289, 0.00001771, -0.00001899, 0.00001628, -0.00001189, 0.00000343, -0.00001389, 0.00001652, -0.00000481, -0.00000131, -0.00000672, 0.00001291, -0.00000860, 0.00000289, -0.00001148, 0.00001703, -0.00001007, -0.00000184, 0.00001378, -0.00002396, 0.00001762, -0.00000873, 0.00000643, -0.00000349, 0.00000467, -0.00001315, 0.00000812, -0.00001086, 0.00000301, 0.00000394, -0.00000403, -0.00000704, 0.00000866, -0.00001255, 0.00001239, -0.00002521, 0.00003262, -0.00002409, 0.00000448, 0.00000695, -0.00001324, 0.00001457, },
		{ -0.00000622, 0.00001046, -0.00000186, 0.00000331, -0.00001654, 0.00002181, -0.00002175, 0.00002083, -0.00001696, 0.00002199, -0.00002185, 0.00003386, -0.00002886, 0.00002444, -0.00001726, 0.00000241, -0.00000366, 0.00000833, -0.00001979, 0.00001477, -0.00001365, 0.00001902, -0.00002840, 0.00002490, -0.00001253, 0.00001726, -0.00002950, 0.00002426, -0.00002029, 0.00001849, -0.00002339, 0.00002439, -0.00002265, 0.00002729, -0.00003554, 0.00003843, -0.00003841, 0.00003525, -0.00003935, 0.00004648, -0.00005379, 0.00005038, -0.00006304, 0.00006500, -0.00006505, 0.00005871, -0.00007645, 0.00008199, -0.00009774, 0.00008740, -0.00010181, 0.00011889, -0.00014591, 0.00013041, -0.00021015, 0.00021152, -0.00034946, 0.00032827, -0.00117829, 0.00154200, 0.00062098, -0.00093307, 0.00033858, -0.00035240, 0.00020083, -0.00019664, 0.00014624, -0.00016258, 0.00013015, -0.00010872, 0.00008503, -0.00009590, 0.00007692, -0.00007467, 0.00006869, -0.00007300, 0.00005374, -0.00005024, 0.00004504, -0.00004850, 0.00004871, -0.00004502, 0.00003911, -0.00004491, 0.00003541, -0.00003289, 0.00004032, -0.00004043, 0.00003622, -0.00003743, 0.00003656, -0.00003423, 0.00002186, -0.00002768, 0.00003486, -0.00002686, 0.00001584, -0.00001753, 0.00002425, -0.00002342, 0.00001530, -0.00002085, 0.00003171, -0.00003493, 0.00002976, -0.00001923, 0.00000167, 0.00000386, 0.00000008, -0.00000109, 0.00000449, -0.00000643, -0.00000291, 0.00000167, -0.00000812, 0.00000561, 0.00000205, -0.00000989, 0.00000459, -0.00000350, 0.00000118, -0.00000553, -0.00000528, 0.00002550, -0.00003912, 0.00003609, -0.00002805, 0.00001729, },
		{ 0.00001858, -0.00000808, 0.00001034, -0.00000107, -0.00000716, 0.00000118, 0.00000470, -0.00000789, 0.00000908, -0.00000484, 0.00000780, -0.00000582, 0.00001979, -0.00002273, 0.00002387, -0.00002244, 0.00000686, 0.00000032, -0.00000342, -0.00000636, 0.00000304, 0.00000184, -0.00000150, -0.00000895, 0.00001217, 0.00000099, -0.00000459, -0.00000695, 0.00000585, -0.00000188, -0.00000405, 0.00000320, -0.00000367, 0.00000899, -0.00001072, 0.00000644, -0.00000340, 0.00000200, -0.00000857, 0.00001148, -0.00000963, 0.00000320, -0.00001069, 0.00000326, -0.00000083, -0.00000059, -0.00001448, 0.00000865, -0.00001030, -0.00000449, -0.00001311, 0.00001825, -0.00001908, -0.00000169, -0.00004671, 0.00000542, -0.00005831, -0.00002855, -0.00021831, -0.00068028, 0.00232110, -0.00094202, -0.00001190, -0.00013186, 0.00000071, -0.00005284, 0.00002064, -0.00004543, 0.00000644, -0.00000206, 0.00000268, -0.00002202, 0.00000468, -0.00001124, 0.00000854, -0.00001122, -0.00000272, -0.00000587, 0.00000761, -0.00001390, 0.00001067, -0.00000619, 0.00000502, -0.00001070, 0.00000244, -0.00000683, 0.00001173, -0.00000558, 0.00000121, -0.00000205, -0.00000129, 0.00000408, -0.00000773, -0.00000370, 0.00000173, 0.00000941, -0.00000912, -0.00000090, 0.00000362, 0.00000216, -0.00000533, -0.00000335, 0.00000294, 0.00000816, -0.00002010, 0.00002651, -0.00002742, 0.00001417, -0.00000286, 0.00000204, -0.00000070, 0.00000334, -0.00000914, 0.00000271, -0.00000394, -0.00000354, 0.00000698, -0.00000424, -0.00000634, 0.00000735, -0.00000924, 0.00001025, -0.00002426, 0.00003129, -0.00002131, 0.00000142, 0.00001233, -0.00001961, },
		{ 0.00001599, -0.00000650, 0.00001077, -0.00000577, 0.00000944, -0.00002096, 0.00002283, -0.00002126, 0.00001843, -0.00001627, 0.00002073, -0.00002024, 0.00002693, -0.00001630, 0.00000925, -0.00000296, -0.00000447, -0.00000437, 0.00001323, -0.00002251, 0.00001592, -0.00001722, 0.00002392, -0.00003001, 0.00002298, -0.00001408, 0.00002363, -0.00003379, 0.00002600, -0.00002414, 0.00002537, -0.00003217, 0.00003418, -0.00003529, 0.00004206, -0.00004908, 0.00005098, -0.00005131, 0.00004871, -0.00005572, 0.00006448, -0.00007026, 0.00006678, -0.00007883, 0.00007984, -0.00008210, 0.00007920, -0.00009888, 0.00010534, -0.00012040, 0.00011047, -0.00013259, 0.00015767, -0.00018758, 0.00017992, -0.00026392, 0.00027583, -0.00042291, 0.00044064, -0.00117470, 0.00102907, 0.00132269, -0.00128603, 0.00049304, -0.00046466, 0.00028218, -0.00026689, 0.00020353, -0.00021225, 0.00016830, -0.00013872, 0.00011419, -0.00012433, 0.00010307, -0.00009891, 0.00008956, -0.00009083, 0.00007183, -0.00006919, 0.00006405, -0.00006844, 0.00006658, -0.00006014, 0.00005378, -0.00005860, 0.00004976, -0.00004865, 0.00005380, -0.00005067, 0.00004541, -0.00004469, 0.00004069, -0.00003631, 0.00002693, -0.00003506, 0.00003723, -0.00002561, 0.00001859, -0.00002402, 0.00002877, -0.00002521, 0.00001879, -0.00002650, 0.00003259, -0.00002844, 0.00001643, -0.00000365, -0.00000967, 0.00000859, 0.00000048, -0.00000273, 0.00000584, -0.00000554, -0.00000228, -0.00000044, -0.00000328, -0.00000143, 0.00000761, -0.00001223, 0.00000593, -0.00000402, 0.00000347, -0.00000705, -0.00000592, 0.00002761, -0.00004116, 0.00003892, -0.00002912, },
		{ -0.00002706, 0.00002341, -0.00001340, 0.00001469, -0.00000841, 0.00000460, -0.00001238, 0.00001716, -0.00001980, 0.00001889, -0.00001642, 0.00002030, -0.00002070, 0.00003222, -0.00002908, 0.00002353, -0.00001652, 0.00000229, -0.00000070, 0.00000411, -0.00001459, 0.00000992, -0.00000740, 0.00001135, -0.00002083, 0.00001912, -0.00000649, 0.00000836, -0.00002014, 0.00001595, -0.00001199, 0.00000814, -0.00001107, 0.00001166, -0.00000958, 0.00001221, -0.00001790, 0.00002131, -0.00002370, 0.00001777, -0.00001851, 0.00002477, -0.00003252, 0.00002704, -0.00003591, 0.00003823, -0.00004189, 0.00003163, -0.00004248, 0.00004582, -0.00006346, 0.00004730, -0.00005144, 0.00006540, -0.00009844, 0.00007054, -0.00013411, 0.00011963, -0.00026097, 0.00018298, -0.00123582, 0.00246054, -0.00057949, -0.00034338, 0.00007702, -0.00016324, 0.00007585, -0.00009177, 0.00005187, -0.00008229, 0.00007403, -0.00006024, 0.00003469, -0.00004857, 0.00003822, -0.00003802, 0.00003108, -0.00003962, 0.00002826, -0.00002539, 0.00001800, -0.00002235, 0.00002583, -0.00002401, 0.00001703, -0.00002342, 0.00001870, -0.00001599, 0.00002187, -0.00002380, 0.00002207, -0.00002405, 0.00002344, -0.00002254, 0.00001166, -0.00001719, 0.00002494, -0.00001834, 0.00000874, -0.00000996, 0.00001573, -0.00001599, 0.00000815, -0.00001329, 0.00002425, -0.00003115, 0.00002865, -0.00002101, 0.00000454, 0.00000503, -0.00000187, 0.00000117, 0.00000234, -0.00000542, -0.00000125, -0.00000000, -0.00000537, 0.00000514, -0.00000119, -0.00000751, 0.00000752, -0.00000777, 0.00001050, -0.00002503, 0.00003041, -0.00001806, -0.00000317, 0.00001950, },
		{ -0.00002612, 0.00001066, -0.00000302, 0.00000745, -0.00000502, 0.00000835, -0.00001615, 0.00001456, -0.00001208, 0.00000831, -0.00000805, 0.00001119, -0.00000926, 0.00001108, 0.00000202, -0.00000701, 0.00000747, -0.00000586, -0.00000825, 0.00001725, -0.00002176, 0.00001361, -0.00001696, 0.00002316, -0.00002513, 0.00001650, -0.00001263, 0.00002526, -0.00003158, 0.00002216, -0.00002350, 0.00002720, -0.00003462, 0.00003655, -0.00003897, 0.00004531, -0.00004916, 0.00004948, -0.00004975, 0.00004714, -0.00005569, 0.00006356, -0.00006598, 0.00006140, -0.00007203, 0.00007198, -0.00007588, 0.00007421, -0.00009291, 0.00009749, -0.00010970, 0.00009913, -0.00012599, 0.00015189, -0.00017598, 0.00016570, -0.00024084, 0.00024547, -0.00037321, 0.00036892, -0.00090272, 0.00029445, 0.00203494, -0.00151243, 0.00052627, -0.00048443, 0.00028821, -0.00027300, 0.00020406, -0.00021089, 0.00016073, -0.00012999, 0.00011037, -0.00012264, 0.00010091, -0.00009647, 0.00008585, -0.00008702, 0.00007055, -0.00006965, 0.00006500, -0.00007008, 0.00006582, -0.00005760, 0.00005204, -0.00005684, 0.00004905, -0.00004844, 0.00005053, -0.00004490, 0.00003955, -0.00003769, 0.00003224, -0.00002844, 0.00002326, -0.00003241, 0.00002961, -0.00001750, 0.00001630, -0.00002400, 0.00002553, -0.00002002, 0.00001573, -0.00002374, 0.00002369, -0.00001407, 0.00000016, 0.00000853, -0.00001348, 0.00000511, 0.00000728, -0.00000854, 0.00001038, -0.00000831, 0.00000280, -0.00000684, 0.00000490, -0.00000961, 0.00001257, -0.00001419, 0.00000789, -0.00000583, 0.00000698, -0.00000882, -0.00000658, 0.00002901, -0.00004174, 0.00003883, },
		{ 0.00002587, -0.00003251, 0.00002598, -0.00001685, 0.00001808, -0.00001519, 0.00001413, -0.00002109, 0.00002327, -0.00002467, 0.00002196, -0.00002139, 0.00002524, -0.00002565, 0.00003235, -0.00002304, 0.00001359, -0.00000676, -0.00000084, -0.00000532, 0.00001398, -0.00002315, 0.00001721, -0.00001749, 0.00002367, -0.00003042, 0.00002492, -0.00001555, 0.00002288, -0.00003337, 0.00002672, -0.00002478, 0.00002427, -0.00002973, 0.00003195, -0.00003347, 0.00003933, -0.00004491, 0.00004822, -0.00005169, 0.00004695, -0.00005159, 0.00006100, -0.00006868, 0.00006441, -0.00007412, 0.00007692, -0.00008409, 0.00007890, -0.00009410, 0.00010141, -0.00012135, 0.00010853, -0.00012452, 0.00015284, -0.00019627, 0.00018575, -0.00026712, 0.00028436, -0.00046698, 0.00049872, -0.00154362, 0.00223040, 0.00005995, -0.00079523, 0.00033963, -0.00036121, 0.00022877, -0.00022230, 0.00016220, -0.00017963, 0.00015490, -0.00012759, 0.00009711, -0.00010761, 0.00009270, -0.00008847, 0.00007645, -0.00008021, 0.00006723, -0.00006359, 0.00005516, -0.00005983, 0.00006067, -0.00005518, 0.00004690, -0.00005168, 0.00004687, -0.00004514, 0.00004864, -0.00004699, 0.00004362, -0.00004322, 0.00003890, -0.00003497, 0.00002617, -0.00003387, 0.00003649, -0.00002529, 0.00001866, -0.00002341, 0.00002747, -0.00002473, 0.00001817, -0.00002581, 0.00003265, -0.00003193, 0.00002124, -0.00000966, -0.00000454, 0.00000852, -0.00000025, -0.00000176, 0.00000531, -0.00000598, 0.00000036, -0.00000266, -0.00000009, -0.00000252, 0.00000542, -0.00001118, 0.00001032, -0.00000943, 0.00001405, -0.00002796, 0.00003044, -0.00001529, -0.00000765, },
		{ 0.00003637, -0.00002010, 0.00000287, 0.00000263, 0.00000165, -0.00000040, 0.00000137, -0.00000501, 0.00000139, 0.00000052, -0.00000404, 0.00000227, -0.00000089, 0.00000415, -0.00000514, 0.00001674, -0.00001596, 0.00000886, -0.00000075, -0.00001353, 0.00001808, -0.00001619, 0.00000712, -0.00001202, 0.00001621, -0.00001423, 0.00000611, -0.00000753, 0.00002049, -0.00002157, 0.00001164, -0.00001631, 0.00002073, -0.00002674, 0.00002703, -0.00002890, 0.00003246, -0.00003209, 0.00003094, -0.00003086, 0.00002744, -0.00003591, 0.00004071, -0.00003880, 0.00003288, -0.00004169, 0.00004041, -0.00004469, 0.00004143, -0.00005628, 0.00005662, -0.00006412, 0.00005118, -0.00007811, 0.00009663, -0.00010675, 0.00008494, -0.00014064, 0.00012389, -0.00021090, 0.00014627, -0.00045727, -0.00048803, 0.00256776, -0.00148086, 0.00036751, -0.00037009, 0.00018987, -0.00019313, 0.00013115, -0.00014606, 0.00009852, -0.00007566, 0.00006687, -0.00008399, 0.00006406, -0.00006180, 0.00005283, -0.00005702, 0.00004491, -0.00004620, 0.00004244, -0.00004833, 0.00004227, -0.00003415, 0.00003104, -0.00003679, 0.00003039, -0.00002986, 0.00002938, -0.00002326, 0.00001961, -0.00001811, 0.00001342, -0.00001244, 0.00001172, -0.00002060, 0.00001389, -0.00000440, 0.00000949, -0.00001733, 0.00001506, -0.00000900, 0.00000733, -0.00001459, 0.00000928, 0.00000168, -0.00001191, 0.00001210, -0.00000824, -0.00000468, 0.00001661, -0.00001436, 0.00001438, -0.00001108, 0.00000808, -0.00001246, 0.00001081, -0.00001362, 0.00001291, -0.00001294, 0.00000806, -0.00000637, 0.00000896, -0.00000843, -0.00000883, 0.00003061, -0.00004138, },
		{ -0.00001093, 0.00002964, -0.00003400, 0.00002450, -0.00001659, 0.00001806, -0.00001778, 0.00001700, -0.00002096, 0.00002014, -0.00002066, 0.00001702, -0.00001853, 0.00002145, -0.00002045, 0.00002222, -0.00000947, 0.00000096, 0.00000015, 0.00000144, -0.00001403, 0.00002439, -0.00002968, 0.00002248, -0.00002535, 0.00003194, -0.00003492, 0.00002757, -0.00002341, 0.00003483, -0.00004229, 0.00003414, -0.00003569, 0.00003847, -0.00004557, 0.00004869, -0.00005266, 0.00005940, -0.00006296, 0.00006564, -0.00006994, 0.00006607, -0.00007359, 0.00008388, -0.00008972, 0.00008570, -0.00009553, 0.00009899, -0.00010960, 0.00010786, -0.00012503, 0.00013395, -0.00015447, 0.00014469, -0.00017078, 0.00020756, -0.00025376, 0.00025114, -0.00033804, 0.00036942, -0.00056330, 0.00063061, -0.00153390, 0.00165456, 0.00086614, -0.00124349, 0.00055827, -0.00051818, 0.00034473, -0.00031901, 0.00024066, -0.00024670, 0.00020751, -0.00017127, 0.00014007, -0.00014890, 0.00013019, -0.00012252, 0.00010657, -0.00010776, 0.00009499, -0.00009116, 0.00008184, -0.00008621, 0.00008355, -0.00007483, 0.00006613, -0.00006997, 0.00006526, -0.00006335, 0.00006323, -0.00005819, 0.00005370, -0.00005123, 0.00004445, -0.00003990, 0.00003442, -0.00004296, 0.00004011, -0.00002719, 0.00002553, -0.00003253, 0.00003347, -0.00002820, 0.00002325, -0.00003131, 0.00003193, -0.00002453, 0.00001000, -0.00000083, -0.00000643, 0.00000328, 0.00000889, -0.00001026, 0.00001288, -0.00001153, 0.00000783, -0.00001122, 0.00001032, -0.00001309, 0.00001297, -0.00001566, 0.00001471, -0.00001327, 0.00001965, -0.00003177, 0.00003071, -0.00001316, },
		{ -0.00004110, 0.00003316, -0.00001358, -0.00000434, 0.00000743, -0.00000377, 0.00000498, -0.00000717, 0.00000645, -0.00000973, 0.00000959, -0.00001208, 0.00000857, -0.00000925, 0.00001301, -0.00001435, 0.00002159, -0.00001414, 0.00000169, 0.00000732, -0.00001613, 0.00001319, -0.00000534, -0.00000304, -0.00000312, 0.00000443, 0.00000042, -0.00000619, 0.00000056, 0.00000968, -0.00000505, -0.00000404, -0.00000330, 0.00000612, -0.00000894, 0.00000659, -0.00000671, 0.00000641, -0.00000225, 0.00000064, -0.00000023, -0.00000459, -0.00000266, 0.00000334, 0.00000249, -0.00000911, 0.00000199, -0.00000457, 0.00000131, -0.00000849, -0.00000087, -0.00000394, 0.00000180, -0.00001832, -0.00000458, 0.00001002, -0.00000243, -0.00003409, 0.00000118, -0.00004365, 0.00000334, -0.00013543, 0.00001310, -0.00112628, 0.00276537, -0.00112628, 0.00001310, -0.00013543, 0.00000334, -0.00004365, 0.00000118, -0.00003409, -0.00000243, 0.00001002, -0.00000458, -0.00001832, 0.00000180, -0.00000394, -0.00000087, -0.00000849, 0.00000131, -0.00000457, 0.00000199, -0.00000911, 0.00000249, 0.00000334, -0.00000266, -0.00000459, -0.00000023, 0.00000064, -0.00000225, 0.00000641, -0.00000671, 0.00000659, -0.00000894, 0.00000612, -0.00000330, -0.00000404, -0.00000505, 0.00000968, 0.00000056, -0.00000619, 0.00000042, 0.00000443, -0.00000312, -0.00000304, -0.00000534, 0.00001319, -0.00001613, 0.00000732, 0.00000169, -0.00001414, 0.00002159, -0.00001435, 0.00001301, -0.00000925, 0.00000857, -0.00001208, 0.00000959, -0.00000973, 0.00000645, -0.00000717, 0.00000498, -0.00000377, 0.00000743, -0.00000434, -0.00001358, 0.00003316, },
		{ -0.00001093, -0.00001316, 0.00003071, -0.00003177, 0.00001965, -0.00001327, 0.00001471, -0.00001566, 0.00001297, -0.00001309, 0.00001032, -0.00001122, 0.00000783, -0.00001153, 0.00001288, -0.00001026, 0.00000889, 0.00000328, -0.00000643, -0.00000083, 0.00001000, -0.00002453, 0.00003193, -0.00003131, 0.00002325, -0.00002820, 0.00003347, -0.00003253, 0.00002553, -0.00002719, 0.00004011, -0.00004296, 0.00003442, -0.00003990, 0.00004445, -0.00005123, 0.00005370, -0.00005819, 0.00006323, -0.00006335, 0.00006526, -0.00006997, 0.00006613, -0.00007483, 0.00008355, -0.00008621, 0.00008184, -0.00009116, 0.00009499, -0.00010776, 0.00010657, -0.00012252, 0.00013019, -0.00014890, 0.00014007, -0.00017127, 0.00020751, -0.00024670, 0.00024066, -0.00031901, 0.00034473, -0.00051818, 0.00055827, -0.00124349, 0.00086614, 0.00165456, -0.00153390, 0.00063061, -0.00056330, 0.00036942, -0.00033804, 0.00025114, -0.00025376, 0.00020756, -0.00017078, 0.00014469, -0.00015447, 0.00013395, -0.00012503, 0.00010786, -0.00010960, 0.00009899, -0.00009553, 0.00008570, -0.00008972, 0.00008388, -0.00007359, 0.00006607, -0.00006994, 0.00006564, -0.00006296, 0.00005940, -0.00005266, 0.00004869, -0.00004557, 0.00003847, -0.00003569, 0.00003414, -0.00004229, 0.00003483, -0.00002341, 0.00002757, -0.00003492, 0.00003194, -0.00002535, 0.00002248, -0.00002968, 0.00002439, -0.00001403, 0.00000144, 0.00000015, 0.00000096, -0.00000947, 0.00002222, -0.00002045, 0.00002145, -0.00001853, 0.00001702, -0.00002066, 0.00002014, -0.00002096, 0.00001700, -0.00001778, 0.00001806, -0.00001659, 0.00002450, -0.00003400, 0.00002964, },
		{ 0.00003637, -0.00004138, 0.00003061, -0.00000883, -0.00000843, 0.00000896, -0.00000637, 0.00000806, -0.00001294, 0.00001291, -0.00001362, 0.00001081, -0.00001246, 0.00000808, -0.00001108, 0.00001438, -0.00001436, 0.00001661, -0.00000468, -0.00000824, 0.00001210, -0.00001191, 0.00000168, 0.00000928, -0.00001459, 0.00000733, -0.00000900, 0.00001506, -0.00001733, 0.00000949, -0.00000440, 0.00001389, -0.00002060, 0.00001172, -0.00001244, 0.00001342, -0.00001811, 0.00001961, -0.00002326, 0.00002938, -0.00002986, 0.00003039, -0.00003679, 0.00003104, -0.00003415, 0.00004227, -0.00004833, 0.00004244, -0.00004620, 0.00004491, -0.00005702, 0.00005283, -0.00006180, 0.00006406, -0.00008399, 0.00006687, -0.00007566, 0.00009852, -0.00014606, 0.00013115, -0.00019313, 0.00018987, -0.00037009, 0.00036751, -0.00148086, 0.00256776, -0.00048803, -0.00045727, 0.00014627, -0.00021090, 0.00012389, -0.00014064, 0.00008494, -0.00010675, 0.00009663, -0.00007811, 0.00005118, -0.00006412, 0.00005662, -0.00005628, 0.00004143, -0.00004469, 0.00004041, -0.00004169, 0.00003288, -0.00003880, 0.00004071, -0.00003591, 0.00002744, -0.00003086, 0.00003094, -0.00003209, 0.00003246, -0.00002890, 0.00002703, -0.00002674, 0.00002073, -0.00001631, 0.00001164, -0.00002157, 0.00002049, -0.00000753, 0.00000611, -0.00001423, 0.00001621, -0.00001202, 0.00000712, -0.00001619, 0.00001808, -0.00001353, -0.00000075, 0.00000886, -0.00001596, 0.00001674, -0.00000514, 0.00000415, -0.00000089, 0.00000227, -0.00000404, 0.00000052, 0.00000139, -0.00000501, 0.00000137, -0.00000040, 0.00000165, 0.00000263, 0.00000287, -0.00002010, },
		{ 0.00002587, -0.00000765, -0.00001529, 0.00003044, -0.00002796, 0.00001405, -0.00000943, 0.00001032, -0.00001118, 0.00000542, -0.00000252, -0.00000009, -0.00000266, 0.00000036, -0.00000598, 0.00000531, -0.00000176, -0.00000025, 0.00000852, -0.00000454, -0.00000966, 0.00002124, -0.00003193, 0.00003265, -0.00002581, 0.00001817, -0.00002473, 0.00002747, -0.00002341, 0.00001866, -0.00002529, 0.00003649, -0.00003387, 0.00002617, -0.00003497, 0.00003890, -0.00004322, 0.00004362, -0.00004699, 0.00004864, -0.00004514, 0.00004687, -0.00005168, 0.00004690, -0.00005518, 0.00006067, -0.00005983, 0.00005516, -0.00006359, 0.00006723, -0.00008021, 0.00007645, -0.00008847, 0.00009270, -0.00010761, 0.00009711, -0.00012759, 0.00015490, -0.00017963, 0.00016220, -0.00022230, 0.00022877, -0.00036121, 0.00033963, -0.00079523, 0.00005995, 0.00223040, -0.00154362, 0.00049872, -0.00046698, 0.00028436, -0.00026712, 0.00018575, -0.00019627, 0.00015284, -0.00012452, 0.00010853, -0.00012135, 0.00010141, -0.00009410, 0.00007890, -0.00008409, 0.00007692, -0.00007412, 0.00006441, -0.00006868, 0.00006100, -0.00005159, 0.00004695, -0.00005169, 0.00004822, -0.00004491, 0.00003933, -0.00003347, 0.00003195, -0.00002973, 0.00002427, -0.00002478, 0.00002672, -0.00003337, 0.00002288, -0.00001555, 0.00002492, -0.00003042, 0.00002367, -0.00001749, 0.00001721, -0.00002315, 0.00001398, -0.00000532, -0.00000084, -0.00000676, 0.00001359, -0.00002304, 0.00003235, -0.00002565, 0.00002524, -0.00002139, 0.00002196, -0.00002467, 0.00002327, -0.00002109, 0.00001413, -0.00001519, 0.00001808, -0.00001685, 0.00002598, -0.00003251, },
		{ -0.00002612, 0.00003883, -0.00004174, 0.00002901, -0.00000658, -0.00000882, 0.00000698, -0.00000583, 0.00000789, -0.00001419, 0.00001257, -0.00000961, 0.00000490, -0.00000684, 0.00000280, -0.00000831, 0.00001038, -0.00000854, 0.00000728, 0.00000511, -0.00001348, 0.00000853, 0.00000016, -0.00001407, 0.00002369, -0.00002374, 0.00001573, -0.00002002, 0.00002553, -0.00002400, 0.00001630, -0.00001750, 0.00002961, -0.00003241, 0.00002326, -0.00002844, 0.00003224, -0.00003769, 0.00003955, -0.00004490, 0.00005053, -0.00004844, 0.00004905, -0.00005684, 0.00005204, -0.00005760, 0.00006582, -0.00007008, 0.00006500, -0.00006965, 0.00007055, -0.00008702, 0.00008585, -0.00009647, 0.00010091, -0.00012264, 0.00011037, -0.00012999, 0.00016073, -0.00021089, 0.00020406, -0.00027300, 0.00028821, -0.00048443, 0.00052627, -0.00151243, 0.00203494, 0.00029445, -0.00090272, 0.00036892, -0.00037321, 0.00024547, -0.00024084, 0.00016570, -0.00017598, 0.00015189, -0.00012599, 0.00009913, -0.00010970, 0.00009749, -0.00009291, 0.00007421, -0.00007588, 0.00007198, -0.00007203, 0.00006140, -0.00006598, 0.00006356, -0.00005569, 0.00004714, -0.00004975, 0.00004948, -0.00004916, 0.00004531, -0.00003897, 0.00003655, -0.00003462, 0.00002720, -0.00002350, 0.00002216, -0.00003158, 0.00002526, -0.00001263, 0.00001650, -0.00002513, 0.00002316, -0.00001696, 0.00001361, -0.00002176, 0.00001725, -0.00000825, -0.00000586, 0.00000747, -0.00000701, 0.00000202, 0.00001108, -0.00000926, 0.00001119, -0.00000805, 0.00000831, -0.00001208, 0.00001456, -0.00001615, 0.00000835, -0.00000502, 0.00000745, -0.00000302, 0.00001066, },
		{ -0.00002706, 0.00001950, -0.00000317, -0.00001806, 0.00003041, -0.00002503, 0.00001050, -0.00000777, 0.00000752, -0.00000751, -0.00000119, 0.00000514, -0.00000537, -0.00000000, -0.00000125, -0.00000542, 0.00000234, 0.00000117, -0.00000187, 0.00000503, 0.00000454, -0.00002101, 0.00002865, -0.00003115, 0.00002425, -0.00001329, 0.00000815, -0.00001599, 0.00001573, -0.00000996, 0.00000874, -0.00001834, 0.00002494, -0.00001719, 0.00001166, -0.00002254, 0.00002344, -0.00002405, 0.00002207, -0.00002380, 0.00002187, -0.00001599, 0.00001870, -0.00002342, 0.00001703, -0.00002401, 0.00002583, -0.00002235, 0.00001800, -0.00002539, 0.00002826, -0.00003962, 0.00003108, -0.00003802, 0.00003822, -0.00004857, 0.00003469, -0.00006024, 0.00007403, -0.00008229, 0.00005187, -0.00009177, 0.00007585, -0.00016324, 0.00007702, -0.00034338, -0.00057949, 0.00246054, -0.00123582, 0.00018298, -0.00026097, 0.00011963, -0.00013411, 0.00007054, -0.00009844, 0.00006540, -0.00005144, 0.00004730, -0.00006346, 0.00004582, -0.00004248, 0.00003163, -0.00004189, 0.00003823, -0.00003591, 0.00002704, -0.00003252, 0.00002477, -0.00001851, 0.00001777, -0.00002370, 0.00002131, -0.00001790, 0.00001221, -0.00000958, 0.00001166, -0.00001107, 0.00000814, -0.00001199, 0.00001595, -0.00002014, 0.00000836, -0.00000649, 0.00001912, -0.00002083, 0.00001135, -0.00000740, 0.00000992, -0.00001459, 0.00000411, -0.00000070, 0.00000229, -0.00001652, 0.00002353, -0.00002908, 0.00003222, -0.00002070, 0.00002030, -0.00001642, 0.00001889, -0.00001980, 0.00001716, -0.00001238, 0.00000460, -0.00000841, 0.00001469, -0.00001340, 0.00002341, },
		{ 0.00001599, -0.00002912, 0.00003892, -0.00004116, 0.00002761, -0.00000592, -0.00000705, 0.00000347, -0.00000402, 0.00000593, -0.00001223, 0.00000761, -0.00000143, -0.00000328, -0.00000044, -0.00000228, -0.00000554, 0.00000584, -0.00000273, 0.00000048, 0.00000859, -0.00000967, -0.00000365, 0.00001643, -0.00002844, 0.00003259, -0.00002650, 0.00001879, -0.00002521, 0.00002877, -0.00002402, 0.00001859, -0.00002561, 0.00003723, -0.00003506, 0.00002693, -0.00003631, 0.00004069, -0.00004469, 0.00004541, -0.00005067, 0.00005380, -0.00004865, 0.00004976, -0.00005860, 0.00005378, -0.00006014, 0.00006658, -0.00006844, 0.00006405, -0.00006919, 0.00007183, -0.00009083, 0.00008956, -0.00009891, 0.00010307, -0.00012433, 0.00011419, -0.00013872, 0.00016830, -0.00021225, 0.00020353, -0.00026689, 0.00028218, -0.00046466, 0.00049304, -0.00128603, 0.00132269, 0.00102907, -0.00117470, 0.00044064, -0.00042291, 0.00027583, -0.00026392, 0.00017992, -0.00018758, 0.00015767, -0.00013259, 0.00011047, -0.00012040, 0.00010534, -0.00009888, 0.00007920, -0.00008210, 0.00007984, -0.00007883, 0.00006678, -0.00007026, 0.00006448, -0.00005572, 0.00004871, -0.00005131, 0.00005098, -0.00004908, 0.00004206, -0.00003529, 0.00003418, -0.00003217, 0.00002537, -0.00002414, 0.00002600, -0.00003379, 0.00002363, -0.00001408, 0.00002298, -0.00003001, 0.00002392, -0.00001722, 0.00001592, -0.00002251, 0.00001323, -0.00000437, -0.00000447, -0.00000296, 0.00000925, -0.00001630, 0.00002693, -0.00002024, 0.00002073, -0.00001627, 0.00001843, -0.00002126, 0.00002283, -0.00002096, 0.00000944, -0.00000577, 0.00001077, -0.00000650, },
		{ 0.00001858, -0.00001961, 0.00001233, 0.00000142, -0.00002131, 0.00003129, -0.00002426, 0.00001025, -0.00000924, 0.00000735, -0.00000634, -0.00000424, 0.00000698, -0.00000354, -0.00000394, 0.00000271, -0.00000914, 0.00000334, -0.00000070, 0.00000204, -0.00000286, 0.00001417, -0.00002742, 0.00002651, -0.00002010, 0.00000816, 0.00000294, -0.00000335, -0.00000533, 0.00000216, 0.00000362, -0.00000090, -0.00000912, 0.00000941, 0.00000173, -0.00000370, -0.00000773, 0.00000408, -0.00000129, -0.00000205, 0.00000121, -0.00000558, 0.00001173, -0.00000683, 0.00000244, -0.00001070, 0.00000502, -0.00000619, 0.00001067, -0.00001390, 0.00000761, -0.00000587, -0.00000272, -0.00001122, 0.00000854, -0.00001124, 0.00000468, -0.00002202, 0.00000268, -0.00000206, 0.00000644, -0.00004543, 0.00002064, -0.00005284, 0.00000071, -0.00013186, -0.00001190, -0.00094202, 0.00232110, -0.00068028, -0.00021831, -0.00002855, -0.00005831, 0.00000542, -0.00004671, -0.00000169, -0.00001908, 0.00001825, -0.00001311, -0.00000449, -0.00001030, 0.00000865, -0.00001448, -0.00000059, -0.00000083, 0.00000326, -0.00001069, 0.00000320, -0.00000963, 0.00001148, -0.00000857, 0.00000200, -0.00000340, 0.00000644, -0.00001072, 0.00000899, -0.00000367, 0.00000320, -0.00000405, -0.00000188, 0.00000585, -0.00000695, -0.00000459, 0.00000099, 0.00001217, -0.00000895, -0.00000150, 0.00000184, 0.00000304, -0.00000636, -0.00000342, 0.00000032, 0.00000686, -0.00002244, 0.00002387, -0.00002273, 0.00001979, -0.00000582, 0.00000780, -0.00000484, 0.00000908, -0.00000789, 0.00000470, 0.00000118, -0.00000716, -0.00000107, 0.00001034, -0.00000808, },
		{ -0.00000622, 0.00001729, -0.00002805, 0.00003609, -0.00003912, 0.00002550, -0.00000528, -0.00000553, 0.00000118, -0.00000350, 0.00000459, -0.00000989, 0.00000205, 0.00000561, -0.00000812, 0.00000167, -0.00000291, -0.00000643, 0.00000449, -0.00000109, 0.00000008, 0.00000386, 0.00000167, -0.00001923, 0.00002976, -0.00003493, 0.00003171, -0.00002085, 0.00001530, -0.00002342, 0.00002425, -0.00001753, 0.00001584, -0.00002686, 0.00003486, -0.00002768, 0.00002186, -0.00003423, 0.00003656, -0.00003743, 0.00003622, -0.00004043, 0.00004032, -0.00003289, 0.00003541, -0.00004491, 0.00003911, -0.00004502, 0.00004871, -0.00004850, 0.00004504, -0.00005024, 0.00005374, -0.00007300, 0.00006869, -0.00007467, 0.00007692, -0.00009590, 0.00008503, -0.00010872, 0.00013015, -0.00016258, 0.00014624, -0.00019664, 0.00020083, -0.00035240, 0.00033858, -0.00093307, 0.00062098, 0.00154200, -0.00117829, 0.00032827, -0.00034946, 0.00021152, -0.00021015, 0.00013041, -0.00014591, 0.00011889, -0.00010181, 0.00008740, -0.00009774, 0.00008199, -0.00007645, 0.00005871, -0.00006505, 0.00006500, -0.00006304, 0.00005038, -0.00005379, 0.00004648, -0.00003935, 0.00003525, -0.00003841, 0.00003843, -0.00003554, 0.00002729, -0.00002265, 0.00002439, -0.00002339, 0.00001849, -0.00002029, 0.00002426, -0.00002950, 0.00001726, -0.00001253, 0.00002490, -0.00002840, 0.00001902, -0.00001365, 0.00001477, -0.00001979, 0.00000833, -0.00000366, 0.00000241, -0.00001726, 0.00002444, -0.00002886, 0.00003386, -0.00002185, 0.00002199, -0.00001696, 0.00002083, -0.00002175, 0.00002181, -0.00001654, 0.00000331, -0.00000186, 0.00001046, },
		{ -0.00000386, 0.00001457, -0.00001324, 0.00000695, 0.00000448, -0.00002409, 0.00003262, -0.00002521, 0.00001239, -0.00001255, 0.00000866, -0.00000704, -0.00000403, 0.00000394, 0.00000301, -0.00001086, 0.00000812, -0.00001315, 0.00000467, -0.00000349, 0.00000643, -0.00000873, 0.00001762, -0.00002396, 0.00001378, -0.00000184, -0.00001007, 0.00001703, -0.00001148, 0.00000289, -0.00000860, 0.00001291, -0.00000672, -0.00000131, -0.00000481, 0.00001652, -0.00001389, 0.00000343, -0.00001189, 0.00001628, -0.00001899, 0.00001771, -0.00002289, 0.00002730, -0.00001964, 0.00001613, -0.00002619, 0.00002117, -0.00002398, 0.00002778, -0.00002943, 0.00002439, -0.00002385, 0.00001825, -0.00003658, 0.00003579, -0.00003915, 0.00003466, -0.00005412, 0.00003960, -0.00004792, 0.00005756, -0.00009782, 0.00007989, -0.00011805, 0.00008527, -0.00023353, 0.00014643, -0.00102071, 0.00189990, -0.00002892, -0.00056633, 0.00013239, -0.00017866, 0.00009570, -0.00012035, 0.00005620, -0.00006926, 0.00005852, -0.00004999, 0.00003411, -0.00004648, 0.00004035, -0.00004229, 0.00002428, -0.00002558, 0.00002888, -0.00003450, 0.00002477, -0.00002944, 0.00002711, -0.00002210, 0.00001611, -0.00001699, 0.00001957, -0.00002184, 0.00001620, -0.00000947, 0.00000950, -0.00000954, 0.00000338, -0.00000120, 0.00000316, -0.00001316, 0.00000516, 0.00000583, 0.00000230, -0.00001160, 0.00000778, -0.00000173, 0.00000007, -0.00000817, -0.00000041, 0.00000891, -0.00002040, 0.00001299, -0.00000609, 0.00000007, 0.00001255, -0.00000586, 0.00000750, -0.00000170, 0.00000478, -0.00000727, 0.00001257, -0.00001486, 0.00000214, 0.00000842, },
		{ 0.00000801, -0.00000312, 0.00001551, -0.00002411, 0.00003155, -0.00003624, 0.00002253, -0.00000366, -0.00000589, 0.00000179, -0.00000571, 0.00000533, -0.00000923, -0.00000103, 0.00000799, -0.00000679, -0.00000214, 0.00000150, -0.00001105, 0.00000646, -0.00000368, 0.00000494, -0.00000550, 0.00001415, -0.00003027, 0.00003318, -0.00002967, 0.00002058, -0.00000838, 0.00000715, -0.00001648, 0.00001439, -0.00000732, 0.00000995, -0.00002236, 0.00002453, -0.00001362, 0.00001136, -0.00002503, 0.00002327, -0.00002067, 0.00001806, -0.00002144, 0.00001870, -0.00001071, 0.00001567, -0.00002532, 0.00001789, -0.00002282, 0.00002388, -0.00002270, 0.00002056, -0.00002533, 0.00002878, -0.00004635, 0.00003730, -0.00003948, 0.00003966, -0.00005570, 0.00004241, -0.00006184, 0.00007246, -0.00009343, 0.00006943, -0.00010633, 0.00009752, -0.00021579, 0.00015852, -0.00059160, 0.00008381, 0.00173995, -0.00091246, 0.00007631, -0.00019935, 0.00009246, -0.00011462, 0.00004856, -0.00007891, 0.00005991, -0.00005408, 0.00004702, -0.00005736, 0.00004259, -0.00004009, 0.00002611, -0.00003670, 0.00003838, -0.00003537, 0.00002314, -0.00002769, 0.00002075, -0.00001711, 0.00001633, -0.00001999, 0.00002060, -0.00001748, 0.00000988, -0.00000906, 0.00001392, -0.00001388, 0.00001097, -0.00001504, 0.00001937, -0.00002146, 0.00000883, -0.00000937, 0.00002281, -0.00002153, 0.00001050, -0.00000804, 0.00001147, -0.00001502, 0.00000367, -0.00000523, 0.00001034, -0.00002782, 0.00003012, -0.00002874, 0.00002741, -0.00001188, 0.00001384, -0.00000919, 0.00001473, -0.00001341, 0.00001245, -0.00000516, -0.00000702, 0.00000418, },
		{ 0.00001075, -0.00000274, 0.00001366, -0.00001017, 0.00000502, 0.00000498, -0.00002563, 0.00003346, -0.00002646, 0.00001475, -0.00001521, 0.00000916, -0.00000758, -0.00000282, -0.00000057, 0.00000956, -0.00001557, 0.00001014, -0.00001334, 0.00000282, -0.00000358, 0.00000710, -0.00000836, 0.00001233, -0.00001146, -0.00000472, 0.00001619, -0.00002288, 0.00002306, -0.00001231, 0.00000551, -0.00001333, 0.00001522, -0.00000673, 0.00000249, -0.00001358, 0.00002272, -0.00001525, 0.00000718, -0.00001953, 0.00002293, -0.00002314, 0.00002045, -0.00002526, 0.00002690, -0.00001701, 0.00001530, -0.00002683, 0.00002156, -0.00002473, 0.00002675, -0.00002701, 0.00002349, -0.00002389, 0.00002068, -0.00004136, 0.00003957, -0.00004141, 0.00003727, -0.00005704, 0.00004463, -0.00005626, 0.00006381, -0.00009918, 0.00008122, -0.00011644, 0.00008909, -0.00023148, 0.00015456, -0.00089639, 0.00135435, 0.00054315, -0.00073944, 0.00015580, -0.00020173, 0.00010858, -0.00012947, 0.00005964, -0.00007345, 0.00006022, -0.00005487, 0.00004338, -0.00005389, 0.00004483, -0.00004512, 0.00002680, -0.00003012, 0.00003477, -0.00003842, 0.00002691, -0.00003043, 0.00002537, -0.00002059, 0.00001650, -0.00001736, 0.00001976, -0.00002035, 0.00001245, -0.00000675, 0.00000889, -0.00000931, 0.00000437, -0.00000484, 0.00000899, -0.00001646, 0.00000563, 0.00000126, 0.00001053, -0.00001668, 0.00000918, -0.00000361, 0.00000400, -0.00001022, -0.00000145, 0.00000695, -0.00001122, -0.00000427, 0.00001263, -0.00001733, 0.00002507, -0.00001268, 0.00001366, -0.00000680, 0.00001151, -0.00001218, 0.00001594, -0.00001414, -0.00000143, },
		{ 0.00000785, 0.00000660, 0.00000011, 0.00001338, -0.00002007, 0.00002765, -0.00003402, 0.00001942, -0.00000121, -0.00000820, 0.00000497, -0.00000995, 0.00000758, -0.00001025, -0.00000114, 0.00000566, -0.00000063, -0.00000927, 0.00000770, -0.00001612, 0.00000876, -0.00000723, 0.00001051, -0.00001317, 0.00002056, -0.00003073, 0.00002415, -0.00001418, 0.00000354, 0.00000575, -0.00000117, -0.00000846, 0.00000367, 0.00000225, 0.00000436, -0.00001549, 0.00001114, 0.00000107, 0.00000121, -0.00001440, 0.00000770, -0.00000289, 0.00000050, -0.00000409, 0.00000008, 0.00000658, 0.00000133, -0.00001037, 0.00000101, -0.00000520, 0.00000448, -0.00000365, 0.00000302, -0.00000683, 0.00000957, -0.00002440, 0.00001053, -0.00001020, 0.00000947, -0.00002314, 0.00000734, -0.00002199, 0.00002350, -0.00003753, 0.00001000, -0.00003768, 0.00002066, -0.00011358, 0.00002833, -0.00035401, -0.00021880, 0.00163690, -0.00046968, -0.00021059, -0.00005375, -0.00001943, -0.00002787, -0.00002318, -0.00002240, 0.00001083, -0.00001448, 0.00001114, -0.00001982, 0.00000689, -0.00000825, -0.00000180, -0.00001224, 0.00001425, -0.00000997, -0.00000092, -0.00000562, 0.00000014, -0.00000056, 0.00000220, -0.00000561, 0.00000674, -0.00000385, -0.00000202, -0.00000120, 0.00000794, -0.00000774, 0.00000589, -0.00001068, 0.00001364, -0.00001254, 0.00000096, -0.00000609, 0.00001806, -0.00001185, 0.00000123, -0.00000257, 0.00000749, -0.00000943, -0.00000041, -0.00000700, 0.00001470, -0.00002922, 0.00002286, -0.00001585, 0.00001039, 0.00000514, 0.00000111, 0.00000252, 0.00000470, -0.00000136, 0.00000066, 0.00000682, -0.00001563, },
		{ -0.00001028, 0.00001601, -0.00000422, 0.00001586, -0.00001034, 0.00000611, 0.00000352, -0.00002614, 0.00003338, -0.00002682, 0.00001548, -0.00001528, 0.00000720, -0.00000636, -0.00000278, -0.00000339, 0.00001231, -0.00001478, 0.00000657, -0.00000856, -0.00000277, -0.00000035, 0.00000342, -0.00000222, 0.00000129, 0.00000392, -0.00002053, 0.00002574, -0.00002442, 0.00001830, -0.00000522, 0.00000234, -0.00001179, 0.00001087, -0.00000166, 0.00000208, -0.00001565, 0.00001972, -0.00000836, 0.00000393, -0.00001850, 0.00001855, -0.00001546, 0.00001150, -0.00001582, 0.00001490, -0.00000426, 0.00000532, -0.00001779, 0.00001143, -0.00001420, 0.00001422, -0.00001389, 0.00001215, -0.00001296, 0.00001112, -0.00003192, 0.00002687, -0.00002606, 0.00002145, -0.00004022, 0.00002746, -0.00003790, 0.00003898, -0.00006722, 0.00004646, -0.00007463, 0.00004545, -0.00017052, 0.00008280, -0.00068959, 0.00084623, 0.00090398, -0.00068416, 0.00003966, -0.00013843, 0.00005632, -0.00008670, 0.00002131, -0.00004375, 0.00003411, -0.00003620, 0.00002956, -0.00003817, 0.00002798, -0.00002900, 0.00001259, -0.00001921, 0.00002500, -0.00002664, 0.00001435, -0.00001793, 0.00001208, -0.00000965, 0.00000820, -0.00000918, 0.00001174, -0.00001143, 0.00000322, -0.00000056, 0.00000549, -0.00000661, 0.00000332, -0.00000605, 0.00001085, -0.00001509, 0.00000314, -0.00000110, 0.00001441, -0.00001604, 0.00000622, -0.00000276, 0.00000526, -0.00000974, -0.00000257, 0.00000240, 0.00000023, -0.00001953, 0.00002370, -0.00002364, 0.00002502, -0.00000805, 0.00001024, -0.00000314, 0.00000941, -0.00000779, 0.00001023, -0.00000542, },
		{ -0.00001727, 0.00000571, 0.00000870, 0.00000101, 0.00001357, -0.00001850, 0.00002639, -0.00003371, 0.00001703, 0.00000122, -0.00001123, 0.00000886, -0.00001403, 0.00000937, -0.00001133, -0.00000002, 0.00000137, 0.00000628, -0.00001495, 0.00001106, -0.00001762, 0.00000793, -0.00000812, 0.00001236, -0.00001436, 0.00001735, -0.00002021, 0.00000639, 0.00000469, -0.00001169, 0.00001492, -0.00000475, -0.00000358, -0.00000357, 0.00000725, 0.00000223, -0.00000992, -0.00000001, 0.00001054, -0.00000343, -0.00000763, -0.00000353, 0.00000821, -0.00000852, 0.00000376, -0.00000785, 0.00001201, -0.00000157, -0.00000583, -0.00000540, 0.00000130, -0.00000271, 0.00000220, -0.00000162, -0.00000069, 0.00000252, -0.00001455, -0.00000279, 0.00000334, -0.00000329, -0.00000913, -0.00000817, -0.00000334, 0.00000004, -0.00001319, -0.00001290, -0.00001093, -0.00000762, -0.00007092, -0.00002316, -0.00024364, -0.00030841, 0.00133559, 0.00000430, -0.00041384, 0.00001212, -0.00007315, 0.00001119, -0.00005399, -0.00000098, -0.00000863, 0.00000074, -0.00000555, 0.00000027, -0.00001098, 0.00000625, -0.00001367, -0.00000190, 0.00000252, 0.00000319, -0.00001210, 0.00000348, -0.00000748, 0.00000385, -0.00000168, -0.00000039, 0.00000178, 0.00000087, -0.00000491, -0.00000115, 0.00000763, -0.00000577, 0.00000386, -0.00000791, 0.00000847, -0.00000490, -0.00000446, -0.00000394, 0.00001243, -0.00000230, -0.00000584, 0.00000080, 0.00000422, -0.00000428, -0.00000359, -0.00000766, 0.00001354, -0.00002131, 0.00000626, 0.00000272, -0.00000834, 0.00002013, -0.00000807, 0.00001093, -0.00000227, 0.00000721, -0.00000640, 0.00001276, },
		{ 0.00000613, -0.00001935, 0.00002035, -0.00000569, 0.00001903, -0.00001156, 0.00000792, 0.00000208, -0.00002682, 0.00003282, -0.00002610, 0.00001411, -0.00001260, 0.00000298, -0.00000351, -0.00000432, -0.00000350, 0.00001051, -0.00000898, -0.00000079, -0.00000143, -0.00000930, 0.00000359, -0.00000172, 0.00000539, -0.00000909, 0.00001428, -0.00002624, 0.00002227, -0.00001402, 0.00000511, 0.00000622, -0.00000347, -0.00000675, 0.00000308, 0.00000528, -0.00000031, -0.00001297, 0.00001077, 0.00000239, -0.00000222, -0.00001263, 0.00000797, -0.00000232, -0.00000162, -0.00000295, 0.00000057, 0.00000903, -0.00000465, -0.00000790, -0.00000010, -0.00000216, 0.00000069, -0.00000084, 0.00000091, -0.00000140, -0.00000006, -0.00001919, 0.00000998, -0.00000692, 0.00000217, -0.00001959, 0.00000531, -0.00001256, 0.00000681, -0.00002995, 0.00000770, -0.00002956, -0.00000192, -0.00010543, 0.00000643, -0.00050588, 0.00048371, 0.00100867, -0.00043502, -0.00015301, -0.00004308, -0.00001783, -0.00002848, -0.00002741, -0.00000702, 0.00000276, -0.00001249, 0.00000845, -0.00001431, 0.00000450, -0.00000782, -0.00000567, -0.00000418, 0.00001018, -0.00000979, -0.00000213, -0.00000270, -0.00000241, 0.00000139, -0.00000061, -0.00000008, 0.00000297, -0.00000245, -0.00000465, 0.00000356, 0.00000345, -0.00000450, 0.00000227, -0.00000600, 0.00000975, -0.00001060, -0.00000096, -0.00000154, 0.00001394, -0.00001068, 0.00000048, -0.00000022, 0.00000429, -0.00000702, -0.00000420, -0.00000192, 0.00000825, -0.00002577, 0.00002153, -0.00001621, 0.00001255, 0.00000593, -0.00000010, 0.00000667, 0.00000128, 0.00000243, -0.00000016, },
		{ 0.00000984, -0.00001070, -0.00000216, 0.00001407, -0.00000081, 0.00001688, -0.00002009, 0.00002797, -0.00003523, 0.00001551, 0.00000302, -0.00001371, 0.00001152, -0.00001588, 0.00000891, -0.00001081, 0.00000022, -0.00000174, 0.00001000, -0.00001558, 0.00000891, -0.00001393, 0.00000295, -0.00000526, 0.00000925, -0.00000874, 0.00000666, -0.00000428, -0.00001166, 0.00001793, -0.00001779, 0.00001453, -0.00000113, -0.00000382, -0.00000526, 0.00000617, 0.00000458, -0.00000760, -0.00000574, 0.00001196, -0.00000074, -0.00000678, -0.00000738, 0.00000943, -0.00000655, 0.00000053, -0.00000436, 0.00000599, 0.00000563, -0.00001031, -0.00000239, -0.00000240, 0.00000108, -0.00000324, 0.00000426, -0.00000469, 0.00000588, -0.00001581, -0.00000298, 0.00000148, 0.00000057, -0.00001256, -0.00000502, -0.00000562, 0.00000192, -0.00001883, -0.00000313, -0.00002009, 0.00000455, -0.00007680, -0.00001267, -0.00022595, -0.00026785, 0.00097579, 0.00037091, -0.00045335, -0.00003322, -0.00005499, -0.00000533, -0.00003955, -0.00001637, 0.00000184, -0.00000836, -0.00000205, 0.00000057, -0.00000865, 0.00000185, -0.00000862, -0.00000637, 0.00000429, 0.00000258, -0.00000921, -0.00000086, -0.00000256, -0.00000258, 0.00000322, -0.00000317, 0.00000467, -0.00000203, -0.00000081, -0.00000628, 0.00001043, -0.00000601, 0.00000353, -0.00000620, 0.00000445, 0.00000019, -0.00000649, -0.00000361, 0.00000766, 0.00000446, -0.00000860, 0.00000087, 0.00000272, -0.00000078, -0.00000514, -0.00000748, 0.00000841, -0.00000890, -0.00001115, 0.00001670, -0.00001893, 0.00002479, -0.00000734, 0.00001087, -0.00000134, 0.00000773, -0.00000474, },
		{ -0.00000815, 0.00001374, -0.00002304, 0.00001992, -0.00000440, 0.00002056, -0.00001104, 0.00000777, 0.00000276, -0.00002901, 0.00003265, -0.00002500, 0.00001155, -0.00000868, -0.00000163, -0.00000071, -0.00000616, -0.00000215, 0.00000632, -0.00000172, -0.00000754, 0.00000357, -0.00001275, 0.00000471, -0.00000452, 0.00000964, -0.00001321, 0.00001473, -0.00001971, 0.00000773, 0.00000296, -0.00000968, 0.00001574, -0.00000712, -0.00000240, -0.00000377, 0.00001004, -0.00000151, -0.00000903, 0.00000100, 0.00001121, -0.00000610, -0.00000716, -0.00000222, 0.00000842, -0.00001054, 0.00000480, -0.00000751, 0.00001481, -0.00000742, -0.00000408, -0.00000576, 0.00000360, -0.00000572, 0.00000425, -0.00000277, 0.00000336, -0.00000508, -0.00001194, -0.00000085, 0.00000454, -0.00000833, -0.00000817, -0.00000753, 0.00000305, -0.00001289, -0.00000959, -0.00001096, -0.00000765, -0.00002292, -0.00007237, -0.00002840, -0.00039958, 0.00029474, 0.00090313, -0.00009486, -0.00032485, 0.00001680, -0.00006555, 0.00000706, -0.00005549, 0.00001152, -0.00001330, -0.00000065, -0.00000480, 0.00000274, -0.00001105, 0.00000505, -0.00001604, 0.00000423, 0.00000061, 0.00000170, -0.00001226, 0.00000564, -0.00000944, 0.00000525, -0.00000367, 0.00000411, -0.00000094, 0.00000138, -0.00000691, 0.00000277, 0.00000452, -0.00000417, 0.00000200, -0.00000531, 0.00000684, -0.00000485, -0.00000503, -0.00000093, 0.00001042, -0.00000301, -0.00000553, 0.00000235, 0.00000217, -0.00000299, -0.00000639, -0.00000412, 0.00000980, -0.00002119, 0.00000781, -0.00000018, -0.00000517, 0.00002142, -0.00001004, 0.00001599, -0.00000632, 0.00001162, },
		{ 0.00000402, 0.00000078, 0.00000032, -0.00001177, 0.00001966, -0.00000340, 0.00002182, -0.00002323, 0.00003063, -0.00003700, 0.00001392, 0.00000438, -0.00001518, 0.00001212, -0.00001494, 0.00000601, -0.00000847, -0.00000116, -0.00000225, 0.00000923, -0.00001091, 0.00000232, -0.00000703, -0.00000384, -0.00000081, 0.00000362, -0.00000022, -0.00000523, 0.00000873, -0.00002142, 0.00001914, -0.00001184, 0.00000475, 0.00000806, -0.00000772, -0.00000253, 0.00000062, 0.00000974, -0.00000787, -0.00000617, 0.00000653, 0.00000696, -0.00000993, -0.00000525, 0.00000294, 0.00000281, -0.00000907, 0.00000494, -0.00000498, 0.00001604, -0.00001730, 0.00000391, -0.00001003, 0.00000905, -0.00001252, 0.00001301, -0.00001148, 0.00001268, -0.00002152, 0.00000312, -0.00000800, 0.00001212, -0.00002371, 0.00000654, -0.00001795, 0.00001650, -0.00003859, 0.00002002, -0.00004267, 0.00003102, -0.00009995, 0.00001972, -0.00024188, -0.00019258, 0.00067597, 0.00055081, -0.00032050, -0.00016245, 0.00000632, -0.00005331, -0.00000115, -0.00004957, 0.00002567, -0.00002799, 0.00001087, -0.00001024, 0.00000552, -0.00001286, 0.00000500, -0.00001805, 0.00001295, -0.00000572, 0.00000144, -0.00001187, 0.00000785, -0.00001284, 0.00001056, -0.00000849, 0.00001036, -0.00000748, 0.00000510, -0.00001168, 0.00001244, -0.00000576, 0.00000320, -0.00000480, 0.00000179, 0.00000224, -0.00000512, -0.00000519, 0.00000485, 0.00000696, -0.00000647, -0.00000225, 0.00000323, 0.00000034, -0.00000436, -0.00000770, 0.00000274, 0.00000145, -0.00002098, 0.00001859, -0.00001638, 0.00001669, 0.00000372, 0.00000274, 0.00000693, 0.00000105, },
		{ 0.00001406, -0.00000854, 0.00001326, -0.00001882, 0.00001368, 0.00000028, 0.00001936, -0.00000737, 0.00000442, 0.00000630, -0.00003302, 0.00003320, -0.00002422, 0.00000912, -0.00000549, -0.00000450, 0.00000019, -0.00000663, -0.00000141, 0.00000283, 0.00000301, -0.00000976, 0.00000328, -0.00001088, 0.00000139, -0.00000323, 0.00000845, -0.00000962, 0.00000610, -0.00000527, -0.00001029, 0.00001760, -0.00001814, 0.00001768, -0.00000498, -0.00000178, -0.00000648, 0.00001000, 0.00000054, -0.00000672, -0.00000539, 0.00001400, -0.00000453, -0.00000555, -0.00000740, 0.00001163, -0.00001063, 0.00000348, -0.00000604, 0.00001078, -0.00000159, -0.00000764, -0.00000384, 0.00000115, -0.00000321, 0.00000011, 0.00000197, 0.00000029, -0.00000235, -0.00001269, -0.00000211, 0.00000443, -0.00000620, -0.00000991, -0.00000643, 0.00000302, -0.00001318, -0.00001279, -0.00000381, -0.00001404, -0.00001241, -0.00007685, -0.00001612, -0.00036849, 0.00023823, 0.00069034, 0.00021313, -0.00039072, -0.00000281, -0.00006097, 0.00000145, -0.00004941, 0.00000211, -0.00000689, -0.00000654, -0.00000407, 0.00000593, -0.00001185, 0.00000382, -0.00001373, 0.00000173, 0.00000072, 0.00000313, -0.00001171, 0.00000351, -0.00000648, 0.00000044, 0.00000005, 0.00000225, 0.00000114, -0.00000068, -0.00000365, -0.00000199, 0.00000755, -0.00000478, 0.00000203, -0.00000411, 0.00000326, 0.00000031, -0.00000752, -0.00000051, 0.00000595, 0.00000390, -0.00000914, 0.00000321, 0.00000039, 0.00000080, -0.00000820, -0.00000435, 0.00000587, -0.00000981, -0.00000964, 0.00001453, -0.00001816, 0.00002945, -0.00001224, 0.00001857, -0.00000748, },
		{ -0.00000727, 0.00001387, -0.00000801, 0.00000939, -0.00001762, 0.00002159, -0.00000405, 0.00002587, -0.00002525, 0.00003173, -0.00003691, 0.00001092, 0.00000609, -0.00001614, 0.00001128, -0.00001235, 0.00000220, -0.00000571, -0.00000318, -0.00000101, 0.00000558, -0.00000396, -0.00000471, -0.00000122, -0.00000843, 0.00000161, -0.00000056, 0.00000599, -0.00001191, 0.00001254, -0.00001883, 0.00000806, 0.00000282, -0.00000915, 0.00001758, -0.00001126, 0.00000118, -0.00000571, 0.00001428, -0.00000824, -0.00000388, -0.00000164, 0.00001488, -0.00001288, -0.00000131, -0.00000582, 0.00001276, -0.00001755, 0.00001227, -0.00001287, 0.00002198, -0.00001994, 0.00000698, -0.00001478, 0.00001393, -0.00001801, 0.00001716, -0.00001405, 0.00001608, -0.00002466, 0.00000776, -0.00001599, 0.00002107, -0.00003155, 0.00001479, -0.00002745, 0.00002867, -0.00005428, 0.00003578, -0.00005582, 0.00004645, -0.00011111, 0.00003815, -0.00024305, -0.00014880, 0.00049456, 0.00054491, -0.00007674, -0.00030466, 0.00005804, -0.00009375, 0.00002968, -0.00007446, 0.00004145, -0.00004081, 0.00001921, -0.00001981, 0.00001922, -0.00002569, 0.00001583, -0.00002679, 0.00001949, -0.00001306, 0.00001100, -0.00002068, 0.00001522, -0.00001915, 0.00001379, -0.00001079, 0.00001368, -0.00001047, 0.00000812, -0.00001345, 0.00001108, -0.00000361, 0.00000213, -0.00000357, 0.00000065, 0.00000145, -0.00000123, -0.00000796, 0.00000408, 0.00000541, -0.00000075, -0.00000704, 0.00000495, -0.00000066, -0.00000138, -0.00000909, -0.00000064, 0.00000535, -0.00001956, 0.00000788, -0.00000347, 0.00000077, 0.00001909, -0.00000759, 0.00001707, },
		{ -0.00000085, 0.00000879, -0.00000119, 0.00000573, -0.00000898, 0.00000448, 0.00000615, 0.00001667, -0.00000137, -0.00000119, 0.00001145, -0.00003765, 0.00003383, -0.00002375, 0.00000748, -0.00000415, -0.00000469, -0.00000139, -0.00000517, -0.00000233, 0.00000183, 0.00000333, -0.00000650, -0.00000203, -0.00000498, -0.00000467, 0.00000060, 0.00000356, -0.00000150, -0.00000591, 0.00000884, -0.00002256, 0.00002214, -0.00001550, 0.00001029, 0.00000296, -0.00000495, -0.00000477, 0.00000534, 0.00000543, -0.00000665, -0.00000700, 0.00001025, 0.00000199, -0.00000773, -0.00000688, 0.00000714, -0.00000296, -0.00000481, 0.00000202, 0.00000084, 0.00000843, -0.00001454, 0.00000204, -0.00000588, 0.00000419, -0.00000865, 0.00001035, -0.00000619, 0.00000424, -0.00001810, 0.00000307, -0.00000360, 0.00000410, -0.00001994, 0.00000363, -0.00000761, -0.00000005, -0.00003122, 0.00001824, -0.00003566, 0.00001430, -0.00010057, 0.00001967, -0.00037511, 0.00024123, 0.00047900, 0.00039743, -0.00031650, -0.00009916, -0.00001402, -0.00003576, -0.00001863, -0.00002602, 0.00001363, -0.00002334, 0.00000604, -0.00000174, -0.00000087, -0.00000795, -0.00000252, -0.00000821, 0.00000779, -0.00000316, -0.00000317, -0.00000576, 0.00000258, -0.00000885, 0.00000675, -0.00000251, 0.00000632, -0.00000560, 0.00000178, -0.00000735, 0.00000988, -0.00000468, 0.00000161, -0.00000256, 0.00000012, 0.00000338, -0.00000734, -0.00000134, 0.00000259, 0.00000756, -0.00000865, 0.00000136, 0.00000013, 0.00000281, -0.00000824, -0.00000426, 0.00000013, 0.00000145, -0.00002160, 0.00001915, -0.00001965, 0.00002545, -0.00000411, 0.00001278, },
		{ 0.00002244, -0.00001084, 0.00001851, -0.00001057, 0.00001148, -0.00001623, 0.00001797, -0.00000145, 0.00002742, -0.00002425, 0.00002954, -0.00003379, 0.00000588, 0.00000867, -0.00001732, 0.00001026, -0.00001004, -0.00000038, -0.00000441, -0.00000418, -0.00000001, 0.00000206, 0.00000125, -0.00000799, -0.00000017, -0.00000799, -0.00000028, -0.00000079, 0.00000680, -0.00001058, 0.00000631, -0.00000646, -0.00000915, 0.00001814, -0.00001939, 0.00002147, -0.00001042, 0.00000244, -0.00000923, 0.00001513, -0.00000638, -0.00000189, -0.00000817, 0.00001856, -0.00001203, 0.00000060, -0.00001175, 0.00001743, -0.00001928, 0.00001245, -0.00001300, 0.00001968, -0.00001539, 0.00000411, -0.00001357, 0.00001230, -0.00001634, 0.00001383, -0.00000994, 0.00001352, -0.00002218, 0.00000691, -0.00001737, 0.00002167, -0.00003012, 0.00001348, -0.00002700, 0.00002973, -0.00005566, 0.00003349, -0.00004906, 0.00004008, -0.00009843, 0.00002857, -0.00021154, -0.00015572, 0.00042456, 0.00041864, 0.00017934, -0.00038211, 0.00005473, -0.00009745, 0.00003104, -0.00007421, 0.00003642, -0.00003706, 0.00001498, -0.00002030, 0.00002386, -0.00002823, 0.00001652, -0.00002632, 0.00001830, -0.00001385, 0.00001360, -0.00002168, 0.00001466, -0.00001762, 0.00001015, -0.00000790, 0.00001250, -0.00000893, 0.00000656, -0.00001079, 0.00000658, -0.00000020, 0.00000092, -0.00000291, 0.00000103, -0.00000132, 0.00000356, -0.00001052, 0.00000433, 0.00000157, 0.00000583, -0.00001103, 0.00000627, -0.00000248, 0.00000251, -0.00001105, -0.00000130, 0.00000272, -0.00000955, -0.00000895, 0.00001101, -0.00001370, 0.00002997, -0.00001256, },
		{ 0.00000276, 0.00000958, 0.00000006, 0.00000909, -0.00000340, 0.00000076, -0.00000256, 0.00000957, 0.00001498, 0.00000472, -0.00000660, 0.00001572, -0.00004086, 0.00003322, -0.00002290, 0.00000633, -0.00000432, -0.00000300, -0.00000428, -0.00000286, -0.00000425, 0.00000288, 0.00000032, -0.00000015, -0.00000888, 0.00000100, -0.00000963, 0.00000346, -0.00000112, 0.00000580, -0.00001433, 0.00001524, -0.00002331, 0.00001423, -0.00000297, -0.00000291, 0.00001282, -0.00000894, -0.00000105, -0.00000115, 0.00001058, -0.00000725, -0.00000531, 0.00000276, 0.00000991, -0.00001079, -0.00000337, -0.00000122, 0.00000724, -0.00001402, 0.00001019, -0.00000814, 0.00001601, -0.00001888, 0.00000638, -0.00001179, 0.00001034, -0.00001555, 0.00001604, -0.00001026, 0.00000917, -0.00002272, 0.00000874, -0.00001248, 0.00001449, -0.00002951, 0.00001343, -0.00001878, 0.00001452, -0.00005012, 0.00003805, -0.00005347, 0.00003607, -0.00011910, 0.00004896, -0.00037778, 0.00023986, 0.00034003, 0.00042984, -0.00013122, -0.00022508, 0.00003754, -0.00007587, 0.00001309, -0.00005280, 0.00003173, -0.00003785, 0.00001525, -0.00001148, 0.00001293, -0.00002109, 0.00000895, -0.00001789, 0.00001509, -0.00001078, 0.00000665, -0.00001522, 0.00001089, -0.00001625, 0.00001107, -0.00000571, 0.00001047, -0.00000933, 0.00000548, -0.00001007, 0.00000950, -0.00000298, 0.00000058, -0.00000117, -0.00000156, 0.00000353, -0.00000434, -0.00000369, 0.00000137, 0.00000705, -0.00000418, -0.00000275, 0.00000155, 0.00000241, -0.00000582, -0.00000537, -0.00000380, 0.00000699, -0.00002272, 0.00001109, -0.00001008, 0.00001200, 0.00001039, },
		{ -0.00000791, 0.00001978, -0.00000675, 0.00001545, -0.00000522, 0.00000609, -0.00000849, 0.00001055, 0.00000305, 0.00002700, -0.00002036, 0.00002437, -0.00002833, -0.00000042, 0.00001174, -0.00001890, 0.00000989, -0.00000934, -0.00000049, -0.00000549, -0.00000329, -0.00000058, 0.00000080, 0.00000233, -0.00000584, -0.00000443, -0.00000294, -0.00000558, 0.00000223, 0.00000300, -0.00000334, -0.00000527, 0.00000813, -0.00002335, 0.00002570, -0.00001999, 0.00001659, -0.00000389, 0.00000011, -0.00000858, 0.00001148, -0.00000182, -0.00000163, -0.00001071, 0.00001633, -0.00000667, -0.00000070, -0.00001270, 0.00001490, -0.00001339, 0.00000558, -0.00000629, 0.00001102, -0.00000607, -0.00000253, -0.00000811, 0.00000586, -0.00000951, 0.00000553, -0.00000184, 0.00000733, -0.00001584, 0.00000163, -0.00001268, 0.00001468, -0.00002071, 0.00000407, -0.00001787, 0.00002052, -0.00004391, 0.00001596, -0.00002707, 0.00001802, -0.00006914, -0.00000041, -0.00015827, -0.00019449, 0.00041721, 0.00026176, 0.00035821, -0.00034877, -0.00001547, -0.00006239, 0.00000294, -0.00005039, 0.00001327, -0.00001964, 0.00000045, -0.00001233, 0.00001875, -0.00001992, 0.00000729, -0.00001736, 0.00001008, -0.00000823, 0.00000905, -0.00001507, 0.00000709, -0.00000990, 0.00000182, -0.00000188, 0.00000838, -0.00000429, 0.00000212, -0.00000583, 0.00000130, 0.00000252, 0.00000069, -0.00000320, 0.00000251, -0.00000463, 0.00000732, -0.00001133, 0.00000413, -0.00000218, 0.00001035, -0.00001196, 0.00000563, -0.00000347, 0.00000538, -0.00001195, -0.00000086, -0.00000301, 0.00000230, -0.00002255, 0.00001771, -0.00001888, 0.00003028, },
		{ 0.00002327, -0.00000433, 0.00001736, -0.00000581, 0.00001590, -0.00000790, 0.00000488, -0.00000337, 0.00000807, 0.00001605, 0.00000896, -0.00000961, 0.00001705, -0.00004114, 0.00003045, -0.00002099, 0.00000491, -0.00000486, -0.00000122, -0.00000642, -0.00000156, -0.00000556, 0.00000415, -0.00000322, 0.00000537, -0.00001295, 0.00000305, -0.00001016, 0.00000253, -0.00000243, 0.00000830, -0.00001509, 0.00001137, -0.00001319, -0.00000183, 0.00001277, -0.00001487, 0.00001885, -0.00000978, 0.00000130, -0.00000594, 0.00001292, -0.00000630, -0.00000301, -0.00000440, 0.00001500, -0.00001134, -0.00000045, -0.00000855, 0.00001408, -0.00001828, 0.00001289, -0.00001094, 0.00001671, -0.00001714, 0.00000587, -0.00001293, 0.00001127, -0.00001660, 0.00001549, -0.00000880, 0.00000934, -0.00002307, 0.00001060, -0.00001676, 0.00001877, -0.00003216, 0.00001614, -0.00002276, 0.00002116, -0.00005832, 0.00004363, -0.00005527, 0.00004006, -0.00011811, 0.00005467, -0.00035406, 0.00020434, 0.00028916, 0.00034493, 0.00009145, -0.00031488, 0.00005213, -0.00009087, 0.00002412, -0.00006123, 0.00003430, -0.00003998, 0.00001550, -0.00001557, 0.00002115, -0.00002727, 0.00001322, -0.00002092, 0.00001697, -0.00001412, 0.00001184, -0.00001909, 0.00001322, -0.00001746, 0.00000983, -0.00000486, 0.00001118, -0.00000954, 0.00000552, -0.00000906, 0.00000635, -0.00000027, -0.00000032, -0.00000068, -0.00000126, 0.00000099, 0.00000051, -0.00000674, 0.00000192, 0.00000340, 0.00000231, -0.00000727, 0.00000353, 0.00000035, -0.00000169, -0.00000766, -0.00000463, 0.00000545, -0.00001406, -0.00000485, 0.00000337, -0.00000292, },
		{ 0.00002035, 0.00000448, 0.00001143, 0.00000264, 0.00000745, 0.00000458, -0.00000249, 0.00000076, 0.00000380, 0.00000620, 0.00002673, -0.00001542, 0.00001826, -0.00002271, -0.00000614, 0.00001409, -0.00002030, 0.00001008, -0.00001025, 0.00000148, -0.00000819, -0.00000121, -0.00000246, 0.00000184, -0.00000029, 0.00000008, -0.00001123, 0.00000329, -0.00001086, 0.00000534, -0.00000187, 0.00000472, -0.00001516, 0.00001690, -0.00002746, 0.00002132, -0.00001018, 0.00000473, 0.00000573, -0.00000391, -0.00000520, 0.00000519, 0.00000365, -0.00000232, -0.00000974, 0.00000983, 0.00000083, -0.00000352, -0.00000998, 0.00000742, -0.00000354, -0.00000379, 0.00000220, 0.00000156, 0.00000251, -0.00000812, -0.00000280, -0.00000088, -0.00000244, -0.00000246, 0.00000511, 0.00000215, -0.00000986, -0.00000413, -0.00000612, 0.00000516, -0.00000925, -0.00000703, -0.00000675, 0.00000832, -0.00002782, -0.00000555, -0.00000327, -0.00000472, -0.00003996, -0.00002951, -0.00010673, -0.00023245, 0.00041751, 0.00014604, 0.00041568, -0.00021085, -0.00012408, -0.00001377, -0.00003487, -0.00001988, -0.00001338, -0.00000071, -0.00001486, -0.00000290, 0.00000951, -0.00000668, -0.00000551, -0.00000606, 0.00000029, -0.00000082, 0.00000168, -0.00000556, -0.00000239, -0.00000124, -0.00000619, 0.00000306, 0.00000475, 0.00000029, -0.00000206, -0.00000174, -0.00000208, 0.00000294, 0.00000196, -0.00000422, 0.00000410, -0.00000689, 0.00000849, -0.00000940, 0.00000246, -0.00000402, 0.00001100, -0.00000896, 0.00000255, -0.00000270, 0.00000586, -0.00001042, -0.00000136, -0.00000770, 0.00000956, -0.00002621, 0.00001234, -0.00001316, },
		{ -0.00001110, 0.00002757, -0.00000311, 0.00001804, -0.00000467, 0.00001551, -0.00000479, 0.00000160, 0.00000239, 0.00000234, 0.00001929, 0.00001122, -0.00000971, 0.00001510, -0.00003860, 0.00002585, -0.00001811, 0.00000287, -0.00000476, -0.00000088, -0.00000629, -0.00000246, -0.00000514, 0.00000405, -0.00000484, 0.00000720, -0.00001190, -0.00000023, -0.00000582, -0.00000217, -0.00000016, 0.00000566, -0.00000896, 0.00000055, 0.00000144, -0.00001755, 0.00002322, -0.00001892, 0.00001696, -0.00000528, 0.00000041, -0.00000695, 0.00001096, -0.00000279, -0.00000194, -0.00000839, 0.00001492, -0.00000794, -0.00000003, -0.00001180, 0.00001447, -0.00001534, 0.00000873, -0.00000690, 0.00001082, -0.00001027, 0.00000128, -0.00000965, 0.00000723, -0.00001229, 0.00000969, -0.00000305, 0.00000568, -0.00001955, 0.00000821, -0.00001535, 0.00001583, -0.00002706, 0.00001095, -0.00001834, 0.00001793, -0.00005373, 0.00003398, -0.00004142, 0.00002756, -0.00009933, 0.00003899, -0.00030665, 0.00014042, 0.00029788, 0.00021215, 0.00027075, -0.00031882, 0.00000992, -0.00007158, 0.00000861, -0.00004789, 0.00001984, -0.00002924, 0.00000624, -0.00001228, 0.00002089, -0.00002374, 0.00000841, -0.00001607, 0.00001240, -0.00001176, 0.00001064, -0.00001595, 0.00000889, -0.00001259, 0.00000384, -0.00000081, 0.00000900, -0.00000669, 0.00000266, -0.00000565, 0.00000223, 0.00000187, -0.00000013, -0.00000148, 0.00000068, -0.00000293, 0.00000533, -0.00000889, 0.00000278, -0.00000109, 0.00000789, -0.00000980, 0.00000436, -0.00000163, 0.00000222, -0.00000959, -0.00000367, -0.00000036, -0.00000168, -0.00001972, 0.00001144, },
		{ -0.00000189, 0.00000650, 0.00001788, 0.00000365, 0.00001149, 0.00000038, 0.00001274, -0.00000813, 0.00000590, 0.00000212, 0.00000522, 0.00002859, -0.00001148, 0.00001354, -0.00001918, -0.00000953, 0.00001459, -0.00002073, 0.00001010, -0.00001174, 0.00000389, -0.00001058, 0.00000034, -0.00000418, 0.00000359, -0.00000408, 0.00000611, -0.00001633, 0.00000658, -0.00001251, 0.00000539, -0.00000425, 0.00000894, -0.00001821, 0.00001577, -0.00002024, 0.00000738, 0.00000498, -0.00000810, 0.00001327, -0.00000598, -0.00000215, -0.00000047, 0.00000715, -0.00000201, -0.00000746, 0.00000257, 0.00000674, -0.00000495, -0.00000655, -0.00000069, 0.00000483, -0.00001002, 0.00000693, -0.00000337, 0.00000568, -0.00000887, -0.00000118, -0.00000411, 0.00000075, -0.00000597, 0.00000710, 0.00000130, -0.00000742, -0.00000697, -0.00000179, -0.00000169, -0.00000168, -0.00001351, -0.00000029, 0.00000058, -0.00001651, -0.00002002, 0.00001029, -0.00001558, -0.00002422, -0.00004373, -0.00007472, -0.00024649, 0.00039202, 0.00010053, 0.00036330, -0.00002024, -0.00021811, 0.00001259, -0.00005743, -0.00000240, -0.00002771, 0.00000742, -0.00002154, 0.00000069, 0.00000299, 0.00000402, -0.00001429, 0.00000069, -0.00000513, 0.00000323, -0.00000345, 0.00000142, -0.00000834, 0.00000332, -0.00000959, 0.00000371, 0.00000412, 0.00000243, -0.00000364, -0.00000049, -0.00000235, 0.00000093, 0.00000414, -0.00000503, 0.00000471, -0.00000702, 0.00000659, -0.00000499, -0.00000053, -0.00000364, 0.00000798, -0.00000313, -0.00000186, -0.00000061, 0.00000408, -0.00000659, -0.00000344, -0.00000920, 0.00000958, -0.00001937, -0.00000207, },
		{ 0.00000855, -0.00000930, 0.00002168, 0.00000650, 0.00001217, 0.00000270, 0.00000916, 0.00000389, -0.00000610, 0.00001095, -0.00000388, 0.00002203, 0.00001315, -0.00000824, 0.00001140, -0.00003497, 0.00002094, -0.00001524, 0.00000063, -0.00000392, -0.00000229, -0.00000390, -0.00000521, -0.00000328, 0.00000245, -0.00000402, 0.00000513, -0.00000658, -0.00000685, 0.00000067, -0.00000780, 0.00000315, 0.00000078, -0.00000042, -0.00001056, 0.00001239, -0.00002504, 0.00002284, -0.00001263, 0.00000738, 0.00000317, -0.00000279, -0.00000472, 0.00000573, 0.00000215, -0.00000196, -0.00000888, 0.00001035, -0.00000197, -0.00000145, -0.00001106, 0.00000924, -0.00000739, 0.00000082, 0.00000027, 0.00000264, -0.00000264, -0.00000360, -0.00000529, 0.00000166, -0.00000642, 0.00000278, 0.00000294, 0.00000176, -0.00001523, 0.00000425, -0.00001090, 0.00000896, -0.00001821, 0.00000222, -0.00001000, 0.00000957, -0.00004226, 0.00001713, -0.00002186, 0.00000987, -0.00007542, 0.00001661, -0.00025355, 0.00007274, 0.00032139, 0.00009836, 0.00035494, -0.00022768, -0.00007478, -0.00003317, -0.00002112, -0.00002387, -0.00000161, -0.00001408, -0.00000601, -0.00000591, 0.00001521, -0.00001387, -0.00000150, -0.00000740, 0.00000497, -0.00000655, 0.00000561, -0.00000877, 0.00000136, -0.00000544, -0.00000309, 0.00000322, 0.00000646, -0.00000313, -0.00000066, -0.00000238, -0.00000062, 0.00000208, 0.00000154, -0.00000318, 0.00000313, -0.00000637, 0.00000809, -0.00000864, 0.00000243, -0.00000414, 0.00001017, -0.00000877, 0.00000290, -0.00000201, 0.00000402, -0.00000935, -0.00000321, -0.00000607, 0.00000745, -0.00002591, },
		{ -0.00001738, 0.00000625, -0.00000330, 0.00002503, 0.00000246, 0.00001462, -0.00000082, 0.00001464, -0.00000689, 0.00000407, 0.00000689, -0.00000009, 0.00003265, -0.00000918, 0.00001122, -0.00001860, -0.00001017, 0.00001314, -0.00001998, 0.00000942, -0.00001271, 0.00000511, -0.00001099, -0.00000007, -0.00000442, 0.00000429, -0.00000645, 0.00000906, -0.00001682, 0.00000479, -0.00000930, 0.00000160, -0.00000300, 0.00000788, -0.00001380, 0.00000649, -0.00000657, -0.00000873, 0.00001741, -0.00001502, 0.00001404, -0.00000334, -0.00000177, -0.00000292, 0.00000670, 0.00000060, -0.00000585, -0.00000244, 0.00000840, -0.00000321, -0.00000470, -0.00000592, 0.00000787, -0.00000991, 0.00000546, -0.00000195, 0.00000253, -0.00000458, -0.00000352, -0.00000314, -0.00000075, -0.00000437, 0.00000401, 0.00000462, -0.00000860, -0.00000634, -0.00000105, -0.00000372, -0.00000040, -0.00001310, -0.00000087, 0.00000024, -0.00001382, -0.00002310, 0.00000983, -0.00001173, -0.00002426, -0.00004065, -0.00006485, -0.00023366, 0.00033761, 0.00011231, 0.00025202, 0.00015352, -0.00024871, -0.00000692, -0.00005152, -0.00000737, -0.00002319, 0.00000059, -0.00001692, -0.00000378, 0.00000223, 0.00000802, -0.00001503, -0.00000029, -0.00000379, 0.00000185, -0.00000399, 0.00000319, -0.00000833, 0.00000201, -0.00000744, -0.00000009, 0.00000644, 0.00000199, -0.00000247, -0.00000191, -0.00000033, -0.00000212, 0.00000581, -0.00000464, 0.00000374, -0.00000502, 0.00000252, 0.00000034, -0.00000349, -0.00000222, 0.00000323, 0.00000293, -0.00000541, 0.00000123, 0.00000158, -0.00000216, -0.00000584, -0.00000828, 0.00000430, -0.00000715, },
		{ -0.00002135, -0.00000393, -0.00000102, 0.00001008, 0.00001927, 0.00000487, 0.00001145, 0.00000196, 0.00001266, -0.00001252, 0.00001686, -0.00000616, 0.00002133, 0.00001678, -0.00000717, 0.00000818, -0.00003246, 0.00001753, -0.00001361, -0.00000101, -0.00000298, -0.00000462, -0.00000063, -0.00000815, -0.00000148, 0.00000055, -0.00000201, 0.00000124, -0.00000024, -0.00001289, 0.00000529, -0.00001074, 0.00000428, -0.00000264, 0.00000552, -0.00001602, 0.00001429, -0.00002135, 0.00001208, 0.00000067, -0.00000512, 0.00001115, -0.00000519, -0.00000183, 0.00000011, 0.00000591, -0.00000146, -0.00000754, 0.00000412, 0.00000346, -0.00000244, -0.00000852, 0.00000187, 0.00000086, -0.00000574, 0.00000544, -0.00000273, 0.00000117, -0.00000502, -0.00000331, -0.00000178, -0.00000296, -0.00000113, 0.00000543, 0.00000082, -0.00001313, 0.00000176, -0.00000703, 0.00000276, -0.00001096, -0.00000435, -0.00000373, 0.00000282, -0.00003221, 0.00000328, -0.00000801, -0.00000093, -0.00005934, 0.00000231, -0.00021247, 0.00002385, 0.00032583, 0.00004002, 0.00033892, -0.00007514, -0.00016106, -0.00000499, -0.00004398, -0.00000620, -0.00001613, -0.00000544, -0.00001303, -0.00000263, 0.00000968, -0.00000390, -0.00001002, -0.00000089, -0.00000017, -0.00000286, 0.00000101, -0.00000221, -0.00000454, -0.00000061, -0.00000698, 0.00000420, 0.00000589, -0.00000105, -0.00000229, -0.00000116, -0.00000099, 0.00000019, 0.00000405, -0.00000472, 0.00000468, -0.00000776, 0.00000767, -0.00000553, 0.00000039, -0.00000477, 0.00000857, -0.00000443, -0.00000054, -0.00000055, 0.00000316, -0.00000643, -0.00000449, -0.00000881, 0.00000957, },
		{ 0.00000337, -0.00002548, 0.00000493, -0.00000464, 0.00002281, 0.00000957, 0.00001102, 0.00000459, 0.00001009, 0.00000060, -0.00000303, 0.00001541, -0.00000681, 0.00003683, -0.00000737, 0.00001048, -0.00002001, -0.00000929, 0.00001085, -0.00001875, 0.00000824, -0.00001283, 0.00000451, -0.00000903, -0.00000247, -0.00000310, 0.00000342, -0.00000638, 0.00000809, -0.00001279, -0.00000093, -0.00000314, -0.00000393, 0.00000006, 0.00000357, -0.00000552, -0.00000501, 0.00000565, -0.00001881, 0.00002062, -0.00001224, 0.00000694, 0.00000372, -0.00000400, -0.00000193, 0.00000261, 0.00000494, -0.00000528, -0.00000427, 0.00000572, 0.00000112, -0.00000457, -0.00000734, 0.00000516, -0.00000428, -0.00000048, 0.00000345, -0.00000390, 0.00000151, -0.00000706, -0.00000031, -0.00000458, -0.00000036, -0.00000112, 0.00000912, -0.00001083, -0.00000428, -0.00000242, -0.00000222, -0.00000369, -0.00000808, -0.00000599, 0.00000484, -0.00001730, -0.00001803, 0.00000035, 0.00000019, -0.00003235, -0.00002895, -0.00006689, -0.00020700, 0.00027208, 0.00014635, 0.00014129, 0.00025650, -0.00019655, -0.00006858, -0.00002438, -0.00002843, -0.00000610, -0.00001499, -0.00000640, -0.00001231, 0.00000509, 0.00000632, -0.00000902, -0.00000698, 0.00000204, -0.00000293, -0.00000134, 0.00000084, -0.00000383, -0.00000313, -0.00000225, -0.00000562, 0.00000937, 0.00000071, -0.00000015, -0.00000421, 0.00000198, -0.00000431, 0.00000576, -0.00000267, 0.00000155, -0.00000197, -0.00000181, 0.00000440, -0.00000462, -0.00000149, -0.00000078, 0.00000655, -0.00000609, 0.00000122, 0.00000023, 0.00000066, -0.00000653, -0.00000729, -0.00000196, },
		{ 0.00000561, -0.00001002, -0.00001891, 0.00000604, -0.00000002, 0.00002825, 0.00000221, 0.00001610, -0.00000078, 0.00001642, -0.00001303, 0.00001648, -0.00000234, 0.00001632, 0.00002270, -0.00000751, 0.00000677, -0.00003234, 0.00001641, -0.00001366, -0.00000160, -0.00000257, -0.00000674, 0.00000184, -0.00000950, -0.00000128, -0.00000024, -0.00000062, -0.00000182, 0.00000379, -0.00001507, 0.00000536, -0.00000908, 0.00000175, -0.00000273, 0.00000641, -0.00001390, 0.00000731, -0.00000972, -0.00000305, 0.00001374, -0.00001382, 0.00001376, -0.00000371, -0.00000086, -0.00000307, 0.00000623, 0.00000101, -0.00000633, -0.00000078, 0.00000565, -0.00000117, -0.00000646, -0.00000403, 0.00000521, -0.00000717, 0.00000544, -0.00000265, -0.00000061, -0.00000195, -0.00000475, -0.00000166, -0.00000351, -0.00000064, 0.00000345, 0.00000352, -0.00001401, 0.00000192, -0.00000567, 0.00000002, -0.00000846, -0.00000560, -0.00000289, 0.00000166, -0.00002861, -0.00000170, -0.00000552, 0.00000027, -0.00005608, 0.00000168, -0.00018991, 0.00000114, 0.00030164, 0.00003625, 0.00025739, 0.00008132, -0.00020379, -0.00001159, -0.00004498, -0.00000610, -0.00001545, -0.00000899, -0.00001090, -0.00000552, 0.00000794, 0.00000153, -0.00001254, -0.00000037, 0.00000001, -0.00000327, -0.00000039, 0.00000062, -0.00000586, -0.00000051, -0.00000617, 0.00000134, 0.00000777, -0.00000103, -0.00000165, -0.00000221, 0.00000066, -0.00000269, 0.00000606, -0.00000496, 0.00000445, -0.00000664, 0.00000442, -0.00000067, -0.00000239, -0.00000371, 0.00000447, 0.00000118, -0.00000420, 0.00000152, 0.00000080, -0.00000215, -0.00000670, -0.00000866, },
		{ -0.00000573, 0.00000747, -0.00002302, -0.00000595, 0.00000065, 0.00001382, 0.00002144, 0.00000440, 0.00001291, 0.00000316, 0.00000975, -0.00001037, 0.00002269, -0.00001073, 0.00003828, -0.00000412, 0.00000952, -0.00002139, -0.00000893, 0.00000942, -0.00001821, 0.00000732, -0.00001258, 0.00000267, -0.00000577, -0.00000553, -0.00000143, 0.00000188, -0.00000473, 0.00000476, -0.00000692, -0.00000714, 0.00000225, -0.00000776, 0.00000188, -0.00000043, 0.00000160, -0.00001239, 0.00001017, -0.00001846, 0.00001330, -0.00000138, -0.00000459, 0.00001165, -0.00000636, 0.00000054, -0.00000272, 0.00000876, -0.00000447, -0.00000404, 0.00000087, 0.00000560, -0.00000460, -0.00000625, -0.00000080, 0.00000306, -0.00000655, 0.00000833, -0.00000901, 0.00000518, -0.00000836, 0.00000127, -0.00000743, 0.00000247, -0.00000454, 0.00001133, -0.00001115, -0.00000343, -0.00000335, -0.00000008, -0.00000781, -0.00000286, -0.00001091, 0.00000940, -0.00002137, -0.00001177, -0.00000933, 0.00001014, -0.00003773, -0.00002026, -0.00006738, -0.00018311, 0.00021658, 0.00017008, 0.00007111, 0.00027170, -0.00008039, -0.00014254, 0.00000123, -0.00004884, 0.00000965, -0.00002771, 0.00000083, -0.00001806, 0.00000678, 0.00000305, -0.00000104, -0.00001422, 0.00000748, -0.00000691, 0.00000107, -0.00000235, 0.00000145, -0.00000813, 0.00000204, -0.00000936, 0.00001025, 0.00000063, 0.00000140, -0.00000549, 0.00000285, -0.00000450, 0.00000378, 0.00000022, -0.00000070, 0.00000051, -0.00000447, 0.00000550, -0.00000290, -0.00000249, -0.00000249, 0.00000642, -0.00000337, -0.00000106, 0.00000095, 0.00000069, -0.00000442, -0.00000795, },
		{ -0.00000781, -0.00000041, 0.00000116, -0.00002819, 0.00000526, -0.00000359, 0.00002922, 0.00000718, 0.00001435, 0.00000295, 0.00001369, -0.00000702, 0.00001029, 0.00000600, 0.00000905, 0.00002944, -0.00000860, 0.00000686, -0.00003411, 0.00001685, -0.00001468, -0.00000154, -0.00000270, -0.00000811, 0.00000263, -0.00000856, -0.00000309, 0.00000051, -0.00000069, -0.00000261, 0.00000409, -0.00001274, 0.00000118, -0.00000394, -0.00000310, -0.00000044, 0.00000329, -0.00000668, -0.00000362, 0.00000265, -0.00001449, 0.00001956, -0.00001395, 0.00000878, 0.00000229, -0.00000252, -0.00000284, 0.00000278, 0.00000536, -0.00000592, -0.00000304, 0.00000397, 0.00000236, -0.00000561, -0.00000675, 0.00000421, -0.00000313, 0.00000081, 0.00000170, -0.00000613, 0.00000345, -0.00000783, 0.00000063, -0.00000666, 0.00000259, -0.00000097, 0.00000779, -0.00001620, 0.00000357, -0.00000628, 0.00000057, -0.00001033, -0.00000228, -0.00000661, 0.00000541, -0.00003102, 0.00000129, -0.00001212, 0.00000988, -0.00006126, 0.00000988, -0.00018017, -0.00000334, 0.00026016, 0.00006132, 0.00016102, 0.00018920, -0.00017769, -0.00005678, -0.00002532, -0.00002189, -0.00000205, -0.00002162, -0.00000261, -0.00001250, 0.00000947, 0.00000184, -0.00000875, -0.00000540, 0.00000463, -0.00000696, 0.00000112, -0.00000035, -0.00000287, -0.00000427, -0.00000212, -0.00000357, 0.00001052, -0.00000197, 0.00000022, -0.00000426, 0.00000278, -0.00000492, 0.00000643, -0.00000347, 0.00000263, -0.00000390, 0.00000009, 0.00000383, -0.00000410, -0.00000269, 0.00000030, 0.00000533, -0.00000588, 0.00000241, -0.00000101, 0.00000117, -0.00000775, },
		{ -0.00000995, -0.00000639, 0.00000504, -0.00001279, -0.00002045, 0.00000609, 0.00000415, 0.00003185, 0.00000046, 0.00001884, -0.00000084, 0.00001530, -0.00001303, 0.00002460, -0.00000909, 0.00003556, 0.00000158, 0.00000708, -0.00002120, -0.00001057, 0.00000992, -0.00001907, 0.00000724, -0.00001264, 0.00000068, -0.00000283, -0.00000748, -0.00000095, 0.00000105, -0.00000321, 0.00000163, -0.00000252, -0.00001040, 0.00000384, -0.00000753, 0.00000051, -0.00000167, 0.00000427, -0.00001252, 0.00000561, -0.00000925, -0.00000010, 0.00001145, -0.00001438, 0.00001565, -0.00000574, 0.00000184, -0.00000638, 0.00000966, -0.00000192, -0.00000356, -0.00000342, 0.00000770, -0.00000316, -0.00000465, -0.00000664, 0.00000795, -0.00000879, 0.00000919, -0.00000970, 0.00000411, -0.00000583, 0.00000006, -0.00000745, 0.00000209, -0.00000435, 0.00000978, -0.00000845, -0.00000494, -0.00000243, 0.00000058, -0.00000979, -0.00000089, -0.00001207, 0.00001015, -0.00002163, -0.00000987, -0.00001268, 0.00001149, -0.00003403, -0.00002105, -0.00005886, -0.00017117, 0.00018188, 0.00016917, 0.00005001, 0.00021975, 0.00005436, -0.00018913, 0.00000188, -0.00005349, 0.00001243, -0.00002870, -0.00000153, -0.00001669, 0.00000399, 0.00000183, 0.00000435, -0.00001736, 0.00000855, -0.00000690, 0.00000059, -0.00000361, 0.00000437, -0.00000987, 0.00000275, -0.00000932, 0.00000793, 0.00000250, 0.00000138, -0.00000497, 0.00000184, -0.00000289, 0.00000084, 0.00000273, -0.00000173, 0.00000122, -0.00000451, 0.00000341, 0.00000119, -0.00000480, -0.00000208, 0.00000329, 0.00000138, -0.00000442, 0.00000299, -0.00000139, -0.00000034, },
		{ -0.00000607, -0.00000836, -0.00000464, 0.00000677, -0.00002742, -0.00000457, -0.00000107, 0.00002311, 0.00001794, 0.00000844, 0.00001079, 0.00000736, 0.00000200, 0.00000244, 0.00001450, 0.00000334, 0.00003436, -0.00000870, 0.00000687, -0.00003606, 0.00001706, -0.00001516, -0.00000188, -0.00000274, -0.00000902, 0.00000205, -0.00000604, -0.00000589, 0.00000192, -0.00000169, -0.00000159, 0.00000168, -0.00000794, -0.00000438, 0.00000144, -0.00000720, 0.00000152, -0.00000065, 0.00000087, -0.00001215, 0.00000897, -0.00001677, 0.00001531, -0.00000552, -0.00000171, 0.00001020, -0.00000501, -0.00000056, -0.00000258, 0.00000969, -0.00000537, -0.00000332, -0.00000003, 0.00000632, -0.00000512, -0.00000672, -0.00000067, 0.00000359, -0.00000502, 0.00000664, -0.00001146, 0.00000751, -0.00000958, 0.00000246, -0.00000964, 0.00000549, -0.00000460, 0.00001064, -0.00001719, 0.00000456, -0.00000698, 0.00000234, -0.00001382, 0.00000223, -0.00001128, 0.00001030, -0.00003524, 0.00000686, -0.00002076, 0.00001968, -0.00006601, 0.00001742, -0.00017219, -0.00000383, 0.00022003, 0.00008545, 0.00009041, 0.00022361, -0.00008971, -0.00012032, -0.00000164, -0.00004106, 0.00001335, -0.00003410, 0.00000449, -0.00001821, 0.00001083, -0.00000025, -0.00000197, -0.00001216, 0.00000996, -0.00001073, 0.00000307, -0.00000282, 0.00000173, -0.00000884, 0.00000201, -0.00000760, 0.00001191, -0.00000220, 0.00000178, -0.00000571, 0.00000384, -0.00000544, 0.00000494, -0.00000083, 0.00000036, -0.00000120, -0.00000319, 0.00000588, -0.00000324, -0.00000318, -0.00000198, 0.00000620, -0.00000443, 0.00000107, -0.00000083, 0.00000182, },
		{ 0.00000345, -0.00001135, -0.00000583, -0.00000042, -0.00000129, -0.00003059, 0.00000524, -0.00000111, 0.00003605, 0.00000305, 0.00001915, 0.00000110, 0.00001474, -0.00000929, 0.00002053, -0.00000246, 0.00002990, 0.00000885, 0.00000334, -0.00001934, -0.00001418, 0.00001203, -0.00002094, 0.00000787, -0.00001319, -0.00000077, -0.00000129, -0.00000735, -0.00000234, 0.00000162, -0.00000294, 0.00000043, -0.00000145, -0.00000942, 0.00000119, -0.00000358, -0.00000349, -0.00000028, 0.00000254, -0.00000669, -0.00000434, 0.00000265, -0.00001209, 0.00001915, -0.00001703, 0.00001266, -0.00000076, 0.00000069, -0.00000685, 0.00000676, 0.00000262, -0.00000366, -0.00000562, 0.00000651, 0.00000005, -0.00000360, -0.00001014, 0.00000819, -0.00000595, 0.00000560, -0.00000616, -0.00000081, -0.00000074, -0.00000296, -0.00000522, -0.00000095, -0.00000135, 0.00000566, -0.00000398, -0.00000784, -0.00000014, -0.00000041, -0.00000896, -0.00000275, -0.00000913, 0.00000672, -0.00001739, -0.00001339, -0.00000880, 0.00000437, -0.00002239, -0.00002967, -0.00004292, -0.00016955, 0.00016491, 0.00014886, 0.00006109, 0.00014187, 0.00015921, -0.00018045, -0.00003113, -0.00003942, 0.00000053, -0.00001762, -0.00001264, -0.00000938, -0.00000254, 0.00000321, 0.00000553, -0.00001499, 0.00000457, -0.00000279, -0.00000281, -0.00000247, 0.00000410, -0.00000778, -0.00000012, -0.00000611, 0.00000346, 0.00000534, 0.00000043, -0.00000324, -0.00000020, -0.00000073, -0.00000163, 0.00000372, -0.00000095, 0.00000007, -0.00000246, -0.00000051, 0.00000575, -0.00000675, -0.00000112, -0.00000066, 0.00000565, -0.00000676, 0.00000460, -0.00000359, },
		{ -0.00000010, -0.00000197, -0.00001052, -0.00000576, 0.00000570, -0.00001823, -0.00001867, 0.00000238, 0.00001461, 0.00002934, 0.00000335, 0.00001793, 0.00000239, 0.00000894, -0.00000216, 0.00001883, 0.00000228, 0.00003569, -0.00000652, 0.00000535, -0.00003651, 0.00001542, -0.00001387, -0.00000348, -0.00000201, -0.00001014, 0.00000107, -0.00000343, -0.00000803, 0.00000252, -0.00000233, -0.00000027, -0.00000110, -0.00000381, -0.00000793, 0.00000390, -0.00000795, 0.00000093, -0.00000263, 0.00000491, -0.00001418, 0.00000653, -0.00000983, 0.00000382, 0.00000667, -0.00001229, 0.00001554, -0.00000537, 0.00000126, -0.00000700, 0.00001156, -0.00000320, -0.00000319, -0.00000393, 0.00000853, -0.00000369, -0.00000545, -0.00000655, 0.00000912, -0.00000820, 0.00000863, -0.00001339, 0.00000768, -0.00000815, 0.00000210, -0.00001043, 0.00000586, -0.00000529, 0.00001035, -0.00001562, 0.00000362, -0.00000633, 0.00000333, -0.00001611, 0.00000466, -0.00001344, 0.00001265, -0.00003696, 0.00000960, -0.00002494, 0.00002289, -0.00006365, 0.00001743, -0.00015801, -0.00001061, 0.00019388, 0.00009133, 0.00006098, 0.00019347, 0.00002605, -0.00016886, 0.00000508, -0.00004959, 0.00001966, -0.00003776, 0.00000414, -0.00001837, 0.00000893, -0.00000154, 0.00000370, -0.00001634, 0.00001222, -0.00001155, 0.00000301, -0.00000433, 0.00000506, -0.00001127, 0.00000358, -0.00000868, 0.00001065, -0.00000088, 0.00000216, -0.00000572, 0.00000330, -0.00000434, 0.00000241, 0.00000167, -0.00000095, -0.00000001, -0.00000404, 0.00000468, 0.00000030, -0.00000529, -0.00000197, 0.00000387, -0.00000049, -0.00000196, 0.00000123, },
		{ -0.00000405, 0.00000485, -0.00001022, -0.00000635, -0.00000483, 0.00000562, -0.00003135, -0.00000388, -0.00000138, 0.00003351, 0.00001185, 0.00001463, 0.00000791, 0.00000972, -0.00000153, 0.00001360, 0.00000558, 0.00002456, 0.00001547, -0.00000026, -0.00001705, -0.00001834, 0.00001425, -0.00002256, 0.00000832, -0.00001374, -0.00000179, -0.00000109, -0.00000553, -0.00000495, 0.00000303, -0.00000372, 0.00000115, -0.00000324, -0.00000559, -0.00000352, 0.00000145, -0.00000755, 0.00000157, -0.00000104, 0.00000091, -0.00001360, 0.00001031, -0.00001651, 0.00001773, -0.00001114, 0.00000350, 0.00000691, -0.00000181, -0.00000483, 0.00000139, 0.00000758, -0.00000366, -0.00000605, 0.00000309, 0.00000373, -0.00000280, -0.00001100, 0.00000428, 0.00000018, 0.00000007, -0.00000123, -0.00000645, 0.00000388, -0.00000538, -0.00000283, -0.00000443, 0.00000197, 0.00000161, -0.00000021, -0.00001010, 0.00000188, -0.00000177, -0.00000662, -0.00000665, -0.00000439, 0.00000162, -0.00001123, -0.00001950, -0.00000147, -0.00000597, -0.00000915, -0.00003926, -0.00002680, -0.00017006, 0.00015463, 0.00012428, 0.00007866, 0.00007685, 0.00020487, -0.00011465, -0.00008542, -0.00001791, -0.00001735, -0.00000227, -0.00002551, -0.00000182, -0.00000879, 0.00000494, 0.00000400, -0.00000915, -0.00000183, 0.00000273, -0.00000681, -0.00000059, 0.00000202, -0.00000363, -0.00000441, -0.00000212, -0.00000083, 0.00000735, -0.00000018, -0.00000150, -0.00000196, 0.00000067, -0.00000266, 0.00000291, 0.00000119, -0.00000190, 0.00000010, -0.00000409, 0.00000853, -0.00000658, -0.00000130, -0.00000324, 0.00000728, -0.00000647, 0.00000427, },
		{ 0.00000358, -0.00000282, 0.00000251, -0.00001257, -0.00000524, 0.00000083, -0.00000654, -0.00002949, 0.00000081, 0.00000840, 0.00003659, 0.00000330, 0.00002059, 0.00000243, 0.00001058, -0.00000094, 0.00001750, 0.00000616, 0.00003390, -0.00000235, 0.00000205, -0.00003496, 0.00001157, -0.00001076, -0.00000628, -0.00000042, -0.00001181, 0.00000049, -0.00000192, -0.00000838, 0.00000143, -0.00000176, 0.00000006, -0.00000234, -0.00000252, -0.00000771, 0.00000237, -0.00000501, -0.00000231, -0.00000203, 0.00000448, -0.00000977, -0.00000233, 0.00000140, -0.00000833, 0.00001586, -0.00001734, 0.00001471, -0.00000169, 0.00000094, -0.00000853, 0.00000965, 0.00000111, -0.00000355, -0.00000621, 0.00000796, -0.00000096, -0.00000404, -0.00001103, 0.00001085, -0.00000688, 0.00000657, -0.00001132, 0.00000404, -0.00000408, -0.00000015, -0.00000898, 0.00000362, -0.00000322, 0.00000744, -0.00001215, 0.00000119, -0.00000439, 0.00000294, -0.00001602, 0.00000376, -0.00001197, 0.00001125, -0.00003459, 0.00000740, -0.00002249, 0.00001809, -0.00005357, 0.00000963, -0.00013701, -0.00002505, 0.00018262, 0.00007892, 0.00006382, 0.00013054, 0.00012677, -0.00017429, -0.00001657, -0.00004190, 0.00001296, -0.00003034, -0.00000441, -0.00001296, 0.00000370, -0.00000087, 0.00000610, -0.00001591, 0.00001007, -0.00000867, 0.00000043, -0.00000399, 0.00000586, -0.00001045, 0.00000198, -0.00000680, 0.00000727, 0.00000145, 0.00000164, -0.00000456, 0.00000173, -0.00000259, 0.00000014, 0.00000285, -0.00000050, -0.00000077, -0.00000250, 0.00000110, 0.00000504, -0.00000766, -0.00000092, 0.00000010, 0.00000381, -0.00000487, },
		{ 0.00000186, -0.00000230, 0.00000337, -0.00000632, -0.00000863, -0.00000630, 0.00000594, -0.00002335, -0.00001744, -0.00000062, 0.00002750, 0.00002300, 0.00000925, 0.00001562, 0.00000458, 0.00000567, 0.00000846, 0.00001081, 0.00002269, 0.00001950, -0.00000232, -0.00001585, -0.00002135, 0.00001492, -0.00002263, 0.00000767, -0.00001357, -0.00000302, -0.00000140, -0.00000333, -0.00000730, 0.00000399, -0.00000447, 0.00000251, -0.00000589, -0.00000173, -0.00000710, 0.00000443, -0.00000904, 0.00000160, -0.00000359, 0.00000612, -0.00001742, 0.00000992, -0.00001186, 0.00000841, -0.00000010, -0.00000738, 0.00001328, -0.00000299, -0.00000259, -0.00000377, 0.00001055, -0.00000211, -0.00000608, -0.00000050, 0.00000605, -0.00000139, -0.00001012, -0.00000152, 0.00000624, -0.00000391, 0.00000181, -0.00000966, 0.00000548, -0.00000531, -0.00000204, -0.00000634, 0.00000342, -0.00000019, 0.00000105, -0.00001030, 0.00000236, -0.00000221, -0.00000449, -0.00001013, -0.00000076, -0.00000206, -0.00000644, -0.00002432, 0.00000436, -0.00001348, -0.00000071, -0.00004340, -0.00001703, -0.00016520, 0.00014103, 0.00010832, 0.00008423, 0.00004456, 0.00019098, -0.00001579, -0.00013404, -0.00000686, -0.00002890, 0.00000738, -0.00003211, 0.00000016, -0.00001080, 0.00000437, 0.00000237, -0.00000327, -0.00000694, 0.00000631, -0.00000872, -0.00000002, 0.00000018, 0.00000015, -0.00000750, 0.00000021, -0.00000295, 0.00000728, 0.00000038, -0.00000060, -0.00000260, 0.00000074, -0.00000221, 0.00000099, 0.00000345, -0.00000325, 0.00000153, -0.00000558, 0.00000821, -0.00000363, -0.00000322, -0.00000358, 0.00000576, -0.00000351, },
		{ -0.00000581, 0.00000446, -0.00000421, 0.00000499, -0.00001239, -0.00000544, -0.00000374, 0.00000154, -0.00003162, -0.00000809, 0.00000561, 0.00003815, 0.00000921, 0.00001834, 0.00000762, 0.00000740, 0.00000482, 0.00001260, 0.00001228, 0.00003159, 0.00000209, -0.00000203, -0.00003224, 0.00000654, -0.00000697, -0.00000929, 0.00000136, -0.00001368, 0.00000035, -0.00000170, -0.00000703, -0.00000101, -0.00000022, -0.00000072, -0.00000165, -0.00000411, -0.00000446, -0.00000166, -0.00000038, -0.00000624, -0.00000037, 0.00000138, -0.00000233, -0.00001215, 0.00001023, -0.00001468, 0.00001722, -0.00001429, 0.00000741, 0.00000520, -0.00000109, -0.00000727, 0.00000478, 0.00000630, -0.00000387, -0.00000699, 0.00000539, 0.00000211, -0.00000259, -0.00001320, 0.00000846, -0.00000195, 0.00000208, -0.00000726, -0.00000098, 0.00000026, -0.00000243, -0.00000684, 0.00000042, -0.00000028, 0.00000400, -0.00000873, -0.00000116, -0.00000233, 0.00000192, -0.00001421, 0.00000052, -0.00000825, 0.00000759, -0.00002960, 0.00000186, -0.00001577, 0.00000893, -0.00004033, -0.00000094, -0.00011437, -0.00004164, 0.00017856, 0.00005938, 0.00007809, 0.00007008, 0.00018145, -0.00012869, -0.00006038, -0.00002485, -0.00000130, -0.00001709, -0.00001610, -0.00000620, -0.00000196, 0.00000041, 0.00000578, -0.00001187, 0.00000509, -0.00000391, -0.00000308, -0.00000278, 0.00000481, -0.00000739, -0.00000139, -0.00000368, 0.00000354, 0.00000338, 0.00000117, -0.00000316, 0.00000019, -0.00000133, -0.00000094, 0.00000233, 0.00000143, -0.00000268, 0.00000005, -0.00000285, 0.00000871, -0.00000843, -0.00000059, -0.00000287, 0.00000622, },
		{ 0.00000051, -0.00000123, 0.00000047, 0.00000043, -0.00000140, -0.00001127, -0.00000581, 0.00000188, -0.00001188, -0.00002845, -0.00000404, 0.00002189, 0.00003216, 0.00000719, 0.00002026, 0.00000325, 0.00000866, 0.00000818, 0.00001112, 0.00002524, 0.00002086, -0.00000269, -0.00001637, -0.00002237, 0.00001333, -0.00002075, 0.00000566, -0.00001236, -0.00000490, -0.00000139, -0.00000196, -0.00000817, 0.00000350, -0.00000421, 0.00000317, -0.00000741, -0.00000017, -0.00000747, 0.00000385, -0.00000708, -0.00000082, -0.00000386, 0.00000712, -0.00001470, 0.00000249, -0.00000169, -0.00000332, 0.00000995, -0.00001440, 0.00001449, -0.00000063, -0.00000207, -0.00000644, 0.00000985, 0.00000173, -0.00000662, -0.00000281, 0.00000605, 0.00000088, -0.00000843, -0.00000691, 0.00000947, -0.00000413, 0.00000138, -0.00000935, 0.00000358, -0.00000275, -0.00000300, -0.00000620, 0.00000247, 0.00000056, -0.00000016, -0.00000864, 0.00000139, -0.00000150, -0.00000341, -0.00001175, 0.00000014, -0.00000277, -0.00000476, -0.00002570, 0.00000594, -0.00001519, 0.00000046, -0.00004023, -0.00001523, -0.00015297, 0.00012084, 0.00010471, 0.00007366, 0.00004282, 0.00014050, 0.00007969, -0.00015038, -0.00001878, -0.00002716, 0.00000611, -0.00002898, -0.00000524, -0.00000779, 0.00000093, 0.00000203, 0.00000031, -0.00000842, 0.00000622, -0.00000743, -0.00000152, -0.00000041, 0.00000213, -0.00000806, -0.00000008, -0.00000249, 0.00000524, 0.00000187, -0.00000047, -0.00000219, -0.00000016, -0.00000108, -0.00000086, 0.00000464, -0.00000300, 0.00000107, -0.00000459, 0.00000513, 0.00000109, -0.00000592, -0.00000249, 0.00000234, },
		{ 0.00000573, -0.00000418, 0.00000314, -0.00000324, 0.00000446, -0.00000924, -0.00000747, -0.00000566, 0.00000333, -0.00002490, -0.00002118, 0.00000323, 0.00003593, 0.00001861, 0.00001397, 0.00001497, 0.00000288, 0.00001126, 0.00000824, 0.00001677, 0.00003174, 0.00000494, -0.00000558, -0.00002977, 0.00000190, -0.00000408, -0.00001125, 0.00000240, -0.00001503, 0.00000008, -0.00000208, -0.00000503, -0.00000356, 0.00000122, -0.00000176, -0.00000001, -0.00000696, -0.00000058, -0.00000534, 0.00000304, -0.00000837, 0.00000023, -0.00000166, 0.00000398, -0.00001780, 0.00001205, -0.00001253, 0.00001054, -0.00000514, -0.00000294, 0.00001196, -0.00000257, -0.00000514, -0.00000059, 0.00001000, -0.00000270, -0.00000747, 0.00000257, 0.00000400, -0.00000066, -0.00001336, 0.00000349, 0.00000387, -0.00000183, -0.00000415, -0.00000456, 0.00000246, -0.00000296, -0.00000564, -0.00000189, 0.00000149, 0.00000206, -0.00000709, -0.00000207, -0.00000130, 0.00000133, -0.00001202, -0.00000307, -0.00000471, 0.00000426, -0.00002477, -0.00000373, -0.00000905, 0.00000073, -0.00002972, -0.00000842, -0.00009603, -0.00005376, 0.00017274, 0.00004474, 0.00008561, 0.00003428, 0.00018287, -0.00004760, -0.00010541, -0.00001308, -0.00001285, -0.00000659, -0.00002384, -0.00000336, -0.00000456, 0.00000018, 0.00000469, -0.00000691, 0.00000036, -0.00000013, -0.00000521, -0.00000236, 0.00000343, -0.00000397, -0.00000439, -0.00000141, 0.00000123, 0.00000378, 0.00000151, -0.00000224, -0.00000054, -0.00000115, -0.00000073, 0.00000069, 0.00000373, -0.00000431, 0.00000191, -0.00000520, 0.00000962, -0.00000650, -0.00000196, -0.00000378, },
		{ -0.00000086, 0.00000342, -0.00000314, 0.00000236, -0.00000171, 0.00000204, -0.00001204, -0.00000570, -0.00000259, -0.00000307, -0.00003133, -0.00001389, 0.00001783, 0.00003726, 0.00001022, 0.00002048, 0.00000684, 0.00000693, 0.00001248, 0.00000766, 0.00003032, 0.00002148, -0.00000257, -0.00001805, -0.00002178, 0.00001007, -0.00001768, 0.00000295, -0.00001055, -0.00000727, -0.00000086, -0.00000181, -0.00000738, 0.00000161, -0.00000291, 0.00000273, -0.00000706, -0.00000141, -0.00000479, 0.00000055, -0.00000305, -0.00000434, -0.00000268, 0.00000488, -0.00000791, -0.00000737, 0.00000783, -0.00001112, 0.00001373, -0.00001414, 0.00000927, 0.00000524, -0.00000344, -0.00000619, 0.00000581, 0.00000690, -0.00000720, -0.00000389, 0.00000438, 0.00000320, -0.00000622, -0.00001056, 0.00000891, -0.00000075, -0.00000164, -0.00000681, -0.00000010, 0.00000055, -0.00000440, -0.00000504, 0.00000023, 0.00000252, -0.00000229, -0.00000660, 0.00000011, -0.00000044, -0.00000305, -0.00001161, -0.00000135, -0.00000110, -0.00000552, -0.00002416, 0.00000383, -0.00001223, -0.00000344, -0.00003274, -0.00001799, -0.00013678, 0.00009736, 0.00010876, 0.00005427, 0.00005585, 0.00008409, 0.00014117, -0.00012260, -0.00005208, -0.00001553, -0.00000328, -0.00001926, -0.00001457, -0.00000280, -0.00000330, 0.00000231, 0.00000144, -0.00000650, 0.00000326, -0.00000409, -0.00000403, -0.00000026, 0.00000245, -0.00000645, -0.00000220, -0.00000060, 0.00000255, 0.00000325, -0.00000045, -0.00000139, -0.00000118, -0.00000028, -0.00000173, 0.00000425, -0.00000119, -0.00000079, -0.00000214, 0.00000101, 0.00000549, -0.00000760, -0.00000166, },
		{ -0.00000303, 0.00000309, -0.00000099, 0.00000050, -0.00000067, 0.00000193, -0.00000451, -0.00001025, -0.00000545, 0.00000037, -0.00001403, -0.00003195, -0.00000314, 0.00003271, 0.00002796, 0.00001121, 0.00002076, 0.00000091, 0.00001464, 0.00000772, 0.00001713, 0.00003570, 0.00000577, -0.00000819, -0.00002836, -0.00000127, -0.00000304, -0.00001150, 0.00000220, -0.00001538, -0.00000083, -0.00000224, -0.00000360, -0.00000500, 0.00000153, -0.00000208, 0.00000127, -0.00000913, 0.00000158, -0.00000646, 0.00000346, -0.00000748, -0.00000122, -0.00000296, 0.00000666, -0.00001720, 0.00000662, -0.00000410, 0.00000003, 0.00000483, -0.00001111, 0.00001464, -0.00000109, -0.00000423, -0.00000399, 0.00001023, 0.00000090, -0.00000857, 0.00000082, 0.00000396, 0.00000177, -0.00001210, -0.00000203, 0.00000787, -0.00000285, -0.00000376, -0.00000531, 0.00000172, -0.00000138, -0.00000584, -0.00000253, 0.00000126, 0.00000227, -0.00000754, -0.00000141, -0.00000147, 0.00000155, -0.00001036, -0.00000550, -0.00000307, 0.00000300, -0.00002201, -0.00000697, -0.00000537, -0.00000314, -0.00002477, -0.00001017, -0.00008448, -0.00005856, 0.00016083, 0.00004051, 0.00007938, 0.00002632, 0.00014614, 0.00003910, -0.00012759, -0.00001904, -0.00001425, -0.00000463, -0.00002350, -0.00000680, -0.00000287, -0.00000235, 0.00000421, -0.00000329, -0.00000179, 0.00000083, -0.00000474, -0.00000351, 0.00000269, -0.00000160, -0.00000560, -0.00000108, 0.00000098, 0.00000254, 0.00000258, -0.00000190, -0.00000041, -0.00000180, 0.00000012, -0.00000101, 0.00000520, -0.00000456, 0.00000205, -0.00000512, 0.00000756, -0.00000238, -0.00000451, },
		{ -0.00000236, -0.00000234, 0.00000388, -0.00000280, 0.00000220, -0.00000158, 0.00000254, -0.00000982, -0.00000736, -0.00000478, -0.00000016, -0.00002541, -0.00002738, 0.00001290, 0.00003907, 0.00001716, 0.00001789, 0.00001343, 0.00000305, 0.00001831, 0.00000386, 0.00003453, 0.00002406, -0.00000359, -0.00001984, -0.00002073, 0.00000646, -0.00001476, 0.00000067, -0.00000899, -0.00000946, -0.00000025, -0.00000239, -0.00000575, -0.00000070, -0.00000142, 0.00000177, -0.00000551, -0.00000423, -0.00000101, -0.00000307, 0.00000053, -0.00000678, -0.00000187, 0.00000176, -0.00000097, -0.00001449, 0.00001165, -0.00001140, 0.00000979, -0.00000729, -0.00000001, 0.00001192, -0.00000488, -0.00000451, 0.00000059, 0.00001108, -0.00000635, -0.00000491, 0.00000259, 0.00000432, -0.00000340, -0.00001224, 0.00000539, 0.00000419, -0.00000480, -0.00000445, -0.00000314, 0.00000249, -0.00000472, -0.00000423, -0.00000170, 0.00000390, -0.00000353, -0.00000571, -0.00000027, 0.00000001, -0.00000263, -0.00001068, -0.00000375, 0.00000110, -0.00000677, -0.00002182, 0.00000057, -0.00000803, -0.00000799, -0.00002580, -0.00002033, -0.00012176, 0.00007609, 0.00011297, 0.00003657, 0.00006686, 0.00004456, 0.00015669, -0.00005936, -0.00009114, -0.00000546, -0.00001261, -0.00001001, -0.00002180, -0.00000039, -0.00000536, 0.00000158, 0.00000148, -0.00000316, -0.00000015, -0.00000103, -0.00000565, -0.00000060, 0.00000217, -0.00000405, -0.00000450, 0.00000105, 0.00000069, 0.00000355, 0.00000008, -0.00000083, -0.00000163, -0.00000033, -0.00000147, 0.00000272, 0.00000124, -0.00000276, 0.00000013, -0.00000211, 0.00000763, -0.00000686, },
		{ -0.00000661, -0.00000212, 0.00000014, 0.00000181, -0.00000175, 0.00000169, -0.00000040, -0.00000070, -0.00001154, -0.00000533, -0.00000357, -0.00000507, -0.00003485, -0.00001541, 0.00002923, 0.00003517, 0.00001225, 0.00002301, 0.00000329, 0.00001370, 0.00001156, 0.00001377, 0.00004226, 0.00000597, -0.00001066, -0.00002785, -0.00000300, -0.00000367, -0.00001041, 0.00000110, -0.00001491, -0.00000239, -0.00000184, -0.00000326, -0.00000493, 0.00000053, -0.00000145, 0.00000158, -0.00000963, 0.00000118, -0.00000470, 0.00000116, -0.00000437, -0.00000395, -0.00000271, 0.00000582, -0.00001180, -0.00000244, 0.00000526, -0.00000834, 0.00001023, -0.00001310, 0.00001124, 0.00000394, -0.00000521, -0.00000448, 0.00000685, 0.00000626, -0.00000985, 0.00000004, 0.00000261, 0.00000387, -0.00000970, -0.00000662, 0.00000870, -0.00000059, -0.00000577, -0.00000389, -0.00000086, 0.00000109, -0.00000662, -0.00000208, -0.00000032, 0.00000375, -0.00000899, -0.00000023, -0.00000206, 0.00000210, -0.00000922, -0.00000649, -0.00000350, 0.00000380, -0.00002137, -0.00000763, -0.00000505, -0.00000277, -0.00002453, -0.00000777, -0.00007788, -0.00005788, 0.00014424, 0.00004457, 0.00006325, 0.00003531, 0.00009682, 0.00010263, -0.00011301, -0.00004432, -0.00000634, -0.00001042, -0.00001655, -0.00001441, 0.00000098, -0.00000571, 0.00000411, -0.00000160, -0.00000105, -0.00000080, -0.00000233, -0.00000552, 0.00000236, -0.00000043, -0.00000500, -0.00000239, 0.00000210, 0.00000056, 0.00000366, -0.00000170, 0.00000013, -0.00000261, 0.00000073, -0.00000191, 0.00000523, -0.00000323, 0.00000064, -0.00000325, 0.00000389, 0.00000211, },
		{ -0.00000371, -0.00000445, -0.00000209, 0.00000215, -0.00000069, 0.00000029, 0.00000057, 0.00000062, -0.00000558, -0.00001005, -0.00000482, -0.00000215, -0.00001491, -0.00003830, 0.00000338, 0.00003926, 0.00002552, 0.00001545, 0.00001982, 0.00000055, 0.00002212, 0.00000301, 0.00003529, 0.00003012, -0.00000635, -0.00002125, -0.00002006, 0.00000362, -0.00001305, -0.00000040, -0.00000830, -0.00001091, -0.00000013, -0.00000291, -0.00000441, -0.00000228, -0.00000070, 0.00000119, -0.00000393, -0.00000685, 0.00000167, -0.00000483, 0.00000176, -0.00000674, -0.00000257, -0.00000035, 0.00000320, -0.00001600, 0.00000844, -0.00000508, 0.00000108, 0.00000187, -0.00000863, 0.00001562, -0.00000398, -0.00000354, -0.00000319, 0.00001204, -0.00000288, -0.00000679, 0.00000187, 0.00000366, -0.00000025, -0.00001208, 0.00000052, 0.00000826, -0.00000591, -0.00000406, -0.00000406, 0.00000212, -0.00000341, -0.00000441, -0.00000242, 0.00000375, -0.00000307, -0.00000645, 0.00000054, -0.00000044, -0.00000175, -0.00000976, -0.00000569, 0.00000220, -0.00000684, -0.00002055, -0.00000149, -0.00000559, -0.00000982, -0.00002259, -0.00001930, -0.00011085, 0.00006034, 0.00011251, 0.00002693, 0.00006710, 0.00002909, 0.00013465, 0.00001577, -0.00011505, -0.00000854, -0.00001464, -0.00000705, -0.00002256, -0.00000322, -0.00000383, -0.00000097, 0.00000162, -0.00000042, -0.00000187, -0.00000004, -0.00000519, -0.00000213, 0.00000204, -0.00000204, -0.00000572, 0.00000138, 0.00000036, 0.00000254, 0.00000114, -0.00000065, -0.00000140, -0.00000109, -0.00000063, 0.00000100, 0.00000311, -0.00000363, 0.00000098, -0.00000308, 0.00000685, },
		{ 0.00000497, -0.00000668, -0.00000244, -0.00000150, 0.00000271, -0.00000225, 0.00000230, -0.00000069, 0.00000036, -0.00000989, -0.00000689, -0.00000559, -0.00000165, -0.00002888, -0.00003072, 0.00002318, 0.00004028, 0.00001696, 0.00002229, 0.00000909, 0.00001010, 0.00001752, 0.00000941, 0.00004858, 0.00000788, -0.00001435, -0.00002746, -0.00000409, -0.00000496, -0.00000906, 0.00000001, -0.00001432, -0.00000407, -0.00000121, -0.00000371, -0.00000396, -0.00000106, -0.00000045, 0.00000125, -0.00000880, -0.00000098, -0.00000151, -0.00000196, -0.00000112, -0.00000615, -0.00000226, 0.00000330, -0.00000503, -0.00001020, 0.00001036, -0.00001045, 0.00000864, -0.00000844, 0.00000313, 0.00001054, -0.00000675, -0.00000311, 0.00000168, 0.00001114, -0.00000976, -0.00000098, 0.00000145, 0.00000450, -0.00000631, -0.00000957, 0.00000656, 0.00000357, -0.00000835, -0.00000222, -0.00000331, 0.00000274, -0.00000675, -0.00000162, -0.00000197, 0.00000503, -0.00000993, 0.00000019, -0.00000206, 0.00000221, -0.00000809, -0.00000660, -0.00000502, 0.00000539, -0.00002150, -0.00000715, -0.00000629, -0.00000071, -0.00002558, -0.00000500, -0.00007233, -0.00005573, 0.00012718, 0.00005106, 0.00004571, 0.00004665, 0.00005694, 0.00012817, -0.00006497, -0.00007786, 0.00000230, -0.00001776, -0.00000859, -0.00002115, 0.00000313, -0.00000741, 0.00000317, -0.00000098, 0.00000119, -0.00000319, 0.00000017, -0.00000686, 0.00000154, 0.00000014, -0.00000348, -0.00000416, 0.00000332, -0.00000098, 0.00000395, -0.00000117, 0.00000057, -0.00000301, 0.00000063, -0.00000181, 0.00000412, -0.00000108, -0.00000123, -0.00000106, 0.00000055, },
		{ 0.00000405, 0.00000037, -0.00000655, -0.00000136, -0.00000025, 0.00000161, -0.00000190, 0.00000311, -0.00000162, -0.00000176, -0.00001157, -0.00000477, -0.00000550, -0.00000587, -0.00004111, -0.00001209, 0.00003772, 0.00003363, 0.00001537, 0.00002379, 0.00000162, 0.00002219, 0.00000617, 0.00003239, 0.00003901, -0.00000987, -0.00002280, -0.00001983, 0.00000175, -0.00001264, -0.00000034, -0.00000839, -0.00001158, -0.00000063, -0.00000296, -0.00000396, -0.00000263, -0.00000106, 0.00000135, -0.00000305, -0.00000814, 0.00000218, -0.00000406, 0.00000048, -0.00000456, -0.00000445, -0.00000110, 0.00000393, -0.00001236, 0.00000076, 0.00000343, -0.00000711, 0.00000813, -0.00001231, 0.00001378, 0.00000037, -0.00000436, -0.00000416, 0.00000916, 0.00000282, -0.00000915, 0.00000200, 0.00000198, 0.00000223, -0.00001018, -0.00000429, 0.00000990, -0.00000421, -0.00000573, -0.00000311, 0.00000008, -0.00000126, -0.00000510, -0.00000215, 0.00000233, -0.00000144, -0.00000813, 0.00000184, -0.00000126, -0.00000069, -0.00000905, -0.00000665, 0.00000172, -0.00000534, -0.00002083, -0.00000161, -0.00000579, -0.00000831, -0.00002301, -0.00001543, -0.00010341, 0.00004959, 0.00010726, 0.00002535, 0.00005742, 0.00003128, 0.00009538, 0.00007689, -0.00010926, -0.00002834, -0.00000851, -0.00001100, -0.00001710, -0.00001012, -0.00000015, -0.00000428, 0.00000186, 0.00000095, -0.00000129, -0.00000124, -0.00000298, -0.00000433, 0.00000194, -0.00000072, -0.00000559, 0.00000041, 0.00000121, 0.00000085, 0.00000223, -0.00000058, -0.00000079, -0.00000203, 0.00000004, -0.00000009, 0.00000375, -0.00000304, 0.00000029, -0.00000209, },
		{ -0.00000105, 0.00000514, -0.00000436, -0.00000420, -0.00000150, 0.00000159, -0.00000099, 0.00000098, 0.00000129, -0.00000126, -0.00000591, -0.00000968, -0.00000550, -0.00000320, -0.00001805, -0.00004286, 0.00001122, 0.00004389, 0.00002399, 0.00002065, 0.00001585, 0.00000682, 0.00002244, 0.00000711, 0.00005213, 0.00001309, -0.00001994, -0.00002669, -0.00000531, -0.00000591, -0.00000841, -0.00000032, -0.00001424, -0.00000526, -0.00000089, -0.00000423, -0.00000307, -0.00000221, 0.00000008, 0.00000102, -0.00000763, -0.00000334, 0.00000112, -0.00000387, 0.00000043, -0.00000645, -0.00000273, 0.00000096, -0.00000002, -0.00001327, 0.00000900, -0.00000607, 0.00000178, -0.00000035, -0.00000566, 0.00001516, -0.00000654, -0.00000196, -0.00000259, 0.00001307, -0.00000685, -0.00000327, 0.00000160, 0.00000325, -0.00000248, -0.00001057, 0.00000251, 0.00000760, -0.00000953, -0.00000192, -0.00000425, 0.00000262, -0.00000568, -0.00000177, -0.00000280, 0.00000514, -0.00000954, -0.00000070, -0.00000113, 0.00000160, -0.00000667, -0.00000647, -0.00000652, 0.00000641, -0.00002098, -0.00000712, -0.00000702, 0.00000034, -0.00002485, -0.00000477, -0.00006511, -0.00005497, 0.00011288, 0.00005528, 0.00003329, 0.00005083, 0.00003634, 0.00011869, -0.00000133, -0.00010207, 0.00000140, -0.00002008, -0.00000508, -0.00002275, 0.00000086, -0.00000601, 0.00000067, -0.00000047, 0.00000327, -0.00000448, 0.00000113, -0.00000643, -0.00000031, 0.00000063, -0.00000192, -0.00000535, 0.00000371, -0.00000142, 0.00000324, -0.00000025, 0.00000076, -0.00000283, -0.00000011, -0.00000115, 0.00000269, 0.00000082, -0.00000239, 0.00000014, },
		{ -0.00000036, 0.00000098, 0.00000342, -0.00000706, -0.00000157, -0.00000174, 0.00000257, -0.00000292, 0.00000431, -0.00000211, -0.00000048, -0.00001020, -0.00000639, -0.00000717, -0.00000219, -0.00003470, -0.00003027, 0.00003174, 0.00004130, 0.00001820, 0.00002510, 0.00000618, 0.00001932, 0.00001188, 0.00002789, 0.00004844, -0.00001219, -0.00002561, -0.00001949, 0.00000035, -0.00001283, 0.00000007, -0.00000863, -0.00001199, -0.00000135, -0.00000274, -0.00000423, -0.00000215, -0.00000199, 0.00000189, -0.00000278, -0.00000815, 0.00000096, -0.00000179, -0.00000187, -0.00000189, -0.00000614, -0.00000132, 0.00000240, -0.00000636, -0.00000693, 0.00000908, -0.00001039, 0.00000841, -0.00000963, 0.00000686, 0.00000685, -0.00000613, -0.00000293, 0.00000392, 0.00000861, -0.00001030, 0.00000165, 0.00000086, 0.00000286, -0.00000672, -0.00000805, 0.00000883, -0.00000056, -0.00000811, -0.00000173, -0.00000209, 0.00000039, -0.00000540, -0.00000167, 0.00000063, 0.00000022, -0.00000952, 0.00000261, -0.00000166, -0.00000002, -0.00000825, -0.00000688, 0.00000022, -0.00000304, -0.00002186, -0.00000061, -0.00000752, -0.00000524, -0.00002456, -0.00001146, -0.00009675, 0.00004107, 0.00010000, 0.00002780, 0.00004407, 0.00003921, 0.00005906, 0.00010789, -0.00007322, -0.00005744, -0.00000040, -0.00001738, -0.00000971, -0.00001699, 0.00000240, -0.00000623, 0.00000126, 0.00000146, 0.00000065, -0.00000326, -0.00000053, -0.00000589, 0.00000118, 0.00000023, -0.00000462, -0.00000109, 0.00000231, -0.00000064, 0.00000274, -0.00000029, -0.00000018, -0.00000264, 0.00000010, -0.00000032, 0.00000330, -0.00000151, -0.00000113, },
		{ -0.00000005, -0.00000073, 0.00000309, -0.00000070, -0.00000631, -0.00000084, -0.00000039, 0.00000083, -0.00000118, 0.00000418, -0.00000369, -0.00000189, -0.00001156, -0.00000524, -0.00000632, -0.00000845, -0.00004618, -0.00000746, 0.00004485, 0.00003250, 0.00001989, 0.00002147, 0.00000612, 0.00002440, 0.00000826, 0.00005230, 0.00002136, -0.00002677, -0.00002591, -0.00000682, -0.00000615, -0.00000876, 0.00000025, -0.00001480, -0.00000580, -0.00000110, -0.00000440, -0.00000287, -0.00000240, -0.00000022, 0.00000129, -0.00000687, -0.00000475, 0.00000203, -0.00000373, -0.00000016, -0.00000488, -0.00000411, -0.00000043, 0.00000193, -0.00001120, 0.00000272, 0.00000136, -0.00000589, 0.00000640, -0.00001079, 0.00001491, -0.00000308, -0.00000243, -0.00000421, 0.00001097, -0.00000117, -0.00000652, 0.00000273, 0.00000107, 0.00000051, -0.00000936, -0.00000226, 0.00000995, -0.00000843, -0.00000328, -0.00000368, 0.00000111, -0.00000392, -0.00000232, -0.00000278, 0.00000413, -0.00000806, -0.00000244, 0.00000027, 0.00000062, -0.00000513, -0.00000631, -0.00000741, 0.00000625, -0.00001927, -0.00000825, -0.00000626, -0.00000078, -0.00002141, -0.00000741, -0.00005619, -0.00005569, 0.00010172, 0.00005601, 0.00002752, 0.00004636, 0.00003248, 0.00008952, 0.00005543, -0.00010278, -0.00001365, -0.00001548, -0.00000754, -0.00001863, -0.00000535, -0.00000246, -0.00000259, -0.00000002, 0.00000439, -0.00000398, 0.00000026, -0.00000440, -0.00000270, 0.00000087, -0.00000060, -0.00000565, 0.00000316, -0.00000094, 0.00000193, 0.00000068, 0.00000087, -0.00000229, -0.00000103, -0.00000060, 0.00000169, 0.00000176, -0.00000230, },
		{ -0.00000218, 0.00000082, -0.00000082, 0.00000415, -0.00000528, -0.00000318, -0.00000175, 0.00000179, -0.00000234, 0.00000352, -0.00000019, -0.00000208, -0.00000620, -0.00000946, -0.00000673, -0.00000359, -0.00002305, -0.00004455, 0.00001800, 0.00004832, 0.00002366, 0.00002482, 0.00001260, 0.00001582, 0.00001754, 0.00002443, 0.00005611, -0.00001180, -0.00003037, -0.00001861, -0.00000122, -0.00001278, -0.00000002, -0.00000833, -0.00001273, -0.00000173, -0.00000274, -0.00000462, -0.00000167, -0.00000264, 0.00000212, -0.00000256, -0.00000765, -0.00000070, 0.00000031, -0.00000353, -0.00000038, -0.00000643, -0.00000195, 0.00000028, -0.00000106, -0.00001096, 0.00000909, -0.00000766, 0.00000331, -0.00000274, -0.00000188, 0.00001230, -0.00000671, -0.00000141, -0.00000102, 0.00001180, -0.00000847, -0.00000046, 0.00000147, 0.00000128, -0.00000248, -0.00001001, 0.00000558, 0.00000348, -0.00000958, -0.00000128, -0.00000323, 0.00000074, -0.00000487, -0.00000149, -0.00000058, 0.00000102, -0.00000986, 0.00000233, -0.00000132, -0.00000000, -0.00000716, -0.00000680, -0.00000148, -0.00000100, -0.00002254, 0.00000022, -0.00000906, -0.00000291, -0.00002462, -0.00000982, -0.00008867, 0.00003249, 0.00009342, 0.00003019, 0.00003304, 0.00004358, 0.00003692, 0.00010768, -0.00001989, -0.00008142, 0.00000094, -0.00002046, -0.00000531, -0.00001981, 0.00000118, -0.00000538, -0.00000082, 0.00000186, 0.00000258, -0.00000455, 0.00000073, -0.00000584, -0.00000061, 0.00000107, -0.00000340, -0.00000235, 0.00000295, -0.00000140, 0.00000251, 0.00000024, 0.00000030, -0.00000279, -0.00000039, -0.00000005, 0.00000245, 0.00000003, },
		{ -0.00000128, -0.00000106, 0.00000059, 0.00000030, 0.00000246, -0.00000722, -0.00000087, -0.00000177, 0.00000183, -0.00000273, 0.00000617, -0.00000469, -0.00000021, -0.00001059, -0.00000673, -0.00000787, -0.00000424, -0.00003951, -0.00002895, 0.00003957, 0.00004253, 0.00002083, 0.00002521, 0.00000853, 0.00002355, 0.00001202, 0.00005054, 0.00003093, -0.00003328, -0.00002599, -0.00000826, -0.00000601, -0.00000963, 0.00000118, -0.00001558, -0.00000601, -0.00000156, -0.00000431, -0.00000330, -0.00000189, -0.00000100, 0.00000185, -0.00000658, -0.00000506, 0.00000139, -0.00000218, -0.00000187, -0.00000266, -0.00000543, -0.00000120, 0.00000133, -0.00000603, -0.00000465, 0.00000722, -0.00000999, 0.00000830, -0.00001011, 0.00000945, 0.00000292, -0.00000415, -0.00000332, 0.00000593, 0.00000530, -0.00000891, 0.00000327, -0.00000025, 0.00000126, -0.00000598, -0.00000669, 0.00001003, -0.00000548, -0.00000533, -0.00000267, -0.00000063, -0.00000248, -0.00000264, -0.00000243, 0.00000268, -0.00000639, -0.00000408, 0.00000134, -0.00000013, -0.00000391, -0.00000599, -0.00000773, 0.00000517, -0.00001676, -0.00001013, -0.00000439, -0.00000345, -0.00001646, -0.00001109, -0.00004757, -0.00005613, 0.00009206, 0.00005474, 0.00002609, 0.00003729, 0.00003635, 0.00005827, 0.00008919, -0.00007691, -0.00003815, -0.00000826, -0.00001288, -0.00001196, -0.00001217, 0.00000044, -0.00000473, -0.00000042, 0.00000489, -0.00000233, -0.00000142, -0.00000203, -0.00000448, 0.00000025, 0.00000066, -0.00000533, 0.00000215, -0.00000016, 0.00000067, 0.00000122, 0.00000110, -0.00000167, -0.00000173, -0.00000054, 0.00000137, 0.00000177, },
		{ 0.00000088, -0.00000226, 0.00000087, -0.00000091, 0.00000261, -0.00000194, -0.00000538, -0.00000102, 0.00000009, -0.00000097, 0.00000147, 0.00000311, -0.00000494, -0.00000175, -0.00001187, -0.00000607, -0.00000674, -0.00001252, -0.00004890, -0.00000396, 0.00005239, 0.00003201, 0.00002402, 0.00001924, 0.00001373, 0.00002127, 0.00002346, 0.00006126, -0.00000867, -0.00003664, -0.00001742, -0.00000319, -0.00001209, -0.00000094, -0.00000727, -0.00001405, -0.00000149, -0.00000323, -0.00000471, -0.00000177, -0.00000252, 0.00000170, -0.00000201, -0.00000734, -0.00000173, 0.00000112, -0.00000361, -0.00000058, -0.00000518, -0.00000311, -0.00000143, 0.00000177, -0.00001016, 0.00000405, -0.00000133, -0.00000360, 0.00000416, -0.00000821, 0.00001365, -0.00000435, -0.00000128, -0.00000353, 0.00001086, -0.00000332, -0.00000416, 0.00000343, -0.00000136, 0.00000094, -0.00000948, 0.00000095, 0.00000655, -0.00000929, -0.00000206, -0.00000325, 0.00000002, -0.00000387, -0.00000154, -0.00000114, 0.00000089, -0.00000927, 0.00000132, -0.00000055, -0.00000038, -0.00000595, -0.00000657, -0.00000286, 0.00000021, -0.00002232, 0.00000019, -0.00000941, -0.00000253, -0.00002218, -0.00001095, -0.00007905, 0.00002353, 0.00008824, 0.00003080, 0.00002671, 0.00004134, 0.00002943, 0.00008714, 0.00003182, -0.00008717, -0.00000948, -0.00001785, -0.00000593, -0.00001743, -0.00000384, -0.00000237, -0.00000371, 0.00000216, 0.00000385, -0.00000444, 0.00000045, -0.00000428, -0.00000293, 0.00000148, -0.00000204, -0.00000314, 0.00000301, -0.00000153, 0.00000183, 0.00000075, 0.00000074, -0.00000260, -0.00000106, 0.00000015, 0.00000188, },
		{ 0.00000139, -0.00000010, -0.00000199, 0.00000176, -0.00000172, 0.00000370, -0.00000603, -0.00000222, -0.00000188, 0.00000141, -0.00000298, 0.00000627, -0.00000321, -0.00000157, -0.00000669, -0.00000991, -0.00000724, -0.00000535, -0.00002703, -0.00004581, 0.00002418, 0.00005349, 0.00002407, 0.00002725, 0.00001325, 0.00002148, 0.00001644, 0.00004899, 0.00003990, -0.00003819, -0.00002750, -0.00000932, -0.00000596, -0.00001038, 0.00000181, -0.00001601, -0.00000641, -0.00000174, -0.00000440, -0.00000386, -0.00000136, -0.00000157, 0.00000218, -0.00000639, -0.00000481, 0.00000023, -0.00000059, -0.00000324, -0.00000119, -0.00000572, -0.00000198, -0.00000052, -0.00000057, -0.00000944, 0.00000837, -0.00000879, 0.00000507, -0.00000477, 0.00000128, 0.00000877, -0.00000525, -0.00000165, 0.00000054, 0.00000965, -0.00000829, 0.00000161, 0.00000067, -0.00000065, -0.00000130, -0.00000970, 0.00000787, -0.00000183, -0.00000681, -0.00000227, -0.00000170, -0.00000194, -0.00000247, -0.00000212, 0.00000144, -0.00000525, -0.00000498, 0.00000166, -0.00000037, -0.00000320, -0.00000541, -0.00000771, 0.00000372, -0.00001420, -0.00001201, -0.00000233, -0.00000635, -0.00001187, -0.00001361, -0.00004116, -0.00005471, 0.00008230, 0.00005331, 0.00002589, 0.00002849, 0.00003959, 0.00003644, 0.00009613, -0.00003349, -0.00006051, -0.00000572, -0.00001628, -0.00000708, -0.00001589, 0.00000012, -0.00000435, -0.00000226, 0.00000535, -0.00000065, -0.00000258, -0.00000060, -0.00000472, -0.00000152, 0.00000193, -0.00000468, 0.00000117, 0.00000043, -0.00000018, 0.00000125, 0.00000144, -0.00000105, -0.00000208, -0.00000093, 0.00000143, },
		{ 0.00000180, 0.00000101, -0.00000153, 0.00000011, 0.00000016, -0.00000002, 0.00000137, -0.00000672, -0.00000079, -0.00000123, -0.00000002, -0.00000053, 0.00000600, -0.00000674, 0.00000063, -0.00001144, -0.00000728, -0.00000829, -0.00000763, -0.00004201, -0.00002942, 0.00004865, 0.00004406, 0.00002336, 0.00002524, 0.00001387, 0.00002286, 0.00002461, 0.00006487, -0.00000414, -0.00004325, -0.00001654, -0.00000535, -0.00001088, -0.00000246, -0.00000574, -0.00001572, -0.00000079, -0.00000403, -0.00000454, -0.00000244, -0.00000180, 0.00000089, -0.00000126, -0.00000737, -0.00000185, 0.00000064, -0.00000252, -0.00000179, -0.00000330, -0.00000417, -0.00000265, 0.00000201, -0.00000579, -0.00000292, 0.00000450, -0.00000822, 0.00000746, -0.00000948, 0.00000988, 0.00000081, -0.00000261, -0.00000332, 0.00000641, 0.00000339, -0.00000754, 0.00000486, -0.00000295, 0.00000181, -0.00000618, -0.00000412, 0.00000792, -0.00000743, -0.00000339, -0.00000298, -0.00000086, -0.00000314, -0.00000142, -0.00000135, 0.00000029, -0.00000840, 0.00000028, 0.00000005, -0.00000069, -0.00000496, -0.00000613, -0.00000386, 0.00000067, -0.00002136, -0.00000061, -0.00000859, -0.00000384, -0.00001804, -0.00001344, -0.00006941, 0.00001537, 0.00008357, 0.00003011, 0.00002420, 0.00003454, 0.00003040, 0.00006097, 0.00006669, -0.00007015, -0.00002888, -0.00001246, -0.00000955, -0.00001216, -0.00000996, 0.00000039, -0.00000573, 0.00000162, 0.00000471, -0.00000343, -0.00000055, -0.00000232, -0.00000464, 0.00000083, -0.00000037, -0.00000363, 0.00000282, -0.00000144, 0.00000115, 0.00000096, 0.00000125, -0.00000228, -0.00000159, -0.00000006, },
		{ 0.00000141, 0.00000118, 0.00000064, -0.00000225, 0.00000213, -0.00000229, 0.00000276, -0.00000314, -0.00000430, -0.00000124, 0.00000016, -0.00000239, 0.00000482, 0.00000005, -0.00000461, -0.00000190, -0.00001270, -0.00000635, -0.00000837, -0.00001560, -0.00005126, -0.00000175, 0.00006203, 0.00003106, 0.00002772, 0.00001936, 0.00001973, 0.00001997, 0.00004888, 0.00004761, -0.00004147, -0.00003009, -0.00001020, -0.00000617, -0.00001063, 0.00000181, -0.00001585, -0.00000730, -0.00000133, -0.00000496, -0.00000415, -0.00000133, -0.00000151, 0.00000201, -0.00000601, -0.00000459, -0.00000052, 0.00000004, -0.00000347, -0.00000100, -0.00000483, -0.00000289, -0.00000259, 0.00000317, -0.00000992, 0.00000464, -0.00000379, -0.00000064, 0.00000173, -0.00000570, 0.00001132, -0.00000380, -0.00000104, -0.00000278, 0.00000986, -0.00000384, -0.00000236, 0.00000347, -0.00000398, 0.00000278, -0.00001013, 0.00000384, 0.00000149, -0.00000702, -0.00000270, -0.00000209, -0.00000205, -0.00000213, -0.00000177, 0.00000055, -0.00000478, -0.00000521, 0.00000147, -0.00000034, -0.00000282, -0.00000475, -0.00000743, 0.00000227, -0.00001201, -0.00001343, -0.00000070, -0.00000850, -0.00000883, -0.00001402, -0.00003737, -0.00005142, 0.00007229, 0.00005232, 0.00002524, 0.00002264, 0.00003837, 0.00002663, 0.00008337, 0.00001198, -0.00006900, -0.00001273, -0.00001521, -0.00000622, -0.00001502, -0.00000377, -0.00000180, -0.00000501, 0.00000576, 0.00000053, -0.00000266, -0.00000046, -0.00000350, -0.00000389, 0.00000271, -0.00000362, 0.00000029, 0.00000082, -0.00000072, 0.00000104, 0.00000165, -0.00000039, -0.00000216, -0.00000147, },
		{ -0.00000063, 0.00000198, 0.00000081, -0.00000056, -0.00000077, 0.00000141, -0.00000236, 0.00000320, -0.00000618, -0.00000177, -0.00000149, -0.00000010, -0.00000164, 0.00000724, -0.00000613, -0.00000014, -0.00000780, -0.00001050, -0.00000750, -0.00000848, -0.00002842, -0.00004970, 0.00003207, 0.00005935, 0.00002392, 0.00003011, 0.00001614, 0.00002324, 0.00002647, 0.00006857, 0.00000037, -0.00004928, -0.00001633, -0.00000760, -0.00000943, -0.00000412, -0.00000422, -0.00001736, -0.00000005, -0.00000469, -0.00000448, -0.00000325, -0.00000106, 0.00000024, -0.00000069, -0.00000752, -0.00000142, -0.00000037, -0.00000133, -0.00000287, -0.00000190, -0.00000443, -0.00000362, 0.00000048, -0.00000022, -0.00000835, 0.00000660, -0.00000840, 0.00000614, -0.00000599, 0.00000274, 0.00000654, -0.00000382, -0.00000184, 0.00000093, 0.00000860, -0.00000804, 0.00000373, -0.00000177, -0.00000057, -0.00000090, -0.00000835, 0.00000727, -0.00000475, -0.00000428, -0.00000318, -0.00000127, -0.00000304, -0.00000104, -0.00000139, -0.00000034, -0.00000783, -0.00000027, 0.00000018, -0.00000073, -0.00000433, -0.00000547, -0.00000455, 0.00000069, -0.00002011, -0.00000173, -0.00000721, -0.00000586, -0.00001372, -0.00001549, -0.00006127, 0.00000908, 0.00007852, 0.00002909, 0.00002351, 0.00002678, 0.00003290, 0.00004007, 0.00007927, -0.00003636, -0.00004805, -0.00001028, -0.00001219, -0.00000772, -0.00001377, 0.00000043, -0.00000543, -0.00000040, 0.00000572, -0.00000249, -0.00000111, -0.00000115, -0.00000482, -0.00000118, 0.00000154, -0.00000389, 0.00000260, -0.00000139, 0.00000066, 0.00000086, 0.00000176, -0.00000186, -0.00000185, },
		{ -0.00000188, 0.00000103, 0.00000131, 0.00000087, -0.00000186, 0.00000179, -0.00000171, 0.00000051, 0.00000008, -0.00000579, -0.00000100, -0.00000082, -0.00000192, 0.00000295, 0.00000337, -0.00000699, 0.00000106, -0.00001265, -0.00000753, -0.00000962, -0.00001035, -0.00004361, -0.00003271, 0.00006171, 0.00004383, 0.00002665, 0.00002626, 0.00001906, 0.00002232, 0.00005000, 0.00005486, -0.00004422, -0.00003282, -0.00001135, -0.00000659, -0.00001040, 0.00000127, -0.00001525, -0.00000863, -0.00000036, -0.00000589, -0.00000419, -0.00000180, -0.00000098, 0.00000157, -0.00000553, -0.00000458, -0.00000055, -0.00000035, -0.00000285, -0.00000160, -0.00000347, -0.00000348, -0.00000450, 0.00000452, -0.00000662, -0.00000161, 0.00000154, -0.00000517, 0.00000584, -0.00000852, 0.00000904, 0.00000056, -0.00000198, -0.00000323, 0.00000608, 0.00000298, -0.00000672, 0.00000605, -0.00000621, 0.00000409, -0.00000726, -0.00000141, 0.00000378, -0.00000609, -0.00000335, -0.00000249, -0.00000208, -0.00000220, -0.00000115, -0.00000010, -0.00000465, -0.00000526, 0.00000133, -0.00000049, -0.00000239, -0.00000431, -0.00000681, 0.00000091, -0.00001021, -0.00001444, 0.00000044, -0.00000975, -0.00000729, -0.00001288, -0.00003507, -0.00004753, 0.00006292, 0.00005133, 0.00002398, 0.00001981, 0.00003318, 0.00002513, 0.00006270, 0.00004576, -0.00005845, -0.00002782, -0.00001158, -0.00000816, -0.00001116, -0.00000907, 0.00000081, -0.00000709, 0.00000535, 0.00000153, -0.00000214, -0.00000089, -0.00000187, -0.00000564, 0.00000222, -0.00000188, -0.00000068, 0.00000125, -0.00000125, 0.00000090, 0.00000158, 0.00000036, -0.00000209, },
		},
		{ { -0.00000125, -0.00000204, -0.00000870, -0.00000356, -0.00000215, -0.00000051, -0.00000059, -0.00000076, -0.00000926, -0.00000340, -0.00000090, -0.00000884, -0.00001321, -0.00001466, -0.00006784, 0.00000540, 0.00007196, 0.00002660, 0.00003297, 0.00002146, 0.00002297, 0.00002559, 0.00007525, 0.00000676, -0.00006225, -0.00001442, -0.00001503, -0.00000752, -0.00000598, -0.00000659, 0.00000464, -0.00000513, -0.00000330, -0.00000161, -0.00000318, 0.00000056, 0.00000083, 0.00000130, 0.00000100, 0.00000032, -0.00000075, -0.00000528, 0.00001360, -0.00000199, -0.00000137, -0.00000258, 0.00000020, 0.00000118, -0.00000556, 0.00000130, -0.00000059, -0.00000255, 0.00000055, 0.00000200, -0.00000716, -0.00000376, 0.00001182, -0.00000355, -0.00000889, -0.00000365, -0.00000029, 0.00000422, -0.00001207, 0.00000110, -0.00000137, 0.00000110, -0.00001207, 0.00000422, -0.00000029, -0.00000365, -0.00000889, -0.00000355, 0.00001182, -0.00000376, -0.00000716, 0.00000200, 0.00000055, -0.00000255, -0.00000059, 0.00000130, -0.00000556, 0.00000118, 0.00000020, -0.00000258, -0.00000137, -0.00000199, 0.00001360, -0.00000528, -0.00000075, 0.00000032, 0.00000100, 0.00000130, 0.00000083, 0.00000056, -0.00000318, -0.00000161, -0.00000330, -0.00000513, 0.00000464, -0.00000659, -0.00000598, -0.00000752, -0.00001503, -0.00001442, -0.00006225, 0.00000676, 0.00007525, 0.00002559, 0.00002297, 0.00002146, 0.00003297, 0.00002660, 0.00007196, 0.00000540, -0.00006784, -0.00001466, -0.00001321, -0.00000884, -0.00000090, -0.00000340, -0.00000926, -0.00000076, -0.00000059, -0.00000051, -0.00000215, -0.00000356, -0.00000870, -0.00000204, },
		{ -0.00000086, -0.00000134, -0.00000376, -0.00000878, -0.00000230, -0.00000164, -0.00000061, 0.00000009, -0.00000352, -0.00000895, -0.00000105, -0.00000284, -0.00001158, -0.00001098, -0.00002631, -0.00007043, 0.00003799, 0.00006391, 0.00002747, 0.00003162, 0.00002208, 0.00002200, 0.00004317, 0.00007434, -0.00003796, -0.00004888, -0.00000947, -0.00001589, -0.00000315, -0.00000910, -0.00000219, 0.00000278, -0.00000618, -0.00000241, -0.00000223, -0.00000247, 0.00000156, 0.00000021, 0.00000220, -0.00000048, 0.00000197, -0.00000384, -0.00000122, 0.00001334, -0.00000549, 0.00000102, -0.00000452, 0.00000279, -0.00000146, -0.00000429, 0.00000159, -0.00000164, -0.00000200, 0.00000151, 0.00000012, -0.00000929, 0.00000290, 0.00000950, -0.00000808, -0.00000627, -0.00000523, 0.00000543, -0.00000400, -0.00000788, 0.00000158, -0.00000021, -0.00000367, -0.00000980, 0.00000814, -0.00000576, -0.00000067, -0.00001300, 0.00000492, 0.00000804, -0.00000721, -0.00000445, 0.00000313, -0.00000092, -0.00000239, 0.00000050, 0.00000012, -0.00000562, 0.00000279, -0.00000184, -0.00000085, -0.00000385, 0.00000299, 0.00001092, -0.00000686, 0.00000103, -0.00000059, 0.00000213, 0.00000044, 0.00000146, -0.00000095, -0.00000321, -0.00000104, -0.00000452, -0.00000277, 0.00000436, -0.00000984, -0.00000276, -0.00001431, -0.00000938, -0.00003094, -0.00005358, 0.00004635, 0.00005803, 0.00001885, 0.00002139, 0.00002485, 0.00002911, 0.00003255, 0.00007000, -0.00002468, -0.00005608, -0.00000914, -0.00001408, -0.00000597, 0.00000033, -0.00000701, -0.00000678, 0.00000006, -0.00000061, -0.00000101, -0.00000247, -0.00000494, -0.00000770, },
		{ -0.00000584, -0.00000043, -0.00000163, -0.00000574, -0.00000796, -0.00000110, -0.00000151, -0.00000043, 0.00000015, -0.00000678, -0.00000625, -0.00000096, -0.00000456, -0.00001400, -0.00000875, -0.00004347, -0.00005949, 0.00006512, 0.00005170, 0.00003200, 0.00002698, 0.00002566, 0.00002138, 0.00006859, 0.00004577, -0.00006690, -0.00002696, -0.00001390, -0.00001032, -0.00000408, -0.00000891, 0.00000054, 0.00000072, -0.00000686, -0.00000133, -0.00000319, -0.00000077, 0.00000132, 0.00000058, 0.00000198, -0.00000093, 0.00000288, -0.00000682, 0.00000442, 0.00001006, -0.00000661, 0.00000226, -0.00000577, 0.00000520, -0.00000472, -0.00000177, 0.00000042, -0.00000161, -0.00000176, 0.00000256, -0.00000348, -0.00000803, 0.00000872, 0.00000344, -0.00000825, -0.00000691, -0.00000227, 0.00000602, -0.00001025, -0.00000153, -0.00000037, 0.00000148, -0.00001040, -0.00000142, 0.00000455, -0.00000585, -0.00000138, -0.00001358, 0.00001258, 0.00000052, -0.00000664, -0.00000272, 0.00000405, -0.00000304, -0.00000096, 0.00000064, -0.00000102, -0.00000505, 0.00000359, -0.00000315, 0.00000015, -0.00000547, 0.00000813, 0.00000628, -0.00000617, 0.00000122, -0.00000016, 0.00000207, 0.00000051, 0.00000110, -0.00000201, -0.00000309, -0.00000034, -0.00000622, 0.00000127, 0.00000058, -0.00000909, -0.00000399, -0.00001720, -0.00000732, -0.00004948, -0.00002314, 0.00006581, 0.00003736, 0.00001831, 0.00001924, 0.00002933, 0.00002244, 0.00004414, 0.00005583, -0.00004552, -0.00004131, -0.00000785, -0.00001362, -0.00000275, -0.00000057, -0.00000955, -0.00000322, -0.00000066, 0.00000018, -0.00000238, -0.00000225, -0.00000649, },
		{ -0.00000785, -0.00000351, -0.00000074, -0.00000210, -0.00000769, -0.00000619, -0.00000033, -0.00000163, 0.00000026, -0.00000138, -0.00000858, -0.00000302, -0.00000255, -0.00000596, -0.00001553, -0.00000903, -0.00006238, -0.00003471, 0.00008107, 0.00004146, 0.00003527, 0.00002365, 0.00002802, 0.00002870, 0.00008801, -0.00000468, -0.00006752, -0.00001309, -0.00001769, -0.00000405, -0.00000727, -0.00000650, 0.00000166, -0.00000147, -0.00000678, -0.00000068, -0.00000370, 0.00000099, 0.00000048, 0.00000160, 0.00000072, 0.00000008, 0.00000203, -0.00000820, 0.00001010, 0.00000446, -0.00000491, 0.00000136, -0.00000511, 0.00000603, -0.00000722, 0.00000082, -0.00000142, -0.00000074, -0.00000179, 0.00000319, -0.00000802, -0.00000240, 0.00000984, -0.00000179, -0.00000736, -0.00000826, 0.00000353, 0.00000042, -0.00001033, 0.00000216, -0.00000130, 0.00000014, -0.00001370, 0.00000747, -0.00000359, -0.00000048, -0.00000757, -0.00000719, 0.00001401, -0.00000556, -0.00000460, -0.00000121, 0.00000387, -0.00000451, 0.00000080, 0.00000011, -0.00000201, -0.00000396, 0.00000354, -0.00000353, -0.00000006, -0.00000525, 0.00001186, 0.00000132, -0.00000444, 0.00000080, 0.00000080, 0.00000143, 0.00000102, 0.00000003, -0.00000265, -0.00000242, -0.00000049, -0.00000687, 0.00000493, -0.00000483, -0.00000540, -0.00001006, -0.00001348, -0.00001569, -0.00005577, 0.00001570, 0.00006344, 0.00002319, 0.00001842, 0.00001970, 0.00003056, 0.00001812, 0.00005601, 0.00003243, -0.00005453, -0.00002837, -0.00000833, -0.00001223, 0.00000054, -0.00000402, -0.00000902, -0.00000102, -0.00000117, 0.00000059, -0.00000378, -0.00000187, },
		{ -0.00000213, -0.00000840, -0.00000136, -0.00000153, -0.00000283, -0.00000926, -0.00000355, -0.00000059, -0.00000121, 0.00000055, -0.00000397, -0.00000810, -0.00000101, -0.00000460, -0.00000737, -0.00001581, -0.00001426, -0.00007729, -0.00000013, 0.00008446, 0.00003637, 0.00003444, 0.00002480, 0.00002674, 0.00004972, 0.00008291, -0.00005529, -0.00004554, -0.00001367, -0.00001371, -0.00000307, -0.00000808, -0.00000490, 0.00000283, -0.00000471, -0.00000494, -0.00000153, -0.00000254, 0.00000136, 0.00000050, 0.00000188, -0.00000031, 0.00000166, -0.00000042, -0.00000710, 0.00001413, -0.00000193, -0.00000102, -0.00000181, -0.00000190, 0.00000414, -0.00000750, 0.00000188, -0.00000236, -0.00000034, -0.00000109, 0.00000176, -0.00001051, 0.00000497, 0.00000602, -0.00000359, -0.00000848, -0.00000653, 0.00000722, -0.00000735, -0.00000507, 0.00000173, 0.00000005, -0.00000555, -0.00000958, 0.00001003, -0.00000920, 0.00000483, -0.00001522, 0.00000405, 0.00000805, -0.00000677, -0.00000383, 0.00000129, 0.00000162, -0.00000389, 0.00000135, -0.00000005, -0.00000339, -0.00000203, 0.00000243, -0.00000279, -0.00000169, -0.00000266, 0.00001304, -0.00000254, -0.00000290, 0.00000086, 0.00000110, 0.00000137, 0.00000083, -0.00000080, -0.00000341, -0.00000067, -0.00000257, -0.00000458, 0.00000512, -0.00000781, -0.00000347, -0.00001597, -0.00000790, -0.00003221, -0.00004208, 0.00004647, 0.00004796, 0.00001745, 0.00001696, 0.00002358, 0.00002629, 0.00002049, 0.00006147, 0.00000659, -0.00005393, -0.00001891, -0.00000936, -0.00000965, 0.00000236, -0.00000832, -0.00000575, -0.00000115, -0.00000030, -0.00000049, -0.00000408, },
		{ -0.00000304, -0.00000341, -0.00000770, -0.00000001, -0.00000242, -0.00000403, -0.00000980, -0.00000074, -0.00000175, 0.00000017, -0.00000061, -0.00000614, -0.00000599, -0.00000100, -0.00000604, -0.00000926, -0.00001499, -0.00002569, -0.00008234, 0.00003723, 0.00007823, 0.00003542, 0.00003099, 0.00002881, 0.00002599, 0.00007939, 0.00004404, -0.00008120, -0.00002068, -0.00002098, -0.00000462, -0.00000698, -0.00000582, -0.00000473, 0.00000401, -0.00000837, -0.00000185, -0.00000351, 0.00000001, 0.00000020, 0.00000178, 0.00000080, -0.00000008, 0.00000261, -0.00000365, -0.00000329, 0.00001490, -0.00000681, 0.00000309, -0.00000579, 0.00000286, -0.00000007, -0.00000545, 0.00000099, -0.00000172, -0.00000102, 0.00000066, -0.00000244, -0.00000819, 0.00000953, 0.00000071, -0.00000321, -0.00001129, -0.00000069, 0.00000499, -0.00001161, 0.00000100, -0.00000047, 0.00000134, -0.00001241, 0.00000043, 0.00000400, -0.00000708, 0.00000381, -0.00001743, 0.00001335, -0.00000093, -0.00000416, -0.00000441, 0.00000423, -0.00000195, -0.00000136, 0.00000037, 0.00000045, -0.00000514, 0.00000055, 0.00000057, -0.00000145, -0.00000409, 0.00000181, 0.00001159, -0.00000490, -0.00000169, 0.00000136, 0.00000075, 0.00000196, -0.00000026, -0.00000115, -0.00000394, 0.00000129, -0.00000568, 0.00000014, 0.00000128, -0.00000647, -0.00000674, -0.00001629, -0.00000850, -0.00004612, -0.00001266, 0.00005956, 0.00003056, 0.00001675, 0.00001533, 0.00002837, 0.00001833, 0.00002991, 0.00005656, -0.00001489, -0.00004832, -0.00001209, -0.00001088, -0.00000539, 0.00000090, -0.00001045, -0.00000232, -0.00000224, 0.00000126, -0.00000279, },
		{ -0.00000481, -0.00000150, -0.00000533, -0.00000583, 0.00000016, -0.00000304, -0.00000586, -0.00000857, 0.00000091, -0.00000262, 0.00000160, -0.00000308, -0.00000684, -0.00000348, -0.00000285, -0.00000615, -0.00001216, -0.00001367, -0.00004283, -0.00007333, 0.00006988, 0.00006741, 0.00003533, 0.00002912, 0.00003084, 0.00003541, 0.00009927, -0.00001721, -0.00007240, -0.00001077, -0.00002225, 0.00000093, -0.00001000, -0.00000407, -0.00000355, 0.00000293, -0.00000988, 0.00000005, -0.00000445, 0.00000189, -0.00000073, 0.00000291, -0.00000062, 0.00000104, 0.00000250, -0.00000673, 0.00000272, 0.00001154, -0.00000805, 0.00000489, -0.00000810, 0.00000678, -0.00000462, -0.00000258, -0.00000059, -0.00000051, -0.00000185, 0.00000193, -0.00000745, -0.00000142, 0.00000863, -0.00000223, -0.00000387, -0.00001255, 0.00000613, -0.00000251, -0.00000945, 0.00000364, -0.00000113, -0.00000094, -0.00001462, 0.00000946, -0.00000583, 0.00000120, -0.00000459, -0.00001066, 0.00001469, -0.00000670, -0.00000166, -0.00000419, 0.00000551, -0.00000446, 0.00000105, -0.00000079, 0.00000062, -0.00000624, 0.00000272, -0.00000097, -0.00000074, -0.00000576, 0.00000667, 0.00000845, -0.00000626, -0.00000011, 0.00000124, 0.00000089, 0.00000199, -0.00000120, -0.00000187, -0.00000298, 0.00000131, -0.00000689, 0.00000381, -0.00000320, -0.00000376, -0.00001308, -0.00001146, -0.00001846, -0.00004670, 0.00002020, 0.00005534, 0.00001870, 0.00001674, 0.00001611, 0.00002995, 0.00001174, 0.00004193, 0.00004251, -0.00002891, -0.00004116, -0.00000710, -0.00001239, -0.00000027, -0.00000385, -0.00000866, -0.00000129, -0.00000211, 0.00000171, },
		{ 0.00000005, -0.00000522, -0.00000062, -0.00000705, -0.00000339, -0.00000075, -0.00000334, -0.00000804, -0.00000549, 0.00000044, -0.00000181, 0.00000138, -0.00000535, -0.00000626, -0.00000157, -0.00000576, -0.00000501, -0.00001622, -0.00001298, -0.00006311, -0.00004929, 0.00009242, 0.00005604, 0.00003426, 0.00003099, 0.00002913, 0.00006075, 0.00008771, -0.00007211, -0.00004174, -0.00001769, -0.00001274, -0.00000194, -0.00000796, -0.00000563, 0.00000036, -0.00000180, -0.00000752, -0.00000104, -0.00000264, 0.00000130, 0.00000031, 0.00000220, -0.00000073, 0.00000187, 0.00000150, -0.00000864, 0.00000939, 0.00000487, -0.00000516, 0.00000309, -0.00000701, 0.00000780, -0.00000752, -0.00000057, -0.00000137, -0.00000013, -0.00000148, 0.00000061, -0.00000955, 0.00000594, 0.00000362, -0.00000181, -0.00000726, -0.00000932, 0.00000904, -0.00001051, -0.00000279, 0.00000222, 0.00000040, -0.00000745, -0.00000876, 0.00001050, -0.00001112, 0.00000817, -0.00001445, 0.00000154, 0.00000796, -0.00000649, -0.00000186, -0.00000174, 0.00000385, -0.00000419, 0.00000149, -0.00000065, -0.00000067, -0.00000562, 0.00000351, -0.00000133, -0.00000161, -0.00000535, 0.00001042, 0.00000458, -0.00000682, 0.00000195, 0.00000000, 0.00000227, 0.00000054, -0.00000113, -0.00000332, -0.00000020, -0.00000159, -0.00000424, 0.00000325, -0.00000419, -0.00000471, -0.00001656, -0.00000801, -0.00003223, -0.00003121, 0.00004463, 0.00004075, 0.00001433, 0.00001513, 0.00002042, 0.00002555, 0.00001161, 0.00004992, 0.00002478, -0.00003677, -0.00003312, -0.00000448, -0.00001222, 0.00000310, -0.00000913, -0.00000433, -0.00000277, -0.00000008, },
		{ 0.00000227, -0.00000288, -0.00000394, -0.00000098, -0.00000785, -0.00000118, -0.00000221, -0.00000373, -0.00000959, -0.00000174, -0.00000153, 0.00000065, -0.00000122, -0.00000602, -0.00000518, -0.00000098, -0.00000824, -0.00000399, -0.00002045, -0.00001497, -0.00008179, -0.00001304, 0.00010288, 0.00004551, 0.00003389, 0.00003362, 0.00003022, 0.00009328, 0.00003776, -0.00009333, -0.00001470, -0.00002802, 0.00000027, -0.00000920, -0.00000309, -0.00000863, 0.00000484, -0.00000790, -0.00000296, -0.00000388, 0.00000085, -0.00000103, 0.00000287, -0.00000020, 0.00000079, 0.00000183, -0.00000031, -0.00000821, 0.00001425, -0.00000243, -0.00000001, -0.00000129, -0.00000293, 0.00000568, -0.00000827, 0.00000030, -0.00000120, -0.00000061, -0.00000020, -0.00000338, -0.00000648, 0.00000943, -0.00000151, -0.00000030, -0.00001182, -0.00000194, 0.00000528, -0.00001364, 0.00000349, -0.00000033, 0.00000121, -0.00001410, 0.00000225, 0.00000259, -0.00000714, 0.00000697, -0.00001775, 0.00001112, -0.00000110, -0.00000274, -0.00000381, 0.00000173, 0.00000057, -0.00000193, 0.00000022, 0.00000064, -0.00000318, -0.00000343, 0.00000307, -0.00000088, -0.00000375, -0.00000272, 0.00001240, 0.00000030, -0.00000595, 0.00000334, -0.00000120, 0.00000388, -0.00000166, -0.00000068, -0.00000425, 0.00000239, -0.00000525, 0.00000047, -0.00000063, -0.00000164, -0.00001060, -0.00001389, -0.00001209, -0.00003988, -0.00000575, 0.00005481, 0.00002422, 0.00001518, 0.00001260, 0.00002636, 0.00001633, 0.00001927, 0.00004948, 0.00000891, -0.00004140, -0.00002358, -0.00000513, -0.00000891, 0.00000213, -0.00001117, -0.00000097, -0.00000436, },
		{ -0.00000364, 0.00000273, -0.00000517, -0.00000200, -0.00000239, -0.00000741, 0.00000005, -0.00000351, -0.00000475, -0.00000930, 0.00000071, -0.00000298, 0.00000283, -0.00000491, -0.00000511, -0.00000400, -0.00000232, -0.00000852, -0.00000545, -0.00002273, -0.00002246, -0.00009266, 0.00002967, 0.00010192, 0.00003607, 0.00003705, 0.00003210, 0.00004543, 0.00010958, -0.00003218, -0.00007528, -0.00000880, -0.00002722, 0.00000573, -0.00001265, -0.00000114, -0.00000849, 0.00000555, -0.00001094, -0.00000015, -0.00000511, 0.00000292, -0.00000230, 0.00000447, -0.00000217, 0.00000255, 0.00000141, -0.00000300, -0.00000463, 0.00001506, -0.00000719, 0.00000437, -0.00000566, 0.00000206, 0.00000176, -0.00000765, 0.00000091, -0.00000120, -0.00000060, -0.00000009, -0.00000723, 0.00000022, 0.00000752, -0.00000359, -0.00000070, -0.00001406, 0.00000577, -0.00000348, -0.00000988, 0.00000549, -0.00000080, -0.00000214, -0.00001478, 0.00001046, -0.00000769, 0.00000273, -0.00000226, -0.00001148, 0.00001226, -0.00000610, 0.00000008, -0.00000472, 0.00000366, -0.00000162, -0.00000014, -0.00000081, 0.00000143, -0.00000534, -0.00000100, 0.00000256, -0.00000096, -0.00000568, 0.00000117, 0.00001265, -0.00000423, -0.00000302, 0.00000261, -0.00000060, 0.00000379, -0.00000284, -0.00000125, -0.00000289, 0.00000201, -0.00000592, 0.00000280, -0.00000320, -0.00000031, -0.00001683, -0.00000869, -0.00002279, -0.00003527, 0.00002025, 0.00005115, 0.00001232, 0.00001708, 0.00001181, 0.00002971, 0.00000735, 0.00003084, 0.00004139, -0.00000320, -0.00004382, -0.00001334, -0.00000814, -0.00000338, -0.00000296, -0.00000847, -0.00000108, },
		{ -0.00000385, -0.00000057, 0.00000062, -0.00000555, -0.00000055, -0.00000422, -0.00000590, -0.00000001, -0.00000430, -0.00000639, -0.00000687, 0.00000066, -0.00000213, 0.00000246, -0.00000733, -0.00000384, -0.00000279, -0.00000544, -0.00000593, -0.00001100, -0.00002138, -0.00003745, -0.00009001, 0.00007204, 0.00009083, 0.00002988, 0.00004263, 0.00002744, 0.00007829, 0.00008676, -0.00008715, -0.00003828, -0.00002070, -0.00001351, 0.00000036, -0.00000895, -0.00000451, -0.00000359, 0.00000092, -0.00000886, -0.00000122, -0.00000285, 0.00000167, -0.00000065, 0.00000324, -0.00000178, 0.00000298, 0.00000120, -0.00000610, 0.00000145, 0.00001144, -0.00000780, 0.00000588, -0.00000793, 0.00000597, -0.00000262, -0.00000623, 0.00000162, -0.00000190, 0.00000059, -0.00000276, -0.00000739, 0.00000629, 0.00000240, -0.00000251, -0.00000402, -0.00001154, 0.00000906, -0.00001201, -0.00000193, 0.00000332, 0.00000080, -0.00000946, -0.00000705, 0.00000929, -0.00001141, 0.00000968, -0.00001219, -0.00000042, 0.00000606, -0.00000526, -0.00000011, -0.00000343, 0.00000316, -0.00000150, -0.00000029, -0.00000023, 0.00000047, -0.00000593, 0.00000067, 0.00000267, -0.00000244, -0.00000611, 0.00000531, 0.00001106, -0.00000795, 0.00000109, -0.00000004, 0.00000207, 0.00000137, -0.00000235, -0.00000284, 0.00000048, -0.00000164, -0.00000236, 0.00000036, -0.00000111, -0.00000469, -0.00001741, -0.00000829, -0.00003235, -0.00001998, 0.00003974, 0.00003781, 0.00000821, 0.00001664, 0.00001493, 0.00002679, 0.00000440, 0.00003978, 0.00003042, -0.00001342, -0.00004242, -0.00000523, -0.00001053, 0.00000099, -0.00000861, -0.00000336, },
		{ -0.00000003, -0.00000607, 0.00000250, -0.00000263, -0.00000427, -0.00000011, -0.00000583, -0.00000386, -0.00000110, -0.00000478, -0.00000773, -0.00000353, -0.00000108, 0.00000066, -0.00000094, -0.00000713, -0.00000317, -0.00000210, -0.00000876, -0.00000218, -0.00001957, -0.00001722, -0.00005914, -0.00007052, 0.00010770, 0.00007116, 0.00003130, 0.00004419, 0.00003068, 0.00011370, 0.00002335, -0.00010019, -0.00001167, -0.00003236, 0.00000191, -0.00000875, -0.00000270, -0.00000936, 0.00000234, -0.00000558, -0.00000429, -0.00000416, 0.00000110, -0.00000139, 0.00000272, 0.00000001, 0.00000071, 0.00000197, 0.00000093, -0.00000824, 0.00000784, 0.00000525, -0.00000504, 0.00000450, -0.00000764, 0.00000806, -0.00000700, -0.00000355, 0.00000139, -0.00000211, 0.00000146, -0.00000695, -0.00000296, 0.00000824, -0.00000230, -0.00000049, -0.00000857, -0.00000494, 0.00000557, -0.00001505, 0.00000494, 0.00000043, 0.00000102, -0.00001561, 0.00000453, -0.00000035, -0.00000521, 0.00000734, -0.00001478, 0.00000714, -0.00000116, -0.00000180, -0.00000174, -0.00000152, 0.00000197, -0.00000055, -0.00000134, 0.00000121, -0.00000166, -0.00000522, 0.00000192, 0.00000274, -0.00000479, -0.00000487, 0.00000923, 0.00000714, -0.00000903, 0.00000399, -0.00000235, 0.00000483, -0.00000183, -0.00000148, -0.00000350, 0.00000286, -0.00000510, 0.00000207, -0.00000397, 0.00000328, -0.00001341, -0.00001186, -0.00001598, -0.00003306, -0.00000033, 0.00004941, 0.00002086, 0.00001094, 0.00001340, 0.00002119, 0.00001771, 0.00000971, 0.00004168, 0.00002091, -0.00002393, -0.00003556, -0.00000181, -0.00000973, 0.00000105, -0.00001059, },
		{ -0.00000772, -0.00000077, -0.00000522, 0.00000312, -0.00000486, -0.00000251, -0.00000052, -0.00000683, -0.00000200, -0.00000267, -0.00000533, -0.00000773, -0.00000122, -0.00000228, 0.00000302, -0.00000543, -0.00000497, -0.00000306, -0.00000283, -0.00000991, -0.00000082, -0.00002749, -0.00001402, -0.00008328, -0.00003395, 0.00013040, 0.00004711, 0.00004215, 0.00003638, 0.00005514, 0.00012228, -0.00005310, -0.00007286, -0.00001018, -0.00002943, 0.00000722, -0.00001241, -0.00000066, -0.00001013, 0.00000407, -0.00000878, -0.00000179, -0.00000507, 0.00000312, -0.00000292, 0.00000472, -0.00000241, 0.00000317, 0.00000070, -0.00000020, -0.00000800, 0.00001213, -0.00000085, -0.00000087, 0.00000142, -0.00000534, 0.00000810, -0.00001092, 0.00000062, -0.00000084, -0.00000019, -0.00000059, -0.00000897, 0.00000325, 0.00000583, -0.00000441, 0.00000018, -0.00001150, 0.00000227, -0.00000287, -0.00001090, 0.00000698, -0.00000003, -0.00000346, -0.00001442, 0.00001109, -0.00001007, 0.00000538, -0.00000227, -0.00000856, 0.00000677, -0.00000401, 0.00000025, -0.00000241, -0.00000097, 0.00000229, -0.00000107, -0.00000114, 0.00000158, -0.00000333, -0.00000440, 0.00000373, 0.00000140, -0.00000654, -0.00000256, 0.00001251, 0.00000117, -0.00000650, 0.00000371, -0.00000203, 0.00000525, -0.00000366, -0.00000176, -0.00000153, 0.00000147, -0.00000441, 0.00000244, -0.00000465, 0.00000374, -0.00001970, -0.00000664, -0.00002690, -0.00002372, 0.00001819, 0.00004838, 0.00000698, 0.00001614, 0.00001008, 0.00002658, 0.00000710, 0.00002010, 0.00003722, 0.00001335, -0.00003393, -0.00002454, -0.00000271, -0.00000619, -0.00000296, },
		{ -0.00000736, -0.00000284, -0.00000412, -0.00000178, 0.00000097, -0.00000503, -0.00000126, -0.00000141, -0.00000706, -0.00000080, -0.00000432, -0.00000584, -0.00000634, -0.00000080, -0.00000145, 0.00000274, -0.00000839, -0.00000253, -0.00000317, -0.00000513, -0.00000777, -0.00000462, -0.00003076, -0.00001669, -0.00010325, 0.00001699, 0.00013394, 0.00002784, 0.00005613, 0.00002488, 0.00009991, 0.00008202, -0.00010248, -0.00003313, -0.00002452, -0.00001391, 0.00000150, -0.00000883, -0.00000405, -0.00000603, 0.00000070, -0.00000723, -0.00000293, -0.00000256, 0.00000173, -0.00000135, 0.00000368, -0.00000213, 0.00000385, 0.00000014, -0.00000233, -0.00000522, 0.00001330, -0.00000523, 0.00000323, -0.00000245, -0.00000135, 0.00000538, -0.00001268, 0.00000461, -0.00000402, 0.00000324, -0.00000600, -0.00000608, 0.00000719, 0.00000153, -0.00000414, -0.00000116, -0.00001093, 0.00000593, -0.00001120, -0.00000257, 0.00000474, 0.00000138, -0.00001146, -0.00000469, 0.00000702, -0.00001097, 0.00001070, -0.00001057, 0.00000025, 0.00000128, -0.00000240, 0.00000014, -0.00000211, -0.00000133, 0.00000385, -0.00000329, 0.00000087, 0.00000014, -0.00000380, -0.00000378, 0.00000608, -0.00000172, -0.00000685, 0.00000053, 0.00001398, -0.00000495, -0.00000179, 0.00000085, 0.00000092, 0.00000290, -0.00000373, -0.00000264, 0.00000183, -0.00000275, 0.00000030, -0.00000218, 0.00000031, -0.00000321, -0.00001823, -0.00000894, -0.00003217, -0.00000932, 0.00003302, 0.00003758, 0.00000119, 0.00001881, 0.00001009, 0.00002667, 0.00000098, 0.00002957, 0.00003055, 0.00000508, -0.00003994, -0.00001337, -0.00000489, -0.00000310, },
		{ -0.00000314, -0.00000835, -0.00000011, -0.00000655, 0.00000148, -0.00000213, -0.00000375, -0.00000071, -0.00000256, -0.00000660, -0.00000043, -0.00000592, -0.00000566, -0.00000475, -0.00000123, 0.00000064, -0.00000027, -0.00000866, -0.00000088, -0.00000379, -0.00000756, -0.00000394, -0.00001328, -0.00002791, -0.00002889, -0.00011128, 0.00007578, 0.00011497, 0.00002424, 0.00006050, 0.00002783, 0.00014033, 0.00000079, -0.00010183, -0.00001155, -0.00003381, 0.00000041, -0.00000616, -0.00000413, -0.00000782, -0.00000177, -0.00000346, -0.00000440, -0.00000492, 0.00000087, -0.00000092, 0.00000152, 0.00000098, 0.00000013, 0.00000288, 0.00000010, -0.00000460, -0.00000101, 0.00001187, -0.00000754, 0.00000667, -0.00000666, 0.00000416, -0.00000083, -0.00001009, 0.00000541, -0.00000502, 0.00000478, -0.00001154, 0.00000063, 0.00000696, -0.00000212, -0.00000283, -0.00000363, -0.00000717, 0.00000386, -0.00001442, 0.00000458, 0.00000197, 0.00000072, -0.00001687, 0.00000715, -0.00000443, -0.00000187, 0.00000587, -0.00001046, 0.00000384, -0.00000286, -0.00000017, 0.00000002, -0.00000291, -0.00000004, 0.00000394, -0.00000481, 0.00000285, -0.00000177, -0.00000399, -0.00000225, 0.00000738, -0.00000532, -0.00000603, 0.00000457, 0.00001224, -0.00000842, 0.00000201, -0.00000182, 0.00000403, -0.00000033, -0.00000353, -0.00000192, 0.00000295, -0.00000524, 0.00000415, -0.00000697, 0.00000625, -0.00001364, -0.00001128, -0.00001890, -0.00002701, 0.00000468, 0.00004256, 0.00002096, 0.00000413, 0.00001694, 0.00001418, 0.00002037, 0.00000229, 0.00003400, 0.00002490, -0.00000576, -0.00003931, -0.00000558, -0.00000569, },
		{ -0.00000532, -0.00000561, -0.00000551, -0.00000117, -0.00000588, 0.00000226, -0.00000399, -0.00000237, -0.00000046, -0.00000395, -0.00000551, -0.00000103, -0.00000712, -0.00000456, -0.00000414, -0.00000080, 0.00000188, -0.00000410, -0.00000685, -0.00000026, -0.00000535, -0.00000818, -0.00000167, -0.00002362, -0.00002066, -0.00005173, -0.00009925, 0.00013041, 0.00007879, 0.00003984, 0.00004775, 0.00006172, 0.00013955, -0.00008204, -0.00006332, -0.00001671, -0.00002679, 0.00000347, -0.00000780, -0.00000402, -0.00000735, -0.00000166, -0.00000405, -0.00000426, -0.00000426, 0.00000188, -0.00000172, 0.00000281, -0.00000067, 0.00000224, 0.00000146, -0.00000008, -0.00000617, 0.00000354, 0.00000850, -0.00000753, 0.00000853, -0.00000995, 0.00000959, -0.00000926, -0.00000310, 0.00000162, -0.00000167, 0.00000114, -0.00001245, 0.00000691, 0.00000387, -0.00000432, -0.00000144, -0.00000584, -0.00000249, -0.00000256, -0.00001091, 0.00000701, 0.00000160, -0.00000497, -0.00001360, 0.00001157, -0.00001306, 0.00000921, -0.00000452, -0.00000298, 0.00000019, -0.00000214, -0.00000006, 0.00000118, -0.00000565, 0.00000423, 0.00000048, -0.00000325, 0.00000257, -0.00000238, -0.00000509, 0.00000130, 0.00000592, -0.00000768, -0.00000469, 0.00000920, 0.00000719, -0.00000787, 0.00000273, -0.00000193, 0.00000495, -0.00000254, -0.00000400, 0.00000128, -0.00000020, -0.00000255, 0.00000236, -0.00000609, 0.00000615, -0.00001959, -0.00000709, -0.00002874, -0.00001452, 0.00001663, 0.00004430, 0.00000532, 0.00001159, 0.00001264, 0.00001939, 0.00001105, 0.00000907, 0.00003366, 0.00001991, -0.00001759, -0.00003294, -0.00000162, },
		{ 0.00000032, -0.00000598, -0.00000721, -0.00000192, -0.00000413, -0.00000300, 0.00000056, -0.00000383, -0.00000177, -0.00000014, -0.00000564, -0.00000382, -0.00000281, -0.00000705, -0.00000352, -0.00000411, 0.00000070, 0.00000114, -0.00000698, -0.00000396, -0.00000104, -0.00000709, -0.00000682, -0.00000312, -0.00003197, -0.00001285, -0.00008318, -0.00006056, 0.00016442, 0.00004170, 0.00006367, 0.00002824, 0.00012031, 0.00007762, -0.00012164, -0.00002329, -0.00003195, -0.00001083, -0.00000166, -0.00000480, -0.00000704, -0.00000424, -0.00000433, -0.00000184, -0.00000651, -0.00000109, 0.00000032, -0.00000030, 0.00000191, -0.00000029, 0.00000303, 0.00000028, -0.00000045, -0.00000705, 0.00000823, 0.00000312, -0.00000426, 0.00000699, -0.00000984, 0.00001154, -0.00001577, 0.00000467, -0.00000420, 0.00000414, -0.00000697, -0.00000687, 0.00000903, 0.00000049, -0.00000541, -0.00000027, -0.00000683, 0.00000018, -0.00000912, -0.00000355, 0.00000551, 0.00000249, -0.00001341, -0.00000205, 0.00000436, -0.00001055, 0.00001211, -0.00001064, 0.00000386, -0.00000560, 0.00000118, -0.00000083, 0.00000174, -0.00000779, 0.00000911, -0.00000505, 0.00000089, 0.00000001, -0.00000169, -0.00000658, 0.00000612, 0.00000174, -0.00000820, -0.00000268, 0.00001278, 0.00000094, -0.00000475, 0.00000110, 0.00000018, 0.00000363, -0.00000386, -0.00000375, 0.00000437, -0.00000492, 0.00000329, -0.00000408, 0.00000065, -0.00000184, -0.00001721, -0.00001150, -0.00002993, -0.00000153, 0.00002736, 0.00003661, -0.00000290, 0.00001767, 0.00000965, 0.00002173, 0.00000359, 0.00001703, 0.00003119, 0.00001323, -0.00002669, -0.00002426, },
		{ -0.00001590, 0.00000130, -0.00000853, -0.00000562, -0.00000083, -0.00000585, -0.00000049, -0.00000175, -0.00000249, -0.00000185, 0.00000019, -0.00000744, -0.00000177, -0.00000572, -0.00000493, -0.00000409, -0.00000291, 0.00000158, -0.00000067, -0.00000859, -0.00000051, -0.00000393, -0.00000709, -0.00000553, -0.00000794, -0.00003592, -0.00000963, -0.00011647, 0.00000492, 0.00016433, 0.00002385, 0.00007547, 0.00002824, 0.00016755, -0.00002534, -0.00010209, -0.00001124, -0.00003506, -0.00000126, -0.00000458, -0.00000439, -0.00000715, -0.00000409, -0.00000441, -0.00000152, -0.00000711, 0.00000105, -0.00000080, 0.00000081, 0.00000095, 0.00000077, 0.00000302, -0.00000084, -0.00000071, -0.00000744, 0.00001285, -0.00000397, 0.00000251, 0.00000099, -0.00000457, 0.00000708, -0.00001588, 0.00000801, -0.00000694, 0.00000715, -0.00001435, 0.00000214, 0.00000680, -0.00000202, -0.00000549, 0.00000043, -0.00000648, -0.00000085, -0.00001154, 0.00000273, 0.00000377, 0.00000049, -0.00001765, 0.00000959, -0.00000867, 0.00000166, 0.00000406, -0.00000685, 0.00000301, -0.00000690, 0.00000250, 0.00000049, -0.00000129, -0.00000542, 0.00001011, -0.00000849, 0.00000436, -0.00000249, -0.00000153, -0.00000637, 0.00000969, -0.00000322, -0.00000762, 0.00000067, 0.00001340, -0.00000360, -0.00000195, -0.00000030, 0.00000220, 0.00000180, -0.00000556, -0.00000063, 0.00000336, -0.00000580, 0.00000609, -0.00000869, 0.00000676, -0.00001177, -0.00001123, -0.00002152, -0.00002109, 0.00000825, 0.00003590, 0.00002195, -0.00000188, 0.00001921, 0.00000964, 0.00001975, 0.00000053, 0.00002351, 0.00002786, 0.00000424, -0.00003089, },
		{ -0.00003121, -0.00000802, 0.00000008, -0.00001077, -0.00000204, -0.00000255, -0.00000512, -0.00000002, -0.00000281, -0.00000140, -0.00000180, -0.00000012, -0.00000861, -0.00000026, -0.00000857, -0.00000154, -0.00000644, 0.00000031, 0.00000031, -0.00000205, -0.00000914, 0.00000267, -0.00000843, -0.00000455, -0.00000631, -0.00001469, -0.00003397, -0.00001751, -0.00013839, 0.00008501, 0.00013068, 0.00003425, 0.00006373, 0.00006823, 0.00015813, -0.00011606, -0.00004919, -0.00002647, -0.00002067, -0.00000394, -0.00000072, -0.00000937, -0.00000231, -0.00000887, 0.00000045, -0.00000566, -0.00000362, -0.00000021, 0.00000064, -0.00000036, 0.00000184, 0.00000092, 0.00000288, -0.00000176, -0.00000137, -0.00000633, 0.00001561, -0.00001036, 0.00001038, -0.00000745, 0.00000417, -0.00000285, -0.00000886, 0.00000473, -0.00000371, 0.00000335, -0.00001534, 0.00000907, 0.00000292, -0.00000413, -0.00000407, 0.00000029, -0.00000564, -0.00000465, -0.00000861, 0.00000501, 0.00000407, -0.00000664, -0.00001218, 0.00001155, -0.00001586, 0.00001307, -0.00000749, 0.00000287, -0.00000481, -0.00000233, 0.00000050, 0.00000403, -0.00000771, 0.00000226, 0.00000500, -0.00000687, 0.00000435, -0.00000278, -0.00000335, -0.00000293, 0.00000976, -0.00000690, -0.00000666, 0.00000515, 0.00001075, -0.00000516, -0.00000106, 0.00000007, 0.00000264, 0.00000033, -0.00000730, 0.00000472, -0.00000229, -0.00000062, 0.00000199, -0.00000619, 0.00000552, -0.00001582, -0.00001015, -0.00002795, -0.00000833, 0.00001631, 0.00003849, 0.00000698, 0.00000479, 0.00001720, 0.00001118, 0.00001526, 0.00000108, 0.00002858, 0.00002245, -0.00000471, },
		{ -0.00001091, -0.00002969, -0.00000015, -0.00000473, -0.00000954, 0.00000010, -0.00000463, -0.00000332, -0.00000111, -0.00000216, -0.00000146, -0.00000085, -0.00000192, -0.00000809, -0.00000077, -0.00000914, 0.00000049, -0.00000817, 0.00000371, -0.00000221, -0.00000315, -0.00000785, 0.00000329, -0.00001155, -0.00000187, -0.00000850, -0.00002296, -0.00002520, -0.00004265, -0.00013177, 0.00015546, 0.00008261, 0.00006093, 0.00004149, 0.00013703, 0.00007476, -0.00014515, -0.00000856, -0.00004339, -0.00000335, -0.00000992, 0.00000350, -0.00001377, 0.00000173, -0.00001319, 0.00000575, -0.00001077, 0.00000124, -0.00000275, 0.00000286, -0.00000231, 0.00000372, 0.00000053, 0.00000211, -0.00000145, -0.00000385, -0.00000174, 0.00001349, -0.00001189, 0.00001454, -0.00001337, 0.00001107, -0.00001275, 0.00000065, -0.00000183, 0.00000283, -0.00000571, -0.00000879, 0.00001058, -0.00000015, -0.00000623, -0.00000108, -0.00000083, -0.00000570, -0.00000804, -0.00000307, 0.00000452, 0.00000446, -0.00001528, 0.00000077, 0.00000146, -0.00000999, 0.00001351, -0.00001170, 0.00000871, -0.00001205, 0.00000340, -0.00000148, 0.00000590, -0.00001286, 0.00001095, -0.00000342, -0.00000138, 0.00000125, -0.00000126, -0.00000624, 0.00000275, 0.00000639, -0.00000869, -0.00000503, 0.00000914, 0.00000667, -0.00000508, -0.00000098, 0.00000116, 0.00000234, -0.00000191, -0.00000636, 0.00000788, -0.00000771, 0.00000628, -0.00000569, 0.00000132, -0.00000254, -0.00001279, -0.00001697, -0.00002451, 0.00000177, 0.00002488, 0.00003244, -0.00000191, 0.00001178, 0.00001416, 0.00001227, 0.00001058, 0.00000382, 0.00003247, 0.00001396, },
		{ 0.00000457, -0.00001455, -0.00002633, 0.00000600, -0.00001156, -0.00000444, -0.00000115, -0.00000479, -0.00000249, -0.00000209, -0.00000073, -0.00000247, 0.00000101, -0.00000536, -0.00000546, -0.00000394, -0.00000591, -0.00000141, -0.00000618, 0.00000392, -0.00000323, -0.00000547, -0.00000349, -0.00000097, -0.00000963, -0.00000312, -0.00000955, -0.00003262, -0.00001142, -0.00008541, -0.00008479, 0.00019125, 0.00004762, 0.00007897, 0.00004086, 0.00018802, -0.00004938, -0.00010527, -0.00000747, -0.00003903, 0.00000012, -0.00000721, -0.00000074, -0.00000997, -0.00000226, -0.00000981, 0.00000464, -0.00001084, 0.00000232, -0.00000225, 0.00000208, -0.00000161, 0.00000387, 0.00000117, -0.00000008, 0.00000056, -0.00000843, 0.00000681, 0.00000516, -0.00000598, 0.00001128, -0.00001246, 0.00001108, -0.00001648, 0.00000621, -0.00000558, 0.00000647, -0.00001372, 0.00000093, 0.00000774, -0.00000232, -0.00000732, 0.00000231, -0.00000273, -0.00000741, -0.00000798, 0.00000095, 0.00000466, 0.00000075, -0.00001790, 0.00001145, -0.00001233, 0.00000466, 0.00000259, -0.00000462, 0.00000444, -0.00001197, 0.00000492, 0.00000039, 0.00000211, -0.00001140, 0.00001445, -0.00000948, 0.00000371, -0.00000195, -0.00000033, -0.00000755, 0.00000776, 0.00000159, -0.00000931, -0.00000203, 0.00001091, 0.00000339, -0.00000540, 0.00000025, 0.00000105, 0.00000275, -0.00000595, -0.00000071, 0.00000465, -0.00000685, 0.00000769, -0.00000924, 0.00000593, -0.00000984, -0.00000966, -0.00002540, -0.00001379, 0.00000850, 0.00003210, 0.00002027, -0.00000313, 0.00001635, 0.00001101, 0.00001292, 0.00000594, 0.00000922, 0.00003305, },
		{ 0.00002709, -0.00000140, -0.00001845, -0.00001917, 0.00000675, -0.00001577, 0.00000106, -0.00000421, -0.00000331, -0.00000314, -0.00000180, 0.00000016, -0.00000332, 0.00000262, -0.00000922, -0.00000198, -0.00000821, -0.00000027, -0.00000666, -0.00000016, -0.00000023, -0.00000140, -0.00000921, 0.00000293, -0.00000917, -0.00000258, -0.00000948, -0.00000896, -0.00004076, -0.00000068, -0.00013402, -0.00000235, 0.00018324, 0.00004150, 0.00007393, 0.00008428, 0.00016928, -0.00014761, -0.00003675, -0.00003452, -0.00001530, -0.00001068, 0.00000442, -0.00001262, 0.00000092, -0.00001347, 0.00000144, -0.00000399, -0.00000443, -0.00000173, 0.00000232, -0.00000258, 0.00000282, 0.00000129, 0.00000302, -0.00000280, 0.00000250, -0.00001242, 0.00001597, -0.00000619, 0.00000511, 0.00000174, -0.00000522, 0.00000401, -0.00001249, 0.00000516, -0.00000352, 0.00000324, -0.00001495, 0.00000794, 0.00000384, -0.00000474, -0.00000593, 0.00000442, -0.00000532, -0.00000969, -0.00000430, 0.00000176, 0.00000654, -0.00000811, -0.00001006, 0.00001051, -0.00001742, 0.00001569, -0.00000966, 0.00000707, -0.00000673, -0.00000494, 0.00000202, 0.00000540, -0.00000653, -0.00000268, 0.00001014, -0.00000955, 0.00000507, -0.00000287, -0.00000157, -0.00000559, 0.00000980, -0.00000274, -0.00000919, 0.00000194, 0.00001049, 0.00000124, -0.00000622, 0.00000258, -0.00000028, 0.00000326, -0.00000983, 0.00000719, -0.00000367, 0.00000086, 0.00000124, -0.00000473, 0.00000237, -0.00001001, -0.00001384, -0.00002618, -0.00000373, 0.00001572, 0.00003307, 0.00000860, 0.00000005, 0.00001885, 0.00000709, 0.00001443, 0.00000041, 0.00001851, },
		{ 0.00002992, 0.00001466, -0.00000188, -0.00002409, -0.00000782, 0.00000036, -0.00001354, 0.00000269, -0.00000572, -0.00000239, -0.00000400, -0.00000035, -0.00000034, -0.00000270, 0.00000221, -0.00001126, -0.00000027, -0.00001015, 0.00000355, -0.00001095, 0.00000588, -0.00000562, 0.00000094, -0.00001146, 0.00000693, -0.00001573, 0.00000408, -0.00001661, -0.00001058, -0.00004099, -0.00000625, -0.00016546, 0.00009132, 0.00014589, 0.00005589, 0.00005963, 0.00015527, 0.00006787, -0.00016753, 0.00000606, -0.00005485, 0.00000530, -0.00001996, 0.00001251, -0.00002080, 0.00000824, -0.00002183, 0.00001177, -0.00001311, 0.00000282, -0.00000615, 0.00000660, -0.00000708, 0.00000774, -0.00000170, 0.00000402, -0.00000355, 0.00000112, -0.00001133, 0.00001990, -0.00001393, 0.00001487, -0.00000832, 0.00000306, -0.00000498, -0.00000514, 0.00000086, 0.00000135, -0.00000458, -0.00000919, 0.00000973, 0.00000075, -0.00000741, -0.00000201, 0.00000433, -0.00000886, -0.00000990, -0.00000013, 0.00000158, 0.00000699, -0.00001689, 0.00000379, -0.00000198, -0.00000845, 0.00001366, -0.00001212, 0.00001236, -0.00001552, 0.00000256, -0.00000053, 0.00000834, -0.00001404, 0.00000800, 0.00000128, -0.00000489, 0.00000304, -0.00000192, -0.00000405, -0.00000130, 0.00000888, -0.00000615, -0.00000779, 0.00000528, 0.00000957, -0.00000145, -0.00000546, 0.00000368, -0.00000062, 0.00000130, -0.00000923, 0.00001101, -0.00001000, 0.00000866, -0.00000722, 0.00000306, -0.00000582, -0.00000548, -0.00002417, -0.00001696, 0.00000137, 0.00002510, 0.00002571, 0.00000249, 0.00000402, 0.00001970, 0.00000304, 0.00001625, -0.00000471, },
		{ -0.00000483, 0.00003730, 0.00000128, 0.00000001, -0.00002873, 0.00000371, -0.00000987, -0.00000584, -0.00000046, -0.00000420, -0.00000355, -0.00000359, 0.00000127, -0.00000209, -0.00000020, -0.00000123, -0.00000995, -0.00000206, -0.00000726, 0.00000182, -0.00000958, 0.00000665, -0.00000722, -0.00000020, -0.00000854, 0.00000377, -0.00001435, 0.00000308, -0.00001800, -0.00001914, -0.00002865, -0.00003801, -0.00015720, 0.00016525, 0.00010632, 0.00006814, 0.00006892, 0.00019961, -0.00007063, -0.00011102, -0.00000115, -0.00004510, 0.00000448, -0.00001380, 0.00000612, -0.00001549, 0.00000254, -0.00001762, 0.00001166, -0.00001432, 0.00000390, -0.00000500, 0.00000512, -0.00000617, 0.00000860, -0.00000193, 0.00000215, -0.00000090, -0.00000491, -0.00000301, 0.00001491, -0.00001292, 0.00001743, -0.00001321, 0.00000687, -0.00001001, -0.00000029, -0.00000121, 0.00000314, -0.00001064, -0.00000128, 0.00000811, -0.00000220, -0.00000848, 0.00000261, 0.00000213, -0.00001305, -0.00000622, 0.00000106, 0.00000364, 0.00000173, -0.00001759, 0.00001263, -0.00001535, 0.00000741, 0.00000090, -0.00000283, 0.00000620, -0.00001545, 0.00000498, 0.00000136, 0.00000480, -0.00001442, 0.00001378, -0.00000607, 0.00000024, 0.00000039, -0.00000124, -0.00000541, 0.00000284, 0.00000650, -0.00000902, -0.00000466, 0.00000684, 0.00000933, -0.00000572, -0.00000169, 0.00000173, 0.00000152, -0.00000417, -0.00000211, 0.00000639, -0.00000783, 0.00000867, -0.00000914, 0.00000523, -0.00000963, -0.00000546, -0.00003088, -0.00000488, 0.00000495, 0.00003217, 0.00001446, 0.00000135, 0.00000834, 0.00001722, 0.00000198, 0.00001490, },
		{ 0.00000737, 0.00000412, 0.00003416, -0.00000555, -0.00000158, -0.00002739, 0.00000907, -0.00001701, 0.00000180, -0.00000501, -0.00000147, -0.00000599, -0.00000156, 0.00000195, -0.00000389, 0.00000296, -0.00000660, -0.00000619, -0.00000610, -0.00000091, -0.00000477, -0.00000228, 0.00000082, -0.00000327, -0.00000572, -0.00000070, -0.00000618, -0.00000451, -0.00000741, -0.00001197, -0.00003363, -0.00000828, -0.00009176, -0.00010278, 0.00020138, 0.00008221, 0.00006722, 0.00011971, 0.00016452, -0.00016991, -0.00003113, -0.00003704, -0.00001404, -0.00001313, 0.00000421, -0.00001095, -0.00000024, -0.00001331, -0.00000213, 0.00000080, -0.00000672, -0.00000193, 0.00000201, -0.00000231, 0.00000087, 0.00000438, 0.00000093, -0.00000166, 0.00000299, -0.00001242, 0.00000918, 0.00000333, -0.00000388, 0.00001214, -0.00001193, 0.00000536, -0.00000980, 0.00000039, 0.00000062, -0.00000059, -0.00001057, 0.00000385, 0.00000585, -0.00000597, -0.00000635, 0.00000579, -0.00000198, -0.00001595, -0.00000005, -0.00000081, 0.00000765, -0.00000883, -0.00000736, 0.00000828, -0.00001730, 0.00001666, -0.00001071, 0.00000937, -0.00000625, -0.00000829, 0.00000295, 0.00000631, -0.00000388, -0.00000724, 0.00001202, -0.00000833, 0.00000279, -0.00000097, -0.00000189, -0.00000455, 0.00000533, 0.00000364, -0.00001099, -0.00000075, 0.00000746, 0.00000861, -0.00000980, 0.00000336, -0.00000183, 0.00000421, -0.00000985, 0.00000753, -0.00000370, 0.00000163, 0.00000031, -0.00000248, -0.00000145, -0.00000486, -0.00001557, -0.00002543, 0.00000104, 0.00001303, 0.00003050, 0.00000679, 0.00000103, 0.00001429, 0.00000980, 0.00000663, },
		{ 0.00001498, -0.00000409, 0.00002017, 0.00002002, -0.00000287, -0.00000931, -0.00001820, 0.00000510, -0.00001613, 0.00000412, -0.00000664, -0.00000072, -0.00000743, 0.00000118, 0.00000109, -0.00000414, 0.00000443, -0.00001111, -0.00000294, -0.00000861, 0.00000397, -0.00001079, 0.00000559, -0.00000698, 0.00000241, -0.00001142, 0.00000637, -0.00001585, 0.00000628, -0.00002061, -0.00000487, -0.00004527, 0.00000532, -0.00014713, -0.00001735, 0.00020398, 0.00006982, 0.00006958, 0.00018647, 0.00004642, -0.00017940, 0.00001263, -0.00005994, 0.00000956, -0.00002616, 0.00001669, -0.00002308, 0.00001042, -0.00002558, 0.00001256, -0.00001122, 0.00000201, -0.00000793, 0.00000843, -0.00000958, 0.00000895, -0.00000119, 0.00000375, -0.00000396, 0.00000371, -0.00001548, 0.00001819, -0.00000783, 0.00000727, 0.00000333, -0.00000802, 0.00000221, -0.00000797, 0.00000031, 0.00000255, -0.00000624, -0.00000588, 0.00000543, 0.00000339, -0.00000935, -0.00000172, 0.00000668, -0.00000804, -0.00001460, 0.00000434, -0.00000200, 0.00000890, -0.00001765, 0.00000691, -0.00000618, -0.00000527, 0.00001172, -0.00001082, 0.00001342, -0.00001516, -0.00000107, 0.00000163, 0.00000877, -0.00001144, 0.00000217, 0.00000584, -0.00000657, 0.00000306, -0.00000158, -0.00000273, -0.00000269, 0.00000669, -0.00000006, -0.00001083, 0.00000204, 0.00000908, 0.00000494, -0.00001025, 0.00000582, -0.00000336, 0.00000361, -0.00001018, 0.00001196, -0.00001047, 0.00000967, -0.00000834, 0.00000537, -0.00001020, 0.00000210, -0.00003005, -0.00001007, -0.00000016, 0.00002551, 0.00001945, 0.00000604, -0.00000043, 0.00002070, -0.00000033, },
		{ -0.00000680, 0.00002025, -0.00001176, 0.00003458, 0.00000229, 0.00000477, -0.00001951, -0.00000547, -0.00000477, -0.00000850, 0.00000060, -0.00000419, -0.00000332, -0.00000621, 0.00000328, -0.00000106, -0.00000206, 0.00000251, -0.00001241, -0.00000253, -0.00000672, 0.00000314, -0.00001070, 0.00000752, -0.00001012, 0.00000395, -0.00001174, 0.00000617, -0.00001702, 0.00000816, -0.00002654, -0.00000575, -0.00004419, -0.00000354, -0.00017899, 0.00007361, 0.00018903, 0.00005628, 0.00010412, 0.00020970, -0.00009639, -0.00011221, 0.00000118, -0.00004787, 0.00000724, -0.00001966, 0.00001132, -0.00001904, 0.00000558, -0.00002286, 0.00001510, -0.00001444, 0.00000368, -0.00000707, 0.00000759, -0.00000997, 0.00001196, -0.00000355, 0.00000350, -0.00000249, -0.00000099, -0.00001067, 0.00001892, -0.00001246, 0.00001461, -0.00000436, -0.00000490, 0.00000075, -0.00000802, 0.00000309, -0.00000002, -0.00000805, -0.00000158, 0.00000591, -0.00000074, -0.00000969, 0.00000286, 0.00000551, -0.00001520, -0.00000778, 0.00000361, 0.00000093, 0.00000287, -0.00001644, 0.00001317, -0.00001800, 0.00001063, -0.00000203, -0.00000009, 0.00000621, -0.00001530, 0.00000155, 0.00000425, 0.00000512, -0.00001262, 0.00000775, 0.00000031, -0.00000403, 0.00000286, -0.00000276, -0.00000218, -0.00000159, 0.00000796, -0.00000490, -0.00000808, 0.00000303, 0.00001205, -0.00000216, -0.00000564, 0.00000359, -0.00000108, -0.00000152, -0.00000320, 0.00000700, -0.00000754, 0.00000839, -0.00000850, 0.00000507, -0.00001104, 0.00000011, -0.00003593, 0.00000355, -0.00000035, 0.00003428, 0.00000656, 0.00000836, -0.00000030, 0.00002281, },
		{ 0.00001656, -0.00000423, 0.00001644, -0.00000843, 0.00003788, -0.00000891, 0.00000921, -0.00002532, 0.00000338, -0.00001298, -0.00000033, -0.00000452, -0.00000045, -0.00000755, -0.00000254, 0.00000375, -0.00000327, 0.00000128, -0.00000220, -0.00001073, -0.00000422, -0.00000170, -0.00000259, -0.00000429, 0.00000207, -0.00000637, -0.00000044, -0.00000590, -0.00000184, -0.00000830, -0.00000192, -0.00002125, -0.00001726, -0.00002934, -0.00003892, -0.00017322, 0.00015230, 0.00016466, 0.00004440, 0.00017497, 0.00014361, -0.00018357, -0.00003090, -0.00003596, -0.00001540, -0.00001224, -0.00000017, -0.00000602, -0.00000401, -0.00001052, -0.00000740, 0.00000564, -0.00000831, -0.00000190, 0.00000035, -0.00000032, -0.00000280, 0.00000855, -0.00000190, 0.00000066, 0.00000090, -0.00000834, -0.00000042, 0.00001223, -0.00000979, 0.00001649, -0.00000994, -0.00000213, -0.00000011, -0.00000866, 0.00000733, -0.00000657, -0.00000432, -0.00000054, 0.00000675, -0.00000662, -0.00000592, 0.00000547, 0.00000205, -0.00002044, 0.00000160, -0.00000093, 0.00000652, -0.00000860, -0.00000419, 0.00000511, -0.00001583, 0.00001661, -0.00001147, 0.00001097, -0.00000547, -0.00000980, 0.00000136, 0.00000832, -0.00000228, -0.00000798, 0.00000805, -0.00000235, -0.00000227, 0.00000265, -0.00000419, -0.00000061, -0.00000118, 0.00000874, -0.00000964, -0.00000419, 0.00000385, 0.00001380, -0.00000946, 0.00000119, -0.00000112, 0.00000251, -0.00000721, 0.00000615, -0.00000302, 0.00000224, -0.00000092, -0.00000021, -0.00000433, -0.00000210, -0.00001442, -0.00002580, 0.00000587, 0.00000839, 0.00003095, 0.00000113, 0.00000741, 0.00000508, },
		{ 0.00001464, 0.00000380, 0.00000616, 0.00000449, 0.00000573, 0.00002760, -0.00000870, 0.00000585, -0.00002314, 0.00000410, -0.00001420, 0.00000282, -0.00000648, 0.00000071, -0.00001027, 0.00000189, 0.00000233, -0.00000399, 0.00000361, -0.00000725, -0.00000837, -0.00000494, 0.00000191, -0.00000782, 0.00000281, -0.00000570, 0.00000015, -0.00000662, 0.00000077, -0.00001088, 0.00000274, -0.00001714, -0.00001050, -0.00003261, -0.00001091, -0.00008995, -0.00013198, 0.00021469, 0.00012551, 0.00006089, 0.00023940, 0.00000285, -0.00017462, 0.00000646, -0.00005521, 0.00000629, -0.00002508, 0.00001290, -0.00001814, 0.00000614, -0.00002280, 0.00000756, -0.00000543, -0.00000108, -0.00000732, 0.00000703, -0.00000834, 0.00000620, 0.00000269, 0.00000075, -0.00000155, 0.00000209, -0.00001284, 0.00000914, 0.00000342, -0.00000337, 0.00001471, -0.00001475, 0.00000280, -0.00000411, -0.00000534, 0.00000738, -0.00001115, 0.00000073, -0.00000088, 0.00000617, -0.00001122, -0.00000023, 0.00000629, -0.00000448, -0.00001976, 0.00000778, -0.00000400, 0.00000867, -0.00001690, 0.00000976, -0.00001090, -0.00000064, 0.00000799, -0.00000817, 0.00001240, -0.00001235, -0.00000527, 0.00000309, 0.00000863, -0.00000745, -0.00000263, 0.00000623, -0.00000366, -0.00000030, 0.00000107, -0.00000385, -0.00000024, 0.00000038, 0.00000724, -0.00001211, -0.00000143, 0.00000678, 0.00001102, -0.00001230, 0.00000531, -0.00000390, 0.00000307, -0.00000771, 0.00000986, -0.00000870, 0.00000911, -0.00000867, 0.00000714, -0.00001344, 0.00000690, -0.00003181, -0.00000609, -0.00000068, 0.00002400, 0.00001619, 0.00000545, 0.00000166, },
		{ -0.00000383, 0.00002192, -0.00000777, 0.00001668, -0.00000764, 0.00002240, 0.00001028, -0.00000055, -0.00000306, -0.00001561, -0.00000101, -0.00000931, 0.00000046, -0.00000418, -0.00000222, -0.00000931, 0.00000504, -0.00000019, -0.00000267, 0.00000334, -0.00001060, -0.00000691, -0.00000290, 0.00000092, -0.00000781, 0.00000464, -0.00000937, 0.00000356, -0.00000922, 0.00000228, -0.00001330, 0.00000650, -0.00002716, -0.00000374, -0.00004223, -0.00000201, -0.00014039, -0.00006370, 0.00025845, 0.00007075, 0.00013062, 0.00023106, -0.00013788, -0.00009902, -0.00000884, -0.00004057, 0.00000237, -0.00001865, 0.00000889, -0.00001526, 0.00000181, -0.00002086, 0.00001137, -0.00000894, -0.00000011, -0.00000619, 0.00000665, -0.00000982, 0.00001088, -0.00000112, 0.00000163, -0.00000132, -0.00000075, -0.00001151, 0.00001375, -0.00000314, 0.00000359, 0.00001029, -0.00001817, 0.00000941, -0.00001174, 0.00000354, -0.00000014, -0.00000841, 0.00000167, 0.00000082, 0.00000164, -0.00001090, 0.00000388, 0.00000613, -0.00001339, -0.00001181, 0.00000705, -0.00000170, 0.00000276, -0.00001379, 0.00001290, -0.00002033, 0.00001455, -0.00000647, 0.00000396, 0.00000389, -0.00001158, -0.00000424, 0.00000795, 0.00000351, -0.00000711, -0.00000062, 0.00000559, -0.00000551, 0.00000280, -0.00000254, -0.00000072, -0.00000247, 0.00000428, 0.00000256, -0.00001132, -0.00000035, 0.00001169, 0.00000340, -0.00000906, 0.00000437, -0.00000284, -0.00000037, -0.00000177, 0.00000484, -0.00000485, 0.00000622, -0.00000689, 0.00000465, -0.00001218, 0.00000397, -0.00003717, 0.00000832, -0.00000416, 0.00003510, 0.00000040, 0.00001285, },
		{ 0.00001469, -0.00000298, 0.00002070, -0.00001084, 0.00001965, -0.00001158, 0.00003171, -0.00000417, 0.00000794, -0.00001174, -0.00000846, -0.00000609, -0.00000351, -0.00000351, -0.00000041, -0.00000742, -0.00000476, 0.00000553, -0.00000221, -0.00000050, 0.00000098, -0.00001238, -0.00000535, -0.00000012, -0.00000283, -0.00000349, 0.00000081, -0.00000742, 0.00000240, -0.00000771, -0.00000137, -0.00000846, 0.00000034, -0.00002760, -0.00000512, -0.00004284, -0.00000848, -0.00017989, 0.00002877, 0.00026902, 0.00002650, 0.00023662, 0.00011744, -0.00019837, -0.00002712, -0.00003923, -0.00001280, -0.00001368, -0.00000291, -0.00000377, -0.00000479, -0.00001060, -0.00000883, 0.00000574, -0.00000578, -0.00000411, -0.00000014, 0.00000041, -0.00000473, 0.00001019, -0.00000221, 0.00000128, -0.00000048, -0.00000496, -0.00000645, 0.00001404, -0.00000704, 0.00001081, 0.00000170, -0.00001608, 0.00001225, -0.00001739, 0.00001260, -0.00001116, 0.00000027, -0.00000209, 0.00000463, -0.00000589, -0.00000538, 0.00000502, 0.00000434, -0.00002098, -0.00000038, 0.00000144, 0.00000378, -0.00000819, -0.00000033, 0.00000125, -0.00001358, 0.00001640, -0.00001291, 0.00001303, -0.00000608, -0.00000801, -0.00000322, 0.00001171, -0.00000284, -0.00000391, -0.00000092, 0.00000576, -0.00000713, 0.00000549, -0.00000615, 0.00000318, -0.00000573, 0.00000910, -0.00000384, -0.00000864, 0.00000096, 0.00001529, -0.00000517, -0.00000283, 0.00000064, -0.00000047, -0.00000377, 0.00000506, -0.00000334, 0.00000394, -0.00000304, 0.00000206, -0.00000613, -0.00000119, -0.00001216, -0.00002490, 0.00000836, 0.00000435, 0.00003176, -0.00000528, },
		{ 0.00000117, 0.00000870, 0.00000461, 0.00001106, -0.00000462, 0.00001382, -0.00000549, 0.00002975, -0.00001003, 0.00001169, -0.00001668, -0.00000487, -0.00000735, -0.00000104, -0.00000510, 0.00000112, -0.00001152, 0.00000134, 0.00000329, -0.00000236, 0.00000068, -0.00000182, -0.00001377, -0.00000172, -0.00000037, -0.00000436, -0.00000025, -0.00000371, -0.00000374, 0.00000012, -0.00000635, -0.00000450, -0.00000300, -0.00000997, -0.00002216, -0.00001174, -0.00003812, -0.00002886, -0.00020302, 0.00014424, 0.00022670, 0.00004038, 0.00030939, -0.00005892, -0.00015713, -0.00000841, -0.00004463, -0.00000138, -0.00001914, 0.00000379, -0.00000907, -0.00000149, -0.00001684, 0.00000064, 0.00000047, -0.00000372, -0.00000594, 0.00000368, -0.00000491, 0.00000156, 0.00000748, -0.00000270, 0.00000184, -0.00000190, -0.00000663, -0.00000175, 0.00001286, -0.00000950, 0.00001835, -0.00001167, -0.00000540, 0.00000581, -0.00001397, 0.00001338, -0.00001671, 0.00000751, -0.00000580, 0.00000646, -0.00001154, 0.00000144, 0.00000474, -0.00000091, -0.00002224, 0.00000774, -0.00000287, 0.00000570, -0.00001459, 0.00001216, -0.00001554, 0.00000455, 0.00000371, -0.00000555, 0.00001090, -0.00000949, -0.00000740, 0.00000200, 0.00000946, -0.00000462, -0.00000323, 0.00000058, 0.00000372, -0.00000588, 0.00000488, -0.00000633, 0.00000414, -0.00000652, 0.00001168, -0.00000915, -0.00000623, 0.00000465, 0.00001401, -0.00000979, 0.00000151, -0.00000188, -0.00000036, -0.00000239, 0.00000576, -0.00000581, 0.00000785, -0.00000849, 0.00000772, -0.00001423, 0.00000772, -0.00002914, -0.00000466, -0.00000059, 0.00002104, 0.00001553, },
		{ -0.00000166, 0.00001194, 0.00000021, 0.00001321, -0.00000075, 0.00000527, 0.00000442, 0.00000548, 0.00002006, -0.00000822, 0.00001061, -0.00001860, -0.00000347, -0.00000579, -0.00000191, -0.00000353, -0.00000113, -0.00001182, 0.00000614, -0.00000003, -0.00000101, 0.00000047, -0.00000467, -0.00001419, 0.00000319, -0.00000416, -0.00000155, -0.00000160, -0.00000475, -0.00000198, -0.00000058, -0.00000787, -0.00000307, -0.00000289, -0.00001774, -0.00001586, -0.00002050, -0.00003079, -0.00006378, -0.00019703, 0.00026465, 0.00013629, 0.00013938, 0.00027018, -0.00020026, -0.00006762, -0.00003386, -0.00002144, -0.00001194, -0.00000851, -0.00000314, -0.00000287, -0.00000969, -0.00001115, 0.00000098, 0.00000103, -0.00000680, -0.00000204, 0.00000142, -0.00000484, 0.00000492, 0.00000525, -0.00000335, 0.00000303, -0.00000509, -0.00000542, 0.00000161, 0.00001056, -0.00000919, 0.00002261, -0.00002497, 0.00001011, -0.00000829, -0.00000101, 0.00000299, -0.00001136, 0.00000713, -0.00000483, 0.00000279, -0.00001086, 0.00000514, 0.00000476, -0.00000959, -0.00001540, 0.00000850, -0.00000194, -0.00000002, -0.00000914, 0.00001152, -0.00002193, 0.00001854, -0.00001142, 0.00000823, 0.00000052, -0.00000645, -0.00000959, 0.00001015, 0.00000190, -0.00000091, -0.00000689, 0.00000568, -0.00000180, -0.00000092, 0.00000021, -0.00000191, 0.00000037, -0.00000277, 0.00000998, -0.00001177, -0.00000448, 0.00001012, 0.00000781, -0.00000902, 0.00000206, -0.00000204, -0.00000231, 0.00000313, -0.00000038, 0.00000012, 0.00000238, -0.00000419, 0.00000295, -0.00001098, 0.00000351, -0.00003242, 0.00000778, -0.00000471, 0.00003270, },
		{ 0.00002956, -0.00000837, 0.00001761, -0.00000412, 0.00001672, -0.00000818, 0.00001265, -0.00000269, 0.00001489, 0.00000871, -0.00000252, 0.00000659, -0.00001933, -0.00000151, -0.00000463, -0.00000336, -0.00000089, -0.00000599, -0.00000765, 0.00000723, -0.00000178, -0.00000028, 0.00000039, -0.00000901, -0.00001080, 0.00000499, -0.00000692, 0.00000172, -0.00000485, -0.00000373, -0.00000078, -0.00000217, -0.00000873, -0.00000055, -0.00000652, -0.00002265, -0.00000874, -0.00003334, -0.00001800, -0.00012007, -0.00013745, 0.00034605, 0.00004863, 0.00028489, 0.00010083, -0.00022653, -0.00000967, -0.00005513, 0.00000040, -0.00002348, 0.00000217, -0.00001011, 0.00000261, -0.00001828, -0.00000220, -0.00000197, 0.00000270, -0.00001020, 0.00000282, -0.00000302, -0.00000175, 0.00000637, 0.00000237, -0.00000200, 0.00000158, -0.00000605, -0.00000497, 0.00000639, 0.00000448, -0.00000241, 0.00001741, -0.00002907, 0.00002017, -0.00002031, 0.00001271, -0.00001163, 0.00000109, 0.00000023, -0.00000014, -0.00000464, -0.00000441, 0.00000493, 0.00000420, -0.00001785, -0.00000442, 0.00000419, 0.00000155, -0.00000910, 0.00000478, -0.00000330, -0.00001072, 0.00001618, -0.00001492, 0.00001544, -0.00000800, -0.00000376, -0.00000896, 0.00001481, -0.00000446, 0.00000289, -0.00001081, 0.00001119, -0.00000791, 0.00000477, -0.00000539, 0.00000413, -0.00000585, 0.00000406, 0.00000478, -0.00001228, -0.00000174, 0.00001395, 0.00000059, -0.00000577, 0.00000094, -0.00000230, -0.00000205, 0.00000643, -0.00000616, 0.00000761, -0.00000651, 0.00000474, -0.00000784, -0.00000001, -0.00001190, -0.00001959, 0.00000574, 0.00000392, },
		{ 0.00001938, 0.00001441, -0.00000297, 0.00001526, -0.00000265, 0.00001400, -0.00000945, 0.00001562, -0.00000587, 0.00002027, -0.00000097, 0.00000475, -0.00000009, -0.00001823, 0.00000102, -0.00000516, -0.00000338, 0.00000035, -0.00001077, -0.00000057, 0.00000422, -0.00000049, -0.00000183, 0.00000164, -0.00001547, -0.00000206, 0.00000003, -0.00000350, -0.00000040, -0.00000438, -0.00000515, 0.00000352, -0.00000890, -0.00000309, -0.00000403, -0.00000755, -0.00002829, 0.00000086, -0.00005176, 0.00000028, -0.00019779, -0.00000698, 0.00034225, 0.00003900, 0.00037587, -0.00012271, -0.00014078, -0.00002003, -0.00003840, -0.00000507, -0.00001569, -0.00000327, -0.00000320, -0.00000564, -0.00001418, -0.00000198, 0.00000123, -0.00000214, -0.00000674, 0.00000157, -0.00000298, -0.00000080, 0.00000897, -0.00000286, 0.00000280, -0.00000435, -0.00000220, -0.00000785, 0.00001414, -0.00000605, 0.00001137, 0.00000058, -0.00001843, 0.00001629, -0.00002021, 0.00001610, -0.00001904, 0.00001080, -0.00000635, 0.00000277, -0.00000980, 0.00000261, 0.00000353, 0.00000051, -0.00002041, 0.00000385, 0.00000080, 0.00000115, -0.00001181, 0.00001442, -0.00001972, 0.00000946, -0.00000006, -0.00000394, 0.00001002, -0.00000800, -0.00000646, -0.00000160, 0.00001114, -0.00000361, 0.00000056, -0.00000910, 0.00001187, -0.00000995, 0.00000678, -0.00000733, 0.00000696, -0.00000994, 0.00001028, -0.00000149, -0.00001185, 0.00000326, 0.00001324, -0.00000362, -0.00000371, 0.00000080, -0.00000465, 0.00000326, 0.00000222, -0.00000392, 0.00000740, -0.00000862, 0.00000752, -0.00001306, 0.00000591, -0.00002461, -0.00000283, -0.00000253, },
		{ -0.00000335, 0.00002827, -0.00000082, 0.00000717, 0.00000880, 0.00000139, 0.00000821, -0.00000667, 0.00001560, -0.00000671, 0.00002258, -0.00000842, 0.00001278, -0.00001012, -0.00001293, 0.00000146, -0.00000581, -0.00000227, -0.00000086, -0.00001272, 0.00000600, -0.00000036, 0.00000226, -0.00000434, 0.00000235, -0.00002074, 0.00000820, -0.00000902, 0.00000495, -0.00000808, 0.00000074, -0.00000925, 0.00000973, -0.00001896, 0.00000803, -0.00001423, -0.00000489, -0.00003310, 0.00000725, -0.00006630, 0.00000757, -0.00027026, 0.00017349, 0.00025093, 0.00014458, 0.00031600, -0.00027409, -0.00002677, -0.00006582, 0.00000230, -0.00002982, 0.00000589, -0.00001982, 0.00001287, -0.00002377, 0.00000107, -0.00001068, 0.00001051, -0.00001272, 0.00000283, -0.00000576, 0.00000229, -0.00000269, 0.00001202, -0.00000818, 0.00000775, -0.00001102, 0.00000308, -0.00001081, 0.00002075, -0.00001557, 0.00002522, -0.00002065, 0.00000187, 0.00000051, -0.00000750, 0.00000618, -0.00001364, 0.00001115, -0.00000743, 0.00000018, -0.00000824, 0.00000554, 0.00000322, -0.00000666, -0.00001554, 0.00000588, 0.00000127, -0.00000555, -0.00000288, 0.00000913, -0.00002226, 0.00002149, -0.00001533, 0.00001108, -0.00000205, -0.00000242, -0.00001192, 0.00000914, 0.00000176, 0.00000344, -0.00000824, -0.00000043, 0.00000596, -0.00000625, 0.00000357, -0.00000385, 0.00000412, -0.00000892, 0.00001305, -0.00000721, -0.00001012, 0.00000893, 0.00000886, -0.00000447, -0.00000325, 0.00000098, -0.00000665, 0.00001011, -0.00000680, 0.00000557, -0.00000175, -0.00000110, -0.00000008, -0.00000696, -0.00000135, -0.00002268, 0.00000354, },
		{ -0.00000240, 0.00000773, 0.00002331, -0.00000659, 0.00001422, 0.00000280, 0.00000490, 0.00000207, -0.00000142, 0.00001292, -0.00000542, 0.00002187, -0.00001227, 0.00001891, -0.00002142, -0.00000350, -0.00000179, -0.00000478, -0.00000135, -0.00000420, -0.00001009, 0.00000838, -0.00000241, 0.00000262, -0.00000414, -0.00000157, -0.00001880, 0.00001189, -0.00001344, 0.00000999, -0.00001378, 0.00000406, -0.00000955, 0.00000978, -0.00002237, 0.00001390, -0.00002158, -0.00000664, -0.00002767, -0.00000329, -0.00005757, -0.00002714, -0.00028407, 0.00033703, 0.00013784, 0.00031212, 0.00009806, -0.00027069, 0.00002259, -0.00008380, 0.00002373, -0.00004156, 0.00001564, -0.00002535, 0.00001781, -0.00003284, 0.00001131, -0.00001556, 0.00001489, -0.00001862, 0.00000881, -0.00001067, 0.00000609, -0.00000232, 0.00001071, -0.00000813, 0.00000660, -0.00001141, 0.00000274, -0.00000701, 0.00001879, -0.00001534, 0.00002815, -0.00003301, 0.00001832, -0.00001511, 0.00000738, -0.00000860, -0.00000068, 0.00000426, -0.00000463, -0.00000533, -0.00000170, 0.00000444, 0.00000284, -0.00001353, -0.00000728, 0.00000435, 0.00000204, -0.00001249, 0.00001134, -0.00000851, -0.00000701, 0.00001523, -0.00001628, 0.00001677, -0.00000958, 0.00000040, -0.00001279, 0.00001515, -0.00000498, 0.00000900, -0.00001696, 0.00001016, -0.00000283, 0.00000011, -0.00000184, 0.00000221, -0.00000241, -0.00000344, 0.00001200, -0.00001209, -0.00000566, 0.00001182, 0.00000465, -0.00000493, -0.00000191, -0.00000154, -0.00000329, 0.00001075, -0.00001122, 0.00001266, -0.00001076, 0.00000778, -0.00001030, 0.00000315, -0.00001555, -0.00000874, },
		{ 0.00000249, -0.00000886, 0.00002149, 0.00000984, -0.00000309, 0.00001628, -0.00000182, 0.00000791, -0.00000439, 0.00000690, 0.00000552, 0.00000056, 0.00001568, -0.00000909, 0.00001790, -0.00002805, 0.00000554, -0.00000657, -0.00000253, -0.00000120, -0.00000842, -0.00000333, 0.00000528, 0.00000028, -0.00000165, 0.00000060, -0.00001156, -0.00000706, 0.00000409, -0.00000657, 0.00000393, -0.00000963, -0.00000128, -0.00000004, -0.00000304, -0.00001044, 0.00000393, -0.00001570, -0.00002070, -0.00000583, -0.00003788, -0.00001618, -0.00012242, -0.00019339, 0.00040921, 0.00010084, 0.00041429, -0.00017076, -0.00013964, -0.00001732, -0.00004525, 0.00000211, -0.00002095, -0.00000189, -0.00000655, -0.00000118, -0.00001931, 0.00000350, -0.00000576, 0.00000520, -0.00001129, 0.00000302, -0.00000551, 0.00000219, 0.00000449, 0.00000228, -0.00000056, -0.00000279, -0.00000288, -0.00000600, 0.00000600, 0.00000570, -0.00000211, 0.00001492, -0.00002800, 0.00002003, -0.00001892, 0.00001237, -0.00001610, 0.00000930, -0.00000245, -0.00000362, -0.00000750, 0.00000396, 0.00000281, -0.00000033, -0.00001530, -0.00000161, 0.00000433, -0.00000249, -0.00001026, 0.00001715, -0.00002339, 0.00001384, -0.00000318, -0.00000307, 0.00000933, -0.00000741, -0.00000381, -0.00000545, 0.00001169, -0.00000292, 0.00000599, -0.00001802, 0.00001565, -0.00000868, 0.00000426, -0.00000482, 0.00000607, -0.00000827, 0.00000357, 0.00000807, -0.00001562, 0.00000141, 0.00001029, 0.00000316, -0.00000704, 0.00000142, -0.00000712, 0.00000649, 0.00000152, -0.00000450, 0.00000864, -0.00000966, 0.00000728, -0.00001147, 0.00000411, -0.00002165, },
		{ -0.00001179, -0.00000036, -0.00000346, 0.00002509, -0.00000068, 0.00000309, 0.00001566, -0.00000602, 0.00001114, -0.00001094, 0.00001756, -0.00000685, 0.00001241, 0.00000292, 0.00000261, 0.00000637, -0.00002454, 0.00000846, -0.00000885, -0.00000114, -0.00000175, -0.00001140, 0.00000426, -0.00000042, 0.00000520, -0.00000782, 0.00000626, -0.00002235, 0.00000850, -0.00001019, 0.00000788, -0.00001076, 0.00000268, -0.00001285, 0.00001505, -0.00002344, 0.00001206, -0.00001875, 0.00000064, -0.00004112, 0.00002130, -0.00008046, 0.00003532, -0.00025269, -0.00000564, 0.00037074, 0.00018831, 0.00034001, -0.00033713, 0.00000481, -0.00008894, 0.00001742, -0.00004044, 0.00001501, -0.00003171, 0.00002279, -0.00003202, 0.00000799, -0.00001640, 0.00001351, -0.00001339, 0.00000462, -0.00001074, 0.00000684, -0.00000687, 0.00001427, -0.00000857, 0.00000888, -0.00001404, 0.00000813, -0.00001670, 0.00002131, -0.00001119, 0.00001648, -0.00000745, -0.00001000, 0.00000836, -0.00001027, 0.00000503, -0.00001165, 0.00001052, -0.00000476, -0.00000678, -0.00000331, 0.00000491, 0.00000261, -0.00000621, -0.00001137, -0.00000031, 0.00000650, -0.00001196, 0.00000335, 0.00000645, -0.00002123, 0.00002275, -0.00001730, 0.00001173, -0.00000302, -0.00000053, -0.00001076, 0.00000552, 0.00000255, 0.00000572, -0.00000528, -0.00000972, 0.00001320, -0.00000892, 0.00000414, -0.00000347, 0.00000524, -0.00001042, 0.00000962, 0.00000180, -0.00001587, 0.00000794, 0.00000676, 0.00000281, -0.00000895, 0.00000371, -0.00001074, 0.00001607, -0.00001133, 0.00000897, -0.00000427, 0.00000110, -0.00000341, -0.00000159, -0.00000807, },
		{ -0.00002207, 0.00000518, -0.00001286, 0.00001299, 0.00001543, -0.00000230, 0.00000744, 0.00001318, -0.00000835, 0.00001223, -0.00001369, 0.00002496, -0.00001848, 0.00002522, -0.00001154, 0.00001781, -0.00001202, -0.00001170, 0.00000381, -0.00000691, -0.00000161, -0.00000316, -0.00001072, 0.00000807, -0.00000328, 0.00000665, -0.00000979, 0.00000583, -0.00002469, 0.00001626, -0.00001829, 0.00001701, -0.00002152, 0.00001140, -0.00001861, 0.00002156, -0.00003461, 0.00002669, -0.00003605, 0.00001008, -0.00004818, 0.00002758, -0.00009583, 0.00004656, -0.00034808, 0.00020785, 0.00028571, 0.00034058, 0.00009126, -0.00031578, 0.00005591, -0.00011285, 0.00004634, -0.00005898, 0.00002985, -0.00004191, 0.00003311, -0.00004685, 0.00002446, -0.00002789, 0.00002441, -0.00002452, 0.00001408, -0.00001885, 0.00001465, -0.00001121, 0.00001796, -0.00001275, 0.00001070, -0.00001684, 0.00001093, -0.00001842, 0.00002729, -0.00001957, 0.00002699, -0.00002446, 0.00000719, -0.00000484, 0.00000054, -0.00000583, -0.00000139, 0.00000618, -0.00000537, -0.00001000, 0.00000358, 0.00000279, 0.00000192, -0.00001065, -0.00000638, 0.00000046, 0.00000561, -0.00001779, 0.00001839, -0.00001384, -0.00000234, 0.00001277, -0.00001566, 0.00001560, -0.00000920, 0.00000227, -0.00001261, 0.00001158, -0.00000333, 0.00001224, -0.00001720, 0.00000268, 0.00000583, -0.00000552, 0.00000179, 0.00000008, 0.00000104, -0.00000871, 0.00001361, -0.00000636, -0.00001104, 0.00001008, 0.00000542, -0.00000001, -0.00000712, 0.00000082, -0.00000618, 0.00001598, -0.00001608, 0.00001680, -0.00001403, 0.00001013, -0.00001309, 0.00000796, },
		{ 0.00000394, -0.00002178, 0.00001167, -0.00001890, 0.00002682, 0.00000186, 0.00000131, 0.00001106, 0.00000670, -0.00000514, 0.00000698, -0.00000718, 0.00002125, -0.00002004, 0.00002975, -0.00001840, 0.00002640, -0.00002724, 0.00000277, -0.00000393, -0.00000305, -0.00000290, -0.00000564, -0.00000563, 0.00000581, -0.00000015, 0.00000132, -0.00000428, -0.00000393, -0.00001404, 0.00000928, -0.00001137, 0.00001054, -0.00001741, 0.00000617, -0.00000879, 0.00000867, -0.00002316, 0.00001737, -0.00003113, -0.00000266, -0.00002719, -0.00000472, -0.00005973, -0.00002263, -0.00033900, 0.00035960, 0.00025520, 0.00041878, -0.00020128, -0.00015398, -0.00000156, -0.00006295, 0.00001758, -0.00003306, 0.00000682, -0.00001776, 0.00000989, -0.00002993, 0.00001436, -0.00001717, 0.00001497, -0.00001723, 0.00000686, -0.00001168, 0.00000942, -0.00000419, 0.00001046, -0.00000617, 0.00000131, -0.00000733, 0.00000115, -0.00000641, 0.00001846, -0.00001318, 0.00002188, -0.00002649, 0.00001284, -0.00000909, 0.00000314, -0.00000961, 0.00000522, 0.00000281, -0.00000913, -0.00000737, 0.00000690, 0.00000167, -0.00000192, -0.00000972, -0.00000519, 0.00000478, -0.00000318, -0.00001088, 0.00002038, -0.00002642, 0.00001781, -0.00000634, -0.00000166, 0.00000738, -0.00000599, -0.00000198, -0.00000660, 0.00000888, -0.00000054, 0.00000964, -0.00002176, 0.00001196, -0.00000129, -0.00000210, 0.00000036, 0.00000244, -0.00000353, -0.00000445, 0.00001468, -0.00001439, -0.00000237, 0.00000730, 0.00000746, -0.00000613, -0.00000137, -0.00000668, 0.00000645, 0.00000373, -0.00000695, 0.00001071, -0.00001095, 0.00000699, -0.00001035, },
		{ 0.00000303, -0.00001337, -0.00000369, -0.00000051, -0.00000742, 0.00002539, -0.00000400, 0.00000336, 0.00001607, -0.00000436, 0.00000430, -0.00000483, 0.00000879, 0.00000427, -0.00000700, 0.00002045, -0.00001170, 0.00002097, -0.00003029, 0.00001026, -0.00000860, -0.00000091, -0.00000361, -0.00000818, 0.00000127, 0.00000021, 0.00000615, -0.00000753, 0.00000458, -0.00001735, 0.00000321, -0.00000674, 0.00000586, -0.00000793, -0.00000108, -0.00000977, 0.00001159, -0.00001749, 0.00000531, -0.00001216, -0.00000691, -0.00003200, 0.00001154, -0.00006142, 0.00001067, -0.00016056, -0.00021205, 0.00041805, 0.00032402, 0.00031314, -0.00036913, 0.00001130, -0.00009104, 0.00001458, -0.00003647, 0.00001220, -0.00003210, 0.00002068, -0.00002924, 0.00000518, -0.00001258, 0.00000768, -0.00000739, 0.00000169, -0.00001096, 0.00000573, -0.00000460, 0.00000962, -0.00000289, 0.00000490, -0.00001198, 0.00000693, -0.00001380, 0.00001218, 0.00000128, 0.00000202, 0.00000636, -0.00001680, 0.00000834, -0.00000505, -0.00000256, -0.00000450, 0.00000522, 0.00000193, -0.00001557, 0.00000150, 0.00000458, 0.00000245, -0.00000761, -0.00000477, -0.00000700, 0.00001049, -0.00001634, 0.00000752, 0.00000442, -0.00001920, 0.00002244, -0.00001764, 0.00001097, -0.00000331, 0.00000023, -0.00000802, 0.00000195, 0.00000206, 0.00000772, -0.00000122, -0.00001696, 0.00001479, -0.00000550, -0.00000002, 0.00000065, 0.00000235, -0.00000675, 0.00000148, 0.00001140, -0.00001846, 0.00000546, 0.00000363, 0.00000936, -0.00001158, 0.00000338, -0.00001190, 0.00001828, -0.00001183, 0.00000894, -0.00000432, 0.00000167, -0.00000589, },
		{ -0.00001476, 0.00001225, -0.00002798, 0.00001734, -0.00002058, 0.00001540, 0.00000992, 0.00000018, 0.00000262, 0.00001993, -0.00001494, 0.00001407, -0.00001596, 0.00002538, -0.00001756, 0.00001444, 0.00000185, 0.00000479, 0.00000374, -0.00002059, 0.00000823, -0.00000784, -0.00000178, -0.00000376, -0.00000851, 0.00000537, -0.00000288, 0.00000896, -0.00001237, 0.00000817, -0.00002366, 0.00001389, -0.00001729, 0.00001809, -0.00002294, 0.00001251, -0.00002099, 0.00002477, -0.00003619, 0.00002799, -0.00003890, 0.00001403, -0.00005275, 0.00003601, -0.00010115, 0.00006051, -0.00030156, -0.00002202, 0.00043428, 0.00042438, 0.00004384, -0.00033383, 0.00006702, -0.00012286, 0.00005213, -0.00006244, 0.00003301, -0.00004833, 0.00003757, -0.00005035, 0.00002812, -0.00003069, 0.00002446, -0.00002263, 0.00001389, -0.00002235, 0.00001812, -0.00001426, 0.00001850, -0.00001104, 0.00000943, -0.00001735, 0.00001340, -0.00002107, 0.00002434, -0.00001167, 0.00001378, -0.00000730, -0.00000671, 0.00000361, -0.00000214, -0.00000738, 0.00000199, 0.00000363, -0.00000125, -0.00001804, 0.00001012, 0.00000062, 0.00000181, -0.00001001, -0.00000182, -0.00000601, 0.00000999, -0.00002245, 0.00002376, -0.00001804, 0.00000280, 0.00000873, -0.00001279, 0.00001190, -0.00000682, 0.00000164, -0.00000885, 0.00000551, -0.00000072, 0.00001307, -0.00001308, -0.00000715, 0.00001273, -0.00000748, 0.00000195, 0.00000084, 0.00000113, -0.00000857, 0.00000846, 0.00000328, -0.00001574, 0.00000788, 0.00000384, 0.00000649, -0.00001155, 0.00000192, -0.00000774, 0.00001878, -0.00001753, 0.00001746, -0.00001437, 0.00001033, },
		{ 0.00000555, -0.00000911, 0.00000461, -0.00002325, 0.00002135, -0.00002839, 0.00003143, -0.00000597, 0.00000669, 0.00000386, 0.00001676, -0.00001723, 0.00001564, -0.00001660, 0.00003011, -0.00003008, 0.00003089, -0.00001413, 0.00001971, -0.00001475, -0.00000596, 0.00000121, -0.00000426, -0.00000405, -0.00000416, -0.00000556, 0.00000429, -0.00000015, 0.00000444, -0.00000894, 0.00000206, -0.00001725, 0.00001025, -0.00001272, 0.00001393, -0.00002187, 0.00001089, -0.00001532, 0.00001723, -0.00003101, 0.00002523, -0.00004162, 0.00001119, -0.00004418, 0.00002133, -0.00009072, 0.00003984, -0.00037404, 0.00015991, 0.00048602, 0.00042103, -0.00023977, -0.00016279, 0.00000857, -0.00007496, 0.00002705, -0.00004009, 0.00001248, -0.00002670, 0.00001754, -0.00003651, 0.00002147, -0.00002427, 0.00001949, -0.00001856, 0.00000815, -0.00001647, 0.00001532, -0.00001098, 0.00001556, -0.00000859, 0.00000305, -0.00001023, 0.00000675, -0.00001455, 0.00002321, -0.00001356, 0.00001546, -0.00001205, -0.00000317, 0.00000488, -0.00000672, -0.00000393, 0.00000256, 0.00000553, -0.00001057, -0.00001091, 0.00001166, -0.00000023, -0.00000295, -0.00000591, -0.00000490, 0.00000136, -0.00000119, -0.00001259, 0.00002284, -0.00002804, 0.00002125, -0.00001001, 0.00000123, 0.00000323, -0.00000269, -0.00000255, -0.00000376, 0.00000240, 0.00000387, 0.00001005, -0.00001914, 0.00000212, 0.00000869, -0.00000835, 0.00000471, -0.00000058, 0.00000016, -0.00000891, 0.00001455, -0.00000717, -0.00000769, 0.00000489, 0.00000837, -0.00000142, -0.00000609, -0.00000489, 0.00000497, 0.00000638, -0.00000849, 0.00001118, -0.00001062, },
		{ 0.00000144, -0.00000753, 0.00000595, -0.00001577, 0.00000036, 0.00000303, -0.00001394, 0.00002783, -0.00000985, 0.00000756, 0.00001053, 0.00000443, -0.00000850, 0.00000632, -0.00000379, 0.00001753, -0.00002466, 0.00003250, -0.00001806, 0.00002346, -0.00002437, 0.00000454, -0.00000445, -0.00000178, -0.00000572, -0.00000450, -0.00000140, 0.00000066, 0.00000542, -0.00000405, -0.00000112, -0.00000865, -0.00000404, -0.00000205, 0.00000188, -0.00000258, -0.00000759, -0.00000283, 0.00000263, -0.00000550, -0.00000685, -0.00000014, -0.00002166, -0.00001249, -0.00001275, -0.00002396, -0.00003843, -0.00005277, -0.00035543, 0.00031543, 0.00058272, 0.00023246, -0.00037203, -0.00000427, -0.00007636, -0.00000136, -0.00002272, 0.00000124, -0.00002388, 0.00000967, -0.00001909, -0.00000348, -0.00000342, -0.00000237, 0.00000093, -0.00000283, -0.00000835, 0.00000065, 0.00000197, 0.00000096, 0.00000545, -0.00000112, -0.00000735, 0.00000202, -0.00000613, -0.00000017, 0.00001341, -0.00000843, 0.00001128, -0.00001183, -0.00000229, 0.00000760, -0.00001421, 0.00000483, -0.00000137, 0.00000857, -0.00002197, 0.00000315, 0.00000615, 0.00000163, -0.00000906, 0.00000119, -0.00001069, 0.00001042, -0.00001686, 0.00000885, 0.00000318, -0.00001630, 0.00002103, -0.00001732, 0.00001031, -0.00000454, 0.00000167, -0.00000622, 0.00000117, -0.00000161, 0.00001119, 0.00000073, -0.00001827, 0.00000859, 0.00000361, -0.00000726, 0.00000668, -0.00000255, -0.00000099, -0.00000651, 0.00001646, -0.00001491, 0.00000030, 0.00000135, 0.00001248, -0.00000932, -0.00000074, -0.00000962, 0.00001650, -0.00000879, 0.00000659, -0.00000308, },
		{ -0.00001130, 0.00000789, -0.00001417, 0.00001380, -0.00002991, 0.00002374, -0.00002214, 0.00001264, 0.00000885, -0.00000169, 0.00000318, 0.00001883, -0.00001108, 0.00000506, -0.00000698, 0.00001433, -0.00000520, -0.00000511, 0.00002035, -0.00001023, 0.00001554, -0.00002277, 0.00000748, -0.00000608, -0.00000192, -0.00000629, -0.00000349, 0.00000036, -0.00000040, 0.00000755, -0.00000926, 0.00000306, -0.00001427, 0.00000413, -0.00000960, 0.00001182, -0.00001618, 0.00000502, -0.00001346, 0.00001561, -0.00002343, 0.00001427, -0.00002591, -0.00000043, -0.00003394, 0.00001316, -0.00006442, 0.00001167, -0.00017198, -0.00027552, 0.00048998, 0.00062704, -0.00007767, -0.00030289, 0.00003928, -0.00010129, 0.00003164, -0.00004460, 0.00001839, -0.00003786, 0.00002512, -0.00003841, 0.00001815, -0.00002073, 0.00001300, -0.00001168, 0.00000653, -0.00001845, 0.00001336, -0.00000859, 0.00001026, -0.00000181, 0.00000169, -0.00001125, 0.00000804, -0.00001375, 0.00001116, 0.00000427, -0.00000440, 0.00000921, -0.00001456, 0.00000399, 0.00000250, -0.00001410, 0.00000902, -0.00000205, 0.00000510, -0.00002574, 0.00001458, 0.00000001, 0.00000130, -0.00001019, 0.00000370, -0.00001132, 0.00001149, -0.00002335, 0.00002529, -0.00001989, 0.00000757, 0.00000392, -0.00000877, 0.00000726, -0.00000414, 0.00000027, -0.00000411, 0.00000016, 0.00000026, 0.00001369, -0.00000833, -0.00001382, 0.00001293, -0.00000292, -0.00000262, 0.00000519, -0.00000263, -0.00000366, -0.00000043, 0.00001213, -0.00001612, 0.00000338, 0.00000222, 0.00001105, -0.00001190, -0.00000076, -0.00000563, 0.00001686, -0.00001392, 0.00001374, },
		{ 0.00000767, -0.00000695, 0.00000161, -0.00000619, 0.00000368, -0.00002229, 0.00002644, -0.00003197, 0.00003064, -0.00000928, 0.00000828, 0.00000059, 0.00002084, -0.00002013, 0.00001357, -0.00001371, 0.00002493, -0.00002400, 0.00001628, 0.00000377, 0.00000156, 0.00000296, -0.00001488, 0.00000537, -0.00000506, -0.00000386, -0.00000568, -0.00000136, -0.00000094, 0.00000273, 0.00000381, -0.00000815, 0.00000054, -0.00001118, 0.00000216, -0.00000568, 0.00000846, -0.00001667, 0.00000569, -0.00001103, 0.00001254, -0.00002299, 0.00001581, -0.00003385, 0.00000457, -0.00003604, 0.00001355, -0.00007419, 0.00002284, -0.00027425, -0.00015897, 0.00071963, 0.00048893, -0.00033216, -0.00013201, -0.00001465, -0.00005844, 0.00001170, -0.00002648, 0.00000123, -0.00002006, 0.00000930, -0.00002782, 0.00001463, -0.00001792, 0.00001124, -0.00000927, 0.00000124, -0.00001365, 0.00001309, -0.00000901, 0.00001134, -0.00000253, -0.00000247, -0.00000621, 0.00000444, -0.00001191, 0.00001500, -0.00000092, -0.00000295, 0.00000989, -0.00002044, 0.00001578, -0.00001179, -0.00000257, 0.00000361, 0.00000433, -0.00000799, -0.00001635, 0.00001600, -0.00000141, -0.00000385, -0.00000377, -0.00000184, -0.00000349, 0.00000045, -0.00001241, 0.00002217, -0.00002684, 0.00002332, -0.00001374, 0.00000516, -0.00000231, 0.00000162, -0.00000478, 0.00000164, -0.00000539, 0.00000830, 0.00000853, -0.00001277, -0.00000873, 0.00001531, -0.00000967, 0.00000474, -0.00000004, -0.00000015, -0.00000745, 0.00000770, 0.00000329, -0.00001165, 0.00000168, 0.00000745, 0.00000403, -0.00000922, -0.00000487, 0.00000520, 0.00000608, -0.00000607, },
		{ 0.00000509, -0.00000326, 0.00000244, -0.00000963, 0.00000814, -0.00001658, 0.00000274, 0.00000634, -0.00001821, 0.00002785, -0.00001398, 0.00001127, 0.00000442, 0.00001295, -0.00001821, 0.00001258, -0.00000955, 0.00002146, -0.00002917, 0.00002922, -0.00000883, 0.00001023, -0.00000748, -0.00000643, 0.00000182, -0.00000341, -0.00000674, -0.00000283, -0.00000111, -0.00000141, 0.00000568, -0.00000129, -0.00000534, -0.00000301, -0.00000624, -0.00000264, 0.00000237, -0.00000164, -0.00000935, -0.00000029, -0.00000212, 0.00000095, -0.00001140, 0.00000320, -0.00002774, -0.00000178, -0.00002677, -0.00000153, -0.00006361, -0.00000069, -0.00035924, 0.00003035, 0.00093971, 0.00014084, -0.00038086, -0.00001363, -0.00006932, -0.00000918, -0.00001735, -0.00000261, -0.00002048, 0.00000270, -0.00001409, -0.00000633, 0.00000010, -0.00000633, 0.00000260, -0.00000176, -0.00000914, -0.00000198, 0.00000582, -0.00000418, 0.00000912, -0.00000262, -0.00000616, -0.00000019, -0.00000153, -0.00000654, 0.00001596, -0.00000694, 0.00000239, 0.00000520, -0.00001991, 0.00002221, -0.00002443, 0.00001166, -0.00000520, 0.00001145, -0.00002328, 0.00000080, 0.00000930, 0.00000025, -0.00000969, 0.00000485, -0.00001012, 0.00000634, -0.00001455, 0.00000889, 0.00000126, -0.00001182, 0.00001845, -0.00001682, 0.00001054, -0.00000730, 0.00000438, -0.00000634, 0.00000376, -0.00000800, 0.00001577, -0.00000011, -0.00001357, -0.00000274, 0.00001377, -0.00001288, 0.00001058, -0.00000566, 0.00000248, -0.00000967, 0.00001394, -0.00000528, -0.00000631, -0.00000014, 0.00001209, -0.00000354, -0.00000636, -0.00000621, 0.00001325, -0.00000529, },
		{ -0.00000668, 0.00000755, -0.00000664, 0.00000410, -0.00001176, 0.00001219, -0.00002710, 0.00002399, -0.00001841, 0.00000671, 0.00001017, -0.00000622, 0.00000774, 0.00001162, -0.00000030, -0.00000926, 0.00000620, 0.00000032, 0.00000806, -0.00002128, 0.00003151, -0.00001533, 0.00001446, -0.00001402, 0.00000050, -0.00000186, -0.00000188, -0.00000994, 0.00000264, -0.00000511, 0.00000295, 0.00000328, -0.00000188, -0.00000686, -0.00000061, -0.00000827, 0.00000038, 0.00000219, -0.00000528, -0.00000695, -0.00000026, -0.00000090, -0.00000282, -0.00000677, -0.00000531, -0.00002441, -0.00000245, -0.00002634, -0.00000769, -0.00006166, -0.00003241, -0.00044696, 0.00036484, 0.00098190, -0.00027286, -0.00022974, -0.00001978, -0.00005633, -0.00000685, -0.00001310, -0.00000773, -0.00001575, 0.00000115, -0.00001677, 0.00000031, -0.00000386, -0.00000392, 0.00000283, -0.00000392, -0.00001006, 0.00000315, 0.00000246, -0.00000258, 0.00001038, -0.00000836, -0.00000215, -0.00000134, -0.00000185, -0.00000452, 0.00001882, -0.00001729, 0.00001584, -0.00001076, -0.00000495, 0.00001250, -0.00002242, 0.00001576, -0.00000665, 0.00000900, -0.00002860, 0.00001388, 0.00000257, -0.00000085, -0.00000924, 0.00000710, -0.00001215, 0.00000769, -0.00001918, 0.00002263, -0.00001942, 0.00001164, -0.00000081, -0.00000502, 0.00000360, -0.00000308, 0.00000016, -0.00000101, -0.00000185, -0.00000206, 0.00001566, -0.00000592, -0.00001405, 0.00000535, 0.00000664, -0.00000930, 0.00001048, -0.00000738, 0.00000211, -0.00000768, 0.00001510, -0.00000974, -0.00000406, 0.00000198, 0.00001164, -0.00000718, -0.00000709, -0.00000020, 0.00001089, },
		{ 0.00000111, 0.00000130, -0.00000005, 0.00000000, -0.00000491, -0.00000076, -0.00000076, -0.00001602, 0.00002365, -0.00002697, 0.00002276, -0.00000653, 0.00000407, 0.00000410, 0.00001552, -0.00001128, -0.00000018, 0.00000096, 0.00000777, -0.00000629, -0.00000677, 0.00002586, -0.00001698, 0.00001571, -0.00001756, 0.00000574, -0.00000509, -0.00000150, -0.00001103, 0.00000733, -0.00001023, 0.00000976, -0.00000292, 0.00000083, -0.00001080, 0.00000535, -0.00001542, 0.00001047, -0.00000726, 0.00000006, -0.00001190, 0.00000739, -0.00000911, 0.00000424, -0.00001401, -0.00000397, -0.00002776, 0.00000383, -0.00003518, -0.00000436, -0.00007012, -0.00007933, -0.00050231, 0.00084472, 0.00069813, -0.00051777, -0.00003742, -0.00008865, -0.00000071, -0.00003780, 0.00001501, -0.00003358, 0.00000871, -0.00002049, 0.00000056, -0.00000975, 0.00000459, -0.00001138, 0.00001169, -0.00001553, -0.00000051, -0.00000028, 0.00000433, -0.00000384, 0.00001275, -0.00001592, 0.00000593, -0.00000714, 0.00000167, -0.00000372, 0.00001939, -0.00002515, 0.00002953, -0.00003052, 0.00001841, -0.00001021, -0.00000506, 0.00000665, 0.00000182, -0.00000521, -0.00001909, 0.00001595, 0.00000074, -0.00000645, -0.00000115, 0.00000063, -0.00000559, -0.00000212, -0.00000731, 0.00001645, -0.00002176, 0.00002308, -0.00001632, 0.00000844, -0.00000699, 0.00000460, -0.00000623, 0.00000637, -0.00001080, 0.00000983, 0.00000769, -0.00000686, -0.00001495, 0.00001404, -0.00000395, -0.00000004, 0.00000400, -0.00000405, -0.00000184, -0.00000172, 0.00001159, -0.00001045, -0.00000423, 0.00000706, 0.00000679, -0.00000774, -0.00000876, 0.00000903, },
		{ 0.00001199, -0.00000488, 0.00000750, -0.00000716, 0.00000637, -0.00001364, 0.00001142, -0.00001861, 0.00000747, 0.00000436, -0.00001508, 0.00002093, -0.00001207, 0.00000963, 0.00000386, 0.00001359, -0.00001680, 0.00000655, -0.00000179, 0.00001016, -0.00001658, 0.00000967, 0.00001382, -0.00001299, 0.00001299, -0.00001699, 0.00000743, -0.00000572, -0.00000480, -0.00000585, 0.00000478, -0.00000863, 0.00001054, -0.00000437, -0.00000172, -0.00000731, 0.00000300, -0.00001425, 0.00001330, -0.00001191, 0.00000134, -0.00001095, 0.00000775, -0.00001015, 0.00000458, -0.00001590, -0.00000935, -0.00001917, -0.00000636, -0.00002451, -0.00002651, -0.00004816, -0.00020290, -0.00039463, 0.00130454, 0.00012193, -0.00045306, 0.00002511, -0.00010324, 0.00001814, -0.00004244, 0.00001898, -0.00003827, 0.00001528, -0.00002851, 0.00000931, -0.00001341, 0.00000592, -0.00001080, 0.00001187, -0.00002002, 0.00000518, -0.00000086, 0.00000187, 0.00000123, 0.00000618, -0.00001377, 0.00000613, -0.00000655, -0.00000066, 0.00000463, 0.00000781, -0.00001765, 0.00002775, -0.00003657, 0.00003189, -0.00002861, 0.00001352, -0.00000511, 0.00001048, -0.00002094, -0.00000260, 0.00001084, 0.00000057, -0.00001079, 0.00000731, -0.00000745, 0.00000164, -0.00001316, 0.00001106, -0.00000377, -0.00000443, 0.00001392, -0.00001547, 0.00001077, -0.00001021, 0.00000685, -0.00000692, 0.00000745, -0.00001401, 0.00001884, -0.00000171, -0.00000640, -0.00001329, 0.00001872, -0.00001219, 0.00000921, -0.00000444, 0.00000133, -0.00000669, 0.00000520, 0.00000653, -0.00001070, -0.00000286, 0.00001040, 0.00000215, -0.00000958, -0.00000509, },
		{ 0.00000543, 0.00000425, 0.00000034, 0.00000269, -0.00000362, 0.00000153, -0.00000946, 0.00000921, -0.00002193, 0.00002160, -0.00001420, 0.00000304, 0.00000877, -0.00000855, 0.00001116, 0.00000470, 0.00000892, -0.00001887, 0.00001297, -0.00000563, 0.00001079, -0.00002315, 0.00002643, -0.00000454, -0.00000104, 0.00000340, -0.00000974, 0.00000309, -0.00000189, -0.00001254, 0.00000650, -0.00000747, 0.00000378, 0.00000010, 0.00000435, -0.00001449, 0.00000898, -0.00001481, 0.00000462, -0.00000273, 0.00000099, -0.00001381, 0.00000793, -0.00001225, 0.00001095, -0.00001881, 0.00000519, -0.00003778, 0.00001730, -0.00005131, 0.00002772, -0.00010159, 0.00003522, -0.00045183, 0.00002595, 0.00145656, -0.00048754, -0.00016003, -0.00007331, -0.00001925, -0.00003659, 0.00000930, -0.00002616, 0.00000093, -0.00001804, -0.00000091, -0.00001119, 0.00000673, -0.00001413, 0.00001033, -0.00000874, -0.00000500, -0.00000446, 0.00001024, -0.00001106, 0.00001690, -0.00001303, 0.00000281, -0.00000718, 0.00000581, -0.00001291, 0.00002265, -0.00001739, 0.00000873, 0.00000361, -0.00001854, 0.00002186, -0.00002687, 0.00001787, -0.00000663, 0.00000753, -0.00002497, 0.00000824, 0.00000717, -0.00000386, -0.00000696, 0.00000754, -0.00000830, -0.00000023, -0.00001201, 0.00001814, -0.00001855, 0.00001585, -0.00000549, -0.00000201, 0.00000153, -0.00000393, 0.00000152, -0.00000001, -0.00000058, -0.00000659, 0.00001810, -0.00000587, -0.00000880, -0.00000619, 0.00001568, -0.00001294, 0.00001246, -0.00000922, 0.00000445, -0.00000915, 0.00001023, 0.00000216, -0.00001225, 0.00000217, 0.00000911, 0.00000028, -0.00001394, },
		{ -0.00001563, 0.00001534, -0.00000698, 0.00001131, -0.00000929, 0.00000776, -0.00001213, 0.00000614, -0.00000831, -0.00000485, 0.00001405, -0.00001568, 0.00001100, -0.00000081, -0.00000329, 0.00001196, 0.00000393, 0.00000417, -0.00001838, 0.00001874, -0.00001134, 0.00001129, -0.00002527, 0.00003938, -0.00002459, 0.00001609, -0.00001140, 0.00000243, -0.00000521, 0.00000329, -0.00001924, 0.00001845, -0.00002135, 0.00001925, -0.00001422, 0.00001564, -0.00002815, 0.00002704, -0.00003658, 0.00003021, -0.00002806, 0.00002300, -0.00003629, 0.00003392, -0.00004055, 0.00004172, -0.00005382, 0.00003682, -0.00007335, 0.00006023, -0.00010430, 0.00008875, -0.00018965, 0.00012257, -0.00074765, 0.00076060, 0.00108692, -0.00079483, 0.00011028, -0.00020170, 0.00008618, -0.00010978, 0.00007384, -0.00008313, 0.00005193, -0.00006407, 0.00004065, -0.00004394, 0.00003567, -0.00004090, 0.00003764, -0.00003703, 0.00001893, -0.00002081, 0.00002442, -0.00002457, 0.00003156, -0.00003214, 0.00002154, -0.00002301, 0.00001960, -0.00002423, 0.00003730, -0.00004078, 0.00003836, -0.00002922, 0.00001306, -0.00000528, -0.00000690, 0.00000709, 0.00000266, -0.00000709, -0.00001483, 0.00000887, 0.00000751, -0.00001185, 0.00000378, -0.00000032, -0.00000210, -0.00001060, 0.00000319, 0.00000612, -0.00001344, 0.00002049, -0.00001693, 0.00000956, -0.00000880, 0.00000432, -0.00000496, 0.00000797, -0.00001154, 0.00000714, 0.00000883, -0.00000402, -0.00001402, 0.00000495, 0.00000615, -0.00000615, 0.00000820, -0.00000804, 0.00000342, -0.00000811, 0.00001302, -0.00000238, -0.00001280, 0.00000802, 0.00000557, -0.00000154, },
		{ -0.00000767, -0.00000812, 0.00001432, -0.00000889, 0.00001435, -0.00001461, 0.00001292, -0.00001955, 0.00001661, -0.00002368, 0.00001700, -0.00000534, -0.00000297, 0.00000641, -0.00000343, 0.00000137, 0.00001097, 0.00000392, -0.00000309, -0.00001115, 0.00001724, -0.00001242, 0.00000676, -0.00001716, 0.00003907, -0.00003460, 0.00002751, -0.00002199, 0.00001092, -0.00000951, 0.00000180, -0.00001501, 0.00001715, -0.00002189, 0.00002180, -0.00001611, 0.00001273, -0.00002390, 0.00002483, -0.00003734, 0.00003555, -0.00003506, 0.00002666, -0.00003760, 0.00003580, -0.00004257, 0.00004427, -0.00006035, 0.00003718, -0.00006959, 0.00005437, -0.00009843, 0.00007555, -0.00018537, 0.00006200, -0.00084712, 0.00154717, 0.00026573, -0.00063639, 0.00014051, -0.00019783, 0.00009456, -0.00010815, 0.00007394, -0.00008452, 0.00005435, -0.00006813, 0.00004780, -0.00004741, 0.00003696, -0.00004083, 0.00003923, -0.00004292, 0.00002500, -0.00002110, 0.00002156, -0.00001955, 0.00002578, -0.00003067, 0.00002189, -0.00002202, 0.00001683, -0.00001724, 0.00002970, -0.00004015, 0.00004630, -0.00004496, 0.00003314, -0.00002672, 0.00001256, -0.00000432, 0.00000971, -0.00002000, -0.00000159, 0.00000635, 0.00000564, -0.00001475, 0.00001135, -0.00000666, 0.00000094, -0.00001669, 0.00001825, -0.00001354, 0.00000661, 0.00000663, -0.00001204, 0.00000912, -0.00001073, 0.00000643, -0.00000522, 0.00000882, -0.00001592, 0.00001750, -0.00000136, -0.00000110, -0.00001764, 0.00001470, -0.00000413, 0.00000305, 0.00000015, -0.00000303, -0.00000043, -0.00000465, 0.00001461, -0.00000904, -0.00000864, 0.00000979, 0.00000442, },
		{ 0.00000637, -0.00001707, 0.00000755, 0.00000068, 0.00000335, 0.00000237, -0.00000468, 0.00000216, -0.00000919, 0.00000740, -0.00001807, 0.00002132, -0.00001491, 0.00000675, 0.00000029, -0.00000446, 0.00000852, 0.00000431, 0.00000896, -0.00001613, 0.00000660, 0.00000224, -0.00000163, -0.00000832, 0.00000589, 0.00001812, -0.00002391, 0.00002190, -0.00001831, 0.00000717, -0.00000319, -0.00001092, 0.00000424, -0.00000294, -0.00000180, 0.00000305, 0.00000282, -0.00001205, 0.00000634, -0.00000789, -0.00000393, 0.00000457, -0.00000593, -0.00000592, 0.00000014, -0.00000564, 0.00000341, -0.00000640, -0.00001095, -0.00002154, 0.00000232, -0.00003340, 0.00000698, -0.00006532, -0.00001594, -0.00027142, -0.00047465, 0.00194545, -0.00062254, -0.00016368, -0.00007032, -0.00003010, -0.00002538, -0.00000364, -0.00001492, -0.00000746, -0.00001407, -0.00000753, -0.00000167, -0.00000202, -0.00000606, 0.00000105, 0.00000043, -0.00001145, -0.00000066, 0.00000570, -0.00000648, 0.00001012, -0.00000588, -0.00000241, -0.00000307, 0.00000237, -0.00000788, 0.00001196, -0.00000414, -0.00000860, 0.00002162, -0.00002924, 0.00002489, -0.00002443, 0.00001458, -0.00000260, 0.00000244, -0.00001814, 0.00000220, 0.00000944, -0.00000449, -0.00000571, 0.00000733, -0.00000318, -0.00000771, -0.00000654, 0.00001594, -0.00002016, 0.00002179, -0.00001112, 0.00000122, -0.00000024, -0.00000479, 0.00000230, 0.00000094, 0.00000114, -0.00000979, 0.00001807, -0.00000570, -0.00000223, -0.00001533, 0.00001806, -0.00000941, 0.00000868, -0.00000634, 0.00000193, -0.00000473, 0.00000023, 0.00001455, -0.00001684, 0.00000025, 0.00000628, },
		{ 0.00000206, 0.00000626, -0.00002168, 0.00002042, -0.00001422, 0.00001953, -0.00001580, 0.00001269, -0.00001705, 0.00001193, -0.00001622, 0.00000755, 0.00000261, -0.00000424, 0.00000175, 0.00000202, -0.00000825, 0.00001817, -0.00000654, 0.00001671, -0.00003025, 0.00002850, -0.00002115, 0.00001744, -0.00002896, 0.00003654, -0.00001696, 0.00000430, 0.00000036, -0.00000151, -0.00000660, 0.00000937, -0.00002703, 0.00002755, -0.00002960, 0.00002658, -0.00002459, 0.00002914, -0.00004270, 0.00004389, -0.00005121, 0.00004384, -0.00004395, 0.00004120, -0.00005543, 0.00005510, -0.00006612, 0.00007128, -0.00008354, 0.00006667, -0.00010725, 0.00010295, -0.00015564, 0.00015536, -0.00026757, 0.00023786, -0.00080060, 0.00043087, 0.00161879, -0.00110077, 0.00025646, -0.00030913, 0.00016449, -0.00017219, 0.00012284, -0.00012434, 0.00008893, -0.00010183, 0.00007402, -0.00007114, 0.00005994, -0.00006324, 0.00005633, -0.00005290, 0.00003512, -0.00003864, 0.00004094, -0.00004042, 0.00004412, -0.00004244, 0.00003247, -0.00003463, 0.00003226, -0.00003640, 0.00004367, -0.00004308, 0.00003377, -0.00001889, 0.00000517, -0.00000302, -0.00000311, 0.00000143, 0.00000935, -0.00001522, -0.00000349, -0.00000353, 0.00001658, -0.00001832, 0.00001026, -0.00000445, 0.00000593, -0.00002252, 0.00001579, -0.00000537, -0.00000465, 0.00001705, -0.00001611, 0.00000853, -0.00000769, 0.00000108, -0.00000140, 0.00000669, -0.00000845, 0.00000203, 0.00001051, -0.00000362, -0.00000789, -0.00000711, 0.00001447, -0.00000826, 0.00000841, -0.00000839, 0.00000443, -0.00000813, 0.00000679, 0.00001014, -0.00002095, 0.00000863, },
		{ 0.00001052, 0.00000268, -0.00000124, -0.00001373, 0.00001857, -0.00001526, 0.00002285, -0.00002258, 0.00001937, -0.00002528, 0.00002243, -0.00003091, 0.00003005, -0.00002040, 0.00001442, -0.00001127, 0.00000778, -0.00000990, 0.00002224, -0.00001163, 0.00001632, -0.00003188, 0.00003761, -0.00003516, 0.00002829, -0.00003795, 0.00005441, -0.00004444, 0.00003047, -0.00002045, 0.00001343, -0.00001608, 0.00001332, -0.00002906, 0.00003363, -0.00003863, 0.00003739, -0.00003432, 0.00003478, -0.00004802, 0.00005303, -0.00006521, 0.00006285, -0.00006501, 0.00006019, -0.00007335, 0.00007445, -0.00008725, 0.00009654, -0.00011732, 0.00009791, -0.00013728, 0.00013599, -0.00019717, 0.00020506, -0.00035250, 0.00033289, -0.00118798, 0.00155278, 0.00061842, -0.00093126, 0.00032094, -0.00033935, 0.00020588, -0.00020084, 0.00015006, -0.00014891, 0.00011072, -0.00012365, 0.00009967, -0.00009281, 0.00007806, -0.00007911, 0.00007298, -0.00007210, 0.00005274, -0.00004999, 0.00004924, -0.00004688, 0.00004972, -0.00005109, 0.00004181, -0.00004220, 0.00003860, -0.00004049, 0.00004867, -0.00005542, 0.00005383, -0.00004292, 0.00002819, -0.00002356, 0.00001433, -0.00000810, 0.00001418, -0.00002536, 0.00000775, -0.00000623, 0.00001628, -0.00002243, 0.00001865, -0.00001029, 0.00000655, -0.00002619, 0.00003023, -0.00002690, 0.00002003, -0.00000293, -0.00000604, 0.00000434, -0.00000718, 0.00000160, 0.00000026, 0.00000594, -0.00001208, 0.00001110, 0.00000171, 0.00000030, -0.00001424, 0.00000309, 0.00000749, -0.00000361, 0.00000418, -0.00000662, 0.00000424, -0.00001005, 0.00001484, -0.00000043, -0.00001680, },
		{ -0.00000544, 0.00000544, 0.00000800, -0.00001415, 0.00000467, 0.00000151, 0.00000129, 0.00000685, -0.00000953, 0.00000569, -0.00001118, 0.00000797, -0.00001768, 0.00002569, -0.00002264, 0.00001874, -0.00001519, 0.00000607, -0.00000088, 0.00001197, -0.00000173, -0.00000100, -0.00001103, 0.00002038, -0.00002394, 0.00001560, -0.00002011, 0.00004132, -0.00004298, 0.00003345, -0.00002195, 0.00001064, -0.00000802, -0.00000235, -0.00000640, 0.00001011, -0.00001550, 0.00001478, -0.00000998, 0.00000394, -0.00001128, 0.00001409, -0.00002637, 0.00002578, -0.00002861, 0.00001984, -0.00002735, 0.00002333, -0.00002998, 0.00003435, -0.00005731, 0.00002870, -0.00005298, 0.00003360, -0.00007508, 0.00004823, -0.00017184, 0.00003351, -0.00100993, 0.00230186, -0.00057368, -0.00029820, 0.00002391, -0.00011257, 0.00004352, -0.00006420, 0.00003575, -0.00004983, 0.00002138, -0.00004345, 0.00003405, -0.00003423, 0.00002348, -0.00002726, 0.00002561, -0.00003221, 0.00001815, -0.00001464, 0.00001377, -0.00001135, 0.00001368, -0.00001847, 0.00001200, -0.00001282, 0.00000953, -0.00000950, 0.00001611, -0.00002773, 0.00003491, -0.00003182, 0.00002050, -0.00001749, 0.00001026, 0.00000026, -0.00000045, -0.00001465, 0.00000224, 0.00000433, 0.00000085, -0.00000856, 0.00001003, -0.00000147, -0.00000962, -0.00000712, 0.00001910, -0.00002599, 0.00003030, -0.00001863, 0.00000613, -0.00000387, -0.00000289, -0.00000031, 0.00000469, -0.00000015, -0.00000812, 0.00001293, -0.00000288, 0.00000148, -0.00001723, 0.00001108, 0.00000114, 0.00000071, -0.00000066, -0.00000308, 0.00000175, -0.00000908, 0.00002138, -0.00001421, },
		{ -0.00002398, 0.00000607, -0.00000056, 0.00001130, -0.00002267, 0.00002060, -0.00001688, 0.00002223, -0.00001646, 0.00001237, -0.00001753, 0.00001431, -0.00002129, 0.00001664, -0.00000503, 0.00000134, 0.00000039, -0.00000246, -0.00000651, 0.00001772, -0.00000949, 0.00001882, -0.00002856, 0.00002420, -0.00001675, 0.00000865, -0.00001670, 0.00001945, 0.00000068, -0.00001215, 0.00001086, -0.00000264, -0.00000910, 0.00001368, -0.00002962, 0.00002932, -0.00002982, 0.00002630, -0.00002736, 0.00003308, -0.00004514, 0.00004630, -0.00005028, 0.00004267, -0.00004551, 0.00004389, -0.00005725, 0.00005733, -0.00007021, 0.00007551, -0.00008462, 0.00006559, -0.00010653, 0.00010452, -0.00015587, 0.00015527, -0.00025543, 0.00021483, -0.00064126, -0.00008731, 0.00217778, -0.00132444, 0.00032072, -0.00035177, 0.00018788, -0.00018785, 0.00013167, -0.00013163, 0.00009679, -0.00011243, 0.00008143, -0.00007449, 0.00006246, -0.00006519, 0.00005644, -0.00005315, 0.00003816, -0.00004335, 0.00004343, -0.00004164, 0.00004200, -0.00003968, 0.00003201, -0.00003498, 0.00003254, -0.00003427, 0.00003521, -0.00003218, 0.00001988, -0.00000640, 0.00000123, -0.00000743, 0.00000722, -0.00000889, 0.00001910, -0.00002569, 0.00000955, -0.00001491, 0.00002222, -0.00002148, 0.00001478, -0.00000816, 0.00001381, -0.00003216, 0.00002481, -0.00001308, 0.00000096, 0.00001507, -0.00001556, 0.00000705, -0.00000575, -0.00000238, 0.00000164, 0.00000534, -0.00000503, -0.00000150, 0.00000950, -0.00000282, -0.00000119, -0.00001559, 0.00001525, -0.00000301, 0.00000302, -0.00000418, 0.00000079, -0.00000288, -0.00000319, 0.00002129, },
		{ 0.00001164, -0.00002358, 0.00001034, -0.00000074, 0.00000600, -0.00001779, 0.00002083, -0.00001982, 0.00002846, -0.00002680, 0.00002215, -0.00002785, 0.00002615, -0.00003710, 0.00004221, -0.00003488, 0.00002995, -0.00002493, 0.00001512, -0.00001807, 0.00003109, -0.00002494, 0.00003132, -0.00004460, 0.00004824, -0.00004710, 0.00003735, -0.00004347, 0.00005351, -0.00003997, 0.00002356, -0.00001615, 0.00001764, -0.00002572, 0.00002739, -0.00004347, 0.00004859, -0.00005306, 0.00005142, -0.00005203, 0.00005572, -0.00006930, 0.00007617, -0.00008643, 0.00008407, -0.00009013, 0.00008904, -0.00010391, 0.00010770, -0.00012568, 0.00014012, -0.00016336, 0.00014715, -0.00019277, 0.00020285, -0.00027605, 0.00030442, -0.00047177, 0.00050491, -0.00131224, 0.00127885, 0.00114375, -0.00127122, 0.00050654, -0.00047821, 0.00031014, -0.00028494, 0.00021723, -0.00020572, 0.00016127, -0.00017315, 0.00014445, -0.00013106, 0.00011237, -0.00011035, 0.00009976, -0.00009601, 0.00007750, -0.00007624, 0.00007329, -0.00006925, 0.00006745, -0.00006574, 0.00005715, -0.00005799, 0.00005472, -0.00005515, 0.00005628, -0.00005795, 0.00004938, -0.00003423, 0.00002321, -0.00002517, 0.00002319, -0.00001901, 0.00002525, -0.00003728, 0.00002394, -0.00002368, 0.00002898, -0.00003116, 0.00002752, -0.00001704, 0.00001632, -0.00003812, 0.00004279, -0.00003963, 0.00003235, -0.00001256, 0.00000127, -0.00000278, -0.00000038, -0.00000640, 0.00000815, 0.00000005, -0.00000439, 0.00000237, 0.00000534, -0.00000080, -0.00000607, -0.00001033, 0.00001605, -0.00000545, 0.00000378, -0.00000611, 0.00000408, -0.00000864, 0.00000766, },
		{ 0.00001936, -0.00000443, -0.00001351, 0.00000624, 0.00000529, -0.00000654, -0.00000107, 0.00000465, -0.00000339, 0.00001308, -0.00001504, 0.00000951, -0.00001354, 0.00000978, -0.00001988, 0.00003319, -0.00003447, 0.00003452, -0.00003250, 0.00001828, -0.00001285, 0.00002351, -0.00001783, 0.00001899, -0.00003054, 0.00003841, -0.00004477, 0.00003614, -0.00003886, 0.00005380, -0.00005008, 0.00003476, -0.00002153, 0.00001578, -0.00001820, 0.00001354, -0.00002406, 0.00002922, -0.00003471, 0.00003309, -0.00003152, 0.00002973, -0.00003855, 0.00004449, -0.00005578, 0.00005454, -0.00006079, 0.00005670, -0.00006709, 0.00006635, -0.00007916, 0.00009111, -0.00011981, 0.00009751, -0.00013083, 0.00012744, -0.00018850, 0.00019625, -0.00036496, 0.00034921, -0.00142704, 0.00239825, -0.00028607, -0.00056846, 0.00019941, -0.00025335, 0.00015588, -0.00015867, 0.00011361, -0.00011582, 0.00007893, -0.00009909, 0.00008622, -0.00008059, 0.00006568, -0.00006629, 0.00005962, -0.00006170, 0.00004713, -0.00004555, 0.00004364, -0.00004079, 0.00003900, -0.00003952, 0.00003264, -0.00003383, 0.00003184, -0.00003275, 0.00003400, -0.00004040, 0.00003865, -0.00002662, 0.00001322, -0.00001265, 0.00001167, -0.00000410, 0.00000438, -0.00001958, 0.00001205, -0.00000979, 0.00001266, -0.00001621, 0.00001722, -0.00000550, -0.00000403, -0.00001433, 0.00002708, -0.00003473, 0.00004003, -0.00002745, 0.00001307, -0.00001043, 0.00000311, -0.00000740, 0.00001222, -0.00000572, -0.00000076, 0.00000292, 0.00000265, 0.00000104, -0.00001146, -0.00000257, 0.00001381, -0.00000648, 0.00000349, -0.00000633, 0.00000538, -0.00001253, },
		{ -0.00001073, 0.00002521, -0.00001881, -0.00000088, -0.00000037, 0.00001103, -0.00001693, 0.00001504, -0.00001416, 0.00001886, -0.00001144, 0.00000737, -0.00001394, 0.00001293, -0.00002205, 0.00002001, -0.00000618, -0.00000098, 0.00000749, -0.00001380, 0.00000161, 0.00001021, -0.00000381, 0.00000964, -0.00001422, 0.00000868, -0.00000248, -0.00000914, 0.00000336, -0.00000174, 0.00001751, -0.00002359, 0.00001441, -0.00000020, -0.00001027, 0.00001225, -0.00002294, 0.00002047, -0.00001935, 0.00001564, -0.00001876, 0.00002298, -0.00003120, 0.00003090, -0.00003146, 0.00002459, -0.00002974, 0.00002709, -0.00003718, 0.00003543, -0.00004697, 0.00004847, -0.00005174, 0.00002836, -0.00006498, 0.00005756, -0.00009721, 0.00008208, -0.00015167, 0.00006877, -0.00033531, -0.00065758, 0.00260554, -0.00135089, 0.00023732, -0.00028917, 0.00012858, -0.00013694, 0.00008551, -0.00009297, 0.00006455, -0.00008587, 0.00005463, -0.00004760, 0.00003800, -0.00004250, 0.00003458, -0.00003452, 0.00002396, -0.00003028, 0.00002767, -0.00002514, 0.00002349, -0.00002295, 0.00001909, -0.00002276, 0.00001972, -0.00001900, 0.00001576, -0.00001425, 0.00000431, 0.00000162, 0.00000401, -0.00001683, 0.00001908, -0.00001720, 0.00002475, -0.00003093, 0.00001624, -0.00001772, 0.00001869, -0.00001714, 0.00001357, -0.00000716, 0.00001624, -0.00003401, 0.00002569, -0.00001369, 0.00000126, 0.00001593, -0.00001680, 0.00000720, -0.00000577, -0.00000284, 0.00000102, 0.00000699, -0.00000485, -0.00000005, 0.00000348, 0.00000067, 0.00000217, -0.00001635, 0.00000694, 0.00000823, -0.00000535, 0.00000179, -0.00000432, 0.00000307, },
		{ -0.00000271, -0.00000198, 0.00002084, -0.00002423, 0.00000640, -0.00000258, 0.00000974, -0.00001632, 0.00001781, -0.00001924, 0.00002791, -0.00002465, 0.00001925, -0.00002540, 0.00002552, -0.00003894, 0.00004871, -0.00004286, 0.00003742, -0.00002920, 0.00001425, -0.00001893, 0.00003253, -0.00002972, 0.00003476, -0.00004285, 0.00004408, -0.00004438, 0.00003154, -0.00003383, 0.00003963, -0.00002718, 0.00001416, -0.00001467, 0.00002505, -0.00003616, 0.00003896, -0.00005176, 0.00005530, -0.00005858, 0.00005693, -0.00006096, 0.00006571, -0.00007706, 0.00008348, -0.00009076, 0.00008911, -0.00009923, 0.00010042, -0.00011497, 0.00011953, -0.00013970, 0.00015494, -0.00017710, 0.00016169, -0.00020935, 0.00022320, -0.00029701, 0.00032822, -0.00048903, 0.00052075, -0.00119659, 0.00078341, 0.00172089, -0.00153849, 0.00061050, -0.00054995, 0.00035684, -0.00031997, 0.00024233, -0.00022688, 0.00018075, -0.00019332, 0.00016122, -0.00014384, 0.00012364, -0.00012016, 0.00010707, -0.00010339, 0.00008797, -0.00008812, 0.00008219, -0.00007616, 0.00007005, -0.00006706, 0.00006075, -0.00006207, 0.00005809, -0.00005565, 0.00004991, -0.00004861, 0.00003755, -0.00002550, 0.00002347, -0.00003367, 0.00003793, -0.00003346, 0.00003809, -0.00004985, 0.00003972, -0.00003810, 0.00003683, -0.00003554, 0.00003344, -0.00002226, 0.00002455, -0.00004599, 0.00004972, -0.00004643, 0.00003948, -0.00001932, 0.00000739, -0.00000964, 0.00000658, -0.00001390, 0.00001477, -0.00000522, 0.00000288, -0.00000418, 0.00000597, -0.00000121, 0.00000185, -0.00001901, 0.00001635, 0.00000000, -0.00000175, -0.00000151, -0.00000030, },
		{ 0.00000374, -0.00000896, 0.00001030, 0.00000798, -0.00001958, 0.00000596, 0.00000113, 0.00000156, -0.00000523, 0.00000607, -0.00000652, 0.00001664, -0.00001720, 0.00001033, -0.00001353, 0.00001026, -0.00002147, 0.00003926, -0.00004422, 0.00004674, -0.00004436, 0.00002569, -0.00002122, 0.00003220, -0.00003105, 0.00003372, -0.00004132, 0.00004608, -0.00005421, 0.00004395, -0.00004281, 0.00005119, -0.00004477, 0.00002911, -0.00002133, 0.00002489, -0.00003314, 0.00003323, -0.00004313, 0.00004786, -0.00005280, 0.00005112, -0.00005373, 0.00005543, -0.00006407, 0.00007105, -0.00008004, 0.00007896, -0.00008985, 0.00009052, -0.00010333, 0.00010569, -0.00012375, 0.00014066, -0.00017289, 0.00015666, -0.00019859, 0.00020862, -0.00028395, 0.00031550, -0.00051046, 0.00056217, -0.00161224, 0.00218577, 0.00020679, -0.00090786, 0.00039476, -0.00040101, 0.00026815, -0.00025019, 0.00018752, -0.00017872, 0.00013445, -0.00015168, 0.00013369, -0.00012202, 0.00010325, -0.00010046, 0.00008889, -0.00008801, 0.00007479, -0.00007506, 0.00007087, -0.00006628, 0.00005935, -0.00005623, 0.00005003, -0.00005153, 0.00004937, -0.00004851, 0.00004226, -0.00004268, 0.00003373, -0.00001891, 0.00000985, -0.00001581, 0.00002255, -0.00001730, 0.00001725, -0.00003199, 0.00002888, -0.00002850, 0.00002652, -0.00002528, 0.00002646, -0.00001312, 0.00000604, -0.00002397, 0.00003524, -0.00004203, 0.00004745, -0.00003522, 0.00002046, -0.00001864, 0.00001173, -0.00001692, 0.00002136, -0.00001347, 0.00000946, -0.00000844, 0.00000809, -0.00000143, -0.00000187, -0.00001643, 0.00002183, -0.00000785, 0.00000270, -0.00000508, },
		{ -0.00000665, 0.00000509, -0.00001139, 0.00001980, -0.00000666, -0.00000998, 0.00000159, 0.00000634, -0.00000666, 0.00000659, -0.00000876, 0.00001236, -0.00000416, 0.00000101, -0.00000897, 0.00000966, -0.00001967, 0.00001865, -0.00000254, -0.00000801, 0.00001829, -0.00002660, 0.00001120, -0.00000001, 0.00000603, -0.00000505, 0.00000474, -0.00000854, 0.00001150, -0.00002511, 0.00001967, -0.00001576, 0.00002431, -0.00002455, 0.00001052, 0.00000281, -0.00000647, 0.00000252, -0.00000637, 0.00000225, -0.00000049, -0.00000308, -0.00000081, 0.00000139, -0.00000481, 0.00000310, -0.00000129, -0.00000382, -0.00000256, -0.00000338, -0.00000167, -0.00000334, -0.00000471, 0.00000033, 0.00000271, -0.00003092, 0.00000179, -0.00001964, -0.00000295, -0.00003256, -0.00000163, -0.00012874, -0.00000298, -0.00111849, 0.00276597, -0.00111849, -0.00000298, -0.00012874, -0.00000163, -0.00003256, -0.00000295, -0.00001964, 0.00000179, -0.00003092, 0.00000271, 0.00000033, -0.00000471, -0.00000334, -0.00000167, -0.00000338, -0.00000256, -0.00000382, -0.00000129, 0.00000310, -0.00000481, 0.00000139, -0.00000081, -0.00000308, -0.00000049, 0.00000225, -0.00000637, 0.00000252, -0.00000647, 0.00000281, 0.00001052, -0.00002455, 0.00002431, -0.00001576, 0.00001967, -0.00002511, 0.00001150, -0.00000854, 0.00000474, -0.00000505, 0.00000603, -0.00000001, 0.00001120, -0.00002660, 0.00001829, -0.00000801, -0.00000254, 0.00001865, -0.00001967, 0.00000966, -0.00000897, 0.00000101, -0.00000416, 0.00001236, -0.00000876, 0.00000659, -0.00000666, 0.00000634, 0.00000159, -0.00000998, -0.00000666, 0.00001980, -0.00001139, 0.00000509, },
		{ 0.00000374, -0.00000508, 0.00000270, -0.00000785, 0.00002183, -0.00001643, -0.00000187, -0.00000143, 0.00000809, -0.00000844, 0.00000946, -0.00001347, 0.00002136, -0.00001692, 0.00001173, -0.00001864, 0.00002046, -0.00003522, 0.00004745, -0.00004203, 0.00003524, -0.00002397, 0.00000604, -0.00001312, 0.00002646, -0.00002528, 0.00002652, -0.00002850, 0.00002888, -0.00003199, 0.00001725, -0.00001730, 0.00002255, -0.00001581, 0.00000985, -0.00001891, 0.00003373, -0.00004268, 0.00004226, -0.00004851, 0.00004937, -0.00005153, 0.00005003, -0.00005623, 0.00005935, -0.00006628, 0.00007087, -0.00007506, 0.00007479, -0.00008801, 0.00008889, -0.00010046, 0.00010325, -0.00012202, 0.00013369, -0.00015168, 0.00013445, -0.00017872, 0.00018752, -0.00025019, 0.00026815, -0.00040101, 0.00039476, -0.00090786, 0.00020679, 0.00218577, -0.00161224, 0.00056217, -0.00051046, 0.00031550, -0.00028395, 0.00020862, -0.00019859, 0.00015666, -0.00017289, 0.00014066, -0.00012375, 0.00010569, -0.00010333, 0.00009052, -0.00008985, 0.00007896, -0.00008004, 0.00007105, -0.00006407, 0.00005543, -0.00005373, 0.00005112, -0.00005280, 0.00004786, -0.00004313, 0.00003323, -0.00003314, 0.00002489, -0.00002133, 0.00002911, -0.00004477, 0.00005119, -0.00004281, 0.00004395, -0.00005421, 0.00004608, -0.00004132, 0.00003372, -0.00003105, 0.00003220, -0.00002122, 0.00002569, -0.00004436, 0.00004674, -0.00004422, 0.00003926, -0.00002147, 0.00001026, -0.00001353, 0.00001033, -0.00001720, 0.00001664, -0.00000652, 0.00000607, -0.00000523, 0.00000156, 0.00000113, 0.00000596, -0.00001958, 0.00000798, 0.00001030, -0.00000896, },
		{ -0.00000271, -0.00000030, -0.00000151, -0.00000175, 0.00000000, 0.00001635, -0.00001901, 0.00000185, -0.00000121, 0.00000597, -0.00000418, 0.00000288, -0.00000522, 0.00001477, -0.00001390, 0.00000658, -0.00000964, 0.00000739, -0.00001932, 0.00003948, -0.00004643, 0.00004972, -0.00004599, 0.00002455, -0.00002226, 0.00003344, -0.00003554, 0.00003683, -0.00003810, 0.00003972, -0.00004985, 0.00003809, -0.00003346, 0.00003793, -0.00003367, 0.00002347, -0.00002550, 0.00003755, -0.00004861, 0.00004991, -0.00005565, 0.00005809, -0.00006207, 0.00006075, -0.00006706, 0.00007005, -0.00007616, 0.00008219, -0.00008812, 0.00008797, -0.00010339, 0.00010707, -0.00012016, 0.00012364, -0.00014384, 0.00016122, -0.00019332, 0.00018075, -0.00022688, 0.00024233, -0.00031997, 0.00035684, -0.00054995, 0.00061050, -0.00153849, 0.00172089, 0.00078341, -0.00119659, 0.00052075, -0.00048903, 0.00032822, -0.00029701, 0.00022320, -0.00020935, 0.00016169, -0.00017710, 0.00015494, -0.00013970, 0.00011953, -0.00011497, 0.00010042, -0.00009923, 0.00008911, -0.00009076, 0.00008348, -0.00007706, 0.00006571, -0.00006096, 0.00005693, -0.00005858, 0.00005530, -0.00005176, 0.00003896, -0.00003616, 0.00002505, -0.00001467, 0.00001416, -0.00002718, 0.00003963, -0.00003383, 0.00003154, -0.00004438, 0.00004408, -0.00004285, 0.00003476, -0.00002972, 0.00003253, -0.00001893, 0.00001425, -0.00002920, 0.00003742, -0.00004286, 0.00004871, -0.00003894, 0.00002552, -0.00002540, 0.00001925, -0.00002465, 0.00002791, -0.00001924, 0.00001781, -0.00001632, 0.00000974, -0.00000258, 0.00000640, -0.00002423, 0.00002084, -0.00000198, },
		{ -0.00001073, 0.00000307, -0.00000432, 0.00000179, -0.00000535, 0.00000823, 0.00000694, -0.00001635, 0.00000217, 0.00000067, 0.00000348, -0.00000005, -0.00000485, 0.00000699, 0.00000102, -0.00000284, -0.00000577, 0.00000720, -0.00001680, 0.00001593, 0.00000126, -0.00001369, 0.00002569, -0.00003401, 0.00001624, -0.00000716, 0.00001357, -0.00001714, 0.00001869, -0.00001772, 0.00001624, -0.00003093, 0.00002475, -0.00001720, 0.00001908, -0.00001683, 0.00000401, 0.00000162, 0.00000431, -0.00001425, 0.00001576, -0.00001900, 0.00001972, -0.00002276, 0.00001909, -0.00002295, 0.00002349, -0.00002514, 0.00002767, -0.00003028, 0.00002396, -0.00003452, 0.00003458, -0.00004250, 0.00003800, -0.00004760, 0.00005463, -0.00008587, 0.00006455, -0.00009297, 0.00008551, -0.00013694, 0.00012858, -0.00028917, 0.00023732, -0.00135089, 0.00260554, -0.00065758, -0.00033531, 0.00006877, -0.00015167, 0.00008208, -0.00009721, 0.00005756, -0.00006498, 0.00002836, -0.00005174, 0.00004847, -0.00004697, 0.00003543, -0.00003718, 0.00002709, -0.00002974, 0.00002459, -0.00003146, 0.00003090, -0.00003120, 0.00002298, -0.00001876, 0.00001564, -0.00001935, 0.00002047, -0.00002294, 0.00001225, -0.00001027, -0.00000020, 0.00001441, -0.00002359, 0.00001751, -0.00000174, 0.00000336, -0.00000914, -0.00000248, 0.00000868, -0.00001422, 0.00000964, -0.00000381, 0.00001021, 0.00000161, -0.00001380, 0.00000749, -0.00000098, -0.00000618, 0.00002001, -0.00002205, 0.00001293, -0.00001394, 0.00000737, -0.00001144, 0.00001886, -0.00001416, 0.00001504, -0.00001693, 0.00001103, -0.00000037, -0.00000088, -0.00001881, 0.00002521, },
		{ 0.00001936, -0.00001253, 0.00000538, -0.00000633, 0.00000349, -0.00000648, 0.00001381, -0.00000257, -0.00001146, 0.00000104, 0.00000265, 0.00000292, -0.00000076, -0.00000572, 0.00001222, -0.00000740, 0.00000311, -0.00001043, 0.00001307, -0.00002745, 0.00004003, -0.00003473, 0.00002708, -0.00001433, -0.00000403, -0.00000550, 0.00001722, -0.00001621, 0.00001266, -0.00000979, 0.00001205, -0.00001958, 0.00000438, -0.00000410, 0.00001167, -0.00001265, 0.00001322, -0.00002662, 0.00003865, -0.00004040, 0.00003400, -0.00003275, 0.00003184, -0.00003383, 0.00003264, -0.00003952, 0.00003900, -0.00004079, 0.00004364, -0.00004555, 0.00004713, -0.00006170, 0.00005962, -0.00006629, 0.00006568, -0.00008059, 0.00008622, -0.00009909, 0.00007893, -0.00011582, 0.00011361, -0.00015867, 0.00015588, -0.00025335, 0.00019941, -0.00056846, -0.00028607, 0.00239825, -0.00142704, 0.00034921, -0.00036496, 0.00019625, -0.00018850, 0.00012744, -0.00013083, 0.00009751, -0.00011981, 0.00009111, -0.00007916, 0.00006635, -0.00006709, 0.00005670, -0.00006079, 0.00005454, -0.00005578, 0.00004449, -0.00003855, 0.00002973, -0.00003152, 0.00003309, -0.00003471, 0.00002922, -0.00002406, 0.00001354, -0.00001820, 0.00001578, -0.00002153, 0.00003476, -0.00005008, 0.00005380, -0.00003886, 0.00003614, -0.00004477, 0.00003841, -0.00003054, 0.00001899, -0.00001783, 0.00002351, -0.00001285, 0.00001828, -0.00003250, 0.00003452, -0.00003447, 0.00003319, -0.00001988, 0.00000978, -0.00001354, 0.00000951, -0.00001504, 0.00001308, -0.00000339, 0.00000465, -0.00000107, -0.00000654, 0.00000529, 0.00000624, -0.00001351, -0.00000443, },
		{ 0.00001164, 0.00000766, -0.00000864, 0.00000408, -0.00000611, 0.00000378, -0.00000545, 0.00001605, -0.00001033, -0.00000607, -0.00000080, 0.00000534, 0.00000237, -0.00000439, 0.00000005, 0.00000815, -0.00000640, -0.00000038, -0.00000278, 0.00000127, -0.00001256, 0.00003235, -0.00003963, 0.00004279, -0.00003812, 0.00001632, -0.00001704, 0.00002752, -0.00003116, 0.00002898, -0.00002368, 0.00002394, -0.00003728, 0.00002525, -0.00001901, 0.00002319, -0.00002517, 0.00002321, -0.00003423, 0.00004938, -0.00005795, 0.00005628, -0.00005515, 0.00005472, -0.00005799, 0.00005715, -0.00006574, 0.00006745, -0.00006925, 0.00007329, -0.00007624, 0.00007750, -0.00009601, 0.00009976, -0.00011035, 0.00011237, -0.00013106, 0.00014445, -0.00017315, 0.00016127, -0.00020572, 0.00021723, -0.00028494, 0.00031014, -0.00047821, 0.00050654, -0.00127122, 0.00114375, 0.00127885, -0.00131224, 0.00050491, -0.00047177, 0.00030442, -0.00027605, 0.00020285, -0.00019277, 0.00014715, -0.00016336, 0.00014012, -0.00012568, 0.00010770, -0.00010391, 0.00008904, -0.00009013, 0.00008407, -0.00008643, 0.00007617, -0.00006930, 0.00005572, -0.00005203, 0.00005142, -0.00005306, 0.00004859, -0.00004347, 0.00002739, -0.00002572, 0.00001764, -0.00001615, 0.00002356, -0.00003997, 0.00005351, -0.00004347, 0.00003735, -0.00004710, 0.00004824, -0.00004460, 0.00003132, -0.00002494, 0.00003109, -0.00001807, 0.00001512, -0.00002493, 0.00002995, -0.00003488, 0.00004221, -0.00003710, 0.00002615, -0.00002785, 0.00002215, -0.00002680, 0.00002846, -0.00001982, 0.00002083, -0.00001779, 0.00000600, -0.00000074, 0.00001034, -0.00002358, },
		{ -0.00002398, 0.00002129, -0.00000319, -0.00000288, 0.00000079, -0.00000418, 0.00000302, -0.00000301, 0.00001525, -0.00001559, -0.00000119, -0.00000282, 0.00000950, -0.00000150, -0.00000503, 0.00000534, 0.00000164, -0.00000238, -0.00000575, 0.00000705, -0.00001556, 0.00001507, 0.00000096, -0.00001308, 0.00002481, -0.00003216, 0.00001381, -0.00000816, 0.00001478, -0.00002148, 0.00002222, -0.00001491, 0.00000955, -0.00002569, 0.00001910, -0.00000889, 0.00000722, -0.00000743, 0.00000123, -0.00000640, 0.00001988, -0.00003218, 0.00003521, -0.00003427, 0.00003254, -0.00003498, 0.00003201, -0.00003968, 0.00004200, -0.00004164, 0.00004343, -0.00004335, 0.00003816, -0.00005315, 0.00005644, -0.00006519, 0.00006246, -0.00007449, 0.00008143, -0.00011243, 0.00009679, -0.00013163, 0.00013167, -0.00018785, 0.00018788, -0.00035177, 0.00032072, -0.00132444, 0.00217778, -0.00008731, -0.00064126, 0.00021483, -0.00025543, 0.00015527, -0.00015587, 0.00010452, -0.00010653, 0.00006559, -0.00008462, 0.00007551, -0.00007021, 0.00005733, -0.00005725, 0.00004389, -0.00004551, 0.00004267, -0.00005028, 0.00004630, -0.00004514, 0.00003308, -0.00002736, 0.00002630, -0.00002982, 0.00002932, -0.00002962, 0.00001368, -0.00000910, -0.00000264, 0.00001086, -0.00001215, 0.00000068, 0.00001945, -0.00001670, 0.00000865, -0.00001675, 0.00002420, -0.00002856, 0.00001882, -0.00000949, 0.00001772, -0.00000651, -0.00000246, 0.00000039, 0.00000134, -0.00000503, 0.00001664, -0.00002129, 0.00001431, -0.00001753, 0.00001237, -0.00001646, 0.00002223, -0.00001688, 0.00002060, -0.00002267, 0.00001130, -0.00000056, 0.00000607, },
		{ -0.00000544, -0.00001421, 0.00002138, -0.00000908, 0.00000175, -0.00000308, -0.00000066, 0.00000071, 0.00000114, 0.00001108, -0.00001723, 0.00000148, -0.00000288, 0.00001293, -0.00000812, -0.00000015, 0.00000469, -0.00000031, -0.00000289, -0.00000387, 0.00000613, -0.00001863, 0.00003030, -0.00002599, 0.00001910, -0.00000712, -0.00000962, -0.00000147, 0.00001003, -0.00000856, 0.00000085, 0.00000433, 0.00000224, -0.00001465, -0.00000045, 0.00000026, 0.00001026, -0.00001749, 0.00002050, -0.00003182, 0.00003491, -0.00002773, 0.00001611, -0.00000950, 0.00000953, -0.00001282, 0.00001200, -0.00001847, 0.00001368, -0.00001135, 0.00001377, -0.00001464, 0.00001815, -0.00003221, 0.00002561, -0.00002726, 0.00002348, -0.00003423, 0.00003405, -0.00004345, 0.00002138, -0.00004983, 0.00003575, -0.00006420, 0.00004352, -0.00011257, 0.00002391, -0.00029820, -0.00057368, 0.00230186, -0.00100993, 0.00003351, -0.00017184, 0.00004823, -0.00007508, 0.00003360, -0.00005298, 0.00002870, -0.00005731, 0.00003435, -0.00002998, 0.00002333, -0.00002735, 0.00001984, -0.00002861, 0.00002578, -0.00002637, 0.00001409, -0.00001128, 0.00000394, -0.00000998, 0.00001478, -0.00001550, 0.00001011, -0.00000640, -0.00000235, -0.00000802, 0.00001064, -0.00002195, 0.00003345, -0.00004298, 0.00004132, -0.00002011, 0.00001560, -0.00002394, 0.00002038, -0.00001103, -0.00000100, -0.00000173, 0.00001197, -0.00000088, 0.00000607, -0.00001519, 0.00001874, -0.00002264, 0.00002569, -0.00001768, 0.00000797, -0.00001118, 0.00000569, -0.00000953, 0.00000685, 0.00000129, 0.00000151, 0.00000467, -0.00001415, 0.00000800, 0.00000544, },
		{ 0.00001052, -0.00001680, -0.00000043, 0.00001484, -0.00001005, 0.00000424, -0.00000662, 0.00000418, -0.00000361, 0.00000749, 0.00000309, -0.00001424, 0.00000030, 0.00000171, 0.00001110, -0.00001208, 0.00000594, 0.00000026, 0.00000160, -0.00000718, 0.00000434, -0.00000604, -0.00000293, 0.00002003, -0.00002690, 0.00003023, -0.00002619, 0.00000655, -0.00001029, 0.00001865, -0.00002243, 0.00001628, -0.00000623, 0.00000775, -0.00002536, 0.00001418, -0.00000810, 0.00001433, -0.00002356, 0.00002819, -0.00004292, 0.00005383, -0.00005542, 0.00004867, -0.00004049, 0.00003860, -0.00004220, 0.00004181, -0.00005109, 0.00004972, -0.00004688, 0.00004924, -0.00004999, 0.00005274, -0.00007210, 0.00007298, -0.00007911, 0.00007806, -0.00009281, 0.00009967, -0.00012365, 0.00011072, -0.00014891, 0.00015006, -0.00020084, 0.00020588, -0.00033935, 0.00032094, -0.00093126, 0.00061842, 0.00155278, -0.00118798, 0.00033289, -0.00035250, 0.00020506, -0.00019717, 0.00013599, -0.00013728, 0.00009791, -0.00011732, 0.00009654, -0.00008725, 0.00007445, -0.00007335, 0.00006019, -0.00006501, 0.00006285, -0.00006521, 0.00005303, -0.00004802, 0.00003478, -0.00003432, 0.00003739, -0.00003863, 0.00003363, -0.00002906, 0.00001332, -0.00001608, 0.00001343, -0.00002045, 0.00003047, -0.00004444, 0.00005441, -0.00003795, 0.00002829, -0.00003516, 0.00003761, -0.00003188, 0.00001632, -0.00001163, 0.00002224, -0.00000990, 0.00000778, -0.00001127, 0.00001442, -0.00002040, 0.00003005, -0.00003091, 0.00002243, -0.00002528, 0.00001937, -0.00002258, 0.00002285, -0.00001526, 0.00001857, -0.00001373, -0.00000124, 0.00000268, },
		{ 0.00000206, 0.00000863, -0.00002095, 0.00001014, 0.00000679, -0.00000813, 0.00000443, -0.00000839, 0.00000841, -0.00000826, 0.00001447, -0.00000711, -0.00000789, -0.00000362, 0.00001051, 0.00000203, -0.00000845, 0.00000669, -0.00000140, 0.00000108, -0.00000769, 0.00000853, -0.00001611, 0.00001705, -0.00000465, -0.00000537, 0.00001579, -0.00002252, 0.00000593, -0.00000445, 0.00001026, -0.00001832, 0.00001658, -0.00000353, -0.00000349, -0.00001522, 0.00000935, 0.00000143, -0.00000311, -0.00000302, 0.00000517, -0.00001889, 0.00003377, -0.00004308, 0.00004367, -0.00003640, 0.00003226, -0.00003463, 0.00003247, -0.00004244, 0.00004412, -0.00004042, 0.00004094, -0.00003864, 0.00003512, -0.00005290, 0.00005633, -0.00006324, 0.00005994, -0.00007114, 0.00007402, -0.00010183, 0.00008893, -0.00012434, 0.00012284, -0.00017219, 0.00016449, -0.00030913, 0.00025646, -0.00110077, 0.00161879, 0.00043087, -0.00080060, 0.00023786, -0.00026757, 0.00015536, -0.00015564, 0.00010295, -0.00010725, 0.00006667, -0.00008354, 0.00007128, -0.00006612, 0.00005510, -0.00005543, 0.00004120, -0.00004395, 0.00004384, -0.00005121, 0.00004389, -0.00004270, 0.00002914, -0.00002459, 0.00002658, -0.00002960, 0.00002755, -0.00002703, 0.00000937, -0.00000660, -0.00000151, 0.00000036, 0.00000430, -0.00001696, 0.00003654, -0.00002896, 0.00001744, -0.00002115, 0.00002850, -0.00003025, 0.00001671, -0.00000654, 0.00001817, -0.00000825, 0.00000202, 0.00000175, -0.00000424, 0.00000261, 0.00000755, -0.00001622, 0.00001193, -0.00001705, 0.00001269, -0.00001580, 0.00001953, -0.00001422, 0.00002042, -0.00002168, 0.00000626, },
		{ 0.00000637, 0.00000628, 0.00000025, -0.00001684, 0.00001455, 0.00000023, -0.00000473, 0.00000193, -0.00000634, 0.00000868, -0.00000941, 0.00001806, -0.00001533, -0.00000223, -0.00000570, 0.00001807, -0.00000979, 0.00000114, 0.00000094, 0.00000230, -0.00000479, -0.00000024, 0.00000122, -0.00001112, 0.00002179, -0.00002016, 0.00001594, -0.00000654, -0.00000771, -0.00000318, 0.00000733, -0.00000571, -0.00000449, 0.00000944, 0.00000220, -0.00001814, 0.00000244, -0.00000260, 0.00001458, -0.00002443, 0.00002489, -0.00002924, 0.00002162, -0.00000860, -0.00000414, 0.00001196, -0.00000788, 0.00000237, -0.00000307, -0.00000241, -0.00000588, 0.00001012, -0.00000648, 0.00000570, -0.00000066, -0.00001145, 0.00000043, 0.00000105, -0.00000606, -0.00000202, -0.00000167, -0.00000753, -0.00001407, -0.00000746, -0.00001492, -0.00000364, -0.00002538, -0.00003010, -0.00007032, -0.00016368, -0.00062254, 0.00194545, -0.00047465, -0.00027142, -0.00001594, -0.00006532, 0.00000698, -0.00003340, 0.00000232, -0.00002154, -0.00001095, -0.00000640, 0.00000341, -0.00000564, 0.00000014, -0.00000592, -0.00000593, 0.00000457, -0.00000393, -0.00000789, 0.00000634, -0.00001205, 0.00000282, 0.00000305, -0.00000180, -0.00000294, 0.00000424, -0.00001092, -0.00000319, 0.00000717, -0.00001831, 0.00002190, -0.00002391, 0.00001812, 0.00000589, -0.00000832, -0.00000163, 0.00000224, 0.00000660, -0.00001613, 0.00000896, 0.00000431, 0.00000852, -0.00000446, 0.00000029, 0.00000675, -0.00001491, 0.00002132, -0.00001807, 0.00000740, -0.00000919, 0.00000216, -0.00000468, 0.00000237, 0.00000335, 0.00000068, 0.00000755, -0.00001707, },
		{ -0.00000767, 0.00000442, 0.00000979, -0.00000864, -0.00000904, 0.00001461, -0.00000465, -0.00000043, -0.00000303, 0.00000015, 0.00000305, -0.00000413, 0.00001470, -0.00001764, -0.00000110, -0.00000136, 0.00001750, -0.00001592, 0.00000882, -0.00000522, 0.00000643, -0.00001073, 0.00000912, -0.00001204, 0.00000663, 0.00000661, -0.00001354, 0.00001825, -0.00001669, 0.00000094, -0.00000666, 0.00001135, -0.00001475, 0.00000564, 0.00000635, -0.00000159, -0.00002000, 0.00000971, -0.00000432, 0.00001256, -0.00002672, 0.00003314, -0.00004496, 0.00004630, -0.00004015, 0.00002970, -0.00001724, 0.00001683, -0.00002202, 0.00002189, -0.00003067, 0.00002578, -0.00001955, 0.00002156, -0.00002110, 0.00002500, -0.00004292, 0.00003923, -0.00004083, 0.00003696, -0.00004741, 0.00004780, -0.00006813, 0.00005435, -0.00008452, 0.00007394, -0.00010815, 0.00009456, -0.00019783, 0.00014051, -0.00063639, 0.00026573, 0.00154717, -0.00084712, 0.00006200, -0.00018537, 0.00007555, -0.00009843, 0.00005437, -0.00006959, 0.00003718, -0.00006035, 0.00004427, -0.00004257, 0.00003580, -0.00003760, 0.00002666, -0.00003506, 0.00003555, -0.00003734, 0.00002483, -0.00002390, 0.00001273, -0.00001611, 0.00002180, -0.00002189, 0.00001715, -0.00001501, 0.00000180, -0.00000951, 0.00001092, -0.00002199, 0.00002751, -0.00003460, 0.00003907, -0.00001716, 0.00000676, -0.00001242, 0.00001724, -0.00001115, -0.00000309, 0.00000392, 0.00001097, 0.00000137, -0.00000343, 0.00000641, -0.00000297, -0.00000534, 0.00001700, -0.00002368, 0.00001661, -0.00001955, 0.00001292, -0.00001461, 0.00001435, -0.00000889, 0.00001432, -0.00000812, },
		{ -0.00001563, -0.00000154, 0.00000557, 0.00000802, -0.00001280, -0.00000238, 0.00001302, -0.00000811, 0.00000342, -0.00000804, 0.00000820, -0.00000615, 0.00000615, 0.00000495, -0.00001402, -0.00000402, 0.00000883, 0.00000714, -0.00001154, 0.00000797, -0.00000496, 0.00000432, -0.00000880, 0.00000956, -0.00001693, 0.00002049, -0.00001344, 0.00000612, 0.00000319, -0.00001060, -0.00000210, -0.00000032, 0.00000378, -0.00001185, 0.00000751, 0.00000887, -0.00001483, -0.00000709, 0.00000266, 0.00000709, -0.00000690, -0.00000528, 0.00001306, -0.00002922, 0.00003836, -0.00004078, 0.00003730, -0.00002423, 0.00001960, -0.00002301, 0.00002154, -0.00003214, 0.00003156, -0.00002457, 0.00002442, -0.00002081, 0.00001893, -0.00003703, 0.00003764, -0.00004090, 0.00003567, -0.00004394, 0.00004065, -0.00006407, 0.00005193, -0.00008313, 0.00007384, -0.00010978, 0.00008618, -0.00020170, 0.00011028, -0.00079483, 0.00108692, 0.00076060, -0.00074765, 0.00012257, -0.00018965, 0.00008875, -0.00010430, 0.00006023, -0.00007335, 0.00003682, -0.00005382, 0.00004172, -0.00004055, 0.00003392, -0.00003629, 0.00002300, -0.00002806, 0.00003021, -0.00003658, 0.00002704, -0.00002815, 0.00001564, -0.00001422, 0.00001925, -0.00002135, 0.00001845, -0.00001924, 0.00000329, -0.00000521, 0.00000243, -0.00001140, 0.00001609, -0.00002459, 0.00003938, -0.00002527, 0.00001129, -0.00001134, 0.00001874, -0.00001838, 0.00000417, 0.00000393, 0.00001196, -0.00000329, -0.00000081, 0.00001100, -0.00001568, 0.00001405, -0.00000485, -0.00000831, 0.00000614, -0.00001213, 0.00000776, -0.00000929, 0.00001131, -0.00000698, 0.00001534, },
		{ 0.00000543, -0.00001394, 0.00000028, 0.00000911, 0.00000217, -0.00001225, 0.00000216, 0.00001023, -0.00000915, 0.00000445, -0.00000922, 0.00001246, -0.00001294, 0.00001568, -0.00000619, -0.00000880, -0.00000587, 0.00001810, -0.00000659, -0.00000058, -0.00000001, 0.00000152, -0.00000393, 0.00000153, -0.00000201, -0.00000549, 0.00001585, -0.00001855, 0.00001814, -0.00001201, -0.00000023, -0.00000830, 0.00000754, -0.00000696, -0.00000386, 0.00000717, 0.00000824, -0.00002497, 0.00000753, -0.00000663, 0.00001787, -0.00002687, 0.00002186, -0.00001854, 0.00000361, 0.00000873, -0.00001739, 0.00002265, -0.00001291, 0.00000581, -0.00000718, 0.00000281, -0.00001303, 0.00001690, -0.00001106, 0.00001024, -0.00000446, -0.00000500, -0.00000874, 0.00001033, -0.00001413, 0.00000673, -0.00001119, -0.00000091, -0.00001804, 0.00000093, -0.00002616, 0.00000930, -0.00003659, -0.00001925, -0.00007331, -0.00016003, -0.00048754, 0.00145656, 0.00002595, -0.00045183, 0.00003522, -0.00010159, 0.00002772, -0.00005131, 0.00001730, -0.00003778, 0.00000519, -0.00001881, 0.00001095, -0.00001225, 0.00000793, -0.00001381, 0.00000099, -0.00000273, 0.00000462, -0.00001481, 0.00000898, -0.00001449, 0.00000435, 0.00000010, 0.00000378, -0.00000747, 0.00000650, -0.00001254, -0.00000189, 0.00000309, -0.00000974, 0.00000340, -0.00000104, -0.00000454, 0.00002643, -0.00002315, 0.00001079, -0.00000563, 0.00001297, -0.00001887, 0.00000892, 0.00000470, 0.00001116, -0.00000855, 0.00000877, 0.00000304, -0.00001420, 0.00002160, -0.00002193, 0.00000921, -0.00000946, 0.00000153, -0.00000362, 0.00000269, 0.00000034, 0.00000425, },
		{ 0.00001199, -0.00000509, -0.00000958, 0.00000215, 0.00001040, -0.00000286, -0.00001070, 0.00000653, 0.00000520, -0.00000669, 0.00000133, -0.00000444, 0.00000921, -0.00001219, 0.00001872, -0.00001329, -0.00000640, -0.00000171, 0.00001884, -0.00001401, 0.00000745, -0.00000692, 0.00000685, -0.00001021, 0.00001077, -0.00001547, 0.00001392, -0.00000443, -0.00000377, 0.00001106, -0.00001316, 0.00000164, -0.00000745, 0.00000731, -0.00001079, 0.00000057, 0.00001084, -0.00000260, -0.00002094, 0.00001048, -0.00000511, 0.00001352, -0.00002861, 0.00003189, -0.00003657, 0.00002775, -0.00001765, 0.00000781, 0.00000463, -0.00000066, -0.00000655, 0.00000613, -0.00001377, 0.00000618, 0.00000123, 0.00000187, -0.00000086, 0.00000518, -0.00002002, 0.00001187, -0.00001080, 0.00000592, -0.00001341, 0.00000931, -0.00002851, 0.00001528, -0.00003827, 0.00001898, -0.00004244, 0.00001814, -0.00010324, 0.00002511, -0.00045306, 0.00012193, 0.00130454, -0.00039463, -0.00020290, -0.00004816, -0.00002651, -0.00002451, -0.00000636, -0.00001917, -0.00000935, -0.00001590, 0.00000458, -0.00001015, 0.00000775, -0.00001095, 0.00000134, -0.00001191, 0.00001330, -0.00001425, 0.00000300, -0.00000731, -0.00000172, -0.00000437, 0.00001054, -0.00000863, 0.00000478, -0.00000585, -0.00000480, -0.00000572, 0.00000743, -0.00001699, 0.00001299, -0.00001299, 0.00001382, 0.00000967, -0.00001658, 0.00001016, -0.00000179, 0.00000655, -0.00001680, 0.00001359, 0.00000386, 0.00000963, -0.00001207, 0.00002093, -0.00001508, 0.00000436, 0.00000747, -0.00001861, 0.00001142, -0.00001364, 0.00000637, -0.00000716, 0.00000750, -0.00000488, },
		{ 0.00000111, 0.00000903, -0.00000876, -0.00000774, 0.00000679, 0.00000706, -0.00000423, -0.00001045, 0.00001159, -0.00000172, -0.00000184, -0.00000405, 0.00000400, -0.00000004, -0.00000395, 0.00001404, -0.00001495, -0.00000686, 0.00000769, 0.00000983, -0.00001080, 0.00000637, -0.00000623, 0.00000460, -0.00000699, 0.00000844, -0.00001632, 0.00002308, -0.00002176, 0.00001645, -0.00000731, -0.00000212, -0.00000559, 0.00000063, -0.00000115, -0.00000645, 0.00000074, 0.00001595, -0.00001909, -0.00000521, 0.00000182, 0.00000665, -0.00000506, -0.00001021, 0.00001841, -0.00003052, 0.00002953, -0.00002515, 0.00001939, -0.00000372, 0.00000167, -0.00000714, 0.00000593, -0.00001592, 0.00001275, -0.00000384, 0.00000433, -0.00000028, -0.00000051, -0.00001553, 0.00001169, -0.00001138, 0.00000459, -0.00000975, 0.00000056, -0.00002049, 0.00000871, -0.00003358, 0.00001501, -0.00003780, -0.00000071, -0.00008865, -0.00003742, -0.00051777, 0.00069813, 0.00084472, -0.00050231, -0.00007933, -0.00007012, -0.00000436, -0.00003518, 0.00000383, -0.00002776, -0.00000397, -0.00001401, 0.00000424, -0.00000911, 0.00000739, -0.00001190, 0.00000006, -0.00000726, 0.00001047, -0.00001542, 0.00000535, -0.00001080, 0.00000083, -0.00000292, 0.00000976, -0.00001023, 0.00000733, -0.00001103, -0.00000150, -0.00000509, 0.00000574, -0.00001756, 0.00001571, -0.00001698, 0.00002586, -0.00000677, -0.00000629, 0.00000777, 0.00000096, -0.00000018, -0.00001128, 0.00001552, 0.00000410, 0.00000407, -0.00000653, 0.00002276, -0.00002697, 0.00002365, -0.00001602, -0.00000076, -0.00000076, -0.00000491, 0.00000000, -0.00000005, 0.00000130, },
		{ -0.00000668, 0.00001089, -0.00000020, -0.00000709, -0.00000718, 0.00001164, 0.00000198, -0.00000406, -0.00000974, 0.00001510, -0.00000768, 0.00000211, -0.00000738, 0.00001048, -0.00000930, 0.00000664, 0.00000535, -0.00001405, -0.00000592, 0.00001566, -0.00000206, -0.00000185, -0.00000101, 0.00000016, -0.00000308, 0.00000360, -0.00000502, -0.00000081, 0.00001164, -0.00001942, 0.00002263, -0.00001918, 0.00000769, -0.00001215, 0.00000710, -0.00000924, -0.00000085, 0.00000257, 0.00001388, -0.00002860, 0.00000900, -0.00000665, 0.00001576, -0.00002242, 0.00001250, -0.00000495, -0.00001076, 0.00001584, -0.00001729, 0.00001882, -0.00000452, -0.00000185, -0.00000134, -0.00000215, -0.00000836, 0.00001038, -0.00000258, 0.00000246, 0.00000315, -0.00001006, -0.00000392, 0.00000283, -0.00000392, -0.00000386, 0.00000031, -0.00001677, 0.00000115, -0.00001575, -0.00000773, -0.00001310, -0.00000685, -0.00005633, -0.00001978, -0.00022974, -0.00027286, 0.00098190, 0.00036484, -0.00044696, -0.00003241, -0.00006166, -0.00000769, -0.00002634, -0.00000245, -0.00002441, -0.00000531, -0.00000677, -0.00000282, -0.00000090, -0.00000026, -0.00000695, -0.00000528, 0.00000219, 0.00000038, -0.00000827, -0.00000061, -0.00000686, -0.00000188, 0.00000328, 0.00000295, -0.00000511, 0.00000264, -0.00000994, -0.00000188, -0.00000186, 0.00000050, -0.00001402, 0.00001446, -0.00001533, 0.00003151, -0.00002128, 0.00000806, 0.00000032, 0.00000620, -0.00000926, -0.00000030, 0.00001162, 0.00000774, -0.00000622, 0.00001017, 0.00000671, -0.00001841, 0.00002399, -0.00002710, 0.00001219, -0.00001176, 0.00000410, -0.00000664, 0.00000755, },
		{ 0.00000509, -0.00000529, 0.00001325, -0.00000621, -0.00000636, -0.00000354, 0.00001209, -0.00000014, -0.00000631, -0.00000528, 0.00001394, -0.00000967, 0.00000248, -0.00000566, 0.00001058, -0.00001288, 0.00001377, -0.00000274, -0.00001357, -0.00000011, 0.00001577, -0.00000800, 0.00000376, -0.00000634, 0.00000438, -0.00000730, 0.00001054, -0.00001682, 0.00001845, -0.00001182, 0.00000126, 0.00000889, -0.00001455, 0.00000634, -0.00001012, 0.00000485, -0.00000969, 0.00000025, 0.00000930, 0.00000080, -0.00002328, 0.00001145, -0.00000520, 0.00001166, -0.00002443, 0.00002221, -0.00001991, 0.00000520, 0.00000239, -0.00000694, 0.00001596, -0.00000654, -0.00000153, -0.00000019, -0.00000616, -0.00000262, 0.00000912, -0.00000418, 0.00000582, -0.00000198, -0.00000914, -0.00000176, 0.00000260, -0.00000633, 0.00000010, -0.00000633, -0.00001409, 0.00000270, -0.00002048, -0.00000261, -0.00001735, -0.00000918, -0.00006932, -0.00001363, -0.00038086, 0.00014084, 0.00093971, 0.00003035, -0.00035924, -0.00000069, -0.00006361, -0.00000153, -0.00002677, -0.00000178, -0.00002774, 0.00000320, -0.00001140, 0.00000095, -0.00000212, -0.00000029, -0.00000935, -0.00000164, 0.00000237, -0.00000264, -0.00000624, -0.00000301, -0.00000534, -0.00000129, 0.00000568, -0.00000141, -0.00000111, -0.00000283, -0.00000674, -0.00000341, 0.00000182, -0.00000643, -0.00000748, 0.00001023, -0.00000883, 0.00002922, -0.00002917, 0.00002146, -0.00000955, 0.00001258, -0.00001821, 0.00001295, 0.00000442, 0.00001127, -0.00001398, 0.00002785, -0.00001821, 0.00000634, 0.00000274, -0.00001658, 0.00000814, -0.00000963, 0.00000244, -0.00000326, },
		{ 0.00000767, -0.00000607, 0.00000608, 0.00000520, -0.00000487, -0.00000922, 0.00000403, 0.00000745, 0.00000168, -0.00001165, 0.00000329, 0.00000770, -0.00000745, -0.00000015, -0.00000004, 0.00000474, -0.00000967, 0.00001531, -0.00000873, -0.00001277, 0.00000853, 0.00000830, -0.00000539, 0.00000164, -0.00000478, 0.00000162, -0.00000231, 0.00000516, -0.00001374, 0.00002332, -0.00002684, 0.00002217, -0.00001241, 0.00000045, -0.00000349, -0.00000184, -0.00000377, -0.00000385, -0.00000141, 0.00001600, -0.00001635, -0.00000799, 0.00000433, 0.00000361, -0.00000257, -0.00001179, 0.00001578, -0.00002044, 0.00000989, -0.00000295, -0.00000092, 0.00001500, -0.00001191, 0.00000444, -0.00000621, -0.00000247, -0.00000253, 0.00001134, -0.00000901, 0.00001309, -0.00001365, 0.00000124, -0.00000927, 0.00001124, -0.00001792, 0.00001463, -0.00002782, 0.00000930, -0.00002006, 0.00000123, -0.00002648, 0.00001170, -0.00005844, -0.00001465, -0.00013201, -0.00033216, 0.00048893, 0.00071963, -0.00015897, -0.00027425, 0.00002284, -0.00007419, 0.00001355, -0.00003604, 0.00000457, -0.00003385, 0.00001581, -0.00002299, 0.00001254, -0.00001103, 0.00000569, -0.00001667, 0.00000846, -0.00000568, 0.00000216, -0.00001118, 0.00000054, -0.00000815, 0.00000381, 0.00000273, -0.00000094, -0.00000136, -0.00000568, -0.00000386, -0.00000506, 0.00000537, -0.00001488, 0.00000296, 0.00000156, 0.00000377, 0.00001628, -0.00002400, 0.00002493, -0.00001371, 0.00001357, -0.00002013, 0.00002084, 0.00000059, 0.00000828, -0.00000928, 0.00003064, -0.00003197, 0.00002644, -0.00002229, 0.00000368, -0.00000619, 0.00000161, -0.00000695, },
		{ -0.00001130, 0.00001374, -0.00001392, 0.00001686, -0.00000563, -0.00000076, -0.00001190, 0.00001105, 0.00000222, 0.00000338, -0.00001612, 0.00001213, -0.00000043, -0.00000366, -0.00000263, 0.00000519, -0.00000262, -0.00000292, 0.00001293, -0.00001382, -0.00000833, 0.00001369, 0.00000026, 0.00000016, -0.00000411, 0.00000027, -0.00000414, 0.00000726, -0.00000877, 0.00000392, 0.00000757, -0.00001989, 0.00002529, -0.00002335, 0.00001149, -0.00001132, 0.00000370, -0.00001019, 0.00000130, 0.00000001, 0.00001458, -0.00002574, 0.00000510, -0.00000205, 0.00000902, -0.00001410, 0.00000250, 0.00000399, -0.00001456, 0.00000921, -0.00000440, 0.00000427, 0.00001116, -0.00001375, 0.00000804, -0.00001125, 0.00000169, -0.00000181, 0.00001026, -0.00000859, 0.00001336, -0.00001845, 0.00000653, -0.00001168, 0.00001300, -0.00002073, 0.00001815, -0.00003841, 0.00002512, -0.00003786, 0.00001839, -0.00004460, 0.00003164, -0.00010129, 0.00003928, -0.00030289, -0.00007767, 0.00062704, 0.00048998, -0.00027552, -0.00017198, 0.00001167, -0.00006442, 0.00001316, -0.00003394, -0.00000043, -0.00002591, 0.00001427, -0.00002343, 0.00001561, -0.00001346, 0.00000502, -0.00001618, 0.00001182, -0.00000960, 0.00000413, -0.00001427, 0.00000306, -0.00000926, 0.00000755, -0.00000040, 0.00000036, -0.00000349, -0.00000629, -0.00000192, -0.00000608, 0.00000748, -0.00002277, 0.00001554, -0.00001023, 0.00002035, -0.00000511, -0.00000520, 0.00001433, -0.00000698, 0.00000506, -0.00001108, 0.00001883, 0.00000318, -0.00000169, 0.00000885, 0.00001264, -0.00002214, 0.00002374, -0.00002991, 0.00001380, -0.00001417, 0.00000789, },
		{ 0.00000144, -0.00000308, 0.00000659, -0.00000879, 0.00001650, -0.00000962, -0.00000074, -0.00000932, 0.00001248, 0.00000135, 0.00000030, -0.00001491, 0.00001646, -0.00000651, -0.00000099, -0.00000255, 0.00000668, -0.00000726, 0.00000361, 0.00000859, -0.00001827, 0.00000073, 0.00001119, -0.00000161, 0.00000117, -0.00000622, 0.00000167, -0.00000454, 0.00001031, -0.00001732, 0.00002103, -0.00001630, 0.00000318, 0.00000885, -0.00001686, 0.00001042, -0.00001069, 0.00000119, -0.00000906, 0.00000163, 0.00000615, 0.00000315, -0.00002197, 0.00000857, -0.00000137, 0.00000483, -0.00001421, 0.00000760, -0.00000229, -0.00001183, 0.00001128, -0.00000843, 0.00001341, -0.00000017, -0.00000613, 0.00000202, -0.00000735, -0.00000112, 0.00000545, 0.00000096, 0.00000197, 0.00000065, -0.00000835, -0.00000283, 0.00000093, -0.00000237, -0.00000342, -0.00000348, -0.00001909, 0.00000967, -0.00002388, 0.00000124, -0.00002272, -0.00000136, -0.00007636, -0.00000427, -0.00037203, 0.00023246, 0.00058272, 0.00031543, -0.00035543, -0.00005277, -0.00003843, -0.00002396, -0.00001275, -0.00001249, -0.00002166, -0.00000014, -0.00000685, -0.00000550, 0.00000263, -0.00000283, -0.00000759, -0.00000258, 0.00000188, -0.00000205, -0.00000404, -0.00000865, -0.00000112, -0.00000405, 0.00000542, 0.00000066, -0.00000140, -0.00000450, -0.00000572, -0.00000178, -0.00000445, 0.00000454, -0.00002437, 0.00002346, -0.00001806, 0.00003250, -0.00002466, 0.00001753, -0.00000379, 0.00000632, -0.00000850, 0.00000443, 0.00001053, 0.00000756, -0.00000985, 0.00002783, -0.00001394, 0.00000303, 0.00000036, -0.00001577, 0.00000595, -0.00000753, },
		{ 0.00000555, -0.00001062, 0.00001118, -0.00000849, 0.00000638, 0.00000497, -0.00000489, -0.00000609, -0.00000142, 0.00000837, 0.00000489, -0.00000769, -0.00000717, 0.00001455, -0.00000891, 0.00000016, -0.00000058, 0.00000471, -0.00000835, 0.00000869, 0.00000212, -0.00001914, 0.00001005, 0.00000387, 0.00000240, -0.00000376, -0.00000255, -0.00000269, 0.00000323, 0.00000123, -0.00001001, 0.00002125, -0.00002804, 0.00002284, -0.00001259, -0.00000119, 0.00000136, -0.00000490, -0.00000591, -0.00000295, -0.00000023, 0.00001166, -0.00001091, -0.00001057, 0.00000553, 0.00000256, -0.00000393, -0.00000672, 0.00000488, -0.00000317, -0.00001205, 0.00001546, -0.00001356, 0.00002321, -0.00001455, 0.00000675, -0.00001023, 0.00000305, -0.00000859, 0.00001556, -0.00001098, 0.00001532, -0.00001647, 0.00000815, -0.00001856, 0.00001949, -0.00002427, 0.00002147, -0.00003651, 0.00001754, -0.00002670, 0.00001248, -0.00004009, 0.00002705, -0.00007496, 0.00000857, -0.00016279, -0.00023977, 0.00042103, 0.00048602, 0.00015991, -0.00037404, 0.00003984, -0.00009072, 0.00002133, -0.00004418, 0.00001119, -0.00004162, 0.00002523, -0.00003101, 0.00001723, -0.00001532, 0.00001089, -0.00002187, 0.00001393, -0.00001272, 0.00001025, -0.00001725, 0.00000206, -0.00000894, 0.00000444, -0.00000015, 0.00000429, -0.00000556, -0.00000416, -0.00000405, -0.00000426, 0.00000121, -0.00000596, -0.00001475, 0.00001971, -0.00001413, 0.00003089, -0.00003008, 0.00003011, -0.00001660, 0.00001564, -0.00001723, 0.00001676, 0.00000386, 0.00000669, -0.00000597, 0.00003143, -0.00002839, 0.00002135, -0.00002325, 0.00000461, -0.00000911, },
		{ -0.00001476, 0.00001033, -0.00001437, 0.00001746, -0.00001753, 0.00001878, -0.00000774, 0.00000192, -0.00001155, 0.00000649, 0.00000384, 0.00000788, -0.00001574, 0.00000328, 0.00000846, -0.00000857, 0.00000113, 0.00000084, 0.00000195, -0.00000748, 0.00001273, -0.00000715, -0.00001308, 0.00001307, -0.00000072, 0.00000551, -0.00000885, 0.00000164, -0.00000682, 0.00001190, -0.00001279, 0.00000873, 0.00000280, -0.00001804, 0.00002376, -0.00002245, 0.00000999, -0.00000601, -0.00000182, -0.00001001, 0.00000181, 0.00000062, 0.00001012, -0.00001804, -0.00000125, 0.00000363, 0.00000199, -0.00000738, -0.00000214, 0.00000361, -0.00000671, -0.00000730, 0.00001378, -0.00001167, 0.00002434, -0.00002107, 0.00001340, -0.00001735, 0.00000943, -0.00001104, 0.00001850, -0.00001426, 0.00001812, -0.00002235, 0.00001389, -0.00002263, 0.00002446, -0.00003069, 0.00002812, -0.00005035, 0.00003757, -0.00004833, 0.00003301, -0.00006244, 0.00005213, -0.00012286, 0.00006702, -0.00033383, 0.00004384, 0.00042438, 0.00043428, -0.00002202, -0.00030156, 0.00006051, -0.00010115, 0.00003601, -0.00005275, 0.00001403, -0.00003890, 0.00002799, -0.00003619, 0.00002477, -0.00002099, 0.00001251, -0.00002294, 0.00001809, -0.00001729, 0.00001389, -0.00002366, 0.00000817, -0.00001237, 0.00000896, -0.00000288, 0.00000537, -0.00000851, -0.00000376, -0.00000178, -0.00000784, 0.00000823, -0.00002059, 0.00000374, 0.00000479, 0.00000185, 0.00001444, -0.00001756, 0.00002538, -0.00001596, 0.00001407, -0.00001494, 0.00001993, 0.00000262, 0.00000018, 0.00000992, 0.00001540, -0.00002058, 0.00001734, -0.00002798, 0.00001225, },
		{ 0.00000303, -0.00000589, 0.00000167, -0.00000432, 0.00000894, -0.00001183, 0.00001828, -0.00001190, 0.00000338, -0.00001158, 0.00000936, 0.00000363, 0.00000546, -0.00001846, 0.00001140, 0.00000148, -0.00000675, 0.00000235, 0.00000065, -0.00000002, -0.00000550, 0.00001479, -0.00001696, -0.00000122, 0.00000772, 0.00000206, 0.00000195, -0.00000802, 0.00000023, -0.00000331, 0.00001097, -0.00001764, 0.00002244, -0.00001920, 0.00000442, 0.00000752, -0.00001634, 0.00001049, -0.00000700, -0.00000477, -0.00000761, 0.00000245, 0.00000458, 0.00000150, -0.00001557, 0.00000193, 0.00000522, -0.00000450, -0.00000256, -0.00000505, 0.00000834, -0.00001680, 0.00000636, 0.00000202, 0.00000128, 0.00001218, -0.00001380, 0.00000693, -0.00001198, 0.00000490, -0.00000289, 0.00000962, -0.00000460, 0.00000573, -0.00001096, 0.00000169, -0.00000739, 0.00000768, -0.00001258, 0.00000518, -0.00002924, 0.00002068, -0.00003210, 0.00001220, -0.00003647, 0.00001458, -0.00009104, 0.00001130, -0.00036913, 0.00031314, 0.00032402, 0.00041805, -0.00021205, -0.00016056, 0.00001067, -0.00006142, 0.00001154, -0.00003200, -0.00000691, -0.00001216, 0.00000531, -0.00001749, 0.00001159, -0.00000977, -0.00000108, -0.00000793, 0.00000586, -0.00000674, 0.00000321, -0.00001735, 0.00000458, -0.00000753, 0.00000615, 0.00000021, 0.00000127, -0.00000818, -0.00000361, -0.00000091, -0.00000860, 0.00001026, -0.00003029, 0.00002097, -0.00001170, 0.00002045, -0.00000700, 0.00000427, 0.00000879, -0.00000483, 0.00000430, -0.00000436, 0.00001607, 0.00000336, -0.00000400, 0.00002539, -0.00000742, -0.00000051, -0.00000369, -0.00001337, },
		{ 0.00000394, -0.00001035, 0.00000699, -0.00001095, 0.00001071, -0.00000695, 0.00000373, 0.00000645, -0.00000668, -0.00000137, -0.00000613, 0.00000746, 0.00000730, -0.00000237, -0.00001439, 0.00001468, -0.00000445, -0.00000353, 0.00000244, 0.00000036, -0.00000210, -0.00000129, 0.00001196, -0.00002176, 0.00000964, -0.00000054, 0.00000888, -0.00000660, -0.00000198, -0.00000599, 0.00000738, -0.00000166, -0.00000634, 0.00001781, -0.00002642, 0.00002038, -0.00001088, -0.00000318, 0.00000478, -0.00000519, -0.00000972, -0.00000192, 0.00000167, 0.00000690, -0.00000737, -0.00000913, 0.00000281, 0.00000522, -0.00000961, 0.00000314, -0.00000909, 0.00001284, -0.00002649, 0.00002188, -0.00001318, 0.00001846, -0.00000641, 0.00000115, -0.00000733, 0.00000131, -0.00000617, 0.00001046, -0.00000419, 0.00000942, -0.00001168, 0.00000686, -0.00001723, 0.00001497, -0.00001717, 0.00001436, -0.00002993, 0.00000989, -0.00001776, 0.00000682, -0.00003306, 0.00001758, -0.00006295, -0.00000156, -0.00015398, -0.00020128, 0.00041878, 0.00025520, 0.00035960, -0.00033900, -0.00002263, -0.00005973, -0.00000472, -0.00002719, -0.00000266, -0.00003113, 0.00001737, -0.00002316, 0.00000867, -0.00000879, 0.00000617, -0.00001741, 0.00001054, -0.00001137, 0.00000928, -0.00001404, -0.00000393, -0.00000428, 0.00000132, -0.00000015, 0.00000581, -0.00000563, -0.00000564, -0.00000290, -0.00000305, -0.00000393, 0.00000277, -0.00002724, 0.00002640, -0.00001840, 0.00002975, -0.00002004, 0.00002125, -0.00000718, 0.00000698, -0.00000514, 0.00000670, 0.00001106, 0.00000131, 0.00000186, 0.00002682, -0.00001890, 0.00001167, -0.00002178, },
		{ -0.00002207, 0.00000796, -0.00001309, 0.00001013, -0.00001403, 0.00001680, -0.00001608, 0.00001598, -0.00000618, 0.00000082, -0.00000712, -0.00000001, 0.00000542, 0.00001008, -0.00001104, -0.00000636, 0.00001361, -0.00000871, 0.00000104, 0.00000008, 0.00000179, -0.00000552, 0.00000583, 0.00000268, -0.00001720, 0.00001224, -0.00000333, 0.00001158, -0.00001261, 0.00000227, -0.00000920, 0.00001560, -0.00001566, 0.00001277, -0.00000234, -0.00001384, 0.00001839, -0.00001779, 0.00000561, 0.00000046, -0.00000638, -0.00001065, 0.00000192, 0.00000279, 0.00000358, -0.00001000, -0.00000537, 0.00000618, -0.00000139, -0.00000583, 0.00000054, -0.00000484, 0.00000719, -0.00002446, 0.00002699, -0.00001957, 0.00002729, -0.00001842, 0.00001093, -0.00001684, 0.00001070, -0.00001275, 0.00001796, -0.00001121, 0.00001465, -0.00001885, 0.00001408, -0.00002452, 0.00002441, -0.00002789, 0.00002446, -0.00004685, 0.00003311, -0.00004191, 0.00002985, -0.00005898, 0.00004634, -0.00011285, 0.00005591, -0.00031578, 0.00009126, 0.00034058, 0.00028571, 0.00020785, -0.00034808, 0.00004656, -0.00009583, 0.00002758, -0.00004818, 0.00001008, -0.00003605, 0.00002669, -0.00003461, 0.00002156, -0.00001861, 0.00001140, -0.00002152, 0.00001701, -0.00001829, 0.00001626, -0.00002469, 0.00000583, -0.00000979, 0.00000665, -0.00000328, 0.00000807, -0.00001072, -0.00000316, -0.00000161, -0.00000691, 0.00000381, -0.00001170, -0.00001202, 0.00001781, -0.00001154, 0.00002522, -0.00001848, 0.00002496, -0.00001369, 0.00001223, -0.00000835, 0.00001318, 0.00000744, -0.00000230, 0.00001543, 0.00001299, -0.00001286, 0.00000518, },
		{ -0.00001179, -0.00000807, -0.00000159, -0.00000341, 0.00000110, -0.00000427, 0.00000897, -0.00001133, 0.00001607, -0.00001074, 0.00000371, -0.00000895, 0.00000281, 0.00000676, 0.00000794, -0.00001587, 0.00000180, 0.00000962, -0.00001042, 0.00000524, -0.00000347, 0.00000414, -0.00000892, 0.00001320, -0.00000972, -0.00000528, 0.00000572, 0.00000255, 0.00000552, -0.00001076, -0.00000053, -0.00000302, 0.00001173, -0.00001730, 0.00002275, -0.00002123, 0.00000645, 0.00000335, -0.00001196, 0.00000650, -0.00000031, -0.00001137, -0.00000621, 0.00000261, 0.00000491, -0.00000331, -0.00000678, -0.00000476, 0.00001052, -0.00001165, 0.00000503, -0.00001027, 0.00000836, -0.00001000, -0.00000745, 0.00001648, -0.00001119, 0.00002131, -0.00001670, 0.00000813, -0.00001404, 0.00000888, -0.00000857, 0.00001427, -0.00000687, 0.00000684, -0.00001074, 0.00000462, -0.00001339, 0.00001351, -0.00001640, 0.00000799, -0.00003202, 0.00002279, -0.00003171, 0.00001501, -0.00004044, 0.00001742, -0.00008894, 0.00000481, -0.00033713, 0.00034001, 0.00018831, 0.00037074, -0.00000564, -0.00025269, 0.00003532, -0.00008046, 0.00002130, -0.00004112, 0.00000064, -0.00001875, 0.00001206, -0.00002344, 0.00001505, -0.00001285, 0.00000268, -0.00001076, 0.00000788, -0.00001019, 0.00000850, -0.00002235, 0.00000626, -0.00000782, 0.00000520, -0.00000042, 0.00000426, -0.00001140, -0.00000175, -0.00000114, -0.00000885, 0.00000846, -0.00002454, 0.00000637, 0.00000261, 0.00000292, 0.00001241, -0.00000685, 0.00001756, -0.00001094, 0.00001114, -0.00000602, 0.00001566, 0.00000309, -0.00000068, 0.00002509, -0.00000346, -0.00000036, },
		{ 0.00000249, -0.00002165, 0.00000411, -0.00001147, 0.00000728, -0.00000966, 0.00000864, -0.00000450, 0.00000152, 0.00000649, -0.00000712, 0.00000142, -0.00000704, 0.00000316, 0.00001029, 0.00000141, -0.00001562, 0.00000807, 0.00000357, -0.00000827, 0.00000607, -0.00000482, 0.00000426, -0.00000868, 0.00001565, -0.00001802, 0.00000599, -0.00000292, 0.00001169, -0.00000545, -0.00000381, -0.00000741, 0.00000933, -0.00000307, -0.00000318, 0.00001384, -0.00002339, 0.00001715, -0.00001026, -0.00000249, 0.00000433, -0.00000161, -0.00001530, -0.00000033, 0.00000281, 0.00000396, -0.00000750, -0.00000362, -0.00000245, 0.00000930, -0.00001610, 0.00001237, -0.00001892, 0.00002003, -0.00002800, 0.00001492, -0.00000211, 0.00000570, 0.00000600, -0.00000600, -0.00000288, -0.00000279, -0.00000056, 0.00000228, 0.00000449, 0.00000219, -0.00000551, 0.00000302, -0.00001129, 0.00000520, -0.00000576, 0.00000350, -0.00001931, -0.00000118, -0.00000655, -0.00000189, -0.00002095, 0.00000211, -0.00004525, -0.00001732, -0.00013964, -0.00017076, 0.00041429, 0.00010084, 0.00040921, -0.00019339, -0.00012242, -0.00001618, -0.00003788, -0.00000583, -0.00002070, -0.00001570, 0.00000393, -0.00001044, -0.00000304, -0.00000004, -0.00000128, -0.00000963, 0.00000393, -0.00000657, 0.00000409, -0.00000706, -0.00001156, 0.00000060, -0.00000165, 0.00000028, 0.00000528, -0.00000333, -0.00000842, -0.00000120, -0.00000253, -0.00000657, 0.00000554, -0.00002805, 0.00001790, -0.00000909, 0.00001568, 0.00000056, 0.00000552, 0.00000690, -0.00000439, 0.00000791, -0.00000182, 0.00001628, -0.00000309, 0.00000984, 0.00002149, -0.00000886, },
		{ -0.00000240, -0.00000874, -0.00001555, 0.00000315, -0.00001030, 0.00000778, -0.00001076, 0.00001266, -0.00001122, 0.00001075, -0.00000329, -0.00000154, -0.00000191, -0.00000493, 0.00000465, 0.00001182, -0.00000566, -0.00001209, 0.00001200, -0.00000344, -0.00000241, 0.00000221, -0.00000184, 0.00000011, -0.00000283, 0.00001016, -0.00001696, 0.00000900, -0.00000498, 0.00001515, -0.00001279, 0.00000040, -0.00000958, 0.00001677, -0.00001628, 0.00001523, -0.00000701, -0.00000851, 0.00001134, -0.00001249, 0.00000204, 0.00000435, -0.00000728, -0.00001353, 0.00000284, 0.00000444, -0.00000170, -0.00000533, -0.00000463, 0.00000426, -0.00000068, -0.00000860, 0.00000738, -0.00001511, 0.00001832, -0.00003301, 0.00002815, -0.00001534, 0.00001879, -0.00000701, 0.00000274, -0.00001141, 0.00000660, -0.00000813, 0.00001071, -0.00000232, 0.00000609, -0.00001067, 0.00000881, -0.00001862, 0.00001489, -0.00001556, 0.00001131, -0.00003284, 0.00001781, -0.00002535, 0.00001564, -0.00004156, 0.00002373, -0.00008380, 0.00002259, -0.00027069, 0.00009806, 0.00031212, 0.00013784, 0.00033703, -0.00028407, -0.00002714, -0.00005757, -0.00000329, -0.00002767, -0.00000664, -0.00002158, 0.00001390, -0.00002237, 0.00000978, -0.00000955, 0.00000406, -0.00001378, 0.00000999, -0.00001344, 0.00001189, -0.00001880, -0.00000157, -0.00000414, 0.00000262, -0.00000241, 0.00000838, -0.00001009, -0.00000420, -0.00000135, -0.00000478, -0.00000179, -0.00000350, -0.00002142, 0.00001891, -0.00001227, 0.00002187, -0.00000542, 0.00001292, -0.00000142, 0.00000207, 0.00000490, 0.00000280, 0.00001422, -0.00000659, 0.00002331, 0.00000773, },
		{ -0.00000335, 0.00000354, -0.00002268, -0.00000135, -0.00000696, -0.00000008, -0.00000110, -0.00000175, 0.00000557, -0.00000680, 0.00001011, -0.00000665, 0.00000098, -0.00000325, -0.00000447, 0.00000886, 0.00000893, -0.00001012, -0.00000721, 0.00001305, -0.00000892, 0.00000412, -0.00000385, 0.00000357, -0.00000625, 0.00000596, -0.00000043, -0.00000824, 0.00000344, 0.00000176, 0.00000914, -0.00001192, -0.00000242, -0.00000205, 0.00001108, -0.00001533, 0.00002149, -0.00002226, 0.00000913, -0.00000288, -0.00000555, 0.00000127, 0.00000588, -0.00001554, -0.00000666, 0.00000322, 0.00000554, -0.00000824, 0.00000018, -0.00000743, 0.00001115, -0.00001364, 0.00000618, -0.00000750, 0.00000051, 0.00000187, -0.00002065, 0.00002522, -0.00001557, 0.00002075, -0.00001081, 0.00000308, -0.00001102, 0.00000775, -0.00000818, 0.00001202, -0.00000269, 0.00000229, -0.00000576, 0.00000283, -0.00001272, 0.00001051, -0.00001068, 0.00000107, -0.00002377, 0.00001287, -0.00001982, 0.00000589, -0.00002982, 0.00000230, -0.00006582, -0.00002677, -0.00027409, 0.00031600, 0.00014458, 0.00025093, 0.00017349, -0.00027026, 0.00000757, -0.00006630, 0.00000725, -0.00003310, -0.00000489, -0.00001423, 0.00000803, -0.00001896, 0.00000973, -0.00000925, 0.00000074, -0.00000808, 0.00000495, -0.00000902, 0.00000820, -0.00002074, 0.00000235, -0.00000434, 0.00000226, -0.00000036, 0.00000600, -0.00001272, -0.00000086, -0.00000227, -0.00000581, 0.00000146, -0.00001293, -0.00001012, 0.00001278, -0.00000842, 0.00002258, -0.00000671, 0.00001560, -0.00000667, 0.00000821, 0.00000139, 0.00000880, 0.00000717, -0.00000082, 0.00002827, },
		{ 0.00001938, -0.00000253, -0.00000283, -0.00002461, 0.00000591, -0.00001306, 0.00000752, -0.00000862, 0.00000740, -0.00000392, 0.00000222, 0.00000326, -0.00000465, 0.00000080, -0.00000371, -0.00000362, 0.00001324, 0.00000326, -0.00001185, -0.00000149, 0.00001028, -0.00000994, 0.00000696, -0.00000733, 0.00000678, -0.00000995, 0.00001187, -0.00000910, 0.00000056, -0.00000361, 0.00001114, -0.00000160, -0.00000646, -0.00000800, 0.00001002, -0.00000394, -0.00000006, 0.00000946, -0.00001972, 0.00001442, -0.00001181, 0.00000115, 0.00000080, 0.00000385, -0.00002041, 0.00000051, 0.00000353, 0.00000261, -0.00000980, 0.00000277, -0.00000635, 0.00001080, -0.00001904, 0.00001610, -0.00002021, 0.00001629, -0.00001843, 0.00000058, 0.00001137, -0.00000605, 0.00001414, -0.00000785, -0.00000220, -0.00000435, 0.00000280, -0.00000286, 0.00000897, -0.00000080, -0.00000298, 0.00000157, -0.00000674, -0.00000214, 0.00000123, -0.00000198, -0.00001418, -0.00000564, -0.00000320, -0.00000327, -0.00001569, -0.00000507, -0.00003840, -0.00002003, -0.00014078, -0.00012271, 0.00037587, 0.00003900, 0.00034225, -0.00000698, -0.00019779, 0.00000028, -0.00005176, 0.00000086, -0.00002829, -0.00000755, -0.00000403, -0.00000309, -0.00000890, 0.00000352, -0.00000515, -0.00000438, -0.00000040, -0.00000350, 0.00000003, -0.00000206, -0.00001547, 0.00000164, -0.00000183, -0.00000049, 0.00000422, -0.00000057, -0.00001077, 0.00000035, -0.00000338, -0.00000516, 0.00000102, -0.00001823, -0.00000009, 0.00000475, -0.00000097, 0.00002027, -0.00000587, 0.00001562, -0.00000945, 0.00001400, -0.00000265, 0.00001526, -0.00000297, 0.00001441, },
		{ 0.00002956, 0.00000392, 0.00000574, -0.00001959, -0.00001190, -0.00000001, -0.00000784, 0.00000474, -0.00000651, 0.00000761, -0.00000616, 0.00000643, -0.00000205, -0.00000230, 0.00000094, -0.00000577, 0.00000059, 0.00001395, -0.00000174, -0.00001228, 0.00000478, 0.00000406, -0.00000585, 0.00000413, -0.00000539, 0.00000477, -0.00000791, 0.00001119, -0.00001081, 0.00000289, -0.00000446, 0.00001481, -0.00000896, -0.00000376, -0.00000800, 0.00001544, -0.00001492, 0.00001618, -0.00001072, -0.00000330, 0.00000478, -0.00000910, 0.00000155, 0.00000419, -0.00000442, -0.00001785, 0.00000420, 0.00000493, -0.00000441, -0.00000464, -0.00000014, 0.00000023, 0.00000109, -0.00001163, 0.00001271, -0.00002031, 0.00002017, -0.00002907, 0.00001741, -0.00000241, 0.00000448, 0.00000639, -0.00000497, -0.00000605, 0.00000158, -0.00000200, 0.00000237, 0.00000637, -0.00000175, -0.00000302, 0.00000282, -0.00001020, 0.00000270, -0.00000197, -0.00000220, -0.00001828, 0.00000261, -0.00001011, 0.00000217, -0.00002348, 0.00000040, -0.00005513, -0.00000967, -0.00022653, 0.00010083, 0.00028489, 0.00004863, 0.00034605, -0.00013745, -0.00012007, -0.00001800, -0.00003334, -0.00000874, -0.00002265, -0.00000652, -0.00000055, -0.00000873, -0.00000217, -0.00000078, -0.00000373, -0.00000485, 0.00000172, -0.00000692, 0.00000499, -0.00001080, -0.00000901, 0.00000039, -0.00000028, -0.00000178, 0.00000723, -0.00000765, -0.00000599, -0.00000089, -0.00000336, -0.00000463, -0.00000151, -0.00001933, 0.00000659, -0.00000252, 0.00000871, 0.00001489, -0.00000269, 0.00001265, -0.00000818, 0.00001672, -0.00000412, 0.00001761, -0.00000837, },
		{ -0.00000166, 0.00003270, -0.00000471, 0.00000778, -0.00003242, 0.00000351, -0.00001098, 0.00000295, -0.00000419, 0.00000238, 0.00000012, -0.00000038, 0.00000313, -0.00000231, -0.00000204, 0.00000206, -0.00000902, 0.00000781, 0.00001012, -0.00000448, -0.00001177, 0.00000998, -0.00000277, 0.00000037, -0.00000191, 0.00000021, -0.00000092, -0.00000180, 0.00000568, -0.00000689, -0.00000091, 0.00000190, 0.00001015, -0.00000959, -0.00000645, 0.00000052, 0.00000823, -0.00001142, 0.00001854, -0.00002193, 0.00001152, -0.00000914, -0.00000002, -0.00000194, 0.00000850, -0.00001540, -0.00000959, 0.00000476, 0.00000514, -0.00001086, 0.00000279, -0.00000483, 0.00000713, -0.00001136, 0.00000299, -0.00000101, -0.00000829, 0.00001011, -0.00002497, 0.00002261, -0.00000919, 0.00001056, 0.00000161, -0.00000542, -0.00000509, 0.00000303, -0.00000335, 0.00000525, 0.00000492, -0.00000484, 0.00000142, -0.00000204, -0.00000680, 0.00000103, 0.00000098, -0.00001115, -0.00000969, -0.00000287, -0.00000314, -0.00000851, -0.00001194, -0.00002144, -0.00003386, -0.00006762, -0.00020026, 0.00027018, 0.00013938, 0.00013629, 0.00026465, -0.00019703, -0.00006378, -0.00003079, -0.00002050, -0.00001586, -0.00001774, -0.00000289, -0.00000307, -0.00000787, -0.00000058, -0.00000198, -0.00000475, -0.00000160, -0.00000155, -0.00000416, 0.00000319, -0.00001419, -0.00000467, 0.00000047, -0.00000101, -0.00000003, 0.00000614, -0.00001182, -0.00000113, -0.00000353, -0.00000191, -0.00000579, -0.00000347, -0.00001860, 0.00001061, -0.00000822, 0.00002006, 0.00000548, 0.00000442, 0.00000527, -0.00000075, 0.00001321, 0.00000021, 0.00001194, },
		{ 0.00000117, 0.00001553, 0.00002104, -0.00000059, -0.00000466, -0.00002914, 0.00000772, -0.00001423, 0.00000772, -0.00000849, 0.00000785, -0.00000581, 0.00000576, -0.00000239, -0.00000036, -0.00000188, 0.00000151, -0.00000979, 0.00001401, 0.00000465, -0.00000623, -0.00000915, 0.00001168, -0.00000652, 0.00000414, -0.00000633, 0.00000488, -0.00000588, 0.00000372, 0.00000058, -0.00000323, -0.00000462, 0.00000946, 0.00000200, -0.00000740, -0.00000949, 0.00001090, -0.00000555, 0.00000371, 0.00000455, -0.00001554, 0.00001216, -0.00001459, 0.00000570, -0.00000287, 0.00000774, -0.00002224, -0.00000091, 0.00000474, 0.00000144, -0.00001154, 0.00000646, -0.00000580, 0.00000751, -0.00001671, 0.00001338, -0.00001397, 0.00000581, -0.00000540, -0.00001167, 0.00001835, -0.00000950, 0.00001286, -0.00000175, -0.00000663, -0.00000190, 0.00000184, -0.00000270, 0.00000748, 0.00000156, -0.00000491, 0.00000368, -0.00000594, -0.00000372, 0.00000047, 0.00000064, -0.00001684, -0.00000149, -0.00000907, 0.00000379, -0.00001914, -0.00000138, -0.00004463, -0.00000841, -0.00015713, -0.00005892, 0.00030939, 0.00004038, 0.00022670, 0.00014424, -0.00020302, -0.00002886, -0.00003812, -0.00001174, -0.00002216, -0.00000997, -0.00000300, -0.00000450, -0.00000635, 0.00000012, -0.00000374, -0.00000371, -0.00000025, -0.00000436, -0.00000037, -0.00000172, -0.00001377, -0.00000182, 0.00000068, -0.00000236, 0.00000329, 0.00000134, -0.00001152, 0.00000112, -0.00000510, -0.00000104, -0.00000735, -0.00000487, -0.00001668, 0.00001169, -0.00001003, 0.00002975, -0.00000549, 0.00001382, -0.00000462, 0.00001106, 0.00000461, 0.00000870, },
		{ 0.00001469, -0.00000528, 0.00003176, 0.00000435, 0.00000836, -0.00002490, -0.00001216, -0.00000119, -0.00000613, 0.00000206, -0.00000304, 0.00000394, -0.00000334, 0.00000506, -0.00000377, -0.00000047, 0.00000064, -0.00000283, -0.00000517, 0.00001529, 0.00000096, -0.00000864, -0.00000384, 0.00000910, -0.00000573, 0.00000318, -0.00000615, 0.00000549, -0.00000713, 0.00000576, -0.00000092, -0.00000391, -0.00000284, 0.00001171, -0.00000322, -0.00000801, -0.00000608, 0.00001303, -0.00001291, 0.00001640, -0.00001358, 0.00000125, -0.00000033, -0.00000819, 0.00000378, 0.00000144, -0.00000038, -0.00002098, 0.00000434, 0.00000502, -0.00000538, -0.00000589, 0.00000463, -0.00000209, 0.00000027, -0.00001116, 0.00001260, -0.00001739, 0.00001225, -0.00001608, 0.00000170, 0.00001081, -0.00000704, 0.00001404, -0.00000645, -0.00000496, -0.00000048, 0.00000128, -0.00000221, 0.00001019, -0.00000473, 0.00000041, -0.00000014, -0.00000411, -0.00000578, 0.00000574, -0.00000883, -0.00001060, -0.00000479, -0.00000377, -0.00000291, -0.00001368, -0.00001280, -0.00003923, -0.00002712, -0.00019837, 0.00011744, 0.00023662, 0.00002650, 0.00026902, 0.00002877, -0.00017989, -0.00000848, -0.00004284, -0.00000512, -0.00002760, 0.00000034, -0.00000846, -0.00000137, -0.00000771, 0.00000240, -0.00000742, 0.00000081, -0.00000349, -0.00000283, -0.00000012, -0.00000535, -0.00001238, 0.00000098, -0.00000050, -0.00000221, 0.00000553, -0.00000476, -0.00000742, -0.00000041, -0.00000351, -0.00000351, -0.00000609, -0.00000846, -0.00001174, 0.00000794, -0.00000417, 0.00003171, -0.00001158, 0.00001965, -0.00001084, 0.00002070, -0.00000298, },
		{ -0.00000383, 0.00001285, 0.00000040, 0.00003510, -0.00000416, 0.00000832, -0.00003717, 0.00000397, -0.00001218, 0.00000465, -0.00000689, 0.00000622, -0.00000485, 0.00000484, -0.00000177, -0.00000037, -0.00000284, 0.00000437, -0.00000906, 0.00000340, 0.00001169, -0.00000035, -0.00001132, 0.00000256, 0.00000428, -0.00000247, -0.00000072, -0.00000254, 0.00000280, -0.00000551, 0.00000559, -0.00000062, -0.00000711, 0.00000351, 0.00000795, -0.00000424, -0.00001158, 0.00000389, 0.00000396, -0.00000647, 0.00001455, -0.00002033, 0.00001290, -0.00001379, 0.00000276, -0.00000170, 0.00000705, -0.00001181, -0.00001339, 0.00000613, 0.00000388, -0.00001090, 0.00000164, 0.00000082, 0.00000167, -0.00000841, -0.00000014, 0.00000354, -0.00001174, 0.00000941, -0.00001817, 0.00001029, 0.00000359, -0.00000314, 0.00001375, -0.00001151, -0.00000075, -0.00000132, 0.00000163, -0.00000112, 0.00001088, -0.00000982, 0.00000665, -0.00000619, -0.00000011, -0.00000894, 0.00001137, -0.00002086, 0.00000181, -0.00001526, 0.00000889, -0.00001865, 0.00000237, -0.00004057, -0.00000884, -0.00009902, -0.00013788, 0.00023106, 0.00013062, 0.00007075, 0.00025845, -0.00006370, -0.00014039, -0.00000201, -0.00004223, -0.00000374, -0.00002716, 0.00000650, -0.00001330, 0.00000228, -0.00000922, 0.00000356, -0.00000937, 0.00000464, -0.00000781, 0.00000092, -0.00000290, -0.00000691, -0.00001060, 0.00000334, -0.00000267, -0.00000019, 0.00000504, -0.00000931, -0.00000222, -0.00000418, 0.00000046, -0.00000931, -0.00000101, -0.00001561, -0.00000306, -0.00000055, 0.00001028, 0.00002240, -0.00000764, 0.00001668, -0.00000777, 0.00002192, },
		{ 0.00001464, 0.00000166, 0.00000545, 0.00001619, 0.00002400, -0.00000068, -0.00000609, -0.00003181, 0.00000690, -0.00001344, 0.00000714, -0.00000867, 0.00000911, -0.00000870, 0.00000986, -0.00000771, 0.00000307, -0.00000390, 0.00000531, -0.00001230, 0.00001102, 0.00000678, -0.00000143, -0.00001211, 0.00000724, 0.00000038, -0.00000024, -0.00000385, 0.00000107, -0.00000030, -0.00000366, 0.00000623, -0.00000263, -0.00000745, 0.00000863, 0.00000309, -0.00000527, -0.00001235, 0.00001240, -0.00000817, 0.00000799, -0.00000064, -0.00001090, 0.00000976, -0.00001690, 0.00000867, -0.00000400, 0.00000778, -0.00001976, -0.00000448, 0.00000629, -0.00000023, -0.00001122, 0.00000617, -0.00000088, 0.00000073, -0.00001115, 0.00000738, -0.00000534, -0.00000411, 0.00000280, -0.00001475, 0.00001471, -0.00000337, 0.00000342, 0.00000914, -0.00001284, 0.00000209, -0.00000155, 0.00000075, 0.00000269, 0.00000620, -0.00000834, 0.00000703, -0.00000732, -0.00000108, -0.00000543, 0.00000756, -0.00002280, 0.00000614, -0.00001814, 0.00001290, -0.00002508, 0.00000629, -0.00005521, 0.00000646, -0.00017462, 0.00000285, 0.00023940, 0.00006089, 0.00012551, 0.00021469, -0.00013198, -0.00008995, -0.00001091, -0.00003261, -0.00001050, -0.00001714, 0.00000274, -0.00001088, 0.00000077, -0.00000662, 0.00000015, -0.00000570, 0.00000281, -0.00000782, 0.00000191, -0.00000494, -0.00000837, -0.00000725, 0.00000361, -0.00000399, 0.00000233, 0.00000189, -0.00001027, 0.00000071, -0.00000648, 0.00000282, -0.00001420, 0.00000410, -0.00002314, 0.00000585, -0.00000870, 0.00002760, 0.00000573, 0.00000449, 0.00000616, 0.00000380, },
		{ 0.00001656, 0.00000508, 0.00000741, 0.00000113, 0.00003095, 0.00000839, 0.00000587, -0.00002580, -0.00001442, -0.00000210, -0.00000433, -0.00000021, -0.00000092, 0.00000224, -0.00000302, 0.00000615, -0.00000721, 0.00000251, -0.00000112, 0.00000119, -0.00000946, 0.00001380, 0.00000385, -0.00000419, -0.00000964, 0.00000874, -0.00000118, -0.00000061, -0.00000419, 0.00000265, -0.00000227, -0.00000235, 0.00000805, -0.00000798, -0.00000228, 0.00000832, 0.00000136, -0.00000980, -0.00000547, 0.00001097, -0.00001147, 0.00001661, -0.00001583, 0.00000511, -0.00000419, -0.00000860, 0.00000652, -0.00000093, 0.00000160, -0.00002044, 0.00000205, 0.00000547, -0.00000592, -0.00000662, 0.00000675, -0.00000054, -0.00000432, -0.00000657, 0.00000733, -0.00000866, -0.00000011, -0.00000213, -0.00000994, 0.00001649, -0.00000979, 0.00001223, -0.00000042, -0.00000834, 0.00000090, 0.00000066, -0.00000190, 0.00000855, -0.00000280, -0.00000032, 0.00000035, -0.00000190, -0.00000831, 0.00000564, -0.00000740, -0.00001052, -0.00000401, -0.00000602, -0.00000017, -0.00001224, -0.00001540, -0.00003596, -0.00003090, -0.00018357, 0.00014361, 0.00017497, 0.00004440, 0.00016466, 0.00015230, -0.00017322, -0.00003892, -0.00002934, -0.00001726, -0.00002125, -0.00000192, -0.00000830, -0.00000184, -0.00000590, -0.00000044, -0.00000637, 0.00000207, -0.00000429, -0.00000259, -0.00000170, -0.00000422, -0.00001073, -0.00000220, 0.00000128, -0.00000327, 0.00000375, -0.00000254, -0.00000755, -0.00000045, -0.00000452, -0.00000033, -0.00001298, 0.00000338, -0.00002532, 0.00000921, -0.00000891, 0.00003788, -0.00000843, 0.00001644, -0.00000423, },
		{ -0.00000680, 0.00002281, -0.00000030, 0.00000836, 0.00000656, 0.00003428, -0.00000035, 0.00000355, -0.00003593, 0.00000011, -0.00001104, 0.00000507, -0.00000850, 0.00000839, -0.00000754, 0.00000700, -0.00000320, -0.00000152, -0.00000108, 0.00000359, -0.00000564, -0.00000216, 0.00001205, 0.00000303, -0.00000808, -0.00000490, 0.00000796, -0.00000159, -0.00000218, -0.00000276, 0.00000286, -0.00000403, 0.00000031, 0.00000775, -0.00001262, 0.00000512, 0.00000425, 0.00000155, -0.00001530, 0.00000621, -0.00000009, -0.00000203, 0.00001063, -0.00001800, 0.00001317, -0.00001644, 0.00000287, 0.00000093, 0.00000361, -0.00000778, -0.00001520, 0.00000551, 0.00000286, -0.00000969, -0.00000074, 0.00000591, -0.00000158, -0.00000805, -0.00000002, 0.00000309, -0.00000802, 0.00000075, -0.00000490, -0.00000436, 0.00001461, -0.00001246, 0.00001892, -0.00001067, -0.00000099, -0.00000249, 0.00000350, -0.00000355, 0.00001196, -0.00000997, 0.00000759, -0.00000707, 0.00000368, -0.00001444, 0.00001510, -0.00002286, 0.00000558, -0.00001904, 0.00001132, -0.00001966, 0.00000724, -0.00004787, 0.00000118, -0.00011221, -0.00009639, 0.00020970, 0.00010412, 0.00005628, 0.00018903, 0.00007361, -0.00017899, -0.00000354, -0.00004419, -0.00000575, -0.00002654, 0.00000816, -0.00001702, 0.00000617, -0.00001174, 0.00000395, -0.00001012, 0.00000752, -0.00001070, 0.00000314, -0.00000672, -0.00000253, -0.00001241, 0.00000251, -0.00000206, -0.00000106, 0.00000328, -0.00000621, -0.00000332, -0.00000419, 0.00000060, -0.00000850, -0.00000477, -0.00000547, -0.00001951, 0.00000477, 0.00000229, 0.00003458, -0.00001176, 0.00002025, },
		{ 0.00001498, -0.00000033, 0.00002070, -0.00000043, 0.00000604, 0.00001945, 0.00002551, -0.00000016, -0.00001007, -0.00003005, 0.00000210, -0.00001020, 0.00000537, -0.00000834, 0.00000967, -0.00001047, 0.00001196, -0.00001018, 0.00000361, -0.00000336, 0.00000582, -0.00001025, 0.00000494, 0.00000908, 0.00000204, -0.00001083, -0.00000006, 0.00000669, -0.00000269, -0.00000273, -0.00000158, 0.00000306, -0.00000657, 0.00000584, 0.00000217, -0.00001144, 0.00000877, 0.00000163, -0.00000107, -0.00001516, 0.00001342, -0.00001082, 0.00001172, -0.00000527, -0.00000618, 0.00000691, -0.00001765, 0.00000890, -0.00000200, 0.00000434, -0.00001460, -0.00000804, 0.00000668, -0.00000172, -0.00000935, 0.00000339, 0.00000543, -0.00000588, -0.00000624, 0.00000255, 0.00000031, -0.00000797, 0.00000221, -0.00000802, 0.00000333, 0.00000727, -0.00000783, 0.00001819, -0.00001548, 0.00000371, -0.00000396, 0.00000375, -0.00000119, 0.00000895, -0.00000958, 0.00000843, -0.00000793, 0.00000201, -0.00001122, 0.00001256, -0.00002558, 0.00001042, -0.00002308, 0.00001669, -0.00002616, 0.00000956, -0.00005994, 0.00001263, -0.00017940, 0.00004642, 0.00018647, 0.00006958, 0.00006982, 0.00020398, -0.00001735, -0.00014713, 0.00000532, -0.00004527, -0.00000487, -0.00002061, 0.00000628, -0.00001585, 0.00000637, -0.00001142, 0.00000241, -0.00000698, 0.00000559, -0.00001079, 0.00000397, -0.00000861, -0.00000294, -0.00001111, 0.00000443, -0.00000414, 0.00000109, 0.00000118, -0.00000743, -0.00000072, -0.00000664, 0.00000412, -0.00001613, 0.00000510, -0.00001820, -0.00000931, -0.00000287, 0.00002002, 0.00002017, -0.00000409, },
		{ 0.00000737, 0.00000663, 0.00000980, 0.00001429, 0.00000103, 0.00000679, 0.00003050, 0.00001303, 0.00000104, -0.00002543, -0.00001557, -0.00000486, -0.00000145, -0.00000248, 0.00000031, 0.00000163, -0.00000370, 0.00000753, -0.00000985, 0.00000421, -0.00000183, 0.00000336, -0.00000980, 0.00000861, 0.00000746, -0.00000075, -0.00001099, 0.00000364, 0.00000533, -0.00000455, -0.00000189, -0.00000097, 0.00000279, -0.00000833, 0.00001202, -0.00000724, -0.00000388, 0.00000631, 0.00000295, -0.00000829, -0.00000625, 0.00000937, -0.00001071, 0.00001666, -0.00001730, 0.00000828, -0.00000736, -0.00000883, 0.00000765, -0.00000081, -0.00000005, -0.00001595, -0.00000198, 0.00000579, -0.00000635, -0.00000597, 0.00000585, 0.00000385, -0.00001057, -0.00000059, 0.00000062, 0.00000039, -0.00000980, 0.00000536, -0.00001193, 0.00001214, -0.00000388, 0.00000333, 0.00000918, -0.00001242, 0.00000299, -0.00000166, 0.00000093, 0.00000438, 0.00000087, -0.00000231, 0.00000201, -0.00000193, -0.00000672, 0.00000080, -0.00000213, -0.00001331, -0.00000024, -0.00001095, 0.00000421, -0.00001313, -0.00001404, -0.00003704, -0.00003113, -0.00016991, 0.00016452, 0.00011971, 0.00006722, 0.00008221, 0.00020138, -0.00010278, -0.00009176, -0.00000828, -0.00003363, -0.00001197, -0.00000741, -0.00000451, -0.00000618, -0.00000070, -0.00000572, -0.00000327, 0.00000082, -0.00000228, -0.00000477, -0.00000091, -0.00000610, -0.00000619, -0.00000660, 0.00000296, -0.00000389, 0.00000195, -0.00000156, -0.00000599, -0.00000147, -0.00000501, 0.00000180, -0.00001701, 0.00000907, -0.00002739, -0.00000158, -0.00000555, 0.00003416, 0.00000412, },
		{ -0.00000483, 0.00001490, 0.00000198, 0.00001722, 0.00000834, 0.00000135, 0.00001446, 0.00003217, 0.00000495, -0.00000488, -0.00003088, -0.00000546, -0.00000963, 0.00000523, -0.00000914, 0.00000867, -0.00000783, 0.00000639, -0.00000211, -0.00000417, 0.00000152, 0.00000173, -0.00000169, -0.00000572, 0.00000933, 0.00000684, -0.00000466, -0.00000902, 0.00000650, 0.00000284, -0.00000541, -0.00000124, 0.00000039, 0.00000024, -0.00000607, 0.00001378, -0.00001442, 0.00000480, 0.00000136, 0.00000498, -0.00001545, 0.00000620, -0.00000283, 0.00000090, 0.00000741, -0.00001535, 0.00001263, -0.00001759, 0.00000173, 0.00000364, 0.00000106, -0.00000622, -0.00001305, 0.00000213, 0.00000261, -0.00000848, -0.00000220, 0.00000811, -0.00000128, -0.00001064, 0.00000314, -0.00000121, -0.00000029, -0.00001001, 0.00000687, -0.00001321, 0.00001743, -0.00001292, 0.00001491, -0.00000301, -0.00000491, -0.00000090, 0.00000215, -0.00000193, 0.00000860, -0.00000617, 0.00000512, -0.00000500, 0.00000390, -0.00001432, 0.00001166, -0.00001762, 0.00000254, -0.00001549, 0.00000612, -0.00001380, 0.00000448, -0.00004510, -0.00000115, -0.00011102, -0.00007063, 0.00019961, 0.00006892, 0.00006814, 0.00010632, 0.00016525, -0.00015720, -0.00003801, -0.00002865, -0.00001914, -0.00001800, 0.00000308, -0.00001435, 0.00000377, -0.00000854, -0.00000020, -0.00000722, 0.00000665, -0.00000958, 0.00000182, -0.00000726, -0.00000206, -0.00000995, -0.00000123, -0.00000020, -0.00000209, 0.00000127, -0.00000359, -0.00000355, -0.00000420, -0.00000046, -0.00000584, -0.00000987, 0.00000371, -0.00002873, 0.00000001, 0.00000128, 0.00003730, },
		{ 0.00002992, -0.00000471, 0.00001625, 0.00000304, 0.00001970, 0.00000402, 0.00000249, 0.00002571, 0.00002510, 0.00000137, -0.00001696, -0.00002417, -0.00000548, -0.00000582, 0.00000306, -0.00000722, 0.00000866, -0.00001000, 0.00001101, -0.00000923, 0.00000130, -0.00000062, 0.00000368, -0.00000546, -0.00000145, 0.00000957, 0.00000528, -0.00000779, -0.00000615, 0.00000888, -0.00000130, -0.00000405, -0.00000192, 0.00000304, -0.00000489, 0.00000128, 0.00000800, -0.00001404, 0.00000834, -0.00000053, 0.00000256, -0.00001552, 0.00001236, -0.00001212, 0.00001366, -0.00000845, -0.00000198, 0.00000379, -0.00001689, 0.00000699, 0.00000158, -0.00000013, -0.00000990, -0.00000886, 0.00000433, -0.00000201, -0.00000741, 0.00000075, 0.00000973, -0.00000919, -0.00000458, 0.00000135, 0.00000086, -0.00000514, -0.00000498, 0.00000306, -0.00000832, 0.00001487, -0.00001393, 0.00001990, -0.00001133, 0.00000112, -0.00000355, 0.00000402, -0.00000170, 0.00000774, -0.00000708, 0.00000660, -0.00000615, 0.00000282, -0.00001311, 0.00001177, -0.00002183, 0.00000824, -0.00002080, 0.00001251, -0.00001996, 0.00000530, -0.00005485, 0.00000606, -0.00016753, 0.00006787, 0.00015527, 0.00005963, 0.00005589, 0.00014589, 0.00009132, -0.00016546, -0.00000625, -0.00004099, -0.00001058, -0.00001661, 0.00000408, -0.00001573, 0.00000693, -0.00001146, 0.00000094, -0.00000562, 0.00000588, -0.00001095, 0.00000355, -0.00001015, -0.00000027, -0.00001126, 0.00000221, -0.00000270, -0.00000034, -0.00000035, -0.00000400, -0.00000239, -0.00000572, 0.00000269, -0.00001354, 0.00000036, -0.00000782, -0.00002409, -0.00000188, 0.00001466, },
		{ 0.00002709, 0.00001851, 0.00000041, 0.00001443, 0.00000709, 0.00001885, 0.00000005, 0.00000860, 0.00003307, 0.00001572, -0.00000373, -0.00002618, -0.00001384, -0.00001001, 0.00000237, -0.00000473, 0.00000124, 0.00000086, -0.00000367, 0.00000719, -0.00000983, 0.00000326, -0.00000028, 0.00000258, -0.00000622, 0.00000124, 0.00001049, 0.00000194, -0.00000919, -0.00000274, 0.00000980, -0.00000559, -0.00000157, -0.00000287, 0.00000507, -0.00000955, 0.00001014, -0.00000268, -0.00000653, 0.00000540, 0.00000202, -0.00000494, -0.00000673, 0.00000707, -0.00000966, 0.00001569, -0.00001742, 0.00001051, -0.00001006, -0.00000811, 0.00000654, 0.00000176, -0.00000430, -0.00000969, -0.00000532, 0.00000442, -0.00000593, -0.00000474, 0.00000384, 0.00000794, -0.00001495, 0.00000324, -0.00000352, 0.00000516, -0.00001249, 0.00000401, -0.00000522, 0.00000174, 0.00000511, -0.00000619, 0.00001597, -0.00001242, 0.00000250, -0.00000280, 0.00000302, 0.00000129, 0.00000282, -0.00000258, 0.00000232, -0.00000173, -0.00000443, -0.00000399, 0.00000144, -0.00001347, 0.00000092, -0.00001262, 0.00000442, -0.00001068, -0.00001530, -0.00003452, -0.00003675, -0.00014761, 0.00016928, 0.00008428, 0.00007393, 0.00004150, 0.00018324, -0.00000235, -0.00013402, -0.00000068, -0.00004076, -0.00000896, -0.00000948, -0.00000258, -0.00000917, 0.00000293, -0.00000921, -0.00000140, -0.00000023, -0.00000016, -0.00000666, -0.00000027, -0.00000821, -0.00000198, -0.00000922, 0.00000262, -0.00000332, 0.00000016, -0.00000180, -0.00000314, -0.00000331, -0.00000421, 0.00000106, -0.00001577, 0.00000675, -0.00001917, -0.00001845, -0.00000140, },
		{ 0.00000457, 0.00003305, 0.00000922, 0.00000594, 0.00001292, 0.00001101, 0.00001635, -0.00000313, 0.00002027, 0.00003210, 0.00000850, -0.00001379, -0.00002540, -0.00000966, -0.00000984, 0.00000593, -0.00000924, 0.00000769, -0.00000685, 0.00000465, -0.00000071, -0.00000595, 0.00000275, 0.00000105, 0.00000025, -0.00000540, 0.00000339, 0.00001091, -0.00000203, -0.00000931, 0.00000159, 0.00000776, -0.00000755, -0.00000033, -0.00000195, 0.00000371, -0.00000948, 0.00001445, -0.00001140, 0.00000211, 0.00000039, 0.00000492, -0.00001197, 0.00000444, -0.00000462, 0.00000259, 0.00000466, -0.00001233, 0.00001145, -0.00001790, 0.00000075, 0.00000466, 0.00000095, -0.00000798, -0.00000741, -0.00000273, 0.00000231, -0.00000732, -0.00000232, 0.00000774, 0.00000093, -0.00001372, 0.00000647, -0.00000558, 0.00000621, -0.00001648, 0.00001108, -0.00001246, 0.00001128, -0.00000598, 0.00000516, 0.00000681, -0.00000843, 0.00000056, -0.00000008, 0.00000117, 0.00000387, -0.00000161, 0.00000208, -0.00000225, 0.00000232, -0.00001084, 0.00000464, -0.00000981, -0.00000226, -0.00000997, -0.00000074, -0.00000721, 0.00000012, -0.00003903, -0.00000747, -0.00010527, -0.00004938, 0.00018802, 0.00004086, 0.00007897, 0.00004762, 0.00019125, -0.00008479, -0.00008541, -0.00001142, -0.00003262, -0.00000955, -0.00000312, -0.00000963, -0.00000097, -0.00000349, -0.00000547, -0.00000323, 0.00000392, -0.00000618, -0.00000141, -0.00000591, -0.00000394, -0.00000546, -0.00000536, 0.00000101, -0.00000247, -0.00000073, -0.00000209, -0.00000249, -0.00000479, -0.00000115, -0.00000444, -0.00001156, 0.00000600, -0.00002633, -0.00001455, },
		{ -0.00001091, 0.00001396, 0.00003247, 0.00000382, 0.00001058, 0.00001227, 0.00001416, 0.00001178, -0.00000191, 0.00003244, 0.00002488, 0.00000177, -0.00002451, -0.00001697, -0.00001279, -0.00000254, 0.00000132, -0.00000569, 0.00000628, -0.00000771, 0.00000788, -0.00000636, -0.00000191, 0.00000234, 0.00000116, -0.00000098, -0.00000508, 0.00000667, 0.00000914, -0.00000503, -0.00000869, 0.00000639, 0.00000275, -0.00000624, -0.00000126, 0.00000125, -0.00000138, -0.00000342, 0.00001095, -0.00001286, 0.00000590, -0.00000148, 0.00000340, -0.00001205, 0.00000871, -0.00001170, 0.00001351, -0.00000999, 0.00000146, 0.00000077, -0.00001528, 0.00000446, 0.00000452, -0.00000307, -0.00000804, -0.00000570, -0.00000083, -0.00000108, -0.00000623, -0.00000015, 0.00001058, -0.00000879, -0.00000571, 0.00000283, -0.00000183, 0.00000065, -0.00001275, 0.00001107, -0.00001337, 0.00001454, -0.00001189, 0.00001349, -0.00000174, -0.00000385, -0.00000145, 0.00000211, 0.00000053, 0.00000372, -0.00000231, 0.00000286, -0.00000275, 0.00000124, -0.00001077, 0.00000575, -0.00001319, 0.00000173, -0.00001377, 0.00000350, -0.00000992, -0.00000335, -0.00004339, -0.00000856, -0.00014515, 0.00007476, 0.00013703, 0.00004149, 0.00006093, 0.00008261, 0.00015546, -0.00013177, -0.00004265, -0.00002520, -0.00002296, -0.00000850, -0.00000187, -0.00001155, 0.00000329, -0.00000785, -0.00000315, -0.00000221, 0.00000371, -0.00000817, 0.00000049, -0.00000914, -0.00000077, -0.00000809, -0.00000192, -0.00000085, -0.00000146, -0.00000216, -0.00000111, -0.00000332, -0.00000463, 0.00000010, -0.00000954, -0.00000473, -0.00000015, -0.00002969, },
		{ -0.00003121, -0.00000471, 0.00002245, 0.00002858, 0.00000108, 0.00001526, 0.00001118, 0.00001720, 0.00000479, 0.00000698, 0.00003849, 0.00001631, -0.00000833, -0.00002795, -0.00001015, -0.00001582, 0.00000552, -0.00000619, 0.00000199, -0.00000062, -0.00000229, 0.00000472, -0.00000730, 0.00000033, 0.00000264, 0.00000007, -0.00000106, -0.00000516, 0.00001075, 0.00000515, -0.00000666, -0.00000690, 0.00000976, -0.00000293, -0.00000335, -0.00000278, 0.00000435, -0.00000687, 0.00000500, 0.00000226, -0.00000771, 0.00000403, 0.00000050, -0.00000233, -0.00000481, 0.00000287, -0.00000749, 0.00001307, -0.00001586, 0.00001155, -0.00001218, -0.00000664, 0.00000407, 0.00000501, -0.00000861, -0.00000465, -0.00000564, 0.00000029, -0.00000407, -0.00000413, 0.00000292, 0.00000907, -0.00001534, 0.00000335, -0.00000371, 0.00000473, -0.00000886, -0.00000285, 0.00000417, -0.00000745, 0.00001038, -0.00001036, 0.00001561, -0.00000633, -0.00000137, -0.00000176, 0.00000288, 0.00000092, 0.00000184, -0.00000036, 0.00000064, -0.00000021, -0.00000362, -0.00000566, 0.00000045, -0.00000887, -0.00000231, -0.00000937, -0.00000072, -0.00000394, -0.00002067, -0.00002647, -0.00004919, -0.00011606, 0.00015813, 0.00006823, 0.00006373, 0.00003425, 0.00013068, 0.00008501, -0.00013839, -0.00001751, -0.00003397, -0.00001469, -0.00000631, -0.00000455, -0.00000843, 0.00000267, -0.00000914, -0.00000205, 0.00000031, 0.00000031, -0.00000644, -0.00000154, -0.00000857, -0.00000026, -0.00000861, -0.00000012, -0.00000180, -0.00000140, -0.00000281, -0.00000002, -0.00000512, -0.00000255, -0.00000204, -0.00001077, 0.00000008, -0.00000802, },
		{ -0.00001590, -0.00003089, 0.00000424, 0.00002786, 0.00002351, 0.00000053, 0.00001975, 0.00000964, 0.00001921, -0.00000188, 0.00002195, 0.00003590, 0.00000825, -0.00002109, -0.00002152, -0.00001123, -0.00001177, 0.00000676, -0.00000869, 0.00000609, -0.00000580, 0.00000336, -0.00000063, -0.00000556, 0.00000180, 0.00000220, -0.00000030, -0.00000195, -0.00000360, 0.00001340, 0.00000067, -0.00000762, -0.00000322, 0.00000969, -0.00000637, -0.00000153, -0.00000249, 0.00000436, -0.00000849, 0.00001011, -0.00000542, -0.00000129, 0.00000049, 0.00000250, -0.00000690, 0.00000301, -0.00000685, 0.00000406, 0.00000166, -0.00000867, 0.00000959, -0.00001765, 0.00000049, 0.00000377, 0.00000273, -0.00001154, -0.00000085, -0.00000648, 0.00000043, -0.00000549, -0.00000202, 0.00000680, 0.00000214, -0.00001435, 0.00000715, -0.00000694, 0.00000801, -0.00001588, 0.00000708, -0.00000457, 0.00000099, 0.00000251, -0.00000397, 0.00001285, -0.00000744, -0.00000071, -0.00000084, 0.00000302, 0.00000077, 0.00000095, 0.00000081, -0.00000080, 0.00000105, -0.00000711, -0.00000152, -0.00000441, -0.00000409, -0.00000715, -0.00000439, -0.00000458, -0.00000126, -0.00003506, -0.00001124, -0.00010209, -0.00002534, 0.00016755, 0.00002824, 0.00007547, 0.00002385, 0.00016433, 0.00000492, -0.00011647, -0.00000963, -0.00003592, -0.00000794, -0.00000553, -0.00000709, -0.00000393, -0.00000051, -0.00000859, -0.00000067, 0.00000158, -0.00000291, -0.00000409, -0.00000493, -0.00000572, -0.00000177, -0.00000744, 0.00000019, -0.00000185, -0.00000249, -0.00000175, -0.00000049, -0.00000585, -0.00000083, -0.00000562, -0.00000853, 0.00000130, },
		{ 0.00000032, -0.00002426, -0.00002669, 0.00001323, 0.00003119, 0.00001703, 0.00000359, 0.00002173, 0.00000965, 0.00001767, -0.00000290, 0.00003661, 0.00002736, -0.00000153, -0.00002993, -0.00001150, -0.00001721, -0.00000184, 0.00000065, -0.00000408, 0.00000329, -0.00000492, 0.00000437, -0.00000375, -0.00000386, 0.00000363, 0.00000018, 0.00000110, -0.00000475, 0.00000094, 0.00001278, -0.00000268, -0.00000820, 0.00000174, 0.00000612, -0.00000658, -0.00000169, 0.00000001, 0.00000089, -0.00000505, 0.00000911, -0.00000779, 0.00000174, -0.00000083, 0.00000118, -0.00000560, 0.00000386, -0.00001064, 0.00001211, -0.00001055, 0.00000436, -0.00000205, -0.00001341, 0.00000249, 0.00000551, -0.00000355, -0.00000912, 0.00000018, -0.00000683, -0.00000027, -0.00000541, 0.00000049, 0.00000903, -0.00000687, -0.00000697, 0.00000414, -0.00000420, 0.00000467, -0.00001577, 0.00001154, -0.00000984, 0.00000699, -0.00000426, 0.00000312, 0.00000823, -0.00000705, -0.00000045, 0.00000028, 0.00000303, -0.00000029, 0.00000191, -0.00000030, 0.00000032, -0.00000109, -0.00000651, -0.00000184, -0.00000433, -0.00000424, -0.00000704, -0.00000480, -0.00000166, -0.00001083, -0.00003195, -0.00002329, -0.00012164, 0.00007762, 0.00012031, 0.00002824, 0.00006367, 0.00004170, 0.00016442, -0.00006056, -0.00008318, -0.00001285, -0.00003197, -0.00000312, -0.00000682, -0.00000709, -0.00000104, -0.00000396, -0.00000698, 0.00000114, 0.00000070, -0.00000411, -0.00000352, -0.00000705, -0.00000281, -0.00000382, -0.00000564, -0.00000014, -0.00000177, -0.00000383, 0.00000056, -0.00000300, -0.00000413, -0.00000192, -0.00000721, -0.00000598, },
		{ -0.00000532, -0.00000162, -0.00003294, -0.00001759, 0.00001991, 0.00003366, 0.00000907, 0.00001105, 0.00001939, 0.00001264, 0.00001159, 0.00000532, 0.00004430, 0.00001663, -0.00001452, -0.00002874, -0.00000709, -0.00001959, 0.00000615, -0.00000609, 0.00000236, -0.00000255, -0.00000020, 0.00000128, -0.00000400, -0.00000254, 0.00000495, -0.00000193, 0.00000273, -0.00000787, 0.00000719, 0.00000920, -0.00000469, -0.00000768, 0.00000592, 0.00000130, -0.00000509, -0.00000238, 0.00000257, -0.00000325, 0.00000048, 0.00000423, -0.00000565, 0.00000118, -0.00000006, -0.00000214, 0.00000019, -0.00000298, -0.00000452, 0.00000921, -0.00001306, 0.00001157, -0.00001360, -0.00000497, 0.00000160, 0.00000701, -0.00001091, -0.00000256, -0.00000249, -0.00000584, -0.00000144, -0.00000432, 0.00000387, 0.00000691, -0.00001245, 0.00000114, -0.00000167, 0.00000162, -0.00000310, -0.00000926, 0.00000959, -0.00000995, 0.00000853, -0.00000753, 0.00000850, 0.00000354, -0.00000617, -0.00000008, 0.00000146, 0.00000224, -0.00000067, 0.00000281, -0.00000172, 0.00000188, -0.00000426, -0.00000426, -0.00000405, -0.00000166, -0.00000735, -0.00000402, -0.00000780, 0.00000347, -0.00002679, -0.00001671, -0.00006332, -0.00008204, 0.00013955, 0.00006172, 0.00004775, 0.00003984, 0.00007879, 0.00013041, -0.00009925, -0.00005173, -0.00002066, -0.00002362, -0.00000167, -0.00000818, -0.00000535, -0.00000026, -0.00000685, -0.00000410, 0.00000188, -0.00000080, -0.00000414, -0.00000456, -0.00000712, -0.00000103, -0.00000551, -0.00000395, -0.00000046, -0.00000237, -0.00000399, 0.00000226, -0.00000588, -0.00000117, -0.00000551, -0.00000561, },
		{ -0.00000314, -0.00000569, -0.00000558, -0.00003931, -0.00000576, 0.00002490, 0.00003400, 0.00000229, 0.00002037, 0.00001418, 0.00001694, 0.00000413, 0.00002096, 0.00004256, 0.00000468, -0.00002701, -0.00001890, -0.00001128, -0.00001364, 0.00000625, -0.00000697, 0.00000415, -0.00000524, 0.00000295, -0.00000192, -0.00000353, -0.00000033, 0.00000403, -0.00000182, 0.00000201, -0.00000842, 0.00001224, 0.00000457, -0.00000603, -0.00000532, 0.00000738, -0.00000225, -0.00000399, -0.00000177, 0.00000285, -0.00000481, 0.00000394, -0.00000004, -0.00000291, 0.00000002, -0.00000017, -0.00000286, 0.00000384, -0.00001046, 0.00000587, -0.00000187, -0.00000443, 0.00000715, -0.00001687, 0.00000072, 0.00000197, 0.00000458, -0.00001442, 0.00000386, -0.00000717, -0.00000363, -0.00000283, -0.00000212, 0.00000696, 0.00000063, -0.00001154, 0.00000478, -0.00000502, 0.00000541, -0.00001009, -0.00000083, 0.00000416, -0.00000666, 0.00000667, -0.00000754, 0.00001187, -0.00000101, -0.00000460, 0.00000010, 0.00000288, 0.00000013, 0.00000098, 0.00000152, -0.00000092, 0.00000087, -0.00000492, -0.00000440, -0.00000346, -0.00000177, -0.00000782, -0.00000413, -0.00000616, 0.00000041, -0.00003381, -0.00001155, -0.00010183, 0.00000079, 0.00014033, 0.00002783, 0.00006050, 0.00002424, 0.00011497, 0.00007578, -0.00011128, -0.00002889, -0.00002791, -0.00001328, -0.00000394, -0.00000756, -0.00000379, -0.00000088, -0.00000866, -0.00000027, 0.00000064, -0.00000123, -0.00000475, -0.00000566, -0.00000592, -0.00000043, -0.00000660, -0.00000256, -0.00000071, -0.00000375, -0.00000213, 0.00000148, -0.00000655, -0.00000011, -0.00000835, },
		{ -0.00000736, -0.00000310, -0.00000489, -0.00001337, -0.00003994, 0.00000508, 0.00003055, 0.00002957, 0.00000098, 0.00002667, 0.00001009, 0.00001881, 0.00000119, 0.00003758, 0.00003302, -0.00000932, -0.00003217, -0.00000894, -0.00001823, -0.00000321, 0.00000031, -0.00000218, 0.00000030, -0.00000275, 0.00000183, -0.00000264, -0.00000373, 0.00000290, 0.00000092, 0.00000085, -0.00000179, -0.00000495, 0.00001398, 0.00000053, -0.00000685, -0.00000172, 0.00000608, -0.00000378, -0.00000380, 0.00000014, 0.00000087, -0.00000329, 0.00000385, -0.00000133, -0.00000211, 0.00000014, -0.00000240, 0.00000128, 0.00000025, -0.00001057, 0.00001070, -0.00001097, 0.00000702, -0.00000469, -0.00001146, 0.00000138, 0.00000474, -0.00000257, -0.00001120, 0.00000593, -0.00001093, -0.00000116, -0.00000414, 0.00000153, 0.00000719, -0.00000608, -0.00000600, 0.00000324, -0.00000402, 0.00000461, -0.00001268, 0.00000538, -0.00000135, -0.00000245, 0.00000323, -0.00000523, 0.00001330, -0.00000522, -0.00000233, 0.00000014, 0.00000385, -0.00000213, 0.00000368, -0.00000135, 0.00000173, -0.00000256, -0.00000293, -0.00000723, 0.00000070, -0.00000603, -0.00000405, -0.00000883, 0.00000150, -0.00001391, -0.00002452, -0.00003313, -0.00010248, 0.00008202, 0.00009991, 0.00002488, 0.00005613, 0.00002784, 0.00013394, 0.00001699, -0.00010325, -0.00001669, -0.00003076, -0.00000462, -0.00000777, -0.00000513, -0.00000317, -0.00000253, -0.00000839, 0.00000274, -0.00000145, -0.00000080, -0.00000634, -0.00000584, -0.00000432, -0.00000080, -0.00000706, -0.00000141, -0.00000126, -0.00000503, 0.00000097, -0.00000178, -0.00000412, -0.00000284, },
		{ -0.00000772, -0.00000296, -0.00000619, -0.00000271, -0.00002454, -0.00003393, 0.00001335, 0.00003722, 0.00002010, 0.00000710, 0.00002658, 0.00001008, 0.00001614, 0.00000698, 0.00004838, 0.00001819, -0.00002372, -0.00002690, -0.00000664, -0.00001970, 0.00000374, -0.00000465, 0.00000244, -0.00000441, 0.00000147, -0.00000153, -0.00000176, -0.00000366, 0.00000525, -0.00000203, 0.00000371, -0.00000650, 0.00000117, 0.00001251, -0.00000256, -0.00000654, 0.00000140, 0.00000373, -0.00000440, -0.00000333, 0.00000158, -0.00000114, -0.00000107, 0.00000229, -0.00000097, -0.00000241, 0.00000025, -0.00000401, 0.00000677, -0.00000856, -0.00000227, 0.00000538, -0.00001007, 0.00001109, -0.00001442, -0.00000346, -0.00000003, 0.00000698, -0.00001090, -0.00000287, 0.00000227, -0.00001150, 0.00000018, -0.00000441, 0.00000583, 0.00000325, -0.00000897, -0.00000059, -0.00000019, -0.00000084, 0.00000062, -0.00001092, 0.00000810, -0.00000534, 0.00000142, -0.00000087, -0.00000085, 0.00001213, -0.00000800, -0.00000020, 0.00000070, 0.00000317, -0.00000241, 0.00000472, -0.00000292, 0.00000312, -0.00000507, -0.00000179, -0.00000878, 0.00000407, -0.00001013, -0.00000066, -0.00001241, 0.00000722, -0.00002943, -0.00001018, -0.00007286, -0.00005310, 0.00012228, 0.00005514, 0.00003638, 0.00004215, 0.00004711, 0.00013040, -0.00003395, -0.00008328, -0.00001402, -0.00002749, -0.00000082, -0.00000991, -0.00000283, -0.00000306, -0.00000497, -0.00000543, 0.00000302, -0.00000228, -0.00000122, -0.00000773, -0.00000533, -0.00000267, -0.00000200, -0.00000683, -0.00000052, -0.00000251, -0.00000486, 0.00000312, -0.00000522, -0.00000077, },
		{ -0.00000003, -0.00001059, 0.00000105, -0.00000973, -0.00000181, -0.00003556, -0.00002393, 0.00002091, 0.00004168, 0.00000971, 0.00001771, 0.00002119, 0.00001340, 0.00001094, 0.00002086, 0.00004941, -0.00000033, -0.00003306, -0.00001598, -0.00001186, -0.00001341, 0.00000328, -0.00000397, 0.00000207, -0.00000510, 0.00000286, -0.00000350, -0.00000148, -0.00000183, 0.00000483, -0.00000235, 0.00000399, -0.00000903, 0.00000714, 0.00000923, -0.00000487, -0.00000479, 0.00000274, 0.00000192, -0.00000522, -0.00000166, 0.00000121, -0.00000134, -0.00000055, 0.00000197, -0.00000152, -0.00000174, -0.00000180, -0.00000116, 0.00000714, -0.00001478, 0.00000734, -0.00000521, -0.00000035, 0.00000453, -0.00001561, 0.00000102, 0.00000043, 0.00000494, -0.00001505, 0.00000557, -0.00000494, -0.00000857, -0.00000049, -0.00000230, 0.00000824, -0.00000296, -0.00000695, 0.00000146, -0.00000211, 0.00000139, -0.00000355, -0.00000700, 0.00000806, -0.00000764, 0.00000450, -0.00000504, 0.00000525, 0.00000784, -0.00000824, 0.00000093, 0.00000197, 0.00000071, 0.00000001, 0.00000272, -0.00000139, 0.00000110, -0.00000416, -0.00000429, -0.00000558, 0.00000234, -0.00000936, -0.00000270, -0.00000875, 0.00000191, -0.00003236, -0.00001167, -0.00010019, 0.00002335, 0.00011370, 0.00003068, 0.00004419, 0.00003130, 0.00007116, 0.00010770, -0.00007052, -0.00005914, -0.00001722, -0.00001957, -0.00000218, -0.00000876, -0.00000210, -0.00000317, -0.00000713, -0.00000094, 0.00000066, -0.00000108, -0.00000353, -0.00000773, -0.00000478, -0.00000110, -0.00000386, -0.00000583, -0.00000011, -0.00000427, -0.00000263, 0.00000250, -0.00000607, },
		{ -0.00000385, -0.00000336, -0.00000861, 0.00000099, -0.00001053, -0.00000523, -0.00004242, -0.00001342, 0.00003042, 0.00003978, 0.00000440, 0.00002679, 0.00001493, 0.00001664, 0.00000821, 0.00003781, 0.00003974, -0.00001998, -0.00003235, -0.00000829, -0.00001741, -0.00000469, -0.00000111, 0.00000036, -0.00000236, -0.00000164, 0.00000048, -0.00000284, -0.00000235, 0.00000137, 0.00000207, -0.00000004, 0.00000109, -0.00000795, 0.00001106, 0.00000531, -0.00000611, -0.00000244, 0.00000267, 0.00000067, -0.00000593, 0.00000047, -0.00000023, -0.00000029, -0.00000150, 0.00000316, -0.00000343, -0.00000011, -0.00000526, 0.00000606, -0.00000042, -0.00001219, 0.00000968, -0.00001141, 0.00000929, -0.00000705, -0.00000946, 0.00000080, 0.00000332, -0.00000193, -0.00001201, 0.00000906, -0.00001154, -0.00000402, -0.00000251, 0.00000240, 0.00000629, -0.00000739, -0.00000276, 0.00000059, -0.00000190, 0.00000162, -0.00000623, -0.00000262, 0.00000597, -0.00000793, 0.00000588, -0.00000780, 0.00001144, 0.00000145, -0.00000610, 0.00000120, 0.00000298, -0.00000178, 0.00000324, -0.00000065, 0.00000167, -0.00000285, -0.00000122, -0.00000886, 0.00000092, -0.00000359, -0.00000451, -0.00000895, 0.00000036, -0.00001351, -0.00002070, -0.00003828, -0.00008715, 0.00008676, 0.00007829, 0.00002744, 0.00004263, 0.00002988, 0.00009083, 0.00007204, -0.00009001, -0.00003745, -0.00002138, -0.00001100, -0.00000593, -0.00000544, -0.00000279, -0.00000384, -0.00000733, 0.00000246, -0.00000213, 0.00000066, -0.00000687, -0.00000639, -0.00000430, -0.00000001, -0.00000590, -0.00000422, -0.00000055, -0.00000555, 0.00000062, -0.00000057, },
		{ -0.00000364, -0.00000108, -0.00000847, -0.00000296, -0.00000338, -0.00000814, -0.00001334, -0.00004382, -0.00000320, 0.00004139, 0.00003084, 0.00000735, 0.00002971, 0.00001181, 0.00001708, 0.00001232, 0.00005115, 0.00002025, -0.00003527, -0.00002279, -0.00000869, -0.00001683, -0.00000031, -0.00000320, 0.00000280, -0.00000592, 0.00000201, -0.00000289, -0.00000125, -0.00000284, 0.00000379, -0.00000060, 0.00000261, -0.00000302, -0.00000423, 0.00001265, 0.00000117, -0.00000568, -0.00000096, 0.00000256, -0.00000100, -0.00000534, 0.00000143, -0.00000081, -0.00000014, -0.00000162, 0.00000366, -0.00000472, 0.00000008, -0.00000610, 0.00001226, -0.00001148, -0.00000226, 0.00000273, -0.00000769, 0.00001046, -0.00001478, -0.00000214, -0.00000080, 0.00000549, -0.00000988, -0.00000348, 0.00000577, -0.00001406, -0.00000070, -0.00000359, 0.00000752, 0.00000022, -0.00000723, -0.00000009, -0.00000060, -0.00000120, 0.00000091, -0.00000765, 0.00000176, 0.00000206, -0.00000566, 0.00000437, -0.00000719, 0.00001506, -0.00000463, -0.00000300, 0.00000141, 0.00000255, -0.00000217, 0.00000447, -0.00000230, 0.00000292, -0.00000511, -0.00000015, -0.00001094, 0.00000555, -0.00000849, -0.00000114, -0.00001265, 0.00000573, -0.00002722, -0.00000880, -0.00007528, -0.00003218, 0.00010958, 0.00004543, 0.00003210, 0.00003705, 0.00003607, 0.00010192, 0.00002967, -0.00009266, -0.00002246, -0.00002273, -0.00000545, -0.00000852, -0.00000232, -0.00000400, -0.00000511, -0.00000491, 0.00000283, -0.00000298, 0.00000071, -0.00000930, -0.00000475, -0.00000351, 0.00000005, -0.00000741, -0.00000239, -0.00000200, -0.00000517, 0.00000273, },
		{ 0.00000227, -0.00000436, -0.00000097, -0.00001117, 0.00000213, -0.00000891, -0.00000513, -0.00002358, -0.00004140, 0.00000891, 0.00004948, 0.00001927, 0.00001633, 0.00002636, 0.00001260, 0.00001518, 0.00002422, 0.00005481, -0.00000575, -0.00003988, -0.00001209, -0.00001389, -0.00001060, -0.00000164, -0.00000063, 0.00000047, -0.00000525, 0.00000239, -0.00000425, -0.00000068, -0.00000166, 0.00000388, -0.00000120, 0.00000334, -0.00000595, 0.00000030, 0.00001240, -0.00000272, -0.00000375, -0.00000088, 0.00000307, -0.00000343, -0.00000318, 0.00000064, 0.00000022, -0.00000193, 0.00000057, 0.00000173, -0.00000381, -0.00000274, -0.00000110, 0.00001112, -0.00001775, 0.00000697, -0.00000714, 0.00000259, 0.00000225, -0.00001410, 0.00000121, -0.00000033, 0.00000349, -0.00001364, 0.00000528, -0.00000194, -0.00001182, -0.00000030, -0.00000151, 0.00000943, -0.00000648, -0.00000338, -0.00000020, -0.00000061, -0.00000120, 0.00000030, -0.00000827, 0.00000568, -0.00000293, -0.00000129, -0.00000001, -0.00000243, 0.00001425, -0.00000821, -0.00000031, 0.00000183, 0.00000079, -0.00000020, 0.00000287, -0.00000103, 0.00000085, -0.00000388, -0.00000296, -0.00000790, 0.00000484, -0.00000863, -0.00000309, -0.00000920, 0.00000027, -0.00002802, -0.00001470, -0.00009333, 0.00003776, 0.00009328, 0.00003022, 0.00003362, 0.00003389, 0.00004551, 0.00010288, -0.00001304, -0.00008179, -0.00001497, -0.00002045, -0.00000399, -0.00000824, -0.00000098, -0.00000518, -0.00000602, -0.00000122, 0.00000065, -0.00000153, -0.00000174, -0.00000959, -0.00000373, -0.00000221, -0.00000118, -0.00000785, -0.00000098, -0.00000394, -0.00000288, },
		{ 0.00000005, -0.00000008, -0.00000277, -0.00000433, -0.00000913, 0.00000310, -0.00001222, -0.00000448, -0.00003312, -0.00003677, 0.00002478, 0.00004992, 0.00001161, 0.00002555, 0.00002042, 0.00001513, 0.00001433, 0.00004075, 0.00004463, -0.00003121, -0.00003223, -0.00000801, -0.00001656, -0.00000471, -0.00000419, 0.00000325, -0.00000424, -0.00000159, -0.00000020, -0.00000332, -0.00000113, 0.00000054, 0.00000227, 0.00000000, 0.00000195, -0.00000682, 0.00000458, 0.00001042, -0.00000535, -0.00000161, -0.00000133, 0.00000351, -0.00000562, -0.00000067, -0.00000065, 0.00000149, -0.00000419, 0.00000385, -0.00000174, -0.00000186, -0.00000649, 0.00000796, 0.00000154, -0.00001445, 0.00000817, -0.00001112, 0.00001050, -0.00000876, -0.00000745, 0.00000040, 0.00000222, -0.00000279, -0.00001051, 0.00000904, -0.00000932, -0.00000726, -0.00000181, 0.00000362, 0.00000594, -0.00000955, 0.00000061, -0.00000148, -0.00000013, -0.00000137, -0.00000057, -0.00000752, 0.00000780, -0.00000701, 0.00000309, -0.00000516, 0.00000487, 0.00000939, -0.00000864, 0.00000150, 0.00000187, -0.00000073, 0.00000220, 0.00000031, 0.00000130, -0.00000264, -0.00000104, -0.00000752, -0.00000180, 0.00000036, -0.00000563, -0.00000796, -0.00000194, -0.00001274, -0.00001769, -0.00004174, -0.00007211, 0.00008771, 0.00006075, 0.00002913, 0.00003099, 0.00003426, 0.00005604, 0.00009242, -0.00004929, -0.00006311, -0.00001298, -0.00001622, -0.00000501, -0.00000576, -0.00000157, -0.00000626, -0.00000535, 0.00000138, -0.00000181, 0.00000044, -0.00000549, -0.00000804, -0.00000334, -0.00000075, -0.00000339, -0.00000705, -0.00000062, -0.00000522, },
		{ -0.00000481, 0.00000171, -0.00000211, -0.00000129, -0.00000866, -0.00000385, -0.00000027, -0.00001239, -0.00000710, -0.00004116, -0.00002891, 0.00004251, 0.00004193, 0.00001174, 0.00002995, 0.00001611, 0.00001674, 0.00001870, 0.00005534, 0.00002020, -0.00004670, -0.00001846, -0.00001146, -0.00001308, -0.00000376, -0.00000320, 0.00000381, -0.00000689, 0.00000131, -0.00000298, -0.00000187, -0.00000120, 0.00000199, 0.00000089, 0.00000124, -0.00000011, -0.00000626, 0.00000845, 0.00000667, -0.00000576, -0.00000074, -0.00000097, 0.00000272, -0.00000624, 0.00000062, -0.00000079, 0.00000105, -0.00000446, 0.00000551, -0.00000419, -0.00000166, -0.00000670, 0.00001469, -0.00001066, -0.00000459, 0.00000120, -0.00000583, 0.00000946, -0.00001462, -0.00000094, -0.00000113, 0.00000364, -0.00000945, -0.00000251, 0.00000613, -0.00001255, -0.00000387, -0.00000223, 0.00000863, -0.00000142, -0.00000745, 0.00000193, -0.00000185, -0.00000051, -0.00000059, -0.00000258, -0.00000462, 0.00000678, -0.00000810, 0.00000489, -0.00000805, 0.00001154, 0.00000272, -0.00000673, 0.00000250, 0.00000104, -0.00000062, 0.00000291, -0.00000073, 0.00000189, -0.00000445, 0.00000005, -0.00000988, 0.00000293, -0.00000355, -0.00000407, -0.00001000, 0.00000093, -0.00002225, -0.00001077, -0.00007240, -0.00001721, 0.00009927, 0.00003541, 0.00003084, 0.00002912, 0.00003533, 0.00006741, 0.00006988, -0.00007333, -0.00004283, -0.00001367, -0.00001216, -0.00000615, -0.00000285, -0.00000348, -0.00000684, -0.00000308, 0.00000160, -0.00000262, 0.00000091, -0.00000857, -0.00000586, -0.00000304, 0.00000016, -0.00000583, -0.00000533, -0.00000150, },
		{ -0.00000304, -0.00000279, 0.00000126, -0.00000224, -0.00000232, -0.00001045, 0.00000090, -0.00000539, -0.00001088, -0.00001209, -0.00004832, -0.00001489, 0.00005656, 0.00002991, 0.00001833, 0.00002837, 0.00001533, 0.00001675, 0.00003056, 0.00005956, -0.00001266, -0.00004612, -0.00000850, -0.00001629, -0.00000674, -0.00000647, 0.00000128, 0.00000014, -0.00000568, 0.00000129, -0.00000394, -0.00000115, -0.00000026, 0.00000196, 0.00000075, 0.00000136, -0.00000169, -0.00000490, 0.00001159, 0.00000181, -0.00000409, -0.00000145, 0.00000057, 0.00000055, -0.00000514, 0.00000045, 0.00000037, -0.00000136, -0.00000195, 0.00000423, -0.00000441, -0.00000416, -0.00000093, 0.00001335, -0.00001743, 0.00000381, -0.00000708, 0.00000400, 0.00000043, -0.00001241, 0.00000134, -0.00000047, 0.00000100, -0.00001161, 0.00000499, -0.00000069, -0.00001129, -0.00000321, 0.00000071, 0.00000953, -0.00000819, -0.00000244, 0.00000066, -0.00000102, -0.00000172, 0.00000099, -0.00000545, -0.00000007, 0.00000286, -0.00000579, 0.00000309, -0.00000681, 0.00001490, -0.00000329, -0.00000365, 0.00000261, -0.00000008, 0.00000080, 0.00000178, 0.00000020, 0.00000001, -0.00000351, -0.00000185, -0.00000837, 0.00000401, -0.00000473, -0.00000582, -0.00000698, -0.00000462, -0.00002098, -0.00002068, -0.00008120, 0.00004404, 0.00007939, 0.00002599, 0.00002881, 0.00003099, 0.00003542, 0.00007823, 0.00003723, -0.00008234, -0.00002569, -0.00001499, -0.00000926, -0.00000604, -0.00000100, -0.00000599, -0.00000614, -0.00000061, 0.00000017, -0.00000175, -0.00000074, -0.00000980, -0.00000403, -0.00000242, -0.00000001, -0.00000770, -0.00000341, },
		{ -0.00000213, -0.00000408, -0.00000049, -0.00000030, -0.00000115, -0.00000575, -0.00000832, 0.00000236, -0.00000965, -0.00000936, -0.00001891, -0.00005393, 0.00000659, 0.00006147, 0.00002049, 0.00002629, 0.00002358, 0.00001696, 0.00001745, 0.00004796, 0.00004647, -0.00004208, -0.00003221, -0.00000790, -0.00001597, -0.00000347, -0.00000781, 0.00000512, -0.00000458, -0.00000257, -0.00000067, -0.00000341, -0.00000080, 0.00000083, 0.00000137, 0.00000110, 0.00000086, -0.00000290, -0.00000254, 0.00001304, -0.00000266, -0.00000169, -0.00000279, 0.00000243, -0.00000203, -0.00000339, -0.00000005, 0.00000135, -0.00000389, 0.00000162, 0.00000129, -0.00000383, -0.00000677, 0.00000805, 0.00000405, -0.00001522, 0.00000483, -0.00000920, 0.00001003, -0.00000958, -0.00000555, 0.00000005, 0.00000173, -0.00000507, -0.00000735, 0.00000722, -0.00000653, -0.00000848, -0.00000359, 0.00000602, 0.00000497, -0.00001051, 0.00000176, -0.00000109, -0.00000034, -0.00000236, 0.00000188, -0.00000750, 0.00000414, -0.00000190, -0.00000181, -0.00000102, -0.00000193, 0.00001413, -0.00000710, -0.00000042, 0.00000166, -0.00000031, 0.00000188, 0.00000050, 0.00000136, -0.00000254, -0.00000153, -0.00000494, -0.00000471, 0.00000283, -0.00000490, -0.00000808, -0.00000307, -0.00001371, -0.00001367, -0.00004554, -0.00005529, 0.00008291, 0.00004972, 0.00002674, 0.00002480, 0.00003444, 0.00003637, 0.00008446, -0.00000013, -0.00007729, -0.00001426, -0.00001581, -0.00000737, -0.00000460, -0.00000101, -0.00000810, -0.00000397, 0.00000055, -0.00000121, -0.00000059, -0.00000355, -0.00000926, -0.00000283, -0.00000153, -0.00000136, -0.00000840, },
		{ -0.00000785, -0.00000187, -0.00000378, 0.00000059, -0.00000117, -0.00000102, -0.00000902, -0.00000402, 0.00000054, -0.00001223, -0.00000833, -0.00002837, -0.00005453, 0.00003243, 0.00005601, 0.00001812, 0.00003056, 0.00001970, 0.00001842, 0.00002319, 0.00006344, 0.00001570, -0.00005577, -0.00001569, -0.00001348, -0.00001006, -0.00000540, -0.00000483, 0.00000493, -0.00000687, -0.00000049, -0.00000242, -0.00000265, 0.00000003, 0.00000102, 0.00000143, 0.00000080, 0.00000080, -0.00000444, 0.00000132, 0.00001186, -0.00000525, -0.00000006, -0.00000353, 0.00000354, -0.00000396, -0.00000201, 0.00000011, 0.00000080, -0.00000451, 0.00000387, -0.00000121, -0.00000460, -0.00000556, 0.00001401, -0.00000719, -0.00000757, -0.00000048, -0.00000359, 0.00000747, -0.00001370, 0.00000014, -0.00000130, 0.00000216, -0.00001033, 0.00000042, 0.00000353, -0.00000826, -0.00000736, -0.00000179, 0.00000984, -0.00000240, -0.00000802, 0.00000319, -0.00000179, -0.00000074, -0.00000142, 0.00000082, -0.00000722, 0.00000603, -0.00000511, 0.00000136, -0.00000491, 0.00000446, 0.00001010, -0.00000820, 0.00000203, 0.00000008, 0.00000072, 0.00000160, 0.00000048, 0.00000099, -0.00000370, -0.00000068, -0.00000678, -0.00000147, 0.00000166, -0.00000650, -0.00000727, -0.00000405, -0.00001769, -0.00001309, -0.00006752, -0.00000468, 0.00008801, 0.00002870, 0.00002802, 0.00002365, 0.00003527, 0.00004146, 0.00008107, -0.00003471, -0.00006238, -0.00000903, -0.00001553, -0.00000596, -0.00000255, -0.00000302, -0.00000858, -0.00000138, 0.00000026, -0.00000163, -0.00000033, -0.00000619, -0.00000769, -0.00000210, -0.00000074, -0.00000351, },
		{ -0.00000584, -0.00000649, -0.00000225, -0.00000238, 0.00000018, -0.00000066, -0.00000322, -0.00000955, -0.00000057, -0.00000275, -0.00001362, -0.00000785, -0.00004131, -0.00004552, 0.00005583, 0.00004414, 0.00002244, 0.00002933, 0.00001924, 0.00001831, 0.00003736, 0.00006581, -0.00002314, -0.00004948, -0.00000732, -0.00001720, -0.00000399, -0.00000909, 0.00000058, 0.00000127, -0.00000622, -0.00000034, -0.00000309, -0.00000201, 0.00000110, 0.00000051, 0.00000207, -0.00000016, 0.00000122, -0.00000617, 0.00000628, 0.00000813, -0.00000547, 0.00000015, -0.00000315, 0.00000359, -0.00000505, -0.00000102, 0.00000064, -0.00000096, -0.00000304, 0.00000405, -0.00000272, -0.00000664, 0.00000052, 0.00001258, -0.00001358, -0.00000138, -0.00000585, 0.00000455, -0.00000142, -0.00001040, 0.00000148, -0.00000037, -0.00000153, -0.00001025, 0.00000602, -0.00000227, -0.00000691, -0.00000825, 0.00000344, 0.00000872, -0.00000803, -0.00000348, 0.00000256, -0.00000176, -0.00000161, 0.00000042, -0.00000177, -0.00000472, 0.00000520, -0.00000577, 0.00000226, -0.00000661, 0.00001006, 0.00000442, -0.00000682, 0.00000288, -0.00000093, 0.00000198, 0.00000058, 0.00000132, -0.00000077, -0.00000319, -0.00000133, -0.00000686, 0.00000072, 0.00000054, -0.00000891, -0.00000408, -0.00001032, -0.00001390, -0.00002696, -0.00006690, 0.00004577, 0.00006859, 0.00002138, 0.00002566, 0.00002698, 0.00003200, 0.00005170, 0.00006512, -0.00005949, -0.00004347, -0.00000875, -0.00001400, -0.00000456, -0.00000096, -0.00000625, -0.00000678, 0.00000015, -0.00000043, -0.00000151, -0.00000110, -0.00000796, -0.00000574, -0.00000163, -0.00000043, },
		{ -0.00000086, -0.00000770, -0.00000494, -0.00000247, -0.00000101, -0.00000061, 0.00000006, -0.00000678, -0.00000701, 0.00000033, -0.00000597, -0.00001408, -0.00000914, -0.00005608, -0.00002468, 0.00007000, 0.00003255, 0.00002911, 0.00002485, 0.00002139, 0.00001885, 0.00005803, 0.00004635, -0.00005358, -0.00003094, -0.00000938, -0.00001431, -0.00000276, -0.00000984, 0.00000436, -0.00000277, -0.00000452, -0.00000104, -0.00000321, -0.00000095, 0.00000146, 0.00000044, 0.00000213, -0.00000059, 0.00000103, -0.00000686, 0.00001092, 0.00000299, -0.00000385, -0.00000085, -0.00000184, 0.00000279, -0.00000562, 0.00000012, 0.00000050, -0.00000239, -0.00000092, 0.00000313, -0.00000445, -0.00000721, 0.00000804, 0.00000492, -0.00001300, -0.00000067, -0.00000576, 0.00000814, -0.00000980, -0.00000367, -0.00000021, 0.00000158, -0.00000788, -0.00000400, 0.00000543, -0.00000523, -0.00000627, -0.00000808, 0.00000950, 0.00000290, -0.00000929, 0.00000012, 0.00000151, -0.00000200, -0.00000164, 0.00000159, -0.00000429, -0.00000146, 0.00000279, -0.00000452, 0.00000102, -0.00000549, 0.00001334, -0.00000122, -0.00000384, 0.00000197, -0.00000048, 0.00000220, 0.00000021, 0.00000156, -0.00000247, -0.00000223, -0.00000241, -0.00000618, 0.00000278, -0.00000219, -0.00000910, -0.00000315, -0.00001589, -0.00000947, -0.00004888, -0.00003796, 0.00007434, 0.00004317, 0.00002200, 0.00002208, 0.00003162, 0.00002747, 0.00006391, 0.00003799, -0.00007043, -0.00002631, -0.00001098, -0.00001158, -0.00000284, -0.00000105, -0.00000895, -0.00000352, 0.00000009, -0.00000061, -0.00000164, -0.00000230, -0.00000878, -0.00000376, -0.00000134, },
		},
		{ { -0.00000581, -0.00000589, -0.00000289, -0.00001038, -0.00001513, -0.00001677, -0.00006593, 0.00000377, 0.00007279, 0.00002611, 0.00002482, 0.00002336, 0.00002457, 0.00002598, 0.00007698, 0.00000531, -0.00006293, -0.00001924, 0.00000272, -0.00001209, -0.00000715, -0.00000409, -0.00000220, 0.00000025, -0.00001559, -0.00000050, -0.00000822, -0.00000077, 0.00000056, 0.00000172, -0.00001615, 0.00000156, 0.00000008, -0.00000136, -0.00000002, -0.00000128, -0.00000207, -0.00000454, -0.00000168, -0.00000467, 0.00000639, 0.00000127, -0.00001084, -0.00000238, -0.00000170, 0.00000121, -0.00000094, 0.00000032, 0.00000655, -0.00000241, -0.00000392, 0.00000156, 0.00000162, 0.00000138, -0.00000029, 0.00000174, -0.00000179, -0.00000118, 0.00000975, 0.00000012, 0.00000066, -0.00000032, 0.00000150, -0.00000239, -0.00000248, -0.00000239, 0.00000150, -0.00000032, 0.00000066, 0.00000012, 0.00000975, -0.00000118, -0.00000179, 0.00000174, -0.00000029, 0.00000138, 0.00000162, 0.00000156, -0.00000392, -0.00000241, 0.00000655, 0.00000032, -0.00000094, 0.00000121, -0.00000170, -0.00000238, -0.00001084, 0.00000127, 0.00000639, -0.00000467, -0.00000168, -0.00000454, -0.00000207, -0.00000128, -0.00000002, -0.00000136, 0.00000008, 0.00000156, -0.00001615, 0.00000172, 0.00000056, -0.00000077, -0.00000822, -0.00000050, -0.00001559, 0.00000025, -0.00000220, -0.00000409, -0.00000715, -0.00001209, 0.00000272, -0.00001924, -0.00006293, 0.00000531, 0.00007698, 0.00002598, 0.00002457, 0.00002336, 0.00002482, 0.00002611, 0.00007279, 0.00000377, -0.00006593, -0.00001677, -0.00001513, -0.00001038, -0.00000289, -0.00000589, },
		{ -0.00000621, -0.00000590, -0.00000513, -0.00000339, -0.00001245, -0.00001414, -0.00002137, -0.00006919, 0.00001859, 0.00007300, 0.00002333, 0.00002678, 0.00002520, 0.00002358, 0.00003671, 0.00007964, -0.00002425, -0.00005753, -0.00001219, 0.00000139, -0.00001303, -0.00000500, -0.00000527, -0.00000052, -0.00000231, -0.00001429, -0.00000118, -0.00000778, 0.00000015, 0.00000111, -0.00000078, -0.00001607, 0.00000430, -0.00000188, -0.00000008, -0.00000135, -0.00000050, -0.00000307, -0.00000391, -0.00000191, -0.00000421, 0.00000851, -0.00000325, -0.00001001, 0.00000002, -0.00000334, 0.00000265, -0.00000278, 0.00000322, 0.00000518, -0.00000323, -0.00000322, 0.00000214, 0.00000092, 0.00000215, -0.00000124, 0.00000237, -0.00000350, 0.00000217, 0.00000874, -0.00000125, 0.00000126, -0.00000072, 0.00000144, -0.00000236, -0.00000260, -0.00000231, 0.00000164, 0.00000005, -0.00000014, 0.00000246, 0.00000901, -0.00000341, -0.00000029, 0.00000135, 0.00000037, 0.00000064, 0.00000243, 0.00000076, -0.00000464, -0.00000078, 0.00000660, -0.00000179, 0.00000074, -0.00000010, 0.00000009, -0.00000626, -0.00000898, 0.00000415, 0.00000418, -0.00000470, -0.00000177, -0.00000477, -0.00000136, -0.00000185, 0.00000121, -0.00000264, 0.00000216, -0.00000208, -0.00001465, 0.00000325, 0.00000026, -0.00000236, -0.00000757, -0.00000073, -0.00001619, 0.00000241, -0.00000368, -0.00000323, -0.00000916, -0.00001103, 0.00000428, -0.00003023, -0.00005884, 0.00003191, 0.00006692, 0.00001999, 0.00002455, 0.00002200, 0.00002280, 0.00002991, 0.00007011, -0.00000960, -0.00006132, -0.00001315, -0.00001611, -0.00000781, -0.00000328, },
		{ -0.00000436, -0.00000608, -0.00000612, -0.00000411, -0.00000475, -0.00001380, -0.00001348, -0.00002672, -0.00007086, 0.00003454, 0.00007059, 0.00002195, 0.00002851, 0.00002731, 0.00002233, 0.00005201, 0.00007133, -0.00005093, -0.00004534, -0.00000973, 0.00000139, -0.00001451, -0.00000233, -0.00000704, 0.00000152, -0.00000526, -0.00001246, -0.00000256, -0.00000630, 0.00000028, 0.00000192, -0.00000412, -0.00001426, 0.00000563, -0.00000316, 0.00000075, -0.00000241, 0.00000014, -0.00000408, -0.00000302, -0.00000250, -0.00000304, 0.00000978, -0.00000857, -0.00000654, 0.00000004, -0.00000363, 0.00000320, -0.00000385, 0.00000599, 0.00000305, -0.00000333, -0.00000259, 0.00000244, 0.00000056, 0.00000257, -0.00000203, 0.00000280, -0.00000481, 0.00000596, 0.00000620, -0.00000137, 0.00000129, -0.00000087, 0.00000135, -0.00000226, -0.00000293, -0.00000208, 0.00000190, 0.00000014, -0.00000063, 0.00000515, 0.00000676, -0.00000428, 0.00000065, 0.00000150, 0.00000037, 0.00000030, 0.00000313, -0.00000024, -0.00000516, 0.00000142, 0.00000525, -0.00000261, 0.00000157, -0.00000044, 0.00000076, -0.00001017, -0.00000522, 0.00000517, 0.00000240, -0.00000461, -0.00000205, -0.00000456, -0.00000113, -0.00000191, 0.00000193, -0.00000346, 0.00000373, -0.00000586, -0.00001203, 0.00000394, 0.00000000, -0.00000422, -0.00000606, -0.00000185, -0.00001601, 0.00000418, -0.00000517, -0.00000231, -0.00001160, -0.00000880, 0.00000406, -0.00004246, -0.00004543, 0.00005137, 0.00005351, 0.00001778, 0.00002342, 0.00002116, 0.00002088, 0.00003429, 0.00006517, -0.00002123, -0.00005560, -0.00001070, -0.00001666, -0.00000516, },
		{ -0.00000293, -0.00000574, -0.00000568, -0.00000627, -0.00000317, -0.00000664, -0.00001450, -0.00001327, -0.00003277, -0.00007058, 0.00005105, 0.00006573, 0.00002220, 0.00003001, 0.00002921, 0.00002234, 0.00006987, 0.00005043, -0.00006887, -0.00003104, -0.00001034, 0.00000239, -0.00001629, 0.00000063, -0.00000921, 0.00000379, -0.00000840, -0.00001044, -0.00000410, -0.00000424, -0.00000013, 0.00000267, -0.00000774, -0.00001109, 0.00000553, -0.00000357, 0.00000094, -0.00000300, 0.00000046, -0.00000487, -0.00000204, -0.00000341, -0.00000093, 0.00000938, -0.00001330, -0.00000142, -0.00000230, -0.00000203, 0.00000213, -0.00000331, 0.00000765, 0.00000100, -0.00000311, -0.00000193, 0.00000241, 0.00000064, 0.00000238, -0.00000228, 0.00000262, -0.00000507, 0.00000926, 0.00000282, -0.00000041, 0.00000064, -0.00000061, 0.00000115, -0.00000213, -0.00000342, -0.00000167, 0.00000227, -0.00000018, -0.00000035, 0.00000739, 0.00000370, -0.00000399, 0.00000105, 0.00000215, -0.00000041, 0.00000064, 0.00000345, -0.00000137, -0.00000521, 0.00000365, 0.00000287, -0.00000216, 0.00000146, 0.00000044, -0.00000051, -0.00001265, -0.00000077, 0.00000464, 0.00000127, -0.00000464, -0.00000233, -0.00000409, -0.00000134, -0.00000138, 0.00000199, -0.00000354, 0.00000431, -0.00000907, -0.00000889, 0.00000403, -0.00000043, -0.00000594, -0.00000407, -0.00000378, -0.00001496, 0.00000539, -0.00000657, -0.00000135, -0.00001457, -0.00000486, 0.00000020, -0.00005221, -0.00002539, 0.00006195, 0.00004009, 0.00001796, 0.00002140, 0.00002081, 0.00001920, 0.00003888, 0.00005819, -0.00003083, -0.00004916, -0.00000945, -0.00001643, },
		{ -0.00001527, -0.00000154, -0.00000700, -0.00000525, -0.00000612, -0.00000276, -0.00000853, -0.00001494, -0.00001327, -0.00003974, -0.00006770, 0.00006721, 0.00005897, 0.00002394, 0.00003147, 0.00003023, 0.00002577, 0.00008618, 0.00001851, -0.00007421, -0.00001957, -0.00001125, 0.00000297, -0.00001735, 0.00000298, -0.00001107, 0.00000576, -0.00001128, -0.00000876, -0.00000516, -0.00000231, -0.00000056, 0.00000281, -0.00001083, -0.00000738, 0.00000444, -0.00000331, 0.00000063, -0.00000323, 0.00000050, -0.00000542, -0.00000106, -0.00000453, 0.00000211, 0.00000665, -0.00001590, 0.00000363, -0.00000596, 0.00000111, -0.00000050, -0.00000089, 0.00000757, -0.00000025, -0.00000302, -0.00000107, 0.00000200, 0.00000116, 0.00000155, -0.00000186, 0.00000165, -0.00000387, 0.00001120, -0.00000050, 0.00000114, -0.00000055, 0.00000007, 0.00000082, -0.00000203, -0.00000400, -0.00000105, 0.00000264, -0.00000089, 0.00000092, 0.00000849, 0.00000067, -0.00000306, 0.00000123, 0.00000295, -0.00000176, 0.00000169, 0.00000324, -0.00000251, -0.00000459, 0.00000530, 0.00000014, -0.00000091, 0.00000080, 0.00000214, -0.00000378, -0.00001279, 0.00000314, 0.00000322, 0.00000074, -0.00000494, -0.00000234, -0.00000365, -0.00000174, -0.00000046, 0.00000151, -0.00000298, 0.00000380, -0.00001129, -0.00000568, 0.00000365, -0.00000112, -0.00000723, -0.00000194, -0.00000635, -0.00001287, 0.00000565, -0.00000739, -0.00000089, -0.00001743, 0.00000024, -0.00000785, -0.00005623, -0.00000278, 0.00006408, 0.00002890, 0.00001914, 0.00001893, 0.00002079, 0.00001785, 0.00004332, 0.00004944, -0.00003811, -0.00004252, -0.00000921, },
		{ -0.00000962, -0.00001325, -0.00000122, -0.00000774, -0.00000509, -0.00000544, -0.00000328, -0.00000986, -0.00001566, -0.00001310, -0.00004800, -0.00006138, 0.00008180, 0.00005124, 0.00002677, 0.00003317, 0.00002988, 0.00003476, 0.00009534, -0.00001928, -0.00006660, -0.00001426, -0.00000994, 0.00000155, -0.00001647, 0.00000366, -0.00001173, 0.00000676, -0.00001342, -0.00000780, -0.00000517, -0.00000115, -0.00000049, 0.00000183, -0.00001258, -0.00000407, 0.00000307, -0.00000283, 0.00000016, -0.00000337, 0.00000049, -0.00000582, -0.00000012, -0.00000567, 0.00000577, 0.00000149, -0.00001518, 0.00000675, -0.00000922, 0.00000455, -0.00000385, 0.00000288, 0.00000576, -0.00000036, -0.00000336, 0.00000010, 0.00000124, 0.00000197, 0.00000025, -0.00000089, 0.00000006, -0.00000118, 0.00001122, -0.00000288, 0.00000259, -0.00000187, 0.00000100, 0.00000038, -0.00000203, -0.00000455, -0.00000022, 0.00000285, -0.00000172, 0.00000310, 0.00000804, -0.00000157, -0.00000209, 0.00000165, 0.00000336, -0.00000317, 0.00000321, 0.00000248, -0.00000352, -0.00000326, 0.00000592, -0.00000224, 0.00000052, 0.00000025, 0.00000374, -0.00000833, -0.00001039, 0.00000558, 0.00000163, 0.00000055, -0.00000547, -0.00000194, -0.00000354, -0.00000195, 0.00000048, 0.00000083, -0.00000211, 0.00000244, -0.00001237, -0.00000271, 0.00000288, -0.00000196, -0.00000794, 0.00000003, -0.00000936, -0.00000971, 0.00000459, -0.00000702, -0.00000181, -0.00001889, 0.00000479, -0.00001885, -0.00005297, 0.00001845, 0.00005960, 0.00002103, 0.00002017, 0.00001656, 0.00002081, 0.00001707, 0.00004717, 0.00003936, -0.00004295, -0.00003619, },
		{ -0.00003056, -0.00001027, -0.00001067, -0.00000200, -0.00000773, -0.00000535, -0.00000418, -0.00000492, -0.00001021, -0.00001709, -0.00001246, -0.00005782, -0.00005088, 0.00009359, 0.00004358, 0.00003027, 0.00003512, 0.00002841, 0.00005024, 0.00009182, -0.00005522, -0.00004965, -0.00001548, -0.00000541, -0.00000253, -0.00001309, 0.00000211, -0.00001064, 0.00000637, -0.00001463, -0.00000759, -0.00000404, -0.00000105, 0.00000025, -0.00000035, -0.00001258, -0.00000185, 0.00000201, -0.00000255, -0.00000018, -0.00000365, 0.00000063, -0.00000618, 0.00000069, -0.00000648, 0.00000931, -0.00000543, -0.00001088, 0.00000660, -0.00001041, 0.00000675, -0.00000651, 0.00000682, 0.00000294, 0.00000046, -0.00000411, 0.00000150, 0.00000027, 0.00000285, -0.00000120, 0.00000025, -0.00000167, 0.00000251, 0.00000927, -0.00000378, 0.00000335, -0.00000287, 0.00000195, -0.00000010, -0.00000220, -0.00000497, 0.00000081, 0.00000273, -0.00000228, 0.00000571, 0.00000611, -0.00000261, -0.00000154, 0.00000268, 0.00000283, -0.00000404, 0.00000477, 0.00000128, -0.00000425, -0.00000138, 0.00000531, -0.00000375, 0.00000166, 0.00000037, 0.00000425, -0.00001293, -0.00000600, 0.00000616, 0.00000050, 0.00000033, -0.00000604, -0.00000121, -0.00000390, -0.00000171, 0.00000113, 0.00000031, -0.00000136, 0.00000064, -0.00001244, -0.00000016, 0.00000173, -0.00000281, -0.00000805, 0.00000150, -0.00001239, -0.00000574, 0.00000218, -0.00000518, -0.00000473, -0.00001774, 0.00000663, -0.00003007, -0.00004299, 0.00003547, 0.00005082, 0.00001675, 0.00002016, 0.00001491, 0.00002043, 0.00001720, 0.00004984, 0.00002863, -0.00004548, },
		{ -0.00004612, -0.00002581, -0.00001075, -0.00000795, -0.00000360, -0.00000699, -0.00000597, -0.00000255, -0.00000749, -0.00000961, -0.00001923, -0.00001145, -0.00006903, -0.00003591, 0.00010167, 0.00003683, 0.00003423, 0.00003681, 0.00002735, 0.00007073, 0.00007233, -0.00008121, -0.00002966, -0.00002079, 0.00000128, -0.00000853, -0.00000776, -0.00000136, -0.00000795, 0.00000472, -0.00001515, -0.00000768, -0.00000225, -0.00000168, 0.00000132, -0.00000326, -0.00001107, -0.00000087, 0.00000150, -0.00000256, -0.00000037, -0.00000406, 0.00000091, -0.00000645, 0.00000122, -0.00000650, 0.00001167, -0.00001265, -0.00000388, 0.00000299, -0.00000857, 0.00000647, -0.00000712, 0.00000951, 0.00000025, 0.00000155, -0.00000490, 0.00000285, -0.00000070, 0.00000352, -0.00000243, 0.00000113, -0.00000285, 0.00000636, 0.00000588, -0.00000315, 0.00000308, -0.00000320, 0.00000266, -0.00000054, -0.00000256, -0.00000516, 0.00000198, 0.00000215, -0.00000216, 0.00000807, 0.00000320, -0.00000249, -0.00000148, 0.00000437, 0.00000110, -0.00000383, 0.00000586, -0.00000008, -0.00000459, 0.00000073, 0.00000358, -0.00000421, 0.00000231, 0.00000132, 0.00000295, -0.00001628, -0.00000065, 0.00000500, 0.00000019, -0.00000027, -0.00000629, -0.00000042, -0.00000458, -0.00000101, 0.00000136, 0.00000012, -0.00000104, -0.00000119, -0.00001165, 0.00000179, 0.00000035, -0.00000360, -0.00000752, 0.00000211, -0.00001478, -0.00000171, -0.00000095, -0.00000231, -0.00000941, -0.00001366, 0.00000420, -0.00003838, -0.00002850, 0.00004696, 0.00003998, 0.00001569, 0.00001862, 0.00001451, 0.00001923, 0.00001859, 0.00005072, 0.00001808, },
		{ 0.00000853, -0.00004546, -0.00002189, -0.00001079, -0.00000561, -0.00000552, -0.00000584, -0.00000666, -0.00000106, -0.00001040, -0.00000857, -0.00002161, -0.00001073, -0.00008079, -0.00001686, 0.00010575, 0.00003140, 0.00003881, 0.00003723, 0.00002950, 0.00009174, 0.00003764, -0.00009151, -0.00001337, -0.00002610, 0.00000766, -0.00001451, -0.00000203, -0.00000557, -0.00000449, 0.00000240, -0.00001557, -0.00000725, -0.00000070, -0.00000229, 0.00000194, -0.00000594, -0.00000889, -0.00000069, 0.00000128, -0.00000263, -0.00000071, -0.00000427, 0.00000109, -0.00000643, 0.00000123, -0.00000533, 0.00001175, -0.00001823, 0.00000387, -0.00000295, -0.00000400, 0.00000346, -0.00000499, 0.00000989, -0.00000120, 0.00000209, -0.00000518, 0.00000375, -0.00000137, 0.00000376, -0.00000314, 0.00000140, -0.00000289, 0.00000937, 0.00000196, -0.00000141, 0.00000185, -0.00000273, 0.00000297, -0.00000089, -0.00000310, -0.00000506, 0.00000318, 0.00000112, -0.00000107, 0.00000946, 0.00000013, -0.00000166, -0.00000161, 0.00000634, -0.00000167, -0.00000234, 0.00000607, -0.00000130, -0.00000451, 0.00000269, 0.00000109, -0.00000371, 0.00000259, 0.00000284, -0.00000025, -0.00001739, 0.00000444, 0.00000266, 0.00000073, -0.00000148, -0.00000589, 0.00000001, -0.00000523, -0.00000010, 0.00000134, 0.00000015, -0.00000118, -0.00000278, -0.00001019, 0.00000295, -0.00000100, -0.00000438, -0.00000631, 0.00000151, -0.00001582, 0.00000137, -0.00000366, 0.00000038, -0.00001460, -0.00000772, -0.00000251, -0.00004151, -0.00001234, 0.00005280, 0.00002899, 0.00001703, 0.00001555, 0.00001551, 0.00001700, 0.00002141, 0.00004936, },
		{ 0.00004571, 0.00000055, -0.00004406, -0.00001867, -0.00001024, -0.00000412, -0.00000715, -0.00000478, -0.00000696, -0.00000035, -0.00001284, -0.00000791, -0.00002336, -0.00001142, -0.00009167, 0.00000519, 0.00010613, 0.00002738, 0.00004429, 0.00003535, 0.00003810, 0.00010630, -0.00000660, -0.00008492, -0.00000541, -0.00002776, 0.00001112, -0.00001836, 0.00000231, -0.00000905, -0.00000134, 0.00000013, -0.00001643, -0.00000562, -0.00000027, -0.00000210, 0.00000134, -0.00000737, -0.00000709, -0.00000056, 0.00000088, -0.00000233, -0.00000162, -0.00000381, 0.00000075, -0.00000585, 0.00000056, -0.00000282, 0.00000887, -0.00002040, 0.00000995, -0.00000910, 0.00000182, -0.00000139, -0.00000049, 0.00000773, -0.00000080, 0.00000149, -0.00000452, 0.00000386, -0.00000153, 0.00000345, -0.00000325, 0.00000099, -0.00000151, 0.00001072, -0.00000147, 0.00000069, 0.00000006, -0.00000158, 0.00000285, -0.00000118, -0.00000375, -0.00000465, 0.00000427, -0.00000021, 0.00000104, 0.00000934, -0.00000229, -0.00000076, -0.00000137, 0.00000789, -0.00000486, 0.00000020, 0.00000528, -0.00000214, -0.00000403, 0.00000416, -0.00000173, -0.00000246, 0.00000279, 0.00000429, -0.00000488, -0.00001586, 0.00000818, -0.00000005, 0.00000184, -0.00000323, -0.00000471, -0.00000027, -0.00000543, 0.00000064, 0.00000135, 0.00000012, -0.00000160, -0.00000404, -0.00000821, 0.00000317, -0.00000204, -0.00000520, -0.00000446, -0.00000041, -0.00001500, 0.00000259, -0.00000476, 0.00000138, -0.00001840, -0.00000211, -0.00001174, -0.00003888, 0.00000308, 0.00005355, 0.00001943, 0.00001950, 0.00001156, 0.00001760, 0.00001395, 0.00002541, },
		{ 0.00002989, 0.00004022, -0.00000570, -0.00004232, -0.00001592, -0.00000912, -0.00000375, -0.00000798, -0.00000425, -0.00000650, -0.00000095, -0.00001406, -0.00000841, -0.00002361, -0.00001470, -0.00010004, 0.00002882, 0.00010336, 0.00002482, 0.00005054, 0.00003109, 0.00005523, 0.00010689, -0.00005102, -0.00006538, -0.00000665, -0.00002418, 0.00001037, -0.00001889, 0.00000414, -0.00001081, 0.00000076, -0.00000172, -0.00001781, -0.00000264, -0.00000130, -0.00000078, -0.00000079, -0.00000696, -0.00000647, 0.00000015, -0.00000012, -0.00000139, -0.00000338, -0.00000236, -0.00000038, -0.00000457, -0.00000069, 0.00000064, 0.00000319, -0.00001824, 0.00001239, -0.00001311, 0.00000675, -0.00000628, 0.00000502, 0.00000378, 0.00000128, -0.00000031, -0.00000282, 0.00000304, -0.00000111, 0.00000266, -0.00000290, 0.00000017, 0.00000112, 0.00001006, -0.00000358, 0.00000235, -0.00000172, -0.00000005, 0.00000238, -0.00000144, -0.00000442, -0.00000394, 0.00000511, -0.00000152, 0.00000387, 0.00000757, -0.00000347, -0.00000036, -0.00000025, 0.00000821, -0.00000755, 0.00000317, 0.00000371, -0.00000252, -0.00000322, 0.00000486, -0.00000441, -0.00000075, 0.00000314, 0.00000495, -0.00000999, -0.00001191, 0.00000989, -0.00000230, 0.00000298, -0.00000518, -0.00000291, -0.00000131, -0.00000497, 0.00000095, 0.00000168, -0.00000024, -0.00000207, -0.00000490, -0.00000597, 0.00000250, -0.00000264, -0.00000602, -0.00000230, -0.00000331, -0.00001245, 0.00000169, -0.00000363, -0.00000033, -0.00001919, 0.00000069, -0.00002058, -0.00003167, 0.00001634, 0.00004997, 0.00001260, 0.00002164, 0.00000769, 0.00001994, 0.00001077, },
		{ 0.00000834, 0.00003391, 0.00003369, -0.00001041, -0.00004048, -0.00001339, -0.00000771, -0.00000444, -0.00000781, -0.00000443, -0.00000523, -0.00000297, -0.00001377, -0.00001044, -0.00002197, -0.00002132, -0.00010454, 0.00005275, 0.00009787, 0.00002422, 0.00005650, 0.00002617, 0.00008003, 0.00008833, -0.00008521, -0.00004056, -0.00001415, -0.00001651, 0.00000608, -0.00001645, 0.00000358, -0.00001086, 0.00000183, -0.00000342, -0.00001910, 0.00000103, -0.00000334, 0.00000124, -0.00000397, -0.00000493, -0.00000716, 0.00000160, -0.00000168, 0.00000007, -0.00000582, -0.00000001, -0.00000226, -0.00000277, -0.00000208, 0.00000411, -0.00000413, -0.00001218, 0.00001041, -0.00001339, 0.00000886, -0.00000921, 0.00000964, -0.00000050, 0.00000416, -0.00000281, -0.00000042, 0.00000146, -0.00000024, 0.00000160, -0.00000239, -0.00000055, 0.00000436, 0.00000762, -0.00000403, 0.00000300, -0.00000293, 0.00000146, 0.00000169, -0.00000171, -0.00000504, -0.00000294, 0.00000551, -0.00000242, 0.00000684, 0.00000449, -0.00000327, -0.00000068, 0.00000196, 0.00000677, -0.00000886, 0.00000573, 0.00000185, -0.00000255, -0.00000213, 0.00000463, -0.00000657, 0.00000124, 0.00000369, 0.00000427, -0.00001443, -0.00000633, 0.00000938, -0.00000346, 0.00000356, -0.00000678, -0.00000098, -0.00000283, -0.00000395, 0.00000084, 0.00000233, -0.00000107, -0.00000241, -0.00000529, -0.00000387, 0.00000125, -0.00000290, -0.00000659, -0.00000037, -0.00000640, -0.00000899, -0.00000067, -0.00000070, -0.00000462, -0.00001651, -0.00000088, -0.00002611, -0.00002209, 0.00002697, 0.00004292, 0.00000930, 0.00002212, 0.00000511, 0.00002152, },
		{ 0.00002153, 0.00000743, 0.00003657, 0.00002702, -0.00001400, -0.00003858, -0.00001083, -0.00000646, -0.00000575, -0.00000689, -0.00000512, -0.00000351, -0.00000600, -0.00001229, -0.00001370, -0.00001875, -0.00003125, -0.00010449, 0.00007614, 0.00008974, 0.00002659, 0.00006018, 0.00002437, 0.00010757, 0.00005037, -0.00010128, -0.00001905, -0.00002279, -0.00000787, 0.00000051, -0.00001275, 0.00000189, -0.00001008, 0.00000256, -0.00000573, -0.00001918, 0.00000412, -0.00000531, 0.00000293, -0.00000702, -0.00000226, -0.00000855, 0.00000333, -0.00000333, 0.00000144, -0.00000829, 0.00000270, -0.00000443, -0.00000087, -0.00000298, 0.00000629, -0.00001109, -0.00000404, 0.00000485, -0.00000978, 0.00000729, -0.00000884, 0.00001175, -0.00000351, 0.00000661, -0.00000512, 0.00000203, -0.00000042, 0.00000081, 0.00000054, -0.00000207, -0.00000056, 0.00000725, 0.00000417, -0.00000304, 0.00000249, -0.00000325, 0.00000266, 0.00000090, -0.00000199, -0.00000558, -0.00000166, 0.00000538, -0.00000249, 0.00000918, 0.00000083, -0.00000204, -0.00000149, 0.00000496, 0.00000354, -0.00000826, 0.00000714, 0.00000028, -0.00000246, -0.00000078, 0.00000338, -0.00000786, 0.00000334, 0.00000428, 0.00000209, -0.00001722, -0.00000023, 0.00000702, -0.00000326, 0.00000315, -0.00000747, 0.00000048, -0.00000430, -0.00000275, 0.00000057, 0.00000310, -0.00000223, -0.00000259, -0.00000510, -0.00000237, -0.00000007, -0.00000313, -0.00000657, 0.00000064, -0.00000861, -0.00000595, -0.00000307, 0.00000257, -0.00001005, -0.00001153, -0.00000658, -0.00002669, -0.00001230, 0.00003490, 0.00003353, 0.00000957, 0.00002021, 0.00000465, },
		{ 0.00000642, 0.00001970, 0.00000832, 0.00003736, 0.00002092, -0.00001688, -0.00003660, -0.00000804, -0.00000592, -0.00000698, -0.00000575, -0.00000580, -0.00000203, -0.00000921, -0.00001043, -0.00001733, -0.00001493, -0.00004375, -0.00009983, 0.00009855, 0.00007886, 0.00003303, 0.00005929, 0.00003060, 0.00012922, -0.00000100, -0.00009683, -0.00000704, -0.00002777, -0.00000162, -0.00000371, -0.00000987, 0.00000074, -0.00000973, 0.00000382, -0.00000945, -0.00001700, 0.00000533, -0.00000597, 0.00000307, -0.00000857, -0.00000032, -0.00000962, 0.00000457, -0.00000432, 0.00000192, -0.00000991, 0.00000495, -0.00000627, 0.00000059, -0.00000283, 0.00000606, -0.00001556, 0.00000356, -0.00000209, -0.00000369, 0.00000267, -0.00000509, 0.00001059, -0.00000417, 0.00000760, -0.00000639, 0.00000379, -0.00000209, 0.00000173, -0.00000032, -0.00000212, 0.00000056, 0.00000890, 0.00000074, -0.00000129, 0.00000112, -0.00000266, 0.00000336, 0.00000008, -0.00000224, -0.00000605, -0.00000013, 0.00000467, -0.00000149, 0.00001019, -0.00000248, -0.00000048, -0.00000215, 0.00000797, -0.00000086, -0.00000580, 0.00000705, -0.00000053, -0.00000250, 0.00000076, 0.00000113, -0.00000803, 0.00000538, 0.00000461, -0.00000138, -0.00001770, 0.00000519, 0.00000358, -0.00000191, 0.00000169, -0.00000698, 0.00000097, -0.00000518, -0.00000185, 0.00000056, 0.00000361, -0.00000348, -0.00000262, -0.00000431, -0.00000181, -0.00000098, -0.00000363, -0.00000574, 0.00000032, -0.00000915, -0.00000456, -0.00000396, 0.00000432, -0.00001443, -0.00000672, -0.00001432, -0.00002253, -0.00000357, 0.00004000, 0.00002331, 0.00001257, 0.00001605, },
		{ 0.00001062, 0.00000978, 0.00001643, 0.00001073, 0.00003629, 0.00001575, -0.00001939, -0.00003444, -0.00000500, -0.00000644, -0.00000750, -0.00000501, -0.00000595, -0.00000148, -0.00001175, -0.00000906, -0.00002036, -0.00001171, -0.00005772, -0.00009075, 0.00011948, 0.00006553, 0.00004393, 0.00005265, 0.00004880, 0.00013496, -0.00005454, -0.00007595, -0.00000607, -0.00002683, 0.00000044, -0.00000507, -0.00000908, 0.00000116, -0.00001055, 0.00000598, -0.00001467, -0.00001225, 0.00000402, -0.00000467, 0.00000101, -0.00000769, -0.00000021, -0.00000946, 0.00000468, -0.00000410, 0.00000091, -0.00000996, 0.00000604, -0.00000723, 0.00000133, -0.00000155, 0.00000301, -0.00001603, 0.00000813, -0.00000764, 0.00000245, -0.00000311, 0.00000078, 0.00000666, -0.00000232, 0.00000671, -0.00000615, 0.00000435, -0.00000313, 0.00000231, -0.00000090, -0.00000245, 0.00000283, 0.00000872, -0.00000169, 0.00000034, -0.00000055, -0.00000142, 0.00000356, -0.00000072, -0.00000242, -0.00000646, 0.00000158, 0.00000352, 0.00000054, 0.00000946, -0.00000463, 0.00000068, -0.00000197, 0.00000993, -0.00000533, -0.00000220, 0.00000562, -0.00000048, -0.00000276, 0.00000236, -0.00000194, -0.00000690, 0.00000713, 0.00000436, -0.00000557, -0.00001571, 0.00000892, 0.00000005, 0.00000000, -0.00000045, -0.00000546, 0.00000038, -0.00000519, -0.00000157, 0.00000109, 0.00000355, -0.00000457, -0.00000247, -0.00000314, -0.00000222, -0.00000127, -0.00000446, -0.00000422, -0.00000121, -0.00000791, -0.00000531, -0.00000245, 0.00000323, -0.00001583, -0.00000474, -0.00002096, -0.00001545, 0.00000392, 0.00004187, 0.00001409, 0.00001673, },
		{ 0.00002022, 0.00000535, 0.00001362, 0.00001255, 0.00001395, 0.00003379, 0.00001155, -0.00002173, -0.00003188, -0.00000198, -0.00000802, -0.00000700, -0.00000498, -0.00000533, -0.00000215, -0.00001316, -0.00000869, -0.00002222, -0.00000993, -0.00007232, -0.00007720, 0.00013790, 0.00005111, 0.00005805, 0.00004164, 0.00007943, 0.00011694, -0.00009706, -0.00004782, -0.00001288, -0.00002115, -0.00000104, -0.00000401, -0.00001006, 0.00000289, -0.00001222, 0.00000844, -0.00002053, -0.00000571, 0.00000064, -0.00000174, -0.00000290, -0.00000444, -0.00000222, -0.00000777, 0.00000352, -0.00000269, -0.00000159, -0.00000827, 0.00000573, -0.00000719, 0.00000149, 0.00000023, -0.00000213, -0.00001240, 0.00000832, -0.00000965, 0.00000616, -0.00000762, 0.00000666, 0.00000148, 0.00000124, 0.00000432, -0.00000454, 0.00000363, -0.00000336, 0.00000250, -0.00000135, -0.00000269, 0.00000578, 0.00000672, -0.00000257, 0.00000116, -0.00000186, 0.00000007, 0.00000335, -0.00000147, -0.00000254, -0.00000677, 0.00000334, 0.00000222, 0.00000323, 0.00000703, -0.00000520, 0.00000098, -0.00000046, 0.00000995, -0.00000860, 0.00000146, 0.00000353, 0.00000017, -0.00000310, 0.00000369, -0.00000545, -0.00000446, 0.00000832, 0.00000331, -0.00000968, -0.00001162, 0.00001038, -0.00000259, 0.00000167, -0.00000256, -0.00000351, -0.00000093, -0.00000448, -0.00000186, 0.00000218, 0.00000283, -0.00000540, -0.00000196, -0.00000207, -0.00000321, -0.00000113, -0.00000530, -0.00000262, -0.00000315, -0.00000571, -0.00000755, 0.00000106, -0.00000067, -0.00001373, -0.00000713, -0.00002370, -0.00000791, 0.00001067, 0.00004010, 0.00000756, },
		{ 0.00000467, 0.00002166, 0.00000150, 0.00001686, 0.00000890, 0.00001726, 0.00003041, 0.00000819, -0.00002406, -0.00002861, 0.00000044, -0.00001011, -0.00000574, -0.00000547, -0.00000426, -0.00000372, -0.00001365, -0.00000909, -0.00002311, -0.00000970, -0.00008729, -0.00005852, 0.00015194, 0.00003820, 0.00007221, 0.00003084, 0.00011757, 0.00007321, -0.00011789, -0.00002316, -0.00002121, -0.00001447, -0.00000331, -0.00000267, -0.00001100, 0.00000443, -0.00001343, 0.00000980, -0.00002529, 0.00000085, -0.00000335, 0.00000149, -0.00000730, 0.00000007, -0.00000566, -0.00000502, 0.00000166, -0.00000075, -0.00000489, -0.00000535, 0.00000436, -0.00000649, 0.00000166, 0.00000132, -0.00000759, -0.00000614, 0.00000454, -0.00000745, 0.00000598, -0.00000893, 0.00001041, -0.00000303, 0.00000507, 0.00000137, -0.00000223, 0.00000200, -0.00000294, 0.00000242, -0.00000190, -0.00000230, 0.00000858, 0.00000353, -0.00000195, 0.00000084, -0.00000238, 0.00000144, 0.00000283, -0.00000209, -0.00000267, -0.00000687, 0.00000491, 0.00000116, 0.00000591, 0.00000347, -0.00000434, 0.00000045, 0.00000230, 0.00000768, -0.00000972, 0.00000405, 0.00000166, 0.00000088, -0.00000321, 0.00000431, -0.00000880, -0.00000093, 0.00000870, 0.00000143, -0.00001293, -0.00000621, 0.00000951, -0.00000375, 0.00000238, -0.00000387, -0.00000191, -0.00000229, -0.00000356, -0.00000235, 0.00000352, 0.00000164, -0.00000602, -0.00000084, -0.00000171, -0.00000409, -0.00000104, -0.00000564, -0.00000174, -0.00000433, -0.00000402, -0.00000969, 0.00000494, -0.00000581, -0.00000944, -0.00001342, -0.00002130, -0.00000178, 0.00001706, 0.00003487, },
		{ 0.00002725, 0.00000521, 0.00002060, -0.00000031, 0.00001890, 0.00000595, 0.00002017, 0.00002659, 0.00000554, -0.00002651, -0.00002426, 0.00000148, -0.00001180, -0.00000445, -0.00000585, -0.00000349, -0.00000538, -0.00001400, -0.00000948, -0.00002386, -0.00001047, -0.00010288, -0.00003371, 0.00015929, 0.00003010, 0.00008213, 0.00002723, 0.00015291, 0.00000976, -0.00011286, -0.00001007, -0.00002501, -0.00001096, -0.00000316, -0.00000371, -0.00000963, 0.00000385, -0.00001257, 0.00000850, -0.00002711, 0.00000542, -0.00000612, 0.00000330, -0.00001036, 0.00000411, -0.00000923, -0.00000223, 0.00000003, 0.00000068, -0.00000788, -0.00000219, 0.00000266, -0.00000573, 0.00000250, 0.00000044, -0.00001112, 0.00000015, -0.00000110, -0.00000220, 0.00000214, -0.00000643, 0.00001077, -0.00000532, 0.00000781, -0.00000102, -0.00000013, 0.00000012, -0.00000219, 0.00000226, -0.00000272, -0.00000080, 0.00001025, 0.00000016, -0.00000042, -0.00000044, -0.00000198, 0.00000246, 0.00000206, -0.00000248, -0.00000298, -0.00000659, 0.00000608, 0.00000076, 0.00000781, -0.00000033, -0.00000264, -0.00000036, 0.00000560, 0.00000354, -0.00000845, 0.00000488, 0.00000077, 0.00000110, -0.00000274, 0.00000376, -0.00001126, 0.00000317, 0.00000811, -0.00000114, -0.00001463, -0.00000057, 0.00000685, -0.00000336, 0.00000184, -0.00000390, -0.00000134, -0.00000301, -0.00000303, -0.00000251, 0.00000468, 0.00000026, -0.00000649, 0.00000097, -0.00000253, -0.00000424, -0.00000150, -0.00000505, -0.00000219, -0.00000375, -0.00000430, -0.00000989, 0.00000706, -0.00000972, -0.00000569, -0.00002111, -0.00001468, 0.00000244, 0.00002271, },
		{ 0.00002651, 0.00001914, 0.00000789, 0.00001764, -0.00000031, 0.00001975, 0.00000369, 0.00002265, 0.00002250, 0.00000355, -0.00002910, -0.00001868, 0.00000052, -0.00001218, -0.00000396, -0.00000540, -0.00000382, -0.00000623, -0.00001504, -0.00000900, -0.00002541, -0.00001164, -0.00011919, -0.00000216, 0.00015821, 0.00002950, 0.00008421, 0.00003766, 0.00017214, -0.00006010, -0.00008609, -0.00001101, -0.00002131, -0.00001287, 0.00000119, -0.00000865, -0.00000456, -0.00000007, -0.00000874, 0.00000378, -0.00002495, 0.00000666, -0.00000640, 0.00000247, -0.00001062, 0.00000609, -0.00001162, -0.00000036, -0.00000055, 0.00000071, -0.00000956, 0.00000028, 0.00000133, -0.00000535, 0.00000428, -0.00000309, -0.00001095, 0.00000373, -0.00000566, 0.00000353, -0.00000353, -0.00000110, 0.00000785, -0.00000485, 0.00000869, -0.00000208, 0.00000098, -0.00000140, -0.00000145, 0.00000214, -0.00000381, 0.00000195, 0.00001007, -0.00000231, 0.00000112, -0.00000209, -0.00000087, 0.00000308, 0.00000107, -0.00000251, -0.00000361, -0.00000575, 0.00000670, 0.00000126, 0.00000827, -0.00000339, -0.00000092, -0.00000066, 0.00000830, -0.00000132, -0.00000537, 0.00000393, 0.00000117, 0.00000051, -0.00000154, 0.00000179, -0.00001214, 0.00000717, 0.00000658, -0.00000411, -0.00001438, 0.00000416, 0.00000339, -0.00000191, 0.00000030, -0.00000271, -0.00000202, -0.00000268, -0.00000333, -0.00000195, 0.00000528, -0.00000110, -0.00000670, 0.00000324, -0.00000456, -0.00000338, -0.00000266, -0.00000355, -0.00000394, -0.00000125, -0.00000711, -0.00000711, 0.00000584, -0.00001024, -0.00000540, -0.00002671, -0.00000635, 0.00000558, },
		{ 0.00000904, 0.00002717, 0.00001267, 0.00001086, 0.00001400, 0.00000071, 0.00001983, 0.00000180, 0.00002488, 0.00001816, 0.00000208, -0.00003165, -0.00001214, -0.00000251, -0.00001080, -0.00000470, -0.00000380, -0.00000559, -0.00000587, -0.00001709, -0.00000738, -0.00002805, -0.00001326, -0.00013540, 0.00003543, 0.00014862, 0.00003696, 0.00007780, 0.00006542, 0.00016345, -0.00011945, -0.00004865, -0.00002212, -0.00001160, -0.00001917, 0.00000893, -0.00001679, 0.00000374, -0.00000697, -0.00000242, -0.00000372, -0.00001924, 0.00000464, -0.00000415, -0.00000103, -0.00000774, 0.00000533, -0.00001219, 0.00000022, 0.00000012, -0.00000087, -0.00000951, 0.00000162, 0.00000064, -0.00000528, 0.00000650, -0.00000876, -0.00000666, 0.00000294, -0.00000663, 0.00000692, -0.00000835, 0.00000487, 0.00000304, -0.00000220, 0.00000782, -0.00000172, 0.00000081, -0.00000217, -0.00000093, 0.00000200, -0.00000484, 0.00000560, 0.00000788, -0.00000316, 0.00000180, -0.00000340, 0.00000053, 0.00000336, -0.00000009, -0.00000215, -0.00000463, -0.00000432, 0.00000678, 0.00000260, 0.00000703, -0.00000505, 0.00000014, 0.00000024, 0.00000924, -0.00000541, -0.00000171, 0.00000190, 0.00000258, -0.00000077, 0.00000021, -0.00000151, -0.00001099, 0.00001032, 0.00000429, -0.00000708, -0.00001213, 0.00000708, 0.00000024, -0.00000027, -0.00000148, -0.00000090, -0.00000359, -0.00000147, -0.00000443, -0.00000063, 0.00000524, -0.00000241, -0.00000633, 0.00000530, -0.00000727, -0.00000180, -0.00000415, -0.00000178, -0.00000610, 0.00000221, -0.00001171, -0.00000177, 0.00000131, -0.00000680, -0.00001017, -0.00002730, 0.00000081, },
		{ 0.00000501, 0.00001367, 0.00002410, 0.00000929, 0.00001253, 0.00001088, 0.00000198, 0.00001957, 0.00000004, 0.00002701, 0.00001366, 0.00000090, -0.00003361, -0.00000545, -0.00000695, -0.00000797, -0.00000643, -0.00000147, -0.00000832, -0.00000479, -0.00001951, -0.00000541, -0.00003098, -0.00001660, -0.00014920, 0.00007645, 0.00013280, 0.00005007, 0.00006660, 0.00010726, 0.00012139, -0.00015332, -0.00001444, -0.00003531, -0.00000091, -0.00002597, 0.00001678, -0.00002530, 0.00001287, -0.00001486, 0.00000450, -0.00001198, -0.00001187, 0.00000084, -0.00000070, -0.00000588, -0.00000273, 0.00000234, -0.00001124, -0.00000004, 0.00000143, -0.00000343, -0.00000818, 0.00000209, 0.00000023, -0.00000487, 0.00000788, -0.00001479, 0.00000038, -0.00000202, -0.00000302, 0.00000607, -0.00000996, 0.00000907, -0.00000162, 0.00000124, 0.00000610, -0.00000054, -0.00000037, -0.00000221, -0.00000061, 0.00000164, -0.00000526, 0.00000934, 0.00000417, -0.00000226, 0.00000119, -0.00000382, 0.00000179, 0.00000339, -0.00000131, -0.00000152, -0.00000594, -0.00000239, 0.00000654, 0.00000435, 0.00000434, -0.00000513, 0.00000033, 0.00000249, 0.00000778, -0.00000746, 0.00000112, -0.00000009, 0.00000427, -0.00000223, 0.00000191, -0.00000556, -0.00000784, 0.00001204, 0.00000153, -0.00000957, -0.00000830, 0.00000775, -0.00000169, 0.00000066, -0.00000259, 0.00000059, -0.00000523, 0.00000003, -0.00000590, 0.00000116, 0.00000475, -0.00000382, -0.00000493, 0.00000627, -0.00000970, -0.00000025, -0.00000521, -0.00000077, -0.00000723, 0.00000482, -0.00001611, 0.00000416, -0.00000468, -0.00000101, -0.00001920, -0.00002199, },
		{ -0.00001235, 0.00000635, 0.00001899, 0.00001801, 0.00000918, 0.00001232, 0.00000882, 0.00000317, 0.00001895, -0.00000145, 0.00002885, 0.00000930, -0.00000048, -0.00003423, 0.00000013, -0.00001148, -0.00000476, -0.00000819, 0.00000050, -0.00001081, -0.00000424, -0.00002086, -0.00000479, -0.00003238, -0.00002392, -0.00015710, 0.00011675, 0.00011495, 0.00006398, 0.00005831, 0.00015259, 0.00005032, -0.00015407, 0.00000509, -0.00004218, 0.00000479, -0.00002845, 0.00002053, -0.00003038, 0.00001953, -0.00002098, 0.00000950, -0.00001839, -0.00000540, -0.00000251, 0.00000189, -0.00000995, 0.00000248, -0.00000148, -0.00000981, -0.00000015, 0.00000223, -0.00000576, -0.00000660, 0.00000251, -0.00000064, -0.00000319, 0.00000683, -0.00001866, 0.00000738, -0.00000894, 0.00000405, 0.00000095, -0.00000735, 0.00000985, -0.00000432, 0.00000392, 0.00000471, 0.00000048, -0.00000188, -0.00000184, -0.00000032, 0.00000083, -0.00000450, 0.00001213, -0.00000004, -0.00000016, -0.00000063, -0.00000315, 0.00000267, 0.00000320, -0.00000241, -0.00000087, -0.00000727, -0.00000023, 0.00000634, 0.00000583, 0.00000095, -0.00000404, 0.00000003, 0.00000555, 0.00000408, -0.00000687, 0.00000211, -0.00000100, 0.00000536, -0.00000320, 0.00000281, -0.00000938, -0.00000324, 0.00001208, -0.00000141, -0.00001108, -0.00000367, 0.00000637, -0.00000205, 0.00000034, -0.00000234, 0.00000086, -0.00000603, 0.00000100, -0.00000707, 0.00000290, 0.00000412, -0.00000547, -0.00000221, 0.00000539, -0.00001086, 0.00000043, -0.00000513, -0.00000141, -0.00000603, 0.00000470, -0.00001795, 0.00000798, -0.00000911, 0.00000385, -0.00002941, },
		{ -0.00003656, -0.00000181, 0.00000666, 0.00002327, 0.00001096, 0.00001109, 0.00001075, 0.00000751, 0.00000458, 0.00001748, -0.00000204, 0.00002986, 0.00000562, -0.00000255, -0.00003276, 0.00000333, -0.00001458, -0.00000253, -0.00000874, 0.00000081, -0.00001163, -0.00000564, -0.00001955, -0.00000740, -0.00003029, -0.00003763, -0.00015542, 0.00015184, 0.00009967, 0.00007341, 0.00006198, 0.00018564, -0.00003558, -0.00012435, 0.00000522, -0.00003787, 0.00000171, -0.00002340, 0.00001721, -0.00002915, 0.00002108, -0.00002311, 0.00001069, -0.00002098, -0.00000189, -0.00000351, 0.00000182, -0.00001128, 0.00000591, -0.00000457, -0.00000900, 0.00000084, 0.00000147, -0.00000669, -0.00000585, 0.00000369, -0.00000257, 0.00000030, 0.00000227, -0.00001813, 0.00001114, -0.00001450, 0.00001174, -0.00000654, -0.00000133, 0.00000703, -0.00000424, 0.00000491, 0.00000456, 0.00000044, -0.00000294, -0.00000149, 0.00000011, -0.00000050, -0.00000223, 0.00001302, -0.00000352, 0.00000217, -0.00000306, -0.00000158, 0.00000310, 0.00000275, -0.00000313, -0.00000054, -0.00000828, 0.00000184, 0.00000654, 0.00000633, -0.00000217, -0.00000255, -0.00000000, 0.00000830, -0.00000077, -0.00000405, 0.00000100, -0.00000026, 0.00000527, -0.00000317, 0.00000223, -0.00001192, 0.00000187, 0.00001058, -0.00000431, -0.00001119, 0.00000069, 0.00000371, -0.00000110, -0.00000114, -0.00000062, -0.00000046, -0.00000542, 0.00000083, -0.00000740, 0.00000418, 0.00000353, -0.00000723, 0.00000164, 0.00000246, -0.00001026, -0.00000015, -0.00000369, -0.00000386, -0.00000209, 0.00000089, -0.00001567, 0.00000755, -0.00000922, 0.00000418, },
		{ -0.00000210, -0.00003716, 0.00000596, 0.00000837, 0.00002427, 0.00000554, 0.00001299, 0.00000915, 0.00000609, 0.00000685, 0.00001442, -0.00000094, 0.00002936, 0.00000309, -0.00000559, -0.00002896, 0.00000345, -0.00001521, -0.00000227, -0.00000727, -0.00000137, -0.00000996, -0.00000976, -0.00001480, -0.00001420, -0.00002372, -0.00005892, -0.00014178, 0.00017862, 0.00008982, 0.00007563, 0.00008382, 0.00019052, -0.00011597, -0.00007632, -0.00001058, -0.00002325, -0.00000967, -0.00001102, 0.00000674, -0.00002115, 0.00001683, -0.00002075, 0.00000781, -0.00001939, -0.00000192, -0.00000165, -0.00000141, -0.00000906, 0.00000647, -0.00000602, -0.00000931, 0.00000321, -0.00000115, -0.00000573, -0.00000642, 0.00000591, -0.00000553, 0.00000525, -0.00000561, -0.00001234, 0.00000947, -0.00001554, 0.00001662, -0.00001334, 0.00000578, 0.00000194, -0.00000185, 0.00000421, 0.00000579, -0.00000107, -0.00000304, -0.00000148, 0.00000070, -0.00000214, 0.00000141, 0.00001158, -0.00000528, 0.00000368, -0.00000526, 0.00000040, 0.00000321, 0.00000200, -0.00000332, -0.00000076, -0.00000865, 0.00000361, 0.00000730, 0.00000537, -0.00000419, -0.00000144, 0.00000102, 0.00000942, -0.00000515, -0.00000028, -0.00000156, 0.00000194, 0.00000397, -0.00000214, -0.00000002, -0.00001234, 0.00000641, 0.00000795, -0.00000693, -0.00000968, 0.00000378, 0.00000089, 0.00000034, -0.00000309, 0.00000200, -0.00000308, -0.00000344, -0.00000060, -0.00000671, 0.00000488, 0.00000289, -0.00000860, 0.00000584, -0.00000195, -0.00000813, -0.00000170, -0.00000147, -0.00000726, 0.00000366, -0.00000594, -0.00000945, 0.00000256, -0.00000403, },
		{ 0.00000484, -0.00001432, -0.00003009, 0.00000880, 0.00001308, 0.00002055, 0.00000380, 0.00001301, 0.00000880, 0.00000376, 0.00001043, 0.00000933, 0.00000240, 0.00002690, 0.00000191, -0.00000936, -0.00002331, 0.00000074, -0.00001327, -0.00000404, -0.00000394, -0.00000577, -0.00000611, -0.00001606, -0.00000732, -0.00002447, -0.00001358, -0.00008684, -0.00011621, 0.00019664, 0.00008498, 0.00007276, 0.00012300, 0.00015744, -0.00017096, -0.00002714, -0.00003267, -0.00000432, -0.00002475, 0.00000489, -0.00000780, -0.00000868, 0.00000848, -0.00001540, 0.00000248, -0.00001516, -0.00000425, 0.00000193, -0.00000664, -0.00000407, 0.00000445, -0.00000604, -0.00001027, 0.00000622, -0.00000489, -0.00000344, -0.00000785, 0.00000859, -0.00000863, 0.00001020, -0.00001503, -0.00000245, 0.00000222, -0.00001059, 0.00001617, -0.00001646, 0.00001117, -0.00000317, 0.00000139, 0.00000279, 0.00000777, -0.00000375, -0.00000213, -0.00000187, 0.00000129, -0.00000357, 0.00000573, 0.00000805, -0.00000493, 0.00000366, -0.00000647, 0.00000224, 0.00000316, 0.00000097, -0.00000297, -0.00000162, -0.00000829, 0.00000507, 0.00000840, 0.00000299, -0.00000472, -0.00000109, 0.00000345, 0.00000799, -0.00000746, 0.00000280, -0.00000430, 0.00000477, 0.00000202, -0.00000062, -0.00000353, -0.00001040, 0.00000948, 0.00000471, -0.00000905, -0.00000668, 0.00000494, -0.00000104, 0.00000126, -0.00000448, 0.00000445, -0.00000610, -0.00000072, -0.00000289, -0.00000526, 0.00000521, 0.00000188, -0.00000885, 0.00000919, -0.00000665, -0.00000538, -0.00000335, 0.00000032, -0.00000991, 0.00000906, -0.00001352, -0.00000146, -0.00000504, },
		{ -0.00001168, 0.00001358, -0.00002896, -0.00001732, 0.00000705, 0.00002047, 0.00001240, 0.00000628, 0.00001038, 0.00001020, 0.00000039, 0.00001512, 0.00000251, 0.00000786, 0.00002257, 0.00000180, -0.00001313, -0.00001709, -0.00000355, -0.00000977, -0.00000687, 0.00000010, -0.00001102, -0.00000160, -0.00002270, 0.00000066, -0.00003577, -0.00000275, -0.00011812, -0.00008128, 0.00020817, 0.00008147, 0.00007212, 0.00016973, 0.00008770, -0.00018806, 0.00000747, -0.00004974, 0.00001083, -0.00003693, 0.00001863, -0.00002145, 0.00000415, -0.00000077, -0.00000981, -0.00000255, -0.00001096, -0.00000646, 0.00000492, -0.00001152, 0.00000161, 0.00000133, -0.00000575, -0.00001065, 0.00000842, -0.00000819, -0.00000122, -0.00000893, 0.00001052, -0.00001042, 0.00001304, -0.00002303, 0.00000846, -0.00000836, -0.00000055, 0.00000989, -0.00001433, 0.00001267, -0.00000616, 0.00000371, 0.00000201, 0.00000929, -0.00000675, -0.00000062, -0.00000244, 0.00000157, -0.00000414, 0.00000966, 0.00000339, -0.00000283, 0.00000202, -0.00000631, 0.00000354, 0.00000307, -0.00000020, -0.00000227, -0.00000294, -0.00000731, 0.00000643, 0.00000933, -0.00000028, -0.00000396, -0.00000129, 0.00000688, 0.00000397, -0.00000681, 0.00000382, -0.00000586, 0.00000710, 0.00000029, 0.00000044, -0.00000728, -0.00000656, 0.00001068, 0.00000132, -0.00001032, -0.00000277, 0.00000409, -0.00000149, 0.00000095, -0.00000443, 0.00000564, -0.00000841, 0.00000179, -0.00000530, -0.00000355, 0.00000553, 0.00000018, -0.00000736, 0.00001054, -0.00001026, -0.00000319, -0.00000403, 0.00000043, -0.00001003, 0.00001164, -0.00001886, 0.00000502, },
		{ 0.00000703, -0.00001375, 0.00001771, -0.00004091, -0.00000312, 0.00000337, 0.00002824, 0.00000202, 0.00001171, 0.00000577, 0.00001278, -0.00000330, 0.00001987, -0.00000496, 0.00001459, 0.00001697, 0.00000215, -0.00001592, -0.00001184, -0.00000763, -0.00000634, -0.00000920, 0.00000308, -0.00001518, 0.00000149, -0.00002724, 0.00000619, -0.00004474, 0.00000486, -0.00014824, -0.00004089, 0.00021668, 0.00007438, 0.00008372, 0.00020684, -0.00000506, -0.00016675, 0.00001904, -0.00005391, 0.00001595, -0.00004079, 0.00002524, -0.00002961, 0.00001337, -0.00000775, -0.00000653, -0.00000482, -0.00000927, -0.00000625, 0.00000504, -0.00001362, 0.00000564, -0.00000108, -0.00000650, -0.00000918, 0.00000835, -0.00000944, -0.00000061, -0.00000833, 0.00001048, -0.00000963, 0.00001187, -0.00002656, 0.00001653, -0.00001838, 0.00001143, -0.00000037, -0.00000746, 0.00000974, -0.00000589, 0.00000392, 0.00000302, 0.00000917, -0.00000902, 0.00000085, -0.00000289, 0.00000132, -0.00000330, 0.00001206, -0.00000111, 0.00000008, -0.00000076, -0.00000484, 0.00000418, 0.00000292, -0.00000127, -0.00000155, -0.00000436, -0.00000598, 0.00000801, 0.00000938, -0.00000350, -0.00000260, -0.00000131, 0.00001019, -0.00000164, -0.00000342, 0.00000227, -0.00000539, 0.00000806, -0.00000044, 0.00000012, -0.00000997, -0.00000186, 0.00001012, -0.00000194, -0.00001035, 0.00000113, 0.00000182, -0.00000057, -0.00000066, -0.00000265, 0.00000495, -0.00000917, 0.00000325, -0.00000717, -0.00000202, 0.00000601, -0.00000217, -0.00000404, 0.00000928, -0.00001184, -0.00000236, -0.00000317, -0.00000169, -0.00000671, 0.00000978, -0.00001959, },
		{ -0.00001524, 0.00000347, -0.00000938, 0.00001414, -0.00004567, 0.00000777, 0.00000143, 0.00003310, -0.00000730, 0.00001755, 0.00000091, 0.00001515, -0.00000600, 0.00002321, -0.00001151, 0.00002129, 0.00001099, 0.00000242, -0.00001706, -0.00000878, -0.00000997, -0.00000445, -0.00000968, 0.00000359, -0.00001672, 0.00000154, -0.00002784, 0.00000697, -0.00004867, 0.00000586, -0.00017317, 0.00000167, 0.00022440, 0.00006092, 0.00011549, 0.00021520, -0.00009826, -0.00011866, 0.00000897, -0.00004401, 0.00000964, -0.00003475, 0.00002292, -0.00003024, 0.00001698, -0.00001098, -0.00000651, -0.00000348, -0.00001106, -0.00000267, 0.00000132, -0.00001163, 0.00000653, -0.00000162, -0.00000897, -0.00000537, 0.00000543, -0.00000783, -0.00000246, -0.00000533, 0.00000799, -0.00000597, 0.00000608, -0.00002384, 0.00001858, -0.00002380, 0.00002119, -0.00001109, 0.00000176, 0.00000359, -0.00000268, 0.00000192, 0.00000613, 0.00000676, -0.00000974, 0.00000171, -0.00000299, 0.00000051, -0.00000092, 0.00001221, -0.00000421, 0.00000260, -0.00000378, -0.00000257, 0.00000425, 0.00000265, -0.00000203, -0.00000116, -0.00000553, -0.00000457, 0.00001001, 0.00000797, -0.00000574, -0.00000150, -0.00000027, 0.00001190, -0.00000706, 0.00000144, -0.00000131, -0.00000291, 0.00000744, 0.00000006, -0.00000204, -0.00001052, 0.00000241, 0.00000831, -0.00000496, -0.00000884, 0.00000397, -0.00000080, 0.00000098, -0.00000296, 0.00000041, 0.00000251, -0.00000820, 0.00000330, -0.00000816, -0.00000082, 0.00000650, -0.00000467, 0.00000043, 0.00000574, -0.00001134, -0.00000287, -0.00000111, -0.00000542, -0.00000055, 0.00000362, },
		{ -0.00000470, -0.00000765, -0.00000416, 0.00000041, 0.00000278, -0.00004127, 0.00001223, 0.00000399, 0.00003235, -0.00001248, 0.00002121, -0.00000228, 0.00001592, -0.00000655, 0.00002385, -0.00001571, 0.00002675, 0.00000532, 0.00000245, -0.00001663, -0.00000813, -0.00001002, -0.00000462, -0.00000794, 0.00000131, -0.00001528, -0.00000166, -0.00002429, 0.00000248, -0.00004673, -0.00000108, -0.00019088, 0.00004546, 0.00023021, 0.00004369, 0.00016792, 0.00018119, -0.00016769, -0.00006310, -0.00001286, -0.00002561, -0.00000409, -0.00002183, 0.00001380, -0.00002463, 0.00001577, -0.00001129, -0.00000855, 0.00000016, -0.00001519, 0.00000319, -0.00000525, -0.00000613, 0.00000436, -0.00000041, -0.00001266, 0.00000002, 0.00000035, -0.00000386, -0.00000642, -0.00000037, 0.00000369, -0.00000051, -0.00000310, -0.00001530, 0.00001358, -0.00002213, 0.00002520, -0.00001851, 0.00000999, -0.00000328, 0.00000188, -0.00000126, 0.00001066, 0.00000225, -0.00000866, 0.00000168, -0.00000272, -0.00000051, 0.00000258, 0.00001001, -0.00000519, 0.00000375, -0.00000608, -0.00000013, 0.00000399, 0.00000220, -0.00000237, -0.00000126, -0.00000621, -0.00000317, 0.00001233, 0.00000496, -0.00000640, -0.00000118, 0.00000236, 0.00001082, -0.00001038, 0.00000586, -0.00000551, 0.00000071, 0.00000577, 0.00000130, -0.00000572, -0.00000849, 0.00000518, 0.00000593, -0.00000761, -0.00000575, 0.00000493, -0.00000263, 0.00000212, -0.00000488, 0.00000366, -0.00000086, -0.00000599, 0.00000220, -0.00000840, 0.00000028, 0.00000654, -0.00000639, 0.00000476, 0.00000111, -0.00000958, -0.00000392, 0.00000097, -0.00000906, 0.00000632, },
		{ 0.00001105, -0.00001190, -0.00000029, -0.00001217, 0.00001182, -0.00001309, -0.00002932, 0.00001014, 0.00001153, 0.00002526, -0.00001204, 0.00002112, -0.00000269, 0.00001438, -0.00000452, 0.00002129, -0.00001679, 0.00003026, 0.00000023, 0.00000267, -0.00001558, -0.00000897, -0.00000847, -0.00000617, -0.00000487, -0.00000274, -0.00001203, -0.00000668, -0.00001826, -0.00000562, -0.00004049, -0.00001485, -0.00020185, 0.00009227, 0.00022921, 0.00003171, 0.00023080, 0.00010337, -0.00019611, -0.00001953, -0.00003316, -0.00000793, -0.00001794, -0.00000807, 0.00000285, -0.00001668, 0.00001269, -0.00001128, -0.00000985, 0.00000319, -0.00001888, 0.00000867, -0.00001205, 0.00000069, 0.00000074, 0.00000122, -0.00001604, 0.00000522, -0.00000500, 0.00000078, -0.00001098, 0.00000511, -0.00000077, 0.00000450, -0.00001279, -0.00000369, 0.00000324, -0.00001351, 0.00002201, -0.00002011, 0.00001437, -0.00000822, 0.00000567, -0.00000393, 0.00001517, -0.00000340, -0.00000611, 0.00000084, -0.00000228, -0.00000117, 0.00000628, 0.00000614, -0.00000410, 0.00000313, -0.00000704, 0.00000191, 0.00000363, 0.00000162, -0.00000240, -0.00000177, -0.00000644, -0.00000163, 0.00001449, 0.00000078, -0.00000553, -0.00000160, 0.00000635, 0.00000662, -0.00001032, 0.00000808, -0.00000867, 0.00000412, 0.00000410, 0.00000220, -0.00000981, -0.00000433, 0.00000594, 0.00000349, -0.00000964, -0.00000149, 0.00000374, -0.00000290, 0.00000198, -0.00000538, 0.00000586, -0.00000395, -0.00000351, 0.00000059, -0.00000835, 0.00000167, 0.00000560, -0.00000643, 0.00000749, -0.00000303, -0.00000788, -0.00000429, 0.00000159, -0.00001049, },
		{ -0.00000833, 0.00001142, -0.00001497, 0.00000327, -0.00001638, 0.00001988, -0.00002805, -0.00001431, 0.00000434, 0.00002185, 0.00001350, -0.00000661, 0.00001739, -0.00000042, 0.00001085, -0.00000050, 0.00001604, -0.00001489, 0.00003171, -0.00000437, 0.00000379, -0.00001532, -0.00000966, -0.00000686, -0.00000757, -0.00000218, -0.00000665, -0.00000898, -0.00001117, -0.00001256, -0.00001432, -0.00003303, -0.00003269, -0.00020794, 0.00014508, 0.00021481, 0.00003808, 0.00028411, -0.00000471, -0.00017972, -0.00000015, -0.00004142, 0.00000092, -0.00002507, 0.00000053, -0.00000467, -0.00001069, 0.00001104, -0.00001369, -0.00000763, 0.00000265, -0.00001922, 0.00001082, -0.00001605, 0.00000598, -0.00000213, 0.00000148, -0.00001730, 0.00000817, -0.00000842, 0.00000395, -0.00001421, 0.00000929, -0.00000354, 0.00000663, -0.00001948, 0.00000683, -0.00000853, -0.00000081, 0.00001290, -0.00001565, 0.00001362, -0.00000952, 0.00000700, -0.00000448, 0.00001801, -0.00000880, -0.00000290, -0.00000043, -0.00000198, -0.00000086, 0.00000905, 0.00000182, -0.00000169, 0.00000101, -0.00000651, 0.00000324, 0.00000325, 0.00000105, -0.00000239, -0.00000240, -0.00000646, 0.00000042, 0.00001572, -0.00000364, -0.00000377, -0.00000207, 0.00001063, 0.00000014, -0.00000677, 0.00000718, -0.00000956, 0.00000617, 0.00000345, 0.00000158, -0.00001278, 0.00000075, 0.00000490, 0.00000121, -0.00001062, 0.00000304, 0.00000090, -0.00000160, 0.00000032, -0.00000397, 0.00000619, -0.00000576, -0.00000166, -0.00000085, -0.00000840, 0.00000357, 0.00000353, -0.00000438, 0.00000767, -0.00000545, -0.00000730, -0.00000315, -0.00000015, },
		{ -0.00000394, -0.00000276, 0.00000699, -0.00001274, 0.00000118, -0.00001409, 0.00002061, -0.00003681, -0.00000168, -0.00000075, 0.00003114, 0.00000054, 0.00000143, 0.00001152, 0.00000334, 0.00000650, 0.00000422, 0.00000935, -0.00001075, 0.00003131, -0.00000854, 0.00000623, -0.00001694, -0.00000871, -0.00000669, -0.00000738, -0.00000144, -0.00000873, -0.00000790, -0.00001312, -0.00000966, -0.00002085, -0.00002726, -0.00005219, -0.00021014, 0.00020517, 0.00018283, 0.00007449, 0.00030367, -0.00011728, -0.00012975, -0.00000565, -0.00003452, -0.00000205, -0.00002256, 0.00000108, -0.00000594, -0.00000907, 0.00001250, -0.00001960, -0.00000081, -0.00000281, -0.00001470, 0.00000800, -0.00001528, 0.00000771, -0.00000262, -0.00000076, -0.00001540, 0.00000768, -0.00000848, 0.00000411, -0.00001468, 0.00001096, -0.00000358, 0.00000445, -0.00002059, 0.00001240, -0.00001713, 0.00001152, 0.00000132, -0.00000717, 0.00000849, -0.00000703, 0.00000533, -0.00000210, 0.00001803, -0.00001268, 0.00000008, -0.00000172, -0.00000198, 0.00000076, 0.00000994, -0.00000162, 0.00000084, -0.00000180, -0.00000486, 0.00000387, 0.00000287, 0.00000068, -0.00000265, -0.00000277, -0.00000658, 0.00000330, 0.00001525, -0.00000720, -0.00000204, -0.00000161, 0.00001358, -0.00000678, -0.00000091, 0.00000347, -0.00000792, 0.00000636, 0.00000430, -0.00000131, -0.00001319, 0.00000519, 0.00000277, -0.00000098, -0.00001005, 0.00000667, -0.00000252, 0.00000059, -0.00000231, -0.00000101, 0.00000460, -0.00000593, -0.00000091, -0.00000177, -0.00000862, 0.00000577, 0.00000079, -0.00000073, 0.00000536, -0.00000579, -0.00000807, -0.00000057, },
		{ 0.00000229, -0.00000813, 0.00000428, -0.00000045, -0.00000655, -0.00000565, -0.00000558, 0.00001308, -0.00003658, 0.00000461, -0.00000136, 0.00003569, -0.00000993, 0.00000913, 0.00000557, 0.00000707, 0.00000260, 0.00000829, 0.00000253, -0.00000522, 0.00002920, -0.00001189, 0.00000955, -0.00002046, -0.00000570, -0.00000844, -0.00000520, -0.00000306, -0.00000856, -0.00000909, -0.00001221, -0.00001017, -0.00002436, -0.00002408, -0.00007314, -0.00020637, 0.00026916, 0.00013585, 0.00014435, 0.00027005, -0.00020442, -0.00006813, -0.00002588, -0.00001768, -0.00001328, -0.00001294, -0.00000471, -0.00000202, -0.00001107, 0.00001602, -0.00002749, 0.00000907, -0.00001199, -0.00000625, 0.00000087, -0.00000993, 0.00000563, -0.00000061, -0.00000525, -0.00001081, 0.00000410, -0.00000524, 0.00000113, -0.00001226, 0.00001021, -0.00000134, -0.00000154, -0.00001570, 0.00001109, -0.00001920, 0.00001921, -0.00000853, 0.00000189, 0.00000135, -0.00000214, 0.00000143, 0.00000284, 0.00001493, -0.00001428, 0.00000217, -0.00000277, -0.00000216, 0.00000350, 0.00000860, -0.00000322, 0.00000237, -0.00000429, -0.00000273, 0.00000400, 0.00000245, 0.00000057, -0.00000335, -0.00000265, -0.00000692, 0.00000703, 0.00001269, -0.00000905, -0.00000111, 0.00000056, 0.00001367, -0.00001189, 0.00000518, -0.00000163, -0.00000454, 0.00000516, 0.00000624, -0.00000631, -0.00001037, 0.00000760, 0.00000049, -0.00000322, -0.00000761, 0.00000832, -0.00000519, 0.00000247, -0.00000475, 0.00000243, 0.00000185, -0.00000485, -0.00000109, -0.00000236, -0.00000863, 0.00000762, -0.00000160, 0.00000324, 0.00000164, -0.00000478, -0.00000944, },
		{ -0.00001005, 0.00000376, -0.00001041, 0.00000974, -0.00000752, 0.00000015, -0.00001371, 0.00000569, 0.00000002, -0.00002836, 0.00000380, 0.00000387, 0.00003368, -0.00001547, 0.00001432, 0.00000107, 0.00000978, -0.00000017, 0.00001100, -0.00000361, 0.00000121, 0.00002512, -0.00001334, 0.00001228, -0.00002453, -0.00000165, -0.00001116, -0.00000210, -0.00000586, -0.00000740, -0.00001099, -0.00001027, -0.00001231, -0.00002658, -0.00002176, -0.00009823, -0.00019076, 0.00032803, 0.00008540, 0.00023777, 0.00017746, -0.00024256, -0.00001894, -0.00004522, -0.00000142, -0.00002460, -0.00000282, -0.00001144, 0.00000280, -0.00001316, 0.00001824, -0.00003371, 0.00001829, -0.00002147, 0.00000301, -0.00000772, -0.00000228, 0.00000139, 0.00000245, -0.00001030, -0.00000538, -0.00000068, -0.00000037, -0.00000362, -0.00000820, 0.00000841, 0.00000122, -0.00000886, -0.00000698, 0.00000388, -0.00001405, 0.00001995, -0.00001341, 0.00000821, -0.00000490, 0.00000292, -0.00000296, 0.00000902, 0.00000943, -0.00001357, 0.00000312, -0.00000354, -0.00000212, 0.00000668, 0.00000542, -0.00000273, 0.00000224, -0.00000565, -0.00000076, 0.00000388, 0.00000202, 0.00000062, -0.00000437, -0.00000209, -0.00000730, 0.00001122, 0.00000824, -0.00000895, -0.00000117, 0.00000449, 0.00001019, -0.00001345, 0.00000932, -0.00000620, -0.00000090, 0.00000373, 0.00000806, -0.00001223, -0.00000466, 0.00000730, -0.00000123, -0.00000547, -0.00000344, 0.00000744, -0.00000607, 0.00000296, -0.00000577, 0.00000496, -0.00000078, -0.00000350, -0.00000152, -0.00000312, -0.00000786, 0.00000832, -0.00000245, 0.00000585, -0.00000176, -0.00000386, },
		{ -0.00000435, -0.00000874, 0.00000255, -0.00000892, 0.00001091, -0.00001080, 0.00000344, -0.00001849, 0.00001457, -0.00001320, -0.00001652, -0.00000141, 0.00001324, 0.00002598, -0.00001588, 0.00001640, -0.00000160, 0.00001143, -0.00000202, 0.00001253, -0.00000897, 0.00000850, 0.00001855, -0.00001156, 0.00001234, -0.00002690, 0.00000133, -0.00001284, -0.00000022, -0.00000759, -0.00000764, -0.00001086, -0.00001049, -0.00001268, -0.00003095, -0.00001683, -0.00013176, -0.00015538, 0.00036921, 0.00005013, 0.00033107, 0.00003873, -0.00022341, 0.00000098, -0.00005005, 0.00000402, -0.00002759, 0.00000059, -0.00001292, 0.00000332, -0.00001108, 0.00001531, -0.00003422, 0.00002260, -0.00002714, 0.00000917, -0.00001399, 0.00000429, -0.00000235, 0.00000433, -0.00001366, -0.00000159, -0.00000408, 0.00000364, -0.00000793, -0.00000445, 0.00000754, 0.00000156, -0.00001397, 0.00000152, -0.00000563, -0.00000400, 0.00001427, -0.00001229, 0.00000986, -0.00000808, 0.00000609, -0.00000594, 0.00001461, 0.00000287, -0.00001116, 0.00000306, -0.00000417, -0.00000135, 0.00000926, 0.00000143, -0.00000069, 0.00000049, -0.00000556, 0.00000063, 0.00000368, 0.00000172, 0.00000052, -0.00000535, -0.00000145, -0.00000724, 0.00001501, 0.00000276, -0.00000737, -0.00000168, 0.00000925, 0.00000368, -0.00001089, 0.00001012, -0.00000865, 0.00000155, 0.00000337, 0.00000824, -0.00001725, 0.00000254, 0.00000452, -0.00000214, -0.00000736, 0.00000175, 0.00000423, -0.00000486, 0.00000155, -0.00000462, 0.00000551, -0.00000212, -0.00000288, -0.00000146, -0.00000456, -0.00000589, 0.00000743, -0.00000102, 0.00000584, -0.00000332, },
		{ -0.00000244, -0.00000673, -0.00000532, -0.00000138, -0.00000339, 0.00000682, -0.00000850, 0.00000084, -0.00001667, 0.00001654, -0.00002092, -0.00000680, -0.00000630, 0.00002285, 0.00001564, -0.00001289, 0.00001625, -0.00000310, 0.00001271, -0.00000383, 0.00001366, -0.00001389, 0.00001669, 0.00000925, -0.00000563, 0.00000805, -0.00002548, 0.00000119, -0.00001151, -0.00000158, -0.00000616, -0.00001135, -0.00000638, -0.00001566, -0.00000824, -0.00004051, -0.00000661, -0.00017667, -0.00009402, 0.00038159, 0.00004954, 0.00039248, -0.00011630, -0.00015749, -0.00001160, -0.00003509, -0.00000595, -0.00001806, -0.00000660, -0.00000552, -0.00000356, -0.00000241, 0.00000521, -0.00002680, 0.00001942, -0.00002629, 0.00000947, -0.00001506, 0.00000700, -0.00000334, 0.00000330, -0.00001369, -0.00000120, -0.00000412, 0.00000474, -0.00000994, -0.00000257, 0.00000896, -0.00000213, -0.00001391, 0.00000566, -0.00001272, 0.00000659, 0.00000527, -0.00000669, 0.00000701, -0.00000751, 0.00000635, -0.00000625, 0.00001805, -0.00000325, -0.00000796, 0.00000234, -0.00000478, 0.00000052, 0.00001025, -0.00000204, 0.00000181, -0.00000219, -0.00000424, 0.00000134, 0.00000342, 0.00000172, -0.00000010, -0.00000578, -0.00000121, -0.00000617, 0.00001738, -0.00000251, -0.00000527, -0.00000164, 0.00001321, -0.00000409, -0.00000514, 0.00000751, -0.00000832, 0.00000202, 0.00000488, 0.00000564, -0.00001947, 0.00000930, 0.00000021, -0.00000248, -0.00000828, 0.00000673, -0.00000027, -0.00000213, -0.00000138, -0.00000146, 0.00000383, -0.00000165, -0.00000355, -0.00000056, -0.00000678, -0.00000283, 0.00000519, 0.00000246, 0.00000298, },
		{ -0.00000171, 0.00000021, -0.00001030, -0.00000084, -0.00000653, 0.00000443, -0.00000107, -0.00000157, -0.00000719, -0.00000797, 0.00001004, -0.00001984, -0.00000363, -0.00000649, 0.00002856, 0.00000639, -0.00000908, 0.00001540, -0.00000439, 0.00001437, -0.00000636, 0.00001506, -0.00001852, 0.00002532, -0.00000209, 0.00000403, -0.00000082, -0.00001957, -0.00000287, -0.00000653, -0.00000674, -0.00000110, -0.00001880, 0.00000266, -0.00002620, 0.00000149, -0.00005545, 0.00000814, -0.00023124, -0.00000659, 0.00036166, 0.00009522, 0.00039252, -0.00024928, -0.00007058, -0.00004569, -0.00000572, -0.00002771, 0.00000142, -0.00002260, 0.00000957, -0.00001694, 0.00001176, -0.00001073, -0.00001253, 0.00000926, -0.00001905, 0.00000380, -0.00001042, 0.00000506, -0.00000104, -0.00000078, -0.00001045, -0.00000424, -0.00000048, 0.00000237, -0.00000913, -0.00000278, 0.00001247, -0.00000969, -0.00000781, 0.00000303, -0.00001359, 0.00001316, -0.00000281, 0.00000021, 0.00000169, -0.00000418, 0.00000404, -0.00000375, 0.00001857, -0.00000777, -0.00000484, 0.00000131, -0.00000533, 0.00000346, 0.00000910, -0.00000386, 0.00000355, -0.00000478, -0.00000231, 0.00000154, 0.00000309, 0.00000212, -0.00000147, -0.00000530, -0.00000174, -0.00000370, 0.00001756, -0.00000643, -0.00000359, -0.00000007, 0.00001457, -0.00001069, 0.00000172, 0.00000274, -0.00000581, 0.00000079, 0.00000804, 0.00000008, -0.00001772, 0.00001374, -0.00000429, -0.00000271, -0.00000759, 0.00001018, -0.00000459, 0.00000088, -0.00000465, 0.00000266, 0.00000065, 0.00000026, -0.00000531, 0.00000098, -0.00000931, 0.00000061, 0.00000264, 0.00000675, },
		{ 0.00000997, -0.00000629, 0.00000298, -0.00001347, 0.00000277, -0.00001033, 0.00001130, -0.00000925, 0.00000651, -0.00001719, 0.00000434, -0.00000249, -0.00001071, -0.00000804, -0.00000012, 0.00002799, 0.00000084, -0.00000643, 0.00001495, -0.00000592, 0.00001647, -0.00000959, 0.00001665, -0.00002229, 0.00003305, -0.00001355, 0.00001535, -0.00001251, -0.00001050, -0.00000960, 0.00000070, -0.00001411, 0.00000573, -0.00002766, 0.00001346, -0.00003913, 0.00001328, -0.00007187, 0.00002182, -0.00028736, 0.00009848, 0.00031774, 0.00018317, 0.00031582, -0.00032608, 0.00000584, -0.00008153, 0.00002482, -0.00005099, 0.00002250, -0.00004047, 0.00002660, -0.00003193, 0.00002686, -0.00002792, 0.00000429, -0.00000418, -0.00000860, -0.00000507, -0.00000227, 0.00000005, 0.00000312, -0.00000620, -0.00000581, -0.00000883, 0.00000520, -0.00000216, -0.00000666, -0.00000368, 0.00001607, -0.00001866, 0.00000231, -0.00000570, -0.00000712, 0.00001296, -0.00000644, 0.00000497, -0.00000333, -0.00000011, 0.00000061, 0.00000063, 0.00001636, -0.00001022, -0.00000236, 0.00000016, -0.00000555, 0.00000695, 0.00000595, -0.00000353, 0.00000365, -0.00000633, -0.00000049, 0.00000153, 0.00000271, 0.00000280, -0.00000354, -0.00000384, -0.00000304, 0.00000006, 0.00001533, -0.00000836, -0.00000284, 0.00000338, 0.00001223, -0.00001395, 0.00000722, -0.00000216, -0.00000262, -0.00000095, 0.00001163, -0.00000741, -0.00001200, 0.00001473, -0.00000783, -0.00000320, -0.00000499, 0.00001110, -0.00000734, 0.00000279, -0.00000679, 0.00000615, -0.00000263, 0.00000251, -0.00000735, 0.00000240, -0.00001132, 0.00000338, 0.00000120, },
		{ 0.00000201, 0.00001040, -0.00000879, 0.00000408, -0.00001456, 0.00000368, -0.00001036, 0.00001392, -0.00001372, 0.00001124, -0.00002415, 0.00001472, -0.00001542, 0.00000189, -0.00001715, 0.00001119, 0.00002168, -0.00000061, -0.00000539, 0.00001481, -0.00000708, 0.00001799, -0.00001241, 0.00001739, -0.00002389, 0.00003787, -0.00002231, 0.00002500, -0.00002370, -0.00000132, -0.00001608, 0.00000707, -0.00002033, 0.00001068, -0.00003363, 0.00002098, -0.00004887, 0.00002101, -0.00008271, 0.00002504, -0.00033185, 0.00020543, 0.00026958, 0.00029090, 0.00016927, -0.00032894, 0.00004648, -0.00009959, 0.00004191, -0.00006377, 0.00003490, -0.00005126, 0.00003784, -0.00004188, 0.00003685, -0.00004043, 0.00001789, -0.00001565, 0.00000031, -0.00001272, 0.00000550, -0.00000488, 0.00000642, -0.00001025, -0.00000267, -0.00001195, 0.00001004, -0.00000636, -0.00000477, -0.00000294, 0.00001675, -0.00002509, 0.00001216, -0.00001684, 0.00000447, 0.00000625, -0.00000429, 0.00000546, -0.00000576, 0.00000265, -0.00000217, 0.00000531, 0.00001241, -0.00001080, -0.00000069, -0.00000107, -0.00000500, 0.00001007, 0.00000166, -0.00000134, 0.00000191, -0.00000632, 0.00000067, 0.00000153, 0.00000238, 0.00000344, -0.00000588, -0.00000176, -0.00000469, 0.00000446, 0.00001121, -0.00000841, -0.00000278, 0.00000810, 0.00000632, -0.00001281, 0.00000952, -0.00000527, -0.00000042, -0.00000163, 0.00001388, -0.00001485, -0.00000360, 0.00001217, -0.00000983, -0.00000393, -0.00000082, 0.00000926, -0.00000776, 0.00000268, -0.00000668, 0.00000760, -0.00000452, 0.00000382, -0.00000862, 0.00000289, -0.00001202, 0.00000463, },
		{ 0.00000420, 0.00000530, 0.00000735, -0.00000810, 0.00000248, -0.00001278, 0.00000122, -0.00000567, 0.00001055, -0.00001187, 0.00000932, -0.00002386, 0.00001782, -0.00002237, 0.00001146, -0.00002548, 0.00002313, 0.00001274, 0.00000030, -0.00000486, 0.00001388, -0.00000653, 0.00001729, -0.00001310, 0.00001578, -0.00002184, 0.00003784, -0.00002574, 0.00002964, -0.00003078, 0.00000448, -0.00001895, 0.00000915, -0.00002183, 0.00000997, -0.00003244, 0.00002011, -0.00004963, 0.00001828, -0.00008072, 0.00000832, -0.00035081, 0.00029663, 0.00024222, 0.00038200, -0.00001710, -0.00026326, 0.00004161, -0.00008925, 0.00003664, -0.00005819, 0.00003148, -0.00004838, 0.00003737, -0.00004170, 0.00003714, -0.00004367, 0.00002364, -0.00002076, 0.00000357, -0.00001516, 0.00000920, -0.00000670, 0.00000639, -0.00001061, -0.00000346, -0.00001097, 0.00001136, -0.00000783, -0.00000546, 0.00000134, 0.00001203, -0.00002526, 0.00001688, -0.00002512, 0.00001651, -0.00000384, 0.00000228, 0.00000168, -0.00000477, 0.00000296, -0.00000296, 0.00000877, 0.00000805, -0.00001015, 0.00000029, -0.00000241, -0.00000332, 0.00001182, -0.00000251, 0.00000174, -0.00000115, -0.00000481, 0.00000098, 0.00000162, 0.00000231, 0.00000358, -0.00000786, 0.00000029, -0.00000599, 0.00000846, 0.00000631, -0.00000735, -0.00000252, 0.00001252, -0.00000160, -0.00000779, 0.00000811, -0.00000557, -0.00000032, -0.00000013, 0.00001321, -0.00001998, 0.00000537, 0.00000699, -0.00001037, -0.00000447, 0.00000399, 0.00000533, -0.00000604, 0.00000055, -0.00000408, 0.00000639, -0.00000414, 0.00000327, -0.00000841, 0.00000198, -0.00001112, },
		{ -0.00000906, 0.00000287, 0.00001005, 0.00000156, -0.00000455, -0.00000152, -0.00000878, -0.00000354, 0.00000249, 0.00000206, -0.00000397, 0.00000039, -0.00001515, 0.00001115, -0.00001916, 0.00001263, -0.00002760, 0.00003070, 0.00000529, 0.00000082, -0.00000310, 0.00001086, -0.00000304, 0.00001299, -0.00001027, 0.00001077, -0.00001540, 0.00003217, -0.00002259, 0.00002743, -0.00003148, 0.00000470, -0.00001624, 0.00000504, -0.00001675, 0.00000186, -0.00002215, 0.00000839, -0.00003846, 0.00000185, -0.00006240, -0.00003330, -0.00033554, 0.00036035, 0.00025557, 0.00041754, -0.00019821, -0.00015582, -0.00000003, -0.00005327, 0.00001024, -0.00003462, 0.00001197, -0.00003104, 0.00002413, -0.00003044, 0.00002703, -0.00003676, 0.00002032, -0.00001818, -0.00000015, -0.00001089, 0.00000721, -0.00000415, 0.00000230, -0.00000677, -0.00000871, -0.00000505, 0.00000810, -0.00000563, -0.00000929, 0.00000936, 0.00000146, -0.00001757, 0.00001327, -0.00002596, 0.00002379, -0.00001267, 0.00000994, -0.00000439, -0.00000123, 0.00000100, -0.00000145, 0.00001012, 0.00000439, -0.00000902, 0.00000081, -0.00000378, -0.00000045, 0.00001147, -0.00000528, 0.00000442, -0.00000449, -0.00000238, 0.00000059, 0.00000178, 0.00000265, 0.00000280, -0.00000885, 0.00000167, -0.00000624, 0.00001102, 0.00000189, -0.00000616, -0.00000101, 0.00001475, -0.00000908, -0.00000086, 0.00000396, -0.00000340, -0.00000227, 0.00000365, 0.00000898, -0.00002110, 0.00001263, 0.00000071, -0.00000998, -0.00000413, 0.00000811, 0.00000065, -0.00000320, -0.00000266, 0.00000019, 0.00000297, -0.00000159, 0.00000079, -0.00000667, -0.00000016, },
		{ -0.00000275, -0.00000683, 0.00000206, 0.00001435, -0.00000501, 0.00000023, -0.00000636, -0.00000441, -0.00000808, 0.00001092, -0.00000814, 0.00000669, -0.00001236, -0.00000085, -0.00000338, -0.00000594, 0.00000366, -0.00002072, 0.00003070, 0.00000239, -0.00000118, 0.00000099, 0.00000526, 0.00000353, 0.00000497, -0.00000376, 0.00000255, -0.00000531, 0.00002190, -0.00001388, 0.00001899, -0.00002608, -0.00000039, -0.00000853, -0.00000437, -0.00000627, -0.00001200, -0.00000479, -0.00001217, -0.00001736, -0.00002598, -0.00003078, -0.00009679, -0.00028707, 0.00039753, 0.00031378, 0.00037066, -0.00032841, -0.00004493, -0.00005590, -0.00000614, -0.00002634, -0.00000173, -0.00001659, -0.00000485, 0.00000264, -0.00001208, 0.00001030, -0.00002310, 0.00001073, -0.00001033, -0.00000864, -0.00000166, 0.00000078, 0.00000153, -0.00000430, -0.00000056, -0.00001655, 0.00000419, 0.00000148, -0.00000095, -0.00001473, 0.00001902, -0.00001261, -0.00000366, 0.00000145, -0.00001756, 0.00002286, -0.00001613, 0.00001492, -0.00000983, 0.00000283, -0.00000188, 0.00000161, 0.00000936, 0.00000202, -0.00000798, 0.00000103, -0.00000495, 0.00000325, 0.00000892, -0.00000588, 0.00000558, -0.00000701, 0.00000012, -0.00000014, 0.00000194, 0.00000341, 0.00000100, -0.00000853, 0.00000202, -0.00000515, 0.00001154, -0.00000112, -0.00000554, 0.00000233, 0.00001334, -0.00001360, 0.00000536, -0.00000086, -0.00000021, -0.00000511, 0.00000863, 0.00000190, -0.00001770, 0.00001655, -0.00000517, -0.00000927, -0.00000238, 0.00001034, -0.00000325, -0.00000069, -0.00000547, 0.00000459, -0.00000132, 0.00000212, -0.00000287, -0.00000405, },
		{ -0.00000152, -0.00000488, -0.00000547, 0.00000317, 0.00001612, -0.00001001, 0.00000407, -0.00001000, -0.00000187, -0.00000958, 0.00001583, -0.00001551, 0.00001513, -0.00002351, 0.00001319, -0.00001999, 0.00001254, -0.00001259, -0.00000608, 0.00002316, 0.00000468, -0.00000602, 0.00000705, -0.00000194, 0.00001177, -0.00000525, 0.00000498, -0.00000720, 0.00000620, 0.00000983, -0.00000272, 0.00000740, -0.00001741, -0.00000795, 0.00000099, -0.00001552, 0.00000554, -0.00002677, 0.00001393, -0.00003528, 0.00000682, -0.00005746, 0.00000469, -0.00017109, -0.00021658, 0.00042362, 0.00039908, 0.00023887, -0.00037738, 0.00003392, -0.00009990, 0.00003297, -0.00005759, 0.00002739, -0.00004296, 0.00002062, -0.00001905, 0.00000632, -0.00000644, -0.00000884, 0.00000039, -0.00000225, -0.00001716, 0.00000827, -0.00000654, 0.00000720, -0.00001019, 0.00000458, -0.00002336, 0.00001332, -0.00000548, 0.00000344, -0.00001877, 0.00002657, -0.00002563, 0.00001183, -0.00001485, -0.00000190, 0.00001353, -0.00001241, 0.00001469, -0.00001211, 0.00000528, -0.00000394, 0.00000478, 0.00000731, 0.00000091, -0.00000726, 0.00000099, -0.00000551, 0.00000699, 0.00000478, -0.00000428, 0.00000472, -0.00000790, 0.00000196, -0.00000083, 0.00000211, 0.00000435, -0.00000158, -0.00000701, 0.00000146, -0.00000298, 0.00001011, -0.00000249, -0.00000549, 0.00000714, 0.00000804, -0.00001364, 0.00000863, -0.00000418, 0.00000212, -0.00000710, 0.00001291, -0.00000617, -0.00001071, 0.00001658, -0.00000967, -0.00000855, 0.00000078, 0.00001007, -0.00000529, 0.00000032, -0.00000650, 0.00000746, -0.00000477, 0.00000548, -0.00000651, },
		{ -0.00000906, 0.00000014, -0.00000599, -0.00000539, 0.00000678, 0.00001418, -0.00001174, 0.00000531, -0.00001103, -0.00000247, -0.00000633, 0.00001460, -0.00001650, 0.00001698, -0.00002777, 0.00002058, -0.00003131, 0.00002889, -0.00002987, 0.00001159, 0.00001131, 0.00001018, -0.00001214, 0.00001330, -0.00000857, 0.00001910, -0.00001495, 0.00001341, -0.00001593, 0.00001620, -0.00000059, 0.00000689, -0.00000317, -0.00000955, -0.00001391, 0.00000800, -0.00002367, 0.00001353, -0.00003646, 0.00002691, -0.00005290, 0.00002517, -0.00008246, 0.00003164, -0.00024128, -0.00014047, 0.00046351, 0.00047412, 0.00004848, -0.00034112, 0.00006106, -0.00011372, 0.00004918, -0.00007071, 0.00004140, -0.00005692, 0.00003628, -0.00003315, 0.00001798, -0.00001690, 0.00000006, -0.00000521, 0.00000096, -0.00002077, 0.00001427, -0.00001082, 0.00000952, -0.00001217, 0.00000528, -0.00002550, 0.00001865, -0.00000946, 0.00000456, -0.00001836, 0.00002828, -0.00003267, 0.00002303, -0.00002958, 0.00001591, -0.00000098, -0.00000269, 0.00000901, -0.00001021, 0.00000485, -0.00000381, 0.00000659, 0.00000514, 0.00000053, -0.00000680, 0.00000067, -0.00000505, 0.00000976, 0.00000018, -0.00000122, 0.00000207, -0.00000697, 0.00000272, -0.00000125, 0.00000239, 0.00000505, -0.00000433, -0.00000481, 0.00000050, -0.00000051, 0.00000752, -0.00000270, -0.00000532, 0.00001206, 0.00000009, -0.00000930, 0.00000797, -0.00000463, 0.00000218, -0.00000675, 0.00001458, -0.00001279, -0.00000215, 0.00001333, -0.00001254, -0.00000765, 0.00000471, 0.00000759, -0.00000526, -0.00000055, -0.00000515, 0.00000778, -0.00000612, 0.00000724, },
		{ 0.00000687, -0.00001008, 0.00000078, -0.00000624, -0.00000607, 0.00001219, 0.00000880, -0.00000994, 0.00000367, -0.00000949, -0.00000577, 0.00000108, 0.00000728, -0.00001034, 0.00001072, -0.00002265, 0.00001745, -0.00003173, 0.00003622, -0.00004127, 0.00002619, -0.00000001, 0.00001550, -0.00001702, 0.00001748, -0.00001232, 0.00002296, -0.00002145, 0.00001908, -0.00002147, 0.00002249, -0.00000683, 0.00001197, -0.00000938, -0.00000576, -0.00001513, 0.00000923, -0.00002541, 0.00001408, -0.00003695, 0.00002909, -0.00005904, 0.00003093, -0.00009329, 0.00004046, -0.00029543, -0.00007146, 0.00054006, 0.00049343, -0.00015175, -0.00024338, 0.00003908, -0.00009436, 0.00003876, -0.00006169, 0.00003609, -0.00005409, 0.00003787, -0.00003593, 0.00001984, -0.00001827, 0.00000083, -0.00000349, -0.00000314, -0.00001685, 0.00001367, -0.00000985, 0.00000682, -0.00000884, 0.00000009, -0.00002115, 0.00001813, -0.00000859, 0.00000091, -0.00001218, 0.00002234, -0.00003067, 0.00002528, -0.00003689, 0.00002965, -0.00001540, 0.00000938, -0.00000009, -0.00000495, 0.00000167, -0.00000112, 0.00000616, 0.00000394, 0.00000017, -0.00000638, 0.00000004, -0.00000335, 0.00001071, -0.00000354, 0.00000204, -0.00000145, -0.00000465, 0.00000241, -0.00000133, 0.00000294, 0.00000503, -0.00000656, -0.00000264, -0.00000015, 0.00000127, 0.00000492, -0.00000263, -0.00000399, 0.00001514, -0.00000809, -0.00000231, 0.00000407, -0.00000224, -0.00000030, -0.00000358, 0.00001256, -0.00001590, 0.00000560, 0.00000821, -0.00001414, -0.00000604, 0.00000828, 0.00000398, -0.00000385, -0.00000271, -0.00000185, 0.00000560, -0.00000502, },
		{ -0.00000222, 0.00000480, -0.00000989, 0.00000088, -0.00000646, -0.00000614, 0.00001754, 0.00000176, -0.00000597, 0.00000035, -0.00000691, -0.00000954, 0.00000980, -0.00000320, 0.00000032, -0.00000131, -0.00000994, 0.00000446, -0.00002005, 0.00003125, -0.00004234, 0.00003304, -0.00000666, 0.00001765, -0.00001869, 0.00001819, -0.00001202, 0.00002206, -0.00002331, 0.00002084, -0.00002313, 0.00002469, -0.00000861, 0.00001199, -0.00001050, -0.00000673, -0.00001122, 0.00000444, -0.00002062, 0.00000734, -0.00002843, 0.00002026, -0.00005302, 0.00002318, -0.00008891, 0.00002934, -0.00033112, -0.00000936, 0.00066040, 0.00042085, -0.00030628, -0.00012647, -0.00001000, -0.00005447, 0.00001059, -0.00003709, 0.00001640, -0.00003798, 0.00002786, -0.00002948, 0.00001407, -0.00001266, -0.00000464, 0.00000381, -0.00001294, -0.00000661, 0.00000719, -0.00000437, 0.00000027, -0.00000166, -0.00000953, -0.00001142, 0.00001243, -0.00000355, -0.00000649, -0.00000164, 0.00001013, -0.00002006, 0.00001738, -0.00003357, 0.00003457, -0.00002460, 0.00001940, -0.00000932, 0.00000150, -0.00000291, 0.00000338, 0.00000346, 0.00000431, -0.00000074, -0.00000573, -0.00000080, -0.00000056, 0.00000944, -0.00000533, 0.00000422, -0.00000464, -0.00000179, 0.00000140, -0.00000117, 0.00000382, 0.00000402, -0.00000771, -0.00000105, -0.00000002, 0.00000164, 0.00000329, -0.00000305, -0.00000074, 0.00001467, -0.00001377, 0.00000463, -0.00000108, 0.00000165, -0.00000428, 0.00000155, 0.00000714, -0.00001457, 0.00001060, 0.00000281, -0.00001506, -0.00000317, 0.00001026, 0.00000069, -0.00000237, -0.00000484, 0.00000203, 0.00000203, },
		{ -0.00000125, 0.00000088, 0.00000209, -0.00000929, 0.00000126, -0.00000770, -0.00000403, 0.00002052, -0.00000436, -0.00000209, -0.00000255, -0.00000555, -0.00001078, 0.00001580, -0.00001211, 0.00001041, -0.00001376, 0.00000479, -0.00001308, -0.00000026, 0.00001580, -0.00003297, 0.00003090, -0.00000702, 0.00001554, -0.00001682, 0.00001571, -0.00000835, 0.00001712, -0.00002110, 0.00001939, -0.00002214, 0.00002459, -0.00000814, 0.00000929, -0.00000874, -0.00001007, -0.00000494, -0.00000320, -0.00001288, -0.00000242, -0.00001585, 0.00000564, -0.00004026, 0.00000779, -0.00007611, 0.00000580, -0.00035758, 0.00006336, 0.00080595, 0.00024638, -0.00037263, -0.00003485, -0.00005562, -0.00001551, -0.00001850, -0.00001070, -0.00000618, -0.00001810, 0.00001400, -0.00002047, 0.00000695, -0.00000607, -0.00001078, 0.00001138, -0.00002334, 0.00000539, -0.00000137, 0.00000221, -0.00000661, 0.00000555, -0.00001961, 0.00000000, 0.00000474, 0.00000266, -0.00001436, 0.00000940, -0.00000399, -0.00000499, 0.00000227, -0.00002058, 0.00002934, -0.00002569, 0.00002400, -0.00001561, 0.00000668, -0.00000696, 0.00000819, -0.00000069, 0.00000623, -0.00000253, -0.00000472, -0.00000162, 0.00000280, 0.00000620, -0.00000477, 0.00000444, -0.00000643, 0.00000066, 0.00000019, -0.00000084, 0.00000492, 0.00000210, -0.00000767, -0.00000020, 0.00000088, 0.00000045, 0.00000304, -0.00000412, 0.00000432, 0.00000999, -0.00001496, 0.00000891, -0.00000508, 0.00000499, -0.00000790, 0.00000678, -0.00000001, -0.00000936, 0.00001197, -0.00000173, -0.00001569, 0.00000110, 0.00000990, -0.00000110, -0.00000206, -0.00000551, 0.00000481, },
		{ 0.00000526, -0.00000280, 0.00000288, -0.00000014, -0.00000905, 0.00000249, -0.00001050, 0.00000115, 0.00001946, -0.00000739, -0.00000038, -0.00000325, -0.00000717, -0.00000725, 0.00001577, -0.00001511, 0.00001482, -0.00002065, 0.00001459, -0.00002743, 0.00002005, -0.00000403, -0.00001734, 0.00002223, -0.00000246, 0.00001030, -0.00001285, 0.00001193, -0.00000363, 0.00001053, -0.00001695, 0.00001679, -0.00002083, 0.00002508, -0.00000896, 0.00000782, -0.00000815, -0.00001159, -0.00000094, -0.00000861, -0.00000776, -0.00000878, -0.00000670, -0.00000646, -0.00002973, -0.00000537, -0.00006633, -0.00001705, -0.00039141, 0.00017784, 0.00093168, -0.00000452, -0.00033757, 0.00000216, -0.00007279, 0.00000301, -0.00003226, 0.00000347, -0.00001935, -0.00000501, 0.00000513, -0.00001643, 0.00000541, -0.00000516, -0.00001124, 0.00001310, -0.00002831, 0.00001348, -0.00000716, 0.00000565, -0.00000967, 0.00000845, -0.00002551, 0.00000854, -0.00000081, 0.00000622, -0.00001874, 0.00001631, -0.00001441, 0.00000829, -0.00001396, -0.00000273, 0.00001653, -0.00001906, 0.00002218, -0.00001734, 0.00000890, -0.00000881, 0.00001166, -0.00000499, 0.00000912, -0.00000510, -0.00000335, -0.00000209, 0.00000591, 0.00000186, -0.00000226, 0.00000257, -0.00000628, 0.00000203, -0.00000080, -0.00000038, 0.00000594, -0.00000033, -0.00000674, 0.00000023, 0.00000199, -0.00000169, 0.00000384, -0.00000525, 0.00000999, 0.00000203, -0.00001130, 0.00000911, -0.00000618, 0.00000605, -0.00000938, 0.00000996, -0.00000643, -0.00000212, 0.00001011, -0.00000500, -0.00001593, 0.00000624, 0.00000725, -0.00000093, -0.00000349, -0.00000391, },
		{ -0.00000027, 0.00000318, -0.00000203, 0.00000307, -0.00000133, -0.00000935, 0.00000435, -0.00001435, 0.00000892, 0.00001421, -0.00000659, -0.00000160, -0.00000134, -0.00001176, 0.00000111, 0.00000891, -0.00001044, 0.00001092, -0.00001832, 0.00001439, -0.00003186, 0.00003303, -0.00002065, -0.00000175, 0.00001177, 0.00000378, 0.00000437, -0.00000901, 0.00000919, -0.00000046, 0.00000500, -0.00001321, 0.00001499, -0.00002114, 0.00002836, -0.00001386, 0.00001083, -0.00001214, -0.00000778, -0.00000299, -0.00000772, -0.00000962, -0.00000678, -0.00000696, -0.00000915, -0.00002895, -0.00000806, -0.00006903, -0.00002863, -0.00044612, 0.00036615, 0.00097720, -0.00027439, -0.00022315, -0.00001866, -0.00005335, -0.00000672, -0.00002333, -0.00000154, -0.00001651, -0.00000477, 0.00000629, -0.00002139, 0.00001306, -0.00001348, -0.00000251, 0.00000542, -0.00002402, 0.00001383, -0.00000687, 0.00000329, -0.00000650, 0.00000446, -0.00002425, 0.00001102, -0.00000131, 0.00000455, -0.00001707, 0.00001600, -0.00001686, 0.00001412, -0.00002470, 0.00001341, 0.00000138, -0.00000794, 0.00001543, -0.00001482, 0.00000783, -0.00000774, 0.00001257, -0.00000808, 0.00001205, -0.00000807, -0.00000174, -0.00000193, 0.00000787, -0.00000232, 0.00000111, -0.00000071, -0.00000438, 0.00000213, -0.00000136, 0.00000030, 0.00000643, -0.00000259, -0.00000551, 0.00000082, 0.00000238, -0.00000370, 0.00000482, -0.00000537, 0.00001430, -0.00000691, -0.00000423, 0.00000551, -0.00000400, 0.00000416, -0.00000781, 0.00000962, -0.00000981, 0.00000473, 0.00000633, -0.00000732, -0.00001520, 0.00001112, 0.00000329, 0.00000065, -0.00000625, },
		{ -0.00000901, 0.00000408, -0.00000040, 0.00000046, 0.00000175, -0.00000175, -0.00000962, 0.00000583, -0.00001769, 0.00001734, 0.00000645, -0.00000306, -0.00000475, 0.00000182, -0.00001728, 0.00001168, -0.00000230, -0.00000009, 0.00000016, -0.00000752, 0.00000368, -0.00002387, 0.00003408, -0.00002847, 0.00000838, 0.00000396, 0.00000856, -0.00000020, -0.00000677, 0.00000871, -0.00000019, 0.00000194, -0.00001090, 0.00001438, -0.00002293, 0.00003412, -0.00002274, 0.00001863, -0.00002108, 0.00000163, -0.00001136, -0.00000031, -0.00001846, 0.00000362, -0.00001711, -0.00000157, -0.00003882, 0.00000034, -0.00008460, -0.00002994, -0.00051942, 0.00064346, 0.00088733, -0.00049015, -0.00007936, -0.00007416, -0.00000939, -0.00003666, 0.00000263, -0.00002169, -0.00000037, -0.00001547, 0.00001565, -0.00003316, 0.00002762, -0.00002900, 0.00001368, -0.00001029, -0.00001122, 0.00000662, -0.00000077, -0.00000416, 0.00000190, -0.00000551, -0.00001624, 0.00000740, 0.00000323, -0.00000206, -0.00000986, 0.00000877, -0.00001064, 0.00001013, -0.00002569, 0.00002227, -0.00001058, 0.00000321, 0.00000676, -0.00000979, 0.00000442, -0.00000405, 0.00001046, -0.00000891, 0.00001398, -0.00001079, -0.00000014, -0.00000094, 0.00000798, -0.00000506, 0.00000394, -0.00000417, -0.00000154, 0.00000125, -0.00000153, 0.00000128, 0.00000605, -0.00000408, -0.00000461, 0.00000212, 0.00000125, -0.00000448, 0.00000505, -0.00000348, 0.00001527, -0.00001393, 0.00000356, -0.00000006, 0.00000030, 0.00000012, -0.00000369, 0.00000565, -0.00000887, 0.00000903, 0.00000228, -0.00000937, -0.00001280, 0.00001435, -0.00000047, 0.00000234, },
		{ 0.00000265, -0.00001013, 0.00000728, -0.00000369, 0.00000318, 0.00000006, -0.00000236, -0.00000876, 0.00000543, -0.00001837, 0.00002358, -0.00000091, 0.00000067, -0.00000752, 0.00000362, -0.00002030, 0.00002003, -0.00001286, 0.00001084, -0.00001223, 0.00000654, -0.00001284, -0.00000665, 0.00002405, -0.00002637, 0.00001106, 0.00000080, 0.00001066, -0.00000310, -0.00000565, 0.00000957, -0.00000184, 0.00000047, -0.00000881, 0.00001300, -0.00002342, 0.00003889, -0.00003193, 0.00002762, -0.00003128, 0.00001261, -0.00002168, 0.00000884, -0.00002875, 0.00001612, -0.00003045, 0.00000916, -0.00005132, 0.00001013, -0.00010084, -0.00003856, -0.00058444, 0.00099198, 0.00063556, -0.00058766, 0.00003434, -0.00012576, 0.00003282, -0.00006661, 0.00002929, -0.00004347, 0.00001793, -0.00002785, 0.00002498, -0.00004379, 0.00004134, -0.00004445, 0.00003053, -0.00002779, 0.00000469, -0.00000372, 0.00000712, -0.00001256, 0.00001117, -0.00001708, -0.00000542, 0.00000104, 0.00000965, -0.00001025, -0.00000085, -0.00000136, 0.00000066, -0.00000144, -0.00001708, 0.00002173, -0.00001586, 0.00001081, -0.00000084, -0.00000445, 0.00000021, 0.00000119, 0.00000563, -0.00000690, 0.00001400, -0.00001255, 0.00000118, 0.00000084, 0.00000600, -0.00000547, 0.00000496, -0.00000646, 0.00000112, 0.00000002, -0.00000152, 0.00000261, 0.00000466, -0.00000446, -0.00000432, 0.00000425, -0.00000171, -0.00000339, 0.00000410, 0.00000065, 0.00001181, -0.00001660, 0.00000916, -0.00000503, 0.00000464, -0.00000428, 0.00000125, -0.00000055, -0.00000383, 0.00000965, -0.00000071, -0.00001169, -0.00000831, 0.00001485, -0.00000256, },
		{ -0.00000211, 0.00000061, -0.00000838, 0.00000768, -0.00000486, 0.00000441, -0.00000061, -0.00000421, -0.00000579, 0.00000198, -0.00001469, 0.00002511, -0.00000493, 0.00000187, -0.00000742, 0.00000157, -0.00001765, 0.00002184, -0.00001745, 0.00001633, -0.00001950, 0.00001629, -0.00002705, 0.00001215, 0.00000878, -0.00001807, 0.00000838, 0.00000106, 0.00001131, -0.00000607, -0.00000332, 0.00000889, -0.00000237, -0.00000228, -0.00000398, 0.00000728, -0.00001811, 0.00003717, -0.00003527, 0.00003144, -0.00003616, 0.00001810, -0.00002643, 0.00001156, -0.00003127, 0.00002015, -0.00003527, 0.00001020, -0.00005189, 0.00000393, -0.00009599, -0.00008574, -0.00059035, 0.00135443, 0.00024106, -0.00053555, 0.00007176, -0.00013693, 0.00004554, -0.00007393, 0.00003750, -0.00005035, 0.00002417, -0.00002984, 0.00002368, -0.00004325, 0.00004444, -0.00005042, 0.00003909, -0.00003856, 0.00001597, -0.00001054, 0.00001095, -0.00001619, 0.00001543, -0.00002422, 0.00000245, -0.00000287, 0.00001311, -0.00001509, 0.00000454, -0.00000826, 0.00001048, -0.00001448, -0.00000339, 0.00001389, -0.00001440, 0.00001356, -0.00000571, -0.00000033, -0.00000339, 0.00000666, -0.00000105, -0.00000211, 0.00001157, -0.00001272, 0.00000195, 0.00000314, 0.00000221, -0.00000335, 0.00000350, -0.00000657, 0.00000255, -0.00000085, -0.00000158, 0.00000418, 0.00000247, -0.00000385, -0.00000450, 0.00000676, -0.00000606, -0.00000058, 0.00000231, 0.00000617, 0.00000436, -0.00001400, 0.00001062, -0.00000727, 0.00000698, -0.00000700, 0.00000476, -0.00000660, 0.00000365, 0.00000676, -0.00000210, -0.00001434, -0.00000196, 0.00001227, },
		{ 0.00000727, 0.00000066, -0.00000371, -0.00000358, 0.00000463, -0.00000289, 0.00000305, 0.00000063, -0.00000779, -0.00000053, -0.00000458, -0.00000633, 0.00002085, -0.00000396, -0.00000105, -0.00000323, -0.00000529, -0.00000819, 0.00001501, -0.00001293, 0.00001235, -0.00001662, 0.00001537, -0.00003107, 0.00002379, -0.00000359, -0.00001007, 0.00000512, 0.00000135, 0.00001317, -0.00001170, 0.00000311, 0.00000335, 0.00000167, -0.00000947, 0.00000650, -0.00000589, -0.00000322, 0.00002415, -0.00002710, 0.00002406, -0.00002944, 0.00001145, -0.00001853, 0.00000032, -0.00001757, 0.00000591, -0.00002044, -0.00001082, -0.00002651, -0.00003510, -0.00004866, -0.00020426, -0.00047530, 0.00164218, -0.00022831, -0.00034908, 0.00001818, -0.00008990, 0.00001342, -0.00004507, 0.00001509, -0.00003127, 0.00000853, -0.00001307, 0.00000457, -0.00002477, 0.00003014, -0.00004015, 0.00003265, -0.00003590, 0.00001628, -0.00000838, 0.00000608, -0.00001074, 0.00001024, -0.00002219, 0.00000260, 0.00000000, 0.00000960, -0.00001263, 0.00000195, -0.00000654, 0.00001225, -0.00002168, 0.00000845, 0.00000398, -0.00000919, 0.00001269, -0.00000800, 0.00000221, -0.00000569, 0.00001124, -0.00000838, 0.00000486, 0.00000663, -0.00001094, 0.00000205, 0.00000547, -0.00000254, 0.00000067, -0.00000027, -0.00000423, 0.00000218, -0.00000088, -0.00000180, 0.00000569, -0.00000001, -0.00000272, -0.00000459, 0.00000874, -0.00001072, 0.00000312, 0.00000070, 0.00001123, -0.00000504, -0.00000713, 0.00000775, -0.00000597, 0.00000629, -0.00000680, 0.00000508, -0.00000999, 0.00001109, 0.00000164, -0.00000219, -0.00001677, 0.00000527, },
		{ 0.00001188, 0.00000135, 0.00000457, -0.00000914, 0.00000315, -0.00000109, 0.00000187, -0.00000082, 0.00000364, -0.00001259, 0.00000606, -0.00001286, 0.00000516, 0.00001190, 0.00000149, -0.00000761, 0.00000417, -0.00001543, 0.00000618, 0.00000112, -0.00000019, -0.00000082, -0.00000308, 0.00000193, -0.00002111, 0.00002254, -0.00000664, -0.00000823, 0.00000590, -0.00000151, 0.00001832, -0.00002147, 0.00001499, -0.00000857, 0.00001189, -0.00002226, 0.00002317, -0.00002672, 0.00002159, -0.00000118, -0.00000564, 0.00000338, -0.00000899, -0.00000955, 0.00000422, -0.00002695, 0.00001463, -0.00002960, 0.00001773, -0.00005805, 0.00002940, -0.00011251, 0.00004864, -0.00041048, -0.00018912, 0.00175818, -0.00067061, -0.00008796, -0.00010344, 0.00000547, -0.00005783, 0.00001648, -0.00003589, 0.00001281, -0.00002859, 0.00002194, -0.00003143, 0.00001077, -0.00000122, -0.00001349, 0.00001059, -0.00001863, 0.00000413, 0.00000389, -0.00000803, 0.00000401, -0.00000473, -0.00001019, -0.00000610, 0.00001067, -0.00000157, -0.00000234, -0.00000935, 0.00000550, 0.00000263, -0.00001799, 0.00001234, -0.00000228, -0.00000455, 0.00001091, -0.00000915, 0.00000383, -0.00000672, 0.00001417, -0.00001507, 0.00001283, -0.00000020, -0.00000735, 0.00000155, 0.00000719, -0.00000705, 0.00000533, -0.00000530, -0.00000004, 0.00000013, -0.00000001, -0.00000214, 0.00000677, -0.00000214, -0.00000173, -0.00000393, 0.00000914, -0.00001431, 0.00000657, 0.00000044, 0.00001367, -0.00001341, 0.00000142, 0.00000213, -0.00000196, 0.00000302, -0.00000383, 0.00000175, -0.00000911, 0.00001610, -0.00000388, -0.00000183, -0.00001805, },
		{ -0.00001731, 0.00001631, -0.00000366, 0.00000784, -0.00001377, 0.00000970, -0.00000752, 0.00000778, -0.00000585, 0.00000719, -0.00001712, 0.00001200, -0.00002028, 0.00001667, 0.00000131, 0.00000877, -0.00001530, 0.00001190, -0.00002513, 0.00002083, -0.00001480, 0.00001575, -0.00001819, 0.00001627, -0.00001985, 0.00000028, 0.00000865, 0.00000113, -0.00001458, 0.00001241, -0.00000822, 0.00002631, -0.00003402, 0.00003044, -0.00002491, 0.00002632, -0.00003820, 0.00004267, -0.00005098, 0.00005155, -0.00003412, 0.00002479, -0.00002635, 0.00002066, -0.00003999, 0.00003632, -0.00006386, 0.00005802, -0.00007853, 0.00007081, -0.00012192, 0.00010439, -0.00021435, 0.00017831, -0.00068737, 0.00028124, 0.00162833, -0.00097848, 0.00016222, -0.00024103, 0.00011469, -0.00014213, 0.00008991, -0.00009829, 0.00006764, -0.00007515, 0.00006438, -0.00007403, 0.00005372, -0.00004092, 0.00002188, -0.00002035, 0.00000754, -0.00001574, 0.00002187, -0.00002674, 0.00002324, -0.00002473, 0.00000757, -0.00002001, 0.00002567, -0.00001683, 0.00001191, -0.00002532, 0.00002432, -0.00001630, -0.00000328, 0.00000597, -0.00000144, -0.00000372, 0.00001064, -0.00001069, 0.00000540, -0.00000681, 0.00001497, -0.00001975, 0.00002020, -0.00000769, -0.00000259, 0.00000078, 0.00000764, -0.00001003, 0.00000900, -0.00000997, 0.00000469, -0.00000283, 0.00000140, -0.00000237, 0.00000708, -0.00000336, -0.00000142, -0.00000208, 0.00000724, -0.00001574, 0.00000887, 0.00000224, 0.00001193, -0.00001791, 0.00000850, -0.00000364, 0.00000271, -0.00000114, 0.00000038, -0.00000415, -0.00000383, 0.00001718, -0.00000815, -0.00000197, },
		{ -0.00000317, -0.00001409, 0.00001749, -0.00000622, 0.00000882, -0.00001564, 0.00001372, -0.00001214, 0.00001246, -0.00000998, 0.00000944, -0.00001945, 0.00001504, -0.00002395, 0.00002455, -0.00000696, 0.00001417, -0.00002057, 0.00001623, -0.00002987, 0.00003001, -0.00002608, 0.00002763, -0.00003184, 0.00003280, -0.00004094, 0.00002470, -0.00001138, 0.00001530, -0.00002610, 0.00002217, -0.00001599, 0.00003350, -0.00004479, 0.00004420, -0.00004014, 0.00003936, -0.00005127, 0.00005803, -0.00007047, 0.00007732, -0.00006473, 0.00005417, -0.00005491, 0.00004869, -0.00006825, 0.00006510, -0.00009609, 0.00009634, -0.00012277, 0.00011865, -0.00017932, 0.00017090, -0.00030645, 0.00029497, -0.00097672, 0.00089368, 0.00123102, -0.00107302, 0.00031875, -0.00033362, 0.00019247, -0.00020308, 0.00014474, -0.00014601, 0.00011072, -0.00011166, 0.00009685, -0.00010694, 0.00008866, -0.00007468, 0.00005282, -0.00004810, 0.00003184, -0.00003393, 0.00003700, -0.00004172, 0.00003855, -0.00004144, 0.00002323, -0.00003198, 0.00003823, -0.00002940, 0.00002304, -0.00003826, 0.00004198, -0.00003732, 0.00001707, -0.00000789, 0.00000587, -0.00000716, 0.00001255, -0.00001306, 0.00000721, -0.00000616, 0.00001327, -0.00002116, 0.00002516, -0.00001413, 0.00000217, 0.00000028, 0.00000629, -0.00001043, 0.00001021, -0.00001258, 0.00000832, -0.00000551, 0.00000271, -0.00000218, 0.00000641, -0.00000341, -0.00000197, 0.00000091, 0.00000298, -0.00001457, 0.00000977, 0.00000580, 0.00000584, -0.00001699, 0.00001164, -0.00000705, 0.00000580, -0.00000405, 0.00000360, -0.00001035, 0.00000450, 0.00001408, -0.00001016, },
		{ -0.00000981, -0.00000542, -0.00000868, 0.00001524, -0.00000560, 0.00000661, -0.00001357, 0.00001351, -0.00001288, 0.00001375, -0.00001123, 0.00000873, -0.00001807, 0.00001357, -0.00002189, 0.00002600, -0.00000961, 0.00001444, -0.00002040, 0.00001427, -0.00002635, 0.00002935, -0.00002721, 0.00002904, -0.00003430, 0.00003768, -0.00005110, 0.00004124, -0.00002738, 0.00002738, -0.00003588, 0.00002947, -0.00001956, 0.00003435, -0.00004757, 0.00004931, -0.00004691, 0.00004370, -0.00005398, 0.00006103, -0.00007577, 0.00008807, -0.00008114, 0.00007018, -0.00006958, 0.00006169, -0.00008004, 0.00007509, -0.00010634, 0.00010980, -0.00013988, 0.00013580, -0.00020081, 0.00019334, -0.00034351, 0.00033532, -0.00118641, 0.00154905, 0.00061337, -0.00093093, 0.00033045, -0.00033545, 0.00020162, -0.00020923, 0.00015367, -0.00015485, 0.00012064, -0.00011929, 0.00010253, -0.00011445, 0.00010056, -0.00008827, 0.00006602, -0.00006020, 0.00004282, -0.00004035, 0.00004030, -0.00004447, 0.00004153, -0.00004637, 0.00002852, -0.00003432, 0.00004114, -0.00003228, 0.00002374, -0.00003998, 0.00004924, -0.00005076, 0.00003417, -0.00002239, 0.00001524, -0.00001227, 0.00001506, -0.00001521, 0.00000857, -0.00000447, 0.00000879, -0.00001835, 0.00002612, -0.00001777, 0.00000562, 0.00000064, 0.00000293, -0.00000780, 0.00000810, -0.00001190, 0.00000945, -0.00000673, 0.00000317, -0.00000132, 0.00000479, -0.00000235, -0.00000318, 0.00000445, -0.00000296, -0.00001113, 0.00000979, 0.00000977, -0.00000310, -0.00001102, 0.00001003, -0.00000681, 0.00000592, -0.00000421, 0.00000386, -0.00001430, 0.00001351, 0.00000780, },
		{ 0.00000004, -0.00000772, -0.00000815, -0.00000203, 0.00001029, -0.00000219, 0.00000150, -0.00000768, 0.00000874, -0.00000894, 0.00001058, -0.00000858, 0.00000436, -0.00001263, 0.00000749, -0.00001393, 0.00002033, -0.00000538, 0.00000824, -0.00001381, 0.00000549, -0.00001420, 0.00001794, -0.00001645, 0.00001746, -0.00002213, 0.00002602, -0.00004344, 0.00004116, -0.00003004, 0.00002873, -0.00003656, 0.00002829, -0.00001388, 0.00002412, -0.00003735, 0.00004026, -0.00003937, 0.00003362, -0.00004082, 0.00004590, -0.00006035, 0.00007603, -0.00007444, 0.00006337, -0.00006063, 0.00004945, -0.00006459, 0.00005482, -0.00008184, 0.00008356, -0.00011257, 0.00010231, -0.00016280, 0.00014243, -0.00028653, 0.00023919, -0.00121365, 0.00210941, -0.00011337, -0.00059479, 0.00019570, -0.00023444, 0.00012936, -0.00014819, 0.00010489, -0.00011353, 0.00008681, -0.00008856, 0.00007302, -0.00008861, 0.00008148, -0.00007389, 0.00005394, -0.00004931, 0.00003355, -0.00002895, 0.00002666, -0.00003041, 0.00002767, -0.00003477, 0.00001864, -0.00002255, 0.00003035, -0.00002169, 0.00001012, -0.00002564, 0.00003968, -0.00004866, 0.00003935, -0.00002964, 0.00002056, -0.00001474, 0.00001522, -0.00001503, 0.00000811, -0.00000123, 0.00000153, -0.00001107, 0.00002219, -0.00001735, 0.00000667, 0.00000219, -0.00000215, -0.00000250, 0.00000277, -0.00000760, 0.00000738, -0.00000571, 0.00000230, 0.00000029, 0.00000248, -0.00000055, -0.00000455, 0.00000760, -0.00000938, -0.00000633, 0.00000988, 0.00001216, -0.00001215, -0.00000223, 0.00000482, -0.00000335, 0.00000317, -0.00000146, 0.00000042, -0.00001415, 0.00002073, },
		{ 0.00002440, -0.00000740, -0.00000483, -0.00001053, 0.00000452, 0.00000413, 0.00000271, -0.00000513, 0.00000056, 0.00000067, -0.00000118, 0.00000348, -0.00000246, -0.00000297, -0.00000424, -0.00000163, -0.00000204, 0.00000934, 0.00000436, -0.00000313, -0.00000239, -0.00000774, 0.00000344, -0.00000089, 0.00000307, -0.00000413, 0.00000203, -0.00000044, -0.00001793, 0.00002208, -0.00001532, 0.00001485, -0.00002408, 0.00001564, 0.00000287, 0.00000175, -0.00001331, 0.00001604, -0.00001636, 0.00000825, -0.00001151, 0.00001279, -0.00002429, 0.00004060, -0.00004322, 0.00003200, -0.00002640, 0.00001047, -0.00002055, 0.00000321, -0.00002149, 0.00001582, -0.00003806, 0.00001437, -0.00006022, 0.00001057, -0.00012238, -0.00002423, -0.00097840, 0.00243313, -0.00080000, -0.00016186, -0.00003717, -0.00005763, -0.00000598, -0.00003336, 0.00000854, -0.00002975, 0.00001518, -0.00002457, 0.00001309, -0.00003367, 0.00003483, -0.00003420, 0.00001872, -0.00001700, 0.00000515, -0.00000097, -0.00000216, -0.00000165, -0.00000092, -0.00000837, -0.00000512, 0.00000208, 0.00000731, 0.00000054, -0.00001577, 0.00000331, 0.00001312, -0.00002860, 0.00002818, -0.00002437, 0.00001707, -0.00001089, 0.00001041, -0.00001068, 0.00000465, 0.00000377, -0.00000786, -0.00000010, 0.00001370, -0.00001266, 0.00000495, 0.00000479, -0.00000809, 0.00000429, -0.00000464, -0.00000052, 0.00000241, -0.00000244, 0.00000002, 0.00000251, -0.00000010, 0.00000145, -0.00000550, 0.00000940, -0.00001494, -0.00000121, 0.00001085, 0.00001126, -0.00001830, 0.00000619, -0.00000147, 0.00000142, -0.00000088, 0.00000282, -0.00000584, -0.00000947, },
		{ -0.00000137, 0.00002390, -0.00001318, -0.00000204, -0.00001181, 0.00000965, -0.00000156, 0.00000734, -0.00001127, 0.00000877, -0.00000825, 0.00000811, -0.00000558, 0.00000536, -0.00001137, 0.00000485, -0.00001092, 0.00001023, -0.00000316, 0.00001599, -0.00001615, 0.00001009, -0.00002087, 0.00002084, -0.00002056, 0.00002424, -0.00002807, 0.00002993, -0.00003328, 0.00001788, -0.00001039, 0.00001329, -0.00001205, -0.00000020, -0.00000623, 0.00002740, -0.00002848, 0.00001963, -0.00001824, 0.00001703, -0.00002690, 0.00002755, -0.00003091, 0.00002427, -0.00000995, 0.00000450, -0.00001578, 0.00002429, -0.00004563, 0.00004139, -0.00006755, 0.00006092, -0.00007833, 0.00006730, -0.00010993, 0.00008639, -0.00017874, 0.00012258, -0.00043498, -0.00045620, 0.00241670, -0.00129983, 0.00024540, -0.00028713, 0.00013839, -0.00016071, 0.00009981, -0.00010566, 0.00007131, -0.00007270, 0.00005372, -0.00005987, 0.00003440, -0.00002498, 0.00001781, -0.00002780, 0.00002606, -0.00003291, 0.00003478, -0.00003744, 0.00003303, -0.00003555, 0.00002464, -0.00003522, 0.00003241, -0.00002086, 0.00002697, -0.00004592, 0.00003881, -0.00002343, 0.00000455, 0.00000281, -0.00000663, 0.00000392, 0.00000015, 0.00000006, -0.00000186, -0.00000183, 0.00000992, -0.00001780, 0.00001253, 0.00000238, -0.00000480, 0.00000106, 0.00000770, -0.00001354, 0.00001079, -0.00001217, 0.00000755, -0.00000414, 0.00000230, -0.00000319, 0.00000492, -0.00000241, 0.00000316, -0.00000566, 0.00000923, -0.00001864, 0.00000352, 0.00001281, 0.00000639, -0.00001943, 0.00001131, -0.00000606, 0.00000508, -0.00000397, 0.00000633, -0.00001275, },
		{ -0.00001766, 0.00000799, 0.00001979, -0.00001665, 0.00000008, -0.00001160, 0.00001243, -0.00000536, 0.00001009, -0.00001494, 0.00001448, -0.00001516, 0.00001599, -0.00001382, 0.00001237, -0.00001837, 0.00001198, -0.00001727, 0.00001898, -0.00001272, 0.00002501, -0.00002643, 0.00001922, -0.00002893, 0.00003186, -0.00003373, 0.00003874, -0.00004490, 0.00005075, -0.00006033, 0.00005126, -0.00004443, 0.00004564, -0.00004342, 0.00002758, -0.00003005, 0.00005181, -0.00005769, 0.00005176, -0.00005240, 0.00005046, -0.00006107, 0.00006465, -0.00007221, 0.00007102, -0.00006041, 0.00005312, -0.00006381, 0.00007423, -0.00010021, 0.00010083, -0.00013464, 0.00013948, -0.00016980, 0.00017093, -0.00023344, 0.00023340, -0.00037274, 0.00038201, -0.00091531, 0.00030187, 0.00202967, -0.00151076, 0.00051716, -0.00046921, 0.00028847, -0.00028185, 0.00020660, -0.00019906, 0.00015593, -0.00014732, 0.00012008, -0.00012187, 0.00009336, -0.00007753, 0.00006341, -0.00006829, 0.00006390, -0.00006617, 0.00006502, -0.00006652, 0.00006120, -0.00006392, 0.00005237, -0.00006051, 0.00005789, -0.00004389, 0.00004712, -0.00006903, 0.00006874, -0.00005785, 0.00003992, -0.00002833, 0.00001781, -0.00001503, 0.00001555, -0.00001349, 0.00000948, -0.00000977, 0.00001551, -0.00002591, 0.00002392, -0.00000891, 0.00000396, -0.00000352, 0.00000970, -0.00001694, 0.00001504, -0.00001757, 0.00001429, -0.00001031, 0.00000709, -0.00000644, 0.00000695, -0.00000395, 0.00000418, -0.00000494, 0.00000702, -0.00002003, 0.00000770, 0.00001500, -0.00000160, -0.00001522, 0.00001164, -0.00000712, 0.00000583, -0.00000428, 0.00000686, },
		{ 0.00000337, -0.00001854, 0.00001623, 0.00001335, -0.00001779, 0.00000133, -0.00000993, 0.00001259, -0.00000652, 0.00001003, -0.00001496, 0.00001593, -0.00001778, 0.00001988, -0.00001864, 0.00001620, -0.00002184, 0.00001512, -0.00001848, 0.00002137, -0.00001585, 0.00002774, -0.00003043, 0.00002173, -0.00002858, 0.00003237, -0.00003520, 0.00004040, -0.00004728, 0.00005544, -0.00007048, 0.00006926, -0.00006651, 0.00006890, -0.00006768, 0.00004896, -0.00004620, 0.00006638, -0.00007550, 0.00007185, -0.00007462, 0.00007190, -0.00008200, 0.00008681, -0.00009692, 0.00010023, -0.00009377, 0.00008489, -0.00009359, 0.00010370, -0.00013203, 0.00013467, -0.00017229, 0.00018482, -0.00022501, 0.00023442, -0.00031150, 0.00032717, -0.00050405, 0.00056530, -0.00134261, 0.00117128, 0.00132995, -0.00140485, 0.00058962, -0.00052385, 0.00034291, -0.00032694, 0.00024995, -0.00023866, 0.00019457, -0.00018240, 0.00015105, -0.00015139, 0.00012312, -0.00010417, 0.00008535, -0.00008657, 0.00008141, -0.00008090, 0.00007734, -0.00007778, 0.00007161, -0.00007486, 0.00006388, -0.00007061, 0.00006879, -0.00005254, 0.00005172, -0.00007493, 0.00008152, -0.00007740, 0.00006462, -0.00005367, 0.00003963, -0.00003257, 0.00002967, -0.00002565, 0.00001961, -0.00001634, 0.00001821, -0.00002957, 0.00003094, -0.00001695, 0.00001085, -0.00000686, 0.00000950, -0.00001703, 0.00001551, -0.00001897, 0.00001758, -0.00001405, 0.00001032, -0.00000864, 0.00000795, -0.00000430, 0.00000430, -0.00000357, 0.00000318, -0.00001920, 0.00001157, 0.00001598, -0.00001047, -0.00000724, 0.00000769, -0.00000452, 0.00000336, -0.00000128, },
		{ 0.00000383, -0.00000350, -0.00001478, 0.00002160, 0.00000603, -0.00001694, 0.00000179, -0.00000721, 0.00001049, -0.00000512, 0.00000720, -0.00001128, 0.00001271, -0.00001520, 0.00001841, -0.00001851, 0.00001552, -0.00002077, 0.00001362, -0.00001416, 0.00001670, -0.00001132, 0.00002267, -0.00002677, 0.00001682, -0.00001955, 0.00002200, -0.00002414, 0.00002790, -0.00003318, 0.00004071, -0.00005831, 0.00006394, -0.00006687, 0.00007279, -0.00007495, 0.00005509, -0.00004688, 0.00006365, -0.00007417, 0.00007158, -0.00007604, 0.00007237, -0.00008077, 0.00008489, -0.00009521, 0.00010085, -0.00009776, 0.00008680, -0.00009160, 0.00009837, -0.00012569, 0.00012622, -0.00016190, 0.00017541, -0.00021872, 0.00022812, -0.00030869, 0.00032356, -0.00051426, 0.00058599, -0.00158079, 0.00198021, 0.00045321, -0.00103543, 0.00046242, -0.00043633, 0.00028534, -0.00027990, 0.00021448, -0.00020972, 0.00017321, -0.00016506, 0.00013508, -0.00013780, 0.00011359, -0.00009539, 0.00007479, -0.00007421, 0.00007047, -0.00006970, 0.00006529, -0.00006552, 0.00005890, -0.00006296, 0.00005364, -0.00006023, 0.00006031, -0.00004260, 0.00003685, -0.00005904, 0.00007093, -0.00007390, 0.00006887, -0.00006314, 0.00004979, -0.00004118, 0.00003639, -0.00003137, 0.00002445, -0.00001858, 0.00001600, -0.00002691, 0.00003137, -0.00001921, 0.00001349, -0.00000732, 0.00000627, -0.00001328, 0.00001164, -0.00001553, 0.00001618, -0.00001397, 0.00001069, -0.00000888, 0.00000746, -0.00000322, 0.00000345, -0.00000193, -0.00000158, -0.00001656, 0.00001545, 0.00001431, -0.00001743, 0.00000160, 0.00000160, 0.00000015, -0.00000101, },
		{ -0.00000495, 0.00000866, -0.00001155, -0.00000739, 0.00002347, -0.00000106, -0.00001454, 0.00000158, -0.00000401, 0.00000692, -0.00000192, 0.00000250, -0.00000500, 0.00000587, -0.00000814, 0.00001191, -0.00001349, 0.00001044, -0.00001556, 0.00000841, -0.00000582, 0.00000669, -0.00000055, 0.00001101, -0.00001677, 0.00000645, -0.00000482, 0.00000441, -0.00000448, 0.00000543, -0.00000700, 0.00001066, -0.00002666, 0.00003582, -0.00004366, 0.00005395, -0.00006139, 0.00004263, -0.00002968, 0.00004193, -0.00005223, 0.00004924, -0.00005462, 0.00004986, -0.00005577, 0.00005765, -0.00006593, 0.00007128, -0.00007007, 0.00005627, -0.00005532, 0.00005570, -0.00007852, 0.00007281, -0.00010052, 0.00010718, -0.00014513, 0.00014408, -0.00021414, 0.00020683, -0.00037804, 0.00039400, -0.00152138, 0.00255341, -0.00042177, -0.00051931, 0.00019462, -0.00023961, 0.00013767, -0.00015657, 0.00011218, -0.00012134, 0.00009876, -0.00010089, 0.00007719, -0.00008564, 0.00006868, -0.00005465, 0.00003496, -0.00003405, 0.00003341, -0.00003473, 0.00003132, -0.00003252, 0.00002606, -0.00003096, 0.00002398, -0.00003153, 0.00003476, -0.00001694, 0.00000599, -0.00002473, 0.00003910, -0.00004727, 0.00005008, -0.00005233, 0.00004342, -0.00003640, 0.00003192, -0.00002745, 0.00002136, -0.00001478, 0.00000823, -0.00001772, 0.00002477, -0.00001486, 0.00001082, -0.00000426, 0.00000009, -0.00000626, 0.00000413, -0.00000782, 0.00001022, -0.00000976, 0.00000774, -0.00000678, 0.00000535, -0.00000071, 0.00000167, -0.00000038, -0.00000648, -0.00001252, 0.00001926, 0.00000928, -0.00002020, 0.00000823, -0.00000379, 0.00000443, },
		{ 0.00000615, -0.00000622, 0.00001066, -0.00001792, 0.00000144, 0.00002231, -0.00000732, -0.00001094, 0.00000083, -0.00000091, 0.00000288, 0.00000194, -0.00000264, 0.00000201, -0.00000246, 0.00000126, 0.00000224, -0.00000511, 0.00000242, -0.00000786, 0.00000162, 0.00000359, -0.00000517, 0.00001292, -0.00000387, -0.00000376, -0.00000549, 0.00001053, -0.00001419, 0.00001669, -0.00001891, 0.00002201, -0.00002449, 0.00001413, -0.00000601, -0.00000391, 0.00001752, -0.00003097, 0.00001538, 0.00000088, 0.00000660, -0.00001564, 0.00001101, -0.00001647, 0.00001070, -0.00001397, 0.00001297, -0.00001778, 0.00002066, -0.00001991, 0.00000288, 0.00000484, -0.00001294, -0.00000294, -0.00001170, -0.00000370, -0.00000289, -0.00002236, 0.00000176, -0.00004907, -0.00000017, -0.00011836, -0.00000020, -0.00112098, 0.00276051, -0.00112098, -0.00000020, -0.00011836, -0.00000017, -0.00004907, 0.00000176, -0.00002236, -0.00000289, -0.00000370, -0.00001170, -0.00000294, -0.00001294, 0.00000484, 0.00000288, -0.00001991, 0.00002066, -0.00001778, 0.00001297, -0.00001397, 0.00001070, -0.00001647, 0.00001101, -0.00001564, 0.00000660, 0.00000088, 0.00001538, -0.00003097, 0.00001752, -0.00000391, -0.00000601, 0.00001413, -0.00002449, 0.00002201, -0.00001891, 0.00001669, -0.00001419, 0.00001053, -0.00000549, -0.00000376, -0.00000387, 0.00001292, -0.00000517, 0.00000359, 0.00000162, -0.00000786, 0.00000242, -0.00000511, 0.00000224, 0.00000126, -0.00000246, 0.00000201, -0.00000264, 0.00000194, 0.00000288, -0.00000091, 0.00000083, -0.00001094, -0.00000732, 0.00002231, 0.00000144, -0.00001792, 0.00001066, -0.00000622, },
		{ -0.00000495, 0.00000443, -0.00000379, 0.00000823, -0.00002020, 0.00000928, 0.00001926, -0.00001252, -0.00000648, -0.00000038, 0.00000167, -0.00000071, 0.00000535, -0.00000678, 0.00000774, -0.00000976, 0.00001022, -0.00000782, 0.00000413, -0.00000626, 0.00000009, -0.00000426, 0.00001082, -0.00001486, 0.00002477, -0.00001772, 0.00000823, -0.00001478, 0.00002136, -0.00002745, 0.00003192, -0.00003640, 0.00004342, -0.00005233, 0.00005008, -0.00004727, 0.00003910, -0.00002473, 0.00000599, -0.00001694, 0.00003476, -0.00003153, 0.00002398, -0.00003096, 0.00002606, -0.00003252, 0.00003132, -0.00003473, 0.00003341, -0.00003405, 0.00003496, -0.00005465, 0.00006868, -0.00008564, 0.00007719, -0.00010089, 0.00009876, -0.00012134, 0.00011218, -0.00015657, 0.00013767, -0.00023961, 0.00019462, -0.00051931, -0.00042177, 0.00255341, -0.00152138, 0.00039400, -0.00037804, 0.00020683, -0.00021414, 0.00014408, -0.00014513, 0.00010718, -0.00010052, 0.00007281, -0.00007852, 0.00005570, -0.00005532, 0.00005627, -0.00007007, 0.00007128, -0.00006593, 0.00005765, -0.00005577, 0.00004986, -0.00005462, 0.00004924, -0.00005223, 0.00004193, -0.00002968, 0.00004263, -0.00006139, 0.00005395, -0.00004366, 0.00003582, -0.00002666, 0.00001066, -0.00000700, 0.00000543, -0.00000448, 0.00000441, -0.00000482, 0.00000645, -0.00001677, 0.00001101, -0.00000055, 0.00000669, -0.00000582, 0.00000841, -0.00001556, 0.00001044, -0.00001349, 0.00001191, -0.00000814, 0.00000587, -0.00000500, 0.00000250, -0.00000192, 0.00000692, -0.00000401, 0.00000158, -0.00001454, -0.00000106, 0.00002347, -0.00000739, -0.00001155, 0.00000866, },
		{ 0.00000383, -0.00000101, 0.00000015, 0.00000160, 0.00000160, -0.00001743, 0.00001431, 0.00001545, -0.00001656, -0.00000158, -0.00000193, 0.00000345, -0.00000322, 0.00000746, -0.00000888, 0.00001069, -0.00001397, 0.00001618, -0.00001553, 0.00001164, -0.00001328, 0.00000627, -0.00000732, 0.00001349, -0.00001921, 0.00003137, -0.00002691, 0.00001600, -0.00001858, 0.00002445, -0.00003137, 0.00003639, -0.00004118, 0.00004979, -0.00006314, 0.00006887, -0.00007390, 0.00007093, -0.00005904, 0.00003685, -0.00004260, 0.00006031, -0.00006023, 0.00005364, -0.00006296, 0.00005890, -0.00006552, 0.00006529, -0.00006970, 0.00007047, -0.00007421, 0.00007479, -0.00009539, 0.00011359, -0.00013780, 0.00013508, -0.00016506, 0.00017321, -0.00020972, 0.00021448, -0.00027990, 0.00028534, -0.00043633, 0.00046242, -0.00103543, 0.00045321, 0.00198021, -0.00158079, 0.00058599, -0.00051426, 0.00032356, -0.00030869, 0.00022812, -0.00021872, 0.00017541, -0.00016190, 0.00012622, -0.00012569, 0.00009837, -0.00009160, 0.00008680, -0.00009776, 0.00010085, -0.00009521, 0.00008489, -0.00008077, 0.00007237, -0.00007604, 0.00007158, -0.00007417, 0.00006365, -0.00004688, 0.00005509, -0.00007495, 0.00007279, -0.00006687, 0.00006394, -0.00005831, 0.00004071, -0.00003318, 0.00002790, -0.00002414, 0.00002200, -0.00001955, 0.00001682, -0.00002677, 0.00002267, -0.00001132, 0.00001670, -0.00001416, 0.00001362, -0.00002077, 0.00001552, -0.00001851, 0.00001841, -0.00001520, 0.00001271, -0.00001128, 0.00000720, -0.00000512, 0.00001049, -0.00000721, 0.00000179, -0.00001694, 0.00000603, 0.00002160, -0.00001478, -0.00000350, },
		{ 0.00000337, -0.00000128, 0.00000336, -0.00000452, 0.00000769, -0.00000724, -0.00001047, 0.00001598, 0.00001157, -0.00001920, 0.00000318, -0.00000357, 0.00000430, -0.00000430, 0.00000795, -0.00000864, 0.00001032, -0.00001405, 0.00001758, -0.00001897, 0.00001551, -0.00001703, 0.00000950, -0.00000686, 0.00001085, -0.00001695, 0.00003094, -0.00002957, 0.00001821, -0.00001634, 0.00001961, -0.00002565, 0.00002967, -0.00003257, 0.00003963, -0.00005367, 0.00006462, -0.00007740, 0.00008152, -0.00007493, 0.00005172, -0.00005254, 0.00006879, -0.00007061, 0.00006388, -0.00007486, 0.00007161, -0.00007778, 0.00007734, -0.00008090, 0.00008141, -0.00008657, 0.00008535, -0.00010417, 0.00012312, -0.00015139, 0.00015105, -0.00018240, 0.00019457, -0.00023866, 0.00024995, -0.00032694, 0.00034291, -0.00052385, 0.00058962, -0.00140485, 0.00132995, 0.00117128, -0.00134261, 0.00056530, -0.00050405, 0.00032717, -0.00031150, 0.00023442, -0.00022501, 0.00018482, -0.00017229, 0.00013467, -0.00013203, 0.00010370, -0.00009359, 0.00008489, -0.00009377, 0.00010023, -0.00009692, 0.00008681, -0.00008200, 0.00007190, -0.00007462, 0.00007185, -0.00007550, 0.00006638, -0.00004620, 0.00004896, -0.00006768, 0.00006890, -0.00006651, 0.00006926, -0.00007048, 0.00005544, -0.00004728, 0.00004040, -0.00003520, 0.00003237, -0.00002858, 0.00002173, -0.00003043, 0.00002774, -0.00001585, 0.00002137, -0.00001848, 0.00001512, -0.00002184, 0.00001620, -0.00001864, 0.00001988, -0.00001778, 0.00001593, -0.00001496, 0.00001003, -0.00000652, 0.00001259, -0.00000993, 0.00000133, -0.00001779, 0.00001335, 0.00001623, -0.00001854, },
		{ -0.00001766, 0.00000686, -0.00000428, 0.00000583, -0.00000712, 0.00001164, -0.00001522, -0.00000160, 0.00001500, 0.00000770, -0.00002003, 0.00000702, -0.00000494, 0.00000418, -0.00000395, 0.00000695, -0.00000644, 0.00000709, -0.00001031, 0.00001429, -0.00001757, 0.00001504, -0.00001694, 0.00000970, -0.00000352, 0.00000396, -0.00000891, 0.00002392, -0.00002591, 0.00001551, -0.00000977, 0.00000948, -0.00001349, 0.00001555, -0.00001503, 0.00001781, -0.00002833, 0.00003992, -0.00005785, 0.00006874, -0.00006903, 0.00004712, -0.00004389, 0.00005789, -0.00006051, 0.00005237, -0.00006392, 0.00006120, -0.00006652, 0.00006502, -0.00006617, 0.00006390, -0.00006829, 0.00006341, -0.00007753, 0.00009336, -0.00012187, 0.00012008, -0.00014732, 0.00015593, -0.00019906, 0.00020660, -0.00028185, 0.00028847, -0.00046921, 0.00051716, -0.00151076, 0.00202967, 0.00030187, -0.00091531, 0.00038201, -0.00037274, 0.00023340, -0.00023344, 0.00017093, -0.00016980, 0.00013948, -0.00013464, 0.00010083, -0.00010021, 0.00007423, -0.00006381, 0.00005312, -0.00006041, 0.00007102, -0.00007221, 0.00006465, -0.00006107, 0.00005046, -0.00005240, 0.00005176, -0.00005769, 0.00005181, -0.00003005, 0.00002758, -0.00004342, 0.00004564, -0.00004443, 0.00005126, -0.00006033, 0.00005075, -0.00004490, 0.00003874, -0.00003373, 0.00003186, -0.00002893, 0.00001922, -0.00002643, 0.00002501, -0.00001272, 0.00001898, -0.00001727, 0.00001198, -0.00001837, 0.00001237, -0.00001382, 0.00001599, -0.00001516, 0.00001448, -0.00001494, 0.00001009, -0.00000536, 0.00001243, -0.00001160, 0.00000008, -0.00001665, 0.00001979, 0.00000799, },
		{ -0.00000137, -0.00001275, 0.00000633, -0.00000397, 0.00000508, -0.00000606, 0.00001131, -0.00001943, 0.00000639, 0.00001281, 0.00000352, -0.00001864, 0.00000923, -0.00000566, 0.00000316, -0.00000241, 0.00000492, -0.00000319, 0.00000230, -0.00000414, 0.00000755, -0.00001217, 0.00001079, -0.00001354, 0.00000770, 0.00000106, -0.00000480, 0.00000238, 0.00001253, -0.00001780, 0.00000992, -0.00000183, -0.00000186, 0.00000006, 0.00000015, 0.00000392, -0.00000663, 0.00000281, 0.00000455, -0.00002343, 0.00003881, -0.00004592, 0.00002697, -0.00002086, 0.00003241, -0.00003522, 0.00002464, -0.00003555, 0.00003303, -0.00003744, 0.00003478, -0.00003291, 0.00002606, -0.00002780, 0.00001781, -0.00002498, 0.00003440, -0.00005987, 0.00005372, -0.00007270, 0.00007131, -0.00010566, 0.00009981, -0.00016071, 0.00013839, -0.00028713, 0.00024540, -0.00129983, 0.00241670, -0.00045620, -0.00043498, 0.00012258, -0.00017874, 0.00008639, -0.00010993, 0.00006730, -0.00007833, 0.00006092, -0.00006755, 0.00004139, -0.00004563, 0.00002429, -0.00001578, 0.00000450, -0.00000995, 0.00002427, -0.00003091, 0.00002755, -0.00002690, 0.00001703, -0.00001824, 0.00001963, -0.00002848, 0.00002740, -0.00000623, -0.00000020, -0.00001205, 0.00001329, -0.00001039, 0.00001788, -0.00003328, 0.00002993, -0.00002807, 0.00002424, -0.00002056, 0.00002084, -0.00002087, 0.00001009, -0.00001615, 0.00001599, -0.00000316, 0.00001023, -0.00001092, 0.00000485, -0.00001137, 0.00000536, -0.00000558, 0.00000811, -0.00000825, 0.00000877, -0.00001127, 0.00000734, -0.00000156, 0.00000965, -0.00001181, -0.00000204, -0.00001318, 0.00002390, },
		{ 0.00002440, -0.00000947, -0.00000584, 0.00000282, -0.00000088, 0.00000142, -0.00000147, 0.00000619, -0.00001830, 0.00001126, 0.00001085, -0.00000121, -0.00001494, 0.00000940, -0.00000550, 0.00000145, -0.00000010, 0.00000251, 0.00000002, -0.00000244, 0.00000241, -0.00000052, -0.00000464, 0.00000429, -0.00000809, 0.00000479, 0.00000495, -0.00001266, 0.00001370, -0.00000010, -0.00000786, 0.00000377, 0.00000465, -0.00001068, 0.00001041, -0.00001089, 0.00001707, -0.00002437, 0.00002818, -0.00002860, 0.00001312, 0.00000331, -0.00001577, 0.00000054, 0.00000731, 0.00000208, -0.00000512, -0.00000837, -0.00000092, -0.00000165, -0.00000216, -0.00000097, 0.00000515, -0.00001700, 0.00001872, -0.00003420, 0.00003483, -0.00003367, 0.00001309, -0.00002457, 0.00001518, -0.00002975, 0.00000854, -0.00003336, -0.00000598, -0.00005763, -0.00003717, -0.00016186, -0.00080000, 0.00243313, -0.00097840, -0.00002423, -0.00012238, 0.00001057, -0.00006022, 0.00001437, -0.00003806, 0.00001582, -0.00002149, 0.00000321, -0.00002055, 0.00001047, -0.00002640, 0.00003200, -0.00004322, 0.00004060, -0.00002429, 0.00001279, -0.00001151, 0.00000825, -0.00001636, 0.00001604, -0.00001331, 0.00000175, 0.00000287, 0.00001564, -0.00002408, 0.00001485, -0.00001532, 0.00002208, -0.00001793, -0.00000044, 0.00000203, -0.00000413, 0.00000307, -0.00000089, 0.00000344, -0.00000774, -0.00000239, -0.00000313, 0.00000436, 0.00000934, -0.00000204, -0.00000163, -0.00000424, -0.00000297, -0.00000246, 0.00000348, -0.00000118, 0.00000067, 0.00000056, -0.00000513, 0.00000271, 0.00000413, 0.00000452, -0.00001053, -0.00000483, -0.00000740, },
		{ 0.00000004, 0.00002073, -0.00001415, 0.00000042, -0.00000146, 0.00000317, -0.00000335, 0.00000482, -0.00000223, -0.00001215, 0.00001216, 0.00000988, -0.00000633, -0.00000938, 0.00000760, -0.00000455, -0.00000055, 0.00000248, 0.00000029, 0.00000230, -0.00000571, 0.00000738, -0.00000760, 0.00000277, -0.00000250, -0.00000215, 0.00000219, 0.00000667, -0.00001735, 0.00002219, -0.00001107, 0.00000153, -0.00000123, 0.00000811, -0.00001503, 0.00001522, -0.00001474, 0.00002056, -0.00002964, 0.00003935, -0.00004866, 0.00003968, -0.00002564, 0.00001012, -0.00002169, 0.00003035, -0.00002255, 0.00001864, -0.00003477, 0.00002767, -0.00003041, 0.00002666, -0.00002895, 0.00003355, -0.00004931, 0.00005394, -0.00007389, 0.00008148, -0.00008861, 0.00007302, -0.00008856, 0.00008681, -0.00011353, 0.00010489, -0.00014819, 0.00012936, -0.00023444, 0.00019570, -0.00059479, -0.00011337, 0.00210941, -0.00121365, 0.00023919, -0.00028653, 0.00014243, -0.00016280, 0.00010231, -0.00011257, 0.00008356, -0.00008184, 0.00005482, -0.00006459, 0.00004945, -0.00006063, 0.00006337, -0.00007444, 0.00007603, -0.00006035, 0.00004590, -0.00004082, 0.00003362, -0.00003937, 0.00004026, -0.00003735, 0.00002412, -0.00001388, 0.00002829, -0.00003656, 0.00002873, -0.00003004, 0.00004116, -0.00004344, 0.00002602, -0.00002213, 0.00001746, -0.00001645, 0.00001794, -0.00001420, 0.00000549, -0.00001381, 0.00000824, -0.00000538, 0.00002033, -0.00001393, 0.00000749, -0.00001263, 0.00000436, -0.00000858, 0.00001058, -0.00000894, 0.00000874, -0.00000768, 0.00000150, -0.00000219, 0.00001029, -0.00000203, -0.00000815, -0.00000772, },
		{ -0.00000981, 0.00000780, 0.00001351, -0.00001430, 0.00000386, -0.00000421, 0.00000592, -0.00000681, 0.00001003, -0.00001102, -0.00000310, 0.00000977, 0.00000979, -0.00001113, -0.00000296, 0.00000445, -0.00000318, -0.00000235, 0.00000479, -0.00000132, 0.00000317, -0.00000673, 0.00000945, -0.00001190, 0.00000810, -0.00000780, 0.00000293, 0.00000064, 0.00000562, -0.00001777, 0.00002612, -0.00001835, 0.00000879, -0.00000447, 0.00000857, -0.00001521, 0.00001506, -0.00001227, 0.00001524, -0.00002239, 0.00003417, -0.00005076, 0.00004924, -0.00003998, 0.00002374, -0.00003228, 0.00004114, -0.00003432, 0.00002852, -0.00004637, 0.00004153, -0.00004447, 0.00004030, -0.00004035, 0.00004282, -0.00006020, 0.00006602, -0.00008827, 0.00010056, -0.00011445, 0.00010253, -0.00011929, 0.00012064, -0.00015485, 0.00015367, -0.00020923, 0.00020162, -0.00033545, 0.00033045, -0.00093093, 0.00061337, 0.00154905, -0.00118641, 0.00033532, -0.00034351, 0.00019334, -0.00020081, 0.00013580, -0.00013988, 0.00010980, -0.00010634, 0.00007509, -0.00008004, 0.00006169, -0.00006958, 0.00007018, -0.00008114, 0.00008807, -0.00007577, 0.00006103, -0.00005398, 0.00004370, -0.00004691, 0.00004931, -0.00004757, 0.00003435, -0.00001956, 0.00002947, -0.00003588, 0.00002738, -0.00002738, 0.00004124, -0.00005110, 0.00003768, -0.00003430, 0.00002904, -0.00002721, 0.00002935, -0.00002635, 0.00001427, -0.00002040, 0.00001444, -0.00000961, 0.00002600, -0.00002189, 0.00001357, -0.00001807, 0.00000873, -0.00001123, 0.00001375, -0.00001288, 0.00001351, -0.00001357, 0.00000661, -0.00000560, 0.00001524, -0.00000868, -0.00000542, },
		{ -0.00000317, -0.00001016, 0.00001408, 0.00000450, -0.00001035, 0.00000360, -0.00000405, 0.00000580, -0.00000705, 0.00001164, -0.00001699, 0.00000584, 0.00000580, 0.00000977, -0.00001457, 0.00000298, 0.00000091, -0.00000197, -0.00000341, 0.00000641, -0.00000218, 0.00000271, -0.00000551, 0.00000832, -0.00001258, 0.00001021, -0.00001043, 0.00000629, 0.00000028, 0.00000217, -0.00001413, 0.00002516, -0.00002116, 0.00001327, -0.00000616, 0.00000721, -0.00001306, 0.00001255, -0.00000716, 0.00000587, -0.00000789, 0.00001707, -0.00003732, 0.00004198, -0.00003826, 0.00002304, -0.00002940, 0.00003823, -0.00003198, 0.00002323, -0.00004144, 0.00003855, -0.00004172, 0.00003700, -0.00003393, 0.00003184, -0.00004810, 0.00005282, -0.00007468, 0.00008866, -0.00010694, 0.00009685, -0.00011166, 0.00011072, -0.00014601, 0.00014474, -0.00020308, 0.00019247, -0.00033362, 0.00031875, -0.00107302, 0.00123102, 0.00089368, -0.00097672, 0.00029497, -0.00030645, 0.00017090, -0.00017932, 0.00011865, -0.00012277, 0.00009634, -0.00009609, 0.00006510, -0.00006825, 0.00004869, -0.00005491, 0.00005417, -0.00006473, 0.00007732, -0.00007047, 0.00005803, -0.00005127, 0.00003936, -0.00004014, 0.00004420, -0.00004479, 0.00003350, -0.00001599, 0.00002217, -0.00002610, 0.00001530, -0.00001138, 0.00002470, -0.00004094, 0.00003280, -0.00003184, 0.00002763, -0.00002608, 0.00003001, -0.00002987, 0.00001623, -0.00002057, 0.00001417, -0.00000696, 0.00002455, -0.00002395, 0.00001504, -0.00001945, 0.00000944, -0.00000998, 0.00001246, -0.00001214, 0.00001372, -0.00001564, 0.00000882, -0.00000622, 0.00001749, -0.00001409, },
		{ -0.00001731, -0.00000197, -0.00000815, 0.00001718, -0.00000383, -0.00000415, 0.00000038, -0.00000114, 0.00000271, -0.00000364, 0.00000850, -0.00001791, 0.00001193, 0.00000224, 0.00000887, -0.00001574, 0.00000724, -0.00000208, -0.00000142, -0.00000336, 0.00000708, -0.00000237, 0.00000140, -0.00000283, 0.00000469, -0.00000997, 0.00000900, -0.00001003, 0.00000764, 0.00000078, -0.00000259, -0.00000769, 0.00002020, -0.00001975, 0.00001497, -0.00000681, 0.00000540, -0.00001069, 0.00001064, -0.00000372, -0.00000144, 0.00000597, -0.00000328, -0.00001630, 0.00002432, -0.00002532, 0.00001191, -0.00001683, 0.00002567, -0.00002001, 0.00000757, -0.00002473, 0.00002324, -0.00002674, 0.00002187, -0.00001574, 0.00000754, -0.00002035, 0.00002188, -0.00004092, 0.00005372, -0.00007403, 0.00006438, -0.00007515, 0.00006764, -0.00009829, 0.00008991, -0.00014213, 0.00011469, -0.00024103, 0.00016222, -0.00097848, 0.00162833, 0.00028124, -0.00068737, 0.00017831, -0.00021435, 0.00010439, -0.00012192, 0.00007081, -0.00007853, 0.00005802, -0.00006386, 0.00003632, -0.00003999, 0.00002066, -0.00002635, 0.00002479, -0.00003412, 0.00005155, -0.00005098, 0.00004267, -0.00003820, 0.00002632, -0.00002491, 0.00003044, -0.00003402, 0.00002631, -0.00000822, 0.00001241, -0.00001458, 0.00000113, 0.00000865, 0.00000028, -0.00001985, 0.00001627, -0.00001819, 0.00001575, -0.00001480, 0.00002083, -0.00002513, 0.00001190, -0.00001530, 0.00000877, 0.00000131, 0.00001667, -0.00002028, 0.00001200, -0.00001712, 0.00000719, -0.00000585, 0.00000778, -0.00000752, 0.00000970, -0.00001377, 0.00000784, -0.00000366, 0.00001631, },
		{ 0.00001188, -0.00001805, -0.00000183, -0.00000388, 0.00001610, -0.00000911, 0.00000175, -0.00000383, 0.00000302, -0.00000196, 0.00000213, 0.00000142, -0.00001341, 0.00001367, 0.00000044, 0.00000657, -0.00001431, 0.00000914, -0.00000393, -0.00000173, -0.00000214, 0.00000677, -0.00000214, -0.00000001, 0.00000013, -0.00000004, -0.00000530, 0.00000533, -0.00000705, 0.00000719, 0.00000155, -0.00000735, -0.00000020, 0.00001283, -0.00001507, 0.00001417, -0.00000672, 0.00000383, -0.00000915, 0.00001091, -0.00000455, -0.00000228, 0.00001234, -0.00001799, 0.00000263, 0.00000550, -0.00000935, -0.00000234, -0.00000157, 0.00001067, -0.00000610, -0.00001019, -0.00000473, 0.00000401, -0.00000803, 0.00000389, 0.00000413, -0.00001863, 0.00001059, -0.00001349, -0.00000122, 0.00001077, -0.00003143, 0.00002194, -0.00002859, 0.00001281, -0.00003589, 0.00001648, -0.00005783, 0.00000547, -0.00010344, -0.00008796, -0.00067061, 0.00175818, -0.00018912, -0.00041048, 0.00004864, -0.00011251, 0.00002940, -0.00005805, 0.00001773, -0.00002960, 0.00001463, -0.00002695, 0.00000422, -0.00000955, -0.00000899, 0.00000338, -0.00000564, -0.00000118, 0.00002159, -0.00002672, 0.00002317, -0.00002226, 0.00001189, -0.00000857, 0.00001499, -0.00002147, 0.00001832, -0.00000151, 0.00000590, -0.00000823, -0.00000664, 0.00002254, -0.00002111, 0.00000193, -0.00000308, -0.00000082, -0.00000019, 0.00000112, 0.00000618, -0.00001543, 0.00000417, -0.00000761, 0.00000149, 0.00001190, 0.00000516, -0.00001286, 0.00000606, -0.00001259, 0.00000364, -0.00000082, 0.00000187, -0.00000109, 0.00000315, -0.00000914, 0.00000457, 0.00000135, },
		{ 0.00000727, 0.00000527, -0.00001677, -0.00000219, 0.00000164, 0.00001109, -0.00000999, 0.00000508, -0.00000680, 0.00000629, -0.00000597, 0.00000775, -0.00000713, -0.00000504, 0.00001123, 0.00000070, 0.00000312, -0.00001072, 0.00000874, -0.00000459, -0.00000272, -0.00000001, 0.00000569, -0.00000180, -0.00000088, 0.00000218, -0.00000423, -0.00000027, 0.00000067, -0.00000254, 0.00000547, 0.00000205, -0.00001094, 0.00000663, 0.00000486, -0.00000838, 0.00001124, -0.00000569, 0.00000221, -0.00000800, 0.00001269, -0.00000919, 0.00000398, 0.00000845, -0.00002168, 0.00001225, -0.00000654, 0.00000195, -0.00001263, 0.00000960, 0.00000000, 0.00000260, -0.00002219, 0.00001024, -0.00001074, 0.00000608, -0.00000838, 0.00001628, -0.00003590, 0.00003265, -0.00004015, 0.00003014, -0.00002477, 0.00000457, -0.00001307, 0.00000853, -0.00003127, 0.00001509, -0.00004507, 0.00001342, -0.00008990, 0.00001818, -0.00034908, -0.00022831, 0.00164218, -0.00047530, -0.00020426, -0.00004866, -0.00003510, -0.00002651, -0.00001082, -0.00002044, 0.00000591, -0.00001757, 0.00000032, -0.00001853, 0.00001145, -0.00002944, 0.00002406, -0.00002710, 0.00002415, -0.00000322, -0.00000589, 0.00000650, -0.00000947, 0.00000167, 0.00000335, 0.00000311, -0.00001170, 0.00001317, 0.00000135, 0.00000512, -0.00001007, -0.00000359, 0.00002379, -0.00003107, 0.00001537, -0.00001662, 0.00001235, -0.00001293, 0.00001501, -0.00000819, -0.00000529, -0.00000323, -0.00000105, -0.00000396, 0.00002085, -0.00000633, -0.00000458, -0.00000053, -0.00000779, 0.00000063, 0.00000305, -0.00000289, 0.00000463, -0.00000358, -0.00000371, 0.00000066, },
		{ -0.00000211, 0.00001227, -0.00000196, -0.00001434, -0.00000210, 0.00000676, 0.00000365, -0.00000660, 0.00000476, -0.00000700, 0.00000698, -0.00000727, 0.00001062, -0.00001400, 0.00000436, 0.00000617, 0.00000231, -0.00000058, -0.00000606, 0.00000676, -0.00000450, -0.00000385, 0.00000247, 0.00000418, -0.00000158, -0.00000085, 0.00000255, -0.00000657, 0.00000350, -0.00000335, 0.00000221, 0.00000314, 0.00000195, -0.00001272, 0.00001157, -0.00000211, -0.00000105, 0.00000666, -0.00000339, -0.00000033, -0.00000571, 0.00001356, -0.00001440, 0.00001389, -0.00000339, -0.00001448, 0.00001048, -0.00000826, 0.00000454, -0.00001509, 0.00001311, -0.00000287, 0.00000245, -0.00002422, 0.00001543, -0.00001619, 0.00001095, -0.00001054, 0.00001597, -0.00003856, 0.00003909, -0.00005042, 0.00004444, -0.00004325, 0.00002368, -0.00002984, 0.00002417, -0.00005035, 0.00003750, -0.00007393, 0.00004554, -0.00013693, 0.00007176, -0.00053555, 0.00024106, 0.00135443, -0.00059035, -0.00008574, -0.00009599, 0.00000393, -0.00005189, 0.00001020, -0.00003527, 0.00002015, -0.00003127, 0.00001156, -0.00002643, 0.00001810, -0.00003616, 0.00003144, -0.00003527, 0.00003717, -0.00001811, 0.00000728, -0.00000398, -0.00000228, -0.00000237, 0.00000889, -0.00000332, -0.00000607, 0.00001131, 0.00000106, 0.00000838, -0.00001807, 0.00000878, 0.00001215, -0.00002705, 0.00001629, -0.00001950, 0.00001633, -0.00001745, 0.00002184, -0.00001765, 0.00000157, -0.00000742, 0.00000187, -0.00000493, 0.00002511, -0.00001469, 0.00000198, -0.00000579, -0.00000421, -0.00000061, 0.00000441, -0.00000486, 0.00000768, -0.00000838, 0.00000061, },
		{ 0.00000265, -0.00000256, 0.00001485, -0.00000831, -0.00001169, -0.00000071, 0.00000965, -0.00000383, -0.00000055, 0.00000125, -0.00000428, 0.00000464, -0.00000503, 0.00000916, -0.00001660, 0.00001181, 0.00000065, 0.00000410, -0.00000339, -0.00000171, 0.00000425, -0.00000432, -0.00000446, 0.00000466, 0.00000261, -0.00000152, 0.00000002, 0.00000112, -0.00000646, 0.00000496, -0.00000547, 0.00000600, 0.00000084, 0.00000118, -0.00001255, 0.00001400, -0.00000690, 0.00000563, 0.00000119, 0.00000021, -0.00000445, -0.00000084, 0.00001081, -0.00001586, 0.00002173, -0.00001708, -0.00000144, 0.00000066, -0.00000136, -0.00000085, -0.00001025, 0.00000965, 0.00000104, -0.00000542, -0.00001708, 0.00001117, -0.00001256, 0.00000712, -0.00000372, 0.00000469, -0.00002779, 0.00003053, -0.00004445, 0.00004134, -0.00004379, 0.00002498, -0.00002785, 0.00001793, -0.00004347, 0.00002929, -0.00006661, 0.00003282, -0.00012576, 0.00003434, -0.00058766, 0.00063556, 0.00099198, -0.00058444, -0.00003856, -0.00010084, 0.00001013, -0.00005132, 0.00000916, -0.00003045, 0.00001612, -0.00002875, 0.00000884, -0.00002168, 0.00001261, -0.00003128, 0.00002762, -0.00003193, 0.00003889, -0.00002342, 0.00001300, -0.00000881, 0.00000047, -0.00000184, 0.00000957, -0.00000565, -0.00000310, 0.00001066, 0.00000080, 0.00001106, -0.00002637, 0.00002405, -0.00000665, -0.00001284, 0.00000654, -0.00001223, 0.00001084, -0.00001286, 0.00002003, -0.00002030, 0.00000362, -0.00000752, 0.00000067, -0.00000091, 0.00002358, -0.00001837, 0.00000543, -0.00000876, -0.00000236, 0.00000006, 0.00000318, -0.00000369, 0.00000728, -0.00001013, },
		{ -0.00000901, 0.00000234, -0.00000047, 0.00001435, -0.00001280, -0.00000937, 0.00000228, 0.00000903, -0.00000887, 0.00000565, -0.00000369, 0.00000012, 0.00000030, -0.00000006, 0.00000356, -0.00001393, 0.00001527, -0.00000348, 0.00000505, -0.00000448, 0.00000125, 0.00000212, -0.00000461, -0.00000408, 0.00000605, 0.00000128, -0.00000153, 0.00000125, -0.00000154, -0.00000417, 0.00000394, -0.00000506, 0.00000798, -0.00000094, -0.00000014, -0.00001079, 0.00001398, -0.00000891, 0.00001046, -0.00000405, 0.00000442, -0.00000979, 0.00000676, 0.00000321, -0.00001058, 0.00002227, -0.00002569, 0.00001013, -0.00001064, 0.00000877, -0.00000986, -0.00000206, 0.00000323, 0.00000740, -0.00001624, -0.00000551, 0.00000190, -0.00000416, -0.00000077, 0.00000662, -0.00001122, -0.00001029, 0.00001368, -0.00002900, 0.00002762, -0.00003316, 0.00001565, -0.00001547, -0.00000037, -0.00002169, 0.00000263, -0.00003666, -0.00000939, -0.00007416, -0.00007936, -0.00049015, 0.00088733, 0.00064346, -0.00051942, -0.00002994, -0.00008460, 0.00000034, -0.00003882, -0.00000157, -0.00001711, 0.00000362, -0.00001846, -0.00000031, -0.00001136, 0.00000163, -0.00002108, 0.00001863, -0.00002274, 0.00003412, -0.00002293, 0.00001438, -0.00001090, 0.00000194, -0.00000019, 0.00000871, -0.00000677, -0.00000020, 0.00000856, 0.00000396, 0.00000838, -0.00002847, 0.00003408, -0.00002387, 0.00000368, -0.00000752, 0.00000016, -0.00000009, -0.00000230, 0.00001168, -0.00001728, 0.00000182, -0.00000475, -0.00000306, 0.00000645, 0.00001734, -0.00001769, 0.00000583, -0.00000962, -0.00000175, 0.00000175, 0.00000046, -0.00000040, 0.00000408, },
		{ -0.00000027, -0.00000625, 0.00000065, 0.00000329, 0.00001112, -0.00001520, -0.00000732, 0.00000633, 0.00000473, -0.00000981, 0.00000962, -0.00000781, 0.00000416, -0.00000400, 0.00000551, -0.00000423, -0.00000691, 0.00001430, -0.00000537, 0.00000482, -0.00000370, 0.00000238, 0.00000082, -0.00000551, -0.00000259, 0.00000643, 0.00000030, -0.00000136, 0.00000213, -0.00000438, -0.00000071, 0.00000111, -0.00000232, 0.00000787, -0.00000193, -0.00000174, -0.00000807, 0.00001205, -0.00000808, 0.00001257, -0.00000774, 0.00000783, -0.00001482, 0.00001543, -0.00000794, 0.00000138, 0.00001341, -0.00002470, 0.00001412, -0.00001686, 0.00001600, -0.00001707, 0.00000455, -0.00000131, 0.00001102, -0.00002425, 0.00000446, -0.00000650, 0.00000329, -0.00000687, 0.00001383, -0.00002402, 0.00000542, -0.00000251, -0.00001348, 0.00001306, -0.00002139, 0.00000629, -0.00000477, -0.00001651, -0.00000154, -0.00002333, -0.00000672, -0.00005335, -0.00001866, -0.00022315, -0.00027439, 0.00097720, 0.00036615, -0.00044612, -0.00002863, -0.00006903, -0.00000806, -0.00002895, -0.00000915, -0.00000696, -0.00000678, -0.00000962, -0.00000772, -0.00000299, -0.00000778, -0.00001214, 0.00001083, -0.00001386, 0.00002836, -0.00002114, 0.00001499, -0.00001321, 0.00000500, -0.00000046, 0.00000919, -0.00000901, 0.00000437, 0.00000378, 0.00001177, -0.00000175, -0.00002065, 0.00003303, -0.00003186, 0.00001439, -0.00001832, 0.00001092, -0.00001044, 0.00000891, 0.00000111, -0.00001176, -0.00000134, -0.00000160, -0.00000659, 0.00001421, 0.00000892, -0.00001435, 0.00000435, -0.00000935, -0.00000133, 0.00000307, -0.00000203, 0.00000318, },
		{ 0.00000526, -0.00000391, -0.00000349, -0.00000093, 0.00000725, 0.00000624, -0.00001593, -0.00000500, 0.00001011, -0.00000212, -0.00000643, 0.00000996, -0.00000938, 0.00000605, -0.00000618, 0.00000911, -0.00001130, 0.00000203, 0.00000999, -0.00000525, 0.00000384, -0.00000169, 0.00000199, 0.00000023, -0.00000674, -0.00000033, 0.00000594, -0.00000038, -0.00000080, 0.00000203, -0.00000628, 0.00000257, -0.00000226, 0.00000186, 0.00000591, -0.00000209, -0.00000335, -0.00000510, 0.00000912, -0.00000499, 0.00001166, -0.00000881, 0.00000890, -0.00001734, 0.00002218, -0.00001906, 0.00001653, -0.00000273, -0.00001396, 0.00000829, -0.00001441, 0.00001631, -0.00001874, 0.00000622, -0.00000081, 0.00000854, -0.00002551, 0.00000845, -0.00000967, 0.00000565, -0.00000716, 0.00001348, -0.00002831, 0.00001310, -0.00001124, -0.00000516, 0.00000541, -0.00001643, 0.00000513, -0.00000501, -0.00001935, 0.00000347, -0.00003226, 0.00000301, -0.00007279, 0.00000216, -0.00033757, -0.00000452, 0.00093168, 0.00017784, -0.00039141, -0.00001705, -0.00006633, -0.00000537, -0.00002973, -0.00000646, -0.00000670, -0.00000878, -0.00000776, -0.00000861, -0.00000094, -0.00001159, -0.00000815, 0.00000782, -0.00000896, 0.00002508, -0.00002083, 0.00001679, -0.00001695, 0.00001053, -0.00000363, 0.00001193, -0.00001285, 0.00001030, -0.00000246, 0.00002223, -0.00001734, -0.00000403, 0.00002005, -0.00002743, 0.00001459, -0.00002065, 0.00001482, -0.00001511, 0.00001577, -0.00000725, -0.00000717, -0.00000325, -0.00000038, -0.00000739, 0.00001946, 0.00000115, -0.00001050, 0.00000249, -0.00000905, -0.00000014, 0.00000288, -0.00000280, },
		{ -0.00000125, 0.00000481, -0.00000551, -0.00000206, -0.00000110, 0.00000990, 0.00000110, -0.00001569, -0.00000173, 0.00001197, -0.00000936, -0.00000001, 0.00000678, -0.00000790, 0.00000499, -0.00000508, 0.00000891, -0.00001496, 0.00000999, 0.00000432, -0.00000412, 0.00000304, 0.00000045, 0.00000088, -0.00000020, -0.00000767, 0.00000210, 0.00000492, -0.00000084, 0.00000019, 0.00000066, -0.00000643, 0.00000444, -0.00000477, 0.00000620, 0.00000280, -0.00000162, -0.00000472, -0.00000253, 0.00000623, -0.00000069, 0.00000819, -0.00000696, 0.00000668, -0.00001561, 0.00002400, -0.00002569, 0.00002934, -0.00002058, 0.00000227, -0.00000499, -0.00000399, 0.00000940, -0.00001436, 0.00000266, 0.00000474, 0.00000000, -0.00001961, 0.00000555, -0.00000661, 0.00000221, -0.00000137, 0.00000539, -0.00002334, 0.00001138, -0.00001078, -0.00000607, 0.00000695, -0.00002047, 0.00001400, -0.00001810, -0.00000618, -0.00001070, -0.00001850, -0.00001551, -0.00005562, -0.00003485, -0.00037263, 0.00024638, 0.00080595, 0.00006336, -0.00035758, 0.00000580, -0.00007611, 0.00000779, -0.00004026, 0.00000564, -0.00001585, -0.00000242, -0.00001288, -0.00000320, -0.00000494, -0.00001007, -0.00000874, 0.00000929, -0.00000814, 0.00002459, -0.00002214, 0.00001939, -0.00002110, 0.00001712, -0.00000835, 0.00001571, -0.00001682, 0.00001554, -0.00000702, 0.00003090, -0.00003297, 0.00001580, -0.00000026, -0.00001308, 0.00000479, -0.00001376, 0.00001041, -0.00001211, 0.00001580, -0.00001078, -0.00000555, -0.00000255, -0.00000209, -0.00000436, 0.00002052, -0.00000403, -0.00000770, 0.00000126, -0.00000929, 0.00000209, 0.00000088, },
		{ -0.00000222, 0.00000203, 0.00000203, -0.00000484, -0.00000237, 0.00000069, 0.00001026, -0.00000317, -0.00001506, 0.00000281, 0.00001060, -0.00001457, 0.00000714, 0.00000155, -0.00000428, 0.00000165, -0.00000108, 0.00000463, -0.00001377, 0.00001467, -0.00000074, -0.00000305, 0.00000329, 0.00000164, -0.00000002, -0.00000105, -0.00000771, 0.00000402, 0.00000382, -0.00000117, 0.00000140, -0.00000179, -0.00000464, 0.00000422, -0.00000533, 0.00000944, -0.00000056, -0.00000080, -0.00000573, -0.00000074, 0.00000431, 0.00000346, 0.00000338, -0.00000291, 0.00000150, -0.00000932, 0.00001940, -0.00002460, 0.00003457, -0.00003357, 0.00001738, -0.00002006, 0.00001013, -0.00000164, -0.00000649, -0.00000355, 0.00001243, -0.00001142, -0.00000953, -0.00000166, 0.00000027, -0.00000437, 0.00000719, -0.00000661, -0.00001294, 0.00000381, -0.00000464, -0.00001266, 0.00001407, -0.00002948, 0.00002786, -0.00003798, 0.00001640, -0.00003709, 0.00001059, -0.00005447, -0.00001000, -0.00012647, -0.00030628, 0.00042085, 0.00066040, -0.00000936, -0.00033112, 0.00002934, -0.00008891, 0.00002318, -0.00005302, 0.00002026, -0.00002843, 0.00000734, -0.00002062, 0.00000444, -0.00001122, -0.00000673, -0.00001050, 0.00001199, -0.00000861, 0.00002469, -0.00002313, 0.00002084, -0.00002331, 0.00002206, -0.00001202, 0.00001819, -0.00001869, 0.00001765, -0.00000666, 0.00003304, -0.00004234, 0.00003125, -0.00002005, 0.00000446, -0.00000994, -0.00000131, 0.00000032, -0.00000320, 0.00000980, -0.00000954, -0.00000691, 0.00000035, -0.00000597, 0.00000176, 0.00001754, -0.00000614, -0.00000646, 0.00000088, -0.00000989, 0.00000480, },
		{ 0.00000687, -0.00000502, 0.00000560, -0.00000185, -0.00000271, -0.00000385, 0.00000398, 0.00000828, -0.00000604, -0.00001414, 0.00000821, 0.00000560, -0.00001590, 0.00001256, -0.00000358, -0.00000030, -0.00000224, 0.00000407, -0.00000231, -0.00000809, 0.00001514, -0.00000399, -0.00000263, 0.00000492, 0.00000127, -0.00000015, -0.00000264, -0.00000656, 0.00000503, 0.00000294, -0.00000133, 0.00000241, -0.00000465, -0.00000145, 0.00000204, -0.00000354, 0.00001071, -0.00000335, 0.00000004, -0.00000638, 0.00000017, 0.00000394, 0.00000616, -0.00000112, 0.00000167, -0.00000495, -0.00000009, 0.00000938, -0.00001540, 0.00002965, -0.00003689, 0.00002528, -0.00003067, 0.00002234, -0.00001218, 0.00000091, -0.00000859, 0.00001813, -0.00002115, 0.00000009, -0.00000884, 0.00000682, -0.00000985, 0.00001367, -0.00001685, -0.00000314, -0.00000349, 0.00000083, -0.00001827, 0.00001984, -0.00003593, 0.00003787, -0.00005409, 0.00003609, -0.00006169, 0.00003876, -0.00009436, 0.00003908, -0.00024338, -0.00015175, 0.00049343, 0.00054006, -0.00007146, -0.00029543, 0.00004046, -0.00009329, 0.00003093, -0.00005904, 0.00002909, -0.00003695, 0.00001408, -0.00002541, 0.00000923, -0.00001513, -0.00000576, -0.00000938, 0.00001197, -0.00000683, 0.00002249, -0.00002147, 0.00001908, -0.00002145, 0.00002296, -0.00001232, 0.00001748, -0.00001702, 0.00001550, -0.00000001, 0.00002619, -0.00004127, 0.00003622, -0.00003173, 0.00001745, -0.00002265, 0.00001072, -0.00001034, 0.00000728, 0.00000108, -0.00000577, -0.00000949, 0.00000367, -0.00000994, 0.00000880, 0.00001219, -0.00000607, -0.00000624, 0.00000078, -0.00001008, },
		{ -0.00000906, 0.00000724, -0.00000612, 0.00000778, -0.00000515, -0.00000055, -0.00000526, 0.00000759, 0.00000471, -0.00000765, -0.00001254, 0.00001333, -0.00000215, -0.00001279, 0.00001458, -0.00000675, 0.00000218, -0.00000463, 0.00000797, -0.00000930, 0.00000009, 0.00001206, -0.00000532, -0.00000270, 0.00000752, -0.00000051, 0.00000050, -0.00000481, -0.00000433, 0.00000505, 0.00000239, -0.00000125, 0.00000272, -0.00000697, 0.00000207, -0.00000122, 0.00000018, 0.00000976, -0.00000505, 0.00000067, -0.00000680, 0.00000053, 0.00000514, 0.00000659, -0.00000381, 0.00000485, -0.00001021, 0.00000901, -0.00000269, -0.00000098, 0.00001591, -0.00002958, 0.00002303, -0.00003267, 0.00002828, -0.00001836, 0.00000456, -0.00000946, 0.00001865, -0.00002550, 0.00000528, -0.00001217, 0.00000952, -0.00001082, 0.00001427, -0.00002077, 0.00000096, -0.00000521, 0.00000006, -0.00001690, 0.00001798, -0.00003315, 0.00003628, -0.00005692, 0.00004140, -0.00007071, 0.00004918, -0.00011372, 0.00006106, -0.00034112, 0.00004848, 0.00047412, 0.00046351, -0.00014047, -0.00024128, 0.00003164, -0.00008246, 0.00002517, -0.00005290, 0.00002691, -0.00003646, 0.00001353, -0.00002367, 0.00000800, -0.00001391, -0.00000955, -0.00000317, 0.00000689, -0.00000059, 0.00001620, -0.00001593, 0.00001341, -0.00001495, 0.00001910, -0.00000857, 0.00001330, -0.00001214, 0.00001018, 0.00001131, 0.00001159, -0.00002987, 0.00002889, -0.00003131, 0.00002058, -0.00002777, 0.00001698, -0.00001650, 0.00001460, -0.00000633, -0.00000247, -0.00001103, 0.00000531, -0.00001174, 0.00001418, 0.00000678, -0.00000539, -0.00000599, 0.00000014, },
		{ -0.00000152, -0.00000651, 0.00000548, -0.00000477, 0.00000746, -0.00000650, 0.00000032, -0.00000529, 0.00001007, 0.00000078, -0.00000855, -0.00000967, 0.00001658, -0.00001071, -0.00000617, 0.00001291, -0.00000710, 0.00000212, -0.00000418, 0.00000863, -0.00001364, 0.00000804, 0.00000714, -0.00000549, -0.00000249, 0.00001011, -0.00000298, 0.00000146, -0.00000701, -0.00000158, 0.00000435, 0.00000211, -0.00000083, 0.00000196, -0.00000790, 0.00000472, -0.00000428, 0.00000478, 0.00000699, -0.00000551, 0.00000099, -0.00000726, 0.00000091, 0.00000731, 0.00000478, -0.00000394, 0.00000528, -0.00001211, 0.00001469, -0.00001241, 0.00001353, -0.00000190, -0.00001485, 0.00001183, -0.00002563, 0.00002657, -0.00001877, 0.00000344, -0.00000548, 0.00001332, -0.00002336, 0.00000458, -0.00001019, 0.00000720, -0.00000654, 0.00000827, -0.00001716, -0.00000225, 0.00000039, -0.00000884, -0.00000644, 0.00000632, -0.00001905, 0.00002062, -0.00004296, 0.00002739, -0.00005759, 0.00003297, -0.00009990, 0.00003392, -0.00037738, 0.00023887, 0.00039908, 0.00042362, -0.00021658, -0.00017109, 0.00000469, -0.00005746, 0.00000682, -0.00003528, 0.00001393, -0.00002677, 0.00000554, -0.00001552, 0.00000099, -0.00000795, -0.00001741, 0.00000740, -0.00000272, 0.00000983, 0.00000620, -0.00000720, 0.00000498, -0.00000525, 0.00001177, -0.00000194, 0.00000705, -0.00000602, 0.00000468, 0.00002316, -0.00000608, -0.00001259, 0.00001254, -0.00001999, 0.00001319, -0.00002351, 0.00001513, -0.00001551, 0.00001583, -0.00000958, -0.00000187, -0.00001000, 0.00000407, -0.00001001, 0.00001612, 0.00000317, -0.00000547, -0.00000488, },
		{ -0.00000275, -0.00000405, -0.00000287, 0.00000212, -0.00000132, 0.00000459, -0.00000547, -0.00000069, -0.00000325, 0.00001034, -0.00000238, -0.00000927, -0.00000517, 0.00001655, -0.00001770, 0.00000190, 0.00000863, -0.00000511, -0.00000021, -0.00000086, 0.00000536, -0.00001360, 0.00001334, 0.00000233, -0.00000554, -0.00000112, 0.00001154, -0.00000515, 0.00000202, -0.00000853, 0.00000100, 0.00000341, 0.00000194, -0.00000014, 0.00000012, -0.00000701, 0.00000558, -0.00000588, 0.00000892, 0.00000325, -0.00000495, 0.00000103, -0.00000798, 0.00000202, 0.00000936, 0.00000161, -0.00000188, 0.00000283, -0.00000983, 0.00001492, -0.00001613, 0.00002286, -0.00001756, 0.00000145, -0.00000366, -0.00001261, 0.00001902, -0.00001473, -0.00000095, 0.00000148, 0.00000419, -0.00001655, -0.00000056, -0.00000430, 0.00000153, 0.00000078, -0.00000166, -0.00000864, -0.00001033, 0.00001073, -0.00002310, 0.00001030, -0.00001208, 0.00000264, -0.00000485, -0.00001659, -0.00000173, -0.00002634, -0.00000614, -0.00005590, -0.00004493, -0.00032841, 0.00037066, 0.00031378, 0.00039753, -0.00028707, -0.00009679, -0.00003078, -0.00002598, -0.00001736, -0.00001217, -0.00000479, -0.00001200, -0.00000627, -0.00000437, -0.00000853, -0.00000039, -0.00002608, 0.00001899, -0.00001388, 0.00002190, -0.00000531, 0.00000255, -0.00000376, 0.00000497, 0.00000353, 0.00000526, 0.00000099, -0.00000118, 0.00000239, 0.00003070, -0.00002072, 0.00000366, -0.00000594, -0.00000338, -0.00000085, -0.00001236, 0.00000669, -0.00000814, 0.00001092, -0.00000808, -0.00000441, -0.00000636, 0.00000023, -0.00000501, 0.00001435, 0.00000206, -0.00000683, },
		{ -0.00000906, -0.00000016, -0.00000667, 0.00000079, -0.00000159, 0.00000297, 0.00000019, -0.00000266, -0.00000320, 0.00000065, 0.00000811, -0.00000413, -0.00000998, 0.00000071, 0.00001263, -0.00002110, 0.00000898, 0.00000365, -0.00000227, -0.00000340, 0.00000396, -0.00000086, -0.00000908, 0.00001475, -0.00000101, -0.00000616, 0.00000189, 0.00001102, -0.00000624, 0.00000167, -0.00000885, 0.00000280, 0.00000265, 0.00000178, 0.00000059, -0.00000238, -0.00000449, 0.00000442, -0.00000528, 0.00001147, -0.00000045, -0.00000378, 0.00000081, -0.00000902, 0.00000439, 0.00001012, -0.00000145, 0.00000100, -0.00000123, -0.00000439, 0.00000994, -0.00001267, 0.00002379, -0.00002596, 0.00001327, -0.00001757, 0.00000146, 0.00000936, -0.00000929, -0.00000563, 0.00000810, -0.00000505, -0.00000871, -0.00000677, 0.00000230, -0.00000415, 0.00000721, -0.00001089, -0.00000015, -0.00001818, 0.00002032, -0.00003676, 0.00002703, -0.00003044, 0.00002413, -0.00003104, 0.00001197, -0.00003462, 0.00001024, -0.00005327, -0.00000003, -0.00015582, -0.00019821, 0.00041754, 0.00025557, 0.00036035, -0.00033554, -0.00003330, -0.00006240, 0.00000185, -0.00003846, 0.00000839, -0.00002215, 0.00000186, -0.00001675, 0.00000504, -0.00001624, 0.00000470, -0.00003148, 0.00002743, -0.00002259, 0.00003217, -0.00001540, 0.00001077, -0.00001027, 0.00001299, -0.00000304, 0.00001086, -0.00000310, 0.00000082, 0.00000529, 0.00003070, -0.00002760, 0.00001263, -0.00001916, 0.00001115, -0.00001515, 0.00000039, -0.00000397, 0.00000206, 0.00000249, -0.00000354, -0.00000878, -0.00000152, -0.00000455, 0.00000156, 0.00001005, 0.00000287, },
		{ 0.00000420, -0.00001112, 0.00000198, -0.00000841, 0.00000327, -0.00000414, 0.00000639, -0.00000408, 0.00000055, -0.00000604, 0.00000533, 0.00000399, -0.00000447, -0.00001037, 0.00000699, 0.00000537, -0.00001998, 0.00001321, -0.00000013, -0.00000032, -0.00000557, 0.00000811, -0.00000779, -0.00000160, 0.00001252, -0.00000252, -0.00000735, 0.00000631, 0.00000846, -0.00000599, 0.00000029, -0.00000786, 0.00000358, 0.00000231, 0.00000162, 0.00000098, -0.00000481, -0.00000115, 0.00000174, -0.00000251, 0.00001182, -0.00000332, -0.00000241, 0.00000029, -0.00001015, 0.00000805, 0.00000877, -0.00000296, 0.00000296, -0.00000477, 0.00000168, 0.00000228, -0.00000384, 0.00001651, -0.00002512, 0.00001688, -0.00002526, 0.00001203, 0.00000134, -0.00000546, -0.00000783, 0.00001136, -0.00001097, -0.00000346, -0.00001061, 0.00000639, -0.00000670, 0.00000920, -0.00001516, 0.00000357, -0.00002076, 0.00002364, -0.00004367, 0.00003714, -0.00004170, 0.00003737, -0.00004838, 0.00003148, -0.00005819, 0.00003664, -0.00008925, 0.00004161, -0.00026326, -0.00001710, 0.00038200, 0.00024222, 0.00029663, -0.00035081, 0.00000832, -0.00008072, 0.00001828, -0.00004963, 0.00002011, -0.00003244, 0.00000997, -0.00002183, 0.00000915, -0.00001895, 0.00000448, -0.00003078, 0.00002964, -0.00002574, 0.00003784, -0.00002184, 0.00001578, -0.00001310, 0.00001729, -0.00000653, 0.00001388, -0.00000486, 0.00000030, 0.00001274, 0.00002313, -0.00002548, 0.00001146, -0.00002237, 0.00001782, -0.00002386, 0.00000932, -0.00001187, 0.00001055, -0.00000567, 0.00000122, -0.00001278, 0.00000248, -0.00000810, 0.00000735, 0.00000530, },
		{ 0.00000201, 0.00000463, -0.00001202, 0.00000289, -0.00000862, 0.00000382, -0.00000452, 0.00000760, -0.00000668, 0.00000268, -0.00000776, 0.00000926, -0.00000082, -0.00000393, -0.00000983, 0.00001217, -0.00000360, -0.00001485, 0.00001388, -0.00000163, -0.00000042, -0.00000527, 0.00000952, -0.00001281, 0.00000632, 0.00000810, -0.00000278, -0.00000841, 0.00001121, 0.00000446, -0.00000469, -0.00000176, -0.00000588, 0.00000344, 0.00000238, 0.00000153, 0.00000067, -0.00000632, 0.00000191, -0.00000134, 0.00000166, 0.00001007, -0.00000500, -0.00000107, -0.00000069, -0.00001080, 0.00001241, 0.00000531, -0.00000217, 0.00000265, -0.00000576, 0.00000546, -0.00000429, 0.00000625, 0.00000447, -0.00001684, 0.00001216, -0.00002509, 0.00001675, -0.00000294, -0.00000477, -0.00000636, 0.00001004, -0.00001195, -0.00000267, -0.00001025, 0.00000642, -0.00000488, 0.00000550, -0.00001272, 0.00000031, -0.00001565, 0.00001789, -0.00004043, 0.00003685, -0.00004188, 0.00003784, -0.00005126, 0.00003490, -0.00006377, 0.00004191, -0.00009959, 0.00004648, -0.00032894, 0.00016927, 0.00029090, 0.00026958, 0.00020543, -0.00033185, 0.00002504, -0.00008271, 0.00002101, -0.00004887, 0.00002098, -0.00003363, 0.00001068, -0.00002033, 0.00000707, -0.00001608, -0.00000132, -0.00002370, 0.00002500, -0.00002231, 0.00003787, -0.00002389, 0.00001739, -0.00001241, 0.00001799, -0.00000708, 0.00001481, -0.00000539, -0.00000061, 0.00002168, 0.00001119, -0.00001715, 0.00000189, -0.00001542, 0.00001472, -0.00002415, 0.00001124, -0.00001372, 0.00001392, -0.00001036, 0.00000368, -0.00001456, 0.00000408, -0.00000879, 0.00001040, },
		{ 0.00000997, 0.00000120, 0.00000338, -0.00001132, 0.00000240, -0.00000735, 0.00000251, -0.00000263, 0.00000615, -0.00000679, 0.00000279, -0.00000734, 0.00001110, -0.00000499, -0.00000320, -0.00000783, 0.00001473, -0.00001200, -0.00000741, 0.00001163, -0.00000095, -0.00000262, -0.00000216, 0.00000722, -0.00001395, 0.00001223, 0.00000338, -0.00000284, -0.00000836, 0.00001533, 0.00000006, -0.00000304, -0.00000384, -0.00000354, 0.00000280, 0.00000271, 0.00000153, -0.00000049, -0.00000633, 0.00000365, -0.00000353, 0.00000595, 0.00000695, -0.00000555, 0.00000016, -0.00000236, -0.00001022, 0.00001636, 0.00000063, 0.00000061, -0.00000011, -0.00000333, 0.00000497, -0.00000644, 0.00001296, -0.00000712, -0.00000570, 0.00000231, -0.00001866, 0.00001607, -0.00000368, -0.00000666, -0.00000216, 0.00000520, -0.00000883, -0.00000581, -0.00000620, 0.00000312, 0.00000005, -0.00000227, -0.00000507, -0.00000860, -0.00000418, 0.00000429, -0.00002792, 0.00002686, -0.00003193, 0.00002660, -0.00004047, 0.00002250, -0.00005099, 0.00002482, -0.00008153, 0.00000584, -0.00032608, 0.00031582, 0.00018317, 0.00031774, 0.00009848, -0.00028736, 0.00002182, -0.00007187, 0.00001328, -0.00003913, 0.00001346, -0.00002766, 0.00000573, -0.00001411, 0.00000070, -0.00000960, -0.00001050, -0.00001251, 0.00001535, -0.00001355, 0.00003305, -0.00002229, 0.00001665, -0.00000959, 0.00001647, -0.00000592, 0.00001495, -0.00000643, 0.00000084, 0.00002799, -0.00000012, -0.00000804, -0.00001071, -0.00000249, 0.00000434, -0.00001719, 0.00000651, -0.00000925, 0.00001130, -0.00001033, 0.00000277, -0.00001347, 0.00000298, -0.00000629, },
		{ -0.00000171, 0.00000675, 0.00000264, 0.00000061, -0.00000931, 0.00000098, -0.00000531, 0.00000026, 0.00000065, 0.00000266, -0.00000465, 0.00000088, -0.00000459, 0.00001018, -0.00000759, -0.00000271, -0.00000429, 0.00001374, -0.00001772, 0.00000008, 0.00000804, 0.00000079, -0.00000581, 0.00000274, 0.00000172, -0.00001069, 0.00001457, -0.00000007, -0.00000359, -0.00000643, 0.00001756, -0.00000370, -0.00000174, -0.00000530, -0.00000147, 0.00000212, 0.00000309, 0.00000154, -0.00000231, -0.00000478, 0.00000355, -0.00000386, 0.00000910, 0.00000346, -0.00000533, 0.00000131, -0.00000484, -0.00000777, 0.00001857, -0.00000375, 0.00000404, -0.00000418, 0.00000169, 0.00000021, -0.00000281, 0.00001316, -0.00001359, 0.00000303, -0.00000781, -0.00000969, 0.00001247, -0.00000278, -0.00000913, 0.00000237, -0.00000048, -0.00000424, -0.00001045, -0.00000078, -0.00000104, 0.00000506, -0.00001042, 0.00000380, -0.00001905, 0.00000926, -0.00001253, -0.00001073, 0.00001176, -0.00001694, 0.00000957, -0.00002260, 0.00000142, -0.00002771, -0.00000572, -0.00004569, -0.00007058, -0.00024928, 0.00039252, 0.00009522, 0.00036166, -0.00000659, -0.00023124, 0.00000814, -0.00005545, 0.00000149, -0.00002620, 0.00000266, -0.00001880, -0.00000110, -0.00000674, -0.00000653, -0.00000287, -0.00001957, -0.00000082, 0.00000403, -0.00000209, 0.00002532, -0.00001852, 0.00001506, -0.00000636, 0.00001437, -0.00000439, 0.00001540, -0.00000908, 0.00000639, 0.00002856, -0.00000649, -0.00000363, -0.00001984, 0.00001004, -0.00000797, -0.00000719, -0.00000157, -0.00000107, 0.00000443, -0.00000653, -0.00000084, -0.00001030, 0.00000021, },
		{ -0.00000244, 0.00000298, 0.00000246, 0.00000519, -0.00000283, -0.00000678, -0.00000056, -0.00000355, -0.00000165, 0.00000383, -0.00000146, -0.00000138, -0.00000213, -0.00000027, 0.00000673, -0.00000828, -0.00000248, 0.00000021, 0.00000930, -0.00001947, 0.00000564, 0.00000488, 0.00000202, -0.00000832, 0.00000751, -0.00000514, -0.00000409, 0.00001321, -0.00000164, -0.00000527, -0.00000251, 0.00001738, -0.00000617, -0.00000121, -0.00000578, -0.00000010, 0.00000172, 0.00000342, 0.00000134, -0.00000424, -0.00000219, 0.00000181, -0.00000204, 0.00001025, 0.00000052, -0.00000478, 0.00000234, -0.00000796, -0.00000325, 0.00001805, -0.00000625, 0.00000635, -0.00000751, 0.00000701, -0.00000669, 0.00000527, 0.00000659, -0.00001272, 0.00000566, -0.00001391, -0.00000213, 0.00000896, -0.00000257, -0.00000994, 0.00000474, -0.00000412, -0.00000120, -0.00001369, 0.00000330, -0.00000334, 0.00000700, -0.00001506, 0.00000947, -0.00002629, 0.00001942, -0.00002680, 0.00000521, -0.00000241, -0.00000356, -0.00000552, -0.00000660, -0.00001806, -0.00000595, -0.00003509, -0.00001160, -0.00015749, -0.00011630, 0.00039248, 0.00004954, 0.00038159, -0.00009402, -0.00017667, -0.00000661, -0.00004051, -0.00000824, -0.00001566, -0.00000638, -0.00001135, -0.00000616, -0.00000158, -0.00001151, 0.00000119, -0.00002548, 0.00000805, -0.00000563, 0.00000925, 0.00001669, -0.00001389, 0.00001366, -0.00000383, 0.00001271, -0.00000310, 0.00001625, -0.00001289, 0.00001564, 0.00002285, -0.00000630, -0.00000680, -0.00002092, 0.00001654, -0.00001667, 0.00000084, -0.00000850, 0.00000682, -0.00000339, -0.00000138, -0.00000532, -0.00000673, },
		{ -0.00000435, -0.00000332, 0.00000584, -0.00000102, 0.00000743, -0.00000589, -0.00000456, -0.00000146, -0.00000288, -0.00000212, 0.00000551, -0.00000462, 0.00000155, -0.00000486, 0.00000423, 0.00000175, -0.00000736, -0.00000214, 0.00000452, 0.00000254, -0.00001725, 0.00000824, 0.00000337, 0.00000155, -0.00000865, 0.00001012, -0.00001089, 0.00000368, 0.00000925, -0.00000168, -0.00000737, 0.00000276, 0.00001501, -0.00000724, -0.00000145, -0.00000535, 0.00000052, 0.00000172, 0.00000368, 0.00000063, -0.00000556, 0.00000049, -0.00000069, 0.00000143, 0.00000926, -0.00000135, -0.00000417, 0.00000306, -0.00001116, 0.00000287, 0.00001461, -0.00000594, 0.00000609, -0.00000808, 0.00000986, -0.00001229, 0.00001427, -0.00000400, -0.00000563, 0.00000152, -0.00001397, 0.00000156, 0.00000754, -0.00000445, -0.00000793, 0.00000364, -0.00000408, -0.00000159, -0.00001366, 0.00000433, -0.00000235, 0.00000429, -0.00001399, 0.00000917, -0.00002714, 0.00002260, -0.00003422, 0.00001531, -0.00001108, 0.00000332, -0.00001292, 0.00000059, -0.00002759, 0.00000402, -0.00005005, 0.00000098, -0.00022341, 0.00003873, 0.00033107, 0.00005013, 0.00036921, -0.00015538, -0.00013176, -0.00001683, -0.00003095, -0.00001268, -0.00001049, -0.00001086, -0.00000764, -0.00000759, -0.00000022, -0.00001284, 0.00000133, -0.00002690, 0.00001234, -0.00001156, 0.00001855, 0.00000850, -0.00000897, 0.00001253, -0.00000202, 0.00001143, -0.00000160, 0.00001640, -0.00001588, 0.00002598, 0.00001324, -0.00000141, -0.00001652, -0.00001320, 0.00001457, -0.00001849, 0.00000344, -0.00001080, 0.00001091, -0.00000892, 0.00000255, -0.00000874, },
		{ -0.00001005, -0.00000386, -0.00000176, 0.00000585, -0.00000245, 0.00000832, -0.00000786, -0.00000312, -0.00000152, -0.00000350, -0.00000078, 0.00000496, -0.00000577, 0.00000296, -0.00000607, 0.00000744, -0.00000344, -0.00000547, -0.00000123, 0.00000730, -0.00000466, -0.00001223, 0.00000806, 0.00000373, -0.00000090, -0.00000620, 0.00000932, -0.00001345, 0.00001019, 0.00000449, -0.00000117, -0.00000895, 0.00000824, 0.00001122, -0.00000730, -0.00000209, -0.00000437, 0.00000062, 0.00000202, 0.00000388, -0.00000076, -0.00000565, 0.00000224, -0.00000273, 0.00000542, 0.00000668, -0.00000212, -0.00000354, 0.00000312, -0.00001357, 0.00000943, 0.00000902, -0.00000296, 0.00000292, -0.00000490, 0.00000821, -0.00001341, 0.00001995, -0.00001405, 0.00000388, -0.00000698, -0.00000886, 0.00000122, 0.00000841, -0.00000820, -0.00000362, -0.00000037, -0.00000068, -0.00000538, -0.00001030, 0.00000245, 0.00000139, -0.00000228, -0.00000772, 0.00000301, -0.00002147, 0.00001829, -0.00003371, 0.00001824, -0.00001316, 0.00000280, -0.00001144, -0.00000282, -0.00002460, -0.00000142, -0.00004522, -0.00001894, -0.00024256, 0.00017746, 0.00023777, 0.00008540, 0.00032803, -0.00019076, -0.00009823, -0.00002176, -0.00002658, -0.00001231, -0.00001027, -0.00001099, -0.00000740, -0.00000586, -0.00000210, -0.00001116, -0.00000165, -0.00002453, 0.00001228, -0.00001334, 0.00002512, 0.00000121, -0.00000361, 0.00001100, -0.00000017, 0.00000978, 0.00000107, 0.00001432, -0.00001547, 0.00003368, 0.00000387, 0.00000380, -0.00002836, 0.00000002, 0.00000569, -0.00001371, 0.00000015, -0.00000752, 0.00000974, -0.00001041, 0.00000376, },
		{ 0.00000229, -0.00000944, -0.00000478, 0.00000164, 0.00000324, -0.00000160, 0.00000762, -0.00000863, -0.00000236, -0.00000109, -0.00000485, 0.00000185, 0.00000243, -0.00000475, 0.00000247, -0.00000519, 0.00000832, -0.00000761, -0.00000322, 0.00000049, 0.00000760, -0.00001037, -0.00000631, 0.00000624, 0.00000516, -0.00000454, -0.00000163, 0.00000518, -0.00001189, 0.00001367, 0.00000056, -0.00000111, -0.00000905, 0.00001269, 0.00000703, -0.00000692, -0.00000265, -0.00000335, 0.00000057, 0.00000245, 0.00000400, -0.00000273, -0.00000429, 0.00000237, -0.00000322, 0.00000860, 0.00000350, -0.00000216, -0.00000277, 0.00000217, -0.00001428, 0.00001493, 0.00000284, 0.00000143, -0.00000214, 0.00000135, 0.00000189, -0.00000853, 0.00001921, -0.00001920, 0.00001109, -0.00001570, -0.00000154, -0.00000134, 0.00001021, -0.00001226, 0.00000113, -0.00000524, 0.00000410, -0.00001081, -0.00000525, -0.00000061, 0.00000563, -0.00000993, 0.00000087, -0.00000625, -0.00001199, 0.00000907, -0.00002749, 0.00001602, -0.00001107, -0.00000202, -0.00000471, -0.00001294, -0.00001328, -0.00001768, -0.00002588, -0.00006813, -0.00020442, 0.00027005, 0.00014435, 0.00013585, 0.00026916, -0.00020637, -0.00007314, -0.00002408, -0.00002436, -0.00001017, -0.00001221, -0.00000909, -0.00000856, -0.00000306, -0.00000520, -0.00000844, -0.00000570, -0.00002046, 0.00000955, -0.00001189, 0.00002920, -0.00000522, 0.00000253, 0.00000829, 0.00000260, 0.00000707, 0.00000557, 0.00000913, -0.00000993, 0.00003569, -0.00000136, 0.00000461, -0.00003658, 0.00001308, -0.00000558, -0.00000565, -0.00000655, -0.00000045, 0.00000428, -0.00000813, },
		{ -0.00000394, -0.00000057, -0.00000807, -0.00000579, 0.00000536, -0.00000073, 0.00000079, 0.00000577, -0.00000862, -0.00000177, -0.00000091, -0.00000593, 0.00000460, -0.00000101, -0.00000231, 0.00000059, -0.00000252, 0.00000667, -0.00001005, -0.00000098, 0.00000277, 0.00000519, -0.00001319, -0.00000131, 0.00000430, 0.00000636, -0.00000792, 0.00000347, -0.00000091, -0.00000678, 0.00001358, -0.00000161, -0.00000204, -0.00000720, 0.00001525, 0.00000330, -0.00000658, -0.00000277, -0.00000265, 0.00000068, 0.00000287, 0.00000387, -0.00000486, -0.00000180, 0.00000084, -0.00000162, 0.00000994, 0.00000076, -0.00000198, -0.00000172, 0.00000008, -0.00001268, 0.00001803, -0.00000210, 0.00000533, -0.00000703, 0.00000849, -0.00000717, 0.00000132, 0.00001152, -0.00001713, 0.00001240, -0.00002059, 0.00000445, -0.00000358, 0.00001096, -0.00001468, 0.00000411, -0.00000848, 0.00000768, -0.00001540, -0.00000076, -0.00000262, 0.00000771, -0.00001528, 0.00000800, -0.00001470, -0.00000281, -0.00000081, -0.00001960, 0.00001250, -0.00000907, -0.00000594, 0.00000108, -0.00002256, -0.00000205, -0.00003452, -0.00000565, -0.00012975, -0.00011728, 0.00030367, 0.00007449, 0.00018283, 0.00020517, -0.00021014, -0.00005219, -0.00002726, -0.00002085, -0.00000966, -0.00001312, -0.00000790, -0.00000873, -0.00000144, -0.00000738, -0.00000669, -0.00000871, -0.00001694, 0.00000623, -0.00000854, 0.00003131, -0.00001075, 0.00000935, 0.00000422, 0.00000650, 0.00000334, 0.00001152, 0.00000143, 0.00000054, 0.00003114, -0.00000075, -0.00000168, -0.00003681, 0.00002061, -0.00001409, 0.00000118, -0.00001274, 0.00000699, -0.00000276, },
		{ -0.00000833, -0.00000015, -0.00000315, -0.00000730, -0.00000545, 0.00000767, -0.00000438, 0.00000353, 0.00000357, -0.00000840, -0.00000085, -0.00000166, -0.00000576, 0.00000619, -0.00000397, 0.00000032, -0.00000160, 0.00000090, 0.00000304, -0.00001062, 0.00000121, 0.00000490, 0.00000075, -0.00001278, 0.00000158, 0.00000345, 0.00000617, -0.00000956, 0.00000718, -0.00000677, 0.00000014, 0.00001063, -0.00000207, -0.00000377, -0.00000364, 0.00001572, 0.00000042, -0.00000646, -0.00000240, -0.00000239, 0.00000105, 0.00000325, 0.00000324, -0.00000651, 0.00000101, -0.00000169, 0.00000182, 0.00000905, -0.00000086, -0.00000198, -0.00000043, -0.00000290, -0.00000880, 0.00001801, -0.00000448, 0.00000700, -0.00000952, 0.00001362, -0.00001565, 0.00001290, -0.00000081, -0.00000853, 0.00000683, -0.00001948, 0.00000663, -0.00000354, 0.00000929, -0.00001421, 0.00000395, -0.00000842, 0.00000817, -0.00001730, 0.00000148, -0.00000213, 0.00000598, -0.00001605, 0.00001082, -0.00001922, 0.00000265, -0.00000763, -0.00001369, 0.00001104, -0.00001069, -0.00000467, 0.00000053, -0.00002507, 0.00000092, -0.00004142, -0.00000015, -0.00017972, -0.00000471, 0.00028411, 0.00003808, 0.00021481, 0.00014508, -0.00020794, -0.00003269, -0.00003303, -0.00001432, -0.00001256, -0.00001117, -0.00000898, -0.00000665, -0.00000218, -0.00000757, -0.00000686, -0.00000966, -0.00001532, 0.00000379, -0.00000437, 0.00003171, -0.00001489, 0.00001604, -0.00000050, 0.00001085, -0.00000042, 0.00001739, -0.00000661, 0.00001350, 0.00002185, 0.00000434, -0.00001431, -0.00002805, 0.00001988, -0.00001638, 0.00000327, -0.00001497, 0.00001142, },
		{ 0.00001105, -0.00001049, 0.00000159, -0.00000429, -0.00000788, -0.00000303, 0.00000749, -0.00000643, 0.00000560, 0.00000167, -0.00000835, 0.00000059, -0.00000351, -0.00000395, 0.00000586, -0.00000538, 0.00000198, -0.00000290, 0.00000374, -0.00000149, -0.00000964, 0.00000349, 0.00000594, -0.00000433, -0.00000981, 0.00000220, 0.00000410, 0.00000412, -0.00000867, 0.00000808, -0.00001032, 0.00000662, 0.00000635, -0.00000160, -0.00000553, 0.00000078, 0.00001449, -0.00000163, -0.00000644, -0.00000177, -0.00000240, 0.00000162, 0.00000363, 0.00000191, -0.00000704, 0.00000313, -0.00000410, 0.00000614, 0.00000628, -0.00000117, -0.00000228, 0.00000084, -0.00000611, -0.00000340, 0.00001517, -0.00000393, 0.00000567, -0.00000822, 0.00001437, -0.00002011, 0.00002201, -0.00001351, 0.00000324, -0.00000369, -0.00001279, 0.00000450, -0.00000077, 0.00000511, -0.00001098, 0.00000078, -0.00000500, 0.00000522, -0.00001604, 0.00000122, 0.00000074, 0.00000069, -0.00001205, 0.00000867, -0.00001888, 0.00000319, -0.00000985, -0.00001128, 0.00001269, -0.00001668, 0.00000285, -0.00000807, -0.00001794, -0.00000793, -0.00003316, -0.00001953, -0.00019611, 0.00010337, 0.00023080, 0.00003171, 0.00022921, 0.00009227, -0.00020185, -0.00001485, -0.00004049, -0.00000562, -0.00001826, -0.00000668, -0.00001203, -0.00000274, -0.00000487, -0.00000617, -0.00000847, -0.00000897, -0.00001558, 0.00000267, 0.00000023, 0.00003026, -0.00001679, 0.00002129, -0.00000452, 0.00001438, -0.00000269, 0.00002112, -0.00001204, 0.00002526, 0.00001153, 0.00001014, -0.00002932, -0.00001309, 0.00001182, -0.00001217, -0.00000029, -0.00001190, },
		{ -0.00000470, 0.00000632, -0.00000906, 0.00000097, -0.00000392, -0.00000958, 0.00000111, 0.00000476, -0.00000639, 0.00000654, 0.00000028, -0.00000840, 0.00000220, -0.00000599, -0.00000086, 0.00000366, -0.00000488, 0.00000212, -0.00000263, 0.00000493, -0.00000575, -0.00000761, 0.00000593, 0.00000518, -0.00000849, -0.00000572, 0.00000130, 0.00000577, 0.00000071, -0.00000551, 0.00000586, -0.00001038, 0.00001082, 0.00000236, -0.00000118, -0.00000640, 0.00000496, 0.00001233, -0.00000317, -0.00000621, -0.00000126, -0.00000237, 0.00000220, 0.00000399, -0.00000013, -0.00000608, 0.00000375, -0.00000519, 0.00001001, 0.00000258, -0.00000051, -0.00000272, 0.00000168, -0.00000866, 0.00000225, 0.00001066, -0.00000126, 0.00000188, -0.00000328, 0.00000999, -0.00001851, 0.00002520, -0.00002213, 0.00001358, -0.00001530, -0.00000310, -0.00000051, 0.00000369, -0.00000037, -0.00000642, -0.00000386, 0.00000035, 0.00000002, -0.00001266, -0.00000041, 0.00000436, -0.00000613, -0.00000525, 0.00000319, -0.00001519, 0.00000016, -0.00000855, -0.00001129, 0.00001577, -0.00002463, 0.00001380, -0.00002183, -0.00000409, -0.00002561, -0.00001286, -0.00006310, -0.00016769, 0.00018119, 0.00016792, 0.00004369, 0.00023021, 0.00004546, -0.00019088, -0.00000108, -0.00004673, 0.00000248, -0.00002429, -0.00000166, -0.00001528, 0.00000131, -0.00000794, -0.00000462, -0.00001002, -0.00000813, -0.00001663, 0.00000245, 0.00000532, 0.00002675, -0.00001571, 0.00002385, -0.00000655, 0.00001592, -0.00000228, 0.00002121, -0.00001248, 0.00003235, 0.00000399, 0.00001223, -0.00004127, 0.00000278, 0.00000041, -0.00000416, -0.00000765, },
		{ -0.00001524, 0.00000362, -0.00000055, -0.00000542, -0.00000111, -0.00000287, -0.00001134, 0.00000574, 0.00000043, -0.00000467, 0.00000650, -0.00000082, -0.00000816, 0.00000330, -0.00000820, 0.00000251, 0.00000041, -0.00000296, 0.00000098, -0.00000080, 0.00000397, -0.00000884, -0.00000496, 0.00000831, 0.00000241, -0.00001052, -0.00000204, 0.00000006, 0.00000744, -0.00000291, -0.00000131, 0.00000144, -0.00000706, 0.00001190, -0.00000027, -0.00000150, -0.00000574, 0.00000797, 0.00001001, -0.00000457, -0.00000553, -0.00000116, -0.00000203, 0.00000265, 0.00000425, -0.00000257, -0.00000378, 0.00000260, -0.00000421, 0.00001221, -0.00000092, 0.00000051, -0.00000299, 0.00000171, -0.00000974, 0.00000676, 0.00000613, 0.00000192, -0.00000268, 0.00000359, 0.00000176, -0.00001109, 0.00002119, -0.00002380, 0.00001858, -0.00002384, 0.00000608, -0.00000597, 0.00000799, -0.00000533, -0.00000246, -0.00000783, 0.00000543, -0.00000537, -0.00000897, -0.00000162, 0.00000653, -0.00001163, 0.00000132, -0.00000267, -0.00001106, -0.00000348, -0.00000651, -0.00001098, 0.00001698, -0.00003024, 0.00002292, -0.00003475, 0.00000964, -0.00004401, 0.00000897, -0.00011866, -0.00009826, 0.00021520, 0.00011549, 0.00006092, 0.00022440, 0.00000167, -0.00017317, 0.00000586, -0.00004867, 0.00000697, -0.00002784, 0.00000154, -0.00001672, 0.00000359, -0.00000968, -0.00000445, -0.00000997, -0.00000878, -0.00001706, 0.00000242, 0.00001099, 0.00002129, -0.00001151, 0.00002321, -0.00000600, 0.00001515, 0.00000091, 0.00001755, -0.00000730, 0.00003310, 0.00000143, 0.00000777, -0.00004567, 0.00001414, -0.00000938, 0.00000347, },
		{ 0.00000703, -0.00001959, 0.00000978, -0.00000671, -0.00000169, -0.00000317, -0.00000236, -0.00001184, 0.00000928, -0.00000404, -0.00000217, 0.00000601, -0.00000202, -0.00000717, 0.00000325, -0.00000917, 0.00000495, -0.00000265, -0.00000066, -0.00000057, 0.00000182, 0.00000113, -0.00001035, -0.00000194, 0.00001012, -0.00000186, -0.00000997, 0.00000012, -0.00000044, 0.00000806, -0.00000539, 0.00000227, -0.00000342, -0.00000164, 0.00001019, -0.00000131, -0.00000260, -0.00000350, 0.00000938, 0.00000801, -0.00000598, -0.00000436, -0.00000155, -0.00000127, 0.00000292, 0.00000418, -0.00000484, -0.00000076, 0.00000008, -0.00000111, 0.00001206, -0.00000330, 0.00000132, -0.00000289, 0.00000085, -0.00000902, 0.00000917, 0.00000302, 0.00000392, -0.00000589, 0.00000974, -0.00000746, -0.00000037, 0.00001143, -0.00001838, 0.00001653, -0.00002656, 0.00001187, -0.00000963, 0.00001048, -0.00000833, -0.00000061, -0.00000944, 0.00000835, -0.00000918, -0.00000650, -0.00000108, 0.00000564, -0.00001362, 0.00000504, -0.00000625, -0.00000927, -0.00000482, -0.00000653, -0.00000775, 0.00001337, -0.00002961, 0.00002524, -0.00004079, 0.00001595, -0.00005391, 0.00001904, -0.00016675, -0.00000506, 0.00020684, 0.00008372, 0.00007438, 0.00021668, -0.00004089, -0.00014824, 0.00000486, -0.00004474, 0.00000619, -0.00002724, 0.00000149, -0.00001518, 0.00000308, -0.00000920, -0.00000634, -0.00000763, -0.00001184, -0.00001592, 0.00000215, 0.00001697, 0.00001459, -0.00000496, 0.00001987, -0.00000330, 0.00001278, 0.00000577, 0.00001171, 0.00000202, 0.00002824, 0.00000337, -0.00000312, -0.00004091, 0.00001771, -0.00001375, },
		{ -0.00001168, 0.00000502, -0.00001886, 0.00001164, -0.00001003, 0.00000043, -0.00000403, -0.00000319, -0.00001026, 0.00001054, -0.00000736, 0.00000018, 0.00000553, -0.00000355, -0.00000530, 0.00000179, -0.00000841, 0.00000564, -0.00000443, 0.00000095, -0.00000149, 0.00000409, -0.00000277, -0.00001032, 0.00000132, 0.00001068, -0.00000656, -0.00000728, 0.00000044, 0.00000029, 0.00000710, -0.00000586, 0.00000382, -0.00000681, 0.00000397, 0.00000688, -0.00000129, -0.00000396, -0.00000028, 0.00000933, 0.00000643, -0.00000731, -0.00000294, -0.00000227, -0.00000020, 0.00000307, 0.00000354, -0.00000631, 0.00000202, -0.00000283, 0.00000339, 0.00000966, -0.00000414, 0.00000157, -0.00000244, -0.00000062, -0.00000675, 0.00000929, 0.00000201, 0.00000371, -0.00000616, 0.00001267, -0.00001433, 0.00000989, -0.00000055, -0.00000836, 0.00000846, -0.00002303, 0.00001304, -0.00001042, 0.00001052, -0.00000893, -0.00000122, -0.00000819, 0.00000842, -0.00001065, -0.00000575, 0.00000133, 0.00000161, -0.00001152, 0.00000492, -0.00000646, -0.00001096, -0.00000255, -0.00000981, -0.00000077, 0.00000415, -0.00002145, 0.00001863, -0.00003693, 0.00001083, -0.00004974, 0.00000747, -0.00018806, 0.00008770, 0.00016973, 0.00007212, 0.00008147, 0.00020817, -0.00008128, -0.00011812, -0.00000275, -0.00003577, 0.00000066, -0.00002270, -0.00000160, -0.00001102, 0.00000010, -0.00000687, -0.00000977, -0.00000355, -0.00001709, -0.00001313, 0.00000180, 0.00002257, 0.00000786, 0.00000251, 0.00001512, 0.00000039, 0.00001020, 0.00001038, 0.00000628, 0.00001240, 0.00002047, 0.00000705, -0.00001732, -0.00002896, 0.00001358, },
		{ 0.00000484, -0.00000504, -0.00000146, -0.00001352, 0.00000906, -0.00000991, 0.00000032, -0.00000335, -0.00000538, -0.00000665, 0.00000919, -0.00000885, 0.00000188, 0.00000521, -0.00000526, -0.00000289, -0.00000072, -0.00000610, 0.00000445, -0.00000448, 0.00000126, -0.00000104, 0.00000494, -0.00000668, -0.00000905, 0.00000471, 0.00000948, -0.00001040, -0.00000353, -0.00000062, 0.00000202, 0.00000477, -0.00000430, 0.00000280, -0.00000746, 0.00000799, 0.00000345, -0.00000109, -0.00000472, 0.00000299, 0.00000840, 0.00000507, -0.00000829, -0.00000162, -0.00000297, 0.00000097, 0.00000316, 0.00000224, -0.00000647, 0.00000366, -0.00000493, 0.00000805, 0.00000573, -0.00000357, 0.00000129, -0.00000187, -0.00000213, -0.00000375, 0.00000777, 0.00000279, 0.00000139, -0.00000317, 0.00001117, -0.00001646, 0.00001617, -0.00001059, 0.00000222, -0.00000245, -0.00001503, 0.00001020, -0.00000863, 0.00000859, -0.00000785, -0.00000344, -0.00000489, 0.00000622, -0.00001027, -0.00000604, 0.00000445, -0.00000407, -0.00000664, 0.00000193, -0.00000425, -0.00001516, 0.00000248, -0.00001540, 0.00000848, -0.00000868, -0.00000780, 0.00000489, -0.00002475, -0.00000432, -0.00003267, -0.00002714, -0.00017096, 0.00015744, 0.00012300, 0.00007276, 0.00008498, 0.00019664, -0.00011621, -0.00008684, -0.00001358, -0.00002447, -0.00000732, -0.00001606, -0.00000611, -0.00000577, -0.00000394, -0.00000404, -0.00001327, 0.00000074, -0.00002331, -0.00000936, 0.00000191, 0.00002690, 0.00000240, 0.00000933, 0.00001043, 0.00000376, 0.00000880, 0.00001301, 0.00000380, 0.00002055, 0.00001308, 0.00000880, -0.00003009, -0.00001432, },
		{ -0.00000210, -0.00000403, 0.00000256, -0.00000945, -0.00000594, 0.00000366, -0.00000726, -0.00000147, -0.00000170, -0.00000813, -0.00000195, 0.00000584, -0.00000860, 0.00000289, 0.00000488, -0.00000671, -0.00000060, -0.00000344, -0.00000308, 0.00000200, -0.00000309, 0.00000034, 0.00000089, 0.00000378, -0.00000968, -0.00000693, 0.00000795, 0.00000641, -0.00001234, -0.00000002, -0.00000214, 0.00000397, 0.00000194, -0.00000156, -0.00000028, -0.00000515, 0.00000942, 0.00000102, -0.00000144, -0.00000419, 0.00000537, 0.00000730, 0.00000361, -0.00000865, -0.00000076, -0.00000332, 0.00000200, 0.00000321, 0.00000040, -0.00000526, 0.00000368, -0.00000528, 0.00001158, 0.00000141, -0.00000214, 0.00000070, -0.00000148, -0.00000304, -0.00000107, 0.00000579, 0.00000421, -0.00000185, 0.00000194, 0.00000578, -0.00001334, 0.00001662, -0.00001554, 0.00000947, -0.00001234, -0.00000561, 0.00000525, -0.00000553, 0.00000591, -0.00000642, -0.00000573, -0.00000115, 0.00000321, -0.00000931, -0.00000602, 0.00000647, -0.00000906, -0.00000141, -0.00000165, -0.00000192, -0.00001939, 0.00000781, -0.00002075, 0.00001683, -0.00002115, 0.00000674, -0.00001102, -0.00000967, -0.00002325, -0.00001058, -0.00007632, -0.00011597, 0.00019052, 0.00008382, 0.00007563, 0.00008982, 0.00017862, -0.00014178, -0.00005892, -0.00002372, -0.00001420, -0.00001480, -0.00000976, -0.00000996, -0.00000137, -0.00000727, -0.00000227, -0.00001521, 0.00000345, -0.00002896, -0.00000559, 0.00000309, 0.00002936, -0.00000094, 0.00001442, 0.00000685, 0.00000609, 0.00000915, 0.00001299, 0.00000554, 0.00002427, 0.00000837, 0.00000596, -0.00003716, },
		{ -0.00003656, 0.00000418, -0.00000922, 0.00000755, -0.00001567, 0.00000089, -0.00000209, -0.00000386, -0.00000369, -0.00000015, -0.00001026, 0.00000246, 0.00000164, -0.00000723, 0.00000353, 0.00000418, -0.00000740, 0.00000083, -0.00000542, -0.00000046, -0.00000062, -0.00000114, -0.00000110, 0.00000371, 0.00000069, -0.00001119, -0.00000431, 0.00001058, 0.00000187, -0.00001192, 0.00000223, -0.00000317, 0.00000527, -0.00000026, 0.00000100, -0.00000405, -0.00000077, 0.00000830, -0.00000000, -0.00000255, -0.00000217, 0.00000633, 0.00000654, 0.00000184, -0.00000828, -0.00000054, -0.00000313, 0.00000275, 0.00000310, -0.00000158, -0.00000306, 0.00000217, -0.00000352, 0.00001302, -0.00000223, -0.00000050, 0.00000011, -0.00000149, -0.00000294, 0.00000044, 0.00000456, 0.00000491, -0.00000424, 0.00000703, -0.00000133, -0.00000654, 0.00001174, -0.00001450, 0.00001114, -0.00001813, 0.00000227, 0.00000030, -0.00000257, 0.00000369, -0.00000585, -0.00000669, 0.00000147, 0.00000084, -0.00000900, -0.00000457, 0.00000591, -0.00001128, 0.00000182, -0.00000351, -0.00000189, -0.00002098, 0.00001069, -0.00002311, 0.00002108, -0.00002915, 0.00001721, -0.00002340, 0.00000171, -0.00003787, 0.00000522, -0.00012435, -0.00003558, 0.00018564, 0.00006198, 0.00007341, 0.00009967, 0.00015184, -0.00015542, -0.00003763, -0.00003029, -0.00000740, -0.00001955, -0.00000564, -0.00001163, 0.00000081, -0.00000874, -0.00000253, -0.00001458, 0.00000333, -0.00003276, -0.00000255, 0.00000562, 0.00002986, -0.00000204, 0.00001748, 0.00000458, 0.00000751, 0.00001075, 0.00001109, 0.00001096, 0.00002327, 0.00000666, -0.00000181, },
		{ -0.00001235, -0.00002941, 0.00000385, -0.00000911, 0.00000798, -0.00001795, 0.00000470, -0.00000603, -0.00000141, -0.00000513, 0.00000043, -0.00001086, 0.00000539, -0.00000221, -0.00000547, 0.00000412, 0.00000290, -0.00000707, 0.00000100, -0.00000603, 0.00000086, -0.00000234, 0.00000034, -0.00000205, 0.00000637, -0.00000367, -0.00001108, -0.00000141, 0.00001208, -0.00000324, -0.00000938, 0.00000281, -0.00000320, 0.00000536, -0.00000100, 0.00000211, -0.00000687, 0.00000408, 0.00000555, 0.00000003, -0.00000404, 0.00000095, 0.00000583, 0.00000634, -0.00000023, -0.00000727, -0.00000087, -0.00000241, 0.00000320, 0.00000267, -0.00000315, -0.00000063, -0.00000016, -0.00000004, 0.00001213, -0.00000450, 0.00000083, -0.00000032, -0.00000184, -0.00000188, 0.00000048, 0.00000471, 0.00000392, -0.00000432, 0.00000985, -0.00000735, 0.00000095, 0.00000405, -0.00000894, 0.00000738, -0.00001866, 0.00000683, -0.00000319, -0.00000064, 0.00000251, -0.00000660, -0.00000576, 0.00000223, -0.00000015, -0.00000981, -0.00000148, 0.00000248, -0.00000995, 0.00000189, -0.00000251, -0.00000540, -0.00001839, 0.00000950, -0.00002098, 0.00001953, -0.00003038, 0.00002053, -0.00002845, 0.00000479, -0.00004218, 0.00000509, -0.00015407, 0.00005032, 0.00015259, 0.00005831, 0.00006398, 0.00011495, 0.00011675, -0.00015710, -0.00002392, -0.00003238, -0.00000479, -0.00002086, -0.00000424, -0.00001081, 0.00000050, -0.00000819, -0.00000476, -0.00001148, 0.00000013, -0.00003423, -0.00000048, 0.00000930, 0.00002885, -0.00000145, 0.00001895, 0.00000317, 0.00000882, 0.00001232, 0.00000918, 0.00001801, 0.00001899, 0.00000635, },
		{ 0.00000501, -0.00002199, -0.00001920, -0.00000101, -0.00000468, 0.00000416, -0.00001611, 0.00000482, -0.00000723, -0.00000077, -0.00000521, -0.00000025, -0.00000970, 0.00000627, -0.00000493, -0.00000382, 0.00000475, 0.00000116, -0.00000590, 0.00000003, -0.00000523, 0.00000059, -0.00000259, 0.00000066, -0.00000169, 0.00000775, -0.00000830, -0.00000957, 0.00000153, 0.00001204, -0.00000784, -0.00000556, 0.00000191, -0.00000223, 0.00000427, -0.00000009, 0.00000112, -0.00000746, 0.00000778, 0.00000249, 0.00000033, -0.00000513, 0.00000434, 0.00000435, 0.00000654, -0.00000239, -0.00000594, -0.00000152, -0.00000131, 0.00000339, 0.00000179, -0.00000382, 0.00000119, -0.00000226, 0.00000417, 0.00000934, -0.00000526, 0.00000164, -0.00000061, -0.00000221, -0.00000037, -0.00000054, 0.00000610, 0.00000124, -0.00000162, 0.00000907, -0.00000996, 0.00000607, -0.00000302, -0.00000202, 0.00000038, -0.00001479, 0.00000788, -0.00000487, 0.00000023, 0.00000209, -0.00000818, -0.00000343, 0.00000143, -0.00000004, -0.00001124, 0.00000234, -0.00000273, -0.00000588, -0.00000070, 0.00000084, -0.00001187, -0.00001198, 0.00000450, -0.00001486, 0.00001287, -0.00002530, 0.00001678, -0.00002597, -0.00000091, -0.00003531, -0.00001444, -0.00015332, 0.00012139, 0.00010726, 0.00006660, 0.00005007, 0.00013280, 0.00007645, -0.00014920, -0.00001660, -0.00003098, -0.00000541, -0.00001951, -0.00000479, -0.00000832, -0.00000147, -0.00000643, -0.00000797, -0.00000695, -0.00000545, -0.00003361, 0.00000090, 0.00001366, 0.00002701, 0.00000004, 0.00001957, 0.00000198, 0.00001088, 0.00001253, 0.00000929, 0.00002410, 0.00001367, },
		{ 0.00000904, 0.00000081, -0.00002730, -0.00001017, -0.00000680, 0.00000131, -0.00000177, -0.00001171, 0.00000221, -0.00000610, -0.00000178, -0.00000415, -0.00000180, -0.00000727, 0.00000530, -0.00000633, -0.00000241, 0.00000524, -0.00000063, -0.00000443, -0.00000147, -0.00000359, -0.00000090, -0.00000148, -0.00000027, 0.00000024, 0.00000708, -0.00001213, -0.00000708, 0.00000429, 0.00001032, -0.00001099, -0.00000151, 0.00000021, -0.00000077, 0.00000258, 0.00000190, -0.00000171, -0.00000541, 0.00000924, 0.00000024, 0.00000014, -0.00000505, 0.00000703, 0.00000260, 0.00000678, -0.00000432, -0.00000463, -0.00000215, -0.00000009, 0.00000336, 0.00000053, -0.00000340, 0.00000180, -0.00000316, 0.00000788, 0.00000560, -0.00000484, 0.00000200, -0.00000093, -0.00000217, 0.00000081, -0.00000172, 0.00000782, -0.00000220, 0.00000304, 0.00000487, -0.00000835, 0.00000692, -0.00000663, 0.00000294, -0.00000666, -0.00000876, 0.00000650, -0.00000528, 0.00000064, 0.00000162, -0.00000951, -0.00000087, 0.00000012, 0.00000022, -0.00001219, 0.00000533, -0.00000774, -0.00000103, -0.00000415, 0.00000464, -0.00001924, -0.00000372, -0.00000242, -0.00000697, 0.00000374, -0.00001679, 0.00000893, -0.00001917, -0.00001160, -0.00002212, -0.00004865, -0.00011945, 0.00016345, 0.00006542, 0.00007780, 0.00003696, 0.00014862, 0.00003543, -0.00013540, -0.00001326, -0.00002805, -0.00000738, -0.00001709, -0.00000587, -0.00000559, -0.00000380, -0.00000470, -0.00001080, -0.00000251, -0.00001214, -0.00003165, 0.00000208, 0.00001816, 0.00002488, 0.00000180, 0.00001983, 0.00000071, 0.00001400, 0.00001086, 0.00001267, 0.00002717, },
		{ 0.00002651, 0.00000558, -0.00000635, -0.00002671, -0.00000540, -0.00001024, 0.00000584, -0.00000711, -0.00000711, -0.00000125, -0.00000394, -0.00000355, -0.00000266, -0.00000338, -0.00000456, 0.00000324, -0.00000670, -0.00000110, 0.00000528, -0.00000195, -0.00000333, -0.00000268, -0.00000202, -0.00000271, 0.00000030, -0.00000191, 0.00000339, 0.00000416, -0.00001438, -0.00000411, 0.00000658, 0.00000717, -0.00001214, 0.00000179, -0.00000154, 0.00000051, 0.00000117, 0.00000393, -0.00000537, -0.00000132, 0.00000830, -0.00000066, -0.00000092, -0.00000339, 0.00000827, 0.00000126, 0.00000670, -0.00000575, -0.00000361, -0.00000251, 0.00000107, 0.00000308, -0.00000087, -0.00000209, 0.00000112, -0.00000231, 0.00001007, 0.00000195, -0.00000381, 0.00000214, -0.00000145, -0.00000140, 0.00000098, -0.00000208, 0.00000869, -0.00000485, 0.00000785, -0.00000110, -0.00000353, 0.00000353, -0.00000566, 0.00000373, -0.00001095, -0.00000309, 0.00000428, -0.00000535, 0.00000133, 0.00000028, -0.00000956, 0.00000071, -0.00000055, -0.00000036, -0.00001162, 0.00000609, -0.00001062, 0.00000247, -0.00000640, 0.00000666, -0.00002495, 0.00000378, -0.00000874, -0.00000007, -0.00000456, -0.00000865, 0.00000119, -0.00001287, -0.00002131, -0.00001101, -0.00008609, -0.00006010, 0.00017214, 0.00003766, 0.00008421, 0.00002950, 0.00015821, -0.00000216, -0.00011919, -0.00001164, -0.00002541, -0.00000900, -0.00001504, -0.00000623, -0.00000382, -0.00000540, -0.00000396, -0.00001218, 0.00000052, -0.00001868, -0.00002910, 0.00000355, 0.00002250, 0.00002265, 0.00000369, 0.00001975, -0.00000031, 0.00001764, 0.00000789, 0.00001914, },
		{ 0.00002725, 0.00002271, 0.00000244, -0.00001468, -0.00002111, -0.00000569, -0.00000972, 0.00000706, -0.00000989, -0.00000430, -0.00000375, -0.00000219, -0.00000505, -0.00000150, -0.00000424, -0.00000253, 0.00000097, -0.00000649, 0.00000026, 0.00000468, -0.00000251, -0.00000303, -0.00000301, -0.00000134, -0.00000390, 0.00000184, -0.00000336, 0.00000685, -0.00000057, -0.00001463, -0.00000114, 0.00000811, 0.00000317, -0.00001126, 0.00000376, -0.00000274, 0.00000110, 0.00000077, 0.00000488, -0.00000845, 0.00000354, 0.00000560, -0.00000036, -0.00000264, -0.00000033, 0.00000781, 0.00000076, 0.00000608, -0.00000659, -0.00000298, -0.00000248, 0.00000206, 0.00000246, -0.00000198, -0.00000044, -0.00000042, 0.00000016, 0.00001025, -0.00000080, -0.00000272, 0.00000226, -0.00000219, 0.00000012, -0.00000013, -0.00000102, 0.00000781, -0.00000532, 0.00001077, -0.00000643, 0.00000214, -0.00000220, -0.00000110, 0.00000015, -0.00001112, 0.00000044, 0.00000250, -0.00000573, 0.00000266, -0.00000219, -0.00000788, 0.00000068, 0.00000003, -0.00000223, -0.00000923, 0.00000411, -0.00001036, 0.00000330, -0.00000612, 0.00000542, -0.00002711, 0.00000850, -0.00001257, 0.00000385, -0.00000963, -0.00000371, -0.00000316, -0.00001096, -0.00002501, -0.00001007, -0.00011286, 0.00000976, 0.00015291, 0.00002723, 0.00008213, 0.00003010, 0.00015929, -0.00003371, -0.00010288, -0.00001047, -0.00002386, -0.00000948, -0.00001400, -0.00000538, -0.00000349, -0.00000585, -0.00000445, -0.00001180, 0.00000148, -0.00002426, -0.00002651, 0.00000554, 0.00002659, 0.00002017, 0.00000595, 0.00001890, -0.00000031, 0.00002060, 0.00000521, },
		{ 0.00000467, 0.00003487, 0.00001706, -0.00000178, -0.00002130, -0.00001342, -0.00000944, -0.00000581, 0.00000494, -0.00000969, -0.00000402, -0.00000433, -0.00000174, -0.00000564, -0.00000104, -0.00000409, -0.00000171, -0.00000084, -0.00000602, 0.00000164, 0.00000352, -0.00000235, -0.00000356, -0.00000229, -0.00000191, -0.00000387, 0.00000238, -0.00000375, 0.00000951, -0.00000621, -0.00001293, 0.00000143, 0.00000870, -0.00000093, -0.00000880, 0.00000431, -0.00000321, 0.00000088, 0.00000166, 0.00000405, -0.00000972, 0.00000768, 0.00000230, 0.00000045, -0.00000434, 0.00000347, 0.00000591, 0.00000116, 0.00000491, -0.00000687, -0.00000267, -0.00000209, 0.00000283, 0.00000144, -0.00000238, 0.00000084, -0.00000195, 0.00000353, 0.00000858, -0.00000230, -0.00000190, 0.00000242, -0.00000294, 0.00000200, -0.00000223, 0.00000137, 0.00000507, -0.00000303, 0.00001041, -0.00000893, 0.00000598, -0.00000745, 0.00000454, -0.00000614, -0.00000759, 0.00000132, 0.00000166, -0.00000649, 0.00000436, -0.00000535, -0.00000489, -0.00000075, 0.00000166, -0.00000502, -0.00000566, 0.00000007, -0.00000730, 0.00000149, -0.00000335, 0.00000085, -0.00002529, 0.00000980, -0.00001343, 0.00000443, -0.00001100, -0.00000267, -0.00000331, -0.00001447, -0.00002121, -0.00002316, -0.00011789, 0.00007321, 0.00011757, 0.00003084, 0.00007221, 0.00003820, 0.00015194, -0.00005852, -0.00008729, -0.00000970, -0.00002311, -0.00000909, -0.00001365, -0.00000372, -0.00000426, -0.00000547, -0.00000574, -0.00001011, 0.00000044, -0.00002861, -0.00002406, 0.00000819, 0.00003041, 0.00001726, 0.00000890, 0.00001686, 0.00000150, 0.00002166, },
		{ 0.00002022, 0.00000756, 0.00004010, 0.00001067, -0.00000791, -0.00002370, -0.00000713, -0.00001373, -0.00000067, 0.00000106, -0.00000755, -0.00000571, -0.00000315, -0.00000262, -0.00000530, -0.00000113, -0.00000321, -0.00000207, -0.00000196, -0.00000540, 0.00000283, 0.00000218, -0.00000186, -0.00000448, -0.00000093, -0.00000351, -0.00000256, 0.00000167, -0.00000259, 0.00001038, -0.00001162, -0.00000968, 0.00000331, 0.00000832, -0.00000446, -0.00000545, 0.00000369, -0.00000310, 0.00000017, 0.00000353, 0.00000146, -0.00000860, 0.00000995, -0.00000046, 0.00000098, -0.00000520, 0.00000703, 0.00000323, 0.00000222, 0.00000334, -0.00000677, -0.00000254, -0.00000147, 0.00000335, 0.00000007, -0.00000186, 0.00000116, -0.00000257, 0.00000672, 0.00000578, -0.00000269, -0.00000135, 0.00000250, -0.00000336, 0.00000363, -0.00000454, 0.00000432, 0.00000124, 0.00000148, 0.00000666, -0.00000762, 0.00000616, -0.00000965, 0.00000832, -0.00001240, -0.00000213, 0.00000023, 0.00000149, -0.00000719, 0.00000573, -0.00000827, -0.00000159, -0.00000269, 0.00000352, -0.00000777, -0.00000222, -0.00000444, -0.00000290, -0.00000174, 0.00000064, -0.00000571, -0.00002053, 0.00000844, -0.00001222, 0.00000289, -0.00001006, -0.00000401, -0.00000104, -0.00002115, -0.00001288, -0.00004782, -0.00009706, 0.00011694, 0.00007943, 0.00004164, 0.00005805, 0.00005111, 0.00013790, -0.00007720, -0.00007232, -0.00000993, -0.00002222, -0.00000869, -0.00001316, -0.00000215, -0.00000533, -0.00000498, -0.00000700, -0.00000802, -0.00000198, -0.00003188, -0.00002173, 0.00001155, 0.00003379, 0.00001395, 0.00001255, 0.00001362, 0.00000535, },
		{ 0.00001062, 0.00001673, 0.00001409, 0.00004187, 0.00000392, -0.00001545, -0.00002096, -0.00000474, -0.00001583, 0.00000323, -0.00000245, -0.00000531, -0.00000791, -0.00000121, -0.00000422, -0.00000446, -0.00000127, -0.00000222, -0.00000314, -0.00000247, -0.00000457, 0.00000355, 0.00000109, -0.00000157, -0.00000519, 0.00000038, -0.00000546, -0.00000045, 0.00000000, 0.00000005, 0.00000892, -0.00001571, -0.00000557, 0.00000436, 0.00000713, -0.00000690, -0.00000194, 0.00000236, -0.00000276, -0.00000048, 0.00000562, -0.00000220, -0.00000533, 0.00000993, -0.00000197, 0.00000068, -0.00000463, 0.00000946, 0.00000054, 0.00000352, 0.00000158, -0.00000646, -0.00000242, -0.00000072, 0.00000356, -0.00000142, -0.00000055, 0.00000034, -0.00000169, 0.00000872, 0.00000283, -0.00000245, -0.00000090, 0.00000231, -0.00000313, 0.00000435, -0.00000615, 0.00000671, -0.00000232, 0.00000666, 0.00000078, -0.00000311, 0.00000245, -0.00000764, 0.00000813, -0.00001603, 0.00000301, -0.00000155, 0.00000133, -0.00000723, 0.00000604, -0.00000996, 0.00000091, -0.00000410, 0.00000468, -0.00000946, -0.00000021, -0.00000769, 0.00000101, -0.00000467, 0.00000402, -0.00001225, -0.00001467, 0.00000598, -0.00001055, 0.00000116, -0.00000908, -0.00000507, 0.00000044, -0.00002683, -0.00000607, -0.00007595, -0.00005454, 0.00013496, 0.00004880, 0.00005265, 0.00004393, 0.00006553, 0.00011948, -0.00009075, -0.00005772, -0.00001171, -0.00002036, -0.00000906, -0.00001175, -0.00000148, -0.00000595, -0.00000501, -0.00000750, -0.00000644, -0.00000500, -0.00003444, -0.00001939, 0.00001575, 0.00003629, 0.00001073, 0.00001643, 0.00000978, },
		{ 0.00000642, 0.00001605, 0.00001257, 0.00002331, 0.00004000, -0.00000357, -0.00002253, -0.00001432, -0.00000672, -0.00001443, 0.00000432, -0.00000396, -0.00000456, -0.00000915, 0.00000032, -0.00000574, -0.00000363, -0.00000098, -0.00000181, -0.00000431, -0.00000262, -0.00000348, 0.00000361, 0.00000056, -0.00000185, -0.00000518, 0.00000097, -0.00000698, 0.00000169, -0.00000191, 0.00000358, 0.00000519, -0.00001770, -0.00000138, 0.00000461, 0.00000538, -0.00000803, 0.00000113, 0.00000076, -0.00000250, -0.00000053, 0.00000705, -0.00000580, -0.00000086, 0.00000797, -0.00000215, -0.00000048, -0.00000248, 0.00001019, -0.00000149, 0.00000467, -0.00000013, -0.00000605, -0.00000224, 0.00000008, 0.00000336, -0.00000266, 0.00000112, -0.00000129, 0.00000074, 0.00000890, 0.00000056, -0.00000212, -0.00000032, 0.00000173, -0.00000209, 0.00000379, -0.00000639, 0.00000760, -0.00000417, 0.00001059, -0.00000509, 0.00000267, -0.00000369, -0.00000209, 0.00000356, -0.00001556, 0.00000606, -0.00000283, 0.00000059, -0.00000627, 0.00000495, -0.00000991, 0.00000192, -0.00000432, 0.00000457, -0.00000962, -0.00000032, -0.00000857, 0.00000307, -0.00000597, 0.00000533, -0.00001700, -0.00000945, 0.00000382, -0.00000973, 0.00000074, -0.00000987, -0.00000371, -0.00000162, -0.00002777, -0.00000704, -0.00009683, -0.00000100, 0.00012922, 0.00003060, 0.00005929, 0.00003303, 0.00007886, 0.00009855, -0.00009983, -0.00004375, -0.00001493, -0.00001733, -0.00001043, -0.00000921, -0.00000203, -0.00000580, -0.00000575, -0.00000698, -0.00000592, -0.00000804, -0.00003660, -0.00001688, 0.00002092, 0.00003736, 0.00000832, 0.00001970, },
		{ 0.00002153, 0.00000465, 0.00002021, 0.00000957, 0.00003353, 0.00003490, -0.00001230, -0.00002669, -0.00000658, -0.00001153, -0.00001005, 0.00000257, -0.00000307, -0.00000595, -0.00000861, 0.00000064, -0.00000657, -0.00000313, -0.00000007, -0.00000237, -0.00000510, -0.00000259, -0.00000223, 0.00000310, 0.00000057, -0.00000275, -0.00000430, 0.00000048, -0.00000747, 0.00000315, -0.00000326, 0.00000702, -0.00000023, -0.00001722, 0.00000209, 0.00000428, 0.00000334, -0.00000786, 0.00000338, -0.00000078, -0.00000246, 0.00000028, 0.00000714, -0.00000826, 0.00000354, 0.00000496, -0.00000149, -0.00000204, 0.00000083, 0.00000918, -0.00000249, 0.00000538, -0.00000166, -0.00000558, -0.00000199, 0.00000090, 0.00000266, -0.00000325, 0.00000249, -0.00000304, 0.00000417, 0.00000725, -0.00000056, -0.00000207, 0.00000054, 0.00000081, -0.00000042, 0.00000203, -0.00000512, 0.00000661, -0.00000351, 0.00001175, -0.00000884, 0.00000729, -0.00000978, 0.00000485, -0.00000404, -0.00001109, 0.00000629, -0.00000298, -0.00000087, -0.00000443, 0.00000270, -0.00000829, 0.00000144, -0.00000333, 0.00000333, -0.00000855, -0.00000226, -0.00000702, 0.00000293, -0.00000531, 0.00000412, -0.00001918, -0.00000573, 0.00000256, -0.00001008, 0.00000189, -0.00001275, 0.00000051, -0.00000787, -0.00002279, -0.00001905, -0.00010128, 0.00005037, 0.00010757, 0.00002437, 0.00006018, 0.00002659, 0.00008974, 0.00007614, -0.00010449, -0.00003125, -0.00001875, -0.00001370, -0.00001229, -0.00000600, -0.00000351, -0.00000512, -0.00000689, -0.00000575, -0.00000646, -0.00001083, -0.00003858, -0.00001400, 0.00002702, 0.00003657, 0.00000743, },
		{ 0.00000834, 0.00002152, 0.00000511, 0.00002212, 0.00000930, 0.00004292, 0.00002697, -0.00002209, -0.00002611, -0.00000088, -0.00001651, -0.00000462, -0.00000070, -0.00000067, -0.00000899, -0.00000640, -0.00000037, -0.00000659, -0.00000290, 0.00000125, -0.00000387, -0.00000529, -0.00000241, -0.00000107, 0.00000233, 0.00000084, -0.00000395, -0.00000283, -0.00000098, -0.00000678, 0.00000356, -0.00000346, 0.00000938, -0.00000633, -0.00001443, 0.00000427, 0.00000369, 0.00000124, -0.00000657, 0.00000463, -0.00000213, -0.00000255, 0.00000185, 0.00000573, -0.00000886, 0.00000677, 0.00000196, -0.00000068, -0.00000327, 0.00000449, 0.00000684, -0.00000242, 0.00000551, -0.00000294, -0.00000504, -0.00000171, 0.00000169, 0.00000146, -0.00000293, 0.00000300, -0.00000403, 0.00000762, 0.00000436, -0.00000055, -0.00000239, 0.00000160, -0.00000024, 0.00000146, -0.00000042, -0.00000281, 0.00000416, -0.00000050, 0.00000964, -0.00000921, 0.00000886, -0.00001339, 0.00001041, -0.00001218, -0.00000413, 0.00000411, -0.00000208, -0.00000277, -0.00000226, -0.00000001, -0.00000582, 0.00000007, -0.00000168, 0.00000160, -0.00000716, -0.00000493, -0.00000397, 0.00000124, -0.00000334, 0.00000103, -0.00001910, -0.00000342, 0.00000183, -0.00001086, 0.00000358, -0.00001645, 0.00000608, -0.00001651, -0.00001415, -0.00004056, -0.00008521, 0.00008833, 0.00008003, 0.00002617, 0.00005650, 0.00002422, 0.00009787, 0.00005275, -0.00010454, -0.00002132, -0.00002197, -0.00001044, -0.00001377, -0.00000297, -0.00000523, -0.00000443, -0.00000781, -0.00000444, -0.00000771, -0.00001339, -0.00004048, -0.00001041, 0.00003369, 0.00003391, },
		{ 0.00002989, 0.00001077, 0.00001994, 0.00000769, 0.00002164, 0.00001260, 0.00004997, 0.00001634, -0.00003167, -0.00002058, 0.00000069, -0.00001919, -0.00000033, -0.00000363, 0.00000169, -0.00001245, -0.00000331, -0.00000230, -0.00000602, -0.00000264, 0.00000250, -0.00000597, -0.00000490, -0.00000207, -0.00000024, 0.00000168, 0.00000095, -0.00000497, -0.00000131, -0.00000291, -0.00000518, 0.00000298, -0.00000230, 0.00000989, -0.00001191, -0.00000999, 0.00000495, 0.00000314, -0.00000075, -0.00000441, 0.00000486, -0.00000322, -0.00000252, 0.00000371, 0.00000317, -0.00000755, 0.00000821, -0.00000025, -0.00000036, -0.00000347, 0.00000757, 0.00000387, -0.00000152, 0.00000511, -0.00000394, -0.00000442, -0.00000144, 0.00000238, -0.00000005, -0.00000172, 0.00000235, -0.00000358, 0.00001006, 0.00000112, 0.00000017, -0.00000290, 0.00000266, -0.00000111, 0.00000304, -0.00000282, -0.00000031, 0.00000128, 0.00000378, 0.00000502, -0.00000628, 0.00000675, -0.00001311, 0.00001239, -0.00001824, 0.00000319, 0.00000064, -0.00000069, -0.00000457, -0.00000038, -0.00000236, -0.00000338, -0.00000139, -0.00000012, 0.00000015, -0.00000647, -0.00000696, -0.00000079, -0.00000078, -0.00000130, -0.00000264, -0.00001781, -0.00000172, 0.00000076, -0.00001081, 0.00000414, -0.00001889, 0.00001037, -0.00002418, -0.00000665, -0.00006538, -0.00005102, 0.00010689, 0.00005523, 0.00003109, 0.00005054, 0.00002482, 0.00010336, 0.00002882, -0.00010004, -0.00001470, -0.00002361, -0.00000841, -0.00001406, -0.00000095, -0.00000650, -0.00000425, -0.00000798, -0.00000375, -0.00000912, -0.00001592, -0.00004232, -0.00000570, 0.00004022, },
		{ 0.00004571, 0.00002541, 0.00001395, 0.00001760, 0.00001156, 0.00001950, 0.00001943, 0.00005355, 0.00000308, -0.00003888, -0.00001174, -0.00000211, -0.00001840, 0.00000138, -0.00000476, 0.00000259, -0.00001500, -0.00000041, -0.00000446, -0.00000520, -0.00000204, 0.00000317, -0.00000821, -0.00000404, -0.00000160, 0.00000012, 0.00000135, 0.00000064, -0.00000543, -0.00000027, -0.00000471, -0.00000323, 0.00000184, -0.00000005, 0.00000818, -0.00001586, -0.00000488, 0.00000429, 0.00000279, -0.00000246, -0.00000173, 0.00000416, -0.00000403, -0.00000214, 0.00000528, 0.00000020, -0.00000486, 0.00000789, -0.00000137, -0.00000076, -0.00000229, 0.00000934, 0.00000104, -0.00000021, 0.00000427, -0.00000465, -0.00000375, -0.00000118, 0.00000285, -0.00000158, 0.00000006, 0.00000069, -0.00000147, 0.00001072, -0.00000151, 0.00000099, -0.00000325, 0.00000345, -0.00000153, 0.00000386, -0.00000452, 0.00000149, -0.00000080, 0.00000773, -0.00000049, -0.00000139, 0.00000182, -0.00000910, 0.00000995, -0.00002040, 0.00000887, -0.00000282, 0.00000056, -0.00000585, 0.00000075, -0.00000381, -0.00000162, -0.00000233, 0.00000088, -0.00000056, -0.00000709, -0.00000737, 0.00000134, -0.00000210, -0.00000027, -0.00000562, -0.00001643, 0.00000013, -0.00000134, -0.00000905, 0.00000231, -0.00001836, 0.00001112, -0.00002776, -0.00000541, -0.00008492, -0.00000660, 0.00010630, 0.00003810, 0.00003535, 0.00004429, 0.00002738, 0.00010613, 0.00000519, -0.00009167, -0.00001142, -0.00002336, -0.00000791, -0.00001284, -0.00000035, -0.00000696, -0.00000478, -0.00000715, -0.00000412, -0.00001024, -0.00001867, -0.00004406, 0.00000055, },
		{ 0.00000853, 0.00004936, 0.00002141, 0.00001700, 0.00001551, 0.00001555, 0.00001703, 0.00002899, 0.00005280, -0.00001234, -0.00004151, -0.00000251, -0.00000772, -0.00001460, 0.00000038, -0.00000366, 0.00000137, -0.00001582, 0.00000151, -0.00000631, -0.00000438, -0.00000100, 0.00000295, -0.00001019, -0.00000278, -0.00000118, 0.00000015, 0.00000134, -0.00000010, -0.00000523, 0.00000001, -0.00000589, -0.00000148, 0.00000073, 0.00000266, 0.00000444, -0.00001739, -0.00000025, 0.00000284, 0.00000259, -0.00000371, 0.00000109, 0.00000269, -0.00000451, -0.00000130, 0.00000607, -0.00000234, -0.00000167, 0.00000634, -0.00000161, -0.00000166, 0.00000013, 0.00000946, -0.00000107, 0.00000112, 0.00000318, -0.00000506, -0.00000310, -0.00000089, 0.00000297, -0.00000273, 0.00000185, -0.00000141, 0.00000196, 0.00000937, -0.00000289, 0.00000140, -0.00000314, 0.00000376, -0.00000137, 0.00000375, -0.00000518, 0.00000209, -0.00000120, 0.00000989, -0.00000499, 0.00000346, -0.00000400, -0.00000295, 0.00000387, -0.00001823, 0.00001175, -0.00000533, 0.00000123, -0.00000643, 0.00000109, -0.00000427, -0.00000071, -0.00000263, 0.00000128, -0.00000069, -0.00000889, -0.00000594, 0.00000194, -0.00000229, -0.00000070, -0.00000725, -0.00001557, 0.00000240, -0.00000449, -0.00000557, -0.00000203, -0.00001451, 0.00000766, -0.00002610, -0.00001337, -0.00009151, 0.00003764, 0.00009174, 0.00002950, 0.00003723, 0.00003881, 0.00003140, 0.00010575, -0.00001686, -0.00008079, -0.00001073, -0.00002161, -0.00000857, -0.00001040, -0.00000106, -0.00000666, -0.00000584, -0.00000552, -0.00000561, -0.00001079, -0.00002189, -0.00004546, },
		{ -0.00004612, 0.00001808, 0.00005072, 0.00001859, 0.00001923, 0.00001451, 0.00001862, 0.00001569, 0.00003998, 0.00004696, -0.00002850, -0.00003838, 0.00000420, -0.00001366, -0.00000941, -0.00000231, -0.00000095, -0.00000171, -0.00001478, 0.00000211, -0.00000752, -0.00000360, 0.00000035, 0.00000179, -0.00001165, -0.00000119, -0.00000104, 0.00000012, 0.00000136, -0.00000101, -0.00000458, -0.00000042, -0.00000629, -0.00000027, 0.00000019, 0.00000500, -0.00000065, -0.00001628, 0.00000295, 0.00000132, 0.00000231, -0.00000421, 0.00000358, 0.00000073, -0.00000459, -0.00000008, 0.00000586, -0.00000383, 0.00000110, 0.00000437, -0.00000148, -0.00000249, 0.00000320, 0.00000807, -0.00000216, 0.00000215, 0.00000198, -0.00000516, -0.00000256, -0.00000054, 0.00000266, -0.00000320, 0.00000308, -0.00000315, 0.00000588, 0.00000636, -0.00000285, 0.00000113, -0.00000243, 0.00000352, -0.00000070, 0.00000285, -0.00000490, 0.00000155, 0.00000025, 0.00000951, -0.00000712, 0.00000647, -0.00000857, 0.00000299, -0.00000388, -0.00001265, 0.00001167, -0.00000650, 0.00000122, -0.00000645, 0.00000091, -0.00000406, -0.00000037, -0.00000256, 0.00000150, -0.00000087, -0.00001107, -0.00000326, 0.00000132, -0.00000168, -0.00000225, -0.00000768, -0.00001515, 0.00000472, -0.00000795, -0.00000136, -0.00000776, -0.00000853, 0.00000128, -0.00002079, -0.00002966, -0.00008121, 0.00007233, 0.00007073, 0.00002735, 0.00003681, 0.00003423, 0.00003683, 0.00010167, -0.00003591, -0.00006903, -0.00001145, -0.00001923, -0.00000961, -0.00000749, -0.00000255, -0.00000597, -0.00000699, -0.00000360, -0.00000795, -0.00001075, -0.00002581, },
		{ -0.00003056, -0.00004548, 0.00002863, 0.00004984, 0.00001720, 0.00002043, 0.00001491, 0.00002016, 0.00001675, 0.00005082, 0.00003547, -0.00004299, -0.00003007, 0.00000663, -0.00001774, -0.00000473, -0.00000518, 0.00000218, -0.00000574, -0.00001239, 0.00000150, -0.00000805, -0.00000281, 0.00000173, -0.00000016, -0.00001244, 0.00000064, -0.00000136, 0.00000031, 0.00000113, -0.00000171, -0.00000390, -0.00000121, -0.00000604, 0.00000033, 0.00000050, 0.00000616, -0.00000600, -0.00001293, 0.00000425, 0.00000037, 0.00000166, -0.00000375, 0.00000531, -0.00000138, -0.00000425, 0.00000128, 0.00000477, -0.00000404, 0.00000283, 0.00000268, -0.00000154, -0.00000261, 0.00000611, 0.00000571, -0.00000228, 0.00000273, 0.00000081, -0.00000497, -0.00000220, -0.00000010, 0.00000195, -0.00000287, 0.00000335, -0.00000378, 0.00000927, 0.00000251, -0.00000167, 0.00000025, -0.00000120, 0.00000285, 0.00000027, 0.00000150, -0.00000411, 0.00000046, 0.00000294, 0.00000682, -0.00000651, 0.00000675, -0.00001041, 0.00000660, -0.00001088, -0.00000543, 0.00000931, -0.00000648, 0.00000069, -0.00000618, 0.00000063, -0.00000365, -0.00000018, -0.00000255, 0.00000201, -0.00000185, -0.00001258, -0.00000035, 0.00000025, -0.00000105, -0.00000404, -0.00000759, -0.00001463, 0.00000637, -0.00001064, 0.00000211, -0.00001309, -0.00000253, -0.00000541, -0.00001548, -0.00004965, -0.00005522, 0.00009182, 0.00005024, 0.00002841, 0.00003512, 0.00003027, 0.00004358, 0.00009359, -0.00005088, -0.00005782, -0.00001246, -0.00001709, -0.00001021, -0.00000492, -0.00000418, -0.00000535, -0.00000773, -0.00000200, -0.00001067, -0.00001027, },
		{ -0.00000962, -0.00003619, -0.00004295, 0.00003936, 0.00004717, 0.00001707, 0.00002081, 0.00001656, 0.00002017, 0.00002103, 0.00005960, 0.00001845, -0.00005297, -0.00001885, 0.00000479, -0.00001889, -0.00000181, -0.00000702, 0.00000459, -0.00000971, -0.00000936, 0.00000003, -0.00000794, -0.00000196, 0.00000288, -0.00000271, -0.00001237, 0.00000244, -0.00000211, 0.00000083, 0.00000048, -0.00000195, -0.00000354, -0.00000194, -0.00000547, 0.00000055, 0.00000163, 0.00000558, -0.00001039, -0.00000833, 0.00000374, 0.00000025, 0.00000052, -0.00000224, 0.00000592, -0.00000326, -0.00000352, 0.00000248, 0.00000321, -0.00000317, 0.00000336, 0.00000165, -0.00000209, -0.00000157, 0.00000804, 0.00000310, -0.00000172, 0.00000285, -0.00000022, -0.00000455, -0.00000203, 0.00000038, 0.00000100, -0.00000187, 0.00000259, -0.00000288, 0.00001122, -0.00000118, 0.00000006, -0.00000089, 0.00000025, 0.00000197, 0.00000124, 0.00000010, -0.00000336, -0.00000036, 0.00000576, 0.00000288, -0.00000385, 0.00000455, -0.00000922, 0.00000675, -0.00001518, 0.00000149, 0.00000577, -0.00000567, -0.00000012, -0.00000582, 0.00000049, -0.00000337, 0.00000016, -0.00000283, 0.00000307, -0.00000407, -0.00001258, 0.00000183, -0.00000049, -0.00000115, -0.00000517, -0.00000780, -0.00001342, 0.00000676, -0.00001173, 0.00000366, -0.00001647, 0.00000155, -0.00000994, -0.00001426, -0.00006660, -0.00001928, 0.00009534, 0.00003476, 0.00002988, 0.00003317, 0.00002677, 0.00005124, 0.00008180, -0.00006138, -0.00004800, -0.00001310, -0.00001566, -0.00000986, -0.00000328, -0.00000544, -0.00000509, -0.00000774, -0.00000122, -0.00001325, },
		{ -0.00001527, -0.00000921, -0.00004252, -0.00003811, 0.00004944, 0.00004332, 0.00001785, 0.00002079, 0.00001893, 0.00001914, 0.00002890, 0.00006408, -0.00000278, -0.00005623, -0.00000785, 0.00000024, -0.00001743, -0.00000089, -0.00000739, 0.00000565, -0.00001287, -0.00000635, -0.00000194, -0.00000723, -0.00000112, 0.00000365, -0.00000568, -0.00001129, 0.00000380, -0.00000298, 0.00000151, -0.00000046, -0.00000174, -0.00000365, -0.00000234, -0.00000494, 0.00000074, 0.00000322, 0.00000314, -0.00001279, -0.00000378, 0.00000214, 0.00000080, -0.00000091, 0.00000014, 0.00000530, -0.00000459, -0.00000251, 0.00000324, 0.00000169, -0.00000176, 0.00000295, 0.00000123, -0.00000306, 0.00000067, 0.00000849, 0.00000092, -0.00000089, 0.00000264, -0.00000105, -0.00000400, -0.00000203, 0.00000082, 0.00000007, -0.00000055, 0.00000114, -0.00000050, 0.00001120, -0.00000387, 0.00000165, -0.00000186, 0.00000155, 0.00000116, 0.00000200, -0.00000107, -0.00000302, -0.00000025, 0.00000757, -0.00000089, -0.00000050, 0.00000111, -0.00000596, 0.00000363, -0.00001590, 0.00000665, 0.00000211, -0.00000453, -0.00000106, -0.00000542, 0.00000050, -0.00000323, 0.00000063, -0.00000331, 0.00000444, -0.00000738, -0.00001083, 0.00000281, -0.00000056, -0.00000231, -0.00000516, -0.00000876, -0.00001128, 0.00000576, -0.00001107, 0.00000298, -0.00001735, 0.00000297, -0.00001125, -0.00001957, -0.00007421, 0.00001851, 0.00008618, 0.00002577, 0.00003023, 0.00003147, 0.00002394, 0.00005897, 0.00006721, -0.00006770, -0.00003974, -0.00001327, -0.00001494, -0.00000853, -0.00000276, -0.00000612, -0.00000525, -0.00000700, -0.00000154, },
		{ -0.00000293, -0.00001643, -0.00000945, -0.00004916, -0.00003083, 0.00005819, 0.00003888, 0.00001920, 0.00002081, 0.00002140, 0.00001796, 0.00004009, 0.00006195, -0.00002539, -0.00005221, 0.00000020, -0.00000486, -0.00001457, -0.00000135, -0.00000657, 0.00000539, -0.00001496, -0.00000378, -0.00000407, -0.00000594, -0.00000043, 0.00000403, -0.00000889, -0.00000907, 0.00000431, -0.00000354, 0.00000199, -0.00000138, -0.00000134, -0.00000409, -0.00000233, -0.00000464, 0.00000127, 0.00000464, -0.00000077, -0.00001265, -0.00000051, 0.00000044, 0.00000146, -0.00000216, 0.00000287, 0.00000365, -0.00000521, -0.00000137, 0.00000345, 0.00000064, -0.00000041, 0.00000215, 0.00000105, -0.00000399, 0.00000370, 0.00000739, -0.00000035, -0.00000018, 0.00000227, -0.00000167, -0.00000342, -0.00000213, 0.00000115, -0.00000061, 0.00000064, -0.00000041, 0.00000282, 0.00000926, -0.00000507, 0.00000262, -0.00000228, 0.00000238, 0.00000064, 0.00000241, -0.00000193, -0.00000311, 0.00000100, 0.00000765, -0.00000331, 0.00000213, -0.00000203, -0.00000230, -0.00000142, -0.00001330, 0.00000938, -0.00000093, -0.00000341, -0.00000204, -0.00000487, 0.00000046, -0.00000300, 0.00000094, -0.00000357, 0.00000553, -0.00001109, -0.00000774, 0.00000267, -0.00000013, -0.00000424, -0.00000410, -0.00001044, -0.00000840, 0.00000379, -0.00000921, 0.00000063, -0.00001629, 0.00000239, -0.00001034, -0.00003104, -0.00006887, 0.00005043, 0.00006987, 0.00002234, 0.00002921, 0.00003001, 0.00002220, 0.00006573, 0.00005105, -0.00007058, -0.00003277, -0.00001327, -0.00001450, -0.00000664, -0.00000317, -0.00000627, -0.00000568, -0.00000574, },
		{ -0.00000436, -0.00000516, -0.00001666, -0.00001070, -0.00005560, -0.00002123, 0.00006517, 0.00003429, 0.00002088, 0.00002116, 0.00002342, 0.00001778, 0.00005351, 0.00005137, -0.00004543, -0.00004246, 0.00000406, -0.00000880, -0.00001160, -0.00000231, -0.00000517, 0.00000418, -0.00001601, -0.00000185, -0.00000606, -0.00000422, 0.00000000, 0.00000394, -0.00001203, -0.00000586, 0.00000373, -0.00000346, 0.00000193, -0.00000191, -0.00000113, -0.00000456, -0.00000205, -0.00000461, 0.00000240, 0.00000517, -0.00000522, -0.00001017, 0.00000076, -0.00000044, 0.00000157, -0.00000261, 0.00000525, 0.00000142, -0.00000516, -0.00000024, 0.00000313, 0.00000030, 0.00000037, 0.00000150, 0.00000065, -0.00000428, 0.00000676, 0.00000515, -0.00000063, 0.00000014, 0.00000190, -0.00000208, -0.00000293, -0.00000226, 0.00000135, -0.00000087, 0.00000129, -0.00000137, 0.00000620, 0.00000596, -0.00000481, 0.00000280, -0.00000203, 0.00000257, 0.00000056, 0.00000244, -0.00000259, -0.00000333, 0.00000305, 0.00000599, -0.00000385, 0.00000320, -0.00000363, 0.00000004, -0.00000654, -0.00000857, 0.00000978, -0.00000304, -0.00000250, -0.00000302, -0.00000408, 0.00000014, -0.00000241, 0.00000075, -0.00000316, 0.00000563, -0.00001426, -0.00000412, 0.00000192, 0.00000028, -0.00000630, -0.00000256, -0.00001246, -0.00000526, 0.00000152, -0.00000704, -0.00000233, -0.00001451, 0.00000139, -0.00000973, -0.00004534, -0.00005093, 0.00007133, 0.00005201, 0.00002233, 0.00002731, 0.00002851, 0.00002195, 0.00007059, 0.00003454, -0.00007086, -0.00002672, -0.00001348, -0.00001380, -0.00000475, -0.00000411, -0.00000612, -0.00000608, },
		{ -0.00000621, -0.00000328, -0.00000781, -0.00001611, -0.00001315, -0.00006132, -0.00000960, 0.00007011, 0.00002991, 0.00002280, 0.00002200, 0.00002455, 0.00001999, 0.00006692, 0.00003191, -0.00005884, -0.00003023, 0.00000428, -0.00001103, -0.00000916, -0.00000323, -0.00000368, 0.00000241, -0.00001619, -0.00000073, -0.00000757, -0.00000236, 0.00000026, 0.00000325, -0.00001465, -0.00000208, 0.00000216, -0.00000264, 0.00000121, -0.00000185, -0.00000136, -0.00000477, -0.00000177, -0.00000470, 0.00000418, 0.00000415, -0.00000898, -0.00000626, 0.00000009, -0.00000010, 0.00000074, -0.00000179, 0.00000660, -0.00000078, -0.00000464, 0.00000076, 0.00000243, 0.00000064, 0.00000037, 0.00000135, -0.00000029, -0.00000341, 0.00000901, 0.00000246, -0.00000014, 0.00000005, 0.00000164, -0.00000231, -0.00000260, -0.00000236, 0.00000144, -0.00000072, 0.00000126, -0.00000125, 0.00000874, 0.00000217, -0.00000350, 0.00000237, -0.00000124, 0.00000215, 0.00000092, 0.00000214, -0.00000322, -0.00000323, 0.00000518, 0.00000322, -0.00000278, 0.00000265, -0.00000334, 0.00000002, -0.00001001, -0.00000325, 0.00000851, -0.00000421, -0.00000191, -0.00000391, -0.00000307, -0.00000050, -0.00000135, -0.00000008, -0.00000188, 0.00000430, -0.00001607, -0.00000078, 0.00000111, 0.00000015, -0.00000778, -0.00000118, -0.00001429, -0.00000231, -0.00000052, -0.00000527, -0.00000500, -0.00001303, 0.00000139, -0.00001219, -0.00005753, -0.00002425, 0.00007964, 0.00003671, 0.00002358, 0.00002520, 0.00002678, 0.00002333, 0.00007300, 0.00001859, -0.00006919, -0.00002137, -0.00001414, -0.00001245, -0.00000339, -0.00000513, -0.00000590, },
		},
		{ { 0.00012788, 0.00003688, -0.00003689, 0.00000354, 0.00001201, 0.00001452, 0.00008374, -0.00000109, -0.00007054, -0.00001123, -0.00002086, -0.00001074, -0.00000704, -0.00000743, -0.00000174, -0.00000433, -0.00000860, 0.00000329, -0.00001795, 0.00000116, -0.00000184, -0.00000213, 0.00000117, -0.00000072, 0.00000555, -0.00000173, -0.00000098, 0.00000028, -0.00000365, -0.00000110, 0.00000878, 0.00000297, -0.00000697, 0.00000145, -0.00000361, 0.00000160, -0.00000378, 0.00000351, -0.00000280, 0.00000759, -0.00001162, 0.00000459, -0.00000787, 0.00000673, -0.00000292, 0.00000344, -0.00000473, 0.00000437, -0.00000680, 0.00000359, -0.00000029, 0.00000444, -0.00000658, 0.00000275, -0.00000134, 0.00000354, -0.00000541, 0.00000414, -0.00000457, 0.00000544, -0.00000643, 0.00000461, -0.00000893, 0.00000777, -0.00000272, 0.00000777, -0.00000893, 0.00000461, -0.00000643, 0.00000544, -0.00000457, 0.00000414, -0.00000541, 0.00000354, -0.00000134, 0.00000275, -0.00000658, 0.00000444, -0.00000029, 0.00000359, -0.00000680, 0.00000437, -0.00000473, 0.00000344, -0.00000292, 0.00000673, -0.00000787, 0.00000459, -0.00001162, 0.00000759, -0.00000280, 0.00000351, -0.00000378, 0.00000160, -0.00000361, 0.00000145, -0.00000697, 0.00000297, 0.00000878, -0.00000110, -0.00000365, 0.00000028, -0.00000098, -0.00000173, 0.00000555, -0.00000072, 0.00000117, -0.00000213, -0.00000184, 0.00000116, -0.00001795, 0.00000329, -0.00000860, -0.00000433, -0.00000174, -0.00000743, -0.00000704, -0.00001074, -0.00002086, -0.00001123, -0.00007054, -0.00000109, 0.00008374, 0.00001452, 0.00001201, 0.00000354, -0.00003689, 0.00003688, },
		{ 0.00003492, 0.00012804, 0.00003871, -0.00003253, 0.00000396, 0.00001313, 0.00001819, 0.00008901, -0.00001497, -0.00007023, -0.00000830, -0.00002196, -0.00000949, -0.00000748, -0.00000757, -0.00000195, -0.00000361, -0.00000832, 0.00000197, -0.00001820, 0.00000363, -0.00000427, -0.00000052, -0.00000058, 0.00000182, 0.00000473, -0.00000257, -0.00000036, 0.00000007, -0.00000391, -0.00000032, 0.00000883, 0.00000187, -0.00000622, 0.00000060, -0.00000275, 0.00000113, -0.00000321, 0.00000294, -0.00000227, 0.00000688, -0.00001131, 0.00000461, -0.00000829, 0.00000791, -0.00000393, 0.00000394, -0.00000535, 0.00000503, -0.00000745, 0.00000446, -0.00000074, 0.00000438, -0.00000680, 0.00000335, -0.00000163, 0.00000323, -0.00000496, 0.00000373, -0.00000417, 0.00000503, -0.00000613, 0.00000399, -0.00000826, 0.00000812, -0.00000267, 0.00000722, -0.00000939, 0.00000517, -0.00000671, 0.00000582, -0.00000499, 0.00000461, -0.00000597, 0.00000405, -0.00000127, 0.00000227, -0.00000643, 0.00000466, -0.00000007, 0.00000290, -0.00000621, 0.00000371, -0.00000410, 0.00000295, -0.00000191, 0.00000537, -0.00000714, 0.00000423, -0.00001158, 0.00000803, -0.00000315, 0.00000393, -0.00000421, 0.00000190, -0.00000439, 0.00000232, -0.00000777, 0.00000429, 0.00000834, -0.00000163, -0.00000350, 0.00000060, -0.00000181, -0.00000041, 0.00000568, -0.00000272, 0.00000258, -0.00000342, 0.00000048, -0.00000166, -0.00001708, 0.00000423, -0.00000879, -0.00000510, -0.00000142, -0.00000736, -0.00000658, -0.00001212, -0.00001935, -0.00001535, -0.00006830, 0.00001121, 0.00007731, 0.00001189, 0.00001072, 0.00000328, -0.00004120, },
		{ -0.00004545, 0.00003279, 0.00012849, 0.00004043, -0.00002817, 0.00000457, 0.00001405, 0.00002308, 0.00009254, -0.00002987, -0.00006742, -0.00000684, -0.00002233, -0.00000859, -0.00000772, -0.00000793, -0.00000192, -0.00000304, -0.00000789, 0.00000019, -0.00001768, 0.00000548, -0.00000651, 0.00000119, -0.00000249, 0.00000475, 0.00000328, -0.00000293, -0.00000001, -0.00000002, -0.00000426, 0.00000072, 0.00000847, 0.00000104, -0.00000560, -0.00000013, -0.00000189, 0.00000058, -0.00000253, 0.00000227, -0.00000161, 0.00000594, -0.00001066, 0.00000427, -0.00000833, 0.00000880, -0.00000482, 0.00000438, -0.00000588, 0.00000559, -0.00000807, 0.00000544, -0.00000140, 0.00000447, -0.00000708, 0.00000405, -0.00000213, 0.00000314, -0.00000466, 0.00000343, -0.00000382, 0.00000464, -0.00000586, 0.00000339, -0.00000743, 0.00000829, -0.00000252, 0.00000645, -0.00000961, 0.00000563, -0.00000693, 0.00000614, -0.00000538, 0.00000507, -0.00000659, 0.00000471, -0.00000142, 0.00000191, -0.00000634, 0.00000504, -0.00000012, 0.00000245, -0.00000578, 0.00000313, -0.00000352, 0.00000256, -0.00000097, 0.00000393, -0.00000619, 0.00000358, -0.00001122, 0.00000819, -0.00000330, 0.00000418, -0.00000447, 0.00000198, -0.00000499, 0.00000313, -0.00000854, 0.00000572, 0.00000757, -0.00000190, -0.00000344, 0.00000100, -0.00000280, 0.00000131, 0.00000512, -0.00000409, 0.00000351, -0.00000418, 0.00000241, -0.00000457, -0.00001576, 0.00000486, -0.00000899, -0.00000580, -0.00000112, -0.00000722, -0.00000630, -0.00001333, -0.00001779, -0.00002025, -0.00006367, 0.00002155, 0.00007023, 0.00001011, 0.00000928, 0.00000315, },
		{ 0.00000313, -0.00004963, 0.00003049, 0.00012923, 0.00004207, -0.00002384, 0.00000540, 0.00001477, 0.00002934, 0.00009370, -0.00004503, -0.00006238, -0.00000695, -0.00002181, -0.00000822, -0.00000765, -0.00000859, -0.00000156, -0.00000271, -0.00000727, -0.00000205, -0.00001632, 0.00000651, -0.00000829, 0.00000272, -0.00000433, 0.00000786, 0.00000132, -0.00000282, 0.00000009, 0.00000003, -0.00000469, 0.00000198, 0.00000769, 0.00000055, -0.00000519, -0.00000065, -0.00000116, 0.00000002, -0.00000183, 0.00000155, -0.00000088, 0.00000483, -0.00000971, 0.00000356, -0.00000797, 0.00000935, -0.00000553, 0.00000471, -0.00000627, 0.00000599, -0.00000856, 0.00000642, -0.00000219, 0.00000469, -0.00000741, 0.00000481, -0.00000281, 0.00000327, -0.00000454, 0.00000326, -0.00000358, 0.00000431, -0.00000566, 0.00000287, -0.00000653, 0.00000832, -0.00000232, 0.00000550, -0.00000955, 0.00000595, -0.00000706, 0.00000636, -0.00000569, 0.00000549, -0.00000720, 0.00000549, -0.00000175, 0.00000166, -0.00000629, 0.00000554, -0.00000042, 0.00000228, -0.00000555, 0.00000269, -0.00000306, 0.00000234, -0.00000021, 0.00000254, -0.00000512, 0.00000273, -0.00001058, 0.00000811, -0.00000328, 0.00000429, -0.00000458, 0.00000183, -0.00000536, 0.00000380, -0.00000918, 0.00000717, 0.00000656, -0.00000194, -0.00000344, 0.00000141, -0.00000384, 0.00000333, 0.00000391, -0.00000482, 0.00000390, -0.00000430, 0.00000375, -0.00000730, -0.00001419, 0.00000529, -0.00000927, -0.00000632, -0.00000095, -0.00000689, -0.00000637, -0.00001411, -0.00001658, -0.00002539, -0.00005701, 0.00002974, 0.00006291, 0.00000901, 0.00000771, },
		{ 0.00000603, 0.00000322, -0.00005374, 0.00002802, 0.00013024, 0.00004369, -0.00001956, 0.00000647, 0.00001534, 0.00003704, 0.00009190, -0.00005959, -0.00005562, -0.00000853, -0.00002034, -0.00000846, -0.00000719, -0.00000958, -0.00000087, -0.00000263, -0.00000648, -0.00000471, -0.00001410, 0.00000656, -0.00000938, 0.00000385, -0.00000587, 0.00001092, -0.00000097, -0.00000233, -0.00000006, 0.00000018, -0.00000514, 0.00000340, 0.00000652, 0.00000043, -0.00000503, -0.00000089, -0.00000064, -0.00000045, -0.00000119, 0.00000087, -0.00000015, 0.00000364, -0.00000855, 0.00000255, -0.00000721, 0.00000953, -0.00000600, 0.00000487, -0.00000648, 0.00000617, -0.00000885, 0.00000731, -0.00000304, 0.00000499, -0.00000774, 0.00000559, -0.00000362, 0.00000361, -0.00000459, 0.00000324, -0.00000348, 0.00000407, -0.00000557, 0.00000248, -0.00000563, 0.00000827, -0.00000212, 0.00000440, -0.00000924, 0.00000611, -0.00000707, 0.00000644, -0.00000588, 0.00000580, -0.00000775, 0.00000631, -0.00000223, 0.00000152, -0.00000624, 0.00000612, -0.00000093, 0.00000238, -0.00000557, 0.00000243, -0.00000275, 0.00000232, 0.00000028, 0.00000130, -0.00000406, 0.00000177, -0.00000973, 0.00000783, -0.00000310, 0.00000426, -0.00000455, 0.00000146, -0.00000546, 0.00000426, -0.00000962, 0.00000854, 0.00000537, -0.00000181, -0.00000348, 0.00000177, -0.00000483, 0.00000550, 0.00000219, -0.00000495, 0.00000375, -0.00000373, 0.00000433, -0.00000964, -0.00001253, 0.00000560, -0.00000968, -0.00000659, -0.00000102, -0.00000627, -0.00000692, -0.00001425, -0.00001608, -0.00003023, -0.00004883, 0.00003580, 0.00005564, 0.00000847, },
		{ 0.00000838, 0.00000421, 0.00000344, -0.00005778, 0.00002536, 0.00013152, 0.00004529, -0.00001534, 0.00000777, 0.00001587, 0.00004611, 0.00008662, -0.00007263, -0.00004780, -0.00001130, -0.00001803, -0.00000927, -0.00000638, -0.00001087, 0.00000014, -0.00000277, -0.00000558, -0.00000765, -0.00001111, 0.00000562, -0.00000962, 0.00000436, -0.00000688, 0.00001371, -0.00000341, -0.00000156, -0.00000040, 0.00000042, -0.00000558, 0.00000491, 0.00000501, 0.00000065, -0.00000514, -0.00000081, -0.00000042, -0.00000071, -0.00000069, 0.00000031, 0.00000047, 0.00000248, -0.00000728, 0.00000133, -0.00000611, 0.00000933, -0.00000621, 0.00000485, -0.00000649, 0.00000610, -0.00000885, 0.00000802, -0.00000386, 0.00000531, -0.00000804, 0.00000634, -0.00000452, 0.00000410, -0.00000481, 0.00000338, -0.00000353, 0.00000395, -0.00000560, 0.00000228, -0.00000478, 0.00000819, -0.00000197, 0.00000324, -0.00000870, 0.00000610, -0.00000696, 0.00000639, -0.00000593, 0.00000598, -0.00000819, 0.00000713, -0.00000283, 0.00000144, -0.00000616, 0.00000669, -0.00000159, 0.00000271, -0.00000583, 0.00000238, -0.00000262, 0.00000253, 0.00000047, 0.00000030, -0.00000309, 0.00000080, -0.00000876, 0.00000738, -0.00000281, 0.00000415, -0.00000445, 0.00000089, -0.00000528, 0.00000448, -0.00000980, 0.00000974, 0.00000410, -0.00000156, -0.00000352, 0.00000203, -0.00000566, 0.00000766, 0.00000010, -0.00000459, 0.00000316, -0.00000252, 0.00000410, -0.00001143, -0.00001092, 0.00000583, -0.00001025, -0.00000657, -0.00000134, -0.00000533, -0.00000801, -0.00001361, -0.00001658, -0.00003423, -0.00003970, 0.00003989, 0.00004863, },
		{ 0.00004198, 0.00000866, 0.00000226, 0.00000381, -0.00006177, 0.00002252, 0.00013306, 0.00004687, -0.00001116, 0.00000922, 0.00001657, 0.00005633, 0.00007754, -0.00008325, -0.00003969, -0.00001483, -0.00001510, -0.00001051, -0.00000534, -0.00001233, 0.00000135, -0.00000303, -0.00000472, -0.00001070, -0.00000754, 0.00000373, -0.00000895, 0.00000414, -0.00000720, 0.00001602, -0.00000580, -0.00000063, -0.00000086, 0.00000068, -0.00000592, 0.00000641, 0.00000323, 0.00000121, -0.00000551, -0.00000040, -0.00000056, -0.00000069, -0.00000042, -0.00000004, 0.00000087, 0.00000147, -0.00000602, 0.00000001, -0.00000477, 0.00000880, -0.00000617, 0.00000466, -0.00000632, 0.00000578, -0.00000854, 0.00000848, -0.00000459, 0.00000561, -0.00000827, 0.00000700, -0.00000545, 0.00000472, -0.00000517, 0.00000367, -0.00000372, 0.00000395, -0.00000577, 0.00000229, -0.00000407, 0.00000815, -0.00000193, 0.00000208, -0.00000798, 0.00000593, -0.00000674, 0.00000620, -0.00000583, 0.00000599, -0.00000845, 0.00000787, -0.00000349, 0.00000141, -0.00000601, 0.00000721, -0.00000231, 0.00000321, -0.00000631, 0.00000253, -0.00000266, 0.00000297, 0.00000033, -0.00000039, -0.00000232, -0.00000008, -0.00000774, 0.00000685, -0.00000245, 0.00000400, -0.00000433, 0.00000020, -0.00000485, 0.00000444, -0.00000967, 0.00001069, 0.00000284, -0.00000123, -0.00000351, 0.00000213, -0.00000622, 0.00000963, -0.00000217, -0.00000390, 0.00000229, -0.00000080, 0.00000308, -0.00001260, -0.00000944, 0.00000601, -0.00001094, -0.00000629, -0.00000190, -0.00000412, -0.00000958, -0.00001218, -0.00001823, -0.00003696, -0.00003021, 0.00004225, },
		{ 0.00004320, 0.00003573, 0.00000925, 0.00000015, 0.00000431, -0.00006569, 0.00001947, 0.00013490, 0.00004842, -0.00000694, 0.00001072, 0.00001771, 0.00006729, 0.00006455, -0.00009069, -0.00003210, -0.00001860, -0.00001190, -0.00001196, -0.00000422, -0.00001380, 0.00000262, -0.00000328, -0.00000405, -0.00001360, -0.00000364, 0.00000109, -0.00000741, 0.00000314, -0.00000673, 0.00001766, -0.00000798, 0.00000033, -0.00000138, 0.00000092, -0.00000612, 0.00000780, 0.00000128, 0.00000203, -0.00000609, 0.00000033, -0.00000108, -0.00000034, -0.00000046, -0.00000010, 0.00000098, 0.00000072, -0.00000493, -0.00000127, -0.00000330, 0.00000803, -0.00000594, 0.00000433, -0.00000601, 0.00000524, -0.00000790, 0.00000864, -0.00000515, 0.00000583, -0.00000839, 0.00000753, -0.00000632, 0.00000539, -0.00000563, 0.00000408, -0.00000406, 0.00000406, -0.00000607, 0.00000253, -0.00000352, 0.00000820, -0.00000208, 0.00000101, -0.00000714, 0.00000565, -0.00000642, 0.00000590, -0.00000558, 0.00000584, -0.00000852, 0.00000848, -0.00000417, 0.00000141, -0.00000576, 0.00000759, -0.00000301, 0.00000380, -0.00000694, 0.00000284, -0.00000283, 0.00000360, -0.00000014, -0.00000075, -0.00000179, -0.00000080, -0.00000674, 0.00000626, -0.00000206, 0.00000385, -0.00000424, -0.00000054, -0.00000422, 0.00000416, -0.00000922, 0.00001132, 0.00000167, -0.00000087, -0.00000344, 0.00000203, -0.00000641, 0.00001125, -0.00000441, -0.00000305, 0.00000130, 0.00000124, 0.00000139, -0.00001314, -0.00000813, 0.00000610, -0.00001168, -0.00000582, -0.00000261, -0.00000275, -0.00001152, -0.00001007, -0.00002104, -0.00003812, -0.00002089, },
		{ -0.00001220, 0.00004304, 0.00002991, 0.00001009, -0.00000208, 0.00000495, -0.00006953, 0.00001616, 0.00013708, 0.00004987, -0.00000259, 0.00001212, 0.00001968, 0.00007836, 0.00004782, -0.00009436, -0.00002576, -0.00002205, -0.00000880, -0.00001335, -0.00000326, -0.00001508, 0.00000374, -0.00000335, -0.00000377, -0.00001607, 0.00000024, -0.00000204, -0.00000517, 0.00000142, -0.00000545, 0.00001855, -0.00000977, 0.00000122, -0.00000190, 0.00000111, -0.00000612, 0.00000900, -0.00000073, 0.00000303, -0.00000682, 0.00000131, -0.00000197, 0.00000038, -0.00000083, 0.00000017, 0.00000072, 0.00000034, -0.00000411, -0.00000237, -0.00000183, 0.00000711, -0.00000557, 0.00000393, -0.00000563, 0.00000455, -0.00000699, 0.00000848, -0.00000550, 0.00000593, -0.00000837, 0.00000789, -0.00000710, 0.00000606, -0.00000613, 0.00000456, -0.00000450, 0.00000426, -0.00000645, 0.00000297, -0.00000316, 0.00000836, -0.00000246, 0.00000012, -0.00000626, 0.00000528, -0.00000605, 0.00000550, -0.00000522, 0.00000552, -0.00000838, 0.00000893, -0.00000483, 0.00000142, -0.00000538, 0.00000779, -0.00000360, 0.00000438, -0.00000765, 0.00000326, -0.00000309, 0.00000436, -0.00000088, -0.00000078, -0.00000154, -0.00000132, -0.00000582, 0.00000567, -0.00000166, 0.00000372, -0.00000425, -0.00000126, -0.00000343, 0.00000366, -0.00000845, 0.00001160, 0.00000063, -0.00000054, -0.00000328, 0.00000171, -0.00000620, 0.00001239, -0.00000646, -0.00000224, 0.00000042, 0.00000337, -0.00000080, -0.00001309, -0.00000698, 0.00000606, -0.00001237, -0.00000529, -0.00000334, -0.00000138, -0.00001359, -0.00000751, -0.00002487, -0.00003758, },
		{ -0.00003538, -0.00000445, 0.00004205, 0.00002453, 0.00001110, -0.00000442, 0.00000569, -0.00007324, 0.00001253, 0.00013968, 0.00005113, 0.00000202, 0.00001322, 0.00002294, 0.00008875, 0.00002785, -0.00009399, -0.00002129, -0.00002468, -0.00000619, -0.00001440, -0.00000267, -0.00001596, 0.00000454, -0.00000308, -0.00000408, -0.00001784, 0.00000375, -0.00000532, -0.00000247, -0.00000087, -0.00000343, 0.00001863, -0.00001110, 0.00000194, -0.00000234, 0.00000120, -0.00000587, 0.00000990, -0.00000269, 0.00000413, -0.00000761, 0.00000246, -0.00000318, 0.00000145, -0.00000155, 0.00000081, 0.00000004, 0.00000039, -0.00000368, -0.00000315, -0.00000051, 0.00000616, -0.00000516, 0.00000353, -0.00000527, 0.00000381, -0.00000585, 0.00000803, -0.00000564, 0.00000590, -0.00000819, 0.00000803, -0.00000771, 0.00000667, -0.00000662, 0.00000507, -0.00000501, 0.00000452, -0.00000690, 0.00000360, -0.00000298, 0.00000864, -0.00000310, -0.00000054, -0.00000541, 0.00000489, -0.00000566, 0.00000507, -0.00000477, 0.00000506, -0.00000802, 0.00000918, -0.00000544, 0.00000144, -0.00000488, 0.00000777, -0.00000401, 0.00000487, -0.00000835, 0.00000372, -0.00000335, 0.00000517, -0.00000182, -0.00000052, -0.00000157, -0.00000160, -0.00000501, 0.00000510, -0.00000128, 0.00000363, -0.00000439, -0.00000188, -0.00000258, 0.00000300, -0.00000740, 0.00001152, -0.00000021, -0.00000025, -0.00000304, 0.00000119, -0.00000556, 0.00001296, -0.00000815, -0.00000161, -0.00000018, 0.00000533, -0.00000328, -0.00001257, -0.00000595, 0.00000581, -0.00001289, -0.00000483, -0.00000394, -0.00000023, -0.00001554, -0.00000481, -0.00002940, },
		{ -0.00003425, -0.00003174, 0.00000221, 0.00004045, 0.00001961, 0.00001221, -0.00000679, 0.00000646, -0.00007676, 0.00000850, 0.00014277, 0.00005213, 0.00000703, 0.00001384, 0.00002794, 0.00009748, 0.00000543, -0.00008958, -0.00001909, -0.00002607, -0.00000438, -0.00001487, -0.00000265, -0.00001626, 0.00000483, -0.00000231, -0.00000513, -0.00001866, 0.00000654, -0.00000837, 0.00000040, -0.00000350, -0.00000081, 0.00001795, -0.00001189, 0.00000243, -0.00000268, 0.00000119, -0.00000535, 0.00001045, -0.00000449, 0.00000523, -0.00000836, 0.00000366, -0.00000460, 0.00000280, -0.00000258, 0.00000178, -0.00000104, 0.00000091, -0.00000370, -0.00000352, 0.00000054, 0.00000530, -0.00000478, 0.00000319, -0.00000501, 0.00000312, -0.00000460, 0.00000733, -0.00000557, 0.00000573, -0.00000786, 0.00000796, -0.00000812, 0.00000717, -0.00000705, 0.00000555, -0.00000554, 0.00000479, -0.00000734, 0.00000434, -0.00000296, 0.00000903, -0.00000400, -0.00000091, -0.00000466, 0.00000451, -0.00000530, 0.00000463, -0.00000429, 0.00000450, -0.00000747, 0.00000922, -0.00000599, 0.00000146, -0.00000426, 0.00000752, -0.00000418, 0.00000518, -0.00000897, 0.00000416, -0.00000357, 0.00000594, -0.00000288, -0.00000004, -0.00000184, -0.00000165, -0.00000433, 0.00000458, -0.00000091, 0.00000357, -0.00000469, -0.00000234, -0.00000174, 0.00000226, -0.00000612, 0.00001109, -0.00000084, -0.00000002, -0.00000273, 0.00000052, -0.00000453, 0.00001293, -0.00000941, -0.00000127, -0.00000033, 0.00000692, -0.00000583, -0.00001170, -0.00000499, 0.00000528, -0.00001312, -0.00000458, -0.00000425, 0.00000050, -0.00001708, -0.00000235, },
		{ -0.00000049, -0.00003893, -0.00002701, 0.00000773, 0.00003838, 0.00001519, 0.00001328, -0.00000912, 0.00000720, -0.00008002, 0.00000399, 0.00014645, 0.00005277, 0.00001256, 0.00001382, 0.00003512, 0.00010351, -0.00001836, -0.00008150, -0.00001933, -0.00002599, -0.00000360, -0.00001459, -0.00000333, -0.00001586, 0.00000452, -0.00000097, -0.00000699, -0.00001835, 0.00000830, -0.00001083, 0.00000309, -0.00000620, 0.00000220, 0.00001659, -0.00001216, 0.00000265, -0.00000290, 0.00000108, -0.00000457, 0.00001058, -0.00000605, 0.00000623, -0.00000897, 0.00000477, -0.00000612, 0.00000436, -0.00000385, 0.00000303, -0.00000248, 0.00000188, -0.00000420, -0.00000340, 0.00000122, 0.00000461, -0.00000451, 0.00000300, -0.00000494, 0.00000259, -0.00000333, 0.00000647, -0.00000532, 0.00000545, -0.00000740, 0.00000768, -0.00000833, 0.00000751, -0.00000737, 0.00000596, -0.00000605, 0.00000504, -0.00000772, 0.00000515, -0.00000305, 0.00000948, -0.00000513, -0.00000096, -0.00000405, 0.00000419, -0.00000500, 0.00000423, -0.00000381, 0.00000390, -0.00000677, 0.00000905, -0.00000646, 0.00000152, -0.00000357, 0.00000705, -0.00000409, 0.00000525, -0.00000944, 0.00000450, -0.00000366, 0.00000659, -0.00000395, 0.00000059, -0.00000231, -0.00000149, -0.00000379, 0.00000409, -0.00000054, 0.00000353, -0.00000514, -0.00000258, -0.00000097, 0.00000150, -0.00000468, 0.00001035, -0.00000126, 0.00000016, -0.00000239, -0.00000025, -0.00000317, 0.00001231, -0.00001017, -0.00000129, 0.00000006, 0.00000793, -0.00000822, -0.00001062, -0.00000405, 0.00000442, -0.00001296, -0.00000465, -0.00000415, 0.00000065, -0.00001795, },
		{ -0.00001795, 0.00000042, -0.00004296, -0.00002161, 0.00001222, 0.00003593, 0.00001133, 0.00001421, -0.00001131, 0.00000782, -0.00008294, -0.00000107, 0.00015076, 0.00005300, 0.00001869, 0.00001310, 0.00004484, 0.00010575, -0.00004223, -0.00007041, -0.00002186, -0.00002439, -0.00000393, -0.00001349, -0.00000474, -0.00001472, 0.00000357, 0.00000096, -0.00000969, -0.00001683, 0.00000880, -0.00001236, 0.00000526, -0.00000866, 0.00000533, 0.00001474, -0.00001196, 0.00000262, -0.00000300, 0.00000090, -0.00000356, 0.00001029, -0.00000730, 0.00000706, -0.00000934, 0.00000568, -0.00000759, 0.00000599, -0.00000527, 0.00000447, -0.00000421, 0.00000324, -0.00000515, -0.00000277, 0.00000147, 0.00000419, -0.00000442, 0.00000300, -0.00000512, 0.00000232, -0.00000217, 0.00000552, -0.00000495, 0.00000508, -0.00000684, 0.00000721, -0.00000831, 0.00000768, -0.00000755, 0.00000626, -0.00000651, 0.00000523, -0.00000800, 0.00000595, -0.00000320, 0.00000995, -0.00000645, -0.00000069, -0.00000363, 0.00000397, -0.00000480, 0.00000393, -0.00000340, 0.00000330, -0.00000595, 0.00000870, -0.00000685, 0.00000162, -0.00000285, 0.00000640, -0.00000375, 0.00000505, -0.00000969, 0.00000470, -0.00000358, 0.00000703, -0.00000494, 0.00000128, -0.00000290, -0.00000114, -0.00000337, 0.00000365, -0.00000016, 0.00000348, -0.00000573, -0.00000259, -0.00000036, 0.00000080, -0.00000316, 0.00000934, -0.00000147, 0.00000029, -0.00000207, -0.00000103, -0.00000159, 0.00001117, -0.00001047, -0.00000164, 0.00000100, 0.00000825, -0.00001029, -0.00000942, -0.00000312, 0.00000322, -0.00001237, -0.00000511, -0.00000357, 0.00000010, },
		{ -0.00000116, -0.00001699, 0.00000017, -0.00004592, -0.00001602, 0.00001585, 0.00003313, 0.00000809, 0.00001487, -0.00001327, 0.00000825, -0.00008547, -0.00000674, 0.00015573, 0.00005284, 0.00002542, 0.00001172, 0.00005724, 0.00010325, -0.00006476, -0.00005726, -0.00002633, -0.00002140, -0.00000530, -0.00001163, -0.00000682, -0.00001290, 0.00000203, 0.00000338, -0.00001310, -0.00001414, 0.00000795, -0.00001274, 0.00000662, -0.00001058, 0.00000831, 0.00001258, -0.00001139, 0.00000237, -0.00000301, 0.00000069, -0.00000236, 0.00000958, -0.00000818, 0.00000766, -0.00000941, 0.00000626, -0.00000888, 0.00000756, -0.00000669, 0.00000596, -0.00000609, 0.00000489, -0.00000649, -0.00000166, 0.00000126, 0.00000406, -0.00000454, 0.00000321, -0.00000559, 0.00000240, -0.00000123, 0.00000461, -0.00000453, 0.00000467, -0.00000623, 0.00000660, -0.00000810, 0.00000767, -0.00000758, 0.00000642, -0.00000687, 0.00000533, -0.00000812, 0.00000666, -0.00000332, 0.00001034, -0.00000788, -0.00000012, -0.00000342, 0.00000387, -0.00000472, 0.00000373, -0.00000308, 0.00000277, -0.00000508, 0.00000818, -0.00000717, 0.00000179, -0.00000216, 0.00000563, -0.00000320, 0.00000459, -0.00000970, 0.00000473, -0.00000329, 0.00000722, -0.00000577, 0.00000195, -0.00000356, -0.00000066, -0.00000307, 0.00000323, 0.00000024, 0.00000336, -0.00000640, -0.00000234, 0.00000006, 0.00000024, -0.00000165, 0.00000813, -0.00000151, 0.00000039, -0.00000182, -0.00000172, 0.00000007, 0.00000964, -0.00001038, -0.00000225, 0.00000245, 0.00000784, -0.00001191, -0.00000820, -0.00000220, 0.00000172, -0.00001136, -0.00000594, -0.00000251, },
		{ -0.00000107, -0.00000306, -0.00001510, -0.00000136, -0.00004748, -0.00001068, 0.00001883, 0.00002999, 0.00000552, 0.00001517, -0.00001492, 0.00000843, -0.00008758, -0.00001300, 0.00016134, 0.00005234, 0.00003266, 0.00000990, 0.00007225, 0.00009524, -0.00008454, -0.00004316, -0.00003211, -0.00001737, -0.00000750, -0.00000920, -0.00000940, -0.00001056, 0.00000005, 0.00000612, -0.00001702, -0.00001044, 0.00000577, -0.00001185, 0.00000697, -0.00001171, 0.00001087, 0.00001033, -0.00001055, 0.00000196, -0.00000297, 0.00000050, -0.00000106, 0.00000849, -0.00000869, 0.00000800, -0.00000912, 0.00000641, -0.00000983, 0.00000893, -0.00000799, 0.00000737, -0.00000797, 0.00000668, -0.00000809, -0.00000015, 0.00000062, 0.00000423, -0.00000487, 0.00000362, -0.00000636, 0.00000288, -0.00000059, 0.00000381, -0.00000413, 0.00000428, -0.00000562, 0.00000591, -0.00000772, 0.00000750, -0.00000745, 0.00000642, -0.00000713, 0.00000533, -0.00000806, 0.00000724, -0.00000338, 0.00001060, -0.00000934, 0.00000069, -0.00000340, 0.00000389, -0.00000476, 0.00000368, -0.00000291, 0.00000236, -0.00000421, 0.00000753, -0.00000743, 0.00000208, -0.00000157, 0.00000483, -0.00000249, 0.00000389, -0.00000949, 0.00000459, -0.00000280, 0.00000713, -0.00000637, 0.00000251, -0.00000421, -0.00000008, -0.00000286, 0.00000283, 0.00000070, 0.00000314, -0.00000712, -0.00000187, 0.00000025, -0.00000013, -0.00000023, 0.00000680, -0.00000141, 0.00000051, -0.00000169, -0.00000221, 0.00000166, 0.00000786, -0.00001002, -0.00000299, 0.00000427, 0.00000675, -0.00001304, -0.00000698, -0.00000134, -0.00000000, -0.00001002, -0.00000708, },
		{ -0.00000838, 0.00000059, -0.00000540, -0.00001245, -0.00000410, -0.00004746, -0.00000595, 0.00002135, 0.00002652, 0.00000364, 0.00001505, -0.00001621, 0.00000833, -0.00008929, -0.00001983, 0.00016747, 0.00005167, 0.00004021, 0.00000800, 0.00008950, 0.00008124, -0.00010026, -0.00002933, -0.00003847, -0.00001278, -0.00001017, -0.00000648, -0.00001220, -0.00000796, -0.00000214, 0.00000894, -0.00002115, -0.00000602, 0.00000243, -0.00000970, 0.00000620, -0.00001187, 0.00001278, 0.00000820, -0.00000959, 0.00000145, -0.00000293, 0.00000040, 0.00000026, 0.00000709, -0.00000884, 0.00000810, -0.00000847, 0.00000607, -0.00001034, 0.00000996, -0.00000902, 0.00000855, -0.00000971, 0.00000846, -0.00000981, 0.00000164, -0.00000038, 0.00000467, -0.00000537, 0.00000420, -0.00000738, 0.00000376, -0.00000034, 0.00000321, -0.00000381, 0.00000396, -0.00000508, 0.00000518, -0.00000723, 0.00000720, -0.00000718, 0.00000627, -0.00000727, 0.00000523, -0.00000779, 0.00000764, -0.00000330, 0.00001064, -0.00001073, 0.00000168, -0.00000356, 0.00000401, -0.00000492, 0.00000376, -0.00000289, 0.00000210, -0.00000340, 0.00000678, -0.00000763, 0.00000248, -0.00000113, 0.00000409, -0.00000172, 0.00000301, -0.00000908, 0.00000431, -0.00000212, 0.00000675, -0.00000672, 0.00000293, -0.00000481, 0.00000055, -0.00000273, 0.00000246, 0.00000120, 0.00000279, -0.00000781, -0.00000119, 0.00000019, -0.00000025, 0.00000102, 0.00000542, -0.00000121, 0.00000065, -0.00000171, -0.00000243, 0.00000303, 0.00000601, -0.00000954, -0.00000370, 0.00000627, 0.00000509, -0.00001369, -0.00000577, -0.00000064, -0.00000180, -0.00000847, },
		{ -0.00000689, -0.00000966, 0.00000224, -0.00000790, -0.00000931, -0.00000785, -0.00004585, -0.00000207, 0.00002359, 0.00002275, 0.00000243, 0.00001447, -0.00001715, 0.00000799, -0.00009065, -0.00002711, 0.00017398, 0.00005107, 0.00004776, 0.00000661, 0.00010829, 0.00006118, -0.00011088, -0.00001694, -0.00004455, -0.00000822, -0.00001290, -0.00000381, -0.00001490, -0.00000538, -0.00000426, 0.00001152, -0.00002512, -0.00000128, -0.00000177, -0.00000643, 0.00000432, -0.00001098, 0.00001387, 0.00000638, -0.00000864, 0.00000093, -0.00000294, 0.00000045, 0.00000151, 0.00000547, -0.00000868, 0.00000798, -0.00000749, 0.00000523, -0.00001034, 0.00001056, -0.00000967, 0.00000937, -0.00001113, 0.00001004, -0.00001147, 0.00000356, -0.00000163, 0.00000531, -0.00000600, 0.00000487, -0.00000860, 0.00000501, -0.00000047, 0.00000287, -0.00000365, 0.00000377, -0.00000465, 0.00000450, -0.00000669, 0.00000681, -0.00000680, 0.00000598, -0.00000731, 0.00000505, -0.00000734, 0.00000783, -0.00000305, 0.00001041, -0.00001195, 0.00000277, -0.00000386, 0.00000422, -0.00000518, 0.00000396, -0.00000302, 0.00000203, -0.00000270, 0.00000598, -0.00000777, 0.00000302, -0.00000091, 0.00000348, -0.00000098, 0.00000203, -0.00000852, 0.00000393, -0.00000132, 0.00000612, -0.00000681, 0.00000317, -0.00000530, 0.00000121, -0.00000267, 0.00000212, 0.00000174, 0.00000227, -0.00000842, -0.00000035, -0.00000010, -0.00000010, 0.00000203, 0.00000408, -0.00000097, 0.00000084, -0.00000192, -0.00000230, 0.00000405, 0.00000423, -0.00000908, -0.00000418, 0.00000822, 0.00000303, -0.00001393, -0.00000450, -0.00000020, -0.00000349, },
		{ -0.00000491, -0.00000551, -0.00001068, 0.00000359, -0.00001022, -0.00000607, -0.00001223, -0.00004282, 0.00000085, 0.00002563, 0.00001876, 0.00000182, 0.00001349, -0.00001778, 0.00000747, -0.00009177, -0.00003470, 0.00018064, 0.00005085, 0.00005490, 0.00000645, 0.00012757, 0.00003546, -0.00011567, -0.00000699, -0.00004958, -0.00000427, -0.00001521, -0.00000159, -0.00001716, -0.00000316, -0.00000600, 0.00001354, -0.00002853, 0.00000334, -0.00000642, -0.00000232, 0.00000148, -0.00000905, 0.00001405, 0.00000501, -0.00000780, 0.00000045, -0.00000304, 0.00000072, 0.00000257, 0.00000374, -0.00000829, 0.00000771, -0.00000626, 0.00000392, -0.00000980, 0.00001066, -0.00000987, 0.00000972, -0.00001211, 0.00001127, -0.00001289, 0.00000544, -0.00000300, 0.00000607, -0.00000667, 0.00000554, -0.00000989, 0.00000655, -0.00000100, 0.00000282, -0.00000367, 0.00000373, -0.00000438, 0.00000391, -0.00000614, 0.00000639, -0.00000636, 0.00000559, -0.00000726, 0.00000482, -0.00000673, 0.00000784, -0.00000264, 0.00000986, -0.00001291, 0.00000387, -0.00000424, 0.00000449, -0.00000550, 0.00000427, -0.00000331, 0.00000216, -0.00000214, 0.00000514, -0.00000786, 0.00000368, -0.00000092, 0.00000309, -0.00000038, 0.00000105, -0.00000788, 0.00000352, -0.00000046, 0.00000530, -0.00000666, 0.00000323, -0.00000568, 0.00000188, -0.00000270, 0.00000185, 0.00000230, 0.00000158, -0.00000889, 0.00000059, -0.00000060, 0.00000033, 0.00000274, 0.00000284, -0.00000071, 0.00000110, -0.00000232, -0.00000178, 0.00000462, 0.00000268, -0.00000878, -0.00000428, 0.00000988, 0.00000079, -0.00001386, -0.00000312, -0.00000015, },
		{ -0.00000061, -0.00000586, -0.00000450, -0.00001123, 0.00000438, -0.00001201, -0.00000318, -0.00001678, -0.00003869, 0.00000285, 0.00002747, 0.00001469, 0.00000166, 0.00001219, -0.00001818, 0.00000687, -0.00009278, -0.00004243, 0.00018719, 0.00005137, 0.00006119, 0.00000837, 0.00014600, 0.00000493, -0.00011439, -0.00000027, -0.00005287, -0.00000147, -0.00001671, -0.00000015, -0.00001864, -0.00000159, -0.00000708, 0.00001469, -0.00003099, 0.00000736, -0.00001105, 0.00000225, -0.00000209, -0.00000620, 0.00001330, 0.00000418, -0.00000717, 0.00000007, -0.00000325, 0.00000124, 0.00000335, 0.00000202, -0.00000775, 0.00000739, -0.00000489, 0.00000222, -0.00000875, 0.00001025, -0.00000956, 0.00000955, -0.00001256, 0.00001203, -0.00001393, 0.00000714, -0.00000437, 0.00000683, -0.00000730, 0.00000610, -0.00001114, 0.00000829, -0.00000186, 0.00000306, -0.00000389, 0.00000387, -0.00000431, 0.00000347, -0.00000566, 0.00000597, -0.00000590, 0.00000512, -0.00000716, 0.00000458, -0.00000602, 0.00000768, -0.00000206, 0.00000898, -0.00001355, 0.00000490, -0.00000466, 0.00000476, -0.00000585, 0.00000465, -0.00000372, 0.00000248, -0.00000176, 0.00000431, -0.00000787, 0.00000445, -0.00000119, 0.00000297, 0.00000000, 0.00000016, -0.00000725, 0.00000314, 0.00000038, 0.00000437, -0.00000633, 0.00000312, -0.00000594, 0.00000255, -0.00000282, 0.00000167, 0.00000283, 0.00000073, -0.00000919, 0.00000156, -0.00000129, 0.00000104, 0.00000313, 0.00000176, -0.00000048, 0.00000141, -0.00000290, -0.00000089, 0.00000467, 0.00000145, -0.00000874, -0.00000384, 0.00001104, -0.00000141, -0.00001361, -0.00000156, },
		{ 0.00000023, -0.00000166, -0.00000621, -0.00000405, -0.00001112, 0.00000437, -0.00001293, -0.00000108, -0.00002097, -0.00003390, 0.00000411, 0.00002904, 0.00001070, 0.00000179, 0.00001072, -0.00001850, 0.00000632, -0.00009385, -0.00005009, 0.00019339, 0.00005297, 0.00006620, 0.00001324, 0.00016200, -0.00002908, -0.00010724, 0.00000282, -0.00005394, -0.00000022, -0.00001706, 0.00000023, -0.00001911, -0.00000092, -0.00000728, 0.00001474, -0.00003216, 0.00001033, -0.00001517, 0.00000685, -0.00000606, -0.00000265, 0.00001171, 0.00000391, -0.00000679, -0.00000018, -0.00000359, 0.00000204, 0.00000378, 0.00000040, -0.00000715, 0.00000710, -0.00000352, 0.00000028, -0.00000728, 0.00000938, -0.00000878, 0.00000884, -0.00001242, 0.00001223, -0.00001447, 0.00000849, -0.00000559, 0.00000751, -0.00000781, 0.00000647, -0.00001222, 0.00001009, -0.00000297, 0.00000355, -0.00000430, 0.00000419, -0.00000444, 0.00000321, -0.00000528, 0.00000562, -0.00000547, 0.00000464, -0.00000704, 0.00000438, -0.00000525, 0.00000740, -0.00000136, 0.00000778, -0.00001382, 0.00000579, -0.00000506, 0.00000502, -0.00000619, 0.00000505, -0.00000421, 0.00000298, -0.00000156, 0.00000349, -0.00000781, 0.00000528, -0.00000169, 0.00000314, 0.00000010, -0.00000055, -0.00000669, 0.00000287, 0.00000111, 0.00000342, -0.00000588, 0.00000289, -0.00000609, 0.00000323, -0.00000304, 0.00000163, 0.00000328, -0.00000025, -0.00000927, 0.00000252, -0.00000210, 0.00000200, 0.00000318, 0.00000087, -0.00000029, 0.00000177, -0.00000361, 0.00000036, 0.00000421, 0.00000059, -0.00000900, -0.00000280, 0.00001152, -0.00000337, -0.00001330, },
		{ -0.00001303, 0.00000227, -0.00000332, -0.00000588, -0.00000425, -0.00001024, 0.00000338, -0.00001272, -0.00000016, -0.00002428, -0.00002895, 0.00000487, 0.00003022, 0.00000701, 0.00000199, 0.00000922, -0.00001887, 0.00000596, -0.00009510, -0.00005754, 0.00019900, 0.00005596, 0.00006955, 0.00002188, 0.00017386, -0.00006485, -0.00009492, 0.00000223, -0.00005259, -0.00000073, -0.00001609, -0.00000059, -0.00001843, -0.00000128, -0.00000649, 0.00001355, -0.00003182, 0.00001189, -0.00001833, 0.00001104, -0.00001004, 0.00000133, 0.00000941, 0.00000419, -0.00000669, -0.00000029, -0.00000406, 0.00000313, 0.00000379, -0.00000099, -0.00000659, 0.00000696, -0.00000229, -0.00000176, -0.00000551, 0.00000813, -0.00000759, 0.00000764, -0.00001172, 0.00001184, -0.00001442, 0.00000940, -0.00000657, 0.00000802, -0.00000812, 0.00000654, -0.00001299, 0.00001180, -0.00000423, 0.00000424, -0.00000488, 0.00000466, -0.00000476, 0.00000314, -0.00000503, 0.00000537, -0.00000512, 0.00000416, -0.00000694, 0.00000428, -0.00000452, 0.00000708, -0.00000062, 0.00000632, -0.00001371, 0.00000649, -0.00000540, 0.00000521, -0.00000647, 0.00000542, -0.00000474, 0.00000363, -0.00000156, 0.00000270, -0.00000764, 0.00000613, -0.00000239, 0.00000361, -0.00000013, -0.00000103, -0.00000626, 0.00000276, 0.00000166, 0.00000252, -0.00000537, 0.00000258, -0.00000615, 0.00000390, -0.00000339, 0.00000175, 0.00000358, -0.00000130, -0.00000914, 0.00000340, -0.00000298, 0.00000316, 0.00000290, 0.00000020, -0.00000015, 0.00000214, -0.00000440, 0.00000188, 0.00000326, 0.00000010, -0.00000955, -0.00000114, 0.00001122, -0.00000492, },
		{ -0.00000597, -0.00001283, 0.00000452, -0.00000555, -0.00000487, -0.00000512, -0.00000856, 0.00000137, -0.00001124, -0.00000071, -0.00002624, -0.00002435, 0.00000545, 0.00003082, 0.00000384, 0.00000205, 0.00000788, -0.00001941, 0.00000588, -0.00009665, -0.00006465, 0.00020389, 0.00006054, 0.00007105, 0.00003494, 0.00017989, -0.00010037, -0.00007859, -0.00000168, -0.00004889, -0.00000300, -0.00001380, -0.00000259, -0.00001659, -0.00000267, -0.00000471, 0.00001114, -0.00002990, 0.00001181, -0.00002014, 0.00001439, -0.00001365, 0.00000542, 0.00000661, 0.00000493, -0.00000683, -0.00000027, -0.00000462, 0.00000447, 0.00000336, -0.00000209, -0.00000615, 0.00000706, -0.00000136, -0.00000369, -0.00000361, 0.00000664, -0.00000611, 0.00000606, -0.00001052, 0.00001089, -0.00001378, 0.00000980, -0.00000723, 0.00000830, -0.00000819, 0.00000626, -0.00001334, 0.00001328, -0.00000551, 0.00000504, -0.00000557, 0.00000525, -0.00000526, 0.00000327, -0.00000495, 0.00000524, -0.00000488, 0.00000375, -0.00000689, 0.00000432, -0.00000388, 0.00000678, 0.00000010, 0.00000467, -0.00001322, 0.00000696, -0.00000565, 0.00000531, -0.00000667, 0.00000574, -0.00000525, 0.00000439, -0.00000175, 0.00000195, -0.00000734, 0.00000693, -0.00000320, 0.00000432, -0.00000070, -0.00000124, -0.00000601, 0.00000286, 0.00000197, 0.00000176, -0.00000489, 0.00000224, -0.00000615, 0.00000459, -0.00000385, 0.00000208, 0.00000367, -0.00000236, -0.00000879, 0.00000417, -0.00000387, 0.00000447, 0.00000233, -0.00000025, -0.00000007, 0.00000251, -0.00000518, 0.00000356, 0.00000191, -0.00000008, -0.00001030, 0.00000107, 0.00001009, },
		{ 0.00000821, -0.00000649, -0.00001271, 0.00000688, -0.00000825, -0.00000327, -0.00000658, -0.00000616, -0.00000159, -0.00000849, -0.00000285, -0.00002652, -0.00002055, 0.00000615, 0.00003068, 0.00000138, 0.00000178, 0.00000682, -0.00002024, 0.00000616, -0.00009854, -0.00007141, 0.00020803, 0.00006676, 0.00007070, 0.00005278, 0.00017857, -0.00013349, -0.00005975, -0.00000819, -0.00004323, -0.00000679, -0.00001037, -0.00000560, -0.00001376, -0.00000495, -0.00000209, 0.00000766, -0.00002647, 0.00001000, -0.00002038, 0.00001657, -0.00001654, 0.00000928, 0.00000354, 0.00000601, -0.00000717, -0.00000015, -0.00000523, 0.00000603, 0.00000247, -0.00000283, -0.00000589, 0.00000748, -0.00000088, -0.00000534, -0.00000177, 0.00000510, -0.00000451, 0.00000425, -0.00000895, 0.00000948, -0.00001259, 0.00000968, -0.00000753, 0.00000833, -0.00000802, 0.00000562, -0.00001322, 0.00001443, -0.00000669, 0.00000588, -0.00000632, 0.00000591, -0.00000590, 0.00000358, -0.00000501, 0.00000527, -0.00000479, 0.00000341, -0.00000689, 0.00000453, -0.00000339, 0.00000658, 0.00000069, 0.00000292, -0.00001240, 0.00000719, -0.00000578, 0.00000531, -0.00000675, 0.00000595, -0.00000571, 0.00000520, -0.00000210, 0.00000126, -0.00000689, 0.00000761, -0.00000406, 0.00000522, -0.00000156, -0.00000116, -0.00000593, 0.00000317, 0.00000201, 0.00000120, -0.00000449, 0.00000191, -0.00000609, 0.00000526, -0.00000444, 0.00000262, 0.00000350, -0.00000337, -0.00000825, 0.00000477, -0.00000470, 0.00000585, 0.00000150, -0.00000048, -0.00000003, 0.00000281, -0.00000587, 0.00000528, 0.00000028, -0.00000005, -0.00001113, 0.00000368, },
		{ 0.00000648, 0.00000572, -0.00000651, -0.00001261, 0.00000921, -0.00001121, -0.00000124, -0.00000848, -0.00000324, -0.00000531, -0.00000465, -0.00000653, -0.00002496, -0.00001788, 0.00000723, 0.00002970, -0.00000024, 0.00000105, 0.00000613, -0.00002139, 0.00000677, -0.00010071, -0.00007791, 0.00021152, 0.00007450, 0.00006880, 0.00007535, 0.00016867, -0.00016209, -0.00004011, -0.00001630, -0.00003622, -0.00001165, -0.00000617, -0.00000930, -0.00001023, -0.00000785, 0.00000107, 0.00000340, -0.00002179, 0.00000658, -0.00001897, 0.00001737, -0.00001842, 0.00001262, 0.00000043, 0.00000732, -0.00000767, 0.00000002, -0.00000584, 0.00000773, 0.00000116, -0.00000318, -0.00000586, 0.00000827, -0.00000095, -0.00000653, -0.00000018, 0.00000368, -0.00000297, 0.00000239, -0.00000719, 0.00000775, -0.00001094, 0.00000909, -0.00000750, 0.00000813, -0.00000763, 0.00000464, -0.00001257, 0.00001513, -0.00000768, 0.00000665, -0.00000706, 0.00000659, -0.00000661, 0.00000402, -0.00000522, 0.00000543, -0.00000484, 0.00000317, -0.00000697, 0.00000492, -0.00000312, 0.00000656, 0.00000107, 0.00000118, -0.00001132, 0.00000719, -0.00000581, 0.00000521, -0.00000672, 0.00000605, -0.00000606, 0.00000603, -0.00000260, 0.00000062, -0.00000627, 0.00000810, -0.00000487, 0.00000621, -0.00000266, -0.00000083, -0.00000603, 0.00000367, 0.00000176, 0.00000086, -0.00000421, 0.00000162, -0.00000598, 0.00000592, -0.00000511, 0.00000336, 0.00000303, -0.00000425, -0.00000755, 0.00000519, -0.00000541, 0.00000722, 0.00000048, -0.00000053, -0.00000002, 0.00000302, -0.00000639, 0.00000690, -0.00000147, 0.00000007, -0.00001187, },
		{ -0.00001231, 0.00000921, 0.00000281, -0.00000616, -0.00001242, 0.00001133, -0.00001420, 0.00000097, -0.00001056, -0.00000010, -0.00000945, -0.00000002, -0.00001151, -0.00002160, -0.00001651, 0.00000883, 0.00002782, -0.00000095, -0.00000022, 0.00000583, -0.00002282, 0.00000764, -0.00010302, -0.00008436, 0.00021464, 0.00008342, 0.00006595, 0.00010214, 0.00014948, -0.00018427, -0.00002147, -0.00002483, -0.00002870, -0.00001698, -0.00000170, -0.00001322, -0.00000642, -0.00001098, 0.00000440, -0.00000121, -0.00001626, 0.00000183, -0.00001599, 0.00001671, -0.00001911, 0.00001520, -0.00000249, 0.00000871, -0.00000824, 0.00000022, -0.00000639, 0.00000949, -0.00000052, -0.00000313, -0.00000607, 0.00000943, -0.00000164, -0.00000710, 0.00000097, 0.00000257, -0.00000168, 0.00000069, -0.00000543, 0.00000588, -0.00000899, 0.00000813, -0.00000718, 0.00000774, -0.00000711, 0.00000339, -0.00001143, 0.00001536, -0.00000838, 0.00000729, -0.00000774, 0.00000722, -0.00000735, 0.00000457, -0.00000554, 0.00000572, -0.00000505, 0.00000303, -0.00000711, 0.00000549, -0.00000307, 0.00000675, 0.00000115, -0.00000043, -0.00001007, 0.00000700, -0.00000573, 0.00000502, -0.00000657, 0.00000600, -0.00000627, 0.00000681, -0.00000323, 0.00000006, -0.00000549, 0.00000835, -0.00000552, 0.00000717, -0.00000391, -0.00000030, -0.00000626, 0.00000433, 0.00000127, 0.00000074, -0.00000406, 0.00000140, -0.00000583, 0.00000653, -0.00000584, 0.00000428, 0.00000224, -0.00000495, -0.00000672, 0.00000542, -0.00000595, 0.00000850, -0.00000066, -0.00000042, -0.00000001, 0.00000309, -0.00000664, 0.00000828, -0.00000321, 0.00000013, },
		{ 0.00000000, -0.00001228, 0.00001162, -0.00000026, -0.00000558, -0.00001202, 0.00001301, -0.00001695, 0.00000308, -0.00001250, 0.00000293, -0.00001359, 0.00000492, -0.00001738, -0.00001669, -0.00001644, 0.00001099, 0.00002509, -0.00000080, -0.00000198, 0.00000587, -0.00002443, 0.00000858, -0.00010523, -0.00009108, 0.00021778, 0.00009300, 0.00006306, 0.00013212, 0.00012088, -0.00019852, -0.00000552, -0.00003252, -0.00002155, -0.00002208, 0.00000246, -0.00001685, -0.00000282, -0.00001389, 0.00000741, -0.00000564, -0.00001042, -0.00000381, -0.00001172, 0.00001467, -0.00001855, 0.00001687, -0.00000505, 0.00001007, -0.00000884, 0.00000039, -0.00000682, 0.00001123, -0.00000250, -0.00000270, -0.00000648, 0.00001092, -0.00000297, -0.00000697, 0.00000156, 0.00000193, -0.00000082, -0.00000067, -0.00000389, 0.00000409, -0.00000692, 0.00000693, -0.00000667, 0.00000725, -0.00000654, 0.00000198, -0.00000986, 0.00001509, -0.00000874, 0.00000774, -0.00000828, 0.00000776, -0.00000806, 0.00000517, -0.00000594, 0.00000611, -0.00000539, 0.00000299, -0.00000729, 0.00000621, -0.00000325, 0.00000718, 0.00000086, -0.00000180, -0.00000873, 0.00000666, -0.00000557, 0.00000475, -0.00000633, 0.00000582, -0.00000631, 0.00000752, -0.00000396, -0.00000042, -0.00000455, 0.00000832, -0.00000593, 0.00000800, -0.00000521, 0.00000036, -0.00000655, 0.00000507, 0.00000058, 0.00000083, -0.00000403, 0.00000123, -0.00000562, 0.00000704, -0.00000654, 0.00000532, 0.00000115, -0.00000542, -0.00000582, 0.00000543, -0.00000627, 0.00000962, -0.00000184, -0.00000020, -0.00000000, 0.00000301, -0.00000658, 0.00000932, -0.00000479, },
		{ -0.00000610, -0.00000043, -0.00001162, 0.00001342, -0.00000322, -0.00000493, -0.00001126, 0.00001405, -0.00001917, 0.00000479, -0.00001400, 0.00000544, -0.00001727, 0.00000965, -0.00002356, -0.00001066, -0.00001749, 0.00001360, 0.00002168, 0.00000010, -0.00000413, 0.00000610, -0.00002604, 0.00000936, -0.00010702, -0.00009849, 0.00022148, 0.00010258, 0.00006132, 0.00016372, 0.00008341, -0.00020385, 0.00000629, -0.00003821, -0.00001568, -0.00002623, 0.00000572, -0.00001964, 0.00000008, -0.00001610, 0.00000962, -0.00000934, -0.00000487, -0.00000978, -0.00000657, 0.00001149, -0.00001682, 0.00001755, -0.00000712, 0.00001129, -0.00000940, 0.00000051, -0.00000709, 0.00001286, -0.00000470, -0.00000196, -0.00000706, 0.00001268, -0.00000490, -0.00000609, 0.00000147, 0.00000187, -0.00000053, -0.00000151, -0.00000275, 0.00000257, -0.00000493, 0.00000564, -0.00000609, 0.00000676, -0.00000605, 0.00000055, -0.00000799, 0.00001437, -0.00000875, 0.00000795, -0.00000867, 0.00000817, -0.00000870, 0.00000577, -0.00000637, 0.00000655, -0.00000584, 0.00000303, -0.00000748, 0.00000704, -0.00000364, 0.00000784, 0.00000019, -0.00000284, -0.00000741, 0.00000623, -0.00000537, 0.00000444, -0.00000600, 0.00000552, -0.00000618, 0.00000810, -0.00000477, -0.00000080, -0.00000348, 0.00000800, -0.00000605, 0.00000860, -0.00000645, 0.00000107, -0.00000683, 0.00000582, -0.00000023, 0.00000108, -0.00000412, 0.00000111, -0.00000533, 0.00000740, -0.00000715, 0.00000640, -0.00000021, -0.00000563, -0.00000488, 0.00000525, -0.00000633, 0.00001050, -0.00000299, 0.00000008, 0.00000003, 0.00000276, -0.00000617, 0.00000993, },
		{ 0.00001005, -0.00000707, -0.00000121, -0.00001022, 0.00001443, -0.00000584, -0.00000439, -0.00001003, 0.00001427, -0.00002062, 0.00000581, -0.00001475, 0.00000708, -0.00002002, 0.00001358, -0.00002941, -0.00000406, -0.00001930, 0.00001643, 0.00001784, 0.00000150, -0.00000648, 0.00000631, -0.00002741, 0.00000966, -0.00010802, -0.00010707, 0.00022632, 0.00011141, 0.00006211, 0.00019489, 0.00003835, -0.00019993, 0.00001294, -0.00004097, -0.00001186, -0.00002881, 0.00000753, -0.00002110, 0.00000181, -0.00001717, 0.00001058, -0.00001178, -0.00000022, -0.00001545, -0.00000107, 0.00000752, -0.00001409, 0.00001729, -0.00000866, 0.00001233, -0.00000990, 0.00000057, -0.00000717, 0.00001429, -0.00000700, -0.00000096, -0.00000772, 0.00001459, -0.00000732, -0.00000450, 0.00000070, 0.00000246, -0.00000089, -0.00000170, -0.00000218, 0.00000151, -0.00000320, 0.00000443, -0.00000554, 0.00000637, -0.00000575, -0.00000074, -0.00000595, 0.00001329, -0.00000844, 0.00000792, -0.00000888, 0.00000841, -0.00000921, 0.00000632, -0.00000678, 0.00000701, -0.00000636, 0.00000311, -0.00000762, 0.00000792, -0.00000418, 0.00000871, -0.00000089, -0.00000348, -0.00000619, 0.00000576, -0.00000516, 0.00000413, -0.00000563, 0.00000512, -0.00000588, 0.00000853, -0.00000563, -0.00000104, -0.00000233, 0.00000740, -0.00000582, 0.00000887, -0.00000752, 0.00000173, -0.00000704, 0.00000648, -0.00000108, 0.00000141, -0.00000426, 0.00000100, -0.00000492, 0.00000756, -0.00000758, 0.00000741, -0.00000176, -0.00000558, -0.00000394, 0.00000488, -0.00000612, 0.00001109, -0.00000402, 0.00000038, 0.00000010, 0.00000237, -0.00000542, },
		{ -0.00000436, 0.00000971, -0.00000767, -0.00000235, -0.00000805, 0.00001448, -0.00000791, -0.00000409, -0.00000825, 0.00001357, -0.00002112, 0.00000595, -0.00001451, 0.00000756, -0.00002144, 0.00001617, -0.00003428, 0.00000244, -0.00002144, 0.00001915, 0.00001391, 0.00000312, -0.00000877, 0.00000628, -0.00002828, 0.00000921, -0.00010788, -0.00011728, 0.00023287, 0.00011874, 0.00006688, 0.00022325, -0.00001233, -0.00018714, 0.00001389, -0.00004016, -0.00001063, -0.00002936, 0.00000751, -0.00002087, 0.00000203, -0.00001678, 0.00000995, -0.00001253, 0.00000295, -0.00002020, 0.00000422, 0.00000322, -0.00001067, 0.00001621, -0.00000966, 0.00001315, -0.00001033, 0.00000058, -0.00000704, 0.00001544, -0.00000930, 0.00000021, -0.00000838, 0.00001649, -0.00001009, -0.00000230, -0.00000073, 0.00000369, -0.00000194, -0.00000118, -0.00000228, 0.00000106, -0.00000190, 0.00000343, -0.00000516, 0.00000619, -0.00000578, -0.00000171, -0.00000392, 0.00001197, -0.00000785, 0.00000767, -0.00000891, 0.00000848, -0.00000957, 0.00000679, -0.00000714, 0.00000743, -0.00000691, 0.00000322, -0.00000769, 0.00000877, -0.00000480, 0.00000971, -0.00000233, -0.00000369, -0.00000516, 0.00000530, -0.00000498, 0.00000385, -0.00000526, 0.00000465, -0.00000542, 0.00000880, -0.00000652, -0.00000112, -0.00000116, 0.00000656, -0.00000524, 0.00000877, -0.00000834, 0.00000227, -0.00000709, 0.00000698, -0.00000187, 0.00000176, -0.00000443, 0.00000089, -0.00000438, 0.00000747, -0.00000776, 0.00000825, -0.00000341, -0.00000528, -0.00000303, 0.00000434, -0.00000562, 0.00001134, -0.00000488, 0.00000067, 0.00000020, 0.00000184, },
		{ 0.00000125, -0.00000305, 0.00000893, -0.00000794, -0.00000378, -0.00000517, 0.00001353, -0.00000929, -0.00000409, -0.00000593, 0.00001191, -0.00002057, 0.00000508, -0.00001315, 0.00000667, -0.00002123, 0.00001700, -0.00003756, 0.00000819, -0.00002338, 0.00002138, 0.00001025, 0.00000463, -0.00001075, 0.00000575, -0.00002841, 0.00000771, -0.00010629, -0.00012949, 0.00024162, 0.00012398, 0.00007701, 0.00024617, -0.00006612, -0.00016658, 0.00000918, -0.00003555, -0.00001227, -0.00002763, 0.00000547, -0.00001876, 0.00000054, -0.00001475, 0.00000752, -0.00001129, 0.00000421, -0.00002347, 0.00000871, -0.00000093, -0.00000692, 0.00001451, -0.00001019, 0.00001377, -0.00001070, 0.00000056, -0.00000671, 0.00001627, -0.00001151, 0.00000147, -0.00000894, 0.00001824, -0.00001301, 0.00000037, -0.00000271, 0.00000548, -0.00000363, 0.00000005, -0.00000309, 0.00000130, -0.00000115, 0.00000278, -0.00000502, 0.00000627, -0.00000622, -0.00000224, -0.00000208, 0.00001054, -0.00000708, 0.00000724, -0.00000878, 0.00000839, -0.00000977, 0.00000714, -0.00000741, 0.00000779, -0.00000746, 0.00000333, -0.00000763, 0.00000952, -0.00000541, 0.00001075, -0.00000407, -0.00000344, -0.00000435, 0.00000490, -0.00000486, 0.00000363, -0.00000492, 0.00000417, -0.00000483, 0.00000890, -0.00000743, -0.00000102, -0.00000004, 0.00000555, -0.00000436, 0.00000826, -0.00000885, 0.00000264, -0.00000693, 0.00000724, -0.00000253, 0.00000207, -0.00000456, 0.00000075, -0.00000369, 0.00000709, -0.00000760, 0.00000880, -0.00000507, -0.00000477, -0.00000218, 0.00000368, -0.00000486, 0.00001123, -0.00000552, 0.00000090, 0.00000033, },
		{ 0.00000046, 0.00000064, -0.00000160, 0.00000781, -0.00000796, -0.00000537, -0.00000172, 0.00001161, -0.00000996, -0.00000442, -0.00000313, 0.00000936, -0.00001900, 0.00000320, -0.00001067, 0.00000440, -0.00001929, 0.00001578, -0.00003878, 0.00001260, -0.00002464, 0.00002275, 0.00000722, 0.00000574, -0.00001216, 0.00000451, -0.00002762, 0.00000501, -0.00010304, -0.00014398, 0.00025288, 0.00012675, 0.00009361, 0.00026104, -0.00012005, -0.00013994, -0.00000058, -0.00002735, -0.00001669, -0.00002368, 0.00000143, -0.00001480, -0.00000264, -0.00001108, 0.00000329, -0.00000796, 0.00000329, -0.00002483, 0.00001191, -0.00000446, -0.00000321, 0.00001244, -0.00001035, 0.00001422, -0.00001104, 0.00000054, -0.00000620, 0.00001672, -0.00001353, 0.00000273, -0.00000931, 0.00001965, -0.00001586, 0.00000329, -0.00000507, 0.00000770, -0.00000584, 0.00000189, -0.00000457, 0.00000225, -0.00000101, 0.00000253, -0.00000519, 0.00000666, -0.00000711, -0.00000220, -0.00000057, 0.00000915, -0.00000621, 0.00000668, -0.00000854, 0.00000815, -0.00000982, 0.00000738, -0.00000758, 0.00000804, -0.00000798, 0.00000341, -0.00000742, 0.00001010, -0.00000593, 0.00001174, -0.00000601, -0.00000278, -0.00000381, 0.00000459, -0.00000481, 0.00000349, -0.00000465, 0.00000371, -0.00000413, 0.00000881, -0.00000833, -0.00000070, 0.00000095, 0.00000447, -0.00000322, 0.00000737, -0.00000903, 0.00000279, -0.00000655, 0.00000723, -0.00000300, 0.00000228, -0.00000462, 0.00000056, -0.00000285, 0.00000642, -0.00000706, 0.00000898, -0.00000664, -0.00000409, -0.00000142, 0.00000292, -0.00000387, 0.00001076, -0.00000591, 0.00000107, },
		{ 0.00000116, 0.00000056, 0.00000009, -0.00000012, 0.00000648, -0.00000785, -0.00000694, 0.00000209, 0.00000887, -0.00000994, -0.00000501, -0.00000000, 0.00000610, -0.00001655, 0.00000044, -0.00000720, 0.00000087, -0.00001568, 0.00001247, -0.00003767, 0.00001522, -0.00002479, 0.00002294, 0.00000512, 0.00000617, -0.00001283, 0.00000245, -0.00002581, 0.00000102, -0.00009810, -0.00016079, 0.00026669, 0.00012706, 0.00011739, 0.00026547, -0.00017095, -0.00010943, -0.00001428, -0.00001618, -0.00002350, -0.00001780, -0.00000433, -0.00000922, -0.00000730, -0.00000599, -0.00000254, -0.00000265, 0.00000013, -0.00002403, 0.00001344, -0.00000695, 0.00000007, 0.00001025, -0.00001028, 0.00001455, -0.00001138, 0.00000057, -0.00000555, 0.00001679, -0.00001528, 0.00000393, -0.00000941, 0.00002058, -0.00001841, 0.00000624, -0.00000760, 0.00001016, -0.00000840, 0.00000421, -0.00000661, 0.00000384, -0.00000148, 0.00000272, -0.00000567, 0.00000734, -0.00000846, -0.00000152, 0.00000046, 0.00000794, -0.00000535, 0.00000607, -0.00000822, 0.00000782, -0.00000973, 0.00000749, -0.00000763, 0.00000818, -0.00000844, 0.00000346, -0.00000703, 0.00001046, -0.00000628, 0.00001256, -0.00000804, -0.00000176, -0.00000353, 0.00000439, -0.00000484, 0.00000346, -0.00000447, 0.00000330, -0.00000338, 0.00000854, -0.00000920, -0.00000017, 0.00000174, 0.00000340, -0.00000194, 0.00000616, -0.00000888, 0.00000274, -0.00000594, 0.00000693, -0.00000325, 0.00000235, -0.00000460, 0.00000033, -0.00000188, 0.00000547, -0.00000613, 0.00000871, -0.00000800, -0.00000328, -0.00000075, 0.00000212, -0.00000269, 0.00000995, -0.00000605, },
		{ -0.00000593, 0.00000119, 0.00000062, -0.00000032, 0.00000126, 0.00000506, -0.00000775, -0.00000829, 0.00000597, 0.00000552, -0.00000936, -0.00000570, 0.00000321, 0.00000239, -0.00001346, -0.00000296, -0.00000302, -0.00000362, -0.00001067, 0.00000722, -0.00003419, 0.00001580, -0.00002356, 0.00002175, 0.00000415, 0.00000578, -0.00001267, -0.00000046, -0.00002304, -0.00000416, -0.00009161, -0.00017974, 0.00028280, 0.00012536, 0.00014844, 0.00025752, -0.00021570, -0.00007760, -0.00003041, -0.00000301, -0.00003198, -0.00001058, -0.00001130, -0.00000248, -0.00001303, 0.00000011, -0.00000958, 0.00000429, -0.00000507, -0.00002103, 0.00001307, -0.00000809, 0.00000262, 0.00000819, -0.00001010, 0.00001483, -0.00001175, 0.00000068, -0.00000481, 0.00001647, -0.00001671, 0.00000500, -0.00000918, 0.00002090, -0.00002044, 0.00000897, -0.00001008, 0.00001263, -0.00001109, 0.00000681, -0.00000905, 0.00000596, -0.00000249, 0.00000331, -0.00000642, 0.00000826, -0.00001021, -0.00000021, 0.00000093, 0.00000701, -0.00000461, 0.00000547, -0.00000788, 0.00000742, -0.00000954, 0.00000749, -0.00000756, 0.00000817, -0.00000883, 0.00000348, -0.00000646, 0.00001056, -0.00000638, 0.00001310, -0.00001003, -0.00000046, -0.00000351, 0.00000429, -0.00000496, 0.00000352, -0.00000441, 0.00000299, -0.00000261, 0.00000811, -0.00001002, 0.00000059, 0.00000226, 0.00000247, -0.00000062, 0.00000471, -0.00000845, 0.00000251, -0.00000514, 0.00000636, -0.00000326, 0.00000227, -0.00000449, 0.00000007, -0.00000085, 0.00000430, -0.00000483, 0.00000796, -0.00000908, -0.00000240, -0.00000021, 0.00000134, -0.00000139, 0.00000886, },
		{ 0.00000754, -0.00000560, 0.00000116, 0.00000059, -0.00000052, 0.00000243, 0.00000371, -0.00000781, -0.00000918, 0.00000960, 0.00000187, -0.00000842, -0.00000632, 0.00000624, -0.00000144, -0.00001007, -0.00000663, 0.00000147, -0.00000863, -0.00000472, 0.00000043, -0.00002855, 0.00001433, -0.00002086, 0.00001910, 0.00000437, 0.00000454, -0.00001171, -0.00000411, -0.00001948, -0.00001026, -0.00008393, -0.00020038, 0.00030061, 0.00012261, 0.00018614, 0.00023594, -0.00025145, -0.00004706, -0.00004720, 0.00001092, -0.00004120, -0.00000278, -0.00001879, 0.00000480, -0.00001925, 0.00000667, -0.00001723, 0.00001231, -0.00001191, -0.00001604, 0.00001079, -0.00000773, 0.00000420, 0.00000647, -0.00000994, 0.00001510, -0.00001217, 0.00000091, -0.00000404, 0.00001579, -0.00001778, 0.00000591, -0.00000862, 0.00002052, -0.00002178, 0.00001124, -0.00001226, 0.00001488, -0.00001367, 0.00000944, -0.00001166, 0.00000843, -0.00000392, 0.00000422, -0.00000738, 0.00000933, -0.00001226, 0.00000168, 0.00000081, 0.00000645, -0.00000405, 0.00000495, -0.00000758, 0.00000700, -0.00000926, 0.00000741, -0.00000740, 0.00000804, -0.00000913, 0.00000347, -0.00000574, 0.00001038, -0.00000619, 0.00001328, -0.00001185, 0.00000102, -0.00000369, 0.00000428, -0.00000514, 0.00000368, -0.00000445, 0.00000280, -0.00000185, 0.00000752, -0.00001076, 0.00000156, 0.00000247, 0.00000177, 0.00000061, 0.00000313, -0.00000781, 0.00000216, -0.00000422, 0.00000558, -0.00000308, 0.00000207, -0.00000431, -0.00000016, 0.00000019, 0.00000299, -0.00000323, 0.00000673, -0.00000982, -0.00000148, 0.00000020, 0.00000064, -0.00000006, },
		{ 0.00000121, 0.00000610, -0.00000510, 0.00000110, 0.00000045, -0.00000046, 0.00000328, 0.00000255, -0.00000815, -0.00000942, 0.00001268, -0.00000177, -0.00000735, -0.00000663, 0.00000877, -0.00000500, -0.00000678, -0.00001018, 0.00000581, -0.00001361, 0.00000156, -0.00000729, -0.00002122, 0.00001102, -0.00001678, 0.00001511, 0.00000571, 0.00000253, -0.00001011, -0.00000827, -0.00001545, -0.00001687, -0.00007561, -0.00022205, 0.00031921, 0.00012030, 0.00022906, 0.00020035, -0.00027591, -0.00002029, -0.00006281, 0.00002425, -0.00005009, 0.00000472, -0.00002601, 0.00001190, -0.00002529, 0.00001302, -0.00002481, 0.00002072, -0.00001979, -0.00000946, 0.00000677, -0.00000585, 0.00000469, 0.00000523, -0.00000988, 0.00001539, -0.00001266, 0.00000130, -0.00000329, 0.00001477, -0.00001846, 0.00000666, -0.00000772, 0.00001941, -0.00002229, 0.00001288, -0.00001393, 0.00001667, -0.00001589, 0.00001185, -0.00001419, 0.00001103, -0.00000560, 0.00000533, -0.00000842, 0.00001039, -0.00001445, 0.00000405, 0.00000013, 0.00000630, -0.00000374, 0.00000456, -0.00000734, 0.00000661, -0.00000895, 0.00000728, -0.00000715, 0.00000778, -0.00000936, 0.00000347, -0.00000489, 0.00000995, -0.00000569, 0.00001303, -0.00001337, 0.00000257, -0.00000404, 0.00000434, -0.00000536, 0.00000391, -0.00000461, 0.00000276, -0.00000115, 0.00000679, -0.00001138, 0.00000271, 0.00000233, 0.00000139, 0.00000162, 0.00000155, -0.00000705, 0.00000176, -0.00000324, 0.00000466, -0.00000275, 0.00000179, -0.00000410, -0.00000031, 0.00000115, 0.00000165, -0.00000143, 0.00000508, -0.00001020, -0.00000057, 0.00000044, 0.00000009, },
		{ -0.00000024, 0.00000231, 0.00000462, -0.00000450, 0.00000104, 0.00000018, -0.00000009, 0.00000372, 0.00000168, -0.00000888, -0.00000883, 0.00001490, -0.00000508, -0.00000639, -0.00000643, 0.00001050, -0.00000792, -0.00000397, -0.00001317, 0.00000949, -0.00001796, 0.00000746, -0.00001520, -0.00001286, 0.00000631, -0.00001163, 0.00001003, 0.00000793, -0.00000001, -0.00000815, -0.00001259, -0.00001140, -0.00002343, -0.00006734, -0.00024387, 0.00033741, 0.00012038, 0.00027495, 0.00015137, -0.00028756, 0.00000058, -0.00007553, 0.00003564, -0.00005759, 0.00001100, -0.00003216, 0.00001806, -0.00003044, 0.00001846, -0.00003156, 0.00002872, -0.00002796, -0.00000186, 0.00000138, -0.00000261, 0.00000409, 0.00000454, -0.00000997, 0.00001570, -0.00001319, 0.00000184, -0.00000262, 0.00001348, -0.00001877, 0.00000724, -0.00000656, 0.00001760, -0.00002190, 0.00001374, -0.00001494, 0.00001782, -0.00001753, 0.00001381, -0.00001640, 0.00001352, -0.00000736, 0.00000649, -0.00000943, 0.00001131, -0.00001660, 0.00000673, -0.00000105, 0.00000654, -0.00000369, 0.00000432, -0.00000721, 0.00000627, -0.00000863, 0.00000711, -0.00000686, 0.00000742, -0.00000952, 0.00000351, -0.00000399, 0.00000930, -0.00000489, 0.00001232, -0.00001451, 0.00000408, -0.00000448, 0.00000443, -0.00000561, 0.00000419, -0.00000486, 0.00000286, -0.00000054, 0.00000594, -0.00001185, 0.00000400, 0.00000185, 0.00000140, 0.00000230, 0.00000009, -0.00000627, 0.00000141, -0.00000232, 0.00000369, -0.00000235, 0.00000147, -0.00000392, -0.00000031, 0.00000191, 0.00000041, 0.00000042, 0.00000310, -0.00001022, 0.00000033, 0.00000049, },
		{ 0.00000034, -0.00000031, 0.00000317, 0.00000320, -0.00000385, 0.00000101, -0.00000023, 0.00000058, 0.00000371, 0.00000116, -0.00001003, -0.00000734, 0.00001605, -0.00000777, -0.00000577, -0.00000552, 0.00001119, -0.00000987, -0.00000204, -0.00001519, 0.00001204, -0.00002109, 0.00001229, -0.00002247, -0.00000426, 0.00000082, -0.00000586, 0.00000429, 0.00001068, -0.00000275, -0.00000620, -0.00001664, -0.00000782, -0.00002934, -0.00005991, -0.00026487, 0.00035387, 0.00012515, 0.00032078, 0.00009067, -0.00028579, 0.00001401, -0.00008395, 0.00004397, -0.00006275, 0.00001527, -0.00003650, 0.00002260, -0.00003407, 0.00002233, -0.00003677, 0.00003552, -0.00003561, 0.00000604, -0.00000488, 0.00000170, 0.00000251, 0.00000439, -0.00001021, 0.00001602, -0.00001373, 0.00000254, -0.00000208, 0.00001198, -0.00001874, 0.00000770, -0.00000523, 0.00001518, -0.00002062, 0.00001375, -0.00001518, 0.00001820, -0.00001843, 0.00001511, -0.00001807, 0.00001568, -0.00000899, 0.00000755, -0.00001026, 0.00001193, -0.00001849, 0.00000953, -0.00000260, 0.00000714, -0.00000391, 0.00000424, -0.00000719, 0.00000602, -0.00000833, 0.00000694, -0.00000655, 0.00000699, -0.00000964, 0.00000361, -0.00000308, 0.00000851, -0.00000386, 0.00001115, -0.00001518, 0.00000546, -0.00000497, 0.00000453, -0.00000583, 0.00000448, -0.00000519, 0.00000310, -0.00000005, 0.00000499, -0.00001213, 0.00000535, 0.00000105, 0.00000180, 0.00000257, -0.00000113, -0.00000558, 0.00000117, -0.00000155, 0.00000277, -0.00000196, 0.00000120, -0.00000383, -0.00000010, 0.00000238, -0.00000058, 0.00000217, 0.00000090, -0.00000992, 0.00000120, },
		{ 0.00000204, -0.00000002, -0.00000006, 0.00000370, 0.00000194, -0.00000323, 0.00000103, -0.00000076, 0.00000155, 0.00000325, 0.00000101, -0.00001160, -0.00000491, 0.00001598, -0.00000963, -0.00000566, -0.00000380, 0.00001066, -0.00001060, -0.00000124, -0.00001595, 0.00001309, -0.00002252, 0.00001537, -0.00002831, 0.00000372, -0.00000474, -0.00000003, -0.00000163, 0.00001352, -0.00000530, -0.00000464, -0.00002000, -0.00000523, -0.00003398, -0.00005408, -0.00028409, 0.00036725, 0.00013714, 0.00036293, 0.00002097, -0.00027102, 0.00001918, -0.00008714, 0.00004838, -0.00006486, 0.00001693, -0.00003848, 0.00002501, -0.00003566, 0.00002411, -0.00003983, 0.00004040, -0.00004195, 0.00001350, -0.00001143, 0.00000668, 0.00000016, 0.00000471, -0.00001058, 0.00001631, -0.00001424, 0.00000336, -0.00000170, 0.00001033, -0.00001839, 0.00000809, -0.00000385, 0.00001230, -0.00001855, 0.00001294, -0.00001462, 0.00001774, -0.00001850, 0.00001562, -0.00001902, 0.00001732, -0.00001032, 0.00000838, -0.00001080, 0.00001211, -0.00001994, 0.00001224, -0.00000436, 0.00000800, -0.00000435, 0.00000432, -0.00000728, 0.00000586, -0.00000808, 0.00000680, -0.00000625, 0.00000650, -0.00000971, 0.00000382, -0.00000226, 0.00000765, -0.00000267, 0.00000956, -0.00001535, 0.00000662, -0.00000545, 0.00000460, -0.00000601, 0.00000476, -0.00000555, 0.00000348, 0.00000031, 0.00000395, -0.00001219, 0.00000670, 0.00000001, 0.00000260, 0.00000238, -0.00000202, -0.00000505, 0.00000114, -0.00000101, 0.00000200, -0.00000166, 0.00000103, -0.00000387, 0.00000038, 0.00000246, -0.00000119, 0.00000365, -0.00000136, -0.00000937, },
		{ -0.00000864, 0.00000287, -0.00000059, 0.00000052, 0.00000384, 0.00000092, -0.00000270, 0.00000111, -0.00000140, 0.00000277, 0.00000238, 0.00000118, -0.00001347, -0.00000166, 0.00001466, -0.00001052, -0.00000612, -0.00000124, 0.00000886, -0.00000998, -0.00000175, -0.00001525, 0.00001241, -0.00002190, 0.00001623, -0.00003201, 0.00001028, -0.00000964, 0.00000524, -0.00000719, 0.00001602, -0.00000727, -0.00000387, -0.00002227, -0.00000407, -0.00003683, -0.00005051, -0.00030070, 0.00037636, 0.00015879, 0.00039733, -0.00005414, -0.00024469, 0.00001609, -0.00008473, 0.00004848, -0.00006356, 0.00001563, -0.00003778, 0.00002496, -0.00003493, 0.00002350, -0.00004033, 0.00004279, -0.00004629, 0.00001980, -0.00001763, 0.00001186, -0.00000268, 0.00000538, -0.00001101, 0.00001651, -0.00001467, 0.00000428, -0.00000153, 0.00000862, -0.00001779, 0.00000848, -0.00000258, 0.00000916, -0.00001583, 0.00001140, -0.00001332, 0.00001649, -0.00001771, 0.00001528, -0.00001915, 0.00001829, -0.00001121, 0.00000886, -0.00001098, 0.00001174, -0.00002077, 0.00001464, -0.00000616, 0.00000902, -0.00000497, 0.00000451, -0.00000747, 0.00000579, -0.00000788, 0.00000671, -0.00000600, 0.00000599, -0.00000977, 0.00000417, -0.00000158, 0.00000684, -0.00000143, 0.00000764, -0.00001502, 0.00000752, -0.00000587, 0.00000462, -0.00000613, 0.00000500, -0.00000592, 0.00000398, 0.00000050, 0.00000285, -0.00001200, 0.00000796, -0.00000120, 0.00000375, 0.00000171, -0.00000253, -0.00000474, 0.00000135, -0.00000076, 0.00000145, -0.00000151, 0.00000101, -0.00000408, 0.00000117, 0.00000211, -0.00000132, 0.00000471, -0.00000353, },
		{ -0.00000547, -0.00000781, 0.00000369, -0.00000137, 0.00000145, 0.00000355, 0.00000020, -0.00000229, 0.00000126, -0.00000211, 0.00000415, 0.00000119, 0.00000157, -0.00001551, 0.00000225, 0.00001218, -0.00001044, -0.00000716, 0.00000207, 0.00000587, -0.00000804, -0.00000358, -0.00001309, 0.00000996, -0.00001911, 0.00001455, -0.00003307, 0.00001475, -0.00001321, 0.00000940, -0.00001191, 0.00001777, -0.00000833, -0.00000417, -0.00002315, -0.00000463, -0.00003757, -0.00004961, -0.00031419, 0.00038038, 0.00019230, 0.00041982, -0.00013035, -0.00020921, 0.00000555, -0.00007698, 0.00004435, -0.00005886, 0.00001140, -0.00003441, 0.00002245, -0.00003184, 0.00002043, -0.00003813, 0.00004238, -0.00004812, 0.00002433, -0.00002290, 0.00001679, -0.00000572, 0.00000624, -0.00001143, 0.00001657, -0.00001495, 0.00000525, -0.00000160, 0.00000693, -0.00001701, 0.00000894, -0.00000158, 0.00000600, -0.00001270, 0.00000931, -0.00001143, 0.00001455, -0.00001615, 0.00001412, -0.00001843, 0.00001854, -0.00001160, 0.00000895, -0.00001075, 0.00001078, -0.00002088, 0.00001656, -0.00000784, 0.00001008, -0.00000569, 0.00000480, -0.00000773, 0.00000580, -0.00000774, 0.00000667, -0.00000581, 0.00000548, -0.00000981, 0.00000467, -0.00000112, 0.00000616, -0.00000024, 0.00000550, -0.00001423, 0.00000813, -0.00000620, 0.00000458, -0.00000616, 0.00000517, -0.00000627, 0.00000458, 0.00000052, 0.00000174, -0.00001154, 0.00000903, -0.00000247, 0.00000515, 0.00000061, -0.00000265, -0.00000468, 0.00000182, -0.00000084, 0.00000117, -0.00000157, 0.00000118, -0.00000447, 0.00000225, 0.00000129, -0.00000089, 0.00000521, },
		{ 0.00000508, -0.00000704, -0.00000695, 0.00000450, -0.00000230, 0.00000268, 0.00000283, -0.00000019, -0.00000203, 0.00000147, -0.00000281, 0.00000559, -0.00000021, 0.00000206, -0.00001748, 0.00000653, 0.00000874, -0.00000949, -0.00000864, 0.00000594, 0.00000189, -0.00000495, -0.00000655, -0.00000964, 0.00000590, -0.00001427, 0.00001034, -0.00003127, 0.00001666, -0.00001495, 0.00001202, -0.00001539, 0.00001847, -0.00000827, -0.00000571, -0.00002255, -0.00000700, -0.00003611, -0.00005150, -0.00032442, 0.00037900, 0.00023930, 0.00042640, -0.00020294, -0.00016774, -0.00001086, -0.00006477, 0.00003657, -0.00005122, 0.00000463, -0.00002867, 0.00001776, -0.00002666, 0.00001514, -0.00003337, 0.00003913, -0.00004719, 0.00002667, -0.00002679, 0.00002104, -0.00000866, 0.00000715, -0.00001178, 0.00001646, -0.00001504, 0.00000625, -0.00000192, 0.00000534, -0.00001612, 0.00000955, -0.00000102, 0.00000307, -0.00000941, 0.00000690, -0.00000915, 0.00001210, -0.00001397, 0.00001226, -0.00001692, 0.00001807, -0.00001146, 0.00000865, -0.00001014, 0.00000925, -0.00002019, 0.00001785, -0.00000924, 0.00001107, -0.00000643, 0.00000512, -0.00000803, 0.00000588, -0.00000766, 0.00000668, -0.00000569, 0.00000499, -0.00000985, 0.00000533, -0.00000093, 0.00000570, 0.00000075, 0.00000326, -0.00001305, 0.00000844, -0.00000644, 0.00000447, -0.00000611, 0.00000525, -0.00000656, 0.00000524, 0.00000036, 0.00000063, -0.00001080, 0.00000985, -0.00000368, 0.00000669, -0.00000085, -0.00000238, -0.00000486, 0.00000254, -0.00000125, 0.00000118, -0.00000185, 0.00000154, -0.00000503, 0.00000360, 0.00000004, 0.00000010, },
		{ 0.00000163, 0.00000427, -0.00000815, -0.00000614, 0.00000527, -0.00000335, 0.00000416, 0.00000174, -0.00000025, -0.00000193, 0.00000171, -0.00000344, 0.00000697, -0.00000168, 0.00000248, -0.00001915, 0.00001085, 0.00000465, -0.00000788, -0.00001036, 0.00001005, -0.00000272, -0.00000103, -0.00001035, -0.00000525, 0.00000063, -0.00000775, 0.00000386, -0.00002666, 0.00001586, -0.00001460, 0.00001283, -0.00001743, 0.00001798, -0.00000705, -0.00000842, -0.00002056, -0.00001099, -0.00003271, -0.00005588, -0.00033179, 0.00037266, 0.00030060, 0.00041361, -0.00026709, -0.00012397, -0.00003092, -0.00004953, 0.00002619, -0.00004147, -0.00000399, -0.00002122, 0.00001145, -0.00001992, 0.00000815, -0.00002649, 0.00003332, -0.00004352, 0.00002663, -0.00002897, 0.00002430, -0.00001127, 0.00000799, -0.00001201, 0.00001617, -0.00001493, 0.00000725, -0.00000254, 0.00000396, -0.00001520, 0.00001036, -0.00000104, 0.00000064, -0.00000627, 0.00000443, -0.00000675, 0.00000941, -0.00001140, 0.00000989, -0.00001478, 0.00001696, -0.00001084, 0.00000800, -0.00000924, 0.00000723, -0.00001872, 0.00001846, -0.00001025, 0.00001189, -0.00000714, 0.00000543, -0.00000834, 0.00000601, -0.00000761, 0.00000676, -0.00000567, 0.00000453, -0.00000987, 0.00000615, -0.00000103, 0.00000555, 0.00000143, 0.00000108, -0.00001157, 0.00000848, -0.00000659, 0.00000430, -0.00000598, 0.00000524, -0.00000678, 0.00000594, 0.00000002, -0.00000043, -0.00000979, 0.00001034, -0.00000471, 0.00000824, -0.00000257, -0.00000180, -0.00000523, 0.00000345, -0.00000193, 0.00000144, -0.00000232, 0.00000206, -0.00000569, 0.00000514, -0.00000158, },
		{ -0.00000343, 0.00000358, 0.00000284, -0.00000877, -0.00000539, 0.00000598, -0.00000441, 0.00000579, 0.00000033, -0.00000003, -0.00000195, 0.00000194, -0.00000393, 0.00000817, -0.00000305, 0.00000266, -0.00002024, 0.00001481, 0.00000028, -0.00000588, -0.00001202, 0.00001402, -0.00000753, 0.00000328, -0.00001453, -0.00000042, -0.00000530, -0.00000015, -0.00000434, -0.00001963, 0.00001248, -0.00001214, 0.00001183, -0.00001802, 0.00001637, -0.00000483, -0.00001206, -0.00001757, -0.00001613, -0.00002798, -0.00006199, -0.00033729, 0.00036259, 0.00037596, 0.00037887, -0.00031830, -0.00008186, -0.00005203, -0.00003310, 0.00001460, -0.00003074, -0.00001343, -0.00001295, 0.00000435, -0.00001241, 0.00000023, -0.00001820, 0.00002550, -0.00003744, 0.00002427, -0.00002934, 0.00002638, -0.00001337, 0.00000868, -0.00001213, 0.00001571, -0.00001463, 0.00000824, -0.00000348, 0.00000287, -0.00001434, 0.00001144, -0.00000176, -0.00000106, -0.00000356, 0.00000222, -0.00000451, 0.00000677, -0.00000874, 0.00000728, -0.00001222, 0.00001537, -0.00000987, 0.00000712, -0.00000817, 0.00000487, -0.00001658, 0.00001835, -0.00001081, 0.00001248, -0.00000776, 0.00000571, -0.00000864, 0.00000617, -0.00000761, 0.00000688, -0.00000575, 0.00000412, -0.00000986, 0.00000709, -0.00000144, 0.00000574, 0.00000171, -0.00000091, -0.00000991, 0.00000830, -0.00000665, 0.00000410, -0.00000579, 0.00000515, -0.00000690, 0.00000664, -0.00000051, -0.00000140, -0.00000854, 0.00001046, -0.00000546, 0.00000964, -0.00000440, -0.00000100, -0.00000572, 0.00000448, -0.00000281, 0.00000189, -0.00000292, 0.00000266, -0.00000638, 0.00000675, },
		{ 0.00000829, -0.00000534, 0.00000580, 0.00000086, -0.00000891, -0.00000471, 0.00000656, -0.00000540, 0.00000743, -0.00000127, 0.00000040, -0.00000205, 0.00000210, -0.00000418, 0.00000904, -0.00000419, 0.00000243, -0.00002050, 0.00001804, -0.00000395, -0.00000386, -0.00001328, 0.00001741, -0.00001204, 0.00000745, -0.00001854, 0.00000422, -0.00001116, 0.00000774, -0.00001345, -0.00001088, 0.00000698, -0.00000786, 0.00000924, -0.00001739, 0.00001393, -0.00000199, -0.00001615, -0.00001417, -0.00002165, -0.00002287, -0.00006867, -0.00034249, 0.00035092, 0.00046395, 0.00032075, -0.00035269, -0.00004524, -0.00007137, -0.00001754, 0.00000343, -0.00002041, -0.00002251, -0.00000493, -0.00000256, -0.00000508, -0.00000768, -0.00000944, 0.00001651, -0.00002957, 0.00001993, -0.00002800, 0.00002726, -0.00001493, 0.00000922, -0.00001219, 0.00001517, -0.00001419, 0.00000924, -0.00000479, 0.00000217, -0.00001361, 0.00001281, -0.00000326, -0.00000183, -0.00000155, 0.00000052, -0.00000274, 0.00000448, -0.00000628, 0.00000472, -0.00000950, 0.00001352, -0.00000871, 0.00000617, -0.00000713, 0.00000239, -0.00001391, 0.00001761, -0.00001090, 0.00001278, -0.00000827, 0.00000594, -0.00000892, 0.00000635, -0.00000763, 0.00000706, -0.00000592, 0.00000377, -0.00000981, 0.00000812, -0.00000211, 0.00000629, 0.00000150, -0.00000257, -0.00000819, 0.00000795, -0.00000666, 0.00000389, -0.00000555, 0.00000498, -0.00000692, 0.00000732, -0.00000121, -0.00000223, -0.00000708, 0.00001019, -0.00000583, 0.00001077, -0.00000620, -0.00000009, -0.00000623, 0.00000551, -0.00000379, 0.00000246, -0.00000358, 0.00000328, -0.00000699, },
		{ -0.00000739, 0.00000958, -0.00000712, 0.00000808, -0.00000150, -0.00000863, -0.00000408, 0.00000696, -0.00000619, 0.00000896, -0.00000293, 0.00000095, -0.00000215, 0.00000212, -0.00000411, 0.00000948, -0.00000494, 0.00000166, -0.00001972, 0.00002015, -0.00000761, -0.00000217, -0.00001381, 0.00001981, -0.00001572, 0.00001093, -0.00002178, 0.00000801, -0.00001618, 0.00001502, -0.00002248, -0.00000134, 0.00000011, -0.00000233, 0.00000555, -0.00001599, 0.00001113, 0.00000089, -0.00002000, -0.00001118, -0.00002657, -0.00001861, -0.00007437, -0.00034949, 0.00034063, 0.00056186, 0.00023924, -0.00036736, -0.00001756, -0.00008624, -0.00000496, -0.00000562, -0.00001189, -0.00002996, 0.00000169, -0.00000820, 0.00000103, -0.00001452, -0.00000129, 0.00000736, -0.00002072, 0.00001415, -0.00002527, 0.00002706, -0.00001598, 0.00000967, -0.00001229, 0.00001465, -0.00001368, 0.00001028, -0.00000650, 0.00000194, -0.00001308, 0.00001448, -0.00000557, -0.00000155, -0.00000044, -0.00000041, -0.00000168, 0.00000284, -0.00000434, 0.00000254, -0.00000693, 0.00001164, -0.00000756, 0.00000532, -0.00000631, 0.00000002, -0.00001095, 0.00001634, -0.00001058, 0.00001282, -0.00000865, 0.00000612, -0.00000916, 0.00000654, -0.00000769, 0.00000727, -0.00000620, 0.00000348, -0.00000970, 0.00000917, -0.00000301, 0.00000718, 0.00000079, -0.00000379, -0.00000652, 0.00000749, -0.00000665, 0.00000368, -0.00000529, 0.00000474, -0.00000681, 0.00000794, -0.00000207, -0.00000288, -0.00000548, 0.00000953, -0.00000575, 0.00001150, -0.00000783, 0.00000081, -0.00000666, 0.00000643, -0.00000474, 0.00000301, -0.00000419, 0.00000379, },
		{ 0.00000410, -0.00000746, 0.00001047, -0.00000856, 0.00001018, -0.00000405, -0.00000802, -0.00000345, 0.00000709, -0.00000666, 0.00001019, -0.00000447, 0.00000148, -0.00000216, 0.00000194, -0.00000367, 0.00000939, -0.00000521, 0.00000026, -0.00001774, 0.00002086, -0.00001032, -0.00000114, -0.00001328, 0.00002080, -0.00001809, 0.00001320, -0.00002369, 0.00001032, -0.00001959, 0.00002075, -0.00003038, 0.00000791, -0.00000720, 0.00000369, 0.00000142, -0.00001442, 0.00000860, 0.00000313, -0.00002280, -0.00000957, -0.00002972, -0.00001665, -0.00007728, -0.00036079, 0.00033544, 0.00066568, 0.00013585, -0.00036068, -0.00000153, -0.00009429, 0.00000275, -0.00001098, -0.00000652, -0.00003460, 0.00000581, -0.00001155, 0.00000493, -0.00001925, 0.00000517, -0.00000085, -0.00001188, 0.00000769, -0.00002162, 0.00002607, -0.00001668, 0.00001016, -0.00001256, 0.00001431, -0.00001323, 0.00001142, -0.00000864, 0.00000227, -0.00001281, 0.00001643, -0.00000866, -0.00000014, -0.00000040, -0.00000038, -0.00000155, 0.00000207, -0.00000318, 0.00000101, -0.00000478, 0.00001000, -0.00000661, 0.00000476, -0.00000592, -0.00000196, -0.00000794, 0.00001474, -0.00000993, 0.00001263, -0.00000893, 0.00000626, -0.00000939, 0.00000677, -0.00000777, 0.00000752, -0.00000657, 0.00000325, -0.00000950, 0.00001019, -0.00000405, 0.00000835, -0.00000044, -0.00000449, -0.00000501, 0.00000699, -0.00000664, 0.00000352, -0.00000505, 0.00000446, -0.00000659, 0.00000847, -0.00000308, -0.00000330, -0.00000381, 0.00000853, -0.00000522, 0.00001174, -0.00000915, 0.00000158, -0.00000690, 0.00000711, -0.00000553, 0.00000345, -0.00000464, },
		{ -0.00000483, 0.00000411, -0.00000709, 0.00001081, -0.00000945, 0.00001187, -0.00000656, -0.00000723, -0.00000273, 0.00000687, -0.00000669, 0.00001099, -0.00000575, 0.00000186, -0.00000200, 0.00000151, -0.00000279, 0.00000871, -0.00000494, -0.00000179, -0.00001452, 0.00001996, -0.00001179, -0.00000106, -0.00001146, 0.00002010, -0.00001875, 0.00001382, -0.00002381, 0.00001058, -0.00002070, 0.00002407, -0.00003610, 0.00001572, -0.00001388, 0.00000929, -0.00000235, -0.00001344, 0.00000708, 0.00000392, -0.00002367, -0.00001035, -0.00002989, -0.00001848, -0.00007547, -0.00037911, 0.00033963, 0.00077031, 0.00001366, -0.00033241, 0.00000110, -0.00009378, 0.00000409, -0.00001137, -0.00000541, -0.00003542, 0.00000652, -0.00001175, 0.00000576, -0.00002096, 0.00000892, -0.00000704, -0.00000408, 0.00000140, -0.00001767, 0.00002467, -0.00001726, 0.00001086, -0.00001318, 0.00001430, -0.00001297, 0.00001272, -0.00001126, 0.00000321, -0.00001285, 0.00001862, -0.00001245, 0.00000236, -0.00000146, 0.00000069, -0.00000249, 0.00000235, -0.00000301, 0.00000037, -0.00000332, 0.00000881, -0.00000607, 0.00000464, -0.00000616, -0.00000331, -0.00000518, 0.00001302, -0.00000908, 0.00001228, -0.00000915, 0.00000640, -0.00000963, 0.00000703, -0.00000790, 0.00000780, -0.00000704, 0.00000311, -0.00000921, 0.00001111, -0.00000516, 0.00000972, -0.00000214, -0.00000465, -0.00000375, 0.00000650, -0.00000666, 0.00000341, -0.00000484, 0.00000416, -0.00000625, 0.00000888, -0.00000419, -0.00000347, -0.00000215, 0.00000724, -0.00000423, 0.00001142, -0.00001005, 0.00000211, -0.00000687, 0.00000746, -0.00000606, 0.00000365, },
		{ 0.00000353, -0.00000468, 0.00000374, -0.00000620, 0.00001047, -0.00000961, 0.00001291, -0.00000879, -0.00000642, -0.00000187, 0.00000624, -0.00000617, 0.00001122, -0.00000660, 0.00000196, -0.00000157, 0.00000076, -0.00000146, 0.00000742, -0.00000411, -0.00000446, -0.00001008, 0.00001737, -0.00001180, -0.00000208, -0.00000823, 0.00001752, -0.00001746, 0.00001250, -0.00002181, 0.00000843, -0.00001902, 0.00002429, -0.00003873, 0.00002103, -0.00001888, 0.00001353, -0.00000491, -0.00001379, 0.00000730, 0.00000252, -0.00002179, -0.00001445, -0.00002597, -0.00002550, -0.00006712, -0.00040715, 0.00035775, 0.00086971, -0.00012278, -0.00028384, -0.00001031, -0.00008371, -0.00000187, -0.00000596, -0.00000930, -0.00003175, 0.00000318, -0.00000819, 0.00000291, -0.00001898, 0.00000915, -0.00001026, 0.00000170, -0.00000388, -0.00001407, 0.00002331, -0.00001801, 0.00001198, -0.00001432, 0.00001481, -0.00001304, 0.00001424, -0.00001436, 0.00000481, -0.00001324, 0.00002096, -0.00001680, 0.00000587, -0.00000362, 0.00000281, -0.00000452, 0.00000377, -0.00000396, 0.00000080, -0.00000273, 0.00000826, -0.00000607, 0.00000510, -0.00000716, -0.00000382, -0.00000291, 0.00001139, -0.00000820, 0.00001187, -0.00000937, 0.00000659, -0.00000991, 0.00000736, -0.00000808, 0.00000812, -0.00000760, 0.00000305, -0.00000881, 0.00001187, -0.00000624, 0.00001118, -0.00000422, -0.00000425, -0.00000278, 0.00000607, -0.00000672, 0.00000337, -0.00000468, 0.00000386, -0.00000581, 0.00000914, -0.00000538, -0.00000336, -0.00000060, 0.00000576, -0.00000284, 0.00001054, -0.00001047, 0.00000233, -0.00000649, 0.00000740, -0.00000622, },
		{ -0.00000597, 0.00000303, -0.00000413, 0.00000295, -0.00000475, 0.00000940, -0.00000894, 0.00001311, -0.00001054, -0.00000572, -0.00000078, 0.00000513, -0.00000505, 0.00001075, -0.00000689, 0.00000167, -0.00000081, -0.00000034, 0.00000032, 0.00000555, -0.00000277, -0.00000763, -0.00000458, 0.00001313, -0.00001031, -0.00000426, -0.00000360, 0.00001307, -0.00001413, 0.00000910, -0.00001755, 0.00000369, -0.00001428, 0.00002097, -0.00003759, 0.00002295, -0.00002126, 0.00001555, -0.00000549, -0.00001616, 0.00000988, -0.00000172, -0.00001648, -0.00002263, -0.00001705, -0.00003886, -0.00005073, -0.00044729, 0.00039427, 0.00095728, -0.00026767, -0.00021769, -0.00003516, -0.00006401, -0.00001540, 0.00000554, -0.00001846, -0.00002332, -0.00000447, -0.00000062, -0.00000388, -0.00001295, 0.00000535, -0.00000982, 0.00000465, -0.00000736, -0.00001147, 0.00002243, -0.00001920, 0.00001370, -0.00001616, 0.00001599, -0.00001357, 0.00001603, -0.00001794, 0.00000708, -0.00001397, 0.00002337, -0.00002150, 0.00001021, -0.00000673, 0.00000590, -0.00000758, 0.00000628, -0.00000603, 0.00000236, -0.00000313, 0.00000847, -0.00000671, 0.00000620, -0.00000899, -0.00000332, -0.00000136, 0.00001008, -0.00000742, 0.00001150, -0.00000966, 0.00000687, -0.00001027, 0.00000778, -0.00000833, 0.00000848, -0.00000826, 0.00000309, -0.00000832, 0.00001242, -0.00000719, 0.00001262, -0.00000656, -0.00000334, -0.00000214, 0.00000573, -0.00000684, 0.00000341, -0.00000458, 0.00000358, -0.00000526, 0.00000922, -0.00000659, -0.00000295, 0.00000077, 0.00000419, -0.00000115, 0.00000914, -0.00001039, 0.00000221, -0.00000575, 0.00000688, },
		{ 0.00000592, -0.00000528, 0.00000213, -0.00000319, 0.00000176, -0.00000278, 0.00000760, -0.00000738, 0.00001235, -0.00001161, -0.00000527, 0.00000058, 0.00000354, -0.00000328, 0.00000954, -0.00000653, 0.00000091, 0.00000034, -0.00000177, 0.00000250, 0.00000314, -0.00000101, -0.00001111, 0.00000172, 0.00000744, -0.00000738, -0.00000751, 0.00000228, 0.00000689, -0.00000888, 0.00000374, -0.00001115, -0.00000352, -0.00000652, 0.00001402, -0.00003236, 0.00002091, -0.00002031, 0.00001467, -0.00000348, -0.00002106, 0.00001528, -0.00000920, -0.00000734, -0.00003534, -0.00000262, -0.00005925, -0.00002528, -0.00050134, 0.00045324, 0.00102627, -0.00041427, -0.00013799, -0.00007170, -0.00003553, -0.00003600, 0.00002281, -0.00003263, -0.00001035, -0.00001626, 0.00001081, -0.00001451, -0.00000290, -0.00000264, -0.00000533, 0.00000421, -0.00000842, -0.00001041, 0.00002244, -0.00002108, 0.00001617, -0.00001879, 0.00001794, -0.00001465, 0.00001811, -0.00002196, 0.00001000, -0.00001504, 0.00002573, -0.00002635, 0.00001512, -0.00001061, 0.00000978, -0.00001152, 0.00000977, -0.00000915, 0.00000502, -0.00000454, 0.00000948, -0.00000801, 0.00000792, -0.00001165, -0.00000174, -0.00000069, 0.00000925, -0.00000690, 0.00001126, -0.00001006, 0.00000727, -0.00001073, 0.00000831, -0.00000866, 0.00000888, -0.00000901, 0.00000326, -0.00000774, 0.00001273, -0.00000791, 0.00001388, -0.00000904, -0.00000199, -0.00000182, 0.00000548, -0.00000700, 0.00000352, -0.00000454, 0.00000332, -0.00000462, 0.00000909, -0.00000778, -0.00000226, 0.00000188, 0.00000266, 0.00000074, 0.00000728, -0.00000982, 0.00000174, -0.00000464, },
		{ -0.00000323, 0.00000456, -0.00000420, 0.00000088, -0.00000189, 0.00000023, -0.00000036, 0.00000515, -0.00000497, 0.00001059, -0.00001191, -0.00000517, 0.00000222, 0.00000149, -0.00000090, 0.00000759, -0.00000548, -0.00000036, 0.00000188, -0.00000352, 0.00000500, 0.00000031, 0.00000104, -0.00001467, 0.00000848, 0.00000061, -0.00000324, -0.00001162, 0.00000909, -0.00000068, -0.00000201, -0.00000330, -0.00000293, -0.00001286, 0.00000389, 0.00000368, -0.00002308, 0.00001467, -0.00001565, 0.00001048, 0.00000147, -0.00002877, 0.00002367, -0.00002005, 0.00000572, -0.00005262, 0.00001736, -0.00008677, 0.00000947, -0.00057020, 0.00053786, 0.00107022, -0.00055530, -0.00004980, -0.00011716, 0.00000003, -0.00006248, 0.00004492, -0.00005103, 0.00000649, -0.00003157, 0.00002554, -0.00002844, 0.00001072, -0.00001454, 0.00000319, 0.00000013, -0.00000668, -0.00001126, 0.00002363, -0.00002382, 0.00001946, -0.00002226, 0.00002072, -0.00001632, 0.00002046, -0.00002633, 0.00001351, -0.00001641, 0.00002791, -0.00003107, 0.00002032, -0.00001497, 0.00001419, -0.00001609, 0.00001400, -0.00001312, 0.00000864, -0.00000688, 0.00001123, -0.00000991, 0.00001017, -0.00001504, 0.00000088, -0.00000099, 0.00000903, -0.00000674, 0.00001122, -0.00001063, 0.00000784, -0.00001131, 0.00000895, -0.00000909, 0.00000932, -0.00000984, 0.00000355, -0.00000709, 0.00001278, -0.00000835, 0.00001487, -0.00001149, -0.00000030, -0.00000181, 0.00000532, -0.00000720, 0.00000368, -0.00000457, 0.00000311, -0.00000391, 0.00000874, -0.00000888, -0.00000132, 0.00000268, 0.00000128, 0.00000268, 0.00000509, -0.00000883, 0.00000097, },
		{ -0.00000001, -0.00000161, 0.00000290, -0.00000281, -0.00000065, -0.00000032, -0.00000153, 0.00000237, 0.00000219, -0.00000182, 0.00000785, -0.00001138, -0.00000544, 0.00000411, -0.00000093, 0.00000200, 0.00000495, -0.00000378, -0.00000213, 0.00000378, -0.00000551, 0.00000772, -0.00000280, 0.00000320, -0.00001803, 0.00001530, -0.00000694, 0.00000177, -0.00001624, 0.00001641, -0.00000914, 0.00000600, -0.00001151, 0.00000657, -0.00002370, 0.00001632, -0.00000944, -0.00001020, 0.00000444, -0.00000730, 0.00000294, 0.00000938, -0.00003921, 0.00003489, -0.00003402, 0.00002234, -0.00007402, 0.00004238, -0.00012085, 0.00005294, -0.00065368, 0.00065013, 0.00108342, -0.00068343, 0.00004116, -0.00016790, 0.00004021, -0.00009298, 0.00007039, -0.00007241, 0.00002608, -0.00004941, 0.00004263, -0.00004478, 0.00002709, -0.00002969, 0.00001533, -0.00000747, -0.00000207, -0.00001419, 0.00002614, -0.00002745, 0.00002356, -0.00002651, 0.00002428, -0.00001856, 0.00002302, -0.00003093, 0.00001750, -0.00001800, 0.00002977, -0.00003541, 0.00002548, -0.00001949, 0.00001880, -0.00002095, 0.00001868, -0.00001766, 0.00001298, -0.00000997, 0.00001360, -0.00001229, 0.00001281, -0.00001899, 0.00000444, -0.00000224, 0.00000949, -0.00000700, 0.00001142, -0.00001138, 0.00000857, -0.00001202, 0.00000972, -0.00000962, 0.00000979, -0.00001074, 0.00000400, -0.00000642, 0.00001257, -0.00000846, 0.00001546, -0.00001376, 0.00000159, -0.00000206, 0.00000524, -0.00000741, 0.00000387, -0.00000465, 0.00000296, -0.00000315, 0.00000815, -0.00000984, -0.00000015, 0.00000312, 0.00000016, 0.00000453, 0.00000271, -0.00000752, },
		{ -0.00000602, -0.00000111, 0.00000010, 0.00000105, -0.00000122, -0.00000235, 0.00000139, -0.00000338, 0.00000522, -0.00000107, 0.00000186, 0.00000429, -0.00001006, -0.00000608, 0.00000618, -0.00000361, 0.00000527, 0.00000175, -0.00000150, -0.00000432, 0.00000597, -0.00000764, 0.00001051, -0.00000602, 0.00000530, -0.00002090, 0.00002172, -0.00001472, 0.00000724, -0.00002096, 0.00002371, -0.00001792, 0.00001453, -0.00002026, 0.00001662, -0.00003523, 0.00002984, -0.00002440, 0.00000544, -0.00000914, 0.00000434, -0.00000762, 0.00001994, -0.00005200, 0.00004844, -0.00005045, 0.00004174, -0.00009861, 0.00007129, -0.00016015, 0.00010356, -0.00075032, 0.00079055, 0.00106142, -0.00079178, 0.00012890, -0.00021981, 0.00008203, -0.00012520, 0.00009733, -0.00009513, 0.00004697, -0.00006844, 0.00006083, -0.00006230, 0.00004503, -0.00004702, 0.00003029, -0.00001810, 0.00000518, -0.00001912, 0.00002992, -0.00003191, 0.00002832, -0.00003140, 0.00002850, -0.00002128, 0.00002568, -0.00003558, 0.00002181, -0.00001974, 0.00003119, -0.00003911, 0.00003027, -0.00002383, 0.00002327, -0.00002576, 0.00002342, -0.00002242, 0.00001773, -0.00001356, 0.00001638, -0.00001494, 0.00001559, -0.00002324, 0.00000873, -0.00000435, 0.00001059, -0.00000770, 0.00001187, -0.00001229, 0.00000945, -0.00001284, 0.00001059, -0.00001023, 0.00001026, -0.00001169, 0.00000459, -0.00000577, 0.00001213, -0.00000822, 0.00001559, -0.00001572, 0.00000357, -0.00000249, 0.00000521, -0.00000761, 0.00000406, -0.00000476, 0.00000285, -0.00000235, 0.00000732, -0.00001059, 0.00000118, 0.00000319, -0.00000061, 0.00000613, 0.00000030, },
		{ -0.00000197, -0.00000446, -0.00000219, 0.00000175, -0.00000082, 0.00000041, -0.00000406, 0.00000310, -0.00000514, 0.00000797, -0.00000436, 0.00000581, 0.00000010, -0.00000808, -0.00000701, 0.00000829, -0.00000635, 0.00000872, -0.00000180, 0.00000118, -0.00000682, 0.00000833, -0.00000979, 0.00001320, -0.00000917, 0.00000712, -0.00002300, 0.00002730, -0.00002220, 0.00001267, -0.00002528, 0.00003043, -0.00002632, 0.00002287, -0.00002879, 0.00002639, -0.00004647, 0.00004334, -0.00003999, 0.00002264, -0.00002508, 0.00001846, -0.00002051, 0.00003245, -0.00006638, 0.00006344, -0.00006835, 0.00006273, -0.00012497, 0.00010241, -0.00020263, 0.00015882, -0.00085729, 0.00095789, 0.00100137, -0.00087436, 0.00020760, -0.00026854, 0.00012222, -0.00015653, 0.00012357, -0.00011729, 0.00006746, -0.00008711, 0.00007866, -0.00007958, 0.00006312, -0.00006522, 0.00004694, -0.00003093, 0.00001457, -0.00002574, 0.00003474, -0.00003696, 0.00003350, -0.00003667, 0.00003316, -0.00002431, 0.00002825, -0.00004006, 0.00002626, -0.00002151, 0.00003204, -0.00004195, 0.00003438, -0.00002765, 0.00002725, -0.00003014, 0.00002786, -0.00002700, 0.00002253, -0.00001734, 0.00001931, -0.00001763, 0.00001826, -0.00002746, 0.00001344, -0.00000713, 0.00001224, -0.00000878, 0.00001253, -0.00001332, 0.00001043, -0.00001372, 0.00001153, -0.00001089, 0.00001073, -0.00001266, 0.00000534, -0.00000517, 0.00001151, -0.00000765, 0.00001520, -0.00001724, 0.00000551, -0.00000305, 0.00000521, -0.00000777, 0.00000424, -0.00000488, 0.00000281, -0.00000155, 0.00000626, -0.00001109, 0.00000260, 0.00000290, -0.00000097, 0.00000735, },
		{ 0.00000810, -0.00000396, -0.00000301, -0.00000312, 0.00000321, -0.00000256, 0.00000193, -0.00000562, 0.00000462, -0.00000661, 0.00001037, -0.00000739, 0.00000970, -0.00000441, -0.00000561, -0.00000809, 0.00001030, -0.00000895, 0.00001209, -0.00000547, 0.00000407, -0.00000943, 0.00001069, -0.00001179, 0.00001562, -0.00001206, 0.00000850, -0.00002408, 0.00003162, -0.00002885, 0.00001758, -0.00002874, 0.00003596, -0.00003367, 0.00003030, -0.00003632, 0.00003500, -0.00005642, 0.00005564, -0.00005485, 0.00004000, -0.00004205, 0.00003390, -0.00003469, 0.00004593, -0.00008130, 0.00007873, -0.00008639, 0.00008378, -0.00015129, 0.00013353, -0.00024561, 0.00021530, -0.00097044, 0.00114908, 0.00090233, -0.00092653, 0.00027206, -0.00030995, 0.00015754, -0.00018432, 0.00014688, -0.00013694, 0.00008580, -0.00010380, 0.00009459, -0.00009507, 0.00007982, -0.00008275, 0.00006390, -0.00004488, 0.00002528, -0.00003348, 0.00004021, -0.00004224, 0.00003873, -0.00004197, 0.00003795, -0.00002743, 0.00003053, -0.00004412, 0.00003062, -0.00002319, 0.00003224, -0.00004373, 0.00003752, -0.00003064, 0.00003041, -0.00003373, 0.00003161, -0.00003103, 0.00002699, -0.00002098, 0.00002210, -0.00002010, 0.00002052, -0.00003132, 0.00001825, -0.00001032, 0.00001429, -0.00001014, 0.00001333, -0.00001440, 0.00001144, -0.00001460, 0.00001247, -0.00001156, 0.00001114, -0.00001360, 0.00000624, -0.00000468, 0.00001078, -0.00000679, 0.00001429, -0.00001824, 0.00000728, -0.00000366, 0.00000519, -0.00000787, 0.00000436, -0.00000500, 0.00000283, -0.00000079, 0.00000499, -0.00001129, 0.00000404, 0.00000228, -0.00000089, },
		{ -0.00000036, 0.00000829, -0.00000552, -0.00000178, -0.00000377, 0.00000433, -0.00000400, 0.00000317, -0.00000687, 0.00000580, -0.00000760, 0.00001218, -0.00000986, 0.00001319, -0.00000892, -0.00000291, -0.00000914, 0.00001200, -0.00001118, 0.00001512, -0.00000897, 0.00000692, -0.00001196, 0.00001288, -0.00001346, 0.00001757, -0.00001449, 0.00000928, -0.00002393, 0.00003433, -0.00003422, 0.00002150, -0.00003088, 0.00003981, -0.00003935, 0.00003611, -0.00004212, 0.00004164, -0.00006408, 0.00006553, -0.00006759, 0.00005597, -0.00005854, 0.00004930, -0.00004890, 0.00005915, -0.00009551, 0.00009294, -0.00010301, 0.00010310, -0.00017547, 0.00016212, -0.00028595, 0.00026889, -0.00108443, 0.00135923, 0.00076545, -0.00094533, 0.00031805, -0.00034040, 0.00018499, -0.00020608, 0.00016515, -0.00015223, 0.00010029, -0.00011696, 0.00010715, -0.00010732, 0.00009360, -0.00009805, 0.00007969, -0.00005868, 0.00003634, -0.00004162, 0.00004579, -0.00004729, 0.00004359, -0.00004688, 0.00004251, -0.00003034, 0.00003227, -0.00004749, 0.00003464, -0.00002465, 0.00003170, -0.00004431, 0.00003949, -0.00003257, 0.00003247, -0.00003624, 0.00003433, -0.00003413, 0.00003075, -0.00002413, 0.00002447, -0.00002207, 0.00002211, -0.00003447, 0.00002277, -0.00001362, 0.00001651, -0.00001165, 0.00001415, -0.00001543, 0.00001238, -0.00001541, 0.00001335, -0.00001219, 0.00001144, -0.00001445, 0.00000724, -0.00000432, 0.00000999, -0.00000570, 0.00001289, -0.00001866, 0.00000880, -0.00000426, 0.00000514, -0.00000788, 0.00000443, -0.00000510, 0.00000292, -0.00000009, 0.00000355, -0.00001117, 0.00000542, 0.00000141, },
		{ 0.00000034, 0.00000057, 0.00000788, -0.00000657, -0.00000089, -0.00000406, 0.00000500, -0.00000502, 0.00000400, -0.00000766, 0.00000648, -0.00000796, 0.00001321, -0.00001151, 0.00001594, -0.00001308, -0.00000024, -0.00000997, 0.00001321, -0.00001279, 0.00001751, -0.00001200, 0.00000944, -0.00001416, 0.00001468, -0.00001463, 0.00001887, -0.00001631, 0.00000933, -0.00002245, 0.00003516, -0.00003790, 0.00002404, -0.00003135, 0.00004153, -0.00004282, 0.00003972, -0.00004555, 0.00004559, -0.00006863, 0.00007197, -0.00007691, 0.00006905, -0.00007300, 0.00006316, -0.00006174, 0.00007077, -0.00010762, 0.00010457, -0.00011658, 0.00011882, -0.00019531, 0.00018551, -0.00032024, 0.00031505, -0.00119303, 0.00158180, 0.00059401, -0.00092968, 0.00034264, -0.00035706, 0.00020217, -0.00021976, 0.00017659, -0.00016157, 0.00010951, -0.00012525, 0.00011505, -0.00011503, 0.00010309, -0.00010967, 0.00009283, -0.00007101, 0.00004668, -0.00004936, 0.00005084, -0.00005160, 0.00004759, -0.00005096, 0.00004643, -0.00003274, 0.00003322, -0.00004989, 0.00003807, -0.00002575, 0.00003037, -0.00004362, 0.00004014, -0.00003324, 0.00003324, -0.00003742, 0.00003575, -0.00003601, 0.00003348, -0.00002650, 0.00002615, -0.00002334, 0.00002277, -0.00003660, 0.00002665, -0.00001671, 0.00001869, -0.00001313, 0.00001488, -0.00001629, 0.00001317, -0.00001604, 0.00001409, -0.00001272, 0.00001159, -0.00001517, 0.00000832, -0.00000411, 0.00000923, -0.00000449, 0.00001107, -0.00001848, 0.00001000, -0.00000481, 0.00000504, -0.00000781, 0.00000443, -0.00000519, 0.00000309, 0.00000048, 0.00000201, -0.00001072, 0.00000667, },
		{ 0.00000773, -0.00000083, 0.00000186, 0.00000689, -0.00000704, -0.00000042, -0.00000389, 0.00000512, -0.00000549, 0.00000429, -0.00000788, 0.00000655, -0.00000756, 0.00001329, -0.00001210, 0.00001764, -0.00001653, 0.00000210, -0.00001037, 0.00001372, -0.00001357, 0.00001902, -0.00001428, 0.00001136, -0.00001580, 0.00001589, -0.00001512, 0.00001938, -0.00001737, 0.00000861, -0.00001960, 0.00003399, -0.00003963, 0.00002493, -0.00002991, 0.00004085, -0.00004372, 0.00004072, -0.00004617, 0.00004636, -0.00006943, 0.00007413, -0.00008170, 0.00007788, -0.00008392, 0.00007403, -0.00007182, 0.00007944, -0.00011629, 0.00011221, -0.00012551, 0.00012914, -0.00020868, 0.00020105, -0.00034513, 0.00034907, -0.00128938, 0.00180891, 0.00039328, -0.00088049, 0.00034439, -0.00035813, 0.00020742, -0.00022387, 0.00017992, -0.00016381, 0.00011240, -0.00012771, 0.00011738, -0.00011726, 0.00010722, -0.00011638, 0.00010202, -0.00008060, 0.00005521, -0.00005582, 0.00005472, -0.00005464, 0.00005026, -0.00005375, 0.00004933, -0.00003432, 0.00003313, -0.00005106, 0.00004066, -0.00002637, 0.00002827, -0.00004166, 0.00003941, -0.00003258, 0.00003262, -0.00003714, 0.00003572, -0.00003645, 0.00003493, -0.00002785, 0.00002693, -0.00002373, 0.00002233, -0.00003744, 0.00002956, -0.00001928, 0.00002055, -0.00001439, 0.00001536, -0.00001687, 0.00001369, -0.00001640, 0.00001460, -0.00001308, 0.00001153, -0.00001567, 0.00000942, -0.00000408, 0.00000855, -0.00000326, 0.00000893, -0.00001774, 0.00001082, -0.00000525, 0.00000489, -0.00000766, 0.00000435, -0.00000524, 0.00000335, 0.00000089, 0.00000043, -0.00000996, },
		{ -0.00000892, 0.00000856, -0.00000203, 0.00000340, 0.00000537, -0.00000691, -0.00000041, -0.00000324, 0.00000465, -0.00000537, 0.00000398, -0.00000746, 0.00000595, -0.00000634, 0.00001234, -0.00001148, 0.00001805, -0.00001899, 0.00000385, -0.00001017, 0.00001340, -0.00001334, 0.00001941, -0.00001555, 0.00001244, -0.00001664, 0.00001633, -0.00001478, 0.00001895, -0.00001758, 0.00000709, -0.00001543, 0.00003079, -0.00003927, 0.00002404, -0.00002647, 0.00003766, -0.00004185, 0.00003888, -0.00004374, 0.00004369, -0.00006616, 0.00007150, -0.00008121, 0.00008140, -0.00009008, 0.00008063, -0.00007790, 0.00008397, -0.00012032, 0.00011458, -0.00012844, 0.00013250, -0.00021374, 0.00020647, -0.00035756, 0.00036651, -0.00136652, 0.00203168, 0.00017031, -0.00080061, 0.00032343, -0.00034299, 0.00019995, -0.00021765, 0.00017445, -0.00015832, 0.00010839, -0.00012381, 0.00011363, -0.00011344, 0.00010531, -0.00011731, 0.00010620, -0.00008639, 0.00006099, -0.00006023, 0.00005681, -0.00005591, 0.00005117, -0.00005485, 0.00005083, -0.00003482, 0.00003182, -0.00005080, 0.00004219, -0.00002641, 0.00002541, -0.00003847, 0.00003732, -0.00003059, 0.00003060, -0.00003539, 0.00003417, -0.00003534, 0.00003496, -0.00002801, 0.00002669, -0.00002314, 0.00002071, -0.00003684, 0.00003124, -0.00002106, 0.00002188, -0.00001527, 0.00001547, -0.00001706, 0.00001384, -0.00001641, 0.00001481, -0.00001320, 0.00001120, -0.00001590, 0.00001047, -0.00000420, 0.00000802, -0.00000210, 0.00000658, -0.00001650, 0.00001127, -0.00000558, 0.00000469, -0.00000745, 0.00000422, -0.00000528, 0.00000372, 0.00000107, -0.00000111, },
		{ -0.00000252, -0.00000767, 0.00000912, -0.00000315, 0.00000510, 0.00000338, -0.00000621, -0.00000086, -0.00000212, 0.00000360, -0.00000464, 0.00000305, -0.00000639, 0.00000468, -0.00000432, 0.00001034, -0.00000960, 0.00001704, -0.00002022, 0.00000481, -0.00000921, 0.00001212, -0.00001197, 0.00001854, -0.00001562, 0.00001247, -0.00001652, 0.00001586, -0.00001351, 0.00001751, -0.00001688, 0.00000482, -0.00001009, 0.00002569, -0.00003684, 0.00002135, -0.00002110, 0.00003203, -0.00003727, 0.00003421, -0.00003828, 0.00003761, -0.00005881, 0.00006397, -0.00007508, 0.00007895, -0.00009057, 0.00008198, -0.00007899, 0.00008342, -0.00011880, 0.00011077, -0.00012435, 0.00012776, -0.00020913, 0.00019998, -0.00035510, 0.00036354, -0.00141782, 0.00224083, -0.00006653, -0.00069463, 0.00028145, -0.00031223, 0.00017998, -0.00020116, 0.00016017, -0.00014510, 0.00009746, -0.00011352, 0.00010378, -0.00010353, 0.00009717, -0.00011203, 0.00010472, -0.00008759, 0.00006323, -0.00006194, 0.00005664, -0.00005504, 0.00005000, -0.00005395, 0.00005064, -0.00003401, 0.00002914, -0.00004894, 0.00004248, -0.00002580, 0.00002187, -0.00003420, 0.00003399, -0.00002738, 0.00002729, -0.00003224, 0.00003115, -0.00003269, 0.00003352, -0.00002692, 0.00002538, -0.00002156, 0.00001791, -0.00003473, 0.00003152, -0.00002183, 0.00002250, -0.00001562, 0.00001512, -0.00001677, 0.00001355, -0.00001599, 0.00001465, -0.00001304, 0.00001057, -0.00001579, 0.00001141, -0.00000446, 0.00000768, -0.00000113, 0.00000417, -0.00001484, 0.00001135, -0.00000578, 0.00000445, -0.00000719, 0.00000405, -0.00000531, 0.00000419, 0.00000098, },
		{ 0.00000058, -0.00000369, -0.00000627, 0.00000941, -0.00000414, 0.00000684, 0.00000105, -0.00000500, -0.00000174, -0.00000058, 0.00000202, -0.00000333, 0.00000155, -0.00000471, 0.00000279, -0.00000158, 0.00000739, -0.00000651, 0.00001454, -0.00002008, 0.00000482, -0.00000743, 0.00000985, -0.00000945, 0.00001633, -0.00001437, 0.00001131, -0.00001530, 0.00001437, -0.00001125, 0.00001503, -0.00001527, 0.00000190, -0.00000382, 0.00001892, -0.00003250, 0.00001701, -0.00001402, 0.00002422, -0.00003019, 0.00002693, -0.00003005, 0.00002843, -0.00004773, 0.00005181, -0.00006341, 0.00007036, -0.00008496, 0.00007751, -0.00007452, 0.00007725, -0.00011122, 0.00010027, -0.00011271, 0.00011431, -0.00019410, 0.00018057, -0.00033618, 0.00033731, -0.00143746, 0.00242716, -0.00030796, -0.00056852, 0.00022150, -0.00026758, 0.00014865, -0.00017522, 0.00013778, -0.00012471, 0.00008013, -0.00009735, 0.00008832, -0.00008798, 0.00008312, -0.00010064, 0.00009738, -0.00008380, 0.00006148, -0.00006053, 0.00005387, -0.00005180, 0.00004657, -0.00005088, 0.00004858, -0.00003177, 0.00002505, -0.00004542, 0.00004143, -0.00002449, 0.00001777, -0.00002901, 0.00002959, -0.00002312, 0.00002287, -0.00002789, 0.00002683, -0.00002863, 0.00003068, -0.00002462, 0.00002304, -0.00001909, 0.00001404, -0.00003117, 0.00003035, -0.00002148, 0.00002226, -0.00001535, 0.00001423, -0.00001597, 0.00001279, -0.00001511, 0.00001407, -0.00001257, 0.00000960, -0.00001528, 0.00001216, -0.00000480, 0.00000754, -0.00000043, 0.00000182, -0.00001288, 0.00001111, -0.00000588, 0.00000419, -0.00000692, 0.00000387, -0.00000533, 0.00000478, },
		{ 0.00000548, -0.00000016, -0.00000455, -0.00000482, 0.00000945, -0.00000493, 0.00000850, -0.00000150, -0.00000337, -0.00000297, 0.00000131, -0.00000000, -0.00000153, -0.00000046, -0.00000250, 0.00000039, 0.00000175, 0.00000363, -0.00000234, 0.00001065, -0.00001855, 0.00000383, -0.00000481, 0.00000664, -0.00000580, 0.00001280, -0.00001177, 0.00000893, -0.00001295, 0.00001183, -0.00000800, 0.00001151, -0.00001279, -0.00000153, 0.00000309, 0.00001084, -0.00002652, 0.00001130, -0.00000560, 0.00001466, -0.00002104, 0.00001749, -0.00001955, 0.00001675, -0.00003359, 0.00003570, -0.00004676, 0.00005595, -0.00007331, 0.00006713, -0.00006435, 0.00006536, -0.00009754, 0.00008307, -0.00009355, 0.00009219, -0.00016867, 0.00014812, -0.00030033, 0.00028628, -0.00142090, 0.00258214, -0.00054442, -0.00042929, 0.00014778, -0.00021178, 0.00010794, -0.00014139, 0.00010857, -0.00009832, 0.00005744, -0.00007627, 0.00006821, -0.00006770, 0.00006397, -0.00008372, 0.00008445, -0.00007502, 0.00005560, -0.00005584, 0.00004840, -0.00004614, 0.00004085, -0.00004559, 0.00004460, -0.00002808, 0.00001958, -0.00004027, 0.00003899, -0.00002249, 0.00001323, -0.00002314, 0.00002438, -0.00001807, 0.00001762, -0.00002261, 0.00002148, -0.00002339, 0.00002662, -0.00002124, 0.00001980, -0.00001588, 0.00000932, -0.00002632, 0.00002778, -0.00001999, 0.00002114, -0.00001441, 0.00001280, -0.00001463, 0.00001153, -0.00001377, 0.00001307, -0.00001177, 0.00000831, -0.00001436, 0.00001266, -0.00000518, 0.00000759, -0.00000006, -0.00000032, -0.00001074, 0.00001060, -0.00000588, 0.00000394, -0.00000666, 0.00000370, -0.00000537, },
		{ -0.00000543, 0.00000629, -0.00000125, -0.00000502, -0.00000342, 0.00000929, -0.00000549, 0.00000999, -0.00000413, -0.00000145, -0.00000446, 0.00000343, -0.00000235, 0.00000063, -0.00000284, 0.00000009, -0.00000234, 0.00000546, -0.00000068, 0.00000266, 0.00000554, -0.00001571, 0.00000187, -0.00000143, 0.00000260, -0.00000118, 0.00000808, -0.00000790, 0.00000536, -0.00000949, 0.00000830, -0.00000383, 0.00000705, -0.00000952, -0.00000530, 0.00001031, 0.00000187, -0.00001929, 0.00000460, 0.00000370, 0.00000391, -0.00001042, 0.00000650, -0.00000746, 0.00000336, -0.00001732, 0.00001666, -0.00002614, 0.00003653, -0.00005620, 0.00005125, -0.00004883, 0.00004812, -0.00007822, 0.00005974, -0.00006751, 0.00006214, -0.00013367, 0.00010349, -0.00024829, 0.00021043, -0.00136531, 0.00269851, -0.00076660, -0.00028445, 0.00006525, -0.00014829, 0.00006050, -0.00010179, 0.00007436, -0.00006754, 0.00003085, -0.00005166, 0.00004478, -0.00004402, 0.00004096, -0.00006231, 0.00006668, -0.00006169, 0.00004580, -0.00004798, 0.00004036, -0.00003823, 0.00003301, -0.00003822, 0.00003879, -0.00002306, 0.00001291, -0.00003362, 0.00003521, -0.00001985, 0.00000844, -0.00001686, 0.00001862, -0.00001252, 0.00001186, -0.00001673, 0.00001543, -0.00001728, 0.00002160, -0.00001699, 0.00001586, -0.00001216, 0.00000402, -0.00002045, 0.00002398, -0.00001742, 0.00001914, -0.00001282, 0.00001084, -0.00001281, 0.00000983, -0.00001198, 0.00001165, -0.00001066, 0.00000670, -0.00001300, 0.00001284, -0.00000553, 0.00000782, -0.00000005, -0.00000216, -0.00000853, 0.00000988, -0.00000582, 0.00000374, -0.00000645, 0.00000357, },
		{ 0.00000350, -0.00000551, 0.00000717, -0.00000268, -0.00000503, -0.00000217, 0.00000897, -0.00000581, 0.00001122, -0.00000671, 0.00000064, -0.00000608, 0.00000566, -0.00000487, 0.00000301, -0.00000543, 0.00000290, -0.00000524, 0.00000930, -0.00000526, 0.00000817, -0.00000051, -0.00001172, -0.00000094, 0.00000257, -0.00000206, 0.00000419, 0.00000238, -0.00000295, 0.00000075, -0.00000504, 0.00000387, 0.00000114, 0.00000177, -0.00000560, -0.00000921, 0.00001746, -0.00000754, -0.00001126, -0.00000265, 0.00001333, -0.00000741, 0.00000101, -0.00000533, 0.00000542, -0.00001079, -0.00000005, -0.00000403, -0.00000286, 0.00001335, -0.00003467, 0.00003077, -0.00002878, 0.00002639, -0.00005420, 0.00003132, -0.00003581, 0.00002554, -0.00009069, 0.00004849, -0.00018206, 0.00011140, -0.00126978, 0.00277067, -0.00096603, -0.00014151, -0.00002073, -0.00008106, 0.00000941, -0.00005897, 0.00003734, -0.00003428, 0.00000212, -0.00002516, 0.00001965, -0.00001854, 0.00001567, -0.00003786, 0.00004520, -0.00004462, 0.00003265, -0.00003738, 0.00003011, -0.00002841, 0.00002340, -0.00002908, 0.00003138, -0.00001691, 0.00000529, -0.00002572, 0.00003021, -0.00001666, 0.00000357, -0.00001043, 0.00001263, -0.00000681, 0.00000593, -0.00001062, 0.00000908, -0.00001068, 0.00001594, -0.00001217, 0.00001149, -0.00000823, -0.00000151, -0.00001393, 0.00001922, -0.00001394, 0.00001638, -0.00001066, 0.00000845, -0.00001058, 0.00000776, -0.00000983, 0.00000989, -0.00000929, 0.00000483, -0.00001124, 0.00001267, -0.00000579, 0.00000816, -0.00000042, -0.00000358, -0.00000639, 0.00000902, -0.00000572, 0.00000358, -0.00000632, },
		{ -0.00000627, 0.00000350, -0.00000561, 0.00000810, -0.00000441, -0.00000455, -0.00000114, 0.00000857, -0.00000590, 0.00001212, -0.00000912, 0.00000277, -0.00000773, 0.00000785, -0.00000742, 0.00000544, -0.00000807, 0.00000574, -0.00000808, 0.00001303, -0.00000978, 0.00001382, -0.00000714, -0.00000688, -0.00000439, 0.00000696, -0.00000709, 0.00001001, -0.00000400, 0.00000282, -0.00000468, 0.00000019, -0.00000125, 0.00000671, -0.00000413, -0.00000117, -0.00001306, 0.00002421, -0.00001689, -0.00000291, -0.00000997, 0.00002274, -0.00001862, 0.00001252, -0.00001722, 0.00001822, -0.00002470, 0.00001701, -0.00002495, 0.00002147, -0.00001204, -0.00001015, 0.00000697, -0.00000545, 0.00000144, -0.00002686, -0.00000064, -0.00000016, -0.00001563, -0.00004204, -0.00001417, -0.00010481, -0.00000753, -0.00113555, 0.00279512, -0.00113555, -0.00000753, -0.00010481, -0.00001417, -0.00004204, -0.00001563, -0.00000016, -0.00000064, -0.00002686, 0.00000144, -0.00000545, 0.00000697, -0.00001015, -0.00001204, 0.00002147, -0.00002495, 0.00001701, -0.00002470, 0.00001822, -0.00001722, 0.00001252, -0.00001862, 0.00002274, -0.00000997, -0.00000291, -0.00001689, 0.00002421, -0.00001306, -0.00000117, -0.00000413, 0.00000671, -0.00000125, 0.00000019, -0.00000468, 0.00000282, -0.00000400, 0.00001001, -0.00000709, 0.00000696, -0.00000439, -0.00000688, -0.00000714, 0.00001382, -0.00000978, 0.00001303, -0.00000808, 0.00000574, -0.00000807, 0.00000544, -0.00000742, 0.00000785, -0.00000773, 0.00000277, -0.00000912, 0.00001212, -0.00000590, 0.00000857, -0.00000114, -0.00000455, -0.00000441, 0.00000810, -0.00000561, 0.00000350, },
		{ 0.00000350, -0.00000632, 0.00000358, -0.00000572, 0.00000902, -0.00000639, -0.00000358, -0.00000042, 0.00000816, -0.00000579, 0.00001267, -0.00001124, 0.00000483, -0.00000929, 0.00000989, -0.00000983, 0.00000776, -0.00001058, 0.00000845, -0.00001066, 0.00001638, -0.00001394, 0.00001922, -0.00001393, -0.00000151, -0.00000823, 0.00001149, -0.00001217, 0.00001594, -0.00001068, 0.00000908, -0.00001062, 0.00000593, -0.00000681, 0.00001263, -0.00001043, 0.00000357, -0.00001666, 0.00003021, -0.00002572, 0.00000529, -0.00001691, 0.00003138, -0.00002908, 0.00002340, -0.00002841, 0.00003011, -0.00003738, 0.00003265, -0.00004462, 0.00004520, -0.00003786, 0.00001567, -0.00001854, 0.00001965, -0.00002516, 0.00000212, -0.00003428, 0.00003734, -0.00005897, 0.00000941, -0.00008106, -0.00002073, -0.00014151, -0.00096603, 0.00277067, -0.00126978, 0.00011140, -0.00018206, 0.00004849, -0.00009069, 0.00002554, -0.00003581, 0.00003132, -0.00005420, 0.00002639, -0.00002878, 0.00003077, -0.00003467, 0.00001335, -0.00000286, -0.00000403, -0.00000005, -0.00001079, 0.00000542, -0.00000533, 0.00000101, -0.00000741, 0.00001333, -0.00000265, -0.00001126, -0.00000754, 0.00001746, -0.00000921, -0.00000560, 0.00000177, 0.00000114, 0.00000387, -0.00000504, 0.00000075, -0.00000295, 0.00000238, 0.00000419, -0.00000206, 0.00000257, -0.00000094, -0.00001172, -0.00000051, 0.00000817, -0.00000526, 0.00000930, -0.00000524, 0.00000290, -0.00000543, 0.00000301, -0.00000487, 0.00000566, -0.00000608, 0.00000064, -0.00000671, 0.00001122, -0.00000581, 0.00000897, -0.00000217, -0.00000503, -0.00000268, 0.00000717, -0.00000551, },
		{ -0.00000543, 0.00000357, -0.00000645, 0.00000374, -0.00000582, 0.00000988, -0.00000853, -0.00000216, -0.00000005, 0.00000782, -0.00000553, 0.00001284, -0.00001300, 0.00000670, -0.00001066, 0.00001165, -0.00001198, 0.00000983, -0.00001281, 0.00001084, -0.00001282, 0.00001914, -0.00001742, 0.00002398, -0.00002045, 0.00000402, -0.00001216, 0.00001586, -0.00001699, 0.00002160, -0.00001728, 0.00001543, -0.00001673, 0.00001186, -0.00001252, 0.00001862, -0.00001686, 0.00000844, -0.00001985, 0.00003521, -0.00003362, 0.00001291, -0.00002306, 0.00003879, -0.00003822, 0.00003301, -0.00003823, 0.00004036, -0.00004798, 0.00004580, -0.00006169, 0.00006668, -0.00006231, 0.00004096, -0.00004402, 0.00004478, -0.00005166, 0.00003085, -0.00006754, 0.00007436, -0.00010179, 0.00006050, -0.00014829, 0.00006525, -0.00028445, -0.00076660, 0.00269851, -0.00136531, 0.00021043, -0.00024829, 0.00010349, -0.00013367, 0.00006214, -0.00006751, 0.00005974, -0.00007822, 0.00004812, -0.00004883, 0.00005125, -0.00005620, 0.00003653, -0.00002614, 0.00001666, -0.00001732, 0.00000336, -0.00000746, 0.00000650, -0.00001042, 0.00000391, 0.00000370, 0.00000460, -0.00001929, 0.00000187, 0.00001031, -0.00000530, -0.00000952, 0.00000705, -0.00000383, 0.00000830, -0.00000949, 0.00000536, -0.00000790, 0.00000808, -0.00000118, 0.00000260, -0.00000143, 0.00000187, -0.00001571, 0.00000554, 0.00000266, -0.00000068, 0.00000546, -0.00000234, 0.00000009, -0.00000284, 0.00000063, -0.00000235, 0.00000343, -0.00000446, -0.00000145, -0.00000413, 0.00000999, -0.00000549, 0.00000929, -0.00000342, -0.00000502, -0.00000125, 0.00000629, },
		{ 0.00000548, -0.00000537, 0.00000370, -0.00000666, 0.00000394, -0.00000588, 0.00001060, -0.00001074, -0.00000032, -0.00000006, 0.00000759, -0.00000518, 0.00001266, -0.00001436, 0.00000831, -0.00001177, 0.00001307, -0.00001377, 0.00001153, -0.00001463, 0.00001280, -0.00001441, 0.00002114, -0.00001999, 0.00002778, -0.00002632, 0.00000932, -0.00001588, 0.00001980, -0.00002124, 0.00002662, -0.00002339, 0.00002148, -0.00002261, 0.00001762, -0.00001807, 0.00002438, -0.00002314, 0.00001323, -0.00002249, 0.00003899, -0.00004027, 0.00001958, -0.00002808, 0.00004460, -0.00004559, 0.00004085, -0.00004614, 0.00004840, -0.00005584, 0.00005560, -0.00007502, 0.00008445, -0.00008372, 0.00006397, -0.00006770, 0.00006821, -0.00007627, 0.00005744, -0.00009832, 0.00010857, -0.00014139, 0.00010794, -0.00021178, 0.00014778, -0.00042929, -0.00054442, 0.00258214, -0.00142090, 0.00028628, -0.00030033, 0.00014812, -0.00016867, 0.00009219, -0.00009355, 0.00008307, -0.00009754, 0.00006536, -0.00006435, 0.00006713, -0.00007331, 0.00005595, -0.00004676, 0.00003570, -0.00003359, 0.00001675, -0.00001955, 0.00001749, -0.00002104, 0.00001466, -0.00000560, 0.00001130, -0.00002652, 0.00001084, 0.00000309, -0.00000153, -0.00001279, 0.00001151, -0.00000800, 0.00001183, -0.00001295, 0.00000893, -0.00001177, 0.00001280, -0.00000580, 0.00000664, -0.00000481, 0.00000383, -0.00001855, 0.00001065, -0.00000234, 0.00000363, 0.00000175, 0.00000039, -0.00000250, -0.00000046, -0.00000153, -0.00000000, 0.00000131, -0.00000297, -0.00000337, -0.00000150, 0.00000850, -0.00000493, 0.00000945, -0.00000482, -0.00000455, -0.00000016, },
		{ 0.00000058, 0.00000478, -0.00000533, 0.00000387, -0.00000692, 0.00000419, -0.00000588, 0.00001111, -0.00001288, 0.00000182, -0.00000043, 0.00000754, -0.00000480, 0.00001216, -0.00001528, 0.00000960, -0.00001257, 0.00001407, -0.00001511, 0.00001279, -0.00001597, 0.00001423, -0.00001535, 0.00002226, -0.00002148, 0.00003035, -0.00003117, 0.00001404, -0.00001909, 0.00002304, -0.00002462, 0.00003068, -0.00002863, 0.00002683, -0.00002789, 0.00002287, -0.00002312, 0.00002959, -0.00002901, 0.00001777, -0.00002449, 0.00004143, -0.00004542, 0.00002505, -0.00003177, 0.00004858, -0.00005088, 0.00004657, -0.00005180, 0.00005387, -0.00006053, 0.00006148, -0.00008380, 0.00009738, -0.00010064, 0.00008312, -0.00008798, 0.00008832, -0.00009735, 0.00008013, -0.00012471, 0.00013778, -0.00017522, 0.00014865, -0.00026758, 0.00022150, -0.00056852, -0.00030796, 0.00242716, -0.00143746, 0.00033731, -0.00033618, 0.00018057, -0.00019410, 0.00011431, -0.00011271, 0.00010027, -0.00011122, 0.00007725, -0.00007452, 0.00007751, -0.00008496, 0.00007036, -0.00006341, 0.00005181, -0.00004773, 0.00002843, -0.00003005, 0.00002693, -0.00003019, 0.00002422, -0.00001402, 0.00001701, -0.00003250, 0.00001892, -0.00000382, 0.00000190, -0.00001527, 0.00001503, -0.00001125, 0.00001437, -0.00001530, 0.00001131, -0.00001437, 0.00001633, -0.00000945, 0.00000985, -0.00000743, 0.00000482, -0.00002008, 0.00001454, -0.00000651, 0.00000739, -0.00000158, 0.00000279, -0.00000471, 0.00000155, -0.00000333, 0.00000202, -0.00000058, -0.00000174, -0.00000500, 0.00000105, 0.00000684, -0.00000414, 0.00000941, -0.00000627, -0.00000369, },
		{ -0.00000252, 0.00000098, 0.00000419, -0.00000531, 0.00000405, -0.00000719, 0.00000445, -0.00000578, 0.00001135, -0.00001484, 0.00000417, -0.00000113, 0.00000768, -0.00000446, 0.00001141, -0.00001579, 0.00001057, -0.00001304, 0.00001465, -0.00001599, 0.00001355, -0.00001677, 0.00001512, -0.00001562, 0.00002250, -0.00002183, 0.00003152, -0.00003473, 0.00001791, -0.00002156, 0.00002538, -0.00002692, 0.00003352, -0.00003269, 0.00003115, -0.00003224, 0.00002729, -0.00002738, 0.00003399, -0.00003420, 0.00002187, -0.00002580, 0.00004248, -0.00004894, 0.00002914, -0.00003401, 0.00005064, -0.00005395, 0.00005000, -0.00005504, 0.00005664, -0.00006194, 0.00006323, -0.00008759, 0.00010472, -0.00011203, 0.00009717, -0.00010353, 0.00010378, -0.00011352, 0.00009746, -0.00014510, 0.00016017, -0.00020116, 0.00017998, -0.00031223, 0.00028145, -0.00069463, -0.00006653, 0.00224083, -0.00141782, 0.00036354, -0.00035510, 0.00019998, -0.00020913, 0.00012776, -0.00012435, 0.00011077, -0.00011880, 0.00008342, -0.00007899, 0.00008198, -0.00009057, 0.00007895, -0.00007508, 0.00006397, -0.00005881, 0.00003761, -0.00003828, 0.00003421, -0.00003727, 0.00003203, -0.00002110, 0.00002135, -0.00003684, 0.00002569, -0.00001009, 0.00000482, -0.00001688, 0.00001751, -0.00001351, 0.00001586, -0.00001652, 0.00001247, -0.00001562, 0.00001854, -0.00001197, 0.00001212, -0.00000921, 0.00000481, -0.00002022, 0.00001704, -0.00000960, 0.00001034, -0.00000432, 0.00000468, -0.00000639, 0.00000305, -0.00000464, 0.00000360, -0.00000212, -0.00000086, -0.00000621, 0.00000338, 0.00000510, -0.00000315, 0.00000912, -0.00000767, },
		{ -0.00000892, -0.00000111, 0.00000107, 0.00000372, -0.00000528, 0.00000422, -0.00000745, 0.00000469, -0.00000558, 0.00001127, -0.00001650, 0.00000658, -0.00000210, 0.00000802, -0.00000420, 0.00001047, -0.00001590, 0.00001120, -0.00001320, 0.00001481, -0.00001641, 0.00001384, -0.00001706, 0.00001547, -0.00001527, 0.00002188, -0.00002106, 0.00003124, -0.00003684, 0.00002071, -0.00002314, 0.00002669, -0.00002801, 0.00003496, -0.00003534, 0.00003417, -0.00003539, 0.00003060, -0.00003059, 0.00003732, -0.00003847, 0.00002541, -0.00002641, 0.00004219, -0.00005080, 0.00003182, -0.00003482, 0.00005083, -0.00005485, 0.00005117, -0.00005591, 0.00005681, -0.00006023, 0.00006099, -0.00008639, 0.00010620, -0.00011731, 0.00010531, -0.00011344, 0.00011363, -0.00012381, 0.00010839, -0.00015832, 0.00017445, -0.00021765, 0.00019995, -0.00034299, 0.00032343, -0.00080061, 0.00017031, 0.00203168, -0.00136652, 0.00036651, -0.00035756, 0.00020647, -0.00021374, 0.00013250, -0.00012844, 0.00011458, -0.00012032, 0.00008397, -0.00007790, 0.00008063, -0.00009008, 0.00008140, -0.00008121, 0.00007150, -0.00006616, 0.00004369, -0.00004374, 0.00003888, -0.00004185, 0.00003766, -0.00002647, 0.00002404, -0.00003927, 0.00003079, -0.00001543, 0.00000709, -0.00001758, 0.00001895, -0.00001478, 0.00001633, -0.00001664, 0.00001244, -0.00001555, 0.00001941, -0.00001334, 0.00001340, -0.00001017, 0.00000385, -0.00001899, 0.00001805, -0.00001148, 0.00001234, -0.00000634, 0.00000595, -0.00000746, 0.00000398, -0.00000537, 0.00000465, -0.00000324, -0.00000041, -0.00000691, 0.00000537, 0.00000340, -0.00000203, 0.00000856, },
		{ 0.00000773, -0.00000996, 0.00000043, 0.00000089, 0.00000335, -0.00000524, 0.00000435, -0.00000766, 0.00000489, -0.00000525, 0.00001082, -0.00001774, 0.00000893, -0.00000326, 0.00000855, -0.00000408, 0.00000942, -0.00001567, 0.00001153, -0.00001308, 0.00001460, -0.00001640, 0.00001369, -0.00001687, 0.00001536, -0.00001439, 0.00002055, -0.00001928, 0.00002956, -0.00003744, 0.00002233, -0.00002373, 0.00002693, -0.00002785, 0.00003493, -0.00003645, 0.00003572, -0.00003714, 0.00003262, -0.00003258, 0.00003941, -0.00004166, 0.00002827, -0.00002637, 0.00004066, -0.00005106, 0.00003313, -0.00003432, 0.00004933, -0.00005375, 0.00005026, -0.00005464, 0.00005472, -0.00005582, 0.00005521, -0.00008060, 0.00010202, -0.00011638, 0.00010722, -0.00011726, 0.00011738, -0.00012771, 0.00011240, -0.00016381, 0.00017992, -0.00022387, 0.00020742, -0.00035813, 0.00034439, -0.00088049, 0.00039328, 0.00180891, -0.00128938, 0.00034907, -0.00034513, 0.00020105, -0.00020868, 0.00012914, -0.00012551, 0.00011221, -0.00011629, 0.00007944, -0.00007182, 0.00007403, -0.00008392, 0.00007788, -0.00008170, 0.00007413, -0.00006943, 0.00004636, -0.00004617, 0.00004072, -0.00004372, 0.00004085, -0.00002991, 0.00002493, -0.00003963, 0.00003399, -0.00001960, 0.00000861, -0.00001737, 0.00001938, -0.00001512, 0.00001589, -0.00001580, 0.00001136, -0.00001428, 0.00001902, -0.00001357, 0.00001372, -0.00001037, 0.00000210, -0.00001653, 0.00001764, -0.00001210, 0.00001329, -0.00000756, 0.00000655, -0.00000788, 0.00000429, -0.00000549, 0.00000512, -0.00000389, -0.00000042, -0.00000704, 0.00000689, 0.00000186, -0.00000083, },
		{ 0.00000034, 0.00000667, -0.00001072, 0.00000201, 0.00000048, 0.00000309, -0.00000519, 0.00000443, -0.00000781, 0.00000504, -0.00000481, 0.00001000, -0.00001848, 0.00001107, -0.00000449, 0.00000923, -0.00000411, 0.00000832, -0.00001517, 0.00001159, -0.00001272, 0.00001409, -0.00001604, 0.00001317, -0.00001629, 0.00001488, -0.00001313, 0.00001869, -0.00001671, 0.00002665, -0.00003660, 0.00002277, -0.00002334, 0.00002615, -0.00002650, 0.00003348, -0.00003601, 0.00003575, -0.00003742, 0.00003324, -0.00003324, 0.00004014, -0.00004362, 0.00003037, -0.00002575, 0.00003807, -0.00004989, 0.00003322, -0.00003274, 0.00004643, -0.00005096, 0.00004759, -0.00005160, 0.00005084, -0.00004936, 0.00004668, -0.00007101, 0.00009283, -0.00010967, 0.00010309, -0.00011503, 0.00011505, -0.00012525, 0.00010951, -0.00016157, 0.00017659, -0.00021976, 0.00020217, -0.00035706, 0.00034264, -0.00092968, 0.00059401, 0.00158180, -0.00119303, 0.00031505, -0.00032024, 0.00018551, -0.00019531, 0.00011882, -0.00011658, 0.00010457, -0.00010762, 0.00007077, -0.00006174, 0.00006316, -0.00007300, 0.00006905, -0.00007691, 0.00007197, -0.00006863, 0.00004559, -0.00004555, 0.00003972, -0.00004282, 0.00004153, -0.00003135, 0.00002404, -0.00003790, 0.00003516, -0.00002245, 0.00000933, -0.00001631, 0.00001887, -0.00001463, 0.00001468, -0.00001416, 0.00000944, -0.00001200, 0.00001751, -0.00001279, 0.00001321, -0.00000997, -0.00000024, -0.00001308, 0.00001594, -0.00001151, 0.00001321, -0.00000796, 0.00000648, -0.00000766, 0.00000400, -0.00000502, 0.00000500, -0.00000406, -0.00000089, -0.00000657, 0.00000788, 0.00000057, },
		{ -0.00000036, 0.00000141, 0.00000542, -0.00001117, 0.00000355, -0.00000009, 0.00000292, -0.00000510, 0.00000443, -0.00000788, 0.00000514, -0.00000426, 0.00000880, -0.00001866, 0.00001289, -0.00000570, 0.00000999, -0.00000432, 0.00000724, -0.00001445, 0.00001144, -0.00001219, 0.00001335, -0.00001541, 0.00001238, -0.00001543, 0.00001415, -0.00001165, 0.00001651, -0.00001362, 0.00002277, -0.00003447, 0.00002211, -0.00002207, 0.00002447, -0.00002413, 0.00003075, -0.00003413, 0.00003433, -0.00003624, 0.00003247, -0.00003257, 0.00003949, -0.00004431, 0.00003170, -0.00002465, 0.00003464, -0.00004749, 0.00003227, -0.00003034, 0.00004251, -0.00004688, 0.00004359, -0.00004729, 0.00004579, -0.00004162, 0.00003634, -0.00005868, 0.00007969, -0.00009805, 0.00009360, -0.00010732, 0.00010715, -0.00011696, 0.00010029, -0.00015223, 0.00016515, -0.00020608, 0.00018499, -0.00034040, 0.00031805, -0.00094533, 0.00076545, 0.00135923, -0.00108443, 0.00026889, -0.00028595, 0.00016212, -0.00017547, 0.00010310, -0.00010301, 0.00009294, -0.00009551, 0.00005915, -0.00004890, 0.00004930, -0.00005854, 0.00005597, -0.00006759, 0.00006553, -0.00006408, 0.00004164, -0.00004212, 0.00003611, -0.00003935, 0.00003981, -0.00003088, 0.00002150, -0.00003422, 0.00003433, -0.00002393, 0.00000928, -0.00001449, 0.00001757, -0.00001346, 0.00001288, -0.00001196, 0.00000692, -0.00000897, 0.00001512, -0.00001118, 0.00001200, -0.00000914, -0.00000291, -0.00000892, 0.00001319, -0.00000986, 0.00001218, -0.00000760, 0.00000580, -0.00000687, 0.00000317, -0.00000400, 0.00000433, -0.00000377, -0.00000178, -0.00000552, 0.00000829, },
		{ 0.00000810, -0.00000089, 0.00000228, 0.00000404, -0.00001129, 0.00000499, -0.00000079, 0.00000283, -0.00000500, 0.00000436, -0.00000787, 0.00000519, -0.00000366, 0.00000728, -0.00001824, 0.00001429, -0.00000679, 0.00001078, -0.00000468, 0.00000624, -0.00001360, 0.00001114, -0.00001156, 0.00001247, -0.00001460, 0.00001144, -0.00001440, 0.00001333, -0.00001014, 0.00001429, -0.00001032, 0.00001825, -0.00003132, 0.00002052, -0.00002010, 0.00002210, -0.00002098, 0.00002699, -0.00003103, 0.00003161, -0.00003373, 0.00003041, -0.00003064, 0.00003752, -0.00004373, 0.00003224, -0.00002319, 0.00003062, -0.00004412, 0.00003053, -0.00002743, 0.00003795, -0.00004197, 0.00003873, -0.00004224, 0.00004021, -0.00003348, 0.00002528, -0.00004488, 0.00006390, -0.00008275, 0.00007982, -0.00009507, 0.00009459, -0.00010380, 0.00008580, -0.00013694, 0.00014688, -0.00018432, 0.00015754, -0.00030995, 0.00027206, -0.00092653, 0.00090233, 0.00114908, -0.00097044, 0.00021530, -0.00024561, 0.00013353, -0.00015129, 0.00008378, -0.00008639, 0.00007873, -0.00008130, 0.00004593, -0.00003469, 0.00003390, -0.00004205, 0.00004000, -0.00005485, 0.00005564, -0.00005642, 0.00003500, -0.00003632, 0.00003030, -0.00003367, 0.00003596, -0.00002874, 0.00001758, -0.00002885, 0.00003162, -0.00002408, 0.00000850, -0.00001206, 0.00001562, -0.00001179, 0.00001069, -0.00000943, 0.00000407, -0.00000547, 0.00001209, -0.00000895, 0.00001030, -0.00000809, -0.00000561, -0.00000441, 0.00000970, -0.00000739, 0.00001037, -0.00000661, 0.00000462, -0.00000562, 0.00000193, -0.00000256, 0.00000321, -0.00000312, -0.00000301, -0.00000396, },
		{ -0.00000197, 0.00000735, -0.00000097, 0.00000290, 0.00000260, -0.00001109, 0.00000626, -0.00000155, 0.00000281, -0.00000488, 0.00000424, -0.00000777, 0.00000521, -0.00000305, 0.00000551, -0.00001724, 0.00001520, -0.00000765, 0.00001151, -0.00000517, 0.00000534, -0.00001266, 0.00001073, -0.00001089, 0.00001153, -0.00001372, 0.00001043, -0.00001332, 0.00001253, -0.00000878, 0.00001224, -0.00000713, 0.00001344, -0.00002746, 0.00001826, -0.00001763, 0.00001931, -0.00001734, 0.00002253, -0.00002700, 0.00002786, -0.00003014, 0.00002725, -0.00002765, 0.00003438, -0.00004195, 0.00003204, -0.00002151, 0.00002626, -0.00004006, 0.00002825, -0.00002431, 0.00003316, -0.00003667, 0.00003350, -0.00003696, 0.00003474, -0.00002574, 0.00001457, -0.00003093, 0.00004694, -0.00006522, 0.00006312, -0.00007958, 0.00007866, -0.00008711, 0.00006746, -0.00011729, 0.00012357, -0.00015653, 0.00012222, -0.00026854, 0.00020760, -0.00087436, 0.00100137, 0.00095789, -0.00085729, 0.00015882, -0.00020263, 0.00010241, -0.00012497, 0.00006273, -0.00006835, 0.00006344, -0.00006638, 0.00003245, -0.00002051, 0.00001846, -0.00002508, 0.00002264, -0.00003999, 0.00004334, -0.00004647, 0.00002639, -0.00002879, 0.00002287, -0.00002632, 0.00003043, -0.00002528, 0.00001267, -0.00002220, 0.00002730, -0.00002300, 0.00000712, -0.00000917, 0.00001320, -0.00000979, 0.00000833, -0.00000682, 0.00000118, -0.00000180, 0.00000872, -0.00000635, 0.00000829, -0.00000701, -0.00000808, 0.00000010, 0.00000581, -0.00000436, 0.00000797, -0.00000514, 0.00000310, -0.00000406, 0.00000041, -0.00000082, 0.00000175, -0.00000219, -0.00000446, },
		{ -0.00000602, 0.00000030, 0.00000613, -0.00000061, 0.00000319, 0.00000118, -0.00001059, 0.00000732, -0.00000235, 0.00000285, -0.00000476, 0.00000406, -0.00000761, 0.00000521, -0.00000249, 0.00000357, -0.00001572, 0.00001559, -0.00000822, 0.00001213, -0.00000577, 0.00000459, -0.00001169, 0.00001026, -0.00001023, 0.00001059, -0.00001284, 0.00000945, -0.00001229, 0.00001187, -0.00000770, 0.00001059, -0.00000435, 0.00000873, -0.00002324, 0.00001559, -0.00001494, 0.00001638, -0.00001356, 0.00001773, -0.00002242, 0.00002342, -0.00002576, 0.00002327, -0.00002383, 0.00003027, -0.00003911, 0.00003119, -0.00001974, 0.00002181, -0.00003558, 0.00002568, -0.00002128, 0.00002850, -0.00003140, 0.00002832, -0.00003191, 0.00002992, -0.00001912, 0.00000518, -0.00001810, 0.00003029, -0.00004702, 0.00004503, -0.00006230, 0.00006083, -0.00006844, 0.00004697, -0.00009513, 0.00009733, -0.00012520, 0.00008203, -0.00021981, 0.00012890, -0.00079178, 0.00106142, 0.00079055, -0.00075032, 0.00010356, -0.00016015, 0.00007129, -0.00009861, 0.00004174, -0.00005045, 0.00004844, -0.00005200, 0.00001994, -0.00000762, 0.00000434, -0.00000914, 0.00000544, -0.00002440, 0.00002984, -0.00003523, 0.00001662, -0.00002026, 0.00001453, -0.00001792, 0.00002371, -0.00002096, 0.00000724, -0.00001472, 0.00002172, -0.00002090, 0.00000530, -0.00000602, 0.00001051, -0.00000764, 0.00000597, -0.00000432, -0.00000150, 0.00000175, 0.00000527, -0.00000361, 0.00000618, -0.00000608, -0.00001006, 0.00000429, 0.00000186, -0.00000107, 0.00000522, -0.00000338, 0.00000139, -0.00000235, -0.00000122, 0.00000105, 0.00000010, -0.00000111, },
		{ -0.00000001, -0.00000752, 0.00000271, 0.00000453, 0.00000016, 0.00000312, -0.00000015, -0.00000984, 0.00000815, -0.00000315, 0.00000296, -0.00000465, 0.00000387, -0.00000741, 0.00000524, -0.00000206, 0.00000159, -0.00001376, 0.00001546, -0.00000846, 0.00001257, -0.00000642, 0.00000400, -0.00001074, 0.00000979, -0.00000962, 0.00000972, -0.00001202, 0.00000857, -0.00001138, 0.00001142, -0.00000700, 0.00000949, -0.00000224, 0.00000444, -0.00001899, 0.00001281, -0.00001229, 0.00001360, -0.00000997, 0.00001298, -0.00001766, 0.00001868, -0.00002095, 0.00001880, -0.00001949, 0.00002548, -0.00003541, 0.00002977, -0.00001800, 0.00001750, -0.00003093, 0.00002302, -0.00001856, 0.00002428, -0.00002651, 0.00002356, -0.00002745, 0.00002614, -0.00001419, -0.00000207, -0.00000747, 0.00001533, -0.00002969, 0.00002709, -0.00004478, 0.00004263, -0.00004941, 0.00002608, -0.00007241, 0.00007039, -0.00009298, 0.00004021, -0.00016790, 0.00004116, -0.00068343, 0.00108342, 0.00065013, -0.00065368, 0.00005294, -0.00012085, 0.00004238, -0.00007402, 0.00002234, -0.00003402, 0.00003489, -0.00003921, 0.00000938, 0.00000294, -0.00000730, 0.00000444, -0.00001020, -0.00000944, 0.00001632, -0.00002370, 0.00000657, -0.00001151, 0.00000600, -0.00000914, 0.00001641, -0.00001624, 0.00000177, -0.00000694, 0.00001530, -0.00001803, 0.00000320, -0.00000280, 0.00000772, -0.00000551, 0.00000378, -0.00000213, -0.00000378, 0.00000495, 0.00000200, -0.00000093, 0.00000411, -0.00000544, -0.00001138, 0.00000785, -0.00000182, 0.00000219, 0.00000237, -0.00000153, -0.00000032, -0.00000065, -0.00000281, 0.00000290, -0.00000161, },
		{ -0.00000323, 0.00000097, -0.00000883, 0.00000509, 0.00000268, 0.00000128, 0.00000268, -0.00000132, -0.00000888, 0.00000874, -0.00000391, 0.00000311, -0.00000457, 0.00000368, -0.00000720, 0.00000532, -0.00000181, -0.00000030, -0.00001149, 0.00001487, -0.00000835, 0.00001278, -0.00000709, 0.00000355, -0.00000984, 0.00000932, -0.00000909, 0.00000895, -0.00001131, 0.00000784, -0.00001063, 0.00001122, -0.00000674, 0.00000903, -0.00000099, 0.00000088, -0.00001504, 0.00001017, -0.00000991, 0.00001123, -0.00000688, 0.00000864, -0.00001312, 0.00001400, -0.00001609, 0.00001419, -0.00001497, 0.00002032, -0.00003107, 0.00002791, -0.00001641, 0.00001351, -0.00002633, 0.00002046, -0.00001632, 0.00002072, -0.00002226, 0.00001946, -0.00002382, 0.00002363, -0.00001126, -0.00000668, 0.00000013, 0.00000319, -0.00001454, 0.00001072, -0.00002844, 0.00002554, -0.00003157, 0.00000649, -0.00005103, 0.00004492, -0.00006248, 0.00000003, -0.00011716, -0.00004980, -0.00055530, 0.00107022, 0.00053786, -0.00057020, 0.00000947, -0.00008677, 0.00001736, -0.00005262, 0.00000572, -0.00002005, 0.00002367, -0.00002877, 0.00000147, 0.00001048, -0.00001565, 0.00001467, -0.00002308, 0.00000368, 0.00000389, -0.00001286, -0.00000293, -0.00000330, -0.00000201, -0.00000068, 0.00000909, -0.00001162, -0.00000324, 0.00000061, 0.00000848, -0.00001467, 0.00000104, 0.00000031, 0.00000500, -0.00000352, 0.00000188, -0.00000036, -0.00000548, 0.00000759, -0.00000090, 0.00000149, 0.00000222, -0.00000517, -0.00001191, 0.00001059, -0.00000497, 0.00000515, -0.00000036, 0.00000023, -0.00000189, 0.00000088, -0.00000420, 0.00000456, },
		{ 0.00000592, -0.00000464, 0.00000174, -0.00000982, 0.00000728, 0.00000074, 0.00000266, 0.00000188, -0.00000226, -0.00000778, 0.00000909, -0.00000462, 0.00000332, -0.00000454, 0.00000352, -0.00000700, 0.00000548, -0.00000182, -0.00000199, -0.00000904, 0.00001388, -0.00000791, 0.00001273, -0.00000774, 0.00000326, -0.00000901, 0.00000888, -0.00000866, 0.00000831, -0.00001073, 0.00000727, -0.00001006, 0.00001126, -0.00000690, 0.00000925, -0.00000069, -0.00000174, -0.00001165, 0.00000792, -0.00000801, 0.00000948, -0.00000454, 0.00000502, -0.00000915, 0.00000977, -0.00001152, 0.00000978, -0.00001061, 0.00001512, -0.00002635, 0.00002573, -0.00001504, 0.00001000, -0.00002196, 0.00001811, -0.00001465, 0.00001794, -0.00001879, 0.00001617, -0.00002108, 0.00002244, -0.00001041, -0.00000842, 0.00000421, -0.00000533, -0.00000264, -0.00000290, -0.00001451, 0.00001081, -0.00001626, -0.00001035, -0.00003263, 0.00002281, -0.00003600, -0.00003553, -0.00007170, -0.00013799, -0.00041427, 0.00102627, 0.00045324, -0.00050134, -0.00002528, -0.00005925, -0.00000262, -0.00003534, -0.00000734, -0.00000920, 0.00001528, -0.00002106, -0.00000348, 0.00001467, -0.00002031, 0.00002091, -0.00003236, 0.00001402, -0.00000652, -0.00000352, -0.00001115, 0.00000374, -0.00000888, 0.00000689, 0.00000228, -0.00000751, -0.00000738, 0.00000744, 0.00000172, -0.00001111, -0.00000101, 0.00000314, 0.00000250, -0.00000177, 0.00000034, 0.00000091, -0.00000653, 0.00000954, -0.00000328, 0.00000354, 0.00000058, -0.00000527, -0.00001161, 0.00001235, -0.00000738, 0.00000760, -0.00000278, 0.00000176, -0.00000319, 0.00000213, -0.00000528, },
		{ -0.00000597, 0.00000688, -0.00000575, 0.00000221, -0.00001039, 0.00000914, -0.00000115, 0.00000419, 0.00000077, -0.00000295, -0.00000659, 0.00000922, -0.00000526, 0.00000358, -0.00000458, 0.00000341, -0.00000684, 0.00000573, -0.00000214, -0.00000334, -0.00000656, 0.00001262, -0.00000719, 0.00001242, -0.00000832, 0.00000309, -0.00000826, 0.00000848, -0.00000833, 0.00000778, -0.00001027, 0.00000687, -0.00000966, 0.00001150, -0.00000742, 0.00001008, -0.00000136, -0.00000332, -0.00000899, 0.00000620, -0.00000671, 0.00000847, -0.00000313, 0.00000236, -0.00000603, 0.00000628, -0.00000758, 0.00000590, -0.00000673, 0.00001021, -0.00002150, 0.00002337, -0.00001397, 0.00000708, -0.00001794, 0.00001603, -0.00001357, 0.00001599, -0.00001616, 0.00001370, -0.00001920, 0.00002243, -0.00001147, -0.00000736, 0.00000465, -0.00000982, 0.00000535, -0.00001295, -0.00000388, -0.00000062, -0.00000447, -0.00002332, -0.00001846, 0.00000554, -0.00001540, -0.00006401, -0.00003516, -0.00021769, -0.00026767, 0.00095728, 0.00039427, -0.00044729, -0.00005073, -0.00003886, -0.00001705, -0.00002263, -0.00001648, -0.00000172, 0.00000988, -0.00001616, -0.00000549, 0.00001555, -0.00002126, 0.00002295, -0.00003759, 0.00002097, -0.00001428, 0.00000369, -0.00001755, 0.00000910, -0.00001413, 0.00001307, -0.00000360, -0.00000426, -0.00001031, 0.00001313, -0.00000458, -0.00000763, -0.00000277, 0.00000555, 0.00000032, -0.00000034, -0.00000081, 0.00000167, -0.00000689, 0.00001075, -0.00000505, 0.00000513, -0.00000078, -0.00000572, -0.00001054, 0.00001311, -0.00000894, 0.00000940, -0.00000475, 0.00000295, -0.00000413, 0.00000303, },
		{ 0.00000353, -0.00000622, 0.00000740, -0.00000649, 0.00000233, -0.00001047, 0.00001054, -0.00000284, 0.00000576, -0.00000060, -0.00000336, -0.00000538, 0.00000914, -0.00000581, 0.00000386, -0.00000468, 0.00000337, -0.00000672, 0.00000607, -0.00000278, -0.00000425, -0.00000422, 0.00001118, -0.00000624, 0.00001187, -0.00000881, 0.00000305, -0.00000760, 0.00000812, -0.00000808, 0.00000736, -0.00000991, 0.00000659, -0.00000937, 0.00001187, -0.00000820, 0.00001139, -0.00000291, -0.00000382, -0.00000716, 0.00000510, -0.00000607, 0.00000826, -0.00000273, 0.00000080, -0.00000396, 0.00000377, -0.00000452, 0.00000281, -0.00000362, 0.00000587, -0.00001680, 0.00002096, -0.00001324, 0.00000481, -0.00001436, 0.00001424, -0.00001304, 0.00001481, -0.00001432, 0.00001198, -0.00001801, 0.00002331, -0.00001407, -0.00000388, 0.00000170, -0.00001026, 0.00000915, -0.00001898, 0.00000291, -0.00000819, 0.00000318, -0.00003175, -0.00000930, -0.00000596, -0.00000187, -0.00008371, -0.00001031, -0.00028384, -0.00012278, 0.00086971, 0.00035775, -0.00040715, -0.00006712, -0.00002550, -0.00002597, -0.00001445, -0.00002179, 0.00000252, 0.00000730, -0.00001379, -0.00000491, 0.00001353, -0.00001888, 0.00002103, -0.00003873, 0.00002429, -0.00001902, 0.00000843, -0.00002181, 0.00001250, -0.00001746, 0.00001752, -0.00000823, -0.00000208, -0.00001180, 0.00001737, -0.00001008, -0.00000446, -0.00000411, 0.00000742, -0.00000146, 0.00000076, -0.00000157, 0.00000196, -0.00000660, 0.00001122, -0.00000617, 0.00000624, -0.00000187, -0.00000642, -0.00000879, 0.00001291, -0.00000961, 0.00001047, -0.00000620, 0.00000374, -0.00000468, },
		{ -0.00000483, 0.00000365, -0.00000606, 0.00000746, -0.00000687, 0.00000211, -0.00001005, 0.00001142, -0.00000423, 0.00000724, -0.00000215, -0.00000347, -0.00000419, 0.00000888, -0.00000625, 0.00000416, -0.00000484, 0.00000341, -0.00000666, 0.00000650, -0.00000375, -0.00000465, -0.00000214, 0.00000972, -0.00000516, 0.00001111, -0.00000921, 0.00000311, -0.00000704, 0.00000780, -0.00000790, 0.00000703, -0.00000963, 0.00000640, -0.00000915, 0.00001228, -0.00000908, 0.00001302, -0.00000518, -0.00000331, -0.00000616, 0.00000464, -0.00000607, 0.00000881, -0.00000332, 0.00000037, -0.00000301, 0.00000235, -0.00000249, 0.00000069, -0.00000146, 0.00000236, -0.00001245, 0.00001862, -0.00001285, 0.00000321, -0.00001126, 0.00001272, -0.00001297, 0.00001430, -0.00001318, 0.00001086, -0.00001726, 0.00002467, -0.00001767, 0.00000140, -0.00000408, -0.00000704, 0.00000892, -0.00002096, 0.00000576, -0.00001175, 0.00000652, -0.00003542, -0.00000541, -0.00001137, 0.00000409, -0.00009378, 0.00000110, -0.00033241, 0.00001366, 0.00077031, 0.00033963, -0.00037911, -0.00007547, -0.00001848, -0.00002989, -0.00001035, -0.00002367, 0.00000392, 0.00000708, -0.00001344, -0.00000235, 0.00000929, -0.00001388, 0.00001572, -0.00003610, 0.00002407, -0.00002070, 0.00001058, -0.00002381, 0.00001382, -0.00001875, 0.00002010, -0.00001146, -0.00000106, -0.00001179, 0.00001996, -0.00001452, -0.00000179, -0.00000494, 0.00000871, -0.00000279, 0.00000151, -0.00000200, 0.00000186, -0.00000575, 0.00001099, -0.00000669, 0.00000687, -0.00000273, -0.00000723, -0.00000656, 0.00001187, -0.00000945, 0.00001081, -0.00000709, 0.00000411, },
		{ 0.00000410, -0.00000464, 0.00000345, -0.00000553, 0.00000711, -0.00000690, 0.00000158, -0.00000915, 0.00001174, -0.00000522, 0.00000853, -0.00000381, -0.00000330, -0.00000308, 0.00000847, -0.00000659, 0.00000446, -0.00000505, 0.00000352, -0.00000664, 0.00000699, -0.00000501, -0.00000449, -0.00000044, 0.00000835, -0.00000405, 0.00001019, -0.00000950, 0.00000325, -0.00000657, 0.00000752, -0.00000777, 0.00000677, -0.00000939, 0.00000626, -0.00000893, 0.00001263, -0.00000993, 0.00001474, -0.00000794, -0.00000196, -0.00000592, 0.00000476, -0.00000661, 0.00001000, -0.00000478, 0.00000101, -0.00000318, 0.00000207, -0.00000155, -0.00000038, -0.00000040, -0.00000014, -0.00000866, 0.00001643, -0.00001281, 0.00000227, -0.00000864, 0.00001142, -0.00001323, 0.00001431, -0.00001256, 0.00001016, -0.00001668, 0.00002607, -0.00002162, 0.00000769, -0.00001188, -0.00000085, 0.00000517, -0.00001925, 0.00000493, -0.00001155, 0.00000581, -0.00003460, -0.00000652, -0.00001098, 0.00000275, -0.00009429, -0.00000153, -0.00036068, 0.00013585, 0.00066568, 0.00033544, -0.00036079, -0.00007728, -0.00001665, -0.00002972, -0.00000957, -0.00002280, 0.00000313, 0.00000860, -0.00001442, 0.00000142, 0.00000369, -0.00000720, 0.00000791, -0.00003038, 0.00002075, -0.00001959, 0.00001032, -0.00002369, 0.00001320, -0.00001809, 0.00002080, -0.00001328, -0.00000114, -0.00001032, 0.00002086, -0.00001774, 0.00000026, -0.00000521, 0.00000939, -0.00000367, 0.00000194, -0.00000216, 0.00000148, -0.00000447, 0.00001019, -0.00000666, 0.00000709, -0.00000345, -0.00000802, -0.00000405, 0.00001018, -0.00000856, 0.00001047, -0.00000746, },
		{ -0.00000739, 0.00000379, -0.00000419, 0.00000301, -0.00000474, 0.00000643, -0.00000666, 0.00000081, -0.00000783, 0.00001150, -0.00000575, 0.00000953, -0.00000548, -0.00000288, -0.00000207, 0.00000794, -0.00000681, 0.00000474, -0.00000529, 0.00000368, -0.00000665, 0.00000749, -0.00000652, -0.00000379, 0.00000079, 0.00000718, -0.00000301, 0.00000917, -0.00000970, 0.00000348, -0.00000620, 0.00000727, -0.00000769, 0.00000654, -0.00000916, 0.00000612, -0.00000865, 0.00001282, -0.00001058, 0.00001634, -0.00001095, 0.00000002, -0.00000631, 0.00000532, -0.00000756, 0.00001164, -0.00000693, 0.00000254, -0.00000434, 0.00000284, -0.00000168, -0.00000041, -0.00000044, -0.00000155, -0.00000557, 0.00001448, -0.00001308, 0.00000194, -0.00000650, 0.00001028, -0.00001368, 0.00001465, -0.00001229, 0.00000967, -0.00001598, 0.00002706, -0.00002527, 0.00001415, -0.00002072, 0.00000736, -0.00000129, -0.00001452, 0.00000103, -0.00000820, 0.00000169, -0.00002996, -0.00001189, -0.00000562, -0.00000496, -0.00008624, -0.00001756, -0.00036736, 0.00023924, 0.00056186, 0.00034063, -0.00034949, -0.00007437, -0.00001861, -0.00002657, -0.00001118, -0.00002000, 0.00000089, 0.00001113, -0.00001599, 0.00000555, -0.00000233, 0.00000011, -0.00000134, -0.00002248, 0.00001502, -0.00001618, 0.00000801, -0.00002178, 0.00001093, -0.00001572, 0.00001981, -0.00001381, -0.00000217, -0.00000761, 0.00002015, -0.00001972, 0.00000166, -0.00000494, 0.00000948, -0.00000411, 0.00000212, -0.00000215, 0.00000095, -0.00000293, 0.00000896, -0.00000619, 0.00000696, -0.00000408, -0.00000863, -0.00000150, 0.00000808, -0.00000712, 0.00000958, },
		{ 0.00000829, -0.00000699, 0.00000328, -0.00000358, 0.00000246, -0.00000379, 0.00000551, -0.00000623, -0.00000009, -0.00000620, 0.00001077, -0.00000583, 0.00001019, -0.00000708, -0.00000223, -0.00000121, 0.00000732, -0.00000692, 0.00000498, -0.00000555, 0.00000389, -0.00000666, 0.00000795, -0.00000819, -0.00000257, 0.00000150, 0.00000629, -0.00000211, 0.00000812, -0.00000981, 0.00000377, -0.00000592, 0.00000706, -0.00000763, 0.00000635, -0.00000892, 0.00000594, -0.00000827, 0.00001278, -0.00001090, 0.00001761, -0.00001391, 0.00000239, -0.00000713, 0.00000617, -0.00000871, 0.00001352, -0.00000950, 0.00000472, -0.00000628, 0.00000448, -0.00000274, 0.00000052, -0.00000155, -0.00000183, -0.00000326, 0.00001281, -0.00001361, 0.00000217, -0.00000479, 0.00000924, -0.00001419, 0.00001517, -0.00001219, 0.00000922, -0.00001493, 0.00002726, -0.00002800, 0.00001993, -0.00002957, 0.00001651, -0.00000944, -0.00000768, -0.00000508, -0.00000256, -0.00000493, -0.00002251, -0.00002041, 0.00000343, -0.00001754, -0.00007137, -0.00004524, -0.00035269, 0.00032075, 0.00046395, 0.00035092, -0.00034249, -0.00006867, -0.00002287, -0.00002165, -0.00001417, -0.00001615, -0.00000199, 0.00001393, -0.00001739, 0.00000924, -0.00000786, 0.00000698, -0.00001088, -0.00001345, 0.00000774, -0.00001116, 0.00000422, -0.00001854, 0.00000745, -0.00001204, 0.00001741, -0.00001328, -0.00000386, -0.00000395, 0.00001804, -0.00002050, 0.00000243, -0.00000419, 0.00000904, -0.00000418, 0.00000210, -0.00000205, 0.00000040, -0.00000127, 0.00000743, -0.00000540, 0.00000656, -0.00000471, -0.00000891, 0.00000086, 0.00000580, -0.00000534, },
		{ -0.00000343, 0.00000675, -0.00000638, 0.00000266, -0.00000292, 0.00000189, -0.00000281, 0.00000448, -0.00000572, -0.00000100, -0.00000440, 0.00000964, -0.00000546, 0.00001046, -0.00000854, -0.00000140, -0.00000051, 0.00000664, -0.00000690, 0.00000515, -0.00000579, 0.00000410, -0.00000665, 0.00000830, -0.00000991, -0.00000091, 0.00000171, 0.00000574, -0.00000144, 0.00000709, -0.00000986, 0.00000412, -0.00000575, 0.00000688, -0.00000761, 0.00000617, -0.00000864, 0.00000571, -0.00000776, 0.00001248, -0.00001081, 0.00001835, -0.00001658, 0.00000487, -0.00000817, 0.00000712, -0.00000987, 0.00001537, -0.00001222, 0.00000728, -0.00000874, 0.00000677, -0.00000451, 0.00000222, -0.00000356, -0.00000106, -0.00000176, 0.00001144, -0.00001434, 0.00000287, -0.00000348, 0.00000824, -0.00001463, 0.00001571, -0.00001213, 0.00000868, -0.00001337, 0.00002638, -0.00002934, 0.00002427, -0.00003744, 0.00002550, -0.00001820, 0.00000023, -0.00001241, 0.00000435, -0.00001295, -0.00001343, -0.00003074, 0.00001460, -0.00003310, -0.00005203, -0.00008186, -0.00031830, 0.00037887, 0.00037596, 0.00036259, -0.00033729, -0.00006199, -0.00002798, -0.00001613, -0.00001757, -0.00001206, -0.00000483, 0.00001637, -0.00001802, 0.00001183, -0.00001214, 0.00001248, -0.00001963, -0.00000434, -0.00000015, -0.00000530, -0.00000042, -0.00001453, 0.00000328, -0.00000753, 0.00001402, -0.00001202, -0.00000588, 0.00000028, 0.00001481, -0.00002024, 0.00000266, -0.00000305, 0.00000817, -0.00000393, 0.00000194, -0.00000195, -0.00000003, 0.00000033, 0.00000579, -0.00000441, 0.00000598, -0.00000539, -0.00000877, 0.00000284, 0.00000358, },
		{ 0.00000163, -0.00000158, 0.00000514, -0.00000569, 0.00000206, -0.00000232, 0.00000144, -0.00000193, 0.00000345, -0.00000523, -0.00000180, -0.00000257, 0.00000824, -0.00000471, 0.00001034, -0.00000979, -0.00000043, 0.00000002, 0.00000594, -0.00000678, 0.00000524, -0.00000598, 0.00000430, -0.00000659, 0.00000848, -0.00001157, 0.00000108, 0.00000143, 0.00000555, -0.00000103, 0.00000615, -0.00000987, 0.00000453, -0.00000567, 0.00000676, -0.00000761, 0.00000601, -0.00000834, 0.00000543, -0.00000714, 0.00001189, -0.00001025, 0.00001846, -0.00001872, 0.00000723, -0.00000924, 0.00000800, -0.00001084, 0.00001696, -0.00001478, 0.00000989, -0.00001140, 0.00000941, -0.00000675, 0.00000443, -0.00000627, 0.00000064, -0.00000104, 0.00001036, -0.00001520, 0.00000396, -0.00000254, 0.00000725, -0.00001493, 0.00001617, -0.00001201, 0.00000799, -0.00001127, 0.00002430, -0.00002897, 0.00002663, -0.00004352, 0.00003332, -0.00002649, 0.00000815, -0.00001992, 0.00001145, -0.00002122, -0.00000399, -0.00004147, 0.00002619, -0.00004953, -0.00003092, -0.00012397, -0.00026709, 0.00041361, 0.00030060, 0.00037266, -0.00033179, -0.00005588, -0.00003271, -0.00001099, -0.00002056, -0.00000842, -0.00000705, 0.00001798, -0.00001743, 0.00001283, -0.00001460, 0.00001586, -0.00002666, 0.00000386, -0.00000775, 0.00000063, -0.00000525, -0.00001035, -0.00000103, -0.00000272, 0.00001005, -0.00001036, -0.00000788, 0.00000465, 0.00001085, -0.00001915, 0.00000248, -0.00000168, 0.00000697, -0.00000344, 0.00000171, -0.00000193, -0.00000025, 0.00000174, 0.00000416, -0.00000335, 0.00000527, -0.00000614, -0.00000815, 0.00000427, },
		{ 0.00000508, 0.00000010, 0.00000004, 0.00000360, -0.00000503, 0.00000154, -0.00000185, 0.00000118, -0.00000125, 0.00000254, -0.00000486, -0.00000238, -0.00000085, 0.00000669, -0.00000368, 0.00000985, -0.00001080, 0.00000063, 0.00000036, 0.00000524, -0.00000656, 0.00000525, -0.00000611, 0.00000447, -0.00000644, 0.00000844, -0.00001305, 0.00000326, 0.00000075, 0.00000570, -0.00000093, 0.00000533, -0.00000985, 0.00000499, -0.00000569, 0.00000668, -0.00000766, 0.00000588, -0.00000803, 0.00000512, -0.00000643, 0.00001107, -0.00000924, 0.00001785, -0.00002019, 0.00000925, -0.00001014, 0.00000865, -0.00001146, 0.00001807, -0.00001692, 0.00001226, -0.00001397, 0.00001210, -0.00000915, 0.00000690, -0.00000941, 0.00000307, -0.00000102, 0.00000955, -0.00001612, 0.00000534, -0.00000192, 0.00000625, -0.00001504, 0.00001646, -0.00001178, 0.00000715, -0.00000866, 0.00002104, -0.00002679, 0.00002667, -0.00004719, 0.00003913, -0.00003337, 0.00001514, -0.00002666, 0.00001776, -0.00002867, 0.00000463, -0.00005122, 0.00003657, -0.00006477, -0.00001086, -0.00016774, -0.00020294, 0.00042640, 0.00023930, 0.00037900, -0.00032442, -0.00005150, -0.00003611, -0.00000700, -0.00002255, -0.00000571, -0.00000827, 0.00001847, -0.00001539, 0.00001202, -0.00001495, 0.00001666, -0.00003127, 0.00001034, -0.00001427, 0.00000590, -0.00000964, -0.00000655, -0.00000495, 0.00000189, 0.00000594, -0.00000864, -0.00000949, 0.00000874, 0.00000653, -0.00001748, 0.00000206, -0.00000021, 0.00000559, -0.00000281, 0.00000147, -0.00000203, -0.00000019, 0.00000283, 0.00000268, -0.00000230, 0.00000450, -0.00000695, -0.00000704, },
		{ -0.00000547, 0.00000521, -0.00000089, 0.00000129, 0.00000225, -0.00000447, 0.00000118, -0.00000157, 0.00000117, -0.00000084, 0.00000182, -0.00000468, -0.00000265, 0.00000061, 0.00000515, -0.00000247, 0.00000903, -0.00001154, 0.00000174, 0.00000052, 0.00000458, -0.00000627, 0.00000517, -0.00000616, 0.00000458, -0.00000620, 0.00000813, -0.00001423, 0.00000550, -0.00000024, 0.00000616, -0.00000112, 0.00000467, -0.00000981, 0.00000548, -0.00000581, 0.00000667, -0.00000774, 0.00000580, -0.00000773, 0.00000480, -0.00000569, 0.00001008, -0.00000784, 0.00001656, -0.00002088, 0.00001078, -0.00001075, 0.00000895, -0.00001160, 0.00001854, -0.00001843, 0.00001412, -0.00001615, 0.00001455, -0.00001143, 0.00000931, -0.00001270, 0.00000600, -0.00000158, 0.00000894, -0.00001701, 0.00000693, -0.00000160, 0.00000525, -0.00001495, 0.00001657, -0.00001143, 0.00000624, -0.00000572, 0.00001679, -0.00002290, 0.00002433, -0.00004812, 0.00004238, -0.00003813, 0.00002043, -0.00003184, 0.00002245, -0.00003441, 0.00001140, -0.00005886, 0.00004435, -0.00007698, 0.00000555, -0.00020921, -0.00013035, 0.00041982, 0.00019230, 0.00038038, -0.00031419, -0.00004961, -0.00003757, -0.00000463, -0.00002315, -0.00000417, -0.00000833, 0.00001777, -0.00001191, 0.00000940, -0.00001321, 0.00001475, -0.00003307, 0.00001455, -0.00001911, 0.00000996, -0.00001309, -0.00000358, -0.00000804, 0.00000587, 0.00000207, -0.00000716, -0.00001044, 0.00001218, 0.00000225, -0.00001551, 0.00000157, 0.00000119, 0.00000415, -0.00000211, 0.00000126, -0.00000229, 0.00000020, 0.00000355, 0.00000145, -0.00000137, 0.00000369, -0.00000781, },
		{ -0.00000864, -0.00000353, 0.00000471, -0.00000132, 0.00000211, 0.00000117, -0.00000408, 0.00000101, -0.00000151, 0.00000145, -0.00000076, 0.00000135, -0.00000474, -0.00000253, 0.00000171, 0.00000375, -0.00000120, 0.00000796, -0.00001200, 0.00000285, 0.00000050, 0.00000398, -0.00000592, 0.00000500, -0.00000613, 0.00000462, -0.00000587, 0.00000752, -0.00001502, 0.00000764, -0.00000143, 0.00000684, -0.00000158, 0.00000417, -0.00000977, 0.00000599, -0.00000600, 0.00000671, -0.00000788, 0.00000579, -0.00000747, 0.00000451, -0.00000497, 0.00000902, -0.00000616, 0.00001464, -0.00002077, 0.00001174, -0.00001098, 0.00000886, -0.00001121, 0.00001829, -0.00001915, 0.00001528, -0.00001771, 0.00001649, -0.00001332, 0.00001140, -0.00001583, 0.00000916, -0.00000258, 0.00000848, -0.00001779, 0.00000862, -0.00000153, 0.00000428, -0.00001467, 0.00001651, -0.00001101, 0.00000538, -0.00000268, 0.00001186, -0.00001763, 0.00001980, -0.00004629, 0.00004279, -0.00004033, 0.00002350, -0.00003493, 0.00002496, -0.00003778, 0.00001563, -0.00006356, 0.00004848, -0.00008473, 0.00001609, -0.00024469, -0.00005414, 0.00039733, 0.00015879, 0.00037636, -0.00030070, -0.00005051, -0.00003683, -0.00000407, -0.00002227, -0.00000387, -0.00000727, 0.00001602, -0.00000719, 0.00000524, -0.00000964, 0.00001028, -0.00003201, 0.00001623, -0.00002190, 0.00001241, -0.00001525, -0.00000175, -0.00000998, 0.00000886, -0.00000124, -0.00000612, -0.00001052, 0.00001466, -0.00000166, -0.00001347, 0.00000118, 0.00000238, 0.00000277, -0.00000140, 0.00000111, -0.00000270, 0.00000092, 0.00000384, 0.00000052, -0.00000059, 0.00000287, },
		{ 0.00000204, -0.00000937, -0.00000136, 0.00000365, -0.00000119, 0.00000246, 0.00000038, -0.00000387, 0.00000103, -0.00000166, 0.00000200, -0.00000101, 0.00000114, -0.00000505, -0.00000202, 0.00000238, 0.00000260, 0.00000001, 0.00000670, -0.00001219, 0.00000395, 0.00000031, 0.00000348, -0.00000555, 0.00000476, -0.00000601, 0.00000460, -0.00000545, 0.00000662, -0.00001535, 0.00000956, -0.00000267, 0.00000765, -0.00000226, 0.00000382, -0.00000971, 0.00000650, -0.00000625, 0.00000680, -0.00000808, 0.00000586, -0.00000728, 0.00000432, -0.00000435, 0.00000800, -0.00000436, 0.00001224, -0.00001994, 0.00001211, -0.00001080, 0.00000838, -0.00001032, 0.00001732, -0.00001902, 0.00001562, -0.00001850, 0.00001774, -0.00001462, 0.00001294, -0.00001855, 0.00001230, -0.00000385, 0.00000809, -0.00001839, 0.00001033, -0.00000170, 0.00000336, -0.00001424, 0.00001631, -0.00001058, 0.00000471, 0.00000016, 0.00000668, -0.00001143, 0.00001350, -0.00004195, 0.00004040, -0.00003983, 0.00002411, -0.00003566, 0.00002501, -0.00003848, 0.00001693, -0.00006486, 0.00004838, -0.00008714, 0.00001918, -0.00027102, 0.00002097, 0.00036293, 0.00013714, 0.00036725, -0.00028409, -0.00005408, -0.00003398, -0.00000523, -0.00002000, -0.00000464, -0.00000530, 0.00001352, -0.00000163, -0.00000003, -0.00000474, 0.00000372, -0.00002831, 0.00001537, -0.00002252, 0.00001309, -0.00001595, -0.00000124, -0.00001060, 0.00001066, -0.00000380, -0.00000566, -0.00000963, 0.00001598, -0.00000491, -0.00001160, 0.00000101, 0.00000325, 0.00000155, -0.00000076, 0.00000103, -0.00000323, 0.00000194, 0.00000370, -0.00000006, -0.00000002, },
		{ 0.00000034, 0.00000120, -0.00000992, 0.00000090, 0.00000217, -0.00000058, 0.00000238, -0.00000010, -0.00000383, 0.00000120, -0.00000196, 0.00000277, -0.00000155, 0.00000117, -0.00000558, -0.00000113, 0.00000257, 0.00000180, 0.00000105, 0.00000535, -0.00001213, 0.00000499, -0.00000005, 0.00000310, -0.00000519, 0.00000448, -0.00000583, 0.00000453, -0.00000497, 0.00000546, -0.00001518, 0.00001115, -0.00000386, 0.00000851, -0.00000308, 0.00000361, -0.00000964, 0.00000699, -0.00000655, 0.00000694, -0.00000833, 0.00000602, -0.00000719, 0.00000424, -0.00000391, 0.00000714, -0.00000260, 0.00000953, -0.00001849, 0.00001193, -0.00001026, 0.00000755, -0.00000899, 0.00001568, -0.00001807, 0.00001511, -0.00001843, 0.00001820, -0.00001518, 0.00001375, -0.00002062, 0.00001518, -0.00000523, 0.00000770, -0.00001874, 0.00001198, -0.00000208, 0.00000254, -0.00001373, 0.00001602, -0.00001021, 0.00000439, 0.00000251, 0.00000170, -0.00000488, 0.00000604, -0.00003561, 0.00003552, -0.00003677, 0.00002233, -0.00003407, 0.00002260, -0.00003650, 0.00001527, -0.00006275, 0.00004397, -0.00008395, 0.00001401, -0.00028579, 0.00009067, 0.00032078, 0.00012515, 0.00035387, -0.00026487, -0.00005991, -0.00002934, -0.00000782, -0.00001664, -0.00000620, -0.00000275, 0.00001068, 0.00000429, -0.00000586, 0.00000082, -0.00000426, -0.00002247, 0.00001229, -0.00002109, 0.00001204, -0.00001519, -0.00000204, -0.00000987, 0.00001119, -0.00000552, -0.00000577, -0.00000777, 0.00001605, -0.00000734, -0.00001003, 0.00000116, 0.00000371, 0.00000058, -0.00000023, 0.00000101, -0.00000385, 0.00000320, 0.00000317, -0.00000031, },
		{ -0.00000024, 0.00000049, 0.00000033, -0.00001022, 0.00000310, 0.00000042, 0.00000041, 0.00000191, -0.00000031, -0.00000392, 0.00000147, -0.00000235, 0.00000369, -0.00000232, 0.00000141, -0.00000627, 0.00000009, 0.00000230, 0.00000140, 0.00000185, 0.00000400, -0.00001185, 0.00000594, -0.00000054, 0.00000286, -0.00000486, 0.00000419, -0.00000561, 0.00000443, -0.00000448, 0.00000408, -0.00001451, 0.00001232, -0.00000489, 0.00000930, -0.00000399, 0.00000351, -0.00000952, 0.00000742, -0.00000686, 0.00000711, -0.00000863, 0.00000627, -0.00000721, 0.00000432, -0.00000369, 0.00000654, -0.00000105, 0.00000673, -0.00001660, 0.00001131, -0.00000943, 0.00000649, -0.00000736, 0.00001352, -0.00001640, 0.00001381, -0.00001753, 0.00001782, -0.00001494, 0.00001374, -0.00002190, 0.00001760, -0.00000656, 0.00000724, -0.00001877, 0.00001348, -0.00000262, 0.00000184, -0.00001319, 0.00001570, -0.00000997, 0.00000454, 0.00000409, -0.00000261, 0.00000138, -0.00000186, -0.00002796, 0.00002872, -0.00003156, 0.00001846, -0.00003044, 0.00001806, -0.00003216, 0.00001100, -0.00005759, 0.00003564, -0.00007553, 0.00000058, -0.00028756, 0.00015137, 0.00027495, 0.00012038, 0.00033741, -0.00024387, -0.00006734, -0.00002343, -0.00001140, -0.00001259, -0.00000815, -0.00000001, 0.00000793, 0.00001003, -0.00001163, 0.00000631, -0.00001286, -0.00001520, 0.00000746, -0.00001796, 0.00000949, -0.00001317, -0.00000397, -0.00000792, 0.00001050, -0.00000643, -0.00000639, -0.00000508, 0.00001490, -0.00000883, -0.00000888, 0.00000168, 0.00000372, -0.00000009, 0.00000018, 0.00000104, -0.00000450, 0.00000462, 0.00000231, },
		{ 0.00000121, 0.00000009, 0.00000044, -0.00000057, -0.00001020, 0.00000508, -0.00000143, 0.00000165, 0.00000115, -0.00000031, -0.00000410, 0.00000179, -0.00000275, 0.00000466, -0.00000324, 0.00000176, -0.00000705, 0.00000155, 0.00000162, 0.00000139, 0.00000233, 0.00000271, -0.00001138, 0.00000679, -0.00000115, 0.00000276, -0.00000461, 0.00000391, -0.00000536, 0.00000434, -0.00000404, 0.00000257, -0.00001337, 0.00001303, -0.00000569, 0.00000995, -0.00000489, 0.00000347, -0.00000936, 0.00000778, -0.00000715, 0.00000728, -0.00000895, 0.00000661, -0.00000734, 0.00000456, -0.00000374, 0.00000630, 0.00000013, 0.00000405, -0.00001445, 0.00001039, -0.00000842, 0.00000533, -0.00000560, 0.00001103, -0.00001419, 0.00001185, -0.00001589, 0.00001667, -0.00001393, 0.00001288, -0.00002229, 0.00001941, -0.00000772, 0.00000666, -0.00001846, 0.00001477, -0.00000329, 0.00000130, -0.00001266, 0.00001539, -0.00000988, 0.00000523, 0.00000469, -0.00000585, 0.00000677, -0.00000946, -0.00001979, 0.00002072, -0.00002481, 0.00001302, -0.00002529, 0.00001190, -0.00002601, 0.00000472, -0.00005009, 0.00002425, -0.00006281, -0.00002029, -0.00027591, 0.00020035, 0.00022906, 0.00012030, 0.00031921, -0.00022205, -0.00007561, -0.00001687, -0.00001545, -0.00000827, -0.00001011, 0.00000253, 0.00000571, 0.00001511, -0.00001678, 0.00001102, -0.00002122, -0.00000729, 0.00000156, -0.00001361, 0.00000581, -0.00001018, -0.00000678, -0.00000500, 0.00000877, -0.00000663, -0.00000735, -0.00000177, 0.00001268, -0.00000942, -0.00000815, 0.00000255, 0.00000328, -0.00000046, 0.00000045, 0.00000110, -0.00000510, 0.00000610, },
		{ 0.00000754, -0.00000006, 0.00000064, 0.00000020, -0.00000148, -0.00000982, 0.00000673, -0.00000323, 0.00000299, 0.00000019, -0.00000016, -0.00000431, 0.00000207, -0.00000308, 0.00000558, -0.00000422, 0.00000216, -0.00000781, 0.00000313, 0.00000061, 0.00000177, 0.00000247, 0.00000156, -0.00001076, 0.00000752, -0.00000185, 0.00000280, -0.00000445, 0.00000368, -0.00000514, 0.00000428, -0.00000369, 0.00000102, -0.00001185, 0.00001328, -0.00000619, 0.00001038, -0.00000574, 0.00000347, -0.00000913, 0.00000804, -0.00000740, 0.00000741, -0.00000926, 0.00000700, -0.00000758, 0.00000495, -0.00000405, 0.00000645, 0.00000081, 0.00000168, -0.00001226, 0.00000933, -0.00000738, 0.00000422, -0.00000392, 0.00000843, -0.00001166, 0.00000944, -0.00001367, 0.00001488, -0.00001226, 0.00001124, -0.00002178, 0.00002052, -0.00000862, 0.00000591, -0.00001778, 0.00001579, -0.00000404, 0.00000091, -0.00001217, 0.00001510, -0.00000994, 0.00000647, 0.00000420, -0.00000773, 0.00001079, -0.00001604, -0.00001191, 0.00001231, -0.00001723, 0.00000667, -0.00001925, 0.00000480, -0.00001879, -0.00000278, -0.00004120, 0.00001092, -0.00004720, -0.00004706, -0.00025145, 0.00023594, 0.00018614, 0.00012261, 0.00030061, -0.00020038, -0.00008393, -0.00001026, -0.00001948, -0.00000411, -0.00001171, 0.00000454, 0.00000437, 0.00001910, -0.00002086, 0.00001433, -0.00002855, 0.00000043, -0.00000472, -0.00000863, 0.00000147, -0.00000663, -0.00001007, -0.00000144, 0.00000624, -0.00000632, -0.00000842, 0.00000187, 0.00000960, -0.00000918, -0.00000781, 0.00000371, 0.00000243, -0.00000052, 0.00000059, 0.00000116, -0.00000560, },
		{ -0.00000593, 0.00000886, -0.00000139, 0.00000134, -0.00000021, -0.00000240, -0.00000908, 0.00000796, -0.00000483, 0.00000430, -0.00000085, 0.00000007, -0.00000449, 0.00000227, -0.00000326, 0.00000636, -0.00000514, 0.00000251, -0.00000845, 0.00000471, -0.00000062, 0.00000247, 0.00000226, 0.00000059, -0.00001002, 0.00000811, -0.00000261, 0.00000299, -0.00000441, 0.00000352, -0.00000496, 0.00000429, -0.00000351, -0.00000046, -0.00001003, 0.00001310, -0.00000638, 0.00001056, -0.00000646, 0.00000348, -0.00000883, 0.00000817, -0.00000756, 0.00000749, -0.00000954, 0.00000742, -0.00000788, 0.00000547, -0.00000461, 0.00000701, 0.00000093, -0.00000021, -0.00001021, 0.00000826, -0.00000642, 0.00000331, -0.00000249, 0.00000596, -0.00000905, 0.00000681, -0.00001109, 0.00001263, -0.00001008, 0.00000897, -0.00002044, 0.00002090, -0.00000918, 0.00000500, -0.00001671, 0.00001647, -0.00000481, 0.00000068, -0.00001175, 0.00001483, -0.00001010, 0.00000819, 0.00000262, -0.00000809, 0.00001307, -0.00002103, -0.00000507, 0.00000429, -0.00000958, 0.00000011, -0.00001303, -0.00000248, -0.00001130, -0.00001058, -0.00003198, -0.00000301, -0.00003041, -0.00007760, -0.00021570, 0.00025752, 0.00014844, 0.00012536, 0.00028280, -0.00017974, -0.00009161, -0.00000416, -0.00002304, -0.00000046, -0.00001267, 0.00000578, 0.00000415, 0.00002175, -0.00002356, 0.00001580, -0.00003419, 0.00000722, -0.00001067, -0.00000362, -0.00000302, -0.00000296, -0.00001346, 0.00000239, 0.00000321, -0.00000570, -0.00000936, 0.00000552, 0.00000597, -0.00000829, -0.00000775, 0.00000506, 0.00000126, -0.00000032, 0.00000062, 0.00000119, },
		{ 0.00000116, -0.00000605, 0.00000995, -0.00000269, 0.00000212, -0.00000075, -0.00000328, -0.00000800, 0.00000871, -0.00000613, 0.00000547, -0.00000188, 0.00000033, -0.00000460, 0.00000235, -0.00000325, 0.00000693, -0.00000594, 0.00000274, -0.00000888, 0.00000616, -0.00000194, 0.00000340, 0.00000174, -0.00000017, -0.00000920, 0.00000854, -0.00000338, 0.00000330, -0.00000447, 0.00000346, -0.00000484, 0.00000439, -0.00000353, -0.00000176, -0.00000804, 0.00001256, -0.00000628, 0.00001046, -0.00000703, 0.00000346, -0.00000844, 0.00000818, -0.00000763, 0.00000749, -0.00000973, 0.00000782, -0.00000822, 0.00000607, -0.00000535, 0.00000794, 0.00000046, -0.00000152, -0.00000846, 0.00000734, -0.00000567, 0.00000272, -0.00000148, 0.00000384, -0.00000661, 0.00000421, -0.00000840, 0.00001016, -0.00000760, 0.00000624, -0.00001841, 0.00002058, -0.00000941, 0.00000393, -0.00001528, 0.00001679, -0.00000555, 0.00000057, -0.00001138, 0.00001455, -0.00001028, 0.00001025, 0.00000007, -0.00000695, 0.00001344, -0.00002403, 0.00000013, -0.00000265, -0.00000254, -0.00000599, -0.00000730, -0.00000922, -0.00000433, -0.00001780, -0.00002350, -0.00001618, -0.00001428, -0.00010943, -0.00017095, 0.00026547, 0.00011739, 0.00012706, 0.00026669, -0.00016079, -0.00009810, 0.00000102, -0.00002581, 0.00000245, -0.00001283, 0.00000617, 0.00000512, 0.00002294, -0.00002479, 0.00001522, -0.00003767, 0.00001247, -0.00001568, 0.00000087, -0.00000720, 0.00000044, -0.00001655, 0.00000610, -0.00000000, -0.00000501, -0.00000994, 0.00000887, 0.00000209, -0.00000694, -0.00000785, 0.00000648, -0.00000012, 0.00000009, 0.00000056, },
		{ 0.00000046, 0.00000107, -0.00000591, 0.00001076, -0.00000387, 0.00000292, -0.00000142, -0.00000409, -0.00000664, 0.00000898, -0.00000706, 0.00000642, -0.00000285, 0.00000056, -0.00000462, 0.00000228, -0.00000300, 0.00000723, -0.00000655, 0.00000279, -0.00000903, 0.00000737, -0.00000322, 0.00000447, 0.00000095, -0.00000070, -0.00000833, 0.00000881, -0.00000413, 0.00000371, -0.00000465, 0.00000349, -0.00000481, 0.00000459, -0.00000381, -0.00000278, -0.00000601, 0.00001174, -0.00000593, 0.00001010, -0.00000742, 0.00000341, -0.00000798, 0.00000804, -0.00000758, 0.00000738, -0.00000982, 0.00000815, -0.00000854, 0.00000668, -0.00000621, 0.00000915, -0.00000057, -0.00000220, -0.00000711, 0.00000666, -0.00000519, 0.00000253, -0.00000101, 0.00000225, -0.00000457, 0.00000189, -0.00000584, 0.00000770, -0.00000507, 0.00000329, -0.00001586, 0.00001965, -0.00000931, 0.00000273, -0.00001353, 0.00001672, -0.00000620, 0.00000054, -0.00001104, 0.00001422, -0.00001035, 0.00001244, -0.00000321, -0.00000446, 0.00001191, -0.00002483, 0.00000329, -0.00000796, 0.00000329, -0.00001108, -0.00000264, -0.00001480, 0.00000143, -0.00002368, -0.00001669, -0.00002735, -0.00000058, -0.00013994, -0.00012005, 0.00026104, 0.00009361, 0.00012675, 0.00025288, -0.00014398, -0.00010304, 0.00000501, -0.00002762, 0.00000451, -0.00001216, 0.00000574, 0.00000722, 0.00002275, -0.00002464, 0.00001260, -0.00003878, 0.00001578, -0.00001929, 0.00000440, -0.00001067, 0.00000320, -0.00001900, 0.00000936, -0.00000313, -0.00000442, -0.00000996, 0.00001161, -0.00000172, -0.00000537, -0.00000796, 0.00000781, -0.00000160, 0.00000064, },
		{ 0.00000125, 0.00000033, 0.00000090, -0.00000552, 0.00001123, -0.00000486, 0.00000368, -0.00000218, -0.00000477, -0.00000507, 0.00000880, -0.00000760, 0.00000709, -0.00000369, 0.00000075, -0.00000456, 0.00000207, -0.00000253, 0.00000724, -0.00000693, 0.00000264, -0.00000885, 0.00000826, -0.00000436, 0.00000555, -0.00000004, -0.00000102, -0.00000743, 0.00000890, -0.00000483, 0.00000417, -0.00000492, 0.00000363, -0.00000486, 0.00000490, -0.00000435, -0.00000344, -0.00000407, 0.00001075, -0.00000541, 0.00000952, -0.00000763, 0.00000333, -0.00000746, 0.00000779, -0.00000741, 0.00000714, -0.00000977, 0.00000839, -0.00000878, 0.00000724, -0.00000708, 0.00001054, -0.00000208, -0.00000224, -0.00000622, 0.00000627, -0.00000502, 0.00000278, -0.00000115, 0.00000130, -0.00000309, 0.00000005, -0.00000363, 0.00000548, -0.00000271, 0.00000037, -0.00001301, 0.00001824, -0.00000894, 0.00000147, -0.00001151, 0.00001627, -0.00000671, 0.00000056, -0.00001070, 0.00001377, -0.00001019, 0.00001451, -0.00000692, -0.00000093, 0.00000871, -0.00002347, 0.00000421, -0.00001129, 0.00000752, -0.00001475, 0.00000054, -0.00001876, 0.00000547, -0.00002763, -0.00001227, -0.00003555, 0.00000918, -0.00016658, -0.00006612, 0.00024617, 0.00007701, 0.00012398, 0.00024162, -0.00012949, -0.00010629, 0.00000771, -0.00002841, 0.00000575, -0.00001075, 0.00000463, 0.00001025, 0.00002138, -0.00002338, 0.00000819, -0.00003756, 0.00001700, -0.00002123, 0.00000667, -0.00001315, 0.00000508, -0.00002057, 0.00001191, -0.00000593, -0.00000409, -0.00000929, 0.00001353, -0.00000517, -0.00000378, -0.00000794, 0.00000893, -0.00000305, },
		{ -0.00000436, 0.00000184, 0.00000020, 0.00000067, -0.00000488, 0.00001134, -0.00000562, 0.00000434, -0.00000303, -0.00000528, -0.00000341, 0.00000825, -0.00000776, 0.00000747, -0.00000438, 0.00000089, -0.00000443, 0.00000176, -0.00000187, 0.00000698, -0.00000709, 0.00000227, -0.00000834, 0.00000877, -0.00000524, 0.00000656, -0.00000116, -0.00000112, -0.00000652, 0.00000880, -0.00000542, 0.00000465, -0.00000526, 0.00000385, -0.00000498, 0.00000530, -0.00000516, -0.00000369, -0.00000233, 0.00000971, -0.00000480, 0.00000877, -0.00000769, 0.00000322, -0.00000691, 0.00000743, -0.00000714, 0.00000679, -0.00000957, 0.00000848, -0.00000891, 0.00000767, -0.00000785, 0.00001197, -0.00000392, -0.00000171, -0.00000578, 0.00000619, -0.00000516, 0.00000343, -0.00000190, 0.00000106, -0.00000228, -0.00000118, -0.00000194, 0.00000369, -0.00000073, -0.00000230, -0.00001009, 0.00001649, -0.00000838, 0.00000021, -0.00000930, 0.00001544, -0.00000704, 0.00000058, -0.00001033, 0.00001315, -0.00000966, 0.00001621, -0.00001067, 0.00000322, 0.00000422, -0.00002020, 0.00000295, -0.00001253, 0.00000995, -0.00001678, 0.00000203, -0.00002087, 0.00000751, -0.00002936, -0.00001063, -0.00004016, 0.00001389, -0.00018714, -0.00001233, 0.00022325, 0.00006688, 0.00011874, 0.00023287, -0.00011728, -0.00010788, 0.00000921, -0.00002828, 0.00000628, -0.00000877, 0.00000312, 0.00001391, 0.00001915, -0.00002144, 0.00000244, -0.00003428, 0.00001617, -0.00002144, 0.00000756, -0.00001451, 0.00000595, -0.00002112, 0.00001357, -0.00000825, -0.00000409, -0.00000791, 0.00001448, -0.00000805, -0.00000235, -0.00000767, 0.00000971, },
		{ 0.00001005, -0.00000542, 0.00000237, 0.00000010, 0.00000038, -0.00000402, 0.00001109, -0.00000612, 0.00000488, -0.00000394, -0.00000558, -0.00000176, 0.00000741, -0.00000758, 0.00000756, -0.00000492, 0.00000100, -0.00000426, 0.00000141, -0.00000108, 0.00000648, -0.00000704, 0.00000173, -0.00000752, 0.00000887, -0.00000582, 0.00000740, -0.00000233, -0.00000104, -0.00000563, 0.00000853, -0.00000588, 0.00000512, -0.00000563, 0.00000413, -0.00000516, 0.00000576, -0.00000619, -0.00000348, -0.00000089, 0.00000871, -0.00000418, 0.00000792, -0.00000762, 0.00000311, -0.00000636, 0.00000701, -0.00000678, 0.00000632, -0.00000921, 0.00000841, -0.00000888, 0.00000792, -0.00000844, 0.00001329, -0.00000595, -0.00000074, -0.00000575, 0.00000637, -0.00000554, 0.00000443, -0.00000320, 0.00000151, -0.00000218, -0.00000170, -0.00000089, 0.00000246, 0.00000070, -0.00000450, -0.00000732, 0.00001459, -0.00000772, -0.00000096, -0.00000700, 0.00001429, -0.00000717, 0.00000057, -0.00000990, 0.00001233, -0.00000866, 0.00001729, -0.00001409, 0.00000752, -0.00000107, -0.00001545, -0.00000022, -0.00001178, 0.00001058, -0.00001717, 0.00000181, -0.00002110, 0.00000753, -0.00002881, -0.00001186, -0.00004097, 0.00001294, -0.00019993, 0.00003835, 0.00019489, 0.00006211, 0.00011141, 0.00022632, -0.00010707, -0.00010802, 0.00000966, -0.00002741, 0.00000631, -0.00000648, 0.00000150, 0.00001784, 0.00001643, -0.00001930, -0.00000406, -0.00002941, 0.00001358, -0.00002002, 0.00000708, -0.00001475, 0.00000581, -0.00002062, 0.00001427, -0.00001003, -0.00000439, -0.00000584, 0.00001443, -0.00001022, -0.00000121, -0.00000707, },
		{ -0.00000610, 0.00000993, -0.00000617, 0.00000276, 0.00000003, 0.00000008, -0.00000299, 0.00001050, -0.00000633, 0.00000525, -0.00000488, -0.00000563, -0.00000021, 0.00000640, -0.00000715, 0.00000740, -0.00000533, 0.00000111, -0.00000412, 0.00000108, -0.00000023, 0.00000582, -0.00000683, 0.00000107, -0.00000645, 0.00000860, -0.00000605, 0.00000800, -0.00000348, -0.00000080, -0.00000477, 0.00000810, -0.00000618, 0.00000552, -0.00000600, 0.00000444, -0.00000537, 0.00000623, -0.00000741, -0.00000284, 0.00000019, 0.00000784, -0.00000364, 0.00000704, -0.00000748, 0.00000303, -0.00000584, 0.00000655, -0.00000637, 0.00000577, -0.00000870, 0.00000817, -0.00000867, 0.00000795, -0.00000875, 0.00001437, -0.00000799, 0.00000055, -0.00000605, 0.00000676, -0.00000609, 0.00000564, -0.00000493, 0.00000257, -0.00000275, -0.00000151, -0.00000053, 0.00000187, 0.00000147, -0.00000609, -0.00000490, 0.00001268, -0.00000706, -0.00000196, -0.00000470, 0.00001286, -0.00000709, 0.00000051, -0.00000940, 0.00001129, -0.00000712, 0.00001755, -0.00001682, 0.00001149, -0.00000657, -0.00000978, -0.00000487, -0.00000934, 0.00000962, -0.00001610, 0.00000008, -0.00001964, 0.00000572, -0.00002623, -0.00001568, -0.00003821, 0.00000629, -0.00020385, 0.00008341, 0.00016372, 0.00006132, 0.00010258, 0.00022148, -0.00009849, -0.00010702, 0.00000936, -0.00002604, 0.00000610, -0.00000413, 0.00000010, 0.00002168, 0.00001360, -0.00001749, -0.00001066, -0.00002356, 0.00000965, -0.00001727, 0.00000544, -0.00001400, 0.00000479, -0.00001917, 0.00001405, -0.00001126, -0.00000493, -0.00000322, 0.00001342, -0.00001162, -0.00000043, },
		{ 0.00000000, -0.00000479, 0.00000932, -0.00000658, 0.00000301, -0.00000000, -0.00000020, -0.00000184, 0.00000962, -0.00000627, 0.00000543, -0.00000582, -0.00000542, 0.00000115, 0.00000532, -0.00000654, 0.00000704, -0.00000562, 0.00000123, -0.00000403, 0.00000083, 0.00000058, 0.00000507, -0.00000655, 0.00000036, -0.00000521, 0.00000800, -0.00000593, 0.00000832, -0.00000455, -0.00000042, -0.00000396, 0.00000752, -0.00000631, 0.00000582, -0.00000633, 0.00000475, -0.00000557, 0.00000666, -0.00000873, -0.00000180, 0.00000086, 0.00000718, -0.00000325, 0.00000621, -0.00000729, 0.00000299, -0.00000539, 0.00000611, -0.00000594, 0.00000517, -0.00000806, 0.00000776, -0.00000828, 0.00000774, -0.00000874, 0.00001509, -0.00000986, 0.00000198, -0.00000654, 0.00000725, -0.00000667, 0.00000693, -0.00000692, 0.00000409, -0.00000389, -0.00000067, -0.00000082, 0.00000193, 0.00000156, -0.00000697, -0.00000297, 0.00001092, -0.00000648, -0.00000270, -0.00000250, 0.00001123, -0.00000682, 0.00000039, -0.00000884, 0.00001007, -0.00000505, 0.00001687, -0.00001855, 0.00001467, -0.00001172, -0.00000381, -0.00001042, -0.00000564, 0.00000741, -0.00001389, -0.00000282, -0.00001685, 0.00000246, -0.00002208, -0.00002155, -0.00003252, -0.00000552, -0.00019852, 0.00012088, 0.00013212, 0.00006306, 0.00009300, 0.00021778, -0.00009108, -0.00010523, 0.00000858, -0.00002443, 0.00000587, -0.00000198, -0.00000080, 0.00002509, 0.00001099, -0.00001644, -0.00001669, -0.00001738, 0.00000492, -0.00001359, 0.00000293, -0.00001250, 0.00000308, -0.00001695, 0.00001301, -0.00001202, -0.00000558, -0.00000026, 0.00001162, -0.00001228, },
		{ -0.00001231, 0.00000013, -0.00000321, 0.00000828, -0.00000664, 0.00000309, -0.00000001, -0.00000042, -0.00000066, 0.00000850, -0.00000595, 0.00000542, -0.00000672, -0.00000495, 0.00000224, 0.00000428, -0.00000584, 0.00000653, -0.00000583, 0.00000140, -0.00000406, 0.00000074, 0.00000127, 0.00000433, -0.00000626, -0.00000030, -0.00000391, 0.00000717, -0.00000552, 0.00000835, -0.00000549, 0.00000006, -0.00000323, 0.00000681, -0.00000627, 0.00000600, -0.00000657, 0.00000502, -0.00000573, 0.00000700, -0.00001007, -0.00000043, 0.00000115, 0.00000675, -0.00000307, 0.00000549, -0.00000711, 0.00000303, -0.00000505, 0.00000572, -0.00000554, 0.00000457, -0.00000735, 0.00000722, -0.00000774, 0.00000729, -0.00000838, 0.00001536, -0.00001143, 0.00000339, -0.00000711, 0.00000774, -0.00000718, 0.00000813, -0.00000899, 0.00000588, -0.00000543, 0.00000069, -0.00000168, 0.00000257, 0.00000097, -0.00000710, -0.00000164, 0.00000943, -0.00000607, -0.00000313, -0.00000052, 0.00000949, -0.00000639, 0.00000022, -0.00000824, 0.00000871, -0.00000249, 0.00001520, -0.00001911, 0.00001671, -0.00001599, 0.00000183, -0.00001626, -0.00000121, 0.00000440, -0.00001098, -0.00000642, -0.00001322, -0.00000170, -0.00001698, -0.00002870, -0.00002483, -0.00002147, -0.00018427, 0.00014948, 0.00010214, 0.00006595, 0.00008342, 0.00021464, -0.00008436, -0.00010302, 0.00000764, -0.00002282, 0.00000583, -0.00000022, -0.00000095, 0.00002782, 0.00000883, -0.00001651, -0.00002160, -0.00001151, -0.00000002, -0.00000945, -0.00000010, -0.00001056, 0.00000097, -0.00001420, 0.00001133, -0.00001242, -0.00000616, 0.00000281, 0.00000921, },
		{ 0.00000648, -0.00001187, 0.00000007, -0.00000147, 0.00000690, -0.00000639, 0.00000302, -0.00000002, -0.00000053, 0.00000048, 0.00000722, -0.00000541, 0.00000519, -0.00000755, -0.00000425, 0.00000303, 0.00000336, -0.00000511, 0.00000592, -0.00000598, 0.00000162, -0.00000421, 0.00000086, 0.00000176, 0.00000367, -0.00000603, -0.00000083, -0.00000266, 0.00000621, -0.00000487, 0.00000810, -0.00000627, 0.00000062, -0.00000260, 0.00000603, -0.00000606, 0.00000605, -0.00000672, 0.00000521, -0.00000581, 0.00000719, -0.00001132, 0.00000118, 0.00000107, 0.00000656, -0.00000312, 0.00000492, -0.00000697, 0.00000317, -0.00000484, 0.00000543, -0.00000522, 0.00000402, -0.00000661, 0.00000659, -0.00000706, 0.00000665, -0.00000768, 0.00001513, -0.00001257, 0.00000464, -0.00000763, 0.00000813, -0.00000750, 0.00000909, -0.00001094, 0.00000775, -0.00000719, 0.00000239, -0.00000297, 0.00000368, -0.00000018, -0.00000653, -0.00000095, 0.00000827, -0.00000586, -0.00000318, 0.00000116, 0.00000773, -0.00000584, 0.00000002, -0.00000767, 0.00000732, 0.00000043, 0.00001262, -0.00001842, 0.00001737, -0.00001897, 0.00000658, -0.00002179, 0.00000340, 0.00000107, -0.00000785, -0.00001023, -0.00000930, -0.00000617, -0.00001165, -0.00003622, -0.00001630, -0.00004011, -0.00016209, 0.00016867, 0.00007535, 0.00006880, 0.00007450, 0.00021152, -0.00007791, -0.00010071, 0.00000677, -0.00002139, 0.00000613, 0.00000105, -0.00000024, 0.00002970, 0.00000723, -0.00001788, -0.00002496, -0.00000653, -0.00000465, -0.00000531, -0.00000324, -0.00000848, -0.00000124, -0.00001121, 0.00000921, -0.00001261, -0.00000651, 0.00000572, },
		{ 0.00000821, 0.00000368, -0.00001113, -0.00000005, 0.00000028, 0.00000528, -0.00000587, 0.00000281, -0.00000003, -0.00000048, 0.00000150, 0.00000585, -0.00000470, 0.00000477, -0.00000825, -0.00000337, 0.00000350, 0.00000262, -0.00000444, 0.00000526, -0.00000609, 0.00000191, -0.00000449, 0.00000120, 0.00000201, 0.00000317, -0.00000593, -0.00000116, -0.00000156, 0.00000522, -0.00000406, 0.00000761, -0.00000689, 0.00000126, -0.00000210, 0.00000520, -0.00000571, 0.00000595, -0.00000675, 0.00000531, -0.00000578, 0.00000719, -0.00001240, 0.00000292, 0.00000069, 0.00000658, -0.00000339, 0.00000453, -0.00000689, 0.00000341, -0.00000479, 0.00000527, -0.00000501, 0.00000358, -0.00000590, 0.00000591, -0.00000632, 0.00000588, -0.00000669, 0.00001443, -0.00001322, 0.00000562, -0.00000802, 0.00000833, -0.00000753, 0.00000968, -0.00001259, 0.00000948, -0.00000895, 0.00000425, -0.00000451, 0.00000510, -0.00000177, -0.00000534, -0.00000088, 0.00000748, -0.00000589, -0.00000283, 0.00000247, 0.00000603, -0.00000523, -0.00000015, -0.00000717, 0.00000601, 0.00000354, 0.00000928, -0.00001654, 0.00001657, -0.00002038, 0.00001000, -0.00002647, 0.00000766, -0.00000209, -0.00000495, -0.00001376, -0.00000560, -0.00001037, -0.00000679, -0.00004323, -0.00000819, -0.00005975, -0.00013349, 0.00017857, 0.00005278, 0.00007070, 0.00006676, 0.00020803, -0.00007141, -0.00009854, 0.00000616, -0.00002024, 0.00000682, 0.00000178, 0.00000138, 0.00003068, 0.00000615, -0.00002055, -0.00002652, -0.00000285, -0.00000849, -0.00000159, -0.00000616, -0.00000658, -0.00000327, -0.00000825, 0.00000688, -0.00001271, -0.00000649, },
		{ -0.00000597, 0.00001009, 0.00000107, -0.00001030, -0.00000008, 0.00000191, 0.00000356, -0.00000518, 0.00000251, -0.00000007, -0.00000025, 0.00000233, 0.00000447, -0.00000387, 0.00000417, -0.00000879, -0.00000236, 0.00000367, 0.00000208, -0.00000385, 0.00000459, -0.00000615, 0.00000224, -0.00000489, 0.00000176, 0.00000197, 0.00000286, -0.00000601, -0.00000124, -0.00000070, 0.00000432, -0.00000320, 0.00000693, -0.00000734, 0.00000195, -0.00000175, 0.00000439, -0.00000525, 0.00000574, -0.00000667, 0.00000531, -0.00000565, 0.00000696, -0.00001322, 0.00000467, 0.00000010, 0.00000678, -0.00000388, 0.00000432, -0.00000689, 0.00000375, -0.00000488, 0.00000524, -0.00000495, 0.00000327, -0.00000526, 0.00000525, -0.00000557, 0.00000504, -0.00000551, 0.00001328, -0.00001334, 0.00000626, -0.00000819, 0.00000830, -0.00000723, 0.00000980, -0.00001378, 0.00001089, -0.00001052, 0.00000606, -0.00000611, 0.00000664, -0.00000361, -0.00000369, -0.00000136, 0.00000706, -0.00000615, -0.00000209, 0.00000336, 0.00000447, -0.00000462, -0.00000027, -0.00000683, 0.00000493, 0.00000661, 0.00000542, -0.00001365, 0.00001439, -0.00002014, 0.00001181, -0.00002990, 0.00001114, -0.00000471, -0.00000267, -0.00001659, -0.00000259, -0.00001380, -0.00000300, -0.00004889, -0.00000168, -0.00007859, -0.00010037, 0.00017989, 0.00003494, 0.00007105, 0.00006054, 0.00020389, -0.00006465, -0.00009665, 0.00000588, -0.00001941, 0.00000788, 0.00000205, 0.00000384, 0.00003082, 0.00000545, -0.00002435, -0.00002624, -0.00000071, -0.00001124, 0.00000137, -0.00000856, -0.00000512, -0.00000487, -0.00000555, 0.00000452, -0.00001283, },
		{ -0.00001303, -0.00000492, 0.00001122, -0.00000114, -0.00000955, 0.00000010, 0.00000326, 0.00000188, -0.00000440, 0.00000214, -0.00000015, 0.00000020, 0.00000290, 0.00000316, -0.00000298, 0.00000340, -0.00000914, -0.00000130, 0.00000358, 0.00000175, -0.00000339, 0.00000390, -0.00000615, 0.00000258, -0.00000537, 0.00000252, 0.00000166, 0.00000276, -0.00000626, -0.00000103, -0.00000013, 0.00000361, -0.00000239, 0.00000613, -0.00000764, 0.00000270, -0.00000156, 0.00000363, -0.00000474, 0.00000542, -0.00000647, 0.00000521, -0.00000540, 0.00000649, -0.00001371, 0.00000632, -0.00000062, 0.00000708, -0.00000452, 0.00000428, -0.00000694, 0.00000416, -0.00000512, 0.00000537, -0.00000503, 0.00000314, -0.00000476, 0.00000466, -0.00000488, 0.00000424, -0.00000423, 0.00001180, -0.00001299, 0.00000654, -0.00000812, 0.00000802, -0.00000657, 0.00000940, -0.00001442, 0.00001184, -0.00001172, 0.00000764, -0.00000759, 0.00000813, -0.00000551, -0.00000176, -0.00000229, 0.00000696, -0.00000659, -0.00000099, 0.00000379, 0.00000313, -0.00000406, -0.00000029, -0.00000669, 0.00000419, 0.00000941, 0.00000133, -0.00001004, 0.00001104, -0.00001833, 0.00001189, -0.00003182, 0.00001355, -0.00000649, -0.00000128, -0.00001843, -0.00000059, -0.00001609, -0.00000073, -0.00005259, 0.00000223, -0.00009492, -0.00006485, 0.00017386, 0.00002188, 0.00006955, 0.00005596, 0.00019900, -0.00005754, -0.00009510, 0.00000596, -0.00001887, 0.00000922, 0.00000199, 0.00000701, 0.00003022, 0.00000487, -0.00002895, -0.00002428, -0.00000016, -0.00001272, 0.00000338, -0.00001024, -0.00000425, -0.00000588, -0.00000332, 0.00000227, },
		{ 0.00000023, -0.00001330, -0.00000337, 0.00001152, -0.00000280, -0.00000900, 0.00000059, 0.00000421, 0.00000036, -0.00000361, 0.00000177, -0.00000029, 0.00000087, 0.00000318, 0.00000200, -0.00000210, 0.00000252, -0.00000927, -0.00000025, 0.00000328, 0.00000163, -0.00000304, 0.00000323, -0.00000609, 0.00000289, -0.00000588, 0.00000342, 0.00000111, 0.00000287, -0.00000669, -0.00000055, 0.00000010, 0.00000314, -0.00000169, 0.00000528, -0.00000781, 0.00000349, -0.00000156, 0.00000298, -0.00000421, 0.00000505, -0.00000619, 0.00000502, -0.00000506, 0.00000579, -0.00001382, 0.00000778, -0.00000136, 0.00000740, -0.00000525, 0.00000438, -0.00000704, 0.00000464, -0.00000547, 0.00000562, -0.00000528, 0.00000321, -0.00000444, 0.00000419, -0.00000430, 0.00000355, -0.00000297, 0.00001009, -0.00001222, 0.00000647, -0.00000781, 0.00000751, -0.00000559, 0.00000849, -0.00001447, 0.00001223, -0.00001242, 0.00000884, -0.00000878, 0.00000938, -0.00000728, 0.00000028, -0.00000352, 0.00000710, -0.00000715, 0.00000040, 0.00000378, 0.00000204, -0.00000359, -0.00000018, -0.00000679, 0.00000391, 0.00001171, -0.00000265, -0.00000606, 0.00000685, -0.00001517, 0.00001033, -0.00003216, 0.00001474, -0.00000728, -0.00000092, -0.00001911, 0.00000023, -0.00001706, -0.00000022, -0.00005394, 0.00000282, -0.00010724, -0.00002908, 0.00016200, 0.00001324, 0.00006620, 0.00005297, 0.00019339, -0.00005009, -0.00009385, 0.00000632, -0.00001850, 0.00001072, 0.00000179, 0.00001070, 0.00002904, 0.00000411, -0.00003390, -0.00002097, -0.00000108, -0.00001293, 0.00000437, -0.00001112, -0.00000405, -0.00000621, -0.00000166, },
		{ -0.00000061, -0.00000156, -0.00001361, -0.00000141, 0.00001104, -0.00000384, -0.00000874, 0.00000145, 0.00000467, -0.00000089, -0.00000290, 0.00000141, -0.00000048, 0.00000176, 0.00000313, 0.00000104, -0.00000129, 0.00000156, -0.00000919, 0.00000073, 0.00000283, 0.00000167, -0.00000282, 0.00000255, -0.00000594, 0.00000312, -0.00000633, 0.00000437, 0.00000038, 0.00000314, -0.00000725, 0.00000016, 0.00000000, 0.00000297, -0.00000119, 0.00000445, -0.00000787, 0.00000431, -0.00000176, 0.00000248, -0.00000372, 0.00000465, -0.00000585, 0.00000476, -0.00000466, 0.00000490, -0.00001355, 0.00000898, -0.00000206, 0.00000768, -0.00000602, 0.00000458, -0.00000716, 0.00000512, -0.00000590, 0.00000597, -0.00000566, 0.00000347, -0.00000431, 0.00000387, -0.00000389, 0.00000306, -0.00000186, 0.00000829, -0.00001114, 0.00000610, -0.00000730, 0.00000683, -0.00000437, 0.00000714, -0.00001393, 0.00001203, -0.00001256, 0.00000955, -0.00000956, 0.00001025, -0.00000875, 0.00000222, -0.00000489, 0.00000739, -0.00000775, 0.00000202, 0.00000335, 0.00000124, -0.00000325, 0.00000007, -0.00000717, 0.00000418, 0.00001330, -0.00000620, -0.00000209, 0.00000225, -0.00001105, 0.00000736, -0.00003099, 0.00001469, -0.00000708, -0.00000159, -0.00001864, -0.00000015, -0.00001671, -0.00000147, -0.00005287, -0.00000027, -0.00011439, 0.00000493, 0.00014600, 0.00000837, 0.00006119, 0.00005137, 0.00018719, -0.00004243, -0.00009278, 0.00000687, -0.00001818, 0.00001219, 0.00000166, 0.00001469, 0.00002747, 0.00000285, -0.00003869, -0.00001678, -0.00000318, -0.00001201, 0.00000438, -0.00001123, -0.00000450, -0.00000586, },
		{ -0.00000491, -0.00000015, -0.00000312, -0.00001386, 0.00000079, 0.00000988, -0.00000428, -0.00000878, 0.00000268, 0.00000462, -0.00000178, -0.00000232, 0.00000110, -0.00000071, 0.00000284, 0.00000274, 0.00000033, -0.00000060, 0.00000059, -0.00000889, 0.00000158, 0.00000230, 0.00000185, -0.00000270, 0.00000188, -0.00000568, 0.00000323, -0.00000666, 0.00000530, -0.00000046, 0.00000352, -0.00000788, 0.00000105, -0.00000038, 0.00000309, -0.00000092, 0.00000368, -0.00000786, 0.00000514, -0.00000214, 0.00000216, -0.00000331, 0.00000427, -0.00000550, 0.00000449, -0.00000424, 0.00000387, -0.00001291, 0.00000986, -0.00000264, 0.00000784, -0.00000673, 0.00000482, -0.00000726, 0.00000559, -0.00000636, 0.00000639, -0.00000614, 0.00000391, -0.00000438, 0.00000373, -0.00000367, 0.00000282, -0.00000100, 0.00000655, -0.00000989, 0.00000554, -0.00000667, 0.00000607, -0.00000300, 0.00000544, -0.00001289, 0.00001127, -0.00001211, 0.00000972, -0.00000987, 0.00001066, -0.00000980, 0.00000392, -0.00000626, 0.00000771, -0.00000829, 0.00000374, 0.00000257, 0.00000072, -0.00000304, 0.00000045, -0.00000780, 0.00000501, 0.00001405, -0.00000905, 0.00000148, -0.00000232, -0.00000642, 0.00000334, -0.00002853, 0.00001354, -0.00000600, -0.00000316, -0.00001716, -0.00000159, -0.00001521, -0.00000427, -0.00004958, -0.00000699, -0.00011567, 0.00003546, 0.00012757, 0.00000645, 0.00005490, 0.00005085, 0.00018064, -0.00003470, -0.00009177, 0.00000747, -0.00001778, 0.00001349, 0.00000182, 0.00001876, 0.00002563, 0.00000085, -0.00004282, -0.00001223, -0.00000607, -0.00001022, 0.00000359, -0.00001068, -0.00000551, },
		{ -0.00000689, -0.00000349, -0.00000020, -0.00000450, -0.00001393, 0.00000303, 0.00000822, -0.00000418, -0.00000908, 0.00000423, 0.00000405, -0.00000230, -0.00000192, 0.00000084, -0.00000097, 0.00000408, 0.00000203, -0.00000010, -0.00000010, -0.00000035, -0.00000842, 0.00000227, 0.00000174, 0.00000212, -0.00000267, 0.00000121, -0.00000530, 0.00000317, -0.00000681, 0.00000612, -0.00000132, 0.00000393, -0.00000852, 0.00000203, -0.00000098, 0.00000348, -0.00000091, 0.00000302, -0.00000777, 0.00000598, -0.00000270, 0.00000203, -0.00000302, 0.00000396, -0.00000518, 0.00000422, -0.00000386, 0.00000277, -0.00001195, 0.00001041, -0.00000305, 0.00000783, -0.00000734, 0.00000505, -0.00000731, 0.00000598, -0.00000680, 0.00000681, -0.00000669, 0.00000450, -0.00000465, 0.00000377, -0.00000365, 0.00000287, -0.00000047, 0.00000501, -0.00000860, 0.00000487, -0.00000600, 0.00000531, -0.00000163, 0.00000356, -0.00001147, 0.00001004, -0.00001113, 0.00000937, -0.00000967, 0.00001056, -0.00001034, 0.00000523, -0.00000749, 0.00000798, -0.00000868, 0.00000547, 0.00000151, 0.00000045, -0.00000294, 0.00000093, -0.00000864, 0.00000638, 0.00001387, -0.00001098, 0.00000432, -0.00000643, -0.00000177, -0.00000128, -0.00002512, 0.00001152, -0.00000426, -0.00000538, -0.00001490, -0.00000381, -0.00001290, -0.00000822, -0.00004455, -0.00001694, -0.00011088, 0.00006118, 0.00010829, 0.00000661, 0.00004776, 0.00005107, 0.00017398, -0.00002711, -0.00009065, 0.00000799, -0.00001715, 0.00001447, 0.00000243, 0.00002275, 0.00002359, -0.00000207, -0.00004585, -0.00000785, -0.00000931, -0.00000790, 0.00000224, -0.00000966, },
		{ -0.00000838, -0.00000847, -0.00000180, -0.00000064, -0.00000577, -0.00001369, 0.00000509, 0.00000627, -0.00000370, -0.00000954, 0.00000601, 0.00000303, -0.00000243, -0.00000171, 0.00000065, -0.00000121, 0.00000542, 0.00000102, -0.00000025, 0.00000019, -0.00000119, -0.00000781, 0.00000279, 0.00000120, 0.00000246, -0.00000273, 0.00000055, -0.00000481, 0.00000293, -0.00000672, 0.00000675, -0.00000212, 0.00000431, -0.00000908, 0.00000301, -0.00000172, 0.00000409, -0.00000113, 0.00000248, -0.00000763, 0.00000678, -0.00000340, 0.00000210, -0.00000289, 0.00000376, -0.00000492, 0.00000401, -0.00000356, 0.00000168, -0.00001073, 0.00001064, -0.00000330, 0.00000764, -0.00000779, 0.00000523, -0.00000727, 0.00000627, -0.00000718, 0.00000720, -0.00000723, 0.00000518, -0.00000508, 0.00000396, -0.00000381, 0.00000321, -0.00000034, 0.00000376, -0.00000738, 0.00000420, -0.00000537, 0.00000467, -0.00000038, 0.00000164, -0.00000981, 0.00000846, -0.00000971, 0.00000855, -0.00000902, 0.00000996, -0.00001034, 0.00000607, -0.00000847, 0.00000810, -0.00000884, 0.00000709, 0.00000026, 0.00000040, -0.00000293, 0.00000145, -0.00000959, 0.00000820, 0.00001278, -0.00001187, 0.00000620, -0.00000970, 0.00000243, -0.00000602, -0.00002115, 0.00000894, -0.00000214, -0.00000796, -0.00001220, -0.00000648, -0.00001017, -0.00001278, -0.00003847, -0.00002933, -0.00010026, 0.00008124, 0.00008950, 0.00000800, 0.00004021, 0.00005167, 0.00016747, -0.00001983, -0.00008929, 0.00000833, -0.00001621, 0.00001505, 0.00000364, 0.00002652, 0.00002135, -0.00000595, -0.00004746, -0.00000410, -0.00001245, -0.00000540, 0.00000059, },
		{ -0.00000107, -0.00000708, -0.00001002, -0.00000000, -0.00000134, -0.00000698, -0.00001304, 0.00000675, 0.00000427, -0.00000299, -0.00001002, 0.00000786, 0.00000166, -0.00000221, -0.00000169, 0.00000051, -0.00000141, 0.00000680, -0.00000023, -0.00000013, 0.00000025, -0.00000187, -0.00000712, 0.00000314, 0.00000070, 0.00000283, -0.00000286, -0.00000008, -0.00000421, 0.00000251, -0.00000637, 0.00000713, -0.00000280, 0.00000459, -0.00000949, 0.00000389, -0.00000249, 0.00000483, -0.00000157, 0.00000208, -0.00000743, 0.00000753, -0.00000421, 0.00000236, -0.00000291, 0.00000368, -0.00000476, 0.00000389, -0.00000340, 0.00000069, -0.00000934, 0.00001060, -0.00000338, 0.00000724, -0.00000806, 0.00000533, -0.00000713, 0.00000642, -0.00000745, 0.00000750, -0.00000772, 0.00000591, -0.00000562, 0.00000428, -0.00000413, 0.00000381, -0.00000059, 0.00000288, -0.00000636, 0.00000362, -0.00000487, 0.00000423, 0.00000062, -0.00000015, -0.00000809, 0.00000668, -0.00000797, 0.00000737, -0.00000799, 0.00000893, -0.00000983, 0.00000641, -0.00000912, 0.00000800, -0.00000869, 0.00000849, -0.00000106, 0.00000050, -0.00000297, 0.00000196, -0.00001055, 0.00001033, 0.00001087, -0.00001171, 0.00000697, -0.00001185, 0.00000577, -0.00001044, -0.00001702, 0.00000612, 0.00000005, -0.00001056, -0.00000940, -0.00000920, -0.00000750, -0.00001737, -0.00003211, -0.00004316, -0.00008454, 0.00009524, 0.00007225, 0.00000990, 0.00003266, 0.00005234, 0.00016134, -0.00001300, -0.00008758, 0.00000843, -0.00001492, 0.00001517, 0.00000552, 0.00002999, 0.00001883, -0.00001068, -0.00004748, -0.00000136, -0.00001510, -0.00000306, },
		{ -0.00000116, -0.00000251, -0.00000594, -0.00001136, 0.00000172, -0.00000220, -0.00000820, -0.00001191, 0.00000784, 0.00000245, -0.00000225, -0.00001038, 0.00000964, 0.00000007, -0.00000172, -0.00000182, 0.00000039, -0.00000151, 0.00000813, -0.00000165, 0.00000024, 0.00000006, -0.00000234, -0.00000640, 0.00000336, 0.00000024, 0.00000323, -0.00000307, -0.00000066, -0.00000356, 0.00000195, -0.00000577, 0.00000722, -0.00000329, 0.00000473, -0.00000970, 0.00000459, -0.00000320, 0.00000563, -0.00000216, 0.00000179, -0.00000717, 0.00000818, -0.00000508, 0.00000277, -0.00000308, 0.00000373, -0.00000472, 0.00000387, -0.00000342, -0.00000012, -0.00000788, 0.00001034, -0.00000332, 0.00000666, -0.00000812, 0.00000533, -0.00000687, 0.00000642, -0.00000758, 0.00000767, -0.00000810, 0.00000660, -0.00000623, 0.00000467, -0.00000453, 0.00000461, -0.00000123, 0.00000240, -0.00000559, 0.00000321, -0.00000454, 0.00000406, 0.00000126, -0.00000166, -0.00000649, 0.00000489, -0.00000609, 0.00000596, -0.00000669, 0.00000756, -0.00000888, 0.00000626, -0.00000941, 0.00000766, -0.00000818, 0.00000958, -0.00000236, 0.00000069, -0.00000301, 0.00000237, -0.00001139, 0.00001258, 0.00000831, -0.00001058, 0.00000662, -0.00001274, 0.00000795, -0.00001414, -0.00001310, 0.00000338, 0.00000203, -0.00001290, -0.00000682, -0.00001163, -0.00000530, -0.00002140, -0.00002633, -0.00005726, -0.00006476, 0.00010325, 0.00005724, 0.00001172, 0.00002542, 0.00005284, 0.00015573, -0.00000674, -0.00008547, 0.00000825, -0.00001327, 0.00001487, 0.00000809, 0.00003313, 0.00001585, -0.00001602, -0.00004592, 0.00000017, -0.00001699, },
		{ -0.00001795, 0.00000010, -0.00000357, -0.00000511, -0.00001237, 0.00000322, -0.00000312, -0.00000942, -0.00001029, 0.00000825, 0.00000100, -0.00000164, -0.00001047, 0.00001117, -0.00000159, -0.00000103, -0.00000207, 0.00000029, -0.00000147, 0.00000934, -0.00000316, 0.00000080, -0.00000036, -0.00000259, -0.00000573, 0.00000348, -0.00000016, 0.00000365, -0.00000337, -0.00000114, -0.00000290, 0.00000128, -0.00000494, 0.00000703, -0.00000358, 0.00000470, -0.00000969, 0.00000505, -0.00000375, 0.00000640, -0.00000285, 0.00000162, -0.00000685, 0.00000870, -0.00000595, 0.00000330, -0.00000340, 0.00000393, -0.00000480, 0.00000397, -0.00000363, -0.00000069, -0.00000645, 0.00000995, -0.00000320, 0.00000595, -0.00000800, 0.00000523, -0.00000651, 0.00000626, -0.00000755, 0.00000768, -0.00000831, 0.00000721, -0.00000684, 0.00000508, -0.00000495, 0.00000552, -0.00000217, 0.00000232, -0.00000512, 0.00000300, -0.00000442, 0.00000419, 0.00000147, -0.00000277, -0.00000515, 0.00000324, -0.00000421, 0.00000447, -0.00000527, 0.00000599, -0.00000759, 0.00000568, -0.00000934, 0.00000706, -0.00000730, 0.00001029, -0.00000356, 0.00000090, -0.00000300, 0.00000262, -0.00001196, 0.00001474, 0.00000533, -0.00000866, 0.00000526, -0.00001236, 0.00000880, -0.00001683, -0.00000969, 0.00000096, 0.00000357, -0.00001472, -0.00000474, -0.00001349, -0.00000393, -0.00002439, -0.00002186, -0.00007041, -0.00004223, 0.00010575, 0.00004484, 0.00001310, 0.00001869, 0.00005300, 0.00015076, -0.00000107, -0.00008294, 0.00000782, -0.00001131, 0.00001421, 0.00001133, 0.00003593, 0.00001222, -0.00002161, -0.00004296, 0.00000042, },
		{ -0.00000049, -0.00001795, 0.00000065, -0.00000415, -0.00000465, -0.00001296, 0.00000442, -0.00000405, -0.00001062, -0.00000822, 0.00000793, 0.00000006, -0.00000129, -0.00001017, 0.00001231, -0.00000317, -0.00000025, -0.00000239, 0.00000016, -0.00000126, 0.00001035, -0.00000468, 0.00000150, -0.00000097, -0.00000258, -0.00000514, 0.00000353, -0.00000054, 0.00000409, -0.00000379, -0.00000149, -0.00000231, 0.00000059, -0.00000395, 0.00000659, -0.00000366, 0.00000450, -0.00000944, 0.00000525, -0.00000409, 0.00000705, -0.00000357, 0.00000152, -0.00000646, 0.00000905, -0.00000677, 0.00000390, -0.00000381, 0.00000423, -0.00000500, 0.00000419, -0.00000405, -0.00000096, -0.00000513, 0.00000948, -0.00000305, 0.00000515, -0.00000772, 0.00000504, -0.00000605, 0.00000596, -0.00000737, 0.00000751, -0.00000833, 0.00000768, -0.00000740, 0.00000545, -0.00000532, 0.00000647, -0.00000333, 0.00000259, -0.00000494, 0.00000300, -0.00000451, 0.00000461, 0.00000122, -0.00000340, -0.00000420, 0.00000188, -0.00000248, 0.00000303, -0.00000385, 0.00000436, -0.00000612, 0.00000477, -0.00000897, 0.00000623, -0.00000605, 0.00001058, -0.00000457, 0.00000108, -0.00000290, 0.00000265, -0.00001216, 0.00001659, 0.00000220, -0.00000620, 0.00000309, -0.00001083, 0.00000830, -0.00001835, -0.00000699, -0.00000097, 0.00000452, -0.00001586, -0.00000333, -0.00001459, -0.00000360, -0.00002599, -0.00001933, -0.00008150, -0.00001836, 0.00010351, 0.00003512, 0.00001382, 0.00001256, 0.00005277, 0.00014645, 0.00000399, -0.00008002, 0.00000720, -0.00000912, 0.00001328, 0.00001519, 0.00003838, 0.00000773, -0.00002701, -0.00003893, },
		{ -0.00003425, -0.00000235, -0.00001708, 0.00000050, -0.00000425, -0.00000458, -0.00001312, 0.00000528, -0.00000499, -0.00001170, -0.00000583, 0.00000692, -0.00000033, -0.00000127, -0.00000941, 0.00001293, -0.00000453, 0.00000052, -0.00000273, -0.00000002, -0.00000084, 0.00001109, -0.00000612, 0.00000226, -0.00000174, -0.00000234, -0.00000469, 0.00000357, -0.00000091, 0.00000458, -0.00000433, -0.00000165, -0.00000184, -0.00000004, -0.00000288, 0.00000594, -0.00000357, 0.00000416, -0.00000897, 0.00000518, -0.00000418, 0.00000752, -0.00000426, 0.00000146, -0.00000599, 0.00000922, -0.00000747, 0.00000450, -0.00000429, 0.00000463, -0.00000530, 0.00000451, -0.00000466, -0.00000091, -0.00000400, 0.00000903, -0.00000296, 0.00000434, -0.00000734, 0.00000479, -0.00000554, 0.00000555, -0.00000705, 0.00000717, -0.00000812, 0.00000796, -0.00000786, 0.00000573, -0.00000557, 0.00000733, -0.00000460, 0.00000312, -0.00000501, 0.00000319, -0.00000478, 0.00000530, 0.00000054, -0.00000352, -0.00000370, 0.00000091, -0.00000104, 0.00000178, -0.00000258, 0.00000280, -0.00000460, 0.00000366, -0.00000836, 0.00000523, -0.00000449, 0.00001045, -0.00000535, 0.00000119, -0.00000268, 0.00000243, -0.00001189, 0.00001795, -0.00000081, -0.00000350, 0.00000040, -0.00000837, 0.00000654, -0.00001866, -0.00000513, -0.00000231, 0.00000483, -0.00001626, -0.00000265, -0.00001487, -0.00000438, -0.00002607, -0.00001909, -0.00008958, 0.00000543, 0.00009748, 0.00002794, 0.00001384, 0.00000703, 0.00005213, 0.00014277, 0.00000850, -0.00007676, 0.00000646, -0.00000679, 0.00001221, 0.00001961, 0.00004045, 0.00000221, -0.00003174, },
		{ -0.00003538, -0.00002940, -0.00000481, -0.00001554, -0.00000023, -0.00000394, -0.00000483, -0.00001289, 0.00000581, -0.00000595, -0.00001257, -0.00000328, 0.00000533, -0.00000018, -0.00000161, -0.00000815, 0.00001296, -0.00000556, 0.00000119, -0.00000304, -0.00000025, -0.00000021, 0.00001152, -0.00000740, 0.00000300, -0.00000258, -0.00000188, -0.00000439, 0.00000363, -0.00000128, 0.00000510, -0.00000501, -0.00000160, -0.00000157, -0.00000052, -0.00000182, 0.00000517, -0.00000335, 0.00000372, -0.00000835, 0.00000487, -0.00000401, 0.00000777, -0.00000488, 0.00000144, -0.00000544, 0.00000918, -0.00000802, 0.00000506, -0.00000477, 0.00000507, -0.00000566, 0.00000489, -0.00000541, -0.00000054, -0.00000310, 0.00000864, -0.00000298, 0.00000360, -0.00000690, 0.00000452, -0.00000501, 0.00000507, -0.00000662, 0.00000667, -0.00000771, 0.00000803, -0.00000819, 0.00000590, -0.00000564, 0.00000803, -0.00000585, 0.00000381, -0.00000527, 0.00000353, -0.00000516, 0.00000616, -0.00000051, -0.00000315, -0.00000368, 0.00000039, 0.00000004, 0.00000081, -0.00000155, 0.00000145, -0.00000318, 0.00000246, -0.00000761, 0.00000413, -0.00000269, 0.00000990, -0.00000587, 0.00000120, -0.00000234, 0.00000194, -0.00001110, 0.00001863, -0.00000343, -0.00000087, -0.00000247, -0.00000532, 0.00000375, -0.00001784, -0.00000408, -0.00000308, 0.00000454, -0.00001596, -0.00000267, -0.00001440, -0.00000619, -0.00002468, -0.00002129, -0.00009399, 0.00002785, 0.00008875, 0.00002294, 0.00001322, 0.00000202, 0.00005113, 0.00013968, 0.00001253, -0.00007324, 0.00000569, -0.00000442, 0.00001110, 0.00002453, 0.00004205, -0.00000445, },
		{ -0.00001220, -0.00003758, -0.00002487, -0.00000751, -0.00001359, -0.00000138, -0.00000334, -0.00000529, -0.00001237, 0.00000606, -0.00000698, -0.00001309, -0.00000080, 0.00000337, 0.00000042, -0.00000224, -0.00000646, 0.00001239, -0.00000620, 0.00000171, -0.00000328, -0.00000054, 0.00000063, 0.00001160, -0.00000845, 0.00000366, -0.00000343, -0.00000126, -0.00000425, 0.00000372, -0.00000166, 0.00000567, -0.00000582, -0.00000132, -0.00000154, -0.00000078, -0.00000088, 0.00000436, -0.00000309, 0.00000326, -0.00000765, 0.00000438, -0.00000360, 0.00000779, -0.00000538, 0.00000142, -0.00000483, 0.00000893, -0.00000838, 0.00000552, -0.00000522, 0.00000550, -0.00000605, 0.00000528, -0.00000626, 0.00000012, -0.00000246, 0.00000836, -0.00000316, 0.00000297, -0.00000645, 0.00000426, -0.00000450, 0.00000456, -0.00000613, 0.00000606, -0.00000710, 0.00000789, -0.00000837, 0.00000593, -0.00000550, 0.00000848, -0.00000699, 0.00000455, -0.00000563, 0.00000393, -0.00000557, 0.00000711, -0.00000183, -0.00000237, -0.00000411, 0.00000034, 0.00000072, 0.00000017, -0.00000083, 0.00000038, -0.00000197, 0.00000131, -0.00000682, 0.00000303, -0.00000073, 0.00000900, -0.00000612, 0.00000111, -0.00000190, 0.00000122, -0.00000977, 0.00001855, -0.00000545, 0.00000142, -0.00000517, -0.00000204, 0.00000024, -0.00001607, -0.00000377, -0.00000335, 0.00000374, -0.00001508, -0.00000326, -0.00001335, -0.00000880, -0.00002205, -0.00002576, -0.00009436, 0.00004782, 0.00007836, 0.00001968, 0.00001212, -0.00000259, 0.00004987, 0.00013708, 0.00001616, -0.00006953, 0.00000495, -0.00000208, 0.00001009, 0.00002991, 0.00004304, },
		{ 0.00004320, -0.00002089, -0.00003812, -0.00002104, -0.00001007, -0.00001152, -0.00000275, -0.00000261, -0.00000582, -0.00001168, 0.00000610, -0.00000813, -0.00001314, 0.00000139, 0.00000124, 0.00000130, -0.00000305, -0.00000441, 0.00001125, -0.00000641, 0.00000203, -0.00000344, -0.00000087, 0.00000167, 0.00001132, -0.00000922, 0.00000416, -0.00000422, -0.00000054, -0.00000424, 0.00000385, -0.00000206, 0.00000626, -0.00000674, -0.00000080, -0.00000179, -0.00000075, -0.00000014, 0.00000360, -0.00000283, 0.00000284, -0.00000694, 0.00000380, -0.00000301, 0.00000759, -0.00000576, 0.00000141, -0.00000417, 0.00000848, -0.00000852, 0.00000584, -0.00000558, 0.00000590, -0.00000642, 0.00000565, -0.00000714, 0.00000101, -0.00000208, 0.00000820, -0.00000352, 0.00000253, -0.00000607, 0.00000406, -0.00000406, 0.00000408, -0.00000563, 0.00000539, -0.00000632, 0.00000753, -0.00000839, 0.00000583, -0.00000515, 0.00000864, -0.00000790, 0.00000524, -0.00000601, 0.00000433, -0.00000594, 0.00000803, -0.00000330, -0.00000127, -0.00000493, 0.00000072, 0.00000098, -0.00000010, -0.00000046, -0.00000034, -0.00000108, 0.00000033, -0.00000609, 0.00000203, 0.00000128, 0.00000780, -0.00000612, 0.00000092, -0.00000138, 0.00000033, -0.00000798, 0.00001766, -0.00000673, 0.00000314, -0.00000741, 0.00000109, -0.00000364, -0.00001360, -0.00000405, -0.00000328, 0.00000262, -0.00001380, -0.00000422, -0.00001196, -0.00001190, -0.00001860, -0.00003210, -0.00009069, 0.00006455, 0.00006729, 0.00001771, 0.00001072, -0.00000694, 0.00004842, 0.00013490, 0.00001947, -0.00006569, 0.00000431, 0.00000015, 0.00000925, 0.00003573, },
		{ 0.00004198, 0.00004225, -0.00003021, -0.00003696, -0.00001823, -0.00001218, -0.00000958, -0.00000412, -0.00000190, -0.00000629, -0.00001094, 0.00000601, -0.00000944, -0.00001260, 0.00000308, -0.00000080, 0.00000229, -0.00000390, -0.00000217, 0.00000963, -0.00000622, 0.00000213, -0.00000351, -0.00000123, 0.00000284, 0.00001069, -0.00000967, 0.00000444, -0.00000485, 0.00000020, -0.00000433, 0.00000400, -0.00000245, 0.00000685, -0.00000774, -0.00000008, -0.00000232, -0.00000039, 0.00000033, 0.00000297, -0.00000266, 0.00000253, -0.00000631, 0.00000321, -0.00000231, 0.00000721, -0.00000601, 0.00000141, -0.00000349, 0.00000787, -0.00000845, 0.00000599, -0.00000583, 0.00000620, -0.00000674, 0.00000593, -0.00000798, 0.00000208, -0.00000193, 0.00000815, -0.00000407, 0.00000229, -0.00000577, 0.00000395, -0.00000372, 0.00000367, -0.00000517, 0.00000472, -0.00000545, 0.00000700, -0.00000827, 0.00000561, -0.00000459, 0.00000848, -0.00000854, 0.00000578, -0.00000632, 0.00000466, -0.00000617, 0.00000880, -0.00000477, 0.00000001, -0.00000602, 0.00000147, 0.00000087, -0.00000004, -0.00000042, -0.00000069, -0.00000056, -0.00000040, -0.00000551, 0.00000121, 0.00000323, 0.00000641, -0.00000592, 0.00000068, -0.00000086, -0.00000063, -0.00000580, 0.00001602, -0.00000720, 0.00000414, -0.00000895, 0.00000373, -0.00000754, -0.00001070, -0.00000472, -0.00000303, 0.00000135, -0.00001233, -0.00000534, -0.00001051, -0.00001510, -0.00001483, -0.00003969, -0.00008325, 0.00007754, 0.00005633, 0.00001657, 0.00000922, -0.00001116, 0.00004687, 0.00013306, 0.00002252, -0.00006177, 0.00000381, 0.00000226, 0.00000866, },
		{ 0.00000838, 0.00004863, 0.00003989, -0.00003970, -0.00003423, -0.00001658, -0.00001361, -0.00000801, -0.00000533, -0.00000134, -0.00000657, -0.00001025, 0.00000583, -0.00001092, -0.00001143, 0.00000410, -0.00000252, 0.00000316, -0.00000459, 0.00000010, 0.00000766, -0.00000566, 0.00000203, -0.00000352, -0.00000156, 0.00000410, 0.00000974, -0.00000980, 0.00000448, -0.00000528, 0.00000089, -0.00000445, 0.00000415, -0.00000281, 0.00000738, -0.00000876, 0.00000080, -0.00000309, 0.00000030, 0.00000047, 0.00000253, -0.00000262, 0.00000238, -0.00000583, 0.00000271, -0.00000159, 0.00000669, -0.00000616, 0.00000144, -0.00000283, 0.00000713, -0.00000819, 0.00000598, -0.00000593, 0.00000639, -0.00000696, 0.00000610, -0.00000870, 0.00000324, -0.00000197, 0.00000819, -0.00000478, 0.00000228, -0.00000560, 0.00000395, -0.00000353, 0.00000338, -0.00000481, 0.00000410, -0.00000452, 0.00000634, -0.00000804, 0.00000531, -0.00000386, 0.00000802, -0.00000885, 0.00000610, -0.00000649, 0.00000485, -0.00000621, 0.00000933, -0.00000611, 0.00000133, -0.00000728, 0.00000248, 0.00000047, 0.00000031, -0.00000069, -0.00000071, -0.00000042, -0.00000081, -0.00000514, 0.00000065, 0.00000501, 0.00000491, -0.00000558, 0.00000042, -0.00000040, -0.00000156, -0.00000341, 0.00001371, -0.00000688, 0.00000436, -0.00000962, 0.00000562, -0.00001111, -0.00000765, -0.00000558, -0.00000277, 0.00000014, -0.00001087, -0.00000638, -0.00000927, -0.00001803, -0.00001130, -0.00004780, -0.00007263, 0.00008662, 0.00004611, 0.00001587, 0.00000777, -0.00001534, 0.00004529, 0.00013152, 0.00002536, -0.00005778, 0.00000344, 0.00000421, },
		{ 0.00000603, 0.00000847, 0.00005564, 0.00003580, -0.00004883, -0.00003023, -0.00001608, -0.00001425, -0.00000692, -0.00000627, -0.00000102, -0.00000659, -0.00000968, 0.00000560, -0.00001253, -0.00000964, 0.00000433, -0.00000373, 0.00000375, -0.00000495, 0.00000219, 0.00000550, -0.00000483, 0.00000177, -0.00000348, -0.00000181, 0.00000537, 0.00000854, -0.00000962, 0.00000426, -0.00000546, 0.00000146, -0.00000455, 0.00000426, -0.00000310, 0.00000783, -0.00000973, 0.00000177, -0.00000406, 0.00000130, 0.00000028, 0.00000232, -0.00000275, 0.00000243, -0.00000557, 0.00000238, -0.00000093, 0.00000612, -0.00000624, 0.00000152, -0.00000223, 0.00000631, -0.00000775, 0.00000580, -0.00000588, 0.00000644, -0.00000707, 0.00000611, -0.00000924, 0.00000440, -0.00000212, 0.00000827, -0.00000563, 0.00000248, -0.00000557, 0.00000407, -0.00000348, 0.00000324, -0.00000459, 0.00000361, -0.00000362, 0.00000559, -0.00000774, 0.00000499, -0.00000304, 0.00000731, -0.00000885, 0.00000617, -0.00000648, 0.00000487, -0.00000600, 0.00000953, -0.00000721, 0.00000255, -0.00000855, 0.00000364, -0.00000015, 0.00000087, -0.00000119, -0.00000045, -0.00000064, -0.00000089, -0.00000503, 0.00000043, 0.00000652, 0.00000340, -0.00000514, 0.00000018, -0.00000006, -0.00000233, -0.00000097, 0.00001092, -0.00000587, 0.00000385, -0.00000938, 0.00000656, -0.00001410, -0.00000471, -0.00000648, -0.00000263, -0.00000087, -0.00000958, -0.00000719, -0.00000846, -0.00002034, -0.00000853, -0.00005562, -0.00005959, 0.00009190, 0.00003704, 0.00001534, 0.00000647, -0.00001956, 0.00004369, 0.00013024, 0.00002802, -0.00005374, 0.00000322, },
		{ 0.00000313, 0.00000771, 0.00000901, 0.00006291, 0.00002974, -0.00005701, -0.00002539, -0.00001658, -0.00001411, -0.00000637, -0.00000689, -0.00000095, -0.00000632, -0.00000927, 0.00000529, -0.00001419, -0.00000730, 0.00000375, -0.00000430, 0.00000390, -0.00000482, 0.00000391, 0.00000333, -0.00000384, 0.00000141, -0.00000344, -0.00000194, 0.00000656, 0.00000717, -0.00000918, 0.00000380, -0.00000536, 0.00000183, -0.00000458, 0.00000429, -0.00000328, 0.00000811, -0.00001058, 0.00000273, -0.00000512, 0.00000254, -0.00000021, 0.00000234, -0.00000306, 0.00000269, -0.00000555, 0.00000228, -0.00000042, 0.00000554, -0.00000629, 0.00000166, -0.00000175, 0.00000549, -0.00000720, 0.00000549, -0.00000569, 0.00000636, -0.00000706, 0.00000595, -0.00000955, 0.00000550, -0.00000232, 0.00000832, -0.00000653, 0.00000287, -0.00000566, 0.00000431, -0.00000358, 0.00000326, -0.00000454, 0.00000327, -0.00000281, 0.00000481, -0.00000741, 0.00000469, -0.00000219, 0.00000642, -0.00000856, 0.00000599, -0.00000627, 0.00000471, -0.00000553, 0.00000935, -0.00000797, 0.00000356, -0.00000971, 0.00000483, -0.00000088, 0.00000155, -0.00000183, 0.00000002, -0.00000116, -0.00000065, -0.00000519, 0.00000055, 0.00000769, 0.00000198, -0.00000469, 0.00000003, 0.00000009, -0.00000282, 0.00000132, 0.00000786, -0.00000433, 0.00000272, -0.00000829, 0.00000651, -0.00001632, -0.00000205, -0.00000727, -0.00000271, -0.00000156, -0.00000859, -0.00000765, -0.00000822, -0.00002181, -0.00000695, -0.00006238, -0.00004503, 0.00009370, 0.00002934, 0.00001477, 0.00000540, -0.00002384, 0.00004207, 0.00012923, 0.00003049, -0.00004963, },
		{ -0.00004545, 0.00000315, 0.00000928, 0.00001011, 0.00007023, 0.00002155, -0.00006367, -0.00002025, -0.00001779, -0.00001333, -0.00000630, -0.00000722, -0.00000112, -0.00000580, -0.00000899, 0.00000486, -0.00001576, -0.00000457, 0.00000241, -0.00000418, 0.00000351, -0.00000409, 0.00000512, 0.00000131, -0.00000280, 0.00000100, -0.00000344, -0.00000190, 0.00000757, 0.00000572, -0.00000854, 0.00000313, -0.00000499, 0.00000198, -0.00000447, 0.00000418, -0.00000330, 0.00000819, -0.00001122, 0.00000358, -0.00000619, 0.00000393, -0.00000097, 0.00000256, -0.00000352, 0.00000313, -0.00000578, 0.00000245, -0.00000012, 0.00000504, -0.00000634, 0.00000191, -0.00000142, 0.00000471, -0.00000659, 0.00000507, -0.00000538, 0.00000614, -0.00000693, 0.00000563, -0.00000961, 0.00000645, -0.00000252, 0.00000829, -0.00000743, 0.00000339, -0.00000586, 0.00000464, -0.00000382, 0.00000343, -0.00000466, 0.00000314, -0.00000213, 0.00000405, -0.00000708, 0.00000447, -0.00000140, 0.00000544, -0.00000807, 0.00000559, -0.00000588, 0.00000438, -0.00000482, 0.00000880, -0.00000833, 0.00000427, -0.00001066, 0.00000594, -0.00000161, 0.00000227, -0.00000253, 0.00000058, -0.00000189, -0.00000013, -0.00000560, 0.00000104, 0.00000847, 0.00000072, -0.00000426, -0.00000002, -0.00000001, -0.00000293, 0.00000328, 0.00000475, -0.00000249, 0.00000119, -0.00000651, 0.00000548, -0.00001768, 0.00000019, -0.00000789, -0.00000304, -0.00000192, -0.00000793, -0.00000772, -0.00000859, -0.00002233, -0.00000684, -0.00006742, -0.00002987, 0.00009254, 0.00002308, 0.00001405, 0.00000457, -0.00002817, 0.00004043, 0.00012849, 0.00003279, },
		{ 0.00003492, -0.00004120, 0.00000328, 0.00001072, 0.00001189, 0.00007731, 0.00001121, -0.00006830, -0.00001535, -0.00001935, -0.00001212, -0.00000658, -0.00000736, -0.00000142, -0.00000510, -0.00000879, 0.00000423, -0.00001708, -0.00000166, 0.00000048, -0.00000342, 0.00000258, -0.00000272, 0.00000568, -0.00000041, -0.00000181, 0.00000060, -0.00000350, -0.00000163, 0.00000834, 0.00000429, -0.00000777, 0.00000232, -0.00000439, 0.00000190, -0.00000421, 0.00000393, -0.00000315, 0.00000803, -0.00001158, 0.00000423, -0.00000714, 0.00000537, -0.00000191, 0.00000295, -0.00000410, 0.00000371, -0.00000621, 0.00000290, -0.00000007, 0.00000466, -0.00000643, 0.00000227, -0.00000127, 0.00000405, -0.00000597, 0.00000461, -0.00000499, 0.00000582, -0.00000671, 0.00000517, -0.00000939, 0.00000722, -0.00000267, 0.00000812, -0.00000826, 0.00000399, -0.00000613, 0.00000503, -0.00000417, 0.00000373, -0.00000496, 0.00000323, -0.00000163, 0.00000335, -0.00000680, 0.00000438, -0.00000074, 0.00000446, -0.00000745, 0.00000503, -0.00000535, 0.00000394, -0.00000393, 0.00000791, -0.00000829, 0.00000461, -0.00001131, 0.00000688, -0.00000227, 0.00000294, -0.00000321, 0.00000113, -0.00000275, 0.00000060, -0.00000622, 0.00000187, 0.00000883, -0.00000032, -0.00000391, 0.00000007, -0.00000036, -0.00000257, 0.00000473, 0.00000182, -0.00000058, -0.00000052, -0.00000427, 0.00000363, -0.00001820, 0.00000197, -0.00000832, -0.00000361, -0.00000195, -0.00000757, -0.00000748, -0.00000949, -0.00002196, -0.00000830, -0.00007023, -0.00001497, 0.00008901, 0.00001819, 0.00001313, 0.00000396, -0.00003253, 0.00003871, 0.00012804, },
		},
	}
	;

	float f64_6[64][64][64] = {
		{ { -0.00024641, 0.00000296, 0.00027698, 0.00011909, -0.00008617, -0.00003547, -0.00008603, -0.00003638, -0.00000041, 0.00001976, -0.00003370, 0.00001948, -0.00002586, 0.00004374, -0.00001108, -0.00003535, -0.00000512, 0.00002152, 0.00001365, -0.00002183, 0.00003298, -0.00003111, 0.00001358, -0.00002216, 0.00001153, -0.00000614, 0.00001259, -0.00002795, 0.00001993, -0.00001897, 0.00000092, 0.00001644, -0.00000593, 0.00001644, 0.00000092, -0.00001897, 0.00001993, -0.00002795, 0.00001259, -0.00000614, 0.00001153, -0.00002216, 0.00001358, -0.00003111, 0.00003298, -0.00002183, 0.00001365, 0.00002152, -0.00000512, -0.00003535, -0.00001108, 0.00004374, -0.00002586, 0.00001948, -0.00003370, 0.00001976, -0.00000041, -0.00003638, -0.00008603, -0.00003547, -0.00008617, 0.00011909, 0.00027698, 0.00000296, },
		{ 0.00001717, -0.00024790, -0.00000758, 0.00023920, 0.00013123, -0.00006910, -0.00003252, -0.00009160, -0.00003100, -0.00001014, 0.00002867, -0.00003640, 0.00001733, -0.00002393, 0.00003733, -0.00000194, -0.00003842, -0.00000417, 0.00001362, 0.00002378, -0.00002961, 0.00003788, -0.00003248, 0.00001384, -0.00002180, 0.00001235, -0.00000840, 0.00001498, -0.00002831, 0.00001760, -0.00001486, -0.00000467, 0.00001889, -0.00000549, 0.00001327, 0.00000642, -0.00002211, 0.00002095, -0.00002633, 0.00000939, -0.00000358, 0.00001065, -0.00002212, 0.00001250, -0.00002871, 0.00002710, -0.00001366, 0.00000427, 0.00002748, -0.00000460, -0.00003256, -0.00001942, 0.00004819, -0.00002571, 0.00001922, -0.00002812, 0.00000919, 0.00000867, -0.00003941, -0.00008359, -0.00003415, -0.00010391, 0.00010129, 0.00031635, },
		{ 0.00035687, 0.00003577, -0.00025248, -0.00001485, 0.00020300, 0.00013913, -0.00005432, -0.00002493, -0.00009971, -0.00002438, -0.00001919, 0.00003484, -0.00003571, 0.00001267, -0.00001994, 0.00002914, 0.00000782, -0.00004178, -0.00000175, 0.00000411, 0.00003409, -0.00003647, 0.00004146, -0.00003263, 0.00001325, -0.00002101, 0.00001298, -0.00001017, 0.00001639, -0.00002733, 0.00001407, -0.00001007, -0.00001004, 0.00002043, -0.00000417, 0.00000958, 0.00001152, -0.00002408, 0.00002064, -0.00002360, 0.00000564, -0.00000094, 0.00000980, -0.00002171, 0.00001071, -0.00002551, 0.00002062, -0.00000559, -0.00000387, 0.00003130, -0.00000277, -0.00002996, -0.00002685, 0.00005062, -0.00002356, 0.00001684, -0.00002036, -0.00000186, 0.00001591, -0.00003936, -0.00008439, -0.00002947, -0.00012039, 0.00007655, },
		{ 0.00004404, 0.00039759, 0.00005981, -0.00026047, -0.00001899, 0.00016811, 0.00014425, -0.00004303, -0.00001297, -0.00010926, -0.00001790, -0.00002621, 0.00003735, -0.00003137, 0.00000560, -0.00001404, 0.00001943, 0.00001794, -0.00004540, 0.00000205, -0.00000655, 0.00004393, -0.00004190, 0.00004341, -0.00003146, 0.00001177, -0.00001980, 0.00001334, -0.00001127, 0.00001669, -0.00002504, 0.00000954, -0.00000492, -0.00001488, 0.00002095, -0.00000210, 0.00000563, 0.00001592, -0.00002475, 0.00001907, -0.00001998, 0.00000159, 0.00000158, 0.00000908, -0.00002097, 0.00000835, -0.00002177, 0.00001392, 0.00000192, -0.00001041, 0.00003297, 0.00000013, -0.00002745, -0.00003327, 0.00005105, -0.00001966, 0.00001276, -0.00001124, -0.00001228, 0.00002036, -0.00003591, -0.00008800, -0.00002277, -0.00013358, },
		{ -0.00014163, 0.00000348, 0.00043694, 0.00009062, -0.00027228, -0.00002005, 0.00013416, 0.00014780, -0.00003584, 0.00000251, -0.00011880, -0.00001303, -0.00002995, 0.00003557, -0.00002338, -0.00000362, -0.00000655, 0.00000856, 0.00002815, -0.00004916, 0.00000697, -0.00001776, 0.00005263, -0.00004546, 0.00004355, -0.00002894, 0.00000948, -0.00001820, 0.00001334, -0.00001157, 0.00001582, -0.00002153, 0.00000428, 0.00000026, -0.00001894, 0.00002041, 0.00000058, 0.00000169, 0.00001940, -0.00002407, 0.00001641, -0.00001574, -0.00000247, 0.00000379, 0.00000855, -0.00001997, 0.00000557, -0.00001776, 0.00000735, 0.00000850, -0.00001514, 0.00003262, 0.00000377, -0.00002487, -0.00003865, 0.00004959, -0.00001432, 0.00000748, -0.00000160, -0.00002104, 0.00002142, -0.00002919, -0.00009357, -0.00001567, },
		{ -0.00000991, -0.00014303, -0.00004468, 0.00047280, 0.00012975, -0.00028832, -0.00001805, 0.00010087, 0.00015063, -0.00003271, 0.00002012, -0.00012664, -0.00001122, -0.00002942, 0.00002921, -0.00001205, -0.00001451, 0.00000213, -0.00000305, 0.00003810, -0.00005286, 0.00001269, -0.00002887, 0.00005956, -0.00004679, 0.00004177, -0.00002510, 0.00000645, -0.00001624, 0.00001290, -0.00001100, 0.00001382, -0.00001701, -0.00000139, 0.00000512, -0.00002199, 0.00001883, 0.00000368, -0.00000198, 0.00002178, -0.00002210, 0.00001287, -0.00001118, -0.00000630, 0.00000554, 0.00000827, -0.00001875, 0.00000255, -0.00001374, 0.00000123, 0.00001388, -0.00001797, 0.00003050, 0.00000778, -0.00002211, -0.00004296, 0.00004641, -0.00000794, 0.00000159, 0.00000772, -0.00002737, 0.00001889, -0.00001979, -0.00009988, },
		{ -0.00010553, -0.00000709, -0.00013679, -0.00009921, 0.00050251, 0.00017874, -0.00030884, -0.00001321, 0.00006818, 0.00015310, -0.00003297, 0.00003805, -0.00013101, -0.00001375, -0.00002397, 0.00001833, 0.00000204, -0.00002647, 0.00001151, -0.00001490, 0.00004741, -0.00005624, 0.00001876, -0.00003918, 0.00006415, -0.00004567, 0.00003809, -0.00002009, 0.00000284, -0.00001399, 0.00001200, -0.00000952, 0.00001079, -0.00001172, -0.00000713, 0.00000936, -0.00002391, 0.00001632, 0.00000695, -0.00000515, 0.00002297, -0.00001896, 0.00000875, -0.00000661, -0.00000964, 0.00000671, 0.00000825, -0.00001738, -0.00000054, -0.00000996, -0.00000422, 0.00001789, -0.00001896, 0.00002696, 0.00001175, -0.00001907, -0.00004619, 0.00004174, -0.00000096, -0.00000436, 0.00001601, -0.00003073, 0.00001295, -0.00000866, },
		{ 0.00000299, -0.00010906, -0.00000856, -0.00012260, -0.00015810, 0.00052312, 0.00023898, -0.00033373, -0.00000606, 0.00003636, 0.00015504, -0.00003540, 0.00005423, -0.00013025, -0.00002153, -0.00001336, 0.00000341, 0.00001806, -0.00003875, 0.00002103, -0.00002649, 0.00005569, -0.00005903, 0.00002470, -0.00004805, 0.00006595, -0.00004202, 0.00003266, -0.00001412, -0.00000117, -0.00001152, 0.00001062, -0.00000719, 0.00000690, -0.00000598, -0.00001258, 0.00001271, -0.00002463, 0.00001303, 0.00001016, -0.00000764, 0.00002295, -0.00001487, 0.00000435, -0.00000233, -0.00001231, 0.00000724, 0.00000850, -0.00001591, -0.00000354, -0.00000658, -0.00000883, 0.00002048, -0.00001830, 0.00002244, 0.00001532, -0.00001568, -0.00004836, 0.00003585, 0.00000616, -0.00000983, 0.00002270, -0.00003092, 0.00000416, },
		{ -0.00000660, 0.00001379, -0.00010915, -0.00001519, -0.00010094, -0.00021864, 0.00053150, 0.00031147, -0.00036240, 0.00000248, 0.00000614, 0.00015575, -0.00003834, 0.00006656, -0.00012300, -0.00003507, 0.00000216, -0.00001469, 0.00003497, -0.00005056, 0.00003012, -0.00003728, 0.00006254, -0.00006089, 0.00003000, -0.00005485, 0.00006466, -0.00003591, 0.00002572, -0.00000748, -0.00000538, -0.00000894, 0.00000879, -0.00000409, 0.00000239, -0.00000011, -0.00001741, 0.00001497, -0.00002419, 0.00000919, 0.00001308, -0.00000930, 0.00002178, -0.00001008, -0.00000000, 0.00000141, -0.00001419, 0.00000712, 0.00000897, -0.00001437, -0.00000632, -0.00000375, -0.00001252, 0.00002170, -0.00001624, 0.00001739, 0.00001818, -0.00001195, -0.00004947, 0.00002904, 0.00001300, -0.00001439, 0.00002743, -0.00002806, },
		{ -0.00002255, -0.00001824, 0.00002240, -0.00010479, -0.00002727, -0.00007301, -0.00027759, 0.00052465, 0.00039660, -0.00039362, 0.00001102, -0.00002137, 0.00015411, -0.00003989, 0.00007308, -0.00010836, -0.00005433, 0.00002187, -0.00003481, 0.00005162, -0.00006112, 0.00003821, -0.00004677, 0.00006759, -0.00006152, 0.00003418, -0.00005910, 0.00006016, -0.00002758, 0.00001764, -0.00000048, -0.00000954, -0.00000633, 0.00000656, -0.00000038, -0.00000245, 0.00000553, -0.00002134, 0.00001603, -0.00002267, 0.00000502, 0.00001550, -0.00001008, 0.00001960, -0.00000491, -0.00000401, 0.00000442, -0.00001521, 0.00000638, 0.00000961, -0.00001280, -0.00000877, -0.00000154, -0.00001530, 0.00002167, -0.00001313, 0.00001224, 0.00002009, -0.00000794, -0.00004953, 0.00002161, 0.00001917, -0.00001774, 0.00003005, },
		{ 0.00003063, -0.00001504, -0.00002956, 0.00002763, -0.00009537, -0.00004448, -0.00004075, -0.00033137, 0.00049997, 0.00049394, -0.00042547, 0.00001769, -0.00004470, 0.00014870, -0.00003809, 0.00007219, -0.00008600, -0.00007870, 0.00004463, -0.00005555, 0.00006681, -0.00006966, 0.00004479, -0.00005450, 0.00007050, -0.00006063, 0.00003683, -0.00006047, 0.00005252, -0.00001741, 0.00000885, 0.00000651, -0.00001346, -0.00000382, 0.00000403, 0.00000374, -0.00000734, 0.00001063, -0.00002414, 0.00001587, -0.00002025, 0.00000079, 0.00001724, -0.00000998, 0.00001660, 0.00000034, -0.00000739, 0.00000656, -0.00001537, 0.00000513, 0.00001035, -0.00001124, -0.00001081, 0.00000004, -0.00001724, 0.00002059, -0.00000931, 0.00000739, 0.00002092, -0.00000379, -0.00004854, 0.00001385, 0.00002436, -0.00001969, },
		{ -0.00002022, 0.00002944, -0.00000634, -0.00003935, 0.00002855, -0.00008080, -0.00006582, -0.00000666, -0.00037630, 0.00045554, 0.00060210, -0.00045533, 0.00002025, -0.00006214, 0.00013798, -0.00003120, 0.00006286, -0.00005623, -0.00010705, 0.00006893, -0.00007538, 0.00007939, -0.00007556, 0.00004946, -0.00006008, 0.00007104, -0.00005803, 0.00003763, -0.00005876, 0.00004201, -0.00000594, -0.00000015, 0.00001313, -0.00001691, -0.00000149, 0.00000130, 0.00000805, -0.00001196, 0.00001491, -0.00002567, 0.00001453, -0.00001715, -0.00000324, 0.00001820, -0.00000906, 0.00001302, 0.00000536, -0.00000995, 0.00000777, -0.00001474, 0.00000346, 0.00001110, -0.00000968, -0.00001241, 0.00000103, -0.00001847, 0.00001870, -0.00000514, 0.00000316, 0.00002063, 0.00000032, -0.00004652, 0.00000604, 0.00002835, },
		{ 0.00003101, -0.00001945, 0.00002691, 0.00000264, -0.00004657, 0.00002467, -0.00006157, -0.00008972, 0.00002637, -0.00040897, 0.00039033, 0.00071865, -0.00047991, 0.00001618, -0.00007193, 0.00012051, -0.00001785, 0.00004472, -0.00002007, -0.00013775, 0.00009301, -0.00009273, 0.00008837, -0.00007834, 0.00005190, -0.00006324, 0.00006905, -0.00005358, 0.00003638, -0.00005401, 0.00002906, 0.00000623, -0.00000886, 0.00001904, -0.00001971, 0.00000057, -0.00000152, 0.00001232, -0.00001603, 0.00001816, -0.00002589, 0.00001217, -0.00001363, -0.00000685, 0.00001831, -0.00000744, 0.00000912, 0.00000986, -0.00001156, 0.00000807, -0.00001342, 0.00000153, 0.00001177, -0.00000815, -0.00001356, 0.00000153, -0.00001914, 0.00001624, -0.00000094, -0.00000022, 0.00001930, 0.00000418, -0.00004350, -0.00000158, },
		{ -0.00000876, 0.00003230, -0.00001760, 0.00002356, 0.00001101, -0.00005041, 0.00001594, -0.00003871, -0.00011412, 0.00005530, -0.00042643, 0.00030446, 0.00084010, -0.00049542, 0.00000291, -0.00007245, 0.00009519, 0.00000274, 0.00001817, 0.00002083, -0.00016879, 0.00011503, -0.00010617, 0.00009294, -0.00007771, 0.00005194, -0.00006381, 0.00006448, -0.00004726, 0.00003304, -0.00004642, 0.00001428, 0.00001843, -0.00001677, 0.00002396, -0.00002174, 0.00000228, -0.00000428, 0.00001631, -0.00001931, 0.00002022, -0.00002481, 0.00000898, -0.00000997, -0.00000985, 0.00001760, -0.00000530, 0.00000517, 0.00001363, -0.00001215, 0.00000756, -0.00001159, -0.00000051, 0.00001229, -0.00000663, -0.00001430, 0.00000166, -0.00001942, 0.00001342, 0.00000304, -0.00000264, 0.00001712, 0.00000752, -0.00003952, },
		{ -0.00003467, -0.00001533, 0.00003228, -0.00001500, 0.00001996, 0.00001799, -0.00005039, 0.00000282, -0.00001375, -0.00013668, 0.00007719, -0.00042654, 0.00019926, 0.00096198, -0.00049777, -0.00002199, -0.00006244, 0.00006138, 0.00003063, -0.00001562, 0.00006427, -0.00019796, 0.00013314, -0.00011447, 0.00009259, -0.00007362, 0.00004955, -0.00006173, 0.00005740, -0.00003918, 0.00002771, -0.00003637, -0.00000163, 0.00002996, -0.00002345, 0.00002765, -0.00002290, 0.00000360, -0.00000686, 0.00001980, -0.00002163, 0.00002105, -0.00002257, 0.00000522, -0.00000642, -0.00001210, 0.00001611, -0.00000284, 0.00000142, 0.00001650, -0.00001174, 0.00000636, -0.00000942, -0.00000251, 0.00001258, -0.00000514, -0.00001467, 0.00000156, -0.00001947, 0.00001043, 0.00000663, -0.00000408, 0.00001432, 0.00001012, },
		{ 0.00001176, -0.00002905, -0.00002112, 0.00003107, -0.00001201, 0.00001666, 0.00002296, -0.00004642, -0.00001375, 0.00001140, -0.00015493, 0.00008947, -0.00040810, 0.00007738, 0.00107902, -0.00048285, -0.00006049, -0.00004122, 0.00001915, 0.00006513, -0.00005475, 0.00010763, -0.00022298, 0.00014570, -0.00011671, 0.00008714, -0.00006623, 0.00004484, -0.00005709, 0.00004803, -0.00002957, 0.00002063, -0.00002438, -0.00001787, 0.00004017, -0.00002853, 0.00002996, -0.00002316, 0.00000449, -0.00000916, 0.00002262, -0.00002285, 0.00002068, -0.00001935, 0.00000117, -0.00000323, -0.00001352, 0.00001398, -0.00000028, -0.00000193, 0.00001838, -0.00001042, 0.00000467, -0.00000710, -0.00000433, 0.00001260, -0.00000366, -0.00001474, 0.00000136, -0.00001943, 0.00000742, 0.00000971, -0.00000463, 0.00001121, },
		{ 0.00000811, 0.00001228, -0.00002284, -0.00002598, 0.00002884, -0.00000902, 0.00001411, 0.00002558, -0.00003884, -0.00003240, 0.00003462, -0.00016655, 0.00009018, -0.00037111, -0.00005733, 0.00118539, -0.00044681, -0.00011392, -0.00000884, -0.00003069, 0.00010473, -0.00009672, 0.00014811, -0.00024170, 0.00015141, -0.00011239, 0.00007674, -0.00005592, 0.00003807, -0.00005011, 0.00003669, -0.00001875, 0.00001218, -0.00001111, -0.00003364, 0.00004846, -0.00003174, 0.00003082, -0.00002253, 0.00000495, -0.00001107, 0.00002464, -0.00002294, 0.00001920, -0.00001540, -0.00000288, -0.00000061, -0.00001410, 0.00001136, 0.00000218, -0.00000470, 0.00001925, -0.00000837, 0.00000271, -0.00000481, -0.00000586, 0.00001230, -0.00000221, -0.00001458, 0.00000120, -0.00001939, 0.00000444, 0.00001226, -0.00000443, },
		{ -0.00000370, 0.00000531, 0.00001161, -0.00001624, -0.00002981, 0.00002580, -0.00000636, 0.00001264, 0.00002576, -0.00002836, -0.00005146, 0.00005374, -0.00016955, 0.00007815, -0.00031676, -0.00019992, 0.00127504, -0.00038643, -0.00018268, 0.00003375, -0.00008653, 0.00014723, -0.00013858, 0.00018286, -0.00025228, 0.00014939, -0.00010144, 0.00006188, -0.00004325, 0.00002958, -0.00004110, 0.00002381, -0.00000716, 0.00000283, 0.00000274, -0.00004815, 0.00005438, -0.00003294, 0.00003026, -0.00002110, 0.00000501, -0.00001253, 0.00002576, -0.00002194, 0.00001680, -0.00001099, -0.00000666, 0.00000132, -0.00001389, 0.00000845, 0.00000434, -0.00000680, 0.00001919, -0.00000576, 0.00000067, -0.00000272, -0.00000702, 0.00001170, -0.00000078, -0.00001426, 0.00000117, -0.00001938, 0.00000152, 0.00001433, },
		{ 0.00001602, -0.00000267, 0.00000309, 0.00000974, -0.00000951, -0.00003251, 0.00002217, -0.00000430, 0.00001240, 0.00002374, -0.00001604, -0.00006907, 0.00006676, -0.00016250, 0.00005316, -0.00024741, -0.00034459, 0.00134207, -0.00029944, -0.00026609, 0.00008470, -0.00014597, 0.00018986, -0.00017719, 0.00020929, -0.00025335, 0.00013929, -0.00008425, 0.00004337, -0.00002895, 0.00001984, -0.00003048, 0.00000991, 0.00000469, -0.00000688, 0.00001644, -0.00006071, 0.00005758, -0.00003213, 0.00002840, -0.00001898, 0.00000471, -0.00001349, 0.00002597, -0.00001996, 0.00001372, -0.00000642, -0.00000993, 0.00000248, -0.00001300, 0.00000545, 0.00000605, -0.00000818, 0.00001829, -0.00000284, -0.00000123, -0.00000093, -0.00000777, 0.00001080, 0.00000061, -0.00001385, 0.00000132, -0.00001940, -0.00000138, },
		{ -0.00000435, 0.00001744, -0.00000156, 0.00000162, 0.00000679, -0.00000293, -0.00003401, 0.00001816, -0.00000303, 0.00001334, 0.00002002, -0.00000318, -0.00008337, 0.00007203, -0.00014466, 0.00001601, -0.00016656, -0.00048504, 0.00138111, -0.00018475, -0.00036228, 0.00014121, -0.00020602, 0.00022950, -0.00020943, 0.00022521, -0.00024416, 0.00012132, -0.00006166, 0.00002224, -0.00001381, 0.00000934, -0.00001875, -0.00000440, 0.00001625, -0.00001636, 0.00002927, -0.00007073, 0.00005791, -0.00002942, 0.00002542, -0.00001632, 0.00000412, -0.00001395, 0.00002529, -0.00001716, 0.00001021, -0.00000197, -0.00001251, 0.00000285, -0.00001156, 0.00000255, 0.00000719, -0.00000887, 0.00001674, 0.00000017, -0.00000284, 0.00000047, -0.00000811, 0.00000964, 0.00000193, -0.00001340, 0.00000169, -0.00001936, },
		{ -0.00001916, -0.00000750, 0.00001872, -0.00000055, 0.00000101, 0.00000296, 0.00000316, -0.00003425, 0.00001393, -0.00000260, 0.00001523, 0.00001532, 0.00000877, -0.00009266, 0.00006839, -0.00011611, -0.00003152, -0.00007855, -0.00061479, 0.00138771, -0.00004276, -0.00046813, 0.00019972, -0.00026318, 0.00026288, -0.00023249, 0.00022908, -0.00022461, 0.00009626, -0.00003487, -0.00000029, 0.00000130, -0.00000139, -0.00000644, -0.00001853, 0.00002696, -0.00002507, 0.00004061, -0.00007780, 0.00005537, -0.00002505, 0.00002157, -0.00001332, 0.00000332, -0.00001393, 0.00002383, -0.00001375, 0.00000654, 0.00000211, -0.00001428, 0.00000248, -0.00000977, -0.00000006, 0.00000773, -0.00000893, 0.00001472, 0.00000304, -0.00000402, 0.00000145, -0.00000808, 0.00000830, 0.00000317, -0.00001292, 0.00000224, },
		{ 0.00000292, -0.00001867, -0.00001091, 0.00001995, 0.00000022, 0.00000125, -0.00000148, 0.00000845, -0.00003320, 0.00000963, -0.00000293, 0.00001766, 0.00001054, 0.00001838, -0.00009554, 0.00005535, -0.00007781, -0.00008673, 0.00001165, -0.00072763, 0.00135868, 0.00012458, -0.00057938, 0.00025603, -0.00031378, 0.00028690, -0.00024405, 0.00022008, -0.00019532, 0.00006540, -0.00000536, -0.00002291, 0.00001556, -0.00001184, 0.00000589, -0.00003184, 0.00003630, -0.00003250, 0.00004995, -0.00008168, 0.00005013, -0.00001936, 0.00001715, -0.00001016, 0.00000240, -0.00001348, 0.00002169, -0.00000998, 0.00000299, 0.00000562, -0.00001519, 0.00000150, -0.00000780, -0.00000226, 0.00000765, -0.00000851, 0.00001245, 0.00000558, -0.00000471, 0.00000205, -0.00000772, 0.00000683, 0.00000428, -0.00001243, },
		{ -0.00001191, 0.00000362, -0.00001775, -0.00001466, 0.00002118, 0.00000070, 0.00000223, -0.00000615, 0.00001260, -0.00003086, 0.00000535, -0.00000385, 0.00002008, 0.00000664, 0.00002434, -0.00009110, 0.00003317, -0.00003151, -0.00014622, 0.00009878, -0.00081797, 0.00129242, 0.00031368, -0.00069076, 0.00030555, -0.00035420, 0.00029886, -0.00024254, 0.00019828, -0.00015762, 0.00003046, 0.00002517, -0.00004432, 0.00002821, -0.00002154, 0.00001770, -0.00004375, 0.00004383, -0.00003825, 0.00005691, -0.00008230, 0.00004253, -0.00001277, 0.00001246, -0.00000705, 0.00000146, -0.00001267, 0.00001906, -0.00000611, -0.00000021, 0.00000843, -0.00001525, 0.00000006, -0.00000584, -0.00000394, 0.00000701, -0.00000775, 0.00001013, 0.00000767, -0.00000491, 0.00000232, -0.00000714, 0.00000533, 0.00000523, },
		{ 0.00000598, -0.00001135, 0.00000424, -0.00001626, -0.00001879, 0.00002241, 0.00000093, 0.00000375, -0.00001065, 0.00001531, -0.00002724, 0.00000111, -0.00000506, 0.00002189, 0.00000456, 0.00002560, -0.00007896, 0.00000282, 0.00002027, -0.00020607, 0.00017764, -0.00088125, 0.00118906, 0.00051941, -0.00079628, 0.00034368, -0.00038117, 0.00029673, -0.00022727, 0.00016458, -0.00011343, -0.00000655, 0.00005499, -0.00006332, 0.00003865, -0.00003009, 0.00002849, -0.00005377, 0.00004917, -0.00004202, 0.00006129, -0.00007980, 0.00003303, -0.00000574, 0.00000782, -0.00000415, 0.00000057, -0.00001160, 0.00001610, -0.00000236, -0.00000287, 0.00001046, -0.00001456, -0.00000168, -0.00000405, -0.00000506, 0.00000592, -0.00000681, 0.00000793, 0.00000920, -0.00000465, 0.00000235, -0.00000642, 0.00000385, },
		{ 0.00000247, 0.00000651, -0.00001070, 0.00000466, -0.00001411, -0.00002326, 0.00002356, 0.00000101, 0.00000554, -0.00001458, 0.00001635, -0.00002240, -0.00000309, -0.00000621, 0.00002244, 0.00000513, 0.00002148, -0.00005937, -0.00003399, 0.00007447, -0.00026214, 0.00024346, -0.00091423, 0.00105057, 0.00073526, -0.00088952, 0.00036606, -0.00039206, 0.00027933, -0.00019849, 0.00012073, -0.00006515, -0.00004344, 0.00008242, -0.00007889, 0.00004641, -0.00003718, 0.00003782, -0.00006148, 0.00005211, -0.00004368, 0.00006306, -0.00007448, 0.00002218, 0.00000127, 0.00000351, -0.00000162, -0.00000017, -0.00001036, 0.00001301, 0.00000103, -0.00000487, 0.00001171, -0.00001323, -0.00000351, -0.00000258, -0.00000562, 0.00000450, -0.00000588, 0.00000600, 0.00001016, -0.00000403, 0.00000226, -0.00000565, },
		{ -0.00000491, 0.00000124, 0.00000679, -0.00000992, 0.00000479, -0.00001126, -0.00002798, 0.00002447, 0.00000113, 0.00000728, -0.00001751, 0.00001555, -0.00001643, -0.00000727, -0.00000692, 0.00002116, 0.00000895, 0.00001176, -0.00003325, -0.00007496, 0.00012766, -0.00031042, 0.00029223, -0.00091522, 0.00088074, 0.00095368, -0.00096406, 0.00036897, -0.00038514, 0.00024646, -0.00015744, 0.00006918, -0.00001550, -0.00007803, 0.00010597, -0.00009026, 0.00005125, -0.00004263, 0.00004534, -0.00006660, 0.00005254, -0.00004321, 0.00006234, -0.00006678, 0.00001059, 0.00000780, -0.00000023, 0.00000043, -0.00000071, -0.00000904, 0.00000995, 0.00000391, -0.00000615, 0.00001224, -0.00001144, -0.00000526, -0.00000150, -0.00000568, 0.00000291, -0.00000508, 0.00000441, 0.00001058, -0.00000317, 0.00000217, },
		{ 0.00000219, -0.00000428, 0.00000019, 0.00000680, -0.00000899, 0.00000454, -0.00000772, -0.00003275, 0.00002496, 0.00000150, 0.00000862, -0.00001910, 0.00001282, -0.00000947, -0.00001149, -0.00000682, 0.00001760, 0.00001634, -0.00000333, -0.00000206, -0.00011733, 0.00017633, -0.00034724, 0.00032098, -0.00088423, 0.00068503, 0.00116637, -0.00101386, 0.00034962, -0.00035976, 0.00019900, -0.00010624, 0.00001293, 0.00003266, -0.00010832, 0.00012444, -0.00009695, 0.00005311, -0.00004633, 0.00005080, -0.00006897, 0.00005054, -0.00004076, 0.00005941, -0.00005725, -0.00000111, 0.00001347, -0.00000322, 0.00000190, -0.00000099, -0.00000774, 0.00000708, 0.00000615, -0.00000671, 0.00001218, -0.00000937, -0.00000677, -0.00000085, -0.00000533, 0.00000131, -0.00000454, 0.00000321, 0.00001055, -0.00000223, },
		{ -0.00000135, 0.00000239, -0.00000377, -0.00000066, 0.00000655, -0.00000789, 0.00000387, -0.00000358, -0.00003734, 0.00002476, 0.00000237, 0.00000923, -0.00001908, 0.00000820, -0.00000170, -0.00001580, -0.00000558, 0.00001147, 0.00002730, -0.00002302, 0.00003219, -0.00015806, 0.00021710, -0.00036963, 0.00032797, -0.00082297, 0.00047032, 0.00136481, -0.00103369, 0.00030644, -0.00031644, 0.00013887, -0.00004780, -0.00004466, 0.00007658, -0.00013260, 0.00013701, -0.00009883, 0.00005215, -0.00004831, 0.00005406, -0.00006860, 0.00004631, -0.00003661, 0.00005466, -0.00004649, -0.00001233, 0.00001795, -0.00000535, 0.00000278, -0.00000100, -0.00000653, 0.00000452, 0.00000770, -0.00000662, 0.00001167, -0.00000722, -0.00000792, -0.00000062, -0.00000470, -0.00000017, -0.00000433, 0.00000237, 0.00001018, },
		{ 0.00000962, -0.00000068, 0.00000284, -0.00000340, -0.00000132, 0.00000606, -0.00000660, 0.00000278, 0.00000099, -0.00004144, 0.00002361, 0.00000399, 0.00000884, -0.00001729, 0.00000182, 0.00000663, -0.00002022, -0.00000299, 0.00000271, 0.00004145, -0.00004613, 0.00006716, -0.00019413, 0.00024705, -0.00037551, 0.00031287, -0.00073475, 0.00024451, 0.00154068, -0.00101952, 0.00023925, -0.00025694, 0.00006891, 0.00001441, -0.00010017, 0.00011379, -0.00014961, 0.00014326, -0.00009605, 0.00004870, -0.00004867, 0.00005508, -0.00006563, 0.00004018, -0.00003112, 0.00004856, -0.00003515, -0.00002253, 0.00002102, -0.00000658, 0.00000306, -0.00000073, -0.00000547, 0.00000235, 0.00000854, -0.00000600, 0.00001090, -0.00000516, -0.00000863, -0.00000076, -0.00000391, -0.00000139, -0.00000447, 0.00000182, },
		{ 0.00000146, 0.00000903, -0.00000032, 0.00000355, -0.00000313, -0.00000182, 0.00000535, -0.00000516, 0.00000133, 0.00000578, -0.00004472, 0.00002125, 0.00000654, 0.00000725, -0.00001367, -0.00000608, 0.00001526, -0.00002475, 0.00000106, -0.00000848, 0.00005810, -0.00007110, 0.00010032, -0.00022267, 0.00026388, -0.00036386, 0.00027672, -0.00062431, 0.00001616, 0.00168637, -0.00096883, 0.00014937, -0.00018416, -0.00000726, 0.00007661, -0.00015027, 0.00014226, -0.00015860, 0.00014324, -0.00008909, 0.00004323, -0.00004759, 0.00005394, -0.00006033, 0.00003261, -0.00002475, 0.00004162, -0.00002381, -0.00003129, 0.00002256, -0.00000694, 0.00000278, -0.00000021, -0.00000458, 0.00000061, 0.00000875, -0.00000501, 0.00001003, -0.00000331, -0.00000889, -0.00000116, -0.00000313, -0.00000228, -0.00000493, },
		{ -0.00000565, 0.00000117, 0.00000854, -0.00000035, 0.00000448, -0.00000290, -0.00000221, 0.00000447, -0.00000361, -0.00000038, 0.00001050, -0.00004684, 0.00001748, 0.00001016, 0.00000435, -0.00000833, -0.00001512, 0.00002385, -0.00002934, 0.00000652, -0.00002167, 0.00007623, -0.00009611, 0.00012912, -0.00024125, 0.00026615, -0.00033483, 0.00022189, -0.00049749, -0.00020605, 0.00179542, -0.00088086, 0.00003963, -0.00010201, -0.00008551, 0.00013494, -0.00019207, 0.00016059, -0.00015940, 0.00013739, -0.00007868, 0.00003633, -0.00004531, 0.00005079, -0.00005311, 0.00002409, -0.00001799, 0.00003435, -0.00001303, -0.00003828, 0.00002258, -0.00000652, 0.00000204, 0.00000054, -0.00000387, -0.00000071, 0.00000841, -0.00000382, 0.00000923, -0.00000177, -0.00000871, -0.00000171, -0.00000246, -0.00000278, },
		{ -0.00000291, -0.00000652, 0.00000081, 0.00000827, -0.00000077, 0.00000556, -0.00000264, -0.00000255, 0.00000348, -0.00000202, -0.00000220, 0.00001484, -0.00004747, 0.00001215, 0.00001488, 0.00000018, -0.00000149, -0.00002488, 0.00003207, -0.00003387, 0.00001318, -0.00003619, 0.00009461, -0.00011923, 0.00015122, -0.00024808, 0.00025338, -0.00028970, 0.00015192, -0.00036086, -0.00041376, 0.00186291, -0.00075673, -0.00008575, -0.00001517, -0.00016146, 0.00018581, -0.00022321, 0.00016805, -0.00015241, 0.00012655, -0.00006571, 0.00002861, -0.00004209, 0.00004589, -0.00004443, 0.00001521, -0.00001129, 0.00002722, -0.00000324, -0.00004332, 0.00002116, -0.00000547, 0.00000092, 0.00000144, -0.00000332, -0.00000164, 0.00000767, -0.00000261, 0.00000862, -0.00000059, -0.00000819, -0.00000227, -0.00000202, },
		{ -0.00000186, -0.00000271, -0.00000741, 0.00000025, 0.00000829, -0.00000155, 0.00000667, -0.00000225, -0.00000290, 0.00000245, -0.00000047, -0.00000397, 0.00001850, -0.00004636, 0.00000522, 0.00002063, -0.00000512, 0.00000651, -0.00003483, 0.00003953, -0.00003819, 0.00002071, -0.00005120, 0.00011185, -0.00013856, 0.00016467, -0.00024213, 0.00022609, -0.00023083, 0.00007122, -0.00022134, -0.00059944, 0.00188575, -0.00059944, -0.00022134, 0.00007122, -0.00023083, 0.00022609, -0.00024213, 0.00016467, -0.00013856, 0.00011185, -0.00005120, 0.00002071, -0.00003819, 0.00003953, -0.00003483, 0.00000651, -0.00000512, 0.00002063, 0.00000522, -0.00004636, 0.00001850, -0.00000397, -0.00000047, 0.00000245, -0.00000290, -0.00000225, 0.00000667, -0.00000155, 0.00000829, 0.00000025, -0.00000741, -0.00000271, },
		{ -0.00000291, -0.00000202, -0.00000227, -0.00000819, -0.00000059, 0.00000862, -0.00000261, 0.00000767, -0.00000164, -0.00000332, 0.00000144, 0.00000092, -0.00000547, 0.00002116, -0.00004332, -0.00000324, 0.00002722, -0.00001129, 0.00001521, -0.00004443, 0.00004589, -0.00004209, 0.00002861, -0.00006571, 0.00012655, -0.00015241, 0.00016805, -0.00022321, 0.00018581, -0.00016146, -0.00001517, -0.00008575, -0.00075673, 0.00186291, -0.00041376, -0.00036086, 0.00015192, -0.00028970, 0.00025338, -0.00024808, 0.00015122, -0.00011923, 0.00009461, -0.00003619, 0.00001318, -0.00003387, 0.00003207, -0.00002488, -0.00000149, 0.00000018, 0.00001488, 0.00001215, -0.00004747, 0.00001484, -0.00000220, -0.00000202, 0.00000348, -0.00000255, -0.00000264, 0.00000556, -0.00000077, 0.00000827, 0.00000081, -0.00000652, },
		{ -0.00000565, -0.00000278, -0.00000246, -0.00000171, -0.00000871, -0.00000177, 0.00000923, -0.00000382, 0.00000841, -0.00000071, -0.00000387, 0.00000054, 0.00000204, -0.00000652, 0.00002258, -0.00003828, -0.00001303, 0.00003435, -0.00001799, 0.00002409, -0.00005311, 0.00005079, -0.00004531, 0.00003633, -0.00007868, 0.00013739, -0.00015940, 0.00016059, -0.00019207, 0.00013494, -0.00008551, -0.00010201, 0.00003963, -0.00088086, 0.00179542, -0.00020605, -0.00049749, 0.00022189, -0.00033483, 0.00026615, -0.00024125, 0.00012912, -0.00009611, 0.00007623, -0.00002167, 0.00000652, -0.00002934, 0.00002385, -0.00001512, -0.00000833, 0.00000435, 0.00001016, 0.00001748, -0.00004684, 0.00001050, -0.00000038, -0.00000361, 0.00000447, -0.00000221, -0.00000290, 0.00000448, -0.00000035, 0.00000854, 0.00000117, },
		{ 0.00000146, -0.00000493, -0.00000228, -0.00000313, -0.00000116, -0.00000889, -0.00000331, 0.00001003, -0.00000501, 0.00000875, 0.00000061, -0.00000458, -0.00000021, 0.00000278, -0.00000694, 0.00002256, -0.00003129, -0.00002381, 0.00004162, -0.00002475, 0.00003261, -0.00006033, 0.00005394, -0.00004759, 0.00004323, -0.00008909, 0.00014324, -0.00015860, 0.00014226, -0.00015027, 0.00007661, -0.00000726, -0.00018416, 0.00014937, -0.00096883, 0.00168637, 0.00001616, -0.00062431, 0.00027672, -0.00036386, 0.00026388, -0.00022267, 0.00010032, -0.00007110, 0.00005810, -0.00000848, 0.00000106, -0.00002475, 0.00001526, -0.00000608, -0.00001367, 0.00000725, 0.00000654, 0.00002125, -0.00004472, 0.00000578, 0.00000133, -0.00000516, 0.00000535, -0.00000182, -0.00000313, 0.00000355, -0.00000032, 0.00000903, },
		{ 0.00000962, 0.00000182, -0.00000447, -0.00000139, -0.00000391, -0.00000076, -0.00000863, -0.00000516, 0.00001090, -0.00000600, 0.00000854, 0.00000235, -0.00000547, -0.00000073, 0.00000306, -0.00000658, 0.00002102, -0.00002253, -0.00003515, 0.00004856, -0.00003112, 0.00004018, -0.00006563, 0.00005508, -0.00004867, 0.00004870, -0.00009605, 0.00014326, -0.00014961, 0.00011379, -0.00010017, 0.00001441, 0.00006891, -0.00025694, 0.00023925, -0.00101952, 0.00154068, 0.00024451, -0.00073475, 0.00031287, -0.00037551, 0.00024705, -0.00019413, 0.00006716, -0.00004613, 0.00004145, 0.00000271, -0.00000299, -0.00002022, 0.00000663, 0.00000182, -0.00001729, 0.00000884, 0.00000399, 0.00002361, -0.00004144, 0.00000099, 0.00000278, -0.00000660, 0.00000606, -0.00000132, -0.00000340, 0.00000284, -0.00000068, },
		{ -0.00000135, 0.00001018, 0.00000237, -0.00000433, -0.00000017, -0.00000470, -0.00000062, -0.00000792, -0.00000722, 0.00001167, -0.00000662, 0.00000770, 0.00000452, -0.00000653, -0.00000100, 0.00000278, -0.00000535, 0.00001795, -0.00001233, -0.00004649, 0.00005466, -0.00003661, 0.00004631, -0.00006860, 0.00005406, -0.00004831, 0.00005215, -0.00009883, 0.00013701, -0.00013260, 0.00007658, -0.00004466, -0.00004780, 0.00013887, -0.00031644, 0.00030644, -0.00103369, 0.00136481, 0.00047032, -0.00082297, 0.00032797, -0.00036963, 0.00021710, -0.00015806, 0.00003219, -0.00002302, 0.00002730, 0.00001147, -0.00000558, -0.00001580, -0.00000170, 0.00000820, -0.00001908, 0.00000923, 0.00000237, 0.00002476, -0.00003734, -0.00000358, 0.00000387, -0.00000789, 0.00000655, -0.00000066, -0.00000377, 0.00000239, },
		{ 0.00000219, -0.00000223, 0.00001055, 0.00000321, -0.00000454, 0.00000131, -0.00000533, -0.00000085, -0.00000677, -0.00000937, 0.00001218, -0.00000671, 0.00000615, 0.00000708, -0.00000774, -0.00000099, 0.00000190, -0.00000322, 0.00001347, -0.00000111, -0.00005725, 0.00005941, -0.00004076, 0.00005054, -0.00006897, 0.00005080, -0.00004633, 0.00005311, -0.00009695, 0.00012444, -0.00010832, 0.00003266, 0.00001293, -0.00010624, 0.00019900, -0.00035976, 0.00034962, -0.00101386, 0.00116637, 0.00068503, -0.00088423, 0.00032098, -0.00034724, 0.00017633, -0.00011733, -0.00000206, -0.00000333, 0.00001634, 0.00001760, -0.00000682, -0.00001149, -0.00000947, 0.00001282, -0.00001910, 0.00000862, 0.00000150, 0.00002496, -0.00003275, -0.00000772, 0.00000454, -0.00000899, 0.00000680, 0.00000019, -0.00000428, },
		{ -0.00000491, 0.00000217, -0.00000317, 0.00001058, 0.00000441, -0.00000508, 0.00000291, -0.00000568, -0.00000150, -0.00000526, -0.00001144, 0.00001224, -0.00000615, 0.00000391, 0.00000995, -0.00000904, -0.00000071, 0.00000043, -0.00000023, 0.00000780, 0.00001059, -0.00006678, 0.00006234, -0.00004321, 0.00005254, -0.00006660, 0.00004534, -0.00004263, 0.00005125, -0.00009026, 0.00010597, -0.00007803, -0.00001550, 0.00006918, -0.00015744, 0.00024646, -0.00038514, 0.00036897, -0.00096406, 0.00095368, 0.00088074, -0.00091522, 0.00029223, -0.00031042, 0.00012766, -0.00007496, -0.00003325, 0.00001176, 0.00000895, 0.00002116, -0.00000692, -0.00000727, -0.00001643, 0.00001555, -0.00001751, 0.00000728, 0.00000113, 0.00002447, -0.00002798, -0.00001126, 0.00000479, -0.00000992, 0.00000679, 0.00000124, },
		{ 0.00000247, -0.00000565, 0.00000226, -0.00000403, 0.00001016, 0.00000600, -0.00000588, 0.00000450, -0.00000562, -0.00000258, -0.00000351, -0.00001323, 0.00001171, -0.00000487, 0.00000103, 0.00001301, -0.00001036, -0.00000017, -0.00000162, 0.00000351, 0.00000127, 0.00002218, -0.00007448, 0.00006306, -0.00004368, 0.00005211, -0.00006148, 0.00003782, -0.00003718, 0.00004641, -0.00007889, 0.00008242, -0.00004344, -0.00006515, 0.00012073, -0.00019849, 0.00027933, -0.00039206, 0.00036606, -0.00088952, 0.00073526, 0.00105057, -0.00091423, 0.00024346, -0.00026214, 0.00007447, -0.00003399, -0.00005937, 0.00002148, 0.00000513, 0.00002244, -0.00000621, -0.00000309, -0.00002240, 0.00001635, -0.00001458, 0.00000554, 0.00000101, 0.00002356, -0.00002326, -0.00001411, 0.00000466, -0.00001070, 0.00000651, },
		{ 0.00000598, 0.00000385, -0.00000642, 0.00000235, -0.00000465, 0.00000920, 0.00000793, -0.00000681, 0.00000592, -0.00000506, -0.00000405, -0.00000168, -0.00001456, 0.00001046, -0.00000287, -0.00000236, 0.00001610, -0.00001160, 0.00000057, -0.00000415, 0.00000782, -0.00000574, 0.00003303, -0.00007980, 0.00006129, -0.00004202, 0.00004917, -0.00005377, 0.00002849, -0.00003009, 0.00003865, -0.00006332, 0.00005499, -0.00000655, -0.00011343, 0.00016458, -0.00022727, 0.00029673, -0.00038117, 0.00034368, -0.00079628, 0.00051941, 0.00118906, -0.00088125, 0.00017764, -0.00020607, 0.00002027, 0.00000282, -0.00007896, 0.00002560, 0.00000456, 0.00002189, -0.00000506, 0.00000111, -0.00002724, 0.00001531, -0.00001065, 0.00000375, 0.00000093, 0.00002241, -0.00001879, -0.00001626, 0.00000424, -0.00001135, },
		{ -0.00001191, 0.00000523, 0.00000533, -0.00000714, 0.00000232, -0.00000491, 0.00000767, 0.00001013, -0.00000775, 0.00000701, -0.00000394, -0.00000584, 0.00000006, -0.00001525, 0.00000843, -0.00000021, -0.00000611, 0.00001906, -0.00001267, 0.00000146, -0.00000705, 0.00001246, -0.00001277, 0.00004253, -0.00008230, 0.00005691, -0.00003825, 0.00004383, -0.00004375, 0.00001770, -0.00002154, 0.00002821, -0.00004432, 0.00002517, 0.00003046, -0.00015762, 0.00019828, -0.00024254, 0.00029886, -0.00035420, 0.00030555, -0.00069076, 0.00031368, 0.00129242, -0.00081797, 0.00009878, -0.00014622, -0.00003151, 0.00003317, -0.00009110, 0.00002434, 0.00000664, 0.00002008, -0.00000385, 0.00000535, -0.00003086, 0.00001260, -0.00000615, 0.00000223, 0.00000070, 0.00002118, -0.00001466, -0.00001775, 0.00000362, },
		{ 0.00000292, -0.00001243, 0.00000428, 0.00000683, -0.00000772, 0.00000205, -0.00000471, 0.00000558, 0.00001245, -0.00000851, 0.00000765, -0.00000226, -0.00000780, 0.00000150, -0.00001519, 0.00000562, 0.00000299, -0.00000998, 0.00002169, -0.00001348, 0.00000240, -0.00001016, 0.00001715, -0.00001936, 0.00005013, -0.00008168, 0.00004995, -0.00003250, 0.00003630, -0.00003184, 0.00000589, -0.00001184, 0.00001556, -0.00002291, -0.00000536, 0.00006540, -0.00019532, 0.00022008, -0.00024405, 0.00028690, -0.00031378, 0.00025603, -0.00057938, 0.00012458, 0.00135868, -0.00072763, 0.00001165, -0.00008673, -0.00007781, 0.00005535, -0.00009554, 0.00001838, 0.00001054, 0.00001766, -0.00000293, 0.00000963, -0.00003320, 0.00000845, -0.00000148, 0.00000125, 0.00000022, 0.00001995, -0.00001091, -0.00001867, },
		{ -0.00001916, 0.00000224, -0.00001292, 0.00000317, 0.00000830, -0.00000808, 0.00000145, -0.00000402, 0.00000304, 0.00001472, -0.00000893, 0.00000773, -0.00000006, -0.00000977, 0.00000248, -0.00001428, 0.00000211, 0.00000654, -0.00001375, 0.00002383, -0.00001393, 0.00000332, -0.00001332, 0.00002157, -0.00002505, 0.00005537, -0.00007780, 0.00004061, -0.00002507, 0.00002696, -0.00001853, -0.00000644, -0.00000139, 0.00000130, -0.00000029, -0.00003487, 0.00009626, -0.00022461, 0.00022908, -0.00023249, 0.00026288, -0.00026318, 0.00019972, -0.00046813, -0.00004276, 0.00138771, -0.00061479, -0.00007855, -0.00003152, -0.00011611, 0.00006839, -0.00009266, 0.00000877, 0.00001532, 0.00001523, -0.00000260, 0.00001393, -0.00003425, 0.00000316, 0.00000296, 0.00000101, -0.00000055, 0.00001872, -0.00000750, },
		{ -0.00000435, -0.00001936, 0.00000169, -0.00001340, 0.00000193, 0.00000964, -0.00000811, 0.00000047, -0.00000284, 0.00000017, 0.00001674, -0.00000887, 0.00000719, 0.00000255, -0.00001156, 0.00000285, -0.00001251, -0.00000197, 0.00001021, -0.00001716, 0.00002529, -0.00001395, 0.00000412, -0.00001632, 0.00002542, -0.00002942, 0.00005791, -0.00007073, 0.00002927, -0.00001636, 0.00001625, -0.00000440, -0.00001875, 0.00000934, -0.00001381, 0.00002224, -0.00006166, 0.00012132, -0.00024416, 0.00022521, -0.00020943, 0.00022950, -0.00020602, 0.00014121, -0.00036228, -0.00018475, 0.00138111, -0.00048504, -0.00016656, 0.00001601, -0.00014466, 0.00007203, -0.00008337, -0.00000318, 0.00002002, 0.00001334, -0.00000303, 0.00001816, -0.00003401, -0.00000293, 0.00000679, 0.00000162, -0.00000156, 0.00001744, },
		{ 0.00001602, -0.00000138, -0.00001940, 0.00000132, -0.00001385, 0.00000061, 0.00001080, -0.00000777, -0.00000093, -0.00000123, -0.00000284, 0.00001829, -0.00000818, 0.00000605, 0.00000545, -0.00001300, 0.00000248, -0.00000993, -0.00000642, 0.00001372, -0.00001996, 0.00002597, -0.00001349, 0.00000471, -0.00001898, 0.00002840, -0.00003213, 0.00005758, -0.00006071, 0.00001644, -0.00000688, 0.00000469, 0.00000991, -0.00003048, 0.00001984, -0.00002895, 0.00004337, -0.00008425, 0.00013929, -0.00025335, 0.00020929, -0.00017719, 0.00018986, -0.00014597, 0.00008470, -0.00026609, -0.00029944, 0.00134207, -0.00034459, -0.00024741, 0.00005316, -0.00016250, 0.00006676, -0.00006907, -0.00001604, 0.00002374, 0.00001240, -0.00000430, 0.00002217, -0.00003251, -0.00000951, 0.00000974, 0.00000309, -0.00000267, },
		{ -0.00000370, 0.00001433, 0.00000152, -0.00001938, 0.00000117, -0.00001426, -0.00000078, 0.00001170, -0.00000702, -0.00000272, 0.00000067, -0.00000576, 0.00001919, -0.00000680, 0.00000434, 0.00000845, -0.00001389, 0.00000132, -0.00000666, -0.00001099, 0.00001680, -0.00002194, 0.00002576, -0.00001253, 0.00000501, -0.00002110, 0.00003026, -0.00003294, 0.00005438, -0.00004815, 0.00000274, 0.00000283, -0.00000716, 0.00002381, -0.00004110, 0.00002958, -0.00004325, 0.00006188, -0.00010144, 0.00014939, -0.00025228, 0.00018286, -0.00013858, 0.00014723, -0.00008653, 0.00003375, -0.00018268, -0.00038643, 0.00127504, -0.00019992, -0.00031676, 0.00007815, -0.00016955, 0.00005374, -0.00005146, -0.00002836, 0.00002576, 0.00001264, -0.00000636, 0.00002580, -0.00002981, -0.00001624, 0.00001161, 0.00000531, },
		{ 0.00000811, -0.00000443, 0.00001226, 0.00000444, -0.00001939, 0.00000120, -0.00001458, -0.00000221, 0.00001230, -0.00000586, -0.00000481, 0.00000271, -0.00000837, 0.00001925, -0.00000470, 0.00000218, 0.00001136, -0.00001410, -0.00000061, -0.00000288, -0.00001540, 0.00001920, -0.00002294, 0.00002464, -0.00001107, 0.00000495, -0.00002253, 0.00003082, -0.00003174, 0.00004846, -0.00003364, -0.00001111, 0.00001218, -0.00001875, 0.00003669, -0.00005011, 0.00003807, -0.00005592, 0.00007674, -0.00011239, 0.00015141, -0.00024170, 0.00014811, -0.00009672, 0.00010473, -0.00003069, -0.00000884, -0.00011392, -0.00044681, 0.00118539, -0.00005733, -0.00037111, 0.00009018, -0.00016655, 0.00003462, -0.00003240, -0.00003884, 0.00002558, 0.00001411, -0.00000902, 0.00002884, -0.00002598, -0.00002284, 0.00001228, },
		{ 0.00001176, 0.00001121, -0.00000463, 0.00000971, 0.00000742, -0.00001943, 0.00000136, -0.00001474, -0.00000366, 0.00001260, -0.00000433, -0.00000710, 0.00000467, -0.00001042, 0.00001838, -0.00000193, -0.00000028, 0.00001398, -0.00001352, -0.00000323, 0.00000117, -0.00001935, 0.00002068, -0.00002285, 0.00002262, -0.00000916, 0.00000449, -0.00002316, 0.00002996, -0.00002853, 0.00004017, -0.00001787, -0.00002438, 0.00002063, -0.00002957, 0.00004803, -0.00005709, 0.00004484, -0.00006623, 0.00008714, -0.00011671, 0.00014570, -0.00022298, 0.00010763, -0.00005475, 0.00006513, 0.00001915, -0.00004122, -0.00006049, -0.00048285, 0.00107902, 0.00007738, -0.00040810, 0.00008947, -0.00015493, 0.00001140, -0.00001375, -0.00004642, 0.00002296, 0.00001666, -0.00001201, 0.00003107, -0.00002112, -0.00002905, },
		{ -0.00003467, 0.00001012, 0.00001432, -0.00000408, 0.00000663, 0.00001043, -0.00001947, 0.00000156, -0.00001467, -0.00000514, 0.00001258, -0.00000251, -0.00000942, 0.00000636, -0.00001174, 0.00001650, 0.00000142, -0.00000284, 0.00001611, -0.00001210, -0.00000642, 0.00000522, -0.00002257, 0.00002105, -0.00002163, 0.00001980, -0.00000686, 0.00000360, -0.00002290, 0.00002765, -0.00002345, 0.00002996, -0.00000163, -0.00003637, 0.00002771, -0.00003918, 0.00005740, -0.00006173, 0.00004955, -0.00007362, 0.00009259, -0.00011447, 0.00013314, -0.00019796, 0.00006427, -0.00001562, 0.00003063, 0.00006138, -0.00006244, -0.00002199, -0.00049777, 0.00096198, 0.00019926, -0.00042654, 0.00007719, -0.00013668, -0.00001375, 0.00000282, -0.00005039, 0.00001799, 0.00001996, -0.00001500, 0.00003228, -0.00001533, },
		{ -0.00000876, -0.00003952, 0.00000752, 0.00001712, -0.00000264, 0.00000304, 0.00001342, -0.00001942, 0.00000166, -0.00001430, -0.00000663, 0.00001229, -0.00000051, -0.00001159, 0.00000756, -0.00001215, 0.00001363, 0.00000517, -0.00000530, 0.00001760, -0.00000985, -0.00000997, 0.00000898, -0.00002481, 0.00002022, -0.00001931, 0.00001631, -0.00000428, 0.00000228, -0.00002174, 0.00002396, -0.00001677, 0.00001843, 0.00001428, -0.00004642, 0.00003304, -0.00004726, 0.00006448, -0.00006381, 0.00005194, -0.00007771, 0.00009294, -0.00010617, 0.00011503, -0.00016879, 0.00002083, 0.00001817, 0.00000274, 0.00009519, -0.00007245, 0.00000291, -0.00049542, 0.00084010, 0.00030446, -0.00042643, 0.00005530, -0.00011412, -0.00003871, 0.00001594, -0.00005041, 0.00001101, 0.00002356, -0.00001760, 0.00003230, },
		{ 0.00003101, -0.00000158, -0.00004350, 0.00000418, 0.00001930, -0.00000022, -0.00000094, 0.00001624, -0.00001914, 0.00000153, -0.00001356, -0.00000815, 0.00001177, 0.00000153, -0.00001342, 0.00000807, -0.00001156, 0.00000986, 0.00000912, -0.00000744, 0.00001831, -0.00000685, -0.00001363, 0.00001217, -0.00002589, 0.00001816, -0.00001603, 0.00001232, -0.00000152, 0.00000057, -0.00001971, 0.00001904, -0.00000886, 0.00000623, 0.00002906, -0.00005401, 0.00003638, -0.00005358, 0.00006905, -0.00006324, 0.00005190, -0.00007834, 0.00008837, -0.00009273, 0.00009301, -0.00013775, -0.00002007, 0.00004472, -0.00001785, 0.00012051, -0.00007193, 0.00001618, -0.00047991, 0.00071865, 0.00039033, -0.00040897, 0.00002637, -0.00008972, -0.00006157, 0.00002467, -0.00004657, 0.00000264, 0.00002691, -0.00001945, },
		{ -0.00002022, 0.00002835, 0.00000604, -0.00004652, 0.00000032, 0.00002063, 0.00000316, -0.00000514, 0.00001870, -0.00001847, 0.00000103, -0.00001241, -0.00000968, 0.00001110, 0.00000346, -0.00001474, 0.00000777, -0.00000995, 0.00000536, 0.00001302, -0.00000906, 0.00001820, -0.00000324, -0.00001715, 0.00001453, -0.00002567, 0.00001491, -0.00001196, 0.00000805, 0.00000130, -0.00000149, -0.00001691, 0.00001313, -0.00000015, -0.00000594, 0.00004201, -0.00005876, 0.00003763, -0.00005803, 0.00007104, -0.00006008, 0.00004946, -0.00007556, 0.00007939, -0.00007538, 0.00006893, -0.00010705, -0.00005623, 0.00006286, -0.00003120, 0.00013798, -0.00006214, 0.00002025, -0.00045533, 0.00060210, 0.00045554, -0.00037630, -0.00000666, -0.00006582, -0.00008080, 0.00002855, -0.00003935, -0.00000634, 0.00002944, },
		{ 0.00003063, -0.00001969, 0.00002436, 0.00001385, -0.00004854, -0.00000379, 0.00002092, 0.00000739, -0.00000931, 0.00002059, -0.00001724, 0.00000004, -0.00001081, -0.00001124, 0.00001035, 0.00000513, -0.00001537, 0.00000656, -0.00000739, 0.00000034, 0.00001660, -0.00000998, 0.00001724, 0.00000079, -0.00002025, 0.00001587, -0.00002414, 0.00001063, -0.00000734, 0.00000374, 0.00000403, -0.00000382, -0.00001346, 0.00000651, 0.00000885, -0.00001741, 0.00005252, -0.00006047, 0.00003683, -0.00006063, 0.00007050, -0.00005450, 0.00004479, -0.00006966, 0.00006681, -0.00005555, 0.00004463, -0.00007870, -0.00008600, 0.00007219, -0.00003809, 0.00014870, -0.00004470, 0.00001769, -0.00042547, 0.00049394, 0.00049997, -0.00033137, -0.00004075, -0.00004448, -0.00009537, 0.00002763, -0.00002956, -0.00001504, },
		{ -0.00002255, 0.00003005, -0.00001774, 0.00001917, 0.00002161, -0.00004953, -0.00000794, 0.00002009, 0.00001224, -0.00001313, 0.00002167, -0.00001530, -0.00000154, -0.00000877, -0.00001280, 0.00000961, 0.00000638, -0.00001521, 0.00000442, -0.00000401, -0.00000491, 0.00001960, -0.00001008, 0.00001550, 0.00000502, -0.00002267, 0.00001603, -0.00002134, 0.00000553, -0.00000245, -0.00000038, 0.00000656, -0.00000633, -0.00000954, -0.00000048, 0.00001764, -0.00002758, 0.00006016, -0.00005910, 0.00003418, -0.00006152, 0.00006759, -0.00004677, 0.00003821, -0.00006112, 0.00005162, -0.00003481, 0.00002187, -0.00005433, -0.00010836, 0.00007308, -0.00003989, 0.00015411, -0.00002137, 0.00001102, -0.00039362, 0.00039660, 0.00052465, -0.00027759, -0.00007301, -0.00002727, -0.00010479, 0.00002240, -0.00001824, },
		{ -0.00000660, -0.00002806, 0.00002743, -0.00001439, 0.00001300, 0.00002904, -0.00004947, -0.00001195, 0.00001818, 0.00001739, -0.00001624, 0.00002170, -0.00001252, -0.00000375, -0.00000632, -0.00001437, 0.00000897, 0.00000712, -0.00001419, 0.00000141, -0.00000000, -0.00001008, 0.00002178, -0.00000930, 0.00001308, 0.00000919, -0.00002419, 0.00001497, -0.00001741, -0.00000011, 0.00000239, -0.00000409, 0.00000879, -0.00000894, -0.00000538, -0.00000748, 0.00002572, -0.00003591, 0.00006466, -0.00005485, 0.00003000, -0.00006089, 0.00006254, -0.00003728, 0.00003012, -0.00005056, 0.00003497, -0.00001469, 0.00000216, -0.00003507, -0.00012300, 0.00006656, -0.00003834, 0.00015575, 0.00000614, 0.00000248, -0.00036240, 0.00031147, 0.00053150, -0.00021864, -0.00010094, -0.00001519, -0.00010915, 0.00001379, },
		{ 0.00000299, 0.00000416, -0.00003092, 0.00002270, -0.00000983, 0.00000616, 0.00003585, -0.00004836, -0.00001568, 0.00001532, 0.00002244, -0.00001830, 0.00002048, -0.00000883, -0.00000658, -0.00000354, -0.00001591, 0.00000850, 0.00000724, -0.00001231, -0.00000233, 0.00000435, -0.00001487, 0.00002295, -0.00000764, 0.00001016, 0.00001303, -0.00002463, 0.00001271, -0.00001258, -0.00000598, 0.00000690, -0.00000719, 0.00001062, -0.00001152, -0.00000117, -0.00001412, 0.00003266, -0.00004202, 0.00006595, -0.00004805, 0.00002470, -0.00005903, 0.00005569, -0.00002649, 0.00002103, -0.00003875, 0.00001806, 0.00000341, -0.00001336, -0.00002153, -0.00013025, 0.00005423, -0.00003540, 0.00015504, 0.00003636, -0.00000606, -0.00033373, 0.00023898, 0.00052312, -0.00015810, -0.00012260, -0.00000856, -0.00010906, },
		{ -0.00010553, -0.00000866, 0.00001295, -0.00003073, 0.00001601, -0.00000436, -0.00000096, 0.00004174, -0.00004619, -0.00001907, 0.00001175, 0.00002696, -0.00001896, 0.00001789, -0.00000422, -0.00000996, -0.00000054, -0.00001738, 0.00000825, 0.00000671, -0.00000964, -0.00000661, 0.00000875, -0.00001896, 0.00002297, -0.00000515, 0.00000695, 0.00001632, -0.00002391, 0.00000936, -0.00000713, -0.00001172, 0.00001079, -0.00000952, 0.00001200, -0.00001399, 0.00000284, -0.00002009, 0.00003809, -0.00004567, 0.00006415, -0.00003918, 0.00001876, -0.00005624, 0.00004741, -0.00001490, 0.00001151, -0.00002647, 0.00000204, 0.00001833, -0.00002397, -0.00001375, -0.00013101, 0.00003805, -0.00003297, 0.00015310, 0.00006818, -0.00001321, -0.00030884, 0.00017874, 0.00050251, -0.00009921, -0.00013679, -0.00000709, },
		{ -0.00000991, -0.00009988, -0.00001979, 0.00001889, -0.00002737, 0.00000772, 0.00000159, -0.00000794, 0.00004641, -0.00004296, -0.00002211, 0.00000778, 0.00003050, -0.00001797, 0.00001388, 0.00000123, -0.00001374, 0.00000255, -0.00001875, 0.00000827, 0.00000554, -0.00000630, -0.00001118, 0.00001287, -0.00002210, 0.00002178, -0.00000198, 0.00000368, 0.00001883, -0.00002199, 0.00000512, -0.00000139, -0.00001701, 0.00001382, -0.00001100, 0.00001290, -0.00001624, 0.00000645, -0.00002510, 0.00004177, -0.00004679, 0.00005956, -0.00002887, 0.00001269, -0.00005286, 0.00003810, -0.00000305, 0.00000213, -0.00001451, -0.00001205, 0.00002921, -0.00002942, -0.00001122, -0.00012664, 0.00002012, -0.00003271, 0.00015063, 0.00010087, -0.00001805, -0.00028832, 0.00012975, 0.00047280, -0.00004468, -0.00014303, },
		{ -0.00014163, -0.00001567, -0.00009357, -0.00002919, 0.00002142, -0.00002104, -0.00000160, 0.00000748, -0.00001432, 0.00004959, -0.00003865, -0.00002487, 0.00000377, 0.00003262, -0.00001514, 0.00000850, 0.00000735, -0.00001776, 0.00000557, -0.00001997, 0.00000855, 0.00000379, -0.00000247, -0.00001574, 0.00001641, -0.00002407, 0.00001940, 0.00000169, 0.00000058, 0.00002041, -0.00001894, 0.00000026, 0.00000428, -0.00002153, 0.00001582, -0.00001157, 0.00001334, -0.00001820, 0.00000948, -0.00002894, 0.00004355, -0.00004546, 0.00005263, -0.00001776, 0.00000697, -0.00004916, 0.00002815, 0.00000856, -0.00000655, -0.00000362, -0.00002338, 0.00003557, -0.00002995, -0.00001303, -0.00011880, 0.00000251, -0.00003584, 0.00014780, 0.00013416, -0.00002005, -0.00027228, 0.00009062, 0.00043694, 0.00000348, },
		{ 0.00004404, -0.00013358, -0.00002277, -0.00008800, -0.00003591, 0.00002036, -0.00001228, -0.00001124, 0.00001276, -0.00001966, 0.00005105, -0.00003327, -0.00002745, 0.00000013, 0.00003297, -0.00001041, 0.00000192, 0.00001392, -0.00002177, 0.00000835, -0.00002097, 0.00000908, 0.00000158, 0.00000159, -0.00001998, 0.00001907, -0.00002475, 0.00001592, 0.00000563, -0.00000210, 0.00002095, -0.00001488, -0.00000492, 0.00000954, -0.00002504, 0.00001669, -0.00001127, 0.00001334, -0.00001980, 0.00001177, -0.00003146, 0.00004341, -0.00004190, 0.00004393, -0.00000655, 0.00000205, -0.00004540, 0.00001794, 0.00001943, -0.00001404, 0.00000560, -0.00003137, 0.00003735, -0.00002621, -0.00001790, -0.00010926, -0.00001297, -0.00004303, 0.00014425, 0.00016811, -0.00001899, -0.00026047, 0.00005981, 0.00039759, },
		{ 0.00035687, 0.00007655, -0.00012039, -0.00002947, -0.00008439, -0.00003936, 0.00001591, -0.00000186, -0.00002036, 0.00001684, -0.00002356, 0.00005062, -0.00002685, -0.00002996, -0.00000277, 0.00003130, -0.00000387, -0.00000559, 0.00002062, -0.00002551, 0.00001071, -0.00002171, 0.00000980, -0.00000094, 0.00000564, -0.00002360, 0.00002064, -0.00002408, 0.00001152, 0.00000958, -0.00000417, 0.00002043, -0.00001004, -0.00001007, 0.00001407, -0.00002733, 0.00001639, -0.00001017, 0.00001298, -0.00002101, 0.00001325, -0.00003263, 0.00004146, -0.00003647, 0.00003409, 0.00000411, -0.00000175, -0.00004178, 0.00000782, 0.00002914, -0.00001994, 0.00001267, -0.00003571, 0.00003484, -0.00001919, -0.00002438, -0.00009971, -0.00002493, -0.00005432, 0.00013913, 0.00020300, -0.00001485, -0.00025248, 0.00003577, },
		{ 0.00001717, 0.00031635, 0.00010129, -0.00010391, -0.00003415, -0.00008359, -0.00003941, 0.00000867, 0.00000919, -0.00002812, 0.00001922, -0.00002571, 0.00004819, -0.00001942, -0.00003256, -0.00000460, 0.00002748, 0.00000427, -0.00001366, 0.00002710, -0.00002871, 0.00001250, -0.00002212, 0.00001065, -0.00000358, 0.00000939, -0.00002633, 0.00002095, -0.00002211, 0.00000642, 0.00001327, -0.00000549, 0.00001889, -0.00000467, -0.00001486, 0.00001760, -0.00002831, 0.00001498, -0.00000840, 0.00001235, -0.00002180, 0.00001384, -0.00003248, 0.00003788, -0.00002961, 0.00002378, 0.00001362, -0.00000417, -0.00003842, -0.00000194, 0.00003733, -0.00002393, 0.00001733, -0.00003640, 0.00002867, -0.00001014, -0.00003100, -0.00009160, -0.00003252, -0.00006910, 0.00013123, 0.00023920, -0.00000758, -0.00024790, },
		},
		{ { -0.00005222, -0.00003211, -0.00010349, -0.00003039, 0.00019038, 0.00019085, 0.00001706, -0.00002996, -0.00006793, -0.00002439, 0.00000774, -0.00004971, 0.00003743, -0.00005525, -0.00002261, 0.00000139, 0.00003227, 0.00002173, -0.00003843, 0.00000049, -0.00003378, 0.00004037, -0.00002234, -0.00000153, 0.00001039, -0.00001198, 0.00002473, 0.00001262, -0.00001410, -0.00002604, 0.00001909, 0.00000153, -0.00000532, 0.00000153, 0.00001909, -0.00002604, -0.00001410, 0.00001262, 0.00002473, -0.00001198, 0.00001039, -0.00000153, -0.00002234, 0.00004037, -0.00003378, 0.00000049, -0.00003843, 0.00002173, 0.00003227, 0.00000139, -0.00002261, -0.00005525, 0.00003743, -0.00004971, 0.00000774, -0.00002439, -0.00006793, -0.00002996, 0.00001706, 0.00019085, 0.00019038, -0.00003039, -0.00010349, -0.00003211, },
		{ -0.00003250, -0.00005205, -0.00003219, -0.00009226, -0.00004008, 0.00014809, 0.00018957, 0.00003755, -0.00001612, -0.00006089, -0.00004671, 0.00002677, -0.00005854, 0.00003629, -0.00004317, -0.00003353, 0.00000748, 0.00001752, 0.00003694, -0.00004014, 0.00000257, -0.00003827, 0.00003632, -0.00001289, -0.00001220, 0.00001855, -0.00001843, 0.00002436, 0.00001463, -0.00000742, -0.00002990, 0.00001168, 0.00000678, -0.00000424, -0.00000449, 0.00002366, -0.00001866, -0.00002121, 0.00001074, 0.00002333, -0.00000450, 0.00000281, 0.00000704, -0.00002819, 0.00004013, -0.00002614, -0.00000376, -0.00003292, 0.00000448, 0.00004394, -0.00000019, -0.00001572, -0.00006151, 0.00003184, -0.00003639, -0.00001062, -0.00000916, -0.00006583, -0.00004364, -0.00000699, 0.00018368, 0.00023575, -0.00001298, -0.00011755, },
		{ -0.00013401, -0.00003390, -0.00005164, -0.00003217, -0.00008402, -0.00004395, 0.00011042, 0.00018094, 0.00005552, -0.00000601, -0.00004360, -0.00007308, 0.00004192, -0.00005925, 0.00002679, -0.00002527, -0.00004760, 0.00001616, 0.00000278, 0.00004745, -0.00003703, 0.00000209, -0.00003921, 0.00002807, -0.00000061, -0.00002375, 0.00002591, -0.00002270, 0.00002165, 0.00001721, -0.00000195, -0.00003011, 0.00000253, 0.00001026, -0.00000120, -0.00001017, 0.00002467, -0.00000843, -0.00002773, 0.00000848, 0.00002087, 0.00000287, -0.00000309, 0.00001277, -0.00003024, 0.00003602, -0.00001621, -0.00000930, -0.00002521, -0.00001201, 0.00005023, 0.00000331, -0.00001238, -0.00006320, 0.00002204, -0.00002215, -0.00002508, -0.00000175, -0.00005764, -0.00005339, -0.00003403, 0.00016625, 0.00028241, 0.00001395, },
		{ 0.00005238, -0.00015225, -0.00003677, -0.00005117, -0.00003171, -0.00007838, -0.00004419, 0.00007903, 0.00016524, 0.00007276, -0.00000240, -0.00001755, -0.00009840, 0.00004825, -0.00004924, 0.00000888, -0.00000313, -0.00006262, 0.00002460, -0.00000874, 0.00005132, -0.00002899, -0.00000071, -0.00003668, 0.00001628, 0.00001324, -0.00003466, 0.00003103, -0.00002381, 0.00001630, 0.00002052, 0.00000196, -0.00002712, -0.00000699, 0.00001124, 0.00000327, -0.00001445, 0.00002195, 0.00000350, -0.00003252, 0.00000537, 0.00001797, 0.00000918, -0.00000661, 0.00001542, -0.00002881, 0.00002894, -0.00000522, -0.00001489, -0.00001715, -0.00002532, 0.00005002, 0.00001107, -0.00001102, -0.00006220, 0.00001063, -0.00000974, -0.00003425, -0.00000082, -0.00004724, -0.00005674, -0.00006161, 0.00013652, 0.00032777, },
		{ 0.00036791, 0.00010451, -0.00017141, -0.00004171, -0.00005063, -0.00003098, -0.00007426, -0.00004338, 0.00005562, 0.00014244, 0.00009079, -0.00000593, 0.00001310, -0.00011644, 0.00004165, -0.00002784, -0.00001546, 0.00002017, -0.00007545, 0.00002967, -0.00001443, 0.00004780, -0.00001687, -0.00000508, -0.00003124, 0.00000207, 0.00002704, -0.00004329, 0.00003265, -0.00002119, 0.00000842, 0.00002438, 0.00000438, -0.00002183, -0.00001552, 0.00000938, 0.00000836, -0.00001654, 0.00001588, 0.00001566, -0.00003459, 0.00000117, 0.00001511, 0.00001379, -0.00000750, 0.00001524, -0.00002462, 0.00002011, 0.00000545, -0.00001927, -0.00001039, -0.00003391, 0.00004360, 0.00002114, -0.00000963, -0.00006035, -0.00000042, -0.00000057, -0.00003842, -0.00000383, -0.00003813, -0.00005330, -0.00008573, 0.00009316, },
		{ 0.00003664, 0.00039731, 0.00017231, -0.00018988, -0.00004974, -0.00004946, -0.00003090, -0.00006991, -0.00004422, 0.00004154, 0.00011291, 0.00010985, -0.00001467, 0.00004222, -0.00012128, 0.00002007, 0.00000311, -0.00004233, 0.00004057, -0.00008276, 0.00002877, -0.00001285, 0.00003748, -0.00000239, -0.00000989, -0.00002375, -0.00001311, 0.00003904, -0.00004812, 0.00002994, -0.00001479, -0.00000140, 0.00002825, 0.00000577, -0.00001536, -0.00002187, 0.00000480, 0.00001312, -0.00001596, 0.00000731, 0.00002651, -0.00003328, -0.00000407, 0.00001255, 0.00001640, -0.00000596, 0.00001289, -0.00001871, 0.00001087, 0.00001448, -0.00002136, -0.00000608, -0.00003732, 0.00003250, 0.00003093, -0.00000635, -0.00005882, -0.00001014, 0.00000528, -0.00003913, -0.00000799, -0.00003241, -0.00004486, -0.00010178, },
		{ -0.00010580, -0.00002994, 0.00040915, 0.00025685, -0.00020479, -0.00006271, -0.00004643, -0.00003303, -0.00006323, -0.00004903, 0.00003717, 0.00007817, 0.00012832, -0.00002444, 0.00006300, -0.00010874, -0.00001554, 0.00003916, -0.00006644, 0.00005379, -0.00008181, 0.00002043, -0.00000407, 0.00002228, 0.00001216, -0.00001390, -0.00001523, -0.00002763, 0.00004757, -0.00004806, 0.00002264, -0.00000522, -0.00001213, 0.00003135, 0.00000680, -0.00000891, -0.00002527, -0.00000196, 0.00001659, -0.00001265, -0.00000255, 0.00003466, -0.00002838, -0.00000997, 0.00001031, 0.00001704, -0.00000250, 0.00000923, -0.00001221, 0.00000249, 0.00002080, -0.00002056, -0.00000457, -0.00003617, 0.00001907, 0.00003797, -0.00000010, -0.00005784, -0.00001864, 0.00000873, -0.00003829, -0.00001122, -0.00003037, -0.00003491, },
		{ -0.00002757, -0.00009583, -0.00010057, 0.00039624, 0.00035711, -0.00021147, -0.00008333, -0.00003978, -0.00003915, -0.00005249, -0.00005897, 0.00004137, 0.00004128, 0.00014283, -0.00002976, 0.00006959, -0.00007772, -0.00006094, 0.00007385, -0.00008214, 0.00005638, -0.00007127, 0.00000481, 0.00001025, 0.00000510, 0.00002437, -0.00001605, -0.00000670, -0.00003994, 0.00005130, -0.00004263, 0.00001125, 0.00000637, -0.00002242, 0.00003275, 0.00000822, -0.00000346, -0.00002540, -0.00000998, 0.00001798, -0.00000700, -0.00001234, 0.00003912, -0.00002023, -0.00001589, 0.00000828, 0.00001595, 0.00000211, 0.00000525, -0.00000617, -0.00000399, 0.00002381, -0.00001686, -0.00000549, -0.00003191, 0.00000593, 0.00004042, 0.00000908, -0.00005662, -0.00002671, 0.00001103, -0.00003747, -0.00001277, -0.00003049, },
		{ -0.00003021, -0.00002632, -0.00007278, -0.00016675, 0.00035244, 0.00046899, -0.00020327, -0.00011480, -0.00002781, -0.00005055, -0.00003704, -0.00007354, 0.00005114, 0.00000677, 0.00014888, -0.00002534, 0.00005866, -0.00003093, -0.00010898, 0.00009986, -0.00008463, 0.00004662, -0.00005165, -0.00001625, 0.00002719, -0.00001077, 0.00003216, -0.00001568, 0.00000099, -0.00004872, 0.00004947, -0.00003208, -0.00000306, 0.00001833, -0.00003082, 0.00003163, 0.00001057, 0.00000036, -0.00002252, -0.00001812, 0.00001676, 0.00000029, -0.00002075, 0.00003939, -0.00000964, -0.00002103, 0.00000625, 0.00001357, 0.00000704, 0.00000183, -0.00000146, -0.00000794, 0.00002347, -0.00001087, -0.00000780, -0.00002641, -0.00000466, 0.00003762, 0.00001999, -0.00005377, -0.00003534, 0.00001317, -0.00003731, -0.00001328, },
		{ -0.00001447, -0.00002682, -0.00003282, -0.00004075, -0.00021874, 0.00027444, 0.00058447, -0.00017199, -0.00016003, -0.00000974, -0.00006719, -0.00001799, -0.00009026, 0.00006187, -0.00002003, 0.00014211, -0.00000761, 0.00003051, 0.00002519, -0.00015080, 0.00011044, -0.00007118, 0.00002515, -0.00002545, -0.00003949, 0.00004317, -0.00002222, 0.00003420, -0.00001267, 0.00000724, -0.00005306, 0.00004198, -0.00001744, -0.00001854, 0.00002886, -0.00003600, 0.00002747, 0.00001411, 0.00000236, -0.00001729, -0.00002523, 0.00001275, 0.00000823, -0.00002672, 0.00003554, 0.00000221, -0.00002462, 0.00000408, 0.00001037, 0.00001148, -0.00000033, 0.00000143, -0.00000917, 0.00002026, -0.00000369, -0.00001016, -0.00002148, -0.00001126, 0.00003016, 0.00003057, -0.00004777, -0.00004514, 0.00001535, -0.00003732, },
		{ -0.00003617, -0.00001832, -0.00001859, -0.00004624, -0.00000649, -0.00024734, 0.00016335, 0.00069161, -0.00010900, -0.00022030, 0.00001325, -0.00008705, 0.00000158, -0.00010491, 0.00006807, -0.00003410, 0.00011965, 0.00002397, -0.00001054, 0.00008142, -0.00017744, 0.00010101, -0.00004196, -0.00000489, 0.00000324, -0.00006091, 0.00005466, -0.00002696, 0.00003017, -0.00000749, 0.00001174, -0.00005256, 0.00002950, -0.00000039, -0.00003314, 0.00003623, -0.00003705, 0.00002019, 0.00001869, 0.00000279, -0.00001072, -0.00003028, 0.00000618, 0.00001575, -0.00002953, 0.00002819, 0.00001399, -0.00002606, 0.00000171, 0.00000680, 0.00001478, -0.00000081, 0.00000232, -0.00000795, 0.00001513, 0.00000331, -0.00001122, -0.00001843, -0.00001352, 0.00001976, 0.00003847, -0.00003762, -0.00005584, 0.00001685, },
		{ 0.00001623, -0.00003224, -0.00002626, -0.00000554, -0.00006318, 0.00002189, -0.00024598, 0.00002573, 0.00077539, -0.00000698, -0.00029385, 0.00003731, -0.00010587, 0.00001687, -0.00011238, 0.00006456, -0.00003209, 0.00008137, 0.00006629, -0.00005674, 0.00012742, -0.00018177, 0.00007040, -0.00000039, -0.00003829, 0.00002943, -0.00007657, 0.00005890, -0.00002396, 0.00002086, -0.00000110, 0.00001447, -0.00004739, 0.00001338, 0.00001694, -0.00004477, 0.00003911, -0.00003361, 0.00001025, 0.00002373, 0.00000227, -0.00000393, -0.00003255, -0.00000235, 0.00002180, -0.00002895, 0.00001835, 0.00002438, -0.00002505, -0.00000077, 0.00000329, 0.00001647, 0.00000049, 0.00000144, -0.00000493, 0.00000930, 0.00000879, -0.00001003, -0.00001777, -0.00001220, 0.00000879, 0.00004167, -0.00002333, -0.00006621, },
		{ -0.00007422, 0.00001183, -0.00002435, -0.00003842, 0.00001030, -0.00007840, 0.00003674, -0.00021246, -0.00012626, 0.00081960, 0.00013780, -0.00037467, 0.00005561, -0.00011758, 0.00002228, -0.00010792, 0.00004785, -0.00001346, 0.00003055, 0.00011282, -0.00009830, 0.00015376, -0.00016013, 0.00002142, 0.00004722, -0.00006859, 0.00004828, -0.00008344, 0.00005450, -0.00001364, 0.00000799, 0.00000526, 0.00001560, -0.00003819, -0.00000452, 0.00003229, -0.00005173, 0.00003682, -0.00002604, -0.00000139, 0.00002840, 0.00000157, 0.00000205, -0.00003172, -0.00001197, 0.00002552, -0.00002519, 0.00000729, 0.00003232, -0.00002162, -0.00000316, 0.00000017, 0.00001632, 0.00000336, -0.00000071, -0.00000102, 0.00000405, 0.00001169, -0.00000624, -0.00001919, -0.00000886, -0.00000029, 0.00003907, -0.00000612, },
		{ 0.00001173, -0.00007757, 0.00000248, -0.00001243, -0.00005331, 0.00002517, -0.00008608, 0.00003269, -0.00015013, -0.00027602, 0.00080943, 0.00032346, -0.00045188, 0.00005892, -0.00011547, 0.00001290, -0.00008864, 0.00001735, 0.00001884, -0.00002624, 0.00015471, -0.00012537, 0.00015402, -0.00011346, -0.00003925, 0.00009274, -0.00008944, 0.00005612, -0.00008006, 0.00004172, 0.00000224, -0.00000609, 0.00001029, 0.00001546, -0.00002609, -0.00002207, 0.00004361, -0.00005290, 0.00002946, -0.00001535, -0.00001341, 0.00003171, 0.00000147, 0.00000643, -0.00002793, -0.00002160, 0.00002634, -0.00001885, -0.00000367, 0.00003711, -0.00001615, -0.00000520, -0.00000231, 0.00001435, 0.00000733, -0.00000343, 0.00000279, 0.00000046, 0.00001149, -0.00000019, -0.00002161, -0.00000544, -0.00000557, 0.00003088, },
		{ 0.00001859, 0.00002728, -0.00007437, -0.00001193, 0.00000218, -0.00006792, 0.00003454, -0.00008138, 0.00000822, -0.00006788, -0.00040460, 0.00073450, 0.00054086, -0.00051016, 0.00003707, -0.00009384, -0.00001396, -0.00005471, -0.00002399, 0.00005851, -0.00008011, 0.00018261, -0.00013022, 0.00012627, -0.00004747, -0.00010204, 0.00012765, -0.00009581, 0.00005127, -0.00006691, 0.00002239, 0.00002094, -0.00001884, 0.00001297, 0.00001439, -0.00001245, -0.00003716, 0.00004938, -0.00004803, 0.00001794, -0.00000305, -0.00002431, 0.00003276, 0.00000254, 0.00000882, -0.00002172, -0.00003016, 0.00002403, -0.00001083, -0.00001337, 0.00003847, -0.00000929, -0.00000664, -0.00000393, 0.00001078, 0.00001175, -0.00000594, 0.00000563, -0.00000075, 0.00000827, 0.00000710, -0.00002357, -0.00000370, -0.00000615, },
		{ -0.00000234, 0.00000475, 0.00003769, -0.00006382, -0.00003005, 0.00001676, -0.00007844, 0.00003426, -0.00006192, -0.00003361, 0.00002097, -0.00049377, 0.00059166, 0.00077315, -0.00053135, -0.00001882, -0.00005001, -0.00005732, -0.00001014, -0.00006975, 0.00009683, -0.00012154, 0.00018870, -0.00010909, 0.00007386, 0.00002808, -0.00015623, 0.00014475, -0.00008520, 0.00003449, -0.00004634, -0.00000044, 0.00003926, -0.00002800, 0.00001266, 0.00001272, 0.00000124, -0.00004798, 0.00004884, -0.00003774, 0.00000386, 0.00000906, -0.00003271, 0.00003093, 0.00000504, 0.00000924, -0.00001396, -0.00003667, 0.00001875, -0.00000218, -0.00002088, 0.00003658, -0.00000182, -0.00000732, -0.00000456, 0.00000603, 0.00001589, -0.00000755, 0.00000683, 0.00000066, 0.00000273, 0.00001420, -0.00002360, -0.00000479, },
		{ -0.00000893, 0.00000443, -0.00000771, 0.00004093, -0.00004664, -0.00004913, 0.00002782, -0.00008124, 0.00002190, -0.00002873, -0.00008535, 0.00010081, -0.00052933, 0.00038687, 0.00099675, -0.00049718, -0.00011377, 0.00001402, -0.00011179, 0.00003745, -0.00011109, 0.00012438, -0.00014238, 0.00016866, -0.00006328, 0.00000502, 0.00010105, -0.00019187, 0.00013966, -0.00005823, 0.00000894, -0.00002208, -0.00002317, 0.00005416, -0.00003198, 0.00000931, 0.00001068, 0.00001356, -0.00005330, 0.00004215, -0.00002349, -0.00001080, 0.00001920, -0.00003755, 0.00002603, 0.00000881, 0.00000808, -0.00000566, -0.00004041, 0.00001101, 0.00000603, -0.00002568, 0.00003197, 0.00000546, -0.00000721, -0.00000414, 0.00000066, 0.00001902, -0.00000771, 0.00000610, 0.00000442, -0.00000394, 0.00001959, -0.00002065, },
		{ -0.00001435, -0.00001538, 0.00001197, -0.00001610, 0.00003637, -0.00002520, -0.00006546, 0.00003192, -0.00007398, -0.00000236, 0.00001358, -0.00013651, 0.00015667, -0.00050396, 0.00013579, 0.00118380, -0.00039269, -0.00024641, 0.00009050, -0.00016775, 0.00007789, -0.00013850, 0.00013304, -0.00013782, 0.00012296, 0.00000077, -0.00006846, 0.00015908, -0.00020180, 0.00011187, -0.00001869, -0.00002042, 0.00000145, -0.00004233, 0.00006333, -0.00003010, 0.00000339, 0.00000845, 0.00002335, -0.00005261, 0.00003032, -0.00000730, -0.00002393, 0.00002594, -0.00003832, 0.00001837, 0.00001333, 0.00000602, 0.00000214, -0.00004100, 0.00000158, 0.00001286, -0.00002760, 0.00002540, 0.00001189, -0.00000640, -0.00000271, -0.00000467, 0.00002054, -0.00000617, 0.00000351, 0.00000979, -0.00001034, 0.00002198, },
		{ 0.00002059, -0.00000516, -0.00002262, 0.00001793, -0.00001867, 0.00002504, -0.00000312, -0.00007525, 0.00002671, -0.00005644, -0.00003536, 0.00005749, -0.00017560, 0.00017709, -0.00041927, -0.00013740, 0.00130593, -0.00020995, -0.00040734, 0.00016613, -0.00021268, 0.00010046, -0.00014397, 0.00011805, -0.00010770, 0.00005716, 0.00007249, -0.00013345, 0.00019199, -0.00018314, 0.00006498, 0.00002704, -0.00004777, 0.00002015, -0.00005525, 0.00006553, -0.00002269, -0.00000415, 0.00000611, 0.00002979, -0.00004622, 0.00001513, 0.00000853, -0.00003372, 0.00002841, -0.00003511, 0.00000871, 0.00001782, 0.00000382, 0.00000858, -0.00003844, -0.00000856, 0.00001758, -0.00002681, 0.00001772, 0.00001699, -0.00000512, -0.00000042, -0.00000927, 0.00002005, -0.00000300, -0.00000050, 0.00001578, -0.00001510, },
		{ -0.00001713, 0.00001536, 0.00000572, -0.00002876, 0.00002040, -0.00001509, 0.00000958, 0.00001541, -0.00007555, 0.00001177, -0.00003099, -0.00007138, 0.00009408, -0.00019256, 0.00015668, -0.00028618, -0.00040287, 0.00133880, 0.00004891, -0.00057870, 0.00022354, -0.00023343, 0.00009630, -0.00012311, 0.00007939, -0.00005702, -0.00001888, 0.00013915, -0.00017796, 0.00019377, -0.00013804, 0.00000621, 0.00007119, -0.00006762, 0.00003097, -0.00006060, 0.00006077, -0.00001094, -0.00001214, 0.00000372, 0.00003253, -0.00003516, -0.00000121, 0.00002184, -0.00003887, 0.00002645, -0.00002862, -0.00000184, 0.00002139, 0.00000221, 0.00001310, -0.00003310, -0.00001838, 0.00001972, -0.00002378, 0.00000976, 0.00002053, -0.00000363, 0.00000242, -0.00001251, 0.00001742, 0.00000145, -0.00000525, 0.00002128, },
		{ 0.00002528, -0.00001585, 0.00000693, 0.00001658, -0.00003202, 0.00001839, -0.00000651, -0.00000637, 0.00002660, -0.00006517, -0.00001096, -0.00000229, -0.00010320, 0.00011494, -0.00018113, 0.00009746, -0.00012369, -0.00062925, 0.00126657, 0.00036989, -0.00073547, 0.00024402, -0.00021916, 0.00006082, -0.00007673, 0.00002232, 0.00000481, -0.00009273, 0.00018834, -0.00019351, 0.00016382, -0.00007336, -0.00005488, 0.00010609, -0.00007590, 0.00003257, -0.00005847, 0.00005018, 0.00000338, -0.00001935, 0.00000131, 0.00003167, -0.00002105, -0.00001637, 0.00003090, -0.00003889, 0.00002061, -0.00002002, -0.00001201, 0.00002324, 0.00000170, 0.00001549, -0.00002568, -0.00002690, 0.00001914, -0.00001914, 0.00000223, 0.00002248, -0.00000216, 0.00000541, -0.00001388, 0.00001284, 0.00000655, -0.00000990, },
		{ -0.00001366, 0.00002706, -0.00001124, -0.00000346, 0.00002560, -0.00003120, 0.00001212, 0.00000465, -0.00001894, 0.00002805, -0.00004524, -0.00003730, 0.00002355, -0.00012372, 0.00011417, -0.00014043, 0.00000884, 0.00004391, -0.00078851, 0.00108556, 0.00072719, -0.00084844, 0.00021118, -0.00016423, -0.00000453, -0.00001136, -0.00004333, 0.00006562, -0.00015182, 0.00021044, -0.00017686, 0.00010719, 0.00000042, -0.00010815, 0.00012562, -0.00007079, 0.00002554, -0.00005033, 0.00003570, 0.00001827, -0.00002473, -0.00000107, 0.00002773, -0.00000582, -0.00002830, 0.00003472, -0.00003407, 0.00001207, -0.00001071, -0.00002061, 0.00002279, 0.00000248, 0.00001587, -0.00001706, -0.00003331, 0.00001598, -0.00001360, -0.00000436, 0.00002297, -0.00000083, 0.00000805, -0.00001310, 0.00000680, 0.00001153, },
		{ 0.00001556, -0.00001589, 0.00002629, -0.00000383, -0.00001426, 0.00003122, -0.00002598, 0.00000294, 0.00001530, -0.00002497, 0.00001942, -0.00001924, -0.00006152, 0.00004033, -0.00012764, 0.00008996, -0.00007553, -0.00009385, 0.00019102, -0.00086060, 0.00080618, 0.00108541, -0.00088855, 0.00011482, -0.00007051, -0.00009155, 0.00006145, -0.00010486, 0.00011290, -0.00018597, 0.00020067, -0.00013073, 0.00003368, 0.00007118, -0.00014478, 0.00012645, -0.00005311, 0.00001218, -0.00003856, 0.00001962, 0.00003184, -0.00002759, -0.00000336, 0.00002156, 0.00000853, -0.00003548, 0.00003318, -0.00002546, 0.00000233, -0.00000214, -0.00002668, 0.00001983, 0.00000443, 0.00001470, -0.00000822, -0.00003703, 0.00001066, -0.00000789, -0.00000967, 0.00002220, 0.00000033, 0.00000985, -0.00001015, 0.00000005, },
		{ -0.00000647, 0.00001793, -0.00001623, 0.00002304, 0.00000544, -0.00002385, 0.00003249, -0.00001704, -0.00000693, 0.00002247, -0.00002282, 0.00000270, 0.00000755, -0.00007760, 0.00004327, -0.00011289, 0.00004526, 0.00000329, -0.00019212, 0.00029509, -0.00083699, 0.00045270, 0.00140370, -0.00083200, -0.00004575, 0.00005156, -0.00018607, 0.00012704, -0.00014925, 0.00013641, -0.00018953, 0.00016010, -0.00006328, -0.00004397, 0.00012737, -0.00015895, 0.00010866, -0.00002611, -0.00000400, -0.00002587, 0.00000412, 0.00004264, -0.00002762, -0.00000548, 0.00001423, 0.00002026, -0.00003718, 0.00002699, -0.00001463, -0.00000695, 0.00000451, -0.00002970, 0.00001457, 0.00000710, 0.00001262, -0.00000006, -0.00003781, 0.00000379, -0.00000266, -0.00001350, 0.00002040, 0.00000145, 0.00001041, -0.00000536, },
		{ 0.00000064, -0.00001180, 0.00001818, -0.00001465, 0.00001777, 0.00001537, -0.00003088, 0.00002925, -0.00000589, -0.00001502, 0.00002402, -0.00001286, -0.00001801, 0.00002906, -0.00008060, 0.00003034, -0.00008136, -0.00001248, 0.00008236, -0.00026779, 0.00034053, -0.00072253, 0.00006076, 0.00164137, -0.00066512, -0.00025968, 0.00018362, -0.00027008, 0.00017044, -0.00016598, 0.00013046, -0.00016267, 0.00009557, 0.00001342, -0.00011248, 0.00016015, -0.00014891, 0.00007572, 0.00000521, -0.00001914, -0.00001473, -0.00000915, 0.00004981, -0.00002493, -0.00000740, 0.00000683, 0.00002810, -0.00003353, 0.00001758, -0.00000340, -0.00001434, 0.00000847, -0.00002961, 0.00000757, 0.00000983, 0.00001032, 0.00000669, -0.00003571, -0.00000385, 0.00000153, -0.00001582, 0.00001773, 0.00000272, 0.00000953, },
		{ 0.00000726, 0.00000698, -0.00001506, 0.00001613, -0.00001147, 0.00001119, 0.00002479, -0.00003449, 0.00002210, 0.00000543, -0.00001923, 0.00001928, 0.00000249, -0.00003735, 0.00003984, -0.00006804, 0.00000304, -0.00003886, -0.00007269, 0.00014733, -0.00030647, 0.00032156, -0.00053500, -0.00032732, 0.00176406, -0.00038818, -0.00050456, 0.00030175, -0.00032501, 0.00017956, -0.00014945, 0.00009527, -0.00011140, 0.00001830, 0.00008572, -0.00016055, 0.00016503, -0.00011725, 0.00003376, 0.00003511, -0.00002983, -0.00000680, -0.00001925, 0.00005314, -0.00001994, -0.00000910, 0.00000037, 0.00003140, -0.00002544, 0.00000676, 0.00000652, -0.00001884, 0.00000949, -0.00002682, -0.00000032, 0.00001191, 0.00000843, 0.00001158, -0.00003111, -0.00001146, 0.00000427, -0.00001666, 0.00001433, 0.00000430, },
		{ 0.00000626, 0.00000398, 0.00001263, -0.00001562, 0.00001201, -0.00000727, 0.00000414, 0.00003269, -0.00003437, 0.00001228, 0.00001489, -0.00001835, 0.00000925, 0.00001926, -0.00004983, 0.00003634, -0.00004070, -0.00003357, 0.00000597, -0.00012373, 0.00018604, -0.00030044, 0.00024322, -0.00030253, -0.00066843, 0.00174943, -0.00001720, -0.00074845, 0.00038028, -0.00033549, 0.00014799, -0.00010059, 0.00003717, -0.00004645, -0.00005822, 0.00014105, -0.00018089, 0.00014256, -0.00007035, -0.00000964, 0.00005834, -0.00003386, -0.00000268, -0.00002605, 0.00005293, -0.00001327, -0.00001061, -0.00000441, 0.00003023, -0.00001445, -0.00000352, 0.00001378, -0.00001998, 0.00000783, -0.00002209, -0.00000812, 0.00001266, 0.00000737, 0.00001446, -0.00002464, -0.00001821, 0.00000530, -0.00001614, 0.00001032, },
		{ 0.00000586, 0.00000850, 0.00000028, 0.00001662, -0.00001326, 0.00000641, -0.00000279, -0.00000259, 0.00003837, -0.00003078, 0.00000134, 0.00002089, -0.00001242, -0.00000352, 0.00003279, -0.00005117, 0.00001798, -0.00000286, -0.00007169, 0.00004336, -0.00015538, 0.00019109, -0.00025015, 0.00012044, -0.00005920, -0.00092500, 0.00159129, 0.00041669, -0.00095406, 0.00039650, -0.00029291, 0.00007697, -0.00002708, -0.00003249, 0.00001887, -0.00012093, 0.00017031, -0.00017150, 0.00009816, -0.00001705, -0.00004698, 0.00007113, -0.00003063, -0.00000190, -0.00003001, 0.00004980, -0.00000557, -0.00001200, -0.00000706, 0.00002527, -0.00000245, -0.00001160, 0.00001761, -0.00001793, 0.00000420, -0.00001638, -0.00001488, 0.00001164, 0.00000730, 0.00001551, -0.00001709, -0.00002339, 0.00000452, -0.00001446, },
		{ -0.00001188, 0.00000118, 0.00001071, -0.00000305, 0.00001817, -0.00000817, 0.00000023, 0.00000120, -0.00000835, 0.00004151, -0.00002438, -0.00000918, 0.00002263, -0.00000281, -0.00001557, 0.00003882, -0.00003939, -0.00001245, 0.00003830, -0.00010231, 0.00006459, -0.00016108, 0.00016145, -0.00016422, -0.00002467, 0.00016042, -0.00107069, 0.00130143, 0.00087052, -0.00108425, 0.00033535, -0.00019794, -0.00002423, 0.00005779, -0.00009962, 0.00007138, -0.00015978, 0.00016950, -0.00013591, 0.00004086, 0.00003308, -0.00007230, 0.00007192, -0.00002123, -0.00000314, -0.00003195, 0.00004452, 0.00000257, -0.00001333, -0.00000753, 0.00001769, 0.00000868, -0.00001629, 0.00001791, -0.00001334, -0.00000046, -0.00001069, -0.00001983, 0.00000866, 0.00000809, 0.00001514, -0.00000933, -0.00002647, 0.00000202, },
		{ -0.00000191, -0.00000878, -0.00000337, 0.00001244, -0.00000521, 0.00001685, -0.00000105, -0.00000549, 0.00000403, -0.00001273, 0.00004211, -0.00001605, -0.00001811, 0.00002029, 0.00000817, -0.00002328, 0.00003461, -0.00001545, -0.00004908, 0.00007409, -0.00011718, 0.00006395, -0.00013941, 0.00010277, -0.00005779, -0.00016686, 0.00032680, -0.00109421, 0.00090866, 0.00129491, -0.00110823, 0.00019316, -0.00006145, -0.00013950, 0.00013752, -0.00015019, 0.00010069, -0.00016957, 0.00014034, -0.00008223, -0.00001849, 0.00007163, -0.00008222, 0.00006161, -0.00000808, -0.00000469, -0.00003270, 0.00003776, 0.00001067, -0.00001463, -0.00000611, 0.00000897, 0.00001732, -0.00001709, 0.00001520, -0.00000722, -0.00000513, -0.00000588, -0.00002250, 0.00000388, 0.00000937, 0.00001393, -0.00000218, -0.00002713, },
		{ -0.00002534, -0.00000682, -0.00000562, -0.00000734, 0.00001314, -0.00000559, 0.00001273, 0.00000701, -0.00000969, 0.00000527, -0.00001561, 0.00004042, -0.00000669, -0.00002477, 0.00001492, 0.00001770, -0.00002379, 0.00001965, 0.00001667, -0.00008404, 0.00009603, -0.00011068, 0.00004026, -0.00009464, 0.00002634, 0.00005040, -0.00028230, 0.00041999, -0.00100093, 0.00045515, 0.00164089, -0.00100708, -0.00002032, 0.00009653, -0.00024855, 0.00019563, -0.00017317, 0.00010135, -0.00015090, 0.00008973, -0.00002144, -0.00006937, 0.00009287, -0.00007654, 0.00004319, 0.00000565, -0.00000494, -0.00003285, 0.00003001, 0.00001840, -0.00001586, -0.00000339, 0.00000058, 0.00002236, -0.00001420, 0.00001048, -0.00000071, -0.00000892, -0.00000257, -0.00002273, -0.00000225, 0.00001057, 0.00001249, 0.00000371, },
		{ 0.00000793, -0.00002135, -0.00001207, -0.00000293, -0.00001024, 0.00001230, -0.00000386, 0.00000640, 0.00001467, -0.00001148, 0.00000481, -0.00001707, 0.00003682, 0.00000299, -0.00002910, 0.00000815, 0.00002323, -0.00001580, -0.00000395, 0.00005063, -0.00010896, 0.00009780, -0.00008134, -0.00000264, -0.00003590, -0.00005311, 0.00014162, -0.00035277, 0.00043254, -0.00081165, -0.00000943, 0.00186704, -0.00077779, -0.00028220, 0.00024976, -0.00033067, 0.00021902, -0.00016292, 0.00007387, -0.00010982, 0.00002817, 0.00003487, -0.00010362, 0.00009473, -0.00005805, 0.00002107, 0.00001682, -0.00000285, -0.00003254, 0.00002153, 0.00002550, -0.00001686, -0.00000014, -0.00000616, 0.00002335, -0.00000846, 0.00000495, 0.00000517, -0.00001118, -0.00000103, -0.00002073, -0.00000904, 0.00001108, 0.00001138, },
		{ 0.00001096, 0.00001034, -0.00001568, -0.00001695, -0.00000125, -0.00001161, 0.00000964, -0.00000015, -0.00000112, 0.00002052, -0.00001033, 0.00000283, -0.00001737, 0.00003173, 0.00001248, -0.00003148, 0.00000175, 0.00002305, 0.00000003, -0.00003181, 0.00007898, -0.00011683, 0.00007676, -0.00003247, -0.00005642, 0.00002454, -0.00012039, 0.00020074, -0.00036872, 0.00037010, -0.00055905, -0.00043492, 0.00194566, -0.00043492, -0.00055905, 0.00037010, -0.00036872, 0.00020074, -0.00012039, 0.00002454, -0.00005642, -0.00003247, 0.00007676, -0.00011683, 0.00007898, -0.00003181, 0.00000003, 0.00002305, 0.00000175, -0.00003148, 0.00001248, 0.00003173, -0.00001737, 0.00000283, -0.00001033, 0.00002052, -0.00000112, -0.00000015, 0.00000964, -0.00001161, -0.00000125, -0.00001695, -0.00001568, 0.00001034, },
		{ 0.00000793, 0.00001138, 0.00001108, -0.00000904, -0.00002073, -0.00000103, -0.00001118, 0.00000517, 0.00000495, -0.00000846, 0.00002335, -0.00000616, -0.00000014, -0.00001686, 0.00002550, 0.00002153, -0.00003254, -0.00000285, 0.00001682, 0.00002107, -0.00005805, 0.00009473, -0.00010362, 0.00003487, 0.00002817, -0.00010982, 0.00007387, -0.00016292, 0.00021902, -0.00033067, 0.00024976, -0.00028220, -0.00077779, 0.00186704, -0.00000943, -0.00081165, 0.00043254, -0.00035277, 0.00014162, -0.00005311, -0.00003590, -0.00000264, -0.00008134, 0.00009780, -0.00010896, 0.00005063, -0.00000395, -0.00001580, 0.00002323, 0.00000815, -0.00002910, 0.00000299, 0.00003682, -0.00001707, 0.00000481, -0.00001148, 0.00001467, 0.00000640, -0.00000386, 0.00001230, -0.00001024, -0.00000293, -0.00001207, -0.00002135, },
		{ -0.00002534, 0.00000371, 0.00001249, 0.00001057, -0.00000225, -0.00002273, -0.00000257, -0.00000892, -0.00000071, 0.00001048, -0.00001420, 0.00002236, 0.00000058, -0.00000339, -0.00001586, 0.00001840, 0.00003001, -0.00003285, -0.00000494, 0.00000565, 0.00004319, -0.00007654, 0.00009287, -0.00006937, -0.00002144, 0.00008973, -0.00015090, 0.00010135, -0.00017317, 0.00019563, -0.00024855, 0.00009653, -0.00002032, -0.00100708, 0.00164089, 0.00045515, -0.00100093, 0.00041999, -0.00028230, 0.00005040, 0.00002634, -0.00009464, 0.00004026, -0.00011068, 0.00009603, -0.00008404, 0.00001667, 0.00001965, -0.00002379, 0.00001770, 0.00001492, -0.00002477, -0.00000669, 0.00004042, -0.00001561, 0.00000527, -0.00000969, 0.00000701, 0.00001273, -0.00000559, 0.00001314, -0.00000734, -0.00000562, -0.00000682, },
		{ -0.00000191, -0.00002713, -0.00000218, 0.00001393, 0.00000937, 0.00000388, -0.00002250, -0.00000588, -0.00000513, -0.00000722, 0.00001520, -0.00001709, 0.00001732, 0.00000897, -0.00000611, -0.00001463, 0.00001067, 0.00003776, -0.00003270, -0.00000469, -0.00000808, 0.00006161, -0.00008222, 0.00007163, -0.00001849, -0.00008223, 0.00014034, -0.00016957, 0.00010069, -0.00015019, 0.00013752, -0.00013950, -0.00006145, 0.00019316, -0.00110823, 0.00129491, 0.00090866, -0.00109421, 0.00032680, -0.00016686, -0.00005779, 0.00010277, -0.00013941, 0.00006395, -0.00011718, 0.00007409, -0.00004908, -0.00001545, 0.00003461, -0.00002328, 0.00000817, 0.00002029, -0.00001811, -0.00001605, 0.00004211, -0.00001273, 0.00000403, -0.00000549, -0.00000105, 0.00001685, -0.00000521, 0.00001244, -0.00000337, -0.00000878, },
		{ -0.00001188, 0.00000202, -0.00002647, -0.00000933, 0.00001514, 0.00000809, 0.00000866, -0.00001983, -0.00001069, -0.00000046, -0.00001334, 0.00001791, -0.00001629, 0.00000868, 0.00001769, -0.00000753, -0.00001333, 0.00000257, 0.00004452, -0.00003195, -0.00000314, -0.00002123, 0.00007192, -0.00007230, 0.00003308, 0.00004086, -0.00013591, 0.00016950, -0.00015978, 0.00007138, -0.00009962, 0.00005779, -0.00002423, -0.00019794, 0.00033535, -0.00108425, 0.00087052, 0.00130143, -0.00107069, 0.00016042, -0.00002467, -0.00016422, 0.00016145, -0.00016108, 0.00006459, -0.00010231, 0.00003830, -0.00001245, -0.00003939, 0.00003882, -0.00001557, -0.00000281, 0.00002263, -0.00000918, -0.00002438, 0.00004151, -0.00000835, 0.00000120, 0.00000023, -0.00000817, 0.00001817, -0.00000305, 0.00001071, 0.00000118, },
		{ 0.00000586, -0.00001446, 0.00000452, -0.00002339, -0.00001709, 0.00001551, 0.00000730, 0.00001164, -0.00001488, -0.00001638, 0.00000420, -0.00001793, 0.00001761, -0.00001160, -0.00000245, 0.00002527, -0.00000706, -0.00001200, -0.00000557, 0.00004980, -0.00003001, -0.00000190, -0.00003063, 0.00007113, -0.00004698, -0.00001705, 0.00009816, -0.00017150, 0.00017031, -0.00012093, 0.00001887, -0.00003249, -0.00002708, 0.00007697, -0.00029291, 0.00039650, -0.00095406, 0.00041669, 0.00159129, -0.00092500, -0.00005920, 0.00012044, -0.00025015, 0.00019109, -0.00015538, 0.00004336, -0.00007169, -0.00000286, 0.00001798, -0.00005117, 0.00003279, -0.00000352, -0.00001242, 0.00002089, 0.00000134, -0.00003078, 0.00003837, -0.00000259, -0.00000279, 0.00000641, -0.00001326, 0.00001662, 0.00000028, 0.00000850, },
		{ 0.00000626, 0.00001032, -0.00001614, 0.00000530, -0.00001821, -0.00002464, 0.00001446, 0.00000737, 0.00001266, -0.00000812, -0.00002209, 0.00000783, -0.00001998, 0.00001378, -0.00000352, -0.00001445, 0.00003023, -0.00000441, -0.00001061, -0.00001327, 0.00005293, -0.00002605, -0.00000268, -0.00003386, 0.00005834, -0.00000964, -0.00007035, 0.00014256, -0.00018089, 0.00014105, -0.00005822, -0.00004645, 0.00003717, -0.00010059, 0.00014799, -0.00033549, 0.00038028, -0.00074845, -0.00001720, 0.00174943, -0.00066843, -0.00030253, 0.00024322, -0.00030044, 0.00018604, -0.00012373, 0.00000597, -0.00003357, -0.00004070, 0.00003634, -0.00004983, 0.00001926, 0.00000925, -0.00001835, 0.00001489, 0.00001228, -0.00003437, 0.00003269, 0.00000414, -0.00000727, 0.00001201, -0.00001562, 0.00001263, 0.00000398, },
		{ 0.00000726, 0.00000430, 0.00001433, -0.00001666, 0.00000427, -0.00001146, -0.00003111, 0.00001158, 0.00000843, 0.00001191, -0.00000032, -0.00002682, 0.00000949, -0.00001884, 0.00000652, 0.00000676, -0.00002544, 0.00003140, 0.00000037, -0.00000910, -0.00001994, 0.00005314, -0.00001925, -0.00000680, -0.00002983, 0.00003511, 0.00003376, -0.00011725, 0.00016503, -0.00016055, 0.00008572, 0.00001830, -0.00011140, 0.00009527, -0.00014945, 0.00017956, -0.00032501, 0.00030175, -0.00050456, -0.00038818, 0.00176406, -0.00032732, -0.00053500, 0.00032156, -0.00030647, 0.00014733, -0.00007269, -0.00003886, 0.00000304, -0.00006804, 0.00003984, -0.00003735, 0.00000249, 0.00001928, -0.00001923, 0.00000543, 0.00002210, -0.00003449, 0.00002479, 0.00001119, -0.00001147, 0.00001613, -0.00001506, 0.00000698, },
		{ 0.00000064, 0.00000953, 0.00000272, 0.00001773, -0.00001582, 0.00000153, -0.00000385, -0.00003571, 0.00000669, 0.00001032, 0.00000983, 0.00000757, -0.00002961, 0.00000847, -0.00001434, -0.00000340, 0.00001758, -0.00003353, 0.00002810, 0.00000683, -0.00000740, -0.00002493, 0.00004981, -0.00000915, -0.00001473, -0.00001914, 0.00000521, 0.00007572, -0.00014891, 0.00016015, -0.00011248, 0.00001342, 0.00009557, -0.00016267, 0.00013046, -0.00016598, 0.00017044, -0.00027008, 0.00018362, -0.00025968, -0.00066512, 0.00164137, 0.00006076, -0.00072253, 0.00034053, -0.00026779, 0.00008236, -0.00001248, -0.00008136, 0.00003034, -0.00008060, 0.00002906, -0.00001801, -0.00001286, 0.00002402, -0.00001502, -0.00000589, 0.00002925, -0.00003088, 0.00001537, 0.00001777, -0.00001465, 0.00001818, -0.00001180, },
		{ -0.00000647, -0.00000536, 0.00001041, 0.00000145, 0.00002040, -0.00001350, -0.00000266, 0.00000379, -0.00003781, -0.00000006, 0.00001262, 0.00000710, 0.00001457, -0.00002970, 0.00000451, -0.00000695, -0.00001463, 0.00002699, -0.00003718, 0.00002026, 0.00001423, -0.00000548, -0.00002762, 0.00004264, 0.00000412, -0.00002587, -0.00000400, -0.00002611, 0.00010866, -0.00015895, 0.00012737, -0.00004397, -0.00006328, 0.00016010, -0.00018953, 0.00013641, -0.00014925, 0.00012704, -0.00018607, 0.00005156, -0.00004575, -0.00083200, 0.00140370, 0.00045270, -0.00083699, 0.00029509, -0.00019212, 0.00000329, 0.00004526, -0.00011289, 0.00004327, -0.00007760, 0.00000755, 0.00000270, -0.00002282, 0.00002247, -0.00000693, -0.00001704, 0.00003249, -0.00002385, 0.00000544, 0.00002304, -0.00001623, 0.00001793, },
		{ 0.00001556, 0.00000005, -0.00001015, 0.00000985, 0.00000033, 0.00002220, -0.00000967, -0.00000789, 0.00001066, -0.00003703, -0.00000822, 0.00001470, 0.00000443, 0.00001983, -0.00002668, -0.00000214, 0.00000233, -0.00002546, 0.00003318, -0.00003548, 0.00000853, 0.00002156, -0.00000336, -0.00002759, 0.00003184, 0.00001962, -0.00003856, 0.00001218, -0.00005311, 0.00012645, -0.00014478, 0.00007118, 0.00003368, -0.00013073, 0.00020067, -0.00018597, 0.00011290, -0.00010486, 0.00006145, -0.00009155, -0.00007051, 0.00011482, -0.00088855, 0.00108541, 0.00080618, -0.00086060, 0.00019102, -0.00009385, -0.00007553, 0.00008996, -0.00012764, 0.00004033, -0.00006152, -0.00001924, 0.00001942, -0.00002497, 0.00001530, 0.00000294, -0.00002598, 0.00003122, -0.00001426, -0.00000383, 0.00002629, -0.00001589, },
		{ -0.00001366, 0.00001153, 0.00000680, -0.00001310, 0.00000805, -0.00000083, 0.00002297, -0.00000436, -0.00001360, 0.00001598, -0.00003331, -0.00001706, 0.00001587, 0.00000248, 0.00002279, -0.00002061, -0.00001071, 0.00001207, -0.00003407, 0.00003472, -0.00002830, -0.00000582, 0.00002773, -0.00000107, -0.00002473, 0.00001827, 0.00003570, -0.00005033, 0.00002554, -0.00007079, 0.00012562, -0.00010815, 0.00000042, 0.00010719, -0.00017686, 0.00021044, -0.00015182, 0.00006562, -0.00004333, -0.00001136, -0.00000453, -0.00016423, 0.00021118, -0.00084844, 0.00072719, 0.00108556, -0.00078851, 0.00004391, 0.00000884, -0.00014043, 0.00011417, -0.00012372, 0.00002355, -0.00003730, -0.00004524, 0.00002805, -0.00001894, 0.00000465, 0.00001212, -0.00003120, 0.00002560, -0.00000346, -0.00001124, 0.00002706, },
		{ 0.00002528, -0.00000990, 0.00000655, 0.00001284, -0.00001388, 0.00000541, -0.00000216, 0.00002248, 0.00000223, -0.00001914, 0.00001914, -0.00002690, -0.00002568, 0.00001549, 0.00000170, 0.00002324, -0.00001201, -0.00002002, 0.00002061, -0.00003889, 0.00003090, -0.00001637, -0.00002105, 0.00003167, 0.00000131, -0.00001935, 0.00000338, 0.00005018, -0.00005847, 0.00003257, -0.00007590, 0.00010609, -0.00005488, -0.00007336, 0.00016382, -0.00019351, 0.00018834, -0.00009273, 0.00000481, 0.00002232, -0.00007673, 0.00006082, -0.00021916, 0.00024402, -0.00073547, 0.00036989, 0.00126657, -0.00062925, -0.00012369, 0.00009746, -0.00018113, 0.00011494, -0.00010320, -0.00000229, -0.00001096, -0.00006517, 0.00002660, -0.00000637, -0.00000651, 0.00001839, -0.00003202, 0.00001658, 0.00000693, -0.00001585, },
		{ -0.00001713, 0.00002128, -0.00000525, 0.00000145, 0.00001742, -0.00001251, 0.00000242, -0.00000363, 0.00002053, 0.00000976, -0.00002378, 0.00001972, -0.00001838, -0.00003310, 0.00001310, 0.00000221, 0.00002139, -0.00000184, -0.00002862, 0.00002645, -0.00003887, 0.00002184, -0.00000121, -0.00003516, 0.00003253, 0.00000372, -0.00001214, -0.00001094, 0.00006077, -0.00006060, 0.00003097, -0.00006762, 0.00007119, 0.00000621, -0.00013804, 0.00019377, -0.00017796, 0.00013915, -0.00001888, -0.00005702, 0.00007939, -0.00012311, 0.00009630, -0.00023343, 0.00022354, -0.00057870, 0.00004891, 0.00133880, -0.00040287, -0.00028618, 0.00015668, -0.00019256, 0.00009408, -0.00007138, -0.00003099, 0.00001177, -0.00007555, 0.00001541, 0.00000958, -0.00001509, 0.00002040, -0.00002876, 0.00000572, 0.00001536, },
		{ 0.00002059, -0.00001510, 0.00001578, -0.00000050, -0.00000300, 0.00002005, -0.00000927, -0.00000042, -0.00000512, 0.00001699, 0.00001772, -0.00002681, 0.00001758, -0.00000856, -0.00003844, 0.00000858, 0.00000382, 0.00001782, 0.00000871, -0.00003511, 0.00002841, -0.00003372, 0.00000853, 0.00001513, -0.00004622, 0.00002979, 0.00000611, -0.00000415, -0.00002269, 0.00006553, -0.00005525, 0.00002015, -0.00004777, 0.00002704, 0.00006498, -0.00018314, 0.00019199, -0.00013345, 0.00007249, 0.00005716, -0.00010770, 0.00011805, -0.00014397, 0.00010046, -0.00021268, 0.00016613, -0.00040734, -0.00020995, 0.00130593, -0.00013740, -0.00041927, 0.00017709, -0.00017560, 0.00005749, -0.00003536, -0.00005644, 0.00002671, -0.00007525, -0.00000312, 0.00002504, -0.00001867, 0.00001793, -0.00002262, -0.00000516, },
		{ -0.00001435, 0.00002198, -0.00001034, 0.00000979, 0.00000351, -0.00000617, 0.00002054, -0.00000467, -0.00000271, -0.00000640, 0.00001189, 0.00002540, -0.00002760, 0.00001286, 0.00000158, -0.00004100, 0.00000214, 0.00000602, 0.00001333, 0.00001837, -0.00003832, 0.00002594, -0.00002393, -0.00000730, 0.00003032, -0.00005261, 0.00002335, 0.00000845, 0.00000339, -0.00003010, 0.00006333, -0.00004233, 0.00000145, -0.00002042, -0.00001869, 0.00011187, -0.00020180, 0.00015908, -0.00006846, 0.00000077, 0.00012296, -0.00013782, 0.00013304, -0.00013850, 0.00007789, -0.00016775, 0.00009050, -0.00024641, -0.00039269, 0.00118380, 0.00013579, -0.00050396, 0.00015667, -0.00013651, 0.00001358, -0.00000236, -0.00007398, 0.00003192, -0.00006546, -0.00002520, 0.00003637, -0.00001610, 0.00001197, -0.00001538, },
		{ -0.00000893, -0.00002065, 0.00001959, -0.00000394, 0.00000442, 0.00000610, -0.00000771, 0.00001902, 0.00000066, -0.00000414, -0.00000721, 0.00000546, 0.00003197, -0.00002568, 0.00000603, 0.00001101, -0.00004041, -0.00000566, 0.00000808, 0.00000881, 0.00002603, -0.00003755, 0.00001920, -0.00001080, -0.00002349, 0.00004215, -0.00005330, 0.00001356, 0.00001068, 0.00000931, -0.00003198, 0.00005416, -0.00002317, -0.00002208, 0.00000894, -0.00005823, 0.00013966, -0.00019187, 0.00010105, 0.00000502, -0.00006328, 0.00016866, -0.00014238, 0.00012438, -0.00011109, 0.00003745, -0.00011179, 0.00001402, -0.00011377, -0.00049718, 0.00099675, 0.00038687, -0.00052933, 0.00010081, -0.00008535, -0.00002873, 0.00002190, -0.00008124, 0.00002782, -0.00004913, -0.00004664, 0.00004093, -0.00000771, 0.00000443, },
		{ -0.00000234, -0.00000479, -0.00002360, 0.00001420, 0.00000273, 0.00000066, 0.00000683, -0.00000755, 0.00001589, 0.00000603, -0.00000456, -0.00000732, -0.00000182, 0.00003658, -0.00002088, -0.00000218, 0.00001875, -0.00003667, -0.00001396, 0.00000924, 0.00000504, 0.00003093, -0.00003271, 0.00000906, 0.00000386, -0.00003774, 0.00004884, -0.00004798, 0.00000124, 0.00001272, 0.00001266, -0.00002800, 0.00003926, -0.00000044, -0.00004634, 0.00003449, -0.00008520, 0.00014475, -0.00015623, 0.00002808, 0.00007386, -0.00010909, 0.00018870, -0.00012154, 0.00009683, -0.00006975, -0.00001014, -0.00005732, -0.00005001, -0.00001882, -0.00053135, 0.00077315, 0.00059166, -0.00049377, 0.00002097, -0.00003361, -0.00006192, 0.00003426, -0.00007844, 0.00001676, -0.00003005, -0.00006382, 0.00003769, 0.00000475, },
		{ 0.00001859, -0.00000615, -0.00000370, -0.00002357, 0.00000710, 0.00000827, -0.00000075, 0.00000563, -0.00000594, 0.00001175, 0.00001078, -0.00000393, -0.00000664, -0.00000929, 0.00003847, -0.00001337, -0.00001083, 0.00002403, -0.00003016, -0.00002172, 0.00000882, 0.00000254, 0.00003276, -0.00002431, -0.00000305, 0.00001794, -0.00004803, 0.00004938, -0.00003716, -0.00001245, 0.00001439, 0.00001297, -0.00001884, 0.00002094, 0.00002239, -0.00006691, 0.00005127, -0.00009581, 0.00012765, -0.00010204, -0.00004747, 0.00012627, -0.00013022, 0.00018261, -0.00008011, 0.00005851, -0.00002399, -0.00005471, -0.00001396, -0.00009384, 0.00003707, -0.00051016, 0.00054086, 0.00073450, -0.00040460, -0.00006788, 0.00000822, -0.00008138, 0.00003454, -0.00006792, 0.00000218, -0.00001193, -0.00007437, 0.00002728, },
		{ 0.00001173, 0.00003088, -0.00000557, -0.00000544, -0.00002161, -0.00000019, 0.00001149, 0.00000046, 0.00000279, -0.00000343, 0.00000733, 0.00001435, -0.00000231, -0.00000520, -0.00001615, 0.00003711, -0.00000367, -0.00001885, 0.00002634, -0.00002160, -0.00002793, 0.00000643, 0.00000147, 0.00003171, -0.00001341, -0.00001535, 0.00002946, -0.00005290, 0.00004361, -0.00002207, -0.00002609, 0.00001546, 0.00001029, -0.00000609, 0.00000224, 0.00004172, -0.00008006, 0.00005612, -0.00008944, 0.00009274, -0.00003925, -0.00011346, 0.00015402, -0.00012537, 0.00015471, -0.00002624, 0.00001884, 0.00001735, -0.00008864, 0.00001290, -0.00011547, 0.00005892, -0.00045188, 0.00032346, 0.00080943, -0.00027602, -0.00015013, 0.00003269, -0.00008608, 0.00002517, -0.00005331, -0.00001243, 0.00000248, -0.00007757, },
		{ -0.00007422, -0.00000612, 0.00003907, -0.00000029, -0.00000886, -0.00001919, -0.00000624, 0.00001169, 0.00000405, -0.00000102, -0.00000071, 0.00000336, 0.00001632, 0.00000017, -0.00000316, -0.00002162, 0.00003232, 0.00000729, -0.00002519, 0.00002552, -0.00001197, -0.00003172, 0.00000205, 0.00000157, 0.00002840, -0.00000139, -0.00002604, 0.00003682, -0.00005173, 0.00003229, -0.00000452, -0.00003819, 0.00001560, 0.00000526, 0.00000799, -0.00001364, 0.00005450, -0.00008344, 0.00004828, -0.00006859, 0.00004722, 0.00002142, -0.00016013, 0.00015376, -0.00009830, 0.00011282, 0.00003055, -0.00001346, 0.00004785, -0.00010792, 0.00002228, -0.00011758, 0.00005561, -0.00037467, 0.00013780, 0.00081960, -0.00012626, -0.00021246, 0.00003674, -0.00007840, 0.00001030, -0.00003842, -0.00002435, 0.00001183, },
		{ 0.00001623, -0.00006621, -0.00002333, 0.00004167, 0.00000879, -0.00001220, -0.00001777, -0.00001003, 0.00000879, 0.00000930, -0.00000493, 0.00000144, 0.00000049, 0.00001647, 0.00000329, -0.00000077, -0.00002505, 0.00002438, 0.00001835, -0.00002895, 0.00002180, -0.00000235, -0.00003255, -0.00000393, 0.00000227, 0.00002373, 0.00001025, -0.00003361, 0.00003911, -0.00004477, 0.00001694, 0.00001338, -0.00004739, 0.00001447, -0.00000110, 0.00002086, -0.00002396, 0.00005890, -0.00007657, 0.00002943, -0.00003829, -0.00000039, 0.00007040, -0.00018177, 0.00012742, -0.00005674, 0.00006629, 0.00008137, -0.00003209, 0.00006456, -0.00011238, 0.00001687, -0.00010587, 0.00003731, -0.00029385, -0.00000698, 0.00077539, 0.00002573, -0.00024598, 0.00002189, -0.00006318, -0.00000554, -0.00002626, -0.00003224, },
		{ -0.00003617, 0.00001685, -0.00005584, -0.00003762, 0.00003847, 0.00001976, -0.00001352, -0.00001843, -0.00001122, 0.00000331, 0.00001513, -0.00000795, 0.00000232, -0.00000081, 0.00001478, 0.00000680, 0.00000171, -0.00002606, 0.00001399, 0.00002819, -0.00002953, 0.00001575, 0.00000618, -0.00003028, -0.00001072, 0.00000279, 0.00001869, 0.00002019, -0.00003705, 0.00003623, -0.00003314, -0.00000039, 0.00002950, -0.00005256, 0.00001174, -0.00000749, 0.00003017, -0.00002696, 0.00005466, -0.00006091, 0.00000324, -0.00000489, -0.00004196, 0.00010101, -0.00017744, 0.00008142, -0.00001054, 0.00002397, 0.00011965, -0.00003410, 0.00006807, -0.00010491, 0.00000158, -0.00008705, 0.00001325, -0.00022030, -0.00010900, 0.00069161, 0.00016335, -0.00024734, -0.00000649, -0.00004624, -0.00001859, -0.00001832, },
		{ -0.00001447, -0.00003732, 0.00001535, -0.00004514, -0.00004777, 0.00003057, 0.00003016, -0.00001126, -0.00002148, -0.00001016, -0.00000369, 0.00002026, -0.00000917, 0.00000143, -0.00000033, 0.00001148, 0.00001037, 0.00000408, -0.00002462, 0.00000221, 0.00003554, -0.00002672, 0.00000823, 0.00001275, -0.00002523, -0.00001729, 0.00000236, 0.00001411, 0.00002747, -0.00003600, 0.00002886, -0.00001854, -0.00001744, 0.00004198, -0.00005306, 0.00000724, -0.00001267, 0.00003420, -0.00002222, 0.00004317, -0.00003949, -0.00002545, 0.00002515, -0.00007118, 0.00011044, -0.00015080, 0.00002519, 0.00003051, -0.00000761, 0.00014211, -0.00002003, 0.00006187, -0.00009026, -0.00001799, -0.00006719, -0.00000974, -0.00016003, -0.00017199, 0.00058447, 0.00027444, -0.00021874, -0.00004075, -0.00003282, -0.00002682, },
		{ -0.00003021, -0.00001328, -0.00003731, 0.00001317, -0.00003534, -0.00005377, 0.00001999, 0.00003762, -0.00000466, -0.00002641, -0.00000780, -0.00001087, 0.00002347, -0.00000794, -0.00000146, 0.00000183, 0.00000704, 0.00001357, 0.00000625, -0.00002103, -0.00000964, 0.00003939, -0.00002075, 0.00000029, 0.00001676, -0.00001812, -0.00002252, 0.00000036, 0.00001057, 0.00003163, -0.00003082, 0.00001833, -0.00000306, -0.00003208, 0.00004947, -0.00004872, 0.00000099, -0.00001568, 0.00003216, -0.00001077, 0.00002719, -0.00001625, -0.00005165, 0.00004662, -0.00008463, 0.00009986, -0.00010898, -0.00003093, 0.00005866, -0.00002534, 0.00014888, 0.00000677, 0.00005114, -0.00007354, -0.00003704, -0.00005055, -0.00002781, -0.00011480, -0.00020327, 0.00046899, 0.00035244, -0.00016675, -0.00007278, -0.00002632, },
		{ -0.00002757, -0.00003049, -0.00001277, -0.00003747, 0.00001103, -0.00002671, -0.00005662, 0.00000908, 0.00004042, 0.00000593, -0.00003191, -0.00000549, -0.00001686, 0.00002381, -0.00000399, -0.00000617, 0.00000525, 0.00000211, 0.00001595, 0.00000828, -0.00001589, -0.00002023, 0.00003912, -0.00001234, -0.00000700, 0.00001798, -0.00000998, -0.00002540, -0.00000346, 0.00000822, 0.00003275, -0.00002242, 0.00000637, 0.00001125, -0.00004263, 0.00005130, -0.00003994, -0.00000670, -0.00001605, 0.00002437, 0.00000510, 0.00001025, 0.00000481, -0.00007127, 0.00005638, -0.00008214, 0.00007385, -0.00006094, -0.00007772, 0.00006959, -0.00002976, 0.00014283, 0.00004128, 0.00004137, -0.00005897, -0.00005249, -0.00003915, -0.00003978, -0.00008333, -0.00021147, 0.00035711, 0.00039624, -0.00010057, -0.00009583, },
		{ -0.00010580, -0.00003491, -0.00003037, -0.00001122, -0.00003829, 0.00000873, -0.00001864, -0.00005784, -0.00000010, 0.00003797, 0.00001907, -0.00003617, -0.00000457, -0.00002056, 0.00002080, 0.00000249, -0.00001221, 0.00000923, -0.00000250, 0.00001704, 0.00001031, -0.00000997, -0.00002838, 0.00003466, -0.00000255, -0.00001265, 0.00001659, -0.00000196, -0.00002527, -0.00000891, 0.00000680, 0.00003135, -0.00001213, -0.00000522, 0.00002264, -0.00004806, 0.00004757, -0.00002763, -0.00001523, -0.00001390, 0.00001216, 0.00002228, -0.00000407, 0.00002043, -0.00008181, 0.00005379, -0.00006644, 0.00003916, -0.00001554, -0.00010874, 0.00006300, -0.00002444, 0.00012832, 0.00007817, 0.00003717, -0.00004903, -0.00006323, -0.00003303, -0.00004643, -0.00006271, -0.00020479, 0.00025685, 0.00040915, -0.00002994, },
		{ 0.00003664, -0.00010178, -0.00004486, -0.00003241, -0.00000799, -0.00003913, 0.00000528, -0.00001014, -0.00005882, -0.00000635, 0.00003093, 0.00003250, -0.00003732, -0.00000608, -0.00002136, 0.00001448, 0.00001087, -0.00001871, 0.00001289, -0.00000596, 0.00001640, 0.00001255, -0.00000407, -0.00003328, 0.00002651, 0.00000731, -0.00001596, 0.00001312, 0.00000480, -0.00002187, -0.00001536, 0.00000577, 0.00002825, -0.00000140, -0.00001479, 0.00002994, -0.00004812, 0.00003904, -0.00001311, -0.00002375, -0.00000989, -0.00000239, 0.00003748, -0.00001285, 0.00002877, -0.00008276, 0.00004057, -0.00004233, 0.00000311, 0.00002007, -0.00012128, 0.00004222, -0.00001467, 0.00010985, 0.00011291, 0.00004154, -0.00004422, -0.00006991, -0.00003090, -0.00004946, -0.00004974, -0.00018988, 0.00017231, 0.00039731, },
		{ 0.00036791, 0.00009316, -0.00008573, -0.00005330, -0.00003813, -0.00000383, -0.00003842, -0.00000057, -0.00000042, -0.00006035, -0.00000963, 0.00002114, 0.00004360, -0.00003391, -0.00001039, -0.00001927, 0.00000545, 0.00002011, -0.00002462, 0.00001524, -0.00000750, 0.00001379, 0.00001511, 0.00000117, -0.00003459, 0.00001566, 0.00001588, -0.00001654, 0.00000836, 0.00000938, -0.00001552, -0.00002183, 0.00000438, 0.00002438, 0.00000842, -0.00002119, 0.00003265, -0.00004329, 0.00002704, 0.00000207, -0.00003124, -0.00000508, -0.00001687, 0.00004780, -0.00001443, 0.00002967, -0.00007545, 0.00002017, -0.00001546, -0.00002784, 0.00004165, -0.00011644, 0.00001310, -0.00000593, 0.00009079, 0.00014244, 0.00005562, -0.00004338, -0.00007426, -0.00003098, -0.00005063, -0.00004171, -0.00017141, 0.00010451, },
		{ 0.00005238, 0.00032777, 0.00013652, -0.00006161, -0.00005674, -0.00004724, -0.00000082, -0.00003425, -0.00000974, 0.00001063, -0.00006220, -0.00001102, 0.00001107, 0.00005002, -0.00002532, -0.00001715, -0.00001489, -0.00000522, 0.00002894, -0.00002881, 0.00001542, -0.00000661, 0.00000918, 0.00001797, 0.00000537, -0.00003252, 0.00000350, 0.00002195, -0.00001445, 0.00000327, 0.00001124, -0.00000699, -0.00002712, 0.00000196, 0.00002052, 0.00001630, -0.00002381, 0.00003103, -0.00003466, 0.00001324, 0.00001628, -0.00003668, -0.00000071, -0.00002899, 0.00005132, -0.00000874, 0.00002460, -0.00006262, -0.00000313, 0.00000888, -0.00004924, 0.00004825, -0.00009840, -0.00001755, -0.00000240, 0.00007276, 0.00016524, 0.00007903, -0.00004419, -0.00007838, -0.00003171, -0.00005117, -0.00003677, -0.00015225, },
		{ -0.00013401, 0.00001395, 0.00028241, 0.00016625, -0.00003403, -0.00005339, -0.00005764, -0.00000175, -0.00002508, -0.00002215, 0.00002204, -0.00006320, -0.00001238, 0.00000331, 0.00005023, -0.00001201, -0.00002521, -0.00000930, -0.00001621, 0.00003602, -0.00003024, 0.00001277, -0.00000309, 0.00000287, 0.00002087, 0.00000848, -0.00002773, -0.00000843, 0.00002467, -0.00001017, -0.00000120, 0.00001026, 0.00000253, -0.00003011, -0.00000195, 0.00001721, 0.00002165, -0.00002270, 0.00002591, -0.00002375, -0.00000061, 0.00002807, -0.00003921, 0.00000209, -0.00003703, 0.00004745, 0.00000278, 0.00001616, -0.00004760, -0.00002527, 0.00002679, -0.00005925, 0.00004192, -0.00007308, -0.00004360, -0.00000601, 0.00005552, 0.00018094, 0.00011042, -0.00004395, -0.00008402, -0.00003217, -0.00005164, -0.00003390, },
		{ -0.00003250, -0.00011755, -0.00001298, 0.00023575, 0.00018368, -0.00000699, -0.00004364, -0.00006583, -0.00000916, -0.00001062, -0.00003639, 0.00003184, -0.00006151, -0.00001572, -0.00000019, 0.00004394, 0.00000448, -0.00003292, -0.00000376, -0.00002614, 0.00004013, -0.00002819, 0.00000704, 0.00000281, -0.00000450, 0.00002333, 0.00001074, -0.00002121, -0.00001866, 0.00002366, -0.00000449, -0.00000424, 0.00000678, 0.00001168, -0.00002990, -0.00000742, 0.00001463, 0.00002436, -0.00001843, 0.00001855, -0.00001220, -0.00001289, 0.00003632, -0.00003827, 0.00000257, -0.00004014, 0.00003694, 0.00001752, 0.00000748, -0.00003353, -0.00004317, 0.00003629, -0.00005854, 0.00002677, -0.00004671, -0.00006089, -0.00001612, 0.00003755, 0.00018957, 0.00014809, -0.00004008, -0.00009226, -0.00003219, -0.00005205, },
		},
		{ { 0.00000494, -0.00000168, -0.00003261, -0.00002493, -0.00006287, -0.00008761, 0.00003642, 0.00017839, 0.00015722, -0.00000433, -0.00002050, -0.00005804, -0.00003435, 0.00004289, -0.00004615, 0.00004155, -0.00000604, -0.00007728, 0.00005453, -0.00000594, -0.00001275, 0.00000339, -0.00001611, -0.00000513, -0.00002911, 0.00002409, 0.00003805, -0.00002321, -0.00003053, 0.00004217, -0.00000983, -0.00003673, 0.00003957, -0.00003673, -0.00000983, 0.00004217, -0.00003053, -0.00002321, 0.00003805, 0.00002409, -0.00002911, -0.00000513, -0.00001611, 0.00000339, -0.00001275, -0.00000594, 0.00005453, -0.00007728, -0.00000604, 0.00004155, -0.00004615, 0.00004289, -0.00003435, -0.00005804, -0.00002050, -0.00000433, 0.00015722, 0.00017839, 0.00003642, -0.00008761, -0.00006287, -0.00002493, -0.00003261, -0.00000168, },
		{ -0.00000662, 0.00000499, 0.00000160, -0.00002678, -0.00002367, -0.00005285, -0.00008271, 0.00000699, 0.00013294, 0.00017345, 0.00001588, -0.00001309, -0.00003275, -0.00007163, 0.00005512, -0.00003333, 0.00000868, 0.00003522, -0.00009487, 0.00003762, 0.00002081, -0.00002827, 0.00001008, -0.00001587, -0.00000196, -0.00002716, 0.00000500, 0.00004331, -0.00000304, -0.00003883, 0.00002742, 0.00001329, -0.00004896, 0.00003313, -0.00001595, -0.00003131, 0.00004647, -0.00001334, -0.00003888, 0.00002362, 0.00004074, -0.00002302, -0.00001247, -0.00001239, -0.00000774, 0.00000213, -0.00002165, 0.00005180, -0.00004492, -0.00004441, 0.00005563, -0.00003599, 0.00001608, -0.00000162, -0.00006375, -0.00003705, -0.00002296, 0.00013074, 0.00021527, 0.00008115, -0.00008894, -0.00007681, -0.00002464, -0.00003843, },
		{ -0.00004261, -0.00001377, 0.00000473, 0.00000415, -0.00002187, -0.00002149, -0.00004387, -0.00007897, -0.00000783, 0.00008583, 0.00017519, 0.00004259, -0.00001919, 0.00000445, -0.00009749, 0.00004194, 0.00000127, -0.00003133, 0.00006468, -0.00008906, 0.00000264, 0.00005061, -0.00003685, 0.00000858, -0.00001056, -0.00000318, -0.00001973, -0.00001171, 0.00003908, 0.00001655, -0.00003612, 0.00000679, 0.00003113, -0.00004888, 0.00001600, 0.00000681, -0.00004488, 0.00003913, 0.00000802, -0.00004598, 0.00000282, 0.00005019, -0.00000868, -0.00002149, -0.00000764, -0.00001851, 0.00001028, -0.00002356, 0.00003461, -0.00001006, -0.00006805, 0.00004770, -0.00001071, -0.00001095, 0.00001555, -0.00005138, -0.00005360, -0.00004376, 0.00009635, 0.00023882, 0.00013869, -0.00008083, -0.00009611, -0.00002396, },
		{ -0.00002532, -0.00004372, -0.00002273, 0.00000327, 0.00000641, -0.00001737, -0.00002035, -0.00003324, -0.00007835, -0.00001247, 0.00004556, 0.00015842, 0.00007773, -0.00003523, 0.00003883, -0.00009744, 0.00000174, 0.00004639, -0.00006214, 0.00007045, -0.00005889, -0.00004166, 0.00007269, -0.00003238, -0.00000186, -0.00000157, -0.00000688, -0.00001054, -0.00002250, 0.00002741, 0.00003095, -0.00002353, -0.00001328, 0.00003854, -0.00003690, -0.00000613, 0.00002452, -0.00004707, 0.00002260, 0.00002773, -0.00004274, -0.00001969, 0.00004921, 0.00001133, -0.00002826, -0.00000508, -0.00002509, 0.00000908, -0.00001428, 0.00001210, 0.00001620, -0.00007185, 0.00002302, 0.00001723, -0.00002654, 0.00001459, -0.00002960, -0.00006128, -0.00006632, 0.00005458, 0.00024556, 0.00020450, -0.00005770, -0.00012013, },
		{ -0.00014617, -0.00003155, -0.00004117, -0.00003252, 0.00000020, 0.00000764, -0.00001170, -0.00002214, -0.00002030, -0.00007916, -0.00001409, 0.00001983, 0.00012265, 0.00011713, -0.00005034, 0.00005436, -0.00006528, -0.00005546, 0.00008395, -0.00006897, 0.00004865, -0.00001235, -0.00008185, 0.00007736, -0.00001298, -0.00001819, 0.00000754, -0.00001013, -0.00000306, -0.00002597, 0.00001190, 0.00003731, -0.00000498, -0.00002661, 0.00003397, -0.00001739, -0.00002609, 0.00003217, -0.00003824, 0.00000203, 0.00004074, -0.00003018, -0.00003870, 0.00003722, 0.00003223, -0.00002891, -0.00000660, -0.00002604, -0.00000025, -0.00000001, -0.00000650, 0.00002764, -0.00005825, -0.00000791, 0.00003627, -0.00002586, 0.00000088, -0.00000945, -0.00005589, -0.00008585, 0.00000622, 0.00023220, 0.00027254, -0.00001497, },
		{ 0.00005074, -0.00016975, -0.00004542, -0.00003509, -0.00004231, -0.00000389, 0.00000622, -0.00000351, -0.00002684, -0.00000741, -0.00007684, -0.00001998, 0.00001232, 0.00007341, 0.00015052, -0.00005083, 0.00004097, -0.00000726, -0.00011025, 0.00009579, -0.00004509, 0.00000590, 0.00003572, -0.00010442, 0.00005994, 0.00001765, -0.00003431, 0.00001255, -0.00001037, 0.00000065, -0.00002290, -0.00000345, 0.00003514, 0.00001418, -0.00002932, 0.00001993, 0.00000292, -0.00003787, 0.00002846, -0.00002229, -0.00001667, 0.00004427, -0.00001175, -0.00004991, 0.00001655, 0.00004840, -0.00002129, -0.00001187, -0.00002245, -0.00001380, 0.00001226, -0.00001515, 0.00002432, -0.00003515, -0.00003405, 0.00004011, -0.00001164, -0.00001594, 0.00000079, -0.00003968, -0.00009574, -0.00004520, 0.00019563, 0.00033488, },
		{ 0.00038107, 0.00014094, -0.00018435, -0.00006958, -0.00002588, -0.00005184, -0.00000770, 0.00000101, 0.00000661, -0.00003168, 0.00000073, -0.00006650, -0.00003418, 0.00002035, 0.00002215, 0.00016499, -0.00002674, -0.00000006, 0.00005869, -0.00014059, 0.00007137, 0.00000432, -0.00004224, 0.00006883, -0.00010089, 0.00002306, 0.00005098, -0.00004316, 0.00001043, -0.00000663, 0.00000043, -0.00001564, -0.00001536, 0.00002610, 0.00002889, -0.00002105, 0.00000207, 0.00001747, -0.00003869, 0.00001598, -0.00000498, -0.00002881, 0.00003854, 0.00000778, -0.00005113, -0.00000816, 0.00005519, -0.00000597, -0.00001840, -0.00001702, -0.00002689, 0.00001742, -0.00001255, 0.00001140, -0.00001209, -0.00004798, 0.00002918, 0.00000830, -0.00002683, -0.00000122, -0.00001979, -0.00009149, -0.00009236, 0.00013466, },
		{ 0.00005278, 0.00039836, 0.00025356, -0.00018087, -0.00010649, -0.00001426, -0.00006101, -0.00001053, -0.00000713, 0.00001537, -0.00003184, -0.00000045, -0.00004619, -0.00005504, 0.00003500, -0.00001702, 0.00015079, 0.00002239, -0.00005503, 0.00010875, -0.00013052, 0.00001327, 0.00006289, -0.00007670, 0.00007503, -0.00007146, -0.00002361, 0.00007635, -0.00003947, 0.00000079, 0.00000005, -0.00000231, -0.00000712, -0.00002194, 0.00001326, 0.00003585, -0.00000490, -0.00001288, 0.00002219, -0.00002980, 0.00000006, 0.00000819, -0.00003239, 0.00002638, 0.00002384, -0.00004280, -0.00003133, 0.00005036, 0.00001373, -0.00002263, -0.00001251, -0.00003600, 0.00001362, -0.00000174, -0.00000368, 0.00000346, -0.00004796, 0.00000926, 0.00002521, -0.00002708, -0.00001189, -0.00000437, -0.00007366, -0.00012562, },
		{ -0.00013670, -0.00003977, 0.00037396, 0.00038041, -0.00014748, -0.00015748, -0.00000225, -0.00006858, -0.00001371, -0.00001468, 0.00001771, -0.00002297, -0.00001232, -0.00001920, -0.00007528, 0.00004423, -0.00003268, 0.00010670, 0.00008548, -0.00010220, 0.00012226, -0.00007745, -0.00006212, 0.00010840, -0.00008221, 0.00005191, -0.00002538, -0.00006647, 0.00008482, -0.00002201, -0.00001366, 0.00000709, -0.00000541, 0.00000011, -0.00002291, -0.00000004, 0.00003433, 0.00001382, -0.00001953, 0.00001687, -0.00001581, -0.00001344, 0.00001394, -0.00002845, 0.00001203, 0.00003336, -0.00002767, -0.00004801, 0.00003479, 0.00003274, -0.00002160, -0.00001039, -0.00003967, 0.00000219, 0.00001184, -0.00001430, 0.00000833, -0.00003734, -0.00001177, 0.00003309, -0.00001747, -0.00002445, 0.00000139, -0.00004844, },
		{ -0.00002506, -0.00012291, -0.00012638, 0.00029957, 0.00050498, -0.00007130, -0.00022036, 0.00000520, -0.00007096, -0.00002096, -0.00001658, 0.00000942, -0.00000436, -0.00003082, 0.00000616, -0.00008497, 0.00003805, -0.00002134, 0.00004242, 0.00014289, -0.00012001, 0.00009036, 0.00000521, -0.00012951, 0.00012115, -0.00005371, 0.00000803, 0.00002204, -0.00009267, 0.00007249, 0.00000561, -0.00002797, 0.00001149, -0.00000699, 0.00000451, -0.00001922, -0.00001111, 0.00002597, 0.00002941, -0.00001559, 0.00000507, -0.00000267, -0.00002031, 0.00001216, -0.00002030, -0.00000048, 0.00003554, -0.00000983, -0.00005521, 0.00001214, 0.00004596, -0.00001417, -0.00001035, -0.00003842, -0.00001352, 0.00002270, -0.00001681, 0.00000384, -0.00002222, -0.00002742, 0.00003056, -0.00000273, -0.00003268, -0.00000248, },
		{ -0.00001142, -0.00001151, -0.00008973, -0.00018779, 0.00017663, 0.00060250, 0.00005752, -0.00028621, -0.00000096, -0.00006248, -0.00003686, -0.00000913, -0.00000960, 0.00001903, -0.00004696, 0.00001959, -0.00007647, 0.00001347, 0.00001022, -0.00002376, 0.00017362, -0.00009609, 0.00002061, 0.00009162, -0.00016362, 0.00009201, 0.00000086, -0.00003988, 0.00005498, -0.00009484, 0.00004213, 0.00003600, -0.00003673, 0.00001119, -0.00000616, 0.00000569, -0.00001257, -0.00001843, 0.00001388, 0.00003771, -0.00000255, -0.00000736, 0.00000462, -0.00001944, 0.00000554, -0.00001199, -0.00000861, 0.00003168, 0.00000662, -0.00005253, -0.00001233, 0.00004993, -0.00000167, -0.00001071, -0.00003381, -0.00002976, 0.00002707, -0.00001112, -0.00000583, -0.00000852, -0.00003461, 0.00002020, 0.00001145, -0.00003381, },
		{ -0.00002909, -0.00001932, -0.00001061, -0.00004988, -0.00020849, 0.00002011, 0.00064395, 0.00023961, -0.00033694, -0.00003290, -0.00003785, -0.00006349, 0.00000667, -0.00003341, 0.00003728, -0.00005019, 0.00001384, -0.00004908, -0.00002325, 0.00004688, -0.00007135, 0.00016385, -0.00003322, -0.00006459, 0.00015226, -0.00014895, 0.00002692, 0.00006282, -0.00007300, 0.00006299, -0.00007368, 0.00000234, 0.00006045, -0.00003615, 0.00000589, -0.00000316, 0.00000422, -0.00000474, -0.00002165, 0.00000143, 0.00003718, 0.00001511, -0.00001458, 0.00000393, -0.00001296, -0.00000179, -0.00000667, -0.00001194, 0.00002446, 0.00001870, -0.00004192, -0.00003344, 0.00004384, 0.00001267, -0.00000954, -0.00002733, -0.00004356, 0.00002357, 0.00000021, -0.00001584, 0.00000004, -0.00003381, 0.00000643, 0.00002124, },
		{ 0.00002585, -0.00002210, -0.00002216, -0.00001854, -0.00001885, -0.00018334, -0.00014196, 0.00060389, 0.00046039, -0.00034584, -0.00010178, 0.00000342, -0.00009690, 0.00002328, -0.00005097, 0.00003925, -0.00003363, -0.00001074, -0.00001101, -0.00005849, 0.00007104, -0.00008568, 0.00011333, 0.00005058, -0.00013562, 0.00016481, -0.00008644, -0.00005423, 0.00010932, -0.00007773, 0.00004495, -0.00003770, -0.00003550, 0.00007204, -0.00002546, -0.00000278, 0.00000090, 0.00000114, 0.00000270, -0.00002122, -0.00000887, 0.00002907, 0.00003160, -0.00001289, -0.00000337, -0.00000511, -0.00000609, -0.00000547, -0.00001191, 0.00001672, 0.00002519, -0.00002680, -0.00004742, 0.00002957, 0.00002496, -0.00000582, -0.00001973, -0.00005313, 0.00001283, 0.00001348, -0.00002252, 0.00000252, -0.00002763, -0.00000690, },
		{ -0.00001793, 0.00002672, -0.00001596, -0.00002033, -0.00002677, -0.00000848, -0.00012132, -0.00027422, 0.00047039, 0.00068712, -0.00028286, -0.00021089, 0.00005250, -0.00012551, 0.00002816, -0.00005049, 0.00001825, 0.00000121, -0.00004451, 0.00002285, -0.00007694, 0.00007014, -0.00006411, 0.00003648, 0.00012806, -0.00016649, 0.00012281, 0.00000581, -0.00012419, 0.00012193, -0.00005103, 0.00000946, 0.00000005, -0.00006118, 0.00006792, -0.00000721, -0.00001199, 0.00000468, -0.00000241, 0.00000858, -0.00001789, -0.00001583, 0.00001660, 0.00004187, -0.00000201, -0.00001305, -0.00000034, -0.00000553, -0.00000735, -0.00001085, 0.00001052, 0.00002659, -0.00001084, -0.00005279, 0.00001086, 0.00003216, -0.00000005, -0.00001114, -0.00005754, -0.00000334, 0.00002518, -0.00002407, -0.00000002, -0.00001892, },
		{ -0.00000956, -0.00002648, 0.00002555, -0.00001112, -0.00001819, -0.00002666, -0.00002142, -0.00004431, -0.00034416, 0.00025341, 0.00087209, -0.00012451, -0.00034844, 0.00008945, -0.00013245, 0.00000877, -0.00002557, -0.00002301, 0.00004287, -0.00007123, 0.00003695, -0.00006884, 0.00004251, -0.00001759, -0.00004241, 0.00017279, -0.00014476, 0.00003879, 0.00009828, -0.00015817, 0.00009396, -0.00000191, -0.00002853, 0.00002696, -0.00006924, 0.00005017, 0.00001373, -0.00001879, 0.00000718, -0.00000549, 0.00001209, -0.00001240, -0.00001955, 0.00000351, 0.00004327, 0.00001498, -0.00001982, -0.00000142, -0.00000084, -0.00000980, -0.00001087, 0.00000657, 0.00002451, 0.00000304, -0.00005021, -0.00000812, 0.00003303, 0.00000603, -0.00000172, -0.00005621, -0.00002271, 0.00003266, -0.00002033, -0.00000583, },
		{ -0.00001345, -0.00000040, -0.00003274, 0.00002248, -0.00000515, -0.00002114, -0.00001442, -0.00004921, 0.00001970, -0.00033354, -0.00001249, 0.00096279, 0.00013413, -0.00048342, 0.00008654, -0.00010217, -0.00004019, 0.00001941, -0.00007093, 0.00007310, -0.00007460, 0.00002242, -0.00003545, -0.00000124, 0.00003348, -0.00009825, 0.00016924, -0.00007655, -0.00005983, 0.00016038, -0.00014313, 0.00003338, 0.00005182, -0.00005356, 0.00003551, -0.00006046, 0.00002477, 0.00003196, -0.00002124, 0.00000801, -0.00000735, 0.00001275, -0.00000537, -0.00002077, -0.00000720, 0.00003607, 0.00003287, -0.00001942, -0.00000838, 0.00000522, -0.00001014, -0.00001287, 0.00000428, 0.00002086, 0.00001327, -0.00004181, -0.00002406, 0.00002839, 0.00001036, 0.00000768, -0.00004870, -0.00004285, 0.00003411, -0.00001197, },
		{ -0.00000003, -0.00002176, 0.00000787, -0.00003594, 0.00001570, 0.00000536, -0.00003176, 0.00000599, -0.00007536, 0.00004672, -0.00024659, -0.00027353, 0.00091721, 0.00046975, -0.00056825, 0.00001734, -0.00002956, -0.00011086, 0.00006788, -0.00010468, 0.00007412, -0.00004616, -0.00001755, 0.00001007, -0.00004233, 0.00006756, -0.00011474, 0.00011885, 0.00001544, -0.00014035, 0.00017142, -0.00008296, -0.00003970, 0.00009029, -0.00005567, 0.00002577, -0.00004108, -0.00000087, 0.00004345, -0.00001892, 0.00000733, -0.00000745, 0.00001048, 0.00000251, -0.00002025, -0.00001412, 0.00002309, 0.00004622, -0.00001018, -0.00001845, 0.00000907, -0.00000680, -0.00001632, 0.00000239, 0.00001713, 0.00001956, -0.00003014, -0.00003526, 0.00002034, 0.00001171, 0.00001521, -0.00003502, -0.00006096, 0.00002831, },
		{ 0.00001459, 0.00001410, -0.00002949, 0.00001375, -0.00003396, 0.00000282, 0.00002178, -0.00004707, 0.00002397, -0.00008245, 0.00002650, -0.00011121, -0.00047007, 0.00071997, 0.00082834, -0.00054953, -0.00013055, 0.00007240, -0.00018084, 0.00009555, -0.00010441, 0.00003728, 0.00000904, -0.00006714, 0.00004724, -0.00006189, 0.00007100, -0.00009071, 0.00003951, 0.00010028, -0.00017666, 0.00012817, 0.00000269, -0.00010033, 0.00009912, -0.00003422, 0.00000489, -0.00001985, -0.00002057, 0.00004667, -0.00001289, 0.00000554, -0.00000547, 0.00000565, 0.00001032, -0.00001829, -0.00001747, 0.00000834, 0.00005125, 0.00000634, -0.00002724, 0.00000796, 0.00000000, -0.00001954, -0.00000031, 0.00001398, 0.00002243, -0.00001733, -0.00004149, 0.00001119, 0.00001037, 0.00001862, -0.00001637, -0.00007376, },
		{ -0.00007779, -0.00000659, 0.00002815, -0.00003454, 0.00001477, -0.00002446, -0.00001679, 0.00004113, -0.00005879, 0.00002698, -0.00006041, -0.00003248, 0.00002841, -0.00055445, 0.00039294, 0.00113315, -0.00038528, -0.00034357, 0.00017064, -0.00021873, 0.00007983, -0.00006094, -0.00003076, 0.00007212, -0.00010325, 0.00005685, -0.00004969, 0.00004395, -0.00004022, -0.00004138, 0.00015023, -0.00015848, 0.00004591, 0.00008532, -0.00012784, 0.00007494, 0.00000213, -0.00001619, -0.00000444, -0.00003139, 0.00004269, -0.00000492, 0.00000296, -0.00000139, -0.00000076, 0.00001681, -0.00001467, -0.00001853, -0.00000444, 0.00004691, 0.00002602, -0.00003045, 0.00000108, 0.00000834, -0.00002061, -0.00000435, 0.00001136, 0.00002266, -0.00000478, -0.00004346, 0.00000238, 0.00000802, 0.00001630, 0.00000439, },
		{ 0.00002284, -0.00007052, -0.00003295, 0.00003866, -0.00003414, 0.00000848, -0.00000686, -0.00003996, 0.00005605, -0.00005680, 0.00000713, -0.00001228, -0.00010517, 0.00012590, -0.00050726, -0.00000453, 0.00130288, -0.00006370, -0.00057690, 0.00021868, -0.00019589, 0.00001124, 0.00001725, -0.00010772, 0.00011640, -0.00010498, 0.00003002, -0.00000961, 0.00000038, 0.00001350, -0.00009749, 0.00015087, -0.00009471, -0.00004726, 0.00013788, -0.00011344, 0.00002660, 0.00003896, -0.00002751, 0.00000147, -0.00003408, 0.00003422, 0.00000327, -0.00000031, 0.00000433, -0.00000722, 0.00002061, -0.00000902, -0.00001870, -0.00001299, 0.00003493, 0.00004362, -0.00002543, -0.00001001, 0.00001534, -0.00001819, -0.00000934, 0.00000887, 0.00002084, 0.00000675, -0.00004187, -0.00000588, 0.00000680, 0.00000845, },
		{ -0.00000249, 0.00003413, -0.00005164, -0.00005992, 0.00004141, -0.00002586, -0.00000583, 0.00001598, -0.00005963, 0.00005745, -0.00003442, -0.00003364, 0.00004550, -0.00015906, 0.00014890, -0.00034591, -0.00038788, 0.00127564, 0.00038373, -0.00075986, 0.00017161, -0.00010060, -0.00009848, 0.00010372, -0.00016256, 0.00011806, -0.00006354, -0.00002641, 0.00004093, -0.00003799, 0.00004788, -0.00011346, 0.00010581, -0.00000956, -0.00012076, 0.00014442, -0.00006330, -0.00002833, 0.00006200, -0.00002408, -0.00000086, -0.00003196, 0.00002418, 0.00001073, -0.00000421, 0.00001062, -0.00001189, 0.00002067, -0.00000128, -0.00001878, -0.00001694, 0.00001898, 0.00005459, -0.00001221, -0.00002196, 0.00001835, -0.00001218, -0.00001424, 0.00000620, 0.00001731, 0.00001664, -0.00003696, -0.00001423, 0.00000806, },
		{ 0.00001139, -0.00001238, 0.00003472, -0.00002412, -0.00008119, 0.00003277, -0.00000931, -0.00002564, 0.00003755, -0.00006707, 0.00003905, 0.00000695, -0.00008240, 0.00008889, -0.00016710, 0.00009112, -0.00012152, -0.00066968, 0.00103116, 0.00088091, -0.00081318, 0.00000521, 0.00005119, -0.00021582, 0.00016187, -0.00016753, 0.00006723, 0.00001196, -0.00009101, 0.00007777, -0.00005131, 0.00005035, -0.00009024, 0.00003448, 0.00006692, -0.00015252, 0.00010538, 0.00000404, -0.00007036, 0.00006273, -0.00000776, -0.00000639, -0.00002888, 0.00001445, 0.00001728, -0.00000846, 0.00001591, -0.00001320, 0.00001667, 0.00000788, -0.00001862, -0.00001751, 0.00000322, 0.00005644, 0.00000663, -0.00003081, 0.00001595, -0.00000372, -0.00001782, 0.00000337, 0.00001236, 0.00002404, -0.00002851, -0.00002326, },
		{ -0.00003253, 0.00001453, -0.00001688, 0.00002411, 0.00000588, -0.00009035, 0.00001165, 0.00001258, -0.00004507, 0.00004959, -0.00005572, 0.00000145, 0.00005643, -0.00011925, 0.00009618, -0.00011929, -0.00002449, 0.00009605, -0.00078535, 0.00060314, 0.00132096, -0.00067399, -0.00026757, 0.00021415, -0.00029463, 0.00015885, -0.00011085, -0.00002538, 0.00009584, -0.00013545, 0.00008030, -0.00003081, 0.00002346, -0.00004363, -0.00003608, 0.00011006, -0.00013639, 0.00003677, 0.00006486, -0.00008547, 0.00004160, 0.00001390, -0.00000912, -0.00002721, 0.00000536, 0.00002313, -0.00001225, 0.00001851, -0.00001030, 0.00000937, 0.00001704, -0.00001733, -0.00001663, -0.00000904, 0.00004927, 0.00002678, -0.00003340, 0.00000848, 0.00000520, -0.00001918, 0.00000073, 0.00000658, 0.00002779, -0.00001639, },
		{ -0.00000132, -0.00004019, 0.00001429, -0.00001342, 0.00000557, 0.00003046, -0.00008335, -0.00001901, 0.00003349, -0.00005662, 0.00004550, -0.00002502, -0.00004618, 0.00009657, -0.00012526, 0.00005790, -0.00002799, -0.00015337, 0.00024024, -0.00071392, 0.00007588, 0.00159276, -0.00032112, -0.00059046, 0.00032388, -0.00029289, 0.00008062, -0.00000463, -0.00012978, 0.00015460, -0.00013616, 0.00004115, 0.00001667, -0.00001669, 0.00000306, -0.00008197, 0.00010918, -0.00008332, -0.00003639, 0.00009908, -0.00007014, 0.00000753, 0.00003106, -0.00000510, -0.00002678, -0.00000386, 0.00002823, -0.00001429, 0.00001713, -0.00000354, 0.00000053, 0.00002438, -0.00001381, -0.00001601, -0.00001622, 0.00003551, 0.00004368, -0.00002831, -0.00000214, 0.00001248, -0.00001809, -0.00000133, 0.00000101, 0.00002680, },
		{ 0.00002067, 0.00001469, -0.00004340, 0.00000812, -0.00000252, -0.00001453, 0.00004260, -0.00006057, -0.00005213, 0.00004552, -0.00005422, 0.00002375, 0.00001789, -0.00008835, 0.00010996, -0.00009107, -0.00001756, 0.00007552, -0.00024532, 0.00026974, -0.00048598, -0.00043541, 0.00161338, 0.00020882, -0.00087243, 0.00031752, -0.00019144, -0.00005846, 0.00011634, -0.00020656, 0.00016014, -0.00008514, -0.00002924, 0.00007079, -0.00004784, 0.00002915, -0.00009128, 0.00007045, -0.00001616, -0.00008968, 0.00009740, -0.00003226, -0.00002592, 0.00003600, 0.00000562, -0.00002524, -0.00001400, 0.00003179, -0.00001316, 0.00001148, 0.00000554, -0.00000759, 0.00002824, -0.00000745, -0.00001645, -0.00001855, 0.00001885, 0.00005380, -0.00001629, -0.00001308, 0.00001660, -0.00001495, -0.00000265, -0.00000296, },
		{ -0.00000409, 0.00001026, 0.00002844, -0.00003951, -0.00000438, 0.00001196, -0.00002884, 0.00003890, -0.00002763, -0.00007819, 0.00004227, -0.00003640, -0.00001032, 0.00005973, -0.00010871, 0.00008677, -0.00002238, -0.00010541, 0.00015242, -0.00026318, 0.00018220, -0.00017540, -0.00081578, 0.00135570, 0.00082181, -0.00101094, 0.00015861, -0.00000707, -0.00021618, 0.00020548, -0.00022190, 0.00010220, 0.00000492, -0.00010545, 0.00010602, -0.00005163, 0.00002515, -0.00006767, 0.00001309, 0.00003968, -0.00010806, 0.00006402, 0.00001253, -0.00004614, 0.00002635, 0.00001896, -0.00001964, -0.00002477, 0.00003227, -0.00000797, 0.00000254, 0.00001453, -0.00001285, 0.00002772, 0.00000154, -0.00001764, -0.00001763, 0.00000304, 0.00005554, 0.00000018, -0.00002152, 0.00001700, -0.00001053, -0.00000352, },
		{ -0.00000454, -0.00000183, -0.00000216, 0.00003641, -0.00002745, -0.00002072, 0.00002415, -0.00003182, 0.00002127, 0.00000588, -0.00008856, 0.00002201, -0.00000795, -0.00004576, 0.00008566, -0.00009681, 0.00003012, 0.00006049, -0.00017309, 0.00017145, -0.00019692, 0.00001465, 0.00012317, -0.00098360, 0.00086167, 0.00138566, -0.00092384, -0.00014301, 0.00020643, -0.00033491, 0.00022186, -0.00016210, -0.00000543, 0.00010373, -0.00015654, 0.00010260, -0.00002209, -0.00000374, -0.00002747, -0.00003911, 0.00006581, -0.00009058, 0.00001412, 0.00004722, -0.00004644, 0.00000607, 0.00002875, -0.00000831, -0.00003434, 0.00002807, 0.00000106, -0.00000754, 0.00002091, -0.00001395, 0.00002297, 0.00001203, -0.00001845, -0.00001554, -0.00000914, 0.00004938, 0.00001765, -0.00002545, 0.00001407, -0.00000560, },
		{ -0.00000066, -0.00000633, 0.00000325, -0.00001334, 0.00003600, -0.00000868, -0.00003609, 0.00002843, -0.00002215, -0.00000331, 0.00002954, -0.00007910, -0.00001054, 0.00002114, -0.00006985, 0.00008531, -0.00005324, -0.00004356, 0.00012919, -0.00019271, 0.00012165, -0.00006793, -0.00016987, 0.00032388, -0.00091460, 0.00023531, 0.00176411, -0.00057977, -0.00052457, 0.00037168, -0.00036220, 0.00014750, -0.00004148, -0.00012792, 0.00017412, -0.00015863, 0.00005564, 0.00003124, -0.00004060, 0.00000813, -0.00006737, 0.00005708, -0.00004949, -0.00003267, 0.00006002, -0.00002841, -0.00001639, 0.00002956, 0.00000770, -0.00003970, 0.00001836, 0.00001227, -0.00001600, 0.00002284, -0.00001080, 0.00001518, 0.00002232, -0.00001746, -0.00001407, -0.00001643, 0.00003743, 0.00003278, -0.00002410, 0.00000887, },
		{ 0.00000265, 0.00000410, -0.00000906, 0.00000947, -0.00002001, 0.00002667, 0.00001279, -0.00004501, 0.00002190, -0.00000347, -0.00002514, 0.00003606, -0.00005247, -0.00004509, 0.00003929, -0.00007343, 0.00005742, 0.00000947, -0.00010970, 0.00015802, -0.00015253, 0.00001767, 0.00007873, -0.00030475, 0.00037623, -0.00064872, -0.00038374, 0.00185578, -0.00001613, -0.00088184, 0.00041270, -0.00027253, -0.00000240, 0.00010040, -0.00022134, 0.00018645, -0.00010527, -0.00002176, 0.00008695, -0.00006465, 0.00002241, -0.00006481, 0.00002219, -0.00000437, -0.00006005, 0.00004855, -0.00000084, -0.00003172, 0.00001939, 0.00002454, -0.00003790, 0.00000394, 0.00002288, -0.00002039, 0.00001979, -0.00000452, 0.00000611, 0.00003060, -0.00001358, -0.00001412, -0.00001907, 0.00002264, 0.00004306, -0.00001790, },
		{ -0.00000816, -0.00000347, 0.00000856, -0.00001222, 0.00001448, -0.00001998, 0.00001042, 0.00003146, -0.00004349, 0.00000581, 0.00001681, -0.00003547, 0.00002436, -0.00001794, -0.00006930, 0.00003824, -0.00005513, 0.00001115, 0.00007080, -0.00014466, 0.00013421, -0.00006257, -0.00010481, 0.00019089, -0.00034322, 0.00027963, -0.00027702, -0.00085848, 0.00162566, 0.00066305, -0.00109692, 0.00028342, -0.00008145, -0.00018144, 0.00021181, -0.00024921, 0.00013094, -0.00001133, -0.00010279, 0.00012059, -0.00006071, 0.00000991, -0.00003789, -0.00002106, 0.00002603, -0.00006097, 0.00002006, 0.00002420, -0.00003371, 0.00000077, 0.00003681, -0.00002755, -0.00001250, 0.00002976, -0.00001938, 0.00001265, 0.00000301, -0.00000241, 0.00003545, -0.00000639, -0.00001562, -0.00001837, 0.00000793, 0.00004725, },
		{ 0.00004534, 0.00000342, -0.00000865, 0.00001248, -0.00001463, 0.00001608, -0.00001296, -0.00000856, 0.00004211, -0.00003080, -0.00001463, 0.00003040, -0.00003013, 0.00000048, 0.00001172, -0.00007368, 0.00001714, -0.00002263, -0.00003705, 0.00010985, -0.00013492, 0.00006370, 0.00004812, -0.00020210, 0.00022878, -0.00027422, 0.00008100, 0.00008710, -0.00109284, 0.00111838, 0.00131061, -0.00107449, -0.00001063, 0.00015432, -0.00032661, 0.00024804, -0.00019720, 0.00002279, 0.00009163, -0.00015710, 0.00011503, -0.00002603, -0.00002190, -0.00000250, -0.00005371, 0.00003092, -0.00003951, -0.00001216, 0.00003647, -0.00002180, -0.00002005, 0.00003972, -0.00000986, -0.00002698, 0.00003053, -0.00001327, 0.00000344, 0.00000981, -0.00000896, 0.00003612, 0.00000372, -0.00001764, -0.00001612, -0.00000440, },
		{ -0.00001303, 0.00003833, 0.00001505, -0.00001227, 0.00001528, -0.00001487, 0.00001307, -0.00000083, -0.00002514, 0.00004164, -0.00001001, -0.00003212, 0.00003150, -0.00001137, -0.00002449, 0.00002512, -0.00005605, -0.00001620, 0.00001005, -0.00006989, 0.00011353, -0.00008288, -0.00003028, 0.00014187, -0.00023878, 0.00017969, -0.00012509, -0.00014344, 0.00034295, -0.00105761, 0.00044809, 0.00177523, -0.00077494, -0.00040623, 0.00035244, -0.00038082, 0.00018922, -0.00008008, -0.00010235, 0.00016684, -0.00016282, 0.00006802, 0.00002830, -0.00005659, 0.00002359, -0.00006322, 0.00001095, -0.00000833, -0.00003454, 0.00003165, -0.00000122, -0.00003559, 0.00003118, 0.00001126, -0.00003536, 0.00002447, -0.00000394, -0.00000540, 0.00001440, -0.00001280, 0.00003253, 0.00001565, -0.00001880, -0.00001399, },
		{ -0.00001314, -0.00001765, 0.00002781, 0.00002519, -0.00001381, 0.00001616, -0.00001184, 0.00000578, 0.00001297, -0.00003487, 0.00003023, 0.00001290, -0.00004015, 0.00001937, 0.00001277, -0.00003871, 0.00001717, -0.00002323, -0.00004783, 0.00002833, -0.00007637, 0.00008189, -0.00000635, -0.00011551, 0.00018635, -0.00020116, 0.00006244, 0.00004972, -0.00031536, 0.00043202, -0.00079583, -0.00023260, 0.00194401, -0.00023260, -0.00079583, 0.00043202, -0.00031536, 0.00004972, 0.00006244, -0.00020116, 0.00018635, -0.00011551, -0.00000635, 0.00008189, -0.00007637, 0.00002833, -0.00004783, -0.00002323, 0.00001717, -0.00003871, 0.00001277, 0.00001937, -0.00004015, 0.00001290, 0.00003023, -0.00003487, 0.00001297, 0.00000578, -0.00001184, 0.00001616, -0.00001381, 0.00002519, 0.00002781, -0.00001765, },
		{ -0.00001303, -0.00001399, -0.00001880, 0.00001565, 0.00003253, -0.00001280, 0.00001440, -0.00000540, -0.00000394, 0.00002447, -0.00003536, 0.00001126, 0.00003118, -0.00003559, -0.00000122, 0.00003165, -0.00003454, -0.00000833, 0.00001095, -0.00006322, 0.00002359, -0.00005659, 0.00002830, 0.00006802, -0.00016282, 0.00016684, -0.00010235, -0.00008008, 0.00018922, -0.00038082, 0.00035244, -0.00040623, -0.00077494, 0.00177523, 0.00044809, -0.00105761, 0.00034295, -0.00014344, -0.00012509, 0.00017969, -0.00023878, 0.00014187, -0.00003028, -0.00008288, 0.00011353, -0.00006989, 0.00001005, -0.00001620, -0.00005605, 0.00002512, -0.00002449, -0.00001137, 0.00003150, -0.00003212, -0.00001001, 0.00004164, -0.00002514, -0.00000083, 0.00001307, -0.00001487, 0.00001528, -0.00001227, 0.00001505, 0.00003833, },
		{ 0.00004534, -0.00000440, -0.00001612, -0.00001764, 0.00000372, 0.00003612, -0.00000896, 0.00000981, 0.00000344, -0.00001327, 0.00003053, -0.00002698, -0.00000986, 0.00003972, -0.00002005, -0.00002180, 0.00003647, -0.00001216, -0.00003951, 0.00003092, -0.00005371, -0.00000250, -0.00002190, -0.00002603, 0.00011503, -0.00015710, 0.00009163, 0.00002279, -0.00019720, 0.00024804, -0.00032661, 0.00015432, -0.00001063, -0.00107449, 0.00131061, 0.00111838, -0.00109284, 0.00008710, 0.00008100, -0.00027422, 0.00022878, -0.00020210, 0.00004812, 0.00006370, -0.00013492, 0.00010985, -0.00003705, -0.00002263, 0.00001714, -0.00007368, 0.00001172, 0.00000048, -0.00003013, 0.00003040, -0.00001463, -0.00003080, 0.00004211, -0.00000856, -0.00001296, 0.00001608, -0.00001463, 0.00001248, -0.00000865, 0.00000342, },
		{ -0.00000816, 0.00004725, 0.00000793, -0.00001837, -0.00001562, -0.00000639, 0.00003545, -0.00000241, 0.00000301, 0.00001265, -0.00001938, 0.00002976, -0.00001250, -0.00002755, 0.00003681, 0.00000077, -0.00003371, 0.00002420, 0.00002006, -0.00006097, 0.00002603, -0.00002106, -0.00003789, 0.00000991, -0.00006071, 0.00012059, -0.00010279, -0.00001133, 0.00013094, -0.00024921, 0.00021181, -0.00018144, -0.00008145, 0.00028342, -0.00109692, 0.00066305, 0.00162566, -0.00085848, -0.00027702, 0.00027963, -0.00034322, 0.00019089, -0.00010481, -0.00006257, 0.00013421, -0.00014466, 0.00007080, 0.00001115, -0.00005513, 0.00003824, -0.00006930, -0.00001794, 0.00002436, -0.00003547, 0.00001681, 0.00000581, -0.00004349, 0.00003146, 0.00001042, -0.00001998, 0.00001448, -0.00001222, 0.00000856, -0.00000347, },
		{ 0.00000265, -0.00001790, 0.00004306, 0.00002264, -0.00001907, -0.00001412, -0.00001358, 0.00003060, 0.00000611, -0.00000452, 0.00001979, -0.00002039, 0.00002288, 0.00000394, -0.00003790, 0.00002454, 0.00001939, -0.00003172, -0.00000084, 0.00004855, -0.00006005, -0.00000437, 0.00002219, -0.00006481, 0.00002241, -0.00006465, 0.00008695, -0.00002176, -0.00010527, 0.00018645, -0.00022134, 0.00010040, -0.00000240, -0.00027253, 0.00041270, -0.00088184, -0.00001613, 0.00185578, -0.00038374, -0.00064872, 0.00037623, -0.00030475, 0.00007873, 0.00001767, -0.00015253, 0.00015802, -0.00010970, 0.00000947, 0.00005742, -0.00007343, 0.00003929, -0.00004509, -0.00005247, 0.00003606, -0.00002514, -0.00000347, 0.00002190, -0.00004501, 0.00001279, 0.00002667, -0.00002001, 0.00000947, -0.00000906, 0.00000410, },
		{ -0.00000066, 0.00000887, -0.00002410, 0.00003278, 0.00003743, -0.00001643, -0.00001407, -0.00001746, 0.00002232, 0.00001518, -0.00001080, 0.00002284, -0.00001600, 0.00001227, 0.00001836, -0.00003970, 0.00000770, 0.00002956, -0.00001639, -0.00002841, 0.00006002, -0.00003267, -0.00004949, 0.00005708, -0.00006737, 0.00000813, -0.00004060, 0.00003124, 0.00005564, -0.00015863, 0.00017412, -0.00012792, -0.00004148, 0.00014750, -0.00036220, 0.00037168, -0.00052457, -0.00057977, 0.00176411, 0.00023531, -0.00091460, 0.00032388, -0.00016987, -0.00006793, 0.00012165, -0.00019271, 0.00012919, -0.00004356, -0.00005324, 0.00008531, -0.00006985, 0.00002114, -0.00001054, -0.00007910, 0.00002954, -0.00000331, -0.00002215, 0.00002843, -0.00003609, -0.00000868, 0.00003600, -0.00001334, 0.00000325, -0.00000633, },
		{ -0.00000454, -0.00000560, 0.00001407, -0.00002545, 0.00001765, 0.00004938, -0.00000914, -0.00001554, -0.00001845, 0.00001203, 0.00002297, -0.00001395, 0.00002091, -0.00000754, 0.00000106, 0.00002807, -0.00003434, -0.00000831, 0.00002875, 0.00000607, -0.00004644, 0.00004722, 0.00001412, -0.00009058, 0.00006581, -0.00003911, -0.00002747, -0.00000374, -0.00002209, 0.00010260, -0.00015654, 0.00010373, -0.00000543, -0.00016210, 0.00022186, -0.00033491, 0.00020643, -0.00014301, -0.00092384, 0.00138566, 0.00086167, -0.00098360, 0.00012317, 0.00001465, -0.00019692, 0.00017145, -0.00017309, 0.00006049, 0.00003012, -0.00009681, 0.00008566, -0.00004576, -0.00000795, 0.00002201, -0.00008856, 0.00000588, 0.00002127, -0.00003182, 0.00002415, -0.00002072, -0.00002745, 0.00003641, -0.00000216, -0.00000183, },
		{ -0.00000409, -0.00000352, -0.00001053, 0.00001700, -0.00002152, 0.00000018, 0.00005554, 0.00000304, -0.00001763, -0.00001764, 0.00000154, 0.00002772, -0.00001285, 0.00001453, 0.00000254, -0.00000797, 0.00003227, -0.00002477, -0.00001964, 0.00001896, 0.00002635, -0.00004614, 0.00001253, 0.00006402, -0.00010806, 0.00003968, 0.00001309, -0.00006767, 0.00002515, -0.00005163, 0.00010602, -0.00010545, 0.00000492, 0.00010220, -0.00022190, 0.00020548, -0.00021618, -0.00000707, 0.00015861, -0.00101094, 0.00082181, 0.00135570, -0.00081578, -0.00017540, 0.00018220, -0.00026318, 0.00015242, -0.00010541, -0.00002238, 0.00008677, -0.00010871, 0.00005973, -0.00001032, -0.00003640, 0.00004227, -0.00007819, -0.00002763, 0.00003890, -0.00002884, 0.00001196, -0.00000438, -0.00003951, 0.00002844, 0.00001026, },
		{ 0.00002067, -0.00000296, -0.00000265, -0.00001495, 0.00001660, -0.00001308, -0.00001629, 0.00005380, 0.00001885, -0.00001855, -0.00001645, -0.00000745, 0.00002824, -0.00000759, 0.00000554, 0.00001148, -0.00001316, 0.00003179, -0.00001400, -0.00002524, 0.00000562, 0.00003600, -0.00002592, -0.00003226, 0.00009740, -0.00008968, -0.00001616, 0.00007045, -0.00009128, 0.00002915, -0.00004784, 0.00007079, -0.00002924, -0.00008514, 0.00016014, -0.00020656, 0.00011634, -0.00005846, -0.00019144, 0.00031752, -0.00087243, 0.00020882, 0.00161338, -0.00043541, -0.00048598, 0.00026974, -0.00024532, 0.00007552, -0.00001756, -0.00009107, 0.00010996, -0.00008835, 0.00001789, 0.00002375, -0.00005422, 0.00004552, -0.00005213, -0.00006057, 0.00004260, -0.00001453, -0.00000252, 0.00000812, -0.00004340, 0.00001469, },
		{ -0.00000132, 0.00002680, 0.00000101, -0.00000133, -0.00001809, 0.00001248, -0.00000214, -0.00002831, 0.00004368, 0.00003551, -0.00001622, -0.00001601, -0.00001381, 0.00002438, 0.00000053, -0.00000354, 0.00001713, -0.00001429, 0.00002823, -0.00000386, -0.00002678, -0.00000510, 0.00003106, 0.00000753, -0.00007014, 0.00009908, -0.00003639, -0.00008332, 0.00010918, -0.00008197, 0.00000306, -0.00001669, 0.00001667, 0.00004115, -0.00013616, 0.00015460, -0.00012978, -0.00000463, 0.00008062, -0.00029289, 0.00032388, -0.00059046, -0.00032112, 0.00159276, 0.00007588, -0.00071392, 0.00024024, -0.00015337, -0.00002799, 0.00005790, -0.00012526, 0.00009657, -0.00004618, -0.00002502, 0.00004550, -0.00005662, 0.00003349, -0.00001901, -0.00008335, 0.00003046, 0.00000557, -0.00001342, 0.00001429, -0.00004019, },
		{ -0.00003253, -0.00001639, 0.00002779, 0.00000658, 0.00000073, -0.00001918, 0.00000520, 0.00000848, -0.00003340, 0.00002678, 0.00004927, -0.00000904, -0.00001663, -0.00001733, 0.00001704, 0.00000937, -0.00001030, 0.00001851, -0.00001225, 0.00002313, 0.00000536, -0.00002721, -0.00000912, 0.00001390, 0.00004160, -0.00008547, 0.00006486, 0.00003677, -0.00013639, 0.00011006, -0.00003608, -0.00004363, 0.00002346, -0.00003081, 0.00008030, -0.00013545, 0.00009584, -0.00002538, -0.00011085, 0.00015885, -0.00029463, 0.00021415, -0.00026757, -0.00067399, 0.00132096, 0.00060314, -0.00078535, 0.00009605, -0.00002449, -0.00011929, 0.00009618, -0.00011925, 0.00005643, 0.00000145, -0.00005572, 0.00004959, -0.00004507, 0.00001258, 0.00001165, -0.00009035, 0.00000588, 0.00002411, -0.00001688, 0.00001453, },
		{ 0.00001139, -0.00002326, -0.00002851, 0.00002404, 0.00001236, 0.00000337, -0.00001782, -0.00000372, 0.00001595, -0.00003081, 0.00000663, 0.00005644, 0.00000322, -0.00001751, -0.00001862, 0.00000788, 0.00001667, -0.00001320, 0.00001591, -0.00000846, 0.00001728, 0.00001445, -0.00002888, -0.00000639, -0.00000776, 0.00006273, -0.00007036, 0.00000404, 0.00010538, -0.00015252, 0.00006692, 0.00003448, -0.00009024, 0.00005035, -0.00005131, 0.00007777, -0.00009101, 0.00001196, 0.00006723, -0.00016753, 0.00016187, -0.00021582, 0.00005119, 0.00000521, -0.00081318, 0.00088091, 0.00103116, -0.00066968, -0.00012152, 0.00009112, -0.00016710, 0.00008889, -0.00008240, 0.00000695, 0.00003905, -0.00006707, 0.00003755, -0.00002564, -0.00000931, 0.00003277, -0.00008119, -0.00002412, 0.00003472, -0.00001238, },
		{ -0.00000249, 0.00000806, -0.00001423, -0.00003696, 0.00001664, 0.00001731, 0.00000620, -0.00001424, -0.00001218, 0.00001835, -0.00002196, -0.00001221, 0.00005459, 0.00001898, -0.00001694, -0.00001878, -0.00000128, 0.00002067, -0.00001189, 0.00001062, -0.00000421, 0.00001073, 0.00002418, -0.00003196, -0.00000086, -0.00002408, 0.00006200, -0.00002833, -0.00006330, 0.00014442, -0.00012076, -0.00000956, 0.00010581, -0.00011346, 0.00004788, -0.00003799, 0.00004093, -0.00002641, -0.00006354, 0.00011806, -0.00016256, 0.00010372, -0.00009848, -0.00010060, 0.00017161, -0.00075986, 0.00038373, 0.00127564, -0.00038788, -0.00034591, 0.00014890, -0.00015906, 0.00004550, -0.00003364, -0.00003442, 0.00005745, -0.00005963, 0.00001598, -0.00000583, -0.00002586, 0.00004141, -0.00005992, -0.00005164, 0.00003413, },
		{ 0.00002284, 0.00000845, 0.00000680, -0.00000588, -0.00004187, 0.00000675, 0.00002084, 0.00000887, -0.00000934, -0.00001819, 0.00001534, -0.00001001, -0.00002543, 0.00004362, 0.00003493, -0.00001299, -0.00001870, -0.00000902, 0.00002061, -0.00000722, 0.00000433, -0.00000031, 0.00000327, 0.00003422, -0.00003408, 0.00000147, -0.00002751, 0.00003896, 0.00002660, -0.00011344, 0.00013788, -0.00004726, -0.00009471, 0.00015087, -0.00009749, 0.00001350, 0.00000038, -0.00000961, 0.00003002, -0.00010498, 0.00011640, -0.00010772, 0.00001725, 0.00001124, -0.00019589, 0.00021868, -0.00057690, -0.00006370, 0.00130288, -0.00000453, -0.00050726, 0.00012590, -0.00010517, -0.00001228, 0.00000713, -0.00005680, 0.00005605, -0.00003996, -0.00000686, 0.00000848, -0.00003414, 0.00003866, -0.00003295, -0.00007052, },
		{ -0.00007779, 0.00000439, 0.00001630, 0.00000802, 0.00000238, -0.00004346, -0.00000478, 0.00002266, 0.00001136, -0.00000435, -0.00002061, 0.00000834, 0.00000108, -0.00003045, 0.00002602, 0.00004691, -0.00000444, -0.00001853, -0.00001467, 0.00001681, -0.00000076, -0.00000139, 0.00000296, -0.00000492, 0.00004269, -0.00003139, -0.00000444, -0.00001619, 0.00000213, 0.00007494, -0.00012784, 0.00008532, 0.00004591, -0.00015848, 0.00015023, -0.00004138, -0.00004022, 0.00004395, -0.00004969, 0.00005685, -0.00010325, 0.00007212, -0.00003076, -0.00006094, 0.00007983, -0.00021873, 0.00017064, -0.00034357, -0.00038528, 0.00113315, 0.00039294, -0.00055445, 0.00002841, -0.00003248, -0.00006041, 0.00002698, -0.00005879, 0.00004113, -0.00001679, -0.00002446, 0.00001477, -0.00003454, 0.00002815, -0.00000659, },
		{ 0.00001459, -0.00007376, -0.00001637, 0.00001862, 0.00001037, 0.00001119, -0.00004149, -0.00001733, 0.00002243, 0.00001398, -0.00000031, -0.00001954, 0.00000000, 0.00000796, -0.00002724, 0.00000634, 0.00005125, 0.00000834, -0.00001747, -0.00001829, 0.00001032, 0.00000565, -0.00000547, 0.00000554, -0.00001289, 0.00004667, -0.00002057, -0.00001985, 0.00000489, -0.00003422, 0.00009912, -0.00010033, 0.00000269, 0.00012817, -0.00017666, 0.00010028, 0.00003951, -0.00009071, 0.00007100, -0.00006189, 0.00004724, -0.00006714, 0.00000904, 0.00003728, -0.00010441, 0.00009555, -0.00018084, 0.00007240, -0.00013055, -0.00054953, 0.00082834, 0.00071997, -0.00047007, -0.00011121, 0.00002650, -0.00008245, 0.00002397, -0.00004707, 0.00002178, 0.00000282, -0.00003396, 0.00001375, -0.00002949, 0.00001410, },
		{ -0.00000003, 0.00002831, -0.00006096, -0.00003502, 0.00001521, 0.00001171, 0.00002034, -0.00003526, -0.00003014, 0.00001956, 0.00001713, 0.00000239, -0.00001632, -0.00000680, 0.00000907, -0.00001845, -0.00001018, 0.00004622, 0.00002309, -0.00001412, -0.00002025, 0.00000251, 0.00001048, -0.00000745, 0.00000733, -0.00001892, 0.00004345, -0.00000087, -0.00004108, 0.00002577, -0.00005567, 0.00009029, -0.00003970, -0.00008296, 0.00017142, -0.00014035, 0.00001544, 0.00011885, -0.00011474, 0.00006756, -0.00004233, 0.00001007, -0.00001755, -0.00004616, 0.00007412, -0.00010468, 0.00006788, -0.00011086, -0.00002956, 0.00001734, -0.00056825, 0.00046975, 0.00091721, -0.00027353, -0.00024659, 0.00004672, -0.00007536, 0.00000599, -0.00003176, 0.00000536, 0.00001570, -0.00003594, 0.00000787, -0.00002176, },
		{ -0.00001345, -0.00001197, 0.00003411, -0.00004285, -0.00004870, 0.00000768, 0.00001036, 0.00002839, -0.00002406, -0.00004181, 0.00001327, 0.00002086, 0.00000428, -0.00001287, -0.00001014, 0.00000522, -0.00000838, -0.00001942, 0.00003287, 0.00003607, -0.00000720, -0.00002077, -0.00000537, 0.00001275, -0.00000735, 0.00000801, -0.00002124, 0.00003196, 0.00002477, -0.00006046, 0.00003551, -0.00005356, 0.00005182, 0.00003338, -0.00014313, 0.00016038, -0.00005983, -0.00007655, 0.00016924, -0.00009825, 0.00003348, -0.00000124, -0.00003545, 0.00002242, -0.00007460, 0.00007310, -0.00007093, 0.00001941, -0.00004019, -0.00010217, 0.00008654, -0.00048342, 0.00013413, 0.00096279, -0.00001249, -0.00033354, 0.00001970, -0.00004921, -0.00001442, -0.00002114, -0.00000515, 0.00002248, -0.00003274, -0.00000040, },
		{ -0.00000956, -0.00000583, -0.00002033, 0.00003266, -0.00002271, -0.00005621, -0.00000172, 0.00000603, 0.00003303, -0.00000812, -0.00005021, 0.00000304, 0.00002451, 0.00000657, -0.00001087, -0.00000980, -0.00000084, -0.00000142, -0.00001982, 0.00001498, 0.00004327, 0.00000351, -0.00001955, -0.00001240, 0.00001209, -0.00000549, 0.00000718, -0.00001879, 0.00001373, 0.00005017, -0.00006924, 0.00002696, -0.00002853, -0.00000191, 0.00009396, -0.00015817, 0.00009828, 0.00003879, -0.00014476, 0.00017279, -0.00004241, -0.00001759, 0.00004251, -0.00006884, 0.00003695, -0.00007123, 0.00004287, -0.00002301, -0.00002557, 0.00000877, -0.00013245, 0.00008945, -0.00034844, -0.00012451, 0.00087209, 0.00025341, -0.00034416, -0.00004431, -0.00002142, -0.00002666, -0.00001819, -0.00001112, 0.00002555, -0.00002648, },
		{ -0.00001793, -0.00001892, -0.00000002, -0.00002407, 0.00002518, -0.00000334, -0.00005754, -0.00001114, -0.00000005, 0.00003216, 0.00001086, -0.00005279, -0.00001084, 0.00002659, 0.00001052, -0.00001085, -0.00000735, -0.00000553, -0.00000034, -0.00001305, -0.00000201, 0.00004187, 0.00001660, -0.00001583, -0.00001789, 0.00000858, -0.00000241, 0.00000468, -0.00001199, -0.00000721, 0.00006792, -0.00006118, 0.00000005, 0.00000946, -0.00005103, 0.00012193, -0.00012419, 0.00000581, 0.00012281, -0.00016649, 0.00012806, 0.00003648, -0.00006411, 0.00007014, -0.00007694, 0.00002285, -0.00004451, 0.00000121, 0.00001825, -0.00005049, 0.00002816, -0.00012551, 0.00005250, -0.00021089, -0.00028286, 0.00068712, 0.00047039, -0.00027422, -0.00012132, -0.00000848, -0.00002677, -0.00002033, -0.00001596, 0.00002672, },
		{ 0.00002585, -0.00000690, -0.00002763, 0.00000252, -0.00002252, 0.00001348, 0.00001283, -0.00005313, -0.00001973, -0.00000582, 0.00002496, 0.00002957, -0.00004742, -0.00002680, 0.00002519, 0.00001672, -0.00001191, -0.00000547, -0.00000609, -0.00000511, -0.00000337, -0.00001289, 0.00003160, 0.00002907, -0.00000887, -0.00002122, 0.00000270, 0.00000114, 0.00000090, -0.00000278, -0.00002546, 0.00007204, -0.00003550, -0.00003770, 0.00004495, -0.00007773, 0.00010932, -0.00005423, -0.00008644, 0.00016481, -0.00013562, 0.00005058, 0.00011333, -0.00008568, 0.00007104, -0.00005849, -0.00001101, -0.00001074, -0.00003363, 0.00003925, -0.00005097, 0.00002328, -0.00009690, 0.00000342, -0.00010178, -0.00034584, 0.00046039, 0.00060389, -0.00014196, -0.00018334, -0.00001885, -0.00001854, -0.00002216, -0.00002210, },
		{ -0.00002909, 0.00002124, 0.00000643, -0.00003381, 0.00000004, -0.00001584, 0.00000021, 0.00002357, -0.00004356, -0.00002733, -0.00000954, 0.00001267, 0.00004384, -0.00003344, -0.00004192, 0.00001870, 0.00002446, -0.00001194, -0.00000667, -0.00000179, -0.00001296, 0.00000393, -0.00001458, 0.00001511, 0.00003718, 0.00000143, -0.00002165, -0.00000474, 0.00000422, -0.00000316, 0.00000589, -0.00003615, 0.00006045, 0.00000234, -0.00007368, 0.00006299, -0.00007300, 0.00006282, 0.00002692, -0.00014895, 0.00015226, -0.00006459, -0.00003322, 0.00016385, -0.00007135, 0.00004688, -0.00002325, -0.00004908, 0.00001384, -0.00005019, 0.00003728, -0.00003341, 0.00000667, -0.00006349, -0.00003785, -0.00003290, -0.00033694, 0.00023961, 0.00064395, 0.00002011, -0.00020849, -0.00004988, -0.00001061, -0.00001932, },
		{ -0.00001142, -0.00003381, 0.00001145, 0.00002020, -0.00003461, -0.00000852, -0.00000583, -0.00001112, 0.00002707, -0.00002976, -0.00003381, -0.00001071, -0.00000167, 0.00004993, -0.00001233, -0.00005253, 0.00000662, 0.00003168, -0.00000861, -0.00001199, 0.00000554, -0.00001944, 0.00000462, -0.00000736, -0.00000255, 0.00003771, 0.00001388, -0.00001843, -0.00001257, 0.00000569, -0.00000616, 0.00001119, -0.00003673, 0.00003600, 0.00004213, -0.00009484, 0.00005498, -0.00003988, 0.00000086, 0.00009201, -0.00016362, 0.00009162, 0.00002061, -0.00009609, 0.00017362, -0.00002376, 0.00001022, 0.00001347, -0.00007647, 0.00001959, -0.00004696, 0.00001903, -0.00000960, -0.00000913, -0.00003686, -0.00006248, -0.00000096, -0.00028621, 0.00005752, 0.00060250, 0.00017663, -0.00018779, -0.00008973, -0.00001151, },
		{ -0.00002506, -0.00000248, -0.00003268, -0.00000273, 0.00003056, -0.00002742, -0.00002222, 0.00000384, -0.00001681, 0.00002270, -0.00001352, -0.00003842, -0.00001035, -0.00001417, 0.00004596, 0.00001214, -0.00005521, -0.00000983, 0.00003554, -0.00000048, -0.00002030, 0.00001216, -0.00002031, -0.00000267, 0.00000507, -0.00001559, 0.00002941, 0.00002597, -0.00001111, -0.00001922, 0.00000451, -0.00000699, 0.00001149, -0.00002797, 0.00000561, 0.00007249, -0.00009267, 0.00002204, 0.00000803, -0.00005371, 0.00012115, -0.00012951, 0.00000521, 0.00009036, -0.00012001, 0.00014289, 0.00004242, -0.00002134, 0.00003805, -0.00008497, 0.00000616, -0.00003082, -0.00000436, 0.00000942, -0.00001658, -0.00002096, -0.00007096, 0.00000520, -0.00022036, -0.00007130, 0.00050498, 0.00029957, -0.00012638, -0.00012291, },
		{ -0.00013670, -0.00004844, 0.00000139, -0.00002445, -0.00001747, 0.00003309, -0.00001177, -0.00003734, 0.00000833, -0.00001430, 0.00001184, 0.00000219, -0.00003967, -0.00001039, -0.00002160, 0.00003274, 0.00003479, -0.00004801, -0.00002767, 0.00003336, 0.00001203, -0.00002845, 0.00001394, -0.00001344, -0.00001581, 0.00001687, -0.00001953, 0.00001382, 0.00003433, -0.00000004, -0.00002291, 0.00000011, -0.00000541, 0.00000709, -0.00001366, -0.00002201, 0.00008482, -0.00006647, -0.00002538, 0.00005191, -0.00008221, 0.00010840, -0.00006212, -0.00007745, 0.00012226, -0.00010220, 0.00008548, 0.00010670, -0.00003268, 0.00004423, -0.00007528, -0.00001920, -0.00001232, -0.00002297, 0.00001771, -0.00001468, -0.00001371, -0.00006858, -0.00000225, -0.00015748, -0.00014748, 0.00038041, 0.00037396, -0.00003977, },
		{ 0.00005278, -0.00012562, -0.00007366, -0.00000437, -0.00001189, -0.00002708, 0.00002521, 0.00000926, -0.00004796, 0.00000346, -0.00000368, -0.00000174, 0.00001362, -0.00003600, -0.00001251, -0.00002263, 0.00001373, 0.00005036, -0.00003133, -0.00004280, 0.00002384, 0.00002638, -0.00003239, 0.00000819, 0.00000006, -0.00002980, 0.00002219, -0.00001288, -0.00000490, 0.00003585, 0.00001326, -0.00002194, -0.00000712, -0.00000231, 0.00000005, 0.00000079, -0.00003947, 0.00007635, -0.00002361, -0.00007146, 0.00007503, -0.00007670, 0.00006289, 0.00001327, -0.00013052, 0.00010875, -0.00005503, 0.00002239, 0.00015079, -0.00001702, 0.00003500, -0.00005504, -0.00004619, -0.00000045, -0.00003184, 0.00001537, -0.00000713, -0.00001053, -0.00006101, -0.00001426, -0.00010649, -0.00018087, 0.00025356, 0.00039836, },
		{ 0.00038107, 0.00013466, -0.00009236, -0.00009149, -0.00001979, -0.00000122, -0.00002683, 0.00000830, 0.00002918, -0.00004798, -0.00001209, 0.00001140, -0.00001255, 0.00001742, -0.00002689, -0.00001702, -0.00001840, -0.00000597, 0.00005519, -0.00000816, -0.00005113, 0.00000778, 0.00003854, -0.00002881, -0.00000498, 0.00001598, -0.00003869, 0.00001747, 0.00000207, -0.00002105, 0.00002889, 0.00002610, -0.00001536, -0.00001564, 0.00000043, -0.00000663, 0.00001043, -0.00004316, 0.00005098, 0.00002306, -0.00010089, 0.00006883, -0.00004224, 0.00000432, 0.00007137, -0.00014059, 0.00005869, -0.00000006, -0.00002674, 0.00016499, 0.00002215, 0.00002035, -0.00003418, -0.00006650, 0.00000073, -0.00003168, 0.00000661, 0.00000101, -0.00000770, -0.00005184, -0.00002588, -0.00006958, -0.00018435, 0.00014094, },
		{ 0.00005074, 0.00033488, 0.00019563, -0.00004520, -0.00009574, -0.00003968, 0.00000079, -0.00001594, -0.00001164, 0.00004011, -0.00003405, -0.00003515, 0.00002432, -0.00001515, 0.00001226, -0.00001380, -0.00002245, -0.00001187, -0.00002129, 0.00004840, 0.00001655, -0.00004991, -0.00001175, 0.00004427, -0.00001667, -0.00002229, 0.00002846, -0.00003787, 0.00000292, 0.00001993, -0.00002932, 0.00001418, 0.00003514, -0.00000345, -0.00002290, 0.00000065, -0.00001037, 0.00001255, -0.00003431, 0.00001765, 0.00005994, -0.00010442, 0.00003572, 0.00000590, -0.00004509, 0.00009579, -0.00011025, -0.00000726, 0.00004097, -0.00005083, 0.00015052, 0.00007341, 0.00001232, -0.00001998, -0.00007684, -0.00000741, -0.00002684, -0.00000351, 0.00000622, -0.00000389, -0.00004231, -0.00003509, -0.00004542, -0.00016975, },
		{ -0.00014617, -0.00001497, 0.00027254, 0.00023220, 0.00000622, -0.00008585, -0.00005589, -0.00000945, 0.00000088, -0.00002586, 0.00003627, -0.00000791, -0.00005825, 0.00002764, -0.00000650, -0.00000001, -0.00000025, -0.00002604, -0.00000660, -0.00002891, 0.00003223, 0.00003722, -0.00003870, -0.00003018, 0.00004074, 0.00000203, -0.00003824, 0.00003217, -0.00002609, -0.00001739, 0.00003397, -0.00002661, -0.00000498, 0.00003731, 0.00001190, -0.00002597, -0.00000306, -0.00001013, 0.00000754, -0.00001819, -0.00001298, 0.00007736, -0.00008185, -0.00001235, 0.00004865, -0.00006897, 0.00008395, -0.00005546, -0.00006528, 0.00005436, -0.00005034, 0.00011713, 0.00012265, 0.00001983, -0.00001409, -0.00007916, -0.00002030, -0.00002214, -0.00001170, 0.00000764, 0.00000020, -0.00003252, -0.00004117, -0.00003155, },
		{ -0.00002532, -0.00012013, -0.00005770, 0.00020450, 0.00024556, 0.00005458, -0.00006632, -0.00006128, -0.00002960, 0.00001459, -0.00002654, 0.00001723, 0.00002302, -0.00007185, 0.00001620, 0.00001210, -0.00001428, 0.00000908, -0.00002509, -0.00000508, -0.00002826, 0.00001133, 0.00004921, -0.00001969, -0.00004274, 0.00002773, 0.00002260, -0.00004707, 0.00002452, -0.00000613, -0.00003690, 0.00003854, -0.00001328, -0.00002353, 0.00003095, 0.00002741, -0.00002250, -0.00001054, -0.00000688, -0.00000157, -0.00000186, -0.00003238, 0.00007269, -0.00004166, -0.00005889, 0.00007045, -0.00006214, 0.00004639, 0.00000174, -0.00009744, 0.00003883, -0.00003523, 0.00007773, 0.00015842, 0.00004556, -0.00001247, -0.00007835, -0.00003324, -0.00002035, -0.00001737, 0.00000641, 0.00000327, -0.00002273, -0.00004372, },
		{ -0.00004261, -0.00002396, -0.00009611, -0.00008083, 0.00013869, 0.00023882, 0.00009635, -0.00004376, -0.00005360, -0.00005138, 0.00001555, -0.00001095, -0.00001071, 0.00004770, -0.00006805, -0.00001006, 0.00003461, -0.00002356, 0.00001028, -0.00001851, -0.00000764, -0.00002149, -0.00000868, 0.00005019, 0.00000282, -0.00004598, 0.00000802, 0.00003913, -0.00004488, 0.00000681, 0.00001600, -0.00004888, 0.00003113, 0.00000679, -0.00003612, 0.00001655, 0.00003908, -0.00001171, -0.00001973, -0.00000318, -0.00001056, 0.00000858, -0.00003685, 0.00005061, 0.00000264, -0.00008906, 0.00006468, -0.00003133, 0.00000127, 0.00004194, -0.00009749, 0.00000445, -0.00001919, 0.00004259, 0.00017519, 0.00008583, -0.00000783, -0.00007897, -0.00004387, -0.00002149, -0.00002187, 0.00000415, 0.00000473, -0.00001377, },
		{ -0.00000662, -0.00003843, -0.00002464, -0.00007681, -0.00008894, 0.00008115, 0.00021527, 0.00013074, -0.00002296, -0.00003705, -0.00006375, -0.00000162, 0.00001608, -0.00003599, 0.00005563, -0.00004441, -0.00004492, 0.00005180, -0.00002165, 0.00000213, -0.00000774, -0.00001239, -0.00001247, -0.00002302, 0.00004074, 0.00002362, -0.00003888, -0.00001334, 0.00004647, -0.00003131, -0.00001595, 0.00003313, -0.00004896, 0.00001329, 0.00002742, -0.00003883, -0.00000304, 0.00004331, 0.00000500, -0.00002716, -0.00000196, -0.00001587, 0.00001008, -0.00002827, 0.00002081, 0.00003762, -0.00009487, 0.00003522, 0.00000868, -0.00003333, 0.00005512, -0.00007163, -0.00003275, -0.00001309, 0.00001588, 0.00017345, 0.00013294, 0.00000699, -0.00008271, -0.00005285, -0.00002367, -0.00002678, 0.00000160, 0.00000499, },
		},
		{ { -0.00001601, -0.00001098, -0.00000741, -0.00000362, -0.00001643, -0.00001387, -0.00002086, -0.00007438, -0.00003424, 0.00003500, 0.00010516, 0.00015433, 0.00001484, -0.00001598, 0.00000898, -0.00010334, 0.00001518, 0.00005851, -0.00007171, 0.00001747, -0.00000272, -0.00004299, 0.00004370, -0.00001003, 0.00002654, 0.00001243, -0.00010770, 0.00004338, 0.00009269, -0.00005577, -0.00000738, 0.00001748, -0.00004602, 0.00001748, -0.00000738, -0.00005577, 0.00009269, 0.00004338, -0.00010770, 0.00001243, 0.00002654, -0.00001003, 0.00004370, -0.00004299, -0.00000272, 0.00001747, -0.00007171, 0.00005851, 0.00001518, -0.00010334, 0.00000898, -0.00001598, 0.00001484, 0.00015433, 0.00010516, 0.00003500, -0.00003424, -0.00007438, -0.00002086, -0.00001387, -0.00001643, -0.00000362, -0.00000741, -0.00001098, },
		{ -0.00001052, -0.00001687, -0.00001057, -0.00000861, -0.00000566, -0.00001286, -0.00001516, -0.00000861, -0.00006604, -0.00004663, 0.00002385, 0.00006009, 0.00015100, 0.00005591, -0.00003795, 0.00004323, -0.00007952, -0.00005475, 0.00009246, -0.00004904, -0.00002215, 0.00003316, -0.00005869, 0.00002788, 0.00001542, -0.00000496, 0.00004490, -0.00008972, -0.00002626, 0.00011823, -0.00001685, -0.00004242, 0.00003012, -0.00003544, -0.00000963, 0.00002169, -0.00006180, 0.00003916, 0.00009495, -0.00008410, -0.00003909, 0.00004951, -0.00001391, 0.00003366, -0.00001045, -0.00003412, 0.00003242, -0.00005801, 0.00000521, 0.00005923, -0.00008202, -0.00003664, -0.00000746, -0.00001155, 0.00013414, 0.00014931, 0.00006180, -0.00002272, -0.00007906, -0.00003399, -0.00001423, -0.00002054, -0.00000399, -0.00000421, },
		{ -0.00000101, -0.00000762, -0.00001842, -0.00001167, -0.00000705, -0.00000869, -0.00001022, -0.00001768, -0.00000015, -0.00005075, -0.00006195, 0.00002060, 0.00002861, 0.00011972, 0.00010497, -0.00005251, 0.00004132, -0.00001559, -0.00011591, 0.00007996, 0.00000839, -0.00006518, 0.00004923, -0.00004377, -0.00001010, 0.00004433, -0.00002172, 0.00004493, -0.00004246, -0.00008382, 0.00010109, 0.00004107, -0.00006223, 0.00001949, -0.00001033, -0.00003442, 0.00002963, -0.00003916, -0.00001581, 0.00010785, -0.00002809, -0.00008353, 0.00004517, 0.00000559, 0.00001115, 0.00001760, -0.00004399, 0.00002077, -0.00002621, -0.00003735, 0.00006349, -0.00003498, -0.00006637, -0.00001817, -0.00002875, 0.00009944, 0.00018036, 0.00010440, -0.00000532, -0.00008271, -0.00004745, -0.00001561, -0.00002461, -0.00000730, },
		{ -0.00001304, 0.00000001, -0.00000264, -0.00001804, -0.00001588, -0.00000384, -0.00001062, -0.00000914, -0.00001953, 0.00000117, -0.00002858, -0.00007538, 0.00001324, 0.00002013, 0.00006820, 0.00014244, -0.00003704, -0.00000101, 0.00005706, -0.00013088, 0.00001640, 0.00007452, -0.00008189, 0.00003043, -0.00000334, -0.00004987, 0.00005414, -0.00001129, 0.00001933, 0.00000785, -0.00010582, 0.00004715, 0.00009233, -0.00005205, -0.00000908, 0.00001381, -0.00004294, 0.00001526, -0.00000588, -0.00004857, 0.00008146, 0.00003476, -0.00009754, 0.00001162, 0.00003353, -0.00000412, 0.00002708, -0.00003182, -0.00000323, 0.00000093, -0.00005329, 0.00003746, 0.00000962, -0.00006719, -0.00003791, -0.00004515, 0.00005660, 0.00019117, 0.00015671, 0.00002532, -0.00008460, -0.00006266, -0.00001759, -0.00002777, },
		{ -0.00002928, -0.00001992, -0.00000211, 0.00000188, -0.00001315, -0.00002251, -0.00000196, -0.00000925, -0.00001036, -0.00001868, -0.00000517, -0.00000519, -0.00007789, -0.00000698, 0.00003180, 0.00001661, 0.00014652, 0.00001537, -0.00006078, 0.00009645, -0.00008330, -0.00007189, 0.00011074, -0.00005352, -0.00001649, 0.00003954, -0.00006691, 0.00003341, 0.00001958, -0.00001025, 0.00003663, -0.00008770, -0.00002013, 0.00011316, -0.00001273, -0.00003886, 0.00002254, -0.00003237, -0.00000976, 0.00001793, -0.00005036, 0.00003378, 0.00007705, -0.00007413, -0.00003499, 0.00004892, -0.00000020, 0.00001916, -0.00001060, -0.00002180, 0.00001074, -0.00004570, 0.00000272, 0.00003241, -0.00004509, -0.00005076, -0.00006264, 0.00000812, 0.00017798, 0.00020907, 0.00007415, -0.00008022, -0.00008133, -0.00002080, },
		{ -0.00002698, -0.00002872, -0.00002693, -0.00000625, 0.00000256, -0.00000359, -0.00002784, -0.00000518, -0.00000351, -0.00001344, -0.00001529, -0.00001504, 0.00001028, -0.00006313, -0.00003834, 0.00004809, -0.00001199, 0.00010872, 0.00008702, -0.00009909, 0.00007544, 0.00000814, -0.00013848, 0.00008910, 0.00001233, -0.00006406, 0.00005761, -0.00004898, -0.00000955, 0.00004950, -0.00002185, 0.00003485, -0.00004481, -0.00007250, 0.00009441, 0.00003919, -0.00005159, 0.00001163, -0.00001171, -0.00002968, 0.00002154, -0.00002965, -0.00001018, 0.00008428, -0.00002605, -0.00007070, 0.00003892, 0.00001952, 0.00000698, 0.00000461, -0.00002586, 0.00000536, -0.00002891, -0.00002255, 0.00003106, -0.00001719, -0.00004764, -0.00007462, -0.00004287, 0.00013853, 0.00024955, 0.00014210, -0.00006207, -0.00010399, },
		{ -0.00012865, -0.00003914, -0.00002554, -0.00003425, -0.00001004, -0.00000214, 0.00000683, -0.00002632, -0.00001551, 0.00000463, -0.00001553, -0.00001249, -0.00002169, 0.00001127, -0.00003389, -0.00006796, 0.00004954, -0.00000586, 0.00004299, 0.00014292, -0.00008483, 0.00000083, 0.00009630, -0.00014280, 0.00001240, 0.00008201, -0.00008000, 0.00003765, -0.00000526, -0.00005109, 0.00005665, -0.00000727, 0.00001249, -0.00000200, -0.00009174, 0.00004581, 0.00008051, -0.00003837, -0.00001068, 0.00000516, -0.00003481, 0.00000792, -0.00000436, -0.00003256, 0.00006104, 0.00002271, -0.00007839, 0.00000692, 0.00003982, 0.00000447, 0.00000754, -0.00001880, -0.00000439, -0.00001596, -0.00003264, 0.00001625, 0.00000150, -0.00003189, -0.00007250, -0.00008792, 0.00007369, 0.00026405, 0.00022442, -0.00002055, },
		{ 0.00005406, -0.00014858, -0.00006200, -0.00001891, -0.00004240, -0.00001252, -0.00000949, 0.00001125, -0.00001431, -0.00003013, 0.00000945, -0.00001212, -0.00001403, -0.00002075, -0.00000067, -0.00000273, -0.00007951, 0.00002575, 0.00002645, -0.00001940, 0.00015137, -0.00001596, -0.00008732, 0.00013117, -0.00007594, -0.00008280, 0.00011388, -0.00004859, -0.00000930, 0.00003847, -0.00006722, 0.00003335, 0.00002325, -0.00000973, 0.00002118, -0.00007777, -0.00001039, 0.00009400, -0.00000462, -0.00002945, 0.00000893, -0.00002637, -0.00001087, 0.00001000, -0.00003086, 0.00002569, 0.00005130, -0.00005723, -0.00003038, 0.00004497, 0.00001601, 0.00000345, -0.00001057, -0.00000934, -0.00001078, -0.00003275, 0.00000066, 0.00000665, -0.00001485, -0.00005431, -0.00011440, -0.00000795, 0.00023862, 0.00030792, },
		{ 0.00036947, 0.00016644, -0.00014957, -0.00010114, -0.00000967, -0.00004963, -0.00001663, -0.00001317, 0.00000421, 0.00000559, -0.00004064, 0.00000391, -0.00000066, -0.00001999, -0.00001442, -0.00001575, 0.00001507, -0.00006491, -0.00001621, 0.00006002, -0.00004766, 0.00010474, 0.00007630, -0.00013751, 0.00008986, 0.00003043, -0.00014460, 0.00008404, 0.00001684, -0.00005352, 0.00005615, -0.00004903, -0.00000883, 0.00004867, -0.00001454, 0.00002050, -0.00004578, -0.00005209, 0.00007682, 0.00003374, -0.00003271, 0.00000036, -0.00001365, -0.00002308, 0.00000866, -0.00001585, -0.00000209, 0.00005287, -0.00002155, -0.00005425, 0.00002946, 0.00003396, 0.00000300, -0.00000829, -0.00000836, -0.00000913, -0.00003069, -0.00000972, 0.00000366, -0.00000618, -0.00002818, -0.00011280, -0.00008782, 0.00016603, },
		{ 0.00005426, 0.00037994, 0.00030798, -0.00010961, -0.00015833, -0.00000431, -0.00004966, -0.00002837, -0.00000799, -0.00001280, 0.00002310, -0.00003670, -0.00001492, 0.00001517, -0.00002453, -0.00001046, -0.00002243, 0.00001200, -0.00003160, -0.00005516, 0.00006902, -0.00003067, 0.00002614, 0.00014438, -0.00011747, -0.00000785, 0.00011961, -0.00013597, 0.00000348, 0.00007956, -0.00006523, 0.00003847, -0.00000834, -0.00004663, 0.00005130, 0.00000120, 0.00000666, -0.00001528, -0.00006822, 0.00004018, 0.00005992, -0.00001851, -0.00001149, -0.00000613, -0.00002461, -0.00000238, -0.00000223, -0.00001240, 0.00003586, 0.00000946, -0.00005543, 0.00000103, 0.00004507, 0.00001290, -0.00001076, -0.00000694, -0.00000557, -0.00002955, -0.00001661, 0.00000013, -0.00000696, -0.00000728, -0.00008477, -0.00014218, },
		{ -0.00015355, -0.00006975, 0.00031614, 0.00044968, -0.00000545, -0.00022326, -0.00001770, -0.00003387, -0.00005119, 0.00000363, -0.00002909, 0.00002475, -0.00001365, -0.00004060, 0.00002473, -0.00001891, -0.00001578, -0.00001636, -0.00000567, 0.00000082, -0.00006962, 0.00004395, 0.00001564, -0.00004337, 0.00015254, -0.00003427, -0.00010916, 0.00014184, -0.00005813, -0.00008553, 0.00010108, -0.00003442, -0.00000097, 0.00003032, -0.00006001, 0.00002816, 0.00002596, -0.00000420, 0.00000148, -0.00006191, 0.00000123, 0.00006535, 0.00000490, -0.00001649, -0.00000673, -0.00001978, -0.00001308, 0.00000118, -0.00000832, 0.00001598, 0.00002348, -0.00003840, -0.00002542, 0.00004034, 0.00003042, -0.00001048, -0.00001037, 0.00000047, -0.00002674, -0.00002396, -0.00000130, -0.00001080, -0.00000042, -0.00004483, },
		{ -0.00001278, -0.00012206, -0.00016717, 0.00017663, 0.00054383, 0.00017286, -0.00026652, -0.00006903, 0.00000042, -0.00007855, 0.00000918, -0.00002979, 0.00000344, 0.00002073, -0.00005757, 0.00001610, 0.00000094, -0.00002895, -0.00000476, -0.00002162, 0.00001344, -0.00005306, -0.00000162, 0.00005850, -0.00006951, 0.00009850, 0.00006901, -0.00015831, 0.00008372, 0.00004788, -0.00013314, 0.00006556, 0.00002150, -0.00003471, 0.00004580, -0.00004439, -0.00000798, 0.00004255, -0.00000190, 0.00000451, -0.00004502, -0.00002683, 0.00005274, 0.00002546, -0.00001061, -0.00001076, -0.00001598, -0.00001732, -0.00000469, -0.00000062, 0.00000536, 0.00002081, -0.00001656, -0.00003801, 0.00002105, 0.00004568, -0.00000078, -0.00001812, 0.00000429, -0.00001871, -0.00003205, -0.00000377, -0.00001084, -0.00000590, },
		{ -0.00001368, -0.00000138, -0.00006802, -0.00020468, -0.00000794, 0.00053943, 0.00040478, -0.00024199, -0.00016848, 0.00003964, -0.00009152, -0.00000786, -0.00000725, -0.00003284, 0.00004649, -0.00004931, -0.00001415, 0.00002771, -0.00003846, -0.00000138, -0.00002150, 0.00000158, -0.00001933, -0.00004018, 0.00006981, -0.00004542, 0.00001441, 0.00013864, -0.00012890, -0.00002322, 0.00012393, -0.00011165, -0.00000887, 0.00006796, -0.00003985, 0.00003369, -0.00001209, -0.00003767, 0.00003984, 0.00001203, 0.00000249, -0.00002892, -0.00004025, 0.00003192, 0.00003510, 0.00000193, -0.00001076, -0.00001688, -0.00001622, -0.00001268, 0.00000126, 0.00000606, 0.00001132, -0.00000241, -0.00003475, -0.00000289, 0.00004904, 0.00001814, -0.00002369, 0.00000081, -0.00000614, -0.00003650, -0.00001135, -0.00000643, },
		{ -0.00000345, -0.00001427, -0.00000829, -0.00002162, -0.00017340, -0.00018025, 0.00040817, 0.00063126, -0.00010465, -0.00029913, 0.00005043, -0.00006750, -0.00005418, 0.00002855, -0.00006105, 0.00004203, -0.00001115, -0.00005379, 0.00004356, -0.00003001, -0.00001503, -0.00000465, -0.00002120, 0.00000647, -0.00004946, 0.00004440, 0.00000567, -0.00005355, 0.00014177, -0.00003870, -0.00012070, 0.00012842, -0.00003287, -0.00008001, 0.00007515, -0.00001344, 0.00000778, 0.00001712, -0.00004728, 0.00001939, 0.00002720, 0.00000440, -0.00001805, -0.00004333, 0.00001237, 0.00003389, 0.00001264, -0.00000311, -0.00001968, -0.00001529, -0.00001640, -0.00000479, 0.00001143, 0.00000592, -0.00000007, -0.00002305, -0.00002025, 0.00003786, 0.00003884, -0.00001982, -0.00001006, 0.00000518, -0.00003253, -0.00002332, },
		{ -0.00003340, -0.00000828, -0.00000700, -0.00001864, -0.00000502, -0.00009688, -0.00027930, 0.00016807, 0.00076720, 0.00016005, -0.00040664, -0.00000830, 0.00000110, -0.00011388, 0.00004984, -0.00005559, 0.00000018, 0.00004106, -0.00007788, 0.00003080, 0.00000099, -0.00003984, 0.00001399, -0.00003267, 0.00000590, -0.00002840, 0.00000243, 0.00004718, -0.00007351, 0.00008448, 0.00006214, -0.00015756, 0.00006035, 0.00005829, -0.00010692, 0.00003751, 0.00002569, -0.00001075, 0.00002952, -0.00003642, -0.00000706, 0.00003281, 0.00001266, -0.00000977, -0.00004220, -0.00000220, 0.00002785, 0.00001554, 0.00000880, -0.00001804, -0.00001844, -0.00001513, -0.00001400, 0.00001290, 0.00000909, -0.00000459, -0.00001308, -0.00002568, 0.00001773, 0.00005176, -0.00000439, -0.00002235, 0.00000864, -0.00002021, },
		{ -0.00000566, -0.00003455, -0.00002083, -0.00000074, -0.00001719, -0.00001850, -0.00002045, -0.00027143, -0.00011175, 0.00073555, 0.00050986, -0.00040984, -0.00015808, 0.00008895, -0.00014862, 0.00002622, -0.00000786, -0.00006087, 0.00007526, -0.00006333, -0.00001327, 0.00004128, -0.00005629, 0.00001447, -0.00001899, -0.00001875, 0.00000200, -0.00002588, 0.00005406, -0.00004588, 0.00000520, 0.00012177, -0.00011868, -0.00004086, 0.00011044, -0.00007501, -0.00002232, 0.00004992, -0.00000867, 0.00002531, -0.00001578, -0.00002631, 0.00002518, 0.00002244, 0.00000047, -0.00003937, -0.00001382, 0.00002304, 0.00001143, 0.00001711, -0.00000811, -0.00002396, -0.00001316, -0.00002008, 0.00000660, 0.00001731, -0.00000741, -0.00001057, -0.00002184, -0.00000213, 0.00005154, 0.00001775, -0.00002800, 0.00000220, },
		{ -0.00000970, 0.00000322, -0.00002537, -0.00003308, -0.00000597, -0.00000105, -0.00004074, 0.00001407, -0.00017133, -0.00033397, 0.00050987, 0.00084211, -0.00023623, -0.00037104, 0.00013636, -0.00011717, -0.00005115, 0.00006008, -0.00010246, 0.00006148, -0.00000720, -0.00006848, 0.00006415, -0.00004433, -0.00001158, 0.00001320, -0.00004520, 0.00001399, -0.00002259, 0.00002854, 0.00000053, -0.00005109, 0.00011644, -0.00003133, -0.00011963, 0.00009597, -0.00000488, -0.00006779, 0.00004175, 0.00001022, 0.00001605, 0.00000206, -0.00003205, 0.00000927, 0.00002655, 0.00001359, -0.00003269, -0.00002563, 0.00002057, 0.00000670, 0.00001578, 0.00000737, -0.00002582, -0.00001498, -0.00002062, -0.00000490, 0.00002331, -0.00000323, -0.00001403, -0.00001568, -0.00001504, 0.00003970, 0.00003818, -0.00002256, },
		{ -0.00000779, -0.00001949, 0.00000275, -0.00001205, -0.00003494, -0.00002459, 0.00001689, -0.00004442, -0.00000648, -0.00003828, -0.00041706, 0.00014171, 0.00102461, 0.00013035, -0.00055896, 0.00007431, -0.00000744, -0.00015228, 0.00010102, -0.00008769, -0.00000521, 0.00006424, -0.00009939, 0.00004640, -0.00000050, -0.00005110, 0.00003951, -0.00004726, -0.00000583, 0.00000474, -0.00000380, 0.00003057, -0.00006064, 0.00006137, 0.00005382, -0.00013692, 0.00002639, 0.00006082, -0.00007178, 0.00000584, 0.00002869, 0.00001403, 0.00001145, -0.00002705, -0.00000612, 0.00002163, 0.00002533, -0.00001905, -0.00003719, 0.00001636, 0.00000772, 0.00000555, 0.00002022, -0.00001886, -0.00002061, -0.00001825, -0.00001589, 0.00002205, 0.00000691, -0.00001767, -0.00001265, -0.00002004, 0.00002222, 0.00005028, },
		{ 0.00005244, 0.00001049, -0.00002208, -0.00000379, -0.00000393, -0.00002306, -0.00004587, 0.00001746, -0.00001703, -0.00005813, 0.00005416, -0.00033815, -0.00024806, 0.00095392, 0.00061624, -0.00059783, -0.00013050, 0.00013871, -0.00021449, 0.00006948, -0.00000963, -0.00009345, 0.00010577, -0.00007711, -0.00001233, 0.00005413, -0.00007537, 0.00003503, -0.00001523, -0.00004511, 0.00002857, -0.00001444, 0.00002723, -0.00003265, -0.00000312, 0.00009420, -0.00009137, -0.00005643, 0.00008381, -0.00003372, -0.00003413, 0.00002933, 0.00002216, 0.00001597, -0.00001858, -0.00001501, 0.00001064, 0.00002963, 0.00000086, -0.00004347, 0.00000560, 0.00001535, -0.00000616, 0.00002257, -0.00000382, -0.00002517, -0.00001737, -0.00002256, 0.00001371, 0.00001773, -0.00001664, -0.00001348, -0.00002024, 0.00000482, },
		{ -0.00001006, 0.00004672, 0.00002719, -0.00001789, -0.00000933, -0.00000575, -0.00000502, -0.00005310, -0.00000885, 0.00002791, -0.00009516, 0.00005904, -0.00014929, -0.00051709, 0.00061421, 0.00106497, -0.00038594, -0.00043436, 0.00022984, -0.00017612, -0.00004313, 0.00009416, -0.00014773, 0.00008042, -0.00000244, -0.00008260, 0.00008360, -0.00005956, -0.00000562, 0.00003428, -0.00007310, 0.00002212, 0.00000592, 0.00000173, 0.00000160, -0.00003950, 0.00007911, -0.00001638, -0.00010681, 0.00005310, 0.00002057, -0.00005157, 0.00000789, 0.00003175, 0.00002290, -0.00001137, -0.00001762, 0.00000011, 0.00002378, 0.00002095, -0.00003855, -0.00001194, 0.00002381, -0.00001050, 0.00001281, 0.00001223, -0.00002299, -0.00001954, -0.00002529, 0.00000210, 0.00002431, -0.00001038, -0.00001536, -0.00001870, },
		{ -0.00001588, -0.00002224, 0.00003519, 0.00004074, -0.00001085, -0.00000944, -0.00001332, 0.00000538, -0.00003660, -0.00005145, 0.00005831, -0.00008085, -0.00001639, 0.00004609, -0.00056833, 0.00010248, 0.00129298, 0.00008947, -0.00070905, 0.00016690, -0.00002276, -0.00018912, 0.00015540, -0.00012141, -0.00001203, 0.00008677, -0.00011888, 0.00006056, -0.00000321, -0.00006043, 0.00006754, -0.00006250, -0.00001925, 0.00003967, -0.00001729, 0.00001331, -0.00003598, 0.00003151, 0.00004334, -0.00010238, -0.00000985, 0.00005601, -0.00003505, -0.00002286, 0.00002980, 0.00003505, -0.00000406, -0.00001821, -0.00000524, 0.00001115, 0.00003287, -0.00002098, -0.00003004, 0.00002490, -0.00000372, -0.00000282, 0.00002045, -0.00001267, -0.00002190, -0.00002679, -0.00000927, 0.00002522, -0.00000169, -0.00001578, },
		{ -0.00001497, -0.00001075, -0.00003068, 0.00001783, 0.00005161, -0.00000336, -0.00000631, -0.00001672, -0.00000113, -0.00000316, -0.00008383, 0.00004583, -0.00001381, -0.00011489, 0.00014677, -0.00040073, -0.00040092, 0.00117438, 0.00071567, -0.00078509, -0.00009278, 0.00018352, -0.00027974, 0.00011507, -0.00001210, -0.00012573, 0.00013321, -0.00008850, -0.00001141, 0.00006409, -0.00009198, 0.00005694, -0.00001079, -0.00007205, 0.00005479, -0.00000705, -0.00000327, -0.00001009, -0.00001127, 0.00005973, -0.00005486, -0.00006639, 0.00005083, 0.00000395, -0.00004180, 0.00001022, 0.00004661, 0.00000823, -0.00001974, -0.00000596, -0.00000088, 0.00003155, 0.00000359, -0.00003934, 0.00001438, 0.00000994, -0.00001431, 0.00001659, 0.00000092, -0.00001985, -0.00002833, -0.00001962, 0.00002179, 0.00000667, },
		{ 0.00001434, -0.00001478, -0.00000404, -0.00003187, -0.00000561, 0.00005824, 0.00000682, -0.00000641, -0.00000902, -0.00002071, 0.00002610, -0.00008090, -0.00001187, 0.00006814, -0.00016438, 0.00010857, -0.00011338, -0.00070925, 0.00071277, 0.00127804, -0.00053468, -0.00048536, 0.00031808, -0.00023473, -0.00003212, 0.00012563, -0.00018976, 0.00009564, 0.00000231, -0.00009364, 0.00009847, -0.00007272, 0.00000231, 0.00005469, -0.00009903, 0.00002903, 0.00003066, -0.00002837, 0.00000802, -0.00002361, 0.00003596, 0.00000087, -0.00008555, 0.00000975, 0.00003879, -0.00003445, -0.00001972, 0.00004677, 0.00002754, -0.00002011, -0.00000671, -0.00000626, 0.00001894, 0.00002474, -0.00003369, -0.00000417, 0.00002107, -0.00001477, 0.00000416, 0.00000990, -0.00001177, -0.00002778, -0.00002976, 0.00001513, },
		{ 0.00000425, 0.00002226, -0.00001538, 0.00000007, -0.00002232, -0.00003159, 0.00005494, 0.00002418, -0.00001334, 0.00000645, -0.00003747, 0.00002976, -0.00003804, -0.00008519, 0.00011087, -0.00012347, -0.00003187, 0.00014316, -0.00071758, 0.00005787, 0.00154424, 0.00004537, -0.00084093, 0.00025455, -0.00004084, -0.00021925, 0.00020415, -0.00015118, -0.00001937, 0.00010474, -0.00013281, 0.00007073, -0.00000660, -0.00006539, 0.00009292, -0.00007505, -0.00003164, 0.00006904, -0.00003337, -0.00000064, -0.00000691, -0.00000039, 0.00003106, -0.00006233, -0.00004030, 0.00004552, -0.00000373, -0.00004288, 0.00002852, 0.00004855, -0.00001355, -0.00001138, -0.00000448, 0.00000298, 0.00003342, -0.00001480, -0.00002110, 0.00002187, -0.00000511, -0.00000798, 0.00000923, -0.00000142, -0.00002160, -0.00003892, },
		{ -0.00004299, -0.00001241, 0.00002941, -0.00001330, -0.00000361, -0.00000323, -0.00005089, 0.00003504, 0.00004893, -0.00002198, 0.00001695, -0.00003475, 0.00000229, 0.00002239, -0.00012907, 0.00007983, -0.00000903, -0.00017956, 0.00024458, -0.00045539, -0.00055007, 0.00137371, 0.00079955, -0.00094845, -0.00005412, 0.00021817, -0.00033392, 0.00015503, -0.00001468, -0.00015243, 0.00015260, -0.00009534, -0.00001056, 0.00006915, -0.00010233, 0.00007645, -0.00000637, -0.00009383, 0.00007482, -0.00000396, -0.00002990, 0.00001534, -0.00001911, 0.00002402, -0.00001818, -0.00006859, 0.00001874, 0.00003108, -0.00004362, -0.00000411, 0.00006006, 0.00000399, -0.00001873, -0.00000061, -0.00000759, 0.00002742, 0.00000823, -0.00002694, 0.00001168, 0.00000690, -0.00001199, 0.00000081, 0.00000418, -0.00000919, },
		{ 0.00000445, -0.00003679, -0.00003323, 0.00003097, -0.00000369, -0.00001616, 0.00001690, -0.00005229, -0.00000208, 0.00007209, -0.00001997, 0.00000999, -0.00000882, -0.00003949, 0.00006335, -0.00011201, -0.00001467, 0.00011246, -0.00023793, 0.00015699, -0.00007253, -0.00088665, 0.00079677, 0.00145815, -0.00066174, -0.00052137, 0.00038791, -0.00028250, -0.00002058, 0.00014922, -0.00022135, 0.00010411, 0.00000624, -0.00009929, 0.00010568, -0.00008089, 0.00001146, 0.00007051, -0.00011734, 0.00003286, 0.00004705, -0.00005389, 0.00001740, -0.00000838, -0.00000535, 0.00001536, -0.00006039, -0.00002511, 0.00004679, -0.00001917, -0.00003629, 0.00005227, 0.00002948, -0.00002233, -0.00000086, -0.00000890, 0.00001240, 0.00002430, -0.00001865, -0.00000250, 0.00001256, -0.00000668, -0.00000772, 0.00000039, },
		{ -0.00001055, 0.00001076, -0.00001903, -0.00005118, 0.00002068, 0.00001451, -0.00003115, 0.00002546, -0.00003052, -0.00004572, 0.00007780, 0.00000301, -0.00001500, 0.00002565, -0.00006600, 0.00005722, -0.00003878, -0.00012041, 0.00016559, -0.00016649, -0.00004870, 0.00023690, -0.00084806, 0.00001659, 0.00174958, 0.00000676, -0.00093954, 0.00032219, -0.00005706, -0.00023797, 0.00023874, -0.00017169, -0.00002631, 0.00011467, -0.00013816, 0.00007479, -0.00001004, -0.00006402, 0.00011084, -0.00008202, -0.00004063, 0.00008672, -0.00004709, -0.00000862, 0.00001899, -0.00002888, 0.00001798, -0.00002525, -0.00005889, 0.00003171, 0.00001716, -0.00005076, 0.00002458, 0.00005242, -0.00001522, -0.00000723, -0.00000389, -0.00000215, 0.00002699, -0.00000162, -0.00001097, 0.00000863, 0.00000190, -0.00000873, },
		{ -0.00000076, -0.00002002, 0.00000366, 0.00000403, -0.00005634, -0.00000353, 0.00003483, -0.00003698, 0.00001388, 0.00000722, -0.00007617, 0.00005325, 0.00004436, -0.00004242, 0.00004388, -0.00005502, 0.00000564, 0.00004974, -0.00017406, 0.00011367, -0.00000471, -0.00024089, 0.00033316, -0.00049895, -0.00067435, 0.00152920, 0.00086015, -0.00106628, -0.00002379, 0.00023850, -0.00036755, 0.00018209, -0.00001681, -0.00016898, 0.00015991, -0.00009593, -0.00000984, 0.00006788, -0.00010346, 0.00009024, -0.00000262, -0.00010591, 0.00008410, -0.00000455, -0.00004673, 0.00003708, -0.00002605, -0.00000718, 0.00001082, -0.00006261, -0.00000646, 0.00004370, -0.00003904, -0.00001193, 0.00006044, 0.00000403, -0.00001540, 0.00000071, -0.00000917, 0.00001796, 0.00001404, -0.00000818, -0.00000031, 0.00000535, },
		{ -0.00000029, 0.00000979, -0.00001882, -0.00001472, 0.00002021, -0.00004257, -0.00003496, 0.00004501, -0.00002459, -0.00001480, 0.00004213, -0.00007511, 0.00000018, 0.00008515, -0.00004883, 0.00002854, -0.00000879, -0.00005920, 0.00010079, -0.00014100, -0.00001684, 0.00015253, -0.00030406, 0.00019852, -0.00003646, -0.00102596, 0.00085820, 0.00158438, -0.00074839, -0.00054083, 0.00042830, -0.00031031, -0.00001087, 0.00016075, -0.00023681, 0.00010365, 0.00000869, -0.00009797, 0.00010323, -0.00008195, 0.00002096, 0.00007873, -0.00012405, 0.00003232, 0.00005247, -0.00006867, 0.00002685, 0.00000229, -0.00003792, 0.00002341, -0.00003598, -0.00004549, 0.00004398, -0.00000756, -0.00003993, 0.00004736, 0.00002860, -0.00001775, -0.00000006, -0.00000802, 0.00000483, 0.00002018, 0.00000351, -0.00000543, },
		{ -0.00000046, -0.00001090, 0.00001313, -0.00000447, -0.00003350, 0.00001852, -0.00001368, -0.00005936, 0.00003464, 0.00000452, -0.00004517, 0.00005361, -0.00003839, -0.00006116, 0.00009894, -0.00001951, -0.00001556, 0.00004494, -0.00009432, 0.00008320, -0.00003827, -0.00015108, 0.00021269, -0.00020385, -0.00006238, 0.00031141, -0.00094435, -0.00001328, 0.00188410, -0.00001868, -0.00099302, 0.00035761, -0.00006735, -0.00024202, 0.00025282, -0.00017910, -0.00003204, 0.00011434, -0.00013318, 0.00007155, -0.00001294, -0.00005539, 0.00011804, -0.00008173, -0.00004476, 0.00008931, -0.00005454, -0.00000967, 0.00003587, -0.00004943, 0.00000542, 0.00000224, -0.00006283, 0.00001724, 0.00002561, -0.00004600, 0.00001811, 0.00004664, -0.00000930, -0.00000548, -0.00000351, -0.00000448, 0.00001545, 0.00001571, },
		{ 0.00002005, 0.00001338, -0.00001690, 0.00000361, 0.00001566, -0.00003912, -0.00000271, 0.00001612, -0.00006272, 0.00000374, 0.00003654, -0.00005784, 0.00003246, 0.00001815, -0.00010045, 0.00006870, 0.00003797, -0.00006120, 0.00006993, -0.00007423, 0.00000820, 0.00007470, -0.00021095, 0.00014136, -0.00000280, -0.00028797, 0.00039910, -0.00052884, -0.00075549, 0.00162225, 0.00089164, -0.00112224, -0.00000906, 0.00024195, -0.00037424, 0.00019117, -0.00001801, -0.00017252, 0.00015303, -0.00008958, -0.00000930, 0.00005992, -0.00009407, 0.00009624, -0.00000009, -0.00010602, 0.00008058, -0.00000758, -0.00005068, 0.00005019, -0.00003134, -0.00002954, 0.00002711, -0.00004973, -0.00002113, 0.00004162, -0.00002887, -0.00001334, 0.00004855, 0.00000787, -0.00000994, -0.00000124, -0.00000683, 0.00000526, },
		{ -0.00000293, 0.00001398, 0.00002783, -0.00001017, -0.00001505, 0.00002828, -0.00002459, -0.00003261, 0.00002944, -0.00004110, -0.00003439, 0.00005254, -0.00004213, -0.00001151, 0.00006530, -0.00009469, 0.00000088, 0.00009410, -0.00007494, 0.00004613, -0.00000894, -0.00007677, 0.00013234, -0.00016253, -0.00001803, 0.00018060, -0.00035204, 0.00022806, -0.00001358, -0.00110824, 0.00089094, 0.00164143, -0.00078140, -0.00054319, 0.00043242, -0.00031247, -0.00000472, 0.00015802, -0.00023318, 0.00009352, 0.00000934, -0.00008929, 0.00009079, -0.00007512, 0.00003002, 0.00007795, -0.00011785, 0.00002713, 0.00004679, -0.00006975, 0.00003376, 0.00000632, -0.00005736, 0.00002344, -0.00001605, -0.00004942, 0.00003228, -0.00000018, -0.00003207, 0.00003333, 0.00002522, -0.00000770, -0.00000287, -0.00000484, },
		{ -0.00000338, -0.00000508, 0.00000207, 0.00003322, 0.00000952, -0.00003110, 0.00002291, 0.00000455, -0.00005299, 0.00001677, -0.00000541, -0.00005957, 0.00004088, -0.00000455, -0.00005377, 0.00007703, -0.00004373, -0.00007420, 0.00011356, -0.00003995, -0.00001486, 0.00006111, -0.00011791, 0.00010333, -0.00003608, -0.00017185, 0.00024363, -0.00023022, -0.00006924, 0.00035400, -0.00099472, -0.00002586, 0.00193092, -0.00002586, -0.00099472, 0.00035400, -0.00006924, -0.00023022, 0.00024363, -0.00017185, -0.00003608, 0.00010333, -0.00011791, 0.00006111, -0.00001486, -0.00003995, 0.00011356, -0.00007420, -0.00004373, 0.00007703, -0.00005377, -0.00000455, 0.00004088, -0.00005957, -0.00000541, 0.00001677, -0.00005299, 0.00000455, 0.00002291, -0.00003110, 0.00000952, 0.00003322, 0.00000207, -0.00000508, },
		{ -0.00000293, -0.00000484, -0.00000287, -0.00000770, 0.00002522, 0.00003333, -0.00003207, -0.00000018, 0.00003228, -0.00004942, -0.00001605, 0.00002344, -0.00005736, 0.00000632, 0.00003376, -0.00006975, 0.00004679, 0.00002713, -0.00011785, 0.00007795, 0.00003002, -0.00007512, 0.00009079, -0.00008929, 0.00000934, 0.00009352, -0.00023318, 0.00015802, -0.00000472, -0.00031247, 0.00043242, -0.00054319, -0.00078140, 0.00164143, 0.00089094, -0.00110824, -0.00001358, 0.00022806, -0.00035204, 0.00018060, -0.00001803, -0.00016253, 0.00013234, -0.00007677, -0.00000894, 0.00004613, -0.00007494, 0.00009410, 0.00000088, -0.00009469, 0.00006530, -0.00001151, -0.00004213, 0.00005254, -0.00003439, -0.00004110, 0.00002944, -0.00003261, -0.00002459, 0.00002828, -0.00001505, -0.00001017, 0.00002783, 0.00001398, },
		{ 0.00002005, 0.00000526, -0.00000683, -0.00000124, -0.00000994, 0.00000787, 0.00004855, -0.00001334, -0.00002887, 0.00004162, -0.00002113, -0.00004973, 0.00002711, -0.00002954, -0.00003134, 0.00005019, -0.00005068, -0.00000758, 0.00008058, -0.00010602, -0.00000009, 0.00009624, -0.00009407, 0.00005992, -0.00000930, -0.00008958, 0.00015303, -0.00017252, -0.00001801, 0.00019117, -0.00037424, 0.00024195, -0.00000906, -0.00112224, 0.00089164, 0.00162225, -0.00075549, -0.00052884, 0.00039910, -0.00028797, -0.00000280, 0.00014136, -0.00021095, 0.00007470, 0.00000820, -0.00007423, 0.00006993, -0.00006120, 0.00003797, 0.00006870, -0.00010045, 0.00001815, 0.00003246, -0.00005784, 0.00003654, 0.00000374, -0.00006272, 0.00001612, -0.00000271, -0.00003912, 0.00001566, 0.00000361, -0.00001690, 0.00001338, },
		{ -0.00000046, 0.00001571, 0.00001545, -0.00000448, -0.00000351, -0.00000548, -0.00000930, 0.00004664, 0.00001811, -0.00004600, 0.00002561, 0.00001724, -0.00006283, 0.00000224, 0.00000542, -0.00004943, 0.00003587, -0.00000967, -0.00005454, 0.00008931, -0.00004476, -0.00008173, 0.00011804, -0.00005539, -0.00001294, 0.00007155, -0.00013318, 0.00011434, -0.00003204, -0.00017910, 0.00025282, -0.00024202, -0.00006735, 0.00035761, -0.00099302, -0.00001868, 0.00188410, -0.00001328, -0.00094435, 0.00031141, -0.00006238, -0.00020385, 0.00021269, -0.00015108, -0.00003827, 0.00008320, -0.00009432, 0.00004494, -0.00001556, -0.00001951, 0.00009894, -0.00006116, -0.00003839, 0.00005361, -0.00004517, 0.00000452, 0.00003464, -0.00005936, -0.00001368, 0.00001852, -0.00003350, -0.00000447, 0.00001313, -0.00001090, },
		{ -0.00000029, -0.00000543, 0.00000351, 0.00002018, 0.00000483, -0.00000802, -0.00000006, -0.00001775, 0.00002860, 0.00004736, -0.00003993, -0.00000756, 0.00004398, -0.00004549, -0.00003598, 0.00002341, -0.00003792, 0.00000229, 0.00002685, -0.00006867, 0.00005247, 0.00003232, -0.00012405, 0.00007873, 0.00002096, -0.00008195, 0.00010323, -0.00009797, 0.00000869, 0.00010365, -0.00023681, 0.00016075, -0.00001087, -0.00031031, 0.00042830, -0.00054083, -0.00074839, 0.00158438, 0.00085820, -0.00102596, -0.00003646, 0.00019852, -0.00030406, 0.00015253, -0.00001684, -0.00014100, 0.00010079, -0.00005920, -0.00000879, 0.00002854, -0.00004883, 0.00008515, 0.00000018, -0.00007511, 0.00004213, -0.00001480, -0.00002459, 0.00004501, -0.00003496, -0.00004257, 0.00002021, -0.00001472, -0.00001882, 0.00000979, },
		{ -0.00000076, 0.00000535, -0.00000031, -0.00000818, 0.00001404, 0.00001796, -0.00000917, 0.00000071, -0.00001540, 0.00000403, 0.00006044, -0.00001193, -0.00003904, 0.00004370, -0.00000646, -0.00006261, 0.00001082, -0.00000718, -0.00002605, 0.00003708, -0.00004673, -0.00000455, 0.00008410, -0.00010591, -0.00000262, 0.00009024, -0.00010346, 0.00006788, -0.00000984, -0.00009593, 0.00015991, -0.00016898, -0.00001681, 0.00018209, -0.00036755, 0.00023850, -0.00002379, -0.00106628, 0.00086015, 0.00152920, -0.00067435, -0.00049895, 0.00033316, -0.00024089, -0.00000471, 0.00011367, -0.00017406, 0.00004974, 0.00000564, -0.00005502, 0.00004388, -0.00004242, 0.00004436, 0.00005325, -0.00007617, 0.00000722, 0.00001388, -0.00003698, 0.00003483, -0.00000353, -0.00005634, 0.00000403, 0.00000366, -0.00002002, },
		{ -0.00001055, -0.00000873, 0.00000190, 0.00000863, -0.00001097, -0.00000162, 0.00002699, -0.00000215, -0.00000389, -0.00000723, -0.00001522, 0.00005242, 0.00002458, -0.00005076, 0.00001716, 0.00003171, -0.00005889, -0.00002525, 0.00001798, -0.00002888, 0.00001899, -0.00000862, -0.00004709, 0.00008672, -0.00004063, -0.00008202, 0.00011084, -0.00006402, -0.00001004, 0.00007479, -0.00013816, 0.00011467, -0.00002631, -0.00017169, 0.00023874, -0.00023797, -0.00005706, 0.00032219, -0.00093954, 0.00000676, 0.00174958, 0.00001659, -0.00084806, 0.00023690, -0.00004870, -0.00016649, 0.00016559, -0.00012041, -0.00003878, 0.00005722, -0.00006600, 0.00002565, -0.00001500, 0.00000301, 0.00007780, -0.00004572, -0.00003052, 0.00002546, -0.00003115, 0.00001451, 0.00002068, -0.00005118, -0.00001903, 0.00001076, },
		{ 0.00000445, 0.00000039, -0.00000772, -0.00000668, 0.00001256, -0.00000250, -0.00001865, 0.00002430, 0.00001240, -0.00000890, -0.00000086, -0.00002233, 0.00002948, 0.00005227, -0.00003629, -0.00001917, 0.00004679, -0.00002511, -0.00006039, 0.00001536, -0.00000535, -0.00000838, 0.00001740, -0.00005389, 0.00004705, 0.00003286, -0.00011734, 0.00007051, 0.00001146, -0.00008089, 0.00010568, -0.00009929, 0.00000624, 0.00010411, -0.00022135, 0.00014922, -0.00002058, -0.00028250, 0.00038791, -0.00052137, -0.00066174, 0.00145815, 0.00079677, -0.00088665, -0.00007253, 0.00015699, -0.00023793, 0.00011246, -0.00001467, -0.00011201, 0.00006335, -0.00003949, -0.00000882, 0.00000999, -0.00001997, 0.00007209, -0.00000208, -0.00005229, 0.00001690, -0.00001616, -0.00000369, 0.00003097, -0.00003323, -0.00003679, },
		{ -0.00004299, -0.00000919, 0.00000418, 0.00000081, -0.00001199, 0.00000690, 0.00001168, -0.00002694, 0.00000823, 0.00002742, -0.00000759, -0.00000061, -0.00001873, 0.00000399, 0.00006006, -0.00000411, -0.00004362, 0.00003108, 0.00001874, -0.00006859, -0.00001818, 0.00002402, -0.00001911, 0.00001534, -0.00002990, -0.00000396, 0.00007482, -0.00009383, -0.00000637, 0.00007645, -0.00010233, 0.00006915, -0.00001056, -0.00009534, 0.00015260, -0.00015243, -0.00001468, 0.00015503, -0.00033392, 0.00021817, -0.00005412, -0.00094845, 0.00079955, 0.00137371, -0.00055007, -0.00045539, 0.00024458, -0.00017956, -0.00000903, 0.00007983, -0.00012907, 0.00002239, 0.00000229, -0.00003475, 0.00001695, -0.00002198, 0.00004893, 0.00003504, -0.00005089, -0.00000323, -0.00000361, -0.00001330, 0.00002941, -0.00001241, },
		{ 0.00000425, -0.00003892, -0.00002160, -0.00000142, 0.00000923, -0.00000798, -0.00000511, 0.00002187, -0.00002110, -0.00001480, 0.00003342, 0.00000298, -0.00000448, -0.00001138, -0.00001355, 0.00004855, 0.00002852, -0.00004288, -0.00000373, 0.00004552, -0.00004030, -0.00006233, 0.00003106, -0.00000039, -0.00000691, -0.00000064, -0.00003337, 0.00006904, -0.00003164, -0.00007505, 0.00009292, -0.00006539, -0.00000660, 0.00007073, -0.00013281, 0.00010474, -0.00001937, -0.00015118, 0.00020415, -0.00021925, -0.00004084, 0.00025455, -0.00084093, 0.00004537, 0.00154424, 0.00005787, -0.00071758, 0.00014316, -0.00003187, -0.00012347, 0.00011087, -0.00008519, -0.00003804, 0.00002976, -0.00003747, 0.00000645, -0.00001334, 0.00002418, 0.00005494, -0.00003159, -0.00002232, 0.00000007, -0.00001538, 0.00002226, },
		{ 0.00001434, 0.00001513, -0.00002976, -0.00002778, -0.00001177, 0.00000990, 0.00000416, -0.00001477, 0.00002107, -0.00000417, -0.00003369, 0.00002474, 0.00001894, -0.00000626, -0.00000671, -0.00002011, 0.00002754, 0.00004677, -0.00001972, -0.00003445, 0.00003879, 0.00000975, -0.00008555, 0.00000087, 0.00003596, -0.00002361, 0.00000802, -0.00002837, 0.00003066, 0.00002903, -0.00009903, 0.00005469, 0.00000231, -0.00007272, 0.00009847, -0.00009364, 0.00000231, 0.00009564, -0.00018976, 0.00012563, -0.00003212, -0.00023473, 0.00031808, -0.00048536, -0.00053468, 0.00127804, 0.00071277, -0.00070925, -0.00011338, 0.00010857, -0.00016438, 0.00006814, -0.00001187, -0.00008090, 0.00002610, -0.00002071, -0.00000902, -0.00000641, 0.00000682, 0.00005824, -0.00000561, -0.00003187, -0.00000404, -0.00001478, },
		{ -0.00001497, 0.00000667, 0.00002179, -0.00001962, -0.00002833, -0.00001985, 0.00000092, 0.00001659, -0.00001431, 0.00000994, 0.00001438, -0.00003934, 0.00000359, 0.00003155, -0.00000088, -0.00000596, -0.00001974, 0.00000823, 0.00004661, 0.00001022, -0.00004180, 0.00000395, 0.00005083, -0.00006639, -0.00005486, 0.00005973, -0.00001127, -0.00001009, -0.00000327, -0.00000705, 0.00005479, -0.00007205, -0.00001079, 0.00005694, -0.00009198, 0.00006409, -0.00001141, -0.00008850, 0.00013321, -0.00012573, -0.00001210, 0.00011507, -0.00027974, 0.00018352, -0.00009278, -0.00078509, 0.00071567, 0.00117438, -0.00040092, -0.00040073, 0.00014677, -0.00011489, -0.00001381, 0.00004583, -0.00008383, -0.00000316, -0.00000113, -0.00001672, -0.00000631, -0.00000336, 0.00005161, 0.00001783, -0.00003068, -0.00001075, },
		{ -0.00001588, -0.00001578, -0.00000169, 0.00002522, -0.00000927, -0.00002679, -0.00002190, -0.00001267, 0.00002045, -0.00000282, -0.00000372, 0.00002490, -0.00003004, -0.00002098, 0.00003287, 0.00001115, -0.00000524, -0.00001821, -0.00000406, 0.00003505, 0.00002980, -0.00002286, -0.00003505, 0.00005601, -0.00000985, -0.00010238, 0.00004334, 0.00003151, -0.00003598, 0.00001331, -0.00001729, 0.00003967, -0.00001925, -0.00006250, 0.00006754, -0.00006043, -0.00000321, 0.00006056, -0.00011888, 0.00008677, -0.00001203, -0.00012141, 0.00015540, -0.00018912, -0.00002276, 0.00016690, -0.00070905, 0.00008947, 0.00129298, 0.00010248, -0.00056833, 0.00004609, -0.00001639, -0.00008085, 0.00005831, -0.00005145, -0.00003660, 0.00000538, -0.00001332, -0.00000944, -0.00001085, 0.00004074, 0.00003519, -0.00002224, },
		{ -0.00001006, -0.00001870, -0.00001536, -0.00001038, 0.00002431, 0.00000210, -0.00002529, -0.00001954, -0.00002299, 0.00001223, 0.00001281, -0.00001050, 0.00002381, -0.00001194, -0.00003855, 0.00002095, 0.00002378, 0.00000011, -0.00001762, -0.00001137, 0.00002290, 0.00003175, 0.00000789, -0.00005157, 0.00002057, 0.00005310, -0.00010681, -0.00001638, 0.00007911, -0.00003950, 0.00000160, 0.00000173, 0.00000592, 0.00002212, -0.00007310, 0.00003428, -0.00000562, -0.00005956, 0.00008360, -0.00008260, -0.00000244, 0.00008042, -0.00014773, 0.00009416, -0.00004313, -0.00017612, 0.00022984, -0.00043436, -0.00038594, 0.00106497, 0.00061421, -0.00051709, -0.00014929, 0.00005904, -0.00009516, 0.00002791, -0.00000885, -0.00005310, -0.00000502, -0.00000575, -0.00000933, -0.00001789, 0.00002719, 0.00004672, },
		{ 0.00005244, 0.00000482, -0.00002024, -0.00001348, -0.00001664, 0.00001773, 0.00001371, -0.00002256, -0.00001737, -0.00002517, -0.00000382, 0.00002257, -0.00000616, 0.00001535, 0.00000560, -0.00004347, 0.00000086, 0.00002963, 0.00001064, -0.00001501, -0.00001858, 0.00001597, 0.00002216, 0.00002933, -0.00003413, -0.00003372, 0.00008381, -0.00005643, -0.00009137, 0.00009420, -0.00000312, -0.00003265, 0.00002723, -0.00001444, 0.00002857, -0.00004511, -0.00001523, 0.00003503, -0.00007537, 0.00005413, -0.00001233, -0.00007711, 0.00010577, -0.00009345, -0.00000963, 0.00006948, -0.00021449, 0.00013871, -0.00013050, -0.00059783, 0.00061624, 0.00095392, -0.00024806, -0.00033815, 0.00005416, -0.00005813, -0.00001703, 0.00001746, -0.00004587, -0.00002306, -0.00000393, -0.00000379, -0.00002208, 0.00001049, },
		{ -0.00000779, 0.00005028, 0.00002222, -0.00002004, -0.00001265, -0.00001767, 0.00000691, 0.00002205, -0.00001589, -0.00001825, -0.00002061, -0.00001886, 0.00002022, 0.00000555, 0.00000772, 0.00001636, -0.00003719, -0.00001905, 0.00002533, 0.00002163, -0.00000612, -0.00002705, 0.00001145, 0.00001403, 0.00002869, 0.00000584, -0.00007178, 0.00006082, 0.00002639, -0.00013692, 0.00005382, 0.00006137, -0.00006064, 0.00003057, -0.00000380, 0.00000474, -0.00000583, -0.00004726, 0.00003951, -0.00005110, -0.00000050, 0.00004640, -0.00009939, 0.00006424, -0.00000521, -0.00008769, 0.00010102, -0.00015228, -0.00000744, 0.00007431, -0.00055896, 0.00013035, 0.00102461, 0.00014171, -0.00041706, -0.00003828, -0.00000648, -0.00004442, 0.00001689, -0.00002459, -0.00003494, -0.00001205, 0.00000275, -0.00001949, },
		{ -0.00000970, -0.00002256, 0.00003818, 0.00003970, -0.00001504, -0.00001568, -0.00001403, -0.00000323, 0.00002331, -0.00000490, -0.00002062, -0.00001498, -0.00002582, 0.00000737, 0.00001578, 0.00000670, 0.00002057, -0.00002563, -0.00003269, 0.00001359, 0.00002655, 0.00000927, -0.00003205, 0.00000206, 0.00001605, 0.00001022, 0.00004175, -0.00006779, -0.00000488, 0.00009597, -0.00011963, -0.00003133, 0.00011644, -0.00005109, 0.00000053, 0.00002854, -0.00002259, 0.00001399, -0.00004520, 0.00001320, -0.00001158, -0.00004433, 0.00006415, -0.00006848, -0.00000720, 0.00006148, -0.00010246, 0.00006008, -0.00005115, -0.00011717, 0.00013636, -0.00037104, -0.00023623, 0.00084211, 0.00050987, -0.00033397, -0.00017133, 0.00001407, -0.00004074, -0.00000105, -0.00000597, -0.00003308, -0.00002537, 0.00000322, },
		{ -0.00000566, 0.00000220, -0.00002800, 0.00001775, 0.00005154, -0.00000213, -0.00002184, -0.00001057, -0.00000741, 0.00001731, 0.00000660, -0.00002008, -0.00001316, -0.00002396, -0.00000811, 0.00001711, 0.00001143, 0.00002304, -0.00001382, -0.00003937, 0.00000047, 0.00002244, 0.00002518, -0.00002631, -0.00001578, 0.00002531, -0.00000867, 0.00004992, -0.00002232, -0.00007501, 0.00011044, -0.00004086, -0.00011868, 0.00012177, 0.00000520, -0.00004588, 0.00005406, -0.00002588, 0.00000200, -0.00001875, -0.00001899, 0.00001447, -0.00005629, 0.00004128, -0.00001327, -0.00006333, 0.00007526, -0.00006087, -0.00000786, 0.00002622, -0.00014862, 0.00008895, -0.00015808, -0.00040984, 0.00050986, 0.00073555, -0.00011175, -0.00027143, -0.00002045, -0.00001850, -0.00001719, -0.00000074, -0.00002083, -0.00003455, },
		{ -0.00003340, -0.00002021, 0.00000864, -0.00002235, -0.00000439, 0.00005176, 0.00001773, -0.00002568, -0.00001308, -0.00000459, 0.00000909, 0.00001290, -0.00001400, -0.00001513, -0.00001844, -0.00001804, 0.00000880, 0.00001554, 0.00002785, -0.00000220, -0.00004220, -0.00000977, 0.00001266, 0.00003281, -0.00000706, -0.00003642, 0.00002952, -0.00001075, 0.00002569, 0.00003751, -0.00010692, 0.00005829, 0.00006035, -0.00015756, 0.00006214, 0.00008448, -0.00007351, 0.00004718, 0.00000243, -0.00002840, 0.00000590, -0.00003267, 0.00001399, -0.00003984, 0.00000099, 0.00003080, -0.00007788, 0.00004106, 0.00000018, -0.00005559, 0.00004984, -0.00011388, 0.00000110, -0.00000830, -0.00040664, 0.00016005, 0.00076720, 0.00016807, -0.00027930, -0.00009688, -0.00000502, -0.00001864, -0.00000700, -0.00000828, },
		{ -0.00000345, -0.00002332, -0.00003253, 0.00000518, -0.00001006, -0.00001982, 0.00003884, 0.00003786, -0.00002025, -0.00002305, -0.00000007, 0.00000592, 0.00001143, -0.00000479, -0.00001640, -0.00001529, -0.00001968, -0.00000311, 0.00001264, 0.00003389, 0.00001237, -0.00004333, -0.00001805, 0.00000440, 0.00002720, 0.00001939, -0.00004728, 0.00001712, 0.00000778, -0.00001344, 0.00007515, -0.00008001, -0.00003287, 0.00012842, -0.00012070, -0.00003870, 0.00014177, -0.00005355, 0.00000567, 0.00004440, -0.00004946, 0.00000647, -0.00002120, -0.00000465, -0.00001503, -0.00003001, 0.00004356, -0.00005379, -0.00001115, 0.00004203, -0.00006105, 0.00002855, -0.00005418, -0.00006750, 0.00005043, -0.00029913, -0.00010465, 0.00063126, 0.00040817, -0.00018025, -0.00017340, -0.00002162, -0.00000829, -0.00001427, },
		{ -0.00001368, -0.00000643, -0.00001135, -0.00003650, -0.00000614, 0.00000081, -0.00002369, 0.00001814, 0.00004904, -0.00000289, -0.00003475, -0.00000241, 0.00001132, 0.00000606, 0.00000126, -0.00001268, -0.00001622, -0.00001688, -0.00001076, 0.00000193, 0.00003510, 0.00003192, -0.00004025, -0.00002892, 0.00000249, 0.00001203, 0.00003984, -0.00003767, -0.00001209, 0.00003369, -0.00003985, 0.00006796, -0.00000887, -0.00011165, 0.00012393, -0.00002322, -0.00012890, 0.00013864, 0.00001441, -0.00004542, 0.00006981, -0.00004018, -0.00001933, 0.00000158, -0.00002150, -0.00000138, -0.00003846, 0.00002771, -0.00001415, -0.00004931, 0.00004649, -0.00003284, -0.00000725, -0.00000786, -0.00009152, 0.00003964, -0.00016848, -0.00024199, 0.00040478, 0.00053943, -0.00000794, -0.00020468, -0.00006802, -0.00000138, },
		{ -0.00001278, -0.00000590, -0.00001084, -0.00000377, -0.00003205, -0.00001871, 0.00000429, -0.00001812, -0.00000078, 0.00004568, 0.00002105, -0.00003801, -0.00001656, 0.00002081, 0.00000536, -0.00000062, -0.00000469, -0.00001732, -0.00001598, -0.00001076, -0.00001061, 0.00002546, 0.00005274, -0.00002683, -0.00004502, 0.00000451, -0.00000190, 0.00004255, -0.00000798, -0.00004439, 0.00004580, -0.00003471, 0.00002150, 0.00006556, -0.00013314, 0.00004788, 0.00008372, -0.00015831, 0.00006901, 0.00009850, -0.00006951, 0.00005850, -0.00000162, -0.00005306, 0.00001344, -0.00002162, -0.00000476, -0.00002895, 0.00000094, 0.00001610, -0.00005757, 0.00002073, 0.00000344, -0.00002979, 0.00000918, -0.00007855, 0.00000042, -0.00006903, -0.00026652, 0.00017286, 0.00054383, 0.00017663, -0.00016717, -0.00012206, },
		{ -0.00015355, -0.00004483, -0.00000042, -0.00001080, -0.00000130, -0.00002396, -0.00002674, 0.00000047, -0.00001037, -0.00001048, 0.00003042, 0.00004034, -0.00002542, -0.00003840, 0.00002348, 0.00001598, -0.00000832, 0.00000118, -0.00001308, -0.00001978, -0.00000673, -0.00001649, 0.00000490, 0.00006535, 0.00000123, -0.00006191, 0.00000148, -0.00000420, 0.00002596, 0.00002816, -0.00006001, 0.00003032, -0.00000097, -0.00003442, 0.00010108, -0.00008553, -0.00005813, 0.00014184, -0.00010916, -0.00003427, 0.00015254, -0.00004337, 0.00001564, 0.00004395, -0.00006962, 0.00000082, -0.00000567, -0.00001636, -0.00001578, -0.00001891, 0.00002473, -0.00004060, -0.00001365, 0.00002475, -0.00002909, 0.00000363, -0.00005119, -0.00003387, -0.00001770, -0.00022326, -0.00000545, 0.00044968, 0.00031614, -0.00006975, },
		{ 0.00005426, -0.00014218, -0.00008477, -0.00000728, -0.00000696, 0.00000013, -0.00001661, -0.00002955, -0.00000557, -0.00000694, -0.00001076, 0.00001290, 0.00004507, 0.00000103, -0.00005543, 0.00000946, 0.00003586, -0.00001240, -0.00000223, -0.00000238, -0.00002461, -0.00000613, -0.00001149, -0.00001851, 0.00005992, 0.00004018, -0.00006822, -0.00001528, 0.00000666, 0.00000120, 0.00005130, -0.00004663, -0.00000834, 0.00003847, -0.00006523, 0.00007956, 0.00000348, -0.00013597, 0.00011961, -0.00000785, -0.00011747, 0.00014438, 0.00002614, -0.00003067, 0.00006902, -0.00005516, -0.00003160, 0.00001200, -0.00002243, -0.00001046, -0.00002453, 0.00001517, -0.00001492, -0.00003670, 0.00002310, -0.00001280, -0.00000799, -0.00002837, -0.00004966, -0.00000431, -0.00015833, -0.00010961, 0.00030798, 0.00037994, },
		{ 0.00036947, 0.00016603, -0.00008782, -0.00011280, -0.00002818, -0.00000618, 0.00000366, -0.00000972, -0.00003069, -0.00000913, -0.00000836, -0.00000829, 0.00000300, 0.00003396, 0.00002946, -0.00005425, -0.00002155, 0.00005287, -0.00000209, -0.00001585, 0.00000866, -0.00002308, -0.00001365, 0.00000036, -0.00003271, 0.00003374, 0.00007682, -0.00005209, -0.00004578, 0.00002050, -0.00001454, 0.00004867, -0.00000883, -0.00004903, 0.00005615, -0.00005352, 0.00001684, 0.00008404, -0.00014460, 0.00003043, 0.00008986, -0.00013751, 0.00007630, 0.00010474, -0.00004766, 0.00006002, -0.00001621, -0.00006491, 0.00001507, -0.00001575, -0.00001442, -0.00001999, -0.00000066, 0.00000391, -0.00004064, 0.00000559, 0.00000421, -0.00001317, -0.00001663, -0.00004963, -0.00000967, -0.00010114, -0.00014957, 0.00016644, },
		{ 0.00005406, 0.00030792, 0.00023862, -0.00000795, -0.00011440, -0.00005431, -0.00001485, 0.00000665, 0.00000066, -0.00003275, -0.00001078, -0.00000934, -0.00001057, 0.00000345, 0.00001601, 0.00004497, -0.00003038, -0.00005723, 0.00005130, 0.00002569, -0.00003086, 0.00001000, -0.00001087, -0.00002637, 0.00000893, -0.00002945, -0.00000462, 0.00009400, -0.00001039, -0.00007777, 0.00002118, -0.00000973, 0.00002325, 0.00003335, -0.00006722, 0.00003847, -0.00000930, -0.00004859, 0.00011388, -0.00008280, -0.00007594, 0.00013117, -0.00008732, -0.00001596, 0.00015137, -0.00001940, 0.00002645, 0.00002575, -0.00007951, -0.00000273, -0.00000067, -0.00002075, -0.00001403, -0.00001212, 0.00000945, -0.00003013, -0.00001431, 0.00001125, -0.00000949, -0.00001252, -0.00004240, -0.00001891, -0.00006200, -0.00014858, },
		{ -0.00012865, -0.00002055, 0.00022442, 0.00026405, 0.00007369, -0.00008792, -0.00007250, -0.00003189, 0.00000150, 0.00001625, -0.00003264, -0.00001596, -0.00000439, -0.00001880, 0.00000754, 0.00000447, 0.00003982, 0.00000692, -0.00007839, 0.00002271, 0.00006104, -0.00003256, -0.00000436, 0.00000792, -0.00003481, 0.00000516, -0.00001068, -0.00003837, 0.00008051, 0.00004581, -0.00009174, -0.00000200, 0.00001249, -0.00000727, 0.00005665, -0.00005109, -0.00000526, 0.00003765, -0.00008000, 0.00008201, 0.00001240, -0.00014280, 0.00009630, 0.00000083, -0.00008483, 0.00014292, 0.00004299, -0.00000586, 0.00004954, -0.00006796, -0.00003389, 0.00001127, -0.00002169, -0.00001249, -0.00001553, 0.00000463, -0.00001551, -0.00002632, 0.00000683, -0.00000214, -0.00001004, -0.00003425, -0.00002554, -0.00003914, },
		{ -0.00002698, -0.00010399, -0.00006207, 0.00014210, 0.00024955, 0.00013853, -0.00004287, -0.00007462, -0.00004764, -0.00001719, 0.00003106, -0.00002255, -0.00002891, 0.00000536, -0.00002586, 0.00000461, 0.00000698, 0.00001952, 0.00003892, -0.00007070, -0.00002605, 0.00008428, -0.00001018, -0.00002965, 0.00002154, -0.00002968, -0.00001171, 0.00001163, -0.00005159, 0.00003919, 0.00009441, -0.00007250, -0.00004481, 0.00003485, -0.00002185, 0.00004950, -0.00000955, -0.00004898, 0.00005761, -0.00006406, 0.00001233, 0.00008910, -0.00013848, 0.00000814, 0.00007544, -0.00009909, 0.00008702, 0.00010872, -0.00001199, 0.00004809, -0.00003834, -0.00006313, 0.00001028, -0.00001504, -0.00001529, -0.00001344, -0.00000351, -0.00000518, -0.00002784, -0.00000359, 0.00000256, -0.00000625, -0.00002693, -0.00002872, },
		{ -0.00002928, -0.00002080, -0.00008133, -0.00008022, 0.00007415, 0.00020907, 0.00017798, 0.00000812, -0.00006264, -0.00005076, -0.00004509, 0.00003241, 0.00000272, -0.00004570, 0.00001074, -0.00002180, -0.00001060, 0.00001916, -0.00000020, 0.00004892, -0.00003499, -0.00007413, 0.00007705, 0.00003378, -0.00005036, 0.00001793, -0.00000976, -0.00003237, 0.00002254, -0.00003886, -0.00001273, 0.00011316, -0.00002013, -0.00008770, 0.00003663, -0.00001025, 0.00001958, 0.00003341, -0.00006691, 0.00003954, -0.00001649, -0.00005352, 0.00011074, -0.00007189, -0.00008330, 0.00009645, -0.00006078, 0.00001537, 0.00014652, 0.00001661, 0.00003180, -0.00000698, -0.00007789, -0.00000519, -0.00000517, -0.00001868, -0.00001036, -0.00000925, -0.00000196, -0.00002251, -0.00001315, 0.00000188, -0.00000211, -0.00001992, },
		{ -0.00001304, -0.00002777, -0.00001759, -0.00006266, -0.00008460, 0.00002532, 0.00015671, 0.00019117, 0.00005660, -0.00004515, -0.00003791, -0.00006719, 0.00000962, 0.00003746, -0.00005329, 0.00000093, -0.00000323, -0.00003182, 0.00002708, -0.00000412, 0.00003353, 0.00001162, -0.00009754, 0.00003476, 0.00008146, -0.00004857, -0.00000588, 0.00001526, -0.00004294, 0.00001381, -0.00000908, -0.00005205, 0.00009233, 0.00004715, -0.00010582, 0.00000785, 0.00001933, -0.00001129, 0.00005414, -0.00004987, -0.00000334, 0.00003043, -0.00008189, 0.00007452, 0.00001640, -0.00013088, 0.00005706, -0.00000101, -0.00003704, 0.00014244, 0.00006820, 0.00002013, 0.00001324, -0.00007538, -0.00002858, 0.00000117, -0.00001953, -0.00000914, -0.00001062, -0.00000384, -0.00001588, -0.00001804, -0.00000264, 0.00000001, },
		{ -0.00000101, -0.00000730, -0.00002461, -0.00001561, -0.00004745, -0.00008271, -0.00000532, 0.00010440, 0.00018036, 0.00009944, -0.00002875, -0.00001817, -0.00006637, -0.00003498, 0.00006349, -0.00003735, -0.00002621, 0.00002077, -0.00004399, 0.00001760, 0.00001115, 0.00000559, 0.00004517, -0.00008353, -0.00002809, 0.00010785, -0.00001581, -0.00003916, 0.00002963, -0.00003442, -0.00001033, 0.00001949, -0.00006223, 0.00004107, 0.00010109, -0.00008382, -0.00004246, 0.00004493, -0.00002172, 0.00004433, -0.00001010, -0.00004377, 0.00004923, -0.00006518, 0.00000839, 0.00007996, -0.00011591, -0.00001559, 0.00004132, -0.00005251, 0.00010497, 0.00011972, 0.00002861, 0.00002060, -0.00006195, -0.00005075, -0.00000015, -0.00001768, -0.00001022, -0.00000869, -0.00000705, -0.00001167, -0.00001842, -0.00000762, },
		{ -0.00001052, -0.00000421, -0.00000399, -0.00002054, -0.00001423, -0.00003399, -0.00007906, -0.00002272, 0.00006180, 0.00014931, 0.00013414, -0.00001155, -0.00000746, -0.00003664, -0.00008202, 0.00005923, 0.00000521, -0.00005801, 0.00003242, -0.00003412, -0.00001045, 0.00003366, -0.00001391, 0.00004951, -0.00003909, -0.00008410, 0.00009495, 0.00003916, -0.00006180, 0.00002169, -0.00000963, -0.00003544, 0.00003012, -0.00004242, -0.00001685, 0.00011823, -0.00002626, -0.00008972, 0.00004490, -0.00000496, 0.00001542, 0.00002788, -0.00005869, 0.00003316, -0.00002215, -0.00004904, 0.00009246, -0.00005475, -0.00007952, 0.00004323, -0.00003795, 0.00005591, 0.00015100, 0.00006009, 0.00002385, -0.00004663, -0.00006604, -0.00000861, -0.00001516, -0.00001286, -0.00000566, -0.00000861, -0.00001057, -0.00001687, },
		},
		{ { -0.00001506, -0.00001624, -0.00000992, 0.00000553, 0.00000616, -0.00000064, -0.00000411, -0.00001677, -0.00001268, -0.00002681, -0.00006266, -0.00002150, -0.00001135, -0.00002705, 0.00013661, 0.00017230, -0.00004687, -0.00002849, 0.00006628, -0.00006875, -0.00003857, 0.00007072, -0.00002833, -0.00003016, 0.00004008, -0.00000172, 0.00000122, -0.00001014, -0.00004588, 0.00001248, 0.00002678, -0.00000514, -0.00000322, -0.00000514, 0.00002678, 0.00001248, -0.00004588, -0.00001014, 0.00000122, -0.00000172, 0.00004008, -0.00003016, -0.00002833, 0.00007072, -0.00003857, -0.00006875, 0.00006628, -0.00002849, -0.00004687, 0.00017230, 0.00013661, -0.00002705, -0.00001135, -0.00002150, -0.00006266, -0.00002681, -0.00001268, -0.00001677, -0.00000411, -0.00000064, 0.00000616, 0.00000553, -0.00000992, -0.00001624, },
		{ -0.00001427, -0.00001412, -0.00001746, -0.00001474, 0.00000102, 0.00000738, 0.00000191, 0.00000134, -0.00001517, -0.00001782, -0.00000757, -0.00005411, -0.00004381, 0.00001449, -0.00004285, 0.00002761, 0.00020058, 0.00004635, -0.00009235, 0.00005997, 0.00001498, -0.00009193, 0.00003797, 0.00003301, -0.00005376, 0.00001349, 0.00002580, -0.00000128, -0.00000118, -0.00004022, -0.00001324, 0.00002786, 0.00000535, 0.00000015, -0.00001156, 0.00001066, 0.00003677, -0.00003355, -0.00003219, 0.00001303, -0.00001863, 0.00002518, 0.00002036, -0.00005853, 0.00003544, 0.00003916, -0.00008972, 0.00000126, 0.00003344, -0.00006869, 0.00008208, 0.00020083, 0.00004996, -0.00002904, -0.00002019, -0.00005446, -0.00004543, -0.00001925, -0.00001546, -0.00000622, -0.00000446, 0.00000235, 0.00000888, -0.00000490, },
		{ 0.00000101, -0.00001229, -0.00001232, -0.00001650, -0.00001895, -0.00000492, 0.00000744, 0.00000276, 0.00000474, -0.00000603, -0.00002588, -0.00000354, -0.00002724, -0.00006504, 0.00001194, -0.00000486, -0.00005724, 0.00013514, 0.00015653, -0.00008491, -0.00002081, 0.00008830, -0.00006611, -0.00003891, 0.00006573, -0.00002455, -0.00003039, 0.00003129, 0.00001439, 0.00000188, -0.00003003, -0.00003003, 0.00001837, 0.00001179, 0.00000565, -0.00000610, -0.00001212, 0.00004199, 0.00000039, -0.00005362, 0.00001019, -0.00000660, -0.00001306, 0.00004804, -0.00003184, -0.00002839, 0.00007166, -0.00003643, -0.00006964, 0.00003501, -0.00003274, -0.00000277, 0.00018721, 0.00014708, -0.00000135, -0.00003439, -0.00004813, -0.00005201, -0.00003256, -0.00002025, -0.00000528, -0.00000517, -0.00000430, 0.00000928, },
		{ 0.00000444, 0.00000674, -0.00000921, -0.00001111, -0.00001370, -0.00002076, -0.00001208, 0.00000538, 0.00000448, 0.00000339, 0.00000497, -0.00002384, -0.00001643, -0.00000144, -0.00006066, -0.00002175, 0.00003875, -0.00006786, 0.00001373, 0.00020060, 0.00001204, -0.00010639, 0.00007994, 0.00002394, -0.00008935, 0.00003321, 0.00003128, -0.00004729, 0.00000342, 0.00003101, 0.00001307, -0.00002056, -0.00003944, 0.00000562, 0.00001426, 0.00000568, 0.00000791, -0.00002384, 0.00002064, 0.00003856, -0.00005061, -0.00001763, 0.00001908, -0.00003472, 0.00002803, 0.00002262, -0.00005990, 0.00003705, 0.00003769, -0.00008265, -0.00001836, 0.00000335, -0.00003843, 0.00011815, 0.00020890, 0.00007534, -0.00003393, -0.00005654, -0.00005201, -0.00004064, -0.00003283, -0.00000863, -0.00000045, -0.00000974, },
		{ -0.00000845, -0.00000424, 0.00000905, -0.00000439, -0.00001019, -0.00001069, -0.00001965, -0.00001799, -0.00000039, 0.00000707, 0.00000124, 0.00000898, -0.00000958, -0.00003046, 0.00000275, -0.00003097, -0.00005587, 0.00004325, -0.00001373, -0.00008254, 0.00013799, 0.00013672, -0.00011258, -0.00001157, 0.00010060, -0.00006045, -0.00003915, 0.00005806, -0.00001758, -0.00003311, 0.00002444, 0.00003192, -0.00000510, -0.00004469, -0.00000966, 0.00001721, 0.00000052, 0.00001682, -0.00001369, -0.00001322, 0.00005349, -0.00001487, -0.00005184, 0.00002603, -0.00001889, -0.00001798, 0.00005240, -0.00003389, -0.00002528, 0.00006811, -0.00003122, -0.00007067, -0.00000314, -0.00003369, 0.00004292, 0.00020790, 0.00016748, 0.00000873, -0.00006972, -0.00006057, -0.00003951, -0.00004422, -0.00002239, 0.00000467, },
		{ 0.00000071, 0.00000005, -0.00001112, 0.00000596, 0.00000037, -0.00000856, -0.00000783, -0.00001732, -0.00002056, -0.00000803, 0.00000647, 0.00000299, 0.00000526, 0.00000545, -0.00002883, -0.00001413, -0.00000171, -0.00005953, 0.00000621, 0.00004786, -0.00009483, 0.00000889, 0.00019480, -0.00001765, -0.00011038, 0.00009134, 0.00003002, -0.00008329, 0.00002494, 0.00003128, -0.00004088, -0.00000742, 0.00004006, 0.00002107, -0.00004052, -0.00003019, 0.00001918, -0.00000029, 0.00001053, 0.00000979, -0.00003307, 0.00003247, 0.00003299, -0.00006032, -0.00000078, 0.00001645, -0.00004572, 0.00003190, 0.00002035, -0.00005698, 0.00003642, 0.00003511, -0.00007204, -0.00004344, -0.00002762, -0.00000669, 0.00015698, 0.00023146, 0.00009376, -0.00006059, -0.00008383, -0.00003938, -0.00004410, -0.00004299, },
		{ -0.00005816, -0.00001736, 0.00000879, -0.00001110, -0.00000039, 0.00000274, -0.00000727, -0.00000404, -0.00001444, -0.00002146, -0.00001296, 0.00000058, 0.00000648, 0.00000199, 0.00000965, -0.00001191, -0.00003075, 0.00000291, -0.00003322, -0.00003587, 0.00006463, -0.00003247, -0.00009733, 0.00013908, 0.00011311, -0.00012725, -0.00000340, 0.00010444, -0.00005166, -0.00004071, 0.00004945, -0.00000882, -0.00003769, 0.00002127, 0.00004685, -0.00001772, -0.00005099, 0.00001055, 0.00001045, -0.00000508, 0.00002483, -0.00002326, -0.00000840, 0.00005827, -0.00002902, -0.00004183, 0.00003373, -0.00003158, -0.00001693, 0.00005137, -0.00003456, -0.00002037, 0.00006087, -0.00002284, -0.00007337, -0.00004445, -0.00003409, 0.00008753, 0.00024315, 0.00019347, -0.00000673, -0.00010804, -0.00005537, -0.00003277, },
		{ -0.00002357, -0.00005680, -0.00004452, 0.00000835, -0.00000445, -0.00000592, 0.00000342, -0.00000855, -0.00000082, -0.00000819, -0.00002372, -0.00001434, -0.00000578, 0.00000462, 0.00000495, 0.00000495, 0.00000387, -0.00002813, -0.00001462, -0.00000624, -0.00004757, 0.00003264, 0.00004195, -0.00011545, 0.00000983, 0.00018011, -0.00003911, -0.00010613, 0.00009374, 0.00003493, -0.00007575, 0.00001424, 0.00003316, -0.00003571, -0.00001638, 0.00005052, 0.00002139, -0.00005624, -0.00001521, 0.00002554, -0.00001203, 0.00001713, 0.00000588, -0.00003656, 0.00004250, 0.00002213, -0.00006053, 0.00001348, 0.00000712, -0.00004941, 0.00003398, 0.00001489, -0.00005007, 0.00003296, 0.00003287, -0.00006023, -0.00007154, -0.00005842, 0.00002076, 0.00020490, 0.00027158, 0.00009249, -0.00010605, -0.00009230, },
		{ -0.00013694, -0.00003435, -0.00003814, -0.00006832, -0.00000646, 0.00000402, -0.00000917, 0.00000508, -0.00001071, -0.00000383, 0.00000342, -0.00002484, -0.00001818, -0.00000577, -0.00000371, 0.00000843, 0.00000357, 0.00000582, -0.00001001, -0.00002915, -0.00000396, -0.00002913, -0.00000944, 0.00007148, -0.00005335, -0.00010049, 0.00013317, 0.00008845, -0.00012869, 0.00000146, 0.00010213, -0.00004080, -0.00004420, 0.00004145, 0.00000006, -0.00004244, 0.00002194, 0.00005571, -0.00003216, -0.00004824, 0.00002637, 0.00000123, -0.00000517, 0.00002718, -0.00003134, 0.00000036, 0.00005524, -0.00003788, -0.00002716, 0.00003202, -0.00004072, -0.00001206, 0.00004473, -0.00003271, -0.00001522, 0.00005161, -0.00001270, -0.00007798, -0.00008503, -0.00003921, 0.00013191, 0.00030151, 0.00021518, -0.00005335, },
		{ 0.00005911, -0.00016105, -0.00007601, -0.00001377, -0.00007571, -0.00003425, 0.00000913, -0.00001113, 0.00000656, -0.00000661, -0.00001635, 0.00001381, -0.00001577, -0.00002949, -0.00000172, -0.00000859, 0.00000204, 0.00001072, 0.00000012, 0.00000286, -0.00002195, -0.00002088, -0.00001086, -0.00002512, 0.00005055, 0.00002537, -0.00012452, 0.00001139, 0.00015680, -0.00004985, -0.00009668, 0.00008838, 0.00003959, -0.00006870, 0.00000269, 0.00003625, -0.00003191, -0.00002150, 0.00005904, 0.00001505, -0.00006416, 0.00000131, 0.00002514, -0.00001842, 0.00002212, -0.00000179, -0.00003348, 0.00004746, 0.00000972, -0.00005173, 0.00002119, -0.00000519, -0.00004577, 0.00003223, 0.00000814, -0.00004056, 0.00002732, 0.00003143, -0.00004961, -0.00009854, -0.00009075, 0.00003953, 0.00027123, 0.00032664, },
		{ 0.00038831, 0.00021869, -0.00013010, -0.00014395, -0.00000459, -0.00005851, -0.00006808, 0.00000604, -0.00001019, 0.00000098, 0.00000824, -0.00002990, 0.00000877, 0.00000785, -0.00004058, -0.00000690, -0.00000001, -0.00001281, 0.00001510, 0.00000330, -0.00000140, -0.00000196, -0.00002915, -0.00001597, -0.00001563, 0.00001688, 0.00006349, -0.00006900, -0.00009476, 0.00011751, 0.00006645, -0.00011897, 0.00000195, 0.00009617, -0.00002968, -0.00004938, 0.00003490, 0.00000778, -0.00004532, 0.00002492, 0.00005676, -0.00004401, -0.00003831, 0.00003521, -0.00000688, -0.00000178, 0.00002327, -0.00003500, 0.00000979, 0.00004569, -0.00003884, -0.00001249, 0.00002232, -0.00004371, -0.00000624, 0.00003378, -0.00002841, -0.00001080, 0.00004224, -0.00000350, -0.00008265, -0.00012236, -0.00005553, 0.00018145, },
		{ 0.00004884, 0.00036606, 0.00038739, -0.00001502, -0.00021169, -0.00003493, -0.00001871, -0.00009435, -0.00001241, -0.00000131, -0.00001537, 0.00002445, -0.00002608, -0.00001843, 0.00003250, -0.00003241, -0.00003022, 0.00001518, -0.00001846, 0.00000189, 0.00001868, -0.00001020, 0.00000554, -0.00001269, -0.00003390, -0.00001087, 0.00000405, 0.00005548, 0.00000496, -0.00012086, 0.00000864, 0.00012784, -0.00004943, -0.00008519, 0.00007769, 0.00004388, -0.00006355, -0.00000813, 0.00003948, -0.00002862, -0.00002226, 0.00006257, 0.00000508, -0.00006297, 0.00001516, 0.00002041, -0.00001892, 0.00002280, -0.00001008, -0.00002515, 0.00004574, -0.00000015, -0.00003690, 0.00002053, -0.00001638, -0.00003681, 0.00002618, 0.00000167, -0.00003033, 0.00002132, 0.00002971, -0.00004147, -0.00011969, -0.00012925, },
		{ -0.00015675, -0.00008873, 0.00024565, 0.00050294, 0.00018889, -0.00022837, -0.00011991, 0.00002325, -0.00009211, -0.00005239, 0.00001406, -0.00003133, 0.00002117, 0.00000459, -0.00005096, 0.00003100, 0.00000475, -0.00005786, 0.00001316, 0.00000035, -0.00002416, 0.00002700, -0.00000152, -0.00000916, 0.00001210, -0.00003397, -0.00002906, 0.00000734, 0.00003644, 0.00004441, -0.00007442, -0.00008512, 0.00009240, 0.00005025, -0.00010165, -0.00000150, 0.00008855, -0.00002029, -0.00005536, 0.00002992, 0.00001385, -0.00004476, 0.00002764, 0.00005050, -0.00004971, -0.00002479, 0.00003681, -0.00001108, 0.00000205, 0.00001457, -0.00003288, 0.00001657, 0.00003286, -0.00003172, -0.00000206, 0.00000813, -0.00004015, -0.00000191, 0.00002076, -0.00002297, -0.00000681, 0.00003389, 0.00000161, -0.00008349, },
		{ -0.00003499, -0.00013125, -0.00018069, 0.00005299, 0.00049716, 0.00043882, -0.00013163, -0.00024240, 0.00002451, -0.00004398, -0.00010454, 0.00001706, -0.00002489, -0.00001219, 0.00004344, -0.00005442, -0.00001079, 0.00005026, -0.00005720, -0.00002201, 0.00003215, -0.00003513, 0.00000698, 0.00002573, -0.00002498, 0.00001420, -0.00000415, -0.00005206, -0.00000285, 0.00003384, 0.00004670, -0.00001223, -0.00010708, -0.00000145, 0.00009756, -0.00003955, -0.00007401, 0.00006449, 0.00004689, -0.00006087, -0.00001712, 0.00004195, -0.00002455, -0.00001969, 0.00005943, -0.00000460, -0.00005393, 0.00002344, 0.00001466, -0.00001499, 0.00001814, -0.00001592, -0.00001448, 0.00003797, -0.00000446, -0.00002053, 0.00001243, -0.00002344, -0.00002565, 0.00001705, -0.00000409, -0.00002086, 0.00001727, 0.00002502, },
		{ 0.00002609, 0.00000030, -0.00007536, -0.00019208, -0.00014125, 0.00033629, 0.00063725, 0.00011335, -0.00033303, -0.00006180, 0.00003762, -0.00013252, -0.00001968, 0.00001149, -0.00005423, 0.00005042, -0.00000932, -0.00006846, 0.00006039, -0.00000892, -0.00007133, 0.00003999, -0.00000756, -0.00003387, 0.00004254, -0.00001242, -0.00001383, 0.00002978, -0.00004514, -0.00003833, 0.00003663, 0.00004455, 0.00002060, -0.00006820, -0.00007688, 0.00006089, 0.00004136, -0.00008072, -0.00000732, 0.00008035, -0.00001421, -0.00006106, 0.00002615, 0.00001867, -0.00004034, 0.00002752, 0.00003947, -0.00004770, -0.00001177, 0.00003304, -0.00001036, 0.00000353, 0.00000394, -0.00002573, 0.00001858, 0.00002081, -0.00001899, 0.00000180, -0.00000625, -0.00003190, -0.00000005, 0.00000779, -0.00001849, -0.00000157, },
		{ -0.00001236, 0.00001696, 0.00001375, -0.00002678, -0.00013228, -0.00025205, 0.00006165, 0.00066672, 0.00046277, -0.00028384, -0.00023978, 0.00009054, -0.00009005, -0.00009880, 0.00004727, -0.00006009, 0.00000205, 0.00005793, -0.00008812, 0.00000846, 0.00006141, -0.00008632, -0.00000302, 0.00004416, -0.00005593, 0.00001858, 0.00002841, -0.00004123, 0.00002925, -0.00000014, -0.00007151, 0.00001405, 0.00005766, 0.00002731, -0.00002096, -0.00008821, -0.00001916, 0.00007014, -0.00002345, -0.00006422, 0.00005125, 0.00004740, -0.00006055, -0.00002380, 0.00004334, -0.00001863, -0.00001595, 0.00004997, -0.00001050, -0.00004025, 0.00002539, 0.00001082, -0.00000941, 0.00000909, -0.00001749, -0.00000487, 0.00002691, -0.00000239, -0.00000714, 0.00000015, -0.00002544, -0.00001527, 0.00000724, -0.00001016, },
		{ -0.00000410, -0.00001687, 0.00000747, 0.00001638, -0.00000761, -0.00005347, -0.00023545, -0.00021013, 0.00046431, 0.00077560, -0.00001397, -0.00042523, 0.00002590, 0.00002926, -0.00017019, 0.00002726, -0.00000581, -0.00007242, 0.00008475, -0.00003452, -0.00007691, 0.00009182, -0.00003342, -0.00007574, 0.00006903, -0.00002433, -0.00003805, 0.00005831, -0.00002892, -0.00001232, 0.00004708, -0.00006177, -0.00003970, 0.00006669, 0.00003977, -0.00000092, -0.00005249, -0.00007390, 0.00002753, 0.00003940, -0.00005960, -0.00001340, 0.00007180, -0.00001230, -0.00006565, 0.00002320, 0.00002315, -0.00003308, 0.00002304, 0.00002739, -0.00003883, -0.00000238, 0.00002701, -0.00000573, 0.00000116, -0.00000540, -0.00001610, 0.00001580, 0.00001295, -0.00000491, -0.00000062, -0.00001726, -0.00002226, 0.00000022, },
		{ -0.00000729, -0.00000067, -0.00001863, -0.00000342, 0.00002056, -0.00000736, -0.00001140, -0.00012584, -0.00034790, 0.00008854, 0.00086897, 0.00044190, -0.00045639, -0.00019611, 0.00014870, -0.00015173, -0.00007271, 0.00007018, -0.00010195, 0.00002827, 0.00006262, -0.00012141, 0.00003809, 0.00006123, -0.00011381, 0.00002559, 0.00004695, -0.00007636, 0.00003567, 0.00002417, -0.00005485, 0.00004879, -0.00000314, -0.00008726, 0.00003764, 0.00007041, 0.00000312, -0.00001913, -0.00006995, -0.00004222, 0.00004849, -0.00000501, -0.00005572, 0.00003964, 0.00004437, -0.00006207, -0.00002819, 0.00004405, -0.00001071, -0.00001343, 0.00003658, -0.00001083, -0.00002596, 0.00002247, 0.00001032, -0.00000514, -0.00000172, -0.00001491, 0.00000107, 0.00001648, 0.00000427, 0.00000033, -0.00001187, -0.00002389, },
		{ -0.00002336, -0.00001467, 0.00000128, -0.00001551, -0.00001882, 0.00002274, 0.00000077, -0.00001934, -0.00001538, -0.00029557, -0.00028610, 0.00063627, 0.00089695, -0.00019000, -0.00048637, 0.00013161, -0.00000611, -0.00019486, 0.00008162, -0.00003850, -0.00007996, 0.00011800, -0.00006890, -0.00007247, 0.00011869, -0.00006577, -0.00006759, 0.00009436, -0.00004782, -0.00003419, 0.00007012, -0.00004841, -0.00000322, 0.00005961, -0.00008065, -0.00003136, 0.00009115, 0.00002429, -0.00001453, -0.00003198, -0.00007762, -0.00000295, 0.00004235, -0.00004060, -0.00001778, 0.00006262, -0.00001468, -0.00006873, 0.00002104, 0.00002810, -0.00002507, 0.00001446, 0.00001776, -0.00002607, 0.00000216, 0.00002185, 0.00000023, -0.00000460, -0.00001112, -0.00000741, 0.00001037, 0.00001079, 0.00000584, -0.00000644, },
		{ 0.00000152, -0.00001958, -0.00002207, -0.00000129, -0.00000512, -0.00003179, 0.00000882, 0.00002597, -0.00003971, 0.00001727, -0.00011927, -0.00046788, 0.00014861, 0.00108930, 0.00036394, -0.00062794, -0.00011173, 0.00018532, -0.00021907, -0.00002641, 0.00007877, -0.00014314, 0.00006399, 0.00005417, -0.00014766, 0.00007375, 0.00004760, -0.00013318, 0.00005969, 0.00003848, -0.00009150, 0.00005509, 0.00001241, -0.00006190, 0.00006905, -0.00001336, -0.00009475, 0.00006321, 0.00006987, -0.00001912, -0.00000805, -0.00005692, -0.00006665, 0.00003370, 0.00001225, -0.00004770, 0.00003043, 0.00003730, -0.00006490, -0.00003043, 0.00004492, -0.00000185, -0.00001367, 0.00002283, -0.00000608, -0.00001448, 0.00001762, 0.00001264, -0.00000413, -0.00001101, -0.00001027, 0.00000261, 0.00001041, 0.00001161, },
		{ 0.00001323, 0.00000979, -0.00001120, -0.00002556, -0.00001134, 0.00000603, -0.00002861, -0.00002312, 0.00004529, -0.00002516, -0.00002916, 0.00003734, -0.00038023, -0.00035416, 0.00085107, 0.00097867, -0.00040184, -0.00050239, 0.00023896, -0.00006640, -0.00019834, 0.00013401, -0.00008325, -0.00007321, 0.00014339, -0.00010742, -0.00005431, 0.00013471, -0.00010021, -0.00004643, 0.00011032, -0.00007357, -0.00002209, 0.00007434, -0.00006674, 0.00001253, 0.00006396, -0.00009708, -0.00001439, 0.00010469, 0.00000309, -0.00001741, -0.00001225, -0.00008696, -0.00002691, 0.00004739, -0.00002476, -0.00001915, 0.00005242, -0.00002096, -0.00007025, 0.00002014, 0.00003358, -0.00001867, 0.00000393, 0.00001271, -0.00001339, 0.00000285, 0.00001943, 0.00000449, -0.00001153, -0.00001273, -0.00000263, 0.00000594, },
		{ 0.00000124, 0.00001077, 0.00001492, -0.00000115, -0.00002085, -0.00002379, 0.00000505, -0.00000640, -0.00005095, 0.00002769, 0.00002879, -0.00008159, 0.00006396, -0.00012309, -0.00060492, 0.00025418, 0.00130561, 0.00022539, -0.00077504, 0.00000447, 0.00019089, -0.00027971, 0.00003495, 0.00006910, -0.00017578, 0.00010374, 0.00003241, -0.00016117, 0.00010882, 0.00002182, -0.00013919, 0.00009307, 0.00001965, -0.00009743, 0.00007236, -0.00000511, -0.00005989, 0.00008544, -0.00002849, -0.00009134, 0.00008491, 0.00005730, -0.00003366, 0.00000837, -0.00005154, -0.00008809, 0.00002516, 0.00002583, -0.00003919, 0.00002368, 0.00002627, -0.00006868, -0.00003045, 0.00004670, 0.00000591, -0.00001666, 0.00001232, 0.00000115, -0.00000749, 0.00001407, 0.00001555, -0.00000649, -0.00001608, -0.00000681, },
		{ -0.00001177, -0.00000309, 0.00000639, 0.00001679, 0.00000597, -0.00001047, -0.00002714, -0.00001227, 0.00001657, -0.00004630, -0.00002656, 0.00007372, -0.00006463, -0.00002943, 0.00009814, -0.00049334, -0.00039591, 0.00109819, 0.00100122, -0.00062822, -0.00046707, 0.00033029, -0.00014331, -0.00017679, 0.00017484, -0.00013329, -0.00005203, 0.00015537, -0.00014323, -0.00002509, 0.00013574, -0.00012966, -0.00001522, 0.00011316, -0.00009602, -0.00000411, 0.00006912, -0.00007944, 0.00003181, 0.00005889, -0.00010639, 0.00000747, 0.00010457, -0.00001758, -0.00001000, 0.00000187, -0.00009901, -0.00004242, 0.00005176, -0.00001214, -0.00001700, 0.00004101, -0.00003055, -0.00007028, 0.00002127, 0.00003861, -0.00001556, -0.00000516, 0.00001216, -0.00000437, 0.00000238, 0.00001975, 0.00000473, -0.00001641, },
		{ -0.00001041, -0.00001589, -0.00000775, 0.00000031, 0.00001709, 0.00001031, -0.00000322, -0.00001628, -0.00003185, 0.00001511, -0.00000688, -0.00007536, 0.00005566, 0.00002328, -0.00012762, 0.00012992, -0.00014864, -0.00074498, 0.00041112, 0.00149159, 0.00003347, -0.00087680, 0.00013656, 0.00016241, -0.00032181, 0.00010232, 0.00004145, -0.00019350, 0.00014034, 0.00000059, -0.00015881, 0.00013613, -0.00001143, -0.00012949, 0.00011909, -0.00000577, -0.00009252, 0.00008305, -0.00002468, -0.00004865, 0.00009395, -0.00004445, -0.00007718, 0.00009746, 0.00003688, -0.00003724, 0.00002504, -0.00005371, -0.00010300, 0.00002112, 0.00003450, -0.00002953, 0.00001907, 0.00001186, -0.00007319, -0.00002782, 0.00004943, 0.00001029, -0.00002055, 0.00000736, 0.00000708, -0.00000441, 0.00001407, 0.00001603, },
		{ 0.00002091, 0.00000039, -0.00001631, -0.00001131, -0.00000786, 0.00001426, 0.00001672, -0.00000356, -0.00000150, -0.00003236, -0.00001507, 0.00003306, -0.00007095, -0.00002467, 0.00010381, -0.00011799, -0.00001258, 0.00015600, -0.00063247, -0.00039333, 0.00135784, 0.00095266, -0.00084300, -0.00038378, 0.00039059, -0.00022418, -0.00013181, 0.00019663, -0.00018006, -0.00001991, 0.00015115, -0.00016947, 0.00000955, 0.00012102, -0.00014756, 0.00002031, 0.00010220, -0.00011014, 0.00001550, 0.00005508, -0.00008313, 0.00005018, 0.00004571, -0.00010539, 0.00002889, 0.00009132, -0.00003207, 0.00000452, 0.00000756, -0.00011015, -0.00004939, 0.00005346, -0.00000233, -0.00001159, 0.00002856, -0.00004281, -0.00006865, 0.00002504, 0.00004131, -0.00001588, -0.00000935, 0.00001387, -0.00000099, 0.00000398, },
		{ -0.00000280, 0.00001797, 0.00001153, -0.00001279, -0.00001154, -0.00001494, 0.00000382, 0.00002625, -0.00000380, 0.00000027, -0.00000916, -0.00004678, 0.00003190, -0.00001036, -0.00010189, 0.00009295, 0.00000394, -0.00016940, 0.00021135, -0.00020501, -0.00086864, 0.00061598, 0.00162145, -0.00019425, -0.00091925, 0.00026490, 0.00010423, -0.00033697, 0.00016479, 0.00000038, -0.00019306, 0.00016680, -0.00003536, -0.00014101, 0.00014989, -0.00004530, -0.00010535, 0.00013253, -0.00003218, -0.00007793, 0.00008415, -0.00004167, -0.00003046, 0.00009243, -0.00005660, -0.00005521, 0.00009772, 0.00001438, -0.00002973, 0.00003729, -0.00006128, -0.00010941, 0.00001956, 0.00003811, -0.00001866, 0.00001621, -0.00000518, -0.00007816, -0.00002192, 0.00005215, 0.00000967, -0.00002225, 0.00000812, 0.00000809, },
		{ 0.00001420, -0.00000293, 0.00001014, 0.00001990, -0.00000706, -0.00000971, -0.00001481, -0.00001383, 0.00002970, 0.00000857, -0.00001427, 0.00001568, -0.00004494, -0.00001566, 0.00005322, -0.00010315, -0.00001308, 0.00012904, -0.00018170, 0.00002653, 0.00019726, -0.00078760, -0.00033303, 0.00160351, 0.00083186, -0.00102005, -0.00026488, 0.00041077, -0.00029490, -0.00007019, 0.00019591, -0.00021529, 0.00001690, 0.00013158, -0.00018111, 0.00004244, 0.00009350, -0.00014975, 0.00005316, 0.00008001, -0.00011284, 0.00003187, 0.00003522, -0.00007653, 0.00006325, 0.00002789, -0.00009343, 0.00004466, 0.00006849, -0.00003684, 0.00002149, 0.00000443, -0.00011720, -0.00004908, 0.00005155, 0.00000516, -0.00000357, 0.00001548, -0.00005713, -0.00006484, 0.00003130, 0.00003953, -0.00001795, -0.00000634, },
		{ -0.00001841, 0.00001245, 0.00000209, 0.00000068, 0.00002409, 0.00000107, -0.00001037, -0.00000602, -0.00002799, 0.00001496, 0.00003331, -0.00002587, 0.00001300, -0.00000446, -0.00006634, 0.00005669, -0.00002081, -0.00012532, 0.00013669, -0.00003260, -0.00019712, 0.00029905, -0.00029594, -0.00095444, 0.00085494, 0.00167507, -0.00043251, -0.00089818, 0.00037023, 0.00002690, -0.00032221, 0.00021225, -0.00004640, -0.00017521, 0.00017812, -0.00006837, -0.00011174, 0.00014727, -0.00007253, -0.00007142, 0.00013098, -0.00005377, -0.00005728, 0.00007506, -0.00005196, -0.00000939, 0.00008122, -0.00006126, -0.00003026, 0.00008561, -0.00000455, -0.00001379, 0.00004213, -0.00007110, -0.00010712, 0.00001880, 0.00003723, -0.00000699, 0.00001483, -0.00002404, -0.00008303, -0.00001237, 0.00005297, 0.00000398, },
		{ 0.00003180, -0.00001873, 0.00000400, 0.00000946, -0.00000773, 0.00002154, 0.00001382, -0.00001449, 0.00000271, -0.00002478, -0.00001674, 0.00005004, -0.00001014, -0.00002046, 0.00003791, -0.00006647, -0.00000963, 0.00007259, -0.00014080, 0.00001089, 0.00014198, -0.00024826, 0.00008844, 0.00020859, -0.00094173, -0.00021004, 0.00180626, 0.00064949, -0.00113825, -0.00012917, 0.00038952, -0.00034324, -0.00000227, 0.00017396, -0.00023307, 0.00005103, 0.00010086, -0.00017631, 0.00006676, 0.00005912, -0.00013561, 0.00007695, 0.00005175, -0.00010388, 0.00004098, 0.00001403, -0.00006094, 0.00006799, 0.00000998, -0.00007262, 0.00005115, 0.00004153, -0.00003126, 0.00003631, -0.00000574, -0.00011826, -0.00004340, 0.00004610, 0.00001056, 0.00000622, 0.00000226, -0.00007270, -0.00005816, 0.00003884, },
		{ 0.00000046, 0.00004966, -0.00000498, -0.00000679, 0.00001656, -0.00001070, 0.00000891, 0.00002924, -0.00001417, -0.00000103, -0.00000319, -0.00004412, 0.00003475, 0.00003307, -0.00005139, 0.00003597, -0.00000718, -0.00008589, 0.00008810, -0.00004089, -0.00013964, 0.00018093, -0.00008597, -0.00020199, 0.00037956, -0.00041768, -0.00098317, 0.00110513, 0.00164226, -0.00065292, -0.00081992, 0.00043774, -0.00005529, -0.00028050, 0.00023768, -0.00009005, -0.00014431, 0.00017263, -0.00009212, -0.00007738, 0.00012923, -0.00008747, -0.00003448, 0.00011552, -0.00006611, -0.00003554, 0.00005776, -0.00005317, 0.00000992, 0.00006307, -0.00005688, -0.00000764, 0.00006401, -0.00001599, 0.00000617, 0.00003887, -0.00008008, -0.00009736, 0.00001790, 0.00003268, 0.00000474, 0.00001466, -0.00004389, -0.00008678, },
		{ -0.00008836, -0.00004809, 0.00004546, 0.00001813, -0.00001470, 0.00001934, -0.00000274, -0.00001166, 0.00003663, 0.00000111, -0.00001814, 0.00001811, -0.00004167, -0.00001256, 0.00007028, -0.00003882, -0.00001713, 0.00005969, -0.00009555, 0.00000609, 0.00008571, -0.00017912, 0.00004691, 0.00013632, -0.00030736, 0.00016796, 0.00018040, -0.00107339, -0.00003010, 0.00194005, 0.00042649, -0.00118544, 0.00000212, 0.00033328, -0.00036150, 0.00006063, 0.00013628, -0.00023120, 0.00007596, 0.00006539, -0.00015690, 0.00007793, 0.00002518, -0.00010825, 0.00008764, 0.00002358, -0.00008588, 0.00004089, -0.00000384, -0.00003980, 0.00006369, -0.00000385, -0.00004726, 0.00004744, 0.00001615, -0.00001749, 0.00004571, -0.00001992, -0.00011301, -0.00003424, 0.00003784, 0.00001387, 0.00001695, -0.00001071, },
		{ 0.00001542, -0.00006373, -0.00008815, 0.00001508, 0.00004058, -0.00001385, 0.00001278, 0.00001628, -0.00002766, 0.00002312, 0.00002992, -0.00003107, 0.00001527, -0.00000647, -0.00005887, 0.00006137, 0.00002174, -0.00007407, 0.00006681, -0.00002061, -0.00009972, 0.00012197, -0.00007067, -0.00013965, 0.00021767, -0.00015135, -0.00017878, 0.00043769, -0.00055841, -0.00094228, 0.00133816, 0.00152518, -0.00082925, -0.00069975, 0.00045991, -0.00012775, -0.00021990, 0.00023867, -0.00012277, -0.00010707, 0.00015227, -0.00010273, -0.00004505, 0.00010021, -0.00008758, -0.00000160, 0.00009032, -0.00006742, -0.00001752, 0.00003624, -0.00004540, 0.00002365, 0.00004224, -0.00004457, 0.00000849, 0.00003789, -0.00001863, 0.00002550, 0.00002893, -0.00008604, -0.00008217, 0.00001667, 0.00002528, 0.00001574, },
		{ 0.00001500, 0.00002778, -0.00002305, -0.00010244, -0.00003489, 0.00004848, 0.00000048, -0.00000327, 0.00003547, -0.00002248, -0.00001110, 0.00005206, -0.00001769, -0.00001498, 0.00003227, -0.00006343, 0.00000103, 0.00008452, -0.00007362, -0.00000165, 0.00007476, -0.00012779, 0.00003199, 0.00008772, -0.00021151, 0.00009104, 0.00010881, -0.00034807, 0.00025439, 0.00010978, -0.00116059, 0.00019036, 0.00198680, 0.00019036, -0.00116059, 0.00010978, 0.00025439, -0.00034807, 0.00010881, 0.00009104, -0.00021151, 0.00008772, 0.00003199, -0.00012779, 0.00007476, -0.00000165, -0.00007362, 0.00008452, 0.00000103, -0.00006343, 0.00003227, -0.00001498, -0.00001769, 0.00005206, -0.00001110, -0.00002248, 0.00003547, -0.00000327, 0.00000048, 0.00004848, -0.00003489, -0.00010244, -0.00002305, 0.00002778, },
		{ 0.00001542, 0.00001574, 0.00002528, 0.00001667, -0.00008217, -0.00008604, 0.00002893, 0.00002550, -0.00001863, 0.00003789, 0.00000849, -0.00004457, 0.00004224, 0.00002365, -0.00004540, 0.00003624, -0.00001752, -0.00006742, 0.00009032, -0.00000160, -0.00008758, 0.00010021, -0.00004505, -0.00010273, 0.00015227, -0.00010707, -0.00012277, 0.00023867, -0.00021990, -0.00012775, 0.00045991, -0.00069975, -0.00082925, 0.00152518, 0.00133816, -0.00094228, -0.00055841, 0.00043769, -0.00017878, -0.00015135, 0.00021767, -0.00013965, -0.00007067, 0.00012197, -0.00009972, -0.00002061, 0.00006681, -0.00007407, 0.00002174, 0.00006137, -0.00005887, -0.00000647, 0.00001527, -0.00003107, 0.00002992, 0.00002312, -0.00002766, 0.00001628, 0.00001278, -0.00001385, 0.00004058, 0.00001508, -0.00008815, -0.00006373, },
		{ -0.00008836, -0.00001071, 0.00001695, 0.00001387, 0.00003784, -0.00003424, -0.00011301, -0.00001992, 0.00004571, -0.00001749, 0.00001615, 0.00004744, -0.00004726, -0.00000385, 0.00006369, -0.00003980, -0.00000384, 0.00004089, -0.00008588, 0.00002358, 0.00008764, -0.00010825, 0.00002518, 0.00007793, -0.00015690, 0.00006539, 0.00007596, -0.00023120, 0.00013628, 0.00006063, -0.00036150, 0.00033328, 0.00000212, -0.00118544, 0.00042649, 0.00194005, -0.00003010, -0.00107339, 0.00018040, 0.00016796, -0.00030736, 0.00013632, 0.00004691, -0.00017912, 0.00008571, 0.00000609, -0.00009555, 0.00005969, -0.00001713, -0.00003882, 0.00007028, -0.00001256, -0.00004167, 0.00001811, -0.00001814, 0.00000111, 0.00003663, -0.00001166, -0.00000274, 0.00001934, -0.00001470, 0.00001813, 0.00004546, -0.00004809, },
		{ 0.00000046, -0.00008678, -0.00004389, 0.00001466, 0.00000474, 0.00003268, 0.00001790, -0.00009736, -0.00008008, 0.00003887, 0.00000617, -0.00001599, 0.00006401, -0.00000764, -0.00005688, 0.00006307, 0.00000992, -0.00005317, 0.00005776, -0.00003554, -0.00006611, 0.00011552, -0.00003448, -0.00008747, 0.00012923, -0.00007738, -0.00009212, 0.00017263, -0.00014431, -0.00009005, 0.00023768, -0.00028050, -0.00005529, 0.00043774, -0.00081992, -0.00065292, 0.00164226, 0.00110513, -0.00098317, -0.00041768, 0.00037956, -0.00020199, -0.00008597, 0.00018093, -0.00013964, -0.00004089, 0.00008810, -0.00008589, -0.00000718, 0.00003597, -0.00005139, 0.00003307, 0.00003475, -0.00004412, -0.00000319, -0.00000103, -0.00001417, 0.00002924, 0.00000891, -0.00001070, 0.00001656, -0.00000679, -0.00000498, 0.00004966, },
		{ 0.00003180, 0.00003884, -0.00005816, -0.00007270, 0.00000226, 0.00000622, 0.00001056, 0.00004610, -0.00004340, -0.00011826, -0.00000574, 0.00003631, -0.00003126, 0.00004153, 0.00005115, -0.00007262, 0.00000998, 0.00006799, -0.00006094, 0.00001403, 0.00004098, -0.00010388, 0.00005175, 0.00007695, -0.00013561, 0.00005912, 0.00006676, -0.00017631, 0.00010086, 0.00005103, -0.00023307, 0.00017396, -0.00000227, -0.00034324, 0.00038952, -0.00012917, -0.00113825, 0.00064949, 0.00180626, -0.00021004, -0.00094173, 0.00020859, 0.00008844, -0.00024826, 0.00014198, 0.00001089, -0.00014080, 0.00007259, -0.00000963, -0.00006647, 0.00003791, -0.00002046, -0.00001014, 0.00005004, -0.00001674, -0.00002478, 0.00000271, -0.00001449, 0.00001382, 0.00002154, -0.00000773, 0.00000946, 0.00000400, -0.00001873, },
		{ -0.00001841, 0.00000398, 0.00005297, -0.00001237, -0.00008303, -0.00002404, 0.00001483, -0.00000699, 0.00003723, 0.00001880, -0.00010712, -0.00007110, 0.00004213, -0.00001379, -0.00000455, 0.00008561, -0.00003026, -0.00006126, 0.00008122, -0.00000939, -0.00005196, 0.00007506, -0.00005728, -0.00005377, 0.00013098, -0.00007142, -0.00007253, 0.00014727, -0.00011174, -0.00006837, 0.00017812, -0.00017521, -0.00004640, 0.00021225, -0.00032221, 0.00002690, 0.00037023, -0.00089818, -0.00043251, 0.00167507, 0.00085494, -0.00095444, -0.00029594, 0.00029905, -0.00019712, -0.00003260, 0.00013669, -0.00012532, -0.00002081, 0.00005669, -0.00006634, -0.00000446, 0.00001300, -0.00002587, 0.00003331, 0.00001496, -0.00002799, -0.00000602, -0.00001037, 0.00000107, 0.00002409, 0.00000068, 0.00000209, 0.00001245, },
		{ 0.00001420, -0.00000634, -0.00001795, 0.00003953, 0.00003130, -0.00006484, -0.00005713, 0.00001548, -0.00000357, 0.00000516, 0.00005155, -0.00004908, -0.00011720, 0.00000443, 0.00002149, -0.00003684, 0.00006849, 0.00004466, -0.00009343, 0.00002789, 0.00006325, -0.00007653, 0.00003522, 0.00003187, -0.00011284, 0.00008001, 0.00005316, -0.00014975, 0.00009350, 0.00004244, -0.00018111, 0.00013158, 0.00001690, -0.00021529, 0.00019591, -0.00007019, -0.00029490, 0.00041077, -0.00026488, -0.00102005, 0.00083186, 0.00160351, -0.00033303, -0.00078760, 0.00019726, 0.00002653, -0.00018170, 0.00012904, -0.00001308, -0.00010315, 0.00005322, -0.00001566, -0.00004494, 0.00001568, -0.00001427, 0.00000857, 0.00002970, -0.00001383, -0.00001481, -0.00000971, -0.00000706, 0.00001990, 0.00001014, -0.00000293, },
		{ -0.00000280, 0.00000809, 0.00000812, -0.00002225, 0.00000967, 0.00005215, -0.00002192, -0.00007816, -0.00000518, 0.00001621, -0.00001866, 0.00003811, 0.00001956, -0.00010941, -0.00006128, 0.00003729, -0.00002973, 0.00001438, 0.00009772, -0.00005521, -0.00005660, 0.00009243, -0.00003046, -0.00004167, 0.00008415, -0.00007793, -0.00003218, 0.00013253, -0.00010535, -0.00004530, 0.00014989, -0.00014101, -0.00003536, 0.00016680, -0.00019306, 0.00000038, 0.00016479, -0.00033697, 0.00010423, 0.00026490, -0.00091925, -0.00019425, 0.00162145, 0.00061598, -0.00086864, -0.00020501, 0.00021135, -0.00016940, 0.00000394, 0.00009295, -0.00010189, -0.00001036, 0.00003190, -0.00004678, -0.00000916, 0.00000027, -0.00000380, 0.00002625, 0.00000382, -0.00001494, -0.00001154, -0.00001279, 0.00001153, 0.00001797, },
		{ 0.00002091, 0.00000398, -0.00000099, 0.00001387, -0.00000935, -0.00001588, 0.00004131, 0.00002504, -0.00006865, -0.00004281, 0.00002856, -0.00001159, -0.00000233, 0.00005346, -0.00004939, -0.00011015, 0.00000756, 0.00000452, -0.00003207, 0.00009132, 0.00002889, -0.00010539, 0.00004571, 0.00005018, -0.00008313, 0.00005508, 0.00001550, -0.00011014, 0.00010220, 0.00002031, -0.00014756, 0.00012102, 0.00000955, -0.00016947, 0.00015115, -0.00001991, -0.00018006, 0.00019663, -0.00013181, -0.00022418, 0.00039059, -0.00038378, -0.00084300, 0.00095266, 0.00135784, -0.00039333, -0.00063247, 0.00015600, -0.00001258, -0.00011799, 0.00010381, -0.00002467, -0.00007095, 0.00003306, -0.00001507, -0.00003236, -0.00000150, -0.00000356, 0.00001672, 0.00001426, -0.00000786, -0.00001131, -0.00001631, 0.00000039, },
		{ -0.00001041, 0.00001603, 0.00001407, -0.00000441, 0.00000708, 0.00000736, -0.00002055, 0.00001029, 0.00004943, -0.00002782, -0.00007319, 0.00001186, 0.00001907, -0.00002953, 0.00003450, 0.00002112, -0.00010300, -0.00005371, 0.00002504, -0.00003724, 0.00003688, 0.00009746, -0.00007718, -0.00004445, 0.00009395, -0.00004865, -0.00002468, 0.00008305, -0.00009252, -0.00000577, 0.00011909, -0.00012949, -0.00001143, 0.00013613, -0.00015881, 0.00000059, 0.00014034, -0.00019350, 0.00004145, 0.00010232, -0.00032181, 0.00016241, 0.00013656, -0.00087680, 0.00003347, 0.00149159, 0.00041112, -0.00074498, -0.00014864, 0.00012992, -0.00012762, 0.00002328, 0.00005566, -0.00007536, -0.00000688, 0.00001511, -0.00003185, -0.00001628, -0.00000322, 0.00001031, 0.00001709, 0.00000031, -0.00000775, -0.00001589, },
		{ -0.00001177, -0.00001641, 0.00000473, 0.00001975, 0.00000238, -0.00000437, 0.00001216, -0.00000516, -0.00001556, 0.00003861, 0.00002127, -0.00007028, -0.00003055, 0.00004101, -0.00001700, -0.00001214, 0.00005176, -0.00004242, -0.00009901, 0.00000187, -0.00001000, -0.00001758, 0.00010457, 0.00000747, -0.00010639, 0.00005889, 0.00003181, -0.00007944, 0.00006912, -0.00000411, -0.00009602, 0.00011316, -0.00001522, -0.00012966, 0.00013574, -0.00002509, -0.00014323, 0.00015537, -0.00005203, -0.00013329, 0.00017484, -0.00017679, -0.00014331, 0.00033029, -0.00046707, -0.00062822, 0.00100122, 0.00109819, -0.00039591, -0.00049334, 0.00009814, -0.00002943, -0.00006463, 0.00007372, -0.00002656, -0.00004630, 0.00001657, -0.00001227, -0.00002714, -0.00001047, 0.00000597, 0.00001679, 0.00000639, -0.00000309, },
		{ 0.00000124, -0.00000681, -0.00001608, -0.00000649, 0.00001555, 0.00001407, -0.00000749, 0.00000115, 0.00001232, -0.00001666, 0.00000591, 0.00004670, -0.00003045, -0.00006868, 0.00002627, 0.00002368, -0.00003919, 0.00002583, 0.00002516, -0.00008809, -0.00005154, 0.00000837, -0.00003366, 0.00005730, 0.00008491, -0.00009134, -0.00002849, 0.00008544, -0.00005989, -0.00000511, 0.00007236, -0.00009743, 0.00001965, 0.00009307, -0.00013919, 0.00002182, 0.00010882, -0.00016117, 0.00003241, 0.00010374, -0.00017578, 0.00006910, 0.00003495, -0.00027971, 0.00019089, 0.00000447, -0.00077504, 0.00022539, 0.00130561, 0.00025418, -0.00060492, -0.00012309, 0.00006396, -0.00008159, 0.00002879, 0.00002769, -0.00005095, -0.00000640, 0.00000505, -0.00002379, -0.00002085, -0.00000115, 0.00001492, 0.00001077, },
		{ 0.00001323, 0.00000594, -0.00000263, -0.00001273, -0.00001153, 0.00000449, 0.00001943, 0.00000285, -0.00001339, 0.00001271, 0.00000393, -0.00001867, 0.00003358, 0.00002014, -0.00007025, -0.00002096, 0.00005242, -0.00001915, -0.00002476, 0.00004739, -0.00002691, -0.00008696, -0.00001225, -0.00001741, 0.00000309, 0.00010469, -0.00001439, -0.00009708, 0.00006396, 0.00001253, -0.00006674, 0.00007434, -0.00002209, -0.00007357, 0.00011032, -0.00004643, -0.00010021, 0.00013471, -0.00005431, -0.00010742, 0.00014339, -0.00007321, -0.00008325, 0.00013401, -0.00019834, -0.00006640, 0.00023896, -0.00050239, -0.00040184, 0.00097867, 0.00085107, -0.00035416, -0.00038023, 0.00003734, -0.00002916, -0.00002516, 0.00004529, -0.00002312, -0.00002861, 0.00000603, -0.00001134, -0.00002556, -0.00001120, 0.00000979, },
		{ 0.00000152, 0.00001161, 0.00001041, 0.00000261, -0.00001027, -0.00001101, -0.00000413, 0.00001264, 0.00001762, -0.00001448, -0.00000608, 0.00002283, -0.00001367, -0.00000185, 0.00004492, -0.00003043, -0.00006490, 0.00003730, 0.00003043, -0.00004770, 0.00001225, 0.00003370, -0.00006665, -0.00005692, -0.00000805, -0.00001912, 0.00006987, 0.00006321, -0.00009475, -0.00001336, 0.00006905, -0.00006190, 0.00001241, 0.00005509, -0.00009150, 0.00003848, 0.00005969, -0.00013318, 0.00004760, 0.00007375, -0.00014766, 0.00005417, 0.00006399, -0.00014314, 0.00007877, -0.00002641, -0.00021907, 0.00018532, -0.00011173, -0.00062794, 0.00036394, 0.00108930, 0.00014861, -0.00046788, -0.00011927, 0.00001727, -0.00003971, 0.00002597, 0.00000882, -0.00003179, -0.00000512, -0.00000129, -0.00002207, -0.00001958, },
		{ -0.00002336, -0.00000644, 0.00000584, 0.00001079, 0.00001037, -0.00000741, -0.00001112, -0.00000460, 0.00000023, 0.00002185, 0.00000216, -0.00002607, 0.00001776, 0.00001446, -0.00002507, 0.00002810, 0.00002104, -0.00006873, -0.00001468, 0.00006262, -0.00001778, -0.00004060, 0.00004235, -0.00000295, -0.00007762, -0.00003198, -0.00001453, 0.00002429, 0.00009115, -0.00003136, -0.00008065, 0.00005961, -0.00000322, -0.00004841, 0.00007012, -0.00003419, -0.00004782, 0.00009436, -0.00006759, -0.00006577, 0.00011869, -0.00007247, -0.00006890, 0.00011800, -0.00007996, -0.00003850, 0.00008162, -0.00019486, -0.00000611, 0.00013161, -0.00048637, -0.00019000, 0.00089695, 0.00063627, -0.00028610, -0.00029557, -0.00001538, -0.00001934, 0.00000077, 0.00002274, -0.00001882, -0.00001551, 0.00000128, -0.00001467, },
		{ -0.00000729, -0.00002389, -0.00001187, 0.00000033, 0.00000427, 0.00001648, 0.00000107, -0.00001491, -0.00000172, -0.00000514, 0.00001032, 0.00002247, -0.00002596, -0.00001083, 0.00003658, -0.00001343, -0.00001071, 0.00004405, -0.00002819, -0.00006207, 0.00004437, 0.00003964, -0.00005572, -0.00000501, 0.00004849, -0.00004222, -0.00006995, -0.00001913, 0.00000312, 0.00007041, 0.00003764, -0.00008726, -0.00000314, 0.00004879, -0.00005485, 0.00002417, 0.00003567, -0.00007636, 0.00004695, 0.00002559, -0.00011381, 0.00006123, 0.00003809, -0.00012141, 0.00006262, 0.00002827, -0.00010195, 0.00007018, -0.00007271, -0.00015173, 0.00014870, -0.00019611, -0.00045639, 0.00044190, 0.00086897, 0.00008854, -0.00034790, -0.00012584, -0.00001140, -0.00000736, 0.00002056, -0.00000342, -0.00001863, -0.00000067, },
		{ -0.00000410, 0.00000022, -0.00002226, -0.00001726, -0.00000062, -0.00000491, 0.00001295, 0.00001580, -0.00001610, -0.00000540, 0.00000116, -0.00000573, 0.00002701, -0.00000238, -0.00003883, 0.00002739, 0.00002304, -0.00003308, 0.00002315, 0.00002320, -0.00006565, -0.00001230, 0.00007180, -0.00001340, -0.00005960, 0.00003940, 0.00002753, -0.00007390, -0.00005249, -0.00000092, 0.00003977, 0.00006669, -0.00003970, -0.00006177, 0.00004708, -0.00001232, -0.00002892, 0.00005831, -0.00003805, -0.00002433, 0.00006903, -0.00007574, -0.00003342, 0.00009182, -0.00007691, -0.00003452, 0.00008475, -0.00007242, -0.00000581, 0.00002726, -0.00017019, 0.00002926, 0.00002590, -0.00042523, -0.00001397, 0.00077560, 0.00046431, -0.00021013, -0.00023545, -0.00005347, -0.00000761, 0.00001638, 0.00000747, -0.00001687, },
		{ -0.00001236, -0.00001016, 0.00000724, -0.00001527, -0.00002544, 0.00000015, -0.00000714, -0.00000239, 0.00002691, -0.00000487, -0.00001749, 0.00000909, -0.00000941, 0.00001082, 0.00002539, -0.00004025, -0.00001050, 0.00004997, -0.00001595, -0.00001863, 0.00004334, -0.00002380, -0.00006055, 0.00004740, 0.00005125, -0.00006422, -0.00002345, 0.00007014, -0.00001916, -0.00008821, -0.00002096, 0.00002731, 0.00005766, 0.00001405, -0.00007151, -0.00000014, 0.00002925, -0.00004123, 0.00002841, 0.00001858, -0.00005593, 0.00004416, -0.00000302, -0.00008632, 0.00006141, 0.00000846, -0.00008812, 0.00005793, 0.00000205, -0.00006009, 0.00004727, -0.00009880, -0.00009005, 0.00009054, -0.00023978, -0.00028384, 0.00046277, 0.00066672, 0.00006165, -0.00025205, -0.00013228, -0.00002678, 0.00001375, 0.00001696, },
		{ 0.00002609, -0.00000157, -0.00001849, 0.00000779, -0.00000005, -0.00003190, -0.00000625, 0.00000180, -0.00001899, 0.00002081, 0.00001858, -0.00002573, 0.00000394, 0.00000353, -0.00001036, 0.00003304, -0.00001177, -0.00004770, 0.00003947, 0.00002752, -0.00004034, 0.00001867, 0.00002615, -0.00006106, -0.00001421, 0.00008035, -0.00000732, -0.00008072, 0.00004136, 0.00006089, -0.00007688, -0.00006820, 0.00002060, 0.00004455, 0.00003663, -0.00003833, -0.00004514, 0.00002978, -0.00001383, -0.00001242, 0.00004254, -0.00003387, -0.00000756, 0.00003999, -0.00007133, -0.00000892, 0.00006039, -0.00006846, -0.00000932, 0.00005042, -0.00005423, 0.00001149, -0.00001968, -0.00013252, 0.00003762, -0.00006180, -0.00033303, 0.00011335, 0.00063725, 0.00033629, -0.00014125, -0.00019208, -0.00007536, 0.00000030, },
		{ -0.00003499, 0.00002502, 0.00001727, -0.00002086, -0.00000409, 0.00001705, -0.00002565, -0.00002344, 0.00001243, -0.00002053, -0.00000446, 0.00003797, -0.00001448, -0.00001592, 0.00001814, -0.00001499, 0.00001466, 0.00002344, -0.00005393, -0.00000460, 0.00005943, -0.00001969, -0.00002455, 0.00004195, -0.00001712, -0.00006087, 0.00004689, 0.00006449, -0.00007401, -0.00003955, 0.00009756, -0.00000145, -0.00010708, -0.00001223, 0.00004670, 0.00003384, -0.00000285, -0.00005206, -0.00000415, 0.00001420, -0.00002498, 0.00002573, 0.00000698, -0.00003513, 0.00003215, -0.00002201, -0.00005720, 0.00005026, -0.00001079, -0.00005442, 0.00004344, -0.00001219, -0.00002489, 0.00001706, -0.00010454, -0.00004398, 0.00002451, -0.00024240, -0.00013163, 0.00043882, 0.00049716, 0.00005299, -0.00018069, -0.00013125, },
		{ -0.00015675, -0.00008349, 0.00000161, 0.00003389, -0.00000681, -0.00002297, 0.00002076, -0.00000191, -0.00004015, 0.00000813, -0.00000206, -0.00003172, 0.00003286, 0.00001657, -0.00003288, 0.00001457, 0.00000205, -0.00001108, 0.00003681, -0.00002479, -0.00004971, 0.00005050, 0.00002764, -0.00004476, 0.00001385, 0.00002992, -0.00005536, -0.00002029, 0.00008855, -0.00000150, -0.00010165, 0.00005025, 0.00009240, -0.00008512, -0.00007442, 0.00004441, 0.00003644, 0.00000734, -0.00002906, -0.00003397, 0.00001210, -0.00000916, -0.00000152, 0.00002700, -0.00002416, 0.00000035, 0.00001316, -0.00005786, 0.00000475, 0.00003100, -0.00005096, 0.00000459, 0.00002117, -0.00003133, 0.00001406, -0.00005239, -0.00009211, 0.00002325, -0.00011991, -0.00022837, 0.00018889, 0.00050294, 0.00024565, -0.00008873, },
		{ 0.00004884, -0.00012925, -0.00011969, -0.00004147, 0.00002971, 0.00002132, -0.00003033, 0.00000167, 0.00002618, -0.00003681, -0.00001638, 0.00002053, -0.00003690, -0.00000015, 0.00004574, -0.00002515, -0.00001008, 0.00002280, -0.00001892, 0.00002041, 0.00001516, -0.00006297, 0.00000508, 0.00006257, -0.00002226, -0.00002862, 0.00003948, -0.00000813, -0.00006355, 0.00004388, 0.00007769, -0.00008519, -0.00004943, 0.00012784, 0.00000864, -0.00012086, 0.00000496, 0.00005548, 0.00000405, -0.00001087, -0.00003390, -0.00001269, 0.00000554, -0.00001020, 0.00001868, 0.00000189, -0.00001846, 0.00001518, -0.00003022, -0.00003241, 0.00003250, -0.00001843, -0.00002608, 0.00002445, -0.00001537, -0.00000131, -0.00001241, -0.00009435, -0.00001871, -0.00003493, -0.00021169, -0.00001502, 0.00038739, 0.00036606, },
		{ 0.00038831, 0.00018145, -0.00005553, -0.00012236, -0.00008265, -0.00000350, 0.00004224, -0.00001080, -0.00002841, 0.00003378, -0.00000624, -0.00004371, 0.00002232, -0.00001249, -0.00003884, 0.00004569, 0.00000979, -0.00003500, 0.00002327, -0.00000178, -0.00000688, 0.00003521, -0.00003831, -0.00004401, 0.00005676, 0.00002492, -0.00004532, 0.00000778, 0.00003490, -0.00004938, -0.00002968, 0.00009617, 0.00000195, -0.00011897, 0.00006645, 0.00011751, -0.00009476, -0.00006900, 0.00006349, 0.00001688, -0.00001563, -0.00001597, -0.00002915, -0.00000196, -0.00000140, 0.00000330, 0.00001510, -0.00001281, -0.00000001, -0.00000690, -0.00004058, 0.00000785, 0.00000877, -0.00002990, 0.00000824, 0.00000098, -0.00001019, 0.00000604, -0.00006808, -0.00005851, -0.00000459, -0.00014395, -0.00013010, 0.00021869, },
		{ 0.00005911, 0.00032664, 0.00027123, 0.00003953, -0.00009075, -0.00009854, -0.00004961, 0.00003143, 0.00002732, -0.00004056, 0.00000814, 0.00003223, -0.00004577, -0.00000519, 0.00002119, -0.00005173, 0.00000972, 0.00004746, -0.00003348, -0.00000179, 0.00002212, -0.00001842, 0.00002514, 0.00000131, -0.00006416, 0.00001505, 0.00005904, -0.00002150, -0.00003191, 0.00003625, 0.00000269, -0.00006870, 0.00003959, 0.00008838, -0.00009668, -0.00004985, 0.00015680, 0.00001139, -0.00012452, 0.00002537, 0.00005055, -0.00002512, -0.00001086, -0.00002088, -0.00002195, 0.00000286, 0.00000012, 0.00001072, 0.00000204, -0.00000859, -0.00000172, -0.00002949, -0.00001577, 0.00001381, -0.00001635, -0.00000661, 0.00000656, -0.00001113, 0.00000913, -0.00003425, -0.00007571, -0.00001377, -0.00007601, -0.00016105, },
		{ -0.00013694, -0.00005335, 0.00021518, 0.00030151, 0.00013191, -0.00003921, -0.00008503, -0.00007798, -0.00001270, 0.00005161, -0.00001522, -0.00003271, 0.00004473, -0.00001206, -0.00004072, 0.00003202, -0.00002716, -0.00003788, 0.00005524, 0.00000036, -0.00003134, 0.00002718, -0.00000517, 0.00000123, 0.00002637, -0.00004824, -0.00003216, 0.00005571, 0.00002194, -0.00004244, 0.00000006, 0.00004145, -0.00004420, -0.00004080, 0.00010213, 0.00000146, -0.00012869, 0.00008845, 0.00013317, -0.00010049, -0.00005335, 0.00007148, -0.00000944, -0.00002913, -0.00000396, -0.00002915, -0.00001001, 0.00000582, 0.00000357, 0.00000843, -0.00000371, -0.00000577, -0.00001818, -0.00002484, 0.00000342, -0.00000383, -0.00001071, 0.00000508, -0.00000917, 0.00000402, -0.00000646, -0.00006832, -0.00003814, -0.00003435, },
		{ -0.00002357, -0.00009230, -0.00010605, 0.00009249, 0.00027158, 0.00020490, 0.00002076, -0.00005842, -0.00007154, -0.00006023, 0.00003287, 0.00003296, -0.00005007, 0.00001489, 0.00003398, -0.00004941, 0.00000712, 0.00001348, -0.00006053, 0.00002213, 0.00004250, -0.00003656, 0.00000588, 0.00001713, -0.00001203, 0.00002554, -0.00001521, -0.00005624, 0.00002139, 0.00005052, -0.00001638, -0.00003571, 0.00003316, 0.00001424, -0.00007575, 0.00003493, 0.00009374, -0.00010613, -0.00003911, 0.00018011, 0.00000983, -0.00011545, 0.00004195, 0.00003264, -0.00004757, -0.00000624, -0.00001462, -0.00002813, 0.00000387, 0.00000495, 0.00000495, 0.00000462, -0.00000578, -0.00001434, -0.00002372, -0.00000819, -0.00000082, -0.00000855, 0.00000342, -0.00000592, -0.00000445, 0.00000835, -0.00004452, -0.00005680, },
		{ -0.00005816, -0.00003277, -0.00005537, -0.00010804, -0.00000673, 0.00019347, 0.00024315, 0.00008753, -0.00003409, -0.00004445, -0.00007337, -0.00002284, 0.00006087, -0.00002037, -0.00003456, 0.00005137, -0.00001693, -0.00003158, 0.00003373, -0.00004183, -0.00002902, 0.00005827, -0.00000840, -0.00002326, 0.00002483, -0.00000508, 0.00001045, 0.00001055, -0.00005099, -0.00001772, 0.00004685, 0.00002127, -0.00003769, -0.00000882, 0.00004945, -0.00004071, -0.00005166, 0.00010444, -0.00000340, -0.00012725, 0.00011311, 0.00013908, -0.00009733, -0.00003247, 0.00006463, -0.00003587, -0.00003322, 0.00000291, -0.00003075, -0.00001191, 0.00000965, 0.00000199, 0.00000648, 0.00000058, -0.00001296, -0.00002146, -0.00001444, -0.00000404, -0.00000727, 0.00000274, -0.00000039, -0.00001110, 0.00000879, -0.00001736, },
		{ 0.00000071, -0.00004299, -0.00004410, -0.00003938, -0.00008383, -0.00006059, 0.00009376, 0.00023146, 0.00015698, -0.00000669, -0.00002762, -0.00004344, -0.00007204, 0.00003511, 0.00003642, -0.00005698, 0.00002035, 0.00003190, -0.00004572, 0.00001645, -0.00000078, -0.00006032, 0.00003299, 0.00003247, -0.00003307, 0.00000979, 0.00001053, -0.00000029, 0.00001918, -0.00003019, -0.00004052, 0.00002107, 0.00004006, -0.00000742, -0.00004088, 0.00003128, 0.00002494, -0.00008329, 0.00003002, 0.00009134, -0.00011038, -0.00001765, 0.00019480, 0.00000889, -0.00009483, 0.00004786, 0.00000621, -0.00005953, -0.00000171, -0.00001413, -0.00002883, 0.00000545, 0.00000526, 0.00000299, 0.00000647, -0.00000803, -0.00002056, -0.00001732, -0.00000783, -0.00000856, 0.00000037, 0.00000596, -0.00001112, 0.00000005, },
		{ -0.00000845, 0.00000467, -0.00002239, -0.00004422, -0.00003951, -0.00006057, -0.00006972, 0.00000873, 0.00016748, 0.00020790, 0.00004292, -0.00003369, -0.00000314, -0.00007067, -0.00003122, 0.00006811, -0.00002528, -0.00003389, 0.00005240, -0.00001798, -0.00001889, 0.00002603, -0.00005184, -0.00001487, 0.00005349, -0.00001322, -0.00001369, 0.00001682, 0.00000052, 0.00001721, -0.00000966, -0.00004469, -0.00000510, 0.00003192, 0.00002444, -0.00003311, -0.00001758, 0.00005806, -0.00003915, -0.00006045, 0.00010060, -0.00001157, -0.00011258, 0.00013672, 0.00013799, -0.00008254, -0.00001373, 0.00004325, -0.00005587, -0.00003097, 0.00000275, -0.00003046, -0.00000958, 0.00000898, 0.00000124, 0.00000707, -0.00000039, -0.00001799, -0.00001965, -0.00001069, -0.00001019, -0.00000439, 0.00000905, -0.00000424, },
		{ 0.00000444, -0.00000974, -0.00000045, -0.00000863, -0.00003283, -0.00004064, -0.00005201, -0.00005654, -0.00003393, 0.00007534, 0.00020890, 0.00011815, -0.00003843, 0.00000335, -0.00001836, -0.00008265, 0.00003769, 0.00003705, -0.00005990, 0.00002262, 0.00002803, -0.00003472, 0.00001908, -0.00001763, -0.00005061, 0.00003856, 0.00002064, -0.00002384, 0.00000791, 0.00000568, 0.00001426, 0.00000562, -0.00003944, -0.00002056, 0.00001307, 0.00003101, 0.00000342, -0.00004729, 0.00003128, 0.00003321, -0.00008935, 0.00002394, 0.00007994, -0.00010639, 0.00001204, 0.00020060, 0.00001373, -0.00006786, 0.00003875, -0.00002175, -0.00006066, -0.00000144, -0.00001643, -0.00002384, 0.00000497, 0.00000339, 0.00000448, 0.00000538, -0.00001208, -0.00002076, -0.00001370, -0.00001111, -0.00000921, 0.00000674, },
		{ 0.00000101, 0.00000928, -0.00000430, -0.00000517, -0.00000528, -0.00002025, -0.00003256, -0.00005201, -0.00004813, -0.00003439, -0.00000135, 0.00014708, 0.00018721, -0.00000277, -0.00003274, 0.00003501, -0.00006964, -0.00003643, 0.00007166, -0.00002839, -0.00003184, 0.00004804, -0.00001306, -0.00000660, 0.00001019, -0.00005362, 0.00000039, 0.00004199, -0.00001212, -0.00000610, 0.00000565, 0.00001179, 0.00001837, -0.00003003, -0.00003003, 0.00000188, 0.00001439, 0.00003129, -0.00003039, -0.00002455, 0.00006573, -0.00003891, -0.00006611, 0.00008830, -0.00002081, -0.00008491, 0.00015653, 0.00013514, -0.00005724, -0.00000486, 0.00001194, -0.00006504, -0.00002724, -0.00000354, -0.00002588, -0.00000603, 0.00000474, 0.00000276, 0.00000744, -0.00000492, -0.00001895, -0.00001650, -0.00001232, -0.00001229, },
		{ -0.00001427, -0.00000490, 0.00000888, 0.00000235, -0.00000446, -0.00000622, -0.00001546, -0.00001925, -0.00004543, -0.00005446, -0.00002019, -0.00002904, 0.00004996, 0.00020083, 0.00008208, -0.00006869, 0.00003344, 0.00000126, -0.00008972, 0.00003916, 0.00003544, -0.00005853, 0.00002036, 0.00002518, -0.00001863, 0.00001303, -0.00003219, -0.00003355, 0.00003677, 0.00001066, -0.00001156, 0.00000015, 0.00000535, 0.00002786, -0.00001324, -0.00004022, -0.00000118, -0.00000128, 0.00002580, 0.00001349, -0.00005376, 0.00003301, 0.00003797, -0.00009193, 0.00001498, 0.00005997, -0.00009235, 0.00004635, 0.00020058, 0.00002761, -0.00004285, 0.00001449, -0.00004381, -0.00005411, -0.00000757, -0.00001782, -0.00001517, 0.00000134, 0.00000191, 0.00000738, 0.00000102, -0.00001474, -0.00001746, -0.00001412, },
		},
		{ { 0.00000924, 0.00000711, 0.00000223, -0.00001039, -0.00002662, -0.00001181, 0.00002478, 0.00001686, -0.00001040, -0.00000119, -0.00002143, -0.00004767, 0.00002538, 0.00004386, -0.00007478, -0.00005942, 0.00005206, -0.00002189, -0.00004069, 0.00014715, 0.00010889, -0.00010046, -0.00000908, 0.00012165, -0.00003799, -0.00010518, 0.00003682, 0.00003972, -0.00004470, -0.00001344, 0.00002492, 0.00000271, -0.00001277, 0.00000271, 0.00002492, -0.00001344, -0.00004470, 0.00003972, 0.00003682, -0.00010518, -0.00003799, 0.00012165, -0.00000908, -0.00010046, 0.00010889, 0.00014715, -0.00004069, -0.00002189, 0.00005206, -0.00005942, -0.00007478, 0.00004386, 0.00002538, -0.00004767, -0.00002143, -0.00000119, -0.00001040, 0.00001686, 0.00002478, -0.00001181, -0.00002662, -0.00001039, 0.00000223, 0.00000711, },
		{ 0.00000568, 0.00001088, 0.00000767, 0.00000307, 0.00000024, -0.00001486, -0.00002654, 0.00000290, 0.00002900, 0.00000245, -0.00000846, -0.00000102, -0.00004187, -0.00002519, 0.00005883, -0.00000447, -0.00009876, 0.00000972, 0.00005031, -0.00007507, 0.00001785, 0.00017650, 0.00000828, -0.00011148, 0.00006971, 0.00008629, -0.00009066, -0.00005497, 0.00005800, 0.00000653, -0.00004019, -0.00000020, 0.00002086, 0.00000273, -0.00001722, 0.00000904, 0.00002811, -0.00003737, -0.00002899, 0.00006751, -0.00001681, -0.00012175, 0.00003665, 0.00010104, -0.00007647, -0.00002688, 0.00017075, 0.00008088, -0.00004857, 0.00002342, 0.00000323, -0.00009304, -0.00001909, 0.00005216, -0.00001610, -0.00004281, -0.00000678, -0.00000889, -0.00000218, 0.00002669, 0.00001240, -0.00002143, -0.00002360, -0.00000710, },
		{ -0.00002040, -0.00000367, 0.00001054, 0.00001257, 0.00000368, 0.00000022, -0.00000156, -0.00001984, -0.00002054, 0.00001761, 0.00002290, -0.00000785, -0.00000177, -0.00001432, -0.00004924, 0.00001786, 0.00005581, -0.00006535, -0.00006745, 0.00006833, -0.00000489, -0.00008895, 0.00010027, 0.00014112, -0.00008242, -0.00006047, 0.00011419, 0.00001982, -0.00009918, -0.00000317, 0.00004993, -0.00001367, -0.00003041, 0.00000465, 0.00002389, -0.00000349, -0.00002319, 0.00002635, 0.00001647, -0.00005851, 0.00000912, 0.00006538, -0.00008087, -0.00008884, 0.00009184, 0.00003524, -0.00008992, 0.00007261, 0.00017309, 0.00002290, -0.00002007, 0.00002437, -0.00005715, -0.00007483, 0.00002703, 0.00002488, -0.00003842, -0.00002456, -0.00000636, -0.00001234, 0.00001202, 0.00002649, -0.00000167, -0.00002528, },
		{ -0.00001226, -0.00002601, -0.00001756, 0.00000121, 0.00001536, 0.00001138, -0.00000135, -0.00000116, -0.00000414, -0.00002283, -0.00000901, 0.00002627, 0.00001081, -0.00000983, -0.00000071, -0.00003319, -0.00003015, 0.00005460, 0.00001097, -0.00009659, 0.00000394, 0.00007412, -0.00007589, -0.00004280, 0.00015699, 0.00005368, -0.00011989, 0.00001620, 0.00010576, -0.00003855, -0.00007201, 0.00002711, 0.00003186, -0.00002046, -0.00002595, 0.00001247, 0.00002769, -0.00002101, -0.00001643, 0.00004383, -0.00001386, -0.00005740, 0.00005126, 0.00002399, -0.00012033, -0.00002258, 0.00009605, -0.00003390, -0.00003925, 0.00015060, 0.00013378, 0.00000026, 0.00000440, -0.00001740, -0.00008580, -0.00002675, 0.00003677, -0.00001038, -0.00003507, -0.00001207, -0.00001388, -0.00000543, 0.00002317, 0.00001789, },
		{ 0.00002571, 0.00000676, -0.00001877, -0.00002587, -0.00001382, 0.00000723, 0.00001881, 0.00000752, -0.00000646, -0.00000132, -0.00000699, -0.00002185, 0.00000420, 0.00002652, -0.00000031, -0.00000744, -0.00001018, -0.00004054, 0.00000905, 0.00005707, -0.00004960, -0.00007288, 0.00006961, 0.00002074, -0.00011182, 0.00004092, 0.00015227, -0.00004071, -0.00009511, 0.00007350, 0.00006088, -0.00006405, -0.00003539, 0.00003503, 0.00001925, -0.00002515, -0.00002341, 0.00002840, 0.00001900, -0.00003962, 0.00000830, 0.00004363, -0.00004878, -0.00002694, 0.00006962, -0.00003931, -0.00011294, 0.00003778, 0.00005143, -0.00006187, 0.00004850, 0.00018147, 0.00008931, 0.00000342, -0.00000626, -0.00006622, -0.00006821, 0.00001230, 0.00001482, -0.00002785, -0.00002016, -0.00001329, -0.00001729, 0.00000788, },
		{ -0.00001080, 0.00001907, 0.00002116, -0.00000280, -0.00002324, -0.00002464, -0.00000871, 0.00001328, 0.00002060, 0.00000101, -0.00001095, 0.00000017, -0.00000915, -0.00001734, 0.00001495, 0.00002090, -0.00000721, -0.00000833, -0.00002306, -0.00002482, 0.00004337, 0.00001925, -0.00008499, -0.00000660, 0.00008580, -0.00005678, -0.00008614, 0.00011317, 0.00008969, -0.00009666, -0.00003567, 0.00008819, 0.00001290, -0.00005935, -0.00000987, 0.00003488, 0.00001206, -0.00003493, -0.00001165, 0.00004340, -0.00000690, -0.00004184, 0.00003897, 0.00001802, -0.00006548, 0.00001792, 0.00004876, -0.00009052, -0.00006722, 0.00005770, -0.00000730, -0.00003109, 0.00013181, 0.00017346, 0.00006211, -0.00000001, -0.00004792, -0.00008374, -0.00002635, 0.00001841, -0.00001225, -0.00002275, -0.00001204, -0.00002139, },
		{ -0.00001692, -0.00002469, 0.00000135, 0.00002363, 0.00001430, -0.00001103, -0.00002634, -0.00002126, -0.00000349, 0.00001939, 0.00002033, -0.00000878, -0.00001299, 0.00000405, -0.00001104, -0.00001127, 0.00002198, 0.00001320, -0.00001262, -0.00001483, -0.00002515, 0.00000741, 0.00004720, -0.00003520, -0.00006961, 0.00005950, 0.00004254, -0.00010788, -0.00001488, 0.00013426, 0.00000720, -0.00009826, 0.00001834, 0.00006919, -0.00001494, -0.00004433, 0.00000307, 0.00003786, 0.00000017, -0.00004366, 0.00001335, 0.00004172, -0.00003816, -0.00001955, 0.00005452, -0.00002127, -0.00005141, 0.00004938, -0.00000063, -0.00010396, -0.00001686, 0.00003223, -0.00003799, 0.00004175, 0.00018457, 0.00015163, 0.00004390, -0.00003211, -0.00008653, -0.00006045, 0.00000341, -0.00000144, -0.00002141, -0.00000996, },
		{ -0.00001133, -0.00000648, -0.00002660, -0.00001885, 0.00001158, 0.00002375, 0.00000704, -0.00001937, -0.00002614, -0.00001697, 0.00000102, 0.00002691, 0.00001561, -0.00002127, -0.00000911, 0.00000913, -0.00001461, -0.00000420, 0.00002648, 0.00000381, -0.00002049, -0.00001958, -0.00000860, 0.00003361, 0.00001546, -0.00006870, -0.00001305, 0.00008299, -0.00003003, -0.00010289, 0.00005756, 0.00010029, -0.00005246, -0.00006425, 0.00004345, 0.00004247, -0.00002290, -0.00003469, 0.00001496, 0.00004113, -0.00002152, -0.00003823, 0.00004062, 0.00001664, -0.00005423, 0.00001671, 0.00004241, -0.00005082, -0.00001592, 0.00004824, -0.00004912, -0.00008152, 0.00000468, -0.00001212, -0.00001910, 0.00012429, 0.00020758, 0.00012913, 0.00000968, -0.00008193, -0.00008950, -0.00002111, 0.00000187, -0.00002084, },
		{ -0.00001936, -0.00002012, 0.00000114, -0.00001550, -0.00003002, -0.00000920, 0.00001760, 0.00002307, -0.00000285, -0.00002593, -0.00002158, -0.00001516, 0.00000726, 0.00003544, 0.00000244, -0.00003193, 0.00000324, 0.00001077, -0.00002044, 0.00000591, 0.00002813, -0.00001022, -0.00002875, -0.00001139, 0.00001810, 0.00003219, -0.00003014, -0.00005765, 0.00004652, 0.00005203, -0.00008523, -0.00005181, 0.00009213, 0.00004085, -0.00007080, -0.00002731, 0.00004401, 0.00002659, -0.00002623, -0.00002952, 0.00003357, 0.00003313, -0.00004440, -0.00001330, 0.00005152, -0.00002074, -0.00004293, 0.00004482, 0.00000975, -0.00005423, 0.00001670, 0.00001873, -0.00007147, -0.00004986, -0.00001159, -0.00003869, 0.00004188, 0.00019427, 0.00021342, 0.00009455, -0.00005119, -0.00011395, -0.00005517, -0.00000059, },
		{ -0.00001444, -0.00001235, -0.00003233, -0.00000312, 0.00000036, -0.00002636, -0.00002539, -0.00000234, 0.00002354, 0.00002019, -0.00001661, -0.00002573, -0.00001646, -0.00001701, 0.00002000, 0.00003952, -0.00002040, -0.00003233, 0.00002130, 0.00000310, -0.00002451, 0.00002069, 0.00002237, -0.00002944, -0.00002767, 0.00001255, 0.00003386, 0.00000115, -0.00005617, -0.00001048, 0.00007068, -0.00000823, -0.00009395, 0.00000768, 0.00008107, -0.00000734, -0.00006015, -0.00000730, 0.00003968, 0.00002008, -0.00003577, -0.00001698, 0.00005272, 0.00000833, -0.00005244, 0.00002032, 0.00003591, -0.00004858, -0.00001102, 0.00004852, -0.00002259, -0.00003489, 0.00002680, -0.00001617, -0.00006561, -0.00003824, -0.00004662, -0.00002550, 0.00012355, 0.00024818, 0.00020219, 0.00002818, -0.00011664, -0.00010083, },
		{ -0.00014562, -0.00005108, -0.00000273, -0.00003762, -0.00002075, 0.00000969, -0.00001262, -0.00002842, -0.00002050, 0.00000289, 0.00003235, 0.00000921, -0.00002898, -0.00001677, -0.00001810, -0.00001710, 0.00004059, 0.00002975, -0.00004629, -0.00001544, 0.00003511, -0.00001511, -0.00001909, 0.00003621, 0.00000348, -0.00004611, -0.00000765, 0.00003953, 0.00002152, -0.00003776, -0.00004308, 0.00003896, 0.00004841, -0.00005703, -0.00006258, 0.00004099, 0.00004705, -0.00002756, -0.00004586, -0.00000048, 0.00004333, 0.00000922, -0.00004674, 0.00000963, 0.00005781, -0.00002402, -0.00003718, 0.00004256, 0.00000182, -0.00005120, 0.00002115, 0.00002959, -0.00003720, -0.00001152, 0.00001432, -0.00003486, -0.00005170, -0.00005660, -0.00007082, 0.00002725, 0.00021123, 0.00028635, 0.00015457, -0.00006787, },
		{ 0.00005300, -0.00015620, -0.00011446, -0.00000548, -0.00002734, -0.00004460, 0.00000468, 0.00000357, -0.00002296, -0.00002368, -0.00002152, 0.00001448, 0.00003841, -0.00001219, -0.00002999, -0.00000580, -0.00002988, -0.00000515, 0.00006113, 0.00000023, -0.00006067, 0.00001662, 0.00003223, -0.00003586, 0.00000087, 0.00004164, -0.00002759, -0.00004566, 0.00002906, 0.00004767, -0.00001602, -0.00005452, -0.00000055, 0.00005759, 0.00000162, -0.00006905, -0.00002244, 0.00004061, 0.00001965, -0.00002957, -0.00004219, 0.00000931, 0.00004991, -0.00001394, -0.00004466, 0.00004216, 0.00004058, -0.00004588, -0.00000682, 0.00003997, -0.00003022, -0.00003072, 0.00003601, 0.00000422, -0.00003310, -0.00000104, -0.00000412, -0.00003439, -0.00005338, -0.00009172, -0.00006456, 0.00011007, 0.00030059, 0.00029256, },
		{ 0.00038050, 0.00023450, -0.00008722, -0.00018536, -0.00004361, -0.00000303, -0.00005974, -0.00001977, 0.00001754, -0.00001486, -0.00001969, -0.00002453, -0.00002272, 0.00003621, 0.00002983, -0.00003517, -0.00001460, -0.00000611, -0.00004408, 0.00002498, 0.00006545, -0.00004282, -0.00004906, 0.00004927, 0.00000739, -0.00004423, 0.00003012, 0.00002568, -0.00005770, -0.00001741, 0.00006386, 0.00002297, -0.00005453, -0.00003433, 0.00004012, 0.00003765, -0.00003482, -0.00005039, -0.00000185, 0.00002544, 0.00001101, -0.00003365, -0.00004286, 0.00003056, 0.00004429, -0.00004136, -0.00002122, 0.00006316, 0.00000841, -0.00004535, 0.00001951, 0.00001638, -0.00004311, -0.00000351, 0.00003119, -0.00001212, -0.00002314, -0.00000497, -0.00001220, -0.00003005, -0.00007923, -0.00012179, -0.00002259, 0.00021925, },
		{ 0.00005830, 0.00035358, 0.00041552, 0.00009241, -0.00020913, -0.00013246, 0.00001195, -0.00004524, -0.00006099, 0.00001757, 0.00000277, -0.00002411, -0.00000968, -0.00003832, -0.00000964, 0.00005804, 0.00000071, -0.00004273, 0.00000840, -0.00002635, -0.00004339, 0.00006542, 0.00003962, -0.00007930, -0.00000913, 0.00006179, -0.00003000, -0.00002854, 0.00005199, -0.00001241, -0.00006577, 0.00003304, 0.00007140, -0.00002638, -0.00006520, 0.00000985, 0.00005072, 0.00000035, -0.00004112, -0.00002816, -0.00000485, 0.00001929, 0.00001026, -0.00004886, -0.00003188, 0.00005436, 0.00001856, -0.00005529, 0.00001597, 0.00006073, -0.00002087, -0.00002653, 0.00002685, -0.00001121, -0.00003510, 0.00001430, 0.00001758, -0.00001661, -0.00001963, -0.00001159, -0.00000841, -0.00003961, -0.00012055, -0.00013178, },
		{ -0.00016410, -0.00011285, 0.00018926, 0.00049222, 0.00035420, -0.00010594, -0.00024463, -0.00003096, 0.00000513, -0.00009102, -0.00001862, 0.00003182, -0.00002635, -0.00001084, -0.00001279, -0.00005456, 0.00002529, 0.00005976, -0.00003787, -0.00002345, 0.00001842, -0.00005903, -0.00001344, 0.00009373, -0.00001461, -0.00008527, 0.00004277, 0.00004123, -0.00005832, 0.00000832, 0.00004772, -0.00005621, -0.00003699, 0.00008112, 0.00003827, -0.00007179, -0.00003709, 0.00004684, 0.00002788, -0.00002364, -0.00002581, -0.00002485, -0.00001280, 0.00002898, -0.00000109, -0.00006507, -0.00000247, 0.00006260, -0.00001816, -0.00004409, 0.00004843, 0.00003841, -0.00003329, -0.00000527, 0.00001524, -0.00002686, -0.00001777, 0.00001833, 0.00000678, -0.00001732, -0.00002318, -0.00001014, -0.00000475, -0.00006892, },
		{ -0.00001296, -0.00011301, -0.00020221, -0.00004823, 0.00037757, 0.00057771, 0.00016823, -0.00027979, -0.00016666, 0.00004495, -0.00006161, -0.00008969, 0.00003810, 0.00000483, -0.00003843, 0.00001309, -0.00004087, -0.00004959, 0.00006914, 0.00002805, -0.00006060, 0.00001468, -0.00000205, -0.00008028, 0.00004228, 0.00008561, -0.00007465, -0.00004983, 0.00007808, -0.00000577, -0.00005716, 0.00004631, 0.00001149, -0.00007819, 0.00002270, 0.00009604, -0.00002346, -0.00008307, 0.00001297, 0.00005151, -0.00000571, -0.00001987, -0.00001468, -0.00003965, -0.00000494, 0.00004041, -0.00002907, -0.00006415, 0.00003417, 0.00004511, -0.00004600, -0.00001252, 0.00006139, 0.00001132, -0.00002890, 0.00000440, -0.00000282, -0.00002683, -0.00000391, 0.00001503, 0.00000115, -0.00002151, -0.00002531, -0.00000180, },
		{ -0.00002063, 0.00000396, -0.00003477, -0.00016875, -0.00022533, 0.00009286, 0.00059629, 0.00053216, -0.00010506, -0.00033644, -0.00002117, 0.00003356, -0.00013220, -0.00002615, 0.00006017, -0.00004358, -0.00001384, 0.00001734, -0.00008037, -0.00000766, 0.00009206, -0.00002714, -0.00004600, 0.00004435, -0.00005017, -0.00006484, 0.00009813, 0.00003367, -0.00010745, 0.00001356, 0.00007375, -0.00005370, -0.00002317, 0.00005939, -0.00004054, -0.00005736, 0.00008451, 0.00006211, -0.00008042, -0.00005090, 0.00005201, 0.00002436, -0.00002131, -0.00000064, -0.00002628, -0.00005165, 0.00002157, 0.00003400, -0.00006049, -0.00003830, 0.00005730, 0.00000952, -0.00005004, 0.00002248, 0.00005402, -0.00000685, -0.00001898, 0.00000048, -0.00001449, -0.00001855, 0.00000233, 0.00001146, -0.00000419, -0.00002748, },
		{ -0.00001161, -0.00002897, -0.00001384, 0.00000231, -0.00006798, -0.00024002, -0.00020322, 0.00033381, 0.00076461, 0.00030938, -0.00035798, -0.00023660, 0.00009346, -0.00006322, -0.00013829, 0.00006054, 0.00002146, -0.00006870, 0.00003254, -0.00001937, -0.00009682, 0.00005931, 0.00006911, -0.00007311, 0.00000351, 0.00003657, -0.00009854, -0.00000681, 0.00011945, -0.00004215, -0.00009062, 0.00007165, 0.00002923, -0.00007360, 0.00002430, 0.00003431, -0.00007650, 0.00000334, 0.00011201, -0.00000739, -0.00009915, 0.00000494, 0.00005571, -0.00000958, -0.00000822, 0.00000739, -0.00005546, -0.00004033, 0.00004916, 0.00000400, -0.00007444, 0.00000149, 0.00005367, -0.00002448, -0.00003095, 0.00004531, 0.00003706, -0.00001354, -0.00001455, -0.00000863, -0.00001618, -0.00001068, 0.00000434, 0.00000808, },
		{ 0.00000475, 0.00000212, -0.00001635, -0.00002629, -0.00001231, -0.00000176, -0.00012122, -0.00032176, -0.00008057, 0.00064879, 0.00077041, -0.00005062, -0.00046406, -0.00003481, 0.00009211, -0.00017376, -0.00005566, 0.00010370, -0.00005426, -0.00003900, 0.00005818, -0.00008473, -0.00006150, 0.00011292, 0.00000376, -0.00007563, 0.00005717, -0.00001566, -0.00011083, 0.00007028, 0.00008625, -0.00010291, -0.00002963, 0.00009133, -0.00003073, -0.00005321, 0.00005435, -0.00001675, -0.00006898, 0.00007399, 0.00008511, -0.00007565, -0.00006949, 0.00004916, 0.00002678, -0.00002166, 0.00001773, -0.00001272, -0.00007819, -0.00000454, 0.00005573, -0.00003539, -0.00006053, 0.00003365, 0.00002764, -0.00004021, -0.00000204, 0.00005178, 0.00002162, -0.00001561, -0.00001757, -0.00001409, -0.00001243, -0.00000570, },
		{ -0.00000783, -0.00000218, 0.00000137, -0.00000388, -0.00001393, -0.00002918, -0.00001234, -0.00000621, -0.00021542, -0.00037320, 0.00019203, 0.00093583, 0.00053721, -0.00041333, -0.00035994, 0.00014308, -0.00003218, -0.00020586, 0.00007005, 0.00006225, -0.00010710, 0.00003263, 0.00002691, -0.00013461, 0.00002082, 0.00011365, -0.00006775, -0.00002614, 0.00007437, -0.00008691, -0.00006783, 0.00012464, 0.00001008, -0.00011453, 0.00004274, 0.00006173, -0.00007128, -0.00000690, 0.00004610, -0.00006113, -0.00001560, 0.00011413, 0.00001544, -0.00010558, -0.00001209, 0.00005665, -0.00000795, -0.00000307, 0.00002813, -0.00005318, -0.00007242, 0.00003605, 0.00003373, -0.00006162, -0.00002735, 0.00004266, -0.00000364, -0.00003455, 0.00002264, 0.00004788, 0.00001079, -0.00001996, -0.00002252, -0.00001391, },
		{ -0.00002448, -0.00001085, -0.00000245, -0.00000184, -0.00000703, -0.00000100, -0.00001140, -0.00004426, 0.00000202, -0.00003153, -0.00035835, -0.00031160, 0.00060154, 0.00104267, 0.00009540, -0.00061009, -0.00009978, 0.00017879, -0.00019552, -0.00011696, 0.00015206, -0.00004218, -0.00008738, 0.00009268, -0.00005704, -0.00012383, 0.00010753, 0.00005241, -0.00009850, 0.00004701, 0.00003419, -0.00013280, 0.00001526, 0.00012300, -0.00006990, -0.00007073, 0.00008628, 0.00000195, -0.00007099, 0.00003356, 0.00000507, -0.00006682, 0.00005591, 0.00009887, -0.00005928, -0.00008442, 0.00003637, 0.00003037, -0.00001933, 0.00002826, 0.00000601, -0.00008750, -0.00003648, 0.00005708, -0.00000397, -0.00006164, 0.00000502, 0.00002916, -0.00002392, -0.00001626, 0.00003745, 0.00004068, 0.00000098, -0.00002692, },
		{ -0.00001050, -0.00003256, -0.00002380, -0.00000557, 0.00000376, -0.00001045, -0.00001222, 0.00001441, -0.00002660, -0.00005823, 0.00003814, -0.00012097, -0.00050627, -0.00005123, 0.00103300, 0.00084857, -0.00039972, -0.00053882, 0.00016443, 0.00004416, -0.00027770, 0.00004828, 0.00012665, -0.00013604, 0.00000179, 0.00008616, -0.00014465, -0.00004308, 0.00014463, -0.00003757, -0.00006464, 0.00009484, -0.00004642, -0.00011982, 0.00009540, 0.00006403, -0.00011087, 0.00000181, 0.00007926, -0.00005186, -0.00003624, 0.00004174, -0.00003898, -0.00002564, 0.00010391, 0.00003615, -0.00009953, -0.00003147, 0.00004960, -0.00000252, 0.00000096, 0.00004258, -0.00004003, -0.00009167, 0.00000984, 0.00004835, -0.00003513, -0.00004057, 0.00002058, 0.00000632, -0.00002779, 0.00000402, 0.00004440, 0.00003190, },
		{ 0.00004607, 0.00001990, -0.00002142, -0.00003567, -0.00002210, 0.00000497, 0.00000413, -0.00002843, 0.00000113, 0.00002745, -0.00006890, -0.00003923, 0.00006713, -0.00030559, -0.00055092, 0.00042328, 0.00129329, 0.00035676, -0.00073298, -0.00023830, 0.00027308, -0.00017363, -0.00020748, 0.00018465, 0.00000388, -0.00014696, 0.00010129, 0.00000132, -0.00017417, 0.00006823, 0.00010567, -0.00010036, 0.00001476, 0.00008126, -0.00012105, -0.00004970, 0.00012883, -0.00001922, -0.00009393, 0.00006088, 0.00003160, -0.00007138, 0.00000465, 0.00001595, -0.00005268, 0.00003924, 0.00010005, -0.00003806, -0.00008959, 0.00001702, 0.00002953, -0.00001292, 0.00003443, 0.00002233, -0.00008252, -0.00006233, 0.00004127, 0.00001987, -0.00004528, -0.00001517, 0.00001716, -0.00001210, -0.00001968, 0.00002119, },
		{ -0.00000587, 0.00003458, 0.00004162, 0.00000588, -0.00002875, -0.00003947, -0.00001570, 0.00002005, -0.00001249, -0.00004094, 0.00003840, 0.00000824, -0.00011680, 0.00003372, 0.00002122, -0.00055133, -0.00036292, 0.00099929, 0.00120547, -0.00027068, -0.00074901, 0.00012174, 0.00016212, -0.00032710, -0.00001664, 0.00019995, -0.00013534, -0.00005920, 0.00013704, -0.00011612, -0.00011665, 0.00014572, 0.00001324, -0.00009565, 0.00008862, 0.00000939, -0.00014439, 0.00004012, 0.00009990, -0.00008075, -0.00003535, 0.00007634, -0.00002421, -0.00005424, 0.00002328, -0.00001965, -0.00002345, 0.00008804, 0.00004810, -0.00008399, -0.00004596, 0.00003443, 0.00000209, 0.00000737, 0.00005115, -0.00002425, -0.00009630, -0.00001772, 0.00004550, -0.00000838, -0.00003625, 0.00000013, 0.00000359, -0.00002026, },
		{ -0.00001020, -0.00001920, 0.00001059, 0.00004239, 0.00003031, -0.00000490, -0.00003503, -0.00004487, 0.00000306, 0.00002722, -0.00004770, -0.00002294, 0.00007748, -0.00006152, -0.00011906, 0.00013165, -0.00016393, -0.00073348, 0.00011818, 0.00145045, 0.00072512, -0.00077556, -0.00045147, 0.00033762, -0.00009283, -0.00030784, 0.00017859, 0.00008174, -0.00019554, 0.00007176, 0.00007488, -0.00019227, 0.00000038, 0.00014425, -0.00007469, -0.00002749, 0.00010832, -0.00007956, -0.00010274, 0.00010112, 0.00003043, -0.00009137, 0.00002629, 0.00004783, -0.00005701, -0.00002225, 0.00001233, -0.00003372, 0.00003077, 0.00009140, -0.00001960, -0.00008404, -0.00000258, 0.00002146, -0.00000493, 0.00004028, 0.00003350, -0.00007018, -0.00007623, 0.00001775, 0.00002920, -0.00002233, -0.00002060, 0.00000187, },
		{ -0.00000961, -0.00000496, -0.00001960, -0.00001059, 0.00002755, 0.00004049, 0.00001703, -0.00000974, -0.00004747, -0.00004312, 0.00003418, 0.00000711, -0.00008037, 0.00003747, 0.00007310, -0.00015748, -0.00003195, 0.00016569, -0.00047991, -0.00067849, 0.00080323, 0.00153829, -0.00000061, -0.00093915, -0.00001117, 0.00029587, -0.00032499, -0.00012064, 0.00025656, -0.00009265, -0.00013464, 0.00015749, -0.00005257, -0.00017670, 0.00010998, 0.00006936, -0.00010431, 0.00005855, 0.00006081, -0.00013427, -0.00002217, 0.00010540, -0.00003671, -0.00005542, 0.00005693, 0.00000052, -0.00005741, -0.00000184, -0.00001045, -0.00001196, 0.00007426, 0.00004987, -0.00006535, -0.00005159, 0.00001520, 0.00000185, 0.00001599, 0.00005696, -0.00001091, -0.00009044, -0.00003812, 0.00003095, 0.00000880, -0.00002157, },
		{ -0.00000370, -0.00001449, -0.00000642, -0.00001491, -0.00002337, 0.00000565, 0.00003904, 0.00002821, 0.00001092, -0.00001549, -0.00006746, -0.00001959, 0.00005951, -0.00004701, -0.00007272, 0.00011477, -0.00000950, -0.00020965, 0.00012795, 0.00003263, -0.00080196, -0.00026773, 0.00145009, 0.00115199, -0.00068433, -0.00070929, 0.00033013, 0.00004183, -0.00038589, 0.00012044, 0.00017297, -0.00020999, 0.00000670, 0.00013916, -0.00016787, -0.00007772, 0.00015149, -0.00002738, -0.00006215, 0.00010737, -0.00002408, -0.00012732, 0.00005092, 0.00006137, -0.00006688, -0.00000345, 0.00004762, -0.00003611, -0.00003958, -0.00000297, -0.00001826, 0.00003198, 0.00007923, -0.00000837, -0.00007147, -0.00001701, 0.00000757, 0.00000081, 0.00004741, 0.00004074, -0.00005653, -0.00007892, -0.00000412, 0.00002572, },
		{ 0.00001337, 0.00001444, -0.00000709, -0.00000819, -0.00000899, -0.00002592, -0.00001856, 0.00002693, 0.00003676, 0.00001434, 0.00001535, -0.00003524, -0.00008105, 0.00002979, 0.00004869, -0.00011058, -0.00000002, 0.00014794, -0.00015246, -0.00014472, 0.00026109, -0.00030028, -0.00092134, 0.00045635, 0.00176396, 0.00039742, -0.00104317, -0.00023452, 0.00040183, -0.00025407, -0.00024064, 0.00027024, -0.00001172, -0.00019895, 0.00013531, 0.00002772, -0.00020240, 0.00004475, 0.00011060, -0.00008524, 0.00001838, 0.00009134, -0.00009619, -0.00007059, 0.00008130, 0.00000361, -0.00005400, 0.00003167, 0.00001475, -0.00004890, -0.00002507, -0.00001331, 0.00000211, 0.00006721, 0.00004491, -0.00004934, -0.00004901, -0.00000287, -0.00000404, 0.00002414, 0.00006234, -0.00000094, -0.00007975, -0.00004898, },
		{ -0.00007465, -0.00001967, 0.00001641, 0.00000577, -0.00000612, -0.00000406, -0.00001669, -0.00003556, -0.00000084, 0.00004315, 0.00001906, 0.00001339, 0.00001903, -0.00007351, -0.00006402, 0.00008334, -0.00001559, -0.00013232, 0.00011613, 0.00007936, -0.00027347, 0.00004696, 0.00022741, -0.00072937, -0.00065665, 0.00126129, 0.00155633, -0.00043271, -0.00095285, 0.00022156, 0.00020062, -0.00040880, 0.00001553, 0.00024844, -0.00017758, -0.00007542, 0.00017097, -0.00010654, -0.00014169, 0.00012186, 0.00002504, -0.00007501, 0.00008320, 0.00002568, -0.00011931, -0.00000253, 0.00006518, -0.00003345, -0.00001861, 0.00003593, -0.00001736, -0.00004544, -0.00002326, -0.00001153, 0.00003927, 0.00006940, -0.00000410, -0.00005716, -0.00002447, -0.00000816, 0.00000224, 0.00005455, 0.00004612, -0.00004418, },
		{ 0.00000686, -0.00006771, -0.00005272, -0.00000127, 0.00001344, 0.00000025, -0.00000312, -0.00000205, -0.00003126, -0.00003618, 0.00002800, 0.00004050, -0.00000146, 0.00003153, 0.00000076, -0.00011276, -0.00000355, 0.00009869, -0.00011133, -0.00006892, 0.00020326, -0.00009093, -0.00026838, 0.00027124, -0.00005712, -0.00102934, 0.00001582, 0.00181334, 0.00086652, -0.00100285, -0.00051401, 0.00043329, -0.00012044, -0.00034079, 0.00022680, 0.00008639, -0.00022754, 0.00007486, 0.00009866, -0.00018482, -0.00003017, 0.00012127, -0.00004580, -0.00001444, 0.00009518, -0.00004689, -0.00009186, 0.00004041, 0.00002619, -0.00003706, 0.00001164, 0.00001780, -0.00003550, -0.00004112, -0.00002500, 0.00001232, 0.00006666, 0.00003848, -0.00003814, -0.00004194, -0.00001657, -0.00001319, 0.00002943, 0.00006715, },
		{ 0.00005959, 0.00005002, -0.00003290, -0.00006699, -0.00002959, 0.00000778, 0.00000885, -0.00000404, 0.00000365, -0.00000655, -0.00005040, -0.00001786, 0.00005396, 0.00001436, -0.00000317, 0.00005436, -0.00005111, -0.00011803, 0.00008014, 0.00004273, -0.00017690, 0.00007136, 0.00017864, -0.00028018, -0.00009035, 0.00036229, -0.00053100, -0.00096302, 0.00090207, 0.00184717, -0.00003781, -0.00111037, 0.00001262, 0.00033604, -0.00035821, -0.00011059, 0.00028020, -0.00010354, -0.00014727, 0.00015903, -0.00002833, -0.00017155, 0.00006449, 0.00006292, -0.00006251, 0.00005017, 0.00005545, -0.00008773, -0.00004112, 0.00004596, -0.00000612, -0.00001839, 0.00002183, -0.00000522, -0.00004300, -0.00004216, -0.00001393, 0.00004696, 0.00006422, -0.00000321, -0.00004456, -0.00002681, -0.00002217, 0.00000035, },
		{ -0.00002236, 0.00003196, 0.00006961, 0.00001295, -0.00005473, -0.00005261, -0.00001232, 0.00001253, 0.00000107, -0.00000428, 0.00001351, -0.00002461, -0.00006199, 0.00002048, 0.00005624, -0.00002016, 0.00002633, 0.00005137, -0.00011959, -0.00006295, 0.00013499, -0.00007320, -0.00014786, 0.00021381, 0.00001455, -0.00035966, 0.00019009, 0.00018392, -0.00097737, -0.00043059, 0.00165692, 0.00131790, -0.00079187, -0.00078619, 0.00036002, 0.00004384, -0.00038647, 0.00013269, 0.00017080, -0.00020831, -0.00000373, 0.00013752, -0.00013146, -0.00009081, 0.00009799, -0.00000272, -0.00002767, 0.00007883, -0.00000440, -0.00008545, 0.00000051, 0.00002716, -0.00001681, 0.00000229, 0.00001490, -0.00002286, -0.00004945, -0.00004014, 0.00001565, 0.00006884, 0.00003414, -0.00003017, -0.00003400, -0.00002614, },
		{ -0.00002711, -0.00003367, -0.00000208, 0.00006207, 0.00005084, -0.00002227, -0.00005661, -0.00003693, 0.00000163, 0.00001248, -0.00001016, 0.00000598, 0.00001712, -0.00005603, -0.00004895, 0.00006225, 0.00002417, -0.00003327, 0.00007291, -0.00000049, -0.00016059, 0.00004075, 0.00010941, -0.00018456, -0.00001034, 0.00025441, -0.00022054, -0.00024113, 0.00039971, -0.00025981, -0.00112191, 0.00043783, 0.00195312, 0.00043783, -0.00112191, -0.00025981, 0.00039971, -0.00024113, -0.00022054, 0.00025441, -0.00001034, -0.00018456, 0.00010941, 0.00004075, -0.00016059, -0.00000049, 0.00007291, -0.00003327, 0.00002417, 0.00006225, -0.00004895, -0.00005603, 0.00001712, 0.00000598, -0.00001016, 0.00001248, 0.00000163, -0.00003693, -0.00005661, -0.00002227, 0.00005084, 0.00006207, -0.00000208, -0.00003367, },
		{ -0.00002236, -0.00002614, -0.00003400, -0.00003017, 0.00003414, 0.00006884, 0.00001565, -0.00004014, -0.00004945, -0.00002286, 0.00001490, 0.00000229, -0.00001681, 0.00002716, 0.00000051, -0.00008545, -0.00000440, 0.00007883, -0.00002767, -0.00000272, 0.00009799, -0.00009081, -0.00013146, 0.00013752, -0.00000373, -0.00020831, 0.00017080, 0.00013269, -0.00038647, 0.00004384, 0.00036002, -0.00078619, -0.00079187, 0.00131790, 0.00165692, -0.00043059, -0.00097737, 0.00018392, 0.00019009, -0.00035966, 0.00001455, 0.00021381, -0.00014786, -0.00007320, 0.00013499, -0.00006295, -0.00011959, 0.00005137, 0.00002633, -0.00002016, 0.00005624, 0.00002048, -0.00006199, -0.00002461, 0.00001351, -0.00000428, 0.00000107, 0.00001253, -0.00001232, -0.00005261, -0.00005473, 0.00001295, 0.00006961, 0.00003196, },
		{ 0.00005959, 0.00000035, -0.00002217, -0.00002681, -0.00004456, -0.00000321, 0.00006422, 0.00004696, -0.00001393, -0.00004216, -0.00004300, -0.00000522, 0.00002183, -0.00001839, -0.00000612, 0.00004596, -0.00004112, -0.00008773, 0.00005545, 0.00005017, -0.00006251, 0.00006292, 0.00006449, -0.00017155, -0.00002833, 0.00015903, -0.00014727, -0.00010354, 0.00028020, -0.00011059, -0.00035821, 0.00033604, 0.00001262, -0.00111037, -0.00003781, 0.00184717, 0.00090207, -0.00096302, -0.00053100, 0.00036229, -0.00009035, -0.00028018, 0.00017864, 0.00007136, -0.00017690, 0.00004273, 0.00008014, -0.00011803, -0.00005111, 0.00005436, -0.00000317, 0.00001436, 0.00005396, -0.00001786, -0.00005040, -0.00000655, 0.00000365, -0.00000404, 0.00000885, 0.00000778, -0.00002959, -0.00006699, -0.00003290, 0.00005002, },
		{ 0.00000686, 0.00006715, 0.00002943, -0.00001319, -0.00001657, -0.00004194, -0.00003814, 0.00003848, 0.00006666, 0.00001232, -0.00002500, -0.00004112, -0.00003550, 0.00001780, 0.00001164, -0.00003706, 0.00002619, 0.00004041, -0.00009186, -0.00004689, 0.00009518, -0.00001444, -0.00004580, 0.00012127, -0.00003017, -0.00018482, 0.00009866, 0.00007486, -0.00022754, 0.00008639, 0.00022680, -0.00034079, -0.00012044, 0.00043329, -0.00051401, -0.00100285, 0.00086652, 0.00181334, 0.00001582, -0.00102934, -0.00005712, 0.00027124, -0.00026838, -0.00009093, 0.00020326, -0.00006892, -0.00011133, 0.00009869, -0.00000355, -0.00011276, 0.00000076, 0.00003153, -0.00000146, 0.00004050, 0.00002800, -0.00003618, -0.00003126, -0.00000205, -0.00000312, 0.00000025, 0.00001344, -0.00000127, -0.00005272, -0.00006771, },
		{ -0.00007465, -0.00004418, 0.00004612, 0.00005455, 0.00000224, -0.00000816, -0.00002447, -0.00005716, -0.00000410, 0.00006940, 0.00003927, -0.00001153, -0.00002326, -0.00004544, -0.00001736, 0.00003593, -0.00001861, -0.00003345, 0.00006518, -0.00000253, -0.00011931, 0.00002568, 0.00008320, -0.00007501, 0.00002504, 0.00012186, -0.00014169, -0.00010654, 0.00017097, -0.00007542, -0.00017758, 0.00024844, 0.00001553, -0.00040880, 0.00020062, 0.00022156, -0.00095285, -0.00043271, 0.00155633, 0.00126129, -0.00065665, -0.00072937, 0.00022741, 0.00004696, -0.00027347, 0.00007936, 0.00011613, -0.00013232, -0.00001559, 0.00008334, -0.00006402, -0.00007351, 0.00001903, 0.00001339, 0.00001906, 0.00004315, -0.00000084, -0.00003556, -0.00001669, -0.00000406, -0.00000612, 0.00000577, 0.00001641, -0.00001967, },
		{ 0.00001337, -0.00004898, -0.00007975, -0.00000094, 0.00006234, 0.00002414, -0.00000404, -0.00000287, -0.00004901, -0.00004934, 0.00004491, 0.00006721, 0.00000211, -0.00001331, -0.00002507, -0.00004890, 0.00001475, 0.00003167, -0.00005400, 0.00000361, 0.00008130, -0.00007059, -0.00009619, 0.00009134, 0.00001838, -0.00008524, 0.00011060, 0.00004475, -0.00020240, 0.00002772, 0.00013531, -0.00019895, -0.00001172, 0.00027024, -0.00024064, -0.00025407, 0.00040183, -0.00023452, -0.00104317, 0.00039742, 0.00176396, 0.00045635, -0.00092134, -0.00030028, 0.00026109, -0.00014472, -0.00015246, 0.00014794, -0.00000002, -0.00011058, 0.00004869, 0.00002979, -0.00008105, -0.00003524, 0.00001535, 0.00001434, 0.00003676, 0.00002693, -0.00001856, -0.00002592, -0.00000899, -0.00000819, -0.00000709, 0.00001444, },
		{ -0.00000370, 0.00002572, -0.00000412, -0.00007892, -0.00005653, 0.00004074, 0.00004741, 0.00000081, 0.00000757, -0.00001701, -0.00007147, -0.00000837, 0.00007923, 0.00003198, -0.00001826, -0.00000297, -0.00003958, -0.00003611, 0.00004762, -0.00000345, -0.00006688, 0.00006137, 0.00005092, -0.00012732, -0.00002408, 0.00010737, -0.00006215, -0.00002738, 0.00015149, -0.00007772, -0.00016787, 0.00013916, 0.00000670, -0.00020999, 0.00017297, 0.00012044, -0.00038589, 0.00004183, 0.00033013, -0.00070929, -0.00068433, 0.00115199, 0.00145009, -0.00026773, -0.00080196, 0.00003263, 0.00012795, -0.00020965, -0.00000950, 0.00011477, -0.00007272, -0.00004701, 0.00005951, -0.00001959, -0.00006746, -0.00001549, 0.00001092, 0.00002821, 0.00003904, 0.00000565, -0.00002337, -0.00001491, -0.00000642, -0.00001449, },
		{ -0.00000961, -0.00002157, 0.00000880, 0.00003095, -0.00003812, -0.00009044, -0.00001091, 0.00005696, 0.00001599, 0.00000185, 0.00001520, -0.00005159, -0.00006535, 0.00004987, 0.00007426, -0.00001196, -0.00001045, -0.00000184, -0.00005741, 0.00000052, 0.00005693, -0.00005542, -0.00003671, 0.00010540, -0.00002217, -0.00013427, 0.00006081, 0.00005855, -0.00010431, 0.00006936, 0.00010998, -0.00017670, -0.00005257, 0.00015749, -0.00013464, -0.00009265, 0.00025656, -0.00012064, -0.00032499, 0.00029587, -0.00001117, -0.00093915, -0.00000061, 0.00153829, 0.00080323, -0.00067849, -0.00047991, 0.00016569, -0.00003195, -0.00015748, 0.00007310, 0.00003747, -0.00008037, 0.00000711, 0.00003418, -0.00004312, -0.00004747, -0.00000974, 0.00001703, 0.00004049, 0.00002755, -0.00001059, -0.00001960, -0.00000496, },
		{ -0.00001020, 0.00000187, -0.00002060, -0.00002233, 0.00002920, 0.00001775, -0.00007623, -0.00007018, 0.00003350, 0.00004028, -0.00000493, 0.00002146, -0.00000258, -0.00008404, -0.00001960, 0.00009140, 0.00003077, -0.00003372, 0.00001233, -0.00002225, -0.00005701, 0.00004783, 0.00002629, -0.00009137, 0.00003043, 0.00010112, -0.00010274, -0.00007956, 0.00010832, -0.00002749, -0.00007469, 0.00014425, 0.00000038, -0.00019227, 0.00007488, 0.00007176, -0.00019554, 0.00008174, 0.00017859, -0.00030784, -0.00009283, 0.00033762, -0.00045147, -0.00077556, 0.00072512, 0.00145045, 0.00011818, -0.00073348, -0.00016393, 0.00013165, -0.00011906, -0.00006152, 0.00007748, -0.00002294, -0.00004770, 0.00002722, 0.00000306, -0.00004487, -0.00003503, -0.00000490, 0.00003031, 0.00004239, 0.00001059, -0.00001920, },
		{ -0.00000587, -0.00002026, 0.00000359, 0.00000013, -0.00003625, -0.00000838, 0.00004550, -0.00001772, -0.00009630, -0.00002425, 0.00005115, 0.00000737, 0.00000209, 0.00003443, -0.00004596, -0.00008399, 0.00004810, 0.00008804, -0.00002345, -0.00001965, 0.00002328, -0.00005424, -0.00002421, 0.00007634, -0.00003535, -0.00008075, 0.00009990, 0.00004012, -0.00014439, 0.00000939, 0.00008862, -0.00009565, 0.00001324, 0.00014572, -0.00011665, -0.00011612, 0.00013704, -0.00005920, -0.00013534, 0.00019995, -0.00001664, -0.00032710, 0.00016212, 0.00012174, -0.00074901, -0.00027068, 0.00120547, 0.00099929, -0.00036292, -0.00055133, 0.00002122, 0.00003372, -0.00011680, 0.00000824, 0.00003840, -0.00004094, -0.00001249, 0.00002005, -0.00001570, -0.00003947, -0.00002875, 0.00000588, 0.00004162, 0.00003458, },
		{ 0.00004607, 0.00002119, -0.00001968, -0.00001210, 0.00001716, -0.00001517, -0.00004528, 0.00001987, 0.00004127, -0.00006233, -0.00008252, 0.00002233, 0.00003443, -0.00001292, 0.00002953, 0.00001702, -0.00008959, -0.00003806, 0.00010005, 0.00003924, -0.00005268, 0.00001595, 0.00000465, -0.00007138, 0.00003160, 0.00006088, -0.00009393, -0.00001922, 0.00012883, -0.00004970, -0.00012105, 0.00008126, 0.00001476, -0.00010036, 0.00010567, 0.00006823, -0.00017417, 0.00000132, 0.00010129, -0.00014696, 0.00000388, 0.00018465, -0.00020748, -0.00017363, 0.00027308, -0.00023830, -0.00073298, 0.00035676, 0.00129329, 0.00042328, -0.00055092, -0.00030559, 0.00006713, -0.00003923, -0.00006890, 0.00002745, 0.00000113, -0.00002843, 0.00000413, 0.00000497, -0.00002210, -0.00003567, -0.00002142, 0.00001990, },
		{ -0.00001050, 0.00003190, 0.00004440, 0.00000402, -0.00002779, 0.00000632, 0.00002058, -0.00004057, -0.00003513, 0.00004835, 0.00000984, -0.00009167, -0.00004003, 0.00004258, 0.00000096, -0.00000252, 0.00004960, -0.00003147, -0.00009953, 0.00003615, 0.00010391, -0.00002564, -0.00003898, 0.00004174, -0.00003624, -0.00005186, 0.00007926, 0.00000181, -0.00011087, 0.00006403, 0.00009540, -0.00011982, -0.00004642, 0.00009484, -0.00006464, -0.00003757, 0.00014463, -0.00004308, -0.00014465, 0.00008616, 0.00000179, -0.00013604, 0.00012665, 0.00004828, -0.00027770, 0.00004416, 0.00016443, -0.00053882, -0.00039972, 0.00084857, 0.00103300, -0.00005123, -0.00050627, -0.00012097, 0.00003814, -0.00005823, -0.00002660, 0.00001441, -0.00001222, -0.00001045, 0.00000376, -0.00000557, -0.00002380, -0.00003256, },
		{ -0.00002448, -0.00002692, 0.00000098, 0.00004068, 0.00003745, -0.00001626, -0.00002392, 0.00002916, 0.00000502, -0.00006164, -0.00000397, 0.00005708, -0.00003648, -0.00008750, 0.00000601, 0.00002826, -0.00001933, 0.00003037, 0.00003637, -0.00008442, -0.00005928, 0.00009887, 0.00005591, -0.00006682, 0.00000507, 0.00003356, -0.00007099, 0.00000195, 0.00008628, -0.00007073, -0.00006990, 0.00012300, 0.00001526, -0.00013280, 0.00003419, 0.00004701, -0.00009850, 0.00005241, 0.00010753, -0.00012383, -0.00005704, 0.00009268, -0.00008738, -0.00004218, 0.00015206, -0.00011696, -0.00019552, 0.00017879, -0.00009978, -0.00061009, 0.00009540, 0.00104267, 0.00060154, -0.00031160, -0.00035835, -0.00003153, 0.00000202, -0.00004426, -0.00001140, -0.00000100, -0.00000703, -0.00000184, -0.00000245, -0.00001085, },
		{ -0.00000783, -0.00001391, -0.00002252, -0.00001996, 0.00001079, 0.00004788, 0.00002264, -0.00003455, -0.00000364, 0.00004266, -0.00002735, -0.00006162, 0.00003373, 0.00003605, -0.00007242, -0.00005318, 0.00002813, -0.00000307, -0.00000795, 0.00005665, -0.00001209, -0.00010558, 0.00001544, 0.00011413, -0.00001560, -0.00006113, 0.00004610, -0.00000690, -0.00007128, 0.00006173, 0.00004274, -0.00011453, 0.00001008, 0.00012464, -0.00006783, -0.00008691, 0.00007437, -0.00002614, -0.00006775, 0.00011365, 0.00002082, -0.00013461, 0.00002691, 0.00003263, -0.00010710, 0.00006225, 0.00007005, -0.00020586, -0.00003218, 0.00014308, -0.00035994, -0.00041333, 0.00053721, 0.00093583, 0.00019203, -0.00037320, -0.00021542, -0.00000621, -0.00001234, -0.00002918, -0.00001393, -0.00000388, 0.00000137, -0.00000218, },
		{ 0.00000475, -0.00000570, -0.00001243, -0.00001409, -0.00001757, -0.00001561, 0.00002162, 0.00005178, -0.00000204, -0.00004021, 0.00002764, 0.00003365, -0.00006053, -0.00003539, 0.00005573, -0.00000454, -0.00007819, -0.00001272, 0.00001773, -0.00002166, 0.00002678, 0.00004916, -0.00006949, -0.00007565, 0.00008511, 0.00007399, -0.00006898, -0.00001675, 0.00005435, -0.00005321, -0.00003073, 0.00009133, -0.00002963, -0.00010291, 0.00008625, 0.00007028, -0.00011083, -0.00001566, 0.00005717, -0.00007563, 0.00000376, 0.00011292, -0.00006150, -0.00008473, 0.00005818, -0.00003900, -0.00005426, 0.00010370, -0.00005566, -0.00017376, 0.00009211, -0.00003481, -0.00046406, -0.00005062, 0.00077041, 0.00064879, -0.00008057, -0.00032176, -0.00012122, -0.00000176, -0.00001231, -0.00002629, -0.00001635, 0.00000212, },
		{ -0.00001161, 0.00000808, 0.00000434, -0.00001068, -0.00001618, -0.00000863, -0.00001455, -0.00001354, 0.00003706, 0.00004531, -0.00003095, -0.00002448, 0.00005367, 0.00000149, -0.00007444, 0.00000400, 0.00004916, -0.00004033, -0.00005546, 0.00000739, -0.00000822, -0.00000958, 0.00005571, 0.00000494, -0.00009915, -0.00000739, 0.00011201, 0.00000334, -0.00007650, 0.00003431, 0.00002430, -0.00007360, 0.00002923, 0.00007165, -0.00009062, -0.00004215, 0.00011945, -0.00000681, -0.00009854, 0.00003657, 0.00000351, -0.00007311, 0.00006911, 0.00005931, -0.00009682, -0.00001937, 0.00003254, -0.00006870, 0.00002146, 0.00006054, -0.00013829, -0.00006322, 0.00009346, -0.00023660, -0.00035798, 0.00030938, 0.00076461, 0.00033381, -0.00020322, -0.00024002, -0.00006798, 0.00000231, -0.00001384, -0.00002897, },
		{ -0.00002063, -0.00002748, -0.00000419, 0.00001146, 0.00000233, -0.00001855, -0.00001449, 0.00000048, -0.00001898, -0.00000685, 0.00005402, 0.00002248, -0.00005004, 0.00000952, 0.00005730, -0.00003830, -0.00006049, 0.00003400, 0.00002157, -0.00005165, -0.00002628, -0.00000064, -0.00002131, 0.00002436, 0.00005201, -0.00005090, -0.00008042, 0.00006211, 0.00008451, -0.00005736, -0.00004054, 0.00005939, -0.00002317, -0.00005370, 0.00007375, 0.00001356, -0.00010745, 0.00003367, 0.00009813, -0.00006484, -0.00005017, 0.00004435, -0.00004600, -0.00002714, 0.00009206, -0.00000766, -0.00008037, 0.00001734, -0.00001384, -0.00004358, 0.00006017, -0.00002615, -0.00013220, 0.00003356, -0.00002117, -0.00033644, -0.00010506, 0.00053216, 0.00059629, 0.00009286, -0.00022533, -0.00016875, -0.00003477, 0.00000396, },
		{ -0.00001296, -0.00000180, -0.00002531, -0.00002151, 0.00000115, 0.00001503, -0.00000391, -0.00002683, -0.00000282, 0.00000440, -0.00002890, 0.00001132, 0.00006139, -0.00001252, -0.00004600, 0.00004511, 0.00003417, -0.00006415, -0.00002907, 0.00004041, -0.00000494, -0.00003965, -0.00001468, -0.00001987, -0.00000571, 0.00005151, 0.00001297, -0.00008307, -0.00002346, 0.00009604, 0.00002270, -0.00007819, 0.00001149, 0.00004631, -0.00005716, -0.00000577, 0.00007808, -0.00004983, -0.00007465, 0.00008561, 0.00004228, -0.00008028, -0.00000205, 0.00001468, -0.00006060, 0.00002805, 0.00006914, -0.00004959, -0.00004087, 0.00001309, -0.00003843, 0.00000483, 0.00003810, -0.00008969, -0.00006161, 0.00004495, -0.00016666, -0.00027979, 0.00016823, 0.00057771, 0.00037757, -0.00004823, -0.00020221, -0.00011301, },
		{ -0.00016410, -0.00006892, -0.00000475, -0.00001014, -0.00002318, -0.00001732, 0.00000678, 0.00001833, -0.00001777, -0.00002686, 0.00001524, -0.00000527, -0.00003329, 0.00003841, 0.00004843, -0.00004409, -0.00001816, 0.00006260, -0.00000247, -0.00006507, -0.00000109, 0.00002898, -0.00001280, -0.00002485, -0.00002581, -0.00002364, 0.00002788, 0.00004684, -0.00003709, -0.00007179, 0.00003827, 0.00008112, -0.00003699, -0.00005621, 0.00004772, 0.00000832, -0.00005832, 0.00004123, 0.00004277, -0.00008527, -0.00001461, 0.00009373, -0.00001344, -0.00005903, 0.00001842, -0.00002345, -0.00003787, 0.00005976, 0.00002529, -0.00005456, -0.00001279, -0.00001084, -0.00002635, 0.00003182, -0.00001862, -0.00009102, 0.00000513, -0.00003096, -0.00024463, -0.00010594, 0.00035420, 0.00049222, 0.00018926, -0.00011285, },
		{ 0.00005830, -0.00013178, -0.00012055, -0.00003961, -0.00000841, -0.00001159, -0.00001963, -0.00001661, 0.00001758, 0.00001430, -0.00003510, -0.00001121, 0.00002685, -0.00002653, -0.00002087, 0.00006073, 0.00001597, -0.00005529, 0.00001856, 0.00005436, -0.00003188, -0.00004886, 0.00001026, 0.00001929, -0.00000485, -0.00002816, -0.00004112, 0.00000035, 0.00005072, 0.00000985, -0.00006520, -0.00002638, 0.00007140, 0.00003304, -0.00006577, -0.00001241, 0.00005199, -0.00002854, -0.00003000, 0.00006179, -0.00000913, -0.00007930, 0.00003962, 0.00006542, -0.00004339, -0.00002635, 0.00000840, -0.00004273, 0.00000071, 0.00005804, -0.00000964, -0.00003832, -0.00000968, -0.00002411, 0.00000277, 0.00001757, -0.00006099, -0.00004524, 0.00001195, -0.00013246, -0.00020913, 0.00009241, 0.00041552, 0.00035358, },
		{ 0.00038050, 0.00021925, -0.00002259, -0.00012179, -0.00007923, -0.00003005, -0.00001220, -0.00000497, -0.00002314, -0.00001212, 0.00003119, -0.00000351, -0.00004311, 0.00001638, 0.00001951, -0.00004535, 0.00000841, 0.00006316, -0.00002122, -0.00004136, 0.00004429, 0.00003056, -0.00004286, -0.00003365, 0.00001101, 0.00002544, -0.00000185, -0.00005039, -0.00003482, 0.00003765, 0.00004012, -0.00003433, -0.00005453, 0.00002297, 0.00006386, -0.00001741, -0.00005770, 0.00002568, 0.00003012, -0.00004423, 0.00000739, 0.00004927, -0.00004906, -0.00004282, 0.00006545, 0.00002498, -0.00004408, -0.00000611, -0.00001460, -0.00003517, 0.00002983, 0.00003621, -0.00002272, -0.00002453, -0.00001969, -0.00001486, 0.00001754, -0.00001977, -0.00005974, -0.00000303, -0.00004361, -0.00018536, -0.00008722, 0.00023450, },
		{ 0.00005300, 0.00029256, 0.00030059, 0.00011007, -0.00006456, -0.00009172, -0.00005338, -0.00003439, -0.00000412, -0.00000104, -0.00003310, 0.00000422, 0.00003601, -0.00003072, -0.00003022, 0.00003997, -0.00000682, -0.00004588, 0.00004058, 0.00004216, -0.00004466, -0.00001394, 0.00004991, 0.00000931, -0.00004219, -0.00002957, 0.00001965, 0.00004061, -0.00002244, -0.00006905, 0.00000162, 0.00005759, -0.00000055, -0.00005452, -0.00001602, 0.00004767, 0.00002906, -0.00004566, -0.00002759, 0.00004164, 0.00000087, -0.00003586, 0.00003223, 0.00001662, -0.00006067, 0.00000023, 0.00006113, -0.00000515, -0.00002988, -0.00000580, -0.00002999, -0.00001219, 0.00003841, 0.00001448, -0.00002152, -0.00002368, -0.00002296, 0.00000357, 0.00000468, -0.00004460, -0.00002734, -0.00000548, -0.00011446, -0.00015620, },
		{ -0.00014562, -0.00006787, 0.00015457, 0.00028635, 0.00021123, 0.00002725, -0.00007082, -0.00005660, -0.00005170, -0.00003486, 0.00001432, -0.00001152, -0.00003720, 0.00002959, 0.00002115, -0.00005120, 0.00000182, 0.00004256, -0.00003718, -0.00002402, 0.00005781, 0.00000963, -0.00004674, 0.00000922, 0.00004333, -0.00000048, -0.00004586, -0.00002756, 0.00004705, 0.00004099, -0.00006258, -0.00005703, 0.00004841, 0.00003896, -0.00004308, -0.00003776, 0.00002152, 0.00003953, -0.00000765, -0.00004611, 0.00000348, 0.00003621, -0.00001909, -0.00001511, 0.00003511, -0.00001544, -0.00004629, 0.00002975, 0.00004059, -0.00001710, -0.00001810, -0.00001677, -0.00002898, 0.00000921, 0.00003235, 0.00000289, -0.00002050, -0.00002842, -0.00001262, 0.00000969, -0.00002075, -0.00003762, -0.00000273, -0.00005108, },
		{ -0.00001444, -0.00010083, -0.00011664, 0.00002818, 0.00020219, 0.00024818, 0.00012355, -0.00002550, -0.00004662, -0.00003824, -0.00006561, -0.00001617, 0.00002680, -0.00003489, -0.00002259, 0.00004852, -0.00001102, -0.00004858, 0.00003591, 0.00002032, -0.00005244, 0.00000833, 0.00005272, -0.00001698, -0.00003577, 0.00002008, 0.00003968, -0.00000730, -0.00006015, -0.00000734, 0.00008107, 0.00000768, -0.00009395, -0.00000823, 0.00007068, -0.00001048, -0.00005617, 0.00000115, 0.00003386, 0.00001255, -0.00002767, -0.00002944, 0.00002237, 0.00002069, -0.00002451, 0.00000310, 0.00002130, -0.00003233, -0.00002040, 0.00003952, 0.00002000, -0.00001701, -0.00001646, -0.00002573, -0.00001661, 0.00002019, 0.00002354, -0.00000234, -0.00002539, -0.00002636, 0.00000036, -0.00000312, -0.00003233, -0.00001235, },
		{ -0.00001936, -0.00000059, -0.00005517, -0.00011395, -0.00005119, 0.00009455, 0.00021342, 0.00019427, 0.00004188, -0.00003869, -0.00001159, -0.00004986, -0.00007147, 0.00001873, 0.00001670, -0.00005423, 0.00000975, 0.00004482, -0.00004293, -0.00002074, 0.00005152, -0.00001330, -0.00004440, 0.00003313, 0.00003357, -0.00002952, -0.00002623, 0.00002659, 0.00004401, -0.00002731, -0.00007080, 0.00004085, 0.00009213, -0.00005181, -0.00008523, 0.00005203, 0.00004652, -0.00005765, -0.00003014, 0.00003219, 0.00001810, -0.00001139, -0.00002875, -0.00001022, 0.00002813, 0.00000591, -0.00002044, 0.00001077, 0.00000324, -0.00003193, 0.00000244, 0.00003544, 0.00000726, -0.00001516, -0.00002158, -0.00002593, -0.00000285, 0.00002307, 0.00001760, -0.00000920, -0.00003002, -0.00001550, 0.00000114, -0.00002012, },
		{ -0.00001133, -0.00002084, 0.00000187, -0.00002111, -0.00008950, -0.00008193, 0.00000968, 0.00012913, 0.00020758, 0.00012429, -0.00001910, -0.00001212, 0.00000468, -0.00008152, -0.00004912, 0.00004824, -0.00001592, -0.00005082, 0.00004241, 0.00001671, -0.00005423, 0.00001664, 0.00004062, -0.00003823, -0.00002152, 0.00004113, 0.00001496, -0.00003469, -0.00002290, 0.00004247, 0.00004345, -0.00006425, -0.00005246, 0.00010029, 0.00005756, -0.00010289, -0.00003003, 0.00008299, -0.00001305, -0.00006870, 0.00001546, 0.00003361, -0.00000860, -0.00001958, -0.00002049, 0.00000381, 0.00002648, -0.00000420, -0.00001461, 0.00000913, -0.00000911, -0.00002127, 0.00001561, 0.00002691, 0.00000102, -0.00001697, -0.00002614, -0.00001937, 0.00000704, 0.00002375, 0.00001158, -0.00001885, -0.00002660, -0.00000648, },
		{ -0.00001692, -0.00000996, -0.00002141, -0.00000144, 0.00000341, -0.00006045, -0.00008653, -0.00003211, 0.00004390, 0.00015163, 0.00018457, 0.00004175, -0.00003799, 0.00003223, -0.00001686, -0.00010396, -0.00000063, 0.00004938, -0.00005141, -0.00002127, 0.00005452, -0.00001955, -0.00003816, 0.00004172, 0.00001335, -0.00004366, 0.00000017, 0.00003786, 0.00000307, -0.00004433, -0.00001494, 0.00006919, 0.00001834, -0.00009826, 0.00000720, 0.00013426, -0.00001488, -0.00010788, 0.00004254, 0.00005950, -0.00006961, -0.00003520, 0.00004720, 0.00000741, -0.00002515, -0.00001483, -0.00001262, 0.00001320, 0.00002198, -0.00001127, -0.00001104, 0.00000405, -0.00001299, -0.00000878, 0.00002033, 0.00001939, -0.00000349, -0.00002126, -0.00002634, -0.00001103, 0.00001430, 0.00002363, 0.00000135, -0.00002469, },
		{ -0.00001080, -0.00002139, -0.00001204, -0.00002275, -0.00001225, 0.00001841, -0.00002635, -0.00008374, -0.00004792, -0.00000001, 0.00006211, 0.00017346, 0.00013181, -0.00003109, -0.00000730, 0.00005770, -0.00006722, -0.00009052, 0.00004876, 0.00001792, -0.00006548, 0.00001802, 0.00003897, -0.00004184, -0.00000690, 0.00004340, -0.00001165, -0.00003493, 0.00001206, 0.00003488, -0.00000987, -0.00005935, 0.00001290, 0.00008819, -0.00003567, -0.00009666, 0.00008969, 0.00011317, -0.00008614, -0.00005678, 0.00008580, -0.00000660, -0.00008499, 0.00001925, 0.00004337, -0.00002482, -0.00002306, -0.00000833, -0.00000721, 0.00002090, 0.00001495, -0.00001734, -0.00000915, 0.00000017, -0.00001095, 0.00000101, 0.00002060, 0.00001328, -0.00000871, -0.00002464, -0.00002324, -0.00000280, 0.00002116, 0.00001907, },
		{ 0.00002571, 0.00000788, -0.00001729, -0.00001329, -0.00002016, -0.00002785, 0.00001482, 0.00001230, -0.00006821, -0.00006622, -0.00000626, 0.00000342, 0.00008931, 0.00018147, 0.00004850, -0.00006187, 0.00005143, 0.00003778, -0.00011294, -0.00003931, 0.00006962, -0.00002694, -0.00004878, 0.00004363, 0.00000830, -0.00003962, 0.00001900, 0.00002840, -0.00002341, -0.00002515, 0.00001925, 0.00003503, -0.00003539, -0.00006405, 0.00006088, 0.00007350, -0.00009511, -0.00004071, 0.00015227, 0.00004092, -0.00011182, 0.00002074, 0.00006961, -0.00007288, -0.00004960, 0.00005707, 0.00000905, -0.00004054, -0.00001018, -0.00000744, -0.00000031, 0.00002652, 0.00000420, -0.00002185, -0.00000699, -0.00000132, -0.00000646, 0.00000752, 0.00001881, 0.00000723, -0.00001382, -0.00002587, -0.00001877, 0.00000676, },
		{ -0.00001226, 0.00001789, 0.00002317, -0.00000543, -0.00001388, -0.00001207, -0.00003507, -0.00001038, 0.00003677, -0.00002675, -0.00008580, -0.00001740, 0.00000440, 0.00000026, 0.00013378, 0.00015060, -0.00003925, -0.00003390, 0.00009605, -0.00002258, -0.00012033, 0.00002399, 0.00005126, -0.00005740, -0.00001386, 0.00004383, -0.00001643, -0.00002101, 0.00002769, 0.00001247, -0.00002595, -0.00002046, 0.00003186, 0.00002711, -0.00007201, -0.00003855, 0.00010576, 0.00001620, -0.00011989, 0.00005368, 0.00015699, -0.00004280, -0.00007589, 0.00007412, 0.00000394, -0.00009659, 0.00001097, 0.00005460, -0.00003015, -0.00003319, -0.00000071, -0.00000983, 0.00001081, 0.00002627, -0.00000901, -0.00002283, -0.00000414, -0.00000116, -0.00000135, 0.00001138, 0.00001536, 0.00000121, -0.00001756, -0.00002601, },
		{ -0.00002040, -0.00002528, -0.00000167, 0.00002649, 0.00001202, -0.00001234, -0.00000636, -0.00002456, -0.00003842, 0.00002488, 0.00002703, -0.00007483, -0.00005715, 0.00002437, -0.00002007, 0.00002290, 0.00017309, 0.00007261, -0.00008992, 0.00003524, 0.00009184, -0.00008884, -0.00008087, 0.00006538, 0.00000912, -0.00005851, 0.00001647, 0.00002635, -0.00002319, -0.00000349, 0.00002389, 0.00000465, -0.00003041, -0.00001367, 0.00004993, -0.00000317, -0.00009918, 0.00001982, 0.00011419, -0.00006047, -0.00008242, 0.00014112, 0.00010027, -0.00008895, -0.00000489, 0.00006833, -0.00006745, -0.00006535, 0.00005581, 0.00001786, -0.00004924, -0.00001432, -0.00000177, -0.00000785, 0.00002290, 0.00001761, -0.00002054, -0.00001984, -0.00000156, 0.00000022, 0.00000368, 0.00001257, 0.00001054, -0.00000367, },
		{ 0.00000568, -0.00000710, -0.00002360, -0.00002143, 0.00001240, 0.00002669, -0.00000218, -0.00000889, -0.00000678, -0.00004281, -0.00001610, 0.00005216, -0.00001909, -0.00009304, 0.00000323, 0.00002342, -0.00004857, 0.00008088, 0.00017075, -0.00002688, -0.00007647, 0.00010104, 0.00003665, -0.00012175, -0.00001681, 0.00006751, -0.00002899, -0.00003737, 0.00002811, 0.00000904, -0.00001722, 0.00000273, 0.00002086, -0.00000020, -0.00004019, 0.00000653, 0.00005800, -0.00005497, -0.00009066, 0.00008629, 0.00006971, -0.00011148, 0.00000828, 0.00017650, 0.00001785, -0.00007507, 0.00005031, 0.00000972, -0.00009876, -0.00000447, 0.00005883, -0.00002519, -0.00004187, -0.00000102, -0.00000846, 0.00000245, 0.00002900, 0.00000290, -0.00002654, -0.00001486, 0.00000024, 0.00000307, 0.00000767, 0.00001088, },
		},
		{ { -0.00000068, -0.00000253, -0.00000145, 0.00000930, 0.00001426, -0.00000338, -0.00002242, -0.00002124, -0.00001531, -0.00000255, 0.00003030, 0.00003712, -0.00000490, -0.00002327, -0.00000384, -0.00001563, -0.00001991, 0.00003383, 0.00002093, -0.00008359, -0.00006300, 0.00007733, 0.00005591, -0.00007721, -0.00001629, 0.00012734, 0.00004837, -0.00009904, -0.00003600, 0.00008935, 0.00005912, -0.00003820, -0.00007559, -0.00003820, 0.00005912, 0.00008935, -0.00003600, -0.00009904, 0.00004837, 0.00012734, -0.00001629, -0.00007721, 0.00005591, 0.00007733, -0.00006300, -0.00008359, 0.00002093, 0.00003383, -0.00001991, -0.00001563, -0.00000384, -0.00002327, -0.00000490, 0.00003712, 0.00003030, -0.00000255, -0.00001531, -0.00002124, -0.00002242, -0.00000338, 0.00001426, 0.00000930, -0.00000145, -0.00000253, },
		{ -0.00000335, -0.00000451, 0.00000072, -0.00000090, -0.00000185, 0.00001052, 0.00001327, -0.00001034, -0.00002541, -0.00001623, -0.00001071, 0.00000192, 0.00003453, 0.00003113, -0.00001423, -0.00002244, -0.00000401, -0.00001787, -0.00000508, 0.00004312, -0.00000455, -0.00009427, -0.00002362, 0.00009358, 0.00001580, -0.00008824, 0.00001930, 0.00012095, 0.00000480, -0.00009747, -0.00001047, 0.00008010, 0.00004689, -0.00002994, -0.00008095, -0.00004286, 0.00007985, 0.00008731, -0.00006562, -0.00008270, 0.00009268, 0.00011575, -0.00004534, -0.00004848, 0.00008335, 0.00004265, -0.00009090, -0.00005930, 0.00003702, 0.00001648, -0.00002856, -0.00001027, -0.00000615, -0.00002271, 0.00000595, 0.00004108, 0.00002453, -0.00000783, -0.00002010, -0.00002391, -0.00001803, 0.00000228, 0.00001551, 0.00000866, },
		{ 0.00001744, 0.00000703, -0.00000744, -0.00000479, 0.00000423, -0.00000149, -0.00000345, 0.00001346, 0.00001144, -0.00001863, -0.00002586, -0.00000934, -0.00000720, 0.00000548, 0.00003731, 0.00002389, -0.00002189, -0.00002161, -0.00000531, -0.00001456, 0.00001196, 0.00004077, -0.00003249, -0.00008744, 0.00001724, 0.00008773, -0.00002552, -0.00007926, 0.00005094, 0.00009787, -0.00002911, -0.00008262, 0.00000472, 0.00006602, 0.00004517, -0.00002522, -0.00009481, -0.00003798, 0.00010311, 0.00007090, -0.00009061, -0.00004870, 0.00012851, 0.00009053, -0.00005934, -0.00001021, 0.00009004, -0.00000149, -0.00010092, -0.00002902, 0.00004007, -0.00000319, -0.00002957, -0.00000508, -0.00001078, -0.00001941, 0.00001746, 0.00004235, 0.00001737, -0.00001361, -0.00002423, -0.00002431, -0.00001320, 0.00000720, },
		{ -0.00000846, 0.00001216, 0.00001959, 0.00000303, -0.00001245, -0.00000189, 0.00000807, -0.00000526, -0.00000439, 0.00001827, 0.00000735, -0.00002746, -0.00002279, -0.00000159, -0.00000551, 0.00000834, 0.00003893, 0.00001601, -0.00002839, -0.00002143, -0.00000537, -0.00000519, 0.00002589, 0.00002668, -0.00005477, -0.00006437, 0.00004896, 0.00006241, -0.00005684, -0.00005429, 0.00006970, 0.00006443, -0.00004764, -0.00006231, 0.00000762, 0.00005446, 0.00005155, -0.00002969, -0.00011130, -0.00002113, 0.00012108, 0.00004180, -0.00010262, -0.00000220, 0.00014942, 0.00006021, -0.00005481, 0.00002600, 0.00007382, -0.00004364, -0.00009212, -0.00000157, 0.00003071, -0.00001937, -0.00002424, -0.00000284, -0.00001620, -0.00001263, 0.00002830, 0.00004067, 0.00000916, -0.00001948, -0.00002700, -0.00002288, },
		{ -0.00002796, -0.00002026, -0.00000388, 0.00001772, 0.00002081, -0.00000398, -0.00001635, 0.00000448, 0.00000992, -0.00001187, -0.00000269, 0.00002399, -0.00000007, -0.00003534, -0.00001574, 0.00000560, -0.00000601, 0.00001098, 0.00003974, 0.00000765, -0.00003446, -0.00002142, -0.00000175, 0.00000831, 0.00003183, 0.00000456, -0.00006462, -0.00003124, 0.00006360, 0.00002573, -0.00007061, -0.00002207, 0.00007102, 0.00002981, -0.00004998, -0.00004483, 0.00000094, 0.00005100, 0.00006015, -0.00004532, -0.00012309, 0.00000557, 0.00012678, 0.00000612, -0.00009635, 0.00004847, 0.00015393, 0.00003461, -0.00003464, 0.00004875, 0.00003928, -0.00007336, -0.00006907, 0.00001585, 0.00001327, -0.00002804, -0.00001589, -0.00000481, -0.00002025, -0.00000270, 0.00003713, 0.00003624, 0.00000026, -0.00002493, },
		{ -0.00000905, -0.00002936, -0.00002691, -0.00001711, 0.00000082, 0.00002391, 0.00001962, -0.00001350, -0.00001697, 0.00001327, 0.00000775, -0.00001965, 0.00000302, 0.00002876, -0.00001110, -0.00004034, -0.00000507, 0.00001062, -0.00000851, 0.00001409, 0.00003988, -0.00000148, -0.00004053, -0.00002004, 0.00000698, 0.00002201, 0.00002704, -0.00001906, -0.00005887, 0.00000271, 0.00005823, -0.00001139, -0.00006503, 0.00000695, 0.00005599, 0.00000287, -0.00004015, -0.00003622, -0.00000934, 0.00005741, 0.00006401, -0.00006971, -0.00012408, 0.00003600, 0.00011658, -0.00002750, -0.00007121, 0.00009474, 0.00014586, 0.00002116, -0.00000746, 0.00005061, -0.00000373, -0.00008442, -0.00004009, 0.00001982, -0.00000591, -0.00002821, -0.00000845, -0.00001036, -0.00002105, 0.00000910, 0.00004301, 0.00002956, },
		{ 0.00004573, 0.00002115, -0.00001855, -0.00003212, -0.00002398, -0.00001394, 0.00000609, 0.00003013, 0.00001478, -0.00002397, -0.00001280, 0.00002226, 0.00000055, -0.00002610, 0.00001287, 0.00003031, -0.00002490, -0.00004060, 0.00000792, 0.00001211, -0.00001222, 0.00001833, 0.00003904, -0.00001190, -0.00004624, -0.00001518, 0.00002033, 0.00003110, 0.00001207, -0.00003685, -0.00003908, 0.00002797, 0.00003581, -0.00003858, -0.00004435, 0.00002403, 0.00003068, -0.00001063, -0.00002537, -0.00003836, -0.00001632, 0.00007115, 0.00005767, -0.00009686, -0.00011168, 0.00006213, 0.00009153, -0.00005059, -0.00003123, 0.00013071, 0.00013280, 0.00002212, 0.00001518, 0.00003061, -0.00004321, -0.00007654, -0.00001444, 0.00001159, -0.00002076, -0.00002190, -0.00000492, -0.00001732, -0.00001774, 0.00002101, },
		{ -0.00001088, 0.00003145, 0.00004569, 0.00001129, -0.00002794, -0.00003245, -0.00001958, -0.00001112, 0.00001226, 0.00003522, 0.00000575, -0.00003307, -0.00000351, 0.00002867, -0.00001125, -0.00002855, 0.00002560, 0.00002671, -0.00003949, -0.00003495, 0.00002117, 0.00000928, -0.00001583, 0.00002402, 0.00003650, -0.00002395, -0.00005024, -0.00000512, 0.00003576, 0.00003155, -0.00000915, -0.00004301, -0.00001115, 0.00003771, 0.00000414, -0.00004915, -0.00001718, 0.00002483, 0.00000389, -0.00000966, -0.00001340, -0.00004874, -0.00001488, 0.00008646, 0.00003926, -0.00011935, -0.00008779, 0.00007665, 0.00005700, -0.00005749, 0.00001670, 0.00015513, 0.00012294, 0.00003333, 0.00002271, -0.00000528, -0.00006880, -0.00005522, 0.00000080, -0.00000377, -0.00002754, -0.00001314, -0.00000630, -0.00002303, },
		{ -0.00001132, -0.00002540, -0.00000214, 0.00003963, 0.00004358, -0.00000008, -0.00003663, -0.00002969, -0.00001442, -0.00000875, 0.00001940, 0.00003781, -0.00000699, -0.00003836, 0.00000979, 0.00003000, -0.00002571, -0.00002506, 0.00003876, 0.00001703, -0.00005217, -0.00002327, 0.00003224, 0.00000226, -0.00001781, 0.00003089, 0.00003120, -0.00003726, -0.00005048, 0.00001057, 0.00004914, 0.00002154, -0.00003043, -0.00003517, 0.00001647, 0.00002985, -0.00002663, -0.00004198, 0.00000648, 0.00001045, -0.00001589, 0.00000211, -0.00000984, -0.00006172, -0.00000345, 0.00009653, 0.00001120, -0.00013097, -0.00005824, 0.00007545, 0.00002085, -0.00004680, 0.00006562, 0.00017153, 0.00012157, 0.00004543, 0.00000934, -0.00004600, -0.00007512, -0.00002961, 0.00000293, -0.00001942, -0.00002601, -0.00000625, },
		{ -0.00001923, -0.00000403, -0.00001715, -0.00002392, 0.00000642, 0.00004577, 0.00003982, -0.00001328, -0.00004357, -0.00002344, -0.00000949, -0.00000674, 0.00002721, 0.00003663, -0.00002196, -0.00003797, 0.00002479, 0.00002474, -0.00003989, -0.00001501, 0.00004926, 0.00000176, -0.00006013, -0.00000675, 0.00003887, -0.00000782, -0.00001680, 0.00003795, 0.00002224, -0.00005053, -0.00004483, 0.00003044, 0.00005596, 0.00000235, -0.00004513, -0.00001525, 0.00003541, 0.00000764, -0.00004721, -0.00002154, 0.00001848, -0.00001323, -0.00002316, 0.00001797, -0.00001644, -0.00007087, 0.00001542, 0.00009607, -0.00002071, -0.00012891, -0.00003073, 0.00005871, -0.00000921, -0.00002117, 0.00011135, 0.00018629, 0.00012834, 0.00004700, -0.00002359, -0.00007901, -0.00006340, -0.00000925, -0.00000569, -0.00002935, },
		{ -0.00001879, -0.00003060, -0.00001207, -0.00000672, -0.00002066, -0.00001981, 0.00001331, 0.00005085, 0.00003418, -0.00002841, -0.00004724, -0.00001385, -0.00000593, -0.00000481, 0.00003505, 0.00003094, -0.00003694, -0.00003109, 0.00003846, 0.00001287, -0.00005052, 0.00000056, 0.00005426, -0.00001713, -0.00006120, 0.00001225, 0.00003951, -0.00001907, -0.00001207, 0.00004362, 0.00000927, -0.00006154, -0.00003193, 0.00005121, 0.00005283, -0.00002177, -0.00004830, 0.00001103, 0.00003996, -0.00002141, -0.00005198, 0.00000382, 0.00001494, -0.00003782, -0.00001692, 0.00003072, -0.00003082, -0.00007140, 0.00003620, 0.00008323, -0.00004932, -0.00011459, -0.00001230, 0.00003060, -0.00002797, 0.00001461, 0.00015404, 0.00020532, 0.00013663, 0.00002881, -0.00006756, -0.00009446, -0.00004086, -0.00000067, },
		{ -0.00001814, -0.00000516, -0.00002877, -0.00002436, -0.00000919, -0.00001186, -0.00001992, -0.00001559, 0.00001833, 0.00005600, 0.00002567, -0.00004498, -0.00004590, -0.00000186, -0.00000477, -0.00000253, 0.00004200, 0.00002071, -0.00004944, -0.00001826, 0.00004785, -0.00000404, -0.00005492, 0.00001916, 0.00005189, -0.00003657, -0.00005440, 0.00003070, 0.00003376, -0.00002917, -0.00000380, 0.00004604, -0.00000701, -0.00006756, -0.00001201, 0.00006839, 0.00003868, -0.00004474, -0.00003826, 0.00003589, 0.00002886, -0.00004801, -0.00004053, 0.00002468, -0.00000280, -0.00005529, -0.00000061, 0.00003514, -0.00004775, -0.00006189, 0.00005249, 0.00006011, -0.00006868, -0.00009302, -0.00000715, -0.00000265, -0.00003392, 0.00005634, 0.00019782, 0.00023055, 0.00013518, -0.00001229, -0.00010931, -0.00008862, },
		{ -0.00013506, -0.00006533, -0.00000540, -0.00001833, -0.00002967, -0.00001539, -0.00001295, -0.00001539, -0.00001519, -0.00001390, 0.00002276, 0.00006167, 0.00001282, -0.00006143, -0.00003806, 0.00001074, -0.00000674, 0.00000063, 0.00004710, 0.00000673, -0.00005726, -0.00000129, 0.00005076, -0.00002319, -0.00005169, 0.00003740, 0.00004185, -0.00005308, -0.00004031, 0.00004550, 0.00002251, -0.00003595, 0.00000680, 0.00004365, -0.00002477, -0.00006613, 0.00001280, 0.00007750, 0.00001538, -0.00006039, -0.00001731, 0.00005195, 0.00000565, -0.00006394, -0.00001757, 0.00003360, -0.00002900, -0.00006050, 0.00001955, 0.00002957, -0.00006128, -0.00004464, 0.00005933, 0.00003194, -0.00007593, -0.00007085, -0.00001562, -0.00003519, -0.00002866, 0.00010303, 0.00024843, 0.00025748, 0.00011198, -0.00007120, },
		{ 0.00005895, -0.00013510, -0.00013529, -0.00003492, -0.00000862, -0.00003175, -0.00001964, -0.00001018, -0.00002223, -0.00001332, -0.00000911, -0.00001632, 0.00002893, 0.00006699, -0.00000563, -0.00007511, -0.00002310, 0.00002151, -0.00001190, 0.00000519, 0.00004943, -0.00000955, -0.00005893, 0.00001717, 0.00004633, -0.00004125, -0.00004106, 0.00005185, 0.00002551, -0.00006361, -0.00002097, 0.00005425, 0.00000777, -0.00003799, 0.00001771, 0.00003573, -0.00004114, -0.00005589, 0.00003860, 0.00007544, -0.00001253, -0.00006436, 0.00000901, 0.00005445, -0.00002250, -0.00006461, 0.00000881, 0.00002726, -0.00005567, -0.00005253, 0.00003685, 0.00001612, -0.00006705, -0.00002460, 0.00005492, 0.00000498, -0.00007204, -0.00005396, -0.00003489, -0.00006361, -0.00001479, 0.00015792, 0.00030954, 0.00027507, },
		{ 0.00037905, 0.00026820, -0.00002391, -0.00018671, -0.00010830, -0.00001092, -0.00002719, -0.00003606, -0.00000198, -0.00001427, -0.00003259, -0.00000365, -0.00000585, -0.00002231, 0.00003920, 0.00006963, -0.00002986, -0.00008264, -0.00000180, 0.00002775, -0.00001954, 0.00001147, 0.00004830, -0.00002627, -0.00005403, 0.00003423, 0.00003515, -0.00005513, -0.00002486, 0.00005986, 0.00000564, -0.00006623, 0.00000051, 0.00005569, -0.00000779, -0.00003493, 0.00002653, 0.00002280, -0.00005290, -0.00003724, 0.00006058, 0.00006155, -0.00003899, -0.00005547, 0.00003349, 0.00004279, -0.00004709, -0.00005030, 0.00002982, 0.00000725, -0.00007527, -0.00003453, 0.00004647, -0.00000058, -0.00006373, -0.00000738, 0.00004101, -0.00001569, -0.00006091, -0.00004560, -0.00006089, -0.00008767, 0.00000698, 0.00022759, },
		{ 0.00004013, 0.00031879, 0.00044695, 0.00022240, -0.00012658, -0.00020933, -0.00006158, -0.00000539, -0.00005369, -0.00002464, 0.00001581, -0.00002979, -0.00003771, 0.00001241, -0.00000959, -0.00002905, 0.00005475, 0.00006626, -0.00005829, -0.00008065, 0.00002341, 0.00002717, -0.00002810, 0.00001949, 0.00004333, -0.00004144, -0.00004319, 0.00004734, 0.00001906, -0.00006271, -0.00000602, 0.00006017, -0.00001428, -0.00006056, 0.00002073, 0.00005005, -0.00002148, -0.00002767, 0.00003116, 0.00000674, -0.00005720, -0.00001259, 0.00007417, 0.00003800, -0.00005816, -0.00003614, 0.00004954, 0.00002064, -0.00006093, -0.00002584, 0.00003876, -0.00002086, -0.00008304, -0.00001218, 0.00004682, -0.00001536, -0.00005322, 0.00000287, 0.00002184, -0.00002788, -0.00004749, -0.00004580, -0.00009066, -0.00010920, },
		{ -0.00012414, -0.00012947, 0.00009301, 0.00043418, 0.00049573, 0.00012930, -0.00023007, -0.00019242, -0.00001017, -0.00002461, -0.00007614, 0.00000357, 0.00002444, -0.00005420, -0.00003173, 0.00003032, -0.00002274, -0.00003209, 0.00007460, 0.00005362, -0.00008732, -0.00006686, 0.00004859, 0.00001856, -0.00003548, 0.00002877, 0.00003455, -0.00005327, -0.00002791, 0.00005479, 0.00000067, -0.00006323, 0.00001218, 0.00005318, -0.00003085, -0.00004787, 0.00003668, 0.00003885, -0.00003123, -0.00001812, 0.00003036, -0.00000949, -0.00005245, 0.00001397, 0.00007636, 0.00000938, -0.00006605, -0.00001167, 0.00005319, -0.00000526, -0.00006035, 0.00000104, 0.00003290, -0.00004959, -0.00007812, 0.00000856, 0.00003958, -0.00002458, -0.00003959, 0.00000488, 0.00000225, -0.00003259, -0.00003548, -0.00005241, },
		{ -0.00002587, -0.00006331, -0.00016429, -0.00014605, 0.00017706, 0.00056823, 0.00050378, -0.00000873, -0.00031063, -0.00013568, 0.00002753, -0.00006637, -0.00008184, 0.00004305, 0.00001620, -0.00008044, -0.00001164, 0.00004335, -0.00004461, -0.00002679, 0.00009529, 0.00002978, -0.00011180, -0.00004098, 0.00006896, 0.00000228, -0.00003940, 0.00003832, 0.00002239, -0.00006043, -0.00001020, 0.00005595, -0.00001726, -0.00005738, 0.00002685, 0.00004075, -0.00004151, -0.00003063, 0.00004637, 0.00002453, -0.00003607, -0.00000870, 0.00002425, -0.00002252, -0.00003893, 0.00003756, 0.00006657, -0.00001844, -0.00006170, 0.00001139, 0.00004428, -0.00002727, -0.00004617, 0.00002234, 0.00001399, -0.00007206, -0.00006326, 0.00002342, 0.00002850, -0.00002725, -0.00002711, 0.00000027, -0.00001448, -0.00003291, },
		{ -0.00002783, -0.00003189, -0.00001684, -0.00007888, -0.00021530, -0.00015045, 0.00030280, 0.00070506, 0.00045122, -0.00017658, -0.00034597, -0.00005036, 0.00003534, -0.00011985, -0.00006183, 0.00008277, -0.00001202, -0.00009896, 0.00002107, 0.00004463, -0.00007107, -0.00001006, 0.00011148, -0.00000470, -0.00012601, -0.00000534, 0.00007995, -0.00001954, -0.00003800, 0.00004668, 0.00000774, -0.00006212, 0.00000777, 0.00005119, -0.00003246, -0.00004692, 0.00003617, 0.00002567, -0.00004508, -0.00001196, 0.00004921, 0.00000975, -0.00003626, -0.00000167, 0.00001433, -0.00002940, -0.00001890, 0.00005367, 0.00004691, -0.00003997, -0.00004744, 0.00002722, 0.00002639, -0.00003927, -0.00002319, 0.00003220, -0.00001271, -0.00008404, -0.00004335, 0.00003104, 0.00001753, -0.00002508, -0.00001840, -0.00000759, },
		{ -0.00001345, -0.00001562, -0.00003997, -0.00003039, -0.00000537, -0.00010351, -0.00027928, -0.00012783, 0.00047455, 0.00081932, 0.00032647, -0.00034795, -0.00032187, 0.00004273, 0.00000457, -0.00016836, -0.00001437, 0.00010936, -0.00005699, -0.00010062, 0.00006022, 0.00002955, -0.00009531, 0.00001820, 0.00011728, -0.00004623, -0.00012518, 0.00003520, 0.00007827, -0.00004329, -0.00003040, 0.00005219, -0.00000809, -0.00005817, 0.00002403, 0.00004169, -0.00004351, -0.00003420, 0.00003964, 0.00001093, -0.00004184, 0.00000516, 0.00004593, -0.00000321, -0.00003315, 0.00000150, 0.00000318, -0.00002841, 0.00000378, 0.00005938, 0.00002175, -0.00005154, -0.00002807, 0.00003255, 0.00000565, -0.00003848, 0.00000146, 0.00002855, -0.00004047, -0.00008482, -0.00002330, 0.00003274, 0.00000905, -0.00002114, },
		{ 0.00000305, -0.00001804, -0.00000988, -0.00002299, -0.00005415, -0.00002590, 0.00000999, -0.00014524, -0.00035240, -0.00005953, 0.00068568, 0.00088052, 0.00013175, -0.00049046, -0.00023709, 0.00011837, -0.00006262, -0.00019408, 0.00005369, 0.00011122, -0.00010940, -0.00007984, 0.00009625, -0.00000239, -0.00010972, 0.00005446, 0.00010805, -0.00008850, -0.00010692, 0.00007400, 0.00006293, -0.00006456, -0.00001705, 0.00005335, -0.00002344, -0.00004901, 0.00003700, 0.00002902, -0.00004993, -0.00002150, 0.00003801, -0.00000098, -0.00003332, 0.00001847, 0.00003824, -0.00001298, -0.00002869, 0.00000054, -0.00000613, -0.00001962, 0.00002456, 0.00005408, -0.00000354, -0.00005232, -0.00000938, 0.00002754, -0.00001135, -0.00002620, 0.00002078, 0.00001334, -0.00006338, -0.00007680, -0.00000633, 0.00003128, },
		{ 0.00000688, 0.00002908, -0.00000254, -0.00001628, -0.00000454, -0.00003197, -0.00007229, -0.00001297, 0.00002636, -0.00021322, -0.00042226, 0.00007193, 0.00091598, 0.00086003, -0.00011452, -0.00057314, -0.00010518, 0.00015362, -0.00015301, -0.00018343, 0.00012813, 0.00008236, -0.00015608, -0.00003678, 0.00011878, -0.00004690, -0.00010815, 0.00009207, 0.00008209, -0.00012373, -0.00007226, 0.00010402, 0.00003570, -0.00007914, 0.00000019, 0.00004918, -0.00003648, -0.00003557, 0.00004564, 0.00001480, -0.00005212, -0.00001053, 0.00003292, -0.00000852, -0.00002179, 0.00002683, 0.00002826, -0.00001926, -0.00002482, -0.00000353, -0.00001094, -0.00000495, 0.00003930, 0.00003969, -0.00002400, -0.00004442, 0.00000377, 0.00001561, -0.00001967, -0.00000712, 0.00003002, -0.00000857, -0.00007812, -0.00006392, },
		{ -0.00008455, -0.00004974, 0.00001777, 0.00002669, -0.00001032, -0.00001356, 0.00000431, -0.00004737, -0.00009264, 0.00001443, 0.00003493, -0.00031340, -0.00046767, 0.00027633, 0.00113268, 0.00073911, -0.00037937, -0.00057444, 0.00004783, 0.00013432, -0.00024523, -0.00013153, 0.00019098, 0.00002476, -0.00018363, 0.00002202, 0.00011969, -0.00009561, -0.00008797, 0.00012280, 0.00004162, -0.00014450, -0.00002584, 0.00011943, 0.00000088, -0.00008403, 0.00001840, 0.00003967, -0.00004541, -0.00001929, 0.00004940, 0.00000044, -0.00005094, -0.00000217, 0.00002638, -0.00001133, -0.00000963, 0.00003024, 0.00001795, -0.00002275, -0.00002287, -0.00000868, -0.00000968, 0.00001220, 0.00004534, 0.00002009, -0.00003645, -0.00003203, 0.00000887, 0.00000215, -0.00001761, 0.00001238, 0.00002806, -0.00003136, },
		{ 0.00001730, -0.00005044, -0.00008504, -0.00003592, 0.00002848, 0.00002226, -0.00002150, -0.00000558, 0.00001506, -0.00007430, -0.00010872, 0.00005946, 0.00002166, -0.00044382, -0.00046160, 0.00054992, 0.00129547, 0.00051554, -0.00062109, -0.00048839, 0.00018917, 0.00005940, -0.00031534, -0.00004421, 0.00022542, -0.00005155, -0.00018232, 0.00008507, 0.00009558, -0.00013818, -0.00005119, 0.00013904, -0.00000731, -0.00014572, 0.00002483, 0.00011706, -0.00003553, -0.00007822, 0.00003422, 0.00002587, -0.00004882, -0.00000190, 0.00004820, -0.00001299, -0.00004746, 0.00000352, 0.00002016, -0.00001005, 0.00000120, 0.00002957, 0.00000864, -0.00002468, -0.00002316, -0.00001246, -0.00000240, 0.00002789, 0.00004220, 0.00000001, -0.00004028, -0.00002001, 0.00000653, -0.00000742, -0.00000708, 0.00002648, },
		{ 0.00000719, 0.00003190, 0.00000245, -0.00006388, -0.00008283, -0.00002171, 0.00003999, 0.00001234, -0.00003428, 0.00001197, 0.00002203, -0.00011527, -0.00011026, 0.00011899, -0.00002951, -0.00059118, -0.00037721, 0.00087191, 0.00136461, 0.00020711, -0.00079792, -0.00032721, 0.00028721, -0.00005819, -0.00034321, 0.00006298, 0.00022056, -0.00013083, -0.00014904, 0.00013875, 0.00004911, -0.00016506, -0.00000422, 0.00013607, -0.00005599, -0.00012613, 0.00007095, 0.00009727, -0.00006717, -0.00006300, 0.00004465, 0.00000990, -0.00004598, 0.00001464, 0.00004233, -0.00002471, -0.00004254, 0.00000708, 0.00001548, -0.00000596, 0.00000951, 0.00002615, 0.00000085, -0.00002639, -0.00002497, -0.00001276, 0.00000922, 0.00003858, 0.00003164, -0.00001636, -0.00003747, -0.00001221, 0.00000008, -0.00000952, },
		{ -0.00000569, -0.00000371, 0.00001944, 0.00002882, -0.00001150, -0.00007273, -0.00008016, -0.00000469, 0.00005082, -0.00000629, -0.00004351, 0.00004080, 0.00001662, -0.00016739, -0.00008614, 0.00018158, -0.00013139, -0.00073023, -0.00019554, 0.00120510, 0.00131036, -0.00014973, -0.00087757, -0.00011941, 0.00031927, -0.00019444, -0.00031808, 0.00016851, 0.00017469, -0.00019565, -0.00008848, 0.00017093, -0.00001154, -0.00017015, 0.00004364, 0.00011351, -0.00009514, -0.00008889, 0.00010445, 0.00006392, -0.00008879, -0.00004165, 0.00004784, -0.00000547, -0.00003706, 0.00002837, 0.00003249, -0.00003419, -0.00003681, 0.00000932, 0.00001283, -0.00000057, 0.00001493, 0.00002128, -0.00000565, -0.00002872, -0.00002676, -0.00000845, 0.00002228, 0.00004214, 0.00001710, -0.00002648, -0.00003170, -0.00001018, },
		{ -0.00002701, -0.00001266, -0.00000640, 0.00000717, 0.00002503, 0.00002062, -0.00002130, -0.00008006, -0.00007686, 0.00001753, 0.00005671, -0.00003442, -0.00004182, 0.00007784, -0.00001001, -0.00022121, -0.00002874, 0.00022811, -0.00028773, -0.00082682, 0.00008723, 0.00150106, 0.00112131, -0.00050588, -0.00084494, 0.00009631, 0.00027734, -0.00032016, -0.00024158, 0.00025044, 0.00009592, -0.00023148, -0.00001209, 0.00017427, -0.00007472, -0.00015251, 0.00008268, 0.00007568, -0.00011728, -0.00004100, 0.00011985, 0.00002354, -0.00009746, -0.00001868, 0.00004360, -0.00001737, -0.00002329, 0.00003760, 0.00001968, -0.00004104, -0.00003055, 0.00001098, 0.00001206, 0.00000487, 0.00001773, 0.00001590, -0.00001150, -0.00003178, -0.00002668, 0.00000023, 0.00003343, 0.00003838, 0.00000264, -0.00003019, },
		{ -0.00000825, -0.00002957, -0.00002627, -0.00001608, 0.00000023, 0.00001796, 0.00002214, 0.00001242, -0.00002672, -0.00008914, -0.00007017, 0.00004533, 0.00005182, -0.00006874, -0.00002220, 0.00011460, -0.00006319, -0.00026178, 0.00006176, 0.00023543, -0.00048871, -0.00084449, 0.00045500, 0.00170899, 0.00080910, -0.00080914, -0.00070601, 0.00027976, 0.00016996, -0.00040829, -0.00012742, 0.00029203, 0.00000022, -0.00023052, 0.00006489, 0.00014803, -0.00012840, -0.00011659, 0.00010539, 0.00003060, -0.00011867, 0.00000829, 0.00011558, -0.00001628, -0.00009317, 0.00000136, 0.00003349, -0.00002347, -0.00000679, 0.00004116, 0.00000525, -0.00004497, -0.00002393, 0.00001255, 0.00001261, 0.00000952, 0.00001855, 0.00001041, -0.00001740, -0.00003489, -0.00002322, 0.00001168, 0.00003980, 0.00002908, },
		{ 0.00003981, 0.00001744, -0.00001376, -0.00002806, -0.00003010, -0.00001592, 0.00001333, 0.00002307, 0.00001258, 0.00000895, -0.00003056, -0.00010150, -0.00005579, 0.00007572, 0.00003092, -0.00010152, 0.00001890, 0.00013847, -0.00014144, -0.00027205, 0.00017463, 0.00018227, -0.00070940, -0.00075314, 0.00087144, 0.00178643, 0.00040819, -0.00101503, -0.00048690, 0.00039836, 0.00001994, -0.00044056, 0.00000217, 0.00028596, -0.00009281, -0.00019359, 0.00012734, 0.00009818, -0.00016323, -0.00007088, 0.00010851, -0.00001211, -0.00010030, 0.00004975, 0.00009422, -0.00004865, -0.00007872, 0.00001498, 0.00002048, -0.00002262, 0.00000974, 0.00003868, -0.00000930, -0.00004583, -0.00001706, 0.00001416, 0.00001380, 0.00001304, 0.00001803, 0.00000473, -0.00002369, -0.00003676, -0.00001571, 0.00002328, },
		{ -0.00000466, 0.00003215, 0.00003361, 0.00000711, -0.00001420, -0.00002924, -0.00003651, -0.00000837, 0.00002888, 0.00001853, 0.00000117, 0.00001255, -0.00003742, -0.00011555, -0.00002984, 0.00010220, -0.00000797, -0.00012206, 0.00007902, 0.00013596, -0.00023478, -0.00023801, 0.00028871, 0.00005620, -0.00091234, -0.00053776, 0.00128462, 0.00170912, -0.00002976, -0.00109612, -0.00022804, 0.00043428, -0.00014143, -0.00041169, 0.00012217, 0.00023583, -0.00016490, -0.00012962, 0.00016408, 0.00003557, -0.00017488, -0.00002553, 0.00009375, -0.00004381, -0.00006760, 0.00007636, 0.00006174, -0.00006903, -0.00005880, 0.00002064, 0.00000816, -0.00001512, 0.00002351, 0.00003063, -0.00002239, -0.00004370, -0.00001016, 0.00001571, 0.00001507, 0.00001545, 0.00001660, -0.00000144, -0.00003018, -0.00003597, },
		{ -0.00003608, -0.00003142, 0.00000833, 0.00003588, 0.00002312, 0.00000105, -0.00001192, -0.00003538, -0.00004143, 0.00000797, 0.00004089, 0.00000366, -0.00000586, 0.00002204, -0.00005166, -0.00012654, 0.00000882, 0.00011613, -0.00006166, -0.00011962, 0.00014868, 0.00009698, -0.00032555, -0.00015390, 0.00037591, -0.00014064, -0.00105414, -0.00020465, 0.00163608, 0.00147747, -0.00044567, -0.00104762, 0.00002487, 0.00038770, -0.00028221, -0.00033005, 0.00021098, 0.00015479, -0.00020362, -0.00005302, 0.00017043, -0.00002720, -0.00016473, 0.00001057, 0.00006713, -0.00005906, -0.00002893, 0.00008494, 0.00002584, -0.00007608, -0.00003865, 0.00001901, -0.00000016, -0.00000278, 0.00003220, 0.00001839, -0.00003267, -0.00003893, -0.00000356, 0.00001698, 0.00001608, 0.00001698, 0.00001442, -0.00000839, },
		{ 0.00001141, -0.00001612, -0.00004029, -0.00002272, 0.00002085, 0.00003322, 0.00001143, 0.00000068, -0.00001044, -0.00004651, -0.00004012, 0.00003124, 0.00004330, -0.00001835, -0.00000315, 0.00003280, -0.00007524, -0.00012781, 0.00005708, 0.00010931, -0.00012130, -0.00008706, 0.00021263, 0.00001899, -0.00039193, -0.00002594, 0.00040749, -0.00038873, -0.00109470, 0.00021679, 0.00187220, 0.00111775, -0.00078472, -0.00088773, 0.00023078, 0.00027567, -0.00037724, -0.00021484, 0.00025543, 0.00006169, -0.00020501, 0.00002030, 0.00014879, -0.00007905, -0.00013895, 0.00003205, 0.00003704, -0.00005697, 0.00000672, 0.00007667, -0.00000610, -0.00007168, -0.00002267, 0.00001270, -0.00000235, 0.00001153, 0.00003448, 0.00000394, -0.00003919, -0.00003218, 0.00000239, 0.00001780, 0.00001669, 0.00001782, },
		{ 0.00001689, 0.00001808, 0.00000742, -0.00002428, -0.00004162, -0.00001045, 0.00003029, 0.00002450, 0.00000199, 0.00000539, -0.00001329, -0.00006010, -0.00002877, 0.00005645, 0.00003205, -0.00004114, 0.00001171, 0.00003824, -0.00010629, -0.00011309, 0.00010731, 0.00007701, -0.00017381, -0.00002401, 0.00025328, -0.00009064, -0.00041350, 0.00012688, 0.00036171, -0.00065282, -0.00100700, 0.00067761, 0.00195541, 0.00067761, -0.00100700, -0.00065282, 0.00036171, 0.00012688, -0.00041350, -0.00009064, 0.00025328, -0.00002401, -0.00017381, 0.00007701, 0.00010731, -0.00011309, -0.00010629, 0.00003824, 0.00001171, -0.00004114, 0.00003205, 0.00005645, -0.00002877, -0.00006010, -0.00001329, 0.00000539, 0.00000199, 0.00002450, 0.00003029, -0.00001045, -0.00004162, -0.00002428, 0.00000742, 0.00001808, },
		{ 0.00001141, 0.00001782, 0.00001669, 0.00001780, 0.00000239, -0.00003218, -0.00003919, 0.00000394, 0.00003448, 0.00001153, -0.00000235, 0.00001270, -0.00002267, -0.00007168, -0.00000610, 0.00007667, 0.00000672, -0.00005697, 0.00003704, 0.00003205, -0.00013895, -0.00007905, 0.00014879, 0.00002030, -0.00020501, 0.00006169, 0.00025543, -0.00021484, -0.00037724, 0.00027567, 0.00023078, -0.00088773, -0.00078472, 0.00111775, 0.00187220, 0.00021679, -0.00109470, -0.00038873, 0.00040749, -0.00002594, -0.00039193, 0.00001899, 0.00021263, -0.00008706, -0.00012130, 0.00010931, 0.00005708, -0.00012781, -0.00007524, 0.00003280, -0.00000315, -0.00001835, 0.00004330, 0.00003124, -0.00004012, -0.00004651, -0.00001044, 0.00000068, 0.00001143, 0.00003322, 0.00002085, -0.00002272, -0.00004029, -0.00001612, },
		{ -0.00003608, -0.00000839, 0.00001442, 0.00001698, 0.00001608, 0.00001698, -0.00000356, -0.00003893, -0.00003267, 0.00001839, 0.00003220, -0.00000278, -0.00000016, 0.00001901, -0.00003865, -0.00007608, 0.00002584, 0.00008494, -0.00002893, -0.00005906, 0.00006713, 0.00001057, -0.00016473, -0.00002720, 0.00017043, -0.00005302, -0.00020362, 0.00015479, 0.00021098, -0.00033005, -0.00028221, 0.00038770, 0.00002487, -0.00104762, -0.00044567, 0.00147747, 0.00163608, -0.00020465, -0.00105414, -0.00014064, 0.00037591, -0.00015390, -0.00032555, 0.00009698, 0.00014868, -0.00011962, -0.00006166, 0.00011613, 0.00000882, -0.00012654, -0.00005166, 0.00002204, -0.00000586, 0.00000366, 0.00004089, 0.00000797, -0.00004143, -0.00003538, -0.00001192, 0.00000105, 0.00002312, 0.00003588, 0.00000833, -0.00003142, },
		{ -0.00000466, -0.00003597, -0.00003018, -0.00000144, 0.00001660, 0.00001545, 0.00001507, 0.00001571, -0.00001016, -0.00004370, -0.00002239, 0.00003063, 0.00002351, -0.00001512, 0.00000816, 0.00002064, -0.00005880, -0.00006903, 0.00006174, 0.00007636, -0.00006760, -0.00004381, 0.00009375, -0.00002553, -0.00017488, 0.00003557, 0.00016408, -0.00012962, -0.00016490, 0.00023583, 0.00012217, -0.00041169, -0.00014143, 0.00043428, -0.00022804, -0.00109612, -0.00002976, 0.00170912, 0.00128462, -0.00053776, -0.00091234, 0.00005620, 0.00028871, -0.00023801, -0.00023478, 0.00013596, 0.00007902, -0.00012206, -0.00000797, 0.00010220, -0.00002984, -0.00011555, -0.00003742, 0.00001255, 0.00000117, 0.00001853, 0.00002888, -0.00000837, -0.00003651, -0.00002924, -0.00001420, 0.00000711, 0.00003361, 0.00003215, },
		{ 0.00003981, 0.00002328, -0.00001571, -0.00003676, -0.00002369, 0.00000473, 0.00001803, 0.00001304, 0.00001380, 0.00001416, -0.00001706, -0.00004583, -0.00000930, 0.00003868, 0.00000974, -0.00002262, 0.00002048, 0.00001498, -0.00007872, -0.00004865, 0.00009422, 0.00004975, -0.00010030, -0.00001211, 0.00010851, -0.00007088, -0.00016323, 0.00009818, 0.00012734, -0.00019359, -0.00009281, 0.00028596, 0.00000217, -0.00044056, 0.00001994, 0.00039836, -0.00048690, -0.00101503, 0.00040819, 0.00178643, 0.00087144, -0.00075314, -0.00070940, 0.00018227, 0.00017463, -0.00027205, -0.00014144, 0.00013847, 0.00001890, -0.00010152, 0.00003092, 0.00007572, -0.00005579, -0.00010150, -0.00003056, 0.00000895, 0.00001258, 0.00002307, 0.00001333, -0.00001592, -0.00003010, -0.00002806, -0.00001376, 0.00001744, },
		{ -0.00000825, 0.00002908, 0.00003980, 0.00001168, -0.00002322, -0.00003489, -0.00001740, 0.00001041, 0.00001855, 0.00000952, 0.00001261, 0.00001255, -0.00002393, -0.00004497, 0.00000525, 0.00004116, -0.00000679, -0.00002347, 0.00003349, 0.00000136, -0.00009317, -0.00001628, 0.00011558, 0.00000829, -0.00011867, 0.00003060, 0.00010539, -0.00011659, -0.00012840, 0.00014803, 0.00006489, -0.00023052, 0.00000022, 0.00029203, -0.00012742, -0.00040829, 0.00016996, 0.00027976, -0.00070601, -0.00080914, 0.00080910, 0.00170899, 0.00045500, -0.00084449, -0.00048871, 0.00023543, 0.00006176, -0.00026178, -0.00006319, 0.00011460, -0.00002220, -0.00006874, 0.00005182, 0.00004533, -0.00007017, -0.00008914, -0.00002672, 0.00001242, 0.00002214, 0.00001796, 0.00000023, -0.00001608, -0.00002627, -0.00002957, },
		{ -0.00002701, -0.00003019, 0.00000264, 0.00003838, 0.00003343, 0.00000023, -0.00002668, -0.00003178, -0.00001150, 0.00001590, 0.00001773, 0.00000487, 0.00001206, 0.00001098, -0.00003055, -0.00004104, 0.00001968, 0.00003760, -0.00002329, -0.00001737, 0.00004360, -0.00001868, -0.00009746, 0.00002354, 0.00011985, -0.00004100, -0.00011728, 0.00007568, 0.00008268, -0.00015251, -0.00007472, 0.00017427, -0.00001209, -0.00023148, 0.00009592, 0.00025044, -0.00024158, -0.00032016, 0.00027734, 0.00009631, -0.00084494, -0.00050588, 0.00112131, 0.00150106, 0.00008723, -0.00082682, -0.00028773, 0.00022811, -0.00002874, -0.00022121, -0.00001001, 0.00007784, -0.00004182, -0.00003442, 0.00005671, 0.00001753, -0.00007686, -0.00008006, -0.00002130, 0.00002062, 0.00002503, 0.00000717, -0.00000640, -0.00001266, },
		{ -0.00000569, -0.00001018, -0.00003170, -0.00002648, 0.00001710, 0.00004214, 0.00002228, -0.00000845, -0.00002676, -0.00002872, -0.00000565, 0.00002128, 0.00001493, -0.00000057, 0.00001283, 0.00000932, -0.00003681, -0.00003419, 0.00003249, 0.00002837, -0.00003706, -0.00000547, 0.00004784, -0.00004165, -0.00008879, 0.00006392, 0.00010445, -0.00008889, -0.00009514, 0.00011351, 0.00004364, -0.00017015, -0.00001154, 0.00017093, -0.00008848, -0.00019565, 0.00017469, 0.00016851, -0.00031808, -0.00019444, 0.00031927, -0.00011941, -0.00087757, -0.00014973, 0.00131036, 0.00120510, -0.00019554, -0.00073023, -0.00013139, 0.00018158, -0.00008614, -0.00016739, 0.00001662, 0.00004080, -0.00004351, -0.00000629, 0.00005082, -0.00000469, -0.00008016, -0.00007273, -0.00001150, 0.00002882, 0.00001944, -0.00000371, },
		{ 0.00000719, -0.00000952, 0.00000008, -0.00001221, -0.00003747, -0.00001636, 0.00003164, 0.00003858, 0.00000922, -0.00001276, -0.00002497, -0.00002639, 0.00000085, 0.00002615, 0.00000951, -0.00000596, 0.00001548, 0.00000708, -0.00004254, -0.00002471, 0.00004233, 0.00001464, -0.00004598, 0.00000990, 0.00004465, -0.00006300, -0.00006717, 0.00009727, 0.00007095, -0.00012613, -0.00005599, 0.00013607, -0.00000422, -0.00016506, 0.00004911, 0.00013875, -0.00014904, -0.00013083, 0.00022056, 0.00006298, -0.00034321, -0.00005819, 0.00028721, -0.00032721, -0.00079792, 0.00020711, 0.00136461, 0.00087191, -0.00037721, -0.00059118, -0.00002951, 0.00011899, -0.00011026, -0.00011527, 0.00002203, 0.00001197, -0.00003428, 0.00001234, 0.00003999, -0.00002171, -0.00008283, -0.00006388, 0.00000245, 0.00003190, },
		{ 0.00001730, 0.00002648, -0.00000708, -0.00000742, 0.00000653, -0.00002001, -0.00004028, 0.00000001, 0.00004220, 0.00002789, -0.00000240, -0.00001246, -0.00002316, -0.00002468, 0.00000864, 0.00002957, 0.00000120, -0.00001005, 0.00002016, 0.00000352, -0.00004746, -0.00001299, 0.00004820, -0.00000190, -0.00004882, 0.00002587, 0.00003422, -0.00007822, -0.00003553, 0.00011706, 0.00002483, -0.00014572, -0.00000731, 0.00013904, -0.00005119, -0.00013818, 0.00009558, 0.00008507, -0.00018232, -0.00005155, 0.00022542, -0.00004421, -0.00031534, 0.00005940, 0.00018917, -0.00048839, -0.00062109, 0.00051554, 0.00129547, 0.00054992, -0.00046160, -0.00044382, 0.00002166, 0.00005946, -0.00010872, -0.00007430, 0.00001506, -0.00000558, -0.00002150, 0.00002226, 0.00002848, -0.00003592, -0.00008504, -0.00005044, },
		{ -0.00008455, -0.00003136, 0.00002806, 0.00001238, -0.00001761, 0.00000215, 0.00000887, -0.00003203, -0.00003645, 0.00002009, 0.00004534, 0.00001220, -0.00000968, -0.00000868, -0.00002287, -0.00002275, 0.00001795, 0.00003024, -0.00000963, -0.00001133, 0.00002638, -0.00000217, -0.00005094, 0.00000044, 0.00004940, -0.00001929, -0.00004541, 0.00003967, 0.00001840, -0.00008403, 0.00000088, 0.00011943, -0.00002584, -0.00014450, 0.00004162, 0.00012280, -0.00008797, -0.00009561, 0.00011969, 0.00002202, -0.00018363, 0.00002476, 0.00019098, -0.00013153, -0.00024523, 0.00013432, 0.00004783, -0.00057444, -0.00037937, 0.00073911, 0.00113268, 0.00027633, -0.00046767, -0.00031340, 0.00003493, 0.00001443, -0.00009264, -0.00004737, 0.00000431, -0.00001356, -0.00001032, 0.00002669, 0.00001777, -0.00004974, },
		{ 0.00000688, -0.00006392, -0.00007812, -0.00000857, 0.00003002, -0.00000712, -0.00001967, 0.00001561, 0.00000377, -0.00004442, -0.00002400, 0.00003969, 0.00003930, -0.00000495, -0.00001094, -0.00000353, -0.00002482, -0.00001926, 0.00002826, 0.00002683, -0.00002179, -0.00000852, 0.00003292, -0.00001053, -0.00005212, 0.00001480, 0.00004564, -0.00003557, -0.00003648, 0.00004918, 0.00000019, -0.00007914, 0.00003570, 0.00010402, -0.00007226, -0.00012373, 0.00008209, 0.00009207, -0.00010815, -0.00004690, 0.00011878, -0.00003678, -0.00015608, 0.00008236, 0.00012813, -0.00018343, -0.00015301, 0.00015362, -0.00010518, -0.00057314, -0.00011452, 0.00086003, 0.00091598, 0.00007193, -0.00042226, -0.00021322, 0.00002636, -0.00001297, -0.00007229, -0.00003197, -0.00000454, -0.00001628, -0.00000254, 0.00002908, },
		{ 0.00000305, 0.00003128, -0.00000633, -0.00007680, -0.00006338, 0.00001334, 0.00002078, -0.00002620, -0.00001135, 0.00002754, -0.00000938, -0.00005232, -0.00000354, 0.00005408, 0.00002456, -0.00001962, -0.00000613, 0.00000054, -0.00002869, -0.00001298, 0.00003824, 0.00001847, -0.00003332, -0.00000098, 0.00003801, -0.00002150, -0.00004993, 0.00002902, 0.00003700, -0.00004901, -0.00002344, 0.00005335, -0.00001705, -0.00006456, 0.00006293, 0.00007400, -0.00010692, -0.00008850, 0.00010805, 0.00005446, -0.00010972, -0.00000239, 0.00009625, -0.00007984, -0.00010940, 0.00011122, 0.00005369, -0.00019408, -0.00006262, 0.00011837, -0.00023709, -0.00049046, 0.00013175, 0.00088052, 0.00068568, -0.00005953, -0.00035240, -0.00014524, 0.00000999, -0.00002590, -0.00005415, -0.00002299, -0.00000988, -0.00001804, },
		{ -0.00001345, -0.00002114, 0.00000905, 0.00003274, -0.00002330, -0.00008482, -0.00004047, 0.00002855, 0.00000146, -0.00003848, 0.00000565, 0.00003255, -0.00002807, -0.00005154, 0.00002175, 0.00005938, 0.00000378, -0.00002841, 0.00000318, 0.00000150, -0.00003315, -0.00000321, 0.00004593, 0.00000516, -0.00004184, 0.00001093, 0.00003964, -0.00003420, -0.00004351, 0.00004169, 0.00002403, -0.00005817, -0.00000809, 0.00005219, -0.00003040, -0.00004329, 0.00007827, 0.00003520, -0.00012518, -0.00004623, 0.00011728, 0.00001820, -0.00009531, 0.00002955, 0.00006022, -0.00010062, -0.00005699, 0.00010936, -0.00001437, -0.00016836, 0.00000457, 0.00004273, -0.00032187, -0.00034795, 0.00032647, 0.00081932, 0.00047455, -0.00012783, -0.00027928, -0.00010351, -0.00000537, -0.00003039, -0.00003997, -0.00001562, },
		{ -0.00002783, -0.00000759, -0.00001840, -0.00002508, 0.00001753, 0.00003104, -0.00004335, -0.00008404, -0.00001271, 0.00003220, -0.00002319, -0.00003927, 0.00002639, 0.00002722, -0.00004744, -0.00003997, 0.00004691, 0.00005367, -0.00001890, -0.00002940, 0.00001433, -0.00000167, -0.00003626, 0.00000975, 0.00004921, -0.00001196, -0.00004508, 0.00002567, 0.00003617, -0.00004692, -0.00003246, 0.00005119, 0.00000777, -0.00006212, 0.00000774, 0.00004668, -0.00003800, -0.00001954, 0.00007995, -0.00000534, -0.00012601, -0.00000470, 0.00011148, -0.00001006, -0.00007107, 0.00004463, 0.00002107, -0.00009896, -0.00001202, 0.00008277, -0.00006183, -0.00011985, 0.00003534, -0.00005036, -0.00034597, -0.00017658, 0.00045122, 0.00070506, 0.00030280, -0.00015045, -0.00021530, -0.00007888, -0.00001684, -0.00003189, },
		{ -0.00002587, -0.00003291, -0.00001448, 0.00000027, -0.00002711, -0.00002725, 0.00002850, 0.00002342, -0.00006326, -0.00007206, 0.00001399, 0.00002234, -0.00004617, -0.00002727, 0.00004428, 0.00001139, -0.00006170, -0.00001844, 0.00006657, 0.00003756, -0.00003893, -0.00002252, 0.00002425, -0.00000870, -0.00003607, 0.00002453, 0.00004637, -0.00003063, -0.00004151, 0.00004075, 0.00002685, -0.00005738, -0.00001726, 0.00005595, -0.00001020, -0.00006043, 0.00002239, 0.00003832, -0.00003940, 0.00000228, 0.00006896, -0.00004098, -0.00011180, 0.00002978, 0.00009529, -0.00002679, -0.00004461, 0.00004335, -0.00001164, -0.00008044, 0.00001620, 0.00004305, -0.00008184, -0.00006637, 0.00002753, -0.00013568, -0.00031063, -0.00000873, 0.00050378, 0.00056823, 0.00017706, -0.00014605, -0.00016429, -0.00006331, },
		{ -0.00012414, -0.00005241, -0.00003548, -0.00003259, 0.00000225, 0.00000488, -0.00003959, -0.00002458, 0.00003958, 0.00000856, -0.00007812, -0.00004959, 0.00003290, 0.00000104, -0.00006035, -0.00000526, 0.00005319, -0.00001167, -0.00006605, 0.00000938, 0.00007636, 0.00001397, -0.00005245, -0.00000949, 0.00003036, -0.00001812, -0.00003123, 0.00003885, 0.00003668, -0.00004787, -0.00003085, 0.00005318, 0.00001218, -0.00006323, 0.00000067, 0.00005479, -0.00002791, -0.00005327, 0.00003455, 0.00002877, -0.00003548, 0.00001856, 0.00004859, -0.00006686, -0.00008732, 0.00005362, 0.00007460, -0.00003209, -0.00002274, 0.00003032, -0.00003173, -0.00005420, 0.00002444, 0.00000357, -0.00007614, -0.00002461, -0.00001017, -0.00019242, -0.00023007, 0.00012930, 0.00049573, 0.00043418, 0.00009301, -0.00012947, },
		{ 0.00004013, -0.00010920, -0.00009066, -0.00004580, -0.00004749, -0.00002788, 0.00002184, 0.00000287, -0.00005322, -0.00001536, 0.00004682, -0.00001218, -0.00008304, -0.00002086, 0.00003876, -0.00002584, -0.00006093, 0.00002064, 0.00004954, -0.00003614, -0.00005816, 0.00003800, 0.00007417, -0.00001259, -0.00005720, 0.00000674, 0.00003116, -0.00002767, -0.00002148, 0.00005005, 0.00002073, -0.00006056, -0.00001428, 0.00006017, -0.00000602, -0.00006271, 0.00001906, 0.00004734, -0.00004319, -0.00004144, 0.00004333, 0.00001949, -0.00002810, 0.00002717, 0.00002341, -0.00008065, -0.00005829, 0.00006626, 0.00005475, -0.00002905, -0.00000959, 0.00001241, -0.00003771, -0.00002979, 0.00001581, -0.00002464, -0.00005369, -0.00000539, -0.00006158, -0.00020933, -0.00012658, 0.00022240, 0.00044695, 0.00031879, },
		{ 0.00037905, 0.00022759, 0.00000698, -0.00008767, -0.00006089, -0.00004560, -0.00006091, -0.00001569, 0.00004101, -0.00000738, -0.00006373, -0.00000058, 0.00004647, -0.00003453, -0.00007527, 0.00000725, 0.00002982, -0.00005030, -0.00004709, 0.00004279, 0.00003349, -0.00005547, -0.00003899, 0.00006155, 0.00006058, -0.00003724, -0.00005290, 0.00002280, 0.00002653, -0.00003493, -0.00000779, 0.00005569, 0.00000051, -0.00006623, 0.00000564, 0.00005986, -0.00002486, -0.00005513, 0.00003515, 0.00003423, -0.00005403, -0.00002627, 0.00004830, 0.00001147, -0.00001954, 0.00002775, -0.00000180, -0.00008264, -0.00002986, 0.00006963, 0.00003920, -0.00002231, -0.00000585, -0.00000365, -0.00003259, -0.00001427, -0.00000198, -0.00003606, -0.00002719, -0.00001092, -0.00010830, -0.00018671, -0.00002391, 0.00026820, },
		{ 0.00005895, 0.00027507, 0.00030954, 0.00015792, -0.00001479, -0.00006361, -0.00003489, -0.00005396, -0.00007204, 0.00000498, 0.00005492, -0.00002460, -0.00006705, 0.00001612, 0.00003685, -0.00005253, -0.00005567, 0.00002726, 0.00000881, -0.00006461, -0.00002250, 0.00005445, 0.00000901, -0.00006436, -0.00001253, 0.00007544, 0.00003860, -0.00005589, -0.00004114, 0.00003573, 0.00001771, -0.00003799, 0.00000777, 0.00005425, -0.00002097, -0.00006361, 0.00002551, 0.00005185, -0.00004106, -0.00004125, 0.00004633, 0.00001717, -0.00005893, -0.00000955, 0.00004943, 0.00000519, -0.00001190, 0.00002151, -0.00002310, -0.00007511, -0.00000563, 0.00006699, 0.00002893, -0.00001632, -0.00000911, -0.00001332, -0.00002223, -0.00001018, -0.00001964, -0.00003175, -0.00000862, -0.00003492, -0.00013529, -0.00013510, },
		{ -0.00013506, -0.00007120, 0.00011198, 0.00025748, 0.00024843, 0.00010303, -0.00002866, -0.00003519, -0.00001562, -0.00007085, -0.00007593, 0.00003194, 0.00005933, -0.00004464, -0.00006128, 0.00002957, 0.00001955, -0.00006050, -0.00002900, 0.00003360, -0.00001757, -0.00006394, 0.00000565, 0.00005195, -0.00001731, -0.00006039, 0.00001538, 0.00007750, 0.00001280, -0.00006613, -0.00002477, 0.00004365, 0.00000680, -0.00003595, 0.00002251, 0.00004550, -0.00004031, -0.00005308, 0.00004185, 0.00003740, -0.00005169, -0.00002319, 0.00005076, -0.00000129, -0.00005726, 0.00000673, 0.00004710, 0.00000063, -0.00000674, 0.00001074, -0.00003806, -0.00006143, 0.00001282, 0.00006167, 0.00002276, -0.00001390, -0.00001519, -0.00001539, -0.00001295, -0.00001539, -0.00002967, -0.00001833, -0.00000540, -0.00006533, },
		{ -0.00001814, -0.00008862, -0.00010931, -0.00001229, 0.00013518, 0.00023055, 0.00019782, 0.00005634, -0.00003392, -0.00000265, -0.00000715, -0.00009302, -0.00006868, 0.00006011, 0.00005249, -0.00006189, -0.00004775, 0.00003514, -0.00000061, -0.00005529, -0.00000280, 0.00002468, -0.00004053, -0.00004801, 0.00002886, 0.00003589, -0.00003826, -0.00004474, 0.00003868, 0.00006839, -0.00001201, -0.00006756, -0.00000701, 0.00004604, -0.00000380, -0.00002917, 0.00003376, 0.00003070, -0.00005440, -0.00003657, 0.00005189, 0.00001916, -0.00005492, -0.00000404, 0.00004785, -0.00001826, -0.00004944, 0.00002071, 0.00004200, -0.00000253, -0.00000477, -0.00000186, -0.00004590, -0.00004498, 0.00002567, 0.00005600, 0.00001833, -0.00001559, -0.00001992, -0.00001186, -0.00000919, -0.00002436, -0.00002877, -0.00000516, },
		{ -0.00001879, -0.00000067, -0.00004086, -0.00009446, -0.00006756, 0.00002881, 0.00013663, 0.00020532, 0.00015404, 0.00001461, -0.00002797, 0.00003060, -0.00001230, -0.00011459, -0.00004932, 0.00008323, 0.00003620, -0.00007140, -0.00003082, 0.00003072, -0.00001692, -0.00003782, 0.00001494, 0.00000382, -0.00005198, -0.00002141, 0.00003996, 0.00001103, -0.00004830, -0.00002177, 0.00005283, 0.00005121, -0.00003193, -0.00006154, 0.00000927, 0.00004362, -0.00001207, -0.00001907, 0.00003951, 0.00001225, -0.00006120, -0.00001713, 0.00005426, 0.00000056, -0.00005052, 0.00001287, 0.00003846, -0.00003109, -0.00003694, 0.00003094, 0.00003505, -0.00000481, -0.00000593, -0.00001385, -0.00004724, -0.00002841, 0.00003418, 0.00005085, 0.00001331, -0.00001981, -0.00002066, -0.00000672, -0.00001207, -0.00003060, },
		{ -0.00001923, -0.00002935, -0.00000569, -0.00000925, -0.00006340, -0.00007901, -0.00002359, 0.00004700, 0.00012834, 0.00018629, 0.00011135, -0.00002117, -0.00000921, 0.00005871, -0.00003073, -0.00012891, -0.00002071, 0.00009607, 0.00001542, -0.00007087, -0.00001644, 0.00001797, -0.00002316, -0.00001323, 0.00001848, -0.00002154, -0.00004721, 0.00000764, 0.00003541, -0.00001525, -0.00004513, 0.00000235, 0.00005596, 0.00003044, -0.00004483, -0.00005053, 0.00002224, 0.00003795, -0.00001680, -0.00000782, 0.00003887, -0.00000675, -0.00006013, 0.00000176, 0.00004926, -0.00001501, -0.00003989, 0.00002474, 0.00002479, -0.00003797, -0.00002196, 0.00003663, 0.00002721, -0.00000674, -0.00000949, -0.00002344, -0.00004357, -0.00001328, 0.00003982, 0.00004577, 0.00000642, -0.00002392, -0.00001715, -0.00000403, },
		{ -0.00001132, -0.00000625, -0.00002601, -0.00001942, 0.00000293, -0.00002961, -0.00007512, -0.00004600, 0.00000934, 0.00004543, 0.00012157, 0.00017153, 0.00006562, -0.00004680, 0.00002085, 0.00007545, -0.00005824, -0.00013097, 0.00001120, 0.00009653, -0.00000345, -0.00006172, -0.00000984, 0.00000211, -0.00001589, 0.00001045, 0.00000648, -0.00004198, -0.00002663, 0.00002985, 0.00001647, -0.00003517, -0.00003043, 0.00002154, 0.00004914, 0.00001057, -0.00005048, -0.00003726, 0.00003120, 0.00003089, -0.00001781, 0.00000226, 0.00003224, -0.00002327, -0.00005217, 0.00001703, 0.00003876, -0.00002506, -0.00002571, 0.00003000, 0.00000979, -0.00003836, -0.00000699, 0.00003781, 0.00001940, -0.00000875, -0.00001442, -0.00002969, -0.00003663, -0.00000008, 0.00004358, 0.00003963, -0.00000214, -0.00002540, },
		{ -0.00001088, -0.00002303, -0.00000630, -0.00001314, -0.00002754, -0.00000377, 0.00000080, -0.00005522, -0.00006880, -0.00000528, 0.00002271, 0.00003333, 0.00012294, 0.00015513, 0.00001670, -0.00005749, 0.00005700, 0.00007665, -0.00008779, -0.00011935, 0.00003926, 0.00008646, -0.00001488, -0.00004874, -0.00001340, -0.00000966, 0.00000389, 0.00002483, -0.00001718, -0.00004915, 0.00000414, 0.00003771, -0.00001115, -0.00004301, -0.00000915, 0.00003155, 0.00003576, -0.00000512, -0.00005024, -0.00002395, 0.00003650, 0.00002402, -0.00001583, 0.00000928, 0.00002117, -0.00003495, -0.00003949, 0.00002671, 0.00002560, -0.00002855, -0.00001125, 0.00002867, -0.00000351, -0.00003307, 0.00000575, 0.00003522, 0.00001226, -0.00001112, -0.00001958, -0.00003245, -0.00002794, 0.00001129, 0.00004569, 0.00003145, },
		{ 0.00004573, 0.00002101, -0.00001774, -0.00001732, -0.00000492, -0.00002190, -0.00002076, 0.00001159, -0.00001444, -0.00007654, -0.00004321, 0.00003061, 0.00001518, 0.00002212, 0.00013280, 0.00013071, -0.00003123, -0.00005059, 0.00009153, 0.00006213, -0.00011168, -0.00009686, 0.00005767, 0.00007115, -0.00001632, -0.00003836, -0.00002537, -0.00001063, 0.00003068, 0.00002403, -0.00004435, -0.00003858, 0.00003581, 0.00002797, -0.00003908, -0.00003685, 0.00001207, 0.00003110, 0.00002033, -0.00001518, -0.00004624, -0.00001190, 0.00003904, 0.00001833, -0.00001222, 0.00001211, 0.00000792, -0.00004060, -0.00002490, 0.00003031, 0.00001287, -0.00002610, 0.00000055, 0.00002226, -0.00001280, -0.00002397, 0.00001478, 0.00003013, 0.00000609, -0.00001394, -0.00002398, -0.00003212, -0.00001855, 0.00002115, },
		{ -0.00000905, 0.00002956, 0.00004301, 0.00000910, -0.00002105, -0.00001036, -0.00000845, -0.00002821, -0.00000591, 0.00001982, -0.00004009, -0.00008442, -0.00000373, 0.00005061, -0.00000746, 0.00002116, 0.00014586, 0.00009474, -0.00007121, -0.00002750, 0.00011658, 0.00003600, -0.00012408, -0.00006971, 0.00006401, 0.00005741, -0.00000934, -0.00003622, -0.00004015, 0.00000287, 0.00005599, 0.00000695, -0.00006503, -0.00001139, 0.00005823, 0.00000271, -0.00005887, -0.00001906, 0.00002704, 0.00002201, 0.00000698, -0.00002004, -0.00004053, -0.00000148, 0.00003988, 0.00001409, -0.00000851, 0.00001062, -0.00000507, -0.00004034, -0.00001110, 0.00002876, 0.00000302, -0.00001965, 0.00000775, 0.00001327, -0.00001697, -0.00001350, 0.00001962, 0.00002391, 0.00000082, -0.00001711, -0.00002691, -0.00002936, },
		{ -0.00002796, -0.00002493, 0.00000026, 0.00003624, 0.00003713, -0.00000270, -0.00002025, -0.00000481, -0.00001589, -0.00002804, 0.00001327, 0.00001585, -0.00006907, -0.00007336, 0.00003928, 0.00004875, -0.00003464, 0.00003461, 0.00015393, 0.00004847, -0.00009635, 0.00000612, 0.00012678, 0.00000557, -0.00012309, -0.00004532, 0.00006015, 0.00005100, 0.00000094, -0.00004483, -0.00004998, 0.00002981, 0.00007102, -0.00002207, -0.00007061, 0.00002573, 0.00006360, -0.00003124, -0.00006462, 0.00000456, 0.00003183, 0.00000831, -0.00000175, -0.00002142, -0.00003446, 0.00000765, 0.00003974, 0.00001098, -0.00000601, 0.00000560, -0.00001574, -0.00003534, -0.00000007, 0.00002399, -0.00000269, -0.00001187, 0.00000992, 0.00000448, -0.00001635, -0.00000398, 0.00002081, 0.00001772, -0.00000388, -0.00002026, },
		{ -0.00000846, -0.00002288, -0.00002700, -0.00001948, 0.00000916, 0.00004067, 0.00002830, -0.00001263, -0.00001620, -0.00000284, -0.00002424, -0.00001937, 0.00003071, -0.00000157, -0.00009212, -0.00004364, 0.00007382, 0.00002600, -0.00005481, 0.00006021, 0.00014942, -0.00000220, -0.00010262, 0.00004180, 0.00012108, -0.00002113, -0.00011130, -0.00002969, 0.00005155, 0.00005446, 0.00000762, -0.00006231, -0.00004764, 0.00006443, 0.00006970, -0.00005429, -0.00005684, 0.00006241, 0.00004896, -0.00006437, -0.00005477, 0.00002668, 0.00002589, -0.00000519, -0.00000537, -0.00002143, -0.00002839, 0.00001601, 0.00003893, 0.00000834, -0.00000551, -0.00000159, -0.00002279, -0.00002746, 0.00000735, 0.00001827, -0.00000439, -0.00000526, 0.00000807, -0.00000189, -0.00001245, 0.00000303, 0.00001959, 0.00001216, },
		{ 0.00001744, 0.00000720, -0.00001320, -0.00002431, -0.00002423, -0.00001361, 0.00001737, 0.00004235, 0.00001746, -0.00001941, -0.00001078, -0.00000508, -0.00002957, -0.00000319, 0.00004007, -0.00002902, -0.00010092, -0.00000149, 0.00009004, -0.00001021, -0.00005934, 0.00009053, 0.00012851, -0.00004870, -0.00009061, 0.00007090, 0.00010311, -0.00003798, -0.00009481, -0.00002522, 0.00004517, 0.00006602, 0.00000472, -0.00008262, -0.00002911, 0.00009787, 0.00005094, -0.00007926, -0.00002552, 0.00008773, 0.00001724, -0.00008744, -0.00003249, 0.00004077, 0.00001196, -0.00001456, -0.00000531, -0.00002161, -0.00002189, 0.00002389, 0.00003731, 0.00000548, -0.00000720, -0.00000934, -0.00002586, -0.00001863, 0.00001144, 0.00001346, -0.00000345, -0.00000149, 0.00000423, -0.00000479, -0.00000744, 0.00000703, },
		{ -0.00000335, 0.00000866, 0.00001551, 0.00000228, -0.00001803, -0.00002391, -0.00002010, -0.00000783, 0.00002453, 0.00004108, 0.00000595, -0.00002271, -0.00000615, -0.00001027, -0.00002856, 0.00001648, 0.00003702, -0.00005930, -0.00009090, 0.00004265, 0.00008335, -0.00004848, -0.00004534, 0.00011575, 0.00009268, -0.00008270, -0.00006562, 0.00008731, 0.00007985, -0.00004286, -0.00008095, -0.00002994, 0.00004689, 0.00008010, -0.00001047, -0.00009747, 0.00000480, 0.00012095, 0.00001930, -0.00008824, 0.00001580, 0.00009358, -0.00002362, -0.00009427, -0.00000455, 0.00004312, -0.00000508, -0.00001787, -0.00000401, -0.00002244, -0.00001423, 0.00003113, 0.00003453, 0.00000192, -0.00001071, -0.00001623, -0.00002541, -0.00001034, 0.00001327, 0.00001052, -0.00000185, -0.00000090, 0.00000072, -0.00000451, },
		},
		{ { -0.00001122, -0.00000873, -0.00000719, -0.00000527, 0.00000780, 0.00001837, 0.00000503, -0.00001220, -0.00000394, 0.00000801, -0.00000616, -0.00001981, -0.00000260, 0.00001631, 0.00000647, -0.00000781, -0.00000345, 0.00000286, 0.00000242, 0.00000321, -0.00000553, -0.00002004, -0.00000369, 0.00002247, -0.00001622, -0.00007448, -0.00000902, 0.00012698, 0.00008616, -0.00013120, -0.00018242, 0.00005843, 0.00023008, 0.00005843, -0.00018242, -0.00013120, 0.00008616, 0.00012698, -0.00000902, -0.00007448, -0.00001622, 0.00002247, -0.00000369, -0.00002004, -0.00000553, 0.00000321, 0.00000242, 0.00000286, -0.00000345, -0.00000781, 0.00000647, 0.00001631, -0.00000260, -0.00001981, -0.00000616, 0.00000801, -0.00000394, -0.00001220, 0.00000503, 0.00001837, 0.00000780, -0.00000527, -0.00000719, -0.00000873, },
		{ -0.00000616, -0.00001166, -0.00001122, -0.00000583, -0.00000810, -0.00000912, 0.00000800, 0.00002216, 0.00000648, -0.00001322, -0.00000417, 0.00000771, -0.00000749, -0.00001949, -0.00000062, 0.00001618, 0.00000423, -0.00000877, -0.00000286, 0.00000467, 0.00000723, 0.00000743, -0.00000962, -0.00002867, -0.00000375, 0.00003003, -0.00001584, -0.00008173, -0.00000786, 0.00013849, 0.00009177, -0.00013611, -0.00018617, 0.00005843, 0.00022709, 0.00005813, -0.00017138, -0.00012169, 0.00007701, 0.00011037, -0.00001057, -0.00006539, -0.00001669, 0.00001370, -0.00000369, -0.00001072, -0.00000108, -0.00000091, -0.00000248, 0.00000074, -0.00000409, -0.00000664, 0.00000840, 0.00001578, -0.00000443, -0.00001933, -0.00000469, 0.00000782, -0.00000369, -0.00001072, 0.00000349, 0.00001404, 0.00000748, -0.00000116, },
		{ 0.00000704, 0.00000318, -0.00000504, -0.00001459, -0.00001122, -0.00000297, -0.00000891, -0.00001270, 0.00000809, 0.00002543, 0.00000783, -0.00001383, -0.00000440, 0.00000698, -0.00000867, -0.00001852, 0.00000144, 0.00001552, 0.00000181, -0.00000956, -0.00000235, 0.00000616, 0.00001172, 0.00001145, -0.00001320, -0.00003618, -0.00000384, 0.00003604, -0.00001558, -0.00008677, -0.00000714, 0.00014438, 0.00009366, -0.00013611, -0.00018242, 0.00005813, 0.00021835, 0.00005766, -0.00015357, -0.00010817, 0.00006461, 0.00008940, -0.00001245, -0.00005493, -0.00001726, 0.00000416, -0.00000380, -0.00000123, 0.00000356, -0.00000463, -0.00000724, -0.00000171, -0.00000478, -0.00000525, 0.00000986, 0.00001446, -0.00000606, -0.00001791, -0.00000308, 0.00000706, -0.00000343, -0.00000875, 0.00000187, 0.00000921, },
		{ 0.00000018, 0.00000390, 0.00000649, 0.00000769, -0.00000383, -0.00001749, -0.00001122, -0.00000020, -0.00000960, -0.00001597, 0.00000807, 0.00002819, 0.00000907, -0.00001407, -0.00000461, 0.00000592, -0.00000970, -0.00001706, 0.00000348, 0.00001451, -0.00000065, -0.00001018, -0.00000190, 0.00000738, 0.00001571, 0.00001502, -0.00001614, -0.00004217, -0.00000392, 0.00004021, -0.00001541, -0.00008936, -0.00000690, 0.00014438, 0.00009177, -0.00013120, -0.00017138, 0.00005766, 0.00020454, 0.00005727, -0.00012987, -0.00009155, 0.00004936, 0.00006503, -0.00001457, -0.00004364, -0.00001792, -0.00000570, -0.00000406, 0.00000791, 0.00000823, -0.00000764, -0.00001164, -0.00000451, -0.00000553, -0.00000362, 0.00001075, 0.00001222, -0.00000741, -0.00001547, -0.00000134, 0.00000570, -0.00000318, -0.00000631, },
		{ -0.00000291, -0.00000338, -0.00000157, -0.00000184, 0.00000583, 0.00001234, -0.00000255, -0.00002032, -0.00001122, 0.00000247, -0.00001018, -0.00001894, 0.00000797, 0.00003047, 0.00001019, -0.00001401, -0.00000480, 0.00000464, -0.00001060, -0.00001528, 0.00000545, 0.00001328, -0.00000301, -0.00001066, -0.00000153, 0.00000832, 0.00001900, 0.00001790, -0.00001832, -0.00004634, -0.00000399, 0.00004235, -0.00001536, -0.00008936, -0.00000714, 0.00013849, 0.00008616, -0.00012169, -0.00015357, 0.00005727, 0.00018677, 0.00005733, -0.00010146, -0.00007292, 0.00003173, 0.00003838, -0.00001684, -0.00003212, -0.00001867, -0.00001539, -0.00000452, 0.00001617, 0.00001274, -0.00000964, -0.00001548, -0.00000766, -0.00000631, -0.00000176, 0.00001096, 0.00000899, -0.00000843, -0.00001193, 0.00000053, 0.00000371, },
		{ 0.00000251, 0.00000110, -0.00000265, -0.00000001, -0.00000334, -0.00000794, 0.00000507, 0.00001706, -0.00000122, -0.00002307, -0.00001122, 0.00000501, -0.00001066, -0.00002159, 0.00000780, 0.00003231, 0.00001120, -0.00001370, -0.00000498, 0.00000322, -0.00001136, -0.00001337, 0.00000726, 0.00001199, -0.00000514, -0.00001102, -0.00000124, 0.00000901, 0.00002147, 0.00001993, -0.00001966, -0.00004848, -0.00000401, 0.00004235, -0.00001541, -0.00008677, -0.00000786, 0.00012698, 0.00007701, -0.00010817, -0.00012987, 0.00005733, 0.00016652, 0.00005830, -0.00006977, -0.00005360, 0.00001228, 0.00001068, -0.00001916, -0.00002101, -0.00001949, -0.00002440, -0.00000522, 0.00002303, 0.00001693, -0.00001038, -0.00001854, -0.00001116, -0.00000713, 0.00000035, 0.00001041, 0.00000472, -0.00000910, -0.00000727, },
		{ -0.00000938, -0.00000150, 0.00000458, -0.00000211, -0.00000239, 0.00000377, -0.00000514, -0.00001431, 0.00000422, 0.00002179, 0.00000014, -0.00002570, -0.00001122, 0.00000740, -0.00001104, -0.00002393, 0.00000758, 0.00003375, 0.00001209, -0.00001323, -0.00000514, 0.00000177, -0.00001199, -0.00001149, 0.00000884, 0.00001078, -0.00000692, -0.00001127, -0.00000103, 0.00000947, 0.00002299, 0.00002098, -0.00002011, -0.00004848, -0.00000399, 0.00004021, -0.00001558, -0.00008173, -0.00000902, 0.00011037, 0.00006461, -0.00009155, -0.00010146, 0.00005830, 0.00014559, 0.00006069, -0.00003642, -0.00003501, -0.00000838, -0.00001675, -0.00002142, -0.00001095, -0.00002038, -0.00003225, -0.00000619, 0.00002803, 0.00002064, -0.00000962, -0.00002067, -0.00001500, -0.00000797, 0.00000269, 0.00000906, -0.00000061, },
		{ 0.00000687, -0.00000698, -0.00000927, 0.00000533, 0.00000672, -0.00000588, -0.00000214, 0.00000789, -0.00000692, -0.00002085, 0.00000331, 0.00002647, 0.00000150, -0.00002818, -0.00001122, 0.00000961, -0.00001134, -0.00002596, 0.00000732, 0.00003485, 0.00001286, -0.00001266, -0.00000527, 0.00000039, -0.00001249, -0.00000978, 0.00001013, 0.00000977, -0.00000827, -0.00001143, -0.00000090, 0.00000969, 0.00002351, 0.00002098, -0.00001966, -0.00004634, -0.00000392, 0.00003604, -0.00001584, -0.00007448, -0.00001057, 0.00008940, 0.00004936, -0.00007292, -0.00006977, 0.00006069, 0.00012599, 0.00006509, -0.00000315, -0.00001866, -0.00002964, -0.00004260, -0.00002352, -0.00000255, -0.00002133, -0.00003850, -0.00000747, 0.00003073, 0.00002374, -0.00000717, -0.00002174, -0.00001917, -0.00000883, 0.00000527, },
		{ -0.00000970, 0.00000805, 0.00000386, -0.00001432, -0.00000878, 0.00001310, 0.00000891, -0.00001012, -0.00000189, 0.00001227, -0.00000868, -0.00002746, 0.00000237, 0.00003103, 0.00000285, -0.00003050, -0.00001122, 0.00001163, -0.00001156, -0.00002770, 0.00000704, 0.00003566, 0.00001351, -0.00001207, -0.00000538, -0.00000082, -0.00001288, -0.00000837, 0.00001110, 0.00000904, -0.00000910, -0.00001150, -0.00000086, 0.00000969, 0.00002299, 0.00001993, -0.00001832, -0.00004217, -0.00000384, 0.00003003, -0.00001622, -0.00006539, -0.00001245, 0.00006503, 0.00003173, -0.00005360, -0.00003642, 0.00006509, 0.00010991, 0.00007209, 0.00002823, -0.00000606, -0.00005084, -0.00006556, -0.00002535, 0.00000360, -0.00002233, -0.00004275, -0.00000909, 0.00003080, 0.00002613, -0.00000291, -0.00002163, -0.00002363, },
		{ -0.00002032, -0.00002835, -0.00001056, 0.00001101, 0.00000007, -0.00002254, -0.00000792, 0.00002170, 0.00001111, -0.00001475, -0.00000165, 0.00001682, -0.00001039, -0.00003402, 0.00000140, 0.00003539, 0.00000414, -0.00003263, -0.00001122, 0.00001343, -0.00001171, -0.00002915, 0.00000677, 0.00003624, 0.00001404, -0.00001151, -0.00000547, -0.00000181, -0.00001315, -0.00000738, 0.00001169, 0.00000865, -0.00000938, -0.00001150, -0.00000090, 0.00000947, 0.00002147, 0.00001790, -0.00001614, -0.00003618, -0.00000375, 0.00002247, -0.00001669, -0.00005493, -0.00001457, 0.00003838, 0.00001228, -0.00003501, -0.00000315, 0.00007209, 0.00009955, 0.00008227, 0.00005592, 0.00000135, -0.00007139, -0.00008443, -0.00002684, 0.00000701, -0.00002337, -0.00004467, -0.00001105, 0.00002799, 0.00002770, 0.00000322, },
		{ 0.00002842, 0.00001122, -0.00001778, -0.00003328, -0.00001142, 0.00001412, -0.00000442, -0.00003150, -0.00000673, 0.00003095, 0.00001331, -0.00001966, -0.00000143, 0.00002145, -0.00001202, -0.00004040, 0.00000044, 0.00003948, 0.00000536, -0.00003454, -0.00001122, 0.00001501, -0.00001182, -0.00003034, 0.00000653, 0.00003664, 0.00001445, -0.00001104, -0.00000553, -0.00000250, -0.00001331, -0.00000686, 0.00001189, 0.00000865, -0.00000910, -0.00001143, -0.00000103, 0.00000901, 0.00001900, 0.00001502, -0.00001320, -0.00002867, -0.00000369, 0.00001370, -0.00001726, -0.00004364, -0.00001684, 0.00001068, -0.00000838, -0.00001866, 0.00002823, 0.00008227, 0.00009707, 0.00009616, 0.00007823, 0.00000226, -0.00009068, -0.00009815, -0.00002789, 0.00000726, -0.00002445, -0.00004402, -0.00001336, 0.00002217, },
		{ -0.00001602, 0.00001329, 0.00002826, 0.00002101, -0.00001407, -0.00003838, -0.00001226, 0.00001734, -0.00000953, -0.00004103, -0.00000526, 0.00004065, 0.00001546, -0.00002473, -0.00000122, 0.00002604, -0.00001356, -0.00004647, -0.00000049, 0.00004324, 0.00000649, -0.00003622, -0.00001122, 0.00001635, -0.00001189, -0.00003127, 0.00000631, 0.00003689, 0.00001474, -0.00001071, -0.00000557, -0.00000286, -0.00001336, -0.00000686, 0.00001169, 0.00000904, -0.00000827, -0.00001127, -0.00000124, 0.00000832, 0.00001571, 0.00001145, -0.00000962, -0.00002004, -0.00000369, 0.00000416, -0.00001792, -0.00003212, -0.00001916, -0.00001675, -0.00002964, -0.00000606, 0.00005592, 0.00009616, 0.00010446, 0.00011423, 0.00009364, -0.00000448, -0.00010821, -0.00010585, -0.00002845, 0.00000406, -0.00002554, -0.00004065, },
		{ -0.00002665, -0.00003452, -0.00001899, 0.00000144, 0.00002724, 0.00003245, -0.00000926, -0.00004357, -0.00001308, 0.00002062, -0.00001512, -0.00005093, -0.00000357, 0.00005056, 0.00001753, -0.00002982, -0.00000102, 0.00003048, -0.00001498, -0.00005211, -0.00000135, 0.00004660, 0.00000748, -0.00003765, -0.00001122, 0.00001743, -0.00001194, -0.00003195, 0.00000615, 0.00003704, 0.00001492, -0.00001053, -0.00000559, -0.00000286, -0.00001331, -0.00000738, 0.00001110, 0.00000977, -0.00000692, -0.00001102, -0.00000153, 0.00000738, 0.00001172, 0.00000743, -0.00000553, -0.00001072, -0.00000380, -0.00000570, -0.00001867, -0.00002101, -0.00002142, -0.00004260, -0.00005084, 0.00000135, 0.00007823, 0.00011423, 0.00012349, 0.00013684, 0.00010083, -0.00001975, -0.00012354, -0.00010684, -0.00002846, -0.00000277, },
		{ -0.00002792, -0.00001327, -0.00002776, -0.00002568, -0.00002226, -0.00001316, 0.00002540, 0.00004531, -0.00000350, -0.00004880, -0.00001386, 0.00002391, -0.00002105, -0.00006099, -0.00000173, 0.00006046, 0.00001950, -0.00003481, -0.00000085, 0.00003466, -0.00001626, -0.00005720, -0.00000213, 0.00004950, 0.00000833, -0.00003881, -0.00001122, 0.00001825, -0.00001196, -0.00003240, 0.00000605, 0.00003711, 0.00001498, -0.00001053, -0.00000557, -0.00000250, -0.00001315, -0.00000837, 0.00001013, 0.00001078, -0.00000514, -0.00001066, -0.00000190, 0.00000616, 0.00000723, 0.00000321, -0.00000108, -0.00000123, -0.00000406, -0.00001539, -0.00001949, -0.00001095, -0.00002352, -0.00006556, -0.00007139, 0.00000226, 0.00009364, 0.00013684, 0.00015556, 0.00016424, 0.00009880, -0.00004415, -0.00013630, -0.00010072, },
		{ -0.00014625, -0.00008733, -0.00002681, -0.00002733, -0.00002885, -0.00001430, -0.00002577, -0.00003023, 0.00002283, 0.00005933, 0.00000305, -0.00005397, -0.00001461, 0.00002714, -0.00002715, -0.00007097, 0.00000019, 0.00007008, 0.00002133, -0.00003954, -0.00000069, 0.00003848, -0.00001738, -0.00006163, -0.00000280, 0.00005189, 0.00000901, -0.00003969, -0.00001122, 0.00001880, -0.00001197, -0.00003263, 0.00000602, 0.00003711, 0.00001492, -0.00001071, -0.00000553, -0.00000181, -0.00001288, -0.00000978, 0.00000884, 0.00001199, -0.00000301, -0.00001018, -0.00000235, 0.00000467, 0.00000242, -0.00000091, 0.00000356, 0.00000791, -0.00000452, -0.00002440, -0.00002038, -0.00000255, -0.00002535, -0.00008443, -0.00009068, -0.00000448, 0.00010083, 0.00016424, 0.00020169, 0.00019651, 0.00008685, -0.00007796, },
		{ 0.00006468, -0.00012113, -0.00015325, -0.00006681, -0.00002515, -0.00004471, -0.00002993, -0.00000064, -0.00002946, -0.00004933, 0.00001962, 0.00007418, 0.00001020, -0.00005902, -0.00001530, 0.00003028, -0.00003324, -0.00008062, 0.00000212, 0.00007917, 0.00002298, -0.00004389, -0.00000056, 0.00004183, -0.00001832, -0.00006530, -0.00000335, 0.00005372, 0.00000951, -0.00004028, -0.00001122, 0.00001908, -0.00001198, -0.00003263, 0.00000605, 0.00003704, 0.00001474, -0.00001104, -0.00000547, -0.00000082, -0.00001249, -0.00001149, 0.00000726, 0.00001328, -0.00000065, -0.00000956, -0.00000286, 0.00000286, -0.00000248, -0.00000463, 0.00000823, 0.00001617, -0.00000522, -0.00003225, -0.00002133, 0.00000360, -0.00002684, -0.00009815, -0.00010821, -0.00001975, 0.00009880, 0.00019651, 0.00026239, 0.00023360, },
		{ 0.00033770, 0.00027527, 0.00003233, -0.00017323, -0.00015728, -0.00003956, -0.00002299, -0.00006505, -0.00003097, 0.00001494, -0.00003328, -0.00006998, 0.00001591, 0.00008950, 0.00001770, -0.00006387, -0.00001595, 0.00003324, -0.00003916, -0.00008972, 0.00000399, 0.00008749, 0.00002444, -0.00004773, -0.00000044, 0.00004463, -0.00001906, -0.00006812, -0.00000375, 0.00005496, 0.00000981, -0.00004058, -0.00001122, 0.00001908, -0.00001197, -0.00003240, 0.00000615, 0.00003689, 0.00001445, -0.00001151, -0.00000538, 0.00000039, -0.00001199, -0.00001337, 0.00000545, 0.00001451, 0.00000181, -0.00000877, -0.00000345, 0.00000074, -0.00000724, -0.00000764, 0.00001274, 0.00002303, -0.00000619, -0.00003850, -0.00002233, 0.00000701, -0.00002789, -0.00010585, -0.00012354, -0.00004415, 0.00008685, 0.00023360, },
		{ 0.00006468, 0.00027527, 0.00042707, 0.00032113, -0.00000971, -0.00023350, -0.00015843, -0.00000626, -0.00002037, -0.00008785, -0.00003197, 0.00003201, -0.00003714, -0.00009161, 0.00001185, 0.00010487, 0.00002532, -0.00006842, -0.00001653, 0.00003599, -0.00004472, -0.00009802, 0.00000570, 0.00009482, 0.00002567, -0.00005094, -0.00000036, 0.00004679, -0.00001961, -0.00007005, -0.00000399, 0.00005559, 0.00000991, -0.00004058, -0.00001122, 0.00001880, -0.00001196, -0.00003195, 0.00000631, 0.00003664, 0.00001404, -0.00001207, -0.00000527, 0.00000177, -0.00001136, -0.00001528, 0.00000348, 0.00001552, 0.00000423, -0.00000781, -0.00000409, -0.00000171, -0.00001164, -0.00000964, 0.00001693, 0.00002803, -0.00000747, -0.00004275, -0.00002337, 0.00000726, -0.00002845, -0.00010684, -0.00013630, -0.00007796, },
		{ -0.00014625, -0.00012113, 0.00003233, 0.00032113, 0.00052946, 0.00037060, -0.00006063, -0.00030085, -0.00015688, 0.00003220, -0.00001739, -0.00011252, -0.00003292, 0.00005006, -0.00004097, -0.00011360, 0.00000761, 0.00011991, 0.00003281, -0.00007261, -0.00001705, 0.00003846, -0.00004975, -0.00010531, 0.00000721, 0.00010094, 0.00002665, -0.00005344, -0.00000029, 0.00004827, -0.00001994, -0.00007102, -0.00000407, 0.00005559, 0.00000981, -0.00004028, -0.00001122, 0.00001825, -0.00001194, -0.00003127, 0.00000653, 0.00003624, 0.00001351, -0.00001266, -0.00000514, 0.00000322, -0.00001060, -0.00001706, 0.00000144, 0.00001618, 0.00000647, -0.00000664, -0.00000478, -0.00000451, -0.00001548, -0.00001038, 0.00002064, 0.00003073, -0.00000909, -0.00004467, -0.00002445, 0.00000406, -0.00002846, -0.00010072, },
		{ -0.00002792, -0.00008733, -0.00015325, -0.00017323, -0.00000971, 0.00037060, 0.00064326, 0.00042296, -0.00011921, -0.00037388, -0.00015293, 0.00007468, -0.00001413, -0.00013837, -0.00003381, 0.00006855, -0.00004469, -0.00013529, 0.00000334, 0.00013418, 0.00003991, -0.00007636, -0.00001749, 0.00004060, -0.00005411, -0.00011140, 0.00000845, 0.00010570, 0.00002736, -0.00005515, -0.00000025, 0.00004902, -0.00002005, -0.00007102, -0.00000399, 0.00005496, 0.00000951, -0.00003969, -0.00001122, 0.00001743, -0.00001189, -0.00003034, 0.00000677, 0.00003566, 0.00001286, -0.00001323, -0.00000498, 0.00000464, -0.00000970, -0.00001852, -0.00000062, 0.00001631, 0.00000840, -0.00000525, -0.00000553, -0.00000766, -0.00001854, -0.00000962, 0.00002374, 0.00003080, -0.00001105, -0.00004402, -0.00002554, -0.00000277, },
		{ -0.00002665, -0.00001327, -0.00002681, -0.00006681, -0.00015728, -0.00023350, -0.00006063, 0.00042296, 0.00076642, 0.00047735, -0.00018396, -0.00045092, -0.00014694, 0.00011990, -0.00001070, -0.00016466, -0.00003464, 0.00008691, -0.00004822, -0.00015605, -0.00000078, 0.00014728, 0.00004639, -0.00007960, -0.00001786, 0.00004238, -0.00005764, -0.00011611, 0.00000937, 0.00010894, 0.00002779, -0.00005602, -0.00000024, 0.00004902, -0.00001994, -0.00007005, -0.00000375, 0.00005372, 0.00000901, -0.00003881, -0.00001122, 0.00001635, -0.00001182, -0.00002915, 0.00000704, 0.00003485, 0.00001209, -0.00001370, -0.00000480, 0.00000592, -0.00000867, -0.00001949, -0.00000260, 0.00001578, 0.00000986, -0.00000362, -0.00000631, -0.00001116, -0.00002067, -0.00000717, 0.00002613, 0.00002799, -0.00001336, -0.00004065, },
		{ -0.00001602, -0.00003452, -0.00002776, -0.00002733, -0.00002515, -0.00003956, -0.00015843, -0.00030085, -0.00011921, 0.00047735, 0.00089644, 0.00053278, -0.00025310, -0.00053012, -0.00013937, 0.00016644, -0.00000722, -0.00019061, -0.00003538, 0.00010455, -0.00005147, -0.00017523, -0.00000458, 0.00015885, 0.00005203, -0.00008227, -0.00001815, 0.00004374, -0.00006025, -0.00011932, 0.00000994, 0.00011059, 0.00002794, -0.00005602, -0.00000025, 0.00004827, -0.00001961, -0.00006812, -0.00000335, 0.00005189, 0.00000833, -0.00003765, -0.00001122, 0.00001501, -0.00001171, -0.00002770, 0.00000732, 0.00003375, 0.00001120, -0.00001401, -0.00000461, 0.00000698, -0.00000749, -0.00001981, -0.00000443, 0.00001446, 0.00001075, -0.00000176, -0.00000713, -0.00001500, -0.00002174, -0.00000291, 0.00002770, 0.00002217, },
		{ 0.00002842, 0.00001329, -0.00001899, -0.00002568, -0.00002885, -0.00004471, -0.00002299, -0.00000626, -0.00015688, -0.00037388, -0.00018396, 0.00053278, 0.00103046, 0.00058819, -0.00032463, -0.00060946, -0.00013069, 0.00021280, -0.00000378, -0.00021545, -0.00003604, 0.00012091, -0.00005436, -0.00019223, -0.00000792, 0.00016854, 0.00005662, -0.00008431, -0.00001837, 0.00004467, -0.00006185, -0.00012095, 0.00001013, 0.00011059, 0.00002779, -0.00005515, -0.00000029, 0.00004679, -0.00001906, -0.00006530, -0.00000280, 0.00004950, 0.00000748, -0.00003622, -0.00001122, 0.00001343, -0.00001156, -0.00002596, 0.00000758, 0.00003231, 0.00001019, -0.00001407, -0.00000440, 0.00000771, -0.00000616, -0.00001933, -0.00000606, 0.00001222, 0.00001096, 0.00000035, -0.00000797, -0.00001917, -0.00002163, 0.00000322, },
		{ -0.00002032, 0.00001122, 0.00002826, 0.00000144, -0.00002226, -0.00001430, -0.00002993, -0.00006505, -0.00002037, 0.00003220, -0.00015293, -0.00045092, -0.00025310, 0.00058819, 0.00116538, 0.00064244, -0.00039645, -0.00068685, -0.00012143, 0.00025748, -0.00000053, -0.00023840, -0.00003661, 0.00013545, -0.00005684, -0.00020653, -0.00001067, 0.00017607, 0.00006002, -0.00008570, -0.00001849, 0.00004513, -0.00006238, -0.00012095, 0.00000994, 0.00010894, 0.00002736, -0.00005344, -0.00000036, 0.00004463, -0.00001832, -0.00006163, -0.00000213, 0.00004660, 0.00000649, -0.00003454, -0.00001122, 0.00001163, -0.00001134, -0.00002393, 0.00000780, 0.00003047, 0.00000907, -0.00001383, -0.00000417, 0.00000801, -0.00000469, -0.00001791, -0.00000741, 0.00000899, 0.00001041, 0.00000269, -0.00000883, -0.00002363, },
		{ -0.00000970, -0.00002835, -0.00001778, 0.00002101, 0.00002724, -0.00001316, -0.00002577, -0.00000064, -0.00003097, -0.00008785, -0.00001739, 0.00007468, -0.00014694, -0.00053012, -0.00032463, 0.00064244, 0.00129790, 0.00069438, -0.00046637, -0.00076021, -0.00011209, 0.00029900, 0.00000245, -0.00025875, -0.00003708, 0.00014772, -0.00005882, -0.00021768, -0.00001271, 0.00018122, 0.00006211, -0.00008640, -0.00001854, 0.00004513, -0.00006185, -0.00011932, 0.00000937, 0.00010570, 0.00002665, -0.00005094, -0.00000044, 0.00004183, -0.00001738, -0.00005720, -0.00000135, 0.00004324, 0.00000536, -0.00003263, -0.00001122, 0.00000961, -0.00001104, -0.00002159, 0.00000797, 0.00002819, 0.00000783, -0.00001322, -0.00000394, 0.00000782, -0.00000308, -0.00001547, -0.00000843, 0.00000472, 0.00000906, 0.00000527, },
		{ 0.00000687, 0.00000805, -0.00001056, -0.00003328, -0.00001407, 0.00003245, 0.00002540, -0.00003023, -0.00002946, 0.00001494, -0.00003197, -0.00011252, -0.00001413, 0.00011990, -0.00013937, -0.00060946, -0.00039645, 0.00069438, 0.00142471, 0.00074285, -0.00053224, -0.00082750, -0.00010319, 0.00033599, 0.00000503, -0.00027588, -0.00003746, 0.00015728, -0.00006028, -0.00022531, -0.00001397, 0.00018383, 0.00006282, -0.00008640, -0.00001849, 0.00004467, -0.00006025, -0.00011611, 0.00000845, 0.00010094, 0.00002567, -0.00004773, -0.00000056, 0.00003848, -0.00001626, -0.00005211, -0.00000049, 0.00003948, 0.00000414, -0.00003050, -0.00001122, 0.00000740, -0.00001066, -0.00001894, 0.00000807, 0.00002543, 0.00000648, -0.00001220, -0.00000369, 0.00000706, -0.00000134, -0.00001193, -0.00000910, -0.00000061, },
		{ -0.00000938, -0.00000698, 0.00000386, 0.00001101, -0.00001142, -0.00003838, -0.00000926, 0.00004531, 0.00002283, -0.00004933, -0.00003328, 0.00003201, -0.00003292, -0.00013837, -0.00001070, 0.00016644, -0.00013069, -0.00068685, -0.00046637, 0.00074285, 0.00154251, 0.00078676, -0.00059200, -0.00088682, -0.00009520, 0.00036720, 0.00000714, -0.00028922, -0.00003773, 0.00016385, -0.00006117, -0.00022919, -0.00001439, 0.00018383, 0.00006211, -0.00008570, -0.00001837, 0.00004374, -0.00005764, -0.00011140, 0.00000721, 0.00009482, 0.00002444, -0.00004389, -0.00000069, 0.00003466, -0.00001498, -0.00004647, 0.00000044, 0.00003539, 0.00000285, -0.00002818, -0.00001122, 0.00000501, -0.00001018, -0.00001597, 0.00000809, 0.00002216, 0.00000503, -0.00001072, -0.00000343, 0.00000570, 0.00000053, -0.00000727, },
		{ 0.00000251, -0.00000150, -0.00000927, -0.00001432, 0.00000007, 0.00001412, -0.00001226, -0.00004357, -0.00000350, 0.00005933, 0.00001962, -0.00006998, -0.00003714, 0.00005006, -0.00003381, -0.00016466, -0.00000722, 0.00021280, -0.00012143, -0.00076021, -0.00053224, 0.00078676, 0.00164821, 0.00082509, -0.00064375, -0.00093648, -0.00008852, 0.00039159, 0.00000871, -0.00029836, -0.00003789, 0.00016718, -0.00006147, -0.00022919, -0.00001397, 0.00018122, 0.00006002, -0.00008431, -0.00001815, 0.00004238, -0.00005411, -0.00010531, 0.00000570, 0.00008749, 0.00002298, -0.00003954, -0.00000085, 0.00003048, -0.00001356, -0.00004040, 0.00000140, 0.00003103, 0.00000150, -0.00002570, -0.00001122, 0.00000247, -0.00000960, -0.00001270, 0.00000800, 0.00001837, 0.00000349, -0.00000875, -0.00000318, 0.00000371, },
		{ -0.00000291, 0.00000110, 0.00000458, 0.00000533, -0.00000878, -0.00002254, -0.00000442, 0.00001734, -0.00001308, -0.00004880, 0.00000305, 0.00007418, 0.00001591, -0.00009161, -0.00004097, 0.00006855, -0.00003464, -0.00019061, -0.00000378, 0.00025748, -0.00011209, -0.00082750, -0.00059200, 0.00082509, 0.00173895, 0.00085692, -0.00068583, -0.00097505, -0.00008349, 0.00040832, 0.00000967, -0.00030300, -0.00003794, 0.00016718, -0.00006117, -0.00022531, -0.00001271, 0.00017607, 0.00005662, -0.00008227, -0.00001786, 0.00004060, -0.00004975, -0.00009802, 0.00000399, 0.00007917, 0.00002133, -0.00003481, -0.00000102, 0.00002604, -0.00001202, -0.00003402, 0.00000237, 0.00002647, 0.00000014, -0.00002307, -0.00001122, -0.00000020, -0.00000891, -0.00000912, 0.00000780, 0.00001404, 0.00000187, -0.00000631, },
		{ 0.00000018, -0.00000338, -0.00000265, -0.00000211, 0.00000672, 0.00001310, -0.00000792, -0.00003150, -0.00000953, 0.00002062, -0.00001386, -0.00005397, 0.00001020, 0.00008950, 0.00001185, -0.00011360, -0.00004469, 0.00008691, -0.00003538, -0.00021545, -0.00000053, 0.00029900, -0.00010319, -0.00088682, -0.00064375, 0.00085692, 0.00181226, 0.00088150, -0.00071690, -0.00100141, -0.00008037, 0.00041683, 0.00000999, -0.00030300, -0.00003789, 0.00016385, -0.00006028, -0.00021768, -0.00001067, 0.00016854, 0.00005203, -0.00007960, -0.00001749, 0.00003846, -0.00004472, -0.00008972, 0.00000212, 0.00007008, 0.00001950, -0.00002982, -0.00000122, 0.00002145, -0.00001039, -0.00002746, 0.00000331, 0.00002179, -0.00000122, -0.00002032, -0.00001122, -0.00000297, -0.00000810, -0.00000527, 0.00000748, 0.00000921, },
		{ 0.00000704, 0.00000390, -0.00000157, -0.00000001, -0.00000239, -0.00000588, 0.00000891, 0.00002170, -0.00000673, -0.00004103, -0.00001512, 0.00002391, -0.00001461, -0.00005902, 0.00001770, 0.00010487, 0.00000761, -0.00013529, -0.00004822, 0.00010455, -0.00003604, -0.00023840, 0.00000245, 0.00033599, -0.00009520, -0.00093648, -0.00068583, 0.00088150, 0.00186614, 0.00089823, -0.00073596, -0.00101479, -0.00007931, 0.00041683, 0.00000967, -0.00029836, -0.00003773, 0.00015728, -0.00005882, -0.00020653, -0.00000792, 0.00015885, 0.00004639, -0.00007636, -0.00001705, 0.00003599, -0.00003916, -0.00008062, 0.00000019, 0.00006046, 0.00001753, -0.00002473, -0.00000143, 0.00001682, -0.00000868, -0.00002085, 0.00000422, 0.00001706, -0.00000255, -0.00001749, -0.00001122, -0.00000583, -0.00000719, -0.00000116, },
		{ -0.00000616, 0.00000318, 0.00000649, -0.00000184, -0.00000334, 0.00000377, -0.00000214, -0.00001012, 0.00001111, 0.00003095, -0.00000526, -0.00005093, -0.00002105, 0.00002714, -0.00001530, -0.00006387, 0.00002532, 0.00011991, 0.00000334, -0.00015605, -0.00005147, 0.00012091, -0.00003661, -0.00025875, 0.00000503, 0.00036720, -0.00008852, -0.00097505, -0.00071690, 0.00089823, 0.00189907, 0.00090670, -0.00074238, -0.00101479, -0.00008037, 0.00040832, 0.00000871, -0.00028922, -0.00003746, 0.00014772, -0.00005684, -0.00019223, -0.00000458, 0.00014728, 0.00003991, -0.00007261, -0.00001653, 0.00003324, -0.00003324, -0.00007097, -0.00000173, 0.00005056, 0.00001546, -0.00001966, -0.00000165, 0.00001227, -0.00000692, -0.00001431, 0.00000507, 0.00001234, -0.00000383, -0.00001459, -0.00001122, -0.00000873, },
		{ -0.00001122, -0.00001166, -0.00000504, 0.00000769, 0.00000583, -0.00000794, -0.00000514, 0.00000789, -0.00000189, -0.00001475, 0.00001331, 0.00004065, -0.00000357, -0.00006099, -0.00002715, 0.00003028, -0.00001595, -0.00006842, 0.00003281, 0.00013418, -0.00000078, -0.00017523, -0.00005436, 0.00013545, -0.00003708, -0.00027588, 0.00000714, 0.00039159, -0.00008349, -0.00100141, -0.00073596, 0.00090670, 0.00191015, 0.00090670, -0.00073596, -0.00100141, -0.00008349, 0.00039159, 0.00000714, -0.00027588, -0.00003708, 0.00013545, -0.00005436, -0.00017523, -0.00000078, 0.00013418, 0.00003281, -0.00006842, -0.00001595, 0.00003028, -0.00002715, -0.00006099, -0.00000357, 0.00004065, 0.00001331, -0.00001475, -0.00000189, 0.00000789, -0.00000514, -0.00000794, 0.00000583, 0.00000769, -0.00000504, -0.00001166, },
		{ -0.00000616, -0.00000873, -0.00001122, -0.00001459, -0.00000383, 0.00001234, 0.00000507, -0.00001431, -0.00000692, 0.00001227, -0.00000165, -0.00001966, 0.00001546, 0.00005056, -0.00000173, -0.00007097, -0.00003324, 0.00003324, -0.00001653, -0.00007261, 0.00003991, 0.00014728, -0.00000458, -0.00019223, -0.00005684, 0.00014772, -0.00003746, -0.00028922, 0.00000871, 0.00040832, -0.00008037, -0.00101479, -0.00074238, 0.00090670, 0.00189907, 0.00089823, -0.00071690, -0.00097505, -0.00008852, 0.00036720, 0.00000503, -0.00025875, -0.00003661, 0.00012091, -0.00005147, -0.00015605, 0.00000334, 0.00011991, 0.00002532, -0.00006387, -0.00001530, 0.00002714, -0.00002105, -0.00005093, -0.00000526, 0.00003095, 0.00001111, -0.00001012, -0.00000214, 0.00000377, -0.00000334, -0.00000184, 0.00000649, 0.00000318, },
		{ 0.00000704, -0.00000116, -0.00000719, -0.00000583, -0.00001122, -0.00001749, -0.00000255, 0.00001706, 0.00000422, -0.00002085, -0.00000868, 0.00001682, -0.00000143, -0.00002473, 0.00001753, 0.00006046, 0.00000019, -0.00008062, -0.00003916, 0.00003599, -0.00001705, -0.00007636, 0.00004639, 0.00015885, -0.00000792, -0.00020653, -0.00005882, 0.00015728, -0.00003773, -0.00029836, 0.00000967, 0.00041683, -0.00007931, -0.00101479, -0.00073596, 0.00089823, 0.00186614, 0.00088150, -0.00068583, -0.00093648, -0.00009520, 0.00033599, 0.00000245, -0.00023840, -0.00003604, 0.00010455, -0.00004822, -0.00013529, 0.00000761, 0.00010487, 0.00001770, -0.00005902, -0.00001461, 0.00002391, -0.00001512, -0.00004103, -0.00000673, 0.00002170, 0.00000891, -0.00000588, -0.00000239, -0.00000001, -0.00000157, 0.00000390, },
		{ 0.00000018, 0.00000921, 0.00000748, -0.00000527, -0.00000810, -0.00000297, -0.00001122, -0.00002032, -0.00000122, 0.00002179, 0.00000331, -0.00002746, -0.00001039, 0.00002145, -0.00000122, -0.00002982, 0.00001950, 0.00007008, 0.00000212, -0.00008972, -0.00004472, 0.00003846, -0.00001749, -0.00007960, 0.00005203, 0.00016854, -0.00001067, -0.00021768, -0.00006028, 0.00016385, -0.00003789, -0.00030300, 0.00000999, 0.00041683, -0.00008037, -0.00100141, -0.00071690, 0.00088150, 0.00181226, 0.00085692, -0.00064375, -0.00088682, -0.00010319, 0.00029900, -0.00000053, -0.00021545, -0.00003538, 0.00008691, -0.00004469, -0.00011360, 0.00001185, 0.00008950, 0.00001020, -0.00005397, -0.00001386, 0.00002062, -0.00000953, -0.00003150, -0.00000792, 0.00001310, 0.00000672, -0.00000211, -0.00000265, -0.00000338, },
		{ -0.00000291, -0.00000631, 0.00000187, 0.00001404, 0.00000780, -0.00000912, -0.00000891, -0.00000020, -0.00001122, -0.00002307, 0.00000014, 0.00002647, 0.00000237, -0.00003402, -0.00001202, 0.00002604, -0.00000102, -0.00003481, 0.00002133, 0.00007917, 0.00000399, -0.00009802, -0.00004975, 0.00004060, -0.00001786, -0.00008227, 0.00005662, 0.00017607, -0.00001271, -0.00022531, -0.00006117, 0.00016718, -0.00003794, -0.00030300, 0.00000967, 0.00040832, -0.00008349, -0.00097505, -0.00068583, 0.00085692, 0.00173895, 0.00082509, -0.00059200, -0.00082750, -0.00011209, 0.00025748, -0.00000378, -0.00019061, -0.00003464, 0.00006855, -0.00004097, -0.00009161, 0.00001591, 0.00007418, 0.00000305, -0.00004880, -0.00001308, 0.00001734, -0.00000442, -0.00002254, -0.00000878, 0.00000533, 0.00000458, 0.00000110, },
		{ 0.00000251, 0.00000371, -0.00000318, -0.00000875, 0.00000349, 0.00001837, 0.00000800, -0.00001270, -0.00000960, 0.00000247, -0.00001122, -0.00002570, 0.00000150, 0.00003103, 0.00000140, -0.00004040, -0.00001356, 0.00003048, -0.00000085, -0.00003954, 0.00002298, 0.00008749, 0.00000570, -0.00010531, -0.00005411, 0.00004238, -0.00001815, -0.00008431, 0.00006002, 0.00018122, -0.00001397, -0.00022919, -0.00006147, 0.00016718, -0.00003789, -0.00029836, 0.00000871, 0.00039159, -0.00008852, -0.00093648, -0.00064375, 0.00082509, 0.00164821, 0.00078676, -0.00053224, -0.00076021, -0.00012143, 0.00021280, -0.00000722, -0.00016466, -0.00003381, 0.00005006, -0.00003714, -0.00006998, 0.00001962, 0.00005933, -0.00000350, -0.00004357, -0.00001226, 0.00001412, 0.00000007, -0.00001432, -0.00000927, -0.00000150, },
		{ -0.00000938, -0.00000727, 0.00000053, 0.00000570, -0.00000343, -0.00001072, 0.00000503, 0.00002216, 0.00000809, -0.00001597, -0.00001018, 0.00000501, -0.00001122, -0.00002818, 0.00000285, 0.00003539, 0.00000044, -0.00004647, -0.00001498, 0.00003466, -0.00000069, -0.00004389, 0.00002444, 0.00009482, 0.00000721, -0.00011140, -0.00005764, 0.00004374, -0.00001837, -0.00008570, 0.00006211, 0.00018383, -0.00001439, -0.00022919, -0.00006117, 0.00016385, -0.00003773, -0.00028922, 0.00000714, 0.00036720, -0.00009520, -0.00088682, -0.00059200, 0.00078676, 0.00154251, 0.00074285, -0.00046637, -0.00068685, -0.00013069, 0.00016644, -0.00001070, -0.00013837, -0.00003292, 0.00003201, -0.00003328, -0.00004933, 0.00002283, 0.00004531, -0.00000926, -0.00003838, -0.00001142, 0.00001101, 0.00000386, -0.00000698, },
		{ 0.00000687, -0.00000061, -0.00000910, -0.00001193, -0.00000134, 0.00000706, -0.00000369, -0.00001220, 0.00000648, 0.00002543, 0.00000807, -0.00001894, -0.00001066, 0.00000740, -0.00001122, -0.00003050, 0.00000414, 0.00003948, -0.00000049, -0.00005211, -0.00001626, 0.00003848, -0.00000056, -0.00004773, 0.00002567, 0.00010094, 0.00000845, -0.00011611, -0.00006025, 0.00004467, -0.00001849, -0.00008640, 0.00006282, 0.00018383, -0.00001397, -0.00022531, -0.00006028, 0.00015728, -0.00003746, -0.00027588, 0.00000503, 0.00033599, -0.00010319, -0.00082750, -0.00053224, 0.00074285, 0.00142471, 0.00069438, -0.00039645, -0.00060946, -0.00013937, 0.00011990, -0.00001413, -0.00011252, -0.00003197, 0.00001494, -0.00002946, -0.00003023, 0.00002540, 0.00003245, -0.00001407, -0.00003328, -0.00001056, 0.00000805, },
		{ -0.00000970, 0.00000527, 0.00000906, 0.00000472, -0.00000843, -0.00001547, -0.00000308, 0.00000782, -0.00000394, -0.00001322, 0.00000783, 0.00002819, 0.00000797, -0.00002159, -0.00001104, 0.00000961, -0.00001122, -0.00003263, 0.00000536, 0.00004324, -0.00000135, -0.00005720, -0.00001738, 0.00004183, -0.00000044, -0.00005094, 0.00002665, 0.00010570, 0.00000937, -0.00011932, -0.00006185, 0.00004513, -0.00001854, -0.00008640, 0.00006211, 0.00018122, -0.00001271, -0.00021768, -0.00005882, 0.00014772, -0.00003708, -0.00025875, 0.00000245, 0.00029900, -0.00011209, -0.00076021, -0.00046637, 0.00069438, 0.00129790, 0.00064244, -0.00032463, -0.00053012, -0.00014694, 0.00007468, -0.00001739, -0.00008785, -0.00003097, -0.00000064, -0.00002577, -0.00001316, 0.00002724, 0.00002101, -0.00001778, -0.00002835, },
		{ -0.00002032, -0.00002363, -0.00000883, 0.00000269, 0.00001041, 0.00000899, -0.00000741, -0.00001791, -0.00000469, 0.00000801, -0.00000417, -0.00001383, 0.00000907, 0.00003047, 0.00000780, -0.00002393, -0.00001134, 0.00001163, -0.00001122, -0.00003454, 0.00000649, 0.00004660, -0.00000213, -0.00006163, -0.00001832, 0.00004463, -0.00000036, -0.00005344, 0.00002736, 0.00010894, 0.00000994, -0.00012095, -0.00006238, 0.00004513, -0.00001849, -0.00008570, 0.00006002, 0.00017607, -0.00001067, -0.00020653, -0.00005684, 0.00013545, -0.00003661, -0.00023840, -0.00000053, 0.00025748, -0.00012143, -0.00068685, -0.00039645, 0.00064244, 0.00116538, 0.00058819, -0.00025310, -0.00045092, -0.00015293, 0.00003220, -0.00002037, -0.00006505, -0.00002993, -0.00001430, -0.00002226, 0.00000144, 0.00002826, 0.00001122, },
		{ 0.00002842, 0.00000322, -0.00002163, -0.00001917, -0.00000797, 0.00000035, 0.00001096, 0.00001222, -0.00000606, -0.00001933, -0.00000616, 0.00000771, -0.00000440, -0.00001407, 0.00001019, 0.00003231, 0.00000758, -0.00002596, -0.00001156, 0.00001343, -0.00001122, -0.00003622, 0.00000748, 0.00004950, -0.00000280, -0.00006530, -0.00001906, 0.00004679, -0.00000029, -0.00005515, 0.00002779, 0.00011059, 0.00001013, -0.00012095, -0.00006185, 0.00004467, -0.00001837, -0.00008431, 0.00005662, 0.00016854, -0.00000792, -0.00019223, -0.00005436, 0.00012091, -0.00003604, -0.00021545, -0.00000378, 0.00021280, -0.00013069, -0.00060946, -0.00032463, 0.00058819, 0.00103046, 0.00053278, -0.00018396, -0.00037388, -0.00015688, -0.00000626, -0.00002299, -0.00004471, -0.00002885, -0.00002568, -0.00001899, 0.00001329, },
		{ -0.00001602, 0.00002217, 0.00002770, -0.00000291, -0.00002174, -0.00001500, -0.00000713, -0.00000176, 0.00001075, 0.00001446, -0.00000443, -0.00001981, -0.00000749, 0.00000698, -0.00000461, -0.00001401, 0.00001120, 0.00003375, 0.00000732, -0.00002770, -0.00001171, 0.00001501, -0.00001122, -0.00003765, 0.00000833, 0.00005189, -0.00000335, -0.00006812, -0.00001961, 0.00004827, -0.00000025, -0.00005602, 0.00002794, 0.00011059, 0.00000994, -0.00011932, -0.00006025, 0.00004374, -0.00001815, -0.00008227, 0.00005203, 0.00015885, -0.00000458, -0.00017523, -0.00005147, 0.00010455, -0.00003538, -0.00019061, -0.00000722, 0.00016644, -0.00013937, -0.00053012, -0.00025310, 0.00053278, 0.00089644, 0.00047735, -0.00011921, -0.00030085, -0.00015843, -0.00003956, -0.00002515, -0.00002733, -0.00002776, -0.00003452, },
		{ -0.00002665, -0.00004065, -0.00001336, 0.00002799, 0.00002613, -0.00000717, -0.00002067, -0.00001116, -0.00000631, -0.00000362, 0.00000986, 0.00001578, -0.00000260, -0.00001949, -0.00000867, 0.00000592, -0.00000480, -0.00001370, 0.00001209, 0.00003485, 0.00000704, -0.00002915, -0.00001182, 0.00001635, -0.00001122, -0.00003881, 0.00000901, 0.00005372, -0.00000375, -0.00007005, -0.00001994, 0.00004902, -0.00000024, -0.00005602, 0.00002779, 0.00010894, 0.00000937, -0.00011611, -0.00005764, 0.00004238, -0.00001786, -0.00007960, 0.00004639, 0.00014728, -0.00000078, -0.00015605, -0.00004822, 0.00008691, -0.00003464, -0.00016466, -0.00001070, 0.00011990, -0.00014694, -0.00045092, -0.00018396, 0.00047735, 0.00076642, 0.00042296, -0.00006063, -0.00023350, -0.00015728, -0.00006681, -0.00002681, -0.00001327, },
		{ -0.00002792, -0.00000277, -0.00002554, -0.00004402, -0.00001105, 0.00003080, 0.00002374, -0.00000962, -0.00001854, -0.00000766, -0.00000553, -0.00000525, 0.00000840, 0.00001631, -0.00000062, -0.00001852, -0.00000970, 0.00000464, -0.00000498, -0.00001323, 0.00001286, 0.00003566, 0.00000677, -0.00003034, -0.00001189, 0.00001743, -0.00001122, -0.00003969, 0.00000951, 0.00005496, -0.00000399, -0.00007102, -0.00002005, 0.00004902, -0.00000025, -0.00005515, 0.00002736, 0.00010570, 0.00000845, -0.00011140, -0.00005411, 0.00004060, -0.00001749, -0.00007636, 0.00003991, 0.00013418, 0.00000334, -0.00013529, -0.00004469, 0.00006855, -0.00003381, -0.00013837, -0.00001413, 0.00007468, -0.00015293, -0.00037388, -0.00011921, 0.00042296, 0.00064326, 0.00037060, -0.00000971, -0.00017323, -0.00015325, -0.00008733, },
		{ -0.00014625, -0.00010072, -0.00002846, 0.00000406, -0.00002445, -0.00004467, -0.00000909, 0.00003073, 0.00002064, -0.00001038, -0.00001548, -0.00000451, -0.00000478, -0.00000664, 0.00000647, 0.00001618, 0.00000144, -0.00001706, -0.00001060, 0.00000322, -0.00000514, -0.00001266, 0.00001351, 0.00003624, 0.00000653, -0.00003127, -0.00001194, 0.00001825, -0.00001122, -0.00004028, 0.00000981, 0.00005559, -0.00000407, -0.00007102, -0.00001994, 0.00004827, -0.00000029, -0.00005344, 0.00002665, 0.00010094, 0.00000721, -0.00010531, -0.00004975, 0.00003846, -0.00001705, -0.00007261, 0.00003281, 0.00011991, 0.00000761, -0.00011360, -0.00004097, 0.00005006, -0.00003292, -0.00011252, -0.00001739, 0.00003220, -0.00015688, -0.00030085, -0.00006063, 0.00037060, 0.00052946, 0.00032113, 0.00003233, -0.00012113, },
		{ 0.00006468, -0.00007796, -0.00013630, -0.00010684, -0.00002845, 0.00000726, -0.00002337, -0.00004275, -0.00000747, 0.00002803, 0.00001693, -0.00000964, -0.00001164, -0.00000171, -0.00000409, -0.00000781, 0.00000423, 0.00001552, 0.00000348, -0.00001528, -0.00001136, 0.00000177, -0.00000527, -0.00001207, 0.00001404, 0.00003664, 0.00000631, -0.00003195, -0.00001196, 0.00001880, -0.00001122, -0.00004058, 0.00000991, 0.00005559, -0.00000399, -0.00007005, -0.00001961, 0.00004679, -0.00000036, -0.00005094, 0.00002567, 0.00009482, 0.00000570, -0.00009802, -0.00004472, 0.00003599, -0.00001653, -0.00006842, 0.00002532, 0.00010487, 0.00001185, -0.00009161, -0.00003714, 0.00003201, -0.00003197, -0.00008785, -0.00002037, -0.00000626, -0.00015843, -0.00023350, -0.00000971, 0.00032113, 0.00042707, 0.00027527, },
		{ 0.00033770, 0.00023360, 0.00008685, -0.00004415, -0.00012354, -0.00010585, -0.00002789, 0.00000701, -0.00002233, -0.00003850, -0.00000619, 0.00002303, 0.00001274, -0.00000764, -0.00000724, 0.00000074, -0.00000345, -0.00000877, 0.00000181, 0.00001451, 0.00000545, -0.00001337, -0.00001199, 0.00000039, -0.00000538, -0.00001151, 0.00001445, 0.00003689, 0.00000615, -0.00003240, -0.00001197, 0.00001908, -0.00001122, -0.00004058, 0.00000981, 0.00005496, -0.00000375, -0.00006812, -0.00001906, 0.00004463, -0.00000044, -0.00004773, 0.00002444, 0.00008749, 0.00000399, -0.00008972, -0.00003916, 0.00003324, -0.00001595, -0.00006387, 0.00001770, 0.00008950, 0.00001591, -0.00006998, -0.00003328, 0.00001494, -0.00003097, -0.00006505, -0.00002299, -0.00003956, -0.00015728, -0.00017323, 0.00003233, 0.00027527, },
		{ 0.00006468, 0.00023360, 0.00026239, 0.00019651, 0.00009880, -0.00001975, -0.00010821, -0.00009815, -0.00002684, 0.00000360, -0.00002133, -0.00003225, -0.00000522, 0.00001617, 0.00000823, -0.00000463, -0.00000248, 0.00000286, -0.00000286, -0.00000956, -0.00000065, 0.00001328, 0.00000726, -0.00001149, -0.00001249, -0.00000082, -0.00000547, -0.00001104, 0.00001474, 0.00003704, 0.00000605, -0.00003263, -0.00001198, 0.00001908, -0.00001122, -0.00004028, 0.00000951, 0.00005372, -0.00000335, -0.00006530, -0.00001832, 0.00004183, -0.00000056, -0.00004389, 0.00002298, 0.00007917, 0.00000212, -0.00008062, -0.00003324, 0.00003028, -0.00001530, -0.00005902, 0.00001020, 0.00007418, 0.00001962, -0.00004933, -0.00002946, -0.00000064, -0.00002993, -0.00004471, -0.00002515, -0.00006681, -0.00015325, -0.00012113, },
		{ -0.00014625, -0.00007796, 0.00008685, 0.00019651, 0.00020169, 0.00016424, 0.00010083, -0.00000448, -0.00009068, -0.00008443, -0.00002535, -0.00000255, -0.00002038, -0.00002440, -0.00000452, 0.00000791, 0.00000356, -0.00000091, 0.00000242, 0.00000467, -0.00000235, -0.00001018, -0.00000301, 0.00001199, 0.00000884, -0.00000978, -0.00001288, -0.00000181, -0.00000553, -0.00001071, 0.00001492, 0.00003711, 0.00000602, -0.00003263, -0.00001197, 0.00001880, -0.00001122, -0.00003969, 0.00000901, 0.00005189, -0.00000280, -0.00006163, -0.00001738, 0.00003848, -0.00000069, -0.00003954, 0.00002133, 0.00007008, 0.00000019, -0.00007097, -0.00002715, 0.00002714, -0.00001461, -0.00005397, 0.00000305, 0.00005933, 0.00002283, -0.00003023, -0.00002577, -0.00001430, -0.00002885, -0.00002733, -0.00002681, -0.00008733, },
		{ -0.00002792, -0.00010072, -0.00013630, -0.00004415, 0.00009880, 0.00016424, 0.00015556, 0.00013684, 0.00009364, 0.00000226, -0.00007139, -0.00006556, -0.00002352, -0.00001095, -0.00001949, -0.00001539, -0.00000406, -0.00000123, -0.00000108, 0.00000321, 0.00000723, 0.00000616, -0.00000190, -0.00001066, -0.00000514, 0.00001078, 0.00001013, -0.00000837, -0.00001315, -0.00000250, -0.00000557, -0.00001053, 0.00001498, 0.00003711, 0.00000605, -0.00003240, -0.00001196, 0.00001825, -0.00001122, -0.00003881, 0.00000833, 0.00004950, -0.00000213, -0.00005720, -0.00001626, 0.00003466, -0.00000085, -0.00003481, 0.00001950, 0.00006046, -0.00000173, -0.00006099, -0.00002105, 0.00002391, -0.00001386, -0.00004880, -0.00000350, 0.00004531, 0.00002540, -0.00001316, -0.00002226, -0.00002568, -0.00002776, -0.00001327, },
		{ -0.00002665, -0.00000277, -0.00002846, -0.00010684, -0.00012354, -0.00001975, 0.00010083, 0.00013684, 0.00012349, 0.00011423, 0.00007823, 0.00000135, -0.00005084, -0.00004260, -0.00002142, -0.00002101, -0.00001867, -0.00000570, -0.00000380, -0.00001072, -0.00000553, 0.00000743, 0.00001172, 0.00000738, -0.00000153, -0.00001102, -0.00000692, 0.00000977, 0.00001110, -0.00000738, -0.00001331, -0.00000286, -0.00000559, -0.00001053, 0.00001492, 0.00003704, 0.00000615, -0.00003195, -0.00001194, 0.00001743, -0.00001122, -0.00003765, 0.00000748, 0.00004660, -0.00000135, -0.00005211, -0.00001498, 0.00003048, -0.00000102, -0.00002982, 0.00001753, 0.00005056, -0.00000357, -0.00005093, -0.00001512, 0.00002062, -0.00001308, -0.00004357, -0.00000926, 0.00003245, 0.00002724, 0.00000144, -0.00001899, -0.00003452, },
		{ -0.00001602, -0.00004065, -0.00002554, 0.00000406, -0.00002845, -0.00010585, -0.00010821, -0.00000448, 0.00009364, 0.00011423, 0.00010446, 0.00009616, 0.00005592, -0.00000606, -0.00002964, -0.00001675, -0.00001916, -0.00003212, -0.00001792, 0.00000416, -0.00000369, -0.00002004, -0.00000962, 0.00001145, 0.00001571, 0.00000832, -0.00000124, -0.00001127, -0.00000827, 0.00000904, 0.00001169, -0.00000686, -0.00001336, -0.00000286, -0.00000557, -0.00001071, 0.00001474, 0.00003689, 0.00000631, -0.00003127, -0.00001189, 0.00001635, -0.00001122, -0.00003622, 0.00000649, 0.00004324, -0.00000049, -0.00004647, -0.00001356, 0.00002604, -0.00000122, -0.00002473, 0.00001546, 0.00004065, -0.00000526, -0.00004103, -0.00000953, 0.00001734, -0.00001226, -0.00003838, -0.00001407, 0.00002101, 0.00002826, 0.00001329, },
		{ 0.00002842, 0.00002217, -0.00001336, -0.00004402, -0.00002445, 0.00000726, -0.00002789, -0.00009815, -0.00009068, 0.00000226, 0.00007823, 0.00009616, 0.00009707, 0.00008227, 0.00002823, -0.00001866, -0.00000838, 0.00001068, -0.00001684, -0.00004364, -0.00001726, 0.00001370, -0.00000369, -0.00002867, -0.00001320, 0.00001502, 0.00001900, 0.00000901, -0.00000103, -0.00001143, -0.00000910, 0.00000865, 0.00001189, -0.00000686, -0.00001331, -0.00000250, -0.00000553, -0.00001104, 0.00001445, 0.00003664, 0.00000653, -0.00003034, -0.00001182, 0.00001501, -0.00001122, -0.00003454, 0.00000536, 0.00003948, 0.00000044, -0.00004040, -0.00001202, 0.00002145, -0.00000143, -0.00001966, 0.00001331, 0.00003095, -0.00000673, -0.00003150, -0.00000442, 0.00001412, -0.00001142, -0.00003328, -0.00001778, 0.00001122, },
		{ -0.00002032, 0.00000322, 0.00002770, 0.00002799, -0.00001105, -0.00004467, -0.00002337, 0.00000701, -0.00002684, -0.00008443, -0.00007139, 0.00000135, 0.00005592, 0.00008227, 0.00009955, 0.00007209, -0.00000315, -0.00003501, 0.00001228, 0.00003838, -0.00001457, -0.00005493, -0.00001669, 0.00002247, -0.00000375, -0.00003618, -0.00001614, 0.00001790, 0.00002147, 0.00000947, -0.00000090, -0.00001150, -0.00000938, 0.00000865, 0.00001169, -0.00000738, -0.00001315, -0.00000181, -0.00000547, -0.00001151, 0.00001404, 0.00003624, 0.00000677, -0.00002915, -0.00001171, 0.00001343, -0.00001122, -0.00003263, 0.00000414, 0.00003539, 0.00000140, -0.00003402, -0.00001039, 0.00001682, -0.00000165, -0.00001475, 0.00001111, 0.00002170, -0.00000792, -0.00002254, 0.00000007, 0.00001101, -0.00001056, -0.00002835, },
		{ -0.00000970, -0.00002363, -0.00002163, -0.00000291, 0.00002613, 0.00003080, -0.00000909, -0.00004275, -0.00002233, 0.00000360, -0.00002535, -0.00006556, -0.00005084, -0.00000606, 0.00002823, 0.00007209, 0.00010991, 0.00006509, -0.00003642, -0.00005360, 0.00003173, 0.00006503, -0.00001245, -0.00006539, -0.00001622, 0.00003003, -0.00000384, -0.00004217, -0.00001832, 0.00001993, 0.00002299, 0.00000969, -0.00000086, -0.00001150, -0.00000910, 0.00000904, 0.00001110, -0.00000837, -0.00001288, -0.00000082, -0.00000538, -0.00001207, 0.00001351, 0.00003566, 0.00000704, -0.00002770, -0.00001156, 0.00001163, -0.00001122, -0.00003050, 0.00000285, 0.00003103, 0.00000237, -0.00002746, -0.00000868, 0.00001227, -0.00000189, -0.00001012, 0.00000891, 0.00001310, -0.00000878, -0.00001432, 0.00000386, 0.00000805, },
		{ 0.00000687, 0.00000527, -0.00000883, -0.00001917, -0.00002174, -0.00000717, 0.00002374, 0.00003073, -0.00000747, -0.00003850, -0.00002133, -0.00000255, -0.00002352, -0.00004260, -0.00002964, -0.00001866, -0.00000315, 0.00006509, 0.00012599, 0.00006069, -0.00006977, -0.00007292, 0.00004936, 0.00008940, -0.00001057, -0.00007448, -0.00001584, 0.00003604, -0.00000392, -0.00004634, -0.00001966, 0.00002098, 0.00002351, 0.00000969, -0.00000090, -0.00001143, -0.00000827, 0.00000977, 0.00001013, -0.00000978, -0.00001249, 0.00000039, -0.00000527, -0.00001266, 0.00001286, 0.00003485, 0.00000732, -0.00002596, -0.00001134, 0.00000961, -0.00001122, -0.00002818, 0.00000150, 0.00002647, 0.00000331, -0.00002085, -0.00000692, 0.00000789, -0.00000214, -0.00000588, 0.00000672, 0.00000533, -0.00000927, -0.00000698, },
		{ -0.00000938, -0.00000061, 0.00000906, 0.00000269, -0.00000797, -0.00001500, -0.00002067, -0.00000962, 0.00002064, 0.00002803, -0.00000619, -0.00003225, -0.00002038, -0.00001095, -0.00002142, -0.00001675, -0.00000838, -0.00003501, -0.00003642, 0.00006069, 0.00014559, 0.00005830, -0.00010146, -0.00009155, 0.00006461, 0.00011037, -0.00000902, -0.00008173, -0.00001558, 0.00004021, -0.00000399, -0.00004848, -0.00002011, 0.00002098, 0.00002299, 0.00000947, -0.00000103, -0.00001127, -0.00000692, 0.00001078, 0.00000884, -0.00001149, -0.00001199, 0.00000177, -0.00000514, -0.00001323, 0.00001209, 0.00003375, 0.00000758, -0.00002393, -0.00001104, 0.00000740, -0.00001122, -0.00002570, 0.00000014, 0.00002179, 0.00000422, -0.00001431, -0.00000514, 0.00000377, -0.00000239, -0.00000211, 0.00000458, -0.00000150, },
		{ 0.00000251, -0.00000727, -0.00000910, 0.00000472, 0.00001041, 0.00000035, -0.00000713, -0.00001116, -0.00001854, -0.00001038, 0.00001693, 0.00002303, -0.00000522, -0.00002440, -0.00001949, -0.00002101, -0.00001916, 0.00001068, 0.00001228, -0.00005360, -0.00006977, 0.00005830, 0.00016652, 0.00005733, -0.00012987, -0.00010817, 0.00007701, 0.00012698, -0.00000786, -0.00008677, -0.00001541, 0.00004235, -0.00000401, -0.00004848, -0.00001966, 0.00001993, 0.00002147, 0.00000901, -0.00000124, -0.00001102, -0.00000514, 0.00001199, 0.00000726, -0.00001337, -0.00001136, 0.00000322, -0.00000498, -0.00001370, 0.00001120, 0.00003231, 0.00000780, -0.00002159, -0.00001066, 0.00000501, -0.00001122, -0.00002307, -0.00000122, 0.00001706, 0.00000507, -0.00000794, -0.00000334, -0.00000001, -0.00000265, 0.00000110, },
		{ -0.00000291, 0.00000371, 0.00000053, -0.00001193, -0.00000843, 0.00000899, 0.00001096, -0.00000176, -0.00000631, -0.00000766, -0.00001548, -0.00000964, 0.00001274, 0.00001617, -0.00000452, -0.00001539, -0.00001867, -0.00003212, -0.00001684, 0.00003838, 0.00003173, -0.00007292, -0.00010146, 0.00005733, 0.00018677, 0.00005727, -0.00015357, -0.00012169, 0.00008616, 0.00013849, -0.00000714, -0.00008936, -0.00001536, 0.00004235, -0.00000399, -0.00004634, -0.00001832, 0.00001790, 0.00001900, 0.00000832, -0.00000153, -0.00001066, -0.00000301, 0.00001328, 0.00000545, -0.00001528, -0.00001060, 0.00000464, -0.00000480, -0.00001401, 0.00001019, 0.00003047, 0.00000797, -0.00001894, -0.00001018, 0.00000247, -0.00001122, -0.00002032, -0.00000255, 0.00001234, 0.00000583, -0.00000184, -0.00000157, -0.00000338, },
		{ 0.00000018, -0.00000631, -0.00000318, 0.00000570, -0.00000134, -0.00001547, -0.00000741, 0.00001222, 0.00001075, -0.00000362, -0.00000553, -0.00000451, -0.00001164, -0.00000764, 0.00000823, 0.00000791, -0.00000406, -0.00000570, -0.00001792, -0.00004364, -0.00001457, 0.00006503, 0.00004936, -0.00009155, -0.00012987, 0.00005727, 0.00020454, 0.00005766, -0.00017138, -0.00013120, 0.00009177, 0.00014438, -0.00000690, -0.00008936, -0.00001541, 0.00004021, -0.00000392, -0.00004217, -0.00001614, 0.00001502, 0.00001571, 0.00000738, -0.00000190, -0.00001018, -0.00000065, 0.00001451, 0.00000348, -0.00001706, -0.00000970, 0.00000592, -0.00000461, -0.00001407, 0.00000907, 0.00002819, 0.00000807, -0.00001597, -0.00000960, -0.00000020, -0.00001122, -0.00001749, -0.00000383, 0.00000769, 0.00000649, 0.00000390, },
		{ 0.00000704, 0.00000921, 0.00000187, -0.00000875, -0.00000343, 0.00000706, -0.00000308, -0.00001791, -0.00000606, 0.00001446, 0.00000986, -0.00000525, -0.00000478, -0.00000171, -0.00000724, -0.00000463, 0.00000356, -0.00000123, -0.00000380, 0.00000416, -0.00001726, -0.00005493, -0.00001245, 0.00008940, 0.00006461, -0.00010817, -0.00015357, 0.00005766, 0.00021835, 0.00005813, -0.00018242, -0.00013611, 0.00009366, 0.00014438, -0.00000714, -0.00008677, -0.00001558, 0.00003604, -0.00000384, -0.00003618, -0.00001320, 0.00001145, 0.00001172, 0.00000616, -0.00000235, -0.00000956, 0.00000181, 0.00001552, 0.00000144, -0.00001852, -0.00000867, 0.00000698, -0.00000440, -0.00001383, 0.00000783, 0.00002543, 0.00000809, -0.00001270, -0.00000891, -0.00000297, -0.00001122, -0.00001459, -0.00000504, 0.00000318, },
		{ -0.00000616, -0.00000116, 0.00000748, 0.00001404, 0.00000349, -0.00001072, -0.00000369, 0.00000782, -0.00000469, -0.00001933, -0.00000443, 0.00001578, 0.00000840, -0.00000664, -0.00000409, 0.00000074, -0.00000248, -0.00000091, -0.00000108, -0.00001072, -0.00000369, 0.00001370, -0.00001669, -0.00006539, -0.00001057, 0.00011037, 0.00007701, -0.00012169, -0.00017138, 0.00005813, 0.00022709, 0.00005843, -0.00018617, -0.00013611, 0.00009177, 0.00013849, -0.00000786, -0.00008173, -0.00001584, 0.00003003, -0.00000375, -0.00002867, -0.00000962, 0.00000743, 0.00000723, 0.00000467, -0.00000286, -0.00000877, 0.00000423, 0.00001618, -0.00000062, -0.00001949, -0.00000749, 0.00000771, -0.00000417, -0.00001322, 0.00000648, 0.00002216, 0.00000800, -0.00000912, -0.00000810, -0.00000583, -0.00001122, -0.00001166, },
		},
		{ { 0.00001689, 0.00001782, 0.00001442, -0.00000144, -0.00002369, -0.00003489, -0.00002668, -0.00000845, 0.00000922, 0.00002789, 0.00004534, 0.00003969, -0.00000354, -0.00005154, -0.00004744, 0.00001139, 0.00005319, 0.00002064, -0.00004709, -0.00006461, -0.00001757, 0.00002468, 0.00001494, -0.00001323, -0.00001589, -0.00000966, -0.00002537, -0.00003622, 0.00000094, 0.00005446, 0.00004517, -0.00002994, -0.00007559, -0.00002994, 0.00004517, 0.00005446, 0.00000094, -0.00003622, -0.00002537, -0.00000966, -0.00001589, -0.00001323, 0.00001494, 0.00002468, -0.00001757, -0.00006461, -0.00004709, 0.00002064, 0.00005319, 0.00001139, -0.00004744, -0.00005154, -0.00000354, 0.00003969, 0.00004534, 0.00002789, 0.00000922, -0.00000845, -0.00002668, -0.00003489, -0.00002369, -0.00000144, 0.00001442, 0.00001782, },
		{ 0.00001141, 0.00001808, 0.00001669, 0.00001698, 0.00001660, 0.00000473, -0.00001740, -0.00003178, -0.00002676, -0.00001276, -0.00000240, 0.00001220, 0.00003930, 0.00005408, 0.00002175, -0.00003997, -0.00006170, -0.00001167, 0.00004954, 0.00004279, -0.00002250, -0.00006394, -0.00004053, 0.00000382, 0.00001848, 0.00001045, 0.00000389, -0.00001063, -0.00004015, -0.00004483, 0.00000762, 0.00006602, 0.00004689, -0.00003820, -0.00008095, -0.00002522, 0.00005155, 0.00005100, -0.00000934, -0.00003836, -0.00001340, 0.00000211, -0.00002316, -0.00003782, -0.00000280, 0.00003360, 0.00000881, -0.00005030, -0.00006093, -0.00000526, 0.00004428, 0.00002722, -0.00002807, -0.00005232, -0.00002400, 0.00002009, 0.00004220, 0.00003858, 0.00002228, 0.00000023, -0.00002322, -0.00003676, -0.00003018, -0.00000839, },
		{ -0.00003608, -0.00001612, 0.00000742, 0.00001780, 0.00001608, 0.00001545, 0.00001803, 0.00001041, -0.00001150, -0.00002872, -0.00002497, -0.00001246, -0.00000968, -0.00000495, 0.00002456, 0.00005938, 0.00004691, -0.00001844, -0.00006605, -0.00003614, 0.00003349, 0.00005445, 0.00000565, -0.00004801, -0.00005198, -0.00002154, 0.00000648, 0.00002483, 0.00003068, 0.00000287, -0.00004998, -0.00006231, 0.00000472, 0.00008010, 0.00005912, -0.00004286, -0.00009481, -0.00002969, 0.00006015, 0.00005741, -0.00001632, -0.00004874, -0.00000984, 0.00001797, -0.00001692, -0.00005529, -0.00002900, 0.00002726, 0.00002982, -0.00002584, -0.00006035, -0.00002727, 0.00002639, 0.00003255, -0.00000938, -0.00004442, -0.00003645, 0.00000001, 0.00003164, 0.00004214, 0.00003343, 0.00001168, -0.00001571, -0.00003597, },
		{ -0.00000466, -0.00003142, -0.00004029, -0.00002428, 0.00000239, 0.00001698, 0.00001507, 0.00001304, 0.00001855, 0.00001590, -0.00000565, -0.00002639, -0.00002316, -0.00000868, -0.00001094, -0.00001962, 0.00000378, 0.00005367, 0.00006657, 0.00000938, -0.00005816, -0.00005547, 0.00000901, 0.00005195, 0.00002886, -0.00002141, -0.00004721, -0.00004198, -0.00001718, 0.00002403, 0.00005599, 0.00002981, -0.00004764, -0.00008262, -0.00001047, 0.00008935, 0.00007985, -0.00003798, -0.00011130, -0.00004532, 0.00006401, 0.00007115, -0.00001488, -0.00006172, -0.00001644, 0.00003072, -0.00000061, -0.00006050, -0.00005567, 0.00000725, 0.00003876, 0.00000104, -0.00004617, -0.00003927, 0.00000565, 0.00002754, 0.00000377, -0.00003203, -0.00004028, -0.00001636, 0.00001710, 0.00003838, 0.00003980, 0.00002328, },
		{ 0.00003981, 0.00003215, 0.00000833, -0.00002272, -0.00004162, -0.00003218, -0.00000356, 0.00001571, 0.00001380, 0.00000952, 0.00001773, 0.00002128, 0.00000085, -0.00002468, -0.00002287, -0.00000353, -0.00000613, -0.00002841, -0.00001890, 0.00003756, 0.00007636, 0.00003800, -0.00003899, -0.00006436, -0.00001731, 0.00003589, 0.00003996, 0.00000764, -0.00002663, -0.00004915, -0.00004435, 0.00000695, 0.00007102, 0.00006443, -0.00002911, -0.00009747, -0.00003600, 0.00008731, 0.00010311, -0.00002113, -0.00012309, -0.00006971, 0.00005767, 0.00008646, -0.00000345, -0.00007087, -0.00003082, 0.00003514, 0.00001955, -0.00005253, -0.00007527, -0.00002086, 0.00003290, 0.00002234, -0.00002319, -0.00003848, -0.00001135, 0.00001561, 0.00000887, -0.00002001, -0.00003747, -0.00002648, 0.00000264, 0.00002908, },
		{ -0.00000825, 0.00001744, 0.00003361, 0.00003588, 0.00002085, -0.00001045, -0.00003919, -0.00003893, -0.00001016, 0.00001416, 0.00001261, 0.00000487, 0.00001493, 0.00002615, 0.00000864, -0.00002275, -0.00002482, 0.00000054, 0.00000318, -0.00002940, -0.00003893, 0.00001397, 0.00007417, 0.00006155, -0.00001253, -0.00006039, -0.00003826, 0.00001103, 0.00003541, 0.00002985, 0.00000414, -0.00003858, -0.00006503, -0.00002207, 0.00006970, 0.00009787, 0.00000480, -0.00009904, -0.00006562, 0.00007090, 0.00012108, 0.00000557, -0.00012408, -0.00009686, 0.00003926, 0.00009653, 0.00001542, -0.00007140, -0.00004775, 0.00002957, 0.00003685, -0.00003453, -0.00008304, -0.00004959, 0.00001399, 0.00003220, 0.00000146, -0.00002620, -0.00001967, 0.00000215, 0.00000653, -0.00001221, -0.00003170, -0.00003019, },
		{ -0.00002701, -0.00002957, -0.00001376, 0.00000711, 0.00002312, 0.00003322, 0.00003029, 0.00000394, -0.00003267, -0.00004370, -0.00001706, 0.00001255, 0.00001206, -0.00000057, 0.00000951, 0.00002957, 0.00001795, -0.00001926, -0.00002869, 0.00000150, 0.00001433, -0.00002252, -0.00005245, -0.00001259, 0.00006058, 0.00007544, 0.00001538, -0.00004474, -0.00004830, -0.00001525, 0.00001647, 0.00003771, 0.00003581, -0.00001139, -0.00007061, -0.00005429, 0.00005094, 0.00012095, 0.00004837, -0.00008270, -0.00009061, 0.00004180, 0.00012678, 0.00003600, -0.00011168, -0.00011935, 0.00001120, 0.00009607, 0.00003620, -0.00006189, -0.00006128, 0.00001612, 0.00004647, -0.00001218, -0.00007812, -0.00007206, -0.00001271, 0.00002855, 0.00002078, -0.00000712, -0.00001761, -0.00000742, 0.00000008, -0.00001018, },
		{ -0.00000569, -0.00001266, -0.00002627, -0.00002806, -0.00001420, 0.00000105, 0.00001143, 0.00002450, 0.00003448, 0.00001839, -0.00002239, -0.00004583, -0.00002393, 0.00001098, 0.00001283, -0.00000596, 0.00000120, 0.00003024, 0.00002826, -0.00001298, -0.00003315, -0.00000167, 0.00002425, -0.00000949, -0.00005720, -0.00003724, 0.00003860, 0.00007750, 0.00003868, -0.00002177, -0.00004513, -0.00003517, -0.00001115, 0.00002797, 0.00005823, 0.00002573, -0.00005684, -0.00007926, 0.00001930, 0.00012734, 0.00009268, -0.00004870, -0.00010262, 0.00000612, 0.00011658, 0.00006213, -0.00008779, -0.00013097, -0.00002071, 0.00008323, 0.00005249, -0.00004464, -0.00006705, -0.00000058, 0.00004682, 0.00000856, -0.00006326, -0.00008404, -0.00004047, 0.00001334, 0.00003002, 0.00001238, -0.00000708, -0.00000952, },
		{ 0.00000719, -0.00000371, -0.00000640, -0.00001608, -0.00003010, -0.00002924, -0.00001192, 0.00000068, 0.00000199, 0.00001153, 0.00003220, 0.00003063, -0.00000930, -0.00004497, -0.00003055, 0.00000932, 0.00001548, -0.00001005, -0.00000963, 0.00002683, 0.00003824, -0.00000321, -0.00003626, -0.00000870, 0.00003036, 0.00000674, -0.00005290, -0.00005589, 0.00001280, 0.00006839, 0.00005283, 0.00000235, -0.00003043, -0.00004301, -0.00003908, 0.00000271, 0.00006360, 0.00006241, -0.00002552, -0.00008824, -0.00001629, 0.00011575, 0.00012851, -0.00000220, -0.00009635, -0.00002750, 0.00009153, 0.00007665, -0.00005824, -0.00012891, -0.00004932, 0.00006011, 0.00005933, -0.00002460, -0.00006373, -0.00001536, 0.00003958, 0.00002342, -0.00004335, -0.00008482, -0.00006338, -0.00000857, 0.00002806, 0.00002648, },
		{ 0.00001730, 0.00003190, 0.00001944, 0.00000717, 0.00000023, -0.00001592, -0.00003651, -0.00003538, -0.00001044, 0.00000539, -0.00000235, -0.00000278, 0.00002351, 0.00003868, 0.00000525, -0.00004104, -0.00003681, 0.00000708, 0.00002016, -0.00001133, -0.00002179, 0.00001847, 0.00004593, 0.00000975, -0.00003607, -0.00001812, 0.00003116, 0.00002280, -0.00004114, -0.00006613, -0.00001201, 0.00005121, 0.00005596, 0.00002154, -0.00000915, -0.00003685, -0.00005887, -0.00003124, 0.00004896, 0.00008773, 0.00001580, -0.00007721, -0.00004534, 0.00009053, 0.00014942, 0.00004847, -0.00007121, -0.00005059, 0.00005700, 0.00007545, -0.00003073, -0.00011459, -0.00006868, 0.00003194, 0.00005492, -0.00000738, -0.00005322, -0.00002458, 0.00002850, 0.00003104, -0.00002330, -0.00007680, -0.00007812, -0.00003136, },
		{ -0.00008455, -0.00005044, 0.00000245, 0.00002882, 0.00002503, 0.00001796, 0.00001333, -0.00000837, -0.00004143, -0.00004651, -0.00001329, 0.00001270, -0.00000016, -0.00001512, 0.00000974, 0.00004116, 0.00001968, -0.00003419, -0.00004254, 0.00000352, 0.00002638, -0.00000852, -0.00003332, 0.00000516, 0.00004921, 0.00002453, -0.00003123, -0.00002767, 0.00002653, 0.00003573, -0.00002477, -0.00006756, -0.00003193, 0.00003044, 0.00004914, 0.00003155, 0.00001207, -0.00001906, -0.00006462, -0.00006437, 0.00001724, 0.00009358, 0.00005591, -0.00004848, -0.00005934, 0.00006021, 0.00015393, 0.00009474, -0.00003123, -0.00005749, 0.00002085, 0.00005871, -0.00001230, -0.00009302, -0.00007593, 0.00000498, 0.00004101, 0.00000287, -0.00003959, -0.00002725, 0.00001753, 0.00003274, -0.00000633, -0.00006392, },
		{ 0.00000688, -0.00004974, -0.00008504, -0.00006388, -0.00001150, 0.00002062, 0.00002214, 0.00002307, 0.00002888, 0.00000797, -0.00004012, -0.00006010, -0.00002267, 0.00001901, 0.00000816, -0.00002262, -0.00000679, 0.00003760, 0.00003249, -0.00002471, -0.00004746, -0.00000217, 0.00003292, -0.00000098, -0.00004184, -0.00001196, 0.00004637, 0.00003885, -0.00002148, -0.00003493, 0.00001771, 0.00004365, -0.00000701, -0.00006154, -0.00004483, 0.00001057, 0.00003576, 0.00003110, 0.00002704, 0.00000456, -0.00005477, -0.00008744, -0.00002362, 0.00007733, 0.00008335, -0.00001021, -0.00005481, 0.00003461, 0.00014586, 0.00013071, 0.00001670, -0.00004680, -0.00000921, 0.00003060, -0.00000715, -0.00007085, -0.00007204, -0.00001569, 0.00002184, 0.00000488, -0.00002711, -0.00002508, 0.00000905, 0.00003128, },
		{ 0.00000305, 0.00002908, 0.00001777, -0.00003592, -0.00008283, -0.00007273, -0.00002130, 0.00001242, 0.00001258, 0.00001853, 0.00004089, 0.00003124, -0.00002877, -0.00007168, -0.00003865, 0.00002064, 0.00002048, -0.00002347, -0.00002329, 0.00002837, 0.00004233, -0.00001299, -0.00005094, -0.00001053, 0.00003801, 0.00001093, -0.00004508, -0.00003063, 0.00003668, 0.00005005, -0.00000779, -0.00003799, 0.00000680, 0.00004604, 0.00000927, -0.00005053, -0.00005048, -0.00000512, 0.00002033, 0.00002201, 0.00003183, 0.00002668, -0.00003249, -0.00009427, -0.00006300, 0.00004265, 0.00009004, 0.00002600, -0.00003464, 0.00002116, 0.00013280, 0.00015513, 0.00006562, -0.00002117, -0.00002797, -0.00000265, -0.00001562, -0.00005396, -0.00006091, -0.00002788, 0.00000225, 0.00000027, -0.00001840, -0.00002114, },
		{ -0.00001345, -0.00001804, -0.00000254, 0.00002669, 0.00002848, -0.00002171, -0.00008016, -0.00008006, -0.00002672, 0.00000895, 0.00000117, 0.00000366, 0.00004330, 0.00005645, -0.00000610, -0.00007608, -0.00005880, 0.00001498, 0.00003349, -0.00001737, -0.00003706, 0.00001464, 0.00004820, 0.00000044, -0.00005212, -0.00002150, 0.00003964, 0.00002567, -0.00004151, -0.00004787, 0.00002073, 0.00005569, 0.00000777, -0.00003595, -0.00000380, 0.00004362, 0.00002224, -0.00003726, -0.00005024, -0.00001518, 0.00000698, 0.00000831, 0.00002589, 0.00004077, -0.00000455, -0.00008359, -0.00009090, -0.00000149, 0.00007382, 0.00004875, -0.00000746, 0.00002212, 0.00012294, 0.00017153, 0.00011135, 0.00001461, -0.00003392, -0.00003519, -0.00003489, -0.00004560, -0.00004749, -0.00003259, -0.00001448, -0.00000759, },
		{ -0.00002783, -0.00001562, -0.00000988, -0.00001628, -0.00001032, 0.00002226, 0.00003999, -0.00000469, -0.00007686, -0.00008914, -0.00003056, 0.00001255, -0.00000586, -0.00001835, 0.00003205, 0.00007667, 0.00002584, -0.00006903, -0.00007872, 0.00000136, 0.00004360, -0.00000547, -0.00004598, -0.00000190, 0.00004940, 0.00001480, -0.00004993, -0.00003420, 0.00003617, 0.00004075, -0.00003085, -0.00006056, 0.00000051, 0.00005425, 0.00002251, -0.00002917, -0.00001207, 0.00003795, 0.00003120, -0.00002395, -0.00004624, -0.00002004, -0.00000175, -0.00000519, 0.00001196, 0.00004312, 0.00002093, -0.00005930, -0.00010092, -0.00004364, 0.00003928, 0.00005061, 0.00001518, 0.00003333, 0.00012157, 0.00018629, 0.00015404, 0.00005634, -0.00002866, -0.00006361, -0.00006089, -0.00004580, -0.00003548, -0.00003291, },
		{ -0.00002587, -0.00003189, -0.00003997, -0.00002299, -0.00000454, -0.00001356, -0.00002150, 0.00001234, 0.00005082, 0.00001753, -0.00007017, -0.00010150, -0.00003742, 0.00002204, -0.00000315, -0.00004114, 0.00000672, 0.00008494, 0.00006174, -0.00004865, -0.00009317, -0.00001868, 0.00004784, 0.00000990, -0.00004882, -0.00001929, 0.00004564, 0.00002902, -0.00004351, -0.00004692, 0.00002685, 0.00005318, -0.00001428, -0.00006623, -0.00002097, 0.00004550, 0.00003376, -0.00001907, -0.00001680, 0.00003089, 0.00003650, -0.00001190, -0.00004053, -0.00002142, -0.00000537, -0.00001456, -0.00000508, 0.00003383, 0.00003702, -0.00002902, -0.00009212, -0.00007336, -0.00000373, 0.00003061, 0.00002271, 0.00004543, 0.00012834, 0.00020532, 0.00019782, 0.00010303, -0.00001479, -0.00008767, -0.00009066, -0.00005241, },
		{ -0.00012414, -0.00006331, -0.00001684, -0.00003039, -0.00005415, -0.00003197, 0.00000431, -0.00000558, -0.00003428, -0.00000629, 0.00005671, 0.00004533, -0.00005579, -0.00011555, -0.00005166, 0.00003280, 0.00001171, -0.00005697, -0.00002893, 0.00007636, 0.00009422, -0.00001628, -0.00009746, -0.00004165, 0.00004465, 0.00002587, -0.00004541, -0.00003557, 0.00003700, 0.00004169, -0.00003246, -0.00005738, 0.00001218, 0.00006017, 0.00000564, -0.00006361, -0.00004031, 0.00003070, 0.00003951, -0.00000782, -0.00001781, 0.00002402, 0.00003904, -0.00000148, -0.00003446, -0.00002143, -0.00000531, -0.00001787, -0.00001991, 0.00001648, 0.00004007, -0.00000157, -0.00006907, -0.00008442, -0.00004321, -0.00000528, 0.00000934, 0.00004700, 0.00013663, 0.00023055, 0.00024843, 0.00015792, 0.00000698, -0.00010920, },
		{ 0.00004013, -0.00012947, -0.00016429, -0.00007888, -0.00000537, -0.00002590, -0.00007229, -0.00004737, 0.00001506, 0.00001197, -0.00004351, -0.00003442, 0.00005182, 0.00007572, -0.00002984, -0.00012654, -0.00007524, 0.00003824, 0.00003704, -0.00005906, -0.00006760, 0.00004975, 0.00011558, 0.00002354, -0.00008879, -0.00006300, 0.00003422, 0.00003967, -0.00003648, -0.00004901, 0.00002403, 0.00005119, -0.00001726, -0.00006323, -0.00000602, 0.00005986, 0.00002551, -0.00005308, -0.00005440, 0.00001225, 0.00003887, 0.00000226, -0.00001583, 0.00001833, 0.00003988, 0.00000765, -0.00002839, -0.00002161, -0.00000401, -0.00001563, -0.00002856, -0.00000319, 0.00003071, 0.00001585, -0.00004009, -0.00007654, -0.00006880, -0.00004600, -0.00002359, 0.00002881, 0.00013518, 0.00025748, 0.00030954, 0.00022759, },
		{ 0.00037905, 0.00031879, 0.00009301, -0.00014605, -0.00021530, -0.00010351, 0.00000999, -0.00001297, -0.00009264, -0.00007430, 0.00002203, 0.00004080, -0.00004182, -0.00006874, 0.00003092, 0.00010220, 0.00000882, -0.00012781, -0.00010629, 0.00003205, 0.00006713, -0.00004381, -0.00010030, 0.00000829, 0.00011985, 0.00006392, -0.00006717, -0.00007822, 0.00001840, 0.00004918, -0.00002344, -0.00005817, 0.00000777, 0.00005595, 0.00000067, -0.00006271, -0.00002486, 0.00005185, 0.00004185, -0.00003657, -0.00006120, -0.00000675, 0.00003224, 0.00000928, -0.00001222, 0.00001409, 0.00003974, 0.00001601, -0.00002189, -0.00002244, -0.00000384, -0.00001027, -0.00002957, -0.00001937, 0.00001327, 0.00001982, -0.00001444, -0.00005522, -0.00007512, -0.00007901, -0.00006756, -0.00001229, 0.00011198, 0.00027507, },
		{ 0.00005895, 0.00026820, 0.00044695, 0.00043418, 0.00017706, -0.00015045, -0.00027928, -0.00014524, 0.00002636, 0.00001443, -0.00010872, -0.00011527, 0.00001662, 0.00007784, -0.00002220, -0.00010152, -0.00000797, 0.00011613, 0.00005708, -0.00011309, -0.00013895, 0.00001057, 0.00009375, -0.00001211, -0.00011867, -0.00004100, 0.00010445, 0.00009727, -0.00003553, -0.00008403, 0.00000019, 0.00005335, -0.00000809, -0.00006212, -0.00001020, 0.00005479, 0.00001906, -0.00005513, -0.00004106, 0.00003740, 0.00005189, -0.00001713, -0.00006013, -0.00002327, 0.00002117, 0.00001211, -0.00000851, 0.00001098, 0.00003893, 0.00002389, -0.00001423, -0.00002327, -0.00000615, -0.00000508, -0.00002424, -0.00002804, -0.00000591, 0.00001159, 0.00000080, -0.00002961, -0.00006340, -0.00009446, -0.00010931, -0.00007120, },
		{ -0.00013506, -0.00013510, -0.00002391, 0.00022240, 0.00049573, 0.00056823, 0.00030280, -0.00012783, -0.00035240, -0.00021322, 0.00003493, 0.00005946, -0.00011026, -0.00016739, -0.00001001, 0.00011460, 0.00001890, -0.00012206, -0.00006166, 0.00010931, 0.00010731, -0.00007905, -0.00016473, -0.00002553, 0.00010851, 0.00003060, -0.00011728, -0.00008889, 0.00007095, 0.00011706, 0.00000088, -0.00007914, -0.00001705, 0.00005219, 0.00000774, -0.00006043, -0.00002791, 0.00004734, 0.00003515, -0.00004125, -0.00005169, 0.00001916, 0.00005426, 0.00000176, -0.00005217, -0.00003495, 0.00000792, 0.00001062, -0.00000601, 0.00000834, 0.00003731, 0.00003113, -0.00000490, -0.00002271, -0.00001078, -0.00000284, -0.00001589, -0.00002821, -0.00002076, -0.00000377, 0.00000293, -0.00000925, -0.00004086, -0.00008862, },
		{ -0.00001814, -0.00006533, -0.00013529, -0.00018671, -0.00012658, 0.00012930, 0.00050378, 0.00070506, 0.00047455, -0.00005953, -0.00042226, -0.00031340, 0.00002166, 0.00011899, -0.00008614, -0.00022121, -0.00006319, 0.00013847, 0.00007902, -0.00011962, -0.00012130, 0.00007701, 0.00014879, -0.00002720, -0.00017488, -0.00007088, 0.00010539, 0.00007568, -0.00009514, -0.00012613, 0.00002483, 0.00011943, 0.00003570, -0.00006456, -0.00003040, 0.00004668, 0.00002239, -0.00005327, -0.00004319, 0.00003423, 0.00004633, -0.00002319, -0.00005492, 0.00000056, 0.00004926, 0.00001703, -0.00003949, -0.00004060, -0.00000507, 0.00000560, -0.00000551, 0.00000548, 0.00003453, 0.00003712, 0.00000595, -0.00001941, -0.00001620, -0.00000481, -0.00000845, -0.00002190, -0.00002754, -0.00001942, -0.00000569, -0.00000067, },
		{ -0.00001879, -0.00000516, -0.00000540, -0.00003492, -0.00010830, -0.00020933, -0.00023007, -0.00000873, 0.00045122, 0.00081932, 0.00068568, 0.00007193, -0.00046767, -0.00044382, -0.00002951, 0.00018158, -0.00002874, -0.00026178, -0.00014144, 0.00013596, 0.00014868, -0.00008706, -0.00017381, 0.00002030, 0.00017043, 0.00003557, -0.00016323, -0.00011659, 0.00008268, 0.00011351, -0.00005599, -0.00014572, -0.00002584, 0.00010402, 0.00006293, -0.00004329, -0.00003800, 0.00003832, 0.00003455, -0.00004144, -0.00005403, 0.00001717, 0.00005076, -0.00000404, -0.00005052, -0.00001501, 0.00003876, 0.00002671, -0.00002490, -0.00004034, -0.00001574, -0.00000159, -0.00000720, 0.00000192, 0.00003030, 0.00004108, 0.00001746, -0.00001263, -0.00002025, -0.00001036, -0.00000492, -0.00001314, -0.00002601, -0.00002935, },
		{ -0.00001923, -0.00003060, -0.00002877, -0.00001833, -0.00000862, -0.00001092, -0.00006158, -0.00019242, -0.00031063, -0.00017658, 0.00032647, 0.00088052, 0.00091598, 0.00027633, -0.00046160, -0.00059118, -0.00013139, 0.00022811, 0.00006176, -0.00027205, -0.00023478, 0.00009698, 0.00021263, -0.00002401, -0.00020501, -0.00005302, 0.00016408, 0.00009818, -0.00012840, -0.00015251, 0.00004364, 0.00013607, -0.00000731, -0.00014450, -0.00007226, 0.00007400, 0.00007827, -0.00001954, -0.00003940, 0.00002877, 0.00004333, -0.00002627, -0.00005893, -0.00000129, 0.00004785, 0.00001287, -0.00003989, -0.00002506, 0.00002560, 0.00003031, -0.00001110, -0.00003534, -0.00002279, -0.00000934, -0.00001071, -0.00000255, 0.00002453, 0.00004235, 0.00002830, -0.00000270, -0.00002105, -0.00001732, -0.00000630, -0.00000625, },
		{ -0.00001132, -0.00000403, -0.00001207, -0.00002436, -0.00002967, -0.00003175, -0.00002719, -0.00000539, -0.00001017, -0.00013568, -0.00034597, -0.00034795, 0.00013175, 0.00086003, 0.00113268, 0.00054992, -0.00037721, -0.00073023, -0.00028773, 0.00023543, 0.00017463, -0.00023801, -0.00032555, 0.00001899, 0.00025328, 0.00006169, -0.00020362, -0.00012962, 0.00012734, 0.00014803, -0.00007472, -0.00017015, -0.00000422, 0.00013904, 0.00004162, -0.00012373, -0.00010692, 0.00003520, 0.00007995, 0.00000228, -0.00003548, 0.00001949, 0.00004830, -0.00000955, -0.00005726, -0.00001826, 0.00003846, 0.00002474, -0.00002571, -0.00002855, 0.00001287, 0.00002876, -0.00000007, -0.00002746, -0.00002586, -0.00001623, -0.00001531, -0.00000783, 0.00001737, 0.00004067, 0.00003713, 0.00000910, -0.00001774, -0.00002303, },
		{ -0.00001088, -0.00002540, -0.00001715, -0.00000672, -0.00000919, -0.00001539, -0.00001964, -0.00003606, -0.00005369, -0.00002461, 0.00002753, -0.00005036, -0.00032187, -0.00049046, -0.00011452, 0.00073911, 0.00129547, 0.00087191, -0.00019554, -0.00082682, -0.00048871, 0.00018227, 0.00028871, -0.00015390, -0.00039193, -0.00009064, 0.00025543, 0.00015479, -0.00016490, -0.00019359, 0.00006489, 0.00017427, -0.00001154, -0.00016506, -0.00005119, 0.00012280, 0.00008209, -0.00008850, -0.00012518, -0.00000534, 0.00006896, 0.00001856, -0.00002810, 0.00001147, 0.00004943, 0.00000673, -0.00004944, -0.00003109, 0.00002479, 0.00003000, -0.00001125, -0.00002610, 0.00000302, 0.00002399, 0.00000735, -0.00001863, -0.00002541, -0.00002124, -0.00002010, -0.00001361, 0.00000916, 0.00003624, 0.00004301, 0.00002101, },
		{ 0.00004573, 0.00003145, -0.00000214, -0.00002392, -0.00002066, -0.00001186, -0.00001295, -0.00001018, -0.00000198, -0.00002464, -0.00007614, -0.00006637, 0.00003534, 0.00004273, -0.00023709, -0.00057314, -0.00037937, 0.00051554, 0.00136461, 0.00120510, 0.00008723, -0.00084449, -0.00070940, 0.00005620, 0.00037591, -0.00002594, -0.00041350, -0.00021484, 0.00021098, 0.00023583, -0.00009281, -0.00023052, -0.00001209, 0.00017093, 0.00004911, -0.00013818, -0.00008797, 0.00009207, 0.00010805, -0.00004623, -0.00012601, -0.00004098, 0.00004859, 0.00002717, -0.00001954, 0.00000519, 0.00004710, 0.00002071, -0.00003694, -0.00003797, 0.00000979, 0.00002867, 0.00000055, -0.00001965, -0.00000269, 0.00001827, 0.00001144, -0.00001034, -0.00002242, -0.00002391, -0.00002423, -0.00001948, 0.00000026, 0.00002956, },
		{ -0.00000905, 0.00002115, 0.00004569, 0.00003963, 0.00000642, -0.00001981, -0.00001992, -0.00001539, -0.00002223, -0.00001427, 0.00001581, 0.00000357, -0.00008184, -0.00011985, 0.00000457, 0.00011837, -0.00010518, -0.00057444, -0.00062109, 0.00020711, 0.00131036, 0.00150106, 0.00045500, -0.00075314, -0.00091234, -0.00014064, 0.00040749, 0.00012688, -0.00037724, -0.00033005, 0.00012217, 0.00028596, 0.00000022, -0.00023148, -0.00008848, 0.00013875, 0.00009558, -0.00009561, -0.00010815, 0.00005446, 0.00011728, -0.00000470, -0.00011180, -0.00006686, 0.00002341, 0.00002775, -0.00001190, 0.00000063, 0.00004200, 0.00003094, -0.00002196, -0.00003836, -0.00000351, 0.00002226, 0.00000775, -0.00001187, -0.00000439, 0.00001346, 0.00001327, -0.00000338, -0.00001803, -0.00002431, -0.00002700, -0.00002493, },
		{ -0.00002796, -0.00002936, -0.00001855, 0.00001129, 0.00004358, 0.00004577, 0.00001331, -0.00001559, -0.00001519, -0.00001332, -0.00003259, -0.00002979, 0.00002444, 0.00004305, -0.00006183, -0.00016836, -0.00006262, 0.00015362, 0.00004783, -0.00048839, -0.00079792, -0.00014973, 0.00112131, 0.00170899, 0.00087144, -0.00053776, -0.00105414, -0.00038873, 0.00036171, 0.00027567, -0.00028221, -0.00041169, 0.00000217, 0.00029203, 0.00009592, -0.00019565, -0.00014904, 0.00008507, 0.00011969, -0.00004690, -0.00010972, 0.00001820, 0.00011148, 0.00002978, -0.00008732, -0.00008065, -0.00000180, 0.00002151, -0.00000674, -0.00000253, 0.00003505, 0.00003663, -0.00000699, -0.00003307, -0.00001280, 0.00001327, 0.00000992, -0.00000526, -0.00000345, 0.00001052, 0.00001426, 0.00000228, -0.00001320, -0.00002288, },
		{ -0.00000846, -0.00002026, -0.00002691, -0.00003212, -0.00002794, -0.00000008, 0.00003982, 0.00005085, 0.00001833, -0.00001390, -0.00000911, -0.00000365, -0.00003771, -0.00005420, 0.00001620, 0.00008277, -0.00001437, -0.00019408, -0.00015301, 0.00013432, 0.00018917, -0.00032721, -0.00087757, -0.00050588, 0.00080910, 0.00178643, 0.00128462, -0.00020465, -0.00109470, -0.00065282, 0.00023078, 0.00038770, -0.00014143, -0.00044056, -0.00012742, 0.00025044, 0.00017469, -0.00013083, -0.00018232, 0.00002202, 0.00011878, -0.00000239, -0.00009531, -0.00001006, 0.00009529, 0.00005362, -0.00005829, -0.00008264, -0.00002310, 0.00001074, -0.00000477, -0.00000481, 0.00002721, 0.00003781, 0.00000575, -0.00002397, -0.00001697, 0.00000448, 0.00000807, -0.00000149, -0.00000185, 0.00000930, 0.00001551, 0.00000720, },
		{ 0.00001744, 0.00001216, -0.00000388, -0.00001711, -0.00002398, -0.00003245, -0.00003663, -0.00001328, 0.00003418, 0.00005600, 0.00002276, -0.00001632, -0.00000585, 0.00001241, -0.00003173, -0.00008044, -0.00001202, 0.00010936, 0.00005369, -0.00018343, -0.00024523, 0.00005940, 0.00028721, -0.00011941, -0.00084494, -0.00080914, 0.00040819, 0.00170912, 0.00163608, 0.00021679, -0.00100700, -0.00088773, 0.00002487, 0.00043428, 0.00001994, -0.00040829, -0.00024158, 0.00016851, 0.00022056, -0.00005155, -0.00018363, -0.00003678, 0.00009625, 0.00002955, -0.00007107, -0.00002679, 0.00007460, 0.00006626, -0.00002986, -0.00007511, -0.00003806, -0.00000186, -0.00000593, -0.00000674, 0.00001940, 0.00003522, 0.00001478, -0.00001350, -0.00001635, -0.00000189, 0.00000423, -0.00000090, -0.00000145, 0.00000866, },
		{ -0.00000335, 0.00000703, 0.00001959, 0.00001772, 0.00000082, -0.00001394, -0.00001958, -0.00002969, -0.00004357, -0.00002841, 0.00002567, 0.00006167, 0.00002893, -0.00002231, -0.00000959, 0.00003032, -0.00001164, -0.00009896, -0.00005699, 0.00011122, 0.00012813, -0.00013153, -0.00031534, -0.00005819, 0.00031927, 0.00009631, -0.00070601, -0.00101503, -0.00002976, 0.00147747, 0.00187220, 0.00067761, -0.00078472, -0.00104762, -0.00022804, 0.00039836, 0.00016996, -0.00032016, -0.00031808, 0.00006298, 0.00022542, 0.00002476, -0.00015608, -0.00007984, 0.00006022, 0.00004463, -0.00004461, -0.00003209, 0.00005475, 0.00006963, -0.00000563, -0.00006143, -0.00004590, -0.00001385, -0.00000949, -0.00000875, 0.00001226, 0.00003013, 0.00001962, -0.00000398, -0.00001245, -0.00000479, 0.00000072, -0.00000253, },
		{ -0.00000068, -0.00000451, -0.00000744, 0.00000303, 0.00002081, 0.00002391, 0.00000609, -0.00001112, -0.00001442, -0.00002344, -0.00004724, -0.00004498, 0.00001282, 0.00006699, 0.00003920, -0.00002905, -0.00002274, 0.00004335, 0.00002107, -0.00010062, -0.00010940, 0.00008236, 0.00019098, -0.00004421, -0.00034321, -0.00019444, 0.00027734, 0.00027976, -0.00048690, -0.00109612, -0.00044567, 0.00111775, 0.00195541, 0.00111775, -0.00044567, -0.00109612, -0.00048690, 0.00027976, 0.00027734, -0.00019444, -0.00034321, -0.00004421, 0.00019098, 0.00008236, -0.00010940, -0.00010062, 0.00002107, 0.00004335, -0.00002274, -0.00002905, 0.00003920, 0.00006699, 0.00001282, -0.00004498, -0.00004724, -0.00002344, -0.00001442, -0.00001112, 0.00000609, 0.00002391, 0.00002081, 0.00000303, -0.00000744, -0.00000451, },
		{ -0.00000335, -0.00000253, 0.00000072, -0.00000479, -0.00001245, -0.00000398, 0.00001962, 0.00003013, 0.00001226, -0.00000875, -0.00000949, -0.00001385, -0.00004590, -0.00006143, -0.00000563, 0.00006963, 0.00005475, -0.00003209, -0.00004461, 0.00004463, 0.00006022, -0.00007984, -0.00015608, 0.00002476, 0.00022542, 0.00006298, -0.00031808, -0.00032016, 0.00016996, 0.00039836, -0.00022804, -0.00104762, -0.00078472, 0.00067761, 0.00187220, 0.00147747, -0.00002976, -0.00101503, -0.00070601, 0.00009631, 0.00031927, -0.00005819, -0.00031534, -0.00013153, 0.00012813, 0.00011122, -0.00005699, -0.00009896, -0.00001164, 0.00003032, -0.00000959, -0.00002231, 0.00002893, 0.00006167, 0.00002567, -0.00002841, -0.00004357, -0.00002969, -0.00001958, -0.00001394, 0.00000082, 0.00001772, 0.00001959, 0.00000703, },
		{ 0.00001744, 0.00000866, -0.00000145, -0.00000090, 0.00000423, -0.00000189, -0.00001635, -0.00001350, 0.00001478, 0.00003522, 0.00001940, -0.00000674, -0.00000593, -0.00000186, -0.00003806, -0.00007511, -0.00002986, 0.00006626, 0.00007460, -0.00002679, -0.00007107, 0.00002955, 0.00009625, -0.00003678, -0.00018363, -0.00005155, 0.00022056, 0.00016851, -0.00024158, -0.00040829, 0.00001994, 0.00043428, 0.00002487, -0.00088773, -0.00100700, 0.00021679, 0.00163608, 0.00170912, 0.00040819, -0.00080914, -0.00084494, -0.00011941, 0.00028721, 0.00005940, -0.00024523, -0.00018343, 0.00005369, 0.00010936, -0.00001202, -0.00008044, -0.00003173, 0.00001241, -0.00000585, -0.00001632, 0.00002276, 0.00005600, 0.00003418, -0.00001328, -0.00003663, -0.00003245, -0.00002398, -0.00001711, -0.00000388, 0.00001216, },
		{ -0.00000846, 0.00000720, 0.00001551, 0.00000930, -0.00000185, -0.00000149, 0.00000807, 0.00000448, -0.00001697, -0.00002397, 0.00000575, 0.00003781, 0.00002721, -0.00000481, -0.00000477, 0.00001074, -0.00002310, -0.00008264, -0.00005829, 0.00005362, 0.00009529, -0.00001006, -0.00009531, -0.00000239, 0.00011878, 0.00002202, -0.00018232, -0.00013083, 0.00017469, 0.00025044, -0.00012742, -0.00044056, -0.00014143, 0.00038770, 0.00023078, -0.00065282, -0.00109470, -0.00020465, 0.00128462, 0.00178643, 0.00080910, -0.00050588, -0.00087757, -0.00032721, 0.00018917, 0.00013432, -0.00015301, -0.00019408, -0.00001437, 0.00008277, 0.00001620, -0.00005420, -0.00003771, -0.00000365, -0.00000911, -0.00001390, 0.00001833, 0.00005085, 0.00003982, -0.00000008, -0.00002794, -0.00003212, -0.00002691, -0.00002026, },
		{ -0.00002796, -0.00002288, -0.00001320, 0.00000228, 0.00001426, 0.00001052, -0.00000345, -0.00000526, 0.00000992, 0.00001327, -0.00001280, -0.00003307, -0.00000699, 0.00003663, 0.00003505, -0.00000253, -0.00000674, 0.00002151, -0.00000180, -0.00008065, -0.00008732, 0.00002978, 0.00011148, 0.00001820, -0.00010972, -0.00004690, 0.00011969, 0.00008507, -0.00014904, -0.00019565, 0.00009592, 0.00029203, 0.00000217, -0.00041169, -0.00028221, 0.00027567, 0.00036171, -0.00038873, -0.00105414, -0.00053776, 0.00087144, 0.00170899, 0.00112131, -0.00014973, -0.00079792, -0.00048839, 0.00004783, 0.00015362, -0.00006262, -0.00016836, -0.00006183, 0.00004305, 0.00002444, -0.00002979, -0.00003259, -0.00001332, -0.00001519, -0.00001559, 0.00001331, 0.00004577, 0.00004358, 0.00001129, -0.00001855, -0.00002936, },
		{ -0.00000905, -0.00002493, -0.00002700, -0.00002431, -0.00001803, -0.00000338, 0.00001327, 0.00001346, -0.00000439, -0.00001187, 0.00000775, 0.00002226, -0.00000351, -0.00003836, -0.00002196, 0.00003094, 0.00004200, 0.00000063, -0.00001190, 0.00002775, 0.00002341, -0.00006686, -0.00011180, -0.00000470, 0.00011728, 0.00005446, -0.00010815, -0.00009561, 0.00009558, 0.00013875, -0.00008848, -0.00023148, 0.00000022, 0.00028596, 0.00012217, -0.00033005, -0.00037724, 0.00012688, 0.00040749, -0.00014064, -0.00091234, -0.00075314, 0.00045500, 0.00150106, 0.00131036, 0.00020711, -0.00062109, -0.00057444, -0.00010518, 0.00011837, 0.00000457, -0.00011985, -0.00008184, 0.00000357, 0.00001581, -0.00001427, -0.00002223, -0.00001539, -0.00001992, -0.00001981, 0.00000642, 0.00003963, 0.00004569, 0.00002115, },
		{ 0.00004573, 0.00002956, 0.00000026, -0.00001948, -0.00002423, -0.00002391, -0.00002242, -0.00001034, 0.00001144, 0.00001827, -0.00000269, -0.00001965, 0.00000055, 0.00002867, 0.00000979, -0.00003797, -0.00003694, 0.00002071, 0.00004710, 0.00000519, -0.00001954, 0.00002717, 0.00004859, -0.00004098, -0.00012601, -0.00004623, 0.00010805, 0.00009207, -0.00008797, -0.00013818, 0.00004911, 0.00017093, -0.00001209, -0.00023052, -0.00009281, 0.00023583, 0.00021098, -0.00021484, -0.00041350, -0.00002594, 0.00037591, 0.00005620, -0.00070940, -0.00084449, 0.00008723, 0.00120510, 0.00136461, 0.00051554, -0.00037937, -0.00057314, -0.00023709, 0.00004273, 0.00003534, -0.00006637, -0.00007614, -0.00002464, -0.00000198, -0.00001018, -0.00001295, -0.00001186, -0.00002066, -0.00002392, -0.00000214, 0.00003145, },
		{ -0.00001088, 0.00002101, 0.00004301, 0.00003624, 0.00000916, -0.00001361, -0.00002010, -0.00002124, -0.00002541, -0.00001863, 0.00000735, 0.00002399, 0.00000302, -0.00002610, -0.00001125, 0.00003000, 0.00002479, -0.00003109, -0.00004944, 0.00000673, 0.00004943, 0.00001147, -0.00002810, 0.00001856, 0.00006896, -0.00000534, -0.00012518, -0.00008850, 0.00008209, 0.00012280, -0.00005119, -0.00016506, -0.00001154, 0.00017427, 0.00006489, -0.00019359, -0.00016490, 0.00015479, 0.00025543, -0.00009064, -0.00039193, -0.00015390, 0.00028871, 0.00018227, -0.00048871, -0.00082682, -0.00019554, 0.00087191, 0.00129547, 0.00073911, -0.00011452, -0.00049046, -0.00032187, -0.00005036, 0.00002753, -0.00002461, -0.00005369, -0.00003606, -0.00001964, -0.00001539, -0.00000919, -0.00000672, -0.00001715, -0.00002540, },
		{ -0.00001132, -0.00002303, -0.00001774, 0.00000910, 0.00003713, 0.00004067, 0.00001737, -0.00000783, -0.00001531, -0.00001623, -0.00002586, -0.00002746, -0.00000007, 0.00002876, 0.00001287, -0.00002855, -0.00002571, 0.00002474, 0.00003846, -0.00001826, -0.00005726, -0.00000955, 0.00004830, 0.00001949, -0.00003548, 0.00000228, 0.00007995, 0.00003520, -0.00010692, -0.00012373, 0.00004162, 0.00013904, -0.00000422, -0.00017015, -0.00007472, 0.00014803, 0.00012734, -0.00012962, -0.00020362, 0.00006169, 0.00025328, 0.00001899, -0.00032555, -0.00023801, 0.00017463, 0.00023543, -0.00028773, -0.00073023, -0.00037721, 0.00054992, 0.00113268, 0.00086003, 0.00013175, -0.00034795, -0.00034597, -0.00013568, -0.00001017, -0.00000539, -0.00002719, -0.00003175, -0.00002967, -0.00002436, -0.00001207, -0.00000403, },
		{ -0.00001923, -0.00000625, -0.00000630, -0.00001732, -0.00002105, -0.00000270, 0.00002830, 0.00004235, 0.00002453, -0.00000255, -0.00001071, -0.00000934, -0.00002279, -0.00003534, -0.00001110, 0.00003031, 0.00002560, -0.00002506, -0.00003989, 0.00001287, 0.00004785, -0.00000129, -0.00005893, -0.00002627, 0.00004333, 0.00002877, -0.00003940, -0.00001954, 0.00007827, 0.00007400, -0.00007226, -0.00014450, -0.00000731, 0.00013607, 0.00004364, -0.00015251, -0.00012840, 0.00009818, 0.00016408, -0.00005302, -0.00020501, -0.00002401, 0.00021263, 0.00009698, -0.00023478, -0.00027205, 0.00006176, 0.00022811, -0.00013139, -0.00059118, -0.00046160, 0.00027633, 0.00091598, 0.00088052, 0.00032647, -0.00017658, -0.00031063, -0.00019242, -0.00006158, -0.00001092, -0.00000862, -0.00001833, -0.00002877, -0.00003060, },
		{ -0.00001879, -0.00002935, -0.00002601, -0.00001314, -0.00000492, -0.00001036, -0.00002025, -0.00001263, 0.00001746, 0.00004108, 0.00003030, 0.00000192, -0.00000720, -0.00000159, -0.00001574, -0.00004034, -0.00002490, 0.00002671, 0.00003876, -0.00001501, -0.00005052, -0.00000404, 0.00005076, 0.00001717, -0.00005403, -0.00004144, 0.00003455, 0.00003832, -0.00003800, -0.00004329, 0.00006293, 0.00010402, -0.00002584, -0.00014572, -0.00005599, 0.00011351, 0.00008268, -0.00011659, -0.00016323, 0.00003557, 0.00017043, 0.00002030, -0.00017381, -0.00008706, 0.00014868, 0.00013596, -0.00014144, -0.00026178, -0.00002874, 0.00018158, -0.00002951, -0.00044382, -0.00046767, 0.00007193, 0.00068568, 0.00081932, 0.00045122, -0.00000873, -0.00023007, -0.00020933, -0.00010830, -0.00003492, -0.00000540, -0.00000516, },
		{ -0.00001814, -0.00000067, -0.00000569, -0.00001942, -0.00002754, -0.00002190, -0.00000845, -0.00000481, -0.00001620, -0.00001941, 0.00000595, 0.00003712, 0.00003453, 0.00000548, -0.00000551, 0.00000560, -0.00000507, -0.00004060, -0.00003949, 0.00001703, 0.00004926, 0.00000056, -0.00005492, -0.00002319, 0.00004633, 0.00003423, -0.00004319, -0.00005327, 0.00002239, 0.00004668, -0.00003040, -0.00006456, 0.00003570, 0.00011943, 0.00002483, -0.00012613, -0.00009514, 0.00007568, 0.00010539, -0.00007088, -0.00017488, -0.00002720, 0.00014879, 0.00007701, -0.00012130, -0.00011962, 0.00007902, 0.00013847, -0.00006319, -0.00022121, -0.00008614, 0.00011899, 0.00002166, -0.00031340, -0.00042226, -0.00005953, 0.00047455, 0.00070506, 0.00050378, 0.00012930, -0.00012658, -0.00018671, -0.00013529, -0.00006533, },
		{ -0.00013506, -0.00008862, -0.00004086, -0.00000925, 0.00000293, -0.00000377, -0.00002076, -0.00002821, -0.00001589, -0.00000284, -0.00001078, -0.00002271, -0.00000490, 0.00003113, 0.00003731, 0.00000834, -0.00000601, 0.00001062, 0.00000792, -0.00003495, -0.00005217, 0.00000176, 0.00005426, 0.00001916, -0.00005169, -0.00004125, 0.00003515, 0.00004734, -0.00002791, -0.00006043, 0.00000774, 0.00005219, -0.00001705, -0.00007914, 0.00000088, 0.00011706, 0.00007095, -0.00008889, -0.00011728, 0.00003060, 0.00010851, -0.00002553, -0.00016473, -0.00007905, 0.00010731, 0.00010931, -0.00006166, -0.00012206, 0.00001890, 0.00011460, -0.00001001, -0.00016739, -0.00011026, 0.00005946, 0.00003493, -0.00021322, -0.00035240, -0.00012783, 0.00030280, 0.00056823, 0.00049573, 0.00022240, -0.00002391, -0.00013510, },
		{ 0.00005895, -0.00007120, -0.00010931, -0.00009446, -0.00006340, -0.00002961, 0.00000080, 0.00001159, -0.00000591, -0.00002804, -0.00002424, -0.00000508, -0.00000615, -0.00002327, -0.00001423, 0.00002389, 0.00003893, 0.00001098, -0.00000851, 0.00001211, 0.00002117, -0.00002327, -0.00006013, -0.00001713, 0.00005189, 0.00003740, -0.00004106, -0.00005513, 0.00001906, 0.00005479, -0.00001020, -0.00006212, -0.00000809, 0.00005335, 0.00000019, -0.00008403, -0.00003553, 0.00009727, 0.00010445, -0.00004100, -0.00011867, -0.00001211, 0.00009375, 0.00001057, -0.00013895, -0.00011309, 0.00005708, 0.00011613, -0.00000797, -0.00010152, -0.00002220, 0.00007784, 0.00001662, -0.00011527, -0.00010872, 0.00001443, 0.00002636, -0.00014524, -0.00027928, -0.00015045, 0.00017706, 0.00043418, 0.00044695, 0.00026820, },
		{ 0.00037905, 0.00027507, 0.00011198, -0.00001229, -0.00006756, -0.00007901, -0.00007512, -0.00005522, -0.00001444, 0.00001982, 0.00001327, -0.00001937, -0.00002957, -0.00001027, -0.00000384, -0.00002244, -0.00002189, 0.00001601, 0.00003974, 0.00001409, -0.00001222, 0.00000928, 0.00003224, -0.00000675, -0.00006120, -0.00003657, 0.00004185, 0.00005185, -0.00002486, -0.00006271, 0.00000067, 0.00005595, 0.00000777, -0.00005817, -0.00002344, 0.00004918, 0.00001840, -0.00007822, -0.00006717, 0.00006392, 0.00011985, 0.00000829, -0.00010030, -0.00004381, 0.00006713, 0.00003205, -0.00010629, -0.00012781, 0.00000882, 0.00010220, 0.00003092, -0.00006874, -0.00004182, 0.00004080, 0.00002203, -0.00007430, -0.00009264, -0.00001297, 0.00000999, -0.00010351, -0.00021530, -0.00014605, 0.00009301, 0.00031879, },
		{ 0.00004013, 0.00022759, 0.00030954, 0.00025748, 0.00013518, 0.00002881, -0.00002359, -0.00004600, -0.00006880, -0.00007654, -0.00004009, 0.00001585, 0.00003071, -0.00000319, -0.00002856, -0.00001563, -0.00000401, -0.00002161, -0.00002839, 0.00000765, 0.00003988, 0.00001833, -0.00001583, 0.00000226, 0.00003887, 0.00001225, -0.00005440, -0.00005308, 0.00002551, 0.00005986, -0.00000602, -0.00006323, -0.00001726, 0.00005119, 0.00002403, -0.00004901, -0.00003648, 0.00003967, 0.00003422, -0.00006300, -0.00008879, 0.00002354, 0.00011558, 0.00004975, -0.00006760, -0.00005906, 0.00003704, 0.00003824, -0.00007524, -0.00012654, -0.00002984, 0.00007572, 0.00005182, -0.00003442, -0.00004351, 0.00001197, 0.00001506, -0.00004737, -0.00007229, -0.00002590, -0.00000537, -0.00007888, -0.00016429, -0.00012947, },
		{ -0.00012414, -0.00010920, 0.00000698, 0.00015792, 0.00024843, 0.00023055, 0.00013663, 0.00004700, 0.00000934, -0.00000528, -0.00004321, -0.00008442, -0.00006907, -0.00000157, 0.00004007, 0.00001648, -0.00001991, -0.00001787, -0.00000531, -0.00002143, -0.00003446, -0.00000148, 0.00003904, 0.00002402, -0.00001781, -0.00000782, 0.00003951, 0.00003070, -0.00004031, -0.00006361, 0.00000564, 0.00006017, 0.00001218, -0.00005738, -0.00003246, 0.00004169, 0.00003700, -0.00003557, -0.00004541, 0.00002587, 0.00004465, -0.00004165, -0.00009746, -0.00001628, 0.00009422, 0.00007636, -0.00002893, -0.00005697, 0.00001171, 0.00003280, -0.00005166, -0.00011555, -0.00005579, 0.00004533, 0.00005671, -0.00000629, -0.00003428, -0.00000558, 0.00000431, -0.00003197, -0.00005415, -0.00003039, -0.00001684, -0.00006331, },
		{ -0.00002587, -0.00005241, -0.00009066, -0.00008767, -0.00001479, 0.00010303, 0.00019782, 0.00020532, 0.00012834, 0.00004543, 0.00002271, 0.00003061, -0.00000373, -0.00007336, -0.00009212, -0.00002902, 0.00003702, 0.00003383, -0.00000508, -0.00001456, -0.00000537, -0.00002142, -0.00004053, -0.00001190, 0.00003650, 0.00003089, -0.00001680, -0.00001907, 0.00003376, 0.00004550, -0.00002097, -0.00006623, -0.00001428, 0.00005318, 0.00002685, -0.00004692, -0.00004351, 0.00002902, 0.00004564, -0.00001929, -0.00004882, 0.00000990, 0.00004784, -0.00001868, -0.00009317, -0.00004865, 0.00006174, 0.00008494, 0.00000672, -0.00004114, -0.00000315, 0.00002204, -0.00003742, -0.00010150, -0.00007017, 0.00001753, 0.00005082, 0.00001234, -0.00002150, -0.00001356, -0.00000454, -0.00002299, -0.00003997, -0.00003189, },
		{ -0.00002783, -0.00003291, -0.00003548, -0.00004580, -0.00006089, -0.00006361, -0.00002866, 0.00005634, 0.00015404, 0.00018629, 0.00012157, 0.00003333, 0.00001518, 0.00005061, 0.00003928, -0.00004364, -0.00010092, -0.00005930, 0.00002093, 0.00004312, 0.00001196, -0.00000519, -0.00000175, -0.00002004, -0.00004624, -0.00002395, 0.00003120, 0.00003795, -0.00001207, -0.00002917, 0.00002251, 0.00005425, 0.00000051, -0.00006056, -0.00003085, 0.00004075, 0.00003617, -0.00003420, -0.00004993, 0.00001480, 0.00004940, -0.00000190, -0.00004598, -0.00000547, 0.00004360, 0.00000136, -0.00007872, -0.00006903, 0.00002584, 0.00007667, 0.00003205, -0.00001835, -0.00000586, 0.00001255, -0.00003056, -0.00008914, -0.00007686, -0.00000469, 0.00003999, 0.00002226, -0.00001032, -0.00001628, -0.00000988, -0.00001562, },
		{ -0.00001345, -0.00000759, -0.00001448, -0.00003259, -0.00004749, -0.00004560, -0.00003489, -0.00003519, -0.00003392, 0.00001461, 0.00011135, 0.00017153, 0.00012294, 0.00002212, -0.00000746, 0.00004875, 0.00007382, -0.00000149, -0.00009090, -0.00008359, -0.00000455, 0.00004077, 0.00002589, 0.00000831, 0.00000698, -0.00001518, -0.00005024, -0.00003726, 0.00002224, 0.00004362, -0.00000380, -0.00003595, 0.00000777, 0.00005569, 0.00002073, -0.00004787, -0.00004151, 0.00002567, 0.00003964, -0.00002150, -0.00005212, 0.00000044, 0.00004820, 0.00001464, -0.00003706, -0.00001737, 0.00003349, 0.00001498, -0.00005880, -0.00007608, -0.00000610, 0.00005645, 0.00004330, 0.00000366, 0.00000117, 0.00000895, -0.00002672, -0.00008006, -0.00008016, -0.00002171, 0.00002848, 0.00002669, -0.00000254, -0.00001804, },
		{ 0.00000305, -0.00002114, -0.00001840, 0.00000027, 0.00000225, -0.00002788, -0.00006091, -0.00005396, -0.00001562, -0.00000265, -0.00002797, -0.00002117, 0.00006562, 0.00015513, 0.00013280, 0.00002116, -0.00003464, 0.00002600, 0.00009004, 0.00004265, -0.00006300, -0.00009427, -0.00003249, 0.00002668, 0.00003183, 0.00002201, 0.00002033, -0.00000512, -0.00005048, -0.00005053, 0.00000927, 0.00004604, 0.00000680, -0.00003799, -0.00000779, 0.00005005, 0.00003668, -0.00003063, -0.00004508, 0.00001093, 0.00003801, -0.00001053, -0.00005094, -0.00001299, 0.00004233, 0.00002837, -0.00002329, -0.00002347, 0.00002048, 0.00002064, -0.00003865, -0.00007168, -0.00002877, 0.00003124, 0.00004089, 0.00001853, 0.00001258, 0.00001242, -0.00002130, -0.00007273, -0.00008283, -0.00003592, 0.00001777, 0.00002908, },
		{ 0.00000688, 0.00003128, 0.00000905, -0.00002508, -0.00002711, 0.00000488, 0.00002184, -0.00001569, -0.00007204, -0.00007085, -0.00000715, 0.00003060, -0.00000921, -0.00004680, 0.00001670, 0.00013071, 0.00014586, 0.00003461, -0.00005481, -0.00001021, 0.00008335, 0.00007733, -0.00002362, -0.00008744, -0.00005477, 0.00000456, 0.00002704, 0.00003110, 0.00003576, 0.00001057, -0.00004483, -0.00006154, -0.00000701, 0.00004365, 0.00001771, -0.00003493, -0.00002148, 0.00003885, 0.00004637, -0.00001196, -0.00004184, -0.00000098, 0.00003292, -0.00000217, -0.00004746, -0.00002471, 0.00003249, 0.00003760, -0.00000679, -0.00002262, 0.00000816, 0.00001901, -0.00002267, -0.00006010, -0.00004012, 0.00000797, 0.00002888, 0.00002307, 0.00002214, 0.00002062, -0.00001150, -0.00006388, -0.00008504, -0.00004974, },
		{ -0.00008455, -0.00006392, -0.00000633, 0.00003274, 0.00001753, -0.00002725, -0.00003959, 0.00000287, 0.00004101, 0.00000498, -0.00007593, -0.00009302, -0.00001230, 0.00005871, 0.00002085, -0.00005749, -0.00003123, 0.00009474, 0.00015393, 0.00006021, -0.00005934, -0.00004848, 0.00005591, 0.00009358, 0.00001724, -0.00006437, -0.00006462, -0.00001906, 0.00001207, 0.00003155, 0.00004914, 0.00003044, -0.00003193, -0.00006756, -0.00002477, 0.00003573, 0.00002653, -0.00002767, -0.00003123, 0.00002453, 0.00004921, 0.00000516, -0.00003332, -0.00000852, 0.00002638, 0.00000352, -0.00004254, -0.00003419, 0.00001968, 0.00004116, 0.00000974, -0.00001512, -0.00000016, 0.00001270, -0.00001329, -0.00004651, -0.00004143, -0.00000837, 0.00001333, 0.00001796, 0.00002503, 0.00002882, 0.00000245, -0.00005044, },
		{ 0.00001730, -0.00003136, -0.00007812, -0.00007680, -0.00002330, 0.00003104, 0.00002850, -0.00002458, -0.00005322, -0.00000738, 0.00005492, 0.00003194, -0.00006868, -0.00011459, -0.00003073, 0.00007545, 0.00005700, -0.00005059, -0.00007121, 0.00004847, 0.00014942, 0.00009053, -0.00004534, -0.00007721, 0.00001580, 0.00008773, 0.00004896, -0.00003124, -0.00005887, -0.00003685, -0.00000915, 0.00002154, 0.00005596, 0.00005121, -0.00001201, -0.00006613, -0.00004114, 0.00002280, 0.00003116, -0.00001812, -0.00003607, 0.00000975, 0.00004593, 0.00001847, -0.00002179, -0.00001133, 0.00002016, 0.00000708, -0.00003681, -0.00004104, 0.00000525, 0.00003868, 0.00002351, -0.00000278, -0.00000235, 0.00000539, -0.00001044, -0.00003538, -0.00003651, -0.00001592, 0.00000023, 0.00000717, 0.00001944, 0.00003190, },
		{ 0.00000719, 0.00002648, 0.00002806, -0.00000857, -0.00006338, -0.00008482, -0.00004335, 0.00002342, 0.00003958, -0.00001536, -0.00006373, -0.00002460, 0.00005933, 0.00006011, -0.00004932, -0.00012891, -0.00005824, 0.00007665, 0.00009153, -0.00002750, -0.00009635, -0.00000220, 0.00012851, 0.00011575, -0.00001629, -0.00008824, -0.00002552, 0.00006241, 0.00006360, 0.00000271, -0.00003908, -0.00004301, -0.00003043, 0.00000235, 0.00005283, 0.00006839, 0.00001280, -0.00005589, -0.00005290, 0.00000674, 0.00003036, -0.00000870, -0.00003626, -0.00000321, 0.00003824, 0.00002683, -0.00000963, -0.00001005, 0.00001548, 0.00000932, -0.00003055, -0.00004497, -0.00000930, 0.00003063, 0.00003220, 0.00001153, 0.00000199, 0.00000068, -0.00001192, -0.00002924, -0.00003010, -0.00001608, -0.00000640, -0.00000371, },
		{ -0.00000569, -0.00000952, -0.00000708, 0.00001238, 0.00003002, 0.00001334, -0.00004047, -0.00008404, -0.00006326, 0.00000856, 0.00004682, -0.00000058, -0.00006705, -0.00004464, 0.00005249, 0.00008323, -0.00002071, -0.00013097, -0.00008779, 0.00006213, 0.00011658, 0.00000612, -0.00010262, -0.00004870, 0.00009268, 0.00012734, 0.00001930, -0.00007926, -0.00005684, 0.00002573, 0.00005823, 0.00002797, -0.00001115, -0.00003517, -0.00004513, -0.00002177, 0.00003868, 0.00007750, 0.00003860, -0.00003724, -0.00005720, -0.00000949, 0.00002425, -0.00000167, -0.00003315, -0.00001298, 0.00002826, 0.00003024, 0.00000120, -0.00000596, 0.00001283, 0.00001098, -0.00002393, -0.00004583, -0.00002239, 0.00001839, 0.00003448, 0.00002450, 0.00001143, 0.00000105, -0.00001420, -0.00002806, -0.00002627, -0.00001266, },
		{ -0.00002701, -0.00001018, 0.00000008, -0.00000742, -0.00001761, -0.00000712, 0.00002078, 0.00002855, -0.00001271, -0.00007206, -0.00007812, -0.00001218, 0.00004647, 0.00001612, -0.00006128, -0.00006189, 0.00003620, 0.00009607, 0.00001120, -0.00011935, -0.00011168, 0.00003600, 0.00012678, 0.00004180, -0.00009061, -0.00008270, 0.00004837, 0.00012095, 0.00005094, -0.00005429, -0.00007061, -0.00001139, 0.00003581, 0.00003771, 0.00001647, -0.00001525, -0.00004830, -0.00004474, 0.00001538, 0.00007544, 0.00006058, -0.00001259, -0.00005245, -0.00002252, 0.00001433, 0.00000150, -0.00002869, -0.00001926, 0.00001795, 0.00002957, 0.00000951, -0.00000057, 0.00001206, 0.00001255, -0.00001706, -0.00004370, -0.00003267, 0.00000394, 0.00003029, 0.00003322, 0.00002312, 0.00000711, -0.00001376, -0.00002957, },
		{ -0.00000825, -0.00003019, -0.00003170, -0.00001221, 0.00000653, 0.00000215, -0.00001967, -0.00002620, 0.00000146, 0.00003220, 0.00001399, -0.00004959, -0.00008304, -0.00003453, 0.00003685, 0.00002957, -0.00004775, -0.00007140, 0.00001542, 0.00009653, 0.00003926, -0.00009686, -0.00012408, 0.00000557, 0.00012108, 0.00007090, -0.00006562, -0.00009904, 0.00000480, 0.00009787, 0.00006970, -0.00002207, -0.00006503, -0.00003858, 0.00000414, 0.00002985, 0.00003541, 0.00001103, -0.00003826, -0.00006039, -0.00001253, 0.00006155, 0.00007417, 0.00001397, -0.00003893, -0.00002940, 0.00000318, 0.00000054, -0.00002482, -0.00002275, 0.00000864, 0.00002615, 0.00001493, 0.00000487, 0.00001261, 0.00001416, -0.00001016, -0.00003893, -0.00003919, -0.00001045, 0.00002085, 0.00003588, 0.00003361, 0.00001744, },
		{ 0.00003981, 0.00002908, 0.00000264, -0.00002648, -0.00003747, -0.00002001, 0.00000887, 0.00001561, -0.00001135, -0.00003848, -0.00002319, 0.00002234, 0.00003290, -0.00002086, -0.00007527, -0.00005253, 0.00001955, 0.00003514, -0.00003082, -0.00007087, -0.00000345, 0.00008646, 0.00005767, -0.00006971, -0.00012309, -0.00002113, 0.00010311, 0.00008731, -0.00003600, -0.00009747, -0.00002911, 0.00006443, 0.00007102, 0.00000695, -0.00004435, -0.00004915, -0.00002663, 0.00000764, 0.00003996, 0.00003589, -0.00001731, -0.00006436, -0.00003899, 0.00003800, 0.00007636, 0.00003756, -0.00001890, -0.00002841, -0.00000613, -0.00000353, -0.00002287, -0.00002468, 0.00000085, 0.00002128, 0.00001773, 0.00000952, 0.00001380, 0.00001571, -0.00000356, -0.00003218, -0.00004162, -0.00002272, 0.00000833, 0.00003215, },
		{ -0.00000466, 0.00002328, 0.00003980, 0.00003838, 0.00001710, -0.00001636, -0.00004028, -0.00003203, 0.00000377, 0.00002754, 0.00000565, -0.00003927, -0.00004617, 0.00000104, 0.00003876, 0.00000725, -0.00005567, -0.00006050, -0.00000061, 0.00003072, -0.00001644, -0.00006172, -0.00001488, 0.00007115, 0.00006401, -0.00004532, -0.00011130, -0.00003798, 0.00007985, 0.00008935, -0.00001047, -0.00008262, -0.00004764, 0.00002981, 0.00005599, 0.00002403, -0.00001718, -0.00004198, -0.00004721, -0.00002141, 0.00002886, 0.00005195, 0.00000901, -0.00005547, -0.00005816, 0.00000938, 0.00006657, 0.00005367, 0.00000378, -0.00001962, -0.00001094, -0.00000868, -0.00002316, -0.00002639, -0.00000565, 0.00001590, 0.00001855, 0.00001304, 0.00001507, 0.00001698, 0.00000239, -0.00002428, -0.00004029, -0.00003142, },
		{ -0.00003608, -0.00003597, -0.00001571, 0.00001168, 0.00003343, 0.00004214, 0.00003164, 0.00000001, -0.00003645, -0.00004442, -0.00000938, 0.00003255, 0.00002639, -0.00002727, -0.00006035, -0.00002584, 0.00002982, 0.00002726, -0.00002900, -0.00005529, -0.00001692, 0.00001797, -0.00000984, -0.00004874, -0.00001632, 0.00005741, 0.00006015, -0.00002969, -0.00009481, -0.00004286, 0.00005912, 0.00008010, 0.00000472, -0.00006231, -0.00004998, 0.00000287, 0.00003068, 0.00002483, 0.00000648, -0.00002154, -0.00005198, -0.00004801, 0.00000565, 0.00005445, 0.00003349, -0.00003614, -0.00006605, -0.00001844, 0.00004691, 0.00005938, 0.00002456, -0.00000495, -0.00000968, -0.00001246, -0.00002497, -0.00002872, -0.00001150, 0.00001041, 0.00001803, 0.00001545, 0.00001608, 0.00001780, 0.00000742, -0.00001612, },
		{ 0.00001141, -0.00000839, -0.00003018, -0.00003676, -0.00002322, 0.00000023, 0.00002228, 0.00003858, 0.00004220, 0.00002009, -0.00002400, -0.00005232, -0.00002807, 0.00002722, 0.00004428, -0.00000526, -0.00006093, -0.00005030, 0.00000881, 0.00003360, -0.00000280, -0.00003782, -0.00002316, 0.00000211, -0.00001340, -0.00003836, -0.00000934, 0.00005100, 0.00005155, -0.00002522, -0.00008095, -0.00003820, 0.00004689, 0.00006602, 0.00000762, -0.00004483, -0.00004015, -0.00001063, 0.00000389, 0.00001045, 0.00001848, 0.00000382, -0.00004053, -0.00006394, -0.00002250, 0.00004279, 0.00004954, -0.00001167, -0.00006170, -0.00003997, 0.00002175, 0.00005408, 0.00003930, 0.00001220, -0.00000240, -0.00001276, -0.00002676, -0.00003178, -0.00001740, 0.00000473, 0.00001660, 0.00001698, 0.00001669, 0.00001808, },
		},
		{ { -0.00002711, -0.00002614, -0.00002217, -0.00001319, 0.00000224, 0.00002414, 0.00004741, 0.00005696, 0.00003350, -0.00002425, -0.00008252, -0.00009167, -0.00003648, 0.00003605, 0.00005573, 0.00000400, -0.00006049, -0.00006415, -0.00000247, 0.00005436, 0.00004429, -0.00001394, -0.00004674, -0.00001698, 0.00003357, 0.00004113, 0.00000017, -0.00003493, -0.00002341, 0.00001247, 0.00002389, 0.00000273, -0.00001277, 0.00000273, 0.00002389, 0.00001247, -0.00002341, -0.00003493, 0.00000017, 0.00004113, 0.00003357, -0.00001698, -0.00004674, -0.00001394, 0.00004429, 0.00005436, -0.00000247, -0.00006415, -0.00006049, 0.00000400, 0.00005573, 0.00003605, -0.00003648, -0.00009167, -0.00008252, -0.00002425, 0.00003350, 0.00005696, 0.00004741, 0.00002414, 0.00000224, -0.00001319, -0.00002217, -0.00002614, },
		{ -0.00002236, -0.00003367, -0.00003400, -0.00002681, -0.00001657, -0.00000816, -0.00000404, 0.00000081, 0.00001599, 0.00004028, 0.00005115, 0.00002233, -0.00004003, -0.00008750, -0.00007242, -0.00000454, 0.00004916, 0.00003400, -0.00002907, -0.00006507, -0.00003188, 0.00003056, 0.00004991, 0.00000922, -0.00003577, -0.00002952, 0.00001496, 0.00003786, 0.00001206, -0.00002515, -0.00002595, 0.00000465, 0.00002086, 0.00000271, -0.00001722, -0.00000349, 0.00002769, 0.00002840, -0.00001165, -0.00004366, -0.00002152, 0.00003313, 0.00005272, 0.00000963, -0.00004466, -0.00004136, 0.00001856, 0.00006260, 0.00003417, -0.00003830, -0.00007444, -0.00003539, 0.00003373, 0.00005708, 0.00000984, -0.00006233, -0.00009630, -0.00007018, -0.00001091, 0.00004074, 0.00006234, 0.00005455, 0.00002943, 0.00000035, },
		{ 0.00005959, 0.00003196, -0.00000208, -0.00003017, -0.00004456, -0.00004194, -0.00002447, -0.00000287, 0.00000757, 0.00000185, -0.00000493, 0.00000737, 0.00003443, 0.00004258, 0.00000601, -0.00005318, -0.00007819, -0.00004033, 0.00002157, 0.00004041, -0.00000109, -0.00004886, -0.00004286, 0.00000931, 0.00004333, 0.00002008, -0.00002623, -0.00003469, 0.00000307, 0.00003488, 0.00001925, -0.00002046, -0.00003041, -0.00000020, 0.00002492, 0.00000904, -0.00002319, -0.00002101, 0.00001900, 0.00004340, 0.00001335, -0.00003823, -0.00004440, 0.00000833, 0.00005781, 0.00004216, -0.00002122, -0.00005529, -0.00001816, 0.00004511, 0.00005730, 0.00000149, -0.00006053, -0.00006162, -0.00000397, 0.00004835, 0.00004127, -0.00001772, -0.00007623, -0.00009044, -0.00005653, -0.00000094, 0.00004612, 0.00006715, },
		{ 0.00000686, 0.00005002, 0.00006961, 0.00006207, 0.00003414, -0.00000321, -0.00003814, -0.00005716, -0.00004901, -0.00001701, 0.00001520, 0.00002146, 0.00000209, -0.00001292, 0.00000096, 0.00002826, 0.00002813, -0.00001272, -0.00005546, -0.00005165, -0.00000494, 0.00002898, 0.00001026, -0.00003365, -0.00004219, -0.00000048, 0.00003968, 0.00002659, -0.00002290, -0.00004433, -0.00000987, 0.00003503, 0.00003186, -0.00001367, -0.00004019, -0.00001344, 0.00002811, 0.00002635, -0.00001643, -0.00003962, -0.00000690, 0.00004172, 0.00004062, -0.00001330, -0.00005244, -0.00002402, 0.00004058, 0.00006316, 0.00001597, -0.00004409, -0.00004600, 0.00000952, 0.00005367, 0.00003365, -0.00002735, -0.00006164, -0.00003513, 0.00001987, 0.00004550, 0.00001775, -0.00003812, -0.00007892, -0.00007975, -0.00004418, },
		{ -0.00007465, -0.00006771, -0.00003290, 0.00001295, 0.00005084, 0.00006884, 0.00006422, 0.00003848, -0.00000410, -0.00004934, -0.00007147, -0.00005159, -0.00000258, 0.00003443, 0.00002953, -0.00000252, -0.00001933, -0.00000307, 0.00001773, 0.00000739, -0.00002628, -0.00003965, -0.00001280, 0.00001929, 0.00001101, -0.00002957, -0.00004586, -0.00000730, 0.00004401, 0.00004247, -0.00001494, -0.00005935, -0.00003539, 0.00002711, 0.00004993, 0.00000653, -0.00004470, -0.00003737, 0.00001647, 0.00004383, 0.00000830, -0.00004184, -0.00003816, 0.00001664, 0.00005152, 0.00002032, -0.00003718, -0.00004588, 0.00000841, 0.00006073, 0.00004843, -0.00001252, -0.00005004, -0.00002448, 0.00002764, 0.00004266, 0.00000502, -0.00004057, -0.00004528, -0.00000838, 0.00002920, 0.00003095, -0.00000412, -0.00004898, },
		{ 0.00001337, -0.00001967, -0.00005272, -0.00006699, -0.00005473, -0.00002227, 0.00001565, 0.00004696, 0.00006666, 0.00006940, 0.00004491, -0.00000837, -0.00006535, -0.00008404, -0.00004596, 0.00001702, 0.00004960, 0.00003037, -0.00000795, -0.00002166, -0.00000822, -0.00000064, -0.00001468, -0.00002485, -0.00000485, 0.00002544, 0.00001965, -0.00002756, -0.00006015, -0.00002731, 0.00004345, 0.00006919, 0.00001290, -0.00006405, -0.00007201, -0.00000317, 0.00005800, 0.00003972, -0.00002899, -0.00005851, -0.00001386, 0.00004363, 0.00003897, -0.00001955, -0.00005423, -0.00002074, 0.00003591, 0.00004256, -0.00000682, -0.00004535, -0.00002087, 0.00003841, 0.00006139, 0.00002248, -0.00003095, -0.00004021, -0.00000364, 0.00002916, 0.00002058, -0.00001517, -0.00003625, -0.00002233, 0.00000880, 0.00002572, },
		{ -0.00000370, 0.00001444, 0.00001641, -0.00000127, -0.00002959, -0.00005261, -0.00005661, -0.00004014, -0.00001393, 0.00001232, 0.00003927, 0.00006721, 0.00007923, 0.00004987, -0.00001960, -0.00008399, -0.00008959, -0.00003147, 0.00003637, 0.00005665, 0.00002678, -0.00000958, -0.00002131, -0.00001987, -0.00002581, -0.00002816, -0.00000185, 0.00004061, 0.00004705, -0.00000734, -0.00007080, -0.00006425, 0.00001834, 0.00008819, 0.00006088, -0.00003855, -0.00009918, -0.00005497, 0.00003682, 0.00006751, 0.00000912, -0.00005740, -0.00004878, 0.00001802, 0.00005452, 0.00001671, -0.00004293, -0.00004858, 0.00000182, 0.00003997, 0.00001951, -0.00002653, -0.00003329, 0.00001132, 0.00005402, 0.00004531, -0.00000204, -0.00003455, -0.00002392, 0.00000632, 0.00001716, 0.00000013, -0.00002060, -0.00002157, },
		{ -0.00000961, -0.00001449, -0.00000709, 0.00000577, 0.00001344, 0.00000778, -0.00001232, -0.00003693, -0.00004945, -0.00004216, -0.00002500, -0.00001153, 0.00000211, 0.00003198, 0.00007426, 0.00009140, 0.00004810, -0.00003806, -0.00009953, -0.00008442, -0.00001209, 0.00004916, 0.00005571, 0.00002436, -0.00000571, -0.00002364, -0.00004112, -0.00005039, -0.00002244, 0.00004099, 0.00008107, 0.00004085, -0.00005246, -0.00009826, -0.00003567, 0.00007350, 0.00010576, 0.00001982, -0.00009066, -0.00010518, -0.00001681, 0.00006538, 0.00005126, -0.00002694, -0.00006548, -0.00002127, 0.00004241, 0.00004482, -0.00001102, -0.00005120, -0.00003022, 0.00001638, 0.00002685, -0.00000527, -0.00002890, -0.00000685, 0.00003706, 0.00005178, 0.00002264, -0.00001626, -0.00002779, -0.00001210, 0.00000359, 0.00000187, },
		{ -0.00001020, -0.00000496, -0.00000642, -0.00000819, -0.00000612, 0.00000025, 0.00000885, 0.00001253, 0.00000163, -0.00002286, -0.00004300, -0.00004112, -0.00002326, -0.00001331, -0.00001826, -0.00001196, 0.00003077, 0.00008804, 0.00010005, 0.00003615, -0.00005928, -0.00010558, -0.00006949, 0.00000494, 0.00005201, 0.00005151, 0.00002788, 0.00000035, -0.00003482, -0.00006905, -0.00006258, 0.00000768, 0.00009213, 0.00010029, 0.00000720, -0.00009666, -0.00009511, 0.00001620, 0.00011419, 0.00008629, -0.00003799, -0.00012175, -0.00008087, 0.00002399, 0.00006962, 0.00001792, -0.00005141, -0.00005082, 0.00000975, 0.00004852, 0.00002115, -0.00003072, -0.00004311, -0.00001121, 0.00001524, 0.00000440, -0.00001898, -0.00001354, 0.00002162, 0.00004788, 0.00003745, 0.00000402, -0.00001968, -0.00002026, },
		{ -0.00000587, -0.00001920, -0.00001960, -0.00001491, -0.00000899, -0.00000406, -0.00000312, -0.00000404, 0.00000107, 0.00001248, 0.00001490, -0.00000522, -0.00003550, -0.00004544, -0.00002507, -0.00000297, -0.00001045, -0.00003372, -0.00002345, 0.00003924, 0.00010391, 0.00009887, 0.00001544, -0.00007565, -0.00009915, -0.00005090, 0.00001297, 0.00004684, 0.00005072, 0.00003765, 0.00000162, -0.00005703, -0.00009395, -0.00005181, 0.00005756, 0.00013426, 0.00008969, -0.00004071, -0.00011989, -0.00006047, 0.00006971, 0.00012165, 0.00003665, -0.00008884, -0.00012033, -0.00003931, 0.00004876, 0.00004938, -0.00001592, -0.00005423, -0.00002259, 0.00002959, 0.00003601, -0.00000351, -0.00003510, -0.00002686, -0.00000282, 0.00000048, -0.00001455, -0.00001561, 0.00001079, 0.00004068, 0.00004440, 0.00002119, },
		{ 0.00004607, 0.00003458, 0.00001059, -0.00001059, -0.00002337, -0.00002592, -0.00001669, -0.00000205, 0.00000365, -0.00000428, -0.00001016, 0.00000229, 0.00002183, 0.00001780, -0.00001736, -0.00004890, -0.00003958, -0.00000184, 0.00001233, -0.00001965, -0.00005268, -0.00002564, 0.00005591, 0.00011413, 0.00008511, -0.00000739, -0.00008042, -0.00008307, -0.00003709, 0.00000985, 0.00004012, 0.00005759, 0.00004841, -0.00000823, -0.00008523, -0.00010289, -0.00001488, 0.00011317, 0.00015227, 0.00005368, -0.00008242, -0.00011148, -0.00000908, 0.00010104, 0.00009184, -0.00002258, -0.00011294, -0.00009052, -0.00000063, 0.00004824, 0.00001670, -0.00003489, -0.00003720, 0.00000422, 0.00003119, 0.00001430, -0.00001777, -0.00002683, -0.00001449, -0.00000863, -0.00001757, -0.00001996, 0.00000098, 0.00003190, },
		{ -0.00001050, 0.00001990, 0.00004162, 0.00004239, 0.00002755, 0.00000565, -0.00001856, -0.00003556, -0.00003126, -0.00000655, 0.00001351, 0.00000598, -0.00001681, -0.00001839, 0.00001164, 0.00003593, 0.00001475, -0.00003611, -0.00005741, -0.00002225, 0.00002328, 0.00001595, -0.00003898, -0.00006682, -0.00001560, 0.00007399, 0.00011201, 0.00006211, -0.00002346, -0.00007179, -0.00006520, -0.00003433, -0.00000055, 0.00003896, 0.00007068, 0.00005203, -0.00003003, -0.00010788, -0.00008614, 0.00004092, 0.00015699, 0.00014112, 0.00000828, -0.00010046, -0.00007647, 0.00003524, 0.00009605, 0.00003778, -0.00006722, -0.00010396, -0.00004912, 0.00001873, 0.00002680, -0.00001152, -0.00003310, -0.00001212, 0.00001758, 0.00001833, -0.00000391, -0.00001855, -0.00001618, -0.00001409, -0.00002252, -0.00002692, },
		{ -0.00002448, -0.00003256, -0.00002142, 0.00000588, 0.00003031, 0.00004049, 0.00003904, 0.00002693, -0.00000084, -0.00003618, -0.00005040, -0.00002461, 0.00001712, 0.00002716, -0.00000612, -0.00003706, -0.00001861, 0.00003167, 0.00004762, 0.00000052, -0.00005701, -0.00005424, 0.00000465, 0.00004174, 0.00000507, -0.00006113, -0.00006898, 0.00000334, 0.00008451, 0.00009604, 0.00003827, -0.00002638, -0.00005453, -0.00005452, -0.00004308, -0.00001048, 0.00004652, 0.00008299, 0.00004254, -0.00005678, -0.00011182, -0.00004280, 0.00010027, 0.00017650, 0.00010889, -0.00002688, -0.00008992, -0.00003390, 0.00005143, 0.00005770, -0.00001686, -0.00008152, -0.00007147, -0.00001617, 0.00001432, -0.00000104, -0.00002314, -0.00001661, 0.00000678, 0.00001503, 0.00000233, -0.00001068, -0.00001243, -0.00001391, },
		{ -0.00000783, -0.00001085, -0.00002380, -0.00003567, -0.00002875, -0.00000490, 0.00001703, 0.00002821, 0.00003676, 0.00004315, 0.00002800, -0.00001786, -0.00006199, -0.00005603, 0.00000051, 0.00004596, 0.00002619, -0.00003345, -0.00005400, -0.00000345, 0.00005693, 0.00004783, -0.00002421, -0.00007138, -0.00003624, 0.00003356, 0.00004610, -0.00001675, -0.00007650, -0.00005736, 0.00002270, 0.00008112, 0.00007140, 0.00002297, -0.00001602, -0.00003776, -0.00005617, -0.00005765, -0.00001305, 0.00005950, 0.00008580, 0.00002074, -0.00007589, -0.00008895, 0.00001785, 0.00014715, 0.00017075, 0.00007261, -0.00003925, -0.00006187, -0.00000730, 0.00003223, 0.00000468, -0.00004986, -0.00006561, -0.00003486, -0.00000412, -0.00000497, -0.00001963, -0.00001732, 0.00000115, 0.00001146, 0.00000434, -0.00000570, },
		{ 0.00000475, -0.00000218, -0.00000245, -0.00000557, -0.00002210, -0.00003947, -0.00003503, -0.00000974, 0.00001092, 0.00001434, 0.00001906, 0.00004050, 0.00005396, 0.00002048, -0.00004895, -0.00008545, -0.00004112, 0.00004041, 0.00006518, 0.00000361, -0.00006688, -0.00005542, 0.00002629, 0.00007634, 0.00003160, -0.00005186, -0.00007099, -0.00000690, 0.00005435, 0.00003431, -0.00004054, -0.00007819, -0.00003699, 0.00003304, 0.00006386, 0.00004767, 0.00002152, 0.00000115, -0.00003014, -0.00006870, -0.00006961, -0.00000660, 0.00006961, 0.00007412, -0.00000489, -0.00007507, -0.00004069, 0.00008088, 0.00017309, 0.00015060, 0.00004850, -0.00003109, -0.00003799, -0.00001212, -0.00001159, -0.00003824, -0.00005170, -0.00003439, -0.00001220, -0.00001159, -0.00002318, -0.00002151, -0.00000419, 0.00000808, },
		{ -0.00001161, 0.00000212, 0.00000137, -0.00000184, 0.00000376, 0.00000497, -0.00001570, -0.00004487, -0.00004747, -0.00001549, 0.00001535, 0.00001339, -0.00000146, 0.00001436, 0.00005624, 0.00006225, -0.00000440, -0.00008773, -0.00009186, -0.00000253, 0.00008130, 0.00006137, -0.00003671, -0.00009137, -0.00003535, 0.00006088, 0.00007926, 0.00000195, -0.00007128, -0.00005321, 0.00002430, 0.00005939, 0.00001149, -0.00005621, -0.00006577, -0.00001741, 0.00002906, 0.00003953, 0.00003386, 0.00003219, 0.00001546, -0.00003520, -0.00008499, -0.00007288, 0.00000394, 0.00006833, 0.00005031, -0.00002189, -0.00004857, 0.00002290, 0.00013378, 0.00018147, 0.00013181, 0.00004175, -0.00001910, -0.00003869, -0.00004662, -0.00005660, -0.00005338, -0.00003005, -0.00000841, -0.00001014, -0.00002531, -0.00002748, },
		{ -0.00002063, -0.00002897, -0.00001635, -0.00000388, -0.00000703, -0.00001045, 0.00000413, 0.00002005, 0.00000306, -0.00004312, -0.00006746, -0.00003524, 0.00001903, 0.00003153, -0.00000317, -0.00002016, 0.00002417, 0.00007883, 0.00005545, -0.00004689, -0.00011931, -0.00007059, 0.00005092, 0.00010540, 0.00003043, -0.00008075, -0.00009393, 0.00000181, 0.00008628, 0.00006173, -0.00003073, -0.00007360, -0.00002317, 0.00004631, 0.00004772, -0.00001241, -0.00005770, -0.00004566, -0.00000765, 0.00001255, 0.00001810, 0.00003361, 0.00004720, 0.00001925, -0.00004960, -0.00009659, -0.00006745, 0.00000972, 0.00005206, 0.00002342, -0.00002007, 0.00000026, 0.00008931, 0.00017346, 0.00018457, 0.00012429, 0.00004188, -0.00002550, -0.00007082, -0.00009172, -0.00007923, -0.00003961, -0.00000475, -0.00000180, },
		{ -0.00001296, 0.00000396, -0.00001384, -0.00002629, -0.00001393, -0.00000100, -0.00001222, -0.00002843, -0.00001249, 0.00002722, 0.00003418, -0.00001959, -0.00008105, -0.00007351, 0.00000076, 0.00005436, 0.00002633, -0.00003327, -0.00002767, 0.00005017, 0.00009518, 0.00002568, -0.00009619, -0.00012732, -0.00002217, 0.00010112, 0.00009990, -0.00001922, -0.00011087, -0.00007073, 0.00004274, 0.00009133, 0.00002923, -0.00005370, -0.00005716, 0.00000832, 0.00005199, 0.00002568, -0.00002759, -0.00004611, -0.00002767, -0.00001139, -0.00000860, 0.00000741, 0.00004337, 0.00005707, 0.00001097, -0.00006535, -0.00009876, -0.00005942, 0.00000323, 0.00002437, 0.00000440, 0.00000342, 0.00006211, 0.00015163, 0.00020758, 0.00019427, 0.00012355, 0.00002725, -0.00006456, -0.00012179, -0.00012055, -0.00006892, },
		{ -0.00016410, -0.00011301, -0.00003477, 0.00000231, -0.00001231, -0.00002918, -0.00001140, 0.00001441, 0.00000113, -0.00004094, -0.00004770, 0.00000711, 0.00005951, 0.00002979, -0.00006402, -0.00011276, -0.00005111, 0.00005137, 0.00007291, -0.00000272, -0.00006251, -0.00001444, 0.00008320, 0.00009134, -0.00002408, -0.00013427, -0.00010274, 0.00004012, 0.00012883, 0.00006403, -0.00006990, -0.00011453, -0.00002963, 0.00007165, 0.00007375, -0.00000577, -0.00005832, -0.00002854, 0.00003012, 0.00004164, 0.00000348, -0.00002944, -0.00002875, -0.00001958, -0.00002515, -0.00002482, 0.00000905, 0.00005460, 0.00005581, -0.00000447, -0.00007478, -0.00009304, -0.00005715, -0.00001740, -0.00000626, -0.00000001, 0.00004390, 0.00012913, 0.00021342, 0.00024818, 0.00021123, 0.00011007, -0.00002259, -0.00013178, },
		{ 0.00005830, -0.00011285, -0.00020221, -0.00016875, -0.00006798, -0.00000176, -0.00001234, -0.00004426, -0.00002660, 0.00002745, 0.00003840, -0.00002294, -0.00008037, -0.00004701, 0.00004869, 0.00008334, -0.00000355, -0.00011803, -0.00011959, -0.00000049, 0.00009799, 0.00006292, -0.00004580, -0.00007501, 0.00001838, 0.00010737, 0.00006081, -0.00007956, -0.00014439, -0.00004970, 0.00009540, 0.00012300, 0.00001008, -0.00010291, -0.00009062, 0.00001356, 0.00007808, 0.00004123, -0.00003000, -0.00004423, 0.00000087, 0.00003621, 0.00002237, -0.00001022, -0.00002049, -0.00001483, -0.00002306, -0.00004054, -0.00003015, 0.00001786, 0.00005883, 0.00004386, -0.00001909, -0.00007483, -0.00008580, -0.00006622, -0.00004792, -0.00003211, 0.00000968, 0.00009455, 0.00020219, 0.00028635, 0.00030059, 0.00021925, },
		{ 0.00038050, 0.00035358, 0.00018926, -0.00004823, -0.00022533, -0.00024002, -0.00012122, -0.00000621, 0.00000202, -0.00005823, -0.00006890, 0.00000824, 0.00007748, 0.00003747, -0.00007272, -0.00011058, -0.00001559, 0.00009869, 0.00008014, -0.00006295, -0.00016059, -0.00009081, 0.00006449, 0.00012127, 0.00002504, -0.00008524, -0.00006215, 0.00005855, 0.00010832, 0.00000939, -0.00012105, -0.00011982, 0.00001526, 0.00012464, 0.00008625, -0.00004215, -0.00010745, -0.00004983, 0.00004277, 0.00006179, 0.00000739, -0.00003586, -0.00001909, 0.00002069, 0.00002813, 0.00000381, -0.00001262, -0.00000833, -0.00001018, -0.00003319, -0.00004924, -0.00002519, 0.00002538, 0.00005216, 0.00002703, -0.00002675, -0.00006821, -0.00008374, -0.00008653, -0.00008193, -0.00005119, 0.00002818, 0.00015457, 0.00029256, },
		{ 0.00005300, 0.00023450, 0.00041552, 0.00049222, 0.00037757, 0.00009286, -0.00020322, -0.00032176, -0.00021542, -0.00003153, 0.00003814, -0.00003923, -0.00011680, -0.00006152, 0.00007310, 0.00011477, -0.00000002, -0.00013232, -0.00011133, 0.00004273, 0.00013499, 0.00004075, -0.00013146, -0.00017155, -0.00003017, 0.00012186, 0.00011060, -0.00002738, -0.00010431, -0.00002749, 0.00008862, 0.00008126, -0.00004642, -0.00013280, -0.00006783, 0.00007028, 0.00011945, 0.00003367, -0.00007465, -0.00008527, -0.00000913, 0.00004927, 0.00003223, -0.00001511, -0.00002451, 0.00000591, 0.00002648, 0.00001320, -0.00000721, -0.00000744, -0.00000071, -0.00001432, -0.00004187, -0.00004767, -0.00001610, 0.00002488, 0.00003677, 0.00001230, -0.00002635, -0.00006045, -0.00008950, -0.00011395, -0.00011664, -0.00006787, },
		{ -0.00014562, -0.00015620, -0.00008722, 0.00009241, 0.00035420, 0.00057771, 0.00059629, 0.00033381, -0.00008057, -0.00037320, -0.00035835, -0.00012097, 0.00006713, 0.00003372, -0.00011906, -0.00015748, -0.00000950, 0.00014794, 0.00011613, -0.00006892, -0.00017690, -0.00007320, 0.00010941, 0.00013752, -0.00002833, -0.00018482, -0.00014169, 0.00004475, 0.00015149, 0.00006936, -0.00007469, -0.00009565, 0.00001476, 0.00009484, 0.00003419, -0.00008691, -0.00011083, -0.00000681, 0.00009813, 0.00008561, -0.00001461, -0.00007930, -0.00004906, 0.00001662, 0.00003511, 0.00000310, -0.00002044, -0.00000420, 0.00002198, 0.00002090, -0.00000031, -0.00000983, -0.00000177, -0.00000102, -0.00002143, -0.00004281, -0.00003842, -0.00001038, 0.00001482, 0.00001841, 0.00000341, -0.00002111, -0.00005517, -0.00010083, },
		{ -0.00001444, -0.00005108, -0.00011446, -0.00018536, -0.00020913, -0.00010594, 0.00016823, 0.00053216, 0.00076461, 0.00064879, 0.00019203, -0.00031160, -0.00050627, -0.00030559, 0.00002122, 0.00013165, -0.00003195, -0.00020965, -0.00015246, 0.00007936, 0.00020326, 0.00007136, -0.00014786, -0.00018456, -0.00000373, 0.00015903, 0.00009866, -0.00010654, -0.00020240, -0.00007772, 0.00010998, 0.00014425, 0.00001324, -0.00010036, -0.00006464, 0.00004701, 0.00007437, -0.00001566, -0.00009854, -0.00006484, 0.00004228, 0.00009373, 0.00003962, -0.00004282, -0.00006067, -0.00001544, 0.00002130, 0.00001077, -0.00001461, -0.00001127, 0.00001495, 0.00002652, 0.00001081, -0.00000785, -0.00000846, -0.00000119, -0.00000678, -0.00002456, -0.00003507, -0.00002785, -0.00001225, -0.00000144, 0.00000187, -0.00000059, },
		{ -0.00001936, -0.00001235, -0.00000273, -0.00000548, -0.00004361, -0.00013246, -0.00024463, -0.00027979, -0.00010506, 0.00030938, 0.00077041, 0.00093583, 0.00060154, -0.00005123, -0.00055092, -0.00055133, -0.00016393, 0.00016569, 0.00012795, -0.00014472, -0.00027347, -0.00009093, 0.00017864, 0.00021381, -0.00001034, -0.00020831, -0.00014727, 0.00007486, 0.00017097, 0.00002772, -0.00016787, -0.00017670, 0.00000038, 0.00014572, 0.00010567, -0.00003757, -0.00009850, -0.00002614, 0.00005717, 0.00003657, -0.00005017, -0.00008028, -0.00001344, 0.00006542, 0.00006545, 0.00000023, -0.00004629, -0.00003233, 0.00000324, 0.00000913, -0.00001104, -0.00001734, 0.00000420, 0.00002627, 0.00002290, 0.00000245, -0.00001040, -0.00000889, -0.00000636, -0.00001207, -0.00002016, -0.00002275, -0.00002141, -0.00002084, },
		{ -0.00001133, -0.00002012, -0.00003233, -0.00003762, -0.00002734, -0.00000303, 0.00001195, -0.00003096, -0.00016666, -0.00033644, -0.00035798, -0.00005062, 0.00053721, 0.00104267, 0.00103300, 0.00042328, -0.00036292, -0.00073348, -0.00047991, 0.00003263, 0.00026109, 0.00004696, -0.00026838, -0.00028018, 0.00001455, 0.00025441, 0.00017080, -0.00010354, -0.00022754, -0.00007542, 0.00013531, 0.00013916, -0.00005257, -0.00019227, -0.00011665, 0.00006823, 0.00014463, 0.00005241, -0.00006775, -0.00007563, 0.00000351, 0.00004435, -0.00000205, -0.00005903, -0.00004339, 0.00002498, 0.00006113, 0.00002975, -0.00002040, -0.00003193, -0.00000911, 0.00000405, -0.00000915, -0.00002185, -0.00000901, 0.00001761, 0.00002900, 0.00001686, -0.00000218, -0.00001234, -0.00001388, -0.00001329, -0.00001204, -0.00000996, },
		{ -0.00001692, -0.00000648, 0.00000114, -0.00000312, -0.00002075, -0.00004460, -0.00005974, -0.00004524, 0.00000513, 0.00004495, -0.00002117, -0.00023660, -0.00046406, -0.00041333, 0.00009540, 0.00084857, 0.00129329, 0.00099929, 0.00011818, -0.00067849, -0.00080196, -0.00030028, 0.00022741, 0.00027124, -0.00009035, -0.00035966, -0.00022054, 0.00013269, 0.00028020, 0.00008639, -0.00017758, -0.00019895, 0.00000670, 0.00015749, 0.00007488, -0.00011612, -0.00017417, -0.00004308, 0.00010753, 0.00011365, 0.00000376, -0.00007311, -0.00004600, 0.00001468, 0.00001842, -0.00002635, -0.00004408, -0.00000515, 0.00004059, 0.00003952, 0.00000244, -0.00002127, -0.00001299, 0.00000017, -0.00000699, -0.00002283, -0.00002054, 0.00000290, 0.00002478, 0.00002669, 0.00001202, -0.00000543, -0.00001729, -0.00002139, },
		{ -0.00001080, -0.00002469, -0.00002660, -0.00001550, 0.00000036, 0.00000969, 0.00000468, -0.00001977, -0.00006099, -0.00009102, -0.00006161, 0.00003356, 0.00009346, -0.00003481, -0.00035994, -0.00061009, -0.00039972, 0.00035676, 0.00120547, 0.00145045, 0.00080323, -0.00026773, -0.00092134, -0.00072937, -0.00005712, 0.00036229, 0.00019009, -0.00024113, -0.00038647, -0.00011059, 0.00022680, 0.00024844, -0.00001172, -0.00020999, -0.00013464, 0.00007176, 0.00013704, 0.00000132, -0.00014465, -0.00012383, 0.00002082, 0.00011292, 0.00006911, -0.00002714, -0.00006060, -0.00002345, 0.00000840, -0.00000611, -0.00002988, -0.00001710, 0.00002000, 0.00003544, 0.00001561, -0.00000878, -0.00001095, -0.00000132, -0.00000414, -0.00001984, -0.00002654, -0.00001181, 0.00001240, 0.00002649, 0.00002317, 0.00000788, },
		{ 0.00002571, 0.00001907, 0.00000135, -0.00001885, -0.00003002, -0.00002636, -0.00001262, 0.00000357, 0.00001754, 0.00001757, -0.00001862, -0.00008969, -0.00013220, -0.00006322, 0.00009211, 0.00014308, -0.00009978, -0.00053882, -0.00073298, -0.00027068, 0.00072512, 0.00153829, 0.00145009, 0.00045635, -0.00065665, -0.00102934, -0.00053100, 0.00018392, 0.00039971, 0.00004384, -0.00035821, -0.00034079, 0.00001553, 0.00027024, 0.00017297, -0.00009265, -0.00019554, -0.00005920, 0.00010129, 0.00008616, -0.00005704, -0.00013461, -0.00006150, 0.00005931, 0.00009206, 0.00002805, -0.00003787, -0.00004273, -0.00001460, -0.00000580, -0.00001810, -0.00001701, 0.00000726, 0.00002691, 0.00002033, 0.00000101, -0.00000646, -0.00000116, -0.00000156, -0.00001486, -0.00002662, -0.00002143, -0.00000167, 0.00001789, },
		{ -0.00001226, 0.00000676, 0.00002116, 0.00002363, 0.00001158, -0.00000920, -0.00002539, -0.00002842, -0.00002296, -0.00001486, 0.00000277, 0.00003182, 0.00003810, -0.00002615, -0.00013829, -0.00017376, -0.00003218, 0.00017879, 0.00016443, -0.00023830, -0.00074901, -0.00077556, -0.00000061, 0.00115199, 0.00176396, 0.00126129, 0.00001582, -0.00096302, -0.00097737, -0.00025981, 0.00036002, 0.00033604, -0.00012044, -0.00040880, -0.00024064, 0.00012044, 0.00025656, 0.00008174, -0.00013534, -0.00014696, 0.00000179, 0.00009268, 0.00002691, -0.00008473, -0.00009682, -0.00000766, 0.00006914, 0.00005976, 0.00000071, -0.00003517, -0.00002999, -0.00001677, -0.00001646, -0.00001516, 0.00000102, 0.00001939, 0.00002060, 0.00000752, -0.00000135, 0.00000022, 0.00000024, -0.00001039, -0.00002360, -0.00002528, },
		{ -0.00002040, -0.00002601, -0.00001877, -0.00000280, 0.00001430, 0.00002375, 0.00001760, -0.00000234, -0.00002050, -0.00002368, -0.00001969, -0.00002411, -0.00002635, 0.00000483, 0.00006017, 0.00006054, -0.00005566, -0.00020586, -0.00019552, 0.00004416, 0.00027308, 0.00012174, -0.00045147, -0.00093915, -0.00068433, 0.00039742, 0.00155633, 0.00181334, 0.00090207, -0.00043059, -0.00112191, -0.00078619, 0.00001262, 0.00043329, 0.00020062, -0.00025407, -0.00038589, -0.00012064, 0.00017859, 0.00019995, 0.00000388, -0.00013604, -0.00008738, 0.00003263, 0.00005818, -0.00001937, -0.00008037, -0.00004959, 0.00002529, 0.00005804, 0.00002983, -0.00001219, -0.00002898, -0.00002573, -0.00002158, -0.00001697, -0.00000349, 0.00001328, 0.00001881, 0.00001138, 0.00000368, 0.00000307, 0.00000223, -0.00000710, },
		{ 0.00000568, -0.00000367, -0.00001756, -0.00002587, -0.00002324, -0.00001103, 0.00000704, 0.00002307, 0.00002354, 0.00000289, -0.00002152, -0.00002453, -0.00000968, -0.00001084, -0.00003843, -0.00004358, 0.00002146, 0.00010370, 0.00007005, -0.00011696, -0.00027770, -0.00017363, 0.00016212, 0.00033762, -0.00001117, -0.00070929, -0.00104317, -0.00043271, 0.00086652, 0.00184717, 0.00165692, 0.00043783, -0.00079187, -0.00111037, -0.00051401, 0.00022156, 0.00040183, 0.00004183, -0.00032499, -0.00030784, -0.00001664, 0.00018465, 0.00012665, -0.00004218, -0.00010710, -0.00003900, 0.00003254, 0.00001734, -0.00004087, -0.00005456, -0.00000964, 0.00003621, 0.00003841, 0.00000921, -0.00001661, -0.00002593, -0.00002614, -0.00002126, -0.00000871, 0.00000723, 0.00001536, 0.00001257, 0.00000767, 0.00000711, },
		{ 0.00000924, 0.00001088, 0.00001054, 0.00000121, -0.00001382, -0.00002464, -0.00002634, -0.00001937, -0.00000285, 0.00002019, 0.00003235, 0.00001448, -0.00002272, -0.00003832, -0.00001279, 0.00001309, -0.00001384, -0.00006870, -0.00005426, 0.00006225, 0.00015206, 0.00004828, -0.00020748, -0.00032710, -0.00009283, 0.00029587, 0.00033013, -0.00023452, -0.00095285, -0.00100285, -0.00003781, 0.00131790, 0.00195312, 0.00131790, -0.00003781, -0.00100285, -0.00095285, -0.00023452, 0.00033013, 0.00029587, -0.00009283, -0.00032710, -0.00020748, 0.00004828, 0.00015206, 0.00006225, -0.00005426, -0.00006870, -0.00001384, 0.00001309, -0.00001279, -0.00003832, -0.00002272, 0.00001448, 0.00003235, 0.00002019, -0.00000285, -0.00001937, -0.00002634, -0.00002464, -0.00001382, 0.00000121, 0.00001054, 0.00001088, },
		{ 0.00000568, 0.00000711, 0.00000767, 0.00001257, 0.00001536, 0.00000723, -0.00000871, -0.00002126, -0.00002614, -0.00002593, -0.00001661, 0.00000921, 0.00003841, 0.00003621, -0.00000964, -0.00005456, -0.00004087, 0.00001734, 0.00003254, -0.00003900, -0.00010710, -0.00004218, 0.00012665, 0.00018465, -0.00001664, -0.00030784, -0.00032499, 0.00004183, 0.00040183, 0.00022156, -0.00051401, -0.00111037, -0.00079187, 0.00043783, 0.00165692, 0.00184717, 0.00086652, -0.00043271, -0.00104317, -0.00070929, -0.00001117, 0.00033762, 0.00016212, -0.00017363, -0.00027770, -0.00011696, 0.00007005, 0.00010370, 0.00002146, -0.00004358, -0.00003843, -0.00001084, -0.00000968, -0.00002453, -0.00002152, 0.00000289, 0.00002354, 0.00002307, 0.00000704, -0.00001103, -0.00002324, -0.00002587, -0.00001756, -0.00000367, },
		{ -0.00002040, -0.00000710, 0.00000223, 0.00000307, 0.00000368, 0.00001138, 0.00001881, 0.00001328, -0.00000349, -0.00001697, -0.00002158, -0.00002573, -0.00002898, -0.00001219, 0.00002983, 0.00005804, 0.00002529, -0.00004959, -0.00008037, -0.00001937, 0.00005818, 0.00003263, -0.00008738, -0.00013604, 0.00000388, 0.00019995, 0.00017859, -0.00012064, -0.00038589, -0.00025407, 0.00020062, 0.00043329, 0.00001262, -0.00078619, -0.00112191, -0.00043059, 0.00090207, 0.00181334, 0.00155633, 0.00039742, -0.00068433, -0.00093915, -0.00045147, 0.00012174, 0.00027308, 0.00004416, -0.00019552, -0.00020586, -0.00005566, 0.00006054, 0.00006017, 0.00000483, -0.00002635, -0.00002411, -0.00001969, -0.00002368, -0.00002050, -0.00000234, 0.00001760, 0.00002375, 0.00001430, -0.00000280, -0.00001877, -0.00002601, },
		{ -0.00001226, -0.00002528, -0.00002360, -0.00001039, 0.00000024, 0.00000022, -0.00000135, 0.00000752, 0.00002060, 0.00001939, 0.00000102, -0.00001516, -0.00001646, -0.00001677, -0.00002999, -0.00003517, 0.00000071, 0.00005976, 0.00006914, -0.00000766, -0.00009682, -0.00008473, 0.00002691, 0.00009268, 0.00000179, -0.00014696, -0.00013534, 0.00008174, 0.00025656, 0.00012044, -0.00024064, -0.00040880, -0.00012044, 0.00033604, 0.00036002, -0.00025981, -0.00097737, -0.00096302, 0.00001582, 0.00126129, 0.00176396, 0.00115199, -0.00000061, -0.00077556, -0.00074901, -0.00023830, 0.00016443, 0.00017879, -0.00003218, -0.00017376, -0.00013829, -0.00002615, 0.00003810, 0.00003182, 0.00000277, -0.00001486, -0.00002296, -0.00002842, -0.00002539, -0.00000920, 0.00001158, 0.00002363, 0.00002116, 0.00000676, },
		{ 0.00002571, 0.00001789, -0.00000167, -0.00002143, -0.00002662, -0.00001486, -0.00000156, -0.00000116, -0.00000646, 0.00000101, 0.00002033, 0.00002691, 0.00000726, -0.00001701, -0.00001810, -0.00000580, -0.00001460, -0.00004273, -0.00003787, 0.00002805, 0.00009206, 0.00005931, -0.00006150, -0.00013461, -0.00005704, 0.00008616, 0.00010129, -0.00005920, -0.00019554, -0.00009265, 0.00017297, 0.00027024, 0.00001553, -0.00034079, -0.00035821, 0.00004384, 0.00039971, 0.00018392, -0.00053100, -0.00102934, -0.00065665, 0.00045635, 0.00145009, 0.00153829, 0.00072512, -0.00027068, -0.00073298, -0.00053882, -0.00009978, 0.00014308, 0.00009211, -0.00006322, -0.00013220, -0.00008969, -0.00001862, 0.00001757, 0.00001754, 0.00000357, -0.00001262, -0.00002636, -0.00003002, -0.00001885, 0.00000135, 0.00001907, },
		{ -0.00001080, 0.00000788, 0.00002317, 0.00002649, 0.00001240, -0.00001181, -0.00002654, -0.00001984, -0.00000414, -0.00000132, -0.00001095, -0.00000878, 0.00001561, 0.00003544, 0.00002000, -0.00001710, -0.00002988, -0.00000611, 0.00000840, -0.00002345, -0.00006060, -0.00002714, 0.00006911, 0.00011292, 0.00002082, -0.00012383, -0.00014465, 0.00000132, 0.00013704, 0.00007176, -0.00013464, -0.00020999, -0.00001172, 0.00024844, 0.00022680, -0.00011059, -0.00038647, -0.00024113, 0.00019009, 0.00036229, -0.00005712, -0.00072937, -0.00092134, -0.00026773, 0.00080323, 0.00145045, 0.00120547, 0.00035676, -0.00039972, -0.00061009, -0.00035994, -0.00003481, 0.00009346, 0.00003356, -0.00006161, -0.00009102, -0.00006099, -0.00001977, 0.00000468, 0.00000969, 0.00000036, -0.00001550, -0.00002660, -0.00002469, },
		{ -0.00001692, -0.00002139, -0.00001729, -0.00000543, 0.00001202, 0.00002669, 0.00002478, 0.00000290, -0.00002054, -0.00002283, -0.00000699, 0.00000017, -0.00001299, -0.00002127, 0.00000244, 0.00003952, 0.00004059, -0.00000515, -0.00004408, -0.00002635, 0.00001842, 0.00001468, -0.00004600, -0.00007311, 0.00000376, 0.00011365, 0.00010753, -0.00004308, -0.00017417, -0.00011612, 0.00007488, 0.00015749, 0.00000670, -0.00019895, -0.00017758, 0.00008639, 0.00028020, 0.00013269, -0.00022054, -0.00035966, -0.00009035, 0.00027124, 0.00022741, -0.00030028, -0.00080196, -0.00067849, 0.00011818, 0.00099929, 0.00129329, 0.00084857, 0.00009540, -0.00041333, -0.00046406, -0.00023660, -0.00002117, 0.00004495, 0.00000513, -0.00004524, -0.00005974, -0.00004460, -0.00002075, -0.00000312, 0.00000114, -0.00000648, },
		{ -0.00001133, -0.00000996, -0.00001204, -0.00001329, -0.00001388, -0.00001234, -0.00000218, 0.00001686, 0.00002900, 0.00001761, -0.00000901, -0.00002185, -0.00000915, 0.00000405, -0.00000911, -0.00003193, -0.00002040, 0.00002975, 0.00006113, 0.00002498, -0.00004339, -0.00005903, -0.00000205, 0.00004435, 0.00000351, -0.00007563, -0.00006775, 0.00005241, 0.00014463, 0.00006823, -0.00011665, -0.00019227, -0.00005257, 0.00013916, 0.00013531, -0.00007542, -0.00022754, -0.00010354, 0.00017080, 0.00025441, 0.00001455, -0.00028018, -0.00026838, 0.00004696, 0.00026109, 0.00003263, -0.00047991, -0.00073348, -0.00036292, 0.00042328, 0.00103300, 0.00104267, 0.00053721, -0.00005062, -0.00035798, -0.00033644, -0.00016666, -0.00003096, 0.00001195, -0.00000303, -0.00002734, -0.00003762, -0.00003233, -0.00002012, },
		{ -0.00001936, -0.00002084, -0.00002141, -0.00002275, -0.00002016, -0.00001207, -0.00000636, -0.00000889, -0.00001040, 0.00000245, 0.00002290, 0.00002627, 0.00000420, -0.00001734, -0.00001104, 0.00000913, 0.00000324, -0.00003233, -0.00004629, 0.00000023, 0.00006545, 0.00006542, -0.00001344, -0.00008028, -0.00005017, 0.00003657, 0.00005717, -0.00002614, -0.00009850, -0.00003757, 0.00010567, 0.00014572, 0.00000038, -0.00017670, -0.00016787, 0.00002772, 0.00017097, 0.00007486, -0.00014727, -0.00020831, -0.00001034, 0.00021381, 0.00017864, -0.00009093, -0.00027347, -0.00014472, 0.00012795, 0.00016569, -0.00016393, -0.00055133, -0.00055092, -0.00005123, 0.00060154, 0.00093583, 0.00077041, 0.00030938, -0.00010506, -0.00027979, -0.00024463, -0.00013246, -0.00004361, -0.00000548, -0.00000273, -0.00001235, },
		{ -0.00001444, -0.00000059, 0.00000187, -0.00000144, -0.00001225, -0.00002785, -0.00003507, -0.00002456, -0.00000678, -0.00000119, -0.00000846, -0.00000785, 0.00001081, 0.00002652, 0.00001495, -0.00001127, -0.00001461, 0.00001077, 0.00002130, -0.00001544, -0.00006067, -0.00004282, 0.00003962, 0.00009373, 0.00004228, -0.00006484, -0.00009854, -0.00001566, 0.00007437, 0.00004701, -0.00006464, -0.00010036, 0.00001324, 0.00014425, 0.00010998, -0.00007772, -0.00020240, -0.00010654, 0.00009866, 0.00015903, -0.00000373, -0.00018456, -0.00014786, 0.00007136, 0.00020326, 0.00007936, -0.00015246, -0.00020965, -0.00003195, 0.00013165, 0.00002122, -0.00030559, -0.00050627, -0.00031160, 0.00019203, 0.00064879, 0.00076461, 0.00053216, 0.00016823, -0.00010594, -0.00020913, -0.00018536, -0.00011446, -0.00005108, },
		{ -0.00014562, -0.00010083, -0.00005517, -0.00002111, 0.00000341, 0.00001841, 0.00001482, -0.00001038, -0.00003842, -0.00004281, -0.00002143, -0.00000102, -0.00000177, -0.00000983, -0.00000031, 0.00002090, 0.00002198, -0.00000420, -0.00002044, 0.00000310, 0.00003511, 0.00001662, -0.00004906, -0.00007930, -0.00001461, 0.00008561, 0.00009813, -0.00000681, -0.00011083, -0.00008691, 0.00003419, 0.00009484, 0.00001476, -0.00009565, -0.00007469, 0.00006936, 0.00015149, 0.00004475, -0.00014169, -0.00018482, -0.00002833, 0.00013752, 0.00010941, -0.00007320, -0.00017690, -0.00006892, 0.00011613, 0.00014794, -0.00000950, -0.00015748, -0.00011906, 0.00003372, 0.00006713, -0.00012097, -0.00035835, -0.00037320, -0.00008057, 0.00033381, 0.00059629, 0.00057771, 0.00035420, 0.00009241, -0.00008722, -0.00015620, },
		{ 0.00005300, -0.00006787, -0.00011664, -0.00011395, -0.00008950, -0.00006045, -0.00002635, 0.00001230, 0.00003677, 0.00002488, -0.00001610, -0.00004767, -0.00004187, -0.00001432, -0.00000071, -0.00000744, -0.00000721, 0.00001320, 0.00002648, 0.00000591, -0.00002451, -0.00001511, 0.00003223, 0.00004927, -0.00000913, -0.00008527, -0.00007465, 0.00003367, 0.00011945, 0.00007028, -0.00006783, -0.00013280, -0.00004642, 0.00008126, 0.00008862, -0.00002749, -0.00010431, -0.00002738, 0.00011060, 0.00012186, -0.00003017, -0.00017155, -0.00013146, 0.00004075, 0.00013499, 0.00004273, -0.00011133, -0.00013232, -0.00000002, 0.00011477, 0.00007310, -0.00006152, -0.00011680, -0.00003923, 0.00003814, -0.00003153, -0.00021542, -0.00032176, -0.00020322, 0.00009286, 0.00037757, 0.00049222, 0.00041552, 0.00023450, },
		{ 0.00038050, 0.00029256, 0.00015457, 0.00002818, -0.00005119, -0.00008193, -0.00008653, -0.00008374, -0.00006821, -0.00002675, 0.00002703, 0.00005216, 0.00002538, -0.00002519, -0.00004924, -0.00003319, -0.00001018, -0.00000833, -0.00001262, 0.00000381, 0.00002813, 0.00002069, -0.00001909, -0.00003586, 0.00000739, 0.00006179, 0.00004277, -0.00004983, -0.00010745, -0.00004215, 0.00008625, 0.00012464, 0.00001526, -0.00011982, -0.00012105, 0.00000939, 0.00010832, 0.00005855, -0.00006215, -0.00008524, 0.00002504, 0.00012127, 0.00006449, -0.00009081, -0.00016059, -0.00006295, 0.00008014, 0.00009869, -0.00001559, -0.00011058, -0.00007272, 0.00003747, 0.00007748, 0.00000824, -0.00006890, -0.00005823, 0.00000202, -0.00000621, -0.00012122, -0.00024002, -0.00022533, -0.00004823, 0.00018926, 0.00035358, },
		{ 0.00005830, 0.00021925, 0.00030059, 0.00028635, 0.00020219, 0.00009455, 0.00000968, -0.00003211, -0.00004792, -0.00006622, -0.00008580, -0.00007483, -0.00001909, 0.00004386, 0.00005883, 0.00001786, -0.00003015, -0.00004054, -0.00002306, -0.00001483, -0.00002049, -0.00001022, 0.00002237, 0.00003621, 0.00000087, -0.00004423, -0.00003000, 0.00004123, 0.00007808, 0.00001356, -0.00009062, -0.00010291, 0.00001008, 0.00012300, 0.00009540, -0.00004970, -0.00014439, -0.00007956, 0.00006081, 0.00010737, 0.00001838, -0.00007501, -0.00004580, 0.00006292, 0.00009799, -0.00000049, -0.00011959, -0.00011803, -0.00000355, 0.00008334, 0.00004869, -0.00004701, -0.00008037, -0.00002294, 0.00003840, 0.00002745, -0.00002660, -0.00004426, -0.00001234, -0.00000176, -0.00006798, -0.00016875, -0.00020221, -0.00011285, },
		{ -0.00016410, -0.00013178, -0.00002259, 0.00011007, 0.00021123, 0.00024818, 0.00021342, 0.00012913, 0.00004390, -0.00000001, -0.00000626, -0.00001740, -0.00005715, -0.00009304, -0.00007478, -0.00000447, 0.00005581, 0.00005460, 0.00000905, -0.00002482, -0.00002515, -0.00001958, -0.00002875, -0.00002944, 0.00000348, 0.00004164, 0.00003012, -0.00002854, -0.00005832, -0.00000577, 0.00007375, 0.00007165, -0.00002963, -0.00011453, -0.00006990, 0.00006403, 0.00012883, 0.00004012, -0.00010274, -0.00013427, -0.00002408, 0.00009134, 0.00008320, -0.00001444, -0.00006251, -0.00000272, 0.00007291, 0.00005137, -0.00005111, -0.00011276, -0.00006402, 0.00002979, 0.00005951, 0.00000711, -0.00004770, -0.00004094, 0.00000113, 0.00001441, -0.00001140, -0.00002918, -0.00001231, 0.00000231, -0.00003477, -0.00011301, },
		{ -0.00001296, -0.00006892, -0.00012055, -0.00012179, -0.00006456, 0.00002725, 0.00012355, 0.00019427, 0.00020758, 0.00015163, 0.00006211, 0.00000342, 0.00000440, 0.00002437, 0.00000323, -0.00005942, -0.00009876, -0.00006535, 0.00001097, 0.00005707, 0.00004337, 0.00000741, -0.00000860, -0.00001139, -0.00002767, -0.00004611, -0.00002759, 0.00002568, 0.00005199, 0.00000832, -0.00005716, -0.00005370, 0.00002923, 0.00009133, 0.00004274, -0.00007073, -0.00011087, -0.00001922, 0.00009990, 0.00010112, -0.00002217, -0.00012732, -0.00009619, 0.00002568, 0.00009518, 0.00005017, -0.00002767, -0.00003327, 0.00002633, 0.00005436, 0.00000076, -0.00007351, -0.00008105, -0.00001959, 0.00003418, 0.00002722, -0.00001249, -0.00002843, -0.00001222, -0.00000100, -0.00001393, -0.00002629, -0.00001384, 0.00000396, },
		{ -0.00002063, -0.00000180, -0.00000475, -0.00003961, -0.00007923, -0.00009172, -0.00007082, -0.00002550, 0.00004188, 0.00012429, 0.00018457, 0.00017346, 0.00008931, 0.00000026, -0.00002007, 0.00002342, 0.00005206, 0.00000972, -0.00006745, -0.00009659, -0.00004960, 0.00001925, 0.00004720, 0.00003361, 0.00001810, 0.00001255, -0.00000765, -0.00004566, -0.00005770, -0.00001241, 0.00004772, 0.00004631, -0.00002317, -0.00007360, -0.00003073, 0.00006173, 0.00008628, 0.00000181, -0.00009393, -0.00008075, 0.00003043, 0.00010540, 0.00005092, -0.00007059, -0.00011931, -0.00004689, 0.00005545, 0.00007883, 0.00002417, -0.00002016, -0.00000317, 0.00003153, 0.00001903, -0.00003524, -0.00006746, -0.00004312, 0.00000306, 0.00002005, 0.00000413, -0.00001045, -0.00000703, -0.00000388, -0.00001635, -0.00002897, },
		{ -0.00001161, -0.00002748, -0.00002531, -0.00001014, -0.00000841, -0.00003005, -0.00005338, -0.00005660, -0.00004662, -0.00003869, -0.00001910, 0.00004175, 0.00013181, 0.00018147, 0.00013378, 0.00002290, -0.00004857, -0.00002189, 0.00005031, 0.00006833, 0.00000394, -0.00007288, -0.00008499, -0.00003520, 0.00001546, 0.00003219, 0.00003386, 0.00003953, 0.00002906, -0.00001741, -0.00006577, -0.00005621, 0.00001149, 0.00005939, 0.00002430, -0.00005321, -0.00007128, 0.00000195, 0.00007926, 0.00006088, -0.00003535, -0.00009137, -0.00003671, 0.00006137, 0.00008130, -0.00000253, -0.00009186, -0.00008773, -0.00000440, 0.00006225, 0.00005624, 0.00001436, -0.00000146, 0.00001339, 0.00001535, -0.00001549, -0.00004747, -0.00004487, -0.00001570, 0.00000497, 0.00000376, -0.00000184, 0.00000137, 0.00000212, },
		{ 0.00000475, 0.00000808, -0.00000419, -0.00002151, -0.00002318, -0.00001159, -0.00001220, -0.00003439, -0.00005170, -0.00003824, -0.00001159, -0.00001212, -0.00003799, -0.00003109, 0.00004850, 0.00015060, 0.00017309, 0.00008088, -0.00004069, -0.00007507, -0.00000489, 0.00007412, 0.00006961, -0.00000660, -0.00006961, -0.00006870, -0.00003014, 0.00000115, 0.00002152, 0.00004767, 0.00006386, 0.00003304, -0.00003699, -0.00007819, -0.00004054, 0.00003431, 0.00005435, -0.00000690, -0.00007099, -0.00005186, 0.00003160, 0.00007634, 0.00002629, -0.00005542, -0.00006688, 0.00000361, 0.00006518, 0.00004041, -0.00004112, -0.00008545, -0.00004895, 0.00002048, 0.00005396, 0.00004050, 0.00001906, 0.00001434, 0.00001092, -0.00000974, -0.00003503, -0.00003947, -0.00002210, -0.00000557, -0.00000245, -0.00000218, },
		{ -0.00000783, -0.00000570, 0.00000434, 0.00001146, 0.00000115, -0.00001732, -0.00001963, -0.00000497, -0.00000412, -0.00003486, -0.00006561, -0.00004986, 0.00000468, 0.00003223, -0.00000730, -0.00006187, -0.00003925, 0.00007261, 0.00017075, 0.00014715, 0.00001785, -0.00008895, -0.00007589, 0.00002074, 0.00008580, 0.00005950, -0.00001305, -0.00005765, -0.00005617, -0.00003776, -0.00001602, 0.00002297, 0.00007140, 0.00008112, 0.00002270, -0.00005736, -0.00007650, -0.00001675, 0.00004610, 0.00003356, -0.00003624, -0.00007138, -0.00002421, 0.00004783, 0.00005693, -0.00000345, -0.00005400, -0.00003345, 0.00002619, 0.00004596, 0.00000051, -0.00005603, -0.00006199, -0.00001786, 0.00002800, 0.00004315, 0.00003676, 0.00002821, 0.00001703, -0.00000490, -0.00002875, -0.00003567, -0.00002380, -0.00001085, },
		{ -0.00002448, -0.00001391, -0.00001243, -0.00001068, 0.00000233, 0.00001503, 0.00000678, -0.00001661, -0.00002314, -0.00000104, 0.00001432, -0.00001617, -0.00007147, -0.00008152, -0.00001686, 0.00005770, 0.00005143, -0.00003390, -0.00008992, -0.00002688, 0.00010889, 0.00017650, 0.00010027, -0.00004280, -0.00011182, -0.00005678, 0.00004254, 0.00008299, 0.00004652, -0.00001048, -0.00004308, -0.00005452, -0.00005453, -0.00002638, 0.00003827, 0.00009604, 0.00008451, 0.00000334, -0.00006898, -0.00006113, 0.00000507, 0.00004174, 0.00000465, -0.00005424, -0.00005701, 0.00000052, 0.00004762, 0.00003167, -0.00001861, -0.00003706, -0.00000612, 0.00002716, 0.00001712, -0.00002461, -0.00005040, -0.00003618, -0.00000084, 0.00002693, 0.00003904, 0.00004049, 0.00003031, 0.00000588, -0.00002142, -0.00003256, },
		{ -0.00001050, -0.00002692, -0.00002252, -0.00001409, -0.00001618, -0.00001855, -0.00000391, 0.00001833, 0.00001758, -0.00001212, -0.00003310, -0.00001152, 0.00002680, 0.00001873, -0.00004912, -0.00010396, -0.00006722, 0.00003778, 0.00009605, 0.00003524, -0.00007647, -0.00010046, 0.00000828, 0.00014112, 0.00015699, 0.00004092, -0.00008614, -0.00010788, -0.00003003, 0.00005203, 0.00007068, 0.00003896, -0.00000055, -0.00003433, -0.00006520, -0.00007179, -0.00002346, 0.00006211, 0.00011201, 0.00007399, -0.00001560, -0.00006682, -0.00003898, 0.00001595, 0.00002328, -0.00002225, -0.00005741, -0.00003611, 0.00001475, 0.00003593, 0.00001164, -0.00001839, -0.00001681, 0.00000598, 0.00001351, -0.00000655, -0.00003126, -0.00003556, -0.00001856, 0.00000565, 0.00002755, 0.00004239, 0.00004162, 0.00001990, },
		{ 0.00004607, 0.00003190, 0.00000098, -0.00001996, -0.00001757, -0.00000863, -0.00001449, -0.00002683, -0.00001777, 0.00001430, 0.00003119, 0.00000422, -0.00003720, -0.00003489, 0.00001670, 0.00004824, -0.00000063, -0.00009052, -0.00011294, -0.00002258, 0.00009184, 0.00010104, -0.00000908, -0.00011148, -0.00008242, 0.00005368, 0.00015227, 0.00011317, -0.00001488, -0.00010289, -0.00008523, -0.00000823, 0.00004841, 0.00005759, 0.00004012, 0.00000985, -0.00003709, -0.00008307, -0.00008042, -0.00000739, 0.00008511, 0.00011413, 0.00005591, -0.00002564, -0.00005268, -0.00001965, 0.00001233, -0.00000184, -0.00003958, -0.00004890, -0.00001736, 0.00001780, 0.00002183, 0.00000229, -0.00001016, -0.00000428, 0.00000365, -0.00000205, -0.00001669, -0.00002592, -0.00002337, -0.00001059, 0.00001059, 0.00003458, },
		{ -0.00000587, 0.00002119, 0.00004440, 0.00004068, 0.00001079, -0.00001561, -0.00001455, 0.00000048, -0.00000282, -0.00002686, -0.00003510, -0.00000351, 0.00003601, 0.00002959, -0.00002259, -0.00005423, -0.00001592, 0.00004938, 0.00004876, -0.00003931, -0.00012033, -0.00008884, 0.00003665, 0.00012165, 0.00006971, -0.00006047, -0.00011989, -0.00004071, 0.00008969, 0.00013426, 0.00005756, -0.00005181, -0.00009395, -0.00005703, 0.00000162, 0.00003765, 0.00005072, 0.00004684, 0.00001297, -0.00005090, -0.00009915, -0.00007565, 0.00001544, 0.00009887, 0.00010391, 0.00003924, -0.00002345, -0.00003372, -0.00001045, -0.00000297, -0.00002507, -0.00004544, -0.00003550, -0.00000522, 0.00001490, 0.00001248, 0.00000107, -0.00000404, -0.00000312, -0.00000406, -0.00000899, -0.00001491, -0.00001960, -0.00001920, },
		{ -0.00001020, -0.00002026, -0.00001968, 0.00000402, 0.00003745, 0.00004788, 0.00002162, -0.00001354, -0.00001898, 0.00000440, 0.00001524, -0.00001121, -0.00004311, -0.00003072, 0.00002115, 0.00004852, 0.00000975, -0.00005082, -0.00005141, 0.00001792, 0.00006962, 0.00002399, -0.00008087, -0.00012175, -0.00003799, 0.00008629, 0.00011419, 0.00001620, -0.00009511, -0.00009666, 0.00000720, 0.00010029, 0.00009213, 0.00000768, -0.00006258, -0.00006905, -0.00003482, 0.00000035, 0.00002788, 0.00005151, 0.00005201, 0.00000494, -0.00006949, -0.00010558, -0.00005928, 0.00003615, 0.00010005, 0.00008804, 0.00003077, -0.00001196, -0.00001826, -0.00001331, -0.00002326, -0.00004112, -0.00004300, -0.00002286, 0.00000163, 0.00001253, 0.00000885, 0.00000025, -0.00000612, -0.00000819, -0.00000642, -0.00000496, },
		{ -0.00000961, 0.00000187, 0.00000359, -0.00001210, -0.00002779, -0.00001626, 0.00002264, 0.00005178, 0.00003706, -0.00000685, -0.00002890, -0.00000527, 0.00002685, 0.00001638, -0.00003022, -0.00005120, -0.00001102, 0.00004482, 0.00004241, -0.00002127, -0.00006548, -0.00002694, 0.00005126, 0.00006538, -0.00001681, -0.00010518, -0.00009066, 0.00001982, 0.00010576, 0.00007350, -0.00003567, -0.00009826, -0.00005246, 0.00004085, 0.00008107, 0.00004099, -0.00002244, -0.00005039, -0.00004112, -0.00002364, -0.00000571, 0.00002436, 0.00005571, 0.00004916, -0.00001209, -0.00008442, -0.00009953, -0.00003806, 0.00004810, 0.00009140, 0.00007426, 0.00003198, 0.00000211, -0.00001153, -0.00002500, -0.00004216, -0.00004945, -0.00003693, -0.00001232, 0.00000778, 0.00001344, 0.00000577, -0.00000709, -0.00001449, },
		{ -0.00000370, -0.00002157, -0.00002060, 0.00000013, 0.00001716, 0.00000632, -0.00002392, -0.00003455, -0.00000204, 0.00004531, 0.00005402, 0.00001132, -0.00003329, -0.00002653, 0.00001951, 0.00003997, 0.00000182, -0.00004858, -0.00004293, 0.00001671, 0.00005452, 0.00001802, -0.00004878, -0.00005740, 0.00000912, 0.00006751, 0.00003682, -0.00005497, -0.00009918, -0.00003855, 0.00006088, 0.00008819, 0.00001834, -0.00006425, -0.00007080, -0.00000734, 0.00004705, 0.00004061, -0.00000185, -0.00002816, -0.00002581, -0.00001987, -0.00002131, -0.00000958, 0.00002678, 0.00005665, 0.00003637, -0.00003147, -0.00008959, -0.00008399, -0.00001960, 0.00004987, 0.00007923, 0.00006721, 0.00003927, 0.00001232, -0.00001393, -0.00004014, -0.00005661, -0.00005261, -0.00002959, -0.00000127, 0.00001641, 0.00001444, },
		{ 0.00001337, 0.00002572, 0.00000880, -0.00002233, -0.00003625, -0.00001517, 0.00002058, 0.00002916, -0.00000364, -0.00004021, -0.00003095, 0.00002248, 0.00006139, 0.00003841, -0.00002087, -0.00004535, -0.00000682, 0.00004256, 0.00003591, -0.00002074, -0.00005423, -0.00001955, 0.00003897, 0.00004363, -0.00001386, -0.00005851, -0.00002899, 0.00003972, 0.00005800, -0.00000317, -0.00007201, -0.00006405, 0.00001290, 0.00006919, 0.00004345, -0.00002731, -0.00006015, -0.00002756, 0.00001965, 0.00002544, -0.00000485, -0.00002485, -0.00001468, -0.00000064, -0.00000822, -0.00002166, -0.00000795, 0.00003037, 0.00004960, 0.00001702, -0.00004596, -0.00008404, -0.00006535, -0.00000837, 0.00004491, 0.00006940, 0.00006666, 0.00004696, 0.00001565, -0.00002227, -0.00005473, -0.00006699, -0.00005272, -0.00001967, },
		{ -0.00007465, -0.00004898, -0.00000412, 0.00003095, 0.00002920, -0.00000838, -0.00004528, -0.00004057, 0.00000502, 0.00004266, 0.00002764, -0.00002448, -0.00005004, -0.00001252, 0.00004843, 0.00006073, 0.00000841, -0.00004588, -0.00003718, 0.00002032, 0.00005152, 0.00001664, -0.00003816, -0.00004184, 0.00000830, 0.00004383, 0.00001647, -0.00003737, -0.00004470, 0.00000653, 0.00004993, 0.00002711, -0.00003539, -0.00005935, -0.00001494, 0.00004247, 0.00004401, -0.00000730, -0.00004586, -0.00002957, 0.00001101, 0.00001929, -0.00001280, -0.00003965, -0.00002628, 0.00000739, 0.00001773, -0.00000307, -0.00001933, -0.00000252, 0.00002953, 0.00003443, -0.00000258, -0.00005159, -0.00007147, -0.00004934, -0.00000410, 0.00003848, 0.00006422, 0.00006884, 0.00005084, 0.00001295, -0.00003290, -0.00006771, },
		{ 0.00000686, -0.00004418, -0.00007975, -0.00007892, -0.00003812, 0.00001775, 0.00004550, 0.00001987, -0.00003513, -0.00006164, -0.00002735, 0.00003365, 0.00005367, 0.00000952, -0.00004600, -0.00004409, 0.00001597, 0.00006316, 0.00004058, -0.00002402, -0.00005244, -0.00001330, 0.00004062, 0.00004172, -0.00000690, -0.00003962, -0.00001643, 0.00002635, 0.00002811, -0.00001344, -0.00004019, -0.00001367, 0.00003186, 0.00003503, -0.00000987, -0.00004433, -0.00002290, 0.00002659, 0.00003968, -0.00000048, -0.00004219, -0.00003365, 0.00001026, 0.00002898, -0.00000494, -0.00005165, -0.00005546, -0.00001272, 0.00002813, 0.00002826, 0.00000096, -0.00001292, 0.00000209, 0.00002146, 0.00001520, -0.00001701, -0.00004901, -0.00005716, -0.00003814, -0.00000321, 0.00003414, 0.00006207, 0.00006961, 0.00005002, },
		{ 0.00005959, 0.00006715, 0.00004612, -0.00000094, -0.00005653, -0.00009044, -0.00007623, -0.00001772, 0.00004127, 0.00004835, -0.00000397, -0.00006162, -0.00006053, 0.00000149, 0.00005730, 0.00004511, -0.00001816, -0.00005529, -0.00002122, 0.00004216, 0.00005781, 0.00000833, -0.00004440, -0.00003823, 0.00001335, 0.00004340, 0.00001900, -0.00002101, -0.00002319, 0.00000904, 0.00002492, -0.00000020, -0.00003041, -0.00002046, 0.00001925, 0.00003488, 0.00000307, -0.00003469, -0.00002623, 0.00002008, 0.00004333, 0.00000931, -0.00004286, -0.00004886, -0.00000109, 0.00004041, 0.00002157, -0.00004033, -0.00007819, -0.00005318, 0.00000601, 0.00004258, 0.00003443, 0.00000737, -0.00000493, 0.00000185, 0.00000757, -0.00000287, -0.00002447, -0.00004194, -0.00004456, -0.00003017, -0.00000208, 0.00003196, },
		{ -0.00002236, 0.00000035, 0.00002943, 0.00005455, 0.00006234, 0.00004074, -0.00001091, -0.00007018, -0.00009630, -0.00006233, 0.00000984, 0.00005708, 0.00003373, -0.00003539, -0.00007444, -0.00003830, 0.00003417, 0.00006260, 0.00001856, -0.00004136, -0.00004466, 0.00000963, 0.00005272, 0.00003313, -0.00002152, -0.00004366, -0.00001165, 0.00002840, 0.00002769, -0.00000349, -0.00001722, 0.00000271, 0.00002086, 0.00000465, -0.00002595, -0.00002515, 0.00001206, 0.00003786, 0.00001496, -0.00002952, -0.00003577, 0.00000922, 0.00004991, 0.00003056, -0.00003188, -0.00006507, -0.00002907, 0.00003400, 0.00004916, -0.00000454, -0.00007242, -0.00008750, -0.00004003, 0.00002233, 0.00005115, 0.00004028, 0.00001599, 0.00000081, -0.00000404, -0.00000816, -0.00001657, -0.00002681, -0.00003400, -0.00003367, },
		},
		{ { 0.00001500, 0.00001574, 0.00001695, 0.00001466, 0.00000226, -0.00002404, -0.00005713, -0.00007816, -0.00006865, -0.00002782, 0.00002127, 0.00004670, 0.00003358, -0.00000185, -0.00002507, -0.00001343, 0.00002304, 0.00004997, 0.00003947, -0.00000460, -0.00004971, -0.00006297, -0.00003831, 0.00000131, 0.00002637, 0.00002554, 0.00001045, -0.00000029, 0.00000052, 0.00000568, 0.00000565, 0.00000015, -0.00000322, 0.00000015, 0.00000565, 0.00000568, 0.00000052, -0.00000029, 0.00001045, 0.00002554, 0.00002637, 0.00000131, -0.00003831, -0.00006297, -0.00004971, -0.00000460, 0.00003947, 0.00004997, 0.00002304, -0.00001343, -0.00002507, -0.00000185, 0.00003358, 0.00004670, 0.00002127, -0.00002782, -0.00006865, -0.00007816, -0.00005713, -0.00002404, 0.00000226, 0.00001466, 0.00001695, 0.00001574, },
		{ 0.00001542, 0.00002778, 0.00002528, 0.00001387, 0.00000474, 0.00000622, 0.00001483, 0.00001548, -0.00000518, -0.00004281, -0.00007319, -0.00007028, -0.00003045, 0.00002014, 0.00004492, 0.00002810, -0.00001071, -0.00003308, -0.00001595, 0.00002752, 0.00005943, 0.00005050, 0.00000508, -0.00004401, -0.00006416, -0.00004824, -0.00001521, 0.00001055, 0.00001918, 0.00001721, 0.00001426, 0.00001179, 0.00000535, -0.00000514, -0.00001156, -0.00000610, 0.00000791, 0.00001682, 0.00001053, -0.00000508, -0.00001203, 0.00000123, 0.00002514, 0.00003521, 0.00001516, -0.00002479, -0.00005393, -0.00004770, -0.00001050, 0.00002739, 0.00003658, 0.00001446, -0.00001367, -0.00001867, 0.00000591, 0.00003861, 0.00004943, 0.00002504, -0.00002192, -0.00006484, -0.00008303, -0.00007270, -0.00004389, -0.00001071, },
		{ -0.00008836, -0.00006373, -0.00002305, 0.00001667, 0.00003784, 0.00003268, 0.00001056, -0.00000699, -0.00000357, 0.00001621, 0.00002856, 0.00001186, -0.00003055, -0.00006868, -0.00007025, -0.00003043, 0.00002104, 0.00004405, 0.00002315, -0.00001863, -0.00004034, -0.00001969, 0.00002764, 0.00006257, 0.00005676, 0.00001505, -0.00003216, -0.00005624, -0.00005099, -0.00003019, -0.00000966, 0.00000562, 0.00001837, 0.00002786, 0.00002678, 0.00001066, -0.00001212, -0.00002384, -0.00001369, 0.00000979, 0.00002483, 0.00001713, -0.00000517, -0.00001842, -0.00000688, 0.00002041, 0.00003681, 0.00002344, -0.00001177, -0.00004025, -0.00003883, -0.00001083, 0.00001776, 0.00002283, 0.00000393, -0.00001666, -0.00001556, 0.00001029, 0.00004131, 0.00005215, 0.00003130, -0.00001237, -0.00005816, -0.00008678, },
		{ 0.00000046, -0.00004809, -0.00008815, -0.00010244, -0.00008217, -0.00003424, 0.00001790, 0.00004610, 0.00003723, 0.00000516, -0.00001866, -0.00001159, 0.00001907, 0.00004101, 0.00002627, -0.00002096, -0.00006490, -0.00006873, -0.00002819, 0.00002320, 0.00004334, 0.00001867, -0.00002455, -0.00004476, -0.00002226, 0.00002492, 0.00005904, 0.00005571, 0.00002139, -0.00001772, -0.00004052, -0.00004469, -0.00003944, -0.00003003, -0.00001324, 0.00001248, 0.00003677, 0.00004199, 0.00002064, -0.00001322, -0.00003307, -0.00002326, 0.00000588, 0.00002718, 0.00002212, -0.00000178, -0.00001892, -0.00001108, 0.00001466, 0.00003304, 0.00002539, -0.00000238, -0.00002596, -0.00002607, -0.00000608, 0.00001271, 0.00001232, -0.00000516, -0.00002055, -0.00001588, 0.00000967, 0.00003953, 0.00005297, 0.00003884, },
		{ 0.00003180, 0.00004966, 0.00004546, 0.00001508, -0.00003489, -0.00008604, -0.00011301, -0.00009736, -0.00004340, 0.00001880, 0.00005155, 0.00003811, -0.00000233, -0.00002953, -0.00001700, 0.00002368, 0.00005242, 0.00003730, -0.00001468, -0.00006207, -0.00006565, -0.00002380, 0.00002615, 0.00004195, 0.00001385, -0.00002862, -0.00004532, -0.00002150, 0.00002194, 0.00005052, 0.00004685, 0.00002107, -0.00000510, -0.00002056, -0.00003003, -0.00004022, -0.00004588, -0.00003355, 0.00000039, 0.00003856, 0.00005349, 0.00003247, -0.00000840, -0.00003656, -0.00003134, -0.00000179, 0.00002327, 0.00002280, 0.00000205, -0.00001499, -0.00001036, 0.00001082, 0.00002701, 0.00002247, 0.00000216, -0.00001448, -0.00001339, 0.00000115, 0.00001216, 0.00000736, -0.00000935, -0.00002225, -0.00001795, 0.00000398, },
		{ -0.00001841, -0.00001873, -0.00000498, 0.00001813, 0.00004058, 0.00004848, 0.00002893, -0.00001992, -0.00008008, -0.00011826, -0.00010712, -0.00004908, 0.00001956, 0.00005346, 0.00003450, -0.00001214, -0.00003919, -0.00001915, 0.00003043, 0.00006262, 0.00004437, -0.00001230, -0.00006055, -0.00006106, -0.00001712, 0.00002992, 0.00003948, 0.00000778, -0.00003191, -0.00004244, -0.00001638, 0.00002127, 0.00004006, 0.00003192, 0.00001307, 0.00000188, -0.00000118, -0.00001014, -0.00003219, -0.00005362, -0.00005061, -0.00001487, 0.00003299, 0.00005827, 0.00004250, 0.00000036, -0.00003348, -0.00003500, -0.00001008, 0.00001457, 0.00001814, 0.00000353, -0.00000941, -0.00000573, 0.00001032, 0.00002185, 0.00001762, 0.00000285, -0.00000749, -0.00000437, 0.00000708, 0.00001387, 0.00000812, -0.00000634, },
		{ 0.00001420, 0.00001245, 0.00000400, -0.00000679, -0.00001470, -0.00001385, 0.00000048, 0.00002550, 0.00004571, 0.00003887, -0.00000574, -0.00007110, -0.00011720, -0.00010941, -0.00004939, 0.00002112, 0.00005176, 0.00002583, -0.00002476, -0.00004770, -0.00001778, 0.00003964, 0.00007180, 0.00004740, -0.00001421, -0.00006087, -0.00005536, -0.00000813, 0.00003490, 0.00003625, 0.00000006, -0.00003571, -0.00003769, -0.00000742, 0.00002444, 0.00003101, 0.00001439, -0.00000128, 0.00000122, 0.00001303, 0.00001019, -0.00001763, -0.00005184, -0.00006032, -0.00002902, 0.00002213, 0.00005524, 0.00004746, 0.00000979, -0.00002515, -0.00003288, -0.00001592, 0.00000394, 0.00000909, 0.00000116, -0.00000514, 0.00000023, 0.00001264, 0.00001943, 0.00001407, 0.00000238, -0.00000441, -0.00000099, 0.00000809, },
		{ -0.00000280, -0.00000293, 0.00000209, 0.00000946, 0.00001656, 0.00001934, 0.00001278, -0.00000327, -0.00001863, -0.00001749, 0.00000617, 0.00003631, 0.00004213, 0.00000443, -0.00006128, -0.00011015, -0.00010300, -0.00004242, 0.00002516, 0.00004739, 0.00001225, -0.00004060, -0.00005572, -0.00001340, 0.00005125, 0.00008035, 0.00004689, -0.00002029, -0.00006355, -0.00004938, 0.00000269, 0.00004145, 0.00003316, -0.00000882, -0.00004088, -0.00003311, 0.00000342, 0.00003129, 0.00002580, -0.00000172, -0.00001863, -0.00000660, 0.00001908, 0.00002603, -0.00000078, -0.00004183, -0.00006053, -0.00003788, 0.00000972, 0.00004569, 0.00004574, 0.00001657, -0.00001448, -0.00002573, -0.00001749, -0.00000540, -0.00000172, -0.00000460, -0.00000413, 0.00000449, 0.00001555, 0.00001975, 0.00001407, 0.00000398, },
		{ 0.00002091, 0.00001797, 0.00001014, 0.00000068, -0.00000773, -0.00001070, -0.00000274, 0.00001628, 0.00003547, 0.00003789, 0.00001615, -0.00001599, -0.00003126, -0.00001379, 0.00002149, 0.00003729, 0.00000756, -0.00005371, -0.00009901, -0.00008809, -0.00002691, 0.00003370, 0.00004235, -0.00000501, -0.00005960, -0.00006422, -0.00000732, 0.00006449, 0.00008855, 0.00004388, -0.00002968, -0.00006870, -0.00004420, 0.00001424, 0.00004945, 0.00003128, -0.00001758, -0.00004729, -0.00003039, 0.00001349, 0.00004008, 0.00002518, -0.00001306, -0.00003472, -0.00001889, 0.00001645, 0.00003373, 0.00001348, -0.00002716, -0.00005173, -0.00003884, -0.00000015, 0.00003286, 0.00003797, 0.00001858, -0.00000487, -0.00001610, -0.00001491, -0.00001112, -0.00001101, -0.00001153, -0.00000649, 0.00000473, 0.00001603, },
		{ -0.00001041, 0.00000039, 0.00001153, 0.00001990, 0.00002409, 0.00002154, 0.00000891, -0.00001166, -0.00002766, -0.00002248, 0.00000849, 0.00004744, 0.00006401, 0.00004153, -0.00000455, -0.00003684, -0.00002973, 0.00000452, 0.00002504, 0.00000187, -0.00005154, -0.00008696, -0.00006665, -0.00000295, 0.00004849, 0.00003940, -0.00002345, -0.00008072, -0.00007401, -0.00000150, 0.00007769, 0.00009617, 0.00003959, -0.00004080, -0.00007575, -0.00004071, 0.00002494, 0.00005806, 0.00003128, -0.00002455, -0.00005376, -0.00003016, 0.00002036, 0.00004804, 0.00002803, -0.00001798, -0.00004572, -0.00003158, 0.00000712, 0.00003202, 0.00002119, -0.00001249, -0.00003690, -0.00003172, -0.00000446, 0.00002081, 0.00002691, 0.00001580, 0.00000107, -0.00000741, -0.00001027, -0.00001273, -0.00001608, -0.00001641, },
		{ -0.00001177, -0.00001589, -0.00001631, -0.00001279, -0.00000706, 0.00000107, 0.00001382, 0.00002924, 0.00003663, 0.00002312, -0.00001110, -0.00004457, -0.00004726, -0.00000764, 0.00005115, 0.00008561, 0.00006849, 0.00001438, -0.00003207, -0.00003724, -0.00001000, 0.00000837, -0.00001225, -0.00005692, -0.00007762, -0.00004222, 0.00002753, 0.00007014, 0.00004136, -0.00003955, -0.00010165, -0.00008519, 0.00000195, 0.00008838, 0.00010213, 0.00003493, -0.00005166, -0.00008329, -0.00003915, 0.00003321, 0.00006573, 0.00003301, -0.00002833, -0.00005853, -0.00003184, 0.00002262, 0.00005240, 0.00003190, -0.00001693, -0.00004941, -0.00004072, -0.00000519, 0.00002232, 0.00002053, -0.00000206, -0.00002053, -0.00001899, -0.00000239, 0.00001295, 0.00001648, 0.00001037, 0.00000261, -0.00000263, -0.00000681, },
		{ 0.00000124, -0.00000309, -0.00000775, -0.00001131, -0.00001154, -0.00000971, -0.00001037, -0.00001449, -0.00001417, 0.00000111, 0.00002992, 0.00005206, 0.00004224, -0.00000385, -0.00005688, -0.00007262, -0.00003026, 0.00004466, 0.00009772, 0.00009132, 0.00003688, -0.00001758, -0.00003366, -0.00001741, -0.00000805, -0.00003198, -0.00006995, -0.00007390, -0.00001916, 0.00006089, 0.00009756, 0.00005025, -0.00004943, -0.00011897, -0.00009668, 0.00000146, 0.00009374, 0.00010444, 0.00003002, -0.00006045, -0.00008935, -0.00003891, 0.00003797, 0.00007072, 0.00003544, -0.00002839, -0.00005990, -0.00003389, 0.00002035, 0.00005137, 0.00003398, -0.00001206, -0.00004577, -0.00004371, -0.00001638, 0.00000813, 0.00001243, 0.00000180, -0.00000714, -0.00000491, 0.00000427, 0.00001079, 0.00001041, 0.00000594, },
		{ 0.00001323, 0.00001077, 0.00000639, 0.00000031, -0.00000786, -0.00001494, -0.00001481, -0.00000602, 0.00000271, -0.00000103, -0.00001814, -0.00003107, -0.00001769, 0.00002365, 0.00006369, 0.00006307, 0.00000998, -0.00006126, -0.00009343, -0.00005521, 0.00002889, 0.00009746, 0.00010457, 0.00005730, 0.00000309, -0.00001912, -0.00001453, -0.00001913, -0.00005249, -0.00008821, -0.00007688, -0.00000145, 0.00009240, 0.00012784, 0.00006645, -0.00004985, -0.00012869, -0.00010613, -0.00000340, 0.00009134, 0.00010060, 0.00002394, -0.00006611, -0.00009193, -0.00003857, 0.00003916, 0.00007166, 0.00003705, -0.00002528, -0.00005698, -0.00003456, 0.00001489, 0.00004473, 0.00003223, -0.00000624, -0.00003681, -0.00004015, -0.00002344, -0.00000625, 0.00000015, -0.00000062, 0.00000033, 0.00000584, 0.00001161, },
		{ 0.00000152, 0.00000979, 0.00001492, 0.00001679, 0.00001709, 0.00001426, 0.00000382, -0.00001383, -0.00002799, -0.00002478, -0.00000319, 0.00001811, 0.00001527, -0.00001498, -0.00004540, -0.00003980, 0.00000992, 0.00006799, 0.00008122, 0.00002789, -0.00005660, -0.00010539, -0.00007718, 0.00000747, 0.00008491, 0.00010469, 0.00006987, 0.00002429, 0.00000312, -0.00000092, -0.00002096, -0.00006820, -0.00010708, -0.00008512, 0.00000864, 0.00011751, 0.00015680, 0.00008845, -0.00003911, -0.00012725, -0.00011038, -0.00001157, 0.00007994, 0.00008830, 0.00001498, -0.00006875, -0.00008972, -0.00003643, 0.00003769, 0.00006811, 0.00003642, -0.00002037, -0.00005007, -0.00003271, 0.00000814, 0.00003378, 0.00002618, -0.00000191, -0.00002565, -0.00003190, -0.00002544, -0.00001726, -0.00001187, -0.00000644, },
		{ -0.00002336, -0.00001958, -0.00001120, -0.00000115, 0.00000597, 0.00001031, 0.00001672, 0.00002625, 0.00002970, 0.00001496, -0.00001674, -0.00004412, -0.00004167, -0.00000647, 0.00003227, 0.00003624, -0.00000384, -0.00005317, -0.00006094, -0.00000939, 0.00006325, 0.00009243, 0.00004571, -0.00004445, -0.00010639, -0.00009134, -0.00001439, 0.00006321, 0.00009115, 0.00007041, 0.00003977, 0.00002731, 0.00002060, -0.00001223, -0.00007442, -0.00012086, -0.00009476, 0.00001139, 0.00013317, 0.00018011, 0.00011311, -0.00001765, -0.00011258, -0.00010639, -0.00002081, 0.00005997, 0.00006628, 0.00000126, -0.00006964, -0.00008265, -0.00003122, 0.00003511, 0.00006087, 0.00003296, -0.00001522, -0.00004056, -0.00002841, 0.00000167, 0.00002076, 0.00001705, -0.00000005, -0.00001527, -0.00002226, -0.00002389, },
		{ -0.00000729, -0.00001467, -0.00002207, -0.00002556, -0.00002085, -0.00001047, -0.00000322, -0.00000356, -0.00000380, 0.00000857, 0.00003331, 0.00005004, 0.00003475, -0.00001256, -0.00005887, -0.00006343, -0.00001752, 0.00004089, 0.00005776, 0.00001403, -0.00005196, -0.00007653, -0.00003046, 0.00005018, 0.00009395, 0.00005889, -0.00002849, -0.00009708, -0.00009475, -0.00003136, 0.00003764, 0.00006669, 0.00005766, 0.00004455, 0.00004670, 0.00004441, 0.00000496, -0.00006900, -0.00012452, -0.00010049, 0.00000983, 0.00013908, 0.00019480, 0.00013672, 0.00001204, -0.00008491, -0.00009235, -0.00002849, 0.00003344, 0.00003501, -0.00001836, -0.00007067, -0.00007204, -0.00002284, 0.00003287, 0.00005161, 0.00002732, -0.00001080, -0.00003033, -0.00002297, -0.00000409, 0.00000779, 0.00000724, 0.00000022, },
		{ -0.00000410, -0.00000067, 0.00000128, -0.00000129, -0.00001134, -0.00002379, -0.00002714, -0.00001628, -0.00000150, 0.00000027, -0.00001427, -0.00002587, -0.00001014, 0.00003307, 0.00007028, 0.00006137, 0.00000103, -0.00006742, -0.00008588, -0.00003554, 0.00004098, 0.00007506, 0.00003522, -0.00004167, -0.00008313, -0.00004865, 0.00003181, 0.00008544, 0.00006396, -0.00001336, -0.00008065, -0.00008726, -0.00003970, 0.00001405, 0.00003663, 0.00003384, 0.00003644, 0.00005548, 0.00006349, 0.00002537, -0.00005335, -0.00011545, -0.00009733, 0.00000889, 0.00013799, 0.00020060, 0.00015653, 0.00004635, -0.00004687, -0.00006869, -0.00003274, 0.00000335, -0.00000314, -0.00004344, -0.00007337, -0.00006023, -0.00001270, 0.00003143, 0.00004224, 0.00002132, -0.00000681, -0.00002086, -0.00001849, -0.00001016, },
		{ -0.00001236, -0.00001687, -0.00001863, -0.00001551, -0.00000512, 0.00000603, 0.00000505, -0.00001227, -0.00003185, -0.00003236, -0.00000916, 0.00001568, 0.00001300, -0.00002046, -0.00005139, -0.00003882, 0.00002174, 0.00008452, 0.00009032, 0.00002358, -0.00006611, -0.00010388, -0.00005728, 0.00003187, 0.00008415, 0.00005508, -0.00002468, -0.00007944, -0.00005989, 0.00001253, 0.00006905, 0.00005961, -0.00000314, -0.00006177, -0.00007151, -0.00003833, -0.00000285, 0.00000734, 0.00000405, 0.00001688, 0.00005055, 0.00007148, 0.00004195, -0.00003247, -0.00009483, -0.00008254, 0.00001373, 0.00013514, 0.00020058, 0.00017230, 0.00008208, -0.00000277, -0.00003843, -0.00003369, -0.00002762, -0.00004445, -0.00007154, -0.00007798, -0.00004961, -0.00000350, 0.00002971, 0.00003389, 0.00001727, -0.00000157, },
		{ 0.00002609, 0.00001696, 0.00000747, -0.00000342, -0.00001882, -0.00003179, -0.00002861, -0.00000640, 0.00001657, 0.00001511, -0.00001507, -0.00004678, -0.00004494, -0.00000446, 0.00003791, 0.00003597, -0.00001713, -0.00007407, -0.00007362, -0.00000160, 0.00008764, 0.00011552, 0.00005175, -0.00005377, -0.00011284, -0.00007793, 0.00001550, 0.00008305, 0.00006912, -0.00000511, -0.00006674, -0.00006190, -0.00000322, 0.00004879, 0.00004708, -0.00000014, -0.00004514, -0.00005206, -0.00002906, -0.00001087, -0.00001563, -0.00002512, -0.00000944, 0.00003264, 0.00006463, 0.00004786, -0.00001373, -0.00006786, -0.00005724, 0.00002761, 0.00013661, 0.00020083, 0.00018721, 0.00011815, 0.00004292, -0.00000669, -0.00003409, -0.00005842, -0.00008503, -0.00009854, -0.00008265, -0.00004147, 0.00000161, 0.00002502, },
		{ -0.00003499, 0.00000030, 0.00001375, 0.00001638, 0.00002056, 0.00002274, 0.00000882, -0.00002312, -0.00005095, -0.00004630, -0.00000688, 0.00003306, 0.00003190, -0.00001566, -0.00006634, -0.00006647, -0.00000718, 0.00005969, 0.00006681, -0.00000165, -0.00008758, -0.00010825, -0.00003448, 0.00007695, 0.00013098, 0.00008001, -0.00003218, -0.00011014, -0.00009252, -0.00000411, 0.00007236, 0.00007434, 0.00001241, -0.00004841, -0.00005485, -0.00001232, 0.00002925, 0.00002978, -0.00000415, -0.00003397, -0.00003390, -0.00001597, -0.00001086, -0.00002913, -0.00004757, -0.00003587, 0.00000621, 0.00004325, 0.00003875, -0.00000486, -0.00004285, -0.00002705, 0.00004996, 0.00014708, 0.00020890, 0.00020790, 0.00015698, 0.00008753, 0.00002076, -0.00003921, -0.00009075, -0.00012236, -0.00011969, -0.00008349, },
		{ -0.00015675, -0.00013125, -0.00007536, -0.00002678, -0.00000761, -0.00000736, 0.00000077, 0.00002597, 0.00004529, 0.00002769, -0.00002656, -0.00007536, -0.00007095, -0.00001036, 0.00005322, 0.00005669, -0.00000963, -0.00008589, -0.00009555, -0.00002061, 0.00007476, 0.00010021, 0.00002518, -0.00008747, -0.00013561, -0.00007142, 0.00005316, 0.00013253, 0.00010220, -0.00000577, -0.00009602, -0.00009743, -0.00002209, 0.00005509, 0.00007012, 0.00002417, -0.00002892, -0.00004123, -0.00001383, 0.00001420, 0.00001210, -0.00001269, -0.00002915, -0.00002088, -0.00000396, -0.00000624, -0.00003322, -0.00005953, -0.00005587, -0.00002175, 0.00001194, 0.00001449, -0.00001135, -0.00002904, -0.00000135, 0.00007534, 0.00016748, 0.00023146, 0.00024315, 0.00020490, 0.00013191, 0.00003953, -0.00005553, -0.00012925, },
		{ 0.00004884, -0.00008873, -0.00018069, -0.00019208, -0.00013228, -0.00005347, -0.00001140, -0.00001934, -0.00003971, -0.00002516, 0.00002879, 0.00007372, 0.00005566, -0.00002467, -0.00010189, -0.00010315, -0.00002081, 0.00007259, 0.00008810, 0.00000609, -0.00009972, -0.00012779, -0.00004505, 0.00007793, 0.00012923, 0.00005912, -0.00007253, -0.00014975, -0.00010535, 0.00002031, 0.00011909, 0.00011316, 0.00001965, -0.00007357, -0.00009150, -0.00003419, 0.00003567, 0.00005831, 0.00002841, -0.00001242, -0.00002498, -0.00000916, 0.00000554, -0.00000196, -0.00002195, -0.00002915, -0.00001462, 0.00000291, -0.00000171, -0.00003097, -0.00006066, -0.00006504, -0.00004381, -0.00002150, -0.00002019, -0.00003439, -0.00003393, 0.00000873, 0.00009376, 0.00019347, 0.00027158, 0.00030151, 0.00027123, 0.00018145, },
		{ 0.00038831, 0.00036606, 0.00024565, 0.00005299, -0.00014125, -0.00025205, -0.00023545, -0.00012584, -0.00001538, 0.00001727, -0.00002916, -0.00008159, -0.00006463, 0.00002328, 0.00010381, 0.00009295, -0.00001308, -0.00012532, -0.00014080, -0.00004089, 0.00008571, 0.00012197, 0.00003199, -0.00010273, -0.00015690, -0.00007738, 0.00006676, 0.00014727, 0.00009350, -0.00004530, -0.00014756, -0.00012949, -0.00001522, 0.00009307, 0.00011032, 0.00003848, -0.00004782, -0.00007636, -0.00003805, 0.00001858, 0.00004254, 0.00002573, -0.00000152, -0.00001020, -0.00000140, 0.00000286, -0.00001001, -0.00002813, -0.00003075, -0.00001413, 0.00000275, -0.00000144, -0.00002724, -0.00005411, -0.00006266, -0.00005446, -0.00004813, -0.00005654, -0.00006972, -0.00006059, -0.00000673, 0.00009249, 0.00021518, 0.00032664, },
		{ 0.00005911, 0.00021869, 0.00038739, 0.00050294, 0.00049716, 0.00033629, 0.00006165, -0.00021013, -0.00034790, -0.00029557, -0.00011927, 0.00003734, 0.00006396, -0.00002943, -0.00012762, -0.00011799, 0.00000394, 0.00012904, 0.00013669, 0.00001089, -0.00013964, -0.00017912, -0.00007067, 0.00008772, 0.00015227, 0.00006539, -0.00009212, -0.00017631, -0.00011174, 0.00004244, 0.00014989, 0.00012102, -0.00001143, -0.00012966, -0.00013919, -0.00004643, 0.00005969, 0.00009436, 0.00004695, -0.00002433, -0.00005593, -0.00003387, 0.00000698, 0.00002700, 0.00001868, 0.00000330, 0.00000012, 0.00000582, 0.00000387, -0.00001191, -0.00002883, -0.00003046, -0.00001643, -0.00000354, -0.00000757, -0.00002681, -0.00004543, -0.00005201, -0.00005201, -0.00006057, -0.00008383, -0.00010804, -0.00010605, -0.00005335, },
		{ -0.00013694, -0.00016105, -0.00013010, -0.00001502, 0.00018889, 0.00043882, 0.00063725, 0.00066672, 0.00046431, 0.00008854, -0.00028610, -0.00046788, -0.00038023, -0.00012309, 0.00009814, 0.00012992, -0.00001258, -0.00016940, -0.00018170, -0.00003260, 0.00014198, 0.00018093, 0.00004691, -0.00013965, -0.00021151, -0.00010707, 0.00007596, 0.00017263, 0.00010086, -0.00006837, -0.00018111, -0.00014101, 0.00000955, 0.00013613, 0.00013574, 0.00002182, -0.00010021, -0.00013318, -0.00006759, 0.00002559, 0.00006903, 0.00004416, -0.00000756, -0.00003513, -0.00002416, 0.00000189, 0.00001510, 0.00001072, 0.00000357, 0.00000495, 0.00000965, 0.00000545, -0.00000958, -0.00002384, -0.00002588, -0.00001782, -0.00001268, -0.00001925, -0.00003256, -0.00004064, -0.00003951, -0.00003938, -0.00005537, -0.00009230, },
		{ -0.00002357, -0.00003435, -0.00007601, -0.00014395, -0.00021169, -0.00022837, -0.00013163, 0.00011335, 0.00046277, 0.00077560, 0.00086897, 0.00063627, 0.00014861, -0.00035416, -0.00060492, -0.00049334, -0.00014864, 0.00015600, 0.00021135, 0.00002653, -0.00019712, -0.00024826, -0.00008597, 0.00013632, 0.00021767, 0.00009104, -0.00012277, -0.00023120, -0.00014431, 0.00005103, 0.00017812, 0.00013158, -0.00003536, -0.00016947, -0.00015881, -0.00002509, 0.00010882, 0.00013471, 0.00004760, -0.00006577, -0.00011381, -0.00007574, -0.00000302, 0.00003999, 0.00003215, 0.00000035, -0.00001846, -0.00001281, 0.00000204, 0.00000843, 0.00000495, 0.00000199, 0.00000526, 0.00000898, 0.00000497, -0.00000603, -0.00001517, -0.00001677, -0.00001546, -0.00002025, -0.00003283, -0.00004422, -0.00004410, -0.00003277, },
		{ -0.00005816, -0.00005680, -0.00003814, -0.00001377, -0.00000459, -0.00003493, -0.00011991, -0.00024240, -0.00033303, -0.00028384, -0.00001397, 0.00044190, 0.00089695, 0.00108930, 0.00085107, 0.00025418, -0.00039591, -0.00074498, -0.00063247, -0.00020501, 0.00019726, 0.00029905, 0.00008844, -0.00020199, -0.00030736, -0.00015135, 0.00010881, 0.00023867, 0.00013628, -0.00009005, -0.00023307, -0.00017521, 0.00001690, 0.00016680, 0.00015115, 0.00000059, -0.00014323, -0.00016117, -0.00005431, 0.00007375, 0.00011869, 0.00006123, -0.00003342, -0.00008632, -0.00007133, -0.00002201, 0.00001316, 0.00001518, -0.00000001, -0.00000859, -0.00000371, 0.00000462, 0.00000648, 0.00000299, 0.00000124, 0.00000339, 0.00000474, 0.00000134, -0.00000411, -0.00000622, -0.00000528, -0.00000863, -0.00002239, -0.00004299, },
		{ 0.00000071, -0.00001736, -0.00004452, -0.00006832, -0.00007571, -0.00005851, -0.00001871, 0.00002325, 0.00002451, -0.00006180, -0.00023978, -0.00042523, -0.00045639, -0.00019000, 0.00036394, 0.00097867, 0.00130561, 0.00109819, 0.00041112, -0.00039333, -0.00086864, -0.00078760, -0.00029594, 0.00020859, 0.00037956, 0.00016796, -0.00017878, -0.00034807, -0.00021990, 0.00006063, 0.00023768, 0.00017396, -0.00004640, -0.00021529, -0.00019306, -0.00001991, 0.00014034, 0.00015537, 0.00003241, -0.00010742, -0.00014766, -0.00007247, 0.00003809, 0.00009182, 0.00006141, -0.00000892, -0.00005720, -0.00005786, -0.00003022, -0.00000690, -0.00000172, -0.00000577, -0.00000578, 0.00000058, 0.00000647, 0.00000707, 0.00000448, 0.00000276, 0.00000191, -0.00000064, -0.00000446, -0.00000517, -0.00000045, 0.00000467, },
		{ -0.00000845, 0.00000005, 0.00000879, 0.00000835, -0.00000646, -0.00003425, -0.00006808, -0.00009435, -0.00009211, -0.00004398, 0.00003762, 0.00009054, 0.00002590, -0.00019611, -0.00048637, -0.00062794, -0.00040184, 0.00022539, 0.00100122, 0.00149159, 0.00135784, 0.00061598, -0.00033303, -0.00095444, -0.00094173, -0.00041768, 0.00018040, 0.00043769, 0.00025439, -0.00012775, -0.00036150, -0.00028050, -0.00000227, 0.00021225, 0.00019591, 0.00000038, -0.00018006, -0.00019350, -0.00005203, 0.00010374, 0.00014339, 0.00005417, -0.00006890, -0.00012141, -0.00007691, 0.00000846, 0.00006039, 0.00005026, 0.00000475, -0.00003241, -0.00004058, -0.00002949, -0.00001818, -0.00001434, -0.00001296, -0.00000803, -0.00000039, 0.00000538, 0.00000744, 0.00000738, 0.00000616, 0.00000235, -0.00000430, -0.00000974, },
		{ 0.00000444, -0.00000424, -0.00001112, -0.00001110, -0.00000445, 0.00000402, 0.00000913, 0.00000604, -0.00001241, -0.00005239, -0.00010454, -0.00013252, -0.00009005, 0.00002926, 0.00014870, 0.00013161, -0.00011173, -0.00050239, -0.00077504, -0.00062822, 0.00003347, 0.00095266, 0.00162145, 0.00160351, 0.00085494, -0.00021004, -0.00098317, -0.00107339, -0.00055841, 0.00010978, 0.00045991, 0.00033328, -0.00005529, -0.00034324, -0.00032221, -0.00007019, 0.00016479, 0.00019663, 0.00004145, -0.00013329, -0.00017578, -0.00007321, 0.00006399, 0.00011800, 0.00006262, -0.00003452, -0.00008812, -0.00006846, -0.00001079, 0.00003100, 0.00003250, 0.00000785, -0.00001577, -0.00002484, -0.00002372, -0.00002146, -0.00002056, -0.00001799, -0.00001208, -0.00000492, 0.00000102, 0.00000553, 0.00000888, 0.00000928, },
		{ 0.00000101, 0.00000674, 0.00000905, 0.00000596, -0.00000039, -0.00000592, -0.00000917, -0.00001113, -0.00001019, -0.00000131, 0.00001406, 0.00001706, -0.00001968, -0.00009880, -0.00017019, -0.00015173, -0.00000611, 0.00018532, 0.00023896, 0.00000447, -0.00046707, -0.00087680, -0.00084300, -0.00019425, 0.00083186, 0.00167507, 0.00180626, 0.00110513, -0.00003010, -0.00094228, -0.00116059, -0.00069975, 0.00000212, 0.00043774, 0.00038952, 0.00002690, -0.00029490, -0.00033697, -0.00013181, 0.00010232, 0.00017484, 0.00006910, -0.00008325, -0.00014314, -0.00007996, 0.00002827, 0.00008475, 0.00005793, -0.00000932, -0.00005442, -0.00005096, -0.00001843, 0.00000877, 0.00001381, 0.00000342, -0.00000819, -0.00001444, -0.00001732, -0.00001965, -0.00002076, -0.00001895, -0.00001474, -0.00000992, -0.00000490, },
		{ -0.00001427, -0.00001229, -0.00000921, -0.00000439, 0.00000037, 0.00000274, 0.00000342, 0.00000508, 0.00000656, 0.00000098, -0.00001537, -0.00003133, -0.00002489, 0.00001149, 0.00004727, 0.00002726, -0.00007271, -0.00019486, -0.00021907, -0.00006640, 0.00019089, 0.00033029, 0.00013656, -0.00038378, -0.00091925, -0.00102005, -0.00043251, 0.00064949, 0.00164226, 0.00194005, 0.00133816, 0.00019036, -0.00082925, -0.00118544, -0.00081992, -0.00012917, 0.00037023, 0.00041077, 0.00010423, -0.00022418, -0.00032181, -0.00017679, 0.00003495, 0.00013401, 0.00007877, -0.00003850, -0.00010195, -0.00007242, 0.00000205, 0.00005042, 0.00004344, 0.00000459, -0.00002608, -0.00002990, -0.00001635, -0.00000383, -0.00000082, -0.00000404, -0.00000783, -0.00001069, -0.00001370, -0.00001650, -0.00001746, -0.00001624, },
		{ -0.00001506, -0.00001412, -0.00001232, -0.00001111, -0.00001019, -0.00000856, -0.00000727, -0.00000855, -0.00001071, -0.00000661, 0.00000824, 0.00002445, 0.00002117, -0.00001219, -0.00005423, -0.00006009, -0.00000581, 0.00007018, 0.00008162, -0.00002641, -0.00019834, -0.00027971, -0.00014331, 0.00016241, 0.00039059, 0.00026490, -0.00026488, -0.00089818, -0.00113825, -0.00065292, 0.00042649, 0.00152518, 0.00198680, 0.00152518, 0.00042649, -0.00065292, -0.00113825, -0.00089818, -0.00026488, 0.00026490, 0.00039059, 0.00016241, -0.00014331, -0.00027971, -0.00019834, -0.00002641, 0.00008162, 0.00007018, -0.00000581, -0.00006009, -0.00005423, -0.00001219, 0.00002117, 0.00002445, 0.00000824, -0.00000661, -0.00001071, -0.00000855, -0.00000727, -0.00000856, -0.00001019, -0.00001111, -0.00001232, -0.00001412, },
		{ -0.00001427, -0.00001624, -0.00001746, -0.00001650, -0.00001370, -0.00001069, -0.00000783, -0.00000404, -0.00000082, -0.00000383, -0.00001635, -0.00002990, -0.00002608, 0.00000459, 0.00004344, 0.00005042, 0.00000205, -0.00007242, -0.00010195, -0.00003850, 0.00007877, 0.00013401, 0.00003495, -0.00017679, -0.00032181, -0.00022418, 0.00010423, 0.00041077, 0.00037023, -0.00012917, -0.00081992, -0.00118544, -0.00082925, 0.00019036, 0.00133816, 0.00194005, 0.00164226, 0.00064949, -0.00043251, -0.00102005, -0.00091925, -0.00038378, 0.00013656, 0.00033029, 0.00019089, -0.00006640, -0.00021907, -0.00019486, -0.00007271, 0.00002726, 0.00004727, 0.00001149, -0.00002489, -0.00003133, -0.00001537, 0.00000098, 0.00000656, 0.00000508, 0.00000342, 0.00000274, 0.00000037, -0.00000439, -0.00000921, -0.00001229, },
		{ 0.00000101, -0.00000490, -0.00000992, -0.00001474, -0.00001895, -0.00002076, -0.00001965, -0.00001732, -0.00001444, -0.00000819, 0.00000342, 0.00001381, 0.00000877, -0.00001843, -0.00005096, -0.00005442, -0.00000932, 0.00005793, 0.00008475, 0.00002827, -0.00007996, -0.00014314, -0.00008325, 0.00006910, 0.00017484, 0.00010232, -0.00013181, -0.00033697, -0.00029490, 0.00002690, 0.00038952, 0.00043774, 0.00000212, -0.00069975, -0.00116059, -0.00094228, -0.00003010, 0.00110513, 0.00180626, 0.00167507, 0.00083186, -0.00019425, -0.00084300, -0.00087680, -0.00046707, 0.00000447, 0.00023896, 0.00018532, -0.00000611, -0.00015173, -0.00017019, -0.00009880, -0.00001968, 0.00001706, 0.00001406, -0.00000131, -0.00001019, -0.00001113, -0.00000917, -0.00000592, -0.00000039, 0.00000596, 0.00000905, 0.00000674, },
		{ 0.00000444, 0.00000928, 0.00000888, 0.00000553, 0.00000102, -0.00000492, -0.00001208, -0.00001799, -0.00002056, -0.00002146, -0.00002372, -0.00002484, -0.00001577, 0.00000785, 0.00003250, 0.00003100, -0.00001079, -0.00006846, -0.00008812, -0.00003452, 0.00006262, 0.00011800, 0.00006399, -0.00007321, -0.00017578, -0.00013329, 0.00004145, 0.00019663, 0.00016479, -0.00007019, -0.00032221, -0.00034324, -0.00005529, 0.00033328, 0.00045991, 0.00010978, -0.00055841, -0.00107339, -0.00098317, -0.00021004, 0.00085494, 0.00160351, 0.00162145, 0.00095266, 0.00003347, -0.00062822, -0.00077504, -0.00050239, -0.00011173, 0.00013161, 0.00014870, 0.00002926, -0.00009005, -0.00013252, -0.00010454, -0.00005239, -0.00001241, 0.00000604, 0.00000913, 0.00000402, -0.00000445, -0.00001110, -0.00001112, -0.00000424, },
		{ -0.00000845, -0.00000974, -0.00000430, 0.00000235, 0.00000616, 0.00000738, 0.00000744, 0.00000538, -0.00000039, -0.00000803, -0.00001296, -0.00001434, -0.00001818, -0.00002949, -0.00004058, -0.00003241, 0.00000475, 0.00005026, 0.00006039, 0.00000846, -0.00007691, -0.00012141, -0.00006890, 0.00005417, 0.00014339, 0.00010374, -0.00005203, -0.00019350, -0.00018006, 0.00000038, 0.00019591, 0.00021225, -0.00000227, -0.00028050, -0.00036150, -0.00012775, 0.00025439, 0.00043769, 0.00018040, -0.00041768, -0.00094173, -0.00095444, -0.00033303, 0.00061598, 0.00135784, 0.00149159, 0.00100122, 0.00022539, -0.00040184, -0.00062794, -0.00048637, -0.00019611, 0.00002590, 0.00009054, 0.00003762, -0.00004398, -0.00009211, -0.00009435, -0.00006808, -0.00003425, -0.00000646, 0.00000835, 0.00000879, 0.00000005, },
		{ 0.00000071, 0.00000467, -0.00000045, -0.00000517, -0.00000446, -0.00000064, 0.00000191, 0.00000276, 0.00000448, 0.00000707, 0.00000647, 0.00000058, -0.00000578, -0.00000577, -0.00000172, -0.00000690, -0.00003022, -0.00005786, -0.00005720, -0.00000892, 0.00006141, 0.00009182, 0.00003809, -0.00007247, -0.00014766, -0.00010742, 0.00003241, 0.00015537, 0.00014034, -0.00001991, -0.00019306, -0.00021529, -0.00004640, 0.00017396, 0.00023768, 0.00006063, -0.00021990, -0.00034807, -0.00017878, 0.00016796, 0.00037956, 0.00020859, -0.00029594, -0.00078760, -0.00086864, -0.00039333, 0.00041112, 0.00109819, 0.00130561, 0.00097867, 0.00036394, -0.00019000, -0.00045639, -0.00042523, -0.00023978, -0.00006180, 0.00002451, 0.00002325, -0.00001871, -0.00005851, -0.00007571, -0.00006832, -0.00004452, -0.00001736, },
		{ -0.00005816, -0.00004299, -0.00002239, -0.00000863, -0.00000528, -0.00000622, -0.00000411, 0.00000134, 0.00000474, 0.00000339, 0.00000124, 0.00000299, 0.00000648, 0.00000462, -0.00000371, -0.00000859, -0.00000001, 0.00001518, 0.00001316, -0.00002201, -0.00007133, -0.00008632, -0.00003342, 0.00006123, 0.00011869, 0.00007375, -0.00005431, -0.00016117, -0.00014323, 0.00000059, 0.00015115, 0.00016680, 0.00001690, -0.00017521, -0.00023307, -0.00009005, 0.00013628, 0.00023867, 0.00010881, -0.00015135, -0.00030736, -0.00020199, 0.00008844, 0.00029905, 0.00019726, -0.00020501, -0.00063247, -0.00074498, -0.00039591, 0.00025418, 0.00085107, 0.00108930, 0.00089695, 0.00044190, -0.00001397, -0.00028384, -0.00033303, -0.00024240, -0.00011991, -0.00003493, -0.00000459, -0.00001377, -0.00003814, -0.00005680, },
		{ -0.00002357, -0.00003277, -0.00004410, -0.00004422, -0.00003283, -0.00002025, -0.00001546, -0.00001677, -0.00001517, -0.00000603, 0.00000497, 0.00000898, 0.00000526, 0.00000199, 0.00000495, 0.00000843, 0.00000204, -0.00001281, -0.00001846, 0.00000035, 0.00003215, 0.00003999, -0.00000302, -0.00007574, -0.00011381, -0.00006577, 0.00004760, 0.00013471, 0.00010882, -0.00002509, -0.00015881, -0.00016947, -0.00003536, 0.00013158, 0.00017812, 0.00005103, -0.00014431, -0.00023120, -0.00012277, 0.00009104, 0.00021767, 0.00013632, -0.00008597, -0.00024826, -0.00019712, 0.00002653, 0.00021135, 0.00015600, -0.00014864, -0.00049334, -0.00060492, -0.00035416, 0.00014861, 0.00063627, 0.00086897, 0.00077560, 0.00046277, 0.00011335, -0.00013163, -0.00022837, -0.00021169, -0.00014395, -0.00007601, -0.00003435, },
		{ -0.00013694, -0.00009230, -0.00005537, -0.00003938, -0.00003951, -0.00004064, -0.00003256, -0.00001925, -0.00001268, -0.00001782, -0.00002588, -0.00002384, -0.00000958, 0.00000545, 0.00000965, 0.00000495, 0.00000357, 0.00001072, 0.00001510, 0.00000189, -0.00002416, -0.00003513, -0.00000756, 0.00004416, 0.00006903, 0.00002559, -0.00006759, -0.00013318, -0.00010021, 0.00002182, 0.00013574, 0.00013613, 0.00000955, -0.00014101, -0.00018111, -0.00006837, 0.00010086, 0.00017263, 0.00007596, -0.00010707, -0.00021151, -0.00013965, 0.00004691, 0.00018093, 0.00014198, -0.00003260, -0.00018170, -0.00016940, -0.00001258, 0.00012992, 0.00009814, -0.00012309, -0.00038023, -0.00046788, -0.00028610, 0.00008854, 0.00046431, 0.00066672, 0.00063725, 0.00043882, 0.00018889, -0.00001502, -0.00013010, -0.00016105, },
		{ 0.00005911, -0.00005335, -0.00010605, -0.00010804, -0.00008383, -0.00006057, -0.00005201, -0.00005201, -0.00004543, -0.00002681, -0.00000757, -0.00000354, -0.00001643, -0.00003046, -0.00002883, -0.00001191, 0.00000387, 0.00000582, 0.00000012, 0.00000330, 0.00001868, 0.00002700, 0.00000698, -0.00003387, -0.00005593, -0.00002433, 0.00004695, 0.00009436, 0.00005969, -0.00004643, -0.00013919, -0.00012966, -0.00001143, 0.00012102, 0.00014989, 0.00004244, -0.00011174, -0.00017631, -0.00009212, 0.00006539, 0.00015227, 0.00008772, -0.00007067, -0.00017912, -0.00013964, 0.00001089, 0.00013669, 0.00012904, 0.00000394, -0.00011799, -0.00012762, -0.00002943, 0.00006396, 0.00003734, -0.00011927, -0.00029557, -0.00034790, -0.00021013, 0.00006165, 0.00033629, 0.00049716, 0.00050294, 0.00038739, 0.00021869, },
		{ 0.00038831, 0.00032664, 0.00021518, 0.00009249, -0.00000673, -0.00006059, -0.00006972, -0.00005654, -0.00004813, -0.00005446, -0.00006266, -0.00005411, -0.00002724, -0.00000144, 0.00000275, -0.00001413, -0.00003075, -0.00002813, -0.00001001, 0.00000286, -0.00000140, -0.00001020, -0.00000152, 0.00002573, 0.00004254, 0.00001858, -0.00003805, -0.00007636, -0.00004782, 0.00003848, 0.00011032, 0.00009307, -0.00001522, -0.00012949, -0.00014756, -0.00004530, 0.00009350, 0.00014727, 0.00006676, -0.00007738, -0.00015690, -0.00010273, 0.00003199, 0.00012197, 0.00008571, -0.00004089, -0.00014080, -0.00012532, -0.00001308, 0.00009295, 0.00010381, 0.00002328, -0.00006463, -0.00008159, -0.00002916, 0.00001727, -0.00001538, -0.00012584, -0.00023545, -0.00025205, -0.00014125, 0.00005299, 0.00024565, 0.00036606, },
		{ 0.00004884, 0.00018145, 0.00027123, 0.00030151, 0.00027158, 0.00019347, 0.00009376, 0.00000873, -0.00003393, -0.00003439, -0.00002019, -0.00002150, -0.00004381, -0.00006504, -0.00006066, -0.00003097, -0.00000171, 0.00000291, -0.00001462, -0.00002915, -0.00002195, -0.00000196, 0.00000554, -0.00000916, -0.00002498, -0.00001242, 0.00002841, 0.00005831, 0.00003567, -0.00003419, -0.00009150, -0.00007357, 0.00001965, 0.00011316, 0.00011909, 0.00002031, -0.00010535, -0.00014975, -0.00007253, 0.00005912, 0.00012923, 0.00007793, -0.00004505, -0.00012779, -0.00009972, 0.00000609, 0.00008810, 0.00007259, -0.00002081, -0.00010315, -0.00010189, -0.00002467, 0.00005566, 0.00007372, 0.00002879, -0.00002516, -0.00003971, -0.00001934, -0.00001140, -0.00005347, -0.00013228, -0.00019208, -0.00018069, -0.00008873, },
		{ -0.00015675, -0.00012925, -0.00005553, 0.00003953, 0.00013191, 0.00020490, 0.00024315, 0.00023146, 0.00016748, 0.00007534, -0.00000135, -0.00002904, -0.00001135, 0.00001449, 0.00001194, -0.00002175, -0.00005587, -0.00005953, -0.00003322, -0.00000624, -0.00000396, -0.00002088, -0.00002915, -0.00001269, 0.00001210, 0.00001420, -0.00001383, -0.00004123, -0.00002892, 0.00002417, 0.00007012, 0.00005509, -0.00002209, -0.00009743, -0.00009602, -0.00000577, 0.00010220, 0.00013253, 0.00005316, -0.00007142, -0.00013561, -0.00008747, 0.00002518, 0.00010021, 0.00007476, -0.00002061, -0.00009555, -0.00008589, -0.00000963, 0.00005669, 0.00005322, -0.00001036, -0.00007095, -0.00007536, -0.00002656, 0.00002769, 0.00004529, 0.00002597, 0.00000077, -0.00000736, -0.00000761, -0.00002678, -0.00007536, -0.00013125, },
		{ -0.00003499, -0.00008349, -0.00011969, -0.00012236, -0.00009075, -0.00003921, 0.00002076, 0.00008753, 0.00015698, 0.00020790, 0.00020890, 0.00014708, 0.00004996, -0.00002705, -0.00004285, -0.00000486, 0.00003875, 0.00004325, 0.00000621, -0.00003587, -0.00004757, -0.00002913, -0.00001086, -0.00001597, -0.00003390, -0.00003397, -0.00000415, 0.00002978, 0.00002925, -0.00001232, -0.00005485, -0.00004841, 0.00001241, 0.00007434, 0.00007236, -0.00000411, -0.00009252, -0.00011014, -0.00003218, 0.00008001, 0.00013098, 0.00007695, -0.00003448, -0.00010825, -0.00008758, -0.00000165, 0.00006681, 0.00005969, -0.00000718, -0.00006647, -0.00006634, -0.00001566, 0.00003190, 0.00003306, -0.00000688, -0.00004630, -0.00005095, -0.00002312, 0.00000882, 0.00002274, 0.00002056, 0.00001638, 0.00001375, 0.00000030, },
		{ 0.00002609, 0.00002502, 0.00000161, -0.00004147, -0.00008265, -0.00009854, -0.00008503, -0.00005842, -0.00003409, -0.00000669, 0.00004292, 0.00011815, 0.00018721, 0.00020083, 0.00013661, 0.00002761, -0.00005724, -0.00006786, -0.00001373, 0.00004786, 0.00006463, 0.00003264, -0.00000944, -0.00002512, -0.00001563, -0.00001087, -0.00002906, -0.00005206, -0.00004514, -0.00000014, 0.00004708, 0.00004879, -0.00000322, -0.00006190, -0.00006674, -0.00000511, 0.00006912, 0.00008305, 0.00001550, -0.00007793, -0.00011284, -0.00005377, 0.00005175, 0.00011552, 0.00008764, -0.00000160, -0.00007362, -0.00007407, -0.00001713, 0.00003597, 0.00003791, -0.00000446, -0.00004494, -0.00004678, -0.00001507, 0.00001511, 0.00001657, -0.00000640, -0.00002861, -0.00003179, -0.00001882, -0.00000342, 0.00000747, 0.00001696, },
		{ -0.00001236, -0.00000157, 0.00001727, 0.00003389, 0.00002971, -0.00000350, -0.00004961, -0.00007798, -0.00007154, -0.00004445, -0.00002762, -0.00003369, -0.00003843, -0.00000277, 0.00008208, 0.00017230, 0.00020058, 0.00013514, 0.00001373, -0.00008254, -0.00009483, -0.00003247, 0.00004195, 0.00007148, 0.00005055, 0.00001688, 0.00000405, 0.00000734, -0.00000285, -0.00003833, -0.00007151, -0.00006177, -0.00000314, 0.00005961, 0.00006905, 0.00001253, -0.00005989, -0.00007944, -0.00002468, 0.00005508, 0.00008415, 0.00003187, -0.00005728, -0.00010388, -0.00006611, 0.00002358, 0.00009032, 0.00008452, 0.00002174, -0.00003882, -0.00005139, -0.00002046, 0.00001300, 0.00001568, -0.00000916, -0.00003236, -0.00003185, -0.00001227, 0.00000505, 0.00000603, -0.00000512, -0.00001551, -0.00001863, -0.00001687, },
		{ -0.00000410, -0.00001016, -0.00001849, -0.00002086, -0.00000681, 0.00002132, 0.00004224, 0.00003143, -0.00001270, -0.00006023, -0.00007337, -0.00004344, -0.00000314, 0.00000335, -0.00003274, -0.00006869, -0.00004687, 0.00004635, 0.00015653, 0.00020060, 0.00013799, 0.00000889, -0.00009733, -0.00011545, -0.00005335, 0.00002537, 0.00006349, 0.00005548, 0.00003644, 0.00003384, 0.00003663, 0.00001405, -0.00003970, -0.00008726, -0.00008065, -0.00001336, 0.00006396, 0.00008544, 0.00003181, -0.00004865, -0.00008313, -0.00004167, 0.00003522, 0.00007506, 0.00004098, -0.00003554, -0.00008588, -0.00006742, 0.00000103, 0.00006137, 0.00007028, 0.00003307, -0.00001014, -0.00002587, -0.00001427, 0.00000027, -0.00000150, -0.00001628, -0.00002714, -0.00002379, -0.00001134, -0.00000129, 0.00000128, -0.00000067, },
		{ -0.00000729, 0.00000022, 0.00000724, 0.00000779, -0.00000409, -0.00002297, -0.00003033, -0.00001080, 0.00002732, 0.00005161, 0.00003287, -0.00002284, -0.00007204, -0.00007067, -0.00001836, 0.00003501, 0.00003344, -0.00002849, -0.00009235, -0.00008491, 0.00001204, 0.00013672, 0.00019480, 0.00013908, 0.00000983, -0.00010049, -0.00012452, -0.00006900, 0.00000496, 0.00004441, 0.00004670, 0.00004455, 0.00005766, 0.00006669, 0.00003764, -0.00003136, -0.00009475, -0.00009708, -0.00002849, 0.00005889, 0.00009395, 0.00005018, -0.00003046, -0.00007653, -0.00005196, 0.00001403, 0.00005776, 0.00004089, -0.00001752, -0.00006343, -0.00005887, -0.00001256, 0.00003475, 0.00005004, 0.00003331, 0.00000857, -0.00000380, -0.00000356, -0.00000322, -0.00001047, -0.00002085, -0.00002556, -0.00002207, -0.00001467, },
		{ -0.00002336, -0.00002389, -0.00002226, -0.00001527, -0.00000005, 0.00001705, 0.00002076, 0.00000167, -0.00002841, -0.00004056, -0.00001522, 0.00003296, 0.00006087, 0.00003511, -0.00003122, -0.00008265, -0.00006964, 0.00000126, 0.00006628, 0.00005997, -0.00002081, -0.00010639, -0.00011258, -0.00001765, 0.00011311, 0.00018011, 0.00013317, 0.00001139, -0.00009476, -0.00012086, -0.00007442, -0.00001223, 0.00002060, 0.00002731, 0.00003977, 0.00007041, 0.00009115, 0.00006321, -0.00001439, -0.00009134, -0.00010639, -0.00004445, 0.00004571, 0.00009243, 0.00006325, -0.00000939, -0.00006094, -0.00005317, -0.00000384, 0.00003624, 0.00003227, -0.00000647, -0.00004167, -0.00004412, -0.00001674, 0.00001496, 0.00002970, 0.00002625, 0.00001672, 0.00001031, 0.00000597, -0.00000115, -0.00001120, -0.00001958, },
		{ 0.00000152, -0.00000644, -0.00001187, -0.00001726, -0.00002544, -0.00003190, -0.00002565, -0.00000191, 0.00002618, 0.00003378, 0.00000814, -0.00003271, -0.00005007, -0.00002037, 0.00003642, 0.00006811, 0.00003769, -0.00003643, -0.00008972, -0.00006875, 0.00001498, 0.00008830, 0.00007994, -0.00001157, -0.00011038, -0.00012725, -0.00003911, 0.00008845, 0.00015680, 0.00011751, 0.00000864, -0.00008512, -0.00010708, -0.00006820, -0.00002096, -0.00000092, 0.00000312, 0.00002429, 0.00006987, 0.00010469, 0.00008491, 0.00000747, -0.00007718, -0.00010539, -0.00005660, 0.00002789, 0.00008122, 0.00006799, 0.00000992, -0.00003980, -0.00004540, -0.00001498, 0.00001527, 0.00001811, -0.00000319, -0.00002478, -0.00002799, -0.00001383, 0.00000382, 0.00001426, 0.00001709, 0.00001679, 0.00001492, 0.00000979, },
		{ 0.00001323, 0.00001161, 0.00000584, 0.00000033, -0.00000062, 0.00000015, -0.00000625, -0.00002344, -0.00004015, -0.00003681, -0.00000624, 0.00003223, 0.00004473, 0.00001489, -0.00003456, -0.00005698, -0.00002528, 0.00003705, 0.00007166, 0.00003916, -0.00003857, -0.00009193, -0.00006611, 0.00002394, 0.00010060, 0.00009134, -0.00000340, -0.00010613, -0.00012869, -0.00004985, 0.00006645, 0.00012784, 0.00009240, -0.00000145, -0.00007688, -0.00008821, -0.00005249, -0.00001913, -0.00001453, -0.00001912, 0.00000309, 0.00005730, 0.00010457, 0.00009746, 0.00002889, -0.00005521, -0.00009343, -0.00006126, 0.00000998, 0.00006307, 0.00006369, 0.00002365, -0.00001769, -0.00003107, -0.00001814, -0.00000103, 0.00000271, -0.00000602, -0.00001481, -0.00001494, -0.00000786, 0.00000031, 0.00000639, 0.00001077, },
		{ 0.00000124, 0.00000594, 0.00001041, 0.00001079, 0.00000427, -0.00000491, -0.00000714, 0.00000180, 0.00001243, 0.00000813, -0.00001638, -0.00004371, -0.00004577, -0.00001206, 0.00003398, 0.00005137, 0.00002035, -0.00003389, -0.00005990, -0.00002839, 0.00003544, 0.00007072, 0.00003797, -0.00003891, -0.00008935, -0.00006045, 0.00003002, 0.00010444, 0.00009374, 0.00000146, -0.00009668, -0.00011897, -0.00004943, 0.00005025, 0.00009756, 0.00006089, -0.00001916, -0.00007390, -0.00006995, -0.00003198, -0.00000805, -0.00001741, -0.00003366, -0.00001758, 0.00003688, 0.00009132, 0.00009772, 0.00004466, -0.00003026, -0.00007262, -0.00005688, -0.00000385, 0.00004224, 0.00005206, 0.00002992, 0.00000111, -0.00001417, -0.00001449, -0.00001037, -0.00000971, -0.00001154, -0.00001131, -0.00000775, -0.00000309, },
		{ -0.00001177, -0.00000681, -0.00000263, 0.00000261, 0.00001037, 0.00001648, 0.00001295, -0.00000239, -0.00001899, -0.00002053, -0.00000206, 0.00002053, 0.00002232, -0.00000519, -0.00004072, -0.00004941, -0.00001693, 0.00003190, 0.00005240, 0.00002262, -0.00003184, -0.00005853, -0.00002833, 0.00003301, 0.00006573, 0.00003321, -0.00003915, -0.00008329, -0.00005166, 0.00003493, 0.00010213, 0.00008838, 0.00000195, -0.00008519, -0.00010165, -0.00003955, 0.00004136, 0.00007014, 0.00002753, -0.00004222, -0.00007762, -0.00005692, -0.00001225, 0.00000837, -0.00001000, -0.00003724, -0.00003207, 0.00001438, 0.00006849, 0.00008561, 0.00005115, -0.00000764, -0.00004726, -0.00004457, -0.00001110, 0.00002312, 0.00003663, 0.00002924, 0.00001382, 0.00000107, -0.00000706, -0.00001279, -0.00001631, -0.00001589, },
		{ -0.00001041, -0.00001641, -0.00001608, -0.00001273, -0.00001027, -0.00000741, 0.00000107, 0.00001580, 0.00002691, 0.00002081, -0.00000446, -0.00003172, -0.00003690, -0.00001249, 0.00002119, 0.00003202, 0.00000712, -0.00003158, -0.00004572, -0.00001798, 0.00002803, 0.00004804, 0.00002036, -0.00003016, -0.00005376, -0.00002455, 0.00003128, 0.00005806, 0.00002494, -0.00004071, -0.00007575, -0.00004080, 0.00003959, 0.00009617, 0.00007769, -0.00000150, -0.00007401, -0.00008072, -0.00002345, 0.00003940, 0.00004849, -0.00000295, -0.00006665, -0.00008696, -0.00005154, 0.00000187, 0.00002504, 0.00000452, -0.00002973, -0.00003684, -0.00000455, 0.00004153, 0.00006401, 0.00004744, 0.00000849, -0.00002248, -0.00002766, -0.00001166, 0.00000891, 0.00002154, 0.00002409, 0.00001990, 0.00001153, 0.00000039, },
		{ 0.00002091, 0.00001603, 0.00000473, -0.00000649, -0.00001153, -0.00001101, -0.00001112, -0.00001491, -0.00001610, -0.00000487, 0.00001858, 0.00003797, 0.00003286, -0.00000015, -0.00003884, -0.00005173, -0.00002716, 0.00001348, 0.00003373, 0.00001645, -0.00001889, -0.00003472, -0.00001306, 0.00002518, 0.00004008, 0.00001349, -0.00003039, -0.00004729, -0.00001758, 0.00003128, 0.00004945, 0.00001424, -0.00004420, -0.00006870, -0.00002968, 0.00004388, 0.00008855, 0.00006449, -0.00000732, -0.00006422, -0.00005960, -0.00000501, 0.00004235, 0.00003370, -0.00002691, -0.00008809, -0.00009901, -0.00005371, 0.00000756, 0.00003729, 0.00002149, -0.00001379, -0.00003126, -0.00001599, 0.00001615, 0.00003789, 0.00003547, 0.00001628, -0.00000274, -0.00001070, -0.00000773, 0.00000068, 0.00001014, 0.00001797, },
		{ -0.00000280, 0.00000398, 0.00001407, 0.00001975, 0.00001555, 0.00000449, -0.00000413, -0.00000460, -0.00000172, -0.00000540, -0.00001749, -0.00002573, -0.00001448, 0.00001657, 0.00004574, 0.00004569, 0.00000972, -0.00003788, -0.00006053, -0.00004183, -0.00000078, 0.00002603, 0.00001908, -0.00000660, -0.00001863, -0.00000172, 0.00002580, 0.00003129, 0.00000342, -0.00003311, -0.00004088, -0.00000882, 0.00003316, 0.00004145, 0.00000269, -0.00004938, -0.00006355, -0.00002029, 0.00004689, 0.00008035, 0.00005125, -0.00001340, -0.00005572, -0.00004060, 0.00001225, 0.00004739, 0.00002516, -0.00004242, -0.00010300, -0.00011015, -0.00006128, 0.00000443, 0.00004213, 0.00003631, 0.00000617, -0.00001749, -0.00001863, -0.00000327, 0.00001278, 0.00001934, 0.00001656, 0.00000946, 0.00000209, -0.00000293, },
		{ 0.00001420, 0.00000809, -0.00000099, -0.00000441, 0.00000238, 0.00001407, 0.00001943, 0.00001264, 0.00000023, -0.00000514, 0.00000116, 0.00000909, 0.00000394, -0.00001592, -0.00003288, -0.00002515, 0.00000979, 0.00004746, 0.00005524, 0.00002213, -0.00002902, -0.00006032, -0.00005184, -0.00001763, 0.00001019, 0.00001303, 0.00000122, -0.00000128, 0.00001439, 0.00003101, 0.00002444, -0.00000742, -0.00003769, -0.00003571, 0.00000006, 0.00003625, 0.00003490, -0.00000813, -0.00005536, -0.00006087, -0.00001421, 0.00004740, 0.00007180, 0.00003964, -0.00001778, -0.00004770, -0.00002476, 0.00002583, 0.00005176, 0.00002112, -0.00004939, -0.00010941, -0.00011720, -0.00007110, -0.00000574, 0.00003887, 0.00004571, 0.00002550, 0.00000048, -0.00001385, -0.00001470, -0.00000679, 0.00000400, 0.00001245, },
		{ -0.00001841, -0.00000634, 0.00000812, 0.00001387, 0.00000708, -0.00000437, -0.00000749, 0.00000285, 0.00001762, 0.00002185, 0.00001032, -0.00000573, -0.00000941, 0.00000353, 0.00001814, 0.00001457, -0.00001008, -0.00003500, -0.00003348, 0.00000036, 0.00004250, 0.00005827, 0.00003299, -0.00001487, -0.00005061, -0.00005362, -0.00003219, -0.00001014, -0.00000118, 0.00000188, 0.00001307, 0.00003192, 0.00004006, 0.00002127, -0.00001638, -0.00004244, -0.00003191, 0.00000778, 0.00003948, 0.00002992, -0.00001712, -0.00006106, -0.00006055, -0.00001230, 0.00004437, 0.00006262, 0.00003043, -0.00001915, -0.00003919, -0.00001214, 0.00003450, 0.00005346, 0.00001956, -0.00004908, -0.00010712, -0.00011826, -0.00008008, -0.00001992, 0.00002893, 0.00004848, 0.00004058, 0.00001813, -0.00000498, -0.00001873, },
		{ 0.00003180, 0.00000398, -0.00001795, -0.00002225, -0.00000935, 0.00000736, 0.00001216, 0.00000115, -0.00001339, -0.00001448, 0.00000216, 0.00002247, 0.00002701, 0.00001082, -0.00001036, -0.00001499, 0.00000205, 0.00002280, 0.00002327, -0.00000179, -0.00003134, -0.00003656, -0.00000840, 0.00003247, 0.00005349, 0.00003856, 0.00000039, -0.00003355, -0.00004588, -0.00004022, -0.00003003, -0.00002056, -0.00000510, 0.00002107, 0.00004685, 0.00005052, 0.00002194, -0.00002150, -0.00004532, -0.00002862, 0.00001385, 0.00004195, 0.00002615, -0.00002380, -0.00006565, -0.00006207, -0.00001468, 0.00003730, 0.00005242, 0.00002368, -0.00001700, -0.00002953, -0.00000233, 0.00003811, 0.00005155, 0.00001880, -0.00004340, -0.00009736, -0.00011301, -0.00008604, -0.00003489, 0.00001508, 0.00004546, 0.00004966, },
		{ 0.00000046, 0.00003884, 0.00005297, 0.00003953, 0.00000967, -0.00001588, -0.00002055, -0.00000516, 0.00001232, 0.00001271, -0.00000608, -0.00002607, -0.00002596, -0.00000238, 0.00002539, 0.00003304, 0.00001466, -0.00001108, -0.00001892, -0.00000178, 0.00002212, 0.00002718, 0.00000588, -0.00002326, -0.00003307, -0.00001322, 0.00002064, 0.00004199, 0.00003677, 0.00001248, -0.00001324, -0.00003003, -0.00003944, -0.00004469, -0.00004052, -0.00001772, 0.00002139, 0.00005571, 0.00005904, 0.00002492, -0.00002226, -0.00004476, -0.00002455, 0.00001867, 0.00004334, 0.00002320, -0.00002819, -0.00006873, -0.00006490, -0.00002096, 0.00002627, 0.00004101, 0.00001907, -0.00001159, -0.00001866, 0.00000516, 0.00003723, 0.00004610, 0.00001790, -0.00003424, -0.00008217, -0.00010244, -0.00008815, -0.00004809, },
		{ -0.00008836, -0.00008678, -0.00005816, -0.00001237, 0.00003130, 0.00005215, 0.00004131, 0.00001029, -0.00001556, -0.00001666, 0.00000393, 0.00002283, 0.00001776, -0.00001083, -0.00003883, -0.00004025, -0.00001177, 0.00002344, 0.00003681, 0.00002041, -0.00000688, -0.00001842, -0.00000517, 0.00001713, 0.00002483, 0.00000979, -0.00001369, -0.00002384, -0.00001212, 0.00001066, 0.00002678, 0.00002786, 0.00001837, 0.00000562, -0.00000966, -0.00003019, -0.00005099, -0.00005624, -0.00003216, 0.00001505, 0.00005676, 0.00006257, 0.00002764, -0.00001969, -0.00004034, -0.00001863, 0.00002315, 0.00004405, 0.00002104, -0.00003043, -0.00007025, -0.00006868, -0.00003055, 0.00001186, 0.00002856, 0.00001621, -0.00000357, -0.00000699, 0.00001056, 0.00003268, 0.00003784, 0.00001667, -0.00002305, -0.00006373, },
		{ 0.00001542, -0.00001071, -0.00004389, -0.00007270, -0.00008303, -0.00006484, -0.00002192, 0.00002504, 0.00004943, 0.00003861, 0.00000591, -0.00001867, -0.00001367, 0.00001446, 0.00003658, 0.00002739, -0.00001050, -0.00004770, -0.00005393, -0.00002479, 0.00001516, 0.00003521, 0.00002514, 0.00000123, -0.00001203, -0.00000508, 0.00001053, 0.00001682, 0.00000791, -0.00000610, -0.00001156, -0.00000514, 0.00000535, 0.00001179, 0.00001426, 0.00001721, 0.00001918, 0.00001055, -0.00001521, -0.00004824, -0.00006416, -0.00004401, 0.00000508, 0.00005050, 0.00005943, 0.00002752, -0.00001595, -0.00003308, -0.00001071, 0.00002810, 0.00004492, 0.00002014, -0.00003045, -0.00007028, -0.00007319, -0.00004281, -0.00000518, 0.00001548, 0.00001483, 0.00000622, 0.00000474, 0.00001387, 0.00002528, 0.00002778, },
		},
		{ { -0.00000338, -0.00000484, -0.00000683, -0.00000448, 0.00000483, 0.00001796, 0.00002699, 0.00002430, 0.00000823, -0.00001480, -0.00003369, -0.00003934, -0.00003004, -0.00001194, 0.00000560, 0.00001636, 0.00002057, 0.00002304, 0.00002785, 0.00003389, 0.00003510, 0.00002546, 0.00000490, -0.00001851, -0.00003271, -0.00002945, -0.00001068, 0.00001163, 0.00002254, 0.00001381, -0.00001033, -0.00003544, -0.00004602, -0.00003544, -0.00001033, 0.00001381, 0.00002254, 0.00001163, -0.00001068, -0.00002945, -0.00003271, -0.00001851, 0.00000490, 0.00002546, 0.00003510, 0.00003389, 0.00002785, 0.00002304, 0.00002057, 0.00001636, 0.00000560, -0.00001194, -0.00003004, -0.00003934, -0.00003369, -0.00001480, 0.00000823, 0.00002430, 0.00002699, 0.00001796, 0.00000483, -0.00000448, -0.00000683, -0.00000484, },
		{ -0.00000293, -0.00000508, -0.00000287, -0.00000124, -0.00000351, -0.00000802, -0.00000917, -0.00000215, 0.00001240, 0.00002742, 0.00003342, 0.00002474, 0.00000359, -0.00002098, -0.00003855, -0.00004347, -0.00003719, -0.00002563, -0.00001382, -0.00000220, 0.00001237, 0.00003192, 0.00005274, 0.00006535, 0.00005992, 0.00003374, -0.00000462, -0.00003837, -0.00005159, -0.00003886, -0.00000908, 0.00001949, 0.00003012, 0.00001748, -0.00000963, -0.00003442, -0.00004294, -0.00003237, -0.00001171, 0.00000516, 0.00000893, 0.00000036, -0.00001149, -0.00001649, -0.00001061, 0.00000193, 0.00001264, 0.00001554, 0.00001143, 0.00000670, 0.00000772, 0.00001535, 0.00002381, 0.00002490, 0.00001438, -0.00000417, -0.00002110, -0.00002694, -0.00001865, -0.00000162, 0.00001404, 0.00002018, 0.00001545, 0.00000526, },
		{ 0.00002005, 0.00001398, 0.00000207, -0.00000770, -0.00000994, -0.00000548, -0.00000006, 0.00000071, -0.00000389, -0.00000890, -0.00000759, 0.00000298, 0.00001894, 0.00003155, 0.00003287, 0.00002095, 0.00000086, -0.00001905, -0.00003269, -0.00003937, -0.00004220, -0.00004333, -0.00004025, -0.00002683, 0.00000123, 0.00004018, 0.00007682, 0.00009400, 0.00008051, 0.00003919, -0.00001273, -0.00005205, -0.00006223, -0.00004242, -0.00000738, 0.00002169, 0.00002963, 0.00001526, -0.00000976, -0.00002968, -0.00003481, -0.00002637, -0.00001365, -0.00000613, -0.00000673, -0.00001076, -0.00001076, -0.00000311, 0.00000880, 0.00001711, 0.00001578, 0.00000555, -0.00000616, -0.00001050, -0.00000372, 0.00000994, 0.00002107, 0.00002187, 0.00001168, -0.00000250, -0.00001097, -0.00000818, 0.00000351, 0.00001571, },
		{ -0.00000046, 0.00001338, 0.00002783, 0.00003322, 0.00002522, 0.00000787, -0.00000930, -0.00001775, -0.00001540, -0.00000723, -0.00000086, -0.00000061, -0.00000448, -0.00000626, -0.00000088, 0.00001115, 0.00002378, 0.00002963, 0.00002533, 0.00001359, 0.00000047, -0.00000977, -0.00001805, -0.00002892, -0.00004502, -0.00006191, -0.00006822, -0.00005209, -0.00001039, 0.00004581, 0.00009441, 0.00011316, 0.00009233, 0.00004107, -0.00001685, -0.00005577, -0.00006180, -0.00003916, -0.00000588, 0.00001793, 0.00002154, 0.00000792, -0.00001087, -0.00002308, -0.00002461, -0.00001978, -0.00001598, -0.00001688, -0.00001968, -0.00001804, -0.00000811, 0.00000737, 0.00002022, 0.00002257, 0.00001281, -0.00000282, -0.00001431, -0.00001477, -0.00000511, 0.00000690, 0.00001256, 0.00000863, -0.00000031, -0.00000543, },
		{ -0.00000029, -0.00001090, -0.00001690, -0.00001017, 0.00000952, 0.00003333, 0.00004855, 0.00004664, 0.00002860, 0.00000403, -0.00001522, -0.00002233, -0.00001873, -0.00001138, -0.00000671, -0.00000596, -0.00000524, 0.00000011, 0.00001064, 0.00002163, 0.00002655, 0.00002244, 0.00001266, 0.00000440, 0.00000249, 0.00000451, 0.00000148, -0.00001528, -0.00004578, -0.00007777, -0.00009174, -0.00007250, -0.00002013, 0.00004715, 0.00010109, 0.00011823, 0.00009269, 0.00003916, -0.00001581, -0.00004857, -0.00005036, -0.00002965, -0.00000436, 0.00001000, 0.00000866, -0.00000238, -0.00001308, -0.00001732, -0.00001622, -0.00001529, -0.00001844, -0.00002396, -0.00002582, -0.00001886, -0.00000382, 0.00001223, 0.00002045, 0.00001659, 0.00000416, -0.00000798, -0.00001199, -0.00000668, 0.00000190, 0.00000535, },
		{ -0.00000076, 0.00000979, 0.00001313, 0.00000361, -0.00001505, -0.00003110, -0.00003207, -0.00001334, 0.00001811, 0.00004736, 0.00006044, 0.00005242, 0.00002948, 0.00000399, -0.00001355, -0.00002011, -0.00001974, -0.00001821, -0.00001762, -0.00001501, -0.00000612, 0.00000927, 0.00002518, 0.00003281, 0.00002720, 0.00001203, -0.00000190, -0.00000420, 0.00000666, 0.00002050, 0.00002118, -0.00000200, -0.00004481, -0.00008770, -0.00010582, -0.00008382, -0.00002626, 0.00004338, 0.00009495, 0.00010785, 0.00008146, 0.00003378, -0.00001018, -0.00003256, -0.00003086, -0.00001585, -0.00000223, 0.00000118, -0.00000469, -0.00001268, -0.00001640, -0.00001513, -0.00001316, -0.00001498, -0.00002061, -0.00002517, -0.00002299, -0.00001267, 0.00000092, 0.00000990, 0.00000923, 0.00000081, -0.00000772, -0.00000873, },
		{ -0.00001055, -0.00002002, -0.00001882, -0.00000447, 0.00001566, 0.00002828, 0.00002291, -0.00000018, -0.00002887, -0.00004600, -0.00003993, -0.00001193, 0.00002458, 0.00005227, 0.00006006, 0.00004855, 0.00002754, 0.00000823, -0.00000406, -0.00001137, -0.00001858, -0.00002705, -0.00003205, -0.00002631, -0.00000706, 0.00001939, 0.00003984, 0.00004255, 0.00002596, 0.00000120, -0.00001454, -0.00000973, 0.00001249, 0.00003485, 0.00003663, 0.00000785, -0.00004246, -0.00008972, -0.00010770, -0.00008410, -0.00002809, 0.00003476, 0.00007705, 0.00008428, 0.00006104, 0.00002569, -0.00000209, -0.00001240, -0.00000832, -0.00000062, 0.00000126, -0.00000479, -0.00001400, -0.00002008, -0.00002062, -0.00001825, -0.00001737, -0.00001954, -0.00002190, -0.00001985, -0.00001177, -0.00000142, 0.00000418, 0.00000039, },
		{ 0.00000445, 0.00001076, 0.00000366, -0.00001472, -0.00003350, -0.00003912, -0.00002459, 0.00000455, 0.00003228, 0.00004162, 0.00002561, -0.00000756, -0.00003904, -0.00005076, -0.00003629, -0.00000411, 0.00002852, 0.00004677, 0.00004661, 0.00003505, 0.00002290, 0.00001597, 0.00001145, 0.00000206, -0.00001578, -0.00003642, -0.00004728, -0.00003767, -0.00000798, 0.00002816, 0.00005130, 0.00004867, 0.00002325, -0.00000727, -0.00002185, -0.00001025, 0.00001933, 0.00004493, 0.00004490, 0.00001243, -0.00003909, -0.00008353, -0.00009754, -0.00007413, -0.00002605, 0.00002271, 0.00005130, 0.00005287, 0.00003586, 0.00001598, 0.00000536, 0.00000606, 0.00001143, 0.00001290, 0.00000660, -0.00000490, -0.00001589, -0.00002256, -0.00002529, -0.00002679, -0.00002833, -0.00002778, -0.00002160, -0.00000919, },
		{ -0.00004299, -0.00003679, -0.00001903, 0.00000403, 0.00002021, 0.00001852, -0.00000271, -0.00003261, -0.00005299, -0.00004942, -0.00002113, 0.00001724, 0.00004398, 0.00004370, 0.00001716, -0.00001917, -0.00004362, -0.00004288, -0.00001972, 0.00001022, 0.00002980, 0.00003175, 0.00002216, 0.00001403, 0.00001605, 0.00002531, 0.00002952, 0.00001712, -0.00001209, -0.00004439, -0.00006001, -0.00004663, -0.00000883, 0.00003335, 0.00005665, 0.00004950, 0.00001958, -0.00001129, -0.00002172, -0.00000496, 0.00002654, 0.00004951, 0.00004517, 0.00001162, -0.00003499, -0.00007070, -0.00007839, -0.00005723, -0.00002155, 0.00000946, 0.00002348, 0.00002081, 0.00001132, 0.00000592, 0.00000909, 0.00001731, 0.00002331, 0.00002205, 0.00001371, 0.00000210, -0.00000927, -0.00001962, -0.00002976, -0.00003892, },
		{ 0.00000425, -0.00001241, -0.00003323, -0.00005118, -0.00005634, -0.00004257, -0.00001368, 0.00001612, 0.00002944, 0.00001677, -0.00001605, -0.00004973, -0.00006283, -0.00004549, -0.00000646, 0.00003171, 0.00004679, 0.00003108, -0.00000373, -0.00003445, -0.00004180, -0.00002286, 0.00000789, 0.00002933, 0.00002869, 0.00001022, -0.00000867, -0.00001075, 0.00000778, 0.00003369, 0.00004580, 0.00003032, -0.00000834, -0.00004903, -0.00006722, -0.00005109, -0.00000955, 0.00003341, 0.00005414, 0.00004433, 0.00001542, -0.00001003, -0.00001391, 0.00000559, 0.00003353, 0.00004892, 0.00003892, 0.00000692, -0.00003038, -0.00005425, -0.00005543, -0.00003840, -0.00001656, -0.00000241, -0.00000007, -0.00000459, -0.00000741, -0.00000323, 0.00000691, 0.00001773, 0.00002431, 0.00002522, 0.00002179, 0.00001513, },
		{ 0.00001434, 0.00002226, 0.00002941, 0.00003097, 0.00002068, -0.00000353, -0.00003496, -0.00005936, -0.00006272, -0.00004110, -0.00000541, 0.00002344, 0.00002711, 0.00000224, -0.00003598, -0.00006261, -0.00005889, -0.00002511, 0.00001874, 0.00004552, 0.00003879, 0.00000395, -0.00003505, -0.00005157, -0.00003413, 0.00000584, 0.00004175, 0.00004992, 0.00002569, -0.00001344, -0.00003985, -0.00003471, -0.00000097, 0.00003847, 0.00005615, 0.00003847, -0.00000526, -0.00004898, -0.00006691, -0.00004987, -0.00001010, 0.00002788, 0.00004370, 0.00003366, 0.00001115, -0.00000412, -0.00000020, 0.00001952, 0.00003982, 0.00004497, 0.00002946, 0.00000103, -0.00002542, -0.00003801, -0.00003475, -0.00002305, -0.00001308, -0.00001057, -0.00001403, -0.00001767, -0.00001664, -0.00001038, -0.00000169, 0.00000667, },
		{ -0.00001497, -0.00001478, -0.00001538, -0.00001330, -0.00000369, 0.00001451, 0.00003483, 0.00004501, 0.00003464, 0.00000374, -0.00003439, -0.00005957, -0.00005736, -0.00002954, 0.00000542, 0.00002341, 0.00001082, -0.00002525, -0.00006039, -0.00006859, -0.00004030, 0.00000975, 0.00005083, 0.00005601, 0.00002057, -0.00003372, -0.00007178, -0.00006779, -0.00002232, 0.00003751, 0.00007515, 0.00006796, 0.00002150, -0.00003442, -0.00006523, -0.00005352, -0.00000930, 0.00003765, 0.00005761, 0.00003954, -0.00000334, -0.00004377, -0.00005869, -0.00004299, -0.00001045, 0.00001760, 0.00002708, 0.00001916, 0.00000698, 0.00000447, 0.00001601, 0.00003396, 0.00004507, 0.00004034, 0.00002105, -0.00000289, -0.00002025, -0.00002568, -0.00002184, -0.00001568, -0.00001265, -0.00001348, -0.00001536, -0.00001578, },
		{ -0.00001588, -0.00001075, -0.00000404, 0.00000007, -0.00000361, -0.00001616, -0.00003115, -0.00003698, -0.00002459, 0.00000452, 0.00003654, 0.00005254, 0.00004088, 0.00000632, -0.00003134, -0.00004943, -0.00003792, -0.00000718, 0.00001798, 0.00001536, -0.00001818, -0.00006233, -0.00008555, -0.00006639, -0.00000985, 0.00005310, 0.00008381, 0.00006082, -0.00000488, -0.00007501, -0.00010692, -0.00008001, -0.00000887, 0.00006556, 0.00010108, 0.00007956, 0.00001684, -0.00004859, -0.00008000, -0.00006406, -0.00001649, 0.00003043, 0.00004923, 0.00003316, -0.00000272, -0.00003412, -0.00004399, -0.00003182, -0.00001060, 0.00000461, 0.00000754, 0.00000345, 0.00000300, 0.00001290, 0.00003042, 0.00004568, 0.00004904, 0.00003786, 0.00001773, -0.00000213, -0.00001504, -0.00002004, -0.00002024, -0.00001870, },
		{ -0.00001006, -0.00002224, -0.00003068, -0.00003187, -0.00002232, -0.00000323, 0.00001690, 0.00002546, 0.00001388, -0.00001480, -0.00004517, -0.00005784, -0.00004213, -0.00000455, 0.00003376, 0.00005019, 0.00003587, 0.00000229, -0.00002605, -0.00002888, -0.00000535, 0.00002402, 0.00003106, 0.00000087, -0.00005486, -0.00010238, -0.00010681, -0.00005643, 0.00002639, 0.00009597, 0.00011044, 0.00005829, -0.00003287, -0.00011165, -0.00013314, -0.00008553, 0.00000348, 0.00008404, 0.00011388, 0.00008201, 0.00001233, -0.00005352, -0.00008189, -0.00006518, -0.00002215, 0.00001747, 0.00003242, 0.00002077, -0.00000323, -0.00002180, -0.00002586, -0.00001880, -0.00001057, -0.00000829, -0.00001076, -0.00001048, -0.00000078, 0.00001814, 0.00003884, 0.00005176, 0.00005154, 0.00003970, 0.00002222, 0.00000482, },
		{ 0.00005244, 0.00004672, 0.00003519, 0.00001783, -0.00000561, -0.00003159, -0.00005089, -0.00005229, -0.00003052, 0.00000722, 0.00004213, 0.00005361, 0.00003246, -0.00001151, -0.00005377, -0.00006975, -0.00005068, -0.00000967, 0.00002685, 0.00003708, 0.00001899, -0.00000838, -0.00001911, -0.00000039, 0.00003596, 0.00005973, 0.00004334, -0.00001638, -0.00009137, -0.00013692, -0.00011963, -0.00004086, 0.00006035, 0.00012842, 0.00012393, 0.00004788, -0.00005813, -0.00013597, -0.00014460, -0.00008280, 0.00001240, 0.00008910, 0.00011074, 0.00007452, 0.00000839, -0.00004904, -0.00007171, -0.00005801, -0.00002621, 0.00000093, 0.00001074, 0.00000536, -0.00000439, -0.00000934, -0.00000836, -0.00000694, -0.00001037, -0.00001812, -0.00002369, -0.00001982, -0.00000439, 0.00001775, 0.00003818, 0.00005028, },
		{ -0.00000779, 0.00001049, 0.00002719, 0.00004074, 0.00005161, 0.00005824, 0.00005494, 0.00003504, -0.00000208, -0.00004572, -0.00007617, -0.00007511, -0.00003839, 0.00001815, 0.00006530, 0.00007703, 0.00004679, -0.00000758, -0.00005454, -0.00006867, -0.00004673, -0.00000862, 0.00001740, 0.00001534, -0.00000691, -0.00002361, -0.00001127, 0.00003151, 0.00007911, 0.00009420, 0.00005382, -0.00003133, -0.00011868, -0.00015756, -0.00012070, -0.00002322, 0.00008372, 0.00014184, 0.00011961, 0.00003043, -0.00007594, -0.00014280, -0.00013848, -0.00007189, 0.00001640, 0.00007996, 0.00009246, 0.00005851, 0.00000521, -0.00003735, -0.00005329, -0.00004570, -0.00002891, -0.00001596, -0.00001078, -0.00000913, -0.00000557, 0.00000047, 0.00000429, 0.00000081, -0.00001006, -0.00002235, -0.00002800, -0.00002256, },
		{ -0.00000970, -0.00001949, -0.00002208, -0.00001789, -0.00001085, -0.00000336, 0.00000682, 0.00002418, 0.00004893, 0.00007209, 0.00007780, 0.00005325, 0.00000018, -0.00006116, -0.00010045, -0.00009469, -0.00004373, 0.00002713, 0.00008058, 0.00008931, 0.00005247, -0.00000455, -0.00004709, -0.00005389, -0.00002990, -0.00000064, 0.00000802, -0.00001009, -0.00003598, -0.00003950, -0.00000312, 0.00006137, 0.00011644, 0.00012177, 0.00006214, -0.00003870, -0.00012890, -0.00015831, -0.00010916, -0.00000785, 0.00008986, 0.00013117, 0.00009630, 0.00000814, -0.00008330, -0.00013088, -0.00011591, -0.00005475, 0.00001518, 0.00005923, 0.00006349, 0.00003746, 0.00000272, -0.00002255, -0.00003264, -0.00003275, -0.00003069, -0.00002955, -0.00002674, -0.00001871, -0.00000614, 0.00000518, 0.00000864, 0.00000220, },
		{ -0.00000566, 0.00000322, 0.00000275, -0.00000379, -0.00000933, -0.00000944, -0.00000631, -0.00000641, -0.00001334, -0.00002198, -0.00001997, 0.00000301, 0.00004436, 0.00008515, 0.00009894, 0.00006870, 0.00000088, -0.00007420, -0.00011785, -0.00010602, -0.00004476, 0.00003232, 0.00008410, 0.00008672, 0.00004705, -0.00000396, -0.00003337, -0.00002837, -0.00000327, 0.00001331, 0.00000160, -0.00003265, -0.00006064, -0.00005109, 0.00000520, 0.00008448, 0.00014177, 0.00013864, 0.00006901, -0.00003427, -0.00011747, -0.00013751, -0.00008732, 0.00000083, 0.00007544, 0.00009645, 0.00005706, -0.00001559, -0.00007952, -0.00010334, -0.00008202, -0.00003498, 0.00000962, 0.00003241, 0.00003106, 0.00001625, 0.00000066, -0.00000972, -0.00001661, -0.00002396, -0.00003205, -0.00003650, -0.00003253, -0.00002021, },
		{ -0.00003340, -0.00003455, -0.00002537, -0.00001205, -0.00000393, -0.00000575, -0.00001332, -0.00001672, -0.00000902, 0.00000645, 0.00001695, 0.00000999, -0.00001500, -0.00004242, -0.00004883, -0.00001951, 0.00003797, 0.00009410, 0.00011356, 0.00007795, -0.00000009, -0.00008173, -0.00012405, -0.00010591, -0.00004063, 0.00003286, 0.00007482, 0.00006904, 0.00003066, -0.00000705, -0.00001729, 0.00000173, 0.00002723, 0.00003057, 0.00000053, -0.00004588, -0.00007351, -0.00005355, 0.00001441, 0.00009850, 0.00015254, 0.00014438, 0.00007630, -0.00001596, -0.00008483, -0.00009909, -0.00006078, -0.00000101, 0.00004132, 0.00004323, 0.00000898, -0.00003664, -0.00006637, -0.00006719, -0.00004509, -0.00001719, 0.00000150, 0.00000665, 0.00000366, 0.00000013, -0.00000130, -0.00000377, -0.00001135, -0.00002332, },
		{ -0.00000345, -0.00000828, -0.00002083, -0.00003308, -0.00003494, -0.00002306, -0.00000502, 0.00000538, -0.00000113, -0.00002071, -0.00003747, -0.00003475, -0.00000882, 0.00002565, 0.00004388, 0.00002854, -0.00001556, -0.00006120, -0.00007494, -0.00003995, 0.00003002, 0.00009624, 0.00011804, 0.00007873, -0.00000262, -0.00008202, -0.00011734, -0.00009383, -0.00003164, 0.00002903, 0.00005479, 0.00003967, 0.00000592, -0.00001444, -0.00000380, 0.00002854, 0.00005406, 0.00004718, 0.00000567, -0.00004542, -0.00006951, -0.00004337, 0.00002614, 0.00010474, 0.00015137, 0.00014292, 0.00008702, 0.00001537, -0.00003704, -0.00005251, -0.00003795, -0.00001598, -0.00000746, -0.00001817, -0.00003791, -0.00005076, -0.00004764, -0.00003189, -0.00001485, -0.00000618, -0.00000696, -0.00001080, -0.00001084, -0.00000643, },
		{ -0.00001368, -0.00001427, -0.00000700, -0.00000074, -0.00000597, -0.00002459, -0.00004587, -0.00005310, -0.00003660, -0.00000316, 0.00002610, 0.00002976, 0.00000229, -0.00003949, -0.00006600, -0.00005502, -0.00000879, 0.00004494, 0.00006993, 0.00004613, -0.00001486, -0.00007512, -0.00009407, -0.00005539, 0.00002096, 0.00009024, 0.00011084, 0.00007051, -0.00000637, -0.00007505, -0.00009903, -0.00007205, -0.00001925, 0.00002212, 0.00002857, 0.00000474, -0.00002259, -0.00002588, 0.00000243, 0.00004440, 0.00006981, 0.00005850, 0.00001564, -0.00003067, -0.00004766, -0.00001940, 0.00004299, 0.00010872, 0.00014652, 0.00014244, 0.00010497, 0.00005591, 0.00001484, -0.00001155, -0.00002875, -0.00004515, -0.00006264, -0.00007462, -0.00007250, -0.00005431, -0.00002818, -0.00000728, -0.00000042, -0.00000590, },
		{ -0.00001278, -0.00000138, -0.00000829, -0.00001864, -0.00001719, -0.00000105, 0.00001689, 0.00001746, -0.00000885, -0.00005145, -0.00008383, -0.00008090, -0.00003804, 0.00002239, 0.00006335, 0.00005722, 0.00000564, -0.00005920, -0.00009432, -0.00007423, -0.00000894, 0.00006111, 0.00009079, 0.00005992, -0.00001294, -0.00008195, -0.00010346, -0.00006402, 0.00001146, 0.00007645, 0.00009292, 0.00005469, -0.00001079, -0.00006250, -0.00007310, -0.00004511, -0.00000583, 0.00001399, 0.00000200, -0.00002840, -0.00004946, -0.00004018, -0.00000162, 0.00004395, 0.00006902, 0.00006002, 0.00002645, -0.00000586, -0.00001199, 0.00001661, 0.00006820, 0.00011972, 0.00015100, 0.00015433, 0.00013414, 0.00009944, 0.00005660, 0.00000812, -0.00004287, -0.00008792, -0.00011440, -0.00011280, -0.00008477, -0.00004483, },
		{ -0.00015355, -0.00012206, -0.00006802, -0.00002162, -0.00000502, -0.00001850, -0.00004074, -0.00004442, -0.00001703, 0.00002791, 0.00005831, 0.00004583, -0.00001187, -0.00008519, -0.00012907, -0.00011201, -0.00003878, 0.00004974, 0.00010079, 0.00008320, 0.00000820, -0.00007677, -0.00011791, -0.00008929, -0.00000930, 0.00007155, 0.00010323, 0.00006788, -0.00001004, -0.00008089, -0.00010233, -0.00006539, 0.00000231, 0.00005694, 0.00006754, 0.00003428, -0.00001523, -0.00004726, -0.00004520, -0.00001875, 0.00000590, 0.00000647, -0.00001933, -0.00005306, -0.00006962, -0.00005516, -0.00001621, 0.00002575, 0.00004954, 0.00004809, 0.00003180, 0.00002013, 0.00002861, 0.00006009, 0.00010516, 0.00014931, 0.00018036, 0.00019117, 0.00017798, 0.00013853, 0.00007369, -0.00000795, -0.00008782, -0.00014218, },
		{ 0.00005426, -0.00006975, -0.00016717, -0.00020468, -0.00017340, -0.00009688, -0.00002045, 0.00001407, -0.00000648, -0.00005813, -0.00009516, -0.00008085, -0.00001381, 0.00006814, 0.00011087, 0.00007983, -0.00001467, -0.00012041, -0.00017406, -0.00014100, -0.00003827, 0.00007470, 0.00013234, 0.00010333, 0.00000934, -0.00008958, -0.00013318, -0.00009797, -0.00000984, 0.00007479, 0.00010568, 0.00006915, -0.00000660, -0.00007272, -0.00009198, -0.00006043, -0.00000562, 0.00003503, 0.00003951, 0.00001320, -0.00001899, -0.00003267, -0.00002120, 0.00000158, 0.00001344, 0.00000082, -0.00003160, -0.00006491, -0.00007951, -0.00006796, -0.00003834, -0.00000698, 0.00001324, 0.00002060, 0.00002385, 0.00003500, 0.00006180, 0.00010440, 0.00015671, 0.00020907, 0.00024955, 0.00026405, 0.00023862, 0.00016603, },
		{ 0.00036947, 0.00037994, 0.00031614, 0.00017663, -0.00000794, -0.00018025, -0.00027930, -0.00027143, -0.00017133, -0.00003828, 0.00005416, 0.00005904, -0.00001639, -0.00011489, -0.00016438, -0.00012347, -0.00000903, 0.00011246, 0.00016559, 0.00011367, -0.00001684, -0.00015108, -0.00021095, -0.00016253, -0.00003608, 0.00009352, 0.00015303, 0.00011434, 0.00000869, -0.00009593, -0.00013816, -0.00009929, -0.00001056, 0.00007073, 0.00009847, 0.00006409, -0.00000321, -0.00005956, -0.00007537, -0.00005110, -0.00001158, 0.00001447, 0.00001399, -0.00000465, -0.00002150, -0.00002162, -0.00000567, 0.00001200, 0.00001507, -0.00000273, -0.00003389, -0.00006313, -0.00007789, -0.00007538, -0.00006195, -0.00004663, -0.00003424, -0.00002272, -0.00000532, 0.00002532, 0.00007415, 0.00014210, 0.00022442, 0.00030792, },
		{ 0.00005406, 0.00016644, 0.00030798, 0.00044968, 0.00054383, 0.00053943, 0.00040817, 0.00016807, -0.00011175, -0.00033397, -0.00041706, -0.00033815, -0.00014929, 0.00004609, 0.00014677, 0.00010857, -0.00003187, -0.00017956, -0.00023793, -0.00016649, -0.00000471, 0.00015253, 0.00021269, 0.00014136, -0.00001803, -0.00017185, -0.00023318, -0.00017252, -0.00003204, 0.00010365, 0.00015991, 0.00011467, 0.00000624, -0.00009534, -0.00013281, -0.00009364, -0.00001141, 0.00006056, 0.00008360, 0.00005413, -0.00000050, -0.00004433, -0.00005629, -0.00003984, -0.00001503, -0.00000138, -0.00000476, -0.00001636, -0.00002243, -0.00001575, -0.00000067, 0.00001127, 0.00001028, -0.00000519, -0.00002858, -0.00005075, -0.00006604, -0.00007438, -0.00007906, -0.00008271, -0.00008460, -0.00008022, -0.00006207, -0.00002055, },
		{ -0.00012865, -0.00014858, -0.00014957, -0.00010961, -0.00000545, 0.00017286, 0.00040478, 0.00063126, 0.00076720, 0.00073555, 0.00050987, 0.00014171, -0.00024806, -0.00051709, -0.00056833, -0.00040073, -0.00011338, 0.00014316, 0.00024458, 0.00015699, -0.00004870, -0.00024089, -0.00030406, -0.00020385, -0.00000280, 0.00018060, 0.00024363, 0.00015802, -0.00001801, -0.00017910, -0.00023681, -0.00016898, -0.00002631, 0.00010411, 0.00015260, 0.00010474, 0.00000231, -0.00008850, -0.00011888, -0.00008260, -0.00001233, 0.00004640, 0.00006415, 0.00004128, 0.00000099, -0.00003001, -0.00003846, -0.00002895, -0.00001578, -0.00001046, -0.00001442, -0.00002075, -0.00002169, -0.00001504, -0.00000517, 0.00000117, -0.00000015, -0.00000861, -0.00002086, -0.00003399, -0.00004745, -0.00006266, -0.00008133, -0.00010399, },
		{ -0.00002698, -0.00003914, -0.00006200, -0.00010114, -0.00015833, -0.00022326, -0.00026652, -0.00024199, -0.00010465, 0.00016005, 0.00050986, 0.00084211, 0.00102461, 0.00095392, 0.00061421, 0.00010248, -0.00040092, -0.00070925, -0.00071758, -0.00045539, -0.00007253, 0.00023690, 0.00033316, 0.00019852, -0.00006238, -0.00028797, -0.00035204, -0.00023022, -0.00000472, 0.00019117, 0.00025282, 0.00016075, -0.00001681, -0.00017169, -0.00022135, -0.00015243, -0.00001937, 0.00009564, 0.00013321, 0.00008677, -0.00000244, -0.00007711, -0.00009939, -0.00006848, -0.00001327, 0.00003080, 0.00004356, 0.00002771, 0.00000094, -0.00001891, -0.00002453, -0.00001999, -0.00001403, -0.00001249, -0.00001529, -0.00001868, -0.00001953, -0.00001768, -0.00001516, -0.00001387, -0.00001423, -0.00001561, -0.00001759, -0.00002080, },
		{ -0.00002928, -0.00002872, -0.00002554, -0.00001891, -0.00000967, -0.00000431, -0.00001770, -0.00006903, -0.00016848, -0.00029913, -0.00040664, -0.00040984, -0.00023623, 0.00013035, 0.00061624, 0.00106497, 0.00129298, 0.00117438, 0.00071277, 0.00005787, -0.00055007, -0.00088665, -0.00084806, -0.00049895, -0.00003646, 0.00031141, 0.00039910, 0.00022806, -0.00006924, -0.00031247, -0.00037424, -0.00024202, -0.00001087, 0.00018209, 0.00023874, 0.00014922, -0.00001468, -0.00015118, -0.00018976, -0.00012573, -0.00001203, 0.00008042, 0.00010577, 0.00006424, -0.00000720, -0.00006333, -0.00007788, -0.00005379, -0.00001415, 0.00001610, 0.00002473, 0.00001517, -0.00000066, -0.00001212, -0.00001553, -0.00001344, -0.00001036, -0.00000914, -0.00001022, -0.00001286, -0.00001643, -0.00002054, -0.00002461, -0.00002777, },
		{ -0.00001304, -0.00001992, -0.00002693, -0.00003425, -0.00004240, -0.00004963, -0.00004966, -0.00003387, 0.00000042, 0.00003964, 0.00005043, -0.00000830, -0.00015808, -0.00037104, -0.00055896, -0.00059783, -0.00038594, 0.00008947, 0.00071567, 0.00127804, 0.00154424, 0.00137371, 0.00079677, 0.00001659, -0.00067435, -0.00102596, -0.00094435, -0.00052884, -0.00001358, 0.00035400, 0.00043242, 0.00024195, -0.00006735, -0.00031031, -0.00036755, -0.00023797, -0.00002058, 0.00015503, 0.00020415, 0.00012563, -0.00001210, -0.00012141, -0.00014773, -0.00009345, -0.00000521, 0.00006148, 0.00007526, 0.00004106, -0.00001115, -0.00004931, -0.00005757, -0.00004060, -0.00001492, 0.00000391, 0.00000945, 0.00000463, -0.00000351, -0.00000925, -0.00001062, -0.00000869, -0.00000566, -0.00000362, -0.00000399, -0.00000730, },
		{ -0.00000101, 0.00000001, -0.00000211, -0.00000625, -0.00001004, -0.00001252, -0.00001663, -0.00002837, -0.00005119, -0.00007855, -0.00009152, -0.00006750, 0.00000110, 0.00008895, 0.00013636, 0.00007431, -0.00013050, -0.00043436, -0.00070905, -0.00078509, -0.00053468, 0.00004537, 0.00079955, 0.00145815, 0.00174958, 0.00152920, 0.00085820, -0.00001328, -0.00075549, -0.00110824, -0.00099472, -0.00054319, -0.00000906, 0.00035761, 0.00042830, 0.00023850, -0.00005706, -0.00028250, -0.00033392, -0.00021925, -0.00003212, 0.00011507, 0.00015540, 0.00009416, -0.00000963, -0.00008769, -0.00010246, -0.00006087, 0.00000018, 0.00004203, 0.00004649, 0.00002073, -0.00001365, -0.00003670, -0.00004064, -0.00003013, -0.00001551, -0.00000518, -0.00000196, -0.00000384, -0.00000705, -0.00000861, -0.00000741, -0.00000421, },
		{ -0.00001052, -0.00000762, -0.00000264, 0.00000188, 0.00000256, -0.00000214, -0.00000949, -0.00001317, -0.00000799, 0.00000363, 0.00000918, -0.00000786, -0.00005418, -0.00011388, -0.00014862, -0.00011717, -0.00000744, 0.00013871, 0.00022984, 0.00016690, -0.00009278, -0.00048536, -0.00084093, -0.00094845, -0.00066174, 0.00000676, 0.00086015, 0.00158438, 0.00188410, 0.00162225, 0.00089094, -0.00002586, -0.00078140, -0.00112224, -0.00099302, -0.00054083, -0.00002379, 0.00032219, 0.00038791, 0.00021817, -0.00004084, -0.00023473, -0.00027974, -0.00018912, -0.00004313, 0.00006948, 0.00010102, 0.00006008, -0.00000786, -0.00005559, -0.00006105, -0.00003284, 0.00000344, 0.00002475, 0.00002310, 0.00000559, -0.00001431, -0.00002632, -0.00002784, -0.00002251, -0.00001588, -0.00001167, -0.00001057, -0.00001098, },
		{ -0.00001601, -0.00001687, -0.00001842, -0.00001804, -0.00001315, -0.00000359, 0.00000683, 0.00001125, 0.00000421, -0.00001280, -0.00002909, -0.00002979, -0.00000725, 0.00002855, 0.00004984, 0.00002622, -0.00005115, -0.00015228, -0.00021449, -0.00017612, -0.00002276, 0.00018352, 0.00031808, 0.00025455, -0.00005412, -0.00052137, -0.00093954, -0.00106628, -0.00074839, -0.00001868, 0.00089164, 0.00164143, 0.00193092, 0.00164143, 0.00089164, -0.00001868, -0.00074839, -0.00106628, -0.00093954, -0.00052137, -0.00005412, 0.00025455, 0.00031808, 0.00018352, -0.00002276, -0.00017612, -0.00021449, -0.00015228, -0.00005115, 0.00002622, 0.00004984, 0.00002855, -0.00000725, -0.00002979, -0.00002909, -0.00001280, 0.00000421, 0.00001125, 0.00000683, -0.00000359, -0.00001315, -0.00001804, -0.00001842, -0.00001687, },
		{ -0.00001052, -0.00001098, -0.00001057, -0.00001167, -0.00001588, -0.00002251, -0.00002784, -0.00002632, -0.00001431, 0.00000559, 0.00002310, 0.00002475, 0.00000344, -0.00003284, -0.00006105, -0.00005559, -0.00000786, 0.00006008, 0.00010102, 0.00006948, -0.00004313, -0.00018912, -0.00027974, -0.00023473, -0.00004084, 0.00021817, 0.00038791, 0.00032219, -0.00002379, -0.00054083, -0.00099302, -0.00112224, -0.00078140, -0.00002586, 0.00089094, 0.00162225, 0.00188410, 0.00158438, 0.00086015, 0.00000676, -0.00066174, -0.00094845, -0.00084093, -0.00048536, -0.00009278, 0.00016690, 0.00022984, 0.00013871, -0.00000744, -0.00011717, -0.00014862, -0.00011388, -0.00005418, -0.00000786, 0.00000918, 0.00000363, -0.00000799, -0.00001317, -0.00000949, -0.00000214, 0.00000256, 0.00000188, -0.00000264, -0.00000762, },
		{ -0.00000101, -0.00000421, -0.00000741, -0.00000861, -0.00000705, -0.00000384, -0.00000196, -0.00000518, -0.00001551, -0.00003013, -0.00004064, -0.00003670, -0.00001365, 0.00002073, 0.00004649, 0.00004203, 0.00000018, -0.00006087, -0.00010246, -0.00008769, -0.00000963, 0.00009416, 0.00015540, 0.00011507, -0.00003212, -0.00021925, -0.00033392, -0.00028250, -0.00005706, 0.00023850, 0.00042830, 0.00035761, -0.00000906, -0.00054319, -0.00099472, -0.00110824, -0.00075549, -0.00001328, 0.00085820, 0.00152920, 0.00174958, 0.00145815, 0.00079955, 0.00004537, -0.00053468, -0.00078509, -0.00070905, -0.00043436, -0.00013050, 0.00007431, 0.00013636, 0.00008895, 0.00000110, -0.00006750, -0.00009152, -0.00007855, -0.00005119, -0.00002837, -0.00001663, -0.00001252, -0.00001004, -0.00000625, -0.00000211, 0.00000001, },
		{ -0.00001304, -0.00000730, -0.00000399, -0.00000362, -0.00000566, -0.00000869, -0.00001062, -0.00000925, -0.00000351, 0.00000463, 0.00000945, 0.00000391, -0.00001492, -0.00004060, -0.00005757, -0.00004931, -0.00001115, 0.00004106, 0.00007526, 0.00006148, -0.00000521, -0.00009345, -0.00014773, -0.00012141, -0.00001210, 0.00012563, 0.00020415, 0.00015503, -0.00002058, -0.00023797, -0.00036755, -0.00031031, -0.00006735, 0.00024195, 0.00043242, 0.00035400, -0.00001358, -0.00052884, -0.00094435, -0.00102596, -0.00067435, 0.00001659, 0.00079677, 0.00137371, 0.00154424, 0.00127804, 0.00071567, 0.00008947, -0.00038594, -0.00059783, -0.00055896, -0.00037104, -0.00015808, -0.00000830, 0.00005043, 0.00003964, 0.00000042, -0.00003387, -0.00004966, -0.00004963, -0.00004240, -0.00003425, -0.00002693, -0.00001992, },
		{ -0.00002928, -0.00002777, -0.00002461, -0.00002054, -0.00001643, -0.00001286, -0.00001022, -0.00000914, -0.00001036, -0.00001344, -0.00001553, -0.00001212, -0.00000066, 0.00001517, 0.00002473, 0.00001610, -0.00001415, -0.00005379, -0.00007788, -0.00006333, -0.00000720, 0.00006424, 0.00010577, 0.00008042, -0.00001203, -0.00012573, -0.00018976, -0.00015118, -0.00001468, 0.00014922, 0.00023874, 0.00018209, -0.00001087, -0.00024202, -0.00037424, -0.00031247, -0.00006924, 0.00022806, 0.00039910, 0.00031141, -0.00003646, -0.00049895, -0.00084806, -0.00088665, -0.00055007, 0.00005787, 0.00071277, 0.00117438, 0.00129298, 0.00106497, 0.00061624, 0.00013035, -0.00023623, -0.00040984, -0.00040664, -0.00029913, -0.00016848, -0.00006903, -0.00001770, -0.00000431, -0.00000967, -0.00001891, -0.00002554, -0.00002872, },
		{ -0.00002698, -0.00002080, -0.00001759, -0.00001561, -0.00001423, -0.00001387, -0.00001516, -0.00001768, -0.00001953, -0.00001868, -0.00001529, -0.00001249, -0.00001403, -0.00001999, -0.00002453, -0.00001891, 0.00000094, 0.00002771, 0.00004356, 0.00003080, -0.00001327, -0.00006848, -0.00009939, -0.00007711, -0.00000244, 0.00008677, 0.00013321, 0.00009564, -0.00001937, -0.00015243, -0.00022135, -0.00017169, -0.00001681, 0.00016075, 0.00025282, 0.00019117, -0.00000472, -0.00023022, -0.00035204, -0.00028797, -0.00006238, 0.00019852, 0.00033316, 0.00023690, -0.00007253, -0.00045539, -0.00071758, -0.00070925, -0.00040092, 0.00010248, 0.00061421, 0.00095392, 0.00102461, 0.00084211, 0.00050986, 0.00016005, -0.00010465, -0.00024199, -0.00026652, -0.00022326, -0.00015833, -0.00010114, -0.00006200, -0.00003914, },
		{ -0.00012865, -0.00010399, -0.00008133, -0.00006266, -0.00004745, -0.00003399, -0.00002086, -0.00000861, -0.00000015, 0.00000117, -0.00000517, -0.00001504, -0.00002169, -0.00002075, -0.00001442, -0.00001046, -0.00001578, -0.00002895, -0.00003846, -0.00003001, 0.00000099, 0.00004128, 0.00006415, 0.00004640, -0.00001233, -0.00008260, -0.00011888, -0.00008850, 0.00000231, 0.00010474, 0.00015260, 0.00010411, -0.00002631, -0.00016898, -0.00023681, -0.00017910, -0.00001801, 0.00015802, 0.00024363, 0.00018060, -0.00000280, -0.00020385, -0.00030406, -0.00024089, -0.00004870, 0.00015699, 0.00024458, 0.00014316, -0.00011338, -0.00040073, -0.00056833, -0.00051709, -0.00024806, 0.00014171, 0.00050987, 0.00073555, 0.00076720, 0.00063126, 0.00040478, 0.00017286, -0.00000545, -0.00010961, -0.00014957, -0.00014858, },
		{ 0.00005406, -0.00002055, -0.00006207, -0.00008022, -0.00008460, -0.00008271, -0.00007906, -0.00007438, -0.00006604, -0.00005075, -0.00002858, -0.00000519, 0.00001028, 0.00001127, -0.00000067, -0.00001575, -0.00002243, -0.00001636, -0.00000476, -0.00000138, -0.00001503, -0.00003984, -0.00005629, -0.00004433, -0.00000050, 0.00005413, 0.00008360, 0.00006056, -0.00001141, -0.00009364, -0.00013281, -0.00009534, 0.00000624, 0.00011467, 0.00015991, 0.00010365, -0.00003204, -0.00017252, -0.00023318, -0.00017185, -0.00001803, 0.00014136, 0.00021269, 0.00015253, -0.00000471, -0.00016649, -0.00023793, -0.00017956, -0.00003187, 0.00010857, 0.00014677, 0.00004609, -0.00014929, -0.00033815, -0.00041706, -0.00033397, -0.00011175, 0.00016807, 0.00040817, 0.00053943, 0.00054383, 0.00044968, 0.00030798, 0.00016644, },
		{ 0.00036947, 0.00030792, 0.00022442, 0.00014210, 0.00007415, 0.00002532, -0.00000532, -0.00002272, -0.00003424, -0.00004663, -0.00006195, -0.00007538, -0.00007789, -0.00006313, -0.00003389, -0.00000273, 0.00001507, 0.00001200, -0.00000567, -0.00002162, -0.00002150, -0.00000465, 0.00001399, 0.00001447, -0.00001158, -0.00005110, -0.00007537, -0.00005956, -0.00000321, 0.00006409, 0.00009847, 0.00007073, -0.00001056, -0.00009929, -0.00013816, -0.00009593, 0.00000869, 0.00011434, 0.00015303, 0.00009352, -0.00003608, -0.00016253, -0.00021095, -0.00015108, -0.00001684, 0.00011367, 0.00016559, 0.00011246, -0.00000903, -0.00012347, -0.00016438, -0.00011489, -0.00001639, 0.00005904, 0.00005416, -0.00003828, -0.00017133, -0.00027143, -0.00027930, -0.00018025, -0.00000794, 0.00017663, 0.00031614, 0.00037994, },
		{ 0.00005426, 0.00016603, 0.00023862, 0.00026405, 0.00024955, 0.00020907, 0.00015671, 0.00010440, 0.00006180, 0.00003500, 0.00002385, 0.00002060, 0.00001324, -0.00000698, -0.00003834, -0.00006796, -0.00007951, -0.00006491, -0.00003160, 0.00000082, 0.00001344, 0.00000158, -0.00002120, -0.00003267, -0.00001899, 0.00001320, 0.00003951, 0.00003503, -0.00000562, -0.00006043, -0.00009198, -0.00007272, -0.00000660, 0.00006915, 0.00010568, 0.00007479, -0.00000984, -0.00009797, -0.00013318, -0.00008958, 0.00000934, 0.00010333, 0.00013234, 0.00007470, -0.00003827, -0.00014100, -0.00017406, -0.00012041, -0.00001467, 0.00007983, 0.00011087, 0.00006814, -0.00001381, -0.00008085, -0.00009516, -0.00005813, -0.00000648, 0.00001407, -0.00002045, -0.00009688, -0.00017340, -0.00020468, -0.00016717, -0.00006975, },
		{ -0.00015355, -0.00014218, -0.00008782, -0.00000795, 0.00007369, 0.00013853, 0.00017798, 0.00019117, 0.00018036, 0.00014931, 0.00010516, 0.00006009, 0.00002861, 0.00002013, 0.00003180, 0.00004809, 0.00004954, 0.00002575, -0.00001621, -0.00005516, -0.00006962, -0.00005306, -0.00001933, 0.00000647, 0.00000590, -0.00001875, -0.00004520, -0.00004726, -0.00001523, 0.00003428, 0.00006754, 0.00005694, 0.00000231, -0.00006539, -0.00010233, -0.00008089, -0.00001004, 0.00006788, 0.00010323, 0.00007155, -0.00000930, -0.00008929, -0.00011791, -0.00007677, 0.00000820, 0.00008320, 0.00010079, 0.00004974, -0.00003878, -0.00011201, -0.00012907, -0.00008519, -0.00001187, 0.00004583, 0.00005831, 0.00002791, -0.00001703, -0.00004442, -0.00004074, -0.00001850, -0.00000502, -0.00002162, -0.00006802, -0.00012206, },
		{ -0.00001278, -0.00004483, -0.00008477, -0.00011280, -0.00011440, -0.00008792, -0.00004287, 0.00000812, 0.00005660, 0.00009944, 0.00013414, 0.00015433, 0.00015100, 0.00011972, 0.00006820, 0.00001661, -0.00001199, -0.00000586, 0.00002645, 0.00006002, 0.00006902, 0.00004395, -0.00000162, -0.00004018, -0.00004946, -0.00002840, 0.00000200, 0.00001399, -0.00000583, -0.00004511, -0.00007310, -0.00006250, -0.00001079, 0.00005469, 0.00009292, 0.00007645, 0.00001146, -0.00006402, -0.00010346, -0.00008195, -0.00001294, 0.00005992, 0.00009079, 0.00006111, -0.00000894, -0.00007423, -0.00009432, -0.00005920, 0.00000564, 0.00005722, 0.00006335, 0.00002239, -0.00003804, -0.00008090, -0.00008383, -0.00005145, -0.00000885, 0.00001746, 0.00001689, -0.00000105, -0.00001719, -0.00001864, -0.00000829, -0.00000138, },
		{ -0.00001368, -0.00000590, -0.00000042, -0.00000728, -0.00002818, -0.00005431, -0.00007250, -0.00007462, -0.00006264, -0.00004515, -0.00002875, -0.00001155, 0.00001484, 0.00005591, 0.00010497, 0.00014244, 0.00014652, 0.00010872, 0.00004299, -0.00001940, -0.00004766, -0.00003067, 0.00001564, 0.00005850, 0.00006981, 0.00004440, 0.00000243, -0.00002588, -0.00002259, 0.00000474, 0.00002857, 0.00002212, -0.00001925, -0.00007205, -0.00009903, -0.00007505, -0.00000637, 0.00007051, 0.00011084, 0.00009024, 0.00002096, -0.00005539, -0.00009407, -0.00007512, -0.00001486, 0.00004613, 0.00006993, 0.00004494, -0.00000879, -0.00005502, -0.00006600, -0.00003949, 0.00000229, 0.00002976, 0.00002610, -0.00000316, -0.00003660, -0.00005310, -0.00004587, -0.00002459, -0.00000597, -0.00000074, -0.00000700, -0.00001427, },
		{ -0.00000345, -0.00000643, -0.00001084, -0.00001080, -0.00000696, -0.00000618, -0.00001485, -0.00003189, -0.00004764, -0.00005076, -0.00003791, -0.00001817, -0.00000746, -0.00001598, -0.00003795, -0.00005251, -0.00003704, 0.00001537, 0.00008702, 0.00014292, 0.00015137, 0.00010474, 0.00002614, -0.00004337, -0.00006951, -0.00004542, 0.00000567, 0.00004718, 0.00005406, 0.00002854, -0.00000380, -0.00001444, 0.00000592, 0.00003967, 0.00005479, 0.00002903, -0.00003164, -0.00009383, -0.00011734, -0.00008202, -0.00000262, 0.00007873, 0.00011804, 0.00009624, 0.00003002, -0.00003995, -0.00007494, -0.00006120, -0.00001556, 0.00002854, 0.00004388, 0.00002565, -0.00000882, -0.00003475, -0.00003747, -0.00002071, -0.00000113, 0.00000538, -0.00000502, -0.00002306, -0.00003494, -0.00003308, -0.00002083, -0.00000828, },
		{ -0.00003340, -0.00002332, -0.00001135, -0.00000377, -0.00000130, 0.00000013, 0.00000366, 0.00000665, 0.00000150, -0.00001719, -0.00004509, -0.00006719, -0.00006637, -0.00003664, 0.00000898, 0.00004323, 0.00004132, -0.00000101, -0.00006078, -0.00009909, -0.00008483, -0.00001596, 0.00007630, 0.00014438, 0.00015254, 0.00009850, 0.00001441, -0.00005355, -0.00007351, -0.00004588, 0.00000053, 0.00003057, 0.00002723, 0.00000173, -0.00001729, -0.00000705, 0.00003066, 0.00006904, 0.00007482, 0.00003286, -0.00004063, -0.00010591, -0.00012405, -0.00008173, -0.00000009, 0.00007795, 0.00011356, 0.00009410, 0.00003797, -0.00001951, -0.00004883, -0.00004242, -0.00001500, 0.00000999, 0.00001695, 0.00000645, -0.00000902, -0.00001672, -0.00001332, -0.00000575, -0.00000393, -0.00001205, -0.00002537, -0.00003455, },
		{ -0.00000566, -0.00002021, -0.00003253, -0.00003650, -0.00003205, -0.00002396, -0.00001661, -0.00000972, 0.00000066, 0.00001625, 0.00003106, 0.00003241, 0.00000962, -0.00003498, -0.00008202, -0.00010334, -0.00007952, -0.00001559, 0.00005706, 0.00009645, 0.00007544, 0.00000083, -0.00008732, -0.00013751, -0.00011747, -0.00003427, 0.00006901, 0.00013864, 0.00014177, 0.00008448, 0.00000520, -0.00005109, -0.00006064, -0.00003265, 0.00000160, 0.00001331, -0.00000327, -0.00002837, -0.00003337, -0.00000396, 0.00004705, 0.00008672, 0.00008410, 0.00003232, -0.00004476, -0.00010602, -0.00011785, -0.00007420, 0.00000088, 0.00006870, 0.00009894, 0.00008515, 0.00004436, 0.00000301, -0.00001997, -0.00002198, -0.00001334, -0.00000641, -0.00000631, -0.00000944, -0.00000933, -0.00000379, 0.00000275, 0.00000322, },
		{ -0.00000970, 0.00000220, 0.00000864, 0.00000518, -0.00000614, -0.00001871, -0.00002674, -0.00002955, -0.00003069, -0.00003275, -0.00003264, -0.00002255, 0.00000272, 0.00003746, 0.00006349, 0.00005923, 0.00001518, -0.00005475, -0.00011591, -0.00013088, -0.00008330, 0.00000814, 0.00009630, 0.00013117, 0.00008986, -0.00000785, -0.00010916, -0.00015831, -0.00012890, -0.00003870, 0.00006214, 0.00012177, 0.00011644, 0.00006137, -0.00000312, -0.00003950, -0.00003598, -0.00001009, 0.00000802, -0.00000064, -0.00002990, -0.00005389, -0.00004709, -0.00000455, 0.00005247, 0.00008931, 0.00008058, 0.00002713, -0.00004373, -0.00009469, -0.00010045, -0.00006116, 0.00000018, 0.00005325, 0.00007780, 0.00007209, 0.00004893, 0.00002418, 0.00000682, -0.00000336, -0.00001085, -0.00001789, -0.00002208, -0.00001949, },
		{ -0.00000779, -0.00002256, -0.00002800, -0.00002235, -0.00001006, 0.00000081, 0.00000429, 0.00000047, -0.00000557, -0.00000913, -0.00001078, -0.00001596, -0.00002891, -0.00004570, -0.00005329, -0.00003735, 0.00000521, 0.00005851, 0.00009246, 0.00007996, 0.00001640, -0.00007189, -0.00013848, -0.00014280, -0.00007594, 0.00003043, 0.00011961, 0.00014184, 0.00008372, -0.00002322, -0.00012070, -0.00015756, -0.00011868, -0.00003133, 0.00005382, 0.00009420, 0.00007911, 0.00003151, -0.00001127, -0.00002361, -0.00000691, 0.00001534, 0.00001740, -0.00000862, -0.00004673, -0.00006867, -0.00005454, -0.00000758, 0.00004679, 0.00007703, 0.00006530, 0.00001815, -0.00003839, -0.00007511, -0.00007617, -0.00004572, -0.00000208, 0.00003504, 0.00005494, 0.00005824, 0.00005161, 0.00004074, 0.00002719, 0.00001049, },
		{ 0.00005244, 0.00005028, 0.00003818, 0.00001775, -0.00000439, -0.00001982, -0.00002369, -0.00001812, -0.00001037, -0.00000694, -0.00000836, -0.00000934, -0.00000439, 0.00000536, 0.00001074, 0.00000093, -0.00002621, -0.00005801, -0.00007171, -0.00004904, 0.00000839, 0.00007452, 0.00011074, 0.00008910, 0.00001240, -0.00008280, -0.00014460, -0.00013597, -0.00005813, 0.00004788, 0.00012393, 0.00012842, 0.00006035, -0.00004086, -0.00011963, -0.00013692, -0.00009137, -0.00001638, 0.00004334, 0.00005973, 0.00003596, -0.00000039, -0.00001911, -0.00000838, 0.00001899, 0.00003708, 0.00002685, -0.00000967, -0.00005068, -0.00006975, -0.00005377, -0.00001151, 0.00003246, 0.00005361, 0.00004213, 0.00000722, -0.00003052, -0.00005229, -0.00005089, -0.00003159, -0.00000561, 0.00001783, 0.00003519, 0.00004672, },
		{ -0.00001006, 0.00000482, 0.00002222, 0.00003970, 0.00005154, 0.00005176, 0.00003884, 0.00001814, -0.00000078, -0.00001048, -0.00001076, -0.00000829, -0.00001057, -0.00001880, -0.00002586, -0.00002180, -0.00000323, 0.00002077, 0.00003242, 0.00001747, -0.00002215, -0.00006518, -0.00008189, -0.00005352, 0.00001233, 0.00008201, 0.00011388, 0.00008404, 0.00000348, -0.00008553, -0.00013314, -0.00011165, -0.00003287, 0.00005829, 0.00011044, 0.00009597, 0.00002639, -0.00005643, -0.00010681, -0.00010238, -0.00005486, 0.00000087, 0.00003106, 0.00002402, -0.00000535, -0.00002888, -0.00002605, 0.00000229, 0.00003587, 0.00005019, 0.00003376, -0.00000455, -0.00004213, -0.00005784, -0.00004517, -0.00001480, 0.00001388, 0.00002546, 0.00001690, -0.00000323, -0.00002232, -0.00003187, -0.00003068, -0.00002224, },
		{ -0.00001588, -0.00001870, -0.00002024, -0.00002004, -0.00001504, -0.00000213, 0.00001773, 0.00003786, 0.00004904, 0.00004568, 0.00003042, 0.00001290, 0.00000300, 0.00000345, 0.00000754, 0.00000461, -0.00001060, -0.00003182, -0.00004399, -0.00003412, -0.00000272, 0.00003316, 0.00004923, 0.00003043, -0.00001649, -0.00006406, -0.00008000, -0.00004859, 0.00001684, 0.00007956, 0.00010108, 0.00006556, -0.00000887, -0.00008001, -0.00010692, -0.00007501, -0.00000488, 0.00006082, 0.00008381, 0.00005310, -0.00000985, -0.00006639, -0.00008555, -0.00006233, -0.00001818, 0.00001536, 0.00001798, -0.00000718, -0.00003792, -0.00004943, -0.00003134, 0.00000632, 0.00004088, 0.00005254, 0.00003654, 0.00000452, -0.00002459, -0.00003698, -0.00003115, -0.00001616, -0.00000361, 0.00000007, -0.00000404, -0.00001075, },
		{ -0.00001497, -0.00001578, -0.00001536, -0.00001348, -0.00001265, -0.00001568, -0.00002184, -0.00002568, -0.00002025, -0.00000289, 0.00002105, 0.00004034, 0.00004507, 0.00003396, 0.00001601, 0.00000447, 0.00000698, 0.00001916, 0.00002708, 0.00001760, -0.00001045, -0.00004299, -0.00005869, -0.00004377, -0.00000334, 0.00003954, 0.00005761, 0.00003765, -0.00000930, -0.00005352, -0.00006523, -0.00003442, 0.00002150, 0.00006796, 0.00007515, 0.00003751, -0.00002232, -0.00006779, -0.00007178, -0.00003372, 0.00002057, 0.00005601, 0.00005083, 0.00000975, -0.00004030, -0.00006859, -0.00006039, -0.00002525, 0.00001082, 0.00002341, 0.00000542, -0.00002954, -0.00005736, -0.00005957, -0.00003439, 0.00000374, 0.00003464, 0.00004501, 0.00003483, 0.00001451, -0.00000369, -0.00001330, -0.00001538, -0.00001478, },
		{ 0.00001434, 0.00000667, -0.00000169, -0.00001038, -0.00001664, -0.00001767, -0.00001403, -0.00001057, -0.00001308, -0.00002305, -0.00003475, -0.00003801, -0.00002542, 0.00000103, 0.00002946, 0.00004497, 0.00003982, 0.00001952, -0.00000020, -0.00000412, 0.00001115, 0.00003366, 0.00004370, 0.00002788, -0.00001010, -0.00004987, -0.00006691, -0.00004898, -0.00000526, 0.00003847, 0.00005615, 0.00003847, -0.00000097, -0.00003471, -0.00003985, -0.00001344, 0.00002569, 0.00004992, 0.00004175, 0.00000584, -0.00003413, -0.00005157, -0.00003505, 0.00000395, 0.00003879, 0.00004552, 0.00001874, -0.00002511, -0.00005889, -0.00006261, -0.00003598, 0.00000224, 0.00002711, 0.00002344, -0.00000541, -0.00004110, -0.00006272, -0.00005936, -0.00003496, -0.00000353, 0.00002068, 0.00003097, 0.00002941, 0.00002226, },
		{ 0.00000425, 0.00001513, 0.00002179, 0.00002522, 0.00002431, 0.00001773, 0.00000691, -0.00000323, -0.00000741, -0.00000459, -0.00000007, -0.00000241, -0.00001656, -0.00003840, -0.00005543, -0.00005425, -0.00003038, 0.00000692, 0.00003892, 0.00004892, 0.00003353, 0.00000559, -0.00001391, -0.00001003, 0.00001542, 0.00004433, 0.00005414, 0.00003341, -0.00000955, -0.00005109, -0.00006722, -0.00004903, -0.00000834, 0.00003032, 0.00004580, 0.00003369, 0.00000778, -0.00001075, -0.00000867, 0.00001022, 0.00002869, 0.00002933, 0.00000789, -0.00002286, -0.00004180, -0.00003445, -0.00000373, 0.00003108, 0.00004679, 0.00003171, -0.00000646, -0.00004549, -0.00006283, -0.00004973, -0.00001605, 0.00001677, 0.00002944, 0.00001612, -0.00001368, -0.00004257, -0.00005634, -0.00005118, -0.00003323, -0.00001241, },
		{ -0.00004299, -0.00003892, -0.00002976, -0.00001962, -0.00000927, 0.00000210, 0.00001371, 0.00002205, 0.00002331, 0.00001731, 0.00000909, 0.00000592, 0.00001132, 0.00002081, 0.00002348, 0.00000946, -0.00002155, -0.00005723, -0.00007839, -0.00007070, -0.00003499, 0.00001162, 0.00004517, 0.00004951, 0.00002654, -0.00000496, -0.00002172, -0.00001129, 0.00001958, 0.00004950, 0.00005665, 0.00003335, -0.00000883, -0.00004663, -0.00006001, -0.00004439, -0.00001209, 0.00001712, 0.00002952, 0.00002531, 0.00001605, 0.00001403, 0.00002216, 0.00003175, 0.00002980, 0.00001022, -0.00001972, -0.00004288, -0.00004362, -0.00001917, 0.00001716, 0.00004370, 0.00004398, 0.00001724, -0.00002113, -0.00004942, -0.00005299, -0.00003261, -0.00000271, 0.00001852, 0.00002021, 0.00000403, -0.00001903, -0.00003679, },
		{ 0.00000445, -0.00000919, -0.00002160, -0.00002778, -0.00002833, -0.00002679, -0.00002529, -0.00002256, -0.00001589, -0.00000490, 0.00000660, 0.00001290, 0.00001143, 0.00000606, 0.00000536, 0.00001598, 0.00003586, 0.00005287, 0.00005130, 0.00002271, -0.00002605, -0.00007413, -0.00009754, -0.00008353, -0.00003909, 0.00001243, 0.00004490, 0.00004493, 0.00001933, -0.00001025, -0.00002185, -0.00000727, 0.00002325, 0.00004867, 0.00005130, 0.00002816, -0.00000798, -0.00003767, -0.00004728, -0.00003642, -0.00001578, 0.00000206, 0.00001145, 0.00001597, 0.00002290, 0.00003505, 0.00004661, 0.00004677, 0.00002852, -0.00000411, -0.00003629, -0.00005076, -0.00003904, -0.00000756, 0.00002561, 0.00004162, 0.00003228, 0.00000455, -0.00002459, -0.00003912, -0.00003350, -0.00001472, 0.00000366, 0.00001076, },
		{ -0.00001055, 0.00000039, 0.00000418, -0.00000142, -0.00001177, -0.00001985, -0.00002190, -0.00001954, -0.00001737, -0.00001825, -0.00002062, -0.00002008, -0.00001400, -0.00000479, 0.00000126, -0.00000062, -0.00000832, -0.00001240, -0.00000209, 0.00002569, 0.00006104, 0.00008428, 0.00007705, 0.00003476, -0.00002809, -0.00008410, -0.00010770, -0.00008972, -0.00004246, 0.00000785, 0.00003663, 0.00003485, 0.00001249, -0.00000973, -0.00001454, 0.00000120, 0.00002596, 0.00004255, 0.00003984, 0.00001939, -0.00000706, -0.00002631, -0.00003205, -0.00002705, -0.00001858, -0.00001137, -0.00000406, 0.00000823, 0.00002754, 0.00004855, 0.00006006, 0.00005227, 0.00002458, -0.00001193, -0.00003993, -0.00004600, -0.00002887, -0.00000018, 0.00002291, 0.00002828, 0.00001566, -0.00000447, -0.00001882, -0.00002002, },
		{ -0.00000076, -0.00000873, -0.00000772, 0.00000081, 0.00000923, 0.00000990, 0.00000092, -0.00001267, -0.00002299, -0.00002517, -0.00002061, -0.00001498, -0.00001316, -0.00001513, -0.00001640, -0.00001268, -0.00000469, 0.00000118, -0.00000223, -0.00001585, -0.00003086, -0.00003256, -0.00001018, 0.00003378, 0.00008146, 0.00010785, 0.00009495, 0.00004338, -0.00002626, -0.00008382, -0.00010582, -0.00008770, -0.00004481, -0.00000200, 0.00002118, 0.00002050, 0.00000666, -0.00000420, -0.00000190, 0.00001203, 0.00002720, 0.00003281, 0.00002518, 0.00000927, -0.00000612, -0.00001501, -0.00001762, -0.00001821, -0.00001974, -0.00002011, -0.00001355, 0.00000399, 0.00002948, 0.00005242, 0.00006044, 0.00004736, 0.00001811, -0.00001334, -0.00003207, -0.00003110, -0.00001505, 0.00000361, 0.00001313, 0.00000979, },
		{ -0.00000029, 0.00000535, 0.00000190, -0.00000668, -0.00001199, -0.00000798, 0.00000416, 0.00001659, 0.00002045, 0.00001223, -0.00000382, -0.00001886, -0.00002582, -0.00002396, -0.00001844, -0.00001529, -0.00001622, -0.00001732, -0.00001308, -0.00000238, 0.00000866, 0.00001000, -0.00000436, -0.00002965, -0.00005036, -0.00004857, -0.00001581, 0.00003916, 0.00009269, 0.00011823, 0.00010109, 0.00004715, -0.00002013, -0.00007250, -0.00009174, -0.00007777, -0.00004578, -0.00001528, 0.00000148, 0.00000451, 0.00000249, 0.00000440, 0.00001266, 0.00002244, 0.00002655, 0.00002163, 0.00001064, 0.00000011, -0.00000524, -0.00000596, -0.00000671, -0.00001138, -0.00001873, -0.00002233, -0.00001522, 0.00000403, 0.00002860, 0.00004664, 0.00004855, 0.00003333, 0.00000952, -0.00001017, -0.00001690, -0.00001090, },
		{ -0.00000046, -0.00000543, -0.00000031, 0.00000863, 0.00001256, 0.00000690, -0.00000511, -0.00001477, -0.00001431, -0.00000282, 0.00001281, 0.00002257, 0.00002022, 0.00000737, -0.00000811, -0.00001804, -0.00001968, -0.00001688, -0.00001598, -0.00001978, -0.00002461, -0.00002308, -0.00001087, 0.00000792, 0.00002154, 0.00001793, -0.00000588, -0.00003916, -0.00006180, -0.00005577, -0.00001685, 0.00004107, 0.00009233, 0.00011316, 0.00009441, 0.00004581, -0.00001039, -0.00005209, -0.00006822, -0.00006191, -0.00004502, -0.00002892, -0.00001805, -0.00000977, 0.00000047, 0.00001359, 0.00002533, 0.00002963, 0.00002378, 0.00001115, -0.00000088, -0.00000626, -0.00000448, -0.00000061, -0.00000086, -0.00000723, -0.00001540, -0.00001775, -0.00000930, 0.00000787, 0.00002522, 0.00003322, 0.00002783, 0.00001338, },
		{ 0.00002005, 0.00001571, 0.00000351, -0.00000818, -0.00001097, -0.00000250, 0.00001168, 0.00002187, 0.00002107, 0.00000994, -0.00000372, -0.00001050, -0.00000616, 0.00000555, 0.00001578, 0.00001711, 0.00000880, -0.00000311, -0.00001076, -0.00001076, -0.00000673, -0.00000613, -0.00001365, -0.00002637, -0.00003481, -0.00002968, -0.00000976, 0.00001526, 0.00002963, 0.00002169, -0.00000738, -0.00004242, -0.00006223, -0.00005205, -0.00001273, 0.00003919, 0.00008051, 0.00009400, 0.00007682, 0.00004018, 0.00000123, -0.00002683, -0.00004025, -0.00004333, -0.00004220, -0.00003937, -0.00003269, -0.00001905, 0.00000086, 0.00002095, 0.00003287, 0.00003155, 0.00001894, 0.00000298, -0.00000759, -0.00000890, -0.00000389, 0.00000071, -0.00000006, -0.00000548, -0.00000994, -0.00000770, 0.00000207, 0.00001398, },
		{ -0.00000293, 0.00000526, 0.00001545, 0.00002018, 0.00001404, -0.00000162, -0.00001865, -0.00002694, -0.00002110, -0.00000417, 0.00001438, 0.00002490, 0.00002381, 0.00001535, 0.00000772, 0.00000670, 0.00001143, 0.00001554, 0.00001264, 0.00000193, -0.00001061, -0.00001649, -0.00001149, 0.00000036, 0.00000893, 0.00000516, -0.00001171, -0.00003237, -0.00004294, -0.00003442, -0.00000963, 0.00001748, 0.00003012, 0.00001949, -0.00000908, -0.00003886, -0.00005159, -0.00003837, -0.00000462, 0.00003374, 0.00005992, 0.00006535, 0.00005274, 0.00003192, 0.00001237, -0.00000220, -0.00001382, -0.00002563, -0.00003719, -0.00004347, -0.00003855, -0.00002098, 0.00000359, 0.00002474, 0.00003342, 0.00002742, 0.00001240, -0.00000215, -0.00000917, -0.00000802, -0.00000351, -0.00000124, -0.00000287, -0.00000508, },
		},
		{ { -0.00001314, -0.00001399, -0.00001612, -0.00001837, -0.00001907, -0.00001643, -0.00000914, 0.00000304, 0.00001885, 0.00003551, 0.00004927, 0.00005644, 0.00005459, 0.00004362, 0.00002602, 0.00000634, -0.00001018, -0.00001942, -0.00001982, -0.00001305, -0.00000337, 0.00000393, 0.00000462, -0.00000267, -0.00001581, -0.00002980, -0.00003869, -0.00003787, -0.00002609, -0.00000613, 0.00001600, 0.00003313, 0.00003957, 0.00003313, 0.00001600, -0.00000613, -0.00002609, -0.00003787, -0.00003869, -0.00002980, -0.00001581, -0.00000267, 0.00000462, 0.00000393, -0.00000337, -0.00001305, -0.00001982, -0.00001942, -0.00001018, 0.00000634, 0.00002602, 0.00004362, 0.00005459, 0.00005644, 0.00004927, 0.00003551, 0.00001885, 0.00000304, -0.00000914, -0.00001643, -0.00001907, -0.00001837, -0.00001612, -0.00001399, },
		{ -0.00001303, -0.00001765, -0.00001880, -0.00001764, -0.00001562, -0.00001412, -0.00001407, -0.00001554, -0.00001763, -0.00001855, -0.00001622, -0.00000904, 0.00000322, 0.00001898, 0.00003493, 0.00004691, 0.00005125, 0.00004622, 0.00003287, 0.00001498, -0.00000201, -0.00001289, -0.00001458, -0.00000736, 0.00000507, 0.00001687, 0.00002219, 0.00001747, 0.00000292, -0.00001739, -0.00003690, -0.00004888, -0.00004896, -0.00003673, -0.00001595, 0.00000681, 0.00002452, 0.00003217, 0.00002846, 0.00001598, 0.00000006, -0.00001344, -0.00002031, -0.00001944, -0.00001296, -0.00000511, -0.00000034, -0.00000142, -0.00000838, -0.00001845, -0.00002724, -0.00003045, -0.00002543, -0.00001221, 0.00000663, 0.00002678, 0.00004368, 0.00005380, 0.00005554, 0.00004938, 0.00003743, 0.00002264, 0.00000793, -0.00000440, },
		{ 0.00004534, 0.00003833, 0.00002781, 0.00001565, 0.00000372, -0.00000639, -0.00001358, -0.00001746, -0.00001845, -0.00001764, -0.00001645, -0.00001601, -0.00001663, -0.00001751, -0.00001694, -0.00001299, -0.00000444, 0.00000834, 0.00002309, 0.00003607, 0.00004327, 0.00004187, 0.00003160, 0.00001511, -0.00000255, -0.00001559, -0.00001953, -0.00001288, 0.00000207, 0.00001993, 0.00003397, 0.00003854, 0.00003113, 0.00001329, -0.00000983, -0.00003131, -0.00004488, -0.00004707, -0.00003824, -0.00002229, -0.00000498, 0.00000819, 0.00001394, 0.00001216, 0.00000554, -0.00000179, -0.00000609, -0.00000553, -0.00000084, 0.00000522, 0.00000907, 0.00000796, 0.00000108, -0.00001001, -0.00002196, -0.00003081, -0.00003340, -0.00002831, -0.00001629, 0.00000018, 0.00001765, 0.00003278, 0.00004306, 0.00004725, },
		{ -0.00000816, 0.00000342, 0.00001505, 0.00002519, 0.00003253, 0.00003612, 0.00003545, 0.00003060, 0.00002232, 0.00001203, 0.00000154, -0.00000745, -0.00001381, -0.00001733, -0.00001862, -0.00001878, -0.00001870, -0.00001853, -0.00001747, -0.00001412, -0.00000720, 0.00000351, 0.00001660, 0.00002907, 0.00003718, 0.00003771, 0.00002941, 0.00001382, -0.00000490, -0.00002105, -0.00002932, -0.00002661, -0.00001328, 0.00000679, 0.00002742, 0.00004217, 0.00004647, 0.00003913, 0.00002260, 0.00000203, -0.00001667, -0.00002881, -0.00003239, -0.00002845, -0.00002030, -0.00001199, -0.00000667, -0.00000547, -0.00000735, -0.00000980, -0.00001014, -0.00000680, 0.00000000, 0.00000834, 0.00001534, 0.00001835, 0.00001595, 0.00000848, -0.00000214, -0.00001308, -0.00002152, -0.00002545, -0.00002410, -0.00001790, },
		{ 0.00000265, -0.00000347, -0.00000865, -0.00001227, -0.00001381, -0.00001280, -0.00000896, -0.00000241, 0.00000611, 0.00001518, 0.00002297, 0.00002772, 0.00002824, 0.00002438, 0.00001704, 0.00000788, -0.00000128, -0.00000902, -0.00001467, -0.00001829, -0.00002025, -0.00002077, -0.00001955, -0.00001583, -0.00000887, 0.00000143, 0.00001388, 0.00002597, 0.00003433, 0.00003585, 0.00002889, 0.00001418, -0.00000498, -0.00002353, -0.00003612, -0.00003883, -0.00003053, -0.00001334, 0.00000802, 0.00002773, 0.00004074, 0.00004427, 0.00003854, 0.00002638, 0.00001203, -0.00000048, -0.00000861, -0.00001194, -0.00001191, -0.00001085, -0.00001087, -0.00001287, -0.00001632, -0.00001954, -0.00002061, -0.00001819, -0.00001218, -0.00000372, 0.00000520, 0.00001248, 0.00001660, 0.00001700, 0.00001407, 0.00000887, },
		{ -0.00000066, 0.00000410, 0.00000856, 0.00001248, 0.00001528, 0.00001616, 0.00001440, 0.00000981, 0.00000301, -0.00000452, -0.00001080, -0.00001395, -0.00001285, -0.00000759, 0.00000053, 0.00000937, 0.00001667, 0.00002067, 0.00002061, 0.00001681, 0.00001032, 0.00000251, -0.00000537, -0.00001240, -0.00001789, -0.00002122, -0.00002165, -0.00001843, -0.00001111, -0.00000004, 0.00001326, 0.00002610, 0.00003514, 0.00003731, 0.00003095, 0.00001655, -0.00000304, -0.00002321, -0.00003888, -0.00004598, -0.00004274, -0.00003018, -0.00001175, 0.00000778, 0.00002384, 0.00003336, 0.00003554, 0.00003168, 0.00002446, 0.00001672, 0.00001052, 0.00000657, 0.00000428, 0.00000239, -0.00000031, -0.00000435, -0.00000934, -0.00001424, -0.00001782, -0.00001918, -0.00001809, -0.00001495, -0.00001053, -0.00000560, },
		{ -0.00000454, -0.00000633, -0.00000906, -0.00001222, -0.00001463, -0.00001487, -0.00001184, -0.00000540, 0.00000344, 0.00001265, 0.00001979, 0.00002284, 0.00002091, 0.00001453, 0.00000554, -0.00000354, -0.00001030, -0.00001320, -0.00001189, -0.00000722, -0.00000076, 0.00000565, 0.00001048, 0.00001275, 0.00001209, 0.00000858, 0.00000270, -0.00000474, -0.00001257, -0.00001922, -0.00002291, -0.00002194, -0.00001536, -0.00000345, 0.00001190, 0.00002741, 0.00003908, 0.00004331, 0.00003805, 0.00002362, 0.00000282, -0.00001969, -0.00003870, -0.00004991, -0.00005113, -0.00004280, -0.00002767, -0.00000983, 0.00000662, 0.00001870, 0.00002519, 0.00002659, 0.00002451, 0.00002086, 0.00001713, 0.00001398, 0.00001136, 0.00000887, 0.00000620, 0.00000337, 0.00000073, -0.00000133, -0.00000265, -0.00000352, },
		{ -0.00000409, -0.00000183, 0.00000325, 0.00000947, 0.00001448, 0.00001608, 0.00001307, 0.00000578, -0.00000394, -0.00001327, -0.00001938, -0.00002039, -0.00001600, -0.00000754, 0.00000254, 0.00001148, 0.00001713, 0.00001851, 0.00001591, 0.00001062, 0.00000433, -0.00000139, -0.00000547, -0.00000745, -0.00000735, -0.00000549, -0.00000241, 0.00000114, 0.00000422, 0.00000569, 0.00000451, 0.00000011, -0.00000712, -0.00001564, -0.00002290, -0.00002597, -0.00002250, -0.00001171, 0.00000500, 0.00002409, 0.00004074, 0.00005019, 0.00004921, 0.00003722, 0.00001655, -0.00000816, -0.00003133, -0.00004801, -0.00005521, -0.00005253, -0.00004192, -0.00002680, -0.00001084, 0.00000304, 0.00001327, 0.00001956, 0.00002243, 0.00002266, 0.00002084, 0.00001731, 0.00001236, 0.00000658, 0.00000101, -0.00000296, },
		{ 0.00002067, 0.00001026, -0.00000216, -0.00001334, -0.00002001, -0.00001998, -0.00001296, -0.00000083, 0.00001297, 0.00002447, 0.00003053, 0.00002976, 0.00002288, 0.00001227, 0.00000106, -0.00000797, -0.00001316, -0.00001429, -0.00001225, -0.00000846, -0.00000421, -0.00000031, 0.00000296, 0.00000554, 0.00000733, 0.00000801, 0.00000718, 0.00000468, 0.00000090, -0.00000316, -0.00000616, -0.00000699, -0.00000541, -0.00000231, 0.00000043, 0.00000065, -0.00000306, -0.00001054, -0.00001973, -0.00002716, -0.00002911, -0.00002302, -0.00000868, 0.00001133, 0.00003223, 0.00004840, 0.00005519, 0.00005036, 0.00003479, 0.00001214, -0.00001233, -0.00003344, -0.00004742, -0.00005279, -0.00005021, -0.00004181, -0.00003014, -0.00001733, -0.00000478, 0.00000675, 0.00001664, 0.00002404, 0.00002779, 0.00002680, },
		{ -0.00000132, 0.00001469, 0.00002844, 0.00003641, 0.00003600, 0.00002667, 0.00001042, -0.00000856, -0.00002514, -0.00003487, -0.00003536, -0.00002698, -0.00001250, 0.00000394, 0.00001836, 0.00002807, 0.00003227, 0.00003179, 0.00002823, 0.00002313, 0.00001728, 0.00001073, 0.00000327, -0.00000492, -0.00001289, -0.00001892, -0.00002124, -0.00001879, -0.00001199, -0.00000278, 0.00000589, 0.00001119, 0.00001149, 0.00000709, 0.00000005, -0.00000663, -0.00001037, -0.00001013, -0.00000688, -0.00000318, -0.00000196, -0.00000513, -0.00001247, -0.00002149, -0.00002826, -0.00002891, -0.00002129, -0.00000597, 0.00001373, 0.00003274, 0.00004596, 0.00004993, 0.00004384, 0.00002957, 0.00001086, -0.00000812, -0.00002406, -0.00003526, -0.00004149, -0.00004346, -0.00004187, -0.00003696, -0.00002851, -0.00001639, },
		{ -0.00003253, -0.00004019, -0.00004340, -0.00003951, -0.00002745, -0.00000868, 0.00001279, 0.00003146, 0.00004211, 0.00004164, 0.00003023, 0.00001126, -0.00000986, -0.00002755, -0.00003790, -0.00003970, -0.00003434, -0.00002477, -0.00001400, -0.00000386, 0.00000536, 0.00001445, 0.00002418, 0.00003422, 0.00004269, 0.00004667, 0.00004345, 0.00003196, 0.00001373, -0.00000721, -0.00002546, -0.00003615, -0.00003673, -0.00002797, -0.00001366, 0.00000079, 0.00001043, 0.00001255, 0.00000754, -0.00000157, -0.00001056, -0.00001587, -0.00001611, -0.00001239, -0.00000764, -0.00000508, -0.00000660, -0.00001187, -0.00001840, -0.00002263, -0.00002160, -0.00001417, -0.00000167, 0.00001267, 0.00002496, 0.00003216, 0.00003303, 0.00002839, 0.00002034, 0.00001119, 0.00000238, -0.00000588, -0.00001423, -0.00002326, },
		{ 0.00001139, 0.00001453, 0.00001429, 0.00000812, -0.00000438, -0.00002072, -0.00003609, -0.00004501, -0.00004349, -0.00003080, -0.00001001, 0.00001290, 0.00003118, 0.00003972, 0.00003681, 0.00002454, 0.00000770, -0.00000831, -0.00001964, -0.00002524, -0.00002678, -0.00002721, -0.00002888, -0.00003196, -0.00003408, -0.00003139, -0.00002057, -0.00000087, 0.00002477, 0.00005017, 0.00006792, 0.00007204, 0.00006045, 0.00003600, 0.00000561, -0.00002201, -0.00003947, -0.00004316, -0.00003431, -0.00001819, -0.00000186, 0.00000858, 0.00001008, 0.00000339, -0.00000774, -0.00001851, -0.00002509, -0.00002604, -0.00002245, -0.00001702, -0.00001251, -0.00001039, -0.00001035, -0.00001071, -0.00000954, -0.00000582, -0.00000005, 0.00000603, 0.00001036, 0.00001171, 0.00001037, 0.00000802, 0.00000680, 0.00000806, },
		{ -0.00000249, -0.00001238, -0.00001688, -0.00001342, -0.00000252, 0.00001196, 0.00002415, 0.00002843, 0.00002190, 0.00000581, -0.00001463, -0.00003212, -0.00004015, -0.00003559, -0.00002005, 0.00000077, 0.00001939, 0.00002956, 0.00002875, 0.00001896, 0.00000562, -0.00000510, -0.00000912, -0.00000639, -0.00000086, 0.00000147, -0.00000444, -0.00001985, -0.00004108, -0.00006046, -0.00006924, -0.00006118, -0.00003550, 0.00000234, 0.00004213, 0.00007249, 0.00008482, 0.00007635, 0.00005098, 0.00001765, -0.00001298, -0.00003238, -0.00003685, -0.00002827, -0.00001275, 0.00000213, 0.00001028, 0.00000908, -0.00000025, -0.00001380, -0.00002689, -0.00003600, -0.00003967, -0.00003842, -0.00003381, -0.00002733, -0.00001973, -0.00001114, -0.00000172, 0.00000768, 0.00001521, 0.00001862, 0.00001630, 0.00000845, },
		{ 0.00002284, 0.00003413, 0.00003472, 0.00002411, 0.00000557, -0.00001453, -0.00002884, -0.00003182, -0.00002215, -0.00000347, 0.00001681, 0.00003040, 0.00003150, 0.00001937, -0.00000122, -0.00002180, -0.00003371, -0.00003172, -0.00001639, 0.00000607, 0.00002635, 0.00003600, 0.00003106, 0.00001390, -0.00000776, -0.00002408, -0.00002751, -0.00001619, 0.00000489, 0.00002577, 0.00003551, 0.00002696, 0.00000005, -0.00003770, -0.00007368, -0.00009484, -0.00009267, -0.00006647, -0.00002361, 0.00002306, 0.00005994, 0.00007736, 0.00007269, 0.00005061, 0.00002081, -0.00000594, -0.00002165, -0.00002356, -0.00001428, -0.00000001, 0.00001226, 0.00001742, 0.00001362, 0.00000219, -0.00001352, -0.00002976, -0.00004356, -0.00005313, -0.00005754, -0.00005621, -0.00004870, -0.00003502, -0.00001637, 0.00000439, },
		{ -0.00007779, -0.00007052, -0.00005164, -0.00002412, 0.00000588, 0.00003046, 0.00004260, 0.00003890, 0.00002127, -0.00000331, -0.00002514, -0.00003547, -0.00003013, -0.00001137, 0.00001277, 0.00003165, 0.00003647, 0.00002420, -0.00000084, -0.00002841, -0.00004644, -0.00004614, -0.00002592, 0.00000753, 0.00004160, 0.00006273, 0.00006200, 0.00003896, 0.00000213, -0.00003422, -0.00005567, -0.00005356, -0.00002853, 0.00000946, 0.00004495, 0.00006299, 0.00005498, 0.00002204, -0.00002538, -0.00007146, -0.00010089, -0.00010442, -0.00008185, -0.00004166, 0.00000264, 0.00003762, 0.00005453, 0.00005180, 0.00003461, 0.00001210, -0.00000650, -0.00001515, -0.00001255, -0.00000174, 0.00001184, 0.00002270, 0.00002707, 0.00002357, 0.00001283, -0.00000334, -0.00002271, -0.00004285, -0.00006096, -0.00007376, },
		{ 0.00001459, -0.00000659, -0.00003295, -0.00005992, -0.00008119, -0.00009035, -0.00008335, -0.00006057, -0.00002763, 0.00000588, 0.00002954, 0.00003606, 0.00002436, 0.00000048, -0.00002449, -0.00003871, -0.00003454, -0.00001216, 0.00002006, 0.00004855, 0.00006002, 0.00004722, 0.00001253, -0.00003226, -0.00007014, -0.00008547, -0.00007036, -0.00002833, 0.00002660, 0.00007494, 0.00009912, 0.00009029, 0.00005182, -0.00000191, -0.00005103, -0.00007773, -0.00007300, -0.00003988, 0.00000803, 0.00005191, 0.00007503, 0.00006883, 0.00003572, -0.00001235, -0.00005889, -0.00008906, -0.00009487, -0.00007728, -0.00004492, -0.00001006, 0.00001620, 0.00002764, 0.00002432, 0.00001140, -0.00000368, -0.00001430, -0.00001681, -0.00001112, 0.00000021, 0.00001348, 0.00002518, 0.00003266, 0.00003411, 0.00002831, },
		{ -0.00000003, 0.00001410, 0.00002815, 0.00003866, 0.00004141, 0.00003277, 0.00001165, -0.00001901, -0.00005213, -0.00007819, -0.00008856, -0.00007910, -0.00005247, -0.00001794, 0.00001172, 0.00002512, 0.00001717, -0.00000833, -0.00003951, -0.00006097, -0.00006005, -0.00003267, 0.00001412, 0.00006402, 0.00009740, 0.00009908, 0.00006486, 0.00000404, -0.00006330, -0.00011344, -0.00012784, -0.00010033, -0.00003970, 0.00003338, 0.00009396, 0.00012193, 0.00010932, 0.00006282, 0.00000086, -0.00005371, -0.00008221, -0.00007670, -0.00004224, 0.00000590, 0.00004865, 0.00007045, 0.00006468, 0.00003522, -0.00000604, -0.00004441, -0.00006805, -0.00007185, -0.00005825, -0.00003515, -0.00001209, 0.00000346, 0.00000833, 0.00000384, -0.00000583, -0.00001584, -0.00002252, -0.00002407, -0.00002033, -0.00001197, },
		{ -0.00001345, -0.00002176, -0.00002949, -0.00003454, -0.00003414, -0.00002586, -0.00000931, 0.00001258, 0.00003349, 0.00004552, 0.00004227, 0.00002201, -0.00001054, -0.00004509, -0.00006930, -0.00007368, -0.00005605, -0.00002323, 0.00001095, 0.00003092, 0.00002603, -0.00000437, -0.00004949, -0.00009058, -0.00010806, -0.00008968, -0.00003639, 0.00003677, 0.00010538, 0.00014442, 0.00013788, 0.00008532, 0.00000269, -0.00008296, -0.00014313, -0.00015817, -0.00012419, -0.00005423, 0.00002692, 0.00009201, 0.00012115, 0.00010840, 0.00006289, 0.00000432, -0.00004509, -0.00006897, -0.00006214, -0.00003133, 0.00000868, 0.00004155, 0.00005563, 0.00004770, 0.00002302, -0.00000791, -0.00003405, -0.00004798, -0.00004796, -0.00003734, -0.00002222, -0.00000852, 0.00000004, 0.00000252, -0.00000002, -0.00000583, },
		{ -0.00000956, -0.00000040, 0.00000787, 0.00001375, 0.00001477, 0.00000848, -0.00000583, -0.00002564, -0.00004507, -0.00005662, -0.00005422, -0.00003640, -0.00000795, 0.00002114, 0.00003929, 0.00003824, 0.00001714, -0.00001620, -0.00004783, -0.00006322, -0.00005371, -0.00002106, 0.00002219, 0.00005708, 0.00006581, 0.00003968, -0.00001616, -0.00008332, -0.00013639, -0.00015252, -0.00012076, -0.00004726, 0.00004591, 0.00012817, 0.00017142, 0.00016038, 0.00009828, 0.00000581, -0.00008644, -0.00014895, -0.00016362, -0.00012951, -0.00006212, 0.00001327, 0.00007137, 0.00009579, 0.00008395, 0.00004639, 0.00000127, -0.00003333, -0.00004615, -0.00003599, -0.00001071, 0.00001723, 0.00003627, 0.00004011, 0.00002918, 0.00000926, -0.00001177, -0.00002742, -0.00003461, -0.00003381, -0.00002763, -0.00001892, },
		{ -0.00001793, -0.00002648, -0.00003274, -0.00003594, -0.00003396, -0.00002446, -0.00000686, 0.00001598, 0.00003755, 0.00004959, 0.00004550, 0.00002375, -0.00001032, -0.00004576, -0.00006985, -0.00007343, -0.00005513, -0.00002263, 0.00001005, 0.00002833, 0.00002359, -0.00000250, -0.00003789, -0.00006481, -0.00006737, -0.00003911, 0.00001309, 0.00007045, 0.00010918, 0.00011006, 0.00006692, -0.00000956, -0.00009471, -0.00015848, -0.00017666, -0.00014035, -0.00005983, 0.00003879, 0.00012281, 0.00016481, 0.00015226, 0.00009162, 0.00000521, -0.00007745, -0.00013052, -0.00014059, -0.00011025, -0.00005546, 0.00000174, 0.00004194, 0.00005512, 0.00004289, 0.00001608, -0.00001095, -0.00002654, -0.00002586, -0.00001164, 0.00000830, 0.00002521, 0.00003309, 0.00003056, 0.00002020, 0.00000643, -0.00000690, },
		{ 0.00002585, 0.00002672, 0.00002555, 0.00002248, 0.00001570, 0.00000282, -0.00001679, -0.00003996, -0.00005963, -0.00006707, -0.00005572, -0.00002502, 0.00001789, 0.00005973, 0.00008566, 0.00008531, 0.00005742, 0.00001115, -0.00003705, -0.00006989, -0.00007637, -0.00005659, -0.00002190, 0.00000991, 0.00002241, 0.00000813, -0.00002747, -0.00006767, -0.00009128, -0.00008197, -0.00003608, 0.00003448, 0.00010581, 0.00015087, 0.00015023, 0.00010028, 0.00001544, -0.00007655, -0.00014476, -0.00016649, -0.00013562, -0.00006459, 0.00002061, 0.00009036, 0.00012226, 0.00010875, 0.00005869, -0.00000726, -0.00006528, -0.00009744, -0.00009749, -0.00007163, -0.00003435, -0.00000162, 0.00001555, 0.00001459, 0.00000088, -0.00001594, -0.00002683, -0.00002708, -0.00001747, -0.00000273, 0.00001145, 0.00002124, },
		{ -0.00002909, -0.00002210, -0.00001596, -0.00001112, -0.00000515, 0.00000536, 0.00002178, 0.00004113, 0.00005605, 0.00005745, 0.00003905, 0.00000145, -0.00004618, -0.00008835, -0.00010871, -0.00009681, -0.00005324, 0.00000947, 0.00007080, 0.00010985, 0.00011353, 0.00008189, 0.00002830, -0.00002603, -0.00006071, -0.00006465, -0.00004060, -0.00000374, 0.00002515, 0.00002915, 0.00000306, -0.00004363, -0.00009024, -0.00011346, -0.00009749, -0.00004138, 0.00003951, 0.00011885, 0.00016924, 0.00017279, 0.00012806, 0.00005058, -0.00003322, -0.00009609, -0.00012001, -0.00010220, -0.00005503, -0.00000006, 0.00004097, 0.00005436, 0.00003883, 0.00000445, -0.00003275, -0.00005804, -0.00006375, -0.00005138, -0.00002960, -0.00000945, 0.00000079, -0.00000122, -0.00001189, -0.00002445, -0.00003268, -0.00003381, },
		{ -0.00001142, -0.00001932, -0.00002216, -0.00002033, -0.00001819, -0.00002114, -0.00003176, -0.00004707, -0.00005879, -0.00005680, -0.00003442, 0.00000695, 0.00005643, 0.00009657, 0.00010996, 0.00008677, 0.00003012, -0.00004356, -0.00010970, -0.00014466, -0.00013492, -0.00008288, -0.00000635, 0.00006802, 0.00011503, 0.00012059, 0.00008695, 0.00003124, -0.00002209, -0.00005163, -0.00004784, -0.00001669, 0.00002346, 0.00005035, 0.00004788, 0.00001350, -0.00004022, -0.00009071, -0.00011474, -0.00009825, -0.00004241, 0.00003648, 0.00011333, 0.00016385, 0.00017362, 0.00014289, 0.00008548, 0.00002239, -0.00002674, -0.00005083, -0.00005034, -0.00003523, -0.00001919, -0.00001309, -0.00002050, -0.00003705, -0.00005360, -0.00006128, -0.00005589, -0.00003968, -0.00001979, -0.00000437, 0.00000139, -0.00000248, },
		{ -0.00002506, -0.00001151, -0.00001061, -0.00001854, -0.00002677, -0.00002666, -0.00001442, 0.00000599, 0.00002397, 0.00002698, 0.00000713, -0.00003364, -0.00008240, -0.00011925, -0.00012526, -0.00009107, -0.00002238, 0.00006049, 0.00012919, 0.00015802, 0.00013421, 0.00006370, -0.00003028, -0.00011551, -0.00016282, -0.00015710, -0.00010279, -0.00002176, 0.00005564, 0.00010260, 0.00010602, 0.00007079, 0.00001667, -0.00003081, -0.00005131, -0.00003799, 0.00000038, 0.00004395, 0.00007100, 0.00006756, 0.00003348, -0.00001759, -0.00006411, -0.00008568, -0.00007135, -0.00002376, 0.00004242, 0.00010670, 0.00015079, 0.00016499, 0.00015052, 0.00011713, 0.00007773, 0.00004259, 0.00001588, -0.00000433, -0.00002296, -0.00004376, -0.00006632, -0.00008585, -0.00009574, -0.00009149, -0.00007366, -0.00004844, },
		{ -0.00013670, -0.00012291, -0.00008973, -0.00004988, -0.00001885, -0.00000848, -0.00002142, -0.00004921, -0.00007536, -0.00008245, -0.00006041, -0.00001228, 0.00004550, 0.00008889, 0.00009618, 0.00005790, -0.00001756, -0.00010541, -0.00017309, -0.00019271, -0.00015253, -0.00006257, 0.00004812, 0.00014187, 0.00018635, 0.00016684, 0.00009163, -0.00001133, -0.00010527, -0.00015863, -0.00015654, -0.00010545, -0.00002924, 0.00004115, 0.00008030, 0.00007777, 0.00004093, -0.00000961, -0.00004969, -0.00006189, -0.00004233, -0.00000124, 0.00004251, 0.00007014, 0.00007104, 0.00004688, 0.00001022, -0.00002134, -0.00003268, -0.00001702, 0.00002215, 0.00007341, 0.00012265, 0.00015842, 0.00017519, 0.00017345, 0.00015722, 0.00013074, 0.00009635, 0.00005458, 0.00000622, -0.00004520, -0.00009236, -0.00012562, },
		{ 0.00005278, -0.00003977, -0.00012638, -0.00018779, -0.00020849, -0.00018334, -0.00012132, -0.00004431, 0.00001970, 0.00004672, 0.00002650, -0.00003248, -0.00010517, -0.00015906, -0.00016710, -0.00011929, -0.00002799, 0.00007552, 0.00015242, 0.00017145, 0.00012165, 0.00001767, -0.00010481, -0.00020210, -0.00023878, -0.00020116, -0.00010235, 0.00002279, 0.00013094, 0.00018645, 0.00017412, 0.00010373, 0.00000492, -0.00008514, -0.00013616, -0.00013545, -0.00009101, -0.00002641, 0.00003002, 0.00005685, 0.00004724, 0.00001007, -0.00003545, -0.00006884, -0.00007694, -0.00005849, -0.00002325, 0.00001347, 0.00003805, 0.00004423, 0.00003500, 0.00002035, 0.00001232, 0.00001983, 0.00004556, 0.00008583, 0.00013294, 0.00017839, 0.00021527, 0.00023882, 0.00024556, 0.00023220, 0.00019563, 0.00013466, },
		{ 0.00038107, 0.00039836, 0.00037396, 0.00029957, 0.00017663, 0.00002011, -0.00014196, -0.00027422, -0.00034416, -0.00033354, -0.00024659, -0.00011121, 0.00002841, 0.00012590, 0.00014890, 0.00009112, -0.00002449, -0.00015337, -0.00024532, -0.00026318, -0.00019692, -0.00006793, 0.00007873, 0.00019089, 0.00022878, 0.00017969, 0.00006244, -0.00008008, -0.00019720, -0.00024921, -0.00022134, -0.00012792, -0.00000543, 0.00010220, 0.00016014, 0.00015460, 0.00009584, 0.00001196, -0.00006354, -0.00010498, -0.00010325, -0.00006714, -0.00001755, 0.00002242, 0.00003695, 0.00002285, -0.00001101, -0.00004908, -0.00007647, -0.00008497, -0.00007528, -0.00005504, -0.00003418, -0.00001998, -0.00001409, -0.00001247, -0.00000783, 0.00000699, 0.00003642, 0.00008115, 0.00013869, 0.00020450, 0.00027254, 0.00033488, },
		{ 0.00005074, 0.00014094, 0.00025356, 0.00038041, 0.00050498, 0.00060250, 0.00064395, 0.00060389, 0.00047039, 0.00025341, -0.00001249, -0.00027353, -0.00047007, -0.00055445, -0.00050726, -0.00034591, -0.00012152, 0.00009605, 0.00024024, 0.00026974, 0.00018220, 0.00001465, -0.00016987, -0.00030475, -0.00034322, -0.00027422, -0.00012509, 0.00004972, 0.00018922, 0.00024804, 0.00021181, 0.00010040, -0.00004148, -0.00016210, -0.00022190, -0.00020656, -0.00012978, -0.00002538, 0.00006723, 0.00011806, 0.00011640, 0.00007212, 0.00000904, -0.00004616, -0.00007460, -0.00007123, -0.00004451, -0.00001074, 0.00001384, 0.00001959, 0.00000616, -0.00001920, -0.00004619, -0.00006650, -0.00007684, -0.00007916, -0.00007835, -0.00007897, -0.00008271, -0.00008761, -0.00008894, -0.00008083, -0.00005770, -0.00001497, },
		{ -0.00014617, -0.00016975, -0.00018435, -0.00018087, -0.00014748, -0.00007130, 0.00005752, 0.00023961, 0.00046039, 0.00068712, 0.00087209, 0.00096279, 0.00091721, 0.00071997, 0.00039294, -0.00000453, -0.00038788, -0.00066968, -0.00078535, -0.00071392, -0.00048598, -0.00017540, 0.00012317, 0.00032388, 0.00037623, 0.00027963, 0.00008100, -0.00014344, -0.00031536, -0.00038082, -0.00032661, -0.00018144, -0.00000240, 0.00014750, 0.00022186, 0.00020548, 0.00011634, -0.00000463, -0.00011085, -0.00016753, -0.00016256, -0.00010772, -0.00003076, 0.00003728, 0.00007412, 0.00007310, 0.00004287, 0.00000121, -0.00003363, -0.00005019, -0.00004696, -0.00003082, -0.00001232, -0.00000045, 0.00000073, -0.00000741, -0.00002030, -0.00003324, -0.00004387, -0.00005285, -0.00006287, -0.00007681, -0.00009611, -0.00012013, },
		{ -0.00002532, -0.00003155, -0.00004542, -0.00006958, -0.00010649, -0.00015748, -0.00022036, -0.00028621, -0.00033694, -0.00034584, -0.00028286, -0.00012451, 0.00013413, 0.00046975, 0.00082834, 0.00113315, 0.00130288, 0.00127564, 0.00103116, 0.00060314, 0.00007588, -0.00043541, -0.00081578, -0.00098360, -0.00091460, -0.00064872, -0.00027702, 0.00008710, 0.00034295, 0.00043202, 0.00035244, 0.00015432, -0.00008145, -0.00027253, -0.00036220, -0.00033491, -0.00021618, -0.00005846, 0.00008062, 0.00015885, 0.00016187, 0.00010372, 0.00001725, -0.00006094, -0.00010441, -0.00010468, -0.00007093, -0.00002301, 0.00001825, 0.00003925, 0.00003728, 0.00001903, -0.00000436, -0.00002297, -0.00003184, -0.00003168, -0.00002684, -0.00002214, -0.00002035, -0.00002149, -0.00002367, -0.00002493, -0.00002464, -0.00002396, },
		{ -0.00004261, -0.00004372, -0.00004117, -0.00003509, -0.00002588, -0.00001426, -0.00000225, 0.00000520, -0.00000096, -0.00003290, -0.00010178, -0.00021089, -0.00034844, -0.00048342, -0.00056825, -0.00054953, -0.00038528, -0.00006370, 0.00038373, 0.00088091, 0.00132096, 0.00159276, 0.00161338, 0.00135570, 0.00086167, 0.00023531, -0.00038374, -0.00085848, -0.00109284, -0.00105761, -0.00079583, -0.00040623, -0.00001063, 0.00028342, 0.00041270, 0.00037168, 0.00020643, -0.00000707, -0.00019144, -0.00029289, -0.00029463, -0.00021582, -0.00009848, 0.00001124, 0.00007983, 0.00009555, 0.00006788, 0.00001941, -0.00002557, -0.00005049, -0.00005097, -0.00003341, -0.00000960, 0.00000942, 0.00001771, 0.00001537, 0.00000661, -0.00000351, -0.00001170, -0.00001737, -0.00002187, -0.00002678, -0.00003261, -0.00003843, },
		{ -0.00000662, -0.00001377, -0.00002273, -0.00003252, -0.00004231, -0.00005184, -0.00006101, -0.00006858, -0.00007096, -0.00006248, -0.00003785, 0.00000342, 0.00005250, 0.00008945, 0.00008654, 0.00001734, -0.00013055, -0.00034357, -0.00057690, -0.00075986, -0.00081318, -0.00067399, -0.00032112, 0.00020882, 0.00082181, 0.00138566, 0.00176411, 0.00185578, 0.00162566, 0.00111838, 0.00044809, -0.00023260, -0.00077494, -0.00107449, -0.00109692, -0.00088184, -0.00052457, -0.00014301, 0.00015861, 0.00031752, 0.00032388, 0.00021415, 0.00005119, -0.00010060, -0.00019589, -0.00021873, -0.00018084, -0.00011086, -0.00004019, 0.00000877, 0.00002816, 0.00002328, 0.00000667, -0.00000913, -0.00001658, -0.00001468, -0.00000713, 0.00000101, 0.00000622, 0.00000764, 0.00000641, 0.00000415, 0.00000160, -0.00000168, },
		{ 0.00000494, 0.00000499, 0.00000473, 0.00000327, 0.00000020, -0.00000389, -0.00000770, -0.00001053, -0.00001371, -0.00002096, -0.00003686, -0.00006349, -0.00009690, -0.00012551, -0.00013245, -0.00010217, -0.00002956, 0.00007240, 0.00017064, 0.00021868, 0.00017161, 0.00000521, -0.00026757, -0.00059046, -0.00087243, -0.00101094, -0.00092384, -0.00057977, -0.00001613, 0.00066305, 0.00131061, 0.00177523, 0.00194401, 0.00177523, 0.00131061, 0.00066305, -0.00001613, -0.00057977, -0.00092384, -0.00101094, -0.00087243, -0.00059046, -0.00026757, 0.00000521, 0.00017161, 0.00021868, 0.00017064, 0.00007240, -0.00002956, -0.00010217, -0.00013245, -0.00012551, -0.00009690, -0.00006349, -0.00003686, -0.00002096, -0.00001371, -0.00001053, -0.00000770, -0.00000389, 0.00000020, 0.00000327, 0.00000473, 0.00000499, },
		{ -0.00000662, -0.00000168, 0.00000160, 0.00000415, 0.00000641, 0.00000764, 0.00000622, 0.00000101, -0.00000713, -0.00001468, -0.00001658, -0.00000913, 0.00000667, 0.00002328, 0.00002816, 0.00000877, -0.00004019, -0.00011086, -0.00018084, -0.00021873, -0.00019589, -0.00010060, 0.00005119, 0.00021415, 0.00032388, 0.00031752, 0.00015861, -0.00014301, -0.00052457, -0.00088184, -0.00109692, -0.00107449, -0.00077494, -0.00023260, 0.00044809, 0.00111838, 0.00162566, 0.00185578, 0.00176411, 0.00138566, 0.00082181, 0.00020882, -0.00032112, -0.00067399, -0.00081318, -0.00075986, -0.00057690, -0.00034357, -0.00013055, 0.00001734, 0.00008654, 0.00008945, 0.00005250, 0.00000342, -0.00003785, -0.00006248, -0.00007096, -0.00006858, -0.00006101, -0.00005184, -0.00004231, -0.00003252, -0.00002273, -0.00001377, },
		{ -0.00004261, -0.00003843, -0.00003261, -0.00002678, -0.00002187, -0.00001737, -0.00001170, -0.00000351, 0.00000661, 0.00001537, 0.00001771, 0.00000942, -0.00000960, -0.00003341, -0.00005097, -0.00005049, -0.00002557, 0.00001941, 0.00006788, 0.00009555, 0.00007983, 0.00001124, -0.00009848, -0.00021582, -0.00029463, -0.00029289, -0.00019144, -0.00000707, 0.00020643, 0.00037168, 0.00041270, 0.00028342, -0.00001063, -0.00040623, -0.00079583, -0.00105761, -0.00109284, -0.00085848, -0.00038374, 0.00023531, 0.00086167, 0.00135570, 0.00161338, 0.00159276, 0.00132096, 0.00088091, 0.00038373, -0.00006370, -0.00038528, -0.00054953, -0.00056825, -0.00048342, -0.00034844, -0.00021089, -0.00010178, -0.00003290, -0.00000096, 0.00000520, -0.00000225, -0.00001426, -0.00002588, -0.00003509, -0.00004117, -0.00004372, },
		{ -0.00002532, -0.00002396, -0.00002464, -0.00002493, -0.00002367, -0.00002149, -0.00002035, -0.00002214, -0.00002684, -0.00003168, -0.00003184, -0.00002297, -0.00000436, 0.00001903, 0.00003728, 0.00003925, 0.00001825, -0.00002301, -0.00007093, -0.00010468, -0.00010441, -0.00006094, 0.00001725, 0.00010372, 0.00016187, 0.00015885, 0.00008062, -0.00005846, -0.00021618, -0.00033491, -0.00036220, -0.00027253, -0.00008145, 0.00015432, 0.00035244, 0.00043202, 0.00034295, 0.00008710, -0.00027702, -0.00064872, -0.00091460, -0.00098360, -0.00081578, -0.00043541, 0.00007588, 0.00060314, 0.00103116, 0.00127564, 0.00130288, 0.00113315, 0.00082834, 0.00046975, 0.00013413, -0.00012451, -0.00028286, -0.00034584, -0.00033694, -0.00028621, -0.00022036, -0.00015748, -0.00010649, -0.00006958, -0.00004542, -0.00003155, },
		{ -0.00014617, -0.00012013, -0.00009611, -0.00007681, -0.00006287, -0.00005285, -0.00004387, -0.00003324, -0.00002030, -0.00000741, 0.00000073, -0.00000045, -0.00001232, -0.00003082, -0.00004696, -0.00005019, -0.00003363, 0.00000121, 0.00004287, 0.00007310, 0.00007412, 0.00003728, -0.00003076, -0.00010772, -0.00016256, -0.00016753, -0.00011085, -0.00000463, 0.00011634, 0.00020548, 0.00022186, 0.00014750, -0.00000240, -0.00018144, -0.00032661, -0.00038082, -0.00031536, -0.00014344, 0.00008100, 0.00027963, 0.00037623, 0.00032388, 0.00012317, -0.00017540, -0.00048598, -0.00071392, -0.00078535, -0.00066968, -0.00038788, -0.00000453, 0.00039294, 0.00071997, 0.00091721, 0.00096279, 0.00087209, 0.00068712, 0.00046039, 0.00023961, 0.00005752, -0.00007130, -0.00014748, -0.00018087, -0.00018435, -0.00016975, },
		{ 0.00005074, -0.00001497, -0.00005770, -0.00008083, -0.00008894, -0.00008761, -0.00008271, -0.00007897, -0.00007835, -0.00007916, -0.00007684, -0.00006650, -0.00004619, -0.00001920, 0.00000616, 0.00001959, 0.00001384, -0.00001074, -0.00004451, -0.00007123, -0.00007460, -0.00004616, 0.00000904, 0.00007212, 0.00011640, 0.00011806, 0.00006723, -0.00002538, -0.00012978, -0.00020656, -0.00022190, -0.00016210, -0.00004148, 0.00010040, 0.00021181, 0.00024804, 0.00018922, 0.00004972, -0.00012509, -0.00027422, -0.00034322, -0.00030475, -0.00016987, 0.00001465, 0.00018220, 0.00026974, 0.00024024, 0.00009605, -0.00012152, -0.00034591, -0.00050726, -0.00055445, -0.00047007, -0.00027353, -0.00001249, 0.00025341, 0.00047039, 0.00060389, 0.00064395, 0.00060250, 0.00050498, 0.00038041, 0.00025356, 0.00014094, },
		{ 0.00038107, 0.00033488, 0.00027254, 0.00020450, 0.00013869, 0.00008115, 0.00003642, 0.00000699, -0.00000783, -0.00001247, -0.00001409, -0.00001998, -0.00003418, -0.00005504, -0.00007528, -0.00008497, -0.00007647, -0.00004908, -0.00001101, 0.00002285, 0.00003695, 0.00002242, -0.00001755, -0.00006714, -0.00010325, -0.00010498, -0.00006354, 0.00001196, 0.00009584, 0.00015460, 0.00016014, 0.00010220, -0.00000543, -0.00012792, -0.00022134, -0.00024921, -0.00019720, -0.00008008, 0.00006244, 0.00017969, 0.00022878, 0.00019089, 0.00007873, -0.00006793, -0.00019692, -0.00026318, -0.00024532, -0.00015337, -0.00002449, 0.00009112, 0.00014890, 0.00012590, 0.00002841, -0.00011121, -0.00024659, -0.00033354, -0.00034416, -0.00027422, -0.00014196, 0.00002011, 0.00017663, 0.00029957, 0.00037396, 0.00039836, },
		{ 0.00005278, 0.00013466, 0.00019563, 0.00023220, 0.00024556, 0.00023882, 0.00021527, 0.00017839, 0.00013294, 0.00008583, 0.00004556, 0.00001983, 0.00001232, 0.00002035, 0.00003500, 0.00004423, 0.00003805, 0.00001347, -0.00002325, -0.00005849, -0.00007694, -0.00006884, -0.00003545, 0.00001007, 0.00004724, 0.00005685, 0.00003002, -0.00002641, -0.00009101, -0.00013545, -0.00013616, -0.00008514, 0.00000492, 0.00010373, 0.00017412, 0.00018645, 0.00013094, 0.00002279, -0.00010235, -0.00020116, -0.00023878, -0.00020210, -0.00010481, 0.00001767, 0.00012165, 0.00017145, 0.00015242, 0.00007552, -0.00002799, -0.00011929, -0.00016710, -0.00015906, -0.00010517, -0.00003248, 0.00002650, 0.00004672, 0.00001970, -0.00004431, -0.00012132, -0.00018334, -0.00020849, -0.00018779, -0.00012638, -0.00003977, },
		{ -0.00013670, -0.00012562, -0.00009236, -0.00004520, 0.00000622, 0.00005458, 0.00009635, 0.00013074, 0.00015722, 0.00017345, 0.00017519, 0.00015842, 0.00012265, 0.00007341, 0.00002215, -0.00001702, -0.00003268, -0.00002134, 0.00001022, 0.00004688, 0.00007104, 0.00007014, 0.00004251, -0.00000124, -0.00004233, -0.00006189, -0.00004969, -0.00000961, 0.00004093, 0.00007777, 0.00008030, 0.00004115, -0.00002924, -0.00010545, -0.00015654, -0.00015863, -0.00010527, -0.00001133, 0.00009163, 0.00016684, 0.00018635, 0.00014187, 0.00004812, -0.00006257, -0.00015253, -0.00019271, -0.00017309, -0.00010541, -0.00001756, 0.00005790, 0.00009618, 0.00008889, 0.00004550, -0.00001228, -0.00006041, -0.00008245, -0.00007536, -0.00004921, -0.00002142, -0.00000848, -0.00001885, -0.00004988, -0.00008973, -0.00012291, },
		{ -0.00002506, -0.00004844, -0.00007366, -0.00009149, -0.00009574, -0.00008585, -0.00006632, -0.00004376, -0.00002296, -0.00000433, 0.00001588, 0.00004259, 0.00007773, 0.00011713, 0.00015052, 0.00016499, 0.00015079, 0.00010670, 0.00004242, -0.00002376, -0.00007135, -0.00008568, -0.00006411, -0.00001759, 0.00003348, 0.00006756, 0.00007100, 0.00004395, 0.00000038, -0.00003799, -0.00005131, -0.00003081, 0.00001667, 0.00007079, 0.00010602, 0.00010260, 0.00005564, -0.00002176, -0.00010279, -0.00015710, -0.00016282, -0.00011551, -0.00003028, 0.00006370, 0.00013421, 0.00015802, 0.00012919, 0.00006049, -0.00002238, -0.00009107, -0.00012526, -0.00011925, -0.00008240, -0.00003364, 0.00000713, 0.00002698, 0.00002397, 0.00000599, -0.00001442, -0.00002666, -0.00002677, -0.00001854, -0.00001061, -0.00001151, },
		{ -0.00001142, -0.00000248, 0.00000139, -0.00000437, -0.00001979, -0.00003968, -0.00005589, -0.00006128, -0.00005360, -0.00003705, -0.00002050, -0.00001309, -0.00001919, -0.00003523, -0.00005034, -0.00005083, -0.00002674, 0.00002239, 0.00008548, 0.00014289, 0.00017362, 0.00016385, 0.00011333, 0.00003648, -0.00004241, -0.00009825, -0.00011474, -0.00009071, -0.00004022, 0.00001350, 0.00004788, 0.00005035, 0.00002346, -0.00001669, -0.00004784, -0.00005163, -0.00002209, 0.00003124, 0.00008695, 0.00012059, 0.00011503, 0.00006802, -0.00000635, -0.00008288, -0.00013492, -0.00014466, -0.00010970, -0.00004356, 0.00003012, 0.00008677, 0.00010996, 0.00009657, 0.00005643, 0.00000695, -0.00003442, -0.00005680, -0.00005879, -0.00004707, -0.00003176, -0.00002114, -0.00001819, -0.00002033, -0.00002216, -0.00001932, },
		{ -0.00002909, -0.00003381, -0.00003268, -0.00002445, -0.00001189, -0.00000122, 0.00000079, -0.00000945, -0.00002960, -0.00005138, -0.00006375, -0.00005804, -0.00003275, 0.00000445, 0.00003883, 0.00005436, 0.00004097, -0.00000006, -0.00005503, -0.00010220, -0.00012001, -0.00009609, -0.00003322, 0.00005058, 0.00012806, 0.00017279, 0.00016924, 0.00011885, 0.00003951, -0.00004138, -0.00009749, -0.00011346, -0.00009024, -0.00004363, 0.00000306, 0.00002915, 0.00002515, -0.00000374, -0.00004060, -0.00006465, -0.00006071, -0.00002603, 0.00002830, 0.00008189, 0.00011353, 0.00010985, 0.00007080, 0.00000947, -0.00005324, -0.00009681, -0.00010871, -0.00008835, -0.00004618, 0.00000145, 0.00003905, 0.00005745, 0.00005605, 0.00004113, 0.00002178, 0.00000536, -0.00000515, -0.00001112, -0.00001596, -0.00002210, },
		{ 0.00002585, 0.00002124, 0.00001145, -0.00000273, -0.00001747, -0.00002708, -0.00002683, -0.00001594, 0.00000088, 0.00001459, 0.00001555, -0.00000162, -0.00003435, -0.00007163, -0.00009749, -0.00009744, -0.00006528, -0.00000726, 0.00005869, 0.00010875, 0.00012226, 0.00009036, 0.00002061, -0.00006459, -0.00013562, -0.00016649, -0.00014476, -0.00007655, 0.00001544, 0.00010028, 0.00015023, 0.00015087, 0.00010581, 0.00003448, -0.00003608, -0.00008197, -0.00009128, -0.00006767, -0.00002747, 0.00000813, 0.00002241, 0.00000991, -0.00002190, -0.00005659, -0.00007637, -0.00006989, -0.00003705, 0.00001115, 0.00005742, 0.00008531, 0.00008566, 0.00005973, 0.00001789, -0.00002502, -0.00005572, -0.00006707, -0.00005963, -0.00003996, -0.00001679, 0.00000282, 0.00001570, 0.00002248, 0.00002555, 0.00002672, },
		{ -0.00001793, -0.00000690, 0.00000643, 0.00002020, 0.00003056, 0.00003309, 0.00002521, 0.00000830, -0.00001164, -0.00002586, -0.00002654, -0.00001095, 0.00001608, 0.00004289, 0.00005512, 0.00004194, 0.00000174, -0.00005546, -0.00011025, -0.00014059, -0.00013052, -0.00007745, 0.00000521, 0.00009162, 0.00015226, 0.00016481, 0.00012281, 0.00003879, -0.00005983, -0.00014035, -0.00017666, -0.00015848, -0.00009471, -0.00000956, 0.00006692, 0.00011006, 0.00010918, 0.00007045, 0.00001309, -0.00003911, -0.00006737, -0.00006481, -0.00003789, -0.00000250, 0.00002359, 0.00002833, 0.00001005, -0.00002263, -0.00005513, -0.00007343, -0.00006985, -0.00004576, -0.00001032, 0.00002375, 0.00004550, 0.00004959, 0.00003755, 0.00001598, -0.00000686, -0.00002446, -0.00003396, -0.00003594, -0.00003274, -0.00002648, },
		{ -0.00000956, -0.00001892, -0.00002763, -0.00003381, -0.00003461, -0.00002742, -0.00001177, 0.00000926, 0.00002918, 0.00004011, 0.00003627, 0.00001723, -0.00001071, -0.00003599, -0.00004615, -0.00003333, 0.00000127, 0.00004639, 0.00008395, 0.00009579, 0.00007137, 0.00001327, -0.00006212, -0.00012951, -0.00016362, -0.00014895, -0.00008644, 0.00000581, 0.00009828, 0.00016038, 0.00017142, 0.00012817, 0.00004591, -0.00004726, -0.00012076, -0.00015252, -0.00013639, -0.00008332, -0.00001616, 0.00003968, 0.00006581, 0.00005708, 0.00002219, -0.00002106, -0.00005371, -0.00006322, -0.00004783, -0.00001620, 0.00001714, 0.00003824, 0.00003929, 0.00002114, -0.00000795, -0.00003640, -0.00005422, -0.00005662, -0.00004507, -0.00002564, -0.00000583, 0.00000848, 0.00001477, 0.00001375, 0.00000787, -0.00000040, },
		{ -0.00001345, -0.00000583, -0.00000002, 0.00000252, 0.00000004, -0.00000852, -0.00002222, -0.00003734, -0.00004796, -0.00004798, -0.00003405, -0.00000791, 0.00002302, 0.00004770, 0.00005563, 0.00004155, 0.00000868, -0.00003133, -0.00006214, -0.00006897, -0.00004509, 0.00000432, 0.00006289, 0.00010840, 0.00012115, 0.00009201, 0.00002692, -0.00005423, -0.00012419, -0.00015817, -0.00014313, -0.00008296, 0.00000269, 0.00008532, 0.00013788, 0.00014442, 0.00010538, 0.00003677, -0.00003639, -0.00008968, -0.00010806, -0.00009058, -0.00004949, -0.00000437, 0.00002603, 0.00003092, 0.00001095, -0.00002323, -0.00005605, -0.00007368, -0.00006930, -0.00004509, -0.00001054, 0.00002201, 0.00004227, 0.00004552, 0.00003349, 0.00001258, -0.00000931, -0.00002586, -0.00003414, -0.00003454, -0.00002949, -0.00002176, },
		{ -0.00000003, -0.00001197, -0.00002033, -0.00002407, -0.00002252, -0.00001584, -0.00000583, 0.00000384, 0.00000833, 0.00000346, -0.00001209, -0.00003515, -0.00005825, -0.00007185, -0.00006805, -0.00004441, -0.00000604, 0.00003522, 0.00006468, 0.00007045, 0.00004865, 0.00000590, -0.00004224, -0.00007670, -0.00008221, -0.00005371, 0.00000086, 0.00006282, 0.00010932, 0.00012193, 0.00009396, 0.00003338, -0.00003970, -0.00010033, -0.00012784, -0.00011344, -0.00006330, 0.00000404, 0.00006486, 0.00009908, 0.00009740, 0.00006402, 0.00001412, -0.00003267, -0.00006005, -0.00006097, -0.00003951, -0.00000833, 0.00001717, 0.00002512, 0.00001172, -0.00001794, -0.00005247, -0.00007910, -0.00008856, -0.00007819, -0.00005213, -0.00001901, 0.00001165, 0.00003277, 0.00004141, 0.00003866, 0.00002815, 0.00001410, },
		{ 0.00001459, 0.00002831, 0.00003411, 0.00003266, 0.00002518, 0.00001348, 0.00000021, -0.00001112, -0.00001681, -0.00001430, -0.00000368, 0.00001140, 0.00002432, 0.00002764, 0.00001620, -0.00001006, -0.00004492, -0.00007728, -0.00009487, -0.00008906, -0.00005889, -0.00001235, 0.00003572, 0.00006883, 0.00007503, 0.00005191, 0.00000803, -0.00003988, -0.00007300, -0.00007773, -0.00005103, -0.00000191, 0.00005182, 0.00009029, 0.00009912, 0.00007494, 0.00002660, -0.00002833, -0.00007036, -0.00008547, -0.00007014, -0.00003226, 0.00001253, 0.00004722, 0.00006002, 0.00004855, 0.00002006, -0.00001216, -0.00003454, -0.00003871, -0.00002449, 0.00000048, 0.00002436, 0.00003606, 0.00002954, 0.00000588, -0.00002763, -0.00006057, -0.00008335, -0.00009035, -0.00008119, -0.00005992, -0.00003295, -0.00000659, },
		{ -0.00007779, -0.00007376, -0.00006096, -0.00004285, -0.00002271, -0.00000334, 0.00001283, 0.00002357, 0.00002707, 0.00002270, 0.00001184, -0.00000174, -0.00001255, -0.00001515, -0.00000650, 0.00001210, 0.00003461, 0.00005180, 0.00005453, 0.00003762, 0.00000264, -0.00004166, -0.00008185, -0.00010442, -0.00010089, -0.00007146, -0.00002538, 0.00002204, 0.00005498, 0.00006299, 0.00004495, 0.00000946, -0.00002853, -0.00005356, -0.00005567, -0.00003422, 0.00000213, 0.00003896, 0.00006200, 0.00006273, 0.00004160, 0.00000753, -0.00002592, -0.00004614, -0.00004644, -0.00002841, -0.00000084, 0.00002420, 0.00003647, 0.00003165, 0.00001277, -0.00001137, -0.00003013, -0.00003547, -0.00002514, -0.00000331, 0.00002127, 0.00003890, 0.00004260, 0.00003046, 0.00000588, -0.00002412, -0.00005164, -0.00007052, },
		{ 0.00002284, 0.00000439, -0.00001637, -0.00003502, -0.00004870, -0.00005621, -0.00005754, -0.00005313, -0.00004356, -0.00002976, -0.00001352, 0.00000219, 0.00001362, 0.00001742, 0.00001226, -0.00000001, -0.00001428, -0.00002356, -0.00002165, -0.00000594, 0.00002081, 0.00005061, 0.00007269, 0.00007736, 0.00005994, 0.00002306, -0.00002361, -0.00006647, -0.00009267, -0.00009484, -0.00007368, -0.00003770, 0.00000005, 0.00002696, 0.00003551, 0.00002577, 0.00000489, -0.00001619, -0.00002751, -0.00002408, -0.00000776, 0.00001390, 0.00003106, 0.00003600, 0.00002635, 0.00000607, -0.00001639, -0.00003172, -0.00003371, -0.00002180, -0.00000122, 0.00001937, 0.00003150, 0.00003040, 0.00001681, -0.00000347, -0.00002215, -0.00003182, -0.00002884, -0.00001453, 0.00000557, 0.00002411, 0.00003472, 0.00003413, },
		{ -0.00000249, 0.00000845, 0.00001630, 0.00001862, 0.00001521, 0.00000768, -0.00000172, -0.00001114, -0.00001973, -0.00002733, -0.00003381, -0.00003842, -0.00003967, -0.00003600, -0.00002689, -0.00001380, -0.00000025, 0.00000908, 0.00001028, 0.00000213, -0.00001275, -0.00002827, -0.00003685, -0.00003238, -0.00001298, 0.00001765, 0.00005098, 0.00007635, 0.00008482, 0.00007249, 0.00004213, 0.00000234, -0.00003550, -0.00006118, -0.00006924, -0.00006046, -0.00004108, -0.00001985, -0.00000444, 0.00000147, -0.00000086, -0.00000639, -0.00000912, -0.00000510, 0.00000562, 0.00001896, 0.00002875, 0.00002956, 0.00001939, 0.00000077, -0.00002005, -0.00003559, -0.00004015, -0.00003212, -0.00001463, 0.00000581, 0.00002190, 0.00002843, 0.00002415, 0.00001196, -0.00000252, -0.00001342, -0.00001688, -0.00001238, },
		{ 0.00001139, 0.00000806, 0.00000680, 0.00000802, 0.00001037, 0.00001171, 0.00001036, 0.00000603, -0.00000005, -0.00000582, -0.00000954, -0.00001071, -0.00001035, -0.00001039, -0.00001251, -0.00001702, -0.00002245, -0.00002604, -0.00002509, -0.00001851, -0.00000774, 0.00000339, 0.00001008, 0.00000858, -0.00000186, -0.00001819, -0.00003431, -0.00004316, -0.00003947, -0.00002201, 0.00000561, 0.00003600, 0.00006045, 0.00007204, 0.00006792, 0.00005017, 0.00002477, -0.00000087, -0.00002057, -0.00003139, -0.00003408, -0.00003196, -0.00002888, -0.00002721, -0.00002678, -0.00002524, -0.00001964, -0.00000831, 0.00000770, 0.00002454, 0.00003681, 0.00003972, 0.00003118, 0.00001290, -0.00001001, -0.00003080, -0.00004349, -0.00004501, -0.00003609, -0.00002072, -0.00000438, 0.00000812, 0.00001429, 0.00001453, },
		{ -0.00003253, -0.00002326, -0.00001423, -0.00000588, 0.00000238, 0.00001119, 0.00002034, 0.00002839, 0.00003303, 0.00003216, 0.00002496, 0.00001267, -0.00000167, -0.00001417, -0.00002160, -0.00002263, -0.00001840, -0.00001187, -0.00000660, -0.00000508, -0.00000764, -0.00001239, -0.00001611, -0.00001587, -0.00001056, -0.00000157, 0.00000754, 0.00001255, 0.00001043, 0.00000079, -0.00001366, -0.00002797, -0.00003673, -0.00003615, -0.00002546, -0.00000721, 0.00001373, 0.00003196, 0.00004345, 0.00004667, 0.00004269, 0.00003422, 0.00002418, 0.00001445, 0.00000536, -0.00000386, -0.00001400, -0.00002477, -0.00003434, -0.00003970, -0.00003790, -0.00002755, -0.00000986, 0.00001126, 0.00003023, 0.00004164, 0.00004211, 0.00003146, 0.00001279, -0.00000868, -0.00002745, -0.00003951, -0.00004340, -0.00004019, },
		{ -0.00000132, -0.00001639, -0.00002851, -0.00003696, -0.00004187, -0.00004346, -0.00004149, -0.00003526, -0.00002406, -0.00000812, 0.00001086, 0.00002957, 0.00004384, 0.00004993, 0.00004596, 0.00003274, 0.00001373, -0.00000597, -0.00002129, -0.00002891, -0.00002826, -0.00002149, -0.00001247, -0.00000513, -0.00000196, -0.00000318, -0.00000688, -0.00001013, -0.00001037, -0.00000663, 0.00000005, 0.00000709, 0.00001149, 0.00001119, 0.00000589, -0.00000278, -0.00001199, -0.00001879, -0.00002124, -0.00001892, -0.00001289, -0.00000492, 0.00000327, 0.00001073, 0.00001728, 0.00002313, 0.00002823, 0.00003179, 0.00003227, 0.00002807, 0.00001836, 0.00000394, -0.00001250, -0.00002698, -0.00003536, -0.00003487, -0.00002514, -0.00000856, 0.00001042, 0.00002667, 0.00003600, 0.00003641, 0.00002844, 0.00001469, },
		{ 0.00002067, 0.00002680, 0.00002779, 0.00002404, 0.00001664, 0.00000675, -0.00000478, -0.00001733, -0.00003014, -0.00004181, -0.00005021, -0.00005279, -0.00004742, -0.00003344, -0.00001233, 0.00001214, 0.00003479, 0.00005036, 0.00005519, 0.00004840, 0.00003223, 0.00001133, -0.00000868, -0.00002302, -0.00002911, -0.00002716, -0.00001973, -0.00001054, -0.00000306, 0.00000065, 0.00000043, -0.00000231, -0.00000541, -0.00000699, -0.00000616, -0.00000316, 0.00000090, 0.00000468, 0.00000718, 0.00000801, 0.00000733, 0.00000554, 0.00000296, -0.00000031, -0.00000421, -0.00000846, -0.00001225, -0.00001429, -0.00001316, -0.00000797, 0.00000106, 0.00001227, 0.00002288, 0.00002976, 0.00003053, 0.00002447, 0.00001297, -0.00000083, -0.00001296, -0.00001998, -0.00002001, -0.00001334, -0.00000216, 0.00001026, },
		{ -0.00000409, -0.00000296, 0.00000101, 0.00000658, 0.00001236, 0.00001731, 0.00002084, 0.00002266, 0.00002243, 0.00001956, 0.00001327, 0.00000304, -0.00001084, -0.00002680, -0.00004192, -0.00005253, -0.00005521, -0.00004801, -0.00003133, -0.00000816, 0.00001655, 0.00003722, 0.00004921, 0.00005019, 0.00004074, 0.00002409, 0.00000500, -0.00001171, -0.00002250, -0.00002597, -0.00002290, -0.00001564, -0.00000712, 0.00000011, 0.00000451, 0.00000569, 0.00000422, 0.00000114, -0.00000241, -0.00000549, -0.00000735, -0.00000745, -0.00000547, -0.00000139, 0.00000433, 0.00001062, 0.00001591, 0.00001851, 0.00001713, 0.00001148, 0.00000254, -0.00000754, -0.00001600, -0.00002039, -0.00001938, -0.00001327, -0.00000394, 0.00000578, 0.00001307, 0.00001608, 0.00001448, 0.00000947, 0.00000325, -0.00000183, },
		{ -0.00000454, -0.00000352, -0.00000265, -0.00000133, 0.00000073, 0.00000337, 0.00000620, 0.00000887, 0.00001136, 0.00001398, 0.00001713, 0.00002086, 0.00002451, 0.00002659, 0.00002519, 0.00001870, 0.00000662, -0.00000983, -0.00002767, -0.00004280, -0.00005113, -0.00004991, -0.00003870, -0.00001969, 0.00000282, 0.00002362, 0.00003805, 0.00004331, 0.00003908, 0.00002741, 0.00001190, -0.00000345, -0.00001536, -0.00002194, -0.00002291, -0.00001922, -0.00001257, -0.00000474, 0.00000270, 0.00000858, 0.00001209, 0.00001275, 0.00001048, 0.00000565, -0.00000076, -0.00000722, -0.00001189, -0.00001320, -0.00001030, -0.00000354, 0.00000554, 0.00001453, 0.00002091, 0.00002284, 0.00001979, 0.00001265, 0.00000344, -0.00000540, -0.00001184, -0.00001487, -0.00001463, -0.00001222, -0.00000906, -0.00000633, },
		{ -0.00000066, -0.00000560, -0.00001053, -0.00001495, -0.00001809, -0.00001918, -0.00001782, -0.00001424, -0.00000934, -0.00000435, -0.00000031, 0.00000239, 0.00000428, 0.00000657, 0.00001052, 0.00001672, 0.00002446, 0.00003168, 0.00003554, 0.00003336, 0.00002384, 0.00000778, -0.00001175, -0.00003018, -0.00004274, -0.00004598, -0.00003888, -0.00002321, -0.00000304, 0.00001655, 0.00003095, 0.00003731, 0.00003514, 0.00002610, 0.00001326, -0.00000004, -0.00001111, -0.00001843, -0.00002165, -0.00002122, -0.00001789, -0.00001240, -0.00000537, 0.00000251, 0.00001032, 0.00001681, 0.00002061, 0.00002067, 0.00001667, 0.00000937, 0.00000053, -0.00000759, -0.00001285, -0.00001395, -0.00001080, -0.00000452, 0.00000301, 0.00000981, 0.00001440, 0.00001616, 0.00001528, 0.00001248, 0.00000856, 0.00000410, },
		{ 0.00000265, 0.00000887, 0.00001407, 0.00001700, 0.00001660, 0.00001248, 0.00000520, -0.00000372, -0.00001218, -0.00001819, -0.00002061, -0.00001954, -0.00001632, -0.00001287, -0.00001087, -0.00001085, -0.00001191, -0.00001194, -0.00000861, -0.00000048, 0.00001203, 0.00002638, 0.00003854, 0.00004427, 0.00004074, 0.00002773, 0.00000802, -0.00001334, -0.00003053, -0.00003883, -0.00003612, -0.00002353, -0.00000498, 0.00001418, 0.00002889, 0.00003585, 0.00003433, 0.00002597, 0.00001388, 0.00000143, -0.00000887, -0.00001583, -0.00001955, -0.00002077, -0.00002025, -0.00001829, -0.00001467, -0.00000902, -0.00000128, 0.00000788, 0.00001704, 0.00002438, 0.00002824, 0.00002772, 0.00002297, 0.00001518, 0.00000611, -0.00000241, -0.00000896, -0.00001280, -0.00001381, -0.00001227, -0.00000865, -0.00000347, },
		{ -0.00000816, -0.00001790, -0.00002410, -0.00002545, -0.00002152, -0.00001308, -0.00000214, 0.00000848, 0.00001595, 0.00001835, 0.00001534, 0.00000834, 0.00000000, -0.00000680, -0.00001014, -0.00000980, -0.00000735, -0.00000547, -0.00000667, -0.00001199, -0.00002030, -0.00002845, -0.00003239, -0.00002881, -0.00001667, 0.00000203, 0.00002260, 0.00003913, 0.00004647, 0.00004217, 0.00002742, 0.00000679, -0.00001328, -0.00002661, -0.00002932, -0.00002105, -0.00000490, 0.00001382, 0.00002941, 0.00003771, 0.00003718, 0.00002907, 0.00001660, 0.00000351, -0.00000720, -0.00001412, -0.00001747, -0.00001853, -0.00001870, -0.00001878, -0.00001862, -0.00001733, -0.00001381, -0.00000745, 0.00000154, 0.00001203, 0.00002232, 0.00003060, 0.00003545, 0.00003612, 0.00003253, 0.00002519, 0.00001505, 0.00000342, },
		{ 0.00004534, 0.00004725, 0.00004306, 0.00003278, 0.00001765, 0.00000018, -0.00001629, -0.00002831, -0.00003340, -0.00003081, -0.00002196, -0.00001001, 0.00000108, 0.00000796, 0.00000907, 0.00000522, -0.00000084, -0.00000553, -0.00000609, -0.00000179, 0.00000554, 0.00001216, 0.00001394, 0.00000819, -0.00000498, -0.00002229, -0.00003824, -0.00004707, -0.00004488, -0.00003131, -0.00000983, 0.00001329, 0.00003113, 0.00003854, 0.00003397, 0.00001993, 0.00000207, -0.00001288, -0.00001953, -0.00001559, -0.00000255, 0.00001511, 0.00003160, 0.00004187, 0.00004327, 0.00003607, 0.00002309, 0.00000834, -0.00000444, -0.00001299, -0.00001694, -0.00001751, -0.00001663, -0.00001601, -0.00001645, -0.00001764, -0.00001845, -0.00001746, -0.00001358, -0.00000639, 0.00000372, 0.00001565, 0.00002781, 0.00003833, },
		{ -0.00001303, -0.00000440, 0.00000793, 0.00002264, 0.00003743, 0.00004938, 0.00005554, 0.00005380, 0.00004368, 0.00002678, 0.00000663, -0.00001221, -0.00002543, -0.00003045, -0.00002724, -0.00001845, -0.00000838, -0.00000142, -0.00000034, -0.00000511, -0.00001296, -0.00001944, -0.00002031, -0.00001344, 0.00000006, 0.00001598, 0.00002846, 0.00003217, 0.00002452, 0.00000681, -0.00001595, -0.00003673, -0.00004896, -0.00004888, -0.00003690, -0.00001739, 0.00000292, 0.00001747, 0.00002219, 0.00001687, 0.00000507, -0.00000736, -0.00001458, -0.00001289, -0.00000201, 0.00001498, 0.00003287, 0.00004622, 0.00005125, 0.00004691, 0.00003493, 0.00001898, 0.00000322, -0.00000904, -0.00001622, -0.00001855, -0.00001763, -0.00001554, -0.00001407, -0.00001412, -0.00001562, -0.00001764, -0.00001880, -0.00001765, },
		},
		{ { 0.00001096, 0.00001138, 0.00001249, 0.00001393, 0.00001514, 0.00001551, 0.00001446, 0.00001158, 0.00000669, -0.00000006, -0.00000822, -0.00001706, -0.00002568, -0.00003310, -0.00003844, -0.00004100, -0.00004041, -0.00003667, -0.00003016, -0.00002160, -0.00001197, -0.00000235, 0.00000618, 0.00001275, 0.00001676, 0.00001798, 0.00001659, 0.00001312, 0.00000836, 0.00000327, -0.00000120, -0.00000424, -0.00000532, -0.00000424, -0.00000120, 0.00000327, 0.00000836, 0.00001312, 0.00001659, 0.00001798, 0.00001676, 0.00001275, 0.00000618, -0.00000235, -0.00001197, -0.00002160, -0.00003016, -0.00003667, -0.00004041, -0.00004100, -0.00003844, -0.00003310, -0.00002568, -0.00001706, -0.00000822, -0.00000006, 0.00000669, 0.00001158, 0.00001446, 0.00001551, 0.00001514, 0.00001393, 0.00001249, 0.00001138, },
		{ 0.00000793, 0.00001034, 0.00001108, 0.00001057, 0.00000937, 0.00000809, 0.00000730, 0.00000737, 0.00000843, 0.00001032, 0.00001262, 0.00001470, 0.00001587, 0.00001549, 0.00001310, 0.00000858, 0.00000214, -0.00000566, -0.00001396, -0.00002172, -0.00002793, -0.00003172, -0.00003255, -0.00003028, -0.00002523, -0.00001812, -0.00000998, -0.00000196, 0.00000480, 0.00000938, 0.00001124, 0.00001026, 0.00000678, 0.00000153, -0.00000449, -0.00001017, -0.00001445, -0.00001654, -0.00001596, -0.00001265, -0.00000700, 0.00000029, 0.00000823, 0.00001575, 0.00002180, 0.00002552, 0.00002634, 0.00002403, 0.00001875, 0.00001101, 0.00000158, -0.00000856, -0.00001838, -0.00002690, -0.00003331, -0.00003703, -0.00003781, -0.00003571, -0.00003111, -0.00002464, -0.00001709, -0.00000933, -0.00000218, 0.00000371, },
		{ -0.00002534, -0.00002135, -0.00001568, -0.00000904, -0.00000225, 0.00000388, 0.00000866, 0.00001164, 0.00001266, 0.00001191, 0.00000983, 0.00000710, 0.00000443, 0.00000248, 0.00000170, 0.00000221, 0.00000382, 0.00000602, 0.00000808, 0.00000924, 0.00000882, 0.00000643, 0.00000205, -0.00000393, -0.00001072, -0.00001729, -0.00002252, -0.00002540, -0.00002527, -0.00002187, -0.00001552, -0.00000699, 0.00000253, 0.00001168, 0.00001909, 0.00002366, 0.00002467, 0.00002195, 0.00001588, 0.00000731, -0.00000255, -0.00001234, -0.00002075, -0.00002672, -0.00002953, -0.00002895, -0.00002519, -0.00001885, -0.00001083, -0.00000218, 0.00000603, 0.00001286, 0.00001758, 0.00001972, 0.00001914, 0.00001598, 0.00001066, 0.00000379, -0.00000385, -0.00001146, -0.00001821, -0.00002339, -0.00002647, -0.00002713, },
		{ -0.00000191, -0.00000682, -0.00001207, -0.00001695, -0.00002073, -0.00002273, -0.00002250, -0.00001983, -0.00001488, -0.00000812, -0.00000032, 0.00000757, 0.00001457, 0.00001983, 0.00002279, 0.00002324, 0.00002139, 0.00001782, 0.00001333, 0.00000881, 0.00000504, 0.00000254, 0.00000147, 0.00000157, 0.00000227, 0.00000279, 0.00000236, 0.00000036, -0.00000346, -0.00000891, -0.00001536, -0.00002183, -0.00002712, -0.00003011, -0.00002990, -0.00002604, -0.00001866, -0.00000843, 0.00000350, 0.00001566, 0.00002651, 0.00003466, 0.00003912, 0.00003939, 0.00003554, 0.00002819, 0.00001835, 0.00000729, -0.00000367, -0.00001337, -0.00002088, -0.00002568, -0.00002760, -0.00002681, -0.00002378, -0.00001914, -0.00001360, -0.00000789, -0.00000266, 0.00000153, 0.00000427, 0.00000530, 0.00000452, 0.00000202, },
		{ -0.00001188, -0.00000878, -0.00000562, -0.00000293, -0.00000125, -0.00000103, -0.00000257, -0.00000588, -0.00001069, -0.00001638, -0.00002209, -0.00002682, -0.00002961, -0.00002970, -0.00002668, -0.00002061, -0.00001201, -0.00000184, 0.00000871, 0.00001837, 0.00002603, 0.00003093, 0.00003276, 0.00003171, 0.00002840, 0.00002373, 0.00001869, 0.00001411, 0.00001057, 0.00000822, 0.00000680, 0.00000577, 0.00000438, 0.00000196, -0.00000195, -0.00000742, -0.00001410, -0.00002121, -0.00002773, -0.00003252, -0.00003459, -0.00003328, -0.00002838, -0.00002023, -0.00000964, 0.00000221, 0.00001399, 0.00002438, 0.00003232, 0.00003711, 0.00003847, 0.00003658, 0.00003197, 0.00002540, 0.00001772, 0.00000976, 0.00000223, -0.00000436, -0.00000967, -0.00001350, -0.00001582, -0.00001666, -0.00001614, -0.00001446, },
		{ 0.00000586, 0.00000118, -0.00000337, -0.00000734, -0.00001024, -0.00001161, -0.00001118, -0.00000892, -0.00000513, -0.00000046, 0.00000420, 0.00000783, 0.00000949, 0.00000847, 0.00000451, -0.00000214, -0.00001071, -0.00002002, -0.00002862, -0.00003511, -0.00003832, -0.00003755, -0.00003271, -0.00002431, -0.00001341, -0.00000139, 0.00001025, 0.00002019, 0.00002747, 0.00003163, 0.00003275, 0.00003135, 0.00002825, 0.00002438, 0.00002052, 0.00001721, 0.00001463, 0.00001262, 0.00001074, 0.00000848, 0.00000537, 0.00000117, -0.00000407, -0.00000997, -0.00001589, -0.00002103, -0.00002462, -0.00002606, -0.00002505, -0.00002162, -0.00001615, -0.00000929, -0.00000182, 0.00000546, 0.00001189, 0.00001699, 0.00002053, 0.00002248, 0.00002297, 0.00002220, 0.00002040, 0.00001773, 0.00001433, 0.00001032, },
		{ 0.00000626, 0.00000850, 0.00001071, 0.00001244, 0.00001314, 0.00001230, 0.00000964, 0.00000517, -0.00000071, -0.00000722, -0.00001334, -0.00001793, -0.00001998, -0.00001884, -0.00001434, -0.00000695, 0.00000233, 0.00001207, 0.00002061, 0.00002645, 0.00002841, 0.00002594, 0.00001920, 0.00000906, -0.00000305, -0.00001535, -0.00002604, -0.00003361, -0.00003705, -0.00003600, -0.00003082, -0.00002242, -0.00001213, -0.00000140, 0.00000842, 0.00001630, 0.00002165, 0.00002436, 0.00002473, 0.00002333, 0.00002087, 0.00001797, 0.00001511, 0.00001255, 0.00001031, 0.00000828, 0.00000625, 0.00000408, 0.00000171, -0.00000077, -0.00000316, -0.00000520, -0.00000664, -0.00000732, -0.00000721, -0.00000640, -0.00000512, -0.00000363, -0.00000216, -0.00000083, 0.00000033, 0.00000145, 0.00000272, 0.00000430, },
		{ 0.00000726, 0.00000398, 0.00000028, -0.00000305, -0.00000521, -0.00000559, -0.00000386, -0.00000015, 0.00000495, 0.00001048, 0.00001520, 0.00001791, 0.00001761, 0.00001378, 0.00000652, -0.00000340, -0.00001463, -0.00002546, -0.00003407, -0.00003889, -0.00003887, -0.00003372, -0.00002393, -0.00001080, 0.00000386, 0.00001794, 0.00002946, 0.00003682, 0.00003911, 0.00003623, 0.00002886, 0.00001833, 0.00000637, -0.00000522, -0.00001479, -0.00002119, -0.00002381, -0.00002270, -0.00001843, -0.00001198, -0.00000450, 0.00000287, 0.00000918, 0.00001379, 0.00001640, 0.00001704, 0.00001595, 0.00001357, 0.00001037, 0.00000680, 0.00000329, 0.00000017, -0.00000231, -0.00000393, -0.00000456, -0.00000414, -0.00000271, -0.00000042, 0.00000242, 0.00000541, 0.00000805, 0.00000985, 0.00001041, 0.00000953, },
		{ 0.00000064, 0.00000698, 0.00001263, 0.00001662, 0.00001817, 0.00001685, 0.00001273, 0.00000640, -0.00000112, -0.00000846, -0.00001420, -0.00001709, -0.00001629, -0.00001160, -0.00000352, 0.00000676, 0.00001758, 0.00002699, 0.00003318, 0.00003472, 0.00003090, 0.00002184, 0.00000853, -0.00000730, -0.00002349, -0.00003774, -0.00004803, -0.00005290, -0.00005173, -0.00004477, -0.00003314, -0.00001854, -0.00000306, 0.00001125, 0.00002264, 0.00002994, 0.00003265, 0.00003103, 0.00002591, 0.00001855, 0.00001039, 0.00000281, -0.00000309, -0.00000661, -0.00000750, -0.00000596, -0.00000250, 0.00000211, 0.00000704, 0.00001148, 0.00001478, 0.00001647, 0.00001632, 0.00001435, 0.00001078, 0.00000603, 0.00000066, -0.00000467, -0.00000927, -0.00001251, -0.00001388, -0.00001310, -0.00001015, -0.00000536, },
		{ -0.00000647, -0.00001180, -0.00001506, -0.00001562, -0.00001326, -0.00000817, -0.00000105, 0.00000701, 0.00001467, 0.00002052, 0.00002335, 0.00002236, 0.00001732, 0.00000868, -0.00000245, -0.00001445, -0.00002544, -0.00003353, -0.00003718, -0.00003548, -0.00002830, -0.00001637, -0.00000121, 0.00001513, 0.00003032, 0.00004215, 0.00004884, 0.00004938, 0.00004361, 0.00003229, 0.00001694, -0.00000039, -0.00001744, -0.00003208, -0.00004263, -0.00004806, -0.00004812, -0.00004329, -0.00003466, -0.00002375, -0.00001220, -0.00000153, 0.00000704, 0.00001277, 0.00001542, 0.00001524, 0.00001289, 0.00000923, 0.00000525, 0.00000183, -0.00000033, -0.00000081, 0.00000049, 0.00000336, 0.00000733, 0.00001175, 0.00001589, 0.00001902, 0.00002054, 0.00002005, 0.00001742, 0.00001284, 0.00000680, 0.00000005, },
		{ 0.00001556, 0.00001793, 0.00001818, 0.00001613, 0.00001201, 0.00000641, 0.00000023, -0.00000549, -0.00000969, -0.00001148, -0.00001033, -0.00000616, 0.00000058, 0.00000897, 0.00001769, 0.00002527, 0.00003023, 0.00003140, 0.00002810, 0.00002026, 0.00000853, -0.00000582, -0.00002105, -0.00003516, -0.00004622, -0.00005261, -0.00005330, -0.00004798, -0.00003716, -0.00002207, -0.00000452, 0.00001338, 0.00002950, 0.00004198, 0.00004947, 0.00005130, 0.00004757, 0.00003904, 0.00002704, 0.00001324, -0.00000061, -0.00001289, -0.00002234, -0.00002819, -0.00003024, -0.00002881, -0.00002462, -0.00001871, -0.00001221, -0.00000617, -0.00000146, 0.00000143, 0.00000232, 0.00000144, -0.00000071, -0.00000343, -0.00000594, -0.00000755, -0.00000771, -0.00000617, -0.00000300, 0.00000145, 0.00000655, 0.00001153, },
		{ -0.00001366, -0.00001589, -0.00001623, -0.00001465, -0.00001147, -0.00000727, -0.00000279, 0.00000120, 0.00000403, 0.00000527, 0.00000481, 0.00000283, -0.00000014, -0.00000339, -0.00000611, -0.00000753, -0.00000706, -0.00000441, 0.00000037, 0.00000683, 0.00001423, 0.00002156, 0.00002773, 0.00003167, 0.00003253, 0.00002979, 0.00002335, 0.00001356, 0.00000124, -0.00001245, -0.00002609, -0.00003819, -0.00004739, -0.00005256, -0.00005306, -0.00004872, -0.00003994, -0.00002763, -0.00001311, 0.00000207, 0.00001628, 0.00002807, 0.00003632, 0.00004037, 0.00004013, 0.00003602, 0.00002894, 0.00002011, 0.00001087, 0.00000249, -0.00000399, -0.00000794, -0.00000917, -0.00000795, -0.00000493, -0.00000102, 0.00000279, 0.00000563, 0.00000683, 0.00000610, 0.00000351, -0.00000050, -0.00000525, -0.00000990, },
		{ 0.00002528, 0.00002706, 0.00002629, 0.00002304, 0.00001777, 0.00001119, 0.00000414, -0.00000259, -0.00000835, -0.00001273, -0.00001561, -0.00001707, -0.00001737, -0.00001686, -0.00001586, -0.00001463, -0.00001333, -0.00001200, -0.00001061, -0.00000910, -0.00000740, -0.00000548, -0.00000336, -0.00000107, 0.00000131, 0.00000372, 0.00000611, 0.00000845, 0.00001068, 0.00001272, 0.00001439, 0.00001546, 0.00001560, 0.00001447, 0.00001174, 0.00000724, 0.00000099, -0.00000670, -0.00001523, -0.00002375, -0.00003124, -0.00003668, -0.00003921, -0.00003827, -0.00003378, -0.00002614, -0.00001621, -0.00000522, 0.00000545, 0.00001448, 0.00002080, 0.00002381, 0.00002347, 0.00002026, 0.00001513, 0.00000930, 0.00000405, 0.00000046, -0.00000075, 0.00000066, 0.00000442, 0.00000979, 0.00001578, 0.00002128, },
		{ -0.00001713, -0.00001585, -0.00001124, -0.00000383, 0.00000544, 0.00001537, 0.00002479, 0.00003269, 0.00003837, 0.00004151, 0.00004211, 0.00004042, 0.00003682, 0.00003173, 0.00002550, 0.00001840, 0.00001067, 0.00000257, -0.00000557, -0.00001327, -0.00001994, -0.00002493, -0.00002762, -0.00002759, -0.00002473, -0.00001935, -0.00001214, -0.00000415, 0.00000339, 0.00000931, 0.00001266, 0.00001297, 0.00001029, 0.00000526, -0.00000110, -0.00000749, -0.00001267, -0.00001568, -0.00001605, -0.00001390, -0.00000989, -0.00000508, -0.00000071, 0.00000209, 0.00000257, 0.00000049, -0.00000376, -0.00000930, -0.00001489, -0.00001927, -0.00002136, -0.00002056, -0.00001686, -0.00001087, -0.00000369, 0.00000331, 0.00000879, 0.00001169, 0.00001149, 0.00000827, 0.00000273, -0.00000394, -0.00001034, -0.00001510, },
		{ 0.00002059, 0.00001536, 0.00000693, -0.00000346, -0.00001426, -0.00002385, -0.00003088, -0.00003449, -0.00003437, -0.00003078, -0.00002438, -0.00001605, -0.00000669, 0.00000299, 0.00001248, 0.00002153, 0.00003001, 0.00003776, 0.00004452, 0.00004980, 0.00005293, 0.00005314, 0.00004981, 0.00004264, 0.00003184, 0.00001827, 0.00000338, -0.00001094, -0.00002269, -0.00003010, -0.00003198, -0.00002800, -0.00001884, -0.00000609, 0.00000799, 0.00002086, 0.00003017, 0.00003420, 0.00003216, 0.00002437, 0.00001216, -0.00000239, -0.00001687, -0.00002899, -0.00003703, -0.00004014, -0.00003843, -0.00003292, -0.00002521, -0.00001715, -0.00001039, -0.00000608, -0.00000457, -0.00000549, -0.00000780, -0.00001016, -0.00001122, -0.00001003, -0.00000624, -0.00000019, 0.00000710, 0.00001420, 0.00001959, 0.00002198, },
		{ -0.00001435, -0.00000516, 0.00000572, 0.00001658, 0.00002560, 0.00003122, 0.00003249, 0.00002925, 0.00002210, 0.00001228, 0.00000134, -0.00000918, -0.00001811, -0.00002477, -0.00002910, -0.00003148, -0.00003254, -0.00003285, -0.00003270, -0.00003195, -0.00003001, -0.00002605, -0.00001925, -0.00000915, 0.00000412, 0.00001962, 0.00003570, 0.00005018, 0.00006077, 0.00006553, 0.00006333, 0.00005416, 0.00003926, 0.00002094, 0.00000224, -0.00001364, -0.00002396, -0.00002696, -0.00002222, -0.00001077, 0.00000510, 0.00002228, 0.00003748, 0.00004780, 0.00005132, 0.00004745, 0.00003694, 0.00002173, 0.00000448, -0.00001201, -0.00002532, -0.00003391, -0.00003732, -0.00003617, -0.00003191, -0.00002641, -0.00002148, -0.00001843, -0.00001777, -0.00001919, -0.00002161, -0.00002357, -0.00002360, -0.00002065, },
		{ -0.00000893, -0.00001538, -0.00002262, -0.00002876, -0.00003202, -0.00003120, -0.00002598, -0.00001704, -0.00000589, 0.00000543, 0.00001489, 0.00002089, 0.00002263, 0.00002029, 0.00001492, 0.00000815, 0.00000175, -0.00000285, -0.00000494, -0.00000469, -0.00000314, -0.00000190, -0.00000268, -0.00000680, -0.00001473, -0.00002587, -0.00003856, -0.00005033, -0.00005847, -0.00006060, -0.00005525, -0.00004233, -0.00002317, -0.00000044, 0.00002239, 0.00004172, 0.00005450, 0.00005890, 0.00005466, 0.00004317, 0.00002719, 0.00001025, -0.00000407, -0.00001285, -0.00001443, -0.00000874, 0.00000278, 0.00001752, 0.00003227, 0.00004394, 0.00005023, 0.00005002, 0.00004360, 0.00003250, 0.00001907, 0.00000593, -0.00000466, -0.00001126, -0.00001352, -0.00001220, -0.00000886, -0.00000544, -0.00000370, -0.00000479, },
		{ -0.00000234, 0.00000443, 0.00001197, 0.00001793, 0.00002040, 0.00001839, 0.00001212, 0.00000294, -0.00000693, -0.00001502, -0.00001923, -0.00001835, -0.00001242, -0.00000281, 0.00000817, 0.00001770, 0.00002323, 0.00002305, 0.00001682, 0.00000565, -0.00000808, -0.00002123, -0.00003063, -0.00003386, -0.00002983, -0.00001914, -0.00000400, 0.00001218, 0.00002554, 0.00003257, 0.00003097, 0.00002015, 0.00000145, -0.00002208, -0.00004634, -0.00006691, -0.00008006, -0.00008344, -0.00007657, -0.00006091, -0.00003949, -0.00001625, 0.00000481, 0.00002043, 0.00002877, 0.00002967, 0.00002460, 0.00001616, 0.00000748, 0.00000139, -0.00000019, 0.00000331, 0.00001107, 0.00002114, 0.00003093, 0.00003797, 0.00004042, 0.00003762, 0.00003016, 0.00001976, 0.00000879, -0.00000029, -0.00000557, -0.00000615, },
		{ 0.00001859, 0.00000475, -0.00000771, -0.00001610, -0.00001867, -0.00001509, -0.00000651, 0.00000465, 0.00001530, 0.00002247, 0.00002402, 0.00001928, 0.00000925, -0.00000352, -0.00001557, -0.00002328, -0.00002379, -0.00001580, 0.00000003, 0.00002107, 0.00004319, 0.00006161, 0.00007192, 0.00007113, 0.00005834, 0.00003511, 0.00000521, -0.00002611, -0.00005311, -0.00007079, -0.00007590, -0.00006762, -0.00004777, -0.00002042, 0.00000894, 0.00003449, 0.00005127, 0.00005612, 0.00004828, 0.00002943, 0.00000324, -0.00002545, -0.00005165, -0.00007127, -0.00008181, -0.00008276, -0.00007545, -0.00006262, -0.00004760, -0.00003353, -0.00002261, -0.00001572, -0.00001238, -0.00001102, -0.00000963, -0.00000635, -0.00000010, 0.00000908, 0.00001999, 0.00003057, 0.00003847, 0.00004167, 0.00003907, 0.00003088, },
		{ 0.00001173, 0.00002728, 0.00003769, 0.00004093, 0.00003637, 0.00002504, 0.00000958, -0.00000637, -0.00001894, -0.00002497, -0.00002282, -0.00001286, 0.00000249, 0.00001926, 0.00003279, 0.00003882, 0.00003461, 0.00001965, -0.00000395, -0.00003181, -0.00005805, -0.00007654, -0.00008222, -0.00007230, -0.00004698, -0.00000964, 0.00003376, 0.00007572, 0.00010866, 0.00012645, 0.00012562, 0.00010609, 0.00007119, 0.00002704, -0.00001869, -0.00005823, -0.00008520, -0.00009581, -0.00008944, -0.00006859, -0.00003829, -0.00000489, 0.00002515, 0.00004662, 0.00005638, 0.00005379, 0.00004057, 0.00002017, -0.00000313, -0.00002527, -0.00004317, -0.00005525, -0.00006151, -0.00006320, -0.00006220, -0.00006035, -0.00005882, -0.00005784, -0.00005662, -0.00005377, -0.00004777, -0.00003762, -0.00002333, -0.00000612, },
		{ -0.00007422, -0.00007757, -0.00007437, -0.00006382, -0.00004664, -0.00002520, -0.00000312, 0.00001541, 0.00002660, 0.00002805, 0.00001942, 0.00000270, -0.00001801, -0.00003735, -0.00004983, -0.00005117, -0.00003939, -0.00001545, 0.00001667, 0.00005063, 0.00007898, 0.00009473, 0.00009287, 0.00007163, 0.00003308, -0.00001705, -0.00007035, -0.00011725, -0.00014891, -0.00015895, -0.00014478, -0.00010815, -0.00005488, 0.00000621, 0.00006498, 0.00011187, 0.00013966, 0.00014475, 0.00012765, 0.00009274, 0.00004722, -0.00000039, -0.00004196, -0.00007118, -0.00008463, -0.00008214, -0.00006644, -0.00004233, -0.00001546, 0.00000888, 0.00002679, 0.00003629, 0.00003743, 0.00003184, 0.00002204, 0.00001063, -0.00000042, -0.00001014, -0.00001864, -0.00002671, -0.00003534, -0.00004514, -0.00005584, -0.00006621, },
		{ 0.00001623, 0.00001183, 0.00000248, -0.00001193, -0.00003005, -0.00004913, -0.00006546, -0.00007525, -0.00007555, -0.00006517, -0.00004524, -0.00001924, 0.00000755, 0.00002906, 0.00003984, 0.00003634, 0.00001798, -0.00001245, -0.00004908, -0.00008404, -0.00010896, -0.00011683, -0.00010362, -0.00006937, -0.00001849, 0.00004086, 0.00009816, 0.00014256, 0.00016503, 0.00016015, 0.00012737, 0.00007118, 0.00000042, -0.00007336, -0.00013804, -0.00018314, -0.00020180, -0.00019187, -0.00015623, -0.00010204, -0.00003925, 0.00002142, 0.00007040, 0.00010101, 0.00011044, 0.00009986, 0.00007385, 0.00003916, 0.00000311, -0.00002784, -0.00004924, -0.00005925, -0.00005854, -0.00004971, -0.00003639, -0.00002215, -0.00000974, -0.00000057, 0.00000528, 0.00000873, 0.00001103, 0.00001317, 0.00001535, 0.00001685, },
		{ -0.00003617, -0.00003224, -0.00002435, -0.00001243, 0.00000218, 0.00001676, 0.00002782, 0.00003192, 0.00002671, 0.00001177, -0.00001096, -0.00003730, -0.00006152, -0.00007760, -0.00008060, -0.00006804, -0.00004070, -0.00000286, 0.00003830, 0.00007409, 0.00009603, 0.00009780, 0.00007676, 0.00003487, -0.00002144, -0.00008223, -0.00013591, -0.00017150, -0.00018089, -0.00016055, -0.00011248, -0.00004397, 0.00003368, 0.00010719, 0.00016382, 0.00019377, 0.00019199, 0.00015908, 0.00010105, 0.00002808, -0.00004747, -0.00011346, -0.00016013, -0.00018177, -0.00017744, -0.00015080, -0.00010898, -0.00006094, -0.00001554, 0.00002007, 0.00004165, 0.00004825, 0.00004192, 0.00002677, 0.00000774, -0.00001062, -0.00002508, -0.00003425, -0.00003842, -0.00003913, -0.00003829, -0.00003747, -0.00003731, -0.00003732, },
		{ -0.00001447, -0.00001832, -0.00002626, -0.00003842, -0.00005331, -0.00006792, -0.00007844, -0.00008124, -0.00007398, -0.00005644, -0.00003099, -0.00000229, 0.00002355, 0.00004033, 0.00004327, 0.00003034, 0.00000304, -0.00003357, -0.00007169, -0.00010231, -0.00011718, -0.00011068, -0.00008134, -0.00003247, 0.00002817, 0.00008973, 0.00014034, 0.00016950, 0.00017031, 0.00014105, 0.00008572, 0.00001342, -0.00006328, -0.00013073, -0.00017686, -0.00019351, -0.00017796, -0.00013345, -0.00006846, 0.00000502, 0.00007386, 0.00012627, 0.00015402, 0.00015376, 0.00012742, 0.00008142, 0.00002519, -0.00003093, -0.00007772, -0.00010874, -0.00012128, -0.00011644, -0.00009840, -0.00007308, -0.00004671, -0.00002439, -0.00000916, -0.00000175, -0.00000082, -0.00000383, -0.00000799, -0.00001122, -0.00001277, -0.00001328, },
		{ -0.00003021, -0.00002682, -0.00001859, -0.00000554, 0.00001030, 0.00002517, 0.00003454, 0.00003426, 0.00002190, -0.00000236, -0.00003536, -0.00007138, -0.00010320, -0.00012372, -0.00012764, -0.00011289, -0.00008136, -0.00003886, 0.00000597, 0.00004336, 0.00006459, 0.00006395, 0.00004026, -0.00000264, -0.00005642, -0.00010982, -0.00015090, -0.00016957, -0.00015978, -0.00012093, -0.00005822, 0.00001830, 0.00009557, 0.00016010, 0.00020067, 0.00021044, 0.00018834, 0.00013915, 0.00007249, 0.00000077, -0.00006328, -0.00010909, -0.00013022, -0.00012537, -0.00009830, -0.00005674, -0.00001054, 0.00003051, 0.00005866, 0.00006959, 0.00006300, 0.00004222, 0.00001310, -0.00001755, -0.00004360, -0.00006089, -0.00006793, -0.00006583, -0.00005764, -0.00004724, -0.00003813, -0.00003241, -0.00003037, -0.00003049, },
		{ -0.00002757, -0.00002632, -0.00003282, -0.00004624, -0.00006318, -0.00007840, -0.00008608, -0.00008138, -0.00006192, -0.00002873, 0.00001358, 0.00005749, 0.00009408, 0.00011494, 0.00011417, 0.00008996, 0.00004526, -0.00001248, -0.00007269, -0.00012373, -0.00015538, -0.00016108, -0.00013941, -0.00009464, -0.00003590, 0.00002454, 0.00007387, 0.00010135, 0.00010069, 0.00007138, 0.00001887, -0.00004645, -0.00011140, -0.00016267, -0.00018953, -0.00018597, -0.00015182, -0.00009273, -0.00001888, 0.00005716, 0.00012296, 0.00016866, 0.00018870, 0.00018261, 0.00015471, 0.00011282, 0.00006629, 0.00002397, -0.00000761, -0.00002534, -0.00002976, -0.00002444, -0.00001467, -0.00000593, -0.00000240, -0.00000601, -0.00001612, -0.00002996, -0.00004364, -0.00005339, -0.00005674, -0.00005330, -0.00004486, -0.00003491, },
		{ -0.00010580, -0.00009583, -0.00007278, -0.00004075, -0.00000649, 0.00002189, 0.00003674, 0.00003269, 0.00000822, -0.00003361, -0.00008535, -0.00013651, -0.00017560, -0.00019256, -0.00018113, -0.00014043, -0.00007553, 0.00000329, 0.00008236, 0.00014733, 0.00018604, 0.00019109, 0.00016145, 0.00010277, 0.00002634, -0.00005311, -0.00012039, -0.00016292, -0.00017317, -0.00015019, -0.00009962, -0.00003249, 0.00003717, 0.00009527, 0.00013046, 0.00013641, 0.00011290, 0.00006562, 0.00000481, -0.00005702, -0.00010770, -0.00013782, -0.00014238, -0.00012154, -0.00008011, -0.00002624, 0.00003055, 0.00008137, 0.00011965, 0.00014211, 0.00014888, 0.00014283, 0.00012832, 0.00010985, 0.00009079, 0.00007276, 0.00005552, 0.00003755, 0.00001706, -0.00000699, -0.00003403, -0.00006161, -0.00008573, -0.00010178, },
		{ 0.00003664, -0.00002994, -0.00010057, -0.00016675, -0.00021874, -0.00024734, -0.00024598, -0.00021246, -0.00015013, -0.00006788, 0.00002097, 0.00010081, 0.00015667, 0.00017709, 0.00015668, 0.00009746, 0.00000884, -0.00009385, -0.00019212, -0.00026779, -0.00030647, -0.00030044, -0.00025015, -0.00016422, -0.00005779, 0.00005040, 0.00014162, 0.00020074, 0.00021902, 0.00019563, 0.00013752, 0.00005779, -0.00002708, -0.00010059, -0.00014945, -0.00016598, -0.00014925, -0.00010486, -0.00004333, 0.00002232, 0.00007939, 0.00011805, 0.00013304, 0.00012438, 0.00009683, 0.00005851, 0.00001884, -0.00001346, -0.00003209, -0.00003410, -0.00002003, 0.00000677, 0.00004128, 0.00007817, 0.00011291, 0.00014244, 0.00016524, 0.00018094, 0.00018957, 0.00019085, 0.00018368, 0.00016625, 0.00013652, 0.00009316, },
		{ 0.00036791, 0.00039731, 0.00040915, 0.00039624, 0.00035244, 0.00027444, 0.00016335, 0.00002573, -0.00012626, -0.00027602, -0.00040460, -0.00049377, -0.00052933, -0.00050396, -0.00041927, -0.00028618, -0.00012369, 0.00004391, 0.00019102, 0.00029509, 0.00034053, 0.00032156, 0.00024322, 0.00012044, -0.00002467, -0.00016686, -0.00028230, -0.00035277, -0.00036872, -0.00033067, -0.00024855, -0.00013950, -0.00002423, 0.00007697, 0.00014799, 0.00017956, 0.00017044, 0.00012704, 0.00006145, -0.00001136, -0.00007673, -0.00012311, -0.00014397, -0.00013850, -0.00011109, -0.00006975, -0.00002399, 0.00001735, 0.00004785, 0.00006456, 0.00006807, 0.00006187, 0.00005114, 0.00004137, 0.00003717, 0.00004154, 0.00005562, 0.00007903, 0.00011042, 0.00014809, 0.00019038, 0.00023575, 0.00028241, 0.00032777, },
		{ 0.00005238, 0.00010451, 0.00017231, 0.00025685, 0.00035711, 0.00046899, 0.00058447, 0.00069161, 0.00077539, 0.00081960, 0.00080943, 0.00073450, 0.00059166, 0.00038687, 0.00013579, -0.00013740, -0.00040287, -0.00062925, -0.00078851, -0.00086060, -0.00083699, -0.00072253, -0.00053500, -0.00030253, -0.00005920, 0.00016042, 0.00032680, 0.00041999, 0.00043254, 0.00037010, 0.00024976, 0.00009653, -0.00006145, -0.00019794, -0.00029291, -0.00033549, -0.00032501, -0.00027008, -0.00018607, -0.00009155, -0.00000453, 0.00006082, 0.00009630, 0.00010046, 0.00007789, 0.00003745, -0.00001014, -0.00005471, -0.00008864, -0.00010792, -0.00011238, -0.00010491, -0.00009026, -0.00007354, -0.00005897, -0.00004903, -0.00004422, -0.00004338, -0.00004419, -0.00004395, -0.00004008, -0.00003039, -0.00001298, 0.00001395, },
		{ -0.00013401, -0.00015225, -0.00017141, -0.00018988, -0.00020479, -0.00021147, -0.00020327, -0.00017199, -0.00010900, -0.00000698, 0.00013780, 0.00032346, 0.00054086, 0.00077315, 0.00099675, 0.00118380, 0.00130593, 0.00133880, 0.00126657, 0.00108556, 0.00080618, 0.00045270, 0.00006076, -0.00032732, -0.00066843, -0.00092500, -0.00107069, -0.00109421, -0.00100093, -0.00081165, -0.00055905, -0.00028220, -0.00002032, 0.00019316, 0.00033535, 0.00039650, 0.00038028, 0.00030175, 0.00018362, 0.00005156, -0.00007051, -0.00016423, -0.00021916, -0.00023343, -0.00021268, -0.00016775, -0.00011179, -0.00005732, -0.00001396, 0.00001290, 0.00002228, 0.00001687, 0.00000158, -0.00001799, -0.00003704, -0.00005249, -0.00006323, -0.00006991, -0.00007426, -0.00007838, -0.00008402, -0.00009226, -0.00010349, -0.00011755, },
		{ -0.00003250, -0.00003390, -0.00003677, -0.00004171, -0.00004974, -0.00006271, -0.00008333, -0.00011480, -0.00016003, -0.00022030, -0.00029385, -0.00037467, -0.00045188, -0.00051016, -0.00053135, -0.00049718, -0.00039269, -0.00020995, 0.00004891, 0.00036989, 0.00072719, 0.00108541, 0.00140370, 0.00164137, 0.00176406, 0.00174943, 0.00159129, 0.00130143, 0.00090866, 0.00045515, -0.00000943, -0.00043492, -0.00077779, -0.00100708, -0.00110823, -0.00108425, -0.00095406, -0.00074845, -0.00050456, -0.00025968, -0.00004575, 0.00011482, 0.00021118, 0.00024402, 0.00022354, 0.00016613, 0.00009050, 0.00001402, -0.00005001, -0.00009384, -0.00011547, -0.00011758, -0.00010587, -0.00008705, -0.00006719, -0.00005055, -0.00003915, -0.00003303, -0.00003090, -0.00003098, -0.00003171, -0.00003217, -0.00003219, -0.00003211, },
		{ -0.00005222, -0.00005205, -0.00005164, -0.00005117, -0.00005063, -0.00004946, -0.00004643, -0.00003978, -0.00002781, -0.00000974, 0.00001325, 0.00003731, 0.00005561, 0.00005892, 0.00003707, -0.00001882, -0.00011377, -0.00024641, -0.00040734, -0.00057870, -0.00073547, -0.00084844, -0.00088855, -0.00083200, -0.00066512, -0.00038818, -0.00001720, 0.00041669, 0.00087052, 0.00129491, 0.00164089, 0.00186704, 0.00194566, 0.00186704, 0.00164089, 0.00129491, 0.00087052, 0.00041669, -0.00001720, -0.00038818, -0.00066512, -0.00083200, -0.00088855, -0.00084844, -0.00073547, -0.00057870, -0.00040734, -0.00024641, -0.00011377, -0.00001882, 0.00003707, 0.00005892, 0.00005561, 0.00003731, 0.00001325, -0.00000974, -0.00002781, -0.00003978, -0.00004643, -0.00004946, -0.00005063, -0.00005117, -0.00005164, -0.00005205, },
		{ -0.00003250, -0.00003211, -0.00003219, -0.00003217, -0.00003171, -0.00003098, -0.00003090, -0.00003303, -0.00003915, -0.00005055, -0.00006719, -0.00008705, -0.00010587, -0.00011758, -0.00011547, -0.00009384, -0.00005001, 0.00001402, 0.00009050, 0.00016613, 0.00022354, 0.00024402, 0.00021118, 0.00011482, -0.00004575, -0.00025968, -0.00050456, -0.00074845, -0.00095406, -0.00108425, -0.00110823, -0.00100708, -0.00077779, -0.00043492, -0.00000943, 0.00045515, 0.00090866, 0.00130143, 0.00159129, 0.00174943, 0.00176406, 0.00164137, 0.00140370, 0.00108541, 0.00072719, 0.00036989, 0.00004891, -0.00020995, -0.00039269, -0.00049718, -0.00053135, -0.00051016, -0.00045188, -0.00037467, -0.00029385, -0.00022030, -0.00016003, -0.00011480, -0.00008333, -0.00006271, -0.00004974, -0.00004171, -0.00003677, -0.00003390, },
		{ -0.00013401, -0.00011755, -0.00010349, -0.00009226, -0.00008402, -0.00007838, -0.00007426, -0.00006991, -0.00006323, -0.00005249, -0.00003704, -0.00001799, 0.00000158, 0.00001687, 0.00002228, 0.00001290, -0.00001396, -0.00005732, -0.00011179, -0.00016775, -0.00021268, -0.00023343, -0.00021916, -0.00016423, -0.00007051, 0.00005156, 0.00018362, 0.00030175, 0.00038028, 0.00039650, 0.00033535, 0.00019316, -0.00002032, -0.00028220, -0.00055905, -0.00081165, -0.00100093, -0.00109421, -0.00107069, -0.00092500, -0.00066843, -0.00032732, 0.00006076, 0.00045270, 0.00080618, 0.00108556, 0.00126657, 0.00133880, 0.00130593, 0.00118380, 0.00099675, 0.00077315, 0.00054086, 0.00032346, 0.00013780, -0.00000698, -0.00010900, -0.00017199, -0.00020327, -0.00021147, -0.00020479, -0.00018988, -0.00017141, -0.00015225, },
		{ 0.00005238, 0.00001395, -0.00001298, -0.00003039, -0.00004008, -0.00004395, -0.00004419, -0.00004338, -0.00004422, -0.00004903, -0.00005897, -0.00007354, -0.00009026, -0.00010491, -0.00011238, -0.00010792, -0.00008864, -0.00005471, -0.00001014, 0.00003745, 0.00007789, 0.00010046, 0.00009630, 0.00006082, -0.00000453, -0.00009155, -0.00018607, -0.00027008, -0.00032501, -0.00033549, -0.00029291, -0.00019794, -0.00006145, 0.00009653, 0.00024976, 0.00037010, 0.00043254, 0.00041999, 0.00032680, 0.00016042, -0.00005920, -0.00030253, -0.00053500, -0.00072253, -0.00083699, -0.00086060, -0.00078851, -0.00062925, -0.00040287, -0.00013740, 0.00013579, 0.00038687, 0.00059166, 0.00073450, 0.00080943, 0.00081960, 0.00077539, 0.00069161, 0.00058447, 0.00046899, 0.00035711, 0.00025685, 0.00017231, 0.00010451, },
		{ 0.00036791, 0.00032777, 0.00028241, 0.00023575, 0.00019038, 0.00014809, 0.00011042, 0.00007903, 0.00005562, 0.00004154, 0.00003717, 0.00004137, 0.00005114, 0.00006187, 0.00006807, 0.00006456, 0.00004785, 0.00001735, -0.00002399, -0.00006975, -0.00011109, -0.00013850, -0.00014397, -0.00012311, -0.00007673, -0.00001136, 0.00006145, 0.00012704, 0.00017044, 0.00017956, 0.00014799, 0.00007697, -0.00002423, -0.00013950, -0.00024855, -0.00033067, -0.00036872, -0.00035277, -0.00028230, -0.00016686, -0.00002467, 0.00012044, 0.00024322, 0.00032156, 0.00034053, 0.00029509, 0.00019102, 0.00004391, -0.00012369, -0.00028618, -0.00041927, -0.00050396, -0.00052933, -0.00049377, -0.00040460, -0.00027602, -0.00012626, 0.00002573, 0.00016335, 0.00027444, 0.00035244, 0.00039624, 0.00040915, 0.00039731, },
		{ 0.00003664, 0.00009316, 0.00013652, 0.00016625, 0.00018368, 0.00019085, 0.00018957, 0.00018094, 0.00016524, 0.00014244, 0.00011291, 0.00007817, 0.00004128, 0.00000677, -0.00002003, -0.00003410, -0.00003209, -0.00001346, 0.00001884, 0.00005851, 0.00009683, 0.00012438, 0.00013304, 0.00011805, 0.00007939, 0.00002232, -0.00004333, -0.00010486, -0.00014925, -0.00016598, -0.00014945, -0.00010059, -0.00002708, 0.00005779, 0.00013752, 0.00019563, 0.00021902, 0.00020074, 0.00014162, 0.00005040, -0.00005779, -0.00016422, -0.00025015, -0.00030044, -0.00030647, -0.00026779, -0.00019212, -0.00009385, 0.00000884, 0.00009746, 0.00015668, 0.00017709, 0.00015667, 0.00010081, 0.00002097, -0.00006788, -0.00015013, -0.00021246, -0.00024598, -0.00024734, -0.00021874, -0.00016675, -0.00010057, -0.00002994, },
		{ -0.00010580, -0.00010178, -0.00008573, -0.00006161, -0.00003403, -0.00000699, 0.00001706, 0.00003755, 0.00005552, 0.00007276, 0.00009079, 0.00010985, 0.00012832, 0.00014283, 0.00014888, 0.00014211, 0.00011965, 0.00008137, 0.00003055, -0.00002624, -0.00008011, -0.00012154, -0.00014238, -0.00013782, -0.00010770, -0.00005702, 0.00000481, 0.00006562, 0.00011290, 0.00013641, 0.00013046, 0.00009527, 0.00003717, -0.00003249, -0.00009962, -0.00015019, -0.00017317, -0.00016292, -0.00012039, -0.00005311, 0.00002634, 0.00010277, 0.00016145, 0.00019109, 0.00018604, 0.00014733, 0.00008236, 0.00000329, -0.00007553, -0.00014043, -0.00018113, -0.00019256, -0.00017560, -0.00013651, -0.00008535, -0.00003361, 0.00000822, 0.00003269, 0.00003674, 0.00002189, -0.00000649, -0.00004075, -0.00007278, -0.00009583, },
		{ -0.00002757, -0.00003491, -0.00004486, -0.00005330, -0.00005674, -0.00005339, -0.00004364, -0.00002996, -0.00001612, -0.00000601, -0.00000240, -0.00000593, -0.00001467, -0.00002444, -0.00002976, -0.00002534, -0.00000761, 0.00002397, 0.00006629, 0.00011282, 0.00015471, 0.00018261, 0.00018870, 0.00016866, 0.00012296, 0.00005716, -0.00001888, -0.00009273, -0.00015182, -0.00018597, -0.00018953, -0.00016267, -0.00011140, -0.00004645, 0.00001887, 0.00007138, 0.00010069, 0.00010135, 0.00007387, 0.00002454, -0.00003590, -0.00009464, -0.00013941, -0.00016108, -0.00015538, -0.00012373, -0.00007269, -0.00001248, 0.00004526, 0.00008996, 0.00011417, 0.00011494, 0.00009408, 0.00005749, 0.00001358, -0.00002873, -0.00006192, -0.00008138, -0.00008608, -0.00007840, -0.00006318, -0.00004624, -0.00003282, -0.00002632, },
		{ -0.00003021, -0.00003049, -0.00003037, -0.00003241, -0.00003813, -0.00004724, -0.00005764, -0.00006583, -0.00006793, -0.00006089, -0.00004360, -0.00001755, 0.00001310, 0.00004222, 0.00006300, 0.00006959, 0.00005866, 0.00003051, -0.00001054, -0.00005674, -0.00009830, -0.00012537, -0.00013022, -0.00010909, -0.00006328, 0.00000077, 0.00007249, 0.00013915, 0.00018834, 0.00021044, 0.00020067, 0.00016010, 0.00009557, 0.00001830, -0.00005822, -0.00012093, -0.00015978, -0.00016957, -0.00015090, -0.00010982, -0.00005642, -0.00000264, 0.00004026, 0.00006395, 0.00006459, 0.00004336, 0.00000597, -0.00003886, -0.00008136, -0.00011289, -0.00012764, -0.00012372, -0.00010320, -0.00007138, -0.00003536, -0.00000236, 0.00002190, 0.00003426, 0.00003454, 0.00002517, 0.00001030, -0.00000554, -0.00001859, -0.00002682, },
		{ -0.00001447, -0.00001328, -0.00001277, -0.00001122, -0.00000799, -0.00000383, -0.00000082, -0.00000175, -0.00000916, -0.00002439, -0.00004671, -0.00007308, -0.00009840, -0.00011644, -0.00012128, -0.00010874, -0.00007772, -0.00003093, 0.00002519, 0.00008142, 0.00012742, 0.00015376, 0.00015402, 0.00012627, 0.00007386, 0.00000502, -0.00006846, -0.00013345, -0.00017796, -0.00019351, -0.00017686, -0.00013073, -0.00006328, 0.00001342, 0.00008572, 0.00014105, 0.00017031, 0.00016950, 0.00014034, 0.00008973, 0.00002817, -0.00003247, -0.00008134, -0.00011068, -0.00011718, -0.00010231, -0.00007169, -0.00003357, 0.00000304, 0.00003034, 0.00004327, 0.00004033, 0.00002355, -0.00000229, -0.00003099, -0.00005644, -0.00007398, -0.00008124, -0.00007844, -0.00006792, -0.00005331, -0.00003842, -0.00002626, -0.00001832, },
		{ -0.00003617, -0.00003732, -0.00003731, -0.00003747, -0.00003829, -0.00003913, -0.00003842, -0.00003425, -0.00002508, -0.00001062, 0.00000774, 0.00002677, 0.00004192, 0.00004825, 0.00004165, 0.00002007, -0.00001554, -0.00006094, -0.00010898, -0.00015080, -0.00017744, -0.00018177, -0.00016013, -0.00011346, -0.00004747, 0.00002808, 0.00010105, 0.00015908, 0.00019199, 0.00019377, 0.00016382, 0.00010719, 0.00003368, -0.00004397, -0.00011248, -0.00016055, -0.00018089, -0.00017150, -0.00013591, -0.00008223, -0.00002144, 0.00003487, 0.00007676, 0.00009780, 0.00009603, 0.00007409, 0.00003830, -0.00000286, -0.00004070, -0.00006804, -0.00008060, -0.00007760, -0.00006152, -0.00003730, -0.00001096, 0.00001177, 0.00002671, 0.00003192, 0.00002782, 0.00001676, 0.00000218, -0.00001243, -0.00002435, -0.00003224, },
		{ 0.00001623, 0.00001685, 0.00001535, 0.00001317, 0.00001103, 0.00000873, 0.00000528, -0.00000057, -0.00000974, -0.00002215, -0.00003639, -0.00004971, -0.00005854, -0.00005925, -0.00004924, -0.00002784, 0.00000311, 0.00003916, 0.00007385, 0.00009986, 0.00011044, 0.00010101, 0.00007040, 0.00002142, -0.00003925, -0.00010204, -0.00015623, -0.00019187, -0.00020180, -0.00018314, -0.00013804, -0.00007336, 0.00000042, 0.00007118, 0.00012737, 0.00016015, 0.00016503, 0.00014256, 0.00009816, 0.00004086, -0.00001849, -0.00006937, -0.00010362, -0.00011683, -0.00010896, -0.00008404, -0.00004908, -0.00001245, 0.00001798, 0.00003634, 0.00003984, 0.00002906, 0.00000755, -0.00001924, -0.00004524, -0.00006517, -0.00007555, -0.00007525, -0.00006546, -0.00004913, -0.00003005, -0.00001193, 0.00000248, 0.00001183, },
		{ -0.00007422, -0.00006621, -0.00005584, -0.00004514, -0.00003534, -0.00002671, -0.00001864, -0.00001014, -0.00000042, 0.00001063, 0.00002204, 0.00003184, 0.00003743, 0.00003629, 0.00002679, 0.00000888, -0.00001546, -0.00004233, -0.00006644, -0.00008214, -0.00008463, -0.00007118, -0.00004196, -0.00000039, 0.00004722, 0.00009274, 0.00012765, 0.00014475, 0.00013966, 0.00011187, 0.00006498, 0.00000621, -0.00005488, -0.00010815, -0.00014478, -0.00015895, -0.00014891, -0.00011725, -0.00007035, -0.00001705, 0.00003308, 0.00007163, 0.00009287, 0.00009473, 0.00007898, 0.00005063, 0.00001667, -0.00001545, -0.00003939, -0.00005117, -0.00004983, -0.00003735, -0.00001801, 0.00000270, 0.00001942, 0.00002805, 0.00002660, 0.00001541, -0.00000312, -0.00002520, -0.00004664, -0.00006382, -0.00007437, -0.00007757, },
		{ 0.00001173, -0.00000612, -0.00002333, -0.00003762, -0.00004777, -0.00005377, -0.00005662, -0.00005784, -0.00005882, -0.00006035, -0.00006220, -0.00006320, -0.00006151, -0.00005525, -0.00004317, -0.00002527, -0.00000313, 0.00002017, 0.00004057, 0.00005379, 0.00005638, 0.00004662, 0.00002515, -0.00000489, -0.00003829, -0.00006859, -0.00008944, -0.00009581, -0.00008520, -0.00005823, -0.00001869, 0.00002704, 0.00007119, 0.00010609, 0.00012562, 0.00012645, 0.00010866, 0.00007572, 0.00003376, -0.00000964, -0.00004698, -0.00007230, -0.00008222, -0.00007654, -0.00005805, -0.00003181, -0.00000395, 0.00001965, 0.00003461, 0.00003882, 0.00003279, 0.00001926, 0.00000249, -0.00001286, -0.00002282, -0.00002497, -0.00001894, -0.00000637, 0.00000958, 0.00002504, 0.00003637, 0.00004093, 0.00003769, 0.00002728, },
		{ 0.00001859, 0.00003088, 0.00003907, 0.00004167, 0.00003847, 0.00003057, 0.00001999, 0.00000908, -0.00000010, -0.00000635, -0.00000963, -0.00001102, -0.00001238, -0.00001572, -0.00002261, -0.00003353, -0.00004760, -0.00006262, -0.00007545, -0.00008276, -0.00008181, -0.00007127, -0.00005165, -0.00002545, 0.00000324, 0.00002943, 0.00004828, 0.00005612, 0.00005127, 0.00003449, 0.00000894, -0.00002042, -0.00004777, -0.00006762, -0.00007590, -0.00007079, -0.00005311, -0.00002611, 0.00000521, 0.00003511, 0.00005834, 0.00007113, 0.00007192, 0.00006161, 0.00004319, 0.00002107, 0.00000003, -0.00001580, -0.00002379, -0.00002328, -0.00001557, -0.00000352, 0.00000925, 0.00001928, 0.00002402, 0.00002247, 0.00001530, 0.00000465, -0.00000651, -0.00001509, -0.00001867, -0.00001610, -0.00000771, 0.00000475, },
		{ -0.00000234, -0.00000615, -0.00000557, -0.00000029, 0.00000879, 0.00001976, 0.00003016, 0.00003762, 0.00004042, 0.00003797, 0.00003093, 0.00002114, 0.00001107, 0.00000331, -0.00000019, 0.00000139, 0.00000748, 0.00001616, 0.00002460, 0.00002967, 0.00002877, 0.00002043, 0.00000481, -0.00001625, -0.00003949, -0.00006091, -0.00007657, -0.00008344, -0.00008006, -0.00006691, -0.00004634, -0.00002208, 0.00000145, 0.00002015, 0.00003097, 0.00003257, 0.00002554, 0.00001218, -0.00000400, -0.00001914, -0.00002983, -0.00003386, -0.00003063, -0.00002123, -0.00000808, 0.00000565, 0.00001682, 0.00002305, 0.00002323, 0.00001770, 0.00000817, -0.00000281, -0.00001242, -0.00001835, -0.00001923, -0.00001502, -0.00000693, 0.00000294, 0.00001212, 0.00001839, 0.00002040, 0.00001793, 0.00001197, 0.00000443, },
		{ -0.00000893, -0.00000479, -0.00000370, -0.00000544, -0.00000886, -0.00001220, -0.00001352, -0.00001126, -0.00000466, 0.00000593, 0.00001907, 0.00003250, 0.00004360, 0.00005002, 0.00005023, 0.00004394, 0.00003227, 0.00001752, 0.00000278, -0.00000874, -0.00001443, -0.00001285, -0.00000407, 0.00001025, 0.00002719, 0.00004317, 0.00005466, 0.00005890, 0.00005450, 0.00004172, 0.00002239, -0.00000044, -0.00002317, -0.00004233, -0.00005525, -0.00006060, -0.00005847, -0.00005033, -0.00003856, -0.00002587, -0.00001473, -0.00000680, -0.00000268, -0.00000190, -0.00000314, -0.00000469, -0.00000494, -0.00000285, 0.00000175, 0.00000815, 0.00001492, 0.00002029, 0.00002263, 0.00002089, 0.00001489, 0.00000543, -0.00000589, -0.00001704, -0.00002598, -0.00003120, -0.00003202, -0.00002876, -0.00002262, -0.00001538, },
		{ -0.00001435, -0.00002065, -0.00002360, -0.00002357, -0.00002161, -0.00001919, -0.00001777, -0.00001843, -0.00002148, -0.00002641, -0.00003191, -0.00003617, -0.00003732, -0.00003391, -0.00002532, -0.00001201, 0.00000448, 0.00002173, 0.00003694, 0.00004745, 0.00005132, 0.00004780, 0.00003748, 0.00002228, 0.00000510, -0.00001077, -0.00002222, -0.00002696, -0.00002396, -0.00001364, 0.00000224, 0.00002094, 0.00003926, 0.00005416, 0.00006333, 0.00006553, 0.00006077, 0.00005018, 0.00003570, 0.00001962, 0.00000412, -0.00000915, -0.00001925, -0.00002605, -0.00003001, -0.00003195, -0.00003270, -0.00003285, -0.00003254, -0.00003148, -0.00002910, -0.00002477, -0.00001811, -0.00000918, 0.00000134, 0.00001228, 0.00002210, 0.00002925, 0.00003249, 0.00003122, 0.00002560, 0.00001658, 0.00000572, -0.00000516, },
		{ 0.00002059, 0.00002198, 0.00001959, 0.00001420, 0.00000710, -0.00000019, -0.00000624, -0.00001003, -0.00001122, -0.00001016, -0.00000780, -0.00000549, -0.00000457, -0.00000608, -0.00001039, -0.00001715, -0.00002521, -0.00003292, -0.00003843, -0.00004014, -0.00003703, -0.00002899, -0.00001687, -0.00000239, 0.00001216, 0.00002437, 0.00003216, 0.00003420, 0.00003017, 0.00002086, 0.00000799, -0.00000609, -0.00001884, -0.00002800, -0.00003198, -0.00003010, -0.00002269, -0.00001094, 0.00000338, 0.00001827, 0.00003184, 0.00004264, 0.00004981, 0.00005314, 0.00005293, 0.00004980, 0.00004452, 0.00003776, 0.00003001, 0.00002153, 0.00001248, 0.00000299, -0.00000669, -0.00001605, -0.00002438, -0.00003078, -0.00003437, -0.00003449, -0.00003088, -0.00002385, -0.00001426, -0.00000346, 0.00000693, 0.00001536, },
		{ -0.00001713, -0.00001510, -0.00001034, -0.00000394, 0.00000273, 0.00000827, 0.00001149, 0.00001169, 0.00000879, 0.00000331, -0.00000369, -0.00001087, -0.00001686, -0.00002056, -0.00002136, -0.00001927, -0.00001489, -0.00000930, -0.00000376, 0.00000049, 0.00000257, 0.00000209, -0.00000071, -0.00000508, -0.00000989, -0.00001390, -0.00001605, -0.00001568, -0.00001267, -0.00000749, -0.00000110, 0.00000526, 0.00001029, 0.00001297, 0.00001266, 0.00000931, 0.00000339, -0.00000415, -0.00001214, -0.00001935, -0.00002473, -0.00002759, -0.00002762, -0.00002493, -0.00001994, -0.00001327, -0.00000557, 0.00000257, 0.00001067, 0.00001840, 0.00002550, 0.00003173, 0.00003682, 0.00004042, 0.00004211, 0.00004151, 0.00003837, 0.00003269, 0.00002479, 0.00001537, 0.00000544, -0.00000383, -0.00001124, -0.00001585, },
		{ 0.00002528, 0.00002128, 0.00001578, 0.00000979, 0.00000442, 0.00000066, -0.00000075, 0.00000046, 0.00000405, 0.00000930, 0.00001513, 0.00002026, 0.00002347, 0.00002381, 0.00002080, 0.00001448, 0.00000545, -0.00000522, -0.00001621, -0.00002614, -0.00003378, -0.00003827, -0.00003921, -0.00003668, -0.00003124, -0.00002375, -0.00001523, -0.00000670, 0.00000099, 0.00000724, 0.00001174, 0.00001447, 0.00001560, 0.00001546, 0.00001439, 0.00001272, 0.00001068, 0.00000845, 0.00000611, 0.00000372, 0.00000131, -0.00000107, -0.00000336, -0.00000548, -0.00000740, -0.00000910, -0.00001061, -0.00001200, -0.00001333, -0.00001463, -0.00001586, -0.00001686, -0.00001737, -0.00001707, -0.00001561, -0.00001273, -0.00000835, -0.00000259, 0.00000414, 0.00001119, 0.00001777, 0.00002304, 0.00002629, 0.00002706, },
		{ -0.00001366, -0.00000990, -0.00000525, -0.00000050, 0.00000351, 0.00000610, 0.00000683, 0.00000563, 0.00000279, -0.00000102, -0.00000493, -0.00000795, -0.00000917, -0.00000794, -0.00000399, 0.00000249, 0.00001087, 0.00002011, 0.00002894, 0.00003602, 0.00004013, 0.00004037, 0.00003632, 0.00002807, 0.00001628, 0.00000207, -0.00001311, -0.00002763, -0.00003994, -0.00004872, -0.00005306, -0.00005256, -0.00004739, -0.00003819, -0.00002609, -0.00001245, 0.00000124, 0.00001356, 0.00002335, 0.00002979, 0.00003253, 0.00003167, 0.00002773, 0.00002156, 0.00001423, 0.00000683, 0.00000037, -0.00000441, -0.00000706, -0.00000753, -0.00000611, -0.00000339, -0.00000014, 0.00000283, 0.00000481, 0.00000527, 0.00000403, 0.00000120, -0.00000279, -0.00000727, -0.00001147, -0.00001465, -0.00001623, -0.00001589, },
		{ 0.00001556, 0.00001153, 0.00000655, 0.00000145, -0.00000300, -0.00000617, -0.00000771, -0.00000755, -0.00000594, -0.00000343, -0.00000071, 0.00000144, 0.00000232, 0.00000143, -0.00000146, -0.00000617, -0.00001221, -0.00001871, -0.00002462, -0.00002881, -0.00003024, -0.00002819, -0.00002234, -0.00001289, -0.00000061, 0.00001324, 0.00002704, 0.00003904, 0.00004757, 0.00005130, 0.00004947, 0.00004198, 0.00002950, 0.00001338, -0.00000452, -0.00002207, -0.00003716, -0.00004798, -0.00005330, -0.00005261, -0.00004622, -0.00003516, -0.00002105, -0.00000582, 0.00000853, 0.00002026, 0.00002810, 0.00003140, 0.00003023, 0.00002527, 0.00001769, 0.00000897, 0.00000058, -0.00000616, -0.00001033, -0.00001148, -0.00000969, -0.00000549, 0.00000023, 0.00000641, 0.00001201, 0.00001613, 0.00001818, 0.00001793, },
		{ -0.00000647, 0.00000005, 0.00000680, 0.00001284, 0.00001742, 0.00002005, 0.00002054, 0.00001902, 0.00001589, 0.00001175, 0.00000733, 0.00000336, 0.00000049, -0.00000081, -0.00000033, 0.00000183, 0.00000525, 0.00000923, 0.00001289, 0.00001524, 0.00001542, 0.00001277, 0.00000704, -0.00000153, -0.00001220, -0.00002375, -0.00003466, -0.00004329, -0.00004812, -0.00004806, -0.00004263, -0.00003208, -0.00001744, -0.00000039, 0.00001694, 0.00003229, 0.00004361, 0.00004938, 0.00004884, 0.00004215, 0.00003032, 0.00001513, -0.00000121, -0.00001637, -0.00002830, -0.00003548, -0.00003718, -0.00003353, -0.00002544, -0.00001445, -0.00000245, 0.00000868, 0.00001732, 0.00002236, 0.00002335, 0.00002052, 0.00001467, 0.00000701, -0.00000105, -0.00000817, -0.00001326, -0.00001562, -0.00001506, -0.00001180, },
		{ 0.00000064, -0.00000536, -0.00001015, -0.00001310, -0.00001388, -0.00001251, -0.00000927, -0.00000467, 0.00000066, 0.00000603, 0.00001078, 0.00001435, 0.00001632, 0.00001647, 0.00001478, 0.00001148, 0.00000704, 0.00000211, -0.00000250, -0.00000596, -0.00000750, -0.00000661, -0.00000309, 0.00000281, 0.00001039, 0.00001855, 0.00002591, 0.00003103, 0.00003265, 0.00002994, 0.00002264, 0.00001125, -0.00000306, -0.00001854, -0.00003314, -0.00004477, -0.00005173, -0.00005290, -0.00004803, -0.00003774, -0.00002349, -0.00000730, 0.00000853, 0.00002184, 0.00003090, 0.00003472, 0.00003318, 0.00002699, 0.00001758, 0.00000676, -0.00000352, -0.00001160, -0.00001629, -0.00001709, -0.00001420, -0.00000846, -0.00000112, 0.00000640, 0.00001273, 0.00001685, 0.00001817, 0.00001662, 0.00001263, 0.00000698, },
		{ 0.00000726, 0.00000953, 0.00001041, 0.00000985, 0.00000805, 0.00000541, 0.00000242, -0.00000042, -0.00000271, -0.00000414, -0.00000456, -0.00000393, -0.00000231, 0.00000017, 0.00000329, 0.00000680, 0.00001037, 0.00001357, 0.00001595, 0.00001704, 0.00001640, 0.00001379, 0.00000918, 0.00000287, -0.00000450, -0.00001198, -0.00001843, -0.00002270, -0.00002381, -0.00002119, -0.00001479, -0.00000522, 0.00000637, 0.00001833, 0.00002886, 0.00003623, 0.00003911, 0.00003682, 0.00002946, 0.00001794, 0.00000386, -0.00001080, -0.00002393, -0.00003372, -0.00003887, -0.00003889, -0.00003407, -0.00002546, -0.00001463, -0.00000340, 0.00000652, 0.00001378, 0.00001761, 0.00001791, 0.00001520, 0.00001048, 0.00000495, -0.00000015, -0.00000386, -0.00000559, -0.00000521, -0.00000305, 0.00000028, 0.00000398, },
		{ 0.00000626, 0.00000430, 0.00000272, 0.00000145, 0.00000033, -0.00000083, -0.00000216, -0.00000363, -0.00000512, -0.00000640, -0.00000721, -0.00000732, -0.00000664, -0.00000520, -0.00000316, -0.00000077, 0.00000171, 0.00000408, 0.00000625, 0.00000828, 0.00001031, 0.00001255, 0.00001511, 0.00001797, 0.00002087, 0.00002333, 0.00002473, 0.00002436, 0.00002165, 0.00001630, 0.00000842, -0.00000140, -0.00001213, -0.00002242, -0.00003082, -0.00003600, -0.00003705, -0.00003361, -0.00002604, -0.00001535, -0.00000305, 0.00000906, 0.00001920, 0.00002594, 0.00002841, 0.00002645, 0.00002061, 0.00001207, 0.00000233, -0.00000695, -0.00001434, -0.00001884, -0.00001998, -0.00001793, -0.00001334, -0.00000722, -0.00000071, 0.00000517, 0.00000964, 0.00001230, 0.00001314, 0.00001244, 0.00001071, 0.00000850, },
		{ 0.00000586, 0.00001032, 0.00001433, 0.00001773, 0.00002040, 0.00002220, 0.00002297, 0.00002248, 0.00002053, 0.00001699, 0.00001189, 0.00000546, -0.00000182, -0.00000929, -0.00001615, -0.00002162, -0.00002505, -0.00002606, -0.00002462, -0.00002103, -0.00001589, -0.00000997, -0.00000407, 0.00000117, 0.00000537, 0.00000848, 0.00001074, 0.00001262, 0.00001463, 0.00001721, 0.00002052, 0.00002438, 0.00002825, 0.00003135, 0.00003275, 0.00003163, 0.00002747, 0.00002019, 0.00001025, -0.00000139, -0.00001341, -0.00002431, -0.00003271, -0.00003755, -0.00003832, -0.00003511, -0.00002862, -0.00002002, -0.00001071, -0.00000214, 0.00000451, 0.00000847, 0.00000949, 0.00000783, 0.00000420, -0.00000046, -0.00000513, -0.00000892, -0.00001118, -0.00001161, -0.00001024, -0.00000734, -0.00000337, 0.00000118, },
		{ -0.00001188, -0.00001446, -0.00001614, -0.00001666, -0.00001582, -0.00001350, -0.00000967, -0.00000436, 0.00000223, 0.00000976, 0.00001772, 0.00002540, 0.00003197, 0.00003658, 0.00003847, 0.00003711, 0.00003232, 0.00002438, 0.00001399, 0.00000221, -0.00000964, -0.00002023, -0.00002838, -0.00003328, -0.00003459, -0.00003252, -0.00002773, -0.00002121, -0.00001410, -0.00000742, -0.00000195, 0.00000196, 0.00000438, 0.00000577, 0.00000680, 0.00000822, 0.00001057, 0.00001411, 0.00001869, 0.00002373, 0.00002840, 0.00003171, 0.00003276, 0.00003093, 0.00002603, 0.00001837, 0.00000871, -0.00000184, -0.00001201, -0.00002061, -0.00002668, -0.00002970, -0.00002961, -0.00002682, -0.00002209, -0.00001638, -0.00001069, -0.00000588, -0.00000257, -0.00000103, -0.00000125, -0.00000293, -0.00000562, -0.00000878, },
		{ -0.00000191, 0.00000202, 0.00000452, 0.00000530, 0.00000427, 0.00000153, -0.00000266, -0.00000789, -0.00001360, -0.00001914, -0.00002378, -0.00002681, -0.00002760, -0.00002568, -0.00002088, -0.00001337, -0.00000367, 0.00000729, 0.00001835, 0.00002819, 0.00003554, 0.00003939, 0.00003912, 0.00003466, 0.00002651, 0.00001566, 0.00000350, -0.00000843, -0.00001866, -0.00002604, -0.00002990, -0.00003011, -0.00002712, -0.00002183, -0.00001536, -0.00000891, -0.00000346, 0.00000036, 0.00000236, 0.00000279, 0.00000227, 0.00000157, 0.00000147, 0.00000254, 0.00000504, 0.00000881, 0.00001333, 0.00001782, 0.00002139, 0.00002324, 0.00002279, 0.00001983, 0.00001457, 0.00000757, -0.00000032, -0.00000812, -0.00001488, -0.00001983, -0.00002250, -0.00002273, -0.00002073, -0.00001695, -0.00001207, -0.00000682, },
		{ -0.00002534, -0.00002713, -0.00002647, -0.00002339, -0.00001821, -0.00001146, -0.00000385, 0.00000379, 0.00001066, 0.00001598, 0.00001914, 0.00001972, 0.00001758, 0.00001286, 0.00000603, -0.00000218, -0.00001083, -0.00001885, -0.00002519, -0.00002895, -0.00002953, -0.00002672, -0.00002075, -0.00001234, -0.00000255, 0.00000731, 0.00001588, 0.00002195, 0.00002467, 0.00002366, 0.00001909, 0.00001168, 0.00000253, -0.00000699, -0.00001552, -0.00002187, -0.00002527, -0.00002540, -0.00002252, -0.00001729, -0.00001072, -0.00000393, 0.00000205, 0.00000643, 0.00000882, 0.00000924, 0.00000808, 0.00000602, 0.00000382, 0.00000221, 0.00000170, 0.00000248, 0.00000443, 0.00000710, 0.00000983, 0.00001191, 0.00001266, 0.00001164, 0.00000866, 0.00000388, -0.00000225, -0.00000904, -0.00001568, -0.00002135, },
		{ 0.00000793, 0.00000371, -0.00000218, -0.00000933, -0.00001709, -0.00002464, -0.00003111, -0.00003571, -0.00003781, -0.00003703, -0.00003331, -0.00002690, -0.00001838, -0.00000856, 0.00000158, 0.00001101, 0.00001875, 0.00002403, 0.00002634, 0.00002552, 0.00002180, 0.00001575, 0.00000823, 0.00000029, -0.00000700, -0.00001265, -0.00001596, -0.00001654, -0.00001445, -0.00001017, -0.00000449, 0.00000153, 0.00000678, 0.00001026, 0.00001124, 0.00000938, 0.00000480, -0.00000196, -0.00000998, -0.00001812, -0.00002523, -0.00003028, -0.00003255, -0.00003172, -0.00002793, -0.00002172, -0.00001396, -0.00000566, 0.00000214, 0.00000858, 0.00001310, 0.00001549, 0.00001587, 0.00001470, 0.00001262, 0.00001032, 0.00000843, 0.00000737, 0.00000730, 0.00000809, 0.00000937, 0.00001057, 0.00001108, 0.00001034, },
		},
		{ { -0.00000186, -0.00000202, -0.00000246, -0.00000313, -0.00000391, -0.00000470, -0.00000533, -0.00000568, -0.00000562, -0.00000506, -0.00000394, -0.00000226, -0.00000006, 0.00000255, 0.00000545, 0.00000845, 0.00001136, 0.00001398, 0.00001611, 0.00001760, 0.00001831, 0.00001820, 0.00001724, 0.00001550, 0.00001308, 0.00001016, 0.00000695, 0.00000368, 0.00000058, -0.00000210, -0.00000417, -0.00000549, -0.00000593, -0.00000549, -0.00000417, -0.00000210, 0.00000058, 0.00000368, 0.00000695, 0.00001016, 0.00001308, 0.00001550, 0.00001724, 0.00001820, 0.00001831, 0.00001760, 0.00001611, 0.00001398, 0.00001136, 0.00000845, 0.00000545, 0.00000255, -0.00000006, -0.00000226, -0.00000394, -0.00000506, -0.00000562, -0.00000568, -0.00000533, -0.00000470, -0.00000391, -0.00000313, -0.00000246, -0.00000202, },
		{ -0.00000291, -0.00000271, -0.00000227, -0.00000171, -0.00000116, -0.00000076, -0.00000062, -0.00000085, -0.00000150, -0.00000258, -0.00000405, -0.00000584, -0.00000780, -0.00000977, -0.00001156, -0.00001300, -0.00001389, -0.00001410, -0.00001352, -0.00001210, -0.00000985, -0.00000685, -0.00000324, 0.00000079, 0.00000502, 0.00000919, 0.00001303, 0.00001632, 0.00001883, 0.00002041, 0.00002095, 0.00002043, 0.00001889, 0.00001644, 0.00001327, 0.00000958, 0.00000563, 0.00000169, -0.00000198, -0.00000515, -0.00000764, -0.00000930, -0.00001008, -0.00000998, -0.00000906, -0.00000744, -0.00000530, -0.00000284, -0.00000028, 0.00000218, 0.00000434, 0.00000605, 0.00000719, 0.00000773, 0.00000765, 0.00000701, 0.00000592, 0.00000450, 0.00000291, 0.00000131, -0.00000017, -0.00000139, -0.00000228, -0.00000278, },
		{ -0.00000565, -0.00000652, -0.00000741, -0.00000819, -0.00000871, -0.00000889, -0.00000863, -0.00000792, -0.00000677, -0.00000526, -0.00000351, -0.00000168, 0.00000006, 0.00000150, 0.00000248, 0.00000285, 0.00000248, 0.00000132, -0.00000061, -0.00000323, -0.00000642, -0.00000997, -0.00001363, -0.00001715, -0.00002025, -0.00002267, -0.00002419, -0.00002463, -0.00002391, -0.00002199, -0.00001894, -0.00001488, -0.00001004, -0.00000467, 0.00000092, 0.00000642, 0.00001152, 0.00001592, 0.00001940, 0.00002178, 0.00002297, 0.00002295, 0.00002178, 0.00001960, 0.00001660, 0.00001302, 0.00000912, 0.00000517, 0.00000142, -0.00000193, -0.00000470, -0.00000680, -0.00000818, -0.00000887, -0.00000893, -0.00000851, -0.00000775, -0.00000681, -0.00000588, -0.00000508, -0.00000454, -0.00000433, -0.00000447, -0.00000493, },
		{ 0.00000146, 0.00000117, 0.00000081, 0.00000025, -0.00000059, -0.00000177, -0.00000331, -0.00000516, -0.00000722, -0.00000937, -0.00001144, -0.00001323, -0.00001456, -0.00001525, -0.00001519, -0.00001428, -0.00001251, -0.00000993, -0.00000666, -0.00000288, 0.00000117, 0.00000522, 0.00000898, 0.00001217, 0.00001453, 0.00001587, 0.00001603, 0.00001497, 0.00001271, 0.00000936, 0.00000512, 0.00000026, -0.00000492, -0.00001007, -0.00001486, -0.00001897, -0.00002211, -0.00002408, -0.00002475, -0.00002407, -0.00002210, -0.00001896, -0.00001487, -0.00001008, -0.00000491, 0.00000034, 0.00000536, 0.00000986, 0.00001363, 0.00001650, 0.00001838, 0.00001925, 0.00001919, 0.00001829, 0.00001674, 0.00001472, 0.00001245, 0.00001013, 0.00000793, 0.00000600, 0.00000441, 0.00000321, 0.00000237, 0.00000182, },
		{ 0.00000962, 0.00000903, 0.00000854, 0.00000827, 0.00000829, 0.00000862, 0.00000923, 0.00001003, 0.00001090, 0.00001167, 0.00001218, 0.00001224, 0.00001171, 0.00001046, 0.00000843, 0.00000562, 0.00000211, -0.00000197, -0.00000642, -0.00001099, -0.00001540, -0.00001935, -0.00002257, -0.00002481, -0.00002589, -0.00002567, -0.00002414, -0.00002134, -0.00001741, -0.00001258, -0.00000713, -0.00000139, 0.00000428, 0.00000954, 0.00001407, 0.00001760, 0.00001993, 0.00002095, 0.00002064, 0.00001907, 0.00001641, 0.00001287, 0.00000875, 0.00000435, -0.00000000, -0.00000401, -0.00000739, -0.00000995, -0.00001156, -0.00001215, -0.00001174, -0.00001042, -0.00000837, -0.00000576, -0.00000284, 0.00000017, 0.00000304, 0.00000558, 0.00000767, 0.00000920, 0.00001016, 0.00001058, 0.00001055, 0.00001018, },
		{ -0.00000135, -0.00000068, -0.00000032, -0.00000035, -0.00000077, -0.00000155, -0.00000261, -0.00000382, -0.00000501, -0.00000600, -0.00000662, -0.00000671, -0.00000615, -0.00000487, -0.00000287, -0.00000021, 0.00000299, 0.00000654, 0.00001021, 0.00001372, 0.00001680, 0.00001920, 0.00002068, 0.00002105, 0.00002022, 0.00001816, 0.00001491, 0.00001063, 0.00000553, -0.00000011, -0.00000598, -0.00001172, -0.00001701, -0.00002153, -0.00002504, -0.00002733, -0.00002831, -0.00002795, -0.00002633, -0.00002360, -0.00001998, -0.00001574, -0.00001118, -0.00000661, -0.00000233, 0.00000141, 0.00000442, 0.00000656, 0.00000777, 0.00000807, 0.00000756, 0.00000636, 0.00000467, 0.00000271, 0.00000067, -0.00000123, -0.00000284, -0.00000402, -0.00000471, -0.00000491, -0.00000465, -0.00000403, -0.00000317, -0.00000223, },
		{ 0.00000219, 0.00000239, 0.00000284, 0.00000355, 0.00000448, 0.00000556, 0.00000667, 0.00000767, 0.00000841, 0.00000875, 0.00000854, 0.00000770, 0.00000615, 0.00000391, 0.00000103, -0.00000236, -0.00000611, -0.00000998, -0.00001375, -0.00001716, -0.00001996, -0.00002194, -0.00002294, -0.00002285, -0.00002163, -0.00001931, -0.00001603, -0.00001196, -0.00000734, -0.00000245, 0.00000239, 0.00000690, 0.00001079, 0.00001382, 0.00001582, 0.00001669, 0.00001639, 0.00001498, 0.00001259, 0.00000939, 0.00000564, 0.00000159, -0.00000247, -0.00000630, -0.00000964, -0.00001231, -0.00001419, -0.00001521, -0.00001537, -0.00001474, -0.00001342, -0.00001159, -0.00000942, -0.00000710, -0.00000481, -0.00000272, -0.00000093, 0.00000047, 0.00000145, 0.00000205, 0.00000232, 0.00000235, 0.00000226, 0.00000217, },
		{ -0.00000491, -0.00000428, -0.00000377, -0.00000340, -0.00000313, -0.00000290, -0.00000264, -0.00000225, -0.00000164, -0.00000071, 0.00000061, 0.00000235, 0.00000452, 0.00000708, 0.00000995, 0.00001301, 0.00001610, 0.00001906, 0.00002169, 0.00002383, 0.00002529, 0.00002597, 0.00002576, 0.00002464, 0.00002262, 0.00001980, 0.00001631, 0.00001232, 0.00000805, 0.00000374, -0.00000038, -0.00000409, -0.00000719, -0.00000952, -0.00001100, -0.00001157, -0.00001127, -0.00001017, -0.00000840, -0.00000614, -0.00000358, -0.00000094, 0.00000158, 0.00000379, 0.00000554, 0.00000671, 0.00000724, 0.00000712, 0.00000638, 0.00000513, 0.00000346, 0.00000153, -0.00000051, -0.00000251, -0.00000433, -0.00000586, -0.00000702, -0.00000777, -0.00000811, -0.00000808, -0.00000772, -0.00000714, -0.00000642, -0.00000565, },
		{ 0.00000247, 0.00000124, 0.00000019, -0.00000066, -0.00000132, -0.00000182, -0.00000221, -0.00000255, -0.00000290, -0.00000332, -0.00000387, -0.00000458, -0.00000547, -0.00000653, -0.00000774, -0.00000904, -0.00001036, -0.00001160, -0.00001267, -0.00001348, -0.00001393, -0.00001395, -0.00001349, -0.00001253, -0.00001107, -0.00000916, -0.00000686, -0.00000428, -0.00000152, 0.00000130, 0.00000403, 0.00000656, 0.00000879, 0.00001062, 0.00001200, 0.00001290, 0.00001334, 0.00001334, 0.00001298, 0.00001235, 0.00001153, 0.00001065, 0.00000980, 0.00000908, 0.00000855, 0.00000827, 0.00000825, 0.00000850, 0.00000897, 0.00000961, 0.00001035, 0.00001110, 0.00001177, 0.00001229, 0.00001258, 0.00001260, 0.00001230, 0.00001170, 0.00001080, 0.00000964, 0.00000830, 0.00000683, 0.00000533, 0.00000385, },
		{ 0.00000598, 0.00000651, 0.00000679, 0.00000680, 0.00000655, 0.00000606, 0.00000535, 0.00000447, 0.00000348, 0.00000245, 0.00000144, 0.00000054, -0.00000021, -0.00000073, -0.00000100, -0.00000099, -0.00000071, -0.00000017, 0.00000057, 0.00000146, 0.00000240, 0.00000332, 0.00000412, 0.00000471, 0.00000501, 0.00000495, 0.00000449, 0.00000360, 0.00000228, 0.00000057, -0.00000149, -0.00000382, -0.00000633, -0.00000894, -0.00001152, -0.00001399, -0.00001624, -0.00001820, -0.00001980, -0.00002101, -0.00002180, -0.00002216, -0.00002212, -0.00002171, -0.00002097, -0.00001997, -0.00001875, -0.00001738, -0.00001591, -0.00001437, -0.00001280, -0.00001124, -0.00000968, -0.00000815, -0.00000663, -0.00000514, -0.00000366, -0.00000221, -0.00000078, 0.00000061, 0.00000193, 0.00000317, 0.00000428, 0.00000523, },
		{ -0.00001191, -0.00001135, -0.00001070, -0.00000992, -0.00000899, -0.00000789, -0.00000660, -0.00000516, -0.00000361, -0.00000202, -0.00000047, 0.00000092, 0.00000204, 0.00000278, 0.00000306, 0.00000278, 0.00000190, 0.00000043, -0.00000162, -0.00000415, -0.00000705, -0.00001016, -0.00001332, -0.00001632, -0.00001898, -0.00002110, -0.00002253, -0.00002316, -0.00002290, -0.00002174, -0.00001971, -0.00001691, -0.00001346, -0.00000954, -0.00000538, -0.00000117, 0.00000284, 0.00000645, 0.00000948, 0.00001177, 0.00001325, 0.00001384, 0.00001358, 0.00001250, 0.00001071, 0.00000835, 0.00000557, 0.00000255, -0.00000054, -0.00000354, -0.00000632, -0.00000877, -0.00001081, -0.00001241, -0.00001356, -0.00001430, -0.00001467, -0.00001474, -0.00001458, -0.00001426, -0.00001385, -0.00001340, -0.00001292, -0.00001243, },
		{ 0.00000292, 0.00000362, 0.00000424, 0.00000466, 0.00000479, 0.00000454, 0.00000387, 0.00000278, 0.00000133, -0.00000038, -0.00000220, -0.00000397, -0.00000547, -0.00000652, -0.00000694, -0.00000658, -0.00000535, -0.00000322, -0.00000023, 0.00000351, 0.00000782, 0.00001246, 0.00001715, 0.00002157, 0.00002542, 0.00002840, 0.00003026, 0.00003082, 0.00002996, 0.00002765, 0.00002396, 0.00001904, 0.00001313, 0.00000651, -0.00000048, -0.00000748, -0.00001412, -0.00002009, -0.00002510, -0.00002894, -0.00003146, -0.00003263, -0.00003248, -0.00003111, -0.00002871, -0.00002551, -0.00002177, -0.00001776, -0.00001374, -0.00000996, -0.00000658, -0.00000375, -0.00000154, 0.00000004, 0.00000103, 0.00000153, 0.00000166, 0.00000156, 0.00000136, 0.00000120, 0.00000117, 0.00000132, 0.00000169, 0.00000224, },
		{ -0.00001916, -0.00001867, -0.00001775, -0.00001626, -0.00001411, -0.00001126, -0.00000772, -0.00000358, 0.00000099, 0.00000578, 0.00001050, 0.00001484, 0.00001850, 0.00002116, 0.00002258, 0.00002256, 0.00002102, 0.00001795, 0.00001347, 0.00000780, 0.00000127, -0.00000574, -0.00001277, -0.00001936, -0.00002505, -0.00002942, -0.00003213, -0.00003294, -0.00003174, -0.00002853, -0.00002345, -0.00001677, -0.00000886, -0.00000015, 0.00000885, 0.00001764, 0.00002572, 0.00003266, 0.00003809, 0.00004177, 0.00004355, 0.00004341, 0.00004146, 0.00003788, 0.00003298, 0.00002710, 0.00002062, 0.00001392, 0.00000735, 0.00000123, -0.00000422, -0.00000883, -0.00001252, -0.00001530, -0.00001724, -0.00001847, -0.00001914, -0.00001942, -0.00001947, -0.00001943, -0.00001939, -0.00001938, -0.00001940, -0.00001936, },
		{ -0.00000435, -0.00000750, -0.00001091, -0.00001466, -0.00001879, -0.00002326, -0.00002798, -0.00003275, -0.00003734, -0.00004144, -0.00004472, -0.00004684, -0.00004747, -0.00004636, -0.00004332, -0.00003828, -0.00003129, -0.00002253, -0.00001233, -0.00000111, 0.00001059, 0.00002218, 0.00003303, 0.00004253, 0.00005013, 0.00005537, 0.00005791, 0.00005758, 0.00005438, 0.00004846, 0.00004017, 0.00002996, 0.00001843, 0.00000623, -0.00000594, -0.00001741, -0.00002758, -0.00003591, -0.00004202, -0.00004567, -0.00004679, -0.00004546, -0.00004190, -0.00003647, -0.00002961, -0.00002183, -0.00001366, -0.00000559, 0.00000192, 0.00000850, 0.00001388, 0.00001789, 0.00002048, 0.00002170, 0.00002167, 0.00002059, 0.00001870, 0.00001624, 0.00001342, 0.00001043, 0.00000742, 0.00000444, 0.00000152, -0.00000138, },
		{ 0.00001602, 0.00001744, 0.00001872, 0.00001995, 0.00002118, 0.00002241, 0.00002356, 0.00002447, 0.00002496, 0.00002476, 0.00002361, 0.00002125, 0.00001748, 0.00001215, 0.00000522, -0.00000324, -0.00001303, -0.00002381, -0.00003515, -0.00004649, -0.00005725, -0.00006678, -0.00007448, -0.00007980, -0.00008230, -0.00008168, -0.00007780, -0.00007073, -0.00006071, -0.00004815, -0.00003364, -0.00001787, -0.00000163, 0.00001428, 0.00002906, 0.00004201, 0.00005252, 0.00006016, 0.00006466, 0.00006595, 0.00006415, 0.00005956, 0.00005263, 0.00004393, 0.00003409, 0.00002378, 0.00001365, 0.00000427, -0.00000387, -0.00001041, -0.00001514, -0.00001797, -0.00001896, -0.00001830, -0.00001624, -0.00001313, -0.00000931, -0.00000514, -0.00000094, 0.00000304, 0.00000663, 0.00000971, 0.00001226, 0.00001433, },
		{ -0.00000370, -0.00000267, -0.00000156, -0.00000055, 0.00000022, 0.00000070, 0.00000093, 0.00000101, 0.00000113, 0.00000150, 0.00000237, 0.00000399, 0.00000654, 0.00001016, 0.00001488, 0.00002063, 0.00002722, 0.00003435, 0.00004162, 0.00004856, 0.00005466, 0.00005941, 0.00006234, 0.00006306, 0.00006129, 0.00005691, 0.00004995, 0.00004061, 0.00002927, 0.00001644, 0.00000274, -0.00001111, -0.00002438, -0.00003637, -0.00004642, -0.00005401, -0.00005876, -0.00006047, -0.00005910, -0.00005485, -0.00004805, -0.00003918, -0.00002887, -0.00001776, -0.00000655, 0.00000411, 0.00001362, 0.00002152, 0.00002748, 0.00003130, 0.00003297, 0.00003262, 0.00003050, 0.00002696, 0.00002244, 0.00001739, 0.00001224, 0.00000739, 0.00000316, -0.00000022, -0.00000264, -0.00000408, -0.00000463, -0.00000443, },
		{ 0.00000811, 0.00000531, 0.00000309, 0.00000162, 0.00000101, 0.00000125, 0.00000223, 0.00000375, 0.00000554, 0.00000728, 0.00000862, 0.00000923, 0.00000884, 0.00000725, 0.00000435, 0.00000018, -0.00000512, -0.00001129, -0.00001799, -0.00002475, -0.00003112, -0.00003661, -0.00004076, -0.00004321, -0.00004368, -0.00004202, -0.00003825, -0.00003250, -0.00002507, -0.00001636, -0.00000688, 0.00000283, 0.00001218, 0.00002063, 0.00002771, 0.00003304, 0.00003638, 0.00003763, 0.00003683, 0.00003418, 0.00003000, 0.00002470, 0.00001876, 0.00001269, 0.00000697, 0.00000205, -0.00000175, -0.00000417, -0.00000512, -0.00000460, -0.00000277, 0.00000013, 0.00000377, 0.00000778, 0.00001175, 0.00001532, 0.00001818, 0.00002009, 0.00002092, 0.00002063, 0.00001930, 0.00001712, 0.00001432, 0.00001121, },
		{ 0.00001176, 0.00001228, 0.00001161, 0.00000974, 0.00000679, 0.00000296, -0.00000148, -0.00000615, -0.00001065, -0.00001458, -0.00001751, -0.00001910, -0.00001908, -0.00001729, -0.00001367, -0.00000833, -0.00000149, 0.00000651, 0.00001521, 0.00002409, 0.00003261, 0.00004018, 0.00004631, 0.00005054, 0.00005254, 0.00005211, 0.00004917, 0.00004383, 0.00003630, 0.00002696, 0.00001625, 0.00000469, -0.00000716, -0.00001875, -0.00002957, -0.00003918, -0.00004726, -0.00005358, -0.00005803, -0.00006063, -0.00006152, -0.00006089, -0.00005903, -0.00005624, -0.00005286, -0.00004916, -0.00004540, -0.00004178, -0.00003842, -0.00003535, -0.00003256, -0.00002996, -0.00002745, -0.00002487, -0.00002211, -0.00001907, -0.00001568, -0.00001195, -0.00000794, -0.00000379, 0.00000032, 0.00000418, 0.00000752, 0.00001012, },
		{ -0.00003467, -0.00002905, -0.00002284, -0.00001624, -0.00000951, -0.00000293, 0.00000316, 0.00000845, 0.00001260, 0.00001531, 0.00001635, 0.00001555, 0.00001282, 0.00000820, 0.00000182, -0.00000608, -0.00001512, -0.00002488, -0.00003483, -0.00004443, -0.00005311, -0.00006033, -0.00006563, -0.00006860, -0.00006897, -0.00006660, -0.00006148, -0.00005377, -0.00004375, -0.00003184, -0.00001853, -0.00000440, 0.00000991, 0.00002381, 0.00003669, 0.00004803, 0.00005740, 0.00006448, 0.00006905, 0.00007104, 0.00007050, 0.00006759, 0.00006254, 0.00005569, 0.00004741, 0.00003810, 0.00002815, 0.00001794, 0.00000782, -0.00000194, -0.00001108, -0.00001942, -0.00002685, -0.00003327, -0.00003865, -0.00004296, -0.00004619, -0.00004836, -0.00004947, -0.00004953, -0.00004854, -0.00004652, -0.00004350, -0.00003952, },
		{ -0.00000876, -0.00001533, -0.00002112, -0.00002598, -0.00002981, -0.00003251, -0.00003401, -0.00003425, -0.00003320, -0.00003086, -0.00002724, -0.00002240, -0.00001643, -0.00000947, -0.00000170, 0.00000663, 0.00001526, 0.00002385, 0.00003207, 0.00003953, 0.00004589, 0.00005079, 0.00005394, 0.00005508, 0.00005406, 0.00005080, 0.00004534, 0.00003782, 0.00002849, 0.00001770, 0.00000589, -0.00000644, -0.00001875, -0.00003048, -0.00004110, -0.00005011, -0.00005709, -0.00006173, -0.00006381, -0.00006324, -0.00006008, -0.00005450, -0.00004677, -0.00003728, -0.00002649, -0.00001490, -0.00000305, 0.00000856, 0.00001943, 0.00002914, 0.00003733, 0.00004374, 0.00004819, 0.00005062, 0.00005105, 0.00004959, 0.00004641, 0.00004174, 0.00003585, 0.00002904, 0.00002161, 0.00001385, 0.00000604, -0.00000158, },
		{ 0.00003101, 0.00003230, 0.00003228, 0.00003107, 0.00002884, 0.00002580, 0.00002217, 0.00001816, 0.00001393, 0.00000963, 0.00000535, 0.00000111, -0.00000309, -0.00000727, -0.00001149, -0.00001580, -0.00002022, -0.00002475, -0.00002934, -0.00003387, -0.00003819, -0.00004209, -0.00004531, -0.00004759, -0.00004867, -0.00004831, -0.00004633, -0.00004263, -0.00003718, -0.00003009, -0.00002154, -0.00001184, -0.00000139, 0.00000934, 0.00001984, 0.00002958, 0.00003807, 0.00004484, 0.00004955, 0.00005194, 0.00005190, 0.00004946, 0.00004479, 0.00003821, 0.00003012, 0.00002103, 0.00001151, 0.00000213, -0.00000655, -0.00001404, -0.00001994, -0.00002393, -0.00002586, -0.00002571, -0.00002356, -0.00001966, -0.00001432, -0.00000794, -0.00000096, 0.00000616, 0.00001300, 0.00001917, 0.00002436, 0.00002835, },
		{ -0.00002022, -0.00001945, -0.00001760, -0.00001500, -0.00001201, -0.00000902, -0.00000636, -0.00000430, -0.00000303, -0.00000260, -0.00000293, -0.00000385, -0.00000506, -0.00000621, -0.00000692, -0.00000682, -0.00000558, -0.00000299, 0.00000106, 0.00000652, 0.00001318, 0.00002071, 0.00002861, 0.00003633, 0.00004323, 0.00004870, 0.00005215, 0.00005311, 0.00005125, 0.00004641, 0.00003865, 0.00002821, 0.00001556, 0.00000130, -0.00001381, -0.00002895, -0.00004325, -0.00005592, -0.00006623, -0.00007362, -0.00007771, -0.00007834, -0.00007556, -0.00006966, -0.00006112, -0.00005056, -0.00003875, -0.00002647, -0.00001451, -0.00000362, 0.00000560, 0.00001267, 0.00001733, 0.00001948, 0.00001922, 0.00001684, 0.00001276, 0.00000748, 0.00000159, -0.00000436, -0.00000983, -0.00001439, -0.00001774, -0.00001969, },
		{ 0.00003063, 0.00002944, 0.00002691, 0.00002356, 0.00001996, 0.00001666, 0.00001411, 0.00001264, 0.00001240, 0.00001334, 0.00001523, 0.00001766, 0.00002008, 0.00002189, 0.00002244, 0.00002116, 0.00001760, 0.00001147, 0.00000271, -0.00000848, -0.00002167, -0.00003619, -0.00005120, -0.00006571, -0.00007868, -0.00008909, -0.00009605, -0.00009883, -0.00009695, -0.00009026, -0.00007889, -0.00006332, -0.00004432, -0.00002291, -0.00000029, 0.00002224, 0.00004337, 0.00006188, 0.00007674, 0.00008714, 0.00009259, 0.00009294, 0.00008837, 0.00007939, 0.00006681, 0.00005162, 0.00003497, 0.00001806, 0.00000204, -0.00001205, -0.00002338, -0.00003137, -0.00003571, -0.00003640, -0.00003370, -0.00002812, -0.00002036, -0.00001124, -0.00000160, 0.00000772, 0.00001601, 0.00002270, 0.00002743, 0.00003005, },
		{ -0.00002255, -0.00001504, -0.00000634, 0.00000264, 0.00001101, 0.00001799, 0.00002296, 0.00002558, 0.00002576, 0.00002374, 0.00002002, 0.00001532, 0.00001054, 0.00000664, 0.00000456, 0.00000513, 0.00000895, 0.00001634, 0.00002730, 0.00004145, 0.00005810, 0.00007623, 0.00009461, 0.00011185, 0.00012655, 0.00013739, 0.00014324, 0.00014326, 0.00013701, 0.00012444, 0.00010597, 0.00008242, 0.00005499, 0.00002517, -0.00000536, -0.00003487, -0.00006166, -0.00008425, -0.00010144, -0.00011239, -0.00011671, -0.00011447, -0.00010617, -0.00009273, -0.00007538, -0.00005555, -0.00003481, -0.00001469, 0.00000341, 0.00001833, 0.00002921, 0.00003557, 0.00003735, 0.00003484, 0.00002867, 0.00001976, 0.00000919, -0.00000186, -0.00001228, -0.00002104, -0.00002737, -0.00003073, -0.00003092, -0.00002806, },
		{ -0.00000660, -0.00001824, -0.00002956, -0.00003935, -0.00004657, -0.00005041, -0.00005039, -0.00004642, -0.00003884, -0.00002836, -0.00001604, -0.00000318, 0.00000877, 0.00001838, 0.00002434, 0.00002560, 0.00002148, 0.00001176, -0.00000333, -0.00002302, -0.00004613, -0.00007110, -0.00009611, -0.00011923, -0.00013856, -0.00015241, -0.00015940, -0.00015860, -0.00014961, -0.00013260, -0.00010832, -0.00007803, -0.00004344, -0.00000655, 0.00003046, 0.00006540, 0.00009626, 0.00012132, 0.00013929, 0.00014939, 0.00015141, 0.00014570, 0.00013314, 0.00011503, 0.00009301, 0.00006893, 0.00004463, 0.00002187, 0.00000216, -0.00001336, -0.00002397, -0.00002942, -0.00002995, -0.00002621, -0.00001919, -0.00001014, -0.00000041, 0.00000867, 0.00001591, 0.00002036, 0.00002142, 0.00001889, 0.00001295, 0.00000416, },
		{ 0.00000299, 0.00001379, 0.00002240, 0.00002763, 0.00002855, 0.00002467, 0.00001594, 0.00000282, -0.00001375, -0.00003240, -0.00005146, -0.00006907, -0.00008337, -0.00009266, -0.00009554, -0.00009110, -0.00007896, -0.00005937, -0.00003325, -0.00000206, 0.00003219, 0.00006716, 0.00010032, 0.00012912, 0.00015122, 0.00016467, 0.00016805, 0.00016059, 0.00014226, 0.00011379, 0.00007658, 0.00003266, -0.00001550, -0.00006515, -0.00011343, -0.00015762, -0.00019532, -0.00022461, -0.00024416, -0.00025335, -0.00025228, -0.00024170, -0.00022298, -0.00019796, -0.00016879, -0.00013775, -0.00010705, -0.00007870, -0.00005433, -0.00003507, -0.00002153, -0.00001375, -0.00001122, -0.00001303, -0.00001790, -0.00002438, -0.00003100, -0.00003638, -0.00003941, -0.00003936, -0.00003591, -0.00002919, -0.00001979, -0.00000866, },
		{ -0.00010553, -0.00010906, -0.00010915, -0.00010479, -0.00009537, -0.00008080, -0.00006157, -0.00003871, -0.00001375, 0.00001140, 0.00003462, 0.00005374, 0.00006676, 0.00007203, 0.00006839, 0.00005535, 0.00003317, 0.00000282, -0.00003399, -0.00007496, -0.00011733, -0.00015806, -0.00019413, -0.00022267, -0.00024125, -0.00024808, -0.00024213, -0.00022321, -0.00019207, -0.00015027, -0.00010017, -0.00004466, 0.00001293, 0.00006918, 0.00012073, 0.00016458, 0.00019828, 0.00022008, 0.00022908, 0.00022521, 0.00020929, 0.00018286, 0.00014811, 0.00010763, 0.00006427, 0.00002083, -0.00002007, -0.00005623, -0.00008600, -0.00010836, -0.00012300, -0.00013025, -0.00013101, -0.00012664, -0.00011880, -0.00010926, -0.00009971, -0.00009160, -0.00008603, -0.00008359, -0.00008439, -0.00008800, -0.00009357, -0.00009988, },
		{ -0.00000991, -0.00000709, -0.00000856, -0.00001519, -0.00002727, -0.00004448, -0.00006582, -0.00008972, -0.00011412, -0.00013668, -0.00015493, -0.00016655, -0.00016955, -0.00016250, -0.00014466, -0.00011611, -0.00007781, -0.00003151, 0.00002027, 0.00007447, 0.00012766, 0.00017633, 0.00021710, 0.00024705, 0.00026388, 0.00026615, 0.00025338, 0.00022609, 0.00018581, 0.00013494, 0.00007661, 0.00001441, -0.00004780, -0.00010624, -0.00015744, -0.00019849, -0.00022727, -0.00024254, -0.00024405, -0.00023249, -0.00020943, -0.00017719, -0.00013858, -0.00009672, -0.00005475, -0.00001562, 0.00001817, 0.00004472, 0.00006286, 0.00007219, 0.00007308, 0.00006656, 0.00005423, 0.00003805, 0.00002012, 0.00000251, -0.00001297, -0.00002493, -0.00003252, -0.00003547, -0.00003415, -0.00002947, -0.00002277, -0.00001567, },
		{ -0.00014163, -0.00014303, -0.00013679, -0.00012260, -0.00010094, -0.00007301, -0.00004075, -0.00000666, 0.00002637, 0.00005530, 0.00007719, 0.00008947, 0.00009018, 0.00007815, 0.00005316, 0.00001601, -0.00003152, -0.00008673, -0.00014622, -0.00020607, -0.00026214, -0.00031042, -0.00034724, -0.00036963, -0.00037551, -0.00036386, -0.00033483, -0.00028970, -0.00023083, -0.00016146, -0.00008551, -0.00000726, 0.00006891, 0.00013887, 0.00019900, 0.00024646, 0.00027933, 0.00029673, 0.00029886, 0.00028690, 0.00026288, 0.00022950, 0.00018986, 0.00014723, 0.00010473, 0.00006513, 0.00003063, 0.00000274, -0.00001785, -0.00003120, -0.00003809, -0.00003989, -0.00003834, -0.00003540, -0.00003297, -0.00003271, -0.00003584, -0.00004303, -0.00005432, -0.00006910, -0.00008617, -0.00010391, -0.00012039, -0.00013358, },
		{ 0.00004404, 0.00000348, -0.00004468, -0.00009921, -0.00015810, -0.00021864, -0.00027759, -0.00033137, -0.00037630, -0.00040897, -0.00042643, -0.00042654, -0.00040810, -0.00037111, -0.00031676, -0.00024741, -0.00016656, -0.00007855, 0.00001165, 0.00009878, 0.00017764, 0.00024346, 0.00029223, 0.00032098, 0.00032797, 0.00031287, 0.00027672, 0.00022189, 0.00015192, 0.00007122, -0.00001517, -0.00010201, -0.00018416, -0.00025694, -0.00031644, -0.00035976, -0.00038514, -0.00039206, -0.00038117, -0.00035420, -0.00031378, -0.00026318, -0.00020602, -0.00014597, -0.00008653, -0.00003069, 0.00001915, 0.00006138, 0.00009519, 0.00012051, 0.00013798, 0.00014870, 0.00015411, 0.00015575, 0.00015504, 0.00015310, 0.00015063, 0.00014780, 0.00014425, 0.00013913, 0.00013123, 0.00011909, 0.00010129, 0.00007655, },
		{ 0.00035687, 0.00039759, 0.00043694, 0.00047280, 0.00050251, 0.00052312, 0.00053150, 0.00052465, 0.00049997, 0.00045554, 0.00039033, 0.00030446, 0.00019926, 0.00007738, -0.00005733, -0.00019992, -0.00034459, -0.00048504, -0.00061479, -0.00072763, -0.00081797, -0.00088125, -0.00091423, -0.00091522, -0.00088423, -0.00082297, -0.00073475, -0.00062431, -0.00049749, -0.00036086, -0.00022134, -0.00008575, 0.00003963, 0.00014937, 0.00023925, 0.00030644, 0.00034962, 0.00036897, 0.00036606, 0.00034368, 0.00030555, 0.00025603, 0.00019972, 0.00014121, 0.00008470, 0.00003375, -0.00000884, -0.00004122, -0.00006244, -0.00007245, -0.00007193, -0.00006214, -0.00004470, -0.00002137, 0.00000614, 0.00003636, 0.00006818, 0.00010087, 0.00013416, 0.00016811, 0.00020300, 0.00023920, 0.00027698, 0.00031635, },
		{ 0.00001717, 0.00003577, 0.00005981, 0.00009062, 0.00012975, 0.00017874, 0.00023898, 0.00031147, 0.00039660, 0.00049394, 0.00060210, 0.00071865, 0.00084010, 0.00096198, 0.00107902, 0.00118539, 0.00127504, 0.00134207, 0.00138111, 0.00138771, 0.00135868, 0.00129242, 0.00118906, 0.00105057, 0.00088074, 0.00068503, 0.00047032, 0.00024451, 0.00001616, -0.00020605, -0.00041376, -0.00059944, -0.00075673, -0.00088086, -0.00096883, -0.00101952, -0.00103369, -0.00101386, -0.00096406, -0.00088952, -0.00079628, -0.00069076, -0.00057938, -0.00046813, -0.00036228, -0.00026609, -0.00018268, -0.00011392, -0.00006049, -0.00002199, 0.00000291, 0.00001618, 0.00002025, 0.00001769, 0.00001102, 0.00000248, -0.00000606, -0.00001321, -0.00001805, -0.00002005, -0.00001899, -0.00001485, -0.00000758, 0.00000296, },
		{ -0.00024641, -0.00024790, -0.00025248, -0.00026047, -0.00027228, -0.00028832, -0.00030884, -0.00033373, -0.00036240, -0.00039362, -0.00042547, -0.00045533, -0.00047991, -0.00049542, -0.00049777, -0.00048285, -0.00044681, -0.00038643, -0.00029944, -0.00018475, -0.00004276, 0.00012458, 0.00031368, 0.00051941, 0.00073526, 0.00095368, 0.00116637, 0.00136481, 0.00154068, 0.00168637, 0.00179542, 0.00186291, 0.00188575, 0.00186291, 0.00179542, 0.00168637, 0.00154068, 0.00136481, 0.00116637, 0.00095368, 0.00073526, 0.00051941, 0.00031368, 0.00012458, -0.00004276, -0.00018475, -0.00029944, -0.00038643, -0.00044681, -0.00048285, -0.00049777, -0.00049542, -0.00047991, -0.00045533, -0.00042547, -0.00039362, -0.00036240, -0.00033373, -0.00030884, -0.00028832, -0.00027228, -0.00026047, -0.00025248, -0.00024790, },
		{ 0.00001717, 0.00000296, -0.00000758, -0.00001485, -0.00001899, -0.00002005, -0.00001805, -0.00001321, -0.00000606, 0.00000248, 0.00001102, 0.00001769, 0.00002025, 0.00001618, 0.00000291, -0.00002199, -0.00006049, -0.00011392, -0.00018268, -0.00026609, -0.00036228, -0.00046813, -0.00057938, -0.00069076, -0.00079628, -0.00088952, -0.00096406, -0.00101386, -0.00103369, -0.00101952, -0.00096883, -0.00088086, -0.00075673, -0.00059944, -0.00041376, -0.00020605, 0.00001616, 0.00024451, 0.00047032, 0.00068503, 0.00088074, 0.00105057, 0.00118906, 0.00129242, 0.00135868, 0.00138771, 0.00138111, 0.00134207, 0.00127504, 0.00118539, 0.00107902, 0.00096198, 0.00084010, 0.00071865, 0.00060210, 0.00049394, 0.00039660, 0.00031147, 0.00023898, 0.00017874, 0.00012975, 0.00009062, 0.00005981, 0.00003577, },
		{ 0.00035687, 0.00031635, 0.00027698, 0.00023920, 0.00020300, 0.00016811, 0.00013416, 0.00010087, 0.00006818, 0.00003636, 0.00000614, -0.00002137, -0.00004470, -0.00006214, -0.00007193, -0.00007245, -0.00006244, -0.00004122, -0.00000884, 0.00003375, 0.00008470, 0.00014121, 0.00019972, 0.00025603, 0.00030555, 0.00034368, 0.00036606, 0.00036897, 0.00034962, 0.00030644, 0.00023925, 0.00014937, 0.00003963, -0.00008575, -0.00022134, -0.00036086, -0.00049749, -0.00062431, -0.00073475, -0.00082297, -0.00088423, -0.00091522, -0.00091423, -0.00088125, -0.00081797, -0.00072763, -0.00061479, -0.00048504, -0.00034459, -0.00019992, -0.00005733, 0.00007738, 0.00019926, 0.00030446, 0.00039033, 0.00045554, 0.00049997, 0.00052465, 0.00053150, 0.00052312, 0.00050251, 0.00047280, 0.00043694, 0.00039759, },
		{ 0.00004404, 0.00007655, 0.00010129, 0.00011909, 0.00013123, 0.00013913, 0.00014425, 0.00014780, 0.00015063, 0.00015310, 0.00015504, 0.00015575, 0.00015411, 0.00014870, 0.00013798, 0.00012051, 0.00009519, 0.00006138, 0.00001915, -0.00003069, -0.00008653, -0.00014597, -0.00020602, -0.00026318, -0.00031378, -0.00035420, -0.00038117, -0.00039206, -0.00038514, -0.00035976, -0.00031644, -0.00025694, -0.00018416, -0.00010201, -0.00001517, 0.00007122, 0.00015192, 0.00022189, 0.00027672, 0.00031287, 0.00032797, 0.00032098, 0.00029223, 0.00024346, 0.00017764, 0.00009878, 0.00001165, -0.00007855, -0.00016656, -0.00024741, -0.00031676, -0.00037111, -0.00040810, -0.00042654, -0.00042643, -0.00040897, -0.00037630, -0.00033137, -0.00027759, -0.00021864, -0.00015810, -0.00009921, -0.00004468, 0.00000348, },
		{ -0.00014163, -0.00013358, -0.00012039, -0.00010391, -0.00008617, -0.00006910, -0.00005432, -0.00004303, -0.00003584, -0.00003271, -0.00003297, -0.00003540, -0.00003834, -0.00003989, -0.00003809, -0.00003120, -0.00001785, 0.00000274, 0.00003063, 0.00006513, 0.00010473, 0.00014723, 0.00018986, 0.00022950, 0.00026288, 0.00028690, 0.00029886, 0.00029673, 0.00027933, 0.00024646, 0.00019900, 0.00013887, 0.00006891, -0.00000726, -0.00008551, -0.00016146, -0.00023083, -0.00028970, -0.00033483, -0.00036386, -0.00037551, -0.00036963, -0.00034724, -0.00031042, -0.00026214, -0.00020607, -0.00014622, -0.00008673, -0.00003152, 0.00001601, 0.00005316, 0.00007815, 0.00009018, 0.00008947, 0.00007719, 0.00005530, 0.00002637, -0.00000666, -0.00004075, -0.00007301, -0.00010094, -0.00012260, -0.00013679, -0.00014303, },
		{ -0.00000991, -0.00001567, -0.00002277, -0.00002947, -0.00003415, -0.00003547, -0.00003252, -0.00002493, -0.00001297, 0.00000251, 0.00002012, 0.00003805, 0.00005423, 0.00006656, 0.00007308, 0.00007219, 0.00006286, 0.00004472, 0.00001817, -0.00001562, -0.00005475, -0.00009672, -0.00013858, -0.00017719, -0.00020943, -0.00023249, -0.00024405, -0.00024254, -0.00022727, -0.00019849, -0.00015744, -0.00010624, -0.00004780, 0.00001441, 0.00007661, 0.00013494, 0.00018581, 0.00022609, 0.00025338, 0.00026615, 0.00026388, 0.00024705, 0.00021710, 0.00017633, 0.00012766, 0.00007447, 0.00002027, -0.00003151, -0.00007781, -0.00011611, -0.00014466, -0.00016250, -0.00016955, -0.00016655, -0.00015493, -0.00013668, -0.00011412, -0.00008972, -0.00006582, -0.00004448, -0.00002727, -0.00001519, -0.00000856, -0.00000709, },
		{ -0.00010553, -0.00009988, -0.00009357, -0.00008800, -0.00008439, -0.00008359, -0.00008603, -0.00009160, -0.00009971, -0.00010926, -0.00011880, -0.00012664, -0.00013101, -0.00013025, -0.00012300, -0.00010836, -0.00008600, -0.00005623, -0.00002007, 0.00002083, 0.00006427, 0.00010763, 0.00014811, 0.00018286, 0.00020929, 0.00022521, 0.00022908, 0.00022008, 0.00019828, 0.00016458, 0.00012073, 0.00006918, 0.00001293, -0.00004466, -0.00010017, -0.00015027, -0.00019207, -0.00022321, -0.00024213, -0.00024808, -0.00024125, -0.00022267, -0.00019413, -0.00015806, -0.00011733, -0.00007496, -0.00003399, 0.00000282, 0.00003317, 0.00005535, 0.00006839, 0.00007203, 0.00006676, 0.00005374, 0.00003462, 0.00001140, -0.00001375, -0.00003871, -0.00006157, -0.00008080, -0.00009537, -0.00010479, -0.00010915, -0.00010906, },
		{ 0.00000299, -0.00000866, -0.00001979, -0.00002919, -0.00003591, -0.00003936, -0.00003941, -0.00003638, -0.00003100, -0.00002438, -0.00001790, -0.00001303, -0.00001122, -0.00001375, -0.00002153, -0.00003507, -0.00005433, -0.00007870, -0.00010705, -0.00013775, -0.00016879, -0.00019796, -0.00022298, -0.00024170, -0.00025228, -0.00025335, -0.00024416, -0.00022461, -0.00019532, -0.00015762, -0.00011343, -0.00006515, -0.00001550, 0.00003266, 0.00007658, 0.00011379, 0.00014226, 0.00016059, 0.00016805, 0.00016467, 0.00015122, 0.00012912, 0.00010032, 0.00006716, 0.00003219, -0.00000206, -0.00003325, -0.00005937, -0.00007896, -0.00009110, -0.00009554, -0.00009266, -0.00008337, -0.00006907, -0.00005146, -0.00003240, -0.00001375, 0.00000282, 0.00001594, 0.00002467, 0.00002855, 0.00002763, 0.00002240, 0.00001379, },
		{ -0.00000660, 0.00000416, 0.00001295, 0.00001889, 0.00002142, 0.00002036, 0.00001591, 0.00000867, -0.00000041, -0.00001014, -0.00001919, -0.00002621, -0.00002995, -0.00002942, -0.00002397, -0.00001336, 0.00000216, 0.00002187, 0.00004463, 0.00006893, 0.00009301, 0.00011503, 0.00013314, 0.00014570, 0.00015141, 0.00014939, 0.00013929, 0.00012132, 0.00009626, 0.00006540, 0.00003046, -0.00000655, -0.00004344, -0.00007803, -0.00010832, -0.00013260, -0.00014961, -0.00015860, -0.00015940, -0.00015241, -0.00013856, -0.00011923, -0.00009611, -0.00007110, -0.00004613, -0.00002302, -0.00000333, 0.00001176, 0.00002148, 0.00002560, 0.00002434, 0.00001838, 0.00000877, -0.00000318, -0.00001604, -0.00002836, -0.00003884, -0.00004642, -0.00005039, -0.00005041, -0.00004657, -0.00003935, -0.00002956, -0.00001824, },
		{ -0.00002255, -0.00002806, -0.00003092, -0.00003073, -0.00002737, -0.00002104, -0.00001228, -0.00000186, 0.00000919, 0.00001976, 0.00002867, 0.00003484, 0.00003735, 0.00003557, 0.00002921, 0.00001833, 0.00000341, -0.00001469, -0.00003481, -0.00005555, -0.00007538, -0.00009273, -0.00010617, -0.00011447, -0.00011671, -0.00011239, -0.00010144, -0.00008425, -0.00006166, -0.00003487, -0.00000536, 0.00002517, 0.00005499, 0.00008242, 0.00010597, 0.00012444, 0.00013701, 0.00014326, 0.00014324, 0.00013739, 0.00012655, 0.00011185, 0.00009461, 0.00007623, 0.00005810, 0.00004145, 0.00002730, 0.00001634, 0.00000895, 0.00000513, 0.00000456, 0.00000664, 0.00001054, 0.00001532, 0.00002002, 0.00002374, 0.00002576, 0.00002558, 0.00002296, 0.00001799, 0.00001101, 0.00000264, -0.00000634, -0.00001504, },
		{ 0.00003063, 0.00003005, 0.00002743, 0.00002270, 0.00001601, 0.00000772, -0.00000160, -0.00001124, -0.00002036, -0.00002812, -0.00003370, -0.00003640, -0.00003571, -0.00003137, -0.00002338, -0.00001205, 0.00000204, 0.00001806, 0.00003497, 0.00005162, 0.00006681, 0.00007939, 0.00008837, 0.00009294, 0.00009259, 0.00008714, 0.00007674, 0.00006188, 0.00004337, 0.00002224, -0.00000029, -0.00002291, -0.00004432, -0.00006332, -0.00007889, -0.00009026, -0.00009695, -0.00009883, -0.00009605, -0.00008909, -0.00007868, -0.00006571, -0.00005120, -0.00003619, -0.00002167, -0.00000848, 0.00000271, 0.00001147, 0.00001760, 0.00002116, 0.00002244, 0.00002189, 0.00002008, 0.00001766, 0.00001523, 0.00001334, 0.00001240, 0.00001264, 0.00001411, 0.00001666, 0.00001996, 0.00002356, 0.00002691, 0.00002944, },
		{ -0.00002022, -0.00001969, -0.00001774, -0.00001439, -0.00000983, -0.00000436, 0.00000159, 0.00000748, 0.00001276, 0.00001684, 0.00001922, 0.00001948, 0.00001733, 0.00001267, 0.00000560, -0.00000362, -0.00001451, -0.00002647, -0.00003875, -0.00005056, -0.00006112, -0.00006966, -0.00007556, -0.00007834, -0.00007771, -0.00007362, -0.00006623, -0.00005592, -0.00004325, -0.00002895, -0.00001381, 0.00000130, 0.00001556, 0.00002821, 0.00003865, 0.00004641, 0.00005125, 0.00005311, 0.00005215, 0.00004870, 0.00004323, 0.00003633, 0.00002861, 0.00002071, 0.00001318, 0.00000652, 0.00000106, -0.00000299, -0.00000558, -0.00000682, -0.00000692, -0.00000621, -0.00000506, -0.00000385, -0.00000293, -0.00000260, -0.00000303, -0.00000430, -0.00000636, -0.00000902, -0.00001201, -0.00001500, -0.00001760, -0.00001945, },
		{ 0.00003101, 0.00002835, 0.00002436, 0.00001917, 0.00001300, 0.00000616, -0.00000096, -0.00000794, -0.00001432, -0.00001966, -0.00002356, -0.00002571, -0.00002586, -0.00002393, -0.00001994, -0.00001404, -0.00000655, 0.00000213, 0.00001151, 0.00002103, 0.00003012, 0.00003821, 0.00004479, 0.00004946, 0.00005190, 0.00005194, 0.00004955, 0.00004484, 0.00003807, 0.00002958, 0.00001984, 0.00000934, -0.00000139, -0.00001184, -0.00002154, -0.00003009, -0.00003718, -0.00004263, -0.00004633, -0.00004831, -0.00004867, -0.00004759, -0.00004531, -0.00004209, -0.00003819, -0.00003387, -0.00002934, -0.00002475, -0.00002022, -0.00001580, -0.00001149, -0.00000727, -0.00000309, 0.00000111, 0.00000535, 0.00000963, 0.00001393, 0.00001816, 0.00002217, 0.00002580, 0.00002884, 0.00003107, 0.00003228, 0.00003230, },
		{ -0.00000876, -0.00000158, 0.00000604, 0.00001385, 0.00002161, 0.00002904, 0.00003585, 0.00004174, 0.00004641, 0.00004959, 0.00005105, 0.00005062, 0.00004819, 0.00004374, 0.00003733, 0.00002914, 0.00001943, 0.00000856, -0.00000305, -0.00001490, -0.00002649, -0.00003728, -0.00004677, -0.00005450, -0.00006008, -0.00006324, -0.00006381, -0.00006173, -0.00005709, -0.00005011, -0.00004110, -0.00003048, -0.00001875, -0.00000644, 0.00000589, 0.00001770, 0.00002849, 0.00003782, 0.00004534, 0.00005080, 0.00005406, 0.00005508, 0.00005394, 0.00005079, 0.00004589, 0.00003953, 0.00003207, 0.00002385, 0.00001526, 0.00000663, -0.00000170, -0.00000947, -0.00001643, -0.00002240, -0.00002724, -0.00003086, -0.00003320, -0.00003425, -0.00003401, -0.00003251, -0.00002981, -0.00002598, -0.00002112, -0.00001533, },
		{ -0.00003467, -0.00003952, -0.00004350, -0.00004652, -0.00004854, -0.00004953, -0.00004947, -0.00004836, -0.00004619, -0.00004296, -0.00003865, -0.00003327, -0.00002685, -0.00001942, -0.00001108, -0.00000194, 0.00000782, 0.00001794, 0.00002815, 0.00003810, 0.00004741, 0.00005569, 0.00006254, 0.00006759, 0.00007050, 0.00007104, 0.00006905, 0.00006448, 0.00005740, 0.00004803, 0.00003669, 0.00002381, 0.00000991, -0.00000440, -0.00001853, -0.00003184, -0.00004375, -0.00005377, -0.00006148, -0.00006660, -0.00006897, -0.00006860, -0.00006563, -0.00006033, -0.00005311, -0.00004443, -0.00003483, -0.00002488, -0.00001512, -0.00000608, 0.00000182, 0.00000820, 0.00001282, 0.00001555, 0.00001635, 0.00001531, 0.00001260, 0.00000845, 0.00000316, -0.00000293, -0.00000951, -0.00001624, -0.00002284, -0.00002905, },
		{ 0.00001176, 0.00001012, 0.00000752, 0.00000418, 0.00000032, -0.00000379, -0.00000794, -0.00001195, -0.00001568, -0.00001907, -0.00002211, -0.00002487, -0.00002745, -0.00002996, -0.00003256, -0.00003535, -0.00003842, -0.00004178, -0.00004540, -0.00004916, -0.00005286, -0.00005624, -0.00005903, -0.00006089, -0.00006152, -0.00006063, -0.00005803, -0.00005358, -0.00004726, -0.00003918, -0.00002957, -0.00001875, -0.00000716, 0.00000469, 0.00001625, 0.00002696, 0.00003630, 0.00004383, 0.00004917, 0.00005211, 0.00005254, 0.00005054, 0.00004631, 0.00004018, 0.00003261, 0.00002409, 0.00001521, 0.00000651, -0.00000149, -0.00000833, -0.00001367, -0.00001729, -0.00001908, -0.00001910, -0.00001751, -0.00001458, -0.00001065, -0.00000615, -0.00000148, 0.00000296, 0.00000679, 0.00000974, 0.00001161, 0.00001228, },
		{ 0.00000811, 0.00001121, 0.00001432, 0.00001712, 0.00001930, 0.00002063, 0.00002092, 0.00002009, 0.00001818, 0.00001532, 0.00001175, 0.00000778, 0.00000377, 0.00000013, -0.00000277, -0.00000460, -0.00000512, -0.00000417, -0.00000175, 0.00000205, 0.00000697, 0.00001269, 0.00001876, 0.00002470, 0.00003000, 0.00003418, 0.00003683, 0.00003763, 0.00003638, 0.00003304, 0.00002771, 0.00002063, 0.00001218, 0.00000283, -0.00000688, -0.00001636, -0.00002507, -0.00003250, -0.00003825, -0.00004202, -0.00004368, -0.00004321, -0.00004076, -0.00003661, -0.00003112, -0.00002475, -0.00001799, -0.00001129, -0.00000512, 0.00000018, 0.00000435, 0.00000725, 0.00000884, 0.00000923, 0.00000862, 0.00000728, 0.00000554, 0.00000375, 0.00000223, 0.00000125, 0.00000101, 0.00000162, 0.00000309, 0.00000531, },
		{ -0.00000370, -0.00000443, -0.00000463, -0.00000408, -0.00000264, -0.00000022, 0.00000316, 0.00000739, 0.00001224, 0.00001739, 0.00002244, 0.00002696, 0.00003050, 0.00003262, 0.00003297, 0.00003130, 0.00002748, 0.00002152, 0.00001362, 0.00000411, -0.00000655, -0.00001776, -0.00002887, -0.00003918, -0.00004805, -0.00005485, -0.00005910, -0.00006047, -0.00005876, -0.00005401, -0.00004642, -0.00003637, -0.00002438, -0.00001111, 0.00000274, 0.00001644, 0.00002927, 0.00004061, 0.00004995, 0.00005691, 0.00006129, 0.00006306, 0.00006234, 0.00005941, 0.00005466, 0.00004856, 0.00004162, 0.00003435, 0.00002722, 0.00002063, 0.00001488, 0.00001016, 0.00000654, 0.00000399, 0.00000237, 0.00000150, 0.00000113, 0.00000101, 0.00000093, 0.00000070, 0.00000022, -0.00000055, -0.00000156, -0.00000267, },
		{ 0.00001602, 0.00001433, 0.00001226, 0.00000971, 0.00000663, 0.00000304, -0.00000094, -0.00000514, -0.00000931, -0.00001313, -0.00001624, -0.00001830, -0.00001896, -0.00001797, -0.00001514, -0.00001041, -0.00000387, 0.00000427, 0.00001365, 0.00002378, 0.00003409, 0.00004393, 0.00005263, 0.00005956, 0.00006415, 0.00006595, 0.00006466, 0.00006016, 0.00005252, 0.00004201, 0.00002906, 0.00001428, -0.00000163, -0.00001787, -0.00003364, -0.00004815, -0.00006071, -0.00007073, -0.00007780, -0.00008168, -0.00008230, -0.00007980, -0.00007448, -0.00006678, -0.00005725, -0.00004649, -0.00003515, -0.00002381, -0.00001303, -0.00000324, 0.00000522, 0.00001215, 0.00001748, 0.00002125, 0.00002361, 0.00002476, 0.00002496, 0.00002447, 0.00002356, 0.00002241, 0.00002118, 0.00001995, 0.00001872, 0.00001744, },
		{ -0.00000435, -0.00000138, 0.00000152, 0.00000444, 0.00000742, 0.00001043, 0.00001342, 0.00001624, 0.00001870, 0.00002059, 0.00002167, 0.00002170, 0.00002048, 0.00001789, 0.00001388, 0.00000850, 0.00000192, -0.00000559, -0.00001366, -0.00002183, -0.00002961, -0.00003647, -0.00004190, -0.00004546, -0.00004679, -0.00004567, -0.00004202, -0.00003591, -0.00002758, -0.00001741, -0.00000594, 0.00000623, 0.00001843, 0.00002996, 0.00004017, 0.00004846, 0.00005438, 0.00005758, 0.00005791, 0.00005537, 0.00005013, 0.00004253, 0.00003303, 0.00002218, 0.00001059, -0.00000111, -0.00001233, -0.00002253, -0.00003129, -0.00003828, -0.00004332, -0.00004636, -0.00004747, -0.00004684, -0.00004472, -0.00004144, -0.00003734, -0.00003275, -0.00002798, -0.00002326, -0.00001879, -0.00001466, -0.00001091, -0.00000750, },
		{ -0.00001916, -0.00001936, -0.00001940, -0.00001938, -0.00001939, -0.00001943, -0.00001947, -0.00001942, -0.00001914, -0.00001847, -0.00001724, -0.00001530, -0.00001252, -0.00000883, -0.00000422, 0.00000123, 0.00000735, 0.00001392, 0.00002062, 0.00002710, 0.00003298, 0.00003788, 0.00004146, 0.00004341, 0.00004355, 0.00004177, 0.00003809, 0.00003266, 0.00002572, 0.00001764, 0.00000885, -0.00000015, -0.00000886, -0.00001677, -0.00002345, -0.00002853, -0.00003174, -0.00003294, -0.00003213, -0.00002942, -0.00002505, -0.00001936, -0.00001277, -0.00000574, 0.00000127, 0.00000780, 0.00001347, 0.00001795, 0.00002102, 0.00002256, 0.00002258, 0.00002116, 0.00001850, 0.00001484, 0.00001050, 0.00000578, 0.00000099, -0.00000358, -0.00000772, -0.00001126, -0.00001411, -0.00001626, -0.00001775, -0.00001867, },
		{ 0.00000292, 0.00000224, 0.00000169, 0.00000132, 0.00000117, 0.00000120, 0.00000136, 0.00000156, 0.00000166, 0.00000153, 0.00000103, 0.00000004, -0.00000154, -0.00000375, -0.00000658, -0.00000996, -0.00001374, -0.00001776, -0.00002177, -0.00002551, -0.00002871, -0.00003111, -0.00003248, -0.00003263, -0.00003146, -0.00002894, -0.00002510, -0.00002009, -0.00001412, -0.00000748, -0.00000048, 0.00000651, 0.00001313, 0.00001904, 0.00002396, 0.00002765, 0.00002996, 0.00003082, 0.00003026, 0.00002840, 0.00002542, 0.00002157, 0.00001715, 0.00001246, 0.00000782, 0.00000351, -0.00000023, -0.00000322, -0.00000535, -0.00000658, -0.00000694, -0.00000652, -0.00000547, -0.00000397, -0.00000220, -0.00000038, 0.00000133, 0.00000278, 0.00000387, 0.00000454, 0.00000479, 0.00000466, 0.00000424, 0.00000362, },
		{ -0.00001191, -0.00001243, -0.00001292, -0.00001340, -0.00001385, -0.00001426, -0.00001458, -0.00001474, -0.00001467, -0.00001430, -0.00001356, -0.00001241, -0.00001081, -0.00000877, -0.00000632, -0.00000354, -0.00000054, 0.00000255, 0.00000557, 0.00000835, 0.00001071, 0.00001250, 0.00001358, 0.00001384, 0.00001325, 0.00001177, 0.00000948, 0.00000645, 0.00000284, -0.00000117, -0.00000538, -0.00000954, -0.00001346, -0.00001691, -0.00001971, -0.00002174, -0.00002290, -0.00002316, -0.00002253, -0.00002110, -0.00001898, -0.00001632, -0.00001332, -0.00001016, -0.00000705, -0.00000415, -0.00000162, 0.00000043, 0.00000190, 0.00000278, 0.00000306, 0.00000278, 0.00000204, 0.00000092, -0.00000047, -0.00000202, -0.00000361, -0.00000516, -0.00000660, -0.00000789, -0.00000899, -0.00000992, -0.00001070, -0.00001135, },
		{ 0.00000598, 0.00000523, 0.00000428, 0.00000317, 0.00000193, 0.00000061, -0.00000078, -0.00000221, -0.00000366, -0.00000514, -0.00000663, -0.00000815, -0.00000968, -0.00001124, -0.00001280, -0.00001437, -0.00001591, -0.00001738, -0.00001875, -0.00001997, -0.00002097, -0.00002171, -0.00002212, -0.00002216, -0.00002180, -0.00002101, -0.00001980, -0.00001820, -0.00001624, -0.00001399, -0.00001152, -0.00000894, -0.00000633, -0.00000382, -0.00000149, 0.00000057, 0.00000228, 0.00000360, 0.00000449, 0.00000495, 0.00000501, 0.00000471, 0.00000412, 0.00000332, 0.00000240, 0.00000146, 0.00000057, -0.00000017, -0.00000071, -0.00000099, -0.00000100, -0.00000073, -0.00000021, 0.00000054, 0.00000144, 0.00000245, 0.00000348, 0.00000447, 0.00000535, 0.00000606, 0.00000655, 0.00000680, 0.00000679, 0.00000651, },
		{ 0.00000247, 0.00000385, 0.00000533, 0.00000683, 0.00000830, 0.00000964, 0.00001080, 0.00001170, 0.00001230, 0.00001260, 0.00001258, 0.00001229, 0.00001177, 0.00001110, 0.00001035, 0.00000961, 0.00000897, 0.00000850, 0.00000825, 0.00000827, 0.00000855, 0.00000908, 0.00000980, 0.00001065, 0.00001153, 0.00001235, 0.00001298, 0.00001334, 0.00001334, 0.00001290, 0.00001200, 0.00001062, 0.00000879, 0.00000656, 0.00000403, 0.00000130, -0.00000152, -0.00000428, -0.00000686, -0.00000916, -0.00001107, -0.00001253, -0.00001349, -0.00001395, -0.00001393, -0.00001348, -0.00001267, -0.00001160, -0.00001036, -0.00000904, -0.00000774, -0.00000653, -0.00000547, -0.00000458, -0.00000387, -0.00000332, -0.00000290, -0.00000255, -0.00000221, -0.00000182, -0.00000132, -0.00000066, 0.00000019, 0.00000124, },
		{ -0.00000491, -0.00000565, -0.00000642, -0.00000714, -0.00000772, -0.00000808, -0.00000811, -0.00000777, -0.00000702, -0.00000586, -0.00000433, -0.00000251, -0.00000051, 0.00000153, 0.00000346, 0.00000513, 0.00000638, 0.00000712, 0.00000724, 0.00000671, 0.00000554, 0.00000379, 0.00000158, -0.00000094, -0.00000358, -0.00000614, -0.00000840, -0.00001017, -0.00001127, -0.00001157, -0.00001100, -0.00000952, -0.00000719, -0.00000409, -0.00000038, 0.00000374, 0.00000805, 0.00001232, 0.00001631, 0.00001980, 0.00002262, 0.00002464, 0.00002576, 0.00002597, 0.00002529, 0.00002383, 0.00002169, 0.00001906, 0.00001610, 0.00001301, 0.00000995, 0.00000708, 0.00000452, 0.00000235, 0.00000061, -0.00000071, -0.00000164, -0.00000225, -0.00000264, -0.00000290, -0.00000313, -0.00000340, -0.00000377, -0.00000428, },
		{ 0.00000219, 0.00000217, 0.00000226, 0.00000235, 0.00000232, 0.00000205, 0.00000145, 0.00000047, -0.00000093, -0.00000272, -0.00000481, -0.00000710, -0.00000942, -0.00001159, -0.00001342, -0.00001474, -0.00001537, -0.00001521, -0.00001419, -0.00001231, -0.00000964, -0.00000630, -0.00000247, 0.00000159, 0.00000564, 0.00000939, 0.00001259, 0.00001498, 0.00001639, 0.00001669, 0.00001582, 0.00001382, 0.00001079, 0.00000690, 0.00000239, -0.00000245, -0.00000734, -0.00001196, -0.00001603, -0.00001931, -0.00002163, -0.00002285, -0.00002294, -0.00002194, -0.00001996, -0.00001716, -0.00001375, -0.00000998, -0.00000611, -0.00000236, 0.00000103, 0.00000391, 0.00000615, 0.00000770, 0.00000854, 0.00000875, 0.00000841, 0.00000767, 0.00000667, 0.00000556, 0.00000448, 0.00000355, 0.00000284, 0.00000239, },
		{ -0.00000135, -0.00000223, -0.00000317, -0.00000403, -0.00000465, -0.00000491, -0.00000471, -0.00000402, -0.00000284, -0.00000123, 0.00000067, 0.00000271, 0.00000467, 0.00000636, 0.00000756, 0.00000807, 0.00000777, 0.00000656, 0.00000442, 0.00000141, -0.00000233, -0.00000661, -0.00001118, -0.00001574, -0.00001998, -0.00002360, -0.00002633, -0.00002795, -0.00002831, -0.00002733, -0.00002504, -0.00002153, -0.00001701, -0.00001172, -0.00000598, -0.00000011, 0.00000553, 0.00001063, 0.00001491, 0.00001816, 0.00002022, 0.00002105, 0.00002068, 0.00001920, 0.00001680, 0.00001372, 0.00001021, 0.00000654, 0.00000299, -0.00000021, -0.00000287, -0.00000487, -0.00000615, -0.00000671, -0.00000662, -0.00000600, -0.00000501, -0.00000382, -0.00000261, -0.00000155, -0.00000077, -0.00000035, -0.00000032, -0.00000068, },
		{ 0.00000962, 0.00001018, 0.00001055, 0.00001058, 0.00001016, 0.00000920, 0.00000767, 0.00000558, 0.00000304, 0.00000017, -0.00000284, -0.00000576, -0.00000837, -0.00001042, -0.00001174, -0.00001215, -0.00001156, -0.00000995, -0.00000739, -0.00000401, -0.00000000, 0.00000435, 0.00000875, 0.00001287, 0.00001641, 0.00001907, 0.00002064, 0.00002095, 0.00001993, 0.00001760, 0.00001407, 0.00000954, 0.00000428, -0.00000139, -0.00000713, -0.00001258, -0.00001741, -0.00002134, -0.00002414, -0.00002567, -0.00002589, -0.00002481, -0.00002257, -0.00001935, -0.00001540, -0.00001099, -0.00000642, -0.00000197, 0.00000211, 0.00000562, 0.00000843, 0.00001046, 0.00001171, 0.00001224, 0.00001218, 0.00001167, 0.00001090, 0.00001003, 0.00000923, 0.00000862, 0.00000829, 0.00000827, 0.00000854, 0.00000903, },
		{ 0.00000146, 0.00000182, 0.00000237, 0.00000321, 0.00000441, 0.00000600, 0.00000793, 0.00001013, 0.00001245, 0.00001472, 0.00001674, 0.00001829, 0.00001919, 0.00001925, 0.00001838, 0.00001650, 0.00001363, 0.00000986, 0.00000536, 0.00000034, -0.00000491, -0.00001008, -0.00001487, -0.00001896, -0.00002210, -0.00002407, -0.00002475, -0.00002408, -0.00002211, -0.00001897, -0.00001486, -0.00001007, -0.00000492, 0.00000026, 0.00000512, 0.00000936, 0.00001271, 0.00001497, 0.00001603, 0.00001587, 0.00001453, 0.00001217, 0.00000898, 0.00000522, 0.00000117, -0.00000288, -0.00000666, -0.00000993, -0.00001251, -0.00001428, -0.00001519, -0.00001525, -0.00001456, -0.00001323, -0.00001144, -0.00000937, -0.00000722, -0.00000516, -0.00000331, -0.00000177, -0.00000059, 0.00000025, 0.00000081, 0.00000117, },
		{ -0.00000565, -0.00000493, -0.00000447, -0.00000433, -0.00000454, -0.00000508, -0.00000588, -0.00000681, -0.00000775, -0.00000851, -0.00000893, -0.00000887, -0.00000818, -0.00000680, -0.00000470, -0.00000193, 0.00000142, 0.00000517, 0.00000912, 0.00001302, 0.00001660, 0.00001960, 0.00002178, 0.00002295, 0.00002297, 0.00002178, 0.00001940, 0.00001592, 0.00001152, 0.00000642, 0.00000092, -0.00000467, -0.00001004, -0.00001488, -0.00001894, -0.00002199, -0.00002391, -0.00002463, -0.00002419, -0.00002267, -0.00002025, -0.00001715, -0.00001363, -0.00000997, -0.00000642, -0.00000323, -0.00000061, 0.00000132, 0.00000248, 0.00000285, 0.00000248, 0.00000150, 0.00000006, -0.00000168, -0.00000351, -0.00000526, -0.00000677, -0.00000792, -0.00000863, -0.00000889, -0.00000871, -0.00000819, -0.00000741, -0.00000652, },
		{ -0.00000291, -0.00000278, -0.00000228, -0.00000139, -0.00000017, 0.00000131, 0.00000291, 0.00000450, 0.00000592, 0.00000701, 0.00000765, 0.00000773, 0.00000719, 0.00000605, 0.00000434, 0.00000218, -0.00000028, -0.00000284, -0.00000530, -0.00000744, -0.00000906, -0.00000998, -0.00001008, -0.00000930, -0.00000764, -0.00000515, -0.00000198, 0.00000169, 0.00000563, 0.00000958, 0.00001327, 0.00001644, 0.00001889, 0.00002043, 0.00002095, 0.00002041, 0.00001883, 0.00001632, 0.00001303, 0.00000919, 0.00000502, 0.00000079, -0.00000324, -0.00000685, -0.00000985, -0.00001210, -0.00001352, -0.00001410, -0.00001389, -0.00001300, -0.00001156, -0.00000977, -0.00000780, -0.00000584, -0.00000405, -0.00000258, -0.00000150, -0.00000085, -0.00000062, -0.00000076, -0.00000116, -0.00000171, -0.00000227, -0.00000271, },
		},
		{ { -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, -0.00000390, },
		{ 0.00000081, 0.00000081, 0.00000083, 0.00000086, 0.00000091, 0.00000096, 0.00000102, 0.00000110, 0.00000118, 0.00000128, 0.00000138, 0.00000149, 0.00000160, 0.00000172, 0.00000184, 0.00000196, 0.00000209, 0.00000221, 0.00000234, 0.00000246, 0.00000258, 0.00000269, 0.00000280, 0.00000290, 0.00000299, 0.00000308, 0.00000315, 0.00000322, 0.00000327, 0.00000331, 0.00000334, 0.00000336, 0.00000337, 0.00000336, 0.00000334, 0.00000331, 0.00000327, 0.00000322, 0.00000315, 0.00000308, 0.00000299, 0.00000290, 0.00000280, 0.00000269, 0.00000258, 0.00000246, 0.00000234, 0.00000221, 0.00000209, 0.00000196, 0.00000184, 0.00000172, 0.00000160, 0.00000149, 0.00000138, 0.00000128, 0.00000118, 0.00000110, 0.00000102, 0.00000096, 0.00000091, 0.00000086, 0.00000083, 0.00000081, },
		{ 0.00000883, 0.00000882, 0.00000880, 0.00000877, 0.00000872, 0.00000865, 0.00000858, 0.00000849, 0.00000839, 0.00000828, 0.00000816, 0.00000804, 0.00000790, 0.00000776, 0.00000762, 0.00000748, 0.00000733, 0.00000718, 0.00000704, 0.00000689, 0.00000675, 0.00000662, 0.00000649, 0.00000637, 0.00000627, 0.00000617, 0.00000608, 0.00000600, 0.00000594, 0.00000589, 0.00000585, 0.00000583, 0.00000583, 0.00000583, 0.00000585, 0.00000589, 0.00000594, 0.00000600, 0.00000608, 0.00000617, 0.00000627, 0.00000637, 0.00000649, 0.00000662, 0.00000675, 0.00000689, 0.00000704, 0.00000718, 0.00000733, 0.00000748, 0.00000762, 0.00000776, 0.00000790, 0.00000804, 0.00000816, 0.00000828, 0.00000839, 0.00000849, 0.00000858, 0.00000865, 0.00000872, 0.00000877, 0.00000880, 0.00000882, },
		{ 0.00000110, 0.00000110, 0.00000109, 0.00000107, 0.00000105, 0.00000102, 0.00000099, 0.00000095, 0.00000091, 0.00000086, 0.00000081, 0.00000075, 0.00000069, 0.00000063, 0.00000057, 0.00000050, 0.00000044, 0.00000037, 0.00000031, 0.00000025, 0.00000018, 0.00000013, 0.00000007, 0.00000002, -0.00000003, -0.00000007, -0.00000011, -0.00000015, -0.00000017, -0.00000020, -0.00000021, -0.00000022, -0.00000023, -0.00000022, -0.00000021, -0.00000020, -0.00000017, -0.00000015, -0.00000011, -0.00000007, -0.00000003, 0.00000002, 0.00000007, 0.00000013, 0.00000018, 0.00000025, 0.00000031, 0.00000037, 0.00000044, 0.00000050, 0.00000057, 0.00000063, 0.00000069, 0.00000075, 0.00000081, 0.00000086, 0.00000091, 0.00000095, 0.00000099, 0.00000102, 0.00000105, 0.00000107, 0.00000109, 0.00000110, },
		{ -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, },
		{ -0.00000205, -0.00000204, -0.00000202, -0.00000200, -0.00000196, -0.00000191, -0.00000186, -0.00000179, -0.00000172, -0.00000164, -0.00000155, -0.00000145, -0.00000135, -0.00000125, -0.00000114, -0.00000104, -0.00000093, -0.00000082, -0.00000071, -0.00000060, -0.00000050, -0.00000040, -0.00000030, -0.00000022, -0.00000013, -0.00000006, 0.00000000, 0.00000006, 0.00000011, 0.00000015, 0.00000017, 0.00000019, 0.00000019, 0.00000019, 0.00000017, 0.00000015, 0.00000011, 0.00000006, 0.00000000, -0.00000006, -0.00000013, -0.00000022, -0.00000030, -0.00000040, -0.00000050, -0.00000060, -0.00000071, -0.00000082, -0.00000093, -0.00000104, -0.00000114, -0.00000125, -0.00000135, -0.00000145, -0.00000155, -0.00000164, -0.00000172, -0.00000179, -0.00000186, -0.00000191, -0.00000196, -0.00000200, -0.00000202, -0.00000204, },
		{ -0.00000600, -0.00000598, -0.00000591, -0.00000580, -0.00000565, -0.00000546, -0.00000522, -0.00000495, -0.00000465, -0.00000431, -0.00000395, -0.00000356, -0.00000314, -0.00000272, -0.00000228, -0.00000183, -0.00000137, -0.00000092, -0.00000047, -0.00000003, 0.00000040, 0.00000081, 0.00000120, 0.00000156, 0.00000190, 0.00000220, 0.00000247, 0.00000271, 0.00000290, 0.00000306, 0.00000317, 0.00000323, 0.00000326, 0.00000323, 0.00000317, 0.00000306, 0.00000290, 0.00000271, 0.00000247, 0.00000220, 0.00000190, 0.00000156, 0.00000120, 0.00000081, 0.00000040, -0.00000003, -0.00000047, -0.00000092, -0.00000137, -0.00000183, -0.00000228, -0.00000272, -0.00000314, -0.00000356, -0.00000395, -0.00000431, -0.00000465, -0.00000495, -0.00000522, -0.00000546, -0.00000565, -0.00000580, -0.00000591, -0.00000598, },
		{ -0.00000139, -0.00000144, -0.00000157, -0.00000179, -0.00000210, -0.00000249, -0.00000296, -0.00000351, -0.00000412, -0.00000480, -0.00000553, -0.00000632, -0.00000715, -0.00000801, -0.00000889, -0.00000980, -0.00001071, -0.00001163, -0.00001253, -0.00001342, -0.00001428, -0.00001511, -0.00001589, -0.00001663, -0.00001730, -0.00001792, -0.00001846, -0.00001893, -0.00001932, -0.00001963, -0.00001985, -0.00001999, -0.00002003, -0.00001999, -0.00001985, -0.00001963, -0.00001932, -0.00001893, -0.00001846, -0.00001792, -0.00001730, -0.00001663, -0.00001589, -0.00001511, -0.00001428, -0.00001342, -0.00001253, -0.00001163, -0.00001071, -0.00000980, -0.00000889, -0.00000801, -0.00000715, -0.00000632, -0.00000553, -0.00000480, -0.00000412, -0.00000351, -0.00000296, -0.00000249, -0.00000210, -0.00000179, -0.00000157, -0.00000144, },
		{ -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, },
		{ -0.00000692, -0.00000686, -0.00000669, -0.00000639, -0.00000599, -0.00000547, -0.00000485, -0.00000413, -0.00000332, -0.00000242, -0.00000145, -0.00000041, 0.00000068, 0.00000182, 0.00000299, 0.00000419, 0.00000540, 0.00000660, 0.00000780, 0.00000897, 0.00001011, 0.00001120, 0.00001224, 0.00001321, 0.00001411, 0.00001492, 0.00001564, 0.00001626, 0.00001678, 0.00001719, 0.00001748, 0.00001766, 0.00001772, 0.00001766, 0.00001748, 0.00001719, 0.00001678, 0.00001626, 0.00001564, 0.00001492, 0.00001411, 0.00001321, 0.00001224, 0.00001120, 0.00001011, 0.00000897, 0.00000780, 0.00000660, 0.00000540, 0.00000419, 0.00000299, 0.00000182, 0.00000068, -0.00000041, -0.00000145, -0.00000242, -0.00000332, -0.00000413, -0.00000485, -0.00000547, -0.00000599, -0.00000639, -0.00000669, -0.00000686, },
		{ 0.00001850, 0.00001846, 0.00001835, 0.00001815, 0.00001788, 0.00001754, 0.00001713, 0.00001665, 0.00001611, 0.00001552, 0.00001488, 0.00001419, 0.00001347, 0.00001271, 0.00001194, 0.00001115, 0.00001035, 0.00000955, 0.00000876, 0.00000798, 0.00000723, 0.00000650, 0.00000582, 0.00000517, 0.00000458, 0.00000404, 0.00000357, 0.00000315, 0.00000281, 0.00000254, 0.00000235, 0.00000223, 0.00000219, 0.00000223, 0.00000235, 0.00000254, 0.00000281, 0.00000315, 0.00000357, 0.00000404, 0.00000458, 0.00000517, 0.00000582, 0.00000650, 0.00000723, 0.00000798, 0.00000876, 0.00000955, 0.00001035, 0.00001115, 0.00001194, 0.00001271, 0.00001347, 0.00001419, 0.00001488, 0.00001552, 0.00001611, 0.00001665, 0.00001713, 0.00001754, 0.00001788, 0.00001815, 0.00001835, 0.00001846, },
		{ 0.00000345, 0.00000344, 0.00000340, 0.00000334, 0.00000325, 0.00000313, 0.00000300, 0.00000284, 0.00000267, 0.00000247, 0.00000226, 0.00000203, 0.00000180, 0.00000155, 0.00000129, 0.00000103, 0.00000077, 0.00000051, 0.00000025, -0.00000001, -0.00000025, -0.00000049, -0.00000072, -0.00000093, -0.00000112, -0.00000130, -0.00000146, -0.00000159, -0.00000170, -0.00000179, -0.00000186, -0.00000189, -0.00000191, -0.00000189, -0.00000186, -0.00000179, -0.00000170, -0.00000159, -0.00000146, -0.00000130, -0.00000112, -0.00000093, -0.00000072, -0.00000049, -0.00000025, -0.00000001, 0.00000025, 0.00000051, 0.00000077, 0.00000103, 0.00000129, 0.00000155, 0.00000180, 0.00000203, 0.00000226, 0.00000247, 0.00000267, 0.00000284, 0.00000300, 0.00000313, 0.00000325, 0.00000334, 0.00000340, 0.00000344, },
		{ -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, },
		{ 0.00000275, 0.00000277, 0.00000281, 0.00000289, 0.00000300, 0.00000314, 0.00000331, 0.00000350, 0.00000372, 0.00000396, 0.00000422, 0.00000450, 0.00000480, 0.00000510, 0.00000542, 0.00000574, 0.00000606, 0.00000639, 0.00000671, 0.00000703, 0.00000733, 0.00000763, 0.00000791, 0.00000817, 0.00000841, 0.00000863, 0.00000882, 0.00000899, 0.00000913, 0.00000924, 0.00000932, 0.00000936, 0.00000938, 0.00000936, 0.00000932, 0.00000924, 0.00000913, 0.00000899, 0.00000882, 0.00000863, 0.00000841, 0.00000817, 0.00000791, 0.00000763, 0.00000733, 0.00000703, 0.00000671, 0.00000639, 0.00000606, 0.00000574, 0.00000542, 0.00000510, 0.00000480, 0.00000450, 0.00000422, 0.00000396, 0.00000372, 0.00000350, 0.00000331, 0.00000314, 0.00000300, 0.00000289, 0.00000281, 0.00000277, },
		{ -0.00001439, -0.00001433, -0.00001415, -0.00001385, -0.00001344, -0.00001291, -0.00001228, -0.00001155, -0.00001072, -0.00000981, -0.00000883, -0.00000777, -0.00000666, -0.00000550, -0.00000431, -0.00000310, -0.00000187, -0.00000064, 0.00000057, 0.00000177, 0.00000292, 0.00000403, 0.00000509, 0.00000608, 0.00000699, 0.00000781, 0.00000854, 0.00000917, 0.00000970, 0.00001011, 0.00001041, 0.00001059, 0.00001065, 0.00001059, 0.00001041, 0.00001011, 0.00000970, 0.00000917, 0.00000854, 0.00000781, 0.00000699, 0.00000608, 0.00000509, 0.00000403, 0.00000292, 0.00000177, 0.00000057, -0.00000064, -0.00000187, -0.00000310, -0.00000431, -0.00000550, -0.00000666, -0.00000777, -0.00000883, -0.00000981, -0.00001072, -0.00001155, -0.00001228, -0.00001291, -0.00001344, -0.00001385, -0.00001415, -0.00001433, },
		{ 0.00000545, 0.00000534, 0.00000500, 0.00000443, 0.00000365, 0.00000266, 0.00000147, 0.00000009, -0.00000146, -0.00000318, -0.00000504, -0.00000703, -0.00000912, -0.00001131, -0.00001355, -0.00001584, -0.00001816, -0.00002047, -0.00002276, -0.00002501, -0.00002719, -0.00002929, -0.00003128, -0.00003314, -0.00003485, -0.00003641, -0.00003779, -0.00003898, -0.00003997, -0.00004075, -0.00004131, -0.00004165, -0.00004177, -0.00004165, -0.00004131, -0.00004075, -0.00003997, -0.00003898, -0.00003779, -0.00003641, -0.00003485, -0.00003314, -0.00003128, -0.00002929, -0.00002719, -0.00002501, -0.00002276, -0.00002047, -0.00001816, -0.00001584, -0.00001355, -0.00001131, -0.00000912, -0.00000703, -0.00000504, -0.00000318, -0.00000146, 0.00000009, 0.00000147, 0.00000266, 0.00000365, 0.00000443, 0.00000500, 0.00000534, },
		{ -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, },
		{ -0.00001777, -0.00001764, -0.00001722, -0.00001654, -0.00001559, -0.00001438, -0.00001293, -0.00001125, -0.00000936, -0.00000727, -0.00000500, -0.00000258, -0.00000003, 0.00000262, 0.00000536, 0.00000815, 0.00001097, 0.00001378, 0.00001657, 0.00001931, 0.00002196, 0.00002451, 0.00002693, 0.00002920, 0.00003129, 0.00003318, 0.00003486, 0.00003631, 0.00003752, 0.00003847, 0.00003915, 0.00003957, 0.00003971, 0.00003957, 0.00003915, 0.00003847, 0.00003752, 0.00003631, 0.00003486, 0.00003318, 0.00003129, 0.00002920, 0.00002693, 0.00002451, 0.00002196, 0.00001931, 0.00001657, 0.00001378, 0.00001097, 0.00000815, 0.00000536, 0.00000262, -0.00000003, -0.00000258, -0.00000500, -0.00000727, -0.00000936, -0.00001125, -0.00001293, -0.00001438, -0.00001559, -0.00001654, -0.00001722, -0.00001764, },
		{ 0.00002848, 0.00002839, 0.00002812, 0.00002768, 0.00002706, 0.00002628, 0.00002534, 0.00002426, 0.00002303, 0.00002168, 0.00002021, 0.00001865, 0.00001700, 0.00001528, 0.00001351, 0.00001171, 0.00000989, 0.00000806, 0.00000626, 0.00000449, 0.00000277, 0.00000112, -0.00000044, -0.00000191, -0.00000326, -0.00000449, -0.00000557, -0.00000651, -0.00000729, -0.00000791, -0.00000835, -0.00000862, -0.00000871, -0.00000862, -0.00000835, -0.00000791, -0.00000729, -0.00000651, -0.00000557, -0.00000449, -0.00000326, -0.00000191, -0.00000044, 0.00000112, 0.00000277, 0.00000449, 0.00000626, 0.00000806, 0.00000989, 0.00001171, 0.00001351, 0.00001528, 0.00001700, 0.00001865, 0.00002021, 0.00002168, 0.00002303, 0.00002426, 0.00002534, 0.00002628, 0.00002706, 0.00002768, 0.00002812, 0.00002839, },
		{ -0.00000240, -0.00000243, -0.00000251, -0.00000266, -0.00000286, -0.00000311, -0.00000341, -0.00000377, -0.00000416, -0.00000460, -0.00000508, -0.00000558, -0.00000612, -0.00000667, -0.00000725, -0.00000783, -0.00000842, -0.00000901, -0.00000960, -0.00001017, -0.00001073, -0.00001126, -0.00001177, -0.00001225, -0.00001268, -0.00001308, -0.00001343, -0.00001374, -0.00001399, -0.00001419, -0.00001433, -0.00001442, -0.00001445, -0.00001442, -0.00001433, -0.00001419, -0.00001399, -0.00001374, -0.00001343, -0.00001308, -0.00001268, -0.00001225, -0.00001177, -0.00001126, -0.00001073, -0.00001017, -0.00000960, -0.00000901, -0.00000842, -0.00000783, -0.00000725, -0.00000667, -0.00000612, -0.00000558, -0.00000508, -0.00000460, -0.00000416, -0.00000377, -0.00000341, -0.00000311, -0.00000286, -0.00000266, -0.00000251, -0.00000243, },
		{ -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, },
		{ 0.00000157, 0.00000160, 0.00000171, 0.00000189, 0.00000214, 0.00000245, 0.00000283, 0.00000326, 0.00000375, 0.00000429, 0.00000488, 0.00000551, 0.00000617, 0.00000686, 0.00000757, 0.00000829, 0.00000903, 0.00000976, 0.00001048, 0.00001119, 0.00001188, 0.00001254, 0.00001317, 0.00001376, 0.00001430, 0.00001479, 0.00001522, 0.00001560, 0.00001591, 0.00001616, 0.00001634, 0.00001645, 0.00001648, 0.00001645, 0.00001634, 0.00001616, 0.00001591, 0.00001560, 0.00001522, 0.00001479, 0.00001430, 0.00001376, 0.00001317, 0.00001254, 0.00001188, 0.00001119, 0.00001048, 0.00000976, 0.00000903, 0.00000829, 0.00000757, 0.00000686, 0.00000617, 0.00000551, 0.00000488, 0.00000429, 0.00000375, 0.00000326, 0.00000283, 0.00000245, 0.00000214, 0.00000189, 0.00000171, 0.00000160, },
		{ -0.00003726, -0.00003713, -0.00003672, -0.00003603, -0.00003509, -0.00003389, -0.00003245, -0.00003078, -0.00002890, -0.00002683, -0.00002458, -0.00002217, -0.00001964, -0.00001700, -0.00001429, -0.00001151, -0.00000872, -0.00000592, -0.00000315, -0.00000043, 0.00000221, 0.00000474, 0.00000714, 0.00000939, 0.00001147, 0.00001335, 0.00001502, 0.00001646, 0.00001766, 0.00001860, 0.00001928, 0.00001969, 0.00001983, 0.00001969, 0.00001928, 0.00001860, 0.00001766, 0.00001646, 0.00001502, 0.00001335, 0.00001147, 0.00000939, 0.00000714, 0.00000474, 0.00000221, -0.00000043, -0.00000315, -0.00000592, -0.00000872, -0.00001151, -0.00001429, -0.00001700, -0.00001964, -0.00002217, -0.00002458, -0.00002683, -0.00002890, -0.00003078, -0.00003245, -0.00003389, -0.00003509, -0.00003603, -0.00003672, -0.00003713, },
		{ 0.00002067, 0.00002040, 0.00001961, 0.00001829, 0.00001647, 0.00001416, 0.00001138, 0.00000816, 0.00000453, 0.00000053, -0.00000381, -0.00000845, -0.00001333, -0.00001842, -0.00002367, -0.00002901, -0.00003441, -0.00003981, -0.00004515, -0.00005040, -0.00005549, -0.00006037, -0.00006501, -0.00006935, -0.00007335, -0.00007698, -0.00008020, -0.00008298, -0.00008529, -0.00008711, -0.00008843, -0.00008922, -0.00008948, -0.00008922, -0.00008843, -0.00008711, -0.00008529, -0.00008298, -0.00008020, -0.00007698, -0.00007335, -0.00006935, -0.00006501, -0.00006037, -0.00005549, -0.00005040, -0.00004515, -0.00003981, -0.00003441, -0.00002901, -0.00002367, -0.00001842, -0.00001333, -0.00000845, -0.00000381, 0.00000053, 0.00000453, 0.00000816, 0.00001138, 0.00001416, 0.00001647, 0.00001829, 0.00001961, 0.00002040, },
		{ -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, },
		{ -0.00003924, -0.00003889, -0.00003784, -0.00003611, -0.00003370, -0.00003065, -0.00002697, -0.00002272, -0.00001792, -0.00001263, -0.00000689, -0.00000076, 0.00000570, 0.00001242, 0.00001935, 0.00002642, 0.00003356, 0.00004069, 0.00004776, 0.00005469, 0.00006142, 0.00006787, 0.00007400, 0.00007974, 0.00008503, 0.00008983, 0.00009409, 0.00009776, 0.00010082, 0.00010322, 0.00010496, 0.00010601, 0.00010636, 0.00010601, 0.00010496, 0.00010322, 0.00010082, 0.00009776, 0.00009409, 0.00008983, 0.00008503, 0.00007974, 0.00007400, 0.00006787, 0.00006142, 0.00005469, 0.00004776, 0.00004069, 0.00003356, 0.00002642, 0.00001935, 0.00001242, 0.00000570, -0.00000076, -0.00000689, -0.00001263, -0.00001792, -0.00002272, -0.00002697, -0.00003065, -0.00003370, -0.00003611, -0.00003784, -0.00003889, },
		{ 0.00006285, 0.00006260, 0.00006188, 0.00006068, 0.00005902, 0.00005691, 0.00005437, 0.00005143, 0.00004811, 0.00004446, 0.00004049, 0.00003626, 0.00003180, 0.00002715, 0.00002236, 0.00001748, 0.00001255, 0.00000762, 0.00000273, -0.00000206, -0.00000670, -0.00001117, -0.00001540, -0.00001937, -0.00002302, -0.00002634, -0.00002928, -0.00003182, -0.00003393, -0.00003559, -0.00003679, -0.00003751, -0.00003776, -0.00003751, -0.00003679, -0.00003559, -0.00003393, -0.00003182, -0.00002928, -0.00002634, -0.00002302, -0.00001937, -0.00001540, -0.00001117, -0.00000670, -0.00000206, 0.00000273, 0.00000762, 0.00001255, 0.00001748, 0.00002236, 0.00002715, 0.00003180, 0.00003626, 0.00004049, 0.00004446, 0.00004811, 0.00005143, 0.00005437, 0.00005691, 0.00005902, 0.00006068, 0.00006188, 0.00006260, },
		{ -0.00002999, -0.00003008, -0.00003033, -0.00003076, -0.00003135, -0.00003210, -0.00003300, -0.00003404, -0.00003521, -0.00003651, -0.00003792, -0.00003942, -0.00004100, -0.00004265, -0.00004435, -0.00004608, -0.00004783, -0.00004958, -0.00005131, -0.00005301, -0.00005466, -0.00005624, -0.00005774, -0.00005915, -0.00006045, -0.00006162, -0.00006267, -0.00006357, -0.00006432, -0.00006490, -0.00006533, -0.00006559, -0.00006567, -0.00006559, -0.00006533, -0.00006490, -0.00006432, -0.00006357, -0.00006267, -0.00006162, -0.00006045, -0.00005915, -0.00005774, -0.00005624, -0.00005466, -0.00005301, -0.00005131, -0.00004958, -0.00004783, -0.00004608, -0.00004435, -0.00004265, -0.00004100, -0.00003942, -0.00003792, -0.00003651, -0.00003521, -0.00003404, -0.00003300, -0.00003210, -0.00003135, -0.00003076, -0.00003033, -0.00003008, },
		{ -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, },
		{ -0.00005694, -0.00005679, -0.00005636, -0.00005564, -0.00005465, -0.00005338, -0.00005186, -0.00005010, -0.00004812, -0.00004592, -0.00004355, -0.00004101, -0.00003834, -0.00003556, -0.00003269, -0.00002976, -0.00002681, -0.00002386, -0.00002093, -0.00001807, -0.00001528, -0.00001261, -0.00001007, -0.00000770, -0.00000551, -0.00000352, -0.00000176, -0.00000024, 0.00000102, 0.00000202, 0.00000274, 0.00000317, 0.00000332, 0.00000317, 0.00000274, 0.00000202, 0.00000102, -0.00000024, -0.00000176, -0.00000352, -0.00000551, -0.00000770, -0.00001007, -0.00001261, -0.00001528, -0.00001807, -0.00002093, -0.00002386, -0.00002681, -0.00002976, -0.00003269, -0.00003556, -0.00003834, -0.00004101, -0.00004355, -0.00004592, -0.00004812, -0.00005010, -0.00005186, -0.00005338, -0.00005465, -0.00005564, -0.00005636, -0.00005679, },
		{ -0.00030133, -0.00030058, -0.00029835, -0.00029466, -0.00028953, -0.00028303, -0.00027522, -0.00026616, -0.00025595, -0.00024469, -0.00023247, -0.00021943, -0.00020569, -0.00019138, -0.00017663, -0.00016159, -0.00014640, -0.00013122, -0.00011618, -0.00010143, -0.00008712, -0.00007337, -0.00006033, -0.00004812, -0.00003685, -0.00002664, -0.00001759, -0.00000977, -0.00000327, 0.00000185, 0.00000555, 0.00000778, 0.00000852, 0.00000778, 0.00000555, 0.00000185, -0.00000327, -0.00000977, -0.00001759, -0.00002664, -0.00003685, -0.00004812, -0.00006033, -0.00007337, -0.00008712, -0.00010143, -0.00011618, -0.00013122, -0.00014640, -0.00016159, -0.00017663, -0.00019138, -0.00020569, -0.00021943, -0.00023247, -0.00024469, -0.00025595, -0.00026616, -0.00027522, -0.00028303, -0.00028953, -0.00029466, -0.00029835, -0.00030058, },
		{ 0.00012854, 0.00012599, 0.00011835, 0.00010571, 0.00008818, 0.00006593, 0.00003918, 0.00000819, -0.00002676, -0.00006531, -0.00010711, -0.00015174, -0.00019878, -0.00024777, -0.00029824, -0.00034971, -0.00040169, -0.00045366, -0.00050513, -0.00055560, -0.00060459, -0.00065163, -0.00069626, -0.00073806, -0.00077661, -0.00081156, -0.00084255, -0.00086930, -0.00089155, -0.00090908, -0.00092172, -0.00092936, -0.00093191, -0.00092936, -0.00092172, -0.00090908, -0.00089155, -0.00086930, -0.00084255, -0.00081156, -0.00077661, -0.00073806, -0.00069626, -0.00065163, -0.00060459, -0.00055560, -0.00050513, -0.00045366, -0.00040169, -0.00034971, -0.00029824, -0.00024777, -0.00019878, -0.00015174, -0.00010711, -0.00006531, -0.00002676, 0.00000819, 0.00003918, 0.00006593, 0.00008818, 0.00010571, 0.00011835, 0.00012599, },
		{ 0.00067929, 0.00068282, 0.00069336, 0.00071083, 0.00073504, 0.00076577, 0.00080273, 0.00084554, 0.00089381, 0.00094707, 0.00100480, 0.00106645, 0.00113143, 0.00119910, 0.00126882, 0.00133992, 0.00141171, 0.00148350, 0.00155460, 0.00162432, 0.00169200, 0.00175697, 0.00181863, 0.00187636, 0.00192961, 0.00197788, 0.00202070, 0.00205765, 0.00208838, 0.00211260, 0.00213006, 0.00214061, 0.00214414, 0.00214061, 0.00213006, 0.00211260, 0.00208838, 0.00205765, 0.00202070, 0.00197788, 0.00192961, 0.00187636, 0.00181863, 0.00175697, 0.00169200, 0.00162432, 0.00155460, 0.00148350, 0.00141171, 0.00133992, 0.00126882, 0.00119910, 0.00113143, 0.00106645, 0.00100480, 0.00094707, 0.00089381, 0.00084554, 0.00080273, 0.00076577, 0.00073504, 0.00071083, 0.00069336, 0.00068282, },
		{ 0.00012854, 0.00012599, 0.00011835, 0.00010571, 0.00008818, 0.00006593, 0.00003918, 0.00000819, -0.00002676, -0.00006531, -0.00010711, -0.00015174, -0.00019878, -0.00024777, -0.00029824, -0.00034971, -0.00040169, -0.00045366, -0.00050513, -0.00055560, -0.00060459, -0.00065163, -0.00069626, -0.00073806, -0.00077661, -0.00081156, -0.00084255, -0.00086930, -0.00089155, -0.00090908, -0.00092172, -0.00092936, -0.00093191, -0.00092936, -0.00092172, -0.00090908, -0.00089155, -0.00086930, -0.00084255, -0.00081156, -0.00077661, -0.00073806, -0.00069626, -0.00065163, -0.00060459, -0.00055560, -0.00050513, -0.00045366, -0.00040169, -0.00034971, -0.00029824, -0.00024777, -0.00019878, -0.00015174, -0.00010711, -0.00006531, -0.00002676, 0.00000819, 0.00003918, 0.00006593, 0.00008818, 0.00010571, 0.00011835, 0.00012599, },
		{ -0.00030133, -0.00030058, -0.00029835, -0.00029466, -0.00028953, -0.00028303, -0.00027522, -0.00026616, -0.00025595, -0.00024469, -0.00023247, -0.00021943, -0.00020569, -0.00019138, -0.00017663, -0.00016159, -0.00014640, -0.00013122, -0.00011618, -0.00010143, -0.00008712, -0.00007337, -0.00006033, -0.00004812, -0.00003685, -0.00002664, -0.00001759, -0.00000977, -0.00000327, 0.00000185, 0.00000555, 0.00000778, 0.00000852, 0.00000778, 0.00000555, 0.00000185, -0.00000327, -0.00000977, -0.00001759, -0.00002664, -0.00003685, -0.00004812, -0.00006033, -0.00007337, -0.00008712, -0.00010143, -0.00011618, -0.00013122, -0.00014640, -0.00016159, -0.00017663, -0.00019138, -0.00020569, -0.00021943, -0.00023247, -0.00024469, -0.00025595, -0.00026616, -0.00027522, -0.00028303, -0.00028953, -0.00029466, -0.00029835, -0.00030058, },
		{ -0.00005694, -0.00005679, -0.00005636, -0.00005564, -0.00005465, -0.00005338, -0.00005186, -0.00005010, -0.00004812, -0.00004592, -0.00004355, -0.00004101, -0.00003834, -0.00003556, -0.00003269, -0.00002976, -0.00002681, -0.00002386, -0.00002093, -0.00001807, -0.00001528, -0.00001261, -0.00001007, -0.00000770, -0.00000551, -0.00000352, -0.00000176, -0.00000024, 0.00000102, 0.00000202, 0.00000274, 0.00000317, 0.00000332, 0.00000317, 0.00000274, 0.00000202, 0.00000102, -0.00000024, -0.00000176, -0.00000352, -0.00000551, -0.00000770, -0.00001007, -0.00001261, -0.00001528, -0.00001807, -0.00002093, -0.00002386, -0.00002681, -0.00002976, -0.00003269, -0.00003556, -0.00003834, -0.00004101, -0.00004355, -0.00004592, -0.00004812, -0.00005010, -0.00005186, -0.00005338, -0.00005465, -0.00005564, -0.00005636, -0.00005679, },
		{ -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, -0.00004811, },
		{ -0.00002999, -0.00003008, -0.00003033, -0.00003076, -0.00003135, -0.00003210, -0.00003300, -0.00003404, -0.00003521, -0.00003651, -0.00003792, -0.00003942, -0.00004100, -0.00004265, -0.00004435, -0.00004608, -0.00004783, -0.00004958, -0.00005131, -0.00005301, -0.00005466, -0.00005624, -0.00005774, -0.00005915, -0.00006045, -0.00006162, -0.00006267, -0.00006357, -0.00006432, -0.00006490, -0.00006533, -0.00006559, -0.00006567, -0.00006559, -0.00006533, -0.00006490, -0.00006432, -0.00006357, -0.00006267, -0.00006162, -0.00006045, -0.00005915, -0.00005774, -0.00005624, -0.00005466, -0.00005301, -0.00005131, -0.00004958, -0.00004783, -0.00004608, -0.00004435, -0.00004265, -0.00004100, -0.00003942, -0.00003792, -0.00003651, -0.00003521, -0.00003404, -0.00003300, -0.00003210, -0.00003135, -0.00003076, -0.00003033, -0.00003008, },
		{ 0.00006285, 0.00006260, 0.00006188, 0.00006068, 0.00005902, 0.00005691, 0.00005437, 0.00005143, 0.00004811, 0.00004446, 0.00004049, 0.00003626, 0.00003180, 0.00002715, 0.00002236, 0.00001748, 0.00001255, 0.00000762, 0.00000273, -0.00000206, -0.00000670, -0.00001117, -0.00001540, -0.00001937, -0.00002302, -0.00002634, -0.00002928, -0.00003182, -0.00003393, -0.00003559, -0.00003679, -0.00003751, -0.00003776, -0.00003751, -0.00003679, -0.00003559, -0.00003393, -0.00003182, -0.00002928, -0.00002634, -0.00002302, -0.00001937, -0.00001540, -0.00001117, -0.00000670, -0.00000206, 0.00000273, 0.00000762, 0.00001255, 0.00001748, 0.00002236, 0.00002715, 0.00003180, 0.00003626, 0.00004049, 0.00004446, 0.00004811, 0.00005143, 0.00005437, 0.00005691, 0.00005902, 0.00006068, 0.00006188, 0.00006260, },
		{ -0.00003924, -0.00003889, -0.00003784, -0.00003611, -0.00003370, -0.00003065, -0.00002697, -0.00002272, -0.00001792, -0.00001263, -0.00000689, -0.00000076, 0.00000570, 0.00001242, 0.00001935, 0.00002642, 0.00003356, 0.00004069, 0.00004776, 0.00005469, 0.00006142, 0.00006787, 0.00007400, 0.00007974, 0.00008503, 0.00008983, 0.00009409, 0.00009776, 0.00010082, 0.00010322, 0.00010496, 0.00010601, 0.00010636, 0.00010601, 0.00010496, 0.00010322, 0.00010082, 0.00009776, 0.00009409, 0.00008983, 0.00008503, 0.00007974, 0.00007400, 0.00006787, 0.00006142, 0.00005469, 0.00004776, 0.00004069, 0.00003356, 0.00002642, 0.00001935, 0.00001242, 0.00000570, -0.00000076, -0.00000689, -0.00001263, -0.00001792, -0.00002272, -0.00002697, -0.00003065, -0.00003370, -0.00003611, -0.00003784, -0.00003889, },
		{ -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, -0.00000991, },
		{ 0.00002067, 0.00002040, 0.00001961, 0.00001829, 0.00001647, 0.00001416, 0.00001138, 0.00000816, 0.00000453, 0.00000053, -0.00000381, -0.00000845, -0.00001333, -0.00001842, -0.00002367, -0.00002901, -0.00003441, -0.00003981, -0.00004515, -0.00005040, -0.00005549, -0.00006037, -0.00006501, -0.00006935, -0.00007335, -0.00007698, -0.00008020, -0.00008298, -0.00008529, -0.00008711, -0.00008843, -0.00008922, -0.00008948, -0.00008922, -0.00008843, -0.00008711, -0.00008529, -0.00008298, -0.00008020, -0.00007698, -0.00007335, -0.00006935, -0.00006501, -0.00006037, -0.00005549, -0.00005040, -0.00004515, -0.00003981, -0.00003441, -0.00002901, -0.00002367, -0.00001842, -0.00001333, -0.00000845, -0.00000381, 0.00000053, 0.00000453, 0.00000816, 0.00001138, 0.00001416, 0.00001647, 0.00001829, 0.00001961, 0.00002040, },
		{ -0.00003726, -0.00003713, -0.00003672, -0.00003603, -0.00003509, -0.00003389, -0.00003245, -0.00003078, -0.00002890, -0.00002683, -0.00002458, -0.00002217, -0.00001964, -0.00001700, -0.00001429, -0.00001151, -0.00000872, -0.00000592, -0.00000315, -0.00000043, 0.00000221, 0.00000474, 0.00000714, 0.00000939, 0.00001147, 0.00001335, 0.00001502, 0.00001646, 0.00001766, 0.00001860, 0.00001928, 0.00001969, 0.00001983, 0.00001969, 0.00001928, 0.00001860, 0.00001766, 0.00001646, 0.00001502, 0.00001335, 0.00001147, 0.00000939, 0.00000714, 0.00000474, 0.00000221, -0.00000043, -0.00000315, -0.00000592, -0.00000872, -0.00001151, -0.00001429, -0.00001700, -0.00001964, -0.00002217, -0.00002458, -0.00002683, -0.00002890, -0.00003078, -0.00003245, -0.00003389, -0.00003509, -0.00003603, -0.00003672, -0.00003713, },
		{ 0.00000157, 0.00000160, 0.00000171, 0.00000189, 0.00000214, 0.00000245, 0.00000283, 0.00000326, 0.00000375, 0.00000429, 0.00000488, 0.00000551, 0.00000617, 0.00000686, 0.00000757, 0.00000829, 0.00000903, 0.00000976, 0.00001048, 0.00001119, 0.00001188, 0.00001254, 0.00001317, 0.00001376, 0.00001430, 0.00001479, 0.00001522, 0.00001560, 0.00001591, 0.00001616, 0.00001634, 0.00001645, 0.00001648, 0.00001645, 0.00001634, 0.00001616, 0.00001591, 0.00001560, 0.00001522, 0.00001479, 0.00001430, 0.00001376, 0.00001317, 0.00001254, 0.00001188, 0.00001119, 0.00001048, 0.00000976, 0.00000903, 0.00000829, 0.00000757, 0.00000686, 0.00000617, 0.00000551, 0.00000488, 0.00000429, 0.00000375, 0.00000326, 0.00000283, 0.00000245, 0.00000214, 0.00000189, 0.00000171, 0.00000160, },
		{ -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, -0.00000494, },
		{ -0.00000240, -0.00000243, -0.00000251, -0.00000266, -0.00000286, -0.00000311, -0.00000341, -0.00000377, -0.00000416, -0.00000460, -0.00000508, -0.00000558, -0.00000612, -0.00000667, -0.00000725, -0.00000783, -0.00000842, -0.00000901, -0.00000960, -0.00001017, -0.00001073, -0.00001126, -0.00001177, -0.00001225, -0.00001268, -0.00001308, -0.00001343, -0.00001374, -0.00001399, -0.00001419, -0.00001433, -0.00001442, -0.00001445, -0.00001442, -0.00001433, -0.00001419, -0.00001399, -0.00001374, -0.00001343, -0.00001308, -0.00001268, -0.00001225, -0.00001177, -0.00001126, -0.00001073, -0.00001017, -0.00000960, -0.00000901, -0.00000842, -0.00000783, -0.00000725, -0.00000667, -0.00000612, -0.00000558, -0.00000508, -0.00000460, -0.00000416, -0.00000377, -0.00000341, -0.00000311, -0.00000286, -0.00000266, -0.00000251, -0.00000243, },
		{ 0.00002848, 0.00002839, 0.00002812, 0.00002768, 0.00002706, 0.00002628, 0.00002534, 0.00002426, 0.00002303, 0.00002168, 0.00002021, 0.00001865, 0.00001700, 0.00001528, 0.00001351, 0.00001171, 0.00000989, 0.00000806, 0.00000626, 0.00000449, 0.00000277, 0.00000112, -0.00000044, -0.00000191, -0.00000326, -0.00000449, -0.00000557, -0.00000651, -0.00000729, -0.00000791, -0.00000835, -0.00000862, -0.00000871, -0.00000862, -0.00000835, -0.00000791, -0.00000729, -0.00000651, -0.00000557, -0.00000449, -0.00000326, -0.00000191, -0.00000044, 0.00000112, 0.00000277, 0.00000449, 0.00000626, 0.00000806, 0.00000989, 0.00001171, 0.00001351, 0.00001528, 0.00001700, 0.00001865, 0.00002021, 0.00002168, 0.00002303, 0.00002426, 0.00002534, 0.00002628, 0.00002706, 0.00002768, 0.00002812, 0.00002839, },
		{ -0.00001777, -0.00001764, -0.00001722, -0.00001654, -0.00001559, -0.00001438, -0.00001293, -0.00001125, -0.00000936, -0.00000727, -0.00000500, -0.00000258, -0.00000003, 0.00000262, 0.00000536, 0.00000815, 0.00001097, 0.00001378, 0.00001657, 0.00001931, 0.00002196, 0.00002451, 0.00002693, 0.00002920, 0.00003129, 0.00003318, 0.00003486, 0.00003631, 0.00003752, 0.00003847, 0.00003915, 0.00003957, 0.00003971, 0.00003957, 0.00003915, 0.00003847, 0.00003752, 0.00003631, 0.00003486, 0.00003318, 0.00003129, 0.00002920, 0.00002693, 0.00002451, 0.00002196, 0.00001931, 0.00001657, 0.00001378, 0.00001097, 0.00000815, 0.00000536, 0.00000262, -0.00000003, -0.00000258, -0.00000500, -0.00000727, -0.00000936, -0.00001125, -0.00001293, -0.00001438, -0.00001559, -0.00001654, -0.00001722, -0.00001764, },
		{ -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, -0.00001122, },
		{ 0.00000545, 0.00000534, 0.00000500, 0.00000443, 0.00000365, 0.00000266, 0.00000147, 0.00000009, -0.00000146, -0.00000318, -0.00000504, -0.00000703, -0.00000912, -0.00001131, -0.00001355, -0.00001584, -0.00001816, -0.00002047, -0.00002276, -0.00002501, -0.00002719, -0.00002929, -0.00003128, -0.00003314, -0.00003485, -0.00003641, -0.00003779, -0.00003898, -0.00003997, -0.00004075, -0.00004131, -0.00004165, -0.00004177, -0.00004165, -0.00004131, -0.00004075, -0.00003997, -0.00003898, -0.00003779, -0.00003641, -0.00003485, -0.00003314, -0.00003128, -0.00002929, -0.00002719, -0.00002501, -0.00002276, -0.00002047, -0.00001816, -0.00001584, -0.00001355, -0.00001131, -0.00000912, -0.00000703, -0.00000504, -0.00000318, -0.00000146, 0.00000009, 0.00000147, 0.00000266, 0.00000365, 0.00000443, 0.00000500, 0.00000534, },
		{ -0.00001439, -0.00001433, -0.00001415, -0.00001385, -0.00001344, -0.00001291, -0.00001228, -0.00001155, -0.00001072, -0.00000981, -0.00000883, -0.00000777, -0.00000666, -0.00000550, -0.00000431, -0.00000310, -0.00000187, -0.00000064, 0.00000057, 0.00000177, 0.00000292, 0.00000403, 0.00000509, 0.00000608, 0.00000699, 0.00000781, 0.00000854, 0.00000917, 0.00000970, 0.00001011, 0.00001041, 0.00001059, 0.00001065, 0.00001059, 0.00001041, 0.00001011, 0.00000970, 0.00000917, 0.00000854, 0.00000781, 0.00000699, 0.00000608, 0.00000509, 0.00000403, 0.00000292, 0.00000177, 0.00000057, -0.00000064, -0.00000187, -0.00000310, -0.00000431, -0.00000550, -0.00000666, -0.00000777, -0.00000883, -0.00000981, -0.00001072, -0.00001155, -0.00001228, -0.00001291, -0.00001344, -0.00001385, -0.00001415, -0.00001433, },
		{ 0.00000275, 0.00000277, 0.00000281, 0.00000289, 0.00000300, 0.00000314, 0.00000331, 0.00000350, 0.00000372, 0.00000396, 0.00000422, 0.00000450, 0.00000480, 0.00000510, 0.00000542, 0.00000574, 0.00000606, 0.00000639, 0.00000671, 0.00000703, 0.00000733, 0.00000763, 0.00000791, 0.00000817, 0.00000841, 0.00000863, 0.00000882, 0.00000899, 0.00000913, 0.00000924, 0.00000932, 0.00000936, 0.00000938, 0.00000936, 0.00000932, 0.00000924, 0.00000913, 0.00000899, 0.00000882, 0.00000863, 0.00000841, 0.00000817, 0.00000791, 0.00000763, 0.00000733, 0.00000703, 0.00000671, 0.00000639, 0.00000606, 0.00000574, 0.00000542, 0.00000510, 0.00000480, 0.00000450, 0.00000422, 0.00000396, 0.00000372, 0.00000350, 0.00000331, 0.00000314, 0.00000300, 0.00000289, 0.00000281, 0.00000277, },
		{ -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, -0.00000089, },
		{ 0.00000345, 0.00000344, 0.00000340, 0.00000334, 0.00000325, 0.00000313, 0.00000300, 0.00000284, 0.00000267, 0.00000247, 0.00000226, 0.00000203, 0.00000180, 0.00000155, 0.00000129, 0.00000103, 0.00000077, 0.00000051, 0.00000025, -0.00000001, -0.00000025, -0.00000049, -0.00000072, -0.00000093, -0.00000112, -0.00000130, -0.00000146, -0.00000159, -0.00000170, -0.00000179, -0.00000186, -0.00000189, -0.00000191, -0.00000189, -0.00000186, -0.00000179, -0.00000170, -0.00000159, -0.00000146, -0.00000130, -0.00000112, -0.00000093, -0.00000072, -0.00000049, -0.00000025, -0.00000001, 0.00000025, 0.00000051, 0.00000077, 0.00000103, 0.00000129, 0.00000155, 0.00000180, 0.00000203, 0.00000226, 0.00000247, 0.00000267, 0.00000284, 0.00000300, 0.00000313, 0.00000325, 0.00000334, 0.00000340, 0.00000344, },
		{ 0.00001850, 0.00001846, 0.00001835, 0.00001815, 0.00001788, 0.00001754, 0.00001713, 0.00001665, 0.00001611, 0.00001552, 0.00001488, 0.00001419, 0.00001347, 0.00001271, 0.00001194, 0.00001115, 0.00001035, 0.00000955, 0.00000876, 0.00000798, 0.00000723, 0.00000650, 0.00000582, 0.00000517, 0.00000458, 0.00000404, 0.00000357, 0.00000315, 0.00000281, 0.00000254, 0.00000235, 0.00000223, 0.00000219, 0.00000223, 0.00000235, 0.00000254, 0.00000281, 0.00000315, 0.00000357, 0.00000404, 0.00000458, 0.00000517, 0.00000582, 0.00000650, 0.00000723, 0.00000798, 0.00000876, 0.00000955, 0.00001035, 0.00001115, 0.00001194, 0.00001271, 0.00001347, 0.00001419, 0.00001488, 0.00001552, 0.00001611, 0.00001665, 0.00001713, 0.00001754, 0.00001788, 0.00001815, 0.00001835, 0.00001846, },
		{ -0.00000692, -0.00000686, -0.00000669, -0.00000639, -0.00000599, -0.00000547, -0.00000485, -0.00000413, -0.00000332, -0.00000242, -0.00000145, -0.00000041, 0.00000068, 0.00000182, 0.00000299, 0.00000419, 0.00000540, 0.00000660, 0.00000780, 0.00000897, 0.00001011, 0.00001120, 0.00001224, 0.00001321, 0.00001411, 0.00001492, 0.00001564, 0.00001626, 0.00001678, 0.00001719, 0.00001748, 0.00001766, 0.00001772, 0.00001766, 0.00001748, 0.00001719, 0.00001678, 0.00001626, 0.00001564, 0.00001492, 0.00001411, 0.00001321, 0.00001224, 0.00001120, 0.00001011, 0.00000897, 0.00000780, 0.00000660, 0.00000540, 0.00000419, 0.00000299, 0.00000182, 0.00000068, -0.00000041, -0.00000145, -0.00000242, -0.00000332, -0.00000413, -0.00000485, -0.00000547, -0.00000599, -0.00000639, -0.00000669, -0.00000686, },
		{ -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, -0.00000948, },
		{ -0.00000139, -0.00000144, -0.00000157, -0.00000179, -0.00000210, -0.00000249, -0.00000296, -0.00000351, -0.00000412, -0.00000480, -0.00000553, -0.00000632, -0.00000715, -0.00000801, -0.00000889, -0.00000980, -0.00001071, -0.00001163, -0.00001253, -0.00001342, -0.00001428, -0.00001511, -0.00001589, -0.00001663, -0.00001730, -0.00001792, -0.00001846, -0.00001893, -0.00001932, -0.00001963, -0.00001985, -0.00001999, -0.00002003, -0.00001999, -0.00001985, -0.00001963, -0.00001932, -0.00001893, -0.00001846, -0.00001792, -0.00001730, -0.00001663, -0.00001589, -0.00001511, -0.00001428, -0.00001342, -0.00001253, -0.00001163, -0.00001071, -0.00000980, -0.00000889, -0.00000801, -0.00000715, -0.00000632, -0.00000553, -0.00000480, -0.00000412, -0.00000351, -0.00000296, -0.00000249, -0.00000210, -0.00000179, -0.00000157, -0.00000144, },
		{ -0.00000600, -0.00000598, -0.00000591, -0.00000580, -0.00000565, -0.00000546, -0.00000522, -0.00000495, -0.00000465, -0.00000431, -0.00000395, -0.00000356, -0.00000314, -0.00000272, -0.00000228, -0.00000183, -0.00000137, -0.00000092, -0.00000047, -0.00000003, 0.00000040, 0.00000081, 0.00000120, 0.00000156, 0.00000190, 0.00000220, 0.00000247, 0.00000271, 0.00000290, 0.00000306, 0.00000317, 0.00000323, 0.00000326, 0.00000323, 0.00000317, 0.00000306, 0.00000290, 0.00000271, 0.00000247, 0.00000220, 0.00000190, 0.00000156, 0.00000120, 0.00000081, 0.00000040, -0.00000003, -0.00000047, -0.00000092, -0.00000137, -0.00000183, -0.00000228, -0.00000272, -0.00000314, -0.00000356, -0.00000395, -0.00000431, -0.00000465, -0.00000495, -0.00000522, -0.00000546, -0.00000565, -0.00000580, -0.00000591, -0.00000598, },
		{ -0.00000205, -0.00000204, -0.00000202, -0.00000200, -0.00000196, -0.00000191, -0.00000186, -0.00000179, -0.00000172, -0.00000164, -0.00000155, -0.00000145, -0.00000135, -0.00000125, -0.00000114, -0.00000104, -0.00000093, -0.00000082, -0.00000071, -0.00000060, -0.00000050, -0.00000040, -0.00000030, -0.00000022, -0.00000013, -0.00000006, 0.00000000, 0.00000006, 0.00000011, 0.00000015, 0.00000017, 0.00000019, 0.00000019, 0.00000019, 0.00000017, 0.00000015, 0.00000011, 0.00000006, 0.00000000, -0.00000006, -0.00000013, -0.00000022, -0.00000030, -0.00000040, -0.00000050, -0.00000060, -0.00000071, -0.00000082, -0.00000093, -0.00000104, -0.00000114, -0.00000125, -0.00000135, -0.00000145, -0.00000155, -0.00000164, -0.00000172, -0.00000179, -0.00000186, -0.00000191, -0.00000196, -0.00000200, -0.00000202, -0.00000204, },
		{ -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, -0.00000519, },
		{ 0.00000110, 0.00000110, 0.00000109, 0.00000107, 0.00000105, 0.00000102, 0.00000099, 0.00000095, 0.00000091, 0.00000086, 0.00000081, 0.00000075, 0.00000069, 0.00000063, 0.00000057, 0.00000050, 0.00000044, 0.00000037, 0.00000031, 0.00000025, 0.00000018, 0.00000013, 0.00000007, 0.00000002, -0.00000003, -0.00000007, -0.00000011, -0.00000015, -0.00000017, -0.00000020, -0.00000021, -0.00000022, -0.00000023, -0.00000022, -0.00000021, -0.00000020, -0.00000017, -0.00000015, -0.00000011, -0.00000007, -0.00000003, 0.00000002, 0.00000007, 0.00000013, 0.00000018, 0.00000025, 0.00000031, 0.00000037, 0.00000044, 0.00000050, 0.00000057, 0.00000063, 0.00000069, 0.00000075, 0.00000081, 0.00000086, 0.00000091, 0.00000095, 0.00000099, 0.00000102, 0.00000105, 0.00000107, 0.00000109, 0.00000110, },
		{ 0.00000883, 0.00000882, 0.00000880, 0.00000877, 0.00000872, 0.00000865, 0.00000858, 0.00000849, 0.00000839, 0.00000828, 0.00000816, 0.00000804, 0.00000790, 0.00000776, 0.00000762, 0.00000748, 0.00000733, 0.00000718, 0.00000704, 0.00000689, 0.00000675, 0.00000662, 0.00000649, 0.00000637, 0.00000627, 0.00000617, 0.00000608, 0.00000600, 0.00000594, 0.00000589, 0.00000585, 0.00000583, 0.00000583, 0.00000583, 0.00000585, 0.00000589, 0.00000594, 0.00000600, 0.00000608, 0.00000617, 0.00000627, 0.00000637, 0.00000649, 0.00000662, 0.00000675, 0.00000689, 0.00000704, 0.00000718, 0.00000733, 0.00000748, 0.00000762, 0.00000776, 0.00000790, 0.00000804, 0.00000816, 0.00000828, 0.00000839, 0.00000849, 0.00000858, 0.00000865, 0.00000872, 0.00000877, 0.00000880, 0.00000882, },
		{ 0.00000081, 0.00000081, 0.00000083, 0.00000086, 0.00000091, 0.00000096, 0.00000102, 0.00000110, 0.00000118, 0.00000128, 0.00000138, 0.00000149, 0.00000160, 0.00000172, 0.00000184, 0.00000196, 0.00000209, 0.00000221, 0.00000234, 0.00000246, 0.00000258, 0.00000269, 0.00000280, 0.00000290, 0.00000299, 0.00000308, 0.00000315, 0.00000322, 0.00000327, 0.00000331, 0.00000334, 0.00000336, 0.00000337, 0.00000336, 0.00000334, 0.00000331, 0.00000327, 0.00000322, 0.00000315, 0.00000308, 0.00000299, 0.00000290, 0.00000280, 0.00000269, 0.00000258, 0.00000246, 0.00000234, 0.00000221, 0.00000209, 0.00000196, 0.00000184, 0.00000172, 0.00000160, 0.00000149, 0.00000138, 0.00000128, 0.00000118, 0.00000110, 0.00000102, 0.00000096, 0.00000091, 0.00000086, 0.00000083, 0.00000081, },
		},
		{ { -0.00000186, -0.00000202, -0.00000246, -0.00000313, -0.00000391, -0.00000470, -0.00000533, -0.00000568, -0.00000562, -0.00000506, -0.00000394, -0.00000226, -0.00000006, 0.00000255, 0.00000545, 0.00000845, 0.00001136, 0.00001398, 0.00001611, 0.00001760, 0.00001831, 0.00001820, 0.00001724, 0.00001550, 0.00001308, 0.00001016, 0.00000695, 0.00000368, 0.00000058, -0.00000210, -0.00000417, -0.00000549, -0.00000593, -0.00000549, -0.00000417, -0.00000210, 0.00000058, 0.00000368, 0.00000695, 0.00001016, 0.00001308, 0.00001550, 0.00001724, 0.00001820, 0.00001831, 0.00001760, 0.00001611, 0.00001398, 0.00001136, 0.00000845, 0.00000545, 0.00000255, -0.00000006, -0.00000226, -0.00000394, -0.00000506, -0.00000562, -0.00000568, -0.00000533, -0.00000470, -0.00000391, -0.00000313, -0.00000246, -0.00000202, },
		{ -0.00000291, -0.00000278, -0.00000228, -0.00000139, -0.00000017, 0.00000131, 0.00000291, 0.00000450, 0.00000592, 0.00000701, 0.00000765, 0.00000773, 0.00000719, 0.00000605, 0.00000434, 0.00000218, -0.00000028, -0.00000284, -0.00000530, -0.00000744, -0.00000906, -0.00000998, -0.00001008, -0.00000930, -0.00000764, -0.00000515, -0.00000198, 0.00000169, 0.00000563, 0.00000958, 0.00001327, 0.00001644, 0.00001889, 0.00002043, 0.00002095, 0.00002041, 0.00001883, 0.00001632, 0.00001303, 0.00000919, 0.00000502, 0.00000079, -0.00000324, -0.00000685, -0.00000985, -0.00001210, -0.00001352, -0.00001410, -0.00001389, -0.00001300, -0.00001156, -0.00000977, -0.00000780, -0.00000584, -0.00000405, -0.00000258, -0.00000150, -0.00000085, -0.00000062, -0.00000076, -0.00000116, -0.00000171, -0.00000227, -0.00000271, },
		{ -0.00000565, -0.00000493, -0.00000447, -0.00000433, -0.00000454, -0.00000508, -0.00000588, -0.00000681, -0.00000775, -0.00000851, -0.00000893, -0.00000887, -0.00000818, -0.00000680, -0.00000470, -0.00000193, 0.00000142, 0.00000517, 0.00000912, 0.00001302, 0.00001660, 0.00001960, 0.00002178, 0.00002295, 0.00002297, 0.00002178, 0.00001940, 0.00001592, 0.00001152, 0.00000642, 0.00000092, -0.00000467, -0.00001004, -0.00001488, -0.00001894, -0.00002199, -0.00002391, -0.00002463, -0.00002419, -0.00002267, -0.00002025, -0.00001715, -0.00001363, -0.00000997, -0.00000642, -0.00000323, -0.00000061, 0.00000132, 0.00000248, 0.00000285, 0.00000248, 0.00000150, 0.00000006, -0.00000168, -0.00000351, -0.00000526, -0.00000677, -0.00000792, -0.00000863, -0.00000889, -0.00000871, -0.00000819, -0.00000741, -0.00000652, },
		{ 0.00000146, 0.00000182, 0.00000237, 0.00000321, 0.00000441, 0.00000600, 0.00000793, 0.00001013, 0.00001245, 0.00001472, 0.00001674, 0.00001829, 0.00001919, 0.00001925, 0.00001838, 0.00001650, 0.00001363, 0.00000986, 0.00000536, 0.00000034, -0.00000491, -0.00001008, -0.00001487, -0.00001896, -0.00002210, -0.00002407, -0.00002475, -0.00002408, -0.00002211, -0.00001897, -0.00001486, -0.00001007, -0.00000492, 0.00000026, 0.00000512, 0.00000936, 0.00001271, 0.00001497, 0.00001603, 0.00001587, 0.00001453, 0.00001217, 0.00000898, 0.00000522, 0.00000117, -0.00000288, -0.00000666, -0.00000993, -0.00001251, -0.00001428, -0.00001519, -0.00001525, -0.00001456, -0.00001323, -0.00001144, -0.00000937, -0.00000722, -0.00000516, -0.00000331, -0.00000177, -0.00000059, 0.00000025, 0.00000081, 0.00000117, },
		{ 0.00000962, 0.00001018, 0.00001055, 0.00001058, 0.00001016, 0.00000920, 0.00000767, 0.00000558, 0.00000304, 0.00000017, -0.00000284, -0.00000576, -0.00000837, -0.00001042, -0.00001174, -0.00001215, -0.00001156, -0.00000995, -0.00000739, -0.00000401, -0.00000000, 0.00000435, 0.00000875, 0.00001287, 0.00001641, 0.00001907, 0.00002064, 0.00002095, 0.00001993, 0.00001760, 0.00001407, 0.00000954, 0.00000428, -0.00000139, -0.00000713, -0.00001258, -0.00001741, -0.00002134, -0.00002414, -0.00002567, -0.00002589, -0.00002481, -0.00002257, -0.00001935, -0.00001540, -0.00001099, -0.00000642, -0.00000197, 0.00000211, 0.00000562, 0.00000843, 0.00001046, 0.00001171, 0.00001224, 0.00001218, 0.00001167, 0.00001090, 0.00001003, 0.00000923, 0.00000862, 0.00000829, 0.00000827, 0.00000854, 0.00000903, },
		{ -0.00000135, -0.00000223, -0.00000317, -0.00000403, -0.00000465, -0.00000491, -0.00000471, -0.00000402, -0.00000284, -0.00000123, 0.00000067, 0.00000271, 0.00000467, 0.00000636, 0.00000756, 0.00000807, 0.00000777, 0.00000656, 0.00000442, 0.00000141, -0.00000233, -0.00000661, -0.00001118, -0.00001574, -0.00001998, -0.00002360, -0.00002633, -0.00002795, -0.00002831, -0.00002733, -0.00002504, -0.00002153, -0.00001701, -0.00001172, -0.00000598, -0.00000011, 0.00000553, 0.00001063, 0.00001491, 0.00001816, 0.00002022, 0.00002105, 0.00002068, 0.00001920, 0.00001680, 0.00001372, 0.00001021, 0.00000654, 0.00000299, -0.00000021, -0.00000287, -0.00000487, -0.00000615, -0.00000671, -0.00000662, -0.00000600, -0.00000501, -0.00000382, -0.00000261, -0.00000155, -0.00000077, -0.00000035, -0.00000032, -0.00000068, },
		{ 0.00000219, 0.00000217, 0.00000226, 0.00000235, 0.00000232, 0.00000205, 0.00000145, 0.00000047, -0.00000093, -0.00000272, -0.00000481, -0.00000710, -0.00000942, -0.00001159, -0.00001342, -0.00001474, -0.00001537, -0.00001521, -0.00001419, -0.00001231, -0.00000964, -0.00000630, -0.00000247, 0.00000159, 0.00000564, 0.00000939, 0.00001259, 0.00001498, 0.00001639, 0.00001669, 0.00001582, 0.00001382, 0.00001079, 0.00000690, 0.00000239, -0.00000245, -0.00000734, -0.00001196, -0.00001603, -0.00001931, -0.00002163, -0.00002285, -0.00002294, -0.00002194, -0.00001996, -0.00001716, -0.00001375, -0.00000998, -0.00000611, -0.00000236, 0.00000103, 0.00000391, 0.00000615, 0.00000770, 0.00000854, 0.00000875, 0.00000841, 0.00000767, 0.00000667, 0.00000556, 0.00000448, 0.00000355, 0.00000284, 0.00000239, },
		{ -0.00000491, -0.00000565, -0.00000642, -0.00000714, -0.00000772, -0.00000808, -0.00000811, -0.00000777, -0.00000702, -0.00000586, -0.00000433, -0.00000251, -0.00000051, 0.00000153, 0.00000346, 0.00000513, 0.00000638, 0.00000712, 0.00000724, 0.00000671, 0.00000554, 0.00000379, 0.00000158, -0.00000094, -0.00000358, -0.00000614, -0.00000840, -0.00001017, -0.00001127, -0.00001157, -0.00001100, -0.00000952, -0.00000719, -0.00000409, -0.00000038, 0.00000374, 0.00000805, 0.00001232, 0.00001631, 0.00001980, 0.00002262, 0.00002464, 0.00002576, 0.00002597, 0.00002529, 0.00002383, 0.00002169, 0.00001906, 0.00001610, 0.00001301, 0.00000995, 0.00000708, 0.00000452, 0.00000235, 0.00000061, -0.00000071, -0.00000164, -0.00000225, -0.00000264, -0.00000290, -0.00000313, -0.00000340, -0.00000377, -0.00000428, },
		{ 0.00000247, 0.00000385, 0.00000533, 0.00000683, 0.00000830, 0.00000964, 0.00001080, 0.00001170, 0.00001230, 0.00001260, 0.00001258, 0.00001229, 0.00001177, 0.00001110, 0.00001035, 0.00000961, 0.00000897, 0.00000850, 0.00000825, 0.00000827, 0.00000855, 0.00000908, 0.00000980, 0.00001065, 0.00001153, 0.00001235, 0.00001298, 0.00001334, 0.00001334, 0.00001290, 0.00001200, 0.00001062, 0.00000879, 0.00000656, 0.00000403, 0.00000130, -0.00000152, -0.00000428, -0.00000686, -0.00000916, -0.00001107, -0.00001253, -0.00001349, -0.00001395, -0.00001393, -0.00001348, -0.00001267, -0.00001160, -0.00001036, -0.00000904, -0.00000774, -0.00000653, -0.00000547, -0.00000458, -0.00000387, -0.00000332, -0.00000290, -0.00000255, -0.00000221, -0.00000182, -0.00000132, -0.00000066, 0.00000019, 0.00000124, },
		{ 0.00000598, 0.00000523, 0.00000428, 0.00000317, 0.00000193, 0.00000061, -0.00000078, -0.00000221, -0.00000366, -0.00000514, -0.00000663, -0.00000815, -0.00000968, -0.00001124, -0.00001280, -0.00001437, -0.00001591, -0.00001738, -0.00001875, -0.00001997, -0.00002097, -0.00002171, -0.00002212, -0.00002216, -0.00002180, -0.00002101, -0.00001980, -0.00001820, -0.00001624, -0.00001399, -0.00001152, -0.00000894, -0.00000633, -0.00000382, -0.00000149, 0.00000057, 0.00000228, 0.00000360, 0.00000449, 0.00000495, 0.00000501, 0.00000471, 0.00000412, 0.00000332, 0.00000240, 0.00000146, 0.00000057, -0.00000017, -0.00000071, -0.00000099, -0.00000100, -0.00000073, -0.00000021, 0.00000054, 0.00000144, 0.00000245, 0.00000348, 0.00000447, 0.00000535, 0.00000606, 0.00000655, 0.00000680, 0.00000679, 0.00000651, },
		{ -0.00001191, -0.00001243, -0.00001292, -0.00001340, -0.00001385, -0.00001426, -0.00001458, -0.00001474, -0.00001467, -0.00001430, -0.00001356, -0.00001241, -0.00001081, -0.00000877, -0.00000632, -0.00000354, -0.00000054, 0.00000255, 0.00000557, 0.00000835, 0.00001071, 0.00001250, 0.00001358, 0.00001384, 0.00001325, 0.00001177, 0.00000948, 0.00000645, 0.00000284, -0.00000117, -0.00000538, -0.00000954, -0.00001346, -0.00001691, -0.00001971, -0.00002174, -0.00002290, -0.00002316, -0.00002253, -0.00002110, -0.00001898, -0.00001632, -0.00001332, -0.00001016, -0.00000705, -0.00000415, -0.00000162, 0.00000043, 0.00000190, 0.00000278, 0.00000306, 0.00000278, 0.00000204, 0.00000092, -0.00000047, -0.00000202, -0.00000361, -0.00000516, -0.00000660, -0.00000789, -0.00000899, -0.00000992, -0.00001070, -0.00001135, },
		{ 0.00000292, 0.00000224, 0.00000169, 0.00000132, 0.00000117, 0.00000120, 0.00000136, 0.00000156, 0.00000166, 0.00000153, 0.00000103, 0.00000004, -0.00000154, -0.00000375, -0.00000658, -0.00000996, -0.00001374, -0.00001776, -0.00002177, -0.00002551, -0.00002871, -0.00003111, -0.00003248, -0.00003263, -0.00003146, -0.00002894, -0.00002510, -0.00002009, -0.00001412, -0.00000748, -0.00000048, 0.00000651, 0.00001313, 0.00001904, 0.00002396, 0.00002765, 0.00002996, 0.00003082, 0.00003026, 0.00002840, 0.00002542, 0.00002157, 0.00001715, 0.00001246, 0.00000782, 0.00000351, -0.00000023, -0.00000322, -0.00000535, -0.00000658, -0.00000694, -0.00000652, -0.00000547, -0.00000397, -0.00000220, -0.00000038, 0.00000133, 0.00000278, 0.00000387, 0.00000454, 0.00000479, 0.00000466, 0.00000424, 0.00000362, },
		{ -0.00001916, -0.00001936, -0.00001940, -0.00001938, -0.00001939, -0.00001943, -0.00001947, -0.00001942, -0.00001914, -0.00001847, -0.00001724, -0.00001530, -0.00001252, -0.00000883, -0.00000422, 0.00000123, 0.00000735, 0.00001392, 0.00002062, 0.00002710, 0.00003298, 0.00003788, 0.00004146, 0.00004341, 0.00004355, 0.00004177, 0.00003809, 0.00003266, 0.00002572, 0.00001764, 0.00000885, -0.00000015, -0.00000886, -0.00001677, -0.00002345, -0.00002853, -0.00003174, -0.00003294, -0.00003213, -0.00002942, -0.00002505, -0.00001936, -0.00001277, -0.00000574, 0.00000127, 0.00000780, 0.00001347, 0.00001795, 0.00002102, 0.00002256, 0.00002258, 0.00002116, 0.00001850, 0.00001484, 0.00001050, 0.00000578, 0.00000099, -0.00000358, -0.00000772, -0.00001126, -0.00001411, -0.00001626, -0.00001775, -0.00001867, },
		{ -0.00000435, -0.00000138, 0.00000152, 0.00000444, 0.00000742, 0.00001043, 0.00001342, 0.00001624, 0.00001870, 0.00002059, 0.00002167, 0.00002170, 0.00002048, 0.00001789, 0.00001388, 0.00000850, 0.00000192, -0.00000559, -0.00001366, -0.00002183, -0.00002961, -0.00003647, -0.00004190, -0.00004546, -0.00004679, -0.00004567, -0.00004202, -0.00003591, -0.00002758, -0.00001741, -0.00000594, 0.00000623, 0.00001843, 0.00002996, 0.00004017, 0.00004846, 0.00005438, 0.00005758, 0.00005791, 0.00005537, 0.00005013, 0.00004253, 0.00003303, 0.00002218, 0.00001059, -0.00000111, -0.00001233, -0.00002253, -0.00003129, -0.00003828, -0.00004332, -0.00004636, -0.00004747, -0.00004684, -0.00004472, -0.00004144, -0.00003734, -0.00003275, -0.00002798, -0.00002326, -0.00001879, -0.00001466, -0.00001091, -0.00000750, },
		{ 0.00001602, 0.00001433, 0.00001226, 0.00000971, 0.00000663, 0.00000304, -0.00000094, -0.00000514, -0.00000931, -0.00001313, -0.00001624, -0.00001830, -0.00001896, -0.00001797, -0.00001514, -0.00001041, -0.00000387, 0.00000427, 0.00001365, 0.00002378, 0.00003409, 0.00004393, 0.00005263, 0.00005956, 0.00006415, 0.00006595, 0.00006466, 0.00006016, 0.00005252, 0.00004201, 0.00002906, 0.00001428, -0.00000163, -0.00001787, -0.00003364, -0.00004815, -0.00006071, -0.00007073, -0.00007780, -0.00008168, -0.00008230, -0.00007980, -0.00007448, -0.00006678, -0.00005725, -0.00004649, -0.00003515, -0.00002381, -0.00001303, -0.00000324, 0.00000522, 0.00001215, 0.00001748, 0.00002125, 0.00002361, 0.00002476, 0.00002496, 0.00002447, 0.00002356, 0.00002241, 0.00002118, 0.00001995, 0.00001872, 0.00001744, },
		{ -0.00000370, -0.00000443, -0.00000463, -0.00000408, -0.00000264, -0.00000022, 0.00000316, 0.00000739, 0.00001224, 0.00001739, 0.00002244, 0.00002696, 0.00003050, 0.00003262, 0.00003297, 0.00003130, 0.00002748, 0.00002152, 0.00001362, 0.00000411, -0.00000655, -0.00001776, -0.00002887, -0.00003918, -0.00004805, -0.00005485, -0.00005910, -0.00006047, -0.00005876, -0.00005401, -0.00004642, -0.00003637, -0.00002438, -0.00001111, 0.00000274, 0.00001644, 0.00002927, 0.00004061, 0.00004995, 0.00005691, 0.00006129, 0.00006306, 0.00006234, 0.00005941, 0.00005466, 0.00004856, 0.00004162, 0.00003435, 0.00002722, 0.00002063, 0.00001488, 0.00001016, 0.00000654, 0.00000399, 0.00000237, 0.00000150, 0.00000113, 0.00000101, 0.00000093, 0.00000070, 0.00000022, -0.00000055, -0.00000156, -0.00000267, },
		{ 0.00000811, 0.00001121, 0.00001432, 0.00001712, 0.00001930, 0.00002063, 0.00002092, 0.00002009, 0.00001818, 0.00001532, 0.00001175, 0.00000778, 0.00000377, 0.00000013, -0.00000277, -0.00000460, -0.00000512, -0.00000417, -0.00000175, 0.00000205, 0.00000697, 0.00001269, 0.00001876, 0.00002470, 0.00003000, 0.00003418, 0.00003683, 0.00003763, 0.00003638, 0.00003304, 0.00002771, 0.00002063, 0.00001218, 0.00000283, -0.00000688, -0.00001636, -0.00002507, -0.00003250, -0.00003825, -0.00004202, -0.00004368, -0.00004321, -0.00004076, -0.00003661, -0.00003112, -0.00002475, -0.00001799, -0.00001129, -0.00000512, 0.00000018, 0.00000435, 0.00000725, 0.00000884, 0.00000923, 0.00000862, 0.00000728, 0.00000554, 0.00000375, 0.00000223, 0.00000125, 0.00000101, 0.00000162, 0.00000309, 0.00000531, },
		{ 0.00001176, 0.00001012, 0.00000752, 0.00000418, 0.00000032, -0.00000379, -0.00000794, -0.00001195, -0.00001568, -0.00001907, -0.00002211, -0.00002487, -0.00002745, -0.00002996, -0.00003256, -0.00003535, -0.00003842, -0.00004178, -0.00004540, -0.00004916, -0.00005286, -0.00005624, -0.00005903, -0.00006089, -0.00006152, -0.00006063, -0.00005803, -0.00005358, -0.00004726, -0.00003918, -0.00002957, -0.00001875, -0.00000716, 0.00000469, 0.00001625, 0.00002696, 0.00003630, 0.00004383, 0.00004917, 0.00005211, 0.00005254, 0.00005054, 0.00004631, 0.00004018, 0.00003261, 0.00002409, 0.00001521, 0.00000651, -0.00000149, -0.00000833, -0.00001367, -0.00001729, -0.00001908, -0.00001910, -0.00001751, -0.00001458, -0.00001065, -0.00000615, -0.00000148, 0.00000296, 0.00000679, 0.00000974, 0.00001161, 0.00001228, },
		{ -0.00003467, -0.00003952, -0.00004350, -0.00004652, -0.00004854, -0.00004953, -0.00004947, -0.00004836, -0.00004619, -0.00004296, -0.00003865, -0.00003327, -0.00002685, -0.00001942, -0.00001108, -0.00000194, 0.00000782, 0.00001794, 0.00002815, 0.00003810, 0.00004741, 0.00005569, 0.00006254, 0.00006759, 0.00007050, 0.00007104, 0.00006905, 0.00006448, 0.00005740, 0.00004803, 0.00003669, 0.00002381, 0.00000991, -0.00000440, -0.00001853, -0.00003184, -0.00004375, -0.00005377, -0.00006148, -0.00006660, -0.00006897, -0.00006860, -0.00006563, -0.00006033, -0.00005311, -0.00004443, -0.00003483, -0.00002488, -0.00001512, -0.00000608, 0.00000182, 0.00000820, 0.00001282, 0.00001555, 0.00001635, 0.00001531, 0.00001260, 0.00000845, 0.00000316, -0.00000293, -0.00000951, -0.00001624, -0.00002284, -0.00002905, },
		{ -0.00000876, -0.00000158, 0.00000604, 0.00001385, 0.00002161, 0.00002904, 0.00003585, 0.00004174, 0.00004641, 0.00004959, 0.00005105, 0.00005062, 0.00004819, 0.00004374, 0.00003733, 0.00002914, 0.00001943, 0.00000856, -0.00000305, -0.00001490, -0.00002649, -0.00003728, -0.00004677, -0.00005450, -0.00006008, -0.00006324, -0.00006381, -0.00006173, -0.00005709, -0.00005011, -0.00004110, -0.00003048, -0.00001875, -0.00000644, 0.00000589, 0.00001770, 0.00002849, 0.00003782, 0.00004534, 0.00005080, 0.00005406, 0.00005508, 0.00005394, 0.00005079, 0.00004589, 0.00003953, 0.00003207, 0.00002385, 0.00001526, 0.00000663, -0.00000170, -0.00000947, -0.00001643, -0.00002240, -0.00002724, -0.00003086, -0.00003320, -0.00003425, -0.00003401, -0.00003251, -0.00002981, -0.00002598, -0.00002112, -0.00001533, },
		{ 0.00003101, 0.00002835, 0.00002436, 0.00001917, 0.00001300, 0.00000616, -0.00000096, -0.00000794, -0.00001432, -0.00001966, -0.00002356, -0.00002571, -0.00002586, -0.00002393, -0.00001994, -0.00001404, -0.00000655, 0.00000213, 0.00001151, 0.00002103, 0.00003012, 0.00003821, 0.00004479, 0.00004946, 0.00005190, 0.00005194, 0.00004955, 0.00004484, 0.00003807, 0.00002958, 0.00001984, 0.00000934, -0.00000139, -0.00001184, -0.00002154, -0.00003009, -0.00003718, -0.00004263, -0.00004633, -0.00004831, -0.00004867, -0.00004759, -0.00004531, -0.00004209, -0.00003819, -0.00003387, -0.00002934, -0.00002475, -0.00002022, -0.00001580, -0.00001149, -0.00000727, -0.00000309, 0.00000111, 0.00000535, 0.00000963, 0.00001393, 0.00001816, 0.00002217, 0.00002580, 0.00002884, 0.00003107, 0.00003228, 0.00003230, },
		{ -0.00002022, -0.00001969, -0.00001774, -0.00001439, -0.00000983, -0.00000436, 0.00000159, 0.00000748, 0.00001276, 0.00001684, 0.00001922, 0.00001948, 0.00001733, 0.00001267, 0.00000560, -0.00000362, -0.00001451, -0.00002647, -0.00003875, -0.00005056, -0.00006112, -0.00006966, -0.00007556, -0.00007834, -0.00007771, -0.00007362, -0.00006623, -0.00005592, -0.00004325, -0.00002895, -0.00001381, 0.00000130, 0.00001556, 0.00002821, 0.00003865, 0.00004641, 0.00005125, 0.00005311, 0.00005215, 0.00004870, 0.00004323, 0.00003633, 0.00002861, 0.00002071, 0.00001318, 0.00000652, 0.00000106, -0.00000299, -0.00000558, -0.00000682, -0.00000692, -0.00000621, -0.00000506, -0.00000385, -0.00000293, -0.00000260, -0.00000303, -0.00000430, -0.00000636, -0.00000902, -0.00001201, -0.00001500, -0.00001760, -0.00001945, },
		{ 0.00003063, 0.00003005, 0.00002743, 0.00002270, 0.00001601, 0.00000772, -0.00000160, -0.00001124, -0.00002036, -0.00002812, -0.00003370, -0.00003640, -0.00003571, -0.00003137, -0.00002338, -0.00001205, 0.00000204, 0.00001806, 0.00003497, 0.00005162, 0.00006681, 0.00007939, 0.00008837, 0.00009294, 0.00009259, 0.00008714, 0.00007674, 0.00006188, 0.00004337, 0.00002224, -0.00000029, -0.00002291, -0.00004432, -0.00006332, -0.00007889, -0.00009026, -0.00009695, -0.00009883, -0.00009605, -0.00008909, -0.00007868, -0.00006571, -0.00005120, -0.00003619, -0.00002167, -0.00000848, 0.00000271, 0.00001147, 0.00001760, 0.00002116, 0.00002244, 0.00002189, 0.00002008, 0.00001766, 0.00001523, 0.00001334, 0.00001240, 0.00001264, 0.00001411, 0.00001666, 0.00001996, 0.00002356, 0.00002691, 0.00002944, },
		{ -0.00002255, -0.00002806, -0.00003092, -0.00003073, -0.00002737, -0.00002104, -0.00001228, -0.00000186, 0.00000919, 0.00001976, 0.00002867, 0.00003484, 0.00003735, 0.00003557, 0.00002921, 0.00001833, 0.00000341, -0.00001469, -0.00003481, -0.00005555, -0.00007538, -0.00009273, -0.00010617, -0.00011447, -0.00011671, -0.00011239, -0.00010144, -0.00008425, -0.00006166, -0.00003487, -0.00000536, 0.00002517, 0.00005499, 0.00008242, 0.00010597, 0.00012444, 0.00013701, 0.00014326, 0.00014324, 0.00013739, 0.00012655, 0.00011185, 0.00009461, 0.00007623, 0.00005810, 0.00004145, 0.00002730, 0.00001634, 0.00000895, 0.00000513, 0.00000456, 0.00000664, 0.00001054, 0.00001532, 0.00002002, 0.00002374, 0.00002576, 0.00002558, 0.00002296, 0.00001799, 0.00001101, 0.00000264, -0.00000634, -0.00001504, },
		{ -0.00000660, 0.00000416, 0.00001295, 0.00001889, 0.00002142, 0.00002036, 0.00001591, 0.00000867, -0.00000041, -0.00001014, -0.00001919, -0.00002621, -0.00002995, -0.00002942, -0.00002397, -0.00001336, 0.00000216, 0.00002187, 0.00004463, 0.00006893, 0.00009301, 0.00011503, 0.00013314, 0.00014570, 0.00015141, 0.00014939, 0.00013929, 0.00012132, 0.00009626, 0.00006540, 0.00003046, -0.00000655, -0.00004344, -0.00007803, -0.00010832, -0.00013260, -0.00014961, -0.00015860, -0.00015940, -0.00015241, -0.00013856, -0.00011923, -0.00009611, -0.00007110, -0.00004613, -0.00002302, -0.00000333, 0.00001176, 0.00002148, 0.00002560, 0.00002434, 0.00001838, 0.00000877, -0.00000318, -0.00001604, -0.00002836, -0.00003884, -0.00004642, -0.00005039, -0.00005041, -0.00004657, -0.00003935, -0.00002956, -0.00001824, },
		{ 0.00000299, -0.00000866, -0.00001979, -0.00002919, -0.00003591, -0.00003936, -0.00003941, -0.00003638, -0.00003100, -0.00002438, -0.00001790, -0.00001303, -0.00001122, -0.00001375, -0.00002153, -0.00003507, -0.00005433, -0.00007870, -0.00010705, -0.00013775, -0.00016879, -0.00019796, -0.00022298, -0.00024170, -0.00025228, -0.00025335, -0.00024416, -0.00022461, -0.00019532, -0.00015762, -0.00011343, -0.00006515, -0.00001550, 0.00003266, 0.00007658, 0.00011379, 0.00014226, 0.00016059, 0.00016805, 0.00016467, 0.00015122, 0.00012912, 0.00010032, 0.00006716, 0.00003219, -0.00000206, -0.00003325, -0.00005937, -0.00007896, -0.00009110, -0.00009554, -0.00009266, -0.00008337, -0.00006907, -0.00005146, -0.00003240, -0.00001375, 0.00000282, 0.00001594, 0.00002467, 0.00002855, 0.00002763, 0.00002240, 0.00001379, },
		{ -0.00010553, -0.00009988, -0.00009357, -0.00008800, -0.00008439, -0.00008359, -0.00008603, -0.00009160, -0.00009971, -0.00010926, -0.00011880, -0.00012664, -0.00013101, -0.00013025, -0.00012300, -0.00010836, -0.00008600, -0.00005623, -0.00002007, 0.00002083, 0.00006427, 0.00010763, 0.00014811, 0.00018286, 0.00020929, 0.00022521, 0.00022908, 0.00022008, 0.00019828, 0.00016458, 0.00012073, 0.00006918, 0.00001293, -0.00004466, -0.00010017, -0.00015027, -0.00019207, -0.00022321, -0.00024213, -0.00024808, -0.00024125, -0.00022267, -0.00019413, -0.00015806, -0.00011733, -0.00007496, -0.00003399, 0.00000282, 0.00003317, 0.00005535, 0.00006839, 0.00007203, 0.00006676, 0.00005374, 0.00003462, 0.00001140, -0.00001375, -0.00003871, -0.00006157, -0.00008080, -0.00009537, -0.00010479, -0.00010915, -0.00010906, },
		{ -0.00000991, -0.00001567, -0.00002277, -0.00002947, -0.00003415, -0.00003547, -0.00003252, -0.00002493, -0.00001297, 0.00000251, 0.00002012, 0.00003805, 0.00005423, 0.00006656, 0.00007308, 0.00007219, 0.00006286, 0.00004472, 0.00001817, -0.00001562, -0.00005475, -0.00009672, -0.00013858, -0.00017719, -0.00020943, -0.00023249, -0.00024405, -0.00024254, -0.00022727, -0.00019849, -0.00015744, -0.00010624, -0.00004780, 0.00001441, 0.00007661, 0.00013494, 0.00018581, 0.00022609, 0.00025338, 0.00026615, 0.00026388, 0.00024705, 0.00021710, 0.00017633, 0.00012766, 0.00007447, 0.00002027, -0.00003151, -0.00007781, -0.00011611, -0.00014466, -0.00016250, -0.00016955, -0.00016655, -0.00015493, -0.00013668, -0.00011412, -0.00008972, -0.00006582, -0.00004448, -0.00002727, -0.00001519, -0.00000856, -0.00000709, },
		{ -0.00014163, -0.00013358, -0.00012039, -0.00010391, -0.00008617, -0.00006910, -0.00005432, -0.00004303, -0.00003584, -0.00003271, -0.00003297, -0.00003540, -0.00003834, -0.00003989, -0.00003809, -0.00003120, -0.00001785, 0.00000274, 0.00003063, 0.00006513, 0.00010473, 0.00014723, 0.00018986, 0.00022950, 0.00026288, 0.00028690, 0.00029886, 0.00029673, 0.00027933, 0.00024646, 0.00019900, 0.00013887, 0.00006891, -0.00000726, -0.00008551, -0.00016146, -0.00023083, -0.00028970, -0.00033483, -0.00036386, -0.00037551, -0.00036963, -0.00034724, -0.00031042, -0.00026214, -0.00020607, -0.00014622, -0.00008673, -0.00003152, 0.00001601, 0.00005316, 0.00007815, 0.00009018, 0.00008947, 0.00007719, 0.00005530, 0.00002637, -0.00000666, -0.00004075, -0.00007301, -0.00010094, -0.00012260, -0.00013679, -0.00014303, },
		{ 0.00004404, 0.00007655, 0.00010129, 0.00011909, 0.00013123, 0.00013913, 0.00014425, 0.00014780, 0.00015063, 0.00015310, 0.00015504, 0.00015575, 0.00015411, 0.00014870, 0.00013798, 0.00012051, 0.00009519, 0.00006138, 0.00001915, -0.00003069, -0.00008653, -0.00014597, -0.00020602, -0.00026318, -0.00031378, -0.00035420, -0.00038117, -0.00039206, -0.00038514, -0.00035976, -0.00031644, -0.00025694, -0.00018416, -0.00010201, -0.00001517, 0.00007122, 0.00015192, 0.00022189, 0.00027672, 0.00031287, 0.00032797, 0.00032098, 0.00029223, 0.00024346, 0.00017764, 0.00009878, 0.00001165, -0.00007855, -0.00016656, -0.00024741, -0.00031676, -0.00037111, -0.00040810, -0.00042654, -0.00042643, -0.00040897, -0.00037630, -0.00033137, -0.00027759, -0.00021864, -0.00015810, -0.00009921, -0.00004468, 0.00000348, },
		{ 0.00035687, 0.00031635, 0.00027698, 0.00023920, 0.00020300, 0.00016811, 0.00013416, 0.00010087, 0.00006818, 0.00003636, 0.00000614, -0.00002137, -0.00004470, -0.00006214, -0.00007193, -0.00007245, -0.00006244, -0.00004122, -0.00000884, 0.00003375, 0.00008470, 0.00014121, 0.00019972, 0.00025603, 0.00030555, 0.00034368, 0.00036606, 0.00036897, 0.00034962, 0.00030644, 0.00023925, 0.00014937, 0.00003963, -0.00008575, -0.00022134, -0.00036086, -0.00049749, -0.00062431, -0.00073475, -0.00082297, -0.00088423, -0.00091522, -0.00091423, -0.00088125, -0.00081797, -0.00072763, -0.00061479, -0.00048504, -0.00034459, -0.00019992, -0.00005733, 0.00007738, 0.00019926, 0.00030446, 0.00039033, 0.00045554, 0.00049997, 0.00052465, 0.00053150, 0.00052312, 0.00050251, 0.00047280, 0.00043694, 0.00039759, },
		{ 0.00001717, 0.00000296, -0.00000758, -0.00001485, -0.00001899, -0.00002005, -0.00001805, -0.00001321, -0.00000606, 0.00000248, 0.00001102, 0.00001769, 0.00002025, 0.00001618, 0.00000291, -0.00002199, -0.00006049, -0.00011392, -0.00018268, -0.00026609, -0.00036228, -0.00046813, -0.00057938, -0.00069076, -0.00079628, -0.00088952, -0.00096406, -0.00101386, -0.00103369, -0.00101952, -0.00096883, -0.00088086, -0.00075673, -0.00059944, -0.00041376, -0.00020605, 0.00001616, 0.00024451, 0.00047032, 0.00068503, 0.00088074, 0.00105057, 0.00118906, 0.00129242, 0.00135868, 0.00138771, 0.00138111, 0.00134207, 0.00127504, 0.00118539, 0.00107902, 0.00096198, 0.00084010, 0.00071865, 0.00060210, 0.00049394, 0.00039660, 0.00031147, 0.00023898, 0.00017874, 0.00012975, 0.00009062, 0.00005981, 0.00003577, },
		{ -0.00024641, -0.00024790, -0.00025248, -0.00026047, -0.00027228, -0.00028832, -0.00030884, -0.00033373, -0.00036240, -0.00039362, -0.00042547, -0.00045533, -0.00047991, -0.00049542, -0.00049777, -0.00048285, -0.00044681, -0.00038643, -0.00029944, -0.00018475, -0.00004276, 0.00012458, 0.00031368, 0.00051941, 0.00073526, 0.00095368, 0.00116637, 0.00136481, 0.00154068, 0.00168637, 0.00179542, 0.00186291, 0.00188575, 0.00186291, 0.00179542, 0.00168637, 0.00154068, 0.00136481, 0.00116637, 0.00095368, 0.00073526, 0.00051941, 0.00031368, 0.00012458, -0.00004276, -0.00018475, -0.00029944, -0.00038643, -0.00044681, -0.00048285, -0.00049777, -0.00049542, -0.00047991, -0.00045533, -0.00042547, -0.00039362, -0.00036240, -0.00033373, -0.00030884, -0.00028832, -0.00027228, -0.00026047, -0.00025248, -0.00024790, },
		{ 0.00001717, 0.00003577, 0.00005981, 0.00009062, 0.00012975, 0.00017874, 0.00023898, 0.00031147, 0.00039660, 0.00049394, 0.00060210, 0.00071865, 0.00084010, 0.00096198, 0.00107902, 0.00118539, 0.00127504, 0.00134207, 0.00138111, 0.00138771, 0.00135868, 0.00129242, 0.00118906, 0.00105057, 0.00088074, 0.00068503, 0.00047032, 0.00024451, 0.00001616, -0.00020605, -0.00041376, -0.00059944, -0.00075673, -0.00088086, -0.00096883, -0.00101952, -0.00103369, -0.00101386, -0.00096406, -0.00088952, -0.00079628, -0.00069076, -0.00057938, -0.00046813, -0.00036228, -0.00026609, -0.00018268, -0.00011392, -0.00006049, -0.00002199, 0.00000291, 0.00001618, 0.00002025, 0.00001769, 0.00001102, 0.00000248, -0.00000606, -0.00001321, -0.00001805, -0.00002005, -0.00001899, -0.00001485, -0.00000758, 0.00000296, },
		{ 0.00035687, 0.00039759, 0.00043694, 0.00047280, 0.00050251, 0.00052312, 0.00053150, 0.00052465, 0.00049997, 0.00045554, 0.00039033, 0.00030446, 0.00019926, 0.00007738, -0.00005733, -0.00019992, -0.00034459, -0.00048504, -0.00061479, -0.00072763, -0.00081797, -0.00088125, -0.00091423, -0.00091522, -0.00088423, -0.00082297, -0.00073475, -0.00062431, -0.00049749, -0.00036086, -0.00022134, -0.00008575, 0.00003963, 0.00014937, 0.00023925, 0.00030644, 0.00034962, 0.00036897, 0.00036606, 0.00034368, 0.00030555, 0.00025603, 0.00019972, 0.00014121, 0.00008470, 0.00003375, -0.00000884, -0.00004122, -0.00006244, -0.00007245, -0.00007193, -0.00006214, -0.00004470, -0.00002137, 0.00000614, 0.00003636, 0.00006818, 0.00010087, 0.00013416, 0.00016811, 0.00020300, 0.00023920, 0.00027698, 0.00031635, },
		{ 0.00004404, 0.00000348, -0.00004468, -0.00009921, -0.00015810, -0.00021864, -0.00027759, -0.00033137, -0.00037630, -0.00040897, -0.00042643, -0.00042654, -0.00040810, -0.00037111, -0.00031676, -0.00024741, -0.00016656, -0.00007855, 0.00001165, 0.00009878, 0.00017764, 0.00024346, 0.00029223, 0.00032098, 0.00032797, 0.00031287, 0.00027672, 0.00022189, 0.00015192, 0.00007122, -0.00001517, -0.00010201, -0.00018416, -0.00025694, -0.00031644, -0.00035976, -0.00038514, -0.00039206, -0.00038117, -0.00035420, -0.00031378, -0.00026318, -0.00020602, -0.00014597, -0.00008653, -0.00003069, 0.00001915, 0.00006138, 0.00009519, 0.00012051, 0.00013798, 0.00014870, 0.00015411, 0.00015575, 0.00015504, 0.00015310, 0.00015063, 0.00014780, 0.00014425, 0.00013913, 0.00013123, 0.00011909, 0.00010129, 0.00007655, },
		{ -0.00014163, -0.00014303, -0.00013679, -0.00012260, -0.00010094, -0.00007301, -0.00004075, -0.00000666, 0.00002637, 0.00005530, 0.00007719, 0.00008947, 0.00009018, 0.00007815, 0.00005316, 0.00001601, -0.00003152, -0.00008673, -0.00014622, -0.00020607, -0.00026214, -0.00031042, -0.00034724, -0.00036963, -0.00037551, -0.00036386, -0.00033483, -0.00028970, -0.00023083, -0.00016146, -0.00008551, -0.00000726, 0.00006891, 0.00013887, 0.00019900, 0.00024646, 0.00027933, 0.00029673, 0.00029886, 0.00028690, 0.00026288, 0.00022950, 0.00018986, 0.00014723, 0.00010473, 0.00006513, 0.00003063, 0.00000274, -0.00001785, -0.00003120, -0.00003809, -0.00003989, -0.00003834, -0.00003540, -0.00003297, -0.00003271, -0.00003584, -0.00004303, -0.00005432, -0.00006910, -0.00008617, -0.00010391, -0.00012039, -0.00013358, },
		{ -0.00000991, -0.00000709, -0.00000856, -0.00001519, -0.00002727, -0.00004448, -0.00006582, -0.00008972, -0.00011412, -0.00013668, -0.00015493, -0.00016655, -0.00016955, -0.00016250, -0.00014466, -0.00011611, -0.00007781, -0.00003151, 0.00002027, 0.00007447, 0.00012766, 0.00017633, 0.00021710, 0.00024705, 0.00026388, 0.00026615, 0.00025338, 0.00022609, 0.00018581, 0.00013494, 0.00007661, 0.00001441, -0.00004780, -0.00010624, -0.00015744, -0.00019849, -0.00022727, -0.00024254, -0.00024405, -0.00023249, -0.00020943, -0.00017719, -0.00013858, -0.00009672, -0.00005475, -0.00001562, 0.00001817, 0.00004472, 0.00006286, 0.00007219, 0.00007308, 0.00006656, 0.00005423, 0.00003805, 0.00002012, 0.00000251, -0.00001297, -0.00002493, -0.00003252, -0.00003547, -0.00003415, -0.00002947, -0.00002277, -0.00001567, },
		{ -0.00010553, -0.00010906, -0.00010915, -0.00010479, -0.00009537, -0.00008080, -0.00006157, -0.00003871, -0.00001375, 0.00001140, 0.00003462, 0.00005374, 0.00006676, 0.00007203, 0.00006839, 0.00005535, 0.00003317, 0.00000282, -0.00003399, -0.00007496, -0.00011733, -0.00015806, -0.00019413, -0.00022267, -0.00024125, -0.00024808, -0.00024213, -0.00022321, -0.00019207, -0.00015027, -0.00010017, -0.00004466, 0.00001293, 0.00006918, 0.00012073, 0.00016458, 0.00019828, 0.00022008, 0.00022908, 0.00022521, 0.00020929, 0.00018286, 0.00014811, 0.00010763, 0.00006427, 0.00002083, -0.00002007, -0.00005623, -0.00008600, -0.00010836, -0.00012300, -0.00013025, -0.00013101, -0.00012664, -0.00011880, -0.00010926, -0.00009971, -0.00009160, -0.00008603, -0.00008359, -0.00008439, -0.00008800, -0.00009357, -0.00009988, },
		{ 0.00000299, 0.00001379, 0.00002240, 0.00002763, 0.00002855, 0.00002467, 0.00001594, 0.00000282, -0.00001375, -0.00003240, -0.00005146, -0.00006907, -0.00008337, -0.00009266, -0.00009554, -0.00009110, -0.00007896, -0.00005937, -0.00003325, -0.00000206, 0.00003219, 0.00006716, 0.00010032, 0.00012912, 0.00015122, 0.00016467, 0.00016805, 0.00016059, 0.00014226, 0.00011379, 0.00007658, 0.00003266, -0.00001550, -0.00006515, -0.00011343, -0.00015762, -0.00019532, -0.00022461, -0.00024416, -0.00025335, -0.00025228, -0.00024170, -0.00022298, -0.00019796, -0.00016879, -0.00013775, -0.00010705, -0.00007870, -0.00005433, -0.00003507, -0.00002153, -0.00001375, -0.00001122, -0.00001303, -0.00001790, -0.00002438, -0.00003100, -0.00003638, -0.00003941, -0.00003936, -0.00003591, -0.00002919, -0.00001979, -0.00000866, },
		{ -0.00000660, -0.00001824, -0.00002956, -0.00003935, -0.00004657, -0.00005041, -0.00005039, -0.00004642, -0.00003884, -0.00002836, -0.00001604, -0.00000318, 0.00000877, 0.00001838, 0.00002434, 0.00002560, 0.00002148, 0.00001176, -0.00000333, -0.00002302, -0.00004613, -0.00007110, -0.00009611, -0.00011923, -0.00013856, -0.00015241, -0.00015940, -0.00015860, -0.00014961, -0.00013260, -0.00010832, -0.00007803, -0.00004344, -0.00000655, 0.00003046, 0.00006540, 0.00009626, 0.00012132, 0.00013929, 0.00014939, 0.00015141, 0.00014570, 0.00013314, 0.00011503, 0.00009301, 0.00006893, 0.00004463, 0.00002187, 0.00000216, -0.00001336, -0.00002397, -0.00002942, -0.00002995, -0.00002621, -0.00001919, -0.00001014, -0.00000041, 0.00000867, 0.00001591, 0.00002036, 0.00002142, 0.00001889, 0.00001295, 0.00000416, },
		{ -0.00002255, -0.00001504, -0.00000634, 0.00000264, 0.00001101, 0.00001799, 0.00002296, 0.00002558, 0.00002576, 0.00002374, 0.00002002, 0.00001532, 0.00001054, 0.00000664, 0.00000456, 0.00000513, 0.00000895, 0.00001634, 0.00002730, 0.00004145, 0.00005810, 0.00007623, 0.00009461, 0.00011185, 0.00012655, 0.00013739, 0.00014324, 0.00014326, 0.00013701, 0.00012444, 0.00010597, 0.00008242, 0.00005499, 0.00002517, -0.00000536, -0.00003487, -0.00006166, -0.00008425, -0.00010144, -0.00011239, -0.00011671, -0.00011447, -0.00010617, -0.00009273, -0.00007538, -0.00005555, -0.00003481, -0.00001469, 0.00000341, 0.00001833, 0.00002921, 0.00003557, 0.00003735, 0.00003484, 0.00002867, 0.00001976, 0.00000919, -0.00000186, -0.00001228, -0.00002104, -0.00002737, -0.00003073, -0.00003092, -0.00002806, },
		{ 0.00003063, 0.00002944, 0.00002691, 0.00002356, 0.00001996, 0.00001666, 0.00001411, 0.00001264, 0.00001240, 0.00001334, 0.00001523, 0.00001766, 0.00002008, 0.00002189, 0.00002244, 0.00002116, 0.00001760, 0.00001147, 0.00000271, -0.00000848, -0.00002167, -0.00003619, -0.00005120, -0.00006571, -0.00007868, -0.00008909, -0.00009605, -0.00009883, -0.00009695, -0.00009026, -0.00007889, -0.00006332, -0.00004432, -0.00002291, -0.00000029, 0.00002224, 0.00004337, 0.00006188, 0.00007674, 0.00008714, 0.00009259, 0.00009294, 0.00008837, 0.00007939, 0.00006681, 0.00005162, 0.00003497, 0.00001806, 0.00000204, -0.00001205, -0.00002338, -0.00003137, -0.00003571, -0.00003640, -0.00003370, -0.00002812, -0.00002036, -0.00001124, -0.00000160, 0.00000772, 0.00001601, 0.00002270, 0.00002743, 0.00003005, },
		{ -0.00002022, -0.00001945, -0.00001760, -0.00001500, -0.00001201, -0.00000902, -0.00000636, -0.00000430, -0.00000303, -0.00000260, -0.00000293, -0.00000385, -0.00000506, -0.00000621, -0.00000692, -0.00000682, -0.00000558, -0.00000299, 0.00000106, 0.00000652, 0.00001318, 0.00002071, 0.00002861, 0.00003633, 0.00004323, 0.00004870, 0.00005215, 0.00005311, 0.00005125, 0.00004641, 0.00003865, 0.00002821, 0.00001556, 0.00000130, -0.00001381, -0.00002895, -0.00004325, -0.00005592, -0.00006623, -0.00007362, -0.00007771, -0.00007834, -0.00007556, -0.00006966, -0.00006112, -0.00005056, -0.00003875, -0.00002647, -0.00001451, -0.00000362, 0.00000560, 0.00001267, 0.00001733, 0.00001948, 0.00001922, 0.00001684, 0.00001276, 0.00000748, 0.00000159, -0.00000436, -0.00000983, -0.00001439, -0.00001774, -0.00001969, },
		{ 0.00003101, 0.00003230, 0.00003228, 0.00003107, 0.00002884, 0.00002580, 0.00002217, 0.00001816, 0.00001393, 0.00000963, 0.00000535, 0.00000111, -0.00000309, -0.00000727, -0.00001149, -0.00001580, -0.00002022, -0.00002475, -0.00002934, -0.00003387, -0.00003819, -0.00004209, -0.00004531, -0.00004759, -0.00004867, -0.00004831, -0.00004633, -0.00004263, -0.00003718, -0.00003009, -0.00002154, -0.00001184, -0.00000139, 0.00000934, 0.00001984, 0.00002958, 0.00003807, 0.00004484, 0.00004955, 0.00005194, 0.00005190, 0.00004946, 0.00004479, 0.00003821, 0.00003012, 0.00002103, 0.00001151, 0.00000213, -0.00000655, -0.00001404, -0.00001994, -0.00002393, -0.00002586, -0.00002571, -0.00002356, -0.00001966, -0.00001432, -0.00000794, -0.00000096, 0.00000616, 0.00001300, 0.00001917, 0.00002436, 0.00002835, },
		{ -0.00000876, -0.00001533, -0.00002112, -0.00002598, -0.00002981, -0.00003251, -0.00003401, -0.00003425, -0.00003320, -0.00003086, -0.00002724, -0.00002240, -0.00001643, -0.00000947, -0.00000170, 0.00000663, 0.00001526, 0.00002385, 0.00003207, 0.00003953, 0.00004589, 0.00005079, 0.00005394, 0.00005508, 0.00005406, 0.00005080, 0.00004534, 0.00003782, 0.00002849, 0.00001770, 0.00000589, -0.00000644, -0.00001875, -0.00003048, -0.00004110, -0.00005011, -0.00005709, -0.00006173, -0.00006381, -0.00006324, -0.00006008, -0.00005450, -0.00004677, -0.00003728, -0.00002649, -0.00001490, -0.00000305, 0.00000856, 0.00001943, 0.00002914, 0.00003733, 0.00004374, 0.00004819, 0.00005062, 0.00005105, 0.00004959, 0.00004641, 0.00004174, 0.00003585, 0.00002904, 0.00002161, 0.00001385, 0.00000604, -0.00000158, },
		{ -0.00003467, -0.00002905, -0.00002284, -0.00001624, -0.00000951, -0.00000293, 0.00000316, 0.00000845, 0.00001260, 0.00001531, 0.00001635, 0.00001555, 0.00001282, 0.00000820, 0.00000182, -0.00000608, -0.00001512, -0.00002488, -0.00003483, -0.00004443, -0.00005311, -0.00006033, -0.00006563, -0.00006860, -0.00006897, -0.00006660, -0.00006148, -0.00005377, -0.00004375, -0.00003184, -0.00001853, -0.00000440, 0.00000991, 0.00002381, 0.00003669, 0.00004803, 0.00005740, 0.00006448, 0.00006905, 0.00007104, 0.00007050, 0.00006759, 0.00006254, 0.00005569, 0.00004741, 0.00003810, 0.00002815, 0.00001794, 0.00000782, -0.00000194, -0.00001108, -0.00001942, -0.00002685, -0.00003327, -0.00003865, -0.00004296, -0.00004619, -0.00004836, -0.00004947, -0.00004953, -0.00004854, -0.00004652, -0.00004350, -0.00003952, },
		{ 0.00001176, 0.00001228, 0.00001161, 0.00000974, 0.00000679, 0.00000296, -0.00000148, -0.00000615, -0.00001065, -0.00001458, -0.00001751, -0.00001910, -0.00001908, -0.00001729, -0.00001367, -0.00000833, -0.00000149, 0.00000651, 0.00001521, 0.00002409, 0.00003261, 0.00004018, 0.00004631, 0.00005054, 0.00005254, 0.00005211, 0.00004917, 0.00004383, 0.00003630, 0.00002696, 0.00001625, 0.00000469, -0.00000716, -0.00001875, -0.00002957, -0.00003918, -0.00004726, -0.00005358, -0.00005803, -0.00006063, -0.00006152, -0.00006089, -0.00005903, -0.00005624, -0.00005286, -0.00004916, -0.00004540, -0.00004178, -0.00003842, -0.00003535, -0.00003256, -0.00002996, -0.00002745, -0.00002487, -0.00002211, -0.00001907, -0.00001568, -0.00001195, -0.00000794, -0.00000379, 0.00000032, 0.00000418, 0.00000752, 0.00001012, },
		{ 0.00000811, 0.00000531, 0.00000309, 0.00000162, 0.00000101, 0.00000125, 0.00000223, 0.00000375, 0.00000554, 0.00000728, 0.00000862, 0.00000923, 0.00000884, 0.00000725, 0.00000435, 0.00000018, -0.00000512, -0.00001129, -0.00001799, -0.00002475, -0.00003112, -0.00003661, -0.00004076, -0.00004321, -0.00004368, -0.00004202, -0.00003825, -0.00003250, -0.00002507, -0.00001636, -0.00000688, 0.00000283, 0.00001218, 0.00002063, 0.00002771, 0.00003304, 0.00003638, 0.00003763, 0.00003683, 0.00003418, 0.00003000, 0.00002470, 0.00001876, 0.00001269, 0.00000697, 0.00000205, -0.00000175, -0.00000417, -0.00000512, -0.00000460, -0.00000277, 0.00000013, 0.00000377, 0.00000778, 0.00001175, 0.00001532, 0.00001818, 0.00002009, 0.00002092, 0.00002063, 0.00001930, 0.00001712, 0.00001432, 0.00001121, },
		{ -0.00000370, -0.00000267, -0.00000156, -0.00000055, 0.00000022, 0.00000070, 0.00000093, 0.00000101, 0.00000113, 0.00000150, 0.00000237, 0.00000399, 0.00000654, 0.00001016, 0.00001488, 0.00002063, 0.00002722, 0.00003435, 0.00004162, 0.00004856, 0.00005466, 0.00005941, 0.00006234, 0.00006306, 0.00006129, 0.00005691, 0.00004995, 0.00004061, 0.00002927, 0.00001644, 0.00000274, -0.00001111, -0.00002438, -0.00003637, -0.00004642, -0.00005401, -0.00005876, -0.00006047, -0.00005910, -0.00005485, -0.00004805, -0.00003918, -0.00002887, -0.00001776, -0.00000655, 0.00000411, 0.00001362, 0.00002152, 0.00002748, 0.00003130, 0.00003297, 0.00003262, 0.00003050, 0.00002696, 0.00002244, 0.00001739, 0.00001224, 0.00000739, 0.00000316, -0.00000022, -0.00000264, -0.00000408, -0.00000463, -0.00000443, },
		{ 0.00001602, 0.00001744, 0.00001872, 0.00001995, 0.00002118, 0.00002241, 0.00002356, 0.00002447, 0.00002496, 0.00002476, 0.00002361, 0.00002125, 0.00001748, 0.00001215, 0.00000522, -0.00000324, -0.00001303, -0.00002381, -0.00003515, -0.00004649, -0.00005725, -0.00006678, -0.00007448, -0.00007980, -0.00008230, -0.00008168, -0.00007780, -0.00007073, -0.00006071, -0.00004815, -0.00003364, -0.00001787, -0.00000163, 0.00001428, 0.00002906, 0.00004201, 0.00005252, 0.00006016, 0.00006466, 0.00006595, 0.00006415, 0.00005956, 0.00005263, 0.00004393, 0.00003409, 0.00002378, 0.00001365, 0.00000427, -0.00000387, -0.00001041, -0.00001514, -0.00001797, -0.00001896, -0.00001830, -0.00001624, -0.00001313, -0.00000931, -0.00000514, -0.00000094, 0.00000304, 0.00000663, 0.00000971, 0.00001226, 0.00001433, },
		{ -0.00000435, -0.00000750, -0.00001091, -0.00001466, -0.00001879, -0.00002326, -0.00002798, -0.00003275, -0.00003734, -0.00004144, -0.00004472, -0.00004684, -0.00004747, -0.00004636, -0.00004332, -0.00003828, -0.00003129, -0.00002253, -0.00001233, -0.00000111, 0.00001059, 0.00002218, 0.00003303, 0.00004253, 0.00005013, 0.00005537, 0.00005791, 0.00005758, 0.00005438, 0.00004846, 0.00004017, 0.00002996, 0.00001843, 0.00000623, -0.00000594, -0.00001741, -0.00002758, -0.00003591, -0.00004202, -0.00004567, -0.00004679, -0.00004546, -0.00004190, -0.00003647, -0.00002961, -0.00002183, -0.00001366, -0.00000559, 0.00000192, 0.00000850, 0.00001388, 0.00001789, 0.00002048, 0.00002170, 0.00002167, 0.00002059, 0.00001870, 0.00001624, 0.00001342, 0.00001043, 0.00000742, 0.00000444, 0.00000152, -0.00000138, },
		{ -0.00001916, -0.00001867, -0.00001775, -0.00001626, -0.00001411, -0.00001126, -0.00000772, -0.00000358, 0.00000099, 0.00000578, 0.00001050, 0.00001484, 0.00001850, 0.00002116, 0.00002258, 0.00002256, 0.00002102, 0.00001795, 0.00001347, 0.00000780, 0.00000127, -0.00000574, -0.00001277, -0.00001936, -0.00002505, -0.00002942, -0.00003213, -0.00003294, -0.00003174, -0.00002853, -0.00002345, -0.00001677, -0.00000886, -0.00000015, 0.00000885, 0.00001764, 0.00002572, 0.00003266, 0.00003809, 0.00004177, 0.00004355, 0.00004341, 0.00004146, 0.00003788, 0.00003298, 0.00002710, 0.00002062, 0.00001392, 0.00000735, 0.00000123, -0.00000422, -0.00000883, -0.00001252, -0.00001530, -0.00001724, -0.00001847, -0.00001914, -0.00001942, -0.00001947, -0.00001943, -0.00001939, -0.00001938, -0.00001940, -0.00001936, },
		{ 0.00000292, 0.00000362, 0.00000424, 0.00000466, 0.00000479, 0.00000454, 0.00000387, 0.00000278, 0.00000133, -0.00000038, -0.00000220, -0.00000397, -0.00000547, -0.00000652, -0.00000694, -0.00000658, -0.00000535, -0.00000322, -0.00000023, 0.00000351, 0.00000782, 0.00001246, 0.00001715, 0.00002157, 0.00002542, 0.00002840, 0.00003026, 0.00003082, 0.00002996, 0.00002765, 0.00002396, 0.00001904, 0.00001313, 0.00000651, -0.00000048, -0.00000748, -0.00001412, -0.00002009, -0.00002510, -0.00002894, -0.00003146, -0.00003263, -0.00003248, -0.00003111, -0.00002871, -0.00002551, -0.00002177, -0.00001776, -0.00001374, -0.00000996, -0.00000658, -0.00000375, -0.00000154, 0.00000004, 0.00000103, 0.00000153, 0.00000166, 0.00000156, 0.00000136, 0.00000120, 0.00000117, 0.00000132, 0.00000169, 0.00000224, },
		{ -0.00001191, -0.00001135, -0.00001070, -0.00000992, -0.00000899, -0.00000789, -0.00000660, -0.00000516, -0.00000361, -0.00000202, -0.00000047, 0.00000092, 0.00000204, 0.00000278, 0.00000306, 0.00000278, 0.00000190, 0.00000043, -0.00000162, -0.00000415, -0.00000705, -0.00001016, -0.00001332, -0.00001632, -0.00001898, -0.00002110, -0.00002253, -0.00002316, -0.00002290, -0.00002174, -0.00001971, -0.00001691, -0.00001346, -0.00000954, -0.00000538, -0.00000117, 0.00000284, 0.00000645, 0.00000948, 0.00001177, 0.00001325, 0.00001384, 0.00001358, 0.00001250, 0.00001071, 0.00000835, 0.00000557, 0.00000255, -0.00000054, -0.00000354, -0.00000632, -0.00000877, -0.00001081, -0.00001241, -0.00001356, -0.00001430, -0.00001467, -0.00001474, -0.00001458, -0.00001426, -0.00001385, -0.00001340, -0.00001292, -0.00001243, },
		{ 0.00000598, 0.00000651, 0.00000679, 0.00000680, 0.00000655, 0.00000606, 0.00000535, 0.00000447, 0.00000348, 0.00000245, 0.00000144, 0.00000054, -0.00000021, -0.00000073, -0.00000100, -0.00000099, -0.00000071, -0.00000017, 0.00000057, 0.00000146, 0.00000240, 0.00000332, 0.00000412, 0.00000471, 0.00000501, 0.00000495, 0.00000449, 0.00000360, 0.00000228, 0.00000057, -0.00000149, -0.00000382, -0.00000633, -0.00000894, -0.00001152, -0.00001399, -0.00001624, -0.00001820, -0.00001980, -0.00002101, -0.00002180, -0.00002216, -0.00002212, -0.00002171, -0.00002097, -0.00001997, -0.00001875, -0.00001738, -0.00001591, -0.00001437, -0.00001280, -0.00001124, -0.00000968, -0.00000815, -0.00000663, -0.00000514, -0.00000366, -0.00000221, -0.00000078, 0.00000061, 0.00000193, 0.00000317, 0.00000428, 0.00000523, },
		{ 0.00000247, 0.00000124, 0.00000019, -0.00000066, -0.00000132, -0.00000182, -0.00000221, -0.00000255, -0.00000290, -0.00000332, -0.00000387, -0.00000458, -0.00000547, -0.00000653, -0.00000774, -0.00000904, -0.00001036, -0.00001160, -0.00001267, -0.00001348, -0.00001393, -0.00001395, -0.00001349, -0.00001253, -0.00001107, -0.00000916, -0.00000686, -0.00000428, -0.00000152, 0.00000130, 0.00000403, 0.00000656, 0.00000879, 0.00001062, 0.00001200, 0.00001290, 0.00001334, 0.00001334, 0.00001298, 0.00001235, 0.00001153, 0.00001065, 0.00000980, 0.00000908, 0.00000855, 0.00000827, 0.00000825, 0.00000850, 0.00000897, 0.00000961, 0.00001035, 0.00001110, 0.00001177, 0.00001229, 0.00001258, 0.00001260, 0.00001230, 0.00001170, 0.00001080, 0.00000964, 0.00000830, 0.00000683, 0.00000533, 0.00000385, },
		{ -0.00000491, -0.00000428, -0.00000377, -0.00000340, -0.00000313, -0.00000290, -0.00000264, -0.00000225, -0.00000164, -0.00000071, 0.00000061, 0.00000235, 0.00000452, 0.00000708, 0.00000995, 0.00001301, 0.00001610, 0.00001906, 0.00002169, 0.00002383, 0.00002529, 0.00002597, 0.00002576, 0.00002464, 0.00002262, 0.00001980, 0.00001631, 0.00001232, 0.00000805, 0.00000374, -0.00000038, -0.00000409, -0.00000719, -0.00000952, -0.00001100, -0.00001157, -0.00001127, -0.00001017, -0.00000840, -0.00000614, -0.00000358, -0.00000094, 0.00000158, 0.00000379, 0.00000554, 0.00000671, 0.00000724, 0.00000712, 0.00000638, 0.00000513, 0.00000346, 0.00000153, -0.00000051, -0.00000251, -0.00000433, -0.00000586, -0.00000702, -0.00000777, -0.00000811, -0.00000808, -0.00000772, -0.00000714, -0.00000642, -0.00000565, },
		{ 0.00000219, 0.00000239, 0.00000284, 0.00000355, 0.00000448, 0.00000556, 0.00000667, 0.00000767, 0.00000841, 0.00000875, 0.00000854, 0.00000770, 0.00000615, 0.00000391, 0.00000103, -0.00000236, -0.00000611, -0.00000998, -0.00001375, -0.00001716, -0.00001996, -0.00002194, -0.00002294, -0.00002285, -0.00002163, -0.00001931, -0.00001603, -0.00001196, -0.00000734, -0.00000245, 0.00000239, 0.00000690, 0.00001079, 0.00001382, 0.00001582, 0.00001669, 0.00001639, 0.00001498, 0.00001259, 0.00000939, 0.00000564, 0.00000159, -0.00000247, -0.00000630, -0.00000964, -0.00001231, -0.00001419, -0.00001521, -0.00001537, -0.00001474, -0.00001342, -0.00001159, -0.00000942, -0.00000710, -0.00000481, -0.00000272, -0.00000093, 0.00000047, 0.00000145, 0.00000205, 0.00000232, 0.00000235, 0.00000226, 0.00000217, },
		{ -0.00000135, -0.00000068, -0.00000032, -0.00000035, -0.00000077, -0.00000155, -0.00000261, -0.00000382, -0.00000501, -0.00000600, -0.00000662, -0.00000671, -0.00000615, -0.00000487, -0.00000287, -0.00000021, 0.00000299, 0.00000654, 0.00001021, 0.00001372, 0.00001680, 0.00001920, 0.00002068, 0.00002105, 0.00002022, 0.00001816, 0.00001491, 0.00001063, 0.00000553, -0.00000011, -0.00000598, -0.00001172, -0.00001701, -0.00002153, -0.00002504, -0.00002733, -0.00002831, -0.00002795, -0.00002633, -0.00002360, -0.00001998, -0.00001574, -0.00001118, -0.00000661, -0.00000233, 0.00000141, 0.00000442, 0.00000656, 0.00000777, 0.00000807, 0.00000756, 0.00000636, 0.00000467, 0.00000271, 0.00000067, -0.00000123, -0.00000284, -0.00000402, -0.00000471, -0.00000491, -0.00000465, -0.00000403, -0.00000317, -0.00000223, },
		{ 0.00000962, 0.00000903, 0.00000854, 0.00000827, 0.00000829, 0.00000862, 0.00000923, 0.00001003, 0.00001090, 0.00001167, 0.00001218, 0.00001224, 0.00001171, 0.00001046, 0.00000843, 0.00000562, 0.00000211, -0.00000197, -0.00000642, -0.00001099, -0.00001540, -0.00001935, -0.00002257, -0.00002481, -0.00002589, -0.00002567, -0.00002414, -0.00002134, -0.00001741, -0.00001258, -0.00000713, -0.00000139, 0.00000428, 0.00000954, 0.00001407, 0.00001760, 0.00001993, 0.00002095, 0.00002064, 0.00001907, 0.00001641, 0.00001287, 0.00000875, 0.00000435, -0.00000000, -0.00000401, -0.00000739, -0.00000995, -0.00001156, -0.00001215, -0.00001174, -0.00001042, -0.00000837, -0.00000576, -0.00000284, 0.00000017, 0.00000304, 0.00000558, 0.00000767, 0.00000920, 0.00001016, 0.00001058, 0.00001055, 0.00001018, },
		{ 0.00000146, 0.00000117, 0.00000081, 0.00000025, -0.00000059, -0.00000177, -0.00000331, -0.00000516, -0.00000722, -0.00000937, -0.00001144, -0.00001323, -0.00001456, -0.00001525, -0.00001519, -0.00001428, -0.00001251, -0.00000993, -0.00000666, -0.00000288, 0.00000117, 0.00000522, 0.00000898, 0.00001217, 0.00001453, 0.00001587, 0.00001603, 0.00001497, 0.00001271, 0.00000936, 0.00000512, 0.00000026, -0.00000492, -0.00001007, -0.00001486, -0.00001897, -0.00002211, -0.00002408, -0.00002475, -0.00002407, -0.00002210, -0.00001896, -0.00001487, -0.00001008, -0.00000491, 0.00000034, 0.00000536, 0.00000986, 0.00001363, 0.00001650, 0.00001838, 0.00001925, 0.00001919, 0.00001829, 0.00001674, 0.00001472, 0.00001245, 0.00001013, 0.00000793, 0.00000600, 0.00000441, 0.00000321, 0.00000237, 0.00000182, },
		{ -0.00000565, -0.00000652, -0.00000741, -0.00000819, -0.00000871, -0.00000889, -0.00000863, -0.00000792, -0.00000677, -0.00000526, -0.00000351, -0.00000168, 0.00000006, 0.00000150, 0.00000248, 0.00000285, 0.00000248, 0.00000132, -0.00000061, -0.00000323, -0.00000642, -0.00000997, -0.00001363, -0.00001715, -0.00002025, -0.00002267, -0.00002419, -0.00002463, -0.00002391, -0.00002199, -0.00001894, -0.00001488, -0.00001004, -0.00000467, 0.00000092, 0.00000642, 0.00001152, 0.00001592, 0.00001940, 0.00002178, 0.00002297, 0.00002295, 0.00002178, 0.00001960, 0.00001660, 0.00001302, 0.00000912, 0.00000517, 0.00000142, -0.00000193, -0.00000470, -0.00000680, -0.00000818, -0.00000887, -0.00000893, -0.00000851, -0.00000775, -0.00000681, -0.00000588, -0.00000508, -0.00000454, -0.00000433, -0.00000447, -0.00000493, },
		{ -0.00000291, -0.00000271, -0.00000227, -0.00000171, -0.00000116, -0.00000076, -0.00000062, -0.00000085, -0.00000150, -0.00000258, -0.00000405, -0.00000584, -0.00000780, -0.00000977, -0.00001156, -0.00001300, -0.00001389, -0.00001410, -0.00001352, -0.00001210, -0.00000985, -0.00000685, -0.00000324, 0.00000079, 0.00000502, 0.00000919, 0.00001303, 0.00001632, 0.00001883, 0.00002041, 0.00002095, 0.00002043, 0.00001889, 0.00001644, 0.00001327, 0.00000958, 0.00000563, 0.00000169, -0.00000198, -0.00000515, -0.00000764, -0.00000930, -0.00001008, -0.00000998, -0.00000906, -0.00000744, -0.00000530, -0.00000284, -0.00000028, 0.00000218, 0.00000434, 0.00000605, 0.00000719, 0.00000773, 0.00000765, 0.00000701, 0.00000592, 0.00000450, 0.00000291, 0.00000131, -0.00000017, -0.00000139, -0.00000228, -0.00000278, },
		},
		{ { 0.00001096, 0.00001138, 0.00001249, 0.00001393, 0.00001514, 0.00001551, 0.00001446, 0.00001158, 0.00000669, -0.00000006, -0.00000822, -0.00001706, -0.00002568, -0.00003310, -0.00003844, -0.00004100, -0.00004041, -0.00003667, -0.00003016, -0.00002160, -0.00001197, -0.00000235, 0.00000618, 0.00001275, 0.00001676, 0.00001798, 0.00001659, 0.00001312, 0.00000836, 0.00000327, -0.00000120, -0.00000424, -0.00000532, -0.00000424, -0.00000120, 0.00000327, 0.00000836, 0.00001312, 0.00001659, 0.00001798, 0.00001676, 0.00001275, 0.00000618, -0.00000235, -0.00001197, -0.00002160, -0.00003016, -0.00003667, -0.00004041, -0.00004100, -0.00003844, -0.00003310, -0.00002568, -0.00001706, -0.00000822, -0.00000006, 0.00000669, 0.00001158, 0.00001446, 0.00001551, 0.00001514, 0.00001393, 0.00001249, 0.00001138, },
		{ 0.00000793, 0.00000371, -0.00000218, -0.00000933, -0.00001709, -0.00002464, -0.00003111, -0.00003571, -0.00003781, -0.00003703, -0.00003331, -0.00002690, -0.00001838, -0.00000856, 0.00000158, 0.00001101, 0.00001875, 0.00002403, 0.00002634, 0.00002552, 0.00002180, 0.00001575, 0.00000823, 0.00000029, -0.00000700, -0.00001265, -0.00001596, -0.00001654, -0.00001445, -0.00001017, -0.00000449, 0.00000153, 0.00000678, 0.00001026, 0.00001124, 0.00000938, 0.00000480, -0.00000196, -0.00000998, -0.00001812, -0.00002523, -0.00003028, -0.00003255, -0.00003172, -0.00002793, -0.00002172, -0.00001396, -0.00000566, 0.00000214, 0.00000858, 0.00001310, 0.00001549, 0.00001587, 0.00001470, 0.00001262, 0.00001032, 0.00000843, 0.00000737, 0.00000730, 0.00000809, 0.00000937, 0.00001057, 0.00001108, 0.00001034, },
		{ -0.00002534, -0.00002713, -0.00002647, -0.00002339, -0.00001821, -0.00001146, -0.00000385, 0.00000379, 0.00001066, 0.00001598, 0.00001914, 0.00001972, 0.00001758, 0.00001286, 0.00000603, -0.00000218, -0.00001083, -0.00001885, -0.00002519, -0.00002895, -0.00002953, -0.00002672, -0.00002075, -0.00001234, -0.00000255, 0.00000731, 0.00001588, 0.00002195, 0.00002467, 0.00002366, 0.00001909, 0.00001168, 0.00000253, -0.00000699, -0.00001552, -0.00002187, -0.00002527, -0.00002540, -0.00002252, -0.00001729, -0.00001072, -0.00000393, 0.00000205, 0.00000643, 0.00000882, 0.00000924, 0.00000808, 0.00000602, 0.00000382, 0.00000221, 0.00000170, 0.00000248, 0.00000443, 0.00000710, 0.00000983, 0.00001191, 0.00001266, 0.00001164, 0.00000866, 0.00000388, -0.00000225, -0.00000904, -0.00001568, -0.00002135, },
		{ -0.00000191, 0.00000202, 0.00000452, 0.00000530, 0.00000427, 0.00000153, -0.00000266, -0.00000789, -0.00001360, -0.00001914, -0.00002378, -0.00002681, -0.00002760, -0.00002568, -0.00002088, -0.00001337, -0.00000367, 0.00000729, 0.00001835, 0.00002819, 0.00003554, 0.00003939, 0.00003912, 0.00003466, 0.00002651, 0.00001566, 0.00000350, -0.00000843, -0.00001866, -0.00002604, -0.00002990, -0.00003011, -0.00002712, -0.00002183, -0.00001536, -0.00000891, -0.00000346, 0.00000036, 0.00000236, 0.00000279, 0.00000227, 0.00000157, 0.00000147, 0.00000254, 0.00000504, 0.00000881, 0.00001333, 0.00001782, 0.00002139, 0.00002324, 0.00002279, 0.00001983, 0.00001457, 0.00000757, -0.00000032, -0.00000812, -0.00001488, -0.00001983, -0.00002250, -0.00002273, -0.00002073, -0.00001695, -0.00001207, -0.00000682, },
		{ -0.00001188, -0.00001446, -0.00001614, -0.00001666, -0.00001582, -0.00001350, -0.00000967, -0.00000436, 0.00000223, 0.00000976, 0.00001772, 0.00002540, 0.00003197, 0.00003658, 0.00003847, 0.00003711, 0.00003232, 0.00002438, 0.00001399, 0.00000221, -0.00000964, -0.00002023, -0.00002838, -0.00003328, -0.00003459, -0.00003252, -0.00002773, -0.00002121, -0.00001410, -0.00000742, -0.00000195, 0.00000196, 0.00000438, 0.00000577, 0.00000680, 0.00000822, 0.00001057, 0.00001411, 0.00001869, 0.00002373, 0.00002840, 0.00003171, 0.00003276, 0.00003093, 0.00002603, 0.00001837, 0.00000871, -0.00000184, -0.00001201, -0.00002061, -0.00002668, -0.00002970, -0.00002961, -0.00002682, -0.00002209, -0.00001638, -0.00001069, -0.00000588, -0.00000257, -0.00000103, -0.00000125, -0.00000293, -0.00000562, -0.00000878, },
		{ 0.00000586, 0.00001032, 0.00001433, 0.00001773, 0.00002040, 0.00002220, 0.00002297, 0.00002248, 0.00002053, 0.00001699, 0.00001189, 0.00000546, -0.00000182, -0.00000929, -0.00001615, -0.00002162, -0.00002505, -0.00002606, -0.00002462, -0.00002103, -0.00001589, -0.00000997, -0.00000407, 0.00000117, 0.00000537, 0.00000848, 0.00001074, 0.00001262, 0.00001463, 0.00001721, 0.00002052, 0.00002438, 0.00002825, 0.00003135, 0.00003275, 0.00003163, 0.00002747, 0.00002019, 0.00001025, -0.00000139, -0.00001341, -0.00002431, -0.00003271, -0.00003755, -0.00003832, -0.00003511, -0.00002862, -0.00002002, -0.00001071, -0.00000214, 0.00000451, 0.00000847, 0.00000949, 0.00000783, 0.00000420, -0.00000046, -0.00000513, -0.00000892, -0.00001118, -0.00001161, -0.00001024, -0.00000734, -0.00000337, 0.00000118, },
		{ 0.00000626, 0.00000430, 0.00000272, 0.00000145, 0.00000033, -0.00000083, -0.00000216, -0.00000363, -0.00000512, -0.00000640, -0.00000721, -0.00000732, -0.00000664, -0.00000520, -0.00000316, -0.00000077, 0.00000171, 0.00000408, 0.00000625, 0.00000828, 0.00001031, 0.00001255, 0.00001511, 0.00001797, 0.00002087, 0.00002333, 0.00002473, 0.00002436, 0.00002165, 0.00001630, 0.00000842, -0.00000140, -0.00001213, -0.00002242, -0.00003082, -0.00003600, -0.00003705, -0.00003361, -0.00002604, -0.00001535, -0.00000305, 0.00000906, 0.00001920, 0.00002594, 0.00002841, 0.00002645, 0.00002061, 0.00001207, 0.00000233, -0.00000695, -0.00001434, -0.00001884, -0.00001998, -0.00001793, -0.00001334, -0.00000722, -0.00000071, 0.00000517, 0.00000964, 0.00001230, 0.00001314, 0.00001244, 0.00001071, 0.00000850, },
		{ 0.00000726, 0.00000953, 0.00001041, 0.00000985, 0.00000805, 0.00000541, 0.00000242, -0.00000042, -0.00000271, -0.00000414, -0.00000456, -0.00000393, -0.00000231, 0.00000017, 0.00000329, 0.00000680, 0.00001037, 0.00001357, 0.00001595, 0.00001704, 0.00001640, 0.00001379, 0.00000918, 0.00000287, -0.00000450, -0.00001198, -0.00001843, -0.00002270, -0.00002381, -0.00002119, -0.00001479, -0.00000522, 0.00000637, 0.00001833, 0.00002886, 0.00003623, 0.00003911, 0.00003682, 0.00002946, 0.00001794, 0.00000386, -0.00001080, -0.00002393, -0.00003372, -0.00003887, -0.00003889, -0.00003407, -0.00002546, -0.00001463, -0.00000340, 0.00000652, 0.00001378, 0.00001761, 0.00001791, 0.00001520, 0.00001048, 0.00000495, -0.00000015, -0.00000386, -0.00000559, -0.00000521, -0.00000305, 0.00000028, 0.00000398, },
		{ 0.00000064, -0.00000536, -0.00001015, -0.00001310, -0.00001388, -0.00001251, -0.00000927, -0.00000467, 0.00000066, 0.00000603, 0.00001078, 0.00001435, 0.00001632, 0.00001647, 0.00001478, 0.00001148, 0.00000704, 0.00000211, -0.00000250, -0.00000596, -0.00000750, -0.00000661, -0.00000309, 0.00000281, 0.00001039, 0.00001855, 0.00002591, 0.00003103, 0.00003265, 0.00002994, 0.00002264, 0.00001125, -0.00000306, -0.00001854, -0.00003314, -0.00004477, -0.00005173, -0.00005290, -0.00004803, -0.00003774, -0.00002349, -0.00000730, 0.00000853, 0.00002184, 0.00003090, 0.00003472, 0.00003318, 0.00002699, 0.00001758, 0.00000676, -0.00000352, -0.00001160, -0.00001629, -0.00001709, -0.00001420, -0.00000846, -0.00000112, 0.00000640, 0.00001273, 0.00001685, 0.00001817, 0.00001662, 0.00001263, 0.00000698, },
		{ -0.00000647, 0.00000005, 0.00000680, 0.00001284, 0.00001742, 0.00002005, 0.00002054, 0.00001902, 0.00001589, 0.00001175, 0.00000733, 0.00000336, 0.00000049, -0.00000081, -0.00000033, 0.00000183, 0.00000525, 0.00000923, 0.00001289, 0.00001524, 0.00001542, 0.00001277, 0.00000704, -0.00000153, -0.00001220, -0.00002375, -0.00003466, -0.00004329, -0.00004812, -0.00004806, -0.00004263, -0.00003208, -0.00001744, -0.00000039, 0.00001694, 0.00003229, 0.00004361, 0.00004938, 0.00004884, 0.00004215, 0.00003032, 0.00001513, -0.00000121, -0.00001637, -0.00002830, -0.00003548, -0.00003718, -0.00003353, -0.00002544, -0.00001445, -0.00000245, 0.00000868, 0.00001732, 0.00002236, 0.00002335, 0.00002052, 0.00001467, 0.00000701, -0.00000105, -0.00000817, -0.00001326, -0.00001562, -0.00001506, -0.00001180, },
		{ 0.00001556, 0.00001153, 0.00000655, 0.00000145, -0.00000300, -0.00000617, -0.00000771, -0.00000755, -0.00000594, -0.00000343, -0.00000071, 0.00000144, 0.00000232, 0.00000143, -0.00000146, -0.00000617, -0.00001221, -0.00001871, -0.00002462, -0.00002881, -0.00003024, -0.00002819, -0.00002234, -0.00001289, -0.00000061, 0.00001324, 0.00002704, 0.00003904, 0.00004757, 0.00005130, 0.00004947, 0.00004198, 0.00002950, 0.00001338, -0.00000452, -0.00002207, -0.00003716, -0.00004798, -0.00005330, -0.00005261, -0.00004622, -0.00003516, -0.00002105, -0.00000582, 0.00000853, 0.00002026, 0.00002810, 0.00003140, 0.00003023, 0.00002527, 0.00001769, 0.00000897, 0.00000058, -0.00000616, -0.00001033, -0.00001148, -0.00000969, -0.00000549, 0.00000023, 0.00000641, 0.00001201, 0.00001613, 0.00001818, 0.00001793, },
		{ -0.00001366, -0.00000990, -0.00000525, -0.00000050, 0.00000351, 0.00000610, 0.00000683, 0.00000563, 0.00000279, -0.00000102, -0.00000493, -0.00000795, -0.00000917, -0.00000794, -0.00000399, 0.00000249, 0.00001087, 0.00002011, 0.00002894, 0.00003602, 0.00004013, 0.00004037, 0.00003632, 0.00002807, 0.00001628, 0.00000207, -0.00001311, -0.00002763, -0.00003994, -0.00004872, -0.00005306, -0.00005256, -0.00004739, -0.00003819, -0.00002609, -0.00001245, 0.00000124, 0.00001356, 0.00002335, 0.00002979, 0.00003253, 0.00003167, 0.00002773, 0.00002156, 0.00001423, 0.00000683, 0.00000037, -0.00000441, -0.00000706, -0.00000753, -0.00000611, -0.00000339, -0.00000014, 0.00000283, 0.00000481, 0.00000527, 0.00000403, 0.00000120, -0.00000279, -0.00000727, -0.00001147, -0.00001465, -0.00001623, -0.00001589, },
		{ 0.00002528, 0.00002128, 0.00001578, 0.00000979, 0.00000442, 0.00000066, -0.00000075, 0.00000046, 0.00000405, 0.00000930, 0.00001513, 0.00002026, 0.00002347, 0.00002381, 0.00002080, 0.00001448, 0.00000545, -0.00000522, -0.00001621, -0.00002614, -0.00003378, -0.00003827, -0.00003921, -0.00003668, -0.00003124, -0.00002375, -0.00001523, -0.00000670, 0.00000099, 0.00000724, 0.00001174, 0.00001447, 0.00001560, 0.00001546, 0.00001439, 0.00001272, 0.00001068, 0.00000845, 0.00000611, 0.00000372, 0.00000131, -0.00000107, -0.00000336, -0.00000548, -0.00000740, -0.00000910, -0.00001061, -0.00001200, -0.00001333, -0.00001463, -0.00001586, -0.00001686, -0.00001737, -0.00001707, -0.00001561, -0.00001273, -0.00000835, -0.00000259, 0.00000414, 0.00001119, 0.00001777, 0.00002304, 0.00002629, 0.00002706, },
		{ -0.00001713, -0.00001510, -0.00001034, -0.00000394, 0.00000273, 0.00000827, 0.00001149, 0.00001169, 0.00000879, 0.00000331, -0.00000369, -0.00001087, -0.00001686, -0.00002056, -0.00002136, -0.00001927, -0.00001489, -0.00000930, -0.00000376, 0.00000049, 0.00000257, 0.00000209, -0.00000071, -0.00000508, -0.00000989, -0.00001390, -0.00001605, -0.00001568, -0.00001267, -0.00000749, -0.00000110, 0.00000526, 0.00001029, 0.00001297, 0.00001266, 0.00000931, 0.00000339, -0.00000415, -0.00001214, -0.00001935, -0.00002473, -0.00002759, -0.00002762, -0.00002493, -0.00001994, -0.00001327, -0.00000557, 0.00000257, 0.00001067, 0.00001840, 0.00002550, 0.00003173, 0.00003682, 0.00004042, 0.00004211, 0.00004151, 0.00003837, 0.00003269, 0.00002479, 0.00001537, 0.00000544, -0.00000383, -0.00001124, -0.00001585, },
		{ 0.00002059, 0.00002198, 0.00001959, 0.00001420, 0.00000710, -0.00000019, -0.00000624, -0.00001003, -0.00001122, -0.00001016, -0.00000780, -0.00000549, -0.00000457, -0.00000608, -0.00001039, -0.00001715, -0.00002521, -0.00003292, -0.00003843, -0.00004014, -0.00003703, -0.00002899, -0.00001687, -0.00000239, 0.00001216, 0.00002437, 0.00003216, 0.00003420, 0.00003017, 0.00002086, 0.00000799, -0.00000609, -0.00001884, -0.00002800, -0.00003198, -0.00003010, -0.00002269, -0.00001094, 0.00000338, 0.00001827, 0.00003184, 0.00004264, 0.00004981, 0.00005314, 0.00005293, 0.00004980, 0.00004452, 0.00003776, 0.00003001, 0.00002153, 0.00001248, 0.00000299, -0.00000669, -0.00001605, -0.00002438, -0.00003078, -0.00003437, -0.00003449, -0.00003088, -0.00002385, -0.00001426, -0.00000346, 0.00000693, 0.00001536, },
		{ -0.00001435, -0.00002065, -0.00002360, -0.00002357, -0.00002161, -0.00001919, -0.00001777, -0.00001843, -0.00002148, -0.00002641, -0.00003191, -0.00003617, -0.00003732, -0.00003391, -0.00002532, -0.00001201, 0.00000448, 0.00002173, 0.00003694, 0.00004745, 0.00005132, 0.00004780, 0.00003748, 0.00002228, 0.00000510, -0.00001077, -0.00002222, -0.00002696, -0.00002396, -0.00001364, 0.00000224, 0.00002094, 0.00003926, 0.00005416, 0.00006333, 0.00006553, 0.00006077, 0.00005018, 0.00003570, 0.00001962, 0.00000412, -0.00000915, -0.00001925, -0.00002605, -0.00003001, -0.00003195, -0.00003270, -0.00003285, -0.00003254, -0.00003148, -0.00002910, -0.00002477, -0.00001811, -0.00000918, 0.00000134, 0.00001228, 0.00002210, 0.00002925, 0.00003249, 0.00003122, 0.00002560, 0.00001658, 0.00000572, -0.00000516, },
		{ -0.00000893, -0.00000479, -0.00000370, -0.00000544, -0.00000886, -0.00001220, -0.00001352, -0.00001126, -0.00000466, 0.00000593, 0.00001907, 0.00003250, 0.00004360, 0.00005002, 0.00005023, 0.00004394, 0.00003227, 0.00001752, 0.00000278, -0.00000874, -0.00001443, -0.00001285, -0.00000407, 0.00001025, 0.00002719, 0.00004317, 0.00005466, 0.00005890, 0.00005450, 0.00004172, 0.00002239, -0.00000044, -0.00002317, -0.00004233, -0.00005525, -0.00006060, -0.00005847, -0.00005033, -0.00003856, -0.00002587, -0.00001473, -0.00000680, -0.00000268, -0.00000190, -0.00000314, -0.00000469, -0.00000494, -0.00000285, 0.00000175, 0.00000815, 0.00001492, 0.00002029, 0.00002263, 0.00002089, 0.00001489, 0.00000543, -0.00000589, -0.00001704, -0.00002598, -0.00003120, -0.00003202, -0.00002876, -0.00002262, -0.00001538, },
		{ -0.00000234, -0.00000615, -0.00000557, -0.00000029, 0.00000879, 0.00001976, 0.00003016, 0.00003762, 0.00004042, 0.00003797, 0.00003093, 0.00002114, 0.00001107, 0.00000331, -0.00000019, 0.00000139, 0.00000748, 0.00001616, 0.00002460, 0.00002967, 0.00002877, 0.00002043, 0.00000481, -0.00001625, -0.00003949, -0.00006091, -0.00007657, -0.00008344, -0.00008006, -0.00006691, -0.00004634, -0.00002208, 0.00000145, 0.00002015, 0.00003097, 0.00003257, 0.00002554, 0.00001218, -0.00000400, -0.00001914, -0.00002983, -0.00003386, -0.00003063, -0.00002123, -0.00000808, 0.00000565, 0.00001682, 0.00002305, 0.00002323, 0.00001770, 0.00000817, -0.00000281, -0.00001242, -0.00001835, -0.00001923, -0.00001502, -0.00000693, 0.00000294, 0.00001212, 0.00001839, 0.00002040, 0.00001793, 0.00001197, 0.00000443, },
		{ 0.00001859, 0.00003088, 0.00003907, 0.00004167, 0.00003847, 0.00003057, 0.00001999, 0.00000908, -0.00000010, -0.00000635, -0.00000963, -0.00001102, -0.00001238, -0.00001572, -0.00002261, -0.00003353, -0.00004760, -0.00006262, -0.00007545, -0.00008276, -0.00008181, -0.00007127, -0.00005165, -0.00002545, 0.00000324, 0.00002943, 0.00004828, 0.00005612, 0.00005127, 0.00003449, 0.00000894, -0.00002042, -0.00004777, -0.00006762, -0.00007590, -0.00007079, -0.00005311, -0.00002611, 0.00000521, 0.00003511, 0.00005834, 0.00007113, 0.00007192, 0.00006161, 0.00004319, 0.00002107, 0.00000003, -0.00001580, -0.00002379, -0.00002328, -0.00001557, -0.00000352, 0.00000925, 0.00001928, 0.00002402, 0.00002247, 0.00001530, 0.00000465, -0.00000651, -0.00001509, -0.00001867, -0.00001610, -0.00000771, 0.00000475, },
		{ 0.00001173, -0.00000612, -0.00002333, -0.00003762, -0.00004777, -0.00005377, -0.00005662, -0.00005784, -0.00005882, -0.00006035, -0.00006220, -0.00006320, -0.00006151, -0.00005525, -0.00004317, -0.00002527, -0.00000313, 0.00002017, 0.00004057, 0.00005379, 0.00005638, 0.00004662, 0.00002515, -0.00000489, -0.00003829, -0.00006859, -0.00008944, -0.00009581, -0.00008520, -0.00005823, -0.00001869, 0.00002704, 0.00007119, 0.00010609, 0.00012562, 0.00012645, 0.00010866, 0.00007572, 0.00003376, -0.00000964, -0.00004698, -0.00007230, -0.00008222, -0.00007654, -0.00005805, -0.00003181, -0.00000395, 0.00001965, 0.00003461, 0.00003882, 0.00003279, 0.00001926, 0.00000249, -0.00001286, -0.00002282, -0.00002497, -0.00001894, -0.00000637, 0.00000958, 0.00002504, 0.00003637, 0.00004093, 0.00003769, 0.00002728, },
		{ -0.00007422, -0.00006621, -0.00005584, -0.00004514, -0.00003534, -0.00002671, -0.00001864, -0.00001014, -0.00000042, 0.00001063, 0.00002204, 0.00003184, 0.00003743, 0.00003629, 0.00002679, 0.00000888, -0.00001546, -0.00004233, -0.00006644, -0.00008214, -0.00008463, -0.00007118, -0.00004196, -0.00000039, 0.00004722, 0.00009274, 0.00012765, 0.00014475, 0.00013966, 0.00011187, 0.00006498, 0.00000621, -0.00005488, -0.00010815, -0.00014478, -0.00015895, -0.00014891, -0.00011725, -0.00007035, -0.00001705, 0.00003308, 0.00007163, 0.00009287, 0.00009473, 0.00007898, 0.00005063, 0.00001667, -0.00001545, -0.00003939, -0.00005117, -0.00004983, -0.00003735, -0.00001801, 0.00000270, 0.00001942, 0.00002805, 0.00002660, 0.00001541, -0.00000312, -0.00002520, -0.00004664, -0.00006382, -0.00007437, -0.00007757, },
		{ 0.00001623, 0.00001685, 0.00001535, 0.00001317, 0.00001103, 0.00000873, 0.00000528, -0.00000057, -0.00000974, -0.00002215, -0.00003639, -0.00004971, -0.00005854, -0.00005925, -0.00004924, -0.00002784, 0.00000311, 0.00003916, 0.00007385, 0.00009986, 0.00011044, 0.00010101, 0.00007040, 0.00002142, -0.00003925, -0.00010204, -0.00015623, -0.00019187, -0.00020180, -0.00018314, -0.00013804, -0.00007336, 0.00000042, 0.00007118, 0.00012737, 0.00016015, 0.00016503, 0.00014256, 0.00009816, 0.00004086, -0.00001849, -0.00006937, -0.00010362, -0.00011683, -0.00010896, -0.00008404, -0.00004908, -0.00001245, 0.00001798, 0.00003634, 0.00003984, 0.00002906, 0.00000755, -0.00001924, -0.00004524, -0.00006517, -0.00007555, -0.00007525, -0.00006546, -0.00004913, -0.00003005, -0.00001193, 0.00000248, 0.00001183, },
		{ -0.00003617, -0.00003732, -0.00003731, -0.00003747, -0.00003829, -0.00003913, -0.00003842, -0.00003425, -0.00002508, -0.00001062, 0.00000774, 0.00002677, 0.00004192, 0.00004825, 0.00004165, 0.00002007, -0.00001554, -0.00006094, -0.00010898, -0.00015080, -0.00017744, -0.00018177, -0.00016013, -0.00011346, -0.00004747, 0.00002808, 0.00010105, 0.00015908, 0.00019199, 0.00019377, 0.00016382, 0.00010719, 0.00003368, -0.00004397, -0.00011248, -0.00016055, -0.00018089, -0.00017150, -0.00013591, -0.00008223, -0.00002144, 0.00003487, 0.00007676, 0.00009780, 0.00009603, 0.00007409, 0.00003830, -0.00000286, -0.00004070, -0.00006804, -0.00008060, -0.00007760, -0.00006152, -0.00003730, -0.00001096, 0.00001177, 0.00002671, 0.00003192, 0.00002782, 0.00001676, 0.00000218, -0.00001243, -0.00002435, -0.00003224, },
		{ -0.00001447, -0.00001328, -0.00001277, -0.00001122, -0.00000799, -0.00000383, -0.00000082, -0.00000175, -0.00000916, -0.00002439, -0.00004671, -0.00007308, -0.00009840, -0.00011644, -0.00012128, -0.00010874, -0.00007772, -0.00003093, 0.00002519, 0.00008142, 0.00012742, 0.00015376, 0.00015402, 0.00012627, 0.00007386, 0.00000502, -0.00006846, -0.00013345, -0.00017796, -0.00019351, -0.00017686, -0.00013073, -0.00006328, 0.00001342, 0.00008572, 0.00014105, 0.00017031, 0.00016950, 0.00014034, 0.00008973, 0.00002817, -0.00003247, -0.00008134, -0.00011068, -0.00011718, -0.00010231, -0.00007169, -0.00003357, 0.00000304, 0.00003034, 0.00004327, 0.00004033, 0.00002355, -0.00000229, -0.00003099, -0.00005644, -0.00007398, -0.00008124, -0.00007844, -0.00006792, -0.00005331, -0.00003842, -0.00002626, -0.00001832, },
		{ -0.00003021, -0.00003049, -0.00003037, -0.00003241, -0.00003813, -0.00004724, -0.00005764, -0.00006583, -0.00006793, -0.00006089, -0.00004360, -0.00001755, 0.00001310, 0.00004222, 0.00006300, 0.00006959, 0.00005866, 0.00003051, -0.00001054, -0.00005674, -0.00009830, -0.00012537, -0.00013022, -0.00010909, -0.00006328, 0.00000077, 0.00007249, 0.00013915, 0.00018834, 0.00021044, 0.00020067, 0.00016010, 0.00009557, 0.00001830, -0.00005822, -0.00012093, -0.00015978, -0.00016957, -0.00015090, -0.00010982, -0.00005642, -0.00000264, 0.00004026, 0.00006395, 0.00006459, 0.00004336, 0.00000597, -0.00003886, -0.00008136, -0.00011289, -0.00012764, -0.00012372, -0.00010320, -0.00007138, -0.00003536, -0.00000236, 0.00002190, 0.00003426, 0.00003454, 0.00002517, 0.00001030, -0.00000554, -0.00001859, -0.00002682, },
		{ -0.00002757, -0.00003491, -0.00004486, -0.00005330, -0.00005674, -0.00005339, -0.00004364, -0.00002996, -0.00001612, -0.00000601, -0.00000240, -0.00000593, -0.00001467, -0.00002444, -0.00002976, -0.00002534, -0.00000761, 0.00002397, 0.00006629, 0.00011282, 0.00015471, 0.00018261, 0.00018870, 0.00016866, 0.00012296, 0.00005716, -0.00001888, -0.00009273, -0.00015182, -0.00018597, -0.00018953, -0.00016267, -0.00011140, -0.00004645, 0.00001887, 0.00007138, 0.00010069, 0.00010135, 0.00007387, 0.00002454, -0.00003590, -0.00009464, -0.00013941, -0.00016108, -0.00015538, -0.00012373, -0.00007269, -0.00001248, 0.00004526, 0.00008996, 0.00011417, 0.00011494, 0.00009408, 0.00005749, 0.00001358, -0.00002873, -0.00006192, -0.00008138, -0.00008608, -0.00007840, -0.00006318, -0.00004624, -0.00003282, -0.00002632, },
		{ -0.00010580, -0.00010178, -0.00008573, -0.00006161, -0.00003403, -0.00000699, 0.00001706, 0.00003755, 0.00005552, 0.00007276, 0.00009079, 0.00010985, 0.00012832, 0.00014283, 0.00014888, 0.00014211, 0.00011965, 0.00008137, 0.00003055, -0.00002624, -0.00008011, -0.00012154, -0.00014238, -0.00013782, -0.00010770, -0.00005702, 0.00000481, 0.00006562, 0.00011290, 0.00013641, 0.00013046, 0.00009527, 0.00003717, -0.00003249, -0.00009962, -0.00015019, -0.00017317, -0.00016292, -0.00012039, -0.00005311, 0.00002634, 0.00010277, 0.00016145, 0.00019109, 0.00018604, 0.00014733, 0.00008236, 0.00000329, -0.00007553, -0.00014043, -0.00018113, -0.00019256, -0.00017560, -0.00013651, -0.00008535, -0.00003361, 0.00000822, 0.00003269, 0.00003674, 0.00002189, -0.00000649, -0.00004075, -0.00007278, -0.00009583, },
		{ 0.00003664, 0.00009316, 0.00013652, 0.00016625, 0.00018368, 0.00019085, 0.00018957, 0.00018094, 0.00016524, 0.00014244, 0.00011291, 0.00007817, 0.00004128, 0.00000677, -0.00002003, -0.00003410, -0.00003209, -0.00001346, 0.00001884, 0.00005851, 0.00009683, 0.00012438, 0.00013304, 0.00011805, 0.00007939, 0.00002232, -0.00004333, -0.00010486, -0.00014925, -0.00016598, -0.00014945, -0.00010059, -0.00002708, 0.00005779, 0.00013752, 0.00019563, 0.00021902, 0.00020074, 0.00014162, 0.00005040, -0.00005779, -0.00016422, -0.00025015, -0.00030044, -0.00030647, -0.00026779, -0.00019212, -0.00009385, 0.00000884, 0.00009746, 0.00015668, 0.00017709, 0.00015667, 0.00010081, 0.00002097, -0.00006788, -0.00015013, -0.00021246, -0.00024598, -0.00024734, -0.00021874, -0.00016675, -0.00010057, -0.00002994, },
		{ 0.00036791, 0.00032777, 0.00028241, 0.00023575, 0.00019038, 0.00014809, 0.00011042, 0.00007903, 0.00005562, 0.00004154, 0.00003717, 0.00004137, 0.00005114, 0.00006187, 0.00006807, 0.00006456, 0.00004785, 0.00001735, -0.00002399, -0.00006975, -0.00011109, -0.00013850, -0.00014397, -0.00012311, -0.00007673, -0.00001136, 0.00006145, 0.00012704, 0.00017044, 0.00017956, 0.00014799, 0.00007697, -0.00002423, -0.00013950, -0.00024855, -0.00033067, -0.00036872, -0.00035277, -0.00028230, -0.00016686, -0.00002467, 0.00012044, 0.00024322, 0.00032156, 0.00034053, 0.00029509, 0.00019102, 0.00004391, -0.00012369, -0.00028618, -0.00041927, -0.00050396, -0.00052933, -0.00049377, -0.00040460, -0.00027602, -0.00012626, 0.00002573, 0.00016335, 0.00027444, 0.00035244, 0.00039624, 0.00040915, 0.00039731, },
		{ 0.00005238, 0.00001395, -0.00001298, -0.00003039, -0.00004008, -0.00004395, -0.00004419, -0.00004338, -0.00004422, -0.00004903, -0.00005897, -0.00007354, -0.00009026, -0.00010491, -0.00011238, -0.00010792, -0.00008864, -0.00005471, -0.00001014, 0.00003745, 0.00007789, 0.00010046, 0.00009630, 0.00006082, -0.00000453, -0.00009155, -0.00018607, -0.00027008, -0.00032501, -0.00033549, -0.00029291, -0.00019794, -0.00006145, 0.00009653, 0.00024976, 0.00037010, 0.00043254, 0.00041999, 0.00032680, 0.00016042, -0.00005920, -0.00030253, -0.00053500, -0.00072253, -0.00083699, -0.00086060, -0.00078851, -0.00062925, -0.00040287, -0.00013740, 0.00013579, 0.00038687, 0.00059166, 0.00073450, 0.00080943, 0.00081960, 0.00077539, 0.00069161, 0.00058447, 0.00046899, 0.00035711, 0.00025685, 0.00017231, 0.00010451, },
		{ -0.00013401, -0.00011755, -0.00010349, -0.00009226, -0.00008402, -0.00007838, -0.00007426, -0.00006991, -0.00006323, -0.00005249, -0.00003704, -0.00001799, 0.00000158, 0.00001687, 0.00002228, 0.00001290, -0.00001396, -0.00005732, -0.00011179, -0.00016775, -0.00021268, -0.00023343, -0.00021916, -0.00016423, -0.00007051, 0.00005156, 0.00018362, 0.00030175, 0.00038028, 0.00039650, 0.00033535, 0.00019316, -0.00002032, -0.00028220, -0.00055905, -0.00081165, -0.00100093, -0.00109421, -0.00107069, -0.00092500, -0.00066843, -0.00032732, 0.00006076, 0.00045270, 0.00080618, 0.00108556, 0.00126657, 0.00133880, 0.00130593, 0.00118380, 0.00099675, 0.00077315, 0.00054086, 0.00032346, 0.00013780, -0.00000698, -0.00010900, -0.00017199, -0.00020327, -0.00021147, -0.00020479, -0.00018988, -0.00017141, -0.00015225, },
		{ -0.00003250, -0.00003211, -0.00003219, -0.00003217, -0.00003171, -0.00003098, -0.00003090, -0.00003303, -0.00003915, -0.00005055, -0.00006719, -0.00008705, -0.00010587, -0.00011758, -0.00011547, -0.00009384, -0.00005001, 0.00001402, 0.00009050, 0.00016613, 0.00022354, 0.00024402, 0.00021118, 0.00011482, -0.00004575, -0.00025968, -0.00050456, -0.00074845, -0.00095406, -0.00108425, -0.00110823, -0.00100708, -0.00077779, -0.00043492, -0.00000943, 0.00045515, 0.00090866, 0.00130143, 0.00159129, 0.00174943, 0.00176406, 0.00164137, 0.00140370, 0.00108541, 0.00072719, 0.00036989, 0.00004891, -0.00020995, -0.00039269, -0.00049718, -0.00053135, -0.00051016, -0.00045188, -0.00037467, -0.00029385, -0.00022030, -0.00016003, -0.00011480, -0.00008333, -0.00006271, -0.00004974, -0.00004171, -0.00003677, -0.00003390, },
		{ -0.00005222, -0.00005205, -0.00005164, -0.00005117, -0.00005063, -0.00004946, -0.00004643, -0.00003978, -0.00002781, -0.00000974, 0.00001325, 0.00003731, 0.00005561, 0.00005892, 0.00003707, -0.00001882, -0.00011377, -0.00024641, -0.00040734, -0.00057870, -0.00073547, -0.00084844, -0.00088855, -0.00083200, -0.00066512, -0.00038818, -0.00001720, 0.00041669, 0.00087052, 0.00129491, 0.00164089, 0.00186704, 0.00194566, 0.00186704, 0.00164089, 0.00129491, 0.00087052, 0.00041669, -0.00001720, -0.00038818, -0.00066512, -0.00083200, -0.00088855, -0.00084844, -0.00073547, -0.00057870, -0.00040734, -0.00024641, -0.00011377, -0.00001882, 0.00003707, 0.00005892, 0.00005561, 0.00003731, 0.00001325, -0.00000974, -0.00002781, -0.00003978, -0.00004643, -0.00004946, -0.00005063, -0.00005117, -0.00005164, -0.00005205, },
		{ -0.00003250, -0.00003390, -0.00003677, -0.00004171, -0.00004974, -0.00006271, -0.00008333, -0.00011480, -0.00016003, -0.00022030, -0.00029385, -0.00037467, -0.00045188, -0.00051016, -0.00053135, -0.00049718, -0.00039269, -0.00020995, 0.00004891, 0.00036989, 0.00072719, 0.00108541, 0.00140370, 0.00164137, 0.00176406, 0.00174943, 0.00159129, 0.00130143, 0.00090866, 0.00045515, -0.00000943, -0.00043492, -0.00077779, -0.00100708, -0.00110823, -0.00108425, -0.00095406, -0.00074845, -0.00050456, -0.00025968, -0.00004575, 0.00011482, 0.00021118, 0.00024402, 0.00022354, 0.00016613, 0.00009050, 0.00001402, -0.00005001, -0.00009384, -0.00011547, -0.00011758, -0.00010587, -0.00008705, -0.00006719, -0.00005055, -0.00003915, -0.00003303, -0.00003090, -0.00003098, -0.00003171, -0.00003217, -0.00003219, -0.00003211, },
		{ -0.00013401, -0.00015225, -0.00017141, -0.00018988, -0.00020479, -0.00021147, -0.00020327, -0.00017199, -0.00010900, -0.00000698, 0.00013780, 0.00032346, 0.00054086, 0.00077315, 0.00099675, 0.00118380, 0.00130593, 0.00133880, 0.00126657, 0.00108556, 0.00080618, 0.00045270, 0.00006076, -0.00032732, -0.00066843, -0.00092500, -0.00107069, -0.00109421, -0.00100093, -0.00081165, -0.00055905, -0.00028220, -0.00002032, 0.00019316, 0.00033535, 0.00039650, 0.00038028, 0.00030175, 0.00018362, 0.00005156, -0.00007051, -0.00016423, -0.00021916, -0.00023343, -0.00021268, -0.00016775, -0.00011179, -0.00005732, -0.00001396, 0.00001290, 0.00002228, 0.00001687, 0.00000158, -0.00001799, -0.00003704, -0.00005249, -0.00006323, -0.00006991, -0.00007426, -0.00007838, -0.00008402, -0.00009226, -0.00010349, -0.00011755, },
		{ 0.00005238, 0.00010451, 0.00017231, 0.00025685, 0.00035711, 0.00046899, 0.00058447, 0.00069161, 0.00077539, 0.00081960, 0.00080943, 0.00073450, 0.00059166, 0.00038687, 0.00013579, -0.00013740, -0.00040287, -0.00062925, -0.00078851, -0.00086060, -0.00083699, -0.00072253, -0.00053500, -0.00030253, -0.00005920, 0.00016042, 0.00032680, 0.00041999, 0.00043254, 0.00037010, 0.00024976, 0.00009653, -0.00006145, -0.00019794, -0.00029291, -0.00033549, -0.00032501, -0.00027008, -0.00018607, -0.00009155, -0.00000453, 0.00006082, 0.00009630, 0.00010046, 0.00007789, 0.00003745, -0.00001014, -0.00005471, -0.00008864, -0.00010792, -0.00011238, -0.00010491, -0.00009026, -0.00007354, -0.00005897, -0.00004903, -0.00004422, -0.00004338, -0.00004419, -0.00004395, -0.00004008, -0.00003039, -0.00001298, 0.00001395, },
		{ 0.00036791, 0.00039731, 0.00040915, 0.00039624, 0.00035244, 0.00027444, 0.00016335, 0.00002573, -0.00012626, -0.00027602, -0.00040460, -0.00049377, -0.00052933, -0.00050396, -0.00041927, -0.00028618, -0.00012369, 0.00004391, 0.00019102, 0.00029509, 0.00034053, 0.00032156, 0.00024322, 0.00012044, -0.00002467, -0.00016686, -0.00028230, -0.00035277, -0.00036872, -0.00033067, -0.00024855, -0.00013950, -0.00002423, 0.00007697, 0.00014799, 0.00017956, 0.00017044, 0.00012704, 0.00006145, -0.00001136, -0.00007673, -0.00012311, -0.00014397, -0.00013850, -0.00011109, -0.00006975, -0.00002399, 0.00001735, 0.00004785, 0.00006456, 0.00006807, 0.00006187, 0.00005114, 0.00004137, 0.00003717, 0.00004154, 0.00005562, 0.00007903, 0.00011042, 0.00014809, 0.00019038, 0.00023575, 0.00028241, 0.00032777, },
		{ 0.00003664, -0.00002994, -0.00010057, -0.00016675, -0.00021874, -0.00024734, -0.00024598, -0.00021246, -0.00015013, -0.00006788, 0.00002097, 0.00010081, 0.00015667, 0.00017709, 0.00015668, 0.00009746, 0.00000884, -0.00009385, -0.00019212, -0.00026779, -0.00030647, -0.00030044, -0.00025015, -0.00016422, -0.00005779, 0.00005040, 0.00014162, 0.00020074, 0.00021902, 0.00019563, 0.00013752, 0.00005779, -0.00002708, -0.00010059, -0.00014945, -0.00016598, -0.00014925, -0.00010486, -0.00004333, 0.00002232, 0.00007939, 0.00011805, 0.00013304, 0.00012438, 0.00009683, 0.00005851, 0.00001884, -0.00001346, -0.00003209, -0.00003410, -0.00002003, 0.00000677, 0.00004128, 0.00007817, 0.00011291, 0.00014244, 0.00016524, 0.00018094, 0.00018957, 0.00019085, 0.00018368, 0.00016625, 0.00013652, 0.00009316, },
		{ -0.00010580, -0.00009583, -0.00007278, -0.00004075, -0.00000649, 0.00002189, 0.00003674, 0.00003269, 0.00000822, -0.00003361, -0.00008535, -0.00013651, -0.00017560, -0.00019256, -0.00018113, -0.00014043, -0.00007553, 0.00000329, 0.00008236, 0.00014733, 0.00018604, 0.00019109, 0.00016145, 0.00010277, 0.00002634, -0.00005311, -0.00012039, -0.00016292, -0.00017317, -0.00015019, -0.00009962, -0.00003249, 0.00003717, 0.00009527, 0.00013046, 0.00013641, 0.00011290, 0.00006562, 0.00000481, -0.00005702, -0.00010770, -0.00013782, -0.00014238, -0.00012154, -0.00008011, -0.00002624, 0.00003055, 0.00008137, 0.00011965, 0.00014211, 0.00014888, 0.00014283, 0.00012832, 0.00010985, 0.00009079, 0.00007276, 0.00005552, 0.00003755, 0.00001706, -0.00000699, -0.00003403, -0.00006161, -0.00008573, -0.00010178, },
		{ -0.00002757, -0.00002632, -0.00003282, -0.00004624, -0.00006318, -0.00007840, -0.00008608, -0.00008138, -0.00006192, -0.00002873, 0.00001358, 0.00005749, 0.00009408, 0.00011494, 0.00011417, 0.00008996, 0.00004526, -0.00001248, -0.00007269, -0.00012373, -0.00015538, -0.00016108, -0.00013941, -0.00009464, -0.00003590, 0.00002454, 0.00007387, 0.00010135, 0.00010069, 0.00007138, 0.00001887, -0.00004645, -0.00011140, -0.00016267, -0.00018953, -0.00018597, -0.00015182, -0.00009273, -0.00001888, 0.00005716, 0.00012296, 0.00016866, 0.00018870, 0.00018261, 0.00015471, 0.00011282, 0.00006629, 0.00002397, -0.00000761, -0.00002534, -0.00002976, -0.00002444, -0.00001467, -0.00000593, -0.00000240, -0.00000601, -0.00001612, -0.00002996, -0.00004364, -0.00005339, -0.00005674, -0.00005330, -0.00004486, -0.00003491, },
		{ -0.00003021, -0.00002682, -0.00001859, -0.00000554, 0.00001030, 0.00002517, 0.00003454, 0.00003426, 0.00002190, -0.00000236, -0.00003536, -0.00007138, -0.00010320, -0.00012372, -0.00012764, -0.00011289, -0.00008136, -0.00003886, 0.00000597, 0.00004336, 0.00006459, 0.00006395, 0.00004026, -0.00000264, -0.00005642, -0.00010982, -0.00015090, -0.00016957, -0.00015978, -0.00012093, -0.00005822, 0.00001830, 0.00009557, 0.00016010, 0.00020067, 0.00021044, 0.00018834, 0.00013915, 0.00007249, 0.00000077, -0.00006328, -0.00010909, -0.00013022, -0.00012537, -0.00009830, -0.00005674, -0.00001054, 0.00003051, 0.00005866, 0.00006959, 0.00006300, 0.00004222, 0.00001310, -0.00001755, -0.00004360, -0.00006089, -0.00006793, -0.00006583, -0.00005764, -0.00004724, -0.00003813, -0.00003241, -0.00003037, -0.00003049, },
		{ -0.00001447, -0.00001832, -0.00002626, -0.00003842, -0.00005331, -0.00006792, -0.00007844, -0.00008124, -0.00007398, -0.00005644, -0.00003099, -0.00000229, 0.00002355, 0.00004033, 0.00004327, 0.00003034, 0.00000304, -0.00003357, -0.00007169, -0.00010231, -0.00011718, -0.00011068, -0.00008134, -0.00003247, 0.00002817, 0.00008973, 0.00014034, 0.00016950, 0.00017031, 0.00014105, 0.00008572, 0.00001342, -0.00006328, -0.00013073, -0.00017686, -0.00019351, -0.00017796, -0.00013345, -0.00006846, 0.00000502, 0.00007386, 0.00012627, 0.00015402, 0.00015376, 0.00012742, 0.00008142, 0.00002519, -0.00003093, -0.00007772, -0.00010874, -0.00012128, -0.00011644, -0.00009840, -0.00007308, -0.00004671, -0.00002439, -0.00000916, -0.00000175, -0.00000082, -0.00000383, -0.00000799, -0.00001122, -0.00001277, -0.00001328, },
		{ -0.00003617, -0.00003224, -0.00002435, -0.00001243, 0.00000218, 0.00001676, 0.00002782, 0.00003192, 0.00002671, 0.00001177, -0.00001096, -0.00003730, -0.00006152, -0.00007760, -0.00008060, -0.00006804, -0.00004070, -0.00000286, 0.00003830, 0.00007409, 0.00009603, 0.00009780, 0.00007676, 0.00003487, -0.00002144, -0.00008223, -0.00013591, -0.00017150, -0.00018089, -0.00016055, -0.00011248, -0.00004397, 0.00003368, 0.00010719, 0.00016382, 0.00019377, 0.00019199, 0.00015908, 0.00010105, 0.00002808, -0.00004747, -0.00011346, -0.00016013, -0.00018177, -0.00017744, -0.00015080, -0.00010898, -0.00006094, -0.00001554, 0.00002007, 0.00004165, 0.00004825, 0.00004192, 0.00002677, 0.00000774, -0.00001062, -0.00002508, -0.00003425, -0.00003842, -0.00003913, -0.00003829, -0.00003747, -0.00003731, -0.00003732, },
		{ 0.00001623, 0.00001183, 0.00000248, -0.00001193, -0.00003005, -0.00004913, -0.00006546, -0.00007525, -0.00007555, -0.00006517, -0.00004524, -0.00001924, 0.00000755, 0.00002906, 0.00003984, 0.00003634, 0.00001798, -0.00001245, -0.00004908, -0.00008404, -0.00010896, -0.00011683, -0.00010362, -0.00006937, -0.00001849, 0.00004086, 0.00009816, 0.00014256, 0.00016503, 0.00016015, 0.00012737, 0.00007118, 0.00000042, -0.00007336, -0.00013804, -0.00018314, -0.00020180, -0.00019187, -0.00015623, -0.00010204, -0.00003925, 0.00002142, 0.00007040, 0.00010101, 0.00011044, 0.00009986, 0.00007385, 0.00003916, 0.00000311, -0.00002784, -0.00004924, -0.00005925, -0.00005854, -0.00004971, -0.00003639, -0.00002215, -0.00000974, -0.00000057, 0.00000528, 0.00000873, 0.00001103, 0.00001317, 0.00001535, 0.00001685, },
		{ -0.00007422, -0.00007757, -0.00007437, -0.00006382, -0.00004664, -0.00002520, -0.00000312, 0.00001541, 0.00002660, 0.00002805, 0.00001942, 0.00000270, -0.00001801, -0.00003735, -0.00004983, -0.00005117, -0.00003939, -0.00001545, 0.00001667, 0.00005063, 0.00007898, 0.00009473, 0.00009287, 0.00007163, 0.00003308, -0.00001705, -0.00007035, -0.00011725, -0.00014891, -0.00015895, -0.00014478, -0.00010815, -0.00005488, 0.00000621, 0.00006498, 0.00011187, 0.00013966, 0.00014475, 0.00012765, 0.00009274, 0.00004722, -0.00000039, -0.00004196, -0.00007118, -0.00008463, -0.00008214, -0.00006644, -0.00004233, -0.00001546, 0.00000888, 0.00002679, 0.00003629, 0.00003743, 0.00003184, 0.00002204, 0.00001063, -0.00000042, -0.00001014, -0.00001864, -0.00002671, -0.00003534, -0.00004514, -0.00005584, -0.00006621, },
		{ 0.00001173, 0.00002728, 0.00003769, 0.00004093, 0.00003637, 0.00002504, 0.00000958, -0.00000637, -0.00001894, -0.00002497, -0.00002282, -0.00001286, 0.00000249, 0.00001926, 0.00003279, 0.00003882, 0.00003461, 0.00001965, -0.00000395, -0.00003181, -0.00005805, -0.00007654, -0.00008222, -0.00007230, -0.00004698, -0.00000964, 0.00003376, 0.00007572, 0.00010866, 0.00012645, 0.00012562, 0.00010609, 0.00007119, 0.00002704, -0.00001869, -0.00005823, -0.00008520, -0.00009581, -0.00008944, -0.00006859, -0.00003829, -0.00000489, 0.00002515, 0.00004662, 0.00005638, 0.00005379, 0.00004057, 0.00002017, -0.00000313, -0.00002527, -0.00004317, -0.00005525, -0.00006151, -0.00006320, -0.00006220, -0.00006035, -0.00005882, -0.00005784, -0.00005662, -0.00005377, -0.00004777, -0.00003762, -0.00002333, -0.00000612, },
		{ 0.00001859, 0.00000475, -0.00000771, -0.00001610, -0.00001867, -0.00001509, -0.00000651, 0.00000465, 0.00001530, 0.00002247, 0.00002402, 0.00001928, 0.00000925, -0.00000352, -0.00001557, -0.00002328, -0.00002379, -0.00001580, 0.00000003, 0.00002107, 0.00004319, 0.00006161, 0.00007192, 0.00007113, 0.00005834, 0.00003511, 0.00000521, -0.00002611, -0.00005311, -0.00007079, -0.00007590, -0.00006762, -0.00004777, -0.00002042, 0.00000894, 0.00003449, 0.00005127, 0.00005612, 0.00004828, 0.00002943, 0.00000324, -0.00002545, -0.00005165, -0.00007127, -0.00008181, -0.00008276, -0.00007545, -0.00006262, -0.00004760, -0.00003353, -0.00002261, -0.00001572, -0.00001238, -0.00001102, -0.00000963, -0.00000635, -0.00000010, 0.00000908, 0.00001999, 0.00003057, 0.00003847, 0.00004167, 0.00003907, 0.00003088, },
		{ -0.00000234, 0.00000443, 0.00001197, 0.00001793, 0.00002040, 0.00001839, 0.00001212, 0.00000294, -0.00000693, -0.00001502, -0.00001923, -0.00001835, -0.00001242, -0.00000281, 0.00000817, 0.00001770, 0.00002323, 0.00002305, 0.00001682, 0.00000565, -0.00000808, -0.00002123, -0.00003063, -0.00003386, -0.00002983, -0.00001914, -0.00000400, 0.00001218, 0.00002554, 0.00003257, 0.00003097, 0.00002015, 0.00000145, -0.00002208, -0.00004634, -0.00006691, -0.00008006, -0.00008344, -0.00007657, -0.00006091, -0.00003949, -0.00001625, 0.00000481, 0.00002043, 0.00002877, 0.00002967, 0.00002460, 0.00001616, 0.00000748, 0.00000139, -0.00000019, 0.00000331, 0.00001107, 0.00002114, 0.00003093, 0.00003797, 0.00004042, 0.00003762, 0.00003016, 0.00001976, 0.00000879, -0.00000029, -0.00000557, -0.00000615, },
		{ -0.00000893, -0.00001538, -0.00002262, -0.00002876, -0.00003202, -0.00003120, -0.00002598, -0.00001704, -0.00000589, 0.00000543, 0.00001489, 0.00002089, 0.00002263, 0.00002029, 0.00001492, 0.00000815, 0.00000175, -0.00000285, -0.00000494, -0.00000469, -0.00000314, -0.00000190, -0.00000268, -0.00000680, -0.00001473, -0.00002587, -0.00003856, -0.00005033, -0.00005847, -0.00006060, -0.00005525, -0.00004233, -0.00002317, -0.00000044, 0.00002239, 0.00004172, 0.00005450, 0.00005890, 0.00005466, 0.00004317, 0.00002719, 0.00001025, -0.00000407, -0.00001285, -0.00001443, -0.00000874, 0.00000278, 0.00001752, 0.00003227, 0.00004394, 0.00005023, 0.00005002, 0.00004360, 0.00003250, 0.00001907, 0.00000593, -0.00000466, -0.00001126, -0.00001352, -0.00001220, -0.00000886, -0.00000544, -0.00000370, -0.00000479, },
		{ -0.00001435, -0.00000516, 0.00000572, 0.00001658, 0.00002560, 0.00003122, 0.00003249, 0.00002925, 0.00002210, 0.00001228, 0.00000134, -0.00000918, -0.00001811, -0.00002477, -0.00002910, -0.00003148, -0.00003254, -0.00003285, -0.00003270, -0.00003195, -0.00003001, -0.00002605, -0.00001925, -0.00000915, 0.00000412, 0.00001962, 0.00003570, 0.00005018, 0.00006077, 0.00006553, 0.00006333, 0.00005416, 0.00003926, 0.00002094, 0.00000224, -0.00001364, -0.00002396, -0.00002696, -0.00002222, -0.00001077, 0.00000510, 0.00002228, 0.00003748, 0.00004780, 0.00005132, 0.00004745, 0.00003694, 0.00002173, 0.00000448, -0.00001201, -0.00002532, -0.00003391, -0.00003732, -0.00003617, -0.00003191, -0.00002641, -0.00002148, -0.00001843, -0.00001777, -0.00001919, -0.00002161, -0.00002357, -0.00002360, -0.00002065, },
		{ 0.00002059, 0.00001536, 0.00000693, -0.00000346, -0.00001426, -0.00002385, -0.00003088, -0.00003449, -0.00003437, -0.00003078, -0.00002438, -0.00001605, -0.00000669, 0.00000299, 0.00001248, 0.00002153, 0.00003001, 0.00003776, 0.00004452, 0.00004980, 0.00005293, 0.00005314, 0.00004981, 0.00004264, 0.00003184, 0.00001827, 0.00000338, -0.00001094, -0.00002269, -0.00003010, -0.00003198, -0.00002800, -0.00001884, -0.00000609, 0.00000799, 0.00002086, 0.00003017, 0.00003420, 0.00003216, 0.00002437, 0.00001216, -0.00000239, -0.00001687, -0.00002899, -0.00003703, -0.00004014, -0.00003843, -0.00003292, -0.00002521, -0.00001715, -0.00001039, -0.00000608, -0.00000457, -0.00000549, -0.00000780, -0.00001016, -0.00001122, -0.00001003, -0.00000624, -0.00000019, 0.00000710, 0.00001420, 0.00001959, 0.00002198, },
		{ -0.00001713, -0.00001585, -0.00001124, -0.00000383, 0.00000544, 0.00001537, 0.00002479, 0.00003269, 0.00003837, 0.00004151, 0.00004211, 0.00004042, 0.00003682, 0.00003173, 0.00002550, 0.00001840, 0.00001067, 0.00000257, -0.00000557, -0.00001327, -0.00001994, -0.00002493, -0.00002762, -0.00002759, -0.00002473, -0.00001935, -0.00001214, -0.00000415, 0.00000339, 0.00000931, 0.00001266, 0.00001297, 0.00001029, 0.00000526, -0.00000110, -0.00000749, -0.00001267, -0.00001568, -0.00001605, -0.00001390, -0.00000989, -0.00000508, -0.00000071, 0.00000209, 0.00000257, 0.00000049, -0.00000376, -0.00000930, -0.00001489, -0.00001927, -0.00002136, -0.00002056, -0.00001686, -0.00001087, -0.00000369, 0.00000331, 0.00000879, 0.00001169, 0.00001149, 0.00000827, 0.00000273, -0.00000394, -0.00001034, -0.00001510, },
		{ 0.00002528, 0.00002706, 0.00002629, 0.00002304, 0.00001777, 0.00001119, 0.00000414, -0.00000259, -0.00000835, -0.00001273, -0.00001561, -0.00001707, -0.00001737, -0.00001686, -0.00001586, -0.00001463, -0.00001333, -0.00001200, -0.00001061, -0.00000910, -0.00000740, -0.00000548, -0.00000336, -0.00000107, 0.00000131, 0.00000372, 0.00000611, 0.00000845, 0.00001068, 0.00001272, 0.00001439, 0.00001546, 0.00001560, 0.00001447, 0.00001174, 0.00000724, 0.00000099, -0.00000670, -0.00001523, -0.00002375, -0.00003124, -0.00003668, -0.00003921, -0.00003827, -0.00003378, -0.00002614, -0.00001621, -0.00000522, 0.00000545, 0.00001448, 0.00002080, 0.00002381, 0.00002347, 0.00002026, 0.00001513, 0.00000930, 0.00000405, 0.00000046, -0.00000075, 0.00000066, 0.00000442, 0.00000979, 0.00001578, 0.00002128, },
		{ -0.00001366, -0.00001589, -0.00001623, -0.00001465, -0.00001147, -0.00000727, -0.00000279, 0.00000120, 0.00000403, 0.00000527, 0.00000481, 0.00000283, -0.00000014, -0.00000339, -0.00000611, -0.00000753, -0.00000706, -0.00000441, 0.00000037, 0.00000683, 0.00001423, 0.00002156, 0.00002773, 0.00003167, 0.00003253, 0.00002979, 0.00002335, 0.00001356, 0.00000124, -0.00001245, -0.00002609, -0.00003819, -0.00004739, -0.00005256, -0.00005306, -0.00004872, -0.00003994, -0.00002763, -0.00001311, 0.00000207, 0.00001628, 0.00002807, 0.00003632, 0.00004037, 0.00004013, 0.00003602, 0.00002894, 0.00002011, 0.00001087, 0.00000249, -0.00000399, -0.00000794, -0.00000917, -0.00000795, -0.00000493, -0.00000102, 0.00000279, 0.00000563, 0.00000683, 0.00000610, 0.00000351, -0.00000050, -0.00000525, -0.00000990, },
		{ 0.00001556, 0.00001793, 0.00001818, 0.00001613, 0.00001201, 0.00000641, 0.00000023, -0.00000549, -0.00000969, -0.00001148, -0.00001033, -0.00000616, 0.00000058, 0.00000897, 0.00001769, 0.00002527, 0.00003023, 0.00003140, 0.00002810, 0.00002026, 0.00000853, -0.00000582, -0.00002105, -0.00003516, -0.00004622, -0.00005261, -0.00005330, -0.00004798, -0.00003716, -0.00002207, -0.00000452, 0.00001338, 0.00002950, 0.00004198, 0.00004947, 0.00005130, 0.00004757, 0.00003904, 0.00002704, 0.00001324, -0.00000061, -0.00001289, -0.00002234, -0.00002819, -0.00003024, -0.00002881, -0.00002462, -0.00001871, -0.00001221, -0.00000617, -0.00000146, 0.00000143, 0.00000232, 0.00000144, -0.00000071, -0.00000343, -0.00000594, -0.00000755, -0.00000771, -0.00000617, -0.00000300, 0.00000145, 0.00000655, 0.00001153, },
		{ -0.00000647, -0.00001180, -0.00001506, -0.00001562, -0.00001326, -0.00000817, -0.00000105, 0.00000701, 0.00001467, 0.00002052, 0.00002335, 0.00002236, 0.00001732, 0.00000868, -0.00000245, -0.00001445, -0.00002544, -0.00003353, -0.00003718, -0.00003548, -0.00002830, -0.00001637, -0.00000121, 0.00001513, 0.00003032, 0.00004215, 0.00004884, 0.00004938, 0.00004361, 0.00003229, 0.00001694, -0.00000039, -0.00001744, -0.00003208, -0.00004263, -0.00004806, -0.00004812, -0.00004329, -0.00003466, -0.00002375, -0.00001220, -0.00000153, 0.00000704, 0.00001277, 0.00001542, 0.00001524, 0.00001289, 0.00000923, 0.00000525, 0.00000183, -0.00000033, -0.00000081, 0.00000049, 0.00000336, 0.00000733, 0.00001175, 0.00001589, 0.00001902, 0.00002054, 0.00002005, 0.00001742, 0.00001284, 0.00000680, 0.00000005, },
		{ 0.00000064, 0.00000698, 0.00001263, 0.00001662, 0.00001817, 0.00001685, 0.00001273, 0.00000640, -0.00000112, -0.00000846, -0.00001420, -0.00001709, -0.00001629, -0.00001160, -0.00000352, 0.00000676, 0.00001758, 0.00002699, 0.00003318, 0.00003472, 0.00003090, 0.00002184, 0.00000853, -0.00000730, -0.00002349, -0.00003774, -0.00004803, -0.00005290, -0.00005173, -0.00004477, -0.00003314, -0.00001854, -0.00000306, 0.00001125, 0.00002264, 0.00002994, 0.00003265, 0.00003103, 0.00002591, 0.00001855, 0.00001039, 0.00000281, -0.00000309, -0.00000661, -0.00000750, -0.00000596, -0.00000250, 0.00000211, 0.00000704, 0.00001148, 0.00001478, 0.00001647, 0.00001632, 0.00001435, 0.00001078, 0.00000603, 0.00000066, -0.00000467, -0.00000927, -0.00001251, -0.00001388, -0.00001310, -0.00001015, -0.00000536, },
		{ 0.00000726, 0.00000398, 0.00000028, -0.00000305, -0.00000521, -0.00000559, -0.00000386, -0.00000015, 0.00000495, 0.00001048, 0.00001520, 0.00001791, 0.00001761, 0.00001378, 0.00000652, -0.00000340, -0.00001463, -0.00002546, -0.00003407, -0.00003889, -0.00003887, -0.00003372, -0.00002393, -0.00001080, 0.00000386, 0.00001794, 0.00002946, 0.00003682, 0.00003911, 0.00003623, 0.00002886, 0.00001833, 0.00000637, -0.00000522, -0.00001479, -0.00002119, -0.00002381, -0.00002270, -0.00001843, -0.00001198, -0.00000450, 0.00000287, 0.00000918, 0.00001379, 0.00001640, 0.00001704, 0.00001595, 0.00001357, 0.00001037, 0.00000680, 0.00000329, 0.00000017, -0.00000231, -0.00000393, -0.00000456, -0.00000414, -0.00000271, -0.00000042, 0.00000242, 0.00000541, 0.00000805, 0.00000985, 0.00001041, 0.00000953, },
		{ 0.00000626, 0.00000850, 0.00001071, 0.00001244, 0.00001314, 0.00001230, 0.00000964, 0.00000517, -0.00000071, -0.00000722, -0.00001334, -0.00001793, -0.00001998, -0.00001884, -0.00001434, -0.00000695, 0.00000233, 0.00001207, 0.00002061, 0.00002645, 0.00002841, 0.00002594, 0.00001920, 0.00000906, -0.00000305, -0.00001535, -0.00002604, -0.00003361, -0.00003705, -0.00003600, -0.00003082, -0.00002242, -0.00001213, -0.00000140, 0.00000842, 0.00001630, 0.00002165, 0.00002436, 0.00002473, 0.00002333, 0.00002087, 0.00001797, 0.00001511, 0.00001255, 0.00001031, 0.00000828, 0.00000625, 0.00000408, 0.00000171, -0.00000077, -0.00000316, -0.00000520, -0.00000664, -0.00000732, -0.00000721, -0.00000640, -0.00000512, -0.00000363, -0.00000216, -0.00000083, 0.00000033, 0.00000145, 0.00000272, 0.00000430, },
		{ 0.00000586, 0.00000118, -0.00000337, -0.00000734, -0.00001024, -0.00001161, -0.00001118, -0.00000892, -0.00000513, -0.00000046, 0.00000420, 0.00000783, 0.00000949, 0.00000847, 0.00000451, -0.00000214, -0.00001071, -0.00002002, -0.00002862, -0.00003511, -0.00003832, -0.00003755, -0.00003271, -0.00002431, -0.00001341, -0.00000139, 0.00001025, 0.00002019, 0.00002747, 0.00003163, 0.00003275, 0.00003135, 0.00002825, 0.00002438, 0.00002052, 0.00001721, 0.00001463, 0.00001262, 0.00001074, 0.00000848, 0.00000537, 0.00000117, -0.00000407, -0.00000997, -0.00001589, -0.00002103, -0.00002462, -0.00002606, -0.00002505, -0.00002162, -0.00001615, -0.00000929, -0.00000182, 0.00000546, 0.00001189, 0.00001699, 0.00002053, 0.00002248, 0.00002297, 0.00002220, 0.00002040, 0.00001773, 0.00001433, 0.00001032, },
		{ -0.00001188, -0.00000878, -0.00000562, -0.00000293, -0.00000125, -0.00000103, -0.00000257, -0.00000588, -0.00001069, -0.00001638, -0.00002209, -0.00002682, -0.00002961, -0.00002970, -0.00002668, -0.00002061, -0.00001201, -0.00000184, 0.00000871, 0.00001837, 0.00002603, 0.00003093, 0.00003276, 0.00003171, 0.00002840, 0.00002373, 0.00001869, 0.00001411, 0.00001057, 0.00000822, 0.00000680, 0.00000577, 0.00000438, 0.00000196, -0.00000195, -0.00000742, -0.00001410, -0.00002121, -0.00002773, -0.00003252, -0.00003459, -0.00003328, -0.00002838, -0.00002023, -0.00000964, 0.00000221, 0.00001399, 0.00002438, 0.00003232, 0.00003711, 0.00003847, 0.00003658, 0.00003197, 0.00002540, 0.00001772, 0.00000976, 0.00000223, -0.00000436, -0.00000967, -0.00001350, -0.00001582, -0.00001666, -0.00001614, -0.00001446, },
		{ -0.00000191, -0.00000682, -0.00001207, -0.00001695, -0.00002073, -0.00002273, -0.00002250, -0.00001983, -0.00001488, -0.00000812, -0.00000032, 0.00000757, 0.00001457, 0.00001983, 0.00002279, 0.00002324, 0.00002139, 0.00001782, 0.00001333, 0.00000881, 0.00000504, 0.00000254, 0.00000147, 0.00000157, 0.00000227, 0.00000279, 0.00000236, 0.00000036, -0.00000346, -0.00000891, -0.00001536, -0.00002183, -0.00002712, -0.00003011, -0.00002990, -0.00002604, -0.00001866, -0.00000843, 0.00000350, 0.00001566, 0.00002651, 0.00003466, 0.00003912, 0.00003939, 0.00003554, 0.00002819, 0.00001835, 0.00000729, -0.00000367, -0.00001337, -0.00002088, -0.00002568, -0.00002760, -0.00002681, -0.00002378, -0.00001914, -0.00001360, -0.00000789, -0.00000266, 0.00000153, 0.00000427, 0.00000530, 0.00000452, 0.00000202, },
		{ -0.00002534, -0.00002135, -0.00001568, -0.00000904, -0.00000225, 0.00000388, 0.00000866, 0.00001164, 0.00001266, 0.00001191, 0.00000983, 0.00000710, 0.00000443, 0.00000248, 0.00000170, 0.00000221, 0.00000382, 0.00000602, 0.00000808, 0.00000924, 0.00000882, 0.00000643, 0.00000205, -0.00000393, -0.00001072, -0.00001729, -0.00002252, -0.00002540, -0.00002527, -0.00002187, -0.00001552, -0.00000699, 0.00000253, 0.00001168, 0.00001909, 0.00002366, 0.00002467, 0.00002195, 0.00001588, 0.00000731, -0.00000255, -0.00001234, -0.00002075, -0.00002672, -0.00002953, -0.00002895, -0.00002519, -0.00001885, -0.00001083, -0.00000218, 0.00000603, 0.00001286, 0.00001758, 0.00001972, 0.00001914, 0.00001598, 0.00001066, 0.00000379, -0.00000385, -0.00001146, -0.00001821, -0.00002339, -0.00002647, -0.00002713, },
		{ 0.00000793, 0.00001034, 0.00001108, 0.00001057, 0.00000937, 0.00000809, 0.00000730, 0.00000737, 0.00000843, 0.00001032, 0.00001262, 0.00001470, 0.00001587, 0.00001549, 0.00001310, 0.00000858, 0.00000214, -0.00000566, -0.00001396, -0.00002172, -0.00002793, -0.00003172, -0.00003255, -0.00003028, -0.00002523, -0.00001812, -0.00000998, -0.00000196, 0.00000480, 0.00000938, 0.00001124, 0.00001026, 0.00000678, 0.00000153, -0.00000449, -0.00001017, -0.00001445, -0.00001654, -0.00001596, -0.00001265, -0.00000700, 0.00000029, 0.00000823, 0.00001575, 0.00002180, 0.00002552, 0.00002634, 0.00002403, 0.00001875, 0.00001101, 0.00000158, -0.00000856, -0.00001838, -0.00002690, -0.00003331, -0.00003703, -0.00003781, -0.00003571, -0.00003111, -0.00002464, -0.00001709, -0.00000933, -0.00000218, 0.00000371, },
		},
		{ { -0.00001314, -0.00001399, -0.00001612, -0.00001837, -0.00001907, -0.00001643, -0.00000914, 0.00000304, 0.00001885, 0.00003551, 0.00004927, 0.00005644, 0.00005459, 0.00004362, 0.00002602, 0.00000634, -0.00001018, -0.00001942, -0.00001982, -0.00001305, -0.00000337, 0.00000393, 0.00000462, -0.00000267, -0.00001581, -0.00002980, -0.00003869, -0.00003787, -0.00002609, -0.00000613, 0.00001600, 0.00003313, 0.00003957, 0.00003313, 0.00001600, -0.00000613, -0.00002609, -0.00003787, -0.00003869, -0.00002980, -0.00001581, -0.00000267, 0.00000462, 0.00000393, -0.00000337, -0.00001305, -0.00001982, -0.00001942, -0.00001018, 0.00000634, 0.00002602, 0.00004362, 0.00005459, 0.00005644, 0.00004927, 0.00003551, 0.00001885, 0.00000304, -0.00000914, -0.00001643, -0.00001907, -0.00001837, -0.00001612, -0.00001399, },
		{ -0.00001303, -0.00000440, 0.00000793, 0.00002264, 0.00003743, 0.00004938, 0.00005554, 0.00005380, 0.00004368, 0.00002678, 0.00000663, -0.00001221, -0.00002543, -0.00003045, -0.00002724, -0.00001845, -0.00000838, -0.00000142, -0.00000034, -0.00000511, -0.00001296, -0.00001944, -0.00002031, -0.00001344, 0.00000006, 0.00001598, 0.00002846, 0.00003217, 0.00002452, 0.00000681, -0.00001595, -0.00003673, -0.00004896, -0.00004888, -0.00003690, -0.00001739, 0.00000292, 0.00001747, 0.00002219, 0.00001687, 0.00000507, -0.00000736, -0.00001458, -0.00001289, -0.00000201, 0.00001498, 0.00003287, 0.00004622, 0.00005125, 0.00004691, 0.00003493, 0.00001898, 0.00000322, -0.00000904, -0.00001622, -0.00001855, -0.00001763, -0.00001554, -0.00001407, -0.00001412, -0.00001562, -0.00001764, -0.00001880, -0.00001765, },
		{ 0.00004534, 0.00004725, 0.00004306, 0.00003278, 0.00001765, 0.00000018, -0.00001629, -0.00002831, -0.00003340, -0.00003081, -0.00002196, -0.00001001, 0.00000108, 0.00000796, 0.00000907, 0.00000522, -0.00000084, -0.00000553, -0.00000609, -0.00000179, 0.00000554, 0.00001216, 0.00001394, 0.00000819, -0.00000498, -0.00002229, -0.00003824, -0.00004707, -0.00004488, -0.00003131, -0.00000983, 0.00001329, 0.00003113, 0.00003854, 0.00003397, 0.00001993, 0.00000207, -0.00001288, -0.00001953, -0.00001559, -0.00000255, 0.00001511, 0.00003160, 0.00004187, 0.00004327, 0.00003607, 0.00002309, 0.00000834, -0.00000444, -0.00001299, -0.00001694, -0.00001751, -0.00001663, -0.00001601, -0.00001645, -0.00001764, -0.00001845, -0.00001746, -0.00001358, -0.00000639, 0.00000372, 0.00001565, 0.00002781, 0.00003833, },
		{ -0.00000816, -0.00001790, -0.00002410, -0.00002545, -0.00002152, -0.00001308, -0.00000214, 0.00000848, 0.00001595, 0.00001835, 0.00001534, 0.00000834, 0.00000000, -0.00000680, -0.00001014, -0.00000980, -0.00000735, -0.00000547, -0.00000667, -0.00001199, -0.00002030, -0.00002845, -0.00003239, -0.00002881, -0.00001667, 0.00000203, 0.00002260, 0.00003913, 0.00004647, 0.00004217, 0.00002742, 0.00000679, -0.00001328, -0.00002661, -0.00002932, -0.00002105, -0.00000490, 0.00001382, 0.00002941, 0.00003771, 0.00003718, 0.00002907, 0.00001660, 0.00000351, -0.00000720, -0.00001412, -0.00001747, -0.00001853, -0.00001870, -0.00001878, -0.00001862, -0.00001733, -0.00001381, -0.00000745, 0.00000154, 0.00001203, 0.00002232, 0.00003060, 0.00003545, 0.00003612, 0.00003253, 0.00002519, 0.00001505, 0.00000342, },
		{ 0.00000265, 0.00000887, 0.00001407, 0.00001700, 0.00001660, 0.00001248, 0.00000520, -0.00000372, -0.00001218, -0.00001819, -0.00002061, -0.00001954, -0.00001632, -0.00001287, -0.00001087, -0.00001085, -0.00001191, -0.00001194, -0.00000861, -0.00000048, 0.00001203, 0.00002638, 0.00003854, 0.00004427, 0.00004074, 0.00002773, 0.00000802, -0.00001334, -0.00003053, -0.00003883, -0.00003612, -0.00002353, -0.00000498, 0.00001418, 0.00002889, 0.00003585, 0.00003433, 0.00002597, 0.00001388, 0.00000143, -0.00000887, -0.00001583, -0.00001955, -0.00002077, -0.00002025, -0.00001829, -0.00001467, -0.00000902, -0.00000128, 0.00000788, 0.00001704, 0.00002438, 0.00002824, 0.00002772, 0.00002297, 0.00001518, 0.00000611, -0.00000241, -0.00000896, -0.00001280, -0.00001381, -0.00001227, -0.00000865, -0.00000347, },
		{ -0.00000066, -0.00000560, -0.00001053, -0.00001495, -0.00001809, -0.00001918, -0.00001782, -0.00001424, -0.00000934, -0.00000435, -0.00000031, 0.00000239, 0.00000428, 0.00000657, 0.00001052, 0.00001672, 0.00002446, 0.00003168, 0.00003554, 0.00003336, 0.00002384, 0.00000778, -0.00001175, -0.00003018, -0.00004274, -0.00004598, -0.00003888, -0.00002321, -0.00000304, 0.00001655, 0.00003095, 0.00003731, 0.00003514, 0.00002610, 0.00001326, -0.00000004, -0.00001111, -0.00001843, -0.00002165, -0.00002122, -0.00001789, -0.00001240, -0.00000537, 0.00000251, 0.00001032, 0.00001681, 0.00002061, 0.00002067, 0.00001667, 0.00000937, 0.00000053, -0.00000759, -0.00001285, -0.00001395, -0.00001080, -0.00000452, 0.00000301, 0.00000981, 0.00001440, 0.00001616, 0.00001528, 0.00001248, 0.00000856, 0.00000410, },
		{ -0.00000454, -0.00000352, -0.00000265, -0.00000133, 0.00000073, 0.00000337, 0.00000620, 0.00000887, 0.00001136, 0.00001398, 0.00001713, 0.00002086, 0.00002451, 0.00002659, 0.00002519, 0.00001870, 0.00000662, -0.00000983, -0.00002767, -0.00004280, -0.00005113, -0.00004991, -0.00003870, -0.00001969, 0.00000282, 0.00002362, 0.00003805, 0.00004331, 0.00003908, 0.00002741, 0.00001190, -0.00000345, -0.00001536, -0.00002194, -0.00002291, -0.00001922, -0.00001257, -0.00000474, 0.00000270, 0.00000858, 0.00001209, 0.00001275, 0.00001048, 0.00000565, -0.00000076, -0.00000722, -0.00001189, -0.00001320, -0.00001030, -0.00000354, 0.00000554, 0.00001453, 0.00002091, 0.00002284, 0.00001979, 0.00001265, 0.00000344, -0.00000540, -0.00001184, -0.00001487, -0.00001463, -0.00001222, -0.00000906, -0.00000633, },
		{ -0.00000409, -0.00000296, 0.00000101, 0.00000658, 0.00001236, 0.00001731, 0.00002084, 0.00002266, 0.00002243, 0.00001956, 0.00001327, 0.00000304, -0.00001084, -0.00002680, -0.00004192, -0.00005253, -0.00005521, -0.00004801, -0.00003133, -0.00000816, 0.00001655, 0.00003722, 0.00004921, 0.00005019, 0.00004074, 0.00002409, 0.00000500, -0.00001171, -0.00002250, -0.00002597, -0.00002290, -0.00001564, -0.00000712, 0.00000011, 0.00000451, 0.00000569, 0.00000422, 0.00000114, -0.00000241, -0.00000549, -0.00000735, -0.00000745, -0.00000547, -0.00000139, 0.00000433, 0.00001062, 0.00001591, 0.00001851, 0.00001713, 0.00001148, 0.00000254, -0.00000754, -0.00001600, -0.00002039, -0.00001938, -0.00001327, -0.00000394, 0.00000578, 0.00001307, 0.00001608, 0.00001448, 0.00000947, 0.00000325, -0.00000183, },
		{ 0.00002067, 0.00002680, 0.00002779, 0.00002404, 0.00001664, 0.00000675, -0.00000478, -0.00001733, -0.00003014, -0.00004181, -0.00005021, -0.00005279, -0.00004742, -0.00003344, -0.00001233, 0.00001214, 0.00003479, 0.00005036, 0.00005519, 0.00004840, 0.00003223, 0.00001133, -0.00000868, -0.00002302, -0.00002911, -0.00002716, -0.00001973, -0.00001054, -0.00000306, 0.00000065, 0.00000043, -0.00000231, -0.00000541, -0.00000699, -0.00000616, -0.00000316, 0.00000090, 0.00000468, 0.00000718, 0.00000801, 0.00000733, 0.00000554, 0.00000296, -0.00000031, -0.00000421, -0.00000846, -0.00001225, -0.00001429, -0.00001316, -0.00000797, 0.00000106, 0.00001227, 0.00002288, 0.00002976, 0.00003053, 0.00002447, 0.00001297, -0.00000083, -0.00001296, -0.00001998, -0.00002001, -0.00001334, -0.00000216, 0.00001026, },
		{ -0.00000132, -0.00001639, -0.00002851, -0.00003696, -0.00004187, -0.00004346, -0.00004149, -0.00003526, -0.00002406, -0.00000812, 0.00001086, 0.00002957, 0.00004384, 0.00004993, 0.00004596, 0.00003274, 0.00001373, -0.00000597, -0.00002129, -0.00002891, -0.00002826, -0.00002149, -0.00001247, -0.00000513, -0.00000196, -0.00000318, -0.00000688, -0.00001013, -0.00001037, -0.00000663, 0.00000005, 0.00000709, 0.00001149, 0.00001119, 0.00000589, -0.00000278, -0.00001199, -0.00001879, -0.00002124, -0.00001892, -0.00001289, -0.00000492, 0.00000327, 0.00001073, 0.00001728, 0.00002313, 0.00002823, 0.00003179, 0.00003227, 0.00002807, 0.00001836, 0.00000394, -0.00001250, -0.00002698, -0.00003536, -0.00003487, -0.00002514, -0.00000856, 0.00001042, 0.00002667, 0.00003600, 0.00003641, 0.00002844, 0.00001469, },
		{ -0.00003253, -0.00002326, -0.00001423, -0.00000588, 0.00000238, 0.00001119, 0.00002034, 0.00002839, 0.00003303, 0.00003216, 0.00002496, 0.00001267, -0.00000167, -0.00001417, -0.00002160, -0.00002263, -0.00001840, -0.00001187, -0.00000660, -0.00000508, -0.00000764, -0.00001239, -0.00001611, -0.00001587, -0.00001056, -0.00000157, 0.00000754, 0.00001255, 0.00001043, 0.00000079, -0.00001366, -0.00002797, -0.00003673, -0.00003615, -0.00002546, -0.00000721, 0.00001373, 0.00003196, 0.00004345, 0.00004667, 0.00004269, 0.00003422, 0.00002418, 0.00001445, 0.00000536, -0.00000386, -0.00001400, -0.00002477, -0.00003434, -0.00003970, -0.00003790, -0.00002755, -0.00000986, 0.00001126, 0.00003023, 0.00004164, 0.00004211, 0.00003146, 0.00001279, -0.00000868, -0.00002745, -0.00003951, -0.00004340, -0.00004019, },
		{ 0.00001139, 0.00000806, 0.00000680, 0.00000802, 0.00001037, 0.00001171, 0.00001036, 0.00000603, -0.00000005, -0.00000582, -0.00000954, -0.00001071, -0.00001035, -0.00001039, -0.00001251, -0.00001702, -0.00002245, -0.00002604, -0.00002509, -0.00001851, -0.00000774, 0.00000339, 0.00001008, 0.00000858, -0.00000186, -0.00001819, -0.00003431, -0.00004316, -0.00003947, -0.00002201, 0.00000561, 0.00003600, 0.00006045, 0.00007204, 0.00006792, 0.00005017, 0.00002477, -0.00000087, -0.00002057, -0.00003139, -0.00003408, -0.00003196, -0.00002888, -0.00002721, -0.00002678, -0.00002524, -0.00001964, -0.00000831, 0.00000770, 0.00002454, 0.00003681, 0.00003972, 0.00003118, 0.00001290, -0.00001001, -0.00003080, -0.00004349, -0.00004501, -0.00003609, -0.00002072, -0.00000438, 0.00000812, 0.00001429, 0.00001453, },
		{ -0.00000249, 0.00000845, 0.00001630, 0.00001862, 0.00001521, 0.00000768, -0.00000172, -0.00001114, -0.00001973, -0.00002733, -0.00003381, -0.00003842, -0.00003967, -0.00003600, -0.00002689, -0.00001380, -0.00000025, 0.00000908, 0.00001028, 0.00000213, -0.00001275, -0.00002827, -0.00003685, -0.00003238, -0.00001298, 0.00001765, 0.00005098, 0.00007635, 0.00008482, 0.00007249, 0.00004213, 0.00000234, -0.00003550, -0.00006118, -0.00006924, -0.00006046, -0.00004108, -0.00001985, -0.00000444, 0.00000147, -0.00000086, -0.00000639, -0.00000912, -0.00000510, 0.00000562, 0.00001896, 0.00002875, 0.00002956, 0.00001939, 0.00000077, -0.00002005, -0.00003559, -0.00004015, -0.00003212, -0.00001463, 0.00000581, 0.00002190, 0.00002843, 0.00002415, 0.00001196, -0.00000252, -0.00001342, -0.00001688, -0.00001238, },
		{ 0.00002284, 0.00000439, -0.00001637, -0.00003502, -0.00004870, -0.00005621, -0.00005754, -0.00005313, -0.00004356, -0.00002976, -0.00001352, 0.00000219, 0.00001362, 0.00001742, 0.00001226, -0.00000001, -0.00001428, -0.00002356, -0.00002165, -0.00000594, 0.00002081, 0.00005061, 0.00007269, 0.00007736, 0.00005994, 0.00002306, -0.00002361, -0.00006647, -0.00009267, -0.00009484, -0.00007368, -0.00003770, 0.00000005, 0.00002696, 0.00003551, 0.00002577, 0.00000489, -0.00001619, -0.00002751, -0.00002408, -0.00000776, 0.00001390, 0.00003106, 0.00003600, 0.00002635, 0.00000607, -0.00001639, -0.00003172, -0.00003371, -0.00002180, -0.00000122, 0.00001937, 0.00003150, 0.00003040, 0.00001681, -0.00000347, -0.00002215, -0.00003182, -0.00002884, -0.00001453, 0.00000557, 0.00002411, 0.00003472, 0.00003413, },
		{ -0.00007779, -0.00007376, -0.00006096, -0.00004285, -0.00002271, -0.00000334, 0.00001283, 0.00002357, 0.00002707, 0.00002270, 0.00001184, -0.00000174, -0.00001255, -0.00001515, -0.00000650, 0.00001210, 0.00003461, 0.00005180, 0.00005453, 0.00003762, 0.00000264, -0.00004166, -0.00008185, -0.00010442, -0.00010089, -0.00007146, -0.00002538, 0.00002204, 0.00005498, 0.00006299, 0.00004495, 0.00000946, -0.00002853, -0.00005356, -0.00005567, -0.00003422, 0.00000213, 0.00003896, 0.00006200, 0.00006273, 0.00004160, 0.00000753, -0.00002592, -0.00004614, -0.00004644, -0.00002841, -0.00000084, 0.00002420, 0.00003647, 0.00003165, 0.00001277, -0.00001137, -0.00003013, -0.00003547, -0.00002514, -0.00000331, 0.00002127, 0.00003890, 0.00004260, 0.00003046, 0.00000588, -0.00002412, -0.00005164, -0.00007052, },
		{ 0.00001459, 0.00002831, 0.00003411, 0.00003266, 0.00002518, 0.00001348, 0.00000021, -0.00001112, -0.00001681, -0.00001430, -0.00000368, 0.00001140, 0.00002432, 0.00002764, 0.00001620, -0.00001006, -0.00004492, -0.00007728, -0.00009487, -0.00008906, -0.00005889, -0.00001235, 0.00003572, 0.00006883, 0.00007503, 0.00005191, 0.00000803, -0.00003988, -0.00007300, -0.00007773, -0.00005103, -0.00000191, 0.00005182, 0.00009029, 0.00009912, 0.00007494, 0.00002660, -0.00002833, -0.00007036, -0.00008547, -0.00007014, -0.00003226, 0.00001253, 0.00004722, 0.00006002, 0.00004855, 0.00002006, -0.00001216, -0.00003454, -0.00003871, -0.00002449, 0.00000048, 0.00002436, 0.00003606, 0.00002954, 0.00000588, -0.00002763, -0.00006057, -0.00008335, -0.00009035, -0.00008119, -0.00005992, -0.00003295, -0.00000659, },
		{ -0.00000003, -0.00001197, -0.00002033, -0.00002407, -0.00002252, -0.00001584, -0.00000583, 0.00000384, 0.00000833, 0.00000346, -0.00001209, -0.00003515, -0.00005825, -0.00007185, -0.00006805, -0.00004441, -0.00000604, 0.00003522, 0.00006468, 0.00007045, 0.00004865, 0.00000590, -0.00004224, -0.00007670, -0.00008221, -0.00005371, 0.00000086, 0.00006282, 0.00010932, 0.00012193, 0.00009396, 0.00003338, -0.00003970, -0.00010033, -0.00012784, -0.00011344, -0.00006330, 0.00000404, 0.00006486, 0.00009908, 0.00009740, 0.00006402, 0.00001412, -0.00003267, -0.00006005, -0.00006097, -0.00003951, -0.00000833, 0.00001717, 0.00002512, 0.00001172, -0.00001794, -0.00005247, -0.00007910, -0.00008856, -0.00007819, -0.00005213, -0.00001901, 0.00001165, 0.00003277, 0.00004141, 0.00003866, 0.00002815, 0.00001410, },
		{ -0.00001345, -0.00000583, -0.00000002, 0.00000252, 0.00000004, -0.00000852, -0.00002222, -0.00003734, -0.00004796, -0.00004798, -0.00003405, -0.00000791, 0.00002302, 0.00004770, 0.00005563, 0.00004155, 0.00000868, -0.00003133, -0.00006214, -0.00006897, -0.00004509, 0.00000432, 0.00006289, 0.00010840, 0.00012115, 0.00009201, 0.00002692, -0.00005423, -0.00012419, -0.00015817, -0.00014313, -0.00008296, 0.00000269, 0.00008532, 0.00013788, 0.00014442, 0.00010538, 0.00003677, -0.00003639, -0.00008968, -0.00010806, -0.00009058, -0.00004949, -0.00000437, 0.00002603, 0.00003092, 0.00001095, -0.00002323, -0.00005605, -0.00007368, -0.00006930, -0.00004509, -0.00001054, 0.00002201, 0.00004227, 0.00004552, 0.00003349, 0.00001258, -0.00000931, -0.00002586, -0.00003414, -0.00003454, -0.00002949, -0.00002176, },
		{ -0.00000956, -0.00001892, -0.00002763, -0.00003381, -0.00003461, -0.00002742, -0.00001177, 0.00000926, 0.00002918, 0.00004011, 0.00003627, 0.00001723, -0.00001071, -0.00003599, -0.00004615, -0.00003333, 0.00000127, 0.00004639, 0.00008395, 0.00009579, 0.00007137, 0.00001327, -0.00006212, -0.00012951, -0.00016362, -0.00014895, -0.00008644, 0.00000581, 0.00009828, 0.00016038, 0.00017142, 0.00012817, 0.00004591, -0.00004726, -0.00012076, -0.00015252, -0.00013639, -0.00008332, -0.00001616, 0.00003968, 0.00006581, 0.00005708, 0.00002219, -0.00002106, -0.00005371, -0.00006322, -0.00004783, -0.00001620, 0.00001714, 0.00003824, 0.00003929, 0.00002114, -0.00000795, -0.00003640, -0.00005422, -0.00005662, -0.00004507, -0.00002564, -0.00000583, 0.00000848, 0.00001477, 0.00001375, 0.00000787, -0.00000040, },
		{ -0.00001793, -0.00000690, 0.00000643, 0.00002020, 0.00003056, 0.00003309, 0.00002521, 0.00000830, -0.00001164, -0.00002586, -0.00002654, -0.00001095, 0.00001608, 0.00004289, 0.00005512, 0.00004194, 0.00000174, -0.00005546, -0.00011025, -0.00014059, -0.00013052, -0.00007745, 0.00000521, 0.00009162, 0.00015226, 0.00016481, 0.00012281, 0.00003879, -0.00005983, -0.00014035, -0.00017666, -0.00015848, -0.00009471, -0.00000956, 0.00006692, 0.00011006, 0.00010918, 0.00007045, 0.00001309, -0.00003911, -0.00006737, -0.00006481, -0.00003789, -0.00000250, 0.00002359, 0.00002833, 0.00001005, -0.00002263, -0.00005513, -0.00007343, -0.00006985, -0.00004576, -0.00001032, 0.00002375, 0.00004550, 0.00004959, 0.00003755, 0.00001598, -0.00000686, -0.00002446, -0.00003396, -0.00003594, -0.00003274, -0.00002648, },
		{ 0.00002585, 0.00002124, 0.00001145, -0.00000273, -0.00001747, -0.00002708, -0.00002683, -0.00001594, 0.00000088, 0.00001459, 0.00001555, -0.00000162, -0.00003435, -0.00007163, -0.00009749, -0.00009744, -0.00006528, -0.00000726, 0.00005869, 0.00010875, 0.00012226, 0.00009036, 0.00002061, -0.00006459, -0.00013562, -0.00016649, -0.00014476, -0.00007655, 0.00001544, 0.00010028, 0.00015023, 0.00015087, 0.00010581, 0.00003448, -0.00003608, -0.00008197, -0.00009128, -0.00006767, -0.00002747, 0.00000813, 0.00002241, 0.00000991, -0.00002190, -0.00005659, -0.00007637, -0.00006989, -0.00003705, 0.00001115, 0.00005742, 0.00008531, 0.00008566, 0.00005973, 0.00001789, -0.00002502, -0.00005572, -0.00006707, -0.00005963, -0.00003996, -0.00001679, 0.00000282, 0.00001570, 0.00002248, 0.00002555, 0.00002672, },
		{ -0.00002909, -0.00003381, -0.00003268, -0.00002445, -0.00001189, -0.00000122, 0.00000079, -0.00000945, -0.00002960, -0.00005138, -0.00006375, -0.00005804, -0.00003275, 0.00000445, 0.00003883, 0.00005436, 0.00004097, -0.00000006, -0.00005503, -0.00010220, -0.00012001, -0.00009609, -0.00003322, 0.00005058, 0.00012806, 0.00017279, 0.00016924, 0.00011885, 0.00003951, -0.00004138, -0.00009749, -0.00011346, -0.00009024, -0.00004363, 0.00000306, 0.00002915, 0.00002515, -0.00000374, -0.00004060, -0.00006465, -0.00006071, -0.00002603, 0.00002830, 0.00008189, 0.00011353, 0.00010985, 0.00007080, 0.00000947, -0.00005324, -0.00009681, -0.00010871, -0.00008835, -0.00004618, 0.00000145, 0.00003905, 0.00005745, 0.00005605, 0.00004113, 0.00002178, 0.00000536, -0.00000515, -0.00001112, -0.00001596, -0.00002210, },
		{ -0.00001142, -0.00000248, 0.00000139, -0.00000437, -0.00001979, -0.00003968, -0.00005589, -0.00006128, -0.00005360, -0.00003705, -0.00002050, -0.00001309, -0.00001919, -0.00003523, -0.00005034, -0.00005083, -0.00002674, 0.00002239, 0.00008548, 0.00014289, 0.00017362, 0.00016385, 0.00011333, 0.00003648, -0.00004241, -0.00009825, -0.00011474, -0.00009071, -0.00004022, 0.00001350, 0.00004788, 0.00005035, 0.00002346, -0.00001669, -0.00004784, -0.00005163, -0.00002209, 0.00003124, 0.00008695, 0.00012059, 0.00011503, 0.00006802, -0.00000635, -0.00008288, -0.00013492, -0.00014466, -0.00010970, -0.00004356, 0.00003012, 0.00008677, 0.00010996, 0.00009657, 0.00005643, 0.00000695, -0.00003442, -0.00005680, -0.00005879, -0.00004707, -0.00003176, -0.00002114, -0.00001819, -0.00002033, -0.00002216, -0.00001932, },
		{ -0.00002506, -0.00004844, -0.00007366, -0.00009149, -0.00009574, -0.00008585, -0.00006632, -0.00004376, -0.00002296, -0.00000433, 0.00001588, 0.00004259, 0.00007773, 0.00011713, 0.00015052, 0.00016499, 0.00015079, 0.00010670, 0.00004242, -0.00002376, -0.00007135, -0.00008568, -0.00006411, -0.00001759, 0.00003348, 0.00006756, 0.00007100, 0.00004395, 0.00000038, -0.00003799, -0.00005131, -0.00003081, 0.00001667, 0.00007079, 0.00010602, 0.00010260, 0.00005564, -0.00002176, -0.00010279, -0.00015710, -0.00016282, -0.00011551, -0.00003028, 0.00006370, 0.00013421, 0.00015802, 0.00012919, 0.00006049, -0.00002238, -0.00009107, -0.00012526, -0.00011925, -0.00008240, -0.00003364, 0.00000713, 0.00002698, 0.00002397, 0.00000599, -0.00001442, -0.00002666, -0.00002677, -0.00001854, -0.00001061, -0.00001151, },
		{ -0.00013670, -0.00012562, -0.00009236, -0.00004520, 0.00000622, 0.00005458, 0.00009635, 0.00013074, 0.00015722, 0.00017345, 0.00017519, 0.00015842, 0.00012265, 0.00007341, 0.00002215, -0.00001702, -0.00003268, -0.00002134, 0.00001022, 0.00004688, 0.00007104, 0.00007014, 0.00004251, -0.00000124, -0.00004233, -0.00006189, -0.00004969, -0.00000961, 0.00004093, 0.00007777, 0.00008030, 0.00004115, -0.00002924, -0.00010545, -0.00015654, -0.00015863, -0.00010527, -0.00001133, 0.00009163, 0.00016684, 0.00018635, 0.00014187, 0.00004812, -0.00006257, -0.00015253, -0.00019271, -0.00017309, -0.00010541, -0.00001756, 0.00005790, 0.00009618, 0.00008889, 0.00004550, -0.00001228, -0.00006041, -0.00008245, -0.00007536, -0.00004921, -0.00002142, -0.00000848, -0.00001885, -0.00004988, -0.00008973, -0.00012291, },
		{ 0.00005278, 0.00013466, 0.00019563, 0.00023220, 0.00024556, 0.00023882, 0.00021527, 0.00017839, 0.00013294, 0.00008583, 0.00004556, 0.00001983, 0.00001232, 0.00002035, 0.00003500, 0.00004423, 0.00003805, 0.00001347, -0.00002325, -0.00005849, -0.00007694, -0.00006884, -0.00003545, 0.00001007, 0.00004724, 0.00005685, 0.00003002, -0.00002641, -0.00009101, -0.00013545, -0.00013616, -0.00008514, 0.00000492, 0.00010373, 0.00017412, 0.00018645, 0.00013094, 0.00002279, -0.00010235, -0.00020116, -0.00023878, -0.00020210, -0.00010481, 0.00001767, 0.00012165, 0.00017145, 0.00015242, 0.00007552, -0.00002799, -0.00011929, -0.00016710, -0.00015906, -0.00010517, -0.00003248, 0.00002650, 0.00004672, 0.00001970, -0.00004431, -0.00012132, -0.00018334, -0.00020849, -0.00018779, -0.00012638, -0.00003977, },
		{ 0.00038107, 0.00033488, 0.00027254, 0.00020450, 0.00013869, 0.00008115, 0.00003642, 0.00000699, -0.00000783, -0.00001247, -0.00001409, -0.00001998, -0.00003418, -0.00005504, -0.00007528, -0.00008497, -0.00007647, -0.00004908, -0.00001101, 0.00002285, 0.00003695, 0.00002242, -0.00001755, -0.00006714, -0.00010325, -0.00010498, -0.00006354, 0.00001196, 0.00009584, 0.00015460, 0.00016014, 0.00010220, -0.00000543, -0.00012792, -0.00022134, -0.00024921, -0.00019720, -0.00008008, 0.00006244, 0.00017969, 0.00022878, 0.00019089, 0.00007873, -0.00006793, -0.00019692, -0.00026318, -0.00024532, -0.00015337, -0.00002449, 0.00009112, 0.00014890, 0.00012590, 0.00002841, -0.00011121, -0.00024659, -0.00033354, -0.00034416, -0.00027422, -0.00014196, 0.00002011, 0.00017663, 0.00029957, 0.00037396, 0.00039836, },
		{ 0.00005074, -0.00001497, -0.00005770, -0.00008083, -0.00008894, -0.00008761, -0.00008271, -0.00007897, -0.00007835, -0.00007916, -0.00007684, -0.00006650, -0.00004619, -0.00001920, 0.00000616, 0.00001959, 0.00001384, -0.00001074, -0.00004451, -0.00007123, -0.00007460, -0.00004616, 0.00000904, 0.00007212, 0.00011640, 0.00011806, 0.00006723, -0.00002538, -0.00012978, -0.00020656, -0.00022190, -0.00016210, -0.00004148, 0.00010040, 0.00021181, 0.00024804, 0.00018922, 0.00004972, -0.00012509, -0.00027422, -0.00034322, -0.00030475, -0.00016987, 0.00001465, 0.00018220, 0.00026974, 0.00024024, 0.00009605, -0.00012152, -0.00034591, -0.00050726, -0.00055445, -0.00047007, -0.00027353, -0.00001249, 0.00025341, 0.00047039, 0.00060389, 0.00064395, 0.00060250, 0.00050498, 0.00038041, 0.00025356, 0.00014094, },
		{ -0.00014617, -0.00012013, -0.00009611, -0.00007681, -0.00006287, -0.00005285, -0.00004387, -0.00003324, -0.00002030, -0.00000741, 0.00000073, -0.00000045, -0.00001232, -0.00003082, -0.00004696, -0.00005019, -0.00003363, 0.00000121, 0.00004287, 0.00007310, 0.00007412, 0.00003728, -0.00003076, -0.00010772, -0.00016256, -0.00016753, -0.00011085, -0.00000463, 0.00011634, 0.00020548, 0.00022186, 0.00014750, -0.00000240, -0.00018144, -0.00032661, -0.00038082, -0.00031536, -0.00014344, 0.00008100, 0.00027963, 0.00037623, 0.00032388, 0.00012317, -0.00017540, -0.00048598, -0.00071392, -0.00078535, -0.00066968, -0.00038788, -0.00000453, 0.00039294, 0.00071997, 0.00091721, 0.00096279, 0.00087209, 0.00068712, 0.00046039, 0.00023961, 0.00005752, -0.00007130, -0.00014748, -0.00018087, -0.00018435, -0.00016975, },
		{ -0.00002532, -0.00002396, -0.00002464, -0.00002493, -0.00002367, -0.00002149, -0.00002035, -0.00002214, -0.00002684, -0.00003168, -0.00003184, -0.00002297, -0.00000436, 0.00001903, 0.00003728, 0.00003925, 0.00001825, -0.00002301, -0.00007093, -0.00010468, -0.00010441, -0.00006094, 0.00001725, 0.00010372, 0.00016187, 0.00015885, 0.00008062, -0.00005846, -0.00021618, -0.00033491, -0.00036220, -0.00027253, -0.00008145, 0.00015432, 0.00035244, 0.00043202, 0.00034295, 0.00008710, -0.00027702, -0.00064872, -0.00091460, -0.00098360, -0.00081578, -0.00043541, 0.00007588, 0.00060314, 0.00103116, 0.00127564, 0.00130288, 0.00113315, 0.00082834, 0.00046975, 0.00013413, -0.00012451, -0.00028286, -0.00034584, -0.00033694, -0.00028621, -0.00022036, -0.00015748, -0.00010649, -0.00006958, -0.00004542, -0.00003155, },
		{ -0.00004261, -0.00003843, -0.00003261, -0.00002678, -0.00002187, -0.00001737, -0.00001170, -0.00000351, 0.00000661, 0.00001537, 0.00001771, 0.00000942, -0.00000960, -0.00003341, -0.00005097, -0.00005049, -0.00002557, 0.00001941, 0.00006788, 0.00009555, 0.00007983, 0.00001124, -0.00009848, -0.00021582, -0.00029463, -0.00029289, -0.00019144, -0.00000707, 0.00020643, 0.00037168, 0.00041270, 0.00028342, -0.00001063, -0.00040623, -0.00079583, -0.00105761, -0.00109284, -0.00085848, -0.00038374, 0.00023531, 0.00086167, 0.00135570, 0.00161338, 0.00159276, 0.00132096, 0.00088091, 0.00038373, -0.00006370, -0.00038528, -0.00054953, -0.00056825, -0.00048342, -0.00034844, -0.00021089, -0.00010178, -0.00003290, -0.00000096, 0.00000520, -0.00000225, -0.00001426, -0.00002588, -0.00003509, -0.00004117, -0.00004372, },
		{ -0.00000662, -0.00000168, 0.00000160, 0.00000415, 0.00000641, 0.00000764, 0.00000622, 0.00000101, -0.00000713, -0.00001468, -0.00001658, -0.00000913, 0.00000667, 0.00002328, 0.00002816, 0.00000877, -0.00004019, -0.00011086, -0.00018084, -0.00021873, -0.00019589, -0.00010060, 0.00005119, 0.00021415, 0.00032388, 0.00031752, 0.00015861, -0.00014301, -0.00052457, -0.00088184, -0.00109692, -0.00107449, -0.00077494, -0.00023260, 0.00044809, 0.00111838, 0.00162566, 0.00185578, 0.00176411, 0.00138566, 0.00082181, 0.00020882, -0.00032112, -0.00067399, -0.00081318, -0.00075986, -0.00057690, -0.00034357, -0.00013055, 0.00001734, 0.00008654, 0.00008945, 0.00005250, 0.00000342, -0.00003785, -0.00006248, -0.00007096, -0.00006858, -0.00006101, -0.00005184, -0.00004231, -0.00003252, -0.00002273, -0.00001377, },
		{ 0.00000494, 0.00000499, 0.00000473, 0.00000327, 0.00000020, -0.00000389, -0.00000770, -0.00001053, -0.00001371, -0.00002096, -0.00003686, -0.00006349, -0.00009690, -0.00012551, -0.00013245, -0.00010217, -0.00002956, 0.00007240, 0.00017064, 0.00021868, 0.00017161, 0.00000521, -0.00026757, -0.00059046, -0.00087243, -0.00101094, -0.00092384, -0.00057977, -0.00001613, 0.00066305, 0.00131061, 0.00177523, 0.00194401, 0.00177523, 0.00131061, 0.00066305, -0.00001613, -0.00057977, -0.00092384, -0.00101094, -0.00087243, -0.00059046, -0.00026757, 0.00000521, 0.00017161, 0.00021868, 0.00017064, 0.00007240, -0.00002956, -0.00010217, -0.00013245, -0.00012551, -0.00009690, -0.00006349, -0.00003686, -0.00002096, -0.00001371, -0.00001053, -0.00000770, -0.00000389, 0.00000020, 0.00000327, 0.00000473, 0.00000499, },
		{ -0.00000662, -0.00001377, -0.00002273, -0.00003252, -0.00004231, -0.00005184, -0.00006101, -0.00006858, -0.00007096, -0.00006248, -0.00003785, 0.00000342, 0.00005250, 0.00008945, 0.00008654, 0.00001734, -0.00013055, -0.00034357, -0.00057690, -0.00075986, -0.00081318, -0.00067399, -0.00032112, 0.00020882, 0.00082181, 0.00138566, 0.00176411, 0.00185578, 0.00162566, 0.00111838, 0.00044809, -0.00023260, -0.00077494, -0.00107449, -0.00109692, -0.00088184, -0.00052457, -0.00014301, 0.00015861, 0.00031752, 0.00032388, 0.00021415, 0.00005119, -0.00010060, -0.00019589, -0.00021873, -0.00018084, -0.00011086, -0.00004019, 0.00000877, 0.00002816, 0.00002328, 0.00000667, -0.00000913, -0.00001658, -0.00001468, -0.00000713, 0.00000101, 0.00000622, 0.00000764, 0.00000641, 0.00000415, 0.00000160, -0.00000168, },
		{ -0.00004261, -0.00004372, -0.00004117, -0.00003509, -0.00002588, -0.00001426, -0.00000225, 0.00000520, -0.00000096, -0.00003290, -0.00010178, -0.00021089, -0.00034844, -0.00048342, -0.00056825, -0.00054953, -0.00038528, -0.00006370, 0.00038373, 0.00088091, 0.00132096, 0.00159276, 0.00161338, 0.00135570, 0.00086167, 0.00023531, -0.00038374, -0.00085848, -0.00109284, -0.00105761, -0.00079583, -0.00040623, -0.00001063, 0.00028342, 0.00041270, 0.00037168, 0.00020643, -0.00000707, -0.00019144, -0.00029289, -0.00029463, -0.00021582, -0.00009848, 0.00001124, 0.00007983, 0.00009555, 0.00006788, 0.00001941, -0.00002557, -0.00005049, -0.00005097, -0.00003341, -0.00000960, 0.00000942, 0.00001771, 0.00001537, 0.00000661, -0.00000351, -0.00001170, -0.00001737, -0.00002187, -0.00002678, -0.00003261, -0.00003843, },
		{ -0.00002532, -0.00003155, -0.00004542, -0.00006958, -0.00010649, -0.00015748, -0.00022036, -0.00028621, -0.00033694, -0.00034584, -0.00028286, -0.00012451, 0.00013413, 0.00046975, 0.00082834, 0.00113315, 0.00130288, 0.00127564, 0.00103116, 0.00060314, 0.00007588, -0.00043541, -0.00081578, -0.00098360, -0.00091460, -0.00064872, -0.00027702, 0.00008710, 0.00034295, 0.00043202, 0.00035244, 0.00015432, -0.00008145, -0.00027253, -0.00036220, -0.00033491, -0.00021618, -0.00005846, 0.00008062, 0.00015885, 0.00016187, 0.00010372, 0.00001725, -0.00006094, -0.00010441, -0.00010468, -0.00007093, -0.00002301, 0.00001825, 0.00003925, 0.00003728, 0.00001903, -0.00000436, -0.00002297, -0.00003184, -0.00003168, -0.00002684, -0.00002214, -0.00002035, -0.00002149, -0.00002367, -0.00002493, -0.00002464, -0.00002396, },
		{ -0.00014617, -0.00016975, -0.00018435, -0.00018087, -0.00014748, -0.00007130, 0.00005752, 0.00023961, 0.00046039, 0.00068712, 0.00087209, 0.00096279, 0.00091721, 0.00071997, 0.00039294, -0.00000453, -0.00038788, -0.00066968, -0.00078535, -0.00071392, -0.00048598, -0.00017540, 0.00012317, 0.00032388, 0.00037623, 0.00027963, 0.00008100, -0.00014344, -0.00031536, -0.00038082, -0.00032661, -0.00018144, -0.00000240, 0.00014750, 0.00022186, 0.00020548, 0.00011634, -0.00000463, -0.00011085, -0.00016753, -0.00016256, -0.00010772, -0.00003076, 0.00003728, 0.00007412, 0.00007310, 0.00004287, 0.00000121, -0.00003363, -0.00005019, -0.00004696, -0.00003082, -0.00001232, -0.00000045, 0.00000073, -0.00000741, -0.00002030, -0.00003324, -0.00004387, -0.00005285, -0.00006287, -0.00007681, -0.00009611, -0.00012013, },
		{ 0.00005074, 0.00014094, 0.00025356, 0.00038041, 0.00050498, 0.00060250, 0.00064395, 0.00060389, 0.00047039, 0.00025341, -0.00001249, -0.00027353, -0.00047007, -0.00055445, -0.00050726, -0.00034591, -0.00012152, 0.00009605, 0.00024024, 0.00026974, 0.00018220, 0.00001465, -0.00016987, -0.00030475, -0.00034322, -0.00027422, -0.00012509, 0.00004972, 0.00018922, 0.00024804, 0.00021181, 0.00010040, -0.00004148, -0.00016210, -0.00022190, -0.00020656, -0.00012978, -0.00002538, 0.00006723, 0.00011806, 0.00011640, 0.00007212, 0.00000904, -0.00004616, -0.00007460, -0.00007123, -0.00004451, -0.00001074, 0.00001384, 0.00001959, 0.00000616, -0.00001920, -0.00004619, -0.00006650, -0.00007684, -0.00007916, -0.00007835, -0.00007897, -0.00008271, -0.00008761, -0.00008894, -0.00008083, -0.00005770, -0.00001497, },
		{ 0.00038107, 0.00039836, 0.00037396, 0.00029957, 0.00017663, 0.00002011, -0.00014196, -0.00027422, -0.00034416, -0.00033354, -0.00024659, -0.00011121, 0.00002841, 0.00012590, 0.00014890, 0.00009112, -0.00002449, -0.00015337, -0.00024532, -0.00026318, -0.00019692, -0.00006793, 0.00007873, 0.00019089, 0.00022878, 0.00017969, 0.00006244, -0.00008008, -0.00019720, -0.00024921, -0.00022134, -0.00012792, -0.00000543, 0.00010220, 0.00016014, 0.00015460, 0.00009584, 0.00001196, -0.00006354, -0.00010498, -0.00010325, -0.00006714, -0.00001755, 0.00002242, 0.00003695, 0.00002285, -0.00001101, -0.00004908, -0.00007647, -0.00008497, -0.00007528, -0.00005504, -0.00003418, -0.00001998, -0.00001409, -0.00001247, -0.00000783, 0.00000699, 0.00003642, 0.00008115, 0.00013869, 0.00020450, 0.00027254, 0.00033488, },
		{ 0.00005278, -0.00003977, -0.00012638, -0.00018779, -0.00020849, -0.00018334, -0.00012132, -0.00004431, 0.00001970, 0.00004672, 0.00002650, -0.00003248, -0.00010517, -0.00015906, -0.00016710, -0.00011929, -0.00002799, 0.00007552, 0.00015242, 0.00017145, 0.00012165, 0.00001767, -0.00010481, -0.00020210, -0.00023878, -0.00020116, -0.00010235, 0.00002279, 0.00013094, 0.00018645, 0.00017412, 0.00010373, 0.00000492, -0.00008514, -0.00013616, -0.00013545, -0.00009101, -0.00002641, 0.00003002, 0.00005685, 0.00004724, 0.00001007, -0.00003545, -0.00006884, -0.00007694, -0.00005849, -0.00002325, 0.00001347, 0.00003805, 0.00004423, 0.00003500, 0.00002035, 0.00001232, 0.00001983, 0.00004556, 0.00008583, 0.00013294, 0.00017839, 0.00021527, 0.00023882, 0.00024556, 0.00023220, 0.00019563, 0.00013466, },
		{ -0.00013670, -0.00012291, -0.00008973, -0.00004988, -0.00001885, -0.00000848, -0.00002142, -0.00004921, -0.00007536, -0.00008245, -0.00006041, -0.00001228, 0.00004550, 0.00008889, 0.00009618, 0.00005790, -0.00001756, -0.00010541, -0.00017309, -0.00019271, -0.00015253, -0.00006257, 0.00004812, 0.00014187, 0.00018635, 0.00016684, 0.00009163, -0.00001133, -0.00010527, -0.00015863, -0.00015654, -0.00010545, -0.00002924, 0.00004115, 0.00008030, 0.00007777, 0.00004093, -0.00000961, -0.00004969, -0.00006189, -0.00004233, -0.00000124, 0.00004251, 0.00007014, 0.00007104, 0.00004688, 0.00001022, -0.00002134, -0.00003268, -0.00001702, 0.00002215, 0.00007341, 0.00012265, 0.00015842, 0.00017519, 0.00017345, 0.00015722, 0.00013074, 0.00009635, 0.00005458, 0.00000622, -0.00004520, -0.00009236, -0.00012562, },
		{ -0.00002506, -0.00001151, -0.00001061, -0.00001854, -0.00002677, -0.00002666, -0.00001442, 0.00000599, 0.00002397, 0.00002698, 0.00000713, -0.00003364, -0.00008240, -0.00011925, -0.00012526, -0.00009107, -0.00002238, 0.00006049, 0.00012919, 0.00015802, 0.00013421, 0.00006370, -0.00003028, -0.00011551, -0.00016282, -0.00015710, -0.00010279, -0.00002176, 0.00005564, 0.00010260, 0.00010602, 0.00007079, 0.00001667, -0.00003081, -0.00005131, -0.00003799, 0.00000038, 0.00004395, 0.00007100, 0.00006756, 0.00003348, -0.00001759, -0.00006411, -0.00008568, -0.00007135, -0.00002376, 0.00004242, 0.00010670, 0.00015079, 0.00016499, 0.00015052, 0.00011713, 0.00007773, 0.00004259, 0.00001588, -0.00000433, -0.00002296, -0.00004376, -0.00006632, -0.00008585, -0.00009574, -0.00009149, -0.00007366, -0.00004844, },
		{ -0.00001142, -0.00001932, -0.00002216, -0.00002033, -0.00001819, -0.00002114, -0.00003176, -0.00004707, -0.00005879, -0.00005680, -0.00003442, 0.00000695, 0.00005643, 0.00009657, 0.00010996, 0.00008677, 0.00003012, -0.00004356, -0.00010970, -0.00014466, -0.00013492, -0.00008288, -0.00000635, 0.00006802, 0.00011503, 0.00012059, 0.00008695, 0.00003124, -0.00002209, -0.00005163, -0.00004784, -0.00001669, 0.00002346, 0.00005035, 0.00004788, 0.00001350, -0.00004022, -0.00009071, -0.00011474, -0.00009825, -0.00004241, 0.00003648, 0.00011333, 0.00016385, 0.00017362, 0.00014289, 0.00008548, 0.00002239, -0.00002674, -0.00005083, -0.00005034, -0.00003523, -0.00001919, -0.00001309, -0.00002050, -0.00003705, -0.00005360, -0.00006128, -0.00005589, -0.00003968, -0.00001979, -0.00000437, 0.00000139, -0.00000248, },
		{ -0.00002909, -0.00002210, -0.00001596, -0.00001112, -0.00000515, 0.00000536, 0.00002178, 0.00004113, 0.00005605, 0.00005745, 0.00003905, 0.00000145, -0.00004618, -0.00008835, -0.00010871, -0.00009681, -0.00005324, 0.00000947, 0.00007080, 0.00010985, 0.00011353, 0.00008189, 0.00002830, -0.00002603, -0.00006071, -0.00006465, -0.00004060, -0.00000374, 0.00002515, 0.00002915, 0.00000306, -0.00004363, -0.00009024, -0.00011346, -0.00009749, -0.00004138, 0.00003951, 0.00011885, 0.00016924, 0.00017279, 0.00012806, 0.00005058, -0.00003322, -0.00009609, -0.00012001, -0.00010220, -0.00005503, -0.00000006, 0.00004097, 0.00005436, 0.00003883, 0.00000445, -0.00003275, -0.00005804, -0.00006375, -0.00005138, -0.00002960, -0.00000945, 0.00000079, -0.00000122, -0.00001189, -0.00002445, -0.00003268, -0.00003381, },
		{ 0.00002585, 0.00002672, 0.00002555, 0.00002248, 0.00001570, 0.00000282, -0.00001679, -0.00003996, -0.00005963, -0.00006707, -0.00005572, -0.00002502, 0.00001789, 0.00005973, 0.00008566, 0.00008531, 0.00005742, 0.00001115, -0.00003705, -0.00006989, -0.00007637, -0.00005659, -0.00002190, 0.00000991, 0.00002241, 0.00000813, -0.00002747, -0.00006767, -0.00009128, -0.00008197, -0.00003608, 0.00003448, 0.00010581, 0.00015087, 0.00015023, 0.00010028, 0.00001544, -0.00007655, -0.00014476, -0.00016649, -0.00013562, -0.00006459, 0.00002061, 0.00009036, 0.00012226, 0.00010875, 0.00005869, -0.00000726, -0.00006528, -0.00009744, -0.00009749, -0.00007163, -0.00003435, -0.00000162, 0.00001555, 0.00001459, 0.00000088, -0.00001594, -0.00002683, -0.00002708, -0.00001747, -0.00000273, 0.00001145, 0.00002124, },
		{ -0.00001793, -0.00002648, -0.00003274, -0.00003594, -0.00003396, -0.00002446, -0.00000686, 0.00001598, 0.00003755, 0.00004959, 0.00004550, 0.00002375, -0.00001032, -0.00004576, -0.00006985, -0.00007343, -0.00005513, -0.00002263, 0.00001005, 0.00002833, 0.00002359, -0.00000250, -0.00003789, -0.00006481, -0.00006737, -0.00003911, 0.00001309, 0.00007045, 0.00010918, 0.00011006, 0.00006692, -0.00000956, -0.00009471, -0.00015848, -0.00017666, -0.00014035, -0.00005983, 0.00003879, 0.00012281, 0.00016481, 0.00015226, 0.00009162, 0.00000521, -0.00007745, -0.00013052, -0.00014059, -0.00011025, -0.00005546, 0.00000174, 0.00004194, 0.00005512, 0.00004289, 0.00001608, -0.00001095, -0.00002654, -0.00002586, -0.00001164, 0.00000830, 0.00002521, 0.00003309, 0.00003056, 0.00002020, 0.00000643, -0.00000690, },
		{ -0.00000956, -0.00000040, 0.00000787, 0.00001375, 0.00001477, 0.00000848, -0.00000583, -0.00002564, -0.00004507, -0.00005662, -0.00005422, -0.00003640, -0.00000795, 0.00002114, 0.00003929, 0.00003824, 0.00001714, -0.00001620, -0.00004783, -0.00006322, -0.00005371, -0.00002106, 0.00002219, 0.00005708, 0.00006581, 0.00003968, -0.00001616, -0.00008332, -0.00013639, -0.00015252, -0.00012076, -0.00004726, 0.00004591, 0.00012817, 0.00017142, 0.00016038, 0.00009828, 0.00000581, -0.00008644, -0.00014895, -0.00016362, -0.00012951, -0.00006212, 0.00001327, 0.00007137, 0.00009579, 0.00008395, 0.00004639, 0.00000127, -0.00003333, -0.00004615, -0.00003599, -0.00001071, 0.00001723, 0.00003627, 0.00004011, 0.00002918, 0.00000926, -0.00001177, -0.00002742, -0.00003461, -0.00003381, -0.00002763, -0.00001892, },
		{ -0.00001345, -0.00002176, -0.00002949, -0.00003454, -0.00003414, -0.00002586, -0.00000931, 0.00001258, 0.00003349, 0.00004552, 0.00004227, 0.00002201, -0.00001054, -0.00004509, -0.00006930, -0.00007368, -0.00005605, -0.00002323, 0.00001095, 0.00003092, 0.00002603, -0.00000437, -0.00004949, -0.00009058, -0.00010806, -0.00008968, -0.00003639, 0.00003677, 0.00010538, 0.00014442, 0.00013788, 0.00008532, 0.00000269, -0.00008296, -0.00014313, -0.00015817, -0.00012419, -0.00005423, 0.00002692, 0.00009201, 0.00012115, 0.00010840, 0.00006289, 0.00000432, -0.00004509, -0.00006897, -0.00006214, -0.00003133, 0.00000868, 0.00004155, 0.00005563, 0.00004770, 0.00002302, -0.00000791, -0.00003405, -0.00004798, -0.00004796, -0.00003734, -0.00002222, -0.00000852, 0.00000004, 0.00000252, -0.00000002, -0.00000583, },
		{ -0.00000003, 0.00001410, 0.00002815, 0.00003866, 0.00004141, 0.00003277, 0.00001165, -0.00001901, -0.00005213, -0.00007819, -0.00008856, -0.00007910, -0.00005247, -0.00001794, 0.00001172, 0.00002512, 0.00001717, -0.00000833, -0.00003951, -0.00006097, -0.00006005, -0.00003267, 0.00001412, 0.00006402, 0.00009740, 0.00009908, 0.00006486, 0.00000404, -0.00006330, -0.00011344, -0.00012784, -0.00010033, -0.00003970, 0.00003338, 0.00009396, 0.00012193, 0.00010932, 0.00006282, 0.00000086, -0.00005371, -0.00008221, -0.00007670, -0.00004224, 0.00000590, 0.00004865, 0.00007045, 0.00006468, 0.00003522, -0.00000604, -0.00004441, -0.00006805, -0.00007185, -0.00005825, -0.00003515, -0.00001209, 0.00000346, 0.00000833, 0.00000384, -0.00000583, -0.00001584, -0.00002252, -0.00002407, -0.00002033, -0.00001197, },
		{ 0.00001459, -0.00000659, -0.00003295, -0.00005992, -0.00008119, -0.00009035, -0.00008335, -0.00006057, -0.00002763, 0.00000588, 0.00002954, 0.00003606, 0.00002436, 0.00000048, -0.00002449, -0.00003871, -0.00003454, -0.00001216, 0.00002006, 0.00004855, 0.00006002, 0.00004722, 0.00001253, -0.00003226, -0.00007014, -0.00008547, -0.00007036, -0.00002833, 0.00002660, 0.00007494, 0.00009912, 0.00009029, 0.00005182, -0.00000191, -0.00005103, -0.00007773, -0.00007300, -0.00003988, 0.00000803, 0.00005191, 0.00007503, 0.00006883, 0.00003572, -0.00001235, -0.00005889, -0.00008906, -0.00009487, -0.00007728, -0.00004492, -0.00001006, 0.00001620, 0.00002764, 0.00002432, 0.00001140, -0.00000368, -0.00001430, -0.00001681, -0.00001112, 0.00000021, 0.00001348, 0.00002518, 0.00003266, 0.00003411, 0.00002831, },
		{ -0.00007779, -0.00007052, -0.00005164, -0.00002412, 0.00000588, 0.00003046, 0.00004260, 0.00003890, 0.00002127, -0.00000331, -0.00002514, -0.00003547, -0.00003013, -0.00001137, 0.00001277, 0.00003165, 0.00003647, 0.00002420, -0.00000084, -0.00002841, -0.00004644, -0.00004614, -0.00002592, 0.00000753, 0.00004160, 0.00006273, 0.00006200, 0.00003896, 0.00000213, -0.00003422, -0.00005567, -0.00005356, -0.00002853, 0.00000946, 0.00004495, 0.00006299, 0.00005498, 0.00002204, -0.00002538, -0.00007146, -0.00010089, -0.00010442, -0.00008185, -0.00004166, 0.00000264, 0.00003762, 0.00005453, 0.00005180, 0.00003461, 0.00001210, -0.00000650, -0.00001515, -0.00001255, -0.00000174, 0.00001184, 0.00002270, 0.00002707, 0.00002357, 0.00001283, -0.00000334, -0.00002271, -0.00004285, -0.00006096, -0.00007376, },
		{ 0.00002284, 0.00003413, 0.00003472, 0.00002411, 0.00000557, -0.00001453, -0.00002884, -0.00003182, -0.00002215, -0.00000347, 0.00001681, 0.00003040, 0.00003150, 0.00001937, -0.00000122, -0.00002180, -0.00003371, -0.00003172, -0.00001639, 0.00000607, 0.00002635, 0.00003600, 0.00003106, 0.00001390, -0.00000776, -0.00002408, -0.00002751, -0.00001619, 0.00000489, 0.00002577, 0.00003551, 0.00002696, 0.00000005, -0.00003770, -0.00007368, -0.00009484, -0.00009267, -0.00006647, -0.00002361, 0.00002306, 0.00005994, 0.00007736, 0.00007269, 0.00005061, 0.00002081, -0.00000594, -0.00002165, -0.00002356, -0.00001428, -0.00000001, 0.00001226, 0.00001742, 0.00001362, 0.00000219, -0.00001352, -0.00002976, -0.00004356, -0.00005313, -0.00005754, -0.00005621, -0.00004870, -0.00003502, -0.00001637, 0.00000439, },
		{ -0.00000249, -0.00001238, -0.00001688, -0.00001342, -0.00000252, 0.00001196, 0.00002415, 0.00002843, 0.00002190, 0.00000581, -0.00001463, -0.00003212, -0.00004015, -0.00003559, -0.00002005, 0.00000077, 0.00001939, 0.00002956, 0.00002875, 0.00001896, 0.00000562, -0.00000510, -0.00000912, -0.00000639, -0.00000086, 0.00000147, -0.00000444, -0.00001985, -0.00004108, -0.00006046, -0.00006924, -0.00006118, -0.00003550, 0.00000234, 0.00004213, 0.00007249, 0.00008482, 0.00007635, 0.00005098, 0.00001765, -0.00001298, -0.00003238, -0.00003685, -0.00002827, -0.00001275, 0.00000213, 0.00001028, 0.00000908, -0.00000025, -0.00001380, -0.00002689, -0.00003600, -0.00003967, -0.00003842, -0.00003381, -0.00002733, -0.00001973, -0.00001114, -0.00000172, 0.00000768, 0.00001521, 0.00001862, 0.00001630, 0.00000845, },
		{ 0.00001139, 0.00001453, 0.00001429, 0.00000812, -0.00000438, -0.00002072, -0.00003609, -0.00004501, -0.00004349, -0.00003080, -0.00001001, 0.00001290, 0.00003118, 0.00003972, 0.00003681, 0.00002454, 0.00000770, -0.00000831, -0.00001964, -0.00002524, -0.00002678, -0.00002721, -0.00002888, -0.00003196, -0.00003408, -0.00003139, -0.00002057, -0.00000087, 0.00002477, 0.00005017, 0.00006792, 0.00007204, 0.00006045, 0.00003600, 0.00000561, -0.00002201, -0.00003947, -0.00004316, -0.00003431, -0.00001819, -0.00000186, 0.00000858, 0.00001008, 0.00000339, -0.00000774, -0.00001851, -0.00002509, -0.00002604, -0.00002245, -0.00001702, -0.00001251, -0.00001039, -0.00001035, -0.00001071, -0.00000954, -0.00000582, -0.00000005, 0.00000603, 0.00001036, 0.00001171, 0.00001037, 0.00000802, 0.00000680, 0.00000806, },
		{ -0.00003253, -0.00004019, -0.00004340, -0.00003951, -0.00002745, -0.00000868, 0.00001279, 0.00003146, 0.00004211, 0.00004164, 0.00003023, 0.00001126, -0.00000986, -0.00002755, -0.00003790, -0.00003970, -0.00003434, -0.00002477, -0.00001400, -0.00000386, 0.00000536, 0.00001445, 0.00002418, 0.00003422, 0.00004269, 0.00004667, 0.00004345, 0.00003196, 0.00001373, -0.00000721, -0.00002546, -0.00003615, -0.00003673, -0.00002797, -0.00001366, 0.00000079, 0.00001043, 0.00001255, 0.00000754, -0.00000157, -0.00001056, -0.00001587, -0.00001611, -0.00001239, -0.00000764, -0.00000508, -0.00000660, -0.00001187, -0.00001840, -0.00002263, -0.00002160, -0.00001417, -0.00000167, 0.00001267, 0.00002496, 0.00003216, 0.00003303, 0.00002839, 0.00002034, 0.00001119, 0.00000238, -0.00000588, -0.00001423, -0.00002326, },
		{ -0.00000132, 0.00001469, 0.00002844, 0.00003641, 0.00003600, 0.00002667, 0.00001042, -0.00000856, -0.00002514, -0.00003487, -0.00003536, -0.00002698, -0.00001250, 0.00000394, 0.00001836, 0.00002807, 0.00003227, 0.00003179, 0.00002823, 0.00002313, 0.00001728, 0.00001073, 0.00000327, -0.00000492, -0.00001289, -0.00001892, -0.00002124, -0.00001879, -0.00001199, -0.00000278, 0.00000589, 0.00001119, 0.00001149, 0.00000709, 0.00000005, -0.00000663, -0.00001037, -0.00001013, -0.00000688, -0.00000318, -0.00000196, -0.00000513, -0.00001247, -0.00002149, -0.00002826, -0.00002891, -0.00002129, -0.00000597, 0.00001373, 0.00003274, 0.00004596, 0.00004993, 0.00004384, 0.00002957, 0.00001086, -0.00000812, -0.00002406, -0.00003526, -0.00004149, -0.00004346, -0.00004187, -0.00003696, -0.00002851, -0.00001639, },
		{ 0.00002067, 0.00001026, -0.00000216, -0.00001334, -0.00002001, -0.00001998, -0.00001296, -0.00000083, 0.00001297, 0.00002447, 0.00003053, 0.00002976, 0.00002288, 0.00001227, 0.00000106, -0.00000797, -0.00001316, -0.00001429, -0.00001225, -0.00000846, -0.00000421, -0.00000031, 0.00000296, 0.00000554, 0.00000733, 0.00000801, 0.00000718, 0.00000468, 0.00000090, -0.00000316, -0.00000616, -0.00000699, -0.00000541, -0.00000231, 0.00000043, 0.00000065, -0.00000306, -0.00001054, -0.00001973, -0.00002716, -0.00002911, -0.00002302, -0.00000868, 0.00001133, 0.00003223, 0.00004840, 0.00005519, 0.00005036, 0.00003479, 0.00001214, -0.00001233, -0.00003344, -0.00004742, -0.00005279, -0.00005021, -0.00004181, -0.00003014, -0.00001733, -0.00000478, 0.00000675, 0.00001664, 0.00002404, 0.00002779, 0.00002680, },
		{ -0.00000409, -0.00000183, 0.00000325, 0.00000947, 0.00001448, 0.00001608, 0.00001307, 0.00000578, -0.00000394, -0.00001327, -0.00001938, -0.00002039, -0.00001600, -0.00000754, 0.00000254, 0.00001148, 0.00001713, 0.00001851, 0.00001591, 0.00001062, 0.00000433, -0.00000139, -0.00000547, -0.00000745, -0.00000735, -0.00000549, -0.00000241, 0.00000114, 0.00000422, 0.00000569, 0.00000451, 0.00000011, -0.00000712, -0.00001564, -0.00002290, -0.00002597, -0.00002250, -0.00001171, 0.00000500, 0.00002409, 0.00004074, 0.00005019, 0.00004921, 0.00003722, 0.00001655, -0.00000816, -0.00003133, -0.00004801, -0.00005521, -0.00005253, -0.00004192, -0.00002680, -0.00001084, 0.00000304, 0.00001327, 0.00001956, 0.00002243, 0.00002266, 0.00002084, 0.00001731, 0.00001236, 0.00000658, 0.00000101, -0.00000296, },
		{ -0.00000454, -0.00000633, -0.00000906, -0.00001222, -0.00001463, -0.00001487, -0.00001184, -0.00000540, 0.00000344, 0.00001265, 0.00001979, 0.00002284, 0.00002091, 0.00001453, 0.00000554, -0.00000354, -0.00001030, -0.00001320, -0.00001189, -0.00000722, -0.00000076, 0.00000565, 0.00001048, 0.00001275, 0.00001209, 0.00000858, 0.00000270, -0.00000474, -0.00001257, -0.00001922, -0.00002291, -0.00002194, -0.00001536, -0.00000345, 0.00001190, 0.00002741, 0.00003908, 0.00004331, 0.00003805, 0.00002362, 0.00000282, -0.00001969, -0.00003870, -0.00004991, -0.00005113, -0.00004280, -0.00002767, -0.00000983, 0.00000662, 0.00001870, 0.00002519, 0.00002659, 0.00002451, 0.00002086, 0.00001713, 0.00001398, 0.00001136, 0.00000887, 0.00000620, 0.00000337, 0.00000073, -0.00000133, -0.00000265, -0.00000352, },
		{ -0.00000066, 0.00000410, 0.00000856, 0.00001248, 0.00001528, 0.00001616, 0.00001440, 0.00000981, 0.00000301, -0.00000452, -0.00001080, -0.00001395, -0.00001285, -0.00000759, 0.00000053, 0.00000937, 0.00001667, 0.00002067, 0.00002061, 0.00001681, 0.00001032, 0.00000251, -0.00000537, -0.00001240, -0.00001789, -0.00002122, -0.00002165, -0.00001843, -0.00001111, -0.00000004, 0.00001326, 0.00002610, 0.00003514, 0.00003731, 0.00003095, 0.00001655, -0.00000304, -0.00002321, -0.00003888, -0.00004598, -0.00004274, -0.00003018, -0.00001175, 0.00000778, 0.00002384, 0.00003336, 0.00003554, 0.00003168, 0.00002446, 0.00001672, 0.00001052, 0.00000657, 0.00000428, 0.00000239, -0.00000031, -0.00000435, -0.00000934, -0.00001424, -0.00001782, -0.00001918, -0.00001809, -0.00001495, -0.00001053, -0.00000560, },
		{ 0.00000265, -0.00000347, -0.00000865, -0.00001227, -0.00001381, -0.00001280, -0.00000896, -0.00000241, 0.00000611, 0.00001518, 0.00002297, 0.00002772, 0.00002824, 0.00002438, 0.00001704, 0.00000788, -0.00000128, -0.00000902, -0.00001467, -0.00001829, -0.00002025, -0.00002077, -0.00001955, -0.00001583, -0.00000887, 0.00000143, 0.00001388, 0.00002597, 0.00003433, 0.00003585, 0.00002889, 0.00001418, -0.00000498, -0.00002353, -0.00003612, -0.00003883, -0.00003053, -0.00001334, 0.00000802, 0.00002773, 0.00004074, 0.00004427, 0.00003854, 0.00002638, 0.00001203, -0.00000048, -0.00000861, -0.00001194, -0.00001191, -0.00001085, -0.00001087, -0.00001287, -0.00001632, -0.00001954, -0.00002061, -0.00001819, -0.00001218, -0.00000372, 0.00000520, 0.00001248, 0.00001660, 0.00001700, 0.00001407, 0.00000887, },
		{ -0.00000816, 0.00000342, 0.00001505, 0.00002519, 0.00003253, 0.00003612, 0.00003545, 0.00003060, 0.00002232, 0.00001203, 0.00000154, -0.00000745, -0.00001381, -0.00001733, -0.00001862, -0.00001878, -0.00001870, -0.00001853, -0.00001747, -0.00001412, -0.00000720, 0.00000351, 0.00001660, 0.00002907, 0.00003718, 0.00003771, 0.00002941, 0.00001382, -0.00000490, -0.00002105, -0.00002932, -0.00002661, -0.00001328, 0.00000679, 0.00002742, 0.00004217, 0.00004647, 0.00003913, 0.00002260, 0.00000203, -0.00001667, -0.00002881, -0.00003239, -0.00002845, -0.00002030, -0.00001199, -0.00000667, -0.00000547, -0.00000735, -0.00000980, -0.00001014, -0.00000680, 0.00000000, 0.00000834, 0.00001534, 0.00001835, 0.00001595, 0.00000848, -0.00000214, -0.00001308, -0.00002152, -0.00002545, -0.00002410, -0.00001790, },
		{ 0.00004534, 0.00003833, 0.00002781, 0.00001565, 0.00000372, -0.00000639, -0.00001358, -0.00001746, -0.00001845, -0.00001764, -0.00001645, -0.00001601, -0.00001663, -0.00001751, -0.00001694, -0.00001299, -0.00000444, 0.00000834, 0.00002309, 0.00003607, 0.00004327, 0.00004187, 0.00003160, 0.00001511, -0.00000255, -0.00001559, -0.00001953, -0.00001288, 0.00000207, 0.00001993, 0.00003397, 0.00003854, 0.00003113, 0.00001329, -0.00000983, -0.00003131, -0.00004488, -0.00004707, -0.00003824, -0.00002229, -0.00000498, 0.00000819, 0.00001394, 0.00001216, 0.00000554, -0.00000179, -0.00000609, -0.00000553, -0.00000084, 0.00000522, 0.00000907, 0.00000796, 0.00000108, -0.00001001, -0.00002196, -0.00003081, -0.00003340, -0.00002831, -0.00001629, 0.00000018, 0.00001765, 0.00003278, 0.00004306, 0.00004725, },
		{ -0.00001303, -0.00001765, -0.00001880, -0.00001764, -0.00001562, -0.00001412, -0.00001407, -0.00001554, -0.00001763, -0.00001855, -0.00001622, -0.00000904, 0.00000322, 0.00001898, 0.00003493, 0.00004691, 0.00005125, 0.00004622, 0.00003287, 0.00001498, -0.00000201, -0.00001289, -0.00001458, -0.00000736, 0.00000507, 0.00001687, 0.00002219, 0.00001747, 0.00000292, -0.00001739, -0.00003690, -0.00004888, -0.00004896, -0.00003673, -0.00001595, 0.00000681, 0.00002452, 0.00003217, 0.00002846, 0.00001598, 0.00000006, -0.00001344, -0.00002031, -0.00001944, -0.00001296, -0.00000511, -0.00000034, -0.00000142, -0.00000838, -0.00001845, -0.00002724, -0.00003045, -0.00002543, -0.00001221, 0.00000663, 0.00002678, 0.00004368, 0.00005380, 0.00005554, 0.00004938, 0.00003743, 0.00002264, 0.00000793, -0.00000440, },
		},
		{ { -0.00000338, -0.00000484, -0.00000683, -0.00000448, 0.00000483, 0.00001796, 0.00002699, 0.00002430, 0.00000823, -0.00001480, -0.00003369, -0.00003934, -0.00003004, -0.00001194, 0.00000560, 0.00001636, 0.00002057, 0.00002304, 0.00002785, 0.00003389, 0.00003510, 0.00002546, 0.00000490, -0.00001851, -0.00003271, -0.00002945, -0.00001068, 0.00001163, 0.00002254, 0.00001381, -0.00001033, -0.00003544, -0.00004602, -0.00003544, -0.00001033, 0.00001381, 0.00002254, 0.00001163, -0.00001068, -0.00002945, -0.00003271, -0.00001851, 0.00000490, 0.00002546, 0.00003510, 0.00003389, 0.00002785, 0.00002304, 0.00002057, 0.00001636, 0.00000560, -0.00001194, -0.00003004, -0.00003934, -0.00003369, -0.00001480, 0.00000823, 0.00002430, 0.00002699, 0.00001796, 0.00000483, -0.00000448, -0.00000683, -0.00000484, },
		{ -0.00000293, 0.00000526, 0.00001545, 0.00002018, 0.00001404, -0.00000162, -0.00001865, -0.00002694, -0.00002110, -0.00000417, 0.00001438, 0.00002490, 0.00002381, 0.00001535, 0.00000772, 0.00000670, 0.00001143, 0.00001554, 0.00001264, 0.00000193, -0.00001061, -0.00001649, -0.00001149, 0.00000036, 0.00000893, 0.00000516, -0.00001171, -0.00003237, -0.00004294, -0.00003442, -0.00000963, 0.00001748, 0.00003012, 0.00001949, -0.00000908, -0.00003886, -0.00005159, -0.00003837, -0.00000462, 0.00003374, 0.00005992, 0.00006535, 0.00005274, 0.00003192, 0.00001237, -0.00000220, -0.00001382, -0.00002563, -0.00003719, -0.00004347, -0.00003855, -0.00002098, 0.00000359, 0.00002474, 0.00003342, 0.00002742, 0.00001240, -0.00000215, -0.00000917, -0.00000802, -0.00000351, -0.00000124, -0.00000287, -0.00000508, },
		{ 0.00002005, 0.00001571, 0.00000351, -0.00000818, -0.00001097, -0.00000250, 0.00001168, 0.00002187, 0.00002107, 0.00000994, -0.00000372, -0.00001050, -0.00000616, 0.00000555, 0.00001578, 0.00001711, 0.00000880, -0.00000311, -0.00001076, -0.00001076, -0.00000673, -0.00000613, -0.00001365, -0.00002637, -0.00003481, -0.00002968, -0.00000976, 0.00001526, 0.00002963, 0.00002169, -0.00000738, -0.00004242, -0.00006223, -0.00005205, -0.00001273, 0.00003919, 0.00008051, 0.00009400, 0.00007682, 0.00004018, 0.00000123, -0.00002683, -0.00004025, -0.00004333, -0.00004220, -0.00003937, -0.00003269, -0.00001905, 0.00000086, 0.00002095, 0.00003287, 0.00003155, 0.00001894, 0.00000298, -0.00000759, -0.00000890, -0.00000389, 0.00000071, -0.00000006, -0.00000548, -0.00000994, -0.00000770, 0.00000207, 0.00001398, },
		{ -0.00000046, -0.00000543, -0.00000031, 0.00000863, 0.00001256, 0.00000690, -0.00000511, -0.00001477, -0.00001431, -0.00000282, 0.00001281, 0.00002257, 0.00002022, 0.00000737, -0.00000811, -0.00001804, -0.00001968, -0.00001688, -0.00001598, -0.00001978, -0.00002461, -0.00002308, -0.00001087, 0.00000792, 0.00002154, 0.00001793, -0.00000588, -0.00003916, -0.00006180, -0.00005577, -0.00001685, 0.00004107, 0.00009233, 0.00011316, 0.00009441, 0.00004581, -0.00001039, -0.00005209, -0.00006822, -0.00006191, -0.00004502, -0.00002892, -0.00001805, -0.00000977, 0.00000047, 0.00001359, 0.00002533, 0.00002963, 0.00002378, 0.00001115, -0.00000088, -0.00000626, -0.00000448, -0.00000061, -0.00000086, -0.00000723, -0.00001540, -0.00001775, -0.00000930, 0.00000787, 0.00002522, 0.00003322, 0.00002783, 0.00001338, },
		{ -0.00000029, 0.00000535, 0.00000190, -0.00000668, -0.00001199, -0.00000798, 0.00000416, 0.00001659, 0.00002045, 0.00001223, -0.00000382, -0.00001886, -0.00002582, -0.00002396, -0.00001844, -0.00001529, -0.00001622, -0.00001732, -0.00001308, -0.00000238, 0.00000866, 0.00001000, -0.00000436, -0.00002965, -0.00005036, -0.00004857, -0.00001581, 0.00003916, 0.00009269, 0.00011823, 0.00010109, 0.00004715, -0.00002013, -0.00007250, -0.00009174, -0.00007777, -0.00004578, -0.00001528, 0.00000148, 0.00000451, 0.00000249, 0.00000440, 0.00001266, 0.00002244, 0.00002655, 0.00002163, 0.00001064, 0.00000011, -0.00000524, -0.00000596, -0.00000671, -0.00001138, -0.00001873, -0.00002233, -0.00001522, 0.00000403, 0.00002860, 0.00004664, 0.00004855, 0.00003333, 0.00000952, -0.00001017, -0.00001690, -0.00001090, },
		{ -0.00000076, -0.00000873, -0.00000772, 0.00000081, 0.00000923, 0.00000990, 0.00000092, -0.00001267, -0.00002299, -0.00002517, -0.00002061, -0.00001498, -0.00001316, -0.00001513, -0.00001640, -0.00001268, -0.00000469, 0.00000118, -0.00000223, -0.00001585, -0.00003086, -0.00003256, -0.00001018, 0.00003378, 0.00008146, 0.00010785, 0.00009495, 0.00004338, -0.00002626, -0.00008382, -0.00010582, -0.00008770, -0.00004481, -0.00000200, 0.00002118, 0.00002050, 0.00000666, -0.00000420, -0.00000190, 0.00001203, 0.00002720, 0.00003281, 0.00002518, 0.00000927, -0.00000612, -0.00001501, -0.00001762, -0.00001821, -0.00001974, -0.00002011, -0.00001355, 0.00000399, 0.00002948, 0.00005242, 0.00006044, 0.00004736, 0.00001811, -0.00001334, -0.00003207, -0.00003110, -0.00001505, 0.00000361, 0.00001313, 0.00000979, },
		{ -0.00001055, 0.00000039, 0.00000418, -0.00000142, -0.00001177, -0.00001985, -0.00002190, -0.00001954, -0.00001737, -0.00001825, -0.00002062, -0.00002008, -0.00001400, -0.00000479, 0.00000126, -0.00000062, -0.00000832, -0.00001240, -0.00000209, 0.00002569, 0.00006104, 0.00008428, 0.00007705, 0.00003476, -0.00002809, -0.00008410, -0.00010770, -0.00008972, -0.00004246, 0.00000785, 0.00003663, 0.00003485, 0.00001249, -0.00000973, -0.00001454, 0.00000120, 0.00002596, 0.00004255, 0.00003984, 0.00001939, -0.00000706, -0.00002631, -0.00003205, -0.00002705, -0.00001858, -0.00001137, -0.00000406, 0.00000823, 0.00002754, 0.00004855, 0.00006006, 0.00005227, 0.00002458, -0.00001193, -0.00003993, -0.00004600, -0.00002887, -0.00000018, 0.00002291, 0.00002828, 0.00001566, -0.00000447, -0.00001882, -0.00002002, },
		{ 0.00000445, -0.00000919, -0.00002160, -0.00002778, -0.00002833, -0.00002679, -0.00002529, -0.00002256, -0.00001589, -0.00000490, 0.00000660, 0.00001290, 0.00001143, 0.00000606, 0.00000536, 0.00001598, 0.00003586, 0.00005287, 0.00005130, 0.00002271, -0.00002605, -0.00007413, -0.00009754, -0.00008353, -0.00003909, 0.00001243, 0.00004490, 0.00004493, 0.00001933, -0.00001025, -0.00002185, -0.00000727, 0.00002325, 0.00004867, 0.00005130, 0.00002816, -0.00000798, -0.00003767, -0.00004728, -0.00003642, -0.00001578, 0.00000206, 0.00001145, 0.00001597, 0.00002290, 0.00003505, 0.00004661, 0.00004677, 0.00002852, -0.00000411, -0.00003629, -0.00005076, -0.00003904, -0.00000756, 0.00002561, 0.00004162, 0.00003228, 0.00000455, -0.00002459, -0.00003912, -0.00003350, -0.00001472, 0.00000366, 0.00001076, },
		{ -0.00004299, -0.00003892, -0.00002976, -0.00001962, -0.00000927, 0.00000210, 0.00001371, 0.00002205, 0.00002331, 0.00001731, 0.00000909, 0.00000592, 0.00001132, 0.00002081, 0.00002348, 0.00000946, -0.00002155, -0.00005723, -0.00007839, -0.00007070, -0.00003499, 0.00001162, 0.00004517, 0.00004951, 0.00002654, -0.00000496, -0.00002172, -0.00001129, 0.00001958, 0.00004950, 0.00005665, 0.00003335, -0.00000883, -0.00004663, -0.00006001, -0.00004439, -0.00001209, 0.00001712, 0.00002952, 0.00002531, 0.00001605, 0.00001403, 0.00002216, 0.00003175, 0.00002980, 0.00001022, -0.00001972, -0.00004288, -0.00004362, -0.00001917, 0.00001716, 0.00004370, 0.00004398, 0.00001724, -0.00002113, -0.00004942, -0.00005299, -0.00003261, -0.00000271, 0.00001852, 0.00002021, 0.00000403, -0.00001903, -0.00003679, },
		{ 0.00000425, 0.00001513, 0.00002179, 0.00002522, 0.00002431, 0.00001773, 0.00000691, -0.00000323, -0.00000741, -0.00000459, -0.00000007, -0.00000241, -0.00001656, -0.00003840, -0.00005543, -0.00005425, -0.00003038, 0.00000692, 0.00003892, 0.00004892, 0.00003353, 0.00000559, -0.00001391, -0.00001003, 0.00001542, 0.00004433, 0.00005414, 0.00003341, -0.00000955, -0.00005109, -0.00006722, -0.00004903, -0.00000834, 0.00003032, 0.00004580, 0.00003369, 0.00000778, -0.00001075, -0.00000867, 0.00001022, 0.00002869, 0.00002933, 0.00000789, -0.00002286, -0.00004180, -0.00003445, -0.00000373, 0.00003108, 0.00004679, 0.00003171, -0.00000646, -0.00004549, -0.00006283, -0.00004973, -0.00001605, 0.00001677, 0.00002944, 0.00001612, -0.00001368, -0.00004257, -0.00005634, -0.00005118, -0.00003323, -0.00001241, },
		{ 0.00001434, 0.00000667, -0.00000169, -0.00001038, -0.00001664, -0.00001767, -0.00001403, -0.00001057, -0.00001308, -0.00002305, -0.00003475, -0.00003801, -0.00002542, 0.00000103, 0.00002946, 0.00004497, 0.00003982, 0.00001952, -0.00000020, -0.00000412, 0.00001115, 0.00003366, 0.00004370, 0.00002788, -0.00001010, -0.00004987, -0.00006691, -0.00004898, -0.00000526, 0.00003847, 0.00005615, 0.00003847, -0.00000097, -0.00003471, -0.00003985, -0.00001344, 0.00002569, 0.00004992, 0.00004175, 0.00000584, -0.00003413, -0.00005157, -0.00003505, 0.00000395, 0.00003879, 0.00004552, 0.00001874, -0.00002511, -0.00005889, -0.00006261, -0.00003598, 0.00000224, 0.00002711, 0.00002344, -0.00000541, -0.00004110, -0.00006272, -0.00005936, -0.00003496, -0.00000353, 0.00002068, 0.00003097, 0.00002941, 0.00002226, },
		{ -0.00001497, -0.00001578, -0.00001536, -0.00001348, -0.00001265, -0.00001568, -0.00002184, -0.00002568, -0.00002025, -0.00000289, 0.00002105, 0.00004034, 0.00004507, 0.00003396, 0.00001601, 0.00000447, 0.00000698, 0.00001916, 0.00002708, 0.00001760, -0.00001045, -0.00004299, -0.00005869, -0.00004377, -0.00000334, 0.00003954, 0.00005761, 0.00003765, -0.00000930, -0.00005352, -0.00006523, -0.00003442, 0.00002150, 0.00006796, 0.00007515, 0.00003751, -0.00002232, -0.00006779, -0.00007178, -0.00003372, 0.00002057, 0.00005601, 0.00005083, 0.00000975, -0.00004030, -0.00006859, -0.00006039, -0.00002525, 0.00001082, 0.00002341, 0.00000542, -0.00002954, -0.00005736, -0.00005957, -0.00003439, 0.00000374, 0.00003464, 0.00004501, 0.00003483, 0.00001451, -0.00000369, -0.00001330, -0.00001538, -0.00001478, },
		{ -0.00001588, -0.00001870, -0.00002024, -0.00002004, -0.00001504, -0.00000213, 0.00001773, 0.00003786, 0.00004904, 0.00004568, 0.00003042, 0.00001290, 0.00000300, 0.00000345, 0.00000754, 0.00000461, -0.00001060, -0.00003182, -0.00004399, -0.00003412, -0.00000272, 0.00003316, 0.00004923, 0.00003043, -0.00001649, -0.00006406, -0.00008000, -0.00004859, 0.00001684, 0.00007956, 0.00010108, 0.00006556, -0.00000887, -0.00008001, -0.00010692, -0.00007501, -0.00000488, 0.00006082, 0.00008381, 0.00005310, -0.00000985, -0.00006639, -0.00008555, -0.00006233, -0.00001818, 0.00001536, 0.00001798, -0.00000718, -0.00003792, -0.00004943, -0.00003134, 0.00000632, 0.00004088, 0.00005254, 0.00003654, 0.00000452, -0.00002459, -0.00003698, -0.00003115, -0.00001616, -0.00000361, 0.00000007, -0.00000404, -0.00001075, },
		{ -0.00001006, 0.00000482, 0.00002222, 0.00003970, 0.00005154, 0.00005176, 0.00003884, 0.00001814, -0.00000078, -0.00001048, -0.00001076, -0.00000829, -0.00001057, -0.00001880, -0.00002586, -0.00002180, -0.00000323, 0.00002077, 0.00003242, 0.00001747, -0.00002215, -0.00006518, -0.00008189, -0.00005352, 0.00001233, 0.00008201, 0.00011388, 0.00008404, 0.00000348, -0.00008553, -0.00013314, -0.00011165, -0.00003287, 0.00005829, 0.00011044, 0.00009597, 0.00002639, -0.00005643, -0.00010681, -0.00010238, -0.00005486, 0.00000087, 0.00003106, 0.00002402, -0.00000535, -0.00002888, -0.00002605, 0.00000229, 0.00003587, 0.00005019, 0.00003376, -0.00000455, -0.00004213, -0.00005784, -0.00004517, -0.00001480, 0.00001388, 0.00002546, 0.00001690, -0.00000323, -0.00002232, -0.00003187, -0.00003068, -0.00002224, },
		{ 0.00005244, 0.00005028, 0.00003818, 0.00001775, -0.00000439, -0.00001982, -0.00002369, -0.00001812, -0.00001037, -0.00000694, -0.00000836, -0.00000934, -0.00000439, 0.00000536, 0.00001074, 0.00000093, -0.00002621, -0.00005801, -0.00007171, -0.00004904, 0.00000839, 0.00007452, 0.00011074, 0.00008910, 0.00001240, -0.00008280, -0.00014460, -0.00013597, -0.00005813, 0.00004788, 0.00012393, 0.00012842, 0.00006035, -0.00004086, -0.00011963, -0.00013692, -0.00009137, -0.00001638, 0.00004334, 0.00005973, 0.00003596, -0.00000039, -0.00001911, -0.00000838, 0.00001899, 0.00003708, 0.00002685, -0.00000967, -0.00005068, -0.00006975, -0.00005377, -0.00001151, 0.00003246, 0.00005361, 0.00004213, 0.00000722, -0.00003052, -0.00005229, -0.00005089, -0.00003159, -0.00000561, 0.00001783, 0.00003519, 0.00004672, },
		{ -0.00000779, -0.00002256, -0.00002800, -0.00002235, -0.00001006, 0.00000081, 0.00000429, 0.00000047, -0.00000557, -0.00000913, -0.00001078, -0.00001596, -0.00002891, -0.00004570, -0.00005329, -0.00003735, 0.00000521, 0.00005851, 0.00009246, 0.00007996, 0.00001640, -0.00007189, -0.00013848, -0.00014280, -0.00007594, 0.00003043, 0.00011961, 0.00014184, 0.00008372, -0.00002322, -0.00012070, -0.00015756, -0.00011868, -0.00003133, 0.00005382, 0.00009420, 0.00007911, 0.00003151, -0.00001127, -0.00002361, -0.00000691, 0.00001534, 0.00001740, -0.00000862, -0.00004673, -0.00006867, -0.00005454, -0.00000758, 0.00004679, 0.00007703, 0.00006530, 0.00001815, -0.00003839, -0.00007511, -0.00007617, -0.00004572, -0.00000208, 0.00003504, 0.00005494, 0.00005824, 0.00005161, 0.00004074, 0.00002719, 0.00001049, },
		{ -0.00000970, 0.00000220, 0.00000864, 0.00000518, -0.00000614, -0.00001871, -0.00002674, -0.00002955, -0.00003069, -0.00003275, -0.00003264, -0.00002255, 0.00000272, 0.00003746, 0.00006349, 0.00005923, 0.00001518, -0.00005475, -0.00011591, -0.00013088, -0.00008330, 0.00000814, 0.00009630, 0.00013117, 0.00008986, -0.00000785, -0.00010916, -0.00015831, -0.00012890, -0.00003870, 0.00006214, 0.00012177, 0.00011644, 0.00006137, -0.00000312, -0.00003950, -0.00003598, -0.00001009, 0.00000802, -0.00000064, -0.00002990, -0.00005389, -0.00004709, -0.00000455, 0.00005247, 0.00008931, 0.00008058, 0.00002713, -0.00004373, -0.00009469, -0.00010045, -0.00006116, 0.00000018, 0.00005325, 0.00007780, 0.00007209, 0.00004893, 0.00002418, 0.00000682, -0.00000336, -0.00001085, -0.00001789, -0.00002208, -0.00001949, },
		{ -0.00000566, -0.00002021, -0.00003253, -0.00003650, -0.00003205, -0.00002396, -0.00001661, -0.00000972, 0.00000066, 0.00001625, 0.00003106, 0.00003241, 0.00000962, -0.00003498, -0.00008202, -0.00010334, -0.00007952, -0.00001559, 0.00005706, 0.00009645, 0.00007544, 0.00000083, -0.00008732, -0.00013751, -0.00011747, -0.00003427, 0.00006901, 0.00013864, 0.00014177, 0.00008448, 0.00000520, -0.00005109, -0.00006064, -0.00003265, 0.00000160, 0.00001331, -0.00000327, -0.00002837, -0.00003337, -0.00000396, 0.00004705, 0.00008672, 0.00008410, 0.00003232, -0.00004476, -0.00010602, -0.00011785, -0.00007420, 0.00000088, 0.00006870, 0.00009894, 0.00008515, 0.00004436, 0.00000301, -0.00001997, -0.00002198, -0.00001334, -0.00000641, -0.00000631, -0.00000944, -0.00000933, -0.00000379, 0.00000275, 0.00000322, },
		{ -0.00003340, -0.00002332, -0.00001135, -0.00000377, -0.00000130, 0.00000013, 0.00000366, 0.00000665, 0.00000150, -0.00001719, -0.00004509, -0.00006719, -0.00006637, -0.00003664, 0.00000898, 0.00004323, 0.00004132, -0.00000101, -0.00006078, -0.00009909, -0.00008483, -0.00001596, 0.00007630, 0.00014438, 0.00015254, 0.00009850, 0.00001441, -0.00005355, -0.00007351, -0.00004588, 0.00000053, 0.00003057, 0.00002723, 0.00000173, -0.00001729, -0.00000705, 0.00003066, 0.00006904, 0.00007482, 0.00003286, -0.00004063, -0.00010591, -0.00012405, -0.00008173, -0.00000009, 0.00007795, 0.00011356, 0.00009410, 0.00003797, -0.00001951, -0.00004883, -0.00004242, -0.00001500, 0.00000999, 0.00001695, 0.00000645, -0.00000902, -0.00001672, -0.00001332, -0.00000575, -0.00000393, -0.00001205, -0.00002537, -0.00003455, },
		{ -0.00000345, -0.00000643, -0.00001084, -0.00001080, -0.00000696, -0.00000618, -0.00001485, -0.00003189, -0.00004764, -0.00005076, -0.00003791, -0.00001817, -0.00000746, -0.00001598, -0.00003795, -0.00005251, -0.00003704, 0.00001537, 0.00008702, 0.00014292, 0.00015137, 0.00010474, 0.00002614, -0.00004337, -0.00006951, -0.00004542, 0.00000567, 0.00004718, 0.00005406, 0.00002854, -0.00000380, -0.00001444, 0.00000592, 0.00003967, 0.00005479, 0.00002903, -0.00003164, -0.00009383, -0.00011734, -0.00008202, -0.00000262, 0.00007873, 0.00011804, 0.00009624, 0.00003002, -0.00003995, -0.00007494, -0.00006120, -0.00001556, 0.00002854, 0.00004388, 0.00002565, -0.00000882, -0.00003475, -0.00003747, -0.00002071, -0.00000113, 0.00000538, -0.00000502, -0.00002306, -0.00003494, -0.00003308, -0.00002083, -0.00000828, },
		{ -0.00001368, -0.00000590, -0.00000042, -0.00000728, -0.00002818, -0.00005431, -0.00007250, -0.00007462, -0.00006264, -0.00004515, -0.00002875, -0.00001155, 0.00001484, 0.00005591, 0.00010497, 0.00014244, 0.00014652, 0.00010872, 0.00004299, -0.00001940, -0.00004766, -0.00003067, 0.00001564, 0.00005850, 0.00006981, 0.00004440, 0.00000243, -0.00002588, -0.00002259, 0.00000474, 0.00002857, 0.00002212, -0.00001925, -0.00007205, -0.00009903, -0.00007505, -0.00000637, 0.00007051, 0.00011084, 0.00009024, 0.00002096, -0.00005539, -0.00009407, -0.00007512, -0.00001486, 0.00004613, 0.00006993, 0.00004494, -0.00000879, -0.00005502, -0.00006600, -0.00003949, 0.00000229, 0.00002976, 0.00002610, -0.00000316, -0.00003660, -0.00005310, -0.00004587, -0.00002459, -0.00000597, -0.00000074, -0.00000700, -0.00001427, },
		{ -0.00001278, -0.00004483, -0.00008477, -0.00011280, -0.00011440, -0.00008792, -0.00004287, 0.00000812, 0.00005660, 0.00009944, 0.00013414, 0.00015433, 0.00015100, 0.00011972, 0.00006820, 0.00001661, -0.00001199, -0.00000586, 0.00002645, 0.00006002, 0.00006902, 0.00004395, -0.00000162, -0.00004018, -0.00004946, -0.00002840, 0.00000200, 0.00001399, -0.00000583, -0.00004511, -0.00007310, -0.00006250, -0.00001079, 0.00005469, 0.00009292, 0.00007645, 0.00001146, -0.00006402, -0.00010346, -0.00008195, -0.00001294, 0.00005992, 0.00009079, 0.00006111, -0.00000894, -0.00007423, -0.00009432, -0.00005920, 0.00000564, 0.00005722, 0.00006335, 0.00002239, -0.00003804, -0.00008090, -0.00008383, -0.00005145, -0.00000885, 0.00001746, 0.00001689, -0.00000105, -0.00001719, -0.00001864, -0.00000829, -0.00000138, },
		{ -0.00015355, -0.00014218, -0.00008782, -0.00000795, 0.00007369, 0.00013853, 0.00017798, 0.00019117, 0.00018036, 0.00014931, 0.00010516, 0.00006009, 0.00002861, 0.00002013, 0.00003180, 0.00004809, 0.00004954, 0.00002575, -0.00001621, -0.00005516, -0.00006962, -0.00005306, -0.00001933, 0.00000647, 0.00000590, -0.00001875, -0.00004520, -0.00004726, -0.00001523, 0.00003428, 0.00006754, 0.00005694, 0.00000231, -0.00006539, -0.00010233, -0.00008089, -0.00001004, 0.00006788, 0.00010323, 0.00007155, -0.00000930, -0.00008929, -0.00011791, -0.00007677, 0.00000820, 0.00008320, 0.00010079, 0.00004974, -0.00003878, -0.00011201, -0.00012907, -0.00008519, -0.00001187, 0.00004583, 0.00005831, 0.00002791, -0.00001703, -0.00004442, -0.00004074, -0.00001850, -0.00000502, -0.00002162, -0.00006802, -0.00012206, },
		{ 0.00005426, 0.00016603, 0.00023862, 0.00026405, 0.00024955, 0.00020907, 0.00015671, 0.00010440, 0.00006180, 0.00003500, 0.00002385, 0.00002060, 0.00001324, -0.00000698, -0.00003834, -0.00006796, -0.00007951, -0.00006491, -0.00003160, 0.00000082, 0.00001344, 0.00000158, -0.00002120, -0.00003267, -0.00001899, 0.00001320, 0.00003951, 0.00003503, -0.00000562, -0.00006043, -0.00009198, -0.00007272, -0.00000660, 0.00006915, 0.00010568, 0.00007479, -0.00000984, -0.00009797, -0.00013318, -0.00008958, 0.00000934, 0.00010333, 0.00013234, 0.00007470, -0.00003827, -0.00014100, -0.00017406, -0.00012041, -0.00001467, 0.00007983, 0.00011087, 0.00006814, -0.00001381, -0.00008085, -0.00009516, -0.00005813, -0.00000648, 0.00001407, -0.00002045, -0.00009688, -0.00017340, -0.00020468, -0.00016717, -0.00006975, },
		{ 0.00036947, 0.00030792, 0.00022442, 0.00014210, 0.00007415, 0.00002532, -0.00000532, -0.00002272, -0.00003424, -0.00004663, -0.00006195, -0.00007538, -0.00007789, -0.00006313, -0.00003389, -0.00000273, 0.00001507, 0.00001200, -0.00000567, -0.00002162, -0.00002150, -0.00000465, 0.00001399, 0.00001447, -0.00001158, -0.00005110, -0.00007537, -0.00005956, -0.00000321, 0.00006409, 0.00009847, 0.00007073, -0.00001056, -0.00009929, -0.00013816, -0.00009593, 0.00000869, 0.00011434, 0.00015303, 0.00009352, -0.00003608, -0.00016253, -0.00021095, -0.00015108, -0.00001684, 0.00011367, 0.00016559, 0.00011246, -0.00000903, -0.00012347, -0.00016438, -0.00011489, -0.00001639, 0.00005904, 0.00005416, -0.00003828, -0.00017133, -0.00027143, -0.00027930, -0.00018025, -0.00000794, 0.00017663, 0.00031614, 0.00037994, },
		{ 0.00005406, -0.00002055, -0.00006207, -0.00008022, -0.00008460, -0.00008271, -0.00007906, -0.00007438, -0.00006604, -0.00005075, -0.00002858, -0.00000519, 0.00001028, 0.00001127, -0.00000067, -0.00001575, -0.00002243, -0.00001636, -0.00000476, -0.00000138, -0.00001503, -0.00003984, -0.00005629, -0.00004433, -0.00000050, 0.00005413, 0.00008360, 0.00006056, -0.00001141, -0.00009364, -0.00013281, -0.00009534, 0.00000624, 0.00011467, 0.00015991, 0.00010365, -0.00003204, -0.00017252, -0.00023318, -0.00017185, -0.00001803, 0.00014136, 0.00021269, 0.00015253, -0.00000471, -0.00016649, -0.00023793, -0.00017956, -0.00003187, 0.00010857, 0.00014677, 0.00004609, -0.00014929, -0.00033815, -0.00041706, -0.00033397, -0.00011175, 0.00016807, 0.00040817, 0.00053943, 0.00054383, 0.00044968, 0.00030798, 0.00016644, },
		{ -0.00012865, -0.00010399, -0.00008133, -0.00006266, -0.00004745, -0.00003399, -0.00002086, -0.00000861, -0.00000015, 0.00000117, -0.00000517, -0.00001504, -0.00002169, -0.00002075, -0.00001442, -0.00001046, -0.00001578, -0.00002895, -0.00003846, -0.00003001, 0.00000099, 0.00004128, 0.00006415, 0.00004640, -0.00001233, -0.00008260, -0.00011888, -0.00008850, 0.00000231, 0.00010474, 0.00015260, 0.00010411, -0.00002631, -0.00016898, -0.00023681, -0.00017910, -0.00001801, 0.00015802, 0.00024363, 0.00018060, -0.00000280, -0.00020385, -0.00030406, -0.00024089, -0.00004870, 0.00015699, 0.00024458, 0.00014316, -0.00011338, -0.00040073, -0.00056833, -0.00051709, -0.00024806, 0.00014171, 0.00050987, 0.00073555, 0.00076720, 0.00063126, 0.00040478, 0.00017286, -0.00000545, -0.00010961, -0.00014957, -0.00014858, },
		{ -0.00002698, -0.00002080, -0.00001759, -0.00001561, -0.00001423, -0.00001387, -0.00001516, -0.00001768, -0.00001953, -0.00001868, -0.00001529, -0.00001249, -0.00001403, -0.00001999, -0.00002453, -0.00001891, 0.00000094, 0.00002771, 0.00004356, 0.00003080, -0.00001327, -0.00006848, -0.00009939, -0.00007711, -0.00000244, 0.00008677, 0.00013321, 0.00009564, -0.00001937, -0.00015243, -0.00022135, -0.00017169, -0.00001681, 0.00016075, 0.00025282, 0.00019117, -0.00000472, -0.00023022, -0.00035204, -0.00028797, -0.00006238, 0.00019852, 0.00033316, 0.00023690, -0.00007253, -0.00045539, -0.00071758, -0.00070925, -0.00040092, 0.00010248, 0.00061421, 0.00095392, 0.00102461, 0.00084211, 0.00050986, 0.00016005, -0.00010465, -0.00024199, -0.00026652, -0.00022326, -0.00015833, -0.00010114, -0.00006200, -0.00003914, },
		{ -0.00002928, -0.00002777, -0.00002461, -0.00002054, -0.00001643, -0.00001286, -0.00001022, -0.00000914, -0.00001036, -0.00001344, -0.00001553, -0.00001212, -0.00000066, 0.00001517, 0.00002473, 0.00001610, -0.00001415, -0.00005379, -0.00007788, -0.00006333, -0.00000720, 0.00006424, 0.00010577, 0.00008042, -0.00001203, -0.00012573, -0.00018976, -0.00015118, -0.00001468, 0.00014922, 0.00023874, 0.00018209, -0.00001087, -0.00024202, -0.00037424, -0.00031247, -0.00006924, 0.00022806, 0.00039910, 0.00031141, -0.00003646, -0.00049895, -0.00084806, -0.00088665, -0.00055007, 0.00005787, 0.00071277, 0.00117438, 0.00129298, 0.00106497, 0.00061624, 0.00013035, -0.00023623, -0.00040984, -0.00040664, -0.00029913, -0.00016848, -0.00006903, -0.00001770, -0.00000431, -0.00000967, -0.00001891, -0.00002554, -0.00002872, },
		{ -0.00001304, -0.00000730, -0.00000399, -0.00000362, -0.00000566, -0.00000869, -0.00001062, -0.00000925, -0.00000351, 0.00000463, 0.00000945, 0.00000391, -0.00001492, -0.00004060, -0.00005757, -0.00004931, -0.00001115, 0.00004106, 0.00007526, 0.00006148, -0.00000521, -0.00009345, -0.00014773, -0.00012141, -0.00001210, 0.00012563, 0.00020415, 0.00015503, -0.00002058, -0.00023797, -0.00036755, -0.00031031, -0.00006735, 0.00024195, 0.00043242, 0.00035400, -0.00001358, -0.00052884, -0.00094435, -0.00102596, -0.00067435, 0.00001659, 0.00079677, 0.00137371, 0.00154424, 0.00127804, 0.00071567, 0.00008947, -0.00038594, -0.00059783, -0.00055896, -0.00037104, -0.00015808, -0.00000830, 0.00005043, 0.00003964, 0.00000042, -0.00003387, -0.00004966, -0.00004963, -0.00004240, -0.00003425, -0.00002693, -0.00001992, },
		{ -0.00000101, -0.00000421, -0.00000741, -0.00000861, -0.00000705, -0.00000384, -0.00000196, -0.00000518, -0.00001551, -0.00003013, -0.00004064, -0.00003670, -0.00001365, 0.00002073, 0.00004649, 0.00004203, 0.00000018, -0.00006087, -0.00010246, -0.00008769, -0.00000963, 0.00009416, 0.00015540, 0.00011507, -0.00003212, -0.00021925, -0.00033392, -0.00028250, -0.00005706, 0.00023850, 0.00042830, 0.00035761, -0.00000906, -0.00054319, -0.00099472, -0.00110824, -0.00075549, -0.00001328, 0.00085820, 0.00152920, 0.00174958, 0.00145815, 0.00079955, 0.00004537, -0.00053468, -0.00078509, -0.00070905, -0.00043436, -0.00013050, 0.00007431, 0.00013636, 0.00008895, 0.00000110, -0.00006750, -0.00009152, -0.00007855, -0.00005119, -0.00002837, -0.00001663, -0.00001252, -0.00001004, -0.00000625, -0.00000211, 0.00000001, },
		{ -0.00001052, -0.00001098, -0.00001057, -0.00001167, -0.00001588, -0.00002251, -0.00002784, -0.00002632, -0.00001431, 0.00000559, 0.00002310, 0.00002475, 0.00000344, -0.00003284, -0.00006105, -0.00005559, -0.00000786, 0.00006008, 0.00010102, 0.00006948, -0.00004313, -0.00018912, -0.00027974, -0.00023473, -0.00004084, 0.00021817, 0.00038791, 0.00032219, -0.00002379, -0.00054083, -0.00099302, -0.00112224, -0.00078140, -0.00002586, 0.00089094, 0.00162225, 0.00188410, 0.00158438, 0.00086015, 0.00000676, -0.00066174, -0.00094845, -0.00084093, -0.00048536, -0.00009278, 0.00016690, 0.00022984, 0.00013871, -0.00000744, -0.00011717, -0.00014862, -0.00011388, -0.00005418, -0.00000786, 0.00000918, 0.00000363, -0.00000799, -0.00001317, -0.00000949, -0.00000214, 0.00000256, 0.00000188, -0.00000264, -0.00000762, },
		{ -0.00001601, -0.00001687, -0.00001842, -0.00001804, -0.00001315, -0.00000359, 0.00000683, 0.00001125, 0.00000421, -0.00001280, -0.00002909, -0.00002979, -0.00000725, 0.00002855, 0.00004984, 0.00002622, -0.00005115, -0.00015228, -0.00021449, -0.00017612, -0.00002276, 0.00018352, 0.00031808, 0.00025455, -0.00005412, -0.00052137, -0.00093954, -0.00106628, -0.00074839, -0.00001868, 0.00089164, 0.00164143, 0.00193092, 0.00164143, 0.00089164, -0.00001868, -0.00074839, -0.00106628, -0.00093954, -0.00052137, -0.00005412, 0.00025455, 0.00031808, 0.00018352, -0.00002276, -0.00017612, -0.00021449, -0.00015228, -0.00005115, 0.00002622, 0.00004984, 0.00002855, -0.00000725, -0.00002979, -0.00002909, -0.00001280, 0.00000421, 0.00001125, 0.00000683, -0.00000359, -0.00001315, -0.00001804, -0.00001842, -0.00001687, },
		{ -0.00001052, -0.00000762, -0.00000264, 0.00000188, 0.00000256, -0.00000214, -0.00000949, -0.00001317, -0.00000799, 0.00000363, 0.00000918, -0.00000786, -0.00005418, -0.00011388, -0.00014862, -0.00011717, -0.00000744, 0.00013871, 0.00022984, 0.00016690, -0.00009278, -0.00048536, -0.00084093, -0.00094845, -0.00066174, 0.00000676, 0.00086015, 0.00158438, 0.00188410, 0.00162225, 0.00089094, -0.00002586, -0.00078140, -0.00112224, -0.00099302, -0.00054083, -0.00002379, 0.00032219, 0.00038791, 0.00021817, -0.00004084, -0.00023473, -0.00027974, -0.00018912, -0.00004313, 0.00006948, 0.00010102, 0.00006008, -0.00000786, -0.00005559, -0.00006105, -0.00003284, 0.00000344, 0.00002475, 0.00002310, 0.00000559, -0.00001431, -0.00002632, -0.00002784, -0.00002251, -0.00001588, -0.00001167, -0.00001057, -0.00001098, },
		{ -0.00000101, 0.00000001, -0.00000211, -0.00000625, -0.00001004, -0.00001252, -0.00001663, -0.00002837, -0.00005119, -0.00007855, -0.00009152, -0.00006750, 0.00000110, 0.00008895, 0.00013636, 0.00007431, -0.00013050, -0.00043436, -0.00070905, -0.00078509, -0.00053468, 0.00004537, 0.00079955, 0.00145815, 0.00174958, 0.00152920, 0.00085820, -0.00001328, -0.00075549, -0.00110824, -0.00099472, -0.00054319, -0.00000906, 0.00035761, 0.00042830, 0.00023850, -0.00005706, -0.00028250, -0.00033392, -0.00021925, -0.00003212, 0.00011507, 0.00015540, 0.00009416, -0.00000963, -0.00008769, -0.00010246, -0.00006087, 0.00000018, 0.00004203, 0.00004649, 0.00002073, -0.00001365, -0.00003670, -0.00004064, -0.00003013, -0.00001551, -0.00000518, -0.00000196, -0.00000384, -0.00000705, -0.00000861, -0.00000741, -0.00000421, },
		{ -0.00001304, -0.00001992, -0.00002693, -0.00003425, -0.00004240, -0.00004963, -0.00004966, -0.00003387, 0.00000042, 0.00003964, 0.00005043, -0.00000830, -0.00015808, -0.00037104, -0.00055896, -0.00059783, -0.00038594, 0.00008947, 0.00071567, 0.00127804, 0.00154424, 0.00137371, 0.00079677, 0.00001659, -0.00067435, -0.00102596, -0.00094435, -0.00052884, -0.00001358, 0.00035400, 0.00043242, 0.00024195, -0.00006735, -0.00031031, -0.00036755, -0.00023797, -0.00002058, 0.00015503, 0.00020415, 0.00012563, -0.00001210, -0.00012141, -0.00014773, -0.00009345, -0.00000521, 0.00006148, 0.00007526, 0.00004106, -0.00001115, -0.00004931, -0.00005757, -0.00004060, -0.00001492, 0.00000391, 0.00000945, 0.00000463, -0.00000351, -0.00000925, -0.00001062, -0.00000869, -0.00000566, -0.00000362, -0.00000399, -0.00000730, },
		{ -0.00002928, -0.00002872, -0.00002554, -0.00001891, -0.00000967, -0.00000431, -0.00001770, -0.00006903, -0.00016848, -0.00029913, -0.00040664, -0.00040984, -0.00023623, 0.00013035, 0.00061624, 0.00106497, 0.00129298, 0.00117438, 0.00071277, 0.00005787, -0.00055007, -0.00088665, -0.00084806, -0.00049895, -0.00003646, 0.00031141, 0.00039910, 0.00022806, -0.00006924, -0.00031247, -0.00037424, -0.00024202, -0.00001087, 0.00018209, 0.00023874, 0.00014922, -0.00001468, -0.00015118, -0.00018976, -0.00012573, -0.00001203, 0.00008042, 0.00010577, 0.00006424, -0.00000720, -0.00006333, -0.00007788, -0.00005379, -0.00001415, 0.00001610, 0.00002473, 0.00001517, -0.00000066, -0.00001212, -0.00001553, -0.00001344, -0.00001036, -0.00000914, -0.00001022, -0.00001286, -0.00001643, -0.00002054, -0.00002461, -0.00002777, },
		{ -0.00002698, -0.00003914, -0.00006200, -0.00010114, -0.00015833, -0.00022326, -0.00026652, -0.00024199, -0.00010465, 0.00016005, 0.00050986, 0.00084211, 0.00102461, 0.00095392, 0.00061421, 0.00010248, -0.00040092, -0.00070925, -0.00071758, -0.00045539, -0.00007253, 0.00023690, 0.00033316, 0.00019852, -0.00006238, -0.00028797, -0.00035204, -0.00023022, -0.00000472, 0.00019117, 0.00025282, 0.00016075, -0.00001681, -0.00017169, -0.00022135, -0.00015243, -0.00001937, 0.00009564, 0.00013321, 0.00008677, -0.00000244, -0.00007711, -0.00009939, -0.00006848, -0.00001327, 0.00003080, 0.00004356, 0.00002771, 0.00000094, -0.00001891, -0.00002453, -0.00001999, -0.00001403, -0.00001249, -0.00001529, -0.00001868, -0.00001953, -0.00001768, -0.00001516, -0.00001387, -0.00001423, -0.00001561, -0.00001759, -0.00002080, },
		{ -0.00012865, -0.00014858, -0.00014957, -0.00010961, -0.00000545, 0.00017286, 0.00040478, 0.00063126, 0.00076720, 0.00073555, 0.00050987, 0.00014171, -0.00024806, -0.00051709, -0.00056833, -0.00040073, -0.00011338, 0.00014316, 0.00024458, 0.00015699, -0.00004870, -0.00024089, -0.00030406, -0.00020385, -0.00000280, 0.00018060, 0.00024363, 0.00015802, -0.00001801, -0.00017910, -0.00023681, -0.00016898, -0.00002631, 0.00010411, 0.00015260, 0.00010474, 0.00000231, -0.00008850, -0.00011888, -0.00008260, -0.00001233, 0.00004640, 0.00006415, 0.00004128, 0.00000099, -0.00003001, -0.00003846, -0.00002895, -0.00001578, -0.00001046, -0.00001442, -0.00002075, -0.00002169, -0.00001504, -0.00000517, 0.00000117, -0.00000015, -0.00000861, -0.00002086, -0.00003399, -0.00004745, -0.00006266, -0.00008133, -0.00010399, },
		{ 0.00005406, 0.00016644, 0.00030798, 0.00044968, 0.00054383, 0.00053943, 0.00040817, 0.00016807, -0.00011175, -0.00033397, -0.00041706, -0.00033815, -0.00014929, 0.00004609, 0.00014677, 0.00010857, -0.00003187, -0.00017956, -0.00023793, -0.00016649, -0.00000471, 0.00015253, 0.00021269, 0.00014136, -0.00001803, -0.00017185, -0.00023318, -0.00017252, -0.00003204, 0.00010365, 0.00015991, 0.00011467, 0.00000624, -0.00009534, -0.00013281, -0.00009364, -0.00001141, 0.00006056, 0.00008360, 0.00005413, -0.00000050, -0.00004433, -0.00005629, -0.00003984, -0.00001503, -0.00000138, -0.00000476, -0.00001636, -0.00002243, -0.00001575, -0.00000067, 0.00001127, 0.00001028, -0.00000519, -0.00002858, -0.00005075, -0.00006604, -0.00007438, -0.00007906, -0.00008271, -0.00008460, -0.00008022, -0.00006207, -0.00002055, },
		{ 0.00036947, 0.00037994, 0.00031614, 0.00017663, -0.00000794, -0.00018025, -0.00027930, -0.00027143, -0.00017133, -0.00003828, 0.00005416, 0.00005904, -0.00001639, -0.00011489, -0.00016438, -0.00012347, -0.00000903, 0.00011246, 0.00016559, 0.00011367, -0.00001684, -0.00015108, -0.00021095, -0.00016253, -0.00003608, 0.00009352, 0.00015303, 0.00011434, 0.00000869, -0.00009593, -0.00013816, -0.00009929, -0.00001056, 0.00007073, 0.00009847, 0.00006409, -0.00000321, -0.00005956, -0.00007537, -0.00005110, -0.00001158, 0.00001447, 0.00001399, -0.00000465, -0.00002150, -0.00002162, -0.00000567, 0.00001200, 0.00001507, -0.00000273, -0.00003389, -0.00006313, -0.00007789, -0.00007538, -0.00006195, -0.00004663, -0.00003424, -0.00002272, -0.00000532, 0.00002532, 0.00007415, 0.00014210, 0.00022442, 0.00030792, },
		{ 0.00005426, -0.00006975, -0.00016717, -0.00020468, -0.00017340, -0.00009688, -0.00002045, 0.00001407, -0.00000648, -0.00005813, -0.00009516, -0.00008085, -0.00001381, 0.00006814, 0.00011087, 0.00007983, -0.00001467, -0.00012041, -0.00017406, -0.00014100, -0.00003827, 0.00007470, 0.00013234, 0.00010333, 0.00000934, -0.00008958, -0.00013318, -0.00009797, -0.00000984, 0.00007479, 0.00010568, 0.00006915, -0.00000660, -0.00007272, -0.00009198, -0.00006043, -0.00000562, 0.00003503, 0.00003951, 0.00001320, -0.00001899, -0.00003267, -0.00002120, 0.00000158, 0.00001344, 0.00000082, -0.00003160, -0.00006491, -0.00007951, -0.00006796, -0.00003834, -0.00000698, 0.00001324, 0.00002060, 0.00002385, 0.00003500, 0.00006180, 0.00010440, 0.00015671, 0.00020907, 0.00024955, 0.00026405, 0.00023862, 0.00016603, },
		{ -0.00015355, -0.00012206, -0.00006802, -0.00002162, -0.00000502, -0.00001850, -0.00004074, -0.00004442, -0.00001703, 0.00002791, 0.00005831, 0.00004583, -0.00001187, -0.00008519, -0.00012907, -0.00011201, -0.00003878, 0.00004974, 0.00010079, 0.00008320, 0.00000820, -0.00007677, -0.00011791, -0.00008929, -0.00000930, 0.00007155, 0.00010323, 0.00006788, -0.00001004, -0.00008089, -0.00010233, -0.00006539, 0.00000231, 0.00005694, 0.00006754, 0.00003428, -0.00001523, -0.00004726, -0.00004520, -0.00001875, 0.00000590, 0.00000647, -0.00001933, -0.00005306, -0.00006962, -0.00005516, -0.00001621, 0.00002575, 0.00004954, 0.00004809, 0.00003180, 0.00002013, 0.00002861, 0.00006009, 0.00010516, 0.00014931, 0.00018036, 0.00019117, 0.00017798, 0.00013853, 0.00007369, -0.00000795, -0.00008782, -0.00014218, },
		{ -0.00001278, -0.00000138, -0.00000829, -0.00001864, -0.00001719, -0.00000105, 0.00001689, 0.00001746, -0.00000885, -0.00005145, -0.00008383, -0.00008090, -0.00003804, 0.00002239, 0.00006335, 0.00005722, 0.00000564, -0.00005920, -0.00009432, -0.00007423, -0.00000894, 0.00006111, 0.00009079, 0.00005992, -0.00001294, -0.00008195, -0.00010346, -0.00006402, 0.00001146, 0.00007645, 0.00009292, 0.00005469, -0.00001079, -0.00006250, -0.00007310, -0.00004511, -0.00000583, 0.00001399, 0.00000200, -0.00002840, -0.00004946, -0.00004018, -0.00000162, 0.00004395, 0.00006902, 0.00006002, 0.00002645, -0.00000586, -0.00001199, 0.00001661, 0.00006820, 0.00011972, 0.00015100, 0.00015433, 0.00013414, 0.00009944, 0.00005660, 0.00000812, -0.00004287, -0.00008792, -0.00011440, -0.00011280, -0.00008477, -0.00004483, },
		{ -0.00001368, -0.00001427, -0.00000700, -0.00000074, -0.00000597, -0.00002459, -0.00004587, -0.00005310, -0.00003660, -0.00000316, 0.00002610, 0.00002976, 0.00000229, -0.00003949, -0.00006600, -0.00005502, -0.00000879, 0.00004494, 0.00006993, 0.00004613, -0.00001486, -0.00007512, -0.00009407, -0.00005539, 0.00002096, 0.00009024, 0.00011084, 0.00007051, -0.00000637, -0.00007505, -0.00009903, -0.00007205, -0.00001925, 0.00002212, 0.00002857, 0.00000474, -0.00002259, -0.00002588, 0.00000243, 0.00004440, 0.00006981, 0.00005850, 0.00001564, -0.00003067, -0.00004766, -0.00001940, 0.00004299, 0.00010872, 0.00014652, 0.00014244, 0.00010497, 0.00005591, 0.00001484, -0.00001155, -0.00002875, -0.00004515, -0.00006264, -0.00007462, -0.00007250, -0.00005431, -0.00002818, -0.00000728, -0.00000042, -0.00000590, },
		{ -0.00000345, -0.00000828, -0.00002083, -0.00003308, -0.00003494, -0.00002306, -0.00000502, 0.00000538, -0.00000113, -0.00002071, -0.00003747, -0.00003475, -0.00000882, 0.00002565, 0.00004388, 0.00002854, -0.00001556, -0.00006120, -0.00007494, -0.00003995, 0.00003002, 0.00009624, 0.00011804, 0.00007873, -0.00000262, -0.00008202, -0.00011734, -0.00009383, -0.00003164, 0.00002903, 0.00005479, 0.00003967, 0.00000592, -0.00001444, -0.00000380, 0.00002854, 0.00005406, 0.00004718, 0.00000567, -0.00004542, -0.00006951, -0.00004337, 0.00002614, 0.00010474, 0.00015137, 0.00014292, 0.00008702, 0.00001537, -0.00003704, -0.00005251, -0.00003795, -0.00001598, -0.00000746, -0.00001817, -0.00003791, -0.00005076, -0.00004764, -0.00003189, -0.00001485, -0.00000618, -0.00000696, -0.00001080, -0.00001084, -0.00000643, },
		{ -0.00003340, -0.00003455, -0.00002537, -0.00001205, -0.00000393, -0.00000575, -0.00001332, -0.00001672, -0.00000902, 0.00000645, 0.00001695, 0.00000999, -0.00001500, -0.00004242, -0.00004883, -0.00001951, 0.00003797, 0.00009410, 0.00011356, 0.00007795, -0.00000009, -0.00008173, -0.00012405, -0.00010591, -0.00004063, 0.00003286, 0.00007482, 0.00006904, 0.00003066, -0.00000705, -0.00001729, 0.00000173, 0.00002723, 0.00003057, 0.00000053, -0.00004588, -0.00007351, -0.00005355, 0.00001441, 0.00009850, 0.00015254, 0.00014438, 0.00007630, -0.00001596, -0.00008483, -0.00009909, -0.00006078, -0.00000101, 0.00004132, 0.00004323, 0.00000898, -0.00003664, -0.00006637, -0.00006719, -0.00004509, -0.00001719, 0.00000150, 0.00000665, 0.00000366, 0.00000013, -0.00000130, -0.00000377, -0.00001135, -0.00002332, },
		{ -0.00000566, 0.00000322, 0.00000275, -0.00000379, -0.00000933, -0.00000944, -0.00000631, -0.00000641, -0.00001334, -0.00002198, -0.00001997, 0.00000301, 0.00004436, 0.00008515, 0.00009894, 0.00006870, 0.00000088, -0.00007420, -0.00011785, -0.00010602, -0.00004476, 0.00003232, 0.00008410, 0.00008672, 0.00004705, -0.00000396, -0.00003337, -0.00002837, -0.00000327, 0.00001331, 0.00000160, -0.00003265, -0.00006064, -0.00005109, 0.00000520, 0.00008448, 0.00014177, 0.00013864, 0.00006901, -0.00003427, -0.00011747, -0.00013751, -0.00008732, 0.00000083, 0.00007544, 0.00009645, 0.00005706, -0.00001559, -0.00007952, -0.00010334, -0.00008202, -0.00003498, 0.00000962, 0.00003241, 0.00003106, 0.00001625, 0.00000066, -0.00000972, -0.00001661, -0.00002396, -0.00003205, -0.00003650, -0.00003253, -0.00002021, },
		{ -0.00000970, -0.00001949, -0.00002208, -0.00001789, -0.00001085, -0.00000336, 0.00000682, 0.00002418, 0.00004893, 0.00007209, 0.00007780, 0.00005325, 0.00000018, -0.00006116, -0.00010045, -0.00009469, -0.00004373, 0.00002713, 0.00008058, 0.00008931, 0.00005247, -0.00000455, -0.00004709, -0.00005389, -0.00002990, -0.00000064, 0.00000802, -0.00001009, -0.00003598, -0.00003950, -0.00000312, 0.00006137, 0.00011644, 0.00012177, 0.00006214, -0.00003870, -0.00012890, -0.00015831, -0.00010916, -0.00000785, 0.00008986, 0.00013117, 0.00009630, 0.00000814, -0.00008330, -0.00013088, -0.00011591, -0.00005475, 0.00001518, 0.00005923, 0.00006349, 0.00003746, 0.00000272, -0.00002255, -0.00003264, -0.00003275, -0.00003069, -0.00002955, -0.00002674, -0.00001871, -0.00000614, 0.00000518, 0.00000864, 0.00000220, },
		{ -0.00000779, 0.00001049, 0.00002719, 0.00004074, 0.00005161, 0.00005824, 0.00005494, 0.00003504, -0.00000208, -0.00004572, -0.00007617, -0.00007511, -0.00003839, 0.00001815, 0.00006530, 0.00007703, 0.00004679, -0.00000758, -0.00005454, -0.00006867, -0.00004673, -0.00000862, 0.00001740, 0.00001534, -0.00000691, -0.00002361, -0.00001127, 0.00003151, 0.00007911, 0.00009420, 0.00005382, -0.00003133, -0.00011868, -0.00015756, -0.00012070, -0.00002322, 0.00008372, 0.00014184, 0.00011961, 0.00003043, -0.00007594, -0.00014280, -0.00013848, -0.00007189, 0.00001640, 0.00007996, 0.00009246, 0.00005851, 0.00000521, -0.00003735, -0.00005329, -0.00004570, -0.00002891, -0.00001596, -0.00001078, -0.00000913, -0.00000557, 0.00000047, 0.00000429, 0.00000081, -0.00001006, -0.00002235, -0.00002800, -0.00002256, },
		{ 0.00005244, 0.00004672, 0.00003519, 0.00001783, -0.00000561, -0.00003159, -0.00005089, -0.00005229, -0.00003052, 0.00000722, 0.00004213, 0.00005361, 0.00003246, -0.00001151, -0.00005377, -0.00006975, -0.00005068, -0.00000967, 0.00002685, 0.00003708, 0.00001899, -0.00000838, -0.00001911, -0.00000039, 0.00003596, 0.00005973, 0.00004334, -0.00001638, -0.00009137, -0.00013692, -0.00011963, -0.00004086, 0.00006035, 0.00012842, 0.00012393, 0.00004788, -0.00005813, -0.00013597, -0.00014460, -0.00008280, 0.00001240, 0.00008910, 0.00011074, 0.00007452, 0.00000839, -0.00004904, -0.00007171, -0.00005801, -0.00002621, 0.00000093, 0.00001074, 0.00000536, -0.00000439, -0.00000934, -0.00000836, -0.00000694, -0.00001037, -0.00001812, -0.00002369, -0.00001982, -0.00000439, 0.00001775, 0.00003818, 0.00005028, },
		{ -0.00001006, -0.00002224, -0.00003068, -0.00003187, -0.00002232, -0.00000323, 0.00001690, 0.00002546, 0.00001388, -0.00001480, -0.00004517, -0.00005784, -0.00004213, -0.00000455, 0.00003376, 0.00005019, 0.00003587, 0.00000229, -0.00002605, -0.00002888, -0.00000535, 0.00002402, 0.00003106, 0.00000087, -0.00005486, -0.00010238, -0.00010681, -0.00005643, 0.00002639, 0.00009597, 0.00011044, 0.00005829, -0.00003287, -0.00011165, -0.00013314, -0.00008553, 0.00000348, 0.00008404, 0.00011388, 0.00008201, 0.00001233, -0.00005352, -0.00008189, -0.00006518, -0.00002215, 0.00001747, 0.00003242, 0.00002077, -0.00000323, -0.00002180, -0.00002586, -0.00001880, -0.00001057, -0.00000829, -0.00001076, -0.00001048, -0.00000078, 0.00001814, 0.00003884, 0.00005176, 0.00005154, 0.00003970, 0.00002222, 0.00000482, },
		{ -0.00001588, -0.00001075, -0.00000404, 0.00000007, -0.00000361, -0.00001616, -0.00003115, -0.00003698, -0.00002459, 0.00000452, 0.00003654, 0.00005254, 0.00004088, 0.00000632, -0.00003134, -0.00004943, -0.00003792, -0.00000718, 0.00001798, 0.00001536, -0.00001818, -0.00006233, -0.00008555, -0.00006639, -0.00000985, 0.00005310, 0.00008381, 0.00006082, -0.00000488, -0.00007501, -0.00010692, -0.00008001, -0.00000887, 0.00006556, 0.00010108, 0.00007956, 0.00001684, -0.00004859, -0.00008000, -0.00006406, -0.00001649, 0.00003043, 0.00004923, 0.00003316, -0.00000272, -0.00003412, -0.00004399, -0.00003182, -0.00001060, 0.00000461, 0.00000754, 0.00000345, 0.00000300, 0.00001290, 0.00003042, 0.00004568, 0.00004904, 0.00003786, 0.00001773, -0.00000213, -0.00001504, -0.00002004, -0.00002024, -0.00001870, },
		{ -0.00001497, -0.00001478, -0.00001538, -0.00001330, -0.00000369, 0.00001451, 0.00003483, 0.00004501, 0.00003464, 0.00000374, -0.00003439, -0.00005957, -0.00005736, -0.00002954, 0.00000542, 0.00002341, 0.00001082, -0.00002525, -0.00006039, -0.00006859, -0.00004030, 0.00000975, 0.00005083, 0.00005601, 0.00002057, -0.00003372, -0.00007178, -0.00006779, -0.00002232, 0.00003751, 0.00007515, 0.00006796, 0.00002150, -0.00003442, -0.00006523, -0.00005352, -0.00000930, 0.00003765, 0.00005761, 0.00003954, -0.00000334, -0.00004377, -0.00005869, -0.00004299, -0.00001045, 0.00001760, 0.00002708, 0.00001916, 0.00000698, 0.00000447, 0.00001601, 0.00003396, 0.00004507, 0.00004034, 0.00002105, -0.00000289, -0.00002025, -0.00002568, -0.00002184, -0.00001568, -0.00001265, -0.00001348, -0.00001536, -0.00001578, },
		{ 0.00001434, 0.00002226, 0.00002941, 0.00003097, 0.00002068, -0.00000353, -0.00003496, -0.00005936, -0.00006272, -0.00004110, -0.00000541, 0.00002344, 0.00002711, 0.00000224, -0.00003598, -0.00006261, -0.00005889, -0.00002511, 0.00001874, 0.00004552, 0.00003879, 0.00000395, -0.00003505, -0.00005157, -0.00003413, 0.00000584, 0.00004175, 0.00004992, 0.00002569, -0.00001344, -0.00003985, -0.00003471, -0.00000097, 0.00003847, 0.00005615, 0.00003847, -0.00000526, -0.00004898, -0.00006691, -0.00004987, -0.00001010, 0.00002788, 0.00004370, 0.00003366, 0.00001115, -0.00000412, -0.00000020, 0.00001952, 0.00003982, 0.00004497, 0.00002946, 0.00000103, -0.00002542, -0.00003801, -0.00003475, -0.00002305, -0.00001308, -0.00001057, -0.00001403, -0.00001767, -0.00001664, -0.00001038, -0.00000169, 0.00000667, },
		{ 0.00000425, -0.00001241, -0.00003323, -0.00005118, -0.00005634, -0.00004257, -0.00001368, 0.00001612, 0.00002944, 0.00001677, -0.00001605, -0.00004973, -0.00006283, -0.00004549, -0.00000646, 0.00003171, 0.00004679, 0.00003108, -0.00000373, -0.00003445, -0.00004180, -0.00002286, 0.00000789, 0.00002933, 0.00002869, 0.00001022, -0.00000867, -0.00001075, 0.00000778, 0.00003369, 0.00004580, 0.00003032, -0.00000834, -0.00004903, -0.00006722, -0.00005109, -0.00000955, 0.00003341, 0.00005414, 0.00004433, 0.00001542, -0.00001003, -0.00001391, 0.00000559, 0.00003353, 0.00004892, 0.00003892, 0.00000692, -0.00003038, -0.00005425, -0.00005543, -0.00003840, -0.00001656, -0.00000241, -0.00000007, -0.00000459, -0.00000741, -0.00000323, 0.00000691, 0.00001773, 0.00002431, 0.00002522, 0.00002179, 0.00001513, },
		{ -0.00004299, -0.00003679, -0.00001903, 0.00000403, 0.00002021, 0.00001852, -0.00000271, -0.00003261, -0.00005299, -0.00004942, -0.00002113, 0.00001724, 0.00004398, 0.00004370, 0.00001716, -0.00001917, -0.00004362, -0.00004288, -0.00001972, 0.00001022, 0.00002980, 0.00003175, 0.00002216, 0.00001403, 0.00001605, 0.00002531, 0.00002952, 0.00001712, -0.00001209, -0.00004439, -0.00006001, -0.00004663, -0.00000883, 0.00003335, 0.00005665, 0.00004950, 0.00001958, -0.00001129, -0.00002172, -0.00000496, 0.00002654, 0.00004951, 0.00004517, 0.00001162, -0.00003499, -0.00007070, -0.00007839, -0.00005723, -0.00002155, 0.00000946, 0.00002348, 0.00002081, 0.00001132, 0.00000592, 0.00000909, 0.00001731, 0.00002331, 0.00002205, 0.00001371, 0.00000210, -0.00000927, -0.00001962, -0.00002976, -0.00003892, },
		{ 0.00000445, 0.00001076, 0.00000366, -0.00001472, -0.00003350, -0.00003912, -0.00002459, 0.00000455, 0.00003228, 0.00004162, 0.00002561, -0.00000756, -0.00003904, -0.00005076, -0.00003629, -0.00000411, 0.00002852, 0.00004677, 0.00004661, 0.00003505, 0.00002290, 0.00001597, 0.00001145, 0.00000206, -0.00001578, -0.00003642, -0.00004728, -0.00003767, -0.00000798, 0.00002816, 0.00005130, 0.00004867, 0.00002325, -0.00000727, -0.00002185, -0.00001025, 0.00001933, 0.00004493, 0.00004490, 0.00001243, -0.00003909, -0.00008353, -0.00009754, -0.00007413, -0.00002605, 0.00002271, 0.00005130, 0.00005287, 0.00003586, 0.00001598, 0.00000536, 0.00000606, 0.00001143, 0.00001290, 0.00000660, -0.00000490, -0.00001589, -0.00002256, -0.00002529, -0.00002679, -0.00002833, -0.00002778, -0.00002160, -0.00000919, },
		{ -0.00001055, -0.00002002, -0.00001882, -0.00000447, 0.00001566, 0.00002828, 0.00002291, -0.00000018, -0.00002887, -0.00004600, -0.00003993, -0.00001193, 0.00002458, 0.00005227, 0.00006006, 0.00004855, 0.00002754, 0.00000823, -0.00000406, -0.00001137, -0.00001858, -0.00002705, -0.00003205, -0.00002631, -0.00000706, 0.00001939, 0.00003984, 0.00004255, 0.00002596, 0.00000120, -0.00001454, -0.00000973, 0.00001249, 0.00003485, 0.00003663, 0.00000785, -0.00004246, -0.00008972, -0.00010770, -0.00008410, -0.00002809, 0.00003476, 0.00007705, 0.00008428, 0.00006104, 0.00002569, -0.00000209, -0.00001240, -0.00000832, -0.00000062, 0.00000126, -0.00000479, -0.00001400, -0.00002008, -0.00002062, -0.00001825, -0.00001737, -0.00001954, -0.00002190, -0.00001985, -0.00001177, -0.00000142, 0.00000418, 0.00000039, },
		{ -0.00000076, 0.00000979, 0.00001313, 0.00000361, -0.00001505, -0.00003110, -0.00003207, -0.00001334, 0.00001811, 0.00004736, 0.00006044, 0.00005242, 0.00002948, 0.00000399, -0.00001355, -0.00002011, -0.00001974, -0.00001821, -0.00001762, -0.00001501, -0.00000612, 0.00000927, 0.00002518, 0.00003281, 0.00002720, 0.00001203, -0.00000190, -0.00000420, 0.00000666, 0.00002050, 0.00002118, -0.00000200, -0.00004481, -0.00008770, -0.00010582, -0.00008382, -0.00002626, 0.00004338, 0.00009495, 0.00010785, 0.00008146, 0.00003378, -0.00001018, -0.00003256, -0.00003086, -0.00001585, -0.00000223, 0.00000118, -0.00000469, -0.00001268, -0.00001640, -0.00001513, -0.00001316, -0.00001498, -0.00002061, -0.00002517, -0.00002299, -0.00001267, 0.00000092, 0.00000990, 0.00000923, 0.00000081, -0.00000772, -0.00000873, },
		{ -0.00000029, -0.00001090, -0.00001690, -0.00001017, 0.00000952, 0.00003333, 0.00004855, 0.00004664, 0.00002860, 0.00000403, -0.00001522, -0.00002233, -0.00001873, -0.00001138, -0.00000671, -0.00000596, -0.00000524, 0.00000011, 0.00001064, 0.00002163, 0.00002655, 0.00002244, 0.00001266, 0.00000440, 0.00000249, 0.00000451, 0.00000148, -0.00001528, -0.00004578, -0.00007777, -0.00009174, -0.00007250, -0.00002013, 0.00004715, 0.00010109, 0.00011823, 0.00009269, 0.00003916, -0.00001581, -0.00004857, -0.00005036, -0.00002965, -0.00000436, 0.00001000, 0.00000866, -0.00000238, -0.00001308, -0.00001732, -0.00001622, -0.00001529, -0.00001844, -0.00002396, -0.00002582, -0.00001886, -0.00000382, 0.00001223, 0.00002045, 0.00001659, 0.00000416, -0.00000798, -0.00001199, -0.00000668, 0.00000190, 0.00000535, },
		{ -0.00000046, 0.00001338, 0.00002783, 0.00003322, 0.00002522, 0.00000787, -0.00000930, -0.00001775, -0.00001540, -0.00000723, -0.00000086, -0.00000061, -0.00000448, -0.00000626, -0.00000088, 0.00001115, 0.00002378, 0.00002963, 0.00002533, 0.00001359, 0.00000047, -0.00000977, -0.00001805, -0.00002892, -0.00004502, -0.00006191, -0.00006822, -0.00005209, -0.00001039, 0.00004581, 0.00009441, 0.00011316, 0.00009233, 0.00004107, -0.00001685, -0.00005577, -0.00006180, -0.00003916, -0.00000588, 0.00001793, 0.00002154, 0.00000792, -0.00001087, -0.00002308, -0.00002461, -0.00001978, -0.00001598, -0.00001688, -0.00001968, -0.00001804, -0.00000811, 0.00000737, 0.00002022, 0.00002257, 0.00001281, -0.00000282, -0.00001431, -0.00001477, -0.00000511, 0.00000690, 0.00001256, 0.00000863, -0.00000031, -0.00000543, },
		{ 0.00002005, 0.00001398, 0.00000207, -0.00000770, -0.00000994, -0.00000548, -0.00000006, 0.00000071, -0.00000389, -0.00000890, -0.00000759, 0.00000298, 0.00001894, 0.00003155, 0.00003287, 0.00002095, 0.00000086, -0.00001905, -0.00003269, -0.00003937, -0.00004220, -0.00004333, -0.00004025, -0.00002683, 0.00000123, 0.00004018, 0.00007682, 0.00009400, 0.00008051, 0.00003919, -0.00001273, -0.00005205, -0.00006223, -0.00004242, -0.00000738, 0.00002169, 0.00002963, 0.00001526, -0.00000976, -0.00002968, -0.00003481, -0.00002637, -0.00001365, -0.00000613, -0.00000673, -0.00001076, -0.00001076, -0.00000311, 0.00000880, 0.00001711, 0.00001578, 0.00000555, -0.00000616, -0.00001050, -0.00000372, 0.00000994, 0.00002107, 0.00002187, 0.00001168, -0.00000250, -0.00001097, -0.00000818, 0.00000351, 0.00001571, },
		{ -0.00000293, -0.00000508, -0.00000287, -0.00000124, -0.00000351, -0.00000802, -0.00000917, -0.00000215, 0.00001240, 0.00002742, 0.00003342, 0.00002474, 0.00000359, -0.00002098, -0.00003855, -0.00004347, -0.00003719, -0.00002563, -0.00001382, -0.00000220, 0.00001237, 0.00003192, 0.00005274, 0.00006535, 0.00005992, 0.00003374, -0.00000462, -0.00003837, -0.00005159, -0.00003886, -0.00000908, 0.00001949, 0.00003012, 0.00001748, -0.00000963, -0.00003442, -0.00004294, -0.00003237, -0.00001171, 0.00000516, 0.00000893, 0.00000036, -0.00001149, -0.00001649, -0.00001061, 0.00000193, 0.00001264, 0.00001554, 0.00001143, 0.00000670, 0.00000772, 0.00001535, 0.00002381, 0.00002490, 0.00001438, -0.00000417, -0.00002110, -0.00002694, -0.00001865, -0.00000162, 0.00001404, 0.00002018, 0.00001545, 0.00000526, },
		},
		{ { 0.00001500, 0.00001574, 0.00001695, 0.00001466, 0.00000226, -0.00002404, -0.00005713, -0.00007816, -0.00006865, -0.00002782, 0.00002127, 0.00004670, 0.00003358, -0.00000185, -0.00002507, -0.00001343, 0.00002304, 0.00004997, 0.00003947, -0.00000460, -0.00004971, -0.00006297, -0.00003831, 0.00000131, 0.00002637, 0.00002554, 0.00001045, -0.00000029, 0.00000052, 0.00000568, 0.00000565, 0.00000015, -0.00000322, 0.00000015, 0.00000565, 0.00000568, 0.00000052, -0.00000029, 0.00001045, 0.00002554, 0.00002637, 0.00000131, -0.00003831, -0.00006297, -0.00004971, -0.00000460, 0.00003947, 0.00004997, 0.00002304, -0.00001343, -0.00002507, -0.00000185, 0.00003358, 0.00004670, 0.00002127, -0.00002782, -0.00006865, -0.00007816, -0.00005713, -0.00002404, 0.00000226, 0.00001466, 0.00001695, 0.00001574, },
		{ 0.00001542, -0.00001071, -0.00004389, -0.00007270, -0.00008303, -0.00006484, -0.00002192, 0.00002504, 0.00004943, 0.00003861, 0.00000591, -0.00001867, -0.00001367, 0.00001446, 0.00003658, 0.00002739, -0.00001050, -0.00004770, -0.00005393, -0.00002479, 0.00001516, 0.00003521, 0.00002514, 0.00000123, -0.00001203, -0.00000508, 0.00001053, 0.00001682, 0.00000791, -0.00000610, -0.00001156, -0.00000514, 0.00000535, 0.00001179, 0.00001426, 0.00001721, 0.00001918, 0.00001055, -0.00001521, -0.00004824, -0.00006416, -0.00004401, 0.00000508, 0.00005050, 0.00005943, 0.00002752, -0.00001595, -0.00003308, -0.00001071, 0.00002810, 0.00004492, 0.00002014, -0.00003045, -0.00007028, -0.00007319, -0.00004281, -0.00000518, 0.00001548, 0.00001483, 0.00000622, 0.00000474, 0.00001387, 0.00002528, 0.00002778, },
		{ -0.00008836, -0.00008678, -0.00005816, -0.00001237, 0.00003130, 0.00005215, 0.00004131, 0.00001029, -0.00001556, -0.00001666, 0.00000393, 0.00002283, 0.00001776, -0.00001083, -0.00003883, -0.00004025, -0.00001177, 0.00002344, 0.00003681, 0.00002041, -0.00000688, -0.00001842, -0.00000517, 0.00001713, 0.00002483, 0.00000979, -0.00001369, -0.00002384, -0.00001212, 0.00001066, 0.00002678, 0.00002786, 0.00001837, 0.00000562, -0.00000966, -0.00003019, -0.00005099, -0.00005624, -0.00003216, 0.00001505, 0.00005676, 0.00006257, 0.00002764, -0.00001969, -0.00004034, -0.00001863, 0.00002315, 0.00004405, 0.00002104, -0.00003043, -0.00007025, -0.00006868, -0.00003055, 0.00001186, 0.00002856, 0.00001621, -0.00000357, -0.00000699, 0.00001056, 0.00003268, 0.00003784, 0.00001667, -0.00002305, -0.00006373, },
		{ 0.00000046, 0.00003884, 0.00005297, 0.00003953, 0.00000967, -0.00001588, -0.00002055, -0.00000516, 0.00001232, 0.00001271, -0.00000608, -0.00002607, -0.00002596, -0.00000238, 0.00002539, 0.00003304, 0.00001466, -0.00001108, -0.00001892, -0.00000178, 0.00002212, 0.00002718, 0.00000588, -0.00002326, -0.00003307, -0.00001322, 0.00002064, 0.00004199, 0.00003677, 0.00001248, -0.00001324, -0.00003003, -0.00003944, -0.00004469, -0.00004052, -0.00001772, 0.00002139, 0.00005571, 0.00005904, 0.00002492, -0.00002226, -0.00004476, -0.00002455, 0.00001867, 0.00004334, 0.00002320, -0.00002819, -0.00006873, -0.00006490, -0.00002096, 0.00002627, 0.00004101, 0.00001907, -0.00001159, -0.00001866, 0.00000516, 0.00003723, 0.00004610, 0.00001790, -0.00003424, -0.00008217, -0.00010244, -0.00008815, -0.00004809, },
		{ 0.00003180, 0.00000398, -0.00001795, -0.00002225, -0.00000935, 0.00000736, 0.00001216, 0.00000115, -0.00001339, -0.00001448, 0.00000216, 0.00002247, 0.00002701, 0.00001082, -0.00001036, -0.00001499, 0.00000205, 0.00002280, 0.00002327, -0.00000179, -0.00003134, -0.00003656, -0.00000840, 0.00003247, 0.00005349, 0.00003856, 0.00000039, -0.00003355, -0.00004588, -0.00004022, -0.00003003, -0.00002056, -0.00000510, 0.00002107, 0.00004685, 0.00005052, 0.00002194, -0.00002150, -0.00004532, -0.00002862, 0.00001385, 0.00004195, 0.00002615, -0.00002380, -0.00006565, -0.00006207, -0.00001468, 0.00003730, 0.00005242, 0.00002368, -0.00001700, -0.00002953, -0.00000233, 0.00003811, 0.00005155, 0.00001880, -0.00004340, -0.00009736, -0.00011301, -0.00008604, -0.00003489, 0.00001508, 0.00004546, 0.00004966, },
		{ -0.00001841, -0.00000634, 0.00000812, 0.00001387, 0.00000708, -0.00000437, -0.00000749, 0.00000285, 0.00001762, 0.00002185, 0.00001032, -0.00000573, -0.00000941, 0.00000353, 0.00001814, 0.00001457, -0.00001008, -0.00003500, -0.00003348, 0.00000036, 0.00004250, 0.00005827, 0.00003299, -0.00001487, -0.00005061, -0.00005362, -0.00003219, -0.00001014, -0.00000118, 0.00000188, 0.00001307, 0.00003192, 0.00004006, 0.00002127, -0.00001638, -0.00004244, -0.00003191, 0.00000778, 0.00003948, 0.00002992, -0.00001712, -0.00006106, -0.00006055, -0.00001230, 0.00004437, 0.00006262, 0.00003043, -0.00001915, -0.00003919, -0.00001214, 0.00003450, 0.00005346, 0.00001956, -0.00004908, -0.00010712, -0.00011826, -0.00008008, -0.00001992, 0.00002893, 0.00004848, 0.00004058, 0.00001813, -0.00000498, -0.00001873, },
		{ 0.00001420, 0.00000809, -0.00000099, -0.00000441, 0.00000238, 0.00001407, 0.00001943, 0.00001264, 0.00000023, -0.00000514, 0.00000116, 0.00000909, 0.00000394, -0.00001592, -0.00003288, -0.00002515, 0.00000979, 0.00004746, 0.00005524, 0.00002213, -0.00002902, -0.00006032, -0.00005184, -0.00001763, 0.00001019, 0.00001303, 0.00000122, -0.00000128, 0.00001439, 0.00003101, 0.00002444, -0.00000742, -0.00003769, -0.00003571, 0.00000006, 0.00003625, 0.00003490, -0.00000813, -0.00005536, -0.00006087, -0.00001421, 0.00004740, 0.00007180, 0.00003964, -0.00001778, -0.00004770, -0.00002476, 0.00002583, 0.00005176, 0.00002112, -0.00004939, -0.00010941, -0.00011720, -0.00007110, -0.00000574, 0.00003887, 0.00004571, 0.00002550, 0.00000048, -0.00001385, -0.00001470, -0.00000679, 0.00000400, 0.00001245, },
		{ -0.00000280, 0.00000398, 0.00001407, 0.00001975, 0.00001555, 0.00000449, -0.00000413, -0.00000460, -0.00000172, -0.00000540, -0.00001749, -0.00002573, -0.00001448, 0.00001657, 0.00004574, 0.00004569, 0.00000972, -0.00003788, -0.00006053, -0.00004183, -0.00000078, 0.00002603, 0.00001908, -0.00000660, -0.00001863, -0.00000172, 0.00002580, 0.00003129, 0.00000342, -0.00003311, -0.00004088, -0.00000882, 0.00003316, 0.00004145, 0.00000269, -0.00004938, -0.00006355, -0.00002029, 0.00004689, 0.00008035, 0.00005125, -0.00001340, -0.00005572, -0.00004060, 0.00001225, 0.00004739, 0.00002516, -0.00004242, -0.00010300, -0.00011015, -0.00006128, 0.00000443, 0.00004213, 0.00003631, 0.00000617, -0.00001749, -0.00001863, -0.00000327, 0.00001278, 0.00001934, 0.00001656, 0.00000946, 0.00000209, -0.00000293, },
		{ 0.00002091, 0.00001603, 0.00000473, -0.00000649, -0.00001153, -0.00001101, -0.00001112, -0.00001491, -0.00001610, -0.00000487, 0.00001858, 0.00003797, 0.00003286, -0.00000015, -0.00003884, -0.00005173, -0.00002716, 0.00001348, 0.00003373, 0.00001645, -0.00001889, -0.00003472, -0.00001306, 0.00002518, 0.00004008, 0.00001349, -0.00003039, -0.00004729, -0.00001758, 0.00003128, 0.00004945, 0.00001424, -0.00004420, -0.00006870, -0.00002968, 0.00004388, 0.00008855, 0.00006449, -0.00000732, -0.00006422, -0.00005960, -0.00000501, 0.00004235, 0.00003370, -0.00002691, -0.00008809, -0.00009901, -0.00005371, 0.00000756, 0.00003729, 0.00002149, -0.00001379, -0.00003126, -0.00001599, 0.00001615, 0.00003789, 0.00003547, 0.00001628, -0.00000274, -0.00001070, -0.00000773, 0.00000068, 0.00001014, 0.00001797, },
		{ -0.00001041, -0.00001641, -0.00001608, -0.00001273, -0.00001027, -0.00000741, 0.00000107, 0.00001580, 0.00002691, 0.00002081, -0.00000446, -0.00003172, -0.00003690, -0.00001249, 0.00002119, 0.00003202, 0.00000712, -0.00003158, -0.00004572, -0.00001798, 0.00002803, 0.00004804, 0.00002036, -0.00003016, -0.00005376, -0.00002455, 0.00003128, 0.00005806, 0.00002494, -0.00004071, -0.00007575, -0.00004080, 0.00003959, 0.00009617, 0.00007769, -0.00000150, -0.00007401, -0.00008072, -0.00002345, 0.00003940, 0.00004849, -0.00000295, -0.00006665, -0.00008696, -0.00005154, 0.00000187, 0.00002504, 0.00000452, -0.00002973, -0.00003684, -0.00000455, 0.00004153, 0.00006401, 0.00004744, 0.00000849, -0.00002248, -0.00002766, -0.00001166, 0.00000891, 0.00002154, 0.00002409, 0.00001990, 0.00001153, 0.00000039, },
		{ -0.00001177, -0.00000681, -0.00000263, 0.00000261, 0.00001037, 0.00001648, 0.00001295, -0.00000239, -0.00001899, -0.00002053, -0.00000206, 0.00002053, 0.00002232, -0.00000519, -0.00004072, -0.00004941, -0.00001693, 0.00003190, 0.00005240, 0.00002262, -0.00003184, -0.00005853, -0.00002833, 0.00003301, 0.00006573, 0.00003321, -0.00003915, -0.00008329, -0.00005166, 0.00003493, 0.00010213, 0.00008838, 0.00000195, -0.00008519, -0.00010165, -0.00003955, 0.00004136, 0.00007014, 0.00002753, -0.00004222, -0.00007762, -0.00005692, -0.00001225, 0.00000837, -0.00001000, -0.00003724, -0.00003207, 0.00001438, 0.00006849, 0.00008561, 0.00005115, -0.00000764, -0.00004726, -0.00004457, -0.00001110, 0.00002312, 0.00003663, 0.00002924, 0.00001382, 0.00000107, -0.00000706, -0.00001279, -0.00001631, -0.00001589, },
		{ 0.00000124, 0.00000594, 0.00001041, 0.00001079, 0.00000427, -0.00000491, -0.00000714, 0.00000180, 0.00001243, 0.00000813, -0.00001638, -0.00004371, -0.00004577, -0.00001206, 0.00003398, 0.00005137, 0.00002035, -0.00003389, -0.00005990, -0.00002839, 0.00003544, 0.00007072, 0.00003797, -0.00003891, -0.00008935, -0.00006045, 0.00003002, 0.00010444, 0.00009374, 0.00000146, -0.00009668, -0.00011897, -0.00004943, 0.00005025, 0.00009756, 0.00006089, -0.00001916, -0.00007390, -0.00006995, -0.00003198, -0.00000805, -0.00001741, -0.00003366, -0.00001758, 0.00003688, 0.00009132, 0.00009772, 0.00004466, -0.00003026, -0.00007262, -0.00005688, -0.00000385, 0.00004224, 0.00005206, 0.00002992, 0.00000111, -0.00001417, -0.00001449, -0.00001037, -0.00000971, -0.00001154, -0.00001131, -0.00000775, -0.00000309, },
		{ 0.00001323, 0.00001161, 0.00000584, 0.00000033, -0.00000062, 0.00000015, -0.00000625, -0.00002344, -0.00004015, -0.00003681, -0.00000624, 0.00003223, 0.00004473, 0.00001489, -0.00003456, -0.00005698, -0.00002528, 0.00003705, 0.00007166, 0.00003916, -0.00003857, -0.00009193, -0.00006611, 0.00002394, 0.00010060, 0.00009134, -0.00000340, -0.00010613, -0.00012869, -0.00004985, 0.00006645, 0.00012784, 0.00009240, -0.00000145, -0.00007688, -0.00008821, -0.00005249, -0.00001913, -0.00001453, -0.00001912, 0.00000309, 0.00005730, 0.00010457, 0.00009746, 0.00002889, -0.00005521, -0.00009343, -0.00006126, 0.00000998, 0.00006307, 0.00006369, 0.00002365, -0.00001769, -0.00003107, -0.00001814, -0.00000103, 0.00000271, -0.00000602, -0.00001481, -0.00001494, -0.00000786, 0.00000031, 0.00000639, 0.00001077, },
		{ 0.00000152, -0.00000644, -0.00001187, -0.00001726, -0.00002544, -0.00003190, -0.00002565, -0.00000191, 0.00002618, 0.00003378, 0.00000814, -0.00003271, -0.00005007, -0.00002037, 0.00003642, 0.00006811, 0.00003769, -0.00003643, -0.00008972, -0.00006875, 0.00001498, 0.00008830, 0.00007994, -0.00001157, -0.00011038, -0.00012725, -0.00003911, 0.00008845, 0.00015680, 0.00011751, 0.00000864, -0.00008512, -0.00010708, -0.00006820, -0.00002096, -0.00000092, 0.00000312, 0.00002429, 0.00006987, 0.00010469, 0.00008491, 0.00000747, -0.00007718, -0.00010539, -0.00005660, 0.00002789, 0.00008122, 0.00006799, 0.00000992, -0.00003980, -0.00004540, -0.00001498, 0.00001527, 0.00001811, -0.00000319, -0.00002478, -0.00002799, -0.00001383, 0.00000382, 0.00001426, 0.00001709, 0.00001679, 0.00001492, 0.00000979, },
		{ -0.00002336, -0.00002389, -0.00002226, -0.00001527, -0.00000005, 0.00001705, 0.00002076, 0.00000167, -0.00002841, -0.00004056, -0.00001522, 0.00003296, 0.00006087, 0.00003511, -0.00003122, -0.00008265, -0.00006964, 0.00000126, 0.00006628, 0.00005997, -0.00002081, -0.00010639, -0.00011258, -0.00001765, 0.00011311, 0.00018011, 0.00013317, 0.00001139, -0.00009476, -0.00012086, -0.00007442, -0.00001223, 0.00002060, 0.00002731, 0.00003977, 0.00007041, 0.00009115, 0.00006321, -0.00001439, -0.00009134, -0.00010639, -0.00004445, 0.00004571, 0.00009243, 0.00006325, -0.00000939, -0.00006094, -0.00005317, -0.00000384, 0.00003624, 0.00003227, -0.00000647, -0.00004167, -0.00004412, -0.00001674, 0.00001496, 0.00002970, 0.00002625, 0.00001672, 0.00001031, 0.00000597, -0.00000115, -0.00001120, -0.00001958, },
		{ -0.00000729, 0.00000022, 0.00000724, 0.00000779, -0.00000409, -0.00002297, -0.00003033, -0.00001080, 0.00002732, 0.00005161, 0.00003287, -0.00002284, -0.00007204, -0.00007067, -0.00001836, 0.00003501, 0.00003344, -0.00002849, -0.00009235, -0.00008491, 0.00001204, 0.00013672, 0.00019480, 0.00013908, 0.00000983, -0.00010049, -0.00012452, -0.00006900, 0.00000496, 0.00004441, 0.00004670, 0.00004455, 0.00005766, 0.00006669, 0.00003764, -0.00003136, -0.00009475, -0.00009708, -0.00002849, 0.00005889, 0.00009395, 0.00005018, -0.00003046, -0.00007653, -0.00005196, 0.00001403, 0.00005776, 0.00004089, -0.00001752, -0.00006343, -0.00005887, -0.00001256, 0.00003475, 0.00005004, 0.00003331, 0.00000857, -0.00000380, -0.00000356, -0.00000322, -0.00001047, -0.00002085, -0.00002556, -0.00002207, -0.00001467, },
		{ -0.00000410, -0.00001016, -0.00001849, -0.00002086, -0.00000681, 0.00002132, 0.00004224, 0.00003143, -0.00001270, -0.00006023, -0.00007337, -0.00004344, -0.00000314, 0.00000335, -0.00003274, -0.00006869, -0.00004687, 0.00004635, 0.00015653, 0.00020060, 0.00013799, 0.00000889, -0.00009733, -0.00011545, -0.00005335, 0.00002537, 0.00006349, 0.00005548, 0.00003644, 0.00003384, 0.00003663, 0.00001405, -0.00003970, -0.00008726, -0.00008065, -0.00001336, 0.00006396, 0.00008544, 0.00003181, -0.00004865, -0.00008313, -0.00004167, 0.00003522, 0.00007506, 0.00004098, -0.00003554, -0.00008588, -0.00006742, 0.00000103, 0.00006137, 0.00007028, 0.00003307, -0.00001014, -0.00002587, -0.00001427, 0.00000027, -0.00000150, -0.00001628, -0.00002714, -0.00002379, -0.00001134, -0.00000129, 0.00000128, -0.00000067, },
		{ -0.00001236, -0.00000157, 0.00001727, 0.00003389, 0.00002971, -0.00000350, -0.00004961, -0.00007798, -0.00007154, -0.00004445, -0.00002762, -0.00003369, -0.00003843, -0.00000277, 0.00008208, 0.00017230, 0.00020058, 0.00013514, 0.00001373, -0.00008254, -0.00009483, -0.00003247, 0.00004195, 0.00007148, 0.00005055, 0.00001688, 0.00000405, 0.00000734, -0.00000285, -0.00003833, -0.00007151, -0.00006177, -0.00000314, 0.00005961, 0.00006905, 0.00001253, -0.00005989, -0.00007944, -0.00002468, 0.00005508, 0.00008415, 0.00003187, -0.00005728, -0.00010388, -0.00006611, 0.00002358, 0.00009032, 0.00008452, 0.00002174, -0.00003882, -0.00005139, -0.00002046, 0.00001300, 0.00001568, -0.00000916, -0.00003236, -0.00003185, -0.00001227, 0.00000505, 0.00000603, -0.00000512, -0.00001551, -0.00001863, -0.00001687, },
		{ 0.00002609, 0.00002502, 0.00000161, -0.00004147, -0.00008265, -0.00009854, -0.00008503, -0.00005842, -0.00003409, -0.00000669, 0.00004292, 0.00011815, 0.00018721, 0.00020083, 0.00013661, 0.00002761, -0.00005724, -0.00006786, -0.00001373, 0.00004786, 0.00006463, 0.00003264, -0.00000944, -0.00002512, -0.00001563, -0.00001087, -0.00002906, -0.00005206, -0.00004514, -0.00000014, 0.00004708, 0.00004879, -0.00000322, -0.00006190, -0.00006674, -0.00000511, 0.00006912, 0.00008305, 0.00001550, -0.00007793, -0.00011284, -0.00005377, 0.00005175, 0.00011552, 0.00008764, -0.00000160, -0.00007362, -0.00007407, -0.00001713, 0.00003597, 0.00003791, -0.00000446, -0.00004494, -0.00004678, -0.00001507, 0.00001511, 0.00001657, -0.00000640, -0.00002861, -0.00003179, -0.00001882, -0.00000342, 0.00000747, 0.00001696, },
		{ -0.00003499, -0.00008349, -0.00011969, -0.00012236, -0.00009075, -0.00003921, 0.00002076, 0.00008753, 0.00015698, 0.00020790, 0.00020890, 0.00014708, 0.00004996, -0.00002705, -0.00004285, -0.00000486, 0.00003875, 0.00004325, 0.00000621, -0.00003587, -0.00004757, -0.00002913, -0.00001086, -0.00001597, -0.00003390, -0.00003397, -0.00000415, 0.00002978, 0.00002925, -0.00001232, -0.00005485, -0.00004841, 0.00001241, 0.00007434, 0.00007236, -0.00000411, -0.00009252, -0.00011014, -0.00003218, 0.00008001, 0.00013098, 0.00007695, -0.00003448, -0.00010825, -0.00008758, -0.00000165, 0.00006681, 0.00005969, -0.00000718, -0.00006647, -0.00006634, -0.00001566, 0.00003190, 0.00003306, -0.00000688, -0.00004630, -0.00005095, -0.00002312, 0.00000882, 0.00002274, 0.00002056, 0.00001638, 0.00001375, 0.00000030, },
		{ -0.00015675, -0.00012925, -0.00005553, 0.00003953, 0.00013191, 0.00020490, 0.00024315, 0.00023146, 0.00016748, 0.00007534, -0.00000135, -0.00002904, -0.00001135, 0.00001449, 0.00001194, -0.00002175, -0.00005587, -0.00005953, -0.00003322, -0.00000624, -0.00000396, -0.00002088, -0.00002915, -0.00001269, 0.00001210, 0.00001420, -0.00001383, -0.00004123, -0.00002892, 0.00002417, 0.00007012, 0.00005509, -0.00002209, -0.00009743, -0.00009602, -0.00000577, 0.00010220, 0.00013253, 0.00005316, -0.00007142, -0.00013561, -0.00008747, 0.00002518, 0.00010021, 0.00007476, -0.00002061, -0.00009555, -0.00008589, -0.00000963, 0.00005669, 0.00005322, -0.00001036, -0.00007095, -0.00007536, -0.00002656, 0.00002769, 0.00004529, 0.00002597, 0.00000077, -0.00000736, -0.00000761, -0.00002678, -0.00007536, -0.00013125, },
		{ 0.00004884, 0.00018145, 0.00027123, 0.00030151, 0.00027158, 0.00019347, 0.00009376, 0.00000873, -0.00003393, -0.00003439, -0.00002019, -0.00002150, -0.00004381, -0.00006504, -0.00006066, -0.00003097, -0.00000171, 0.00000291, -0.00001462, -0.00002915, -0.00002195, -0.00000196, 0.00000554, -0.00000916, -0.00002498, -0.00001242, 0.00002841, 0.00005831, 0.00003567, -0.00003419, -0.00009150, -0.00007357, 0.00001965, 0.00011316, 0.00011909, 0.00002031, -0.00010535, -0.00014975, -0.00007253, 0.00005912, 0.00012923, 0.00007793, -0.00004505, -0.00012779, -0.00009972, 0.00000609, 0.00008810, 0.00007259, -0.00002081, -0.00010315, -0.00010189, -0.00002467, 0.00005566, 0.00007372, 0.00002879, -0.00002516, -0.00003971, -0.00001934, -0.00001140, -0.00005347, -0.00013228, -0.00019208, -0.00018069, -0.00008873, },
		{ 0.00038831, 0.00032664, 0.00021518, 0.00009249, -0.00000673, -0.00006059, -0.00006972, -0.00005654, -0.00004813, -0.00005446, -0.00006266, -0.00005411, -0.00002724, -0.00000144, 0.00000275, -0.00001413, -0.00003075, -0.00002813, -0.00001001, 0.00000286, -0.00000140, -0.00001020, -0.00000152, 0.00002573, 0.00004254, 0.00001858, -0.00003805, -0.00007636, -0.00004782, 0.00003848, 0.00011032, 0.00009307, -0.00001522, -0.00012949, -0.00014756, -0.00004530, 0.00009350, 0.00014727, 0.00006676, -0.00007738, -0.00015690, -0.00010273, 0.00003199, 0.00012197, 0.00008571, -0.00004089, -0.00014080, -0.00012532, -0.00001308, 0.00009295, 0.00010381, 0.00002328, -0.00006463, -0.00008159, -0.00002916, 0.00001727, -0.00001538, -0.00012584, -0.00023545, -0.00025205, -0.00014125, 0.00005299, 0.00024565, 0.00036606, },
		{ 0.00005911, -0.00005335, -0.00010605, -0.00010804, -0.00008383, -0.00006057, -0.00005201, -0.00005201, -0.00004543, -0.00002681, -0.00000757, -0.00000354, -0.00001643, -0.00003046, -0.00002883, -0.00001191, 0.00000387, 0.00000582, 0.00000012, 0.00000330, 0.00001868, 0.00002700, 0.00000698, -0.00003387, -0.00005593, -0.00002433, 0.00004695, 0.00009436, 0.00005969, -0.00004643, -0.00013919, -0.00012966, -0.00001143, 0.00012102, 0.00014989, 0.00004244, -0.00011174, -0.00017631, -0.00009212, 0.00006539, 0.00015227, 0.00008772, -0.00007067, -0.00017912, -0.00013964, 0.00001089, 0.00013669, 0.00012904, 0.00000394, -0.00011799, -0.00012762, -0.00002943, 0.00006396, 0.00003734, -0.00011927, -0.00029557, -0.00034790, -0.00021013, 0.00006165, 0.00033629, 0.00049716, 0.00050294, 0.00038739, 0.00021869, },
		{ -0.00013694, -0.00009230, -0.00005537, -0.00003938, -0.00003951, -0.00004064, -0.00003256, -0.00001925, -0.00001268, -0.00001782, -0.00002588, -0.00002384, -0.00000958, 0.00000545, 0.00000965, 0.00000495, 0.00000357, 0.00001072, 0.00001510, 0.00000189, -0.00002416, -0.00003513, -0.00000756, 0.00004416, 0.00006903, 0.00002559, -0.00006759, -0.00013318, -0.00010021, 0.00002182, 0.00013574, 0.00013613, 0.00000955, -0.00014101, -0.00018111, -0.00006837, 0.00010086, 0.00017263, 0.00007596, -0.00010707, -0.00021151, -0.00013965, 0.00004691, 0.00018093, 0.00014198, -0.00003260, -0.00018170, -0.00016940, -0.00001258, 0.00012992, 0.00009814, -0.00012309, -0.00038023, -0.00046788, -0.00028610, 0.00008854, 0.00046431, 0.00066672, 0.00063725, 0.00043882, 0.00018889, -0.00001502, -0.00013010, -0.00016105, },
		{ -0.00002357, -0.00003277, -0.00004410, -0.00004422, -0.00003283, -0.00002025, -0.00001546, -0.00001677, -0.00001517, -0.00000603, 0.00000497, 0.00000898, 0.00000526, 0.00000199, 0.00000495, 0.00000843, 0.00000204, -0.00001281, -0.00001846, 0.00000035, 0.00003215, 0.00003999, -0.00000302, -0.00007574, -0.00011381, -0.00006577, 0.00004760, 0.00013471, 0.00010882, -0.00002509, -0.00015881, -0.00016947, -0.00003536, 0.00013158, 0.00017812, 0.00005103, -0.00014431, -0.00023120, -0.00012277, 0.00009104, 0.00021767, 0.00013632, -0.00008597, -0.00024826, -0.00019712, 0.00002653, 0.00021135, 0.00015600, -0.00014864, -0.00049334, -0.00060492, -0.00035416, 0.00014861, 0.00063627, 0.00086897, 0.00077560, 0.00046277, 0.00011335, -0.00013163, -0.00022837, -0.00021169, -0.00014395, -0.00007601, -0.00003435, },
		{ -0.00005816, -0.00004299, -0.00002239, -0.00000863, -0.00000528, -0.00000622, -0.00000411, 0.00000134, 0.00000474, 0.00000339, 0.00000124, 0.00000299, 0.00000648, 0.00000462, -0.00000371, -0.00000859, -0.00000001, 0.00001518, 0.00001316, -0.00002201, -0.00007133, -0.00008632, -0.00003342, 0.00006123, 0.00011869, 0.00007375, -0.00005431, -0.00016117, -0.00014323, 0.00000059, 0.00015115, 0.00016680, 0.00001690, -0.00017521, -0.00023307, -0.00009005, 0.00013628, 0.00023867, 0.00010881, -0.00015135, -0.00030736, -0.00020199, 0.00008844, 0.00029905, 0.00019726, -0.00020501, -0.00063247, -0.00074498, -0.00039591, 0.00025418, 0.00085107, 0.00108930, 0.00089695, 0.00044190, -0.00001397, -0.00028384, -0.00033303, -0.00024240, -0.00011991, -0.00003493, -0.00000459, -0.00001377, -0.00003814, -0.00005680, },
		{ 0.00000071, 0.00000467, -0.00000045, -0.00000517, -0.00000446, -0.00000064, 0.00000191, 0.00000276, 0.00000448, 0.00000707, 0.00000647, 0.00000058, -0.00000578, -0.00000577, -0.00000172, -0.00000690, -0.00003022, -0.00005786, -0.00005720, -0.00000892, 0.00006141, 0.00009182, 0.00003809, -0.00007247, -0.00014766, -0.00010742, 0.00003241, 0.00015537, 0.00014034, -0.00001991, -0.00019306, -0.00021529, -0.00004640, 0.00017396, 0.00023768, 0.00006063, -0.00021990, -0.00034807, -0.00017878, 0.00016796, 0.00037956, 0.00020859, -0.00029594, -0.00078760, -0.00086864, -0.00039333, 0.00041112, 0.00109819, 0.00130561, 0.00097867, 0.00036394, -0.00019000, -0.00045639, -0.00042523, -0.00023978, -0.00006180, 0.00002451, 0.00002325, -0.00001871, -0.00005851, -0.00007571, -0.00006832, -0.00004452, -0.00001736, },
		{ -0.00000845, -0.00000974, -0.00000430, 0.00000235, 0.00000616, 0.00000738, 0.00000744, 0.00000538, -0.00000039, -0.00000803, -0.00001296, -0.00001434, -0.00001818, -0.00002949, -0.00004058, -0.00003241, 0.00000475, 0.00005026, 0.00006039, 0.00000846, -0.00007691, -0.00012141, -0.00006890, 0.00005417, 0.00014339, 0.00010374, -0.00005203, -0.00019350, -0.00018006, 0.00000038, 0.00019591, 0.00021225, -0.00000227, -0.00028050, -0.00036150, -0.00012775, 0.00025439, 0.00043769, 0.00018040, -0.00041768, -0.00094173, -0.00095444, -0.00033303, 0.00061598, 0.00135784, 0.00149159, 0.00100122, 0.00022539, -0.00040184, -0.00062794, -0.00048637, -0.00019611, 0.00002590, 0.00009054, 0.00003762, -0.00004398, -0.00009211, -0.00009435, -0.00006808, -0.00003425, -0.00000646, 0.00000835, 0.00000879, 0.00000005, },
		{ 0.00000444, 0.00000928, 0.00000888, 0.00000553, 0.00000102, -0.00000492, -0.00001208, -0.00001799, -0.00002056, -0.00002146, -0.00002372, -0.00002484, -0.00001577, 0.00000785, 0.00003250, 0.00003100, -0.00001079, -0.00006846, -0.00008812, -0.00003452, 0.00006262, 0.00011800, 0.00006399, -0.00007321, -0.00017578, -0.00013329, 0.00004145, 0.00019663, 0.00016479, -0.00007019, -0.00032221, -0.00034324, -0.00005529, 0.00033328, 0.00045991, 0.00010978, -0.00055841, -0.00107339, -0.00098317, -0.00021004, 0.00085494, 0.00160351, 0.00162145, 0.00095266, 0.00003347, -0.00062822, -0.00077504, -0.00050239, -0.00011173, 0.00013161, 0.00014870, 0.00002926, -0.00009005, -0.00013252, -0.00010454, -0.00005239, -0.00001241, 0.00000604, 0.00000913, 0.00000402, -0.00000445, -0.00001110, -0.00001112, -0.00000424, },
		{ 0.00000101, -0.00000490, -0.00000992, -0.00001474, -0.00001895, -0.00002076, -0.00001965, -0.00001732, -0.00001444, -0.00000819, 0.00000342, 0.00001381, 0.00000877, -0.00001843, -0.00005096, -0.00005442, -0.00000932, 0.00005793, 0.00008475, 0.00002827, -0.00007996, -0.00014314, -0.00008325, 0.00006910, 0.00017484, 0.00010232, -0.00013181, -0.00033697, -0.00029490, 0.00002690, 0.00038952, 0.00043774, 0.00000212, -0.00069975, -0.00116059, -0.00094228, -0.00003010, 0.00110513, 0.00180626, 0.00167507, 0.00083186, -0.00019425, -0.00084300, -0.00087680, -0.00046707, 0.00000447, 0.00023896, 0.00018532, -0.00000611, -0.00015173, -0.00017019, -0.00009880, -0.00001968, 0.00001706, 0.00001406, -0.00000131, -0.00001019, -0.00001113, -0.00000917, -0.00000592, -0.00000039, 0.00000596, 0.00000905, 0.00000674, },
		{ -0.00001427, -0.00001624, -0.00001746, -0.00001650, -0.00001370, -0.00001069, -0.00000783, -0.00000404, -0.00000082, -0.00000383, -0.00001635, -0.00002990, -0.00002608, 0.00000459, 0.00004344, 0.00005042, 0.00000205, -0.00007242, -0.00010195, -0.00003850, 0.00007877, 0.00013401, 0.00003495, -0.00017679, -0.00032181, -0.00022418, 0.00010423, 0.00041077, 0.00037023, -0.00012917, -0.00081992, -0.00118544, -0.00082925, 0.00019036, 0.00133816, 0.00194005, 0.00164226, 0.00064949, -0.00043251, -0.00102005, -0.00091925, -0.00038378, 0.00013656, 0.00033029, 0.00019089, -0.00006640, -0.00021907, -0.00019486, -0.00007271, 0.00002726, 0.00004727, 0.00001149, -0.00002489, -0.00003133, -0.00001537, 0.00000098, 0.00000656, 0.00000508, 0.00000342, 0.00000274, 0.00000037, -0.00000439, -0.00000921, -0.00001229, },
		{ -0.00001506, -0.00001412, -0.00001232, -0.00001111, -0.00001019, -0.00000856, -0.00000727, -0.00000855, -0.00001071, -0.00000661, 0.00000824, 0.00002445, 0.00002117, -0.00001219, -0.00005423, -0.00006009, -0.00000581, 0.00007018, 0.00008162, -0.00002641, -0.00019834, -0.00027971, -0.00014331, 0.00016241, 0.00039059, 0.00026490, -0.00026488, -0.00089818, -0.00113825, -0.00065292, 0.00042649, 0.00152518, 0.00198680, 0.00152518, 0.00042649, -0.00065292, -0.00113825, -0.00089818, -0.00026488, 0.00026490, 0.00039059, 0.00016241, -0.00014331, -0.00027971, -0.00019834, -0.00002641, 0.00008162, 0.00007018, -0.00000581, -0.00006009, -0.00005423, -0.00001219, 0.00002117, 0.00002445, 0.00000824, -0.00000661, -0.00001071, -0.00000855, -0.00000727, -0.00000856, -0.00001019, -0.00001111, -0.00001232, -0.00001412, },
		{ -0.00001427, -0.00001229, -0.00000921, -0.00000439, 0.00000037, 0.00000274, 0.00000342, 0.00000508, 0.00000656, 0.00000098, -0.00001537, -0.00003133, -0.00002489, 0.00001149, 0.00004727, 0.00002726, -0.00007271, -0.00019486, -0.00021907, -0.00006640, 0.00019089, 0.00033029, 0.00013656, -0.00038378, -0.00091925, -0.00102005, -0.00043251, 0.00064949, 0.00164226, 0.00194005, 0.00133816, 0.00019036, -0.00082925, -0.00118544, -0.00081992, -0.00012917, 0.00037023, 0.00041077, 0.00010423, -0.00022418, -0.00032181, -0.00017679, 0.00003495, 0.00013401, 0.00007877, -0.00003850, -0.00010195, -0.00007242, 0.00000205, 0.00005042, 0.00004344, 0.00000459, -0.00002608, -0.00002990, -0.00001635, -0.00000383, -0.00000082, -0.00000404, -0.00000783, -0.00001069, -0.00001370, -0.00001650, -0.00001746, -0.00001624, },
		{ 0.00000101, 0.00000674, 0.00000905, 0.00000596, -0.00000039, -0.00000592, -0.00000917, -0.00001113, -0.00001019, -0.00000131, 0.00001406, 0.00001706, -0.00001968, -0.00009880, -0.00017019, -0.00015173, -0.00000611, 0.00018532, 0.00023896, 0.00000447, -0.00046707, -0.00087680, -0.00084300, -0.00019425, 0.00083186, 0.00167507, 0.00180626, 0.00110513, -0.00003010, -0.00094228, -0.00116059, -0.00069975, 0.00000212, 0.00043774, 0.00038952, 0.00002690, -0.00029490, -0.00033697, -0.00013181, 0.00010232, 0.00017484, 0.00006910, -0.00008325, -0.00014314, -0.00007996, 0.00002827, 0.00008475, 0.00005793, -0.00000932, -0.00005442, -0.00005096, -0.00001843, 0.00000877, 0.00001381, 0.00000342, -0.00000819, -0.00001444, -0.00001732, -0.00001965, -0.00002076, -0.00001895, -0.00001474, -0.00000992, -0.00000490, },
		{ 0.00000444, -0.00000424, -0.00001112, -0.00001110, -0.00000445, 0.00000402, 0.00000913, 0.00000604, -0.00001241, -0.00005239, -0.00010454, -0.00013252, -0.00009005, 0.00002926, 0.00014870, 0.00013161, -0.00011173, -0.00050239, -0.00077504, -0.00062822, 0.00003347, 0.00095266, 0.00162145, 0.00160351, 0.00085494, -0.00021004, -0.00098317, -0.00107339, -0.00055841, 0.00010978, 0.00045991, 0.00033328, -0.00005529, -0.00034324, -0.00032221, -0.00007019, 0.00016479, 0.00019663, 0.00004145, -0.00013329, -0.00017578, -0.00007321, 0.00006399, 0.00011800, 0.00006262, -0.00003452, -0.00008812, -0.00006846, -0.00001079, 0.00003100, 0.00003250, 0.00000785, -0.00001577, -0.00002484, -0.00002372, -0.00002146, -0.00002056, -0.00001799, -0.00001208, -0.00000492, 0.00000102, 0.00000553, 0.00000888, 0.00000928, },
		{ -0.00000845, 0.00000005, 0.00000879, 0.00000835, -0.00000646, -0.00003425, -0.00006808, -0.00009435, -0.00009211, -0.00004398, 0.00003762, 0.00009054, 0.00002590, -0.00019611, -0.00048637, -0.00062794, -0.00040184, 0.00022539, 0.00100122, 0.00149159, 0.00135784, 0.00061598, -0.00033303, -0.00095444, -0.00094173, -0.00041768, 0.00018040, 0.00043769, 0.00025439, -0.00012775, -0.00036150, -0.00028050, -0.00000227, 0.00021225, 0.00019591, 0.00000038, -0.00018006, -0.00019350, -0.00005203, 0.00010374, 0.00014339, 0.00005417, -0.00006890, -0.00012141, -0.00007691, 0.00000846, 0.00006039, 0.00005026, 0.00000475, -0.00003241, -0.00004058, -0.00002949, -0.00001818, -0.00001434, -0.00001296, -0.00000803, -0.00000039, 0.00000538, 0.00000744, 0.00000738, 0.00000616, 0.00000235, -0.00000430, -0.00000974, },
		{ 0.00000071, -0.00001736, -0.00004452, -0.00006832, -0.00007571, -0.00005851, -0.00001871, 0.00002325, 0.00002451, -0.00006180, -0.00023978, -0.00042523, -0.00045639, -0.00019000, 0.00036394, 0.00097867, 0.00130561, 0.00109819, 0.00041112, -0.00039333, -0.00086864, -0.00078760, -0.00029594, 0.00020859, 0.00037956, 0.00016796, -0.00017878, -0.00034807, -0.00021990, 0.00006063, 0.00023768, 0.00017396, -0.00004640, -0.00021529, -0.00019306, -0.00001991, 0.00014034, 0.00015537, 0.00003241, -0.00010742, -0.00014766, -0.00007247, 0.00003809, 0.00009182, 0.00006141, -0.00000892, -0.00005720, -0.00005786, -0.00003022, -0.00000690, -0.00000172, -0.00000577, -0.00000578, 0.00000058, 0.00000647, 0.00000707, 0.00000448, 0.00000276, 0.00000191, -0.00000064, -0.00000446, -0.00000517, -0.00000045, 0.00000467, },
		{ -0.00005816, -0.00005680, -0.00003814, -0.00001377, -0.00000459, -0.00003493, -0.00011991, -0.00024240, -0.00033303, -0.00028384, -0.00001397, 0.00044190, 0.00089695, 0.00108930, 0.00085107, 0.00025418, -0.00039591, -0.00074498, -0.00063247, -0.00020501, 0.00019726, 0.00029905, 0.00008844, -0.00020199, -0.00030736, -0.00015135, 0.00010881, 0.00023867, 0.00013628, -0.00009005, -0.00023307, -0.00017521, 0.00001690, 0.00016680, 0.00015115, 0.00000059, -0.00014323, -0.00016117, -0.00005431, 0.00007375, 0.00011869, 0.00006123, -0.00003342, -0.00008632, -0.00007133, -0.00002201, 0.00001316, 0.00001518, -0.00000001, -0.00000859, -0.00000371, 0.00000462, 0.00000648, 0.00000299, 0.00000124, 0.00000339, 0.00000474, 0.00000134, -0.00000411, -0.00000622, -0.00000528, -0.00000863, -0.00002239, -0.00004299, },
		{ -0.00002357, -0.00003435, -0.00007601, -0.00014395, -0.00021169, -0.00022837, -0.00013163, 0.00011335, 0.00046277, 0.00077560, 0.00086897, 0.00063627, 0.00014861, -0.00035416, -0.00060492, -0.00049334, -0.00014864, 0.00015600, 0.00021135, 0.00002653, -0.00019712, -0.00024826, -0.00008597, 0.00013632, 0.00021767, 0.00009104, -0.00012277, -0.00023120, -0.00014431, 0.00005103, 0.00017812, 0.00013158, -0.00003536, -0.00016947, -0.00015881, -0.00002509, 0.00010882, 0.00013471, 0.00004760, -0.00006577, -0.00011381, -0.00007574, -0.00000302, 0.00003999, 0.00003215, 0.00000035, -0.00001846, -0.00001281, 0.00000204, 0.00000843, 0.00000495, 0.00000199, 0.00000526, 0.00000898, 0.00000497, -0.00000603, -0.00001517, -0.00001677, -0.00001546, -0.00002025, -0.00003283, -0.00004422, -0.00004410, -0.00003277, },
		{ -0.00013694, -0.00016105, -0.00013010, -0.00001502, 0.00018889, 0.00043882, 0.00063725, 0.00066672, 0.00046431, 0.00008854, -0.00028610, -0.00046788, -0.00038023, -0.00012309, 0.00009814, 0.00012992, -0.00001258, -0.00016940, -0.00018170, -0.00003260, 0.00014198, 0.00018093, 0.00004691, -0.00013965, -0.00021151, -0.00010707, 0.00007596, 0.00017263, 0.00010086, -0.00006837, -0.00018111, -0.00014101, 0.00000955, 0.00013613, 0.00013574, 0.00002182, -0.00010021, -0.00013318, -0.00006759, 0.00002559, 0.00006903, 0.00004416, -0.00000756, -0.00003513, -0.00002416, 0.00000189, 0.00001510, 0.00001072, 0.00000357, 0.00000495, 0.00000965, 0.00000545, -0.00000958, -0.00002384, -0.00002588, -0.00001782, -0.00001268, -0.00001925, -0.00003256, -0.00004064, -0.00003951, -0.00003938, -0.00005537, -0.00009230, },
		{ 0.00005911, 0.00021869, 0.00038739, 0.00050294, 0.00049716, 0.00033629, 0.00006165, -0.00021013, -0.00034790, -0.00029557, -0.00011927, 0.00003734, 0.00006396, -0.00002943, -0.00012762, -0.00011799, 0.00000394, 0.00012904, 0.00013669, 0.00001089, -0.00013964, -0.00017912, -0.00007067, 0.00008772, 0.00015227, 0.00006539, -0.00009212, -0.00017631, -0.00011174, 0.00004244, 0.00014989, 0.00012102, -0.00001143, -0.00012966, -0.00013919, -0.00004643, 0.00005969, 0.00009436, 0.00004695, -0.00002433, -0.00005593, -0.00003387, 0.00000698, 0.00002700, 0.00001868, 0.00000330, 0.00000012, 0.00000582, 0.00000387, -0.00001191, -0.00002883, -0.00003046, -0.00001643, -0.00000354, -0.00000757, -0.00002681, -0.00004543, -0.00005201, -0.00005201, -0.00006057, -0.00008383, -0.00010804, -0.00010605, -0.00005335, },
		{ 0.00038831, 0.00036606, 0.00024565, 0.00005299, -0.00014125, -0.00025205, -0.00023545, -0.00012584, -0.00001538, 0.00001727, -0.00002916, -0.00008159, -0.00006463, 0.00002328, 0.00010381, 0.00009295, -0.00001308, -0.00012532, -0.00014080, -0.00004089, 0.00008571, 0.00012197, 0.00003199, -0.00010273, -0.00015690, -0.00007738, 0.00006676, 0.00014727, 0.00009350, -0.00004530, -0.00014756, -0.00012949, -0.00001522, 0.00009307, 0.00011032, 0.00003848, -0.00004782, -0.00007636, -0.00003805, 0.00001858, 0.00004254, 0.00002573, -0.00000152, -0.00001020, -0.00000140, 0.00000286, -0.00001001, -0.00002813, -0.00003075, -0.00001413, 0.00000275, -0.00000144, -0.00002724, -0.00005411, -0.00006266, -0.00005446, -0.00004813, -0.00005654, -0.00006972, -0.00006059, -0.00000673, 0.00009249, 0.00021518, 0.00032664, },
		{ 0.00004884, -0.00008873, -0.00018069, -0.00019208, -0.00013228, -0.00005347, -0.00001140, -0.00001934, -0.00003971, -0.00002516, 0.00002879, 0.00007372, 0.00005566, -0.00002467, -0.00010189, -0.00010315, -0.00002081, 0.00007259, 0.00008810, 0.00000609, -0.00009972, -0.00012779, -0.00004505, 0.00007793, 0.00012923, 0.00005912, -0.00007253, -0.00014975, -0.00010535, 0.00002031, 0.00011909, 0.00011316, 0.00001965, -0.00007357, -0.00009150, -0.00003419, 0.00003567, 0.00005831, 0.00002841, -0.00001242, -0.00002498, -0.00000916, 0.00000554, -0.00000196, -0.00002195, -0.00002915, -0.00001462, 0.00000291, -0.00000171, -0.00003097, -0.00006066, -0.00006504, -0.00004381, -0.00002150, -0.00002019, -0.00003439, -0.00003393, 0.00000873, 0.00009376, 0.00019347, 0.00027158, 0.00030151, 0.00027123, 0.00018145, },
		{ -0.00015675, -0.00013125, -0.00007536, -0.00002678, -0.00000761, -0.00000736, 0.00000077, 0.00002597, 0.00004529, 0.00002769, -0.00002656, -0.00007536, -0.00007095, -0.00001036, 0.00005322, 0.00005669, -0.00000963, -0.00008589, -0.00009555, -0.00002061, 0.00007476, 0.00010021, 0.00002518, -0.00008747, -0.00013561, -0.00007142, 0.00005316, 0.00013253, 0.00010220, -0.00000577, -0.00009602, -0.00009743, -0.00002209, 0.00005509, 0.00007012, 0.00002417, -0.00002892, -0.00004123, -0.00001383, 0.00001420, 0.00001210, -0.00001269, -0.00002915, -0.00002088, -0.00000396, -0.00000624, -0.00003322, -0.00005953, -0.00005587, -0.00002175, 0.00001194, 0.00001449, -0.00001135, -0.00002904, -0.00000135, 0.00007534, 0.00016748, 0.00023146, 0.00024315, 0.00020490, 0.00013191, 0.00003953, -0.00005553, -0.00012925, },
		{ -0.00003499, 0.00000030, 0.00001375, 0.00001638, 0.00002056, 0.00002274, 0.00000882, -0.00002312, -0.00005095, -0.00004630, -0.00000688, 0.00003306, 0.00003190, -0.00001566, -0.00006634, -0.00006647, -0.00000718, 0.00005969, 0.00006681, -0.00000165, -0.00008758, -0.00010825, -0.00003448, 0.00007695, 0.00013098, 0.00008001, -0.00003218, -0.00011014, -0.00009252, -0.00000411, 0.00007236, 0.00007434, 0.00001241, -0.00004841, -0.00005485, -0.00001232, 0.00002925, 0.00002978, -0.00000415, -0.00003397, -0.00003390, -0.00001597, -0.00001086, -0.00002913, -0.00004757, -0.00003587, 0.00000621, 0.00004325, 0.00003875, -0.00000486, -0.00004285, -0.00002705, 0.00004996, 0.00014708, 0.00020890, 0.00020790, 0.00015698, 0.00008753, 0.00002076, -0.00003921, -0.00009075, -0.00012236, -0.00011969, -0.00008349, },
		{ 0.00002609, 0.00001696, 0.00000747, -0.00000342, -0.00001882, -0.00003179, -0.00002861, -0.00000640, 0.00001657, 0.00001511, -0.00001507, -0.00004678, -0.00004494, -0.00000446, 0.00003791, 0.00003597, -0.00001713, -0.00007407, -0.00007362, -0.00000160, 0.00008764, 0.00011552, 0.00005175, -0.00005377, -0.00011284, -0.00007793, 0.00001550, 0.00008305, 0.00006912, -0.00000511, -0.00006674, -0.00006190, -0.00000322, 0.00004879, 0.00004708, -0.00000014, -0.00004514, -0.00005206, -0.00002906, -0.00001087, -0.00001563, -0.00002512, -0.00000944, 0.00003264, 0.00006463, 0.00004786, -0.00001373, -0.00006786, -0.00005724, 0.00002761, 0.00013661, 0.00020083, 0.00018721, 0.00011815, 0.00004292, -0.00000669, -0.00003409, -0.00005842, -0.00008503, -0.00009854, -0.00008265, -0.00004147, 0.00000161, 0.00002502, },
		{ -0.00001236, -0.00001687, -0.00001863, -0.00001551, -0.00000512, 0.00000603, 0.00000505, -0.00001227, -0.00003185, -0.00003236, -0.00000916, 0.00001568, 0.00001300, -0.00002046, -0.00005139, -0.00003882, 0.00002174, 0.00008452, 0.00009032, 0.00002358, -0.00006611, -0.00010388, -0.00005728, 0.00003187, 0.00008415, 0.00005508, -0.00002468, -0.00007944, -0.00005989, 0.00001253, 0.00006905, 0.00005961, -0.00000314, -0.00006177, -0.00007151, -0.00003833, -0.00000285, 0.00000734, 0.00000405, 0.00001688, 0.00005055, 0.00007148, 0.00004195, -0.00003247, -0.00009483, -0.00008254, 0.00001373, 0.00013514, 0.00020058, 0.00017230, 0.00008208, -0.00000277, -0.00003843, -0.00003369, -0.00002762, -0.00004445, -0.00007154, -0.00007798, -0.00004961, -0.00000350, 0.00002971, 0.00003389, 0.00001727, -0.00000157, },
		{ -0.00000410, -0.00000067, 0.00000128, -0.00000129, -0.00001134, -0.00002379, -0.00002714, -0.00001628, -0.00000150, 0.00000027, -0.00001427, -0.00002587, -0.00001014, 0.00003307, 0.00007028, 0.00006137, 0.00000103, -0.00006742, -0.00008588, -0.00003554, 0.00004098, 0.00007506, 0.00003522, -0.00004167, -0.00008313, -0.00004865, 0.00003181, 0.00008544, 0.00006396, -0.00001336, -0.00008065, -0.00008726, -0.00003970, 0.00001405, 0.00003663, 0.00003384, 0.00003644, 0.00005548, 0.00006349, 0.00002537, -0.00005335, -0.00011545, -0.00009733, 0.00000889, 0.00013799, 0.00020060, 0.00015653, 0.00004635, -0.00004687, -0.00006869, -0.00003274, 0.00000335, -0.00000314, -0.00004344, -0.00007337, -0.00006023, -0.00001270, 0.00003143, 0.00004224, 0.00002132, -0.00000681, -0.00002086, -0.00001849, -0.00001016, },
		{ -0.00000729, -0.00001467, -0.00002207, -0.00002556, -0.00002085, -0.00001047, -0.00000322, -0.00000356, -0.00000380, 0.00000857, 0.00003331, 0.00005004, 0.00003475, -0.00001256, -0.00005887, -0.00006343, -0.00001752, 0.00004089, 0.00005776, 0.00001403, -0.00005196, -0.00007653, -0.00003046, 0.00005018, 0.00009395, 0.00005889, -0.00002849, -0.00009708, -0.00009475, -0.00003136, 0.00003764, 0.00006669, 0.00005766, 0.00004455, 0.00004670, 0.00004441, 0.00000496, -0.00006900, -0.00012452, -0.00010049, 0.00000983, 0.00013908, 0.00019480, 0.00013672, 0.00001204, -0.00008491, -0.00009235, -0.00002849, 0.00003344, 0.00003501, -0.00001836, -0.00007067, -0.00007204, -0.00002284, 0.00003287, 0.00005161, 0.00002732, -0.00001080, -0.00003033, -0.00002297, -0.00000409, 0.00000779, 0.00000724, 0.00000022, },
		{ -0.00002336, -0.00001958, -0.00001120, -0.00000115, 0.00000597, 0.00001031, 0.00001672, 0.00002625, 0.00002970, 0.00001496, -0.00001674, -0.00004412, -0.00004167, -0.00000647, 0.00003227, 0.00003624, -0.00000384, -0.00005317, -0.00006094, -0.00000939, 0.00006325, 0.00009243, 0.00004571, -0.00004445, -0.00010639, -0.00009134, -0.00001439, 0.00006321, 0.00009115, 0.00007041, 0.00003977, 0.00002731, 0.00002060, -0.00001223, -0.00007442, -0.00012086, -0.00009476, 0.00001139, 0.00013317, 0.00018011, 0.00011311, -0.00001765, -0.00011258, -0.00010639, -0.00002081, 0.00005997, 0.00006628, 0.00000126, -0.00006964, -0.00008265, -0.00003122, 0.00003511, 0.00006087, 0.00003296, -0.00001522, -0.00004056, -0.00002841, 0.00000167, 0.00002076, 0.00001705, -0.00000005, -0.00001527, -0.00002226, -0.00002389, },
		{ 0.00000152, 0.00000979, 0.00001492, 0.00001679, 0.00001709, 0.00001426, 0.00000382, -0.00001383, -0.00002799, -0.00002478, -0.00000319, 0.00001811, 0.00001527, -0.00001498, -0.00004540, -0.00003980, 0.00000992, 0.00006799, 0.00008122, 0.00002789, -0.00005660, -0.00010539, -0.00007718, 0.00000747, 0.00008491, 0.00010469, 0.00006987, 0.00002429, 0.00000312, -0.00000092, -0.00002096, -0.00006820, -0.00010708, -0.00008512, 0.00000864, 0.00011751, 0.00015680, 0.00008845, -0.00003911, -0.00012725, -0.00011038, -0.00001157, 0.00007994, 0.00008830, 0.00001498, -0.00006875, -0.00008972, -0.00003643, 0.00003769, 0.00006811, 0.00003642, -0.00002037, -0.00005007, -0.00003271, 0.00000814, 0.00003378, 0.00002618, -0.00000191, -0.00002565, -0.00003190, -0.00002544, -0.00001726, -0.00001187, -0.00000644, },
		{ 0.00001323, 0.00001077, 0.00000639, 0.00000031, -0.00000786, -0.00001494, -0.00001481, -0.00000602, 0.00000271, -0.00000103, -0.00001814, -0.00003107, -0.00001769, 0.00002365, 0.00006369, 0.00006307, 0.00000998, -0.00006126, -0.00009343, -0.00005521, 0.00002889, 0.00009746, 0.00010457, 0.00005730, 0.00000309, -0.00001912, -0.00001453, -0.00001913, -0.00005249, -0.00008821, -0.00007688, -0.00000145, 0.00009240, 0.00012784, 0.00006645, -0.00004985, -0.00012869, -0.00010613, -0.00000340, 0.00009134, 0.00010060, 0.00002394, -0.00006611, -0.00009193, -0.00003857, 0.00003916, 0.00007166, 0.00003705, -0.00002528, -0.00005698, -0.00003456, 0.00001489, 0.00004473, 0.00003223, -0.00000624, -0.00003681, -0.00004015, -0.00002344, -0.00000625, 0.00000015, -0.00000062, 0.00000033, 0.00000584, 0.00001161, },
		{ 0.00000124, -0.00000309, -0.00000775, -0.00001131, -0.00001154, -0.00000971, -0.00001037, -0.00001449, -0.00001417, 0.00000111, 0.00002992, 0.00005206, 0.00004224, -0.00000385, -0.00005688, -0.00007262, -0.00003026, 0.00004466, 0.00009772, 0.00009132, 0.00003688, -0.00001758, -0.00003366, -0.00001741, -0.00000805, -0.00003198, -0.00006995, -0.00007390, -0.00001916, 0.00006089, 0.00009756, 0.00005025, -0.00004943, -0.00011897, -0.00009668, 0.00000146, 0.00009374, 0.00010444, 0.00003002, -0.00006045, -0.00008935, -0.00003891, 0.00003797, 0.00007072, 0.00003544, -0.00002839, -0.00005990, -0.00003389, 0.00002035, 0.00005137, 0.00003398, -0.00001206, -0.00004577, -0.00004371, -0.00001638, 0.00000813, 0.00001243, 0.00000180, -0.00000714, -0.00000491, 0.00000427, 0.00001079, 0.00001041, 0.00000594, },
		{ -0.00001177, -0.00001589, -0.00001631, -0.00001279, -0.00000706, 0.00000107, 0.00001382, 0.00002924, 0.00003663, 0.00002312, -0.00001110, -0.00004457, -0.00004726, -0.00000764, 0.00005115, 0.00008561, 0.00006849, 0.00001438, -0.00003207, -0.00003724, -0.00001000, 0.00000837, -0.00001225, -0.00005692, -0.00007762, -0.00004222, 0.00002753, 0.00007014, 0.00004136, -0.00003955, -0.00010165, -0.00008519, 0.00000195, 0.00008838, 0.00010213, 0.00003493, -0.00005166, -0.00008329, -0.00003915, 0.00003321, 0.00006573, 0.00003301, -0.00002833, -0.00005853, -0.00003184, 0.00002262, 0.00005240, 0.00003190, -0.00001693, -0.00004941, -0.00004072, -0.00000519, 0.00002232, 0.00002053, -0.00000206, -0.00002053, -0.00001899, -0.00000239, 0.00001295, 0.00001648, 0.00001037, 0.00000261, -0.00000263, -0.00000681, },
		{ -0.00001041, 0.00000039, 0.00001153, 0.00001990, 0.00002409, 0.00002154, 0.00000891, -0.00001166, -0.00002766, -0.00002248, 0.00000849, 0.00004744, 0.00006401, 0.00004153, -0.00000455, -0.00003684, -0.00002973, 0.00000452, 0.00002504, 0.00000187, -0.00005154, -0.00008696, -0.00006665, -0.00000295, 0.00004849, 0.00003940, -0.00002345, -0.00008072, -0.00007401, -0.00000150, 0.00007769, 0.00009617, 0.00003959, -0.00004080, -0.00007575, -0.00004071, 0.00002494, 0.00005806, 0.00003128, -0.00002455, -0.00005376, -0.00003016, 0.00002036, 0.00004804, 0.00002803, -0.00001798, -0.00004572, -0.00003158, 0.00000712, 0.00003202, 0.00002119, -0.00001249, -0.00003690, -0.00003172, -0.00000446, 0.00002081, 0.00002691, 0.00001580, 0.00000107, -0.00000741, -0.00001027, -0.00001273, -0.00001608, -0.00001641, },
		{ 0.00002091, 0.00001797, 0.00001014, 0.00000068, -0.00000773, -0.00001070, -0.00000274, 0.00001628, 0.00003547, 0.00003789, 0.00001615, -0.00001599, -0.00003126, -0.00001379, 0.00002149, 0.00003729, 0.00000756, -0.00005371, -0.00009901, -0.00008809, -0.00002691, 0.00003370, 0.00004235, -0.00000501, -0.00005960, -0.00006422, -0.00000732, 0.00006449, 0.00008855, 0.00004388, -0.00002968, -0.00006870, -0.00004420, 0.00001424, 0.00004945, 0.00003128, -0.00001758, -0.00004729, -0.00003039, 0.00001349, 0.00004008, 0.00002518, -0.00001306, -0.00003472, -0.00001889, 0.00001645, 0.00003373, 0.00001348, -0.00002716, -0.00005173, -0.00003884, -0.00000015, 0.00003286, 0.00003797, 0.00001858, -0.00000487, -0.00001610, -0.00001491, -0.00001112, -0.00001101, -0.00001153, -0.00000649, 0.00000473, 0.00001603, },
		{ -0.00000280, -0.00000293, 0.00000209, 0.00000946, 0.00001656, 0.00001934, 0.00001278, -0.00000327, -0.00001863, -0.00001749, 0.00000617, 0.00003631, 0.00004213, 0.00000443, -0.00006128, -0.00011015, -0.00010300, -0.00004242, 0.00002516, 0.00004739, 0.00001225, -0.00004060, -0.00005572, -0.00001340, 0.00005125, 0.00008035, 0.00004689, -0.00002029, -0.00006355, -0.00004938, 0.00000269, 0.00004145, 0.00003316, -0.00000882, -0.00004088, -0.00003311, 0.00000342, 0.00003129, 0.00002580, -0.00000172, -0.00001863, -0.00000660, 0.00001908, 0.00002603, -0.00000078, -0.00004183, -0.00006053, -0.00003788, 0.00000972, 0.00004569, 0.00004574, 0.00001657, -0.00001448, -0.00002573, -0.00001749, -0.00000540, -0.00000172, -0.00000460, -0.00000413, 0.00000449, 0.00001555, 0.00001975, 0.00001407, 0.00000398, },
		{ 0.00001420, 0.00001245, 0.00000400, -0.00000679, -0.00001470, -0.00001385, 0.00000048, 0.00002550, 0.00004571, 0.00003887, -0.00000574, -0.00007110, -0.00011720, -0.00010941, -0.00004939, 0.00002112, 0.00005176, 0.00002583, -0.00002476, -0.00004770, -0.00001778, 0.00003964, 0.00007180, 0.00004740, -0.00001421, -0.00006087, -0.00005536, -0.00000813, 0.00003490, 0.00003625, 0.00000006, -0.00003571, -0.00003769, -0.00000742, 0.00002444, 0.00003101, 0.00001439, -0.00000128, 0.00000122, 0.00001303, 0.00001019, -0.00001763, -0.00005184, -0.00006032, -0.00002902, 0.00002213, 0.00005524, 0.00004746, 0.00000979, -0.00002515, -0.00003288, -0.00001592, 0.00000394, 0.00000909, 0.00000116, -0.00000514, 0.00000023, 0.00001264, 0.00001943, 0.00001407, 0.00000238, -0.00000441, -0.00000099, 0.00000809, },
		{ -0.00001841, -0.00001873, -0.00000498, 0.00001813, 0.00004058, 0.00004848, 0.00002893, -0.00001992, -0.00008008, -0.00011826, -0.00010712, -0.00004908, 0.00001956, 0.00005346, 0.00003450, -0.00001214, -0.00003919, -0.00001915, 0.00003043, 0.00006262, 0.00004437, -0.00001230, -0.00006055, -0.00006106, -0.00001712, 0.00002992, 0.00003948, 0.00000778, -0.00003191, -0.00004244, -0.00001638, 0.00002127, 0.00004006, 0.00003192, 0.00001307, 0.00000188, -0.00000118, -0.00001014, -0.00003219, -0.00005362, -0.00005061, -0.00001487, 0.00003299, 0.00005827, 0.00004250, 0.00000036, -0.00003348, -0.00003500, -0.00001008, 0.00001457, 0.00001814, 0.00000353, -0.00000941, -0.00000573, 0.00001032, 0.00002185, 0.00001762, 0.00000285, -0.00000749, -0.00000437, 0.00000708, 0.00001387, 0.00000812, -0.00000634, },
		{ 0.00003180, 0.00004966, 0.00004546, 0.00001508, -0.00003489, -0.00008604, -0.00011301, -0.00009736, -0.00004340, 0.00001880, 0.00005155, 0.00003811, -0.00000233, -0.00002953, -0.00001700, 0.00002368, 0.00005242, 0.00003730, -0.00001468, -0.00006207, -0.00006565, -0.00002380, 0.00002615, 0.00004195, 0.00001385, -0.00002862, -0.00004532, -0.00002150, 0.00002194, 0.00005052, 0.00004685, 0.00002107, -0.00000510, -0.00002056, -0.00003003, -0.00004022, -0.00004588, -0.00003355, 0.00000039, 0.00003856, 0.00005349, 0.00003247, -0.00000840, -0.00003656, -0.00003134, -0.00000179, 0.00002327, 0.00002280, 0.00000205, -0.00001499, -0.00001036, 0.00001082, 0.00002701, 0.00002247, 0.00000216, -0.00001448, -0.00001339, 0.00000115, 0.00001216, 0.00000736, -0.00000935, -0.00002225, -0.00001795, 0.00000398, },
		{ 0.00000046, -0.00004809, -0.00008815, -0.00010244, -0.00008217, -0.00003424, 0.00001790, 0.00004610, 0.00003723, 0.00000516, -0.00001866, -0.00001159, 0.00001907, 0.00004101, 0.00002627, -0.00002096, -0.00006490, -0.00006873, -0.00002819, 0.00002320, 0.00004334, 0.00001867, -0.00002455, -0.00004476, -0.00002226, 0.00002492, 0.00005904, 0.00005571, 0.00002139, -0.00001772, -0.00004052, -0.00004469, -0.00003944, -0.00003003, -0.00001324, 0.00001248, 0.00003677, 0.00004199, 0.00002064, -0.00001322, -0.00003307, -0.00002326, 0.00000588, 0.00002718, 0.00002212, -0.00000178, -0.00001892, -0.00001108, 0.00001466, 0.00003304, 0.00002539, -0.00000238, -0.00002596, -0.00002607, -0.00000608, 0.00001271, 0.00001232, -0.00000516, -0.00002055, -0.00001588, 0.00000967, 0.00003953, 0.00005297, 0.00003884, },
		{ -0.00008836, -0.00006373, -0.00002305, 0.00001667, 0.00003784, 0.00003268, 0.00001056, -0.00000699, -0.00000357, 0.00001621, 0.00002856, 0.00001186, -0.00003055, -0.00006868, -0.00007025, -0.00003043, 0.00002104, 0.00004405, 0.00002315, -0.00001863, -0.00004034, -0.00001969, 0.00002764, 0.00006257, 0.00005676, 0.00001505, -0.00003216, -0.00005624, -0.00005099, -0.00003019, -0.00000966, 0.00000562, 0.00001837, 0.00002786, 0.00002678, 0.00001066, -0.00001212, -0.00002384, -0.00001369, 0.00000979, 0.00002483, 0.00001713, -0.00000517, -0.00001842, -0.00000688, 0.00002041, 0.00003681, 0.00002344, -0.00001177, -0.00004025, -0.00003883, -0.00001083, 0.00001776, 0.00002283, 0.00000393, -0.00001666, -0.00001556, 0.00001029, 0.00004131, 0.00005215, 0.00003130, -0.00001237, -0.00005816, -0.00008678, },
		{ 0.00001542, 0.00002778, 0.00002528, 0.00001387, 0.00000474, 0.00000622, 0.00001483, 0.00001548, -0.00000518, -0.00004281, -0.00007319, -0.00007028, -0.00003045, 0.00002014, 0.00004492, 0.00002810, -0.00001071, -0.00003308, -0.00001595, 0.00002752, 0.00005943, 0.00005050, 0.00000508, -0.00004401, -0.00006416, -0.00004824, -0.00001521, 0.00001055, 0.00001918, 0.00001721, 0.00001426, 0.00001179, 0.00000535, -0.00000514, -0.00001156, -0.00000610, 0.00000791, 0.00001682, 0.00001053, -0.00000508, -0.00001203, 0.00000123, 0.00002514, 0.00003521, 0.00001516, -0.00002479, -0.00005393, -0.00004770, -0.00001050, 0.00002739, 0.00003658, 0.00001446, -0.00001367, -0.00001867, 0.00000591, 0.00003861, 0.00004943, 0.00002504, -0.00002192, -0.00006484, -0.00008303, -0.00007270, -0.00004389, -0.00001071, },
		},
		{ { -0.00002711, -0.00002614, -0.00002217, -0.00001319, 0.00000224, 0.00002414, 0.00004741, 0.00005696, 0.00003350, -0.00002425, -0.00008252, -0.00009167, -0.00003648, 0.00003605, 0.00005573, 0.00000400, -0.00006049, -0.00006415, -0.00000247, 0.00005436, 0.00004429, -0.00001394, -0.00004674, -0.00001698, 0.00003357, 0.00004113, 0.00000017, -0.00003493, -0.00002341, 0.00001247, 0.00002389, 0.00000273, -0.00001277, 0.00000273, 0.00002389, 0.00001247, -0.00002341, -0.00003493, 0.00000017, 0.00004113, 0.00003357, -0.00001698, -0.00004674, -0.00001394, 0.00004429, 0.00005436, -0.00000247, -0.00006415, -0.00006049, 0.00000400, 0.00005573, 0.00003605, -0.00003648, -0.00009167, -0.00008252, -0.00002425, 0.00003350, 0.00005696, 0.00004741, 0.00002414, 0.00000224, -0.00001319, -0.00002217, -0.00002614, },
		{ -0.00002236, 0.00000035, 0.00002943, 0.00005455, 0.00006234, 0.00004074, -0.00001091, -0.00007018, -0.00009630, -0.00006233, 0.00000984, 0.00005708, 0.00003373, -0.00003539, -0.00007444, -0.00003830, 0.00003417, 0.00006260, 0.00001856, -0.00004136, -0.00004466, 0.00000963, 0.00005272, 0.00003313, -0.00002152, -0.00004366, -0.00001165, 0.00002840, 0.00002769, -0.00000349, -0.00001722, 0.00000271, 0.00002086, 0.00000465, -0.00002595, -0.00002515, 0.00001206, 0.00003786, 0.00001496, -0.00002952, -0.00003577, 0.00000922, 0.00004991, 0.00003056, -0.00003188, -0.00006507, -0.00002907, 0.00003400, 0.00004916, -0.00000454, -0.00007242, -0.00008750, -0.00004003, 0.00002233, 0.00005115, 0.00004028, 0.00001599, 0.00000081, -0.00000404, -0.00000816, -0.00001657, -0.00002681, -0.00003400, -0.00003367, },
		{ 0.00005959, 0.00006715, 0.00004612, -0.00000094, -0.00005653, -0.00009044, -0.00007623, -0.00001772, 0.00004127, 0.00004835, -0.00000397, -0.00006162, -0.00006053, 0.00000149, 0.00005730, 0.00004511, -0.00001816, -0.00005529, -0.00002122, 0.00004216, 0.00005781, 0.00000833, -0.00004440, -0.00003823, 0.00001335, 0.00004340, 0.00001900, -0.00002101, -0.00002319, 0.00000904, 0.00002492, -0.00000020, -0.00003041, -0.00002046, 0.00001925, 0.00003488, 0.00000307, -0.00003469, -0.00002623, 0.00002008, 0.00004333, 0.00000931, -0.00004286, -0.00004886, -0.00000109, 0.00004041, 0.00002157, -0.00004033, -0.00007819, -0.00005318, 0.00000601, 0.00004258, 0.00003443, 0.00000737, -0.00000493, 0.00000185, 0.00000757, -0.00000287, -0.00002447, -0.00004194, -0.00004456, -0.00003017, -0.00000208, 0.00003196, },
		{ 0.00000686, -0.00004418, -0.00007975, -0.00007892, -0.00003812, 0.00001775, 0.00004550, 0.00001987, -0.00003513, -0.00006164, -0.00002735, 0.00003365, 0.00005367, 0.00000952, -0.00004600, -0.00004409, 0.00001597, 0.00006316, 0.00004058, -0.00002402, -0.00005244, -0.00001330, 0.00004062, 0.00004172, -0.00000690, -0.00003962, -0.00001643, 0.00002635, 0.00002811, -0.00001344, -0.00004019, -0.00001367, 0.00003186, 0.00003503, -0.00000987, -0.00004433, -0.00002290, 0.00002659, 0.00003968, -0.00000048, -0.00004219, -0.00003365, 0.00001026, 0.00002898, -0.00000494, -0.00005165, -0.00005546, -0.00001272, 0.00002813, 0.00002826, 0.00000096, -0.00001292, 0.00000209, 0.00002146, 0.00001520, -0.00001701, -0.00004901, -0.00005716, -0.00003814, -0.00000321, 0.00003414, 0.00006207, 0.00006961, 0.00005002, },
		{ -0.00007465, -0.00004898, -0.00000412, 0.00003095, 0.00002920, -0.00000838, -0.00004528, -0.00004057, 0.00000502, 0.00004266, 0.00002764, -0.00002448, -0.00005004, -0.00001252, 0.00004843, 0.00006073, 0.00000841, -0.00004588, -0.00003718, 0.00002032, 0.00005152, 0.00001664, -0.00003816, -0.00004184, 0.00000830, 0.00004383, 0.00001647, -0.00003737, -0.00004470, 0.00000653, 0.00004993, 0.00002711, -0.00003539, -0.00005935, -0.00001494, 0.00004247, 0.00004401, -0.00000730, -0.00004586, -0.00002957, 0.00001101, 0.00001929, -0.00001280, -0.00003965, -0.00002628, 0.00000739, 0.00001773, -0.00000307, -0.00001933, -0.00000252, 0.00002953, 0.00003443, -0.00000258, -0.00005159, -0.00007147, -0.00004934, -0.00000410, 0.00003848, 0.00006422, 0.00006884, 0.00005084, 0.00001295, -0.00003290, -0.00006771, },
		{ 0.00001337, 0.00002572, 0.00000880, -0.00002233, -0.00003625, -0.00001517, 0.00002058, 0.00002916, -0.00000364, -0.00004021, -0.00003095, 0.00002248, 0.00006139, 0.00003841, -0.00002087, -0.00004535, -0.00000682, 0.00004256, 0.00003591, -0.00002074, -0.00005423, -0.00001955, 0.00003897, 0.00004363, -0.00001386, -0.00005851, -0.00002899, 0.00003972, 0.00005800, -0.00000317, -0.00007201, -0.00006405, 0.00001290, 0.00006919, 0.00004345, -0.00002731, -0.00006015, -0.00002756, 0.00001965, 0.00002544, -0.00000485, -0.00002485, -0.00001468, -0.00000064, -0.00000822, -0.00002166, -0.00000795, 0.00003037, 0.00004960, 0.00001702, -0.00004596, -0.00008404, -0.00006535, -0.00000837, 0.00004491, 0.00006940, 0.00006666, 0.00004696, 0.00001565, -0.00002227, -0.00005473, -0.00006699, -0.00005272, -0.00001967, },
		{ -0.00000370, -0.00002157, -0.00002060, 0.00000013, 0.00001716, 0.00000632, -0.00002392, -0.00003455, -0.00000204, 0.00004531, 0.00005402, 0.00001132, -0.00003329, -0.00002653, 0.00001951, 0.00003997, 0.00000182, -0.00004858, -0.00004293, 0.00001671, 0.00005452, 0.00001802, -0.00004878, -0.00005740, 0.00000912, 0.00006751, 0.00003682, -0.00005497, -0.00009918, -0.00003855, 0.00006088, 0.00008819, 0.00001834, -0.00006425, -0.00007080, -0.00000734, 0.00004705, 0.00004061, -0.00000185, -0.00002816, -0.00002581, -0.00001987, -0.00002131, -0.00000958, 0.00002678, 0.00005665, 0.00003637, -0.00003147, -0.00008959, -0.00008399, -0.00001960, 0.00004987, 0.00007923, 0.00006721, 0.00003927, 0.00001232, -0.00001393, -0.00004014, -0.00005661, -0.00005261, -0.00002959, -0.00000127, 0.00001641, 0.00001444, },
		{ -0.00000961, 0.00000187, 0.00000359, -0.00001210, -0.00002779, -0.00001626, 0.00002264, 0.00005178, 0.00003706, -0.00000685, -0.00002890, -0.00000527, 0.00002685, 0.00001638, -0.00003022, -0.00005120, -0.00001102, 0.00004482, 0.00004241, -0.00002127, -0.00006548, -0.00002694, 0.00005126, 0.00006538, -0.00001681, -0.00010518, -0.00009066, 0.00001982, 0.00010576, 0.00007350, -0.00003567, -0.00009826, -0.00005246, 0.00004085, 0.00008107, 0.00004099, -0.00002244, -0.00005039, -0.00004112, -0.00002364, -0.00000571, 0.00002436, 0.00005571, 0.00004916, -0.00001209, -0.00008442, -0.00009953, -0.00003806, 0.00004810, 0.00009140, 0.00007426, 0.00003198, 0.00000211, -0.00001153, -0.00002500, -0.00004216, -0.00004945, -0.00003693, -0.00001232, 0.00000778, 0.00001344, 0.00000577, -0.00000709, -0.00001449, },
		{ -0.00001020, -0.00002026, -0.00001968, 0.00000402, 0.00003745, 0.00004788, 0.00002162, -0.00001354, -0.00001898, 0.00000440, 0.00001524, -0.00001121, -0.00004311, -0.00003072, 0.00002115, 0.00004852, 0.00000975, -0.00005082, -0.00005141, 0.00001792, 0.00006962, 0.00002399, -0.00008087, -0.00012175, -0.00003799, 0.00008629, 0.00011419, 0.00001620, -0.00009511, -0.00009666, 0.00000720, 0.00010029, 0.00009213, 0.00000768, -0.00006258, -0.00006905, -0.00003482, 0.00000035, 0.00002788, 0.00005151, 0.00005201, 0.00000494, -0.00006949, -0.00010558, -0.00005928, 0.00003615, 0.00010005, 0.00008804, 0.00003077, -0.00001196, -0.00001826, -0.00001331, -0.00002326, -0.00004112, -0.00004300, -0.00002286, 0.00000163, 0.00001253, 0.00000885, 0.00000025, -0.00000612, -0.00000819, -0.00000642, -0.00000496, },
		{ -0.00000587, 0.00002119, 0.00004440, 0.00004068, 0.00001079, -0.00001561, -0.00001455, 0.00000048, -0.00000282, -0.00002686, -0.00003510, -0.00000351, 0.00003601, 0.00002959, -0.00002259, -0.00005423, -0.00001592, 0.00004938, 0.00004876, -0.00003931, -0.00012033, -0.00008884, 0.00003665, 0.00012165, 0.00006971, -0.00006047, -0.00011989, -0.00004071, 0.00008969, 0.00013426, 0.00005756, -0.00005181, -0.00009395, -0.00005703, 0.00000162, 0.00003765, 0.00005072, 0.00004684, 0.00001297, -0.00005090, -0.00009915, -0.00007565, 0.00001544, 0.00009887, 0.00010391, 0.00003924, -0.00002345, -0.00003372, -0.00001045, -0.00000297, -0.00002507, -0.00004544, -0.00003550, -0.00000522, 0.00001490, 0.00001248, 0.00000107, -0.00000404, -0.00000312, -0.00000406, -0.00000899, -0.00001491, -0.00001960, -0.00001920, },
		{ 0.00004607, 0.00003190, 0.00000098, -0.00001996, -0.00001757, -0.00000863, -0.00001449, -0.00002683, -0.00001777, 0.00001430, 0.00003119, 0.00000422, -0.00003720, -0.00003489, 0.00001670, 0.00004824, -0.00000063, -0.00009052, -0.00011294, -0.00002258, 0.00009184, 0.00010104, -0.00000908, -0.00011148, -0.00008242, 0.00005368, 0.00015227, 0.00011317, -0.00001488, -0.00010289, -0.00008523, -0.00000823, 0.00004841, 0.00005759, 0.00004012, 0.00000985, -0.00003709, -0.00008307, -0.00008042, -0.00000739, 0.00008511, 0.00011413, 0.00005591, -0.00002564, -0.00005268, -0.00001965, 0.00001233, -0.00000184, -0.00003958, -0.00004890, -0.00001736, 0.00001780, 0.00002183, 0.00000229, -0.00001016, -0.00000428, 0.00000365, -0.00000205, -0.00001669, -0.00002592, -0.00002337, -0.00001059, 0.00001059, 0.00003458, },
		{ -0.00001050, -0.00002692, -0.00002252, -0.00001409, -0.00001618, -0.00001855, -0.00000391, 0.00001833, 0.00001758, -0.00001212, -0.00003310, -0.00001152, 0.00002680, 0.00001873, -0.00004912, -0.00010396, -0.00006722, 0.00003778, 0.00009605, 0.00003524, -0.00007647, -0.00010046, 0.00000828, 0.00014112, 0.00015699, 0.00004092, -0.00008614, -0.00010788, -0.00003003, 0.00005203, 0.00007068, 0.00003896, -0.00000055, -0.00003433, -0.00006520, -0.00007179, -0.00002346, 0.00006211, 0.00011201, 0.00007399, -0.00001560, -0.00006682, -0.00003898, 0.00001595, 0.00002328, -0.00002225, -0.00005741, -0.00003611, 0.00001475, 0.00003593, 0.00001164, -0.00001839, -0.00001681, 0.00000598, 0.00001351, -0.00000655, -0.00003126, -0.00003556, -0.00001856, 0.00000565, 0.00002755, 0.00004239, 0.00004162, 0.00001990, },
		{ -0.00002448, -0.00001391, -0.00001243, -0.00001068, 0.00000233, 0.00001503, 0.00000678, -0.00001661, -0.00002314, -0.00000104, 0.00001432, -0.00001617, -0.00007147, -0.00008152, -0.00001686, 0.00005770, 0.00005143, -0.00003390, -0.00008992, -0.00002688, 0.00010889, 0.00017650, 0.00010027, -0.00004280, -0.00011182, -0.00005678, 0.00004254, 0.00008299, 0.00004652, -0.00001048, -0.00004308, -0.00005452, -0.00005453, -0.00002638, 0.00003827, 0.00009604, 0.00008451, 0.00000334, -0.00006898, -0.00006113, 0.00000507, 0.00004174, 0.00000465, -0.00005424, -0.00005701, 0.00000052, 0.00004762, 0.00003167, -0.00001861, -0.00003706, -0.00000612, 0.00002716, 0.00001712, -0.00002461, -0.00005040, -0.00003618, -0.00000084, 0.00002693, 0.00003904, 0.00004049, 0.00003031, 0.00000588, -0.00002142, -0.00003256, },
		{ -0.00000783, -0.00000570, 0.00000434, 0.00001146, 0.00000115, -0.00001732, -0.00001963, -0.00000497, -0.00000412, -0.00003486, -0.00006561, -0.00004986, 0.00000468, 0.00003223, -0.00000730, -0.00006187, -0.00003925, 0.00007261, 0.00017075, 0.00014715, 0.00001785, -0.00008895, -0.00007589, 0.00002074, 0.00008580, 0.00005950, -0.00001305, -0.00005765, -0.00005617, -0.00003776, -0.00001602, 0.00002297, 0.00007140, 0.00008112, 0.00002270, -0.00005736, -0.00007650, -0.00001675, 0.00004610, 0.00003356, -0.00003624, -0.00007138, -0.00002421, 0.00004783, 0.00005693, -0.00000345, -0.00005400, -0.00003345, 0.00002619, 0.00004596, 0.00000051, -0.00005603, -0.00006199, -0.00001786, 0.00002800, 0.00004315, 0.00003676, 0.00002821, 0.00001703, -0.00000490, -0.00002875, -0.00003567, -0.00002380, -0.00001085, },
		{ 0.00000475, 0.00000808, -0.00000419, -0.00002151, -0.00002318, -0.00001159, -0.00001220, -0.00003439, -0.00005170, -0.00003824, -0.00001159, -0.00001212, -0.00003799, -0.00003109, 0.00004850, 0.00015060, 0.00017309, 0.00008088, -0.00004069, -0.00007507, -0.00000489, 0.00007412, 0.00006961, -0.00000660, -0.00006961, -0.00006870, -0.00003014, 0.00000115, 0.00002152, 0.00004767, 0.00006386, 0.00003304, -0.00003699, -0.00007819, -0.00004054, 0.00003431, 0.00005435, -0.00000690, -0.00007099, -0.00005186, 0.00003160, 0.00007634, 0.00002629, -0.00005542, -0.00006688, 0.00000361, 0.00006518, 0.00004041, -0.00004112, -0.00008545, -0.00004895, 0.00002048, 0.00005396, 0.00004050, 0.00001906, 0.00001434, 0.00001092, -0.00000974, -0.00003503, -0.00003947, -0.00002210, -0.00000557, -0.00000245, -0.00000218, },
		{ -0.00001161, -0.00002748, -0.00002531, -0.00001014, -0.00000841, -0.00003005, -0.00005338, -0.00005660, -0.00004662, -0.00003869, -0.00001910, 0.00004175, 0.00013181, 0.00018147, 0.00013378, 0.00002290, -0.00004857, -0.00002189, 0.00005031, 0.00006833, 0.00000394, -0.00007288, -0.00008499, -0.00003520, 0.00001546, 0.00003219, 0.00003386, 0.00003953, 0.00002906, -0.00001741, -0.00006577, -0.00005621, 0.00001149, 0.00005939, 0.00002430, -0.00005321, -0.00007128, 0.00000195, 0.00007926, 0.00006088, -0.00003535, -0.00009137, -0.00003671, 0.00006137, 0.00008130, -0.00000253, -0.00009186, -0.00008773, -0.00000440, 0.00006225, 0.00005624, 0.00001436, -0.00000146, 0.00001339, 0.00001535, -0.00001549, -0.00004747, -0.00004487, -0.00001570, 0.00000497, 0.00000376, -0.00000184, 0.00000137, 0.00000212, },
		{ -0.00002063, -0.00000180, -0.00000475, -0.00003961, -0.00007923, -0.00009172, -0.00007082, -0.00002550, 0.00004188, 0.00012429, 0.00018457, 0.00017346, 0.00008931, 0.00000026, -0.00002007, 0.00002342, 0.00005206, 0.00000972, -0.00006745, -0.00009659, -0.00004960, 0.00001925, 0.00004720, 0.00003361, 0.00001810, 0.00001255, -0.00000765, -0.00004566, -0.00005770, -0.00001241, 0.00004772, 0.00004631, -0.00002317, -0.00007360, -0.00003073, 0.00006173, 0.00008628, 0.00000181, -0.00009393, -0.00008075, 0.00003043, 0.00010540, 0.00005092, -0.00007059, -0.00011931, -0.00004689, 0.00005545, 0.00007883, 0.00002417, -0.00002016, -0.00000317, 0.00003153, 0.00001903, -0.00003524, -0.00006746, -0.00004312, 0.00000306, 0.00002005, 0.00000413, -0.00001045, -0.00000703, -0.00000388, -0.00001635, -0.00002897, },
		{ -0.00001296, -0.00006892, -0.00012055, -0.00012179, -0.00006456, 0.00002725, 0.00012355, 0.00019427, 0.00020758, 0.00015163, 0.00006211, 0.00000342, 0.00000440, 0.00002437, 0.00000323, -0.00005942, -0.00009876, -0.00006535, 0.00001097, 0.00005707, 0.00004337, 0.00000741, -0.00000860, -0.00001139, -0.00002767, -0.00004611, -0.00002759, 0.00002568, 0.00005199, 0.00000832, -0.00005716, -0.00005370, 0.00002923, 0.00009133, 0.00004274, -0.00007073, -0.00011087, -0.00001922, 0.00009990, 0.00010112, -0.00002217, -0.00012732, -0.00009619, 0.00002568, 0.00009518, 0.00005017, -0.00002767, -0.00003327, 0.00002633, 0.00005436, 0.00000076, -0.00007351, -0.00008105, -0.00001959, 0.00003418, 0.00002722, -0.00001249, -0.00002843, -0.00001222, -0.00000100, -0.00001393, -0.00002629, -0.00001384, 0.00000396, },
		{ -0.00016410, -0.00013178, -0.00002259, 0.00011007, 0.00021123, 0.00024818, 0.00021342, 0.00012913, 0.00004390, -0.00000001, -0.00000626, -0.00001740, -0.00005715, -0.00009304, -0.00007478, -0.00000447, 0.00005581, 0.00005460, 0.00000905, -0.00002482, -0.00002515, -0.00001958, -0.00002875, -0.00002944, 0.00000348, 0.00004164, 0.00003012, -0.00002854, -0.00005832, -0.00000577, 0.00007375, 0.00007165, -0.00002963, -0.00011453, -0.00006990, 0.00006403, 0.00012883, 0.00004012, -0.00010274, -0.00013427, -0.00002408, 0.00009134, 0.00008320, -0.00001444, -0.00006251, -0.00000272, 0.00007291, 0.00005137, -0.00005111, -0.00011276, -0.00006402, 0.00002979, 0.00005951, 0.00000711, -0.00004770, -0.00004094, 0.00000113, 0.00001441, -0.00001140, -0.00002918, -0.00001231, 0.00000231, -0.00003477, -0.00011301, },
		{ 0.00005830, 0.00021925, 0.00030059, 0.00028635, 0.00020219, 0.00009455, 0.00000968, -0.00003211, -0.00004792, -0.00006622, -0.00008580, -0.00007483, -0.00001909, 0.00004386, 0.00005883, 0.00001786, -0.00003015, -0.00004054, -0.00002306, -0.00001483, -0.00002049, -0.00001022, 0.00002237, 0.00003621, 0.00000087, -0.00004423, -0.00003000, 0.00004123, 0.00007808, 0.00001356, -0.00009062, -0.00010291, 0.00001008, 0.00012300, 0.00009540, -0.00004970, -0.00014439, -0.00007956, 0.00006081, 0.00010737, 0.00001838, -0.00007501, -0.00004580, 0.00006292, 0.00009799, -0.00000049, -0.00011959, -0.00011803, -0.00000355, 0.00008334, 0.00004869, -0.00004701, -0.00008037, -0.00002294, 0.00003840, 0.00002745, -0.00002660, -0.00004426, -0.00001234, -0.00000176, -0.00006798, -0.00016875, -0.00020221, -0.00011285, },
		{ 0.00038050, 0.00029256, 0.00015457, 0.00002818, -0.00005119, -0.00008193, -0.00008653, -0.00008374, -0.00006821, -0.00002675, 0.00002703, 0.00005216, 0.00002538, -0.00002519, -0.00004924, -0.00003319, -0.00001018, -0.00000833, -0.00001262, 0.00000381, 0.00002813, 0.00002069, -0.00001909, -0.00003586, 0.00000739, 0.00006179, 0.00004277, -0.00004983, -0.00010745, -0.00004215, 0.00008625, 0.00012464, 0.00001526, -0.00011982, -0.00012105, 0.00000939, 0.00010832, 0.00005855, -0.00006215, -0.00008524, 0.00002504, 0.00012127, 0.00006449, -0.00009081, -0.00016059, -0.00006295, 0.00008014, 0.00009869, -0.00001559, -0.00011058, -0.00007272, 0.00003747, 0.00007748, 0.00000824, -0.00006890, -0.00005823, 0.00000202, -0.00000621, -0.00012122, -0.00024002, -0.00022533, -0.00004823, 0.00018926, 0.00035358, },
		{ 0.00005300, -0.00006787, -0.00011664, -0.00011395, -0.00008950, -0.00006045, -0.00002635, 0.00001230, 0.00003677, 0.00002488, -0.00001610, -0.00004767, -0.00004187, -0.00001432, -0.00000071, -0.00000744, -0.00000721, 0.00001320, 0.00002648, 0.00000591, -0.00002451, -0.00001511, 0.00003223, 0.00004927, -0.00000913, -0.00008527, -0.00007465, 0.00003367, 0.00011945, 0.00007028, -0.00006783, -0.00013280, -0.00004642, 0.00008126, 0.00008862, -0.00002749, -0.00010431, -0.00002738, 0.00011060, 0.00012186, -0.00003017, -0.00017155, -0.00013146, 0.00004075, 0.00013499, 0.00004273, -0.00011133, -0.00013232, -0.00000002, 0.00011477, 0.00007310, -0.00006152, -0.00011680, -0.00003923, 0.00003814, -0.00003153, -0.00021542, -0.00032176, -0.00020322, 0.00009286, 0.00037757, 0.00049222, 0.00041552, 0.00023450, },
		{ -0.00014562, -0.00010083, -0.00005517, -0.00002111, 0.00000341, 0.00001841, 0.00001482, -0.00001038, -0.00003842, -0.00004281, -0.00002143, -0.00000102, -0.00000177, -0.00000983, -0.00000031, 0.00002090, 0.00002198, -0.00000420, -0.00002044, 0.00000310, 0.00003511, 0.00001662, -0.00004906, -0.00007930, -0.00001461, 0.00008561, 0.00009813, -0.00000681, -0.00011083, -0.00008691, 0.00003419, 0.00009484, 0.00001476, -0.00009565, -0.00007469, 0.00006936, 0.00015149, 0.00004475, -0.00014169, -0.00018482, -0.00002833, 0.00013752, 0.00010941, -0.00007320, -0.00017690, -0.00006892, 0.00011613, 0.00014794, -0.00000950, -0.00015748, -0.00011906, 0.00003372, 0.00006713, -0.00012097, -0.00035835, -0.00037320, -0.00008057, 0.00033381, 0.00059629, 0.00057771, 0.00035420, 0.00009241, -0.00008722, -0.00015620, },
		{ -0.00001444, -0.00000059, 0.00000187, -0.00000144, -0.00001225, -0.00002785, -0.00003507, -0.00002456, -0.00000678, -0.00000119, -0.00000846, -0.00000785, 0.00001081, 0.00002652, 0.00001495, -0.00001127, -0.00001461, 0.00001077, 0.00002130, -0.00001544, -0.00006067, -0.00004282, 0.00003962, 0.00009373, 0.00004228, -0.00006484, -0.00009854, -0.00001566, 0.00007437, 0.00004701, -0.00006464, -0.00010036, 0.00001324, 0.00014425, 0.00010998, -0.00007772, -0.00020240, -0.00010654, 0.00009866, 0.00015903, -0.00000373, -0.00018456, -0.00014786, 0.00007136, 0.00020326, 0.00007936, -0.00015246, -0.00020965, -0.00003195, 0.00013165, 0.00002122, -0.00030559, -0.00050627, -0.00031160, 0.00019203, 0.00064879, 0.00076461, 0.00053216, 0.00016823, -0.00010594, -0.00020913, -0.00018536, -0.00011446, -0.00005108, },
		{ -0.00001936, -0.00002084, -0.00002141, -0.00002275, -0.00002016, -0.00001207, -0.00000636, -0.00000889, -0.00001040, 0.00000245, 0.00002290, 0.00002627, 0.00000420, -0.00001734, -0.00001104, 0.00000913, 0.00000324, -0.00003233, -0.00004629, 0.00000023, 0.00006545, 0.00006542, -0.00001344, -0.00008028, -0.00005017, 0.00003657, 0.00005717, -0.00002614, -0.00009850, -0.00003757, 0.00010567, 0.00014572, 0.00000038, -0.00017670, -0.00016787, 0.00002772, 0.00017097, 0.00007486, -0.00014727, -0.00020831, -0.00001034, 0.00021381, 0.00017864, -0.00009093, -0.00027347, -0.00014472, 0.00012795, 0.00016569, -0.00016393, -0.00055133, -0.00055092, -0.00005123, 0.00060154, 0.00093583, 0.00077041, 0.00030938, -0.00010506, -0.00027979, -0.00024463, -0.00013246, -0.00004361, -0.00000548, -0.00000273, -0.00001235, },
		{ -0.00001133, -0.00000996, -0.00001204, -0.00001329, -0.00001388, -0.00001234, -0.00000218, 0.00001686, 0.00002900, 0.00001761, -0.00000901, -0.00002185, -0.00000915, 0.00000405, -0.00000911, -0.00003193, -0.00002040, 0.00002975, 0.00006113, 0.00002498, -0.00004339, -0.00005903, -0.00000205, 0.00004435, 0.00000351, -0.00007563, -0.00006775, 0.00005241, 0.00014463, 0.00006823, -0.00011665, -0.00019227, -0.00005257, 0.00013916, 0.00013531, -0.00007542, -0.00022754, -0.00010354, 0.00017080, 0.00025441, 0.00001455, -0.00028018, -0.00026838, 0.00004696, 0.00026109, 0.00003263, -0.00047991, -0.00073348, -0.00036292, 0.00042328, 0.00103300, 0.00104267, 0.00053721, -0.00005062, -0.00035798, -0.00033644, -0.00016666, -0.00003096, 0.00001195, -0.00000303, -0.00002734, -0.00003762, -0.00003233, -0.00002012, },
		{ -0.00001692, -0.00002139, -0.00001729, -0.00000543, 0.00001202, 0.00002669, 0.00002478, 0.00000290, -0.00002054, -0.00002283, -0.00000699, 0.00000017, -0.00001299, -0.00002127, 0.00000244, 0.00003952, 0.00004059, -0.00000515, -0.00004408, -0.00002635, 0.00001842, 0.00001468, -0.00004600, -0.00007311, 0.00000376, 0.00011365, 0.00010753, -0.00004308, -0.00017417, -0.00011612, 0.00007488, 0.00015749, 0.00000670, -0.00019895, -0.00017758, 0.00008639, 0.00028020, 0.00013269, -0.00022054, -0.00035966, -0.00009035, 0.00027124, 0.00022741, -0.00030028, -0.00080196, -0.00067849, 0.00011818, 0.00099929, 0.00129329, 0.00084857, 0.00009540, -0.00041333, -0.00046406, -0.00023660, -0.00002117, 0.00004495, 0.00000513, -0.00004524, -0.00005974, -0.00004460, -0.00002075, -0.00000312, 0.00000114, -0.00000648, },
		{ -0.00001080, 0.00000788, 0.00002317, 0.00002649, 0.00001240, -0.00001181, -0.00002654, -0.00001984, -0.00000414, -0.00000132, -0.00001095, -0.00000878, 0.00001561, 0.00003544, 0.00002000, -0.00001710, -0.00002988, -0.00000611, 0.00000840, -0.00002345, -0.00006060, -0.00002714, 0.00006911, 0.00011292, 0.00002082, -0.00012383, -0.00014465, 0.00000132, 0.00013704, 0.00007176, -0.00013464, -0.00020999, -0.00001172, 0.00024844, 0.00022680, -0.00011059, -0.00038647, -0.00024113, 0.00019009, 0.00036229, -0.00005712, -0.00072937, -0.00092134, -0.00026773, 0.00080323, 0.00145045, 0.00120547, 0.00035676, -0.00039972, -0.00061009, -0.00035994, -0.00003481, 0.00009346, 0.00003356, -0.00006161, -0.00009102, -0.00006099, -0.00001977, 0.00000468, 0.00000969, 0.00000036, -0.00001550, -0.00002660, -0.00002469, },
		{ 0.00002571, 0.00001789, -0.00000167, -0.00002143, -0.00002662, -0.00001486, -0.00000156, -0.00000116, -0.00000646, 0.00000101, 0.00002033, 0.00002691, 0.00000726, -0.00001701, -0.00001810, -0.00000580, -0.00001460, -0.00004273, -0.00003787, 0.00002805, 0.00009206, 0.00005931, -0.00006150, -0.00013461, -0.00005704, 0.00008616, 0.00010129, -0.00005920, -0.00019554, -0.00009265, 0.00017297, 0.00027024, 0.00001553, -0.00034079, -0.00035821, 0.00004384, 0.00039971, 0.00018392, -0.00053100, -0.00102934, -0.00065665, 0.00045635, 0.00145009, 0.00153829, 0.00072512, -0.00027068, -0.00073298, -0.00053882, -0.00009978, 0.00014308, 0.00009211, -0.00006322, -0.00013220, -0.00008969, -0.00001862, 0.00001757, 0.00001754, 0.00000357, -0.00001262, -0.00002636, -0.00003002, -0.00001885, 0.00000135, 0.00001907, },
		{ -0.00001226, -0.00002528, -0.00002360, -0.00001039, 0.00000024, 0.00000022, -0.00000135, 0.00000752, 0.00002060, 0.00001939, 0.00000102, -0.00001516, -0.00001646, -0.00001677, -0.00002999, -0.00003517, 0.00000071, 0.00005976, 0.00006914, -0.00000766, -0.00009682, -0.00008473, 0.00002691, 0.00009268, 0.00000179, -0.00014696, -0.00013534, 0.00008174, 0.00025656, 0.00012044, -0.00024064, -0.00040880, -0.00012044, 0.00033604, 0.00036002, -0.00025981, -0.00097737, -0.00096302, 0.00001582, 0.00126129, 0.00176396, 0.00115199, -0.00000061, -0.00077556, -0.00074901, -0.00023830, 0.00016443, 0.00017879, -0.00003218, -0.00017376, -0.00013829, -0.00002615, 0.00003810, 0.00003182, 0.00000277, -0.00001486, -0.00002296, -0.00002842, -0.00002539, -0.00000920, 0.00001158, 0.00002363, 0.00002116, 0.00000676, },
		{ -0.00002040, -0.00000710, 0.00000223, 0.00000307, 0.00000368, 0.00001138, 0.00001881, 0.00001328, -0.00000349, -0.00001697, -0.00002158, -0.00002573, -0.00002898, -0.00001219, 0.00002983, 0.00005804, 0.00002529, -0.00004959, -0.00008037, -0.00001937, 0.00005818, 0.00003263, -0.00008738, -0.00013604, 0.00000388, 0.00019995, 0.00017859, -0.00012064, -0.00038589, -0.00025407, 0.00020062, 0.00043329, 0.00001262, -0.00078619, -0.00112191, -0.00043059, 0.00090207, 0.00181334, 0.00155633, 0.00039742, -0.00068433, -0.00093915, -0.00045147, 0.00012174, 0.00027308, 0.00004416, -0.00019552, -0.00020586, -0.00005566, 0.00006054, 0.00006017, 0.00000483, -0.00002635, -0.00002411, -0.00001969, -0.00002368, -0.00002050, -0.00000234, 0.00001760, 0.00002375, 0.00001430, -0.00000280, -0.00001877, -0.00002601, },
		{ 0.00000568, 0.00000711, 0.00000767, 0.00001257, 0.00001536, 0.00000723, -0.00000871, -0.00002126, -0.00002614, -0.00002593, -0.00001661, 0.00000921, 0.00003841, 0.00003621, -0.00000964, -0.00005456, -0.00004087, 0.00001734, 0.00003254, -0.00003900, -0.00010710, -0.00004218, 0.00012665, 0.00018465, -0.00001664, -0.00030784, -0.00032499, 0.00004183, 0.00040183, 0.00022156, -0.00051401, -0.00111037, -0.00079187, 0.00043783, 0.00165692, 0.00184717, 0.00086652, -0.00043271, -0.00104317, -0.00070929, -0.00001117, 0.00033762, 0.00016212, -0.00017363, -0.00027770, -0.00011696, 0.00007005, 0.00010370, 0.00002146, -0.00004358, -0.00003843, -0.00001084, -0.00000968, -0.00002453, -0.00002152, 0.00000289, 0.00002354, 0.00002307, 0.00000704, -0.00001103, -0.00002324, -0.00002587, -0.00001756, -0.00000367, },
		{ 0.00000924, 0.00001088, 0.00001054, 0.00000121, -0.00001382, -0.00002464, -0.00002634, -0.00001937, -0.00000285, 0.00002019, 0.00003235, 0.00001448, -0.00002272, -0.00003832, -0.00001279, 0.00001309, -0.00001384, -0.00006870, -0.00005426, 0.00006225, 0.00015206, 0.00004828, -0.00020748, -0.00032710, -0.00009283, 0.00029587, 0.00033013, -0.00023452, -0.00095285, -0.00100285, -0.00003781, 0.00131790, 0.00195312, 0.00131790, -0.00003781, -0.00100285, -0.00095285, -0.00023452, 0.00033013, 0.00029587, -0.00009283, -0.00032710, -0.00020748, 0.00004828, 0.00015206, 0.00006225, -0.00005426, -0.00006870, -0.00001384, 0.00001309, -0.00001279, -0.00003832, -0.00002272, 0.00001448, 0.00003235, 0.00002019, -0.00000285, -0.00001937, -0.00002634, -0.00002464, -0.00001382, 0.00000121, 0.00001054, 0.00001088, },
		{ 0.00000568, -0.00000367, -0.00001756, -0.00002587, -0.00002324, -0.00001103, 0.00000704, 0.00002307, 0.00002354, 0.00000289, -0.00002152, -0.00002453, -0.00000968, -0.00001084, -0.00003843, -0.00004358, 0.00002146, 0.00010370, 0.00007005, -0.00011696, -0.00027770, -0.00017363, 0.00016212, 0.00033762, -0.00001117, -0.00070929, -0.00104317, -0.00043271, 0.00086652, 0.00184717, 0.00165692, 0.00043783, -0.00079187, -0.00111037, -0.00051401, 0.00022156, 0.00040183, 0.00004183, -0.00032499, -0.00030784, -0.00001664, 0.00018465, 0.00012665, -0.00004218, -0.00010710, -0.00003900, 0.00003254, 0.00001734, -0.00004087, -0.00005456, -0.00000964, 0.00003621, 0.00003841, 0.00000921, -0.00001661, -0.00002593, -0.00002614, -0.00002126, -0.00000871, 0.00000723, 0.00001536, 0.00001257, 0.00000767, 0.00000711, },
		{ -0.00002040, -0.00002601, -0.00001877, -0.00000280, 0.00001430, 0.00002375, 0.00001760, -0.00000234, -0.00002050, -0.00002368, -0.00001969, -0.00002411, -0.00002635, 0.00000483, 0.00006017, 0.00006054, -0.00005566, -0.00020586, -0.00019552, 0.00004416, 0.00027308, 0.00012174, -0.00045147, -0.00093915, -0.00068433, 0.00039742, 0.00155633, 0.00181334, 0.00090207, -0.00043059, -0.00112191, -0.00078619, 0.00001262, 0.00043329, 0.00020062, -0.00025407, -0.00038589, -0.00012064, 0.00017859, 0.00019995, 0.00000388, -0.00013604, -0.00008738, 0.00003263, 0.00005818, -0.00001937, -0.00008037, -0.00004959, 0.00002529, 0.00005804, 0.00002983, -0.00001219, -0.00002898, -0.00002573, -0.00002158, -0.00001697, -0.00000349, 0.00001328, 0.00001881, 0.00001138, 0.00000368, 0.00000307, 0.00000223, -0.00000710, },
		{ -0.00001226, 0.00000676, 0.00002116, 0.00002363, 0.00001158, -0.00000920, -0.00002539, -0.00002842, -0.00002296, -0.00001486, 0.00000277, 0.00003182, 0.00003810, -0.00002615, -0.00013829, -0.00017376, -0.00003218, 0.00017879, 0.00016443, -0.00023830, -0.00074901, -0.00077556, -0.00000061, 0.00115199, 0.00176396, 0.00126129, 0.00001582, -0.00096302, -0.00097737, -0.00025981, 0.00036002, 0.00033604, -0.00012044, -0.00040880, -0.00024064, 0.00012044, 0.00025656, 0.00008174, -0.00013534, -0.00014696, 0.00000179, 0.00009268, 0.00002691, -0.00008473, -0.00009682, -0.00000766, 0.00006914, 0.00005976, 0.00000071, -0.00003517, -0.00002999, -0.00001677, -0.00001646, -0.00001516, 0.00000102, 0.00001939, 0.00002060, 0.00000752, -0.00000135, 0.00000022, 0.00000024, -0.00001039, -0.00002360, -0.00002528, },
		{ 0.00002571, 0.00001907, 0.00000135, -0.00001885, -0.00003002, -0.00002636, -0.00001262, 0.00000357, 0.00001754, 0.00001757, -0.00001862, -0.00008969, -0.00013220, -0.00006322, 0.00009211, 0.00014308, -0.00009978, -0.00053882, -0.00073298, -0.00027068, 0.00072512, 0.00153829, 0.00145009, 0.00045635, -0.00065665, -0.00102934, -0.00053100, 0.00018392, 0.00039971, 0.00004384, -0.00035821, -0.00034079, 0.00001553, 0.00027024, 0.00017297, -0.00009265, -0.00019554, -0.00005920, 0.00010129, 0.00008616, -0.00005704, -0.00013461, -0.00006150, 0.00005931, 0.00009206, 0.00002805, -0.00003787, -0.00004273, -0.00001460, -0.00000580, -0.00001810, -0.00001701, 0.00000726, 0.00002691, 0.00002033, 0.00000101, -0.00000646, -0.00000116, -0.00000156, -0.00001486, -0.00002662, -0.00002143, -0.00000167, 0.00001789, },
		{ -0.00001080, -0.00002469, -0.00002660, -0.00001550, 0.00000036, 0.00000969, 0.00000468, -0.00001977, -0.00006099, -0.00009102, -0.00006161, 0.00003356, 0.00009346, -0.00003481, -0.00035994, -0.00061009, -0.00039972, 0.00035676, 0.00120547, 0.00145045, 0.00080323, -0.00026773, -0.00092134, -0.00072937, -0.00005712, 0.00036229, 0.00019009, -0.00024113, -0.00038647, -0.00011059, 0.00022680, 0.00024844, -0.00001172, -0.00020999, -0.00013464, 0.00007176, 0.00013704, 0.00000132, -0.00014465, -0.00012383, 0.00002082, 0.00011292, 0.00006911, -0.00002714, -0.00006060, -0.00002345, 0.00000840, -0.00000611, -0.00002988, -0.00001710, 0.00002000, 0.00003544, 0.00001561, -0.00000878, -0.00001095, -0.00000132, -0.00000414, -0.00001984, -0.00002654, -0.00001181, 0.00001240, 0.00002649, 0.00002317, 0.00000788, },
		{ -0.00001692, -0.00000648, 0.00000114, -0.00000312, -0.00002075, -0.00004460, -0.00005974, -0.00004524, 0.00000513, 0.00004495, -0.00002117, -0.00023660, -0.00046406, -0.00041333, 0.00009540, 0.00084857, 0.00129329, 0.00099929, 0.00011818, -0.00067849, -0.00080196, -0.00030028, 0.00022741, 0.00027124, -0.00009035, -0.00035966, -0.00022054, 0.00013269, 0.00028020, 0.00008639, -0.00017758, -0.00019895, 0.00000670, 0.00015749, 0.00007488, -0.00011612, -0.00017417, -0.00004308, 0.00010753, 0.00011365, 0.00000376, -0.00007311, -0.00004600, 0.00001468, 0.00001842, -0.00002635, -0.00004408, -0.00000515, 0.00004059, 0.00003952, 0.00000244, -0.00002127, -0.00001299, 0.00000017, -0.00000699, -0.00002283, -0.00002054, 0.00000290, 0.00002478, 0.00002669, 0.00001202, -0.00000543, -0.00001729, -0.00002139, },
		{ -0.00001133, -0.00002012, -0.00003233, -0.00003762, -0.00002734, -0.00000303, 0.00001195, -0.00003096, -0.00016666, -0.00033644, -0.00035798, -0.00005062, 0.00053721, 0.00104267, 0.00103300, 0.00042328, -0.00036292, -0.00073348, -0.00047991, 0.00003263, 0.00026109, 0.00004696, -0.00026838, -0.00028018, 0.00001455, 0.00025441, 0.00017080, -0.00010354, -0.00022754, -0.00007542, 0.00013531, 0.00013916, -0.00005257, -0.00019227, -0.00011665, 0.00006823, 0.00014463, 0.00005241, -0.00006775, -0.00007563, 0.00000351, 0.00004435, -0.00000205, -0.00005903, -0.00004339, 0.00002498, 0.00006113, 0.00002975, -0.00002040, -0.00003193, -0.00000911, 0.00000405, -0.00000915, -0.00002185, -0.00000901, 0.00001761, 0.00002900, 0.00001686, -0.00000218, -0.00001234, -0.00001388, -0.00001329, -0.00001204, -0.00000996, },
		{ -0.00001936, -0.00001235, -0.00000273, -0.00000548, -0.00004361, -0.00013246, -0.00024463, -0.00027979, -0.00010506, 0.00030938, 0.00077041, 0.00093583, 0.00060154, -0.00005123, -0.00055092, -0.00055133, -0.00016393, 0.00016569, 0.00012795, -0.00014472, -0.00027347, -0.00009093, 0.00017864, 0.00021381, -0.00001034, -0.00020831, -0.00014727, 0.00007486, 0.00017097, 0.00002772, -0.00016787, -0.00017670, 0.00000038, 0.00014572, 0.00010567, -0.00003757, -0.00009850, -0.00002614, 0.00005717, 0.00003657, -0.00005017, -0.00008028, -0.00001344, 0.00006542, 0.00006545, 0.00000023, -0.00004629, -0.00003233, 0.00000324, 0.00000913, -0.00001104, -0.00001734, 0.00000420, 0.00002627, 0.00002290, 0.00000245, -0.00001040, -0.00000889, -0.00000636, -0.00001207, -0.00002016, -0.00002275, -0.00002141, -0.00002084, },
		{ -0.00001444, -0.00005108, -0.00011446, -0.00018536, -0.00020913, -0.00010594, 0.00016823, 0.00053216, 0.00076461, 0.00064879, 0.00019203, -0.00031160, -0.00050627, -0.00030559, 0.00002122, 0.00013165, -0.00003195, -0.00020965, -0.00015246, 0.00007936, 0.00020326, 0.00007136, -0.00014786, -0.00018456, -0.00000373, 0.00015903, 0.00009866, -0.00010654, -0.00020240, -0.00007772, 0.00010998, 0.00014425, 0.00001324, -0.00010036, -0.00006464, 0.00004701, 0.00007437, -0.00001566, -0.00009854, -0.00006484, 0.00004228, 0.00009373, 0.00003962, -0.00004282, -0.00006067, -0.00001544, 0.00002130, 0.00001077, -0.00001461, -0.00001127, 0.00001495, 0.00002652, 0.00001081, -0.00000785, -0.00000846, -0.00000119, -0.00000678, -0.00002456, -0.00003507, -0.00002785, -0.00001225, -0.00000144, 0.00000187, -0.00000059, },
		{ -0.00014562, -0.00015620, -0.00008722, 0.00009241, 0.00035420, 0.00057771, 0.00059629, 0.00033381, -0.00008057, -0.00037320, -0.00035835, -0.00012097, 0.00006713, 0.00003372, -0.00011906, -0.00015748, -0.00000950, 0.00014794, 0.00011613, -0.00006892, -0.00017690, -0.00007320, 0.00010941, 0.00013752, -0.00002833, -0.00018482, -0.00014169, 0.00004475, 0.00015149, 0.00006936, -0.00007469, -0.00009565, 0.00001476, 0.00009484, 0.00003419, -0.00008691, -0.00011083, -0.00000681, 0.00009813, 0.00008561, -0.00001461, -0.00007930, -0.00004906, 0.00001662, 0.00003511, 0.00000310, -0.00002044, -0.00000420, 0.00002198, 0.00002090, -0.00000031, -0.00000983, -0.00000177, -0.00000102, -0.00002143, -0.00004281, -0.00003842, -0.00001038, 0.00001482, 0.00001841, 0.00000341, -0.00002111, -0.00005517, -0.00010083, },
		{ 0.00005300, 0.00023450, 0.00041552, 0.00049222, 0.00037757, 0.00009286, -0.00020322, -0.00032176, -0.00021542, -0.00003153, 0.00003814, -0.00003923, -0.00011680, -0.00006152, 0.00007310, 0.00011477, -0.00000002, -0.00013232, -0.00011133, 0.00004273, 0.00013499, 0.00004075, -0.00013146, -0.00017155, -0.00003017, 0.00012186, 0.00011060, -0.00002738, -0.00010431, -0.00002749, 0.00008862, 0.00008126, -0.00004642, -0.00013280, -0.00006783, 0.00007028, 0.00011945, 0.00003367, -0.00007465, -0.00008527, -0.00000913, 0.00004927, 0.00003223, -0.00001511, -0.00002451, 0.00000591, 0.00002648, 0.00001320, -0.00000721, -0.00000744, -0.00000071, -0.00001432, -0.00004187, -0.00004767, -0.00001610, 0.00002488, 0.00003677, 0.00001230, -0.00002635, -0.00006045, -0.00008950, -0.00011395, -0.00011664, -0.00006787, },
		{ 0.00038050, 0.00035358, 0.00018926, -0.00004823, -0.00022533, -0.00024002, -0.00012122, -0.00000621, 0.00000202, -0.00005823, -0.00006890, 0.00000824, 0.00007748, 0.00003747, -0.00007272, -0.00011058, -0.00001559, 0.00009869, 0.00008014, -0.00006295, -0.00016059, -0.00009081, 0.00006449, 0.00012127, 0.00002504, -0.00008524, -0.00006215, 0.00005855, 0.00010832, 0.00000939, -0.00012105, -0.00011982, 0.00001526, 0.00012464, 0.00008625, -0.00004215, -0.00010745, -0.00004983, 0.00004277, 0.00006179, 0.00000739, -0.00003586, -0.00001909, 0.00002069, 0.00002813, 0.00000381, -0.00001262, -0.00000833, -0.00001018, -0.00003319, -0.00004924, -0.00002519, 0.00002538, 0.00005216, 0.00002703, -0.00002675, -0.00006821, -0.00008374, -0.00008653, -0.00008193, -0.00005119, 0.00002818, 0.00015457, 0.00029256, },
		{ 0.00005830, -0.00011285, -0.00020221, -0.00016875, -0.00006798, -0.00000176, -0.00001234, -0.00004426, -0.00002660, 0.00002745, 0.00003840, -0.00002294, -0.00008037, -0.00004701, 0.00004869, 0.00008334, -0.00000355, -0.00011803, -0.00011959, -0.00000049, 0.00009799, 0.00006292, -0.00004580, -0.00007501, 0.00001838, 0.00010737, 0.00006081, -0.00007956, -0.00014439, -0.00004970, 0.00009540, 0.00012300, 0.00001008, -0.00010291, -0.00009062, 0.00001356, 0.00007808, 0.00004123, -0.00003000, -0.00004423, 0.00000087, 0.00003621, 0.00002237, -0.00001022, -0.00002049, -0.00001483, -0.00002306, -0.00004054, -0.00003015, 0.00001786, 0.00005883, 0.00004386, -0.00001909, -0.00007483, -0.00008580, -0.00006622, -0.00004792, -0.00003211, 0.00000968, 0.00009455, 0.00020219, 0.00028635, 0.00030059, 0.00021925, },
		{ -0.00016410, -0.00011301, -0.00003477, 0.00000231, -0.00001231, -0.00002918, -0.00001140, 0.00001441, 0.00000113, -0.00004094, -0.00004770, 0.00000711, 0.00005951, 0.00002979, -0.00006402, -0.00011276, -0.00005111, 0.00005137, 0.00007291, -0.00000272, -0.00006251, -0.00001444, 0.00008320, 0.00009134, -0.00002408, -0.00013427, -0.00010274, 0.00004012, 0.00012883, 0.00006403, -0.00006990, -0.00011453, -0.00002963, 0.00007165, 0.00007375, -0.00000577, -0.00005832, -0.00002854, 0.00003012, 0.00004164, 0.00000348, -0.00002944, -0.00002875, -0.00001958, -0.00002515, -0.00002482, 0.00000905, 0.00005460, 0.00005581, -0.00000447, -0.00007478, -0.00009304, -0.00005715, -0.00001740, -0.00000626, -0.00000001, 0.00004390, 0.00012913, 0.00021342, 0.00024818, 0.00021123, 0.00011007, -0.00002259, -0.00013178, },
		{ -0.00001296, 0.00000396, -0.00001384, -0.00002629, -0.00001393, -0.00000100, -0.00001222, -0.00002843, -0.00001249, 0.00002722, 0.00003418, -0.00001959, -0.00008105, -0.00007351, 0.00000076, 0.00005436, 0.00002633, -0.00003327, -0.00002767, 0.00005017, 0.00009518, 0.00002568, -0.00009619, -0.00012732, -0.00002217, 0.00010112, 0.00009990, -0.00001922, -0.00011087, -0.00007073, 0.00004274, 0.00009133, 0.00002923, -0.00005370, -0.00005716, 0.00000832, 0.00005199, 0.00002568, -0.00002759, -0.00004611, -0.00002767, -0.00001139, -0.00000860, 0.00000741, 0.00004337, 0.00005707, 0.00001097, -0.00006535, -0.00009876, -0.00005942, 0.00000323, 0.00002437, 0.00000440, 0.00000342, 0.00006211, 0.00015163, 0.00020758, 0.00019427, 0.00012355, 0.00002725, -0.00006456, -0.00012179, -0.00012055, -0.00006892, },
		{ -0.00002063, -0.00002897, -0.00001635, -0.00000388, -0.00000703, -0.00001045, 0.00000413, 0.00002005, 0.00000306, -0.00004312, -0.00006746, -0.00003524, 0.00001903, 0.00003153, -0.00000317, -0.00002016, 0.00002417, 0.00007883, 0.00005545, -0.00004689, -0.00011931, -0.00007059, 0.00005092, 0.00010540, 0.00003043, -0.00008075, -0.00009393, 0.00000181, 0.00008628, 0.00006173, -0.00003073, -0.00007360, -0.00002317, 0.00004631, 0.00004772, -0.00001241, -0.00005770, -0.00004566, -0.00000765, 0.00001255, 0.00001810, 0.00003361, 0.00004720, 0.00001925, -0.00004960, -0.00009659, -0.00006745, 0.00000972, 0.00005206, 0.00002342, -0.00002007, 0.00000026, 0.00008931, 0.00017346, 0.00018457, 0.00012429, 0.00004188, -0.00002550, -0.00007082, -0.00009172, -0.00007923, -0.00003961, -0.00000475, -0.00000180, },
		{ -0.00001161, 0.00000212, 0.00000137, -0.00000184, 0.00000376, 0.00000497, -0.00001570, -0.00004487, -0.00004747, -0.00001549, 0.00001535, 0.00001339, -0.00000146, 0.00001436, 0.00005624, 0.00006225, -0.00000440, -0.00008773, -0.00009186, -0.00000253, 0.00008130, 0.00006137, -0.00003671, -0.00009137, -0.00003535, 0.00006088, 0.00007926, 0.00000195, -0.00007128, -0.00005321, 0.00002430, 0.00005939, 0.00001149, -0.00005621, -0.00006577, -0.00001741, 0.00002906, 0.00003953, 0.00003386, 0.00003219, 0.00001546, -0.00003520, -0.00008499, -0.00007288, 0.00000394, 0.00006833, 0.00005031, -0.00002189, -0.00004857, 0.00002290, 0.00013378, 0.00018147, 0.00013181, 0.00004175, -0.00001910, -0.00003869, -0.00004662, -0.00005660, -0.00005338, -0.00003005, -0.00000841, -0.00001014, -0.00002531, -0.00002748, },
		{ 0.00000475, -0.00000218, -0.00000245, -0.00000557, -0.00002210, -0.00003947, -0.00003503, -0.00000974, 0.00001092, 0.00001434, 0.00001906, 0.00004050, 0.00005396, 0.00002048, -0.00004895, -0.00008545, -0.00004112, 0.00004041, 0.00006518, 0.00000361, -0.00006688, -0.00005542, 0.00002629, 0.00007634, 0.00003160, -0.00005186, -0.00007099, -0.00000690, 0.00005435, 0.00003431, -0.00004054, -0.00007819, -0.00003699, 0.00003304, 0.00006386, 0.00004767, 0.00002152, 0.00000115, -0.00003014, -0.00006870, -0.00006961, -0.00000660, 0.00006961, 0.00007412, -0.00000489, -0.00007507, -0.00004069, 0.00008088, 0.00017309, 0.00015060, 0.00004850, -0.00003109, -0.00003799, -0.00001212, -0.00001159, -0.00003824, -0.00005170, -0.00003439, -0.00001220, -0.00001159, -0.00002318, -0.00002151, -0.00000419, 0.00000808, },
		{ -0.00000783, -0.00001085, -0.00002380, -0.00003567, -0.00002875, -0.00000490, 0.00001703, 0.00002821, 0.00003676, 0.00004315, 0.00002800, -0.00001786, -0.00006199, -0.00005603, 0.00000051, 0.00004596, 0.00002619, -0.00003345, -0.00005400, -0.00000345, 0.00005693, 0.00004783, -0.00002421, -0.00007138, -0.00003624, 0.00003356, 0.00004610, -0.00001675, -0.00007650, -0.00005736, 0.00002270, 0.00008112, 0.00007140, 0.00002297, -0.00001602, -0.00003776, -0.00005617, -0.00005765, -0.00001305, 0.00005950, 0.00008580, 0.00002074, -0.00007589, -0.00008895, 0.00001785, 0.00014715, 0.00017075, 0.00007261, -0.00003925, -0.00006187, -0.00000730, 0.00003223, 0.00000468, -0.00004986, -0.00006561, -0.00003486, -0.00000412, -0.00000497, -0.00001963, -0.00001732, 0.00000115, 0.00001146, 0.00000434, -0.00000570, },
		{ -0.00002448, -0.00003256, -0.00002142, 0.00000588, 0.00003031, 0.00004049, 0.00003904, 0.00002693, -0.00000084, -0.00003618, -0.00005040, -0.00002461, 0.00001712, 0.00002716, -0.00000612, -0.00003706, -0.00001861, 0.00003167, 0.00004762, 0.00000052, -0.00005701, -0.00005424, 0.00000465, 0.00004174, 0.00000507, -0.00006113, -0.00006898, 0.00000334, 0.00008451, 0.00009604, 0.00003827, -0.00002638, -0.00005453, -0.00005452, -0.00004308, -0.00001048, 0.00004652, 0.00008299, 0.00004254, -0.00005678, -0.00011182, -0.00004280, 0.00010027, 0.00017650, 0.00010889, -0.00002688, -0.00008992, -0.00003390, 0.00005143, 0.00005770, -0.00001686, -0.00008152, -0.00007147, -0.00001617, 0.00001432, -0.00000104, -0.00002314, -0.00001661, 0.00000678, 0.00001503, 0.00000233, -0.00001068, -0.00001243, -0.00001391, },
		{ -0.00001050, 0.00001990, 0.00004162, 0.00004239, 0.00002755, 0.00000565, -0.00001856, -0.00003556, -0.00003126, -0.00000655, 0.00001351, 0.00000598, -0.00001681, -0.00001839, 0.00001164, 0.00003593, 0.00001475, -0.00003611, -0.00005741, -0.00002225, 0.00002328, 0.00001595, -0.00003898, -0.00006682, -0.00001560, 0.00007399, 0.00011201, 0.00006211, -0.00002346, -0.00007179, -0.00006520, -0.00003433, -0.00000055, 0.00003896, 0.00007068, 0.00005203, -0.00003003, -0.00010788, -0.00008614, 0.00004092, 0.00015699, 0.00014112, 0.00000828, -0.00010046, -0.00007647, 0.00003524, 0.00009605, 0.00003778, -0.00006722, -0.00010396, -0.00004912, 0.00001873, 0.00002680, -0.00001152, -0.00003310, -0.00001212, 0.00001758, 0.00001833, -0.00000391, -0.00001855, -0.00001618, -0.00001409, -0.00002252, -0.00002692, },
		{ 0.00004607, 0.00003458, 0.00001059, -0.00001059, -0.00002337, -0.00002592, -0.00001669, -0.00000205, 0.00000365, -0.00000428, -0.00001016, 0.00000229, 0.00002183, 0.00001780, -0.00001736, -0.00004890, -0.00003958, -0.00000184, 0.00001233, -0.00001965, -0.00005268, -0.00002564, 0.00005591, 0.00011413, 0.00008511, -0.00000739, -0.00008042, -0.00008307, -0.00003709, 0.00000985, 0.00004012, 0.00005759, 0.00004841, -0.00000823, -0.00008523, -0.00010289, -0.00001488, 0.00011317, 0.00015227, 0.00005368, -0.00008242, -0.00011148, -0.00000908, 0.00010104, 0.00009184, -0.00002258, -0.00011294, -0.00009052, -0.00000063, 0.00004824, 0.00001670, -0.00003489, -0.00003720, 0.00000422, 0.00003119, 0.00001430, -0.00001777, -0.00002683, -0.00001449, -0.00000863, -0.00001757, -0.00001996, 0.00000098, 0.00003190, },
		{ -0.00000587, -0.00001920, -0.00001960, -0.00001491, -0.00000899, -0.00000406, -0.00000312, -0.00000404, 0.00000107, 0.00001248, 0.00001490, -0.00000522, -0.00003550, -0.00004544, -0.00002507, -0.00000297, -0.00001045, -0.00003372, -0.00002345, 0.00003924, 0.00010391, 0.00009887, 0.00001544, -0.00007565, -0.00009915, -0.00005090, 0.00001297, 0.00004684, 0.00005072, 0.00003765, 0.00000162, -0.00005703, -0.00009395, -0.00005181, 0.00005756, 0.00013426, 0.00008969, -0.00004071, -0.00011989, -0.00006047, 0.00006971, 0.00012165, 0.00003665, -0.00008884, -0.00012033, -0.00003931, 0.00004876, 0.00004938, -0.00001592, -0.00005423, -0.00002259, 0.00002959, 0.00003601, -0.00000351, -0.00003510, -0.00002686, -0.00000282, 0.00000048, -0.00001455, -0.00001561, 0.00001079, 0.00004068, 0.00004440, 0.00002119, },
		{ -0.00001020, -0.00000496, -0.00000642, -0.00000819, -0.00000612, 0.00000025, 0.00000885, 0.00001253, 0.00000163, -0.00002286, -0.00004300, -0.00004112, -0.00002326, -0.00001331, -0.00001826, -0.00001196, 0.00003077, 0.00008804, 0.00010005, 0.00003615, -0.00005928, -0.00010558, -0.00006949, 0.00000494, 0.00005201, 0.00005151, 0.00002788, 0.00000035, -0.00003482, -0.00006905, -0.00006258, 0.00000768, 0.00009213, 0.00010029, 0.00000720, -0.00009666, -0.00009511, 0.00001620, 0.00011419, 0.00008629, -0.00003799, -0.00012175, -0.00008087, 0.00002399, 0.00006962, 0.00001792, -0.00005141, -0.00005082, 0.00000975, 0.00004852, 0.00002115, -0.00003072, -0.00004311, -0.00001121, 0.00001524, 0.00000440, -0.00001898, -0.00001354, 0.00002162, 0.00004788, 0.00003745, 0.00000402, -0.00001968, -0.00002026, },
		{ -0.00000961, -0.00001449, -0.00000709, 0.00000577, 0.00001344, 0.00000778, -0.00001232, -0.00003693, -0.00004945, -0.00004216, -0.00002500, -0.00001153, 0.00000211, 0.00003198, 0.00007426, 0.00009140, 0.00004810, -0.00003806, -0.00009953, -0.00008442, -0.00001209, 0.00004916, 0.00005571, 0.00002436, -0.00000571, -0.00002364, -0.00004112, -0.00005039, -0.00002244, 0.00004099, 0.00008107, 0.00004085, -0.00005246, -0.00009826, -0.00003567, 0.00007350, 0.00010576, 0.00001982, -0.00009066, -0.00010518, -0.00001681, 0.00006538, 0.00005126, -0.00002694, -0.00006548, -0.00002127, 0.00004241, 0.00004482, -0.00001102, -0.00005120, -0.00003022, 0.00001638, 0.00002685, -0.00000527, -0.00002890, -0.00000685, 0.00003706, 0.00005178, 0.00002264, -0.00001626, -0.00002779, -0.00001210, 0.00000359, 0.00000187, },
		{ -0.00000370, 0.00001444, 0.00001641, -0.00000127, -0.00002959, -0.00005261, -0.00005661, -0.00004014, -0.00001393, 0.00001232, 0.00003927, 0.00006721, 0.00007923, 0.00004987, -0.00001960, -0.00008399, -0.00008959, -0.00003147, 0.00003637, 0.00005665, 0.00002678, -0.00000958, -0.00002131, -0.00001987, -0.00002581, -0.00002816, -0.00000185, 0.00004061, 0.00004705, -0.00000734, -0.00007080, -0.00006425, 0.00001834, 0.00008819, 0.00006088, -0.00003855, -0.00009918, -0.00005497, 0.00003682, 0.00006751, 0.00000912, -0.00005740, -0.00004878, 0.00001802, 0.00005452, 0.00001671, -0.00004293, -0.00004858, 0.00000182, 0.00003997, 0.00001951, -0.00002653, -0.00003329, 0.00001132, 0.00005402, 0.00004531, -0.00000204, -0.00003455, -0.00002392, 0.00000632, 0.00001716, 0.00000013, -0.00002060, -0.00002157, },
		{ 0.00001337, -0.00001967, -0.00005272, -0.00006699, -0.00005473, -0.00002227, 0.00001565, 0.00004696, 0.00006666, 0.00006940, 0.00004491, -0.00000837, -0.00006535, -0.00008404, -0.00004596, 0.00001702, 0.00004960, 0.00003037, -0.00000795, -0.00002166, -0.00000822, -0.00000064, -0.00001468, -0.00002485, -0.00000485, 0.00002544, 0.00001965, -0.00002756, -0.00006015, -0.00002731, 0.00004345, 0.00006919, 0.00001290, -0.00006405, -0.00007201, -0.00000317, 0.00005800, 0.00003972, -0.00002899, -0.00005851, -0.00001386, 0.00004363, 0.00003897, -0.00001955, -0.00005423, -0.00002074, 0.00003591, 0.00004256, -0.00000682, -0.00004535, -0.00002087, 0.00003841, 0.00006139, 0.00002248, -0.00003095, -0.00004021, -0.00000364, 0.00002916, 0.00002058, -0.00001517, -0.00003625, -0.00002233, 0.00000880, 0.00002572, },
		{ -0.00007465, -0.00006771, -0.00003290, 0.00001295, 0.00005084, 0.00006884, 0.00006422, 0.00003848, -0.00000410, -0.00004934, -0.00007147, -0.00005159, -0.00000258, 0.00003443, 0.00002953, -0.00000252, -0.00001933, -0.00000307, 0.00001773, 0.00000739, -0.00002628, -0.00003965, -0.00001280, 0.00001929, 0.00001101, -0.00002957, -0.00004586, -0.00000730, 0.00004401, 0.00004247, -0.00001494, -0.00005935, -0.00003539, 0.00002711, 0.00004993, 0.00000653, -0.00004470, -0.00003737, 0.00001647, 0.00004383, 0.00000830, -0.00004184, -0.00003816, 0.00001664, 0.00005152, 0.00002032, -0.00003718, -0.00004588, 0.00000841, 0.00006073, 0.00004843, -0.00001252, -0.00005004, -0.00002448, 0.00002764, 0.00004266, 0.00000502, -0.00004057, -0.00004528, -0.00000838, 0.00002920, 0.00003095, -0.00000412, -0.00004898, },
		{ 0.00000686, 0.00005002, 0.00006961, 0.00006207, 0.00003414, -0.00000321, -0.00003814, -0.00005716, -0.00004901, -0.00001701, 0.00001520, 0.00002146, 0.00000209, -0.00001292, 0.00000096, 0.00002826, 0.00002813, -0.00001272, -0.00005546, -0.00005165, -0.00000494, 0.00002898, 0.00001026, -0.00003365, -0.00004219, -0.00000048, 0.00003968, 0.00002659, -0.00002290, -0.00004433, -0.00000987, 0.00003503, 0.00003186, -0.00001367, -0.00004019, -0.00001344, 0.00002811, 0.00002635, -0.00001643, -0.00003962, -0.00000690, 0.00004172, 0.00004062, -0.00001330, -0.00005244, -0.00002402, 0.00004058, 0.00006316, 0.00001597, -0.00004409, -0.00004600, 0.00000952, 0.00005367, 0.00003365, -0.00002735, -0.00006164, -0.00003513, 0.00001987, 0.00004550, 0.00001775, -0.00003812, -0.00007892, -0.00007975, -0.00004418, },
		{ 0.00005959, 0.00003196, -0.00000208, -0.00003017, -0.00004456, -0.00004194, -0.00002447, -0.00000287, 0.00000757, 0.00000185, -0.00000493, 0.00000737, 0.00003443, 0.00004258, 0.00000601, -0.00005318, -0.00007819, -0.00004033, 0.00002157, 0.00004041, -0.00000109, -0.00004886, -0.00004286, 0.00000931, 0.00004333, 0.00002008, -0.00002623, -0.00003469, 0.00000307, 0.00003488, 0.00001925, -0.00002046, -0.00003041, -0.00000020, 0.00002492, 0.00000904, -0.00002319, -0.00002101, 0.00001900, 0.00004340, 0.00001335, -0.00003823, -0.00004440, 0.00000833, 0.00005781, 0.00004216, -0.00002122, -0.00005529, -0.00001816, 0.00004511, 0.00005730, 0.00000149, -0.00006053, -0.00006162, -0.00000397, 0.00004835, 0.00004127, -0.00001772, -0.00007623, -0.00009044, -0.00005653, -0.00000094, 0.00004612, 0.00006715, },
		{ -0.00002236, -0.00003367, -0.00003400, -0.00002681, -0.00001657, -0.00000816, -0.00000404, 0.00000081, 0.00001599, 0.00004028, 0.00005115, 0.00002233, -0.00004003, -0.00008750, -0.00007242, -0.00000454, 0.00004916, 0.00003400, -0.00002907, -0.00006507, -0.00003188, 0.00003056, 0.00004991, 0.00000922, -0.00003577, -0.00002952, 0.00001496, 0.00003786, 0.00001206, -0.00002515, -0.00002595, 0.00000465, 0.00002086, 0.00000271, -0.00001722, -0.00000349, 0.00002769, 0.00002840, -0.00001165, -0.00004366, -0.00002152, 0.00003313, 0.00005272, 0.00000963, -0.00004466, -0.00004136, 0.00001856, 0.00006260, 0.00003417, -0.00003830, -0.00007444, -0.00003539, 0.00003373, 0.00005708, 0.00000984, -0.00006233, -0.00009630, -0.00007018, -0.00001091, 0.00004074, 0.00006234, 0.00005455, 0.00002943, 0.00000035, },
		},
		{ { 0.00001689, 0.00001782, 0.00001442, -0.00000144, -0.00002369, -0.00003489, -0.00002668, -0.00000845, 0.00000922, 0.00002789, 0.00004534, 0.00003969, -0.00000354, -0.00005154, -0.00004744, 0.00001139, 0.00005319, 0.00002064, -0.00004709, -0.00006461, -0.00001757, 0.00002468, 0.00001494, -0.00001323, -0.00001589, -0.00000966, -0.00002537, -0.00003622, 0.00000094, 0.00005446, 0.00004517, -0.00002994, -0.00007559, -0.00002994, 0.00004517, 0.00005446, 0.00000094, -0.00003622, -0.00002537, -0.00000966, -0.00001589, -0.00001323, 0.00001494, 0.00002468, -0.00001757, -0.00006461, -0.00004709, 0.00002064, 0.00005319, 0.00001139, -0.00004744, -0.00005154, -0.00000354, 0.00003969, 0.00004534, 0.00002789, 0.00000922, -0.00000845, -0.00002668, -0.00003489, -0.00002369, -0.00000144, 0.00001442, 0.00001782, },
		{ 0.00001141, -0.00000839, -0.00003018, -0.00003676, -0.00002322, 0.00000023, 0.00002228, 0.00003858, 0.00004220, 0.00002009, -0.00002400, -0.00005232, -0.00002807, 0.00002722, 0.00004428, -0.00000526, -0.00006093, -0.00005030, 0.00000881, 0.00003360, -0.00000280, -0.00003782, -0.00002316, 0.00000211, -0.00001340, -0.00003836, -0.00000934, 0.00005100, 0.00005155, -0.00002522, -0.00008095, -0.00003820, 0.00004689, 0.00006602, 0.00000762, -0.00004483, -0.00004015, -0.00001063, 0.00000389, 0.00001045, 0.00001848, 0.00000382, -0.00004053, -0.00006394, -0.00002250, 0.00004279, 0.00004954, -0.00001167, -0.00006170, -0.00003997, 0.00002175, 0.00005408, 0.00003930, 0.00001220, -0.00000240, -0.00001276, -0.00002676, -0.00003178, -0.00001740, 0.00000473, 0.00001660, 0.00001698, 0.00001669, 0.00001808, },
		{ -0.00003608, -0.00003597, -0.00001571, 0.00001168, 0.00003343, 0.00004214, 0.00003164, 0.00000001, -0.00003645, -0.00004442, -0.00000938, 0.00003255, 0.00002639, -0.00002727, -0.00006035, -0.00002584, 0.00002982, 0.00002726, -0.00002900, -0.00005529, -0.00001692, 0.00001797, -0.00000984, -0.00004874, -0.00001632, 0.00005741, 0.00006015, -0.00002969, -0.00009481, -0.00004286, 0.00005912, 0.00008010, 0.00000472, -0.00006231, -0.00004998, 0.00000287, 0.00003068, 0.00002483, 0.00000648, -0.00002154, -0.00005198, -0.00004801, 0.00000565, 0.00005445, 0.00003349, -0.00003614, -0.00006605, -0.00001844, 0.00004691, 0.00005938, 0.00002456, -0.00000495, -0.00000968, -0.00001246, -0.00002497, -0.00002872, -0.00001150, 0.00001041, 0.00001803, 0.00001545, 0.00001608, 0.00001780, 0.00000742, -0.00001612, },
		{ -0.00000466, 0.00002328, 0.00003980, 0.00003838, 0.00001710, -0.00001636, -0.00004028, -0.00003203, 0.00000377, 0.00002754, 0.00000565, -0.00003927, -0.00004617, 0.00000104, 0.00003876, 0.00000725, -0.00005567, -0.00006050, -0.00000061, 0.00003072, -0.00001644, -0.00006172, -0.00001488, 0.00007115, 0.00006401, -0.00004532, -0.00011130, -0.00003798, 0.00007985, 0.00008935, -0.00001047, -0.00008262, -0.00004764, 0.00002981, 0.00005599, 0.00002403, -0.00001718, -0.00004198, -0.00004721, -0.00002141, 0.00002886, 0.00005195, 0.00000901, -0.00005547, -0.00005816, 0.00000938, 0.00006657, 0.00005367, 0.00000378, -0.00001962, -0.00001094, -0.00000868, -0.00002316, -0.00002639, -0.00000565, 0.00001590, 0.00001855, 0.00001304, 0.00001507, 0.00001698, 0.00000239, -0.00002428, -0.00004029, -0.00003142, },
		{ 0.00003981, 0.00002908, 0.00000264, -0.00002648, -0.00003747, -0.00002001, 0.00000887, 0.00001561, -0.00001135, -0.00003848, -0.00002319, 0.00002234, 0.00003290, -0.00002086, -0.00007527, -0.00005253, 0.00001955, 0.00003514, -0.00003082, -0.00007087, -0.00000345, 0.00008646, 0.00005767, -0.00006971, -0.00012309, -0.00002113, 0.00010311, 0.00008731, -0.00003600, -0.00009747, -0.00002911, 0.00006443, 0.00007102, 0.00000695, -0.00004435, -0.00004915, -0.00002663, 0.00000764, 0.00003996, 0.00003589, -0.00001731, -0.00006436, -0.00003899, 0.00003800, 0.00007636, 0.00003756, -0.00001890, -0.00002841, -0.00000613, -0.00000353, -0.00002287, -0.00002468, 0.00000085, 0.00002128, 0.00001773, 0.00000952, 0.00001380, 0.00001571, -0.00000356, -0.00003218, -0.00004162, -0.00002272, 0.00000833, 0.00003215, },
		{ -0.00000825, -0.00003019, -0.00003170, -0.00001221, 0.00000653, 0.00000215, -0.00001967, -0.00002620, 0.00000146, 0.00003220, 0.00001399, -0.00004959, -0.00008304, -0.00003453, 0.00003685, 0.00002957, -0.00004775, -0.00007140, 0.00001542, 0.00009653, 0.00003926, -0.00009686, -0.00012408, 0.00000557, 0.00012108, 0.00007090, -0.00006562, -0.00009904, 0.00000480, 0.00009787, 0.00006970, -0.00002207, -0.00006503, -0.00003858, 0.00000414, 0.00002985, 0.00003541, 0.00001103, -0.00003826, -0.00006039, -0.00001253, 0.00006155, 0.00007417, 0.00001397, -0.00003893, -0.00002940, 0.00000318, 0.00000054, -0.00002482, -0.00002275, 0.00000864, 0.00002615, 0.00001493, 0.00000487, 0.00001261, 0.00001416, -0.00001016, -0.00003893, -0.00003919, -0.00001045, 0.00002085, 0.00003588, 0.00003361, 0.00001744, },
		{ -0.00002701, -0.00001018, 0.00000008, -0.00000742, -0.00001761, -0.00000712, 0.00002078, 0.00002855, -0.00001271, -0.00007206, -0.00007812, -0.00001218, 0.00004647, 0.00001612, -0.00006128, -0.00006189, 0.00003620, 0.00009607, 0.00001120, -0.00011935, -0.00011168, 0.00003600, 0.00012678, 0.00004180, -0.00009061, -0.00008270, 0.00004837, 0.00012095, 0.00005094, -0.00005429, -0.00007061, -0.00001139, 0.00003581, 0.00003771, 0.00001647, -0.00001525, -0.00004830, -0.00004474, 0.00001538, 0.00007544, 0.00006058, -0.00001259, -0.00005245, -0.00002252, 0.00001433, 0.00000150, -0.00002869, -0.00001926, 0.00001795, 0.00002957, 0.00000951, -0.00000057, 0.00001206, 0.00001255, -0.00001706, -0.00004370, -0.00003267, 0.00000394, 0.00003029, 0.00003322, 0.00002312, 0.00000711, -0.00001376, -0.00002957, },
		{ -0.00000569, -0.00000952, -0.00000708, 0.00001238, 0.00003002, 0.00001334, -0.00004047, -0.00008404, -0.00006326, 0.00000856, 0.00004682, -0.00000058, -0.00006705, -0.00004464, 0.00005249, 0.00008323, -0.00002071, -0.00013097, -0.00008779, 0.00006213, 0.00011658, 0.00000612, -0.00010262, -0.00004870, 0.00009268, 0.00012734, 0.00001930, -0.00007926, -0.00005684, 0.00002573, 0.00005823, 0.00002797, -0.00001115, -0.00003517, -0.00004513, -0.00002177, 0.00003868, 0.00007750, 0.00003860, -0.00003724, -0.00005720, -0.00000949, 0.00002425, -0.00000167, -0.00003315, -0.00001298, 0.00002826, 0.00003024, 0.00000120, -0.00000596, 0.00001283, 0.00001098, -0.00002393, -0.00004583, -0.00002239, 0.00001839, 0.00003448, 0.00002450, 0.00001143, 0.00000105, -0.00001420, -0.00002806, -0.00002627, -0.00001266, },
		{ 0.00000719, 0.00002648, 0.00002806, -0.00000857, -0.00006338, -0.00008482, -0.00004335, 0.00002342, 0.00003958, -0.00001536, -0.00006373, -0.00002460, 0.00005933, 0.00006011, -0.00004932, -0.00012891, -0.00005824, 0.00007665, 0.00009153, -0.00002750, -0.00009635, -0.00000220, 0.00012851, 0.00011575, -0.00001629, -0.00008824, -0.00002552, 0.00006241, 0.00006360, 0.00000271, -0.00003908, -0.00004301, -0.00003043, 0.00000235, 0.00005283, 0.00006839, 0.00001280, -0.00005589, -0.00005290, 0.00000674, 0.00003036, -0.00000870, -0.00003626, -0.00000321, 0.00003824, 0.00002683, -0.00000963, -0.00001005, 0.00001548, 0.00000932, -0.00003055, -0.00004497, -0.00000930, 0.00003063, 0.00003220, 0.00001153, 0.00000199, 0.00000068, -0.00001192, -0.00002924, -0.00003010, -0.00001608, -0.00000640, -0.00000371, },
		{ 0.00001730, -0.00003136, -0.00007812, -0.00007680, -0.00002330, 0.00003104, 0.00002850, -0.00002458, -0.00005322, -0.00000738, 0.00005492, 0.00003194, -0.00006868, -0.00011459, -0.00003073, 0.00007545, 0.00005700, -0.00005059, -0.00007121, 0.00004847, 0.00014942, 0.00009053, -0.00004534, -0.00007721, 0.00001580, 0.00008773, 0.00004896, -0.00003124, -0.00005887, -0.00003685, -0.00000915, 0.00002154, 0.00005596, 0.00005121, -0.00001201, -0.00006613, -0.00004114, 0.00002280, 0.00003116, -0.00001812, -0.00003607, 0.00000975, 0.00004593, 0.00001847, -0.00002179, -0.00001133, 0.00002016, 0.00000708, -0.00003681, -0.00004104, 0.00000525, 0.00003868, 0.00002351, -0.00000278, -0.00000235, 0.00000539, -0.00001044, -0.00003538, -0.00003651, -0.00001592, 0.00000023, 0.00000717, 0.00001944, 0.00003190, },
		{ -0.00008455, -0.00006392, -0.00000633, 0.00003274, 0.00001753, -0.00002725, -0.00003959, 0.00000287, 0.00004101, 0.00000498, -0.00007593, -0.00009302, -0.00001230, 0.00005871, 0.00002085, -0.00005749, -0.00003123, 0.00009474, 0.00015393, 0.00006021, -0.00005934, -0.00004848, 0.00005591, 0.00009358, 0.00001724, -0.00006437, -0.00006462, -0.00001906, 0.00001207, 0.00003155, 0.00004914, 0.00003044, -0.00003193, -0.00006756, -0.00002477, 0.00003573, 0.00002653, -0.00002767, -0.00003123, 0.00002453, 0.00004921, 0.00000516, -0.00003332, -0.00000852, 0.00002638, 0.00000352, -0.00004254, -0.00003419, 0.00001968, 0.00004116, 0.00000974, -0.00001512, -0.00000016, 0.00001270, -0.00001329, -0.00004651, -0.00004143, -0.00000837, 0.00001333, 0.00001796, 0.00002503, 0.00002882, 0.00000245, -0.00005044, },
		{ 0.00000688, 0.00003128, 0.00000905, -0.00002508, -0.00002711, 0.00000488, 0.00002184, -0.00001569, -0.00007204, -0.00007085, -0.00000715, 0.00003060, -0.00000921, -0.00004680, 0.00001670, 0.00013071, 0.00014586, 0.00003461, -0.00005481, -0.00001021, 0.00008335, 0.00007733, -0.00002362, -0.00008744, -0.00005477, 0.00000456, 0.00002704, 0.00003110, 0.00003576, 0.00001057, -0.00004483, -0.00006154, -0.00000701, 0.00004365, 0.00001771, -0.00003493, -0.00002148, 0.00003885, 0.00004637, -0.00001196, -0.00004184, -0.00000098, 0.00003292, -0.00000217, -0.00004746, -0.00002471, 0.00003249, 0.00003760, -0.00000679, -0.00002262, 0.00000816, 0.00001901, -0.00002267, -0.00006010, -0.00004012, 0.00000797, 0.00002888, 0.00002307, 0.00002214, 0.00002062, -0.00001150, -0.00006388, -0.00008504, -0.00004974, },
		{ 0.00000305, -0.00002114, -0.00001840, 0.00000027, 0.00000225, -0.00002788, -0.00006091, -0.00005396, -0.00001562, -0.00000265, -0.00002797, -0.00002117, 0.00006562, 0.00015513, 0.00013280, 0.00002116, -0.00003464, 0.00002600, 0.00009004, 0.00004265, -0.00006300, -0.00009427, -0.00003249, 0.00002668, 0.00003183, 0.00002201, 0.00002033, -0.00000512, -0.00005048, -0.00005053, 0.00000927, 0.00004604, 0.00000680, -0.00003799, -0.00000779, 0.00005005, 0.00003668, -0.00003063, -0.00004508, 0.00001093, 0.00003801, -0.00001053, -0.00005094, -0.00001299, 0.00004233, 0.00002837, -0.00002329, -0.00002347, 0.00002048, 0.00002064, -0.00003865, -0.00007168, -0.00002877, 0.00003124, 0.00004089, 0.00001853, 0.00001258, 0.00001242, -0.00002130, -0.00007273, -0.00008283, -0.00003592, 0.00001777, 0.00002908, },
		{ -0.00001345, -0.00000759, -0.00001448, -0.00003259, -0.00004749, -0.00004560, -0.00003489, -0.00003519, -0.00003392, 0.00001461, 0.00011135, 0.00017153, 0.00012294, 0.00002212, -0.00000746, 0.00004875, 0.00007382, -0.00000149, -0.00009090, -0.00008359, -0.00000455, 0.00004077, 0.00002589, 0.00000831, 0.00000698, -0.00001518, -0.00005024, -0.00003726, 0.00002224, 0.00004362, -0.00000380, -0.00003595, 0.00000777, 0.00005569, 0.00002073, -0.00004787, -0.00004151, 0.00002567, 0.00003964, -0.00002150, -0.00005212, 0.00000044, 0.00004820, 0.00001464, -0.00003706, -0.00001737, 0.00003349, 0.00001498, -0.00005880, -0.00007608, -0.00000610, 0.00005645, 0.00004330, 0.00000366, 0.00000117, 0.00000895, -0.00002672, -0.00008006, -0.00008016, -0.00002171, 0.00002848, 0.00002669, -0.00000254, -0.00001804, },
		{ -0.00002783, -0.00003291, -0.00003548, -0.00004580, -0.00006089, -0.00006361, -0.00002866, 0.00005634, 0.00015404, 0.00018629, 0.00012157, 0.00003333, 0.00001518, 0.00005061, 0.00003928, -0.00004364, -0.00010092, -0.00005930, 0.00002093, 0.00004312, 0.00001196, -0.00000519, -0.00000175, -0.00002004, -0.00004624, -0.00002395, 0.00003120, 0.00003795, -0.00001207, -0.00002917, 0.00002251, 0.00005425, 0.00000051, -0.00006056, -0.00003085, 0.00004075, 0.00003617, -0.00003420, -0.00004993, 0.00001480, 0.00004940, -0.00000190, -0.00004598, -0.00000547, 0.00004360, 0.00000136, -0.00007872, -0.00006903, 0.00002584, 0.00007667, 0.00003205, -0.00001835, -0.00000586, 0.00001255, -0.00003056, -0.00008914, -0.00007686, -0.00000469, 0.00003999, 0.00002226, -0.00001032, -0.00001628, -0.00000988, -0.00001562, },
		{ -0.00002587, -0.00005241, -0.00009066, -0.00008767, -0.00001479, 0.00010303, 0.00019782, 0.00020532, 0.00012834, 0.00004543, 0.00002271, 0.00003061, -0.00000373, -0.00007336, -0.00009212, -0.00002902, 0.00003702, 0.00003383, -0.00000508, -0.00001456, -0.00000537, -0.00002142, -0.00004053, -0.00001190, 0.00003650, 0.00003089, -0.00001680, -0.00001907, 0.00003376, 0.00004550, -0.00002097, -0.00006623, -0.00001428, 0.00005318, 0.00002685, -0.00004692, -0.00004351, 0.00002902, 0.00004564, -0.00001929, -0.00004882, 0.00000990, 0.00004784, -0.00001868, -0.00009317, -0.00004865, 0.00006174, 0.00008494, 0.00000672, -0.00004114, -0.00000315, 0.00002204, -0.00003742, -0.00010150, -0.00007017, 0.00001753, 0.00005082, 0.00001234, -0.00002150, -0.00001356, -0.00000454, -0.00002299, -0.00003997, -0.00003189, },
		{ -0.00012414, -0.00010920, 0.00000698, 0.00015792, 0.00024843, 0.00023055, 0.00013663, 0.00004700, 0.00000934, -0.00000528, -0.00004321, -0.00008442, -0.00006907, -0.00000157, 0.00004007, 0.00001648, -0.00001991, -0.00001787, -0.00000531, -0.00002143, -0.00003446, -0.00000148, 0.00003904, 0.00002402, -0.00001781, -0.00000782, 0.00003951, 0.00003070, -0.00004031, -0.00006361, 0.00000564, 0.00006017, 0.00001218, -0.00005738, -0.00003246, 0.00004169, 0.00003700, -0.00003557, -0.00004541, 0.00002587, 0.00004465, -0.00004165, -0.00009746, -0.00001628, 0.00009422, 0.00007636, -0.00002893, -0.00005697, 0.00001171, 0.00003280, -0.00005166, -0.00011555, -0.00005579, 0.00004533, 0.00005671, -0.00000629, -0.00003428, -0.00000558, 0.00000431, -0.00003197, -0.00005415, -0.00003039, -0.00001684, -0.00006331, },
		{ 0.00004013, 0.00022759, 0.00030954, 0.00025748, 0.00013518, 0.00002881, -0.00002359, -0.00004600, -0.00006880, -0.00007654, -0.00004009, 0.00001585, 0.00003071, -0.00000319, -0.00002856, -0.00001563, -0.00000401, -0.00002161, -0.00002839, 0.00000765, 0.00003988, 0.00001833, -0.00001583, 0.00000226, 0.00003887, 0.00001225, -0.00005440, -0.00005308, 0.00002551, 0.00005986, -0.00000602, -0.00006323, -0.00001726, 0.00005119, 0.00002403, -0.00004901, -0.00003648, 0.00003967, 0.00003422, -0.00006300, -0.00008879, 0.00002354, 0.00011558, 0.00004975, -0.00006760, -0.00005906, 0.00003704, 0.00003824, -0.00007524, -0.00012654, -0.00002984, 0.00007572, 0.00005182, -0.00003442, -0.00004351, 0.00001197, 0.00001506, -0.00004737, -0.00007229, -0.00002590, -0.00000537, -0.00007888, -0.00016429, -0.00012947, },
		{ 0.00037905, 0.00027507, 0.00011198, -0.00001229, -0.00006756, -0.00007901, -0.00007512, -0.00005522, -0.00001444, 0.00001982, 0.00001327, -0.00001937, -0.00002957, -0.00001027, -0.00000384, -0.00002244, -0.00002189, 0.00001601, 0.00003974, 0.00001409, -0.00001222, 0.00000928, 0.00003224, -0.00000675, -0.00006120, -0.00003657, 0.00004185, 0.00005185, -0.00002486, -0.00006271, 0.00000067, 0.00005595, 0.00000777, -0.00005817, -0.00002344, 0.00004918, 0.00001840, -0.00007822, -0.00006717, 0.00006392, 0.00011985, 0.00000829, -0.00010030, -0.00004381, 0.00006713, 0.00003205, -0.00010629, -0.00012781, 0.00000882, 0.00010220, 0.00003092, -0.00006874, -0.00004182, 0.00004080, 0.00002203, -0.00007430, -0.00009264, -0.00001297, 0.00000999, -0.00010351, -0.00021530, -0.00014605, 0.00009301, 0.00031879, },
		{ 0.00005895, -0.00007120, -0.00010931, -0.00009446, -0.00006340, -0.00002961, 0.00000080, 0.00001159, -0.00000591, -0.00002804, -0.00002424, -0.00000508, -0.00000615, -0.00002327, -0.00001423, 0.00002389, 0.00003893, 0.00001098, -0.00000851, 0.00001211, 0.00002117, -0.00002327, -0.00006013, -0.00001713, 0.00005189, 0.00003740, -0.00004106, -0.00005513, 0.00001906, 0.00005479, -0.00001020, -0.00006212, -0.00000809, 0.00005335, 0.00000019, -0.00008403, -0.00003553, 0.00009727, 0.00010445, -0.00004100, -0.00011867, -0.00001211, 0.00009375, 0.00001057, -0.00013895, -0.00011309, 0.00005708, 0.00011613, -0.00000797, -0.00010152, -0.00002220, 0.00007784, 0.00001662, -0.00011527, -0.00010872, 0.00001443, 0.00002636, -0.00014524, -0.00027928, -0.00015045, 0.00017706, 0.00043418, 0.00044695, 0.00026820, },
		{ -0.00013506, -0.00008862, -0.00004086, -0.00000925, 0.00000293, -0.00000377, -0.00002076, -0.00002821, -0.00001589, -0.00000284, -0.00001078, -0.00002271, -0.00000490, 0.00003113, 0.00003731, 0.00000834, -0.00000601, 0.00001062, 0.00000792, -0.00003495, -0.00005217, 0.00000176, 0.00005426, 0.00001916, -0.00005169, -0.00004125, 0.00003515, 0.00004734, -0.00002791, -0.00006043, 0.00000774, 0.00005219, -0.00001705, -0.00007914, 0.00000088, 0.00011706, 0.00007095, -0.00008889, -0.00011728, 0.00003060, 0.00010851, -0.00002553, -0.00016473, -0.00007905, 0.00010731, 0.00010931, -0.00006166, -0.00012206, 0.00001890, 0.00011460, -0.00001001, -0.00016739, -0.00011026, 0.00005946, 0.00003493, -0.00021322, -0.00035240, -0.00012783, 0.00030280, 0.00056823, 0.00049573, 0.00022240, -0.00002391, -0.00013510, },
		{ -0.00001814, -0.00000067, -0.00000569, -0.00001942, -0.00002754, -0.00002190, -0.00000845, -0.00000481, -0.00001620, -0.00001941, 0.00000595, 0.00003712, 0.00003453, 0.00000548, -0.00000551, 0.00000560, -0.00000507, -0.00004060, -0.00003949, 0.00001703, 0.00004926, 0.00000056, -0.00005492, -0.00002319, 0.00004633, 0.00003423, -0.00004319, -0.00005327, 0.00002239, 0.00004668, -0.00003040, -0.00006456, 0.00003570, 0.00011943, 0.00002483, -0.00012613, -0.00009514, 0.00007568, 0.00010539, -0.00007088, -0.00017488, -0.00002720, 0.00014879, 0.00007701, -0.00012130, -0.00011962, 0.00007902, 0.00013847, -0.00006319, -0.00022121, -0.00008614, 0.00011899, 0.00002166, -0.00031340, -0.00042226, -0.00005953, 0.00047455, 0.00070506, 0.00050378, 0.00012930, -0.00012658, -0.00018671, -0.00013529, -0.00006533, },
		{ -0.00001879, -0.00002935, -0.00002601, -0.00001314, -0.00000492, -0.00001036, -0.00002025, -0.00001263, 0.00001746, 0.00004108, 0.00003030, 0.00000192, -0.00000720, -0.00000159, -0.00001574, -0.00004034, -0.00002490, 0.00002671, 0.00003876, -0.00001501, -0.00005052, -0.00000404, 0.00005076, 0.00001717, -0.00005403, -0.00004144, 0.00003455, 0.00003832, -0.00003800, -0.00004329, 0.00006293, 0.00010402, -0.00002584, -0.00014572, -0.00005599, 0.00011351, 0.00008268, -0.00011659, -0.00016323, 0.00003557, 0.00017043, 0.00002030, -0.00017381, -0.00008706, 0.00014868, 0.00013596, -0.00014144, -0.00026178, -0.00002874, 0.00018158, -0.00002951, -0.00044382, -0.00046767, 0.00007193, 0.00068568, 0.00081932, 0.00045122, -0.00000873, -0.00023007, -0.00020933, -0.00010830, -0.00003492, -0.00000540, -0.00000516, },
		{ -0.00001923, -0.00000625, -0.00000630, -0.00001732, -0.00002105, -0.00000270, 0.00002830, 0.00004235, 0.00002453, -0.00000255, -0.00001071, -0.00000934, -0.00002279, -0.00003534, -0.00001110, 0.00003031, 0.00002560, -0.00002506, -0.00003989, 0.00001287, 0.00004785, -0.00000129, -0.00005893, -0.00002627, 0.00004333, 0.00002877, -0.00003940, -0.00001954, 0.00007827, 0.00007400, -0.00007226, -0.00014450, -0.00000731, 0.00013607, 0.00004364, -0.00015251, -0.00012840, 0.00009818, 0.00016408, -0.00005302, -0.00020501, -0.00002401, 0.00021263, 0.00009698, -0.00023478, -0.00027205, 0.00006176, 0.00022811, -0.00013139, -0.00059118, -0.00046160, 0.00027633, 0.00091598, 0.00088052, 0.00032647, -0.00017658, -0.00031063, -0.00019242, -0.00006158, -0.00001092, -0.00000862, -0.00001833, -0.00002877, -0.00003060, },
		{ -0.00001132, -0.00002303, -0.00001774, 0.00000910, 0.00003713, 0.00004067, 0.00001737, -0.00000783, -0.00001531, -0.00001623, -0.00002586, -0.00002746, -0.00000007, 0.00002876, 0.00001287, -0.00002855, -0.00002571, 0.00002474, 0.00003846, -0.00001826, -0.00005726, -0.00000955, 0.00004830, 0.00001949, -0.00003548, 0.00000228, 0.00007995, 0.00003520, -0.00010692, -0.00012373, 0.00004162, 0.00013904, -0.00000422, -0.00017015, -0.00007472, 0.00014803, 0.00012734, -0.00012962, -0.00020362, 0.00006169, 0.00025328, 0.00001899, -0.00032555, -0.00023801, 0.00017463, 0.00023543, -0.00028773, -0.00073023, -0.00037721, 0.00054992, 0.00113268, 0.00086003, 0.00013175, -0.00034795, -0.00034597, -0.00013568, -0.00001017, -0.00000539, -0.00002719, -0.00003175, -0.00002967, -0.00002436, -0.00001207, -0.00000403, },
		{ -0.00001088, 0.00002101, 0.00004301, 0.00003624, 0.00000916, -0.00001361, -0.00002010, -0.00002124, -0.00002541, -0.00001863, 0.00000735, 0.00002399, 0.00000302, -0.00002610, -0.00001125, 0.00003000, 0.00002479, -0.00003109, -0.00004944, 0.00000673, 0.00004943, 0.00001147, -0.00002810, 0.00001856, 0.00006896, -0.00000534, -0.00012518, -0.00008850, 0.00008209, 0.00012280, -0.00005119, -0.00016506, -0.00001154, 0.00017427, 0.00006489, -0.00019359, -0.00016490, 0.00015479, 0.00025543, -0.00009064, -0.00039193, -0.00015390, 0.00028871, 0.00018227, -0.00048871, -0.00082682, -0.00019554, 0.00087191, 0.00129547, 0.00073911, -0.00011452, -0.00049046, -0.00032187, -0.00005036, 0.00002753, -0.00002461, -0.00005369, -0.00003606, -0.00001964, -0.00001539, -0.00000919, -0.00000672, -0.00001715, -0.00002540, },
		{ 0.00004573, 0.00002956, 0.00000026, -0.00001948, -0.00002423, -0.00002391, -0.00002242, -0.00001034, 0.00001144, 0.00001827, -0.00000269, -0.00001965, 0.00000055, 0.00002867, 0.00000979, -0.00003797, -0.00003694, 0.00002071, 0.00004710, 0.00000519, -0.00001954, 0.00002717, 0.00004859, -0.00004098, -0.00012601, -0.00004623, 0.00010805, 0.00009207, -0.00008797, -0.00013818, 0.00004911, 0.00017093, -0.00001209, -0.00023052, -0.00009281, 0.00023583, 0.00021098, -0.00021484, -0.00041350, -0.00002594, 0.00037591, 0.00005620, -0.00070940, -0.00084449, 0.00008723, 0.00120510, 0.00136461, 0.00051554, -0.00037937, -0.00057314, -0.00023709, 0.00004273, 0.00003534, -0.00006637, -0.00007614, -0.00002464, -0.00000198, -0.00001018, -0.00001295, -0.00001186, -0.00002066, -0.00002392, -0.00000214, 0.00003145, },
		{ -0.00000905, -0.00002493, -0.00002700, -0.00002431, -0.00001803, -0.00000338, 0.00001327, 0.00001346, -0.00000439, -0.00001187, 0.00000775, 0.00002226, -0.00000351, -0.00003836, -0.00002196, 0.00003094, 0.00004200, 0.00000063, -0.00001190, 0.00002775, 0.00002341, -0.00006686, -0.00011180, -0.00000470, 0.00011728, 0.00005446, -0.00010815, -0.00009561, 0.00009558, 0.00013875, -0.00008848, -0.00023148, 0.00000022, 0.00028596, 0.00012217, -0.00033005, -0.00037724, 0.00012688, 0.00040749, -0.00014064, -0.00091234, -0.00075314, 0.00045500, 0.00150106, 0.00131036, 0.00020711, -0.00062109, -0.00057444, -0.00010518, 0.00011837, 0.00000457, -0.00011985, -0.00008184, 0.00000357, 0.00001581, -0.00001427, -0.00002223, -0.00001539, -0.00001992, -0.00001981, 0.00000642, 0.00003963, 0.00004569, 0.00002115, },
		{ -0.00002796, -0.00002288, -0.00001320, 0.00000228, 0.00001426, 0.00001052, -0.00000345, -0.00000526, 0.00000992, 0.00001327, -0.00001280, -0.00003307, -0.00000699, 0.00003663, 0.00003505, -0.00000253, -0.00000674, 0.00002151, -0.00000180, -0.00008065, -0.00008732, 0.00002978, 0.00011148, 0.00001820, -0.00010972, -0.00004690, 0.00011969, 0.00008507, -0.00014904, -0.00019565, 0.00009592, 0.00029203, 0.00000217, -0.00041169, -0.00028221, 0.00027567, 0.00036171, -0.00038873, -0.00105414, -0.00053776, 0.00087144, 0.00170899, 0.00112131, -0.00014973, -0.00079792, -0.00048839, 0.00004783, 0.00015362, -0.00006262, -0.00016836, -0.00006183, 0.00004305, 0.00002444, -0.00002979, -0.00003259, -0.00001332, -0.00001519, -0.00001559, 0.00001331, 0.00004577, 0.00004358, 0.00001129, -0.00001855, -0.00002936, },
		{ -0.00000846, 0.00000720, 0.00001551, 0.00000930, -0.00000185, -0.00000149, 0.00000807, 0.00000448, -0.00001697, -0.00002397, 0.00000575, 0.00003781, 0.00002721, -0.00000481, -0.00000477, 0.00001074, -0.00002310, -0.00008264, -0.00005829, 0.00005362, 0.00009529, -0.00001006, -0.00009531, -0.00000239, 0.00011878, 0.00002202, -0.00018232, -0.00013083, 0.00017469, 0.00025044, -0.00012742, -0.00044056, -0.00014143, 0.00038770, 0.00023078, -0.00065282, -0.00109470, -0.00020465, 0.00128462, 0.00178643, 0.00080910, -0.00050588, -0.00087757, -0.00032721, 0.00018917, 0.00013432, -0.00015301, -0.00019408, -0.00001437, 0.00008277, 0.00001620, -0.00005420, -0.00003771, -0.00000365, -0.00000911, -0.00001390, 0.00001833, 0.00005085, 0.00003982, -0.00000008, -0.00002794, -0.00003212, -0.00002691, -0.00002026, },
		{ 0.00001744, 0.00000866, -0.00000145, -0.00000090, 0.00000423, -0.00000189, -0.00001635, -0.00001350, 0.00001478, 0.00003522, 0.00001940, -0.00000674, -0.00000593, -0.00000186, -0.00003806, -0.00007511, -0.00002986, 0.00006626, 0.00007460, -0.00002679, -0.00007107, 0.00002955, 0.00009625, -0.00003678, -0.00018363, -0.00005155, 0.00022056, 0.00016851, -0.00024158, -0.00040829, 0.00001994, 0.00043428, 0.00002487, -0.00088773, -0.00100700, 0.00021679, 0.00163608, 0.00170912, 0.00040819, -0.00080914, -0.00084494, -0.00011941, 0.00028721, 0.00005940, -0.00024523, -0.00018343, 0.00005369, 0.00010936, -0.00001202, -0.00008044, -0.00003173, 0.00001241, -0.00000585, -0.00001632, 0.00002276, 0.00005600, 0.00003418, -0.00001328, -0.00003663, -0.00003245, -0.00002398, -0.00001711, -0.00000388, 0.00001216, },
		{ -0.00000335, -0.00000253, 0.00000072, -0.00000479, -0.00001245, -0.00000398, 0.00001962, 0.00003013, 0.00001226, -0.00000875, -0.00000949, -0.00001385, -0.00004590, -0.00006143, -0.00000563, 0.00006963, 0.00005475, -0.00003209, -0.00004461, 0.00004463, 0.00006022, -0.00007984, -0.00015608, 0.00002476, 0.00022542, 0.00006298, -0.00031808, -0.00032016, 0.00016996, 0.00039836, -0.00022804, -0.00104762, -0.00078472, 0.00067761, 0.00187220, 0.00147747, -0.00002976, -0.00101503, -0.00070601, 0.00009631, 0.00031927, -0.00005819, -0.00031534, -0.00013153, 0.00012813, 0.00011122, -0.00005699, -0.00009896, -0.00001164, 0.00003032, -0.00000959, -0.00002231, 0.00002893, 0.00006167, 0.00002567, -0.00002841, -0.00004357, -0.00002969, -0.00001958, -0.00001394, 0.00000082, 0.00001772, 0.00001959, 0.00000703, },
		{ -0.00000068, -0.00000451, -0.00000744, 0.00000303, 0.00002081, 0.00002391, 0.00000609, -0.00001112, -0.00001442, -0.00002344, -0.00004724, -0.00004498, 0.00001282, 0.00006699, 0.00003920, -0.00002905, -0.00002274, 0.00004335, 0.00002107, -0.00010062, -0.00010940, 0.00008236, 0.00019098, -0.00004421, -0.00034321, -0.00019444, 0.00027734, 0.00027976, -0.00048690, -0.00109612, -0.00044567, 0.00111775, 0.00195541, 0.00111775, -0.00044567, -0.00109612, -0.00048690, 0.00027976, 0.00027734, -0.00019444, -0.00034321, -0.00004421, 0.00019098, 0.00008236, -0.00010940, -0.00010062, 0.00002107, 0.00004335, -0.00002274, -0.00002905, 0.00003920, 0.00006699, 0.00001282, -0.00004498, -0.00004724, -0.00002344, -0.00001442, -0.00001112, 0.00000609, 0.00002391, 0.00002081, 0.00000303, -0.00000744, -0.00000451, },
		{ -0.00000335, 0.00000703, 0.00001959, 0.00001772, 0.00000082, -0.00001394, -0.00001958, -0.00002969, -0.00004357, -0.00002841, 0.00002567, 0.00006167, 0.00002893, -0.00002231, -0.00000959, 0.00003032, -0.00001164, -0.00009896, -0.00005699, 0.00011122, 0.00012813, -0.00013153, -0.00031534, -0.00005819, 0.00031927, 0.00009631, -0.00070601, -0.00101503, -0.00002976, 0.00147747, 0.00187220, 0.00067761, -0.00078472, -0.00104762, -0.00022804, 0.00039836, 0.00016996, -0.00032016, -0.00031808, 0.00006298, 0.00022542, 0.00002476, -0.00015608, -0.00007984, 0.00006022, 0.00004463, -0.00004461, -0.00003209, 0.00005475, 0.00006963, -0.00000563, -0.00006143, -0.00004590, -0.00001385, -0.00000949, -0.00000875, 0.00001226, 0.00003013, 0.00001962, -0.00000398, -0.00001245, -0.00000479, 0.00000072, -0.00000253, },
		{ 0.00001744, 0.00001216, -0.00000388, -0.00001711, -0.00002398, -0.00003245, -0.00003663, -0.00001328, 0.00003418, 0.00005600, 0.00002276, -0.00001632, -0.00000585, 0.00001241, -0.00003173, -0.00008044, -0.00001202, 0.00010936, 0.00005369, -0.00018343, -0.00024523, 0.00005940, 0.00028721, -0.00011941, -0.00084494, -0.00080914, 0.00040819, 0.00170912, 0.00163608, 0.00021679, -0.00100700, -0.00088773, 0.00002487, 0.00043428, 0.00001994, -0.00040829, -0.00024158, 0.00016851, 0.00022056, -0.00005155, -0.00018363, -0.00003678, 0.00009625, 0.00002955, -0.00007107, -0.00002679, 0.00007460, 0.00006626, -0.00002986, -0.00007511, -0.00003806, -0.00000186, -0.00000593, -0.00000674, 0.00001940, 0.00003522, 0.00001478, -0.00001350, -0.00001635, -0.00000189, 0.00000423, -0.00000090, -0.00000145, 0.00000866, },
		{ -0.00000846, -0.00002026, -0.00002691, -0.00003212, -0.00002794, -0.00000008, 0.00003982, 0.00005085, 0.00001833, -0.00001390, -0.00000911, -0.00000365, -0.00003771, -0.00005420, 0.00001620, 0.00008277, -0.00001437, -0.00019408, -0.00015301, 0.00013432, 0.00018917, -0.00032721, -0.00087757, -0.00050588, 0.00080910, 0.00178643, 0.00128462, -0.00020465, -0.00109470, -0.00065282, 0.00023078, 0.00038770, -0.00014143, -0.00044056, -0.00012742, 0.00025044, 0.00017469, -0.00013083, -0.00018232, 0.00002202, 0.00011878, -0.00000239, -0.00009531, -0.00001006, 0.00009529, 0.00005362, -0.00005829, -0.00008264, -0.00002310, 0.00001074, -0.00000477, -0.00000481, 0.00002721, 0.00003781, 0.00000575, -0.00002397, -0.00001697, 0.00000448, 0.00000807, -0.00000149, -0.00000185, 0.00000930, 0.00001551, 0.00000720, },
		{ -0.00002796, -0.00002936, -0.00001855, 0.00001129, 0.00004358, 0.00004577, 0.00001331, -0.00001559, -0.00001519, -0.00001332, -0.00003259, -0.00002979, 0.00002444, 0.00004305, -0.00006183, -0.00016836, -0.00006262, 0.00015362, 0.00004783, -0.00048839, -0.00079792, -0.00014973, 0.00112131, 0.00170899, 0.00087144, -0.00053776, -0.00105414, -0.00038873, 0.00036171, 0.00027567, -0.00028221, -0.00041169, 0.00000217, 0.00029203, 0.00009592, -0.00019565, -0.00014904, 0.00008507, 0.00011969, -0.00004690, -0.00010972, 0.00001820, 0.00011148, 0.00002978, -0.00008732, -0.00008065, -0.00000180, 0.00002151, -0.00000674, -0.00000253, 0.00003505, 0.00003663, -0.00000699, -0.00003307, -0.00001280, 0.00001327, 0.00000992, -0.00000526, -0.00000345, 0.00001052, 0.00001426, 0.00000228, -0.00001320, -0.00002288, },
		{ -0.00000905, 0.00002115, 0.00004569, 0.00003963, 0.00000642, -0.00001981, -0.00001992, -0.00001539, -0.00002223, -0.00001427, 0.00001581, 0.00000357, -0.00008184, -0.00011985, 0.00000457, 0.00011837, -0.00010518, -0.00057444, -0.00062109, 0.00020711, 0.00131036, 0.00150106, 0.00045500, -0.00075314, -0.00091234, -0.00014064, 0.00040749, 0.00012688, -0.00037724, -0.00033005, 0.00012217, 0.00028596, 0.00000022, -0.00023148, -0.00008848, 0.00013875, 0.00009558, -0.00009561, -0.00010815, 0.00005446, 0.00011728, -0.00000470, -0.00011180, -0.00006686, 0.00002341, 0.00002775, -0.00001190, 0.00000063, 0.00004200, 0.00003094, -0.00002196, -0.00003836, -0.00000351, 0.00002226, 0.00000775, -0.00001187, -0.00000439, 0.00001346, 0.00001327, -0.00000338, -0.00001803, -0.00002431, -0.00002700, -0.00002493, },
		{ 0.00004573, 0.00003145, -0.00000214, -0.00002392, -0.00002066, -0.00001186, -0.00001295, -0.00001018, -0.00000198, -0.00002464, -0.00007614, -0.00006637, 0.00003534, 0.00004273, -0.00023709, -0.00057314, -0.00037937, 0.00051554, 0.00136461, 0.00120510, 0.00008723, -0.00084449, -0.00070940, 0.00005620, 0.00037591, -0.00002594, -0.00041350, -0.00021484, 0.00021098, 0.00023583, -0.00009281, -0.00023052, -0.00001209, 0.00017093, 0.00004911, -0.00013818, -0.00008797, 0.00009207, 0.00010805, -0.00004623, -0.00012601, -0.00004098, 0.00004859, 0.00002717, -0.00001954, 0.00000519, 0.00004710, 0.00002071, -0.00003694, -0.00003797, 0.00000979, 0.00002867, 0.00000055, -0.00001965, -0.00000269, 0.00001827, 0.00001144, -0.00001034, -0.00002242, -0.00002391, -0.00002423, -0.00001948, 0.00000026, 0.00002956, },
		{ -0.00001088, -0.00002540, -0.00001715, -0.00000672, -0.00000919, -0.00001539, -0.00001964, -0.00003606, -0.00005369, -0.00002461, 0.00002753, -0.00005036, -0.00032187, -0.00049046, -0.00011452, 0.00073911, 0.00129547, 0.00087191, -0.00019554, -0.00082682, -0.00048871, 0.00018227, 0.00028871, -0.00015390, -0.00039193, -0.00009064, 0.00025543, 0.00015479, -0.00016490, -0.00019359, 0.00006489, 0.00017427, -0.00001154, -0.00016506, -0.00005119, 0.00012280, 0.00008209, -0.00008850, -0.00012518, -0.00000534, 0.00006896, 0.00001856, -0.00002810, 0.00001147, 0.00004943, 0.00000673, -0.00004944, -0.00003109, 0.00002479, 0.00003000, -0.00001125, -0.00002610, 0.00000302, 0.00002399, 0.00000735, -0.00001863, -0.00002541, -0.00002124, -0.00002010, -0.00001361, 0.00000916, 0.00003624, 0.00004301, 0.00002101, },
		{ -0.00001132, -0.00000403, -0.00001207, -0.00002436, -0.00002967, -0.00003175, -0.00002719, -0.00000539, -0.00001017, -0.00013568, -0.00034597, -0.00034795, 0.00013175, 0.00086003, 0.00113268, 0.00054992, -0.00037721, -0.00073023, -0.00028773, 0.00023543, 0.00017463, -0.00023801, -0.00032555, 0.00001899, 0.00025328, 0.00006169, -0.00020362, -0.00012962, 0.00012734, 0.00014803, -0.00007472, -0.00017015, -0.00000422, 0.00013904, 0.00004162, -0.00012373, -0.00010692, 0.00003520, 0.00007995, 0.00000228, -0.00003548, 0.00001949, 0.00004830, -0.00000955, -0.00005726, -0.00001826, 0.00003846, 0.00002474, -0.00002571, -0.00002855, 0.00001287, 0.00002876, -0.00000007, -0.00002746, -0.00002586, -0.00001623, -0.00001531, -0.00000783, 0.00001737, 0.00004067, 0.00003713, 0.00000910, -0.00001774, -0.00002303, },
		{ -0.00001923, -0.00003060, -0.00002877, -0.00001833, -0.00000862, -0.00001092, -0.00006158, -0.00019242, -0.00031063, -0.00017658, 0.00032647, 0.00088052, 0.00091598, 0.00027633, -0.00046160, -0.00059118, -0.00013139, 0.00022811, 0.00006176, -0.00027205, -0.00023478, 0.00009698, 0.00021263, -0.00002401, -0.00020501, -0.00005302, 0.00016408, 0.00009818, -0.00012840, -0.00015251, 0.00004364, 0.00013607, -0.00000731, -0.00014450, -0.00007226, 0.00007400, 0.00007827, -0.00001954, -0.00003940, 0.00002877, 0.00004333, -0.00002627, -0.00005893, -0.00000129, 0.00004785, 0.00001287, -0.00003989, -0.00002506, 0.00002560, 0.00003031, -0.00001110, -0.00003534, -0.00002279, -0.00000934, -0.00001071, -0.00000255, 0.00002453, 0.00004235, 0.00002830, -0.00000270, -0.00002105, -0.00001732, -0.00000630, -0.00000625, },
		{ -0.00001879, -0.00000516, -0.00000540, -0.00003492, -0.00010830, -0.00020933, -0.00023007, -0.00000873, 0.00045122, 0.00081932, 0.00068568, 0.00007193, -0.00046767, -0.00044382, -0.00002951, 0.00018158, -0.00002874, -0.00026178, -0.00014144, 0.00013596, 0.00014868, -0.00008706, -0.00017381, 0.00002030, 0.00017043, 0.00003557, -0.00016323, -0.00011659, 0.00008268, 0.00011351, -0.00005599, -0.00014572, -0.00002584, 0.00010402, 0.00006293, -0.00004329, -0.00003800, 0.00003832, 0.00003455, -0.00004144, -0.00005403, 0.00001717, 0.00005076, -0.00000404, -0.00005052, -0.00001501, 0.00003876, 0.00002671, -0.00002490, -0.00004034, -0.00001574, -0.00000159, -0.00000720, 0.00000192, 0.00003030, 0.00004108, 0.00001746, -0.00001263, -0.00002025, -0.00001036, -0.00000492, -0.00001314, -0.00002601, -0.00002935, },
		{ -0.00001814, -0.00006533, -0.00013529, -0.00018671, -0.00012658, 0.00012930, 0.00050378, 0.00070506, 0.00047455, -0.00005953, -0.00042226, -0.00031340, 0.00002166, 0.00011899, -0.00008614, -0.00022121, -0.00006319, 0.00013847, 0.00007902, -0.00011962, -0.00012130, 0.00007701, 0.00014879, -0.00002720, -0.00017488, -0.00007088, 0.00010539, 0.00007568, -0.00009514, -0.00012613, 0.00002483, 0.00011943, 0.00003570, -0.00006456, -0.00003040, 0.00004668, 0.00002239, -0.00005327, -0.00004319, 0.00003423, 0.00004633, -0.00002319, -0.00005492, 0.00000056, 0.00004926, 0.00001703, -0.00003949, -0.00004060, -0.00000507, 0.00000560, -0.00000551, 0.00000548, 0.00003453, 0.00003712, 0.00000595, -0.00001941, -0.00001620, -0.00000481, -0.00000845, -0.00002190, -0.00002754, -0.00001942, -0.00000569, -0.00000067, },
		{ -0.00013506, -0.00013510, -0.00002391, 0.00022240, 0.00049573, 0.00056823, 0.00030280, -0.00012783, -0.00035240, -0.00021322, 0.00003493, 0.00005946, -0.00011026, -0.00016739, -0.00001001, 0.00011460, 0.00001890, -0.00012206, -0.00006166, 0.00010931, 0.00010731, -0.00007905, -0.00016473, -0.00002553, 0.00010851, 0.00003060, -0.00011728, -0.00008889, 0.00007095, 0.00011706, 0.00000088, -0.00007914, -0.00001705, 0.00005219, 0.00000774, -0.00006043, -0.00002791, 0.00004734, 0.00003515, -0.00004125, -0.00005169, 0.00001916, 0.00005426, 0.00000176, -0.00005217, -0.00003495, 0.00000792, 0.00001062, -0.00000601, 0.00000834, 0.00003731, 0.00003113, -0.00000490, -0.00002271, -0.00001078, -0.00000284, -0.00001589, -0.00002821, -0.00002076, -0.00000377, 0.00000293, -0.00000925, -0.00004086, -0.00008862, },
		{ 0.00005895, 0.00026820, 0.00044695, 0.00043418, 0.00017706, -0.00015045, -0.00027928, -0.00014524, 0.00002636, 0.00001443, -0.00010872, -0.00011527, 0.00001662, 0.00007784, -0.00002220, -0.00010152, -0.00000797, 0.00011613, 0.00005708, -0.00011309, -0.00013895, 0.00001057, 0.00009375, -0.00001211, -0.00011867, -0.00004100, 0.00010445, 0.00009727, -0.00003553, -0.00008403, 0.00000019, 0.00005335, -0.00000809, -0.00006212, -0.00001020, 0.00005479, 0.00001906, -0.00005513, -0.00004106, 0.00003740, 0.00005189, -0.00001713, -0.00006013, -0.00002327, 0.00002117, 0.00001211, -0.00000851, 0.00001098, 0.00003893, 0.00002389, -0.00001423, -0.00002327, -0.00000615, -0.00000508, -0.00002424, -0.00002804, -0.00000591, 0.00001159, 0.00000080, -0.00002961, -0.00006340, -0.00009446, -0.00010931, -0.00007120, },
		{ 0.00037905, 0.00031879, 0.00009301, -0.00014605, -0.00021530, -0.00010351, 0.00000999, -0.00001297, -0.00009264, -0.00007430, 0.00002203, 0.00004080, -0.00004182, -0.00006874, 0.00003092, 0.00010220, 0.00000882, -0.00012781, -0.00010629, 0.00003205, 0.00006713, -0.00004381, -0.00010030, 0.00000829, 0.00011985, 0.00006392, -0.00006717, -0.00007822, 0.00001840, 0.00004918, -0.00002344, -0.00005817, 0.00000777, 0.00005595, 0.00000067, -0.00006271, -0.00002486, 0.00005185, 0.00004185, -0.00003657, -0.00006120, -0.00000675, 0.00003224, 0.00000928, -0.00001222, 0.00001409, 0.00003974, 0.00001601, -0.00002189, -0.00002244, -0.00000384, -0.00001027, -0.00002957, -0.00001937, 0.00001327, 0.00001982, -0.00001444, -0.00005522, -0.00007512, -0.00007901, -0.00006756, -0.00001229, 0.00011198, 0.00027507, },
		{ 0.00004013, -0.00012947, -0.00016429, -0.00007888, -0.00000537, -0.00002590, -0.00007229, -0.00004737, 0.00001506, 0.00001197, -0.00004351, -0.00003442, 0.00005182, 0.00007572, -0.00002984, -0.00012654, -0.00007524, 0.00003824, 0.00003704, -0.00005906, -0.00006760, 0.00004975, 0.00011558, 0.00002354, -0.00008879, -0.00006300, 0.00003422, 0.00003967, -0.00003648, -0.00004901, 0.00002403, 0.00005119, -0.00001726, -0.00006323, -0.00000602, 0.00005986, 0.00002551, -0.00005308, -0.00005440, 0.00001225, 0.00003887, 0.00000226, -0.00001583, 0.00001833, 0.00003988, 0.00000765, -0.00002839, -0.00002161, -0.00000401, -0.00001563, -0.00002856, -0.00000319, 0.00003071, 0.00001585, -0.00004009, -0.00007654, -0.00006880, -0.00004600, -0.00002359, 0.00002881, 0.00013518, 0.00025748, 0.00030954, 0.00022759, },
		{ -0.00012414, -0.00006331, -0.00001684, -0.00003039, -0.00005415, -0.00003197, 0.00000431, -0.00000558, -0.00003428, -0.00000629, 0.00005671, 0.00004533, -0.00005579, -0.00011555, -0.00005166, 0.00003280, 0.00001171, -0.00005697, -0.00002893, 0.00007636, 0.00009422, -0.00001628, -0.00009746, -0.00004165, 0.00004465, 0.00002587, -0.00004541, -0.00003557, 0.00003700, 0.00004169, -0.00003246, -0.00005738, 0.00001218, 0.00006017, 0.00000564, -0.00006361, -0.00004031, 0.00003070, 0.00003951, -0.00000782, -0.00001781, 0.00002402, 0.00003904, -0.00000148, -0.00003446, -0.00002143, -0.00000531, -0.00001787, -0.00001991, 0.00001648, 0.00004007, -0.00000157, -0.00006907, -0.00008442, -0.00004321, -0.00000528, 0.00000934, 0.00004700, 0.00013663, 0.00023055, 0.00024843, 0.00015792, 0.00000698, -0.00010920, },
		{ -0.00002587, -0.00003189, -0.00003997, -0.00002299, -0.00000454, -0.00001356, -0.00002150, 0.00001234, 0.00005082, 0.00001753, -0.00007017, -0.00010150, -0.00003742, 0.00002204, -0.00000315, -0.00004114, 0.00000672, 0.00008494, 0.00006174, -0.00004865, -0.00009317, -0.00001868, 0.00004784, 0.00000990, -0.00004882, -0.00001929, 0.00004564, 0.00002902, -0.00004351, -0.00004692, 0.00002685, 0.00005318, -0.00001428, -0.00006623, -0.00002097, 0.00004550, 0.00003376, -0.00001907, -0.00001680, 0.00003089, 0.00003650, -0.00001190, -0.00004053, -0.00002142, -0.00000537, -0.00001456, -0.00000508, 0.00003383, 0.00003702, -0.00002902, -0.00009212, -0.00007336, -0.00000373, 0.00003061, 0.00002271, 0.00004543, 0.00012834, 0.00020532, 0.00019782, 0.00010303, -0.00001479, -0.00008767, -0.00009066, -0.00005241, },
		{ -0.00002783, -0.00001562, -0.00000988, -0.00001628, -0.00001032, 0.00002226, 0.00003999, -0.00000469, -0.00007686, -0.00008914, -0.00003056, 0.00001255, -0.00000586, -0.00001835, 0.00003205, 0.00007667, 0.00002584, -0.00006903, -0.00007872, 0.00000136, 0.00004360, -0.00000547, -0.00004598, -0.00000190, 0.00004940, 0.00001480, -0.00004993, -0.00003420, 0.00003617, 0.00004075, -0.00003085, -0.00006056, 0.00000051, 0.00005425, 0.00002251, -0.00002917, -0.00001207, 0.00003795, 0.00003120, -0.00002395, -0.00004624, -0.00002004, -0.00000175, -0.00000519, 0.00001196, 0.00004312, 0.00002093, -0.00005930, -0.00010092, -0.00004364, 0.00003928, 0.00005061, 0.00001518, 0.00003333, 0.00012157, 0.00018629, 0.00015404, 0.00005634, -0.00002866, -0.00006361, -0.00006089, -0.00004580, -0.00003548, -0.00003291, },
		{ -0.00001345, -0.00001804, -0.00000254, 0.00002669, 0.00002848, -0.00002171, -0.00008016, -0.00008006, -0.00002672, 0.00000895, 0.00000117, 0.00000366, 0.00004330, 0.00005645, -0.00000610, -0.00007608, -0.00005880, 0.00001498, 0.00003349, -0.00001737, -0.00003706, 0.00001464, 0.00004820, 0.00000044, -0.00005212, -0.00002150, 0.00003964, 0.00002567, -0.00004151, -0.00004787, 0.00002073, 0.00005569, 0.00000777, -0.00003595, -0.00000380, 0.00004362, 0.00002224, -0.00003726, -0.00005024, -0.00001518, 0.00000698, 0.00000831, 0.00002589, 0.00004077, -0.00000455, -0.00008359, -0.00009090, -0.00000149, 0.00007382, 0.00004875, -0.00000746, 0.00002212, 0.00012294, 0.00017153, 0.00011135, 0.00001461, -0.00003392, -0.00003519, -0.00003489, -0.00004560, -0.00004749, -0.00003259, -0.00001448, -0.00000759, },
		{ 0.00000305, 0.00002908, 0.00001777, -0.00003592, -0.00008283, -0.00007273, -0.00002130, 0.00001242, 0.00001258, 0.00001853, 0.00004089, 0.00003124, -0.00002877, -0.00007168, -0.00003865, 0.00002064, 0.00002048, -0.00002347, -0.00002329, 0.00002837, 0.00004233, -0.00001299, -0.00005094, -0.00001053, 0.00003801, 0.00001093, -0.00004508, -0.00003063, 0.00003668, 0.00005005, -0.00000779, -0.00003799, 0.00000680, 0.00004604, 0.00000927, -0.00005053, -0.00005048, -0.00000512, 0.00002033, 0.00002201, 0.00003183, 0.00002668, -0.00003249, -0.00009427, -0.00006300, 0.00004265, 0.00009004, 0.00002600, -0.00003464, 0.00002116, 0.00013280, 0.00015513, 0.00006562, -0.00002117, -0.00002797, -0.00000265, -0.00001562, -0.00005396, -0.00006091, -0.00002788, 0.00000225, 0.00000027, -0.00001840, -0.00002114, },
		{ 0.00000688, -0.00004974, -0.00008504, -0.00006388, -0.00001150, 0.00002062, 0.00002214, 0.00002307, 0.00002888, 0.00000797, -0.00004012, -0.00006010, -0.00002267, 0.00001901, 0.00000816, -0.00002262, -0.00000679, 0.00003760, 0.00003249, -0.00002471, -0.00004746, -0.00000217, 0.00003292, -0.00000098, -0.00004184, -0.00001196, 0.00004637, 0.00003885, -0.00002148, -0.00003493, 0.00001771, 0.00004365, -0.00000701, -0.00006154, -0.00004483, 0.00001057, 0.00003576, 0.00003110, 0.00002704, 0.00000456, -0.00005477, -0.00008744, -0.00002362, 0.00007733, 0.00008335, -0.00001021, -0.00005481, 0.00003461, 0.00014586, 0.00013071, 0.00001670, -0.00004680, -0.00000921, 0.00003060, -0.00000715, -0.00007085, -0.00007204, -0.00001569, 0.00002184, 0.00000488, -0.00002711, -0.00002508, 0.00000905, 0.00003128, },
		{ -0.00008455, -0.00005044, 0.00000245, 0.00002882, 0.00002503, 0.00001796, 0.00001333, -0.00000837, -0.00004143, -0.00004651, -0.00001329, 0.00001270, -0.00000016, -0.00001512, 0.00000974, 0.00004116, 0.00001968, -0.00003419, -0.00004254, 0.00000352, 0.00002638, -0.00000852, -0.00003332, 0.00000516, 0.00004921, 0.00002453, -0.00003123, -0.00002767, 0.00002653, 0.00003573, -0.00002477, -0.00006756, -0.00003193, 0.00003044, 0.00004914, 0.00003155, 0.00001207, -0.00001906, -0.00006462, -0.00006437, 0.00001724, 0.00009358, 0.00005591, -0.00004848, -0.00005934, 0.00006021, 0.00015393, 0.00009474, -0.00003123, -0.00005749, 0.00002085, 0.00005871, -0.00001230, -0.00009302, -0.00007593, 0.00000498, 0.00004101, 0.00000287, -0.00003959, -0.00002725, 0.00001753, 0.00003274, -0.00000633, -0.00006392, },
		{ 0.00001730, 0.00003190, 0.00001944, 0.00000717, 0.00000023, -0.00001592, -0.00003651, -0.00003538, -0.00001044, 0.00000539, -0.00000235, -0.00000278, 0.00002351, 0.00003868, 0.00000525, -0.00004104, -0.00003681, 0.00000708, 0.00002016, -0.00001133, -0.00002179, 0.00001847, 0.00004593, 0.00000975, -0.00003607, -0.00001812, 0.00003116, 0.00002280, -0.00004114, -0.00006613, -0.00001201, 0.00005121, 0.00005596, 0.00002154, -0.00000915, -0.00003685, -0.00005887, -0.00003124, 0.00004896, 0.00008773, 0.00001580, -0.00007721, -0.00004534, 0.00009053, 0.00014942, 0.00004847, -0.00007121, -0.00005059, 0.00005700, 0.00007545, -0.00003073, -0.00011459, -0.00006868, 0.00003194, 0.00005492, -0.00000738, -0.00005322, -0.00002458, 0.00002850, 0.00003104, -0.00002330, -0.00007680, -0.00007812, -0.00003136, },
		{ 0.00000719, -0.00000371, -0.00000640, -0.00001608, -0.00003010, -0.00002924, -0.00001192, 0.00000068, 0.00000199, 0.00001153, 0.00003220, 0.00003063, -0.00000930, -0.00004497, -0.00003055, 0.00000932, 0.00001548, -0.00001005, -0.00000963, 0.00002683, 0.00003824, -0.00000321, -0.00003626, -0.00000870, 0.00003036, 0.00000674, -0.00005290, -0.00005589, 0.00001280, 0.00006839, 0.00005283, 0.00000235, -0.00003043, -0.00004301, -0.00003908, 0.00000271, 0.00006360, 0.00006241, -0.00002552, -0.00008824, -0.00001629, 0.00011575, 0.00012851, -0.00000220, -0.00009635, -0.00002750, 0.00009153, 0.00007665, -0.00005824, -0.00012891, -0.00004932, 0.00006011, 0.00005933, -0.00002460, -0.00006373, -0.00001536, 0.00003958, 0.00002342, -0.00004335, -0.00008482, -0.00006338, -0.00000857, 0.00002806, 0.00002648, },
		{ -0.00000569, -0.00001266, -0.00002627, -0.00002806, -0.00001420, 0.00000105, 0.00001143, 0.00002450, 0.00003448, 0.00001839, -0.00002239, -0.00004583, -0.00002393, 0.00001098, 0.00001283, -0.00000596, 0.00000120, 0.00003024, 0.00002826, -0.00001298, -0.00003315, -0.00000167, 0.00002425, -0.00000949, -0.00005720, -0.00003724, 0.00003860, 0.00007750, 0.00003868, -0.00002177, -0.00004513, -0.00003517, -0.00001115, 0.00002797, 0.00005823, 0.00002573, -0.00005684, -0.00007926, 0.00001930, 0.00012734, 0.00009268, -0.00004870, -0.00010262, 0.00000612, 0.00011658, 0.00006213, -0.00008779, -0.00013097, -0.00002071, 0.00008323, 0.00005249, -0.00004464, -0.00006705, -0.00000058, 0.00004682, 0.00000856, -0.00006326, -0.00008404, -0.00004047, 0.00001334, 0.00003002, 0.00001238, -0.00000708, -0.00000952, },
		{ -0.00002701, -0.00002957, -0.00001376, 0.00000711, 0.00002312, 0.00003322, 0.00003029, 0.00000394, -0.00003267, -0.00004370, -0.00001706, 0.00001255, 0.00001206, -0.00000057, 0.00000951, 0.00002957, 0.00001795, -0.00001926, -0.00002869, 0.00000150, 0.00001433, -0.00002252, -0.00005245, -0.00001259, 0.00006058, 0.00007544, 0.00001538, -0.00004474, -0.00004830, -0.00001525, 0.00001647, 0.00003771, 0.00003581, -0.00001139, -0.00007061, -0.00005429, 0.00005094, 0.00012095, 0.00004837, -0.00008270, -0.00009061, 0.00004180, 0.00012678, 0.00003600, -0.00011168, -0.00011935, 0.00001120, 0.00009607, 0.00003620, -0.00006189, -0.00006128, 0.00001612, 0.00004647, -0.00001218, -0.00007812, -0.00007206, -0.00001271, 0.00002855, 0.00002078, -0.00000712, -0.00001761, -0.00000742, 0.00000008, -0.00001018, },
		{ -0.00000825, 0.00001744, 0.00003361, 0.00003588, 0.00002085, -0.00001045, -0.00003919, -0.00003893, -0.00001016, 0.00001416, 0.00001261, 0.00000487, 0.00001493, 0.00002615, 0.00000864, -0.00002275, -0.00002482, 0.00000054, 0.00000318, -0.00002940, -0.00003893, 0.00001397, 0.00007417, 0.00006155, -0.00001253, -0.00006039, -0.00003826, 0.00001103, 0.00003541, 0.00002985, 0.00000414, -0.00003858, -0.00006503, -0.00002207, 0.00006970, 0.00009787, 0.00000480, -0.00009904, -0.00006562, 0.00007090, 0.00012108, 0.00000557, -0.00012408, -0.00009686, 0.00003926, 0.00009653, 0.00001542, -0.00007140, -0.00004775, 0.00002957, 0.00003685, -0.00003453, -0.00008304, -0.00004959, 0.00001399, 0.00003220, 0.00000146, -0.00002620, -0.00001967, 0.00000215, 0.00000653, -0.00001221, -0.00003170, -0.00003019, },
		{ 0.00003981, 0.00003215, 0.00000833, -0.00002272, -0.00004162, -0.00003218, -0.00000356, 0.00001571, 0.00001380, 0.00000952, 0.00001773, 0.00002128, 0.00000085, -0.00002468, -0.00002287, -0.00000353, -0.00000613, -0.00002841, -0.00001890, 0.00003756, 0.00007636, 0.00003800, -0.00003899, -0.00006436, -0.00001731, 0.00003589, 0.00003996, 0.00000764, -0.00002663, -0.00004915, -0.00004435, 0.00000695, 0.00007102, 0.00006443, -0.00002911, -0.00009747, -0.00003600, 0.00008731, 0.00010311, -0.00002113, -0.00012309, -0.00006971, 0.00005767, 0.00008646, -0.00000345, -0.00007087, -0.00003082, 0.00003514, 0.00001955, -0.00005253, -0.00007527, -0.00002086, 0.00003290, 0.00002234, -0.00002319, -0.00003848, -0.00001135, 0.00001561, 0.00000887, -0.00002001, -0.00003747, -0.00002648, 0.00000264, 0.00002908, },
		{ -0.00000466, -0.00003142, -0.00004029, -0.00002428, 0.00000239, 0.00001698, 0.00001507, 0.00001304, 0.00001855, 0.00001590, -0.00000565, -0.00002639, -0.00002316, -0.00000868, -0.00001094, -0.00001962, 0.00000378, 0.00005367, 0.00006657, 0.00000938, -0.00005816, -0.00005547, 0.00000901, 0.00005195, 0.00002886, -0.00002141, -0.00004721, -0.00004198, -0.00001718, 0.00002403, 0.00005599, 0.00002981, -0.00004764, -0.00008262, -0.00001047, 0.00008935, 0.00007985, -0.00003798, -0.00011130, -0.00004532, 0.00006401, 0.00007115, -0.00001488, -0.00006172, -0.00001644, 0.00003072, -0.00000061, -0.00006050, -0.00005567, 0.00000725, 0.00003876, 0.00000104, -0.00004617, -0.00003927, 0.00000565, 0.00002754, 0.00000377, -0.00003203, -0.00004028, -0.00001636, 0.00001710, 0.00003838, 0.00003980, 0.00002328, },
		{ -0.00003608, -0.00001612, 0.00000742, 0.00001780, 0.00001608, 0.00001545, 0.00001803, 0.00001041, -0.00001150, -0.00002872, -0.00002497, -0.00001246, -0.00000968, -0.00000495, 0.00002456, 0.00005938, 0.00004691, -0.00001844, -0.00006605, -0.00003614, 0.00003349, 0.00005445, 0.00000565, -0.00004801, -0.00005198, -0.00002154, 0.00000648, 0.00002483, 0.00003068, 0.00000287, -0.00004998, -0.00006231, 0.00000472, 0.00008010, 0.00005912, -0.00004286, -0.00009481, -0.00002969, 0.00006015, 0.00005741, -0.00001632, -0.00004874, -0.00000984, 0.00001797, -0.00001692, -0.00005529, -0.00002900, 0.00002726, 0.00002982, -0.00002584, -0.00006035, -0.00002727, 0.00002639, 0.00003255, -0.00000938, -0.00004442, -0.00003645, 0.00000001, 0.00003164, 0.00004214, 0.00003343, 0.00001168, -0.00001571, -0.00003597, },
		{ 0.00001141, 0.00001808, 0.00001669, 0.00001698, 0.00001660, 0.00000473, -0.00001740, -0.00003178, -0.00002676, -0.00001276, -0.00000240, 0.00001220, 0.00003930, 0.00005408, 0.00002175, -0.00003997, -0.00006170, -0.00001167, 0.00004954, 0.00004279, -0.00002250, -0.00006394, -0.00004053, 0.00000382, 0.00001848, 0.00001045, 0.00000389, -0.00001063, -0.00004015, -0.00004483, 0.00000762, 0.00006602, 0.00004689, -0.00003820, -0.00008095, -0.00002522, 0.00005155, 0.00005100, -0.00000934, -0.00003836, -0.00001340, 0.00000211, -0.00002316, -0.00003782, -0.00000280, 0.00003360, 0.00000881, -0.00005030, -0.00006093, -0.00000526, 0.00004428, 0.00002722, -0.00002807, -0.00005232, -0.00002400, 0.00002009, 0.00004220, 0.00003858, 0.00002228, 0.00000023, -0.00002322, -0.00003676, -0.00003018, -0.00000839, },
		},
		{ { -0.00001125, -0.00000876, -0.00000721, -0.00000529, 0.00000777, 0.00001834, 0.00000501, -0.00001222, -0.00000396, 0.00000799, -0.00000618, -0.00001982, -0.00000261, 0.00001631, 0.00000647, -0.00000781, -0.00000345, 0.00000287, 0.00000243, 0.00000322, -0.00000552, -0.00002003, -0.00000367, 0.00002249, -0.00001620, -0.00007446, -0.00000899, 0.00012701, 0.00008619, -0.00013117, -0.00018240, 0.00005845, 0.00023011, 0.00005845, -0.00018240, -0.00013117, 0.00008619, 0.00012701, -0.00000899, -0.00007446, -0.00001620, 0.00002249, -0.00000367, -0.00002003, -0.00000552, 0.00000322, 0.00000243, 0.00000287, -0.00000345, -0.00000781, 0.00000647, 0.00001631, -0.00000261, -0.00001982, -0.00000618, 0.00000799, -0.00000396, -0.00001222, 0.00000501, 0.00001834, 0.00000777, -0.00000529, -0.00000721, -0.00000876, },
		{ -0.00000619, -0.00000118, 0.00000745, 0.00001402, 0.00000346, -0.00001074, -0.00000371, 0.00000780, -0.00000471, -0.00001934, -0.00000445, 0.00001577, 0.00000839, -0.00000664, -0.00000409, 0.00000074, -0.00000247, -0.00000090, -0.00000107, -0.00001071, -0.00000368, 0.00001372, -0.00001667, -0.00006537, -0.00001055, 0.00011039, 0.00007703, -0.00012166, -0.00017135, 0.00005815, 0.00022712, 0.00005845, -0.00018614, -0.00013609, 0.00009180, 0.00013852, -0.00000783, -0.00008170, -0.00001582, 0.00003005, -0.00000373, -0.00002866, -0.00000961, 0.00000744, 0.00000723, 0.00000467, -0.00000286, -0.00000877, 0.00000423, 0.00001617, -0.00000062, -0.00001950, -0.00000750, 0.00000769, -0.00000419, -0.00001324, 0.00000646, 0.00002214, 0.00000797, -0.00000915, -0.00000813, -0.00000586, -0.00001125, -0.00001169, },
		{ 0.00000702, 0.00000918, 0.00000184, -0.00000878, -0.00000346, 0.00000704, -0.00000310, -0.00001793, -0.00000607, 0.00001444, 0.00000985, -0.00000525, -0.00000479, -0.00000171, -0.00000724, -0.00000462, 0.00000357, -0.00000123, -0.00000379, 0.00000417, -0.00001725, -0.00005491, -0.00001243, 0.00008942, 0.00006463, -0.00010815, -0.00015354, 0.00005769, 0.00021837, 0.00005816, -0.00018240, -0.00013609, 0.00009369, 0.00014440, -0.00000711, -0.00008675, -0.00001555, 0.00003606, -0.00000382, -0.00003616, -0.00001319, 0.00001146, 0.00001173, 0.00000617, -0.00000234, -0.00000955, 0.00000181, 0.00001552, 0.00000143, -0.00001852, -0.00000868, 0.00000697, -0.00000441, -0.00001385, 0.00000781, 0.00002541, 0.00000806, -0.00001272, -0.00000893, -0.00000300, -0.00001125, -0.00001462, -0.00000507, 0.00000315, },
		{ 0.00000015, -0.00000633, -0.00000320, 0.00000567, -0.00000136, -0.00001549, -0.00000742, 0.00001220, 0.00001074, -0.00000364, -0.00000554, -0.00000452, -0.00001165, -0.00000764, 0.00000823, 0.00000791, -0.00000405, -0.00000569, -0.00001791, -0.00004362, -0.00001455, 0.00006505, 0.00004938, -0.00009153, -0.00012985, 0.00005730, 0.00020456, 0.00005769, -0.00017135, -0.00013117, 0.00009180, 0.00014440, -0.00000687, -0.00008933, -0.00001539, 0.00004024, -0.00000390, -0.00004215, -0.00001612, 0.00001503, 0.00001572, 0.00000739, -0.00000189, -0.00001017, -0.00000065, 0.00001451, 0.00000348, -0.00001706, -0.00000971, 0.00000591, -0.00000462, -0.00001409, 0.00000905, 0.00002817, 0.00000805, -0.00001600, -0.00000962, -0.00000023, -0.00001125, -0.00001751, -0.00000386, 0.00000766, 0.00000646, 0.00000387, },
		{ -0.00000294, 0.00000368, 0.00000051, -0.00001195, -0.00000845, 0.00000897, 0.00001094, -0.00000177, -0.00000632, -0.00000767, -0.00001548, -0.00000965, 0.00001274, 0.00001617, -0.00000452, -0.00001538, -0.00001866, -0.00003211, -0.00001682, 0.00003839, 0.00003175, -0.00007290, -0.00010144, 0.00005736, 0.00018679, 0.00005730, -0.00015354, -0.00012166, 0.00008619, 0.00013852, -0.00000711, -0.00008933, -0.00001533, 0.00004237, -0.00000396, -0.00004632, -0.00001830, 0.00001792, 0.00001902, 0.00000833, -0.00000152, -0.00001065, -0.00000301, 0.00001328, 0.00000545, -0.00001529, -0.00001060, 0.00000463, -0.00000481, -0.00001402, 0.00001018, 0.00003045, 0.00000795, -0.00001896, -0.00001020, 0.00000244, -0.00001125, -0.00002035, -0.00000258, 0.00001231, 0.00000580, -0.00000187, -0.00000159, -0.00000341, },
		{ 0.00000249, -0.00000729, -0.00000912, 0.00000470, 0.00001040, 0.00000033, -0.00000714, -0.00001117, -0.00001855, -0.00001039, 0.00001692, 0.00002303, -0.00000521, -0.00002439, -0.00001948, -0.00002100, -0.00001914, 0.00001069, 0.00001230, -0.00005358, -0.00006975, 0.00005832, 0.00016654, 0.00005736, -0.00012985, -0.00010815, 0.00007704, 0.00012701, -0.00000783, -0.00008674, -0.00001539, 0.00004237, -0.00000398, -0.00004845, -0.00001963, 0.00001995, 0.00002149, 0.00000903, -0.00000122, -0.00001100, -0.00000513, 0.00001200, 0.00000726, -0.00001337, -0.00001136, 0.00000321, -0.00000499, -0.00001371, 0.00001119, 0.00003229, 0.00000778, -0.00002161, -0.00001068, 0.00000499, -0.00001125, -0.00002309, -0.00000125, 0.00001703, 0.00000504, -0.00000797, -0.00000337, -0.00000003, -0.00000268, 0.00000107, },
		{ -0.00000940, -0.00000063, 0.00000904, 0.00000267, -0.00000799, -0.00001502, -0.00002069, -0.00000963, 0.00002063, 0.00002802, -0.00000619, -0.00003225, -0.00002037, -0.00001094, -0.00002141, -0.00001674, -0.00000837, -0.00003500, -0.00003640, 0.00006071, 0.00014561, 0.00005832, -0.00010144, -0.00009152, 0.00006464, 0.00011039, -0.00000899, -0.00008170, -0.00001555, 0.00004024, -0.00000396, -0.00004845, -0.00002009, 0.00002100, 0.00002301, 0.00000949, -0.00000101, -0.00001125, -0.00000691, 0.00001079, 0.00000884, -0.00001148, -0.00001199, 0.00000177, -0.00000514, -0.00001324, 0.00001208, 0.00003374, 0.00000756, -0.00002394, -0.00001106, 0.00000737, -0.00001124, -0.00002572, 0.00000011, 0.00002176, 0.00000420, -0.00001434, -0.00000516, 0.00000374, -0.00000242, -0.00000214, 0.00000455, -0.00000152, },
		{ 0.00000685, 0.00000525, -0.00000885, -0.00001918, -0.00002175, -0.00000718, 0.00002374, 0.00003072, -0.00000747, -0.00003850, -0.00002132, -0.00000255, -0.00002351, -0.00004259, -0.00002962, -0.00001865, -0.00000313, 0.00006511, 0.00012602, 0.00006071, -0.00006975, -0.00007290, 0.00004938, 0.00008943, -0.00001054, -0.00007445, -0.00001582, 0.00003607, -0.00000390, -0.00004631, -0.00001963, 0.00002100, 0.00002353, 0.00000971, -0.00000088, -0.00001141, -0.00000825, 0.00000978, 0.00001014, -0.00000977, -0.00001249, 0.00000039, -0.00000527, -0.00001267, 0.00001285, 0.00003484, 0.00000730, -0.00002597, -0.00001135, 0.00000959, -0.00001124, -0.00002821, 0.00000148, 0.00002644, 0.00000329, -0.00002088, -0.00000695, 0.00000786, -0.00000217, -0.00000590, 0.00000669, 0.00000530, -0.00000930, -0.00000700, },
		{ -0.00000972, -0.00002365, -0.00002165, -0.00000293, 0.00002612, 0.00003079, -0.00000909, -0.00004275, -0.00002233, 0.00000360, -0.00002535, -0.00006555, -0.00005083, -0.00000605, 0.00002824, 0.00007211, 0.00010993, 0.00006511, -0.00003639, -0.00005358, 0.00003176, 0.00006506, -0.00001242, -0.00006536, -0.00001619, 0.00003006, -0.00000381, -0.00004214, -0.00001829, 0.00001996, 0.00002301, 0.00000972, -0.00000084, -0.00001149, -0.00000909, 0.00000905, 0.00001111, -0.00000837, -0.00001287, -0.00000082, -0.00000538, -0.00001207, 0.00001351, 0.00003565, 0.00000703, -0.00002771, -0.00001157, 0.00001161, -0.00001124, -0.00003053, 0.00000282, 0.00003100, 0.00000234, -0.00002749, -0.00000871, 0.00001224, -0.00000192, -0.00001015, 0.00000888, 0.00001308, -0.00000880, -0.00001434, 0.00000384, 0.00000803, },
		{ -0.00002034, 0.00000320, 0.00002769, 0.00002798, -0.00001106, -0.00004467, -0.00002338, 0.00000701, -0.00002683, -0.00008442, -0.00007138, 0.00000136, 0.00005593, 0.00008228, 0.00009957, 0.00007211, -0.00000313, -0.00003499, 0.00001231, 0.00003840, -0.00001454, -0.00005490, -0.00001666, 0.00002250, -0.00000372, -0.00003615, -0.00001611, 0.00001793, 0.00002149, 0.00000949, -0.00000088, -0.00001148, -0.00000937, 0.00000867, 0.00001170, -0.00000737, -0.00001314, -0.00000180, -0.00000547, -0.00001151, 0.00001404, 0.00003623, 0.00000677, -0.00002916, -0.00001173, 0.00001342, -0.00001124, -0.00003265, 0.00000412, 0.00003536, 0.00000137, -0.00003404, -0.00001042, 0.00001680, -0.00000168, -0.00001478, 0.00001108, 0.00002168, -0.00000794, -0.00002256, 0.00000005, 0.00001099, -0.00001058, -0.00002837, },
		{ 0.00002841, 0.00002215, -0.00001337, -0.00004402, -0.00002445, 0.00000726, -0.00002789, -0.00009815, -0.00009068, 0.00000226, 0.00007824, 0.00009617, 0.00009708, 0.00008228, 0.00002825, -0.00001864, -0.00000836, 0.00001070, -0.00001681, -0.00004361, -0.00001723, 0.00001373, -0.00000366, -0.00002865, -0.00001318, 0.00001505, 0.00001903, 0.00000904, -0.00000100, -0.00001141, -0.00000908, 0.00000867, 0.00001191, -0.00000685, -0.00001330, -0.00000249, -0.00000553, -0.00001104, 0.00001445, 0.00003663, 0.00000652, -0.00003034, -0.00001183, 0.00001500, -0.00001124, -0.00003456, 0.00000534, 0.00003946, 0.00000042, -0.00004042, -0.00001205, 0.00002142, -0.00000146, -0.00001969, 0.00001328, 0.00003092, -0.00000676, -0.00003152, -0.00000445, 0.00001410, -0.00001144, -0.00003331, -0.00001780, 0.00001120, },
		{ -0.00001603, -0.00004066, -0.00002555, 0.00000405, -0.00002845, -0.00010585, -0.00010821, -0.00000447, 0.00009365, 0.00011424, 0.00010447, 0.00009617, 0.00005594, -0.00000604, -0.00002961, -0.00001673, -0.00001913, -0.00003209, -0.00001789, 0.00000418, -0.00000366, -0.00002001, -0.00000959, 0.00001148, 0.00001573, 0.00000835, -0.00000121, -0.00001124, -0.00000824, 0.00000906, 0.00001171, -0.00000685, -0.00001335, -0.00000284, -0.00000557, -0.00001070, 0.00001475, 0.00003689, 0.00000631, -0.00003127, -0.00001190, 0.00001634, -0.00001123, -0.00003624, 0.00000647, 0.00004322, -0.00000051, -0.00004649, -0.00001359, 0.00002601, -0.00000124, -0.00002476, 0.00001543, 0.00004062, -0.00000529, -0.00004105, -0.00000956, 0.00001732, -0.00001229, -0.00003840, -0.00001409, 0.00002099, 0.00002825, 0.00001327, },
		{ -0.00002666, -0.00000278, -0.00002847, -0.00010684, -0.00012354, -0.00001974, 0.00010084, 0.00013685, 0.00012350, 0.00011424, 0.00007825, 0.00000137, -0.00005082, -0.00004257, -0.00002139, -0.00002098, -0.00001864, -0.00000568, -0.00000377, -0.00001070, -0.00000550, 0.00000746, 0.00001175, 0.00000740, -0.00000150, -0.00001099, -0.00000690, 0.00000979, 0.00001112, -0.00000736, -0.00001329, -0.00000284, -0.00000558, -0.00001052, 0.00001492, 0.00003704, 0.00000615, -0.00003195, -0.00001194, 0.00001742, -0.00001123, -0.00003766, 0.00000747, 0.00004659, -0.00000137, -0.00005213, -0.00001500, 0.00003045, -0.00000105, -0.00002985, 0.00001751, 0.00005053, -0.00000360, -0.00005096, -0.00001515, 0.00002059, -0.00001310, -0.00004360, -0.00000929, 0.00003242, 0.00002722, 0.00000143, -0.00001901, -0.00003453, },
		{ -0.00002793, -0.00010072, -0.00013630, -0.00004415, 0.00009880, 0.00016424, 0.00015557, 0.00013685, 0.00009365, 0.00000227, -0.00007137, -0.00006554, -0.00002349, -0.00001092, -0.00001946, -0.00001536, -0.00000404, -0.00000121, -0.00000105, 0.00000324, 0.00000725, 0.00000619, -0.00000187, -0.00001063, -0.00000512, 0.00001081, 0.00001016, -0.00000835, -0.00001313, -0.00000248, -0.00000556, -0.00001052, 0.00001498, 0.00003712, 0.00000606, -0.00003240, -0.00001196, 0.00001825, -0.00001123, -0.00003882, 0.00000832, 0.00004949, -0.00000215, -0.00005722, -0.00001628, 0.00003464, -0.00000087, -0.00003483, 0.00001947, 0.00006043, -0.00000176, -0.00006101, -0.00002108, 0.00002388, -0.00001389, -0.00004882, -0.00000352, 0.00004529, 0.00002538, -0.00001318, -0.00002228, -0.00002569, -0.00002777, -0.00001328, },
		{ -0.00014625, -0.00007796, 0.00008685, 0.00019651, 0.00020169, 0.00016424, 0.00010084, -0.00000447, -0.00009067, -0.00008441, -0.00002533, -0.00000253, -0.00002035, -0.00002437, -0.00000449, 0.00000794, 0.00000359, -0.00000088, 0.00000245, 0.00000469, -0.00000232, -0.00001015, -0.00000299, 0.00001202, 0.00000886, -0.00000975, -0.00001286, -0.00000179, -0.00000552, -0.00001069, 0.00001493, 0.00003712, 0.00000603, -0.00003262, -0.00001197, 0.00001880, -0.00001123, -0.00003970, 0.00000900, 0.00005188, -0.00000282, -0.00006164, -0.00001740, 0.00003846, -0.00000071, -0.00003956, 0.00002130, 0.00007005, 0.00000017, -0.00007099, -0.00002718, 0.00002712, -0.00001463, -0.00005400, 0.00000303, 0.00005931, 0.00002280, -0.00003025, -0.00002579, -0.00001431, -0.00002887, -0.00002734, -0.00002682, -0.00008734, },
		{ 0.00006467, 0.00023360, 0.00026240, 0.00019652, 0.00009881, -0.00001974, -0.00010820, -0.00009814, -0.00002682, 0.00000362, -0.00002131, -0.00003223, -0.00000519, 0.00001620, 0.00000825, -0.00000460, -0.00000245, 0.00000289, -0.00000283, -0.00000953, -0.00000063, 0.00001331, 0.00000728, -0.00001146, -0.00001247, -0.00000080, -0.00000545, -0.00001103, 0.00001476, 0.00003705, 0.00000606, -0.00003262, -0.00001197, 0.00001908, -0.00001122, -0.00004029, 0.00000950, 0.00005371, -0.00000336, -0.00006531, -0.00001833, 0.00004181, -0.00000058, -0.00004391, 0.00002296, 0.00007914, 0.00000210, -0.00008065, -0.00003327, 0.00003025, -0.00001533, -0.00005905, 0.00001017, 0.00007416, 0.00001960, -0.00004935, -0.00002948, -0.00000066, -0.00002995, -0.00004473, -0.00002516, -0.00006682, -0.00015326, -0.00012113, },
		{ 0.00033770, 0.00023360, 0.00008686, -0.00004414, -0.00012353, -0.00010583, -0.00002787, 0.00000703, -0.00002231, -0.00003848, -0.00000617, 0.00002306, 0.00001276, -0.00000761, -0.00000721, 0.00000077, -0.00000342, -0.00000874, 0.00000183, 0.00001453, 0.00000547, -0.00001335, -0.00001196, 0.00000042, -0.00000536, -0.00001149, 0.00001447, 0.00003690, 0.00000617, -0.00003239, -0.00001197, 0.00001908, -0.00001122, -0.00004058, 0.00000981, 0.00005496, -0.00000376, -0.00006814, -0.00001908, 0.00004461, -0.00000046, -0.00004775, 0.00002441, 0.00008747, 0.00000396, -0.00008975, -0.00003919, 0.00003322, -0.00001597, -0.00006390, 0.00001767, 0.00008947, 0.00001589, -0.00007001, -0.00003330, 0.00001492, -0.00003099, -0.00006507, -0.00002300, -0.00003958, -0.00015729, -0.00017324, 0.00003233, 0.00027527, },
		{ 0.00006468, -0.00007795, -0.00013629, -0.00010683, -0.00002843, 0.00000728, -0.00002335, -0.00004273, -0.00000745, 0.00002805, 0.00001695, -0.00000962, -0.00001162, -0.00000168, -0.00000406, -0.00000778, 0.00000426, 0.00001555, 0.00000351, -0.00001526, -0.00001133, 0.00000179, -0.00000525, -0.00001205, 0.00001406, 0.00003665, 0.00000633, -0.00003194, -0.00001195, 0.00001881, -0.00001122, -0.00004058, 0.00000991, 0.00005558, -0.00000400, -0.00007006, -0.00001962, 0.00004678, -0.00000037, -0.00005096, 0.00002564, 0.00009480, 0.00000568, -0.00009805, -0.00004475, 0.00003596, -0.00001656, -0.00006845, 0.00002529, 0.00010485, 0.00001183, -0.00009164, -0.00003717, 0.00003198, -0.00003200, -0.00008787, -0.00002039, -0.00000627, -0.00015844, -0.00023351, -0.00000972, 0.00032112, 0.00042707, 0.00027527, },
		{ -0.00014624, -0.00010071, -0.00002845, 0.00000407, -0.00002443, -0.00004465, -0.00000907, 0.00003075, 0.00002066, -0.00001036, -0.00001545, -0.00000448, -0.00000476, -0.00000661, 0.00000650, 0.00001621, 0.00000146, -0.00001703, -0.00001057, 0.00000324, -0.00000511, -0.00001264, 0.00001353, 0.00003626, 0.00000654, -0.00003125, -0.00001193, 0.00001826, -0.00001122, -0.00004028, 0.00000981, 0.00005559, -0.00000408, -0.00007103, -0.00001995, 0.00004826, -0.00000031, -0.00005346, 0.00002663, 0.00010092, 0.00000719, -0.00010534, -0.00004978, 0.00003843, -0.00001707, -0.00007264, 0.00003278, 0.00011988, 0.00000758, -0.00011362, -0.00004100, 0.00005003, -0.00003295, -0.00011254, -0.00001741, 0.00003218, -0.00015689, -0.00030087, -0.00006064, 0.00037059, 0.00052945, 0.00032112, 0.00003233, -0.00012112, },
		{ -0.00002791, -0.00000276, -0.00002553, -0.00004400, -0.00001103, 0.00003082, 0.00002377, -0.00000960, -0.00001852, -0.00000764, -0.00000550, -0.00000522, 0.00000842, 0.00001634, -0.00000059, -0.00001849, -0.00000968, 0.00000466, -0.00000495, -0.00001321, 0.00001288, 0.00003568, 0.00000679, -0.00003032, -0.00001188, 0.00001744, -0.00001121, -0.00003968, 0.00000951, 0.00005496, -0.00000399, -0.00007102, -0.00002005, 0.00004901, -0.00000027, -0.00005517, 0.00002734, 0.00010568, 0.00000843, -0.00011142, -0.00005413, 0.00004058, -0.00001752, -0.00007639, 0.00003988, 0.00013415, 0.00000331, -0.00013532, -0.00004472, 0.00006852, -0.00003384, -0.00013839, -0.00001415, 0.00007466, -0.00015294, -0.00037389, -0.00011923, 0.00042295, 0.00064326, 0.00037059, -0.00000972, -0.00017323, -0.00015325, -0.00008733, },
		{ -0.00002664, -0.00004064, -0.00001335, 0.00002801, 0.00002615, -0.00000715, -0.00002065, -0.00001114, -0.00000629, -0.00000360, 0.00000989, 0.00001581, -0.00000257, -0.00001946, -0.00000864, 0.00000595, -0.00000478, -0.00001368, 0.00001212, 0.00003487, 0.00000706, -0.00002913, -0.00001181, 0.00001636, -0.00001121, -0.00003880, 0.00000902, 0.00005373, -0.00000375, -0.00007005, -0.00001994, 0.00004901, -0.00000025, -0.00005604, 0.00002778, 0.00010893, 0.00000935, -0.00011613, -0.00005766, 0.00004235, -0.00001789, -0.00007963, 0.00004636, 0.00014726, -0.00000080, -0.00015608, -0.00004825, 0.00008688, -0.00003466, -0.00016468, -0.00001073, 0.00011988, -0.00014696, -0.00045094, -0.00018398, 0.00047734, 0.00076641, 0.00042295, -0.00006063, -0.00023351, -0.00015728, -0.00006681, -0.00002680, -0.00001326, },
		{ -0.00001600, 0.00002218, 0.00002772, -0.00000289, -0.00002171, -0.00001498, -0.00000711, -0.00000174, 0.00001078, 0.00001448, -0.00000441, -0.00001978, -0.00000746, 0.00000701, -0.00000458, -0.00001398, 0.00001123, 0.00003377, 0.00000734, -0.00002768, -0.00001170, 0.00001503, -0.00001121, -0.00003764, 0.00000834, 0.00005190, -0.00000334, -0.00006812, -0.00001961, 0.00004827, -0.00000026, -0.00005603, 0.00002792, 0.00011057, 0.00000992, -0.00011934, -0.00006027, 0.00004372, -0.00001818, -0.00008229, 0.00005200, 0.00015882, -0.00000461, -0.00017526, -0.00005150, 0.00010452, -0.00003541, -0.00019064, -0.00000724, 0.00016641, -0.00013939, -0.00053014, -0.00025311, 0.00053277, 0.00089643, 0.00047734, -0.00011922, -0.00030086, -0.00015843, -0.00003956, -0.00002515, -0.00002732, -0.00002775, -0.00003450, },
		{ 0.00002844, 0.00000324, -0.00002161, -0.00001915, -0.00000795, 0.00000037, 0.00001099, 0.00001225, -0.00000603, -0.00001930, -0.00000614, 0.00000773, -0.00000437, -0.00001404, 0.00001022, 0.00003233, 0.00000760, -0.00002594, -0.00001154, 0.00001345, -0.00001121, -0.00003621, 0.00000749, 0.00004951, -0.00000280, -0.00006529, -0.00001906, 0.00004679, -0.00000030, -0.00005516, 0.00002778, 0.00011058, 0.00001012, -0.00012097, -0.00006187, 0.00004464, -0.00001839, -0.00008434, 0.00005660, 0.00016851, -0.00000795, -0.00019226, -0.00005439, 0.00012088, -0.00003607, -0.00021547, -0.00000381, 0.00021277, -0.00013072, -0.00060948, -0.00032465, 0.00058817, 0.00103045, 0.00053277, -0.00018397, -0.00037389, -0.00015688, -0.00000626, -0.00002299, -0.00004471, -0.00002885, -0.00002567, -0.00001898, 0.00001330, },
		{ -0.00002030, -0.00002361, -0.00000881, 0.00000272, 0.00001044, 0.00000902, -0.00000738, -0.00001788, -0.00000467, 0.00000804, -0.00000415, -0.00001380, 0.00000910, 0.00003050, 0.00000783, -0.00002390, -0.00001131, 0.00001165, -0.00001120, -0.00003453, 0.00000650, 0.00004662, -0.00000212, -0.00006162, -0.00001831, 0.00004463, -0.00000036, -0.00005345, 0.00002735, 0.00010893, 0.00000993, -0.00012097, -0.00006240, 0.00004511, -0.00001851, -0.00008572, 0.00006000, 0.00017605, -0.00001069, -0.00020656, -0.00005686, 0.00013543, -0.00003664, -0.00023843, -0.00000055, 0.00025745, -0.00012145, -0.00068688, -0.00039647, 0.00064242, 0.00116536, 0.00058817, -0.00025311, -0.00045093, -0.00015293, 0.00003219, -0.00002038, -0.00006505, -0.00002993, -0.00001429, -0.00002225, 0.00000146, 0.00002828, 0.00001124, },
		{ -0.00000968, 0.00000529, 0.00000908, 0.00000475, -0.00000841, -0.00001544, -0.00000305, 0.00000784, -0.00000391, -0.00001319, 0.00000786, 0.00002822, 0.00000800, -0.00002156, -0.00001102, 0.00000963, -0.00001120, -0.00003261, 0.00000538, 0.00004326, -0.00000134, -0.00005719, -0.00001737, 0.00004183, -0.00000044, -0.00005095, 0.00002664, 0.00010569, 0.00000936, -0.00011934, -0.00006186, 0.00004512, -0.00001855, -0.00008642, 0.00006209, 0.00018120, -0.00001273, -0.00021771, -0.00005885, 0.00014769, -0.00003711, -0.00025878, 0.00000242, 0.00029897, -0.00011212, -0.00076023, -0.00046640, 0.00069436, 0.00129788, 0.00064242, -0.00032465, -0.00053013, -0.00014695, 0.00007467, -0.00001740, -0.00008786, -0.00003097, -0.00000064, -0.00002576, -0.00001316, 0.00002725, 0.00002102, -0.00001777, -0.00002833, },
		{ 0.00000689, -0.00000059, -0.00000907, -0.00001190, -0.00000131, 0.00000709, -0.00000366, -0.00001217, 0.00000651, 0.00002546, 0.00000810, -0.00001891, -0.00001063, 0.00000742, -0.00001120, -0.00003048, 0.00000416, 0.00003950, -0.00000047, -0.00005210, -0.00001625, 0.00003848, -0.00000055, -0.00004773, 0.00002566, 0.00010094, 0.00000844, -0.00011612, -0.00006026, 0.00004465, -0.00001851, -0.00008642, 0.00006280, 0.00018381, -0.00001399, -0.00022534, -0.00006031, 0.00015726, -0.00003748, -0.00027590, 0.00000501, 0.00033596, -0.00010322, -0.00082752, -0.00053227, 0.00074283, 0.00142469, 0.00069436, -0.00039647, -0.00060947, -0.00013938, 0.00011989, -0.00001414, -0.00011252, -0.00003198, 0.00001494, -0.00002946, -0.00003022, 0.00002541, 0.00003246, -0.00001405, -0.00003327, -0.00001054, 0.00000807, },
		{ -0.00000936, -0.00000724, 0.00000056, 0.00000572, -0.00000341, -0.00001069, 0.00000506, 0.00002219, 0.00000811, -0.00001595, -0.00001016, 0.00000503, -0.00001120, -0.00002816, 0.00000287, 0.00003541, 0.00000045, -0.00004645, -0.00001497, 0.00003467, -0.00000069, -0.00004389, 0.00002444, 0.00009482, 0.00000721, -0.00011140, -0.00005765, 0.00004373, -0.00001838, -0.00008572, 0.00006209, 0.00018381, -0.00001442, -0.00022921, -0.00006120, 0.00016382, -0.00003775, -0.00028925, 0.00000712, 0.00036717, -0.00009523, -0.00088684, -0.00059203, 0.00078674, 0.00154249, 0.00074283, -0.00046639, -0.00068687, -0.00013071, 0.00016642, -0.00001071, -0.00013838, -0.00003293, 0.00003200, -0.00003328, -0.00004933, 0.00002283, 0.00004532, -0.00000925, -0.00003837, -0.00001140, 0.00001103, 0.00000388, -0.00000696, },
		{ 0.00000253, 0.00000374, -0.00000315, -0.00000873, 0.00000352, 0.00001839, 0.00000802, -0.00001267, -0.00000957, 0.00000250, -0.00001120, -0.00002567, 0.00000152, 0.00003105, 0.00000142, -0.00004038, -0.00001355, 0.00003049, -0.00000084, -0.00003953, 0.00002299, 0.00008749, 0.00000570, -0.00010532, -0.00005411, 0.00004237, -0.00001817, -0.00008433, 0.00006001, 0.00018120, -0.00001399, -0.00022921, -0.00006150, 0.00016716, -0.00003791, -0.00029838, 0.00000868, 0.00039156, -0.00008855, -0.00093651, -0.00064377, 0.00082506, 0.00164818, 0.00078674, -0.00053227, -0.00076023, -0.00012145, 0.00021278, -0.00000723, -0.00016467, -0.00003382, 0.00005005, -0.00003714, -0.00006998, 0.00001962, 0.00005934, -0.00000349, -0.00004356, -0.00001225, 0.00001414, 0.00000009, -0.00001430, -0.00000925, -0.00000147, },
		{ -0.00000289, -0.00000628, 0.00000189, 0.00001407, 0.00000782, -0.00000910, -0.00000888, -0.00000017, -0.00001120, -0.00002304, 0.00000016, 0.00002649, 0.00000239, -0.00003400, -0.00001201, 0.00002605, -0.00000101, -0.00003480, 0.00002133, 0.00007917, 0.00000399, -0.00009803, -0.00004976, 0.00004060, -0.00001787, -0.00008228, 0.00005661, 0.00017605, -0.00001273, -0.00022533, -0.00006120, 0.00016716, -0.00003797, -0.00030303, 0.00000964, 0.00040829, -0.00008352, -0.00097508, -0.00068586, 0.00085690, 0.00173892, 0.00082506, -0.00059202, -0.00082752, -0.00011211, 0.00025746, -0.00000380, -0.00019062, -0.00003465, 0.00006854, -0.00004098, -0.00009161, 0.00001591, 0.00007418, 0.00000306, -0.00004879, -0.00001307, 0.00001736, -0.00000441, -0.00002252, -0.00000876, 0.00000535, 0.00000460, 0.00000112, },
		{ 0.00000020, 0.00000923, 0.00000751, -0.00000524, -0.00000807, -0.00000295, -0.00001119, -0.00002030, -0.00000120, 0.00002181, 0.00000334, -0.00002744, -0.00001037, 0.00002146, -0.00000120, -0.00002981, 0.00001951, 0.00007008, 0.00000213, -0.00008972, -0.00004472, 0.00003846, -0.00001750, -0.00007961, 0.00005201, 0.00016853, -0.00001068, -0.00021770, -0.00006030, 0.00016382, -0.00003791, -0.00030303, 0.00000996, 0.00041680, -0.00008040, -0.00100144, -0.00071693, 0.00088148, 0.00181224, 0.00085690, -0.00064377, -0.00088684, -0.00010322, 0.00029898, -0.00000054, -0.00021546, -0.00003539, 0.00008690, -0.00004470, -0.00011360, 0.00001185, 0.00008950, 0.00001020, -0.00005397, -0.00001385, 0.00002063, -0.00000952, -0.00003148, -0.00000790, 0.00001312, 0.00000674, -0.00000209, -0.00000263, -0.00000335, },
		{ 0.00000707, -0.00000113, -0.00000716, -0.00000580, -0.00001119, -0.00001746, -0.00000253, 0.00001708, 0.00000425, -0.00002083, -0.00000866, 0.00001684, -0.00000141, -0.00002472, 0.00001754, 0.00006046, 0.00000020, -0.00008062, -0.00003916, 0.00003599, -0.00001705, -0.00007637, 0.00004638, 0.00015884, -0.00000794, -0.00020655, -0.00005884, 0.00015726, -0.00003775, -0.00029838, 0.00000964, 0.00041680, -0.00007934, -0.00101481, -0.00073599, 0.00089820, 0.00186611, 0.00088148, -0.00068586, -0.00093651, -0.00009522, 0.00033597, 0.00000243, -0.00023842, -0.00003605, 0.00010453, -0.00004823, -0.00013530, 0.00000760, 0.00010487, 0.00001770, -0.00005902, -0.00001460, 0.00002391, -0.00001511, -0.00004101, -0.00000671, 0.00002172, 0.00000893, -0.00000585, -0.00000237, 0.00000002, -0.00000154, 0.00000392, },
		{ -0.00000613, -0.00000871, -0.00001119, -0.00001456, -0.00000380, 0.00001236, 0.00000509, -0.00001429, -0.00000690, 0.00001229, -0.00000164, -0.00001964, 0.00001547, 0.00005057, -0.00000172, -0.00007096, -0.00003324, 0.00003324, -0.00001653, -0.00007262, 0.00003990, 0.00014727, -0.00000459, -0.00019225, -0.00005685, 0.00014770, -0.00003748, -0.00028924, 0.00000868, 0.00040829, -0.00008040, -0.00101481, -0.00074241, 0.00090667, 0.00189904, 0.00089820, -0.00071693, -0.00097508, -0.00008854, 0.00036718, 0.00000501, -0.00025877, -0.00003663, 0.00012089, -0.00005148, -0.00015606, 0.00000333, 0.00011990, 0.00002532, -0.00006387, -0.00001530, 0.00002715, -0.00002104, -0.00005092, -0.00000525, 0.00003097, 0.00001113, -0.00001010, -0.00000212, 0.00000379, -0.00000332, -0.00000182, 0.00000652, 0.00000320, },
		{ -0.00001119, -0.00001164, -0.00000501, 0.00000772, 0.00000585, -0.00000791, -0.00000511, 0.00000791, -0.00000187, -0.00001473, 0.00001332, 0.00004066, -0.00000356, -0.00006098, -0.00002714, 0.00003028, -0.00001595, -0.00006843, 0.00003280, 0.00013417, -0.00000079, -0.00017524, -0.00005438, 0.00013544, -0.00003710, -0.00027590, 0.00000712, 0.00039156, -0.00008352, -0.00100144, -0.00073598, 0.00090667, 0.00191013, 0.00090667, -0.00073598, -0.00100144, -0.00008352, 0.00039156, 0.00000712, -0.00027590, -0.00003710, 0.00013544, -0.00005438, -0.00017524, -0.00000079, 0.00013417, 0.00003280, -0.00006843, -0.00001595, 0.00003028, -0.00002714, -0.00006098, -0.00000356, 0.00004066, 0.00001332, -0.00001473, -0.00000187, 0.00000791, -0.00000511, -0.00000791, 0.00000585, 0.00000772, -0.00000501, -0.00001164, },
		{ -0.00000613, 0.00000320, 0.00000652, -0.00000182, -0.00000332, 0.00000379, -0.00000212, -0.00001010, 0.00001113, 0.00003097, -0.00000525, -0.00005092, -0.00002104, 0.00002715, -0.00001530, -0.00006387, 0.00002532, 0.00011990, 0.00000333, -0.00015606, -0.00005148, 0.00012089, -0.00003663, -0.00025877, 0.00000501, 0.00036718, -0.00008854, -0.00097508, -0.00071693, 0.00089820, 0.00189904, 0.00090667, -0.00074241, -0.00101481, -0.00008040, 0.00040829, 0.00000868, -0.00028924, -0.00003748, 0.00014770, -0.00005685, -0.00019225, -0.00000459, 0.00014727, 0.00003990, -0.00007262, -0.00001653, 0.00003324, -0.00003324, -0.00007096, -0.00000172, 0.00005057, 0.00001547, -0.00001964, -0.00000164, 0.00001229, -0.00000690, -0.00001429, 0.00000509, 0.00001236, -0.00000380, -0.00001456, -0.00001119, -0.00000871, },
		{ 0.00000707, 0.00000392, -0.00000154, 0.00000002, -0.00000237, -0.00000585, 0.00000893, 0.00002172, -0.00000671, -0.00004101, -0.00001511, 0.00002391, -0.00001460, -0.00005902, 0.00001770, 0.00010487, 0.00000760, -0.00013530, -0.00004823, 0.00010453, -0.00003605, -0.00023842, 0.00000243, 0.00033597, -0.00009522, -0.00093651, -0.00068586, 0.00088148, 0.00186611, 0.00089820, -0.00073599, -0.00101481, -0.00007934, 0.00041680, 0.00000964, -0.00029838, -0.00003775, 0.00015726, -0.00005884, -0.00020655, -0.00000794, 0.00015884, 0.00004638, -0.00007637, -0.00001705, 0.00003599, -0.00003916, -0.00008062, 0.00000020, 0.00006046, 0.00001754, -0.00002472, -0.00000141, 0.00001684, -0.00000866, -0.00002083, 0.00000425, 0.00001708, -0.00000253, -0.00001746, -0.00001119, -0.00000580, -0.00000716, -0.00000113, },
		{ 0.00000020, -0.00000335, -0.00000263, -0.00000209, 0.00000674, 0.00001312, -0.00000790, -0.00003148, -0.00000952, 0.00002063, -0.00001385, -0.00005397, 0.00001020, 0.00008950, 0.00001185, -0.00011360, -0.00004470, 0.00008690, -0.00003539, -0.00021546, -0.00000054, 0.00029898, -0.00010322, -0.00088684, -0.00064377, 0.00085690, 0.00181224, 0.00088148, -0.00071693, -0.00100144, -0.00008040, 0.00041680, 0.00000996, -0.00030303, -0.00003791, 0.00016382, -0.00006030, -0.00021770, -0.00001068, 0.00016853, 0.00005201, -0.00007961, -0.00001750, 0.00003846, -0.00004472, -0.00008972, 0.00000213, 0.00007008, 0.00001951, -0.00002981, -0.00000120, 0.00002146, -0.00001037, -0.00002744, 0.00000334, 0.00002181, -0.00000120, -0.00002030, -0.00001119, -0.00000295, -0.00000807, -0.00000524, 0.00000751, 0.00000923, },
		{ -0.00000289, 0.00000112, 0.00000460, 0.00000535, -0.00000876, -0.00002252, -0.00000441, 0.00001736, -0.00001307, -0.00004879, 0.00000306, 0.00007418, 0.00001591, -0.00009161, -0.00004098, 0.00006854, -0.00003465, -0.00019062, -0.00000380, 0.00025746, -0.00011211, -0.00082752, -0.00059202, 0.00082506, 0.00173892, 0.00085690, -0.00068586, -0.00097508, -0.00008352, 0.00040829, 0.00000964, -0.00030303, -0.00003797, 0.00016716, -0.00006120, -0.00022533, -0.00001273, 0.00017605, 0.00005661, -0.00008228, -0.00001787, 0.00004060, -0.00004976, -0.00009803, 0.00000399, 0.00007917, 0.00002133, -0.00003480, -0.00000101, 0.00002605, -0.00001201, -0.00003400, 0.00000239, 0.00002649, 0.00000016, -0.00002304, -0.00001120, -0.00000017, -0.00000888, -0.00000910, 0.00000782, 0.00001407, 0.00000189, -0.00000628, },
		{ 0.00000253, -0.00000147, -0.00000925, -0.00001430, 0.00000009, 0.00001414, -0.00001225, -0.00004356, -0.00000349, 0.00005934, 0.00001962, -0.00006998, -0.00003714, 0.00005005, -0.00003382, -0.00016467, -0.00000723, 0.00021278, -0.00012145, -0.00076023, -0.00053227, 0.00078674, 0.00164818, 0.00082506, -0.00064377, -0.00093651, -0.00008855, 0.00039156, 0.00000868, -0.00029838, -0.00003791, 0.00016716, -0.00006150, -0.00022921, -0.00001399, 0.00018120, 0.00006001, -0.00008433, -0.00001817, 0.00004237, -0.00005411, -0.00010532, 0.00000570, 0.00008749, 0.00002299, -0.00003953, -0.00000084, 0.00003049, -0.00001355, -0.00004038, 0.00000142, 0.00003105, 0.00000152, -0.00002567, -0.00001120, 0.00000250, -0.00000957, -0.00001267, 0.00000802, 0.00001839, 0.00000352, -0.00000873, -0.00000315, 0.00000374, },
		{ -0.00000936, -0.00000696, 0.00000388, 0.00001103, -0.00001140, -0.00003837, -0.00000925, 0.00004532, 0.00002283, -0.00004933, -0.00003328, 0.00003200, -0.00003293, -0.00013838, -0.00001071, 0.00016642, -0.00013071, -0.00068687, -0.00046639, 0.00074283, 0.00154249, 0.00078674, -0.00059203, -0.00088684, -0.00009523, 0.00036717, 0.00000712, -0.00028925, -0.00003775, 0.00016382, -0.00006120, -0.00022921, -0.00001442, 0.00018381, 0.00006209, -0.00008572, -0.00001838, 0.00004373, -0.00005765, -0.00011140, 0.00000721, 0.00009482, 0.00002444, -0.00004389, -0.00000069, 0.00003467, -0.00001497, -0.00004645, 0.00000045, 0.00003541, 0.00000287, -0.00002816, -0.00001120, 0.00000503, -0.00001016, -0.00001595, 0.00000811, 0.00002219, 0.00000506, -0.00001069, -0.00000341, 0.00000572, 0.00000056, -0.00000724, },
		{ 0.00000689, 0.00000807, -0.00001054, -0.00003327, -0.00001405, 0.00003246, 0.00002541, -0.00003022, -0.00002946, 0.00001494, -0.00003198, -0.00011252, -0.00001414, 0.00011989, -0.00013938, -0.00060947, -0.00039647, 0.00069436, 0.00142469, 0.00074283, -0.00053227, -0.00082752, -0.00010322, 0.00033596, 0.00000501, -0.00027590, -0.00003748, 0.00015726, -0.00006031, -0.00022534, -0.00001399, 0.00018381, 0.00006280, -0.00008642, -0.00001851, 0.00004465, -0.00006026, -0.00011612, 0.00000844, 0.00010094, 0.00002566, -0.00004773, -0.00000055, 0.00003848, -0.00001625, -0.00005210, -0.00000047, 0.00003950, 0.00000416, -0.00003048, -0.00001120, 0.00000742, -0.00001063, -0.00001891, 0.00000810, 0.00002546, 0.00000651, -0.00001217, -0.00000366, 0.00000709, -0.00000131, -0.00001190, -0.00000907, -0.00000059, },
		{ -0.00000968, -0.00002833, -0.00001777, 0.00002102, 0.00002725, -0.00001316, -0.00002576, -0.00000064, -0.00003097, -0.00008786, -0.00001740, 0.00007467, -0.00014695, -0.00053013, -0.00032465, 0.00064242, 0.00129788, 0.00069436, -0.00046640, -0.00076023, -0.00011212, 0.00029897, 0.00000242, -0.00025878, -0.00003711, 0.00014769, -0.00005885, -0.00021771, -0.00001273, 0.00018120, 0.00006209, -0.00008642, -0.00001855, 0.00004512, -0.00006186, -0.00011934, 0.00000936, 0.00010569, 0.00002664, -0.00005095, -0.00000044, 0.00004183, -0.00001737, -0.00005719, -0.00000134, 0.00004326, 0.00000538, -0.00003261, -0.00001120, 0.00000963, -0.00001102, -0.00002156, 0.00000800, 0.00002822, 0.00000786, -0.00001319, -0.00000391, 0.00000784, -0.00000305, -0.00001544, -0.00000841, 0.00000475, 0.00000908, 0.00000529, },
		{ -0.00002030, 0.00001124, 0.00002828, 0.00000146, -0.00002225, -0.00001429, -0.00002993, -0.00006505, -0.00002038, 0.00003219, -0.00015293, -0.00045093, -0.00025311, 0.00058817, 0.00116536, 0.00064242, -0.00039647, -0.00068688, -0.00012145, 0.00025745, -0.00000055, -0.00023843, -0.00003664, 0.00013543, -0.00005686, -0.00020656, -0.00001069, 0.00017605, 0.00006000, -0.00008572, -0.00001851, 0.00004511, -0.00006240, -0.00012097, 0.00000993, 0.00010893, 0.00002735, -0.00005345, -0.00000036, 0.00004463, -0.00001831, -0.00006162, -0.00000212, 0.00004662, 0.00000650, -0.00003453, -0.00001120, 0.00001165, -0.00001131, -0.00002390, 0.00000783, 0.00003050, 0.00000910, -0.00001380, -0.00000415, 0.00000804, -0.00000467, -0.00001788, -0.00000738, 0.00000902, 0.00001044, 0.00000272, -0.00000881, -0.00002361, },
		{ 0.00002844, 0.00001330, -0.00001898, -0.00002567, -0.00002885, -0.00004471, -0.00002299, -0.00000626, -0.00015688, -0.00037389, -0.00018397, 0.00053277, 0.00103045, 0.00058817, -0.00032465, -0.00060948, -0.00013072, 0.00021277, -0.00000381, -0.00021547, -0.00003607, 0.00012088, -0.00005439, -0.00019226, -0.00000795, 0.00016851, 0.00005660, -0.00008434, -0.00001839, 0.00004464, -0.00006187, -0.00012097, 0.00001012, 0.00011058, 0.00002778, -0.00005516, -0.00000030, 0.00004679, -0.00001906, -0.00006529, -0.00000280, 0.00004951, 0.00000749, -0.00003621, -0.00001121, 0.00001345, -0.00001154, -0.00002594, 0.00000760, 0.00003233, 0.00001022, -0.00001404, -0.00000437, 0.00000773, -0.00000614, -0.00001930, -0.00000603, 0.00001225, 0.00001099, 0.00000037, -0.00000795, -0.00001915, -0.00002161, 0.00000324, },
		{ -0.00001600, -0.00003450, -0.00002775, -0.00002732, -0.00002515, -0.00003956, -0.00015843, -0.00030086, -0.00011922, 0.00047734, 0.00089643, 0.00053277, -0.00025311, -0.00053014, -0.00013939, 0.00016641, -0.00000724, -0.00019064, -0.00003541, 0.00010452, -0.00005150, -0.00017526, -0.00000461, 0.00015882, 0.00005200, -0.00008229, -0.00001818, 0.00004372, -0.00006027, -0.00011934, 0.00000992, 0.00011057, 0.00002792, -0.00005603, -0.00000026, 0.00004827, -0.00001961, -0.00006812, -0.00000334, 0.00005190, 0.00000834, -0.00003764, -0.00001121, 0.00001503, -0.00001170, -0.00002768, 0.00000734, 0.00003377, 0.00001123, -0.00001398, -0.00000458, 0.00000701, -0.00000746, -0.00001978, -0.00000441, 0.00001448, 0.00001078, -0.00000174, -0.00000711, -0.00001498, -0.00002171, -0.00000289, 0.00002772, 0.00002218, },
		{ -0.00002664, -0.00001326, -0.00002680, -0.00006681, -0.00015728, -0.00023351, -0.00006063, 0.00042295, 0.00076641, 0.00047734, -0.00018398, -0.00045094, -0.00014696, 0.00011988, -0.00001073, -0.00016468, -0.00003466, 0.00008688, -0.00004825, -0.00015608, -0.00000080, 0.00014726, 0.00004636, -0.00007963, -0.00001789, 0.00004235, -0.00005766, -0.00011613, 0.00000935, 0.00010893, 0.00002778, -0.00005604, -0.00000025, 0.00004901, -0.00001994, -0.00007005, -0.00000375, 0.00005373, 0.00000902, -0.00003880, -0.00001121, 0.00001636, -0.00001181, -0.00002913, 0.00000706, 0.00003487, 0.00001212, -0.00001368, -0.00000478, 0.00000595, -0.00000864, -0.00001946, -0.00000257, 0.00001581, 0.00000989, -0.00000360, -0.00000629, -0.00001114, -0.00002065, -0.00000715, 0.00002615, 0.00002801, -0.00001335, -0.00004064, },
		{ -0.00002791, -0.00008733, -0.00015325, -0.00017323, -0.00000972, 0.00037059, 0.00064326, 0.00042295, -0.00011923, -0.00037389, -0.00015294, 0.00007466, -0.00001415, -0.00013839, -0.00003384, 0.00006852, -0.00004472, -0.00013532, 0.00000331, 0.00013415, 0.00003988, -0.00007639, -0.00001752, 0.00004058, -0.00005413, -0.00011142, 0.00000843, 0.00010568, 0.00002734, -0.00005517, -0.00000027, 0.00004901, -0.00002005, -0.00007102, -0.00000399, 0.00005496, 0.00000951, -0.00003968, -0.00001121, 0.00001744, -0.00001188, -0.00003032, 0.00000679, 0.00003568, 0.00001288, -0.00001321, -0.00000495, 0.00000466, -0.00000968, -0.00001849, -0.00000059, 0.00001634, 0.00000842, -0.00000522, -0.00000550, -0.00000764, -0.00001852, -0.00000960, 0.00002377, 0.00003082, -0.00001103, -0.00004400, -0.00002553, -0.00000276, },
		{ -0.00014624, -0.00012112, 0.00003233, 0.00032112, 0.00052945, 0.00037059, -0.00006064, -0.00030087, -0.00015689, 0.00003218, -0.00001741, -0.00011254, -0.00003295, 0.00005003, -0.00004100, -0.00011362, 0.00000758, 0.00011988, 0.00003278, -0.00007264, -0.00001707, 0.00003843, -0.00004978, -0.00010534, 0.00000719, 0.00010092, 0.00002663, -0.00005346, -0.00000031, 0.00004826, -0.00001995, -0.00007103, -0.00000408, 0.00005559, 0.00000981, -0.00004028, -0.00001122, 0.00001826, -0.00001193, -0.00003125, 0.00000654, 0.00003626, 0.00001353, -0.00001264, -0.00000511, 0.00000324, -0.00001057, -0.00001703, 0.00000146, 0.00001621, 0.00000650, -0.00000661, -0.00000476, -0.00000448, -0.00001545, -0.00001036, 0.00002066, 0.00003075, -0.00000907, -0.00004465, -0.00002443, 0.00000407, -0.00002845, -0.00010071, },
		{ 0.00006468, 0.00027527, 0.00042707, 0.00032112, -0.00000972, -0.00023351, -0.00015844, -0.00000627, -0.00002039, -0.00008787, -0.00003200, 0.00003198, -0.00003717, -0.00009164, 0.00001183, 0.00010485, 0.00002529, -0.00006845, -0.00001656, 0.00003596, -0.00004475, -0.00009805, 0.00000568, 0.00009480, 0.00002564, -0.00005096, -0.00000037, 0.00004678, -0.00001962, -0.00007006, -0.00000400, 0.00005558, 0.00000991, -0.00004058, -0.00001122, 0.00001881, -0.00001195, -0.00003194, 0.00000633, 0.00003665, 0.00001406, -0.00001205, -0.00000525, 0.00000179, -0.00001133, -0.00001526, 0.00000351, 0.00001555, 0.00000426, -0.00000778, -0.00000406, -0.00000168, -0.00001162, -0.00000962, 0.00001695, 0.00002805, -0.00000745, -0.00004273, -0.00002335, 0.00000728, -0.00002843, -0.00010683, -0.00013629, -0.00007795, },
		{ 0.00033770, 0.00027527, 0.00003233, -0.00017324, -0.00015729, -0.00003958, -0.00002300, -0.00006507, -0.00003099, 0.00001492, -0.00003330, -0.00007001, 0.00001589, 0.00008947, 0.00001767, -0.00006390, -0.00001597, 0.00003322, -0.00003919, -0.00008975, 0.00000396, 0.00008747, 0.00002441, -0.00004775, -0.00000046, 0.00004461, -0.00001908, -0.00006814, -0.00000376, 0.00005496, 0.00000981, -0.00004058, -0.00001122, 0.00001908, -0.00001197, -0.00003239, 0.00000617, 0.00003690, 0.00001447, -0.00001149, -0.00000536, 0.00000042, -0.00001196, -0.00001335, 0.00000547, 0.00001453, 0.00000183, -0.00000874, -0.00000342, 0.00000077, -0.00000721, -0.00000761, 0.00001276, 0.00002306, -0.00000617, -0.00003848, -0.00002231, 0.00000703, -0.00002787, -0.00010583, -0.00012353, -0.00004414, 0.00008686, 0.00023360, },
		{ 0.00006467, -0.00012113, -0.00015326, -0.00006682, -0.00002516, -0.00004473, -0.00002995, -0.00000066, -0.00002948, -0.00004935, 0.00001960, 0.00007416, 0.00001017, -0.00005905, -0.00001533, 0.00003025, -0.00003327, -0.00008065, 0.00000210, 0.00007914, 0.00002296, -0.00004391, -0.00000058, 0.00004181, -0.00001833, -0.00006531, -0.00000336, 0.00005371, 0.00000950, -0.00004029, -0.00001122, 0.00001908, -0.00001197, -0.00003262, 0.00000606, 0.00003705, 0.00001476, -0.00001103, -0.00000545, -0.00000080, -0.00001247, -0.00001146, 0.00000728, 0.00001331, -0.00000063, -0.00000953, -0.00000283, 0.00000289, -0.00000245, -0.00000460, 0.00000825, 0.00001620, -0.00000519, -0.00003223, -0.00002131, 0.00000362, -0.00002682, -0.00009814, -0.00010820, -0.00001974, 0.00009881, 0.00019652, 0.00026240, 0.00023360, },
		{ -0.00014625, -0.00008734, -0.00002682, -0.00002734, -0.00002887, -0.00001431, -0.00002579, -0.00003025, 0.00002280, 0.00005931, 0.00000303, -0.00005400, -0.00001463, 0.00002712, -0.00002718, -0.00007099, 0.00000017, 0.00007005, 0.00002130, -0.00003956, -0.00000071, 0.00003846, -0.00001740, -0.00006164, -0.00000282, 0.00005188, 0.00000900, -0.00003970, -0.00001123, 0.00001880, -0.00001197, -0.00003262, 0.00000603, 0.00003712, 0.00001493, -0.00001069, -0.00000552, -0.00000179, -0.00001286, -0.00000975, 0.00000886, 0.00001202, -0.00000299, -0.00001015, -0.00000232, 0.00000469, 0.00000245, -0.00000088, 0.00000359, 0.00000794, -0.00000449, -0.00002437, -0.00002035, -0.00000253, -0.00002533, -0.00008441, -0.00009067, -0.00000447, 0.00010084, 0.00016424, 0.00020169, 0.00019651, 0.00008685, -0.00007796, },
		{ -0.00002793, -0.00001328, -0.00002777, -0.00002569, -0.00002228, -0.00001318, 0.00002538, 0.00004529, -0.00000352, -0.00004882, -0.00001389, 0.00002388, -0.00002108, -0.00006101, -0.00000176, 0.00006043, 0.00001947, -0.00003483, -0.00000087, 0.00003464, -0.00001628, -0.00005722, -0.00000215, 0.00004949, 0.00000832, -0.00003882, -0.00001123, 0.00001825, -0.00001196, -0.00003240, 0.00000606, 0.00003712, 0.00001498, -0.00001052, -0.00000556, -0.00000248, -0.00001313, -0.00000835, 0.00001016, 0.00001081, -0.00000512, -0.00001063, -0.00000187, 0.00000619, 0.00000725, 0.00000324, -0.00000105, -0.00000121, -0.00000404, -0.00001536, -0.00001946, -0.00001092, -0.00002349, -0.00006554, -0.00007137, 0.00000227, 0.00009365, 0.00013685, 0.00015557, 0.00016424, 0.00009880, -0.00004415, -0.00013630, -0.00010072, },
		{ -0.00002666, -0.00003453, -0.00001901, 0.00000143, 0.00002722, 0.00003242, -0.00000929, -0.00004360, -0.00001310, 0.00002059, -0.00001515, -0.00005096, -0.00000360, 0.00005053, 0.00001751, -0.00002985, -0.00000105, 0.00003045, -0.00001500, -0.00005213, -0.00000137, 0.00004659, 0.00000747, -0.00003766, -0.00001123, 0.00001742, -0.00001194, -0.00003195, 0.00000615, 0.00003704, 0.00001492, -0.00001052, -0.00000558, -0.00000284, -0.00001329, -0.00000736, 0.00001112, 0.00000979, -0.00000690, -0.00001099, -0.00000150, 0.00000740, 0.00001175, 0.00000746, -0.00000550, -0.00001070, -0.00000377, -0.00000568, -0.00001864, -0.00002098, -0.00002139, -0.00004257, -0.00005082, 0.00000137, 0.00007825, 0.00011424, 0.00012350, 0.00013685, 0.00010084, -0.00001974, -0.00012354, -0.00010684, -0.00002847, -0.00000278, },
		{ -0.00001603, 0.00001327, 0.00002825, 0.00002099, -0.00001409, -0.00003840, -0.00001229, 0.00001732, -0.00000956, -0.00004105, -0.00000529, 0.00004062, 0.00001543, -0.00002476, -0.00000124, 0.00002601, -0.00001359, -0.00004649, -0.00000051, 0.00004322, 0.00000647, -0.00003624, -0.00001123, 0.00001634, -0.00001190, -0.00003127, 0.00000631, 0.00003689, 0.00001475, -0.00001070, -0.00000557, -0.00000284, -0.00001335, -0.00000685, 0.00001171, 0.00000906, -0.00000824, -0.00001124, -0.00000121, 0.00000835, 0.00001573, 0.00001148, -0.00000959, -0.00002001, -0.00000366, 0.00000418, -0.00001789, -0.00003209, -0.00001913, -0.00001673, -0.00002961, -0.00000604, 0.00005594, 0.00009617, 0.00010447, 0.00011424, 0.00009365, -0.00000447, -0.00010821, -0.00010585, -0.00002845, 0.00000405, -0.00002555, -0.00004066, },
		{ 0.00002841, 0.00001120, -0.00001780, -0.00003331, -0.00001144, 0.00001410, -0.00000445, -0.00003152, -0.00000676, 0.00003092, 0.00001328, -0.00001969, -0.00000146, 0.00002142, -0.00001205, -0.00004042, 0.00000042, 0.00003946, 0.00000534, -0.00003456, -0.00001124, 0.00001500, -0.00001183, -0.00003034, 0.00000652, 0.00003663, 0.00001445, -0.00001104, -0.00000553, -0.00000249, -0.00001330, -0.00000685, 0.00001191, 0.00000867, -0.00000908, -0.00001141, -0.00000100, 0.00000904, 0.00001903, 0.00001505, -0.00001318, -0.00002865, -0.00000366, 0.00001373, -0.00001723, -0.00004361, -0.00001681, 0.00001070, -0.00000836, -0.00001864, 0.00002825, 0.00008228, 0.00009708, 0.00009617, 0.00007824, 0.00000226, -0.00009068, -0.00009815, -0.00002789, 0.00000726, -0.00002445, -0.00004402, -0.00001337, 0.00002215, },
		{ -0.00002034, -0.00002837, -0.00001058, 0.00001099, 0.00000005, -0.00002256, -0.00000794, 0.00002168, 0.00001108, -0.00001478, -0.00000168, 0.00001680, -0.00001042, -0.00003404, 0.00000137, 0.00003536, 0.00000412, -0.00003265, -0.00001124, 0.00001342, -0.00001173, -0.00002916, 0.00000677, 0.00003623, 0.00001404, -0.00001151, -0.00000547, -0.00000180, -0.00001314, -0.00000737, 0.00001170, 0.00000867, -0.00000937, -0.00001148, -0.00000088, 0.00000949, 0.00002149, 0.00001793, -0.00001611, -0.00003615, -0.00000372, 0.00002250, -0.00001666, -0.00005490, -0.00001454, 0.00003840, 0.00001231, -0.00003499, -0.00000313, 0.00007211, 0.00009957, 0.00008228, 0.00005593, 0.00000136, -0.00007138, -0.00008442, -0.00002683, 0.00000701, -0.00002338, -0.00004467, -0.00001106, 0.00002798, 0.00002769, 0.00000320, },
		{ -0.00000972, 0.00000803, 0.00000384, -0.00001434, -0.00000880, 0.00001308, 0.00000888, -0.00001015, -0.00000192, 0.00001224, -0.00000871, -0.00002749, 0.00000234, 0.00003100, 0.00000282, -0.00003053, -0.00001124, 0.00001161, -0.00001157, -0.00002771, 0.00000703, 0.00003565, 0.00001351, -0.00001207, -0.00000538, -0.00000082, -0.00001287, -0.00000837, 0.00001111, 0.00000905, -0.00000909, -0.00001149, -0.00000084, 0.00000972, 0.00002301, 0.00001996, -0.00001829, -0.00004214, -0.00000381, 0.00003006, -0.00001619, -0.00006536, -0.00001242, 0.00006506, 0.00003176, -0.00005358, -0.00003639, 0.00006511, 0.00010993, 0.00007211, 0.00002824, -0.00000605, -0.00005083, -0.00006555, -0.00002535, 0.00000360, -0.00002233, -0.00004275, -0.00000909, 0.00003079, 0.00002612, -0.00000293, -0.00002165, -0.00002365, },
		{ 0.00000685, -0.00000700, -0.00000930, 0.00000530, 0.00000669, -0.00000590, -0.00000217, 0.00000786, -0.00000695, -0.00002088, 0.00000329, 0.00002644, 0.00000148, -0.00002821, -0.00001124, 0.00000959, -0.00001135, -0.00002597, 0.00000730, 0.00003484, 0.00001285, -0.00001267, -0.00000527, 0.00000039, -0.00001249, -0.00000977, 0.00001014, 0.00000978, -0.00000825, -0.00001141, -0.00000088, 0.00000971, 0.00002353, 0.00002100, -0.00001963, -0.00004631, -0.00000390, 0.00003607, -0.00001582, -0.00007445, -0.00001054, 0.00008943, 0.00004938, -0.00007290, -0.00006975, 0.00006071, 0.00012602, 0.00006511, -0.00000313, -0.00001865, -0.00002962, -0.00004259, -0.00002351, -0.00000255, -0.00002132, -0.00003850, -0.00000747, 0.00003072, 0.00002374, -0.00000718, -0.00002175, -0.00001918, -0.00000885, 0.00000525, },
		{ -0.00000940, -0.00000152, 0.00000455, -0.00000214, -0.00000242, 0.00000374, -0.00000516, -0.00001434, 0.00000420, 0.00002176, 0.00000011, -0.00002572, -0.00001124, 0.00000737, -0.00001106, -0.00002394, 0.00000756, 0.00003374, 0.00001208, -0.00001324, -0.00000514, 0.00000177, -0.00001199, -0.00001148, 0.00000884, 0.00001079, -0.00000691, -0.00001125, -0.00000101, 0.00000949, 0.00002301, 0.00002100, -0.00002009, -0.00004845, -0.00000396, 0.00004024, -0.00001555, -0.00008170, -0.00000899, 0.00011039, 0.00006464, -0.00009152, -0.00010144, 0.00005832, 0.00014561, 0.00006071, -0.00003640, -0.00003500, -0.00000837, -0.00001674, -0.00002141, -0.00001094, -0.00002037, -0.00003225, -0.00000619, 0.00002802, 0.00002063, -0.00000963, -0.00002069, -0.00001502, -0.00000799, 0.00000267, 0.00000904, -0.00000063, },
		{ 0.00000249, 0.00000107, -0.00000268, -0.00000003, -0.00000337, -0.00000797, 0.00000504, 0.00001703, -0.00000125, -0.00002309, -0.00001125, 0.00000499, -0.00001068, -0.00002161, 0.00000778, 0.00003229, 0.00001119, -0.00001371, -0.00000499, 0.00000321, -0.00001136, -0.00001337, 0.00000726, 0.00001200, -0.00000513, -0.00001100, -0.00000122, 0.00000903, 0.00002149, 0.00001995, -0.00001963, -0.00004845, -0.00000398, 0.00004237, -0.00001539, -0.00008674, -0.00000783, 0.00012701, 0.00007704, -0.00010815, -0.00012985, 0.00005736, 0.00016654, 0.00005832, -0.00006975, -0.00005358, 0.00001230, 0.00001069, -0.00001914, -0.00002100, -0.00001948, -0.00002439, -0.00000521, 0.00002303, 0.00001692, -0.00001039, -0.00001855, -0.00001117, -0.00000714, 0.00000033, 0.00001040, 0.00000470, -0.00000912, -0.00000729, },
		{ -0.00000294, -0.00000341, -0.00000159, -0.00000187, 0.00000580, 0.00001231, -0.00000258, -0.00002035, -0.00001125, 0.00000244, -0.00001020, -0.00001896, 0.00000795, 0.00003045, 0.00001018, -0.00001402, -0.00000481, 0.00000463, -0.00001060, -0.00001529, 0.00000545, 0.00001328, -0.00000301, -0.00001065, -0.00000152, 0.00000833, 0.00001902, 0.00001792, -0.00001830, -0.00004632, -0.00000396, 0.00004237, -0.00001533, -0.00008933, -0.00000711, 0.00013852, 0.00008619, -0.00012166, -0.00015354, 0.00005730, 0.00018679, 0.00005736, -0.00010144, -0.00007290, 0.00003175, 0.00003839, -0.00001682, -0.00003211, -0.00001866, -0.00001538, -0.00000452, 0.00001617, 0.00001274, -0.00000965, -0.00001548, -0.00000767, -0.00000632, -0.00000177, 0.00001094, 0.00000897, -0.00000845, -0.00001195, 0.00000051, 0.00000368, },
		{ 0.00000015, 0.00000387, 0.00000646, 0.00000766, -0.00000386, -0.00001751, -0.00001125, -0.00000023, -0.00000962, -0.00001600, 0.00000805, 0.00002817, 0.00000905, -0.00001409, -0.00000462, 0.00000591, -0.00000971, -0.00001706, 0.00000348, 0.00001451, -0.00000065, -0.00001017, -0.00000189, 0.00000739, 0.00001572, 0.00001503, -0.00001612, -0.00004215, -0.00000390, 0.00004024, -0.00001539, -0.00008933, -0.00000687, 0.00014440, 0.00009180, -0.00013117, -0.00017135, 0.00005769, 0.00020456, 0.00005730, -0.00012985, -0.00009153, 0.00004938, 0.00006505, -0.00001455, -0.00004362, -0.00001791, -0.00000569, -0.00000405, 0.00000791, 0.00000823, -0.00000764, -0.00001165, -0.00000452, -0.00000554, -0.00000364, 0.00001074, 0.00001220, -0.00000742, -0.00001549, -0.00000136, 0.00000567, -0.00000320, -0.00000633, },
		{ 0.00000702, 0.00000315, -0.00000507, -0.00001462, -0.00001125, -0.00000300, -0.00000893, -0.00001272, 0.00000806, 0.00002541, 0.00000781, -0.00001385, -0.00000441, 0.00000697, -0.00000868, -0.00001852, 0.00000143, 0.00001552, 0.00000181, -0.00000955, -0.00000234, 0.00000617, 0.00001173, 0.00001146, -0.00001319, -0.00003616, -0.00000382, 0.00003606, -0.00001555, -0.00008675, -0.00000711, 0.00014440, 0.00009369, -0.00013609, -0.00018240, 0.00005816, 0.00021837, 0.00005769, -0.00015354, -0.00010815, 0.00006463, 0.00008942, -0.00001243, -0.00005491, -0.00001725, 0.00000417, -0.00000379, -0.00000123, 0.00000357, -0.00000462, -0.00000724, -0.00000171, -0.00000479, -0.00000525, 0.00000985, 0.00001444, -0.00000607, -0.00001793, -0.00000310, 0.00000704, -0.00000346, -0.00000878, 0.00000184, 0.00000918, },
		{ -0.00000619, -0.00001169, -0.00001125, -0.00000586, -0.00000813, -0.00000915, 0.00000797, 0.00002214, 0.00000646, -0.00001324, -0.00000419, 0.00000769, -0.00000750, -0.00001950, -0.00000062, 0.00001617, 0.00000423, -0.00000877, -0.00000286, 0.00000467, 0.00000723, 0.00000744, -0.00000961, -0.00002866, -0.00000373, 0.00003005, -0.00001582, -0.00008170, -0.00000783, 0.00013852, 0.00009180, -0.00013609, -0.00018614, 0.00005845, 0.00022712, 0.00005815, -0.00017135, -0.00012166, 0.00007703, 0.00011039, -0.00001055, -0.00006537, -0.00001667, 0.00001372, -0.00000368, -0.00001071, -0.00000107, -0.00000090, -0.00000247, 0.00000074, -0.00000409, -0.00000664, 0.00000839, 0.00001577, -0.00000445, -0.00001934, -0.00000471, 0.00000780, -0.00000371, -0.00001074, 0.00000346, 0.00001402, 0.00000745, -0.00000118, },
		},
		{ { -0.00000067, -0.00000253, -0.00000145, 0.00000930, 0.00001426, -0.00000337, -0.00002242, -0.00002124, -0.00001531, -0.00000255, 0.00003030, 0.00003712, -0.00000490, -0.00002327, -0.00000384, -0.00001562, -0.00001991, 0.00003383, 0.00002093, -0.00008360, -0.00006300, 0.00007733, 0.00005591, -0.00007722, -0.00001629, 0.00012734, 0.00004836, -0.00009905, -0.00003600, 0.00008935, 0.00005912, -0.00003820, -0.00007559, -0.00003820, 0.00005912, 0.00008935, -0.00003600, -0.00009905, 0.00004836, 0.00012734, -0.00001629, -0.00007722, 0.00005591, 0.00007733, -0.00006300, -0.00008360, 0.00002093, 0.00003383, -0.00001991, -0.00001562, -0.00000384, -0.00002327, -0.00000490, 0.00003712, 0.00003030, -0.00000255, -0.00001531, -0.00002124, -0.00002242, -0.00000337, 0.00001426, 0.00000930, -0.00000145, -0.00000253, },
		{ -0.00000335, 0.00000866, 0.00001551, 0.00000228, -0.00001803, -0.00002391, -0.00002010, -0.00000782, 0.00002453, 0.00004108, 0.00000595, -0.00002271, -0.00000615, -0.00001027, -0.00002856, 0.00001648, 0.00003702, -0.00005930, -0.00009090, 0.00004265, 0.00008335, -0.00004848, -0.00004534, 0.00011575, 0.00009267, -0.00008270, -0.00006562, 0.00008730, 0.00007984, -0.00004286, -0.00008096, -0.00002994, 0.00004688, 0.00008010, -0.00001047, -0.00009748, 0.00000480, 0.00012095, 0.00001930, -0.00008824, 0.00001580, 0.00009358, -0.00002362, -0.00009427, -0.00000455, 0.00004312, -0.00000508, -0.00001787, -0.00000401, -0.00002244, -0.00001423, 0.00003113, 0.00003453, 0.00000192, -0.00001071, -0.00001623, -0.00002541, -0.00001034, 0.00001328, 0.00001053, -0.00000185, -0.00000090, 0.00000072, -0.00000451, },
		{ 0.00001744, 0.00000720, -0.00001320, -0.00002430, -0.00002423, -0.00001361, 0.00001737, 0.00004235, 0.00001746, -0.00001940, -0.00001078, -0.00000508, -0.00002957, -0.00000319, 0.00004007, -0.00002902, -0.00010092, -0.00000149, 0.00009003, -0.00001021, -0.00005935, 0.00009052, 0.00012850, -0.00004870, -0.00009061, 0.00007089, 0.00010311, -0.00003798, -0.00009481, -0.00002523, 0.00004517, 0.00006601, 0.00000471, -0.00008262, -0.00002911, 0.00009786, 0.00005094, -0.00007927, -0.00002552, 0.00008773, 0.00001723, -0.00008744, -0.00003250, 0.00004077, 0.00001196, -0.00001456, -0.00000531, -0.00002161, -0.00002189, 0.00002389, 0.00003731, 0.00000548, -0.00000720, -0.00000934, -0.00002586, -0.00001863, 0.00001144, 0.00001346, -0.00000345, -0.00000149, 0.00000424, -0.00000478, -0.00000744, 0.00000704, },
		{ -0.00000846, -0.00002288, -0.00002700, -0.00001948, 0.00000917, 0.00004068, 0.00002831, -0.00001263, -0.00001620, -0.00000284, -0.00002424, -0.00001937, 0.00003071, -0.00000157, -0.00009213, -0.00004364, 0.00007382, 0.00002600, -0.00005481, 0.00006021, 0.00014942, -0.00000221, -0.00010262, 0.00004180, 0.00012107, -0.00002113, -0.00011131, -0.00002969, 0.00005154, 0.00005446, 0.00000762, -0.00006231, -0.00004765, 0.00006443, 0.00006970, -0.00005430, -0.00005685, 0.00006241, 0.00004896, -0.00006438, -0.00005478, 0.00002667, 0.00002589, -0.00000520, -0.00000537, -0.00002143, -0.00002839, 0.00001601, 0.00003893, 0.00000834, -0.00000551, -0.00000159, -0.00002279, -0.00002746, 0.00000736, 0.00001827, -0.00000439, -0.00000526, 0.00000807, -0.00000189, -0.00001245, 0.00000304, 0.00001960, 0.00001217, },
		{ -0.00002795, -0.00002492, 0.00000026, 0.00003625, 0.00003713, -0.00000270, -0.00002025, -0.00000481, -0.00001589, -0.00002804, 0.00001327, 0.00001585, -0.00006907, -0.00007336, 0.00003928, 0.00004875, -0.00003464, 0.00003460, 0.00015392, 0.00004847, -0.00009635, 0.00000611, 0.00012677, 0.00000556, -0.00012310, -0.00004533, 0.00006015, 0.00005100, 0.00000093, -0.00004484, -0.00004998, 0.00002981, 0.00007101, -0.00002208, -0.00007061, 0.00002573, 0.00006360, -0.00003124, -0.00006462, 0.00000456, 0.00003183, 0.00000831, -0.00000175, -0.00002142, -0.00003446, 0.00000765, 0.00003975, 0.00001098, -0.00000601, 0.00000560, -0.00001574, -0.00003534, -0.00000006, 0.00002400, -0.00000268, -0.00001186, 0.00000993, 0.00000449, -0.00001634, -0.00000397, 0.00002082, 0.00001772, -0.00000388, -0.00002026, },
		{ -0.00000905, 0.00002956, 0.00004301, 0.00000910, -0.00002104, -0.00001035, -0.00000845, -0.00002821, -0.00000591, 0.00001982, -0.00004009, -0.00008442, -0.00000373, 0.00005061, -0.00000746, 0.00002116, 0.00014586, 0.00009474, -0.00007121, -0.00002751, 0.00011658, 0.00003600, -0.00012409, -0.00006971, 0.00006400, 0.00005741, -0.00000934, -0.00003622, -0.00004015, 0.00000286, 0.00005598, 0.00000694, -0.00006503, -0.00001139, 0.00005822, 0.00000271, -0.00005887, -0.00001906, 0.00002703, 0.00002201, 0.00000698, -0.00002004, -0.00004053, -0.00000148, 0.00003988, 0.00001409, -0.00000851, 0.00001063, -0.00000507, -0.00004033, -0.00001110, 0.00002877, 0.00000302, -0.00001965, 0.00000776, 0.00001328, -0.00001697, -0.00001350, 0.00001963, 0.00002392, 0.00000083, -0.00001710, -0.00002691, -0.00002936, },
		{ 0.00004573, 0.00002101, -0.00001774, -0.00001732, -0.00000492, -0.00002190, -0.00002075, 0.00001159, -0.00001444, -0.00007654, -0.00004321, 0.00003061, 0.00001518, 0.00002212, 0.00013279, 0.00013070, -0.00003123, -0.00005060, 0.00009153, 0.00006212, -0.00011168, -0.00009686, 0.00005767, 0.00007115, -0.00001633, -0.00003836, -0.00002538, -0.00001063, 0.00003068, 0.00002403, -0.00004436, -0.00003858, 0.00003580, 0.00002797, -0.00003908, -0.00003685, 0.00001207, 0.00003110, 0.00002033, -0.00001518, -0.00004624, -0.00001190, 0.00003904, 0.00001833, -0.00001222, 0.00001211, 0.00000792, -0.00004060, -0.00002490, 0.00003032, 0.00001287, -0.00002610, 0.00000055, 0.00002226, -0.00001280, -0.00002396, 0.00001478, 0.00003013, 0.00000609, -0.00001394, -0.00002398, -0.00003211, -0.00001855, 0.00002115, },
		{ -0.00001088, -0.00002302, -0.00000629, -0.00001314, -0.00002754, -0.00000377, 0.00000080, -0.00005522, -0.00006880, -0.00000528, 0.00002271, 0.00003333, 0.00012294, 0.00015513, 0.00001669, -0.00005749, 0.00005700, 0.00007665, -0.00008779, -0.00011935, 0.00003926, 0.00008645, -0.00001488, -0.00004874, -0.00001341, -0.00000967, 0.00000388, 0.00002482, -0.00001719, -0.00004916, 0.00000414, 0.00003770, -0.00001116, -0.00004301, -0.00000916, 0.00003155, 0.00003576, -0.00000512, -0.00005024, -0.00002395, 0.00003650, 0.00002402, -0.00001583, 0.00000928, 0.00002117, -0.00003495, -0.00003948, 0.00002671, 0.00002561, -0.00002855, -0.00001124, 0.00002868, -0.00000351, -0.00003306, 0.00000576, 0.00003522, 0.00001226, -0.00001111, -0.00001957, -0.00003245, -0.00002794, 0.00001130, 0.00004569, 0.00003145, },
		{ -0.00001132, -0.00000624, -0.00002601, -0.00001942, 0.00000293, -0.00002961, -0.00007512, -0.00004600, 0.00000934, 0.00004543, 0.00012156, 0.00017153, 0.00006562, -0.00004680, 0.00002085, 0.00007544, -0.00005824, -0.00013098, 0.00001120, 0.00009652, -0.00000346, -0.00006173, -0.00000985, 0.00000210, -0.00001590, 0.00001045, 0.00000648, -0.00004199, -0.00002663, 0.00002984, 0.00001647, -0.00003517, -0.00003044, 0.00002154, 0.00004913, 0.00001056, -0.00005049, -0.00003726, 0.00003120, 0.00003089, -0.00001781, 0.00000226, 0.00003225, -0.00002327, -0.00005216, 0.00001703, 0.00003876, -0.00002506, -0.00002571, 0.00003000, 0.00000980, -0.00003836, -0.00000698, 0.00003781, 0.00001940, -0.00000874, -0.00001442, -0.00002969, -0.00003663, -0.00000008, 0.00004358, 0.00003964, -0.00000214, -0.00002540, },
		{ -0.00001922, -0.00002934, -0.00000568, -0.00000925, -0.00006340, -0.00007901, -0.00002359, 0.00004700, 0.00012834, 0.00018629, 0.00011135, -0.00002117, -0.00000921, 0.00005871, -0.00003073, -0.00012892, -0.00002071, 0.00009607, 0.00001541, -0.00007088, -0.00001644, 0.00001797, -0.00002316, -0.00001323, 0.00001848, -0.00002154, -0.00004722, 0.00000764, 0.00003541, -0.00001526, -0.00004514, 0.00000235, 0.00005596, 0.00003044, -0.00004483, -0.00005054, 0.00002224, 0.00003795, -0.00001680, -0.00000782, 0.00003887, -0.00000675, -0.00006013, 0.00000176, 0.00004926, -0.00001501, -0.00003989, 0.00002474, 0.00002479, -0.00003796, -0.00002196, 0.00003663, 0.00002722, -0.00000674, -0.00000949, -0.00002343, -0.00004357, -0.00001327, 0.00003982, 0.00004577, 0.00000642, -0.00002392, -0.00001715, -0.00000403, },
		{ -0.00001879, -0.00000067, -0.00004086, -0.00009446, -0.00006756, 0.00002881, 0.00013663, 0.00020532, 0.00015404, 0.00001461, -0.00002797, 0.00003060, -0.00001231, -0.00011460, -0.00004933, 0.00008322, 0.00003620, -0.00007141, -0.00003082, 0.00003072, -0.00001693, -0.00003782, 0.00001494, 0.00000382, -0.00005198, -0.00002141, 0.00003995, 0.00001102, -0.00004830, -0.00002177, 0.00005283, 0.00005121, -0.00003194, -0.00006155, 0.00000927, 0.00004362, -0.00001207, -0.00001907, 0.00003951, 0.00001225, -0.00006120, -0.00001713, 0.00005426, 0.00000056, -0.00005052, 0.00001287, 0.00003847, -0.00003108, -0.00003694, 0.00003094, 0.00003505, -0.00000481, -0.00000592, -0.00001385, -0.00004724, -0.00002841, 0.00003419, 0.00005086, 0.00001332, -0.00001981, -0.00002066, -0.00000671, -0.00001207, -0.00003060, },
		{ -0.00001814, -0.00008862, -0.00010930, -0.00001229, 0.00013518, 0.00023055, 0.00019782, 0.00005634, -0.00003392, -0.00000265, -0.00000715, -0.00009302, -0.00006868, 0.00006011, 0.00005249, -0.00006190, -0.00004775, 0.00003513, -0.00000061, -0.00005530, -0.00000280, 0.00002468, -0.00004053, -0.00004801, 0.00002885, 0.00003589, -0.00003827, -0.00004474, 0.00003867, 0.00006838, -0.00001201, -0.00006757, -0.00000702, 0.00004604, -0.00000380, -0.00002917, 0.00003376, 0.00003070, -0.00005440, -0.00003657, 0.00005189, 0.00001916, -0.00005492, -0.00000404, 0.00004785, -0.00001826, -0.00004944, 0.00002072, 0.00004200, -0.00000253, -0.00000477, -0.00000186, -0.00004589, -0.00004498, 0.00002568, 0.00005601, 0.00001833, -0.00001558, -0.00001992, -0.00001186, -0.00000919, -0.00002436, -0.00002877, -0.00000516, },
		{ -0.00013506, -0.00007120, 0.00011198, 0.00025748, 0.00024843, 0.00010303, -0.00002866, -0.00003519, -0.00001562, -0.00007085, -0.00007593, 0.00003193, 0.00005933, -0.00004464, -0.00006129, 0.00002957, 0.00001954, -0.00006050, -0.00002901, 0.00003359, -0.00001757, -0.00006395, 0.00000565, 0.00005194, -0.00001731, -0.00006039, 0.00001537, 0.00007750, 0.00001280, -0.00006614, -0.00002477, 0.00004365, 0.00000680, -0.00003596, 0.00002251, 0.00004550, -0.00004031, -0.00005308, 0.00004185, 0.00003740, -0.00005168, -0.00002319, 0.00005077, -0.00000129, -0.00005725, 0.00000674, 0.00004710, 0.00000064, -0.00000673, 0.00001075, -0.00003806, -0.00006143, 0.00001283, 0.00006167, 0.00002276, -0.00001390, -0.00001519, -0.00001538, -0.00001295, -0.00001539, -0.00002966, -0.00001833, -0.00000540, -0.00006532, },
		{ 0.00005895, 0.00027507, 0.00030954, 0.00015792, -0.00001479, -0.00006361, -0.00003490, -0.00005396, -0.00007204, 0.00000498, 0.00005492, -0.00002460, -0.00006705, 0.00001611, 0.00003685, -0.00005254, -0.00005567, 0.00002725, 0.00000880, -0.00006462, -0.00002250, 0.00005444, 0.00000900, -0.00006436, -0.00001253, 0.00007544, 0.00003860, -0.00005589, -0.00004115, 0.00003573, 0.00001771, -0.00003799, 0.00000777, 0.00005425, -0.00002098, -0.00006361, 0.00002551, 0.00005186, -0.00004105, -0.00004125, 0.00004634, 0.00001717, -0.00005893, -0.00000955, 0.00004943, 0.00000519, -0.00001190, 0.00002151, -0.00002310, -0.00007511, -0.00000562, 0.00006699, 0.00002893, -0.00001632, -0.00000910, -0.00001331, -0.00002223, -0.00001018, -0.00001964, -0.00003174, -0.00000862, -0.00003492, -0.00013529, -0.00013510, },
		{ 0.00037905, 0.00022759, 0.00000698, -0.00008767, -0.00006089, -0.00004560, -0.00006091, -0.00001569, 0.00004101, -0.00000739, -0.00006373, -0.00000058, 0.00004647, -0.00003454, -0.00007528, 0.00000724, 0.00002982, -0.00005030, -0.00004709, 0.00004278, 0.00003348, -0.00005547, -0.00003900, 0.00006155, 0.00006057, -0.00003724, -0.00005290, 0.00002280, 0.00002653, -0.00003493, -0.00000779, 0.00005569, 0.00000051, -0.00006623, 0.00000564, 0.00005986, -0.00002486, -0.00005513, 0.00003515, 0.00003423, -0.00005403, -0.00002627, 0.00004830, 0.00001148, -0.00001953, 0.00002775, -0.00000180, -0.00008263, -0.00002985, 0.00006963, 0.00003920, -0.00002230, -0.00000584, -0.00000364, -0.00003259, -0.00001426, -0.00000197, -0.00003605, -0.00002719, -0.00001092, -0.00010830, -0.00018670, -0.00002391, 0.00026820, },
		{ 0.00004013, -0.00010920, -0.00009066, -0.00004580, -0.00004749, -0.00002788, 0.00002184, 0.00000287, -0.00005323, -0.00001537, 0.00004682, -0.00001218, -0.00008305, -0.00002087, 0.00003876, -0.00002585, -0.00006093, 0.00002064, 0.00004954, -0.00003614, -0.00005817, 0.00003800, 0.00007417, -0.00001259, -0.00005720, 0.00000674, 0.00003116, -0.00002767, -0.00002148, 0.00005005, 0.00002073, -0.00006057, -0.00001428, 0.00006017, -0.00000602, -0.00006271, 0.00001906, 0.00004734, -0.00004319, -0.00004143, 0.00004334, 0.00001949, -0.00002810, 0.00002718, 0.00002341, -0.00008065, -0.00005828, 0.00006627, 0.00005475, -0.00002904, -0.00000958, 0.00001241, -0.00003771, -0.00002979, 0.00001582, -0.00002464, -0.00005369, -0.00000538, -0.00006158, -0.00020933, -0.00012658, 0.00022240, 0.00044695, 0.00031879, },
		{ -0.00012414, -0.00005241, -0.00003548, -0.00003260, 0.00000225, 0.00000487, -0.00003959, -0.00002458, 0.00003958, 0.00000855, -0.00007813, -0.00004959, 0.00003290, 0.00000104, -0.00006036, -0.00000526, 0.00005319, -0.00001167, -0.00006606, 0.00000938, 0.00007636, 0.00001396, -0.00005245, -0.00000949, 0.00003036, -0.00001812, -0.00003124, 0.00003885, 0.00003667, -0.00004787, -0.00003085, 0.00005318, 0.00001218, -0.00006323, 0.00000067, 0.00005480, -0.00002791, -0.00005327, 0.00003455, 0.00002877, -0.00003548, 0.00001856, 0.00004859, -0.00006685, -0.00008732, 0.00005363, 0.00007460, -0.00003208, -0.00002273, 0.00003032, -0.00003173, -0.00005419, 0.00002445, 0.00000358, -0.00007613, -0.00002461, -0.00001017, -0.00019242, -0.00023007, 0.00012930, 0.00049573, 0.00043418, 0.00009301, -0.00012947, },
		{ -0.00002587, -0.00003291, -0.00001448, 0.00000027, -0.00002711, -0.00002726, 0.00002850, 0.00002342, -0.00006327, -0.00007206, 0.00001399, 0.00002234, -0.00004617, -0.00002727, 0.00004427, 0.00001138, -0.00006171, -0.00001844, 0.00006656, 0.00003755, -0.00003894, -0.00002252, 0.00002425, -0.00000870, -0.00003607, 0.00002453, 0.00004637, -0.00003064, -0.00004152, 0.00004075, 0.00002685, -0.00005738, -0.00001726, 0.00005595, -0.00001020, -0.00006043, 0.00002239, 0.00003832, -0.00003940, 0.00000228, 0.00006897, -0.00004098, -0.00011180, 0.00002978, 0.00009529, -0.00002679, -0.00004461, 0.00004335, -0.00001164, -0.00008044, 0.00001620, 0.00004305, -0.00008183, -0.00006637, 0.00002753, -0.00013567, -0.00031063, -0.00000873, 0.00050378, 0.00056823, 0.00017707, -0.00014605, -0.00016429, -0.00006331, },
		{ -0.00002783, -0.00000759, -0.00001840, -0.00002508, 0.00001753, 0.00003104, -0.00004335, -0.00008404, -0.00001271, 0.00003220, -0.00002319, -0.00003927, 0.00002639, 0.00002722, -0.00004744, -0.00003997, 0.00004691, 0.00005367, -0.00001891, -0.00002940, 0.00001432, -0.00000167, -0.00003626, 0.00000975, 0.00004921, -0.00001196, -0.00004508, 0.00002567, 0.00003617, -0.00004692, -0.00003246, 0.00005119, 0.00000777, -0.00006212, 0.00000774, 0.00004668, -0.00003799, -0.00001954, 0.00007995, -0.00000534, -0.00012601, -0.00000470, 0.00011148, -0.00001006, -0.00007106, 0.00004463, 0.00002108, -0.00009896, -0.00001202, 0.00008277, -0.00006182, -0.00011985, 0.00003534, -0.00005036, -0.00034597, -0.00017658, 0.00045122, 0.00070506, 0.00030280, -0.00015045, -0.00021530, -0.00007888, -0.00001684, -0.00003189, },
		{ -0.00001345, -0.00002114, 0.00000905, 0.00003274, -0.00002330, -0.00008482, -0.00004047, 0.00002855, 0.00000146, -0.00003848, 0.00000565, 0.00003254, -0.00002808, -0.00005154, 0.00002174, 0.00005938, 0.00000378, -0.00002841, 0.00000318, 0.00000150, -0.00003315, -0.00000321, 0.00004593, 0.00000515, -0.00004184, 0.00001092, 0.00003964, -0.00003421, -0.00004351, 0.00004169, 0.00002403, -0.00005817, -0.00000809, 0.00005220, -0.00003039, -0.00004329, 0.00007827, 0.00003520, -0.00012517, -0.00004623, 0.00011728, 0.00001820, -0.00009531, 0.00002956, 0.00006023, -0.00010062, -0.00005699, 0.00010936, -0.00001436, -0.00016836, 0.00000458, 0.00004273, -0.00032186, -0.00034795, 0.00032648, 0.00081932, 0.00047456, -0.00012783, -0.00027928, -0.00010351, -0.00000537, -0.00003039, -0.00003997, -0.00001562, },
		{ 0.00000305, 0.00003128, -0.00000633, -0.00007680, -0.00006339, 0.00001334, 0.00002077, -0.00002620, -0.00001135, 0.00002754, -0.00000938, -0.00005233, -0.00000355, 0.00005408, 0.00002456, -0.00001962, -0.00000613, 0.00000053, -0.00002869, -0.00001298, 0.00003824, 0.00001847, -0.00003332, -0.00000098, 0.00003801, -0.00002150, -0.00004993, 0.00002902, 0.00003700, -0.00004901, -0.00002344, 0.00005335, -0.00001705, -0.00006456, 0.00006294, 0.00007400, -0.00010691, -0.00008850, 0.00010806, 0.00005447, -0.00010972, -0.00000239, 0.00009625, -0.00007983, -0.00010939, 0.00011123, 0.00005369, -0.00019408, -0.00006262, 0.00011837, -0.00023709, -0.00049046, 0.00013175, 0.00088052, 0.00068568, -0.00005952, -0.00035240, -0.00014524, 0.00000999, -0.00002590, -0.00005415, -0.00002299, -0.00000989, -0.00001804, },
		{ 0.00000688, -0.00006392, -0.00007812, -0.00000857, 0.00003002, -0.00000712, -0.00001968, 0.00001561, 0.00000377, -0.00004442, -0.00002401, 0.00003969, 0.00003929, -0.00000496, -0.00001094, -0.00000354, -0.00002482, -0.00001927, 0.00002826, 0.00002683, -0.00002179, -0.00000852, 0.00003292, -0.00001053, -0.00005212, 0.00001480, 0.00004564, -0.00003557, -0.00003648, 0.00004918, 0.00000019, -0.00007914, 0.00003570, 0.00010402, -0.00007226, -0.00012372, 0.00008210, 0.00009208, -0.00010815, -0.00004690, 0.00011879, -0.00003678, -0.00015607, 0.00008236, 0.00012813, -0.00018343, -0.00015300, 0.00015362, -0.00010518, -0.00057314, -0.00011452, 0.00086004, 0.00091598, 0.00007193, -0.00042226, -0.00021322, 0.00002636, -0.00001297, -0.00007229, -0.00003197, -0.00000454, -0.00001628, -0.00000254, 0.00002908, },
		{ -0.00008455, -0.00003137, 0.00002806, 0.00001237, -0.00001762, 0.00000214, 0.00000887, -0.00003204, -0.00003645, 0.00002009, 0.00004533, 0.00001220, -0.00000969, -0.00000868, -0.00002287, -0.00002275, 0.00001794, 0.00003024, -0.00000964, -0.00001133, 0.00002637, -0.00000217, -0.00005095, 0.00000044, 0.00004940, -0.00001929, -0.00004541, 0.00003967, 0.00001840, -0.00008403, 0.00000088, 0.00011943, -0.00002584, -0.00014450, 0.00004162, 0.00012280, -0.00008797, -0.00009561, 0.00011969, 0.00002203, -0.00018363, 0.00002476, 0.00019098, -0.00013153, -0.00024522, 0.00013432, 0.00004783, -0.00057444, -0.00037937, 0.00073911, 0.00113268, 0.00027634, -0.00046767, -0.00031340, 0.00003493, 0.00001443, -0.00009264, -0.00004737, 0.00000431, -0.00001356, -0.00001032, 0.00002669, 0.00001777, -0.00004974, },
		{ 0.00001729, 0.00002648, -0.00000708, -0.00000743, 0.00000652, -0.00002002, -0.00004029, 0.00000000, 0.00004220, 0.00002789, -0.00000241, -0.00001246, -0.00002316, -0.00002469, 0.00000863, 0.00002956, 0.00000120, -0.00001005, 0.00002015, 0.00000352, -0.00004746, -0.00001299, 0.00004820, -0.00000190, -0.00004882, 0.00002587, 0.00003422, -0.00007822, -0.00003553, 0.00011707, 0.00002483, -0.00014572, -0.00000731, 0.00013905, -0.00005118, -0.00013818, 0.00009559, 0.00008508, -0.00018232, -0.00005155, 0.00022542, -0.00004420, -0.00031533, 0.00005940, 0.00018918, -0.00048838, -0.00062108, 0.00051554, 0.00129548, 0.00054992, -0.00046159, -0.00044381, 0.00002166, 0.00005947, -0.00010872, -0.00007430, 0.00001506, -0.00000558, -0.00002150, 0.00002226, 0.00002848, -0.00003592, -0.00008504, -0.00005045, },
		{ 0.00000719, -0.00000953, 0.00000008, -0.00001222, -0.00003747, -0.00001636, 0.00003164, 0.00003858, 0.00000921, -0.00001277, -0.00002498, -0.00002639, 0.00000084, 0.00002614, 0.00000951, -0.00000597, 0.00001548, 0.00000707, -0.00004255, -0.00002471, 0.00004233, 0.00001464, -0.00004598, 0.00000990, 0.00004465, -0.00006300, -0.00006717, 0.00009727, 0.00007095, -0.00012613, -0.00005598, 0.00013607, -0.00000421, -0.00016506, 0.00004912, 0.00013875, -0.00014904, -0.00013083, 0.00022057, 0.00006298, -0.00034320, -0.00005819, 0.00028721, -0.00032720, -0.00079792, 0.00020711, 0.00136461, 0.00087191, -0.00037721, -0.00059118, -0.00002951, 0.00011899, -0.00011026, -0.00011527, 0.00002203, 0.00001198, -0.00003428, 0.00001234, 0.00003999, -0.00002171, -0.00008283, -0.00006388, 0.00000245, 0.00003190, },
		{ -0.00000569, -0.00001018, -0.00003170, -0.00002648, 0.00001709, 0.00004213, 0.00002227, -0.00000846, -0.00002676, -0.00002872, -0.00000565, 0.00002128, 0.00001493, -0.00000057, 0.00001283, 0.00000931, -0.00003681, -0.00003419, 0.00003248, 0.00002837, -0.00003707, -0.00000547, 0.00004784, -0.00004165, -0.00008879, 0.00006392, 0.00010445, -0.00008889, -0.00009514, 0.00011351, 0.00004364, -0.00017015, -0.00001153, 0.00017094, -0.00008848, -0.00019564, 0.00017469, 0.00016851, -0.00031808, -0.00019444, 0.00031928, -0.00011940, -0.00087757, -0.00014973, 0.00131037, 0.00120510, -0.00019553, -0.00073023, -0.00013138, 0.00018158, -0.00008614, -0.00016739, 0.00001662, 0.00004080, -0.00004351, -0.00000629, 0.00005082, -0.00000469, -0.00008016, -0.00007273, -0.00001150, 0.00002882, 0.00001944, -0.00000371, },
		{ -0.00002701, -0.00003020, 0.00000264, 0.00003837, 0.00003343, 0.00000023, -0.00002668, -0.00003179, -0.00001150, 0.00001590, 0.00001773, 0.00000487, 0.00001206, 0.00001098, -0.00003055, -0.00004104, 0.00001968, 0.00003759, -0.00002329, -0.00001737, 0.00004360, -0.00001868, -0.00009746, 0.00002354, 0.00011985, -0.00004100, -0.00011728, 0.00007568, 0.00008268, -0.00015251, -0.00007472, 0.00017427, -0.00001209, -0.00023147, 0.00009593, 0.00025044, -0.00024158, -0.00032015, 0.00027735, 0.00009631, -0.00084493, -0.00050587, 0.00112131, 0.00150107, 0.00008723, -0.00082682, -0.00028773, 0.00022811, -0.00002874, -0.00022121, -0.00001001, 0.00007784, -0.00004182, -0.00003442, 0.00005671, 0.00001753, -0.00007686, -0.00008006, -0.00002130, 0.00002062, 0.00002502, 0.00000717, -0.00000640, -0.00001266, },
		{ -0.00000826, 0.00002908, 0.00003980, 0.00001167, -0.00002322, -0.00003490, -0.00001740, 0.00001041, 0.00001855, 0.00000952, 0.00001261, 0.00001254, -0.00002393, -0.00004497, 0.00000524, 0.00004116, -0.00000679, -0.00002347, 0.00003349, 0.00000136, -0.00009317, -0.00001628, 0.00011558, 0.00000829, -0.00011867, 0.00003061, 0.00010539, -0.00011658, -0.00012840, 0.00014803, 0.00006489, -0.00023052, 0.00000022, 0.00029204, -0.00012742, -0.00040828, 0.00016996, 0.00027976, -0.00070601, -0.00080913, 0.00080910, 0.00170900, 0.00045500, -0.00084449, -0.00048871, 0.00023543, 0.00006176, -0.00026177, -0.00006318, 0.00011460, -0.00002220, -0.00006874, 0.00005182, 0.00004533, -0.00007017, -0.00008914, -0.00002672, 0.00001242, 0.00002214, 0.00001796, 0.00000023, -0.00001609, -0.00002627, -0.00002957, },
		{ 0.00003980, 0.00002328, -0.00001572, -0.00003677, -0.00002370, 0.00000473, 0.00001803, 0.00001303, 0.00001380, 0.00001416, -0.00001706, -0.00004583, -0.00000930, 0.00003867, 0.00000974, -0.00002262, 0.00002047, 0.00001498, -0.00007872, -0.00004865, 0.00009422, 0.00004975, -0.00010030, -0.00001211, 0.00010851, -0.00007088, -0.00016323, 0.00009819, 0.00012734, -0.00019359, -0.00009280, 0.00028596, 0.00000218, -0.00044056, 0.00001995, 0.00039837, -0.00048690, -0.00101502, 0.00040819, 0.00178643, 0.00087144, -0.00075313, -0.00070939, 0.00018227, 0.00017463, -0.00027205, -0.00014144, 0.00013847, 0.00001890, -0.00010152, 0.00003092, 0.00007572, -0.00005579, -0.00010150, -0.00003056, 0.00000895, 0.00001258, 0.00002306, 0.00001332, -0.00001592, -0.00003011, -0.00002806, -0.00001376, 0.00001744, },
		{ -0.00000466, -0.00003598, -0.00003018, -0.00000144, 0.00001660, 0.00001545, 0.00001507, 0.00001570, -0.00001016, -0.00004370, -0.00002240, 0.00003063, 0.00002350, -0.00001512, 0.00000815, 0.00002064, -0.00005880, -0.00006903, 0.00006174, 0.00007636, -0.00006760, -0.00004381, 0.00009375, -0.00002553, -0.00017488, 0.00003557, 0.00016409, -0.00012962, -0.00016489, 0.00023583, 0.00012218, -0.00041168, -0.00014143, 0.00043428, -0.00022804, -0.00109612, -0.00002975, 0.00170912, 0.00128463, -0.00053775, -0.00091233, 0.00005620, 0.00028871, -0.00023801, -0.00023478, 0.00013596, 0.00007902, -0.00012206, -0.00000797, 0.00010220, -0.00002984, -0.00011555, -0.00003742, 0.00001255, 0.00000117, 0.00001853, 0.00002887, -0.00000837, -0.00003651, -0.00002924, -0.00001420, 0.00000710, 0.00003361, 0.00003215, },
		{ -0.00003609, -0.00000839, 0.00001442, 0.00001697, 0.00001607, 0.00001697, -0.00000356, -0.00003893, -0.00003267, 0.00001838, 0.00003219, -0.00000278, -0.00000017, 0.00001901, -0.00003865, -0.00007608, 0.00002583, 0.00008494, -0.00002893, -0.00005906, 0.00006713, 0.00001057, -0.00016472, -0.00002720, 0.00017043, -0.00005302, -0.00020362, 0.00015479, 0.00021098, -0.00033004, -0.00028220, 0.00038771, 0.00002487, -0.00104762, -0.00044567, 0.00147747, 0.00163608, -0.00020464, -0.00105413, -0.00014063, 0.00037591, -0.00015390, -0.00032554, 0.00009699, 0.00014868, -0.00011961, -0.00006166, 0.00011613, 0.00000882, -0.00012654, -0.00005166, 0.00002204, -0.00000586, 0.00000366, 0.00004089, 0.00000796, -0.00004144, -0.00003539, -0.00001192, 0.00000104, 0.00002312, 0.00003588, 0.00000833, -0.00003142, },
		{ 0.00001140, 0.00001781, 0.00001668, 0.00001779, 0.00000239, -0.00003218, -0.00003919, 0.00000393, 0.00003447, 0.00001152, -0.00000236, 0.00001270, -0.00002268, -0.00007168, -0.00000610, 0.00007667, 0.00000672, -0.00005697, 0.00003704, 0.00003205, -0.00013895, -0.00007905, 0.00014879, 0.00002031, -0.00020501, 0.00006170, 0.00025543, -0.00021484, -0.00037724, 0.00027568, 0.00023078, -0.00088773, -0.00078472, 0.00111776, 0.00187220, 0.00021679, -0.00109470, -0.00038873, 0.00040749, -0.00002594, -0.00039192, 0.00001899, 0.00021263, -0.00008706, -0.00012130, 0.00010932, 0.00005708, -0.00012781, -0.00007524, 0.00003280, -0.00000315, -0.00001835, 0.00004330, 0.00003124, -0.00004012, -0.00004651, -0.00001045, 0.00000067, 0.00001142, 0.00003322, 0.00002085, -0.00002273, -0.00004029, -0.00001613, },
		{ 0.00001688, 0.00001808, 0.00000741, -0.00002429, -0.00004162, -0.00001046, 0.00003029, 0.00002450, 0.00000198, 0.00000538, -0.00001329, -0.00006010, -0.00002877, 0.00005645, 0.00003205, -0.00004114, 0.00001171, 0.00003824, -0.00010629, -0.00011309, 0.00010731, 0.00007701, -0.00017381, -0.00002401, 0.00025328, -0.00009064, -0.00041350, 0.00012688, 0.00036172, -0.00065281, -0.00100700, 0.00067762, 0.00195541, 0.00067762, -0.00100700, -0.00065281, 0.00036172, 0.00012688, -0.00041350, -0.00009064, 0.00025328, -0.00002401, -0.00017381, 0.00007701, 0.00010731, -0.00011309, -0.00010629, 0.00003824, 0.00001171, -0.00004114, 0.00003205, 0.00005645, -0.00002877, -0.00006010, -0.00001329, 0.00000538, 0.00000198, 0.00002450, 0.00003029, -0.00001046, -0.00004162, -0.00002429, 0.00000741, 0.00001808, },
		{ 0.00001140, -0.00001613, -0.00004029, -0.00002273, 0.00002085, 0.00003322, 0.00001142, 0.00000067, -0.00001045, -0.00004651, -0.00004012, 0.00003124, 0.00004330, -0.00001835, -0.00000315, 0.00003280, -0.00007524, -0.00012781, 0.00005708, 0.00010932, -0.00012130, -0.00008706, 0.00021263, 0.00001899, -0.00039192, -0.00002594, 0.00040749, -0.00038873, -0.00109470, 0.00021679, 0.00187220, 0.00111776, -0.00078472, -0.00088773, 0.00023078, 0.00027568, -0.00037724, -0.00021484, 0.00025543, 0.00006170, -0.00020501, 0.00002031, 0.00014879, -0.00007905, -0.00013895, 0.00003205, 0.00003704, -0.00005697, 0.00000672, 0.00007667, -0.00000610, -0.00007168, -0.00002268, 0.00001270, -0.00000236, 0.00001152, 0.00003447, 0.00000393, -0.00003919, -0.00003218, 0.00000239, 0.00001779, 0.00001668, 0.00001781, },
		{ -0.00003609, -0.00003142, 0.00000833, 0.00003588, 0.00002312, 0.00000104, -0.00001192, -0.00003539, -0.00004144, 0.00000796, 0.00004089, 0.00000366, -0.00000586, 0.00002204, -0.00005166, -0.00012654, 0.00000882, 0.00011613, -0.00006166, -0.00011961, 0.00014868, 0.00009699, -0.00032554, -0.00015390, 0.00037591, -0.00014063, -0.00105413, -0.00020464, 0.00163608, 0.00147747, -0.00044567, -0.00104762, 0.00002487, 0.00038771, -0.00028220, -0.00033004, 0.00021098, 0.00015479, -0.00020362, -0.00005302, 0.00017043, -0.00002720, -0.00016472, 0.00001057, 0.00006713, -0.00005906, -0.00002893, 0.00008494, 0.00002583, -0.00007608, -0.00003865, 0.00001901, -0.00000017, -0.00000278, 0.00003219, 0.00001838, -0.00003267, -0.00003893, -0.00000356, 0.00001697, 0.00001607, 0.00001697, 0.00001442, -0.00000839, },
		{ -0.00000466, 0.00003215, 0.00003361, 0.00000710, -0.00001420, -0.00002924, -0.00003651, -0.00000837, 0.00002887, 0.00001853, 0.00000117, 0.00001255, -0.00003742, -0.00011555, -0.00002984, 0.00010220, -0.00000797, -0.00012206, 0.00007902, 0.00013596, -0.00023478, -0.00023801, 0.00028871, 0.00005620, -0.00091233, -0.00053775, 0.00128463, 0.00170912, -0.00002975, -0.00109612, -0.00022804, 0.00043428, -0.00014143, -0.00041168, 0.00012218, 0.00023583, -0.00016489, -0.00012962, 0.00016409, 0.00003557, -0.00017488, -0.00002553, 0.00009375, -0.00004381, -0.00006760, 0.00007636, 0.00006174, -0.00006903, -0.00005880, 0.00002064, 0.00000815, -0.00001512, 0.00002350, 0.00003063, -0.00002240, -0.00004370, -0.00001016, 0.00001570, 0.00001507, 0.00001545, 0.00001660, -0.00000144, -0.00003018, -0.00003598, },
		{ 0.00003980, 0.00001744, -0.00001376, -0.00002806, -0.00003011, -0.00001592, 0.00001332, 0.00002306, 0.00001258, 0.00000895, -0.00003056, -0.00010150, -0.00005579, 0.00007572, 0.00003092, -0.00010152, 0.00001890, 0.00013847, -0.00014144, -0.00027205, 0.00017463, 0.00018227, -0.00070939, -0.00075313, 0.00087144, 0.00178643, 0.00040819, -0.00101502, -0.00048690, 0.00039837, 0.00001995, -0.00044056, 0.00000218, 0.00028596, -0.00009280, -0.00019359, 0.00012734, 0.00009819, -0.00016323, -0.00007088, 0.00010851, -0.00001211, -0.00010030, 0.00004975, 0.00009422, -0.00004865, -0.00007872, 0.00001498, 0.00002047, -0.00002262, 0.00000974, 0.00003867, -0.00000930, -0.00004583, -0.00001706, 0.00001416, 0.00001380, 0.00001303, 0.00001803, 0.00000473, -0.00002370, -0.00003677, -0.00001572, 0.00002328, },
		{ -0.00000826, -0.00002957, -0.00002627, -0.00001609, 0.00000023, 0.00001796, 0.00002214, 0.00001242, -0.00002672, -0.00008914, -0.00007017, 0.00004533, 0.00005182, -0.00006874, -0.00002220, 0.00011460, -0.00006318, -0.00026177, 0.00006176, 0.00023543, -0.00048871, -0.00084449, 0.00045500, 0.00170900, 0.00080910, -0.00080913, -0.00070601, 0.00027976, 0.00016996, -0.00040828, -0.00012742, 0.00029204, 0.00000022, -0.00023052, 0.00006489, 0.00014803, -0.00012840, -0.00011658, 0.00010539, 0.00003061, -0.00011867, 0.00000829, 0.00011558, -0.00001628, -0.00009317, 0.00000136, 0.00003349, -0.00002347, -0.00000679, 0.00004116, 0.00000524, -0.00004497, -0.00002393, 0.00001254, 0.00001261, 0.00000952, 0.00001855, 0.00001041, -0.00001740, -0.00003490, -0.00002322, 0.00001167, 0.00003980, 0.00002908, },
		{ -0.00002701, -0.00001266, -0.00000640, 0.00000717, 0.00002502, 0.00002062, -0.00002130, -0.00008006, -0.00007686, 0.00001753, 0.00005671, -0.00003442, -0.00004182, 0.00007784, -0.00001001, -0.00022121, -0.00002874, 0.00022811, -0.00028773, -0.00082682, 0.00008723, 0.00150107, 0.00112131, -0.00050587, -0.00084493, 0.00009631, 0.00027735, -0.00032015, -0.00024158, 0.00025044, 0.00009593, -0.00023147, -0.00001209, 0.00017427, -0.00007472, -0.00015251, 0.00008268, 0.00007568, -0.00011728, -0.00004100, 0.00011985, 0.00002354, -0.00009746, -0.00001868, 0.00004360, -0.00001737, -0.00002329, 0.00003759, 0.00001968, -0.00004104, -0.00003055, 0.00001098, 0.00001206, 0.00000487, 0.00001773, 0.00001590, -0.00001150, -0.00003179, -0.00002668, 0.00000023, 0.00003343, 0.00003837, 0.00000264, -0.00003020, },
		{ -0.00000569, -0.00000371, 0.00001944, 0.00002882, -0.00001150, -0.00007273, -0.00008016, -0.00000469, 0.00005082, -0.00000629, -0.00004351, 0.00004080, 0.00001662, -0.00016739, -0.00008614, 0.00018158, -0.00013138, -0.00073023, -0.00019553, 0.00120510, 0.00131037, -0.00014973, -0.00087757, -0.00011940, 0.00031928, -0.00019444, -0.00031808, 0.00016851, 0.00017469, -0.00019564, -0.00008848, 0.00017094, -0.00001153, -0.00017015, 0.00004364, 0.00011351, -0.00009514, -0.00008889, 0.00010445, 0.00006392, -0.00008879, -0.00004165, 0.00004784, -0.00000547, -0.00003707, 0.00002837, 0.00003248, -0.00003419, -0.00003681, 0.00000931, 0.00001283, -0.00000057, 0.00001493, 0.00002128, -0.00000565, -0.00002872, -0.00002676, -0.00000846, 0.00002227, 0.00004213, 0.00001709, -0.00002648, -0.00003170, -0.00001018, },
		{ 0.00000719, 0.00003190, 0.00000245, -0.00006388, -0.00008283, -0.00002171, 0.00003999, 0.00001234, -0.00003428, 0.00001198, 0.00002203, -0.00011527, -0.00011026, 0.00011899, -0.00002951, -0.00059118, -0.00037721, 0.00087191, 0.00136461, 0.00020711, -0.00079792, -0.00032720, 0.00028721, -0.00005819, -0.00034320, 0.00006298, 0.00022057, -0.00013083, -0.00014904, 0.00013875, 0.00004912, -0.00016506, -0.00000421, 0.00013607, -0.00005598, -0.00012613, 0.00007095, 0.00009727, -0.00006717, -0.00006300, 0.00004465, 0.00000990, -0.00004598, 0.00001464, 0.00004233, -0.00002471, -0.00004255, 0.00000707, 0.00001548, -0.00000597, 0.00000951, 0.00002614, 0.00000084, -0.00002639, -0.00002498, -0.00001277, 0.00000921, 0.00003858, 0.00003164, -0.00001636, -0.00003747, -0.00001222, 0.00000008, -0.00000953, },
		{ 0.00001729, -0.00005045, -0.00008504, -0.00003592, 0.00002848, 0.00002226, -0.00002150, -0.00000558, 0.00001506, -0.00007430, -0.00010872, 0.00005947, 0.00002166, -0.00044381, -0.00046159, 0.00054992, 0.00129548, 0.00051554, -0.00062108, -0.00048838, 0.00018918, 0.00005940, -0.00031533, -0.00004420, 0.00022542, -0.00005155, -0.00018232, 0.00008508, 0.00009559, -0.00013818, -0.00005118, 0.00013905, -0.00000731, -0.00014572, 0.00002483, 0.00011707, -0.00003553, -0.00007822, 0.00003422, 0.00002587, -0.00004882, -0.00000190, 0.00004820, -0.00001299, -0.00004746, 0.00000352, 0.00002015, -0.00001005, 0.00000120, 0.00002956, 0.00000863, -0.00002469, -0.00002316, -0.00001246, -0.00000241, 0.00002789, 0.00004220, 0.00000000, -0.00004029, -0.00002002, 0.00000652, -0.00000743, -0.00000708, 0.00002648, },
		{ -0.00008455, -0.00004974, 0.00001777, 0.00002669, -0.00001032, -0.00001356, 0.00000431, -0.00004737, -0.00009264, 0.00001443, 0.00003493, -0.00031340, -0.00046767, 0.00027634, 0.00113268, 0.00073911, -0.00037937, -0.00057444, 0.00004783, 0.00013432, -0.00024522, -0.00013153, 0.00019098, 0.00002476, -0.00018363, 0.00002203, 0.00011969, -0.00009561, -0.00008797, 0.00012280, 0.00004162, -0.00014450, -0.00002584, 0.00011943, 0.00000088, -0.00008403, 0.00001840, 0.00003967, -0.00004541, -0.00001929, 0.00004940, 0.00000044, -0.00005095, -0.00000217, 0.00002637, -0.00001133, -0.00000964, 0.00003024, 0.00001794, -0.00002275, -0.00002287, -0.00000868, -0.00000969, 0.00001220, 0.00004533, 0.00002009, -0.00003645, -0.00003204, 0.00000887, 0.00000214, -0.00001762, 0.00001237, 0.00002806, -0.00003137, },
		{ 0.00000688, 0.00002908, -0.00000254, -0.00001628, -0.00000454, -0.00003197, -0.00007229, -0.00001297, 0.00002636, -0.00021322, -0.00042226, 0.00007193, 0.00091598, 0.00086004, -0.00011452, -0.00057314, -0.00010518, 0.00015362, -0.00015300, -0.00018343, 0.00012813, 0.00008236, -0.00015607, -0.00003678, 0.00011879, -0.00004690, -0.00010815, 0.00009208, 0.00008210, -0.00012372, -0.00007226, 0.00010402, 0.00003570, -0.00007914, 0.00000019, 0.00004918, -0.00003648, -0.00003557, 0.00004564, 0.00001480, -0.00005212, -0.00001053, 0.00003292, -0.00000852, -0.00002179, 0.00002683, 0.00002826, -0.00001927, -0.00002482, -0.00000354, -0.00001094, -0.00000496, 0.00003929, 0.00003969, -0.00002401, -0.00004442, 0.00000377, 0.00001561, -0.00001968, -0.00000712, 0.00003002, -0.00000857, -0.00007812, -0.00006392, },
		{ 0.00000305, -0.00001804, -0.00000989, -0.00002299, -0.00005415, -0.00002590, 0.00000999, -0.00014524, -0.00035240, -0.00005952, 0.00068568, 0.00088052, 0.00013175, -0.00049046, -0.00023709, 0.00011837, -0.00006262, -0.00019408, 0.00005369, 0.00011123, -0.00010939, -0.00007983, 0.00009625, -0.00000239, -0.00010972, 0.00005447, 0.00010806, -0.00008850, -0.00010691, 0.00007400, 0.00006294, -0.00006456, -0.00001705, 0.00005335, -0.00002344, -0.00004901, 0.00003700, 0.00002902, -0.00004993, -0.00002150, 0.00003801, -0.00000098, -0.00003332, 0.00001847, 0.00003824, -0.00001298, -0.00002869, 0.00000053, -0.00000613, -0.00001962, 0.00002456, 0.00005408, -0.00000355, -0.00005233, -0.00000938, 0.00002754, -0.00001135, -0.00002620, 0.00002077, 0.00001334, -0.00006339, -0.00007680, -0.00000633, 0.00003128, },
		{ -0.00001345, -0.00001562, -0.00003997, -0.00003039, -0.00000537, -0.00010351, -0.00027928, -0.00012783, 0.00047456, 0.00081932, 0.00032648, -0.00034795, -0.00032186, 0.00004273, 0.00000458, -0.00016836, -0.00001436, 0.00010936, -0.00005699, -0.00010062, 0.00006023, 0.00002956, -0.00009531, 0.00001820, 0.00011728, -0.00004623, -0.00012517, 0.00003520, 0.00007827, -0.00004329, -0.00003039, 0.00005220, -0.00000809, -0.00005817, 0.00002403, 0.00004169, -0.00004351, -0.00003421, 0.00003964, 0.00001092, -0.00004184, 0.00000515, 0.00004593, -0.00000321, -0.00003315, 0.00000150, 0.00000318, -0.00002841, 0.00000378, 0.00005938, 0.00002174, -0.00005154, -0.00002808, 0.00003254, 0.00000565, -0.00003848, 0.00000146, 0.00002855, -0.00004047, -0.00008482, -0.00002330, 0.00003274, 0.00000905, -0.00002114, },
		{ -0.00002783, -0.00003189, -0.00001684, -0.00007888, -0.00021530, -0.00015045, 0.00030280, 0.00070506, 0.00045122, -0.00017658, -0.00034597, -0.00005036, 0.00003534, -0.00011985, -0.00006182, 0.00008277, -0.00001202, -0.00009896, 0.00002108, 0.00004463, -0.00007106, -0.00001006, 0.00011148, -0.00000470, -0.00012601, -0.00000534, 0.00007995, -0.00001954, -0.00003799, 0.00004668, 0.00000774, -0.00006212, 0.00000777, 0.00005119, -0.00003246, -0.00004692, 0.00003617, 0.00002567, -0.00004508, -0.00001196, 0.00004921, 0.00000975, -0.00003626, -0.00000167, 0.00001432, -0.00002940, -0.00001891, 0.00005367, 0.00004691, -0.00003997, -0.00004744, 0.00002722, 0.00002639, -0.00003927, -0.00002319, 0.00003220, -0.00001271, -0.00008404, -0.00004335, 0.00003104, 0.00001753, -0.00002508, -0.00001840, -0.00000759, },
		{ -0.00002587, -0.00006331, -0.00016429, -0.00014605, 0.00017707, 0.00056823, 0.00050378, -0.00000873, -0.00031063, -0.00013567, 0.00002753, -0.00006637, -0.00008183, 0.00004305, 0.00001620, -0.00008044, -0.00001164, 0.00004335, -0.00004461, -0.00002679, 0.00009529, 0.00002978, -0.00011180, -0.00004098, 0.00006897, 0.00000228, -0.00003940, 0.00003832, 0.00002239, -0.00006043, -0.00001020, 0.00005595, -0.00001726, -0.00005738, 0.00002685, 0.00004075, -0.00004152, -0.00003064, 0.00004637, 0.00002453, -0.00003607, -0.00000870, 0.00002425, -0.00002252, -0.00003894, 0.00003755, 0.00006656, -0.00001844, -0.00006171, 0.00001138, 0.00004427, -0.00002727, -0.00004617, 0.00002234, 0.00001399, -0.00007206, -0.00006327, 0.00002342, 0.00002850, -0.00002726, -0.00002711, 0.00000027, -0.00001448, -0.00003291, },
		{ -0.00012414, -0.00012947, 0.00009301, 0.00043418, 0.00049573, 0.00012930, -0.00023007, -0.00019242, -0.00001017, -0.00002461, -0.00007613, 0.00000358, 0.00002445, -0.00005419, -0.00003173, 0.00003032, -0.00002273, -0.00003208, 0.00007460, 0.00005363, -0.00008732, -0.00006685, 0.00004859, 0.00001856, -0.00003548, 0.00002877, 0.00003455, -0.00005327, -0.00002791, 0.00005480, 0.00000067, -0.00006323, 0.00001218, 0.00005318, -0.00003085, -0.00004787, 0.00003667, 0.00003885, -0.00003124, -0.00001812, 0.00003036, -0.00000949, -0.00005245, 0.00001396, 0.00007636, 0.00000938, -0.00006606, -0.00001167, 0.00005319, -0.00000526, -0.00006036, 0.00000104, 0.00003290, -0.00004959, -0.00007813, 0.00000855, 0.00003958, -0.00002458, -0.00003959, 0.00000487, 0.00000225, -0.00003260, -0.00003548, -0.00005241, },
		{ 0.00004013, 0.00031879, 0.00044695, 0.00022240, -0.00012658, -0.00020933, -0.00006158, -0.00000538, -0.00005369, -0.00002464, 0.00001582, -0.00002979, -0.00003771, 0.00001241, -0.00000958, -0.00002904, 0.00005475, 0.00006627, -0.00005828, -0.00008065, 0.00002341, 0.00002718, -0.00002810, 0.00001949, 0.00004334, -0.00004143, -0.00004319, 0.00004734, 0.00001906, -0.00006271, -0.00000602, 0.00006017, -0.00001428, -0.00006057, 0.00002073, 0.00005005, -0.00002148, -0.00002767, 0.00003116, 0.00000674, -0.00005720, -0.00001259, 0.00007417, 0.00003800, -0.00005817, -0.00003614, 0.00004954, 0.00002064, -0.00006093, -0.00002585, 0.00003876, -0.00002087, -0.00008305, -0.00001218, 0.00004682, -0.00001537, -0.00005323, 0.00000287, 0.00002184, -0.00002788, -0.00004749, -0.00004580, -0.00009066, -0.00010920, },
		{ 0.00037905, 0.00026820, -0.00002391, -0.00018670, -0.00010830, -0.00001092, -0.00002719, -0.00003605, -0.00000197, -0.00001426, -0.00003259, -0.00000364, -0.00000584, -0.00002230, 0.00003920, 0.00006963, -0.00002985, -0.00008263, -0.00000180, 0.00002775, -0.00001953, 0.00001148, 0.00004830, -0.00002627, -0.00005403, 0.00003423, 0.00003515, -0.00005513, -0.00002486, 0.00005986, 0.00000564, -0.00006623, 0.00000051, 0.00005569, -0.00000779, -0.00003493, 0.00002653, 0.00002280, -0.00005290, -0.00003724, 0.00006057, 0.00006155, -0.00003900, -0.00005547, 0.00003348, 0.00004278, -0.00004709, -0.00005030, 0.00002982, 0.00000724, -0.00007528, -0.00003454, 0.00004647, -0.00000058, -0.00006373, -0.00000739, 0.00004101, -0.00001569, -0.00006091, -0.00004560, -0.00006089, -0.00008767, 0.00000698, 0.00022759, },
		{ 0.00005895, -0.00013510, -0.00013529, -0.00003492, -0.00000862, -0.00003174, -0.00001964, -0.00001018, -0.00002223, -0.00001331, -0.00000910, -0.00001632, 0.00002893, 0.00006699, -0.00000562, -0.00007511, -0.00002310, 0.00002151, -0.00001190, 0.00000519, 0.00004943, -0.00000955, -0.00005893, 0.00001717, 0.00004634, -0.00004125, -0.00004105, 0.00005186, 0.00002551, -0.00006361, -0.00002098, 0.00005425, 0.00000777, -0.00003799, 0.00001771, 0.00003573, -0.00004115, -0.00005589, 0.00003860, 0.00007544, -0.00001253, -0.00006436, 0.00000900, 0.00005444, -0.00002250, -0.00006462, 0.00000880, 0.00002725, -0.00005567, -0.00005254, 0.00003685, 0.00001611, -0.00006705, -0.00002460, 0.00005492, 0.00000498, -0.00007204, -0.00005396, -0.00003490, -0.00006361, -0.00001479, 0.00015792, 0.00030954, 0.00027507, },
		{ -0.00013506, -0.00006532, -0.00000540, -0.00001833, -0.00002966, -0.00001539, -0.00001295, -0.00001538, -0.00001519, -0.00001390, 0.00002276, 0.00006167, 0.00001283, -0.00006143, -0.00003806, 0.00001075, -0.00000673, 0.00000064, 0.00004710, 0.00000674, -0.00005725, -0.00000129, 0.00005077, -0.00002319, -0.00005168, 0.00003740, 0.00004185, -0.00005308, -0.00004031, 0.00004550, 0.00002251, -0.00003596, 0.00000680, 0.00004365, -0.00002477, -0.00006614, 0.00001280, 0.00007750, 0.00001537, -0.00006039, -0.00001731, 0.00005194, 0.00000565, -0.00006395, -0.00001757, 0.00003359, -0.00002901, -0.00006050, 0.00001954, 0.00002957, -0.00006129, -0.00004464, 0.00005933, 0.00003193, -0.00007593, -0.00007085, -0.00001562, -0.00003519, -0.00002866, 0.00010303, 0.00024843, 0.00025748, 0.00011198, -0.00007120, },
		{ -0.00001814, -0.00000516, -0.00002877, -0.00002436, -0.00000919, -0.00001186, -0.00001992, -0.00001558, 0.00001833, 0.00005601, 0.00002568, -0.00004498, -0.00004589, -0.00000186, -0.00000477, -0.00000253, 0.00004200, 0.00002072, -0.00004944, -0.00001826, 0.00004785, -0.00000404, -0.00005492, 0.00001916, 0.00005189, -0.00003657, -0.00005440, 0.00003070, 0.00003376, -0.00002917, -0.00000380, 0.00004604, -0.00000702, -0.00006757, -0.00001201, 0.00006838, 0.00003867, -0.00004474, -0.00003827, 0.00003589, 0.00002885, -0.00004801, -0.00004053, 0.00002468, -0.00000280, -0.00005530, -0.00000061, 0.00003513, -0.00004775, -0.00006190, 0.00005249, 0.00006011, -0.00006868, -0.00009302, -0.00000715, -0.00000265, -0.00003392, 0.00005634, 0.00019782, 0.00023055, 0.00013518, -0.00001229, -0.00010930, -0.00008862, },
		{ -0.00001879, -0.00003060, -0.00001207, -0.00000671, -0.00002066, -0.00001981, 0.00001332, 0.00005086, 0.00003419, -0.00002841, -0.00004724, -0.00001385, -0.00000592, -0.00000481, 0.00003505, 0.00003094, -0.00003694, -0.00003108, 0.00003847, 0.00001287, -0.00005052, 0.00000056, 0.00005426, -0.00001713, -0.00006120, 0.00001225, 0.00003951, -0.00001907, -0.00001207, 0.00004362, 0.00000927, -0.00006155, -0.00003194, 0.00005121, 0.00005283, -0.00002177, -0.00004830, 0.00001102, 0.00003995, -0.00002141, -0.00005198, 0.00000382, 0.00001494, -0.00003782, -0.00001693, 0.00003072, -0.00003082, -0.00007141, 0.00003620, 0.00008322, -0.00004933, -0.00011460, -0.00001231, 0.00003060, -0.00002797, 0.00001461, 0.00015404, 0.00020532, 0.00013663, 0.00002881, -0.00006756, -0.00009446, -0.00004086, -0.00000067, },
		{ -0.00001922, -0.00000403, -0.00001715, -0.00002392, 0.00000642, 0.00004577, 0.00003982, -0.00001327, -0.00004357, -0.00002343, -0.00000949, -0.00000674, 0.00002722, 0.00003663, -0.00002196, -0.00003796, 0.00002479, 0.00002474, -0.00003989, -0.00001501, 0.00004926, 0.00000176, -0.00006013, -0.00000675, 0.00003887, -0.00000782, -0.00001680, 0.00003795, 0.00002224, -0.00005054, -0.00004483, 0.00003044, 0.00005596, 0.00000235, -0.00004514, -0.00001526, 0.00003541, 0.00000764, -0.00004722, -0.00002154, 0.00001848, -0.00001323, -0.00002316, 0.00001797, -0.00001644, -0.00007088, 0.00001541, 0.00009607, -0.00002071, -0.00012892, -0.00003073, 0.00005871, -0.00000921, -0.00002117, 0.00011135, 0.00018629, 0.00012834, 0.00004700, -0.00002359, -0.00007901, -0.00006340, -0.00000925, -0.00000568, -0.00002934, },
		{ -0.00001132, -0.00002540, -0.00000214, 0.00003964, 0.00004358, -0.00000008, -0.00003663, -0.00002969, -0.00001442, -0.00000874, 0.00001940, 0.00003781, -0.00000698, -0.00003836, 0.00000980, 0.00003000, -0.00002571, -0.00002506, 0.00003876, 0.00001703, -0.00005216, -0.00002327, 0.00003225, 0.00000226, -0.00001781, 0.00003089, 0.00003120, -0.00003726, -0.00005049, 0.00001056, 0.00004913, 0.00002154, -0.00003044, -0.00003517, 0.00001647, 0.00002984, -0.00002663, -0.00004199, 0.00000648, 0.00001045, -0.00001590, 0.00000210, -0.00000985, -0.00006173, -0.00000346, 0.00009652, 0.00001120, -0.00013098, -0.00005824, 0.00007544, 0.00002085, -0.00004680, 0.00006562, 0.00017153, 0.00012156, 0.00004543, 0.00000934, -0.00004600, -0.00007512, -0.00002961, 0.00000293, -0.00001942, -0.00002601, -0.00000624, },
		{ -0.00001088, 0.00003145, 0.00004569, 0.00001130, -0.00002794, -0.00003245, -0.00001957, -0.00001111, 0.00001226, 0.00003522, 0.00000576, -0.00003306, -0.00000351, 0.00002868, -0.00001124, -0.00002855, 0.00002561, 0.00002671, -0.00003948, -0.00003495, 0.00002117, 0.00000928, -0.00001583, 0.00002402, 0.00003650, -0.00002395, -0.00005024, -0.00000512, 0.00003576, 0.00003155, -0.00000916, -0.00004301, -0.00001116, 0.00003770, 0.00000414, -0.00004916, -0.00001719, 0.00002482, 0.00000388, -0.00000967, -0.00001341, -0.00004874, -0.00001488, 0.00008645, 0.00003926, -0.00011935, -0.00008779, 0.00007665, 0.00005700, -0.00005749, 0.00001669, 0.00015513, 0.00012294, 0.00003333, 0.00002271, -0.00000528, -0.00006880, -0.00005522, 0.00000080, -0.00000377, -0.00002754, -0.00001314, -0.00000629, -0.00002302, },
		{ 0.00004573, 0.00002115, -0.00001855, -0.00003211, -0.00002398, -0.00001394, 0.00000609, 0.00003013, 0.00001478, -0.00002396, -0.00001280, 0.00002226, 0.00000055, -0.00002610, 0.00001287, 0.00003032, -0.00002490, -0.00004060, 0.00000792, 0.00001211, -0.00001222, 0.00001833, 0.00003904, -0.00001190, -0.00004624, -0.00001518, 0.00002033, 0.00003110, 0.00001207, -0.00003685, -0.00003908, 0.00002797, 0.00003580, -0.00003858, -0.00004436, 0.00002403, 0.00003068, -0.00001063, -0.00002538, -0.00003836, -0.00001633, 0.00007115, 0.00005767, -0.00009686, -0.00011168, 0.00006212, 0.00009153, -0.00005060, -0.00003123, 0.00013070, 0.00013279, 0.00002212, 0.00001518, 0.00003061, -0.00004321, -0.00007654, -0.00001444, 0.00001159, -0.00002075, -0.00002190, -0.00000492, -0.00001732, -0.00001774, 0.00002101, },
		{ -0.00000905, -0.00002936, -0.00002691, -0.00001710, 0.00000083, 0.00002392, 0.00001963, -0.00001350, -0.00001697, 0.00001328, 0.00000776, -0.00001965, 0.00000302, 0.00002877, -0.00001110, -0.00004033, -0.00000507, 0.00001063, -0.00000851, 0.00001409, 0.00003988, -0.00000148, -0.00004053, -0.00002004, 0.00000698, 0.00002201, 0.00002703, -0.00001906, -0.00005887, 0.00000271, 0.00005822, -0.00001139, -0.00006503, 0.00000694, 0.00005598, 0.00000286, -0.00004015, -0.00003622, -0.00000934, 0.00005741, 0.00006400, -0.00006971, -0.00012409, 0.00003600, 0.00011658, -0.00002751, -0.00007121, 0.00009474, 0.00014586, 0.00002116, -0.00000746, 0.00005061, -0.00000373, -0.00008442, -0.00004009, 0.00001982, -0.00000591, -0.00002821, -0.00000845, -0.00001035, -0.00002104, 0.00000910, 0.00004301, 0.00002956, },
		{ -0.00002795, -0.00002026, -0.00000388, 0.00001772, 0.00002082, -0.00000397, -0.00001634, 0.00000449, 0.00000993, -0.00001186, -0.00000268, 0.00002400, -0.00000006, -0.00003534, -0.00001574, 0.00000560, -0.00000601, 0.00001098, 0.00003975, 0.00000765, -0.00003446, -0.00002142, -0.00000175, 0.00000831, 0.00003183, 0.00000456, -0.00006462, -0.00003124, 0.00006360, 0.00002573, -0.00007061, -0.00002208, 0.00007101, 0.00002981, -0.00004998, -0.00004484, 0.00000093, 0.00005100, 0.00006015, -0.00004533, -0.00012310, 0.00000556, 0.00012677, 0.00000611, -0.00009635, 0.00004847, 0.00015392, 0.00003460, -0.00003464, 0.00004875, 0.00003928, -0.00007336, -0.00006907, 0.00001585, 0.00001327, -0.00002804, -0.00001589, -0.00000481, -0.00002025, -0.00000270, 0.00003713, 0.00003625, 0.00000026, -0.00002492, },
		{ -0.00000846, 0.00001217, 0.00001960, 0.00000304, -0.00001245, -0.00000189, 0.00000807, -0.00000526, -0.00000439, 0.00001827, 0.00000736, -0.00002746, -0.00002279, -0.00000159, -0.00000551, 0.00000834, 0.00003893, 0.00001601, -0.00002839, -0.00002143, -0.00000537, -0.00000520, 0.00002589, 0.00002667, -0.00005478, -0.00006438, 0.00004896, 0.00006241, -0.00005685, -0.00005430, 0.00006970, 0.00006443, -0.00004765, -0.00006231, 0.00000762, 0.00005446, 0.00005154, -0.00002969, -0.00011131, -0.00002113, 0.00012107, 0.00004180, -0.00010262, -0.00000221, 0.00014942, 0.00006021, -0.00005481, 0.00002600, 0.00007382, -0.00004364, -0.00009213, -0.00000157, 0.00003071, -0.00001937, -0.00002424, -0.00000284, -0.00001620, -0.00001263, 0.00002831, 0.00004068, 0.00000917, -0.00001948, -0.00002700, -0.00002288, },
		{ 0.00001744, 0.00000704, -0.00000744, -0.00000478, 0.00000424, -0.00000149, -0.00000345, 0.00001346, 0.00001144, -0.00001863, -0.00002586, -0.00000934, -0.00000720, 0.00000548, 0.00003731, 0.00002389, -0.00002189, -0.00002161, -0.00000531, -0.00001456, 0.00001196, 0.00004077, -0.00003250, -0.00008744, 0.00001723, 0.00008773, -0.00002552, -0.00007927, 0.00005094, 0.00009786, -0.00002911, -0.00008262, 0.00000471, 0.00006601, 0.00004517, -0.00002523, -0.00009481, -0.00003798, 0.00010311, 0.00007089, -0.00009061, -0.00004870, 0.00012850, 0.00009052, -0.00005935, -0.00001021, 0.00009003, -0.00000149, -0.00010092, -0.00002902, 0.00004007, -0.00000319, -0.00002957, -0.00000508, -0.00001078, -0.00001940, 0.00001746, 0.00004235, 0.00001737, -0.00001361, -0.00002423, -0.00002430, -0.00001320, 0.00000720, },
		{ -0.00000335, -0.00000451, 0.00000072, -0.00000090, -0.00000185, 0.00001053, 0.00001328, -0.00001034, -0.00002541, -0.00001623, -0.00001071, 0.00000192, 0.00003453, 0.00003113, -0.00001423, -0.00002244, -0.00000401, -0.00001787, -0.00000508, 0.00004312, -0.00000455, -0.00009427, -0.00002362, 0.00009358, 0.00001580, -0.00008824, 0.00001930, 0.00012095, 0.00000480, -0.00009748, -0.00001047, 0.00008010, 0.00004688, -0.00002994, -0.00008096, -0.00004286, 0.00007984, 0.00008730, -0.00006562, -0.00008270, 0.00009267, 0.00011575, -0.00004534, -0.00004848, 0.00008335, 0.00004265, -0.00009090, -0.00005930, 0.00003702, 0.00001648, -0.00002856, -0.00001027, -0.00000615, -0.00002271, 0.00000595, 0.00004108, 0.00002453, -0.00000782, -0.00002010, -0.00002391, -0.00001803, 0.00000228, 0.00001551, 0.00000866, },
		},
		{ { 0.00000925, 0.00000712, 0.00000224, -0.00001039, -0.00002662, -0.00001181, 0.00002478, 0.00001686, -0.00001039, -0.00000118, -0.00002142, -0.00004766, 0.00002538, 0.00004386, -0.00007477, -0.00005942, 0.00005206, -0.00002189, -0.00004069, 0.00014715, 0.00010889, -0.00010046, -0.00000908, 0.00012165, -0.00003799, -0.00010519, 0.00003682, 0.00003972, -0.00004470, -0.00001344, 0.00002492, 0.00000271, -0.00001278, 0.00000271, 0.00002492, -0.00001344, -0.00004470, 0.00003972, 0.00003682, -0.00010519, -0.00003799, 0.00012165, -0.00000908, -0.00010046, 0.00010889, 0.00014715, -0.00004069, -0.00002189, 0.00005206, -0.00005942, -0.00007477, 0.00004386, 0.00002538, -0.00004766, -0.00002142, -0.00000118, -0.00001039, 0.00001686, 0.00002478, -0.00001181, -0.00002662, -0.00001039, 0.00000224, 0.00000712, },
		{ 0.00000568, -0.00000710, -0.00002360, -0.00002143, 0.00001240, 0.00002670, -0.00000217, -0.00000889, -0.00000678, -0.00004281, -0.00001610, 0.00005216, -0.00001909, -0.00009304, 0.00000323, 0.00002342, -0.00004857, 0.00008088, 0.00017075, -0.00002688, -0.00007648, 0.00010103, 0.00003665, -0.00012175, -0.00001681, 0.00006750, -0.00002899, -0.00003737, 0.00002811, 0.00000903, -0.00001722, 0.00000273, 0.00002086, -0.00000020, -0.00004019, 0.00000652, 0.00005800, -0.00005498, -0.00009067, 0.00008629, 0.00006971, -0.00011148, 0.00000827, 0.00017650, 0.00001785, -0.00007507, 0.00005031, 0.00000972, -0.00009876, -0.00000447, 0.00005883, -0.00002518, -0.00004187, -0.00000102, -0.00000846, 0.00000246, 0.00002900, 0.00000290, -0.00002654, -0.00001485, 0.00000025, 0.00000307, 0.00000768, 0.00001088, },
		{ -0.00002040, -0.00002528, -0.00000167, 0.00002649, 0.00001202, -0.00001234, -0.00000636, -0.00002456, -0.00003842, 0.00002488, 0.00002703, -0.00007483, -0.00005715, 0.00002437, -0.00002007, 0.00002290, 0.00017309, 0.00007261, -0.00008992, 0.00003524, 0.00009184, -0.00008884, -0.00008087, 0.00006538, 0.00000912, -0.00005852, 0.00001646, 0.00002635, -0.00002319, -0.00000350, 0.00002389, 0.00000465, -0.00003041, -0.00001368, 0.00004993, -0.00000317, -0.00009918, 0.00001982, 0.00011419, -0.00006048, -0.00008242, 0.00014112, 0.00010027, -0.00008895, -0.00000489, 0.00006833, -0.00006745, -0.00006535, 0.00005581, 0.00001786, -0.00004924, -0.00001431, -0.00000177, -0.00000785, 0.00002291, 0.00001761, -0.00002054, -0.00001984, -0.00000156, 0.00000022, 0.00000368, 0.00001258, 0.00001054, -0.00000367, },
		{ -0.00001226, 0.00001789, 0.00002317, -0.00000543, -0.00001388, -0.00001206, -0.00003507, -0.00001038, 0.00003677, -0.00002675, -0.00008580, -0.00001740, 0.00000440, 0.00000026, 0.00013378, 0.00015060, -0.00003925, -0.00003390, 0.00009605, -0.00002258, -0.00012034, 0.00002399, 0.00005126, -0.00005741, -0.00001387, 0.00004382, -0.00001643, -0.00002102, 0.00002769, 0.00001247, -0.00002595, -0.00002046, 0.00003186, 0.00002710, -0.00007201, -0.00003856, 0.00010576, 0.00001620, -0.00011989, 0.00005368, 0.00015699, -0.00004281, -0.00007589, 0.00007412, 0.00000394, -0.00009659, 0.00001097, 0.00005460, -0.00003015, -0.00003319, -0.00000071, -0.00000983, 0.00001082, 0.00002627, -0.00000901, -0.00002283, -0.00000414, -0.00000115, -0.00000135, 0.00001138, 0.00001537, 0.00000121, -0.00001755, -0.00002600, },
		{ 0.00002571, 0.00000788, -0.00001728, -0.00001329, -0.00002016, -0.00002785, 0.00001482, 0.00001230, -0.00006821, -0.00006622, -0.00000626, 0.00000342, 0.00008931, 0.00018147, 0.00004850, -0.00006187, 0.00005142, 0.00003778, -0.00011294, -0.00003932, 0.00006961, -0.00002694, -0.00004879, 0.00004363, 0.00000829, -0.00003962, 0.00001899, 0.00002840, -0.00002341, -0.00002516, 0.00001925, 0.00003503, -0.00003539, -0.00006405, 0.00006087, 0.00007349, -0.00009511, -0.00004072, 0.00015226, 0.00004092, -0.00011182, 0.00002074, 0.00006961, -0.00007288, -0.00004960, 0.00005707, 0.00000905, -0.00004054, -0.00001018, -0.00000744, -0.00000030, 0.00002652, 0.00000420, -0.00002184, -0.00000699, -0.00000132, -0.00000645, 0.00000752, 0.00001882, 0.00000724, -0.00001381, -0.00002587, -0.00001876, 0.00000677, },
		{ -0.00001080, -0.00002139, -0.00001204, -0.00002275, -0.00001225, 0.00001841, -0.00002635, -0.00008374, -0.00004792, -0.00000000, 0.00006211, 0.00017346, 0.00013181, -0.00003109, -0.00000731, 0.00005770, -0.00006722, -0.00009052, 0.00004875, 0.00001791, -0.00006548, 0.00001801, 0.00003897, -0.00004184, -0.00000690, 0.00004340, -0.00001165, -0.00003493, 0.00001205, 0.00003488, -0.00000987, -0.00005935, 0.00001290, 0.00008819, -0.00003567, -0.00009666, 0.00008969, 0.00011317, -0.00008614, -0.00005678, 0.00008580, -0.00000660, -0.00008499, 0.00001925, 0.00004337, -0.00002482, -0.00002306, -0.00000833, -0.00000721, 0.00002091, 0.00001495, -0.00001734, -0.00000914, 0.00000017, -0.00001095, 0.00000101, 0.00002060, 0.00001328, -0.00000870, -0.00002463, -0.00002324, -0.00000279, 0.00002116, 0.00001907, },
		{ -0.00001692, -0.00000996, -0.00002140, -0.00000144, 0.00000342, -0.00006045, -0.00008653, -0.00003211, 0.00004390, 0.00015163, 0.00018457, 0.00004175, -0.00003799, 0.00003223, -0.00001687, -0.00010396, -0.00000063, 0.00004938, -0.00005142, -0.00002127, 0.00005451, -0.00001955, -0.00003816, 0.00004172, 0.00001335, -0.00004366, 0.00000016, 0.00003786, 0.00000307, -0.00004433, -0.00001494, 0.00006919, 0.00001834, -0.00009826, 0.00000719, 0.00013425, -0.00001488, -0.00010788, 0.00004254, 0.00005950, -0.00006961, -0.00003520, 0.00004720, 0.00000741, -0.00002515, -0.00001483, -0.00001262, 0.00001320, 0.00002198, -0.00001127, -0.00001103, 0.00000405, -0.00001299, -0.00000877, 0.00002033, 0.00001939, -0.00000349, -0.00002126, -0.00002634, -0.00001103, 0.00001430, 0.00002363, 0.00000135, -0.00002469, },
		{ -0.00001133, -0.00002084, 0.00000187, -0.00002111, -0.00008949, -0.00008193, 0.00000968, 0.00012913, 0.00020758, 0.00012429, -0.00001910, -0.00001213, 0.00000468, -0.00008152, -0.00004912, 0.00004824, -0.00001592, -0.00005082, 0.00004241, 0.00001671, -0.00005424, 0.00001664, 0.00004061, -0.00003823, -0.00002153, 0.00004113, 0.00001495, -0.00003470, -0.00002291, 0.00004247, 0.00004344, -0.00006425, -0.00005246, 0.00010028, 0.00005755, -0.00010290, -0.00003003, 0.00008299, -0.00001305, -0.00006870, 0.00001546, 0.00003361, -0.00000860, -0.00001958, -0.00002049, 0.00000381, 0.00002649, -0.00000420, -0.00001461, 0.00000913, -0.00000911, -0.00002127, 0.00001561, 0.00002691, 0.00000102, -0.00001697, -0.00002614, -0.00001937, 0.00000704, 0.00002375, 0.00001158, -0.00001885, -0.00002660, -0.00000647, },
		{ -0.00001936, -0.00000059, -0.00005517, -0.00011395, -0.00005119, 0.00009455, 0.00021343, 0.00019427, 0.00004188, -0.00003869, -0.00001159, -0.00004986, -0.00007147, 0.00001873, 0.00001670, -0.00005423, 0.00000975, 0.00004482, -0.00004293, -0.00002074, 0.00005152, -0.00001331, -0.00004441, 0.00003313, 0.00003357, -0.00002952, -0.00002623, 0.00002659, 0.00004401, -0.00002731, -0.00007080, 0.00004085, 0.00009213, -0.00005181, -0.00008524, 0.00005202, 0.00004652, -0.00005765, -0.00003014, 0.00003219, 0.00001810, -0.00001139, -0.00002875, -0.00001022, 0.00002813, 0.00000591, -0.00002044, 0.00001077, 0.00000325, -0.00003192, 0.00000245, 0.00003545, 0.00000726, -0.00001515, -0.00002158, -0.00002593, -0.00000285, 0.00002307, 0.00001760, -0.00000919, -0.00003002, -0.00001550, 0.00000115, -0.00002012, },
		{ -0.00001444, -0.00010083, -0.00011664, 0.00002818, 0.00020219, 0.00024818, 0.00012355, -0.00002550, -0.00004662, -0.00003824, -0.00006561, -0.00001617, 0.00002680, -0.00003489, -0.00002259, 0.00004851, -0.00001102, -0.00004858, 0.00003591, 0.00002032, -0.00005244, 0.00000833, 0.00005272, -0.00001698, -0.00003578, 0.00002007, 0.00003967, -0.00000731, -0.00006016, -0.00000735, 0.00008107, 0.00000768, -0.00009396, -0.00000823, 0.00007068, -0.00001049, -0.00005617, 0.00000115, 0.00003386, 0.00001255, -0.00002767, -0.00002944, 0.00002237, 0.00002069, -0.00002451, 0.00000311, 0.00002130, -0.00003232, -0.00002039, 0.00003952, 0.00002000, -0.00001701, -0.00001646, -0.00002573, -0.00001661, 0.00002020, 0.00002354, -0.00000234, -0.00002538, -0.00002636, 0.00000036, -0.00000311, -0.00003232, -0.00001235, },
		{ -0.00014561, -0.00006786, 0.00015457, 0.00028635, 0.00021123, 0.00002725, -0.00007082, -0.00005660, -0.00005170, -0.00003487, 0.00001432, -0.00001153, -0.00003720, 0.00002959, 0.00002114, -0.00005120, 0.00000182, 0.00004255, -0.00003718, -0.00002402, 0.00005780, 0.00000963, -0.00004674, 0.00000922, 0.00004332, -0.00000048, -0.00004586, -0.00002757, 0.00004705, 0.00004098, -0.00006259, -0.00005703, 0.00004841, 0.00003896, -0.00004308, -0.00003776, 0.00002151, 0.00003953, -0.00000765, -0.00004611, 0.00000348, 0.00003621, -0.00001909, -0.00001510, 0.00003511, -0.00001544, -0.00004629, 0.00002976, 0.00004060, -0.00001710, -0.00001809, -0.00001677, -0.00002897, 0.00000921, 0.00003235, 0.00000290, -0.00002050, -0.00002842, -0.00001262, 0.00000969, -0.00002074, -0.00003762, -0.00000273, -0.00005108, },
		{ 0.00005300, 0.00029257, 0.00030059, 0.00011007, -0.00006456, -0.00009172, -0.00005338, -0.00003439, -0.00000412, -0.00000104, -0.00003311, 0.00000422, 0.00003601, -0.00003072, -0.00003022, 0.00003997, -0.00000682, -0.00004588, 0.00004058, 0.00004216, -0.00004467, -0.00001395, 0.00004991, 0.00000930, -0.00004219, -0.00002958, 0.00001965, 0.00004061, -0.00002244, -0.00006906, 0.00000162, 0.00005759, -0.00000055, -0.00005452, -0.00001603, 0.00004767, 0.00002906, -0.00004566, -0.00002759, 0.00004164, 0.00000087, -0.00003586, 0.00003223, 0.00001662, -0.00006067, 0.00000023, 0.00006113, -0.00000514, -0.00002988, -0.00000580, -0.00002999, -0.00001218, 0.00003841, 0.00001448, -0.00002151, -0.00002368, -0.00002296, 0.00000357, 0.00000468, -0.00004460, -0.00002734, -0.00000547, -0.00011445, -0.00015620, },
		{ 0.00038050, 0.00021925, -0.00002259, -0.00012179, -0.00007923, -0.00003005, -0.00001220, -0.00000497, -0.00002314, -0.00001212, 0.00003119, -0.00000351, -0.00004311, 0.00001638, 0.00001951, -0.00004536, 0.00000841, 0.00006316, -0.00002122, -0.00004137, 0.00004429, 0.00003056, -0.00004286, -0.00003365, 0.00001100, 0.00002543, -0.00000185, -0.00005039, -0.00003482, 0.00003765, 0.00004012, -0.00003434, -0.00005454, 0.00002297, 0.00006386, -0.00001741, -0.00005770, 0.00002568, 0.00003012, -0.00004423, 0.00000740, 0.00004928, -0.00004906, -0.00004282, 0.00006545, 0.00002498, -0.00004407, -0.00000610, -0.00001460, -0.00003517, 0.00002984, 0.00003622, -0.00002272, -0.00002453, -0.00001969, -0.00001485, 0.00001754, -0.00001977, -0.00005974, -0.00000302, -0.00004361, -0.00018536, -0.00008722, 0.00023450, },
		{ 0.00005830, -0.00013178, -0.00012055, -0.00003961, -0.00000841, -0.00001159, -0.00001963, -0.00001661, 0.00001758, 0.00001429, -0.00003510, -0.00001121, 0.00002684, -0.00002654, -0.00002087, 0.00006073, 0.00001596, -0.00005529, 0.00001855, 0.00005436, -0.00003189, -0.00004886, 0.00001026, 0.00001928, -0.00000485, -0.00002816, -0.00004112, 0.00000034, 0.00005072, 0.00000985, -0.00006521, -0.00002639, 0.00007140, 0.00003304, -0.00006577, -0.00001241, 0.00005199, -0.00002853, -0.00003000, 0.00006180, -0.00000913, -0.00007930, 0.00003962, 0.00006542, -0.00004339, -0.00002634, 0.00000841, -0.00004273, 0.00000072, 0.00005804, -0.00000964, -0.00003831, -0.00000968, -0.00002411, 0.00000277, 0.00001758, -0.00006099, -0.00004524, 0.00001195, -0.00013246, -0.00020913, 0.00009241, 0.00041553, 0.00035358, },
		{ -0.00016410, -0.00006892, -0.00000475, -0.00001014, -0.00002319, -0.00001732, 0.00000678, 0.00001833, -0.00001777, -0.00002686, 0.00001523, -0.00000527, -0.00003330, 0.00003841, 0.00004843, -0.00004410, -0.00001816, 0.00006259, -0.00000247, -0.00006508, -0.00000110, 0.00002897, -0.00001280, -0.00002486, -0.00002582, -0.00002365, 0.00002788, 0.00004684, -0.00003709, -0.00007180, 0.00003827, 0.00008111, -0.00003699, -0.00005621, 0.00004772, 0.00000832, -0.00005832, 0.00004123, 0.00004277, -0.00008527, -0.00001460, 0.00009373, -0.00001344, -0.00005902, 0.00001842, -0.00002345, -0.00003787, 0.00005977, 0.00002529, -0.00005456, -0.00001278, -0.00001083, -0.00002635, 0.00003182, -0.00001861, -0.00009102, 0.00000513, -0.00003096, -0.00024463, -0.00010593, 0.00035420, 0.00049222, 0.00018926, -0.00011285, },
		{ -0.00001296, -0.00000180, -0.00002531, -0.00002151, 0.00000114, 0.00001503, -0.00000391, -0.00002683, -0.00000282, 0.00000440, -0.00002890, 0.00001131, 0.00006138, -0.00001253, -0.00004600, 0.00004510, 0.00003417, -0.00006416, -0.00002908, 0.00004041, -0.00000494, -0.00003965, -0.00001468, -0.00001987, -0.00000572, 0.00005151, 0.00001296, -0.00008307, -0.00002346, 0.00009604, 0.00002270, -0.00007820, 0.00001149, 0.00004631, -0.00005716, -0.00000577, 0.00007808, -0.00004983, -0.00007464, 0.00008561, 0.00004229, -0.00008028, -0.00000205, 0.00001469, -0.00006060, 0.00002805, 0.00006914, -0.00004959, -0.00004087, 0.00001309, -0.00003842, 0.00000483, 0.00003810, -0.00008969, -0.00006160, 0.00004495, -0.00016666, -0.00027978, 0.00016823, 0.00057771, 0.00037757, -0.00004822, -0.00020221, -0.00011301, },
		{ -0.00002063, -0.00002748, -0.00000419, 0.00001146, 0.00000233, -0.00001856, -0.00001449, 0.00000048, -0.00001898, -0.00000686, 0.00005402, 0.00002248, -0.00005005, 0.00000952, 0.00005729, -0.00003830, -0.00006049, 0.00003400, 0.00002156, -0.00005165, -0.00002628, -0.00000064, -0.00002131, 0.00002436, 0.00005200, -0.00005090, -0.00008042, 0.00006210, 0.00008451, -0.00005737, -0.00004054, 0.00005939, -0.00002317, -0.00005370, 0.00007375, 0.00001356, -0.00010745, 0.00003367, 0.00009813, -0.00006484, -0.00005016, 0.00004435, -0.00004599, -0.00002714, 0.00009206, -0.00000766, -0.00008037, 0.00001735, -0.00001383, -0.00004357, 0.00006017, -0.00002614, -0.00013220, 0.00003356, -0.00002116, -0.00033644, -0.00010506, 0.00053216, 0.00059630, 0.00009286, -0.00022533, -0.00016875, -0.00003476, 0.00000396, },
		{ -0.00001161, 0.00000808, 0.00000433, -0.00001068, -0.00001618, -0.00000863, -0.00001455, -0.00001354, 0.00003705, 0.00004531, -0.00003095, -0.00002449, 0.00005367, 0.00000149, -0.00007444, 0.00000400, 0.00004915, -0.00004033, -0.00005546, 0.00000739, -0.00000823, -0.00000958, 0.00005571, 0.00000494, -0.00009915, -0.00000739, 0.00011201, 0.00000334, -0.00007651, 0.00003431, 0.00002430, -0.00007360, 0.00002923, 0.00007165, -0.00009062, -0.00004214, 0.00011945, -0.00000681, -0.00009854, 0.00003657, 0.00000351, -0.00007311, 0.00006911, 0.00005932, -0.00009682, -0.00001937, 0.00003254, -0.00006869, 0.00002146, 0.00006054, -0.00013829, -0.00006321, 0.00009346, -0.00023659, -0.00035797, 0.00030939, 0.00076461, 0.00033381, -0.00020321, -0.00024002, -0.00006798, 0.00000231, -0.00001384, -0.00002897, },
		{ 0.00000475, -0.00000571, -0.00001243, -0.00001410, -0.00001758, -0.00001561, 0.00002162, 0.00005178, -0.00000204, -0.00004021, 0.00002764, 0.00003365, -0.00006053, -0.00003539, 0.00005573, -0.00000455, -0.00007820, -0.00001273, 0.00001772, -0.00002166, 0.00002678, 0.00004916, -0.00006949, -0.00007565, 0.00008511, 0.00007399, -0.00006899, -0.00001675, 0.00005435, -0.00005321, -0.00003073, 0.00009134, -0.00002963, -0.00010291, 0.00008625, 0.00007028, -0.00011082, -0.00001566, 0.00005717, -0.00007563, 0.00000376, 0.00011292, -0.00006149, -0.00008473, 0.00005818, -0.00003899, -0.00005426, 0.00010370, -0.00005565, -0.00017376, 0.00009211, -0.00003481, -0.00046406, -0.00005062, 0.00077042, 0.00064880, -0.00008057, -0.00032176, -0.00012121, -0.00000176, -0.00001231, -0.00002628, -0.00001635, 0.00000212, },
		{ -0.00000783, -0.00001391, -0.00002252, -0.00001997, 0.00001079, 0.00004788, 0.00002263, -0.00003456, -0.00000364, 0.00004266, -0.00002735, -0.00006162, 0.00003372, 0.00003605, -0.00007242, -0.00005318, 0.00002813, -0.00000307, -0.00000795, 0.00005665, -0.00001209, -0.00010559, 0.00001544, 0.00011413, -0.00001561, -0.00006113, 0.00004610, -0.00000691, -0.00007128, 0.00006173, 0.00004274, -0.00011453, 0.00001009, 0.00012464, -0.00006783, -0.00008690, 0.00007438, -0.00002614, -0.00006775, 0.00011365, 0.00002082, -0.00013461, 0.00002692, 0.00003263, -0.00010709, 0.00006225, 0.00007006, -0.00020585, -0.00003217, 0.00014309, -0.00035994, -0.00041333, 0.00053722, 0.00093584, 0.00019204, -0.00037320, -0.00021542, -0.00000620, -0.00001234, -0.00002918, -0.00001393, -0.00000388, 0.00000137, -0.00000218, },
		{ -0.00002448, -0.00002692, 0.00000097, 0.00004067, 0.00003744, -0.00001626, -0.00002392, 0.00002915, 0.00000502, -0.00006165, -0.00000398, 0.00005708, -0.00003648, -0.00008750, 0.00000601, 0.00002826, -0.00001934, 0.00003037, 0.00003636, -0.00008442, -0.00005929, 0.00009886, 0.00005591, -0.00006682, 0.00000507, 0.00003356, -0.00007100, 0.00000195, 0.00008628, -0.00007073, -0.00006989, 0.00012300, 0.00001526, -0.00013280, 0.00003419, 0.00004702, -0.00009849, 0.00005241, 0.00010753, -0.00012383, -0.00005703, 0.00009268, -0.00008738, -0.00004218, 0.00015207, -0.00011696, -0.00019552, 0.00017880, -0.00009978, -0.00061008, 0.00009540, 0.00104267, 0.00060154, -0.00031160, -0.00035835, -0.00003153, 0.00000202, -0.00004426, -0.00001140, -0.00000100, -0.00000703, -0.00000184, -0.00000245, -0.00001085, },
		{ -0.00001050, 0.00003189, 0.00004439, 0.00000402, -0.00002779, 0.00000631, 0.00002058, -0.00004058, -0.00003513, 0.00004835, 0.00000984, -0.00009167, -0.00004003, 0.00004258, 0.00000096, -0.00000252, 0.00004960, -0.00003147, -0.00009953, 0.00003615, 0.00010391, -0.00002564, -0.00003898, 0.00004174, -0.00003624, -0.00005186, 0.00007926, 0.00000181, -0.00011087, 0.00006403, 0.00009540, -0.00011982, -0.00004642, 0.00009485, -0.00006464, -0.00003757, 0.00014463, -0.00004308, -0.00014464, 0.00008616, 0.00000179, -0.00013604, 0.00012666, 0.00004829, -0.00027770, 0.00004417, 0.00016443, -0.00053881, -0.00039972, 0.00084858, 0.00103301, -0.00005123, -0.00050627, -0.00012097, 0.00003814, -0.00005823, -0.00002660, 0.00001441, -0.00001222, -0.00001045, 0.00000376, -0.00000557, -0.00002380, -0.00003256, },
		{ 0.00004607, 0.00002119, -0.00001968, -0.00001210, 0.00001716, -0.00001517, -0.00004528, 0.00001987, 0.00004127, -0.00006234, -0.00008253, 0.00002233, 0.00003443, -0.00001292, 0.00002952, 0.00001701, -0.00008959, -0.00003806, 0.00010005, 0.00003924, -0.00005268, 0.00001595, 0.00000465, -0.00007138, 0.00003160, 0.00006088, -0.00009393, -0.00001922, 0.00012883, -0.00004970, -0.00012105, 0.00008126, 0.00001477, -0.00010035, 0.00010567, 0.00006823, -0.00017417, 0.00000132, 0.00010130, -0.00014696, 0.00000389, 0.00018465, -0.00020748, -0.00017363, 0.00027308, -0.00023830, -0.00073298, 0.00035676, 0.00129329, 0.00042328, -0.00055092, -0.00030559, 0.00006714, -0.00003923, -0.00006890, 0.00002745, 0.00000113, -0.00002843, 0.00000413, 0.00000497, -0.00002210, -0.00003567, -0.00002143, 0.00001990, },
		{ -0.00000587, -0.00002026, 0.00000358, 0.00000013, -0.00003625, -0.00000838, 0.00004550, -0.00001772, -0.00009630, -0.00002425, 0.00005115, 0.00000737, 0.00000208, 0.00003443, -0.00004597, -0.00008400, 0.00004810, 0.00008804, -0.00002345, -0.00001965, 0.00002327, -0.00005424, -0.00002421, 0.00007634, -0.00003535, -0.00008075, 0.00009990, 0.00004012, -0.00014439, 0.00000939, 0.00008862, -0.00009565, 0.00001324, 0.00014572, -0.00011664, -0.00011612, 0.00013704, -0.00005920, -0.00013534, 0.00019996, -0.00001663, -0.00032709, 0.00016212, 0.00012175, -0.00074901, -0.00027068, 0.00120548, 0.00099929, -0.00036292, -0.00055133, 0.00002123, 0.00003372, -0.00011680, 0.00000824, 0.00003840, -0.00004094, -0.00001249, 0.00002005, -0.00001570, -0.00003947, -0.00002875, 0.00000588, 0.00004162, 0.00003457, },
		{ -0.00001020, 0.00000187, -0.00002061, -0.00002233, 0.00002920, 0.00001774, -0.00007623, -0.00007018, 0.00003350, 0.00004028, -0.00000493, 0.00002146, -0.00000258, -0.00008405, -0.00001960, 0.00009140, 0.00003077, -0.00003373, 0.00001233, -0.00002225, -0.00005702, 0.00004783, 0.00002629, -0.00009137, 0.00003043, 0.00010112, -0.00010274, -0.00007956, 0.00010832, -0.00002749, -0.00007469, 0.00014425, 0.00000038, -0.00019227, 0.00007488, 0.00007176, -0.00019554, 0.00008174, 0.00017859, -0.00030784, -0.00009283, 0.00033763, -0.00045147, -0.00077555, 0.00072512, 0.00145045, 0.00011818, -0.00073348, -0.00016392, 0.00013165, -0.00011906, -0.00006151, 0.00007748, -0.00002294, -0.00004770, 0.00002722, 0.00000306, -0.00004487, -0.00003503, -0.00000491, 0.00003031, 0.00004239, 0.00001059, -0.00001920, },
		{ -0.00000962, -0.00002157, 0.00000880, 0.00003094, -0.00003812, -0.00009044, -0.00001092, 0.00005695, 0.00001598, 0.00000185, 0.00001519, -0.00005159, -0.00006535, 0.00004987, 0.00007426, -0.00001196, -0.00001045, -0.00000184, -0.00005741, 0.00000052, 0.00005693, -0.00005543, -0.00003671, 0.00010540, -0.00002217, -0.00013427, 0.00006081, 0.00005856, -0.00010431, 0.00006937, 0.00010998, -0.00017670, -0.00005257, 0.00015750, -0.00013464, -0.00009265, 0.00025656, -0.00012064, -0.00032499, 0.00029587, -0.00001117, -0.00093915, -0.00000061, 0.00153829, 0.00080323, -0.00067849, -0.00047990, 0.00016569, -0.00003195, -0.00015747, 0.00007310, 0.00003747, -0.00008037, 0.00000711, 0.00003418, -0.00004312, -0.00004747, -0.00000974, 0.00001703, 0.00004048, 0.00002754, -0.00001060, -0.00001960, -0.00000496, },
		{ -0.00000370, 0.00002572, -0.00000412, -0.00007893, -0.00005653, 0.00004074, 0.00004741, 0.00000081, 0.00000757, -0.00001701, -0.00007147, -0.00000838, 0.00007922, 0.00003198, -0.00001826, -0.00000297, -0.00003958, -0.00003611, 0.00004762, -0.00000345, -0.00006688, 0.00006137, 0.00005092, -0.00012732, -0.00002408, 0.00010737, -0.00006215, -0.00002738, 0.00015149, -0.00007772, -0.00016786, 0.00013916, 0.00000671, -0.00020998, 0.00017298, 0.00012044, -0.00038589, 0.00004183, 0.00033014, -0.00070929, -0.00068432, 0.00115199, 0.00145009, -0.00026773, -0.00080195, 0.00003263, 0.00012796, -0.00020965, -0.00000950, 0.00011478, -0.00007272, -0.00004701, 0.00005951, -0.00001959, -0.00006746, -0.00001549, 0.00001092, 0.00002821, 0.00003904, 0.00000565, -0.00002338, -0.00001491, -0.00000642, -0.00001449, },
		{ 0.00001337, -0.00004899, -0.00007975, -0.00000095, 0.00006234, 0.00002413, -0.00000404, -0.00000287, -0.00004901, -0.00004935, 0.00004491, 0.00006721, 0.00000211, -0.00001331, -0.00002508, -0.00004891, 0.00001475, 0.00003167, -0.00005400, 0.00000361, 0.00008130, -0.00007059, -0.00009619, 0.00009134, 0.00001838, -0.00008524, 0.00011060, 0.00004476, -0.00020240, 0.00002772, 0.00013531, -0.00019895, -0.00001172, 0.00027024, -0.00024063, -0.00025406, 0.00040183, -0.00023451, -0.00104316, 0.00039743, 0.00176397, 0.00045636, -0.00092134, -0.00030028, 0.00026110, -0.00014471, -0.00015246, 0.00014795, -0.00000002, -0.00011058, 0.00004869, 0.00002979, -0.00008105, -0.00003524, 0.00001535, 0.00001434, 0.00003675, 0.00002693, -0.00001856, -0.00002593, -0.00000899, -0.00000819, -0.00000709, 0.00001444, },
		{ -0.00007466, -0.00004418, 0.00004612, 0.00005455, 0.00000224, -0.00000816, -0.00002448, -0.00005716, -0.00000410, 0.00006939, 0.00003927, -0.00001153, -0.00002326, -0.00004545, -0.00001736, 0.00003593, -0.00001861, -0.00003346, 0.00006517, -0.00000253, -0.00011931, 0.00002568, 0.00008320, -0.00007501, 0.00002505, 0.00012186, -0.00014169, -0.00010654, 0.00017097, -0.00007541, -0.00017758, 0.00024844, 0.00001554, -0.00040879, 0.00020062, 0.00022156, -0.00095285, -0.00043271, 0.00155634, 0.00126130, -0.00065665, -0.00072937, 0.00022741, 0.00004696, -0.00027347, 0.00007936, 0.00011613, -0.00013232, -0.00001558, 0.00008335, -0.00006402, -0.00007351, 0.00001903, 0.00001339, 0.00001906, 0.00004315, -0.00000084, -0.00003556, -0.00001669, -0.00000406, -0.00000612, 0.00000577, 0.00001641, -0.00001967, },
		{ 0.00000686, 0.00006714, 0.00002943, -0.00001320, -0.00001657, -0.00004194, -0.00003815, 0.00003848, 0.00006666, 0.00001232, -0.00002500, -0.00004112, -0.00003550, 0.00001780, 0.00001164, -0.00003706, 0.00002619, 0.00004041, -0.00009186, -0.00004689, 0.00009518, -0.00001444, -0.00004580, 0.00012127, -0.00003016, -0.00018482, 0.00009866, 0.00007487, -0.00022754, 0.00008639, 0.00022681, -0.00034079, -0.00012043, 0.00043329, -0.00051400, -0.00100284, 0.00086652, 0.00181334, 0.00001582, -0.00102934, -0.00005712, 0.00027125, -0.00026838, -0.00009093, 0.00020327, -0.00006892, -0.00011133, 0.00009869, -0.00000355, -0.00011276, 0.00000076, 0.00003153, -0.00000146, 0.00004050, 0.00002800, -0.00003619, -0.00003127, -0.00000205, -0.00000312, 0.00000024, 0.00001343, -0.00000127, -0.00005272, -0.00006771, },
		{ 0.00005958, 0.00000035, -0.00002218, -0.00002682, -0.00004456, -0.00000322, 0.00006422, 0.00004695, -0.00001393, -0.00004216, -0.00004300, -0.00000523, 0.00002183, -0.00001839, -0.00000612, 0.00004596, -0.00004112, -0.00008773, 0.00005545, 0.00005017, -0.00006251, 0.00006292, 0.00006450, -0.00017155, -0.00002833, 0.00015903, -0.00014727, -0.00010354, 0.00028020, -0.00011058, -0.00035820, 0.00033604, 0.00001262, -0.00111036, -0.00003781, 0.00184718, 0.00090207, -0.00096301, -0.00053100, 0.00036229, -0.00009035, -0.00028018, 0.00017864, 0.00007136, -0.00017690, 0.00004273, 0.00008014, -0.00011803, -0.00005111, 0.00005436, -0.00000317, 0.00001436, 0.00005396, -0.00001786, -0.00005040, -0.00000655, 0.00000365, -0.00000404, 0.00000885, 0.00000778, -0.00002960, -0.00006699, -0.00003291, 0.00005002, },
		{ -0.00002237, -0.00002614, -0.00003400, -0.00003017, 0.00003414, 0.00006883, 0.00001565, -0.00004015, -0.00004945, -0.00002286, 0.00001489, 0.00000229, -0.00001681, 0.00002716, 0.00000051, -0.00008545, -0.00000440, 0.00007883, -0.00002767, -0.00000272, 0.00009799, -0.00009081, -0.00013146, 0.00013753, -0.00000372, -0.00020831, 0.00017081, 0.00013270, -0.00038647, 0.00004384, 0.00036002, -0.00078619, -0.00079187, 0.00131791, 0.00165692, -0.00043059, -0.00097736, 0.00018392, 0.00019010, -0.00035966, 0.00001456, 0.00021381, -0.00014785, -0.00007320, 0.00013499, -0.00006295, -0.00011959, 0.00005137, 0.00002633, -0.00002016, 0.00005624, 0.00002048, -0.00006199, -0.00002461, 0.00001351, -0.00000428, 0.00000107, 0.00001253, -0.00001233, -0.00005261, -0.00005474, 0.00001295, 0.00006960, 0.00003196, },
		{ -0.00002711, -0.00003368, -0.00000209, 0.00006207, 0.00005083, -0.00002227, -0.00005661, -0.00003693, 0.00000163, 0.00001248, -0.00001016, 0.00000597, 0.00001712, -0.00005603, -0.00004896, 0.00006225, 0.00002417, -0.00003327, 0.00007291, -0.00000049, -0.00016059, 0.00004076, 0.00010942, -0.00018455, -0.00001034, 0.00025441, -0.00022053, -0.00024113, 0.00039972, -0.00025981, -0.00112191, 0.00043784, 0.00195312, 0.00043784, -0.00112191, -0.00025981, 0.00039972, -0.00024113, -0.00022053, 0.00025441, -0.00001034, -0.00018455, 0.00010942, 0.00004076, -0.00016059, -0.00000049, 0.00007291, -0.00003327, 0.00002417, 0.00006225, -0.00004896, -0.00005603, 0.00001712, 0.00000597, -0.00001016, 0.00001248, 0.00000163, -0.00003693, -0.00005661, -0.00002227, 0.00005083, 0.00006207, -0.00000209, -0.00003368, },
		{ -0.00002237, 0.00003196, 0.00006960, 0.00001295, -0.00005474, -0.00005261, -0.00001233, 0.00001253, 0.00000107, -0.00000428, 0.00001351, -0.00002461, -0.00006199, 0.00002048, 0.00005624, -0.00002016, 0.00002633, 0.00005137, -0.00011959, -0.00006295, 0.00013499, -0.00007320, -0.00014785, 0.00021381, 0.00001456, -0.00035966, 0.00019010, 0.00018392, -0.00097736, -0.00043059, 0.00165692, 0.00131791, -0.00079187, -0.00078619, 0.00036002, 0.00004384, -0.00038647, 0.00013270, 0.00017081, -0.00020831, -0.00000372, 0.00013753, -0.00013146, -0.00009081, 0.00009799, -0.00000272, -0.00002767, 0.00007883, -0.00000440, -0.00008545, 0.00000051, 0.00002716, -0.00001681, 0.00000229, 0.00001489, -0.00002286, -0.00004945, -0.00004015, 0.00001565, 0.00006883, 0.00003414, -0.00003017, -0.00003400, -0.00002614, },
		{ 0.00005958, 0.00005002, -0.00003291, -0.00006699, -0.00002960, 0.00000778, 0.00000885, -0.00000404, 0.00000365, -0.00000655, -0.00005040, -0.00001786, 0.00005396, 0.00001436, -0.00000317, 0.00005436, -0.00005111, -0.00011803, 0.00008014, 0.00004273, -0.00017690, 0.00007136, 0.00017864, -0.00028018, -0.00009035, 0.00036229, -0.00053100, -0.00096301, 0.00090207, 0.00184718, -0.00003781, -0.00111036, 0.00001262, 0.00033604, -0.00035820, -0.00011058, 0.00028020, -0.00010354, -0.00014727, 0.00015903, -0.00002833, -0.00017155, 0.00006450, 0.00006292, -0.00006251, 0.00005017, 0.00005545, -0.00008773, -0.00004112, 0.00004596, -0.00000612, -0.00001839, 0.00002183, -0.00000523, -0.00004300, -0.00004216, -0.00001393, 0.00004695, 0.00006422, -0.00000322, -0.00004456, -0.00002682, -0.00002218, 0.00000035, },
		{ 0.00000686, -0.00006771, -0.00005272, -0.00000127, 0.00001343, 0.00000024, -0.00000312, -0.00000205, -0.00003127, -0.00003619, 0.00002800, 0.00004050, -0.00000146, 0.00003153, 0.00000076, -0.00011276, -0.00000355, 0.00009869, -0.00011133, -0.00006892, 0.00020327, -0.00009093, -0.00026838, 0.00027125, -0.00005712, -0.00102934, 0.00001582, 0.00181334, 0.00086652, -0.00100284, -0.00051400, 0.00043329, -0.00012043, -0.00034079, 0.00022681, 0.00008639, -0.00022754, 0.00007487, 0.00009866, -0.00018482, -0.00003016, 0.00012127, -0.00004580, -0.00001444, 0.00009518, -0.00004689, -0.00009186, 0.00004041, 0.00002619, -0.00003706, 0.00001164, 0.00001780, -0.00003550, -0.00004112, -0.00002500, 0.00001232, 0.00006666, 0.00003848, -0.00003815, -0.00004194, -0.00001657, -0.00001320, 0.00002943, 0.00006714, },
		{ -0.00007466, -0.00001967, 0.00001641, 0.00000577, -0.00000612, -0.00000406, -0.00001669, -0.00003556, -0.00000084, 0.00004315, 0.00001906, 0.00001339, 0.00001903, -0.00007351, -0.00006402, 0.00008335, -0.00001558, -0.00013232, 0.00011613, 0.00007936, -0.00027347, 0.00004696, 0.00022741, -0.00072937, -0.00065665, 0.00126130, 0.00155634, -0.00043271, -0.00095285, 0.00022156, 0.00020062, -0.00040879, 0.00001554, 0.00024844, -0.00017758, -0.00007541, 0.00017097, -0.00010654, -0.00014169, 0.00012186, 0.00002505, -0.00007501, 0.00008320, 0.00002568, -0.00011931, -0.00000253, 0.00006517, -0.00003346, -0.00001861, 0.00003593, -0.00001736, -0.00004545, -0.00002326, -0.00001153, 0.00003927, 0.00006939, -0.00000410, -0.00005716, -0.00002448, -0.00000816, 0.00000224, 0.00005455, 0.00004612, -0.00004418, },
		{ 0.00001337, 0.00001444, -0.00000709, -0.00000819, -0.00000899, -0.00002593, -0.00001856, 0.00002693, 0.00003675, 0.00001434, 0.00001535, -0.00003524, -0.00008105, 0.00002979, 0.00004869, -0.00011058, -0.00000002, 0.00014795, -0.00015246, -0.00014471, 0.00026110, -0.00030028, -0.00092134, 0.00045636, 0.00176397, 0.00039743, -0.00104316, -0.00023451, 0.00040183, -0.00025406, -0.00024063, 0.00027024, -0.00001172, -0.00019895, 0.00013531, 0.00002772, -0.00020240, 0.00004476, 0.00011060, -0.00008524, 0.00001838, 0.00009134, -0.00009619, -0.00007059, 0.00008130, 0.00000361, -0.00005400, 0.00003167, 0.00001475, -0.00004891, -0.00002508, -0.00001331, 0.00000211, 0.00006721, 0.00004491, -0.00004935, -0.00004901, -0.00000287, -0.00000404, 0.00002413, 0.00006234, -0.00000095, -0.00007975, -0.00004899, },
		{ -0.00000370, -0.00001449, -0.00000642, -0.00001491, -0.00002338, 0.00000565, 0.00003904, 0.00002821, 0.00001092, -0.00001549, -0.00006746, -0.00001959, 0.00005951, -0.00004701, -0.00007272, 0.00011478, -0.00000950, -0.00020965, 0.00012796, 0.00003263, -0.00080195, -0.00026773, 0.00145009, 0.00115199, -0.00068432, -0.00070929, 0.00033014, 0.00004183, -0.00038589, 0.00012044, 0.00017298, -0.00020998, 0.00000671, 0.00013916, -0.00016786, -0.00007772, 0.00015149, -0.00002738, -0.00006215, 0.00010737, -0.00002408, -0.00012732, 0.00005092, 0.00006137, -0.00006688, -0.00000345, 0.00004762, -0.00003611, -0.00003958, -0.00000297, -0.00001826, 0.00003198, 0.00007922, -0.00000838, -0.00007147, -0.00001701, 0.00000757, 0.00000081, 0.00004741, 0.00004074, -0.00005653, -0.00007893, -0.00000412, 0.00002572, },
		{ -0.00000962, -0.00000496, -0.00001960, -0.00001060, 0.00002754, 0.00004048, 0.00001703, -0.00000974, -0.00004747, -0.00004312, 0.00003418, 0.00000711, -0.00008037, 0.00003747, 0.00007310, -0.00015747, -0.00003195, 0.00016569, -0.00047990, -0.00067849, 0.00080323, 0.00153829, -0.00000061, -0.00093915, -0.00001117, 0.00029587, -0.00032499, -0.00012064, 0.00025656, -0.00009265, -0.00013464, 0.00015750, -0.00005257, -0.00017670, 0.00010998, 0.00006937, -0.00010431, 0.00005856, 0.00006081, -0.00013427, -0.00002217, 0.00010540, -0.00003671, -0.00005543, 0.00005693, 0.00000052, -0.00005741, -0.00000184, -0.00001045, -0.00001196, 0.00007426, 0.00004987, -0.00006535, -0.00005159, 0.00001519, 0.00000185, 0.00001598, 0.00005695, -0.00001092, -0.00009044, -0.00003812, 0.00003094, 0.00000880, -0.00002157, },
		{ -0.00001020, -0.00001920, 0.00001059, 0.00004239, 0.00003031, -0.00000491, -0.00003503, -0.00004487, 0.00000306, 0.00002722, -0.00004770, -0.00002294, 0.00007748, -0.00006151, -0.00011906, 0.00013165, -0.00016392, -0.00073348, 0.00011818, 0.00145045, 0.00072512, -0.00077555, -0.00045147, 0.00033763, -0.00009283, -0.00030784, 0.00017859, 0.00008174, -0.00019554, 0.00007176, 0.00007488, -0.00019227, 0.00000038, 0.00014425, -0.00007469, -0.00002749, 0.00010832, -0.00007956, -0.00010274, 0.00010112, 0.00003043, -0.00009137, 0.00002629, 0.00004783, -0.00005702, -0.00002225, 0.00001233, -0.00003373, 0.00003077, 0.00009140, -0.00001960, -0.00008405, -0.00000258, 0.00002146, -0.00000493, 0.00004028, 0.00003350, -0.00007018, -0.00007623, 0.00001774, 0.00002920, -0.00002233, -0.00002061, 0.00000187, },
		{ -0.00000587, 0.00003457, 0.00004162, 0.00000588, -0.00002875, -0.00003947, -0.00001570, 0.00002005, -0.00001249, -0.00004094, 0.00003840, 0.00000824, -0.00011680, 0.00003372, 0.00002123, -0.00055133, -0.00036292, 0.00099929, 0.00120548, -0.00027068, -0.00074901, 0.00012175, 0.00016212, -0.00032709, -0.00001663, 0.00019996, -0.00013534, -0.00005920, 0.00013704, -0.00011612, -0.00011664, 0.00014572, 0.00001324, -0.00009565, 0.00008862, 0.00000939, -0.00014439, 0.00004012, 0.00009990, -0.00008075, -0.00003535, 0.00007634, -0.00002421, -0.00005424, 0.00002327, -0.00001965, -0.00002345, 0.00008804, 0.00004810, -0.00008400, -0.00004597, 0.00003443, 0.00000208, 0.00000737, 0.00005115, -0.00002425, -0.00009630, -0.00001772, 0.00004550, -0.00000838, -0.00003625, 0.00000013, 0.00000358, -0.00002026, },
		{ 0.00004607, 0.00001990, -0.00002143, -0.00003567, -0.00002210, 0.00000497, 0.00000413, -0.00002843, 0.00000113, 0.00002745, -0.00006890, -0.00003923, 0.00006714, -0.00030559, -0.00055092, 0.00042328, 0.00129329, 0.00035676, -0.00073298, -0.00023830, 0.00027308, -0.00017363, -0.00020748, 0.00018465, 0.00000389, -0.00014696, 0.00010130, 0.00000132, -0.00017417, 0.00006823, 0.00010567, -0.00010035, 0.00001477, 0.00008126, -0.00012105, -0.00004970, 0.00012883, -0.00001922, -0.00009393, 0.00006088, 0.00003160, -0.00007138, 0.00000465, 0.00001595, -0.00005268, 0.00003924, 0.00010005, -0.00003806, -0.00008959, 0.00001701, 0.00002952, -0.00001292, 0.00003443, 0.00002233, -0.00008253, -0.00006234, 0.00004127, 0.00001987, -0.00004528, -0.00001517, 0.00001716, -0.00001210, -0.00001968, 0.00002119, },
		{ -0.00001050, -0.00003256, -0.00002380, -0.00000557, 0.00000376, -0.00001045, -0.00001222, 0.00001441, -0.00002660, -0.00005823, 0.00003814, -0.00012097, -0.00050627, -0.00005123, 0.00103301, 0.00084858, -0.00039972, -0.00053881, 0.00016443, 0.00004417, -0.00027770, 0.00004829, 0.00012666, -0.00013604, 0.00000179, 0.00008616, -0.00014464, -0.00004308, 0.00014463, -0.00003757, -0.00006464, 0.00009485, -0.00004642, -0.00011982, 0.00009540, 0.00006403, -0.00011087, 0.00000181, 0.00007926, -0.00005186, -0.00003624, 0.00004174, -0.00003898, -0.00002564, 0.00010391, 0.00003615, -0.00009953, -0.00003147, 0.00004960, -0.00000252, 0.00000096, 0.00004258, -0.00004003, -0.00009167, 0.00000984, 0.00004835, -0.00003513, -0.00004058, 0.00002058, 0.00000631, -0.00002779, 0.00000402, 0.00004439, 0.00003189, },
		{ -0.00002448, -0.00001085, -0.00000245, -0.00000184, -0.00000703, -0.00000100, -0.00001140, -0.00004426, 0.00000202, -0.00003153, -0.00035835, -0.00031160, 0.00060154, 0.00104267, 0.00009540, -0.00061008, -0.00009978, 0.00017880, -0.00019552, -0.00011696, 0.00015207, -0.00004218, -0.00008738, 0.00009268, -0.00005703, -0.00012383, 0.00010753, 0.00005241, -0.00009849, 0.00004702, 0.00003419, -0.00013280, 0.00001526, 0.00012300, -0.00006989, -0.00007073, 0.00008628, 0.00000195, -0.00007100, 0.00003356, 0.00000507, -0.00006682, 0.00005591, 0.00009886, -0.00005929, -0.00008442, 0.00003636, 0.00003037, -0.00001934, 0.00002826, 0.00000601, -0.00008750, -0.00003648, 0.00005708, -0.00000398, -0.00006165, 0.00000502, 0.00002915, -0.00002392, -0.00001626, 0.00003744, 0.00004067, 0.00000097, -0.00002692, },
		{ -0.00000783, -0.00000218, 0.00000137, -0.00000388, -0.00001393, -0.00002918, -0.00001234, -0.00000620, -0.00021542, -0.00037320, 0.00019204, 0.00093584, 0.00053722, -0.00041333, -0.00035994, 0.00014309, -0.00003217, -0.00020585, 0.00007006, 0.00006225, -0.00010709, 0.00003263, 0.00002692, -0.00013461, 0.00002082, 0.00011365, -0.00006775, -0.00002614, 0.00007438, -0.00008690, -0.00006783, 0.00012464, 0.00001009, -0.00011453, 0.00004274, 0.00006173, -0.00007128, -0.00000691, 0.00004610, -0.00006113, -0.00001561, 0.00011413, 0.00001544, -0.00010559, -0.00001209, 0.00005665, -0.00000795, -0.00000307, 0.00002813, -0.00005318, -0.00007242, 0.00003605, 0.00003372, -0.00006162, -0.00002735, 0.00004266, -0.00000364, -0.00003456, 0.00002263, 0.00004788, 0.00001079, -0.00001997, -0.00002252, -0.00001391, },
		{ 0.00000475, 0.00000212, -0.00001635, -0.00002628, -0.00001231, -0.00000176, -0.00012121, -0.00032176, -0.00008057, 0.00064880, 0.00077042, -0.00005062, -0.00046406, -0.00003481, 0.00009211, -0.00017376, -0.00005565, 0.00010370, -0.00005426, -0.00003899, 0.00005818, -0.00008473, -0.00006149, 0.00011292, 0.00000376, -0.00007563, 0.00005717, -0.00001566, -0.00011082, 0.00007028, 0.00008625, -0.00010291, -0.00002963, 0.00009134, -0.00003073, -0.00005321, 0.00005435, -0.00001675, -0.00006899, 0.00007399, 0.00008511, -0.00007565, -0.00006949, 0.00004916, 0.00002678, -0.00002166, 0.00001772, -0.00001273, -0.00007820, -0.00000455, 0.00005573, -0.00003539, -0.00006053, 0.00003365, 0.00002764, -0.00004021, -0.00000204, 0.00005178, 0.00002162, -0.00001561, -0.00001758, -0.00001410, -0.00001243, -0.00000571, },
		{ -0.00001161, -0.00002897, -0.00001384, 0.00000231, -0.00006798, -0.00024002, -0.00020321, 0.00033381, 0.00076461, 0.00030939, -0.00035797, -0.00023659, 0.00009346, -0.00006321, -0.00013829, 0.00006054, 0.00002146, -0.00006869, 0.00003254, -0.00001937, -0.00009682, 0.00005932, 0.00006911, -0.00007311, 0.00000351, 0.00003657, -0.00009854, -0.00000681, 0.00011945, -0.00004214, -0.00009062, 0.00007165, 0.00002923, -0.00007360, 0.00002430, 0.00003431, -0.00007651, 0.00000334, 0.00011201, -0.00000739, -0.00009915, 0.00000494, 0.00005571, -0.00000958, -0.00000823, 0.00000739, -0.00005546, -0.00004033, 0.00004915, 0.00000400, -0.00007444, 0.00000149, 0.00005367, -0.00002449, -0.00003095, 0.00004531, 0.00003705, -0.00001354, -0.00001455, -0.00000863, -0.00001618, -0.00001068, 0.00000433, 0.00000808, },
		{ -0.00002063, 0.00000396, -0.00003476, -0.00016875, -0.00022533, 0.00009286, 0.00059630, 0.00053216, -0.00010506, -0.00033644, -0.00002116, 0.00003356, -0.00013220, -0.00002614, 0.00006017, -0.00004357, -0.00001383, 0.00001735, -0.00008037, -0.00000766, 0.00009206, -0.00002714, -0.00004599, 0.00004435, -0.00005016, -0.00006484, 0.00009813, 0.00003367, -0.00010745, 0.00001356, 0.00007375, -0.00005370, -0.00002317, 0.00005939, -0.00004054, -0.00005737, 0.00008451, 0.00006210, -0.00008042, -0.00005090, 0.00005200, 0.00002436, -0.00002131, -0.00000064, -0.00002628, -0.00005165, 0.00002156, 0.00003400, -0.00006049, -0.00003830, 0.00005729, 0.00000952, -0.00005005, 0.00002248, 0.00005402, -0.00000686, -0.00001898, 0.00000048, -0.00001449, -0.00001856, 0.00000233, 0.00001146, -0.00000419, -0.00002748, },
		{ -0.00001296, -0.00011301, -0.00020221, -0.00004822, 0.00037757, 0.00057771, 0.00016823, -0.00027978, -0.00016666, 0.00004495, -0.00006160, -0.00008969, 0.00003810, 0.00000483, -0.00003842, 0.00001309, -0.00004087, -0.00004959, 0.00006914, 0.00002805, -0.00006060, 0.00001469, -0.00000205, -0.00008028, 0.00004229, 0.00008561, -0.00007464, -0.00004983, 0.00007808, -0.00000577, -0.00005716, 0.00004631, 0.00001149, -0.00007820, 0.00002270, 0.00009604, -0.00002346, -0.00008307, 0.00001296, 0.00005151, -0.00000572, -0.00001987, -0.00001468, -0.00003965, -0.00000494, 0.00004041, -0.00002908, -0.00006416, 0.00003417, 0.00004510, -0.00004600, -0.00001253, 0.00006138, 0.00001131, -0.00002890, 0.00000440, -0.00000282, -0.00002683, -0.00000391, 0.00001503, 0.00000114, -0.00002151, -0.00002531, -0.00000180, },
		{ -0.00016410, -0.00011285, 0.00018926, 0.00049222, 0.00035420, -0.00010593, -0.00024463, -0.00003096, 0.00000513, -0.00009102, -0.00001861, 0.00003182, -0.00002635, -0.00001083, -0.00001278, -0.00005456, 0.00002529, 0.00005977, -0.00003787, -0.00002345, 0.00001842, -0.00005902, -0.00001344, 0.00009373, -0.00001460, -0.00008527, 0.00004277, 0.00004123, -0.00005832, 0.00000832, 0.00004772, -0.00005621, -0.00003699, 0.00008111, 0.00003827, -0.00007180, -0.00003709, 0.00004684, 0.00002788, -0.00002365, -0.00002582, -0.00002486, -0.00001280, 0.00002897, -0.00000110, -0.00006508, -0.00000247, 0.00006259, -0.00001816, -0.00004410, 0.00004843, 0.00003841, -0.00003330, -0.00000527, 0.00001523, -0.00002686, -0.00001777, 0.00001833, 0.00000678, -0.00001732, -0.00002319, -0.00001014, -0.00000475, -0.00006892, },
		{ 0.00005830, 0.00035358, 0.00041553, 0.00009241, -0.00020913, -0.00013246, 0.00001195, -0.00004524, -0.00006099, 0.00001758, 0.00000277, -0.00002411, -0.00000968, -0.00003831, -0.00000964, 0.00005804, 0.00000072, -0.00004273, 0.00000841, -0.00002634, -0.00004339, 0.00006542, 0.00003962, -0.00007930, -0.00000913, 0.00006180, -0.00003000, -0.00002853, 0.00005199, -0.00001241, -0.00006577, 0.00003304, 0.00007140, -0.00002639, -0.00006521, 0.00000985, 0.00005072, 0.00000034, -0.00004112, -0.00002816, -0.00000485, 0.00001928, 0.00001026, -0.00004886, -0.00003189, 0.00005436, 0.00001855, -0.00005529, 0.00001596, 0.00006073, -0.00002087, -0.00002654, 0.00002684, -0.00001121, -0.00003510, 0.00001429, 0.00001758, -0.00001661, -0.00001963, -0.00001159, -0.00000841, -0.00003961, -0.00012055, -0.00013178, },
		{ 0.00038050, 0.00023450, -0.00008722, -0.00018536, -0.00004361, -0.00000302, -0.00005974, -0.00001977, 0.00001754, -0.00001485, -0.00001969, -0.00002453, -0.00002272, 0.00003622, 0.00002984, -0.00003517, -0.00001460, -0.00000610, -0.00004407, 0.00002498, 0.00006545, -0.00004282, -0.00004906, 0.00004928, 0.00000740, -0.00004423, 0.00003012, 0.00002568, -0.00005770, -0.00001741, 0.00006386, 0.00002297, -0.00005454, -0.00003434, 0.00004012, 0.00003765, -0.00003482, -0.00005039, -0.00000185, 0.00002543, 0.00001100, -0.00003365, -0.00004286, 0.00003056, 0.00004429, -0.00004137, -0.00002122, 0.00006316, 0.00000841, -0.00004536, 0.00001951, 0.00001638, -0.00004311, -0.00000351, 0.00003119, -0.00001212, -0.00002314, -0.00000497, -0.00001220, -0.00003005, -0.00007923, -0.00012179, -0.00002259, 0.00021925, },
		{ 0.00005300, -0.00015620, -0.00011445, -0.00000547, -0.00002734, -0.00004460, 0.00000468, 0.00000357, -0.00002296, -0.00002368, -0.00002151, 0.00001448, 0.00003841, -0.00001218, -0.00002999, -0.00000580, -0.00002988, -0.00000514, 0.00006113, 0.00000023, -0.00006067, 0.00001662, 0.00003223, -0.00003586, 0.00000087, 0.00004164, -0.00002759, -0.00004566, 0.00002906, 0.00004767, -0.00001603, -0.00005452, -0.00000055, 0.00005759, 0.00000162, -0.00006906, -0.00002244, 0.00004061, 0.00001965, -0.00002958, -0.00004219, 0.00000930, 0.00004991, -0.00001395, -0.00004467, 0.00004216, 0.00004058, -0.00004588, -0.00000682, 0.00003997, -0.00003022, -0.00003072, 0.00003601, 0.00000422, -0.00003311, -0.00000104, -0.00000412, -0.00003439, -0.00005338, -0.00009172, -0.00006456, 0.00011007, 0.00030059, 0.00029257, },
		{ -0.00014561, -0.00005108, -0.00000273, -0.00003762, -0.00002074, 0.00000969, -0.00001262, -0.00002842, -0.00002050, 0.00000290, 0.00003235, 0.00000921, -0.00002897, -0.00001677, -0.00001809, -0.00001710, 0.00004060, 0.00002976, -0.00004629, -0.00001544, 0.00003511, -0.00001510, -0.00001909, 0.00003621, 0.00000348, -0.00004611, -0.00000765, 0.00003953, 0.00002151, -0.00003776, -0.00004308, 0.00003896, 0.00004841, -0.00005703, -0.00006259, 0.00004098, 0.00004705, -0.00002757, -0.00004586, -0.00000048, 0.00004332, 0.00000922, -0.00004674, 0.00000963, 0.00005780, -0.00002402, -0.00003718, 0.00004255, 0.00000182, -0.00005120, 0.00002114, 0.00002959, -0.00003720, -0.00001153, 0.00001432, -0.00003487, -0.00005170, -0.00005660, -0.00007082, 0.00002725, 0.00021123, 0.00028635, 0.00015457, -0.00006786, },
		{ -0.00001444, -0.00001235, -0.00003232, -0.00000311, 0.00000036, -0.00002636, -0.00002538, -0.00000234, 0.00002354, 0.00002020, -0.00001661, -0.00002573, -0.00001646, -0.00001701, 0.00002000, 0.00003952, -0.00002039, -0.00003232, 0.00002130, 0.00000311, -0.00002451, 0.00002069, 0.00002237, -0.00002944, -0.00002767, 0.00001255, 0.00003386, 0.00000115, -0.00005617, -0.00001049, 0.00007068, -0.00000823, -0.00009396, 0.00000768, 0.00008107, -0.00000735, -0.00006016, -0.00000731, 0.00003967, 0.00002007, -0.00003578, -0.00001698, 0.00005272, 0.00000833, -0.00005244, 0.00002032, 0.00003591, -0.00004858, -0.00001102, 0.00004851, -0.00002259, -0.00003489, 0.00002680, -0.00001617, -0.00006561, -0.00003824, -0.00004662, -0.00002550, 0.00012355, 0.00024818, 0.00020219, 0.00002818, -0.00011664, -0.00010083, },
		{ -0.00001936, -0.00002012, 0.00000115, -0.00001550, -0.00003002, -0.00000919, 0.00001760, 0.00002307, -0.00000285, -0.00002593, -0.00002158, -0.00001515, 0.00000726, 0.00003545, 0.00000245, -0.00003192, 0.00000325, 0.00001077, -0.00002044, 0.00000591, 0.00002813, -0.00001022, -0.00002875, -0.00001139, 0.00001810, 0.00003219, -0.00003014, -0.00005765, 0.00004652, 0.00005202, -0.00008524, -0.00005181, 0.00009213, 0.00004085, -0.00007080, -0.00002731, 0.00004401, 0.00002659, -0.00002623, -0.00002952, 0.00003357, 0.00003313, -0.00004441, -0.00001331, 0.00005152, -0.00002074, -0.00004293, 0.00004482, 0.00000975, -0.00005423, 0.00001670, 0.00001873, -0.00007147, -0.00004986, -0.00001159, -0.00003869, 0.00004188, 0.00019427, 0.00021343, 0.00009455, -0.00005119, -0.00011395, -0.00005517, -0.00000059, },
		{ -0.00001133, -0.00000647, -0.00002660, -0.00001885, 0.00001158, 0.00002375, 0.00000704, -0.00001937, -0.00002614, -0.00001697, 0.00000102, 0.00002691, 0.00001561, -0.00002127, -0.00000911, 0.00000913, -0.00001461, -0.00000420, 0.00002649, 0.00000381, -0.00002049, -0.00001958, -0.00000860, 0.00003361, 0.00001546, -0.00006870, -0.00001305, 0.00008299, -0.00003003, -0.00010290, 0.00005755, 0.00010028, -0.00005246, -0.00006425, 0.00004344, 0.00004247, -0.00002291, -0.00003470, 0.00001495, 0.00004113, -0.00002153, -0.00003823, 0.00004061, 0.00001664, -0.00005424, 0.00001671, 0.00004241, -0.00005082, -0.00001592, 0.00004824, -0.00004912, -0.00008152, 0.00000468, -0.00001213, -0.00001910, 0.00012429, 0.00020758, 0.00012913, 0.00000968, -0.00008193, -0.00008949, -0.00002111, 0.00000187, -0.00002084, },
		{ -0.00001692, -0.00002469, 0.00000135, 0.00002363, 0.00001430, -0.00001103, -0.00002634, -0.00002126, -0.00000349, 0.00001939, 0.00002033, -0.00000877, -0.00001299, 0.00000405, -0.00001103, -0.00001127, 0.00002198, 0.00001320, -0.00001262, -0.00001483, -0.00002515, 0.00000741, 0.00004720, -0.00003520, -0.00006961, 0.00005950, 0.00004254, -0.00010788, -0.00001488, 0.00013425, 0.00000719, -0.00009826, 0.00001834, 0.00006919, -0.00001494, -0.00004433, 0.00000307, 0.00003786, 0.00000016, -0.00004366, 0.00001335, 0.00004172, -0.00003816, -0.00001955, 0.00005451, -0.00002127, -0.00005142, 0.00004938, -0.00000063, -0.00010396, -0.00001687, 0.00003223, -0.00003799, 0.00004175, 0.00018457, 0.00015163, 0.00004390, -0.00003211, -0.00008653, -0.00006045, 0.00000342, -0.00000144, -0.00002140, -0.00000996, },
		{ -0.00001080, 0.00001907, 0.00002116, -0.00000279, -0.00002324, -0.00002463, -0.00000870, 0.00001328, 0.00002060, 0.00000101, -0.00001095, 0.00000017, -0.00000914, -0.00001734, 0.00001495, 0.00002091, -0.00000721, -0.00000833, -0.00002306, -0.00002482, 0.00004337, 0.00001925, -0.00008499, -0.00000660, 0.00008580, -0.00005678, -0.00008614, 0.00011317, 0.00008969, -0.00009666, -0.00003567, 0.00008819, 0.00001290, -0.00005935, -0.00000987, 0.00003488, 0.00001205, -0.00003493, -0.00001165, 0.00004340, -0.00000690, -0.00004184, 0.00003897, 0.00001801, -0.00006548, 0.00001791, 0.00004875, -0.00009052, -0.00006722, 0.00005770, -0.00000731, -0.00003109, 0.00013181, 0.00017346, 0.00006211, -0.00000000, -0.00004792, -0.00008374, -0.00002635, 0.00001841, -0.00001225, -0.00002275, -0.00001204, -0.00002139, },
		{ 0.00002571, 0.00000677, -0.00001876, -0.00002587, -0.00001381, 0.00000724, 0.00001882, 0.00000752, -0.00000645, -0.00000132, -0.00000699, -0.00002184, 0.00000420, 0.00002652, -0.00000030, -0.00000744, -0.00001018, -0.00004054, 0.00000905, 0.00005707, -0.00004960, -0.00007288, 0.00006961, 0.00002074, -0.00011182, 0.00004092, 0.00015226, -0.00004072, -0.00009511, 0.00007349, 0.00006087, -0.00006405, -0.00003539, 0.00003503, 0.00001925, -0.00002516, -0.00002341, 0.00002840, 0.00001899, -0.00003962, 0.00000829, 0.00004363, -0.00004879, -0.00002694, 0.00006961, -0.00003932, -0.00011294, 0.00003778, 0.00005142, -0.00006187, 0.00004850, 0.00018147, 0.00008931, 0.00000342, -0.00000626, -0.00006622, -0.00006821, 0.00001230, 0.00001482, -0.00002785, -0.00002016, -0.00001329, -0.00001728, 0.00000788, },
		{ -0.00001226, -0.00002600, -0.00001755, 0.00000121, 0.00001537, 0.00001138, -0.00000135, -0.00000115, -0.00000414, -0.00002283, -0.00000901, 0.00002627, 0.00001082, -0.00000983, -0.00000071, -0.00003319, -0.00003015, 0.00005460, 0.00001097, -0.00009659, 0.00000394, 0.00007412, -0.00007589, -0.00004281, 0.00015699, 0.00005368, -0.00011989, 0.00001620, 0.00010576, -0.00003856, -0.00007201, 0.00002710, 0.00003186, -0.00002046, -0.00002595, 0.00001247, 0.00002769, -0.00002102, -0.00001643, 0.00004382, -0.00001387, -0.00005741, 0.00005126, 0.00002399, -0.00012034, -0.00002258, 0.00009605, -0.00003390, -0.00003925, 0.00015060, 0.00013378, 0.00000026, 0.00000440, -0.00001740, -0.00008580, -0.00002675, 0.00003677, -0.00001038, -0.00003507, -0.00001206, -0.00001388, -0.00000543, 0.00002317, 0.00001789, },
		{ -0.00002040, -0.00000367, 0.00001054, 0.00001258, 0.00000368, 0.00000022, -0.00000156, -0.00001984, -0.00002054, 0.00001761, 0.00002291, -0.00000785, -0.00000177, -0.00001431, -0.00004924, 0.00001786, 0.00005581, -0.00006535, -0.00006745, 0.00006833, -0.00000489, -0.00008895, 0.00010027, 0.00014112, -0.00008242, -0.00006048, 0.00011419, 0.00001982, -0.00009918, -0.00000317, 0.00004993, -0.00001368, -0.00003041, 0.00000465, 0.00002389, -0.00000350, -0.00002319, 0.00002635, 0.00001646, -0.00005852, 0.00000912, 0.00006538, -0.00008087, -0.00008884, 0.00009184, 0.00003524, -0.00008992, 0.00007261, 0.00017309, 0.00002290, -0.00002007, 0.00002437, -0.00005715, -0.00007483, 0.00002703, 0.00002488, -0.00003842, -0.00002456, -0.00000636, -0.00001234, 0.00001202, 0.00002649, -0.00000167, -0.00002528, },
		{ 0.00000568, 0.00001088, 0.00000768, 0.00000307, 0.00000025, -0.00001485, -0.00002654, 0.00000290, 0.00002900, 0.00000246, -0.00000846, -0.00000102, -0.00004187, -0.00002518, 0.00005883, -0.00000447, -0.00009876, 0.00000972, 0.00005031, -0.00007507, 0.00001785, 0.00017650, 0.00000827, -0.00011148, 0.00006971, 0.00008629, -0.00009067, -0.00005498, 0.00005800, 0.00000652, -0.00004019, -0.00000020, 0.00002086, 0.00000273, -0.00001722, 0.00000903, 0.00002811, -0.00003737, -0.00002899, 0.00006750, -0.00001681, -0.00012175, 0.00003665, 0.00010103, -0.00007648, -0.00002688, 0.00017075, 0.00008088, -0.00004857, 0.00002342, 0.00000323, -0.00009304, -0.00001909, 0.00005216, -0.00001610, -0.00004281, -0.00000678, -0.00000889, -0.00000217, 0.00002670, 0.00001240, -0.00002143, -0.00002360, -0.00000710, },
		},
		{ { -0.00001506, -0.00001624, -0.00000992, 0.00000553, 0.00000616, -0.00000064, -0.00000410, -0.00001676, -0.00001268, -0.00002680, -0.00006266, -0.00002149, -0.00001135, -0.00002705, 0.00013661, 0.00017230, -0.00004687, -0.00002849, 0.00006628, -0.00006875, -0.00003857, 0.00007072, -0.00002833, -0.00003017, 0.00004008, -0.00000172, 0.00000122, -0.00001015, -0.00004588, 0.00001248, 0.00002677, -0.00000514, -0.00000322, -0.00000514, 0.00002677, 0.00001248, -0.00004588, -0.00001015, 0.00000122, -0.00000172, 0.00004008, -0.00003017, -0.00002833, 0.00007072, -0.00003857, -0.00006875, 0.00006628, -0.00002849, -0.00004687, 0.00017230, 0.00013661, -0.00002705, -0.00001135, -0.00002149, -0.00006266, -0.00002680, -0.00001268, -0.00001676, -0.00000410, -0.00000064, 0.00000616, 0.00000553, -0.00000992, -0.00001624, },
		{ -0.00001427, -0.00000489, 0.00000889, 0.00000236, -0.00000446, -0.00000622, -0.00001546, -0.00001925, -0.00004542, -0.00005446, -0.00002019, -0.00002904, 0.00004996, 0.00020083, 0.00008208, -0.00006869, 0.00003344, 0.00000126, -0.00008972, 0.00003916, 0.00003544, -0.00005853, 0.00002035, 0.00002518, -0.00001863, 0.00001303, -0.00003219, -0.00003355, 0.00003677, 0.00001065, -0.00001156, 0.00000014, 0.00000534, 0.00002786, -0.00001325, -0.00004022, -0.00000118, -0.00000128, 0.00002580, 0.00001349, -0.00005376, 0.00003300, 0.00003797, -0.00009193, 0.00001498, 0.00005996, -0.00009235, 0.00004635, 0.00020058, 0.00002761, -0.00004285, 0.00001450, -0.00004381, -0.00005410, -0.00000757, -0.00001782, -0.00001516, 0.00000134, 0.00000191, 0.00000738, 0.00000102, -0.00001474, -0.00001745, -0.00001412, },
		{ 0.00000101, 0.00000929, -0.00000430, -0.00000517, -0.00000528, -0.00002025, -0.00003256, -0.00005201, -0.00004813, -0.00003439, -0.00000135, 0.00014708, 0.00018721, -0.00000277, -0.00003274, 0.00003501, -0.00006964, -0.00003643, 0.00007166, -0.00002839, -0.00003184, 0.00004804, -0.00001306, -0.00000661, 0.00001019, -0.00005362, 0.00000039, 0.00004198, -0.00001212, -0.00000610, 0.00000564, 0.00001179, 0.00001837, -0.00003003, -0.00003003, 0.00000187, 0.00001438, 0.00003129, -0.00003039, -0.00002455, 0.00006573, -0.00003891, -0.00006611, 0.00008830, -0.00002081, -0.00008491, 0.00015653, 0.00013514, -0.00005724, -0.00000486, 0.00001194, -0.00006504, -0.00002724, -0.00000354, -0.00002587, -0.00000603, 0.00000474, 0.00000276, 0.00000744, -0.00000491, -0.00001895, -0.00001650, -0.00001232, -0.00001228, },
		{ 0.00000445, -0.00000974, -0.00000045, -0.00000862, -0.00003283, -0.00004064, -0.00005201, -0.00005654, -0.00003392, 0.00007534, 0.00020890, 0.00011815, -0.00003843, 0.00000335, -0.00001836, -0.00008265, 0.00003768, 0.00003705, -0.00005990, 0.00002262, 0.00002803, -0.00003472, 0.00001908, -0.00001764, -0.00005061, 0.00003856, 0.00002063, -0.00002385, 0.00000790, 0.00000568, 0.00001426, 0.00000562, -0.00003944, -0.00002056, 0.00001306, 0.00003101, 0.00000342, -0.00004729, 0.00003128, 0.00003320, -0.00008935, 0.00002393, 0.00007994, -0.00010639, 0.00001203, 0.00020060, 0.00001373, -0.00006786, 0.00003875, -0.00002174, -0.00006066, -0.00000144, -0.00001643, -0.00002384, 0.00000497, 0.00000340, 0.00000449, 0.00000539, -0.00001207, -0.00002075, -0.00001369, -0.00001111, -0.00000920, 0.00000674, },
		{ -0.00000844, 0.00000467, -0.00002239, -0.00004422, -0.00003951, -0.00006057, -0.00006972, 0.00000874, 0.00016748, 0.00020790, 0.00004292, -0.00003369, -0.00000314, -0.00007067, -0.00003122, 0.00006811, -0.00002528, -0.00003389, 0.00005240, -0.00001798, -0.00001889, 0.00002603, -0.00005184, -0.00001487, 0.00005349, -0.00001323, -0.00001370, 0.00001682, 0.00000052, 0.00001721, -0.00000967, -0.00004470, -0.00000511, 0.00003192, 0.00002443, -0.00003312, -0.00001759, 0.00005805, -0.00003915, -0.00006045, 0.00010060, -0.00001157, -0.00011258, 0.00013672, 0.00013799, -0.00008254, -0.00001373, 0.00004325, -0.00005587, -0.00003097, 0.00000275, -0.00003046, -0.00000957, 0.00000899, 0.00000124, 0.00000707, -0.00000039, -0.00001798, -0.00001964, -0.00001069, -0.00001019, -0.00000438, 0.00000905, -0.00000424, },
		{ 0.00000071, -0.00004298, -0.00004410, -0.00003938, -0.00008383, -0.00006059, 0.00009376, 0.00023146, 0.00015698, -0.00000669, -0.00002761, -0.00004344, -0.00007204, 0.00003511, 0.00003642, -0.00005699, 0.00002035, 0.00003189, -0.00004572, 0.00001645, -0.00000078, -0.00006032, 0.00003299, 0.00003247, -0.00003308, 0.00000978, 0.00001053, -0.00000029, 0.00001918, -0.00003019, -0.00004052, 0.00002107, 0.00004006, -0.00000743, -0.00004089, 0.00003128, 0.00002494, -0.00008329, 0.00003001, 0.00009134, -0.00011038, -0.00001765, 0.00019480, 0.00000889, -0.00009483, 0.00004786, 0.00000621, -0.00005953, -0.00000170, -0.00001413, -0.00002882, 0.00000546, 0.00000526, 0.00000299, 0.00000648, -0.00000803, -0.00002055, -0.00001731, -0.00000782, -0.00000856, 0.00000037, 0.00000596, -0.00001111, 0.00000005, },
		{ -0.00005815, -0.00003277, -0.00005537, -0.00010804, -0.00000672, 0.00019347, 0.00024316, 0.00008753, -0.00003409, -0.00004445, -0.00007337, -0.00002284, 0.00006086, -0.00002037, -0.00003456, 0.00005137, -0.00001693, -0.00003159, 0.00003373, -0.00004183, -0.00002902, 0.00005826, -0.00000841, -0.00002327, 0.00002482, -0.00000508, 0.00001045, 0.00001055, -0.00005099, -0.00001772, 0.00004685, 0.00002127, -0.00003769, -0.00000883, 0.00004945, -0.00004071, -0.00005167, 0.00010444, -0.00000340, -0.00012725, 0.00011311, 0.00013908, -0.00009733, -0.00003247, 0.00006463, -0.00003587, -0.00003321, 0.00000291, -0.00003075, -0.00001191, 0.00000965, 0.00000200, 0.00000648, 0.00000058, -0.00001296, -0.00002145, -0.00001443, -0.00000404, -0.00000727, 0.00000274, -0.00000038, -0.00001109, 0.00000880, -0.00001735, },
		{ -0.00002357, -0.00009230, -0.00010605, 0.00009249, 0.00027158, 0.00020490, 0.00002076, -0.00005842, -0.00007154, -0.00006023, 0.00003287, 0.00003296, -0.00005007, 0.00001489, 0.00003397, -0.00004941, 0.00000712, 0.00001348, -0.00006054, 0.00002213, 0.00004250, -0.00003656, 0.00000587, 0.00001712, -0.00001203, 0.00002553, -0.00001522, -0.00005624, 0.00002139, 0.00005051, -0.00001638, -0.00003572, 0.00003316, 0.00001424, -0.00007575, 0.00003493, 0.00009373, -0.00010613, -0.00003911, 0.00018010, 0.00000983, -0.00011545, 0.00004195, 0.00003264, -0.00004757, -0.00000624, -0.00001462, -0.00002812, 0.00000387, 0.00000495, 0.00000496, 0.00000463, -0.00000578, -0.00001434, -0.00002371, -0.00000819, -0.00000082, -0.00000855, 0.00000342, -0.00000592, -0.00000445, 0.00000835, -0.00004451, -0.00005680, },
		{ -0.00013694, -0.00005335, 0.00021519, 0.00030151, 0.00013192, -0.00003921, -0.00008503, -0.00007798, -0.00001270, 0.00005161, -0.00001522, -0.00003271, 0.00004473, -0.00001206, -0.00004072, 0.00003202, -0.00002716, -0.00003789, 0.00005523, 0.00000035, -0.00003135, 0.00002717, -0.00000517, 0.00000122, 0.00002636, -0.00004825, -0.00003216, 0.00005571, 0.00002194, -0.00004245, 0.00000006, 0.00004145, -0.00004421, -0.00004080, 0.00010213, 0.00000146, -0.00012869, 0.00008845, 0.00013317, -0.00010049, -0.00005335, 0.00007148, -0.00000944, -0.00002913, -0.00000396, -0.00002914, -0.00001001, 0.00000582, 0.00000357, 0.00000843, -0.00000370, -0.00000577, -0.00001818, -0.00002483, 0.00000342, -0.00000382, -0.00001071, 0.00000509, -0.00000917, 0.00000403, -0.00000646, -0.00006832, -0.00003813, -0.00003435, },
		{ 0.00005911, 0.00032664, 0.00027123, 0.00003953, -0.00009075, -0.00009854, -0.00004961, 0.00003143, 0.00002732, -0.00004056, 0.00000814, 0.00003223, -0.00004577, -0.00000519, 0.00002119, -0.00005173, 0.00000971, 0.00004746, -0.00003348, -0.00000180, 0.00002211, -0.00001843, 0.00002514, 0.00000130, -0.00006416, 0.00001505, 0.00005904, -0.00002150, -0.00003191, 0.00003625, 0.00000269, -0.00006871, 0.00003959, 0.00008838, -0.00009668, -0.00004985, 0.00015680, 0.00001139, -0.00012452, 0.00002537, 0.00005055, -0.00002512, -0.00001086, -0.00002088, -0.00002195, 0.00000286, 0.00000013, 0.00001072, 0.00000204, -0.00000859, -0.00000171, -0.00002949, -0.00001577, 0.00001382, -0.00001635, -0.00000660, 0.00000657, -0.00001113, 0.00000914, -0.00003425, -0.00007571, -0.00001377, -0.00007601, -0.00016105, },
		{ 0.00038831, 0.00018146, -0.00005553, -0.00012235, -0.00008265, -0.00000350, 0.00004224, -0.00001080, -0.00002841, 0.00003378, -0.00000624, -0.00004371, 0.00002232, -0.00001249, -0.00003884, 0.00004569, 0.00000979, -0.00003500, 0.00002327, -0.00000179, -0.00000688, 0.00003521, -0.00003831, -0.00004401, 0.00005676, 0.00002492, -0.00004533, 0.00000777, 0.00003490, -0.00004939, -0.00002968, 0.00009617, 0.00000195, -0.00011898, 0.00006645, 0.00011750, -0.00009476, -0.00006900, 0.00006349, 0.00001688, -0.00001563, -0.00001597, -0.00002915, -0.00000196, -0.00000139, 0.00000330, 0.00001510, -0.00001280, -0.00000001, -0.00000689, -0.00004058, 0.00000785, 0.00000877, -0.00002990, 0.00000824, 0.00000098, -0.00001019, 0.00000605, -0.00006807, -0.00005851, -0.00000459, -0.00014394, -0.00013010, 0.00021869, },
		{ 0.00004885, -0.00012925, -0.00011969, -0.00004147, 0.00002971, 0.00002132, -0.00003033, 0.00000167, 0.00002618, -0.00003681, -0.00001639, 0.00002053, -0.00003690, -0.00000016, 0.00004574, -0.00002515, -0.00001008, 0.00002280, -0.00001892, 0.00002041, 0.00001515, -0.00006298, 0.00000507, 0.00006257, -0.00002226, -0.00002863, 0.00003948, -0.00000813, -0.00006355, 0.00004388, 0.00007769, -0.00008520, -0.00004943, 0.00012784, 0.00000864, -0.00012086, 0.00000496, 0.00005548, 0.00000405, -0.00001087, -0.00003390, -0.00001269, 0.00000554, -0.00001020, 0.00001868, 0.00000189, -0.00001846, 0.00001518, -0.00003022, -0.00003241, 0.00003250, -0.00001842, -0.00002608, 0.00002445, -0.00001537, -0.00000130, -0.00001241, -0.00009434, -0.00001871, -0.00003493, -0.00021168, -0.00001502, 0.00038739, 0.00036606, },
		{ -0.00015675, -0.00008349, 0.00000161, 0.00003389, -0.00000681, -0.00002297, 0.00002076, -0.00000191, -0.00004015, 0.00000813, -0.00000206, -0.00003173, 0.00003286, 0.00001657, -0.00003288, 0.00001456, 0.00000205, -0.00001108, 0.00003681, -0.00002479, -0.00004972, 0.00005049, 0.00002764, -0.00004476, 0.00001384, 0.00002992, -0.00005536, -0.00002029, 0.00008855, -0.00000150, -0.00010166, 0.00005025, 0.00009240, -0.00008512, -0.00007442, 0.00004441, 0.00003644, 0.00000734, -0.00002906, -0.00003397, 0.00001210, -0.00000915, -0.00000152, 0.00002700, -0.00002416, 0.00000035, 0.00001317, -0.00005786, 0.00000475, 0.00003100, -0.00005096, 0.00000459, 0.00002118, -0.00003133, 0.00001406, -0.00005239, -0.00009211, 0.00002325, -0.00011990, -0.00022837, 0.00018889, 0.00050294, 0.00024566, -0.00008873, },
		{ -0.00003499, 0.00002502, 0.00001727, -0.00002086, -0.00000409, 0.00001704, -0.00002565, -0.00002344, 0.00001243, -0.00002054, -0.00000446, 0.00003797, -0.00001448, -0.00001592, 0.00001814, -0.00001499, 0.00001466, 0.00002344, -0.00005393, -0.00000460, 0.00005942, -0.00001969, -0.00002455, 0.00004195, -0.00001712, -0.00006087, 0.00004689, 0.00006449, -0.00007401, -0.00003955, 0.00009756, -0.00000145, -0.00010708, -0.00001223, 0.00004670, 0.00003384, -0.00000285, -0.00005206, -0.00000415, 0.00001421, -0.00002498, 0.00002573, 0.00000699, -0.00003513, 0.00003215, -0.00002201, -0.00005720, 0.00005027, -0.00001078, -0.00005441, 0.00004344, -0.00001219, -0.00002489, 0.00001706, -0.00010453, -0.00004397, 0.00002451, -0.00024240, -0.00013163, 0.00043882, 0.00049716, 0.00005299, -0.00018069, -0.00013125, },
		{ 0.00002609, -0.00000157, -0.00001849, 0.00000779, -0.00000005, -0.00003191, -0.00000625, 0.00000179, -0.00001899, 0.00002081, 0.00001858, -0.00002573, 0.00000393, 0.00000352, -0.00001036, 0.00003304, -0.00001177, -0.00004770, 0.00003947, 0.00002752, -0.00004035, 0.00001866, 0.00002615, -0.00006106, -0.00001421, 0.00008034, -0.00000732, -0.00008072, 0.00004136, 0.00006089, -0.00007688, -0.00006820, 0.00002060, 0.00004455, 0.00003663, -0.00003833, -0.00004514, 0.00002978, -0.00001383, -0.00001242, 0.00004254, -0.00003387, -0.00000756, 0.00003999, -0.00007133, -0.00000892, 0.00006040, -0.00006846, -0.00000931, 0.00005043, -0.00005423, 0.00001149, -0.00001968, -0.00013252, 0.00003762, -0.00006179, -0.00033303, 0.00011335, 0.00063726, 0.00033629, -0.00014124, -0.00019208, -0.00007536, 0.00000031, },
		{ -0.00001236, -0.00001016, 0.00000724, -0.00001527, -0.00002544, 0.00000015, -0.00000714, -0.00000239, 0.00002691, -0.00000487, -0.00001749, 0.00000909, -0.00000941, 0.00001082, 0.00002539, -0.00004025, -0.00001050, 0.00004997, -0.00001595, -0.00001863, 0.00004333, -0.00002381, -0.00006055, 0.00004740, 0.00005124, -0.00006422, -0.00002345, 0.00007014, -0.00001916, -0.00008821, -0.00002096, 0.00002731, 0.00005766, 0.00001405, -0.00007151, -0.00000014, 0.00002925, -0.00004122, 0.00002841, 0.00001858, -0.00005592, 0.00004416, -0.00000302, -0.00008631, 0.00006141, 0.00000846, -0.00008811, 0.00005793, 0.00000206, -0.00006008, 0.00004727, -0.00009880, -0.00009005, 0.00009054, -0.00023977, -0.00028383, 0.00046278, 0.00066673, 0.00006165, -0.00025204, -0.00013228, -0.00002678, 0.00001375, 0.00001696, },
		{ -0.00000410, 0.00000022, -0.00002226, -0.00001726, -0.00000063, -0.00000491, 0.00001295, 0.00001580, -0.00001611, -0.00000540, 0.00000116, -0.00000574, 0.00002701, -0.00000238, -0.00003884, 0.00002738, 0.00002303, -0.00003309, 0.00002315, 0.00002320, -0.00006565, -0.00001230, 0.00007180, -0.00001341, -0.00005960, 0.00003940, 0.00002753, -0.00007390, -0.00005249, -0.00000092, 0.00003977, 0.00006669, -0.00003970, -0.00006177, 0.00004708, -0.00001232, -0.00002892, 0.00005831, -0.00003805, -0.00002433, 0.00006903, -0.00007574, -0.00003342, 0.00009182, -0.00007691, -0.00003451, 0.00008475, -0.00007241, -0.00000580, 0.00002727, -0.00017018, 0.00002927, 0.00002590, -0.00042523, -0.00001396, 0.00077560, 0.00046431, -0.00021013, -0.00023545, -0.00005347, -0.00000761, 0.00001638, 0.00000747, -0.00001687, },
		{ -0.00000729, -0.00002389, -0.00001187, 0.00000033, 0.00000427, 0.00001648, 0.00000107, -0.00001491, -0.00000172, -0.00000514, 0.00001032, 0.00002247, -0.00002596, -0.00001084, 0.00003658, -0.00001343, -0.00001071, 0.00004404, -0.00002820, -0.00006208, 0.00004437, 0.00003964, -0.00005572, -0.00000501, 0.00004849, -0.00004222, -0.00006995, -0.00001913, 0.00000312, 0.00007041, 0.00003764, -0.00008726, -0.00000314, 0.00004879, -0.00005485, 0.00002417, 0.00003567, -0.00007636, 0.00004695, 0.00002560, -0.00011381, 0.00006124, 0.00003809, -0.00012141, 0.00006262, 0.00002828, -0.00010194, 0.00007018, -0.00007270, -0.00015172, 0.00014870, -0.00019610, -0.00045638, 0.00044190, 0.00086897, 0.00008855, -0.00034790, -0.00012584, -0.00001140, -0.00000736, 0.00002057, -0.00000342, -0.00001863, -0.00000067, },
		{ -0.00002336, -0.00000644, 0.00000584, 0.00001079, 0.00001036, -0.00000741, -0.00001113, -0.00000461, 0.00000023, 0.00002184, 0.00000216, -0.00002608, 0.00001776, 0.00001446, -0.00002507, 0.00002810, 0.00002103, -0.00006873, -0.00001468, 0.00006262, -0.00001779, -0.00004060, 0.00004235, -0.00000296, -0.00007762, -0.00003198, -0.00001453, 0.00002429, 0.00009115, -0.00003136, -0.00008065, 0.00005961, -0.00000322, -0.00004841, 0.00007012, -0.00003419, -0.00004782, 0.00009436, -0.00006759, -0.00006577, 0.00011869, -0.00007247, -0.00006889, 0.00011801, -0.00007995, -0.00003850, 0.00008162, -0.00019485, -0.00000611, 0.00013161, -0.00048637, -0.00019000, 0.00089696, 0.00063627, -0.00028610, -0.00029557, -0.00001537, -0.00001934, 0.00000078, 0.00002274, -0.00001882, -0.00001551, 0.00000128, -0.00001467, },
		{ 0.00000151, 0.00001161, 0.00001041, 0.00000260, -0.00001027, -0.00001101, -0.00000413, 0.00001264, 0.00001761, -0.00001449, -0.00000609, 0.00002283, -0.00001368, -0.00000185, 0.00004492, -0.00003043, -0.00006491, 0.00003729, 0.00003043, -0.00004770, 0.00001225, 0.00003370, -0.00006665, -0.00005693, -0.00000805, -0.00001912, 0.00006987, 0.00006321, -0.00009475, -0.00001336, 0.00006905, -0.00006190, 0.00001241, 0.00005509, -0.00009150, 0.00003848, 0.00005969, -0.00013317, 0.00004760, 0.00007375, -0.00014766, 0.00005417, 0.00006400, -0.00014313, 0.00007878, -0.00002641, -0.00021907, 0.00018532, -0.00011172, -0.00062794, 0.00036394, 0.00108931, 0.00014861, -0.00046788, -0.00011927, 0.00001727, -0.00003971, 0.00002597, 0.00000882, -0.00003179, -0.00000512, -0.00000129, -0.00002207, -0.00001958, },
		{ 0.00001323, 0.00000594, -0.00000263, -0.00001274, -0.00001153, 0.00000448, 0.00001943, 0.00000285, -0.00001339, 0.00001271, 0.00000393, -0.00001868, 0.00003358, 0.00002013, -0.00007026, -0.00002097, 0.00005241, -0.00001915, -0.00002476, 0.00004739, -0.00002691, -0.00008696, -0.00001225, -0.00001741, 0.00000309, 0.00010469, -0.00001439, -0.00009708, 0.00006396, 0.00001253, -0.00006673, 0.00007434, -0.00002209, -0.00007357, 0.00011032, -0.00004643, -0.00010021, 0.00013472, -0.00005431, -0.00010741, 0.00014340, -0.00007321, -0.00008324, 0.00013402, -0.00019834, -0.00006640, 0.00023896, -0.00050238, -0.00040184, 0.00097868, 0.00085107, -0.00035415, -0.00038022, 0.00003734, -0.00002916, -0.00002516, 0.00004529, -0.00002312, -0.00002861, 0.00000603, -0.00001134, -0.00002557, -0.00001120, 0.00000979, },
		{ 0.00000124, -0.00000682, -0.00001608, -0.00000649, 0.00001555, 0.00001407, -0.00000749, 0.00000114, 0.00001232, -0.00001666, 0.00000591, 0.00004669, -0.00003045, -0.00006868, 0.00002627, 0.00002368, -0.00003919, 0.00002583, 0.00002515, -0.00008810, -0.00005154, 0.00000837, -0.00003366, 0.00005730, 0.00008491, -0.00009134, -0.00002849, 0.00008544, -0.00005989, -0.00000511, 0.00007236, -0.00009743, 0.00001965, 0.00009307, -0.00013919, 0.00002182, 0.00010882, -0.00016116, 0.00003242, 0.00010374, -0.00017578, 0.00006910, 0.00003495, -0.00027970, 0.00019089, 0.00000447, -0.00077503, 0.00022539, 0.00130562, 0.00025418, -0.00060492, -0.00012309, 0.00006396, -0.00008158, 0.00002880, 0.00002770, -0.00005094, -0.00000640, 0.00000505, -0.00002379, -0.00002085, -0.00000115, 0.00001492, 0.00001077, },
		{ -0.00001178, -0.00001642, 0.00000473, 0.00001975, 0.00000237, -0.00000437, 0.00001216, -0.00000517, -0.00001556, 0.00003861, 0.00002127, -0.00007028, -0.00003055, 0.00004101, -0.00001701, -0.00001215, 0.00005176, -0.00004242, -0.00009902, 0.00000187, -0.00001000, -0.00001758, 0.00010457, 0.00000747, -0.00010639, 0.00005889, 0.00003181, -0.00007944, 0.00006912, -0.00000411, -0.00009602, 0.00011316, -0.00001522, -0.00012965, 0.00013575, -0.00002509, -0.00014323, 0.00015537, -0.00005203, -0.00013329, 0.00017484, -0.00017678, -0.00014331, 0.00033030, -0.00046707, -0.00062821, 0.00100122, 0.00109819, -0.00039591, -0.00049334, 0.00009814, -0.00002943, -0.00006463, 0.00007372, -0.00002655, -0.00004630, 0.00001657, -0.00001227, -0.00002714, -0.00001048, 0.00000597, 0.00001679, 0.00000638, -0.00000310, },
		{ -0.00001041, 0.00001603, 0.00001406, -0.00000441, 0.00000707, 0.00000736, -0.00002055, 0.00001028, 0.00004942, -0.00002782, -0.00007319, 0.00001185, 0.00001906, -0.00002954, 0.00003450, 0.00002112, -0.00010301, -0.00005371, 0.00002504, -0.00003725, 0.00003688, 0.00009745, -0.00007718, -0.00004445, 0.00009395, -0.00004865, -0.00002468, 0.00008305, -0.00009252, -0.00000577, 0.00011910, -0.00012949, -0.00001143, 0.00013613, -0.00015880, 0.00000060, 0.00014035, -0.00019349, 0.00004145, 0.00010232, -0.00032181, 0.00016242, 0.00013656, -0.00087680, 0.00003347, 0.00149159, 0.00041112, -0.00074497, -0.00014864, 0.00012993, -0.00012762, 0.00002328, 0.00005566, -0.00007536, -0.00000688, 0.00001511, -0.00003185, -0.00001628, -0.00000322, 0.00001031, 0.00001709, 0.00000031, -0.00000776, -0.00001589, },
		{ 0.00002091, 0.00000397, -0.00000099, 0.00001386, -0.00000935, -0.00001588, 0.00004131, 0.00002504, -0.00006865, -0.00004281, 0.00002856, -0.00001159, -0.00000233, 0.00005346, -0.00004939, -0.00011015, 0.00000756, 0.00000451, -0.00003207, 0.00009131, 0.00002889, -0.00010539, 0.00004571, 0.00005018, -0.00008313, 0.00005508, 0.00001550, -0.00011014, 0.00010220, 0.00002031, -0.00014755, 0.00012102, 0.00000956, -0.00016947, 0.00015116, -0.00001990, -0.00018006, 0.00019663, -0.00013180, -0.00022417, 0.00039060, -0.00038377, -0.00084299, 0.00095266, 0.00135785, -0.00039333, -0.00063247, 0.00015600, -0.00001258, -0.00011799, 0.00010381, -0.00002467, -0.00007095, 0.00003306, -0.00001507, -0.00003236, -0.00000150, -0.00000356, 0.00001672, 0.00001426, -0.00000786, -0.00001131, -0.00001631, 0.00000039, },
		{ -0.00000281, 0.00000809, 0.00000812, -0.00002225, 0.00000967, 0.00005214, -0.00002193, -0.00007817, -0.00000519, 0.00001621, -0.00001866, 0.00003811, 0.00001956, -0.00010942, -0.00006129, 0.00003728, -0.00002973, 0.00001438, 0.00009772, -0.00005521, -0.00005660, 0.00009243, -0.00003046, -0.00004167, 0.00008415, -0.00007793, -0.00003218, 0.00013253, -0.00010535, -0.00004529, 0.00014989, -0.00014101, -0.00003536, 0.00016680, -0.00019306, 0.00000039, 0.00016480, -0.00033697, 0.00010423, 0.00026490, -0.00091924, -0.00019425, 0.00162145, 0.00061598, -0.00086864, -0.00020501, 0.00021135, -0.00016939, 0.00000394, 0.00009295, -0.00010189, -0.00001036, 0.00003190, -0.00004678, -0.00000916, 0.00000027, -0.00000380, 0.00002624, 0.00000382, -0.00001494, -0.00001154, -0.00001279, 0.00001153, 0.00001797, },
		{ 0.00001420, -0.00000634, -0.00001796, 0.00003953, 0.00003130, -0.00006485, -0.00005713, 0.00001548, -0.00000357, 0.00000515, 0.00005155, -0.00004908, -0.00011720, 0.00000443, 0.00002149, -0.00003684, 0.00006849, 0.00004465, -0.00009343, 0.00002789, 0.00006325, -0.00007653, 0.00003522, 0.00003187, -0.00011284, 0.00008002, 0.00005316, -0.00014974, 0.00009350, 0.00004244, -0.00018111, 0.00013158, 0.00001691, -0.00021529, 0.00019592, -0.00007019, -0.00029490, 0.00041078, -0.00026487, -0.00102005, 0.00083186, 0.00160352, -0.00033303, -0.00078760, 0.00019726, 0.00002653, -0.00018170, 0.00012904, -0.00001307, -0.00010315, 0.00005322, -0.00001565, -0.00004494, 0.00001568, -0.00001427, 0.00000857, 0.00002970, -0.00001384, -0.00001481, -0.00000971, -0.00000707, 0.00001990, 0.00001014, -0.00000293, },
		{ -0.00001842, 0.00000398, 0.00005296, -0.00001238, -0.00008303, -0.00002405, 0.00001482, -0.00000700, 0.00003722, 0.00001880, -0.00010712, -0.00007111, 0.00004212, -0.00001380, -0.00000455, 0.00008561, -0.00003026, -0.00006126, 0.00008122, -0.00000939, -0.00005196, 0.00007506, -0.00005728, -0.00005377, 0.00013098, -0.00007142, -0.00007253, 0.00014727, -0.00011173, -0.00006836, 0.00017812, -0.00017521, -0.00004640, 0.00021225, -0.00032220, 0.00002691, 0.00037024, -0.00089817, -0.00043251, 0.00167507, 0.00085494, -0.00095444, -0.00029594, 0.00029905, -0.00019711, -0.00003260, 0.00013669, -0.00012532, -0.00002081, 0.00005670, -0.00006634, -0.00000446, 0.00001300, -0.00002587, 0.00003331, 0.00001496, -0.00002799, -0.00000602, -0.00001037, 0.00000107, 0.00002409, 0.00000067, 0.00000209, 0.00001245, },
		{ 0.00003180, 0.00003884, -0.00005816, -0.00007271, 0.00000225, 0.00000622, 0.00001055, 0.00004610, -0.00004341, -0.00011826, -0.00000574, 0.00003630, -0.00003127, 0.00004153, 0.00005114, -0.00007263, 0.00000997, 0.00006799, -0.00006094, 0.00001403, 0.00004098, -0.00010388, 0.00005175, 0.00007695, -0.00013561, 0.00005912, 0.00006676, -0.00017631, 0.00010086, 0.00005103, -0.00023307, 0.00017396, -0.00000227, -0.00034323, 0.00038953, -0.00012917, -0.00113824, 0.00064949, 0.00180627, -0.00021004, -0.00094173, 0.00020859, 0.00008845, -0.00024826, 0.00014199, 0.00001090, -0.00014080, 0.00007259, -0.00000963, -0.00006647, 0.00003791, -0.00002046, -0.00001014, 0.00005004, -0.00001674, -0.00002479, 0.00000271, -0.00001449, 0.00001382, 0.00002154, -0.00000773, 0.00000946, 0.00000399, -0.00001873, },
		{ 0.00000046, -0.00008678, -0.00004389, 0.00001466, 0.00000474, 0.00003267, 0.00001790, -0.00009736, -0.00008008, 0.00003886, 0.00000617, -0.00001599, 0.00006401, -0.00000764, -0.00005689, 0.00006307, 0.00000992, -0.00005317, 0.00005776, -0.00003554, -0.00006611, 0.00011552, -0.00003448, -0.00008747, 0.00012924, -0.00007738, -0.00009211, 0.00017263, -0.00014431, -0.00009005, 0.00023768, -0.00028050, -0.00005529, 0.00043774, -0.00081992, -0.00065292, 0.00164227, 0.00110513, -0.00098316, -0.00041767, 0.00037956, -0.00020198, -0.00008597, 0.00018094, -0.00013963, -0.00004089, 0.00008810, -0.00008589, -0.00000717, 0.00003597, -0.00005139, 0.00003307, 0.00003475, -0.00004412, -0.00000319, -0.00000104, -0.00001418, 0.00002924, 0.00000891, -0.00001070, 0.00001656, -0.00000679, -0.00000498, 0.00004966, },
		{ -0.00008836, -0.00001071, 0.00001694, 0.00001387, 0.00003783, -0.00003425, -0.00011301, -0.00001993, 0.00004570, -0.00001749, 0.00001614, 0.00004743, -0.00004727, -0.00000386, 0.00006369, -0.00003980, -0.00000384, 0.00004089, -0.00008588, 0.00002358, 0.00008764, -0.00010825, 0.00002518, 0.00007793, -0.00015689, 0.00006539, 0.00007597, -0.00023120, 0.00013628, 0.00006063, -0.00036149, 0.00033329, 0.00000212, -0.00118544, 0.00042650, 0.00194005, -0.00003009, -0.00107339, 0.00018040, 0.00016797, -0.00030736, 0.00013633, 0.00004691, -0.00017912, 0.00008571, 0.00000609, -0.00009555, 0.00005970, -0.00001713, -0.00003882, 0.00007028, -0.00001256, -0.00004167, 0.00001811, -0.00001814, 0.00000110, 0.00003663, -0.00001166, -0.00000275, 0.00001934, -0.00001470, 0.00001813, 0.00004545, -0.00004810, },
		{ 0.00001542, 0.00001574, 0.00002527, 0.00001667, -0.00008217, -0.00008604, 0.00002892, 0.00002550, -0.00001863, 0.00003789, 0.00000849, -0.00004457, 0.00004224, 0.00002365, -0.00004540, 0.00003624, -0.00001752, -0.00006742, 0.00009032, -0.00000160, -0.00008758, 0.00010021, -0.00004505, -0.00010273, 0.00015227, -0.00010707, -0.00012277, 0.00023868, -0.00021990, -0.00012774, 0.00045992, -0.00069975, -0.00082924, 0.00152518, 0.00133817, -0.00094228, -0.00055841, 0.00043770, -0.00017878, -0.00015134, 0.00021767, -0.00013965, -0.00007066, 0.00012197, -0.00009972, -0.00002061, 0.00006681, -0.00007407, 0.00002174, 0.00006137, -0.00005887, -0.00000647, 0.00001527, -0.00003107, 0.00002992, 0.00002312, -0.00002766, 0.00001627, 0.00001278, -0.00001386, 0.00004058, 0.00001508, -0.00008815, -0.00006373, },
		{ 0.00001500, 0.00002778, -0.00002305, -0.00010245, -0.00003489, 0.00004848, 0.00000048, -0.00000327, 0.00003546, -0.00002249, -0.00001110, 0.00005206, -0.00001769, -0.00001498, 0.00003227, -0.00006343, 0.00000103, 0.00008452, -0.00007362, -0.00000165, 0.00007476, -0.00012779, 0.00003200, 0.00008773, -0.00021151, 0.00009105, 0.00010881, -0.00034807, 0.00025439, 0.00010978, -0.00116059, 0.00019036, 0.00198680, 0.00019036, -0.00116059, 0.00010978, 0.00025439, -0.00034807, 0.00010881, 0.00009105, -0.00021151, 0.00008773, 0.00003200, -0.00012779, 0.00007476, -0.00000165, -0.00007362, 0.00008452, 0.00000103, -0.00006343, 0.00003227, -0.00001498, -0.00001769, 0.00005206, -0.00001110, -0.00002249, 0.00003546, -0.00000327, 0.00000048, 0.00004848, -0.00003489, -0.00010245, -0.00002305, 0.00002778, },
		{ 0.00001542, -0.00006373, -0.00008815, 0.00001508, 0.00004058, -0.00001386, 0.00001278, 0.00001627, -0.00002766, 0.00002312, 0.00002992, -0.00003107, 0.00001527, -0.00000647, -0.00005887, 0.00006137, 0.00002174, -0.00007407, 0.00006681, -0.00002061, -0.00009972, 0.00012197, -0.00007066, -0.00013965, 0.00021767, -0.00015134, -0.00017878, 0.00043770, -0.00055841, -0.00094228, 0.00133817, 0.00152518, -0.00082924, -0.00069975, 0.00045992, -0.00012774, -0.00021990, 0.00023868, -0.00012277, -0.00010707, 0.00015227, -0.00010273, -0.00004505, 0.00010021, -0.00008758, -0.00000160, 0.00009032, -0.00006742, -0.00001752, 0.00003624, -0.00004540, 0.00002365, 0.00004224, -0.00004457, 0.00000849, 0.00003789, -0.00001863, 0.00002550, 0.00002892, -0.00008604, -0.00008217, 0.00001667, 0.00002527, 0.00001574, },
		{ -0.00008836, -0.00004810, 0.00004545, 0.00001813, -0.00001470, 0.00001934, -0.00000275, -0.00001166, 0.00003663, 0.00000110, -0.00001814, 0.00001811, -0.00004167, -0.00001256, 0.00007028, -0.00003882, -0.00001713, 0.00005970, -0.00009555, 0.00000609, 0.00008571, -0.00017912, 0.00004691, 0.00013633, -0.00030736, 0.00016797, 0.00018040, -0.00107339, -0.00003009, 0.00194005, 0.00042650, -0.00118544, 0.00000212, 0.00033329, -0.00036149, 0.00006063, 0.00013628, -0.00023120, 0.00007597, 0.00006539, -0.00015689, 0.00007793, 0.00002518, -0.00010825, 0.00008764, 0.00002358, -0.00008588, 0.00004089, -0.00000384, -0.00003980, 0.00006369, -0.00000386, -0.00004727, 0.00004743, 0.00001614, -0.00001749, 0.00004570, -0.00001993, -0.00011301, -0.00003425, 0.00003783, 0.00001387, 0.00001694, -0.00001071, },
		{ 0.00000046, 0.00004966, -0.00000498, -0.00000679, 0.00001656, -0.00001070, 0.00000891, 0.00002924, -0.00001418, -0.00000104, -0.00000319, -0.00004412, 0.00003475, 0.00003307, -0.00005139, 0.00003597, -0.00000717, -0.00008589, 0.00008810, -0.00004089, -0.00013963, 0.00018094, -0.00008597, -0.00020198, 0.00037956, -0.00041767, -0.00098316, 0.00110513, 0.00164227, -0.00065292, -0.00081992, 0.00043774, -0.00005529, -0.00028050, 0.00023768, -0.00009005, -0.00014431, 0.00017263, -0.00009211, -0.00007738, 0.00012924, -0.00008747, -0.00003448, 0.00011552, -0.00006611, -0.00003554, 0.00005776, -0.00005317, 0.00000992, 0.00006307, -0.00005689, -0.00000764, 0.00006401, -0.00001599, 0.00000617, 0.00003886, -0.00008008, -0.00009736, 0.00001790, 0.00003267, 0.00000474, 0.00001466, -0.00004389, -0.00008678, },
		{ 0.00003180, -0.00001873, 0.00000399, 0.00000946, -0.00000773, 0.00002154, 0.00001382, -0.00001449, 0.00000271, -0.00002479, -0.00001674, 0.00005004, -0.00001014, -0.00002046, 0.00003791, -0.00006647, -0.00000963, 0.00007259, -0.00014080, 0.00001090, 0.00014199, -0.00024826, 0.00008845, 0.00020859, -0.00094173, -0.00021004, 0.00180627, 0.00064949, -0.00113824, -0.00012917, 0.00038953, -0.00034323, -0.00000227, 0.00017396, -0.00023307, 0.00005103, 0.00010086, -0.00017631, 0.00006676, 0.00005912, -0.00013561, 0.00007695, 0.00005175, -0.00010388, 0.00004098, 0.00001403, -0.00006094, 0.00006799, 0.00000997, -0.00007263, 0.00005114, 0.00004153, -0.00003127, 0.00003630, -0.00000574, -0.00011826, -0.00004341, 0.00004610, 0.00001055, 0.00000622, 0.00000225, -0.00007271, -0.00005816, 0.00003884, },
		{ -0.00001842, 0.00001245, 0.00000209, 0.00000067, 0.00002409, 0.00000107, -0.00001037, -0.00000602, -0.00002799, 0.00001496, 0.00003331, -0.00002587, 0.00001300, -0.00000446, -0.00006634, 0.00005670, -0.00002081, -0.00012532, 0.00013669, -0.00003260, -0.00019711, 0.00029905, -0.00029594, -0.00095444, 0.00085494, 0.00167507, -0.00043251, -0.00089817, 0.00037024, 0.00002691, -0.00032220, 0.00021225, -0.00004640, -0.00017521, 0.00017812, -0.00006836, -0.00011173, 0.00014727, -0.00007253, -0.00007142, 0.00013098, -0.00005377, -0.00005728, 0.00007506, -0.00005196, -0.00000939, 0.00008122, -0.00006126, -0.00003026, 0.00008561, -0.00000455, -0.00001380, 0.00004212, -0.00007111, -0.00010712, 0.00001880, 0.00003722, -0.00000700, 0.00001482, -0.00002405, -0.00008303, -0.00001238, 0.00005296, 0.00000398, },
		{ 0.00001420, -0.00000293, 0.00001014, 0.00001990, -0.00000707, -0.00000971, -0.00001481, -0.00001384, 0.00002970, 0.00000857, -0.00001427, 0.00001568, -0.00004494, -0.00001565, 0.00005322, -0.00010315, -0.00001307, 0.00012904, -0.00018170, 0.00002653, 0.00019726, -0.00078760, -0.00033303, 0.00160352, 0.00083186, -0.00102005, -0.00026487, 0.00041078, -0.00029490, -0.00007019, 0.00019592, -0.00021529, 0.00001691, 0.00013158, -0.00018111, 0.00004244, 0.00009350, -0.00014974, 0.00005316, 0.00008002, -0.00011284, 0.00003187, 0.00003522, -0.00007653, 0.00006325, 0.00002789, -0.00009343, 0.00004465, 0.00006849, -0.00003684, 0.00002149, 0.00000443, -0.00011720, -0.00004908, 0.00005155, 0.00000515, -0.00000357, 0.00001548, -0.00005713, -0.00006485, 0.00003130, 0.00003953, -0.00001796, -0.00000634, },
		{ -0.00000281, 0.00001797, 0.00001153, -0.00001279, -0.00001154, -0.00001494, 0.00000382, 0.00002624, -0.00000380, 0.00000027, -0.00000916, -0.00004678, 0.00003190, -0.00001036, -0.00010189, 0.00009295, 0.00000394, -0.00016939, 0.00021135, -0.00020501, -0.00086864, 0.00061598, 0.00162145, -0.00019425, -0.00091924, 0.00026490, 0.00010423, -0.00033697, 0.00016480, 0.00000039, -0.00019306, 0.00016680, -0.00003536, -0.00014101, 0.00014989, -0.00004529, -0.00010535, 0.00013253, -0.00003218, -0.00007793, 0.00008415, -0.00004167, -0.00003046, 0.00009243, -0.00005660, -0.00005521, 0.00009772, 0.00001438, -0.00002973, 0.00003728, -0.00006129, -0.00010942, 0.00001956, 0.00003811, -0.00001866, 0.00001621, -0.00000519, -0.00007817, -0.00002193, 0.00005214, 0.00000967, -0.00002225, 0.00000812, 0.00000809, },
		{ 0.00002091, 0.00000039, -0.00001631, -0.00001131, -0.00000786, 0.00001426, 0.00001672, -0.00000356, -0.00000150, -0.00003236, -0.00001507, 0.00003306, -0.00007095, -0.00002467, 0.00010381, -0.00011799, -0.00001258, 0.00015600, -0.00063247, -0.00039333, 0.00135785, 0.00095266, -0.00084299, -0.00038377, 0.00039060, -0.00022417, -0.00013180, 0.00019663, -0.00018006, -0.00001990, 0.00015116, -0.00016947, 0.00000956, 0.00012102, -0.00014755, 0.00002031, 0.00010220, -0.00011014, 0.00001550, 0.00005508, -0.00008313, 0.00005018, 0.00004571, -0.00010539, 0.00002889, 0.00009131, -0.00003207, 0.00000451, 0.00000756, -0.00011015, -0.00004939, 0.00005346, -0.00000233, -0.00001159, 0.00002856, -0.00004281, -0.00006865, 0.00002504, 0.00004131, -0.00001588, -0.00000935, 0.00001386, -0.00000099, 0.00000397, },
		{ -0.00001041, -0.00001589, -0.00000776, 0.00000031, 0.00001709, 0.00001031, -0.00000322, -0.00001628, -0.00003185, 0.00001511, -0.00000688, -0.00007536, 0.00005566, 0.00002328, -0.00012762, 0.00012993, -0.00014864, -0.00074497, 0.00041112, 0.00149159, 0.00003347, -0.00087680, 0.00013656, 0.00016242, -0.00032181, 0.00010232, 0.00004145, -0.00019349, 0.00014035, 0.00000060, -0.00015880, 0.00013613, -0.00001143, -0.00012949, 0.00011910, -0.00000577, -0.00009252, 0.00008305, -0.00002468, -0.00004865, 0.00009395, -0.00004445, -0.00007718, 0.00009745, 0.00003688, -0.00003725, 0.00002504, -0.00005371, -0.00010301, 0.00002112, 0.00003450, -0.00002954, 0.00001906, 0.00001185, -0.00007319, -0.00002782, 0.00004942, 0.00001028, -0.00002055, 0.00000736, 0.00000707, -0.00000441, 0.00001406, 0.00001603, },
		{ -0.00001178, -0.00000310, 0.00000638, 0.00001679, 0.00000597, -0.00001048, -0.00002714, -0.00001227, 0.00001657, -0.00004630, -0.00002655, 0.00007372, -0.00006463, -0.00002943, 0.00009814, -0.00049334, -0.00039591, 0.00109819, 0.00100122, -0.00062821, -0.00046707, 0.00033030, -0.00014331, -0.00017678, 0.00017484, -0.00013329, -0.00005203, 0.00015537, -0.00014323, -0.00002509, 0.00013575, -0.00012965, -0.00001522, 0.00011316, -0.00009602, -0.00000411, 0.00006912, -0.00007944, 0.00003181, 0.00005889, -0.00010639, 0.00000747, 0.00010457, -0.00001758, -0.00001000, 0.00000187, -0.00009902, -0.00004242, 0.00005176, -0.00001215, -0.00001701, 0.00004101, -0.00003055, -0.00007028, 0.00002127, 0.00003861, -0.00001556, -0.00000517, 0.00001216, -0.00000437, 0.00000237, 0.00001975, 0.00000473, -0.00001642, },
		{ 0.00000124, 0.00001077, 0.00001492, -0.00000115, -0.00002085, -0.00002379, 0.00000505, -0.00000640, -0.00005094, 0.00002770, 0.00002880, -0.00008158, 0.00006396, -0.00012309, -0.00060492, 0.00025418, 0.00130562, 0.00022539, -0.00077503, 0.00000447, 0.00019089, -0.00027970, 0.00003495, 0.00006910, -0.00017578, 0.00010374, 0.00003242, -0.00016116, 0.00010882, 0.00002182, -0.00013919, 0.00009307, 0.00001965, -0.00009743, 0.00007236, -0.00000511, -0.00005989, 0.00008544, -0.00002849, -0.00009134, 0.00008491, 0.00005730, -0.00003366, 0.00000837, -0.00005154, -0.00008810, 0.00002515, 0.00002583, -0.00003919, 0.00002368, 0.00002627, -0.00006868, -0.00003045, 0.00004669, 0.00000591, -0.00001666, 0.00001232, 0.00000114, -0.00000749, 0.00001407, 0.00001555, -0.00000649, -0.00001608, -0.00000682, },
		{ 0.00001323, 0.00000979, -0.00001120, -0.00002557, -0.00001134, 0.00000603, -0.00002861, -0.00002312, 0.00004529, -0.00002516, -0.00002916, 0.00003734, -0.00038022, -0.00035415, 0.00085107, 0.00097868, -0.00040184, -0.00050238, 0.00023896, -0.00006640, -0.00019834, 0.00013402, -0.00008324, -0.00007321, 0.00014340, -0.00010741, -0.00005431, 0.00013472, -0.00010021, -0.00004643, 0.00011032, -0.00007357, -0.00002209, 0.00007434, -0.00006673, 0.00001253, 0.00006396, -0.00009708, -0.00001439, 0.00010469, 0.00000309, -0.00001741, -0.00001225, -0.00008696, -0.00002691, 0.00004739, -0.00002476, -0.00001915, 0.00005241, -0.00002097, -0.00007026, 0.00002013, 0.00003358, -0.00001868, 0.00000393, 0.00001271, -0.00001339, 0.00000285, 0.00001943, 0.00000448, -0.00001153, -0.00001274, -0.00000263, 0.00000594, },
		{ 0.00000151, -0.00001958, -0.00002207, -0.00000129, -0.00000512, -0.00003179, 0.00000882, 0.00002597, -0.00003971, 0.00001727, -0.00011927, -0.00046788, 0.00014861, 0.00108931, 0.00036394, -0.00062794, -0.00011172, 0.00018532, -0.00021907, -0.00002641, 0.00007878, -0.00014313, 0.00006400, 0.00005417, -0.00014766, 0.00007375, 0.00004760, -0.00013317, 0.00005969, 0.00003848, -0.00009150, 0.00005509, 0.00001241, -0.00006190, 0.00006905, -0.00001336, -0.00009475, 0.00006321, 0.00006987, -0.00001912, -0.00000805, -0.00005693, -0.00006665, 0.00003370, 0.00001225, -0.00004770, 0.00003043, 0.00003729, -0.00006491, -0.00003043, 0.00004492, -0.00000185, -0.00001368, 0.00002283, -0.00000609, -0.00001449, 0.00001761, 0.00001264, -0.00000413, -0.00001101, -0.00001027, 0.00000260, 0.00001041, 0.00001161, },
		{ -0.00002336, -0.00001467, 0.00000128, -0.00001551, -0.00001882, 0.00002274, 0.00000078, -0.00001934, -0.00001537, -0.00029557, -0.00028610, 0.00063627, 0.00089696, -0.00019000, -0.00048637, 0.00013161, -0.00000611, -0.00019485, 0.00008162, -0.00003850, -0.00007995, 0.00011801, -0.00006889, -0.00007247, 0.00011869, -0.00006577, -0.00006759, 0.00009436, -0.00004782, -0.00003419, 0.00007012, -0.00004841, -0.00000322, 0.00005961, -0.00008065, -0.00003136, 0.00009115, 0.00002429, -0.00001453, -0.00003198, -0.00007762, -0.00000296, 0.00004235, -0.00004060, -0.00001779, 0.00006262, -0.00001468, -0.00006873, 0.00002103, 0.00002810, -0.00002507, 0.00001446, 0.00001776, -0.00002608, 0.00000216, 0.00002184, 0.00000023, -0.00000461, -0.00001113, -0.00000741, 0.00001036, 0.00001079, 0.00000584, -0.00000644, },
		{ -0.00000729, -0.00000067, -0.00001863, -0.00000342, 0.00002057, -0.00000736, -0.00001140, -0.00012584, -0.00034790, 0.00008855, 0.00086897, 0.00044190, -0.00045638, -0.00019610, 0.00014870, -0.00015172, -0.00007270, 0.00007018, -0.00010194, 0.00002828, 0.00006262, -0.00012141, 0.00003809, 0.00006124, -0.00011381, 0.00002560, 0.00004695, -0.00007636, 0.00003567, 0.00002417, -0.00005485, 0.00004879, -0.00000314, -0.00008726, 0.00003764, 0.00007041, 0.00000312, -0.00001913, -0.00006995, -0.00004222, 0.00004849, -0.00000501, -0.00005572, 0.00003964, 0.00004437, -0.00006208, -0.00002820, 0.00004404, -0.00001071, -0.00001343, 0.00003658, -0.00001084, -0.00002596, 0.00002247, 0.00001032, -0.00000514, -0.00000172, -0.00001491, 0.00000107, 0.00001648, 0.00000427, 0.00000033, -0.00001187, -0.00002389, },
		{ -0.00000410, -0.00001687, 0.00000747, 0.00001638, -0.00000761, -0.00005347, -0.00023545, -0.00021013, 0.00046431, 0.00077560, -0.00001396, -0.00042523, 0.00002590, 0.00002927, -0.00017018, 0.00002727, -0.00000580, -0.00007241, 0.00008475, -0.00003451, -0.00007691, 0.00009182, -0.00003342, -0.00007574, 0.00006903, -0.00002433, -0.00003805, 0.00005831, -0.00002892, -0.00001232, 0.00004708, -0.00006177, -0.00003970, 0.00006669, 0.00003977, -0.00000092, -0.00005249, -0.00007390, 0.00002753, 0.00003940, -0.00005960, -0.00001341, 0.00007180, -0.00001230, -0.00006565, 0.00002320, 0.00002315, -0.00003309, 0.00002303, 0.00002738, -0.00003884, -0.00000238, 0.00002701, -0.00000574, 0.00000116, -0.00000540, -0.00001611, 0.00001580, 0.00001295, -0.00000491, -0.00000063, -0.00001726, -0.00002226, 0.00000022, },
		{ -0.00001236, 0.00001696, 0.00001375, -0.00002678, -0.00013228, -0.00025204, 0.00006165, 0.00066673, 0.00046278, -0.00028383, -0.00023977, 0.00009054, -0.00009005, -0.00009880, 0.00004727, -0.00006008, 0.00000206, 0.00005793, -0.00008811, 0.00000846, 0.00006141, -0.00008631, -0.00000302, 0.00004416, -0.00005592, 0.00001858, 0.00002841, -0.00004122, 0.00002925, -0.00000014, -0.00007151, 0.00001405, 0.00005766, 0.00002731, -0.00002096, -0.00008821, -0.00001916, 0.00007014, -0.00002345, -0.00006422, 0.00005124, 0.00004740, -0.00006055, -0.00002381, 0.00004333, -0.00001863, -0.00001595, 0.00004997, -0.00001050, -0.00004025, 0.00002539, 0.00001082, -0.00000941, 0.00000909, -0.00001749, -0.00000487, 0.00002691, -0.00000239, -0.00000714, 0.00000015, -0.00002544, -0.00001527, 0.00000724, -0.00001016, },
		{ 0.00002609, 0.00000031, -0.00007536, -0.00019208, -0.00014124, 0.00033629, 0.00063726, 0.00011335, -0.00033303, -0.00006179, 0.00003762, -0.00013252, -0.00001968, 0.00001149, -0.00005423, 0.00005043, -0.00000931, -0.00006846, 0.00006040, -0.00000892, -0.00007133, 0.00003999, -0.00000756, -0.00003387, 0.00004254, -0.00001242, -0.00001383, 0.00002978, -0.00004514, -0.00003833, 0.00003663, 0.00004455, 0.00002060, -0.00006820, -0.00007688, 0.00006089, 0.00004136, -0.00008072, -0.00000732, 0.00008034, -0.00001421, -0.00006106, 0.00002615, 0.00001866, -0.00004035, 0.00002752, 0.00003947, -0.00004770, -0.00001177, 0.00003304, -0.00001036, 0.00000352, 0.00000393, -0.00002573, 0.00001858, 0.00002081, -0.00001899, 0.00000179, -0.00000625, -0.00003191, -0.00000005, 0.00000779, -0.00001849, -0.00000157, },
		{ -0.00003499, -0.00013125, -0.00018069, 0.00005299, 0.00049716, 0.00043882, -0.00013163, -0.00024240, 0.00002451, -0.00004397, -0.00010453, 0.00001706, -0.00002489, -0.00001219, 0.00004344, -0.00005441, -0.00001078, 0.00005027, -0.00005720, -0.00002201, 0.00003215, -0.00003513, 0.00000699, 0.00002573, -0.00002498, 0.00001421, -0.00000415, -0.00005206, -0.00000285, 0.00003384, 0.00004670, -0.00001223, -0.00010708, -0.00000145, 0.00009756, -0.00003955, -0.00007401, 0.00006449, 0.00004689, -0.00006087, -0.00001712, 0.00004195, -0.00002455, -0.00001969, 0.00005942, -0.00000460, -0.00005393, 0.00002344, 0.00001466, -0.00001499, 0.00001814, -0.00001592, -0.00001448, 0.00003797, -0.00000446, -0.00002054, 0.00001243, -0.00002344, -0.00002565, 0.00001704, -0.00000409, -0.00002086, 0.00001727, 0.00002502, },
		{ -0.00015675, -0.00008873, 0.00024566, 0.00050294, 0.00018889, -0.00022837, -0.00011990, 0.00002325, -0.00009211, -0.00005239, 0.00001406, -0.00003133, 0.00002118, 0.00000459, -0.00005096, 0.00003100, 0.00000475, -0.00005786, 0.00001317, 0.00000035, -0.00002416, 0.00002700, -0.00000152, -0.00000915, 0.00001210, -0.00003397, -0.00002906, 0.00000734, 0.00003644, 0.00004441, -0.00007442, -0.00008512, 0.00009240, 0.00005025, -0.00010166, -0.00000150, 0.00008855, -0.00002029, -0.00005536, 0.00002992, 0.00001384, -0.00004476, 0.00002764, 0.00005049, -0.00004972, -0.00002479, 0.00003681, -0.00001108, 0.00000205, 0.00001456, -0.00003288, 0.00001657, 0.00003286, -0.00003173, -0.00000206, 0.00000813, -0.00004015, -0.00000191, 0.00002076, -0.00002297, -0.00000681, 0.00003389, 0.00000161, -0.00008349, },
		{ 0.00004885, 0.00036606, 0.00038739, -0.00001502, -0.00021168, -0.00003493, -0.00001871, -0.00009434, -0.00001241, -0.00000130, -0.00001537, 0.00002445, -0.00002608, -0.00001842, 0.00003250, -0.00003241, -0.00003022, 0.00001518, -0.00001846, 0.00000189, 0.00001868, -0.00001020, 0.00000554, -0.00001269, -0.00003390, -0.00001087, 0.00000405, 0.00005548, 0.00000496, -0.00012086, 0.00000864, 0.00012784, -0.00004943, -0.00008520, 0.00007769, 0.00004388, -0.00006355, -0.00000813, 0.00003948, -0.00002863, -0.00002226, 0.00006257, 0.00000507, -0.00006298, 0.00001515, 0.00002041, -0.00001892, 0.00002280, -0.00001008, -0.00002515, 0.00004574, -0.00000016, -0.00003690, 0.00002053, -0.00001639, -0.00003681, 0.00002618, 0.00000167, -0.00003033, 0.00002132, 0.00002971, -0.00004147, -0.00011969, -0.00012925, },
		{ 0.00038831, 0.00021869, -0.00013010, -0.00014394, -0.00000459, -0.00005851, -0.00006807, 0.00000605, -0.00001019, 0.00000098, 0.00000824, -0.00002990, 0.00000877, 0.00000785, -0.00004058, -0.00000689, -0.00000001, -0.00001280, 0.00001510, 0.00000330, -0.00000139, -0.00000196, -0.00002915, -0.00001597, -0.00001563, 0.00001688, 0.00006349, -0.00006900, -0.00009476, 0.00011750, 0.00006645, -0.00011898, 0.00000195, 0.00009617, -0.00002968, -0.00004939, 0.00003490, 0.00000777, -0.00004533, 0.00002492, 0.00005676, -0.00004401, -0.00003831, 0.00003521, -0.00000688, -0.00000179, 0.00002327, -0.00003500, 0.00000979, 0.00004569, -0.00003884, -0.00001249, 0.00002232, -0.00004371, -0.00000624, 0.00003378, -0.00002841, -0.00001080, 0.00004224, -0.00000350, -0.00008265, -0.00012235, -0.00005553, 0.00018146, },
		{ 0.00005911, -0.00016105, -0.00007601, -0.00001377, -0.00007571, -0.00003425, 0.00000914, -0.00001113, 0.00000657, -0.00000660, -0.00001635, 0.00001382, -0.00001577, -0.00002949, -0.00000171, -0.00000859, 0.00000204, 0.00001072, 0.00000013, 0.00000286, -0.00002195, -0.00002088, -0.00001086, -0.00002512, 0.00005055, 0.00002537, -0.00012452, 0.00001139, 0.00015680, -0.00004985, -0.00009668, 0.00008838, 0.00003959, -0.00006871, 0.00000269, 0.00003625, -0.00003191, -0.00002150, 0.00005904, 0.00001505, -0.00006416, 0.00000130, 0.00002514, -0.00001843, 0.00002211, -0.00000180, -0.00003348, 0.00004746, 0.00000971, -0.00005173, 0.00002119, -0.00000519, -0.00004577, 0.00003223, 0.00000814, -0.00004056, 0.00002732, 0.00003143, -0.00004961, -0.00009854, -0.00009075, 0.00003953, 0.00027123, 0.00032664, },
		{ -0.00013694, -0.00003435, -0.00003813, -0.00006832, -0.00000646, 0.00000403, -0.00000917, 0.00000509, -0.00001071, -0.00000382, 0.00000342, -0.00002483, -0.00001818, -0.00000577, -0.00000370, 0.00000843, 0.00000357, 0.00000582, -0.00001001, -0.00002914, -0.00000396, -0.00002913, -0.00000944, 0.00007148, -0.00005335, -0.00010049, 0.00013317, 0.00008845, -0.00012869, 0.00000146, 0.00010213, -0.00004080, -0.00004421, 0.00004145, 0.00000006, -0.00004245, 0.00002194, 0.00005571, -0.00003216, -0.00004825, 0.00002636, 0.00000122, -0.00000517, 0.00002717, -0.00003135, 0.00000035, 0.00005523, -0.00003789, -0.00002716, 0.00003202, -0.00004072, -0.00001206, 0.00004473, -0.00003271, -0.00001522, 0.00005161, -0.00001270, -0.00007798, -0.00008503, -0.00003921, 0.00013192, 0.00030151, 0.00021519, -0.00005335, },
		{ -0.00002357, -0.00005680, -0.00004451, 0.00000835, -0.00000445, -0.00000592, 0.00000342, -0.00000855, -0.00000082, -0.00000819, -0.00002371, -0.00001434, -0.00000578, 0.00000463, 0.00000496, 0.00000495, 0.00000387, -0.00002812, -0.00001462, -0.00000624, -0.00004757, 0.00003264, 0.00004195, -0.00011545, 0.00000983, 0.00018010, -0.00003911, -0.00010613, 0.00009373, 0.00003493, -0.00007575, 0.00001424, 0.00003316, -0.00003572, -0.00001638, 0.00005051, 0.00002139, -0.00005624, -0.00001522, 0.00002553, -0.00001203, 0.00001712, 0.00000587, -0.00003656, 0.00004250, 0.00002213, -0.00006054, 0.00001348, 0.00000712, -0.00004941, 0.00003397, 0.00001489, -0.00005007, 0.00003296, 0.00003287, -0.00006023, -0.00007154, -0.00005842, 0.00002076, 0.00020490, 0.00027158, 0.00009249, -0.00010605, -0.00009230, },
		{ -0.00005815, -0.00001735, 0.00000880, -0.00001109, -0.00000038, 0.00000274, -0.00000727, -0.00000404, -0.00001443, -0.00002145, -0.00001296, 0.00000058, 0.00000648, 0.00000200, 0.00000965, -0.00001191, -0.00003075, 0.00000291, -0.00003321, -0.00003587, 0.00006463, -0.00003247, -0.00009733, 0.00013908, 0.00011311, -0.00012725, -0.00000340, 0.00010444, -0.00005167, -0.00004071, 0.00004945, -0.00000883, -0.00003769, 0.00002127, 0.00004685, -0.00001772, -0.00005099, 0.00001055, 0.00001045, -0.00000508, 0.00002482, -0.00002327, -0.00000841, 0.00005826, -0.00002902, -0.00004183, 0.00003373, -0.00003159, -0.00001693, 0.00005137, -0.00003456, -0.00002037, 0.00006086, -0.00002284, -0.00007337, -0.00004445, -0.00003409, 0.00008753, 0.00024316, 0.00019347, -0.00000672, -0.00010804, -0.00005537, -0.00003277, },
		{ 0.00000071, 0.00000005, -0.00001111, 0.00000596, 0.00000037, -0.00000856, -0.00000782, -0.00001731, -0.00002055, -0.00000803, 0.00000648, 0.00000299, 0.00000526, 0.00000546, -0.00002882, -0.00001413, -0.00000170, -0.00005953, 0.00000621, 0.00004786, -0.00009483, 0.00000889, 0.00019480, -0.00001765, -0.00011038, 0.00009134, 0.00003001, -0.00008329, 0.00002494, 0.00003128, -0.00004089, -0.00000743, 0.00004006, 0.00002107, -0.00004052, -0.00003019, 0.00001918, -0.00000029, 0.00001053, 0.00000978, -0.00003308, 0.00003247, 0.00003299, -0.00006032, -0.00000078, 0.00001645, -0.00004572, 0.00003189, 0.00002035, -0.00005699, 0.00003642, 0.00003511, -0.00007204, -0.00004344, -0.00002761, -0.00000669, 0.00015698, 0.00023146, 0.00009376, -0.00006059, -0.00008383, -0.00003938, -0.00004410, -0.00004298, },
		{ -0.00000844, -0.00000424, 0.00000905, -0.00000438, -0.00001019, -0.00001069, -0.00001964, -0.00001798, -0.00000039, 0.00000707, 0.00000124, 0.00000899, -0.00000957, -0.00003046, 0.00000275, -0.00003097, -0.00005587, 0.00004325, -0.00001373, -0.00008254, 0.00013799, 0.00013672, -0.00011258, -0.00001157, 0.00010060, -0.00006045, -0.00003915, 0.00005805, -0.00001759, -0.00003312, 0.00002443, 0.00003192, -0.00000511, -0.00004470, -0.00000967, 0.00001721, 0.00000052, 0.00001682, -0.00001370, -0.00001323, 0.00005349, -0.00001487, -0.00005184, 0.00002603, -0.00001889, -0.00001798, 0.00005240, -0.00003389, -0.00002528, 0.00006811, -0.00003122, -0.00007067, -0.00000314, -0.00003369, 0.00004292, 0.00020790, 0.00016748, 0.00000874, -0.00006972, -0.00006057, -0.00003951, -0.00004422, -0.00002239, 0.00000467, },
		{ 0.00000445, 0.00000674, -0.00000920, -0.00001111, -0.00001369, -0.00002075, -0.00001207, 0.00000539, 0.00000449, 0.00000340, 0.00000497, -0.00002384, -0.00001643, -0.00000144, -0.00006066, -0.00002174, 0.00003875, -0.00006786, 0.00001373, 0.00020060, 0.00001203, -0.00010639, 0.00007994, 0.00002393, -0.00008935, 0.00003320, 0.00003128, -0.00004729, 0.00000342, 0.00003101, 0.00001306, -0.00002056, -0.00003944, 0.00000562, 0.00001426, 0.00000568, 0.00000790, -0.00002385, 0.00002063, 0.00003856, -0.00005061, -0.00001764, 0.00001908, -0.00003472, 0.00002803, 0.00002262, -0.00005990, 0.00003705, 0.00003768, -0.00008265, -0.00001836, 0.00000335, -0.00003843, 0.00011815, 0.00020890, 0.00007534, -0.00003392, -0.00005654, -0.00005201, -0.00004064, -0.00003283, -0.00000862, -0.00000045, -0.00000974, },
		{ 0.00000101, -0.00001228, -0.00001232, -0.00001650, -0.00001895, -0.00000491, 0.00000744, 0.00000276, 0.00000474, -0.00000603, -0.00002587, -0.00000354, -0.00002724, -0.00006504, 0.00001194, -0.00000486, -0.00005724, 0.00013514, 0.00015653, -0.00008491, -0.00002081, 0.00008830, -0.00006611, -0.00003891, 0.00006573, -0.00002455, -0.00003039, 0.00003129, 0.00001438, 0.00000187, -0.00003003, -0.00003003, 0.00001837, 0.00001179, 0.00000564, -0.00000610, -0.00001212, 0.00004198, 0.00000039, -0.00005362, 0.00001019, -0.00000661, -0.00001306, 0.00004804, -0.00003184, -0.00002839, 0.00007166, -0.00003643, -0.00006964, 0.00003501, -0.00003274, -0.00000277, 0.00018721, 0.00014708, -0.00000135, -0.00003439, -0.00004813, -0.00005201, -0.00003256, -0.00002025, -0.00000528, -0.00000517, -0.00000430, 0.00000929, },
		{ -0.00001427, -0.00001412, -0.00001745, -0.00001474, 0.00000102, 0.00000738, 0.00000191, 0.00000134, -0.00001516, -0.00001782, -0.00000757, -0.00005410, -0.00004381, 0.00001450, -0.00004285, 0.00002761, 0.00020058, 0.00004635, -0.00009235, 0.00005996, 0.00001498, -0.00009193, 0.00003797, 0.00003300, -0.00005376, 0.00001349, 0.00002580, -0.00000128, -0.00000118, -0.00004022, -0.00001325, 0.00002786, 0.00000534, 0.00000014, -0.00001156, 0.00001065, 0.00003677, -0.00003355, -0.00003219, 0.00001303, -0.00001863, 0.00002518, 0.00002035, -0.00005853, 0.00003544, 0.00003916, -0.00008972, 0.00000126, 0.00003344, -0.00006869, 0.00008208, 0.00020083, 0.00004996, -0.00002904, -0.00002019, -0.00005446, -0.00004542, -0.00001925, -0.00001546, -0.00000622, -0.00000446, 0.00000236, 0.00000889, -0.00000489, },
		},
		{ { -0.00001611, -0.00001108, -0.00000750, -0.00000371, -0.00001650, -0.00001393, -0.00002090, -0.00007440, -0.00003423, 0.00003502, 0.00010520, 0.00015439, 0.00001491, -0.00001590, 0.00000908, -0.00010324, 0.00001529, 0.00005861, -0.00007162, 0.00001756, -0.00000265, -0.00004293, 0.00004374, -0.00001001, 0.00002654, 0.00001241, -0.00010775, 0.00004331, 0.00009261, -0.00005586, -0.00000748, 0.00001737, -0.00004612, 0.00001737, -0.00000748, -0.00005586, 0.00009261, 0.00004331, -0.00010775, 0.00001241, 0.00002654, -0.00001001, 0.00004374, -0.00004293, -0.00000265, 0.00001756, -0.00007162, 0.00005861, 0.00001529, -0.00010324, 0.00000908, -0.00001590, 0.00001491, 0.00015439, 0.00010520, 0.00003502, -0.00003423, -0.00007440, -0.00002090, -0.00001393, -0.00001650, -0.00000371, -0.00000750, -0.00001108, },
		{ -0.00001062, -0.00000430, -0.00000407, -0.00002061, -0.00001429, -0.00003403, -0.00007908, -0.00002272, 0.00006182, 0.00014936, 0.00013420, -0.00001147, -0.00000738, -0.00003654, -0.00008192, 0.00005934, 0.00000532, -0.00005792, 0.00003251, -0.00003405, -0.00001039, 0.00003370, -0.00001389, 0.00004950, -0.00003911, -0.00008414, 0.00009488, 0.00003908, -0.00006189, 0.00002159, -0.00000974, -0.00003555, 0.00003001, -0.00004252, -0.00001694, 0.00011815, -0.00002632, -0.00008976, 0.00004488, -0.00000496, 0.00001544, 0.00002791, -0.00005863, 0.00003323, -0.00002207, -0.00004895, 0.00009257, -0.00005465, -0.00007942, 0.00004333, -0.00003786, 0.00005599, 0.00015106, 0.00006013, 0.00002387, -0.00004663, -0.00006606, -0.00000864, -0.00001521, -0.00001294, -0.00000575, -0.00000870, -0.00001067, -0.00001698, },
		{ -0.00000111, -0.00000739, -0.00002468, -0.00001566, -0.00004749, -0.00008273, -0.00000532, 0.00010443, 0.00018040, 0.00009950, -0.00002867, -0.00001808, -0.00006627, -0.00003488, 0.00006360, -0.00003725, -0.00002611, 0.00002085, -0.00004392, 0.00001766, 0.00001119, 0.00000561, 0.00004517, -0.00008356, -0.00002813, 0.00010779, -0.00001589, -0.00003925, 0.00002953, -0.00003453, -0.00001044, 0.00001939, -0.00006233, 0.00004098, 0.00010102, -0.00008388, -0.00004250, 0.00004491, -0.00002172, 0.00004435, -0.00001006, -0.00004372, 0.00004931, -0.00006510, 0.00000849, 0.00008006, -0.00011580, -0.00001548, 0.00004142, -0.00005242, 0.00010504, 0.00011978, 0.00002865, 0.00002063, -0.00006195, -0.00005077, -0.00000018, -0.00001774, -0.00001029, -0.00000877, -0.00000715, -0.00001177, -0.00001853, -0.00000772, },
		{ -0.00001313, -0.00002784, -0.00001764, -0.00006270, -0.00008462, 0.00002533, 0.00015673, 0.00019121, 0.00005666, -0.00004508, -0.00003783, -0.00006709, 0.00000973, 0.00003757, -0.00005319, 0.00000103, -0.00000315, -0.00003175, 0.00002714, -0.00000408, 0.00003355, 0.00001162, -0.00009757, 0.00003472, 0.00008140, -0.00004864, -0.00000597, 0.00001516, -0.00004304, 0.00001370, -0.00000919, -0.00005216, 0.00009223, 0.00004708, -0.00010588, 0.00000780, 0.00001931, -0.00001130, 0.00005416, -0.00004983, -0.00000328, 0.00003050, -0.00008181, 0.00007462, 0.00001650, -0.00013078, 0.00005716, -0.00000091, -0.00003695, 0.00014251, 0.00006826, 0.00002017, 0.00001326, -0.00007537, -0.00002859, 0.00000113, -0.00001958, -0.00000921, -0.00001070, -0.00000393, -0.00001598, -0.00001814, -0.00000274, -0.00000009, },
		{ -0.00002935, -0.00002086, -0.00008136, -0.00008024, 0.00007416, 0.00020909, 0.00017803, 0.00000818, -0.00006256, -0.00005068, -0.00004500, 0.00003251, 0.00000282, -0.00004560, 0.00001084, -0.00002171, -0.00001053, 0.00001922, -0.00000016, 0.00004894, -0.00003499, -0.00007415, 0.00007701, 0.00003372, -0.00005044, 0.00001784, -0.00000986, -0.00003247, 0.00002243, -0.00003897, -0.00001283, 0.00011307, -0.00002021, -0.00008776, 0.00003659, -0.00001028, 0.00001958, 0.00003342, -0.00006687, 0.00003960, -0.00001642, -0.00005344, 0.00011084, -0.00007179, -0.00008320, 0.00009656, -0.00006068, 0.00001546, 0.00014660, 0.00001667, 0.00003184, -0.00000695, -0.00007789, -0.00000521, -0.00000521, -0.00001873, -0.00001043, -0.00000933, -0.00000205, -0.00002261, -0.00001325, 0.00000179, -0.00000220, -0.00002000, },
		{ -0.00002704, -0.00010403, -0.00006209, 0.00014211, 0.00024957, 0.00013857, -0.00004281, -0.00007455, -0.00004755, -0.00001710, 0.00003117, -0.00002245, -0.00002880, 0.00000545, -0.00002578, 0.00000469, 0.00000704, 0.00001956, 0.00003894, -0.00007071, -0.00002608, 0.00008424, -0.00001024, -0.00002973, 0.00002145, -0.00002978, -0.00001182, 0.00001152, -0.00005170, 0.00003909, 0.00009432, -0.00007258, -0.00004487, 0.00003480, -0.00002187, 0.00004950, -0.00000953, -0.00004894, 0.00005767, -0.00006399, 0.00001241, 0.00008919, -0.00013838, 0.00000825, 0.00007555, -0.00009899, 0.00008711, 0.00010879, -0.00001193, 0.00004814, -0.00003831, -0.00006312, 0.00001026, -0.00001508, -0.00001534, -0.00001351, -0.00000360, -0.00000527, -0.00002793, -0.00000369, 0.00000246, -0.00000634, -0.00002702, -0.00002879, },
		{ -0.00012869, -0.00002056, 0.00022442, 0.00026408, 0.00007373, -0.00008786, -0.00007242, -0.00003180, 0.00000160, 0.00001635, -0.00003254, -0.00001586, -0.00000430, -0.00001871, 0.00000762, 0.00000453, 0.00003986, 0.00000694, -0.00007839, 0.00002269, 0.00006100, -0.00003262, -0.00000444, 0.00000783, -0.00003491, 0.00000505, -0.00001079, -0.00003848, 0.00008041, 0.00004572, -0.00009182, -0.00000206, 0.00001244, -0.00000729, 0.00005665, -0.00005108, -0.00000522, 0.00003771, -0.00007993, 0.00008210, 0.00001249, -0.00014270, 0.00009641, 0.00000093, -0.00008473, 0.00014301, 0.00004306, -0.00000580, 0.00004958, -0.00006794, -0.00003389, 0.00001125, -0.00002173, -0.00001255, -0.00001560, 0.00000454, -0.00001560, -0.00002642, 0.00000673, -0.00000223, -0.00001014, -0.00003433, -0.00002562, -0.00003919, },
		{ 0.00005405, 0.00030792, 0.00023864, -0.00000791, -0.00011434, -0.00005423, -0.00001476, 0.00000675, 0.00000076, -0.00003265, -0.00001068, -0.00000924, -0.00001048, 0.00000352, 0.00001607, 0.00004501, -0.00003036, -0.00005724, 0.00005127, 0.00002565, -0.00003092, 0.00000993, -0.00001096, -0.00002647, 0.00000882, -0.00002956, -0.00000472, 0.00009390, -0.00001049, -0.00007785, 0.00002111, -0.00000977, 0.00002323, 0.00003335, -0.00006721, 0.00003851, -0.00000924, -0.00004852, 0.00011397, -0.00008270, -0.00007584, 0.00013128, -0.00008721, -0.00001586, 0.00015146, -0.00001932, 0.00002651, 0.00002579, -0.00007949, -0.00000273, -0.00000069, -0.00002079, -0.00001408, -0.00001219, 0.00000937, -0.00003022, -0.00001441, 0.00001115, -0.00000959, -0.00001262, -0.00004249, -0.00001898, -0.00006206, -0.00014862, },
		{ 0.00036947, 0.00016605, -0.00008778, -0.00011274, -0.00002810, -0.00000609, 0.00000376, -0.00000962, -0.00003058, -0.00000903, -0.00000827, -0.00000821, 0.00000307, 0.00003402, 0.00002950, -0.00005424, -0.00002155, 0.00005285, -0.00000214, -0.00001591, 0.00000858, -0.00002317, -0.00001375, 0.00000026, -0.00003281, 0.00003364, 0.00007672, -0.00005218, -0.00004586, 0.00002044, -0.00001458, 0.00004864, -0.00000883, -0.00004901, 0.00005619, -0.00005346, 0.00001691, 0.00008413, -0.00014451, 0.00003053, 0.00008997, -0.00013741, 0.00007640, 0.00010483, -0.00004758, 0.00006008, -0.00001617, -0.00006488, 0.00001507, -0.00001577, -0.00001445, -0.00002004, -0.00000073, 0.00000383, -0.00004074, 0.00000549, 0.00000411, -0.00001327, -0.00001672, -0.00004971, -0.00000974, -0.00010119, -0.00014960, 0.00016643, },
		{ 0.00005429, -0.00014214, -0.00008471, -0.00000720, -0.00000688, 0.00000022, -0.00001651, -0.00002944, -0.00000547, -0.00000684, -0.00001068, 0.00001297, 0.00004513, 0.00000107, -0.00005541, 0.00000946, 0.00003584, -0.00001244, -0.00000229, -0.00000246, -0.00002470, -0.00000623, -0.00001160, -0.00001862, 0.00005981, 0.00004008, -0.00006831, -0.00001535, 0.00000660, 0.00000116, 0.00005128, -0.00004663, -0.00000832, 0.00003851, -0.00006517, 0.00007963, 0.00000357, -0.00013587, 0.00011971, -0.00000774, -0.00011737, 0.00014448, 0.00002623, -0.00003060, 0.00006908, -0.00005512, -0.00003157, 0.00001200, -0.00002244, -0.00001050, -0.00002459, 0.00001509, -0.00001500, -0.00003680, 0.00002300, -0.00001290, -0.00000809, -0.00002846, -0.00004974, -0.00000438, -0.00015838, -0.00010965, 0.00030796, 0.00037995, },
		{ -0.00015350, -0.00004477, -0.00000034, -0.00001071, -0.00000120, -0.00002386, -0.00002664, 0.00000057, -0.00001028, -0.00001040, 0.00003049, 0.00004040, -0.00002538, -0.00003838, 0.00002348, 0.00001596, -0.00000836, 0.00000112, -0.00001316, -0.00001987, -0.00000683, -0.00001659, 0.00000480, 0.00006525, 0.00000113, -0.00006200, 0.00000141, -0.00000427, 0.00002592, 0.00002814, -0.00006002, 0.00003034, -0.00000093, -0.00003436, 0.00010115, -0.00008545, -0.00005803, 0.00014194, -0.00010906, -0.00003416, 0.00015264, -0.00004328, 0.00001571, 0.00004401, -0.00006958, 0.00000084, -0.00000567, -0.00001638, -0.00001582, -0.00001897, 0.00002466, -0.00004068, -0.00001374, 0.00002465, -0.00002919, 0.00000353, -0.00005128, -0.00003396, -0.00001777, -0.00022332, -0.00000549, 0.00044966, 0.00031614, -0.00006972, },
		{ -0.00001272, -0.00000582, -0.00001075, -0.00000368, -0.00003195, -0.00001861, 0.00000440, -0.00001803, -0.00000070, 0.00004575, 0.00002110, -0.00003797, -0.00001654, 0.00002081, 0.00000534, -0.00000066, -0.00000475, -0.00001740, -0.00001607, -0.00001086, -0.00001071, 0.00002535, 0.00005263, -0.00002693, -0.00004511, 0.00000443, -0.00000196, 0.00004250, -0.00000800, -0.00004439, 0.00004582, -0.00003467, 0.00002156, 0.00006564, -0.00013306, 0.00004798, 0.00008382, -0.00015820, 0.00006912, 0.00009860, -0.00006942, 0.00005858, -0.00000156, -0.00005302, 0.00001346, -0.00002161, -0.00000477, -0.00002899, 0.00000089, 0.00001603, -0.00005766, 0.00002064, 0.00000334, -0.00002990, 0.00000908, -0.00007864, 0.00000034, -0.00006910, -0.00026657, 0.00017283, 0.00054381, 0.00017663, -0.00016715, -0.00012202, },
		{ -0.00001360, -0.00000635, -0.00001125, -0.00003639, -0.00000604, 0.00000091, -0.00002360, 0.00001822, 0.00004911, -0.00000283, -0.00003471, -0.00000239, 0.00001132, 0.00000604, 0.00000122, -0.00001274, -0.00001629, -0.00001697, -0.00001086, 0.00000183, 0.00003499, 0.00003181, -0.00004035, -0.00002901, 0.00000241, 0.00001197, 0.00003980, -0.00003769, -0.00001209, 0.00003371, -0.00003982, 0.00006801, -0.00000880, -0.00011156, 0.00012403, -0.00002312, -0.00012879, 0.00013874, 0.00001450, -0.00004533, 0.00006989, -0.00004012, -0.00001929, 0.00000160, -0.00002150, -0.00000140, -0.00003849, 0.00002766, -0.00001422, -0.00004939, 0.00004640, -0.00003294, -0.00000735, -0.00000795, -0.00009161, 0.00003955, -0.00016855, -0.00024205, 0.00040474, 0.00053941, -0.00000794, -0.00020466, -0.00006798, -0.00000132, },
		{ -0.00000336, -0.00002323, -0.00003242, 0.00000528, -0.00000996, -0.00001973, 0.00003892, 0.00003793, -0.00002019, -0.00002302, -0.00000005, 0.00000592, 0.00001141, -0.00000483, -0.00001646, -0.00001537, -0.00001977, -0.00000321, 0.00001254, 0.00003378, 0.00001227, -0.00004343, -0.00001814, 0.00000432, 0.00002714, 0.00001935, -0.00004730, 0.00001712, 0.00000779, -0.00001341, 0.00007521, -0.00007994, -0.00003279, 0.00012851, -0.00012060, -0.00003860, 0.00014188, -0.00005345, 0.00000576, 0.00004447, -0.00004940, 0.00000651, -0.00002118, -0.00000465, -0.00001505, -0.00003005, 0.00004350, -0.00005386, -0.00001124, 0.00004194, -0.00006115, 0.00002845, -0.00005428, -0.00006759, 0.00005035, -0.00029920, -0.00010471, 0.00063123, 0.00040816, -0.00018025, -0.00017338, -0.00002158, -0.00000823, -0.00001419, },
		{ -0.00003330, -0.00002011, 0.00000874, -0.00002225, -0.00000430, 0.00005185, 0.00001780, -0.00002563, -0.00001305, -0.00000457, 0.00000909, 0.00001288, -0.00001404, -0.00001519, -0.00001851, -0.00001813, 0.00000870, 0.00001543, 0.00002774, -0.00000231, -0.00004230, -0.00000986, 0.00001258, 0.00003275, -0.00000710, -0.00003644, 0.00002951, -0.00001073, 0.00002573, 0.00003757, -0.00010684, 0.00005838, 0.00006045, -0.00015746, 0.00006225, 0.00008459, -0.00007341, 0.00004727, 0.00000251, -0.00002834, 0.00000594, -0.00003265, 0.00001399, -0.00003986, 0.00000095, 0.00003075, -0.00007795, 0.00004098, 0.00000009, -0.00005569, 0.00004974, -0.00011398, 0.00000101, -0.00000838, -0.00040671, 0.00015999, 0.00076717, 0.00016805, -0.00027930, -0.00009686, -0.00000498, -0.00001858, -0.00000693, -0.00000819, },
		{ -0.00000555, 0.00000231, -0.00002790, 0.00001785, 0.00005163, -0.00000206, -0.00002178, -0.00001053, -0.00000739, 0.00001730, 0.00000657, -0.00002012, -0.00001322, -0.00002404, -0.00000820, 0.00001701, 0.00001132, 0.00002293, -0.00001392, -0.00003947, 0.00000038, 0.00002236, 0.00002512, -0.00002635, -0.00001580, 0.00002531, -0.00000866, 0.00004996, -0.00002226, -0.00007494, 0.00011053, -0.00004076, -0.00011858, 0.00012188, 0.00000531, -0.00004579, 0.00005415, -0.00002581, 0.00000206, -0.00001870, -0.00001896, 0.00001447, -0.00005631, 0.00004125, -0.00001332, -0.00006340, 0.00007518, -0.00006096, -0.00000796, 0.00002612, -0.00014872, 0.00008885, -0.00015816, -0.00040991, 0.00050980, 0.00073551, -0.00011177, -0.00027142, -0.00002043, -0.00001846, -0.00001713, -0.00000066, -0.00002074, -0.00003446, },
		{ -0.00000959, -0.00002245, 0.00003827, 0.00003978, -0.00001497, -0.00001562, -0.00001399, -0.00000321, 0.00002330, -0.00000492, -0.00002066, -0.00001504, -0.00002590, 0.00000728, 0.00001568, 0.00000659, 0.00002046, -0.00002573, -0.00003279, 0.00001350, 0.00002647, 0.00000921, -0.00003209, 0.00000203, 0.00001604, 0.00001024, 0.00004179, -0.00006774, -0.00000481, 0.00009606, -0.00011953, -0.00003123, 0.00011655, -0.00005099, 0.00000063, 0.00002863, -0.00002252, 0.00001405, -0.00004516, 0.00001322, -0.00001158, -0.00004435, 0.00006412, -0.00006853, -0.00000727, 0.00006140, -0.00010255, 0.00005998, -0.00005125, -0.00011727, 0.00013626, -0.00037112, -0.00023630, 0.00084206, 0.00050983, -0.00033398, -0.00017133, 0.00001409, -0.00004069, -0.00000099, -0.00000590, -0.00003299, -0.00002527, 0.00000332, },
		{ -0.00000769, 0.00005038, 0.00002231, -0.00001997, -0.00001260, -0.00001763, 0.00000693, 0.00002204, -0.00001591, -0.00001830, -0.00002067, -0.00001894, 0.00002013, 0.00000545, 0.00000761, 0.00001625, -0.00003729, -0.00001915, 0.00002524, 0.00002155, -0.00000618, -0.00002709, 0.00001143, 0.00001403, 0.00002871, 0.00000587, -0.00007172, 0.00006090, 0.00002648, -0.00013682, 0.00005392, 0.00006147, -0.00006054, 0.00003067, -0.00000371, 0.00000482, -0.00000577, -0.00004722, 0.00003953, -0.00005110, -0.00000052, 0.00004636, -0.00009944, 0.00006417, -0.00000529, -0.00008779, 0.00010092, -0.00015238, -0.00000753, 0.00007422, -0.00055905, 0.00013028, 0.00102456, 0.00014168, -0.00041708, -0.00003827, -0.00000646, -0.00004438, 0.00001695, -0.00002451, -0.00003485, -0.00001196, 0.00000285, -0.00001939, },
		{ 0.00005254, 0.00000491, -0.00002016, -0.00001342, -0.00001660, 0.00001775, 0.00001370, -0.00002259, -0.00001741, -0.00002523, -0.00000390, 0.00002248, -0.00000626, 0.00001524, 0.00000549, -0.00004358, 0.00000076, 0.00002954, 0.00001056, -0.00001507, -0.00001862, 0.00001594, 0.00002216, 0.00002934, -0.00003410, -0.00003366, 0.00008388, -0.00005634, -0.00009127, 0.00009430, -0.00000302, -0.00003255, 0.00002733, -0.00001435, 0.00002864, -0.00004505, -0.00001518, 0.00003505, -0.00007536, 0.00005411, -0.00001237, -0.00007716, 0.00010570, -0.00009354, -0.00000972, 0.00006938, -0.00021459, 0.00013861, -0.00013059, -0.00059791, 0.00061617, 0.00095387, -0.00024809, -0.00033817, 0.00005416, -0.00005811, -0.00001699, 0.00001752, -0.00004579, -0.00002298, -0.00000383, -0.00000368, -0.00002197, 0.00001059, },
		{ -0.00000997, -0.00001863, -0.00001530, -0.00001034, 0.00002433, 0.00000209, -0.00002532, -0.00001958, -0.00002305, 0.00001215, 0.00001272, -0.00001060, 0.00002371, -0.00001205, -0.00003866, 0.00002085, 0.00002369, 0.00000003, -0.00001768, -0.00001141, 0.00002288, 0.00003175, 0.00000791, -0.00005154, 0.00002063, 0.00005317, -0.00010673, -0.00001628, 0.00007921, -0.00003939, 0.00000170, 0.00000183, 0.00000601, 0.00002220, -0.00007304, 0.00003432, -0.00000560, -0.00005956, 0.00008358, -0.00008263, -0.00000249, 0.00008035, -0.00014781, 0.00009407, -0.00004323, -0.00017622, 0.00022974, -0.00043446, -0.00038602, 0.00106490, 0.00061415, -0.00051712, -0.00014931, 0.00005905, -0.00009514, 0.00002795, -0.00000879, -0.00005303, -0.00000493, -0.00000565, -0.00000922, -0.00001778, 0.00002729, 0.00004682, },
		{ -0.00001581, -0.00001572, -0.00000166, 0.00002523, -0.00000928, -0.00002682, -0.00002194, -0.00001274, 0.00002038, -0.00000291, -0.00000382, 0.00002479, -0.00003015, -0.00002108, 0.00003277, 0.00001106, -0.00000532, -0.00001827, -0.00000410, 0.00003502, 0.00002980, -0.00002284, -0.00003502, 0.00005607, -0.00000978, -0.00010229, 0.00004344, 0.00003161, -0.00003588, 0.00001341, -0.00001719, 0.00003976, -0.00001918, -0.00006244, 0.00006758, -0.00006041, -0.00000321, 0.00006054, -0.00011891, 0.00008672, -0.00001210, -0.00012150, 0.00015530, -0.00018922, -0.00002286, 0.00016680, -0.00070914, 0.00008938, 0.00129291, 0.00010242, -0.00056837, 0.00004607, -0.00001638, -0.00008082, 0.00005835, -0.00005139, -0.00003652, 0.00000547, -0.00001322, -0.00000934, -0.00001074, 0.00004084, 0.00003529, -0.00002216, },
		{ -0.00001491, 0.00000671, 0.00002181, -0.00001962, -0.00002836, -0.00001990, 0.00000086, 0.00001652, -0.00001440, 0.00000984, 0.00001427, -0.00003945, 0.00000348, 0.00003145, -0.00000097, -0.00000604, -0.00001980, 0.00000819, 0.00004658, 0.00001022, -0.00004179, 0.00000398, 0.00005089, -0.00006631, -0.00005478, 0.00005983, -0.00001116, -0.00000998, -0.00000316, -0.00000695, 0.00005488, -0.00007197, -0.00001073, 0.00005699, -0.00009196, 0.00006410, -0.00001143, -0.00008854, 0.00013316, -0.00012580, -0.00001218, 0.00011497, -0.00027984, 0.00018342, -0.00009288, -0.00078518, 0.00071558, 0.00117431, -0.00040098, -0.00040076, 0.00014676, -0.00011489, -0.00001378, 0.00004587, -0.00008377, -0.00000309, -0.00000104, -0.00001662, -0.00000621, -0.00000325, 0.00005171, 0.00001793, -0.00003059, -0.00001068, },
		{ 0.00001438, 0.00001515, -0.00002976, -0.00002780, -0.00001181, 0.00000984, 0.00000408, -0.00001486, 0.00002097, -0.00000427, -0.00003380, 0.00002463, 0.00001884, -0.00000635, -0.00000679, -0.00002018, 0.00002750, 0.00004674, -0.00001973, -0.00003443, 0.00003882, 0.00000981, -0.00008547, 0.00000095, 0.00003606, -0.00002351, 0.00000812, -0.00002827, 0.00003075, 0.00002912, -0.00009895, 0.00005475, 0.00000236, -0.00007269, 0.00009848, -0.00009366, 0.00000228, 0.00009559, -0.00018983, 0.00012554, -0.00003221, -0.00023483, 0.00031797, -0.00048546, -0.00053477, 0.00127796, 0.00071270, -0.00070931, -0.00011342, 0.00010855, -0.00016438, 0.00006816, -0.00001182, -0.00008084, 0.00002617, -0.00002062, -0.00000892, -0.00000631, 0.00000693, 0.00005834, -0.00000551, -0.00003179, -0.00000396, -0.00001473, },
		{ 0.00000427, -0.00003892, -0.00002162, -0.00000146, 0.00000917, -0.00000806, -0.00000520, 0.00002177, -0.00002121, -0.00001491, 0.00003332, 0.00000288, -0.00000457, -0.00001146, -0.00001361, 0.00004851, 0.00002849, -0.00004289, -0.00000371, 0.00004556, -0.00004024, -0.00006225, 0.00003114, -0.00000029, -0.00000680, -0.00000053, -0.00003327, 0.00006913, -0.00003155, -0.00007497, 0.00009298, -0.00006535, -0.00000657, 0.00007073, -0.00013283, 0.00010470, -0.00001943, -0.00015125, 0.00020406, -0.00021934, -0.00004094, 0.00025445, -0.00084103, 0.00004528, 0.00154415, 0.00005780, -0.00071763, 0.00014312, -0.00003189, -0.00012346, 0.00011089, -0.00008515, -0.00003798, 0.00002983, -0.00003738, 0.00000655, -0.00001324, 0.00002429, 0.00005504, -0.00003150, -0.00002223, 0.00000015, -0.00001532, 0.00002230, },
		{ -0.00004299, -0.00000922, 0.00000414, 0.00000075, -0.00001207, 0.00000681, 0.00001158, -0.00002705, 0.00000812, 0.00002731, -0.00000769, -0.00000070, -0.00001881, 0.00000393, 0.00006002, -0.00000414, -0.00004362, 0.00003109, 0.00001878, -0.00006854, -0.00001810, 0.00002410, -0.00001901, 0.00001544, -0.00002980, -0.00000386, 0.00007492, -0.00009374, -0.00000629, 0.00007651, -0.00010229, 0.00006917, -0.00001056, -0.00009535, 0.00015257, -0.00015249, -0.00001475, 0.00015495, -0.00033401, 0.00021808, -0.00005422, -0.00094855, 0.00079946, 0.00137363, -0.00055014, -0.00045544, 0.00024454, -0.00017958, -0.00000903, 0.00007986, -0.00012902, 0.00002245, 0.00000236, -0.00003467, 0.00001705, -0.00002188, 0.00004903, 0.00003514, -0.00005080, -0.00000314, -0.00000354, -0.00001324, 0.00002944, -0.00001240, },
		{ 0.00000443, 0.00000035, -0.00000778, -0.00000676, 0.00001247, -0.00000260, -0.00001876, 0.00002419, 0.00001229, -0.00000900, -0.00000095, -0.00002241, 0.00002942, 0.00005223, -0.00003631, -0.00001917, 0.00004681, -0.00002507, -0.00006033, 0.00001543, -0.00000526, -0.00000828, 0.00001751, -0.00005378, 0.00004715, 0.00003295, -0.00011725, 0.00007059, 0.00001152, -0.00008085, 0.00010571, -0.00009929, 0.00000622, 0.00010407, -0.00022141, 0.00014915, -0.00002066, -0.00028259, 0.00038781, -0.00052147, -0.00066183, 0.00145806, 0.00079669, -0.00088672, -0.00007258, 0.00015696, -0.00023795, 0.00011246, -0.00001465, -0.00011197, 0.00006341, -0.00003942, -0.00000874, 0.00001009, -0.00001986, 0.00007219, -0.00000198, -0.00005220, 0.00001699, -0.00001609, -0.00000364, 0.00003101, -0.00003322, -0.00003679, },
		{ -0.00001059, -0.00000879, 0.00000182, 0.00000854, -0.00001107, -0.00000173, 0.00002688, -0.00000226, -0.00000399, -0.00000732, -0.00001530, 0.00005236, 0.00002454, -0.00005079, 0.00001716, 0.00003173, -0.00005886, -0.00002520, 0.00001806, -0.00002879, 0.00001909, -0.00000852, -0.00004698, 0.00008682, -0.00004053, -0.00008193, 0.00011091, -0.00006396, -0.00001000, 0.00007482, -0.00013815, 0.00011465, -0.00002634, -0.00017174, 0.00023867, -0.00023806, -0.00005715, 0.00032209, -0.00093964, 0.00000666, 0.00174948, 0.00001650, -0.00084813, 0.00023685, -0.00004874, -0.00016651, 0.00016559, -0.00012038, -0.00003874, 0.00005728, -0.00006592, 0.00002574, -0.00001490, 0.00000312, 0.00007790, -0.00004562, -0.00003042, 0.00002555, -0.00003108, 0.00001457, 0.00002072, -0.00005116, -0.00001903, 0.00001074, },
		{ -0.00000082, 0.00000527, -0.00000040, -0.00000828, 0.00001393, 0.00001785, -0.00000928, 0.00000061, -0.00001549, 0.00000395, 0.00006038, -0.00001197, -0.00003906, 0.00004370, -0.00000644, -0.00006258, 0.00001087, -0.00000711, -0.00002596, 0.00003717, -0.00004662, -0.00000444, 0.00008420, -0.00010581, -0.00000253, 0.00009031, -0.00010340, 0.00006792, -0.00000982, -0.00009593, 0.00015989, -0.00016902, -0.00001686, 0.00018201, -0.00036764, 0.00023840, -0.00002389, -0.00106638, 0.00086005, 0.00152911, -0.00067443, -0.00049902, 0.00033310, -0.00024092, -0.00000472, 0.00011367, -0.00017404, 0.00004978, 0.00000570, -0.00005495, 0.00004397, -0.00004232, 0.00004446, 0.00005335, -0.00007606, 0.00000732, 0.00001396, -0.00003691, 0.00003489, -0.00000349, -0.00005633, 0.00000403, 0.00000364, -0.00002006, },
		{ -0.00000036, -0.00000553, 0.00000341, 0.00002007, 0.00000472, -0.00000812, -0.00000016, -0.00001784, 0.00002852, 0.00004730, -0.00003998, -0.00000758, 0.00004398, -0.00004547, -0.00003595, 0.00002347, -0.00003784, 0.00000238, 0.00002694, -0.00006857, 0.00005257, 0.00003242, -0.00012395, 0.00007882, 0.00002104, -0.00008189, 0.00010327, -0.00009795, 0.00000870, 0.00010363, -0.00023685, 0.00016070, -0.00001094, -0.00031040, 0.00042821, -0.00054093, -0.00074849, 0.00158428, 0.00085810, -0.00102604, -0.00003653, 0.00019846, -0.00030410, 0.00015251, -0.00001684, -0.00014098, 0.00010083, -0.00005914, -0.00000871, 0.00002863, -0.00004873, 0.00008525, 0.00000028, -0.00007501, 0.00004223, -0.00001471, -0.00002452, 0.00004507, -0.00003492, -0.00004255, 0.00002021, -0.00001475, -0.00001886, 0.00000973, },
		{ -0.00000055, 0.00001561, 0.00001534, -0.00000459, -0.00000362, -0.00000558, -0.00000939, 0.00004656, 0.00001805, -0.00004604, 0.00002559, 0.00001724, -0.00006281, 0.00000228, 0.00000548, -0.00004936, 0.00003595, -0.00000957, -0.00005443, 0.00008942, -0.00004466, -0.00008163, 0.00011813, -0.00005532, -0.00001288, 0.00007159, -0.00013315, 0.00011435, -0.00003206, -0.00017913, 0.00025277, -0.00024209, -0.00006744, 0.00035752, -0.00099312, -0.00001878, 0.00188400, -0.00001337, -0.00094443, 0.00031134, -0.00006244, -0.00020388, 0.00021267, -0.00015108, -0.00003825, 0.00008324, -0.00009426, 0.00004502, -0.00001547, -0.00001941, 0.00009904, -0.00006106, -0.00003829, 0.00005371, -0.00004509, 0.00000460, 0.00003469, -0.00005932, -0.00001366, 0.00001851, -0.00003352, -0.00000452, 0.00001307, -0.00001098, },
		{ 0.00001995, 0.00000515, -0.00000694, -0.00000135, -0.00001004, 0.00000778, 0.00004847, -0.00001340, -0.00002891, 0.00004159, -0.00002114, -0.00004971, 0.00002715, -0.00002948, -0.00003127, 0.00005028, -0.00005058, -0.00000748, 0.00008069, -0.00010592, 0.00000001, 0.00009633, -0.00009400, 0.00005998, -0.00000925, -0.00008956, 0.00015303, -0.00017254, -0.00001805, 0.00019111, -0.00037431, 0.00024187, -0.00000915, -0.00112234, 0.00089154, 0.00162215, -0.00075558, -0.00052893, 0.00039903, -0.00028802, -0.00000284, 0.00014135, -0.00021095, 0.00007472, 0.00000824, -0.00007417, 0.00007001, -0.00006111, 0.00003807, 0.00006881, -0.00010035, 0.00001825, 0.00003256, -0.00005775, 0.00003662, 0.00000380, -0.00006268, 0.00001614, -0.00000271, -0.00003914, 0.00001562, 0.00000355, -0.00001698, 0.00001329, },
		{ -0.00000303, -0.00000494, -0.00000298, -0.00000780, 0.00002513, 0.00003325, -0.00003213, -0.00000022, 0.00003225, -0.00004942, -0.00001603, 0.00002348, -0.00005731, 0.00000639, 0.00003385, -0.00006965, 0.00004690, 0.00002724, -0.00011775, 0.00007805, 0.00003011, -0.00007505, 0.00009085, -0.00008925, 0.00000936, 0.00009353, -0.00023319, 0.00015799, -0.00000477, -0.00031254, 0.00043234, -0.00054329, -0.00078150, 0.00164133, 0.00089084, -0.00110833, -0.00001366, 0.00022799, -0.00035209, 0.00018057, -0.00001804, -0.00016253, 0.00013236, -0.00007673, -0.00000888, 0.00004621, -0.00007485, 0.00009419, 0.00000098, -0.00009459, 0.00006540, -0.00001141, -0.00004204, 0.00005262, -0.00003433, -0.00004106, 0.00002946, -0.00003261, -0.00002461, 0.00002823, -0.00001511, -0.00001024, 0.00002774, 0.00001387, },
		{ -0.00000349, -0.00000519, 0.00000197, 0.00003312, 0.00000945, -0.00003116, 0.00002287, 0.00000452, -0.00005299, 0.00001678, -0.00000537, -0.00005951, 0.00004095, -0.00000447, -0.00005368, 0.00007713, -0.00004362, -0.00007409, 0.00011366, -0.00003986, -0.00001479, 0.00006117, -0.00011787, 0.00010336, -0.00003608, -0.00017187, 0.00024359, -0.00023027, -0.00006931, 0.00035391, -0.00099482, -0.00002595, 0.00193082, -0.00002595, -0.00099482, 0.00035391, -0.00006931, -0.00023027, 0.00024359, -0.00017187, -0.00003608, 0.00010336, -0.00011787, 0.00006117, -0.00001479, -0.00003986, 0.00011366, -0.00007409, -0.00004362, 0.00007713, -0.00005368, -0.00000447, 0.00004095, -0.00005951, -0.00000537, 0.00001678, -0.00005299, 0.00000452, 0.00002287, -0.00003116, 0.00000945, 0.00003312, 0.00000197, -0.00000519, },
		{ -0.00000303, 0.00001387, 0.00002774, -0.00001024, -0.00001511, 0.00002823, -0.00002461, -0.00003261, 0.00002946, -0.00004106, -0.00003433, 0.00005262, -0.00004204, -0.00001141, 0.00006540, -0.00009459, 0.00000098, 0.00009419, -0.00007485, 0.00004621, -0.00000888, -0.00007673, 0.00013236, -0.00016253, -0.00001804, 0.00018057, -0.00035209, 0.00022799, -0.00001366, -0.00110833, 0.00089084, 0.00164133, -0.00078150, -0.00054329, 0.00043234, -0.00031254, -0.00000477, 0.00015799, -0.00023319, 0.00009353, 0.00000936, -0.00008925, 0.00009085, -0.00007505, 0.00003011, 0.00007805, -0.00011775, 0.00002724, 0.00004690, -0.00006965, 0.00003385, 0.00000639, -0.00005731, 0.00002348, -0.00001603, -0.00004942, 0.00003225, -0.00000022, -0.00003213, 0.00003325, 0.00002513, -0.00000780, -0.00000298, -0.00000494, },
		{ 0.00001995, 0.00001329, -0.00001698, 0.00000355, 0.00001562, -0.00003914, -0.00000271, 0.00001614, -0.00006268, 0.00000380, 0.00003662, -0.00005775, 0.00003256, 0.00001825, -0.00010035, 0.00006881, 0.00003807, -0.00006111, 0.00007001, -0.00007417, 0.00000824, 0.00007472, -0.00021095, 0.00014135, -0.00000284, -0.00028802, 0.00039903, -0.00052893, -0.00075558, 0.00162215, 0.00089154, -0.00112234, -0.00000915, 0.00024187, -0.00037431, 0.00019111, -0.00001805, -0.00017254, 0.00015303, -0.00008956, -0.00000925, 0.00005998, -0.00009400, 0.00009633, 0.00000001, -0.00010592, 0.00008069, -0.00000748, -0.00005058, 0.00005028, -0.00003127, -0.00002948, 0.00002715, -0.00004971, -0.00002114, 0.00004159, -0.00002891, -0.00001340, 0.00004847, 0.00000778, -0.00001004, -0.00000135, -0.00000694, 0.00000515, },
		{ -0.00000055, -0.00001098, 0.00001307, -0.00000452, -0.00003352, 0.00001851, -0.00001366, -0.00005932, 0.00003469, 0.00000460, -0.00004509, 0.00005371, -0.00003829, -0.00006106, 0.00009904, -0.00001941, -0.00001547, 0.00004502, -0.00009426, 0.00008324, -0.00003825, -0.00015108, 0.00021267, -0.00020388, -0.00006244, 0.00031134, -0.00094443, -0.00001337, 0.00188400, -0.00001878, -0.00099312, 0.00035752, -0.00006744, -0.00024209, 0.00025277, -0.00017913, -0.00003206, 0.00011435, -0.00013315, 0.00007159, -0.00001288, -0.00005532, 0.00011813, -0.00008163, -0.00004466, 0.00008942, -0.00005443, -0.00000957, 0.00003595, -0.00004936, 0.00000548, 0.00000228, -0.00006281, 0.00001724, 0.00002559, -0.00004604, 0.00001805, 0.00004656, -0.00000939, -0.00000558, -0.00000362, -0.00000459, 0.00001534, 0.00001561, },
		{ -0.00000036, 0.00000973, -0.00001886, -0.00001475, 0.00002021, -0.00004255, -0.00003492, 0.00004507, -0.00002452, -0.00001471, 0.00004223, -0.00007501, 0.00000028, 0.00008525, -0.00004873, 0.00002863, -0.00000871, -0.00005914, 0.00010083, -0.00014098, -0.00001684, 0.00015251, -0.00030410, 0.00019846, -0.00003653, -0.00102604, 0.00085810, 0.00158428, -0.00074849, -0.00054093, 0.00042821, -0.00031040, -0.00001094, 0.00016070, -0.00023685, 0.00010363, 0.00000870, -0.00009795, 0.00010327, -0.00008189, 0.00002104, 0.00007882, -0.00012395, 0.00003242, 0.00005257, -0.00006857, 0.00002694, 0.00000238, -0.00003784, 0.00002347, -0.00003595, -0.00004547, 0.00004398, -0.00000758, -0.00003998, 0.00004730, 0.00002852, -0.00001784, -0.00000016, -0.00000812, 0.00000472, 0.00002007, 0.00000341, -0.00000553, },
		{ -0.00000082, -0.00002006, 0.00000364, 0.00000403, -0.00005633, -0.00000349, 0.00003489, -0.00003691, 0.00001396, 0.00000732, -0.00007606, 0.00005335, 0.00004446, -0.00004232, 0.00004397, -0.00005495, 0.00000570, 0.00004978, -0.00017404, 0.00011367, -0.00000472, -0.00024092, 0.00033310, -0.00049902, -0.00067443, 0.00152911, 0.00086005, -0.00106638, -0.00002389, 0.00023840, -0.00036764, 0.00018201, -0.00001686, -0.00016902, 0.00015989, -0.00009593, -0.00000982, 0.00006792, -0.00010340, 0.00009031, -0.00000253, -0.00010581, 0.00008420, -0.00000444, -0.00004662, 0.00003717, -0.00002596, -0.00000711, 0.00001087, -0.00006258, -0.00000644, 0.00004370, -0.00003906, -0.00001197, 0.00006038, 0.00000395, -0.00001549, 0.00000061, -0.00000928, 0.00001785, 0.00001393, -0.00000828, -0.00000040, 0.00000527, },
		{ -0.00001059, 0.00001074, -0.00001903, -0.00005116, 0.00002072, 0.00001457, -0.00003108, 0.00002555, -0.00003042, -0.00004562, 0.00007790, 0.00000312, -0.00001490, 0.00002574, -0.00006592, 0.00005728, -0.00003874, -0.00012038, 0.00016559, -0.00016651, -0.00004874, 0.00023685, -0.00084813, 0.00001650, 0.00174948, 0.00000666, -0.00093964, 0.00032209, -0.00005715, -0.00023806, 0.00023867, -0.00017174, -0.00002634, 0.00011465, -0.00013815, 0.00007482, -0.00001000, -0.00006396, 0.00011091, -0.00008193, -0.00004053, 0.00008682, -0.00004698, -0.00000852, 0.00001909, -0.00002879, 0.00001806, -0.00002520, -0.00005886, 0.00003173, 0.00001716, -0.00005079, 0.00002454, 0.00005236, -0.00001530, -0.00000732, -0.00000399, -0.00000226, 0.00002688, -0.00000173, -0.00001107, 0.00000854, 0.00000182, -0.00000879, },
		{ 0.00000443, -0.00003679, -0.00003322, 0.00003101, -0.00000364, -0.00001609, 0.00001699, -0.00005220, -0.00000198, 0.00007219, -0.00001986, 0.00001009, -0.00000874, -0.00003942, 0.00006341, -0.00011197, -0.00001465, 0.00011246, -0.00023795, 0.00015696, -0.00007258, -0.00088672, 0.00079669, 0.00145806, -0.00066183, -0.00052147, 0.00038781, -0.00028259, -0.00002066, 0.00014915, -0.00022141, 0.00010407, 0.00000622, -0.00009929, 0.00010571, -0.00008085, 0.00001152, 0.00007059, -0.00011725, 0.00003295, 0.00004715, -0.00005378, 0.00001751, -0.00000828, -0.00000526, 0.00001543, -0.00006033, -0.00002507, 0.00004681, -0.00001917, -0.00003631, 0.00005223, 0.00002942, -0.00002241, -0.00000095, -0.00000900, 0.00001229, 0.00002419, -0.00001876, -0.00000260, 0.00001247, -0.00000676, -0.00000778, 0.00000035, },
		{ -0.00004299, -0.00001240, 0.00002944, -0.00001324, -0.00000354, -0.00000314, -0.00005080, 0.00003514, 0.00004903, -0.00002188, 0.00001705, -0.00003467, 0.00000236, 0.00002245, -0.00012902, 0.00007986, -0.00000903, -0.00017958, 0.00024454, -0.00045544, -0.00055014, 0.00137363, 0.00079946, -0.00094855, -0.00005422, 0.00021808, -0.00033401, 0.00015495, -0.00001475, -0.00015249, 0.00015257, -0.00009535, -0.00001056, 0.00006917, -0.00010229, 0.00007651, -0.00000629, -0.00009374, 0.00007492, -0.00000386, -0.00002980, 0.00001544, -0.00001901, 0.00002410, -0.00001810, -0.00006854, 0.00001878, 0.00003109, -0.00004362, -0.00000414, 0.00006002, 0.00000393, -0.00001881, -0.00000070, -0.00000769, 0.00002731, 0.00000812, -0.00002705, 0.00001158, 0.00000681, -0.00001207, 0.00000075, 0.00000414, -0.00000922, },
		{ 0.00000427, 0.00002230, -0.00001532, 0.00000015, -0.00002223, -0.00003150, 0.00005504, 0.00002429, -0.00001324, 0.00000655, -0.00003738, 0.00002983, -0.00003798, -0.00008515, 0.00011089, -0.00012346, -0.00003189, 0.00014312, -0.00071763, 0.00005780, 0.00154415, 0.00004528, -0.00084103, 0.00025445, -0.00004094, -0.00021934, 0.00020406, -0.00015125, -0.00001943, 0.00010470, -0.00013283, 0.00007073, -0.00000657, -0.00006535, 0.00009298, -0.00007497, -0.00003155, 0.00006913, -0.00003327, -0.00000053, -0.00000680, -0.00000029, 0.00003114, -0.00006225, -0.00004024, 0.00004556, -0.00000371, -0.00004289, 0.00002849, 0.00004851, -0.00001361, -0.00001146, -0.00000457, 0.00000288, 0.00003332, -0.00001491, -0.00002121, 0.00002177, -0.00000520, -0.00000806, 0.00000917, -0.00000146, -0.00002162, -0.00003892, },
		{ 0.00001438, -0.00001473, -0.00000396, -0.00003179, -0.00000551, 0.00005834, 0.00000693, -0.00000631, -0.00000892, -0.00002062, 0.00002617, -0.00008084, -0.00001182, 0.00006816, -0.00016438, 0.00010855, -0.00011342, -0.00070931, 0.00071270, 0.00127796, -0.00053477, -0.00048546, 0.00031797, -0.00023483, -0.00003221, 0.00012554, -0.00018983, 0.00009559, 0.00000228, -0.00009366, 0.00009848, -0.00007269, 0.00000236, 0.00005475, -0.00009895, 0.00002912, 0.00003075, -0.00002827, 0.00000812, -0.00002351, 0.00003606, 0.00000095, -0.00008547, 0.00000981, 0.00003882, -0.00003443, -0.00001973, 0.00004674, 0.00002750, -0.00002018, -0.00000679, -0.00000635, 0.00001884, 0.00002463, -0.00003380, -0.00000427, 0.00002097, -0.00001486, 0.00000408, 0.00000984, -0.00001181, -0.00002780, -0.00002976, 0.00001515, },
		{ -0.00001491, -0.00001068, -0.00003059, 0.00001793, 0.00005171, -0.00000325, -0.00000621, -0.00001662, -0.00000104, -0.00000309, -0.00008377, 0.00004587, -0.00001378, -0.00011489, 0.00014676, -0.00040076, -0.00040098, 0.00117431, 0.00071558, -0.00078518, -0.00009288, 0.00018342, -0.00027984, 0.00011497, -0.00001218, -0.00012580, 0.00013316, -0.00008854, -0.00001143, 0.00006410, -0.00009196, 0.00005699, -0.00001073, -0.00007197, 0.00005488, -0.00000695, -0.00000316, -0.00000998, -0.00001116, 0.00005983, -0.00005478, -0.00006631, 0.00005089, 0.00000398, -0.00004179, 0.00001022, 0.00004658, 0.00000819, -0.00001980, -0.00000604, -0.00000097, 0.00003145, 0.00000348, -0.00003945, 0.00001427, 0.00000984, -0.00001440, 0.00001652, 0.00000086, -0.00001990, -0.00002836, -0.00001962, 0.00002181, 0.00000671, },
		{ -0.00001581, -0.00002216, 0.00003529, 0.00004084, -0.00001074, -0.00000934, -0.00001322, 0.00000547, -0.00003652, -0.00005139, 0.00005835, -0.00008082, -0.00001638, 0.00004607, -0.00056837, 0.00010242, 0.00129291, 0.00008938, -0.00070914, 0.00016680, -0.00002286, -0.00018922, 0.00015530, -0.00012150, -0.00001210, 0.00008672, -0.00011891, 0.00006054, -0.00000321, -0.00006041, 0.00006758, -0.00006244, -0.00001918, 0.00003976, -0.00001719, 0.00001341, -0.00003588, 0.00003161, 0.00004344, -0.00010229, -0.00000978, 0.00005607, -0.00003502, -0.00002284, 0.00002980, 0.00003502, -0.00000410, -0.00001827, -0.00000532, 0.00001106, 0.00003277, -0.00002108, -0.00003015, 0.00002479, -0.00000382, -0.00000291, 0.00002038, -0.00001274, -0.00002194, -0.00002682, -0.00000928, 0.00002523, -0.00000166, -0.00001572, },
		{ -0.00000997, 0.00004682, 0.00002729, -0.00001778, -0.00000922, -0.00000565, -0.00000493, -0.00005303, -0.00000879, 0.00002795, -0.00009514, 0.00005905, -0.00014931, -0.00051712, 0.00061415, 0.00106490, -0.00038602, -0.00043446, 0.00022974, -0.00017622, -0.00004323, 0.00009407, -0.00014781, 0.00008035, -0.00000249, -0.00008263, 0.00008358, -0.00005956, -0.00000560, 0.00003432, -0.00007304, 0.00002220, 0.00000601, 0.00000183, 0.00000170, -0.00003939, 0.00007921, -0.00001628, -0.00010673, 0.00005317, 0.00002063, -0.00005154, 0.00000791, 0.00003175, 0.00002288, -0.00001141, -0.00001768, 0.00000003, 0.00002369, 0.00002085, -0.00003866, -0.00001205, 0.00002371, -0.00001060, 0.00001272, 0.00001215, -0.00002305, -0.00001958, -0.00002532, 0.00000209, 0.00002433, -0.00001034, -0.00001530, -0.00001863, },
		{ 0.00005254, 0.00001059, -0.00002197, -0.00000368, -0.00000383, -0.00002298, -0.00004579, 0.00001752, -0.00001699, -0.00005811, 0.00005416, -0.00033817, -0.00024809, 0.00095387, 0.00061617, -0.00059791, -0.00013059, 0.00013861, -0.00021459, 0.00006938, -0.00000972, -0.00009354, 0.00010570, -0.00007716, -0.00001237, 0.00005411, -0.00007536, 0.00003505, -0.00001518, -0.00004505, 0.00002864, -0.00001435, 0.00002733, -0.00003255, -0.00000302, 0.00009430, -0.00009127, -0.00005634, 0.00008388, -0.00003366, -0.00003410, 0.00002934, 0.00002216, 0.00001594, -0.00001862, -0.00001507, 0.00001056, 0.00002954, 0.00000076, -0.00004358, 0.00000549, 0.00001524, -0.00000626, 0.00002248, -0.00000390, -0.00002523, -0.00001741, -0.00002259, 0.00001370, 0.00001775, -0.00001660, -0.00001342, -0.00002016, 0.00000491, },
		{ -0.00000769, -0.00001939, 0.00000285, -0.00001196, -0.00003485, -0.00002451, 0.00001695, -0.00004438, -0.00000646, -0.00003827, -0.00041708, 0.00014168, 0.00102456, 0.00013028, -0.00055905, 0.00007422, -0.00000753, -0.00015238, 0.00010092, -0.00008779, -0.00000529, 0.00006417, -0.00009944, 0.00004636, -0.00000052, -0.00005110, 0.00003953, -0.00004722, -0.00000577, 0.00000482, -0.00000371, 0.00003067, -0.00006054, 0.00006147, 0.00005392, -0.00013682, 0.00002648, 0.00006090, -0.00007172, 0.00000587, 0.00002871, 0.00001403, 0.00001143, -0.00002709, -0.00000618, 0.00002155, 0.00002524, -0.00001915, -0.00003729, 0.00001625, 0.00000761, 0.00000545, 0.00002013, -0.00001894, -0.00002067, -0.00001830, -0.00001591, 0.00002204, 0.00000693, -0.00001763, -0.00001260, -0.00001997, 0.00002231, 0.00005038, },
		{ -0.00000959, 0.00000332, -0.00002527, -0.00003299, -0.00000590, -0.00000099, -0.00004069, 0.00001409, -0.00017133, -0.00033398, 0.00050983, 0.00084206, -0.00023630, -0.00037112, 0.00013626, -0.00011727, -0.00005125, 0.00005998, -0.00010255, 0.00006140, -0.00000727, -0.00006853, 0.00006412, -0.00004435, -0.00001158, 0.00001322, -0.00004516, 0.00001405, -0.00002252, 0.00002863, 0.00000063, -0.00005099, 0.00011655, -0.00003123, -0.00011953, 0.00009606, -0.00000481, -0.00006774, 0.00004179, 0.00001024, 0.00001604, 0.00000203, -0.00003209, 0.00000921, 0.00002647, 0.00001350, -0.00003279, -0.00002573, 0.00002046, 0.00000659, 0.00001568, 0.00000728, -0.00002590, -0.00001504, -0.00002066, -0.00000492, 0.00002330, -0.00000321, -0.00001399, -0.00001562, -0.00001497, 0.00003978, 0.00003827, -0.00002245, },
		{ -0.00000555, -0.00003446, -0.00002074, -0.00000066, -0.00001713, -0.00001846, -0.00002043, -0.00027142, -0.00011177, 0.00073551, 0.00050980, -0.00040991, -0.00015816, 0.00008885, -0.00014872, 0.00002612, -0.00000796, -0.00006096, 0.00007518, -0.00006340, -0.00001332, 0.00004125, -0.00005631, 0.00001447, -0.00001896, -0.00001870, 0.00000206, -0.00002581, 0.00005415, -0.00004579, 0.00000531, 0.00012188, -0.00011858, -0.00004076, 0.00011053, -0.00007494, -0.00002226, 0.00004996, -0.00000866, 0.00002531, -0.00001580, -0.00002635, 0.00002512, 0.00002236, 0.00000038, -0.00003947, -0.00001392, 0.00002293, 0.00001132, 0.00001701, -0.00000820, -0.00002404, -0.00001322, -0.00002012, 0.00000657, 0.00001730, -0.00000739, -0.00001053, -0.00002178, -0.00000206, 0.00005163, 0.00001785, -0.00002790, 0.00000231, },
		{ -0.00003330, -0.00000819, -0.00000693, -0.00001858, -0.00000498, -0.00009686, -0.00027930, 0.00016805, 0.00076717, 0.00015999, -0.00040671, -0.00000838, 0.00000101, -0.00011398, 0.00004974, -0.00005569, 0.00000009, 0.00004098, -0.00007795, 0.00003075, 0.00000095, -0.00003986, 0.00001399, -0.00003265, 0.00000594, -0.00002834, 0.00000251, 0.00004727, -0.00007341, 0.00008459, 0.00006225, -0.00015746, 0.00006045, 0.00005838, -0.00010684, 0.00003757, 0.00002573, -0.00001073, 0.00002951, -0.00003644, -0.00000710, 0.00003275, 0.00001258, -0.00000986, -0.00004230, -0.00000231, 0.00002774, 0.00001543, 0.00000870, -0.00001813, -0.00001851, -0.00001519, -0.00001404, 0.00001288, 0.00000909, -0.00000457, -0.00001305, -0.00002563, 0.00001780, 0.00005185, -0.00000430, -0.00002225, 0.00000874, -0.00002011, },
		{ -0.00000336, -0.00001419, -0.00000823, -0.00002158, -0.00017338, -0.00018025, 0.00040816, 0.00063123, -0.00010471, -0.00029920, 0.00005035, -0.00006759, -0.00005428, 0.00002845, -0.00006115, 0.00004194, -0.00001124, -0.00005386, 0.00004350, -0.00003005, -0.00001505, -0.00000465, -0.00002118, 0.00000651, -0.00004940, 0.00004447, 0.00000576, -0.00005345, 0.00014188, -0.00003860, -0.00012060, 0.00012851, -0.00003279, -0.00007994, 0.00007521, -0.00001341, 0.00000779, 0.00001712, -0.00004730, 0.00001935, 0.00002714, 0.00000432, -0.00001814, -0.00004343, 0.00001227, 0.00003378, 0.00001254, -0.00000321, -0.00001977, -0.00001537, -0.00001646, -0.00000483, 0.00001141, 0.00000592, -0.00000005, -0.00002302, -0.00002019, 0.00003793, 0.00003892, -0.00001973, -0.00000996, 0.00000528, -0.00003242, -0.00002323, },
		{ -0.00001360, -0.00000132, -0.00006798, -0.00020466, -0.00000794, 0.00053941, 0.00040474, -0.00024205, -0.00016855, 0.00003955, -0.00009161, -0.00000795, -0.00000735, -0.00003294, 0.00004640, -0.00004939, -0.00001422, 0.00002766, -0.00003849, -0.00000140, -0.00002150, 0.00000160, -0.00001929, -0.00004012, 0.00006989, -0.00004533, 0.00001450, 0.00013874, -0.00012879, -0.00002312, 0.00012403, -0.00011156, -0.00000880, 0.00006801, -0.00003982, 0.00003371, -0.00001209, -0.00003769, 0.00003980, 0.00001197, 0.00000241, -0.00002901, -0.00004035, 0.00003181, 0.00003499, 0.00000183, -0.00001086, -0.00001697, -0.00001629, -0.00001274, 0.00000122, 0.00000604, 0.00001132, -0.00000239, -0.00003471, -0.00000283, 0.00004911, 0.00001822, -0.00002360, 0.00000091, -0.00000604, -0.00003639, -0.00001125, -0.00000635, },
		{ -0.00001272, -0.00012202, -0.00016715, 0.00017663, 0.00054381, 0.00017283, -0.00026657, -0.00006910, 0.00000034, -0.00007864, 0.00000908, -0.00002990, 0.00000334, 0.00002064, -0.00005766, 0.00001603, 0.00000089, -0.00002899, -0.00000477, -0.00002161, 0.00001346, -0.00005302, -0.00000156, 0.00005858, -0.00006942, 0.00009860, 0.00006912, -0.00015820, 0.00008382, 0.00004798, -0.00013306, 0.00006564, 0.00002156, -0.00003467, 0.00004582, -0.00004439, -0.00000800, 0.00004250, -0.00000196, 0.00000443, -0.00004511, -0.00002693, 0.00005263, 0.00002535, -0.00001071, -0.00001086, -0.00001607, -0.00001740, -0.00000475, -0.00000066, 0.00000534, 0.00002081, -0.00001654, -0.00003797, 0.00002110, 0.00004575, -0.00000070, -0.00001803, 0.00000440, -0.00001861, -0.00003195, -0.00000368, -0.00001075, -0.00000582, },
		{ -0.00015350, -0.00006972, 0.00031614, 0.00044966, -0.00000549, -0.00022332, -0.00001777, -0.00003396, -0.00005128, 0.00000353, -0.00002919, 0.00002465, -0.00001374, -0.00004068, 0.00002466, -0.00001897, -0.00001582, -0.00001638, -0.00000567, 0.00000084, -0.00006958, 0.00004401, 0.00001571, -0.00004328, 0.00015264, -0.00003416, -0.00010906, 0.00014194, -0.00005803, -0.00008545, 0.00010115, -0.00003436, -0.00000093, 0.00003034, -0.00006002, 0.00002814, 0.00002592, -0.00000427, 0.00000141, -0.00006200, 0.00000113, 0.00006525, 0.00000480, -0.00001659, -0.00000683, -0.00001987, -0.00001316, 0.00000112, -0.00000836, 0.00001596, 0.00002348, -0.00003838, -0.00002538, 0.00004040, 0.00003049, -0.00001040, -0.00001028, 0.00000057, -0.00002664, -0.00002386, -0.00000120, -0.00001071, -0.00000034, -0.00004477, },
		{ 0.00005429, 0.00037995, 0.00030796, -0.00010965, -0.00015838, -0.00000438, -0.00004974, -0.00002846, -0.00000809, -0.00001290, 0.00002300, -0.00003680, -0.00001500, 0.00001509, -0.00002459, -0.00001050, -0.00002244, 0.00001200, -0.00003157, -0.00005512, 0.00006908, -0.00003060, 0.00002623, 0.00014448, -0.00011737, -0.00000774, 0.00011971, -0.00013587, 0.00000357, 0.00007963, -0.00006517, 0.00003851, -0.00000832, -0.00004663, 0.00005128, 0.00000116, 0.00000660, -0.00001535, -0.00006831, 0.00004008, 0.00005981, -0.00001862, -0.00001160, -0.00000623, -0.00002470, -0.00000246, -0.00000229, -0.00001244, 0.00003584, 0.00000946, -0.00005541, 0.00000107, 0.00004513, 0.00001297, -0.00001068, -0.00000684, -0.00000547, -0.00002944, -0.00001651, 0.00000022, -0.00000688, -0.00000720, -0.00008471, -0.00014214, },
		{ 0.00036947, 0.00016643, -0.00014960, -0.00010119, -0.00000974, -0.00004971, -0.00001672, -0.00001327, 0.00000411, 0.00000549, -0.00004074, 0.00000383, -0.00000073, -0.00002004, -0.00001445, -0.00001577, 0.00001507, -0.00006488, -0.00001617, 0.00006008, -0.00004758, 0.00010483, 0.00007640, -0.00013741, 0.00008997, 0.00003053, -0.00014451, 0.00008413, 0.00001691, -0.00005346, 0.00005619, -0.00004901, -0.00000883, 0.00004864, -0.00001458, 0.00002044, -0.00004586, -0.00005218, 0.00007672, 0.00003364, -0.00003281, 0.00000026, -0.00001375, -0.00002317, 0.00000858, -0.00001591, -0.00000214, 0.00005285, -0.00002155, -0.00005424, 0.00002950, 0.00003402, 0.00000307, -0.00000821, -0.00000827, -0.00000903, -0.00003058, -0.00000962, 0.00000376, -0.00000609, -0.00002810, -0.00011274, -0.00008778, 0.00016605, },
		{ 0.00005405, -0.00014862, -0.00006206, -0.00001898, -0.00004249, -0.00001262, -0.00000959, 0.00001115, -0.00001441, -0.00003022, 0.00000937, -0.00001219, -0.00001408, -0.00002079, -0.00000069, -0.00000273, -0.00007949, 0.00002579, 0.00002651, -0.00001932, 0.00015146, -0.00001586, -0.00008721, 0.00013128, -0.00007584, -0.00008270, 0.00011397, -0.00004852, -0.00000924, 0.00003851, -0.00006721, 0.00003335, 0.00002323, -0.00000977, 0.00002111, -0.00007785, -0.00001049, 0.00009390, -0.00000472, -0.00002956, 0.00000882, -0.00002647, -0.00001096, 0.00000993, -0.00003092, 0.00002565, 0.00005127, -0.00005724, -0.00003036, 0.00004501, 0.00001607, 0.00000352, -0.00001048, -0.00000924, -0.00001068, -0.00003265, 0.00000076, 0.00000675, -0.00001476, -0.00005423, -0.00011434, -0.00000791, 0.00023864, 0.00030792, },
		{ -0.00012869, -0.00003919, -0.00002562, -0.00003433, -0.00001014, -0.00000223, 0.00000673, -0.00002642, -0.00001560, 0.00000454, -0.00001560, -0.00001255, -0.00002173, 0.00001125, -0.00003389, -0.00006794, 0.00004958, -0.00000580, 0.00004306, 0.00014301, -0.00008473, 0.00000093, 0.00009641, -0.00014270, 0.00001249, 0.00008210, -0.00007993, 0.00003771, -0.00000522, -0.00005108, 0.00005665, -0.00000729, 0.00001244, -0.00000206, -0.00009182, 0.00004572, 0.00008041, -0.00003848, -0.00001079, 0.00000505, -0.00003491, 0.00000783, -0.00000444, -0.00003262, 0.00006100, 0.00002269, -0.00007839, 0.00000694, 0.00003986, 0.00000453, 0.00000762, -0.00001871, -0.00000430, -0.00001586, -0.00003254, 0.00001635, 0.00000160, -0.00003180, -0.00007242, -0.00008786, 0.00007373, 0.00026408, 0.00022442, -0.00002056, },
		{ -0.00002704, -0.00002879, -0.00002702, -0.00000634, 0.00000246, -0.00000369, -0.00002793, -0.00000527, -0.00000360, -0.00001351, -0.00001534, -0.00001508, 0.00001026, -0.00006312, -0.00003831, 0.00004814, -0.00001193, 0.00010879, 0.00008711, -0.00009899, 0.00007555, 0.00000825, -0.00013838, 0.00008919, 0.00001241, -0.00006399, 0.00005767, -0.00004894, -0.00000953, 0.00004950, -0.00002187, 0.00003480, -0.00004487, -0.00007258, 0.00009432, 0.00003909, -0.00005170, 0.00001152, -0.00001182, -0.00002978, 0.00002145, -0.00002973, -0.00001024, 0.00008424, -0.00002608, -0.00007071, 0.00003894, 0.00001956, 0.00000704, 0.00000469, -0.00002578, 0.00000545, -0.00002880, -0.00002245, 0.00003117, -0.00001710, -0.00004755, -0.00007455, -0.00004281, 0.00013857, 0.00024957, 0.00014211, -0.00006209, -0.00010403, },
		{ -0.00002935, -0.00002000, -0.00000220, 0.00000179, -0.00001325, -0.00002261, -0.00000205, -0.00000933, -0.00001043, -0.00001873, -0.00000521, -0.00000521, -0.00007789, -0.00000695, 0.00003184, 0.00001667, 0.00014660, 0.00001546, -0.00006068, 0.00009656, -0.00008320, -0.00007179, 0.00011084, -0.00005344, -0.00001642, 0.00003960, -0.00006687, 0.00003342, 0.00001958, -0.00001028, 0.00003659, -0.00008776, -0.00002021, 0.00011307, -0.00001283, -0.00003897, 0.00002243, -0.00003247, -0.00000986, 0.00001784, -0.00005044, 0.00003372, 0.00007701, -0.00007415, -0.00003499, 0.00004894, -0.00000016, 0.00001922, -0.00001053, -0.00002171, 0.00001084, -0.00004560, 0.00000282, 0.00003251, -0.00004500, -0.00005068, -0.00006256, 0.00000818, 0.00017803, 0.00020909, 0.00007416, -0.00008024, -0.00008136, -0.00002086, },
		{ -0.00001313, -0.00000009, -0.00000274, -0.00001814, -0.00001598, -0.00000393, -0.00001070, -0.00000921, -0.00001958, 0.00000113, -0.00002859, -0.00007537, 0.00001326, 0.00002017, 0.00006826, 0.00014251, -0.00003695, -0.00000091, 0.00005716, -0.00013078, 0.00001650, 0.00007462, -0.00008181, 0.00003050, -0.00000328, -0.00004983, 0.00005416, -0.00001130, 0.00001931, 0.00000780, -0.00010588, 0.00004708, 0.00009223, -0.00005216, -0.00000919, 0.00001370, -0.00004304, 0.00001516, -0.00000597, -0.00004864, 0.00008140, 0.00003472, -0.00009757, 0.00001162, 0.00003355, -0.00000408, 0.00002714, -0.00003175, -0.00000315, 0.00000103, -0.00005319, 0.00003757, 0.00000973, -0.00006709, -0.00003783, -0.00004508, 0.00005666, 0.00019121, 0.00015673, 0.00002533, -0.00008462, -0.00006270, -0.00001764, -0.00002784, },
		{ -0.00000111, -0.00000772, -0.00001853, -0.00001177, -0.00000715, -0.00000877, -0.00001029, -0.00001774, -0.00000018, -0.00005077, -0.00006195, 0.00002063, 0.00002865, 0.00011978, 0.00010504, -0.00005242, 0.00004142, -0.00001548, -0.00011580, 0.00008006, 0.00000849, -0.00006510, 0.00004931, -0.00004372, -0.00001006, 0.00004435, -0.00002172, 0.00004491, -0.00004250, -0.00008388, 0.00010102, 0.00004098, -0.00006233, 0.00001939, -0.00001044, -0.00003453, 0.00002953, -0.00003925, -0.00001589, 0.00010779, -0.00002813, -0.00008356, 0.00004517, 0.00000561, 0.00001119, 0.00001766, -0.00004392, 0.00002085, -0.00002611, -0.00003725, 0.00006360, -0.00003488, -0.00006627, -0.00001808, -0.00002867, 0.00009950, 0.00018040, 0.00010443, -0.00000532, -0.00008273, -0.00004749, -0.00001566, -0.00002468, -0.00000739, },
		{ -0.00001062, -0.00001698, -0.00001067, -0.00000870, -0.00000575, -0.00001294, -0.00001521, -0.00000864, -0.00006606, -0.00004663, 0.00002387, 0.00006013, 0.00015106, 0.00005599, -0.00003786, 0.00004333, -0.00007942, -0.00005465, 0.00009257, -0.00004895, -0.00002207, 0.00003323, -0.00005863, 0.00002791, 0.00001544, -0.00000496, 0.00004488, -0.00008976, -0.00002632, 0.00011815, -0.00001694, -0.00004252, 0.00003001, -0.00003555, -0.00000974, 0.00002159, -0.00006189, 0.00003908, 0.00009488, -0.00008414, -0.00003911, 0.00004950, -0.00001389, 0.00003370, -0.00001039, -0.00003405, 0.00003251, -0.00005792, 0.00000532, 0.00005934, -0.00008192, -0.00003654, -0.00000738, -0.00001147, 0.00013420, 0.00014936, 0.00006182, -0.00002272, -0.00007908, -0.00003403, -0.00001429, -0.00002061, -0.00000407, -0.00000430, },
		},
		{ { 0.00000497, -0.00000164, -0.00003258, -0.00002490, -0.00006285, -0.00008759, 0.00003644, 0.00017840, 0.00015722, -0.00000433, -0.00002051, -0.00005806, -0.00003437, 0.00004286, -0.00004618, 0.00004152, -0.00000607, -0.00007731, 0.00005450, -0.00000596, -0.00001277, 0.00000337, -0.00001612, -0.00000513, -0.00002911, 0.00002409, 0.00003806, -0.00002319, -0.00003051, 0.00004219, -0.00000980, -0.00003670, 0.00003960, -0.00003670, -0.00000980, 0.00004219, -0.00003051, -0.00002319, 0.00003806, 0.00002409, -0.00002911, -0.00000513, -0.00001612, 0.00000337, -0.00001277, -0.00000596, 0.00005450, -0.00007731, -0.00000607, 0.00004152, -0.00004618, 0.00004286, -0.00003437, -0.00005806, -0.00002051, -0.00000433, 0.00015722, 0.00017840, 0.00003644, -0.00008759, -0.00006285, -0.00002490, -0.00003258, -0.00000164, },
		{ -0.00000659, -0.00003840, -0.00002461, -0.00007679, -0.00008892, 0.00008116, 0.00021528, 0.00013075, -0.00002296, -0.00003706, -0.00006376, -0.00000164, 0.00001606, -0.00003602, 0.00005560, -0.00004444, -0.00004495, 0.00005177, -0.00002167, 0.00000211, -0.00000776, -0.00001240, -0.00001248, -0.00002302, 0.00004075, 0.00002363, -0.00003886, -0.00001332, 0.00004649, -0.00003129, -0.00001592, 0.00003316, -0.00004893, 0.00001331, 0.00002744, -0.00003881, -0.00000302, 0.00004332, 0.00000500, -0.00002716, -0.00000197, -0.00001589, 0.00001006, -0.00002829, 0.00002078, 0.00003759, -0.00009490, 0.00003519, 0.00000865, -0.00003336, 0.00005509, -0.00007165, -0.00003276, -0.00001310, 0.00001588, 0.00017345, 0.00013295, 0.00000701, -0.00008269, -0.00005282, -0.00002364, -0.00002675, 0.00000163, 0.00000503, },
		{ -0.00004258, -0.00002394, -0.00009609, -0.00008081, 0.00013870, 0.00023883, 0.00009635, -0.00004377, -0.00005361, -0.00005139, 0.00001553, -0.00001098, -0.00001074, 0.00004767, -0.00006808, -0.00001009, 0.00003458, -0.00002358, 0.00001026, -0.00001853, -0.00000765, -0.00002150, -0.00000868, 0.00005019, 0.00000283, -0.00004597, 0.00000804, 0.00003915, -0.00004486, 0.00000684, 0.00001602, -0.00004885, 0.00003115, 0.00000681, -0.00003610, 0.00001656, 0.00003909, -0.00001171, -0.00001973, -0.00000318, -0.00001057, 0.00000856, -0.00003687, 0.00005058, 0.00000262, -0.00008909, 0.00006465, -0.00003136, 0.00000124, 0.00004192, -0.00009751, 0.00000444, -0.00001920, 0.00004259, 0.00017519, 0.00008584, -0.00000782, -0.00007895, -0.00004384, -0.00002146, -0.00002184, 0.00000418, 0.00000476, -0.00001374, },
		{ -0.00002529, -0.00012011, -0.00005768, 0.00020452, 0.00024557, 0.00005458, -0.00006633, -0.00006129, -0.00002961, 0.00001457, -0.00002657, 0.00001720, 0.00002299, -0.00007188, 0.00001617, 0.00001207, -0.00001430, 0.00000906, -0.00002511, -0.00000509, -0.00002827, 0.00001133, 0.00004921, -0.00001968, -0.00004273, 0.00002775, 0.00002262, -0.00004704, 0.00002454, -0.00000610, -0.00003687, 0.00003857, -0.00001326, -0.00002351, 0.00003096, 0.00002742, -0.00002250, -0.00001054, -0.00000689, -0.00000159, -0.00000188, -0.00003240, 0.00007266, -0.00004169, -0.00005892, 0.00007042, -0.00006217, 0.00004636, 0.00000171, -0.00009745, 0.00003881, -0.00003524, 0.00007772, 0.00015842, 0.00004557, -0.00001246, -0.00007833, -0.00003322, -0.00002033, -0.00001734, 0.00000644, 0.00000331, -0.00002270, -0.00004369, },
		{ -0.00014615, -0.00001495, 0.00027255, 0.00023221, 0.00000622, -0.00008585, -0.00005590, -0.00000947, 0.00000086, -0.00002589, 0.00003625, -0.00000794, -0.00005828, 0.00002761, -0.00000653, -0.00000004, -0.00000028, -0.00002605, -0.00000662, -0.00002892, 0.00003222, 0.00003722, -0.00003869, -0.00003016, 0.00004075, 0.00000205, -0.00003822, 0.00003220, -0.00002606, -0.00001737, 0.00003399, -0.00002659, -0.00000496, 0.00003733, 0.00001191, -0.00002597, -0.00000307, -0.00001014, 0.00000752, -0.00001821, -0.00001300, 0.00007734, -0.00008188, -0.00001238, 0.00004862, -0.00006900, 0.00008392, -0.00005548, -0.00006530, 0.00005434, -0.00005035, 0.00011713, 0.00012266, 0.00001983, -0.00001408, -0.00007914, -0.00002028, -0.00002211, -0.00001166, 0.00000767, 0.00000023, -0.00003249, -0.00004114, -0.00003152, },
		{ 0.00005076, 0.00033489, 0.00019564, -0.00004519, -0.00009575, -0.00003969, 0.00000077, -0.00001596, -0.00001167, 0.00004008, -0.00003407, -0.00003518, 0.00002429, -0.00001518, 0.00001224, -0.00001382, -0.00002247, -0.00001189, -0.00002130, 0.00004840, 0.00001655, -0.00004990, -0.00001173, 0.00004428, -0.00001665, -0.00002226, 0.00002848, -0.00003784, 0.00000295, 0.00001995, -0.00002930, 0.00001420, 0.00003515, -0.00000345, -0.00002290, 0.00000064, -0.00001038, 0.00001254, -0.00003433, 0.00001763, 0.00005991, -0.00010444, 0.00003569, 0.00000587, -0.00004512, 0.00009576, -0.00011028, -0.00000728, 0.00004095, -0.00005084, 0.00015052, 0.00007341, 0.00001232, -0.00001997, -0.00007682, -0.00000739, -0.00002682, -0.00000348, 0.00000625, -0.00000385, -0.00004227, -0.00003506, -0.00004539, -0.00016972, },
		{ 0.00038108, 0.00013467, -0.00009236, -0.00009149, -0.00001980, -0.00000123, -0.00002685, 0.00000828, 0.00002915, -0.00004801, -0.00001212, 0.00001137, -0.00001258, 0.00001740, -0.00002692, -0.00001704, -0.00001841, -0.00000598, 0.00005519, -0.00000816, -0.00005112, 0.00000780, 0.00003855, -0.00002879, -0.00000496, 0.00001601, -0.00003866, 0.00001750, 0.00000210, -0.00002103, 0.00002890, 0.00002611, -0.00001535, -0.00001564, 0.00000042, -0.00000664, 0.00001042, -0.00004317, 0.00005095, 0.00002303, -0.00010092, 0.00006880, -0.00004227, 0.00000429, 0.00007134, -0.00014062, 0.00005867, -0.00000007, -0.00002675, 0.00016499, 0.00002215, 0.00002036, -0.00003417, -0.00006648, 0.00000076, -0.00003165, 0.00000664, 0.00000104, -0.00000766, -0.00005180, -0.00002585, -0.00006955, -0.00018433, 0.00014096, },
		{ 0.00005279, -0.00012562, -0.00007367, -0.00000438, -0.00001191, -0.00002710, 0.00002519, 0.00000924, -0.00004799, 0.00000343, -0.00000371, -0.00000177, 0.00001359, -0.00003602, -0.00001253, -0.00002265, 0.00001372, 0.00005036, -0.00003133, -0.00004279, 0.00002385, 0.00002640, -0.00003237, 0.00000821, 0.00000008, -0.00002978, 0.00002221, -0.00001286, -0.00000488, 0.00003586, 0.00001328, -0.00002193, -0.00000711, -0.00000231, 0.00000004, 0.00000078, -0.00003949, 0.00007633, -0.00002364, -0.00007149, 0.00007500, -0.00007673, 0.00006286, 0.00001325, -0.00013055, 0.00010873, -0.00005505, 0.00002238, 0.00015078, -0.00001701, 0.00003501, -0.00005503, -0.00004617, -0.00000043, -0.00003181, 0.00001540, -0.00000710, -0.00001049, -0.00006098, -0.00001423, -0.00010646, -0.00018084, 0.00025358, 0.00039837, },
		{ -0.00013669, -0.00004845, 0.00000138, -0.00002447, -0.00001749, 0.00003307, -0.00001179, -0.00003737, 0.00000830, -0.00001432, 0.00001181, 0.00000217, -0.00003969, -0.00001041, -0.00002161, 0.00003273, 0.00003479, -0.00004801, -0.00002766, 0.00003338, 0.00001205, -0.00002843, 0.00001397, -0.00001342, -0.00001579, 0.00001689, -0.00001951, 0.00001384, 0.00003434, -0.00000002, -0.00002290, 0.00000012, -0.00000541, 0.00000708, -0.00001367, -0.00002203, 0.00008480, -0.00006649, -0.00002541, 0.00005188, -0.00008224, 0.00010837, -0.00006215, -0.00007748, 0.00012224, -0.00010222, 0.00008547, 0.00010670, -0.00003268, 0.00004424, -0.00007527, -0.00001918, -0.00001229, -0.00002294, 0.00001774, -0.00001465, -0.00001367, -0.00006854, -0.00000222, -0.00015745, -0.00014746, 0.00038043, 0.00037397, -0.00003977, },
		{ -0.00002506, -0.00000249, -0.00003270, -0.00000275, 0.00003053, -0.00002744, -0.00002225, 0.00000381, -0.00001684, 0.00002267, -0.00001354, -0.00003845, -0.00001037, -0.00001418, 0.00004595, 0.00001214, -0.00005521, -0.00000982, 0.00003555, -0.00000046, -0.00002028, 0.00001219, -0.00002029, -0.00000265, 0.00000510, -0.00001557, 0.00002943, 0.00002598, -0.00001109, -0.00001921, 0.00000451, -0.00000700, 0.00001148, -0.00002798, 0.00000559, 0.00007247, -0.00009270, 0.00002201, 0.00000800, -0.00005374, 0.00012113, -0.00012954, 0.00000518, 0.00009034, -0.00012002, 0.00014289, 0.00004242, -0.00002134, 0.00003806, -0.00008496, 0.00000618, -0.00003079, -0.00000433, 0.00000945, -0.00001654, -0.00002093, -0.00007093, 0.00000523, -0.00022033, -0.00007127, 0.00050500, 0.00029958, -0.00012637, -0.00012291, },
		{ -0.00001143, -0.00003383, 0.00001143, 0.00002017, -0.00003463, -0.00000855, -0.00000586, -0.00001115, 0.00002704, -0.00002978, -0.00003384, -0.00001073, -0.00000168, 0.00004992, -0.00001234, -0.00005253, 0.00000663, 0.00003170, -0.00000860, -0.00001197, 0.00000557, -0.00001942, 0.00000465, -0.00000733, -0.00000252, 0.00003773, 0.00001390, -0.00001842, -0.00001256, 0.00000569, -0.00000616, 0.00001118, -0.00003674, 0.00003598, 0.00004211, -0.00009486, 0.00005495, -0.00003991, 0.00000083, 0.00009199, -0.00016365, 0.00009159, 0.00002059, -0.00009611, 0.00017361, -0.00002376, 0.00001022, 0.00001348, -0.00007645, 0.00001961, -0.00004694, 0.00001906, -0.00000957, -0.00000909, -0.00003683, -0.00006244, -0.00000093, -0.00028618, 0.00005754, 0.00060252, 0.00017665, -0.00018778, -0.00008973, -0.00001151, },
		{ -0.00002910, 0.00002122, 0.00000641, -0.00003384, 0.00000001, -0.00001587, 0.00000018, 0.00002354, -0.00004359, -0.00002735, -0.00000956, 0.00001265, 0.00004383, -0.00003344, -0.00004192, 0.00001871, 0.00002447, -0.00001193, -0.00000665, -0.00000176, -0.00001294, 0.00000395, -0.00001455, 0.00001514, 0.00003720, 0.00000145, -0.00002164, -0.00000473, 0.00000422, -0.00000317, 0.00000588, -0.00003616, 0.00006044, 0.00000231, -0.00007370, 0.00006296, -0.00007303, 0.00006279, 0.00002689, -0.00014898, 0.00015224, -0.00006461, -0.00003324, 0.00016384, -0.00007136, 0.00004688, -0.00002324, -0.00004906, 0.00001386, -0.00005017, 0.00003731, -0.00003337, 0.00000670, -0.00006345, -0.00003781, -0.00003287, -0.00033691, 0.00023964, 0.00064397, 0.00002013, -0.00020848, -0.00004988, -0.00001061, -0.00001933, },
		{ 0.00002583, -0.00000693, -0.00002765, 0.00000249, -0.00002255, 0.00001346, 0.00001280, -0.00005316, -0.00001975, -0.00000584, 0.00002495, 0.00002956, -0.00004742, -0.00002679, 0.00002520, 0.00001673, -0.00001189, -0.00000545, -0.00000606, -0.00000508, -0.00000334, -0.00001286, 0.00003162, 0.00002910, -0.00000885, -0.00002121, 0.00000271, 0.00000115, 0.00000090, -0.00000279, -0.00002548, 0.00007202, -0.00003552, -0.00003773, 0.00004492, -0.00007776, 0.00010929, -0.00005426, -0.00008647, 0.00016479, -0.00013564, 0.00005057, 0.00011332, -0.00008568, 0.00007104, -0.00005848, -0.00001100, -0.00001072, -0.00003360, 0.00003928, -0.00005094, 0.00002332, -0.00009686, 0.00000345, -0.00010175, -0.00034581, 0.00046041, 0.00060390, -0.00014195, -0.00018333, -0.00001885, -0.00001854, -0.00002217, -0.00002212, },
		{ -0.00001796, -0.00001895, -0.00000005, -0.00002410, 0.00002515, -0.00000337, -0.00005757, -0.00001116, -0.00000007, 0.00003214, 0.00001086, -0.00005279, -0.00001084, 0.00002660, 0.00001054, -0.00001083, -0.00000733, -0.00000551, -0.00000031, -0.00001302, -0.00000198, 0.00004190, 0.00001662, -0.00001582, -0.00001788, 0.00000859, -0.00000241, 0.00000468, -0.00001200, -0.00000723, 0.00006790, -0.00006120, 0.00000002, 0.00000943, -0.00005106, 0.00012190, -0.00012422, 0.00000578, 0.00012279, -0.00016651, 0.00012805, 0.00003647, -0.00006411, 0.00007014, -0.00007693, 0.00002287, -0.00004449, 0.00000123, 0.00001828, -0.00005046, 0.00002819, -0.00012548, 0.00005253, -0.00021086, -0.00028283, 0.00068714, 0.00047041, -0.00027420, -0.00012131, -0.00000848, -0.00002677, -0.00002034, -0.00001597, 0.00002670, },
		{ -0.00000959, -0.00000586, -0.00002036, 0.00003263, -0.00002274, -0.00005624, -0.00000174, 0.00000601, 0.00003302, -0.00000812, -0.00005022, 0.00000305, 0.00002452, 0.00000658, -0.00001085, -0.00000977, -0.00000081, -0.00000140, -0.00001980, 0.00001501, 0.00004329, 0.00000353, -0.00001953, -0.00001239, 0.00001209, -0.00000549, 0.00000718, -0.00001880, 0.00001371, 0.00005016, -0.00006926, 0.00002693, -0.00002856, -0.00000194, 0.00009393, -0.00015820, 0.00009825, 0.00003876, -0.00014477, 0.00017278, -0.00004242, -0.00001759, 0.00004251, -0.00006884, 0.00003697, -0.00007121, 0.00004289, -0.00002298, -0.00002554, 0.00000881, -0.00013241, 0.00008948, -0.00034841, -0.00012448, 0.00087212, 0.00025343, -0.00034415, -0.00004430, -0.00002142, -0.00002666, -0.00001820, -0.00001113, 0.00002553, -0.00002650, },
		{ -0.00001347, -0.00001200, 0.00003408, -0.00004288, -0.00004872, 0.00000765, 0.00001034, 0.00002837, -0.00002407, -0.00004182, 0.00001328, 0.00002087, 0.00000429, -0.00001286, -0.00001012, 0.00000524, -0.00000835, -0.00001939, 0.00003289, 0.00003609, -0.00000717, -0.00002075, -0.00000536, 0.00001276, -0.00000735, 0.00000800, -0.00002125, 0.00003195, 0.00002475, -0.00006048, 0.00003548, -0.00005359, 0.00005180, 0.00003335, -0.00014316, 0.00016035, -0.00005985, -0.00007657, 0.00016923, -0.00009825, 0.00003348, -0.00000124, -0.00003545, 0.00002243, -0.00007458, 0.00007313, -0.00007090, 0.00001944, -0.00004016, -0.00010213, 0.00008658, -0.00048339, 0.00013416, 0.00096281, -0.00001247, -0.00033352, 0.00001971, -0.00004921, -0.00001443, -0.00002115, -0.00000516, 0.00002246, -0.00003277, -0.00000043, },
		{ -0.00000006, 0.00002828, -0.00006099, -0.00003505, 0.00001519, 0.00001170, 0.00002033, -0.00003526, -0.00003014, 0.00001956, 0.00001714, 0.00000240, -0.00001630, -0.00000678, 0.00000909, -0.00001842, -0.00001016, 0.00004625, 0.00002311, -0.00001409, -0.00002023, 0.00000252, 0.00001049, -0.00000745, 0.00000732, -0.00001893, 0.00004344, -0.00000089, -0.00004110, 0.00002574, -0.00005570, 0.00009026, -0.00003973, -0.00008298, 0.00017140, -0.00014038, 0.00001542, 0.00011884, -0.00011475, 0.00006756, -0.00004233, 0.00001007, -0.00001754, -0.00004614, 0.00007415, -0.00010465, 0.00006791, -0.00011083, -0.00002952, 0.00001737, -0.00056822, 0.00046978, 0.00091723, -0.00027351, -0.00024657, 0.00004673, -0.00007536, 0.00000599, -0.00003177, 0.00000534, 0.00001568, -0.00003596, 0.00000784, -0.00002179, },
		{ 0.00001457, -0.00007379, -0.00001640, 0.00001859, 0.00001036, 0.00001117, -0.00004150, -0.00001734, 0.00002243, 0.00001399, -0.00000030, -0.00001953, 0.00000002, 0.00000798, -0.00002722, 0.00000637, 0.00005128, 0.00000837, -0.00001745, -0.00001827, 0.00001033, 0.00000565, -0.00000547, 0.00000553, -0.00001289, 0.00004666, -0.00002059, -0.00001987, 0.00000487, -0.00003425, 0.00009909, -0.00010036, 0.00000266, 0.00012814, -0.00017668, 0.00010026, 0.00003949, -0.00009072, 0.00007100, -0.00006189, 0.00004724, -0.00006713, 0.00000906, 0.00003730, -0.00010438, 0.00009558, -0.00018081, 0.00007244, -0.00013052, -0.00054949, 0.00082836, 0.00072000, -0.00047005, -0.00011119, 0.00002651, -0.00008245, 0.00002397, -0.00004708, 0.00002176, 0.00000280, -0.00003399, 0.00001372, -0.00002952, 0.00001407, },
		{ -0.00007782, 0.00000436, 0.00001628, 0.00000800, 0.00000236, -0.00004346, -0.00000478, 0.00002266, 0.00001137, -0.00000433, -0.00002059, 0.00000836, 0.00000110, -0.00003042, 0.00002604, 0.00004693, -0.00000442, -0.00001851, -0.00001466, 0.00001682, -0.00000076, -0.00000138, 0.00000295, -0.00000493, 0.00004267, -0.00003141, -0.00000446, -0.00001622, 0.00000210, 0.00007491, -0.00012787, 0.00008529, 0.00004588, -0.00015850, 0.00015021, -0.00004140, -0.00004023, 0.00004395, -0.00004969, 0.00005686, -0.00010323, 0.00007214, -0.00003074, -0.00006091, 0.00007986, -0.00021870, 0.00017067, -0.00034354, -0.00038525, 0.00113317, 0.00039297, -0.00055443, 0.00002842, -0.00003248, -0.00006041, 0.00002698, -0.00005880, 0.00004112, -0.00001681, -0.00002449, 0.00001474, -0.00003457, 0.00002812, -0.00000662, },
		{ 0.00002281, 0.00000843, 0.00000678, -0.00000589, -0.00004188, 0.00000675, 0.00002085, 0.00000888, -0.00000933, -0.00001818, 0.00001536, -0.00000999, -0.00002541, 0.00004365, 0.00003496, -0.00001297, -0.00001868, -0.00000900, 0.00002063, -0.00000721, 0.00000433, -0.00000031, 0.00000326, 0.00003420, -0.00003410, 0.00000144, -0.00002754, 0.00003893, 0.00002657, -0.00011347, 0.00013786, -0.00004728, -0.00009473, 0.00015085, -0.00009750, 0.00001350, 0.00000037, -0.00000961, 0.00003003, -0.00010497, 0.00011642, -0.00010770, 0.00001728, 0.00001127, -0.00019586, 0.00021872, -0.00057686, -0.00006366, 0.00130291, -0.00000450, -0.00050724, 0.00012592, -0.00010516, -0.00001228, 0.00000712, -0.00005681, 0.00005603, -0.00003998, -0.00000688, 0.00000845, -0.00003416, 0.00003863, -0.00003298, -0.00007055, },
		{ -0.00000251, 0.00000804, -0.00001425, -0.00003697, 0.00001664, 0.00001731, 0.00000621, -0.00001423, -0.00001216, 0.00001837, -0.00002193, -0.00001218, 0.00005462, 0.00001901, -0.00001692, -0.00001876, -0.00000126, 0.00002068, -0.00001189, 0.00001062, -0.00000422, 0.00001072, 0.00002417, -0.00003198, -0.00000088, -0.00002411, 0.00006197, -0.00002836, -0.00006333, 0.00014439, -0.00012078, -0.00000959, 0.00010579, -0.00011347, 0.00004787, -0.00003800, 0.00004094, -0.00002641, -0.00006352, 0.00011808, -0.00016254, 0.00010375, -0.00009845, -0.00010057, 0.00017164, -0.00075983, 0.00038376, 0.00127567, -0.00038786, -0.00034589, 0.00014891, -0.00015905, 0.00004550, -0.00003365, -0.00003443, 0.00005744, -0.00005965, 0.00001596, -0.00000586, -0.00002589, 0.00004138, -0.00005995, -0.00005167, 0.00003410, },
		{ 0.00001137, -0.00002328, -0.00002852, 0.00002403, 0.00001236, 0.00000338, -0.00001780, -0.00000370, 0.00001597, -0.00003079, 0.00000665, 0.00005646, 0.00000325, -0.00001748, -0.00001860, 0.00000790, 0.00001668, -0.00001319, 0.00001591, -0.00000847, 0.00001727, 0.00001444, -0.00002890, -0.00000642, -0.00000778, 0.00006270, -0.00007039, 0.00000401, 0.00010535, -0.00015255, 0.00006689, 0.00003446, -0.00009025, 0.00005034, -0.00005132, 0.00007777, -0.00009100, 0.00001198, 0.00006725, -0.00016750, 0.00016189, -0.00021579, 0.00005122, 0.00000525, -0.00081315, 0.00088094, 0.00103119, -0.00066965, -0.00012150, 0.00009113, -0.00016709, 0.00008890, -0.00008240, 0.00000694, 0.00003903, -0.00006709, 0.00003753, -0.00002566, -0.00000934, 0.00003274, -0.00008122, -0.00002414, 0.00003470, -0.00001240, },
		{ -0.00003254, -0.00001640, 0.00002778, 0.00000658, 0.00000073, -0.00001917, 0.00000522, 0.00000850, -0.00003338, 0.00002681, 0.00004930, -0.00000901, -0.00001660, -0.00001730, 0.00001706, 0.00000939, -0.00001029, 0.00001851, -0.00001226, 0.00002312, 0.00000534, -0.00002723, -0.00000914, 0.00001387, 0.00004157, -0.00008549, 0.00006483, 0.00003674, -0.00013642, 0.00011003, -0.00003610, -0.00004365, 0.00002345, -0.00003081, 0.00008030, -0.00013544, 0.00009585, -0.00002536, -0.00011083, 0.00015888, -0.00029460, 0.00021418, -0.00026754, -0.00067396, 0.00132099, 0.00060316, -0.00078532, 0.00009607, -0.00002448, -0.00011928, 0.00009618, -0.00011925, 0.00005642, 0.00000144, -0.00005574, 0.00004957, -0.00004510, 0.00001255, 0.00001162, -0.00009038, 0.00000585, 0.00002408, -0.00001690, 0.00001451, },
		{ -0.00000133, 0.00002679, 0.00000102, -0.00000132, -0.00001808, 0.00001250, -0.00000212, -0.00002829, 0.00004370, 0.00003553, -0.00001619, -0.00001598, -0.00001379, 0.00002440, 0.00000054, -0.00000353, 0.00001713, -0.00001429, 0.00002823, -0.00000387, -0.00002679, -0.00000512, 0.00003103, 0.00000750, -0.00007017, 0.00009905, -0.00003642, -0.00008335, 0.00010916, -0.00008199, 0.00000304, -0.00001670, 0.00001667, 0.00004115, -0.00013616, 0.00015461, -0.00012976, -0.00000460, 0.00008064, -0.00029286, 0.00032391, -0.00059043, -0.00032109, 0.00159280, 0.00007591, -0.00071389, 0.00024026, -0.00015335, -0.00002798, 0.00005790, -0.00012526, 0.00009657, -0.00004620, -0.00002504, 0.00004547, -0.00005665, 0.00003346, -0.00001904, -0.00008338, 0.00003043, 0.00000554, -0.00001344, 0.00001427, -0.00004020, },
		{ 0.00002066, -0.00000295, -0.00000264, -0.00001494, 0.00001662, -0.00001306, -0.00001626, 0.00005382, 0.00001887, -0.00001852, -0.00001642, -0.00000743, 0.00002826, -0.00000758, 0.00000555, 0.00001148, -0.00001316, 0.00003178, -0.00001401, -0.00002526, 0.00000559, 0.00003597, -0.00002594, -0.00003229, 0.00009737, -0.00008971, -0.00001618, 0.00007043, -0.00009130, 0.00002913, -0.00004785, 0.00007079, -0.00002924, -0.00008513, 0.00016016, -0.00020654, 0.00011636, -0.00005843, -0.00019141, 0.00031755, -0.00087240, 0.00020886, 0.00161341, -0.00043538, -0.00048595, 0.00026976, -0.00024531, 0.00007552, -0.00001756, -0.00009108, 0.00010995, -0.00008837, 0.00001787, 0.00002373, -0.00005424, 0.00004549, -0.00005216, -0.00006060, 0.00004257, -0.00001456, -0.00000254, 0.00000810, -0.00004341, 0.00001468, },
		{ -0.00000408, -0.00000351, -0.00001052, 0.00001702, -0.00002150, 0.00000020, 0.00005556, 0.00000307, -0.00001760, -0.00001762, 0.00000156, 0.00002774, -0.00001284, 0.00001453, 0.00000254, -0.00000797, 0.00003227, -0.00002479, -0.00001966, 0.00001894, 0.00002632, -0.00004617, 0.00001250, 0.00006399, -0.00010809, 0.00003966, 0.00001307, -0.00006769, 0.00002514, -0.00005164, 0.00010601, -0.00010544, 0.00000493, 0.00010221, -0.00022188, 0.00020551, -0.00021615, -0.00000703, 0.00015864, -0.00101091, 0.00082184, 0.00135573, -0.00081575, -0.00017538, 0.00018222, -0.00026316, 0.00015243, -0.00010541, -0.00002238, 0.00008676, -0.00010872, 0.00005971, -0.00001035, -0.00003643, 0.00004224, -0.00007822, -0.00002766, 0.00003887, -0.00002886, 0.00001194, -0.00000440, -0.00003953, 0.00002843, 0.00001026, },
		{ -0.00000453, -0.00000559, 0.00001409, -0.00002543, 0.00001767, 0.00004940, -0.00000911, -0.00001552, -0.00001842, 0.00001205, 0.00002299, -0.00001394, 0.00002091, -0.00000753, 0.00000106, 0.00002806, -0.00003436, -0.00000833, 0.00002872, 0.00000604, -0.00004647, 0.00004719, 0.00001409, -0.00009061, 0.00006579, -0.00003913, -0.00002749, -0.00000376, -0.00002210, 0.00010260, -0.00015654, 0.00010374, -0.00000541, -0.00016208, 0.00022188, -0.00033488, 0.00020646, -0.00014297, -0.00092381, 0.00138569, 0.00086170, -0.00098357, 0.00012319, 0.00001467, -0.00019691, 0.00017146, -0.00017309, 0.00006048, 0.00003011, -0.00009683, 0.00008564, -0.00004578, -0.00000798, 0.00002198, -0.00008859, 0.00000585, 0.00002124, -0.00003185, 0.00002413, -0.00002074, -0.00002747, 0.00003640, -0.00000216, -0.00000183, },
		{ -0.00000064, 0.00000889, -0.00002408, 0.00003280, 0.00003745, -0.00001640, -0.00001404, -0.00001744, 0.00002234, 0.00001520, -0.00001079, 0.00002285, -0.00001600, 0.00001227, 0.00001835, -0.00003972, 0.00000768, 0.00002954, -0.00001642, -0.00002843, 0.00005999, -0.00003270, -0.00004952, 0.00005705, -0.00006740, 0.00000811, -0.00004061, 0.00003123, 0.00005564, -0.00015862, 0.00017413, -0.00012790, -0.00004146, 0.00014752, -0.00036217, 0.00037171, -0.00052453, -0.00057974, 0.00176414, 0.00023534, -0.00091457, 0.00032390, -0.00016985, -0.00006791, 0.00012166, -0.00019271, 0.00012919, -0.00004357, -0.00005326, 0.00008529, -0.00006987, 0.00002111, -0.00001057, -0.00007912, 0.00002951, -0.00000334, -0.00002218, 0.00002841, -0.00003611, -0.00000870, 0.00003599, -0.00001334, 0.00000325, -0.00000632, },
		{ 0.00000267, -0.00001788, 0.00004308, 0.00002266, -0.00001904, -0.00001410, -0.00001356, 0.00003062, 0.00000612, -0.00000451, 0.00001980, -0.00002039, 0.00002287, 0.00000394, -0.00003791, 0.00002452, 0.00001936, -0.00003174, -0.00000087, 0.00004852, -0.00006008, -0.00000440, 0.00002216, -0.00006483, 0.00002239, -0.00006466, 0.00008694, -0.00002177, -0.00010526, 0.00018646, -0.00022133, 0.00010042, -0.00000237, -0.00027250, 0.00041274, -0.00088181, -0.00001609, 0.00185581, -0.00038371, -0.00064869, 0.00037625, -0.00030473, 0.00007874, 0.00001768, -0.00015252, 0.00015802, -0.00010971, 0.00000945, 0.00005740, -0.00007345, 0.00003926, -0.00004512, -0.00005250, 0.00003603, -0.00002517, -0.00000350, 0.00002188, -0.00004503, 0.00001278, 0.00002666, -0.00002001, 0.00000947, -0.00000905, 0.00000411, },
		{ -0.00000813, 0.00004727, 0.00000796, -0.00001835, -0.00001559, -0.00000637, 0.00003548, -0.00000239, 0.00000302, 0.00001266, -0.00001938, 0.00002976, -0.00001251, -0.00002756, 0.00003679, 0.00000075, -0.00003373, 0.00002417, 0.00002003, -0.00006100, 0.00002600, -0.00002109, -0.00003792, 0.00000989, -0.00006072, 0.00012058, -0.00010280, -0.00001132, 0.00013095, -0.00024919, 0.00021183, -0.00018141, -0.00008142, 0.00028345, -0.00109689, 0.00066309, 0.00162569, -0.00085845, -0.00027700, 0.00027965, -0.00034320, 0.00019091, -0.00010480, -0.00006257, 0.00013421, -0.00014466, 0.00007079, 0.00001113, -0.00005515, 0.00003821, -0.00006932, -0.00001797, 0.00002433, -0.00003550, 0.00001679, 0.00000579, -0.00004351, 0.00003145, 0.00001041, -0.00001998, 0.00001448, -0.00001221, 0.00000857, -0.00000345, },
		{ 0.00004536, -0.00000437, -0.00001609, -0.00001762, 0.00000374, 0.00003614, -0.00000894, 0.00000982, 0.00000345, -0.00001327, 0.00003052, -0.00002699, -0.00000987, 0.00003970, -0.00002008, -0.00002183, 0.00003644, -0.00001219, -0.00003954, 0.00003090, -0.00005374, -0.00000253, -0.00002192, -0.00002605, 0.00011502, -0.00015711, 0.00009163, 0.00002280, -0.00019719, 0.00024806, -0.00032658, 0.00015435, -0.00001060, -0.00107446, 0.00131065, 0.00111841, -0.00109281, 0.00008712, 0.00008102, -0.00027420, 0.00022879, -0.00020209, 0.00004813, 0.00006370, -0.00013493, 0.00010984, -0.00003707, -0.00002265, 0.00001711, -0.00007371, 0.00001169, 0.00000045, -0.00003016, 0.00003037, -0.00001465, -0.00003081, 0.00004210, -0.00000856, -0.00001296, 0.00001608, -0.00001463, 0.00001249, -0.00000864, 0.00000344, },
		{ -0.00001300, -0.00001396, -0.00001878, 0.00001568, 0.00003255, -0.00001278, 0.00001442, -0.00000539, -0.00000394, 0.00002446, -0.00003537, 0.00001124, 0.00003116, -0.00003561, -0.00000125, 0.00003162, -0.00003457, -0.00000836, 0.00001092, -0.00006324, 0.00002357, -0.00005661, 0.00002828, 0.00006801, -0.00016283, 0.00016684, -0.00010235, -0.00008007, 0.00018924, -0.00038079, 0.00035247, -0.00040620, -0.00077491, 0.00177526, 0.00044812, -0.00105758, 0.00034297, -0.00014341, -0.00012507, 0.00017970, -0.00023877, 0.00014187, -0.00003028, -0.00008289, 0.00011351, -0.00006991, 0.00001003, -0.00001623, -0.00005608, 0.00002509, -0.00002452, -0.00001139, 0.00003147, -0.00003214, -0.00001003, 0.00004163, -0.00002515, -0.00000083, 0.00001307, -0.00001486, 0.00001529, -0.00001225, 0.00001507, 0.00003835, },
		{ -0.00001311, -0.00001762, 0.00002783, 0.00002521, -0.00001379, 0.00001617, -0.00001184, 0.00000578, 0.00001296, -0.00003488, 0.00003022, 0.00001288, -0.00004017, 0.00001934, 0.00001274, -0.00003874, 0.00001714, -0.00002326, -0.00004786, 0.00002831, -0.00007639, 0.00008188, -0.00000636, -0.00011551, 0.00018635, -0.00020115, 0.00006245, 0.00004974, -0.00031534, 0.00043204, -0.00079580, -0.00023256, 0.00194404, -0.00023256, -0.00079580, 0.00043204, -0.00031534, 0.00004974, 0.00006245, -0.00020115, 0.00018635, -0.00011551, -0.00000636, 0.00008188, -0.00007639, 0.00002831, -0.00004786, -0.00002326, 0.00001714, -0.00003874, 0.00001274, 0.00001934, -0.00004017, 0.00001288, 0.00003022, -0.00003488, 0.00001296, 0.00000578, -0.00001184, 0.00001617, -0.00001379, 0.00002521, 0.00002783, -0.00001762, },
		{ -0.00001300, 0.00003835, 0.00001507, -0.00001225, 0.00001529, -0.00001486, 0.00001307, -0.00000083, -0.00002515, 0.00004163, -0.00001003, -0.00003214, 0.00003147, -0.00001139, -0.00002452, 0.00002509, -0.00005608, -0.00001623, 0.00001003, -0.00006991, 0.00011351, -0.00008289, -0.00003028, 0.00014187, -0.00023877, 0.00017970, -0.00012507, -0.00014341, 0.00034297, -0.00105758, 0.00044812, 0.00177526, -0.00077491, -0.00040620, 0.00035247, -0.00038079, 0.00018924, -0.00008007, -0.00010235, 0.00016684, -0.00016283, 0.00006801, 0.00002828, -0.00005661, 0.00002357, -0.00006324, 0.00001092, -0.00000836, -0.00003457, 0.00003162, -0.00000125, -0.00003561, 0.00003116, 0.00001124, -0.00003537, 0.00002446, -0.00000394, -0.00000539, 0.00001442, -0.00001278, 0.00003255, 0.00001568, -0.00001878, -0.00001396, },
		{ 0.00004536, 0.00000344, -0.00000864, 0.00001249, -0.00001463, 0.00001608, -0.00001296, -0.00000856, 0.00004210, -0.00003081, -0.00001465, 0.00003037, -0.00003016, 0.00000045, 0.00001169, -0.00007371, 0.00001711, -0.00002265, -0.00003707, 0.00010984, -0.00013493, 0.00006370, 0.00004813, -0.00020209, 0.00022879, -0.00027420, 0.00008102, 0.00008712, -0.00109281, 0.00111841, 0.00131065, -0.00107446, -0.00001060, 0.00015435, -0.00032658, 0.00024806, -0.00019719, 0.00002280, 0.00009163, -0.00015711, 0.00011502, -0.00002605, -0.00002192, -0.00000253, -0.00005374, 0.00003090, -0.00003954, -0.00001219, 0.00003644, -0.00002183, -0.00002008, 0.00003970, -0.00000987, -0.00002699, 0.00003052, -0.00001327, 0.00000345, 0.00000982, -0.00000894, 0.00003614, 0.00000374, -0.00001762, -0.00001609, -0.00000437, },
		{ -0.00000813, -0.00000345, 0.00000857, -0.00001221, 0.00001448, -0.00001998, 0.00001041, 0.00003145, -0.00004351, 0.00000579, 0.00001679, -0.00003550, 0.00002433, -0.00001797, -0.00006932, 0.00003821, -0.00005515, 0.00001113, 0.00007079, -0.00014466, 0.00013421, -0.00006257, -0.00010480, 0.00019091, -0.00034320, 0.00027965, -0.00027700, -0.00085845, 0.00162569, 0.00066309, -0.00109689, 0.00028345, -0.00008142, -0.00018141, 0.00021183, -0.00024919, 0.00013095, -0.00001132, -0.00010280, 0.00012058, -0.00006072, 0.00000989, -0.00003792, -0.00002109, 0.00002600, -0.00006100, 0.00002003, 0.00002417, -0.00003373, 0.00000075, 0.00003679, -0.00002756, -0.00001251, 0.00002976, -0.00001938, 0.00001266, 0.00000302, -0.00000239, 0.00003548, -0.00000637, -0.00001559, -0.00001835, 0.00000796, 0.00004727, },
		{ 0.00000267, 0.00000411, -0.00000905, 0.00000947, -0.00002001, 0.00002666, 0.00001278, -0.00004503, 0.00002188, -0.00000350, -0.00002517, 0.00003603, -0.00005250, -0.00004512, 0.00003926, -0.00007345, 0.00005740, 0.00000945, -0.00010971, 0.00015802, -0.00015252, 0.00001768, 0.00007874, -0.00030473, 0.00037625, -0.00064869, -0.00038371, 0.00185581, -0.00001609, -0.00088181, 0.00041274, -0.00027250, -0.00000237, 0.00010042, -0.00022133, 0.00018646, -0.00010526, -0.00002177, 0.00008694, -0.00006466, 0.00002239, -0.00006483, 0.00002216, -0.00000440, -0.00006008, 0.00004852, -0.00000087, -0.00003174, 0.00001936, 0.00002452, -0.00003791, 0.00000394, 0.00002287, -0.00002039, 0.00001980, -0.00000451, 0.00000612, 0.00003062, -0.00001356, -0.00001410, -0.00001904, 0.00002266, 0.00004308, -0.00001788, },
		{ -0.00000064, -0.00000632, 0.00000325, -0.00001334, 0.00003599, -0.00000870, -0.00003611, 0.00002841, -0.00002218, -0.00000334, 0.00002951, -0.00007912, -0.00001057, 0.00002111, -0.00006987, 0.00008529, -0.00005326, -0.00004357, 0.00012919, -0.00019271, 0.00012166, -0.00006791, -0.00016985, 0.00032390, -0.00091457, 0.00023534, 0.00176414, -0.00057974, -0.00052453, 0.00037171, -0.00036217, 0.00014752, -0.00004146, -0.00012790, 0.00017413, -0.00015862, 0.00005564, 0.00003123, -0.00004061, 0.00000811, -0.00006740, 0.00005705, -0.00004952, -0.00003270, 0.00005999, -0.00002843, -0.00001642, 0.00002954, 0.00000768, -0.00003972, 0.00001835, 0.00001227, -0.00001600, 0.00002285, -0.00001079, 0.00001520, 0.00002234, -0.00001744, -0.00001404, -0.00001640, 0.00003745, 0.00003280, -0.00002408, 0.00000889, },
		{ -0.00000453, -0.00000183, -0.00000216, 0.00003640, -0.00002747, -0.00002074, 0.00002413, -0.00003185, 0.00002124, 0.00000585, -0.00008859, 0.00002198, -0.00000798, -0.00004578, 0.00008564, -0.00009683, 0.00003011, 0.00006048, -0.00017309, 0.00017146, -0.00019691, 0.00001467, 0.00012319, -0.00098357, 0.00086170, 0.00138569, -0.00092381, -0.00014297, 0.00020646, -0.00033488, 0.00022188, -0.00016208, -0.00000541, 0.00010374, -0.00015654, 0.00010260, -0.00002210, -0.00000376, -0.00002749, -0.00003913, 0.00006579, -0.00009061, 0.00001409, 0.00004719, -0.00004647, 0.00000604, 0.00002872, -0.00000833, -0.00003436, 0.00002806, 0.00000106, -0.00000753, 0.00002091, -0.00001394, 0.00002299, 0.00001205, -0.00001842, -0.00001552, -0.00000911, 0.00004940, 0.00001767, -0.00002543, 0.00001409, -0.00000559, },
		{ -0.00000408, 0.00001026, 0.00002843, -0.00003953, -0.00000440, 0.00001194, -0.00002886, 0.00003887, -0.00002766, -0.00007822, 0.00004224, -0.00003643, -0.00001035, 0.00005971, -0.00010872, 0.00008676, -0.00002238, -0.00010541, 0.00015243, -0.00026316, 0.00018222, -0.00017538, -0.00081575, 0.00135573, 0.00082184, -0.00101091, 0.00015864, -0.00000703, -0.00021615, 0.00020551, -0.00022188, 0.00010221, 0.00000493, -0.00010544, 0.00010601, -0.00005164, 0.00002514, -0.00006769, 0.00001307, 0.00003966, -0.00010809, 0.00006399, 0.00001250, -0.00004617, 0.00002632, 0.00001894, -0.00001966, -0.00002479, 0.00003227, -0.00000797, 0.00000254, 0.00001453, -0.00001284, 0.00002774, 0.00000156, -0.00001762, -0.00001760, 0.00000307, 0.00005556, 0.00000020, -0.00002150, 0.00001702, -0.00001052, -0.00000351, },
		{ 0.00002066, 0.00001468, -0.00004341, 0.00000810, -0.00000254, -0.00001456, 0.00004257, -0.00006060, -0.00005216, 0.00004549, -0.00005424, 0.00002373, 0.00001787, -0.00008837, 0.00010995, -0.00009108, -0.00001756, 0.00007552, -0.00024531, 0.00026976, -0.00048595, -0.00043538, 0.00161341, 0.00020886, -0.00087240, 0.00031755, -0.00019141, -0.00005843, 0.00011636, -0.00020654, 0.00016016, -0.00008513, -0.00002924, 0.00007079, -0.00004785, 0.00002913, -0.00009130, 0.00007043, -0.00001618, -0.00008971, 0.00009737, -0.00003229, -0.00002594, 0.00003597, 0.00000559, -0.00002526, -0.00001401, 0.00003178, -0.00001316, 0.00001148, 0.00000555, -0.00000758, 0.00002826, -0.00000743, -0.00001642, -0.00001852, 0.00001887, 0.00005382, -0.00001626, -0.00001306, 0.00001662, -0.00001494, -0.00000264, -0.00000295, },
		{ -0.00000133, -0.00004020, 0.00001427, -0.00001344, 0.00000554, 0.00003043, -0.00008338, -0.00001904, 0.00003346, -0.00005665, 0.00004547, -0.00002504, -0.00004620, 0.00009657, -0.00012526, 0.00005790, -0.00002798, -0.00015335, 0.00024026, -0.00071389, 0.00007591, 0.00159280, -0.00032109, -0.00059043, 0.00032391, -0.00029286, 0.00008064, -0.00000460, -0.00012976, 0.00015461, -0.00013616, 0.00004115, 0.00001667, -0.00001670, 0.00000304, -0.00008199, 0.00010916, -0.00008335, -0.00003642, 0.00009905, -0.00007017, 0.00000750, 0.00003103, -0.00000512, -0.00002679, -0.00000387, 0.00002823, -0.00001429, 0.00001713, -0.00000353, 0.00000054, 0.00002440, -0.00001379, -0.00001598, -0.00001619, 0.00003553, 0.00004370, -0.00002829, -0.00000212, 0.00001250, -0.00001808, -0.00000132, 0.00000102, 0.00002679, },
		{ -0.00003254, 0.00001451, -0.00001690, 0.00002408, 0.00000585, -0.00009038, 0.00001162, 0.00001255, -0.00004510, 0.00004957, -0.00005574, 0.00000144, 0.00005642, -0.00011925, 0.00009618, -0.00011928, -0.00002448, 0.00009607, -0.00078532, 0.00060316, 0.00132099, -0.00067396, -0.00026754, 0.00021418, -0.00029460, 0.00015888, -0.00011083, -0.00002536, 0.00009585, -0.00013544, 0.00008030, -0.00003081, 0.00002345, -0.00004365, -0.00003610, 0.00011003, -0.00013642, 0.00003674, 0.00006483, -0.00008549, 0.00004157, 0.00001387, -0.00000914, -0.00002723, 0.00000534, 0.00002312, -0.00001226, 0.00001851, -0.00001029, 0.00000939, 0.00001706, -0.00001730, -0.00001660, -0.00000901, 0.00004930, 0.00002681, -0.00003338, 0.00000850, 0.00000522, -0.00001917, 0.00000073, 0.00000658, 0.00002778, -0.00001640, },
		{ 0.00001137, -0.00001240, 0.00003470, -0.00002414, -0.00008122, 0.00003274, -0.00000934, -0.00002566, 0.00003753, -0.00006709, 0.00003903, 0.00000694, -0.00008240, 0.00008890, -0.00016709, 0.00009113, -0.00012150, -0.00066965, 0.00103119, 0.00088094, -0.00081315, 0.00000525, 0.00005122, -0.00021579, 0.00016189, -0.00016750, 0.00006725, 0.00001198, -0.00009100, 0.00007777, -0.00005132, 0.00005034, -0.00009025, 0.00003446, 0.00006689, -0.00015255, 0.00010535, 0.00000401, -0.00007039, 0.00006270, -0.00000778, -0.00000642, -0.00002890, 0.00001444, 0.00001727, -0.00000847, 0.00001591, -0.00001319, 0.00001668, 0.00000790, -0.00001860, -0.00001748, 0.00000325, 0.00005646, 0.00000665, -0.00003079, 0.00001597, -0.00000370, -0.00001780, 0.00000338, 0.00001236, 0.00002403, -0.00002852, -0.00002328, },
		{ -0.00000251, 0.00003410, -0.00005167, -0.00005995, 0.00004138, -0.00002589, -0.00000586, 0.00001596, -0.00005965, 0.00005744, -0.00003443, -0.00003365, 0.00004550, -0.00015905, 0.00014891, -0.00034589, -0.00038786, 0.00127567, 0.00038376, -0.00075983, 0.00017164, -0.00010057, -0.00009845, 0.00010375, -0.00016254, 0.00011808, -0.00006352, -0.00002641, 0.00004094, -0.00003800, 0.00004787, -0.00011347, 0.00010579, -0.00000959, -0.00012078, 0.00014439, -0.00006333, -0.00002836, 0.00006197, -0.00002411, -0.00000088, -0.00003198, 0.00002417, 0.00001072, -0.00000422, 0.00001062, -0.00001189, 0.00002068, -0.00000126, -0.00001876, -0.00001692, 0.00001901, 0.00005462, -0.00001218, -0.00002193, 0.00001837, -0.00001216, -0.00001423, 0.00000621, 0.00001731, 0.00001664, -0.00003697, -0.00001425, 0.00000804, },
		{ 0.00002281, -0.00007055, -0.00003298, 0.00003863, -0.00003416, 0.00000845, -0.00000688, -0.00003998, 0.00005603, -0.00005681, 0.00000712, -0.00001228, -0.00010516, 0.00012592, -0.00050724, -0.00000450, 0.00130291, -0.00006366, -0.00057686, 0.00021872, -0.00019586, 0.00001127, 0.00001728, -0.00010770, 0.00011642, -0.00010497, 0.00003003, -0.00000961, 0.00000037, 0.00001350, -0.00009750, 0.00015085, -0.00009473, -0.00004728, 0.00013786, -0.00011347, 0.00002657, 0.00003893, -0.00002754, 0.00000144, -0.00003410, 0.00003420, 0.00000326, -0.00000031, 0.00000433, -0.00000721, 0.00002063, -0.00000900, -0.00001868, -0.00001297, 0.00003496, 0.00004365, -0.00002541, -0.00000999, 0.00001536, -0.00001818, -0.00000933, 0.00000888, 0.00002085, 0.00000675, -0.00004188, -0.00000589, 0.00000678, 0.00000843, },
		{ -0.00007782, -0.00000662, 0.00002812, -0.00003457, 0.00001474, -0.00002449, -0.00001681, 0.00004112, -0.00005880, 0.00002698, -0.00006041, -0.00003248, 0.00002842, -0.00055443, 0.00039297, 0.00113317, -0.00038525, -0.00034354, 0.00017067, -0.00021870, 0.00007986, -0.00006091, -0.00003074, 0.00007214, -0.00010323, 0.00005686, -0.00004969, 0.00004395, -0.00004023, -0.00004140, 0.00015021, -0.00015850, 0.00004588, 0.00008529, -0.00012787, 0.00007491, 0.00000210, -0.00001622, -0.00000446, -0.00003141, 0.00004267, -0.00000493, 0.00000295, -0.00000138, -0.00000076, 0.00001682, -0.00001466, -0.00001851, -0.00000442, 0.00004693, 0.00002604, -0.00003042, 0.00000110, 0.00000836, -0.00002059, -0.00000433, 0.00001137, 0.00002266, -0.00000478, -0.00004346, 0.00000236, 0.00000800, 0.00001628, 0.00000436, },
		{ 0.00001457, 0.00001407, -0.00002952, 0.00001372, -0.00003399, 0.00000280, 0.00002176, -0.00004708, 0.00002397, -0.00008245, 0.00002651, -0.00011119, -0.00047005, 0.00072000, 0.00082836, -0.00054949, -0.00013052, 0.00007244, -0.00018081, 0.00009558, -0.00010438, 0.00003730, 0.00000906, -0.00006713, 0.00004724, -0.00006189, 0.00007100, -0.00009072, 0.00003949, 0.00010026, -0.00017668, 0.00012814, 0.00000266, -0.00010036, 0.00009909, -0.00003425, 0.00000487, -0.00001987, -0.00002059, 0.00004666, -0.00001289, 0.00000553, -0.00000547, 0.00000565, 0.00001033, -0.00001827, -0.00001745, 0.00000837, 0.00005128, 0.00000637, -0.00002722, 0.00000798, 0.00000002, -0.00001953, -0.00000030, 0.00001399, 0.00002243, -0.00001734, -0.00004150, 0.00001117, 0.00001036, 0.00001859, -0.00001640, -0.00007379, },
		{ -0.00000006, -0.00002179, 0.00000784, -0.00003596, 0.00001568, 0.00000534, -0.00003177, 0.00000599, -0.00007536, 0.00004673, -0.00024657, -0.00027351, 0.00091723, 0.00046978, -0.00056822, 0.00001737, -0.00002952, -0.00011083, 0.00006791, -0.00010465, 0.00007415, -0.00004614, -0.00001754, 0.00001007, -0.00004233, 0.00006756, -0.00011475, 0.00011884, 0.00001542, -0.00014038, 0.00017140, -0.00008298, -0.00003973, 0.00009026, -0.00005570, 0.00002574, -0.00004110, -0.00000089, 0.00004344, -0.00001893, 0.00000732, -0.00000745, 0.00001049, 0.00000252, -0.00002023, -0.00001409, 0.00002311, 0.00004625, -0.00001016, -0.00001842, 0.00000909, -0.00000678, -0.00001630, 0.00000240, 0.00001714, 0.00001956, -0.00003014, -0.00003526, 0.00002033, 0.00001170, 0.00001519, -0.00003505, -0.00006099, 0.00002828, },
		{ -0.00001347, -0.00000043, -0.00003277, 0.00002246, -0.00000516, -0.00002115, -0.00001443, -0.00004921, 0.00001971, -0.00033352, -0.00001247, 0.00096281, 0.00013416, -0.00048339, 0.00008658, -0.00010213, -0.00004016, 0.00001944, -0.00007090, 0.00007313, -0.00007458, 0.00002243, -0.00003545, -0.00000124, 0.00003348, -0.00009825, 0.00016923, -0.00007657, -0.00005985, 0.00016035, -0.00014316, 0.00003335, 0.00005180, -0.00005359, 0.00003548, -0.00006048, 0.00002475, 0.00003195, -0.00002125, 0.00000800, -0.00000735, 0.00001276, -0.00000536, -0.00002075, -0.00000717, 0.00003609, 0.00003289, -0.00001939, -0.00000835, 0.00000524, -0.00001012, -0.00001286, 0.00000429, 0.00002087, 0.00001328, -0.00004182, -0.00002407, 0.00002837, 0.00001034, 0.00000765, -0.00004872, -0.00004288, 0.00003408, -0.00001200, },
		{ -0.00000959, -0.00002650, 0.00002553, -0.00001113, -0.00001820, -0.00002666, -0.00002142, -0.00004430, -0.00034415, 0.00025343, 0.00087212, -0.00012448, -0.00034841, 0.00008948, -0.00013241, 0.00000881, -0.00002554, -0.00002298, 0.00004289, -0.00007121, 0.00003697, -0.00006884, 0.00004251, -0.00001759, -0.00004242, 0.00017278, -0.00014477, 0.00003876, 0.00009825, -0.00015820, 0.00009393, -0.00000194, -0.00002856, 0.00002693, -0.00006926, 0.00005016, 0.00001371, -0.00001880, 0.00000718, -0.00000549, 0.00001209, -0.00001239, -0.00001953, 0.00000353, 0.00004329, 0.00001501, -0.00001980, -0.00000140, -0.00000081, -0.00000977, -0.00001085, 0.00000658, 0.00002452, 0.00000305, -0.00005022, -0.00000812, 0.00003302, 0.00000601, -0.00000174, -0.00005624, -0.00002274, 0.00003263, -0.00002036, -0.00000586, },
		{ -0.00001796, 0.00002670, -0.00001597, -0.00002034, -0.00002677, -0.00000848, -0.00012131, -0.00027420, 0.00047041, 0.00068714, -0.00028283, -0.00021086, 0.00005253, -0.00012548, 0.00002819, -0.00005046, 0.00001828, 0.00000123, -0.00004449, 0.00002287, -0.00007693, 0.00007014, -0.00006411, 0.00003647, 0.00012805, -0.00016651, 0.00012279, 0.00000578, -0.00012422, 0.00012190, -0.00005106, 0.00000943, 0.00000002, -0.00006120, 0.00006790, -0.00000723, -0.00001200, 0.00000468, -0.00000241, 0.00000859, -0.00001788, -0.00001582, 0.00001662, 0.00004190, -0.00000198, -0.00001302, -0.00000031, -0.00000551, -0.00000733, -0.00001083, 0.00001054, 0.00002660, -0.00001084, -0.00005279, 0.00001086, 0.00003214, -0.00000007, -0.00001116, -0.00005757, -0.00000337, 0.00002515, -0.00002410, -0.00000005, -0.00001895, },
		{ 0.00002583, -0.00002212, -0.00002217, -0.00001854, -0.00001885, -0.00018333, -0.00014195, 0.00060390, 0.00046041, -0.00034581, -0.00010175, 0.00000345, -0.00009686, 0.00002332, -0.00005094, 0.00003928, -0.00003360, -0.00001072, -0.00001100, -0.00005848, 0.00007104, -0.00008568, 0.00011332, 0.00005057, -0.00013564, 0.00016479, -0.00008647, -0.00005426, 0.00010929, -0.00007776, 0.00004492, -0.00003773, -0.00003552, 0.00007202, -0.00002548, -0.00000279, 0.00000090, 0.00000115, 0.00000271, -0.00002121, -0.00000885, 0.00002910, 0.00003162, -0.00001286, -0.00000334, -0.00000508, -0.00000606, -0.00000545, -0.00001189, 0.00001673, 0.00002520, -0.00002679, -0.00004742, 0.00002956, 0.00002495, -0.00000584, -0.00001975, -0.00005316, 0.00001280, 0.00001346, -0.00002255, 0.00000249, -0.00002765, -0.00000693, },
		{ -0.00002910, -0.00001933, -0.00001061, -0.00004988, -0.00020848, 0.00002013, 0.00064397, 0.00023964, -0.00033691, -0.00003287, -0.00003781, -0.00006345, 0.00000670, -0.00003337, 0.00003731, -0.00005017, 0.00001386, -0.00004906, -0.00002324, 0.00004688, -0.00007136, 0.00016384, -0.00003324, -0.00006461, 0.00015224, -0.00014898, 0.00002689, 0.00006279, -0.00007303, 0.00006296, -0.00007370, 0.00000231, 0.00006044, -0.00003616, 0.00000588, -0.00000317, 0.00000422, -0.00000473, -0.00002164, 0.00000145, 0.00003720, 0.00001514, -0.00001455, 0.00000395, -0.00001294, -0.00000176, -0.00000665, -0.00001193, 0.00002447, 0.00001871, -0.00004192, -0.00003344, 0.00004383, 0.00001265, -0.00000956, -0.00002735, -0.00004359, 0.00002354, 0.00000018, -0.00001587, 0.00000001, -0.00003384, 0.00000641, 0.00002122, },
		{ -0.00001143, -0.00001151, -0.00008973, -0.00018778, 0.00017665, 0.00060252, 0.00005754, -0.00028618, -0.00000093, -0.00006244, -0.00003683, -0.00000909, -0.00000957, 0.00001906, -0.00004694, 0.00001961, -0.00007645, 0.00001348, 0.00001022, -0.00002376, 0.00017361, -0.00009611, 0.00002059, 0.00009159, -0.00016365, 0.00009199, 0.00000083, -0.00003991, 0.00005495, -0.00009486, 0.00004211, 0.00003598, -0.00003674, 0.00001118, -0.00000616, 0.00000569, -0.00001256, -0.00001842, 0.00001390, 0.00003773, -0.00000252, -0.00000733, 0.00000465, -0.00001942, 0.00000557, -0.00001197, -0.00000860, 0.00003170, 0.00000663, -0.00005253, -0.00001234, 0.00004992, -0.00000168, -0.00001073, -0.00003384, -0.00002978, 0.00002704, -0.00001115, -0.00000586, -0.00000855, -0.00003463, 0.00002017, 0.00001143, -0.00003383, },
		{ -0.00002506, -0.00012291, -0.00012637, 0.00029958, 0.00050500, -0.00007127, -0.00022033, 0.00000523, -0.00007093, -0.00002093, -0.00001654, 0.00000945, -0.00000433, -0.00003079, 0.00000618, -0.00008496, 0.00003806, -0.00002134, 0.00004242, 0.00014289, -0.00012002, 0.00009034, 0.00000518, -0.00012954, 0.00012113, -0.00005374, 0.00000800, 0.00002201, -0.00009270, 0.00007247, 0.00000559, -0.00002798, 0.00001148, -0.00000700, 0.00000451, -0.00001921, -0.00001109, 0.00002598, 0.00002943, -0.00001557, 0.00000510, -0.00000265, -0.00002029, 0.00001219, -0.00002028, -0.00000046, 0.00003555, -0.00000982, -0.00005521, 0.00001214, 0.00004595, -0.00001418, -0.00001037, -0.00003845, -0.00001354, 0.00002267, -0.00001684, 0.00000381, -0.00002225, -0.00002744, 0.00003053, -0.00000275, -0.00003270, -0.00000249, },
		{ -0.00013669, -0.00003977, 0.00037397, 0.00038043, -0.00014746, -0.00015745, -0.00000222, -0.00006854, -0.00001367, -0.00001465, 0.00001774, -0.00002294, -0.00001229, -0.00001918, -0.00007527, 0.00004424, -0.00003268, 0.00010670, 0.00008547, -0.00010222, 0.00012224, -0.00007748, -0.00006215, 0.00010837, -0.00008224, 0.00005188, -0.00002541, -0.00006649, 0.00008480, -0.00002203, -0.00001367, 0.00000708, -0.00000541, 0.00000012, -0.00002290, -0.00000002, 0.00003434, 0.00001384, -0.00001951, 0.00001689, -0.00001579, -0.00001342, 0.00001397, -0.00002843, 0.00001205, 0.00003338, -0.00002766, -0.00004801, 0.00003479, 0.00003273, -0.00002161, -0.00001041, -0.00003969, 0.00000217, 0.00001181, -0.00001432, 0.00000830, -0.00003737, -0.00001179, 0.00003307, -0.00001749, -0.00002447, 0.00000138, -0.00004845, },
		{ 0.00005279, 0.00039837, 0.00025358, -0.00018084, -0.00010646, -0.00001423, -0.00006098, -0.00001049, -0.00000710, 0.00001540, -0.00003181, -0.00000043, -0.00004617, -0.00005503, 0.00003501, -0.00001701, 0.00015078, 0.00002238, -0.00005505, 0.00010873, -0.00013055, 0.00001325, 0.00006286, -0.00007673, 0.00007500, -0.00007149, -0.00002364, 0.00007633, -0.00003949, 0.00000078, 0.00000004, -0.00000231, -0.00000711, -0.00002193, 0.00001328, 0.00003586, -0.00000488, -0.00001286, 0.00002221, -0.00002978, 0.00000008, 0.00000821, -0.00003237, 0.00002640, 0.00002385, -0.00004279, -0.00003133, 0.00005036, 0.00001372, -0.00002265, -0.00001253, -0.00003602, 0.00001359, -0.00000177, -0.00000371, 0.00000343, -0.00004799, 0.00000924, 0.00002519, -0.00002710, -0.00001191, -0.00000438, -0.00007367, -0.00012562, },
		{ 0.00038108, 0.00014096, -0.00018433, -0.00006955, -0.00002585, -0.00005180, -0.00000766, 0.00000104, 0.00000664, -0.00003165, 0.00000076, -0.00006648, -0.00003417, 0.00002036, 0.00002215, 0.00016499, -0.00002675, -0.00000007, 0.00005867, -0.00014062, 0.00007134, 0.00000429, -0.00004227, 0.00006880, -0.00010092, 0.00002303, 0.00005095, -0.00004317, 0.00001042, -0.00000664, 0.00000042, -0.00001564, -0.00001535, 0.00002611, 0.00002890, -0.00002103, 0.00000210, 0.00001750, -0.00003866, 0.00001601, -0.00000496, -0.00002879, 0.00003855, 0.00000780, -0.00005112, -0.00000816, 0.00005519, -0.00000598, -0.00001841, -0.00001704, -0.00002692, 0.00001740, -0.00001258, 0.00001137, -0.00001212, -0.00004801, 0.00002915, 0.00000828, -0.00002685, -0.00000123, -0.00001980, -0.00009149, -0.00009236, 0.00013467, },
		{ 0.00005076, -0.00016972, -0.00004539, -0.00003506, -0.00004227, -0.00000385, 0.00000625, -0.00000348, -0.00002682, -0.00000739, -0.00007682, -0.00001997, 0.00001232, 0.00007341, 0.00015052, -0.00005084, 0.00004095, -0.00000728, -0.00011028, 0.00009576, -0.00004512, 0.00000587, 0.00003569, -0.00010444, 0.00005991, 0.00001763, -0.00003433, 0.00001254, -0.00001038, 0.00000064, -0.00002290, -0.00000345, 0.00003515, 0.00001420, -0.00002930, 0.00001995, 0.00000295, -0.00003784, 0.00002848, -0.00002226, -0.00001665, 0.00004428, -0.00001173, -0.00004990, 0.00001655, 0.00004840, -0.00002130, -0.00001189, -0.00002247, -0.00001382, 0.00001224, -0.00001518, 0.00002429, -0.00003518, -0.00003407, 0.00004008, -0.00001167, -0.00001596, 0.00000077, -0.00003969, -0.00009575, -0.00004519, 0.00019564, 0.00033489, },
		{ -0.00014615, -0.00003152, -0.00004114, -0.00003249, 0.00000023, 0.00000767, -0.00001166, -0.00002211, -0.00002028, -0.00007914, -0.00001408, 0.00001983, 0.00012266, 0.00011713, -0.00005035, 0.00005434, -0.00006530, -0.00005548, 0.00008392, -0.00006900, 0.00004862, -0.00001238, -0.00008188, 0.00007734, -0.00001300, -0.00001821, 0.00000752, -0.00001014, -0.00000307, -0.00002597, 0.00001191, 0.00003733, -0.00000496, -0.00002659, 0.00003399, -0.00001737, -0.00002606, 0.00003220, -0.00003822, 0.00000205, 0.00004075, -0.00003016, -0.00003869, 0.00003722, 0.00003222, -0.00002892, -0.00000662, -0.00002605, -0.00000028, -0.00000004, -0.00000653, 0.00002761, -0.00005828, -0.00000794, 0.00003625, -0.00002589, 0.00000086, -0.00000947, -0.00005590, -0.00008585, 0.00000622, 0.00023221, 0.00027255, -0.00001495, },
		{ -0.00002529, -0.00004369, -0.00002270, 0.00000331, 0.00000644, -0.00001734, -0.00002033, -0.00003322, -0.00007833, -0.00001246, 0.00004557, 0.00015842, 0.00007772, -0.00003524, 0.00003881, -0.00009745, 0.00000171, 0.00004636, -0.00006217, 0.00007042, -0.00005892, -0.00004169, 0.00007266, -0.00003240, -0.00000188, -0.00000159, -0.00000689, -0.00001054, -0.00002250, 0.00002742, 0.00003096, -0.00002351, -0.00001326, 0.00003857, -0.00003687, -0.00000610, 0.00002454, -0.00004704, 0.00002262, 0.00002775, -0.00004273, -0.00001968, 0.00004921, 0.00001133, -0.00002827, -0.00000509, -0.00002511, 0.00000906, -0.00001430, 0.00001207, 0.00001617, -0.00007188, 0.00002299, 0.00001720, -0.00002657, 0.00001457, -0.00002961, -0.00006129, -0.00006633, 0.00005458, 0.00024557, 0.00020452, -0.00005768, -0.00012011, },
		{ -0.00004258, -0.00001374, 0.00000476, 0.00000418, -0.00002184, -0.00002146, -0.00004384, -0.00007895, -0.00000782, 0.00008584, 0.00017519, 0.00004259, -0.00001920, 0.00000444, -0.00009751, 0.00004192, 0.00000124, -0.00003136, 0.00006465, -0.00008909, 0.00000262, 0.00005058, -0.00003687, 0.00000856, -0.00001057, -0.00000318, -0.00001973, -0.00001171, 0.00003909, 0.00001656, -0.00003610, 0.00000681, 0.00003115, -0.00004885, 0.00001602, 0.00000684, -0.00004486, 0.00003915, 0.00000804, -0.00004597, 0.00000283, 0.00005019, -0.00000868, -0.00002150, -0.00000765, -0.00001853, 0.00001026, -0.00002358, 0.00003458, -0.00001009, -0.00006808, 0.00004767, -0.00001074, -0.00001098, 0.00001553, -0.00005139, -0.00005361, -0.00004377, 0.00009635, 0.00023883, 0.00013870, -0.00008081, -0.00009609, -0.00002394, },
		{ -0.00000659, 0.00000503, 0.00000163, -0.00002675, -0.00002364, -0.00005282, -0.00008269, 0.00000701, 0.00013295, 0.00017345, 0.00001588, -0.00001310, -0.00003276, -0.00007165, 0.00005509, -0.00003336, 0.00000865, 0.00003519, -0.00009490, 0.00003759, 0.00002078, -0.00002829, 0.00001006, -0.00001589, -0.00000197, -0.00002716, 0.00000500, 0.00004332, -0.00000302, -0.00003881, 0.00002744, 0.00001331, -0.00004893, 0.00003316, -0.00001592, -0.00003129, 0.00004649, -0.00001332, -0.00003886, 0.00002363, 0.00004075, -0.00002302, -0.00001248, -0.00001240, -0.00000776, 0.00000211, -0.00002167, 0.00005177, -0.00004495, -0.00004444, 0.00005560, -0.00003602, 0.00001606, -0.00000164, -0.00006376, -0.00003706, -0.00002296, 0.00013075, 0.00021528, 0.00008116, -0.00008892, -0.00007679, -0.00002461, -0.00003840, },
		},
		{ { -0.00005255, -0.00003241, -0.00010372, -0.00003050, 0.00019040, 0.00019100, 0.00001733, -0.00002961, -0.00006757, -0.00002406, 0.00000798, -0.00004959, 0.00003741, -0.00005541, -0.00002289, 0.00000103, 0.00003187, 0.00002137, -0.00003872, 0.00000033, -0.00003381, 0.00004049, -0.00002209, -0.00000120, 0.00001075, -0.00001164, 0.00002499, 0.00001277, -0.00001408, -0.00002616, 0.00001886, 0.00000122, -0.00000565, 0.00000122, 0.00001886, -0.00002616, -0.00001408, 0.00001277, 0.00002499, -0.00001164, 0.00001075, -0.00000120, -0.00002209, 0.00004049, -0.00003381, 0.00000033, -0.00003872, 0.00002137, 0.00003187, 0.00000103, -0.00002289, -0.00005541, 0.00003741, -0.00004959, 0.00000798, -0.00002406, -0.00006757, -0.00002961, 0.00001733, 0.00019100, 0.00019040, -0.00003050, -0.00010372, -0.00003241, },
		{ -0.00003280, -0.00011778, -0.00001309, 0.00023578, 0.00018384, -0.00000671, -0.00004329, -0.00006546, -0.00000883, -0.00001037, -0.00003627, 0.00003182, -0.00006167, -0.00001601, -0.00000056, 0.00004355, 0.00000412, -0.00003320, -0.00000392, -0.00002616, 0.00004025, -0.00002795, 0.00000737, 0.00000317, -0.00000416, 0.00002360, 0.00001089, -0.00002120, -0.00001878, 0.00002342, -0.00000480, -0.00000458, 0.00000647, 0.00001144, -0.00003001, -0.00000741, 0.00001479, 0.00002462, -0.00001809, 0.00001891, -0.00001187, -0.00001264, 0.00003644, -0.00003829, 0.00000240, -0.00004043, 0.00003658, 0.00001712, 0.00000711, -0.00003381, -0.00004333, 0.00003627, -0.00005841, 0.00002701, -0.00004638, -0.00006053, -0.00001577, 0.00003782, 0.00018973, 0.00014811, -0.00004019, -0.00009249, -0.00003249, -0.00005238, },
		{ -0.00013423, 0.00001384, 0.00028243, 0.00016641, -0.00003375, -0.00005304, -0.00005728, -0.00000142, -0.00002484, -0.00002203, 0.00002202, -0.00006337, -0.00001266, 0.00000295, 0.00004983, -0.00001237, -0.00002549, -0.00000946, -0.00001624, 0.00003614, -0.00003000, 0.00001310, -0.00000273, 0.00000320, 0.00002113, 0.00000863, -0.00002771, -0.00000854, 0.00002443, -0.00001048, -0.00000153, 0.00000995, 0.00000230, -0.00003023, -0.00000193, 0.00001736, 0.00002191, -0.00002236, 0.00002627, -0.00002343, -0.00000037, 0.00002819, -0.00003923, 0.00000193, -0.00003732, 0.00004708, 0.00000239, 0.00001580, -0.00004789, -0.00002543, 0.00002677, -0.00005912, 0.00004217, -0.00007275, -0.00004323, -0.00000567, 0.00005579, 0.00018110, 0.00011045, -0.00004406, -0.00008424, -0.00003247, -0.00005197, -0.00003420, },
		{ 0.00005227, 0.00032780, 0.00013668, -0.00006133, -0.00005639, -0.00004688, -0.00000049, -0.00003400, -0.00000962, 0.00001061, -0.00006237, -0.00001131, 0.00001071, 0.00004963, -0.00002569, -0.00001743, -0.00001506, -0.00000524, 0.00002906, -0.00002856, 0.00001575, -0.00000625, 0.00000952, 0.00001823, 0.00000553, -0.00003250, 0.00000338, 0.00002172, -0.00001476, 0.00000293, 0.00001093, -0.00000723, -0.00002724, 0.00000197, 0.00002067, 0.00001656, -0.00002347, 0.00003139, -0.00003434, 0.00001348, 0.00001640, -0.00003670, -0.00000087, -0.00002927, 0.00005096, -0.00000914, 0.00002423, -0.00006290, -0.00000330, 0.00000886, -0.00004912, 0.00004850, -0.00009806, -0.00001718, -0.00000206, 0.00007303, 0.00016540, 0.00007906, -0.00004430, -0.00007861, -0.00003201, -0.00005150, -0.00003708, -0.00015248, },
		{ 0.00036793, 0.00009332, -0.00008546, -0.00005295, -0.00003776, -0.00000350, -0.00003818, -0.00000045, -0.00000044, -0.00006051, -0.00000992, 0.00002077, 0.00004321, -0.00003428, -0.00001068, -0.00001943, 0.00000543, 0.00002023, -0.00002437, 0.00001557, -0.00000714, 0.00001413, 0.00001538, 0.00000132, -0.00003458, 0.00001554, 0.00001564, -0.00001685, 0.00000802, 0.00000907, -0.00001575, -0.00002194, 0.00000440, 0.00002453, 0.00000869, -0.00002085, 0.00003301, -0.00004296, 0.00002728, 0.00000219, -0.00003126, -0.00000524, -0.00001715, 0.00004743, -0.00001483, 0.00002931, -0.00007573, 0.00002001, -0.00001548, -0.00002772, 0.00004190, -0.00011611, 0.00001347, -0.00000559, 0.00009106, 0.00014260, 0.00005565, -0.00004349, -0.00007449, -0.00003128, -0.00005096, -0.00004201, -0.00017164, 0.00010439, },
		{ 0.00003680, -0.00010150, -0.00004451, -0.00003205, -0.00000765, -0.00003888, 0.00000541, -0.00001016, -0.00005898, -0.00000663, 0.00003057, 0.00003211, -0.00003769, -0.00000636, -0.00002152, 0.00001446, 0.00001099, -0.00001846, 0.00001321, -0.00000560, 0.00001674, 0.00001282, -0.00000392, -0.00003326, 0.00002639, 0.00000708, -0.00001627, 0.00001278, 0.00000449, -0.00002211, -0.00001548, 0.00000578, 0.00002840, -0.00000113, -0.00001445, 0.00003030, -0.00004779, 0.00003928, -0.00001298, -0.00002377, -0.00001005, -0.00000268, 0.00003711, -0.00001324, 0.00002841, -0.00008304, 0.00004040, -0.00004235, 0.00000323, 0.00002032, -0.00012095, 0.00004258, -0.00001433, 0.00011012, 0.00011307, 0.00004157, -0.00004434, -0.00007013, -0.00003120, -0.00004979, -0.00005004, -0.00019010, 0.00017220, 0.00039733, },
		{ -0.00010553, -0.00003457, -0.00003000, -0.00001089, -0.00003804, 0.00000885, -0.00001866, -0.00005800, -0.00000038, 0.00003760, 0.00001868, -0.00003653, -0.00000486, -0.00002072, 0.00002078, 0.00000261, -0.00001196, 0.00000956, -0.00000214, 0.00001738, 0.00001058, -0.00000982, -0.00002836, 0.00003454, -0.00000278, -0.00001296, 0.00001625, -0.00000227, -0.00002550, -0.00000903, 0.00000682, 0.00003150, -0.00001186, -0.00000488, 0.00002300, -0.00004774, 0.00004781, -0.00002751, -0.00001525, -0.00001407, 0.00001188, 0.00002192, -0.00000446, 0.00002007, -0.00008210, 0.00005362, -0.00006646, 0.00003928, -0.00001529, -0.00010840, 0.00006336, -0.00002409, 0.00012859, 0.00007833, 0.00003720, -0.00004914, -0.00006345, -0.00003333, -0.00004676, -0.00006302, -0.00020501, 0.00025674, 0.00040918, -0.00002978, },
		{ -0.00002723, -0.00003013, -0.00001244, -0.00003723, 0.00001116, -0.00002673, -0.00005679, 0.00000880, 0.00004006, 0.00000553, -0.00003227, -0.00000577, -0.00001703, 0.00002379, -0.00000387, -0.00000593, 0.00000558, 0.00000247, 0.00001629, 0.00000854, -0.00001574, -0.00002021, 0.00003900, -0.00001257, -0.00000731, 0.00001764, -0.00001029, -0.00002564, -0.00000358, 0.00000823, 0.00003290, -0.00002216, 0.00000671, 0.00001161, -0.00004230, 0.00005155, -0.00003982, -0.00000672, -0.00001622, 0.00002409, 0.00000473, 0.00000986, 0.00000444, -0.00007156, 0.00005622, -0.00008216, 0.00007398, -0.00006069, -0.00007739, 0.00006995, -0.00002941, 0.00014310, 0.00004144, 0.00004139, -0.00005908, -0.00005272, -0.00003945, -0.00004011, -0.00008363, -0.00021169, 0.00035700, 0.00039627, -0.00010041, -0.00009556, },
		{ -0.00002984, -0.00001295, -0.00003706, 0.00001330, -0.00003536, -0.00005393, 0.00001970, 0.00003726, -0.00000505, -0.00002677, -0.00000809, -0.00001104, 0.00002344, -0.00000782, -0.00000121, 0.00000216, 0.00000740, 0.00001391, 0.00000651, -0.00002088, -0.00000963, 0.00003927, -0.00002098, -0.00000002, 0.00001642, -0.00001843, -0.00002275, 0.00000024, 0.00001059, 0.00003178, -0.00003055, 0.00001867, -0.00000270, -0.00003175, 0.00004971, -0.00004860, 0.00000097, -0.00001585, 0.00003188, -0.00001114, 0.00002680, -0.00001661, -0.00005194, 0.00004646, -0.00008465, 0.00009998, -0.00010873, -0.00003060, 0.00005902, -0.00002499, 0.00014915, 0.00000692, 0.00005116, -0.00007365, -0.00003727, -0.00005085, -0.00002814, -0.00011510, -0.00020350, 0.00046888, 0.00035247, -0.00016660, -0.00007251, -0.00002598, },
		{ -0.00001414, -0.00003708, 0.00001548, -0.00004516, -0.00004793, 0.00003029, 0.00002980, -0.00001165, -0.00002185, -0.00001044, -0.00000386, 0.00002023, -0.00000905, 0.00000167, -0.00000000, 0.00001184, 0.00001070, 0.00000434, -0.00002447, 0.00000223, 0.00003542, -0.00002695, 0.00000792, 0.00001241, -0.00002554, -0.00001752, 0.00000224, 0.00001413, 0.00002762, -0.00003574, 0.00002920, -0.00001818, -0.00001711, 0.00004222, -0.00005294, 0.00000722, -0.00001283, 0.00003391, -0.00002258, 0.00004278, -0.00003985, -0.00002573, 0.00002499, -0.00007120, 0.00011056, -0.00015055, 0.00002552, 0.00003088, -0.00000727, 0.00014238, -0.00001987, 0.00006189, -0.00009037, -0.00001821, -0.00006750, -0.00001007, -0.00016033, -0.00017222, 0.00058435, 0.00027447, -0.00021858, -0.00004048, -0.00003248, -0.00002645, },
		{ -0.00003593, 0.00001698, -0.00005586, -0.00003778, 0.00003819, 0.00001939, -0.00001391, -0.00001879, -0.00001151, 0.00000314, 0.00001510, -0.00000783, 0.00000257, -0.00000048, 0.00001514, 0.00000714, 0.00000198, -0.00002591, 0.00001400, 0.00002807, -0.00002976, 0.00001544, 0.00000584, -0.00003059, -0.00001095, 0.00000268, 0.00001870, 0.00002034, -0.00003678, 0.00003657, -0.00003278, -0.00000007, 0.00002974, -0.00005244, 0.00001172, -0.00000765, 0.00002989, -0.00002732, 0.00005427, -0.00006128, 0.00000296, -0.00000506, -0.00004198, 0.00010114, -0.00017719, 0.00008175, -0.00001018, 0.00002431, 0.00011992, -0.00003394, 0.00006809, -0.00010502, 0.00000136, -0.00008735, 0.00001292, -0.00022060, -0.00010922, 0.00069150, 0.00016337, -0.00024718, -0.00000622, -0.00004590, -0.00001822, -0.00001798, },
		{ 0.00001635, -0.00006623, -0.00002349, 0.00004138, 0.00000842, -0.00001259, -0.00001814, -0.00001032, 0.00000862, 0.00000928, -0.00000481, 0.00000168, 0.00000082, 0.00001683, 0.00000363, -0.00000050, -0.00002489, 0.00002440, 0.00001823, -0.00002919, 0.00002149, -0.00000269, -0.00003286, -0.00000416, 0.00000215, 0.00002375, 0.00001040, -0.00003335, 0.00003945, -0.00004441, 0.00001726, 0.00001362, -0.00004726, 0.00001445, -0.00000126, 0.00002057, -0.00002433, 0.00005851, -0.00007694, 0.00002915, -0.00003845, -0.00000041, 0.00007052, -0.00018152, 0.00012775, -0.00005638, 0.00006663, 0.00008164, -0.00003193, 0.00006458, -0.00011249, 0.00001664, -0.00010617, 0.00003699, -0.00029415, -0.00000721, 0.00077528, 0.00002576, -0.00024582, 0.00002217, -0.00006284, -0.00000517, -0.00002593, -0.00003199, },
		{ -0.00007424, -0.00000628, 0.00003879, -0.00000065, -0.00000925, -0.00001955, -0.00000652, 0.00001153, 0.00000402, -0.00000090, -0.00000047, 0.00000369, 0.00001668, 0.00000050, -0.00000289, -0.00002146, 0.00003234, 0.00000717, -0.00002542, 0.00002521, -0.00001230, -0.00003203, 0.00000181, 0.00000145, 0.00002842, -0.00000124, -0.00002578, 0.00003715, -0.00005137, 0.00003262, -0.00000428, -0.00003807, 0.00001558, 0.00000509, 0.00000770, -0.00001401, 0.00005411, -0.00008381, 0.00004800, -0.00006875, 0.00004720, 0.00002154, -0.00015989, 0.00015410, -0.00009793, 0.00011316, 0.00003082, -0.00001330, 0.00004788, -0.00010803, 0.00002206, -0.00011788, 0.00005528, -0.00037497, 0.00013758, 0.00081949, -0.00012624, -0.00021230, 0.00003701, -0.00007806, 0.00001066, -0.00003809, -0.00002411, 0.00001196, },
		{ 0.00001156, 0.00003059, -0.00000593, -0.00000583, -0.00002197, -0.00000048, 0.00001132, 0.00000043, 0.00000291, -0.00000319, 0.00000766, 0.00001471, -0.00000197, -0.00000493, -0.00001600, 0.00003712, -0.00000379, -0.00001908, 0.00002603, -0.00002194, -0.00002824, 0.00000620, 0.00000135, 0.00003173, -0.00001326, -0.00001508, 0.00002980, -0.00005254, 0.00004394, -0.00002183, -0.00002597, 0.00001544, 0.00001013, -0.00000638, 0.00000188, 0.00004133, -0.00008042, 0.00005584, -0.00008960, 0.00009272, -0.00003912, -0.00011321, 0.00015435, -0.00012500, 0.00015506, -0.00002597, 0.00001900, 0.00001737, -0.00008875, 0.00001268, -0.00011577, 0.00005859, -0.00045218, 0.00032324, 0.00080932, -0.00027599, -0.00014997, 0.00003296, -0.00008573, 0.00002554, -0.00005298, -0.00001218, 0.00000260, -0.00007759, },
		{ 0.00001831, -0.00000651, -0.00000409, -0.00002393, 0.00000681, 0.00000810, -0.00000077, 0.00000575, -0.00000570, 0.00001208, 0.00001114, -0.00000359, -0.00000638, -0.00000914, 0.00003849, -0.00001348, -0.00001106, 0.00002372, -0.00003049, -0.00002203, 0.00000859, 0.00000242, 0.00003278, -0.00002416, -0.00000278, 0.00001828, -0.00004767, 0.00004971, -0.00003691, -0.00001233, 0.00001437, 0.00001280, -0.00001913, 0.00002058, 0.00002200, -0.00006728, 0.00005099, -0.00009598, 0.00012763, -0.00010192, -0.00004723, 0.00012660, -0.00012986, 0.00018295, -0.00007984, 0.00005867, -0.00002396, -0.00005483, -0.00001418, -0.00009414, 0.00003674, -0.00051046, 0.00054063, 0.00073439, -0.00040457, -0.00006772, 0.00000849, -0.00008103, 0.00003490, -0.00006758, 0.00000242, -0.00001181, -0.00007439, 0.00002711, },
		{ -0.00000270, -0.00000519, -0.00002397, 0.00001391, 0.00000257, 0.00000064, 0.00000695, -0.00000730, 0.00001622, 0.00000639, -0.00000422, -0.00000706, -0.00000167, 0.00003660, -0.00002100, -0.00000241, 0.00001844, -0.00003701, -0.00001427, 0.00000901, 0.00000492, 0.00003095, -0.00003256, 0.00000932, 0.00000420, -0.00003738, 0.00004917, -0.00004773, 0.00000136, 0.00001270, 0.00001249, -0.00002829, 0.00003889, -0.00000084, -0.00004670, 0.00003421, -0.00008536, 0.00014473, -0.00015611, 0.00002833, 0.00007419, -0.00010873, 0.00018904, -0.00012127, 0.00009699, -0.00006972, -0.00001025, -0.00005754, -0.00005031, -0.00001915, -0.00053165, 0.00077293, 0.00059154, -0.00049375, 0.00002113, -0.00003333, -0.00006158, 0.00003463, -0.00007811, 0.00001701, -0.00002993, -0.00006384, 0.00003753, 0.00000446, },
		{ -0.00000933, -0.00002102, 0.00001930, -0.00000410, 0.00000439, 0.00000622, -0.00000746, 0.00001935, 0.00000102, -0.00000380, -0.00000694, 0.00000561, 0.00003199, -0.00002580, 0.00000580, 0.00001070, -0.00004075, -0.00000597, 0.00000785, 0.00000869, 0.00002605, -0.00003740, 0.00001946, -0.00001046, -0.00002313, 0.00004247, -0.00005305, 0.00001368, 0.00001066, 0.00000914, -0.00003226, 0.00005380, -0.00002357, -0.00002245, 0.00000866, -0.00005839, 0.00013964, -0.00019174, 0.00010130, 0.00000536, -0.00006292, 0.00016901, -0.00014211, 0.00012454, -0.00011107, 0.00003734, -0.00011201, 0.00001371, -0.00011410, -0.00049748, 0.00099653, 0.00038676, -0.00052930, 0.00010097, -0.00008508, -0.00002839, 0.00002227, -0.00008091, 0.00002807, -0.00004900, -0.00004666, 0.00004077, -0.00000800, 0.00000406, },
		{ -0.00001472, 0.00002169, -0.00001051, 0.00000977, 0.00000363, -0.00000593, 0.00002087, -0.00000431, -0.00000237, -0.00000614, 0.00001204, 0.00002542, -0.00002771, 0.00001263, 0.00000127, -0.00004134, 0.00000183, 0.00000579, 0.00001321, 0.00001838, -0.00003817, 0.00002620, -0.00002359, -0.00000694, 0.00003065, -0.00005237, 0.00002347, 0.00000843, 0.00000323, -0.00003039, 0.00006297, -0.00004272, 0.00000109, -0.00002070, -0.00001885, 0.00011185, -0.00020167, 0.00015933, -0.00006813, 0.00000113, 0.00012331, -0.00013754, 0.00013320, -0.00013848, 0.00007778, -0.00016798, 0.00009019, -0.00024674, -0.00039300, 0.00118358, 0.00013568, -0.00050394, 0.00015682, -0.00013624, 0.00001392, -0.00000199, -0.00007365, 0.00003217, -0.00006534, -0.00002522, 0.00003620, -0.00001638, 0.00001160, -0.00001577, },
		{ 0.00002030, -0.00001526, 0.00001576, -0.00000038, -0.00000276, 0.00002037, -0.00000891, -0.00000009, -0.00000486, 0.00001714, 0.00001774, -0.00002693, 0.00001734, -0.00000887, -0.00003877, 0.00000827, 0.00000359, 0.00001770, 0.00000872, -0.00003496, 0.00002867, -0.00003338, 0.00000889, 0.00001546, -0.00004598, 0.00002991, 0.00000609, -0.00000432, -0.00002298, 0.00006517, -0.00005565, 0.00001979, -0.00004806, 0.00002687, 0.00006496, -0.00018302, 0.00019224, -0.00013312, 0.00007285, 0.00005750, -0.00010743, 0.00011821, -0.00014394, 0.00010035, -0.00021290, 0.00016583, -0.00040767, -0.00021026, 0.00130571, -0.00013751, -0.00041925, 0.00017725, -0.00017533, 0.00005784, -0.00003499, -0.00005611, 0.00002696, -0.00007513, -0.00000314, 0.00002488, -0.00001896, 0.00001757, -0.00002301, -0.00000552, },
		{ -0.00001730, 0.00002126, -0.00000513, 0.00000169, 0.00001775, -0.00001215, 0.00000276, -0.00000337, 0.00002068, 0.00000978, -0.00002390, 0.00001948, -0.00001869, -0.00003344, 0.00001279, 0.00000198, 0.00002128, -0.00000182, -0.00002847, 0.00002671, -0.00003853, 0.00002220, -0.00000088, -0.00003492, 0.00003265, 0.00000370, -0.00001231, -0.00001123, 0.00006040, -0.00006099, 0.00003060, -0.00006791, 0.00007103, 0.00000619, -0.00013791, 0.00019402, -0.00017763, 0.00013952, -0.00001854, -0.00005675, 0.00007955, -0.00012308, 0.00009619, -0.00023366, 0.00022324, -0.00057903, 0.00004861, 0.00133858, -0.00040298, -0.00028616, 0.00015684, -0.00019229, 0.00009442, -0.00007101, -0.00003066, 0.00001202, -0.00007543, 0.00001539, 0.00000942, -0.00001537, 0.00002004, -0.00002915, 0.00000535, 0.00001507, },
		{ 0.00002526, -0.00000978, 0.00000679, 0.00001317, -0.00001352, 0.00000575, -0.00000189, 0.00002263, 0.00000224, -0.00001926, 0.00001890, -0.00002721, -0.00002601, 0.00001518, 0.00000146, 0.00002312, -0.00001199, -0.00001986, 0.00002088, -0.00003855, 0.00003126, -0.00001605, -0.00002080, 0.00003179, 0.00000128, -0.00001951, 0.00000310, 0.00004981, -0.00005886, 0.00003220, -0.00007619, 0.00010593, -0.00005490, -0.00007324, 0.00016406, -0.00019318, 0.00018870, -0.00009239, 0.00000508, 0.00002248, -0.00007671, 0.00006071, -0.00021939, 0.00024372, -0.00073580, 0.00036959, 0.00126635, -0.00062936, -0.00012367, 0.00009762, -0.00018086, 0.00011528, -0.00010283, -0.00000196, -0.00001071, -0.00006505, 0.00002658, -0.00000653, -0.00000679, 0.00001803, -0.00003242, 0.00001621, 0.00000665, -0.00001602, },
		{ -0.00001354, 0.00001177, 0.00000713, -0.00001274, 0.00000839, -0.00000057, 0.00002312, -0.00000435, -0.00001372, 0.00001575, -0.00003362, -0.00001740, 0.00001556, 0.00000225, 0.00002267, -0.00002059, -0.00001056, 0.00001233, -0.00003373, 0.00003508, -0.00002798, -0.00000558, 0.00002785, -0.00000110, -0.00002490, 0.00001798, 0.00003533, -0.00005072, 0.00002517, -0.00007108, 0.00012546, -0.00010817, 0.00000054, 0.00010743, -0.00017653, 0.00021080, -0.00015147, 0.00006589, -0.00004317, -0.00001133, -0.00000464, -0.00016445, 0.00021088, -0.00084877, 0.00072689, 0.00108534, -0.00078862, 0.00004393, 0.00000900, -0.00014016, 0.00011452, -0.00012336, 0.00002388, -0.00003705, -0.00004512, 0.00002804, -0.00001910, 0.00000436, 0.00001175, -0.00003159, 0.00002523, -0.00000375, -0.00001141, 0.00002704, },
		{ 0.00001580, 0.00000038, -0.00000979, 0.00001019, 0.00000060, 0.00002235, -0.00000965, -0.00000801, 0.00001043, -0.00003734, -0.00000856, 0.00001439, 0.00000420, 0.00001972, -0.00002666, -0.00000198, 0.00000260, -0.00002512, 0.00003354, -0.00003516, 0.00000878, 0.00002168, -0.00000338, -0.00002775, 0.00003155, 0.00001926, -0.00003895, 0.00001182, -0.00005339, 0.00012629, -0.00014480, 0.00007130, 0.00003393, -0.00013040, 0.00020103, -0.00018562, 0.00011317, -0.00010470, 0.00006148, -0.00009166, -0.00007074, 0.00011451, -0.00088888, 0.00108511, 0.00080595, -0.00086071, 0.00019105, -0.00009369, -0.00007526, 0.00009030, -0.00012728, 0.00004067, -0.00006127, -0.00001911, 0.00001940, -0.00002513, 0.00001502, 0.00000258, -0.00002637, 0.00003085, -0.00001454, -0.00000399, 0.00002627, -0.00001577, },
		{ -0.00000615, -0.00000500, 0.00001075, 0.00000172, 0.00002055, -0.00001348, -0.00000278, 0.00000356, -0.00003812, -0.00000040, 0.00001231, 0.00000686, 0.00001445, -0.00002968, 0.00000467, -0.00000668, -0.00001429, 0.00002735, -0.00003685, 0.00002051, 0.00001435, -0.00000551, -0.00002778, 0.00004235, 0.00000375, -0.00002627, -0.00000436, -0.00002639, 0.00010850, -0.00015897, 0.00012749, -0.00004372, -0.00006295, 0.00016047, -0.00018919, 0.00013668, -0.00014909, 0.00012707, -0.00018618, 0.00005133, -0.00004605, -0.00083233, 0.00140340, 0.00045248, -0.00083710, 0.00029511, -0.00019196, 0.00000356, 0.00004561, -0.00011252, 0.00004361, -0.00007735, 0.00000767, 0.00000268, -0.00002298, 0.00002218, -0.00000729, -0.00001743, 0.00003213, -0.00002413, 0.00000527, 0.00002302, -0.00001611, 0.00001818, },
		{ 0.00000100, 0.00000987, 0.00000298, 0.00001788, -0.00001580, 0.00000141, -0.00000409, -0.00003602, 0.00000635, 0.00001001, 0.00000960, 0.00000745, -0.00002960, 0.00000862, -0.00001408, -0.00000306, 0.00001794, -0.00003320, 0.00002834, 0.00000695, -0.00000742, -0.00002509, 0.00004953, -0.00000951, -0.00001512, -0.00001950, 0.00000492, 0.00007556, -0.00014893, 0.00016028, -0.00011224, 0.00001375, 0.00009593, -0.00016233, 0.00013073, -0.00016582, 0.00017047, -0.00027019, 0.00018340, -0.00025999, -0.00066544, 0.00164107, 0.00006053, -0.00072264, 0.00034056, -0.00026763, 0.00008263, -0.00001214, -0.00008099, 0.00003067, -0.00008036, 0.00002919, -0.00001803, -0.00001302, 0.00002373, -0.00001539, -0.00000629, 0.00002888, -0.00003117, 0.00001521, 0.00001775, -0.00001453, 0.00001842, -0.00001147, },
		{ 0.00000760, 0.00000456, 0.00001448, -0.00001664, 0.00000415, -0.00001169, -0.00003142, 0.00001124, 0.00000812, 0.00001167, -0.00000044, -0.00002681, 0.00000964, -0.00001857, 0.00000686, 0.00000712, -0.00002511, 0.00003164, 0.00000049, -0.00000912, -0.00002011, 0.00005286, -0.00001962, -0.00000719, -0.00003019, 0.00003483, 0.00003360, -0.00011727, 0.00016515, -0.00016030, 0.00008605, 0.00001867, -0.00011106, 0.00009554, -0.00014929, 0.00017958, -0.00032512, 0.00030153, -0.00050486, -0.00038851, 0.00176376, -0.00032755, -0.00053511, 0.00032159, -0.00030631, 0.00014760, -0.00007234, -0.00003849, 0.00000337, -0.00006779, 0.00003996, -0.00003737, 0.00000233, 0.00001899, -0.00001960, 0.00000504, 0.00002173, -0.00003477, 0.00002463, 0.00001117, -0.00001135, 0.00001637, -0.00001473, 0.00000734, },
		{ 0.00000652, 0.00001047, -0.00001613, 0.00000518, -0.00001844, -0.00002495, 0.00001412, 0.00000706, 0.00001243, -0.00000824, -0.00002207, 0.00000798, -0.00001972, 0.00001411, -0.00000316, -0.00001413, 0.00003047, -0.00000429, -0.00001063, -0.00001343, 0.00005264, -0.00002641, -0.00000307, -0.00003422, 0.00005805, -0.00000980, -0.00007037, 0.00014269, -0.00018064, 0.00014138, -0.00005785, -0.00004611, 0.00003744, -0.00010043, 0.00014802, -0.00033560, 0.00038005, -0.00074875, -0.00001753, 0.00174913, -0.00066865, -0.00030264, 0.00024325, -0.00030028, 0.00018631, -0.00012338, 0.00000633, -0.00003324, -0.00004045, 0.00003646, -0.00004985, 0.00001910, 0.00000897, -0.00001871, 0.00001450, 0.00001191, -0.00003465, 0.00003252, 0.00000411, -0.00000715, 0.00001225, -0.00001530, 0.00001299, 0.00000431, },
		{ 0.00000601, -0.00001444, 0.00000440, -0.00002362, -0.00001740, 0.00001517, 0.00000699, 0.00001140, -0.00001500, -0.00001636, 0.00000435, -0.00001766, 0.00001795, -0.00001124, -0.00000212, 0.00002551, -0.00000694, -0.00001202, -0.00000573, 0.00004952, -0.00003037, -0.00000229, -0.00003100, 0.00007084, -0.00004715, -0.00001707, 0.00009828, -0.00017126, 0.00017064, -0.00012057, 0.00001922, -0.00003222, -0.00002692, 0.00007700, -0.00029302, 0.00039628, -0.00095436, 0.00041636, 0.00159098, -0.00092523, -0.00005931, 0.00012047, -0.00024999, 0.00019136, -0.00015504, 0.00004373, -0.00007136, -0.00000262, 0.00001810, -0.00005119, 0.00003262, -0.00000381, -0.00001279, 0.00002049, 0.00000098, -0.00003106, 0.00003821, -0.00000261, -0.00000267, 0.00000666, -0.00001293, 0.00001698, 0.00000062, 0.00000876, },
		{ -0.00001187, 0.00000190, -0.00002670, -0.00000964, 0.00001480, 0.00000778, 0.00000843, -0.00001995, -0.00001067, -0.00000031, -0.00001308, 0.00001825, -0.00001593, 0.00000901, 0.00001794, -0.00000741, -0.00001335, 0.00000240, 0.00004424, -0.00003231, -0.00000353, -0.00002160, 0.00007164, -0.00007246, 0.00003306, 0.00004098, -0.00013566, 0.00016983, -0.00015941, 0.00007172, -0.00009935, 0.00005795, -0.00002421, -0.00019805, 0.00033512, -0.00108455, 0.00087019, 0.00130112, -0.00107091, 0.00016031, -0.00002464, -0.00016406, 0.00016172, -0.00016073, 0.00006496, -0.00010198, 0.00003855, -0.00001232, -0.00003941, 0.00003866, -0.00001586, -0.00000317, 0.00002224, -0.00000955, -0.00002466, 0.00004134, -0.00000837, 0.00000132, 0.00000047, -0.00000784, 0.00001853, -0.00000271, 0.00001098, 0.00000133, },
		{ -0.00000203, -0.00002736, -0.00000249, 0.00001359, 0.00000906, 0.00000365, -0.00002261, -0.00000587, -0.00000498, -0.00000696, 0.00001554, -0.00001673, 0.00001764, 0.00000921, -0.00000599, -0.00001465, 0.00001051, 0.00003748, -0.00003307, -0.00000508, -0.00000845, 0.00006133, -0.00008238, 0.00007161, -0.00001836, -0.00008198, 0.00014068, -0.00016920, 0.00010104, -0.00014992, 0.00013768, -0.00013948, -0.00006157, 0.00019294, -0.00110853, 0.00129458, 0.00090836, -0.00109444, 0.00032669, -0.00016684, -0.00005763, 0.00010304, -0.00013907, 0.00006432, -0.00011685, 0.00007433, -0.00004896, -0.00001547, 0.00003445, -0.00002356, 0.00000780, 0.00001990, -0.00001847, -0.00001633, 0.00004194, -0.00001276, 0.00000415, -0.00000525, -0.00000072, 0.00001721, -0.00000487, 0.00001271, -0.00000322, -0.00000877, },
		{ -0.00002557, 0.00000340, 0.00001215, 0.00001026, -0.00000248, -0.00002285, -0.00000255, -0.00000876, -0.00000044, 0.00001081, -0.00001384, 0.00002268, 0.00000082, -0.00000327, -0.00001588, 0.00001823, 0.00002972, -0.00003322, -0.00000533, 0.00000528, 0.00004291, -0.00007670, 0.00009285, -0.00006924, -0.00002119, 0.00009007, -0.00015053, 0.00010170, -0.00017290, 0.00019579, -0.00024853, 0.00009642, -0.00002055, -0.00100738, 0.00164056, 0.00045484, -0.00100115, 0.00041988, -0.00028228, 0.00005056, 0.00002661, -0.00009429, 0.00004063, -0.00011035, 0.00009628, -0.00008391, 0.00001665, 0.00001949, -0.00002407, 0.00001733, 0.00001453, -0.00002514, -0.00000697, 0.00004025, -0.00001563, 0.00000540, -0.00000944, 0.00000734, 0.00001309, -0.00000525, 0.00001340, -0.00000719, -0.00000560, -0.00000694, },
		{ 0.00000762, 0.00001104, 0.00001077, -0.00000927, -0.00002084, -0.00000101, -0.00001103, 0.00000544, 0.00000529, -0.00000810, 0.00002368, -0.00000592, -0.00000002, -0.00001688, 0.00002533, 0.00002125, -0.00003291, -0.00000325, 0.00001646, 0.00002079, -0.00005821, 0.00009471, -0.00010350, 0.00003511, 0.00002850, -0.00010945, 0.00007421, -0.00016265, 0.00021918, -0.00033064, 0.00024965, -0.00028242, -0.00077809, 0.00186671, -0.00000973, -0.00081188, 0.00043243, -0.00035274, 0.00014178, -0.00005283, -0.00003556, -0.00000227, -0.00008101, 0.00009804, -0.00010883, 0.00005061, -0.00000411, -0.00001608, 0.00002286, 0.00000776, -0.00002947, 0.00000270, 0.00003666, -0.00001709, 0.00000493, -0.00001124, 0.00001499, 0.00000676, -0.00000352, 0.00001257, -0.00001008, -0.00000291, -0.00001219, -0.00002159, },
		{ 0.00001062, 0.00001003, -0.00001592, -0.00001707, -0.00000123, -0.00001146, 0.00000991, 0.00000019, -0.00000075, 0.00002084, -0.00001009, 0.00000295, -0.00001739, 0.00003157, 0.00001220, -0.00003185, 0.00000136, 0.00002269, -0.00000025, -0.00003197, 0.00007896, -0.00011671, 0.00007701, -0.00003214, -0.00005606, 0.00002489, -0.00012012, 0.00020090, -0.00036870, 0.00036999, -0.00055928, -0.00043522, 0.00194533, -0.00043522, -0.00055928, 0.00036999, -0.00036870, 0.00020090, -0.00012012, 0.00002489, -0.00005606, -0.00003214, 0.00007701, -0.00011671, 0.00007896, -0.00003197, -0.00000025, 0.00002269, 0.00000136, -0.00003185, 0.00001220, 0.00003157, -0.00001739, 0.00000295, -0.00001009, 0.00002084, -0.00000075, 0.00000019, 0.00000991, -0.00001146, -0.00000123, -0.00001707, -0.00001592, 0.00001003, },
		{ 0.00000762, -0.00002159, -0.00001219, -0.00000291, -0.00001008, 0.00001257, -0.00000352, 0.00000676, 0.00001499, -0.00001124, 0.00000493, -0.00001709, 0.00003666, 0.00000270, -0.00002947, 0.00000776, 0.00002286, -0.00001608, -0.00000411, 0.00005061, -0.00010883, 0.00009804, -0.00008101, -0.00000227, -0.00003556, -0.00005283, 0.00014178, -0.00035274, 0.00043243, -0.00081188, -0.00000973, 0.00186671, -0.00077809, -0.00028242, 0.00024965, -0.00033064, 0.00021918, -0.00016265, 0.00007421, -0.00010945, 0.00002850, 0.00003511, -0.00010350, 0.00009471, -0.00005821, 0.00002079, 0.00001646, -0.00000325, -0.00003291, 0.00002125, 0.00002533, -0.00001688, -0.00000002, -0.00000592, 0.00002368, -0.00000810, 0.00000529, 0.00000544, -0.00001103, -0.00000101, -0.00002084, -0.00000927, 0.00001077, 0.00001104, },
		{ -0.00002557, -0.00000694, -0.00000560, -0.00000719, 0.00001340, -0.00000525, 0.00001309, 0.00000734, -0.00000944, 0.00000540, -0.00001563, 0.00004025, -0.00000697, -0.00002514, 0.00001453, 0.00001733, -0.00002407, 0.00001949, 0.00001665, -0.00008391, 0.00009628, -0.00011035, 0.00004063, -0.00009429, 0.00002661, 0.00005056, -0.00028228, 0.00041988, -0.00100115, 0.00045484, 0.00164056, -0.00100738, -0.00002055, 0.00009642, -0.00024853, 0.00019579, -0.00017290, 0.00010170, -0.00015053, 0.00009007, -0.00002119, -0.00006924, 0.00009285, -0.00007670, 0.00004291, 0.00000528, -0.00000533, -0.00003322, 0.00002972, 0.00001823, -0.00001588, -0.00000327, 0.00000082, 0.00002268, -0.00001384, 0.00001081, -0.00000044, -0.00000876, -0.00000255, -0.00002285, -0.00000248, 0.00001026, 0.00001215, 0.00000340, },
		{ -0.00000203, -0.00000877, -0.00000322, 0.00001271, -0.00000487, 0.00001721, -0.00000072, -0.00000525, 0.00000415, -0.00001276, 0.00004194, -0.00001633, -0.00001847, 0.00001990, 0.00000780, -0.00002356, 0.00003445, -0.00001547, -0.00004896, 0.00007433, -0.00011685, 0.00006432, -0.00013907, 0.00010304, -0.00005763, -0.00016684, 0.00032669, -0.00109444, 0.00090836, 0.00129458, -0.00110853, 0.00019294, -0.00006157, -0.00013948, 0.00013768, -0.00014992, 0.00010104, -0.00016920, 0.00014068, -0.00008198, -0.00001836, 0.00007161, -0.00008238, 0.00006133, -0.00000845, -0.00000508, -0.00003307, 0.00003748, 0.00001051, -0.00001465, -0.00000599, 0.00000921, 0.00001764, -0.00001673, 0.00001554, -0.00000696, -0.00000498, -0.00000587, -0.00002261, 0.00000365, 0.00000906, 0.00001359, -0.00000249, -0.00002736, },
		{ -0.00001187, 0.00000133, 0.00001098, -0.00000271, 0.00001853, -0.00000784, 0.00000047, 0.00000132, -0.00000837, 0.00004134, -0.00002466, -0.00000955, 0.00002224, -0.00000317, -0.00001586, 0.00003866, -0.00003941, -0.00001232, 0.00003855, -0.00010198, 0.00006496, -0.00016073, 0.00016172, -0.00016406, -0.00002464, 0.00016031, -0.00107091, 0.00130112, 0.00087019, -0.00108455, 0.00033512, -0.00019805, -0.00002421, 0.00005795, -0.00009935, 0.00007172, -0.00015941, 0.00016983, -0.00013566, 0.00004098, 0.00003306, -0.00007246, 0.00007164, -0.00002160, -0.00000353, -0.00003231, 0.00004424, 0.00000240, -0.00001335, -0.00000741, 0.00001794, 0.00000901, -0.00001593, 0.00001825, -0.00001308, -0.00000031, -0.00001067, -0.00001995, 0.00000843, 0.00000778, 0.00001480, -0.00000964, -0.00002670, 0.00000190, },
		{ 0.00000601, 0.00000876, 0.00000062, 0.00001698, -0.00001293, 0.00000666, -0.00000267, -0.00000261, 0.00003821, -0.00003106, 0.00000098, 0.00002049, -0.00001279, -0.00000381, 0.00003262, -0.00005119, 0.00001810, -0.00000262, -0.00007136, 0.00004373, -0.00015504, 0.00019136, -0.00024999, 0.00012047, -0.00005931, -0.00092523, 0.00159098, 0.00041636, -0.00095436, 0.00039628, -0.00029302, 0.00007700, -0.00002692, -0.00003222, 0.00001922, -0.00012057, 0.00017064, -0.00017126, 0.00009828, -0.00001707, -0.00004715, 0.00007084, -0.00003100, -0.00000229, -0.00003037, 0.00004952, -0.00000573, -0.00001202, -0.00000694, 0.00002551, -0.00000212, -0.00001124, 0.00001795, -0.00001766, 0.00000435, -0.00001636, -0.00001500, 0.00001140, 0.00000699, 0.00001517, -0.00001740, -0.00002362, 0.00000440, -0.00001444, },
		{ 0.00000652, 0.00000431, 0.00001299, -0.00001530, 0.00001225, -0.00000715, 0.00000411, 0.00003252, -0.00003465, 0.00001191, 0.00001450, -0.00001871, 0.00000897, 0.00001910, -0.00004985, 0.00003646, -0.00004045, -0.00003324, 0.00000633, -0.00012338, 0.00018631, -0.00030028, 0.00024325, -0.00030264, -0.00066865, 0.00174913, -0.00001753, -0.00074875, 0.00038005, -0.00033560, 0.00014802, -0.00010043, 0.00003744, -0.00004611, -0.00005785, 0.00014138, -0.00018064, 0.00014269, -0.00007037, -0.00000980, 0.00005805, -0.00003422, -0.00000307, -0.00002641, 0.00005264, -0.00001343, -0.00001063, -0.00000429, 0.00003047, -0.00001413, -0.00000316, 0.00001411, -0.00001972, 0.00000798, -0.00002207, -0.00000824, 0.00001243, 0.00000706, 0.00001412, -0.00002495, -0.00001844, 0.00000518, -0.00001613, 0.00001047, },
		{ 0.00000760, 0.00000734, -0.00001473, 0.00001637, -0.00001135, 0.00001117, 0.00002463, -0.00003477, 0.00002173, 0.00000504, -0.00001960, 0.00001899, 0.00000233, -0.00003737, 0.00003996, -0.00006779, 0.00000337, -0.00003849, -0.00007234, 0.00014760, -0.00030631, 0.00032159, -0.00053511, -0.00032755, 0.00176376, -0.00038851, -0.00050486, 0.00030153, -0.00032512, 0.00017958, -0.00014929, 0.00009554, -0.00011106, 0.00001867, 0.00008605, -0.00016030, 0.00016515, -0.00011727, 0.00003360, 0.00003483, -0.00003019, -0.00000719, -0.00001962, 0.00005286, -0.00002011, -0.00000912, 0.00000049, 0.00003164, -0.00002511, 0.00000712, 0.00000686, -0.00001857, 0.00000964, -0.00002681, -0.00000044, 0.00001167, 0.00000812, 0.00001124, -0.00003142, -0.00001169, 0.00000415, -0.00001664, 0.00001448, 0.00000456, },
		{ 0.00000100, -0.00001147, 0.00001842, -0.00001453, 0.00001775, 0.00001521, -0.00003117, 0.00002888, -0.00000629, -0.00001539, 0.00002373, -0.00001302, -0.00001803, 0.00002919, -0.00008036, 0.00003067, -0.00008099, -0.00001214, 0.00008263, -0.00026763, 0.00034056, -0.00072264, 0.00006053, 0.00164107, -0.00066544, -0.00025999, 0.00018340, -0.00027019, 0.00017047, -0.00016582, 0.00013073, -0.00016233, 0.00009593, 0.00001375, -0.00011224, 0.00016028, -0.00014893, 0.00007556, 0.00000492, -0.00001950, -0.00001512, -0.00000951, 0.00004953, -0.00002509, -0.00000742, 0.00000695, 0.00002834, -0.00003320, 0.00001794, -0.00000306, -0.00001408, 0.00000862, -0.00002960, 0.00000745, 0.00000960, 0.00001001, 0.00000635, -0.00003602, -0.00000409, 0.00000141, -0.00001580, 0.00001788, 0.00000298, 0.00000987, },
		{ -0.00000615, 0.00001818, -0.00001611, 0.00002302, 0.00000527, -0.00002413, 0.00003213, -0.00001743, -0.00000729, 0.00002218, -0.00002298, 0.00000268, 0.00000767, -0.00007735, 0.00004361, -0.00011252, 0.00004561, 0.00000356, -0.00019196, 0.00029511, -0.00083710, 0.00045248, 0.00140340, -0.00083233, -0.00004605, 0.00005133, -0.00018618, 0.00012707, -0.00014909, 0.00013668, -0.00018919, 0.00016047, -0.00006295, -0.00004372, 0.00012749, -0.00015897, 0.00010850, -0.00002639, -0.00000436, -0.00002627, 0.00000375, 0.00004235, -0.00002778, -0.00000551, 0.00001435, 0.00002051, -0.00003685, 0.00002735, -0.00001429, -0.00000668, 0.00000467, -0.00002968, 0.00001445, 0.00000686, 0.00001231, -0.00000040, -0.00003812, 0.00000356, -0.00000278, -0.00001348, 0.00002055, 0.00000172, 0.00001075, -0.00000500, },
		{ 0.00001580, -0.00001577, 0.00002627, -0.00000399, -0.00001454, 0.00003085, -0.00002637, 0.00000258, 0.00001502, -0.00002513, 0.00001940, -0.00001911, -0.00006127, 0.00004067, -0.00012728, 0.00009030, -0.00007526, -0.00009369, 0.00019105, -0.00086071, 0.00080595, 0.00108511, -0.00088888, 0.00011451, -0.00007074, -0.00009166, 0.00006148, -0.00010470, 0.00011317, -0.00018562, 0.00020103, -0.00013040, 0.00003393, 0.00007130, -0.00014480, 0.00012629, -0.00005339, 0.00001182, -0.00003895, 0.00001926, 0.00003155, -0.00002775, -0.00000338, 0.00002168, 0.00000878, -0.00003516, 0.00003354, -0.00002512, 0.00000260, -0.00000198, -0.00002666, 0.00001972, 0.00000420, 0.00001439, -0.00000856, -0.00003734, 0.00001043, -0.00000801, -0.00000965, 0.00002235, 0.00000060, 0.00001019, -0.00000979, 0.00000038, },
		{ -0.00001354, 0.00002704, -0.00001141, -0.00000375, 0.00002523, -0.00003159, 0.00001175, 0.00000436, -0.00001910, 0.00002804, -0.00004512, -0.00003705, 0.00002388, -0.00012336, 0.00011452, -0.00014016, 0.00000900, 0.00004393, -0.00078862, 0.00108534, 0.00072689, -0.00084877, 0.00021088, -0.00016445, -0.00000464, -0.00001133, -0.00004317, 0.00006589, -0.00015147, 0.00021080, -0.00017653, 0.00010743, 0.00000054, -0.00010817, 0.00012546, -0.00007108, 0.00002517, -0.00005072, 0.00003533, 0.00001798, -0.00002490, -0.00000110, 0.00002785, -0.00000558, -0.00002798, 0.00003508, -0.00003373, 0.00001233, -0.00001056, -0.00002059, 0.00002267, 0.00000225, 0.00001556, -0.00001740, -0.00003362, 0.00001575, -0.00001372, -0.00000435, 0.00002312, -0.00000057, 0.00000839, -0.00001274, 0.00000713, 0.00001177, },
		{ 0.00002526, -0.00001602, 0.00000665, 0.00001621, -0.00003242, 0.00001803, -0.00000679, -0.00000653, 0.00002658, -0.00006505, -0.00001071, -0.00000196, -0.00010283, 0.00011528, -0.00018086, 0.00009762, -0.00012367, -0.00062936, 0.00126635, 0.00036959, -0.00073580, 0.00024372, -0.00021939, 0.00006071, -0.00007671, 0.00002248, 0.00000508, -0.00009239, 0.00018870, -0.00019318, 0.00016406, -0.00007324, -0.00005490, 0.00010593, -0.00007619, 0.00003220, -0.00005886, 0.00004981, 0.00000310, -0.00001951, 0.00000128, 0.00003179, -0.00002080, -0.00001605, 0.00003126, -0.00003855, 0.00002088, -0.00001986, -0.00001199, 0.00002312, 0.00000146, 0.00001518, -0.00002601, -0.00002721, 0.00001890, -0.00001926, 0.00000224, 0.00002263, -0.00000189, 0.00000575, -0.00001352, 0.00001317, 0.00000679, -0.00000978, },
		{ -0.00001730, 0.00001507, 0.00000535, -0.00002915, 0.00002004, -0.00001537, 0.00000942, 0.00001539, -0.00007543, 0.00001202, -0.00003066, -0.00007101, 0.00009442, -0.00019229, 0.00015684, -0.00028616, -0.00040298, 0.00133858, 0.00004861, -0.00057903, 0.00022324, -0.00023366, 0.00009619, -0.00012308, 0.00007955, -0.00005675, -0.00001854, 0.00013952, -0.00017763, 0.00019402, -0.00013791, 0.00000619, 0.00007103, -0.00006791, 0.00003060, -0.00006099, 0.00006040, -0.00001123, -0.00001231, 0.00000370, 0.00003265, -0.00003492, -0.00000088, 0.00002220, -0.00003853, 0.00002671, -0.00002847, -0.00000182, 0.00002128, 0.00000198, 0.00001279, -0.00003344, -0.00001869, 0.00001948, -0.00002390, 0.00000978, 0.00002068, -0.00000337, 0.00000276, -0.00001215, 0.00001775, 0.00000169, -0.00000513, 0.00002126, },
		{ 0.00002030, -0.00000552, -0.00002301, 0.00001757, -0.00001896, 0.00002488, -0.00000314, -0.00007513, 0.00002696, -0.00005611, -0.00003499, 0.00005784, -0.00017533, 0.00017725, -0.00041925, -0.00013751, 0.00130571, -0.00021026, -0.00040767, 0.00016583, -0.00021290, 0.00010035, -0.00014394, 0.00011821, -0.00010743, 0.00005750, 0.00007285, -0.00013312, 0.00019224, -0.00018302, 0.00006496, 0.00002687, -0.00004806, 0.00001979, -0.00005565, 0.00006517, -0.00002298, -0.00000432, 0.00000609, 0.00002991, -0.00004598, 0.00001546, 0.00000889, -0.00003338, 0.00002867, -0.00003496, 0.00000872, 0.00001770, 0.00000359, 0.00000827, -0.00003877, -0.00000887, 0.00001734, -0.00002693, 0.00001774, 0.00001714, -0.00000486, -0.00000009, -0.00000891, 0.00002037, -0.00000276, -0.00000038, 0.00001576, -0.00001526, },
		{ -0.00001472, -0.00001577, 0.00001160, -0.00001638, 0.00003620, -0.00002522, -0.00006534, 0.00003217, -0.00007365, -0.00000199, 0.00001392, -0.00013624, 0.00015682, -0.00050394, 0.00013568, 0.00118358, -0.00039300, -0.00024674, 0.00009019, -0.00016798, 0.00007778, -0.00013848, 0.00013320, -0.00013754, 0.00012331, 0.00000113, -0.00006813, 0.00015933, -0.00020167, 0.00011185, -0.00001885, -0.00002070, 0.00000109, -0.00004272, 0.00006297, -0.00003039, 0.00000323, 0.00000843, 0.00002347, -0.00005237, 0.00003065, -0.00000694, -0.00002359, 0.00002620, -0.00003817, 0.00001838, 0.00001321, 0.00000579, 0.00000183, -0.00004134, 0.00000127, 0.00001263, -0.00002771, 0.00002542, 0.00001204, -0.00000614, -0.00000237, -0.00000431, 0.00002087, -0.00000593, 0.00000363, 0.00000977, -0.00001051, 0.00002169, },
		{ -0.00000933, 0.00000406, -0.00000800, 0.00004077, -0.00004666, -0.00004900, 0.00002807, -0.00008091, 0.00002227, -0.00002839, -0.00008508, 0.00010097, -0.00052930, 0.00038676, 0.00099653, -0.00049748, -0.00011410, 0.00001371, -0.00011201, 0.00003734, -0.00011107, 0.00012454, -0.00014211, 0.00016901, -0.00006292, 0.00000536, 0.00010130, -0.00019174, 0.00013964, -0.00005839, 0.00000866, -0.00002245, -0.00002357, 0.00005380, -0.00003226, 0.00000914, 0.00001066, 0.00001368, -0.00005305, 0.00004247, -0.00002313, -0.00001046, 0.00001946, -0.00003740, 0.00002605, 0.00000869, 0.00000785, -0.00000597, -0.00004075, 0.00001070, 0.00000580, -0.00002580, 0.00003199, 0.00000561, -0.00000694, -0.00000380, 0.00000102, 0.00001935, -0.00000746, 0.00000622, 0.00000439, -0.00000410, 0.00001930, -0.00002102, },
		{ -0.00000270, 0.00000446, 0.00003753, -0.00006384, -0.00002993, 0.00001701, -0.00007811, 0.00003463, -0.00006158, -0.00003333, 0.00002113, -0.00049375, 0.00059154, 0.00077293, -0.00053165, -0.00001915, -0.00005031, -0.00005754, -0.00001025, -0.00006972, 0.00009699, -0.00012127, 0.00018904, -0.00010873, 0.00007419, 0.00002833, -0.00015611, 0.00014473, -0.00008536, 0.00003421, -0.00004670, -0.00000084, 0.00003889, -0.00002829, 0.00001249, 0.00001270, 0.00000136, -0.00004773, 0.00004917, -0.00003738, 0.00000420, 0.00000932, -0.00003256, 0.00003095, 0.00000492, 0.00000901, -0.00001427, -0.00003701, 0.00001844, -0.00000241, -0.00002100, 0.00003660, -0.00000167, -0.00000706, -0.00000422, 0.00000639, 0.00001622, -0.00000730, 0.00000695, 0.00000064, 0.00000257, 0.00001391, -0.00002397, -0.00000519, },
		{ 0.00001831, 0.00002711, -0.00007439, -0.00001181, 0.00000242, -0.00006758, 0.00003490, -0.00008103, 0.00000849, -0.00006772, -0.00040457, 0.00073439, 0.00054063, -0.00051046, 0.00003674, -0.00009414, -0.00001418, -0.00005483, -0.00002396, 0.00005867, -0.00007984, 0.00018295, -0.00012986, 0.00012660, -0.00004723, -0.00010192, 0.00012763, -0.00009598, 0.00005099, -0.00006728, 0.00002200, 0.00002058, -0.00001913, 0.00001280, 0.00001437, -0.00001233, -0.00003691, 0.00004971, -0.00004767, 0.00001828, -0.00000278, -0.00002416, 0.00003278, 0.00000242, 0.00000859, -0.00002203, -0.00003049, 0.00002372, -0.00001106, -0.00001348, 0.00003849, -0.00000914, -0.00000638, -0.00000359, 0.00001114, 0.00001208, -0.00000570, 0.00000575, -0.00000077, 0.00000810, 0.00000681, -0.00002393, -0.00000409, -0.00000651, },
		{ 0.00001156, -0.00007759, 0.00000260, -0.00001218, -0.00005298, 0.00002554, -0.00008573, 0.00003296, -0.00014997, -0.00027599, 0.00080932, 0.00032324, -0.00045218, 0.00005859, -0.00011577, 0.00001268, -0.00008875, 0.00001737, 0.00001900, -0.00002597, 0.00015506, -0.00012500, 0.00015435, -0.00011321, -0.00003912, 0.00009272, -0.00008960, 0.00005584, -0.00008042, 0.00004133, 0.00000188, -0.00000638, 0.00001013, 0.00001544, -0.00002597, -0.00002183, 0.00004394, -0.00005254, 0.00002980, -0.00001508, -0.00001326, 0.00003173, 0.00000135, 0.00000620, -0.00002824, -0.00002194, 0.00002603, -0.00001908, -0.00000379, 0.00003712, -0.00001600, -0.00000493, -0.00000197, 0.00001471, 0.00000766, -0.00000319, 0.00000291, 0.00000043, 0.00001132, -0.00000048, -0.00002197, -0.00000583, -0.00000593, 0.00003059, },
		{ -0.00007424, 0.00001196, -0.00002411, -0.00003809, 0.00001066, -0.00007806, 0.00003701, -0.00021230, -0.00012624, 0.00081949, 0.00013758, -0.00037497, 0.00005528, -0.00011788, 0.00002206, -0.00010803, 0.00004788, -0.00001330, 0.00003082, 0.00011316, -0.00009793, 0.00015410, -0.00015989, 0.00002154, 0.00004720, -0.00006875, 0.00004800, -0.00008381, 0.00005411, -0.00001401, 0.00000770, 0.00000509, 0.00001558, -0.00003807, -0.00000428, 0.00003262, -0.00005137, 0.00003715, -0.00002578, -0.00000124, 0.00002842, 0.00000145, 0.00000181, -0.00003203, -0.00001230, 0.00002521, -0.00002542, 0.00000717, 0.00003234, -0.00002146, -0.00000289, 0.00000050, 0.00001668, 0.00000369, -0.00000047, -0.00000090, 0.00000402, 0.00001153, -0.00000652, -0.00001955, -0.00000925, -0.00000065, 0.00003879, -0.00000628, },
		{ 0.00001635, -0.00003199, -0.00002593, -0.00000517, -0.00006284, 0.00002217, -0.00024582, 0.00002576, 0.00077528, -0.00000721, -0.00029415, 0.00003699, -0.00010617, 0.00001664, -0.00011249, 0.00006458, -0.00003193, 0.00008164, 0.00006663, -0.00005638, 0.00012775, -0.00018152, 0.00007052, -0.00000041, -0.00003845, 0.00002915, -0.00007694, 0.00005851, -0.00002433, 0.00002057, -0.00000126, 0.00001445, -0.00004726, 0.00001362, 0.00001726, -0.00004441, 0.00003945, -0.00003335, 0.00001040, 0.00002375, 0.00000215, -0.00000416, -0.00003286, -0.00000269, 0.00002149, -0.00002919, 0.00001823, 0.00002440, -0.00002489, -0.00000050, 0.00000363, 0.00001683, 0.00000082, 0.00000168, -0.00000481, 0.00000928, 0.00000862, -0.00001032, -0.00001814, -0.00001259, 0.00000842, 0.00004138, -0.00002349, -0.00006623, },
		{ -0.00003593, -0.00001798, -0.00001822, -0.00004590, -0.00000622, -0.00024718, 0.00016337, 0.00069150, -0.00010922, -0.00022060, 0.00001292, -0.00008735, 0.00000136, -0.00010502, 0.00006809, -0.00003394, 0.00011992, 0.00002431, -0.00001018, 0.00008175, -0.00017719, 0.00010114, -0.00004198, -0.00000506, 0.00000296, -0.00006128, 0.00005427, -0.00002732, 0.00002989, -0.00000765, 0.00001172, -0.00005244, 0.00002974, -0.00000007, -0.00003278, 0.00003657, -0.00003678, 0.00002034, 0.00001870, 0.00000268, -0.00001095, -0.00003059, 0.00000584, 0.00001544, -0.00002976, 0.00002807, 0.00001400, -0.00002591, 0.00000198, 0.00000714, 0.00001514, -0.00000048, 0.00000257, -0.00000783, 0.00001510, 0.00000314, -0.00001151, -0.00001879, -0.00001391, 0.00001939, 0.00003819, -0.00003778, -0.00005586, 0.00001698, },
		{ -0.00001414, -0.00002645, -0.00003248, -0.00004048, -0.00021858, 0.00027447, 0.00058435, -0.00017222, -0.00016033, -0.00001007, -0.00006750, -0.00001821, -0.00009037, 0.00006189, -0.00001987, 0.00014238, -0.00000727, 0.00003088, 0.00002552, -0.00015055, 0.00011056, -0.00007120, 0.00002499, -0.00002573, -0.00003985, 0.00004278, -0.00002258, 0.00003391, -0.00001283, 0.00000722, -0.00005294, 0.00004222, -0.00001711, -0.00001818, 0.00002920, -0.00003574, 0.00002762, 0.00001413, 0.00000224, -0.00001752, -0.00002554, 0.00001241, 0.00000792, -0.00002695, 0.00003542, 0.00000223, -0.00002447, 0.00000434, 0.00001070, 0.00001184, -0.00000000, 0.00000167, -0.00000905, 0.00002023, -0.00000386, -0.00001044, -0.00002185, -0.00001165, 0.00002980, 0.00003029, -0.00004793, -0.00004516, 0.00001548, -0.00003708, },
		{ -0.00002984, -0.00002598, -0.00007251, -0.00016660, 0.00035247, 0.00046888, -0.00020350, -0.00011510, -0.00002814, -0.00005085, -0.00003727, -0.00007365, 0.00005116, 0.00000692, 0.00014915, -0.00002499, 0.00005902, -0.00003060, -0.00010873, 0.00009998, -0.00008465, 0.00004646, -0.00005194, -0.00001661, 0.00002680, -0.00001114, 0.00003188, -0.00001585, 0.00000097, -0.00004860, 0.00004971, -0.00003175, -0.00000270, 0.00001867, -0.00003055, 0.00003178, 0.00001059, 0.00000024, -0.00002275, -0.00001843, 0.00001642, -0.00000002, -0.00002098, 0.00003927, -0.00000963, -0.00002088, 0.00000651, 0.00001391, 0.00000740, 0.00000216, -0.00000121, -0.00000782, 0.00002344, -0.00001104, -0.00000809, -0.00002677, -0.00000505, 0.00003726, 0.00001970, -0.00005393, -0.00003536, 0.00001330, -0.00003706, -0.00001295, },
		{ -0.00002723, -0.00009556, -0.00010041, 0.00039627, 0.00035700, -0.00021169, -0.00008363, -0.00004011, -0.00003945, -0.00005272, -0.00005908, 0.00004139, 0.00004144, 0.00014310, -0.00002941, 0.00006995, -0.00007739, -0.00006069, 0.00007398, -0.00008216, 0.00005622, -0.00007156, 0.00000444, 0.00000986, 0.00000473, 0.00002409, -0.00001622, -0.00000672, -0.00003982, 0.00005155, -0.00004230, 0.00001161, 0.00000671, -0.00002216, 0.00003290, 0.00000823, -0.00000358, -0.00002564, -0.00001029, 0.00001764, -0.00000731, -0.00001257, 0.00003900, -0.00002021, -0.00001574, 0.00000854, 0.00001629, 0.00000247, 0.00000558, -0.00000593, -0.00000387, 0.00002379, -0.00001703, -0.00000577, -0.00003227, 0.00000553, 0.00004006, 0.00000880, -0.00005679, -0.00002673, 0.00001116, -0.00003723, -0.00001244, -0.00003013, },
		{ -0.00010553, -0.00002978, 0.00040918, 0.00025674, -0.00020501, -0.00006302, -0.00004676, -0.00003333, -0.00006345, -0.00004914, 0.00003720, 0.00007833, 0.00012859, -0.00002409, 0.00006336, -0.00010840, -0.00001529, 0.00003928, -0.00006646, 0.00005362, -0.00008210, 0.00002007, -0.00000446, 0.00002192, 0.00001188, -0.00001407, -0.00001525, -0.00002751, 0.00004781, -0.00004774, 0.00002300, -0.00000488, -0.00001186, 0.00003150, 0.00000682, -0.00000903, -0.00002550, -0.00000227, 0.00001625, -0.00001296, -0.00000278, 0.00003454, -0.00002836, -0.00000982, 0.00001058, 0.00001738, -0.00000214, 0.00000956, -0.00001196, 0.00000261, 0.00002078, -0.00002072, -0.00000486, -0.00003653, 0.00001868, 0.00003760, -0.00000038, -0.00005800, -0.00001866, 0.00000885, -0.00003804, -0.00001089, -0.00003000, -0.00003457, },
		{ 0.00003680, 0.00039733, 0.00017220, -0.00019010, -0.00005004, -0.00004979, -0.00003120, -0.00007013, -0.00004434, 0.00004157, 0.00011307, 0.00011012, -0.00001433, 0.00004258, -0.00012095, 0.00002032, 0.00000323, -0.00004235, 0.00004040, -0.00008304, 0.00002841, -0.00001324, 0.00003711, -0.00000268, -0.00001005, -0.00002377, -0.00001298, 0.00003928, -0.00004779, 0.00003030, -0.00001445, -0.00000113, 0.00002840, 0.00000578, -0.00001548, -0.00002211, 0.00000449, 0.00001278, -0.00001627, 0.00000708, 0.00002639, -0.00003326, -0.00000392, 0.00001282, 0.00001674, -0.00000560, 0.00001321, -0.00001846, 0.00001099, 0.00001446, -0.00002152, -0.00000636, -0.00003769, 0.00003211, 0.00003057, -0.00000663, -0.00005898, -0.00001016, 0.00000541, -0.00003888, -0.00000765, -0.00003205, -0.00004451, -0.00010150, },
		{ 0.00036793, 0.00010439, -0.00017164, -0.00004201, -0.00005096, -0.00003128, -0.00007449, -0.00004349, 0.00005565, 0.00014260, 0.00009106, -0.00000559, 0.00001347, -0.00011611, 0.00004190, -0.00002772, -0.00001548, 0.00002001, -0.00007573, 0.00002931, -0.00001483, 0.00004743, -0.00001715, -0.00000524, -0.00003126, 0.00000219, 0.00002728, -0.00004296, 0.00003301, -0.00002085, 0.00000869, 0.00002453, 0.00000440, -0.00002194, -0.00001575, 0.00000907, 0.00000802, -0.00001685, 0.00001564, 0.00001554, -0.00003458, 0.00000132, 0.00001538, 0.00001413, -0.00000714, 0.00001557, -0.00002437, 0.00002023, 0.00000543, -0.00001943, -0.00001068, -0.00003428, 0.00004321, 0.00002077, -0.00000992, -0.00006051, -0.00000044, -0.00000045, -0.00003818, -0.00000350, -0.00003776, -0.00005295, -0.00008546, 0.00009332, },
		{ 0.00005227, -0.00015248, -0.00003708, -0.00005150, -0.00003201, -0.00007861, -0.00004430, 0.00007906, 0.00016540, 0.00007303, -0.00000206, -0.00001718, -0.00009806, 0.00004850, -0.00004912, 0.00000886, -0.00000330, -0.00006290, 0.00002423, -0.00000914, 0.00005096, -0.00002927, -0.00000087, -0.00003670, 0.00001640, 0.00001348, -0.00003434, 0.00003139, -0.00002347, 0.00001656, 0.00002067, 0.00000197, -0.00002724, -0.00000723, 0.00001093, 0.00000293, -0.00001476, 0.00002172, 0.00000338, -0.00003250, 0.00000553, 0.00001823, 0.00000952, -0.00000625, 0.00001575, -0.00002856, 0.00002906, -0.00000524, -0.00001506, -0.00001743, -0.00002569, 0.00004963, 0.00001071, -0.00001131, -0.00006237, 0.00001061, -0.00000962, -0.00003400, -0.00000049, -0.00004688, -0.00005639, -0.00006133, 0.00013668, 0.00032780, },
		{ -0.00013423, -0.00003420, -0.00005197, -0.00003247, -0.00008424, -0.00004406, 0.00011045, 0.00018110, 0.00005579, -0.00000567, -0.00004323, -0.00007275, 0.00004217, -0.00005912, 0.00002677, -0.00002543, -0.00004789, 0.00001580, 0.00000239, 0.00004708, -0.00003732, 0.00000193, -0.00003923, 0.00002819, -0.00000037, -0.00002343, 0.00002627, -0.00002236, 0.00002191, 0.00001736, -0.00000193, -0.00003023, 0.00000230, 0.00000995, -0.00000153, -0.00001048, 0.00002443, -0.00000854, -0.00002771, 0.00000863, 0.00002113, 0.00000320, -0.00000273, 0.00001310, -0.00003000, 0.00003614, -0.00001624, -0.00000946, -0.00002549, -0.00001237, 0.00004983, 0.00000295, -0.00001266, -0.00006337, 0.00002202, -0.00002203, -0.00002484, -0.00000142, -0.00005728, -0.00005304, -0.00003375, 0.00016641, 0.00028243, 0.00001384, },
		{ -0.00003280, -0.00005238, -0.00003249, -0.00009249, -0.00004019, 0.00014811, 0.00018973, 0.00003782, -0.00001577, -0.00006053, -0.00004638, 0.00002701, -0.00005841, 0.00003627, -0.00004333, -0.00003381, 0.00000711, 0.00001712, 0.00003658, -0.00004043, 0.00000240, -0.00003829, 0.00003644, -0.00001264, -0.00001187, 0.00001891, -0.00001809, 0.00002462, 0.00001479, -0.00000741, -0.00003001, 0.00001144, 0.00000647, -0.00000458, -0.00000480, 0.00002342, -0.00001878, -0.00002120, 0.00001089, 0.00002360, -0.00000416, 0.00000317, 0.00000737, -0.00002795, 0.00004025, -0.00002616, -0.00000392, -0.00003320, 0.00000412, 0.00004355, -0.00000056, -0.00001601, -0.00006167, 0.00003182, -0.00003627, -0.00001037, -0.00000883, -0.00006546, -0.00004329, -0.00000671, 0.00018384, 0.00023578, -0.00001309, -0.00011778, },
		},
		{ { -0.00024969, 0.00000249, 0.00027973, 0.00011994, -0.00008768, -0.00003482, -0.00008364, -0.00003731, -0.00000421, 0.00001881, -0.00003133, 0.00002009, -0.00002741, 0.00004452, -0.00000839, -0.00003588, -0.00000847, 0.00002100, 0.00001634, -0.00002105, 0.00003144, -0.00003050, 0.00001593, -0.00002311, 0.00000773, -0.00000708, 0.00001497, -0.00002731, 0.00001842, -0.00001813, 0.00000366, 0.00001597, -0.00000923, 0.00001597, 0.00000366, -0.00001813, 0.00001842, -0.00002731, 0.00001497, -0.00000708, 0.00000773, -0.00002311, 0.00001593, -0.00003050, 0.00003144, -0.00002105, 0.00001634, 0.00002100, -0.00000847, -0.00003588, -0.00000839, 0.00004452, -0.00002741, 0.00002009, -0.00003133, 0.00001881, -0.00000421, -0.00003731, -0.00008364, -0.00003482, -0.00008768, 0.00011994, 0.00027973, 0.00000249, },
		{ 0.00001671, 0.00031910, 0.00010213, -0.00010542, -0.00003351, -0.00008120, -0.00004035, 0.00000487, 0.00000825, -0.00002576, 0.00001983, -0.00002725, 0.00004898, -0.00001674, -0.00003308, -0.00000795, 0.00002695, 0.00000696, -0.00001287, 0.00002556, -0.00002810, 0.00001486, -0.00002307, 0.00000685, -0.00000452, 0.00001177, -0.00002569, 0.00001944, -0.00002128, 0.00000916, 0.00001280, -0.00000878, 0.00001842, -0.00000193, -0.00001403, 0.00001609, -0.00002767, 0.00001736, -0.00000934, 0.00000854, -0.00002275, 0.00001620, -0.00003187, 0.00003634, -0.00002883, 0.00002647, 0.00001310, -0.00000752, -0.00003894, 0.00000075, 0.00003812, -0.00002548, 0.00001794, -0.00003404, 0.00002772, -0.00001394, -0.00003193, -0.00008922, -0.00003187, -0.00007060, 0.00013207, 0.00024194, -0.00000805, -0.00025118, },
		{ 0.00035962, 0.00007739, -0.00012189, -0.00002882, -0.00008200, -0.00004030, 0.00001211, -0.00000281, -0.00001800, 0.00001745, -0.00002511, 0.00005141, -0.00002416, -0.00003049, -0.00000612, 0.00003078, -0.00000119, -0.00000480, 0.00001907, -0.00002490, 0.00001307, -0.00002266, 0.00000600, -0.00000188, 0.00000802, -0.00002296, 0.00001913, -0.00002325, 0.00001426, 0.00000911, -0.00000747, 0.00001996, -0.00000730, -0.00000924, 0.00001256, -0.00002669, 0.00001877, -0.00001111, 0.00000918, -0.00002196, 0.00001560, -0.00003202, 0.00003991, -0.00003568, 0.00003678, 0.00000358, -0.00000510, -0.00004231, 0.00001050, 0.00002993, -0.00002148, 0.00001328, -0.00003335, 0.00003389, -0.00002299, -0.00002532, -0.00009732, -0.00002429, -0.00005582, 0.00013997, 0.00020574, -0.00001531, -0.00025577, 0.00003531, },
		{ 0.00004488, -0.00013508, -0.00002212, -0.00008562, -0.00003684, 0.00001656, -0.00001322, -0.00000887, 0.00001337, -0.00002120, 0.00005184, -0.00003059, -0.00002797, -0.00000321, 0.00003245, -0.00000773, 0.00000271, 0.00001237, -0.00002116, 0.00001070, -0.00002192, 0.00000527, 0.00000064, 0.00000397, -0.00001934, 0.00001756, -0.00002392, 0.00001866, 0.00000516, -0.00000539, 0.00002048, -0.00001215, -0.00000408, 0.00000804, -0.00002440, 0.00001907, -0.00001221, 0.00000954, -0.00002076, 0.00001413, -0.00003086, 0.00004186, -0.00004111, 0.00004661, -0.00000707, -0.00000130, -0.00004593, 0.00002063, 0.00002022, -0.00001559, 0.00000621, -0.00002901, 0.00003640, -0.00003001, -0.00001884, -0.00010688, -0.00001233, -0.00004453, 0.00014509, 0.00017085, -0.00001946, -0.00026375, 0.00005934, 0.00040034, },
		{ -0.00014314, -0.00001503, -0.00009118, -0.00003013, 0.00001762, -0.00002199, 0.00000076, 0.00000809, -0.00001586, 0.00005038, -0.00003596, -0.00002540, 0.00000043, 0.00003210, -0.00001245, 0.00000929, 0.00000580, -0.00001715, 0.00000793, -0.00002092, 0.00000475, 0.00000285, -0.00000009, -0.00001510, 0.00001490, -0.00002324, 0.00002214, 0.00000122, -0.00000271, 0.00001994, -0.00001620, 0.00000109, 0.00000278, -0.00002089, 0.00001820, -0.00001251, 0.00000953, -0.00001915, 0.00001183, -0.00002833, 0.00004200, -0.00004467, 0.00005532, -0.00001828, 0.00000363, -0.00004968, 0.00003084, 0.00000935, -0.00000810, -0.00000301, -0.00002102, 0.00003463, -0.00003375, -0.00001396, -0.00011642, 0.00000316, -0.00003734, 0.00014864, 0.00013691, -0.00002051, -0.00027556, 0.00009016, 0.00043969, 0.00000432, },
		{ -0.00000926, -0.00009750, -0.00002073, 0.00001509, -0.00002831, 0.00001009, 0.00000220, -0.00000948, 0.00004719, -0.00004027, -0.00002264, 0.00000443, 0.00002997, -0.00001528, 0.00001467, -0.00000032, -0.00001314, 0.00000491, -0.00001970, 0.00000446, 0.00000460, -0.00000392, -0.00001054, 0.00001136, -0.00002127, 0.00002452, -0.00000245, 0.00000039, 0.00001836, -0.00001925, 0.00000596, -0.00000290, -0.00001637, 0.00001620, -0.00001194, 0.00000910, -0.00001719, 0.00000881, -0.00002449, 0.00004022, -0.00004600, 0.00006224, -0.00002939, 0.00000934, -0.00005338, 0.00004079, -0.00000226, 0.00000059, -0.00001390, -0.00000969, 0.00002826, -0.00003322, -0.00001216, -0.00012425, 0.00002077, -0.00003421, 0.00015147, 0.00010362, -0.00001851, -0.00029161, 0.00012928, 0.00047554, -0.00004384, -0.00014453, },
		{ -0.00010314, -0.00000960, 0.00000915, -0.00003168, 0.00001838, -0.00000375, -0.00000250, 0.00004252, -0.00004350, -0.00001959, 0.00000840, 0.00002644, -0.00001628, 0.00001868, -0.00000577, -0.00000935, 0.00000182, -0.00001833, 0.00000445, 0.00000577, -0.00000726, -0.00000597, 0.00000724, -0.00001813, 0.00002571, -0.00000562, 0.00000366, 0.00001585, -0.00002117, 0.00001019, -0.00000864, -0.00001108, 0.00001317, -0.00001046, 0.00000819, -0.00001494, 0.00000520, -0.00001948, 0.00003655, -0.00004489, 0.00006683, -0.00003971, 0.00001541, -0.00005677, 0.00005010, -0.00001412, 0.00000997, -0.00002585, 0.00000440, 0.00001738, -0.00002777, -0.00001468, -0.00012862, 0.00003870, -0.00003447, 0.00015394, 0.00007092, -0.00001368, -0.00031213, 0.00017827, 0.00050526, -0.00009837, -0.00013829, -0.00000645, },
		{ 0.00000205, 0.00000036, -0.00003187, 0.00002507, -0.00000922, 0.00000462, 0.00003664, -0.00004567, -0.00001621, 0.00001198, 0.00002192, -0.00001561, 0.00002127, -0.00001038, -0.00000597, -0.00000118, -0.00001686, 0.00000469, 0.00000630, -0.00000993, -0.00000169, 0.00000284, -0.00001403, 0.00002569, -0.00000811, 0.00000687, 0.00001256, -0.00002190, 0.00001354, -0.00001409, -0.00000534, 0.00000928, -0.00000813, 0.00000681, -0.00001247, 0.00000119, -0.00001351, 0.00003111, -0.00004123, 0.00006863, -0.00004857, 0.00002135, -0.00005955, 0.00005838, -0.00002570, 0.00001949, -0.00003814, 0.00002042, 0.00000247, -0.00001716, -0.00002247, -0.00012786, 0.00005488, -0.00003691, 0.00015588, 0.00003911, -0.00000652, -0.00033702, 0.00023852, 0.00052587, -0.00015726, -0.00012410, -0.00000791, -0.00010667, },
		{ -0.00001040, -0.00002901, 0.00002980, -0.00001378, 0.00001146, 0.00002983, -0.00004678, -0.00001248, 0.00001483, 0.00001686, -0.00001356, 0.00002248, -0.00001407, -0.00000315, -0.00000396, -0.00001532, 0.00000517, 0.00000618, -0.00001181, 0.00000205, -0.00000151, -0.00000925, 0.00002452, -0.00000977, 0.00000979, 0.00000872, -0.00002145, 0.00001580, -0.00001892, 0.00000053, 0.00000477, -0.00000503, 0.00000498, -0.00000989, -0.00000302, -0.00000687, 0.00002417, -0.00003512, 0.00006734, -0.00005538, 0.00002665, -0.00006141, 0.00006523, -0.00003649, 0.00002857, -0.00004995, 0.00003733, -0.00001563, -0.00000164, -0.00003601, -0.00012062, 0.00006721, -0.00003985, 0.00015659, 0.00000889, 0.00000202, -0.00036568, 0.00031101, 0.00053424, -0.00021780, -0.00010244, -0.00001454, -0.00010677, 0.00001286, },
		{ -0.00002350, 0.00003241, -0.00001713, 0.00001762, 0.00002240, -0.00004684, -0.00000847, 0.00001674, 0.00001171, -0.00001044, 0.00002245, -0.00001685, -0.00000094, -0.00000641, -0.00001376, 0.00000581, 0.00000544, -0.00001283, 0.00000506, -0.00000551, -0.00000407, 0.00002234, -0.00001055, 0.00001220, 0.00000455, -0.00001993, 0.00001686, -0.00002284, 0.00000617, -0.00000007, -0.00000132, 0.00000276, -0.00000728, -0.00000719, 0.00000013, 0.00001609, -0.00002679, 0.00006285, -0.00005963, 0.00003083, -0.00006204, 0.00007027, -0.00004598, 0.00003666, -0.00006051, 0.00005398, -0.00003576, 0.00001807, -0.00005526, -0.00010598, 0.00007372, -0.00004139, 0.00015495, -0.00001863, 0.00001056, -0.00039690, 0.00039614, 0.00052739, -0.00027675, -0.00007451, -0.00002663, -0.00010240, 0.00002147, -0.00002204, },
		{ 0.00003300, -0.00001908, 0.00002282, 0.00001464, -0.00004585, -0.00000431, 0.00001757, 0.00000687, -0.00000663, 0.00002138, -0.00001879, 0.00000064, -0.00000845, -0.00001219, 0.00000654, 0.00000419, -0.00001299, 0.00000720, -0.00000890, 0.00000118, 0.00001934, -0.00001045, 0.00001395, 0.00000032, -0.00001751, 0.00001670, -0.00002565, 0.00001127, -0.00000496, 0.00000280, 0.00000023, -0.00000477, -0.00001110, 0.00000711, 0.00000730, -0.00001663, 0.00005521, -0.00006099, 0.00003348, -0.00006116, 0.00007319, -0.00005371, 0.00004325, -0.00006905, 0.00006917, -0.00005650, 0.00004083, -0.00007964, -0.00008361, 0.00007284, -0.00003959, 0.00014954, -0.00004195, 0.00001723, -0.00042876, 0.00049348, 0.00050272, -0.00033053, -0.00004225, -0.00004383, -0.00009298, 0.00002669, -0.00003336, -0.00001599, },
		{ -0.00001961, 0.00002681, 0.00000683, -0.00004383, -0.00000020, 0.00001728, 0.00000264, -0.00000246, 0.00001949, -0.00002002, 0.00000164, -0.00001005, -0.00001063, 0.00000729, 0.00000252, -0.00001236, 0.00000841, -0.00001146, 0.00000619, 0.00001576, -0.00000953, 0.00001491, -0.00000371, -0.00001441, 0.00001536, -0.00002718, 0.00001555, -0.00000958, 0.00000711, -0.00000251, -0.00000244, -0.00001455, 0.00001374, -0.00000170, -0.00000515, 0.00004470, -0.00005929, 0.00003428, -0.00005855, 0.00007373, -0.00005929, 0.00004791, -0.00007495, 0.00008176, -0.00007633, 0.00006513, -0.00010798, -0.00005384, 0.00006351, -0.00003270, 0.00013882, -0.00005940, 0.00001979, -0.00045861, 0.00060164, 0.00045828, -0.00037546, -0.00000816, -0.00006518, -0.00007841, 0.00002762, -0.00004315, -0.00000729, 0.00003181, },
		{ 0.00002946, -0.00000079, -0.00004081, 0.00000365, 0.00001595, -0.00000075, 0.00000175, 0.00001702, -0.00002068, 0.00000214, -0.00001120, -0.00000910, 0.00000797, 0.00000059, -0.00001104, 0.00000871, -0.00001307, 0.00001070, 0.00001186, -0.00000791, 0.00001502, -0.00000732, -0.00001089, 0.00001300, -0.00002739, 0.00001880, -0.00001365, 0.00001138, -0.00000532, -0.00000038, -0.00001736, 0.00001965, -0.00001040, 0.00000702, 0.00003175, -0.00005454, 0.00003303, -0.00005410, 0.00007174, -0.00006245, 0.00005035, -0.00007773, 0.00009073, -0.00009368, 0.00008921, -0.00013868, -0.00001768, 0.00004537, -0.00001935, 0.00012136, -0.00006919, 0.00001572, -0.00048319, 0.00071819, 0.00039308, -0.00040813, 0.00002487, -0.00008907, -0.00005918, 0.00002374, -0.00005037, 0.00000169, 0.00002927, -0.00001884, },
		{ -0.00000798, -0.00003683, 0.00000700, 0.00001377, -0.00000317, 0.00000573, 0.00001420, -0.00002096, 0.00000227, -0.00001194, -0.00000759, 0.00000849, -0.00000145, -0.00000921, 0.00000820, -0.00001365, 0.00001446, 0.00000791, -0.00000577, 0.00001430, -0.00001032, -0.00000723, 0.00000981, -0.00002632, 0.00002086, -0.00001693, 0.00001537, -0.00000808, 0.00000133, -0.00001938, 0.00002457, -0.00001832, 0.00001921, 0.00001696, -0.00004695, 0.00002969, -0.00004779, 0.00006717, -0.00006302, 0.00005039, -0.00007710, 0.00009530, -0.00010712, 0.00011123, -0.00016973, 0.00002322, 0.00001882, 0.00000124, 0.00009603, -0.00006971, 0.00000244, -0.00049870, 0.00083964, 0.00030720, -0.00042559, 0.00005380, -0.00011348, -0.00003632, 0.00001500, -0.00005421, 0.00001007, 0.00002592, -0.00001699, 0.00003076, },
		{ -0.00003198, 0.00000960, 0.00001097, -0.00000461, 0.00000931, 0.00001122, -0.00002102, 0.00000217, -0.00001231, -0.00000609, 0.00000878, -0.00000345, -0.00000704, 0.00000700, -0.00001324, 0.00001733, 0.00000416, -0.00000331, 0.00001282, -0.00001257, -0.00000368, 0.00000605, -0.00002408, 0.00002169, -0.00001925, 0.00001886, -0.00001067, 0.00000265, -0.00002054, 0.00002826, -0.00002500, 0.00003075, 0.00000106, -0.00003689, 0.00002436, -0.00003971, 0.00006009, -0.00006094, 0.00004800, -0.00007301, 0.00009495, -0.00011541, 0.00012934, -0.00019890, 0.00006665, -0.00001498, 0.00002913, 0.00006222, -0.00005970, -0.00002245, -0.00050106, 0.00096152, 0.00020201, -0.00042570, 0.00007569, -0.00013603, -0.00001136, 0.00000188, -0.00005419, 0.00001704, 0.00002232, -0.00001439, 0.00003073, -0.00001454, },
		{ 0.00001123, 0.00000786, -0.00000515, 0.00001239, 0.00000820, -0.00002098, 0.00000197, -0.00001238, -0.00000461, 0.00000879, -0.00000527, -0.00000472, 0.00000531, -0.00001193, 0.00001921, 0.00000081, -0.00000075, 0.00001068, -0.00001399, -0.00000050, 0.00000200, -0.00002086, 0.00002132, -0.00002047, 0.00002168, -0.00001297, 0.00000354, -0.00002080, 0.00003057, -0.00003008, 0.00004095, -0.00001518, -0.00002491, 0.00001728, -0.00003009, 0.00005072, -0.00005631, 0.00004330, -0.00006562, 0.00008950, -0.00011765, 0.00014190, -0.00022391, 0.00011002, -0.00005411, 0.00006363, 0.00001999, -0.00003848, -0.00006095, -0.00048613, 0.00107856, 0.00008012, -0.00040726, 0.00008797, -0.00015428, 0.00001379, -0.00001469, -0.00005022, 0.00002201, 0.00001902, -0.00001140, 0.00002952, -0.00002033, -0.00002637, },
		{ 0.00000476, -0.00000496, 0.00001495, 0.00000523, -0.00002094, 0.00000181, -0.00001222, -0.00000316, 0.00000850, -0.00000680, -0.00000243, 0.00000335, -0.00000987, 0.00002009, -0.00000196, 0.00000171, 0.00000807, -0.00001457, 0.00000213, -0.00000205, -0.00001690, 0.00001984, -0.00002056, 0.00002370, -0.00001488, 0.00000400, -0.00002017, 0.00003143, -0.00003329, 0.00004925, -0.00003095, -0.00001163, 0.00000883, -0.00001927, 0.00003938, -0.00004932, 0.00003652, -0.00005531, 0.00007910, -0.00011333, 0.00014761, -0.00024263, 0.00015050, -0.00009607, 0.00010323, -0.00002985, -0.00000610, -0.00011438, -0.00045009, 0.00118493, -0.00005459, -0.00037027, 0.00008868, -0.00016590, 0.00003701, -0.00003334, -0.00004264, 0.00002463, 0.00001647, -0.00000840, 0.00002729, -0.00002519, -0.00002015, 0.00001176, },
		{ -0.00000423, 0.00001702, 0.00000231, -0.00002093, 0.00000177, -0.00001191, -0.00000173, 0.00000789, -0.00000796, -0.00000034, 0.00000131, -0.00000727, 0.00002002, -0.00000406, 0.00000387, 0.00000516, -0.00001436, 0.00000406, -0.00000583, -0.00001250, 0.00001744, -0.00001956, 0.00002482, -0.00001633, 0.00000406, -0.00001874, 0.00003087, -0.00003449, 0.00005516, -0.00004546, 0.00000222, -0.00000052, -0.00000768, 0.00002650, -0.00004031, 0.00002804, -0.00004264, 0.00006425, -0.00010238, 0.00014559, -0.00025321, 0.00018525, -0.00013793, 0.00014573, -0.00008569, 0.00003650, -0.00018314, -0.00038972, 0.00127458, -0.00019717, -0.00031591, 0.00007665, -0.00016891, 0.00005613, -0.00005239, -0.00003216, 0.00002482, 0.00001500, -0.00000574, 0.00002425, -0.00002902, -0.00001355, 0.00001108, 0.00000196, },
		{ 0.00001871, -0.00000059, -0.00002094, 0.00000193, -0.00001150, -0.00000034, 0.00000699, -0.00000871, 0.00000145, -0.00000059, -0.00000435, 0.00001912, -0.00000544, 0.00000558, 0.00000216, -0.00001347, 0.00000522, -0.00000910, -0.00000793, 0.00001436, -0.00001758, 0.00002502, -0.00001729, 0.00000376, -0.00001662, 0.00002901, -0.00003368, 0.00005837, -0.00005802, 0.00001591, -0.00001023, 0.00000416, 0.00001260, -0.00002969, 0.00001829, -0.00002834, 0.00004573, -0.00008520, 0.00013549, -0.00025429, 0.00021168, -0.00017654, 0.00018836, -0.00014513, 0.00008744, -0.00026655, -0.00030272, 0.00134161, -0.00034184, -0.00024657, 0.00005166, -0.00016185, 0.00006915, -0.00007000, -0.00001984, 0.00002280, 0.00001476, -0.00000369, 0.00002062, -0.00003172, -0.00000682, 0.00000922, -0.00000026, -0.00000320, },
		{ -0.00000357, -0.00002091, 0.00000230, -0.00001104, 0.00000098, 0.00000584, -0.00000905, 0.00000285, -0.00000220, -0.00000134, 0.00001757, -0.00000613, 0.00000672, -0.00000074, -0.00001203, 0.00000558, -0.00001168, -0.00000348, 0.00001085, -0.00001478, 0.00002435, -0.00001775, 0.00000317, -0.00001396, 0.00002603, -0.00003097, 0.00005870, -0.00006805, 0.00002874, -0.00001971, 0.00001572, -0.00000172, -0.00001796, 0.00000779, -0.00001320, 0.00002460, -0.00006261, 0.00011752, -0.00024509, 0.00022760, -0.00020879, 0.00022800, -0.00020518, 0.00014396, -0.00036274, -0.00018804, 0.00138065, -0.00048229, -0.00016572, 0.00001451, -0.00014401, 0.00007441, -0.00008430, -0.00000698, 0.00001907, 0.00001571, -0.00000242, 0.00001661, -0.00003322, -0.00000024, 0.00000627, -0.00000173, -0.00000208, 0.00002013, },
		{ -0.00002071, 0.00000285, -0.00001056, 0.00000222, 0.00000449, -0.00000902, 0.00000383, -0.00000338, 0.00000153, 0.00001555, -0.00000620, 0.00000726, -0.00000336, -0.00001024, 0.00000522, -0.00001345, 0.00000060, 0.00000718, -0.00001137, 0.00002289, -0.00001773, 0.00000237, -0.00001096, 0.00002218, -0.00002660, 0.00005615, -0.00007512, 0.00004008, -0.00002842, 0.00002643, -0.00001584, -0.00000565, -0.00000294, 0.00000191, 0.00000207, -0.00003581, 0.00009246, -0.00022554, 0.00023146, -0.00023184, 0.00026138, -0.00026234, 0.00020247, -0.00046859, -0.00004605, 0.00138724, -0.00061205, -0.00007771, -0.00003302, -0.00011547, 0.00007077, -0.00009359, 0.00000497, 0.00001437, 0.00001760, -0.00000199, 0.00001239, -0.00003346, 0.00000585, 0.00000243, -0.00000234, -0.00000108, 0.00002140, -0.00000671, },
		{ 0.00000352, -0.00001007, 0.00000333, 0.00000303, -0.00000866, 0.00000443, -0.00000407, 0.00000408, 0.00001328, -0.00000577, 0.00000718, -0.00000555, -0.00000827, 0.00000424, -0.00001436, 0.00000411, 0.00000363, -0.00000760, 0.00002075, -0.00001728, 0.00000145, -0.00000781, 0.00001775, -0.00002091, 0.00005092, -0.00007899, 0.00004942, -0.00003585, 0.00003578, -0.00002915, 0.00000668, -0.00001339, 0.00001617, -0.00002054, -0.00000631, 0.00006160, -0.00019625, 0.00022247, -0.00024340, 0.00028540, -0.00031294, 0.00025877, -0.00057984, 0.00012129, 0.00135822, -0.00072489, 0.00001249, -0.00008823, -0.00007716, 0.00005774, -0.00009648, 0.00001458, 0.00000959, 0.00002002, -0.00000232, 0.00000809, -0.00003241, 0.00001114, -0.00000200, -0.00000210, -0.00000031, 0.00002263, -0.00001012, -0.00002022, },
		{ -0.00000955, 0.00000427, 0.00000152, -0.00000808, 0.00000470, -0.00000427, 0.00000616, 0.00001096, -0.00000501, 0.00000654, -0.00000723, -0.00000631, 0.00000279, -0.00001442, 0.00000692, 0.00000043, -0.00000373, 0.00001812, -0.00001648, 0.00000051, -0.00000469, 0.00001307, -0.00001432, 0.00004332, -0.00007961, 0.00005638, -0.00004160, 0.00004330, -0.00004107, 0.00001849, -0.00002308, 0.00002883, -0.00004196, 0.00002422, 0.00002666, -0.00015855, 0.00020066, -0.00024190, 0.00029736, -0.00035336, 0.00030830, -0.00069122, 0.00031039, 0.00129196, -0.00081523, 0.00009962, -0.00014772, -0.00003087, 0.00003556, -0.00009203, 0.00002054, 0.00000569, 0.00002245, -0.00000324, 0.00000380, -0.00003007, 0.00001528, -0.00000667, -0.00000112, 0.00000018, 0.00002387, -0.00001388, -0.00001930, 0.00000423, },
		{ 0.00000503, 0.00000004, -0.00000736, 0.00000473, -0.00000401, 0.00000769, 0.00000877, -0.00000408, 0.00000545, -0.00000835, -0.00000452, 0.00000106, -0.00001372, 0.00000895, -0.00000223, 0.00000002, 0.00001516, -0.00001540, -0.00000038, -0.00000179, 0.00000843, -0.00000729, 0.00003382, -0.00007711, 0.00006076, -0.00004537, 0.00004865, -0.00005109, 0.00002927, -0.00003163, 0.00003926, -0.00006096, 0.00005404, -0.00001035, -0.00011436, 0.00016696, -0.00022662, 0.00029523, -0.00038032, 0.00034642, -0.00079674, 0.00051612, 0.00118860, -0.00087851, 0.00017848, -0.00020757, 0.00002092, 0.00000521, -0.00007989, 0.00002180, 0.00000362, 0.00002425, -0.00000445, -0.00000044, -0.00002645, 0.00001800, -0.00001118, 0.00000040, 0.00000041, 0.00002510, -0.00001801, -0.00001781, 0.00000485, -0.00000899, },
		{ -0.00000134, -0.00000659, 0.00000464, -0.00000339, 0.00000865, 0.00000683, -0.00000314, 0.00000403, -0.00000891, -0.00000305, -0.00000077, -0.00001239, 0.00001020, -0.00000423, 0.00000341, 0.00001207, -0.00001416, -0.00000112, 0.00000074, 0.00000412, -0.00000028, 0.00002297, -0.00007179, 0.00006253, -0.00004703, 0.00005158, -0.00005879, 0.00003860, -0.00003873, 0.00004702, -0.00007653, 0.00008147, -0.00004724, -0.00006608, 0.00012311, -0.00019785, 0.00027783, -0.00039122, 0.00036880, -0.00088999, 0.00073198, 0.00105010, -0.00091148, 0.00024430, -0.00026365, 0.00007511, -0.00003160, -0.00006031, 0.00001768, 0.00000419, 0.00002480, -0.00000560, -0.00000463, -0.00002161, 0.00001904, -0.00001510, 0.00000219, 0.00000049, 0.00002624, -0.00002248, -0.00001566, 0.00000527, -0.00000834, 0.00000556, },
		{ -0.00000585, 0.00000455, -0.00000253, 0.00000908, 0.00000525, -0.00000234, 0.00000244, -0.00000897, -0.00000197, -0.00000252, -0.00001060, 0.00001073, -0.00000551, 0.00000629, 0.00000901, -0.00001285, -0.00000166, 0.00000278, 0.00000038, 0.00000626, 0.00001138, -0.00006409, 0.00006182, -0.00004656, 0.00005202, -0.00006391, 0.00004612, -0.00004418, 0.00005186, -0.00008790, 0.00010502, -0.00008183, -0.00001644, 0.00007156, -0.00015679, 0.00024496, -0.00038430, 0.00037171, -0.00096452, 0.00095039, 0.00088028, -0.00091248, 0.00029307, -0.00031192, 0.00012831, -0.00007257, -0.00003418, 0.00000796, 0.00000801, 0.00002352, -0.00000631, -0.00000881, -0.00001564, 0.00001824, -0.00001803, 0.00000393, 0.00000060, 0.00002716, -0.00002719, -0.00001281, 0.00000540, -0.00000756, 0.00000584, -0.00000257, },
		{ 0.00000457, -0.00000159, 0.00000904, 0.00000404, -0.00000180, 0.00000084, -0.00000862, -0.00000132, -0.00000403, -0.00000854, 0.00001067, -0.00000607, 0.00000853, 0.00000614, -0.00001155, -0.00000194, 0.00000426, -0.00000262, 0.00001192, -0.00000032, -0.00005456, 0.00005889, -0.00004411, 0.00005002, -0.00006628, 0.00005159, -0.00004788, 0.00005372, -0.00009459, 0.00012350, -0.00011212, 0.00003173, 0.00001532, -0.00010559, 0.00019750, -0.00035892, 0.00035237, -0.00101432, 0.00116308, 0.00068457, -0.00088149, 0.00032182, -0.00034874, 0.00017697, -0.00011494, -0.00000300, -0.00000713, 0.00001540, 0.00001996, -0.00000620, -0.00001303, -0.00000868, 0.00001551, -0.00001963, 0.00000527, 0.00000097, 0.00002764, -0.00003196, -0.00000927, 0.00000515, -0.00000663, 0.00000585, -0.00000361, -0.00000522, },
		{ -0.00000071, 0.00000867, 0.00000320, -0.00000159, -0.00000064, -0.00000799, -0.00000109, -0.00000518, -0.00000639, 0.00001017, -0.00000598, 0.00001008, 0.00000358, -0.00001034, -0.00000195, 0.00000513, -0.00000475, 0.00001640, -0.00001154, -0.00004381, 0.00005413, -0.00003995, 0.00004579, -0.00006591, 0.00005484, -0.00004986, 0.00005276, -0.00009647, 0.00013606, -0.00013640, 0.00007565, -0.00004228, -0.00004715, 0.00013736, -0.00031560, 0.00030919, -0.00103415, 0.00136152, 0.00046986, -0.00082022, 0.00032881, -0.00037113, 0.00021775, -0.00015568, 0.00003125, -0.00002682, 0.00002635, 0.00001383, -0.00000497, -0.00001734, -0.00000091, 0.00001089, -0.00001961, 0.00000588, 0.00000185, 0.00002744, -0.00003655, -0.00000513, 0.00000447, -0.00000553, 0.00000560, -0.00000446, -0.00000471, 0.00000477, },
		{ 0.00000811, 0.00000265, -0.00000173, -0.00000186, -0.00000721, -0.00000123, -0.00000589, -0.00000432, 0.00000939, -0.00000536, 0.00001092, 0.00000141, -0.00000927, -0.00000168, 0.00000541, -0.00000597, 0.00001947, -0.00002175, -0.00003246, 0.00004803, -0.00003447, 0.00003966, -0.00006294, 0.00005587, -0.00005022, 0.00004931, -0.00009369, 0.00014231, -0.00015341, 0.00011286, -0.00009778, 0.00001506, 0.00006741, -0.00025609, 0.00024200, -0.00101998, 0.00153740, 0.00024405, -0.00073201, 0.00031371, -0.00037701, 0.00024769, -0.00019174, 0.00006623, -0.00004993, 0.00004051, 0.00000507, -0.00000237, -0.00002176, 0.00000742, 0.00000450, -0.00001781, 0.00000549, 0.00000346, 0.00002629, -0.00004065, -0.00000056, 0.00000339, -0.00000425, 0.00000511, -0.00000512, -0.00000434, 0.00000522, -0.00000004, },
		{ 0.00000229, -0.00000219, -0.00000275, -0.00000642, -0.00000163, -0.00000615, -0.00000247, 0.00000852, -0.00000437, 0.00001113, -0.00000033, -0.00000838, -0.00000116, 0.00000514, -0.00000633, 0.00002102, -0.00003050, -0.00002113, 0.00004109, -0.00002810, 0.00003208, -0.00005765, 0.00005472, -0.00004914, 0.00004384, -0.00008673, 0.00014229, -0.00016240, 0.00014133, -0.00014789, 0.00007726, -0.00000876, -0.00018331, 0.00015211, -0.00096929, 0.00168309, 0.00001570, -0.00062157, 0.00027756, -0.00036536, 0.00026452, -0.00022028, 0.00009938, -0.00007490, 0.00005715, -0.00000611, 0.00000167, -0.00002630, 0.00001605, -0.00000339, -0.00001420, 0.00000390, 0.00000601, 0.00002394, -0.00004394, 0.00000423, 0.00000194, -0.00000281, 0.00000440, -0.00000562, -0.00000407, 0.00000593, 0.00000032, 0.00000752, },
		{ -0.00000291, -0.00000325, -0.00000575, -0.00000218, -0.00000598, -0.00000094, 0.00000772, -0.00000318, 0.00001079, -0.00000165, -0.00000767, -0.00000041, 0.00000440, -0.00000591, 0.00002103, -0.00003750, -0.00001034, 0.00003382, -0.00002133, 0.00002357, -0.00005042, 0.00005158, -0.00004685, 0.00003694, -0.00007632, 0.00013644, -0.00016320, 0.00015965, -0.00018968, 0.00013559, -0.00008701, -0.00010117, 0.00004237, -0.00088133, 0.00179214, -0.00020651, -0.00049474, 0.00022273, -0.00033633, 0.00026680, -0.00023887, 0.00012818, -0.00009991, 0.00007529, -0.00001931, 0.00000713, -0.00003088, 0.00002464, -0.00001244, -0.00000885, 0.00000101, 0.00000963, 0.00002017, -0.00004605, 0.00000895, 0.00000023, -0.00000125, 0.00000352, -0.00000602, -0.00000384, 0.00000686, 0.00000029, 0.00000704, 0.00000200, },
		{ -0.00000338, -0.00000531, -0.00000274, -0.00000545, 0.00000025, 0.00000711, -0.00000197, 0.00001005, -0.00000258, -0.00000713, 0.00000049, 0.00000327, -0.00000486, 0.00001961, -0.00004254, -0.00000056, 0.00002669, -0.00001464, 0.00001468, -0.00004174, 0.00004668, -0.00004363, 0.00002922, -0.00006335, 0.00012560, -0.00015621, 0.00016711, -0.00022083, 0.00018646, -0.00016296, -0.00001433, -0.00008300, -0.00075719, 0.00185962, -0.00041423, -0.00035811, 0.00015276, -0.00029120, 0.00025402, -0.00024570, 0.00015029, -0.00012303, 0.00009366, -0.00003383, 0.00001379, -0.00003542, 0.00003285, -0.00002219, -0.00000201, -0.00000316, 0.00001435, 0.00001484, -0.00004669, 0.00001330, -0.00000160, 0.00000034, 0.00000253, -0.00000635, -0.00000358, 0.00000794, -0.00000013, 0.00000676, 0.00000164, -0.00000378, },
		{ -0.00000516, -0.00000318, -0.00000467, 0.00000109, 0.00000678, -0.00000091, 0.00000905, -0.00000319, -0.00000670, 0.00000150, 0.00000188, -0.00000336, 0.00001695, -0.00004558, 0.00000790, 0.00002010, -0.00000847, 0.00000598, -0.00003214, 0.00004032, -0.00003974, 0.00002132, -0.00004884, 0.00011090, -0.00014236, 0.00016374, -0.00023974, 0.00022674, -0.00023233, 0.00007206, -0.00021860, -0.00059990, 0.00188247, -0.00059990, -0.00021860, 0.00007206, -0.00023233, 0.00022674, -0.00023974, 0.00016374, -0.00014236, 0.00011090, -0.00004884, 0.00002132, -0.00003974, 0.00004032, -0.00003214, 0.00000598, -0.00000847, 0.00002010, 0.00000790, -0.00004558, 0.00001695, -0.00000336, 0.00000188, 0.00000150, -0.00000670, -0.00000319, 0.00000905, -0.00000091, 0.00000678, 0.00000109, -0.00000467, -0.00000318, },
		{ -0.00000338, -0.00000378, 0.00000164, 0.00000676, -0.00000013, 0.00000794, -0.00000358, -0.00000635, 0.00000253, 0.00000034, -0.00000160, 0.00001330, -0.00004669, 0.00001484, 0.00001435, -0.00000316, -0.00000201, -0.00002219, 0.00003285, -0.00003542, 0.00001379, -0.00003383, 0.00009366, -0.00012303, 0.00015029, -0.00024570, 0.00025402, -0.00029120, 0.00015276, -0.00035811, -0.00041423, 0.00185962, -0.00075719, -0.00008300, -0.00001433, -0.00016296, 0.00018646, -0.00022083, 0.00016711, -0.00015621, 0.00012560, -0.00006335, 0.00002922, -0.00004363, 0.00004668, -0.00004174, 0.00001468, -0.00001464, 0.00002669, -0.00000056, -0.00004254, 0.00001961, -0.00000486, 0.00000327, 0.00000049, -0.00000713, -0.00000258, 0.00001005, -0.00000197, 0.00000711, 0.00000025, -0.00000545, -0.00000274, -0.00000531, },
		{ -0.00000291, 0.00000200, 0.00000704, 0.00000029, 0.00000686, -0.00000384, -0.00000602, 0.00000352, -0.00000125, 0.00000023, 0.00000895, -0.00004605, 0.00002017, 0.00000963, 0.00000101, -0.00000885, -0.00001244, 0.00002464, -0.00003088, 0.00000713, -0.00001931, 0.00007529, -0.00009991, 0.00012818, -0.00023887, 0.00026680, -0.00033633, 0.00022273, -0.00049474, -0.00020651, 0.00179214, -0.00088133, 0.00004237, -0.00010117, -0.00008701, 0.00013559, -0.00018968, 0.00015965, -0.00016320, 0.00013644, -0.00007632, 0.00003694, -0.00004685, 0.00005158, -0.00005042, 0.00002357, -0.00002133, 0.00003382, -0.00001034, -0.00003750, 0.00002103, -0.00000591, 0.00000440, -0.00000041, -0.00000767, -0.00000165, 0.00001079, -0.00000318, 0.00000772, -0.00000094, -0.00000598, -0.00000218, -0.00000575, -0.00000325, },
		{ 0.00000229, 0.00000752, 0.00000032, 0.00000593, -0.00000407, -0.00000562, 0.00000440, -0.00000281, 0.00000194, 0.00000423, -0.00004394, 0.00002394, 0.00000601, 0.00000390, -0.00001420, -0.00000339, 0.00001605, -0.00002630, 0.00000167, -0.00000611, 0.00005715, -0.00007490, 0.00009938, -0.00022028, 0.00026452, -0.00036536, 0.00027756, -0.00062157, 0.00001570, 0.00168309, -0.00096929, 0.00015211, -0.00018331, -0.00000876, 0.00007726, -0.00014789, 0.00014133, -0.00016240, 0.00014229, -0.00008673, 0.00004384, -0.00004914, 0.00005472, -0.00005765, 0.00003208, -0.00002810, 0.00004109, -0.00002113, -0.00003050, 0.00002102, -0.00000633, 0.00000514, -0.00000116, -0.00000838, -0.00000033, 0.00001113, -0.00000437, 0.00000852, -0.00000247, -0.00000615, -0.00000163, -0.00000642, -0.00000275, -0.00000219, },
		{ 0.00000811, -0.00000004, 0.00000522, -0.00000434, -0.00000512, 0.00000511, -0.00000425, 0.00000339, -0.00000056, -0.00004065, 0.00002629, 0.00000346, 0.00000549, -0.00001781, 0.00000450, 0.00000742, -0.00002176, -0.00000237, 0.00000507, 0.00004051, -0.00004993, 0.00006623, -0.00019174, 0.00024769, -0.00037701, 0.00031371, -0.00073201, 0.00024405, 0.00153740, -0.00101998, 0.00024200, -0.00025609, 0.00006741, 0.00001506, -0.00009778, 0.00011286, -0.00015341, 0.00014231, -0.00009369, 0.00004931, -0.00005022, 0.00005587, -0.00006294, 0.00003966, -0.00003447, 0.00004803, -0.00003246, -0.00002175, 0.00001947, -0.00000597, 0.00000541, -0.00000168, -0.00000927, 0.00000141, 0.00001092, -0.00000536, 0.00000939, -0.00000432, -0.00000589, -0.00000123, -0.00000721, -0.00000186, -0.00000173, 0.00000265, },
		{ -0.00000071, 0.00000477, -0.00000471, -0.00000446, 0.00000560, -0.00000553, 0.00000447, -0.00000513, -0.00003655, 0.00002744, 0.00000185, 0.00000588, -0.00001961, 0.00001089, -0.00000091, -0.00001734, -0.00000497, 0.00001383, 0.00002635, -0.00002682, 0.00003125, -0.00015568, 0.00021775, -0.00037113, 0.00032881, -0.00082022, 0.00046986, 0.00136152, -0.00103415, 0.00030919, -0.00031560, 0.00013736, -0.00004715, -0.00004228, 0.00007565, -0.00013640, 0.00013606, -0.00009647, 0.00005276, -0.00004986, 0.00005484, -0.00006591, 0.00004579, -0.00003995, 0.00005413, -0.00004381, -0.00001154, 0.00001640, -0.00000475, 0.00000513, -0.00000195, -0.00001034, 0.00000358, 0.00001008, -0.00000598, 0.00001017, -0.00000639, -0.00000518, -0.00000109, -0.00000799, -0.00000064, -0.00000159, 0.00000320, 0.00000867, },
		{ 0.00000457, -0.00000522, -0.00000361, 0.00000585, -0.00000663, 0.00000515, -0.00000927, -0.00003196, 0.00002764, 0.00000097, 0.00000527, -0.00001963, 0.00001551, -0.00000868, -0.00001303, -0.00000620, 0.00001996, 0.00001540, -0.00000713, -0.00000300, -0.00011494, 0.00017697, -0.00034874, 0.00032182, -0.00088149, 0.00068457, 0.00116308, -0.00101432, 0.00035237, -0.00035892, 0.00019750, -0.00010559, 0.00001532, 0.00003173, -0.00011212, 0.00012350, -0.00009459, 0.00005372, -0.00004788, 0.00005159, -0.00006628, 0.00005002, -0.00004411, 0.00005889, -0.00005456, -0.00000032, 0.00001192, -0.00000262, 0.00000426, -0.00000194, -0.00001155, 0.00000614, 0.00000853, -0.00000607, 0.00001067, -0.00000854, -0.00000403, -0.00000132, -0.00000862, 0.00000084, -0.00000180, 0.00000404, 0.00000904, -0.00000159, },
		{ -0.00000585, -0.00000257, 0.00000584, -0.00000756, 0.00000540, -0.00001281, -0.00002719, 0.00002716, 0.00000060, 0.00000393, -0.00001803, 0.00001824, -0.00001564, -0.00000881, -0.00000631, 0.00002352, 0.00000801, 0.00000796, -0.00003418, -0.00007257, 0.00012831, -0.00031192, 0.00029307, -0.00091248, 0.00088028, 0.00095039, -0.00096452, 0.00037171, -0.00038430, 0.00024496, -0.00015679, 0.00007156, -0.00001644, -0.00008183, 0.00010502, -0.00008790, 0.00005186, -0.00004418, 0.00004612, -0.00006391, 0.00005202, -0.00004656, 0.00006182, -0.00006409, 0.00001138, 0.00000626, 0.00000038, 0.00000278, -0.00000166, -0.00001285, 0.00000901, 0.00000629, -0.00000551, 0.00001073, -0.00001060, -0.00000252, -0.00000197, -0.00000897, 0.00000244, -0.00000234, 0.00000525, 0.00000908, -0.00000253, 0.00000455, },
		{ -0.00000134, 0.00000556, -0.00000834, 0.00000527, -0.00001566, -0.00002248, 0.00002624, 0.00000049, 0.00000219, -0.00001510, 0.00001904, -0.00002161, -0.00000463, -0.00000560, 0.00002480, 0.00000419, 0.00001768, -0.00006031, -0.00003160, 0.00007511, -0.00026365, 0.00024430, -0.00091148, 0.00105010, 0.00073198, -0.00088999, 0.00036880, -0.00039122, 0.00027783, -0.00019785, 0.00012311, -0.00006608, -0.00004724, 0.00008147, -0.00007653, 0.00004702, -0.00003873, 0.00003860, -0.00005879, 0.00005158, -0.00004703, 0.00006253, -0.00007179, 0.00002297, -0.00000028, 0.00000412, 0.00000074, -0.00000112, -0.00001416, 0.00001207, 0.00000341, -0.00000423, 0.00001020, -0.00001239, -0.00000077, -0.00000305, -0.00000891, 0.00000403, -0.00000314, 0.00000683, 0.00000865, -0.00000339, 0.00000464, -0.00000659, },
		{ 0.00000503, -0.00000899, 0.00000485, -0.00001781, -0.00001801, 0.00002510, 0.00000041, 0.00000040, -0.00001118, 0.00001800, -0.00002645, -0.00000044, -0.00000445, 0.00002425, 0.00000362, 0.00002180, -0.00007989, 0.00000521, 0.00002092, -0.00020757, 0.00017848, -0.00087851, 0.00118860, 0.00051612, -0.00079674, 0.00034642, -0.00038032, 0.00029523, -0.00022662, 0.00016696, -0.00011436, -0.00001035, 0.00005404, -0.00006096, 0.00003926, -0.00003163, 0.00002927, -0.00005109, 0.00004865, -0.00004537, 0.00006076, -0.00007711, 0.00003382, -0.00000729, 0.00000843, -0.00000179, -0.00000038, -0.00001540, 0.00001516, 0.00000002, -0.00000223, 0.00000895, -0.00001372, 0.00000106, -0.00000452, -0.00000835, 0.00000545, -0.00000408, 0.00000877, 0.00000769, -0.00000401, 0.00000473, -0.00000736, 0.00000004, },
		{ -0.00000955, 0.00000423, -0.00001930, -0.00001388, 0.00002387, 0.00000018, -0.00000112, -0.00000667, 0.00001528, -0.00003007, 0.00000380, -0.00000324, 0.00002245, 0.00000569, 0.00002054, -0.00009203, 0.00003556, -0.00003087, -0.00014772, 0.00009962, -0.00081523, 0.00129196, 0.00031039, -0.00069122, 0.00030830, -0.00035336, 0.00029736, -0.00024190, 0.00020066, -0.00015855, 0.00002666, 0.00002422, -0.00004196, 0.00002883, -0.00002308, 0.00001849, -0.00004107, 0.00004330, -0.00004160, 0.00005638, -0.00007961, 0.00004332, -0.00001432, 0.00001307, -0.00000469, 0.00000051, -0.00001648, 0.00001812, -0.00000373, 0.00000043, 0.00000692, -0.00001442, 0.00000279, -0.00000631, -0.00000723, 0.00000654, -0.00000501, 0.00001096, 0.00000616, -0.00000427, 0.00000470, -0.00000808, 0.00000152, 0.00000427, },
		{ 0.00000352, -0.00002022, -0.00001012, 0.00002263, -0.00000031, -0.00000210, -0.00000200, 0.00001114, -0.00003241, 0.00000809, -0.00000232, 0.00002002, 0.00000959, 0.00001458, -0.00009648, 0.00005774, -0.00007716, -0.00008823, 0.00001249, -0.00072489, 0.00135822, 0.00012129, -0.00057984, 0.00025877, -0.00031294, 0.00028540, -0.00024340, 0.00022247, -0.00019625, 0.00006160, -0.00000631, -0.00002054, 0.00001617, -0.00001339, 0.00000668, -0.00002915, 0.00003578, -0.00003585, 0.00004942, -0.00007899, 0.00005092, -0.00002091, 0.00001775, -0.00000781, 0.00000145, -0.00001728, 0.00002075, -0.00000760, 0.00000363, 0.00000411, -0.00001436, 0.00000424, -0.00000827, -0.00000555, 0.00000718, -0.00000577, 0.00001328, 0.00000408, -0.00000407, 0.00000443, -0.00000866, 0.00000303, 0.00000333, -0.00001007, },
		{ -0.00002071, -0.00000671, 0.00002140, -0.00000108, -0.00000234, 0.00000243, 0.00000585, -0.00003346, 0.00001239, -0.00000199, 0.00001760, 0.00001437, 0.00000497, -0.00009359, 0.00007077, -0.00011547, -0.00003302, -0.00007771, -0.00061205, 0.00138724, -0.00004605, -0.00046859, 0.00020247, -0.00026234, 0.00026138, -0.00023184, 0.00023146, -0.00022554, 0.00009246, -0.00003581, 0.00000207, 0.00000191, -0.00000294, -0.00000565, -0.00001584, 0.00002643, -0.00002842, 0.00004008, -0.00007512, 0.00005615, -0.00002660, 0.00002218, -0.00001096, 0.00000237, -0.00001773, 0.00002289, -0.00001137, 0.00000718, 0.00000060, -0.00001345, 0.00000522, -0.00001024, -0.00000336, 0.00000726, -0.00000620, 0.00001555, 0.00000153, -0.00000338, 0.00000383, -0.00000902, 0.00000449, 0.00000222, -0.00001056, 0.00000285, },
		{ -0.00000357, 0.00002013, -0.00000208, -0.00000173, 0.00000627, -0.00000024, -0.00003322, 0.00001661, -0.00000242, 0.00001571, 0.00001907, -0.00000698, -0.00008430, 0.00007441, -0.00014401, 0.00001451, -0.00016572, -0.00048229, 0.00138065, -0.00018804, -0.00036274, 0.00014396, -0.00020518, 0.00022800, -0.00020879, 0.00022760, -0.00024509, 0.00011752, -0.00006261, 0.00002460, -0.00001320, 0.00000779, -0.00001796, -0.00000172, 0.00001572, -0.00001971, 0.00002874, -0.00006805, 0.00005870, -0.00003097, 0.00002603, -0.00001396, 0.00000317, -0.00001775, 0.00002435, -0.00001478, 0.00001085, -0.00000348, -0.00001168, 0.00000558, -0.00001203, -0.00000074, 0.00000672, -0.00000613, 0.00001757, -0.00000134, -0.00000220, 0.00000285, -0.00000905, 0.00000584, 0.00000098, -0.00001104, 0.00000230, -0.00002091, },
		{ 0.00001871, -0.00000320, -0.00000026, 0.00000922, -0.00000682, -0.00003172, 0.00002062, -0.00000369, 0.00001476, 0.00002280, -0.00001984, -0.00007000, 0.00006915, -0.00016185, 0.00005166, -0.00024657, -0.00034184, 0.00134161, -0.00030272, -0.00026655, 0.00008744, -0.00014513, 0.00018836, -0.00017654, 0.00021168, -0.00025429, 0.00013549, -0.00008520, 0.00004573, -0.00002834, 0.00001829, -0.00002969, 0.00001260, 0.00000416, -0.00001023, 0.00001591, -0.00005802, 0.00005837, -0.00003368, 0.00002901, -0.00001662, 0.00000376, -0.00001729, 0.00002502, -0.00001758, 0.00001436, -0.00000793, -0.00000910, 0.00000522, -0.00001347, 0.00000216, 0.00000558, -0.00000544, 0.00001912, -0.00000435, -0.00000059, 0.00000145, -0.00000871, 0.00000699, -0.00000034, -0.00001150, 0.00000193, -0.00002094, -0.00000059, },
		{ -0.00000423, 0.00000196, 0.00001108, -0.00001355, -0.00002902, 0.00002425, -0.00000574, 0.00001500, 0.00002482, -0.00003216, -0.00005239, 0.00005613, -0.00016891, 0.00007665, -0.00031591, -0.00019717, 0.00127458, -0.00038972, -0.00018314, 0.00003650, -0.00008569, 0.00014573, -0.00013793, 0.00018525, -0.00025321, 0.00014559, -0.00010238, 0.00006425, -0.00004264, 0.00002804, -0.00004031, 0.00002650, -0.00000768, -0.00000052, 0.00000222, -0.00004546, 0.00005516, -0.00003449, 0.00003087, -0.00001874, 0.00000406, -0.00001633, 0.00002482, -0.00001956, 0.00001744, -0.00001250, -0.00000583, 0.00000406, -0.00001436, 0.00000516, 0.00000387, -0.00000406, 0.00002002, -0.00000727, 0.00000131, -0.00000034, -0.00000796, 0.00000789, -0.00000173, -0.00001191, 0.00000177, -0.00002093, 0.00000231, 0.00001702, },
		{ 0.00000476, 0.00001176, -0.00002015, -0.00002519, 0.00002729, -0.00000840, 0.00001647, 0.00002463, -0.00004264, -0.00003334, 0.00003701, -0.00016590, 0.00008868, -0.00037027, -0.00005459, 0.00118493, -0.00045009, -0.00011438, -0.00000610, -0.00002985, 0.00010323, -0.00009607, 0.00015050, -0.00024263, 0.00014761, -0.00011333, 0.00007910, -0.00005531, 0.00003652, -0.00004932, 0.00003938, -0.00001927, 0.00000883, -0.00001163, -0.00003095, 0.00004925, -0.00003329, 0.00003143, -0.00002017, 0.00000400, -0.00001488, 0.00002370, -0.00002056, 0.00001984, -0.00001690, -0.00000205, 0.00000213, -0.00001457, 0.00000807, 0.00000171, -0.00000196, 0.00002009, -0.00000987, 0.00000335, -0.00000243, -0.00000680, 0.00000850, -0.00000316, -0.00001222, 0.00000181, -0.00002094, 0.00000523, 0.00001495, -0.00000496, },
		{ 0.00001123, -0.00002637, -0.00002033, 0.00002952, -0.00001140, 0.00001902, 0.00002201, -0.00005022, -0.00001469, 0.00001379, -0.00015428, 0.00008797, -0.00040726, 0.00008012, 0.00107856, -0.00048613, -0.00006095, -0.00003848, 0.00001999, 0.00006363, -0.00005411, 0.00011002, -0.00022391, 0.00014190, -0.00011765, 0.00008950, -0.00006562, 0.00004330, -0.00005631, 0.00005072, -0.00003009, 0.00001728, -0.00002491, -0.00001518, 0.00004095, -0.00003008, 0.00003057, -0.00002080, 0.00000354, -0.00001297, 0.00002168, -0.00002047, 0.00002132, -0.00002086, 0.00000200, -0.00000050, -0.00001399, 0.00001068, -0.00000075, 0.00000081, 0.00001921, -0.00001193, 0.00000531, -0.00000472, -0.00000527, 0.00000879, -0.00000461, -0.00001238, 0.00000197, -0.00002098, 0.00000820, 0.00001239, -0.00000515, 0.00000786, },
		{ -0.00003198, -0.00001454, 0.00003073, -0.00001439, 0.00002232, 0.00001704, -0.00005419, 0.00000188, -0.00001136, -0.00013603, 0.00007569, -0.00042570, 0.00020201, 0.00096152, -0.00050106, -0.00002245, -0.00005970, 0.00006222, 0.00002913, -0.00001498, 0.00006665, -0.00019890, 0.00012934, -0.00011541, 0.00009495, -0.00007301, 0.00004800, -0.00006094, 0.00006009, -0.00003971, 0.00002436, -0.00003689, 0.00000106, 0.00003075, -0.00002500, 0.00002826, -0.00002054, 0.00000265, -0.00001067, 0.00001886, -0.00001925, 0.00002169, -0.00002408, 0.00000605, -0.00000368, -0.00001257, 0.00001282, -0.00000331, 0.00000416, 0.00001733, -0.00001324, 0.00000700, -0.00000704, -0.00000345, 0.00000878, -0.00000609, -0.00001231, 0.00000217, -0.00002102, 0.00001122, 0.00000931, -0.00000461, 0.00001097, 0.00000960, },
		{ -0.00000798, 0.00003076, -0.00001699, 0.00002592, 0.00001007, -0.00005421, 0.00001500, -0.00003632, -0.00011348, 0.00005380, -0.00042559, 0.00030720, 0.00083964, -0.00049870, 0.00000244, -0.00006971, 0.00009603, 0.00000124, 0.00001882, 0.00002322, -0.00016973, 0.00011123, -0.00010712, 0.00009530, -0.00007710, 0.00005039, -0.00006302, 0.00006717, -0.00004779, 0.00002969, -0.00004695, 0.00001696, 0.00001921, -0.00001832, 0.00002457, -0.00001938, 0.00000133, -0.00000808, 0.00001537, -0.00001693, 0.00002086, -0.00002632, 0.00000981, -0.00000723, -0.00001032, 0.00001430, -0.00000577, 0.00000791, 0.00001446, -0.00001365, 0.00000820, -0.00000921, -0.00000145, 0.00000849, -0.00000759, -0.00001194, 0.00000227, -0.00002096, 0.00001420, 0.00000573, -0.00000317, 0.00001377, 0.00000700, -0.00003683, },
		{ 0.00002946, -0.00001884, 0.00002927, 0.00000169, -0.00005037, 0.00002374, -0.00005918, -0.00008907, 0.00002487, -0.00040813, 0.00039308, 0.00071819, -0.00048319, 0.00001572, -0.00006919, 0.00012136, -0.00001935, 0.00004537, -0.00001768, -0.00013868, 0.00008921, -0.00009368, 0.00009073, -0.00007773, 0.00005035, -0.00006245, 0.00007174, -0.00005410, 0.00003303, -0.00005454, 0.00003175, 0.00000702, -0.00001040, 0.00001965, -0.00001736, -0.00000038, -0.00000532, 0.00001138, -0.00001365, 0.00001880, -0.00002739, 0.00001300, -0.00001089, -0.00000732, 0.00001502, -0.00000791, 0.00001186, 0.00001070, -0.00001307, 0.00000871, -0.00001104, 0.00000059, 0.00000797, -0.00000910, -0.00001120, 0.00000214, -0.00002068, 0.00001702, 0.00000175, -0.00000075, 0.00001595, 0.00000365, -0.00004081, -0.00000079, },
		{ -0.00001961, 0.00003181, -0.00000729, -0.00004315, 0.00002762, -0.00007841, -0.00006518, -0.00000816, -0.00037546, 0.00045828, 0.00060164, -0.00045861, 0.00001979, -0.00005940, 0.00013882, -0.00003270, 0.00006351, -0.00005384, -0.00010798, 0.00006513, -0.00007633, 0.00008176, -0.00007495, 0.00004791, -0.00005929, 0.00007373, -0.00005855, 0.00003428, -0.00005929, 0.00004470, -0.00000515, -0.00000170, 0.00001374, -0.00001455, -0.00000244, -0.00000251, 0.00000711, -0.00000958, 0.00001555, -0.00002718, 0.00001536, -0.00001441, -0.00000371, 0.00001491, -0.00000953, 0.00001576, 0.00000619, -0.00001146, 0.00000841, -0.00001236, 0.00000252, 0.00000729, -0.00001063, -0.00001005, 0.00000164, -0.00002002, 0.00001949, -0.00000246, 0.00000264, 0.00001728, -0.00000020, -0.00004383, 0.00000683, 0.00002681, },
		{ 0.00003300, -0.00001599, -0.00003336, 0.00002669, -0.00009298, -0.00004383, -0.00004225, -0.00033053, 0.00050272, 0.00049348, -0.00042876, 0.00001723, -0.00004195, 0.00014954, -0.00003959, 0.00007284, -0.00008361, -0.00007964, 0.00004083, -0.00005650, 0.00006917, -0.00006905, 0.00004325, -0.00005371, 0.00007319, -0.00006116, 0.00003348, -0.00006099, 0.00005521, -0.00001663, 0.00000730, 0.00000711, -0.00001110, -0.00000477, 0.00000023, 0.00000280, -0.00000496, 0.00001127, -0.00002565, 0.00001670, -0.00001751, 0.00000032, 0.00001395, -0.00001045, 0.00001934, 0.00000118, -0.00000890, 0.00000720, -0.00001299, 0.00000419, 0.00000654, -0.00001219, -0.00000845, 0.00000064, -0.00001879, 0.00002138, -0.00000663, 0.00000687, 0.00001757, -0.00000431, -0.00004585, 0.00001464, 0.00002282, -0.00001908, },
		{ -0.00002350, -0.00002204, 0.00002147, -0.00010240, -0.00002663, -0.00007451, -0.00027675, 0.00052739, 0.00039614, -0.00039690, 0.00001056, -0.00001863, 0.00015495, -0.00004139, 0.00007372, -0.00010598, -0.00005526, 0.00001807, -0.00003576, 0.00005398, -0.00006051, 0.00003666, -0.00004598, 0.00007027, -0.00006204, 0.00003083, -0.00005963, 0.00006285, -0.00002679, 0.00001609, 0.00000013, -0.00000719, -0.00000728, 0.00000276, -0.00000132, -0.00000007, 0.00000617, -0.00002284, 0.00001686, -0.00001993, 0.00000455, 0.00001220, -0.00001055, 0.00002234, -0.00000407, -0.00000551, 0.00000506, -0.00001283, 0.00000544, 0.00000581, -0.00001376, -0.00000641, -0.00000094, -0.00001685, 0.00002245, -0.00001044, 0.00001171, 0.00001674, -0.00000847, -0.00004684, 0.00002240, 0.00001762, -0.00001713, 0.00003241, },
		{ -0.00001040, 0.00001286, -0.00010677, -0.00001454, -0.00010244, -0.00021780, 0.00053424, 0.00031101, -0.00036568, 0.00000202, 0.00000889, 0.00015659, -0.00003985, 0.00006721, -0.00012062, -0.00003601, -0.00000164, -0.00001563, 0.00003733, -0.00004995, 0.00002857, -0.00003649, 0.00006523, -0.00006141, 0.00002665, -0.00005538, 0.00006734, -0.00003512, 0.00002417, -0.00000687, -0.00000302, -0.00000989, 0.00000498, -0.00000503, 0.00000477, 0.00000053, -0.00001892, 0.00001580, -0.00002145, 0.00000872, 0.00000979, -0.00000977, 0.00002452, -0.00000925, -0.00000151, 0.00000205, -0.00001181, 0.00000618, 0.00000517, -0.00001532, -0.00000396, -0.00000315, -0.00001407, 0.00002248, -0.00001356, 0.00001686, 0.00001483, -0.00001248, -0.00004678, 0.00002983, 0.00001146, -0.00001378, 0.00002980, -0.00002901, },
		{ 0.00000205, -0.00010667, -0.00000791, -0.00012410, -0.00015726, 0.00052587, 0.00023852, -0.00033702, -0.00000652, 0.00003911, 0.00015588, -0.00003691, 0.00005488, -0.00012786, -0.00002247, -0.00001716, 0.00000247, 0.00002042, -0.00003814, 0.00001949, -0.00002570, 0.00005838, -0.00005955, 0.00002135, -0.00004857, 0.00006863, -0.00004123, 0.00003111, -0.00001351, 0.00000119, -0.00001247, 0.00000681, -0.00000813, 0.00000928, -0.00000534, -0.00001409, 0.00001354, -0.00002190, 0.00001256, 0.00000687, -0.00000811, 0.00002569, -0.00001403, 0.00000284, -0.00000169, -0.00000993, 0.00000630, 0.00000469, -0.00001686, -0.00000118, -0.00000597, -0.00001038, 0.00002127, -0.00001561, 0.00002192, 0.00001198, -0.00001621, -0.00004567, 0.00003664, 0.00000462, -0.00000922, 0.00002507, -0.00003187, 0.00000036, },
		{ -0.00010314, -0.00000645, -0.00013829, -0.00009837, 0.00050526, 0.00017827, -0.00031213, -0.00001368, 0.00007092, 0.00015394, -0.00003447, 0.00003870, -0.00012862, -0.00001468, -0.00002777, 0.00001738, 0.00000440, -0.00002585, 0.00000997, -0.00001412, 0.00005010, -0.00005677, 0.00001541, -0.00003971, 0.00006683, -0.00004489, 0.00003655, -0.00001948, 0.00000520, -0.00001494, 0.00000819, -0.00001046, 0.00001317, -0.00001108, -0.00000864, 0.00001019, -0.00002117, 0.00001585, 0.00000366, -0.00000562, 0.00002571, -0.00001813, 0.00000724, -0.00000597, -0.00000726, 0.00000577, 0.00000445, -0.00001833, 0.00000182, -0.00000935, -0.00000577, 0.00001868, -0.00001628, 0.00002644, 0.00000840, -0.00001959, -0.00004350, 0.00004252, -0.00000250, -0.00000375, 0.00001838, -0.00003168, 0.00000915, -0.00000960, },
		{ -0.00000926, -0.00014453, -0.00004384, 0.00047554, 0.00012928, -0.00029161, -0.00001851, 0.00010362, 0.00015147, -0.00003421, 0.00002077, -0.00012425, -0.00001216, -0.00003322, 0.00002826, -0.00000969, -0.00001390, 0.00000059, -0.00000226, 0.00004079, -0.00005338, 0.00000934, -0.00002939, 0.00006224, -0.00004600, 0.00004022, -0.00002449, 0.00000881, -0.00001719, 0.00000910, -0.00001194, 0.00001620, -0.00001637, -0.00000290, 0.00000596, -0.00001925, 0.00001836, 0.00000039, -0.00000245, 0.00002452, -0.00002127, 0.00001136, -0.00001054, -0.00000392, 0.00000460, 0.00000446, -0.00001970, 0.00000491, -0.00001314, -0.00000032, 0.00001467, -0.00001528, 0.00002997, 0.00000443, -0.00002264, -0.00004027, 0.00004719, -0.00000948, 0.00000220, 0.00001009, -0.00002831, 0.00001509, -0.00002073, -0.00009750, },
		{ -0.00014314, 0.00000432, 0.00043969, 0.00009016, -0.00027556, -0.00002051, 0.00013691, 0.00014864, -0.00003734, 0.00000316, -0.00011642, -0.00001396, -0.00003375, 0.00003463, -0.00002102, -0.00000301, -0.00000810, 0.00000935, 0.00003084, -0.00004968, 0.00000363, -0.00001828, 0.00005532, -0.00004467, 0.00004200, -0.00002833, 0.00001183, -0.00001915, 0.00000953, -0.00001251, 0.00001820, -0.00002089, 0.00000278, 0.00000109, -0.00001620, 0.00001994, -0.00000271, 0.00000122, 0.00002214, -0.00002324, 0.00001490, -0.00001510, -0.00000009, 0.00000285, 0.00000475, -0.00002092, 0.00000793, -0.00001715, 0.00000580, 0.00000929, -0.00001245, 0.00003210, 0.00000043, -0.00002540, -0.00003596, 0.00005038, -0.00001586, 0.00000809, 0.00000076, -0.00002199, 0.00001762, -0.00003013, -0.00009118, -0.00001503, },
		{ 0.00004488, 0.00040034, 0.00005934, -0.00026375, -0.00001946, 0.00017085, 0.00014509, -0.00004453, -0.00001233, -0.00010688, -0.00001884, -0.00003001, 0.00003640, -0.00002901, 0.00000621, -0.00001559, 0.00002022, 0.00002063, -0.00004593, -0.00000130, -0.00000707, 0.00004661, -0.00004111, 0.00004186, -0.00003086, 0.00001413, -0.00002076, 0.00000954, -0.00001221, 0.00001907, -0.00002440, 0.00000804, -0.00000408, -0.00001215, 0.00002048, -0.00000539, 0.00000516, 0.00001866, -0.00002392, 0.00001756, -0.00001934, 0.00000397, 0.00000064, 0.00000527, -0.00002192, 0.00001070, -0.00002116, 0.00001237, 0.00000271, -0.00000773, 0.00003245, -0.00000321, -0.00002797, -0.00003059, 0.00005184, -0.00002120, 0.00001337, -0.00000887, -0.00001322, 0.00001656, -0.00003684, -0.00008562, -0.00002212, -0.00013508, },
		{ 0.00035962, 0.00003531, -0.00025577, -0.00001531, 0.00020574, 0.00013997, -0.00005582, -0.00002429, -0.00009732, -0.00002532, -0.00002299, 0.00003389, -0.00003335, 0.00001328, -0.00002148, 0.00002993, 0.00001050, -0.00004231, -0.00000510, 0.00000358, 0.00003678, -0.00003568, 0.00003991, -0.00003202, 0.00001560, -0.00002196, 0.00000918, -0.00001111, 0.00001877, -0.00002669, 0.00001256, -0.00000924, -0.00000730, 0.00001996, -0.00000747, 0.00000911, 0.00001426, -0.00002325, 0.00001913, -0.00002296, 0.00000802, -0.00000188, 0.00000600, -0.00002266, 0.00001307, -0.00002490, 0.00001907, -0.00000480, -0.00000119, 0.00003078, -0.00000612, -0.00003049, -0.00002416, 0.00005141, -0.00002511, 0.00001745, -0.00001800, -0.00000281, 0.00001211, -0.00004030, -0.00008200, -0.00002882, -0.00012189, 0.00007739, },
		{ 0.00001671, -0.00025118, -0.00000805, 0.00024194, 0.00013207, -0.00007060, -0.00003187, -0.00008922, -0.00003193, -0.00001394, 0.00002772, -0.00003404, 0.00001794, -0.00002548, 0.00003812, 0.00000075, -0.00003894, -0.00000752, 0.00001310, 0.00002647, -0.00002883, 0.00003634, -0.00003187, 0.00001620, -0.00002275, 0.00000854, -0.00000934, 0.00001736, -0.00002767, 0.00001609, -0.00001403, -0.00000193, 0.00001842, -0.00000878, 0.00001280, 0.00000916, -0.00002128, 0.00001944, -0.00002569, 0.00001177, -0.00000452, 0.00000685, -0.00002307, 0.00001486, -0.00002810, 0.00002556, -0.00001287, 0.00000696, 0.00002695, -0.00000795, -0.00003308, -0.00001674, 0.00004898, -0.00002725, 0.00001983, -0.00002576, 0.00000825, 0.00000487, -0.00004035, -0.00008120, -0.00003351, -0.00010542, 0.00010213, 0.00031910, },
		},
		{ { 0.00068669, 0.00010928, -0.00025939, -0.00009803, -0.00001568, -0.00007142, 0.00009528, -0.00004725, -0.00000303, -0.00000431, -0.00000058, -0.00002840, 0.00001919, -0.00003255, 0.00004341, -0.00001401, -0.00000389, -0.00001478, 0.00000657, -0.00001114, 0.00001156, -0.00001099, 0.00001706, 0.00000074, -0.00000260, -0.00001068, -0.00000316, -0.00000495, -0.00000096, -0.00000176, 0.00000289, 0.00000373, 0.00000349, 0.00000373, 0.00000289, -0.00000176, -0.00000096, -0.00000495, -0.00000316, -0.00001068, -0.00000260, 0.00000074, 0.00001706, -0.00001099, 0.00001156, -0.00001114, 0.00000657, -0.00001478, -0.00000389, -0.00001401, 0.00004341, -0.00003255, 0.00001919, -0.00002840, -0.00000058, -0.00000431, -0.00000303, -0.00004725, 0.00009528, -0.00007142, -0.00001568, -0.00009803, -0.00025939, 0.00010928, },
		{ 0.00010928, -0.00025991, -0.00009871, -0.00001525, -0.00007186, 0.00009748, -0.00004973, -0.00000303, -0.00000311, 0.00000005, -0.00003033, 0.00002047, -0.00003410, 0.00004632, -0.00001625, -0.00000389, -0.00001459, 0.00000810, -0.00001316, 0.00001290, -0.00001261, 0.00001920, -0.00000044, -0.00000260, -0.00001099, -0.00000205, -0.00000592, -0.00000039, -0.00000234, 0.00000334, 0.00000363, 0.00000349, 0.00000383, 0.00000244, -0.00000118, -0.00000154, -0.00000396, -0.00000429, -0.00001035, -0.00000260, 0.00000199, 0.00001481, -0.00000927, 0.00001013, -0.00000897, 0.00000491, -0.00001497, -0.00000389, -0.00001152, 0.00004013, -0.00003079, 0.00001771, -0.00002612, -0.00000135, -0.00000579, -0.00000303, -0.00004397, 0.00009224, -0.00007077, -0.00001639, -0.00009667, -0.00025782, 0.00010438, 0.00069021, },
		{ -0.00025939, -0.00009871, -0.00001510, -0.00007208, 0.00009880, -0.00005140, -0.00000303, -0.00000220, 0.00000056, -0.00003190, 0.00002154, -0.00003541, 0.00004883, -0.00001820, -0.00000389, -0.00001443, 0.00000949, -0.00001500, 0.00001413, -0.00001411, 0.00002120, -0.00000156, -0.00000260, -0.00001130, -0.00000099, -0.00000685, 0.00000016, -0.00000290, 0.00000378, 0.00000353, 0.00000349, 0.00000393, 0.00000198, -0.00000059, -0.00000212, -0.00000295, -0.00000545, -0.00001001, -0.00000260, 0.00000327, 0.00001246, -0.00000747, 0.00000863, -0.00000667, 0.00000313, -0.00001519, -0.00000389, -0.00000879, 0.00003651, -0.00002882, 0.00001604, -0.00002351, -0.00000224, -0.00000753, -0.00000303, -0.00003994, 0.00008839, -0.00006992, -0.00001738, -0.00009466, -0.00025523, 0.00009461, 0.00070076, 0.00010438, },
		{ -0.00009803, -0.00001525, -0.00007208, 0.00009924, -0.00005223, -0.00000303, -0.00000159, 0.00000092, -0.00003309, 0.00002238, -0.00003648, 0.00005091, -0.00001986, -0.00000389, -0.00001429, 0.00001073, -0.00001666, 0.00001525, -0.00001549, 0.00002305, -0.00000260, -0.00000260, -0.00001158, 0.00000001, -0.00000774, 0.00000068, -0.00000344, 0.00000421, 0.00000343, 0.00000349, 0.00000403, 0.00000153, -0.00000001, -0.00000271, -0.00000194, -0.00000662, -0.00000967, -0.00000260, 0.00000459, 0.00001004, -0.00000560, 0.00000706, -0.00000426, 0.00000126, -0.00001541, -0.00000389, -0.00000586, 0.00003259, -0.00002667, 0.00001419, -0.00002061, -0.00000324, -0.00000952, -0.00000303, -0.00003519, 0.00008375, -0.00006887, -0.00001864, -0.00009201, -0.00025164, 0.00008009, 0.00071823, 0.00009461, -0.00025782, },
		{ -0.00001568, -0.00007186, 0.00009880, -0.00005223, -0.00000303, -0.00000128, 0.00000114, -0.00003389, 0.00002298, -0.00003728, 0.00005256, -0.00002121, -0.00000389, -0.00001417, 0.00001179, -0.00001811, 0.00001624, -0.00001673, 0.00002473, -0.00000355, -0.00000260, -0.00001185, 0.00000095, -0.00000859, 0.00000118, -0.00000395, 0.00000462, 0.00000334, 0.00000349, 0.00000412, 0.00000109, 0.00000056, -0.00000328, -0.00000093, -0.00000779, -0.00000932, -0.00000260, 0.00000593, 0.00000757, -0.00000369, 0.00000545, -0.00000177, -0.00000069, -0.00001565, -0.00000389, -0.00000276, 0.00002840, -0.00002436, 0.00001219, -0.00001744, -0.00000434, -0.00001174, -0.00000303, -0.00002977, 0.00007837, -0.00006763, -0.00002015, -0.00008874, -0.00024708, 0.00006094, 0.00074244, 0.00008009, -0.00025523, -0.00009667, },
		{ -0.00007142, 0.00009748, -0.00005140, -0.00000303, -0.00000128, 0.00000121, -0.00003430, 0.00002335, -0.00003782, 0.00005375, -0.00002224, -0.00000389, -0.00001407, 0.00001268, -0.00001934, 0.00001709, -0.00001781, 0.00002622, -0.00000440, -0.00000260, -0.00001209, 0.00000182, -0.00000937, 0.00000165, -0.00000444, 0.00000501, 0.00000325, 0.00000349, 0.00000422, 0.00000066, 0.00000112, -0.00000385, 0.00000006, -0.00000895, -0.00000898, -0.00000260, 0.00000727, 0.00000508, -0.00000175, 0.00000380, 0.00000079, -0.00000270, -0.00001589, -0.00000389, 0.00000050, 0.00002399, -0.00002191, 0.00001006, -0.00001402, -0.00000554, -0.00001416, -0.00000303, -0.00002373, 0.00007231, -0.00006621, -0.00002190, -0.00008489, -0.00024160, 0.00003735, 0.00077317, 0.00006094, -0.00025164, -0.00009466, -0.00001639, },
		{ 0.00009528, -0.00004973, -0.00000303, -0.00000159, 0.00000114, -0.00003430, 0.00002347, -0.00003810, 0.00005446, -0.00002292, -0.00000389, -0.00001400, 0.00001338, -0.00002033, 0.00001780, -0.00001873, 0.00002750, -0.00000515, -0.00000260, -0.00001231, 0.00000261, -0.00001008, 0.00000209, -0.00000489, 0.00000538, 0.00000316, 0.00000349, 0.00000431, 0.00000025, 0.00000166, -0.00000440, 0.00000103, -0.00001009, -0.00000864, -0.00000260, 0.00000861, 0.00000258, 0.00000020, 0.00000214, 0.00000339, -0.00000474, -0.00001614, -0.00000389, 0.00000387, 0.00001939, -0.00001934, 0.00000780, -0.00001040, -0.00000682, -0.00001677, -0.00000303, -0.00001712, 0.00006563, -0.00006463, -0.00002387, -0.00008049, -0.00023525, 0.00000955, 0.00081012, 0.00003735, -0.00024708, -0.00009201, -0.00001738, -0.00007077, },
		{ -0.00004725, -0.00000303, -0.00000220, 0.00000092, -0.00003389, 0.00002335, -0.00003810, 0.00005470, -0.00002327, -0.00000389, -0.00001395, 0.00001388, -0.00002109, 0.00001836, -0.00001948, 0.00002856, -0.00000578, -0.00000260, -0.00001250, 0.00000331, -0.00001072, 0.00000248, -0.00000531, 0.00000571, 0.00000308, 0.00000349, 0.00000439, -0.00000014, 0.00000218, -0.00000493, 0.00000196, -0.00001119, -0.00000831, -0.00000260, 0.00000993, 0.00000012, 0.00000214, 0.00000048, 0.00000600, -0.00000681, -0.00001640, -0.00000389, 0.00000733, 0.00001465, -0.00001668, 0.00000546, -0.00000661, -0.00000817, -0.00001954, -0.00000303, -0.00001001, 0.00005838, -0.00006290, -0.00002604, -0.00007558, -0.00022809, -0.00002219, 0.00085294, 0.00000955, -0.00024160, -0.00008874, -0.00001864, -0.00006992, 0.00009224, },
		{ -0.00000303, -0.00000311, 0.00000056, -0.00003309, 0.00002298, -0.00003782, 0.00005446, -0.00002327, -0.00000389, -0.00001392, 0.00001418, -0.00002160, 0.00001877, -0.00002004, 0.00002941, -0.00000630, -0.00000260, -0.00001266, 0.00000391, -0.00001129, 0.00000283, -0.00000568, 0.00000602, 0.00000301, 0.00000349, 0.00000447, -0.00000050, 0.00000267, -0.00000543, 0.00000286, -0.00001225, -0.00000800, -0.00000260, 0.00001122, -0.00000230, 0.00000405, -0.00000117, 0.00000859, -0.00000888, -0.00001665, -0.00000389, 0.00001084, 0.00000982, -0.00001395, 0.00000304, -0.00000268, -0.00000957, -0.00002243, -0.00000303, -0.00000247, 0.00005064, -0.00006104, -0.00002840, -0.00007022, -0.00022020, -0.00005757, 0.00090121, -0.00002219, -0.00023525, -0.00008489, -0.00002015, -0.00006887, 0.00008839, -0.00004397, },
		{ -0.00000431, 0.00000005, -0.00003190, 0.00002238, -0.00003728, 0.00005375, -0.00002292, -0.00000389, -0.00001392, 0.00001428, -0.00002186, 0.00001901, -0.00002042, 0.00003001, -0.00000669, -0.00000260, -0.00001280, 0.00000441, -0.00001176, 0.00000313, -0.00000600, 0.00000629, 0.00000295, 0.00000349, 0.00000454, -0.00000084, 0.00000312, -0.00000590, 0.00000370, -0.00001326, -0.00000769, -0.00000260, 0.00001246, -0.00000465, 0.00000591, -0.00000278, 0.00001115, -0.00001093, -0.00001691, -0.00000389, 0.00001436, 0.00000494, -0.00001118, 0.00000058, 0.00000134, -0.00001102, -0.00002543, -0.00000303, 0.00000543, 0.00004249, -0.00005907, -0.00003092, -0.00006446, -0.00021163, -0.00009625, 0.00095447, -0.00005757, -0.00022809, -0.00008049, -0.00002190, -0.00006763, 0.00008375, -0.00003994, -0.00000303, },
		{ -0.00000058, -0.00003033, 0.00002154, -0.00003648, 0.00005256, -0.00002224, -0.00000389, -0.00001395, 0.00001418, -0.00002186, 0.00001909, -0.00002061, 0.00003038, -0.00000695, -0.00000260, -0.00001290, 0.00000481, -0.00001215, 0.00000338, -0.00000628, 0.00000652, 0.00000289, 0.00000349, 0.00000460, -0.00000115, 0.00000353, -0.00000633, 0.00000448, -0.00001420, -0.00000741, -0.00000260, 0.00001365, -0.00000691, 0.00000771, -0.00000435, 0.00001365, -0.00001294, -0.00001716, -0.00000389, 0.00001787, 0.00000007, -0.00000840, -0.00000191, 0.00000543, -0.00001249, -0.00002851, -0.00000303, 0.00001361, 0.00003399, -0.00005701, -0.00003358, -0.00005834, -0.00020249, -0.00013786, 0.00101220, -0.00009625, -0.00022020, -0.00007558, -0.00002387, -0.00006621, 0.00007837, -0.00003519, -0.00000303, -0.00000579, },
		{ -0.00002840, 0.00002047, -0.00003541, 0.00005091, -0.00002121, -0.00000389, -0.00001400, 0.00001388, -0.00002160, 0.00001901, -0.00002061, 0.00003050, -0.00000708, -0.00000260, -0.00001296, 0.00000509, -0.00001244, 0.00000357, -0.00000650, 0.00000672, 0.00000284, 0.00000349, 0.00000466, -0.00000142, 0.00000390, -0.00000673, 0.00000519, -0.00001507, -0.00000714, -0.00000260, 0.00001476, -0.00000905, 0.00000943, -0.00000585, 0.00001606, -0.00001489, -0.00001740, -0.00000389, 0.00002133, -0.00000476, -0.00000564, -0.00000439, 0.00000953, -0.00001398, -0.00003163, -0.00000303, 0.00002200, 0.00002524, -0.00005487, -0.00003634, -0.00005193, -0.00019285, -0.00018199, 0.00107385, -0.00013786, -0.00021163, -0.00007022, -0.00002604, -0.00006463, 0.00007231, -0.00002977, -0.00000303, -0.00000753, -0.00000135, },
		{ 0.00001919, -0.00003410, 0.00004883, -0.00001986, -0.00000389, -0.00001407, 0.00001338, -0.00002109, 0.00001877, -0.00002042, 0.00003038, -0.00000708, -0.00000260, -0.00001300, 0.00000526, -0.00001264, 0.00000372, -0.00000667, 0.00000687, 0.00000281, 0.00000349, 0.00000471, -0.00000165, 0.00000423, -0.00000707, 0.00000584, -0.00001586, -0.00000690, -0.00000260, 0.00001580, -0.00001105, 0.00001105, -0.00000728, 0.00001836, -0.00001676, -0.00001764, -0.00000389, 0.00002470, -0.00000950, -0.00000291, -0.00000686, 0.00001362, -0.00001547, -0.00003477, -0.00000303, 0.00003051, 0.00001631, -0.00005268, -0.00003919, -0.00004530, -0.00018282, -0.00022822, 0.00113882, -0.00018199, -0.00020249, -0.00006446, -0.00002840, -0.00006290, 0.00006563, -0.00002373, -0.00000303, -0.00000952, -0.00000224, -0.00002612, },
		{ -0.00003255, 0.00004632, -0.00001820, -0.00000389, -0.00001417, 0.00001268, -0.00002033, 0.00001836, -0.00002004, 0.00003001, -0.00000695, -0.00000260, -0.00001300, 0.00000532, -0.00001274, 0.00000380, -0.00000679, 0.00000698, 0.00000278, 0.00000349, 0.00000475, -0.00000184, 0.00000450, -0.00000737, 0.00000640, -0.00001655, -0.00000668, -0.00000260, 0.00001676, -0.00001290, 0.00001256, -0.00000862, 0.00002053, -0.00001853, -0.00001787, -0.00000389, 0.00002795, -0.00001410, -0.00000025, -0.00000927, 0.00001764, -0.00001695, -0.00003789, -0.00000303, 0.00003906, 0.00000730, -0.00005046, -0.00004210, -0.00003850, -0.00017248, -0.00027610, 0.00120650, -0.00022822, -0.00019285, -0.00005834, -0.00003092, -0.00006104, 0.00005838, -0.00001712, -0.00000303, -0.00001174, -0.00000324, -0.00002351, 0.00001771, },
		{ 0.00004341, -0.00001625, -0.00000389, -0.00001429, 0.00001179, -0.00001934, 0.00001780, -0.00001948, 0.00002941, -0.00000669, -0.00000260, -0.00001296, 0.00000526, -0.00001274, 0.00000383, -0.00000684, 0.00000705, 0.00000276, 0.00000349, 0.00000478, -0.00000200, 0.00000473, -0.00000762, 0.00000687, -0.00001716, -0.00000649, -0.00000260, 0.00001761, -0.00001458, 0.00001394, -0.00000985, 0.00002254, -0.00002020, -0.00001808, -0.00000389, 0.00003106, -0.00001851, 0.00000232, -0.00001162, 0.00002157, -0.00001839, -0.00004096, -0.00000303, 0.00004756, -0.00000171, -0.00004823, -0.00004503, -0.00003160, -0.00016193, -0.00032517, 0.00127622, -0.00027610, -0.00018282, -0.00005193, -0.00003358, -0.00005907, 0.00005064, -0.00001001, -0.00000303, -0.00001416, -0.00000434, -0.00002061, 0.00001604, -0.00003079, },
		{ -0.00001401, -0.00000389, -0.00001443, 0.00001073, -0.00001811, 0.00001709, -0.00001873, 0.00002856, -0.00000630, -0.00000260, -0.00001290, 0.00000509, -0.00001264, 0.00000380, -0.00000684, 0.00000707, 0.00000275, 0.00000349, 0.00000480, -0.00000211, 0.00000490, -0.00000782, 0.00000726, -0.00001766, -0.00000633, -0.00000260, 0.00001836, -0.00001606, 0.00001517, -0.00001097, 0.00002439, -0.00002173, -0.00001828, -0.00000389, 0.00003399, -0.00002270, 0.00000478, -0.00001387, 0.00002536, -0.00001980, -0.00004396, -0.00000303, 0.00005595, -0.00001063, -0.00004600, -0.00004797, -0.00002466, -0.00015129, -0.00037497, 0.00134732, -0.00032517, -0.00017248, -0.00004530, -0.00003634, -0.00005701, 0.00004249, -0.00000247, -0.00000303, -0.00001677, -0.00000554, -0.00001744, 0.00001419, -0.00002882, 0.00004013, },
		{ -0.00000389, -0.00001459, 0.00000949, -0.00001666, 0.00001624, -0.00001781, 0.00002750, -0.00000578, -0.00000260, -0.00001280, 0.00000481, -0.00001244, 0.00000372, -0.00000679, 0.00000705, 0.00000275, 0.00000349, 0.00000481, -0.00000217, 0.00000501, -0.00000796, 0.00000756, -0.00001805, -0.00000619, -0.00000260, 0.00001899, -0.00001735, 0.00001626, -0.00001196, 0.00002604, -0.00002312, -0.00001846, -0.00000389, 0.00003672, -0.00002662, 0.00000709, -0.00001600, 0.00002898, -0.00002115, -0.00004686, -0.00000303, 0.00006413, -0.00001939, -0.00004381, -0.00005087, -0.00001776, -0.00014064, -0.00042501, 0.00141911, -0.00037497, -0.00016193, -0.00003850, -0.00003919, -0.00005487, 0.00003399, 0.00000543, -0.00000303, -0.00001954, -0.00000682, -0.00001402, 0.00001219, -0.00002667, 0.00003651, -0.00001152, },
		{ -0.00001478, 0.00000810, -0.00001500, 0.00001525, -0.00001673, 0.00002622, -0.00000515, -0.00000260, -0.00001266, 0.00000441, -0.00001215, 0.00000357, -0.00000667, 0.00000698, 0.00000276, 0.00000349, 0.00000481, -0.00000220, 0.00000507, -0.00000805, 0.00000775, -0.00001834, -0.00000609, -0.00000260, 0.00001950, -0.00001841, 0.00001717, -0.00001281, 0.00002749, -0.00002435, -0.00001862, -0.00000389, 0.00003921, -0.00003024, 0.00000924, -0.00001800, 0.00003239, -0.00002243, -0.00004963, -0.00000303, 0.00007203, -0.00002788, -0.00004167, -0.00005372, -0.00001096, -0.00013010, -0.00047481, 0.00149090, -0.00042501, -0.00015129, -0.00003160, -0.00004210, -0.00005268, 0.00002524, 0.00001361, -0.00000303, -0.00002243, -0.00000817, -0.00001040, 0.00001006, -0.00002436, 0.00003259, -0.00000879, -0.00000389, },
		{ 0.00000657, -0.00001316, 0.00001413, -0.00001549, 0.00002473, -0.00000440, -0.00000260, -0.00001250, 0.00000391, -0.00001176, 0.00000338, -0.00000650, 0.00000687, 0.00000278, 0.00000349, 0.00000480, -0.00000217, 0.00000507, -0.00000808, 0.00000785, -0.00001851, -0.00000603, -0.00000260, 0.00001989, -0.00001925, 0.00001792, -0.00001352, 0.00002872, -0.00002542, -0.00001876, -0.00000389, 0.00004144, -0.00003352, 0.00001120, -0.00001985, 0.00003557, -0.00002362, -0.00005223, -0.00000303, 0.00007957, -0.00003604, -0.00003961, -0.00005649, -0.00000432, -0.00011975, -0.00052388, 0.00156200, -0.00047481, -0.00014064, -0.00002466, -0.00004503, -0.00005046, 0.00001631, 0.00002200, -0.00000303, -0.00002543, -0.00000957, -0.00000661, 0.00000780, -0.00002191, 0.00002840, -0.00000586, -0.00000389, -0.00001497, },
		{ -0.00001114, 0.00001290, -0.00001411, 0.00002305, -0.00000355, -0.00000260, -0.00001231, 0.00000331, -0.00001129, 0.00000313, -0.00000628, 0.00000672, 0.00000281, 0.00000349, 0.00000478, -0.00000211, 0.00000501, -0.00000805, 0.00000785, -0.00001857, -0.00000599, -0.00000260, 0.00002016, -0.00001986, 0.00001849, -0.00001408, 0.00002972, -0.00002630, -0.00001888, -0.00000389, 0.00004340, -0.00003643, 0.00001297, -0.00002152, 0.00003847, -0.00002473, -0.00005466, -0.00000303, 0.00008668, -0.00004378, -0.00003764, -0.00005914, 0.00000208, -0.00010972, -0.00057177, 0.00163172, -0.00052388, -0.00013010, -0.00001776, -0.00004797, -0.00004823, 0.00000730, 0.00003051, -0.00000303, -0.00002851, -0.00001102, -0.00000268, 0.00000546, -0.00001934, 0.00002399, -0.00000276, -0.00000389, -0.00001519, 0.00000491, },
		{ 0.00001156, -0.00001261, 0.00002120, -0.00000260, -0.00000260, -0.00001209, 0.00000261, -0.00001072, 0.00000283, -0.00000600, 0.00000652, 0.00000284, 0.00000349, 0.00000475, -0.00000200, 0.00000490, -0.00000796, 0.00000775, -0.00001851, -0.00000599, -0.00000260, 0.00002029, -0.00002022, 0.00001887, -0.00001449, 0.00003048, -0.00002700, -0.00001898, -0.00000389, 0.00004506, -0.00003894, 0.00001451, -0.00002300, 0.00004107, -0.00002573, -0.00005687, -0.00000303, 0.00009329, -0.00005102, -0.00003578, -0.00006166, 0.00000820, -0.00010008, -0.00061799, 0.00169940, -0.00057177, -0.00011975, -0.00001096, -0.00005087, -0.00004600, -0.00000171, 0.00003906, -0.00000303, -0.00003163, -0.00001249, 0.00000134, 0.00000304, -0.00001668, 0.00001939, 0.00000050, -0.00000389, -0.00001541, 0.00000313, -0.00000897, },
		{ -0.00001099, 0.00001920, -0.00000156, -0.00000260, -0.00001185, 0.00000182, -0.00001008, 0.00000248, -0.00000568, 0.00000629, 0.00000289, 0.00000349, 0.00000471, -0.00000184, 0.00000473, -0.00000782, 0.00000756, -0.00001834, -0.00000603, -0.00000260, 0.00002029, -0.00002035, 0.00001906, -0.00001473, 0.00003099, -0.00002750, -0.00001905, -0.00000389, 0.00004641, -0.00004103, 0.00001582, -0.00002428, 0.00004335, -0.00002662, -0.00005886, -0.00000303, 0.00009933, -0.00005771, -0.00003406, -0.00006402, 0.00001397, -0.00009094, -0.00066212, 0.00176437, -0.00061799, -0.00010972, -0.00000432, -0.00005372, -0.00004381, -0.00001063, 0.00004756, -0.00000303, -0.00003477, -0.00001398, 0.00000543, 0.00000058, -0.00001395, 0.00001465, 0.00000387, -0.00000389, -0.00001565, 0.00000126, -0.00000667, 0.00001013, },
		{ 0.00001706, -0.00000044, -0.00000260, -0.00001158, 0.00000095, -0.00000937, 0.00000209, -0.00000531, 0.00000602, 0.00000295, 0.00000349, 0.00000466, -0.00000165, 0.00000450, -0.00000762, 0.00000726, -0.00001805, -0.00000609, -0.00000260, 0.00002016, -0.00002022, 0.00001906, -0.00001481, 0.00003124, -0.00002781, -0.00001910, -0.00000389, 0.00004744, -0.00004267, 0.00001689, -0.00002535, 0.00004529, -0.00002738, -0.00006061, -0.00000303, 0.00010476, -0.00006377, -0.00003248, -0.00006620, 0.00001933, -0.00008238, -0.00070373, 0.00182602, -0.00066212, -0.00010008, 0.00000208, -0.00005649, -0.00004167, -0.00001939, 0.00005595, -0.00000303, -0.00003789, -0.00001547, 0.00000953, -0.00000191, -0.00001118, 0.00000982, 0.00000733, -0.00000389, -0.00001589, -0.00000069, -0.00000426, 0.00000863, -0.00000927, },
		{ 0.00000074, -0.00000260, -0.00001130, 0.00000001, -0.00000859, 0.00000165, -0.00000489, 0.00000571, 0.00000301, 0.00000349, 0.00000460, -0.00000142, 0.00000423, -0.00000737, 0.00000687, -0.00001766, -0.00000619, -0.00000260, 0.00001989, -0.00001986, 0.00001887, -0.00001473, 0.00003124, -0.00002791, -0.00001913, -0.00000389, 0.00004812, -0.00004386, 0.00001770, -0.00002619, 0.00004686, -0.00002802, -0.00006208, -0.00000303, 0.00010951, -0.00006914, -0.00003106, -0.00006817, 0.00002423, -0.00007448, -0.00074241, 0.00188375, -0.00070373, -0.00009094, 0.00000820, -0.00005914, -0.00003961, -0.00002788, 0.00006413, -0.00000303, -0.00004096, -0.00001695, 0.00001362, -0.00000439, -0.00000840, 0.00000494, 0.00001084, -0.00000389, -0.00001614, -0.00000270, -0.00000177, 0.00000706, -0.00000747, 0.00001481, },
		{ -0.00000260, -0.00001099, -0.00000099, -0.00000774, 0.00000118, -0.00000444, 0.00000538, 0.00000308, 0.00000349, 0.00000454, -0.00000115, 0.00000390, -0.00000707, 0.00000640, -0.00001716, -0.00000633, -0.00000260, 0.00001950, -0.00001925, 0.00001849, -0.00001449, 0.00003099, -0.00002781, -0.00001913, -0.00000389, 0.00004847, -0.00004458, 0.00001824, -0.00002679, 0.00004805, -0.00002852, -0.00006328, -0.00000303, 0.00011354, -0.00007378, -0.00002982, -0.00006992, 0.00002863, -0.00006732, -0.00077779, 0.00193701, -0.00074241, -0.00008238, 0.00001397, -0.00006166, -0.00003764, -0.00003604, 0.00007203, -0.00000303, -0.00004396, -0.00001839, 0.00001764, -0.00000686, -0.00000564, 0.00000007, 0.00001436, -0.00000389, -0.00001640, -0.00000474, 0.00000079, 0.00000545, -0.00000560, 0.00001246, 0.00000199, },
		{ -0.00001068, -0.00000205, -0.00000685, 0.00000068, -0.00000395, 0.00000501, 0.00000316, 0.00000349, 0.00000447, -0.00000084, 0.00000353, -0.00000673, 0.00000584, -0.00001655, -0.00000649, -0.00000260, 0.00001899, -0.00001841, 0.00001792, -0.00001408, 0.00003048, -0.00002750, -0.00001910, -0.00000389, 0.00004847, -0.00004481, 0.00001851, -0.00002716, 0.00004885, -0.00002889, -0.00006419, -0.00000303, 0.00011681, -0.00007764, -0.00002877, -0.00007143, 0.00003248, -0.00006097, -0.00080953, 0.00198528, -0.00077779, -0.00007448, 0.00001933, -0.00006402, -0.00003578, -0.00004378, 0.00007957, -0.00000303, -0.00004686, -0.00001980, 0.00002157, -0.00000927, -0.00000291, -0.00000476, 0.00001787, -0.00000389, -0.00001665, -0.00000681, 0.00000339, 0.00000380, -0.00000369, 0.00001004, 0.00000327, -0.00000260, },
		{ -0.00000316, -0.00000592, 0.00000016, -0.00000344, 0.00000462, 0.00000325, 0.00000349, 0.00000439, -0.00000050, 0.00000312, -0.00000633, 0.00000519, -0.00001586, -0.00000668, -0.00000260, 0.00001836, -0.00001735, 0.00001717, -0.00001352, 0.00002972, -0.00002700, -0.00001905, -0.00000389, 0.00004812, -0.00004458, 0.00001851, -0.00002728, 0.00004925, -0.00002910, -0.00006481, -0.00000303, 0.00011929, -0.00008068, -0.00002791, -0.00007269, 0.00003575, -0.00005549, -0.00083733, 0.00202810, -0.00080953, -0.00006732, 0.00002423, -0.00006620, -0.00003406, -0.00005102, 0.00008668, -0.00000303, -0.00004963, -0.00002115, 0.00002536, -0.00001162, -0.00000025, -0.00000950, 0.00002133, -0.00000389, -0.00001691, -0.00000888, 0.00000600, 0.00000214, -0.00000175, 0.00000757, 0.00000459, -0.00000260, -0.00001035, },
		{ -0.00000495, -0.00000039, -0.00000290, 0.00000421, 0.00000334, 0.00000349, 0.00000431, -0.00000014, 0.00000267, -0.00000590, 0.00000448, -0.00001507, -0.00000690, -0.00000260, 0.00001761, -0.00001606, 0.00001626, -0.00001281, 0.00002872, -0.00002630, -0.00001898, -0.00000389, 0.00004744, -0.00004386, 0.00001824, -0.00002716, 0.00004925, -0.00002918, -0.00006511, -0.00000303, 0.00012096, -0.00008287, -0.00002726, -0.00007367, 0.00003840, -0.00005093, -0.00086092, 0.00206505, -0.00083733, -0.00006097, 0.00002863, -0.00006817, -0.00003248, -0.00005771, 0.00009329, -0.00000303, -0.00005223, -0.00002243, 0.00002898, -0.00001387, 0.00000232, -0.00001410, 0.00002470, -0.00000389, -0.00001716, -0.00001093, 0.00000859, 0.00000048, 0.00000020, 0.00000508, 0.00000593, -0.00000260, -0.00001001, -0.00000429, },
		{ -0.00000096, -0.00000234, 0.00000378, 0.00000343, 0.00000349, 0.00000422, 0.00000025, 0.00000218, -0.00000543, 0.00000370, -0.00001420, -0.00000714, -0.00000260, 0.00001676, -0.00001458, 0.00001517, -0.00001196, 0.00002749, -0.00002542, -0.00001888, -0.00000389, 0.00004641, -0.00004267, 0.00001770, -0.00002679, 0.00004885, -0.00002910, -0.00006511, -0.00000303, 0.00012180, -0.00008420, -0.00002683, -0.00007439, 0.00004042, -0.00004734, -0.00088007, 0.00209578, -0.00086092, -0.00005549, 0.00003248, -0.00006992, -0.00003106, -0.00006377, 0.00009933, -0.00000303, -0.00005466, -0.00002362, 0.00003239, -0.00001600, 0.00000478, -0.00001851, 0.00002795, -0.00000389, -0.00001740, -0.00001294, 0.00001115, -0.00000117, 0.00000214, 0.00000258, 0.00000727, -0.00000260, -0.00000967, -0.00000545, -0.00000396, },
		{ -0.00000176, 0.00000334, 0.00000353, 0.00000349, 0.00000412, 0.00000066, 0.00000166, -0.00000493, 0.00000286, -0.00001326, -0.00000741, -0.00000260, 0.00001580, -0.00001290, 0.00001394, -0.00001097, 0.00002604, -0.00002435, -0.00001876, -0.00000389, 0.00004506, -0.00004103, 0.00001689, -0.00002619, 0.00004805, -0.00002889, -0.00006481, -0.00000303, 0.00012180, -0.00008464, -0.00002661, -0.00007482, 0.00004177, -0.00004475, -0.00089459, 0.00211999, -0.00088007, -0.00005093, 0.00003575, -0.00007143, -0.00002982, -0.00006914, 0.00010476, -0.00000303, -0.00005687, -0.00002473, 0.00003557, -0.00001800, 0.00000709, -0.00002270, 0.00003106, -0.00000389, -0.00001764, -0.00001489, 0.00001365, -0.00000278, 0.00000405, 0.00000012, 0.00000861, -0.00000260, -0.00000932, -0.00000662, -0.00000295, -0.00000154, },
		{ 0.00000289, 0.00000363, 0.00000349, 0.00000403, 0.00000109, 0.00000112, -0.00000440, 0.00000196, -0.00001225, -0.00000769, -0.00000260, 0.00001476, -0.00001105, 0.00001256, -0.00000985, 0.00002439, -0.00002312, -0.00001862, -0.00000389, 0.00004340, -0.00003894, 0.00001582, -0.00002535, 0.00004686, -0.00002852, -0.00006419, -0.00000303, 0.00012096, -0.00008420, -0.00002661, -0.00007496, 0.00004245, -0.00004319, -0.00090436, 0.00213746, -0.00089459, -0.00004734, 0.00003840, -0.00007269, -0.00002877, -0.00007378, 0.00010951, -0.00000303, -0.00005886, -0.00002573, 0.00003847, -0.00001985, 0.00000924, -0.00002662, 0.00003399, -0.00000389, -0.00001787, -0.00001676, 0.00001606, -0.00000435, 0.00000591, -0.00000230, 0.00000993, -0.00000260, -0.00000898, -0.00000779, -0.00000194, -0.00000212, -0.00000118, },
		{ 0.00000373, 0.00000349, 0.00000393, 0.00000153, 0.00000056, -0.00000385, 0.00000103, -0.00001119, -0.00000800, -0.00000260, 0.00001365, -0.00000905, 0.00001105, -0.00000862, 0.00002254, -0.00002173, -0.00001846, -0.00000389, 0.00004144, -0.00003643, 0.00001451, -0.00002428, 0.00004529, -0.00002802, -0.00006328, -0.00000303, 0.00011929, -0.00008287, -0.00002683, -0.00007482, 0.00004245, -0.00004266, -0.00090926, 0.00214801, -0.00090436, -0.00004475, 0.00004042, -0.00007367, -0.00002791, -0.00007764, 0.00011354, -0.00000303, -0.00006061, -0.00002662, 0.00004107, -0.00002152, 0.00001120, -0.00003024, 0.00003672, -0.00000389, -0.00001808, -0.00001853, 0.00001836, -0.00000585, 0.00000771, -0.00000465, 0.00001122, -0.00000260, -0.00000864, -0.00000895, -0.00000093, -0.00000271, -0.00000059, 0.00000244, },
		{ 0.00000349, 0.00000383, 0.00000198, -0.00000001, -0.00000328, 0.00000006, -0.00001009, -0.00000831, -0.00000260, 0.00001246, -0.00000691, 0.00000943, -0.00000728, 0.00002053, -0.00002020, -0.00001828, -0.00000389, 0.00003921, -0.00003352, 0.00001297, -0.00002300, 0.00004335, -0.00002738, -0.00006208, -0.00000303, 0.00011681, -0.00008068, -0.00002726, -0.00007439, 0.00004177, -0.00004319, -0.00090926, 0.00215153, -0.00090926, -0.00004319, 0.00004177, -0.00007439, -0.00002726, -0.00008068, 0.00011681, -0.00000303, -0.00006208, -0.00002738, 0.00004335, -0.00002300, 0.00001297, -0.00003352, 0.00003921, -0.00000389, -0.00001828, -0.00002020, 0.00002053, -0.00000728, 0.00000943, -0.00000691, 0.00001246, -0.00000260, -0.00000831, -0.00001009, 0.00000006, -0.00000328, -0.00000001, 0.00000198, 0.00000383, },
		{ 0.00000373, 0.00000244, -0.00000059, -0.00000271, -0.00000093, -0.00000895, -0.00000864, -0.00000260, 0.00001122, -0.00000465, 0.00000771, -0.00000585, 0.00001836, -0.00001853, -0.00001808, -0.00000389, 0.00003672, -0.00003024, 0.00001120, -0.00002152, 0.00004107, -0.00002662, -0.00006061, -0.00000303, 0.00011354, -0.00007764, -0.00002791, -0.00007367, 0.00004042, -0.00004475, -0.00090436, 0.00214801, -0.00090926, -0.00004266, 0.00004245, -0.00007482, -0.00002683, -0.00008287, 0.00011929, -0.00000303, -0.00006328, -0.00002802, 0.00004529, -0.00002428, 0.00001451, -0.00003643, 0.00004144, -0.00000389, -0.00001846, -0.00002173, 0.00002254, -0.00000862, 0.00001105, -0.00000905, 0.00001365, -0.00000260, -0.00000800, -0.00001119, 0.00000103, -0.00000385, 0.00000056, 0.00000153, 0.00000393, 0.00000349, },
		{ 0.00000289, -0.00000118, -0.00000212, -0.00000194, -0.00000779, -0.00000898, -0.00000260, 0.00000993, -0.00000230, 0.00000591, -0.00000435, 0.00001606, -0.00001676, -0.00001787, -0.00000389, 0.00003399, -0.00002662, 0.00000924, -0.00001985, 0.00003847, -0.00002573, -0.00005886, -0.00000303, 0.00010951, -0.00007378, -0.00002877, -0.00007269, 0.00003840, -0.00004734, -0.00089459, 0.00213746, -0.00090436, -0.00004319, 0.00004245, -0.00007496, -0.00002661, -0.00008420, 0.00012096, -0.00000303, -0.00006419, -0.00002852, 0.00004686, -0.00002535, 0.00001582, -0.00003894, 0.00004340, -0.00000389, -0.00001862, -0.00002312, 0.00002439, -0.00000985, 0.00001256, -0.00001105, 0.00001476, -0.00000260, -0.00000769, -0.00001225, 0.00000196, -0.00000440, 0.00000112, 0.00000109, 0.00000403, 0.00000349, 0.00000363, },
		{ -0.00000176, -0.00000154, -0.00000295, -0.00000662, -0.00000932, -0.00000260, 0.00000861, 0.00000012, 0.00000405, -0.00000278, 0.00001365, -0.00001489, -0.00001764, -0.00000389, 0.00003106, -0.00002270, 0.00000709, -0.00001800, 0.00003557, -0.00002473, -0.00005687, -0.00000303, 0.00010476, -0.00006914, -0.00002982, -0.00007143, 0.00003575, -0.00005093, -0.00088007, 0.00211999, -0.00089459, -0.00004475, 0.00004177, -0.00007482, -0.00002661, -0.00008464, 0.00012180, -0.00000303, -0.00006481, -0.00002889, 0.00004805, -0.00002619, 0.00001689, -0.00004103, 0.00004506, -0.00000389, -0.00001876, -0.00002435, 0.00002604, -0.00001097, 0.00001394, -0.00001290, 0.00001580, -0.00000260, -0.00000741, -0.00001326, 0.00000286, -0.00000493, 0.00000166, 0.00000066, 0.00000412, 0.00000349, 0.00000353, 0.00000334, },
		{ -0.00000096, -0.00000396, -0.00000545, -0.00000967, -0.00000260, 0.00000727, 0.00000258, 0.00000214, -0.00000117, 0.00001115, -0.00001294, -0.00001740, -0.00000389, 0.00002795, -0.00001851, 0.00000478, -0.00001600, 0.00003239, -0.00002362, -0.00005466, -0.00000303, 0.00009933, -0.00006377, -0.00003106, -0.00006992, 0.00003248, -0.00005549, -0.00086092, 0.00209578, -0.00088007, -0.00004734, 0.00004042, -0.00007439, -0.00002683, -0.00008420, 0.00012180, -0.00000303, -0.00006511, -0.00002910, 0.00004885, -0.00002679, 0.00001770, -0.00004267, 0.00004641, -0.00000389, -0.00001888, -0.00002542, 0.00002749, -0.00001196, 0.00001517, -0.00001458, 0.00001676, -0.00000260, -0.00000714, -0.00001420, 0.00000370, -0.00000543, 0.00000218, 0.00000025, 0.00000422, 0.00000349, 0.00000343, 0.00000378, -0.00000234, },
		{ -0.00000495, -0.00000429, -0.00001001, -0.00000260, 0.00000593, 0.00000508, 0.00000020, 0.00000048, 0.00000859, -0.00001093, -0.00001716, -0.00000389, 0.00002470, -0.00001410, 0.00000232, -0.00001387, 0.00002898, -0.00002243, -0.00005223, -0.00000303, 0.00009329, -0.00005771, -0.00003248, -0.00006817, 0.00002863, -0.00006097, -0.00083733, 0.00206505, -0.00086092, -0.00005093, 0.00003840, -0.00007367, -0.00002726, -0.00008287, 0.00012096, -0.00000303, -0.00006511, -0.00002918, 0.00004925, -0.00002716, 0.00001824, -0.00004386, 0.00004744, -0.00000389, -0.00001898, -0.00002630, 0.00002872, -0.00001281, 0.00001626, -0.00001606, 0.00001761, -0.00000260, -0.00000690, -0.00001507, 0.00000448, -0.00000590, 0.00000267, -0.00000014, 0.00000431, 0.00000349, 0.00000334, 0.00000421, -0.00000290, -0.00000039, },
		{ -0.00000316, -0.00001035, -0.00000260, 0.00000459, 0.00000757, -0.00000175, 0.00000214, 0.00000600, -0.00000888, -0.00001691, -0.00000389, 0.00002133, -0.00000950, -0.00000025, -0.00001162, 0.00002536, -0.00002115, -0.00004963, -0.00000303, 0.00008668, -0.00005102, -0.00003406, -0.00006620, 0.00002423, -0.00006732, -0.00080953, 0.00202810, -0.00083733, -0.00005549, 0.00003575, -0.00007269, -0.00002791, -0.00008068, 0.00011929, -0.00000303, -0.00006481, -0.00002910, 0.00004925, -0.00002728, 0.00001851, -0.00004458, 0.00004812, -0.00000389, -0.00001905, -0.00002700, 0.00002972, -0.00001352, 0.00001717, -0.00001735, 0.00001836, -0.00000260, -0.00000668, -0.00001586, 0.00000519, -0.00000633, 0.00000312, -0.00000050, 0.00000439, 0.00000349, 0.00000325, 0.00000462, -0.00000344, 0.00000016, -0.00000592, },
		{ -0.00001068, -0.00000260, 0.00000327, 0.00001004, -0.00000369, 0.00000380, 0.00000339, -0.00000681, -0.00001665, -0.00000389, 0.00001787, -0.00000476, -0.00000291, -0.00000927, 0.00002157, -0.00001980, -0.00004686, -0.00000303, 0.00007957, -0.00004378, -0.00003578, -0.00006402, 0.00001933, -0.00007448, -0.00077779, 0.00198528, -0.00080953, -0.00006097, 0.00003248, -0.00007143, -0.00002877, -0.00007764, 0.00011681, -0.00000303, -0.00006419, -0.00002889, 0.00004885, -0.00002716, 0.00001851, -0.00004481, 0.00004847, -0.00000389, -0.00001910, -0.00002750, 0.00003048, -0.00001408, 0.00001792, -0.00001841, 0.00001899, -0.00000260, -0.00000649, -0.00001655, 0.00000584, -0.00000673, 0.00000353, -0.00000084, 0.00000447, 0.00000349, 0.00000316, 0.00000501, -0.00000395, 0.00000068, -0.00000685, -0.00000205, },
		{ -0.00000260, 0.00000199, 0.00001246, -0.00000560, 0.00000545, 0.00000079, -0.00000474, -0.00001640, -0.00000389, 0.00001436, 0.00000007, -0.00000564, -0.00000686, 0.00001764, -0.00001839, -0.00004396, -0.00000303, 0.00007203, -0.00003604, -0.00003764, -0.00006166, 0.00001397, -0.00008238, -0.00074241, 0.00193701, -0.00077779, -0.00006732, 0.00002863, -0.00006992, -0.00002982, -0.00007378, 0.00011354, -0.00000303, -0.00006328, -0.00002852, 0.00004805, -0.00002679, 0.00001824, -0.00004458, 0.00004847, -0.00000389, -0.00001913, -0.00002781, 0.00003099, -0.00001449, 0.00001849, -0.00001925, 0.00001950, -0.00000260, -0.00000633, -0.00001716, 0.00000640, -0.00000707, 0.00000390, -0.00000115, 0.00000454, 0.00000349, 0.00000308, 0.00000538, -0.00000444, 0.00000118, -0.00000774, -0.00000099, -0.00001099, },
		{ 0.00000074, 0.00001481, -0.00000747, 0.00000706, -0.00000177, -0.00000270, -0.00001614, -0.00000389, 0.00001084, 0.00000494, -0.00000840, -0.00000439, 0.00001362, -0.00001695, -0.00004096, -0.00000303, 0.00006413, -0.00002788, -0.00003961, -0.00005914, 0.00000820, -0.00009094, -0.00070373, 0.00188375, -0.00074241, -0.00007448, 0.00002423, -0.00006817, -0.00003106, -0.00006914, 0.00010951, -0.00000303, -0.00006208, -0.00002802, 0.00004686, -0.00002619, 0.00001770, -0.00004386, 0.00004812, -0.00000389, -0.00001913, -0.00002791, 0.00003124, -0.00001473, 0.00001887, -0.00001986, 0.00001989, -0.00000260, -0.00000619, -0.00001766, 0.00000687, -0.00000737, 0.00000423, -0.00000142, 0.00000460, 0.00000349, 0.00000301, 0.00000571, -0.00000489, 0.00000165, -0.00000859, 0.00000001, -0.00001130, -0.00000260, },
		{ 0.00001706, -0.00000927, 0.00000863, -0.00000426, -0.00000069, -0.00001589, -0.00000389, 0.00000733, 0.00000982, -0.00001118, -0.00000191, 0.00000953, -0.00001547, -0.00003789, -0.00000303, 0.00005595, -0.00001939, -0.00004167, -0.00005649, 0.00000208, -0.00010008, -0.00066212, 0.00182602, -0.00070373, -0.00008238, 0.00001933, -0.00006620, -0.00003248, -0.00006377, 0.00010476, -0.00000303, -0.00006061, -0.00002738, 0.00004529, -0.00002535, 0.00001689, -0.00004267, 0.00004744, -0.00000389, -0.00001910, -0.00002781, 0.00003124, -0.00001481, 0.00001906, -0.00002022, 0.00002016, -0.00000260, -0.00000609, -0.00001805, 0.00000726, -0.00000762, 0.00000450, -0.00000165, 0.00000466, 0.00000349, 0.00000295, 0.00000602, -0.00000531, 0.00000209, -0.00000937, 0.00000095, -0.00001158, -0.00000260, -0.00000044, },
		{ -0.00001099, 0.00001013, -0.00000667, 0.00000126, -0.00001565, -0.00000389, 0.00000387, 0.00001465, -0.00001395, 0.00000058, 0.00000543, -0.00001398, -0.00003477, -0.00000303, 0.00004756, -0.00001063, -0.00004381, -0.00005372, -0.00000432, -0.00010972, -0.00061799, 0.00176437, -0.00066212, -0.00009094, 0.00001397, -0.00006402, -0.00003406, -0.00005771, 0.00009933, -0.00000303, -0.00005886, -0.00002662, 0.00004335, -0.00002428, 0.00001582, -0.00004103, 0.00004641, -0.00000389, -0.00001905, -0.00002750, 0.00003099, -0.00001473, 0.00001906, -0.00002035, 0.00002029, -0.00000260, -0.00000603, -0.00001834, 0.00000756, -0.00000782, 0.00000473, -0.00000184, 0.00000471, 0.00000349, 0.00000289, 0.00000629, -0.00000568, 0.00000248, -0.00001008, 0.00000182, -0.00001185, -0.00000260, -0.00000156, 0.00001920, },
		{ 0.00001156, -0.00000897, 0.00000313, -0.00001541, -0.00000389, 0.00000050, 0.00001939, -0.00001668, 0.00000304, 0.00000134, -0.00001249, -0.00003163, -0.00000303, 0.00003906, -0.00000171, -0.00004600, -0.00005087, -0.00001096, -0.00011975, -0.00057177, 0.00169940, -0.00061799, -0.00010008, 0.00000820, -0.00006166, -0.00003578, -0.00005102, 0.00009329, -0.00000303, -0.00005687, -0.00002573, 0.00004107, -0.00002300, 0.00001451, -0.00003894, 0.00004506, -0.00000389, -0.00001898, -0.00002700, 0.00003048, -0.00001449, 0.00001887, -0.00002022, 0.00002029, -0.00000260, -0.00000599, -0.00001851, 0.00000775, -0.00000796, 0.00000490, -0.00000200, 0.00000475, 0.00000349, 0.00000284, 0.00000652, -0.00000600, 0.00000283, -0.00001072, 0.00000261, -0.00001209, -0.00000260, -0.00000260, 0.00002120, -0.00001261, },
		{ -0.00001114, 0.00000491, -0.00001519, -0.00000389, -0.00000276, 0.00002399, -0.00001934, 0.00000546, -0.00000268, -0.00001102, -0.00002851, -0.00000303, 0.00003051, 0.00000730, -0.00004823, -0.00004797, -0.00001776, -0.00013010, -0.00052388, 0.00163172, -0.00057177, -0.00010972, 0.00000208, -0.00005914, -0.00003764, -0.00004378, 0.00008668, -0.00000303, -0.00005466, -0.00002473, 0.00003847, -0.00002152, 0.00001297, -0.00003643, 0.00004340, -0.00000389, -0.00001888, -0.00002630, 0.00002972, -0.00001408, 0.00001849, -0.00001986, 0.00002016, -0.00000260, -0.00000599, -0.00001857, 0.00000785, -0.00000805, 0.00000501, -0.00000211, 0.00000478, 0.00000349, 0.00000281, 0.00000672, -0.00000628, 0.00000313, -0.00001129, 0.00000331, -0.00001231, -0.00000260, -0.00000355, 0.00002305, -0.00001411, 0.00001290, },
		{ 0.00000657, -0.00001497, -0.00000389, -0.00000586, 0.00002840, -0.00002191, 0.00000780, -0.00000661, -0.00000957, -0.00002543, -0.00000303, 0.00002200, 0.00001631, -0.00005046, -0.00004503, -0.00002466, -0.00014064, -0.00047481, 0.00156200, -0.00052388, -0.00011975, -0.00000432, -0.00005649, -0.00003961, -0.00003604, 0.00007957, -0.00000303, -0.00005223, -0.00002362, 0.00003557, -0.00001985, 0.00001120, -0.00003352, 0.00004144, -0.00000389, -0.00001876, -0.00002542, 0.00002872, -0.00001352, 0.00001792, -0.00001925, 0.00001989, -0.00000260, -0.00000603, -0.00001851, 0.00000785, -0.00000808, 0.00000507, -0.00000217, 0.00000480, 0.00000349, 0.00000278, 0.00000687, -0.00000650, 0.00000338, -0.00001176, 0.00000391, -0.00001250, -0.00000260, -0.00000440, 0.00002473, -0.00001549, 0.00001413, -0.00001316, },
		{ -0.00001478, -0.00000389, -0.00000879, 0.00003259, -0.00002436, 0.00001006, -0.00001040, -0.00000817, -0.00002243, -0.00000303, 0.00001361, 0.00002524, -0.00005268, -0.00004210, -0.00003160, -0.00015129, -0.00042501, 0.00149090, -0.00047481, -0.00013010, -0.00001096, -0.00005372, -0.00004167, -0.00002788, 0.00007203, -0.00000303, -0.00004963, -0.00002243, 0.00003239, -0.00001800, 0.00000924, -0.00003024, 0.00003921, -0.00000389, -0.00001862, -0.00002435, 0.00002749, -0.00001281, 0.00001717, -0.00001841, 0.00001950, -0.00000260, -0.00000609, -0.00001834, 0.00000775, -0.00000805, 0.00000507, -0.00000220, 0.00000481, 0.00000349, 0.00000276, 0.00000698, -0.00000667, 0.00000357, -0.00001215, 0.00000441, -0.00001266, -0.00000260, -0.00000515, 0.00002622, -0.00001673, 0.00001525, -0.00001500, 0.00000810, },
		{ -0.00000389, -0.00001152, 0.00003651, -0.00002667, 0.00001219, -0.00001402, -0.00000682, -0.00001954, -0.00000303, 0.00000543, 0.00003399, -0.00005487, -0.00003919, -0.00003850, -0.00016193, -0.00037497, 0.00141911, -0.00042501, -0.00014064, -0.00001776, -0.00005087, -0.00004381, -0.00001939, 0.00006413, -0.00000303, -0.00004686, -0.00002115, 0.00002898, -0.00001600, 0.00000709, -0.00002662, 0.00003672, -0.00000389, -0.00001846, -0.00002312, 0.00002604, -0.00001196, 0.00001626, -0.00001735, 0.00001899, -0.00000260, -0.00000619, -0.00001805, 0.00000756, -0.00000796, 0.00000501, -0.00000217, 0.00000481, 0.00000349, 0.00000275, 0.00000705, -0.00000679, 0.00000372, -0.00001244, 0.00000481, -0.00001280, -0.00000260, -0.00000578, 0.00002750, -0.00001781, 0.00001624, -0.00001666, 0.00000949, -0.00001459, },
		{ -0.00001401, 0.00004013, -0.00002882, 0.00001419, -0.00001744, -0.00000554, -0.00001677, -0.00000303, -0.00000247, 0.00004249, -0.00005701, -0.00003634, -0.00004530, -0.00017248, -0.00032517, 0.00134732, -0.00037497, -0.00015129, -0.00002466, -0.00004797, -0.00004600, -0.00001063, 0.00005595, -0.00000303, -0.00004396, -0.00001980, 0.00002536, -0.00001387, 0.00000478, -0.00002270, 0.00003399, -0.00000389, -0.00001828, -0.00002173, 0.00002439, -0.00001097, 0.00001517, -0.00001606, 0.00001836, -0.00000260, -0.00000633, -0.00001766, 0.00000726, -0.00000782, 0.00000490, -0.00000211, 0.00000480, 0.00000349, 0.00000275, 0.00000707, -0.00000684, 0.00000380, -0.00001264, 0.00000509, -0.00001290, -0.00000260, -0.00000630, 0.00002856, -0.00001873, 0.00001709, -0.00001811, 0.00001073, -0.00001443, -0.00000389, },
		{ 0.00004341, -0.00003079, 0.00001604, -0.00002061, -0.00000434, -0.00001416, -0.00000303, -0.00001001, 0.00005064, -0.00005907, -0.00003358, -0.00005193, -0.00018282, -0.00027610, 0.00127622, -0.00032517, -0.00016193, -0.00003160, -0.00004503, -0.00004823, -0.00000171, 0.00004756, -0.00000303, -0.00004096, -0.00001839, 0.00002157, -0.00001162, 0.00000232, -0.00001851, 0.00003106, -0.00000389, -0.00001808, -0.00002020, 0.00002254, -0.00000985, 0.00001394, -0.00001458, 0.00001761, -0.00000260, -0.00000649, -0.00001716, 0.00000687, -0.00000762, 0.00000473, -0.00000200, 0.00000478, 0.00000349, 0.00000276, 0.00000705, -0.00000684, 0.00000383, -0.00001274, 0.00000526, -0.00001296, -0.00000260, -0.00000669, 0.00002941, -0.00001948, 0.00001780, -0.00001934, 0.00001179, -0.00001429, -0.00000389, -0.00001625, },
		{ -0.00003255, 0.00001771, -0.00002351, -0.00000324, -0.00001174, -0.00000303, -0.00001712, 0.00005838, -0.00006104, -0.00003092, -0.00005834, -0.00019285, -0.00022822, 0.00120650, -0.00027610, -0.00017248, -0.00003850, -0.00004210, -0.00005046, 0.00000730, 0.00003906, -0.00000303, -0.00003789, -0.00001695, 0.00001764, -0.00000927, -0.00000025, -0.00001410, 0.00002795, -0.00000389, -0.00001787, -0.00001853, 0.00002053, -0.00000862, 0.00001256, -0.00001290, 0.00001676, -0.00000260, -0.00000668, -0.00001655, 0.00000640, -0.00000737, 0.00000450, -0.00000184, 0.00000475, 0.00000349, 0.00000278, 0.00000698, -0.00000679, 0.00000380, -0.00001274, 0.00000532, -0.00001300, -0.00000260, -0.00000695, 0.00003001, -0.00002004, 0.00001836, -0.00002033, 0.00001268, -0.00001417, -0.00000389, -0.00001820, 0.00004632, },
		{ 0.00001919, -0.00002612, -0.00000224, -0.00000952, -0.00000303, -0.00002373, 0.00006563, -0.00006290, -0.00002840, -0.00006446, -0.00020249, -0.00018199, 0.00113882, -0.00022822, -0.00018282, -0.00004530, -0.00003919, -0.00005268, 0.00001631, 0.00003051, -0.00000303, -0.00003477, -0.00001547, 0.00001362, -0.00000686, -0.00000291, -0.00000950, 0.00002470, -0.00000389, -0.00001764, -0.00001676, 0.00001836, -0.00000728, 0.00001105, -0.00001105, 0.00001580, -0.00000260, -0.00000690, -0.00001586, 0.00000584, -0.00000707, 0.00000423, -0.00000165, 0.00000471, 0.00000349, 0.00000281, 0.00000687, -0.00000667, 0.00000372, -0.00001264, 0.00000526, -0.00001300, -0.00000260, -0.00000708, 0.00003038, -0.00002042, 0.00001877, -0.00002109, 0.00001338, -0.00001407, -0.00000389, -0.00001986, 0.00004883, -0.00003410, },
		{ -0.00002840, -0.00000135, -0.00000753, -0.00000303, -0.00002977, 0.00007231, -0.00006463, -0.00002604, -0.00007022, -0.00021163, -0.00013786, 0.00107385, -0.00018199, -0.00019285, -0.00005193, -0.00003634, -0.00005487, 0.00002524, 0.00002200, -0.00000303, -0.00003163, -0.00001398, 0.00000953, -0.00000439, -0.00000564, -0.00000476, 0.00002133, -0.00000389, -0.00001740, -0.00001489, 0.00001606, -0.00000585, 0.00000943, -0.00000905, 0.00001476, -0.00000260, -0.00000714, -0.00001507, 0.00000519, -0.00000673, 0.00000390, -0.00000142, 0.00000466, 0.00000349, 0.00000284, 0.00000672, -0.00000650, 0.00000357, -0.00001244, 0.00000509, -0.00001296, -0.00000260, -0.00000708, 0.00003050, -0.00002061, 0.00001901, -0.00002160, 0.00001388, -0.00001400, -0.00000389, -0.00002121, 0.00005091, -0.00003541, 0.00002047, },
		{ -0.00000058, -0.00000579, -0.00000303, -0.00003519, 0.00007837, -0.00006621, -0.00002387, -0.00007558, -0.00022020, -0.00009625, 0.00101220, -0.00013786, -0.00020249, -0.00005834, -0.00003358, -0.00005701, 0.00003399, 0.00001361, -0.00000303, -0.00002851, -0.00001249, 0.00000543, -0.00000191, -0.00000840, 0.00000007, 0.00001787, -0.00000389, -0.00001716, -0.00001294, 0.00001365, -0.00000435, 0.00000771, -0.00000691, 0.00001365, -0.00000260, -0.00000741, -0.00001420, 0.00000448, -0.00000633, 0.00000353, -0.00000115, 0.00000460, 0.00000349, 0.00000289, 0.00000652, -0.00000628, 0.00000338, -0.00001215, 0.00000481, -0.00001290, -0.00000260, -0.00000695, 0.00003038, -0.00002061, 0.00001909, -0.00002186, 0.00001418, -0.00001395, -0.00000389, -0.00002224, 0.00005256, -0.00003648, 0.00002154, -0.00003033, },
		{ -0.00000431, -0.00000303, -0.00003994, 0.00008375, -0.00006763, -0.00002190, -0.00008049, -0.00022809, -0.00005757, 0.00095447, -0.00009625, -0.00021163, -0.00006446, -0.00003092, -0.00005907, 0.00004249, 0.00000543, -0.00000303, -0.00002543, -0.00001102, 0.00000134, 0.00000058, -0.00001118, 0.00000494, 0.00001436, -0.00000389, -0.00001691, -0.00001093, 0.00001115, -0.00000278, 0.00000591, -0.00000465, 0.00001246, -0.00000260, -0.00000769, -0.00001326, 0.00000370, -0.00000590, 0.00000312, -0.00000084, 0.00000454, 0.00000349, 0.00000295, 0.00000629, -0.00000600, 0.00000313, -0.00001176, 0.00000441, -0.00001280, -0.00000260, -0.00000669, 0.00003001, -0.00002042, 0.00001901, -0.00002186, 0.00001428, -0.00001392, -0.00000389, -0.00002292, 0.00005375, -0.00003728, 0.00002238, -0.00003190, 0.00000005, },
		{ -0.00000303, -0.00004397, 0.00008839, -0.00006887, -0.00002015, -0.00008489, -0.00023525, -0.00002219, 0.00090121, -0.00005757, -0.00022020, -0.00007022, -0.00002840, -0.00006104, 0.00005064, -0.00000247, -0.00000303, -0.00002243, -0.00000957, -0.00000268, 0.00000304, -0.00001395, 0.00000982, 0.00001084, -0.00000389, -0.00001665, -0.00000888, 0.00000859, -0.00000117, 0.00000405, -0.00000230, 0.00001122, -0.00000260, -0.00000800, -0.00001225, 0.00000286, -0.00000543, 0.00000267, -0.00000050, 0.00000447, 0.00000349, 0.00000301, 0.00000602, -0.00000568, 0.00000283, -0.00001129, 0.00000391, -0.00001266, -0.00000260, -0.00000630, 0.00002941, -0.00002004, 0.00001877, -0.00002160, 0.00001418, -0.00001392, -0.00000389, -0.00002327, 0.00005446, -0.00003782, 0.00002298, -0.00003309, 0.00000056, -0.00000311, },
		{ -0.00004725, 0.00009224, -0.00006992, -0.00001864, -0.00008874, -0.00024160, 0.00000955, 0.00085294, -0.00002219, -0.00022809, -0.00007558, -0.00002604, -0.00006290, 0.00005838, -0.00001001, -0.00000303, -0.00001954, -0.00000817, -0.00000661, 0.00000546, -0.00001668, 0.00001465, 0.00000733, -0.00000389, -0.00001640, -0.00000681, 0.00000600, 0.00000048, 0.00000214, 0.00000012, 0.00000993, -0.00000260, -0.00000831, -0.00001119, 0.00000196, -0.00000493, 0.00000218, -0.00000014, 0.00000439, 0.00000349, 0.00000308, 0.00000571, -0.00000531, 0.00000248, -0.00001072, 0.00000331, -0.00001250, -0.00000260, -0.00000578, 0.00002856, -0.00001948, 0.00001836, -0.00002109, 0.00001388, -0.00001395, -0.00000389, -0.00002327, 0.00005470, -0.00003810, 0.00002335, -0.00003389, 0.00000092, -0.00000220, -0.00000303, },
		{ 0.00009528, -0.00007077, -0.00001738, -0.00009201, -0.00024708, 0.00003735, 0.00081012, 0.00000955, -0.00023525, -0.00008049, -0.00002387, -0.00006463, 0.00006563, -0.00001712, -0.00000303, -0.00001677, -0.00000682, -0.00001040, 0.00000780, -0.00001934, 0.00001939, 0.00000387, -0.00000389, -0.00001614, -0.00000474, 0.00000339, 0.00000214, 0.00000020, 0.00000258, 0.00000861, -0.00000260, -0.00000864, -0.00001009, 0.00000103, -0.00000440, 0.00000166, 0.00000025, 0.00000431, 0.00000349, 0.00000316, 0.00000538, -0.00000489, 0.00000209, -0.00001008, 0.00000261, -0.00001231, -0.00000260, -0.00000515, 0.00002750, -0.00001873, 0.00001780, -0.00002033, 0.00001338, -0.00001400, -0.00000389, -0.00002292, 0.00005446, -0.00003810, 0.00002347, -0.00003430, 0.00000114, -0.00000159, -0.00000303, -0.00004973, },
		{ -0.00007142, -0.00001639, -0.00009466, -0.00025164, 0.00006094, 0.00077317, 0.00003735, -0.00024160, -0.00008489, -0.00002190, -0.00006621, 0.00007231, -0.00002373, -0.00000303, -0.00001416, -0.00000554, -0.00001402, 0.00001006, -0.00002191, 0.00002399, 0.00000050, -0.00000389, -0.00001589, -0.00000270, 0.00000079, 0.00000380, -0.00000175, 0.00000508, 0.00000727, -0.00000260, -0.00000898, -0.00000895, 0.00000006, -0.00000385, 0.00000112, 0.00000066, 0.00000422, 0.00000349, 0.00000325, 0.00000501, -0.00000444, 0.00000165, -0.00000937, 0.00000182, -0.00001209, -0.00000260, -0.00000440, 0.00002622, -0.00001781, 0.00001709, -0.00001934, 0.00001268, -0.00001407, -0.00000389, -0.00002224, 0.00005375, -0.00003782, 0.00002335, -0.00003430, 0.00000121, -0.00000128, -0.00000303, -0.00005140, 0.00009748, },
		{ -0.00001568, -0.00009667, -0.00025523, 0.00008009, 0.00074244, 0.00006094, -0.00024708, -0.00008874, -0.00002015, -0.00006763, 0.00007837, -0.00002977, -0.00000303, -0.00001174, -0.00000434, -0.00001744, 0.00001219, -0.00002436, 0.00002840, -0.00000276, -0.00000389, -0.00001565, -0.00000069, -0.00000177, 0.00000545, -0.00000369, 0.00000757, 0.00000593, -0.00000260, -0.00000932, -0.00000779, -0.00000093, -0.00000328, 0.00000056, 0.00000109, 0.00000412, 0.00000349, 0.00000334, 0.00000462, -0.00000395, 0.00000118, -0.00000859, 0.00000095, -0.00001185, -0.00000260, -0.00000355, 0.00002473, -0.00001673, 0.00001624, -0.00001811, 0.00001179, -0.00001417, -0.00000389, -0.00002121, 0.00005256, -0.00003728, 0.00002298, -0.00003389, 0.00000114, -0.00000128, -0.00000303, -0.00005223, 0.00009880, -0.00007186, },
		{ -0.00009803, -0.00025782, 0.00009461, 0.00071823, 0.00008009, -0.00025164, -0.00009201, -0.00001864, -0.00006887, 0.00008375, -0.00003519, -0.00000303, -0.00000952, -0.00000324, -0.00002061, 0.00001419, -0.00002667, 0.00003259, -0.00000586, -0.00000389, -0.00001541, 0.00000126, -0.00000426, 0.00000706, -0.00000560, 0.00001004, 0.00000459, -0.00000260, -0.00000967, -0.00000662, -0.00000194, -0.00000271, -0.00000001, 0.00000153, 0.00000403, 0.00000349, 0.00000343, 0.00000421, -0.00000344, 0.00000068, -0.00000774, 0.00000001, -0.00001158, -0.00000260, -0.00000260, 0.00002305, -0.00001549, 0.00001525, -0.00001666, 0.00001073, -0.00001429, -0.00000389, -0.00001986, 0.00005091, -0.00003648, 0.00002238, -0.00003309, 0.00000092, -0.00000159, -0.00000303, -0.00005223, 0.00009924, -0.00007208, -0.00001525, },
		{ -0.00025939, 0.00010438, 0.00070076, 0.00009461, -0.00025523, -0.00009466, -0.00001738, -0.00006992, 0.00008839, -0.00003994, -0.00000303, -0.00000753, -0.00000224, -0.00002351, 0.00001604, -0.00002882, 0.00003651, -0.00000879, -0.00000389, -0.00001519, 0.00000313, -0.00000667, 0.00000863, -0.00000747, 0.00001246, 0.00000327, -0.00000260, -0.00001001, -0.00000545, -0.00000295, -0.00000212, -0.00000059, 0.00000198, 0.00000393, 0.00000349, 0.00000353, 0.00000378, -0.00000290, 0.00000016, -0.00000685, -0.00000099, -0.00001130, -0.00000260, -0.00000156, 0.00002120, -0.00001411, 0.00001413, -0.00001500, 0.00000949, -0.00001443, -0.00000389, -0.00001820, 0.00004883, -0.00003541, 0.00002154, -0.00003190, 0.00000056, -0.00000220, -0.00000303, -0.00005140, 0.00009880, -0.00007208, -0.00001510, -0.00009871, },
		{ 0.00010928, 0.00069021, 0.00010438, -0.00025782, -0.00009667, -0.00001639, -0.00007077, 0.00009224, -0.00004397, -0.00000303, -0.00000579, -0.00000135, -0.00002612, 0.00001771, -0.00003079, 0.00004013, -0.00001152, -0.00000389, -0.00001497, 0.00000491, -0.00000897, 0.00001013, -0.00000927, 0.00001481, 0.00000199, -0.00000260, -0.00001035, -0.00000429, -0.00000396, -0.00000154, -0.00000118, 0.00000244, 0.00000383, 0.00000349, 0.00000363, 0.00000334, -0.00000234, -0.00000039, -0.00000592, -0.00000205, -0.00001099, -0.00000260, -0.00000044, 0.00001920, -0.00001261, 0.00001290, -0.00001316, 0.00000810, -0.00001459, -0.00000389, -0.00001625, 0.00004632, -0.00003410, 0.00002047, -0.00003033, 0.00000005, -0.00000311, -0.00000303, -0.00004973, 0.00009748, -0.00007186, -0.00001525, -0.00009871, -0.00025991, },
		},
		{ { -0.00024969, 0.00001671, 0.00035962, 0.00004488, -0.00014314, -0.00000926, -0.00010314, 0.00000205, -0.00001040, -0.00002350, 0.00003300, -0.00001961, 0.00002946, -0.00000798, -0.00003198, 0.00001123, 0.00000476, -0.00000423, 0.00001871, -0.00000357, -0.00002071, 0.00000352, -0.00000955, 0.00000503, -0.00000134, -0.00000585, 0.00000457, -0.00000071, 0.00000811, 0.00000229, -0.00000291, -0.00000338, -0.00000516, -0.00000338, -0.00000291, 0.00000229, 0.00000811, -0.00000071, 0.00000457, -0.00000585, -0.00000134, 0.00000503, -0.00000955, 0.00000352, -0.00002071, -0.00000357, 0.00001871, -0.00000423, 0.00000476, 0.00001123, -0.00003198, -0.00000798, 0.00002946, -0.00001961, 0.00003300, -0.00002350, -0.00001040, 0.00000205, -0.00010314, -0.00000926, -0.00014314, 0.00004488, 0.00035962, 0.00001671, },
		{ 0.00000249, 0.00031910, 0.00007739, -0.00013508, -0.00001503, -0.00009750, -0.00000960, 0.00000036, -0.00002901, 0.00003241, -0.00001908, 0.00002681, -0.00000079, -0.00003683, 0.00000960, 0.00000786, -0.00000496, 0.00001702, -0.00000059, -0.00002091, 0.00000285, -0.00001007, 0.00000427, 0.00000004, -0.00000659, 0.00000455, -0.00000159, 0.00000867, 0.00000265, -0.00000219, -0.00000325, -0.00000531, -0.00000318, -0.00000378, 0.00000200, 0.00000752, -0.00000004, 0.00000477, -0.00000522, -0.00000257, 0.00000556, -0.00000899, 0.00000423, -0.00002022, -0.00000671, 0.00002013, -0.00000320, 0.00000196, 0.00001176, -0.00002637, -0.00001454, 0.00003076, -0.00001884, 0.00003181, -0.00001599, -0.00002204, 0.00001286, -0.00010667, -0.00000645, -0.00014453, 0.00000432, 0.00040034, 0.00003531, -0.00025118, },
		{ 0.00027973, 0.00010213, -0.00012189, -0.00002212, -0.00009118, -0.00002073, 0.00000915, -0.00003187, 0.00002980, -0.00001713, 0.00002282, 0.00000683, -0.00004081, 0.00000700, 0.00001097, -0.00000515, 0.00001495, 0.00000231, -0.00002094, 0.00000230, -0.00001056, 0.00000333, 0.00000152, -0.00000736, 0.00000464, -0.00000253, 0.00000904, 0.00000320, -0.00000173, -0.00000275, -0.00000575, -0.00000274, -0.00000467, 0.00000164, 0.00000704, 0.00000032, 0.00000522, -0.00000471, -0.00000361, 0.00000584, -0.00000834, 0.00000485, -0.00001930, -0.00001012, 0.00002140, -0.00000208, -0.00000026, 0.00001108, -0.00002015, -0.00002033, 0.00003073, -0.00001699, 0.00002927, -0.00000729, -0.00003336, 0.00002147, -0.00010677, -0.00000791, -0.00013829, -0.00004384, 0.00043969, 0.00005934, -0.00025577, -0.00000805, },
		{ 0.00011994, -0.00010542, -0.00002882, -0.00008562, -0.00003013, 0.00001509, -0.00003168, 0.00002507, -0.00001378, 0.00001762, 0.00001464, -0.00004383, 0.00000365, 0.00001377, -0.00000461, 0.00001239, 0.00000523, -0.00002093, 0.00000193, -0.00001104, 0.00000222, 0.00000303, -0.00000808, 0.00000473, -0.00000339, 0.00000908, 0.00000404, -0.00000159, -0.00000186, -0.00000642, -0.00000218, -0.00000545, 0.00000109, 0.00000676, 0.00000029, 0.00000593, -0.00000434, -0.00000446, 0.00000585, -0.00000756, 0.00000527, -0.00001781, -0.00001388, 0.00002263, -0.00000108, -0.00000173, 0.00000922, -0.00001355, -0.00002519, 0.00002952, -0.00001439, 0.00002592, 0.00000169, -0.00004315, 0.00002669, -0.00010240, -0.00001454, -0.00012410, -0.00009837, 0.00047554, 0.00009016, -0.00026375, -0.00001531, 0.00024194, },
		{ -0.00008768, -0.00003351, -0.00008200, -0.00003684, 0.00001762, -0.00002831, 0.00001838, -0.00000922, 0.00001146, 0.00002240, -0.00004585, -0.00000020, 0.00001595, -0.00000317, 0.00000931, 0.00000820, -0.00002094, 0.00000177, -0.00001150, 0.00000098, 0.00000449, -0.00000866, 0.00000470, -0.00000401, 0.00000865, 0.00000525, -0.00000180, -0.00000064, -0.00000721, -0.00000163, -0.00000598, 0.00000025, 0.00000678, -0.00000013, 0.00000686, -0.00000407, -0.00000512, 0.00000560, -0.00000663, 0.00000540, -0.00001566, -0.00001801, 0.00002387, -0.00000031, -0.00000234, 0.00000627, -0.00000682, -0.00002902, 0.00002729, -0.00001140, 0.00002232, 0.00001007, -0.00005037, 0.00002762, -0.00009298, -0.00002663, -0.00010244, -0.00015726, 0.00050526, 0.00012928, -0.00027556, -0.00001946, 0.00020574, 0.00013207, },
		{ -0.00003482, -0.00008120, -0.00004030, 0.00001656, -0.00002199, 0.00001009, -0.00000375, 0.00000462, 0.00002983, -0.00004684, -0.00000431, 0.00001728, -0.00000075, 0.00000573, 0.00001122, -0.00002098, 0.00000181, -0.00001191, -0.00000034, 0.00000584, -0.00000902, 0.00000443, -0.00000427, 0.00000769, 0.00000683, -0.00000234, 0.00000084, -0.00000799, -0.00000123, -0.00000615, -0.00000094, 0.00000711, -0.00000091, 0.00000794, -0.00000384, -0.00000562, 0.00000511, -0.00000553, 0.00000515, -0.00001281, -0.00002248, 0.00002510, 0.00000018, -0.00000210, 0.00000243, -0.00000024, -0.00003172, 0.00002425, -0.00000840, 0.00001902, 0.00001704, -0.00005421, 0.00002374, -0.00007841, -0.00004383, -0.00007451, -0.00021780, 0.00052587, 0.00017827, -0.00029161, -0.00002051, 0.00017085, 0.00013997, -0.00007060, },
		{ -0.00008364, -0.00004035, 0.00001211, -0.00001322, 0.00000076, 0.00000220, -0.00000250, 0.00003664, -0.00004678, -0.00000847, 0.00001757, 0.00000264, 0.00000175, 0.00001420, -0.00002102, 0.00000197, -0.00001222, -0.00000173, 0.00000699, -0.00000905, 0.00000383, -0.00000407, 0.00000616, 0.00000877, -0.00000314, 0.00000244, -0.00000862, -0.00000109, -0.00000589, -0.00000247, 0.00000772, -0.00000197, 0.00000905, -0.00000358, -0.00000602, 0.00000440, -0.00000425, 0.00000447, -0.00000927, -0.00002719, 0.00002624, 0.00000041, -0.00000112, -0.00000200, 0.00000585, -0.00003322, 0.00002062, -0.00000574, 0.00001647, 0.00002201, -0.00005419, 0.00001500, -0.00005918, -0.00006518, -0.00004225, -0.00027675, 0.00053424, 0.00023852, -0.00031213, -0.00001851, 0.00013691, 0.00014509, -0.00005582, -0.00003187, },
		{ -0.00003731, 0.00000487, -0.00000281, -0.00000887, 0.00000809, -0.00000948, 0.00004252, -0.00004567, -0.00001248, 0.00001674, 0.00000687, -0.00000246, 0.00001702, -0.00002096, 0.00000217, -0.00001238, -0.00000316, 0.00000789, -0.00000871, 0.00000285, -0.00000338, 0.00000408, 0.00001096, -0.00000408, 0.00000403, -0.00000897, -0.00000132, -0.00000518, -0.00000432, 0.00000852, -0.00000318, 0.00001005, -0.00000319, -0.00000635, 0.00000352, -0.00000281, 0.00000339, -0.00000513, -0.00003196, 0.00002716, 0.00000049, 0.00000040, -0.00000667, 0.00001114, -0.00003346, 0.00001661, -0.00000369, 0.00001500, 0.00002463, -0.00005022, 0.00000188, -0.00003632, -0.00008907, -0.00000816, -0.00033053, 0.00052739, 0.00031101, -0.00033702, -0.00001368, 0.00010362, 0.00014864, -0.00004453, -0.00002429, -0.00008922, },
		{ -0.00000421, 0.00000825, -0.00001800, 0.00001337, -0.00001586, 0.00004719, -0.00004350, -0.00001621, 0.00001483, 0.00001171, -0.00000663, 0.00001949, -0.00002068, 0.00000227, -0.00001231, -0.00000461, 0.00000850, -0.00000796, 0.00000145, -0.00000220, 0.00000153, 0.00001328, -0.00000501, 0.00000545, -0.00000891, -0.00000197, -0.00000403, -0.00000639, 0.00000939, -0.00000437, 0.00001079, -0.00000258, -0.00000670, 0.00000253, -0.00000125, 0.00000194, -0.00000056, -0.00003655, 0.00002764, 0.00000060, 0.00000219, -0.00001118, 0.00001528, -0.00003241, 0.00001239, -0.00000242, 0.00001476, 0.00002482, -0.00004264, -0.00001469, -0.00001136, -0.00011348, 0.00002487, -0.00037546, 0.00050272, 0.00039614, -0.00036568, -0.00000652, 0.00007092, 0.00015147, -0.00003734, -0.00001233, -0.00009732, -0.00003193, },
		{ 0.00001881, -0.00002576, 0.00001745, -0.00002120, 0.00005038, -0.00004027, -0.00001959, 0.00001198, 0.00001686, -0.00001044, 0.00002138, -0.00002002, 0.00000214, -0.00001194, -0.00000609, 0.00000879, -0.00000680, -0.00000034, -0.00000059, -0.00000134, 0.00001555, -0.00000577, 0.00000654, -0.00000835, -0.00000305, -0.00000252, -0.00000854, 0.00001017, -0.00000536, 0.00001113, -0.00000165, -0.00000713, 0.00000150, 0.00000034, 0.00000023, 0.00000423, -0.00004065, 0.00002744, 0.00000097, 0.00000393, -0.00001510, 0.00001800, -0.00003007, 0.00000809, -0.00000199, 0.00001571, 0.00002280, -0.00003216, -0.00003334, 0.00001379, -0.00013603, 0.00005380, -0.00040813, 0.00045828, 0.00049348, -0.00039690, 0.00000202, 0.00003911, 0.00015394, -0.00003421, 0.00000316, -0.00010688, -0.00002532, -0.00001394, },
		{ -0.00003133, 0.00001983, -0.00002511, 0.00005184, -0.00003596, -0.00002264, 0.00000840, 0.00002192, -0.00001356, 0.00002245, -0.00001879, 0.00000164, -0.00001120, -0.00000759, 0.00000878, -0.00000527, -0.00000243, 0.00000131, -0.00000435, 0.00001757, -0.00000620, 0.00000718, -0.00000723, -0.00000452, -0.00000077, -0.00001060, 0.00001067, -0.00000598, 0.00001092, -0.00000033, -0.00000767, 0.00000049, 0.00000188, -0.00000160, 0.00000895, -0.00004394, 0.00002629, 0.00000185, 0.00000527, -0.00001803, 0.00001904, -0.00002645, 0.00000380, -0.00000232, 0.00001760, 0.00001907, -0.00001984, -0.00005239, 0.00003701, -0.00015428, 0.00007569, -0.00042559, 0.00039308, 0.00060164, -0.00042876, 0.00001056, 0.00000889, 0.00015588, -0.00003447, 0.00002077, -0.00011642, -0.00001884, -0.00002299, 0.00002772, },
		{ 0.00002009, -0.00002725, 0.00005141, -0.00003059, -0.00002540, 0.00000443, 0.00002644, -0.00001561, 0.00002248, -0.00001685, 0.00000064, -0.00001005, -0.00000910, 0.00000849, -0.00000345, -0.00000472, 0.00000335, -0.00000727, 0.00001912, -0.00000613, 0.00000726, -0.00000555, -0.00000631, 0.00000106, -0.00001239, 0.00001073, -0.00000607, 0.00001008, 0.00000141, -0.00000838, -0.00000041, 0.00000327, -0.00000336, 0.00001330, -0.00004605, 0.00002394, 0.00000346, 0.00000588, -0.00001963, 0.00001824, -0.00002161, -0.00000044, -0.00000324, 0.00002002, 0.00001437, -0.00000698, -0.00007000, 0.00005613, -0.00016590, 0.00008797, -0.00042570, 0.00030720, 0.00071819, -0.00045861, 0.00001723, -0.00001863, 0.00015659, -0.00003691, 0.00003870, -0.00012425, -0.00001396, -0.00003001, 0.00003389, -0.00003404, },
		{ -0.00002741, 0.00004898, -0.00002416, -0.00002797, 0.00000043, 0.00002997, -0.00001628, 0.00002127, -0.00001407, -0.00000094, -0.00000845, -0.00001063, 0.00000797, -0.00000145, -0.00000704, 0.00000531, -0.00000987, 0.00002002, -0.00000544, 0.00000672, -0.00000336, -0.00000827, 0.00000279, -0.00001372, 0.00001020, -0.00000551, 0.00000853, 0.00000358, -0.00000927, -0.00000116, 0.00000440, -0.00000486, 0.00001695, -0.00004669, 0.00002017, 0.00000601, 0.00000549, -0.00001961, 0.00001551, -0.00001564, -0.00000463, -0.00000445, 0.00002245, 0.00000959, 0.00000497, -0.00008430, 0.00006915, -0.00016891, 0.00008868, -0.00040726, 0.00020201, 0.00083964, -0.00048319, 0.00001979, -0.00004195, 0.00015495, -0.00003985, 0.00005488, -0.00012862, -0.00001216, -0.00003375, 0.00003640, -0.00003335, 0.00001794, },
		{ 0.00004452, -0.00001674, -0.00003049, -0.00000321, 0.00003210, -0.00001528, 0.00001868, -0.00001038, -0.00000315, -0.00000641, -0.00001219, 0.00000729, 0.00000059, -0.00000921, 0.00000700, -0.00001193, 0.00002009, -0.00000406, 0.00000558, -0.00000074, -0.00001024, 0.00000424, -0.00001442, 0.00000895, -0.00000423, 0.00000629, 0.00000614, -0.00001034, -0.00000168, 0.00000514, -0.00000591, 0.00001961, -0.00004558, 0.00001484, 0.00000963, 0.00000390, -0.00001781, 0.00001089, -0.00000868, -0.00000881, -0.00000560, 0.00002425, 0.00000569, 0.00001458, -0.00009359, 0.00007441, -0.00016185, 0.00007665, -0.00037027, 0.00008012, 0.00096152, -0.00049870, 0.00001572, -0.00005940, 0.00014954, -0.00004139, 0.00006721, -0.00012786, -0.00001468, -0.00003322, 0.00003463, -0.00002901, 0.00001328, -0.00002548, },
		{ -0.00000839, -0.00003308, -0.00000612, 0.00003245, -0.00001245, 0.00001467, -0.00000577, -0.00000597, -0.00000396, -0.00001376, 0.00000654, 0.00000252, -0.00001104, 0.00000820, -0.00001324, 0.00001921, -0.00000196, 0.00000387, 0.00000216, -0.00001203, 0.00000522, -0.00001436, 0.00000692, -0.00000223, 0.00000341, 0.00000901, -0.00001155, -0.00000195, 0.00000541, -0.00000633, 0.00002103, -0.00004254, 0.00000790, 0.00001435, 0.00000101, -0.00001420, 0.00000450, -0.00000091, -0.00001303, -0.00000631, 0.00002480, 0.00000362, 0.00002054, -0.00009648, 0.00007077, -0.00014401, 0.00005166, -0.00031591, -0.00005459, 0.00107856, -0.00050106, 0.00000244, -0.00006919, 0.00013882, -0.00003959, 0.00007372, -0.00012062, -0.00002247, -0.00002777, 0.00002826, -0.00002102, 0.00000621, -0.00002148, 0.00003812, },
		{ -0.00003588, -0.00000795, 0.00003078, -0.00000773, 0.00000929, -0.00000032, -0.00000935, -0.00000118, -0.00001532, 0.00000581, 0.00000419, -0.00001236, 0.00000871, -0.00001365, 0.00001733, 0.00000081, 0.00000171, 0.00000516, -0.00001347, 0.00000558, -0.00001345, 0.00000411, 0.00000043, 0.00000002, 0.00001207, -0.00001285, -0.00000194, 0.00000513, -0.00000597, 0.00002102, -0.00003750, -0.00000056, 0.00002010, -0.00000316, -0.00000885, -0.00000339, 0.00000742, -0.00001734, -0.00000620, 0.00002352, 0.00000419, 0.00002180, -0.00009203, 0.00005774, -0.00011547, 0.00001451, -0.00024657, -0.00019717, 0.00118493, -0.00048613, -0.00002245, -0.00006971, 0.00012136, -0.00003270, 0.00007284, -0.00010598, -0.00003601, -0.00001716, 0.00001738, -0.00000969, -0.00000301, -0.00001559, 0.00002993, 0.00000075, },
		{ -0.00000847, 0.00002695, -0.00000119, 0.00000271, 0.00000580, -0.00001314, 0.00000182, -0.00001686, 0.00000517, 0.00000544, -0.00001299, 0.00000841, -0.00001307, 0.00001446, 0.00000416, -0.00000075, 0.00000807, -0.00001436, 0.00000522, -0.00001168, 0.00000060, 0.00000363, -0.00000373, 0.00001516, -0.00001416, -0.00000166, 0.00000426, -0.00000475, 0.00001947, -0.00003050, -0.00001034, 0.00002669, -0.00000847, -0.00000201, -0.00001244, 0.00001605, -0.00002176, -0.00000497, 0.00001996, 0.00000801, 0.00001768, -0.00007989, 0.00003556, -0.00007716, -0.00003302, -0.00016572, -0.00034184, 0.00127458, -0.00045009, -0.00006095, -0.00005970, 0.00009603, -0.00001935, 0.00006351, -0.00008361, -0.00005526, -0.00000164, 0.00000247, 0.00000440, -0.00001390, -0.00000810, 0.00002022, 0.00001050, -0.00003894, },
		{ 0.00002100, 0.00000696, -0.00000480, 0.00001237, -0.00001715, 0.00000491, -0.00001833, 0.00000469, 0.00000618, -0.00001283, 0.00000720, -0.00001146, 0.00001070, 0.00000791, -0.00000331, 0.00001068, -0.00001457, 0.00000406, -0.00000910, -0.00000348, 0.00000718, -0.00000760, 0.00001812, -0.00001540, -0.00000112, 0.00000278, -0.00000262, 0.00001640, -0.00002175, -0.00002113, 0.00003382, -0.00001464, 0.00000598, -0.00002219, 0.00002464, -0.00002630, -0.00000237, 0.00001383, 0.00001540, 0.00000796, -0.00006031, 0.00000521, -0.00003087, -0.00008823, -0.00007771, -0.00048229, 0.00134161, -0.00038972, -0.00011438, -0.00003848, 0.00006222, 0.00000124, 0.00004537, -0.00005384, -0.00007964, 0.00001807, -0.00001563, 0.00002042, -0.00002585, 0.00000059, 0.00000935, 0.00002063, -0.00004231, -0.00000752, },
		{ 0.00001634, -0.00001287, 0.00001907, -0.00002116, 0.00000793, -0.00001970, 0.00000445, 0.00000630, -0.00001181, 0.00000506, -0.00000890, 0.00000619, 0.00001186, -0.00000577, 0.00001282, -0.00001399, 0.00000213, -0.00000583, -0.00000793, 0.00001085, -0.00001137, 0.00002075, -0.00001648, -0.00000038, 0.00000074, 0.00000038, 0.00001192, -0.00001154, -0.00003246, 0.00004109, -0.00002133, 0.00001468, -0.00003214, 0.00003285, -0.00003088, 0.00000167, 0.00000507, 0.00002635, -0.00000713, -0.00003418, -0.00003160, 0.00002092, -0.00014772, 0.00001249, -0.00061205, 0.00138065, -0.00030272, -0.00018314, -0.00000610, 0.00001999, 0.00002913, 0.00001882, -0.00001768, -0.00010798, 0.00004083, -0.00003576, 0.00003733, -0.00003814, 0.00000997, -0.00000226, 0.00003084, -0.00004593, -0.00000510, 0.00001310, },
		{ -0.00002105, 0.00002556, -0.00002490, 0.00001070, -0.00002092, 0.00000446, 0.00000577, -0.00000993, 0.00000205, -0.00000551, 0.00000118, 0.00001576, -0.00000791, 0.00001430, -0.00001257, -0.00000050, -0.00000205, -0.00001250, 0.00001436, -0.00001478, 0.00002289, -0.00001728, 0.00000051, -0.00000179, 0.00000412, 0.00000626, -0.00000032, -0.00004381, 0.00004803, -0.00002810, 0.00002357, -0.00004174, 0.00004032, -0.00003542, 0.00000713, -0.00000611, 0.00004051, -0.00002682, -0.00000300, -0.00007257, 0.00007511, -0.00020757, 0.00009962, -0.00072489, 0.00138724, -0.00018804, -0.00026655, 0.00003650, -0.00002985, 0.00006363, -0.00001498, 0.00002322, -0.00013868, 0.00006513, -0.00005650, 0.00005398, -0.00004995, 0.00001949, -0.00001412, 0.00004079, -0.00004968, -0.00000130, 0.00000358, 0.00002647, },
		{ 0.00003144, -0.00002810, 0.00001307, -0.00002192, 0.00000475, 0.00000460, -0.00000726, -0.00000169, -0.00000151, -0.00000407, 0.00001934, -0.00000953, 0.00001502, -0.00001032, -0.00000368, 0.00000200, -0.00001690, 0.00001744, -0.00001758, 0.00002435, -0.00001773, 0.00000145, -0.00000469, 0.00000843, -0.00000028, 0.00001138, -0.00005456, 0.00005413, -0.00003447, 0.00003208, -0.00005042, 0.00004668, -0.00003974, 0.00001379, -0.00001931, 0.00005715, -0.00004993, 0.00003125, -0.00011494, 0.00012831, -0.00026365, 0.00017848, -0.00081523, 0.00135822, -0.00004605, -0.00036274, 0.00008744, -0.00008569, 0.00010323, -0.00005411, 0.00006665, -0.00016973, 0.00008921, -0.00007633, 0.00006917, -0.00006051, 0.00002857, -0.00002570, 0.00005010, -0.00005338, 0.00000363, -0.00000707, 0.00003678, -0.00002883, },
		{ -0.00003050, 0.00001486, -0.00002266, 0.00000527, 0.00000285, -0.00000392, -0.00000597, 0.00000284, -0.00000925, 0.00002234, -0.00001045, 0.00001491, -0.00000732, -0.00000723, 0.00000605, -0.00002086, 0.00001984, -0.00001956, 0.00002502, -0.00001775, 0.00000237, -0.00000781, 0.00001307, -0.00000729, 0.00002297, -0.00006409, 0.00005889, -0.00003995, 0.00003966, -0.00005765, 0.00005158, -0.00004363, 0.00002132, -0.00003383, 0.00007529, -0.00007490, 0.00006623, -0.00015568, 0.00017697, -0.00031192, 0.00024430, -0.00087851, 0.00129196, 0.00012129, -0.00046859, 0.00014396, -0.00014513, 0.00014573, -0.00009607, 0.00011002, -0.00019890, 0.00011123, -0.00009368, 0.00008176, -0.00006905, 0.00003666, -0.00003649, 0.00005838, -0.00005677, 0.00000934, -0.00001828, 0.00004661, -0.00003568, 0.00003634, },
		{ 0.00001593, -0.00002307, 0.00000600, 0.00000064, -0.00000009, -0.00001054, 0.00000724, -0.00001403, 0.00002452, -0.00001055, 0.00001395, -0.00000371, -0.00001089, 0.00000981, -0.00002408, 0.00002132, -0.00002056, 0.00002482, -0.00001729, 0.00000317, -0.00001096, 0.00001775, -0.00001432, 0.00003382, -0.00007179, 0.00006182, -0.00004411, 0.00004579, -0.00006294, 0.00005472, -0.00004685, 0.00002922, -0.00004884, 0.00009366, -0.00009991, 0.00009938, -0.00019174, 0.00021775, -0.00034874, 0.00029307, -0.00091148, 0.00118860, 0.00031039, -0.00057984, 0.00020247, -0.00020518, 0.00018836, -0.00013793, 0.00015050, -0.00022391, 0.00012934, -0.00010712, 0.00009073, -0.00007495, 0.00004325, -0.00004598, 0.00006523, -0.00005955, 0.00001541, -0.00002939, 0.00005532, -0.00004111, 0.00003991, -0.00003187, },
		{ -0.00002311, 0.00000685, -0.00000188, 0.00000397, -0.00001510, 0.00001136, -0.00001813, 0.00002569, -0.00000977, 0.00001220, 0.00000032, -0.00001441, 0.00001300, -0.00002632, 0.00002169, -0.00002047, 0.00002370, -0.00001633, 0.00000376, -0.00001396, 0.00002218, -0.00002091, 0.00004332, -0.00007711, 0.00006253, -0.00004656, 0.00005002, -0.00006591, 0.00005587, -0.00004914, 0.00003694, -0.00006335, 0.00011090, -0.00012303, 0.00012818, -0.00022028, 0.00024769, -0.00037113, 0.00032182, -0.00091248, 0.00105010, 0.00051612, -0.00069122, 0.00025877, -0.00026234, 0.00022800, -0.00017654, 0.00018525, -0.00024263, 0.00014190, -0.00011541, 0.00009530, -0.00007773, 0.00004791, -0.00005371, 0.00007027, -0.00006141, 0.00002135, -0.00003971, 0.00006224, -0.00004467, 0.00004186, -0.00003202, 0.00001620, },
		{ 0.00000773, -0.00000452, 0.00000802, -0.00001934, 0.00001490, -0.00002127, 0.00002571, -0.00000811, 0.00000979, 0.00000455, -0.00001751, 0.00001536, -0.00002739, 0.00002086, -0.00001925, 0.00002168, -0.00001488, 0.00000406, -0.00001662, 0.00002603, -0.00002660, 0.00005092, -0.00007961, 0.00006076, -0.00004703, 0.00005202, -0.00006628, 0.00005484, -0.00005022, 0.00004384, -0.00007632, 0.00012560, -0.00014236, 0.00015029, -0.00023887, 0.00026452, -0.00037701, 0.00032881, -0.00088149, 0.00088028, 0.00073198, -0.00079674, 0.00030830, -0.00031294, 0.00026138, -0.00020879, 0.00021168, -0.00025321, 0.00014761, -0.00011765, 0.00009495, -0.00007710, 0.00005035, -0.00005929, 0.00007319, -0.00006204, 0.00002665, -0.00004857, 0.00006683, -0.00004600, 0.00004200, -0.00003086, 0.00001560, -0.00002275, },
		{ -0.00000708, 0.00001177, -0.00002296, 0.00001756, -0.00002324, 0.00002452, -0.00000562, 0.00000687, 0.00000872, -0.00001993, 0.00001670, -0.00002718, 0.00001880, -0.00001693, 0.00001886, -0.00001297, 0.00000400, -0.00001874, 0.00002901, -0.00003097, 0.00005615, -0.00007899, 0.00005638, -0.00004537, 0.00005158, -0.00006391, 0.00005159, -0.00004986, 0.00004931, -0.00008673, 0.00013644, -0.00015621, 0.00016374, -0.00024570, 0.00026680, -0.00036536, 0.00031371, -0.00082022, 0.00068457, 0.00095039, -0.00088999, 0.00034642, -0.00035336, 0.00028540, -0.00023184, 0.00022760, -0.00025429, 0.00014559, -0.00011333, 0.00008950, -0.00007301, 0.00005039, -0.00006245, 0.00007373, -0.00006116, 0.00003083, -0.00005538, 0.00006863, -0.00004489, 0.00004022, -0.00002833, 0.00001413, -0.00002196, 0.00000854, },
		{ 0.00001497, -0.00002569, 0.00001913, -0.00002392, 0.00002214, -0.00000245, 0.00000366, 0.00001256, -0.00002145, 0.00001686, -0.00002565, 0.00001555, -0.00001365, 0.00001537, -0.00001067, 0.00000354, -0.00002017, 0.00003087, -0.00003368, 0.00005870, -0.00007512, 0.00004942, -0.00004160, 0.00004865, -0.00005879, 0.00004612, -0.00004788, 0.00005276, -0.00009369, 0.00014229, -0.00016320, 0.00016711, -0.00023974, 0.00025402, -0.00033633, 0.00027756, -0.00073201, 0.00046986, 0.00116308, -0.00096452, 0.00036880, -0.00038032, 0.00029736, -0.00024340, 0.00023146, -0.00024509, 0.00013549, -0.00010238, 0.00007910, -0.00006562, 0.00004800, -0.00006302, 0.00007174, -0.00005855, 0.00003348, -0.00005963, 0.00006734, -0.00004123, 0.00003655, -0.00002449, 0.00001183, -0.00002076, 0.00000918, -0.00000934, },
		{ -0.00002731, 0.00001944, -0.00002325, 0.00001866, 0.00000122, 0.00000039, 0.00001585, -0.00002190, 0.00001580, -0.00002284, 0.00001127, -0.00000958, 0.00001138, -0.00000808, 0.00000265, -0.00002080, 0.00003143, -0.00003449, 0.00005837, -0.00006805, 0.00004008, -0.00003585, 0.00004330, -0.00005109, 0.00003860, -0.00004418, 0.00005372, -0.00009647, 0.00014231, -0.00016240, 0.00015965, -0.00022083, 0.00022674, -0.00029120, 0.00022273, -0.00062157, 0.00024405, 0.00136152, -0.00101432, 0.00037171, -0.00039122, 0.00029523, -0.00024190, 0.00022247, -0.00022554, 0.00011752, -0.00008520, 0.00006425, -0.00005531, 0.00004330, -0.00006094, 0.00006717, -0.00005410, 0.00003428, -0.00006099, 0.00006285, -0.00003512, 0.00003111, -0.00001948, 0.00000881, -0.00001915, 0.00000954, -0.00001111, 0.00001736, },
		{ 0.00001842, -0.00002128, 0.00001426, 0.00000516, -0.00000271, 0.00001836, -0.00002117, 0.00001354, -0.00001892, 0.00000617, -0.00000496, 0.00000711, -0.00000532, 0.00000133, -0.00002054, 0.00003057, -0.00003329, 0.00005516, -0.00005802, 0.00002874, -0.00002842, 0.00003578, -0.00004107, 0.00002927, -0.00003873, 0.00005186, -0.00009459, 0.00013606, -0.00015341, 0.00014133, -0.00018968, 0.00018646, -0.00023233, 0.00015276, -0.00049474, 0.00001570, 0.00153740, -0.00103415, 0.00035237, -0.00038430, 0.00027783, -0.00022662, 0.00020066, -0.00019625, 0.00009246, -0.00006261, 0.00004573, -0.00004264, 0.00003652, -0.00005631, 0.00006009, -0.00004779, 0.00003303, -0.00005929, 0.00005521, -0.00002679, 0.00002417, -0.00001351, 0.00000520, -0.00001719, 0.00000953, -0.00001221, 0.00001877, -0.00002767, },
		{ -0.00001813, 0.00000916, 0.00000911, -0.00000539, 0.00001994, -0.00001925, 0.00001019, -0.00001409, 0.00000053, -0.00000007, 0.00000280, -0.00000251, -0.00000038, -0.00001938, 0.00002826, -0.00003008, 0.00004925, -0.00004546, 0.00001591, -0.00001971, 0.00002643, -0.00002915, 0.00001849, -0.00003163, 0.00004702, -0.00008790, 0.00012350, -0.00013640, 0.00011286, -0.00014789, 0.00013559, -0.00016296, 0.00007206, -0.00035811, -0.00020651, 0.00168309, -0.00101998, 0.00030919, -0.00035892, 0.00024496, -0.00019785, 0.00016696, -0.00015855, 0.00006160, -0.00003581, 0.00002460, -0.00002834, 0.00002804, -0.00004932, 0.00005072, -0.00003971, 0.00002969, -0.00005454, 0.00004470, -0.00001663, 0.00001609, -0.00000687, 0.00000119, -0.00001494, 0.00000910, -0.00001251, 0.00001907, -0.00002669, 0.00001609, },
		{ 0.00000366, 0.00001280, -0.00000747, 0.00002048, -0.00001620, 0.00000596, -0.00000864, -0.00000534, 0.00000477, -0.00000132, 0.00000023, -0.00000244, -0.00001736, 0.00002457, -0.00002500, 0.00004095, -0.00003095, 0.00000222, -0.00001023, 0.00001572, -0.00001584, 0.00000668, -0.00002308, 0.00003926, -0.00007653, 0.00010502, -0.00011212, 0.00007565, -0.00009778, 0.00007726, -0.00008701, -0.00001433, -0.00021860, -0.00041423, 0.00179214, -0.00096929, 0.00024200, -0.00031560, 0.00019750, -0.00015679, 0.00012311, -0.00011436, 0.00002666, -0.00000631, 0.00000207, -0.00001320, 0.00001829, -0.00004031, 0.00003938, -0.00003009, 0.00002436, -0.00004695, 0.00003175, -0.00000515, 0.00000730, 0.00000013, -0.00000302, -0.00001247, 0.00000819, -0.00001194, 0.00001820, -0.00002440, 0.00001256, -0.00001403, },
		{ 0.00001597, -0.00000878, 0.00001996, -0.00001215, 0.00000109, -0.00000290, -0.00001108, 0.00000928, -0.00000503, 0.00000276, -0.00000477, -0.00001455, 0.00001965, -0.00001832, 0.00003075, -0.00001518, -0.00001163, -0.00000052, 0.00000416, -0.00000172, -0.00000565, -0.00001339, 0.00002883, -0.00006096, 0.00008147, -0.00008183, 0.00003173, -0.00004228, 0.00001506, -0.00000876, -0.00010117, -0.00008300, -0.00059990, 0.00185962, -0.00088133, 0.00015211, -0.00025609, 0.00013736, -0.00010559, 0.00007156, -0.00006608, -0.00001035, 0.00002422, -0.00002054, 0.00000191, 0.00000779, -0.00002969, 0.00002650, -0.00001927, 0.00001728, -0.00003689, 0.00001696, 0.00000702, -0.00000170, 0.00000711, -0.00000719, -0.00000989, 0.00000681, -0.00001046, 0.00001620, -0.00002089, 0.00000804, -0.00000924, -0.00000193, },
		{ -0.00000923, 0.00001842, -0.00000730, -0.00000408, 0.00000278, -0.00001637, 0.00001317, -0.00000813, 0.00000498, -0.00000728, -0.00001110, 0.00001374, -0.00001040, 0.00001921, 0.00000106, -0.00002491, 0.00000883, -0.00000768, 0.00001260, -0.00001796, -0.00000294, 0.00001617, -0.00004196, 0.00005404, -0.00004724, -0.00001644, 0.00001532, -0.00004715, 0.00006741, -0.00018331, 0.00004237, -0.00075719, 0.00188247, -0.00075719, 0.00004237, -0.00018331, 0.00006741, -0.00004715, 0.00001532, -0.00001644, -0.00004724, 0.00005404, -0.00004196, 0.00001617, -0.00000294, -0.00001796, 0.00001260, -0.00000768, 0.00000883, -0.00002491, 0.00000106, 0.00001921, -0.00001040, 0.00001374, -0.00001110, -0.00000728, 0.00000498, -0.00000813, 0.00001317, -0.00001637, 0.00000278, -0.00000408, -0.00000730, 0.00001842, },
		{ 0.00001597, -0.00000193, -0.00000924, 0.00000804, -0.00002089, 0.00001620, -0.00001046, 0.00000681, -0.00000989, -0.00000719, 0.00000711, -0.00000170, 0.00000702, 0.00001696, -0.00003689, 0.00001728, -0.00001927, 0.00002650, -0.00002969, 0.00000779, 0.00000191, -0.00002054, 0.00002422, -0.00001035, -0.00006608, 0.00007156, -0.00010559, 0.00013736, -0.00025609, 0.00015211, -0.00088133, 0.00185962, -0.00059990, -0.00008300, -0.00010117, -0.00000876, 0.00001506, -0.00004228, 0.00003173, -0.00008183, 0.00008147, -0.00006096, 0.00002883, -0.00001339, -0.00000565, -0.00000172, 0.00000416, -0.00000052, -0.00001163, -0.00001518, 0.00003075, -0.00001832, 0.00001965, -0.00001455, -0.00000477, 0.00000276, -0.00000503, 0.00000928, -0.00001108, -0.00000290, 0.00000109, -0.00001215, 0.00001996, -0.00000878, },
		{ 0.00000366, -0.00001403, 0.00001256, -0.00002440, 0.00001820, -0.00001194, 0.00000819, -0.00001247, -0.00000302, 0.00000013, 0.00000730, -0.00000515, 0.00003175, -0.00004695, 0.00002436, -0.00003009, 0.00003938, -0.00004031, 0.00001829, -0.00001320, 0.00000207, -0.00000631, 0.00002666, -0.00011436, 0.00012311, -0.00015679, 0.00019750, -0.00031560, 0.00024200, -0.00096929, 0.00179214, -0.00041423, -0.00021860, -0.00001433, -0.00008701, 0.00007726, -0.00009778, 0.00007565, -0.00011212, 0.00010502, -0.00007653, 0.00003926, -0.00002308, 0.00000668, -0.00001584, 0.00001572, -0.00001023, 0.00000222, -0.00003095, 0.00004095, -0.00002500, 0.00002457, -0.00001736, -0.00000244, 0.00000023, -0.00000132, 0.00000477, -0.00000534, -0.00000864, 0.00000596, -0.00001620, 0.00002048, -0.00000747, 0.00001280, },
		{ -0.00001813, 0.00001609, -0.00002669, 0.00001907, -0.00001251, 0.00000910, -0.00001494, 0.00000119, -0.00000687, 0.00001609, -0.00001663, 0.00004470, -0.00005454, 0.00002969, -0.00003971, 0.00005072, -0.00004932, 0.00002804, -0.00002834, 0.00002460, -0.00003581, 0.00006160, -0.00015855, 0.00016696, -0.00019785, 0.00024496, -0.00035892, 0.00030919, -0.00101998, 0.00168309, -0.00020651, -0.00035811, 0.00007206, -0.00016296, 0.00013559, -0.00014789, 0.00011286, -0.00013640, 0.00012350, -0.00008790, 0.00004702, -0.00003163, 0.00001849, -0.00002915, 0.00002643, -0.00001971, 0.00001591, -0.00004546, 0.00004925, -0.00003008, 0.00002826, -0.00001938, -0.00000038, -0.00000251, 0.00000280, -0.00000007, 0.00000053, -0.00001409, 0.00001019, -0.00001925, 0.00001994, -0.00000539, 0.00000911, 0.00000916, },
		{ 0.00001842, -0.00002767, 0.00001877, -0.00001221, 0.00000953, -0.00001719, 0.00000520, -0.00001351, 0.00002417, -0.00002679, 0.00005521, -0.00005929, 0.00003303, -0.00004779, 0.00006009, -0.00005631, 0.00003652, -0.00004264, 0.00004573, -0.00006261, 0.00009246, -0.00019625, 0.00020066, -0.00022662, 0.00027783, -0.00038430, 0.00035237, -0.00103415, 0.00153740, 0.00001570, -0.00049474, 0.00015276, -0.00023233, 0.00018646, -0.00018968, 0.00014133, -0.00015341, 0.00013606, -0.00009459, 0.00005186, -0.00003873, 0.00002927, -0.00004107, 0.00003578, -0.00002842, 0.00002874, -0.00005802, 0.00005516, -0.00003329, 0.00003057, -0.00002054, 0.00000133, -0.00000532, 0.00000711, -0.00000496, 0.00000617, -0.00001892, 0.00001354, -0.00002117, 0.00001836, -0.00000271, 0.00000516, 0.00001426, -0.00002128, },
		{ -0.00002731, 0.00001736, -0.00001111, 0.00000954, -0.00001915, 0.00000881, -0.00001948, 0.00003111, -0.00003512, 0.00006285, -0.00006099, 0.00003428, -0.00005410, 0.00006717, -0.00006094, 0.00004330, -0.00005531, 0.00006425, -0.00008520, 0.00011752, -0.00022554, 0.00022247, -0.00024190, 0.00029523, -0.00039122, 0.00037171, -0.00101432, 0.00136152, 0.00024405, -0.00062157, 0.00022273, -0.00029120, 0.00022674, -0.00022083, 0.00015965, -0.00016240, 0.00014231, -0.00009647, 0.00005372, -0.00004418, 0.00003860, -0.00005109, 0.00004330, -0.00003585, 0.00004008, -0.00006805, 0.00005837, -0.00003449, 0.00003143, -0.00002080, 0.00000265, -0.00000808, 0.00001138, -0.00000958, 0.00001127, -0.00002284, 0.00001580, -0.00002190, 0.00001585, 0.00000039, 0.00000122, 0.00001866, -0.00002325, 0.00001944, },
		{ 0.00001497, -0.00000934, 0.00000918, -0.00002076, 0.00001183, -0.00002449, 0.00003655, -0.00004123, 0.00006734, -0.00005963, 0.00003348, -0.00005855, 0.00007174, -0.00006302, 0.00004800, -0.00006562, 0.00007910, -0.00010238, 0.00013549, -0.00024509, 0.00023146, -0.00024340, 0.00029736, -0.00038032, 0.00036880, -0.00096452, 0.00116308, 0.00046986, -0.00073201, 0.00027756, -0.00033633, 0.00025402, -0.00023974, 0.00016711, -0.00016320, 0.00014229, -0.00009369, 0.00005276, -0.00004788, 0.00004612, -0.00005879, 0.00004865, -0.00004160, 0.00004942, -0.00007512, 0.00005870, -0.00003368, 0.00003087, -0.00002017, 0.00000354, -0.00001067, 0.00001537, -0.00001365, 0.00001555, -0.00002565, 0.00001686, -0.00002145, 0.00001256, 0.00000366, -0.00000245, 0.00002214, -0.00002392, 0.00001913, -0.00002569, },
		{ -0.00000708, 0.00000854, -0.00002196, 0.00001413, -0.00002833, 0.00004022, -0.00004489, 0.00006863, -0.00005538, 0.00003083, -0.00006116, 0.00007373, -0.00006245, 0.00005039, -0.00007301, 0.00008950, -0.00011333, 0.00014559, -0.00025429, 0.00022760, -0.00023184, 0.00028540, -0.00035336, 0.00034642, -0.00088999, 0.00095039, 0.00068457, -0.00082022, 0.00031371, -0.00036536, 0.00026680, -0.00024570, 0.00016374, -0.00015621, 0.00013644, -0.00008673, 0.00004931, -0.00004986, 0.00005159, -0.00006391, 0.00005158, -0.00004537, 0.00005638, -0.00007899, 0.00005615, -0.00003097, 0.00002901, -0.00001874, 0.00000400, -0.00001297, 0.00001886, -0.00001693, 0.00001880, -0.00002718, 0.00001670, -0.00001993, 0.00000872, 0.00000687, -0.00000562, 0.00002452, -0.00002324, 0.00001756, -0.00002296, 0.00001177, },
		{ 0.00000773, -0.00002275, 0.00001560, -0.00003086, 0.00004200, -0.00004600, 0.00006683, -0.00004857, 0.00002665, -0.00006204, 0.00007319, -0.00005929, 0.00005035, -0.00007710, 0.00009495, -0.00011765, 0.00014761, -0.00025321, 0.00021168, -0.00020879, 0.00026138, -0.00031294, 0.00030830, -0.00079674, 0.00073198, 0.00088028, -0.00088149, 0.00032881, -0.00037701, 0.00026452, -0.00023887, 0.00015029, -0.00014236, 0.00012560, -0.00007632, 0.00004384, -0.00005022, 0.00005484, -0.00006628, 0.00005202, -0.00004703, 0.00006076, -0.00007961, 0.00005092, -0.00002660, 0.00002603, -0.00001662, 0.00000406, -0.00001488, 0.00002168, -0.00001925, 0.00002086, -0.00002739, 0.00001536, -0.00001751, 0.00000455, 0.00000979, -0.00000811, 0.00002571, -0.00002127, 0.00001490, -0.00001934, 0.00000802, -0.00000452, },
		{ -0.00002311, 0.00001620, -0.00003202, 0.00004186, -0.00004467, 0.00006224, -0.00003971, 0.00002135, -0.00006141, 0.00007027, -0.00005371, 0.00004791, -0.00007773, 0.00009530, -0.00011541, 0.00014190, -0.00024263, 0.00018525, -0.00017654, 0.00022800, -0.00026234, 0.00025877, -0.00069122, 0.00051612, 0.00105010, -0.00091248, 0.00032182, -0.00037113, 0.00024769, -0.00022028, 0.00012818, -0.00012303, 0.00011090, -0.00006335, 0.00003694, -0.00004914, 0.00005587, -0.00006591, 0.00005002, -0.00004656, 0.00006253, -0.00007711, 0.00004332, -0.00002091, 0.00002218, -0.00001396, 0.00000376, -0.00001633, 0.00002370, -0.00002047, 0.00002169, -0.00002632, 0.00001300, -0.00001441, 0.00000032, 0.00001220, -0.00000977, 0.00002569, -0.00001813, 0.00001136, -0.00001510, 0.00000397, -0.00000188, 0.00000685, },
		{ 0.00001593, -0.00003187, 0.00003991, -0.00004111, 0.00005532, -0.00002939, 0.00001541, -0.00005955, 0.00006523, -0.00004598, 0.00004325, -0.00007495, 0.00009073, -0.00010712, 0.00012934, -0.00022391, 0.00015050, -0.00013793, 0.00018836, -0.00020518, 0.00020247, -0.00057984, 0.00031039, 0.00118860, -0.00091148, 0.00029307, -0.00034874, 0.00021775, -0.00019174, 0.00009938, -0.00009991, 0.00009366, -0.00004884, 0.00002922, -0.00004685, 0.00005472, -0.00006294, 0.00004579, -0.00004411, 0.00006182, -0.00007179, 0.00003382, -0.00001432, 0.00001775, -0.00001096, 0.00000317, -0.00001729, 0.00002482, -0.00002056, 0.00002132, -0.00002408, 0.00000981, -0.00001089, -0.00000371, 0.00001395, -0.00001055, 0.00002452, -0.00001403, 0.00000724, -0.00001054, -0.00000009, 0.00000064, 0.00000600, -0.00002307, },
		{ -0.00003050, 0.00003634, -0.00003568, 0.00004661, -0.00001828, 0.00000934, -0.00005677, 0.00005838, -0.00003649, 0.00003666, -0.00006905, 0.00008176, -0.00009368, 0.00011123, -0.00019890, 0.00011002, -0.00009607, 0.00014573, -0.00014513, 0.00014396, -0.00046859, 0.00012129, 0.00129196, -0.00087851, 0.00024430, -0.00031192, 0.00017697, -0.00015568, 0.00006623, -0.00007490, 0.00007529, -0.00003383, 0.00002132, -0.00004363, 0.00005158, -0.00005765, 0.00003966, -0.00003995, 0.00005889, -0.00006409, 0.00002297, -0.00000729, 0.00001307, -0.00000781, 0.00000237, -0.00001775, 0.00002502, -0.00001956, 0.00001984, -0.00002086, 0.00000605, -0.00000723, -0.00000732, 0.00001491, -0.00001045, 0.00002234, -0.00000925, 0.00000284, -0.00000597, -0.00000392, 0.00000285, 0.00000527, -0.00002266, 0.00001486, },
		{ 0.00003144, -0.00002883, 0.00003678, -0.00000707, 0.00000363, -0.00005338, 0.00005010, -0.00002570, 0.00002857, -0.00006051, 0.00006917, -0.00007633, 0.00008921, -0.00016973, 0.00006665, -0.00005411, 0.00010323, -0.00008569, 0.00008744, -0.00036274, -0.00004605, 0.00135822, -0.00081523, 0.00017848, -0.00026365, 0.00012831, -0.00011494, 0.00003125, -0.00004993, 0.00005715, -0.00001931, 0.00001379, -0.00003974, 0.00004668, -0.00005042, 0.00003208, -0.00003447, 0.00005413, -0.00005456, 0.00001138, -0.00000028, 0.00000843, -0.00000469, 0.00000145, -0.00001773, 0.00002435, -0.00001758, 0.00001744, -0.00001690, 0.00000200, -0.00000368, -0.00001032, 0.00001502, -0.00000953, 0.00001934, -0.00000407, -0.00000151, -0.00000169, -0.00000726, 0.00000460, 0.00000475, -0.00002192, 0.00001307, -0.00002810, },
		{ -0.00002105, 0.00002647, 0.00000358, -0.00000130, -0.00004968, 0.00004079, -0.00001412, 0.00001949, -0.00004995, 0.00005398, -0.00005650, 0.00006513, -0.00013868, 0.00002322, -0.00001498, 0.00006363, -0.00002985, 0.00003650, -0.00026655, -0.00018804, 0.00138724, -0.00072489, 0.00009962, -0.00020757, 0.00007511, -0.00007257, -0.00000300, -0.00002682, 0.00004051, -0.00000611, 0.00000713, -0.00003542, 0.00004032, -0.00004174, 0.00002357, -0.00002810, 0.00004803, -0.00004381, -0.00000032, 0.00000626, 0.00000412, -0.00000179, 0.00000051, -0.00001728, 0.00002289, -0.00001478, 0.00001436, -0.00001250, -0.00000205, -0.00000050, -0.00001257, 0.00001430, -0.00000791, 0.00001576, 0.00000118, -0.00000551, 0.00000205, -0.00000993, 0.00000577, 0.00000446, -0.00002092, 0.00001070, -0.00002490, 0.00002556, },
		{ 0.00001634, 0.00001310, -0.00000510, -0.00004593, 0.00003084, -0.00000226, 0.00000997, -0.00003814, 0.00003733, -0.00003576, 0.00004083, -0.00010798, -0.00001768, 0.00001882, 0.00002913, 0.00001999, -0.00000610, -0.00018314, -0.00030272, 0.00138065, -0.00061205, 0.00001249, -0.00014772, 0.00002092, -0.00003160, -0.00003418, -0.00000713, 0.00002635, 0.00000507, 0.00000167, -0.00003088, 0.00003285, -0.00003214, 0.00001468, -0.00002133, 0.00004109, -0.00003246, -0.00001154, 0.00001192, 0.00000038, 0.00000074, -0.00000038, -0.00001648, 0.00002075, -0.00001137, 0.00001085, -0.00000793, -0.00000583, 0.00000213, -0.00001399, 0.00001282, -0.00000577, 0.00001186, 0.00000619, -0.00000890, 0.00000506, -0.00001181, 0.00000630, 0.00000445, -0.00001970, 0.00000793, -0.00002116, 0.00001907, -0.00001287, },
		{ 0.00002100, -0.00000752, -0.00004231, 0.00002063, 0.00000935, 0.00000059, -0.00002585, 0.00002042, -0.00001563, 0.00001807, -0.00007964, -0.00005384, 0.00004537, 0.00000124, 0.00006222, -0.00003848, -0.00011438, -0.00038972, 0.00134161, -0.00048229, -0.00007771, -0.00008823, -0.00003087, 0.00000521, -0.00006031, 0.00000796, 0.00001540, 0.00001383, -0.00000237, -0.00002630, 0.00002464, -0.00002219, 0.00000598, -0.00001464, 0.00003382, -0.00002113, -0.00002175, 0.00001640, -0.00000262, 0.00000278, -0.00000112, -0.00001540, 0.00001812, -0.00000760, 0.00000718, -0.00000348, -0.00000910, 0.00000406, -0.00001457, 0.00001068, -0.00000331, 0.00000791, 0.00001070, -0.00001146, 0.00000720, -0.00001283, 0.00000618, 0.00000469, -0.00001833, 0.00000491, -0.00001715, 0.00001237, -0.00000480, 0.00000696, },
		{ -0.00000847, -0.00003894, 0.00001050, 0.00002022, -0.00000810, -0.00001390, 0.00000440, 0.00000247, -0.00000164, -0.00005526, -0.00008361, 0.00006351, -0.00001935, 0.00009603, -0.00005970, -0.00006095, -0.00045009, 0.00127458, -0.00034184, -0.00016572, -0.00003302, -0.00007716, 0.00003556, -0.00007989, 0.00001768, 0.00000801, 0.00001996, -0.00000497, -0.00002176, 0.00001605, -0.00001244, -0.00000201, -0.00000847, 0.00002669, -0.00001034, -0.00003050, 0.00001947, -0.00000475, 0.00000426, -0.00000166, -0.00001416, 0.00001516, -0.00000373, 0.00000363, 0.00000060, -0.00001168, 0.00000522, -0.00001436, 0.00000807, -0.00000075, 0.00000416, 0.00001446, -0.00001307, 0.00000841, -0.00001299, 0.00000544, 0.00000517, -0.00001686, 0.00000182, -0.00001314, 0.00000580, 0.00000271, -0.00000119, 0.00002695, },
		{ -0.00003588, 0.00000075, 0.00002993, -0.00001559, -0.00000301, -0.00000969, 0.00001738, -0.00001716, -0.00003601, -0.00010598, 0.00007284, -0.00003270, 0.00012136, -0.00006971, -0.00002245, -0.00048613, 0.00118493, -0.00019717, -0.00024657, 0.00001451, -0.00011547, 0.00005774, -0.00009203, 0.00002180, 0.00000419, 0.00002352, -0.00000620, -0.00001734, 0.00000742, -0.00000339, -0.00000885, -0.00000316, 0.00002010, -0.00000056, -0.00003750, 0.00002102, -0.00000597, 0.00000513, -0.00000194, -0.00001285, 0.00001207, 0.00000002, 0.00000043, 0.00000411, -0.00001345, 0.00000558, -0.00001347, 0.00000516, 0.00000171, 0.00000081, 0.00001733, -0.00001365, 0.00000871, -0.00001236, 0.00000419, 0.00000581, -0.00001532, -0.00000118, -0.00000935, -0.00000032, 0.00000929, -0.00000773, 0.00003078, -0.00000795, },
		{ -0.00000839, 0.00003812, -0.00002148, 0.00000621, -0.00002102, 0.00002826, -0.00002777, -0.00002247, -0.00012062, 0.00007372, -0.00003959, 0.00013882, -0.00006919, 0.00000244, -0.00050106, 0.00107856, -0.00005459, -0.00031591, 0.00005166, -0.00014401, 0.00007077, -0.00009648, 0.00002054, 0.00000362, 0.00002480, -0.00000631, -0.00001303, -0.00000091, 0.00000450, -0.00001420, 0.00000101, 0.00001435, 0.00000790, -0.00004254, 0.00002103, -0.00000633, 0.00000541, -0.00000195, -0.00001155, 0.00000901, 0.00000341, -0.00000223, 0.00000692, -0.00001436, 0.00000522, -0.00001203, 0.00000216, 0.00000387, -0.00000196, 0.00001921, -0.00001324, 0.00000820, -0.00001104, 0.00000252, 0.00000654, -0.00001376, -0.00000396, -0.00000597, -0.00000577, 0.00001467, -0.00001245, 0.00003245, -0.00000612, -0.00003308, },
		{ 0.00004452, -0.00002548, 0.00001328, -0.00002901, 0.00003463, -0.00003322, -0.00001468, -0.00012786, 0.00006721, -0.00004139, 0.00014954, -0.00005940, 0.00001572, -0.00049870, 0.00096152, 0.00008012, -0.00037027, 0.00007665, -0.00016185, 0.00007441, -0.00009359, 0.00001458, 0.00000569, 0.00002425, -0.00000560, -0.00000881, -0.00000868, 0.00001089, -0.00001781, 0.00000390, 0.00000963, 0.00001484, -0.00004558, 0.00001961, -0.00000591, 0.00000514, -0.00000168, -0.00001034, 0.00000614, 0.00000629, -0.00000423, 0.00000895, -0.00001442, 0.00000424, -0.00001024, -0.00000074, 0.00000558, -0.00000406, 0.00002009, -0.00001193, 0.00000700, -0.00000921, 0.00000059, 0.00000729, -0.00001219, -0.00000641, -0.00000315, -0.00001038, 0.00001868, -0.00001528, 0.00003210, -0.00000321, -0.00003049, -0.00001674, },
		{ -0.00002741, 0.00001794, -0.00003335, 0.00003640, -0.00003375, -0.00001216, -0.00012862, 0.00005488, -0.00003985, 0.00015495, -0.00004195, 0.00001979, -0.00048319, 0.00083964, 0.00020201, -0.00040726, 0.00008868, -0.00016891, 0.00006915, -0.00008430, 0.00000497, 0.00000959, 0.00002245, -0.00000445, -0.00000463, -0.00001564, 0.00001551, -0.00001961, 0.00000549, 0.00000601, 0.00002017, -0.00004669, 0.00001695, -0.00000486, 0.00000440, -0.00000116, -0.00000927, 0.00000358, 0.00000853, -0.00000551, 0.00001020, -0.00001372, 0.00000279, -0.00000827, -0.00000336, 0.00000672, -0.00000544, 0.00002002, -0.00000987, 0.00000531, -0.00000704, -0.00000145, 0.00000797, -0.00001063, -0.00000845, -0.00000094, -0.00001407, 0.00002127, -0.00001628, 0.00002997, 0.00000043, -0.00002797, -0.00002416, 0.00004898, },
		{ 0.00002009, -0.00003404, 0.00003389, -0.00003001, -0.00001396, -0.00012425, 0.00003870, -0.00003691, 0.00015659, -0.00001863, 0.00001723, -0.00045861, 0.00071819, 0.00030720, -0.00042570, 0.00008797, -0.00016590, 0.00005613, -0.00007000, -0.00000698, 0.00001437, 0.00002002, -0.00000324, -0.00000044, -0.00002161, 0.00001824, -0.00001963, 0.00000588, 0.00000346, 0.00002394, -0.00004605, 0.00001330, -0.00000336, 0.00000327, -0.00000041, -0.00000838, 0.00000141, 0.00001008, -0.00000607, 0.00001073, -0.00001239, 0.00000106, -0.00000631, -0.00000555, 0.00000726, -0.00000613, 0.00001912, -0.00000727, 0.00000335, -0.00000472, -0.00000345, 0.00000849, -0.00000910, -0.00001005, 0.00000064, -0.00001685, 0.00002248, -0.00001561, 0.00002644, 0.00000443, -0.00002540, -0.00003059, 0.00005141, -0.00002725, },
		{ -0.00003133, 0.00002772, -0.00002299, -0.00001884, -0.00011642, 0.00002077, -0.00003447, 0.00015588, 0.00000889, 0.00001056, -0.00042876, 0.00060164, 0.00039308, -0.00042559, 0.00007569, -0.00015428, 0.00003701, -0.00005239, -0.00001984, 0.00001907, 0.00001760, -0.00000232, 0.00000380, -0.00002645, 0.00001904, -0.00001803, 0.00000527, 0.00000185, 0.00002629, -0.00004394, 0.00000895, -0.00000160, 0.00000188, 0.00000049, -0.00000767, -0.00000033, 0.00001092, -0.00000598, 0.00001067, -0.00001060, -0.00000077, -0.00000452, -0.00000723, 0.00000718, -0.00000620, 0.00001757, -0.00000435, 0.00000131, -0.00000243, -0.00000527, 0.00000878, -0.00000759, -0.00001120, 0.00000164, -0.00001879, 0.00002245, -0.00001356, 0.00002192, 0.00000840, -0.00002264, -0.00003596, 0.00005184, -0.00002511, 0.00001983, },
		{ 0.00001881, -0.00001394, -0.00002532, -0.00010688, 0.00000316, -0.00003421, 0.00015394, 0.00003911, 0.00000202, -0.00039690, 0.00049348, 0.00045828, -0.00040813, 0.00005380, -0.00013603, 0.00001379, -0.00003334, -0.00003216, 0.00002280, 0.00001571, -0.00000199, 0.00000809, -0.00003007, 0.00001800, -0.00001510, 0.00000393, 0.00000097, 0.00002744, -0.00004065, 0.00000423, 0.00000023, 0.00000034, 0.00000150, -0.00000713, -0.00000165, 0.00001113, -0.00000536, 0.00001017, -0.00000854, -0.00000252, -0.00000305, -0.00000835, 0.00000654, -0.00000577, 0.00001555, -0.00000134, -0.00000059, -0.00000034, -0.00000680, 0.00000879, -0.00000609, -0.00001194, 0.00000214, -0.00002002, 0.00002138, -0.00001044, 0.00001686, 0.00001198, -0.00001959, -0.00004027, 0.00005038, -0.00002120, 0.00001745, -0.00002576, },
		{ -0.00000421, -0.00003193, -0.00009732, -0.00001233, -0.00003734, 0.00015147, 0.00007092, -0.00000652, -0.00036568, 0.00039614, 0.00050272, -0.00037546, 0.00002487, -0.00011348, -0.00001136, -0.00001469, -0.00004264, 0.00002482, 0.00001476, -0.00000242, 0.00001239, -0.00003241, 0.00001528, -0.00001118, 0.00000219, 0.00000060, 0.00002764, -0.00003655, -0.00000056, 0.00000194, -0.00000125, 0.00000253, -0.00000670, -0.00000258, 0.00001079, -0.00000437, 0.00000939, -0.00000639, -0.00000403, -0.00000197, -0.00000891, 0.00000545, -0.00000501, 0.00001328, 0.00000153, -0.00000220, 0.00000145, -0.00000796, 0.00000850, -0.00000461, -0.00001231, 0.00000227, -0.00002068, 0.00001949, -0.00000663, 0.00001171, 0.00001483, -0.00001621, -0.00004350, 0.00004719, -0.00001586, 0.00001337, -0.00001800, 0.00000825, },
		{ -0.00003731, -0.00008922, -0.00002429, -0.00004453, 0.00014864, 0.00010362, -0.00001368, -0.00033702, 0.00031101, 0.00052739, -0.00033053, -0.00000816, -0.00008907, -0.00003632, 0.00000188, -0.00005022, 0.00002463, 0.00001500, -0.00000369, 0.00001661, -0.00003346, 0.00001114, -0.00000667, 0.00000040, 0.00000049, 0.00002716, -0.00003196, -0.00000513, 0.00000339, -0.00000281, 0.00000352, -0.00000635, -0.00000319, 0.00001005, -0.00000318, 0.00000852, -0.00000432, -0.00000518, -0.00000132, -0.00000897, 0.00000403, -0.00000408, 0.00001096, 0.00000408, -0.00000338, 0.00000285, -0.00000871, 0.00000789, -0.00000316, -0.00001238, 0.00000217, -0.00002096, 0.00001702, -0.00000246, 0.00000687, 0.00001674, -0.00001248, -0.00004567, 0.00004252, -0.00000948, 0.00000809, -0.00000887, -0.00000281, 0.00000487, },
		{ -0.00008364, -0.00003187, -0.00005582, 0.00014509, 0.00013691, -0.00001851, -0.00031213, 0.00023852, 0.00053424, -0.00027675, -0.00004225, -0.00006518, -0.00005918, 0.00001500, -0.00005419, 0.00002201, 0.00001647, -0.00000574, 0.00002062, -0.00003322, 0.00000585, -0.00000200, -0.00000112, 0.00000041, 0.00002624, -0.00002719, -0.00000927, 0.00000447, -0.00000425, 0.00000440, -0.00000602, -0.00000358, 0.00000905, -0.00000197, 0.00000772, -0.00000247, -0.00000589, -0.00000109, -0.00000862, 0.00000244, -0.00000314, 0.00000877, 0.00000616, -0.00000407, 0.00000383, -0.00000905, 0.00000699, -0.00000173, -0.00001222, 0.00000197, -0.00002102, 0.00001420, 0.00000175, 0.00000264, 0.00001757, -0.00000847, -0.00004678, 0.00003664, -0.00000250, 0.00000220, 0.00000076, -0.00001322, 0.00001211, -0.00004035, },
		{ -0.00003482, -0.00007060, 0.00013997, 0.00017085, -0.00002051, -0.00029161, 0.00017827, 0.00052587, -0.00021780, -0.00007451, -0.00004383, -0.00007841, 0.00002374, -0.00005421, 0.00001704, 0.00001902, -0.00000840, 0.00002425, -0.00003172, -0.00000024, 0.00000243, -0.00000210, 0.00000018, 0.00002510, -0.00002248, -0.00001281, 0.00000515, -0.00000553, 0.00000511, -0.00000562, -0.00000384, 0.00000794, -0.00000091, 0.00000711, -0.00000094, -0.00000615, -0.00000123, -0.00000799, 0.00000084, -0.00000234, 0.00000683, 0.00000769, -0.00000427, 0.00000443, -0.00000902, 0.00000584, -0.00000034, -0.00001191, 0.00000181, -0.00002098, 0.00001122, 0.00000573, -0.00000075, 0.00001728, -0.00000431, -0.00004684, 0.00002983, 0.00000462, -0.00000375, 0.00001009, -0.00002199, 0.00001656, -0.00004030, -0.00008120, },
		{ -0.00008768, 0.00013207, 0.00020574, -0.00001946, -0.00027556, 0.00012928, 0.00050526, -0.00015726, -0.00010244, -0.00002663, -0.00009298, 0.00002762, -0.00005037, 0.00001007, 0.00002232, -0.00001140, 0.00002729, -0.00002902, -0.00000682, 0.00000627, -0.00000234, -0.00000031, 0.00002387, -0.00001801, -0.00001566, 0.00000540, -0.00000663, 0.00000560, -0.00000512, -0.00000407, 0.00000686, -0.00000013, 0.00000678, 0.00000025, -0.00000598, -0.00000163, -0.00000721, -0.00000064, -0.00000180, 0.00000525, 0.00000865, -0.00000401, 0.00000470, -0.00000866, 0.00000449, 0.00000098, -0.00001150, 0.00000177, -0.00002094, 0.00000820, 0.00000931, -0.00000317, 0.00001595, -0.00000020, -0.00004585, 0.00002240, 0.00001146, -0.00000922, 0.00001838, -0.00002831, 0.00001762, -0.00003684, -0.00008200, -0.00003351, },
		{ 0.00011994, 0.00024194, -0.00001531, -0.00026375, 0.00009016, 0.00047554, -0.00009837, -0.00012410, -0.00001454, -0.00010240, 0.00002669, -0.00004315, 0.00000169, 0.00002592, -0.00001439, 0.00002952, -0.00002519, -0.00001355, 0.00000922, -0.00000173, -0.00000108, 0.00002263, -0.00001388, -0.00001781, 0.00000527, -0.00000756, 0.00000585, -0.00000446, -0.00000434, 0.00000593, 0.00000029, 0.00000676, 0.00000109, -0.00000545, -0.00000218, -0.00000642, -0.00000186, -0.00000159, 0.00000404, 0.00000908, -0.00000339, 0.00000473, -0.00000808, 0.00000303, 0.00000222, -0.00001104, 0.00000193, -0.00002093, 0.00000523, 0.00001239, -0.00000461, 0.00001377, 0.00000365, -0.00004383, 0.00001464, 0.00001762, -0.00001378, 0.00002507, -0.00003168, 0.00001509, -0.00003013, -0.00008562, -0.00002882, -0.00010542, },
		{ 0.00027973, -0.00000805, -0.00025577, 0.00005934, 0.00043969, -0.00004384, -0.00013829, -0.00000791, -0.00010677, 0.00002147, -0.00003336, -0.00000729, 0.00002927, -0.00001699, 0.00003073, -0.00002033, -0.00002015, 0.00001108, -0.00000026, -0.00000208, 0.00002140, -0.00001012, -0.00001930, 0.00000485, -0.00000834, 0.00000584, -0.00000361, -0.00000471, 0.00000522, 0.00000032, 0.00000704, 0.00000164, -0.00000467, -0.00000274, -0.00000575, -0.00000275, -0.00000173, 0.00000320, 0.00000904, -0.00000253, 0.00000464, -0.00000736, 0.00000152, 0.00000333, -0.00001056, 0.00000230, -0.00002094, 0.00000231, 0.00001495, -0.00000515, 0.00001097, 0.00000700, -0.00004081, 0.00000683, 0.00002282, -0.00001713, 0.00002980, -0.00003187, 0.00000915, -0.00002073, -0.00009118, -0.00002212, -0.00012189, 0.00010213, },
		{ 0.00000249, -0.00025118, 0.00003531, 0.00040034, 0.00000432, -0.00014453, -0.00000645, -0.00010667, 0.00001286, -0.00002204, -0.00001599, 0.00003181, -0.00001884, 0.00003076, -0.00001454, -0.00002637, 0.00001176, 0.00000196, -0.00000320, 0.00002013, -0.00000671, -0.00002022, 0.00000423, -0.00000899, 0.00000556, -0.00000257, -0.00000522, 0.00000477, -0.00000004, 0.00000752, 0.00000200, -0.00000378, -0.00000318, -0.00000531, -0.00000325, -0.00000219, 0.00000265, 0.00000867, -0.00000159, 0.00000455, -0.00000659, 0.00000004, 0.00000427, -0.00001007, 0.00000285, -0.00002091, -0.00000059, 0.00001702, -0.00000496, 0.00000786, 0.00000960, -0.00003683, -0.00000079, 0.00002681, -0.00001908, 0.00003241, -0.00002901, 0.00000036, -0.00000960, -0.00009750, -0.00001503, -0.00013508, 0.00007739, 0.00031910, },
		},
		{ { -0.00005255, -0.00003280, -0.00013423, 0.00005227, 0.00036793, 0.00003680, -0.00010553, -0.00002723, -0.00002984, -0.00001414, -0.00003593, 0.00001635, -0.00007424, 0.00001156, 0.00001831, -0.00000270, -0.00000933, -0.00001472, 0.00002030, -0.00001730, 0.00002526, -0.00001354, 0.00001580, -0.00000615, 0.00000100, 0.00000760, 0.00000652, 0.00000601, -0.00001187, -0.00000203, -0.00002557, 0.00000762, 0.00001062, 0.00000762, -0.00002557, -0.00000203, -0.00001187, 0.00000601, 0.00000652, 0.00000760, 0.00000100, -0.00000615, 0.00001580, -0.00001354, 0.00002526, -0.00001730, 0.00002030, -0.00001472, -0.00000933, -0.00000270, 0.00001831, 0.00001156, -0.00007424, 0.00001635, -0.00003593, -0.00001414, -0.00002984, -0.00002723, -0.00010553, 0.00003680, 0.00036793, 0.00005227, -0.00013423, -0.00003280, },
		{ -0.00003241, -0.00011778, 0.00001384, 0.00032780, 0.00009332, -0.00010150, -0.00003457, -0.00003013, -0.00001295, -0.00003708, 0.00001698, -0.00006623, -0.00000628, 0.00003059, -0.00000651, -0.00000519, -0.00002102, 0.00002169, -0.00001526, 0.00002126, -0.00000978, 0.00001177, 0.00000038, -0.00000500, 0.00000987, 0.00000456, 0.00001047, -0.00001444, 0.00000190, -0.00002736, 0.00000340, 0.00001104, 0.00001003, -0.00002159, -0.00000694, -0.00000877, 0.00000133, 0.00000876, 0.00000431, 0.00000734, -0.00001147, 0.00001818, -0.00001577, 0.00002704, -0.00001602, 0.00001507, -0.00000552, -0.00001577, 0.00000406, 0.00000446, 0.00002711, -0.00007759, 0.00001196, -0.00003199, -0.00001798, -0.00002645, -0.00002598, -0.00009556, -0.00002978, 0.00039733, 0.00010439, -0.00015248, -0.00003420, -0.00005238, },
		{ -0.00010372, -0.00001309, 0.00028243, 0.00013668, -0.00008546, -0.00004451, -0.00003000, -0.00001244, -0.00003706, 0.00001548, -0.00005586, -0.00002349, 0.00003879, -0.00000593, -0.00000409, -0.00002397, 0.00001930, -0.00001051, 0.00001576, -0.00000513, 0.00000679, 0.00000713, -0.00000979, 0.00001075, 0.00000298, 0.00001448, -0.00001613, 0.00000440, -0.00002670, -0.00000249, 0.00001215, 0.00001077, -0.00001592, -0.00001219, -0.00000560, -0.00000322, 0.00001098, 0.00000062, 0.00001299, -0.00001473, 0.00001842, -0.00001611, 0.00002627, -0.00001141, 0.00000665, 0.00000535, -0.00002301, 0.00001160, -0.00000800, 0.00003753, -0.00007439, 0.00000260, -0.00002411, -0.00002593, -0.00001822, -0.00003248, -0.00007251, -0.00010041, 0.00040918, 0.00017220, -0.00017164, -0.00003708, -0.00005197, -0.00003249, },
		{ -0.00003050, 0.00023578, 0.00016641, -0.00006133, -0.00005295, -0.00003205, -0.00001089, -0.00003723, 0.00001330, -0.00004516, -0.00003778, 0.00004138, -0.00000065, -0.00000583, -0.00002393, 0.00001391, -0.00000410, 0.00000977, -0.00000038, 0.00000169, 0.00001317, -0.00001274, 0.00001019, 0.00000172, 0.00001788, -0.00001664, 0.00000518, -0.00002362, -0.00000964, 0.00001359, 0.00001026, -0.00000927, -0.00001707, -0.00000291, -0.00000719, 0.00001271, -0.00000271, 0.00001698, -0.00001530, 0.00001637, -0.00001453, 0.00002302, -0.00000399, -0.00000375, 0.00001621, -0.00002915, 0.00001757, -0.00001638, 0.00004077, -0.00006384, -0.00001181, -0.00001218, -0.00003809, -0.00000517, -0.00004590, -0.00004048, -0.00016660, 0.00039627, 0.00025674, -0.00019010, -0.00004201, -0.00005150, -0.00003247, -0.00009249, },
		{ 0.00019040, 0.00018384, -0.00003375, -0.00005639, -0.00003776, -0.00000765, -0.00003804, 0.00001116, -0.00003536, -0.00004793, 0.00003819, 0.00000842, -0.00000925, -0.00002197, 0.00000681, 0.00000257, 0.00000439, 0.00000363, -0.00000276, 0.00001775, -0.00001352, 0.00000839, 0.00000060, 0.00002055, -0.00001580, 0.00000415, -0.00001844, -0.00001740, 0.00001480, 0.00000906, -0.00000248, -0.00002084, -0.00000123, -0.00001008, 0.00001340, -0.00000487, 0.00001853, -0.00001293, 0.00001225, -0.00001135, 0.00001775, 0.00000527, -0.00001454, 0.00002523, -0.00003242, 0.00002004, -0.00001896, 0.00003620, -0.00004666, -0.00002993, 0.00000242, -0.00005298, 0.00001066, -0.00006284, -0.00000622, -0.00021858, 0.00035247, 0.00035700, -0.00020501, -0.00005004, -0.00005096, -0.00003201, -0.00008424, -0.00004019, },
		{ 0.00019100, -0.00000671, -0.00005304, -0.00004688, -0.00000350, -0.00003888, 0.00000885, -0.00002673, -0.00005393, 0.00003029, 0.00001939, -0.00001259, -0.00001955, -0.00000048, 0.00000810, 0.00000064, 0.00000622, -0.00000593, 0.00002037, -0.00001215, 0.00000575, -0.00000057, 0.00002235, -0.00001348, 0.00000141, -0.00001169, -0.00002495, 0.00001517, 0.00000778, 0.00000365, -0.00002285, -0.00000101, -0.00001146, 0.00001257, -0.00000525, 0.00001721, -0.00000784, 0.00000666, -0.00000715, 0.00001117, 0.00001521, -0.00002413, 0.00003085, -0.00003159, 0.00001803, -0.00001537, 0.00002488, -0.00002522, -0.00004900, 0.00001701, -0.00006758, 0.00002554, -0.00007806, 0.00002217, -0.00024718, 0.00027447, 0.00046888, -0.00021169, -0.00006302, -0.00004979, -0.00003128, -0.00007861, -0.00004406, 0.00014811, },
		{ 0.00001733, -0.00004329, -0.00005728, -0.00000049, -0.00003818, 0.00000541, -0.00001866, -0.00005679, 0.00001970, 0.00002980, -0.00001391, -0.00001814, -0.00000652, 0.00001132, -0.00000077, 0.00000695, -0.00000746, 0.00002087, -0.00000891, 0.00000276, -0.00000189, 0.00002312, -0.00000965, -0.00000278, -0.00000409, -0.00003142, 0.00001412, 0.00000699, 0.00000843, -0.00002261, -0.00000255, -0.00001103, 0.00000991, -0.00000352, 0.00001309, -0.00000072, 0.00000047, -0.00000267, 0.00000411, 0.00002463, -0.00003117, 0.00003213, -0.00002637, 0.00001175, -0.00000679, 0.00000942, -0.00000314, -0.00006534, 0.00002807, -0.00007811, 0.00003490, -0.00008573, 0.00003701, -0.00024582, 0.00016337, 0.00058435, -0.00020350, -0.00008363, -0.00004676, -0.00003120, -0.00007449, -0.00004430, 0.00011045, 0.00018973, },
		{ -0.00002961, -0.00006546, -0.00000142, -0.00003400, -0.00000045, -0.00001016, -0.00005800, 0.00000880, 0.00003726, -0.00001165, -0.00001879, -0.00001032, 0.00001153, 0.00000043, 0.00000575, -0.00000730, 0.00001935, -0.00000431, -0.00000009, -0.00000337, 0.00002263, -0.00000435, -0.00000801, 0.00000356, -0.00003602, 0.00001124, 0.00000706, 0.00001140, -0.00001995, -0.00000587, -0.00000876, 0.00000544, 0.00000019, 0.00000676, 0.00000734, -0.00000525, 0.00000132, -0.00000261, 0.00003252, -0.00003477, 0.00002888, -0.00001743, 0.00000258, 0.00000436, -0.00000653, 0.00001539, -0.00007513, 0.00003217, -0.00008091, 0.00003463, -0.00008103, 0.00003296, -0.00021230, 0.00002576, 0.00069150, -0.00017222, -0.00011510, -0.00004011, -0.00003333, -0.00007013, -0.00004349, 0.00007906, 0.00018110, 0.00003782, },
		{ -0.00006757, -0.00000883, -0.00002484, -0.00000962, -0.00000044, -0.00005898, -0.00000038, 0.00004006, -0.00000505, -0.00002185, -0.00001151, 0.00000862, 0.00000402, 0.00000291, -0.00000570, 0.00001622, 0.00000102, -0.00000237, -0.00000486, 0.00002068, 0.00000224, -0.00001372, 0.00001043, -0.00003812, 0.00000635, 0.00000812, 0.00001243, -0.00001500, -0.00001067, -0.00000498, -0.00000044, 0.00000529, -0.00000075, 0.00001499, -0.00000944, 0.00000415, -0.00000837, 0.00003821, -0.00003465, 0.00002173, -0.00000629, -0.00000729, 0.00001502, -0.00001910, 0.00002658, -0.00007543, 0.00002696, -0.00007365, 0.00002227, -0.00006158, 0.00000849, -0.00014997, -0.00012624, 0.00077528, -0.00010922, -0.00016033, -0.00002814, -0.00003945, -0.00006345, -0.00004434, 0.00005565, 0.00016540, 0.00005579, -0.00001577, },
		{ -0.00002406, -0.00001037, -0.00002203, 0.00001061, -0.00006051, -0.00000663, 0.00003760, 0.00000553, -0.00002677, -0.00001044, 0.00000314, 0.00000928, -0.00000090, -0.00000319, 0.00001208, 0.00000639, -0.00000380, -0.00000614, 0.00001714, 0.00000978, -0.00001926, 0.00001575, -0.00003734, -0.00000040, 0.00001001, 0.00001167, -0.00000824, -0.00001636, -0.00000031, -0.00000696, 0.00001081, -0.00000810, 0.00002084, -0.00001124, 0.00000540, -0.00001276, 0.00004134, -0.00003106, 0.00001191, 0.00000504, -0.00001539, 0.00002218, -0.00002513, 0.00002804, -0.00006505, 0.00001202, -0.00005611, -0.00000199, -0.00002839, -0.00003333, -0.00006772, -0.00027599, 0.00081949, -0.00000721, -0.00022060, -0.00001007, -0.00005085, -0.00005272, -0.00004914, 0.00004157, 0.00014260, 0.00007303, -0.00000567, -0.00006053, },
		{ 0.00000798, -0.00003627, 0.00002202, -0.00006237, -0.00000992, 0.00003057, 0.00001868, -0.00003227, -0.00000809, -0.00000386, 0.00001510, -0.00000481, -0.00000047, 0.00000766, 0.00001114, -0.00000422, -0.00000694, 0.00001204, 0.00001774, -0.00002390, 0.00001890, -0.00003362, -0.00000856, 0.00001231, 0.00000960, -0.00000044, -0.00002207, 0.00000435, -0.00001308, 0.00001554, -0.00001384, 0.00002368, -0.00001009, 0.00000493, -0.00001563, 0.00004194, -0.00002466, 0.00000098, 0.00001450, -0.00001960, 0.00002373, -0.00002298, 0.00001940, -0.00004512, -0.00001071, -0.00003066, -0.00003499, 0.00001392, -0.00008508, 0.00002113, -0.00040457, 0.00080932, 0.00013758, -0.00029415, 0.00001292, -0.00006750, -0.00003727, -0.00005908, 0.00003720, 0.00011307, 0.00009106, -0.00000206, -0.00004323, -0.00004638, },
		{ -0.00004959, 0.00003182, -0.00006337, -0.00001131, 0.00002077, 0.00003211, -0.00003653, -0.00000577, -0.00001104, 0.00002023, -0.00000783, 0.00000168, 0.00000369, 0.00001471, -0.00000359, -0.00000706, 0.00000561, 0.00002542, -0.00002693, 0.00001948, -0.00002721, -0.00001740, 0.00001439, 0.00000686, 0.00000745, -0.00002681, 0.00000798, -0.00001766, 0.00001825, -0.00001673, 0.00002268, -0.00000592, 0.00000295, -0.00001709, 0.00004025, -0.00001633, -0.00000955, 0.00002049, -0.00001871, 0.00001899, -0.00001302, 0.00000268, -0.00001911, -0.00003705, -0.00000196, -0.00007101, 0.00005784, -0.00013624, 0.00010097, -0.00049375, 0.00073439, 0.00032324, -0.00037497, 0.00003699, -0.00008735, -0.00001821, -0.00007365, 0.00004139, 0.00007833, 0.00011012, -0.00000559, -0.00001718, -0.00007275, 0.00002701, },
		{ 0.00003741, -0.00006167, -0.00001266, 0.00001071, 0.00004321, -0.00003769, -0.00000486, -0.00001703, 0.00002344, -0.00000905, 0.00000257, 0.00000082, 0.00001668, -0.00000197, -0.00000638, -0.00000167, 0.00003199, -0.00002771, 0.00001734, -0.00001869, -0.00002601, 0.00001556, 0.00000420, 0.00001445, -0.00002960, 0.00000964, -0.00001972, 0.00001795, -0.00001593, 0.00001764, 0.00000082, -0.00000002, -0.00001739, 0.00003666, -0.00000697, -0.00001847, 0.00002224, -0.00001279, 0.00000897, 0.00000233, -0.00001803, 0.00000767, -0.00006127, 0.00002388, -0.00010283, 0.00009442, -0.00017533, 0.00015682, -0.00052930, 0.00059154, 0.00054063, -0.00045218, 0.00005528, -0.00010617, 0.00000136, -0.00009037, 0.00005116, 0.00004144, 0.00012859, -0.00001433, 0.00001347, -0.00009806, 0.00004217, -0.00005841, },
		{ -0.00005541, -0.00001601, 0.00000295, 0.00004963, -0.00003428, -0.00000636, -0.00002072, 0.00002379, -0.00000782, 0.00000167, -0.00000048, 0.00001683, 0.00000050, -0.00000493, -0.00000914, 0.00003660, -0.00002580, 0.00001263, -0.00000887, -0.00003344, 0.00001518, 0.00000225, 0.00001972, -0.00002968, 0.00000862, -0.00001857, 0.00001411, -0.00001124, 0.00000901, 0.00000921, -0.00000327, -0.00001688, 0.00003157, 0.00000270, -0.00002514, 0.00001990, -0.00000317, -0.00000381, 0.00001910, -0.00003737, 0.00002919, -0.00007735, 0.00004067, -0.00012336, 0.00011528, -0.00019229, 0.00017725, -0.00050394, 0.00038676, 0.00077293, -0.00051046, 0.00005859, -0.00011788, 0.00001664, -0.00010502, 0.00006189, 0.00000692, 0.00014310, -0.00002409, 0.00004258, -0.00011611, 0.00004850, -0.00005912, 0.00003627, },
		{ -0.00002289, -0.00000056, 0.00004983, -0.00002569, -0.00001068, -0.00002152, 0.00002078, -0.00000387, -0.00000121, -0.00000000, 0.00001514, 0.00000363, -0.00000289, -0.00001600, 0.00003849, -0.00002100, 0.00000580, 0.00000127, -0.00003877, 0.00001279, 0.00000146, 0.00002267, -0.00002666, 0.00000467, -0.00001408, 0.00000686, -0.00000316, -0.00000212, 0.00001794, -0.00000599, -0.00001588, 0.00002533, 0.00001220, -0.00002947, 0.00001453, 0.00000780, -0.00001586, 0.00003262, -0.00004985, 0.00003996, -0.00008036, 0.00004361, -0.00012728, 0.00011452, -0.00018086, 0.00015684, -0.00041925, 0.00013568, 0.00099653, -0.00053165, 0.00003674, -0.00011577, 0.00002206, -0.00011249, 0.00006809, -0.00001987, 0.00014915, -0.00002941, 0.00006336, -0.00012095, 0.00004190, -0.00004912, 0.00002677, -0.00004333, },
		{ 0.00000103, 0.00004355, -0.00001237, -0.00001743, -0.00001943, 0.00001446, 0.00000261, -0.00000593, 0.00000216, 0.00001184, 0.00000714, -0.00000050, -0.00002146, 0.00003712, -0.00001348, -0.00000241, 0.00001070, -0.00004134, 0.00000827, 0.00000198, 0.00002312, -0.00002059, -0.00000198, -0.00000668, -0.00000306, 0.00000712, -0.00001413, 0.00002551, -0.00000741, -0.00001465, 0.00001823, 0.00002125, -0.00003185, 0.00000776, 0.00001733, -0.00002356, 0.00003866, -0.00005119, 0.00003646, -0.00006779, 0.00003067, -0.00011252, 0.00009030, -0.00014016, 0.00009762, -0.00028616, -0.00013751, 0.00118358, -0.00049748, -0.00001915, -0.00009414, 0.00001268, -0.00010803, 0.00006458, -0.00003394, 0.00014238, -0.00002499, 0.00006995, -0.00010840, 0.00002032, -0.00002772, 0.00000886, -0.00002543, -0.00003381, },
		{ 0.00003187, 0.00000412, -0.00002549, -0.00001506, 0.00000543, 0.00001099, -0.00001196, 0.00000558, 0.00000740, 0.00001070, 0.00000198, -0.00002489, 0.00003234, -0.00000379, -0.00001106, 0.00001844, -0.00004075, 0.00000183, 0.00000359, 0.00002128, -0.00001199, -0.00001056, 0.00000260, -0.00001429, 0.00001794, -0.00002511, 0.00003047, -0.00000694, -0.00001335, 0.00001051, 0.00002972, -0.00003291, 0.00000136, 0.00002286, -0.00002407, 0.00003445, -0.00003941, 0.00001810, -0.00004045, 0.00000337, -0.00008099, 0.00004561, -0.00007526, 0.00000900, -0.00012367, -0.00040298, 0.00130571, -0.00039300, -0.00011410, -0.00005031, -0.00001418, -0.00008875, 0.00004788, -0.00003193, 0.00011992, -0.00000727, 0.00005902, -0.00007739, -0.00001529, 0.00000323, -0.00001548, -0.00000330, -0.00004789, 0.00000711, },
		{ 0.00002137, -0.00003320, -0.00000946, -0.00000524, 0.00002023, -0.00001846, 0.00000956, 0.00000247, 0.00001391, 0.00000434, -0.00002591, 0.00002440, 0.00000717, -0.00001908, 0.00002372, -0.00003701, -0.00000597, 0.00000579, 0.00001770, -0.00000182, -0.00001986, 0.00001233, -0.00002512, 0.00002735, -0.00003320, 0.00003164, -0.00000429, -0.00001202, 0.00000240, 0.00003748, -0.00003322, -0.00000325, 0.00002269, -0.00001608, 0.00001949, -0.00001547, -0.00001232, -0.00000262, -0.00003324, -0.00003849, -0.00001214, 0.00000356, -0.00009369, 0.00004393, -0.00062936, 0.00133858, -0.00021026, -0.00024674, 0.00001371, -0.00005754, -0.00005483, 0.00001737, -0.00001330, 0.00008164, 0.00002431, 0.00003088, -0.00003060, -0.00006069, 0.00003928, -0.00004235, 0.00002001, -0.00006290, 0.00001580, 0.00001712, },
		{ -0.00003872, -0.00000392, -0.00001624, 0.00002906, -0.00002437, 0.00001321, -0.00000214, 0.00001629, 0.00000651, -0.00002447, 0.00001400, 0.00001823, -0.00002542, 0.00002603, -0.00003049, -0.00001427, 0.00000785, 0.00001321, 0.00000872, -0.00002847, 0.00002088, -0.00003373, 0.00003354, -0.00003685, 0.00002834, 0.00000049, -0.00001063, -0.00000573, 0.00004424, -0.00003307, -0.00000533, 0.00001646, -0.00000025, -0.00000411, 0.00001665, -0.00004896, 0.00003855, -0.00007136, 0.00000633, -0.00007234, 0.00008263, -0.00019196, 0.00019105, -0.00078862, 0.00126635, 0.00004861, -0.00040767, 0.00009019, -0.00011201, -0.00001025, -0.00002396, 0.00001900, 0.00003082, 0.00006663, -0.00001018, 0.00002552, -0.00010873, 0.00007398, -0.00006646, 0.00004040, -0.00007573, 0.00002423, 0.00000239, 0.00003658, },
		{ 0.00000033, -0.00002616, 0.00003614, -0.00002856, 0.00001557, -0.00000560, 0.00001738, 0.00000854, -0.00002088, 0.00000223, 0.00002807, -0.00002919, 0.00002521, -0.00002194, -0.00002203, 0.00000901, 0.00000869, 0.00001838, -0.00003496, 0.00002671, -0.00003855, 0.00003508, -0.00003516, 0.00002051, 0.00000695, -0.00000912, -0.00001343, 0.00004952, -0.00003231, -0.00000508, 0.00000528, 0.00002079, -0.00003197, 0.00005061, -0.00008391, 0.00007433, -0.00010198, 0.00004373, -0.00012338, 0.00014760, -0.00026763, 0.00029511, -0.00086071, 0.00108534, 0.00036959, -0.00057903, 0.00016583, -0.00016798, 0.00003734, -0.00006972, 0.00005867, -0.00002597, 0.00011316, -0.00005638, 0.00008175, -0.00015055, 0.00009998, -0.00008216, 0.00005362, -0.00008304, 0.00002931, -0.00000914, 0.00004708, -0.00004043, },
		{ -0.00003381, 0.00004025, -0.00003000, 0.00001575, -0.00000714, 0.00001674, 0.00001058, -0.00001574, -0.00000963, 0.00003542, -0.00002976, 0.00002149, -0.00001230, -0.00002824, 0.00000859, 0.00000492, 0.00002605, -0.00003817, 0.00002867, -0.00003853, 0.00003126, -0.00002798, 0.00000878, 0.00001435, -0.00000742, -0.00002011, 0.00005264, -0.00003037, -0.00000353, -0.00000845, 0.00004291, -0.00005821, 0.00007896, -0.00010883, 0.00009628, -0.00011685, 0.00006496, -0.00015504, 0.00018631, -0.00030631, 0.00034056, -0.00083710, 0.00080595, 0.00072689, -0.00073580, 0.00022324, -0.00021290, 0.00007778, -0.00011107, 0.00009699, -0.00007984, 0.00015506, -0.00009793, 0.00012775, -0.00017719, 0.00011056, -0.00008465, 0.00005622, -0.00008210, 0.00002841, -0.00001483, 0.00005096, -0.00003732, 0.00000240, },
		{ 0.00004049, -0.00002795, 0.00001310, -0.00000625, 0.00001413, 0.00001282, -0.00000982, -0.00002021, 0.00003927, -0.00002695, 0.00001544, -0.00000269, -0.00003203, 0.00000620, 0.00000242, 0.00003095, -0.00003740, 0.00002620, -0.00003338, 0.00002220, -0.00001605, -0.00000558, 0.00002168, -0.00000551, -0.00002509, 0.00005286, -0.00002641, -0.00000229, -0.00002160, 0.00006133, -0.00007670, 0.00009471, -0.00011671, 0.00009804, -0.00011035, 0.00006432, -0.00016073, 0.00019136, -0.00030028, 0.00032159, -0.00072264, 0.00045248, 0.00108511, -0.00084877, 0.00024372, -0.00023366, 0.00010035, -0.00013848, 0.00012454, -0.00012127, 0.00018295, -0.00012500, 0.00015410, -0.00018152, 0.00010114, -0.00007120, 0.00004646, -0.00007156, 0.00002007, -0.00001324, 0.00004743, -0.00002927, 0.00000193, -0.00003829, },
		{ -0.00002209, 0.00000737, -0.00000273, 0.00000952, 0.00001538, -0.00000392, -0.00002836, 0.00003900, -0.00002098, 0.00000792, 0.00000584, -0.00003286, 0.00000181, 0.00000135, 0.00003278, -0.00003256, 0.00001946, -0.00002359, 0.00000889, -0.00000088, -0.00002080, 0.00002785, -0.00000338, -0.00002778, 0.00004953, -0.00001962, -0.00000307, -0.00003100, 0.00007164, -0.00008238, 0.00009285, -0.00010350, 0.00007701, -0.00008101, 0.00004063, -0.00013907, 0.00016172, -0.00024999, 0.00024325, -0.00053511, 0.00006053, 0.00140340, -0.00088888, 0.00021088, -0.00021939, 0.00009619, -0.00014394, 0.00013320, -0.00014211, 0.00018904, -0.00012986, 0.00015435, -0.00015989, 0.00007052, -0.00004198, 0.00002499, -0.00005194, 0.00000444, -0.00000446, 0.00003711, -0.00001715, -0.00000087, -0.00003923, 0.00003644, },
		{ -0.00000120, 0.00000317, 0.00000320, 0.00001823, 0.00000132, -0.00003326, 0.00003454, -0.00001257, -0.00000002, 0.00001241, -0.00003059, -0.00000416, 0.00000145, 0.00003173, -0.00002416, 0.00000932, -0.00001046, -0.00000694, 0.00001546, -0.00003492, 0.00003179, -0.00000110, -0.00002775, 0.00004235, -0.00000951, -0.00000719, -0.00003422, 0.00007084, -0.00007246, 0.00007161, -0.00006924, 0.00003511, -0.00003214, -0.00000227, -0.00009429, 0.00010304, -0.00016406, 0.00012047, -0.00030264, -0.00032755, 0.00164107, -0.00083233, 0.00011451, -0.00016445, 0.00006071, -0.00012308, 0.00011821, -0.00013754, 0.00016901, -0.00010873, 0.00012660, -0.00011321, 0.00002154, -0.00000041, -0.00000506, -0.00002573, -0.00001661, 0.00000986, 0.00002192, -0.00000268, -0.00000524, -0.00003670, 0.00002819, -0.00001264, },
		{ 0.00001075, -0.00000416, 0.00002113, 0.00000553, -0.00003458, 0.00002639, -0.00000278, -0.00000731, 0.00001642, -0.00002554, -0.00001095, 0.00000215, 0.00002842, -0.00001326, -0.00000278, 0.00000420, -0.00002313, 0.00003065, -0.00004598, 0.00003265, 0.00000128, -0.00002490, 0.00003155, 0.00000375, -0.00001512, -0.00003019, 0.00005805, -0.00004715, 0.00003306, -0.00001836, -0.00002119, 0.00002850, -0.00005606, -0.00003556, 0.00002661, -0.00005763, -0.00002464, -0.00005931, -0.00066865, 0.00176376, -0.00066544, -0.00004605, -0.00007074, -0.00000464, -0.00007671, 0.00007955, -0.00010743, 0.00012331, -0.00006292, 0.00007419, -0.00004723, -0.00003912, 0.00004720, -0.00003845, 0.00000296, -0.00003985, 0.00002680, 0.00000473, 0.00001188, -0.00001005, -0.00003126, 0.00001640, -0.00000037, -0.00001187, },
		{ -0.00001164, 0.00002360, 0.00000863, -0.00003250, 0.00001554, 0.00000708, -0.00001296, 0.00001764, -0.00001843, -0.00001752, 0.00000268, 0.00002375, -0.00000124, -0.00001508, 0.00001828, -0.00003738, 0.00004247, -0.00005237, 0.00002991, 0.00000370, -0.00001951, 0.00001798, 0.00001926, -0.00002627, -0.00001950, 0.00003483, -0.00000980, -0.00001707, 0.00004098, -0.00008198, 0.00009007, -0.00010945, 0.00002489, -0.00005283, 0.00005056, -0.00016684, 0.00016031, -0.00092523, 0.00174913, -0.00038851, -0.00025999, 0.00005133, -0.00009166, -0.00001133, 0.00002248, -0.00005675, 0.00005750, 0.00000113, 0.00000536, 0.00002833, -0.00010192, 0.00009272, -0.00006875, 0.00002915, -0.00006128, 0.00004278, -0.00001114, 0.00002409, -0.00001407, -0.00002377, 0.00000219, 0.00001348, -0.00002343, 0.00001891, },
		{ 0.00002499, 0.00001089, -0.00002771, 0.00000338, 0.00001564, -0.00001627, 0.00001625, -0.00001029, -0.00002275, 0.00000224, 0.00001870, 0.00001040, -0.00002578, 0.00002980, -0.00004767, 0.00004917, -0.00005305, 0.00002347, 0.00000609, -0.00001231, 0.00000310, 0.00003533, -0.00003895, -0.00000436, 0.00000492, 0.00003360, -0.00007037, 0.00009828, -0.00013566, 0.00014068, -0.00015053, 0.00007421, -0.00012012, 0.00014178, -0.00028228, 0.00032669, -0.00107091, 0.00159098, -0.00001753, -0.00050486, 0.00018340, -0.00018618, 0.00006148, -0.00004317, 0.00000508, -0.00001854, 0.00007285, -0.00006813, 0.00010130, -0.00015611, 0.00012763, -0.00008960, 0.00004800, -0.00007694, 0.00005427, -0.00002258, 0.00003188, -0.00001622, -0.00001525, -0.00001298, 0.00002728, -0.00003434, 0.00002627, -0.00001809, },
		{ 0.00001277, -0.00002120, -0.00000854, 0.00002172, -0.00001685, 0.00001278, -0.00000227, -0.00002564, 0.00000024, 0.00001413, 0.00002034, -0.00003335, 0.00003715, -0.00005254, 0.00004971, -0.00004773, 0.00001368, 0.00000843, -0.00000432, -0.00001123, 0.00004981, -0.00005072, 0.00001182, -0.00002639, 0.00007556, -0.00011727, 0.00014269, -0.00017126, 0.00016983, -0.00016920, 0.00010170, -0.00016265, 0.00020090, -0.00035274, 0.00041988, -0.00109444, 0.00130112, 0.00041636, -0.00074875, 0.00030153, -0.00027019, 0.00012707, -0.00010470, 0.00006589, -0.00009239, 0.00013952, -0.00013312, 0.00015933, -0.00019174, 0.00014473, -0.00009598, 0.00005584, -0.00008381, 0.00005851, -0.00002732, 0.00003391, -0.00001585, -0.00000672, -0.00002751, 0.00003928, -0.00004296, 0.00003139, -0.00002236, 0.00002462, },
		{ -0.00001408, -0.00001878, 0.00002443, -0.00001476, 0.00000802, 0.00000449, -0.00002550, -0.00000358, 0.00001059, 0.00002762, -0.00003678, 0.00003945, -0.00005137, 0.00004394, -0.00003691, 0.00000136, 0.00001066, 0.00000323, -0.00002298, 0.00006040, -0.00005886, 0.00002517, -0.00005339, 0.00010850, -0.00014893, 0.00016515, -0.00018064, 0.00017064, -0.00015941, 0.00010104, -0.00017290, 0.00021918, -0.00036870, 0.00043243, -0.00100115, 0.00090836, 0.00087019, -0.00095436, 0.00038005, -0.00032512, 0.00017047, -0.00014909, 0.00011317, -0.00015147, 0.00018870, -0.00017763, 0.00019224, -0.00020167, 0.00013964, -0.00008536, 0.00005099, -0.00008042, 0.00005411, -0.00002433, 0.00002989, -0.00001283, 0.00000097, -0.00003982, 0.00004781, -0.00004779, 0.00003301, -0.00002347, 0.00002191, 0.00001479, },
		{ -0.00002616, 0.00002342, -0.00001048, 0.00000293, 0.00000907, -0.00002211, -0.00000903, 0.00000823, 0.00003178, -0.00003574, 0.00003657, -0.00004441, 0.00003262, -0.00002183, -0.00001233, 0.00001270, 0.00000914, -0.00003039, 0.00006517, -0.00006099, 0.00003220, -0.00007108, 0.00012629, -0.00015897, 0.00016028, -0.00016030, 0.00014138, -0.00012057, 0.00007172, -0.00014992, 0.00019579, -0.00033064, 0.00036999, -0.00081188, 0.00045484, 0.00129458, -0.00108455, 0.00039628, -0.00033560, 0.00017958, -0.00016582, 0.00013668, -0.00018562, 0.00021080, -0.00019318, 0.00019402, -0.00018302, 0.00011185, -0.00005839, 0.00003421, -0.00006728, 0.00004133, -0.00001401, 0.00002057, -0.00000765, 0.00000722, -0.00004860, 0.00005155, -0.00004774, 0.00003030, -0.00002085, 0.00001656, 0.00001736, -0.00000741, },
		{ 0.00001886, -0.00000480, -0.00000153, 0.00001093, -0.00001575, -0.00001548, 0.00000682, 0.00003290, -0.00003055, 0.00002920, -0.00003278, 0.00001726, -0.00000428, -0.00002597, 0.00001437, 0.00001249, -0.00003226, 0.00006297, -0.00005565, 0.00003060, -0.00007619, 0.00012546, -0.00014480, 0.00012749, -0.00011224, 0.00008605, -0.00005785, 0.00001922, -0.00009935, 0.00013768, -0.00024853, 0.00024965, -0.00055928, -0.00000973, 0.00164056, -0.00110853, 0.00033512, -0.00029302, 0.00014802, -0.00014929, 0.00013073, -0.00018919, 0.00020103, -0.00017653, 0.00016406, -0.00013791, 0.00006496, -0.00001885, 0.00000866, -0.00004670, 0.00002200, 0.00000188, 0.00000770, -0.00000126, 0.00001172, -0.00005294, 0.00004971, -0.00004230, 0.00002300, -0.00001445, 0.00000869, 0.00002067, -0.00000193, -0.00003001, },
		{ 0.00000122, -0.00000458, 0.00000995, -0.00000723, -0.00002194, 0.00000578, 0.00003150, -0.00002216, 0.00001867, -0.00001818, -0.00000007, 0.00001362, -0.00003807, 0.00001544, 0.00001280, -0.00002829, 0.00005380, -0.00004272, 0.00001979, -0.00006791, 0.00010593, -0.00010817, 0.00007130, -0.00004372, 0.00001375, 0.00001867, -0.00004611, -0.00003222, 0.00005795, -0.00013948, 0.00009642, -0.00028242, -0.00043522, 0.00186671, -0.00100738, 0.00019294, -0.00019805, 0.00007700, -0.00010043, 0.00009554, -0.00016233, 0.00016047, -0.00013040, 0.00010743, -0.00007324, 0.00000619, 0.00002687, -0.00002070, -0.00002245, -0.00000084, 0.00002058, -0.00000638, 0.00000509, 0.00001445, -0.00005244, 0.00004222, -0.00003175, 0.00001161, -0.00000488, -0.00000113, 0.00002453, 0.00000197, -0.00003023, 0.00001144, },
		{ -0.00000565, 0.00000647, 0.00000230, -0.00002724, 0.00000440, 0.00002840, -0.00001186, 0.00000671, -0.00000270, -0.00001711, 0.00002974, -0.00004726, 0.00001558, 0.00001013, -0.00001913, 0.00003889, -0.00002357, 0.00000109, -0.00004806, 0.00007103, -0.00005490, 0.00000054, 0.00003393, -0.00006295, 0.00009593, -0.00011106, 0.00003744, -0.00002692, -0.00002421, -0.00006157, -0.00002055, -0.00077809, 0.00194533, -0.00077809, -0.00002055, -0.00006157, -0.00002421, -0.00002692, 0.00003744, -0.00011106, 0.00009593, -0.00006295, 0.00003393, 0.00000054, -0.00005490, 0.00007103, -0.00004806, 0.00000109, -0.00002357, 0.00003889, -0.00001913, 0.00001013, 0.00001558, -0.00004726, 0.00002974, -0.00001711, -0.00000270, 0.00000671, -0.00001186, 0.00002840, 0.00000440, -0.00002724, 0.00000230, 0.00000647, },
		{ 0.00000122, 0.00001144, -0.00003023, 0.00000197, 0.00002453, -0.00000113, -0.00000488, 0.00001161, -0.00003175, 0.00004222, -0.00005244, 0.00001445, 0.00000509, -0.00000638, 0.00002058, -0.00000084, -0.00002245, -0.00002070, 0.00002687, 0.00000619, -0.00007324, 0.00010743, -0.00013040, 0.00016047, -0.00016233, 0.00009554, -0.00010043, 0.00007700, -0.00019805, 0.00019294, -0.00100738, 0.00186671, -0.00043522, -0.00028242, 0.00009642, -0.00013948, 0.00005795, -0.00003222, -0.00004611, 0.00001867, 0.00001375, -0.00004372, 0.00007130, -0.00010817, 0.00010593, -0.00006791, 0.00001979, -0.00004272, 0.00005380, -0.00002829, 0.00001280, 0.00001544, -0.00003807, 0.00001362, -0.00000007, -0.00001818, 0.00001867, -0.00002216, 0.00003150, 0.00000578, -0.00002194, -0.00000723, 0.00000995, -0.00000458, },
		{ 0.00001886, -0.00003001, -0.00000193, 0.00002067, 0.00000869, -0.00001445, 0.00002300, -0.00004230, 0.00004971, -0.00005294, 0.00001172, -0.00000126, 0.00000770, 0.00000188, 0.00002200, -0.00004670, 0.00000866, -0.00001885, 0.00006496, -0.00013791, 0.00016406, -0.00017653, 0.00020103, -0.00018919, 0.00013073, -0.00014929, 0.00014802, -0.00029302, 0.00033512, -0.00110853, 0.00164056, -0.00000973, -0.00055928, 0.00024965, -0.00024853, 0.00013768, -0.00009935, 0.00001922, -0.00005785, 0.00008605, -0.00011224, 0.00012749, -0.00014480, 0.00012546, -0.00007619, 0.00003060, -0.00005565, 0.00006297, -0.00003226, 0.00001249, 0.00001437, -0.00002597, -0.00000428, 0.00001726, -0.00003278, 0.00002920, -0.00003055, 0.00003290, 0.00000682, -0.00001548, -0.00001575, 0.00001093, -0.00000153, -0.00000480, },
		{ -0.00002616, -0.00000741, 0.00001736, 0.00001656, -0.00002085, 0.00003030, -0.00004774, 0.00005155, -0.00004860, 0.00000722, -0.00000765, 0.00002057, -0.00001401, 0.00004133, -0.00006728, 0.00003421, -0.00005839, 0.00011185, -0.00018302, 0.00019402, -0.00019318, 0.00021080, -0.00018562, 0.00013668, -0.00016582, 0.00017958, -0.00033560, 0.00039628, -0.00108455, 0.00129458, 0.00045484, -0.00081188, 0.00036999, -0.00033064, 0.00019579, -0.00014992, 0.00007172, -0.00012057, 0.00014138, -0.00016030, 0.00016028, -0.00015897, 0.00012629, -0.00007108, 0.00003220, -0.00006099, 0.00006517, -0.00003039, 0.00000914, 0.00001270, -0.00001233, -0.00002183, 0.00003262, -0.00004441, 0.00003657, -0.00003574, 0.00003178, 0.00000823, -0.00000903, -0.00002211, 0.00000907, 0.00000293, -0.00001048, 0.00002342, },
		{ -0.00001408, 0.00001479, 0.00002191, -0.00002347, 0.00003301, -0.00004779, 0.00004781, -0.00003982, 0.00000097, -0.00001283, 0.00002989, -0.00002433, 0.00005411, -0.00008042, 0.00005099, -0.00008536, 0.00013964, -0.00020167, 0.00019224, -0.00017763, 0.00018870, -0.00015147, 0.00011317, -0.00014909, 0.00017047, -0.00032512, 0.00038005, -0.00095436, 0.00087019, 0.00090836, -0.00100115, 0.00043243, -0.00036870, 0.00021918, -0.00017290, 0.00010104, -0.00015941, 0.00017064, -0.00018064, 0.00016515, -0.00014893, 0.00010850, -0.00005339, 0.00002517, -0.00005886, 0.00006040, -0.00002298, 0.00000323, 0.00001066, 0.00000136, -0.00003691, 0.00004394, -0.00005137, 0.00003945, -0.00003678, 0.00002762, 0.00001059, -0.00000358, -0.00002550, 0.00000449, 0.00000802, -0.00001476, 0.00002443, -0.00001878, },
		{ 0.00001277, 0.00002462, -0.00002236, 0.00003139, -0.00004296, 0.00003928, -0.00002751, -0.00000672, -0.00001585, 0.00003391, -0.00002732, 0.00005851, -0.00008381, 0.00005584, -0.00009598, 0.00014473, -0.00019174, 0.00015933, -0.00013312, 0.00013952, -0.00009239, 0.00006589, -0.00010470, 0.00012707, -0.00027019, 0.00030153, -0.00074875, 0.00041636, 0.00130112, -0.00109444, 0.00041988, -0.00035274, 0.00020090, -0.00016265, 0.00010170, -0.00016920, 0.00016983, -0.00017126, 0.00014269, -0.00011727, 0.00007556, -0.00002639, 0.00001182, -0.00005072, 0.00004981, -0.00001123, -0.00000432, 0.00000843, 0.00001368, -0.00004773, 0.00004971, -0.00005254, 0.00003715, -0.00003335, 0.00002034, 0.00001413, 0.00000024, -0.00002564, -0.00000227, 0.00001278, -0.00001685, 0.00002172, -0.00000854, -0.00002120, },
		{ 0.00002499, -0.00001809, 0.00002627, -0.00003434, 0.00002728, -0.00001298, -0.00001525, -0.00001622, 0.00003188, -0.00002258, 0.00005427, -0.00007694, 0.00004800, -0.00008960, 0.00012763, -0.00015611, 0.00010130, -0.00006813, 0.00007285, -0.00001854, 0.00000508, -0.00004317, 0.00006148, -0.00018618, 0.00018340, -0.00050486, -0.00001753, 0.00159098, -0.00107091, 0.00032669, -0.00028228, 0.00014178, -0.00012012, 0.00007421, -0.00015053, 0.00014068, -0.00013566, 0.00009828, -0.00007037, 0.00003360, 0.00000492, -0.00000436, -0.00003895, 0.00003533, 0.00000310, -0.00001231, 0.00000609, 0.00002347, -0.00005305, 0.00004917, -0.00004767, 0.00002980, -0.00002578, 0.00001040, 0.00001870, 0.00000224, -0.00002275, -0.00001029, 0.00001625, -0.00001627, 0.00001564, 0.00000338, -0.00002771, 0.00001089, },
		{ -0.00001164, 0.00001891, -0.00002343, 0.00001348, 0.00000219, -0.00002377, -0.00001407, 0.00002409, -0.00001114, 0.00004278, -0.00006128, 0.00002915, -0.00006875, 0.00009272, -0.00010192, 0.00002833, 0.00000536, 0.00000113, 0.00005750, -0.00005675, 0.00002248, -0.00001133, -0.00009166, 0.00005133, -0.00025999, -0.00038851, 0.00174913, -0.00092523, 0.00016031, -0.00016684, 0.00005056, -0.00005283, 0.00002489, -0.00010945, 0.00009007, -0.00008198, 0.00004098, -0.00001707, -0.00000980, 0.00003483, -0.00001950, -0.00002627, 0.00001926, 0.00001798, -0.00001951, 0.00000370, 0.00002991, -0.00005237, 0.00004247, -0.00003738, 0.00001828, -0.00001508, -0.00000124, 0.00002375, 0.00000268, -0.00001752, -0.00001843, 0.00001764, -0.00001296, 0.00000708, 0.00001554, -0.00003250, 0.00000863, 0.00002360, },
		{ 0.00001075, -0.00001187, -0.00000037, 0.00001640, -0.00003126, -0.00001005, 0.00001188, 0.00000473, 0.00002680, -0.00003985, 0.00000296, -0.00003845, 0.00004720, -0.00003912, -0.00004723, 0.00007419, -0.00006292, 0.00012331, -0.00010743, 0.00007955, -0.00007671, -0.00000464, -0.00007074, -0.00004605, -0.00066544, 0.00176376, -0.00066865, -0.00005931, -0.00002464, -0.00005763, 0.00002661, -0.00003556, -0.00005606, 0.00002850, -0.00002119, -0.00001836, 0.00003306, -0.00004715, 0.00005805, -0.00003019, -0.00001512, 0.00000375, 0.00003155, -0.00002490, 0.00000128, 0.00003265, -0.00004598, 0.00003065, -0.00002313, 0.00000420, -0.00000278, -0.00001326, 0.00002842, 0.00000215, -0.00001095, -0.00002554, 0.00001642, -0.00000731, -0.00000278, 0.00002639, -0.00003458, 0.00000553, 0.00002113, -0.00000416, },
		{ -0.00000120, -0.00001264, 0.00002819, -0.00003670, -0.00000524, -0.00000268, 0.00002192, 0.00000986, -0.00001661, -0.00002573, -0.00000506, -0.00000041, 0.00002154, -0.00011321, 0.00012660, -0.00010873, 0.00016901, -0.00013754, 0.00011821, -0.00012308, 0.00006071, -0.00016445, 0.00011451, -0.00083233, 0.00164107, -0.00032755, -0.00030264, 0.00012047, -0.00016406, 0.00010304, -0.00009429, -0.00000227, -0.00003214, 0.00003511, -0.00006924, 0.00007161, -0.00007246, 0.00007084, -0.00003422, -0.00000719, -0.00000951, 0.00004235, -0.00002775, -0.00000110, 0.00003179, -0.00003492, 0.00001546, -0.00000694, -0.00001046, 0.00000932, -0.00002416, 0.00003173, 0.00000145, -0.00000416, -0.00003059, 0.00001241, -0.00000002, -0.00001257, 0.00003454, -0.00003326, 0.00000132, 0.00001823, 0.00000320, 0.00000317, },
		{ -0.00002209, 0.00003644, -0.00003923, -0.00000087, -0.00001715, 0.00003711, -0.00000446, 0.00000444, -0.00005194, 0.00002499, -0.00004198, 0.00007052, -0.00015989, 0.00015435, -0.00012986, 0.00018904, -0.00014211, 0.00013320, -0.00014394, 0.00009619, -0.00021939, 0.00021088, -0.00088888, 0.00140340, 0.00006053, -0.00053511, 0.00024325, -0.00024999, 0.00016172, -0.00013907, 0.00004063, -0.00008101, 0.00007701, -0.00010350, 0.00009285, -0.00008238, 0.00007164, -0.00003100, -0.00000307, -0.00001962, 0.00004953, -0.00002778, -0.00000338, 0.00002785, -0.00002080, -0.00000088, 0.00000889, -0.00002359, 0.00001946, -0.00003256, 0.00003278, 0.00000135, 0.00000181, -0.00003286, 0.00000584, 0.00000792, -0.00002098, 0.00003900, -0.00002836, -0.00000392, 0.00001538, 0.00000952, -0.00000273, 0.00000737, },
		{ 0.00004049, -0.00003829, 0.00000193, -0.00002927, 0.00004743, -0.00001324, 0.00002007, -0.00007156, 0.00004646, -0.00007120, 0.00010114, -0.00018152, 0.00015410, -0.00012500, 0.00018295, -0.00012127, 0.00012454, -0.00013848, 0.00010035, -0.00023366, 0.00024372, -0.00084877, 0.00108511, 0.00045248, -0.00072264, 0.00032159, -0.00030028, 0.00019136, -0.00016073, 0.00006432, -0.00011035, 0.00009804, -0.00011671, 0.00009471, -0.00007670, 0.00006133, -0.00002160, -0.00000229, -0.00002641, 0.00005286, -0.00002509, -0.00000551, 0.00002168, -0.00000558, -0.00001605, 0.00002220, -0.00003338, 0.00002620, -0.00003740, 0.00003095, 0.00000242, 0.00000620, -0.00003203, -0.00000269, 0.00001544, -0.00002695, 0.00003927, -0.00002021, -0.00000982, 0.00001282, 0.00001413, -0.00000625, 0.00001310, -0.00002795, },
		{ -0.00003381, 0.00000240, -0.00003732, 0.00005096, -0.00001483, 0.00002841, -0.00008210, 0.00005622, -0.00008465, 0.00011056, -0.00017719, 0.00012775, -0.00009793, 0.00015506, -0.00007984, 0.00009699, -0.00011107, 0.00007778, -0.00021290, 0.00022324, -0.00073580, 0.00072689, 0.00080595, -0.00083710, 0.00034056, -0.00030631, 0.00018631, -0.00015504, 0.00006496, -0.00011685, 0.00009628, -0.00010883, 0.00007896, -0.00005821, 0.00004291, -0.00000845, -0.00000353, -0.00003037, 0.00005264, -0.00002011, -0.00000742, 0.00001435, 0.00000878, -0.00002798, 0.00003126, -0.00003853, 0.00002867, -0.00003817, 0.00002605, 0.00000492, 0.00000859, -0.00002824, -0.00001230, 0.00002149, -0.00002976, 0.00003542, -0.00000963, -0.00001574, 0.00001058, 0.00001674, -0.00000714, 0.00001575, -0.00003000, 0.00004025, },
		{ 0.00000033, -0.00004043, 0.00004708, -0.00000914, 0.00002931, -0.00008304, 0.00005362, -0.00008216, 0.00009998, -0.00015055, 0.00008175, -0.00005638, 0.00011316, -0.00002597, 0.00005867, -0.00006972, 0.00003734, -0.00016798, 0.00016583, -0.00057903, 0.00036959, 0.00108534, -0.00086071, 0.00029511, -0.00026763, 0.00014760, -0.00012338, 0.00004373, -0.00010198, 0.00007433, -0.00008391, 0.00005061, -0.00003197, 0.00002079, 0.00000528, -0.00000508, -0.00003231, 0.00004952, -0.00001343, -0.00000912, 0.00000695, 0.00002051, -0.00003516, 0.00003508, -0.00003855, 0.00002671, -0.00003496, 0.00001838, 0.00000869, 0.00000901, -0.00002203, -0.00002194, 0.00002521, -0.00002919, 0.00002807, 0.00000223, -0.00002088, 0.00000854, 0.00001738, -0.00000560, 0.00001557, -0.00002856, 0.00003614, -0.00002616, },
		{ -0.00003872, 0.00003658, 0.00000239, 0.00002423, -0.00007573, 0.00004040, -0.00006646, 0.00007398, -0.00010873, 0.00002552, -0.00001018, 0.00006663, 0.00003082, 0.00001900, -0.00002396, -0.00001025, -0.00011201, 0.00009019, -0.00040767, 0.00004861, 0.00126635, -0.00078862, 0.00019105, -0.00019196, 0.00008263, -0.00007234, 0.00000633, -0.00007136, 0.00003855, -0.00004896, 0.00001665, -0.00000411, -0.00000025, 0.00001646, -0.00000533, -0.00003307, 0.00004424, -0.00000573, -0.00001063, 0.00000049, 0.00002834, -0.00003685, 0.00003354, -0.00003373, 0.00002088, -0.00002847, 0.00000872, 0.00001321, 0.00000785, -0.00001427, -0.00003049, 0.00002603, -0.00002542, 0.00001823, 0.00001400, -0.00002447, 0.00000651, 0.00001629, -0.00000214, 0.00001321, -0.00002437, 0.00002906, -0.00001624, -0.00000392, },
		{ 0.00002137, 0.00001712, 0.00001580, -0.00006290, 0.00002001, -0.00004235, 0.00003928, -0.00006069, -0.00003060, 0.00003088, 0.00002431, 0.00008164, -0.00001330, 0.00001737, -0.00005483, -0.00005754, 0.00001371, -0.00024674, -0.00021026, 0.00133858, -0.00062936, 0.00004393, -0.00009369, 0.00000356, -0.00001214, -0.00003849, -0.00003324, -0.00000262, -0.00001232, -0.00001547, 0.00001949, -0.00001608, 0.00002269, -0.00000325, -0.00003322, 0.00003748, 0.00000240, -0.00001202, -0.00000429, 0.00003164, -0.00003320, 0.00002735, -0.00002512, 0.00001233, -0.00001986, -0.00000182, 0.00001770, 0.00000579, -0.00000597, -0.00003701, 0.00002372, -0.00001908, 0.00000717, 0.00002440, -0.00002591, 0.00000434, 0.00001391, 0.00000247, 0.00000956, -0.00001846, 0.00002023, -0.00000524, -0.00000946, -0.00003320, },
		{ 0.00003187, 0.00000711, -0.00004789, -0.00000330, -0.00001548, 0.00000323, -0.00001529, -0.00007739, 0.00005902, -0.00000727, 0.00011992, -0.00003193, 0.00004788, -0.00008875, -0.00001418, -0.00005031, -0.00011410, -0.00039300, 0.00130571, -0.00040298, -0.00012367, 0.00000900, -0.00007526, 0.00004561, -0.00008099, 0.00000337, -0.00004045, 0.00001810, -0.00003941, 0.00003445, -0.00002407, 0.00002286, 0.00000136, -0.00003291, 0.00002972, 0.00001051, -0.00001335, -0.00000694, 0.00003047, -0.00002511, 0.00001794, -0.00001429, 0.00000260, -0.00001056, -0.00001199, 0.00002128, 0.00000359, 0.00000183, -0.00004075, 0.00001844, -0.00001106, -0.00000379, 0.00003234, -0.00002489, 0.00000198, 0.00001070, 0.00000740, 0.00000558, -0.00001196, 0.00001099, 0.00000543, -0.00001506, -0.00002549, 0.00000412, },
		{ 0.00000103, -0.00003381, -0.00002543, 0.00000886, -0.00002772, 0.00002032, -0.00010840, 0.00006995, -0.00002499, 0.00014238, -0.00003394, 0.00006458, -0.00010803, 0.00001268, -0.00009414, -0.00001915, -0.00049748, 0.00118358, -0.00013751, -0.00028616, 0.00009762, -0.00014016, 0.00009030, -0.00011252, 0.00003067, -0.00006779, 0.00003646, -0.00005119, 0.00003866, -0.00002356, 0.00001733, 0.00000776, -0.00003185, 0.00002125, 0.00001823, -0.00001465, -0.00000741, 0.00002551, -0.00001413, 0.00000712, -0.00000306, -0.00000668, -0.00000198, -0.00002059, 0.00002312, 0.00000198, 0.00000827, -0.00004134, 0.00001070, -0.00000241, -0.00001348, 0.00003712, -0.00002146, -0.00000050, 0.00000714, 0.00001184, 0.00000216, -0.00000593, 0.00000261, 0.00001446, -0.00001943, -0.00001743, -0.00001237, 0.00004355, },
		{ -0.00002289, -0.00004333, 0.00002677, -0.00004912, 0.00004190, -0.00012095, 0.00006336, -0.00002941, 0.00014915, -0.00001987, 0.00006809, -0.00011249, 0.00002206, -0.00011577, 0.00003674, -0.00053165, 0.00099653, 0.00013568, -0.00041925, 0.00015684, -0.00018086, 0.00011452, -0.00012728, 0.00004361, -0.00008036, 0.00003996, -0.00004985, 0.00003262, -0.00001586, 0.00000780, 0.00001453, -0.00002947, 0.00001220, 0.00002533, -0.00001588, -0.00000599, 0.00001794, -0.00000212, -0.00000316, 0.00000686, -0.00001408, 0.00000467, -0.00002666, 0.00002267, 0.00000146, 0.00001279, -0.00003877, 0.00000127, 0.00000580, -0.00002100, 0.00003849, -0.00001600, -0.00000289, 0.00000363, 0.00001514, -0.00000000, -0.00000121, -0.00000387, 0.00002078, -0.00002152, -0.00001068, -0.00002569, 0.00004983, -0.00000056, },
		{ -0.00005541, 0.00003627, -0.00005912, 0.00004850, -0.00011611, 0.00004258, -0.00002409, 0.00014310, 0.00000692, 0.00006189, -0.00010502, 0.00001664, -0.00011788, 0.00005859, -0.00051046, 0.00077293, 0.00038676, -0.00050394, 0.00017725, -0.00019229, 0.00011528, -0.00012336, 0.00004067, -0.00007735, 0.00002919, -0.00003737, 0.00001910, -0.00000381, -0.00000317, 0.00001990, -0.00002514, 0.00000270, 0.00003157, -0.00001688, -0.00000327, 0.00000921, 0.00000901, -0.00001124, 0.00001411, -0.00001857, 0.00000862, -0.00002968, 0.00001972, 0.00000225, 0.00001518, -0.00003344, -0.00000887, 0.00001263, -0.00002580, 0.00003660, -0.00000914, -0.00000493, 0.00000050, 0.00001683, -0.00000048, 0.00000167, -0.00000782, 0.00002379, -0.00002072, -0.00000636, -0.00003428, 0.00004963, 0.00000295, -0.00001601, },
		{ 0.00003741, -0.00005841, 0.00004217, -0.00009806, 0.00001347, -0.00001433, 0.00012859, 0.00004144, 0.00005116, -0.00009037, 0.00000136, -0.00010617, 0.00005528, -0.00045218, 0.00054063, 0.00059154, -0.00052930, 0.00015682, -0.00017533, 0.00009442, -0.00010283, 0.00002388, -0.00006127, 0.00000767, -0.00001803, 0.00000233, 0.00000897, -0.00001279, 0.00002224, -0.00001847, -0.00000697, 0.00003666, -0.00001739, -0.00000002, 0.00000082, 0.00001764, -0.00001593, 0.00001795, -0.00001972, 0.00000964, -0.00002960, 0.00001445, 0.00000420, 0.00001556, -0.00002601, -0.00001869, 0.00001734, -0.00002771, 0.00003199, -0.00000167, -0.00000638, -0.00000197, 0.00001668, 0.00000082, 0.00000257, -0.00000905, 0.00002344, -0.00001703, -0.00000486, -0.00003769, 0.00004321, 0.00001071, -0.00001266, -0.00006167, },
		{ -0.00004959, 0.00002701, -0.00007275, -0.00001718, -0.00000559, 0.00011012, 0.00007833, 0.00004139, -0.00007365, -0.00001821, -0.00008735, 0.00003699, -0.00037497, 0.00032324, 0.00073439, -0.00049375, 0.00010097, -0.00013624, 0.00005784, -0.00007101, -0.00000196, -0.00003705, -0.00001911, 0.00000268, -0.00001302, 0.00001899, -0.00001871, 0.00002049, -0.00000955, -0.00001633, 0.00004025, -0.00001709, 0.00000295, -0.00000592, 0.00002268, -0.00001673, 0.00001825, -0.00001766, 0.00000798, -0.00002681, 0.00000745, 0.00000686, 0.00001439, -0.00001740, -0.00002721, 0.00001948, -0.00002693, 0.00002542, 0.00000561, -0.00000706, -0.00000359, 0.00001471, 0.00000369, 0.00000168, -0.00000783, 0.00002023, -0.00001104, -0.00000577, -0.00003653, 0.00003211, 0.00002077, -0.00001131, -0.00006337, 0.00003182, },
		{ 0.00000798, -0.00004638, -0.00004323, -0.00000206, 0.00009106, 0.00011307, 0.00003720, -0.00005908, -0.00003727, -0.00006750, 0.00001292, -0.00029415, 0.00013758, 0.00080932, -0.00040457, 0.00002113, -0.00008508, 0.00001392, -0.00003499, -0.00003066, -0.00001071, -0.00004512, 0.00001940, -0.00002298, 0.00002373, -0.00001960, 0.00001450, 0.00000098, -0.00002466, 0.00004194, -0.00001563, 0.00000493, -0.00001009, 0.00002368, -0.00001384, 0.00001554, -0.00001308, 0.00000435, -0.00002207, -0.00000044, 0.00000960, 0.00001231, -0.00000856, -0.00003362, 0.00001890, -0.00002390, 0.00001774, 0.00001204, -0.00000694, -0.00000422, 0.00001114, 0.00000766, -0.00000047, -0.00000481, 0.00001510, -0.00000386, -0.00000809, -0.00003227, 0.00001868, 0.00003057, -0.00000992, -0.00006237, 0.00002202, -0.00003627, },
		{ -0.00002406, -0.00006053, -0.00000567, 0.00007303, 0.00014260, 0.00004157, -0.00004914, -0.00005272, -0.00005085, -0.00001007, -0.00022060, -0.00000721, 0.00081949, -0.00027599, -0.00006772, -0.00003333, -0.00002839, -0.00000199, -0.00005611, 0.00001202, -0.00006505, 0.00002804, -0.00002513, 0.00002218, -0.00001539, 0.00000504, 0.00001191, -0.00003106, 0.00004134, -0.00001276, 0.00000540, -0.00001124, 0.00002084, -0.00000810, 0.00001081, -0.00000696, -0.00000031, -0.00001636, -0.00000824, 0.00001167, 0.00001001, -0.00000040, -0.00003734, 0.00001575, -0.00001926, 0.00000978, 0.00001714, -0.00000614, -0.00000380, 0.00000639, 0.00001208, -0.00000319, -0.00000090, 0.00000928, 0.00000314, -0.00001044, -0.00002677, 0.00000553, 0.00003760, -0.00000663, -0.00006051, 0.00001061, -0.00002203, -0.00001037, },
		{ -0.00006757, -0.00001577, 0.00005579, 0.00016540, 0.00005565, -0.00004434, -0.00006345, -0.00003945, -0.00002814, -0.00016033, -0.00010922, 0.00077528, -0.00012624, -0.00014997, 0.00000849, -0.00006158, 0.00002227, -0.00007365, 0.00002696, -0.00007543, 0.00002658, -0.00001910, 0.00001502, -0.00000729, -0.00000629, 0.00002173, -0.00003465, 0.00003821, -0.00000837, 0.00000415, -0.00000944, 0.00001499, -0.00000075, 0.00000529, -0.00000044, -0.00000498, -0.00001067, -0.00001500, 0.00001243, 0.00000812, 0.00000635, -0.00003812, 0.00001043, -0.00001372, 0.00000224, 0.00002068, -0.00000486, -0.00000237, 0.00000102, 0.00001622, -0.00000570, 0.00000291, 0.00000402, 0.00000862, -0.00001151, -0.00002185, -0.00000505, 0.00004006, -0.00000038, -0.00005898, -0.00000044, -0.00000962, -0.00002484, -0.00000883, },
		{ -0.00002961, 0.00003782, 0.00018110, 0.00007906, -0.00004349, -0.00007013, -0.00003333, -0.00004011, -0.00011510, -0.00017222, 0.00069150, 0.00002576, -0.00021230, 0.00003296, -0.00008103, 0.00003463, -0.00008091, 0.00003217, -0.00007513, 0.00001539, -0.00000653, 0.00000436, 0.00000258, -0.00001743, 0.00002888, -0.00003477, 0.00003252, -0.00000261, 0.00000132, -0.00000525, 0.00000734, 0.00000676, 0.00000019, 0.00000544, -0.00000876, -0.00000587, -0.00001995, 0.00001140, 0.00000706, 0.00001124, -0.00003602, 0.00000356, -0.00000801, -0.00000435, 0.00002263, -0.00000337, -0.00000009, -0.00000431, 0.00001935, -0.00000730, 0.00000575, 0.00000043, 0.00001153, -0.00001032, -0.00001879, -0.00001165, 0.00003726, 0.00000880, -0.00005800, -0.00001016, -0.00000045, -0.00003400, -0.00000142, -0.00006546, },
		{ 0.00001733, 0.00018973, 0.00011045, -0.00004430, -0.00007449, -0.00003120, -0.00004676, -0.00008363, -0.00020350, 0.00058435, 0.00016337, -0.00024582, 0.00003701, -0.00008573, 0.00003490, -0.00007811, 0.00002807, -0.00006534, -0.00000314, 0.00000942, -0.00000679, 0.00001175, -0.00002637, 0.00003213, -0.00003117, 0.00002463, 0.00000411, -0.00000267, 0.00000047, -0.00000072, 0.00001309, -0.00000352, 0.00000991, -0.00001103, -0.00000255, -0.00002261, 0.00000843, 0.00000699, 0.00001412, -0.00003142, -0.00000409, -0.00000278, -0.00000965, 0.00002312, -0.00000189, 0.00000276, -0.00000891, 0.00002087, -0.00000746, 0.00000695, -0.00000077, 0.00001132, -0.00000652, -0.00001814, -0.00001391, 0.00002980, 0.00001970, -0.00005679, -0.00001866, 0.00000541, -0.00003818, -0.00000049, -0.00005728, -0.00004329, },
		{ 0.00019100, 0.00014811, -0.00004406, -0.00007861, -0.00003128, -0.00004979, -0.00006302, -0.00021169, 0.00046888, 0.00027447, -0.00024718, 0.00002217, -0.00007806, 0.00002554, -0.00006758, 0.00001701, -0.00004900, -0.00002522, 0.00002488, -0.00001537, 0.00001803, -0.00003159, 0.00003085, -0.00002413, 0.00001521, 0.00001117, -0.00000715, 0.00000666, -0.00000784, 0.00001721, -0.00000525, 0.00001257, -0.00001146, -0.00000101, -0.00002285, 0.00000365, 0.00000778, 0.00001517, -0.00002495, -0.00001169, 0.00000141, -0.00001348, 0.00002235, -0.00000057, 0.00000575, -0.00001215, 0.00002037, -0.00000593, 0.00000622, 0.00000064, 0.00000810, -0.00000048, -0.00001955, -0.00001259, 0.00001939, 0.00003029, -0.00005393, -0.00002673, 0.00000885, -0.00003888, -0.00000350, -0.00004688, -0.00005304, -0.00000671, },
		{ 0.00019040, -0.00004019, -0.00008424, -0.00003201, -0.00005096, -0.00005004, -0.00020501, 0.00035700, 0.00035247, -0.00021858, -0.00000622, -0.00006284, 0.00001066, -0.00005298, 0.00000242, -0.00002993, -0.00004666, 0.00003620, -0.00001896, 0.00002004, -0.00003242, 0.00002523, -0.00001454, 0.00000527, 0.00001775, -0.00001135, 0.00001225, -0.00001293, 0.00001853, -0.00000487, 0.00001340, -0.00001008, -0.00000123, -0.00002084, -0.00000248, 0.00000906, 0.00001480, -0.00001740, -0.00001844, 0.00000415, -0.00001580, 0.00002055, 0.00000060, 0.00000839, -0.00001352, 0.00001775, -0.00000276, 0.00000363, 0.00000439, 0.00000257, 0.00000681, -0.00002197, -0.00000925, 0.00000842, 0.00003819, -0.00004793, -0.00003536, 0.00001116, -0.00003804, -0.00000765, -0.00003776, -0.00005639, -0.00003375, 0.00018384, },
		{ -0.00003050, -0.00009249, -0.00003247, -0.00005150, -0.00004201, -0.00019010, 0.00025674, 0.00039627, -0.00016660, -0.00004048, -0.00004590, -0.00000517, -0.00003809, -0.00001218, -0.00001181, -0.00006384, 0.00004077, -0.00001638, 0.00001757, -0.00002915, 0.00001621, -0.00000375, -0.00000399, 0.00002302, -0.00001453, 0.00001637, -0.00001530, 0.00001698, -0.00000271, 0.00001271, -0.00000719, -0.00000291, -0.00001707, -0.00000927, 0.00001026, 0.00001359, -0.00000964, -0.00002362, 0.00000518, -0.00001664, 0.00001788, 0.00000172, 0.00001019, -0.00001274, 0.00001317, 0.00000169, -0.00000038, 0.00000977, -0.00000410, 0.00001391, -0.00002393, -0.00000583, -0.00000065, 0.00004138, -0.00003778, -0.00004516, 0.00001330, -0.00003723, -0.00001089, -0.00003205, -0.00005295, -0.00006133, 0.00016641, 0.00023578, },
		{ -0.00010372, -0.00003249, -0.00005197, -0.00003708, -0.00017164, 0.00017220, 0.00040918, -0.00010041, -0.00007251, -0.00003248, -0.00001822, -0.00002593, -0.00002411, 0.00000260, -0.00007439, 0.00003753, -0.00000800, 0.00001160, -0.00002301, 0.00000535, 0.00000665, -0.00001141, 0.00002627, -0.00001611, 0.00001842, -0.00001473, 0.00001299, 0.00000062, 0.00001098, -0.00000322, -0.00000560, -0.00001219, -0.00001592, 0.00001077, 0.00001215, -0.00000249, -0.00002670, 0.00000440, -0.00001613, 0.00001448, 0.00000298, 0.00001075, -0.00000979, 0.00000713, 0.00000679, -0.00000513, 0.00001576, -0.00001051, 0.00001930, -0.00002397, -0.00000409, -0.00000593, 0.00003879, -0.00002349, -0.00005586, 0.00001548, -0.00003706, -0.00001244, -0.00003000, -0.00004451, -0.00008546, 0.00013668, 0.00028243, -0.00001309, },
		{ -0.00003241, -0.00005238, -0.00003420, -0.00015248, 0.00010439, 0.00039733, -0.00002978, -0.00009556, -0.00002598, -0.00002645, -0.00001798, -0.00003199, 0.00001196, -0.00007759, 0.00002711, 0.00000446, 0.00000406, -0.00001577, -0.00000552, 0.00001507, -0.00001602, 0.00002704, -0.00001577, 0.00001818, -0.00001147, 0.00000734, 0.00000431, 0.00000876, 0.00000133, -0.00000877, -0.00000694, -0.00002159, 0.00001003, 0.00001104, 0.00000340, -0.00002736, 0.00000190, -0.00001444, 0.00001047, 0.00000456, 0.00000987, -0.00000500, 0.00000038, 0.00001177, -0.00000978, 0.00002126, -0.00001526, 0.00002169, -0.00002102, -0.00000519, -0.00000651, 0.00003059, -0.00000628, -0.00006623, 0.00001698, -0.00003708, -0.00001295, -0.00003013, -0.00003457, -0.00010150, 0.00009332, 0.00032780, 0.00001384, -0.00011778, },
		},
		{ { 0.00000497, -0.00000659, -0.00004258, -0.00002529, -0.00014615, 0.00005076, 0.00038108, 0.00005279, -0.00013669, -0.00002506, -0.00001143, -0.00002910, 0.00002583, -0.00001796, -0.00000959, -0.00001347, -0.00000006, 0.00001457, -0.00007782, 0.00002281, -0.00000251, 0.00001137, -0.00003254, -0.00000133, 0.00002066, -0.00000408, -0.00000453, -0.00000064, 0.00000267, -0.00000813, 0.00004536, -0.00001300, -0.00001311, -0.00001300, 0.00004536, -0.00000813, 0.00000267, -0.00000064, -0.00000453, -0.00000408, 0.00002066, -0.00000133, -0.00003254, 0.00001137, -0.00000251, 0.00002281, -0.00007782, 0.00001457, -0.00000006, -0.00001347, -0.00000959, -0.00001796, 0.00002583, -0.00002910, -0.00001143, -0.00002506, -0.00013669, 0.00005279, 0.00038108, 0.00005076, -0.00014615, -0.00002529, -0.00004258, -0.00000659, },
		{ -0.00000164, -0.00003840, -0.00002394, -0.00012011, -0.00001495, 0.00033489, 0.00013467, -0.00012562, -0.00004845, -0.00000249, -0.00003383, 0.00002122, -0.00000693, -0.00001895, -0.00000586, -0.00001200, 0.00002828, -0.00007379, 0.00000436, 0.00000843, 0.00000804, -0.00002328, -0.00001640, 0.00002679, -0.00000295, -0.00000351, -0.00000559, 0.00000889, -0.00001788, 0.00004727, -0.00000437, -0.00001396, -0.00001762, 0.00003835, 0.00000344, -0.00000345, 0.00000411, -0.00000632, -0.00000183, 0.00001026, 0.00001468, -0.00004020, 0.00001451, -0.00001240, 0.00003410, -0.00007055, -0.00000662, 0.00001407, -0.00002179, -0.00000043, -0.00002650, 0.00002670, -0.00002212, -0.00001933, -0.00001151, -0.00012291, -0.00003977, 0.00039837, 0.00014096, -0.00016972, -0.00003152, -0.00004369, -0.00001374, 0.00000503, },
		{ -0.00003258, -0.00002461, -0.00009609, -0.00005768, 0.00027255, 0.00019564, -0.00009236, -0.00007367, 0.00000138, -0.00003270, 0.00001143, 0.00000641, -0.00002765, -0.00000005, -0.00002036, 0.00003408, -0.00006099, -0.00001640, 0.00001628, 0.00000678, -0.00001425, -0.00002852, 0.00002778, 0.00000102, -0.00000264, -0.00001052, 0.00001409, -0.00002408, 0.00004308, 0.00000796, -0.00001609, -0.00001878, 0.00002783, 0.00001507, -0.00000864, 0.00000857, -0.00000905, 0.00000325, -0.00000216, 0.00002843, -0.00004341, 0.00001427, -0.00001690, 0.00003470, -0.00005167, -0.00003298, 0.00002812, -0.00002952, 0.00000784, -0.00003277, 0.00002553, -0.00001597, -0.00002217, -0.00001061, -0.00008973, -0.00012637, 0.00037397, 0.00025358, -0.00018433, -0.00004539, -0.00004114, -0.00002270, 0.00000476, 0.00000163, },
		{ -0.00002490, -0.00007679, -0.00008081, 0.00020452, 0.00023221, -0.00004519, -0.00009149, -0.00000438, -0.00002447, -0.00000275, 0.00002017, -0.00003384, 0.00000249, -0.00002410, 0.00003263, -0.00004288, -0.00003505, 0.00001859, 0.00000800, -0.00000589, -0.00003697, 0.00002403, 0.00000658, -0.00000132, -0.00001494, 0.00001702, -0.00002543, 0.00003280, 0.00002266, -0.00001835, -0.00001762, 0.00001568, 0.00002521, -0.00001225, 0.00001249, -0.00001221, 0.00000947, -0.00001334, 0.00003640, -0.00003953, 0.00000810, -0.00001344, 0.00002408, -0.00002414, -0.00005995, 0.00003863, -0.00003457, 0.00001372, -0.00003596, 0.00002246, -0.00001113, -0.00002034, -0.00001854, -0.00004988, -0.00018778, 0.00029958, 0.00038043, -0.00018084, -0.00006955, -0.00003506, -0.00003249, 0.00000331, 0.00000418, -0.00002675, },
		{ -0.00006285, -0.00008892, 0.00013870, 0.00024557, 0.00000622, -0.00009575, -0.00001980, -0.00001191, -0.00001749, 0.00003053, -0.00003463, 0.00000001, -0.00002255, 0.00002515, -0.00002274, -0.00004872, 0.00001519, 0.00001036, 0.00000236, -0.00004188, 0.00001664, 0.00001236, 0.00000073, -0.00001808, 0.00001662, -0.00002150, 0.00001767, 0.00003745, -0.00001904, -0.00001559, 0.00000374, 0.00003255, -0.00001379, 0.00001529, -0.00001463, 0.00001448, -0.00002001, 0.00003599, -0.00002747, -0.00000440, -0.00000254, 0.00000554, 0.00000585, -0.00008122, 0.00004138, -0.00003416, 0.00001474, -0.00003399, 0.00001568, -0.00000516, -0.00001820, -0.00002677, -0.00001885, -0.00020848, 0.00017665, 0.00050500, -0.00014746, -0.00010646, -0.00002585, -0.00004227, 0.00000023, 0.00000644, -0.00002184, -0.00002364, },
		{ -0.00008759, 0.00008116, 0.00023883, 0.00005458, -0.00008585, -0.00003969, -0.00000123, -0.00002710, 0.00003307, -0.00002744, -0.00000855, -0.00001587, 0.00001346, -0.00000337, -0.00005624, 0.00000765, 0.00001170, 0.00001117, -0.00004346, 0.00000675, 0.00001731, 0.00000338, -0.00001917, 0.00001250, -0.00001306, 0.00000020, 0.00004940, -0.00001640, -0.00001410, -0.00000637, 0.00003614, -0.00001278, 0.00001617, -0.00001486, 0.00001608, -0.00001998, 0.00002666, -0.00000870, -0.00002074, 0.00001194, -0.00001456, 0.00003043, -0.00009038, 0.00003274, -0.00002589, 0.00000845, -0.00002449, 0.00000280, 0.00000534, -0.00002115, -0.00002666, -0.00000848, -0.00018333, 0.00002013, 0.00060252, -0.00007127, -0.00015745, -0.00001423, -0.00005180, -0.00000385, 0.00000767, -0.00001734, -0.00002146, -0.00005282, },
		{ 0.00003644, 0.00021528, 0.00009635, -0.00006633, -0.00005590, 0.00000077, -0.00002685, 0.00002519, -0.00001179, -0.00002225, -0.00000586, 0.00000018, 0.00001280, -0.00005757, -0.00000174, 0.00001034, 0.00002033, -0.00004150, -0.00000478, 0.00002085, 0.00000621, -0.00001780, 0.00000522, -0.00000212, -0.00001626, 0.00005556, -0.00000911, -0.00001404, -0.00001356, 0.00003548, -0.00000894, 0.00001442, -0.00001184, 0.00001307, -0.00001296, 0.00001041, 0.00001278, -0.00003611, 0.00002413, -0.00002886, 0.00004257, -0.00008338, 0.00001162, -0.00000934, -0.00000586, -0.00000688, -0.00001681, 0.00002176, -0.00003177, -0.00001443, -0.00002142, -0.00012131, -0.00014195, 0.00064397, 0.00005754, -0.00022033, -0.00000222, -0.00006098, -0.00000766, 0.00000625, -0.00001166, -0.00002033, -0.00004384, -0.00008269, },
		{ 0.00017840, 0.00013075, -0.00004377, -0.00006129, -0.00000947, -0.00001596, 0.00000828, 0.00000924, -0.00003737, 0.00000381, -0.00001115, 0.00002354, -0.00005316, -0.00001116, 0.00000601, 0.00002837, -0.00003526, -0.00001734, 0.00002266, 0.00000888, -0.00001423, -0.00000370, 0.00000850, -0.00002829, 0.00005382, 0.00000307, -0.00001552, -0.00001744, 0.00003062, -0.00000239, 0.00000982, -0.00000539, 0.00000578, -0.00000083, -0.00000856, 0.00003145, -0.00004503, 0.00002841, -0.00003185, 0.00003887, -0.00006060, -0.00001904, 0.00001255, -0.00002566, 0.00001596, -0.00003998, 0.00004112, -0.00004708, 0.00000599, -0.00004921, -0.00004430, -0.00027420, 0.00060390, 0.00023964, -0.00028618, 0.00000523, -0.00006854, -0.00001049, 0.00000104, -0.00000348, -0.00002211, -0.00003322, -0.00007895, 0.00000701, },
		{ 0.00015722, -0.00002296, -0.00005361, -0.00002961, 0.00000086, -0.00001167, 0.00002915, -0.00004799, 0.00000830, -0.00001684, 0.00002704, -0.00004359, -0.00001975, -0.00000007, 0.00003302, -0.00002407, -0.00003014, 0.00002243, 0.00001137, -0.00000933, -0.00001216, 0.00001597, -0.00003338, 0.00004370, 0.00001887, -0.00001760, -0.00001842, 0.00002234, 0.00000612, 0.00000302, 0.00000345, -0.00000394, 0.00001296, -0.00002515, 0.00004210, -0.00004351, 0.00002188, -0.00002218, 0.00002124, -0.00002766, -0.00005216, 0.00003346, -0.00004510, 0.00003753, -0.00005965, 0.00005603, -0.00005880, 0.00002397, -0.00007536, 0.00001971, -0.00034415, 0.00047041, 0.00046041, -0.00033691, -0.00000093, -0.00007093, -0.00001367, -0.00000710, 0.00000664, -0.00002682, -0.00002028, -0.00007833, -0.00000782, 0.00013295, },
		{ -0.00000433, -0.00003706, -0.00005139, 0.00001457, -0.00002589, 0.00004008, -0.00004801, 0.00000343, -0.00001432, 0.00002267, -0.00002978, -0.00002735, -0.00000584, 0.00003214, -0.00000812, -0.00004182, 0.00001956, 0.00001399, -0.00000433, -0.00001818, 0.00001837, -0.00003079, 0.00002681, 0.00003553, -0.00001852, -0.00001762, 0.00001205, 0.00001520, -0.00000451, 0.00001266, -0.00001327, 0.00002446, -0.00003488, 0.00004163, -0.00003081, 0.00000579, -0.00000350, -0.00000334, 0.00000585, -0.00007822, 0.00004549, -0.00005665, 0.00004957, -0.00006709, 0.00005744, -0.00005681, 0.00002698, -0.00008245, 0.00004673, -0.00033352, 0.00025343, 0.00068714, -0.00034581, -0.00003287, -0.00006244, -0.00002093, -0.00001465, 0.00001540, -0.00003165, -0.00000739, -0.00007914, -0.00001246, 0.00008584, 0.00017345, },
		{ -0.00002051, -0.00006376, 0.00001553, -0.00002657, 0.00003625, -0.00003407, -0.00001212, -0.00000371, 0.00001181, -0.00001354, -0.00003384, -0.00000956, 0.00002495, 0.00001086, -0.00005022, 0.00001328, 0.00001714, -0.00000030, -0.00002059, 0.00001536, -0.00002193, 0.00000665, 0.00004930, -0.00001619, -0.00001642, 0.00000156, 0.00002299, -0.00001079, 0.00001980, -0.00001938, 0.00003052, -0.00003537, 0.00003022, -0.00001003, -0.00001465, 0.00001679, -0.00002517, 0.00002951, -0.00008859, 0.00004224, -0.00005424, 0.00004547, -0.00005574, 0.00003903, -0.00003443, 0.00000712, -0.00006041, 0.00002651, -0.00024657, -0.00001247, 0.00087212, -0.00028283, -0.00010175, -0.00003781, -0.00003683, -0.00001654, 0.00001774, -0.00003181, 0.00000076, -0.00007682, -0.00001408, 0.00004557, 0.00017519, 0.00001588, },
		{ -0.00005806, -0.00000164, -0.00001098, 0.00001720, -0.00000794, -0.00003518, 0.00001137, -0.00000177, 0.00000217, -0.00003845, -0.00001073, 0.00001265, 0.00002956, -0.00005279, 0.00000305, 0.00002087, 0.00000240, -0.00001953, 0.00000836, -0.00000999, -0.00001218, 0.00005646, -0.00000901, -0.00001598, -0.00000743, 0.00002774, -0.00001394, 0.00002285, -0.00002039, 0.00002976, -0.00002699, 0.00001124, 0.00001288, -0.00003214, 0.00003037, -0.00003550, 0.00003603, -0.00007912, 0.00002198, -0.00003643, 0.00002373, -0.00002504, 0.00000144, 0.00000694, -0.00003365, -0.00001228, -0.00003248, -0.00011119, -0.00027351, 0.00096281, -0.00012448, -0.00021086, 0.00000345, -0.00006345, -0.00000909, 0.00000945, -0.00002294, -0.00000043, -0.00006648, -0.00001997, 0.00001983, 0.00015842, 0.00004259, -0.00001310, },
		{ -0.00003437, 0.00001606, -0.00001074, 0.00002299, -0.00005828, 0.00002429, -0.00001258, 0.00001359, -0.00003969, -0.00001037, -0.00000168, 0.00004383, -0.00004742, -0.00001084, 0.00002452, 0.00000429, -0.00001630, 0.00000002, 0.00000110, -0.00002541, 0.00005462, 0.00000325, -0.00001660, -0.00001379, 0.00002826, -0.00001284, 0.00002091, -0.00001600, 0.00002287, -0.00001251, -0.00000987, 0.00003116, -0.00004017, 0.00003147, -0.00003016, 0.00002433, -0.00005250, -0.00001057, -0.00000798, -0.00001035, 0.00001787, -0.00004620, 0.00005642, -0.00008240, 0.00004550, -0.00010516, 0.00002842, -0.00047005, 0.00091723, 0.00013416, -0.00034841, 0.00005253, -0.00009686, 0.00000670, -0.00000957, -0.00000433, -0.00001229, -0.00004617, -0.00003417, 0.00001232, 0.00012266, 0.00007772, -0.00001920, -0.00003276, },
		{ 0.00004286, -0.00003602, 0.00004767, -0.00007188, 0.00002761, -0.00001518, 0.00001740, -0.00003602, -0.00001041, -0.00001418, 0.00004992, -0.00003344, -0.00002679, 0.00002660, 0.00000658, -0.00001286, -0.00000678, 0.00000798, -0.00003042, 0.00004365, 0.00001901, -0.00001748, -0.00001730, 0.00002440, -0.00000758, 0.00001453, -0.00000753, 0.00001227, 0.00000394, -0.00002756, 0.00003970, -0.00003561, 0.00001934, -0.00001139, 0.00000045, -0.00001797, -0.00004512, 0.00002111, -0.00004578, 0.00005971, -0.00008837, 0.00009657, -0.00011925, 0.00008890, -0.00015905, 0.00012592, -0.00055443, 0.00072000, 0.00046978, -0.00048339, 0.00008948, -0.00012548, 0.00002332, -0.00003337, 0.00001906, -0.00003079, -0.00001918, -0.00005503, 0.00002036, 0.00007341, 0.00011713, -0.00003524, 0.00000444, -0.00007165, },
		{ -0.00004618, 0.00005560, -0.00006808, 0.00001617, -0.00000653, 0.00001224, -0.00002692, -0.00001253, -0.00002161, 0.00004595, -0.00001234, -0.00004192, 0.00002520, 0.00001054, -0.00001085, -0.00001012, 0.00000909, -0.00002722, 0.00002604, 0.00003496, -0.00001692, -0.00001860, 0.00001706, 0.00000054, 0.00000555, 0.00000254, 0.00000106, 0.00001835, -0.00003791, 0.00003679, -0.00002008, -0.00000125, 0.00001274, -0.00002452, 0.00001169, -0.00006932, 0.00003926, -0.00006987, 0.00008564, -0.00010872, 0.00010995, -0.00012526, 0.00009618, -0.00016709, 0.00014891, -0.00050724, 0.00039297, 0.00082836, -0.00056822, 0.00008658, -0.00013241, 0.00002819, -0.00005094, 0.00003731, -0.00004694, 0.00000618, -0.00007527, 0.00003501, 0.00002215, 0.00015052, -0.00005035, 0.00003881, -0.00009751, 0.00005509, },
		{ 0.00004152, -0.00004444, -0.00001009, 0.00001207, -0.00000004, -0.00001382, -0.00001704, -0.00002265, 0.00003273, 0.00001214, -0.00005253, 0.00001871, 0.00001673, -0.00001083, -0.00000977, 0.00000524, -0.00001842, 0.00000637, 0.00004693, -0.00001297, -0.00001876, 0.00000790, 0.00000939, -0.00000353, 0.00001148, -0.00000797, 0.00002806, -0.00003972, 0.00002452, 0.00000075, -0.00002183, 0.00003162, -0.00003874, 0.00002509, -0.00007371, 0.00003821, -0.00007345, 0.00008529, -0.00009683, 0.00008676, -0.00009108, 0.00005790, -0.00011928, 0.00009113, -0.00034589, -0.00000450, 0.00113317, -0.00054949, 0.00001737, -0.00010213, 0.00000881, -0.00005046, 0.00003928, -0.00005017, 0.00001961, -0.00008496, 0.00004424, -0.00001701, 0.00016499, -0.00005084, 0.00005434, -0.00009745, 0.00004192, -0.00003336, },
		{ -0.00000607, -0.00004495, 0.00003458, -0.00001430, -0.00000028, -0.00002247, -0.00001841, 0.00001372, 0.00003479, -0.00005521, 0.00000663, 0.00002447, -0.00001189, -0.00000733, -0.00000081, -0.00000835, -0.00001016, 0.00005128, -0.00000442, -0.00001868, -0.00000126, 0.00001668, -0.00001029, 0.00001713, -0.00001316, 0.00003227, -0.00003436, 0.00000768, 0.00001936, -0.00003373, 0.00003644, -0.00003457, 0.00001714, -0.00005608, 0.00001711, -0.00005515, 0.00005740, -0.00005326, 0.00003011, -0.00002238, -0.00001756, -0.00002798, -0.00002448, -0.00012150, -0.00038786, 0.00130291, -0.00038525, -0.00013052, -0.00002952, -0.00004016, -0.00002554, 0.00001828, -0.00003360, 0.00001386, -0.00007645, 0.00003806, -0.00003268, 0.00015078, -0.00002675, 0.00004095, -0.00006530, 0.00000171, 0.00000124, 0.00000865, },
		{ -0.00007731, 0.00005177, -0.00002358, 0.00000906, -0.00002605, -0.00001189, -0.00000598, 0.00005036, -0.00004801, -0.00000982, 0.00003170, -0.00001193, -0.00000545, -0.00000551, -0.00000140, -0.00001939, 0.00004625, 0.00000837, -0.00001851, -0.00000900, 0.00002068, -0.00001319, 0.00001851, -0.00001429, 0.00003178, -0.00002479, -0.00000833, 0.00002954, -0.00003174, 0.00002417, -0.00001219, -0.00000836, -0.00002326, -0.00001623, -0.00002265, 0.00001113, 0.00000945, -0.00004357, 0.00006048, -0.00010541, 0.00007552, -0.00015335, 0.00009607, -0.00066965, 0.00127567, -0.00006366, -0.00034354, 0.00007244, -0.00011083, 0.00001944, -0.00002298, 0.00000123, -0.00001072, -0.00004906, 0.00001348, -0.00002134, 0.00010670, 0.00002238, -0.00000007, -0.00000728, -0.00005548, 0.00004636, -0.00003136, 0.00003519, },
		{ 0.00005450, -0.00002167, 0.00001026, -0.00002511, -0.00000662, -0.00002130, 0.00005519, -0.00003133, -0.00002766, 0.00003555, -0.00000860, -0.00000665, -0.00000606, -0.00000031, -0.00001980, 0.00003289, 0.00002311, -0.00001745, -0.00001466, 0.00002063, -0.00001189, 0.00001591, -0.00001226, 0.00002823, -0.00001401, -0.00001966, 0.00002872, -0.00001642, -0.00000087, 0.00002003, -0.00003954, 0.00001092, -0.00004786, 0.00001003, -0.00003707, 0.00007079, -0.00010971, 0.00012919, -0.00017309, 0.00015243, -0.00024531, 0.00024026, -0.00078532, 0.00103119, 0.00038376, -0.00057686, 0.00017067, -0.00018081, 0.00006791, -0.00007090, 0.00004289, -0.00004449, -0.00001100, -0.00002324, 0.00001022, 0.00004242, 0.00008547, -0.00005505, 0.00005867, -0.00011028, 0.00008392, -0.00006217, 0.00006465, -0.00009490, },
		{ -0.00000596, 0.00000211, -0.00001853, -0.00000509, -0.00002892, 0.00004840, -0.00000816, -0.00004279, 0.00003338, -0.00000046, -0.00001197, -0.00000176, -0.00000508, -0.00001302, 0.00001501, 0.00003609, -0.00001409, -0.00001827, 0.00001682, -0.00000721, 0.00001062, -0.00000847, 0.00002312, -0.00000387, -0.00002526, 0.00001894, 0.00000604, -0.00002843, 0.00004852, -0.00006100, 0.00003090, -0.00006324, 0.00002831, -0.00006991, 0.00010984, -0.00014466, 0.00015802, -0.00019271, 0.00017146, -0.00026316, 0.00026976, -0.00071389, 0.00060316, 0.00088094, -0.00075983, 0.00021872, -0.00021870, 0.00009558, -0.00010465, 0.00007313, -0.00007121, 0.00002287, -0.00005848, 0.00004688, -0.00002376, 0.00014289, -0.00010222, 0.00010873, -0.00014062, 0.00009576, -0.00006900, 0.00007042, -0.00008909, 0.00003759, },
		{ -0.00001277, -0.00000776, -0.00000765, -0.00002827, 0.00003222, 0.00001655, -0.00005112, 0.00002385, 0.00001205, -0.00002028, 0.00000557, -0.00001294, -0.00000334, -0.00000198, 0.00004329, -0.00000717, -0.00002023, 0.00001033, -0.00000076, 0.00000433, -0.00000422, 0.00001727, 0.00000534, -0.00002679, 0.00000559, 0.00002632, -0.00004647, 0.00005999, -0.00006008, 0.00002600, -0.00005374, 0.00002357, -0.00007639, 0.00011351, -0.00013493, 0.00013421, -0.00015252, 0.00012166, -0.00019691, 0.00018222, -0.00048595, 0.00007591, 0.00132099, -0.00081315, 0.00017164, -0.00019586, 0.00007986, -0.00010438, 0.00007415, -0.00007458, 0.00003697, -0.00007693, 0.00007104, -0.00007136, 0.00017361, -0.00012002, 0.00012224, -0.00013055, 0.00007134, -0.00004512, 0.00004862, -0.00005892, 0.00000262, 0.00002078, },
		{ 0.00000337, -0.00001240, -0.00002150, 0.00001133, 0.00003722, -0.00004990, 0.00000780, 0.00002640, -0.00002843, 0.00001219, -0.00001942, 0.00000395, -0.00001286, 0.00004190, 0.00000353, -0.00002075, 0.00000252, 0.00000565, -0.00000138, -0.00000031, 0.00001072, 0.00001444, -0.00002723, -0.00000512, 0.00003597, -0.00004617, 0.00004719, -0.00003270, -0.00000440, -0.00002109, -0.00000253, -0.00005661, 0.00008188, -0.00008289, 0.00006370, -0.00006257, 0.00001768, -0.00006791, 0.00001467, -0.00017538, -0.00043538, 0.00159280, -0.00067396, 0.00000525, -0.00010057, 0.00001127, -0.00006091, 0.00003730, -0.00004614, 0.00002243, -0.00006884, 0.00007014, -0.00008568, 0.00016384, -0.00009611, 0.00009034, -0.00007748, 0.00001325, 0.00000429, 0.00000587, -0.00001238, -0.00004169, 0.00005058, -0.00002829, },
		{ -0.00001612, -0.00001248, -0.00000868, 0.00004921, -0.00003869, -0.00001173, 0.00003855, -0.00003237, 0.00001397, -0.00002029, 0.00000465, -0.00001455, 0.00003162, 0.00001662, -0.00001953, -0.00000536, 0.00001049, -0.00000547, 0.00000295, 0.00000326, 0.00002417, -0.00002890, -0.00000914, 0.00003103, -0.00002594, 0.00001250, 0.00001409, -0.00004952, 0.00002216, -0.00003792, -0.00002192, 0.00002828, -0.00000636, -0.00003028, 0.00004813, -0.00010480, 0.00007874, -0.00016985, 0.00012319, -0.00081575, 0.00161341, -0.00032109, -0.00026754, 0.00005122, -0.00009845, 0.00001728, -0.00003074, 0.00000906, -0.00001754, -0.00003545, 0.00004251, -0.00006411, 0.00011332, -0.00003324, 0.00002059, 0.00000518, -0.00006215, 0.00006286, -0.00004227, 0.00003569, -0.00008188, 0.00007266, -0.00003687, 0.00001006, },
		{ -0.00000513, -0.00002302, 0.00005019, -0.00001968, -0.00003016, 0.00004428, -0.00002879, 0.00000821, -0.00001342, -0.00000265, -0.00000733, 0.00001514, 0.00002910, -0.00001582, -0.00001239, 0.00001276, -0.00000745, 0.00000553, -0.00000493, 0.00003420, -0.00003198, -0.00000642, 0.00001387, 0.00000750, -0.00003229, 0.00006399, -0.00009061, 0.00005705, -0.00006483, 0.00000989, -0.00002605, 0.00006801, -0.00011551, 0.00014187, -0.00020209, 0.00019091, -0.00030473, 0.00032390, -0.00098357, 0.00135573, 0.00020886, -0.00059043, 0.00021418, -0.00021579, 0.00010375, -0.00010770, 0.00007214, -0.00006713, 0.00001007, -0.00000124, -0.00001759, 0.00003647, 0.00005057, -0.00006461, 0.00009159, -0.00012954, 0.00010837, -0.00007673, 0.00006880, -0.00010444, 0.00007734, -0.00003240, 0.00000856, -0.00001589, },
		{ -0.00002911, 0.00004075, 0.00000283, -0.00004273, 0.00004075, -0.00001665, -0.00000496, 0.00000008, -0.00001579, 0.00000510, -0.00000252, 0.00003720, -0.00000885, -0.00001788, 0.00001209, -0.00000735, 0.00000732, -0.00001289, 0.00004267, -0.00003410, -0.00000088, -0.00000778, 0.00004157, -0.00007017, 0.00009737, -0.00010809, 0.00006579, -0.00006740, 0.00002239, -0.00006072, 0.00011502, -0.00016283, 0.00018635, -0.00023877, 0.00022879, -0.00034320, 0.00037625, -0.00091457, 0.00086170, 0.00082184, -0.00087240, 0.00032391, -0.00029460, 0.00016189, -0.00016254, 0.00011642, -0.00010323, 0.00004724, -0.00004233, 0.00003348, -0.00004242, 0.00012805, -0.00013564, 0.00015224, -0.00016365, 0.00012113, -0.00008224, 0.00007500, -0.00010092, 0.00005991, -0.00001300, -0.00000188, -0.00001057, -0.00000197, },
		{ 0.00002409, 0.00002363, -0.00004597, 0.00002775, 0.00000205, -0.00002226, 0.00001601, -0.00002978, 0.00001689, -0.00001557, 0.00003773, 0.00000145, -0.00002121, 0.00000859, -0.00000549, 0.00000800, -0.00001893, 0.00004666, -0.00003141, 0.00000144, -0.00002411, 0.00006270, -0.00008549, 0.00009905, -0.00008971, 0.00003966, -0.00003913, 0.00000811, -0.00006466, 0.00012058, -0.00015711, 0.00016684, -0.00020115, 0.00017970, -0.00027420, 0.00027965, -0.00064869, 0.00023534, 0.00138569, -0.00101091, 0.00031755, -0.00029286, 0.00015888, -0.00016750, 0.00011808, -0.00010497, 0.00005686, -0.00006189, 0.00006756, -0.00009825, 0.00017278, -0.00016651, 0.00016479, -0.00014898, 0.00009199, -0.00005374, 0.00005188, -0.00007149, 0.00002303, 0.00001763, -0.00001821, -0.00000159, -0.00000318, -0.00002716, },
		{ 0.00003806, -0.00003886, 0.00000804, 0.00002262, -0.00003822, 0.00002848, -0.00003866, 0.00002221, -0.00001951, 0.00002943, 0.00001390, -0.00002164, 0.00000271, -0.00000241, 0.00000718, -0.00002125, 0.00004344, -0.00002059, -0.00000446, -0.00002754, 0.00006197, -0.00007039, 0.00006483, -0.00003642, -0.00001618, 0.00001307, -0.00002749, -0.00004061, 0.00008694, -0.00010280, 0.00009163, -0.00010235, 0.00006245, -0.00012507, 0.00008102, -0.00027700, -0.00038371, 0.00176414, -0.00092381, 0.00015864, -0.00019141, 0.00008064, -0.00011083, 0.00006725, -0.00006352, 0.00003003, -0.00004969, 0.00007100, -0.00011475, 0.00016923, -0.00014477, 0.00012279, -0.00008647, 0.00002689, 0.00000083, 0.00000800, -0.00002541, -0.00002364, 0.00005095, -0.00003433, 0.00000752, -0.00000689, -0.00001973, 0.00000500, },
		{ -0.00002319, -0.00001332, 0.00003915, -0.00004704, 0.00003220, -0.00003784, 0.00001750, -0.00001286, 0.00001384, 0.00002598, -0.00001842, -0.00000473, 0.00000115, 0.00000468, -0.00001880, 0.00003195, -0.00000089, -0.00001987, -0.00001622, 0.00003893, -0.00002836, 0.00000401, 0.00003674, -0.00008335, 0.00007043, -0.00006769, -0.00000376, 0.00003123, -0.00002177, -0.00001132, 0.00002280, -0.00008007, 0.00004974, -0.00014341, 0.00008712, -0.00085845, 0.00185581, -0.00057974, -0.00014297, -0.00000703, -0.00005843, -0.00000460, -0.00002536, 0.00001198, -0.00002641, -0.00000961, 0.00004395, -0.00009072, 0.00011884, -0.00007657, 0.00003876, 0.00000578, -0.00005426, 0.00006279, -0.00003991, 0.00002201, -0.00006649, 0.00007633, -0.00004317, 0.00001254, -0.00001014, -0.00001054, -0.00001171, 0.00004332, },
		{ -0.00003051, 0.00004649, -0.00004486, 0.00002454, -0.00002606, 0.00000295, 0.00000210, -0.00000488, 0.00003434, -0.00001109, -0.00001256, 0.00000422, 0.00000090, -0.00001200, 0.00001371, 0.00002475, -0.00004110, 0.00000487, 0.00000210, 0.00002657, -0.00006333, 0.00010535, -0.00013642, 0.00010916, -0.00009130, 0.00002514, -0.00002210, 0.00005564, -0.00010526, 0.00013095, -0.00019719, 0.00018924, -0.00031534, 0.00034297, -0.00109281, 0.00162569, -0.00001609, -0.00052453, 0.00020646, -0.00021615, 0.00011636, -0.00012976, 0.00009585, -0.00009100, 0.00004094, 0.00000037, -0.00004023, 0.00003949, 0.00001542, -0.00005985, 0.00009825, -0.00012422, 0.00010929, -0.00007303, 0.00005495, -0.00009270, 0.00008480, -0.00003949, 0.00001042, -0.00001038, -0.00000307, -0.00002250, 0.00003909, -0.00000302, },
		{ 0.00004219, -0.00003129, 0.00000684, -0.00000610, -0.00001737, 0.00001995, -0.00002103, 0.00003586, -0.00000002, -0.00001921, 0.00000569, -0.00000317, -0.00000279, -0.00000723, 0.00005016, -0.00006048, 0.00002574, -0.00003425, 0.00007491, -0.00011347, 0.00014439, -0.00015255, 0.00011003, -0.00008199, 0.00002913, -0.00005164, 0.00010260, -0.00015862, 0.00018646, -0.00024919, 0.00024806, -0.00038079, 0.00043204, -0.00105758, 0.00111841, 0.00066309, -0.00088181, 0.00037171, -0.00033488, 0.00020551, -0.00020654, 0.00015461, -0.00013544, 0.00007777, -0.00003800, 0.00001350, -0.00004140, 0.00010026, -0.00014038, 0.00016035, -0.00015820, 0.00012190, -0.00007776, 0.00006296, -0.00009486, 0.00007247, -0.00002203, 0.00000078, -0.00000664, 0.00000064, -0.00002597, 0.00002742, 0.00001656, -0.00003881, },
		{ -0.00000980, -0.00001592, 0.00001602, -0.00003687, 0.00003399, -0.00002930, 0.00002890, 0.00001328, -0.00002290, 0.00000451, -0.00000616, 0.00000588, -0.00002548, 0.00006790, -0.00006926, 0.00003548, -0.00005570, 0.00009909, -0.00012787, 0.00013786, -0.00012078, 0.00006689, -0.00003610, 0.00000304, -0.00004785, 0.00010601, -0.00015654, 0.00017413, -0.00022133, 0.00021183, -0.00032658, 0.00035247, -0.00079580, 0.00044812, 0.00131065, -0.00109689, 0.00041274, -0.00036217, 0.00022188, -0.00022188, 0.00016016, -0.00013616, 0.00008030, -0.00005132, 0.00004787, -0.00009750, 0.00015021, -0.00017668, 0.00017140, -0.00014316, 0.00009393, -0.00005106, 0.00004492, -0.00007370, 0.00004211, 0.00000559, -0.00001367, 0.00000004, 0.00000042, -0.00002290, 0.00001191, 0.00003096, -0.00003610, 0.00002744, },
		{ -0.00003670, 0.00003316, -0.00004885, 0.00003857, -0.00002659, 0.00001420, 0.00002611, -0.00002193, 0.00000012, -0.00000700, 0.00001118, -0.00003616, 0.00007202, -0.00006120, 0.00002693, -0.00005359, 0.00009026, -0.00010036, 0.00008529, -0.00004728, -0.00000959, 0.00003446, -0.00004365, -0.00001670, 0.00007079, -0.00010544, 0.00010374, -0.00012790, 0.00010042, -0.00018141, 0.00015435, -0.00040620, -0.00023256, 0.00177526, -0.00107446, 0.00028345, -0.00027250, 0.00014752, -0.00016208, 0.00010221, -0.00008513, 0.00004115, -0.00003081, 0.00005034, -0.00011347, 0.00015085, -0.00015850, 0.00012814, -0.00008298, 0.00003335, -0.00000194, 0.00000943, -0.00003773, 0.00000231, 0.00003598, -0.00002798, 0.00000708, -0.00000231, -0.00001564, -0.00000345, 0.00003733, -0.00002351, 0.00000681, 0.00001331, },
		{ 0.00003960, -0.00004893, 0.00003115, -0.00001326, -0.00000496, 0.00003515, -0.00001535, -0.00000711, -0.00000541, 0.00001148, -0.00003674, 0.00006044, -0.00003552, 0.00000002, -0.00002856, 0.00005180, -0.00003973, 0.00000266, 0.00004588, -0.00009473, 0.00010579, -0.00009025, 0.00002345, 0.00001667, -0.00002924, 0.00000493, -0.00000541, -0.00004146, -0.00000237, -0.00008142, -0.00001060, -0.00077491, 0.00194404, -0.00077491, -0.00001060, -0.00008142, -0.00000237, -0.00004146, -0.00000541, 0.00000493, -0.00002924, 0.00001667, 0.00002345, -0.00009025, 0.00010579, -0.00009473, 0.00004588, 0.00000266, -0.00003973, 0.00005180, -0.00002856, 0.00000002, -0.00003552, 0.00006044, -0.00003674, 0.00001148, -0.00000541, -0.00000711, -0.00001535, 0.00003515, -0.00000496, -0.00001326, 0.00003115, -0.00004893, },
		{ -0.00003670, 0.00001331, 0.00000681, -0.00002351, 0.00003733, -0.00000345, -0.00001564, -0.00000231, 0.00000708, -0.00002798, 0.00003598, 0.00000231, -0.00003773, 0.00000943, -0.00000194, 0.00003335, -0.00008298, 0.00012814, -0.00015850, 0.00015085, -0.00011347, 0.00005034, -0.00003081, 0.00004115, -0.00008513, 0.00010221, -0.00016208, 0.00014752, -0.00027250, 0.00028345, -0.00107446, 0.00177526, -0.00023256, -0.00040620, 0.00015435, -0.00018141, 0.00010042, -0.00012790, 0.00010374, -0.00010544, 0.00007079, -0.00001670, -0.00004365, 0.00003446, -0.00000959, -0.00004728, 0.00008529, -0.00010036, 0.00009026, -0.00005359, 0.00002693, -0.00006120, 0.00007202, -0.00003616, 0.00001118, -0.00000700, 0.00000012, -0.00002193, 0.00002611, 0.00001420, -0.00002659, 0.00003857, -0.00004885, 0.00003316, },
		{ -0.00000980, 0.00002744, -0.00003610, 0.00003096, 0.00001191, -0.00002290, 0.00000042, 0.00000004, -0.00001367, 0.00000559, 0.00004211, -0.00007370, 0.00004492, -0.00005106, 0.00009393, -0.00014316, 0.00017140, -0.00017668, 0.00015021, -0.00009750, 0.00004787, -0.00005132, 0.00008030, -0.00013616, 0.00016016, -0.00022188, 0.00022188, -0.00036217, 0.00041274, -0.00109689, 0.00131065, 0.00044812, -0.00079580, 0.00035247, -0.00032658, 0.00021183, -0.00022133, 0.00017413, -0.00015654, 0.00010601, -0.00004785, 0.00000304, -0.00003610, 0.00006689, -0.00012078, 0.00013786, -0.00012787, 0.00009909, -0.00005570, 0.00003548, -0.00006926, 0.00006790, -0.00002548, 0.00000588, -0.00000616, 0.00000451, -0.00002290, 0.00001328, 0.00002890, -0.00002930, 0.00003399, -0.00003687, 0.00001602, -0.00001592, },
		{ 0.00004219, -0.00003881, 0.00001656, 0.00002742, -0.00002597, 0.00000064, -0.00000664, 0.00000078, -0.00002203, 0.00007247, -0.00009486, 0.00006296, -0.00007776, 0.00012190, -0.00015820, 0.00016035, -0.00014038, 0.00010026, -0.00004140, 0.00001350, -0.00003800, 0.00007777, -0.00013544, 0.00015461, -0.00020654, 0.00020551, -0.00033488, 0.00037171, -0.00088181, 0.00066309, 0.00111841, -0.00105758, 0.00043204, -0.00038079, 0.00024806, -0.00024919, 0.00018646, -0.00015862, 0.00010260, -0.00005164, 0.00002913, -0.00008199, 0.00011003, -0.00015255, 0.00014439, -0.00011347, 0.00007491, -0.00003425, 0.00002574, -0.00006048, 0.00005016, -0.00000723, -0.00000279, -0.00000317, 0.00000569, -0.00001921, -0.00000002, 0.00003586, -0.00002103, 0.00001995, -0.00001737, -0.00000610, 0.00000684, -0.00003129, },
		{ -0.00003051, -0.00000302, 0.00003909, -0.00002250, -0.00000307, -0.00001038, 0.00001042, -0.00003949, 0.00008480, -0.00009270, 0.00005495, -0.00007303, 0.00010929, -0.00012422, 0.00009825, -0.00005985, 0.00001542, 0.00003949, -0.00004023, 0.00000037, 0.00004094, -0.00009100, 0.00009585, -0.00012976, 0.00011636, -0.00021615, 0.00020646, -0.00052453, -0.00001609, 0.00162569, -0.00109281, 0.00034297, -0.00031534, 0.00018924, -0.00019719, 0.00013095, -0.00010526, 0.00005564, -0.00002210, 0.00002514, -0.00009130, 0.00010916, -0.00013642, 0.00010535, -0.00006333, 0.00002657, 0.00000210, 0.00000487, -0.00004110, 0.00002475, 0.00001371, -0.00001200, 0.00000090, 0.00000422, -0.00001256, -0.00001109, 0.00003434, -0.00000488, 0.00000210, 0.00000295, -0.00002606, 0.00002454, -0.00004486, 0.00004649, },
		{ -0.00002319, 0.00004332, -0.00001171, -0.00001054, -0.00001014, 0.00001254, -0.00004317, 0.00007633, -0.00006649, 0.00002201, -0.00003991, 0.00006279, -0.00005426, 0.00000578, 0.00003876, -0.00007657, 0.00011884, -0.00009072, 0.00004395, -0.00000961, -0.00002641, 0.00001198, -0.00002536, -0.00000460, -0.00005843, -0.00000703, -0.00014297, -0.00057974, 0.00185581, -0.00085845, 0.00008712, -0.00014341, 0.00004974, -0.00008007, 0.00002280, -0.00001132, -0.00002177, 0.00003123, -0.00000376, -0.00006769, 0.00007043, -0.00008335, 0.00003674, 0.00000401, -0.00002836, 0.00003893, -0.00001622, -0.00001987, -0.00000089, 0.00003195, -0.00001880, 0.00000468, 0.00000115, -0.00000473, -0.00001842, 0.00002598, 0.00001384, -0.00001286, 0.00001750, -0.00003784, 0.00003220, -0.00004704, 0.00003915, -0.00001332, },
		{ 0.00003806, 0.00000500, -0.00001973, -0.00000689, 0.00000752, -0.00003433, 0.00005095, -0.00002364, -0.00002541, 0.00000800, 0.00000083, 0.00002689, -0.00008647, 0.00012279, -0.00014477, 0.00016923, -0.00011475, 0.00007100, -0.00004969, 0.00003003, -0.00006352, 0.00006725, -0.00011083, 0.00008064, -0.00019141, 0.00015864, -0.00092381, 0.00176414, -0.00038371, -0.00027700, 0.00008102, -0.00012507, 0.00006245, -0.00010235, 0.00009163, -0.00010280, 0.00008694, -0.00004061, -0.00002749, 0.00001307, -0.00001618, -0.00003642, 0.00006483, -0.00007039, 0.00006197, -0.00002754, -0.00000446, -0.00002059, 0.00004344, -0.00002125, 0.00000718, -0.00000241, 0.00000271, -0.00002164, 0.00001390, 0.00002943, -0.00001951, 0.00002221, -0.00003866, 0.00002848, -0.00003822, 0.00002262, 0.00000804, -0.00003886, },
		{ 0.00002409, -0.00002716, -0.00000318, -0.00000159, -0.00001821, 0.00001763, 0.00002303, -0.00007149, 0.00005188, -0.00005374, 0.00009199, -0.00014898, 0.00016479, -0.00016651, 0.00017278, -0.00009825, 0.00006756, -0.00006189, 0.00005686, -0.00010497, 0.00011808, -0.00016750, 0.00015888, -0.00029286, 0.00031755, -0.00101091, 0.00138569, 0.00023534, -0.00064869, 0.00027965, -0.00027420, 0.00017970, -0.00020115, 0.00016684, -0.00015711, 0.00012058, -0.00006466, 0.00000811, -0.00003913, 0.00003966, -0.00008971, 0.00009905, -0.00008549, 0.00006270, -0.00002411, 0.00000144, -0.00003141, 0.00004666, -0.00001893, 0.00000800, -0.00000549, 0.00000859, -0.00002121, 0.00000145, 0.00003773, -0.00001557, 0.00001689, -0.00002978, 0.00001601, -0.00002226, 0.00000205, 0.00002775, -0.00004597, 0.00002363, },
		{ -0.00002911, -0.00000197, -0.00001057, -0.00000188, -0.00001300, 0.00005991, -0.00010092, 0.00007500, -0.00008224, 0.00012113, -0.00016365, 0.00015224, -0.00013564, 0.00012805, -0.00004242, 0.00003348, -0.00004233, 0.00004724, -0.00010323, 0.00011642, -0.00016254, 0.00016189, -0.00029460, 0.00032391, -0.00087240, 0.00082184, 0.00086170, -0.00091457, 0.00037625, -0.00034320, 0.00022879, -0.00023877, 0.00018635, -0.00016283, 0.00011502, -0.00006072, 0.00002239, -0.00006740, 0.00006579, -0.00010809, 0.00009737, -0.00007017, 0.00004157, -0.00000778, -0.00000088, -0.00003410, 0.00004267, -0.00001289, 0.00000732, -0.00000735, 0.00001209, -0.00001788, -0.00000885, 0.00003720, -0.00000252, 0.00000510, -0.00001579, 0.00000008, -0.00000496, -0.00001665, 0.00004075, -0.00004273, 0.00000283, 0.00004075, },
		{ -0.00000513, -0.00001589, 0.00000856, -0.00003240, 0.00007734, -0.00010444, 0.00006880, -0.00007673, 0.00010837, -0.00012954, 0.00009159, -0.00006461, 0.00005057, 0.00003647, -0.00001759, -0.00000124, 0.00001007, -0.00006713, 0.00007214, -0.00010770, 0.00010375, -0.00021579, 0.00021418, -0.00059043, 0.00020886, 0.00135573, -0.00098357, 0.00032390, -0.00030473, 0.00019091, -0.00020209, 0.00014187, -0.00011551, 0.00006801, -0.00002605, 0.00000989, -0.00006483, 0.00005705, -0.00009061, 0.00006399, -0.00003229, 0.00000750, 0.00001387, -0.00000642, -0.00003198, 0.00003420, -0.00000493, 0.00000553, -0.00000745, 0.00001276, -0.00001239, -0.00001582, 0.00002910, 0.00001514, -0.00000733, -0.00000265, -0.00001342, 0.00000821, -0.00002879, 0.00004428, -0.00003016, -0.00001968, 0.00005019, -0.00002302, },
		{ -0.00001612, 0.00001006, -0.00003687, 0.00007266, -0.00008188, 0.00003569, -0.00004227, 0.00006286, -0.00006215, 0.00000518, 0.00002059, -0.00003324, 0.00011332, -0.00006411, 0.00004251, -0.00003545, -0.00001754, 0.00000906, -0.00003074, 0.00001728, -0.00009845, 0.00005122, -0.00026754, -0.00032109, 0.00161341, -0.00081575, 0.00012319, -0.00016985, 0.00007874, -0.00010480, 0.00004813, -0.00003028, -0.00000636, 0.00002828, -0.00002192, -0.00003792, 0.00002216, -0.00004952, 0.00001409, 0.00001250, -0.00002594, 0.00003103, -0.00000914, -0.00002890, 0.00002417, 0.00000326, 0.00000295, -0.00000547, 0.00001049, -0.00000536, -0.00001953, 0.00001662, 0.00003162, -0.00001455, 0.00000465, -0.00002029, 0.00001397, -0.00003237, 0.00003855, -0.00001173, -0.00003869, 0.00004921, -0.00000868, -0.00001248, },
		{ 0.00000337, -0.00002829, 0.00005058, -0.00004169, -0.00001238, 0.00000587, 0.00000429, 0.00001325, -0.00007748, 0.00009034, -0.00009611, 0.00016384, -0.00008568, 0.00007014, -0.00006884, 0.00002243, -0.00004614, 0.00003730, -0.00006091, 0.00001127, -0.00010057, 0.00000525, -0.00067396, 0.00159280, -0.00043538, -0.00017538, 0.00001467, -0.00006791, 0.00001768, -0.00006257, 0.00006370, -0.00008289, 0.00008188, -0.00005661, -0.00000253, -0.00002109, -0.00000440, -0.00003270, 0.00004719, -0.00004617, 0.00003597, -0.00000512, -0.00002723, 0.00001444, 0.00001072, -0.00000031, -0.00000138, 0.00000565, 0.00000252, -0.00002075, 0.00000353, 0.00004190, -0.00001286, 0.00000395, -0.00001942, 0.00001219, -0.00002843, 0.00002640, 0.00000780, -0.00004990, 0.00003722, 0.00001133, -0.00002150, -0.00001240, },
		{ -0.00001277, 0.00002078, 0.00000262, -0.00005892, 0.00004862, -0.00004512, 0.00007134, -0.00013055, 0.00012224, -0.00012002, 0.00017361, -0.00007136, 0.00007104, -0.00007693, 0.00003697, -0.00007458, 0.00007415, -0.00010438, 0.00007986, -0.00019586, 0.00017164, -0.00081315, 0.00132099, 0.00007591, -0.00048595, 0.00018222, -0.00019691, 0.00012166, -0.00015252, 0.00013421, -0.00013493, 0.00011351, -0.00007639, 0.00002357, -0.00005374, 0.00002600, -0.00006008, 0.00005999, -0.00004647, 0.00002632, 0.00000559, -0.00002679, 0.00000534, 0.00001727, -0.00000422, 0.00000433, -0.00000076, 0.00001033, -0.00002023, -0.00000717, 0.00004329, -0.00000198, -0.00000334, -0.00001294, 0.00000557, -0.00002028, 0.00001205, 0.00002385, -0.00005112, 0.00001655, 0.00003222, -0.00002827, -0.00000765, -0.00000776, },
		{ -0.00000596, 0.00003759, -0.00008909, 0.00007042, -0.00006900, 0.00009576, -0.00014062, 0.00010873, -0.00010222, 0.00014289, -0.00002376, 0.00004688, -0.00005848, 0.00002287, -0.00007121, 0.00007313, -0.00010465, 0.00009558, -0.00021870, 0.00021872, -0.00075983, 0.00088094, 0.00060316, -0.00071389, 0.00026976, -0.00026316, 0.00017146, -0.00019271, 0.00015802, -0.00014466, 0.00010984, -0.00006991, 0.00002831, -0.00006324, 0.00003090, -0.00006100, 0.00004852, -0.00002843, 0.00000604, 0.00001894, -0.00002526, -0.00000387, 0.00002312, -0.00000847, 0.00001062, -0.00000721, 0.00001682, -0.00001827, -0.00001409, 0.00003609, 0.00001501, -0.00001302, -0.00000508, -0.00000176, -0.00001197, -0.00000046, 0.00003338, -0.00004279, -0.00000816, 0.00004840, -0.00002892, -0.00000509, -0.00001853, 0.00000211, },
		{ 0.00005450, -0.00009490, 0.00006465, -0.00006217, 0.00008392, -0.00011028, 0.00005867, -0.00005505, 0.00008547, 0.00004242, 0.00001022, -0.00002324, -0.00001100, -0.00004449, 0.00004289, -0.00007090, 0.00006791, -0.00018081, 0.00017067, -0.00057686, 0.00038376, 0.00103119, -0.00078532, 0.00024026, -0.00024531, 0.00015243, -0.00017309, 0.00012919, -0.00010971, 0.00007079, -0.00003707, 0.00001003, -0.00004786, 0.00001092, -0.00003954, 0.00002003, -0.00000087, -0.00001642, 0.00002872, -0.00001966, -0.00001401, 0.00002823, -0.00001226, 0.00001591, -0.00001189, 0.00002063, -0.00001466, -0.00001745, 0.00002311, 0.00003289, -0.00001980, -0.00000031, -0.00000606, -0.00000665, -0.00000860, 0.00003555, -0.00002766, -0.00003133, 0.00005519, -0.00002130, -0.00000662, -0.00002511, 0.00001026, -0.00002167, },
		{ -0.00007731, 0.00003519, -0.00003136, 0.00004636, -0.00005548, -0.00000728, -0.00000007, 0.00002238, 0.00010670, -0.00002134, 0.00001348, -0.00004906, -0.00001072, 0.00000123, -0.00002298, 0.00001944, -0.00011083, 0.00007244, -0.00034354, -0.00006366, 0.00127567, -0.00066965, 0.00009607, -0.00015335, 0.00007552, -0.00010541, 0.00006048, -0.00004357, 0.00000945, 0.00001113, -0.00002265, -0.00001623, -0.00002326, -0.00000836, -0.00001219, 0.00002417, -0.00003174, 0.00002954, -0.00000833, -0.00002479, 0.00003178, -0.00001429, 0.00001851, -0.00001319, 0.00002068, -0.00000900, -0.00001851, 0.00000837, 0.00004625, -0.00001939, -0.00000140, -0.00000551, -0.00000545, -0.00001193, 0.00003170, -0.00000982, -0.00004801, 0.00005036, -0.00000598, -0.00001189, -0.00002605, 0.00000906, -0.00002358, 0.00005177, },
		{ -0.00000607, 0.00000865, 0.00000124, 0.00000171, -0.00006530, 0.00004095, -0.00002675, 0.00015078, -0.00003268, 0.00003806, -0.00007645, 0.00001386, -0.00003360, 0.00001828, -0.00002554, -0.00004016, -0.00002952, -0.00013052, -0.00038525, 0.00130291, -0.00038786, -0.00012150, -0.00002448, -0.00002798, -0.00001756, -0.00002238, 0.00003011, -0.00005326, 0.00005740, -0.00005515, 0.00001711, -0.00005608, 0.00001714, -0.00003457, 0.00003644, -0.00003373, 0.00001936, 0.00000768, -0.00003436, 0.00003227, -0.00001316, 0.00001713, -0.00001029, 0.00001668, -0.00000126, -0.00001868, -0.00000442, 0.00005128, -0.00001016, -0.00000835, -0.00000081, -0.00000733, -0.00001189, 0.00002447, 0.00000663, -0.00005521, 0.00003479, 0.00001372, -0.00001841, -0.00002247, -0.00000028, -0.00001430, 0.00003458, -0.00004495, },
		{ 0.00004152, -0.00003336, 0.00004192, -0.00009745, 0.00005434, -0.00005084, 0.00016499, -0.00001701, 0.00004424, -0.00008496, 0.00001961, -0.00005017, 0.00003928, -0.00005046, 0.00000881, -0.00010213, 0.00001737, -0.00054949, 0.00113317, -0.00000450, -0.00034589, 0.00009113, -0.00011928, 0.00005790, -0.00009108, 0.00008676, -0.00009683, 0.00008529, -0.00007345, 0.00003821, -0.00007371, 0.00002509, -0.00003874, 0.00003162, -0.00002183, 0.00000075, 0.00002452, -0.00003972, 0.00002806, -0.00000797, 0.00001148, -0.00000353, 0.00000939, 0.00000790, -0.00001876, -0.00001297, 0.00004693, 0.00000637, -0.00001842, 0.00000524, -0.00000977, -0.00001083, 0.00001673, 0.00001871, -0.00005253, 0.00001214, 0.00003273, -0.00002265, -0.00001704, -0.00001382, -0.00000004, 0.00001207, -0.00001009, -0.00004444, },
		{ -0.00004618, 0.00005509, -0.00009751, 0.00003881, -0.00005035, 0.00015052, 0.00002215, 0.00003501, -0.00007527, 0.00000618, -0.00004694, 0.00003731, -0.00005094, 0.00002819, -0.00013241, 0.00008658, -0.00056822, 0.00082836, 0.00039297, -0.00050724, 0.00014891, -0.00016709, 0.00009618, -0.00012526, 0.00010995, -0.00010872, 0.00008564, -0.00006987, 0.00003926, -0.00006932, 0.00001169, -0.00002452, 0.00001274, -0.00000125, -0.00002008, 0.00003679, -0.00003791, 0.00001835, 0.00000106, 0.00000254, 0.00000555, 0.00000054, 0.00001706, -0.00001860, -0.00001692, 0.00003496, 0.00002604, -0.00002722, 0.00000909, -0.00001012, -0.00001085, 0.00001054, 0.00002520, -0.00004192, -0.00001234, 0.00004595, -0.00002161, -0.00001253, -0.00002692, 0.00001224, -0.00000653, 0.00001617, -0.00006808, 0.00005560, },
		{ 0.00004286, -0.00007165, 0.00000444, -0.00003524, 0.00011713, 0.00007341, 0.00002036, -0.00005503, -0.00001918, -0.00003079, 0.00001906, -0.00003337, 0.00002332, -0.00012548, 0.00008948, -0.00048339, 0.00046978, 0.00072000, -0.00055443, 0.00012592, -0.00015905, 0.00008890, -0.00011925, 0.00009657, -0.00008837, 0.00005971, -0.00004578, 0.00002111, -0.00004512, -0.00001797, 0.00000045, -0.00001139, 0.00001934, -0.00003561, 0.00003970, -0.00002756, 0.00000394, 0.00001227, -0.00000753, 0.00001453, -0.00000758, 0.00002440, -0.00001730, -0.00001748, 0.00001901, 0.00004365, -0.00003042, 0.00000798, -0.00000678, -0.00001286, 0.00000658, 0.00002660, -0.00002679, -0.00003344, 0.00004992, -0.00001418, -0.00001041, -0.00003602, 0.00001740, -0.00001518, 0.00002761, -0.00007188, 0.00004767, -0.00003602, },
		{ -0.00003437, -0.00003276, -0.00001920, 0.00007772, 0.00012266, 0.00001232, -0.00003417, -0.00004617, -0.00001229, -0.00000433, -0.00000957, 0.00000670, -0.00009686, 0.00005253, -0.00034841, 0.00013416, 0.00091723, -0.00047005, 0.00002842, -0.00010516, 0.00004550, -0.00008240, 0.00005642, -0.00004620, 0.00001787, -0.00001035, -0.00000798, -0.00001057, -0.00005250, 0.00002433, -0.00003016, 0.00003147, -0.00004017, 0.00003116, -0.00000987, -0.00001251, 0.00002287, -0.00001600, 0.00002091, -0.00001284, 0.00002826, -0.00001379, -0.00001660, 0.00000325, 0.00005462, -0.00002541, 0.00000110, 0.00000002, -0.00001630, 0.00000429, 0.00002452, -0.00001084, -0.00004742, 0.00004383, -0.00000168, -0.00001037, -0.00003969, 0.00001359, -0.00001258, 0.00002429, -0.00005828, 0.00002299, -0.00001074, 0.00001606, },
		{ -0.00005806, -0.00001310, 0.00004259, 0.00015842, 0.00001983, -0.00001997, -0.00006648, -0.00000043, -0.00002294, 0.00000945, -0.00000909, -0.00006345, 0.00000345, -0.00021086, -0.00012448, 0.00096281, -0.00027351, -0.00011119, -0.00003248, -0.00001228, -0.00003365, 0.00000694, 0.00000144, -0.00002504, 0.00002373, -0.00003643, 0.00002198, -0.00007912, 0.00003603, -0.00003550, 0.00003037, -0.00003214, 0.00001288, 0.00001124, -0.00002699, 0.00002976, -0.00002039, 0.00002285, -0.00001394, 0.00002774, -0.00000743, -0.00001598, -0.00000901, 0.00005646, -0.00001218, -0.00000999, 0.00000836, -0.00001953, 0.00000240, 0.00002087, 0.00000305, -0.00005279, 0.00002956, 0.00001265, -0.00001073, -0.00003845, 0.00000217, -0.00000177, 0.00001137, -0.00003518, -0.00000794, 0.00001720, -0.00001098, -0.00000164, },
		{ -0.00002051, 0.00001588, 0.00017519, 0.00004557, -0.00001408, -0.00007682, 0.00000076, -0.00003181, 0.00001774, -0.00001654, -0.00003683, -0.00003781, -0.00010175, -0.00028283, 0.00087212, -0.00001247, -0.00024657, 0.00002651, -0.00006041, 0.00000712, -0.00003443, 0.00003903, -0.00005574, 0.00004547, -0.00005424, 0.00004224, -0.00008859, 0.00002951, -0.00002517, 0.00001679, -0.00001465, -0.00001003, 0.00003022, -0.00003537, 0.00003052, -0.00001938, 0.00001980, -0.00001079, 0.00002299, 0.00000156, -0.00001642, -0.00001619, 0.00004930, 0.00000665, -0.00002193, 0.00001536, -0.00002059, -0.00000030, 0.00001714, 0.00001328, -0.00005022, 0.00001086, 0.00002495, -0.00000956, -0.00003384, -0.00001354, 0.00001181, -0.00000371, -0.00001212, -0.00003407, 0.00003625, -0.00002657, 0.00001553, -0.00006376, },
		{ -0.00000433, 0.00017345, 0.00008584, -0.00001246, -0.00007914, -0.00000739, -0.00003165, 0.00001540, -0.00001465, -0.00002093, -0.00006244, -0.00003287, -0.00034581, 0.00068714, 0.00025343, -0.00033352, 0.00004673, -0.00008245, 0.00002698, -0.00005681, 0.00005744, -0.00006709, 0.00004957, -0.00005665, 0.00004549, -0.00007822, 0.00000585, -0.00000334, -0.00000350, 0.00000579, -0.00003081, 0.00004163, -0.00003488, 0.00002446, -0.00001327, 0.00001266, -0.00000451, 0.00001520, 0.00001205, -0.00001762, -0.00001852, 0.00003553, 0.00002681, -0.00003079, 0.00001837, -0.00001818, -0.00000433, 0.00001399, 0.00001956, -0.00004182, -0.00000812, 0.00003214, -0.00000584, -0.00002735, -0.00002978, 0.00002267, -0.00001432, 0.00000343, -0.00004801, 0.00004008, -0.00002589, 0.00001457, -0.00005139, -0.00003706, },
		{ 0.00015722, 0.00013295, -0.00000782, -0.00007833, -0.00002028, -0.00002682, 0.00000664, -0.00000710, -0.00001367, -0.00007093, -0.00000093, -0.00033691, 0.00046041, 0.00047041, -0.00034415, 0.00001971, -0.00007536, 0.00002397, -0.00005880, 0.00005603, -0.00005965, 0.00003753, -0.00004510, 0.00003346, -0.00005216, -0.00002766, 0.00002124, -0.00002218, 0.00002188, -0.00004351, 0.00004210, -0.00002515, 0.00001296, -0.00000394, 0.00000345, 0.00000302, 0.00000612, 0.00002234, -0.00001842, -0.00001760, 0.00001887, 0.00004370, -0.00003338, 0.00001597, -0.00001216, -0.00000933, 0.00001137, 0.00002243, -0.00003014, -0.00002407, 0.00003302, -0.00000007, -0.00001975, -0.00004359, 0.00002704, -0.00001684, 0.00000830, -0.00004799, 0.00002915, -0.00001167, 0.00000086, -0.00002961, -0.00005361, -0.00002296, },
		{ 0.00017840, 0.00000701, -0.00007895, -0.00003322, -0.00002211, -0.00000348, 0.00000104, -0.00001049, -0.00006854, 0.00000523, -0.00028618, 0.00023964, 0.00060390, -0.00027420, -0.00004430, -0.00004921, 0.00000599, -0.00004708, 0.00004112, -0.00003998, 0.00001596, -0.00002566, 0.00001255, -0.00001904, -0.00006060, 0.00003887, -0.00003185, 0.00002841, -0.00004503, 0.00003145, -0.00000856, -0.00000083, 0.00000578, -0.00000539, 0.00000982, -0.00000239, 0.00003062, -0.00001744, -0.00001552, 0.00000307, 0.00005382, -0.00002829, 0.00000850, -0.00000370, -0.00001423, 0.00000888, 0.00002266, -0.00001734, -0.00003526, 0.00002837, 0.00000601, -0.00001116, -0.00005316, 0.00002354, -0.00001115, 0.00000381, -0.00003737, 0.00000924, 0.00000828, -0.00001596, -0.00000947, -0.00006129, -0.00004377, 0.00013075, },
		{ 0.00003644, -0.00008269, -0.00004384, -0.00002033, -0.00001166, 0.00000625, -0.00000766, -0.00006098, -0.00000222, -0.00022033, 0.00005754, 0.00064397, -0.00014195, -0.00012131, -0.00002142, -0.00001443, -0.00003177, 0.00002176, -0.00001681, -0.00000688, -0.00000586, -0.00000934, 0.00001162, -0.00008338, 0.00004257, -0.00002886, 0.00002413, -0.00003611, 0.00001278, 0.00001041, -0.00001296, 0.00001307, -0.00001184, 0.00001442, -0.00000894, 0.00003548, -0.00001356, -0.00001404, -0.00000911, 0.00005556, -0.00001626, -0.00000212, 0.00000522, -0.00001780, 0.00000621, 0.00002085, -0.00000478, -0.00004150, 0.00002033, 0.00001034, -0.00000174, -0.00005757, 0.00001280, 0.00000018, -0.00000586, -0.00002225, -0.00001179, 0.00002519, -0.00002685, 0.00000077, -0.00005590, -0.00006633, 0.00009635, 0.00021528, },
		{ -0.00008759, -0.00005282, -0.00002146, -0.00001734, 0.00000767, -0.00000385, -0.00005180, -0.00001423, -0.00015745, -0.00007127, 0.00060252, 0.00002013, -0.00018333, -0.00000848, -0.00002666, -0.00002115, 0.00000534, 0.00000280, -0.00002449, 0.00000845, -0.00002589, 0.00003274, -0.00009038, 0.00003043, -0.00001456, 0.00001194, -0.00002074, -0.00000870, 0.00002666, -0.00001998, 0.00001608, -0.00001486, 0.00001617, -0.00001278, 0.00003614, -0.00000637, -0.00001410, -0.00001640, 0.00004940, 0.00000020, -0.00001306, 0.00001250, -0.00001917, 0.00000338, 0.00001731, 0.00000675, -0.00004346, 0.00001117, 0.00001170, 0.00000765, -0.00005624, -0.00000337, 0.00001346, -0.00001587, -0.00000855, -0.00002744, 0.00003307, -0.00002710, -0.00000123, -0.00003969, -0.00008585, 0.00005458, 0.00023883, 0.00008116, },
		{ -0.00006285, -0.00002364, -0.00002184, 0.00000644, 0.00000023, -0.00004227, -0.00002585, -0.00010646, -0.00014746, 0.00050500, 0.00017665, -0.00020848, -0.00001885, -0.00002677, -0.00001820, -0.00000516, 0.00001568, -0.00003399, 0.00001474, -0.00003416, 0.00004138, -0.00008122, 0.00000585, 0.00000554, -0.00000254, -0.00000440, -0.00002747, 0.00003599, -0.00002001, 0.00001448, -0.00001463, 0.00001529, -0.00001379, 0.00003255, 0.00000374, -0.00001559, -0.00001904, 0.00003745, 0.00001767, -0.00002150, 0.00001662, -0.00001808, 0.00000073, 0.00001236, 0.00001664, -0.00004188, 0.00000236, 0.00001036, 0.00001519, -0.00004872, -0.00002274, 0.00002515, -0.00002255, 0.00000001, -0.00003463, 0.00003053, -0.00001749, -0.00001191, -0.00001980, -0.00009575, 0.00000622, 0.00024557, 0.00013870, -0.00008892, },
		{ -0.00002490, -0.00002675, 0.00000418, 0.00000331, -0.00003249, -0.00003506, -0.00006955, -0.00018084, 0.00038043, 0.00029958, -0.00018778, -0.00004988, -0.00001854, -0.00002034, -0.00001113, 0.00002246, -0.00003596, 0.00001372, -0.00003457, 0.00003863, -0.00005995, -0.00002414, 0.00002408, -0.00001344, 0.00000810, -0.00003953, 0.00003640, -0.00001334, 0.00000947, -0.00001221, 0.00001249, -0.00001225, 0.00002521, 0.00001568, -0.00001762, -0.00001835, 0.00002266, 0.00003280, -0.00002543, 0.00001702, -0.00001494, -0.00000132, 0.00000658, 0.00002403, -0.00003697, -0.00000589, 0.00000800, 0.00001859, -0.00003505, -0.00004288, 0.00003263, -0.00002410, 0.00000249, -0.00003384, 0.00002017, -0.00000275, -0.00002447, -0.00000438, -0.00009149, -0.00004519, 0.00023221, 0.00020452, -0.00008081, -0.00007679, },
		{ -0.00003258, 0.00000163, 0.00000476, -0.00002270, -0.00004114, -0.00004539, -0.00018433, 0.00025358, 0.00037397, -0.00012637, -0.00008973, -0.00001061, -0.00002217, -0.00001597, 0.00002553, -0.00003277, 0.00000784, -0.00002952, 0.00002812, -0.00003298, -0.00005167, 0.00003470, -0.00001690, 0.00001427, -0.00004341, 0.00002843, -0.00000216, 0.00000325, -0.00000905, 0.00000857, -0.00000864, 0.00001507, 0.00002783, -0.00001878, -0.00001609, 0.00000796, 0.00004308, -0.00002408, 0.00001409, -0.00001052, -0.00000264, 0.00000102, 0.00002778, -0.00002852, -0.00001425, 0.00000678, 0.00001628, -0.00001640, -0.00006099, 0.00003408, -0.00002036, -0.00000005, -0.00002765, 0.00000641, 0.00001143, -0.00003270, 0.00000138, -0.00007367, -0.00009236, 0.00019564, 0.00027255, -0.00005768, -0.00009609, -0.00002461, },
		{ -0.00000164, 0.00000503, -0.00001374, -0.00004369, -0.00003152, -0.00016972, 0.00014096, 0.00039837, -0.00003977, -0.00012291, -0.00001151, -0.00001933, -0.00002212, 0.00002670, -0.00002650, -0.00000043, -0.00002179, 0.00001407, -0.00000662, -0.00007055, 0.00003410, -0.00001240, 0.00001451, -0.00004020, 0.00001468, 0.00001026, -0.00000183, -0.00000632, 0.00000411, -0.00000345, 0.00000344, 0.00003835, -0.00001762, -0.00001396, -0.00000437, 0.00004727, -0.00001788, 0.00000889, -0.00000559, -0.00000351, -0.00000295, 0.00002679, -0.00001640, -0.00002328, 0.00000804, 0.00000843, 0.00000436, -0.00007379, 0.00002828, -0.00001200, -0.00000586, -0.00001895, -0.00000693, 0.00002122, -0.00003383, -0.00000249, -0.00004845, -0.00012562, 0.00013467, 0.00033489, -0.00001495, -0.00012011, -0.00002394, -0.00003840, },
		},
		{ { -0.00001611, -0.00001062, -0.00000111, -0.00001313, -0.00002935, -0.00002704, -0.00012869, 0.00005405, 0.00036947, 0.00005429, -0.00015350, -0.00001272, -0.00001360, -0.00000336, -0.00003330, -0.00000555, -0.00000959, -0.00000769, 0.00005254, -0.00000997, -0.00001581, -0.00001491, 0.00001438, 0.00000427, -0.00004299, 0.00000443, -0.00001059, -0.00000082, -0.00000036, -0.00000055, 0.00001995, -0.00000303, -0.00000349, -0.00000303, 0.00001995, -0.00000055, -0.00000036, -0.00000082, -0.00001059, 0.00000443, -0.00004299, 0.00000427, 0.00001438, -0.00001491, -0.00001581, -0.00000997, 0.00005254, -0.00000769, -0.00000959, -0.00000555, -0.00003330, -0.00000336, -0.00001360, -0.00001272, -0.00015350, 0.00005429, 0.00036947, 0.00005405, -0.00012869, -0.00002704, -0.00002935, -0.00001313, -0.00000111, -0.00001062, },
		{ -0.00001108, -0.00000430, -0.00000739, -0.00002784, -0.00002086, -0.00010403, -0.00002056, 0.00030792, 0.00016605, -0.00014214, -0.00004477, -0.00000582, -0.00000635, -0.00002323, -0.00002011, 0.00000231, -0.00002245, 0.00005038, 0.00000491, -0.00001863, -0.00001572, 0.00000671, 0.00001515, -0.00003892, -0.00000922, 0.00000035, -0.00000879, 0.00000527, -0.00000553, 0.00001561, 0.00000515, -0.00000494, -0.00000519, 0.00001387, 0.00001329, -0.00001098, 0.00000973, -0.00002006, 0.00001074, -0.00003679, -0.00001240, 0.00002230, -0.00001473, -0.00001068, -0.00002216, 0.00004682, 0.00001059, -0.00001939, 0.00000332, -0.00003446, -0.00000819, -0.00001419, -0.00000132, -0.00012202, -0.00006972, 0.00037995, 0.00016643, -0.00014862, -0.00003919, -0.00002879, -0.00002000, -0.00000009, -0.00000772, -0.00001698, },
		{ -0.00000750, -0.00000407, -0.00002468, -0.00001764, -0.00008136, -0.00006209, 0.00022442, 0.00023864, -0.00008778, -0.00008471, -0.00000034, -0.00001075, -0.00001125, -0.00003242, 0.00000874, -0.00002790, 0.00003827, 0.00002231, -0.00002016, -0.00001530, -0.00000166, 0.00002181, -0.00002976, -0.00002162, 0.00000414, -0.00000778, 0.00000182, -0.00000040, 0.00000341, 0.00001534, -0.00000694, -0.00000298, 0.00000197, 0.00002774, -0.00001698, 0.00001307, -0.00001886, 0.00000364, -0.00001903, -0.00003322, 0.00002944, -0.00001532, -0.00000396, -0.00003059, 0.00003529, 0.00002729, -0.00002197, 0.00000285, -0.00002527, -0.00002074, -0.00000693, -0.00000823, -0.00006798, -0.00016715, 0.00031614, 0.00030796, -0.00014960, -0.00006206, -0.00002562, -0.00002702, -0.00000220, -0.00000274, -0.00001853, -0.00001067, },
		{ -0.00000371, -0.00002061, -0.00001566, -0.00006270, -0.00008024, 0.00014211, 0.00026408, -0.00000791, -0.00011274, -0.00000720, -0.00001071, -0.00000368, -0.00003639, 0.00000528, -0.00002225, 0.00001785, 0.00003978, -0.00001997, -0.00001342, -0.00001034, 0.00002523, -0.00001962, -0.00002780, -0.00000146, 0.00000075, -0.00000676, 0.00000854, -0.00000828, 0.00002007, -0.00000459, -0.00000135, -0.00000780, 0.00003312, -0.00001024, 0.00000355, -0.00000452, -0.00001475, 0.00000403, -0.00005116, 0.00003101, -0.00001324, 0.00000015, -0.00003179, 0.00001793, 0.00004084, -0.00001778, -0.00000368, -0.00001196, -0.00003299, -0.00000066, -0.00001858, -0.00002158, -0.00020466, 0.00017663, 0.00044966, -0.00010965, -0.00010119, -0.00001898, -0.00003433, -0.00000634, 0.00000179, -0.00001814, -0.00001177, -0.00000870, },
		{ -0.00001650, -0.00001429, -0.00004749, -0.00008462, 0.00007416, 0.00024957, 0.00007373, -0.00011434, -0.00002810, -0.00000688, -0.00000120, -0.00003195, -0.00000604, -0.00000996, -0.00000430, 0.00005163, -0.00001497, -0.00001260, -0.00001660, 0.00002433, -0.00000928, -0.00002836, -0.00001181, 0.00000917, -0.00001207, 0.00001247, -0.00001107, 0.00001393, 0.00000472, -0.00000362, -0.00001004, 0.00002513, 0.00000945, -0.00001511, 0.00001562, -0.00003352, 0.00002021, -0.00005633, 0.00002072, -0.00000364, -0.00000354, -0.00002223, -0.00000551, 0.00005171, -0.00001074, -0.00000922, -0.00000383, -0.00003485, -0.00000590, -0.00001713, -0.00000498, -0.00017338, -0.00000794, 0.00054381, -0.00000549, -0.00015838, -0.00000974, -0.00004249, -0.00001014, 0.00000246, -0.00001325, -0.00001598, -0.00000715, -0.00000575, },
		{ -0.00001393, -0.00003403, -0.00008273, 0.00002533, 0.00020909, 0.00013857, -0.00008786, -0.00005423, -0.00000609, 0.00000022, -0.00002386, -0.00001861, 0.00000091, -0.00001973, 0.00005185, -0.00000206, -0.00001562, -0.00001763, 0.00001775, 0.00000209, -0.00002682, -0.00001990, 0.00000984, -0.00000806, 0.00000681, -0.00000260, -0.00000173, 0.00001785, -0.00000812, -0.00000558, 0.00000778, 0.00003325, -0.00003116, 0.00002823, -0.00003914, 0.00001851, -0.00004255, -0.00000349, 0.00001457, -0.00001609, -0.00000314, -0.00003150, 0.00005834, -0.00000325, -0.00000934, -0.00000565, -0.00002298, -0.00002451, -0.00000099, -0.00001846, -0.00009686, -0.00018025, 0.00053941, 0.00017283, -0.00022332, -0.00000438, -0.00004971, -0.00001262, -0.00000223, -0.00000369, -0.00002261, -0.00000393, -0.00000877, -0.00001294, },
		{ -0.00002090, -0.00007908, -0.00000532, 0.00015673, 0.00017803, -0.00004281, -0.00007242, -0.00001476, 0.00000376, -0.00001651, -0.00002664, 0.00000440, -0.00002360, 0.00003892, 0.00001780, -0.00002178, -0.00001399, 0.00000693, 0.00001370, -0.00002532, -0.00002194, 0.00000086, 0.00000408, -0.00000520, 0.00001158, -0.00001876, 0.00002688, -0.00000928, -0.00000016, -0.00000939, 0.00004847, -0.00003213, 0.00002287, -0.00002461, -0.00000271, -0.00001366, -0.00003492, 0.00003489, -0.00003108, 0.00001699, -0.00005080, 0.00005504, 0.00000693, -0.00000621, -0.00001322, -0.00000493, -0.00004579, 0.00001695, -0.00004069, -0.00002043, -0.00027930, 0.00040816, 0.00040474, -0.00026657, -0.00001777, -0.00004974, -0.00001672, -0.00000959, 0.00000673, -0.00002793, -0.00000205, -0.00001070, -0.00001029, -0.00001521, },
		{ -0.00007440, -0.00002272, 0.00010443, 0.00019121, 0.00000818, -0.00007455, -0.00003180, 0.00000675, -0.00000962, -0.00002944, 0.00000057, -0.00001803, 0.00001822, 0.00003793, -0.00002563, -0.00001053, -0.00000321, 0.00002204, -0.00002259, -0.00001958, -0.00001274, 0.00001652, -0.00001486, 0.00002177, -0.00002705, 0.00002419, -0.00000226, 0.00000061, -0.00001784, 0.00004656, -0.00001340, -0.00000022, 0.00000452, -0.00003261, 0.00001614, -0.00005932, 0.00004507, -0.00003691, 0.00002555, -0.00005220, 0.00003514, 0.00002429, -0.00000631, -0.00001662, 0.00000547, -0.00005303, 0.00001752, -0.00004438, 0.00001409, -0.00027142, 0.00016805, 0.00063123, -0.00024205, -0.00006910, -0.00003396, -0.00002846, -0.00001327, 0.00001115, -0.00002642, -0.00000527, -0.00000933, -0.00000921, -0.00001774, -0.00000864, },
		{ -0.00003423, 0.00006182, 0.00018040, 0.00005666, -0.00006256, -0.00004755, 0.00000160, 0.00000076, -0.00003058, -0.00000547, -0.00001028, -0.00000070, 0.00004911, -0.00002019, -0.00001305, -0.00000739, 0.00002330, -0.00001591, -0.00001741, -0.00002305, 0.00002038, -0.00001440, 0.00002097, -0.00002121, 0.00000812, 0.00001229, -0.00000399, -0.00001549, 0.00002852, 0.00001805, -0.00002891, 0.00003225, -0.00005299, 0.00002946, -0.00006268, 0.00003469, -0.00002452, 0.00001396, -0.00003042, -0.00000198, 0.00004903, -0.00001324, -0.00000892, -0.00000104, -0.00003652, -0.00000879, -0.00001699, -0.00000646, -0.00017133, -0.00011177, 0.00076717, -0.00010471, -0.00016855, 0.00000034, -0.00005128, -0.00000809, 0.00000411, -0.00001441, -0.00001560, -0.00000360, -0.00001043, -0.00001958, -0.00000018, -0.00006606, },
		{ 0.00003502, 0.00014936, 0.00009950, -0.00004508, -0.00005068, -0.00001710, 0.00001635, -0.00003265, -0.00000903, -0.00000684, -0.00001040, 0.00004575, -0.00000283, -0.00002302, -0.00000457, 0.00001730, -0.00000492, -0.00001830, -0.00002523, 0.00001215, -0.00000291, 0.00000984, -0.00000427, -0.00001491, 0.00002731, -0.00000900, -0.00000732, 0.00000395, 0.00004730, -0.00004604, 0.00004159, -0.00004942, 0.00001678, -0.00004106, 0.00000380, 0.00000460, -0.00001471, 0.00000732, -0.00004562, 0.00007219, -0.00002188, 0.00000655, -0.00002062, -0.00000309, -0.00005139, 0.00002795, -0.00005811, -0.00003827, -0.00033398, 0.00073551, 0.00015999, -0.00029920, 0.00003955, -0.00007864, 0.00000353, -0.00001290, 0.00000549, -0.00003022, 0.00000454, -0.00001351, -0.00001873, 0.00000113, -0.00005077, -0.00004663, },
		{ 0.00010520, 0.00013420, -0.00002867, -0.00003783, -0.00004500, 0.00003117, -0.00003254, -0.00001068, -0.00000827, -0.00001068, 0.00003049, 0.00002110, -0.00003471, -0.00000005, 0.00000909, 0.00000657, -0.00002066, -0.00002067, -0.00000390, 0.00001272, -0.00000382, 0.00001427, -0.00003380, 0.00003332, -0.00000769, -0.00000095, -0.00001530, 0.00006038, -0.00003998, 0.00002559, -0.00002114, -0.00001603, -0.00000537, -0.00003433, 0.00003662, -0.00004509, 0.00004223, -0.00007606, 0.00007790, -0.00001986, 0.00001705, -0.00003738, 0.00002617, -0.00008377, 0.00005835, -0.00009514, 0.00005416, -0.00041708, 0.00050983, 0.00050980, -0.00040671, 0.00005035, -0.00009161, 0.00000908, -0.00002919, 0.00002300, -0.00004074, 0.00000937, -0.00001560, -0.00001534, -0.00000521, -0.00002859, -0.00006195, 0.00002387, },
		{ 0.00015439, -0.00001147, -0.00001808, -0.00006709, 0.00003251, -0.00002245, -0.00001586, -0.00000924, -0.00000821, 0.00001297, 0.00004040, -0.00003797, -0.00000239, 0.00000592, 0.00001288, -0.00002012, -0.00001504, -0.00001894, 0.00002248, -0.00001060, 0.00002479, -0.00003945, 0.00002463, 0.00000288, -0.00000070, -0.00002241, 0.00005236, -0.00001197, -0.00000758, 0.00001724, -0.00004971, 0.00002348, -0.00005951, 0.00005262, -0.00005775, 0.00005371, -0.00007501, 0.00005335, 0.00000312, 0.00001009, -0.00003467, 0.00002983, -0.00008084, 0.00004587, -0.00008082, 0.00005905, -0.00033817, 0.00014168, 0.00084206, -0.00040991, -0.00000838, -0.00006759, -0.00000795, -0.00002990, 0.00002465, -0.00003680, 0.00000383, -0.00001219, -0.00001255, -0.00001508, -0.00000521, -0.00007537, 0.00002063, 0.00006013, },
		{ 0.00001491, -0.00000738, -0.00006627, 0.00000973, 0.00000282, -0.00002880, -0.00000430, -0.00001048, 0.00000307, 0.00004513, -0.00002538, -0.00001654, 0.00001132, 0.00001141, -0.00001404, -0.00001322, -0.00002590, 0.00002013, -0.00000626, 0.00002371, -0.00003015, 0.00000348, 0.00001884, -0.00000457, -0.00001881, 0.00002942, 0.00002454, -0.00003906, 0.00004398, -0.00006281, 0.00002715, -0.00005731, 0.00004095, -0.00004204, 0.00003256, -0.00003829, 0.00000028, 0.00004446, -0.00001490, -0.00000874, 0.00000236, -0.00003798, -0.00001182, -0.00001378, -0.00001638, -0.00014931, -0.00024809, 0.00102456, -0.00023630, -0.00015816, 0.00000101, -0.00005428, -0.00000735, 0.00000334, -0.00001374, -0.00001500, -0.00000073, -0.00001408, -0.00002173, 0.00001026, -0.00007789, 0.00001326, 0.00002865, 0.00015106, },
		{ -0.00001590, -0.00003654, -0.00003488, 0.00003757, -0.00004560, 0.00000545, -0.00001871, 0.00000352, 0.00003402, 0.00000107, -0.00003838, 0.00002081, 0.00000604, -0.00000483, -0.00001519, -0.00002404, 0.00000728, 0.00000545, 0.00001524, -0.00001205, -0.00002108, 0.00003145, -0.00000635, -0.00001146, 0.00000393, 0.00005223, -0.00005079, 0.00004370, -0.00004547, 0.00000228, -0.00002948, 0.00000639, -0.00000447, -0.00001141, 0.00001825, -0.00006106, 0.00008525, -0.00004232, 0.00002574, -0.00003942, 0.00002245, -0.00008515, 0.00006816, -0.00011489, 0.00004607, -0.00051712, 0.00095387, 0.00013028, -0.00037112, 0.00008885, -0.00011398, 0.00002845, -0.00003294, 0.00002064, -0.00004068, 0.00001509, -0.00002004, -0.00002079, 0.00001125, -0.00006312, -0.00000695, 0.00002017, 0.00011978, 0.00005599, },
		{ 0.00000908, -0.00008192, 0.00006360, -0.00005319, 0.00001084, -0.00002578, 0.00000762, 0.00001607, 0.00002950, -0.00005541, 0.00002348, 0.00000534, 0.00000122, -0.00001646, -0.00001851, -0.00000820, 0.00001568, 0.00000761, 0.00000549, -0.00003866, 0.00003277, -0.00000097, -0.00000679, -0.00001361, 0.00006002, -0.00003631, 0.00001716, -0.00000644, -0.00003595, 0.00000548, -0.00003127, 0.00003385, -0.00005368, 0.00006540, -0.00010035, 0.00009904, -0.00004873, 0.00004397, -0.00006592, 0.00006341, -0.00012902, 0.00011089, -0.00016438, 0.00014676, -0.00056837, 0.00061415, 0.00061617, -0.00055905, 0.00013626, -0.00014872, 0.00004974, -0.00006115, 0.00004640, -0.00005766, 0.00002466, -0.00002459, -0.00001445, -0.00000069, -0.00003389, -0.00003831, 0.00003184, 0.00006826, 0.00010504, -0.00003786, },
		{ -0.00010324, 0.00005934, -0.00003725, 0.00000103, -0.00002171, 0.00000469, 0.00000453, 0.00004501, -0.00005424, 0.00000946, 0.00001596, -0.00000066, -0.00001274, -0.00001537, -0.00001813, 0.00001701, 0.00000659, 0.00001625, -0.00004358, 0.00002085, 0.00001106, -0.00000604, -0.00002018, 0.00004851, -0.00000414, -0.00001917, 0.00003173, -0.00006258, 0.00002347, -0.00004936, 0.00005028, -0.00006965, 0.00007713, -0.00009459, 0.00006881, -0.00001941, 0.00002863, -0.00005495, 0.00005728, -0.00011197, 0.00007986, -0.00012346, 0.00010855, -0.00040076, 0.00010242, 0.00106490, -0.00059791, 0.00007422, -0.00011727, 0.00002612, -0.00005569, 0.00004194, -0.00004939, 0.00001603, -0.00001897, -0.00001050, -0.00001577, -0.00000273, -0.00006794, 0.00004814, 0.00001667, 0.00014251, -0.00005242, 0.00004333, },
		{ 0.00001529, 0.00000532, -0.00002611, -0.00000315, -0.00001053, 0.00000704, 0.00003986, -0.00003036, -0.00002155, 0.00003584, -0.00000836, -0.00000475, -0.00001629, -0.00001977, 0.00000870, 0.00001132, 0.00002046, -0.00003729, 0.00000076, 0.00002369, -0.00000532, -0.00001980, 0.00002750, 0.00002849, -0.00004362, 0.00004681, -0.00005886, 0.00001087, -0.00003784, 0.00003595, -0.00005058, 0.00004690, -0.00004362, 0.00000098, 0.00003807, -0.00001547, -0.00000871, 0.00000570, -0.00003874, -0.00001465, -0.00000903, -0.00003189, -0.00011342, -0.00040098, 0.00129291, -0.00038602, -0.00013059, -0.00000753, -0.00005125, -0.00000796, 0.00000009, -0.00001124, -0.00001422, 0.00000089, -0.00001582, -0.00002244, 0.00001507, -0.00007949, 0.00004958, -0.00001193, 0.00014660, -0.00003695, 0.00004142, -0.00007942, },
		{ 0.00005861, -0.00005792, 0.00002085, -0.00003175, 0.00001922, 0.00001956, 0.00000694, -0.00005724, 0.00005285, -0.00001244, 0.00000112, -0.00001740, -0.00001697, -0.00000321, 0.00001543, 0.00002293, -0.00002573, -0.00001915, 0.00002954, 0.00000003, -0.00001827, 0.00000819, 0.00004674, -0.00004289, 0.00003109, -0.00002507, -0.00002520, -0.00000711, 0.00000238, -0.00000957, -0.00000748, 0.00002724, -0.00007409, 0.00009419, -0.00006111, 0.00004502, -0.00005914, 0.00004978, -0.00012038, 0.00011246, -0.00017958, 0.00014312, -0.00070931, 0.00117431, 0.00008938, -0.00043446, 0.00013861, -0.00015238, 0.00005998, -0.00006096, 0.00004098, -0.00005386, 0.00002766, -0.00002899, -0.00001638, 0.00001200, -0.00006488, 0.00002579, -0.00000580, 0.00010879, 0.00001546, -0.00000091, -0.00001548, -0.00005465, },
		{ -0.00007162, 0.00003251, -0.00004392, 0.00002714, -0.00000016, 0.00003894, -0.00007839, 0.00005127, -0.00000214, -0.00000229, -0.00001316, -0.00001607, -0.00001086, 0.00001254, 0.00002774, -0.00001392, -0.00003279, 0.00002524, 0.00001056, -0.00001768, -0.00000410, 0.00004658, -0.00001973, -0.00000371, 0.00001878, -0.00006033, 0.00001806, -0.00002596, 0.00002694, -0.00005443, 0.00008069, -0.00011775, 0.00011366, -0.00007485, 0.00007001, -0.00009426, 0.00010083, -0.00017404, 0.00016559, -0.00023795, 0.00024454, -0.00071763, 0.00071270, 0.00071558, -0.00070914, 0.00022974, -0.00021459, 0.00010092, -0.00010255, 0.00007518, -0.00007795, 0.00004350, -0.00003849, -0.00000477, -0.00000567, -0.00003157, -0.00001617, 0.00002651, 0.00004306, 0.00008711, -0.00006068, 0.00005716, -0.00011580, 0.00009257, },
		{ 0.00001756, -0.00003405, 0.00001766, -0.00000408, 0.00004894, -0.00007071, 0.00002269, 0.00002565, -0.00001591, -0.00000246, -0.00001987, -0.00001086, 0.00000183, 0.00003378, -0.00000231, -0.00003947, 0.00001350, 0.00002155, -0.00001507, -0.00001141, 0.00003502, 0.00001022, -0.00003443, 0.00004556, -0.00006854, 0.00001543, -0.00002879, 0.00003717, -0.00006857, 0.00008942, -0.00010592, 0.00007805, -0.00003986, 0.00004621, -0.00007417, 0.00008324, -0.00014098, 0.00011367, -0.00016651, 0.00015696, -0.00045544, 0.00005780, 0.00127796, -0.00078518, 0.00016680, -0.00017622, 0.00006938, -0.00008779, 0.00006140, -0.00006340, 0.00003075, -0.00003005, -0.00000140, -0.00002161, 0.00000084, -0.00005512, 0.00006008, -0.00001932, 0.00014301, -0.00009899, 0.00009656, -0.00013078, 0.00008006, -0.00004895, },
		{ -0.00000265, -0.00001039, 0.00001119, 0.00003355, -0.00003499, -0.00002608, 0.00006100, -0.00003092, 0.00000858, -0.00002470, -0.00000683, -0.00001071, 0.00003499, 0.00001227, -0.00004230, 0.00000038, 0.00002647, -0.00000618, -0.00001862, 0.00002288, 0.00002980, -0.00004179, 0.00003882, -0.00004024, -0.00001810, -0.00000526, 0.00001909, -0.00004662, 0.00005257, -0.00004466, 0.00000001, 0.00003011, -0.00001479, -0.00000888, 0.00000824, -0.00003825, -0.00001684, -0.00000472, -0.00004874, -0.00007258, -0.00055014, 0.00154415, -0.00053477, -0.00009288, -0.00002286, -0.00004323, -0.00000972, -0.00000529, -0.00000727, -0.00001332, 0.00000095, -0.00001505, -0.00002150, 0.00001346, -0.00006958, 0.00006908, -0.00004758, 0.00015146, -0.00008473, 0.00007555, -0.00008320, 0.00001650, 0.00000849, -0.00002207, },
		{ -0.00004293, 0.00003370, 0.00000561, 0.00001162, -0.00007415, 0.00008424, -0.00003262, 0.00000993, -0.00002317, -0.00000623, -0.00001659, 0.00002535, 0.00003181, -0.00004343, -0.00000986, 0.00002236, 0.00000921, -0.00002709, 0.00001594, 0.00003175, -0.00002284, 0.00000398, 0.00000981, -0.00006225, 0.00002410, -0.00000828, -0.00000852, -0.00000444, 0.00003242, -0.00008163, 0.00009633, -0.00007505, 0.00006117, -0.00007673, 0.00007472, -0.00015108, 0.00015251, -0.00024092, 0.00023685, -0.00088672, 0.00137363, 0.00004528, -0.00048546, 0.00018342, -0.00018922, 0.00009407, -0.00009354, 0.00006417, -0.00006853, 0.00004125, -0.00003986, -0.00000465, 0.00000160, -0.00005302, 0.00004401, -0.00003060, 0.00010483, -0.00001586, 0.00000093, 0.00000825, -0.00007179, 0.00007462, -0.00006510, 0.00003323, },
		{ 0.00004374, -0.00001389, 0.00004517, -0.00009757, 0.00007701, -0.00001024, -0.00000444, -0.00001096, -0.00001375, -0.00001160, 0.00000480, 0.00005263, -0.00004035, -0.00001814, 0.00001258, 0.00002512, -0.00003209, 0.00001143, 0.00002216, 0.00000791, -0.00003502, 0.00005089, -0.00008547, 0.00003114, -0.00001901, 0.00001751, -0.00004698, 0.00008420, -0.00012395, 0.00011813, -0.00009400, 0.00009085, -0.00011787, 0.00013236, -0.00021095, 0.00021267, -0.00030410, 0.00033310, -0.00084813, 0.00079669, 0.00079946, -0.00084103, 0.00031797, -0.00027984, 0.00015530, -0.00014781, 0.00010570, -0.00009944, 0.00006412, -0.00005631, 0.00001399, -0.00002118, -0.00001929, -0.00000156, 0.00001571, 0.00002623, 0.00007640, -0.00008721, 0.00009641, -0.00013838, 0.00011084, -0.00008181, 0.00004931, -0.00005863, },
		{ -0.00001001, 0.00004950, -0.00008356, 0.00003472, 0.00003372, -0.00002973, 0.00000783, -0.00002647, 0.00000026, -0.00001862, 0.00006525, -0.00002693, -0.00002901, 0.00000432, 0.00003275, -0.00002635, 0.00000203, 0.00001403, 0.00002934, -0.00005154, 0.00005607, -0.00006631, 0.00000095, -0.00000029, 0.00001544, -0.00005378, 0.00008682, -0.00010581, 0.00007882, -0.00005532, 0.00005998, -0.00008925, 0.00010336, -0.00016253, 0.00014135, -0.00020388, 0.00019846, -0.00049902, 0.00001650, 0.00145806, -0.00094855, 0.00025445, -0.00023483, 0.00011497, -0.00012150, 0.00008035, -0.00007716, 0.00004636, -0.00004435, 0.00001447, -0.00003265, 0.00000651, -0.00004012, 0.00005858, -0.00004328, 0.00014448, -0.00013741, 0.00013128, -0.00014270, 0.00008919, -0.00005344, 0.00003050, -0.00004372, 0.00002791, },
		{ 0.00002654, -0.00003911, -0.00002813, 0.00008140, -0.00005044, 0.00002145, -0.00003491, 0.00000882, -0.00003281, 0.00005981, 0.00000113, -0.00004511, 0.00000241, 0.00002714, -0.00000710, -0.00001580, 0.00001604, 0.00002871, -0.00003410, 0.00002063, -0.00000978, -0.00005478, 0.00003606, -0.00000680, -0.00002980, 0.00004715, -0.00004053, -0.00000253, 0.00002104, -0.00001288, -0.00000925, 0.00000936, -0.00003608, -0.00001804, -0.00000284, -0.00006244, -0.00003653, -0.00067443, 0.00174948, -0.00066183, -0.00005422, -0.00004094, -0.00003221, -0.00001218, -0.00001210, -0.00000249, -0.00001237, -0.00000052, -0.00001158, -0.00001896, 0.00000594, -0.00004940, 0.00006989, -0.00006942, 0.00015264, -0.00011737, 0.00008997, -0.00007584, 0.00001249, 0.00001241, -0.00001642, -0.00000328, -0.00001006, 0.00001544, },
		{ 0.00001241, -0.00008414, 0.00010779, -0.00004864, 0.00001784, -0.00002978, 0.00000505, -0.00002956, 0.00003364, 0.00004008, -0.00006200, 0.00000443, 0.00001197, 0.00001935, -0.00003644, 0.00002531, 0.00001024, 0.00000587, -0.00003366, 0.00005317, -0.00010229, 0.00005983, -0.00002351, -0.00000053, -0.00000386, 0.00003295, -0.00008193, 0.00009031, -0.00008189, 0.00007159, -0.00008956, 0.00009353, -0.00017187, 0.00018057, -0.00028802, 0.00031134, -0.00102604, 0.00152911, 0.00000666, -0.00052147, 0.00021808, -0.00021934, 0.00012554, -0.00012580, 0.00008672, -0.00008263, 0.00005411, -0.00005110, 0.00001322, -0.00001870, -0.00002834, 0.00004447, -0.00004533, 0.00009860, -0.00003416, -0.00000774, 0.00003053, -0.00008270, 0.00008210, -0.00006399, 0.00003960, -0.00004983, 0.00004435, -0.00000496, },
		{ -0.00010775, 0.00009488, -0.00001589, -0.00000597, -0.00000986, -0.00001182, -0.00001079, -0.00000472, 0.00007672, -0.00006831, 0.00000141, -0.00000196, 0.00003980, -0.00004730, 0.00002951, -0.00000866, 0.00004179, -0.00007172, 0.00008388, -0.00010673, 0.00004344, -0.00001116, 0.00000812, -0.00003327, 0.00007492, -0.00011725, 0.00011091, -0.00010340, 0.00010327, -0.00013315, 0.00015303, -0.00023319, 0.00024359, -0.00035209, 0.00039903, -0.00094443, 0.00085810, 0.00086005, -0.00093964, 0.00038781, -0.00033401, 0.00020406, -0.00018983, 0.00013316, -0.00011891, 0.00008358, -0.00007536, 0.00003953, -0.00004516, 0.00000206, 0.00000251, 0.00000576, 0.00001450, 0.00006912, -0.00010906, 0.00011971, -0.00014451, 0.00011397, -0.00007993, 0.00005767, -0.00006687, 0.00005416, -0.00002172, 0.00004488, },
		{ 0.00004331, 0.00003908, -0.00003925, 0.00001516, -0.00003247, 0.00001152, -0.00003848, 0.00009390, -0.00005218, -0.00001535, -0.00000427, 0.00004250, -0.00003769, 0.00001712, -0.00001073, 0.00004996, -0.00006774, 0.00006090, -0.00005634, -0.00001628, 0.00003161, -0.00000998, -0.00002827, 0.00006913, -0.00009374, 0.00007059, -0.00006396, 0.00006792, -0.00009795, 0.00011435, -0.00017254, 0.00015799, -0.00023027, 0.00022799, -0.00052893, -0.00001337, 0.00158428, -0.00106638, 0.00032209, -0.00028259, 0.00015495, -0.00015125, 0.00009559, -0.00008854, 0.00006054, -0.00005956, 0.00003505, -0.00004722, 0.00001405, -0.00002581, 0.00004727, -0.00005345, 0.00013874, -0.00015820, 0.00014194, -0.00013587, 0.00008413, -0.00004852, 0.00003771, -0.00004894, 0.00003342, -0.00001130, 0.00004491, -0.00008976, },
		{ 0.00009261, -0.00006189, 0.00002953, -0.00004304, 0.00002243, -0.00005170, 0.00008041, -0.00001049, -0.00004586, 0.00000660, 0.00002592, -0.00000800, -0.00001209, 0.00000779, 0.00002573, -0.00002226, -0.00000481, 0.00002648, -0.00009127, 0.00007921, -0.00003588, -0.00000316, 0.00003075, -0.00003155, -0.00000629, 0.00001152, -0.00001000, -0.00000982, 0.00000870, -0.00003206, -0.00001805, -0.00000477, -0.00006931, -0.00001366, -0.00075558, 0.00188400, -0.00074849, -0.00002389, -0.00005715, -0.00002066, -0.00001475, -0.00001943, 0.00000228, -0.00001143, -0.00000321, -0.00000560, -0.00001518, -0.00000577, -0.00002252, 0.00005415, -0.00007341, 0.00014188, -0.00012879, 0.00008382, -0.00005803, 0.00000357, 0.00001691, -0.00000924, -0.00000522, -0.00000953, 0.00001958, 0.00001931, -0.00004250, -0.00002632, },
		{ -0.00005586, 0.00002159, -0.00003453, 0.00001370, -0.00003897, 0.00003909, 0.00004572, -0.00007785, 0.00002044, 0.00000116, 0.00002814, -0.00004439, 0.00003371, -0.00001341, 0.00003757, -0.00007494, 0.00009606, -0.00013682, 0.00009430, -0.00003939, 0.00001341, -0.00000695, 0.00002912, -0.00007497, 0.00007651, -0.00008085, 0.00007482, -0.00009593, 0.00010363, -0.00017913, 0.00019111, -0.00031254, 0.00035391, -0.00110833, 0.00162215, -0.00001878, -0.00054093, 0.00023840, -0.00023806, 0.00014915, -0.00015249, 0.00010470, -0.00009366, 0.00006410, -0.00006041, 0.00003432, -0.00004505, 0.00000482, 0.00002863, -0.00004579, 0.00008459, -0.00003860, -0.00002312, 0.00004798, -0.00008545, 0.00007963, -0.00005346, 0.00003851, -0.00005108, 0.00004950, -0.00001028, 0.00000780, -0.00008388, 0.00011815, },
		{ -0.00000748, -0.00000974, -0.00001044, -0.00000919, -0.00001283, 0.00009432, -0.00009182, 0.00002111, -0.00001458, 0.00005128, -0.00006002, 0.00004582, -0.00003982, 0.00007521, -0.00010684, 0.00011053, -0.00011953, 0.00005392, -0.00000302, 0.00000170, -0.00001719, 0.00005488, -0.00009895, 0.00009298, -0.00010229, 0.00010571, -0.00013815, 0.00015989, -0.00023685, 0.00025277, -0.00037431, 0.00043234, -0.00099482, 0.00089084, 0.00089154, -0.00099312, 0.00042821, -0.00036764, 0.00023867, -0.00022141, 0.00015257, -0.00013283, 0.00009848, -0.00009196, 0.00006758, -0.00007304, 0.00002864, -0.00000371, 0.00000063, 0.00000531, 0.00006225, -0.00012060, 0.00012403, -0.00013306, 0.00010115, -0.00006517, 0.00005619, -0.00006721, 0.00005665, -0.00002187, 0.00003659, -0.00010588, 0.00010102, -0.00001694, },
		{ 0.00001737, -0.00003555, 0.00001939, -0.00005216, 0.00011307, -0.00007258, -0.00000206, -0.00000977, 0.00004864, -0.00004663, 0.00003034, -0.00003467, 0.00006801, -0.00007994, 0.00005838, -0.00004076, -0.00003123, 0.00006147, -0.00003255, 0.00000183, 0.00003976, -0.00007197, 0.00005475, -0.00006535, 0.00006917, -0.00009929, 0.00011465, -0.00016902, 0.00016070, -0.00024209, 0.00024187, -0.00054329, -0.00002595, 0.00164133, -0.00112234, 0.00035752, -0.00031040, 0.00018201, -0.00017174, 0.00010407, -0.00009535, 0.00007073, -0.00007269, 0.00005699, -0.00006244, 0.00002220, -0.00001435, 0.00003067, -0.00005099, 0.00012188, -0.00015746, 0.00012851, -0.00011156, 0.00006564, -0.00003436, 0.00003851, -0.00004901, 0.00003335, -0.00000729, 0.00003480, -0.00008776, 0.00004708, 0.00004098, -0.00004252, },
		{ -0.00004612, 0.00003001, -0.00006233, 0.00009223, -0.00002021, -0.00004487, 0.00001244, 0.00002323, -0.00000883, -0.00000832, -0.00000093, 0.00002156, -0.00000880, -0.00003279, 0.00006045, -0.00011858, 0.00011655, -0.00006054, 0.00002733, 0.00000601, -0.00001918, -0.00001073, 0.00000236, -0.00000657, -0.00001056, 0.00000622, -0.00002634, -0.00001686, -0.00001094, -0.00006744, -0.00000915, -0.00078150, 0.00193082, -0.00078150, -0.00000915, -0.00006744, -0.00001094, -0.00001686, -0.00002634, 0.00000622, -0.00001056, -0.00000657, 0.00000236, -0.00001073, -0.00001918, 0.00000601, 0.00002733, -0.00006054, 0.00011655, -0.00011858, 0.00006045, -0.00003279, -0.00000880, 0.00002156, -0.00000093, -0.00000832, -0.00000883, 0.00002323, 0.00001244, -0.00004487, -0.00002021, 0.00009223, -0.00006233, 0.00003001, },
		{ 0.00001737, -0.00004252, 0.00004098, 0.00004708, -0.00008776, 0.00003480, -0.00000729, 0.00003335, -0.00004901, 0.00003851, -0.00003436, 0.00006564, -0.00011156, 0.00012851, -0.00015746, 0.00012188, -0.00005099, 0.00003067, -0.00001435, 0.00002220, -0.00006244, 0.00005699, -0.00007269, 0.00007073, -0.00009535, 0.00010407, -0.00017174, 0.00018201, -0.00031040, 0.00035752, -0.00112234, 0.00164133, -0.00002595, -0.00054329, 0.00024187, -0.00024209, 0.00016070, -0.00016902, 0.00011465, -0.00009929, 0.00006917, -0.00006535, 0.00005475, -0.00007197, 0.00003976, 0.00000183, -0.00003255, 0.00006147, -0.00003123, -0.00004076, 0.00005838, -0.00007994, 0.00006801, -0.00003467, 0.00003034, -0.00004663, 0.00004864, -0.00000977, -0.00000206, -0.00007258, 0.00011307, -0.00005216, 0.00001939, -0.00003555, },
		{ -0.00000748, -0.00001694, 0.00010102, -0.00010588, 0.00003659, -0.00002187, 0.00005665, -0.00006721, 0.00005619, -0.00006517, 0.00010115, -0.00013306, 0.00012403, -0.00012060, 0.00006225, 0.00000531, 0.00000063, -0.00000371, 0.00002864, -0.00007304, 0.00006758, -0.00009196, 0.00009848, -0.00013283, 0.00015257, -0.00022141, 0.00023867, -0.00036764, 0.00042821, -0.00099312, 0.00089154, 0.00089084, -0.00099482, 0.00043234, -0.00037431, 0.00025277, -0.00023685, 0.00015989, -0.00013815, 0.00010571, -0.00010229, 0.00009298, -0.00009895, 0.00005488, -0.00001719, 0.00000170, -0.00000302, 0.00005392, -0.00011953, 0.00011053, -0.00010684, 0.00007521, -0.00003982, 0.00004582, -0.00006002, 0.00005128, -0.00001458, 0.00002111, -0.00009182, 0.00009432, -0.00001283, -0.00000919, -0.00001044, -0.00000974, },
		{ -0.00005586, 0.00011815, -0.00008388, 0.00000780, -0.00001028, 0.00004950, -0.00005108, 0.00003851, -0.00005346, 0.00007963, -0.00008545, 0.00004798, -0.00002312, -0.00003860, 0.00008459, -0.00004579, 0.00002863, 0.00000482, -0.00004505, 0.00003432, -0.00006041, 0.00006410, -0.00009366, 0.00010470, -0.00015249, 0.00014915, -0.00023806, 0.00023840, -0.00054093, -0.00001878, 0.00162215, -0.00110833, 0.00035391, -0.00031254, 0.00019111, -0.00017913, 0.00010363, -0.00009593, 0.00007482, -0.00008085, 0.00007651, -0.00007497, 0.00002912, -0.00000695, 0.00001341, -0.00003939, 0.00009430, -0.00013682, 0.00009606, -0.00007494, 0.00003757, -0.00001341, 0.00003371, -0.00004439, 0.00002814, 0.00000116, 0.00002044, -0.00007785, 0.00004572, 0.00003909, -0.00003897, 0.00001370, -0.00003453, 0.00002159, },
		{ 0.00009261, -0.00002632, -0.00004250, 0.00001931, 0.00001958, -0.00000953, -0.00000522, -0.00000924, 0.00001691, 0.00000357, -0.00005803, 0.00008382, -0.00012879, 0.00014188, -0.00007341, 0.00005415, -0.00002252, -0.00000577, -0.00001518, -0.00000560, -0.00000321, -0.00001143, 0.00000228, -0.00001943, -0.00001475, -0.00002066, -0.00005715, -0.00002389, -0.00074849, 0.00188400, -0.00075558, -0.00001366, -0.00006931, -0.00000477, -0.00001805, -0.00003206, 0.00000870, -0.00000982, -0.00001000, 0.00001152, -0.00000629, -0.00003155, 0.00003075, -0.00000316, -0.00003588, 0.00007921, -0.00009127, 0.00002648, -0.00000481, -0.00002226, 0.00002573, 0.00000779, -0.00001209, -0.00000800, 0.00002592, 0.00000660, -0.00004586, -0.00001049, 0.00008041, -0.00005170, 0.00002243, -0.00004304, 0.00002953, -0.00006189, },
		{ 0.00004331, -0.00008976, 0.00004491, -0.00001130, 0.00003342, -0.00004894, 0.00003771, -0.00004852, 0.00008413, -0.00013587, 0.00014194, -0.00015820, 0.00013874, -0.00005345, 0.00004727, -0.00002581, 0.00001405, -0.00004722, 0.00003505, -0.00005956, 0.00006054, -0.00008854, 0.00009559, -0.00015125, 0.00015495, -0.00028259, 0.00032209, -0.00106638, 0.00158428, -0.00001337, -0.00052893, 0.00022799, -0.00023027, 0.00015799, -0.00017254, 0.00011435, -0.00009795, 0.00006792, -0.00006396, 0.00007059, -0.00009374, 0.00006913, -0.00002827, -0.00000998, 0.00003161, -0.00001628, -0.00005634, 0.00006090, -0.00006774, 0.00004996, -0.00001073, 0.00001712, -0.00003769, 0.00004250, -0.00000427, -0.00001535, -0.00005218, 0.00009390, -0.00003848, 0.00001152, -0.00003247, 0.00001516, -0.00003925, 0.00003908, },
		{ -0.00010775, 0.00004488, -0.00002172, 0.00005416, -0.00006687, 0.00005767, -0.00007993, 0.00011397, -0.00014451, 0.00011971, -0.00010906, 0.00006912, 0.00001450, 0.00000576, 0.00000251, 0.00000206, -0.00004516, 0.00003953, -0.00007536, 0.00008358, -0.00011891, 0.00013316, -0.00018983, 0.00020406, -0.00033401, 0.00038781, -0.00093964, 0.00086005, 0.00085810, -0.00094443, 0.00039903, -0.00035209, 0.00024359, -0.00023319, 0.00015303, -0.00013315, 0.00010327, -0.00010340, 0.00011091, -0.00011725, 0.00007492, -0.00003327, 0.00000812, -0.00001116, 0.00004344, -0.00010673, 0.00008388, -0.00007172, 0.00004179, -0.00000866, 0.00002951, -0.00004730, 0.00003980, -0.00000196, 0.00000141, -0.00006831, 0.00007672, -0.00000472, -0.00001079, -0.00001182, -0.00000986, -0.00000597, -0.00001589, 0.00009488, },
		{ 0.00001241, -0.00000496, 0.00004435, -0.00004983, 0.00003960, -0.00006399, 0.00008210, -0.00008270, 0.00003053, -0.00000774, -0.00003416, 0.00009860, -0.00004533, 0.00004447, -0.00002834, -0.00001870, 0.00001322, -0.00005110, 0.00005411, -0.00008263, 0.00008672, -0.00012580, 0.00012554, -0.00021934, 0.00021808, -0.00052147, 0.00000666, 0.00152911, -0.00102604, 0.00031134, -0.00028802, 0.00018057, -0.00017187, 0.00009353, -0.00008956, 0.00007159, -0.00008189, 0.00009031, -0.00008193, 0.00003295, -0.00000386, -0.00000053, -0.00002351, 0.00005983, -0.00010229, 0.00005317, -0.00003366, 0.00000587, 0.00001024, 0.00002531, -0.00003644, 0.00001935, 0.00001197, 0.00000443, -0.00006200, 0.00004008, 0.00003364, -0.00002956, 0.00000505, -0.00002978, 0.00001784, -0.00004864, 0.00010779, -0.00008414, },
		{ 0.00002654, 0.00001544, -0.00001006, -0.00000328, -0.00001642, 0.00001241, 0.00001249, -0.00007584, 0.00008997, -0.00011737, 0.00015264, -0.00006942, 0.00006989, -0.00004940, 0.00000594, -0.00001896, -0.00001158, -0.00000052, -0.00001237, -0.00000249, -0.00001210, -0.00001218, -0.00003221, -0.00004094, -0.00005422, -0.00066183, 0.00174948, -0.00067443, -0.00003653, -0.00006244, -0.00000284, -0.00001804, -0.00003608, 0.00000936, -0.00000925, -0.00001288, 0.00002104, -0.00000253, -0.00004053, 0.00004715, -0.00002980, -0.00000680, 0.00003606, -0.00005478, -0.00000978, 0.00002063, -0.00003410, 0.00002871, 0.00001604, -0.00001580, -0.00000710, 0.00002714, 0.00000241, -0.00004511, 0.00000113, 0.00005981, -0.00003281, 0.00000882, -0.00003491, 0.00002145, -0.00005044, 0.00008140, -0.00002813, -0.00003911, },
		{ -0.00001001, 0.00002791, -0.00004372, 0.00003050, -0.00005344, 0.00008919, -0.00014270, 0.00013128, -0.00013741, 0.00014448, -0.00004328, 0.00005858, -0.00004012, 0.00000651, -0.00003265, 0.00001447, -0.00004435, 0.00004636, -0.00007716, 0.00008035, -0.00012150, 0.00011497, -0.00023483, 0.00025445, -0.00094855, 0.00145806, 0.00001650, -0.00049902, 0.00019846, -0.00020388, 0.00014135, -0.00016253, 0.00010336, -0.00008925, 0.00005998, -0.00005532, 0.00007882, -0.00010581, 0.00008682, -0.00005378, 0.00001544, -0.00000029, 0.00000095, -0.00006631, 0.00005607, -0.00005154, 0.00002934, 0.00001403, 0.00000203, -0.00002635, 0.00003275, 0.00000432, -0.00002901, -0.00002693, 0.00006525, -0.00001862, 0.00000026, -0.00002647, 0.00000783, -0.00002973, 0.00003372, 0.00003472, -0.00008356, 0.00004950, },
		{ 0.00004374, -0.00005863, 0.00004931, -0.00008181, 0.00011084, -0.00013838, 0.00009641, -0.00008721, 0.00007640, 0.00002623, 0.00001571, -0.00000156, -0.00001929, -0.00002118, 0.00001399, -0.00005631, 0.00006412, -0.00009944, 0.00010570, -0.00014781, 0.00015530, -0.00027984, 0.00031797, -0.00084103, 0.00079946, 0.00079669, -0.00084813, 0.00033310, -0.00030410, 0.00021267, -0.00021095, 0.00013236, -0.00011787, 0.00009085, -0.00009400, 0.00011813, -0.00012395, 0.00008420, -0.00004698, 0.00001751, -0.00001901, 0.00003114, -0.00008547, 0.00005089, -0.00003502, 0.00000791, 0.00002216, 0.00001143, -0.00003209, 0.00002512, 0.00001258, -0.00001814, -0.00004035, 0.00005263, 0.00000480, -0.00001160, -0.00001375, -0.00001096, -0.00000444, -0.00001024, 0.00007701, -0.00009757, 0.00004517, -0.00001389, },
		{ -0.00004293, 0.00003323, -0.00006510, 0.00007462, -0.00007179, 0.00000825, 0.00000093, -0.00001586, 0.00010483, -0.00003060, 0.00004401, -0.00005302, 0.00000160, -0.00000465, -0.00003986, 0.00004125, -0.00006853, 0.00006417, -0.00009354, 0.00009407, -0.00018922, 0.00018342, -0.00048546, 0.00004528, 0.00137363, -0.00088672, 0.00023685, -0.00024092, 0.00015251, -0.00015108, 0.00007472, -0.00007673, 0.00006117, -0.00007505, 0.00009633, -0.00008163, 0.00003242, -0.00000444, -0.00000852, -0.00000828, 0.00002410, -0.00006225, 0.00000981, 0.00000398, -0.00002284, 0.00003175, 0.00001594, -0.00002709, 0.00000921, 0.00002236, -0.00000986, -0.00004343, 0.00003181, 0.00002535, -0.00001659, -0.00000623, -0.00002317, 0.00000993, -0.00003262, 0.00008424, -0.00007415, 0.00001162, 0.00000561, 0.00003370, },
		{ -0.00000265, -0.00002207, 0.00000849, 0.00001650, -0.00008320, 0.00007555, -0.00008473, 0.00015146, -0.00004758, 0.00006908, -0.00006958, 0.00001346, -0.00002150, -0.00001505, 0.00000095, -0.00001332, -0.00000727, -0.00000529, -0.00000972, -0.00004323, -0.00002286, -0.00009288, -0.00053477, 0.00154415, -0.00055014, -0.00007258, -0.00004874, -0.00000472, -0.00001684, -0.00003825, 0.00000824, -0.00000888, -0.00001479, 0.00003011, 0.00000001, -0.00004466, 0.00005257, -0.00004662, 0.00001909, -0.00000526, -0.00001810, -0.00004024, 0.00003882, -0.00004179, 0.00002980, 0.00002288, -0.00001862, -0.00000618, 0.00002647, 0.00000038, -0.00004230, 0.00001227, 0.00003499, -0.00001071, -0.00000683, -0.00002470, 0.00000858, -0.00003092, 0.00006100, -0.00002608, -0.00003499, 0.00003355, 0.00001119, -0.00001039, },
		{ 0.00001756, -0.00004895, 0.00008006, -0.00013078, 0.00009656, -0.00009899, 0.00014301, -0.00001932, 0.00006008, -0.00005512, 0.00000084, -0.00002161, -0.00000140, -0.00003005, 0.00003075, -0.00006340, 0.00006140, -0.00008779, 0.00006938, -0.00017622, 0.00016680, -0.00078518, 0.00127796, 0.00005780, -0.00045544, 0.00015696, -0.00016651, 0.00011367, -0.00014098, 0.00008324, -0.00007417, 0.00004621, -0.00003986, 0.00007805, -0.00010592, 0.00008942, -0.00006857, 0.00003717, -0.00002879, 0.00001543, -0.00006854, 0.00004556, -0.00003443, 0.00001022, 0.00003502, -0.00001141, -0.00001507, 0.00002155, 0.00001350, -0.00003947, -0.00000231, 0.00003378, 0.00000183, -0.00001086, -0.00001987, -0.00000246, -0.00001591, 0.00002565, 0.00002269, -0.00007071, 0.00004894, -0.00000408, 0.00001766, -0.00003405, },
		{ -0.00007162, 0.00009257, -0.00011580, 0.00005716, -0.00006068, 0.00008711, 0.00004306, 0.00002651, -0.00001617, -0.00003157, -0.00000567, -0.00000477, -0.00003849, 0.00004350, -0.00007795, 0.00007518, -0.00010255, 0.00010092, -0.00021459, 0.00022974, -0.00070914, 0.00071558, 0.00071270, -0.00071763, 0.00024454, -0.00023795, 0.00016559, -0.00017404, 0.00010083, -0.00009426, 0.00007001, -0.00007485, 0.00011366, -0.00011775, 0.00008069, -0.00005443, 0.00002694, -0.00002596, 0.00001806, -0.00006033, 0.00001878, -0.00000371, -0.00001973, 0.00004658, -0.00000410, -0.00001768, 0.00001056, 0.00002524, -0.00003279, -0.00001392, 0.00002774, 0.00001254, -0.00001086, -0.00001607, -0.00001316, -0.00000229, -0.00000214, 0.00005127, -0.00007839, 0.00003894, -0.00000016, 0.00002714, -0.00004392, 0.00003251, },
		{ 0.00005861, -0.00005465, -0.00001548, -0.00000091, 0.00001546, 0.00010879, -0.00000580, 0.00002579, -0.00006488, 0.00001200, -0.00001638, -0.00002899, 0.00002766, -0.00005386, 0.00004098, -0.00006096, 0.00005998, -0.00015238, 0.00013861, -0.00043446, 0.00008938, 0.00117431, -0.00070931, 0.00014312, -0.00017958, 0.00011246, -0.00012038, 0.00004978, -0.00005914, 0.00004502, -0.00006111, 0.00009419, -0.00007409, 0.00002724, -0.00000748, -0.00000957, 0.00000238, -0.00000711, -0.00002520, -0.00002507, 0.00003109, -0.00004289, 0.00004674, 0.00000819, -0.00001827, 0.00000003, 0.00002954, -0.00001915, -0.00002573, 0.00002293, 0.00001543, -0.00000321, -0.00001697, -0.00001740, 0.00000112, -0.00001244, 0.00005285, -0.00005724, 0.00000694, 0.00001956, 0.00001922, -0.00003175, 0.00002085, -0.00005792, },
		{ 0.00001529, -0.00007942, 0.00004142, -0.00003695, 0.00014660, -0.00001193, 0.00004958, -0.00007949, 0.00001507, -0.00002244, -0.00001582, 0.00000089, -0.00001422, -0.00001124, 0.00000009, -0.00000796, -0.00005125, -0.00000753, -0.00013059, -0.00038602, 0.00129291, -0.00040098, -0.00011342, -0.00003189, -0.00000903, -0.00001465, -0.00003874, 0.00000570, -0.00000871, -0.00001547, 0.00003807, 0.00000098, -0.00004362, 0.00004690, -0.00005058, 0.00003595, -0.00003784, 0.00001087, -0.00005886, 0.00004681, -0.00004362, 0.00002849, 0.00002750, -0.00001980, -0.00000532, 0.00002369, 0.00000076, -0.00003729, 0.00002046, 0.00001132, 0.00000870, -0.00001977, -0.00001629, -0.00000475, -0.00000836, 0.00003584, -0.00002155, -0.00003036, 0.00003986, 0.00000704, -0.00001053, -0.00000315, -0.00002611, 0.00000532, },
		{ -0.00010324, 0.00004333, -0.00005242, 0.00014251, 0.00001667, 0.00004814, -0.00006794, -0.00000273, -0.00001577, -0.00001050, -0.00001897, 0.00001603, -0.00004939, 0.00004194, -0.00005569, 0.00002612, -0.00011727, 0.00007422, -0.00059791, 0.00106490, 0.00010242, -0.00040076, 0.00010855, -0.00012346, 0.00007986, -0.00011197, 0.00005728, -0.00005495, 0.00002863, -0.00001941, 0.00006881, -0.00009459, 0.00007713, -0.00006965, 0.00005028, -0.00004936, 0.00002347, -0.00006258, 0.00003173, -0.00001917, -0.00000414, 0.00004851, -0.00002018, -0.00000604, 0.00001106, 0.00002085, -0.00004358, 0.00001625, 0.00000659, 0.00001701, -0.00001813, -0.00001537, -0.00001274, -0.00000066, 0.00001596, 0.00000946, -0.00005424, 0.00004501, 0.00000453, 0.00000469, -0.00002171, 0.00000103, -0.00003725, 0.00005934, },
		{ 0.00000908, -0.00003786, 0.00010504, 0.00006826, 0.00003184, -0.00003831, -0.00003389, -0.00000069, -0.00001445, -0.00002459, 0.00002466, -0.00005766, 0.00004640, -0.00006115, 0.00004974, -0.00014872, 0.00013626, -0.00055905, 0.00061617, 0.00061415, -0.00056837, 0.00014676, -0.00016438, 0.00011089, -0.00012902, 0.00006341, -0.00006592, 0.00004397, -0.00004873, 0.00009904, -0.00010035, 0.00006540, -0.00005368, 0.00003385, -0.00003127, 0.00000548, -0.00003595, -0.00000644, 0.00001716, -0.00003631, 0.00006002, -0.00001361, -0.00000679, -0.00000097, 0.00003277, -0.00003866, 0.00000549, 0.00000761, 0.00001568, -0.00000820, -0.00001851, -0.00001646, 0.00000122, 0.00000534, 0.00002348, -0.00005541, 0.00002950, 0.00001607, 0.00000762, -0.00002578, 0.00001084, -0.00005319, 0.00006360, -0.00008192, },
		{ -0.00001590, 0.00005599, 0.00011978, 0.00002017, -0.00000695, -0.00006312, 0.00001125, -0.00002079, -0.00002004, 0.00001509, -0.00004068, 0.00002064, -0.00003294, 0.00002845, -0.00011398, 0.00008885, -0.00037112, 0.00013028, 0.00095387, -0.00051712, 0.00004607, -0.00011489, 0.00006816, -0.00008515, 0.00002245, -0.00003942, 0.00002574, -0.00004232, 0.00008525, -0.00006106, 0.00001825, -0.00001141, -0.00000447, 0.00000639, -0.00002948, 0.00000228, -0.00004547, 0.00004370, -0.00005079, 0.00005223, 0.00000393, -0.00001146, -0.00000635, 0.00003145, -0.00002108, -0.00001205, 0.00001524, 0.00000545, 0.00000728, -0.00002404, -0.00001519, -0.00000483, 0.00000604, 0.00002081, -0.00003838, 0.00000107, 0.00003402, 0.00000352, -0.00001871, 0.00000545, -0.00004560, 0.00003757, -0.00003488, -0.00003654, },
		{ 0.00001491, 0.00015106, 0.00002865, 0.00001326, -0.00007789, 0.00001026, -0.00002173, -0.00001408, -0.00000073, -0.00001500, -0.00001374, 0.00000334, -0.00000735, -0.00005428, 0.00000101, -0.00015816, -0.00023630, 0.00102456, -0.00024809, -0.00014931, -0.00001638, -0.00001378, -0.00001182, -0.00003798, 0.00000236, -0.00000874, -0.00001490, 0.00004446, 0.00000028, -0.00003829, 0.00003256, -0.00004204, 0.00004095, -0.00005731, 0.00002715, -0.00006281, 0.00004398, -0.00003906, 0.00002454, 0.00002942, -0.00001881, -0.00000457, 0.00001884, 0.00000348, -0.00003015, 0.00002371, -0.00000626, 0.00002013, -0.00002590, -0.00001322, -0.00001404, 0.00001141, 0.00001132, -0.00001654, -0.00002538, 0.00004513, 0.00000307, -0.00001048, -0.00000430, -0.00002880, 0.00000282, 0.00000973, -0.00006627, -0.00000738, },
		{ 0.00015439, 0.00006013, 0.00002063, -0.00007537, -0.00000521, -0.00001508, -0.00001255, -0.00001219, 0.00000383, -0.00003680, 0.00002465, -0.00002990, -0.00000795, -0.00006759, -0.00000838, -0.00040991, 0.00084206, 0.00014168, -0.00033817, 0.00005905, -0.00008082, 0.00004587, -0.00008084, 0.00002983, -0.00003467, 0.00001009, 0.00000312, 0.00005335, -0.00007501, 0.00005371, -0.00005775, 0.00005262, -0.00005951, 0.00002348, -0.00004971, 0.00001724, -0.00000758, -0.00001197, 0.00005236, -0.00002241, -0.00000070, 0.00000288, 0.00002463, -0.00003945, 0.00002479, -0.00001060, 0.00002248, -0.00001894, -0.00001504, -0.00002012, 0.00001288, 0.00000592, -0.00000239, -0.00003797, 0.00004040, 0.00001297, -0.00000821, -0.00000924, -0.00001586, -0.00002245, 0.00003251, -0.00006709, -0.00001808, -0.00001147, },
		{ 0.00010520, 0.00002387, -0.00006195, -0.00002859, -0.00000521, -0.00001534, -0.00001560, 0.00000937, -0.00004074, 0.00002300, -0.00002919, 0.00000908, -0.00009161, 0.00005035, -0.00040671, 0.00050980, 0.00050983, -0.00041708, 0.00005416, -0.00009514, 0.00005835, -0.00008377, 0.00002617, -0.00003738, 0.00001705, -0.00001986, 0.00007790, -0.00007606, 0.00004223, -0.00004509, 0.00003662, -0.00003433, -0.00000537, -0.00001603, -0.00002114, 0.00002559, -0.00003998, 0.00006038, -0.00001530, -0.00000095, -0.00000769, 0.00003332, -0.00003380, 0.00001427, -0.00000382, 0.00001272, -0.00000390, -0.00002067, -0.00002066, 0.00000657, 0.00000909, -0.00000005, -0.00003471, 0.00002110, 0.00003049, -0.00001068, -0.00000827, -0.00001068, -0.00003254, 0.00003117, -0.00004500, -0.00003783, -0.00002867, 0.00013420, },
		{ 0.00003502, -0.00004663, -0.00005077, 0.00000113, -0.00001873, -0.00001351, 0.00000454, -0.00003022, 0.00000549, -0.00001290, 0.00000353, -0.00007864, 0.00003955, -0.00029920, 0.00015999, 0.00073551, -0.00033398, -0.00003827, -0.00005811, 0.00002795, -0.00005139, -0.00000309, -0.00002062, 0.00000655, -0.00002188, 0.00007219, -0.00004562, 0.00000732, -0.00001471, 0.00000460, 0.00000380, -0.00004106, 0.00001678, -0.00004942, 0.00004159, -0.00004604, 0.00004730, 0.00000395, -0.00000732, -0.00000900, 0.00002731, -0.00001491, -0.00000427, 0.00000984, -0.00000291, 0.00001215, -0.00002523, -0.00001830, -0.00000492, 0.00001730, -0.00000457, -0.00002302, -0.00000283, 0.00004575, -0.00001040, -0.00000684, -0.00000903, -0.00003265, 0.00001635, -0.00001710, -0.00005068, -0.00004508, 0.00009950, 0.00014936, },
		{ -0.00003423, -0.00006606, -0.00000018, -0.00001958, -0.00001043, -0.00000360, -0.00001560, -0.00001441, 0.00000411, -0.00000809, -0.00005128, 0.00000034, -0.00016855, -0.00010471, 0.00076717, -0.00011177, -0.00017133, -0.00000646, -0.00001699, -0.00000879, -0.00003652, -0.00000104, -0.00000892, -0.00001324, 0.00004903, -0.00000198, -0.00003042, 0.00001396, -0.00002452, 0.00003469, -0.00006268, 0.00002946, -0.00005299, 0.00003225, -0.00002891, 0.00001805, 0.00002852, -0.00001549, -0.00000399, 0.00001229, 0.00000812, -0.00002121, 0.00002097, -0.00001440, 0.00002038, -0.00002305, -0.00001741, -0.00001591, 0.00002330, -0.00000739, -0.00001305, -0.00002019, 0.00004911, -0.00000070, -0.00001028, -0.00000547, -0.00003058, 0.00000076, 0.00000160, -0.00004755, -0.00006256, 0.00005666, 0.00018040, 0.00006182, },
		{ -0.00007440, -0.00000864, -0.00001774, -0.00000921, -0.00000933, -0.00000527, -0.00002642, 0.00001115, -0.00001327, -0.00002846, -0.00003396, -0.00006910, -0.00024205, 0.00063123, 0.00016805, -0.00027142, 0.00001409, -0.00004438, 0.00001752, -0.00005303, 0.00000547, -0.00001662, -0.00000631, 0.00002429, 0.00003514, -0.00005220, 0.00002555, -0.00003691, 0.00004507, -0.00005932, 0.00001614, -0.00003261, 0.00000452, -0.00000022, -0.00001340, 0.00004656, -0.00001784, 0.00000061, -0.00000226, 0.00002419, -0.00002705, 0.00002177, -0.00001486, 0.00001652, -0.00001274, -0.00001958, -0.00002259, 0.00002204, -0.00000321, -0.00001053, -0.00002563, 0.00003793, 0.00001822, -0.00001803, 0.00000057, -0.00002944, -0.00000962, 0.00000675, -0.00003180, -0.00007455, 0.00000818, 0.00019121, 0.00010443, -0.00002272, },
		{ -0.00002090, -0.00001521, -0.00001029, -0.00001070, -0.00000205, -0.00002793, 0.00000673, -0.00000959, -0.00001672, -0.00004974, -0.00001777, -0.00026657, 0.00040474, 0.00040816, -0.00027930, -0.00002043, -0.00004069, 0.00001695, -0.00004579, -0.00000493, -0.00001322, -0.00000621, 0.00000693, 0.00005504, -0.00005080, 0.00001699, -0.00003108, 0.00003489, -0.00003492, -0.00001366, -0.00000271, -0.00002461, 0.00002287, -0.00003213, 0.00004847, -0.00000939, -0.00000016, -0.00000928, 0.00002688, -0.00001876, 0.00001158, -0.00000520, 0.00000408, 0.00000086, -0.00002194, -0.00002532, 0.00001370, 0.00000693, -0.00001399, -0.00002178, 0.00001780, 0.00003892, -0.00002360, 0.00000440, -0.00002664, -0.00001651, 0.00000376, -0.00001476, -0.00007242, -0.00004281, 0.00017803, 0.00015673, -0.00000532, -0.00007908, },
		{ -0.00001393, -0.00001294, -0.00000877, -0.00000393, -0.00002261, -0.00000369, -0.00000223, -0.00001262, -0.00004971, -0.00000438, -0.00022332, 0.00017283, 0.00053941, -0.00018025, -0.00009686, -0.00001846, -0.00000099, -0.00002451, -0.00002298, -0.00000565, -0.00000934, -0.00000325, 0.00005834, -0.00003150, -0.00000314, -0.00001609, 0.00001457, -0.00000349, -0.00004255, 0.00001851, -0.00003914, 0.00002823, -0.00003116, 0.00003325, 0.00000778, -0.00000558, -0.00000812, 0.00001785, -0.00000173, -0.00000260, 0.00000681, -0.00000806, 0.00000984, -0.00001990, -0.00002682, 0.00000209, 0.00001775, -0.00001763, -0.00001562, -0.00000206, 0.00005185, -0.00001973, 0.00000091, -0.00001861, -0.00002386, 0.00000022, -0.00000609, -0.00005423, -0.00008786, 0.00013857, 0.00020909, 0.00002533, -0.00008273, -0.00003403, },
		{ -0.00001650, -0.00000575, -0.00000715, -0.00001598, -0.00001325, 0.00000246, -0.00001014, -0.00004249, -0.00000974, -0.00015838, -0.00000549, 0.00054381, -0.00000794, -0.00017338, -0.00000498, -0.00001713, -0.00000590, -0.00003485, -0.00000383, -0.00000922, -0.00001074, 0.00005171, -0.00000551, -0.00002223, -0.00000354, -0.00000364, 0.00002072, -0.00005633, 0.00002021, -0.00003352, 0.00001562, -0.00001511, 0.00000945, 0.00002513, -0.00001004, -0.00000362, 0.00000472, 0.00001393, -0.00001107, 0.00001247, -0.00001207, 0.00000917, -0.00001181, -0.00002836, -0.00000928, 0.00002433, -0.00001660, -0.00001260, -0.00001497, 0.00005163, -0.00000430, -0.00000996, -0.00000604, -0.00003195, -0.00000120, -0.00000688, -0.00002810, -0.00011434, 0.00007373, 0.00024957, 0.00007416, -0.00008462, -0.00004749, -0.00001429, },
		{ -0.00000371, -0.00000870, -0.00001177, -0.00001814, 0.00000179, -0.00000634, -0.00003433, -0.00001898, -0.00010119, -0.00010965, 0.00044966, 0.00017663, -0.00020466, -0.00002158, -0.00001858, -0.00000066, -0.00003299, -0.00001196, -0.00000368, -0.00001778, 0.00004084, 0.00001793, -0.00003179, 0.00000015, -0.00001324, 0.00003101, -0.00005116, 0.00000403, -0.00001475, -0.00000452, 0.00000355, -0.00001024, 0.00003312, -0.00000780, -0.00000135, -0.00000459, 0.00002007, -0.00000828, 0.00000854, -0.00000676, 0.00000075, -0.00000146, -0.00002780, -0.00001962, 0.00002523, -0.00001034, -0.00001342, -0.00001997, 0.00003978, 0.00001785, -0.00002225, 0.00000528, -0.00003639, -0.00000368, -0.00001071, -0.00000720, -0.00011274, -0.00000791, 0.00026408, 0.00014211, -0.00008024, -0.00006270, -0.00001566, -0.00002061, },
		{ -0.00000750, -0.00001067, -0.00001853, -0.00000274, -0.00000220, -0.00002702, -0.00002562, -0.00006206, -0.00014960, 0.00030796, 0.00031614, -0.00016715, -0.00006798, -0.00000823, -0.00000693, -0.00002074, -0.00002527, 0.00000285, -0.00002197, 0.00002729, 0.00003529, -0.00003059, -0.00000396, -0.00001532, 0.00002944, -0.00003322, -0.00001903, 0.00000364, -0.00001886, 0.00001307, -0.00001698, 0.00002774, 0.00000197, -0.00000298, -0.00000694, 0.00001534, 0.00000341, -0.00000040, 0.00000182, -0.00000778, 0.00000414, -0.00002162, -0.00002976, 0.00002181, -0.00000166, -0.00001530, -0.00002016, 0.00002231, 0.00003827, -0.00002790, 0.00000874, -0.00003242, -0.00001125, -0.00001075, -0.00000034, -0.00008471, -0.00008778, 0.00023864, 0.00022442, -0.00006209, -0.00008136, -0.00001764, -0.00002468, -0.00000407, },
		{ -0.00001108, -0.00001698, -0.00000772, -0.00000009, -0.00002000, -0.00002879, -0.00003919, -0.00014862, 0.00016643, 0.00037995, -0.00006972, -0.00012202, -0.00000132, -0.00001419, -0.00000819, -0.00003446, 0.00000332, -0.00001939, 0.00001059, 0.00004682, -0.00002216, -0.00001068, -0.00001473, 0.00002230, -0.00001240, -0.00003679, 0.00001074, -0.00002006, 0.00000973, -0.00001098, 0.00001329, 0.00001387, -0.00000519, -0.00000494, 0.00000515, 0.00001561, -0.00000553, 0.00000527, -0.00000879, 0.00000035, -0.00000922, -0.00003892, 0.00001515, 0.00000671, -0.00001572, -0.00001863, 0.00000491, 0.00005038, -0.00002245, 0.00000231, -0.00002011, -0.00002323, -0.00000635, -0.00000582, -0.00004477, -0.00014214, 0.00016605, 0.00030792, -0.00002056, -0.00010403, -0.00002086, -0.00002784, -0.00000739, -0.00000430, },
		},
		{ { -0.00001506, -0.00001427, 0.00000101, 0.00000445, -0.00000844, 0.00000071, -0.00005815, -0.00002357, -0.00013694, 0.00005911, 0.00038831, 0.00004885, -0.00015675, -0.00003499, 0.00002609, -0.00001236, -0.00000410, -0.00000729, -0.00002336, 0.00000151, 0.00001323, 0.00000124, -0.00001178, -0.00001041, 0.00002091, -0.00000281, 0.00001420, -0.00001842, 0.00003180, 0.00000046, -0.00008836, 0.00001542, 0.00001500, 0.00001542, -0.00008836, 0.00000046, 0.00003180, -0.00001842, 0.00001420, -0.00000281, 0.00002091, -0.00001041, -0.00001178, 0.00000124, 0.00001323, 0.00000151, -0.00002336, -0.00000729, -0.00000410, -0.00001236, 0.00002609, -0.00003499, -0.00015675, 0.00004885, 0.00038831, 0.00005911, -0.00013694, -0.00002357, -0.00005815, 0.00000071, -0.00000844, 0.00000445, 0.00000101, -0.00001427, },
		{ -0.00001624, -0.00000489, 0.00000929, -0.00000974, 0.00000467, -0.00004298, -0.00003277, -0.00009230, -0.00005335, 0.00032664, 0.00018146, -0.00012925, -0.00008349, 0.00002502, -0.00000157, -0.00001016, 0.00000022, -0.00002389, -0.00000644, 0.00001161, 0.00000594, -0.00000682, -0.00001642, 0.00001603, 0.00000397, 0.00000809, -0.00000634, 0.00000398, 0.00003884, -0.00008678, -0.00001071, 0.00001574, 0.00002778, -0.00006373, -0.00004810, 0.00004966, -0.00001873, 0.00001245, -0.00000293, 0.00001797, 0.00000039, -0.00001589, -0.00000310, 0.00001077, 0.00000979, -0.00001958, -0.00001467, -0.00000067, -0.00001687, 0.00001696, 0.00000031, -0.00013125, -0.00008873, 0.00036606, 0.00021869, -0.00016105, -0.00003435, -0.00005680, -0.00001735, 0.00000005, -0.00000424, 0.00000674, -0.00001228, -0.00001412, },
		{ -0.00000992, 0.00000889, -0.00000430, -0.00000045, -0.00002239, -0.00004410, -0.00005537, -0.00010605, 0.00021519, 0.00027123, -0.00005553, -0.00011969, 0.00000161, 0.00001727, -0.00001849, 0.00000724, -0.00002226, -0.00001187, 0.00000584, 0.00001041, -0.00000263, -0.00001608, 0.00000473, 0.00001406, -0.00000099, 0.00000812, -0.00001796, 0.00005296, -0.00005816, -0.00004389, 0.00001694, 0.00002527, -0.00002305, -0.00008815, 0.00004545, -0.00000498, 0.00000399, 0.00000209, 0.00001014, 0.00001153, -0.00001631, -0.00000776, 0.00000638, 0.00001492, -0.00001120, -0.00002207, 0.00000128, -0.00001863, 0.00000747, 0.00001375, -0.00007536, -0.00018069, 0.00024566, 0.00038739, -0.00013010, -0.00007601, -0.00003813, -0.00004451, 0.00000880, -0.00001111, 0.00000905, -0.00000920, -0.00001232, -0.00001745, },
		{ 0.00000553, 0.00000236, -0.00000517, -0.00000862, -0.00004422, -0.00003938, -0.00010804, 0.00009249, 0.00030151, 0.00003953, -0.00012235, -0.00004147, 0.00003389, -0.00002086, 0.00000779, -0.00001527, -0.00001726, 0.00000033, 0.00001079, 0.00000260, -0.00001274, -0.00000649, 0.00001975, -0.00000441, 0.00001386, -0.00002225, 0.00003953, -0.00001238, -0.00007271, 0.00001466, 0.00001387, 0.00001667, -0.00010245, 0.00001508, 0.00001813, -0.00000679, 0.00000946, 0.00000067, 0.00001990, -0.00001279, -0.00001131, 0.00000031, 0.00001679, -0.00000115, -0.00002557, -0.00000129, -0.00001551, -0.00000342, 0.00001638, -0.00002678, -0.00019208, 0.00005299, 0.00050294, -0.00001502, -0.00014394, -0.00001377, -0.00006832, 0.00000835, -0.00001109, 0.00000596, -0.00000438, -0.00001111, -0.00001650, -0.00001474, },
		{ 0.00000616, -0.00000446, -0.00000528, -0.00003283, -0.00003951, -0.00008383, -0.00000672, 0.00027158, 0.00013192, -0.00009075, -0.00008265, 0.00002971, -0.00000681, -0.00000409, -0.00000005, -0.00002544, -0.00000063, 0.00000427, 0.00001036, -0.00001027, -0.00001153, 0.00001555, 0.00000237, 0.00000707, -0.00000935, 0.00000967, 0.00003130, -0.00008303, 0.00000225, 0.00000474, 0.00003783, -0.00008217, -0.00003489, 0.00004058, -0.00001470, 0.00001656, -0.00000773, 0.00002409, -0.00000707, -0.00001154, -0.00000786, 0.00001709, 0.00000597, -0.00002085, -0.00001134, -0.00000512, -0.00001882, 0.00002057, -0.00000761, -0.00013228, -0.00014124, 0.00049716, 0.00018889, -0.00021168, -0.00000459, -0.00007571, -0.00000646, -0.00000445, -0.00000038, 0.00000037, -0.00001019, -0.00001369, -0.00001895, 0.00000102, },
		{ -0.00000064, -0.00000622, -0.00002025, -0.00004064, -0.00006057, -0.00006059, 0.00019347, 0.00020490, -0.00003921, -0.00009854, -0.00000350, 0.00002132, -0.00002297, 0.00001704, -0.00003191, 0.00000015, -0.00000491, 0.00001648, -0.00000741, -0.00001101, 0.00000448, 0.00001407, -0.00000437, 0.00000736, -0.00001588, 0.00005214, -0.00006485, -0.00002405, 0.00000622, 0.00003267, -0.00003425, -0.00008604, 0.00004848, -0.00001386, 0.00001934, -0.00001070, 0.00002154, 0.00000107, -0.00000971, -0.00001494, 0.00001426, 0.00001031, -0.00001048, -0.00002379, 0.00000603, -0.00003179, 0.00002274, -0.00000736, -0.00005347, -0.00025204, 0.00033629, 0.00043882, -0.00022837, -0.00003493, -0.00005851, -0.00003425, 0.00000403, -0.00000592, 0.00000274, -0.00000856, -0.00001069, -0.00002075, -0.00000491, 0.00000738, },
		{ -0.00000410, -0.00001546, -0.00003256, -0.00005201, -0.00006972, 0.00009376, 0.00024316, 0.00002076, -0.00008503, -0.00004961, 0.00004224, -0.00003033, 0.00002076, -0.00002565, -0.00000625, -0.00000714, 0.00001295, 0.00000107, -0.00001113, -0.00000413, 0.00001943, -0.00000749, 0.00001216, -0.00002055, 0.00004131, -0.00002193, -0.00005713, 0.00001482, 0.00001055, 0.00001790, -0.00011301, 0.00002892, 0.00000048, 0.00001278, -0.00000275, 0.00000891, 0.00001382, -0.00001037, -0.00001481, 0.00000382, 0.00001672, -0.00000322, -0.00002714, 0.00000505, -0.00002861, 0.00000882, 0.00000078, -0.00001140, -0.00023545, 0.00006165, 0.00063726, -0.00013163, -0.00011990, -0.00001871, -0.00006807, 0.00000914, -0.00000917, 0.00000342, -0.00000727, -0.00000782, -0.00001964, -0.00001207, 0.00000744, 0.00000191, },
		{ -0.00001676, -0.00001925, -0.00005201, -0.00005654, 0.00000874, 0.00023146, 0.00008753, -0.00005842, -0.00007798, 0.00003143, -0.00001080, 0.00000167, -0.00000191, -0.00002344, 0.00000179, -0.00000239, 0.00001580, -0.00001491, -0.00000461, 0.00001264, 0.00000285, 0.00000114, -0.00000517, 0.00001028, 0.00002504, -0.00007817, 0.00001548, -0.00000700, 0.00004610, -0.00009736, -0.00001993, 0.00002550, -0.00000327, 0.00001627, -0.00001166, 0.00002924, -0.00001449, -0.00000602, -0.00001384, 0.00002624, -0.00000356, -0.00001628, -0.00001227, -0.00000640, -0.00002312, 0.00002597, -0.00001934, -0.00012584, -0.00021013, 0.00066673, 0.00011335, -0.00024240, 0.00002325, -0.00009434, 0.00000605, -0.00001113, 0.00000509, -0.00000855, -0.00000404, -0.00001731, -0.00001798, 0.00000539, 0.00000276, 0.00000134, },
		{ -0.00001268, -0.00004542, -0.00004813, -0.00003392, 0.00016748, 0.00015698, -0.00003409, -0.00007154, -0.00001270, 0.00002732, -0.00002841, 0.00002618, -0.00004015, 0.00001243, -0.00001899, 0.00002691, -0.00001611, -0.00000172, 0.00000023, 0.00001761, -0.00001339, 0.00001232, -0.00001556, 0.00004942, -0.00006865, -0.00000519, -0.00000357, 0.00003722, -0.00004341, -0.00008008, 0.00004570, -0.00001863, 0.00003546, -0.00002766, 0.00003663, -0.00001418, 0.00000271, -0.00002799, 0.00002970, -0.00000380, -0.00000150, -0.00003185, 0.00001657, -0.00005094, 0.00004529, -0.00003971, -0.00001537, -0.00034790, 0.00046431, 0.00046278, -0.00033303, 0.00002451, -0.00009211, -0.00001241, -0.00001019, 0.00000657, -0.00001071, -0.00000082, -0.00001443, -0.00002055, -0.00000039, 0.00000449, 0.00000474, -0.00001516, },
		{ -0.00002680, -0.00005446, -0.00003439, 0.00007534, 0.00020790, -0.00000669, -0.00004445, -0.00006023, 0.00005161, -0.00004056, 0.00003378, -0.00003681, 0.00000813, -0.00002054, 0.00002081, -0.00000487, -0.00000540, -0.00000514, 0.00002184, -0.00001449, 0.00001271, -0.00001666, 0.00003861, -0.00002782, -0.00004281, 0.00001621, 0.00000515, 0.00001880, -0.00011826, 0.00003886, -0.00001749, 0.00003789, -0.00002249, 0.00002312, 0.00000110, -0.00000104, -0.00002479, 0.00001496, 0.00000857, 0.00000027, -0.00003236, 0.00001511, -0.00004630, 0.00002770, -0.00002516, 0.00001727, -0.00029557, 0.00008855, 0.00077560, -0.00028383, -0.00006179, -0.00004397, -0.00005239, -0.00000130, 0.00000098, -0.00000660, -0.00000382, -0.00000819, -0.00002145, -0.00000803, 0.00000707, 0.00000340, -0.00000603, -0.00001782, },
		{ -0.00006266, -0.00002019, -0.00000135, 0.00020890, 0.00004292, -0.00002761, -0.00007337, 0.00003287, -0.00001522, 0.00000814, -0.00000624, -0.00001639, -0.00000206, -0.00000446, 0.00001858, -0.00001749, 0.00000116, 0.00001032, 0.00000216, -0.00000609, 0.00000393, 0.00000591, 0.00002127, -0.00007319, 0.00002856, -0.00001866, 0.00005155, -0.00010712, -0.00000574, 0.00000617, 0.00001614, 0.00000849, -0.00001110, 0.00002992, -0.00001814, -0.00000319, -0.00001674, 0.00003331, -0.00001427, -0.00000916, -0.00001507, -0.00000688, -0.00002655, 0.00002880, -0.00002916, -0.00011927, -0.00028610, 0.00086897, -0.00001396, -0.00023977, 0.00003762, -0.00010453, 0.00001406, -0.00001537, 0.00000824, -0.00001635, 0.00000342, -0.00002371, -0.00001296, 0.00000648, 0.00000124, 0.00000497, -0.00002587, -0.00000757, },
		{ -0.00002149, -0.00002904, 0.00014708, 0.00011815, -0.00003369, -0.00004344, -0.00002284, 0.00003296, -0.00003271, 0.00003223, -0.00004371, 0.00002053, -0.00003173, 0.00003797, -0.00002573, 0.00000909, -0.00000574, 0.00002247, -0.00002608, 0.00002283, -0.00001868, 0.00004669, -0.00007028, 0.00001185, -0.00001159, 0.00003811, -0.00004908, -0.00007111, 0.00003630, -0.00001599, 0.00004743, -0.00004457, 0.00005206, -0.00003107, 0.00001811, -0.00004412, 0.00005004, -0.00002587, 0.00001568, -0.00004678, 0.00003306, -0.00007536, 0.00007372, -0.00008158, 0.00003734, -0.00046788, 0.00063627, 0.00044190, -0.00042523, 0.00009054, -0.00013252, 0.00001706, -0.00003133, 0.00002445, -0.00002990, 0.00001382, -0.00002483, -0.00001434, 0.00000058, 0.00000299, 0.00000899, -0.00002384, -0.00000354, -0.00005410, },
		{ -0.00001135, 0.00004996, 0.00018721, -0.00003843, -0.00000314, -0.00007204, 0.00006086, -0.00005007, 0.00004473, -0.00004577, 0.00002232, -0.00003690, 0.00003286, -0.00001448, 0.00000393, -0.00000941, 0.00002701, -0.00002596, 0.00001776, -0.00001368, 0.00003358, -0.00003045, -0.00003055, 0.00001906, -0.00000233, 0.00001956, -0.00011720, 0.00004212, -0.00003127, 0.00006401, -0.00004727, 0.00004224, -0.00001769, 0.00001527, -0.00004167, 0.00003475, -0.00001014, 0.00001300, -0.00004494, 0.00003190, -0.00007095, 0.00005566, -0.00006463, 0.00006396, -0.00038022, 0.00014861, 0.00089696, -0.00045638, 0.00002590, -0.00009005, -0.00001968, -0.00002489, 0.00002118, -0.00002608, 0.00000877, -0.00001577, -0.00001818, -0.00000578, 0.00000648, 0.00000526, -0.00000957, -0.00001643, -0.00002724, -0.00004381, },
		{ -0.00002705, 0.00020083, -0.00000277, 0.00000335, -0.00007067, 0.00003511, -0.00002037, 0.00001489, -0.00001206, -0.00000519, -0.00001249, -0.00000016, 0.00001657, -0.00001592, 0.00000352, 0.00001082, -0.00000238, -0.00001084, 0.00001446, -0.00000185, 0.00002013, -0.00006868, 0.00004101, -0.00002954, 0.00005346, -0.00010942, 0.00000443, -0.00001380, 0.00004153, -0.00000764, -0.00000386, 0.00002365, -0.00001498, -0.00000647, -0.00001256, 0.00003307, -0.00002046, -0.00000446, -0.00001565, -0.00001036, -0.00002467, 0.00002328, -0.00002943, -0.00012309, -0.00035415, 0.00108931, -0.00019000, -0.00019610, 0.00002927, -0.00009880, 0.00001149, -0.00001219, 0.00000459, -0.00001842, 0.00000785, -0.00002949, -0.00000577, 0.00000463, 0.00000200, 0.00000546, -0.00003046, -0.00000144, -0.00006504, 0.00001450, },
		{ 0.00013661, 0.00008208, -0.00003274, -0.00001836, -0.00003122, 0.00003642, -0.00003456, 0.00003397, -0.00004072, 0.00002119, -0.00003884, 0.00004574, -0.00003288, 0.00001814, -0.00001036, 0.00002539, -0.00003884, 0.00003658, -0.00002507, 0.00004492, -0.00007026, 0.00002627, -0.00001701, 0.00003450, -0.00004939, -0.00006129, 0.00002149, -0.00000455, 0.00005114, -0.00005689, 0.00006369, -0.00004540, 0.00003227, -0.00005887, 0.00007028, -0.00005139, 0.00003791, -0.00006634, 0.00005322, -0.00010189, 0.00010381, -0.00012762, 0.00009814, -0.00060492, 0.00085107, 0.00036394, -0.00048637, 0.00014870, -0.00017018, 0.00004727, -0.00005423, 0.00004344, -0.00005096, 0.00003250, -0.00004058, -0.00000171, -0.00000370, 0.00000496, 0.00000965, -0.00002882, 0.00000275, -0.00006066, 0.00001194, -0.00004285, },
		{ 0.00017230, -0.00006869, 0.00003501, -0.00008265, 0.00006811, -0.00005699, 0.00005137, -0.00004941, 0.00003202, -0.00005173, 0.00004569, -0.00002515, 0.00001456, -0.00001499, 0.00003304, -0.00004025, 0.00002738, -0.00001343, 0.00002810, -0.00003043, -0.00002097, 0.00002368, -0.00001215, 0.00002112, -0.00011015, 0.00003728, -0.00003684, 0.00008561, -0.00007263, 0.00006307, -0.00003980, 0.00003624, -0.00006343, 0.00006137, -0.00003882, 0.00003597, -0.00006647, 0.00005670, -0.00010315, 0.00009295, -0.00011799, 0.00012993, -0.00049334, 0.00025418, 0.00097868, -0.00062794, 0.00013161, -0.00015172, 0.00002727, -0.00006008, 0.00005043, -0.00005441, 0.00003100, -0.00003241, -0.00000689, -0.00000859, 0.00000843, 0.00000495, -0.00001191, -0.00001413, -0.00003097, -0.00002174, -0.00000486, 0.00002761, },
		{ -0.00004687, 0.00003344, -0.00006964, 0.00003768, -0.00002528, 0.00002035, -0.00001693, 0.00000712, -0.00002716, 0.00000971, 0.00000979, -0.00001008, 0.00000205, 0.00001466, -0.00001177, -0.00001050, 0.00002303, -0.00001071, 0.00002103, -0.00006491, 0.00005241, -0.00003919, 0.00005176, -0.00010301, 0.00000756, -0.00002973, 0.00006849, -0.00003026, 0.00000997, 0.00000992, -0.00000384, -0.00001752, 0.00000103, 0.00002174, -0.00001713, -0.00000717, -0.00000963, -0.00002081, -0.00001307, 0.00000394, -0.00001258, -0.00014864, -0.00039591, 0.00130562, -0.00040184, -0.00011172, -0.00000611, -0.00007270, -0.00000580, 0.00000206, -0.00000931, -0.00001078, 0.00000475, -0.00003022, -0.00000001, 0.00000204, 0.00000357, 0.00000387, -0.00003075, -0.00000170, -0.00005587, 0.00003875, -0.00005724, 0.00020058, },
		{ -0.00002849, 0.00000126, -0.00003643, 0.00003705, -0.00003389, 0.00003189, -0.00003159, 0.00001348, -0.00003789, 0.00004746, -0.00003500, 0.00002280, -0.00001108, 0.00002344, -0.00004770, 0.00004997, -0.00003309, 0.00004404, -0.00006873, 0.00003729, -0.00001915, 0.00002583, -0.00004242, -0.00005371, 0.00000451, 0.00001438, 0.00004465, -0.00006126, 0.00006799, -0.00005317, 0.00004089, -0.00006742, 0.00008452, -0.00007407, 0.00005970, -0.00008589, 0.00007259, -0.00012532, 0.00012904, -0.00016939, 0.00015600, -0.00074497, 0.00109819, 0.00022539, -0.00050238, 0.00018532, -0.00019485, 0.00007018, -0.00007241, 0.00005793, -0.00006846, 0.00005027, -0.00005786, 0.00001518, -0.00001280, 0.00001072, 0.00000582, -0.00002812, 0.00000291, -0.00005953, 0.00004325, -0.00006786, 0.00013514, 0.00004635, },
		{ 0.00006628, -0.00008972, 0.00007166, -0.00005990, 0.00005240, -0.00004572, 0.00003373, -0.00006054, 0.00005523, -0.00003348, 0.00002327, -0.00001892, 0.00003681, -0.00005393, 0.00003947, -0.00001595, 0.00002315, -0.00002820, -0.00001468, 0.00003043, -0.00002476, 0.00002515, -0.00009902, 0.00002504, -0.00003207, 0.00009772, -0.00009343, 0.00008122, -0.00006094, 0.00005776, -0.00008588, 0.00009032, -0.00007362, 0.00006681, -0.00009555, 0.00008810, -0.00014080, 0.00013669, -0.00018170, 0.00021135, -0.00063247, 0.00041112, 0.00100122, -0.00077503, 0.00023896, -0.00021907, 0.00008162, -0.00010194, 0.00008475, -0.00008811, 0.00006040, -0.00005720, 0.00001317, -0.00001846, 0.00001510, 0.00000013, -0.00001001, -0.00001462, -0.00003321, 0.00000621, -0.00001373, 0.00001373, 0.00015653, -0.00009235, },
		{ -0.00006875, 0.00003916, -0.00002839, 0.00002262, -0.00001798, 0.00001645, -0.00004183, 0.00002213, 0.00000035, -0.00000180, -0.00000179, 0.00002041, -0.00002479, -0.00000460, 0.00002752, -0.00001863, 0.00002320, -0.00006208, 0.00006262, -0.00004770, 0.00004739, -0.00008810, 0.00000187, -0.00003725, 0.00009131, -0.00005521, 0.00002789, -0.00000939, 0.00001403, -0.00003554, 0.00002358, -0.00000160, -0.00000165, -0.00002061, 0.00000609, -0.00004089, 0.00001090, -0.00003260, 0.00002653, -0.00020501, -0.00039333, 0.00149159, -0.00062821, 0.00000447, -0.00006640, -0.00002641, -0.00003850, 0.00002828, -0.00003451, 0.00000846, -0.00000892, -0.00002201, 0.00000035, 0.00000189, 0.00000330, 0.00000286, -0.00002914, -0.00000624, -0.00003587, 0.00004786, -0.00008254, 0.00020060, -0.00008491, 0.00005996, },
		{ -0.00003857, 0.00003544, -0.00003184, 0.00002803, -0.00001889, -0.00000078, -0.00002902, 0.00004250, -0.00003135, 0.00002211, -0.00000688, 0.00001515, -0.00004972, 0.00005942, -0.00004035, 0.00004333, -0.00006565, 0.00004437, -0.00001779, 0.00001225, -0.00002691, -0.00005154, -0.00001000, 0.00003688, 0.00002889, -0.00005660, 0.00006325, -0.00005196, 0.00004098, -0.00006611, 0.00008764, -0.00008758, 0.00007476, -0.00009972, 0.00008571, -0.00013963, 0.00014199, -0.00019711, 0.00019726, -0.00086864, 0.00135785, 0.00003347, -0.00046707, 0.00019089, -0.00019834, 0.00007878, -0.00007995, 0.00006262, -0.00007691, 0.00006141, -0.00007133, 0.00003215, -0.00002416, 0.00001868, -0.00000139, -0.00002195, -0.00000396, -0.00004757, 0.00006463, -0.00009483, 0.00013799, 0.00001203, -0.00002081, 0.00001498, },
		{ 0.00007072, -0.00005853, 0.00004804, -0.00003472, 0.00002603, -0.00006032, 0.00005826, -0.00003656, 0.00002717, -0.00001843, 0.00003521, -0.00006298, 0.00005049, -0.00001969, 0.00001866, -0.00002381, -0.00001230, 0.00003964, -0.00004060, 0.00003370, -0.00008696, 0.00000837, -0.00001758, 0.00009745, -0.00010539, 0.00009243, -0.00007653, 0.00007506, -0.00010388, 0.00011552, -0.00010825, 0.00010021, -0.00012779, 0.00012197, -0.00017912, 0.00018094, -0.00024826, 0.00029905, -0.00078760, 0.00061598, 0.00095266, -0.00087680, 0.00033030, -0.00027970, 0.00013402, -0.00014313, 0.00011801, -0.00012141, 0.00009182, -0.00008631, 0.00003999, -0.00003513, 0.00002700, -0.00001020, -0.00000196, -0.00002088, -0.00002913, 0.00003264, -0.00003247, 0.00000889, 0.00013672, -0.00010639, 0.00008830, -0.00009193, },
		{ -0.00002833, 0.00002035, -0.00001306, 0.00001908, -0.00005184, 0.00003299, -0.00000841, 0.00000587, -0.00000517, 0.00002514, -0.00003831, 0.00000507, 0.00002764, -0.00002455, 0.00002615, -0.00006055, 0.00007180, -0.00005572, 0.00004235, -0.00006665, -0.00001225, -0.00003366, 0.00010457, -0.00007718, 0.00004571, -0.00003046, 0.00003522, -0.00005728, 0.00005175, -0.00003448, 0.00002518, -0.00004505, 0.00003200, -0.00007066, 0.00004691, -0.00008597, 0.00008845, -0.00029594, -0.00033303, 0.00162145, -0.00084299, 0.00013656, -0.00014331, 0.00003495, -0.00008324, 0.00006400, -0.00006889, 0.00003809, -0.00003342, -0.00000302, -0.00000756, 0.00000699, -0.00000152, 0.00000554, -0.00002915, -0.00001086, -0.00000944, 0.00004195, -0.00009733, 0.00019480, -0.00011258, 0.00007994, -0.00006611, 0.00003797, },
		{ -0.00003017, 0.00002518, -0.00000661, -0.00001764, -0.00001487, 0.00003247, -0.00002327, 0.00001712, 0.00000122, 0.00000130, -0.00004401, 0.00006257, -0.00004476, 0.00004195, -0.00006106, 0.00004740, -0.00001341, -0.00000501, -0.00000296, -0.00005693, -0.00001741, 0.00005730, 0.00000747, -0.00004445, 0.00005018, -0.00004167, 0.00003187, -0.00005377, 0.00007695, -0.00008747, 0.00007793, -0.00010273, 0.00008773, -0.00013965, 0.00013633, -0.00020198, 0.00020859, -0.00095444, 0.00160352, -0.00019425, -0.00038377, 0.00016242, -0.00017678, 0.00006910, -0.00007321, 0.00005417, -0.00007247, 0.00006124, -0.00007574, 0.00004416, -0.00003387, 0.00002573, -0.00000915, -0.00001269, -0.00001597, -0.00002512, 0.00007148, -0.00011545, 0.00013908, -0.00001765, -0.00001157, 0.00002393, -0.00003891, 0.00003300, },
		{ 0.00004008, -0.00001863, 0.00001019, -0.00005061, 0.00005349, -0.00003308, 0.00002482, -0.00001203, 0.00002636, -0.00006416, 0.00005676, -0.00002226, 0.00001384, -0.00001712, -0.00001421, 0.00005124, -0.00005960, 0.00004849, -0.00007762, -0.00000805, 0.00000309, 0.00008491, -0.00010639, 0.00009395, -0.00008313, 0.00008415, -0.00011284, 0.00013098, -0.00013561, 0.00012924, -0.00015689, 0.00015227, -0.00021151, 0.00021767, -0.00030736, 0.00037956, -0.00094173, 0.00085494, 0.00083186, -0.00091924, 0.00039060, -0.00032181, 0.00017484, -0.00017578, 0.00014340, -0.00014766, 0.00011869, -0.00011381, 0.00006903, -0.00005592, 0.00004254, -0.00002498, 0.00001210, -0.00003390, -0.00001563, 0.00005055, -0.00005335, 0.00000983, 0.00011311, -0.00011038, 0.00010060, -0.00008935, 0.00006573, -0.00005376, },
		{ -0.00000172, 0.00001303, -0.00005362, 0.00003856, -0.00001323, 0.00000978, -0.00000508, 0.00002553, -0.00004825, 0.00001505, 0.00002492, -0.00002863, 0.00002992, -0.00006087, 0.00008034, -0.00006422, 0.00003940, -0.00004222, -0.00003198, -0.00001912, 0.00010469, -0.00009134, 0.00005889, -0.00004865, 0.00005508, -0.00007793, 0.00008002, -0.00007142, 0.00005912, -0.00007738, 0.00006539, -0.00010707, 0.00009105, -0.00015134, 0.00016797, -0.00041767, -0.00021004, 0.00167507, -0.00102005, 0.00026490, -0.00022417, 0.00010232, -0.00013329, 0.00010374, -0.00010741, 0.00007375, -0.00006577, 0.00002560, -0.00002433, 0.00001858, -0.00001242, 0.00001421, -0.00003397, -0.00001087, 0.00001688, 0.00002537, -0.00010049, 0.00018010, -0.00012725, 0.00009134, -0.00006045, 0.00003320, -0.00002455, 0.00001349, },
		{ 0.00000122, -0.00003219, 0.00000039, 0.00002063, -0.00001370, 0.00001053, 0.00001045, -0.00001522, -0.00003216, 0.00005904, -0.00004533, 0.00003948, -0.00005536, 0.00004689, -0.00000732, -0.00002345, 0.00002753, -0.00006995, -0.00001453, 0.00006987, -0.00001439, -0.00002849, 0.00003181, -0.00002468, 0.00001550, -0.00003218, 0.00005316, -0.00007253, 0.00006676, -0.00009211, 0.00007597, -0.00012277, 0.00010881, -0.00017878, 0.00018040, -0.00098316, 0.00180627, -0.00043251, -0.00026487, 0.00010423, -0.00013180, 0.00004145, -0.00005203, 0.00003242, -0.00005431, 0.00004760, -0.00006759, 0.00004695, -0.00003805, 0.00002841, -0.00001383, -0.00000415, -0.00002906, 0.00000405, 0.00006349, -0.00012452, 0.00013317, -0.00003911, -0.00000340, 0.00003001, -0.00003915, 0.00003128, -0.00003039, 0.00002580, },
		{ -0.00001015, -0.00003355, 0.00004198, -0.00002385, 0.00001682, -0.00000029, 0.00001055, -0.00005624, 0.00005571, -0.00002150, 0.00000777, -0.00000813, -0.00002029, 0.00006449, -0.00008072, 0.00007014, -0.00007390, -0.00001913, 0.00002429, 0.00006321, -0.00009708, 0.00008544, -0.00007944, 0.00008305, -0.00011014, 0.00013253, -0.00014974, 0.00014727, -0.00017631, 0.00017263, -0.00023120, 0.00023868, -0.00034807, 0.00043770, -0.00107339, 0.00110513, 0.00064949, -0.00089817, 0.00041078, -0.00033697, 0.00019663, -0.00019349, 0.00015537, -0.00016116, 0.00013472, -0.00013317, 0.00009436, -0.00007636, 0.00005831, -0.00004122, 0.00002978, -0.00005206, 0.00000734, 0.00005548, -0.00006900, 0.00001139, 0.00008845, -0.00010613, 0.00010444, -0.00008329, 0.00005805, -0.00004729, 0.00003129, -0.00000128, },
		{ -0.00004588, 0.00003677, -0.00001212, 0.00000790, 0.00000052, 0.00001918, -0.00005099, 0.00002139, 0.00002194, -0.00003191, 0.00003490, -0.00006355, 0.00008855, -0.00007401, 0.00004136, -0.00001916, -0.00005249, 0.00000312, 0.00009115, -0.00009475, 0.00006396, -0.00005989, 0.00006912, -0.00009252, 0.00010220, -0.00010535, 0.00009350, -0.00011173, 0.00010086, -0.00014431, 0.00013628, -0.00021990, 0.00025439, -0.00055841, -0.00003009, 0.00164227, -0.00113824, 0.00037024, -0.00029490, 0.00016480, -0.00018006, 0.00014035, -0.00014323, 0.00010882, -0.00010021, 0.00005969, -0.00004782, 0.00003567, -0.00002892, 0.00002925, -0.00004514, -0.00000285, 0.00003644, 0.00000496, -0.00009476, 0.00015680, -0.00012869, 0.00009373, -0.00005167, 0.00002494, -0.00001759, 0.00000342, 0.00001438, -0.00000118, },
		{ 0.00001248, 0.00001065, -0.00000610, 0.00000568, 0.00001721, -0.00003019, -0.00001772, 0.00005051, -0.00004245, 0.00003625, -0.00004939, 0.00004388, -0.00000150, -0.00003955, 0.00006089, -0.00008821, -0.00000092, 0.00007041, -0.00003136, -0.00001336, 0.00001253, -0.00000511, -0.00000411, -0.00000577, 0.00002031, -0.00004529, 0.00004244, -0.00006836, 0.00005103, -0.00009005, 0.00006063, -0.00012774, 0.00010978, -0.00094228, 0.00194005, -0.00065292, -0.00012917, 0.00002691, -0.00007019, 0.00000039, -0.00001990, 0.00000060, -0.00002509, 0.00002182, -0.00004643, 0.00003848, -0.00003419, 0.00002417, -0.00001232, -0.00000014, -0.00003833, 0.00003384, 0.00004441, -0.00012086, 0.00011750, -0.00004985, 0.00000146, 0.00003493, -0.00004071, 0.00003128, -0.00003312, 0.00003101, 0.00000187, -0.00004022, },
		{ 0.00002677, -0.00001156, 0.00000564, 0.00001426, -0.00000967, -0.00004052, 0.00004685, -0.00001638, 0.00000006, 0.00000269, -0.00002968, 0.00007769, -0.00010166, 0.00009756, -0.00007688, -0.00002096, 0.00003977, 0.00003764, -0.00008065, 0.00006905, -0.00006673, 0.00007236, -0.00009602, 0.00011910, -0.00014755, 0.00014989, -0.00018111, 0.00017812, -0.00023307, 0.00023768, -0.00036149, 0.00045992, -0.00116059, 0.00133817, 0.00042650, -0.00081992, 0.00038953, -0.00032220, 0.00019592, -0.00019306, 0.00015116, -0.00015880, 0.00013575, -0.00013919, 0.00011032, -0.00009150, 0.00007012, -0.00005485, 0.00004708, -0.00007151, 0.00003663, 0.00004670, -0.00007442, 0.00000864, 0.00006645, -0.00009668, 0.00010213, -0.00007575, 0.00004945, -0.00004089, 0.00002443, 0.00001306, -0.00003003, -0.00001325, },
		{ -0.00000514, 0.00000014, 0.00001179, 0.00000562, -0.00004470, 0.00002107, 0.00002127, -0.00003572, 0.00004145, -0.00006871, 0.00009617, -0.00008520, 0.00005025, -0.00000145, -0.00006820, 0.00002731, 0.00006669, -0.00008726, 0.00005961, -0.00006190, 0.00007434, -0.00009743, 0.00011316, -0.00012949, 0.00012102, -0.00014101, 0.00013158, -0.00017521, 0.00017396, -0.00028050, 0.00033329, -0.00069975, 0.00019036, 0.00152518, -0.00118544, 0.00043774, -0.00034323, 0.00021225, -0.00021529, 0.00016680, -0.00016947, 0.00013613, -0.00012965, 0.00009307, -0.00007357, 0.00005509, -0.00004841, 0.00004879, -0.00006177, 0.00001405, 0.00004455, -0.00001223, -0.00008512, 0.00012784, -0.00011898, 0.00008838, -0.00004080, 0.00001424, -0.00000883, -0.00000743, 0.00003192, -0.00002056, -0.00003003, 0.00002786, },
		{ -0.00000322, 0.00000534, 0.00001837, -0.00003944, -0.00000511, 0.00004006, -0.00003769, 0.00003316, -0.00004421, 0.00003959, 0.00000195, -0.00004943, 0.00009240, -0.00010708, 0.00002060, 0.00005766, -0.00003970, -0.00000314, -0.00000322, 0.00001241, -0.00002209, 0.00001965, -0.00001522, -0.00001143, 0.00000956, -0.00003536, 0.00001691, -0.00004640, -0.00000227, -0.00005529, 0.00000212, -0.00082924, 0.00198680, -0.00082924, 0.00000212, -0.00005529, -0.00000227, -0.00004640, 0.00001691, -0.00003536, 0.00000956, -0.00001143, -0.00001522, 0.00001965, -0.00002209, 0.00001241, -0.00000322, -0.00000314, -0.00003970, 0.00005766, 0.00002060, -0.00010708, 0.00009240, -0.00004943, 0.00000195, 0.00003959, -0.00004421, 0.00003316, -0.00003769, 0.00004006, -0.00000511, -0.00003944, 0.00001837, 0.00000534, },
		{ -0.00000514, 0.00002786, -0.00003003, -0.00002056, 0.00003192, -0.00000743, -0.00000883, 0.00001424, -0.00004080, 0.00008838, -0.00011898, 0.00012784, -0.00008512, -0.00001223, 0.00004455, 0.00001405, -0.00006177, 0.00004879, -0.00004841, 0.00005509, -0.00007357, 0.00009307, -0.00012965, 0.00013613, -0.00016947, 0.00016680, -0.00021529, 0.00021225, -0.00034323, 0.00043774, -0.00118544, 0.00152518, 0.00019036, -0.00069975, 0.00033329, -0.00028050, 0.00017396, -0.00017521, 0.00013158, -0.00014101, 0.00012102, -0.00012949, 0.00011316, -0.00009743, 0.00007434, -0.00006190, 0.00005961, -0.00008726, 0.00006669, 0.00002731, -0.00006820, -0.00000145, 0.00005025, -0.00008520, 0.00009617, -0.00006871, 0.00004145, -0.00003572, 0.00002127, 0.00002107, -0.00004470, 0.00000562, 0.00001179, 0.00000014, },
		{ 0.00002677, -0.00001325, -0.00003003, 0.00001306, 0.00002443, -0.00004089, 0.00004945, -0.00007575, 0.00010213, -0.00009668, 0.00006645, 0.00000864, -0.00007442, 0.00004670, 0.00003663, -0.00007151, 0.00004708, -0.00005485, 0.00007012, -0.00009150, 0.00011032, -0.00013919, 0.00013575, -0.00015880, 0.00015116, -0.00019306, 0.00019592, -0.00032220, 0.00038953, -0.00081992, 0.00042650, 0.00133817, -0.00116059, 0.00045992, -0.00036149, 0.00023768, -0.00023307, 0.00017812, -0.00018111, 0.00014989, -0.00014755, 0.00011910, -0.00009602, 0.00007236, -0.00006673, 0.00006905, -0.00008065, 0.00003764, 0.00003977, -0.00002096, -0.00007688, 0.00009756, -0.00010166, 0.00007769, -0.00002968, 0.00000269, 0.00000006, -0.00001638, 0.00004685, -0.00004052, -0.00000967, 0.00001426, 0.00000564, -0.00001156, },
		{ 0.00001248, -0.00004022, 0.00000187, 0.00003101, -0.00003312, 0.00003128, -0.00004071, 0.00003493, 0.00000146, -0.00004985, 0.00011750, -0.00012086, 0.00004441, 0.00003384, -0.00003833, -0.00000014, -0.00001232, 0.00002417, -0.00003419, 0.00003848, -0.00004643, 0.00002182, -0.00002509, 0.00000060, -0.00001990, 0.00000039, -0.00007019, 0.00002691, -0.00012917, -0.00065292, 0.00194005, -0.00094228, 0.00010978, -0.00012774, 0.00006063, -0.00009005, 0.00005103, -0.00006836, 0.00004244, -0.00004529, 0.00002031, -0.00000577, -0.00000411, -0.00000511, 0.00001253, -0.00001336, -0.00003136, 0.00007041, -0.00000092, -0.00008821, 0.00006089, -0.00003955, -0.00000150, 0.00004388, -0.00004939, 0.00003625, -0.00004245, 0.00005051, -0.00001772, -0.00003019, 0.00001721, 0.00000568, -0.00000610, 0.00001065, },
		{ -0.00004588, -0.00000118, 0.00001438, 0.00000342, -0.00001759, 0.00002494, -0.00005167, 0.00009373, -0.00012869, 0.00015680, -0.00009476, 0.00000496, 0.00003644, -0.00000285, -0.00004514, 0.00002925, -0.00002892, 0.00003567, -0.00004782, 0.00005969, -0.00010021, 0.00010882, -0.00014323, 0.00014035, -0.00018006, 0.00016480, -0.00029490, 0.00037024, -0.00113824, 0.00164227, -0.00003009, -0.00055841, 0.00025439, -0.00021990, 0.00013628, -0.00014431, 0.00010086, -0.00011173, 0.00009350, -0.00010535, 0.00010220, -0.00009252, 0.00006912, -0.00005989, 0.00006396, -0.00009475, 0.00009115, 0.00000312, -0.00005249, -0.00001916, 0.00004136, -0.00007401, 0.00008855, -0.00006355, 0.00003490, -0.00003191, 0.00002194, 0.00002139, -0.00005099, 0.00001918, 0.00000052, 0.00000790, -0.00001212, 0.00003677, },
		{ -0.00001015, -0.00000128, 0.00003129, -0.00004729, 0.00005805, -0.00008329, 0.00010444, -0.00010613, 0.00008845, 0.00001139, -0.00006900, 0.00005548, 0.00000734, -0.00005206, 0.00002978, -0.00004122, 0.00005831, -0.00007636, 0.00009436, -0.00013317, 0.00013472, -0.00016116, 0.00015537, -0.00019349, 0.00019663, -0.00033697, 0.00041078, -0.00089817, 0.00064949, 0.00110513, -0.00107339, 0.00043770, -0.00034807, 0.00023868, -0.00023120, 0.00017263, -0.00017631, 0.00014727, -0.00014974, 0.00013253, -0.00011014, 0.00008305, -0.00007944, 0.00008544, -0.00009708, 0.00006321, 0.00002429, -0.00001913, -0.00007390, 0.00007014, -0.00008072, 0.00006449, -0.00002029, -0.00000813, 0.00000777, -0.00002150, 0.00005571, -0.00005624, 0.00001055, -0.00000029, 0.00001682, -0.00002385, 0.00004198, -0.00003355, },
		{ 0.00000122, 0.00002580, -0.00003039, 0.00003128, -0.00003915, 0.00003001, -0.00000340, -0.00003911, 0.00013317, -0.00012452, 0.00006349, 0.00000405, -0.00002906, -0.00000415, -0.00001383, 0.00002841, -0.00003805, 0.00004695, -0.00006759, 0.00004760, -0.00005431, 0.00003242, -0.00005203, 0.00004145, -0.00013180, 0.00010423, -0.00026487, -0.00043251, 0.00180627, -0.00098316, 0.00018040, -0.00017878, 0.00010881, -0.00012277, 0.00007597, -0.00009211, 0.00006676, -0.00007253, 0.00005316, -0.00003218, 0.00001550, -0.00002468, 0.00003181, -0.00002849, -0.00001439, 0.00006987, -0.00001453, -0.00006995, 0.00002753, -0.00002345, -0.00000732, 0.00004689, -0.00005536, 0.00003948, -0.00004533, 0.00005904, -0.00003216, -0.00001522, 0.00001045, 0.00001053, -0.00001370, 0.00002063, 0.00000039, -0.00003219, },
		{ -0.00000172, 0.00001349, -0.00002455, 0.00003320, -0.00006045, 0.00009134, -0.00012725, 0.00018010, -0.00010049, 0.00002537, 0.00001688, -0.00001087, -0.00003397, 0.00001421, -0.00001242, 0.00001858, -0.00002433, 0.00002560, -0.00006577, 0.00007375, -0.00010741, 0.00010374, -0.00013329, 0.00010232, -0.00022417, 0.00026490, -0.00102005, 0.00167507, -0.00021004, -0.00041767, 0.00016797, -0.00015134, 0.00009105, -0.00010707, 0.00006539, -0.00007738, 0.00005912, -0.00007142, 0.00008002, -0.00007793, 0.00005508, -0.00004865, 0.00005889, -0.00009134, 0.00010469, -0.00001912, -0.00003198, -0.00004222, 0.00003940, -0.00006422, 0.00008034, -0.00006087, 0.00002992, -0.00002863, 0.00002492, 0.00001505, -0.00004825, 0.00002553, -0.00000508, 0.00000978, -0.00001323, 0.00003856, -0.00005362, 0.00001303, },
		{ 0.00004008, -0.00005376, 0.00006573, -0.00008935, 0.00010060, -0.00011038, 0.00011311, 0.00000983, -0.00005335, 0.00005055, -0.00001563, -0.00003390, 0.00001210, -0.00002498, 0.00004254, -0.00005592, 0.00006903, -0.00011381, 0.00011869, -0.00014766, 0.00014340, -0.00017578, 0.00017484, -0.00032181, 0.00039060, -0.00091924, 0.00083186, 0.00085494, -0.00094173, 0.00037956, -0.00030736, 0.00021767, -0.00021151, 0.00015227, -0.00015689, 0.00012924, -0.00013561, 0.00013098, -0.00011284, 0.00008415, -0.00008313, 0.00009395, -0.00010639, 0.00008491, 0.00000309, -0.00000805, -0.00007762, 0.00004849, -0.00005960, 0.00005124, -0.00001421, -0.00001712, 0.00001384, -0.00002226, 0.00005676, -0.00006416, 0.00002636, -0.00001203, 0.00002482, -0.00003308, 0.00005349, -0.00005061, 0.00001019, -0.00001863, },
		{ -0.00003017, 0.00003300, -0.00003891, 0.00002393, -0.00001157, -0.00001765, 0.00013908, -0.00011545, 0.00007148, -0.00002512, -0.00001597, -0.00001269, -0.00000915, 0.00002573, -0.00003387, 0.00004416, -0.00007574, 0.00006124, -0.00007247, 0.00005417, -0.00007321, 0.00006910, -0.00017678, 0.00016242, -0.00038377, -0.00019425, 0.00160352, -0.00095444, 0.00020859, -0.00020198, 0.00013633, -0.00013965, 0.00008773, -0.00010273, 0.00007793, -0.00008747, 0.00007695, -0.00005377, 0.00003187, -0.00004167, 0.00005018, -0.00004445, 0.00000747, 0.00005730, -0.00001741, -0.00005693, -0.00000296, -0.00000501, -0.00001341, 0.00004740, -0.00006106, 0.00004195, -0.00004476, 0.00006257, -0.00004401, 0.00000130, 0.00000122, 0.00001712, -0.00002327, 0.00003247, -0.00001487, -0.00001764, -0.00000661, 0.00002518, },
		{ -0.00002833, 0.00003797, -0.00006611, 0.00007994, -0.00011258, 0.00019480, -0.00009733, 0.00004195, -0.00000944, -0.00001086, -0.00002915, 0.00000554, -0.00000152, 0.00000699, -0.00000756, -0.00000302, -0.00003342, 0.00003809, -0.00006889, 0.00006400, -0.00008324, 0.00003495, -0.00014331, 0.00013656, -0.00084299, 0.00162145, -0.00033303, -0.00029594, 0.00008845, -0.00008597, 0.00004691, -0.00007066, 0.00003200, -0.00004505, 0.00002518, -0.00003448, 0.00005175, -0.00005728, 0.00003522, -0.00003046, 0.00004571, -0.00007718, 0.00010457, -0.00003366, -0.00001225, -0.00006665, 0.00004235, -0.00005572, 0.00007180, -0.00006055, 0.00002615, -0.00002455, 0.00002764, 0.00000507, -0.00003831, 0.00002514, -0.00000517, 0.00000587, -0.00000841, 0.00003299, -0.00005184, 0.00001908, -0.00001306, 0.00002035, },
		{ 0.00007072, -0.00009193, 0.00008830, -0.00010639, 0.00013672, 0.00000889, -0.00003247, 0.00003264, -0.00002913, -0.00002088, -0.00000196, -0.00001020, 0.00002700, -0.00003513, 0.00003999, -0.00008631, 0.00009182, -0.00012141, 0.00011801, -0.00014313, 0.00013402, -0.00027970, 0.00033030, -0.00087680, 0.00095266, 0.00061598, -0.00078760, 0.00029905, -0.00024826, 0.00018094, -0.00017912, 0.00012197, -0.00012779, 0.00010021, -0.00010825, 0.00011552, -0.00010388, 0.00007506, -0.00007653, 0.00009243, -0.00010539, 0.00009745, -0.00001758, 0.00000837, -0.00008696, 0.00003370, -0.00004060, 0.00003964, -0.00001230, -0.00002381, 0.00001866, -0.00001969, 0.00005049, -0.00006298, 0.00003521, -0.00001843, 0.00002717, -0.00003656, 0.00005826, -0.00006032, 0.00002603, -0.00003472, 0.00004804, -0.00005853, },
		{ -0.00003857, 0.00001498, -0.00002081, 0.00001203, 0.00013799, -0.00009483, 0.00006463, -0.00004757, -0.00000396, -0.00002195, -0.00000139, 0.00001868, -0.00002416, 0.00003215, -0.00007133, 0.00006141, -0.00007691, 0.00006262, -0.00007995, 0.00007878, -0.00019834, 0.00019089, -0.00046707, 0.00003347, 0.00135785, -0.00086864, 0.00019726, -0.00019711, 0.00014199, -0.00013963, 0.00008571, -0.00009972, 0.00007476, -0.00008758, 0.00008764, -0.00006611, 0.00004098, -0.00005196, 0.00006325, -0.00005660, 0.00002889, 0.00003688, -0.00001000, -0.00005154, -0.00002691, 0.00001225, -0.00001779, 0.00004437, -0.00006565, 0.00004333, -0.00004035, 0.00005942, -0.00004972, 0.00001515, -0.00000688, 0.00002211, -0.00003135, 0.00004250, -0.00002902, -0.00000078, -0.00001889, 0.00002803, -0.00003184, 0.00003544, },
		{ -0.00006875, 0.00005996, -0.00008491, 0.00020060, -0.00008254, 0.00004786, -0.00003587, -0.00000624, -0.00002914, 0.00000286, 0.00000330, 0.00000189, 0.00000035, -0.00002201, -0.00000892, 0.00000846, -0.00003451, 0.00002828, -0.00003850, -0.00002641, -0.00006640, 0.00000447, -0.00062821, 0.00149159, -0.00039333, -0.00020501, 0.00002653, -0.00003260, 0.00001090, -0.00004089, 0.00000609, -0.00002061, -0.00000165, -0.00000160, 0.00002358, -0.00003554, 0.00001403, -0.00000939, 0.00002789, -0.00005521, 0.00009131, -0.00003725, 0.00000187, -0.00008810, 0.00004739, -0.00004770, 0.00006262, -0.00006208, 0.00002320, -0.00001863, 0.00002752, -0.00000460, -0.00002479, 0.00002041, -0.00000179, -0.00000180, 0.00000035, 0.00002213, -0.00004183, 0.00001645, -0.00001798, 0.00002262, -0.00002839, 0.00003916, },
		{ 0.00006628, -0.00009235, 0.00015653, 0.00001373, -0.00001373, 0.00000621, -0.00003321, -0.00001462, -0.00001001, 0.00000013, 0.00001510, -0.00001846, 0.00001317, -0.00005720, 0.00006040, -0.00008811, 0.00008475, -0.00010194, 0.00008162, -0.00021907, 0.00023896, -0.00077503, 0.00100122, 0.00041112, -0.00063247, 0.00021135, -0.00018170, 0.00013669, -0.00014080, 0.00008810, -0.00009555, 0.00006681, -0.00007362, 0.00009032, -0.00008588, 0.00005776, -0.00006094, 0.00008122, -0.00009343, 0.00009772, -0.00003207, 0.00002504, -0.00009902, 0.00002515, -0.00002476, 0.00003043, -0.00001468, -0.00002820, 0.00002315, -0.00001595, 0.00003947, -0.00005393, 0.00003681, -0.00001892, 0.00002327, -0.00003348, 0.00005523, -0.00006054, 0.00003373, -0.00004572, 0.00005240, -0.00005990, 0.00007166, -0.00008972, },
		{ -0.00002849, 0.00004635, 0.00013514, -0.00006786, 0.00004325, -0.00005953, 0.00000291, -0.00002812, 0.00000582, 0.00001072, -0.00001280, 0.00001518, -0.00005786, 0.00005027, -0.00006846, 0.00005793, -0.00007241, 0.00007018, -0.00019485, 0.00018532, -0.00050238, 0.00022539, 0.00109819, -0.00074497, 0.00015600, -0.00016939, 0.00012904, -0.00012532, 0.00007259, -0.00008589, 0.00005970, -0.00007407, 0.00008452, -0.00006742, 0.00004089, -0.00005317, 0.00006799, -0.00006126, 0.00004465, 0.00001438, 0.00000451, -0.00005371, -0.00004242, 0.00002583, -0.00001915, 0.00003729, -0.00006873, 0.00004404, -0.00003309, 0.00004997, -0.00004770, 0.00002344, -0.00001108, 0.00002280, -0.00003500, 0.00004746, -0.00003789, 0.00001348, -0.00003159, 0.00003189, -0.00003389, 0.00003705, -0.00003643, 0.00000126, },
		{ -0.00004687, 0.00020058, -0.00005724, 0.00003875, -0.00005587, -0.00000170, -0.00003075, 0.00000387, 0.00000357, 0.00000204, -0.00000001, -0.00003022, 0.00000475, -0.00001078, -0.00000931, 0.00000206, -0.00000580, -0.00007270, -0.00000611, -0.00011172, -0.00040184, 0.00130562, -0.00039591, -0.00014864, -0.00001258, 0.00000394, -0.00001307, -0.00002081, -0.00000963, -0.00000717, -0.00001713, 0.00002174, 0.00000103, -0.00001752, -0.00000384, 0.00000992, 0.00000997, -0.00003026, 0.00006849, -0.00002973, 0.00000756, -0.00010301, 0.00005176, -0.00003919, 0.00005241, -0.00006491, 0.00002103, -0.00001071, 0.00002303, -0.00001050, -0.00001177, 0.00001466, 0.00000205, -0.00001008, 0.00000979, 0.00000971, -0.00002716, 0.00000712, -0.00001693, 0.00002035, -0.00002528, 0.00003768, -0.00006964, 0.00003344, },
		{ 0.00017230, 0.00002761, -0.00000486, -0.00002174, -0.00003097, -0.00001413, -0.00001191, 0.00000495, 0.00000843, -0.00000859, -0.00000689, -0.00003241, 0.00003100, -0.00005441, 0.00005043, -0.00006008, 0.00002727, -0.00015172, 0.00013161, -0.00062794, 0.00097868, 0.00025418, -0.00049334, 0.00012993, -0.00011799, 0.00009295, -0.00010315, 0.00005670, -0.00006647, 0.00003597, -0.00003882, 0.00006137, -0.00006343, 0.00003624, -0.00003980, 0.00006307, -0.00007263, 0.00008561, -0.00003684, 0.00003728, -0.00011015, 0.00002112, -0.00001215, 0.00002368, -0.00002097, -0.00003043, 0.00002810, -0.00001343, 0.00002738, -0.00004025, 0.00003304, -0.00001499, 0.00001456, -0.00002515, 0.00004569, -0.00005173, 0.00003202, -0.00004941, 0.00005137, -0.00005699, 0.00006811, -0.00008265, 0.00003501, -0.00006869, },
		{ 0.00013661, -0.00004285, 0.00001194, -0.00006066, 0.00000275, -0.00002882, 0.00000965, 0.00000496, -0.00000370, -0.00000171, -0.00004058, 0.00003250, -0.00005096, 0.00004344, -0.00005423, 0.00004727, -0.00017018, 0.00014870, -0.00048637, 0.00036394, 0.00085107, -0.00060492, 0.00009814, -0.00012762, 0.00010381, -0.00010189, 0.00005322, -0.00006634, 0.00003791, -0.00005139, 0.00007028, -0.00005887, 0.00003227, -0.00004540, 0.00006369, -0.00005689, 0.00005114, -0.00000455, 0.00002149, -0.00006129, -0.00004939, 0.00003450, -0.00001701, 0.00002627, -0.00007026, 0.00004492, -0.00002507, 0.00003658, -0.00003884, 0.00002539, -0.00001036, 0.00001814, -0.00003288, 0.00004574, -0.00003884, 0.00002119, -0.00004072, 0.00003397, -0.00003456, 0.00003642, -0.00003122, -0.00001836, -0.00003274, 0.00008208, },
		{ -0.00002705, 0.00001450, -0.00006504, -0.00000144, -0.00003046, 0.00000546, 0.00000200, 0.00000463, -0.00000577, -0.00002949, 0.00000785, -0.00001842, 0.00000459, -0.00001219, 0.00001149, -0.00009880, 0.00002927, -0.00019610, -0.00019000, 0.00108931, -0.00035415, -0.00012309, -0.00002943, 0.00002328, -0.00002467, -0.00001036, -0.00001565, -0.00000446, -0.00002046, 0.00003307, -0.00001256, -0.00000647, -0.00001498, 0.00002365, -0.00000386, -0.00000764, 0.00004153, -0.00001380, 0.00000443, -0.00010942, 0.00005346, -0.00002954, 0.00004101, -0.00006868, 0.00002013, -0.00000185, 0.00001446, -0.00001084, -0.00000238, 0.00001082, 0.00000352, -0.00001592, 0.00001657, -0.00000016, -0.00001249, -0.00000519, -0.00001206, 0.00001489, -0.00002037, 0.00003511, -0.00007067, 0.00000335, -0.00000277, 0.00020083, },
		{ -0.00001135, -0.00004381, -0.00002724, -0.00001643, -0.00000957, 0.00000526, 0.00000648, -0.00000578, -0.00001818, -0.00001577, 0.00000877, -0.00002608, 0.00002118, -0.00002489, -0.00001968, -0.00009005, 0.00002590, -0.00045638, 0.00089696, 0.00014861, -0.00038022, 0.00006396, -0.00006463, 0.00005566, -0.00007095, 0.00003190, -0.00004494, 0.00001300, -0.00001014, 0.00003475, -0.00004167, 0.00001527, -0.00001769, 0.00004224, -0.00004727, 0.00006401, -0.00003127, 0.00004212, -0.00011720, 0.00001956, -0.00000233, 0.00001906, -0.00003055, -0.00003045, 0.00003358, -0.00001368, 0.00001776, -0.00002596, 0.00002701, -0.00000941, 0.00000393, -0.00001448, 0.00003286, -0.00003690, 0.00002232, -0.00004577, 0.00004473, -0.00005007, 0.00006086, -0.00007204, -0.00000314, -0.00003843, 0.00018721, 0.00004996, },
		{ -0.00002149, -0.00005410, -0.00000354, -0.00002384, 0.00000899, 0.00000299, 0.00000058, -0.00001434, -0.00002483, 0.00001382, -0.00002990, 0.00002445, -0.00003133, 0.00001706, -0.00013252, 0.00009054, -0.00042523, 0.00044190, 0.00063627, -0.00046788, 0.00003734, -0.00008158, 0.00007372, -0.00007536, 0.00003306, -0.00004678, 0.00001568, -0.00002587, 0.00005004, -0.00004412, 0.00001811, -0.00003107, 0.00005206, -0.00004457, 0.00004743, -0.00001599, 0.00003630, -0.00007111, -0.00004908, 0.00003811, -0.00001159, 0.00001185, -0.00007028, 0.00004669, -0.00001868, 0.00002283, -0.00002608, 0.00002247, -0.00000574, 0.00000909, -0.00002573, 0.00003797, -0.00003173, 0.00002053, -0.00004371, 0.00003223, -0.00003271, 0.00003296, -0.00002284, -0.00004344, -0.00003369, 0.00011815, 0.00014708, -0.00002904, },
		{ -0.00006266, -0.00000757, -0.00002587, 0.00000497, 0.00000124, 0.00000648, -0.00001296, -0.00002371, 0.00000342, -0.00001635, 0.00000824, -0.00001537, 0.00001406, -0.00010453, 0.00003762, -0.00023977, -0.00001396, 0.00086897, -0.00028610, -0.00011927, -0.00002916, 0.00002880, -0.00002655, -0.00000688, -0.00001507, -0.00000916, -0.00001427, 0.00003331, -0.00001674, -0.00000319, -0.00001814, 0.00002992, -0.00001110, 0.00000849, 0.00001614, 0.00000617, -0.00000574, -0.00010712, 0.00005155, -0.00001866, 0.00002856, -0.00007319, 0.00002127, 0.00000591, 0.00000393, -0.00000609, 0.00000216, 0.00001032, 0.00000116, -0.00001749, 0.00001858, -0.00000446, -0.00000206, -0.00001639, -0.00000624, 0.00000814, -0.00001522, 0.00003287, -0.00007337, -0.00002761, 0.00004292, 0.00020890, -0.00000135, -0.00002019, },
		{ -0.00002680, -0.00001782, -0.00000603, 0.00000340, 0.00000707, -0.00000803, -0.00002145, -0.00000819, -0.00000382, -0.00000660, 0.00000098, -0.00000130, -0.00005239, -0.00004397, -0.00006179, -0.00028383, 0.00077560, 0.00008855, -0.00029557, 0.00001727, -0.00002516, 0.00002770, -0.00004630, 0.00001511, -0.00003236, 0.00000027, 0.00000857, 0.00001496, -0.00002479, -0.00000104, 0.00000110, 0.00002312, -0.00002249, 0.00003789, -0.00001749, 0.00003886, -0.00011826, 0.00001880, 0.00000515, 0.00001621, -0.00004281, -0.00002782, 0.00003861, -0.00001666, 0.00001271, -0.00001449, 0.00002184, -0.00000514, -0.00000540, -0.00000487, 0.00002081, -0.00002054, 0.00000813, -0.00003681, 0.00003378, -0.00004056, 0.00005161, -0.00006023, -0.00004445, -0.00000669, 0.00020790, 0.00007534, -0.00003439, -0.00005446, },
		{ -0.00001268, -0.00001516, 0.00000474, 0.00000449, -0.00000039, -0.00002055, -0.00001443, -0.00000082, -0.00001071, 0.00000657, -0.00001019, -0.00001241, -0.00009211, 0.00002451, -0.00033303, 0.00046278, 0.00046431, -0.00034790, -0.00001537, -0.00003971, 0.00004529, -0.00005094, 0.00001657, -0.00003185, -0.00000150, -0.00000380, 0.00002970, -0.00002799, 0.00000271, -0.00001418, 0.00003663, -0.00002766, 0.00003546, -0.00001863, 0.00004570, -0.00008008, -0.00004341, 0.00003722, -0.00000357, -0.00000519, -0.00006865, 0.00004942, -0.00001556, 0.00001232, -0.00001339, 0.00001761, 0.00000023, -0.00000172, -0.00001611, 0.00002691, -0.00001899, 0.00001243, -0.00004015, 0.00002618, -0.00002841, 0.00002732, -0.00001270, -0.00007154, -0.00003409, 0.00015698, 0.00016748, -0.00003392, -0.00004813, -0.00004542, },
		{ -0.00001676, 0.00000134, 0.00000276, 0.00000539, -0.00001798, -0.00001731, -0.00000404, -0.00000855, 0.00000509, -0.00001113, 0.00000605, -0.00009434, 0.00002325, -0.00024240, 0.00011335, 0.00066673, -0.00021013, -0.00012584, -0.00001934, 0.00002597, -0.00002312, -0.00000640, -0.00001227, -0.00001628, -0.00000356, 0.00002624, -0.00001384, -0.00000602, -0.00001449, 0.00002924, -0.00001166, 0.00001627, -0.00000327, 0.00002550, -0.00001993, -0.00009736, 0.00004610, -0.00000700, 0.00001548, -0.00007817, 0.00002504, 0.00001028, -0.00000517, 0.00000114, 0.00000285, 0.00001264, -0.00000461, -0.00001491, 0.00001580, -0.00000239, 0.00000179, -0.00002344, -0.00000191, 0.00000167, -0.00001080, 0.00003143, -0.00007798, -0.00005842, 0.00008753, 0.00023146, 0.00000874, -0.00005654, -0.00005201, -0.00001925, },
		{ -0.00000410, 0.00000191, 0.00000744, -0.00001207, -0.00001964, -0.00000782, -0.00000727, 0.00000342, -0.00000917, 0.00000914, -0.00006807, -0.00001871, -0.00011990, -0.00013163, 0.00063726, 0.00006165, -0.00023545, -0.00001140, 0.00000078, 0.00000882, -0.00002861, 0.00000505, -0.00002714, -0.00000322, 0.00001672, 0.00000382, -0.00001481, -0.00001037, 0.00001382, 0.00000891, -0.00000275, 0.00001278, 0.00000048, 0.00002892, -0.00011301, 0.00001790, 0.00001055, 0.00001482, -0.00005713, -0.00002193, 0.00004131, -0.00002055, 0.00001216, -0.00000749, 0.00001943, -0.00000413, -0.00001113, 0.00000107, 0.00001295, -0.00000714, -0.00000625, -0.00002565, 0.00002076, -0.00003033, 0.00004224, -0.00004961, -0.00008503, 0.00002076, 0.00024316, 0.00009376, -0.00006972, -0.00005201, -0.00003256, -0.00001546, },
		{ -0.00000064, 0.00000738, -0.00000491, -0.00002075, -0.00001069, -0.00000856, 0.00000274, -0.00000592, 0.00000403, -0.00003425, -0.00005851, -0.00003493, -0.00022837, 0.00043882, 0.00033629, -0.00025204, -0.00005347, -0.00000736, 0.00002274, -0.00003179, 0.00000603, -0.00002379, -0.00001048, 0.00001031, 0.00001426, -0.00001494, -0.00000971, 0.00000107, 0.00002154, -0.00001070, 0.00001934, -0.00001386, 0.00004848, -0.00008604, -0.00003425, 0.00003267, 0.00000622, -0.00002405, -0.00006485, 0.00005214, -0.00001588, 0.00000736, -0.00000437, 0.00001407, 0.00000448, -0.00001101, -0.00000741, 0.00001648, -0.00000491, 0.00000015, -0.00003191, 0.00001704, -0.00002297, 0.00002132, -0.00000350, -0.00009854, -0.00003921, 0.00020490, 0.00019347, -0.00006059, -0.00006057, -0.00004064, -0.00002025, -0.00000622, },
		{ 0.00000616, 0.00000102, -0.00001895, -0.00001369, -0.00001019, 0.00000037, -0.00000038, -0.00000445, -0.00000646, -0.00007571, -0.00000459, -0.00021168, 0.00018889, 0.00049716, -0.00014124, -0.00013228, -0.00000761, 0.00002057, -0.00001882, -0.00000512, -0.00001134, -0.00002085, 0.00000597, 0.00001709, -0.00000786, -0.00001154, -0.00000707, 0.00002409, -0.00000773, 0.00001656, -0.00001470, 0.00004058, -0.00003489, -0.00008217, 0.00003783, 0.00000474, 0.00000225, -0.00008303, 0.00003130, 0.00000967, -0.00000935, 0.00000707, 0.00000237, 0.00001555, -0.00001153, -0.00001027, 0.00001036, 0.00000427, -0.00000063, -0.00002544, -0.00000005, -0.00000409, -0.00000681, 0.00002971, -0.00008265, -0.00009075, 0.00013192, 0.00027158, -0.00000672, -0.00008383, -0.00003951, -0.00003283, -0.00000528, -0.00000446, },
		{ 0.00000553, -0.00001474, -0.00001650, -0.00001111, -0.00000438, 0.00000596, -0.00001109, 0.00000835, -0.00006832, -0.00001377, -0.00014394, -0.00001502, 0.00050294, 0.00005299, -0.00019208, -0.00002678, 0.00001638, -0.00000342, -0.00001551, -0.00000129, -0.00002557, -0.00000115, 0.00001679, 0.00000031, -0.00001131, -0.00001279, 0.00001990, 0.00000067, 0.00000946, -0.00000679, 0.00001813, 0.00001508, -0.00010245, 0.00001667, 0.00001387, 0.00001466, -0.00007271, -0.00001238, 0.00003953, -0.00002225, 0.00001386, -0.00000441, 0.00001975, -0.00000649, -0.00001274, 0.00000260, 0.00001079, 0.00000033, -0.00001726, -0.00001527, 0.00000779, -0.00002086, 0.00003389, -0.00004147, -0.00012235, 0.00003953, 0.00030151, 0.00009249, -0.00010804, -0.00003938, -0.00004422, -0.00000862, -0.00000517, 0.00000236, },
		{ -0.00000992, -0.00001745, -0.00001232, -0.00000920, 0.00000905, -0.00001111, 0.00000880, -0.00004451, -0.00003813, -0.00007601, -0.00013010, 0.00038739, 0.00024566, -0.00018069, -0.00007536, 0.00001375, 0.00000747, -0.00001863, 0.00000128, -0.00002207, -0.00001120, 0.00001492, 0.00000638, -0.00000776, -0.00001631, 0.00001153, 0.00001014, 0.00000209, 0.00000399, -0.00000498, 0.00004545, -0.00008815, -0.00002305, 0.00002527, 0.00001694, -0.00004389, -0.00005816, 0.00005296, -0.00001796, 0.00000812, -0.00000099, 0.00001406, 0.00000473, -0.00001608, -0.00000263, 0.00001041, 0.00000584, -0.00001187, -0.00002226, 0.00000724, -0.00001849, 0.00001727, 0.00000161, -0.00011969, -0.00005553, 0.00027123, 0.00021519, -0.00010605, -0.00005537, -0.00004410, -0.00002239, -0.00000045, -0.00000430, 0.00000889, },
		{ -0.00001624, -0.00001412, -0.00001228, 0.00000674, -0.00000424, 0.00000005, -0.00001735, -0.00005680, -0.00003435, -0.00016105, 0.00021869, 0.00036606, -0.00008873, -0.00013125, 0.00000031, 0.00001696, -0.00001687, -0.00000067, -0.00001467, -0.00001958, 0.00000979, 0.00001077, -0.00000310, -0.00001589, 0.00000039, 0.00001797, -0.00000293, 0.00001245, -0.00001873, 0.00004966, -0.00004810, -0.00006373, 0.00002778, 0.00001574, -0.00001071, -0.00008678, 0.00003884, 0.00000398, -0.00000634, 0.00000809, 0.00000397, 0.00001603, -0.00001642, -0.00000682, 0.00000594, 0.00001161, -0.00000644, -0.00002389, 0.00000022, -0.00001016, -0.00000157, 0.00002502, -0.00008349, -0.00012925, 0.00018146, 0.00032664, -0.00005335, -0.00009230, -0.00003277, -0.00004298, 0.00000467, -0.00000974, 0.00000929, -0.00000489, },
		},
		{ { 0.00000925, 0.00000568, -0.00002040, -0.00001226, 0.00002571, -0.00001080, -0.00001692, -0.00001133, -0.00001936, -0.00001444, -0.00014561, 0.00005300, 0.00038050, 0.00005830, -0.00016410, -0.00001296, -0.00002063, -0.00001161, 0.00000475, -0.00000783, -0.00002448, -0.00001050, 0.00004607, -0.00000587, -0.00001020, -0.00000962, -0.00000370, 0.00001337, -0.00007466, 0.00000686, 0.00005958, -0.00002237, -0.00002711, -0.00002237, 0.00005958, 0.00000686, -0.00007466, 0.00001337, -0.00000370, -0.00000962, -0.00001020, -0.00000587, 0.00004607, -0.00001050, -0.00002448, -0.00000783, 0.00000475, -0.00001161, -0.00002063, -0.00001296, -0.00016410, 0.00005830, 0.00038050, 0.00005300, -0.00014561, -0.00001444, -0.00001936, -0.00001133, -0.00001692, -0.00001080, 0.00002571, -0.00001226, -0.00002040, 0.00000568, },
		{ 0.00000712, -0.00000710, -0.00002528, 0.00001789, 0.00000788, -0.00002139, -0.00000996, -0.00002084, -0.00000059, -0.00010083, -0.00006786, 0.00029257, 0.00021925, -0.00013178, -0.00006892, -0.00000180, -0.00002748, 0.00000808, -0.00000571, -0.00001391, -0.00002692, 0.00003189, 0.00002119, -0.00002026, 0.00000187, -0.00002157, 0.00002572, -0.00004899, -0.00004418, 0.00006714, 0.00000035, -0.00002614, -0.00003368, 0.00003196, 0.00005002, -0.00006771, -0.00001967, 0.00001444, -0.00001449, -0.00000496, -0.00001920, 0.00003457, 0.00001990, -0.00003256, -0.00001085, -0.00000218, 0.00000212, -0.00002897, 0.00000396, -0.00011301, -0.00011285, 0.00035358, 0.00023450, -0.00015620, -0.00005108, -0.00001235, -0.00002012, -0.00000647, -0.00002469, 0.00001907, 0.00000677, -0.00002600, -0.00000367, 0.00001088, },
		{ 0.00000224, -0.00002360, -0.00000167, 0.00002317, -0.00001728, -0.00001204, -0.00002140, 0.00000187, -0.00005517, -0.00011664, 0.00015457, 0.00030059, -0.00002259, -0.00012055, -0.00000475, -0.00002531, -0.00000419, 0.00000433, -0.00001243, -0.00002252, 0.00000097, 0.00004439, -0.00001968, 0.00000358, -0.00002061, 0.00000880, -0.00000412, -0.00007975, 0.00004612, 0.00002943, -0.00002218, -0.00003400, -0.00000209, 0.00006960, -0.00003291, -0.00005272, 0.00001641, -0.00000709, -0.00000642, -0.00001960, 0.00001059, 0.00004162, -0.00002143, -0.00002380, -0.00000245, 0.00000137, -0.00001635, -0.00001384, -0.00003476, -0.00020221, 0.00018926, 0.00041553, -0.00008722, -0.00011445, -0.00000273, -0.00003232, 0.00000115, -0.00002660, 0.00000135, 0.00002116, -0.00001876, -0.00001755, 0.00001054, 0.00000768, },
		{ -0.00001039, -0.00002143, 0.00002649, -0.00000543, -0.00001329, -0.00002275, -0.00000144, -0.00002111, -0.00011395, 0.00002818, 0.00028635, 0.00011007, -0.00012179, -0.00003961, -0.00001014, -0.00002151, 0.00001146, -0.00001068, -0.00001410, -0.00001997, 0.00004067, 0.00000402, -0.00001210, 0.00000013, -0.00002233, 0.00003094, -0.00007893, -0.00000095, 0.00005455, -0.00001320, -0.00002682, -0.00003017, 0.00006207, 0.00001295, -0.00006699, -0.00000127, 0.00000577, -0.00000819, -0.00001491, -0.00001060, 0.00004239, 0.00000588, -0.00003567, -0.00000557, -0.00000184, -0.00000388, -0.00002628, 0.00000231, -0.00016875, -0.00004822, 0.00049222, 0.00009241, -0.00018536, -0.00000547, -0.00003762, -0.00000311, -0.00001550, -0.00001885, 0.00002363, -0.00000279, -0.00002587, 0.00000121, 0.00001258, 0.00000307, },
		{ -0.00002662, 0.00001240, 0.00001202, -0.00001388, -0.00002016, -0.00001225, 0.00000342, -0.00008949, -0.00005119, 0.00020219, 0.00021123, -0.00006456, -0.00007923, -0.00000841, -0.00002319, 0.00000114, 0.00000233, -0.00001618, -0.00001758, 0.00001079, 0.00003744, -0.00002779, 0.00001716, -0.00003625, 0.00002920, -0.00003812, -0.00005653, 0.00006234, 0.00000224, -0.00001657, -0.00004456, 0.00003414, 0.00005083, -0.00005474, -0.00002960, 0.00001343, -0.00000612, -0.00000899, -0.00002338, 0.00002754, 0.00003031, -0.00002875, -0.00002210, 0.00000376, -0.00000703, -0.00001393, -0.00001231, -0.00006798, -0.00022533, 0.00037757, 0.00035420, -0.00020913, -0.00004361, -0.00002734, -0.00002074, 0.00000036, -0.00003002, 0.00001158, 0.00001430, -0.00002324, -0.00001381, 0.00001537, 0.00000368, 0.00000025, },
		{ -0.00001181, 0.00002670, -0.00001234, -0.00001206, -0.00002785, 0.00001841, -0.00006045, -0.00008193, 0.00009455, 0.00024818, 0.00002725, -0.00009172, -0.00003005, -0.00001159, -0.00001732, 0.00001503, -0.00001856, -0.00000863, -0.00001561, 0.00004788, -0.00001626, 0.00000631, -0.00001517, -0.00000838, 0.00001774, -0.00009044, 0.00004074, 0.00002413, -0.00000816, -0.00004194, -0.00000322, 0.00006883, -0.00002227, -0.00005261, 0.00000778, 0.00000024, -0.00000406, -0.00002593, 0.00000565, 0.00004048, -0.00000491, -0.00003947, 0.00000497, -0.00001045, -0.00000100, -0.00002918, -0.00000176, -0.00024002, 0.00009286, 0.00057771, -0.00010593, -0.00013246, -0.00000302, -0.00004460, 0.00000969, -0.00002636, -0.00000919, 0.00002375, -0.00001103, -0.00002463, 0.00000724, 0.00001138, 0.00000022, -0.00001485, },
		{ 0.00002478, -0.00000217, -0.00000636, -0.00003507, 0.00001482, -0.00002635, -0.00008653, 0.00000968, 0.00021343, 0.00012355, -0.00007082, -0.00005338, -0.00001220, -0.00001963, 0.00000678, -0.00000391, -0.00001449, -0.00001455, 0.00002162, 0.00002263, -0.00002392, 0.00002058, -0.00004528, 0.00004550, -0.00007623, -0.00001092, 0.00004741, -0.00000404, -0.00002448, -0.00003815, 0.00006422, 0.00001565, -0.00005661, -0.00001233, 0.00000885, -0.00000312, -0.00001669, -0.00001856, 0.00003904, 0.00001703, -0.00003503, -0.00001570, 0.00000413, -0.00001222, -0.00001140, -0.00001234, -0.00012121, -0.00020321, 0.00059630, 0.00016823, -0.00024463, 0.00001195, -0.00005974, 0.00000468, -0.00001262, -0.00002538, 0.00001760, 0.00000704, -0.00002634, -0.00000870, 0.00001882, -0.00000135, -0.00000156, -0.00002654, },
		{ 0.00001686, -0.00000889, -0.00002456, -0.00001038, 0.00001230, -0.00008374, -0.00003211, 0.00012913, 0.00019427, -0.00002550, -0.00005660, -0.00003439, -0.00000497, -0.00001661, 0.00001833, -0.00002683, 0.00000048, -0.00001354, 0.00005178, -0.00003456, 0.00002915, -0.00004058, 0.00001987, -0.00001772, -0.00007018, 0.00005695, 0.00000081, -0.00000287, -0.00005716, 0.00003848, 0.00004695, -0.00004015, -0.00003693, 0.00001253, -0.00000404, -0.00000205, -0.00003556, 0.00002693, 0.00002821, -0.00000974, -0.00004487, 0.00002005, -0.00002843, 0.00001441, -0.00004426, -0.00000620, -0.00032176, 0.00033381, 0.00053216, -0.00027978, -0.00003096, -0.00004524, -0.00001977, 0.00000357, -0.00002842, -0.00000234, 0.00002307, -0.00001937, -0.00002126, 0.00001328, 0.00000752, -0.00000115, -0.00001984, 0.00000290, },
		{ -0.00001039, -0.00000678, -0.00003842, 0.00003677, -0.00006821, -0.00004792, 0.00004390, 0.00020758, 0.00004188, -0.00004662, -0.00005170, -0.00000412, -0.00002314, 0.00001758, -0.00001777, -0.00000282, -0.00001898, 0.00003705, -0.00000204, -0.00000364, 0.00000502, -0.00003513, 0.00004127, -0.00009630, 0.00003350, 0.00001598, 0.00000757, -0.00004901, -0.00000410, 0.00006666, -0.00001393, -0.00004945, 0.00000163, 0.00000107, 0.00000365, -0.00003127, -0.00000084, 0.00003675, 0.00001092, -0.00004747, 0.00000306, -0.00001249, 0.00000113, -0.00002660, 0.00000202, -0.00021542, -0.00008057, 0.00076461, -0.00010506, -0.00016666, 0.00000513, -0.00006099, 0.00001754, -0.00002296, -0.00002050, 0.00002354, -0.00000285, -0.00002614, -0.00000349, 0.00002060, -0.00000645, -0.00000414, -0.00002054, 0.00002900, },
		{ -0.00000118, -0.00004281, 0.00002488, -0.00002675, -0.00006622, -0.00000000, 0.00015163, 0.00012429, -0.00003869, -0.00003824, -0.00003487, -0.00000104, -0.00001212, 0.00001429, -0.00002686, 0.00000440, -0.00000686, 0.00004531, -0.00004021, 0.00004266, -0.00006165, 0.00004835, -0.00006234, -0.00002425, 0.00004028, 0.00000185, -0.00001701, -0.00004935, 0.00006939, 0.00001232, -0.00004216, -0.00002286, 0.00001248, -0.00000428, -0.00000655, -0.00003619, 0.00004315, 0.00001434, -0.00001549, -0.00004312, 0.00002722, -0.00004094, 0.00002745, -0.00005823, -0.00003153, -0.00037320, 0.00064880, 0.00030939, -0.00033644, 0.00004495, -0.00009102, 0.00001758, -0.00001485, -0.00002368, 0.00000290, 0.00002020, -0.00002593, -0.00001697, 0.00001939, 0.00000101, -0.00000132, -0.00002283, 0.00001761, 0.00000246, },
		{ -0.00002142, -0.00001610, 0.00002703, -0.00008580, -0.00000626, 0.00006211, 0.00018457, -0.00001910, -0.00001159, -0.00006561, 0.00001432, -0.00003311, 0.00003119, -0.00003510, 0.00001523, -0.00002890, 0.00005402, -0.00003095, 0.00002764, -0.00002735, -0.00000398, 0.00000984, -0.00008253, 0.00005115, -0.00000493, 0.00001519, -0.00007147, 0.00004491, 0.00003927, -0.00002500, -0.00004300, 0.00001489, -0.00001016, 0.00001351, -0.00005040, 0.00002800, 0.00001906, 0.00001535, -0.00006746, 0.00003418, -0.00004770, 0.00003840, -0.00006890, 0.00003814, -0.00035835, 0.00019204, 0.00077042, -0.00035797, -0.00002116, -0.00006160, -0.00001861, 0.00000277, -0.00001969, -0.00002151, 0.00003235, -0.00001661, -0.00002158, 0.00000102, 0.00002033, -0.00001095, -0.00000699, -0.00000901, 0.00002291, -0.00000846, },
		{ -0.00004766, 0.00005216, -0.00007483, -0.00001740, 0.00000342, 0.00017346, 0.00004175, -0.00001213, -0.00004986, -0.00001617, -0.00001153, 0.00000422, -0.00000351, -0.00001121, -0.00000527, 0.00001131, 0.00002248, -0.00002449, 0.00003365, -0.00006162, 0.00005708, -0.00009167, 0.00002233, 0.00000737, 0.00002146, -0.00005159, -0.00000838, 0.00006721, -0.00001153, -0.00004112, -0.00000523, 0.00000229, 0.00000597, -0.00002461, -0.00001786, 0.00004050, 0.00001339, -0.00003524, -0.00001959, 0.00000711, -0.00002294, 0.00000824, -0.00003923, -0.00012097, -0.00031160, 0.00093584, -0.00005062, -0.00023659, 0.00003356, -0.00008969, 0.00003182, -0.00002411, -0.00002453, 0.00001448, 0.00000921, -0.00002573, -0.00001515, 0.00002691, -0.00000877, 0.00000017, -0.00002184, 0.00002627, -0.00000785, -0.00000102, },
		{ 0.00002538, -0.00001909, -0.00005715, 0.00000440, 0.00008931, 0.00013181, -0.00003799, 0.00000468, -0.00007147, 0.00002680, -0.00003720, 0.00003601, -0.00004311, 0.00002684, -0.00003330, 0.00006138, -0.00005005, 0.00005367, -0.00006053, 0.00003372, -0.00003648, -0.00004003, 0.00003443, 0.00000208, -0.00000258, -0.00006535, 0.00007922, 0.00000211, -0.00002326, -0.00003550, 0.00002183, -0.00001681, 0.00001712, -0.00006199, 0.00005396, -0.00000146, 0.00001903, -0.00008105, 0.00005951, -0.00008037, 0.00007748, -0.00011680, 0.00006714, -0.00050627, 0.00060154, 0.00053722, -0.00046406, 0.00009346, -0.00013220, 0.00003810, -0.00002635, -0.00000968, -0.00002272, 0.00003841, -0.00002897, -0.00001646, 0.00000726, 0.00001561, -0.00001299, -0.00000914, 0.00000420, 0.00001082, -0.00000177, -0.00004187, },
		{ 0.00004386, -0.00009304, 0.00002437, 0.00000026, 0.00018147, -0.00003109, 0.00003223, -0.00008152, 0.00001873, -0.00003489, 0.00002959, -0.00003072, 0.00001638, -0.00002654, 0.00003841, -0.00001253, 0.00000952, 0.00000149, -0.00003539, 0.00003605, -0.00008750, 0.00004258, -0.00001292, 0.00003443, -0.00008405, 0.00004987, 0.00003198, -0.00001331, -0.00004545, 0.00001780, -0.00001839, 0.00002716, -0.00005603, 0.00002048, 0.00001436, 0.00003153, -0.00007351, 0.00002979, -0.00004701, 0.00003747, -0.00006151, 0.00003372, -0.00030559, -0.00005123, 0.00104267, -0.00041333, -0.00003481, -0.00006321, -0.00002614, 0.00000483, -0.00001083, -0.00003831, 0.00003622, -0.00001218, -0.00001677, -0.00001701, 0.00003545, -0.00002127, 0.00000405, -0.00001734, 0.00002652, -0.00000983, -0.00001431, -0.00002518, },
		{ -0.00007477, 0.00000323, -0.00002007, 0.00013378, 0.00004850, -0.00000731, -0.00001687, -0.00004912, 0.00001670, -0.00002259, 0.00002114, -0.00003022, 0.00001951, -0.00002087, 0.00004843, -0.00004600, 0.00005729, -0.00007444, 0.00005573, -0.00007242, 0.00000601, 0.00000096, 0.00002952, -0.00004597, -0.00001960, 0.00007426, -0.00001826, -0.00002508, -0.00001736, 0.00001164, -0.00000612, 0.00000051, -0.00004896, 0.00005624, -0.00000317, 0.00000076, -0.00006402, 0.00004869, -0.00007272, 0.00007310, -0.00011906, 0.00002123, -0.00055092, 0.00103301, 0.00009540, -0.00035994, 0.00009211, -0.00013829, 0.00006017, -0.00003842, -0.00001278, -0.00000964, 0.00002984, -0.00002999, -0.00001809, 0.00002000, 0.00000245, -0.00000911, -0.00001103, 0.00001495, -0.00000030, -0.00000071, -0.00004924, 0.00005883, },
		{ -0.00005942, 0.00002342, 0.00002290, 0.00015060, -0.00006187, 0.00005770, -0.00010396, 0.00004824, -0.00005423, 0.00004851, -0.00005120, 0.00003997, -0.00004536, 0.00006073, -0.00004410, 0.00004510, -0.00003830, 0.00000400, -0.00000455, -0.00005318, 0.00002826, -0.00000252, 0.00001701, -0.00008400, 0.00009140, -0.00001196, -0.00000297, -0.00004891, 0.00003593, -0.00003706, 0.00004596, -0.00008545, 0.00006225, -0.00002016, 0.00005436, -0.00011276, 0.00008335, -0.00011058, 0.00011478, -0.00015747, 0.00013165, -0.00055133, 0.00042328, 0.00084858, -0.00061008, 0.00014309, -0.00017376, 0.00006054, -0.00004357, 0.00001309, -0.00005456, 0.00005804, -0.00003517, -0.00000580, -0.00001710, 0.00003952, -0.00003192, 0.00000913, -0.00001127, 0.00002091, -0.00000744, -0.00003319, 0.00001786, -0.00000447, },
		{ 0.00005206, -0.00004857, 0.00017309, -0.00003925, 0.00005142, -0.00006722, -0.00000063, -0.00001592, 0.00000975, -0.00001102, 0.00000182, -0.00000682, 0.00000841, 0.00001596, -0.00001816, 0.00003417, -0.00006049, 0.00004915, -0.00007820, 0.00002813, -0.00001934, 0.00004960, -0.00008959, 0.00004810, 0.00003077, -0.00001045, -0.00003958, 0.00001475, -0.00001861, 0.00002619, -0.00004112, -0.00000440, 0.00002417, 0.00002633, -0.00005111, -0.00000355, -0.00001558, -0.00000002, -0.00000950, -0.00003195, -0.00016392, -0.00036292, 0.00129329, -0.00039972, -0.00009978, -0.00003217, -0.00005565, 0.00002146, -0.00001383, -0.00004087, 0.00002529, 0.00000072, -0.00001460, -0.00002988, 0.00004060, -0.00002039, 0.00000325, -0.00001461, 0.00002198, -0.00000721, -0.00001018, -0.00003015, 0.00005581, -0.00009876, },
		{ -0.00002189, 0.00008088, 0.00007261, -0.00003390, 0.00003778, -0.00009052, 0.00004938, -0.00005082, 0.00004482, -0.00004858, 0.00004255, -0.00004588, 0.00006316, -0.00005529, 0.00006259, -0.00006416, 0.00003400, -0.00004033, -0.00001273, -0.00000307, 0.00003037, -0.00003147, -0.00003806, 0.00008804, -0.00003373, -0.00000184, -0.00003611, 0.00003167, -0.00003346, 0.00004041, -0.00008773, 0.00007883, -0.00003327, 0.00005137, -0.00011803, 0.00009869, -0.00013232, 0.00014795, -0.00020965, 0.00016569, -0.00073348, 0.00099929, 0.00035676, -0.00053881, 0.00017880, -0.00020585, 0.00010370, -0.00006869, 0.00001735, -0.00004959, 0.00005977, -0.00004273, -0.00000610, -0.00000514, 0.00002976, -0.00003232, 0.00001077, -0.00000420, 0.00001320, -0.00000833, -0.00004054, 0.00005460, -0.00006535, 0.00000972, },
		{ -0.00004069, 0.00017075, -0.00008992, 0.00009605, -0.00011294, 0.00004875, -0.00005142, 0.00004241, -0.00004293, 0.00003591, -0.00003718, 0.00004058, -0.00002122, 0.00001855, -0.00000247, -0.00002908, 0.00002156, -0.00005546, 0.00001772, -0.00000795, 0.00003636, -0.00009953, 0.00010005, -0.00002345, 0.00001233, -0.00005741, 0.00004762, -0.00005400, 0.00006517, -0.00009186, 0.00005545, -0.00002767, 0.00007291, -0.00011959, 0.00008014, -0.00011133, 0.00011613, -0.00015246, 0.00012796, -0.00047990, 0.00011818, 0.00120548, -0.00073298, 0.00016443, -0.00019552, 0.00007006, -0.00005426, 0.00003254, -0.00008037, 0.00006914, -0.00003787, 0.00000841, -0.00004407, 0.00006113, -0.00004629, 0.00002130, -0.00002044, 0.00002649, -0.00001262, -0.00002306, 0.00000905, 0.00001097, -0.00006745, 0.00005031, },
		{ 0.00014715, -0.00002688, 0.00003524, -0.00002258, -0.00003932, 0.00001791, -0.00002127, 0.00001671, -0.00002074, 0.00002032, -0.00002402, 0.00004216, -0.00004137, 0.00005436, -0.00006508, 0.00004041, -0.00005165, 0.00000739, -0.00002166, 0.00005665, -0.00008442, 0.00003615, 0.00003924, -0.00001965, -0.00002225, 0.00000052, -0.00000345, 0.00000361, -0.00000253, -0.00004689, 0.00005017, -0.00000272, -0.00000049, -0.00006295, 0.00004273, -0.00006892, 0.00007936, -0.00014471, 0.00003263, -0.00067849, 0.00145045, -0.00027068, -0.00023830, 0.00004417, -0.00011696, 0.00006225, -0.00003899, -0.00001937, -0.00000766, 0.00002805, -0.00002345, -0.00002634, 0.00002498, 0.00000023, -0.00001544, 0.00000311, 0.00000591, 0.00000381, -0.00001483, -0.00002482, 0.00005707, -0.00009659, 0.00006833, -0.00007507, },
		{ 0.00010889, -0.00007648, 0.00009184, -0.00012034, 0.00006961, -0.00006548, 0.00005451, -0.00005424, 0.00005152, -0.00005244, 0.00005780, -0.00004467, 0.00004429, -0.00003189, -0.00000110, -0.00000494, -0.00002628, -0.00000823, 0.00002678, -0.00001209, -0.00005929, 0.00010391, -0.00005268, 0.00002327, -0.00005702, 0.00005693, -0.00006688, 0.00008130, -0.00011931, 0.00009518, -0.00006251, 0.00009799, -0.00016059, 0.00013499, -0.00017690, 0.00020327, -0.00027347, 0.00026110, -0.00080195, 0.00080323, 0.00072512, -0.00074901, 0.00027308, -0.00027770, 0.00015207, -0.00010709, 0.00005818, -0.00009682, 0.00009206, -0.00006060, 0.00001842, -0.00004339, 0.00006545, -0.00006067, 0.00003511, -0.00002451, 0.00002813, -0.00002049, -0.00002515, 0.00004337, -0.00004960, 0.00000394, -0.00000489, 0.00001785, },
		{ -0.00010046, 0.00010103, -0.00008884, 0.00002399, -0.00002694, 0.00001801, -0.00001955, 0.00001664, -0.00001331, 0.00000833, 0.00000963, -0.00001395, 0.00003056, -0.00004886, 0.00002897, -0.00003965, -0.00000064, -0.00000958, 0.00004916, -0.00010559, 0.00009886, -0.00002564, 0.00001595, -0.00005424, 0.00004783, -0.00005543, 0.00006137, -0.00007059, 0.00002568, -0.00001444, 0.00006292, -0.00009081, 0.00004076, -0.00007320, 0.00007136, -0.00009093, 0.00004696, -0.00030028, -0.00026773, 0.00153829, -0.00077555, 0.00012175, -0.00017363, 0.00004829, -0.00004218, 0.00003263, -0.00008473, 0.00005932, -0.00002714, 0.00001469, -0.00005902, 0.00006542, -0.00004282, 0.00001662, -0.00001510, 0.00002069, -0.00001022, -0.00001958, 0.00000741, 0.00001925, -0.00007288, 0.00007412, -0.00008895, 0.00017650, },
		{ -0.00000908, 0.00003665, -0.00008087, 0.00005126, -0.00004879, 0.00003897, -0.00003816, 0.00004061, -0.00004441, 0.00005272, -0.00004674, 0.00004991, -0.00004286, 0.00001026, -0.00001280, -0.00001468, -0.00002131, 0.00005571, -0.00006949, 0.00001544, 0.00005591, -0.00003898, 0.00000465, -0.00002421, 0.00002629, -0.00003671, 0.00005092, -0.00009619, 0.00008320, -0.00004580, 0.00006450, -0.00013146, 0.00010942, -0.00014785, 0.00017864, -0.00026838, 0.00022741, -0.00092134, 0.00145009, -0.00000061, -0.00045147, 0.00016212, -0.00020748, 0.00012666, -0.00008738, 0.00002692, -0.00006149, 0.00006911, -0.00004599, -0.00000205, -0.00001344, 0.00003962, -0.00004906, 0.00003223, -0.00001909, 0.00002237, -0.00002875, -0.00000860, 0.00004720, -0.00008499, 0.00006961, -0.00007589, 0.00010027, 0.00000827, },
		{ 0.00012165, -0.00012175, 0.00006538, -0.00005741, 0.00004363, -0.00004184, 0.00004172, -0.00003823, 0.00003313, -0.00001698, 0.00000922, 0.00000930, -0.00003365, 0.00001928, -0.00002486, -0.00001987, 0.00002436, 0.00000494, -0.00007565, 0.00011413, -0.00006682, 0.00004174, -0.00007138, 0.00007634, -0.00009137, 0.00010540, -0.00012732, 0.00009134, -0.00007501, 0.00012127, -0.00017155, 0.00013753, -0.00018455, 0.00021381, -0.00028018, 0.00027125, -0.00072937, 0.00045636, 0.00115199, -0.00093915, 0.00033763, -0.00032709, 0.00018465, -0.00013604, 0.00009268, -0.00013461, 0.00011292, -0.00007311, 0.00004435, -0.00008028, 0.00009373, -0.00007930, 0.00004928, -0.00003586, 0.00003621, -0.00002944, -0.00001139, 0.00003361, -0.00003520, -0.00000660, 0.00002074, -0.00004281, 0.00014112, -0.00011148, },
		{ -0.00003799, -0.00001681, 0.00000912, -0.00001387, 0.00000829, -0.00000690, 0.00001335, -0.00002153, 0.00003357, -0.00003578, 0.00004332, -0.00004219, 0.00001100, -0.00000485, -0.00002582, -0.00000572, 0.00005200, -0.00009915, 0.00008511, -0.00001561, 0.00000507, -0.00003624, 0.00003160, -0.00003535, 0.00003043, -0.00002217, -0.00002408, 0.00001838, 0.00002505, -0.00003016, -0.00002833, -0.00000372, -0.00001034, 0.00001456, -0.00009035, -0.00005712, -0.00065665, 0.00176397, -0.00068432, -0.00001117, -0.00009283, -0.00001663, 0.00000389, 0.00000179, -0.00005703, 0.00002082, 0.00000376, 0.00000351, -0.00005016, 0.00004229, -0.00001460, -0.00000913, 0.00000740, 0.00000087, 0.00000348, -0.00002767, 0.00001810, 0.00001546, -0.00006961, 0.00008580, -0.00011182, 0.00015699, -0.00008242, 0.00006971, },
		{ -0.00010519, 0.00006750, -0.00005852, 0.00004382, -0.00003962, 0.00004340, -0.00004366, 0.00004113, -0.00002952, 0.00002007, -0.00000048, -0.00002958, 0.00002543, -0.00002816, -0.00002365, 0.00005151, -0.00005090, -0.00000739, 0.00007399, -0.00006113, 0.00003356, -0.00005186, 0.00006088, -0.00008075, 0.00010112, -0.00013427, 0.00010737, -0.00008524, 0.00012186, -0.00018482, 0.00015903, -0.00020831, 0.00025441, -0.00035966, 0.00036229, -0.00102934, 0.00126130, 0.00039743, -0.00070929, 0.00029587, -0.00030784, 0.00019996, -0.00014696, 0.00008616, -0.00012383, 0.00011365, -0.00007563, 0.00003657, -0.00006484, 0.00008561, -0.00008527, 0.00006180, -0.00004423, 0.00004164, -0.00004611, 0.00001255, 0.00003219, -0.00006870, 0.00005950, -0.00005678, 0.00004092, 0.00005368, -0.00006048, 0.00008629, },
		{ 0.00003682, -0.00002899, 0.00001646, -0.00001643, 0.00001899, -0.00001165, 0.00000016, 0.00001495, -0.00002623, 0.00003967, -0.00004586, 0.00001965, -0.00000185, -0.00004112, 0.00002788, 0.00001296, -0.00008042, 0.00011201, -0.00006899, 0.00004610, -0.00007100, 0.00007926, -0.00009393, 0.00009990, -0.00010274, 0.00006081, -0.00006215, 0.00011060, -0.00014169, 0.00009866, -0.00014727, 0.00017081, -0.00022053, 0.00019010, -0.00053100, 0.00001582, 0.00155634, -0.00104316, 0.00033014, -0.00032499, 0.00017859, -0.00013534, 0.00010130, -0.00014464, 0.00010753, -0.00006775, 0.00005717, -0.00009854, 0.00009813, -0.00007464, 0.00004277, -0.00003000, 0.00003012, -0.00002759, -0.00000765, 0.00003386, -0.00003014, -0.00001305, 0.00004254, -0.00008614, 0.00015226, -0.00011989, 0.00011419, -0.00009067, },
		{ 0.00003972, -0.00003737, 0.00002635, -0.00002102, 0.00002840, -0.00003493, 0.00003786, -0.00003470, 0.00002659, -0.00000731, -0.00002757, 0.00004061, -0.00005039, 0.00000034, 0.00004684, -0.00008307, 0.00006210, 0.00000334, -0.00001675, -0.00000691, 0.00000195, 0.00000181, -0.00001922, 0.00004012, -0.00007956, 0.00005856, -0.00002738, 0.00004476, -0.00010654, 0.00007487, -0.00010354, 0.00013270, -0.00024113, 0.00018392, -0.00096301, 0.00181334, -0.00043271, -0.00023451, 0.00004183, -0.00012064, 0.00008174, -0.00005920, 0.00000132, -0.00004308, 0.00005241, -0.00002614, -0.00001566, -0.00000681, 0.00003367, -0.00004983, 0.00004123, -0.00002853, 0.00002568, -0.00004566, 0.00003953, 0.00000115, -0.00005765, 0.00008299, -0.00010788, 0.00011317, -0.00004072, 0.00001620, 0.00001982, -0.00005498, },
		{ -0.00004470, 0.00002811, -0.00002319, 0.00002769, -0.00002341, 0.00001205, 0.00000307, -0.00002291, 0.00004401, -0.00006016, 0.00004705, -0.00002244, -0.00003482, 0.00005072, -0.00003709, -0.00002346, 0.00008451, -0.00007651, 0.00005435, -0.00007128, 0.00008628, -0.00011087, 0.00012883, -0.00014439, 0.00010832, -0.00010431, 0.00015149, -0.00020240, 0.00017097, -0.00022754, 0.00028020, -0.00038647, 0.00039972, -0.00097736, 0.00090207, 0.00086652, -0.00095285, 0.00040183, -0.00038589, 0.00025656, -0.00019554, 0.00013704, -0.00017417, 0.00014463, -0.00009849, 0.00007438, -0.00011082, 0.00011945, -0.00010745, 0.00007808, -0.00005832, 0.00005199, -0.00005770, 0.00002906, 0.00002151, -0.00005617, 0.00004652, -0.00003003, -0.00001488, 0.00008969, -0.00009511, 0.00010576, -0.00009918, 0.00005800, },
		{ -0.00001344, 0.00000903, -0.00000350, 0.00001247, -0.00002516, 0.00003488, -0.00004433, 0.00004247, -0.00002731, -0.00000735, 0.00004098, -0.00006906, 0.00003765, 0.00000985, -0.00007180, 0.00009604, -0.00005737, 0.00003431, -0.00005321, 0.00006173, -0.00007073, 0.00006403, -0.00004970, 0.00000939, -0.00002749, 0.00006937, -0.00007772, 0.00002772, -0.00007541, 0.00008639, -0.00011058, 0.00004384, -0.00025981, -0.00043059, 0.00184718, -0.00100284, 0.00022156, -0.00025406, 0.00012044, -0.00009265, 0.00007176, -0.00011612, 0.00006823, -0.00003757, 0.00004702, -0.00008690, 0.00007028, -0.00004214, 0.00001356, -0.00000577, 0.00000832, -0.00001241, -0.00001741, 0.00004767, -0.00003776, -0.00001049, 0.00005202, -0.00010290, 0.00013425, -0.00009666, 0.00007349, -0.00003856, -0.00000317, 0.00000652, },
		{ 0.00002492, -0.00001722, 0.00002389, -0.00002595, 0.00001925, -0.00000987, -0.00001494, 0.00004344, -0.00007080, 0.00008107, -0.00006259, 0.00000162, 0.00004012, -0.00006521, 0.00003827, 0.00002270, -0.00004054, 0.00002430, -0.00003073, 0.00004274, -0.00006989, 0.00009540, -0.00012105, 0.00008862, -0.00007469, 0.00010998, -0.00016786, 0.00013531, -0.00017758, 0.00022681, -0.00035820, 0.00036002, -0.00112191, 0.00165692, -0.00003781, -0.00051400, 0.00020062, -0.00024063, 0.00017298, -0.00013464, 0.00007488, -0.00011664, 0.00010567, -0.00006464, 0.00003419, -0.00006783, 0.00008625, -0.00009062, 0.00007375, -0.00005716, 0.00004772, -0.00006577, 0.00006386, -0.00001603, -0.00004308, 0.00007068, -0.00008524, 0.00005755, 0.00000719, -0.00003567, 0.00006087, -0.00007201, 0.00004993, -0.00004019, },
		{ 0.00000271, 0.00000273, 0.00000465, -0.00002046, 0.00003503, -0.00005935, 0.00006919, -0.00006425, 0.00004085, 0.00000768, -0.00005703, 0.00005759, -0.00003434, -0.00002639, 0.00008111, -0.00007820, 0.00005939, -0.00007360, 0.00009134, -0.00011453, 0.00012300, -0.00011982, 0.00008126, -0.00009565, 0.00014425, -0.00017670, 0.00013916, -0.00019895, 0.00024844, -0.00034079, 0.00033604, -0.00078619, 0.00043784, 0.00131791, -0.00111036, 0.00043329, -0.00040879, 0.00027024, -0.00020998, 0.00015750, -0.00019227, 0.00014572, -0.00010035, 0.00009485, -0.00013280, 0.00012464, -0.00010291, 0.00007165, -0.00005370, 0.00004631, -0.00005621, 0.00003304, 0.00002297, -0.00005452, 0.00003896, -0.00000823, -0.00005181, 0.00010028, -0.00009826, 0.00008819, -0.00006405, 0.00002710, -0.00001368, -0.00000020, },
		{ -0.00001278, 0.00002086, -0.00003041, 0.00003186, -0.00003539, 0.00001290, 0.00001834, -0.00005246, 0.00009213, -0.00009396, 0.00004841, -0.00000055, -0.00005454, 0.00007140, -0.00003699, 0.00001149, -0.00002317, 0.00002923, -0.00002963, 0.00001009, 0.00001526, -0.00004642, 0.00001477, 0.00001324, 0.00000038, -0.00005257, 0.00000671, -0.00001172, 0.00001554, -0.00012043, 0.00001262, -0.00079187, 0.00195312, -0.00079187, 0.00001262, -0.00012043, 0.00001554, -0.00001172, 0.00000671, -0.00005257, 0.00000038, 0.00001324, 0.00001477, -0.00004642, 0.00001526, 0.00001009, -0.00002963, 0.00002923, -0.00002317, 0.00001149, -0.00003699, 0.00007140, -0.00005454, -0.00000055, 0.00004841, -0.00009396, 0.00009213, -0.00005246, 0.00001834, 0.00001290, -0.00003539, 0.00003186, -0.00003041, 0.00002086, },
		{ 0.00000271, -0.00000020, -0.00001368, 0.00002710, -0.00006405, 0.00008819, -0.00009826, 0.00010028, -0.00005181, -0.00000823, 0.00003896, -0.00005452, 0.00002297, 0.00003304, -0.00005621, 0.00004631, -0.00005370, 0.00007165, -0.00010291, 0.00012464, -0.00013280, 0.00009485, -0.00010035, 0.00014572, -0.00019227, 0.00015750, -0.00020998, 0.00027024, -0.00040879, 0.00043329, -0.00111036, 0.00131791, 0.00043784, -0.00078619, 0.00033604, -0.00034079, 0.00024844, -0.00019895, 0.00013916, -0.00017670, 0.00014425, -0.00009565, 0.00008126, -0.00011982, 0.00012300, -0.00011453, 0.00009134, -0.00007360, 0.00005939, -0.00007820, 0.00008111, -0.00002639, -0.00003434, 0.00005759, -0.00005703, 0.00000768, 0.00004085, -0.00006425, 0.00006919, -0.00005935, 0.00003503, -0.00002046, 0.00000465, 0.00000273, },
		{ 0.00002492, -0.00004019, 0.00004993, -0.00007201, 0.00006087, -0.00003567, 0.00000719, 0.00005755, -0.00008524, 0.00007068, -0.00004308, -0.00001603, 0.00006386, -0.00006577, 0.00004772, -0.00005716, 0.00007375, -0.00009062, 0.00008625, -0.00006783, 0.00003419, -0.00006464, 0.00010567, -0.00011664, 0.00007488, -0.00013464, 0.00017298, -0.00024063, 0.00020062, -0.00051400, -0.00003781, 0.00165692, -0.00112191, 0.00036002, -0.00035820, 0.00022681, -0.00017758, 0.00013531, -0.00016786, 0.00010998, -0.00007469, 0.00008862, -0.00012105, 0.00009540, -0.00006989, 0.00004274, -0.00003073, 0.00002430, -0.00004054, 0.00002270, 0.00003827, -0.00006521, 0.00004012, 0.00000162, -0.00006259, 0.00008107, -0.00007080, 0.00004344, -0.00001494, -0.00000987, 0.00001925, -0.00002595, 0.00002389, -0.00001722, },
		{ -0.00001344, 0.00000652, -0.00000317, -0.00003856, 0.00007349, -0.00009666, 0.00013425, -0.00010290, 0.00005202, -0.00001049, -0.00003776, 0.00004767, -0.00001741, -0.00001241, 0.00000832, -0.00000577, 0.00001356, -0.00004214, 0.00007028, -0.00008690, 0.00004702, -0.00003757, 0.00006823, -0.00011612, 0.00007176, -0.00009265, 0.00012044, -0.00025406, 0.00022156, -0.00100284, 0.00184718, -0.00043059, -0.00025981, 0.00004384, -0.00011058, 0.00008639, -0.00007541, 0.00002772, -0.00007772, 0.00006937, -0.00002749, 0.00000939, -0.00004970, 0.00006403, -0.00007073, 0.00006173, -0.00005321, 0.00003431, -0.00005737, 0.00009604, -0.00007180, 0.00000985, 0.00003765, -0.00006906, 0.00004098, -0.00000735, -0.00002731, 0.00004247, -0.00004433, 0.00003488, -0.00002516, 0.00001247, -0.00000350, 0.00000903, },
		{ -0.00004470, 0.00005800, -0.00009918, 0.00010576, -0.00009511, 0.00008969, -0.00001488, -0.00003003, 0.00004652, -0.00005617, 0.00002151, 0.00002906, -0.00005770, 0.00005199, -0.00005832, 0.00007808, -0.00010745, 0.00011945, -0.00011082, 0.00007438, -0.00009849, 0.00014463, -0.00017417, 0.00013704, -0.00019554, 0.00025656, -0.00038589, 0.00040183, -0.00095285, 0.00086652, 0.00090207, -0.00097736, 0.00039972, -0.00038647, 0.00028020, -0.00022754, 0.00017097, -0.00020240, 0.00015149, -0.00010431, 0.00010832, -0.00014439, 0.00012883, -0.00011087, 0.00008628, -0.00007128, 0.00005435, -0.00007651, 0.00008451, -0.00002346, -0.00003709, 0.00005072, -0.00003482, -0.00002244, 0.00004705, -0.00006016, 0.00004401, -0.00002291, 0.00000307, 0.00001205, -0.00002341, 0.00002769, -0.00002319, 0.00002811, },
		{ 0.00003972, -0.00005498, 0.00001982, 0.00001620, -0.00004072, 0.00011317, -0.00010788, 0.00008299, -0.00005765, 0.00000115, 0.00003953, -0.00004566, 0.00002568, -0.00002853, 0.00004123, -0.00004983, 0.00003367, -0.00000681, -0.00001566, -0.00002614, 0.00005241, -0.00004308, 0.00000132, -0.00005920, 0.00008174, -0.00012064, 0.00004183, -0.00023451, -0.00043271, 0.00181334, -0.00096301, 0.00018392, -0.00024113, 0.00013270, -0.00010354, 0.00007487, -0.00010654, 0.00004476, -0.00002738, 0.00005856, -0.00007956, 0.00004012, -0.00001922, 0.00000181, 0.00000195, -0.00000691, -0.00001675, 0.00000334, 0.00006210, -0.00008307, 0.00004684, 0.00000034, -0.00005039, 0.00004061, -0.00002757, -0.00000731, 0.00002659, -0.00003470, 0.00003786, -0.00003493, 0.00002840, -0.00002102, 0.00002635, -0.00003737, },
		{ 0.00003682, -0.00009067, 0.00011419, -0.00011989, 0.00015226, -0.00008614, 0.00004254, -0.00001305, -0.00003014, 0.00003386, -0.00000765, -0.00002759, 0.00003012, -0.00003000, 0.00004277, -0.00007464, 0.00009813, -0.00009854, 0.00005717, -0.00006775, 0.00010753, -0.00014464, 0.00010130, -0.00013534, 0.00017859, -0.00032499, 0.00033014, -0.00104316, 0.00155634, 0.00001582, -0.00053100, 0.00019010, -0.00022053, 0.00017081, -0.00014727, 0.00009866, -0.00014169, 0.00011060, -0.00006215, 0.00006081, -0.00010274, 0.00009990, -0.00009393, 0.00007926, -0.00007100, 0.00004610, -0.00006899, 0.00011201, -0.00008042, 0.00001296, 0.00002788, -0.00004112, -0.00000185, 0.00001965, -0.00004586, 0.00003967, -0.00002623, 0.00001495, 0.00000016, -0.00001165, 0.00001899, -0.00001643, 0.00001646, -0.00002899, },
		{ -0.00010519, 0.00008629, -0.00006048, 0.00005368, 0.00004092, -0.00005678, 0.00005950, -0.00006870, 0.00003219, 0.00001255, -0.00004611, 0.00004164, -0.00004423, 0.00006180, -0.00008527, 0.00008561, -0.00006484, 0.00003657, -0.00007563, 0.00011365, -0.00012383, 0.00008616, -0.00014696, 0.00019996, -0.00030784, 0.00029587, -0.00070929, 0.00039743, 0.00126130, -0.00102934, 0.00036229, -0.00035966, 0.00025441, -0.00020831, 0.00015903, -0.00018482, 0.00012186, -0.00008524, 0.00010737, -0.00013427, 0.00010112, -0.00008075, 0.00006088, -0.00005186, 0.00003356, -0.00006113, 0.00007399, -0.00000739, -0.00005090, 0.00005151, -0.00002365, -0.00002816, 0.00002543, -0.00002958, -0.00000048, 0.00002007, -0.00002952, 0.00004113, -0.00004366, 0.00004340, -0.00003962, 0.00004382, -0.00005852, 0.00006750, },
		{ -0.00003799, 0.00006971, -0.00008242, 0.00015699, -0.00011182, 0.00008580, -0.00006961, 0.00001546, 0.00001810, -0.00002767, 0.00000348, 0.00000087, 0.00000740, -0.00000913, -0.00001460, 0.00004229, -0.00005016, 0.00000351, 0.00000376, 0.00002082, -0.00005703, 0.00000179, 0.00000389, -0.00001663, -0.00009283, -0.00001117, -0.00068432, 0.00176397, -0.00065665, -0.00005712, -0.00009035, 0.00001456, -0.00001034, -0.00000372, -0.00002833, -0.00003016, 0.00002505, 0.00001838, -0.00002408, -0.00002217, 0.00003043, -0.00003535, 0.00003160, -0.00003624, 0.00000507, -0.00001561, 0.00008511, -0.00009915, 0.00005200, -0.00000572, -0.00002582, -0.00000485, 0.00001100, -0.00004219, 0.00004332, -0.00003578, 0.00003357, -0.00002153, 0.00001335, -0.00000690, 0.00000829, -0.00001387, 0.00000912, -0.00001681, },
		{ 0.00012165, -0.00011148, 0.00014112, -0.00004281, 0.00002074, -0.00000660, -0.00003520, 0.00003361, -0.00001139, -0.00002944, 0.00003621, -0.00003586, 0.00004928, -0.00007930, 0.00009373, -0.00008028, 0.00004435, -0.00007311, 0.00011292, -0.00013461, 0.00009268, -0.00013604, 0.00018465, -0.00032709, 0.00033763, -0.00093915, 0.00115199, 0.00045636, -0.00072937, 0.00027125, -0.00028018, 0.00021381, -0.00018455, 0.00013753, -0.00017155, 0.00012127, -0.00007501, 0.00009134, -0.00012732, 0.00010540, -0.00009137, 0.00007634, -0.00007138, 0.00004174, -0.00006682, 0.00011413, -0.00007565, 0.00000494, 0.00002436, -0.00001987, -0.00002486, 0.00001928, -0.00003365, 0.00000930, 0.00000922, -0.00001698, 0.00003313, -0.00003823, 0.00004172, -0.00004184, 0.00004363, -0.00005741, 0.00006538, -0.00012175, },
		{ -0.00000908, 0.00000827, 0.00010027, -0.00007589, 0.00006961, -0.00008499, 0.00004720, -0.00000860, -0.00002875, 0.00002237, -0.00001909, 0.00003223, -0.00004906, 0.00003962, -0.00001344, -0.00000205, -0.00004599, 0.00006911, -0.00006149, 0.00002692, -0.00008738, 0.00012666, -0.00020748, 0.00016212, -0.00045147, -0.00000061, 0.00145009, -0.00092134, 0.00022741, -0.00026838, 0.00017864, -0.00014785, 0.00010942, -0.00013146, 0.00006450, -0.00004580, 0.00008320, -0.00009619, 0.00005092, -0.00003671, 0.00002629, -0.00002421, 0.00000465, -0.00003898, 0.00005591, 0.00001544, -0.00006949, 0.00005571, -0.00002131, -0.00001468, -0.00001280, 0.00001026, -0.00004286, 0.00004991, -0.00004674, 0.00005272, -0.00004441, 0.00004061, -0.00003816, 0.00003897, -0.00004879, 0.00005126, -0.00008087, 0.00003665, },
		{ -0.00010046, 0.00017650, -0.00008895, 0.00007412, -0.00007288, 0.00001925, 0.00000741, -0.00001958, -0.00001022, 0.00002069, -0.00001510, 0.00001662, -0.00004282, 0.00006542, -0.00005902, 0.00001469, -0.00002714, 0.00005932, -0.00008473, 0.00003263, -0.00004218, 0.00004829, -0.00017363, 0.00012175, -0.00077555, 0.00153829, -0.00026773, -0.00030028, 0.00004696, -0.00009093, 0.00007136, -0.00007320, 0.00004076, -0.00009081, 0.00006292, -0.00001444, 0.00002568, -0.00007059, 0.00006137, -0.00005543, 0.00004783, -0.00005424, 0.00001595, -0.00002564, 0.00009886, -0.00010559, 0.00004916, -0.00000958, -0.00000064, -0.00003965, 0.00002897, -0.00004886, 0.00003056, -0.00001395, 0.00000963, 0.00000833, -0.00001331, 0.00001664, -0.00001955, 0.00001801, -0.00002694, 0.00002399, -0.00008884, 0.00010103, },
		{ 0.00010889, 0.00001785, -0.00000489, 0.00000394, -0.00004960, 0.00004337, -0.00002515, -0.00002049, 0.00002813, -0.00002451, 0.00003511, -0.00006067, 0.00006545, -0.00004339, 0.00001842, -0.00006060, 0.00009206, -0.00009682, 0.00005818, -0.00010709, 0.00015207, -0.00027770, 0.00027308, -0.00074901, 0.00072512, 0.00080323, -0.00080195, 0.00026110, -0.00027347, 0.00020327, -0.00017690, 0.00013499, -0.00016059, 0.00009799, -0.00006251, 0.00009518, -0.00011931, 0.00008130, -0.00006688, 0.00005693, -0.00005702, 0.00002327, -0.00005268, 0.00010391, -0.00005929, -0.00001209, 0.00002678, -0.00000823, -0.00002628, -0.00000494, -0.00000110, -0.00003189, 0.00004429, -0.00004467, 0.00005780, -0.00005244, 0.00005152, -0.00005424, 0.00005451, -0.00006548, 0.00006961, -0.00012034, 0.00009184, -0.00007648, },
		{ 0.00014715, -0.00007507, 0.00006833, -0.00009659, 0.00005707, -0.00002482, -0.00001483, 0.00000381, 0.00000591, 0.00000311, -0.00001544, 0.00000023, 0.00002498, -0.00002634, -0.00002345, 0.00002805, -0.00000766, -0.00001937, -0.00003899, 0.00006225, -0.00011696, 0.00004417, -0.00023830, -0.00027068, 0.00145045, -0.00067849, 0.00003263, -0.00014471, 0.00007936, -0.00006892, 0.00004273, -0.00006295, -0.00000049, -0.00000272, 0.00005017, -0.00004689, -0.00000253, 0.00000361, -0.00000345, 0.00000052, -0.00002225, -0.00001965, 0.00003924, 0.00003615, -0.00008442, 0.00005665, -0.00002166, 0.00000739, -0.00005165, 0.00004041, -0.00006508, 0.00005436, -0.00004137, 0.00004216, -0.00002402, 0.00002032, -0.00002074, 0.00001671, -0.00002127, 0.00001791, -0.00003932, -0.00002258, 0.00003524, -0.00002688, },
		{ -0.00004069, 0.00005031, -0.00006745, 0.00001097, 0.00000905, -0.00002306, -0.00001262, 0.00002649, -0.00002044, 0.00002130, -0.00004629, 0.00006113, -0.00004407, 0.00000841, -0.00003787, 0.00006914, -0.00008037, 0.00003254, -0.00005426, 0.00007006, -0.00019552, 0.00016443, -0.00073298, 0.00120548, 0.00011818, -0.00047990, 0.00012796, -0.00015246, 0.00011613, -0.00011133, 0.00008014, -0.00011959, 0.00007291, -0.00002767, 0.00005545, -0.00009186, 0.00006517, -0.00005400, 0.00004762, -0.00005741, 0.00001233, -0.00002345, 0.00010005, -0.00009953, 0.00003636, -0.00000795, 0.00001772, -0.00005546, 0.00002156, -0.00002908, -0.00000247, 0.00001855, -0.00002122, 0.00004058, -0.00003718, 0.00003591, -0.00004293, 0.00004241, -0.00005142, 0.00004875, -0.00011294, 0.00009605, -0.00008992, 0.00017075, },
		{ -0.00002189, 0.00000972, -0.00006535, 0.00005460, -0.00004054, -0.00000833, 0.00001320, -0.00000420, 0.00001077, -0.00003232, 0.00002976, -0.00000514, -0.00000610, -0.00004273, 0.00005977, -0.00004959, 0.00001735, -0.00006869, 0.00010370, -0.00020585, 0.00017880, -0.00053881, 0.00035676, 0.00099929, -0.00073348, 0.00016569, -0.00020965, 0.00014795, -0.00013232, 0.00009869, -0.00011803, 0.00005137, -0.00003327, 0.00007883, -0.00008773, 0.00004041, -0.00003346, 0.00003167, -0.00003611, -0.00000184, -0.00003373, 0.00008804, -0.00003806, -0.00003147, 0.00003037, -0.00000307, -0.00001273, -0.00004033, 0.00003400, -0.00006416, 0.00006259, -0.00005529, 0.00006316, -0.00004588, 0.00004255, -0.00004858, 0.00004482, -0.00005082, 0.00004938, -0.00009052, 0.00003778, -0.00003390, 0.00007261, 0.00008088, },
		{ 0.00005206, -0.00009876, 0.00005581, -0.00003015, -0.00001018, -0.00000721, 0.00002198, -0.00001461, 0.00000325, -0.00002039, 0.00004060, -0.00002988, -0.00001460, 0.00000072, 0.00002529, -0.00004087, -0.00001383, 0.00002146, -0.00005565, -0.00003217, -0.00009978, -0.00039972, 0.00129329, -0.00036292, -0.00016392, -0.00003195, -0.00000950, -0.00000002, -0.00001558, -0.00000355, -0.00005111, 0.00002633, 0.00002417, -0.00000440, -0.00004112, 0.00002619, -0.00001861, 0.00001475, -0.00003958, -0.00001045, 0.00003077, 0.00004810, -0.00008959, 0.00004960, -0.00001934, 0.00002813, -0.00007820, 0.00004915, -0.00006049, 0.00003417, -0.00001816, 0.00001596, 0.00000841, -0.00000682, 0.00000182, -0.00001102, 0.00000975, -0.00001592, -0.00000063, -0.00006722, 0.00005142, -0.00003925, 0.00017309, -0.00004857, },
		{ -0.00005942, -0.00000447, 0.00001786, -0.00003319, -0.00000744, 0.00002091, -0.00001127, 0.00000913, -0.00003192, 0.00003952, -0.00001710, -0.00000580, -0.00003517, 0.00005804, -0.00005456, 0.00001309, -0.00004357, 0.00006054, -0.00017376, 0.00014309, -0.00061008, 0.00084858, 0.00042328, -0.00055133, 0.00013165, -0.00015747, 0.00011478, -0.00011058, 0.00008335, -0.00011276, 0.00005436, -0.00002016, 0.00006225, -0.00008545, 0.00004596, -0.00003706, 0.00003593, -0.00004891, -0.00000297, -0.00001196, 0.00009140, -0.00008400, 0.00001701, -0.00000252, 0.00002826, -0.00005318, -0.00000455, 0.00000400, -0.00003830, 0.00004510, -0.00004410, 0.00006073, -0.00004536, 0.00003997, -0.00005120, 0.00004851, -0.00005423, 0.00004824, -0.00010396, 0.00005770, -0.00006187, 0.00015060, 0.00002290, 0.00002342, },
		{ -0.00007477, 0.00005883, -0.00004924, -0.00000071, -0.00000030, 0.00001495, -0.00001103, -0.00000911, 0.00000245, 0.00002000, -0.00001809, -0.00002999, 0.00002984, -0.00000964, -0.00001278, -0.00003842, 0.00006017, -0.00013829, 0.00009211, -0.00035994, 0.00009540, 0.00103301, -0.00055092, 0.00002123, -0.00011906, 0.00007310, -0.00007272, 0.00004869, -0.00006402, 0.00000076, -0.00000317, 0.00005624, -0.00004896, 0.00000051, -0.00000612, 0.00001164, -0.00001736, -0.00002508, -0.00001826, 0.00007426, -0.00001960, -0.00004597, 0.00002952, 0.00000096, 0.00000601, -0.00007242, 0.00005573, -0.00007444, 0.00005729, -0.00004600, 0.00004843, -0.00002087, 0.00001951, -0.00003022, 0.00002114, -0.00002259, 0.00001670, -0.00004912, -0.00001687, -0.00000731, 0.00004850, 0.00013378, -0.00002007, 0.00000323, },
		{ 0.00004386, -0.00002518, -0.00001431, -0.00000983, 0.00002652, -0.00001734, 0.00000405, -0.00002127, 0.00003545, -0.00001701, -0.00001677, -0.00001218, 0.00003622, -0.00003831, -0.00001083, 0.00000483, -0.00002614, -0.00006321, -0.00003481, -0.00041333, 0.00104267, -0.00005123, -0.00030559, 0.00003372, -0.00006151, 0.00003747, -0.00004701, 0.00002979, -0.00007351, 0.00003153, 0.00001436, 0.00002048, -0.00005603, 0.00002716, -0.00001839, 0.00001780, -0.00004545, -0.00001331, 0.00003198, 0.00004987, -0.00008405, 0.00003443, -0.00001292, 0.00004258, -0.00008750, 0.00003605, -0.00003539, 0.00000149, 0.00000952, -0.00001253, 0.00003841, -0.00002654, 0.00001638, -0.00003072, 0.00002959, -0.00003489, 0.00001873, -0.00008152, 0.00003223, -0.00003109, 0.00018147, 0.00000026, 0.00002437, -0.00009304, },
		{ 0.00002538, -0.00004187, -0.00000177, 0.00001082, 0.00000420, -0.00000914, -0.00001299, 0.00001561, 0.00000726, -0.00001646, -0.00002897, 0.00003841, -0.00002272, -0.00000968, -0.00002635, 0.00003810, -0.00013220, 0.00009346, -0.00046406, 0.00053722, 0.00060154, -0.00050627, 0.00006714, -0.00011680, 0.00007748, -0.00008037, 0.00005951, -0.00008105, 0.00001903, -0.00000146, 0.00005396, -0.00006199, 0.00001712, -0.00001681, 0.00002183, -0.00003550, -0.00002326, 0.00000211, 0.00007922, -0.00006535, -0.00000258, 0.00000208, 0.00003443, -0.00004003, -0.00003648, 0.00003372, -0.00006053, 0.00005367, -0.00005005, 0.00006138, -0.00003330, 0.00002684, -0.00004311, 0.00003601, -0.00003720, 0.00002680, -0.00007147, 0.00000468, -0.00003799, 0.00013181, 0.00008931, 0.00000440, -0.00005715, -0.00001909, },
		{ -0.00004766, -0.00000102, -0.00000785, 0.00002627, -0.00002184, 0.00000017, -0.00000877, 0.00002691, -0.00001515, -0.00002573, 0.00000921, 0.00001448, -0.00002453, -0.00002411, 0.00003182, -0.00008969, 0.00003356, -0.00023659, -0.00005062, 0.00093584, -0.00031160, -0.00012097, -0.00003923, 0.00000824, -0.00002294, 0.00000711, -0.00001959, -0.00003524, 0.00001339, 0.00004050, -0.00001786, -0.00002461, 0.00000597, 0.00000229, -0.00000523, -0.00004112, -0.00001153, 0.00006721, -0.00000838, -0.00005159, 0.00002146, 0.00000737, 0.00002233, -0.00009167, 0.00005708, -0.00006162, 0.00003365, -0.00002449, 0.00002248, 0.00001131, -0.00000527, -0.00001121, -0.00000351, 0.00000422, -0.00001153, -0.00001617, -0.00004986, -0.00001213, 0.00004175, 0.00017346, 0.00000342, -0.00001740, -0.00007483, 0.00005216, },
		{ -0.00002142, -0.00000846, 0.00002291, -0.00000901, -0.00000699, -0.00001095, 0.00002033, 0.00000102, -0.00002158, -0.00001661, 0.00003235, -0.00002151, -0.00001969, 0.00000277, -0.00001861, -0.00006160, -0.00002116, -0.00035797, 0.00077042, 0.00019204, -0.00035835, 0.00003814, -0.00006890, 0.00003840, -0.00004770, 0.00003418, -0.00006746, 0.00001535, 0.00001906, 0.00002800, -0.00005040, 0.00001351, -0.00001016, 0.00001489, -0.00004300, -0.00002500, 0.00003927, 0.00004491, -0.00007147, 0.00001519, -0.00000493, 0.00005115, -0.00008253, 0.00000984, -0.00000398, -0.00002735, 0.00002764, -0.00003095, 0.00005402, -0.00002890, 0.00001523, -0.00003510, 0.00003119, -0.00003311, 0.00001432, -0.00006561, -0.00001159, -0.00001910, 0.00018457, 0.00006211, -0.00000626, -0.00008580, 0.00002703, -0.00001610, },
		{ -0.00000118, 0.00000246, 0.00001761, -0.00002283, -0.00000132, 0.00000101, 0.00001939, -0.00001697, -0.00002593, 0.00002020, 0.00000290, -0.00002368, -0.00001485, 0.00001758, -0.00009102, 0.00004495, -0.00033644, 0.00030939, 0.00064880, -0.00037320, -0.00003153, -0.00005823, 0.00002745, -0.00004094, 0.00002722, -0.00004312, -0.00001549, 0.00001434, 0.00004315, -0.00003619, -0.00000655, -0.00000428, 0.00001248, -0.00002286, -0.00004216, 0.00001232, 0.00006939, -0.00004935, -0.00001701, 0.00000185, 0.00004028, -0.00002425, -0.00006234, 0.00004835, -0.00006165, 0.00004266, -0.00004021, 0.00004531, -0.00000686, 0.00000440, -0.00002686, 0.00001429, -0.00001212, -0.00000104, -0.00003487, -0.00003824, -0.00003869, 0.00012429, 0.00015163, -0.00000000, -0.00006622, -0.00002675, 0.00002488, -0.00004281, },
		{ -0.00001039, 0.00002900, -0.00002054, -0.00000414, -0.00000645, 0.00002060, -0.00000349, -0.00002614, -0.00000285, 0.00002354, -0.00002050, -0.00002296, 0.00001754, -0.00006099, 0.00000513, -0.00016666, -0.00010506, 0.00076461, -0.00008057, -0.00021542, 0.00000202, -0.00002660, 0.00000113, -0.00001249, 0.00000306, -0.00004747, 0.00001092, 0.00003675, -0.00000084, -0.00003127, 0.00000365, 0.00000107, 0.00000163, -0.00004945, -0.00001393, 0.00006666, -0.00000410, -0.00004901, 0.00000757, 0.00001598, 0.00003350, -0.00009630, 0.00004127, -0.00003513, 0.00000502, -0.00000364, -0.00000204, 0.00003705, -0.00001898, -0.00000282, -0.00001777, 0.00001758, -0.00002314, -0.00000412, -0.00005170, -0.00004662, 0.00004188, 0.00020758, 0.00004390, -0.00004792, -0.00006821, 0.00003677, -0.00003842, -0.00000678, },
		{ 0.00001686, 0.00000290, -0.00001984, -0.00000115, 0.00000752, 0.00001328, -0.00002126, -0.00001937, 0.00002307, -0.00000234, -0.00002842, 0.00000357, -0.00001977, -0.00004524, -0.00003096, -0.00027978, 0.00053216, 0.00033381, -0.00032176, -0.00000620, -0.00004426, 0.00001441, -0.00002843, 0.00002005, -0.00004487, -0.00000974, 0.00002821, 0.00002693, -0.00003556, -0.00000205, -0.00000404, 0.00001253, -0.00003693, -0.00004015, 0.00004695, 0.00003848, -0.00005716, -0.00000287, 0.00000081, 0.00005695, -0.00007018, -0.00001772, 0.00001987, -0.00004058, 0.00002915, -0.00003456, 0.00005178, -0.00001354, 0.00000048, -0.00002683, 0.00001833, -0.00001661, -0.00000497, -0.00003439, -0.00005660, -0.00002550, 0.00019427, 0.00012913, -0.00003211, -0.00008374, 0.00001230, -0.00001038, -0.00002456, -0.00000889, },
		{ 0.00002478, -0.00002654, -0.00000156, -0.00000135, 0.00001882, -0.00000870, -0.00002634, 0.00000704, 0.00001760, -0.00002538, -0.00001262, 0.00000468, -0.00005974, 0.00001195, -0.00024463, 0.00016823, 0.00059630, -0.00020321, -0.00012121, -0.00001234, -0.00001140, -0.00001222, 0.00000413, -0.00001570, -0.00003503, 0.00001703, 0.00003904, -0.00001856, -0.00001669, -0.00000312, 0.00000885, -0.00001233, -0.00005661, 0.00001565, 0.00006422, -0.00003815, -0.00002448, -0.00000404, 0.00004741, -0.00001092, -0.00007623, 0.00004550, -0.00004528, 0.00002058, -0.00002392, 0.00002263, 0.00002162, -0.00001455, -0.00001449, -0.00000391, 0.00000678, -0.00001963, -0.00001220, -0.00005338, -0.00007082, 0.00012355, 0.00021343, 0.00000968, -0.00008653, -0.00002635, 0.00001482, -0.00003507, -0.00000636, -0.00000217, },
		{ -0.00001181, -0.00001485, 0.00000022, 0.00001138, 0.00000724, -0.00002463, -0.00001103, 0.00002375, -0.00000919, -0.00002636, 0.00000969, -0.00004460, -0.00000302, -0.00013246, -0.00010593, 0.00057771, 0.00009286, -0.00024002, -0.00000176, -0.00002918, -0.00000100, -0.00001045, 0.00000497, -0.00003947, -0.00000491, 0.00004048, 0.00000565, -0.00002593, -0.00000406, 0.00000024, 0.00000778, -0.00005261, -0.00002227, 0.00006883, -0.00000322, -0.00004194, -0.00000816, 0.00002413, 0.00004074, -0.00009044, 0.00001774, -0.00000838, -0.00001517, 0.00000631, -0.00001626, 0.00004788, -0.00001561, -0.00000863, -0.00001856, 0.00001503, -0.00001732, -0.00001159, -0.00003005, -0.00009172, 0.00002725, 0.00024818, 0.00009455, -0.00008193, -0.00006045, 0.00001841, -0.00002785, -0.00001206, -0.00001234, 0.00002670, },
		{ -0.00002662, 0.00000025, 0.00000368, 0.00001537, -0.00001381, -0.00002324, 0.00001430, 0.00001158, -0.00003002, 0.00000036, -0.00002074, -0.00002734, -0.00004361, -0.00020913, 0.00035420, 0.00037757, -0.00022533, -0.00006798, -0.00001231, -0.00001393, -0.00000703, 0.00000376, -0.00002210, -0.00002875, 0.00003031, 0.00002754, -0.00002338, -0.00000899, -0.00000612, 0.00001343, -0.00002960, -0.00005474, 0.00005083, 0.00003414, -0.00004456, -0.00001657, 0.00000224, 0.00006234, -0.00005653, -0.00003812, 0.00002920, -0.00003625, 0.00001716, -0.00002779, 0.00003744, 0.00001079, -0.00001758, -0.00001618, 0.00000233, 0.00000114, -0.00002319, -0.00000841, -0.00007923, -0.00006456, 0.00021123, 0.00020219, -0.00005119, -0.00008949, 0.00000342, -0.00001225, -0.00002016, -0.00001388, 0.00001202, 0.00001240, },
		{ -0.00001039, 0.00000307, 0.00001258, 0.00000121, -0.00002587, -0.00000279, 0.00002363, -0.00001885, -0.00001550, -0.00000311, -0.00003762, -0.00000547, -0.00018536, 0.00009241, 0.00049222, -0.00004822, -0.00016875, 0.00000231, -0.00002628, -0.00000388, -0.00000184, -0.00000557, -0.00003567, 0.00000588, 0.00004239, -0.00001060, -0.00001491, -0.00000819, 0.00000577, -0.00000127, -0.00006699, 0.00001295, 0.00006207, -0.00003017, -0.00002682, -0.00001320, 0.00005455, -0.00000095, -0.00007893, 0.00003094, -0.00002233, 0.00000013, -0.00001210, 0.00000402, 0.00004067, -0.00001997, -0.00001410, -0.00001068, 0.00001146, -0.00002151, -0.00001014, -0.00003961, -0.00012179, 0.00011007, 0.00028635, 0.00002818, -0.00011395, -0.00002111, -0.00000144, -0.00002275, -0.00001329, -0.00000543, 0.00002649, -0.00002143, },
		{ 0.00000224, 0.00000768, 0.00001054, -0.00001755, -0.00001876, 0.00002116, 0.00000135, -0.00002660, 0.00000115, -0.00003232, -0.00000273, -0.00011445, -0.00008722, 0.00041553, 0.00018926, -0.00020221, -0.00003476, -0.00001384, -0.00001635, 0.00000137, -0.00000245, -0.00002380, -0.00002143, 0.00004162, 0.00001059, -0.00001960, -0.00000642, -0.00000709, 0.00001641, -0.00005272, -0.00003291, 0.00006960, -0.00000209, -0.00003400, -0.00002218, 0.00002943, 0.00004612, -0.00007975, -0.00000412, 0.00000880, -0.00002061, 0.00000358, -0.00001968, 0.00004439, 0.00000097, -0.00002252, -0.00001243, 0.00000433, -0.00000419, -0.00002531, -0.00000475, -0.00012055, -0.00002259, 0.00030059, 0.00015457, -0.00011664, -0.00005517, 0.00000187, -0.00002140, -0.00001204, -0.00001728, 0.00002317, -0.00000167, -0.00002360, },
		{ 0.00000712, 0.00001088, -0.00000367, -0.00002600, 0.00000677, 0.00001907, -0.00002469, -0.00000647, -0.00002012, -0.00001235, -0.00005108, -0.00015620, 0.00023450, 0.00035358, -0.00011285, -0.00011301, 0.00000396, -0.00002897, 0.00000212, -0.00000218, -0.00001085, -0.00003256, 0.00001990, 0.00003457, -0.00001920, -0.00000496, -0.00001449, 0.00001444, -0.00001967, -0.00006771, 0.00005002, 0.00003196, -0.00003368, -0.00002614, 0.00000035, 0.00006714, -0.00004418, -0.00004899, 0.00002572, -0.00002157, 0.00000187, -0.00002026, 0.00002119, 0.00003189, -0.00002692, -0.00001391, -0.00000571, 0.00000808, -0.00002748, -0.00000180, -0.00006892, -0.00013178, 0.00021925, 0.00029257, -0.00006786, -0.00010083, -0.00000059, -0.00002084, -0.00000996, -0.00002139, 0.00000788, 0.00001789, -0.00002528, -0.00000710, },
		},
		{ { -0.00000067, -0.00000335, 0.00001744, -0.00000846, -0.00002795, -0.00000905, 0.00004573, -0.00001088, -0.00001132, -0.00001922, -0.00001879, -0.00001814, -0.00013506, 0.00005895, 0.00037905, 0.00004013, -0.00012414, -0.00002587, -0.00002783, -0.00001345, 0.00000305, 0.00000688, -0.00008455, 0.00001729, 0.00000719, -0.00000569, -0.00002701, -0.00000826, 0.00003980, -0.00000466, -0.00003609, 0.00001140, 0.00001688, 0.00001140, -0.00003609, -0.00000466, 0.00003980, -0.00000826, -0.00002701, -0.00000569, 0.00000719, 0.00001729, -0.00008455, 0.00000688, 0.00000305, -0.00001345, -0.00002783, -0.00002587, -0.00012414, 0.00004013, 0.00037905, 0.00005895, -0.00013506, -0.00001814, -0.00001879, -0.00001922, -0.00001132, -0.00001088, 0.00004573, -0.00000905, -0.00002795, -0.00000846, 0.00001744, -0.00000335, },
		{ -0.00000253, 0.00000866, 0.00000720, -0.00002288, -0.00002492, 0.00002956, 0.00002101, -0.00002302, -0.00000624, -0.00002934, -0.00000067, -0.00008862, -0.00007120, 0.00027507, 0.00022759, -0.00010920, -0.00005241, -0.00003291, -0.00000759, -0.00002114, 0.00003128, -0.00006392, -0.00003137, 0.00002648, -0.00000953, -0.00001018, -0.00003020, 0.00002908, 0.00002328, -0.00003598, -0.00000839, 0.00001781, 0.00001808, -0.00001613, -0.00003142, 0.00003215, 0.00001744, -0.00002957, -0.00001266, -0.00000371, 0.00003190, -0.00005045, -0.00004974, 0.00002908, -0.00001804, -0.00001562, -0.00003189, -0.00006331, -0.00012947, 0.00031879, 0.00026820, -0.00013510, -0.00006532, -0.00000516, -0.00003060, -0.00000403, -0.00002540, 0.00003145, 0.00002115, -0.00002936, -0.00002026, 0.00001217, 0.00000704, -0.00000451, },
		{ -0.00000145, 0.00001551, -0.00001320, -0.00002700, 0.00000026, 0.00004301, -0.00001774, -0.00000629, -0.00002601, -0.00000568, -0.00004086, -0.00010930, 0.00011198, 0.00030954, 0.00000698, -0.00009066, -0.00003548, -0.00001448, -0.00001840, 0.00000905, -0.00000633, -0.00007812, 0.00002806, -0.00000708, 0.00000008, -0.00003170, 0.00000264, 0.00003980, -0.00001572, -0.00003018, 0.00001442, 0.00001668, 0.00000741, -0.00004029, 0.00000833, 0.00003361, -0.00001376, -0.00002627, -0.00000640, 0.00001944, 0.00000245, -0.00008504, 0.00001777, -0.00000254, -0.00000989, -0.00003997, -0.00001684, -0.00016429, 0.00009301, 0.00044695, -0.00002391, -0.00013529, -0.00000540, -0.00002877, -0.00001207, -0.00001715, -0.00000214, 0.00004569, -0.00001855, -0.00002691, -0.00000388, 0.00001960, -0.00000744, 0.00000072, },
		{ 0.00000930, 0.00000228, -0.00002430, -0.00001948, 0.00003625, 0.00000910, -0.00001732, -0.00001314, -0.00001942, -0.00000925, -0.00009446, -0.00001229, 0.00025748, 0.00015792, -0.00008767, -0.00004580, -0.00003260, 0.00000027, -0.00002508, 0.00003274, -0.00007680, -0.00000857, 0.00001237, -0.00000743, -0.00001222, -0.00002648, 0.00003837, 0.00001167, -0.00003677, -0.00000144, 0.00001697, 0.00001779, -0.00002429, -0.00002273, 0.00003588, 0.00000710, -0.00002806, -0.00001609, 0.00000717, 0.00002882, -0.00006388, -0.00003592, 0.00002669, -0.00001628, -0.00002299, -0.00003039, -0.00007888, -0.00014605, 0.00043418, 0.00022240, -0.00018670, -0.00003492, -0.00001833, -0.00002436, -0.00000671, -0.00002392, 0.00003964, 0.00001130, -0.00003211, -0.00001710, 0.00001772, 0.00000304, -0.00000478, -0.00000090, },
		{ 0.00001426, -0.00001803, -0.00002423, 0.00000917, 0.00003713, -0.00002104, -0.00000492, -0.00002754, 0.00000293, -0.00006340, -0.00006756, 0.00013518, 0.00024843, -0.00001479, -0.00006089, -0.00004749, 0.00000225, -0.00002711, 0.00001753, -0.00002330, -0.00006339, 0.00003002, -0.00001762, 0.00000652, -0.00003747, 0.00001709, 0.00003343, -0.00002322, -0.00002370, 0.00001660, 0.00001607, 0.00000239, -0.00004162, 0.00002085, 0.00002312, -0.00001420, -0.00003011, 0.00000023, 0.00002502, -0.00001150, -0.00008283, 0.00002848, -0.00001032, -0.00000454, -0.00005415, -0.00000537, -0.00021530, 0.00017707, 0.00049573, -0.00012658, -0.00010830, -0.00000862, -0.00002966, -0.00000919, -0.00002066, 0.00000642, 0.00004358, -0.00002794, -0.00002398, 0.00000083, 0.00002082, -0.00001245, 0.00000424, -0.00000185, },
		{ -0.00000337, -0.00002391, -0.00001361, 0.00004068, -0.00000270, -0.00001035, -0.00002190, -0.00000377, -0.00002961, -0.00007901, 0.00002881, 0.00023055, 0.00010303, -0.00006361, -0.00004560, -0.00002788, 0.00000487, -0.00002726, 0.00003104, -0.00008482, 0.00001334, -0.00000712, 0.00000214, -0.00002002, -0.00001636, 0.00004213, 0.00000023, -0.00003490, 0.00000473, 0.00001545, 0.00001697, -0.00003218, -0.00001046, 0.00003322, 0.00000104, -0.00002924, -0.00001592, 0.00001796, 0.00002062, -0.00007273, -0.00002171, 0.00002226, -0.00001356, -0.00003197, -0.00002590, -0.00010351, -0.00015045, 0.00056823, 0.00012930, -0.00020933, -0.00001092, -0.00003174, -0.00001539, -0.00001186, -0.00001981, 0.00004577, -0.00000008, -0.00003245, -0.00001394, 0.00002392, -0.00000397, -0.00000189, -0.00000149, 0.00001053, },
		{ -0.00002242, -0.00002010, 0.00001737, 0.00002831, -0.00002025, -0.00000845, -0.00002075, 0.00000080, -0.00007512, -0.00002359, 0.00013663, 0.00019782, -0.00002866, -0.00003490, -0.00006091, 0.00002184, -0.00003959, 0.00002850, -0.00004335, -0.00004047, 0.00002077, -0.00001968, 0.00000887, -0.00004029, 0.00003164, 0.00002227, -0.00002668, -0.00001740, 0.00001803, 0.00001507, -0.00000356, -0.00003919, 0.00003029, 0.00001142, -0.00001192, -0.00003651, 0.00001332, 0.00002214, -0.00002130, -0.00008016, 0.00003999, -0.00002150, 0.00000431, -0.00007229, 0.00000999, -0.00027928, 0.00030280, 0.00050378, -0.00023007, -0.00006158, -0.00002719, -0.00001964, -0.00001295, -0.00001992, 0.00001332, 0.00003982, -0.00003663, -0.00001957, 0.00000609, 0.00001963, -0.00001634, 0.00000807, -0.00000345, 0.00001328, },
		{ -0.00002124, -0.00000782, 0.00004235, -0.00001263, -0.00000481, -0.00002821, 0.00001159, -0.00005522, -0.00004600, 0.00004700, 0.00020532, 0.00005634, -0.00003519, -0.00005396, -0.00001569, 0.00000287, -0.00002458, 0.00002342, -0.00008404, 0.00002855, -0.00002620, 0.00001561, -0.00003204, 0.00000000, 0.00003858, -0.00000846, -0.00003179, 0.00001041, 0.00001303, 0.00001570, -0.00003893, 0.00000393, 0.00002450, 0.00000067, -0.00003539, -0.00000837, 0.00002306, 0.00001242, -0.00008006, -0.00000469, 0.00001234, -0.00000558, -0.00004737, -0.00001297, -0.00014524, -0.00012783, 0.00070506, -0.00000873, -0.00019242, -0.00000538, -0.00003605, -0.00001018, -0.00001538, -0.00001558, 0.00005086, -0.00001327, -0.00002969, -0.00001111, 0.00003013, -0.00001350, 0.00000449, -0.00000526, 0.00001346, -0.00001034, },
		{ -0.00001531, 0.00002453, 0.00001746, -0.00001620, -0.00001589, -0.00000591, -0.00001444, -0.00006880, 0.00000934, 0.00012834, 0.00015404, -0.00003392, -0.00001562, -0.00007204, 0.00004101, -0.00005323, 0.00003958, -0.00006327, -0.00001271, 0.00000146, -0.00001135, 0.00000377, -0.00003645, 0.00004220, 0.00000921, -0.00002676, -0.00001150, 0.00001855, 0.00001380, -0.00001016, -0.00003267, 0.00003447, 0.00000198, -0.00001045, -0.00004144, 0.00002887, 0.00001258, -0.00002672, -0.00007686, 0.00005082, -0.00003428, 0.00001506, -0.00009264, 0.00002636, -0.00035240, 0.00047456, 0.00045122, -0.00031063, -0.00001017, -0.00005369, -0.00000197, -0.00002223, -0.00001519, 0.00001833, 0.00003419, -0.00004357, -0.00001442, 0.00001226, 0.00001478, -0.00001697, 0.00000993, -0.00000439, 0.00001144, -0.00002541, },
		{ -0.00000255, 0.00004108, -0.00001940, -0.00000284, -0.00002804, 0.00001982, -0.00007654, -0.00000528, 0.00004543, 0.00018629, 0.00001461, -0.00000265, -0.00007085, 0.00000498, -0.00000739, -0.00001537, 0.00000855, -0.00007206, 0.00003220, -0.00003848, 0.00002754, -0.00004442, 0.00002009, 0.00002789, -0.00001277, -0.00002872, 0.00001590, 0.00000952, 0.00001416, -0.00004370, 0.00001838, 0.00001152, 0.00000538, -0.00004651, 0.00000796, 0.00001853, 0.00000895, -0.00008914, 0.00001753, -0.00000629, 0.00001198, -0.00007430, 0.00001443, -0.00021322, -0.00005952, 0.00081932, -0.00017658, -0.00013567, -0.00002461, -0.00002464, -0.00001426, -0.00001331, -0.00001390, 0.00005601, -0.00002841, -0.00002343, -0.00000874, 0.00003522, -0.00002396, 0.00001328, -0.00001186, 0.00001827, -0.00001863, -0.00001623, },
		{ 0.00003030, 0.00000595, -0.00001078, -0.00002424, 0.00001327, -0.00004009, -0.00004321, 0.00002271, 0.00012156, 0.00011135, -0.00002797, -0.00000715, -0.00007593, 0.00005492, -0.00006373, 0.00004682, -0.00007813, 0.00001399, -0.00002319, 0.00000565, -0.00000938, -0.00002401, 0.00004533, -0.00000241, -0.00002498, -0.00000565, 0.00001773, 0.00001261, -0.00001706, -0.00002240, 0.00003219, -0.00000236, -0.00001329, -0.00004012, 0.00004089, 0.00000117, -0.00003056, -0.00007017, 0.00005671, -0.00004351, 0.00002203, -0.00010872, 0.00003493, -0.00042226, 0.00068568, 0.00032648, -0.00034597, 0.00002753, -0.00007613, 0.00001582, -0.00003259, -0.00000910, 0.00002276, 0.00002568, -0.00004724, -0.00000949, 0.00001940, 0.00000576, -0.00001280, 0.00000776, -0.00000268, 0.00000736, -0.00002586, -0.00001071, },
		{ 0.00003712, -0.00002271, -0.00000508, -0.00001937, 0.00001585, -0.00008442, 0.00003061, 0.00003333, 0.00017153, -0.00002117, 0.00003060, -0.00009302, 0.00003193, -0.00002460, -0.00000058, -0.00001218, -0.00004959, 0.00002234, -0.00003927, 0.00003254, -0.00005233, 0.00003969, 0.00001220, -0.00001246, -0.00002639, 0.00002128, 0.00000487, 0.00001254, -0.00004583, 0.00003063, -0.00000278, 0.00001270, -0.00006010, 0.00003124, 0.00000366, 0.00001255, -0.00010150, 0.00004533, -0.00003442, 0.00004080, -0.00011527, 0.00005947, -0.00031340, 0.00007193, 0.00088052, -0.00034795, -0.00005036, -0.00006637, 0.00000358, -0.00002979, -0.00000364, -0.00001632, 0.00006167, -0.00004498, -0.00001385, -0.00000674, 0.00003781, -0.00003306, 0.00002226, -0.00001965, 0.00002400, -0.00002746, -0.00000934, 0.00000192, },
		{ -0.00000490, -0.00000615, -0.00002957, 0.00003071, -0.00006907, -0.00000373, 0.00001518, 0.00012294, 0.00006562, -0.00000921, -0.00001231, -0.00006868, 0.00005933, -0.00006705, 0.00004647, -0.00008305, 0.00003290, -0.00004617, 0.00002639, -0.00002808, -0.00000355, 0.00003929, -0.00000969, -0.00002316, 0.00000084, 0.00001493, 0.00001206, -0.00002393, -0.00000930, 0.00002350, -0.00000017, -0.00002268, -0.00002877, 0.00004330, -0.00000586, -0.00003742, -0.00005579, 0.00005182, -0.00004182, 0.00001662, -0.00011026, 0.00002166, -0.00046767, 0.00091598, 0.00013175, -0.00032186, 0.00003534, -0.00008183, 0.00002445, -0.00003771, -0.00000584, 0.00002893, 0.00001283, -0.00004589, -0.00000592, 0.00002722, -0.00000698, -0.00000351, 0.00000055, 0.00000302, -0.00000006, -0.00002279, -0.00000720, 0.00003453, },
		{ -0.00002327, -0.00001027, -0.00000319, -0.00000157, -0.00007336, 0.00005061, 0.00002212, 0.00015513, -0.00004680, 0.00005871, -0.00011460, 0.00006011, -0.00004464, 0.00001611, -0.00003454, -0.00002087, 0.00000104, -0.00002727, 0.00002722, -0.00005154, 0.00005408, -0.00000496, -0.00000868, -0.00002469, 0.00002614, -0.00000057, 0.00001098, -0.00004497, 0.00003867, -0.00001512, 0.00001901, -0.00007168, 0.00005645, -0.00001835, 0.00002204, -0.00011555, 0.00007572, -0.00006874, 0.00007784, -0.00016739, 0.00011899, -0.00044381, 0.00027634, 0.00086004, -0.00049046, 0.00004273, -0.00011985, 0.00004305, -0.00005419, 0.00001241, -0.00002230, 0.00006699, -0.00006143, -0.00000186, -0.00000481, 0.00003663, -0.00003836, 0.00002868, -0.00002610, 0.00002877, -0.00003534, -0.00000159, 0.00000548, 0.00003113, },
		{ -0.00000384, -0.00002856, 0.00004007, -0.00009213, 0.00003928, -0.00000746, 0.00013279, 0.00001669, 0.00002085, -0.00003073, -0.00004933, 0.00005249, -0.00006129, 0.00003685, -0.00007528, 0.00003876, -0.00006036, 0.00004427, -0.00004744, 0.00002174, 0.00002456, -0.00001094, -0.00002287, 0.00000863, 0.00000951, 0.00001283, -0.00003055, 0.00000524, 0.00000974, 0.00000815, -0.00003865, -0.00000610, 0.00003205, -0.00000315, -0.00005166, -0.00002984, 0.00003092, -0.00002220, -0.00001001, -0.00008614, -0.00002951, -0.00046159, 0.00113268, -0.00011452, -0.00023709, 0.00000458, -0.00006182, 0.00001620, -0.00003173, -0.00000958, 0.00003920, -0.00000562, -0.00003806, -0.00000477, 0.00003505, -0.00002196, 0.00000980, -0.00001124, 0.00001287, -0.00001110, -0.00001574, -0.00000551, 0.00003731, -0.00001423, },
		{ -0.00001562, 0.00001648, -0.00002902, -0.00004364, 0.00004875, 0.00002116, 0.00013070, -0.00005749, 0.00007544, -0.00012892, 0.00008322, -0.00006190, 0.00002957, -0.00005254, 0.00000724, -0.00002585, -0.00000526, 0.00001138, -0.00003997, 0.00005938, -0.00001962, -0.00000354, -0.00002275, 0.00002956, -0.00000597, 0.00000931, -0.00004104, 0.00004116, -0.00002262, 0.00002064, -0.00007608, 0.00007667, -0.00004114, 0.00003280, -0.00012654, 0.00010220, -0.00010152, 0.00011460, -0.00022121, 0.00018158, -0.00059118, 0.00054992, 0.00073911, -0.00057314, 0.00011837, -0.00016836, 0.00008277, -0.00008044, 0.00003032, -0.00002904, 0.00006963, -0.00007511, 0.00001075, -0.00000253, 0.00003094, -0.00003796, 0.00003000, -0.00002855, 0.00003032, -0.00004033, 0.00000560, 0.00000834, 0.00002389, -0.00002244, },
		{ -0.00001991, 0.00003702, -0.00010092, 0.00007382, -0.00003464, 0.00014586, -0.00003123, 0.00005700, -0.00005824, -0.00002071, 0.00003620, -0.00004775, 0.00001954, -0.00005567, 0.00002982, -0.00006093, 0.00005319, -0.00006171, 0.00004691, 0.00000378, -0.00000613, -0.00002482, 0.00001794, 0.00000120, 0.00001548, -0.00003681, 0.00001968, -0.00000679, 0.00002047, -0.00005880, 0.00002583, 0.00000672, 0.00001171, -0.00007524, 0.00000882, -0.00000797, 0.00001890, -0.00006318, -0.00002874, -0.00013138, -0.00037721, 0.00129548, -0.00037937, -0.00010518, -0.00006262, -0.00001436, -0.00001202, -0.00001164, -0.00002273, 0.00005475, -0.00002985, -0.00002310, -0.00000673, 0.00004200, -0.00003694, 0.00002479, -0.00002571, 0.00002561, -0.00002490, -0.00000507, -0.00000601, 0.00003893, -0.00002189, -0.00000401, },
		{ 0.00003383, -0.00005930, -0.00000149, 0.00002600, 0.00003460, 0.00009474, -0.00005060, 0.00007665, -0.00013098, 0.00009607, -0.00007141, 0.00003513, -0.00006050, 0.00002725, -0.00005030, 0.00002064, -0.00001167, -0.00001844, 0.00005367, -0.00002841, 0.00000053, -0.00001927, 0.00003024, -0.00001005, 0.00000707, -0.00003419, 0.00003759, -0.00002347, 0.00001498, -0.00006903, 0.00008494, -0.00005697, 0.00003824, -0.00012781, 0.00011613, -0.00012206, 0.00013847, -0.00026177, 0.00022811, -0.00073023, 0.00087191, 0.00051554, -0.00057444, 0.00015362, -0.00019408, 0.00010936, -0.00009896, 0.00004335, -0.00003208, 0.00006627, -0.00008263, 0.00002151, 0.00000064, 0.00002072, -0.00003108, 0.00002474, -0.00002506, 0.00002671, -0.00004060, 0.00001063, 0.00001098, 0.00001601, -0.00002161, -0.00001787, },
		{ 0.00002093, -0.00009090, 0.00009003, -0.00005481, 0.00015392, -0.00007121, 0.00009153, -0.00008779, 0.00001120, 0.00001541, -0.00003082, -0.00000061, -0.00002901, 0.00000880, -0.00004709, 0.00004954, -0.00006606, 0.00006656, -0.00001891, 0.00000318, -0.00002869, 0.00002826, -0.00000964, 0.00002015, -0.00004255, 0.00003248, -0.00002329, 0.00003349, -0.00007872, 0.00006174, -0.00002893, 0.00003704, -0.00010629, 0.00005708, -0.00006166, 0.00007902, -0.00014144, 0.00006176, -0.00028773, -0.00019553, 0.00136461, -0.00062108, 0.00004783, -0.00015300, 0.00005369, -0.00005699, 0.00002108, -0.00004461, 0.00007460, -0.00005828, -0.00000180, -0.00001190, 0.00004710, -0.00004944, 0.00003847, -0.00003989, 0.00003876, -0.00003948, 0.00000792, -0.00000851, 0.00003975, -0.00002839, -0.00000531, -0.00000508, },
		{ -0.00008360, 0.00004265, -0.00001021, 0.00006021, 0.00004847, -0.00002751, 0.00006212, -0.00011935, 0.00009652, -0.00007088, 0.00003072, -0.00005530, 0.00003359, -0.00006462, 0.00004278, -0.00003614, 0.00000938, 0.00003755, -0.00002940, 0.00000150, -0.00001298, 0.00002683, -0.00001133, 0.00000352, -0.00002471, 0.00002837, -0.00001737, 0.00000136, -0.00004865, 0.00007636, -0.00005906, 0.00003205, -0.00011309, 0.00010932, -0.00011961, 0.00013596, -0.00027205, 0.00023543, -0.00082682, 0.00120510, 0.00020711, -0.00048838, 0.00013432, -0.00018343, 0.00011123, -0.00010062, 0.00004463, -0.00002679, 0.00005363, -0.00008065, 0.00002775, 0.00000519, 0.00000674, -0.00001826, 0.00001287, -0.00001501, 0.00001703, -0.00003495, 0.00001211, 0.00001409, 0.00000765, -0.00002143, -0.00001456, 0.00004312, },
		{ -0.00006300, 0.00008335, -0.00005935, 0.00014942, -0.00009635, 0.00011658, -0.00011168, 0.00003926, -0.00000346, -0.00001644, -0.00001693, -0.00000280, -0.00001757, -0.00002250, 0.00003348, -0.00005817, 0.00007636, -0.00003894, 0.00001432, -0.00003315, 0.00003824, -0.00002179, 0.00002637, -0.00004746, 0.00004233, -0.00003707, 0.00004360, -0.00009317, 0.00009422, -0.00006760, 0.00006713, -0.00013895, 0.00010731, -0.00012130, 0.00014868, -0.00023478, 0.00017463, -0.00048871, 0.00008723, 0.00131037, -0.00079792, 0.00018918, -0.00024522, 0.00012813, -0.00010939, 0.00006023, -0.00007106, 0.00009529, -0.00008732, 0.00002341, -0.00001953, 0.00004943, -0.00005725, 0.00004785, -0.00005052, 0.00004926, -0.00005216, 0.00002117, -0.00001222, 0.00003988, -0.00003446, -0.00000537, 0.00001196, -0.00000455, },
		{ 0.00007733, -0.00004848, 0.00009052, -0.00000221, 0.00000611, 0.00003600, -0.00009686, 0.00008645, -0.00006173, 0.00001797, -0.00003782, 0.00002468, -0.00006395, 0.00005444, -0.00005547, 0.00003800, 0.00001396, -0.00002252, -0.00000167, -0.00000321, 0.00001847, -0.00000852, -0.00000217, -0.00001299, 0.00001464, -0.00000547, -0.00001868, -0.00001628, 0.00004975, -0.00004381, 0.00001057, -0.00007905, 0.00007701, -0.00008706, 0.00009699, -0.00023801, 0.00018227, -0.00084449, 0.00150107, -0.00014973, -0.00032720, 0.00005940, -0.00013153, 0.00008236, -0.00007983, 0.00002956, -0.00001006, 0.00002978, -0.00006685, 0.00002718, 0.00001148, -0.00000955, -0.00000129, -0.00000404, 0.00000056, 0.00000176, -0.00002327, 0.00000928, 0.00001833, -0.00000148, -0.00002142, -0.00000520, 0.00004077, -0.00009427, },
		{ 0.00005591, -0.00004534, 0.00012850, -0.00010262, 0.00012677, -0.00012409, 0.00005767, -0.00001488, -0.00000985, -0.00002316, 0.00001494, -0.00004053, 0.00000565, 0.00000900, -0.00003900, 0.00007417, -0.00005245, 0.00002425, -0.00003626, 0.00004593, -0.00003332, 0.00003292, -0.00005095, 0.00004820, -0.00004598, 0.00004784, -0.00009746, 0.00011558, -0.00010030, 0.00009375, -0.00016472, 0.00014879, -0.00017381, 0.00021263, -0.00032554, 0.00028871, -0.00070939, 0.00045500, 0.00112131, -0.00087757, 0.00028721, -0.00031533, 0.00019098, -0.00015607, 0.00009625, -0.00009531, 0.00011148, -0.00011180, 0.00004859, -0.00002810, 0.00004830, -0.00005893, 0.00005077, -0.00005492, 0.00005426, -0.00006013, 0.00003225, -0.00001583, 0.00003904, -0.00004053, -0.00000175, 0.00002589, -0.00003250, -0.00002362, },
		{ -0.00007722, 0.00011575, -0.00004870, 0.00004180, 0.00000556, -0.00006971, 0.00007115, -0.00004874, 0.00000210, -0.00001323, 0.00000382, -0.00004801, 0.00005194, -0.00006436, 0.00006155, -0.00001259, -0.00000949, -0.00000870, 0.00000975, 0.00000515, -0.00000098, -0.00001053, 0.00000044, -0.00000190, 0.00000990, -0.00004165, 0.00002354, 0.00000829, -0.00001211, -0.00002553, -0.00002720, 0.00002031, -0.00002401, 0.00001899, -0.00015390, 0.00005620, -0.00075313, 0.00170900, -0.00050587, -0.00011940, -0.00005819, -0.00004420, 0.00002476, -0.00003678, -0.00000239, 0.00001820, -0.00000470, -0.00004098, 0.00001856, 0.00001949, -0.00002627, 0.00001717, -0.00002319, 0.00001916, -0.00001713, -0.00000675, 0.00000226, 0.00002402, -0.00001190, -0.00002004, 0.00000831, 0.00002667, -0.00008744, 0.00009358, },
		{ -0.00001629, 0.00009267, -0.00009061, 0.00012107, -0.00012310, 0.00006400, -0.00001633, -0.00001341, -0.00001590, 0.00001848, -0.00005198, 0.00002885, -0.00001731, -0.00001253, 0.00006057, -0.00005720, 0.00003036, -0.00003607, 0.00004921, -0.00004184, 0.00003801, -0.00005212, 0.00004940, -0.00004882, 0.00004465, -0.00008879, 0.00011985, -0.00011867, 0.00010851, -0.00017488, 0.00017043, -0.00020501, 0.00025328, -0.00039192, 0.00037591, -0.00091233, 0.00087144, 0.00080910, -0.00084493, 0.00031928, -0.00034320, 0.00022542, -0.00018363, 0.00011879, -0.00010972, 0.00011728, -0.00012601, 0.00006897, -0.00003548, 0.00004334, -0.00005403, 0.00004634, -0.00005168, 0.00005189, -0.00006120, 0.00003887, -0.00001781, 0.00003650, -0.00004624, 0.00000698, 0.00003183, -0.00005478, 0.00001723, 0.00001580, },
		{ 0.00012734, -0.00008270, 0.00007089, -0.00002113, -0.00004533, 0.00005741, -0.00003836, -0.00000967, 0.00001045, -0.00002154, -0.00002141, 0.00003589, -0.00006039, 0.00007544, -0.00003724, 0.00000674, -0.00001812, 0.00002453, -0.00001196, 0.00001092, -0.00002150, 0.00001480, -0.00001929, 0.00002587, -0.00006300, 0.00006392, -0.00004100, 0.00003061, -0.00007088, 0.00003557, -0.00005302, 0.00006170, -0.00009064, -0.00002594, -0.00014063, -0.00053775, 0.00178643, -0.00080913, 0.00009631, -0.00019444, 0.00006298, -0.00005155, 0.00002203, -0.00004690, 0.00005447, -0.00004623, -0.00000534, 0.00000228, 0.00002877, -0.00004143, 0.00003423, -0.00004125, 0.00003740, -0.00003657, 0.00001225, -0.00000782, 0.00003089, -0.00002395, -0.00001518, 0.00002201, 0.00000456, -0.00006438, 0.00008773, -0.00008824, },
		{ 0.00004836, -0.00006562, 0.00010311, -0.00011131, 0.00006015, -0.00000934, -0.00002538, 0.00000388, 0.00000648, -0.00004722, 0.00003995, -0.00003827, 0.00001537, 0.00003860, -0.00005290, 0.00003116, -0.00003124, 0.00004637, -0.00004508, 0.00003964, -0.00004993, 0.00004564, -0.00004541, 0.00003422, -0.00006717, 0.00010445, -0.00011728, 0.00010539, -0.00016323, 0.00016409, -0.00020362, 0.00025543, -0.00041350, 0.00040749, -0.00105413, 0.00128463, 0.00040819, -0.00070601, 0.00027735, -0.00031808, 0.00022057, -0.00018232, 0.00011969, -0.00010815, 0.00010806, -0.00012517, 0.00007995, -0.00003940, 0.00003455, -0.00004319, 0.00003515, -0.00004105, 0.00004185, -0.00005440, 0.00003951, -0.00001680, 0.00003120, -0.00005024, 0.00002033, 0.00002703, -0.00006462, 0.00004896, -0.00002552, 0.00001930, },
		{ -0.00009905, 0.00008730, -0.00003798, -0.00002969, 0.00005100, -0.00003622, -0.00001063, 0.00002482, -0.00004199, 0.00000764, 0.00001102, -0.00004474, 0.00007750, -0.00005589, 0.00002280, -0.00002767, 0.00003885, -0.00003064, 0.00002567, -0.00003421, 0.00002902, -0.00003557, 0.00003967, -0.00007822, 0.00009727, -0.00008889, 0.00007568, -0.00011658, 0.00009819, -0.00012962, 0.00015479, -0.00021484, 0.00012688, -0.00038873, -0.00020464, 0.00170912, -0.00101502, 0.00027976, -0.00032015, 0.00016851, -0.00013083, 0.00008508, -0.00009561, 0.00009208, -0.00008850, 0.00003520, -0.00001954, 0.00003832, -0.00005327, 0.00004734, -0.00005513, 0.00005186, -0.00005308, 0.00003070, -0.00001907, 0.00003795, -0.00003726, -0.00000512, 0.00003110, -0.00001906, -0.00003124, 0.00006241, -0.00007927, 0.00012095, },
		{ -0.00003600, 0.00007984, -0.00009481, 0.00005154, 0.00000093, -0.00004015, 0.00003068, -0.00001719, -0.00002663, 0.00003541, -0.00004830, 0.00003867, 0.00001280, -0.00004115, 0.00002653, -0.00002148, 0.00003667, -0.00004152, 0.00003617, -0.00004351, 0.00003700, -0.00003648, 0.00001840, -0.00003553, 0.00007095, -0.00009514, 0.00008268, -0.00012840, 0.00012734, -0.00016489, 0.00021098, -0.00037724, 0.00036172, -0.00109470, 0.00163608, -0.00002975, -0.00048690, 0.00016996, -0.00024158, 0.00017469, -0.00014904, 0.00009559, -0.00008797, 0.00008210, -0.00010691, 0.00007827, -0.00003799, 0.00002239, -0.00002791, 0.00001906, -0.00002486, 0.00002551, -0.00004031, 0.00003376, -0.00001207, 0.00002224, -0.00005049, 0.00003576, 0.00001207, -0.00005887, 0.00006360, -0.00005685, 0.00005094, 0.00000480, },
		{ 0.00008935, -0.00004286, -0.00002523, 0.00005446, -0.00004484, 0.00000286, 0.00002403, -0.00004916, 0.00002984, -0.00001526, -0.00002177, 0.00006838, -0.00006614, 0.00003573, -0.00003493, 0.00005005, -0.00004787, 0.00004075, -0.00004692, 0.00004169, -0.00004901, 0.00004918, -0.00008403, 0.00011707, -0.00012613, 0.00011351, -0.00015251, 0.00014803, -0.00019359, 0.00023583, -0.00033004, 0.00027568, -0.00065281, 0.00021679, 0.00147747, -0.00109612, 0.00039837, -0.00040828, 0.00025044, -0.00019564, 0.00013875, -0.00013818, 0.00012280, -0.00012372, 0.00007400, -0.00004329, 0.00004668, -0.00006043, 0.00005480, -0.00006271, 0.00005986, -0.00006361, 0.00004550, -0.00002917, 0.00004362, -0.00005054, 0.00001056, 0.00003155, -0.00003685, 0.00000271, 0.00002573, -0.00005430, 0.00009786, -0.00009748, },
		{ 0.00005912, -0.00008096, 0.00004517, 0.00000762, -0.00004998, 0.00005598, -0.00004436, 0.00000414, 0.00001647, -0.00004514, 0.00005283, -0.00001201, -0.00002477, 0.00001771, -0.00000779, 0.00002073, -0.00003085, 0.00002685, -0.00003246, 0.00002403, -0.00002344, 0.00000019, 0.00000088, 0.00002483, -0.00005598, 0.00004364, -0.00007472, 0.00006489, -0.00009280, 0.00012218, -0.00028220, 0.00023078, -0.00100700, 0.00187220, -0.00044567, -0.00022804, 0.00001995, -0.00012742, 0.00009593, -0.00008848, 0.00004912, -0.00005118, 0.00004162, -0.00007226, 0.00006294, -0.00003039, 0.00000774, -0.00001020, 0.00000067, -0.00000602, 0.00000564, -0.00002098, 0.00002251, -0.00000380, 0.00000927, -0.00004483, 0.00004913, -0.00000916, -0.00003908, 0.00005822, -0.00007061, 0.00006970, -0.00002911, -0.00001047, },
		{ -0.00003820, -0.00002994, 0.00006601, -0.00006231, 0.00002981, 0.00000694, -0.00003858, 0.00003770, -0.00003517, 0.00000235, 0.00005121, -0.00006757, 0.00004365, -0.00003799, 0.00005569, -0.00006057, 0.00005318, -0.00005738, 0.00005119, -0.00005817, 0.00005335, -0.00007914, 0.00011943, -0.00014572, 0.00013607, -0.00017015, 0.00017427, -0.00023052, 0.00028596, -0.00041168, 0.00038771, -0.00088773, 0.00067762, 0.00111776, -0.00104762, 0.00043428, -0.00044056, 0.00029204, -0.00023147, 0.00017094, -0.00016506, 0.00013905, -0.00014450, 0.00010402, -0.00006456, 0.00005220, -0.00006212, 0.00005595, -0.00006323, 0.00006017, -0.00006623, 0.00005425, -0.00003596, 0.00004604, -0.00006155, 0.00003044, 0.00002154, -0.00004301, 0.00002797, -0.00001139, -0.00002208, 0.00006443, -0.00008262, 0.00008010, },
		{ -0.00007559, 0.00004688, 0.00000471, -0.00004765, 0.00007101, -0.00006503, 0.00003580, -0.00001116, -0.00003044, 0.00005596, -0.00003194, -0.00000702, 0.00000680, 0.00000777, 0.00000051, -0.00001428, 0.00001218, -0.00001726, 0.00000777, -0.00000809, -0.00001705, 0.00003570, -0.00002584, -0.00000731, -0.00000421, -0.00001153, -0.00001209, 0.00000022, 0.00000218, -0.00014143, 0.00002487, -0.00078472, 0.00195541, -0.00078472, 0.00002487, -0.00014143, 0.00000218, 0.00000022, -0.00001209, -0.00001153, -0.00000421, -0.00000731, -0.00002584, 0.00003570, -0.00001705, -0.00000809, 0.00000777, -0.00001726, 0.00001218, -0.00001428, 0.00000051, 0.00000777, 0.00000680, -0.00000702, -0.00003194, 0.00005596, -0.00003044, -0.00001116, 0.00003580, -0.00006503, 0.00007101, -0.00004765, 0.00000471, 0.00004688, },
		{ -0.00003820, 0.00008010, -0.00008262, 0.00006443, -0.00002208, -0.00001139, 0.00002797, -0.00004301, 0.00002154, 0.00003044, -0.00006155, 0.00004604, -0.00003596, 0.00005425, -0.00006623, 0.00006017, -0.00006323, 0.00005595, -0.00006212, 0.00005220, -0.00006456, 0.00010402, -0.00014450, 0.00013905, -0.00016506, 0.00017094, -0.00023147, 0.00029204, -0.00044056, 0.00043428, -0.00104762, 0.00111776, 0.00067762, -0.00088773, 0.00038771, -0.00041168, 0.00028596, -0.00023052, 0.00017427, -0.00017015, 0.00013607, -0.00014572, 0.00011943, -0.00007914, 0.00005335, -0.00005817, 0.00005119, -0.00005738, 0.00005318, -0.00006057, 0.00005569, -0.00003799, 0.00004365, -0.00006757, 0.00005121, 0.00000235, -0.00003517, 0.00003770, -0.00003858, 0.00000694, 0.00002981, -0.00006231, 0.00006601, -0.00002994, },
		{ 0.00005912, -0.00001047, -0.00002911, 0.00006970, -0.00007061, 0.00005822, -0.00003908, -0.00000916, 0.00004913, -0.00004483, 0.00000927, -0.00000380, 0.00002251, -0.00002098, 0.00000564, -0.00000602, 0.00000067, -0.00001020, 0.00000774, -0.00003039, 0.00006294, -0.00007226, 0.00004162, -0.00005118, 0.00004912, -0.00008848, 0.00009593, -0.00012742, 0.00001995, -0.00022804, -0.00044567, 0.00187220, -0.00100700, 0.00023078, -0.00028220, 0.00012218, -0.00009280, 0.00006489, -0.00007472, 0.00004364, -0.00005598, 0.00002483, 0.00000088, 0.00000019, -0.00002344, 0.00002403, -0.00003246, 0.00002685, -0.00003085, 0.00002073, -0.00000779, 0.00001771, -0.00002477, -0.00001201, 0.00005283, -0.00004514, 0.00001647, 0.00000414, -0.00004436, 0.00005598, -0.00004998, 0.00000762, 0.00004517, -0.00008096, },
		{ 0.00008935, -0.00009748, 0.00009786, -0.00005430, 0.00002573, 0.00000271, -0.00003685, 0.00003155, 0.00001056, -0.00005054, 0.00004362, -0.00002917, 0.00004550, -0.00006361, 0.00005986, -0.00006271, 0.00005480, -0.00006043, 0.00004668, -0.00004329, 0.00007400, -0.00012372, 0.00012280, -0.00013818, 0.00013875, -0.00019564, 0.00025044, -0.00040828, 0.00039837, -0.00109612, 0.00147747, 0.00021679, -0.00065281, 0.00027568, -0.00033004, 0.00023583, -0.00019359, 0.00014803, -0.00015251, 0.00011351, -0.00012613, 0.00011707, -0.00008403, 0.00004918, -0.00004901, 0.00004169, -0.00004692, 0.00004075, -0.00004787, 0.00005005, -0.00003493, 0.00003573, -0.00006614, 0.00006838, -0.00002177, -0.00001526, 0.00002984, -0.00004916, 0.00002403, 0.00000286, -0.00004484, 0.00005446, -0.00002523, -0.00004286, },
		{ -0.00003600, 0.00000480, 0.00005094, -0.00005685, 0.00006360, -0.00005887, 0.00001207, 0.00003576, -0.00005049, 0.00002224, -0.00001207, 0.00003376, -0.00004031, 0.00002551, -0.00002486, 0.00001906, -0.00002791, 0.00002239, -0.00003799, 0.00007827, -0.00010691, 0.00008210, -0.00008797, 0.00009559, -0.00014904, 0.00017469, -0.00024158, 0.00016996, -0.00048690, -0.00002975, 0.00163608, -0.00109470, 0.00036172, -0.00037724, 0.00021098, -0.00016489, 0.00012734, -0.00012840, 0.00008268, -0.00009514, 0.00007095, -0.00003553, 0.00001840, -0.00003648, 0.00003700, -0.00004351, 0.00003617, -0.00004152, 0.00003667, -0.00002148, 0.00002653, -0.00004115, 0.00001280, 0.00003867, -0.00004830, 0.00003541, -0.00002663, -0.00001719, 0.00003068, -0.00004015, 0.00000093, 0.00005154, -0.00009481, 0.00007984, },
		{ -0.00009905, 0.00012095, -0.00007927, 0.00006241, -0.00003124, -0.00001906, 0.00003110, -0.00000512, -0.00003726, 0.00003795, -0.00001907, 0.00003070, -0.00005308, 0.00005186, -0.00005513, 0.00004734, -0.00005327, 0.00003832, -0.00001954, 0.00003520, -0.00008850, 0.00009208, -0.00009561, 0.00008508, -0.00013083, 0.00016851, -0.00032015, 0.00027976, -0.00101502, 0.00170912, -0.00020464, -0.00038873, 0.00012688, -0.00021484, 0.00015479, -0.00012962, 0.00009819, -0.00011658, 0.00007568, -0.00008889, 0.00009727, -0.00007822, 0.00003967, -0.00003557, 0.00002902, -0.00003421, 0.00002567, -0.00003064, 0.00003885, -0.00002767, 0.00002280, -0.00005589, 0.00007750, -0.00004474, 0.00001102, 0.00000764, -0.00004199, 0.00002482, -0.00001063, -0.00003622, 0.00005100, -0.00002969, -0.00003798, 0.00008730, },
		{ 0.00004836, 0.00001930, -0.00002552, 0.00004896, -0.00006462, 0.00002703, 0.00002033, -0.00005024, 0.00003120, -0.00001680, 0.00003951, -0.00005440, 0.00004185, -0.00004105, 0.00003515, -0.00004319, 0.00003455, -0.00003940, 0.00007995, -0.00012517, 0.00010806, -0.00010815, 0.00011969, -0.00018232, 0.00022057, -0.00031808, 0.00027735, -0.00070601, 0.00040819, 0.00128463, -0.00105413, 0.00040749, -0.00041350, 0.00025543, -0.00020362, 0.00016409, -0.00016323, 0.00010539, -0.00011728, 0.00010445, -0.00006717, 0.00003422, -0.00004541, 0.00004564, -0.00004993, 0.00003964, -0.00004508, 0.00004637, -0.00003124, 0.00003116, -0.00005290, 0.00003860, 0.00001537, -0.00003827, 0.00003995, -0.00004722, 0.00000648, 0.00000388, -0.00002538, -0.00000934, 0.00006015, -0.00011131, 0.00010311, -0.00006562, },
		{ 0.00012734, -0.00008824, 0.00008773, -0.00006438, 0.00000456, 0.00002201, -0.00001518, -0.00002395, 0.00003089, -0.00000782, 0.00001225, -0.00003657, 0.00003740, -0.00004125, 0.00003423, -0.00004143, 0.00002877, 0.00000228, -0.00000534, -0.00004623, 0.00005447, -0.00004690, 0.00002203, -0.00005155, 0.00006298, -0.00019444, 0.00009631, -0.00080913, 0.00178643, -0.00053775, -0.00014063, -0.00002594, -0.00009064, 0.00006170, -0.00005302, 0.00003557, -0.00007088, 0.00003061, -0.00004100, 0.00006392, -0.00006300, 0.00002587, -0.00001929, 0.00001480, -0.00002150, 0.00001092, -0.00001196, 0.00002453, -0.00001812, 0.00000674, -0.00003724, 0.00007544, -0.00006039, 0.00003589, -0.00002141, -0.00002154, 0.00001045, -0.00000967, -0.00003836, 0.00005741, -0.00004533, -0.00002113, 0.00007089, -0.00008270, },
		{ -0.00001629, 0.00001580, 0.00001723, -0.00005478, 0.00003183, 0.00000698, -0.00004624, 0.00003650, -0.00001781, 0.00003887, -0.00006120, 0.00005189, -0.00005168, 0.00004634, -0.00005403, 0.00004334, -0.00003548, 0.00006897, -0.00012601, 0.00011728, -0.00010972, 0.00011879, -0.00018363, 0.00022542, -0.00034320, 0.00031928, -0.00084493, 0.00080910, 0.00087144, -0.00091233, 0.00037591, -0.00039192, 0.00025328, -0.00020501, 0.00017043, -0.00017488, 0.00010851, -0.00011867, 0.00011985, -0.00008879, 0.00004465, -0.00004882, 0.00004940, -0.00005212, 0.00003801, -0.00004184, 0.00004921, -0.00003607, 0.00003036, -0.00005720, 0.00006057, -0.00001253, -0.00001731, 0.00002885, -0.00005198, 0.00001848, -0.00001590, -0.00001341, -0.00001633, 0.00006400, -0.00012310, 0.00012107, -0.00009061, 0.00009267, },
		{ -0.00007722, 0.00009358, -0.00008744, 0.00002667, 0.00000831, -0.00002004, -0.00001190, 0.00002402, 0.00000226, -0.00000675, -0.00001713, 0.00001916, -0.00002319, 0.00001717, -0.00002627, 0.00001949, 0.00001856, -0.00004098, -0.00000470, 0.00001820, -0.00000239, -0.00003678, 0.00002476, -0.00004420, -0.00005819, -0.00011940, -0.00050587, 0.00170900, -0.00075313, 0.00005620, -0.00015390, 0.00001899, -0.00002401, 0.00002031, -0.00002720, -0.00002553, -0.00001211, 0.00000829, 0.00002354, -0.00004165, 0.00000990, -0.00000190, 0.00000044, -0.00001053, -0.00000098, 0.00000515, 0.00000975, -0.00000870, -0.00000949, -0.00001259, 0.00006155, -0.00006436, 0.00005194, -0.00004801, 0.00000382, -0.00001323, 0.00000210, -0.00004874, 0.00007115, -0.00006971, 0.00000556, 0.00004180, -0.00004870, 0.00011575, },
		{ 0.00005591, -0.00002362, -0.00003250, 0.00002589, -0.00000175, -0.00004053, 0.00003904, -0.00001583, 0.00003225, -0.00006013, 0.00005426, -0.00005492, 0.00005077, -0.00005893, 0.00004830, -0.00002810, 0.00004859, -0.00011180, 0.00011148, -0.00009531, 0.00009625, -0.00015607, 0.00019098, -0.00031533, 0.00028721, -0.00087757, 0.00112131, 0.00045500, -0.00070939, 0.00028871, -0.00032554, 0.00021263, -0.00017381, 0.00014879, -0.00016472, 0.00009375, -0.00010030, 0.00011558, -0.00009746, 0.00004784, -0.00004598, 0.00004820, -0.00005095, 0.00003292, -0.00003332, 0.00004593, -0.00003626, 0.00002425, -0.00005245, 0.00007417, -0.00003900, 0.00000900, 0.00000565, -0.00004053, 0.00001494, -0.00002316, -0.00000985, -0.00001488, 0.00005767, -0.00012409, 0.00012677, -0.00010262, 0.00012850, -0.00004534, },
		{ 0.00007733, -0.00009427, 0.00004077, -0.00000520, -0.00002142, -0.00000148, 0.00001833, 0.00000928, -0.00002327, 0.00000176, 0.00000056, -0.00000404, -0.00000129, -0.00000955, 0.00001148, 0.00002718, -0.00006685, 0.00002978, -0.00001006, 0.00002956, -0.00007983, 0.00008236, -0.00013153, 0.00005940, -0.00032720, -0.00014973, 0.00150107, -0.00084449, 0.00018227, -0.00023801, 0.00009699, -0.00008706, 0.00007701, -0.00007905, 0.00001057, -0.00004381, 0.00004975, -0.00001628, -0.00001868, -0.00000547, 0.00001464, -0.00001299, -0.00000217, -0.00000852, 0.00001847, -0.00000321, -0.00000167, -0.00002252, 0.00001396, 0.00003800, -0.00005547, 0.00005444, -0.00006395, 0.00002468, -0.00003782, 0.00001797, -0.00006173, 0.00008645, -0.00009686, 0.00003600, 0.00000611, -0.00000221, 0.00009052, -0.00004848, },
		{ -0.00006300, -0.00000455, 0.00001196, -0.00000537, -0.00003446, 0.00003988, -0.00001222, 0.00002117, -0.00005216, 0.00004926, -0.00005052, 0.00004785, -0.00005725, 0.00004943, -0.00001953, 0.00002341, -0.00008732, 0.00009529, -0.00007106, 0.00006023, -0.00010939, 0.00012813, -0.00024522, 0.00018918, -0.00079792, 0.00131037, 0.00008723, -0.00048871, 0.00017463, -0.00023478, 0.00014868, -0.00012130, 0.00010731, -0.00013895, 0.00006713, -0.00006760, 0.00009422, -0.00009317, 0.00004360, -0.00003707, 0.00004233, -0.00004746, 0.00002637, -0.00002179, 0.00003824, -0.00003315, 0.00001432, -0.00003894, 0.00007636, -0.00005817, 0.00003348, -0.00002250, -0.00001757, -0.00000280, -0.00001693, -0.00001644, -0.00000346, 0.00003926, -0.00011168, 0.00011658, -0.00009635, 0.00014942, -0.00005935, 0.00008335, },
		{ -0.00008360, 0.00004312, -0.00001456, -0.00002143, 0.00000765, 0.00001409, 0.00001211, -0.00003495, 0.00001703, -0.00001501, 0.00001287, -0.00001826, 0.00000674, 0.00000519, 0.00002775, -0.00008065, 0.00005363, -0.00002679, 0.00004463, -0.00010062, 0.00011123, -0.00018343, 0.00013432, -0.00048838, 0.00020711, 0.00120510, -0.00082682, 0.00023543, -0.00027205, 0.00013596, -0.00011961, 0.00010932, -0.00011309, 0.00003205, -0.00005906, 0.00007636, -0.00004865, 0.00000136, -0.00001737, 0.00002837, -0.00002471, 0.00000352, -0.00001133, 0.00002683, -0.00001298, 0.00000150, -0.00002940, 0.00003755, 0.00000938, -0.00003614, 0.00004278, -0.00006462, 0.00003359, -0.00005530, 0.00003072, -0.00007088, 0.00009652, -0.00011935, 0.00006212, -0.00002751, 0.00004847, 0.00006021, -0.00001021, 0.00004265, },
		{ 0.00002093, -0.00000508, -0.00000531, -0.00002839, 0.00003975, -0.00000851, 0.00000792, -0.00003948, 0.00003876, -0.00003989, 0.00003847, -0.00004944, 0.00004710, -0.00001190, -0.00000180, -0.00005828, 0.00007460, -0.00004461, 0.00002108, -0.00005699, 0.00005369, -0.00015300, 0.00004783, -0.00062108, 0.00136461, -0.00019553, -0.00028773, 0.00006176, -0.00014144, 0.00007902, -0.00006166, 0.00005708, -0.00010629, 0.00003704, -0.00002893, 0.00006174, -0.00007872, 0.00003349, -0.00002329, 0.00003248, -0.00004255, 0.00002015, -0.00000964, 0.00002826, -0.00002869, 0.00000318, -0.00001891, 0.00006656, -0.00006606, 0.00004954, -0.00004709, 0.00000880, -0.00002901, -0.00000061, -0.00003082, 0.00001541, 0.00001120, -0.00008779, 0.00009153, -0.00007121, 0.00015392, -0.00005481, 0.00009003, -0.00009090, },
		{ 0.00003383, -0.00001787, -0.00002161, 0.00001601, 0.00001098, 0.00001063, -0.00004060, 0.00002671, -0.00002506, 0.00002474, -0.00003108, 0.00002072, 0.00000064, 0.00002151, -0.00008263, 0.00006627, -0.00003208, 0.00004335, -0.00009896, 0.00010936, -0.00019408, 0.00015362, -0.00057444, 0.00051554, 0.00087191, -0.00073023, 0.00022811, -0.00026177, 0.00013847, -0.00012206, 0.00011613, -0.00012781, 0.00003824, -0.00005697, 0.00008494, -0.00006903, 0.00001498, -0.00002347, 0.00003759, -0.00003419, 0.00000707, -0.00001005, 0.00003024, -0.00001927, 0.00000053, -0.00002841, 0.00005367, -0.00001844, -0.00001167, 0.00002064, -0.00005030, 0.00002725, -0.00006050, 0.00003513, -0.00007141, 0.00009607, -0.00013098, 0.00007665, -0.00005060, 0.00009474, 0.00003460, 0.00002600, -0.00000149, -0.00005930, },
		{ -0.00001991, -0.00000401, -0.00002189, 0.00003893, -0.00000601, -0.00000507, -0.00002490, 0.00002561, -0.00002571, 0.00002479, -0.00003694, 0.00004200, -0.00000673, -0.00002310, -0.00002985, 0.00005475, -0.00002273, -0.00001164, -0.00001202, -0.00001436, -0.00006262, -0.00010518, -0.00037937, 0.00129548, -0.00037721, -0.00013138, -0.00002874, -0.00006318, 0.00001890, -0.00000797, 0.00000882, -0.00007524, 0.00001171, 0.00000672, 0.00002583, -0.00005880, 0.00002047, -0.00000679, 0.00001968, -0.00003681, 0.00001548, 0.00000120, 0.00001794, -0.00002482, -0.00000613, 0.00000378, 0.00004691, -0.00006171, 0.00005319, -0.00006093, 0.00002982, -0.00005567, 0.00001954, -0.00004775, 0.00003620, -0.00002071, -0.00005824, 0.00005700, -0.00003123, 0.00014586, -0.00003464, 0.00007382, -0.00010092, 0.00003702, },
		{ -0.00001562, -0.00002244, 0.00002389, 0.00000834, 0.00000560, -0.00004033, 0.00003032, -0.00002855, 0.00003000, -0.00003796, 0.00003094, -0.00000253, 0.00001075, -0.00007511, 0.00006963, -0.00002904, 0.00003032, -0.00008044, 0.00008277, -0.00016836, 0.00011837, -0.00057314, 0.00073911, 0.00054992, -0.00059118, 0.00018158, -0.00022121, 0.00011460, -0.00010152, 0.00010220, -0.00012654, 0.00003280, -0.00004114, 0.00007667, -0.00007608, 0.00002064, -0.00002262, 0.00004116, -0.00004104, 0.00000931, -0.00000597, 0.00002956, -0.00002275, -0.00000354, -0.00001962, 0.00005938, -0.00003997, 0.00001138, -0.00000526, -0.00002585, 0.00000724, -0.00005254, 0.00002957, -0.00006190, 0.00008322, -0.00012892, 0.00007544, -0.00005749, 0.00013070, 0.00002116, 0.00004875, -0.00004364, -0.00002902, 0.00001648, },
		{ -0.00000384, -0.00001423, 0.00003731, -0.00000551, -0.00001574, -0.00001110, 0.00001287, -0.00001124, 0.00000980, -0.00002196, 0.00003505, -0.00000477, -0.00003806, -0.00000562, 0.00003920, -0.00000958, -0.00003173, 0.00001620, -0.00006182, 0.00000458, -0.00023709, -0.00011452, 0.00113268, -0.00046159, -0.00002951, -0.00008614, -0.00001001, -0.00002220, 0.00003092, -0.00002984, -0.00005166, -0.00000315, 0.00003205, -0.00000610, -0.00003865, 0.00000815, 0.00000974, 0.00000524, -0.00003055, 0.00001283, 0.00000951, 0.00000863, -0.00002287, -0.00001094, 0.00002456, 0.00002174, -0.00004744, 0.00004427, -0.00006036, 0.00003876, -0.00007528, 0.00003685, -0.00006129, 0.00005249, -0.00004933, -0.00003073, 0.00002085, 0.00001669, 0.00013279, -0.00000746, 0.00003928, -0.00009213, 0.00004007, -0.00002856, },
		{ -0.00002327, 0.00003113, 0.00000548, -0.00000159, -0.00003534, 0.00002877, -0.00002610, 0.00002868, -0.00003836, 0.00003663, -0.00000481, -0.00000186, -0.00006143, 0.00006699, -0.00002230, 0.00001241, -0.00005419, 0.00004305, -0.00011985, 0.00004273, -0.00049046, 0.00086004, 0.00027634, -0.00044381, 0.00011899, -0.00016739, 0.00007784, -0.00006874, 0.00007572, -0.00011555, 0.00002204, -0.00001835, 0.00005645, -0.00007168, 0.00001901, -0.00001512, 0.00003867, -0.00004497, 0.00001098, -0.00000057, 0.00002614, -0.00002469, -0.00000868, -0.00000496, 0.00005408, -0.00005154, 0.00002722, -0.00002727, 0.00000104, -0.00002087, -0.00003454, 0.00001611, -0.00004464, 0.00006011, -0.00011460, 0.00005871, -0.00004680, 0.00015513, 0.00002212, 0.00005061, -0.00007336, -0.00000157, -0.00000319, -0.00001027, },
		{ -0.00000490, 0.00003453, -0.00000720, -0.00002279, -0.00000006, 0.00000302, 0.00000055, -0.00000351, -0.00000698, 0.00002722, -0.00000592, -0.00004589, 0.00001283, 0.00002893, -0.00000584, -0.00003771, 0.00002445, -0.00008183, 0.00003534, -0.00032186, 0.00013175, 0.00091598, -0.00046767, 0.00002166, -0.00011026, 0.00001662, -0.00004182, 0.00005182, -0.00005579, -0.00003742, -0.00000586, 0.00004330, -0.00002877, -0.00002268, -0.00000017, 0.00002350, -0.00000930, -0.00002393, 0.00001206, 0.00001493, 0.00000084, -0.00002316, -0.00000969, 0.00003929, -0.00000355, -0.00002808, 0.00002639, -0.00004617, 0.00003290, -0.00008305, 0.00004647, -0.00006705, 0.00005933, -0.00006868, -0.00001231, -0.00000921, 0.00006562, 0.00012294, 0.00001518, -0.00000373, -0.00006907, 0.00003071, -0.00002957, -0.00000615, },
		{ 0.00003712, 0.00000192, -0.00000934, -0.00002746, 0.00002400, -0.00001965, 0.00002226, -0.00003306, 0.00003781, -0.00000674, -0.00001385, -0.00004498, 0.00006167, -0.00001632, -0.00000364, -0.00002979, 0.00000358, -0.00006637, -0.00005036, -0.00034795, 0.00088052, 0.00007193, -0.00031340, 0.00005947, -0.00011527, 0.00004080, -0.00003442, 0.00004533, -0.00010150, 0.00001255, 0.00000366, 0.00003124, -0.00006010, 0.00001270, -0.00000278, 0.00003063, -0.00004583, 0.00001254, 0.00000487, 0.00002128, -0.00002639, -0.00001246, 0.00001220, 0.00003969, -0.00005233, 0.00003254, -0.00003927, 0.00002234, -0.00004959, -0.00001218, -0.00000058, -0.00002460, 0.00003193, -0.00009302, 0.00003060, -0.00002117, 0.00017153, 0.00003333, 0.00003061, -0.00008442, 0.00001585, -0.00001937, -0.00000508, -0.00002271, },
		{ 0.00003030, -0.00001071, -0.00002586, 0.00000736, -0.00000268, 0.00000776, -0.00001280, 0.00000576, 0.00001940, -0.00000949, -0.00004724, 0.00002568, 0.00002276, -0.00000910, -0.00003259, 0.00001582, -0.00007613, 0.00002753, -0.00034597, 0.00032648, 0.00068568, -0.00042226, 0.00003493, -0.00010872, 0.00002203, -0.00004351, 0.00005671, -0.00007017, -0.00003056, 0.00000117, 0.00004089, -0.00004012, -0.00001329, -0.00000236, 0.00003219, -0.00002240, -0.00001706, 0.00001261, 0.00001773, -0.00000565, -0.00002498, -0.00000241, 0.00004533, -0.00002401, -0.00000938, 0.00000565, -0.00002319, 0.00001399, -0.00007813, 0.00004682, -0.00006373, 0.00005492, -0.00007593, -0.00000715, -0.00002797, 0.00011135, 0.00012156, 0.00002271, -0.00004321, -0.00004009, 0.00001327, -0.00002424, -0.00001078, 0.00000595, },
		{ -0.00000255, -0.00001623, -0.00001863, 0.00001827, -0.00001186, 0.00001328, -0.00002396, 0.00003522, -0.00000874, -0.00002343, -0.00002841, 0.00005601, -0.00001390, -0.00001331, -0.00001426, -0.00002464, -0.00002461, -0.00013567, -0.00017658, 0.00081932, -0.00005952, -0.00021322, 0.00001443, -0.00007430, 0.00001198, -0.00000629, 0.00001753, -0.00008914, 0.00000895, 0.00001853, 0.00000796, -0.00004651, 0.00000538, 0.00001152, 0.00001838, -0.00004370, 0.00001416, 0.00000952, 0.00001590, -0.00002872, -0.00001277, 0.00002789, 0.00002009, -0.00004442, 0.00002754, -0.00003848, 0.00003220, -0.00007206, 0.00000855, -0.00001537, -0.00000739, 0.00000498, -0.00007085, -0.00000265, 0.00001461, 0.00018629, 0.00004543, -0.00000528, -0.00007654, 0.00001982, -0.00002804, -0.00000284, -0.00001940, 0.00004108, },
		{ -0.00001531, -0.00002541, 0.00001144, -0.00000439, 0.00000993, -0.00001697, 0.00001478, 0.00001226, -0.00001442, -0.00004357, 0.00003419, 0.00001833, -0.00001519, -0.00002223, -0.00000197, -0.00005369, -0.00001017, -0.00031063, 0.00045122, 0.00047456, -0.00035240, 0.00002636, -0.00009264, 0.00001506, -0.00003428, 0.00005082, -0.00007686, -0.00002672, 0.00001258, 0.00002887, -0.00004144, -0.00001045, 0.00000198, 0.00003447, -0.00003267, -0.00001016, 0.00001380, 0.00001855, -0.00001150, -0.00002676, 0.00000921, 0.00004220, -0.00003645, 0.00000377, -0.00001135, 0.00000146, -0.00001271, -0.00006327, 0.00003958, -0.00005323, 0.00004101, -0.00007204, -0.00001562, -0.00003392, 0.00015404, 0.00012834, 0.00000934, -0.00006880, -0.00001444, -0.00000591, -0.00001589, -0.00001620, 0.00001746, 0.00002453, },
		{ -0.00002124, -0.00001034, 0.00001346, -0.00000526, 0.00000449, -0.00001350, 0.00003013, -0.00001111, -0.00002969, -0.00001327, 0.00005086, -0.00001558, -0.00001538, -0.00001018, -0.00003605, -0.00000538, -0.00019242, -0.00000873, 0.00070506, -0.00012783, -0.00014524, -0.00001297, -0.00004737, -0.00000558, 0.00001234, -0.00000469, -0.00008006, 0.00001242, 0.00002306, -0.00000837, -0.00003539, 0.00000067, 0.00002450, 0.00000393, -0.00003893, 0.00001570, 0.00001303, 0.00001041, -0.00003179, -0.00000846, 0.00003858, 0.00000000, -0.00003204, 0.00001561, -0.00002620, 0.00002855, -0.00008404, 0.00002342, -0.00002458, 0.00000287, -0.00001569, -0.00005396, -0.00003519, 0.00005634, 0.00020532, 0.00004700, -0.00004600, -0.00005522, 0.00001159, -0.00002821, -0.00000481, -0.00001263, 0.00004235, -0.00000782, },
		{ -0.00002242, 0.00001328, -0.00000345, 0.00000807, -0.00001634, 0.00001963, 0.00000609, -0.00001957, -0.00003663, 0.00003982, 0.00001332, -0.00001992, -0.00001295, -0.00001964, -0.00002719, -0.00006158, -0.00023007, 0.00050378, 0.00030280, -0.00027928, 0.00000999, -0.00007229, 0.00000431, -0.00002150, 0.00003999, -0.00008016, -0.00002130, 0.00002214, 0.00001332, -0.00003651, -0.00001192, 0.00001142, 0.00003029, -0.00003919, -0.00000356, 0.00001507, 0.00001803, -0.00001740, -0.00002668, 0.00002227, 0.00003164, -0.00004029, 0.00000887, -0.00001968, 0.00002077, -0.00004047, -0.00004335, 0.00002850, -0.00003959, 0.00002184, -0.00006091, -0.00003490, -0.00002866, 0.00019782, 0.00013663, -0.00002359, -0.00007512, 0.00000080, -0.00002075, -0.00000845, -0.00002025, 0.00002831, 0.00001737, -0.00002010, },
		{ -0.00000337, 0.00001053, -0.00000149, -0.00000189, -0.00000397, 0.00002392, -0.00001394, -0.00003245, -0.00000008, 0.00004577, -0.00001981, -0.00001186, -0.00001539, -0.00003174, -0.00001092, -0.00020933, 0.00012930, 0.00056823, -0.00015045, -0.00010351, -0.00002590, -0.00003197, -0.00001356, 0.00002226, -0.00002171, -0.00007273, 0.00002062, 0.00001796, -0.00001592, -0.00002924, 0.00000104, 0.00003322, -0.00001046, -0.00003218, 0.00001697, 0.00001545, 0.00000473, -0.00003490, 0.00000023, 0.00004213, -0.00001636, -0.00002002, 0.00000214, -0.00000712, 0.00001334, -0.00008482, 0.00003104, -0.00002726, 0.00000487, -0.00002788, -0.00004560, -0.00006361, 0.00010303, 0.00023055, 0.00002881, -0.00007901, -0.00002961, -0.00000377, -0.00002190, -0.00001035, -0.00000270, 0.00004068, -0.00001361, -0.00002391, },
		{ 0.00001426, -0.00000185, 0.00000424, -0.00001245, 0.00002082, 0.00000083, -0.00002398, -0.00002794, 0.00004358, 0.00000642, -0.00002066, -0.00000919, -0.00002966, -0.00000862, -0.00010830, -0.00012658, 0.00049573, 0.00017707, -0.00021530, -0.00000537, -0.00005415, -0.00000454, -0.00001032, 0.00002848, -0.00008283, -0.00001150, 0.00002502, 0.00000023, -0.00003011, -0.00001420, 0.00002312, 0.00002085, -0.00004162, 0.00000239, 0.00001607, 0.00001660, -0.00002370, -0.00002322, 0.00003343, 0.00001709, -0.00003747, 0.00000652, -0.00001762, 0.00003002, -0.00006339, -0.00002330, 0.00001753, -0.00002711, 0.00000225, -0.00004749, -0.00006089, -0.00001479, 0.00024843, 0.00013518, -0.00006756, -0.00006340, 0.00000293, -0.00002754, -0.00000492, -0.00002104, 0.00003713, 0.00000917, -0.00002423, -0.00001803, },
		{ 0.00000930, -0.00000090, -0.00000478, 0.00000304, 0.00001772, -0.00001710, -0.00003211, 0.00001130, 0.00003964, -0.00002392, -0.00000671, -0.00002436, -0.00001833, -0.00003492, -0.00018670, 0.00022240, 0.00043418, -0.00014605, -0.00007888, -0.00003039, -0.00002299, -0.00001628, 0.00002669, -0.00003592, -0.00006388, 0.00002882, 0.00000717, -0.00001609, -0.00002806, 0.00000710, 0.00003588, -0.00002273, -0.00002429, 0.00001779, 0.00001697, -0.00000144, -0.00003677, 0.00001167, 0.00003837, -0.00002648, -0.00001222, -0.00000743, 0.00001237, -0.00000857, -0.00007680, 0.00003274, -0.00002508, 0.00000027, -0.00003260, -0.00004580, -0.00008767, 0.00015792, 0.00025748, -0.00001229, -0.00009446, -0.00000925, -0.00001942, -0.00001314, -0.00001732, 0.00000910, 0.00003625, -0.00001948, -0.00002430, 0.00000228, },
		{ -0.00000145, 0.00000072, -0.00000744, 0.00001960, -0.00000388, -0.00002691, -0.00001855, 0.00004569, -0.00000214, -0.00001715, -0.00001207, -0.00002877, -0.00000540, -0.00013529, -0.00002391, 0.00044695, 0.00009301, -0.00016429, -0.00001684, -0.00003997, -0.00000989, -0.00000254, 0.00001777, -0.00008504, 0.00000245, 0.00001944, -0.00000640, -0.00002627, -0.00001376, 0.00003361, 0.00000833, -0.00004029, 0.00000741, 0.00001668, 0.00001442, -0.00003018, -0.00001572, 0.00003980, 0.00000264, -0.00003170, 0.00000008, -0.00000708, 0.00002806, -0.00007812, -0.00000633, 0.00000905, -0.00001840, -0.00001448, -0.00003548, -0.00009066, 0.00000698, 0.00030954, 0.00011198, -0.00010930, -0.00004086, -0.00000568, -0.00002601, -0.00000629, -0.00001774, 0.00004301, 0.00000026, -0.00002700, -0.00001320, 0.00001551, },
		{ -0.00000253, -0.00000451, 0.00000704, 0.00001217, -0.00002026, -0.00002936, 0.00002115, 0.00003145, -0.00002540, -0.00000403, -0.00003060, -0.00000516, -0.00006532, -0.00013510, 0.00026820, 0.00031879, -0.00012947, -0.00006331, -0.00003189, -0.00001562, -0.00001804, 0.00002908, -0.00004974, -0.00005045, 0.00003190, -0.00000371, -0.00001266, -0.00002957, 0.00001744, 0.00003215, -0.00003142, -0.00001613, 0.00001808, 0.00001781, -0.00000839, -0.00003598, 0.00002328, 0.00002908, -0.00003020, -0.00001018, -0.00000953, 0.00002648, -0.00003137, -0.00006392, 0.00003128, -0.00002114, -0.00000759, -0.00003291, -0.00005241, -0.00010920, 0.00022759, 0.00027507, -0.00007120, -0.00008862, -0.00000067, -0.00002934, -0.00000624, -0.00002302, 0.00002101, 0.00002956, -0.00002492, -0.00002288, 0.00000720, 0.00000866, },
		},
		{ { -0.00001125, -0.00000619, 0.00000702, 0.00000015, -0.00000294, 0.00000249, -0.00000940, 0.00000685, -0.00000972, -0.00002034, 0.00002841, -0.00001603, -0.00002666, -0.00002793, -0.00014625, 0.00006467, 0.00033770, 0.00006468, -0.00014624, -0.00002791, -0.00002664, -0.00001600, 0.00002844, -0.00002030, -0.00000968, 0.00000689, -0.00000936, 0.00000253, -0.00000289, 0.00000020, 0.00000707, -0.00000613, -0.00001119, -0.00000613, 0.00000707, 0.00000020, -0.00000289, 0.00000253, -0.00000936, 0.00000689, -0.00000968, -0.00002030, 0.00002844, -0.00001600, -0.00002664, -0.00002791, -0.00014624, 0.00006468, 0.00033770, 0.00006467, -0.00014625, -0.00002793, -0.00002666, -0.00001603, 0.00002841, -0.00002034, -0.00000972, 0.00000685, -0.00000940, 0.00000249, -0.00000294, 0.00000015, 0.00000702, -0.00000619, },
		{ -0.00000876, -0.00000118, 0.00000918, -0.00000633, 0.00000368, -0.00000729, -0.00000063, 0.00000525, -0.00002365, 0.00000320, 0.00002215, -0.00004066, -0.00000278, -0.00010072, -0.00007796, 0.00023360, 0.00023360, -0.00007795, -0.00010071, -0.00000276, -0.00004064, 0.00002218, 0.00000324, -0.00002361, 0.00000529, -0.00000059, -0.00000724, 0.00000374, -0.00000628, 0.00000923, -0.00000113, -0.00000871, -0.00001164, 0.00000320, 0.00000392, -0.00000335, 0.00000112, -0.00000147, -0.00000696, 0.00000807, -0.00002833, 0.00001124, 0.00001330, -0.00003450, -0.00001326, -0.00008733, -0.00012112, 0.00027527, 0.00027527, -0.00012113, -0.00008734, -0.00001328, -0.00003453, 0.00001327, 0.00001120, -0.00002837, 0.00000803, -0.00000700, -0.00000152, 0.00000107, -0.00000341, 0.00000387, 0.00000315, -0.00001169, },
		{ -0.00000721, 0.00000745, 0.00000184, -0.00000320, 0.00000051, -0.00000912, 0.00000904, -0.00000885, -0.00002165, 0.00002769, -0.00001337, -0.00002555, -0.00002847, -0.00013630, 0.00008685, 0.00026240, 0.00008686, -0.00013629, -0.00002845, -0.00002553, -0.00001335, 0.00002772, -0.00002161, -0.00000881, 0.00000908, -0.00000907, 0.00000056, -0.00000315, 0.00000189, 0.00000751, -0.00000716, -0.00001119, -0.00000501, 0.00000652, -0.00000154, -0.00000263, 0.00000460, -0.00000925, 0.00000388, -0.00001054, -0.00001777, 0.00002828, -0.00001898, -0.00002775, -0.00002680, -0.00015325, 0.00003233, 0.00042707, 0.00003233, -0.00015326, -0.00002682, -0.00002777, -0.00001901, 0.00002825, -0.00001780, -0.00001058, 0.00000384, -0.00000930, 0.00000455, -0.00000268, -0.00000159, 0.00000646, -0.00000507, -0.00001125, },
		{ -0.00000529, 0.00001402, -0.00000878, 0.00000567, -0.00001195, 0.00000470, 0.00000267, -0.00001918, -0.00000293, 0.00002798, -0.00004402, 0.00000405, -0.00010684, -0.00004415, 0.00019651, 0.00019652, -0.00004414, -0.00010683, 0.00000407, -0.00004400, 0.00002801, -0.00000289, -0.00001915, 0.00000272, 0.00000475, -0.00001190, 0.00000572, -0.00000873, 0.00001407, -0.00000524, -0.00000580, -0.00001456, 0.00000772, -0.00000182, 0.00000002, -0.00000209, 0.00000535, -0.00001430, 0.00001103, -0.00003327, 0.00002102, 0.00000146, -0.00002567, -0.00002732, -0.00006681, -0.00017323, 0.00032112, 0.00032112, -0.00017324, -0.00006682, -0.00002734, -0.00002569, 0.00000143, 0.00002099, -0.00003331, 0.00001099, -0.00001434, 0.00000530, -0.00000214, -0.00000003, -0.00000187, 0.00000766, -0.00001462, -0.00000586, },
		{ 0.00000777, 0.00000346, -0.00000346, -0.00000136, -0.00000845, 0.00001040, -0.00000799, -0.00002175, 0.00002612, -0.00001106, -0.00002445, -0.00002845, -0.00012354, 0.00009880, 0.00020169, 0.00009881, -0.00012353, -0.00002843, -0.00002443, -0.00001103, 0.00002615, -0.00002171, -0.00000795, 0.00001044, -0.00000841, -0.00000131, -0.00000341, 0.00000352, 0.00000782, -0.00000807, -0.00001119, -0.00000380, 0.00000585, -0.00000332, -0.00000237, 0.00000674, -0.00000876, 0.00000009, -0.00001140, -0.00001405, 0.00002725, -0.00002225, -0.00002885, -0.00002515, -0.00015728, -0.00000972, 0.00052945, -0.00000972, -0.00015729, -0.00002516, -0.00002887, -0.00002228, 0.00002722, -0.00001409, -0.00001144, 0.00000005, -0.00000880, 0.00000669, -0.00000242, -0.00000337, 0.00000580, -0.00000386, -0.00001125, -0.00000813, },
		{ 0.00001834, -0.00001074, 0.00000704, -0.00001549, 0.00000897, 0.00000033, -0.00001502, -0.00000718, 0.00003079, -0.00004467, 0.00000726, -0.00010585, -0.00001974, 0.00016424, 0.00016424, -0.00001974, -0.00010583, 0.00000728, -0.00004465, 0.00003082, -0.00000715, -0.00001498, 0.00000037, 0.00000902, -0.00001544, 0.00000709, -0.00001069, 0.00001839, -0.00000910, -0.00000295, -0.00001746, 0.00001236, -0.00000791, 0.00000379, -0.00000585, 0.00001312, -0.00002252, 0.00001414, -0.00003837, 0.00003246, -0.00001316, -0.00001429, -0.00004471, -0.00003956, -0.00023351, 0.00037059, 0.00037059, -0.00023351, -0.00003958, -0.00004473, -0.00001431, -0.00001318, 0.00003242, -0.00003840, 0.00001410, -0.00002256, 0.00001308, -0.00000590, 0.00000374, -0.00000797, 0.00001231, -0.00001751, -0.00000300, -0.00000915, },
		{ 0.00000501, -0.00000371, -0.00000310, -0.00000742, 0.00001094, -0.00000714, -0.00002069, 0.00002374, -0.00000909, -0.00002338, -0.00002789, -0.00010821, 0.00010084, 0.00015557, 0.00010084, -0.00010820, -0.00002787, -0.00002335, -0.00000907, 0.00002377, -0.00002065, -0.00000711, 0.00001099, -0.00000738, -0.00000305, -0.00000366, 0.00000506, 0.00000802, -0.00000888, -0.00001119, -0.00000253, 0.00000509, -0.00000511, -0.00000212, 0.00000893, -0.00000790, -0.00000441, -0.00001225, -0.00000925, 0.00002541, -0.00002576, -0.00002993, -0.00002299, -0.00015843, -0.00006063, 0.00064326, -0.00006064, -0.00015844, -0.00002300, -0.00002995, -0.00002579, 0.00002538, -0.00000929, -0.00001229, -0.00000445, -0.00000794, 0.00000888, -0.00000217, -0.00000516, 0.00000504, -0.00000258, -0.00001125, -0.00000893, 0.00000797, },
		{ -0.00001222, 0.00000780, -0.00001793, 0.00001220, -0.00000177, -0.00001117, -0.00000963, 0.00003072, -0.00004275, 0.00000701, -0.00009815, -0.00000447, 0.00013685, 0.00013685, -0.00000447, -0.00009814, 0.00000703, -0.00004273, 0.00003075, -0.00000960, -0.00001114, -0.00000174, 0.00001225, -0.00001788, 0.00000784, -0.00001217, 0.00002219, -0.00001267, -0.00000017, -0.00002030, 0.00001708, -0.00001429, 0.00000791, -0.00001010, 0.00002172, -0.00003148, 0.00001736, -0.00004356, 0.00004532, -0.00003022, -0.00000064, -0.00006505, -0.00000626, -0.00030086, 0.00042295, 0.00042295, -0.00030087, -0.00000627, -0.00006507, -0.00000066, -0.00003025, 0.00004529, -0.00004360, 0.00001732, -0.00003152, 0.00002168, -0.00001015, 0.00000786, -0.00001434, 0.00001703, -0.00002035, -0.00000023, -0.00001272, 0.00002214, },
		{ -0.00000396, -0.00000471, -0.00000607, 0.00001074, -0.00000632, -0.00001855, 0.00002063, -0.00000747, -0.00002233, -0.00002683, -0.00009068, 0.00009365, 0.00012350, 0.00009365, -0.00009067, -0.00002682, -0.00002231, -0.00000745, 0.00002066, -0.00001852, -0.00000629, 0.00001078, -0.00000603, -0.00000467, -0.00000391, 0.00000651, 0.00000811, -0.00000957, -0.00001120, -0.00000120, 0.00000425, -0.00000690, -0.00000187, 0.00001113, -0.00000671, -0.00000952, -0.00001307, -0.00000349, 0.00002283, -0.00002946, -0.00003097, -0.00002038, -0.00015688, -0.00011922, 0.00076641, -0.00011923, -0.00015689, -0.00002039, -0.00003099, -0.00002948, 0.00002280, -0.00000352, -0.00001310, -0.00000956, -0.00000676, 0.00001108, -0.00000192, -0.00000695, 0.00000420, -0.00000125, -0.00001125, -0.00000962, 0.00000806, 0.00000646, },
		{ 0.00000799, -0.00001934, 0.00001444, -0.00000364, -0.00000767, -0.00001039, 0.00002802, -0.00003850, 0.00000360, -0.00008442, 0.00000226, 0.00011424, 0.00011424, 0.00000227, -0.00008441, 0.00000362, -0.00003848, 0.00002805, -0.00001036, -0.00000764, -0.00000360, 0.00001448, -0.00001930, 0.00000804, -0.00001319, 0.00002546, -0.00001595, 0.00000250, -0.00002304, 0.00002181, -0.00002083, 0.00001229, -0.00001473, 0.00003097, -0.00004101, 0.00002063, -0.00004879, 0.00005934, -0.00004933, 0.00001494, -0.00008786, 0.00003219, -0.00037389, 0.00047734, 0.00047734, -0.00037389, 0.00003218, -0.00008787, 0.00001492, -0.00004935, 0.00005931, -0.00004882, 0.00002059, -0.00004105, 0.00003092, -0.00001478, 0.00001224, -0.00002088, 0.00002176, -0.00002309, 0.00000244, -0.00001600, 0.00002541, -0.00001324, },
		{ -0.00000618, -0.00000445, 0.00000985, -0.00000554, -0.00001548, 0.00001692, -0.00000619, -0.00002132, -0.00002535, -0.00007138, 0.00007824, 0.00010447, 0.00007825, -0.00007137, -0.00002533, -0.00002131, -0.00000617, 0.00001695, -0.00001545, -0.00000550, 0.00000989, -0.00000441, -0.00000614, -0.00000415, 0.00000786, 0.00000810, -0.00001016, -0.00001120, 0.00000016, 0.00000334, -0.00000866, -0.00000164, 0.00001332, -0.00000525, -0.00001511, -0.00001385, 0.00000306, 0.00001962, -0.00003328, -0.00003198, -0.00001740, -0.00015293, -0.00018397, 0.00089643, -0.00018398, -0.00015294, -0.00001741, -0.00003200, -0.00003330, 0.00001960, 0.00000303, -0.00001389, -0.00001515, -0.00000529, 0.00001328, -0.00000168, -0.00000871, 0.00000329, 0.00000011, -0.00001125, -0.00001020, 0.00000805, 0.00000781, -0.00000419, },
		{ -0.00001982, 0.00001577, -0.00000525, -0.00000452, -0.00000965, 0.00002303, -0.00003225, -0.00000255, -0.00006555, 0.00000136, 0.00009617, 0.00009617, 0.00000137, -0.00006554, -0.00000253, -0.00003223, 0.00002306, -0.00000962, -0.00000448, -0.00000522, 0.00001581, -0.00001978, 0.00000773, -0.00001380, 0.00002822, -0.00001891, 0.00000503, -0.00002567, 0.00002649, -0.00002744, 0.00001684, -0.00001964, 0.00004066, -0.00005092, 0.00002391, -0.00005397, 0.00007418, -0.00006998, 0.00003200, -0.00011252, 0.00007467, -0.00045093, 0.00053277, 0.00053277, -0.00045094, 0.00007466, -0.00011254, 0.00003198, -0.00007001, 0.00007416, -0.00005400, 0.00002388, -0.00005096, 0.00004062, -0.00001969, 0.00001680, -0.00002749, 0.00002644, -0.00002572, 0.00000499, -0.00001896, 0.00002817, -0.00001385, 0.00000769, },
		{ -0.00000261, 0.00000839, -0.00000479, -0.00001165, 0.00001274, -0.00000521, -0.00002037, -0.00002351, -0.00005083, 0.00005593, 0.00009708, 0.00005594, -0.00005082, -0.00002349, -0.00002035, -0.00000519, 0.00001276, -0.00001162, -0.00000476, 0.00000842, -0.00000257, -0.00000746, -0.00000437, 0.00000910, 0.00000800, -0.00001063, -0.00001120, 0.00000152, 0.00000239, -0.00001037, -0.00000141, 0.00001547, -0.00000356, -0.00002104, -0.00001460, 0.00001020, 0.00001591, -0.00003714, -0.00003293, -0.00001414, -0.00014695, -0.00025311, 0.00103045, -0.00025311, -0.00014696, -0.00001415, -0.00003295, -0.00003717, 0.00001589, 0.00001017, -0.00001463, -0.00002108, -0.00000360, 0.00001543, -0.00000146, -0.00001042, 0.00000234, 0.00000148, -0.00001124, -0.00001068, 0.00000795, 0.00000905, -0.00000441, -0.00000750, },
		{ 0.00001631, -0.00000664, -0.00000171, -0.00000764, 0.00001617, -0.00002439, -0.00001094, -0.00004259, -0.00000605, 0.00008228, 0.00008228, -0.00000604, -0.00004257, -0.00001092, -0.00002437, 0.00001620, -0.00000761, -0.00000168, -0.00000661, 0.00001634, -0.00001946, 0.00000701, -0.00001404, 0.00003050, -0.00002156, 0.00000742, -0.00002816, 0.00003105, -0.00003400, 0.00002146, -0.00002472, 0.00005057, -0.00006098, 0.00002715, -0.00005902, 0.00008950, -0.00009161, 0.00005005, -0.00013838, 0.00011989, -0.00053013, 0.00058817, 0.00058817, -0.00053014, 0.00011988, -0.00013839, 0.00005003, -0.00009164, 0.00008947, -0.00005905, 0.00002712, -0.00006101, 0.00005053, -0.00002476, 0.00002142, -0.00003404, 0.00003100, -0.00002821, 0.00000737, -0.00002161, 0.00003045, -0.00001409, 0.00000697, -0.00001950, },
		{ 0.00000647, -0.00000409, -0.00000724, 0.00000823, -0.00000452, -0.00001948, -0.00002141, -0.00002962, 0.00002824, 0.00009957, 0.00002825, -0.00002961, -0.00002139, -0.00001946, -0.00000449, 0.00000825, -0.00000721, -0.00000406, 0.00000650, -0.00000059, -0.00000864, -0.00000458, 0.00001022, 0.00000783, -0.00001102, -0.00001120, 0.00000287, 0.00000142, -0.00001201, -0.00000120, 0.00001754, -0.00000172, -0.00002714, -0.00001530, 0.00001770, 0.00001185, -0.00004098, -0.00003382, -0.00001071, -0.00013938, -0.00032465, 0.00116536, -0.00032465, -0.00013939, -0.00001073, -0.00003384, -0.00004100, 0.00001183, 0.00001767, -0.00001533, -0.00002718, -0.00000176, 0.00001751, -0.00000124, -0.00001205, 0.00000137, 0.00000282, -0.00001124, -0.00001106, 0.00000778, 0.00001018, -0.00000462, -0.00000868, -0.00000062, },
		{ -0.00000781, 0.00000074, -0.00000462, 0.00000791, -0.00001538, -0.00002100, -0.00001674, -0.00001865, 0.00007211, 0.00007211, -0.00001864, -0.00001673, -0.00002098, -0.00001536, 0.00000794, -0.00000460, 0.00000077, -0.00000778, 0.00001621, -0.00001849, 0.00000595, -0.00001398, 0.00003233, -0.00002390, 0.00000963, -0.00003048, 0.00003541, -0.00004038, 0.00002605, -0.00002981, 0.00006046, -0.00007096, 0.00003028, -0.00006387, 0.00010487, -0.00011360, 0.00006854, -0.00016467, 0.00016642, -0.00060947, 0.00064242, 0.00064242, -0.00060948, 0.00016641, -0.00016468, 0.00006852, -0.00011362, 0.00010485, -0.00006390, 0.00003025, -0.00007099, 0.00006043, -0.00002985, 0.00002601, -0.00004042, 0.00003536, -0.00003053, 0.00000959, -0.00002394, 0.00003229, -0.00001402, 0.00000591, -0.00001852, 0.00001617, },
		{ -0.00000345, -0.00000247, 0.00000357, -0.00000405, -0.00001866, -0.00001914, -0.00000837, -0.00000313, 0.00010993, -0.00000313, -0.00000836, -0.00001913, -0.00001864, -0.00000404, 0.00000359, -0.00000245, -0.00000342, 0.00000426, 0.00000146, -0.00000968, -0.00000478, 0.00001123, 0.00000760, -0.00001131, -0.00001120, 0.00000416, 0.00000045, -0.00001355, -0.00000101, 0.00001951, 0.00000020, -0.00003324, -0.00001595, 0.00002532, 0.00000760, -0.00004470, -0.00003465, -0.00000723, -0.00013071, -0.00039647, 0.00129788, -0.00039647, -0.00013072, -0.00000724, -0.00003466, -0.00004472, 0.00000758, 0.00002529, -0.00001597, -0.00003327, 0.00000017, 0.00001947, -0.00000105, -0.00001359, 0.00000042, 0.00000412, -0.00001124, -0.00001135, 0.00000756, 0.00001119, -0.00000481, -0.00000971, 0.00000143, 0.00000423, },
		{ 0.00000287, -0.00000090, -0.00000123, -0.00000569, -0.00003211, 0.00001069, -0.00003500, 0.00006511, 0.00006511, -0.00003499, 0.00001070, -0.00003209, -0.00000568, -0.00000121, -0.00000088, 0.00000289, -0.00000874, 0.00001555, -0.00001703, 0.00000466, -0.00001368, 0.00003377, -0.00002594, 0.00001165, -0.00003261, 0.00003950, -0.00004645, 0.00003049, -0.00003480, 0.00007008, -0.00008062, 0.00003324, -0.00006843, 0.00011990, -0.00013530, 0.00008690, -0.00019062, 0.00021278, -0.00068687, 0.00069436, 0.00069436, -0.00068688, 0.00021277, -0.00019064, 0.00008688, -0.00013532, 0.00011988, -0.00006845, 0.00003322, -0.00008065, 0.00007005, -0.00003483, 0.00003045, -0.00004649, 0.00003946, -0.00003265, 0.00001161, -0.00002597, 0.00003374, -0.00001371, 0.00000463, -0.00001706, 0.00001552, -0.00000877, },
		{ 0.00000243, -0.00000107, -0.00000379, -0.00001791, -0.00001682, 0.00001230, -0.00003640, 0.00012602, -0.00003639, 0.00001231, -0.00001681, -0.00001789, -0.00000377, -0.00000105, 0.00000245, -0.00000283, 0.00000183, 0.00000351, -0.00001057, -0.00000495, 0.00001212, 0.00000734, -0.00001154, -0.00001120, 0.00000538, -0.00000047, -0.00001497, -0.00000084, 0.00002133, 0.00000213, -0.00003916, -0.00001653, 0.00003280, 0.00000333, -0.00004823, -0.00003539, -0.00000380, -0.00012145, -0.00046639, 0.00142469, -0.00046640, -0.00012145, -0.00000381, -0.00003541, -0.00004825, 0.00000331, 0.00003278, -0.00001656, -0.00003919, 0.00000210, 0.00002130, -0.00000087, -0.00001500, -0.00000051, 0.00000534, -0.00001124, -0.00001157, 0.00000730, 0.00001208, -0.00000499, -0.00001060, 0.00000348, 0.00000181, -0.00000286, },
		{ 0.00000322, -0.00001071, 0.00000417, -0.00004362, 0.00003839, -0.00005358, 0.00006071, 0.00006071, -0.00005358, 0.00003840, -0.00004361, 0.00000418, -0.00001070, 0.00000324, 0.00000469, -0.00000953, 0.00001453, -0.00001526, 0.00000324, -0.00001321, 0.00003487, -0.00002768, 0.00001345, -0.00003453, 0.00004326, -0.00005210, 0.00003467, -0.00003953, 0.00007917, -0.00008972, 0.00003599, -0.00007262, 0.00013417, -0.00015606, 0.00010453, -0.00021546, 0.00025746, -0.00076023, 0.00074283, 0.00074283, -0.00076023, 0.00025745, -0.00021547, 0.00010452, -0.00015608, 0.00013415, -0.00007264, 0.00003596, -0.00008975, 0.00007914, -0.00003956, 0.00003464, -0.00005213, 0.00004322, -0.00003456, 0.00001342, -0.00002771, 0.00003484, -0.00001324, 0.00000321, -0.00001529, 0.00001451, -0.00000955, 0.00000467, },
		{ -0.00000552, -0.00000368, -0.00001725, -0.00001455, 0.00003175, -0.00006975, 0.00014561, -0.00006975, 0.00003176, -0.00001454, -0.00001723, -0.00000366, -0.00000550, 0.00000725, -0.00000232, -0.00000063, 0.00000547, -0.00001133, -0.00000511, 0.00001288, 0.00000706, -0.00001170, -0.00001121, 0.00000650, -0.00000134, -0.00001625, -0.00000069, 0.00002299, 0.00000399, -0.00004472, -0.00001705, 0.00003990, -0.00000079, -0.00005148, -0.00003605, -0.00000054, -0.00011211, -0.00053227, 0.00154249, -0.00053227, -0.00011212, -0.00000055, -0.00003607, -0.00005150, -0.00000080, 0.00003988, -0.00001707, -0.00004475, 0.00000396, 0.00002296, -0.00000071, -0.00001628, -0.00000137, 0.00000647, -0.00001124, -0.00001173, 0.00000703, 0.00001285, -0.00000514, -0.00001136, 0.00000545, -0.00000065, -0.00000234, 0.00000723, },
		{ -0.00002003, 0.00001372, -0.00005491, 0.00006505, -0.00007290, 0.00005832, 0.00005832, -0.00007290, 0.00006506, -0.00005490, 0.00001373, -0.00002001, 0.00000746, 0.00000619, -0.00001015, 0.00001331, -0.00001335, 0.00000179, -0.00001264, 0.00003568, -0.00002913, 0.00001503, -0.00003621, 0.00004662, -0.00005719, 0.00003848, -0.00004389, 0.00008749, -0.00009803, 0.00003846, -0.00007637, 0.00014727, -0.00017524, 0.00012089, -0.00023842, 0.00029898, -0.00082752, 0.00078674, 0.00078674, -0.00082752, 0.00029897, -0.00023843, 0.00012088, -0.00017526, 0.00014726, -0.00007639, 0.00003843, -0.00009805, 0.00008747, -0.00004391, 0.00003846, -0.00005722, 0.00004659, -0.00003624, 0.00001500, -0.00002916, 0.00003565, -0.00001267, 0.00000177, -0.00001337, 0.00001328, -0.00001017, 0.00000617, 0.00000744, },
		{ -0.00000367, -0.00001667, -0.00001243, 0.00004938, -0.00010144, 0.00016654, -0.00010144, 0.00004938, -0.00001242, -0.00001666, -0.00000366, -0.00000959, 0.00001175, -0.00000187, -0.00000299, 0.00000728, -0.00001196, -0.00000525, 0.00001353, 0.00000679, -0.00001181, -0.00001121, 0.00000749, -0.00000212, -0.00001737, -0.00000055, 0.00002444, 0.00000570, -0.00004976, -0.00001750, 0.00004638, -0.00000459, -0.00005438, -0.00003663, 0.00000243, -0.00010322, -0.00059202, 0.00164818, -0.00059203, -0.00010322, 0.00000242, -0.00003664, -0.00005439, -0.00000461, 0.00004636, -0.00001752, -0.00004978, 0.00000568, 0.00002441, -0.00000058, -0.00001740, -0.00000215, 0.00000747, -0.00001123, -0.00001183, 0.00000677, 0.00001351, -0.00000527, -0.00001199, 0.00000726, -0.00000301, -0.00000189, 0.00001173, -0.00000961, },
		{ 0.00002249, -0.00006537, 0.00008942, -0.00009153, 0.00005736, 0.00005736, -0.00009152, 0.00008943, -0.00006536, 0.00002250, -0.00002865, 0.00001148, 0.00000740, -0.00001063, 0.00001202, -0.00001146, 0.00000042, -0.00001205, 0.00003626, -0.00003032, 0.00001636, -0.00003764, 0.00004951, -0.00006162, 0.00004183, -0.00004773, 0.00009482, -0.00010532, 0.00004060, -0.00007961, 0.00015884, -0.00019225, 0.00013544, -0.00025877, 0.00033597, -0.00088684, 0.00082506, 0.00082506, -0.00088684, 0.00033596, -0.00025878, 0.00013543, -0.00019226, 0.00015882, -0.00007963, 0.00004058, -0.00010534, 0.00009480, -0.00004775, 0.00004181, -0.00006164, 0.00004949, -0.00003766, 0.00001634, -0.00003034, 0.00003623, -0.00001207, 0.00000039, -0.00001148, 0.00001200, -0.00001065, 0.00000739, 0.00001146, -0.00002866, },
		{ -0.00001620, -0.00001055, 0.00006463, -0.00012985, 0.00018679, -0.00012985, 0.00006464, -0.00001054, -0.00001619, -0.00000372, -0.00001318, 0.00001573, -0.00000150, -0.00000512, 0.00000886, -0.00001247, -0.00000536, 0.00001406, 0.00000654, -0.00001188, -0.00001121, 0.00000834, -0.00000280, -0.00001831, -0.00000044, 0.00002566, 0.00000721, -0.00005411, -0.00001787, 0.00005201, -0.00000794, -0.00005685, -0.00003710, 0.00000501, -0.00009522, -0.00064377, 0.00173892, -0.00064377, -0.00009523, 0.00000501, -0.00003711, -0.00005686, -0.00000795, 0.00005200, -0.00001789, -0.00005413, 0.00000719, 0.00002564, -0.00000046, -0.00001833, -0.00000282, 0.00000832, -0.00001123, -0.00001190, 0.00000652, 0.00001404, -0.00000538, -0.00001249, 0.00000884, -0.00000513, -0.00000152, 0.00001572, -0.00001319, -0.00000373, },
		{ -0.00007446, 0.00011039, -0.00010815, 0.00005730, 0.00005730, -0.00010815, 0.00011039, -0.00007445, 0.00003006, -0.00003615, 0.00001505, 0.00000835, -0.00001099, 0.00001081, -0.00000975, -0.00000080, -0.00001149, 0.00003665, -0.00003125, 0.00001744, -0.00003880, 0.00005190, -0.00006529, 0.00004463, -0.00005095, 0.00010094, -0.00011140, 0.00004237, -0.00008228, 0.00016853, -0.00020655, 0.00014770, -0.00027590, 0.00036718, -0.00093651, 0.00085690, 0.00085690, -0.00093651, 0.00036717, -0.00027590, 0.00014769, -0.00020656, 0.00016851, -0.00008229, 0.00004235, -0.00011142, 0.00010092, -0.00005096, 0.00004461, -0.00006531, 0.00005188, -0.00003882, 0.00001742, -0.00003127, 0.00003663, -0.00001151, -0.00000082, -0.00000977, 0.00001079, -0.00001100, 0.00000833, 0.00001503, -0.00003616, 0.00003005, },
		{ -0.00000899, 0.00007703, -0.00015354, 0.00020456, -0.00015354, 0.00007704, -0.00000899, -0.00001582, -0.00000381, -0.00001611, 0.00001903, -0.00000121, -0.00000690, 0.00001016, -0.00001286, -0.00000545, 0.00001447, 0.00000633, -0.00001193, -0.00001121, 0.00000902, -0.00000334, -0.00001906, -0.00000036, 0.00002664, 0.00000844, -0.00005765, -0.00001817, 0.00005661, -0.00001068, -0.00005884, -0.00003748, 0.00000712, -0.00008854, -0.00068586, 0.00181224, -0.00068586, -0.00008855, 0.00000712, -0.00003748, -0.00005885, -0.00001069, 0.00005660, -0.00001818, -0.00005766, 0.00000843, 0.00002663, -0.00000037, -0.00001908, -0.00000336, 0.00000900, -0.00001123, -0.00001194, 0.00000631, 0.00001445, -0.00000547, -0.00001287, 0.00001014, -0.00000691, -0.00000122, 0.00001902, -0.00001612, -0.00000382, -0.00001582, },
		{ 0.00012701, -0.00012166, 0.00005769, 0.00005769, -0.00012166, 0.00012701, -0.00008170, 0.00003607, -0.00004214, 0.00001793, 0.00000904, -0.00001124, 0.00000979, -0.00000835, -0.00000179, -0.00001103, 0.00003690, -0.00003194, 0.00001826, -0.00003968, 0.00005373, -0.00006812, 0.00004679, -0.00005345, 0.00010569, -0.00011612, 0.00004373, -0.00008433, 0.00017605, -0.00021770, 0.00015726, -0.00028924, 0.00039156, -0.00097508, 0.00088148, 0.00088148, -0.00097508, 0.00039156, -0.00028925, 0.00015726, -0.00021771, 0.00017605, -0.00008434, 0.00004372, -0.00011613, 0.00010568, -0.00005346, 0.00004678, -0.00006814, 0.00005371, -0.00003970, 0.00001825, -0.00003195, 0.00003689, -0.00001104, -0.00000180, -0.00000837, 0.00000978, -0.00001125, 0.00000903, 0.00001792, -0.00004215, 0.00003606, -0.00008170, },
		{ 0.00008619, -0.00017135, 0.00021837, -0.00017135, 0.00008619, -0.00000783, -0.00001555, -0.00000390, -0.00001829, 0.00002149, -0.00000100, -0.00000824, 0.00001112, -0.00001313, -0.00000552, 0.00001476, 0.00000617, -0.00001195, -0.00001122, 0.00000951, -0.00000375, -0.00001961, -0.00000030, 0.00002735, 0.00000936, -0.00006026, -0.00001838, 0.00006001, -0.00001273, -0.00006030, -0.00003775, 0.00000868, -0.00008352, -0.00071693, 0.00186611, -0.00071693, -0.00008352, 0.00000868, -0.00003775, -0.00006031, -0.00001273, 0.00006000, -0.00001839, -0.00006027, 0.00000935, 0.00002734, -0.00000031, -0.00001962, -0.00000376, 0.00000950, -0.00001123, -0.00001196, 0.00000615, 0.00001475, -0.00000553, -0.00001314, 0.00001111, -0.00000825, -0.00000101, 0.00002149, -0.00001830, -0.00000390, -0.00001555, -0.00000783, },
		{ -0.00013117, 0.00005815, 0.00005816, -0.00013117, 0.00013852, -0.00008674, 0.00004024, -0.00004631, 0.00001996, 0.00000949, -0.00001141, 0.00000906, -0.00000736, -0.00000248, -0.00001069, 0.00003705, -0.00003239, 0.00001881, -0.00004028, 0.00005496, -0.00007005, 0.00004827, -0.00005516, 0.00010893, -0.00011934, 0.00004465, -0.00008572, 0.00018120, -0.00022533, 0.00016382, -0.00029838, 0.00040829, -0.00100144, 0.00089820, 0.00089820, -0.00100144, 0.00040829, -0.00029838, 0.00016382, -0.00022534, 0.00018120, -0.00008572, 0.00004464, -0.00011934, 0.00010893, -0.00005517, 0.00004826, -0.00007006, 0.00005496, -0.00004029, 0.00001880, -0.00003240, 0.00003704, -0.00001070, -0.00000249, -0.00000737, 0.00000905, -0.00001141, 0.00000949, 0.00001995, -0.00004632, 0.00004024, -0.00008675, 0.00013852, },
		{ -0.00018240, 0.00022712, -0.00018240, 0.00009180, -0.00000711, -0.00001539, -0.00000396, -0.00001963, 0.00002301, -0.00000088, -0.00000908, 0.00001171, -0.00001329, -0.00000556, 0.00001493, 0.00000606, -0.00001197, -0.00001122, 0.00000981, -0.00000399, -0.00001994, -0.00000026, 0.00002778, 0.00000993, -0.00006186, -0.00001851, 0.00006209, -0.00001399, -0.00006120, -0.00003791, 0.00000964, -0.00008040, -0.00073598, 0.00189904, -0.00073599, -0.00008040, 0.00000964, -0.00003791, -0.00006120, -0.00001399, 0.00006209, -0.00001851, -0.00006187, 0.00000992, 0.00002778, -0.00000027, -0.00001995, -0.00000400, 0.00000981, -0.00001122, -0.00001197, 0.00000606, 0.00001492, -0.00000557, -0.00001330, 0.00001170, -0.00000909, -0.00000088, 0.00002301, -0.00001963, -0.00000396, -0.00001539, -0.00000711, 0.00009180, },
		{ 0.00005845, 0.00005845, -0.00013609, 0.00014440, -0.00008933, 0.00004237, -0.00004845, 0.00002100, 0.00000972, -0.00001148, 0.00000867, -0.00000685, -0.00000284, -0.00001052, 0.00003712, -0.00003262, 0.00001908, -0.00004058, 0.00005559, -0.00007102, 0.00004901, -0.00005603, 0.00011058, -0.00012097, 0.00004512, -0.00008642, 0.00018381, -0.00022921, 0.00016716, -0.00030303, 0.00041680, -0.00101481, 0.00090667, 0.00090667, -0.00101481, 0.00041680, -0.00030303, 0.00016716, -0.00022921, 0.00018381, -0.00008642, 0.00004511, -0.00012097, 0.00011057, -0.00005604, 0.00004901, -0.00007103, 0.00005558, -0.00004058, 0.00001908, -0.00003262, 0.00003712, -0.00001052, -0.00000284, -0.00000685, 0.00000867, -0.00001149, 0.00000971, 0.00002100, -0.00004845, 0.00004237, -0.00008933, 0.00014440, -0.00013609, },
		{ 0.00023011, -0.00018614, 0.00009369, -0.00000687, -0.00001533, -0.00000398, -0.00002009, 0.00002353, -0.00000084, -0.00000937, 0.00001191, -0.00001335, -0.00000558, 0.00001498, 0.00000603, -0.00001197, -0.00001122, 0.00000991, -0.00000408, -0.00002005, -0.00000025, 0.00002792, 0.00001012, -0.00006240, -0.00001855, 0.00006280, -0.00001442, -0.00006150, -0.00003797, 0.00000996, -0.00007934, -0.00074241, 0.00191013, -0.00074241, -0.00007934, 0.00000996, -0.00003797, -0.00006150, -0.00001442, 0.00006280, -0.00001855, -0.00006240, 0.00001012, 0.00002792, -0.00000025, -0.00002005, -0.00000408, 0.00000991, -0.00001122, -0.00001197, 0.00000603, 0.00001498, -0.00000558, -0.00001335, 0.00001191, -0.00000937, -0.00000084, 0.00002353, -0.00002009, -0.00000398, -0.00001533, -0.00000687, 0.00009369, -0.00018614, },
		{ 0.00005845, -0.00013609, 0.00014440, -0.00008933, 0.00004237, -0.00004845, 0.00002100, 0.00000971, -0.00001149, 0.00000867, -0.00000685, -0.00000284, -0.00001052, 0.00003712, -0.00003262, 0.00001908, -0.00004058, 0.00005558, -0.00007103, 0.00004901, -0.00005604, 0.00011057, -0.00012097, 0.00004511, -0.00008642, 0.00018381, -0.00022921, 0.00016716, -0.00030303, 0.00041680, -0.00101481, 0.00090667, 0.00090667, -0.00101481, 0.00041680, -0.00030303, 0.00016716, -0.00022921, 0.00018381, -0.00008642, 0.00004512, -0.00012097, 0.00011058, -0.00005603, 0.00004901, -0.00007102, 0.00005559, -0.00004058, 0.00001908, -0.00003262, 0.00003712, -0.00001052, -0.00000284, -0.00000685, 0.00000867, -0.00001148, 0.00000972, 0.00002100, -0.00004845, 0.00004237, -0.00008933, 0.00014440, -0.00013609, 0.00005845, },
		{ -0.00018240, 0.00009180, -0.00000711, -0.00001539, -0.00000396, -0.00001963, 0.00002301, -0.00000088, -0.00000909, 0.00001170, -0.00001330, -0.00000557, 0.00001492, 0.00000606, -0.00001197, -0.00001122, 0.00000981, -0.00000400, -0.00001995, -0.00000027, 0.00002778, 0.00000992, -0.00006187, -0.00001851, 0.00006209, -0.00001399, -0.00006120, -0.00003791, 0.00000964, -0.00008040, -0.00073599, 0.00189904, -0.00073598, -0.00008040, 0.00000964, -0.00003791, -0.00006120, -0.00001399, 0.00006209, -0.00001851, -0.00006186, 0.00000993, 0.00002778, -0.00000026, -0.00001994, -0.00000399, 0.00000981, -0.00001122, -0.00001197, 0.00000606, 0.00001493, -0.00000556, -0.00001329, 0.00001171, -0.00000908, -0.00000088, 0.00002301, -0.00001963, -0.00000396, -0.00001539, -0.00000711, 0.00009180, -0.00018240, 0.00022712, },
		{ -0.00013117, 0.00013852, -0.00008675, 0.00004024, -0.00004632, 0.00001995, 0.00000949, -0.00001141, 0.00000905, -0.00000737, -0.00000249, -0.00001070, 0.00003704, -0.00003240, 0.00001880, -0.00004029, 0.00005496, -0.00007006, 0.00004826, -0.00005517, 0.00010893, -0.00011934, 0.00004464, -0.00008572, 0.00018120, -0.00022534, 0.00016382, -0.00029838, 0.00040829, -0.00100144, 0.00089820, 0.00089820, -0.00100144, 0.00040829, -0.00029838, 0.00016382, -0.00022533, 0.00018120, -0.00008572, 0.00004465, -0.00011934, 0.00010893, -0.00005516, 0.00004827, -0.00007005, 0.00005496, -0.00004028, 0.00001881, -0.00003239, 0.00003705, -0.00001069, -0.00000248, -0.00000736, 0.00000906, -0.00001141, 0.00000949, 0.00001996, -0.00004631, 0.00004024, -0.00008674, 0.00013852, -0.00013117, 0.00005816, 0.00005815, },
		{ 0.00008619, -0.00000783, -0.00001555, -0.00000390, -0.00001830, 0.00002149, -0.00000101, -0.00000825, 0.00001111, -0.00001314, -0.00000553, 0.00001475, 0.00000615, -0.00001196, -0.00001123, 0.00000950, -0.00000376, -0.00001962, -0.00000031, 0.00002734, 0.00000935, -0.00006027, -0.00001839, 0.00006000, -0.00001273, -0.00006031, -0.00003775, 0.00000868, -0.00008352, -0.00071693, 0.00186611, -0.00071693, -0.00008352, 0.00000868, -0.00003775, -0.00006030, -0.00001273, 0.00006001, -0.00001838, -0.00006026, 0.00000936, 0.00002735, -0.00000030, -0.00001961, -0.00000375, 0.00000951, -0.00001122, -0.00001195, 0.00000617, 0.00001476, -0.00000552, -0.00001313, 0.00001112, -0.00000824, -0.00000100, 0.00002149, -0.00001829, -0.00000390, -0.00001555, -0.00000783, 0.00008619, -0.00017135, 0.00021837, -0.00017135, },
		{ 0.00012701, -0.00008170, 0.00003606, -0.00004215, 0.00001792, 0.00000903, -0.00001125, 0.00000978, -0.00000837, -0.00000180, -0.00001104, 0.00003689, -0.00003195, 0.00001825, -0.00003970, 0.00005371, -0.00006814, 0.00004678, -0.00005346, 0.00010568, -0.00011613, 0.00004372, -0.00008434, 0.00017605, -0.00021771, 0.00015726, -0.00028925, 0.00039156, -0.00097508, 0.00088148, 0.00088148, -0.00097508, 0.00039156, -0.00028924, 0.00015726, -0.00021770, 0.00017605, -0.00008433, 0.00004373, -0.00011612, 0.00010569, -0.00005345, 0.00004679, -0.00006812, 0.00005373, -0.00003968, 0.00001826, -0.00003194, 0.00003690, -0.00001103, -0.00000179, -0.00000835, 0.00000979, -0.00001124, 0.00000904, 0.00001793, -0.00004214, 0.00003607, -0.00008170, 0.00012701, -0.00012166, 0.00005769, 0.00005769, -0.00012166, },
		{ -0.00000899, -0.00001582, -0.00000382, -0.00001612, 0.00001902, -0.00000122, -0.00000691, 0.00001014, -0.00001287, -0.00000547, 0.00001445, 0.00000631, -0.00001194, -0.00001123, 0.00000900, -0.00000336, -0.00001908, -0.00000037, 0.00002663, 0.00000843, -0.00005766, -0.00001818, 0.00005660, -0.00001069, -0.00005885, -0.00003748, 0.00000712, -0.00008855, -0.00068586, 0.00181224, -0.00068586, -0.00008854, 0.00000712, -0.00003748, -0.00005884, -0.00001068, 0.00005661, -0.00001817, -0.00005765, 0.00000844, 0.00002664, -0.00000036, -0.00001906, -0.00000334, 0.00000902, -0.00001121, -0.00001193, 0.00000633, 0.00001447, -0.00000545, -0.00001286, 0.00001016, -0.00000690, -0.00000121, 0.00001903, -0.00001611, -0.00000381, -0.00001582, -0.00000899, 0.00007704, -0.00015354, 0.00020456, -0.00015354, 0.00007703, },
		{ -0.00007446, 0.00003005, -0.00003616, 0.00001503, 0.00000833, -0.00001100, 0.00001079, -0.00000977, -0.00000082, -0.00001151, 0.00003663, -0.00003127, 0.00001742, -0.00003882, 0.00005188, -0.00006531, 0.00004461, -0.00005096, 0.00010092, -0.00011142, 0.00004235, -0.00008229, 0.00016851, -0.00020656, 0.00014769, -0.00027590, 0.00036717, -0.00093651, 0.00085690, 0.00085690, -0.00093651, 0.00036718, -0.00027590, 0.00014770, -0.00020655, 0.00016853, -0.00008228, 0.00004237, -0.00011140, 0.00010094, -0.00005095, 0.00004463, -0.00006529, 0.00005190, -0.00003880, 0.00001744, -0.00003125, 0.00003665, -0.00001149, -0.00000080, -0.00000975, 0.00001081, -0.00001099, 0.00000835, 0.00001505, -0.00003615, 0.00003006, -0.00007445, 0.00011039, -0.00010815, 0.00005730, 0.00005730, -0.00010815, 0.00011039, },
		{ -0.00001620, -0.00000373, -0.00001319, 0.00001572, -0.00000152, -0.00000513, 0.00000884, -0.00001249, -0.00000538, 0.00001404, 0.00000652, -0.00001190, -0.00001123, 0.00000832, -0.00000282, -0.00001833, -0.00000046, 0.00002564, 0.00000719, -0.00005413, -0.00001789, 0.00005200, -0.00000795, -0.00005686, -0.00003711, 0.00000501, -0.00009523, -0.00064377, 0.00173892, -0.00064377, -0.00009522, 0.00000501, -0.00003710, -0.00005685, -0.00000794, 0.00005201, -0.00001787, -0.00005411, 0.00000721, 0.00002566, -0.00000044, -0.00001831, -0.00000280, 0.00000834, -0.00001121, -0.00001188, 0.00000654, 0.00001406, -0.00000536, -0.00001247, 0.00000886, -0.00000512, -0.00000150, 0.00001573, -0.00001318, -0.00000372, -0.00001619, -0.00001054, 0.00006464, -0.00012985, 0.00018679, -0.00012985, 0.00006463, -0.00001055, },
		{ 0.00002249, -0.00002866, 0.00001146, 0.00000739, -0.00001065, 0.00001200, -0.00001148, 0.00000039, -0.00001207, 0.00003623, -0.00003034, 0.00001634, -0.00003766, 0.00004949, -0.00006164, 0.00004181, -0.00004775, 0.00009480, -0.00010534, 0.00004058, -0.00007963, 0.00015882, -0.00019226, 0.00013543, -0.00025878, 0.00033596, -0.00088684, 0.00082506, 0.00082506, -0.00088684, 0.00033597, -0.00025877, 0.00013544, -0.00019225, 0.00015884, -0.00007961, 0.00004060, -0.00010532, 0.00009482, -0.00004773, 0.00004183, -0.00006162, 0.00004951, -0.00003764, 0.00001636, -0.00003032, 0.00003626, -0.00001205, 0.00000042, -0.00001146, 0.00001202, -0.00001063, 0.00000740, 0.00001148, -0.00002865, 0.00002250, -0.00006536, 0.00008943, -0.00009152, 0.00005736, 0.00005736, -0.00009153, 0.00008942, -0.00006537, },
		{ -0.00000367, -0.00000961, 0.00001173, -0.00000189, -0.00000301, 0.00000726, -0.00001199, -0.00000527, 0.00001351, 0.00000677, -0.00001183, -0.00001123, 0.00000747, -0.00000215, -0.00001740, -0.00000058, 0.00002441, 0.00000568, -0.00004978, -0.00001752, 0.00004636, -0.00000461, -0.00005439, -0.00003664, 0.00000242, -0.00010322, -0.00059203, 0.00164818, -0.00059202, -0.00010322, 0.00000243, -0.00003663, -0.00005438, -0.00000459, 0.00004638, -0.00001750, -0.00004976, 0.00000570, 0.00002444, -0.00000055, -0.00001737, -0.00000212, 0.00000749, -0.00001121, -0.00001181, 0.00000679, 0.00001353, -0.00000525, -0.00001196, 0.00000728, -0.00000299, -0.00000187, 0.00001175, -0.00000959, -0.00000366, -0.00001666, -0.00001242, 0.00004938, -0.00010144, 0.00016654, -0.00010144, 0.00004938, -0.00001243, -0.00001667, },
		{ -0.00002003, 0.00000744, 0.00000617, -0.00001017, 0.00001328, -0.00001337, 0.00000177, -0.00001267, 0.00003565, -0.00002916, 0.00001500, -0.00003624, 0.00004659, -0.00005722, 0.00003846, -0.00004391, 0.00008747, -0.00009805, 0.00003843, -0.00007639, 0.00014726, -0.00017526, 0.00012088, -0.00023843, 0.00029897, -0.00082752, 0.00078674, 0.00078674, -0.00082752, 0.00029898, -0.00023842, 0.00012089, -0.00017524, 0.00014727, -0.00007637, 0.00003846, -0.00009803, 0.00008749, -0.00004389, 0.00003848, -0.00005719, 0.00004662, -0.00003621, 0.00001503, -0.00002913, 0.00003568, -0.00001264, 0.00000179, -0.00001335, 0.00001331, -0.00001015, 0.00000619, 0.00000746, -0.00002001, 0.00001373, -0.00005490, 0.00006506, -0.00007290, 0.00005832, 0.00005832, -0.00007290, 0.00006505, -0.00005491, 0.00001372, },
		{ -0.00000552, 0.00000723, -0.00000234, -0.00000065, 0.00000545, -0.00001136, -0.00000514, 0.00001285, 0.00000703, -0.00001173, -0.00001124, 0.00000647, -0.00000137, -0.00001628, -0.00000071, 0.00002296, 0.00000396, -0.00004475, -0.00001707, 0.00003988, -0.00000080, -0.00005150, -0.00003607, -0.00000055, -0.00011212, -0.00053227, 0.00154249, -0.00053227, -0.00011211, -0.00000054, -0.00003605, -0.00005148, -0.00000079, 0.00003990, -0.00001705, -0.00004472, 0.00000399, 0.00002299, -0.00000069, -0.00001625, -0.00000134, 0.00000650, -0.00001121, -0.00001170, 0.00000706, 0.00001288, -0.00000511, -0.00001133, 0.00000547, -0.00000063, -0.00000232, 0.00000725, -0.00000550, -0.00000366, -0.00001723, -0.00001454, 0.00003176, -0.00006975, 0.00014561, -0.00006975, 0.00003175, -0.00001455, -0.00001725, -0.00000368, },
		{ 0.00000322, 0.00000467, -0.00000955, 0.00001451, -0.00001529, 0.00000321, -0.00001324, 0.00003484, -0.00002771, 0.00001342, -0.00003456, 0.00004322, -0.00005213, 0.00003464, -0.00003956, 0.00007914, -0.00008975, 0.00003596, -0.00007264, 0.00013415, -0.00015608, 0.00010452, -0.00021547, 0.00025745, -0.00076023, 0.00074283, 0.00074283, -0.00076023, 0.00025746, -0.00021546, 0.00010453, -0.00015606, 0.00013417, -0.00007262, 0.00003599, -0.00008972, 0.00007917, -0.00003953, 0.00003467, -0.00005210, 0.00004326, -0.00003453, 0.00001345, -0.00002768, 0.00003487, -0.00001321, 0.00000324, -0.00001526, 0.00001453, -0.00000953, 0.00000469, 0.00000324, -0.00001070, 0.00000418, -0.00004361, 0.00003840, -0.00005358, 0.00006071, 0.00006071, -0.00005358, 0.00003839, -0.00004362, 0.00000417, -0.00001071, },
		{ 0.00000243, -0.00000286, 0.00000181, 0.00000348, -0.00001060, -0.00000499, 0.00001208, 0.00000730, -0.00001157, -0.00001124, 0.00000534, -0.00000051, -0.00001500, -0.00000087, 0.00002130, 0.00000210, -0.00003919, -0.00001656, 0.00003278, 0.00000331, -0.00004825, -0.00003541, -0.00000381, -0.00012145, -0.00046640, 0.00142469, -0.00046639, -0.00012145, -0.00000380, -0.00003539, -0.00004823, 0.00000333, 0.00003280, -0.00001653, -0.00003916, 0.00000213, 0.00002133, -0.00000084, -0.00001497, -0.00000047, 0.00000538, -0.00001120, -0.00001154, 0.00000734, 0.00001212, -0.00000495, -0.00001057, 0.00000351, 0.00000183, -0.00000283, 0.00000245, -0.00000105, -0.00000377, -0.00001789, -0.00001681, 0.00001231, -0.00003639, 0.00012602, -0.00003640, 0.00001230, -0.00001682, -0.00001791, -0.00000379, -0.00000107, },
		{ 0.00000287, -0.00000877, 0.00001552, -0.00001706, 0.00000463, -0.00001371, 0.00003374, -0.00002597, 0.00001161, -0.00003265, 0.00003946, -0.00004649, 0.00003045, -0.00003483, 0.00007005, -0.00008065, 0.00003322, -0.00006845, 0.00011988, -0.00013532, 0.00008688, -0.00019064, 0.00021277, -0.00068688, 0.00069436, 0.00069436, -0.00068687, 0.00021278, -0.00019062, 0.00008690, -0.00013530, 0.00011990, -0.00006843, 0.00003324, -0.00008062, 0.00007008, -0.00003480, 0.00003049, -0.00004645, 0.00003950, -0.00003261, 0.00001165, -0.00002594, 0.00003377, -0.00001368, 0.00000466, -0.00001703, 0.00001555, -0.00000874, 0.00000289, -0.00000088, -0.00000121, -0.00000568, -0.00003209, 0.00001070, -0.00003499, 0.00006511, 0.00006511, -0.00003500, 0.00001069, -0.00003211, -0.00000569, -0.00000123, -0.00000090, },
		{ -0.00000345, 0.00000423, 0.00000143, -0.00000971, -0.00000481, 0.00001119, 0.00000756, -0.00001135, -0.00001124, 0.00000412, 0.00000042, -0.00001359, -0.00000105, 0.00001947, 0.00000017, -0.00003327, -0.00001597, 0.00002529, 0.00000758, -0.00004472, -0.00003466, -0.00000724, -0.00013072, -0.00039647, 0.00129788, -0.00039647, -0.00013071, -0.00000723, -0.00003465, -0.00004470, 0.00000760, 0.00002532, -0.00001595, -0.00003324, 0.00000020, 0.00001951, -0.00000101, -0.00001355, 0.00000045, 0.00000416, -0.00001120, -0.00001131, 0.00000760, 0.00001123, -0.00000478, -0.00000968, 0.00000146, 0.00000426, -0.00000342, -0.00000245, 0.00000359, -0.00000404, -0.00001864, -0.00001913, -0.00000836, -0.00000313, 0.00010993, -0.00000313, -0.00000837, -0.00001914, -0.00001866, -0.00000405, 0.00000357, -0.00000247, },
		{ -0.00000781, 0.00001617, -0.00001852, 0.00000591, -0.00001402, 0.00003229, -0.00002394, 0.00000959, -0.00003053, 0.00003536, -0.00004042, 0.00002601, -0.00002985, 0.00006043, -0.00007099, 0.00003025, -0.00006390, 0.00010485, -0.00011362, 0.00006852, -0.00016468, 0.00016641, -0.00060948, 0.00064242, 0.00064242, -0.00060947, 0.00016642, -0.00016467, 0.00006854, -0.00011360, 0.00010487, -0.00006387, 0.00003028, -0.00007096, 0.00006046, -0.00002981, 0.00002605, -0.00004038, 0.00003541, -0.00003048, 0.00000963, -0.00002390, 0.00003233, -0.00001398, 0.00000595, -0.00001849, 0.00001621, -0.00000778, 0.00000077, -0.00000460, 0.00000794, -0.00001536, -0.00002098, -0.00001673, -0.00001864, 0.00007211, 0.00007211, -0.00001865, -0.00001674, -0.00002100, -0.00001538, 0.00000791, -0.00000462, 0.00000074, },
		{ 0.00000647, -0.00000062, -0.00000868, -0.00000462, 0.00001018, 0.00000778, -0.00001106, -0.00001124, 0.00000282, 0.00000137, -0.00001205, -0.00000124, 0.00001751, -0.00000176, -0.00002718, -0.00001533, 0.00001767, 0.00001183, -0.00004100, -0.00003384, -0.00001073, -0.00013939, -0.00032465, 0.00116536, -0.00032465, -0.00013938, -0.00001071, -0.00003382, -0.00004098, 0.00001185, 0.00001770, -0.00001530, -0.00002714, -0.00000172, 0.00001754, -0.00000120, -0.00001201, 0.00000142, 0.00000287, -0.00001120, -0.00001102, 0.00000783, 0.00001022, -0.00000458, -0.00000864, -0.00000059, 0.00000650, -0.00000406, -0.00000721, 0.00000825, -0.00000449, -0.00001946, -0.00002139, -0.00002961, 0.00002825, 0.00009957, 0.00002824, -0.00002962, -0.00002141, -0.00001948, -0.00000452, 0.00000823, -0.00000724, -0.00000409, },
		{ 0.00001631, -0.00001950, 0.00000697, -0.00001409, 0.00003045, -0.00002161, 0.00000737, -0.00002821, 0.00003100, -0.00003404, 0.00002142, -0.00002476, 0.00005053, -0.00006101, 0.00002712, -0.00005905, 0.00008947, -0.00009164, 0.00005003, -0.00013839, 0.00011988, -0.00053014, 0.00058817, 0.00058817, -0.00053013, 0.00011989, -0.00013838, 0.00005005, -0.00009161, 0.00008950, -0.00005902, 0.00002715, -0.00006098, 0.00005057, -0.00002472, 0.00002146, -0.00003400, 0.00003105, -0.00002816, 0.00000742, -0.00002156, 0.00003050, -0.00001404, 0.00000701, -0.00001946, 0.00001634, -0.00000661, -0.00000168, -0.00000761, 0.00001620, -0.00002437, -0.00001092, -0.00004257, -0.00000604, 0.00008228, 0.00008228, -0.00000605, -0.00004259, -0.00001094, -0.00002439, 0.00001617, -0.00000764, -0.00000171, -0.00000664, },
		{ -0.00000261, -0.00000750, -0.00000441, 0.00000905, 0.00000795, -0.00001068, -0.00001124, 0.00000148, 0.00000234, -0.00001042, -0.00000146, 0.00001543, -0.00000360, -0.00002108, -0.00001463, 0.00001017, 0.00001589, -0.00003717, -0.00003295, -0.00001415, -0.00014696, -0.00025311, 0.00103045, -0.00025311, -0.00014695, -0.00001414, -0.00003293, -0.00003714, 0.00001591, 0.00001020, -0.00001460, -0.00002104, -0.00000356, 0.00001547, -0.00000141, -0.00001037, 0.00000239, 0.00000152, -0.00001120, -0.00001063, 0.00000800, 0.00000910, -0.00000437, -0.00000746, -0.00000257, 0.00000842, -0.00000476, -0.00001162, 0.00001276, -0.00000519, -0.00002035, -0.00002349, -0.00005082, 0.00005594, 0.00009708, 0.00005593, -0.00005083, -0.00002351, -0.00002037, -0.00000521, 0.00001274, -0.00001165, -0.00000479, 0.00000839, },
		{ -0.00001982, 0.00000769, -0.00001385, 0.00002817, -0.00001896, 0.00000499, -0.00002572, 0.00002644, -0.00002749, 0.00001680, -0.00001969, 0.00004062, -0.00005096, 0.00002388, -0.00005400, 0.00007416, -0.00007001, 0.00003198, -0.00011254, 0.00007466, -0.00045094, 0.00053277, 0.00053277, -0.00045093, 0.00007467, -0.00011252, 0.00003200, -0.00006998, 0.00007418, -0.00005397, 0.00002391, -0.00005092, 0.00004066, -0.00001964, 0.00001684, -0.00002744, 0.00002649, -0.00002567, 0.00000503, -0.00001891, 0.00002822, -0.00001380, 0.00000773, -0.00001978, 0.00001581, -0.00000522, -0.00000448, -0.00000962, 0.00002306, -0.00003223, -0.00000253, -0.00006554, 0.00000137, 0.00009617, 0.00009617, 0.00000136, -0.00006555, -0.00000255, -0.00003225, 0.00002303, -0.00000965, -0.00000452, -0.00000525, 0.00001577, },
		{ -0.00000618, -0.00000419, 0.00000781, 0.00000805, -0.00001020, -0.00001125, 0.00000011, 0.00000329, -0.00000871, -0.00000168, 0.00001328, -0.00000529, -0.00001515, -0.00001389, 0.00000303, 0.00001960, -0.00003330, -0.00003200, -0.00001741, -0.00015294, -0.00018398, 0.00089643, -0.00018397, -0.00015293, -0.00001740, -0.00003198, -0.00003328, 0.00001962, 0.00000306, -0.00001385, -0.00001511, -0.00000525, 0.00001332, -0.00000164, -0.00000866, 0.00000334, 0.00000016, -0.00001120, -0.00001016, 0.00000810, 0.00000786, -0.00000415, -0.00000614, -0.00000441, 0.00000989, -0.00000550, -0.00001545, 0.00001695, -0.00000617, -0.00002131, -0.00002533, -0.00007137, 0.00007825, 0.00010447, 0.00007824, -0.00007138, -0.00002535, -0.00002132, -0.00000619, 0.00001692, -0.00001548, -0.00000554, 0.00000985, -0.00000445, },
		{ 0.00000799, -0.00001324, 0.00002541, -0.00001600, 0.00000244, -0.00002309, 0.00002176, -0.00002088, 0.00001224, -0.00001478, 0.00003092, -0.00004105, 0.00002059, -0.00004882, 0.00005931, -0.00004935, 0.00001492, -0.00008787, 0.00003218, -0.00037389, 0.00047734, 0.00047734, -0.00037389, 0.00003219, -0.00008786, 0.00001494, -0.00004933, 0.00005934, -0.00004879, 0.00002063, -0.00004101, 0.00003097, -0.00001473, 0.00001229, -0.00002083, 0.00002181, -0.00002304, 0.00000250, -0.00001595, 0.00002546, -0.00001319, 0.00000804, -0.00001930, 0.00001448, -0.00000360, -0.00000764, -0.00001036, 0.00002805, -0.00003848, 0.00000362, -0.00008441, 0.00000227, 0.00011424, 0.00011424, 0.00000226, -0.00008442, 0.00000360, -0.00003850, 0.00002802, -0.00001039, -0.00000767, -0.00000364, 0.00001444, -0.00001934, },
		{ -0.00000396, 0.00000646, 0.00000806, -0.00000962, -0.00001125, -0.00000125, 0.00000420, -0.00000695, -0.00000192, 0.00001108, -0.00000676, -0.00000956, -0.00001310, -0.00000352, 0.00002280, -0.00002948, -0.00003099, -0.00002039, -0.00015689, -0.00011923, 0.00076641, -0.00011922, -0.00015688, -0.00002038, -0.00003097, -0.00002946, 0.00002283, -0.00000349, -0.00001307, -0.00000952, -0.00000671, 0.00001113, -0.00000187, -0.00000690, 0.00000425, -0.00000120, -0.00001120, -0.00000957, 0.00000811, 0.00000651, -0.00000391, -0.00000467, -0.00000603, 0.00001078, -0.00000629, -0.00001852, 0.00002066, -0.00000745, -0.00002231, -0.00002682, -0.00009067, 0.00009365, 0.00012350, 0.00009365, -0.00009068, -0.00002683, -0.00002233, -0.00000747, 0.00002063, -0.00001855, -0.00000632, 0.00001074, -0.00000607, -0.00000471, },
		{ -0.00001222, 0.00002214, -0.00001272, -0.00000023, -0.00002035, 0.00001703, -0.00001434, 0.00000786, -0.00001015, 0.00002168, -0.00003152, 0.00001732, -0.00004360, 0.00004529, -0.00003025, -0.00000066, -0.00006507, -0.00000627, -0.00030087, 0.00042295, 0.00042295, -0.00030086, -0.00000626, -0.00006505, -0.00000064, -0.00003022, 0.00004532, -0.00004356, 0.00001736, -0.00003148, 0.00002172, -0.00001010, 0.00000791, -0.00001429, 0.00001708, -0.00002030, -0.00000017, -0.00001267, 0.00002219, -0.00001217, 0.00000784, -0.00001788, 0.00001225, -0.00000174, -0.00001114, -0.00000960, 0.00003075, -0.00004273, 0.00000703, -0.00009814, -0.00000447, 0.00013685, 0.00013685, -0.00000447, -0.00009815, 0.00000701, -0.00004275, 0.00003072, -0.00000963, -0.00001117, -0.00000177, 0.00001220, -0.00001793, 0.00000780, },
		{ 0.00000501, 0.00000797, -0.00000893, -0.00001125, -0.00000258, 0.00000504, -0.00000516, -0.00000217, 0.00000888, -0.00000794, -0.00000445, -0.00001229, -0.00000929, 0.00002538, -0.00002579, -0.00002995, -0.00002300, -0.00015844, -0.00006064, 0.00064326, -0.00006063, -0.00015843, -0.00002299, -0.00002993, -0.00002576, 0.00002541, -0.00000925, -0.00001225, -0.00000441, -0.00000790, 0.00000893, -0.00000212, -0.00000511, 0.00000509, -0.00000253, -0.00001119, -0.00000888, 0.00000802, 0.00000506, -0.00000366, -0.00000305, -0.00000738, 0.00001099, -0.00000711, -0.00002065, 0.00002377, -0.00000907, -0.00002335, -0.00002787, -0.00010820, 0.00010084, 0.00015557, 0.00010084, -0.00010821, -0.00002789, -0.00002338, -0.00000909, 0.00002374, -0.00002069, -0.00000714, 0.00001094, -0.00000742, -0.00000310, -0.00000371, },
		{ 0.00001834, -0.00000915, -0.00000300, -0.00001751, 0.00001231, -0.00000797, 0.00000374, -0.00000590, 0.00001308, -0.00002256, 0.00001410, -0.00003840, 0.00003242, -0.00001318, -0.00001431, -0.00004473, -0.00003958, -0.00023351, 0.00037059, 0.00037059, -0.00023351, -0.00003956, -0.00004471, -0.00001429, -0.00001316, 0.00003246, -0.00003837, 0.00001414, -0.00002252, 0.00001312, -0.00000585, 0.00000379, -0.00000791, 0.00001236, -0.00001746, -0.00000295, -0.00000910, 0.00001839, -0.00001069, 0.00000709, -0.00001544, 0.00000902, 0.00000037, -0.00001498, -0.00000715, 0.00003082, -0.00004465, 0.00000728, -0.00010583, -0.00001974, 0.00016424, 0.00016424, -0.00001974, -0.00010585, 0.00000726, -0.00004467, 0.00003079, -0.00000718, -0.00001502, 0.00000033, 0.00000897, -0.00001549, 0.00000704, -0.00001074, },
		{ 0.00000777, -0.00000813, -0.00001125, -0.00000386, 0.00000580, -0.00000337, -0.00000242, 0.00000669, -0.00000880, 0.00000005, -0.00001144, -0.00001409, 0.00002722, -0.00002228, -0.00002887, -0.00002516, -0.00015729, -0.00000972, 0.00052945, -0.00000972, -0.00015728, -0.00002515, -0.00002885, -0.00002225, 0.00002725, -0.00001405, -0.00001140, 0.00000009, -0.00000876, 0.00000674, -0.00000237, -0.00000332, 0.00000585, -0.00000380, -0.00001119, -0.00000807, 0.00000782, 0.00000352, -0.00000341, -0.00000131, -0.00000841, 0.00001044, -0.00000795, -0.00002171, 0.00002615, -0.00001103, -0.00002443, -0.00002843, -0.00012353, 0.00009881, 0.00020169, 0.00009880, -0.00012354, -0.00002845, -0.00002445, -0.00001106, 0.00002612, -0.00002175, -0.00000799, 0.00001040, -0.00000845, -0.00000136, -0.00000346, 0.00000346, },
		{ -0.00000529, -0.00000586, -0.00001462, 0.00000766, -0.00000187, -0.00000003, -0.00000214, 0.00000530, -0.00001434, 0.00001099, -0.00003331, 0.00002099, 0.00000143, -0.00002569, -0.00002734, -0.00006682, -0.00017324, 0.00032112, 0.00032112, -0.00017323, -0.00006681, -0.00002732, -0.00002567, 0.00000146, 0.00002102, -0.00003327, 0.00001103, -0.00001430, 0.00000535, -0.00000209, 0.00000002, -0.00000182, 0.00000772, -0.00001456, -0.00000580, -0.00000524, 0.00001407, -0.00000873, 0.00000572, -0.00001190, 0.00000475, 0.00000272, -0.00001915, -0.00000289, 0.00002801, -0.00004400, 0.00000407, -0.00010683, -0.00004414, 0.00019652, 0.00019651, -0.00004415, -0.00010684, 0.00000405, -0.00004402, 0.00002798, -0.00000293, -0.00001918, 0.00000267, 0.00000470, -0.00001195, 0.00000567, -0.00000878, 0.00001402, },
		{ -0.00000721, -0.00001125, -0.00000507, 0.00000646, -0.00000159, -0.00000268, 0.00000455, -0.00000930, 0.00000384, -0.00001058, -0.00001780, 0.00002825, -0.00001901, -0.00002777, -0.00002682, -0.00015326, 0.00003233, 0.00042707, 0.00003233, -0.00015325, -0.00002680, -0.00002775, -0.00001898, 0.00002828, -0.00001777, -0.00001054, 0.00000388, -0.00000925, 0.00000460, -0.00000263, -0.00000154, 0.00000652, -0.00000501, -0.00001119, -0.00000716, 0.00000751, 0.00000189, -0.00000315, 0.00000056, -0.00000907, 0.00000908, -0.00000881, -0.00002161, 0.00002772, -0.00001335, -0.00002553, -0.00002845, -0.00013629, 0.00008686, 0.00026240, 0.00008685, -0.00013630, -0.00002847, -0.00002555, -0.00001337, 0.00002769, -0.00002165, -0.00000885, 0.00000904, -0.00000912, 0.00000051, -0.00000320, 0.00000184, 0.00000745, },
		{ -0.00000876, -0.00001169, 0.00000315, 0.00000387, -0.00000341, 0.00000107, -0.00000152, -0.00000700, 0.00000803, -0.00002837, 0.00001120, 0.00001327, -0.00003453, -0.00001328, -0.00008734, -0.00012113, 0.00027527, 0.00027527, -0.00012112, -0.00008733, -0.00001326, -0.00003450, 0.00001330, 0.00001124, -0.00002833, 0.00000807, -0.00000696, -0.00000147, 0.00000112, -0.00000335, 0.00000392, 0.00000320, -0.00001164, -0.00000871, -0.00000113, 0.00000923, -0.00000628, 0.00000374, -0.00000724, -0.00000059, 0.00000529, -0.00002361, 0.00000324, 0.00002218, -0.00004064, -0.00000276, -0.00010071, -0.00007795, 0.00023360, 0.00023360, -0.00007796, -0.00010072, -0.00000278, -0.00004066, 0.00002215, 0.00000320, -0.00002365, 0.00000525, -0.00000063, -0.00000729, 0.00000368, -0.00000633, 0.00000918, -0.00000118, },
		},
		{ { 0.00001689, 0.00001141, -0.00003608, -0.00000466, 0.00003981, -0.00000825, -0.00002701, -0.00000569, 0.00000719, 0.00001730, -0.00008455, 0.00000688, 0.00000305, -0.00001345, -0.00002783, -0.00002587, -0.00012414, 0.00004013, 0.00037905, 0.00005895, -0.00013506, -0.00001814, -0.00001879, -0.00001923, -0.00001132, -0.00001088, 0.00004573, -0.00000905, -0.00002796, -0.00000846, 0.00001744, -0.00000335, -0.00000068, -0.00000335, 0.00001744, -0.00000846, -0.00002796, -0.00000905, 0.00004573, -0.00001088, -0.00001132, -0.00001923, -0.00001879, -0.00001814, -0.00013506, 0.00005895, 0.00037905, 0.00004013, -0.00012414, -0.00002587, -0.00002783, -0.00001345, 0.00000305, 0.00000688, -0.00008455, 0.00001730, 0.00000719, -0.00000569, -0.00002701, -0.00000825, 0.00003981, -0.00000466, -0.00003608, 0.00001141, },
		{ 0.00001782, -0.00000839, -0.00003597, 0.00002328, 0.00002908, -0.00003019, -0.00001018, -0.00000952, 0.00002648, -0.00003136, -0.00006392, 0.00003128, -0.00002114, -0.00000759, -0.00003291, -0.00005241, -0.00010920, 0.00022759, 0.00027507, -0.00007120, -0.00008862, -0.00000067, -0.00002935, -0.00000625, -0.00002303, 0.00002101, 0.00002956, -0.00002493, -0.00002288, 0.00000720, 0.00000866, -0.00000253, -0.00000451, 0.00000703, 0.00001216, -0.00002026, -0.00002936, 0.00002115, 0.00003145, -0.00002540, -0.00000403, -0.00003060, -0.00000516, -0.00006533, -0.00013510, 0.00026820, 0.00031879, -0.00012947, -0.00006331, -0.00003189, -0.00001562, -0.00001804, 0.00002908, -0.00004974, -0.00005044, 0.00003190, -0.00000371, -0.00001266, -0.00002957, 0.00001744, 0.00003215, -0.00003142, -0.00001612, 0.00001808, },
		{ 0.00001442, -0.00003018, -0.00001571, 0.00003980, 0.00000264, -0.00003170, 0.00000008, -0.00000708, 0.00002806, -0.00007812, -0.00000633, 0.00000905, -0.00001840, -0.00001448, -0.00003548, -0.00009066, 0.00000698, 0.00030954, 0.00011198, -0.00010931, -0.00004086, -0.00000569, -0.00002601, -0.00000630, -0.00001774, 0.00004301, 0.00000026, -0.00002700, -0.00001320, 0.00001551, -0.00000145, 0.00000072, -0.00000744, 0.00001959, -0.00000388, -0.00002691, -0.00001855, 0.00004569, -0.00000214, -0.00001715, -0.00001207, -0.00002877, -0.00000540, -0.00013529, -0.00002391, 0.00044695, 0.00009301, -0.00016429, -0.00001684, -0.00003997, -0.00000988, -0.00000254, 0.00001777, -0.00008504, 0.00000245, 0.00001944, -0.00000640, -0.00002627, -0.00001376, 0.00003361, 0.00000833, -0.00004029, 0.00000742, 0.00001669, },
		{ -0.00000144, -0.00003676, 0.00001168, 0.00003838, -0.00002648, -0.00001221, -0.00000742, 0.00001238, -0.00000857, -0.00007680, 0.00003274, -0.00002508, 0.00000027, -0.00003259, -0.00004580, -0.00008767, 0.00015792, 0.00025748, -0.00001229, -0.00009446, -0.00000925, -0.00001942, -0.00001314, -0.00001732, 0.00000910, 0.00003624, -0.00001948, -0.00002431, 0.00000228, 0.00000930, -0.00000090, -0.00000479, 0.00000303, 0.00001772, -0.00001711, -0.00003212, 0.00001129, 0.00003963, -0.00002392, -0.00000672, -0.00002436, -0.00001833, -0.00003492, -0.00018671, 0.00022240, 0.00043418, -0.00014605, -0.00007888, -0.00003039, -0.00002299, -0.00001628, 0.00002669, -0.00003592, -0.00006388, 0.00002882, 0.00000717, -0.00001608, -0.00002806, 0.00000711, 0.00003588, -0.00002272, -0.00002428, 0.00001780, 0.00001698, },
		{ -0.00002369, -0.00002322, 0.00003343, 0.00001710, -0.00003747, 0.00000653, -0.00001761, 0.00003002, -0.00006338, -0.00002330, 0.00001753, -0.00002711, 0.00000225, -0.00004749, -0.00006089, -0.00001479, 0.00024843, 0.00013518, -0.00006756, -0.00006340, 0.00000293, -0.00002754, -0.00000492, -0.00002105, 0.00003713, 0.00000916, -0.00002423, -0.00001803, 0.00001426, -0.00000185, 0.00000423, -0.00001245, 0.00002081, 0.00000082, -0.00002398, -0.00002794, 0.00004358, 0.00000642, -0.00002066, -0.00000919, -0.00002967, -0.00000862, -0.00010830, -0.00012658, 0.00049573, 0.00017706, -0.00021530, -0.00000537, -0.00005415, -0.00000454, -0.00001032, 0.00002848, -0.00008283, -0.00001150, 0.00002503, 0.00000023, -0.00003010, -0.00001420, 0.00002312, 0.00002085, -0.00004162, 0.00000239, 0.00001608, 0.00001660, },
		{ -0.00003489, 0.00000023, 0.00004214, -0.00001636, -0.00002001, 0.00000215, -0.00000712, 0.00001334, -0.00008482, 0.00003104, -0.00002725, 0.00000488, -0.00002788, -0.00004560, -0.00006361, 0.00010303, 0.00023055, 0.00002881, -0.00007901, -0.00002961, -0.00000377, -0.00002190, -0.00001036, -0.00000270, 0.00004067, -0.00001361, -0.00002391, -0.00000338, 0.00001052, -0.00000149, -0.00000189, -0.00000398, 0.00002391, -0.00001394, -0.00003245, -0.00000008, 0.00004577, -0.00001981, -0.00001186, -0.00001539, -0.00003175, -0.00001092, -0.00020933, 0.00012930, 0.00056823, -0.00015045, -0.00010351, -0.00002590, -0.00003197, -0.00001356, 0.00002226, -0.00002171, -0.00007273, 0.00002062, 0.00001796, -0.00001592, -0.00002924, 0.00000105, 0.00003322, -0.00001045, -0.00003218, 0.00001698, 0.00001545, 0.00000473, },
		{ -0.00002668, 0.00002228, 0.00003164, -0.00004028, 0.00000887, -0.00001967, 0.00002078, -0.00004047, -0.00004335, 0.00002850, -0.00003959, 0.00002184, -0.00006091, -0.00003489, -0.00002866, 0.00019782, 0.00013663, -0.00002359, -0.00007512, 0.00000080, -0.00002076, -0.00000845, -0.00002025, 0.00002830, 0.00001737, -0.00002010, -0.00002242, 0.00001327, -0.00000345, 0.00000807, -0.00001635, 0.00001962, 0.00000609, -0.00001958, -0.00003663, 0.00003982, 0.00001331, -0.00001992, -0.00001295, -0.00001964, -0.00002719, -0.00006158, -0.00023007, 0.00050378, 0.00030280, -0.00027928, 0.00000999, -0.00007229, 0.00000431, -0.00002150, 0.00003999, -0.00008016, -0.00002130, 0.00002214, 0.00001333, -0.00003651, -0.00001192, 0.00001143, 0.00003029, -0.00003919, -0.00000356, 0.00001507, 0.00001803, -0.00001740, },
		{ -0.00000845, 0.00003858, 0.00000001, -0.00003203, 0.00001561, -0.00002620, 0.00002855, -0.00008404, 0.00002342, -0.00002458, 0.00000287, -0.00001569, -0.00005396, -0.00003519, 0.00005634, 0.00020532, 0.00004700, -0.00004600, -0.00005522, 0.00001159, -0.00002821, -0.00000481, -0.00001263, 0.00004235, -0.00000783, -0.00002124, -0.00001034, 0.00001346, -0.00000526, 0.00000448, -0.00001350, 0.00003013, -0.00001112, -0.00002969, -0.00001328, 0.00005085, -0.00001559, -0.00001539, -0.00001018, -0.00003606, -0.00000539, -0.00019242, -0.00000873, 0.00070506, -0.00012783, -0.00014524, -0.00001297, -0.00004737, -0.00000558, 0.00001234, -0.00000469, -0.00008006, 0.00001242, 0.00002307, -0.00000837, -0.00003538, 0.00000068, 0.00002450, 0.00000394, -0.00003893, 0.00001571, 0.00001304, 0.00001041, -0.00003178, },
		{ 0.00000922, 0.00004220, -0.00003645, 0.00000377, -0.00001135, 0.00000146, -0.00001271, -0.00006326, 0.00003958, -0.00005322, 0.00004101, -0.00007204, -0.00001562, -0.00003392, 0.00015404, 0.00012834, 0.00000934, -0.00006880, -0.00001444, -0.00000591, -0.00001589, -0.00001620, 0.00001746, 0.00002453, -0.00001531, -0.00002541, 0.00001144, -0.00000439, 0.00000992, -0.00001697, 0.00001478, 0.00001226, -0.00001442, -0.00004357, 0.00003418, 0.00001833, -0.00001519, -0.00002223, -0.00000198, -0.00005369, -0.00001017, -0.00031063, 0.00045122, 0.00047455, -0.00035240, 0.00002636, -0.00009264, 0.00001506, -0.00003428, 0.00005082, -0.00007686, -0.00002672, 0.00001258, 0.00002888, -0.00004143, -0.00001044, 0.00000199, 0.00003448, -0.00003267, -0.00001016, 0.00001380, 0.00001855, -0.00001150, -0.00002676, },
		{ 0.00002789, 0.00002009, -0.00004442, 0.00002754, -0.00003848, 0.00003220, -0.00007206, 0.00000856, -0.00001536, -0.00000738, 0.00000498, -0.00007085, -0.00000265, 0.00001461, 0.00018629, 0.00004543, -0.00000528, -0.00007654, 0.00001982, -0.00002804, -0.00000284, -0.00001941, 0.00004108, -0.00000255, -0.00001623, -0.00001863, 0.00001827, -0.00001187, 0.00001327, -0.00002397, 0.00003522, -0.00000875, -0.00002344, -0.00002841, 0.00005600, -0.00001390, -0.00001332, -0.00001427, -0.00002464, -0.00002461, -0.00013568, -0.00017658, 0.00081932, -0.00005953, -0.00021322, 0.00001443, -0.00007430, 0.00001197, -0.00000629, 0.00001753, -0.00008914, 0.00000895, 0.00001853, 0.00000797, -0.00004651, 0.00000539, 0.00001153, 0.00001839, -0.00004370, 0.00001416, 0.00000952, 0.00001590, -0.00002872, -0.00001276, },
		{ 0.00004534, -0.00002400, -0.00000938, 0.00000565, -0.00002319, 0.00001399, -0.00007812, 0.00004682, -0.00006373, 0.00005492, -0.00007593, -0.00000715, -0.00002797, 0.00011135, 0.00012157, 0.00002271, -0.00004321, -0.00004009, 0.00001327, -0.00002424, -0.00001078, 0.00000595, 0.00003030, -0.00001071, -0.00002586, 0.00000735, -0.00000269, 0.00000775, -0.00001280, 0.00000575, 0.00001940, -0.00000949, -0.00004724, 0.00002567, 0.00002276, -0.00000911, -0.00003259, 0.00001581, -0.00007614, 0.00002753, -0.00034597, 0.00032647, 0.00068568, -0.00042226, 0.00003493, -0.00010872, 0.00002203, -0.00004351, 0.00005671, -0.00007017, -0.00003056, 0.00000117, 0.00004089, -0.00004012, -0.00001329, -0.00000235, 0.00003220, -0.00002239, -0.00001706, 0.00001261, 0.00001773, -0.00000565, -0.00002497, -0.00000240, },
		{ 0.00003969, -0.00005232, 0.00003255, -0.00003927, 0.00002234, -0.00004959, -0.00001218, -0.00000058, -0.00002460, 0.00003194, -0.00009302, 0.00003060, -0.00002117, 0.00017153, 0.00003333, 0.00003061, -0.00008442, 0.00001585, -0.00001937, -0.00000508, -0.00002271, 0.00003712, 0.00000192, -0.00000934, -0.00002746, 0.00002399, -0.00001965, 0.00002226, -0.00003307, 0.00003781, -0.00000674, -0.00001385, -0.00004498, 0.00006167, -0.00001632, -0.00000365, -0.00002979, 0.00000357, -0.00006637, -0.00005036, -0.00034795, 0.00088052, 0.00007193, -0.00031340, 0.00005946, -0.00011527, 0.00004080, -0.00003442, 0.00004533, -0.00010150, 0.00001255, 0.00000366, 0.00003124, -0.00006010, 0.00001270, -0.00000278, 0.00003063, -0.00004583, 0.00001255, 0.00000487, 0.00002128, -0.00002639, -0.00001246, 0.00001220, },
		{ -0.00000354, -0.00002807, 0.00002639, -0.00004617, 0.00003290, -0.00008304, 0.00004647, -0.00006705, 0.00005933, -0.00006868, -0.00001230, -0.00000921, 0.00006562, 0.00012294, 0.00001518, -0.00000373, -0.00006907, 0.00003071, -0.00002957, -0.00000615, -0.00000490, 0.00003453, -0.00000720, -0.00002279, -0.00000007, 0.00000302, 0.00000055, -0.00000351, -0.00000699, 0.00002721, -0.00000593, -0.00004590, 0.00001282, 0.00002893, -0.00000585, -0.00003771, 0.00002444, -0.00008184, 0.00003534, -0.00032187, 0.00013175, 0.00091598, -0.00046767, 0.00002166, -0.00011026, 0.00001662, -0.00004182, 0.00005182, -0.00005579, -0.00003742, -0.00000586, 0.00004330, -0.00002877, -0.00002267, -0.00000016, 0.00002351, -0.00000930, -0.00002393, 0.00001206, 0.00001493, 0.00000085, -0.00002316, -0.00000968, 0.00003930, },
		{ -0.00005154, 0.00002722, -0.00002727, 0.00000104, -0.00002086, -0.00003453, 0.00001612, -0.00004464, 0.00006011, -0.00011459, 0.00005871, -0.00004680, 0.00015513, 0.00002212, 0.00005061, -0.00007336, -0.00000157, -0.00000319, -0.00001027, -0.00002327, 0.00003113, 0.00000548, -0.00000159, -0.00003534, 0.00002876, -0.00002610, 0.00002867, -0.00003836, 0.00003663, -0.00000481, -0.00000186, -0.00006143, 0.00006699, -0.00002231, 0.00001241, -0.00005420, 0.00004305, -0.00011985, 0.00004273, -0.00049046, 0.00086003, 0.00027633, -0.00044382, 0.00011899, -0.00016739, 0.00007784, -0.00006874, 0.00007572, -0.00011555, 0.00002204, -0.00001835, 0.00005645, -0.00007168, 0.00001901, -0.00001512, 0.00003868, -0.00004497, 0.00001098, -0.00000057, 0.00002615, -0.00002468, -0.00000868, -0.00000495, 0.00005408, },
		{ -0.00004744, 0.00004428, -0.00006035, 0.00003876, -0.00007527, 0.00003685, -0.00006128, 0.00005249, -0.00004932, -0.00003073, 0.00002085, 0.00001670, 0.00013280, -0.00000746, 0.00003928, -0.00009212, 0.00004007, -0.00002856, -0.00000384, -0.00001423, 0.00003731, -0.00000551, -0.00001574, -0.00001110, 0.00001287, -0.00001125, 0.00000979, -0.00002196, 0.00003505, -0.00000477, -0.00003806, -0.00000563, 0.00003920, -0.00000959, -0.00003173, 0.00001620, -0.00006183, 0.00000457, -0.00023709, -0.00011452, 0.00113268, -0.00046160, -0.00002951, -0.00008614, -0.00001001, -0.00002220, 0.00003092, -0.00002984, -0.00005166, -0.00000315, 0.00003205, -0.00000610, -0.00003865, 0.00000816, 0.00000974, 0.00000525, -0.00003055, 0.00001283, 0.00000951, 0.00000864, -0.00002287, -0.00001094, 0.00002456, 0.00002175, },
		{ 0.00001139, -0.00000526, -0.00002584, 0.00000725, -0.00005253, 0.00002957, -0.00006189, 0.00008323, -0.00012891, 0.00007545, -0.00005749, 0.00013071, 0.00002116, 0.00004875, -0.00004364, -0.00002902, 0.00001648, -0.00001563, -0.00002244, 0.00002389, 0.00000834, 0.00000560, -0.00004034, 0.00003031, -0.00002855, 0.00003000, -0.00003797, 0.00003094, -0.00000253, 0.00001074, -0.00007511, 0.00006963, -0.00002905, 0.00003032, -0.00008044, 0.00008277, -0.00016836, 0.00011837, -0.00057314, 0.00073911, 0.00054992, -0.00059118, 0.00018158, -0.00022121, 0.00011460, -0.00010152, 0.00010220, -0.00012654, 0.00003280, -0.00004114, 0.00007667, -0.00007608, 0.00002064, -0.00002262, 0.00004116, -0.00004104, 0.00000932, -0.00000596, 0.00002957, -0.00002275, -0.00000353, -0.00001962, 0.00005938, -0.00003997, },
		{ 0.00005319, -0.00006093, 0.00002982, -0.00005567, 0.00001955, -0.00004775, 0.00003620, -0.00002071, -0.00005824, 0.00005700, -0.00003123, 0.00014586, -0.00003464, 0.00007382, -0.00010092, 0.00003702, -0.00001991, -0.00000401, -0.00002189, 0.00003893, -0.00000601, -0.00000507, -0.00002490, 0.00002560, -0.00002571, 0.00002479, -0.00003694, 0.00004200, -0.00000674, -0.00002310, -0.00002986, 0.00005475, -0.00002274, -0.00001164, -0.00001202, -0.00001437, -0.00006262, -0.00010518, -0.00037937, 0.00129547, -0.00037721, -0.00013139, -0.00002874, -0.00006319, 0.00001890, -0.00000797, 0.00000882, -0.00007524, 0.00001171, 0.00000672, 0.00002584, -0.00005880, 0.00002048, -0.00000679, 0.00001968, -0.00003681, 0.00001548, 0.00000120, 0.00001795, -0.00002482, -0.00000613, 0.00000378, 0.00004691, -0.00006170, },
		{ 0.00002064, -0.00005030, 0.00002726, -0.00006050, 0.00003514, -0.00007140, 0.00009607, -0.00013097, 0.00007665, -0.00005059, 0.00009474, 0.00003461, 0.00002600, -0.00000149, -0.00005930, 0.00003383, -0.00001787, -0.00002161, 0.00001601, 0.00001098, 0.00001062, -0.00004060, 0.00002671, -0.00002506, 0.00002474, -0.00003109, 0.00002071, 0.00000063, 0.00002151, -0.00008264, 0.00006626, -0.00003209, 0.00004335, -0.00009896, 0.00010936, -0.00019408, 0.00015362, -0.00057444, 0.00051554, 0.00087191, -0.00073023, 0.00022811, -0.00026178, 0.00013847, -0.00012206, 0.00011613, -0.00012781, 0.00003824, -0.00005697, 0.00008494, -0.00006903, 0.00001498, -0.00002347, 0.00003760, -0.00003419, 0.00000708, -0.00001005, 0.00003024, -0.00001926, 0.00000054, -0.00002841, 0.00005367, -0.00001844, -0.00001167, },
		{ -0.00004709, 0.00000881, -0.00002900, -0.00000061, -0.00003082, 0.00001542, 0.00001120, -0.00008779, 0.00009153, -0.00007121, 0.00015393, -0.00005481, 0.00009004, -0.00009090, 0.00002093, -0.00000508, -0.00000531, -0.00002839, 0.00003974, -0.00000851, 0.00000792, -0.00003949, 0.00003876, -0.00003989, 0.00003846, -0.00004944, 0.00004710, -0.00001190, -0.00000180, -0.00005829, 0.00007460, -0.00004461, 0.00002107, -0.00005699, 0.00005369, -0.00015301, 0.00004783, -0.00062109, 0.00136461, -0.00019554, -0.00028773, 0.00006176, -0.00014144, 0.00007902, -0.00006166, 0.00005708, -0.00010629, 0.00003704, -0.00002893, 0.00006174, -0.00007872, 0.00003349, -0.00002329, 0.00003249, -0.00004254, 0.00002016, -0.00000963, 0.00002826, -0.00002869, 0.00000318, -0.00001890, 0.00006657, -0.00006605, 0.00004954, },
		{ -0.00006461, 0.00003360, -0.00005529, 0.00003072, -0.00007087, 0.00009653, -0.00011935, 0.00006213, -0.00002750, 0.00004847, 0.00006021, -0.00001021, 0.00004265, -0.00008359, 0.00004312, -0.00001456, -0.00002143, 0.00000765, 0.00001409, 0.00001211, -0.00003495, 0.00001703, -0.00001501, 0.00001287, -0.00001826, 0.00000673, 0.00000519, 0.00002775, -0.00008065, 0.00005362, -0.00002679, 0.00004463, -0.00010062, 0.00011122, -0.00018343, 0.00013432, -0.00048839, 0.00020711, 0.00120510, -0.00082682, 0.00023543, -0.00027205, 0.00013596, -0.00011962, 0.00010931, -0.00011309, 0.00003205, -0.00005906, 0.00007636, -0.00004865, 0.00000136, -0.00001737, 0.00002837, -0.00002471, 0.00000352, -0.00001133, 0.00002683, -0.00001298, 0.00000150, -0.00002940, 0.00003756, 0.00000938, -0.00003614, 0.00004279, },
		{ -0.00001757, -0.00000280, -0.00001692, -0.00001644, -0.00000345, 0.00003926, -0.00011168, 0.00011658, -0.00009635, 0.00014942, -0.00005934, 0.00008335, -0.00006300, -0.00000455, 0.00001196, -0.00000537, -0.00003446, 0.00003988, -0.00001222, 0.00002117, -0.00005217, 0.00004926, -0.00005052, 0.00004785, -0.00005726, 0.00004943, -0.00001954, 0.00002341, -0.00008732, 0.00009529, -0.00007107, 0.00006022, -0.00010940, 0.00012813, -0.00024523, 0.00018917, -0.00079792, 0.00131036, 0.00008723, -0.00048871, 0.00017463, -0.00023478, 0.00014868, -0.00012130, 0.00010731, -0.00013895, 0.00006713, -0.00006760, 0.00009422, -0.00009317, 0.00004360, -0.00003706, 0.00004233, -0.00004746, 0.00002638, -0.00002179, 0.00003824, -0.00003315, 0.00001433, -0.00003893, 0.00007636, -0.00005816, 0.00003349, -0.00002250, },
		{ 0.00002468, -0.00003782, 0.00001797, -0.00006172, 0.00008646, -0.00009686, 0.00003600, 0.00000612, -0.00000220, 0.00009053, -0.00004848, 0.00007733, -0.00009427, 0.00004077, -0.00000519, -0.00002142, -0.00000148, 0.00001833, 0.00000928, -0.00002327, 0.00000176, 0.00000056, -0.00000404, -0.00000129, -0.00000955, 0.00001147, 0.00002717, -0.00006686, 0.00002978, -0.00001006, 0.00002955, -0.00007984, 0.00008236, -0.00013153, 0.00005940, -0.00032721, -0.00014973, 0.00150106, -0.00084449, 0.00018227, -0.00023801, 0.00009698, -0.00008706, 0.00007701, -0.00007905, 0.00001057, -0.00004381, 0.00004975, -0.00001628, -0.00001868, -0.00000547, 0.00001464, -0.00001299, -0.00000217, -0.00000852, 0.00001847, -0.00000321, -0.00000167, -0.00002252, 0.00001397, 0.00003800, -0.00005547, 0.00005445, -0.00006394, },
		{ 0.00001494, -0.00002316, -0.00000984, -0.00001488, 0.00005767, -0.00012408, 0.00012678, -0.00010262, 0.00012851, -0.00004534, 0.00005591, -0.00002362, -0.00003249, 0.00002589, -0.00000175, -0.00004053, 0.00003904, -0.00001583, 0.00003224, -0.00006013, 0.00005426, -0.00005492, 0.00005076, -0.00005893, 0.00004830, -0.00002810, 0.00004859, -0.00011180, 0.00011148, -0.00009531, 0.00009625, -0.00015608, 0.00019098, -0.00031534, 0.00028721, -0.00087757, 0.00112131, 0.00045500, -0.00070940, 0.00028871, -0.00032555, 0.00021263, -0.00017381, 0.00014879, -0.00016473, 0.00009375, -0.00010030, 0.00011558, -0.00009746, 0.00004784, -0.00004598, 0.00004820, -0.00005094, 0.00003292, -0.00003332, 0.00004593, -0.00003626, 0.00002425, -0.00005245, 0.00007417, -0.00003899, 0.00000901, 0.00000565, -0.00004053, },
		{ -0.00001323, 0.00000211, -0.00004874, 0.00007115, -0.00006971, 0.00000557, 0.00004180, -0.00004870, 0.00011575, -0.00007721, 0.00009358, -0.00008744, 0.00002668, 0.00000831, -0.00002004, -0.00001190, 0.00002402, 0.00000226, -0.00000675, -0.00001713, 0.00001916, -0.00002319, 0.00001717, -0.00002627, 0.00001949, 0.00001856, -0.00004098, -0.00000470, 0.00001820, -0.00000239, -0.00003678, 0.00002476, -0.00004421, -0.00005819, -0.00011941, -0.00050588, 0.00170899, -0.00075314, 0.00005620, -0.00015390, 0.00001899, -0.00002401, 0.00002030, -0.00002720, -0.00002553, -0.00001211, 0.00000829, 0.00002354, -0.00004165, 0.00000990, -0.00000190, 0.00000044, -0.00001053, -0.00000098, 0.00000516, 0.00000975, -0.00000870, -0.00000949, -0.00001259, 0.00006155, -0.00006436, 0.00005195, -0.00004801, 0.00000382, },
		{ -0.00001589, -0.00001340, -0.00001632, 0.00006401, -0.00012309, 0.00012108, -0.00009061, 0.00009268, -0.00001629, 0.00001580, 0.00001724, -0.00005477, 0.00003183, 0.00000698, -0.00004624, 0.00003650, -0.00001781, 0.00003887, -0.00006120, 0.00005189, -0.00005169, 0.00004633, -0.00005403, 0.00004333, -0.00003548, 0.00006896, -0.00012601, 0.00011728, -0.00010972, 0.00011878, -0.00018363, 0.00022542, -0.00034321, 0.00031927, -0.00084494, 0.00080910, 0.00087144, -0.00091234, 0.00037591, -0.00039193, 0.00025328, -0.00020501, 0.00017043, -0.00017488, 0.00010851, -0.00011867, 0.00011985, -0.00008879, 0.00004465, -0.00004882, 0.00004940, -0.00005212, 0.00003801, -0.00004184, 0.00004921, -0.00003607, 0.00003036, -0.00005720, 0.00006058, -0.00001253, -0.00001731, 0.00002886, -0.00005198, 0.00001848, },
		{ -0.00000966, -0.00003836, 0.00005741, -0.00004532, -0.00002113, 0.00007090, -0.00008270, 0.00012734, -0.00008824, 0.00008773, -0.00006437, 0.00000456, 0.00002201, -0.00001518, -0.00002395, 0.00003089, -0.00000782, 0.00001225, -0.00003657, 0.00003740, -0.00004125, 0.00003423, -0.00004144, 0.00002877, 0.00000228, -0.00000534, -0.00004623, 0.00005446, -0.00004690, 0.00002202, -0.00005155, 0.00006298, -0.00019444, 0.00009631, -0.00080914, 0.00178643, -0.00053776, -0.00014064, -0.00002594, -0.00009064, 0.00006169, -0.00005302, 0.00003557, -0.00007088, 0.00003060, -0.00004100, 0.00006392, -0.00006300, 0.00002587, -0.00001929, 0.00001480, -0.00002150, 0.00001093, -0.00001196, 0.00002453, -0.00001812, 0.00000674, -0.00003724, 0.00007544, -0.00006039, 0.00003589, -0.00002141, -0.00002154, 0.00001045, },
		{ -0.00002537, -0.00000934, 0.00006015, -0.00011130, 0.00010311, -0.00006562, 0.00004837, 0.00001930, -0.00002552, 0.00004896, -0.00006462, 0.00002704, 0.00002033, -0.00005024, 0.00003120, -0.00001680, 0.00003951, -0.00005440, 0.00004185, -0.00004106, 0.00003515, -0.00004319, 0.00003455, -0.00003940, 0.00007995, -0.00012518, 0.00010805, -0.00010815, 0.00011969, -0.00018232, 0.00022056, -0.00031808, 0.00027734, -0.00070601, 0.00040819, 0.00128462, -0.00105414, 0.00040749, -0.00041350, 0.00025543, -0.00020362, 0.00016408, -0.00016323, 0.00010539, -0.00011728, 0.00010445, -0.00006717, 0.00003422, -0.00004541, 0.00004564, -0.00004993, 0.00003964, -0.00004508, 0.00004637, -0.00003123, 0.00003116, -0.00005290, 0.00003860, 0.00001538, -0.00003826, 0.00003996, -0.00004721, 0.00000648, 0.00000389, },
		{ -0.00003622, 0.00005100, -0.00002969, -0.00003798, 0.00008731, -0.00009904, 0.00012095, -0.00007926, 0.00006241, -0.00003124, -0.00001906, 0.00003110, -0.00000512, -0.00003726, 0.00003795, -0.00001907, 0.00003070, -0.00005308, 0.00005185, -0.00005513, 0.00004734, -0.00005327, 0.00003832, -0.00001954, 0.00003520, -0.00008850, 0.00009207, -0.00009561, 0.00008507, -0.00013083, 0.00016851, -0.00032016, 0.00027976, -0.00101503, 0.00170912, -0.00020465, -0.00038873, 0.00012688, -0.00021484, 0.00015479, -0.00012962, 0.00009818, -0.00011659, 0.00007568, -0.00008889, 0.00009727, -0.00007822, 0.00003967, -0.00003557, 0.00002902, -0.00003420, 0.00002567, -0.00003063, 0.00003885, -0.00002767, 0.00002280, -0.00005589, 0.00007750, -0.00004474, 0.00001103, 0.00000764, -0.00004198, 0.00002483, -0.00001063, },
		{ 0.00000094, 0.00005155, -0.00009481, 0.00007985, -0.00003600, 0.00000480, 0.00005094, -0.00005684, 0.00006360, -0.00005887, 0.00001207, 0.00003576, -0.00005048, 0.00002224, -0.00001207, 0.00003376, -0.00004031, 0.00002551, -0.00002486, 0.00001906, -0.00002791, 0.00002239, -0.00003800, 0.00007827, -0.00010692, 0.00008209, -0.00008797, 0.00009558, -0.00014904, 0.00017469, -0.00024158, 0.00016996, -0.00048690, -0.00002976, 0.00163608, -0.00109470, 0.00036171, -0.00037724, 0.00021098, -0.00016490, 0.00012734, -0.00012840, 0.00008268, -0.00009514, 0.00007095, -0.00003553, 0.00001840, -0.00003648, 0.00003700, -0.00004351, 0.00003617, -0.00004151, 0.00003668, -0.00002148, 0.00002653, -0.00004114, 0.00001280, 0.00003868, -0.00004830, 0.00003541, -0.00002663, -0.00001718, 0.00003068, -0.00004015, },
		{ 0.00005446, -0.00002522, -0.00004286, 0.00008935, -0.00009747, 0.00009787, -0.00005429, 0.00002573, 0.00000271, -0.00003685, 0.00003155, 0.00001057, -0.00005053, 0.00004362, -0.00002917, 0.00004550, -0.00006361, 0.00005986, -0.00006271, 0.00005479, -0.00006043, 0.00004668, -0.00004329, 0.00007400, -0.00012373, 0.00012280, -0.00013818, 0.00013875, -0.00019565, 0.00025044, -0.00040829, 0.00039836, -0.00109612, 0.00147747, 0.00021679, -0.00065282, 0.00027567, -0.00033005, 0.00023583, -0.00019359, 0.00014803, -0.00015251, 0.00011351, -0.00012613, 0.00011706, -0.00008403, 0.00004918, -0.00004901, 0.00004169, -0.00004692, 0.00004075, -0.00004787, 0.00005005, -0.00003493, 0.00003573, -0.00006613, 0.00006839, -0.00002177, -0.00001525, 0.00002985, -0.00004915, 0.00002403, 0.00000287, -0.00004483, },
		{ 0.00004517, -0.00008095, 0.00005912, -0.00001047, -0.00002911, 0.00006970, -0.00007061, 0.00005823, -0.00003908, -0.00000915, 0.00004914, -0.00004483, 0.00000927, -0.00000380, 0.00002251, -0.00002097, 0.00000564, -0.00000602, 0.00000067, -0.00001020, 0.00000774, -0.00003040, 0.00006293, -0.00007226, 0.00004162, -0.00005119, 0.00004911, -0.00008848, 0.00009592, -0.00012742, 0.00001994, -0.00022804, -0.00044567, 0.00187220, -0.00100700, 0.00023078, -0.00028221, 0.00012217, -0.00009281, 0.00006489, -0.00007472, 0.00004364, -0.00005599, 0.00002483, 0.00000088, 0.00000019, -0.00002344, 0.00002403, -0.00003246, 0.00002685, -0.00003085, 0.00002073, -0.00000779, 0.00001771, -0.00002477, -0.00001201, 0.00005283, -0.00004513, 0.00001647, 0.00000414, -0.00004435, 0.00005599, -0.00004998, 0.00000762, },
		{ -0.00002994, -0.00003820, 0.00008010, -0.00008262, 0.00006443, -0.00002207, -0.00001139, 0.00002797, -0.00004301, 0.00002154, 0.00003044, -0.00006154, 0.00004604, -0.00003595, 0.00005425, -0.00006623, 0.00006017, -0.00006323, 0.00005595, -0.00006212, 0.00005219, -0.00006456, 0.00010402, -0.00014450, 0.00013904, -0.00016506, 0.00017093, -0.00023148, 0.00029203, -0.00044056, 0.00043428, -0.00104762, 0.00111775, 0.00067761, -0.00088773, 0.00038770, -0.00041169, 0.00028596, -0.00023052, 0.00017427, -0.00017015, 0.00013607, -0.00014572, 0.00011943, -0.00007914, 0.00005335, -0.00005817, 0.00005119, -0.00005738, 0.00005318, -0.00006056, 0.00005569, -0.00003799, 0.00004365, -0.00006756, 0.00005121, 0.00000235, -0.00003517, 0.00003771, -0.00003858, 0.00000695, 0.00002981, -0.00006231, 0.00006602, },
		{ -0.00007559, 0.00004689, 0.00000472, -0.00004764, 0.00007102, -0.00006503, 0.00003581, -0.00001115, -0.00003043, 0.00005596, -0.00003193, -0.00000701, 0.00000680, 0.00000777, 0.00000051, -0.00001428, 0.00001218, -0.00001726, 0.00000777, -0.00000809, -0.00001705, 0.00003570, -0.00002584, -0.00000731, -0.00000422, -0.00001154, -0.00001209, 0.00000022, 0.00000217, -0.00014143, 0.00002487, -0.00078472, 0.00195541, -0.00078472, 0.00002487, -0.00014143, 0.00000217, 0.00000022, -0.00001209, -0.00001154, -0.00000422, -0.00000731, -0.00002584, 0.00003570, -0.00001705, -0.00000809, 0.00000777, -0.00001726, 0.00001218, -0.00001428, 0.00000051, 0.00000777, 0.00000680, -0.00000701, -0.00003193, 0.00005596, -0.00003043, -0.00001115, 0.00003581, -0.00006503, 0.00007102, -0.00004764, 0.00000472, 0.00004689, },
		{ -0.00002994, 0.00006602, -0.00006231, 0.00002981, 0.00000695, -0.00003858, 0.00003771, -0.00003517, 0.00000235, 0.00005121, -0.00006756, 0.00004365, -0.00003799, 0.00005569, -0.00006056, 0.00005318, -0.00005738, 0.00005119, -0.00005817, 0.00005335, -0.00007914, 0.00011943, -0.00014572, 0.00013607, -0.00017015, 0.00017427, -0.00023052, 0.00028596, -0.00041169, 0.00038770, -0.00088773, 0.00067761, 0.00111775, -0.00104762, 0.00043428, -0.00044056, 0.00029203, -0.00023148, 0.00017093, -0.00016506, 0.00013904, -0.00014450, 0.00010402, -0.00006456, 0.00005219, -0.00006212, 0.00005595, -0.00006323, 0.00006017, -0.00006623, 0.00005425, -0.00003595, 0.00004604, -0.00006154, 0.00003044, 0.00002154, -0.00004301, 0.00002797, -0.00001139, -0.00002207, 0.00006443, -0.00008262, 0.00008010, -0.00003820, },
		{ 0.00004517, 0.00000762, -0.00004998, 0.00005599, -0.00004435, 0.00000414, 0.00001647, -0.00004513, 0.00005283, -0.00001201, -0.00002477, 0.00001771, -0.00000779, 0.00002073, -0.00003085, 0.00002685, -0.00003246, 0.00002403, -0.00002344, 0.00000019, 0.00000088, 0.00002483, -0.00005599, 0.00004364, -0.00007472, 0.00006489, -0.00009281, 0.00012217, -0.00028221, 0.00023078, -0.00100700, 0.00187220, -0.00044567, -0.00022804, 0.00001994, -0.00012742, 0.00009592, -0.00008848, 0.00004911, -0.00005119, 0.00004162, -0.00007226, 0.00006293, -0.00003040, 0.00000774, -0.00001020, 0.00000067, -0.00000602, 0.00000564, -0.00002097, 0.00002251, -0.00000380, 0.00000927, -0.00004483, 0.00004914, -0.00000915, -0.00003908, 0.00005823, -0.00007061, 0.00006970, -0.00002911, -0.00001047, 0.00005912, -0.00008095, },
		{ 0.00005446, -0.00004483, 0.00000287, 0.00002403, -0.00004915, 0.00002985, -0.00001525, -0.00002177, 0.00006839, -0.00006613, 0.00003573, -0.00003493, 0.00005005, -0.00004787, 0.00004075, -0.00004692, 0.00004169, -0.00004901, 0.00004918, -0.00008403, 0.00011706, -0.00012613, 0.00011351, -0.00015251, 0.00014803, -0.00019359, 0.00023583, -0.00033005, 0.00027567, -0.00065282, 0.00021679, 0.00147747, -0.00109612, 0.00039836, -0.00040829, 0.00025044, -0.00019565, 0.00013875, -0.00013818, 0.00012280, -0.00012373, 0.00007400, -0.00004329, 0.00004668, -0.00006043, 0.00005479, -0.00006271, 0.00005986, -0.00006361, 0.00004550, -0.00002917, 0.00004362, -0.00005053, 0.00001057, 0.00003155, -0.00003685, 0.00000271, 0.00002573, -0.00005429, 0.00009787, -0.00009747, 0.00008935, -0.00004286, -0.00002522, },
		{ 0.00000094, -0.00004015, 0.00003068, -0.00001718, -0.00002663, 0.00003541, -0.00004830, 0.00003868, 0.00001280, -0.00004114, 0.00002653, -0.00002148, 0.00003668, -0.00004151, 0.00003617, -0.00004351, 0.00003700, -0.00003648, 0.00001840, -0.00003553, 0.00007095, -0.00009514, 0.00008268, -0.00012840, 0.00012734, -0.00016490, 0.00021098, -0.00037724, 0.00036171, -0.00109470, 0.00163608, -0.00002976, -0.00048690, 0.00016996, -0.00024158, 0.00017469, -0.00014904, 0.00009558, -0.00008797, 0.00008209, -0.00010692, 0.00007827, -0.00003800, 0.00002239, -0.00002791, 0.00001906, -0.00002486, 0.00002551, -0.00004031, 0.00003376, -0.00001207, 0.00002224, -0.00005048, 0.00003576, 0.00001207, -0.00005887, 0.00006360, -0.00005684, 0.00005094, 0.00000480, -0.00003600, 0.00007985, -0.00009481, 0.00005155, },
		{ -0.00003622, -0.00001063, 0.00002483, -0.00004198, 0.00000764, 0.00001103, -0.00004474, 0.00007750, -0.00005589, 0.00002280, -0.00002767, 0.00003885, -0.00003063, 0.00002567, -0.00003420, 0.00002902, -0.00003557, 0.00003967, -0.00007822, 0.00009727, -0.00008889, 0.00007568, -0.00011659, 0.00009818, -0.00012962, 0.00015479, -0.00021484, 0.00012688, -0.00038873, -0.00020465, 0.00170912, -0.00101503, 0.00027976, -0.00032016, 0.00016851, -0.00013083, 0.00008507, -0.00009561, 0.00009207, -0.00008850, 0.00003520, -0.00001954, 0.00003832, -0.00005327, 0.00004734, -0.00005513, 0.00005185, -0.00005308, 0.00003070, -0.00001907, 0.00003795, -0.00003726, -0.00000512, 0.00003110, -0.00001906, -0.00003124, 0.00006241, -0.00007926, 0.00012095, -0.00009904, 0.00008731, -0.00003798, -0.00002969, 0.00005100, },
		{ -0.00002537, 0.00000389, 0.00000648, -0.00004721, 0.00003996, -0.00003826, 0.00001538, 0.00003860, -0.00005290, 0.00003116, -0.00003123, 0.00004637, -0.00004508, 0.00003964, -0.00004993, 0.00004564, -0.00004541, 0.00003422, -0.00006717, 0.00010445, -0.00011728, 0.00010539, -0.00016323, 0.00016408, -0.00020362, 0.00025543, -0.00041350, 0.00040749, -0.00105414, 0.00128462, 0.00040819, -0.00070601, 0.00027734, -0.00031808, 0.00022056, -0.00018232, 0.00011969, -0.00010815, 0.00010805, -0.00012518, 0.00007995, -0.00003940, 0.00003455, -0.00004319, 0.00003515, -0.00004106, 0.00004185, -0.00005440, 0.00003951, -0.00001680, 0.00003120, -0.00005024, 0.00002033, 0.00002704, -0.00006462, 0.00004896, -0.00002552, 0.00001930, 0.00004837, -0.00006562, 0.00010311, -0.00011130, 0.00006015, -0.00000934, },
		{ -0.00000966, 0.00001045, -0.00002154, -0.00002141, 0.00003589, -0.00006039, 0.00007544, -0.00003724, 0.00000674, -0.00001812, 0.00002453, -0.00001196, 0.00001093, -0.00002150, 0.00001480, -0.00001929, 0.00002587, -0.00006300, 0.00006392, -0.00004100, 0.00003060, -0.00007088, 0.00003557, -0.00005302, 0.00006169, -0.00009064, -0.00002594, -0.00014064, -0.00053776, 0.00178643, -0.00080914, 0.00009631, -0.00019444, 0.00006298, -0.00005155, 0.00002202, -0.00004690, 0.00005446, -0.00004623, -0.00000534, 0.00000228, 0.00002877, -0.00004144, 0.00003423, -0.00004125, 0.00003740, -0.00003657, 0.00001225, -0.00000782, 0.00003089, -0.00002395, -0.00001518, 0.00002201, 0.00000456, -0.00006437, 0.00008773, -0.00008824, 0.00012734, -0.00008270, 0.00007090, -0.00002113, -0.00004532, 0.00005741, -0.00003836, },
		{ -0.00001589, 0.00001848, -0.00005198, 0.00002886, -0.00001731, -0.00001253, 0.00006058, -0.00005720, 0.00003036, -0.00003607, 0.00004921, -0.00004184, 0.00003801, -0.00005212, 0.00004940, -0.00004882, 0.00004465, -0.00008879, 0.00011985, -0.00011867, 0.00010851, -0.00017488, 0.00017043, -0.00020501, 0.00025328, -0.00039193, 0.00037591, -0.00091234, 0.00087144, 0.00080910, -0.00084494, 0.00031927, -0.00034321, 0.00022542, -0.00018363, 0.00011878, -0.00010972, 0.00011728, -0.00012601, 0.00006896, -0.00003548, 0.00004333, -0.00005403, 0.00004633, -0.00005169, 0.00005189, -0.00006120, 0.00003887, -0.00001781, 0.00003650, -0.00004624, 0.00000698, 0.00003183, -0.00005477, 0.00001724, 0.00001580, -0.00001629, 0.00009268, -0.00009061, 0.00012108, -0.00012309, 0.00006401, -0.00001632, -0.00001340, },
		{ -0.00001323, 0.00000382, -0.00004801, 0.00005195, -0.00006436, 0.00006155, -0.00001259, -0.00000949, -0.00000870, 0.00000975, 0.00000516, -0.00000098, -0.00001053, 0.00000044, -0.00000190, 0.00000990, -0.00004165, 0.00002354, 0.00000829, -0.00001211, -0.00002553, -0.00002720, 0.00002030, -0.00002401, 0.00001899, -0.00015390, 0.00005620, -0.00075314, 0.00170899, -0.00050588, -0.00011941, -0.00005819, -0.00004421, 0.00002476, -0.00003678, -0.00000239, 0.00001820, -0.00000470, -0.00004098, 0.00001856, 0.00001949, -0.00002627, 0.00001717, -0.00002319, 0.00001916, -0.00001713, -0.00000675, 0.00000226, 0.00002402, -0.00001190, -0.00002004, 0.00000831, 0.00002668, -0.00008744, 0.00009358, -0.00007721, 0.00011575, -0.00004870, 0.00004180, 0.00000557, -0.00006971, 0.00007115, -0.00004874, 0.00000211, },
		{ 0.00001494, -0.00004053, 0.00000565, 0.00000901, -0.00003899, 0.00007417, -0.00005245, 0.00002425, -0.00003626, 0.00004593, -0.00003332, 0.00003292, -0.00005094, 0.00004820, -0.00004598, 0.00004784, -0.00009746, 0.00011558, -0.00010030, 0.00009375, -0.00016473, 0.00014879, -0.00017381, 0.00021263, -0.00032555, 0.00028871, -0.00070940, 0.00045500, 0.00112131, -0.00087757, 0.00028721, -0.00031534, 0.00019098, -0.00015608, 0.00009625, -0.00009531, 0.00011148, -0.00011180, 0.00004859, -0.00002810, 0.00004830, -0.00005893, 0.00005076, -0.00005492, 0.00005426, -0.00006013, 0.00003224, -0.00001583, 0.00003904, -0.00004053, -0.00000175, 0.00002589, -0.00003249, -0.00002362, 0.00005591, -0.00004534, 0.00012851, -0.00010262, 0.00012678, -0.00012408, 0.00005767, -0.00001488, -0.00000984, -0.00002316, },
		{ 0.00002468, -0.00006394, 0.00005445, -0.00005547, 0.00003800, 0.00001397, -0.00002252, -0.00000167, -0.00000321, 0.00001847, -0.00000852, -0.00000217, -0.00001299, 0.00001464, -0.00000547, -0.00001868, -0.00001628, 0.00004975, -0.00004381, 0.00001057, -0.00007905, 0.00007701, -0.00008706, 0.00009698, -0.00023801, 0.00018227, -0.00084449, 0.00150106, -0.00014973, -0.00032721, 0.00005940, -0.00013153, 0.00008236, -0.00007984, 0.00002955, -0.00001006, 0.00002978, -0.00006686, 0.00002717, 0.00001147, -0.00000955, -0.00000129, -0.00000404, 0.00000056, 0.00000176, -0.00002327, 0.00000928, 0.00001833, -0.00000148, -0.00002142, -0.00000519, 0.00004077, -0.00009427, 0.00007733, -0.00004848, 0.00009053, -0.00000220, 0.00000612, 0.00003600, -0.00009686, 0.00008646, -0.00006172, 0.00001797, -0.00003782, },
		{ -0.00001757, -0.00002250, 0.00003349, -0.00005816, 0.00007636, -0.00003893, 0.00001433, -0.00003315, 0.00003824, -0.00002179, 0.00002638, -0.00004746, 0.00004233, -0.00003706, 0.00004360, -0.00009317, 0.00009422, -0.00006760, 0.00006713, -0.00013895, 0.00010731, -0.00012130, 0.00014868, -0.00023478, 0.00017463, -0.00048871, 0.00008723, 0.00131036, -0.00079792, 0.00018917, -0.00024523, 0.00012813, -0.00010940, 0.00006022, -0.00007107, 0.00009529, -0.00008732, 0.00002341, -0.00001954, 0.00004943, -0.00005726, 0.00004785, -0.00005052, 0.00004926, -0.00005217, 0.00002117, -0.00001222, 0.00003988, -0.00003446, -0.00000537, 0.00001196, -0.00000455, -0.00006300, 0.00008335, -0.00005934, 0.00014942, -0.00009635, 0.00011658, -0.00011168, 0.00003926, -0.00000345, -0.00001644, -0.00001692, -0.00000280, },
		{ -0.00006461, 0.00004279, -0.00003614, 0.00000938, 0.00003756, -0.00002940, 0.00000150, -0.00001298, 0.00002683, -0.00001133, 0.00000352, -0.00002471, 0.00002837, -0.00001737, 0.00000136, -0.00004865, 0.00007636, -0.00005906, 0.00003205, -0.00011309, 0.00010931, -0.00011962, 0.00013596, -0.00027205, 0.00023543, -0.00082682, 0.00120510, 0.00020711, -0.00048839, 0.00013432, -0.00018343, 0.00011122, -0.00010062, 0.00004463, -0.00002679, 0.00005362, -0.00008065, 0.00002775, 0.00000519, 0.00000673, -0.00001826, 0.00001287, -0.00001501, 0.00001703, -0.00003495, 0.00001211, 0.00001409, 0.00000765, -0.00002143, -0.00001456, 0.00004312, -0.00008359, 0.00004265, -0.00001021, 0.00006021, 0.00004847, -0.00002750, 0.00006213, -0.00011935, 0.00009653, -0.00007087, 0.00003072, -0.00005529, 0.00003360, },
		{ -0.00004709, 0.00004954, -0.00006605, 0.00006657, -0.00001890, 0.00000318, -0.00002869, 0.00002826, -0.00000963, 0.00002016, -0.00004254, 0.00003249, -0.00002329, 0.00003349, -0.00007872, 0.00006174, -0.00002893, 0.00003704, -0.00010629, 0.00005708, -0.00006166, 0.00007902, -0.00014144, 0.00006176, -0.00028773, -0.00019554, 0.00136461, -0.00062109, 0.00004783, -0.00015301, 0.00005369, -0.00005699, 0.00002107, -0.00004461, 0.00007460, -0.00005829, -0.00000180, -0.00001190, 0.00004710, -0.00004944, 0.00003846, -0.00003989, 0.00003876, -0.00003949, 0.00000792, -0.00000851, 0.00003974, -0.00002839, -0.00000531, -0.00000508, 0.00002093, -0.00009090, 0.00009004, -0.00005481, 0.00015393, -0.00007121, 0.00009153, -0.00008779, 0.00001120, 0.00001542, -0.00003082, -0.00000061, -0.00002900, 0.00000881, },
		{ 0.00002064, -0.00001167, -0.00001844, 0.00005367, -0.00002841, 0.00000054, -0.00001926, 0.00003024, -0.00001005, 0.00000708, -0.00003419, 0.00003760, -0.00002347, 0.00001498, -0.00006903, 0.00008494, -0.00005697, 0.00003824, -0.00012781, 0.00011613, -0.00012206, 0.00013847, -0.00026178, 0.00022811, -0.00073023, 0.00087191, 0.00051554, -0.00057444, 0.00015362, -0.00019408, 0.00010936, -0.00009896, 0.00004335, -0.00003209, 0.00006626, -0.00008264, 0.00002151, 0.00000063, 0.00002071, -0.00003109, 0.00002474, -0.00002506, 0.00002671, -0.00004060, 0.00001062, 0.00001098, 0.00001601, -0.00002161, -0.00001787, 0.00003383, -0.00005930, -0.00000149, 0.00002600, 0.00003461, 0.00009474, -0.00005059, 0.00007665, -0.00013097, 0.00009607, -0.00007140, 0.00003514, -0.00006050, 0.00002726, -0.00005030, },
		{ 0.00005319, -0.00006170, 0.00004691, 0.00000378, -0.00000613, -0.00002482, 0.00001795, 0.00000120, 0.00001548, -0.00003681, 0.00001968, -0.00000679, 0.00002048, -0.00005880, 0.00002584, 0.00000672, 0.00001171, -0.00007524, 0.00000882, -0.00000797, 0.00001890, -0.00006319, -0.00002874, -0.00013139, -0.00037721, 0.00129547, -0.00037937, -0.00010518, -0.00006262, -0.00001437, -0.00001202, -0.00001164, -0.00002274, 0.00005475, -0.00002986, -0.00002310, -0.00000674, 0.00004200, -0.00003694, 0.00002479, -0.00002571, 0.00002560, -0.00002490, -0.00000507, -0.00000601, 0.00003893, -0.00002189, -0.00000401, -0.00001991, 0.00003702, -0.00010092, 0.00007382, -0.00003464, 0.00014586, -0.00003123, 0.00005700, -0.00005824, -0.00002071, 0.00003620, -0.00004775, 0.00001955, -0.00005567, 0.00002982, -0.00006093, },
		{ 0.00001139, -0.00003997, 0.00005938, -0.00001962, -0.00000353, -0.00002275, 0.00002957, -0.00000596, 0.00000932, -0.00004104, 0.00004116, -0.00002262, 0.00002064, -0.00007608, 0.00007667, -0.00004114, 0.00003280, -0.00012654, 0.00010220, -0.00010152, 0.00011460, -0.00022121, 0.00018158, -0.00059118, 0.00054992, 0.00073911, -0.00057314, 0.00011837, -0.00016836, 0.00008277, -0.00008044, 0.00003032, -0.00002905, 0.00006963, -0.00007511, 0.00001074, -0.00000253, 0.00003094, -0.00003797, 0.00003000, -0.00002855, 0.00003031, -0.00004034, 0.00000560, 0.00000834, 0.00002389, -0.00002244, -0.00001563, 0.00001648, -0.00002902, -0.00004364, 0.00004875, 0.00002116, 0.00013071, -0.00005749, 0.00007545, -0.00012891, 0.00008323, -0.00006189, 0.00002957, -0.00005253, 0.00000725, -0.00002584, -0.00000526, },
		{ -0.00004744, 0.00002175, 0.00002456, -0.00001094, -0.00002287, 0.00000864, 0.00000951, 0.00001283, -0.00003055, 0.00000525, 0.00000974, 0.00000816, -0.00003865, -0.00000610, 0.00003205, -0.00000315, -0.00005166, -0.00002984, 0.00003092, -0.00002220, -0.00001001, -0.00008614, -0.00002951, -0.00046160, 0.00113268, -0.00011452, -0.00023709, 0.00000457, -0.00006183, 0.00001620, -0.00003173, -0.00000959, 0.00003920, -0.00000563, -0.00003806, -0.00000477, 0.00003505, -0.00002196, 0.00000979, -0.00001125, 0.00001287, -0.00001110, -0.00001574, -0.00000551, 0.00003731, -0.00001423, -0.00000384, -0.00002856, 0.00004007, -0.00009212, 0.00003928, -0.00000746, 0.00013280, 0.00001670, 0.00002085, -0.00003073, -0.00004932, 0.00005249, -0.00006128, 0.00003685, -0.00007527, 0.00003876, -0.00006035, 0.00004428, },
		{ -0.00005154, 0.00005408, -0.00000495, -0.00000868, -0.00002468, 0.00002615, -0.00000057, 0.00001098, -0.00004497, 0.00003868, -0.00001512, 0.00001901, -0.00007168, 0.00005645, -0.00001835, 0.00002204, -0.00011555, 0.00007572, -0.00006874, 0.00007784, -0.00016739, 0.00011899, -0.00044382, 0.00027633, 0.00086003, -0.00049046, 0.00004273, -0.00011985, 0.00004305, -0.00005420, 0.00001241, -0.00002231, 0.00006699, -0.00006143, -0.00000186, -0.00000481, 0.00003663, -0.00003836, 0.00002867, -0.00002610, 0.00002876, -0.00003534, -0.00000159, 0.00000548, 0.00003113, -0.00002327, -0.00001027, -0.00000319, -0.00000157, -0.00007336, 0.00005061, 0.00002212, 0.00015513, -0.00004680, 0.00005871, -0.00011459, 0.00006011, -0.00004464, 0.00001612, -0.00003453, -0.00002086, 0.00000104, -0.00002727, 0.00002722, },
		{ -0.00000354, 0.00003930, -0.00000968, -0.00002316, 0.00000085, 0.00001493, 0.00001206, -0.00002393, -0.00000930, 0.00002351, -0.00000016, -0.00002267, -0.00002877, 0.00004330, -0.00000586, -0.00003742, -0.00005579, 0.00005182, -0.00004182, 0.00001662, -0.00011026, 0.00002166, -0.00046767, 0.00091598, 0.00013175, -0.00032187, 0.00003534, -0.00008184, 0.00002444, -0.00003771, -0.00000585, 0.00002893, 0.00001282, -0.00004590, -0.00000593, 0.00002721, -0.00000699, -0.00000351, 0.00000055, 0.00000302, -0.00000007, -0.00002279, -0.00000720, 0.00003453, -0.00000490, -0.00000615, -0.00002957, 0.00003071, -0.00006907, -0.00000373, 0.00001518, 0.00012294, 0.00006562, -0.00000921, -0.00001230, -0.00006868, 0.00005933, -0.00006705, 0.00004647, -0.00008304, 0.00003290, -0.00004617, 0.00002639, -0.00002807, },
		{ 0.00003969, 0.00001220, -0.00001246, -0.00002639, 0.00002128, 0.00000487, 0.00001255, -0.00004583, 0.00003063, -0.00000278, 0.00001270, -0.00006010, 0.00003124, 0.00000366, 0.00001255, -0.00010150, 0.00004533, -0.00003442, 0.00004080, -0.00011527, 0.00005946, -0.00031340, 0.00007193, 0.00088052, -0.00034795, -0.00005036, -0.00006637, 0.00000357, -0.00002979, -0.00000365, -0.00001632, 0.00006167, -0.00004498, -0.00001385, -0.00000674, 0.00003781, -0.00003307, 0.00002226, -0.00001965, 0.00002399, -0.00002746, -0.00000934, 0.00000192, 0.00003712, -0.00002271, -0.00000508, -0.00001937, 0.00001585, -0.00008442, 0.00003061, 0.00003333, 0.00017153, -0.00002117, 0.00003060, -0.00009302, 0.00003194, -0.00002460, -0.00000058, -0.00001218, -0.00004959, 0.00002234, -0.00003927, 0.00003255, -0.00005232, },
		{ 0.00004534, -0.00000240, -0.00002497, -0.00000565, 0.00001773, 0.00001261, -0.00001706, -0.00002239, 0.00003220, -0.00000235, -0.00001329, -0.00004012, 0.00004089, 0.00000117, -0.00003056, -0.00007017, 0.00005671, -0.00004351, 0.00002203, -0.00010872, 0.00003493, -0.00042226, 0.00068568, 0.00032647, -0.00034597, 0.00002753, -0.00007614, 0.00001581, -0.00003259, -0.00000911, 0.00002276, 0.00002567, -0.00004724, -0.00000949, 0.00001940, 0.00000575, -0.00001280, 0.00000775, -0.00000269, 0.00000735, -0.00002586, -0.00001071, 0.00003030, 0.00000595, -0.00001078, -0.00002424, 0.00001327, -0.00004009, -0.00004321, 0.00002271, 0.00012157, 0.00011135, -0.00002797, -0.00000715, -0.00007593, 0.00005492, -0.00006373, 0.00004682, -0.00007812, 0.00001399, -0.00002319, 0.00000565, -0.00000938, -0.00002400, },
		{ 0.00002789, -0.00001276, -0.00002872, 0.00001590, 0.00000952, 0.00001416, -0.00004370, 0.00001839, 0.00001153, 0.00000539, -0.00004651, 0.00000797, 0.00001853, 0.00000895, -0.00008914, 0.00001753, -0.00000629, 0.00001197, -0.00007430, 0.00001443, -0.00021322, -0.00005953, 0.00081932, -0.00017658, -0.00013568, -0.00002461, -0.00002464, -0.00001427, -0.00001332, -0.00001390, 0.00005600, -0.00002841, -0.00002344, -0.00000875, 0.00003522, -0.00002397, 0.00001327, -0.00001187, 0.00001827, -0.00001863, -0.00001623, -0.00000255, 0.00004108, -0.00001941, -0.00000284, -0.00002804, 0.00001982, -0.00007654, -0.00000528, 0.00004543, 0.00018629, 0.00001461, -0.00000265, -0.00007085, 0.00000498, -0.00000738, -0.00001536, 0.00000856, -0.00007206, 0.00003220, -0.00003848, 0.00002754, -0.00004442, 0.00002009, },
		{ 0.00000922, -0.00002676, -0.00001150, 0.00001855, 0.00001380, -0.00001016, -0.00003267, 0.00003448, 0.00000199, -0.00001044, -0.00004143, 0.00002888, 0.00001258, -0.00002672, -0.00007686, 0.00005082, -0.00003428, 0.00001506, -0.00009264, 0.00002636, -0.00035240, 0.00047455, 0.00045122, -0.00031063, -0.00001017, -0.00005369, -0.00000198, -0.00002223, -0.00001519, 0.00001833, 0.00003418, -0.00004357, -0.00001442, 0.00001226, 0.00001478, -0.00001697, 0.00000992, -0.00000439, 0.00001144, -0.00002541, -0.00001531, 0.00002453, 0.00001746, -0.00001620, -0.00001589, -0.00000591, -0.00001444, -0.00006880, 0.00000934, 0.00012834, 0.00015404, -0.00003392, -0.00001562, -0.00007204, 0.00004101, -0.00005322, 0.00003958, -0.00006326, -0.00001271, 0.00000146, -0.00001135, 0.00000377, -0.00003645, 0.00004220, },
		{ -0.00000845, -0.00003178, 0.00001041, 0.00001304, 0.00001571, -0.00003893, 0.00000394, 0.00002450, 0.00000068, -0.00003538, -0.00000837, 0.00002307, 0.00001242, -0.00008006, -0.00000469, 0.00001234, -0.00000558, -0.00004737, -0.00001297, -0.00014524, -0.00012783, 0.00070506, -0.00000873, -0.00019242, -0.00000539, -0.00003606, -0.00001018, -0.00001539, -0.00001559, 0.00005085, -0.00001328, -0.00002969, -0.00001112, 0.00003013, -0.00001350, 0.00000448, -0.00000526, 0.00001346, -0.00001034, -0.00002124, -0.00000783, 0.00004235, -0.00001263, -0.00000481, -0.00002821, 0.00001159, -0.00005522, -0.00004600, 0.00004700, 0.00020532, 0.00005634, -0.00003519, -0.00005396, -0.00001569, 0.00000287, -0.00002458, 0.00002342, -0.00008404, 0.00002855, -0.00002620, 0.00001561, -0.00003203, 0.00000001, 0.00003858, },
		{ -0.00002668, -0.00001740, 0.00001803, 0.00001507, -0.00000356, -0.00003919, 0.00003029, 0.00001143, -0.00001192, -0.00003651, 0.00001333, 0.00002214, -0.00002130, -0.00008016, 0.00003999, -0.00002150, 0.00000431, -0.00007229, 0.00000999, -0.00027928, 0.00030280, 0.00050378, -0.00023007, -0.00006158, -0.00002719, -0.00001964, -0.00001295, -0.00001992, 0.00001331, 0.00003982, -0.00003663, -0.00001958, 0.00000609, 0.00001962, -0.00001635, 0.00000807, -0.00000345, 0.00001327, -0.00002242, -0.00002010, 0.00001737, 0.00002830, -0.00002025, -0.00000845, -0.00002076, 0.00000080, -0.00007512, -0.00002359, 0.00013663, 0.00019782, -0.00002866, -0.00003489, -0.00006091, 0.00002184, -0.00003959, 0.00002850, -0.00004335, -0.00004047, 0.00002078, -0.00001967, 0.00000887, -0.00004028, 0.00003164, 0.00002228, },
		{ -0.00003489, 0.00000473, 0.00001545, 0.00001698, -0.00003218, -0.00001045, 0.00003322, 0.00000105, -0.00002924, -0.00001592, 0.00001796, 0.00002062, -0.00007273, -0.00002171, 0.00002226, -0.00001356, -0.00003197, -0.00002590, -0.00010351, -0.00015045, 0.00056823, 0.00012930, -0.00020933, -0.00001092, -0.00003175, -0.00001539, -0.00001186, -0.00001981, 0.00004577, -0.00000008, -0.00003245, -0.00001394, 0.00002391, -0.00000398, -0.00000189, -0.00000149, 0.00001052, -0.00000338, -0.00002391, -0.00001361, 0.00004067, -0.00000270, -0.00001036, -0.00002190, -0.00000377, -0.00002961, -0.00007901, 0.00002881, 0.00023055, 0.00010303, -0.00006361, -0.00004560, -0.00002788, 0.00000488, -0.00002725, 0.00003104, -0.00008482, 0.00001334, -0.00000712, 0.00000215, -0.00002001, -0.00001636, 0.00004214, 0.00000023, },
		{ -0.00002369, 0.00001660, 0.00001608, 0.00000239, -0.00004162, 0.00002085, 0.00002312, -0.00001420, -0.00003010, 0.00000023, 0.00002503, -0.00001150, -0.00008283, 0.00002848, -0.00001032, -0.00000454, -0.00005415, -0.00000537, -0.00021530, 0.00017706, 0.00049573, -0.00012658, -0.00010830, -0.00000862, -0.00002967, -0.00000919, -0.00002066, 0.00000642, 0.00004358, -0.00002794, -0.00002398, 0.00000082, 0.00002081, -0.00001245, 0.00000423, -0.00000185, 0.00001426, -0.00001803, -0.00002423, 0.00000916, 0.00003713, -0.00002105, -0.00000492, -0.00002754, 0.00000293, -0.00006340, -0.00006756, 0.00013518, 0.00024843, -0.00001479, -0.00006089, -0.00004749, 0.00000225, -0.00002711, 0.00001753, -0.00002330, -0.00006338, 0.00003002, -0.00001761, 0.00000653, -0.00003747, 0.00001710, 0.00003343, -0.00002322, },
		{ -0.00000144, 0.00001698, 0.00001780, -0.00002428, -0.00002272, 0.00003588, 0.00000711, -0.00002806, -0.00001608, 0.00000717, 0.00002882, -0.00006388, -0.00003592, 0.00002669, -0.00001628, -0.00002299, -0.00003039, -0.00007888, -0.00014605, 0.00043418, 0.00022240, -0.00018671, -0.00003492, -0.00001833, -0.00002436, -0.00000672, -0.00002392, 0.00003963, 0.00001129, -0.00003212, -0.00001711, 0.00001772, 0.00000303, -0.00000479, -0.00000090, 0.00000930, 0.00000228, -0.00002431, -0.00001948, 0.00003624, 0.00000910, -0.00001732, -0.00001314, -0.00001942, -0.00000925, -0.00009446, -0.00001229, 0.00025748, 0.00015792, -0.00008767, -0.00004580, -0.00003259, 0.00000027, -0.00002508, 0.00003274, -0.00007680, -0.00000857, 0.00001238, -0.00000742, -0.00001221, -0.00002648, 0.00003838, 0.00001168, -0.00003676, },
		{ 0.00001442, 0.00001669, 0.00000742, -0.00004029, 0.00000833, 0.00003361, -0.00001376, -0.00002627, -0.00000640, 0.00001944, 0.00000245, -0.00008504, 0.00001777, -0.00000254, -0.00000988, -0.00003997, -0.00001684, -0.00016429, 0.00009301, 0.00044695, -0.00002391, -0.00013529, -0.00000540, -0.00002877, -0.00001207, -0.00001715, -0.00000214, 0.00004569, -0.00001855, -0.00002691, -0.00000388, 0.00001959, -0.00000744, 0.00000072, -0.00000145, 0.00001551, -0.00001320, -0.00002700, 0.00000026, 0.00004301, -0.00001774, -0.00000630, -0.00002601, -0.00000569, -0.00004086, -0.00010931, 0.00011198, 0.00030954, 0.00000698, -0.00009066, -0.00003548, -0.00001448, -0.00001840, 0.00000905, -0.00000633, -0.00007812, 0.00002806, -0.00000708, 0.00000008, -0.00003170, 0.00000264, 0.00003980, -0.00001571, -0.00003018, },
		{ 0.00001782, 0.00001808, -0.00001612, -0.00003142, 0.00003215, 0.00001744, -0.00002957, -0.00001266, -0.00000371, 0.00003190, -0.00005044, -0.00004974, 0.00002908, -0.00001804, -0.00001562, -0.00003189, -0.00006331, -0.00012947, 0.00031879, 0.00026820, -0.00013510, -0.00006533, -0.00000516, -0.00003060, -0.00000403, -0.00002540, 0.00003145, 0.00002115, -0.00002936, -0.00002026, 0.00001216, 0.00000703, -0.00000451, -0.00000253, 0.00000866, 0.00000720, -0.00002288, -0.00002493, 0.00002956, 0.00002101, -0.00002303, -0.00000625, -0.00002935, -0.00000067, -0.00008862, -0.00007120, 0.00027507, 0.00022759, -0.00010920, -0.00005241, -0.00003291, -0.00000759, -0.00002114, 0.00003128, -0.00006392, -0.00003136, 0.00002648, -0.00000952, -0.00001018, -0.00003019, 0.00002908, 0.00002328, -0.00003597, -0.00000839, },
		},
		{ { -0.00002711, -0.00002236, 0.00005959, 0.00000686, -0.00007465, 0.00001337, -0.00000370, -0.00000961, -0.00001020, -0.00000587, 0.00004607, -0.00001050, -0.00002448, -0.00000783, 0.00000475, -0.00001161, -0.00002063, -0.00001296, -0.00016410, 0.00005830, 0.00038050, 0.00005300, -0.00014562, -0.00001444, -0.00001936, -0.00001133, -0.00001692, -0.00001080, 0.00002571, -0.00001226, -0.00002040, 0.00000568, 0.00000924, 0.00000568, -0.00002040, -0.00001226, 0.00002571, -0.00001080, -0.00001692, -0.00001133, -0.00001936, -0.00001444, -0.00014562, 0.00005300, 0.00038050, 0.00005830, -0.00016410, -0.00001296, -0.00002063, -0.00001161, 0.00000475, -0.00000783, -0.00002448, -0.00001050, 0.00004607, -0.00000587, -0.00001020, -0.00000961, -0.00000370, 0.00001337, -0.00007465, 0.00000686, 0.00005959, -0.00002236, },
		{ -0.00002614, 0.00000035, 0.00006715, -0.00004418, -0.00004898, 0.00002572, -0.00002157, 0.00000187, -0.00002026, 0.00002119, 0.00003190, -0.00002692, -0.00001391, -0.00000570, 0.00000808, -0.00002748, -0.00000180, -0.00006892, -0.00013178, 0.00021925, 0.00029256, -0.00006787, -0.00010083, -0.00000059, -0.00002084, -0.00000996, -0.00002139, 0.00000788, 0.00001789, -0.00002528, -0.00000710, 0.00000711, 0.00001088, -0.00000367, -0.00002601, 0.00000676, 0.00001907, -0.00002469, -0.00000648, -0.00002012, -0.00001235, -0.00005108, -0.00015620, 0.00023450, 0.00035358, -0.00011285, -0.00011301, 0.00000396, -0.00002897, 0.00000212, -0.00000218, -0.00001085, -0.00003256, 0.00001990, 0.00003458, -0.00001920, -0.00000496, -0.00001449, 0.00001444, -0.00001967, -0.00006771, 0.00005002, 0.00003196, -0.00003367, },
		{ -0.00002217, 0.00002943, 0.00004612, -0.00007975, -0.00000412, 0.00000880, -0.00002060, 0.00000359, -0.00001968, 0.00004440, 0.00000098, -0.00002252, -0.00001243, 0.00000434, -0.00000419, -0.00002531, -0.00000475, -0.00012055, -0.00002259, 0.00030059, 0.00015457, -0.00011664, -0.00005517, 0.00000187, -0.00002141, -0.00001204, -0.00001729, 0.00002317, -0.00000167, -0.00002360, 0.00000223, 0.00000767, 0.00001054, -0.00001756, -0.00001877, 0.00002116, 0.00000135, -0.00002660, 0.00000114, -0.00003233, -0.00000273, -0.00011446, -0.00008722, 0.00041552, 0.00018926, -0.00020221, -0.00003477, -0.00001384, -0.00001635, 0.00000137, -0.00000245, -0.00002380, -0.00002142, 0.00004162, 0.00001059, -0.00001960, -0.00000642, -0.00000709, 0.00001641, -0.00005272, -0.00003290, 0.00006961, -0.00000208, -0.00003400, },
		{ -0.00001319, 0.00005455, -0.00000094, -0.00007892, 0.00003095, -0.00002233, 0.00000013, -0.00001210, 0.00000402, 0.00004068, -0.00001996, -0.00001409, -0.00001068, 0.00001146, -0.00002151, -0.00001014, -0.00003961, -0.00012179, 0.00011007, 0.00028635, 0.00002818, -0.00011395, -0.00002111, -0.00000144, -0.00002275, -0.00001329, -0.00000543, 0.00002649, -0.00002143, -0.00001039, 0.00000307, 0.00001257, 0.00000121, -0.00002587, -0.00000280, 0.00002363, -0.00001885, -0.00001550, -0.00000312, -0.00003762, -0.00000548, -0.00018536, 0.00009241, 0.00049222, -0.00004823, -0.00016875, 0.00000231, -0.00002629, -0.00000388, -0.00000184, -0.00000557, -0.00003567, 0.00000588, 0.00004239, -0.00001059, -0.00001491, -0.00000819, 0.00000577, -0.00000127, -0.00006699, 0.00001295, 0.00006207, -0.00003017, -0.00002681, },
		{ 0.00000224, 0.00006234, -0.00005653, -0.00003812, 0.00002920, -0.00003625, 0.00001716, -0.00002779, 0.00003745, 0.00001079, -0.00001757, -0.00001618, 0.00000233, 0.00000115, -0.00002318, -0.00000841, -0.00007923, -0.00006456, 0.00021123, 0.00020219, -0.00005119, -0.00008950, 0.00000341, -0.00001225, -0.00002016, -0.00001388, 0.00001202, 0.00001240, -0.00002662, 0.00000024, 0.00000368, 0.00001536, -0.00001382, -0.00002324, 0.00001430, 0.00001158, -0.00003002, 0.00000036, -0.00002075, -0.00002734, -0.00004361, -0.00020913, 0.00035420, 0.00037757, -0.00022533, -0.00006798, -0.00001231, -0.00001393, -0.00000703, 0.00000376, -0.00002210, -0.00002875, 0.00003031, 0.00002755, -0.00002337, -0.00000899, -0.00000612, 0.00001344, -0.00002959, -0.00005473, 0.00005084, 0.00003414, -0.00004456, -0.00001657, },
		{ 0.00002414, 0.00004074, -0.00009044, 0.00001775, -0.00000838, -0.00001517, 0.00000632, -0.00001626, 0.00004788, -0.00001561, -0.00000863, -0.00001855, 0.00001503, -0.00001732, -0.00001159, -0.00003005, -0.00009172, 0.00002725, 0.00024818, 0.00009455, -0.00008193, -0.00006045, 0.00001841, -0.00002785, -0.00001207, -0.00001234, 0.00002669, -0.00001181, -0.00001486, 0.00000022, 0.00001138, 0.00000723, -0.00002464, -0.00001103, 0.00002375, -0.00000920, -0.00002636, 0.00000969, -0.00004460, -0.00000303, -0.00013246, -0.00010594, 0.00057771, 0.00009286, -0.00024002, -0.00000176, -0.00002918, -0.00000100, -0.00001045, 0.00000497, -0.00003947, -0.00000490, 0.00004049, 0.00000565, -0.00002592, -0.00000406, 0.00000025, 0.00000778, -0.00005261, -0.00002227, 0.00006884, -0.00000321, -0.00004194, -0.00000816, },
		{ 0.00004741, -0.00001091, -0.00007623, 0.00004550, -0.00004528, 0.00002058, -0.00002392, 0.00002264, 0.00002162, -0.00001455, -0.00001449, -0.00000391, 0.00000678, -0.00001963, -0.00001220, -0.00005338, -0.00007082, 0.00012355, 0.00021342, 0.00000968, -0.00008653, -0.00002635, 0.00001482, -0.00003507, -0.00000636, -0.00000218, 0.00002478, -0.00002654, -0.00000156, -0.00000135, 0.00001881, -0.00000871, -0.00002634, 0.00000704, 0.00001760, -0.00002539, -0.00001262, 0.00000468, -0.00005974, 0.00001195, -0.00024463, 0.00016823, 0.00059629, -0.00020322, -0.00012122, -0.00001234, -0.00001140, -0.00001222, 0.00000413, -0.00001570, -0.00003503, 0.00001703, 0.00003904, -0.00001856, -0.00001669, -0.00000312, 0.00000885, -0.00001232, -0.00005661, 0.00001565, 0.00006422, -0.00003814, -0.00002447, -0.00000404, },
		{ 0.00005696, -0.00007018, -0.00001772, 0.00001987, -0.00004057, 0.00002916, -0.00003455, 0.00005178, -0.00001354, 0.00000048, -0.00002683, 0.00001833, -0.00001661, -0.00000497, -0.00003439, -0.00005660, -0.00002550, 0.00019427, 0.00012913, -0.00003211, -0.00008374, 0.00001230, -0.00001038, -0.00002456, -0.00000889, 0.00001686, 0.00000290, -0.00001984, -0.00000116, 0.00000752, 0.00001328, -0.00002126, -0.00001937, 0.00002307, -0.00000234, -0.00002842, 0.00000357, -0.00001977, -0.00004524, -0.00003096, -0.00027979, 0.00053216, 0.00033381, -0.00032176, -0.00000621, -0.00004426, 0.00001441, -0.00002843, 0.00002005, -0.00004487, -0.00000974, 0.00002821, 0.00002693, -0.00003556, -0.00000205, -0.00000404, 0.00001253, -0.00003693, -0.00004014, 0.00004696, 0.00003848, -0.00005716, -0.00000287, 0.00000081, },
		{ 0.00003350, -0.00009630, 0.00004127, -0.00003513, 0.00000502, -0.00000364, -0.00000204, 0.00003706, -0.00001898, -0.00000282, -0.00001777, 0.00001758, -0.00002314, -0.00000412, -0.00005170, -0.00004662, 0.00004188, 0.00020758, 0.00004390, -0.00004792, -0.00006821, 0.00003677, -0.00003842, -0.00000678, -0.00001040, 0.00002900, -0.00002054, -0.00000414, -0.00000646, 0.00002060, -0.00000349, -0.00002614, -0.00000285, 0.00002354, -0.00002050, -0.00002296, 0.00001754, -0.00006099, 0.00000513, -0.00016666, -0.00010506, 0.00076461, -0.00008057, -0.00021542, 0.00000202, -0.00002660, 0.00000113, -0.00001249, 0.00000306, -0.00004747, 0.00001092, 0.00003676, -0.00000084, -0.00003126, 0.00000365, 0.00000107, 0.00000163, -0.00004945, -0.00001393, 0.00006666, -0.00000410, -0.00004901, 0.00000757, 0.00001599, },
		{ -0.00002425, -0.00006233, 0.00004835, -0.00006164, 0.00004266, -0.00004021, 0.00004531, -0.00000685, 0.00000440, -0.00002686, 0.00001430, -0.00001212, -0.00000104, -0.00003486, -0.00003824, -0.00003869, 0.00012429, 0.00015163, -0.00000001, -0.00006622, -0.00002675, 0.00002488, -0.00004281, -0.00000119, 0.00000245, 0.00001761, -0.00002283, -0.00000132, 0.00000101, 0.00001939, -0.00001697, -0.00002593, 0.00002019, 0.00000289, -0.00002368, -0.00001486, 0.00001757, -0.00009102, 0.00004495, -0.00033644, 0.00030938, 0.00064879, -0.00037320, -0.00003153, -0.00005823, 0.00002745, -0.00004094, 0.00002722, -0.00004312, -0.00001549, 0.00001434, 0.00004315, -0.00003618, -0.00000655, -0.00000428, 0.00001248, -0.00002286, -0.00004216, 0.00001232, 0.00006940, -0.00004934, -0.00001701, 0.00000185, 0.00004028, },
		{ -0.00008252, 0.00000984, -0.00000397, -0.00002735, 0.00002764, -0.00003095, 0.00005402, -0.00002890, 0.00001524, -0.00003510, 0.00003119, -0.00003310, 0.00001432, -0.00006561, -0.00001159, -0.00001910, 0.00018457, 0.00006211, -0.00000626, -0.00008580, 0.00002703, -0.00001610, -0.00002143, -0.00000846, 0.00002290, -0.00000901, -0.00000699, -0.00001095, 0.00002033, 0.00000102, -0.00002158, -0.00001661, 0.00003235, -0.00002152, -0.00001969, 0.00000277, -0.00001862, -0.00006161, -0.00002117, -0.00035798, 0.00077041, 0.00019203, -0.00035835, 0.00003814, -0.00006890, 0.00003840, -0.00004770, 0.00003418, -0.00006746, 0.00001535, 0.00001906, 0.00002800, -0.00005040, 0.00001351, -0.00001016, 0.00001490, -0.00004300, -0.00002500, 0.00003927, 0.00004491, -0.00007147, 0.00001520, -0.00000493, 0.00005115, },
		{ -0.00009167, 0.00005708, -0.00006162, 0.00003365, -0.00002448, 0.00002248, 0.00001132, -0.00000527, -0.00001121, -0.00000351, 0.00000422, -0.00001152, -0.00001617, -0.00004986, -0.00001212, 0.00004175, 0.00017346, 0.00000342, -0.00001740, -0.00007483, 0.00005216, -0.00004767, -0.00000102, -0.00000785, 0.00002627, -0.00002185, 0.00000017, -0.00000878, 0.00002691, -0.00001516, -0.00002573, 0.00000921, 0.00001448, -0.00002453, -0.00002411, 0.00003182, -0.00008969, 0.00003356, -0.00023660, -0.00005062, 0.00093583, -0.00031160, -0.00012097, -0.00003923, 0.00000824, -0.00002294, 0.00000711, -0.00001959, -0.00003524, 0.00001339, 0.00004050, -0.00001786, -0.00002461, 0.00000598, 0.00000229, -0.00000522, -0.00004112, -0.00001153, 0.00006721, -0.00000837, -0.00005159, 0.00002146, 0.00000737, 0.00002233, },
		{ -0.00003648, 0.00003373, -0.00006053, 0.00005367, -0.00005004, 0.00006139, -0.00003329, 0.00002685, -0.00004311, 0.00003601, -0.00003720, 0.00002680, -0.00007147, 0.00000468, -0.00003799, 0.00013181, 0.00008931, 0.00000440, -0.00005715, -0.00001909, 0.00002538, -0.00004187, -0.00000177, 0.00001081, 0.00000420, -0.00000915, -0.00001299, 0.00001561, 0.00000726, -0.00001646, -0.00002898, 0.00003841, -0.00002272, -0.00000968, -0.00002635, 0.00003810, -0.00013220, 0.00009346, -0.00046406, 0.00053721, 0.00060154, -0.00050627, 0.00006713, -0.00011680, 0.00007748, -0.00008037, 0.00005951, -0.00008105, 0.00001903, -0.00000146, 0.00005396, -0.00006199, 0.00001712, -0.00001681, 0.00002183, -0.00003550, -0.00002326, 0.00000211, 0.00007923, -0.00006535, -0.00000258, 0.00000209, 0.00003443, -0.00004003, },
		{ 0.00003605, -0.00003539, 0.00000149, 0.00000952, -0.00001252, 0.00003841, -0.00002653, 0.00001638, -0.00003072, 0.00002959, -0.00003489, 0.00001873, -0.00008152, 0.00003223, -0.00003109, 0.00018147, 0.00000026, 0.00002437, -0.00009304, 0.00004386, -0.00002519, -0.00001432, -0.00000983, 0.00002652, -0.00001734, 0.00000405, -0.00002127, 0.00003544, -0.00001701, -0.00001677, -0.00001219, 0.00003621, -0.00003832, -0.00001084, 0.00000483, -0.00002615, -0.00006322, -0.00003481, -0.00041333, 0.00104267, -0.00005123, -0.00030559, 0.00003372, -0.00006152, 0.00003747, -0.00004701, 0.00002979, -0.00007351, 0.00003153, 0.00001436, 0.00002048, -0.00005603, 0.00002716, -0.00001839, 0.00001780, -0.00004544, -0.00001331, 0.00003198, 0.00004987, -0.00008404, 0.00003443, -0.00001292, 0.00004258, -0.00008750, },
		{ 0.00005573, -0.00007444, 0.00005730, -0.00004600, 0.00004843, -0.00002087, 0.00001951, -0.00003022, 0.00002115, -0.00002259, 0.00001670, -0.00004912, -0.00001686, -0.00000730, 0.00004850, 0.00013378, -0.00002007, 0.00000323, -0.00007478, 0.00005883, -0.00004924, -0.00000071, -0.00000031, 0.00001495, -0.00001104, -0.00000911, 0.00000244, 0.00002000, -0.00001810, -0.00002999, 0.00002983, -0.00000964, -0.00001279, -0.00003843, 0.00006017, -0.00013829, 0.00009211, -0.00035994, 0.00009540, 0.00103300, -0.00055092, 0.00002122, -0.00011906, 0.00007310, -0.00007272, 0.00004869, -0.00006402, 0.00000076, -0.00000317, 0.00005624, -0.00004895, 0.00000051, -0.00000612, 0.00001164, -0.00001736, -0.00002507, -0.00001826, 0.00007426, -0.00001960, -0.00004596, 0.00002953, 0.00000096, 0.00000601, -0.00007242, },
		{ 0.00000400, -0.00003830, 0.00004511, -0.00004409, 0.00006073, -0.00004535, 0.00003997, -0.00005120, 0.00004852, -0.00005423, 0.00004824, -0.00010396, 0.00005770, -0.00006187, 0.00015060, 0.00002290, 0.00002342, -0.00005942, -0.00000447, 0.00001786, -0.00003319, -0.00000744, 0.00002090, -0.00001127, 0.00000913, -0.00003193, 0.00003952, -0.00001710, -0.00000580, -0.00003517, 0.00005804, -0.00005456, 0.00001309, -0.00004358, 0.00006054, -0.00017376, 0.00014308, -0.00061009, 0.00084857, 0.00042328, -0.00055133, 0.00013165, -0.00015748, 0.00011477, -0.00011058, 0.00008334, -0.00011276, 0.00005436, -0.00002016, 0.00006225, -0.00008545, 0.00004596, -0.00003706, 0.00003593, -0.00004890, -0.00000297, -0.00001196, 0.00009140, -0.00008399, 0.00001702, -0.00000252, 0.00002826, -0.00005318, -0.00000454, },
		{ -0.00006049, 0.00003417, -0.00001816, 0.00001597, 0.00000841, -0.00000682, 0.00000182, -0.00001102, 0.00000975, -0.00001592, -0.00000063, -0.00006722, 0.00005143, -0.00003925, 0.00017309, -0.00004857, 0.00005206, -0.00009876, 0.00005581, -0.00003015, -0.00001018, -0.00000721, 0.00002198, -0.00001461, 0.00000324, -0.00002040, 0.00004059, -0.00002988, -0.00001460, 0.00000071, 0.00002529, -0.00004087, -0.00001384, 0.00002146, -0.00005566, -0.00003218, -0.00009978, -0.00039972, 0.00129329, -0.00036292, -0.00016393, -0.00003195, -0.00000950, -0.00000002, -0.00001559, -0.00000355, -0.00005111, 0.00002633, 0.00002417, -0.00000440, -0.00004112, 0.00002619, -0.00001861, 0.00001475, -0.00003958, -0.00001045, 0.00003077, 0.00004810, -0.00008959, 0.00004960, -0.00001933, 0.00002813, -0.00007819, 0.00004916, },
		{ -0.00006415, 0.00006260, -0.00005529, 0.00006316, -0.00004588, 0.00004256, -0.00004858, 0.00004482, -0.00005082, 0.00004938, -0.00009052, 0.00003778, -0.00003390, 0.00007261, 0.00008088, -0.00002189, 0.00000972, -0.00006535, 0.00005460, -0.00004054, -0.00000833, 0.00001320, -0.00000420, 0.00001077, -0.00003233, 0.00002975, -0.00000515, -0.00000611, -0.00004273, 0.00005976, -0.00004959, 0.00001734, -0.00006870, 0.00010370, -0.00020586, 0.00017879, -0.00053882, 0.00035676, 0.00099929, -0.00073348, 0.00016569, -0.00020965, 0.00014794, -0.00013232, 0.00009869, -0.00011803, 0.00005137, -0.00003327, 0.00007883, -0.00008773, 0.00004041, -0.00003345, 0.00003167, -0.00003611, -0.00000184, -0.00003372, 0.00008804, -0.00003806, -0.00003147, 0.00003037, -0.00000307, -0.00001272, -0.00004033, 0.00003400, },
		{ -0.00000247, 0.00001856, -0.00002122, 0.00004058, -0.00003718, 0.00003591, -0.00004293, 0.00004241, -0.00005141, 0.00004876, -0.00011294, 0.00009605, -0.00008992, 0.00017075, -0.00004069, 0.00005031, -0.00006745, 0.00001097, 0.00000905, -0.00002306, -0.00001262, 0.00002648, -0.00002044, 0.00002130, -0.00004629, 0.00006113, -0.00004408, 0.00000840, -0.00003787, 0.00006914, -0.00008037, 0.00003254, -0.00005426, 0.00007005, -0.00019552, 0.00016443, -0.00073298, 0.00120547, 0.00011818, -0.00047991, 0.00012795, -0.00015246, 0.00011613, -0.00011133, 0.00008014, -0.00011959, 0.00007291, -0.00002767, 0.00005545, -0.00009186, 0.00006518, -0.00005400, 0.00004762, -0.00005741, 0.00001233, -0.00002345, 0.00010005, -0.00009953, 0.00003637, -0.00000795, 0.00001773, -0.00005546, 0.00002157, -0.00002907, },
		{ 0.00005436, -0.00004136, 0.00004216, -0.00002402, 0.00002032, -0.00002074, 0.00001671, -0.00002127, 0.00001792, -0.00003931, -0.00002258, 0.00003524, -0.00002688, 0.00014715, -0.00007507, 0.00006833, -0.00009659, 0.00005707, -0.00002482, -0.00001483, 0.00000381, 0.00000591, 0.00000310, -0.00001544, 0.00000023, 0.00002498, -0.00002635, -0.00002345, 0.00002805, -0.00000766, -0.00001937, -0.00003900, 0.00006225, -0.00011696, 0.00004416, -0.00023830, -0.00027068, 0.00145045, -0.00067849, 0.00003263, -0.00014472, 0.00007936, -0.00006892, 0.00004273, -0.00006295, -0.00000049, -0.00000272, 0.00005017, -0.00004689, -0.00000253, 0.00000361, -0.00000345, 0.00000052, -0.00002225, -0.00001965, 0.00003924, 0.00003615, -0.00008442, 0.00005665, -0.00002166, 0.00000739, -0.00005165, 0.00004041, -0.00006507, },
		{ 0.00004429, -0.00004466, 0.00005781, -0.00005244, 0.00005152, -0.00005423, 0.00005452, -0.00006548, 0.00006962, -0.00012033, 0.00009184, -0.00007647, 0.00010889, 0.00001785, -0.00000489, 0.00000394, -0.00004960, 0.00004337, -0.00002515, -0.00002049, 0.00002813, -0.00002451, 0.00003511, -0.00006067, 0.00006545, -0.00004339, 0.00001842, -0.00006060, 0.00009206, -0.00009682, 0.00005818, -0.00010710, 0.00015206, -0.00027770, 0.00027308, -0.00074901, 0.00072512, 0.00080323, -0.00080196, 0.00026109, -0.00027347, 0.00020326, -0.00017690, 0.00013499, -0.00016059, 0.00009799, -0.00006251, 0.00009518, -0.00011931, 0.00008130, -0.00006688, 0.00005693, -0.00005701, 0.00002328, -0.00005268, 0.00010391, -0.00005928, -0.00001209, 0.00002678, -0.00000822, -0.00002628, -0.00000494, -0.00000109, -0.00003188, },
		{ -0.00001394, 0.00000963, 0.00000833, -0.00001330, 0.00001664, -0.00001955, 0.00001802, -0.00002694, 0.00002399, -0.00008884, 0.00010104, -0.00010046, 0.00017650, -0.00008895, 0.00007412, -0.00007288, 0.00001925, 0.00000741, -0.00001958, -0.00001022, 0.00002069, -0.00001511, 0.00001662, -0.00004282, 0.00006542, -0.00005903, 0.00001468, -0.00002714, 0.00005931, -0.00008473, 0.00003263, -0.00004218, 0.00004828, -0.00017363, 0.00012174, -0.00077556, 0.00153829, -0.00026773, -0.00030028, 0.00004696, -0.00009093, 0.00007136, -0.00007320, 0.00004075, -0.00009081, 0.00006292, -0.00001444, 0.00002568, -0.00007059, 0.00006137, -0.00005542, 0.00004783, -0.00005424, 0.00001595, -0.00002564, 0.00009887, -0.00010558, 0.00004916, -0.00000958, -0.00000064, -0.00003965, 0.00002898, -0.00004886, 0.00003056, },
		{ -0.00004674, 0.00005272, -0.00004440, 0.00004062, -0.00003816, 0.00003897, -0.00004878, 0.00005126, -0.00008087, 0.00003665, -0.00000908, 0.00000828, 0.00010027, -0.00007589, 0.00006961, -0.00008499, 0.00004720, -0.00000860, -0.00002875, 0.00002237, -0.00001909, 0.00003223, -0.00004906, 0.00003962, -0.00001344, -0.00000205, -0.00004600, 0.00006911, -0.00006150, 0.00002691, -0.00008738, 0.00012665, -0.00020748, 0.00016212, -0.00045147, -0.00000061, 0.00145009, -0.00092134, 0.00022741, -0.00026838, 0.00017864, -0.00014786, 0.00010941, -0.00013146, 0.00006449, -0.00004580, 0.00008320, -0.00009619, 0.00005092, -0.00003671, 0.00002629, -0.00002421, 0.00000465, -0.00003898, 0.00005591, 0.00001544, -0.00006949, 0.00005571, -0.00002131, -0.00001468, -0.00001280, 0.00001026, -0.00004286, 0.00004991, },
		{ -0.00001698, 0.00003313, -0.00003823, 0.00004172, -0.00004184, 0.00004363, -0.00005740, 0.00006538, -0.00012175, 0.00012165, -0.00011148, 0.00014112, -0.00004280, 0.00002074, -0.00000660, -0.00003520, 0.00003361, -0.00001139, -0.00002944, 0.00003621, -0.00003586, 0.00004927, -0.00007930, 0.00009373, -0.00008028, 0.00004435, -0.00007311, 0.00011292, -0.00013461, 0.00009268, -0.00013604, 0.00018465, -0.00032710, 0.00033762, -0.00093915, 0.00115199, 0.00045635, -0.00072937, 0.00027124, -0.00028018, 0.00021381, -0.00018456, 0.00013752, -0.00017155, 0.00012127, -0.00007501, 0.00009134, -0.00012732, 0.00010540, -0.00009137, 0.00007634, -0.00007138, 0.00004174, -0.00006682, 0.00011413, -0.00007565, 0.00000494, 0.00002436, -0.00001987, -0.00002485, 0.00001929, -0.00003365, 0.00000931, 0.00000922, },
		{ 0.00003357, -0.00002152, 0.00001335, -0.00000690, 0.00000830, -0.00001386, 0.00000912, -0.00001681, -0.00003799, 0.00006971, -0.00008242, 0.00015699, -0.00011182, 0.00008580, -0.00006961, 0.00001546, 0.00001810, -0.00002767, 0.00000348, 0.00000087, 0.00000739, -0.00000913, -0.00001461, 0.00004228, -0.00005017, 0.00000351, 0.00000376, 0.00002082, -0.00005704, 0.00000179, 0.00000388, -0.00001664, -0.00009283, -0.00001117, -0.00068433, 0.00176396, -0.00065665, -0.00005712, -0.00009035, 0.00001455, -0.00001034, -0.00000373, -0.00002833, -0.00003017, 0.00002504, 0.00001838, -0.00002408, -0.00002217, 0.00003043, -0.00003535, 0.00003160, -0.00003624, 0.00000507, -0.00001560, 0.00008511, -0.00009915, 0.00005201, -0.00000571, -0.00002581, -0.00000485, 0.00001101, -0.00004219, 0.00004333, -0.00003577, },
		{ 0.00004113, -0.00004366, 0.00004340, -0.00003962, 0.00004383, -0.00005851, 0.00006751, -0.00010518, 0.00008629, -0.00006047, 0.00005368, 0.00004092, -0.00005678, 0.00005950, -0.00006870, 0.00003219, 0.00001255, -0.00004611, 0.00004164, -0.00004423, 0.00006179, -0.00008527, 0.00008561, -0.00006484, 0.00003657, -0.00007563, 0.00011365, -0.00012383, 0.00008616, -0.00014696, 0.00019995, -0.00030784, 0.00029587, -0.00070929, 0.00039742, 0.00126129, -0.00102934, 0.00036229, -0.00035966, 0.00025441, -0.00020831, 0.00015903, -0.00018482, 0.00012186, -0.00008524, 0.00010737, -0.00013427, 0.00010112, -0.00008075, 0.00006088, -0.00005186, 0.00003356, -0.00006113, 0.00007399, -0.00000739, -0.00005090, 0.00005151, -0.00002364, -0.00002816, 0.00002544, -0.00002957, -0.00000048, 0.00002008, -0.00002952, },
		{ 0.00000017, -0.00001165, 0.00001900, -0.00001643, 0.00001647, -0.00002899, 0.00003682, -0.00009066, 0.00011419, -0.00011989, 0.00015227, -0.00008614, 0.00004254, -0.00001305, -0.00003014, 0.00003386, -0.00000765, -0.00002759, 0.00003012, -0.00003000, 0.00004277, -0.00007465, 0.00009813, -0.00009854, 0.00005717, -0.00006775, 0.00010753, -0.00014465, 0.00010129, -0.00013534, 0.00017859, -0.00032499, 0.00033013, -0.00104317, 0.00155633, 0.00001582, -0.00053100, 0.00019009, -0.00022054, 0.00017080, -0.00014727, 0.00009866, -0.00014169, 0.00011060, -0.00006215, 0.00006081, -0.00010274, 0.00009990, -0.00009393, 0.00007926, -0.00007099, 0.00004610, -0.00006898, 0.00011201, -0.00008042, 0.00001297, 0.00002788, -0.00004112, -0.00000185, 0.00001965, -0.00004586, 0.00003968, -0.00002623, 0.00001496, },
		{ -0.00003493, 0.00002840, -0.00002101, 0.00002635, -0.00003737, 0.00003972, -0.00005497, 0.00001982, 0.00001620, -0.00004071, 0.00011317, -0.00010788, 0.00008299, -0.00005765, 0.00000115, 0.00003953, -0.00004566, 0.00002568, -0.00002854, 0.00004123, -0.00004983, 0.00003367, -0.00000681, -0.00001566, -0.00002614, 0.00005241, -0.00004308, 0.00000132, -0.00005920, 0.00008174, -0.00012064, 0.00004183, -0.00023452, -0.00043271, 0.00181334, -0.00096302, 0.00018392, -0.00024113, 0.00013269, -0.00010354, 0.00007486, -0.00010654, 0.00004475, -0.00002738, 0.00005855, -0.00007956, 0.00004012, -0.00001922, 0.00000181, 0.00000195, -0.00000690, -0.00001675, 0.00000334, 0.00006211, -0.00008307, 0.00004684, 0.00000035, -0.00005039, 0.00004061, -0.00002756, -0.00000730, 0.00002659, -0.00003469, 0.00003786, },
		{ -0.00002341, 0.00002769, -0.00002319, 0.00002811, -0.00004470, 0.00005800, -0.00009918, 0.00010576, -0.00009511, 0.00008969, -0.00001488, -0.00003003, 0.00004652, -0.00005617, 0.00002152, 0.00002906, -0.00005770, 0.00005199, -0.00005832, 0.00007808, -0.00010745, 0.00011945, -0.00011083, 0.00007437, -0.00009850, 0.00014463, -0.00017417, 0.00013704, -0.00019554, 0.00025656, -0.00038589, 0.00040183, -0.00095285, 0.00086652, 0.00090207, -0.00097737, 0.00039971, -0.00038647, 0.00028020, -0.00022754, 0.00017097, -0.00020240, 0.00015149, -0.00010431, 0.00010832, -0.00014439, 0.00012883, -0.00011087, 0.00008628, -0.00007128, 0.00005435, -0.00007650, 0.00008451, -0.00002346, -0.00003709, 0.00005072, -0.00003482, -0.00002244, 0.00004705, -0.00006015, 0.00004401, -0.00002290, 0.00000307, 0.00001206, },
		{ 0.00001247, -0.00000349, 0.00000904, -0.00001344, 0.00000653, -0.00000317, -0.00003855, 0.00007350, -0.00009666, 0.00013426, -0.00010289, 0.00005203, -0.00001048, -0.00003776, 0.00004767, -0.00001741, -0.00001241, 0.00000832, -0.00000577, 0.00001356, -0.00004215, 0.00007028, -0.00008691, 0.00004701, -0.00003757, 0.00006823, -0.00011612, 0.00007176, -0.00009265, 0.00012044, -0.00025407, 0.00022156, -0.00100285, 0.00184717, -0.00043059, -0.00025981, 0.00004384, -0.00011059, 0.00008639, -0.00007542, 0.00002772, -0.00007772, 0.00006936, -0.00002749, 0.00000939, -0.00004970, 0.00006403, -0.00007073, 0.00006173, -0.00005321, 0.00003431, -0.00005736, 0.00009604, -0.00007179, 0.00000985, 0.00003765, -0.00006905, 0.00004099, -0.00000734, -0.00002731, 0.00004247, -0.00004433, 0.00003488, -0.00002515, },
		{ 0.00002389, -0.00001722, 0.00002492, -0.00004019, 0.00004993, -0.00007201, 0.00006088, -0.00003567, 0.00000720, 0.00005756, -0.00008523, 0.00007068, -0.00004308, -0.00001602, 0.00006386, -0.00006577, 0.00004772, -0.00005716, 0.00007375, -0.00009062, 0.00008625, -0.00006783, 0.00003419, -0.00006464, 0.00010567, -0.00011665, 0.00007488, -0.00013464, 0.00017297, -0.00024064, 0.00020062, -0.00051401, -0.00003781, 0.00165692, -0.00112191, 0.00036002, -0.00035821, 0.00022680, -0.00017758, 0.00013531, -0.00016787, 0.00010998, -0.00007469, 0.00008862, -0.00012105, 0.00009540, -0.00006990, 0.00004274, -0.00003073, 0.00002430, -0.00004054, 0.00002270, 0.00003827, -0.00006520, 0.00004012, 0.00000162, -0.00006258, 0.00008107, -0.00007080, 0.00004345, -0.00001494, -0.00000987, 0.00001925, -0.00002595, },
		{ 0.00000273, 0.00000271, -0.00000020, -0.00001367, 0.00002711, -0.00006405, 0.00008819, -0.00009826, 0.00010029, -0.00005181, -0.00000823, 0.00003896, -0.00005452, 0.00002297, 0.00003304, -0.00005621, 0.00004631, -0.00005370, 0.00007165, -0.00010291, 0.00012464, -0.00013280, 0.00009484, -0.00010036, 0.00014572, -0.00019227, 0.00015749, -0.00020999, 0.00027024, -0.00040880, 0.00043329, -0.00111037, 0.00131790, 0.00043783, -0.00078619, 0.00033604, -0.00034079, 0.00024844, -0.00019895, 0.00013916, -0.00017670, 0.00014425, -0.00009565, 0.00008126, -0.00011982, 0.00012300, -0.00011453, 0.00009133, -0.00007360, 0.00005939, -0.00007819, 0.00008112, -0.00002638, -0.00003433, 0.00005759, -0.00005703, 0.00000768, 0.00004085, -0.00006425, 0.00006919, -0.00005935, 0.00003503, -0.00002046, 0.00000465, },
		{ -0.00001277, 0.00002086, -0.00003041, 0.00003186, -0.00003539, 0.00001290, 0.00001834, -0.00005246, 0.00009213, -0.00009395, 0.00004841, -0.00000055, -0.00005453, 0.00007140, -0.00003699, 0.00001149, -0.00002317, 0.00002923, -0.00002963, 0.00001008, 0.00001526, -0.00004642, 0.00001476, 0.00001324, 0.00000038, -0.00005257, 0.00000670, -0.00001172, 0.00001553, -0.00012044, 0.00001262, -0.00079187, 0.00195312, -0.00079187, 0.00001262, -0.00012044, 0.00001553, -0.00001172, 0.00000670, -0.00005257, 0.00000038, 0.00001324, 0.00001476, -0.00004642, 0.00001526, 0.00001008, -0.00002963, 0.00002923, -0.00002317, 0.00001149, -0.00003699, 0.00007140, -0.00005453, -0.00000055, 0.00004841, -0.00009395, 0.00009213, -0.00005246, 0.00001834, 0.00001290, -0.00003539, 0.00003186, -0.00003041, 0.00002086, },
		{ 0.00000273, 0.00000465, -0.00002046, 0.00003503, -0.00005935, 0.00006919, -0.00006425, 0.00004085, 0.00000768, -0.00005703, 0.00005759, -0.00003433, -0.00002638, 0.00008112, -0.00007819, 0.00005939, -0.00007360, 0.00009133, -0.00011453, 0.00012300, -0.00011982, 0.00008126, -0.00009565, 0.00014425, -0.00017670, 0.00013916, -0.00019895, 0.00024844, -0.00034079, 0.00033604, -0.00078619, 0.00043783, 0.00131790, -0.00111037, 0.00043329, -0.00040880, 0.00027024, -0.00020999, 0.00015749, -0.00019227, 0.00014572, -0.00010036, 0.00009484, -0.00013280, 0.00012464, -0.00010291, 0.00007165, -0.00005370, 0.00004631, -0.00005621, 0.00003304, 0.00002297, -0.00005452, 0.00003896, -0.00000823, -0.00005181, 0.00010029, -0.00009826, 0.00008819, -0.00006405, 0.00002711, -0.00001367, -0.00000020, 0.00000271, },
		{ 0.00002389, -0.00002595, 0.00001925, -0.00000987, -0.00001494, 0.00004345, -0.00007080, 0.00008107, -0.00006258, 0.00000162, 0.00004012, -0.00006520, 0.00003827, 0.00002270, -0.00004054, 0.00002430, -0.00003073, 0.00004274, -0.00006990, 0.00009540, -0.00012105, 0.00008862, -0.00007469, 0.00010998, -0.00016787, 0.00013531, -0.00017758, 0.00022680, -0.00035821, 0.00036002, -0.00112191, 0.00165692, -0.00003781, -0.00051401, 0.00020062, -0.00024064, 0.00017297, -0.00013464, 0.00007488, -0.00011665, 0.00010567, -0.00006464, 0.00003419, -0.00006783, 0.00008625, -0.00009062, 0.00007375, -0.00005716, 0.00004772, -0.00006577, 0.00006386, -0.00001602, -0.00004308, 0.00007068, -0.00008523, 0.00005756, 0.00000720, -0.00003567, 0.00006088, -0.00007201, 0.00004993, -0.00004019, 0.00002492, -0.00001722, },
		{ 0.00001247, -0.00002515, 0.00003488, -0.00004433, 0.00004247, -0.00002731, -0.00000734, 0.00004099, -0.00006905, 0.00003765, 0.00000985, -0.00007179, 0.00009604, -0.00005736, 0.00003431, -0.00005321, 0.00006173, -0.00007073, 0.00006403, -0.00004970, 0.00000939, -0.00002749, 0.00006936, -0.00007772, 0.00002772, -0.00007542, 0.00008639, -0.00011059, 0.00004384, -0.00025981, -0.00043059, 0.00184717, -0.00100285, 0.00022156, -0.00025407, 0.00012044, -0.00009265, 0.00007176, -0.00011612, 0.00006823, -0.00003757, 0.00004701, -0.00008691, 0.00007028, -0.00004215, 0.00001356, -0.00000577, 0.00000832, -0.00001241, -0.00001741, 0.00004767, -0.00003776, -0.00001048, 0.00005203, -0.00010289, 0.00013426, -0.00009666, 0.00007350, -0.00003855, -0.00000317, 0.00000653, -0.00001344, 0.00000904, -0.00000349, },
		{ -0.00002341, 0.00001206, 0.00000307, -0.00002290, 0.00004401, -0.00006015, 0.00004705, -0.00002244, -0.00003482, 0.00005072, -0.00003709, -0.00002346, 0.00008451, -0.00007650, 0.00005435, -0.00007128, 0.00008628, -0.00011087, 0.00012883, -0.00014439, 0.00010832, -0.00010431, 0.00015149, -0.00020240, 0.00017097, -0.00022754, 0.00028020, -0.00038647, 0.00039971, -0.00097737, 0.00090207, 0.00086652, -0.00095285, 0.00040183, -0.00038589, 0.00025656, -0.00019554, 0.00013704, -0.00017417, 0.00014463, -0.00009850, 0.00007437, -0.00011083, 0.00011945, -0.00010745, 0.00007808, -0.00005832, 0.00005199, -0.00005770, 0.00002906, 0.00002152, -0.00005617, 0.00004652, -0.00003003, -0.00001488, 0.00008969, -0.00009511, 0.00010576, -0.00009918, 0.00005800, -0.00004470, 0.00002811, -0.00002319, 0.00002769, },
		{ -0.00003493, 0.00003786, -0.00003469, 0.00002659, -0.00000730, -0.00002756, 0.00004061, -0.00005039, 0.00000035, 0.00004684, -0.00008307, 0.00006211, 0.00000334, -0.00001675, -0.00000690, 0.00000195, 0.00000181, -0.00001922, 0.00004012, -0.00007956, 0.00005855, -0.00002738, 0.00004475, -0.00010654, 0.00007486, -0.00010354, 0.00013269, -0.00024113, 0.00018392, -0.00096302, 0.00181334, -0.00043271, -0.00023452, 0.00004183, -0.00012064, 0.00008174, -0.00005920, 0.00000132, -0.00004308, 0.00005241, -0.00002614, -0.00001566, -0.00000681, 0.00003367, -0.00004983, 0.00004123, -0.00002854, 0.00002568, -0.00004566, 0.00003953, 0.00000115, -0.00005765, 0.00008299, -0.00010788, 0.00011317, -0.00004071, 0.00001620, 0.00001982, -0.00005497, 0.00003972, -0.00003737, 0.00002635, -0.00002101, 0.00002840, },
		{ 0.00000017, 0.00001496, -0.00002623, 0.00003968, -0.00004586, 0.00001965, -0.00000185, -0.00004112, 0.00002788, 0.00001297, -0.00008042, 0.00011201, -0.00006898, 0.00004610, -0.00007099, 0.00007926, -0.00009393, 0.00009990, -0.00010274, 0.00006081, -0.00006215, 0.00011060, -0.00014169, 0.00009866, -0.00014727, 0.00017080, -0.00022054, 0.00019009, -0.00053100, 0.00001582, 0.00155633, -0.00104317, 0.00033013, -0.00032499, 0.00017859, -0.00013534, 0.00010129, -0.00014465, 0.00010753, -0.00006775, 0.00005717, -0.00009854, 0.00009813, -0.00007465, 0.00004277, -0.00003000, 0.00003012, -0.00002759, -0.00000765, 0.00003386, -0.00003014, -0.00001305, 0.00004254, -0.00008614, 0.00015227, -0.00011989, 0.00011419, -0.00009066, 0.00003682, -0.00002899, 0.00001647, -0.00001643, 0.00001900, -0.00001165, },
		{ 0.00004113, -0.00002952, 0.00002008, -0.00000048, -0.00002957, 0.00002544, -0.00002816, -0.00002364, 0.00005151, -0.00005090, -0.00000739, 0.00007399, -0.00006113, 0.00003356, -0.00005186, 0.00006088, -0.00008075, 0.00010112, -0.00013427, 0.00010737, -0.00008524, 0.00012186, -0.00018482, 0.00015903, -0.00020831, 0.00025441, -0.00035966, 0.00036229, -0.00102934, 0.00126129, 0.00039742, -0.00070929, 0.00029587, -0.00030784, 0.00019995, -0.00014696, 0.00008616, -0.00012383, 0.00011365, -0.00007563, 0.00003657, -0.00006484, 0.00008561, -0.00008527, 0.00006179, -0.00004423, 0.00004164, -0.00004611, 0.00001255, 0.00003219, -0.00006870, 0.00005950, -0.00005678, 0.00004092, 0.00005368, -0.00006047, 0.00008629, -0.00010518, 0.00006751, -0.00005851, 0.00004383, -0.00003962, 0.00004340, -0.00004366, },
		{ 0.00003357, -0.00003577, 0.00004333, -0.00004219, 0.00001101, -0.00000485, -0.00002581, -0.00000571, 0.00005201, -0.00009915, 0.00008511, -0.00001560, 0.00000507, -0.00003624, 0.00003160, -0.00003535, 0.00003043, -0.00002217, -0.00002408, 0.00001838, 0.00002504, -0.00003017, -0.00002833, -0.00000373, -0.00001034, 0.00001455, -0.00009035, -0.00005712, -0.00065665, 0.00176396, -0.00068433, -0.00001117, -0.00009283, -0.00001664, 0.00000388, 0.00000179, -0.00005704, 0.00002082, 0.00000376, 0.00000351, -0.00005017, 0.00004228, -0.00001461, -0.00000913, 0.00000739, 0.00000087, 0.00000348, -0.00002767, 0.00001810, 0.00001546, -0.00006961, 0.00008580, -0.00011182, 0.00015699, -0.00008242, 0.00006971, -0.00003799, -0.00001681, 0.00000912, -0.00001386, 0.00000830, -0.00000690, 0.00001335, -0.00002152, },
		{ -0.00001698, 0.00000922, 0.00000931, -0.00003365, 0.00001929, -0.00002485, -0.00001987, 0.00002436, 0.00000494, -0.00007565, 0.00011413, -0.00006682, 0.00004174, -0.00007138, 0.00007634, -0.00009137, 0.00010540, -0.00012732, 0.00009134, -0.00007501, 0.00012127, -0.00017155, 0.00013752, -0.00018456, 0.00021381, -0.00028018, 0.00027124, -0.00072937, 0.00045635, 0.00115199, -0.00093915, 0.00033762, -0.00032710, 0.00018465, -0.00013604, 0.00009268, -0.00013461, 0.00011292, -0.00007311, 0.00004435, -0.00008028, 0.00009373, -0.00007930, 0.00004927, -0.00003586, 0.00003621, -0.00002944, -0.00001139, 0.00003361, -0.00003520, -0.00000660, 0.00002074, -0.00004280, 0.00014112, -0.00011148, 0.00012165, -0.00012175, 0.00006538, -0.00005740, 0.00004363, -0.00004184, 0.00004172, -0.00003823, 0.00003313, },
		{ -0.00004674, 0.00004991, -0.00004286, 0.00001026, -0.00001280, -0.00001468, -0.00002131, 0.00005571, -0.00006949, 0.00001544, 0.00005591, -0.00003898, 0.00000465, -0.00002421, 0.00002629, -0.00003671, 0.00005092, -0.00009619, 0.00008320, -0.00004580, 0.00006449, -0.00013146, 0.00010941, -0.00014786, 0.00017864, -0.00026838, 0.00022741, -0.00092134, 0.00145009, -0.00000061, -0.00045147, 0.00016212, -0.00020748, 0.00012665, -0.00008738, 0.00002691, -0.00006150, 0.00006911, -0.00004600, -0.00000205, -0.00001344, 0.00003962, -0.00004906, 0.00003223, -0.00001909, 0.00002237, -0.00002875, -0.00000860, 0.00004720, -0.00008499, 0.00006961, -0.00007589, 0.00010027, 0.00000828, -0.00000908, 0.00003665, -0.00008087, 0.00005126, -0.00004878, 0.00003897, -0.00003816, 0.00004062, -0.00004440, 0.00005272, },
		{ -0.00001394, 0.00003056, -0.00004886, 0.00002898, -0.00003965, -0.00000064, -0.00000958, 0.00004916, -0.00010558, 0.00009887, -0.00002564, 0.00001595, -0.00005424, 0.00004783, -0.00005542, 0.00006137, -0.00007059, 0.00002568, -0.00001444, 0.00006292, -0.00009081, 0.00004075, -0.00007320, 0.00007136, -0.00009093, 0.00004696, -0.00030028, -0.00026773, 0.00153829, -0.00077556, 0.00012174, -0.00017363, 0.00004828, -0.00004218, 0.00003263, -0.00008473, 0.00005931, -0.00002714, 0.00001468, -0.00005903, 0.00006542, -0.00004282, 0.00001662, -0.00001511, 0.00002069, -0.00001022, -0.00001958, 0.00000741, 0.00001925, -0.00007288, 0.00007412, -0.00008895, 0.00017650, -0.00010046, 0.00010104, -0.00008884, 0.00002399, -0.00002694, 0.00001802, -0.00001955, 0.00001664, -0.00001330, 0.00000833, 0.00000963, },
		{ 0.00004429, -0.00003188, -0.00000109, -0.00000494, -0.00002628, -0.00000822, 0.00002678, -0.00001209, -0.00005928, 0.00010391, -0.00005268, 0.00002328, -0.00005701, 0.00005693, -0.00006688, 0.00008130, -0.00011931, 0.00009518, -0.00006251, 0.00009799, -0.00016059, 0.00013499, -0.00017690, 0.00020326, -0.00027347, 0.00026109, -0.00080196, 0.00080323, 0.00072512, -0.00074901, 0.00027308, -0.00027770, 0.00015206, -0.00010710, 0.00005818, -0.00009682, 0.00009206, -0.00006060, 0.00001842, -0.00004339, 0.00006545, -0.00006067, 0.00003511, -0.00002451, 0.00002813, -0.00002049, -0.00002515, 0.00004337, -0.00004960, 0.00000394, -0.00000489, 0.00001785, 0.00010889, -0.00007647, 0.00009184, -0.00012033, 0.00006962, -0.00006548, 0.00005452, -0.00005423, 0.00005152, -0.00005244, 0.00005781, -0.00004466, },
		{ 0.00005436, -0.00006507, 0.00004041, -0.00005165, 0.00000739, -0.00002166, 0.00005665, -0.00008442, 0.00003615, 0.00003924, -0.00001965, -0.00002225, 0.00000052, -0.00000345, 0.00000361, -0.00000253, -0.00004689, 0.00005017, -0.00000272, -0.00000049, -0.00006295, 0.00004273, -0.00006892, 0.00007936, -0.00014472, 0.00003263, -0.00067849, 0.00145045, -0.00027068, -0.00023830, 0.00004416, -0.00011696, 0.00006225, -0.00003900, -0.00001937, -0.00000766, 0.00002805, -0.00002345, -0.00002635, 0.00002498, 0.00000023, -0.00001544, 0.00000310, 0.00000591, 0.00000381, -0.00001483, -0.00002482, 0.00005707, -0.00009659, 0.00006833, -0.00007507, 0.00014715, -0.00002688, 0.00003524, -0.00002258, -0.00003931, 0.00001792, -0.00002127, 0.00001671, -0.00002074, 0.00002032, -0.00002402, 0.00004216, -0.00004136, },
		{ -0.00000247, -0.00002907, 0.00002157, -0.00005546, 0.00001773, -0.00000795, 0.00003637, -0.00009953, 0.00010005, -0.00002345, 0.00001233, -0.00005741, 0.00004762, -0.00005400, 0.00006518, -0.00009186, 0.00005545, -0.00002767, 0.00007291, -0.00011959, 0.00008014, -0.00011133, 0.00011613, -0.00015246, 0.00012795, -0.00047991, 0.00011818, 0.00120547, -0.00073298, 0.00016443, -0.00019552, 0.00007005, -0.00005426, 0.00003254, -0.00008037, 0.00006914, -0.00003787, 0.00000840, -0.00004408, 0.00006113, -0.00004629, 0.00002130, -0.00002044, 0.00002648, -0.00001262, -0.00002306, 0.00000905, 0.00001097, -0.00006745, 0.00005031, -0.00004069, 0.00017075, -0.00008992, 0.00009605, -0.00011294, 0.00004876, -0.00005141, 0.00004241, -0.00004293, 0.00003591, -0.00003718, 0.00004058, -0.00002122, 0.00001856, },
		{ -0.00006415, 0.00003400, -0.00004033, -0.00001272, -0.00000307, 0.00003037, -0.00003147, -0.00003806, 0.00008804, -0.00003372, -0.00000184, -0.00003611, 0.00003167, -0.00003345, 0.00004041, -0.00008773, 0.00007883, -0.00003327, 0.00005137, -0.00011803, 0.00009869, -0.00013232, 0.00014794, -0.00020965, 0.00016569, -0.00073348, 0.00099929, 0.00035676, -0.00053882, 0.00017879, -0.00020586, 0.00010370, -0.00006870, 0.00001734, -0.00004959, 0.00005976, -0.00004273, -0.00000611, -0.00000515, 0.00002975, -0.00003233, 0.00001077, -0.00000420, 0.00001320, -0.00000833, -0.00004054, 0.00005460, -0.00006535, 0.00000972, -0.00002189, 0.00008088, 0.00007261, -0.00003390, 0.00003778, -0.00009052, 0.00004938, -0.00005082, 0.00004482, -0.00004858, 0.00004256, -0.00004588, 0.00006316, -0.00005529, 0.00006260, },
		{ -0.00006049, 0.00004916, -0.00007819, 0.00002813, -0.00001933, 0.00004960, -0.00008959, 0.00004810, 0.00003077, -0.00001045, -0.00003958, 0.00001475, -0.00001861, 0.00002619, -0.00004112, -0.00000440, 0.00002417, 0.00002633, -0.00005111, -0.00000355, -0.00001559, -0.00000002, -0.00000950, -0.00003195, -0.00016393, -0.00036292, 0.00129329, -0.00039972, -0.00009978, -0.00003218, -0.00005566, 0.00002146, -0.00001384, -0.00004087, 0.00002529, 0.00000071, -0.00001460, -0.00002988, 0.00004059, -0.00002040, 0.00000324, -0.00001461, 0.00002198, -0.00000721, -0.00001018, -0.00003015, 0.00005581, -0.00009876, 0.00005206, -0.00004857, 0.00017309, -0.00003925, 0.00005143, -0.00006722, -0.00000063, -0.00001592, 0.00000975, -0.00001102, 0.00000182, -0.00000682, 0.00000841, 0.00001597, -0.00001816, 0.00003417, },
		{ 0.00000400, -0.00000454, -0.00005318, 0.00002826, -0.00000252, 0.00001702, -0.00008399, 0.00009140, -0.00001196, -0.00000297, -0.00004890, 0.00003593, -0.00003706, 0.00004596, -0.00008545, 0.00006225, -0.00002016, 0.00005436, -0.00011276, 0.00008334, -0.00011058, 0.00011477, -0.00015748, 0.00013165, -0.00055133, 0.00042328, 0.00084857, -0.00061009, 0.00014308, -0.00017376, 0.00006054, -0.00004358, 0.00001309, -0.00005456, 0.00005804, -0.00003517, -0.00000580, -0.00001710, 0.00003952, -0.00003193, 0.00000913, -0.00001127, 0.00002090, -0.00000744, -0.00003319, 0.00001786, -0.00000447, -0.00005942, 0.00002342, 0.00002290, 0.00015060, -0.00006187, 0.00005770, -0.00010396, 0.00004824, -0.00005423, 0.00004852, -0.00005120, 0.00003997, -0.00004535, 0.00006073, -0.00004409, 0.00004511, -0.00003830, },
		{ 0.00005573, -0.00007242, 0.00000601, 0.00000096, 0.00002953, -0.00004596, -0.00001960, 0.00007426, -0.00001826, -0.00002507, -0.00001736, 0.00001164, -0.00000612, 0.00000051, -0.00004895, 0.00005624, -0.00000317, 0.00000076, -0.00006402, 0.00004869, -0.00007272, 0.00007310, -0.00011906, 0.00002122, -0.00055092, 0.00103300, 0.00009540, -0.00035994, 0.00009211, -0.00013829, 0.00006017, -0.00003843, -0.00001279, -0.00000964, 0.00002983, -0.00002999, -0.00001810, 0.00002000, 0.00000244, -0.00000911, -0.00001104, 0.00001495, -0.00000031, -0.00000071, -0.00004924, 0.00005883, -0.00007478, 0.00000323, -0.00002007, 0.00013378, 0.00004850, -0.00000730, -0.00001686, -0.00004912, 0.00001670, -0.00002259, 0.00002115, -0.00003022, 0.00001951, -0.00002087, 0.00004843, -0.00004600, 0.00005730, -0.00007444, },
		{ 0.00003605, -0.00008750, 0.00004258, -0.00001292, 0.00003443, -0.00008404, 0.00004987, 0.00003198, -0.00001331, -0.00004544, 0.00001780, -0.00001839, 0.00002716, -0.00005603, 0.00002048, 0.00001436, 0.00003153, -0.00007351, 0.00002979, -0.00004701, 0.00003747, -0.00006152, 0.00003372, -0.00030559, -0.00005123, 0.00104267, -0.00041333, -0.00003481, -0.00006322, -0.00002615, 0.00000483, -0.00001084, -0.00003832, 0.00003621, -0.00001219, -0.00001677, -0.00001701, 0.00003544, -0.00002127, 0.00000405, -0.00001734, 0.00002652, -0.00000983, -0.00001432, -0.00002519, 0.00004386, -0.00009304, 0.00002437, 0.00000026, 0.00018147, -0.00003109, 0.00003223, -0.00008152, 0.00001873, -0.00003489, 0.00002959, -0.00003072, 0.00001638, -0.00002653, 0.00003841, -0.00001252, 0.00000952, 0.00000149, -0.00003539, },
		{ -0.00003648, -0.00004003, 0.00003443, 0.00000209, -0.00000258, -0.00006535, 0.00007923, 0.00000211, -0.00002326, -0.00003550, 0.00002183, -0.00001681, 0.00001712, -0.00006199, 0.00005396, -0.00000146, 0.00001903, -0.00008105, 0.00005951, -0.00008037, 0.00007748, -0.00011680, 0.00006713, -0.00050627, 0.00060154, 0.00053721, -0.00046406, 0.00009346, -0.00013220, 0.00003810, -0.00002635, -0.00000968, -0.00002272, 0.00003841, -0.00002898, -0.00001646, 0.00000726, 0.00001561, -0.00001299, -0.00000915, 0.00000420, 0.00001081, -0.00000177, -0.00004187, 0.00002538, -0.00001909, -0.00005715, 0.00000440, 0.00008931, 0.00013181, -0.00003799, 0.00000468, -0.00007147, 0.00002680, -0.00003720, 0.00003601, -0.00004311, 0.00002685, -0.00003329, 0.00006139, -0.00005004, 0.00005367, -0.00006053, 0.00003373, },
		{ -0.00009167, 0.00002233, 0.00000737, 0.00002146, -0.00005159, -0.00000837, 0.00006721, -0.00001153, -0.00004112, -0.00000522, 0.00000229, 0.00000598, -0.00002461, -0.00001786, 0.00004050, 0.00001339, -0.00003524, -0.00001959, 0.00000711, -0.00002294, 0.00000824, -0.00003923, -0.00012097, -0.00031160, 0.00093583, -0.00005062, -0.00023660, 0.00003356, -0.00008969, 0.00003182, -0.00002411, -0.00002453, 0.00001448, 0.00000921, -0.00002573, -0.00001516, 0.00002691, -0.00000878, 0.00000017, -0.00002185, 0.00002627, -0.00000785, -0.00000102, -0.00004767, 0.00005216, -0.00007483, -0.00001740, 0.00000342, 0.00017346, 0.00004175, -0.00001212, -0.00004986, -0.00001617, -0.00001152, 0.00000422, -0.00000351, -0.00001121, -0.00000527, 0.00001132, 0.00002248, -0.00002448, 0.00003365, -0.00006162, 0.00005708, },
		{ -0.00008252, 0.00005115, -0.00000493, 0.00001520, -0.00007147, 0.00004491, 0.00003927, -0.00002500, -0.00004300, 0.00001490, -0.00001016, 0.00001351, -0.00005040, 0.00002800, 0.00001906, 0.00001535, -0.00006746, 0.00003418, -0.00004770, 0.00003840, -0.00006890, 0.00003814, -0.00035835, 0.00019203, 0.00077041, -0.00035798, -0.00002117, -0.00006161, -0.00001862, 0.00000277, -0.00001969, -0.00002152, 0.00003235, -0.00001661, -0.00002158, 0.00000102, 0.00002033, -0.00001095, -0.00000699, -0.00000901, 0.00002290, -0.00000846, -0.00002143, -0.00001610, 0.00002703, -0.00008580, -0.00000626, 0.00006211, 0.00018457, -0.00001910, -0.00001159, -0.00006561, 0.00001432, -0.00003310, 0.00003119, -0.00003510, 0.00001524, -0.00002890, 0.00005402, -0.00003095, 0.00002764, -0.00002735, -0.00000397, 0.00000984, },
		{ -0.00002425, 0.00004028, 0.00000185, -0.00001701, -0.00004934, 0.00006940, 0.00001232, -0.00004216, -0.00002286, 0.00001248, -0.00000428, -0.00000655, -0.00003618, 0.00004315, 0.00001434, -0.00001549, -0.00004312, 0.00002722, -0.00004094, 0.00002745, -0.00005823, -0.00003153, -0.00037320, 0.00064879, 0.00030938, -0.00033644, 0.00004495, -0.00009102, 0.00001757, -0.00001486, -0.00002368, 0.00000289, 0.00002019, -0.00002593, -0.00001697, 0.00001939, 0.00000101, -0.00000132, -0.00002283, 0.00001761, 0.00000245, -0.00000119, -0.00004281, 0.00002488, -0.00002675, -0.00006622, -0.00000001, 0.00015163, 0.00012429, -0.00003869, -0.00003824, -0.00003486, -0.00000104, -0.00001212, 0.00001430, -0.00002686, 0.00000440, -0.00000685, 0.00004531, -0.00004021, 0.00004266, -0.00006164, 0.00004835, -0.00006233, },
		{ 0.00003350, 0.00001599, 0.00000757, -0.00004901, -0.00000410, 0.00006666, -0.00001393, -0.00004945, 0.00000163, 0.00000107, 0.00000365, -0.00003126, -0.00000084, 0.00003676, 0.00001092, -0.00004747, 0.00000306, -0.00001249, 0.00000113, -0.00002660, 0.00000202, -0.00021542, -0.00008057, 0.00076461, -0.00010506, -0.00016666, 0.00000513, -0.00006099, 0.00001754, -0.00002296, -0.00002050, 0.00002354, -0.00000285, -0.00002614, -0.00000349, 0.00002060, -0.00000646, -0.00000414, -0.00002054, 0.00002900, -0.00001040, -0.00000678, -0.00003842, 0.00003677, -0.00006821, -0.00004792, 0.00004390, 0.00020758, 0.00004188, -0.00004662, -0.00005170, -0.00000412, -0.00002314, 0.00001758, -0.00001777, -0.00000282, -0.00001898, 0.00003706, -0.00000204, -0.00000364, 0.00000502, -0.00003513, 0.00004127, -0.00009630, },
		{ 0.00005696, 0.00000081, -0.00000287, -0.00005716, 0.00003848, 0.00004696, -0.00004014, -0.00003693, 0.00001253, -0.00000404, -0.00000205, -0.00003556, 0.00002693, 0.00002821, -0.00000974, -0.00004487, 0.00002005, -0.00002843, 0.00001441, -0.00004426, -0.00000621, -0.00032176, 0.00033381, 0.00053216, -0.00027979, -0.00003096, -0.00004524, -0.00001977, 0.00000357, -0.00002842, -0.00000234, 0.00002307, -0.00001937, -0.00002126, 0.00001328, 0.00000752, -0.00000116, -0.00001984, 0.00000290, 0.00001686, -0.00000889, -0.00002456, -0.00001038, 0.00001230, -0.00008374, -0.00003211, 0.00012913, 0.00019427, -0.00002550, -0.00005660, -0.00003439, -0.00000497, -0.00001661, 0.00001833, -0.00002683, 0.00000048, -0.00001354, 0.00005178, -0.00003455, 0.00002916, -0.00004057, 0.00001987, -0.00001772, -0.00007018, },
		{ 0.00004741, -0.00000404, -0.00002447, -0.00003814, 0.00006422, 0.00001565, -0.00005661, -0.00001232, 0.00000885, -0.00000312, -0.00001669, -0.00001856, 0.00003904, 0.00001703, -0.00003503, -0.00001570, 0.00000413, -0.00001222, -0.00001140, -0.00001234, -0.00012122, -0.00020322, 0.00059629, 0.00016823, -0.00024463, 0.00001195, -0.00005974, 0.00000468, -0.00001262, -0.00002539, 0.00001760, 0.00000704, -0.00002634, -0.00000871, 0.00001881, -0.00000135, -0.00000156, -0.00002654, 0.00002478, -0.00000218, -0.00000636, -0.00003507, 0.00001482, -0.00002635, -0.00008653, 0.00000968, 0.00021342, 0.00012355, -0.00007082, -0.00005338, -0.00001220, -0.00001963, 0.00000678, -0.00000391, -0.00001449, -0.00001455, 0.00002162, 0.00002264, -0.00002392, 0.00002058, -0.00004528, 0.00004550, -0.00007623, -0.00001091, },
		{ 0.00002414, -0.00000816, -0.00004194, -0.00000321, 0.00006884, -0.00002227, -0.00005261, 0.00000778, 0.00000025, -0.00000406, -0.00002592, 0.00000565, 0.00004049, -0.00000490, -0.00003947, 0.00000497, -0.00001045, -0.00000100, -0.00002918, -0.00000176, -0.00024002, 0.00009286, 0.00057771, -0.00010594, -0.00013246, -0.00000303, -0.00004460, 0.00000969, -0.00002636, -0.00000920, 0.00002375, -0.00001103, -0.00002464, 0.00000723, 0.00001138, 0.00000022, -0.00001486, -0.00001181, 0.00002669, -0.00001234, -0.00001207, -0.00002785, 0.00001841, -0.00006045, -0.00008193, 0.00009455, 0.00024818, 0.00002725, -0.00009172, -0.00003005, -0.00001159, -0.00001732, 0.00001503, -0.00001855, -0.00000863, -0.00001561, 0.00004788, -0.00001626, 0.00000632, -0.00001517, -0.00000838, 0.00001775, -0.00009044, 0.00004074, },
		{ 0.00000224, -0.00001657, -0.00004456, 0.00003414, 0.00005084, -0.00005473, -0.00002959, 0.00001344, -0.00000612, -0.00000899, -0.00002337, 0.00002755, 0.00003031, -0.00002875, -0.00002210, 0.00000376, -0.00000703, -0.00001393, -0.00001231, -0.00006798, -0.00022533, 0.00037757, 0.00035420, -0.00020913, -0.00004361, -0.00002734, -0.00002075, 0.00000036, -0.00003002, 0.00001158, 0.00001430, -0.00002324, -0.00001382, 0.00001536, 0.00000368, 0.00000024, -0.00002662, 0.00001240, 0.00001202, -0.00001388, -0.00002016, -0.00001225, 0.00000341, -0.00008950, -0.00005119, 0.00020219, 0.00021123, -0.00006456, -0.00007923, -0.00000841, -0.00002318, 0.00000115, 0.00000233, -0.00001618, -0.00001757, 0.00001079, 0.00003745, -0.00002779, 0.00001716, -0.00003625, 0.00002920, -0.00003812, -0.00005653, 0.00006234, },
		{ -0.00001319, -0.00002681, -0.00003017, 0.00006207, 0.00001295, -0.00006699, -0.00000127, 0.00000577, -0.00000819, -0.00001491, -0.00001059, 0.00004239, 0.00000588, -0.00003567, -0.00000557, -0.00000184, -0.00000388, -0.00002629, 0.00000231, -0.00016875, -0.00004823, 0.00049222, 0.00009241, -0.00018536, -0.00000548, -0.00003762, -0.00000312, -0.00001550, -0.00001885, 0.00002363, -0.00000280, -0.00002587, 0.00000121, 0.00001257, 0.00000307, -0.00001039, -0.00002143, 0.00002649, -0.00000543, -0.00001329, -0.00002275, -0.00000144, -0.00002111, -0.00011395, 0.00002818, 0.00028635, 0.00011007, -0.00012179, -0.00003961, -0.00001014, -0.00002151, 0.00001146, -0.00001068, -0.00001409, -0.00001996, 0.00004068, 0.00000402, -0.00001210, 0.00000013, -0.00002233, 0.00003095, -0.00007892, -0.00000094, 0.00005455, },
		{ -0.00002217, -0.00003400, -0.00000208, 0.00006961, -0.00003290, -0.00005272, 0.00001641, -0.00000709, -0.00000642, -0.00001960, 0.00001059, 0.00004162, -0.00002142, -0.00002380, -0.00000245, 0.00000137, -0.00001635, -0.00001384, -0.00003477, -0.00020221, 0.00018926, 0.00041552, -0.00008722, -0.00011446, -0.00000273, -0.00003233, 0.00000114, -0.00002660, 0.00000135, 0.00002116, -0.00001877, -0.00001756, 0.00001054, 0.00000767, 0.00000223, -0.00002360, -0.00000167, 0.00002317, -0.00001729, -0.00001204, -0.00002141, 0.00000187, -0.00005517, -0.00011664, 0.00015457, 0.00030059, -0.00002259, -0.00012055, -0.00000475, -0.00002531, -0.00000419, 0.00000434, -0.00001243, -0.00002252, 0.00000098, 0.00004440, -0.00001968, 0.00000359, -0.00002060, 0.00000880, -0.00000412, -0.00007975, 0.00004612, 0.00002943, },
		{ -0.00002614, -0.00003367, 0.00003196, 0.00005002, -0.00006771, -0.00001967, 0.00001444, -0.00001449, -0.00000496, -0.00001920, 0.00003458, 0.00001990, -0.00003256, -0.00001085, -0.00000218, 0.00000212, -0.00002897, 0.00000396, -0.00011301, -0.00011285, 0.00035358, 0.00023450, -0.00015620, -0.00005108, -0.00001235, -0.00002012, -0.00000648, -0.00002469, 0.00001907, 0.00000676, -0.00002601, -0.00000367, 0.00001088, 0.00000711, -0.00000710, -0.00002528, 0.00001789, 0.00000788, -0.00002139, -0.00000996, -0.00002084, -0.00000059, -0.00010083, -0.00006787, 0.00029256, 0.00021925, -0.00013178, -0.00006892, -0.00000180, -0.00002748, 0.00000808, -0.00000570, -0.00001391, -0.00002692, 0.00003190, 0.00002119, -0.00002026, 0.00000187, -0.00002157, 0.00002572, -0.00004898, -0.00004418, 0.00006715, 0.00000035, },
		},
		{ { 0.00001500, 0.00001542, -0.00008836, 0.00000046, 0.00003180, -0.00001841, 0.00001420, -0.00000280, 0.00002091, -0.00001041, -0.00001177, 0.00000124, 0.00001323, 0.00000152, -0.00002336, -0.00000729, -0.00000410, -0.00001236, 0.00002609, -0.00003499, -0.00015675, 0.00004884, 0.00038831, 0.00005911, -0.00013694, -0.00002357, -0.00005816, 0.00000071, -0.00000845, 0.00000444, 0.00000101, -0.00001427, -0.00001506, -0.00001427, 0.00000101, 0.00000444, -0.00000845, 0.00000071, -0.00005816, -0.00002357, -0.00013694, 0.00005911, 0.00038831, 0.00004884, -0.00015675, -0.00003499, 0.00002609, -0.00001236, -0.00000410, -0.00000729, -0.00002336, 0.00000152, 0.00001323, 0.00000124, -0.00001177, -0.00001041, 0.00002091, -0.00000280, 0.00001420, -0.00001841, 0.00003180, 0.00000046, -0.00008836, 0.00001542, },
		{ 0.00001574, -0.00001071, -0.00008678, 0.00003884, 0.00000398, -0.00000634, 0.00000809, 0.00000398, 0.00001603, -0.00001641, -0.00000681, 0.00000594, 0.00001161, -0.00000644, -0.00002389, 0.00000022, -0.00001016, -0.00000157, 0.00002502, -0.00008349, -0.00012925, 0.00018145, 0.00032664, -0.00005335, -0.00009230, -0.00003277, -0.00004299, 0.00000467, -0.00000974, 0.00000928, -0.00000490, -0.00001624, -0.00001412, -0.00001229, 0.00000674, -0.00000424, 0.00000005, -0.00001736, -0.00005680, -0.00003435, -0.00016105, 0.00021869, 0.00036606, -0.00008873, -0.00013125, 0.00000030, 0.00001696, -0.00001687, -0.00000067, -0.00001467, -0.00001958, 0.00000979, 0.00001077, -0.00000309, -0.00001589, 0.00000039, 0.00001797, -0.00000293, 0.00001245, -0.00001873, 0.00004966, -0.00004809, -0.00006373, 0.00002778, },
		{ 0.00001695, -0.00004389, -0.00005816, 0.00005297, -0.00001795, 0.00000812, -0.00000099, 0.00001407, 0.00000473, -0.00001608, -0.00000263, 0.00001041, 0.00000584, -0.00001187, -0.00002226, 0.00000724, -0.00001849, 0.00001727, 0.00000161, -0.00011969, -0.00005553, 0.00027123, 0.00021518, -0.00010605, -0.00005537, -0.00004410, -0.00002239, -0.00000045, -0.00000430, 0.00000888, -0.00000992, -0.00001746, -0.00001232, -0.00000921, 0.00000905, -0.00001112, 0.00000879, -0.00004452, -0.00003814, -0.00007601, -0.00013010, 0.00038739, 0.00024565, -0.00018069, -0.00007536, 0.00001375, 0.00000747, -0.00001863, 0.00000128, -0.00002207, -0.00001120, 0.00001492, 0.00000639, -0.00000775, -0.00001631, 0.00001153, 0.00001014, 0.00000209, 0.00000400, -0.00000498, 0.00004546, -0.00008815, -0.00002305, 0.00002528, },
		{ 0.00001466, -0.00007270, -0.00001237, 0.00003953, -0.00002225, 0.00001387, -0.00000441, 0.00001975, -0.00000649, -0.00001273, 0.00000261, 0.00001079, 0.00000033, -0.00001726, -0.00001527, 0.00000779, -0.00002086, 0.00003389, -0.00004147, -0.00012236, 0.00003953, 0.00030151, 0.00009249, -0.00010804, -0.00003938, -0.00004422, -0.00000863, -0.00000517, 0.00000235, 0.00000553, -0.00001474, -0.00001650, -0.00001111, -0.00000439, 0.00000596, -0.00001110, 0.00000835, -0.00006832, -0.00001377, -0.00014395, -0.00001502, 0.00050294, 0.00005299, -0.00019208, -0.00002678, 0.00001638, -0.00000342, -0.00001551, -0.00000129, -0.00002556, -0.00000115, 0.00001679, 0.00000031, -0.00001131, -0.00001279, 0.00001990, 0.00000068, 0.00000946, -0.00000679, 0.00001813, 0.00001508, -0.00010244, 0.00001667, 0.00001387, },
		{ 0.00000226, -0.00008303, 0.00003130, 0.00000967, -0.00000935, 0.00000708, 0.00000238, 0.00001555, -0.00001153, -0.00001027, 0.00001037, 0.00000427, -0.00000062, -0.00002544, -0.00000005, -0.00000409, -0.00000681, 0.00002971, -0.00008265, -0.00009075, 0.00013191, 0.00027158, -0.00000673, -0.00008383, -0.00003951, -0.00003283, -0.00000528, -0.00000446, 0.00000616, 0.00000102, -0.00001895, -0.00001370, -0.00001019, 0.00000037, -0.00000039, -0.00000445, -0.00000646, -0.00007571, -0.00000459, -0.00021169, 0.00018889, 0.00049716, -0.00014125, -0.00013228, -0.00000761, 0.00002056, -0.00001882, -0.00000512, -0.00001134, -0.00002085, 0.00000597, 0.00001709, -0.00000786, -0.00001154, -0.00000706, 0.00002409, -0.00000773, 0.00001656, -0.00001470, 0.00004058, -0.00003489, -0.00008217, 0.00003784, 0.00000474, },
		{ -0.00002404, -0.00006484, 0.00005215, -0.00001588, 0.00000736, -0.00000437, 0.00001407, 0.00000449, -0.00001101, -0.00000741, 0.00001648, -0.00000491, 0.00000015, -0.00003190, 0.00001705, -0.00002297, 0.00002132, -0.00000350, -0.00009854, -0.00003921, 0.00020490, 0.00019347, -0.00006059, -0.00006057, -0.00004064, -0.00002025, -0.00000622, -0.00000064, 0.00000738, -0.00000492, -0.00002076, -0.00001069, -0.00000856, 0.00000274, -0.00000592, 0.00000402, -0.00003425, -0.00005851, -0.00003493, -0.00022837, 0.00043882, 0.00033629, -0.00025205, -0.00005347, -0.00000736, 0.00002274, -0.00003179, 0.00000603, -0.00002379, -0.00001047, 0.00001031, 0.00001426, -0.00001494, -0.00000971, 0.00000107, 0.00002154, -0.00001070, 0.00001934, -0.00001385, 0.00004848, -0.00008604, -0.00003424, 0.00003268, 0.00000622, },
		{ -0.00005713, -0.00002192, 0.00004131, -0.00002055, 0.00001216, -0.00000749, 0.00001943, -0.00000413, -0.00001112, 0.00000107, 0.00001295, -0.00000714, -0.00000625, -0.00002565, 0.00002076, -0.00003033, 0.00004224, -0.00004961, -0.00008503, 0.00002076, 0.00024315, 0.00009376, -0.00006972, -0.00005201, -0.00003256, -0.00001546, -0.00000411, 0.00000191, 0.00000744, -0.00001208, -0.00001965, -0.00000783, -0.00000727, 0.00000342, -0.00000917, 0.00000913, -0.00006808, -0.00001871, -0.00011991, -0.00013163, 0.00063725, 0.00006165, -0.00023545, -0.00001140, 0.00000077, 0.00000882, -0.00002861, 0.00000505, -0.00002714, -0.00000322, 0.00001672, 0.00000382, -0.00001481, -0.00001037, 0.00001382, 0.00000891, -0.00000274, 0.00001278, 0.00000048, 0.00002893, -0.00011301, 0.00001790, 0.00001056, 0.00001483, },
		{ -0.00007816, 0.00002504, 0.00001029, -0.00000516, 0.00000115, 0.00000285, 0.00001264, -0.00000460, -0.00001491, 0.00001580, -0.00000239, 0.00000180, -0.00002344, -0.00000191, 0.00000167, -0.00001080, 0.00003143, -0.00007798, -0.00005842, 0.00008753, 0.00023146, 0.00000873, -0.00005654, -0.00005201, -0.00001925, -0.00001677, 0.00000134, 0.00000276, 0.00000538, -0.00001799, -0.00001732, -0.00000404, -0.00000855, 0.00000508, -0.00001113, 0.00000604, -0.00009435, 0.00002325, -0.00024240, 0.00011335, 0.00066672, -0.00021013, -0.00012584, -0.00001934, 0.00002597, -0.00002312, -0.00000640, -0.00001227, -0.00001628, -0.00000356, 0.00002625, -0.00001383, -0.00000602, -0.00001449, 0.00002924, -0.00001166, 0.00001628, -0.00000327, 0.00002550, -0.00001992, -0.00009736, 0.00004610, -0.00000699, 0.00001548, },
		{ -0.00006865, 0.00004943, -0.00001556, 0.00001232, -0.00001339, 0.00001762, 0.00000023, -0.00000172, -0.00001610, 0.00002691, -0.00001899, 0.00001243, -0.00004015, 0.00002618, -0.00002841, 0.00002732, -0.00001270, -0.00007154, -0.00003409, 0.00015698, 0.00016748, -0.00003393, -0.00004813, -0.00004543, -0.00001268, -0.00001517, 0.00000474, 0.00000448, -0.00000039, -0.00002056, -0.00001444, -0.00000082, -0.00001071, 0.00000656, -0.00001019, -0.00001241, -0.00009211, 0.00002451, -0.00033303, 0.00046277, 0.00046431, -0.00034790, -0.00001538, -0.00003971, 0.00004529, -0.00005095, 0.00001657, -0.00003185, -0.00000150, -0.00000380, 0.00002970, -0.00002799, 0.00000271, -0.00001417, 0.00003663, -0.00002766, 0.00003547, -0.00001863, 0.00004571, -0.00008008, -0.00004340, 0.00003723, -0.00000357, -0.00000518, },
		{ -0.00002782, 0.00003861, -0.00001666, 0.00001271, -0.00001448, 0.00002185, -0.00000514, -0.00000540, -0.00000487, 0.00002081, -0.00002053, 0.00000813, -0.00003681, 0.00003378, -0.00004056, 0.00005161, -0.00006023, -0.00004445, -0.00000669, 0.00020790, 0.00007534, -0.00003439, -0.00005446, -0.00002681, -0.00001782, -0.00000603, 0.00000339, 0.00000707, -0.00000803, -0.00002146, -0.00000819, -0.00000383, -0.00000661, 0.00000098, -0.00000131, -0.00005239, -0.00004398, -0.00006180, -0.00028384, 0.00077560, 0.00008854, -0.00029557, 0.00001727, -0.00002516, 0.00002769, -0.00004630, 0.00001511, -0.00003236, 0.00000027, 0.00000857, 0.00001496, -0.00002478, -0.00000103, 0.00000111, 0.00002312, -0.00002248, 0.00003789, -0.00001749, 0.00003887, -0.00011826, 0.00001880, 0.00000516, 0.00001621, -0.00004281, },
		{ 0.00002127, 0.00000591, 0.00000393, -0.00000608, 0.00000216, 0.00001032, 0.00000116, -0.00001749, 0.00001858, -0.00000446, -0.00000206, -0.00001638, -0.00000624, 0.00000814, -0.00001522, 0.00003287, -0.00007337, -0.00002762, 0.00004292, 0.00020890, -0.00000135, -0.00002019, -0.00006266, -0.00000757, -0.00002588, 0.00000497, 0.00000124, 0.00000647, -0.00001296, -0.00002372, 0.00000342, -0.00001635, 0.00000824, -0.00001537, 0.00001406, -0.00010454, 0.00003762, -0.00023978, -0.00001397, 0.00086897, -0.00028610, -0.00011927, -0.00002916, 0.00002879, -0.00002656, -0.00000688, -0.00001507, -0.00000916, -0.00001427, 0.00003331, -0.00001674, -0.00000319, -0.00001814, 0.00002992, -0.00001110, 0.00000849, 0.00001615, 0.00000617, -0.00000574, -0.00010712, 0.00005155, -0.00001866, 0.00002856, -0.00007319, },
		{ 0.00004670, -0.00001867, 0.00002283, -0.00002607, 0.00002247, -0.00000573, 0.00000909, -0.00002573, 0.00003797, -0.00003172, 0.00002053, -0.00004371, 0.00003223, -0.00003271, 0.00003296, -0.00002284, -0.00004344, -0.00003369, 0.00011815, 0.00014708, -0.00002904, -0.00002150, -0.00005411, -0.00000354, -0.00002384, 0.00000898, 0.00000299, 0.00000058, -0.00001434, -0.00002484, 0.00001381, -0.00002990, 0.00002445, -0.00003133, 0.00001706, -0.00013252, 0.00009054, -0.00042523, 0.00044190, 0.00063627, -0.00046788, 0.00003734, -0.00008159, 0.00007372, -0.00007536, 0.00003306, -0.00004678, 0.00001568, -0.00002587, 0.00005004, -0.00004412, 0.00001811, -0.00003107, 0.00005206, -0.00004457, 0.00004744, -0.00001599, 0.00003631, -0.00007110, -0.00004908, 0.00003811, -0.00001159, 0.00001186, -0.00007028, },
		{ 0.00003358, -0.00001367, 0.00001776, -0.00002596, 0.00002701, -0.00000941, 0.00000394, -0.00001448, 0.00003286, -0.00003690, 0.00002232, -0.00004577, 0.00004473, -0.00005007, 0.00006087, -0.00007204, -0.00000314, -0.00003843, 0.00018721, 0.00004996, -0.00001135, -0.00004381, -0.00002724, -0.00001643, -0.00000958, 0.00000526, 0.00000648, -0.00000578, -0.00001818, -0.00001577, 0.00000877, -0.00002608, 0.00002117, -0.00002489, -0.00001968, -0.00009005, 0.00002590, -0.00045639, 0.00089695, 0.00014861, -0.00038023, 0.00006396, -0.00006463, 0.00005566, -0.00007095, 0.00003190, -0.00004494, 0.00001300, -0.00001014, 0.00003475, -0.00004167, 0.00001527, -0.00001769, 0.00004224, -0.00004726, 0.00006401, -0.00003126, 0.00004213, -0.00011720, 0.00001956, -0.00000233, 0.00001907, -0.00003055, -0.00003045, },
		{ -0.00000185, 0.00001446, -0.00001083, -0.00000238, 0.00001082, 0.00000353, -0.00001592, 0.00001657, -0.00000015, -0.00001249, -0.00000519, -0.00001206, 0.00001489, -0.00002037, 0.00003511, -0.00007067, 0.00000335, -0.00000277, 0.00020083, -0.00002705, 0.00001449, -0.00006504, -0.00000144, -0.00003046, 0.00000545, 0.00000199, 0.00000462, -0.00000577, -0.00002949, 0.00000785, -0.00001843, 0.00000459, -0.00001219, 0.00001149, -0.00009880, 0.00002926, -0.00019611, -0.00019000, 0.00108930, -0.00035416, -0.00012309, -0.00002943, 0.00002328, -0.00002467, -0.00001036, -0.00001566, -0.00000446, -0.00002046, 0.00003307, -0.00001256, -0.00000647, -0.00001498, 0.00002365, -0.00000385, -0.00000764, 0.00004153, -0.00001379, 0.00000443, -0.00010941, 0.00005346, -0.00002953, 0.00004101, -0.00006868, 0.00002014, },
		{ -0.00002507, 0.00003658, -0.00003883, 0.00002539, -0.00001036, 0.00001814, -0.00003288, 0.00004574, -0.00003884, 0.00002119, -0.00004072, 0.00003398, -0.00003456, 0.00003642, -0.00003122, -0.00001836, -0.00003274, 0.00008208, 0.00013661, -0.00004285, 0.00001194, -0.00006066, 0.00000275, -0.00002883, 0.00000965, 0.00000495, -0.00000371, -0.00000172, -0.00004058, 0.00003250, -0.00005096, 0.00004344, -0.00005423, 0.00004727, -0.00017019, 0.00014870, -0.00048637, 0.00036394, 0.00085107, -0.00060492, 0.00009814, -0.00012762, 0.00010381, -0.00010189, 0.00005322, -0.00006634, 0.00003791, -0.00005139, 0.00007028, -0.00005887, 0.00003227, -0.00004540, 0.00006369, -0.00005688, 0.00005115, -0.00000455, 0.00002149, -0.00006128, -0.00004939, 0.00003450, -0.00001700, 0.00002627, -0.00007025, 0.00004492, },
		{ -0.00001343, 0.00002739, -0.00004025, 0.00003304, -0.00001499, 0.00001457, -0.00002515, 0.00004569, -0.00005173, 0.00003202, -0.00004941, 0.00005137, -0.00005698, 0.00006811, -0.00008265, 0.00003501, -0.00006869, 0.00017230, 0.00002761, -0.00000486, -0.00002175, -0.00003097, -0.00001413, -0.00001191, 0.00000495, 0.00000843, -0.00000859, -0.00000690, -0.00003241, 0.00003100, -0.00005442, 0.00005042, -0.00006009, 0.00002726, -0.00015173, 0.00013161, -0.00062794, 0.00097867, 0.00025418, -0.00049334, 0.00012992, -0.00011799, 0.00009295, -0.00010315, 0.00005669, -0.00006647, 0.00003597, -0.00003882, 0.00006137, -0.00006343, 0.00003624, -0.00003980, 0.00006307, -0.00007262, 0.00008561, -0.00003684, 0.00003729, -0.00011015, 0.00002112, -0.00001214, 0.00002368, -0.00002096, -0.00003043, 0.00002810, },
		{ 0.00002304, -0.00001050, -0.00001177, 0.00001466, 0.00000205, -0.00001008, 0.00000979, 0.00000972, -0.00002716, 0.00000712, -0.00001693, 0.00002035, -0.00002528, 0.00003769, -0.00006964, 0.00003344, -0.00004687, 0.00020058, -0.00005724, 0.00003875, -0.00005587, -0.00000171, -0.00003075, 0.00000387, 0.00000357, 0.00000204, -0.00000001, -0.00003022, 0.00000475, -0.00001079, -0.00000932, 0.00000205, -0.00000581, -0.00007271, -0.00000611, -0.00011173, -0.00040184, 0.00130561, -0.00039591, -0.00014864, -0.00001258, 0.00000394, -0.00001308, -0.00002081, -0.00000963, -0.00000718, -0.00001713, 0.00002174, 0.00000103, -0.00001752, -0.00000384, 0.00000992, 0.00000998, -0.00003026, 0.00006849, -0.00002973, 0.00000756, -0.00010300, 0.00005176, -0.00003919, 0.00005242, -0.00006490, 0.00002104, -0.00001071, },
		{ 0.00004997, -0.00004770, 0.00002344, -0.00001108, 0.00002280, -0.00003500, 0.00004746, -0.00003788, 0.00001348, -0.00003158, 0.00003190, -0.00003389, 0.00003705, -0.00003643, 0.00000126, -0.00002849, 0.00004635, 0.00013514, -0.00006786, 0.00004325, -0.00005953, 0.00000291, -0.00002813, 0.00000582, 0.00001072, -0.00001281, 0.00001518, -0.00005786, 0.00005026, -0.00006846, 0.00005793, -0.00007242, 0.00007018, -0.00019486, 0.00018532, -0.00050239, 0.00022539, 0.00109819, -0.00074498, 0.00015600, -0.00016940, 0.00012904, -0.00012532, 0.00007259, -0.00008589, 0.00005969, -0.00007407, 0.00008452, -0.00006742, 0.00004089, -0.00005317, 0.00006799, -0.00006126, 0.00004466, 0.00001438, 0.00000452, -0.00005371, -0.00004242, 0.00002583, -0.00001915, 0.00003730, -0.00006873, 0.00004405, -0.00003308, },
		{ 0.00003947, -0.00005393, 0.00003681, -0.00001892, 0.00002327, -0.00003348, 0.00005524, -0.00006053, 0.00003373, -0.00004572, 0.00005240, -0.00005990, 0.00007166, -0.00008972, 0.00006628, -0.00009235, 0.00015653, 0.00001373, -0.00001373, 0.00000621, -0.00003322, -0.00001462, -0.00001001, 0.00000012, 0.00001510, -0.00001846, 0.00001316, -0.00005720, 0.00006039, -0.00008812, 0.00008475, -0.00010195, 0.00008162, -0.00021907, 0.00023896, -0.00077504, 0.00100122, 0.00041112, -0.00063247, 0.00021135, -0.00018170, 0.00013669, -0.00014080, 0.00008810, -0.00009555, 0.00006681, -0.00007362, 0.00009032, -0.00008588, 0.00005776, -0.00006094, 0.00008122, -0.00009343, 0.00009772, -0.00003207, 0.00002504, -0.00009901, 0.00002516, -0.00002476, 0.00003043, -0.00001468, -0.00002819, 0.00002315, -0.00001595, },
		{ -0.00000460, -0.00002479, 0.00002041, -0.00000178, -0.00000179, 0.00000036, 0.00002213, -0.00004183, 0.00001645, -0.00001798, 0.00002262, -0.00002839, 0.00003916, -0.00006875, 0.00005997, -0.00008491, 0.00020060, -0.00008254, 0.00004786, -0.00003587, -0.00000624, -0.00002915, 0.00000286, 0.00000330, 0.00000189, 0.00000035, -0.00002201, -0.00000892, 0.00000846, -0.00003452, 0.00002827, -0.00003850, -0.00002641, -0.00006640, 0.00000447, -0.00062822, 0.00149159, -0.00039333, -0.00020501, 0.00002653, -0.00003260, 0.00001089, -0.00004089, 0.00000609, -0.00002061, -0.00000165, -0.00000160, 0.00002358, -0.00003554, 0.00001403, -0.00000939, 0.00002789, -0.00005521, 0.00009132, -0.00003724, 0.00000187, -0.00008809, 0.00004739, -0.00004770, 0.00006262, -0.00006207, 0.00002320, -0.00001863, 0.00002752, },
		{ -0.00004971, 0.00001516, -0.00000688, 0.00002212, -0.00003134, 0.00004250, -0.00002902, -0.00000078, -0.00001889, 0.00002803, -0.00003184, 0.00003544, -0.00003857, 0.00001498, -0.00002081, 0.00001204, 0.00013799, -0.00009483, 0.00006463, -0.00004757, -0.00000396, -0.00002195, -0.00000140, 0.00001868, -0.00002416, 0.00003215, -0.00007133, 0.00006141, -0.00007691, 0.00006262, -0.00007996, 0.00007877, -0.00019834, 0.00019089, -0.00046707, 0.00003347, 0.00135784, -0.00086864, 0.00019726, -0.00019712, 0.00014198, -0.00013964, 0.00008571, -0.00009972, 0.00007476, -0.00008758, 0.00008764, -0.00006611, 0.00004098, -0.00005196, 0.00006325, -0.00005660, 0.00002889, 0.00003688, -0.00001000, -0.00005154, -0.00002691, 0.00001225, -0.00001778, 0.00004437, -0.00006565, 0.00004334, -0.00004034, 0.00005943, },
		{ -0.00006297, 0.00003521, -0.00001842, 0.00002718, -0.00003656, 0.00005827, -0.00006032, 0.00002603, -0.00003472, 0.00004804, -0.00005853, 0.00007072, -0.00009193, 0.00008830, -0.00010639, 0.00013672, 0.00000889, -0.00003247, 0.00003264, -0.00002913, -0.00002088, -0.00000196, -0.00001020, 0.00002700, -0.00003513, 0.00003999, -0.00008632, 0.00009182, -0.00012141, 0.00011800, -0.00014314, 0.00013401, -0.00027971, 0.00033029, -0.00087680, 0.00095266, 0.00061598, -0.00078760, 0.00029905, -0.00024826, 0.00018093, -0.00017912, 0.00012197, -0.00012779, 0.00010021, -0.00010825, 0.00011552, -0.00010388, 0.00007506, -0.00007653, 0.00009243, -0.00010539, 0.00009746, -0.00001758, 0.00000837, -0.00008696, 0.00003370, -0.00004060, 0.00003964, -0.00001230, -0.00002380, 0.00001867, -0.00001969, 0.00005050, },
		{ -0.00003831, 0.00002514, -0.00000517, 0.00000588, -0.00000840, 0.00003299, -0.00005184, 0.00001908, -0.00001306, 0.00002036, -0.00002833, 0.00003797, -0.00006611, 0.00007994, -0.00011258, 0.00019480, -0.00009733, 0.00004195, -0.00000944, -0.00001086, -0.00002915, 0.00000554, -0.00000152, 0.00000698, -0.00000756, -0.00000302, -0.00003342, 0.00003809, -0.00006890, 0.00006399, -0.00008325, 0.00003495, -0.00014331, 0.00013656, -0.00084300, 0.00162145, -0.00033303, -0.00029594, 0.00008844, -0.00008597, 0.00004691, -0.00007067, 0.00003199, -0.00004505, 0.00002518, -0.00003448, 0.00005175, -0.00005728, 0.00003522, -0.00003046, 0.00004571, -0.00007718, 0.00010457, -0.00003366, -0.00001225, -0.00006665, 0.00004235, -0.00005572, 0.00007180, -0.00006055, 0.00002615, -0.00002455, 0.00002764, 0.00000508, },
		{ 0.00000131, 0.00000123, 0.00001713, -0.00002326, 0.00003247, -0.00001487, -0.00001763, -0.00000660, 0.00002518, -0.00003016, 0.00003301, -0.00003891, 0.00002394, -0.00001157, -0.00001765, 0.00013908, -0.00011545, 0.00007148, -0.00002512, -0.00001597, -0.00001269, -0.00000916, 0.00002573, -0.00003387, 0.00004416, -0.00007574, 0.00006123, -0.00007247, 0.00005417, -0.00007321, 0.00006910, -0.00017679, 0.00016241, -0.00038378, -0.00019425, 0.00160351, -0.00095444, 0.00020859, -0.00020199, 0.00013632, -0.00013965, 0.00008772, -0.00010273, 0.00007793, -0.00008747, 0.00007695, -0.00005377, 0.00003187, -0.00004167, 0.00005018, -0.00004445, 0.00000747, 0.00005730, -0.00001741, -0.00005692, -0.00000295, -0.00000501, -0.00001340, 0.00004740, -0.00006106, 0.00004195, -0.00004476, 0.00006257, -0.00004401, },
		{ 0.00002637, -0.00001203, 0.00002483, -0.00003307, 0.00005349, -0.00005061, 0.00001019, -0.00001863, 0.00004008, -0.00005376, 0.00006573, -0.00008935, 0.00010060, -0.00011038, 0.00011311, 0.00000983, -0.00005335, 0.00005055, -0.00001563, -0.00003390, 0.00001210, -0.00002498, 0.00004254, -0.00005593, 0.00006903, -0.00011381, 0.00011869, -0.00014766, 0.00014339, -0.00017578, 0.00017484, -0.00032181, 0.00039059, -0.00091925, 0.00083186, 0.00085494, -0.00094173, 0.00037956, -0.00030736, 0.00021767, -0.00021151, 0.00015227, -0.00015690, 0.00012923, -0.00013561, 0.00013098, -0.00011284, 0.00008415, -0.00008313, 0.00009395, -0.00010639, 0.00008491, 0.00000309, -0.00000805, -0.00007762, 0.00004849, -0.00005960, 0.00005125, -0.00001421, -0.00001712, 0.00001385, -0.00002226, 0.00005676, -0.00006416, },
		{ 0.00002554, -0.00000508, 0.00000979, -0.00001322, 0.00003856, -0.00005362, 0.00001303, -0.00000172, 0.00001349, -0.00002455, 0.00003321, -0.00006045, 0.00009134, -0.00012725, 0.00018011, -0.00010049, 0.00002537, 0.00001688, -0.00001087, -0.00003397, 0.00001420, -0.00001242, 0.00001858, -0.00002433, 0.00002559, -0.00006577, 0.00007375, -0.00010742, 0.00010374, -0.00013329, 0.00010232, -0.00022418, 0.00026490, -0.00102005, 0.00167507, -0.00021004, -0.00041768, 0.00016796, -0.00015135, 0.00009104, -0.00010707, 0.00006539, -0.00007738, 0.00005912, -0.00007142, 0.00008001, -0.00007793, 0.00005508, -0.00004865, 0.00005889, -0.00009134, 0.00010469, -0.00001912, -0.00003198, -0.00004222, 0.00003940, -0.00006422, 0.00008035, -0.00006087, 0.00002992, -0.00002862, 0.00002492, 0.00001505, -0.00004824, },
		{ 0.00001045, 0.00001053, -0.00001369, 0.00002064, 0.00000039, -0.00003219, 0.00000122, 0.00002580, -0.00003039, 0.00003128, -0.00003915, 0.00003002, -0.00000340, -0.00003911, 0.00013317, -0.00012452, 0.00006349, 0.00000405, -0.00002906, -0.00000415, -0.00001383, 0.00002841, -0.00003805, 0.00004695, -0.00006759, 0.00004760, -0.00005431, 0.00003241, -0.00005203, 0.00004145, -0.00013181, 0.00010423, -0.00026488, -0.00043251, 0.00180626, -0.00098317, 0.00018040, -0.00017878, 0.00010881, -0.00012277, 0.00007596, -0.00009212, 0.00006676, -0.00007253, 0.00005316, -0.00003218, 0.00001550, -0.00002468, 0.00003181, -0.00002849, -0.00001439, 0.00006987, -0.00001453, -0.00006995, 0.00002753, -0.00002345, -0.00000732, 0.00004689, -0.00005536, 0.00003948, -0.00004532, 0.00005904, -0.00003216, -0.00001521, },
		{ -0.00000029, 0.00001682, -0.00002384, 0.00004199, -0.00003355, -0.00001014, -0.00000128, 0.00003129, -0.00004729, 0.00005806, -0.00008329, 0.00010444, -0.00010613, 0.00008845, 0.00001139, -0.00006900, 0.00005548, 0.00000734, -0.00005206, 0.00002978, -0.00004123, 0.00005831, -0.00007636, 0.00009436, -0.00013318, 0.00013471, -0.00016117, 0.00015537, -0.00019350, 0.00019663, -0.00033697, 0.00041077, -0.00089818, 0.00064949, 0.00110513, -0.00107339, 0.00043769, -0.00034807, 0.00023867, -0.00023120, 0.00017263, -0.00017631, 0.00014727, -0.00014975, 0.00013253, -0.00011014, 0.00008305, -0.00007944, 0.00008544, -0.00009708, 0.00006321, 0.00002429, -0.00001913, -0.00007390, 0.00007014, -0.00008072, 0.00006449, -0.00002029, -0.00000813, 0.00000778, -0.00002150, 0.00005571, -0.00005624, 0.00001055, },
		{ 0.00000052, 0.00000791, -0.00001212, 0.00003677, -0.00004588, -0.00000118, 0.00001439, 0.00000342, -0.00001758, 0.00002494, -0.00005166, 0.00009374, -0.00012869, 0.00015680, -0.00009476, 0.00000496, 0.00003644, -0.00000285, -0.00004514, 0.00002925, -0.00002892, 0.00003567, -0.00004782, 0.00005969, -0.00010021, 0.00010882, -0.00014323, 0.00014034, -0.00018006, 0.00016479, -0.00029490, 0.00037023, -0.00113825, 0.00164226, -0.00003010, -0.00055841, 0.00025439, -0.00021990, 0.00013628, -0.00014431, 0.00010086, -0.00011174, 0.00009350, -0.00010535, 0.00010220, -0.00009252, 0.00006912, -0.00005989, 0.00006396, -0.00009475, 0.00009115, 0.00000312, -0.00005249, -0.00001916, 0.00004136, -0.00007401, 0.00008855, -0.00006355, 0.00003490, -0.00003191, 0.00002194, 0.00002139, -0.00005099, 0.00001918, },
		{ 0.00000568, -0.00000610, 0.00001066, 0.00001248, -0.00004022, 0.00000188, 0.00003101, -0.00003311, 0.00003128, -0.00004071, 0.00003493, 0.00000146, -0.00004985, 0.00011751, -0.00012086, 0.00004441, 0.00003384, -0.00003833, -0.00000014, -0.00001232, 0.00002417, -0.00003419, 0.00003848, -0.00004643, 0.00002182, -0.00002509, 0.00000059, -0.00001991, 0.00000038, -0.00007019, 0.00002690, -0.00012917, -0.00065292, 0.00194005, -0.00094228, 0.00010978, -0.00012775, 0.00006063, -0.00009005, 0.00005103, -0.00006837, 0.00004244, -0.00004530, 0.00002031, -0.00000577, -0.00000411, -0.00000511, 0.00001253, -0.00001336, -0.00003136, 0.00007041, -0.00000092, -0.00008821, 0.00006089, -0.00003955, -0.00000150, 0.00004388, -0.00004938, 0.00003625, -0.00004244, 0.00005052, -0.00001772, -0.00003019, 0.00001721, },
		{ 0.00000565, -0.00001156, 0.00002678, -0.00001324, -0.00003003, 0.00001307, 0.00002444, -0.00004088, 0.00004945, -0.00007575, 0.00010213, -0.00009668, 0.00006645, 0.00000864, -0.00007442, 0.00004670, 0.00003663, -0.00007151, 0.00004708, -0.00005485, 0.00007012, -0.00009150, 0.00011032, -0.00013919, 0.00013574, -0.00015881, 0.00015115, -0.00019306, 0.00019591, -0.00032221, 0.00038952, -0.00081992, 0.00042649, 0.00133816, -0.00116059, 0.00045991, -0.00036150, 0.00023768, -0.00023307, 0.00017812, -0.00018111, 0.00014989, -0.00014756, 0.00011909, -0.00009602, 0.00007236, -0.00006674, 0.00006905, -0.00008065, 0.00003764, 0.00003977, -0.00002096, -0.00007688, 0.00009756, -0.00010165, 0.00007769, -0.00002968, 0.00000269, 0.00000006, -0.00001638, 0.00004685, -0.00004052, -0.00000966, 0.00001426, },
		{ 0.00000015, -0.00000514, 0.00002786, -0.00003003, -0.00002056, 0.00003192, -0.00000742, -0.00000882, 0.00001424, -0.00004080, 0.00008838, -0.00011897, 0.00012784, -0.00008512, -0.00001223, 0.00004455, 0.00001405, -0.00006177, 0.00004879, -0.00004841, 0.00005509, -0.00007357, 0.00009307, -0.00012966, 0.00013613, -0.00016947, 0.00016680, -0.00021529, 0.00021225, -0.00034324, 0.00043774, -0.00118544, 0.00152518, 0.00019036, -0.00069975, 0.00033328, -0.00028050, 0.00017396, -0.00017521, 0.00013158, -0.00014101, 0.00012102, -0.00012949, 0.00011316, -0.00009743, 0.00007434, -0.00006190, 0.00005961, -0.00008726, 0.00006669, 0.00002731, -0.00006820, -0.00000145, 0.00005025, -0.00008519, 0.00009617, -0.00006870, 0.00004145, -0.00003571, 0.00002127, 0.00002107, -0.00004469, 0.00000562, 0.00001179, },
		{ -0.00000322, 0.00000535, 0.00001837, -0.00003944, -0.00000510, 0.00004006, -0.00003769, 0.00003316, -0.00004420, 0.00003959, 0.00000195, -0.00004943, 0.00009240, -0.00010708, 0.00002060, 0.00005766, -0.00003970, -0.00000314, -0.00000322, 0.00001241, -0.00002209, 0.00001965, -0.00001522, -0.00001143, 0.00000955, -0.00003536, 0.00001690, -0.00004640, -0.00000227, -0.00005529, 0.00000212, -0.00082925, 0.00198680, -0.00082925, 0.00000212, -0.00005529, -0.00000227, -0.00004640, 0.00001690, -0.00003536, 0.00000955, -0.00001143, -0.00001522, 0.00001965, -0.00002209, 0.00001241, -0.00000322, -0.00000314, -0.00003970, 0.00005766, 0.00002060, -0.00010708, 0.00009240, -0.00004943, 0.00000195, 0.00003959, -0.00004420, 0.00003316, -0.00003769, 0.00004006, -0.00000510, -0.00003944, 0.00001837, 0.00000535, },
		{ 0.00000015, 0.00001179, 0.00000562, -0.00004469, 0.00002107, 0.00002127, -0.00003571, 0.00004145, -0.00006870, 0.00009617, -0.00008519, 0.00005025, -0.00000145, -0.00006820, 0.00002731, 0.00006669, -0.00008726, 0.00005961, -0.00006190, 0.00007434, -0.00009743, 0.00011316, -0.00012949, 0.00012102, -0.00014101, 0.00013158, -0.00017521, 0.00017396, -0.00028050, 0.00033328, -0.00069975, 0.00019036, 0.00152518, -0.00118544, 0.00043774, -0.00034324, 0.00021225, -0.00021529, 0.00016680, -0.00016947, 0.00013613, -0.00012966, 0.00009307, -0.00007357, 0.00005509, -0.00004841, 0.00004879, -0.00006177, 0.00001405, 0.00004455, -0.00001223, -0.00008512, 0.00012784, -0.00011897, 0.00008838, -0.00004080, 0.00001424, -0.00000882, -0.00000742, 0.00003192, -0.00002056, -0.00003003, 0.00002786, -0.00000514, },
		{ 0.00000565, 0.00001426, -0.00000966, -0.00004052, 0.00004685, -0.00001638, 0.00000006, 0.00000269, -0.00002968, 0.00007769, -0.00010165, 0.00009756, -0.00007688, -0.00002096, 0.00003977, 0.00003764, -0.00008065, 0.00006905, -0.00006674, 0.00007236, -0.00009602, 0.00011909, -0.00014756, 0.00014989, -0.00018111, 0.00017812, -0.00023307, 0.00023768, -0.00036150, 0.00045991, -0.00116059, 0.00133816, 0.00042649, -0.00081992, 0.00038952, -0.00032221, 0.00019591, -0.00019306, 0.00015115, -0.00015881, 0.00013574, -0.00013919, 0.00011032, -0.00009150, 0.00007012, -0.00005485, 0.00004708, -0.00007151, 0.00003663, 0.00004670, -0.00007442, 0.00000864, 0.00006645, -0.00009668, 0.00010213, -0.00007575, 0.00004945, -0.00004088, 0.00002444, 0.00001307, -0.00003003, -0.00001324, 0.00002678, -0.00001156, },
		{ 0.00000568, 0.00001721, -0.00003019, -0.00001772, 0.00005052, -0.00004244, 0.00003625, -0.00004938, 0.00004388, -0.00000150, -0.00003955, 0.00006089, -0.00008821, -0.00000092, 0.00007041, -0.00003136, -0.00001336, 0.00001253, -0.00000511, -0.00000411, -0.00000577, 0.00002031, -0.00004530, 0.00004244, -0.00006837, 0.00005103, -0.00009005, 0.00006063, -0.00012775, 0.00010978, -0.00094228, 0.00194005, -0.00065292, -0.00012917, 0.00002690, -0.00007019, 0.00000038, -0.00001991, 0.00000059, -0.00002509, 0.00002182, -0.00004643, 0.00003848, -0.00003419, 0.00002417, -0.00001232, -0.00000014, -0.00003833, 0.00003384, 0.00004441, -0.00012086, 0.00011751, -0.00004985, 0.00000146, 0.00003493, -0.00004071, 0.00003128, -0.00003311, 0.00003101, 0.00000188, -0.00004022, 0.00001248, 0.00001066, -0.00000610, },
		{ 0.00000052, 0.00001918, -0.00005099, 0.00002139, 0.00002194, -0.00003191, 0.00003490, -0.00006355, 0.00008855, -0.00007401, 0.00004136, -0.00001916, -0.00005249, 0.00000312, 0.00009115, -0.00009475, 0.00006396, -0.00005989, 0.00006912, -0.00009252, 0.00010220, -0.00010535, 0.00009350, -0.00011174, 0.00010086, -0.00014431, 0.00013628, -0.00021990, 0.00025439, -0.00055841, -0.00003010, 0.00164226, -0.00113825, 0.00037023, -0.00029490, 0.00016479, -0.00018006, 0.00014034, -0.00014323, 0.00010882, -0.00010021, 0.00005969, -0.00004782, 0.00003567, -0.00002892, 0.00002925, -0.00004514, -0.00000285, 0.00003644, 0.00000496, -0.00009476, 0.00015680, -0.00012869, 0.00009374, -0.00005166, 0.00002494, -0.00001758, 0.00000342, 0.00001439, -0.00000118, -0.00004588, 0.00003677, -0.00001212, 0.00000791, },
		{ -0.00000029, 0.00001055, -0.00005624, 0.00005571, -0.00002150, 0.00000778, -0.00000813, -0.00002029, 0.00006449, -0.00008072, 0.00007014, -0.00007390, -0.00001913, 0.00002429, 0.00006321, -0.00009708, 0.00008544, -0.00007944, 0.00008305, -0.00011014, 0.00013253, -0.00014975, 0.00014727, -0.00017631, 0.00017263, -0.00023120, 0.00023867, -0.00034807, 0.00043769, -0.00107339, 0.00110513, 0.00064949, -0.00089818, 0.00041077, -0.00033697, 0.00019663, -0.00019350, 0.00015537, -0.00016117, 0.00013471, -0.00013318, 0.00009436, -0.00007636, 0.00005831, -0.00004123, 0.00002978, -0.00005206, 0.00000734, 0.00005548, -0.00006900, 0.00001139, 0.00008845, -0.00010613, 0.00010444, -0.00008329, 0.00005806, -0.00004729, 0.00003129, -0.00000128, -0.00001014, -0.00003355, 0.00004199, -0.00002384, 0.00001682, },
		{ 0.00001045, -0.00001521, -0.00003216, 0.00005904, -0.00004532, 0.00003948, -0.00005536, 0.00004689, -0.00000732, -0.00002345, 0.00002753, -0.00006995, -0.00001453, 0.00006987, -0.00001439, -0.00002849, 0.00003181, -0.00002468, 0.00001550, -0.00003218, 0.00005316, -0.00007253, 0.00006676, -0.00009212, 0.00007596, -0.00012277, 0.00010881, -0.00017878, 0.00018040, -0.00098317, 0.00180626, -0.00043251, -0.00026488, 0.00010423, -0.00013181, 0.00004145, -0.00005203, 0.00003241, -0.00005431, 0.00004760, -0.00006759, 0.00004695, -0.00003805, 0.00002841, -0.00001383, -0.00000415, -0.00002906, 0.00000405, 0.00006349, -0.00012452, 0.00013317, -0.00003911, -0.00000340, 0.00003002, -0.00003915, 0.00003128, -0.00003039, 0.00002580, 0.00000122, -0.00003219, 0.00000039, 0.00002064, -0.00001369, 0.00001053, },
		{ 0.00002554, -0.00004824, 0.00001505, 0.00002492, -0.00002862, 0.00002992, -0.00006087, 0.00008035, -0.00006422, 0.00003940, -0.00004222, -0.00003198, -0.00001912, 0.00010469, -0.00009134, 0.00005889, -0.00004865, 0.00005508, -0.00007793, 0.00008001, -0.00007142, 0.00005912, -0.00007738, 0.00006539, -0.00010707, 0.00009104, -0.00015135, 0.00016796, -0.00041768, -0.00021004, 0.00167507, -0.00102005, 0.00026490, -0.00022418, 0.00010232, -0.00013329, 0.00010374, -0.00010742, 0.00007375, -0.00006577, 0.00002559, -0.00002433, 0.00001858, -0.00001242, 0.00001420, -0.00003397, -0.00001087, 0.00001688, 0.00002537, -0.00010049, 0.00018011, -0.00012725, 0.00009134, -0.00006045, 0.00003321, -0.00002455, 0.00001349, -0.00000172, 0.00001303, -0.00005362, 0.00003856, -0.00001322, 0.00000979, -0.00000508, },
		{ 0.00002637, -0.00006416, 0.00005676, -0.00002226, 0.00001385, -0.00001712, -0.00001421, 0.00005125, -0.00005960, 0.00004849, -0.00007762, -0.00000805, 0.00000309, 0.00008491, -0.00010639, 0.00009395, -0.00008313, 0.00008415, -0.00011284, 0.00013098, -0.00013561, 0.00012923, -0.00015690, 0.00015227, -0.00021151, 0.00021767, -0.00030736, 0.00037956, -0.00094173, 0.00085494, 0.00083186, -0.00091925, 0.00039059, -0.00032181, 0.00017484, -0.00017578, 0.00014339, -0.00014766, 0.00011869, -0.00011381, 0.00006903, -0.00005593, 0.00004254, -0.00002498, 0.00001210, -0.00003390, -0.00001563, 0.00005055, -0.00005335, 0.00000983, 0.00011311, -0.00011038, 0.00010060, -0.00008935, 0.00006573, -0.00005376, 0.00004008, -0.00001863, 0.00001019, -0.00005061, 0.00005349, -0.00003307, 0.00002483, -0.00001203, },
		{ 0.00000131, -0.00004401, 0.00006257, -0.00004476, 0.00004195, -0.00006106, 0.00004740, -0.00001340, -0.00000501, -0.00000295, -0.00005692, -0.00001741, 0.00005730, 0.00000747, -0.00004445, 0.00005018, -0.00004167, 0.00003187, -0.00005377, 0.00007695, -0.00008747, 0.00007793, -0.00010273, 0.00008772, -0.00013965, 0.00013632, -0.00020199, 0.00020859, -0.00095444, 0.00160351, -0.00019425, -0.00038378, 0.00016241, -0.00017679, 0.00006910, -0.00007321, 0.00005417, -0.00007247, 0.00006123, -0.00007574, 0.00004416, -0.00003387, 0.00002573, -0.00000916, -0.00001269, -0.00001597, -0.00002512, 0.00007148, -0.00011545, 0.00013908, -0.00001765, -0.00001157, 0.00002394, -0.00003891, 0.00003301, -0.00003016, 0.00002518, -0.00000660, -0.00001763, -0.00001487, 0.00003247, -0.00002326, 0.00001713, 0.00000123, },
		{ -0.00003831, 0.00000508, 0.00002764, -0.00002455, 0.00002615, -0.00006055, 0.00007180, -0.00005572, 0.00004235, -0.00006665, -0.00001225, -0.00003366, 0.00010457, -0.00007718, 0.00004571, -0.00003046, 0.00003522, -0.00005728, 0.00005175, -0.00003448, 0.00002518, -0.00004505, 0.00003199, -0.00007067, 0.00004691, -0.00008597, 0.00008844, -0.00029594, -0.00033303, 0.00162145, -0.00084300, 0.00013656, -0.00014331, 0.00003495, -0.00008325, 0.00006399, -0.00006890, 0.00003809, -0.00003342, -0.00000302, -0.00000756, 0.00000698, -0.00000152, 0.00000554, -0.00002915, -0.00001086, -0.00000944, 0.00004195, -0.00009733, 0.00019480, -0.00011258, 0.00007994, -0.00006611, 0.00003797, -0.00002833, 0.00002036, -0.00001306, 0.00001908, -0.00005184, 0.00003299, -0.00000840, 0.00000588, -0.00000517, 0.00002514, },
		{ -0.00006297, 0.00005050, -0.00001969, 0.00001867, -0.00002380, -0.00001230, 0.00003964, -0.00004060, 0.00003370, -0.00008696, 0.00000837, -0.00001758, 0.00009746, -0.00010539, 0.00009243, -0.00007653, 0.00007506, -0.00010388, 0.00011552, -0.00010825, 0.00010021, -0.00012779, 0.00012197, -0.00017912, 0.00018093, -0.00024826, 0.00029905, -0.00078760, 0.00061598, 0.00095266, -0.00087680, 0.00033029, -0.00027971, 0.00013401, -0.00014314, 0.00011800, -0.00012141, 0.00009182, -0.00008632, 0.00003999, -0.00003513, 0.00002700, -0.00001020, -0.00000196, -0.00002088, -0.00002913, 0.00003264, -0.00003247, 0.00000889, 0.00013672, -0.00010639, 0.00008830, -0.00009193, 0.00007072, -0.00005853, 0.00004804, -0.00003472, 0.00002603, -0.00006032, 0.00005827, -0.00003656, 0.00002718, -0.00001842, 0.00003521, },
		{ -0.00004971, 0.00005943, -0.00004034, 0.00004334, -0.00006565, 0.00004437, -0.00001778, 0.00001225, -0.00002691, -0.00005154, -0.00001000, 0.00003688, 0.00002889, -0.00005660, 0.00006325, -0.00005196, 0.00004098, -0.00006611, 0.00008764, -0.00008758, 0.00007476, -0.00009972, 0.00008571, -0.00013964, 0.00014198, -0.00019712, 0.00019726, -0.00086864, 0.00135784, 0.00003347, -0.00046707, 0.00019089, -0.00019834, 0.00007877, -0.00007996, 0.00006262, -0.00007691, 0.00006141, -0.00007133, 0.00003215, -0.00002416, 0.00001868, -0.00000140, -0.00002195, -0.00000396, -0.00004757, 0.00006463, -0.00009483, 0.00013799, 0.00001204, -0.00002081, 0.00001498, -0.00003857, 0.00003544, -0.00003184, 0.00002803, -0.00001889, -0.00000078, -0.00002902, 0.00004250, -0.00003134, 0.00002212, -0.00000688, 0.00001516, },
		{ -0.00000460, 0.00002752, -0.00001863, 0.00002320, -0.00006207, 0.00006262, -0.00004770, 0.00004739, -0.00008809, 0.00000187, -0.00003724, 0.00009132, -0.00005521, 0.00002789, -0.00000939, 0.00001403, -0.00003554, 0.00002358, -0.00000160, -0.00000165, -0.00002061, 0.00000609, -0.00004089, 0.00001089, -0.00003260, 0.00002653, -0.00020501, -0.00039333, 0.00149159, -0.00062822, 0.00000447, -0.00006640, -0.00002641, -0.00003850, 0.00002827, -0.00003452, 0.00000846, -0.00000892, -0.00002201, 0.00000035, 0.00000189, 0.00000330, 0.00000286, -0.00002915, -0.00000624, -0.00003587, 0.00004786, -0.00008254, 0.00020060, -0.00008491, 0.00005997, -0.00006875, 0.00003916, -0.00002839, 0.00002262, -0.00001798, 0.00001645, -0.00004183, 0.00002213, 0.00000036, -0.00000179, -0.00000178, 0.00002041, -0.00002479, },
		{ 0.00003947, -0.00001595, 0.00002315, -0.00002819, -0.00001468, 0.00003043, -0.00002476, 0.00002516, -0.00009901, 0.00002504, -0.00003207, 0.00009772, -0.00009343, 0.00008122, -0.00006094, 0.00005776, -0.00008588, 0.00009032, -0.00007362, 0.00006681, -0.00009555, 0.00008810, -0.00014080, 0.00013669, -0.00018170, 0.00021135, -0.00063247, 0.00041112, 0.00100122, -0.00077504, 0.00023896, -0.00021907, 0.00008162, -0.00010195, 0.00008475, -0.00008812, 0.00006039, -0.00005720, 0.00001316, -0.00001846, 0.00001510, 0.00000012, -0.00001001, -0.00001462, -0.00003322, 0.00000621, -0.00001373, 0.00001373, 0.00015653, -0.00009235, 0.00006628, -0.00008972, 0.00007166, -0.00005990, 0.00005240, -0.00004572, 0.00003373, -0.00006053, 0.00005524, -0.00003348, 0.00002327, -0.00001892, 0.00003681, -0.00005393, },
		{ 0.00004997, -0.00003308, 0.00004405, -0.00006873, 0.00003730, -0.00001915, 0.00002583, -0.00004242, -0.00005371, 0.00000452, 0.00001438, 0.00004466, -0.00006126, 0.00006799, -0.00005317, 0.00004089, -0.00006742, 0.00008452, -0.00007407, 0.00005969, -0.00008589, 0.00007259, -0.00012532, 0.00012904, -0.00016940, 0.00015600, -0.00074498, 0.00109819, 0.00022539, -0.00050239, 0.00018532, -0.00019486, 0.00007018, -0.00007242, 0.00005793, -0.00006846, 0.00005026, -0.00005786, 0.00001518, -0.00001281, 0.00001072, 0.00000582, -0.00002813, 0.00000291, -0.00005953, 0.00004325, -0.00006786, 0.00013514, 0.00004635, -0.00002849, 0.00000126, -0.00003643, 0.00003705, -0.00003389, 0.00003190, -0.00003158, 0.00001348, -0.00003788, 0.00004746, -0.00003500, 0.00002280, -0.00001108, 0.00002344, -0.00004770, },
		{ 0.00002304, -0.00001071, 0.00002104, -0.00006490, 0.00005242, -0.00003919, 0.00005176, -0.00010300, 0.00000756, -0.00002973, 0.00006849, -0.00003026, 0.00000998, 0.00000992, -0.00000384, -0.00001752, 0.00000103, 0.00002174, -0.00001713, -0.00000718, -0.00000963, -0.00002081, -0.00001308, 0.00000394, -0.00001258, -0.00014864, -0.00039591, 0.00130561, -0.00040184, -0.00011173, -0.00000611, -0.00007271, -0.00000581, 0.00000205, -0.00000932, -0.00001079, 0.00000475, -0.00003022, -0.00000001, 0.00000204, 0.00000357, 0.00000387, -0.00003075, -0.00000171, -0.00005587, 0.00003875, -0.00005724, 0.00020058, -0.00004687, 0.00003344, -0.00006964, 0.00003769, -0.00002528, 0.00002035, -0.00001693, 0.00000712, -0.00002716, 0.00000972, 0.00000979, -0.00001008, 0.00000205, 0.00001466, -0.00001177, -0.00001050, },
		{ -0.00001343, 0.00002810, -0.00003043, -0.00002096, 0.00002368, -0.00001214, 0.00002112, -0.00011015, 0.00003729, -0.00003684, 0.00008561, -0.00007262, 0.00006307, -0.00003980, 0.00003624, -0.00006343, 0.00006137, -0.00003882, 0.00003597, -0.00006647, 0.00005669, -0.00010315, 0.00009295, -0.00011799, 0.00012992, -0.00049334, 0.00025418, 0.00097867, -0.00062794, 0.00013161, -0.00015173, 0.00002726, -0.00006009, 0.00005042, -0.00005442, 0.00003100, -0.00003241, -0.00000690, -0.00000859, 0.00000843, 0.00000495, -0.00001191, -0.00001413, -0.00003097, -0.00002175, -0.00000486, 0.00002761, 0.00017230, -0.00006869, 0.00003501, -0.00008265, 0.00006811, -0.00005698, 0.00005137, -0.00004941, 0.00003202, -0.00005173, 0.00004569, -0.00002515, 0.00001457, -0.00001499, 0.00003304, -0.00004025, 0.00002739, },
		{ -0.00002507, 0.00004492, -0.00007025, 0.00002627, -0.00001700, 0.00003450, -0.00004939, -0.00006128, 0.00002149, -0.00000455, 0.00005115, -0.00005688, 0.00006369, -0.00004540, 0.00003227, -0.00005887, 0.00007028, -0.00005139, 0.00003791, -0.00006634, 0.00005322, -0.00010189, 0.00010381, -0.00012762, 0.00009814, -0.00060492, 0.00085107, 0.00036394, -0.00048637, 0.00014870, -0.00017019, 0.00004727, -0.00005423, 0.00004344, -0.00005096, 0.00003250, -0.00004058, -0.00000172, -0.00000371, 0.00000495, 0.00000965, -0.00002883, 0.00000275, -0.00006066, 0.00001194, -0.00004285, 0.00013661, 0.00008208, -0.00003274, -0.00001836, -0.00003122, 0.00003642, -0.00003456, 0.00003398, -0.00004072, 0.00002119, -0.00003884, 0.00004574, -0.00003288, 0.00001814, -0.00001036, 0.00002539, -0.00003883, 0.00003658, },
		{ -0.00000185, 0.00002014, -0.00006868, 0.00004101, -0.00002953, 0.00005346, -0.00010941, 0.00000443, -0.00001379, 0.00004153, -0.00000764, -0.00000385, 0.00002365, -0.00001498, -0.00000647, -0.00001256, 0.00003307, -0.00002046, -0.00000446, -0.00001566, -0.00001036, -0.00002467, 0.00002328, -0.00002943, -0.00012309, -0.00035416, 0.00108930, -0.00019000, -0.00019611, 0.00002926, -0.00009880, 0.00001149, -0.00001219, 0.00000459, -0.00001843, 0.00000785, -0.00002949, -0.00000577, 0.00000462, 0.00000199, 0.00000545, -0.00003046, -0.00000144, -0.00006504, 0.00001449, -0.00002705, 0.00020083, -0.00000277, 0.00000335, -0.00007067, 0.00003511, -0.00002037, 0.00001489, -0.00001206, -0.00000519, -0.00001249, -0.00000015, 0.00001657, -0.00001592, 0.00000353, 0.00001082, -0.00000238, -0.00001083, 0.00001446, },
		{ 0.00003358, -0.00003045, -0.00003055, 0.00001907, -0.00000233, 0.00001956, -0.00011720, 0.00004213, -0.00003126, 0.00006401, -0.00004726, 0.00004224, -0.00001769, 0.00001527, -0.00004167, 0.00003475, -0.00001014, 0.00001300, -0.00004494, 0.00003190, -0.00007095, 0.00005566, -0.00006463, 0.00006396, -0.00038023, 0.00014861, 0.00089695, -0.00045639, 0.00002590, -0.00009005, -0.00001968, -0.00002489, 0.00002117, -0.00002608, 0.00000877, -0.00001577, -0.00001818, -0.00000578, 0.00000648, 0.00000526, -0.00000958, -0.00001643, -0.00002724, -0.00004381, -0.00001135, 0.00004996, 0.00018721, -0.00003843, -0.00000314, -0.00007204, 0.00006087, -0.00005007, 0.00004473, -0.00004577, 0.00002232, -0.00003690, 0.00003286, -0.00001448, 0.00000394, -0.00000941, 0.00002701, -0.00002596, 0.00001776, -0.00001367, },
		{ 0.00004670, -0.00007028, 0.00001186, -0.00001159, 0.00003811, -0.00004908, -0.00007110, 0.00003631, -0.00001599, 0.00004744, -0.00004457, 0.00005206, -0.00003107, 0.00001811, -0.00004412, 0.00005004, -0.00002587, 0.00001568, -0.00004678, 0.00003306, -0.00007536, 0.00007372, -0.00008159, 0.00003734, -0.00046788, 0.00063627, 0.00044190, -0.00042523, 0.00009054, -0.00013252, 0.00001706, -0.00003133, 0.00002445, -0.00002990, 0.00001381, -0.00002484, -0.00001434, 0.00000058, 0.00000299, 0.00000898, -0.00002384, -0.00000354, -0.00005411, -0.00002150, -0.00002904, 0.00014708, 0.00011815, -0.00003369, -0.00004344, -0.00002284, 0.00003296, -0.00003271, 0.00003223, -0.00004371, 0.00002053, -0.00003172, 0.00003797, -0.00002573, 0.00000909, -0.00000573, 0.00002247, -0.00002607, 0.00002283, -0.00001867, },
		{ 0.00002127, -0.00007319, 0.00002856, -0.00001866, 0.00005155, -0.00010712, -0.00000574, 0.00000617, 0.00001615, 0.00000849, -0.00001110, 0.00002992, -0.00001814, -0.00000319, -0.00001674, 0.00003331, -0.00001427, -0.00000916, -0.00001507, -0.00000688, -0.00002656, 0.00002879, -0.00002916, -0.00011927, -0.00028610, 0.00086897, -0.00001397, -0.00023978, 0.00003762, -0.00010454, 0.00001406, -0.00001537, 0.00000824, -0.00001635, 0.00000342, -0.00002372, -0.00001296, 0.00000647, 0.00000124, 0.00000497, -0.00002588, -0.00000757, -0.00006266, -0.00002019, -0.00000135, 0.00020890, 0.00004292, -0.00002762, -0.00007337, 0.00003287, -0.00001522, 0.00000814, -0.00000624, -0.00001638, -0.00000206, -0.00000446, 0.00001858, -0.00001749, 0.00000116, 0.00001032, 0.00000216, -0.00000608, 0.00000393, 0.00000591, },
		{ -0.00002782, -0.00004281, 0.00001621, 0.00000516, 0.00001880, -0.00011826, 0.00003887, -0.00001749, 0.00003789, -0.00002248, 0.00002312, 0.00000111, -0.00000103, -0.00002478, 0.00001496, 0.00000857, 0.00000027, -0.00003236, 0.00001511, -0.00004630, 0.00002769, -0.00002516, 0.00001727, -0.00029557, 0.00008854, 0.00077560, -0.00028384, -0.00006180, -0.00004398, -0.00005239, -0.00000131, 0.00000098, -0.00000661, -0.00000383, -0.00000819, -0.00002146, -0.00000803, 0.00000707, 0.00000339, -0.00000603, -0.00001782, -0.00002681, -0.00005446, -0.00003439, 0.00007534, 0.00020790, -0.00000669, -0.00004445, -0.00006023, 0.00005161, -0.00004056, 0.00003378, -0.00003681, 0.00000813, -0.00002053, 0.00002081, -0.00000487, -0.00000540, -0.00000514, 0.00002185, -0.00001448, 0.00001271, -0.00001666, 0.00003861, },
		{ -0.00006865, -0.00000518, -0.00000357, 0.00003723, -0.00004340, -0.00008008, 0.00004571, -0.00001863, 0.00003547, -0.00002766, 0.00003663, -0.00001417, 0.00000271, -0.00002799, 0.00002970, -0.00000380, -0.00000150, -0.00003185, 0.00001657, -0.00005095, 0.00004529, -0.00003971, -0.00001538, -0.00034790, 0.00046431, 0.00046277, -0.00033303, 0.00002451, -0.00009211, -0.00001241, -0.00001019, 0.00000656, -0.00001071, -0.00000082, -0.00001444, -0.00002056, -0.00000039, 0.00000448, 0.00000474, -0.00001517, -0.00001268, -0.00004543, -0.00004813, -0.00003393, 0.00016748, 0.00015698, -0.00003409, -0.00007154, -0.00001270, 0.00002732, -0.00002841, 0.00002618, -0.00004015, 0.00001243, -0.00001899, 0.00002691, -0.00001610, -0.00000172, 0.00000023, 0.00001762, -0.00001339, 0.00001232, -0.00001556, 0.00004943, },
		{ -0.00007816, 0.00001548, -0.00000699, 0.00004610, -0.00009736, -0.00001992, 0.00002550, -0.00000327, 0.00001628, -0.00001166, 0.00002924, -0.00001449, -0.00000602, -0.00001383, 0.00002625, -0.00000356, -0.00001628, -0.00001227, -0.00000640, -0.00002312, 0.00002597, -0.00001934, -0.00012584, -0.00021013, 0.00066672, 0.00011335, -0.00024240, 0.00002325, -0.00009435, 0.00000604, -0.00001113, 0.00000508, -0.00000855, -0.00000404, -0.00001732, -0.00001799, 0.00000538, 0.00000276, 0.00000134, -0.00001677, -0.00001925, -0.00005201, -0.00005654, 0.00000873, 0.00023146, 0.00008753, -0.00005842, -0.00007798, 0.00003143, -0.00001080, 0.00000167, -0.00000191, -0.00002344, 0.00000180, -0.00000239, 0.00001580, -0.00001491, -0.00000460, 0.00001264, 0.00000285, 0.00000115, -0.00000516, 0.00001029, 0.00002504, },
		{ -0.00005713, 0.00001483, 0.00001056, 0.00001790, -0.00011301, 0.00002893, 0.00000048, 0.00001278, -0.00000274, 0.00000891, 0.00001382, -0.00001037, -0.00001481, 0.00000382, 0.00001672, -0.00000322, -0.00002714, 0.00000505, -0.00002861, 0.00000882, 0.00000077, -0.00001140, -0.00023545, 0.00006165, 0.00063725, -0.00013163, -0.00011991, -0.00001871, -0.00006808, 0.00000913, -0.00000917, 0.00000342, -0.00000727, -0.00000783, -0.00001965, -0.00001208, 0.00000744, 0.00000191, -0.00000411, -0.00001546, -0.00003256, -0.00005201, -0.00006972, 0.00009376, 0.00024315, 0.00002076, -0.00008503, -0.00004961, 0.00004224, -0.00003033, 0.00002076, -0.00002565, -0.00000625, -0.00000714, 0.00001295, 0.00000107, -0.00001112, -0.00000413, 0.00001943, -0.00000749, 0.00001216, -0.00002055, 0.00004131, -0.00002192, },
		{ -0.00002404, 0.00000622, 0.00003268, -0.00003424, -0.00008604, 0.00004848, -0.00001385, 0.00001934, -0.00001070, 0.00002154, 0.00000107, -0.00000971, -0.00001494, 0.00001426, 0.00001031, -0.00001047, -0.00002379, 0.00000603, -0.00003179, 0.00002274, -0.00000736, -0.00005347, -0.00025205, 0.00033629, 0.00043882, -0.00022837, -0.00003493, -0.00005851, -0.00003425, 0.00000402, -0.00000592, 0.00000274, -0.00000856, -0.00001069, -0.00002076, -0.00000492, 0.00000738, -0.00000064, -0.00000622, -0.00002025, -0.00004064, -0.00006057, -0.00006059, 0.00019347, 0.00020490, -0.00003921, -0.00009854, -0.00000350, 0.00002132, -0.00002297, 0.00001705, -0.00003190, 0.00000015, -0.00000491, 0.00001648, -0.00000741, -0.00001101, 0.00000449, 0.00001407, -0.00000437, 0.00000736, -0.00001588, 0.00005215, -0.00006484, },
		{ 0.00000226, 0.00000474, 0.00003784, -0.00008217, -0.00003489, 0.00004058, -0.00001470, 0.00001656, -0.00000773, 0.00002409, -0.00000706, -0.00001154, -0.00000786, 0.00001709, 0.00000597, -0.00002085, -0.00001134, -0.00000512, -0.00001882, 0.00002056, -0.00000761, -0.00013228, -0.00014125, 0.00049716, 0.00018889, -0.00021169, -0.00000459, -0.00007571, -0.00000646, -0.00000445, -0.00000039, 0.00000037, -0.00001019, -0.00001370, -0.00001895, 0.00000102, 0.00000616, -0.00000446, -0.00000528, -0.00003283, -0.00003951, -0.00008383, -0.00000673, 0.00027158, 0.00013191, -0.00009075, -0.00008265, 0.00002971, -0.00000681, -0.00000409, -0.00000005, -0.00002544, -0.00000062, 0.00000427, 0.00001037, -0.00001027, -0.00001153, 0.00001555, 0.00000238, 0.00000708, -0.00000935, 0.00000967, 0.00003130, -0.00008303, },
		{ 0.00001466, 0.00001387, 0.00001667, -0.00010244, 0.00001508, 0.00001813, -0.00000679, 0.00000946, 0.00000068, 0.00001990, -0.00001279, -0.00001131, 0.00000031, 0.00001679, -0.00000115, -0.00002556, -0.00000129, -0.00001551, -0.00000342, 0.00001638, -0.00002678, -0.00019208, 0.00005299, 0.00050294, -0.00001502, -0.00014395, -0.00001377, -0.00006832, 0.00000835, -0.00001110, 0.00000596, -0.00000439, -0.00001111, -0.00001650, -0.00001474, 0.00000553, 0.00000235, -0.00000517, -0.00000863, -0.00004422, -0.00003938, -0.00010804, 0.00009249, 0.00030151, 0.00003953, -0.00012236, -0.00004147, 0.00003389, -0.00002086, 0.00000779, -0.00001527, -0.00001726, 0.00000033, 0.00001079, 0.00000261, -0.00001273, -0.00000649, 0.00001975, -0.00000441, 0.00001387, -0.00002225, 0.00003953, -0.00001237, -0.00007270, },
		{ 0.00001695, 0.00002528, -0.00002305, -0.00008815, 0.00004546, -0.00000498, 0.00000400, 0.00000209, 0.00001014, 0.00001153, -0.00001631, -0.00000775, 0.00000639, 0.00001492, -0.00001120, -0.00002207, 0.00000128, -0.00001863, 0.00000747, 0.00001375, -0.00007536, -0.00018069, 0.00024565, 0.00038739, -0.00013010, -0.00007601, -0.00003814, -0.00004452, 0.00000879, -0.00001112, 0.00000905, -0.00000921, -0.00001232, -0.00001746, -0.00000992, 0.00000888, -0.00000430, -0.00000045, -0.00002239, -0.00004410, -0.00005537, -0.00010605, 0.00021518, 0.00027123, -0.00005553, -0.00011969, 0.00000161, 0.00001727, -0.00001849, 0.00000724, -0.00002226, -0.00001187, 0.00000584, 0.00001041, -0.00000263, -0.00001608, 0.00000473, 0.00001407, -0.00000099, 0.00000812, -0.00001795, 0.00005297, -0.00005816, -0.00004389, },
		{ 0.00001574, 0.00002778, -0.00006373, -0.00004809, 0.00004966, -0.00001873, 0.00001245, -0.00000293, 0.00001797, 0.00000039, -0.00001589, -0.00000309, 0.00001077, 0.00000979, -0.00001958, -0.00001467, -0.00000067, -0.00001687, 0.00001696, 0.00000030, -0.00013125, -0.00008873, 0.00036606, 0.00021869, -0.00016105, -0.00003435, -0.00005680, -0.00001736, 0.00000005, -0.00000424, 0.00000674, -0.00001229, -0.00001412, -0.00001624, -0.00000490, 0.00000928, -0.00000974, 0.00000467, -0.00004299, -0.00003277, -0.00009230, -0.00005335, 0.00032664, 0.00018145, -0.00012925, -0.00008349, 0.00002502, -0.00000157, -0.00001016, 0.00000022, -0.00002389, -0.00000644, 0.00001161, 0.00000594, -0.00000681, -0.00001641, 0.00001603, 0.00000398, 0.00000809, -0.00000634, 0.00000398, 0.00003884, -0.00008678, -0.00001071, },
		},
		{ { -0.00000338, -0.00000293, 0.00002005, -0.00000046, -0.00000029, -0.00000076, -0.00001055, 0.00000445, -0.00004299, 0.00000425, 0.00001434, -0.00001497, -0.00001588, -0.00001006, 0.00005244, -0.00000779, -0.00000970, -0.00000566, -0.00003340, -0.00000345, -0.00001368, -0.00001278, -0.00015355, 0.00005426, 0.00036947, 0.00005406, -0.00012865, -0.00002698, -0.00002928, -0.00001304, -0.00000101, -0.00001052, -0.00001601, -0.00001052, -0.00000101, -0.00001304, -0.00002928, -0.00002698, -0.00012865, 0.00005406, 0.00036947, 0.00005426, -0.00015355, -0.00001278, -0.00001368, -0.00000345, -0.00003340, -0.00000566, -0.00000970, -0.00000779, 0.00005244, -0.00001006, -0.00001588, -0.00001497, 0.00001434, 0.00000425, -0.00004299, 0.00000445, -0.00001055, -0.00000076, -0.00000029, -0.00000046, 0.00002005, -0.00000293, },
		{ -0.00000484, 0.00000526, 0.00001571, -0.00000543, 0.00000535, -0.00000873, 0.00000039, -0.00000919, -0.00003892, 0.00001513, 0.00000667, -0.00001578, -0.00001870, 0.00000482, 0.00005028, -0.00002256, 0.00000220, -0.00002021, -0.00002332, -0.00000643, -0.00000590, -0.00004483, -0.00014218, 0.00016603, 0.00030792, -0.00002055, -0.00010399, -0.00002080, -0.00002777, -0.00000730, -0.00000421, -0.00001098, -0.00001687, -0.00000762, 0.00000001, -0.00001992, -0.00002872, -0.00003914, -0.00014858, 0.00016644, 0.00037994, -0.00006975, -0.00012206, -0.00000138, -0.00001427, -0.00000828, -0.00003455, 0.00000322, -0.00001949, 0.00001049, 0.00004672, -0.00002224, -0.00001075, -0.00001478, 0.00002226, -0.00001241, -0.00003679, 0.00001076, -0.00002002, 0.00000979, -0.00001090, 0.00001338, 0.00001398, -0.00000508, },
		{ -0.00000683, 0.00001545, 0.00000351, -0.00000031, 0.00000190, -0.00000772, 0.00000418, -0.00002160, -0.00002976, 0.00002179, -0.00000169, -0.00001536, -0.00002024, 0.00002222, 0.00003818, -0.00002800, 0.00000864, -0.00003253, -0.00001135, -0.00001084, -0.00000042, -0.00008477, -0.00008782, 0.00023862, 0.00022442, -0.00006207, -0.00008133, -0.00001759, -0.00002461, -0.00000399, -0.00000741, -0.00001057, -0.00001842, -0.00000264, -0.00000211, -0.00002693, -0.00002554, -0.00006200, -0.00014957, 0.00030798, 0.00031614, -0.00016717, -0.00006802, -0.00000829, -0.00000700, -0.00002083, -0.00002537, 0.00000275, -0.00002208, 0.00002719, 0.00003519, -0.00003068, -0.00000404, -0.00001538, 0.00002941, -0.00003323, -0.00001903, 0.00000366, -0.00001882, 0.00001313, -0.00001690, 0.00002783, 0.00000207, -0.00000287, },
		{ -0.00000448, 0.00002018, -0.00000818, 0.00000863, -0.00000668, 0.00000081, -0.00000142, -0.00002778, -0.00001962, 0.00002522, -0.00001038, -0.00001348, -0.00002004, 0.00003970, 0.00001775, -0.00002235, 0.00000518, -0.00003650, -0.00000377, -0.00001080, -0.00000728, -0.00011280, -0.00000795, 0.00026405, 0.00014210, -0.00008022, -0.00006266, -0.00001561, -0.00002054, -0.00000362, -0.00000861, -0.00001167, -0.00001804, 0.00000188, -0.00000625, -0.00003425, -0.00001891, -0.00010114, -0.00010961, 0.00044968, 0.00017663, -0.00020468, -0.00002162, -0.00001864, -0.00000074, -0.00003308, -0.00001205, -0.00000379, -0.00001789, 0.00004074, 0.00001783, -0.00003187, 0.00000007, -0.00001330, 0.00003097, -0.00005118, 0.00000403, -0.00001472, -0.00000447, 0.00000361, -0.00001017, 0.00003322, -0.00000770, -0.00000124, },
		{ 0.00000483, 0.00001404, -0.00001097, 0.00001256, -0.00001199, 0.00000923, -0.00001177, -0.00002833, -0.00000927, 0.00002431, -0.00001664, -0.00001265, -0.00001504, 0.00005154, -0.00000439, -0.00001006, -0.00000614, -0.00003205, -0.00000130, -0.00000696, -0.00002818, -0.00011440, 0.00007369, 0.00024955, 0.00007415, -0.00008460, -0.00004745, -0.00001423, -0.00001643, -0.00000566, -0.00000705, -0.00001588, -0.00001315, 0.00000256, -0.00001004, -0.00004240, -0.00000967, -0.00015833, -0.00000545, 0.00054383, -0.00000794, -0.00017340, -0.00000502, -0.00001719, -0.00000597, -0.00003494, -0.00000393, -0.00000933, -0.00001085, 0.00005161, -0.00000561, -0.00002232, -0.00000361, -0.00000369, 0.00002068, -0.00005634, 0.00002021, -0.00003350, 0.00001566, -0.00001505, 0.00000952, 0.00002522, -0.00000994, -0.00000351, },
		{ 0.00001796, -0.00000162, -0.00000250, 0.00000690, -0.00000798, 0.00000990, -0.00001985, -0.00002679, 0.00000210, 0.00001773, -0.00001767, -0.00001568, -0.00000213, 0.00005176, -0.00001982, 0.00000081, -0.00001871, -0.00002396, 0.00000013, -0.00000618, -0.00005431, -0.00008792, 0.00013853, 0.00020907, 0.00002532, -0.00008271, -0.00003399, -0.00001387, -0.00001286, -0.00000869, -0.00000384, -0.00002251, -0.00000359, -0.00000214, -0.00001252, -0.00004963, -0.00000431, -0.00022326, 0.00017286, 0.00053943, -0.00018025, -0.00009688, -0.00001850, -0.00000105, -0.00002459, -0.00002306, -0.00000575, -0.00000944, -0.00000336, 0.00005824, -0.00003159, -0.00000323, -0.00001616, 0.00001451, -0.00000353, -0.00004257, 0.00001852, -0.00003912, 0.00002828, -0.00003110, 0.00003333, 0.00000787, -0.00000548, -0.00000802, },
		{ 0.00002699, -0.00001865, 0.00001168, -0.00000511, 0.00000416, 0.00000092, -0.00002190, -0.00002529, 0.00001371, 0.00000691, -0.00001403, -0.00002184, 0.00001773, 0.00003884, -0.00002369, 0.00000429, -0.00002674, -0.00001661, 0.00000366, -0.00001485, -0.00007250, -0.00004287, 0.00017798, 0.00015671, -0.00000532, -0.00007906, -0.00002086, -0.00001516, -0.00001022, -0.00001062, -0.00000196, -0.00002784, 0.00000683, -0.00000949, -0.00001663, -0.00004966, -0.00001770, -0.00026652, 0.00040478, 0.00040817, -0.00027930, -0.00002045, -0.00004074, 0.00001689, -0.00004587, -0.00000502, -0.00001332, -0.00000631, 0.00000682, 0.00005494, -0.00005089, 0.00001690, -0.00003115, 0.00003483, -0.00003496, -0.00001368, -0.00000271, -0.00002459, 0.00002291, -0.00003207, 0.00004855, -0.00000930, -0.00000006, -0.00000917, },
		{ 0.00002430, -0.00002694, 0.00002187, -0.00001477, 0.00001659, -0.00001267, -0.00001954, -0.00002256, 0.00002205, -0.00000323, -0.00001057, -0.00002568, 0.00003786, 0.00001814, -0.00001812, 0.00000047, -0.00002955, -0.00000972, 0.00000665, -0.00003189, -0.00007462, 0.00000812, 0.00019117, 0.00010440, -0.00002272, -0.00007438, -0.00000861, -0.00001768, -0.00000914, -0.00000925, -0.00000518, -0.00002632, 0.00001125, -0.00001317, -0.00002837, -0.00003387, -0.00006903, -0.00024199, 0.00063126, 0.00016807, -0.00027143, 0.00001407, -0.00004442, 0.00001746, -0.00005310, 0.00000538, -0.00001672, -0.00000641, 0.00002418, 0.00003504, -0.00005229, 0.00002546, -0.00003698, 0.00004501, -0.00005936, 0.00001612, -0.00003261, 0.00000455, -0.00000018, -0.00001334, 0.00004664, -0.00001775, 0.00000071, -0.00000215, },
		{ 0.00000823, -0.00002110, 0.00002107, -0.00001431, 0.00002045, -0.00002299, -0.00001737, -0.00001589, 0.00002331, -0.00000741, -0.00001308, -0.00002025, 0.00004904, -0.00000078, -0.00001037, -0.00000557, -0.00003069, 0.00000066, 0.00000150, -0.00004764, -0.00006264, 0.00005660, 0.00018036, 0.00006180, -0.00003424, -0.00006604, -0.00000015, -0.00001953, -0.00001036, -0.00000351, -0.00001551, -0.00001431, 0.00000421, -0.00000799, -0.00005119, 0.00000042, -0.00016848, -0.00010465, 0.00076720, -0.00011175, -0.00017133, -0.00000648, -0.00001703, -0.00000885, -0.00003660, -0.00000113, -0.00000902, -0.00001334, 0.00004893, -0.00000208, -0.00003052, 0.00001388, -0.00002459, 0.00003464, -0.00006272, 0.00002944, -0.00005299, 0.00003228, -0.00002887, 0.00001811, 0.00002860, -0.00001540, -0.00000389, 0.00001240, },
		{ -0.00001480, -0.00000417, 0.00000994, -0.00000282, 0.00001223, -0.00002517, -0.00001825, -0.00000490, 0.00001731, -0.00000459, -0.00002305, -0.00000289, 0.00004568, -0.00001048, -0.00000694, -0.00000913, -0.00003275, 0.00001625, -0.00001719, -0.00005076, -0.00004515, 0.00009944, 0.00014931, 0.00003500, -0.00004663, -0.00005075, 0.00000117, -0.00001868, -0.00001344, 0.00000463, -0.00003013, 0.00000559, -0.00001280, 0.00000363, -0.00007855, 0.00003964, -0.00029913, 0.00016005, 0.00073555, -0.00033397, -0.00003828, -0.00005813, 0.00002791, -0.00005145, -0.00000316, -0.00002071, 0.00000645, -0.00002198, 0.00007209, -0.00004572, 0.00000722, -0.00001480, 0.00000452, 0.00000374, -0.00004110, 0.00001677, -0.00004942, 0.00004162, -0.00004600, 0.00004736, 0.00000403, -0.00000723, -0.00000890, 0.00002742, },
		{ -0.00003369, 0.00001438, -0.00000372, 0.00001281, -0.00000382, -0.00002061, -0.00002062, 0.00000660, 0.00000909, -0.00000007, -0.00003475, 0.00002105, 0.00003042, -0.00001076, -0.00000836, -0.00001078, -0.00003264, 0.00003106, -0.00004509, -0.00003791, -0.00002875, 0.00013414, 0.00010516, 0.00002385, -0.00006195, -0.00002858, -0.00000517, -0.00001529, -0.00001553, 0.00000945, -0.00004064, 0.00002310, -0.00002909, 0.00000918, -0.00009152, 0.00005043, -0.00040664, 0.00050986, 0.00050987, -0.00041706, 0.00005416, -0.00009516, 0.00005831, -0.00008383, 0.00002610, -0.00003747, 0.00001695, -0.00001997, 0.00007780, -0.00007617, 0.00004213, -0.00004517, 0.00003654, -0.00003439, -0.00000541, -0.00001605, -0.00002113, 0.00002561, -0.00003993, 0.00006044, -0.00001522, -0.00000086, -0.00000759, 0.00003342, },
		{ -0.00003934, 0.00002490, -0.00001050, 0.00002257, -0.00001886, -0.00001498, -0.00002008, 0.00001290, 0.00000592, -0.00000241, -0.00003801, 0.00004034, 0.00001290, -0.00000829, -0.00000934, -0.00001596, -0.00002255, 0.00003241, -0.00006719, -0.00001817, -0.00001155, 0.00015433, 0.00006009, 0.00002060, -0.00007538, -0.00000519, -0.00001504, -0.00001249, -0.00001212, 0.00000391, -0.00003670, 0.00002475, -0.00002979, -0.00000786, -0.00006750, -0.00000830, -0.00040984, 0.00084211, 0.00014171, -0.00033815, 0.00005904, -0.00008085, 0.00004583, -0.00008090, 0.00002976, -0.00003475, 0.00000999, 0.00000301, 0.00005325, -0.00007511, 0.00005361, -0.00005784, 0.00005254, -0.00005957, 0.00002344, -0.00004973, 0.00001724, -0.00000756, -0.00001193, 0.00005242, -0.00002233, -0.00000061, 0.00000298, 0.00002474, },
		{ -0.00003004, 0.00002381, -0.00000616, 0.00002022, -0.00002582, -0.00001316, -0.00001400, 0.00001143, 0.00001132, -0.00001656, -0.00002542, 0.00004507, 0.00000300, -0.00001057, -0.00000439, -0.00002891, 0.00000272, 0.00000962, -0.00006637, -0.00000746, 0.00001484, 0.00015100, 0.00002861, 0.00001324, -0.00007789, 0.00001028, -0.00002169, -0.00001403, -0.00000066, -0.00001492, -0.00001365, 0.00000344, -0.00000725, -0.00005418, 0.00000110, -0.00015808, -0.00023623, 0.00102461, -0.00024806, -0.00014929, -0.00001639, -0.00001381, -0.00001187, -0.00003804, 0.00000229, -0.00000882, -0.00001500, 0.00004436, 0.00000018, -0.00003839, 0.00003246, -0.00004213, 0.00004088, -0.00005736, 0.00002711, -0.00006283, 0.00004398, -0.00003904, 0.00002458, 0.00002948, -0.00001873, -0.00000448, 0.00001894, 0.00000359, },
		{ -0.00001194, 0.00001535, 0.00000555, 0.00000737, -0.00002396, -0.00001513, -0.00000479, 0.00000606, 0.00002081, -0.00003840, 0.00000103, 0.00003396, 0.00000345, -0.00001880, 0.00000536, -0.00004570, 0.00003746, -0.00003498, -0.00003664, -0.00001598, 0.00005591, 0.00011972, 0.00002013, -0.00000698, -0.00006313, 0.00001127, -0.00002075, -0.00001999, 0.00001517, -0.00004060, 0.00002073, -0.00003284, 0.00002855, -0.00011388, 0.00008895, -0.00037104, 0.00013035, 0.00095392, -0.00051709, 0.00004609, -0.00011489, 0.00006814, -0.00008519, 0.00002239, -0.00003949, 0.00002565, -0.00004242, 0.00008515, -0.00006116, 0.00001815, -0.00001151, -0.00000455, 0.00000632, -0.00002954, 0.00000224, -0.00004549, 0.00004370, -0.00005076, 0.00005227, 0.00000399, -0.00001138, -0.00000626, 0.00003155, -0.00002098, },
		{ 0.00000560, 0.00000772, 0.00001578, -0.00000811, -0.00001844, -0.00001640, 0.00000126, 0.00000536, 0.00002348, -0.00005543, 0.00002946, 0.00001601, 0.00000754, -0.00002586, 0.00001074, -0.00005329, 0.00006349, -0.00008202, 0.00000898, -0.00003795, 0.00010497, 0.00006820, 0.00003180, -0.00003834, -0.00003389, -0.00000067, -0.00001442, -0.00002453, 0.00002473, -0.00005757, 0.00004649, -0.00006105, 0.00004984, -0.00014862, 0.00013636, -0.00055896, 0.00061624, 0.00061421, -0.00056833, 0.00014677, -0.00016438, 0.00011087, -0.00012907, 0.00006335, -0.00006600, 0.00004388, -0.00004883, 0.00009894, -0.00010045, 0.00006530, -0.00005377, 0.00003376, -0.00003134, 0.00000542, -0.00003598, -0.00000646, 0.00001716, -0.00003629, 0.00006006, -0.00001355, -0.00000671, -0.00000088, 0.00003287, -0.00003855, },
		{ 0.00001636, 0.00000670, 0.00001711, -0.00001804, -0.00001529, -0.00001268, -0.00000062, 0.00001598, 0.00000946, -0.00005425, 0.00004497, 0.00000447, 0.00000461, -0.00002180, 0.00000093, -0.00003735, 0.00005923, -0.00010334, 0.00004323, -0.00005251, 0.00014244, 0.00001661, 0.00004809, -0.00006796, -0.00000273, -0.00001575, -0.00001046, -0.00001891, 0.00001610, -0.00004931, 0.00004203, -0.00005559, 0.00002622, -0.00011717, 0.00007431, -0.00059783, 0.00106497, 0.00010248, -0.00040073, 0.00010857, -0.00012347, 0.00007983, -0.00011201, 0.00005722, -0.00005502, 0.00002854, -0.00001951, 0.00006870, -0.00009469, 0.00007703, -0.00006975, 0.00005019, -0.00004943, 0.00002341, -0.00006261, 0.00003171, -0.00001917, -0.00000411, 0.00004855, -0.00002011, -0.00000596, 0.00001115, 0.00002095, -0.00004347, },
		{ 0.00002057, 0.00001143, 0.00000880, -0.00001968, -0.00001622, -0.00000469, -0.00000832, 0.00003586, -0.00002155, -0.00003038, 0.00003982, 0.00000698, -0.00001060, -0.00000323, -0.00002621, 0.00000521, 0.00001518, -0.00007952, 0.00004132, -0.00003704, 0.00014652, -0.00001199, 0.00004954, -0.00007951, 0.00001507, -0.00002243, -0.00001578, 0.00000094, -0.00001415, -0.00001115, 0.00000018, -0.00000786, -0.00005115, -0.00000744, -0.00013050, -0.00038594, 0.00129298, -0.00040092, -0.00011338, -0.00003187, -0.00000903, -0.00001467, -0.00003878, 0.00000564, -0.00000879, -0.00001556, 0.00003797, 0.00000088, -0.00004373, 0.00004679, -0.00005068, 0.00003587, -0.00003792, 0.00001082, -0.00005889, 0.00004679, -0.00004362, 0.00002852, 0.00002754, -0.00001974, -0.00000524, 0.00002378, 0.00000086, -0.00003719, },
		{ 0.00002304, 0.00001554, -0.00000311, -0.00001688, -0.00001732, 0.00000118, -0.00001240, 0.00005287, -0.00005723, 0.00000692, 0.00001952, 0.00001916, -0.00003182, 0.00002077, -0.00005801, 0.00005851, -0.00005475, -0.00001559, -0.00000101, 0.00001537, 0.00010872, -0.00000586, 0.00002575, -0.00006491, 0.00001200, -0.00001636, -0.00002895, 0.00002771, -0.00005379, 0.00004106, -0.00006087, 0.00006008, -0.00015228, 0.00013871, -0.00043436, 0.00008947, 0.00117438, -0.00070925, 0.00014316, -0.00017956, 0.00011246, -0.00012041, 0.00004974, -0.00005920, 0.00004494, -0.00006120, 0.00009410, -0.00007420, 0.00002713, -0.00000758, -0.00000967, 0.00000229, -0.00000718, -0.00002525, -0.00002511, 0.00003108, -0.00004288, 0.00004677, 0.00000823, -0.00001821, 0.00000011, 0.00002963, -0.00001905, -0.00002563, },
		{ 0.00002785, 0.00001264, -0.00001076, -0.00001598, -0.00001308, -0.00000223, -0.00000209, 0.00005130, -0.00007839, 0.00003892, -0.00000020, 0.00002708, -0.00004399, 0.00003242, -0.00007171, 0.00009246, -0.00011591, 0.00005706, -0.00006078, 0.00008702, 0.00004299, 0.00002645, -0.00001621, -0.00003160, -0.00000567, -0.00000476, -0.00003846, 0.00004356, -0.00007788, 0.00007526, -0.00010246, 0.00010102, -0.00021449, 0.00022984, -0.00070905, 0.00071567, 0.00071277, -0.00071758, 0.00024458, -0.00023793, 0.00016559, -0.00017406, 0.00010079, -0.00009432, 0.00006993, -0.00007494, 0.00011356, -0.00011785, 0.00008058, -0.00005454, 0.00002685, -0.00002605, 0.00001798, -0.00006039, 0.00001874, -0.00000373, -0.00001972, 0.00004661, -0.00000406, -0.00001762, 0.00001064, 0.00002533, -0.00003269, -0.00001382, },
		{ 0.00003389, 0.00000193, -0.00001076, -0.00001978, -0.00000238, -0.00001585, 0.00002569, 0.00002271, -0.00007070, 0.00004892, -0.00000412, 0.00001760, -0.00003412, 0.00001747, -0.00004904, 0.00007996, -0.00013088, 0.00009645, -0.00009909, 0.00014292, -0.00001940, 0.00006002, -0.00005516, 0.00000082, -0.00002162, -0.00000138, -0.00003001, 0.00003080, -0.00006333, 0.00006148, -0.00008769, 0.00006948, -0.00017612, 0.00016690, -0.00078509, 0.00127804, 0.00005787, -0.00045539, 0.00015699, -0.00016649, 0.00011367, -0.00014100, 0.00008320, -0.00007423, 0.00004613, -0.00003995, 0.00007795, -0.00010602, 0.00008931, -0.00006867, 0.00003708, -0.00002888, 0.00001536, -0.00006859, 0.00004552, -0.00003445, 0.00001022, 0.00003505, -0.00001137, -0.00001501, 0.00002163, 0.00001359, -0.00003937, -0.00000220, },
		{ 0.00003510, -0.00001061, -0.00000673, -0.00002461, 0.00000866, -0.00003086, 0.00006104, -0.00002605, -0.00003499, 0.00003353, 0.00001115, -0.00001045, -0.00000272, -0.00002215, 0.00000839, 0.00001640, -0.00008330, 0.00007544, -0.00008483, 0.00015137, -0.00004766, 0.00006902, -0.00006962, 0.00001344, -0.00002150, -0.00001503, 0.00000099, -0.00001327, -0.00000720, -0.00000521, -0.00000963, -0.00004313, -0.00002276, -0.00009278, -0.00053468, 0.00154424, -0.00055007, -0.00007253, -0.00004870, -0.00000471, -0.00001684, -0.00003827, 0.00000820, -0.00000894, -0.00001486, 0.00003002, -0.00000009, -0.00004476, 0.00005247, -0.00004673, 0.00001899, -0.00000535, -0.00001818, -0.00004030, 0.00003879, -0.00004180, 0.00002980, 0.00002290, -0.00001858, -0.00000612, 0.00002655, 0.00000047, -0.00004220, 0.00001237, },
		{ 0.00002546, -0.00001649, -0.00000613, -0.00002308, 0.00001000, -0.00003256, 0.00008428, -0.00007413, 0.00001162, 0.00000559, 0.00003366, -0.00004299, 0.00003316, -0.00006518, 0.00007452, -0.00007189, 0.00000814, 0.00000083, -0.00001596, 0.00010474, -0.00003067, 0.00004395, -0.00005306, 0.00000158, -0.00000465, -0.00003984, 0.00004128, -0.00006848, 0.00006424, -0.00009345, 0.00009416, -0.00018912, 0.00018352, -0.00048536, 0.00004537, 0.00137371, -0.00088665, 0.00023690, -0.00024089, 0.00015253, -0.00015108, 0.00007470, -0.00007677, 0.00006111, -0.00007512, 0.00009624, -0.00008173, 0.00003232, -0.00000455, -0.00000862, -0.00000838, 0.00002402, -0.00006233, 0.00000975, 0.00000395, -0.00002286, 0.00003175, 0.00001597, -0.00002705, 0.00000927, 0.00002244, -0.00000977, -0.00004333, 0.00003192, },
		{ 0.00000490, -0.00001149, -0.00001365, -0.00001087, -0.00000436, -0.00001018, 0.00007705, -0.00009754, 0.00004517, -0.00001391, 0.00004370, -0.00005869, 0.00004923, -0.00008189, 0.00011074, -0.00013848, 0.00009630, -0.00008732, 0.00007630, 0.00002614, 0.00001564, -0.00000162, -0.00001933, -0.00002120, 0.00001399, -0.00005629, 0.00006415, -0.00009939, 0.00010577, -0.00014773, 0.00015540, -0.00027974, 0.00031808, -0.00084093, 0.00079955, 0.00079677, -0.00084806, 0.00033316, -0.00030406, 0.00021269, -0.00021095, 0.00013234, -0.00011791, 0.00009079, -0.00009407, 0.00011804, -0.00012405, 0.00008410, -0.00004709, 0.00001740, -0.00001911, 0.00003106, -0.00008555, 0.00005083, -0.00003505, 0.00000789, 0.00002216, 0.00001145, -0.00003205, 0.00002518, 0.00001266, -0.00001805, -0.00004025, 0.00005274, },
		{ -0.00001851, 0.00000036, -0.00002637, 0.00000792, -0.00002965, 0.00003378, 0.00003476, -0.00008353, 0.00004951, -0.00001003, 0.00002788, -0.00004377, 0.00003043, -0.00005352, 0.00008910, -0.00014280, 0.00013117, -0.00013751, 0.00014438, -0.00004337, 0.00005850, -0.00004018, 0.00000647, -0.00003267, 0.00001447, -0.00004433, 0.00004640, -0.00007711, 0.00008042, -0.00012141, 0.00011507, -0.00023473, 0.00025455, -0.00094845, 0.00145815, 0.00001659, -0.00049895, 0.00019852, -0.00020385, 0.00014136, -0.00016253, 0.00010333, -0.00008929, 0.00005992, -0.00005539, 0.00007873, -0.00010591, 0.00008672, -0.00005389, 0.00001534, -0.00000039, 0.00000087, -0.00006639, 0.00005601, -0.00005157, 0.00002933, 0.00001403, 0.00000206, -0.00002631, 0.00003281, 0.00000440, -0.00002892, -0.00002683, 0.00006535, },
		{ -0.00003271, 0.00000893, -0.00003481, 0.00002154, -0.00005036, 0.00008146, -0.00002809, -0.00003909, 0.00002654, 0.00001542, -0.00001010, -0.00000334, -0.00001649, 0.00001233, 0.00001240, -0.00007594, 0.00008986, -0.00011747, 0.00015254, -0.00006951, 0.00006981, -0.00004946, 0.00000590, -0.00001899, -0.00001158, -0.00000050, -0.00001233, -0.00000244, -0.00001203, -0.00001210, -0.00003212, -0.00004084, -0.00005412, -0.00066174, 0.00174958, -0.00067435, -0.00003646, -0.00006238, -0.00000280, -0.00001803, -0.00003608, 0.00000934, -0.00000930, -0.00001294, 0.00002096, -0.00000262, -0.00004063, 0.00004705, -0.00002990, -0.00000691, 0.00003596, -0.00005486, -0.00000985, 0.00002057, -0.00003413, 0.00002869, 0.00001605, -0.00001578, -0.00000706, 0.00002720, 0.00000249, -0.00004502, 0.00000123, 0.00005992, },
		{ -0.00002945, 0.00000516, -0.00002968, 0.00001793, -0.00004857, 0.00010785, -0.00008410, 0.00001243, -0.00000496, 0.00004433, -0.00004987, 0.00003954, -0.00006406, 0.00008201, -0.00008280, 0.00003043, -0.00000785, -0.00003427, 0.00009850, -0.00004542, 0.00004440, -0.00002840, -0.00001875, 0.00001320, -0.00005110, 0.00005413, -0.00008260, 0.00008677, -0.00012573, 0.00012563, -0.00021925, 0.00021817, -0.00052137, 0.00000676, 0.00152920, -0.00102596, 0.00031141, -0.00028797, 0.00018060, -0.00017185, 0.00009352, -0.00008958, 0.00007155, -0.00008195, 0.00009024, -0.00008202, 0.00003286, -0.00000396, -0.00000064, -0.00002361, 0.00005973, -0.00010238, 0.00005310, -0.00003372, 0.00000584, 0.00001022, 0.00002531, -0.00003642, 0.00001939, 0.00001203, 0.00000451, -0.00006191, 0.00004018, 0.00003374, },
		{ -0.00001068, -0.00001171, -0.00000976, -0.00000588, -0.00001581, 0.00009495, -0.00010770, 0.00004490, -0.00002172, 0.00005414, -0.00006691, 0.00005761, -0.00008000, 0.00011388, -0.00014460, 0.00011961, -0.00010916, 0.00006901, 0.00001441, 0.00000567, 0.00000243, 0.00000200, -0.00004520, 0.00003951, -0.00007537, 0.00008360, -0.00011888, 0.00013321, -0.00018976, 0.00020415, -0.00033392, 0.00038791, -0.00093954, 0.00086015, 0.00085820, -0.00094435, 0.00039910, -0.00035204, 0.00024363, -0.00023318, 0.00015303, -0.00013318, 0.00010323, -0.00010346, 0.00011084, -0.00011734, 0.00007482, -0.00003337, 0.00000802, -0.00001127, 0.00004334, -0.00010681, 0.00008381, -0.00007178, 0.00004175, -0.00000867, 0.00002952, -0.00004728, 0.00003984, -0.00000190, 0.00000148, -0.00006822, 0.00007682, -0.00000462, },
		{ 0.00001163, -0.00003237, 0.00001526, -0.00003916, 0.00003916, 0.00004338, -0.00008972, 0.00004493, -0.00001129, 0.00003341, -0.00004898, 0.00003765, -0.00004859, 0.00008404, -0.00013597, 0.00014184, -0.00015831, 0.00013864, -0.00005355, 0.00004718, -0.00002588, 0.00001399, -0.00004726, 0.00003503, -0.00005956, 0.00006056, -0.00008850, 0.00009564, -0.00015118, 0.00015503, -0.00028250, 0.00032219, -0.00106628, 0.00158438, -0.00001328, -0.00052884, 0.00022806, -0.00023022, 0.00015802, -0.00017252, 0.00011434, -0.00009797, 0.00006788, -0.00006402, 0.00007051, -0.00009383, 0.00006904, -0.00002837, -0.00001009, 0.00003151, -0.00001638, -0.00005643, 0.00006082, -0.00006779, 0.00004992, -0.00001075, 0.00001712, -0.00003767, 0.00004255, -0.00000420, -0.00001528, -0.00005209, 0.00009400, -0.00003837, },
		{ 0.00002254, -0.00004294, 0.00002963, -0.00006180, 0.00009269, -0.00002626, -0.00004246, 0.00001933, 0.00001958, -0.00000955, -0.00000526, -0.00000930, 0.00001684, 0.00000348, -0.00005813, 0.00008372, -0.00012890, 0.00014177, -0.00007351, 0.00005406, -0.00002259, -0.00000583, -0.00001523, -0.00000562, -0.00000321, -0.00001141, 0.00000231, -0.00001937, -0.00001468, -0.00002058, -0.00005706, -0.00002379, -0.00074839, 0.00188410, -0.00075549, -0.00001358, -0.00006924, -0.00000472, -0.00001801, -0.00003204, 0.00000869, -0.00000984, -0.00001004, 0.00001146, -0.00000637, -0.00003164, 0.00003066, -0.00000327, -0.00003598, 0.00007911, -0.00009137, 0.00002639, -0.00000488, -0.00002232, 0.00002569, 0.00000778, -0.00001209, -0.00000798, 0.00002596, 0.00000666, -0.00004578, -0.00001039, 0.00008051, -0.00005159, },
		{ 0.00001381, -0.00003442, 0.00002169, -0.00005577, 0.00011823, -0.00008382, 0.00000785, -0.00001025, 0.00004950, -0.00005109, 0.00003847, -0.00005352, 0.00007956, -0.00008553, 0.00004788, -0.00002322, -0.00003870, 0.00008448, -0.00004588, 0.00002854, 0.00000474, -0.00004511, 0.00003428, -0.00006043, 0.00006409, -0.00009364, 0.00010474, -0.00015243, 0.00014922, -0.00023797, 0.00023850, -0.00054083, -0.00001868, 0.00162225, -0.00110824, 0.00035400, -0.00031247, 0.00019117, -0.00017910, 0.00010365, -0.00009593, 0.00007479, -0.00008089, 0.00007645, -0.00007505, 0.00002903, -0.00000705, 0.00001331, -0.00003950, 0.00009420, -0.00013692, 0.00009597, -0.00007501, 0.00003751, -0.00001344, 0.00003369, -0.00004439, 0.00002816, 0.00000120, 0.00002050, -0.00007777, 0.00004581, 0.00003919, -0.00003886, },
		{ -0.00001033, -0.00000963, -0.00000738, -0.00001685, 0.00010109, -0.00010582, 0.00003663, -0.00002185, 0.00005665, -0.00006722, 0.00005615, -0.00006523, 0.00010108, -0.00013314, 0.00012393, -0.00012070, 0.00006214, 0.00000520, 0.00000053, -0.00000380, 0.00002857, -0.00007310, 0.00006754, -0.00009198, 0.00009847, -0.00013281, 0.00015260, -0.00022135, 0.00023874, -0.00036755, 0.00042830, -0.00099302, 0.00089164, 0.00089094, -0.00099472, 0.00043242, -0.00037424, 0.00025282, -0.00023681, 0.00015991, -0.00013816, 0.00010568, -0.00010233, 0.00009292, -0.00009903, 0.00005479, -0.00001729, 0.00000160, -0.00000312, 0.00005382, -0.00011963, 0.00011044, -0.00010692, 0.00007515, -0.00003985, 0.00004580, -0.00006001, 0.00005130, -0.00001454, 0.00002118, -0.00009174, 0.00009441, -0.00001273, -0.00000908, },
		{ -0.00003544, 0.00001748, -0.00004242, 0.00004107, 0.00004715, -0.00008770, 0.00003485, -0.00000727, 0.00003335, -0.00004903, 0.00003847, -0.00003442, 0.00006556, -0.00011165, 0.00012842, -0.00015756, 0.00012177, -0.00005109, 0.00003057, -0.00001444, 0.00002212, -0.00006250, 0.00005694, -0.00007272, 0.00007073, -0.00009534, 0.00010411, -0.00017169, 0.00018209, -0.00031031, 0.00035761, -0.00112224, 0.00164143, -0.00002586, -0.00054319, 0.00024195, -0.00024202, 0.00016075, -0.00016898, 0.00011467, -0.00009929, 0.00006915, -0.00006539, 0.00005469, -0.00007205, 0.00003967, 0.00000173, -0.00003265, 0.00006137, -0.00003133, -0.00004086, 0.00005829, -0.00008001, 0.00006796, -0.00003471, 0.00003032, -0.00004663, 0.00004867, -0.00000973, -0.00000200, -0.00007250, 0.00011316, -0.00005205, 0.00001949, },
		{ -0.00004602, 0.00003012, -0.00006223, 0.00009233, -0.00002013, -0.00004481, 0.00001249, 0.00002325, -0.00000883, -0.00000834, -0.00000097, 0.00002150, -0.00000887, -0.00003287, 0.00006035, -0.00011868, 0.00011644, -0.00006064, 0.00002723, 0.00000592, -0.00001925, -0.00001079, 0.00000231, -0.00000660, -0.00001056, 0.00000624, -0.00002631, -0.00001681, -0.00001087, -0.00006735, -0.00000906, -0.00078140, 0.00193092, -0.00078140, -0.00000906, -0.00006735, -0.00001087, -0.00001681, -0.00002631, 0.00000624, -0.00001056, -0.00000660, 0.00000231, -0.00001079, -0.00001925, 0.00000592, 0.00002723, -0.00006064, 0.00011644, -0.00011868, 0.00006035, -0.00003287, -0.00000887, 0.00002150, -0.00000097, -0.00000834, -0.00000883, 0.00002325, 0.00001249, -0.00004481, -0.00002013, 0.00009233, -0.00006223, 0.00003012, },
		{ -0.00003544, 0.00001949, -0.00005205, 0.00011316, -0.00007250, -0.00000200, -0.00000973, 0.00004867, -0.00004663, 0.00003032, -0.00003471, 0.00006796, -0.00008001, 0.00005829, -0.00004086, -0.00003133, 0.00006137, -0.00003265, 0.00000173, 0.00003967, -0.00007205, 0.00005469, -0.00006539, 0.00006915, -0.00009929, 0.00011467, -0.00016898, 0.00016075, -0.00024202, 0.00024195, -0.00054319, -0.00002586, 0.00164143, -0.00112224, 0.00035761, -0.00031031, 0.00018209, -0.00017169, 0.00010411, -0.00009534, 0.00007073, -0.00007272, 0.00005694, -0.00006250, 0.00002212, -0.00001444, 0.00003057, -0.00005109, 0.00012177, -0.00015756, 0.00012842, -0.00011165, 0.00006556, -0.00003442, 0.00003847, -0.00004903, 0.00003335, -0.00000727, 0.00003485, -0.00008770, 0.00004715, 0.00004107, -0.00004242, 0.00001748, },
		{ -0.00001033, -0.00000908, -0.00001273, 0.00009441, -0.00009174, 0.00002118, -0.00001454, 0.00005130, -0.00006001, 0.00004580, -0.00003985, 0.00007515, -0.00010692, 0.00011044, -0.00011963, 0.00005382, -0.00000312, 0.00000160, -0.00001729, 0.00005479, -0.00009903, 0.00009292, -0.00010233, 0.00010568, -0.00013816, 0.00015991, -0.00023681, 0.00025282, -0.00037424, 0.00043242, -0.00099472, 0.00089094, 0.00089164, -0.00099302, 0.00042830, -0.00036755, 0.00023874, -0.00022135, 0.00015260, -0.00013281, 0.00009847, -0.00009198, 0.00006754, -0.00007310, 0.00002857, -0.00000380, 0.00000053, 0.00000520, 0.00006214, -0.00012070, 0.00012393, -0.00013314, 0.00010108, -0.00006523, 0.00005615, -0.00006722, 0.00005665, -0.00002185, 0.00003663, -0.00010582, 0.00010109, -0.00001685, -0.00000738, -0.00000963, },
		{ 0.00001381, -0.00003886, 0.00003919, 0.00004581, -0.00007777, 0.00002050, 0.00000120, 0.00002816, -0.00004439, 0.00003369, -0.00001344, 0.00003751, -0.00007501, 0.00009597, -0.00013692, 0.00009420, -0.00003950, 0.00001331, -0.00000705, 0.00002903, -0.00007505, 0.00007645, -0.00008089, 0.00007479, -0.00009593, 0.00010365, -0.00017910, 0.00019117, -0.00031247, 0.00035400, -0.00110824, 0.00162225, -0.00001868, -0.00054083, 0.00023850, -0.00023797, 0.00014922, -0.00015243, 0.00010474, -0.00009364, 0.00006409, -0.00006043, 0.00003428, -0.00004511, 0.00000474, 0.00002854, -0.00004588, 0.00008448, -0.00003870, -0.00002322, 0.00004788, -0.00008553, 0.00007956, -0.00005352, 0.00003847, -0.00005109, 0.00004950, -0.00001025, 0.00000785, -0.00008382, 0.00011823, -0.00005577, 0.00002169, -0.00003442, },
		{ 0.00002254, -0.00005159, 0.00008051, -0.00001039, -0.00004578, 0.00000666, 0.00002596, -0.00000798, -0.00001209, 0.00000778, 0.00002569, -0.00002232, -0.00000488, 0.00002639, -0.00009137, 0.00007911, -0.00003598, -0.00000327, 0.00003066, -0.00003164, -0.00000637, 0.00001146, -0.00001004, -0.00000984, 0.00000869, -0.00003204, -0.00001801, -0.00000472, -0.00006924, -0.00001358, -0.00075549, 0.00188410, -0.00074839, -0.00002379, -0.00005706, -0.00002058, -0.00001468, -0.00001937, 0.00000231, -0.00001141, -0.00000321, -0.00000562, -0.00001523, -0.00000583, -0.00002259, 0.00005406, -0.00007351, 0.00014177, -0.00012890, 0.00008372, -0.00005813, 0.00000348, 0.00001684, -0.00000930, -0.00000526, -0.00000955, 0.00001958, 0.00001933, -0.00004246, -0.00002626, 0.00009269, -0.00006180, 0.00002963, -0.00004294, },
		{ 0.00001163, -0.00003837, 0.00009400, -0.00005209, -0.00001528, -0.00000420, 0.00004255, -0.00003767, 0.00001712, -0.00001075, 0.00004992, -0.00006779, 0.00006082, -0.00005643, -0.00001638, 0.00003151, -0.00001009, -0.00002837, 0.00006904, -0.00009383, 0.00007051, -0.00006402, 0.00006788, -0.00009797, 0.00011434, -0.00017252, 0.00015802, -0.00023022, 0.00022806, -0.00052884, -0.00001328, 0.00158438, -0.00106628, 0.00032219, -0.00028250, 0.00015503, -0.00015118, 0.00009564, -0.00008850, 0.00006056, -0.00005956, 0.00003503, -0.00004726, 0.00001399, -0.00002588, 0.00004718, -0.00005355, 0.00013864, -0.00015831, 0.00014184, -0.00013597, 0.00008404, -0.00004859, 0.00003765, -0.00004898, 0.00003341, -0.00001129, 0.00004493, -0.00008972, 0.00004338, 0.00003916, -0.00003916, 0.00001526, -0.00003237, },
		{ -0.00001068, -0.00000462, 0.00007682, -0.00006822, 0.00000148, -0.00000190, 0.00003984, -0.00004728, 0.00002952, -0.00000867, 0.00004175, -0.00007178, 0.00008381, -0.00010681, 0.00004334, -0.00001127, 0.00000802, -0.00003337, 0.00007482, -0.00011734, 0.00011084, -0.00010346, 0.00010323, -0.00013318, 0.00015303, -0.00023318, 0.00024363, -0.00035204, 0.00039910, -0.00094435, 0.00085820, 0.00086015, -0.00093954, 0.00038791, -0.00033392, 0.00020415, -0.00018976, 0.00013321, -0.00011888, 0.00008360, -0.00007537, 0.00003951, -0.00004520, 0.00000200, 0.00000243, 0.00000567, 0.00001441, 0.00006901, -0.00010916, 0.00011961, -0.00014460, 0.00011388, -0.00008000, 0.00005761, -0.00006691, 0.00005414, -0.00002172, 0.00004490, -0.00010770, 0.00009495, -0.00001581, -0.00000588, -0.00000976, -0.00001171, },
		{ -0.00002945, 0.00003374, 0.00004018, -0.00006191, 0.00000451, 0.00001203, 0.00001939, -0.00003642, 0.00002531, 0.00001022, 0.00000584, -0.00003372, 0.00005310, -0.00010238, 0.00005973, -0.00002361, -0.00000064, -0.00000396, 0.00003286, -0.00008202, 0.00009024, -0.00008195, 0.00007155, -0.00008958, 0.00009352, -0.00017185, 0.00018060, -0.00028797, 0.00031141, -0.00102596, 0.00152920, 0.00000676, -0.00052137, 0.00021817, -0.00021925, 0.00012563, -0.00012573, 0.00008677, -0.00008260, 0.00005413, -0.00005110, 0.00001320, -0.00001875, -0.00002840, 0.00004440, -0.00004542, 0.00009850, -0.00003427, -0.00000785, 0.00003043, -0.00008280, 0.00008201, -0.00006406, 0.00003954, -0.00004987, 0.00004433, -0.00000496, 0.00001243, -0.00008410, 0.00010785, -0.00004857, 0.00001793, -0.00002968, 0.00000516, },
		{ -0.00003271, 0.00005992, 0.00000123, -0.00004502, 0.00000249, 0.00002720, -0.00000706, -0.00001578, 0.00001605, 0.00002869, -0.00003413, 0.00002057, -0.00000985, -0.00005486, 0.00003596, -0.00000691, -0.00002990, 0.00004705, -0.00004063, -0.00000262, 0.00002096, -0.00001294, -0.00000930, 0.00000934, -0.00003608, -0.00001803, -0.00000280, -0.00006238, -0.00003646, -0.00067435, 0.00174958, -0.00066174, -0.00005412, -0.00004084, -0.00003212, -0.00001210, -0.00001203, -0.00000244, -0.00001233, -0.00000050, -0.00001158, -0.00001899, 0.00000590, -0.00004946, 0.00006981, -0.00006951, 0.00015254, -0.00011747, 0.00008986, -0.00007594, 0.00001240, 0.00001233, -0.00001649, -0.00000334, -0.00001010, 0.00001542, 0.00002654, -0.00003909, -0.00002809, 0.00008146, -0.00005036, 0.00002154, -0.00003481, 0.00000893, },
		{ -0.00001851, 0.00006535, -0.00002683, -0.00002892, 0.00000440, 0.00003281, -0.00002631, 0.00000206, 0.00001403, 0.00002933, -0.00005157, 0.00005601, -0.00006639, 0.00000087, -0.00000039, 0.00001534, -0.00005389, 0.00008672, -0.00010591, 0.00007873, -0.00005539, 0.00005992, -0.00008929, 0.00010333, -0.00016253, 0.00014136, -0.00020385, 0.00019852, -0.00049895, 0.00001659, 0.00145815, -0.00094845, 0.00025455, -0.00023473, 0.00011507, -0.00012141, 0.00008042, -0.00007711, 0.00004640, -0.00004433, 0.00001447, -0.00003267, 0.00000647, -0.00004018, 0.00005850, -0.00004337, 0.00014438, -0.00013751, 0.00013117, -0.00014280, 0.00008910, -0.00005352, 0.00003043, -0.00004377, 0.00002788, -0.00001003, 0.00004951, -0.00008353, 0.00003476, 0.00003378, -0.00002965, 0.00000792, -0.00002637, 0.00000036, },
		{ 0.00000490, 0.00005274, -0.00004025, -0.00001805, 0.00001266, 0.00002518, -0.00003205, 0.00001145, 0.00002216, 0.00000789, -0.00003505, 0.00005083, -0.00008555, 0.00003106, -0.00001911, 0.00001740, -0.00004709, 0.00008410, -0.00012405, 0.00011804, -0.00009407, 0.00009079, -0.00011791, 0.00013234, -0.00021095, 0.00021269, -0.00030406, 0.00033316, -0.00084806, 0.00079677, 0.00079955, -0.00084093, 0.00031808, -0.00027974, 0.00015540, -0.00014773, 0.00010577, -0.00009939, 0.00006415, -0.00005629, 0.00001399, -0.00002120, -0.00001933, -0.00000162, 0.00001564, 0.00002614, 0.00007630, -0.00008732, 0.00009630, -0.00013848, 0.00011074, -0.00008189, 0.00004923, -0.00005869, 0.00004370, -0.00001391, 0.00004517, -0.00009754, 0.00007705, -0.00001018, -0.00000436, -0.00001087, -0.00001365, -0.00001149, },
		{ 0.00002546, 0.00003192, -0.00004333, -0.00000977, 0.00002244, 0.00000927, -0.00002705, 0.00001597, 0.00003175, -0.00002286, 0.00000395, 0.00000975, -0.00006233, 0.00002402, -0.00000838, -0.00000862, -0.00000455, 0.00003232, -0.00008173, 0.00009624, -0.00007512, 0.00006111, -0.00007677, 0.00007470, -0.00015108, 0.00015253, -0.00024089, 0.00023690, -0.00088665, 0.00137371, 0.00004537, -0.00048536, 0.00018352, -0.00018912, 0.00009416, -0.00009345, 0.00006424, -0.00006848, 0.00004128, -0.00003984, -0.00000465, 0.00000158, -0.00005306, 0.00004395, -0.00003067, 0.00010474, -0.00001596, 0.00000083, 0.00000814, -0.00007189, 0.00007452, -0.00006518, 0.00003316, -0.00004299, 0.00003366, 0.00000559, 0.00001162, -0.00007413, 0.00008428, -0.00003256, 0.00001000, -0.00002308, -0.00000613, -0.00001649, },
		{ 0.00003510, 0.00001237, -0.00004220, 0.00000047, 0.00002655, -0.00000612, -0.00001858, 0.00002290, 0.00002980, -0.00004180, 0.00003879, -0.00004030, -0.00001818, -0.00000535, 0.00001899, -0.00004673, 0.00005247, -0.00004476, -0.00000009, 0.00003002, -0.00001486, -0.00000894, 0.00000820, -0.00003827, -0.00001684, -0.00000471, -0.00004870, -0.00007253, -0.00055007, 0.00154424, -0.00053468, -0.00009278, -0.00002276, -0.00004313, -0.00000963, -0.00000521, -0.00000720, -0.00001327, 0.00000099, -0.00001503, -0.00002150, 0.00001344, -0.00006962, 0.00006902, -0.00004766, 0.00015137, -0.00008483, 0.00007544, -0.00008330, 0.00001640, 0.00000839, -0.00002215, -0.00000272, -0.00001045, 0.00001115, 0.00003353, -0.00003499, -0.00002605, 0.00006104, -0.00003086, 0.00000866, -0.00002461, -0.00000673, -0.00001061, },
		{ 0.00003389, -0.00000220, -0.00003937, 0.00001359, 0.00002163, -0.00001501, -0.00001137, 0.00003505, 0.00001022, -0.00003445, 0.00004552, -0.00006859, 0.00001536, -0.00002888, 0.00003708, -0.00006867, 0.00008931, -0.00010602, 0.00007795, -0.00003995, 0.00004613, -0.00007423, 0.00008320, -0.00014100, 0.00011367, -0.00016649, 0.00015699, -0.00045539, 0.00005787, 0.00127804, -0.00078509, 0.00016690, -0.00017612, 0.00006948, -0.00008769, 0.00006148, -0.00006333, 0.00003080, -0.00003001, -0.00000138, -0.00002162, 0.00000082, -0.00005516, 0.00006002, -0.00001940, 0.00014292, -0.00009909, 0.00009645, -0.00013088, 0.00007996, -0.00004904, 0.00001747, -0.00003412, 0.00001760, -0.00000412, 0.00004892, -0.00007070, 0.00002271, 0.00002569, -0.00001585, -0.00000238, -0.00001978, -0.00001076, 0.00000193, },
		{ 0.00002785, -0.00001382, -0.00003269, 0.00002533, 0.00001064, -0.00001762, -0.00000406, 0.00004661, -0.00001972, -0.00000373, 0.00001874, -0.00006039, 0.00001798, -0.00002605, 0.00002685, -0.00005454, 0.00008058, -0.00011785, 0.00011356, -0.00007494, 0.00006993, -0.00009432, 0.00010079, -0.00017406, 0.00016559, -0.00023793, 0.00024458, -0.00071758, 0.00071277, 0.00071567, -0.00070905, 0.00022984, -0.00021449, 0.00010102, -0.00010246, 0.00007526, -0.00007788, 0.00004356, -0.00003846, -0.00000476, -0.00000567, -0.00003160, -0.00001621, 0.00002645, 0.00004299, 0.00008702, -0.00006078, 0.00005706, -0.00011591, 0.00009246, -0.00007171, 0.00003242, -0.00004399, 0.00002708, -0.00000020, 0.00003892, -0.00007839, 0.00005130, -0.00000209, -0.00000223, -0.00001308, -0.00001598, -0.00001076, 0.00001264, },
		{ 0.00002304, -0.00002563, -0.00001905, 0.00002963, 0.00000011, -0.00001821, 0.00000823, 0.00004677, -0.00004288, 0.00003108, -0.00002511, -0.00002525, -0.00000718, 0.00000229, -0.00000967, -0.00000758, 0.00002713, -0.00007420, 0.00009410, -0.00006120, 0.00004494, -0.00005920, 0.00004974, -0.00012041, 0.00011246, -0.00017956, 0.00014316, -0.00070925, 0.00117438, 0.00008947, -0.00043436, 0.00013871, -0.00015228, 0.00006008, -0.00006087, 0.00004106, -0.00005379, 0.00002771, -0.00002895, -0.00001636, 0.00001200, -0.00006491, 0.00002575, -0.00000586, 0.00010872, 0.00001537, -0.00000101, -0.00001559, -0.00005475, 0.00005851, -0.00005801, 0.00002077, -0.00003182, 0.00001916, 0.00001952, 0.00000692, -0.00005723, 0.00005287, -0.00001240, 0.00000118, -0.00001732, -0.00001688, -0.00000311, 0.00001554, },
		{ 0.00002057, -0.00003719, 0.00000086, 0.00002378, -0.00000524, -0.00001974, 0.00002754, 0.00002852, -0.00004362, 0.00004679, -0.00005889, 0.00001082, -0.00003792, 0.00003587, -0.00005068, 0.00004679, -0.00004373, 0.00000088, 0.00003797, -0.00001556, -0.00000879, 0.00000564, -0.00003878, -0.00001467, -0.00000903, -0.00003187, -0.00011338, -0.00040092, 0.00129298, -0.00038594, -0.00013050, -0.00000744, -0.00005115, -0.00000786, 0.00000018, -0.00001115, -0.00001415, 0.00000094, -0.00001578, -0.00002243, 0.00001507, -0.00007951, 0.00004954, -0.00001199, 0.00014652, -0.00003704, 0.00004132, -0.00007952, 0.00001518, 0.00000521, -0.00002621, -0.00000323, -0.00001060, 0.00000698, 0.00003982, -0.00003038, -0.00002155, 0.00003586, -0.00000832, -0.00000469, -0.00001622, -0.00001968, 0.00000880, 0.00001143, },
		{ 0.00001636, -0.00004347, 0.00002095, 0.00001115, -0.00000596, -0.00002011, 0.00004855, -0.00000411, -0.00001917, 0.00003171, -0.00006261, 0.00002341, -0.00004943, 0.00005019, -0.00006975, 0.00007703, -0.00009469, 0.00006870, -0.00001951, 0.00002854, -0.00005502, 0.00005722, -0.00011201, 0.00007983, -0.00012347, 0.00010857, -0.00040073, 0.00010248, 0.00106497, -0.00059783, 0.00007431, -0.00011717, 0.00002622, -0.00005559, 0.00004203, -0.00004931, 0.00001610, -0.00001891, -0.00001046, -0.00001575, -0.00000273, -0.00006796, 0.00004809, 0.00001661, 0.00014244, -0.00005251, 0.00004323, -0.00010334, 0.00005923, -0.00003735, 0.00000093, -0.00002180, 0.00000461, 0.00000447, 0.00004497, -0.00005425, 0.00000946, 0.00001598, -0.00000062, -0.00001268, -0.00001529, -0.00001804, 0.00001711, 0.00000670, },
		{ 0.00000560, -0.00003855, 0.00003287, -0.00000088, -0.00000671, -0.00001355, 0.00006006, -0.00003629, 0.00001716, -0.00000646, -0.00003598, 0.00000542, -0.00003134, 0.00003376, -0.00005377, 0.00006530, -0.00010045, 0.00009894, -0.00004883, 0.00004388, -0.00006600, 0.00006335, -0.00012907, 0.00011087, -0.00016438, 0.00014677, -0.00056833, 0.00061421, 0.00061624, -0.00055896, 0.00013636, -0.00014862, 0.00004984, -0.00006105, 0.00004649, -0.00005757, 0.00002473, -0.00002453, -0.00001442, -0.00000067, -0.00003389, -0.00003834, 0.00003180, 0.00006820, 0.00010497, -0.00003795, 0.00000898, -0.00008202, 0.00006349, -0.00005329, 0.00001074, -0.00002586, 0.00000754, 0.00001601, 0.00002946, -0.00005543, 0.00002348, 0.00000536, 0.00000126, -0.00001640, -0.00001844, -0.00000811, 0.00001578, 0.00000772, },
		{ -0.00001194, -0.00002098, 0.00003155, -0.00000626, -0.00001138, 0.00000399, 0.00005227, -0.00005076, 0.00004370, -0.00004549, 0.00000224, -0.00002954, 0.00000632, -0.00000455, -0.00001151, 0.00001815, -0.00006116, 0.00008515, -0.00004242, 0.00002565, -0.00003949, 0.00002239, -0.00008519, 0.00006814, -0.00011489, 0.00004609, -0.00051709, 0.00095392, 0.00013035, -0.00037104, 0.00008895, -0.00011388, 0.00002855, -0.00003284, 0.00002073, -0.00004060, 0.00001517, -0.00001999, -0.00002075, 0.00001127, -0.00006313, -0.00000698, 0.00002013, 0.00011972, 0.00005591, -0.00001598, -0.00003664, -0.00003498, 0.00003746, -0.00004570, 0.00000536, -0.00001880, 0.00000345, 0.00003396, 0.00000103, -0.00003840, 0.00002081, 0.00000606, -0.00000479, -0.00001513, -0.00002396, 0.00000737, 0.00000555, 0.00001535, },
		{ -0.00003004, 0.00000359, 0.00001894, -0.00000448, -0.00001873, 0.00002948, 0.00002458, -0.00003904, 0.00004398, -0.00006283, 0.00002711, -0.00005736, 0.00004088, -0.00004213, 0.00003246, -0.00003839, 0.00000018, 0.00004436, -0.00001500, -0.00000882, 0.00000229, -0.00003804, -0.00001187, -0.00001381, -0.00001639, -0.00014929, -0.00024806, 0.00102461, -0.00023623, -0.00015808, 0.00000110, -0.00005418, -0.00000725, 0.00000344, -0.00001365, -0.00001492, -0.00000066, -0.00001403, -0.00002169, 0.00001028, -0.00007789, 0.00001324, 0.00002861, 0.00015100, 0.00001484, -0.00000746, -0.00006637, 0.00000962, 0.00000272, -0.00002891, -0.00000439, -0.00001057, 0.00000300, 0.00004507, -0.00002542, -0.00001656, 0.00001132, 0.00001143, -0.00001400, -0.00001316, -0.00002582, 0.00002022, -0.00000616, 0.00002381, },
		{ -0.00003934, 0.00002474, 0.00000298, -0.00000061, -0.00002233, 0.00005242, -0.00001193, -0.00000756, 0.00001724, -0.00004973, 0.00002344, -0.00005957, 0.00005254, -0.00005784, 0.00005361, -0.00007511, 0.00005325, 0.00000301, 0.00000999, -0.00003475, 0.00002976, -0.00008090, 0.00004583, -0.00008085, 0.00005904, -0.00033815, 0.00014171, 0.00084211, -0.00040984, -0.00000830, -0.00006750, -0.00000786, -0.00002979, 0.00002475, -0.00003670, 0.00000391, -0.00001212, -0.00001249, -0.00001504, -0.00000519, -0.00007538, 0.00002060, 0.00006009, 0.00015433, -0.00001155, -0.00001817, -0.00006719, 0.00003241, -0.00002255, -0.00001596, -0.00000934, -0.00000829, 0.00001290, 0.00004034, -0.00003801, -0.00000241, 0.00000592, 0.00001290, -0.00002008, -0.00001498, -0.00001886, 0.00002257, -0.00001050, 0.00002490, },
		{ -0.00003369, 0.00003342, -0.00000759, -0.00000086, -0.00001522, 0.00006044, -0.00003993, 0.00002561, -0.00002113, -0.00001605, -0.00000541, -0.00003439, 0.00003654, -0.00004517, 0.00004213, -0.00007617, 0.00007780, -0.00001997, 0.00001695, -0.00003747, 0.00002610, -0.00008383, 0.00005831, -0.00009516, 0.00005416, -0.00041706, 0.00050987, 0.00050986, -0.00040664, 0.00005043, -0.00009152, 0.00000918, -0.00002909, 0.00002310, -0.00004064, 0.00000945, -0.00001553, -0.00001529, -0.00000517, -0.00002858, -0.00006195, 0.00002385, 0.00010516, 0.00013414, -0.00002875, -0.00003791, -0.00004509, 0.00003106, -0.00003264, -0.00001078, -0.00000836, -0.00001076, 0.00003042, 0.00002105, -0.00003475, -0.00000007, 0.00000909, 0.00000660, -0.00002062, -0.00002061, -0.00000382, 0.00001281, -0.00000372, 0.00001438, },
		{ -0.00001480, 0.00002742, -0.00000890, -0.00000723, 0.00000403, 0.00004736, -0.00004600, 0.00004162, -0.00004942, 0.00001677, -0.00004110, 0.00000374, 0.00000452, -0.00001480, 0.00000722, -0.00004572, 0.00007209, -0.00002198, 0.00000645, -0.00002071, -0.00000316, -0.00005145, 0.00002791, -0.00005813, -0.00003828, -0.00033397, 0.00073555, 0.00016005, -0.00029913, 0.00003964, -0.00007855, 0.00000363, -0.00001280, 0.00000559, -0.00003013, 0.00000463, -0.00001344, -0.00001868, 0.00000117, -0.00005075, -0.00004663, 0.00003500, 0.00014931, 0.00009944, -0.00004515, -0.00005076, -0.00001719, 0.00001625, -0.00003275, -0.00000913, -0.00000694, -0.00001048, 0.00004568, -0.00000289, -0.00002305, -0.00000459, 0.00001731, -0.00000490, -0.00001825, -0.00002517, 0.00001223, -0.00000282, 0.00000994, -0.00000417, },
		{ 0.00000823, 0.00001240, -0.00000389, -0.00001540, 0.00002860, 0.00001811, -0.00002887, 0.00003228, -0.00005299, 0.00002944, -0.00006272, 0.00003464, -0.00002459, 0.00001388, -0.00003052, -0.00000208, 0.00004893, -0.00001334, -0.00000902, -0.00000113, -0.00003660, -0.00000885, -0.00001703, -0.00000648, -0.00017133, -0.00011175, 0.00076720, -0.00010465, -0.00016848, 0.00000042, -0.00005119, -0.00000799, 0.00000421, -0.00001431, -0.00001551, -0.00000351, -0.00001036, -0.00001953, -0.00000015, -0.00006604, -0.00003424, 0.00006180, 0.00018036, 0.00005660, -0.00006264, -0.00004764, 0.00000150, 0.00000066, -0.00003069, -0.00000557, -0.00001037, -0.00000078, 0.00004904, -0.00002025, -0.00001308, -0.00000741, 0.00002331, -0.00001589, -0.00001737, -0.00002299, 0.00002045, -0.00001431, 0.00002107, -0.00002110, },
		{ 0.00002430, -0.00000215, 0.00000071, -0.00001775, 0.00004664, -0.00001334, -0.00000018, 0.00000455, -0.00003261, 0.00001612, -0.00005936, 0.00004501, -0.00003698, 0.00002546, -0.00005229, 0.00003504, 0.00002418, -0.00000641, -0.00001672, 0.00000538, -0.00005310, 0.00001746, -0.00004442, 0.00001407, -0.00027143, 0.00016807, 0.00063126, -0.00024199, -0.00006903, -0.00003387, -0.00002837, -0.00001317, 0.00001125, -0.00002632, -0.00000518, -0.00000925, -0.00000914, -0.00001768, -0.00000861, -0.00007438, -0.00002272, 0.00010440, 0.00019117, 0.00000812, -0.00007462, -0.00003189, 0.00000665, -0.00000972, -0.00002955, 0.00000047, -0.00001812, 0.00001814, 0.00003786, -0.00002568, -0.00001057, -0.00000323, 0.00002205, -0.00002256, -0.00001954, -0.00001267, 0.00001659, -0.00001477, 0.00002187, -0.00002694, },
		{ 0.00002699, -0.00000917, -0.00000006, -0.00000930, 0.00004855, -0.00003207, 0.00002291, -0.00002459, -0.00000271, -0.00001368, -0.00003496, 0.00003483, -0.00003115, 0.00001690, -0.00005089, 0.00005494, 0.00000682, -0.00000631, -0.00001332, -0.00000502, -0.00004587, 0.00001689, -0.00004074, -0.00002045, -0.00027930, 0.00040817, 0.00040478, -0.00026652, -0.00001770, -0.00004966, -0.00001663, -0.00000949, 0.00000683, -0.00002784, -0.00000196, -0.00001062, -0.00001022, -0.00001516, -0.00002086, -0.00007906, -0.00000532, 0.00015671, 0.00017798, -0.00004287, -0.00007250, -0.00001485, 0.00000366, -0.00001661, -0.00002674, 0.00000429, -0.00002369, 0.00003884, 0.00001773, -0.00002184, -0.00001403, 0.00000691, 0.00001371, -0.00002529, -0.00002190, 0.00000092, 0.00000416, -0.00000511, 0.00001168, -0.00001865, },
		{ 0.00001796, -0.00000802, -0.00000548, 0.00000787, 0.00003333, -0.00003110, 0.00002828, -0.00003912, 0.00001852, -0.00004257, -0.00000353, 0.00001451, -0.00001616, -0.00000323, -0.00003159, 0.00005824, -0.00000336, -0.00000944, -0.00000575, -0.00002306, -0.00002459, -0.00000105, -0.00001850, -0.00009688, -0.00018025, 0.00053943, 0.00017286, -0.00022326, -0.00000431, -0.00004963, -0.00001252, -0.00000214, -0.00000359, -0.00002251, -0.00000384, -0.00000869, -0.00001286, -0.00001387, -0.00003399, -0.00008271, 0.00002532, 0.00020907, 0.00013853, -0.00008792, -0.00005431, -0.00000618, 0.00000013, -0.00002396, -0.00001871, 0.00000081, -0.00001982, 0.00005176, -0.00000213, -0.00001568, -0.00001767, 0.00001773, 0.00000210, -0.00002679, -0.00001985, 0.00000990, -0.00000798, 0.00000690, -0.00000250, -0.00000162, },
		{ 0.00000483, -0.00000351, -0.00000994, 0.00002522, 0.00000952, -0.00001505, 0.00001566, -0.00003350, 0.00002021, -0.00005634, 0.00002068, -0.00000369, -0.00000361, -0.00002232, -0.00000561, 0.00005161, -0.00001085, -0.00000933, -0.00000393, -0.00003494, -0.00000597, -0.00001719, -0.00000502, -0.00017340, -0.00000794, 0.00054383, -0.00000545, -0.00015833, -0.00000967, -0.00004240, -0.00001004, 0.00000256, -0.00001315, -0.00001588, -0.00000705, -0.00000566, -0.00001643, -0.00001423, -0.00004745, -0.00008460, 0.00007415, 0.00024955, 0.00007369, -0.00011440, -0.00002818, -0.00000696, -0.00000130, -0.00003205, -0.00000614, -0.00001006, -0.00000439, 0.00005154, -0.00001504, -0.00001265, -0.00001664, 0.00002431, -0.00000927, -0.00002833, -0.00001177, 0.00000923, -0.00001199, 0.00001256, -0.00001097, 0.00001404, },
		{ -0.00000448, -0.00000124, -0.00000770, 0.00003322, -0.00001017, 0.00000361, -0.00000447, -0.00001472, 0.00000403, -0.00005118, 0.00003097, -0.00001330, 0.00000007, -0.00003187, 0.00001783, 0.00004074, -0.00001789, -0.00000379, -0.00001205, -0.00003308, -0.00000074, -0.00001864, -0.00002162, -0.00020468, 0.00017663, 0.00044968, -0.00010961, -0.00010114, -0.00001891, -0.00003425, -0.00000625, 0.00000188, -0.00001804, -0.00001167, -0.00000861, -0.00000362, -0.00002054, -0.00001561, -0.00006266, -0.00008022, 0.00014210, 0.00026405, -0.00000795, -0.00011280, -0.00000728, -0.00001080, -0.00000377, -0.00003650, 0.00000518, -0.00002235, 0.00001775, 0.00003970, -0.00002004, -0.00001348, -0.00001038, 0.00002522, -0.00001962, -0.00002778, -0.00000142, 0.00000081, -0.00000668, 0.00000863, -0.00000818, 0.00002018, },
		{ -0.00000683, -0.00000287, 0.00000207, 0.00002783, -0.00001690, 0.00001313, -0.00001882, 0.00000366, -0.00001903, -0.00003323, 0.00002941, -0.00001538, -0.00000404, -0.00003068, 0.00003519, 0.00002719, -0.00002208, 0.00000275, -0.00002537, -0.00002083, -0.00000700, -0.00000829, -0.00006802, -0.00016717, 0.00031614, 0.00030798, -0.00014957, -0.00006200, -0.00002554, -0.00002693, -0.00000211, -0.00000264, -0.00001842, -0.00001057, -0.00000741, -0.00000399, -0.00002461, -0.00001759, -0.00008133, -0.00006207, 0.00022442, 0.00023862, -0.00008782, -0.00008477, -0.00000042, -0.00001084, -0.00001135, -0.00003253, 0.00000864, -0.00002800, 0.00003818, 0.00002222, -0.00002024, -0.00001536, -0.00000169, 0.00002179, -0.00002976, -0.00002160, 0.00000418, -0.00000772, 0.00000190, -0.00000031, 0.00000351, 0.00001545, },
		{ -0.00000484, -0.00000508, 0.00001398, 0.00001338, -0.00001090, 0.00000979, -0.00002002, 0.00001076, -0.00003679, -0.00001241, 0.00002226, -0.00001478, -0.00001075, -0.00002224, 0.00004672, 0.00001049, -0.00001949, 0.00000322, -0.00003455, -0.00000828, -0.00001427, -0.00000138, -0.00012206, -0.00006975, 0.00037994, 0.00016644, -0.00014858, -0.00003914, -0.00002872, -0.00001992, 0.00000001, -0.00000762, -0.00001687, -0.00001098, -0.00000421, -0.00000730, -0.00002777, -0.00002080, -0.00010399, -0.00002055, 0.00030792, 0.00016603, -0.00014218, -0.00004483, -0.00000590, -0.00000643, -0.00002332, -0.00002021, 0.00000220, -0.00002256, 0.00005028, 0.00000482, -0.00001870, -0.00001578, 0.00000667, 0.00001513, -0.00003892, -0.00000919, 0.00000039, -0.00000873, 0.00000535, -0.00000543, 0.00001571, 0.00000526, },
		},
		{ { -0.00001314, -0.00001303, 0.00004534, -0.00000816, 0.00000265, -0.00000066, -0.00000454, -0.00000409, 0.00002067, -0.00000132, -0.00003253, 0.00001139, -0.00000249, 0.00002284, -0.00007779, 0.00001459, -0.00000003, -0.00001345, -0.00000956, -0.00001793, 0.00002585, -0.00002909, -0.00001142, -0.00002506, -0.00013670, 0.00005278, 0.00038107, 0.00005074, -0.00014617, -0.00002532, -0.00004261, -0.00000662, 0.00000494, -0.00000662, -0.00004261, -0.00002532, -0.00014617, 0.00005074, 0.00038107, 0.00005278, -0.00013670, -0.00002506, -0.00001142, -0.00002909, 0.00002585, -0.00001793, -0.00000956, -0.00001345, -0.00000003, 0.00001459, -0.00007779, 0.00002284, -0.00000249, 0.00001139, -0.00003253, -0.00000132, 0.00002067, -0.00000409, -0.00000454, -0.00000066, 0.00000265, -0.00000816, 0.00004534, -0.00001303, },
		{ -0.00001399, -0.00000440, 0.00004725, -0.00001790, 0.00000887, -0.00000560, -0.00000352, -0.00000296, 0.00002680, -0.00001639, -0.00002326, 0.00000806, 0.00000845, 0.00000439, -0.00007376, 0.00002831, -0.00001197, -0.00000583, -0.00001892, -0.00000690, 0.00002124, -0.00003381, -0.00000248, -0.00004844, -0.00012562, 0.00013466, 0.00033488, -0.00001497, -0.00012013, -0.00002396, -0.00003843, -0.00000168, 0.00000499, -0.00001377, -0.00004372, -0.00003155, -0.00016975, 0.00014094, 0.00039836, -0.00003977, -0.00012291, -0.00001151, -0.00001932, -0.00002210, 0.00002672, -0.00002648, -0.00000040, -0.00002176, 0.00001410, -0.00000659, -0.00007052, 0.00003413, -0.00001238, 0.00001453, -0.00004019, 0.00001469, 0.00001026, -0.00000183, -0.00000633, 0.00000410, -0.00000347, 0.00000342, 0.00003833, -0.00001765, },
		{ -0.00001612, 0.00000793, 0.00004306, -0.00002410, 0.00001407, -0.00001053, -0.00000265, 0.00000101, 0.00002779, -0.00002851, -0.00001423, 0.00000680, 0.00001630, -0.00001637, -0.00006096, 0.00003411, -0.00002033, -0.00000002, -0.00002763, 0.00000643, 0.00001145, -0.00003268, 0.00000139, -0.00007366, -0.00009236, 0.00019563, 0.00027254, -0.00005770, -0.00009611, -0.00002464, -0.00003261, 0.00000160, 0.00000473, -0.00002273, -0.00004117, -0.00004542, -0.00018435, 0.00025356, 0.00037396, -0.00012638, -0.00008973, -0.00001061, -0.00002216, -0.00001596, 0.00002555, -0.00003274, 0.00000787, -0.00002949, 0.00002815, -0.00003295, -0.00005164, 0.00003472, -0.00001688, 0.00001429, -0.00004340, 0.00002844, -0.00000216, 0.00000325, -0.00000906, 0.00000856, -0.00000865, 0.00001505, 0.00002781, -0.00001880, },
		{ -0.00001837, 0.00002264, 0.00003278, -0.00002545, 0.00001700, -0.00001495, -0.00000133, 0.00000658, 0.00002404, -0.00003696, -0.00000588, 0.00000802, 0.00001862, -0.00003502, -0.00004285, 0.00003266, -0.00002407, 0.00000252, -0.00003381, 0.00002020, -0.00000273, -0.00002445, -0.00000437, -0.00009149, -0.00004520, 0.00023220, 0.00020450, -0.00008083, -0.00007681, -0.00002493, -0.00002678, 0.00000415, 0.00000327, -0.00003252, -0.00003509, -0.00006958, -0.00018087, 0.00038041, 0.00029957, -0.00018779, -0.00004988, -0.00001854, -0.00002033, -0.00001112, 0.00002248, -0.00003594, 0.00001375, -0.00003454, 0.00003866, -0.00005992, -0.00002412, 0.00002411, -0.00001342, 0.00000812, -0.00003951, 0.00003641, -0.00001334, 0.00000947, -0.00001222, 0.00001248, -0.00001227, 0.00002519, 0.00001565, -0.00001764, },
		{ -0.00001907, 0.00003743, 0.00001765, -0.00002152, 0.00001660, -0.00001809, 0.00000073, 0.00001236, 0.00001664, -0.00004187, 0.00000238, 0.00001037, 0.00001521, -0.00004870, -0.00002271, 0.00002518, -0.00002252, 0.00000004, -0.00003461, 0.00003056, -0.00001747, -0.00001189, -0.00001979, -0.00009574, 0.00000622, 0.00024556, 0.00013869, -0.00008894, -0.00006287, -0.00002367, -0.00002187, 0.00000641, 0.00000020, -0.00004231, -0.00002588, -0.00010649, -0.00014748, 0.00050498, 0.00017663, -0.00020849, -0.00001885, -0.00002677, -0.00001819, -0.00000515, 0.00001570, -0.00003396, 0.00001477, -0.00003414, 0.00004141, -0.00008119, 0.00000588, 0.00000557, -0.00000252, -0.00000438, -0.00002745, 0.00003600, -0.00002001, 0.00001448, -0.00001463, 0.00001528, -0.00001381, 0.00003253, 0.00000372, -0.00001562, },
		{ -0.00001643, 0.00004938, 0.00000018, -0.00001308, 0.00001248, -0.00001918, 0.00000337, 0.00001731, 0.00000675, -0.00004346, 0.00001119, 0.00001171, 0.00000768, -0.00005621, -0.00000334, 0.00001348, -0.00001584, -0.00000852, -0.00002742, 0.00003309, -0.00002708, -0.00000122, -0.00003968, -0.00008585, 0.00005458, 0.00023882, 0.00008115, -0.00008761, -0.00005285, -0.00002149, -0.00001737, 0.00000764, -0.00000389, -0.00005184, -0.00001426, -0.00015748, -0.00007130, 0.00060250, 0.00002011, -0.00018334, -0.00000848, -0.00002666, -0.00002114, 0.00000536, 0.00000282, -0.00002446, 0.00000848, -0.00002586, 0.00003277, -0.00009035, 0.00003046, -0.00001453, 0.00001196, -0.00002072, -0.00000868, 0.00002667, -0.00001998, 0.00001608, -0.00001487, 0.00001616, -0.00001280, 0.00003612, -0.00000639, -0.00001412, },
		{ -0.00000914, 0.00005554, -0.00001629, -0.00000214, 0.00000520, -0.00001782, 0.00000620, 0.00002084, -0.00000478, -0.00004149, 0.00002034, 0.00001036, -0.00000172, -0.00005754, 0.00001283, 0.00000021, -0.00000583, -0.00002222, -0.00001177, 0.00002521, -0.00002683, 0.00000079, -0.00005589, -0.00006632, 0.00009635, 0.00021527, 0.00003642, -0.00008271, -0.00004387, -0.00002035, -0.00001170, 0.00000622, -0.00000770, -0.00006101, -0.00000225, -0.00022036, 0.00005752, 0.00064395, -0.00014196, -0.00012132, -0.00002142, -0.00001442, -0.00003176, 0.00002178, -0.00001679, -0.00000686, -0.00000583, -0.00000931, 0.00001165, -0.00008335, 0.00004260, -0.00002884, 0.00002415, -0.00003609, 0.00001279, 0.00001042, -0.00001296, 0.00001307, -0.00001184, 0.00001440, -0.00000896, 0.00003545, -0.00001358, -0.00001407, },
		{ 0.00000304, 0.00005380, -0.00002831, 0.00000848, -0.00000372, -0.00001424, 0.00000887, 0.00002266, -0.00001733, -0.00003526, 0.00002839, 0.00000603, -0.00001114, -0.00005313, 0.00002357, -0.00001112, 0.00000384, -0.00003734, 0.00000926, 0.00000830, -0.00001594, -0.00000945, -0.00006128, -0.00004376, 0.00013074, 0.00017839, 0.00000699, -0.00007897, -0.00003324, -0.00002214, -0.00000351, 0.00000101, -0.00001053, -0.00006858, 0.00000520, -0.00028621, 0.00023961, 0.00060389, -0.00027422, -0.00004431, -0.00004921, 0.00000599, -0.00004707, 0.00004113, -0.00003996, 0.00001598, -0.00002564, 0.00001258, -0.00001901, -0.00006057, 0.00003890, -0.00003182, 0.00002843, -0.00004501, 0.00003146, -0.00000856, -0.00000083, 0.00000578, -0.00000540, 0.00000981, -0.00000241, 0.00003060, -0.00001746, -0.00001554, },
		{ 0.00001885, 0.00004368, -0.00003340, 0.00001595, -0.00001218, -0.00000934, 0.00001136, 0.00002243, -0.00003014, -0.00002406, 0.00003303, -0.00000005, -0.00001973, -0.00004356, 0.00002707, -0.00001681, 0.00000833, -0.00004796, 0.00002918, -0.00001164, 0.00000088, -0.00002960, -0.00005360, -0.00002296, 0.00015722, 0.00013294, -0.00000783, -0.00007835, -0.00002030, -0.00002684, 0.00000661, -0.00000713, -0.00001371, -0.00007096, -0.00000096, -0.00033694, 0.00046039, 0.00047039, -0.00034416, 0.00001970, -0.00007536, 0.00002397, -0.00005879, 0.00005605, -0.00005963, 0.00003755, -0.00004507, 0.00003349, -0.00005213, -0.00002763, 0.00002127, -0.00002215, 0.00002190, -0.00004349, 0.00004211, -0.00002514, 0.00001297, -0.00000394, 0.00000344, 0.00000301, 0.00000611, 0.00002232, -0.00001845, -0.00001763, },
		{ 0.00003551, 0.00002678, -0.00003081, 0.00001835, -0.00001819, -0.00000435, 0.00001398, 0.00001956, -0.00004181, -0.00000812, 0.00003216, -0.00000582, -0.00002733, -0.00002976, 0.00002270, -0.00001430, 0.00000346, -0.00004798, 0.00004011, -0.00002586, 0.00001459, -0.00005138, -0.00003705, -0.00000433, 0.00017345, 0.00008583, -0.00001247, -0.00007916, -0.00000741, -0.00003168, 0.00001537, -0.00001468, -0.00002096, -0.00006248, -0.00003290, -0.00034584, 0.00068712, 0.00025341, -0.00033354, 0.00004672, -0.00008245, 0.00002698, -0.00005680, 0.00005745, -0.00006707, 0.00004959, -0.00005662, 0.00004552, -0.00007819, 0.00000588, -0.00000331, -0.00000347, 0.00000581, -0.00003080, 0.00004164, -0.00003487, 0.00002447, -0.00001327, 0.00001265, -0.00000452, 0.00001518, 0.00001203, -0.00001764, -0.00001855, },
		{ 0.00004927, 0.00000663, -0.00002196, 0.00001534, -0.00002061, -0.00000031, 0.00001713, 0.00001327, -0.00005021, 0.00001086, 0.00002496, -0.00000954, -0.00003381, -0.00001352, 0.00001184, -0.00000368, -0.00001209, -0.00003405, 0.00003627, -0.00002654, 0.00001555, -0.00006375, -0.00002050, 0.00001588, 0.00017519, 0.00004556, -0.00001409, -0.00007684, 0.00000073, -0.00003184, 0.00001771, -0.00001658, -0.00003686, -0.00003785, -0.00010178, -0.00028286, 0.00087209, -0.00001249, -0.00024659, 0.00002650, -0.00006041, 0.00000713, -0.00003442, 0.00003905, -0.00005572, 0.00004550, -0.00005422, 0.00004227, -0.00008856, 0.00002954, -0.00002514, 0.00001681, -0.00001463, -0.00001001, 0.00003023, -0.00003536, 0.00003053, -0.00001938, 0.00001979, -0.00001080, 0.00002297, 0.00000154, -0.00001645, -0.00001622, },
		{ 0.00005644, -0.00001221, -0.00001001, 0.00000834, -0.00001954, 0.00000239, 0.00002086, 0.00000304, -0.00005279, 0.00002957, 0.00001267, -0.00001071, -0.00003842, 0.00000219, -0.00000174, 0.00001140, -0.00003515, -0.00000791, 0.00001723, -0.00001095, -0.00000162, -0.00005804, -0.00001309, 0.00004259, 0.00015842, 0.00001983, -0.00001998, -0.00006650, -0.00000045, -0.00002297, 0.00000942, -0.00000913, -0.00006349, 0.00000342, -0.00021089, -0.00012451, 0.00096279, -0.00027353, -0.00011121, -0.00003248, -0.00001228, -0.00003364, 0.00000695, 0.00000145, -0.00002502, 0.00002375, -0.00003640, 0.00002201, -0.00007910, 0.00003606, -0.00003547, 0.00003040, -0.00003212, 0.00001290, 0.00001126, -0.00002698, 0.00002976, -0.00002039, 0.00002284, -0.00001395, 0.00002772, -0.00000745, -0.00001601, -0.00000904, },
		{ 0.00005459, -0.00002543, 0.00000108, 0.00000000, -0.00001632, 0.00000428, 0.00002451, -0.00001084, -0.00004742, 0.00004384, -0.00000167, -0.00001035, -0.00003967, 0.00001362, -0.00001255, 0.00002432, -0.00005825, 0.00002302, -0.00001071, 0.00001608, -0.00003435, -0.00003275, -0.00001919, 0.00007773, 0.00012265, 0.00001232, -0.00003418, -0.00004619, -0.00001232, -0.00000436, -0.00000960, 0.00000667, -0.00009690, 0.00005250, -0.00034844, 0.00013413, 0.00091721, -0.00047007, 0.00002841, -0.00010517, 0.00004550, -0.00008240, 0.00005643, -0.00004618, 0.00001789, -0.00001032, -0.00000795, -0.00001054, -0.00005247, 0.00002436, -0.00003013, 0.00003150, -0.00004015, 0.00003118, -0.00000986, -0.00001250, 0.00002288, -0.00001600, 0.00002091, -0.00001285, 0.00002824, -0.00001381, -0.00001663, 0.00000322, },
		{ 0.00004362, -0.00003045, 0.00000796, -0.00000680, -0.00001287, 0.00000657, 0.00002659, -0.00002680, -0.00003344, 0.00004993, -0.00001417, -0.00001039, -0.00003600, 0.00001742, -0.00001515, 0.00002764, -0.00007185, 0.00004770, -0.00003599, 0.00004289, -0.00007163, 0.00000445, -0.00003523, 0.00011713, 0.00007341, 0.00002035, -0.00005504, -0.00001920, -0.00003082, 0.00001903, -0.00003341, 0.00002328, -0.00012551, 0.00008945, -0.00048342, 0.00046975, 0.00071997, -0.00055445, 0.00012590, -0.00015906, 0.00008889, -0.00011925, 0.00009657, -0.00008835, 0.00005973, -0.00004576, 0.00002114, -0.00004509, -0.00001794, 0.00000048, -0.00001137, 0.00001937, -0.00003559, 0.00003972, -0.00002755, 0.00000394, 0.00001227, -0.00000754, 0.00001453, -0.00000759, 0.00002438, -0.00001733, -0.00001751, 0.00001898, },
		{ 0.00002602, -0.00002724, 0.00000907, -0.00001014, -0.00001087, 0.00001052, 0.00002519, -0.00004192, -0.00001233, 0.00004596, -0.00002160, -0.00001251, -0.00002689, 0.00001226, -0.00000650, 0.00001620, -0.00006805, 0.00005563, -0.00004615, 0.00005512, -0.00009749, 0.00003883, -0.00005034, 0.00015052, 0.00002215, 0.00003500, -0.00007528, 0.00000616, -0.00004696, 0.00003728, -0.00005097, 0.00002816, -0.00013245, 0.00008654, -0.00056825, 0.00082834, 0.00039294, -0.00050726, 0.00014890, -0.00016710, 0.00009618, -0.00012526, 0.00010996, -0.00010871, 0.00008566, -0.00006985, 0.00003929, -0.00006930, 0.00001172, -0.00002449, 0.00001277, -0.00000122, -0.00002005, 0.00003681, -0.00003790, 0.00001836, 0.00000106, 0.00000254, 0.00000554, 0.00000053, 0.00001704, -0.00001862, -0.00001694, 0.00003493, },
		{ 0.00000634, -0.00001845, 0.00000522, -0.00000980, -0.00001085, 0.00001672, 0.00001870, -0.00005253, 0.00001214, 0.00003274, -0.00002263, -0.00001702, -0.00001380, -0.00000001, 0.00001210, -0.00001006, -0.00004441, 0.00004155, -0.00003333, 0.00004194, -0.00009744, 0.00005436, -0.00005083, 0.00016499, -0.00001702, 0.00004423, -0.00008497, 0.00001959, -0.00005019, 0.00003925, -0.00005049, 0.00000877, -0.00010217, 0.00001734, -0.00054953, 0.00113315, -0.00000453, -0.00034591, 0.00009112, -0.00011929, 0.00005790, -0.00009107, 0.00008677, -0.00009681, 0.00008531, -0.00007343, 0.00003824, -0.00007368, 0.00002512, -0.00003871, 0.00003165, -0.00002180, 0.00000077, 0.00002454, -0.00003970, 0.00002807, -0.00000797, 0.00001148, -0.00000354, 0.00000937, 0.00000788, -0.00001878, -0.00001299, 0.00004691, },
		{ -0.00001018, -0.00000838, -0.00000084, -0.00000735, -0.00001191, 0.00002446, 0.00000662, -0.00005521, 0.00003479, 0.00001373, -0.00001840, -0.00002245, -0.00000025, -0.00001428, 0.00003461, -0.00004492, -0.00000604, 0.00000868, 0.00000127, 0.00000174, -0.00006528, 0.00004097, -0.00002674, 0.00015079, -0.00003268, 0.00003805, -0.00007647, 0.00001384, -0.00003363, 0.00001825, -0.00002557, -0.00004019, -0.00002956, -0.00013055, -0.00038528, 0.00130288, -0.00038788, -0.00012152, -0.00002449, -0.00002799, -0.00001756, -0.00002238, 0.00003012, -0.00005324, 0.00005742, -0.00005513, 0.00001714, -0.00005605, 0.00001717, -0.00003454, 0.00003647, -0.00003371, 0.00001939, 0.00000770, -0.00003434, 0.00003227, -0.00001316, 0.00001713, -0.00001030, 0.00001667, -0.00000128, -0.00001870, -0.00000444, 0.00005125, },
		{ -0.00001942, -0.00000142, -0.00000553, -0.00000547, -0.00001194, 0.00003168, -0.00000983, -0.00004801, 0.00005036, -0.00000597, -0.00001187, -0.00002604, 0.00000908, -0.00002356, 0.00005180, -0.00007728, 0.00003522, -0.00003133, 0.00004639, -0.00005546, -0.00000726, -0.00000006, 0.00002239, 0.00010670, -0.00002134, 0.00001347, -0.00004908, -0.00001074, 0.00000121, -0.00002301, 0.00001941, -0.00011086, 0.00007240, -0.00034357, -0.00006370, 0.00127564, -0.00066968, 0.00009605, -0.00015337, 0.00007552, -0.00010541, 0.00006049, -0.00004356, 0.00000947, 0.00001115, -0.00002263, -0.00001620, -0.00002323, -0.00000833, -0.00001216, 0.00002420, -0.00003172, 0.00002956, -0.00000831, -0.00002477, 0.00003179, -0.00001429, 0.00001851, -0.00001320, 0.00002067, -0.00000902, -0.00001853, 0.00000834, 0.00004622, },
		{ -0.00001982, -0.00000034, -0.00000609, -0.00000667, -0.00000861, 0.00003554, -0.00002767, -0.00003133, 0.00005519, -0.00002129, -0.00000660, -0.00002509, 0.00001028, -0.00002165, 0.00005453, -0.00009487, 0.00006468, -0.00006214, 0.00008395, -0.00011025, 0.00005869, -0.00005503, 0.00008548, 0.00004242, 0.00001022, -0.00002325, -0.00001101, -0.00004451, 0.00004287, -0.00007093, 0.00006788, -0.00018084, 0.00017064, -0.00057690, 0.00038373, 0.00103116, -0.00078535, 0.00024024, -0.00024532, 0.00015242, -0.00017309, 0.00012919, -0.00010970, 0.00007080, -0.00003705, 0.00001005, -0.00004783, 0.00001095, -0.00003951, 0.00002006, -0.00000084, -0.00001639, 0.00002875, -0.00001964, -0.00001400, 0.00002823, -0.00001225, 0.00001591, -0.00001189, 0.00002061, -0.00001467, -0.00001747, 0.00002309, 0.00003287, },
		{ -0.00001305, -0.00000511, -0.00000179, -0.00001199, -0.00000048, 0.00003336, -0.00004280, -0.00000816, 0.00004840, -0.00002891, -0.00000508, -0.00001851, 0.00000213, -0.00000594, 0.00003762, -0.00008906, 0.00007045, -0.00006897, 0.00009579, -0.00014059, 0.00010875, -0.00010220, 0.00014289, -0.00002376, 0.00004688, -0.00005849, 0.00002285, -0.00007123, 0.00007310, -0.00010468, 0.00009555, -0.00021873, 0.00021868, -0.00075986, 0.00088091, 0.00060314, -0.00071392, 0.00026974, -0.00026318, 0.00017145, -0.00019271, 0.00015802, -0.00014466, 0.00010985, -0.00006989, 0.00002833, -0.00006322, 0.00003092, -0.00006097, 0.00004855, -0.00002841, 0.00000607, 0.00001896, -0.00002524, -0.00000386, 0.00002313, -0.00000846, 0.00001062, -0.00000722, 0.00001681, -0.00001829, -0.00001412, 0.00003607, 0.00001498, },
		{ -0.00000337, -0.00001296, 0.00000554, -0.00002030, 0.00001203, 0.00002384, -0.00005113, 0.00001655, 0.00003223, -0.00002826, -0.00000764, -0.00000774, -0.00001275, 0.00002081, 0.00000264, -0.00005889, 0.00004865, -0.00004509, 0.00007137, -0.00013052, 0.00012226, -0.00012001, 0.00017362, -0.00007135, 0.00007104, -0.00007694, 0.00003695, -0.00007460, 0.00007412, -0.00010441, 0.00007983, -0.00019589, 0.00017161, -0.00081318, 0.00132096, 0.00007588, -0.00048598, 0.00018220, -0.00019692, 0.00012165, -0.00015253, 0.00013421, -0.00013492, 0.00011353, -0.00007637, 0.00002359, -0.00005371, 0.00002603, -0.00006005, 0.00006002, -0.00004644, 0.00002635, 0.00000562, -0.00002678, 0.00000536, 0.00001728, -0.00000421, 0.00000433, -0.00000076, 0.00001032, -0.00002025, -0.00000720, 0.00004327, -0.00000201, },
		{ 0.00000393, -0.00001944, 0.00001216, -0.00002845, 0.00002638, 0.00000778, -0.00004991, 0.00003722, 0.00001133, -0.00002149, -0.00001239, 0.00000339, -0.00002827, 0.00005061, -0.00004166, -0.00001235, 0.00000590, 0.00000432, 0.00001327, -0.00007745, 0.00009036, -0.00009609, 0.00016385, -0.00008568, 0.00007014, -0.00006884, 0.00002242, -0.00004616, 0.00003728, -0.00006094, 0.00001124, -0.00010060, 0.00000521, -0.00067399, 0.00159276, -0.00043541, -0.00017540, 0.00001465, -0.00006793, 0.00001767, -0.00006257, 0.00006370, -0.00008288, 0.00008189, -0.00005659, -0.00000250, -0.00002106, -0.00000437, -0.00003267, 0.00004722, -0.00004614, 0.00003600, -0.00000510, -0.00002721, 0.00001445, 0.00001073, -0.00000031, -0.00000139, 0.00000565, 0.00000251, -0.00002077, 0.00000351, 0.00004187, -0.00001289, },
		{ 0.00000462, -0.00002031, 0.00001394, -0.00003239, 0.00003854, -0.00001175, -0.00003870, 0.00004921, -0.00000868, -0.00001247, -0.00001611, 0.00001008, -0.00003685, 0.00007269, -0.00008185, 0.00003572, -0.00004224, 0.00006289, -0.00006212, 0.00000521, 0.00002061, -0.00003322, 0.00011333, -0.00006411, 0.00004251, -0.00003545, -0.00001755, 0.00000904, -0.00003076, 0.00001725, -0.00009848, 0.00005119, -0.00026757, -0.00032112, 0.00161338, -0.00081578, 0.00012317, -0.00016987, 0.00007873, -0.00010481, 0.00004812, -0.00003028, -0.00000635, 0.00002830, -0.00002190, -0.00003789, 0.00002219, -0.00004949, 0.00001412, 0.00001253, -0.00002592, 0.00003106, -0.00000912, -0.00002888, 0.00002418, 0.00000327, 0.00000296, -0.00000547, 0.00001048, -0.00000537, -0.00001955, 0.00001660, 0.00003160, -0.00001458, },
		{ -0.00000267, -0.00001344, 0.00000819, -0.00002881, 0.00004427, -0.00003018, -0.00001969, 0.00005019, -0.00002302, -0.00000513, -0.00001587, 0.00000858, -0.00003238, 0.00007736, -0.00010442, 0.00006883, -0.00007670, 0.00010840, -0.00012951, 0.00009162, -0.00006459, 0.00005058, 0.00003648, -0.00001759, -0.00000124, 0.00001007, -0.00006714, 0.00007212, -0.00010772, 0.00010372, -0.00021582, 0.00021415, -0.00059046, 0.00020882, 0.00135570, -0.00098360, 0.00032388, -0.00030475, 0.00019089, -0.00020210, 0.00014187, -0.00011551, 0.00006802, -0.00002603, 0.00000991, -0.00006481, 0.00005708, -0.00009058, 0.00006402, -0.00003226, 0.00000753, 0.00001390, -0.00000639, -0.00003196, 0.00003422, -0.00000492, 0.00000554, -0.00000745, 0.00001275, -0.00001240, -0.00001583, 0.00002907, 0.00001511, -0.00000736, },
		{ -0.00001581, 0.00000006, -0.00000498, -0.00001667, 0.00004074, -0.00004274, 0.00000282, 0.00004074, -0.00002911, -0.00000196, -0.00001056, -0.00000186, -0.00001298, 0.00005994, -0.00010089, 0.00007503, -0.00008221, 0.00012115, -0.00016362, 0.00015226, -0.00013562, 0.00012806, -0.00004241, 0.00003348, -0.00004233, 0.00004724, -0.00010325, 0.00011640, -0.00016256, 0.00016187, -0.00029463, 0.00032388, -0.00087243, 0.00082181, 0.00086167, -0.00091460, 0.00037623, -0.00034322, 0.00022878, -0.00023878, 0.00018635, -0.00016282, 0.00011503, -0.00006071, 0.00002241, -0.00006737, 0.00006581, -0.00010806, 0.00009740, -0.00007014, 0.00004160, -0.00000776, -0.00000086, -0.00003408, 0.00004269, -0.00001289, 0.00000733, -0.00000735, 0.00001209, -0.00001789, -0.00000887, 0.00003718, -0.00000255, 0.00000507, },
		{ -0.00002980, 0.00001598, -0.00002229, 0.00000203, 0.00002773, -0.00004598, 0.00002362, 0.00002409, -0.00002716, -0.00000318, -0.00000157, -0.00001819, 0.00001765, 0.00002306, -0.00007146, 0.00005191, -0.00005371, 0.00009201, -0.00014895, 0.00016481, -0.00016649, 0.00017279, -0.00009825, 0.00006756, -0.00006189, 0.00005685, -0.00010498, 0.00011806, -0.00016753, 0.00015885, -0.00029289, 0.00031752, -0.00101094, 0.00138566, 0.00023531, -0.00064872, 0.00027963, -0.00027422, 0.00017969, -0.00020116, 0.00016684, -0.00015710, 0.00012059, -0.00006465, 0.00000813, -0.00003911, 0.00003968, -0.00008968, 0.00009908, -0.00008547, 0.00006273, -0.00002408, 0.00000147, -0.00003139, 0.00004667, -0.00001892, 0.00000801, -0.00000549, 0.00000858, -0.00002122, 0.00000143, 0.00003771, -0.00001559, 0.00001687, },
		{ -0.00003869, 0.00002846, -0.00003824, 0.00002260, 0.00000802, -0.00003888, 0.00003805, 0.00000500, -0.00001973, -0.00000688, 0.00000754, -0.00003431, 0.00005098, -0.00002361, -0.00002538, 0.00000803, 0.00000086, 0.00002692, -0.00008644, 0.00012281, -0.00014476, 0.00016924, -0.00011474, 0.00007100, -0.00004969, 0.00003002, -0.00006354, 0.00006723, -0.00011085, 0.00008062, -0.00019144, 0.00015861, -0.00092384, 0.00176411, -0.00038374, -0.00027702, 0.00008100, -0.00012509, 0.00006244, -0.00010235, 0.00009163, -0.00010279, 0.00008695, -0.00004060, -0.00002747, 0.00001309, -0.00001616, -0.00003639, 0.00006486, -0.00007036, 0.00006200, -0.00002751, -0.00000444, -0.00002057, 0.00004345, -0.00002124, 0.00000718, -0.00000241, 0.00000270, -0.00002165, 0.00001388, 0.00002941, -0.00001953, 0.00002219, },
		{ -0.00003787, 0.00003217, -0.00004707, 0.00003913, -0.00001334, -0.00002321, 0.00004331, -0.00001171, -0.00001054, -0.00001013, 0.00001255, -0.00004316, 0.00007635, -0.00006647, 0.00002204, -0.00003988, 0.00006282, -0.00005423, 0.00000581, 0.00003879, -0.00007655, 0.00011885, -0.00009071, 0.00004395, -0.00000961, -0.00002641, 0.00001196, -0.00002538, -0.00000463, -0.00005846, -0.00000707, -0.00014301, -0.00057977, 0.00185578, -0.00085848, 0.00008710, -0.00014344, 0.00004972, -0.00008008, 0.00002279, -0.00001133, -0.00002176, 0.00003124, -0.00000374, -0.00006767, 0.00007045, -0.00008332, 0.00003677, 0.00000404, -0.00002833, 0.00003896, -0.00001619, -0.00001985, -0.00000087, 0.00003196, -0.00001879, 0.00000468, 0.00000114, -0.00000474, -0.00001843, 0.00002597, 0.00001382, -0.00001288, 0.00001747, },
		{ -0.00002609, 0.00002452, -0.00004488, 0.00004647, -0.00003053, -0.00000304, 0.00003908, -0.00002250, -0.00000306, -0.00001037, 0.00001043, -0.00003947, 0.00008482, -0.00009267, 0.00005498, -0.00007300, 0.00010932, -0.00012419, 0.00009828, -0.00005983, 0.00001544, 0.00003951, -0.00004022, 0.00000038, 0.00004093, -0.00009101, 0.00009584, -0.00012978, 0.00011634, -0.00021618, 0.00020643, -0.00052457, -0.00001613, 0.00162566, -0.00109284, 0.00034295, -0.00031536, 0.00018922, -0.00019720, 0.00013094, -0.00010527, 0.00005564, -0.00002209, 0.00002515, -0.00009128, 0.00010918, -0.00013639, 0.00010538, -0.00006330, 0.00002660, 0.00000213, 0.00000489, -0.00004108, 0.00002477, 0.00001373, -0.00001199, 0.00000090, 0.00000422, -0.00001257, -0.00001111, 0.00003433, -0.00000490, 0.00000207, 0.00000292, },
		{ -0.00000613, 0.00000681, -0.00003131, 0.00004217, -0.00003883, 0.00001655, 0.00002741, -0.00002597, 0.00000065, -0.00000663, 0.00000079, -0.00002201, 0.00007249, -0.00009484, 0.00006299, -0.00007773, 0.00012193, -0.00015817, 0.00016038, -0.00014035, 0.00010028, -0.00004138, 0.00001350, -0.00003799, 0.00007777, -0.00013545, 0.00015460, -0.00020656, 0.00020548, -0.00033491, 0.00037168, -0.00088184, 0.00066305, 0.00111838, -0.00105761, 0.00043202, -0.00038082, 0.00024804, -0.00024921, 0.00018645, -0.00015863, 0.00010260, -0.00005163, 0.00002915, -0.00008197, 0.00011006, -0.00015252, 0.00014442, -0.00011344, 0.00007494, -0.00003422, 0.00002577, -0.00006046, 0.00005017, -0.00000721, -0.00000278, -0.00000316, 0.00000569, -0.00001922, -0.00000004, 0.00003585, -0.00002105, 0.00001993, -0.00001739, },
		{ 0.00001600, -0.00001595, -0.00000983, 0.00002742, -0.00003612, 0.00003095, 0.00001190, -0.00002290, 0.00000043, 0.00000005, -0.00001366, 0.00000561, 0.00004213, -0.00007368, 0.00004495, -0.00005103, 0.00009396, -0.00014313, 0.00017142, -0.00017666, 0.00015023, -0.00009749, 0.00004788, -0.00005131, 0.00008030, -0.00013616, 0.00016014, -0.00022190, 0.00022186, -0.00036220, 0.00041270, -0.00109692, 0.00131061, 0.00044809, -0.00079583, 0.00035244, -0.00032661, 0.00021181, -0.00022134, 0.00017412, -0.00015654, 0.00010602, -0.00004784, 0.00000306, -0.00003608, 0.00006692, -0.00012076, 0.00013788, -0.00012784, 0.00009912, -0.00005567, 0.00003551, -0.00006924, 0.00006792, -0.00002546, 0.00000589, -0.00000616, 0.00000451, -0.00002291, 0.00001326, 0.00002889, -0.00002932, 0.00003397, -0.00003690, },
		{ 0.00003313, -0.00003673, 0.00001329, 0.00000679, -0.00002353, 0.00003731, -0.00000345, -0.00001564, -0.00000231, 0.00000709, -0.00002797, 0.00003600, 0.00000234, -0.00003770, 0.00000946, -0.00000191, 0.00003338, -0.00008296, 0.00012817, -0.00015848, 0.00015087, -0.00011346, 0.00005035, -0.00003081, 0.00004115, -0.00008514, 0.00010220, -0.00016210, 0.00014750, -0.00027253, 0.00028342, -0.00107449, 0.00177523, -0.00023260, -0.00040623, 0.00015432, -0.00018144, 0.00010040, -0.00012792, 0.00010373, -0.00010545, 0.00007079, -0.00001669, -0.00004363, 0.00003448, -0.00000956, -0.00004726, 0.00008532, -0.00010033, 0.00009029, -0.00005356, 0.00002696, -0.00006118, 0.00007204, -0.00003615, 0.00001119, -0.00000699, 0.00000011, -0.00002194, 0.00002610, 0.00001418, -0.00002661, 0.00003854, -0.00004888, },
		{ 0.00003957, -0.00004896, 0.00003113, -0.00001328, -0.00000498, 0.00003514, -0.00001536, -0.00000712, -0.00000541, 0.00001149, -0.00003673, 0.00006045, -0.00003550, 0.00000005, -0.00002853, 0.00005182, -0.00003970, 0.00000269, 0.00004591, -0.00009471, 0.00010581, -0.00009024, 0.00002346, 0.00001667, -0.00002924, 0.00000492, -0.00000543, -0.00004148, -0.00000240, -0.00008145, -0.00001063, -0.00077494, 0.00194401, -0.00077494, -0.00001063, -0.00008145, -0.00000240, -0.00004148, -0.00000543, 0.00000492, -0.00002924, 0.00001667, 0.00002346, -0.00009024, 0.00010581, -0.00009471, 0.00004591, 0.00000269, -0.00003970, 0.00005182, -0.00002853, 0.00000005, -0.00003550, 0.00006045, -0.00003673, 0.00001149, -0.00000541, -0.00000712, -0.00001536, 0.00003514, -0.00000498, -0.00001328, 0.00003113, -0.00004896, },
		{ 0.00003313, -0.00004888, 0.00003854, -0.00002661, 0.00001418, 0.00002610, -0.00002194, 0.00000011, -0.00000699, 0.00001119, -0.00003615, 0.00007204, -0.00006118, 0.00002696, -0.00005356, 0.00009029, -0.00010033, 0.00008532, -0.00004726, -0.00000956, 0.00003448, -0.00004363, -0.00001669, 0.00007079, -0.00010545, 0.00010373, -0.00012792, 0.00010040, -0.00018144, 0.00015432, -0.00040623, -0.00023260, 0.00177523, -0.00107449, 0.00028342, -0.00027253, 0.00014750, -0.00016210, 0.00010220, -0.00008514, 0.00004115, -0.00003081, 0.00005035, -0.00011346, 0.00015087, -0.00015848, 0.00012817, -0.00008296, 0.00003338, -0.00000191, 0.00000946, -0.00003770, 0.00000234, 0.00003600, -0.00002797, 0.00000709, -0.00000231, -0.00001564, -0.00000345, 0.00003731, -0.00002353, 0.00000679, 0.00001329, -0.00003673, },
		{ 0.00001600, -0.00003690, 0.00003397, -0.00002932, 0.00002889, 0.00001326, -0.00002291, 0.00000451, -0.00000616, 0.00000589, -0.00002546, 0.00006792, -0.00006924, 0.00003551, -0.00005567, 0.00009912, -0.00012784, 0.00013788, -0.00012076, 0.00006692, -0.00003608, 0.00000306, -0.00004784, 0.00010602, -0.00015654, 0.00017412, -0.00022134, 0.00021181, -0.00032661, 0.00035244, -0.00079583, 0.00044809, 0.00131061, -0.00109692, 0.00041270, -0.00036220, 0.00022186, -0.00022190, 0.00016014, -0.00013616, 0.00008030, -0.00005131, 0.00004788, -0.00009749, 0.00015023, -0.00017666, 0.00017142, -0.00014313, 0.00009396, -0.00005103, 0.00004495, -0.00007368, 0.00004213, 0.00000561, -0.00001366, 0.00000005, 0.00000043, -0.00002290, 0.00001190, 0.00003095, -0.00003612, 0.00002742, -0.00000983, -0.00001595, },
		{ -0.00000613, -0.00001739, 0.00001993, -0.00002105, 0.00003585, -0.00000004, -0.00001922, 0.00000569, -0.00000316, -0.00000278, -0.00000721, 0.00005017, -0.00006046, 0.00002577, -0.00003422, 0.00007494, -0.00011344, 0.00014442, -0.00015252, 0.00011006, -0.00008197, 0.00002915, -0.00005163, 0.00010260, -0.00015863, 0.00018645, -0.00024921, 0.00024804, -0.00038082, 0.00043202, -0.00105761, 0.00111838, 0.00066305, -0.00088184, 0.00037168, -0.00033491, 0.00020548, -0.00020656, 0.00015460, -0.00013545, 0.00007777, -0.00003799, 0.00001350, -0.00004138, 0.00010028, -0.00014035, 0.00016038, -0.00015817, 0.00012193, -0.00007773, 0.00006299, -0.00009484, 0.00007249, -0.00002201, 0.00000079, -0.00000663, 0.00000065, -0.00002597, 0.00002741, 0.00001655, -0.00003883, 0.00004217, -0.00003131, 0.00000681, },
		{ -0.00002609, 0.00000292, 0.00000207, -0.00000490, 0.00003433, -0.00001111, -0.00001257, 0.00000422, 0.00000090, -0.00001199, 0.00001373, 0.00002477, -0.00004108, 0.00000489, 0.00000213, 0.00002660, -0.00006330, 0.00010538, -0.00013639, 0.00010918, -0.00009128, 0.00002515, -0.00002209, 0.00005564, -0.00010527, 0.00013094, -0.00019720, 0.00018922, -0.00031536, 0.00034295, -0.00109284, 0.00162566, -0.00001613, -0.00052457, 0.00020643, -0.00021618, 0.00011634, -0.00012978, 0.00009584, -0.00009101, 0.00004093, 0.00000038, -0.00004022, 0.00003951, 0.00001544, -0.00005983, 0.00009828, -0.00012419, 0.00010932, -0.00007300, 0.00005498, -0.00009267, 0.00008482, -0.00003947, 0.00001043, -0.00001037, -0.00000306, -0.00002250, 0.00003908, -0.00000304, -0.00003053, 0.00004647, -0.00004488, 0.00002452, },
		{ -0.00003787, 0.00001747, -0.00001288, 0.00001382, 0.00002597, -0.00001843, -0.00000474, 0.00000114, 0.00000468, -0.00001879, 0.00003196, -0.00000087, -0.00001985, -0.00001619, 0.00003896, -0.00002833, 0.00000404, 0.00003677, -0.00008332, 0.00007045, -0.00006767, -0.00000374, 0.00003124, -0.00002176, -0.00001133, 0.00002279, -0.00008008, 0.00004972, -0.00014344, 0.00008710, -0.00085848, 0.00185578, -0.00057977, -0.00014301, -0.00000707, -0.00005846, -0.00000463, -0.00002538, 0.00001196, -0.00002641, -0.00000961, 0.00004395, -0.00009071, 0.00011885, -0.00007655, 0.00003879, 0.00000581, -0.00005423, 0.00006282, -0.00003988, 0.00002204, -0.00006647, 0.00007635, -0.00004316, 0.00001255, -0.00001013, -0.00001054, -0.00001171, 0.00004331, -0.00002321, -0.00001334, 0.00003913, -0.00004707, 0.00003217, },
		{ -0.00003869, 0.00002219, -0.00001953, 0.00002941, 0.00001388, -0.00002165, 0.00000270, -0.00000241, 0.00000718, -0.00002124, 0.00004345, -0.00002057, -0.00000444, -0.00002751, 0.00006200, -0.00007036, 0.00006486, -0.00003639, -0.00001616, 0.00001309, -0.00002747, -0.00004060, 0.00008695, -0.00010279, 0.00009163, -0.00010235, 0.00006244, -0.00012509, 0.00008100, -0.00027702, -0.00038374, 0.00176411, -0.00092384, 0.00015861, -0.00019144, 0.00008062, -0.00011085, 0.00006723, -0.00006354, 0.00003002, -0.00004969, 0.00007100, -0.00011474, 0.00016924, -0.00014476, 0.00012281, -0.00008644, 0.00002692, 0.00000086, 0.00000803, -0.00002538, -0.00002361, 0.00005098, -0.00003431, 0.00000754, -0.00000688, -0.00001973, 0.00000500, 0.00003805, -0.00003888, 0.00000802, 0.00002260, -0.00003824, 0.00002846, },
		{ -0.00002980, 0.00001687, -0.00001559, 0.00003771, 0.00000143, -0.00002122, 0.00000858, -0.00000549, 0.00000801, -0.00001892, 0.00004667, -0.00003139, 0.00000147, -0.00002408, 0.00006273, -0.00008547, 0.00009908, -0.00008968, 0.00003968, -0.00003911, 0.00000813, -0.00006465, 0.00012059, -0.00015710, 0.00016684, -0.00020116, 0.00017969, -0.00027422, 0.00027963, -0.00064872, 0.00023531, 0.00138566, -0.00101094, 0.00031752, -0.00029289, 0.00015885, -0.00016753, 0.00011806, -0.00010498, 0.00005685, -0.00006189, 0.00006756, -0.00009825, 0.00017279, -0.00016649, 0.00016481, -0.00014895, 0.00009201, -0.00005371, 0.00005191, -0.00007146, 0.00002306, 0.00001765, -0.00001819, -0.00000157, -0.00000318, -0.00002716, 0.00002409, 0.00002362, -0.00004598, 0.00002773, 0.00000203, -0.00002229, 0.00001598, },
		{ -0.00001581, 0.00000507, -0.00000255, 0.00003718, -0.00000887, -0.00001789, 0.00001209, -0.00000735, 0.00000733, -0.00001289, 0.00004269, -0.00003408, -0.00000086, -0.00000776, 0.00004160, -0.00007014, 0.00009740, -0.00010806, 0.00006581, -0.00006737, 0.00002241, -0.00006071, 0.00011503, -0.00016282, 0.00018635, -0.00023878, 0.00022878, -0.00034322, 0.00037623, -0.00091460, 0.00086167, 0.00082181, -0.00087243, 0.00032388, -0.00029463, 0.00016187, -0.00016256, 0.00011640, -0.00010325, 0.00004724, -0.00004233, 0.00003348, -0.00004241, 0.00012806, -0.00013562, 0.00015226, -0.00016362, 0.00012115, -0.00008221, 0.00007503, -0.00010089, 0.00005994, -0.00001298, -0.00000186, -0.00001056, -0.00000196, -0.00002911, 0.00004074, 0.00000282, -0.00004274, 0.00004074, -0.00001667, -0.00000498, 0.00000006, },
		{ -0.00000267, -0.00000736, 0.00001511, 0.00002907, -0.00001583, -0.00001240, 0.00001275, -0.00000745, 0.00000554, -0.00000492, 0.00003422, -0.00003196, -0.00000639, 0.00001390, 0.00000753, -0.00003226, 0.00006402, -0.00009058, 0.00005708, -0.00006481, 0.00000991, -0.00002603, 0.00006802, -0.00011551, 0.00014187, -0.00020210, 0.00019089, -0.00030475, 0.00032388, -0.00098360, 0.00135570, 0.00020882, -0.00059046, 0.00021415, -0.00021582, 0.00010372, -0.00010772, 0.00007212, -0.00006714, 0.00001007, -0.00000124, -0.00001759, 0.00003648, 0.00005058, -0.00006459, 0.00009162, -0.00012951, 0.00010840, -0.00007670, 0.00006883, -0.00010442, 0.00007736, -0.00003238, 0.00000858, -0.00001587, -0.00000513, -0.00002302, 0.00005019, -0.00001969, -0.00003018, 0.00004427, -0.00002881, 0.00000819, -0.00001344, },
		{ 0.00000462, -0.00001458, 0.00003160, 0.00001660, -0.00001955, -0.00000537, 0.00001048, -0.00000547, 0.00000296, 0.00000327, 0.00002418, -0.00002888, -0.00000912, 0.00003106, -0.00002592, 0.00001253, 0.00001412, -0.00004949, 0.00002219, -0.00003789, -0.00002190, 0.00002830, -0.00000635, -0.00003028, 0.00004812, -0.00010481, 0.00007873, -0.00016987, 0.00012317, -0.00081578, 0.00161338, -0.00032112, -0.00026757, 0.00005119, -0.00009848, 0.00001725, -0.00003076, 0.00000904, -0.00001755, -0.00003545, 0.00004251, -0.00006411, 0.00011333, -0.00003322, 0.00002061, 0.00000521, -0.00006212, 0.00006289, -0.00004224, 0.00003572, -0.00008185, 0.00007269, -0.00003685, 0.00001008, -0.00001611, -0.00001247, -0.00000868, 0.00004921, -0.00003870, -0.00001175, 0.00003854, -0.00003239, 0.00001394, -0.00002031, },
		{ 0.00000393, -0.00001289, 0.00004187, 0.00000351, -0.00002077, 0.00000251, 0.00000565, -0.00000139, -0.00000031, 0.00001073, 0.00001445, -0.00002721, -0.00000510, 0.00003600, -0.00004614, 0.00004722, -0.00003267, -0.00000437, -0.00002106, -0.00000250, -0.00005659, 0.00008189, -0.00008288, 0.00006370, -0.00006257, 0.00001767, -0.00006793, 0.00001465, -0.00017540, -0.00043541, 0.00159276, -0.00067399, 0.00000521, -0.00010060, 0.00001124, -0.00006094, 0.00003728, -0.00004616, 0.00002242, -0.00006884, 0.00007014, -0.00008568, 0.00016385, -0.00009609, 0.00009036, -0.00007745, 0.00001327, 0.00000432, 0.00000590, -0.00001235, -0.00004166, 0.00005061, -0.00002827, 0.00000339, -0.00001239, -0.00002149, 0.00001133, 0.00003722, -0.00004991, 0.00000778, 0.00002638, -0.00002845, 0.00001216, -0.00001944, },
		{ -0.00000337, -0.00000201, 0.00004327, -0.00000720, -0.00002025, 0.00001032, -0.00000076, 0.00000433, -0.00000421, 0.00001728, 0.00000536, -0.00002678, 0.00000562, 0.00002635, -0.00004644, 0.00006002, -0.00006005, 0.00002603, -0.00005371, 0.00002359, -0.00007637, 0.00011353, -0.00013492, 0.00013421, -0.00015253, 0.00012165, -0.00019692, 0.00018220, -0.00048598, 0.00007588, 0.00132096, -0.00081318, 0.00017161, -0.00019589, 0.00007983, -0.00010441, 0.00007412, -0.00007460, 0.00003695, -0.00007694, 0.00007104, -0.00007135, 0.00017362, -0.00012001, 0.00012226, -0.00013052, 0.00007137, -0.00004509, 0.00004865, -0.00005889, 0.00000264, 0.00002081, -0.00001275, -0.00000774, -0.00000764, -0.00002826, 0.00003223, 0.00001655, -0.00005113, 0.00002384, 0.00001203, -0.00002030, 0.00000554, -0.00001296, },
		{ -0.00001305, 0.00001498, 0.00003607, -0.00001412, -0.00001829, 0.00001681, -0.00000722, 0.00001062, -0.00000846, 0.00002313, -0.00000386, -0.00002524, 0.00001896, 0.00000607, -0.00002841, 0.00004855, -0.00006097, 0.00003092, -0.00006322, 0.00002833, -0.00006989, 0.00010985, -0.00014466, 0.00015802, -0.00019271, 0.00017145, -0.00026318, 0.00026974, -0.00071392, 0.00060314, 0.00088091, -0.00075986, 0.00021868, -0.00021873, 0.00009555, -0.00010468, 0.00007310, -0.00007123, 0.00002285, -0.00005849, 0.00004688, -0.00002376, 0.00014289, -0.00010220, 0.00010875, -0.00014059, 0.00009579, -0.00006897, 0.00007045, -0.00008906, 0.00003762, -0.00000594, 0.00000213, -0.00001851, -0.00000508, -0.00002891, 0.00004840, -0.00000816, -0.00004280, 0.00003336, -0.00000048, -0.00001199, -0.00000179, -0.00000511, },
		{ -0.00001982, 0.00003287, 0.00002309, -0.00001747, -0.00001467, 0.00002061, -0.00001189, 0.00001591, -0.00001225, 0.00002823, -0.00001400, -0.00001964, 0.00002875, -0.00001639, -0.00000084, 0.00002006, -0.00003951, 0.00001095, -0.00004783, 0.00001005, -0.00003705, 0.00007080, -0.00010970, 0.00012919, -0.00017309, 0.00015242, -0.00024532, 0.00024024, -0.00078535, 0.00103116, 0.00038373, -0.00057690, 0.00017064, -0.00018084, 0.00006788, -0.00007093, 0.00004287, -0.00004451, -0.00001101, -0.00002325, 0.00001022, 0.00004242, 0.00008548, -0.00005503, 0.00005869, -0.00011025, 0.00008395, -0.00006214, 0.00006468, -0.00009487, 0.00005453, -0.00002165, 0.00001028, -0.00002509, -0.00000660, -0.00002129, 0.00005519, -0.00003133, -0.00002767, 0.00003554, -0.00000861, -0.00000667, -0.00000609, -0.00000034, },
		{ -0.00001942, 0.00004622, 0.00000834, -0.00001853, -0.00000902, 0.00002067, -0.00001320, 0.00001851, -0.00001429, 0.00003179, -0.00002477, -0.00000831, 0.00002956, -0.00003172, 0.00002420, -0.00001216, -0.00000833, -0.00002323, -0.00001620, -0.00002263, 0.00001115, 0.00000947, -0.00004356, 0.00006049, -0.00010541, 0.00007552, -0.00015337, 0.00009605, -0.00066968, 0.00127564, -0.00006370, -0.00034357, 0.00007240, -0.00011086, 0.00001941, -0.00002301, 0.00000121, -0.00001074, -0.00004908, 0.00001347, -0.00002134, 0.00010670, 0.00002239, -0.00000006, -0.00000726, -0.00005546, 0.00004639, -0.00003133, 0.00003522, -0.00007728, 0.00005180, -0.00002356, 0.00000908, -0.00002604, -0.00001187, -0.00000597, 0.00005036, -0.00004801, -0.00000983, 0.00003168, -0.00001194, -0.00000547, -0.00000553, -0.00000142, },
		{ -0.00001018, 0.00005125, -0.00000444, -0.00001870, -0.00000128, 0.00001667, -0.00001030, 0.00001713, -0.00001316, 0.00003227, -0.00003434, 0.00000770, 0.00001939, -0.00003371, 0.00003647, -0.00003454, 0.00001717, -0.00005605, 0.00001714, -0.00005513, 0.00005742, -0.00005324, 0.00003012, -0.00002238, -0.00001756, -0.00002799, -0.00002449, -0.00012152, -0.00038788, 0.00130288, -0.00038528, -0.00013055, -0.00002956, -0.00004019, -0.00002557, 0.00001825, -0.00003363, 0.00001384, -0.00007647, 0.00003805, -0.00003268, 0.00015079, -0.00002674, 0.00004097, -0.00006528, 0.00000174, 0.00000127, 0.00000868, -0.00000604, -0.00004492, 0.00003461, -0.00001428, -0.00000025, -0.00002245, -0.00001840, 0.00001373, 0.00003479, -0.00005521, 0.00000662, 0.00002446, -0.00001191, -0.00000735, -0.00000084, -0.00000838, },
		{ 0.00000634, 0.00004691, -0.00001299, -0.00001878, 0.00000788, 0.00000937, -0.00000354, 0.00001148, -0.00000797, 0.00002807, -0.00003970, 0.00002454, 0.00000077, -0.00002180, 0.00003165, -0.00003871, 0.00002512, -0.00007368, 0.00003824, -0.00007343, 0.00008531, -0.00009681, 0.00008677, -0.00009107, 0.00005790, -0.00011929, 0.00009112, -0.00034591, -0.00000453, 0.00113315, -0.00054953, 0.00001734, -0.00010217, 0.00000877, -0.00005049, 0.00003925, -0.00005019, 0.00001959, -0.00008497, 0.00004423, -0.00001702, 0.00016499, -0.00005083, 0.00005436, -0.00009744, 0.00004194, -0.00003333, 0.00004155, -0.00004441, -0.00001006, 0.00001210, -0.00000001, -0.00001380, -0.00001702, -0.00002263, 0.00003274, 0.00001214, -0.00005253, 0.00001870, 0.00001672, -0.00001085, -0.00000980, 0.00000522, -0.00001845, },
		{ 0.00002602, 0.00003493, -0.00001694, -0.00001862, 0.00001704, 0.00000053, 0.00000554, 0.00000254, 0.00000106, 0.00001836, -0.00003790, 0.00003681, -0.00002005, -0.00000122, 0.00001277, -0.00002449, 0.00001172, -0.00006930, 0.00003929, -0.00006985, 0.00008566, -0.00010871, 0.00010996, -0.00012526, 0.00009618, -0.00016710, 0.00014890, -0.00050726, 0.00039294, 0.00082834, -0.00056825, 0.00008654, -0.00013245, 0.00002816, -0.00005097, 0.00003728, -0.00004696, 0.00000616, -0.00007528, 0.00003500, 0.00002215, 0.00015052, -0.00005034, 0.00003883, -0.00009749, 0.00005512, -0.00004615, 0.00005563, -0.00006805, 0.00001620, -0.00000650, 0.00001226, -0.00002689, -0.00001251, -0.00002160, 0.00004596, -0.00001233, -0.00004192, 0.00002519, 0.00001052, -0.00001087, -0.00001014, 0.00000907, -0.00002724, },
		{ 0.00004362, 0.00001898, -0.00001751, -0.00001733, 0.00002438, -0.00000759, 0.00001453, -0.00000754, 0.00001227, 0.00000394, -0.00002755, 0.00003972, -0.00003559, 0.00001937, -0.00001137, 0.00000048, -0.00001794, -0.00004509, 0.00002114, -0.00004576, 0.00005973, -0.00008835, 0.00009657, -0.00011925, 0.00008889, -0.00015906, 0.00012590, -0.00055445, 0.00071997, 0.00046975, -0.00048342, 0.00008945, -0.00012551, 0.00002328, -0.00003341, 0.00001903, -0.00003082, -0.00001920, -0.00005504, 0.00002035, 0.00007341, 0.00011713, -0.00003523, 0.00000445, -0.00007163, 0.00004289, -0.00003599, 0.00004770, -0.00007185, 0.00002764, -0.00001515, 0.00001742, -0.00003600, -0.00001039, -0.00001417, 0.00004993, -0.00003344, -0.00002680, 0.00002659, 0.00000657, -0.00001287, -0.00000680, 0.00000796, -0.00003045, },
		{ 0.00005459, 0.00000322, -0.00001663, -0.00001381, 0.00002824, -0.00001285, 0.00002091, -0.00001600, 0.00002288, -0.00001250, -0.00000986, 0.00003118, -0.00004015, 0.00003150, -0.00003013, 0.00002436, -0.00005247, -0.00001054, -0.00000795, -0.00001032, 0.00001789, -0.00004618, 0.00005643, -0.00008240, 0.00004550, -0.00010517, 0.00002841, -0.00047007, 0.00091721, 0.00013413, -0.00034844, 0.00005250, -0.00009690, 0.00000667, -0.00000960, -0.00000436, -0.00001232, -0.00004619, -0.00003418, 0.00001232, 0.00012265, 0.00007773, -0.00001919, -0.00003275, -0.00003435, 0.00001608, -0.00001071, 0.00002302, -0.00005825, 0.00002432, -0.00001255, 0.00001362, -0.00003967, -0.00001035, -0.00000167, 0.00004384, -0.00004742, -0.00001084, 0.00002451, 0.00000428, -0.00001632, 0.00000000, 0.00000108, -0.00002543, },
		{ 0.00005644, -0.00000904, -0.00001601, -0.00000745, 0.00002772, -0.00001395, 0.00002284, -0.00002039, 0.00002976, -0.00002698, 0.00001126, 0.00001290, -0.00003212, 0.00003040, -0.00003547, 0.00003606, -0.00007910, 0.00002201, -0.00003640, 0.00002375, -0.00002502, 0.00000145, 0.00000695, -0.00003364, -0.00001228, -0.00003248, -0.00011121, -0.00027353, 0.00096279, -0.00012451, -0.00021089, 0.00000342, -0.00006349, -0.00000913, 0.00000942, -0.00002297, -0.00000045, -0.00006650, -0.00001998, 0.00001983, 0.00015842, 0.00004259, -0.00001309, -0.00005804, -0.00000162, -0.00001095, 0.00001723, -0.00000791, -0.00003515, 0.00001140, -0.00000174, 0.00000219, -0.00003842, -0.00001071, 0.00001267, 0.00002957, -0.00005279, 0.00000304, 0.00002086, 0.00000239, -0.00001954, 0.00000834, -0.00001001, -0.00001221, },
		{ 0.00004927, -0.00001622, -0.00001645, 0.00000154, 0.00002297, -0.00001080, 0.00001979, -0.00001938, 0.00003053, -0.00003536, 0.00003023, -0.00001001, -0.00001463, 0.00001681, -0.00002514, 0.00002954, -0.00008856, 0.00004227, -0.00005422, 0.00004550, -0.00005572, 0.00003905, -0.00003442, 0.00000713, -0.00006041, 0.00002650, -0.00024659, -0.00001249, 0.00087209, -0.00028286, -0.00010178, -0.00003785, -0.00003686, -0.00001658, 0.00001771, -0.00003184, 0.00000073, -0.00007684, -0.00001409, 0.00004556, 0.00017519, 0.00001588, -0.00002050, -0.00006375, 0.00001555, -0.00002654, 0.00003627, -0.00003405, -0.00001209, -0.00000368, 0.00001184, -0.00001352, -0.00003381, -0.00000954, 0.00002496, 0.00001086, -0.00005021, 0.00001327, 0.00001713, -0.00000031, -0.00002061, 0.00001534, -0.00002196, 0.00000663, },
		{ 0.00003551, -0.00001855, -0.00001764, 0.00001203, 0.00001518, -0.00000452, 0.00001265, -0.00001327, 0.00002447, -0.00003487, 0.00004164, -0.00003080, 0.00000581, -0.00000347, -0.00000331, 0.00000588, -0.00007819, 0.00004552, -0.00005662, 0.00004959, -0.00006707, 0.00005745, -0.00005680, 0.00002698, -0.00008245, 0.00004672, -0.00033354, 0.00025341, 0.00068712, -0.00034584, -0.00003290, -0.00006248, -0.00002096, -0.00001468, 0.00001537, -0.00003168, -0.00000741, -0.00007916, -0.00001247, 0.00008583, 0.00017345, -0.00000433, -0.00003705, -0.00005138, 0.00001459, -0.00002586, 0.00004011, -0.00004798, 0.00000346, -0.00001430, 0.00002270, -0.00002976, -0.00002733, -0.00000582, 0.00003216, -0.00000812, -0.00004181, 0.00001956, 0.00001398, -0.00000435, -0.00001819, 0.00001835, -0.00003081, 0.00002678, },
		{ 0.00001885, -0.00001763, -0.00001845, 0.00002232, 0.00000611, 0.00000301, 0.00000344, -0.00000394, 0.00001297, -0.00002514, 0.00004211, -0.00004349, 0.00002190, -0.00002215, 0.00002127, -0.00002763, -0.00005213, 0.00003349, -0.00004507, 0.00003755, -0.00005963, 0.00005605, -0.00005879, 0.00002397, -0.00007536, 0.00001970, -0.00034416, 0.00047039, 0.00046039, -0.00033694, -0.00000096, -0.00007096, -0.00001371, -0.00000713, 0.00000661, -0.00002684, -0.00002030, -0.00007835, -0.00000783, 0.00013294, 0.00015722, -0.00002296, -0.00005360, -0.00002960, 0.00000088, -0.00001164, 0.00002918, -0.00004796, 0.00000833, -0.00001681, 0.00002707, -0.00004356, -0.00001973, -0.00000005, 0.00003303, -0.00002406, -0.00003014, 0.00002243, 0.00001136, -0.00000934, -0.00001218, 0.00001595, -0.00003340, 0.00004368, },
		{ 0.00000304, -0.00001554, -0.00001746, 0.00003060, -0.00000241, 0.00000981, -0.00000540, 0.00000578, -0.00000083, -0.00000856, 0.00003146, -0.00004501, 0.00002843, -0.00003182, 0.00003890, -0.00006057, -0.00001901, 0.00001258, -0.00002564, 0.00001598, -0.00003996, 0.00004113, -0.00004707, 0.00000599, -0.00004921, -0.00004431, -0.00027422, 0.00060389, 0.00023961, -0.00028621, 0.00000520, -0.00006858, -0.00001053, 0.00000101, -0.00000351, -0.00002214, -0.00003324, -0.00007897, 0.00000699, 0.00017839, 0.00013074, -0.00004376, -0.00006128, -0.00000945, -0.00001594, 0.00000830, 0.00000926, -0.00003734, 0.00000384, -0.00001112, 0.00002357, -0.00005313, -0.00001114, 0.00000603, 0.00002839, -0.00003526, -0.00001733, 0.00002266, 0.00000887, -0.00001424, -0.00000372, 0.00000848, -0.00002831, 0.00005380, },
		{ -0.00000914, -0.00001407, -0.00001358, 0.00003545, -0.00000896, 0.00001440, -0.00001184, 0.00001307, -0.00001296, 0.00001042, 0.00001279, -0.00003609, 0.00002415, -0.00002884, 0.00004260, -0.00008335, 0.00001165, -0.00000931, -0.00000583, -0.00000686, -0.00001679, 0.00002178, -0.00003176, -0.00001442, -0.00002142, -0.00012132, -0.00014196, 0.00064395, 0.00005752, -0.00022036, -0.00000225, -0.00006101, -0.00000770, 0.00000622, -0.00001170, -0.00002035, -0.00004387, -0.00008271, 0.00003642, 0.00021527, 0.00009635, -0.00006632, -0.00005589, 0.00000079, -0.00002683, 0.00002521, -0.00001177, -0.00002222, -0.00000583, 0.00000021, 0.00001283, -0.00005754, -0.00000172, 0.00001036, 0.00002034, -0.00004149, -0.00000478, 0.00002084, 0.00000620, -0.00001782, 0.00000520, -0.00000214, -0.00001629, 0.00005554, },
		{ -0.00001643, -0.00001412, -0.00000639, 0.00003612, -0.00001280, 0.00001616, -0.00001487, 0.00001608, -0.00001998, 0.00002667, -0.00000868, -0.00002072, 0.00001196, -0.00001453, 0.00003046, -0.00009035, 0.00003277, -0.00002586, 0.00000848, -0.00002446, 0.00000282, 0.00000536, -0.00002114, -0.00002666, -0.00000848, -0.00018334, 0.00002011, 0.00060250, -0.00007130, -0.00015748, -0.00001426, -0.00005184, -0.00000389, 0.00000764, -0.00001737, -0.00002149, -0.00005285, -0.00008761, 0.00008115, 0.00023882, 0.00005458, -0.00008585, -0.00003968, -0.00000122, -0.00002708, 0.00003309, -0.00002742, -0.00000852, -0.00001584, 0.00001348, -0.00000334, -0.00005621, 0.00000768, 0.00001171, 0.00001119, -0.00004346, 0.00000675, 0.00001731, 0.00000337, -0.00001918, 0.00001248, -0.00001308, 0.00000018, 0.00004938, },
		{ -0.00001907, -0.00001562, 0.00000372, 0.00003253, -0.00001381, 0.00001528, -0.00001463, 0.00001448, -0.00002001, 0.00003600, -0.00002745, -0.00000438, -0.00000252, 0.00000557, 0.00000588, -0.00008119, 0.00004141, -0.00003414, 0.00001477, -0.00003396, 0.00001570, -0.00000515, -0.00001819, -0.00002677, -0.00001885, -0.00020849, 0.00017663, 0.00050498, -0.00014748, -0.00010649, -0.00002588, -0.00004231, 0.00000020, 0.00000641, -0.00002187, -0.00002367, -0.00006287, -0.00008894, 0.00013869, 0.00024556, 0.00000622, -0.00009574, -0.00001979, -0.00001189, -0.00001747, 0.00003056, -0.00003461, 0.00000004, -0.00002252, 0.00002518, -0.00002271, -0.00004870, 0.00001521, 0.00001037, 0.00000238, -0.00004187, 0.00001664, 0.00001236, 0.00000073, -0.00001809, 0.00001660, -0.00002152, 0.00001765, 0.00003743, },
		{ -0.00001837, -0.00001764, 0.00001565, 0.00002519, -0.00001227, 0.00001248, -0.00001222, 0.00000947, -0.00001334, 0.00003641, -0.00003951, 0.00000812, -0.00001342, 0.00002411, -0.00002412, -0.00005992, 0.00003866, -0.00003454, 0.00001375, -0.00003594, 0.00002248, -0.00001112, -0.00002033, -0.00001854, -0.00004988, -0.00018779, 0.00029957, 0.00038041, -0.00018087, -0.00006958, -0.00003509, -0.00003252, 0.00000327, 0.00000415, -0.00002678, -0.00002493, -0.00007681, -0.00008083, 0.00020450, 0.00023220, -0.00004520, -0.00009149, -0.00000437, -0.00002445, -0.00000273, 0.00002020, -0.00003381, 0.00000252, -0.00002407, 0.00003266, -0.00004285, -0.00003502, 0.00001862, 0.00000802, -0.00000588, -0.00003696, 0.00002404, 0.00000658, -0.00000133, -0.00001495, 0.00001700, -0.00002545, 0.00003278, 0.00002264, },
		{ -0.00001612, -0.00001880, 0.00002781, 0.00001505, -0.00000865, 0.00000856, -0.00000906, 0.00000325, -0.00000216, 0.00002844, -0.00004340, 0.00001429, -0.00001688, 0.00003472, -0.00005164, -0.00003295, 0.00002815, -0.00002949, 0.00000787, -0.00003274, 0.00002555, -0.00001596, -0.00002216, -0.00001061, -0.00008973, -0.00012638, 0.00037396, 0.00025356, -0.00018435, -0.00004542, -0.00004117, -0.00002273, 0.00000473, 0.00000160, -0.00003261, -0.00002464, -0.00009611, -0.00005770, 0.00027254, 0.00019563, -0.00009236, -0.00007366, 0.00000139, -0.00003268, 0.00001145, 0.00000643, -0.00002763, -0.00000002, -0.00002033, 0.00003411, -0.00006096, -0.00001637, 0.00001630, 0.00000680, -0.00001423, -0.00002851, 0.00002779, 0.00000101, -0.00000265, -0.00001053, 0.00001407, -0.00002410, 0.00004306, 0.00000793, },
		{ -0.00001399, -0.00001765, 0.00003833, 0.00000342, -0.00000347, 0.00000410, -0.00000633, -0.00000183, 0.00001026, 0.00001469, -0.00004019, 0.00001453, -0.00001238, 0.00003413, -0.00007052, -0.00000659, 0.00001410, -0.00002176, -0.00000040, -0.00002648, 0.00002672, -0.00002210, -0.00001932, -0.00001151, -0.00012291, -0.00003977, 0.00039836, 0.00014094, -0.00016975, -0.00003155, -0.00004372, -0.00001377, 0.00000499, -0.00000168, -0.00003843, -0.00002396, -0.00012013, -0.00001497, 0.00033488, 0.00013466, -0.00012562, -0.00004844, -0.00000248, -0.00003381, 0.00002124, -0.00000690, -0.00001892, -0.00000583, -0.00001197, 0.00002831, -0.00007376, 0.00000439, 0.00000845, 0.00000806, -0.00002326, -0.00001639, 0.00002680, -0.00000296, -0.00000352, -0.00000560, 0.00000887, -0.00001790, 0.00004725, -0.00000440, },
		},
		{ { 0.00001096, 0.00000793, -0.00002534, -0.00000191, -0.00001188, 0.00000586, 0.00000626, 0.00000726, 0.00000064, -0.00000647, 0.00001556, -0.00001366, 0.00002528, -0.00001713, 0.00002059, -0.00001435, -0.00000893, -0.00000234, 0.00001859, 0.00001173, -0.00007422, 0.00001623, -0.00003617, -0.00001447, -0.00003021, -0.00002757, -0.00010580, 0.00003664, 0.00036791, 0.00005238, -0.00013401, -0.00003250, -0.00005222, -0.00003250, -0.00013401, 0.00005238, 0.00036791, 0.00003664, -0.00010580, -0.00002757, -0.00003021, -0.00001447, -0.00003617, 0.00001623, -0.00007422, 0.00001173, 0.00001859, -0.00000234, -0.00000893, -0.00001435, 0.00002059, -0.00001713, 0.00002528, -0.00001366, 0.00001556, -0.00000647, 0.00000064, 0.00000726, 0.00000626, 0.00000586, -0.00001188, -0.00000191, -0.00002534, 0.00000793, },
		{ 0.00001138, 0.00000371, -0.00002713, 0.00000202, -0.00001446, 0.00001032, 0.00000430, 0.00000953, -0.00000536, 0.00000005, 0.00001153, -0.00000990, 0.00002128, -0.00001510, 0.00002198, -0.00002065, -0.00000479, -0.00000615, 0.00003088, -0.00000612, -0.00006621, 0.00001685, -0.00003732, -0.00001328, -0.00003049, -0.00003491, -0.00010178, 0.00009316, 0.00032777, 0.00001395, -0.00011755, -0.00003211, -0.00005205, -0.00003390, -0.00015225, 0.00010451, 0.00039731, -0.00002994, -0.00009583, -0.00002632, -0.00002682, -0.00001832, -0.00003224, 0.00001183, -0.00007757, 0.00002728, 0.00000475, 0.00000443, -0.00001538, -0.00000516, 0.00001536, -0.00001585, 0.00002706, -0.00001589, 0.00001793, -0.00001180, 0.00000698, 0.00000398, 0.00000850, 0.00000118, -0.00000878, -0.00000682, -0.00002135, 0.00001034, },
		{ 0.00001249, -0.00000218, -0.00002647, 0.00000452, -0.00001614, 0.00001433, 0.00000272, 0.00001041, -0.00001015, 0.00000680, 0.00000655, -0.00000525, 0.00001578, -0.00001034, 0.00001959, -0.00002360, -0.00000370, -0.00000557, 0.00003907, -0.00002333, -0.00005584, 0.00001535, -0.00003731, -0.00001277, -0.00003037, -0.00004486, -0.00008573, 0.00013652, 0.00028241, -0.00001298, -0.00010349, -0.00003219, -0.00005164, -0.00003677, -0.00017141, 0.00017231, 0.00040915, -0.00010057, -0.00007278, -0.00003282, -0.00001859, -0.00002626, -0.00002435, 0.00000248, -0.00007437, 0.00003769, -0.00000771, 0.00001197, -0.00002262, 0.00000572, 0.00000693, -0.00001124, 0.00002629, -0.00001623, 0.00001818, -0.00001506, 0.00001263, 0.00000028, 0.00001071, -0.00000337, -0.00000562, -0.00001207, -0.00001568, 0.00001108, },
		{ 0.00001393, -0.00000933, -0.00002339, 0.00000530, -0.00001666, 0.00001773, 0.00000145, 0.00000985, -0.00001310, 0.00001284, 0.00000145, -0.00000050, 0.00000979, -0.00000394, 0.00001420, -0.00002357, -0.00000544, -0.00000029, 0.00004167, -0.00003762, -0.00004514, 0.00001317, -0.00003747, -0.00001122, -0.00003241, -0.00005330, -0.00006161, 0.00016625, 0.00023575, -0.00003039, -0.00009226, -0.00003217, -0.00005117, -0.00004171, -0.00018988, 0.00025685, 0.00039624, -0.00016675, -0.00004075, -0.00004624, -0.00000554, -0.00003842, -0.00001243, -0.00001193, -0.00006382, 0.00004093, -0.00001610, 0.00001793, -0.00002876, 0.00001658, -0.00000346, -0.00000383, 0.00002304, -0.00001465, 0.00001613, -0.00001562, 0.00001662, -0.00000305, 0.00001244, -0.00000734, -0.00000293, -0.00001695, -0.00000904, 0.00001057, },
		{ 0.00001514, -0.00001709, -0.00001821, 0.00000427, -0.00001582, 0.00002040, 0.00000033, 0.00000805, -0.00001388, 0.00001742, -0.00000300, 0.00000351, 0.00000442, 0.00000273, 0.00000710, -0.00002161, -0.00000886, 0.00000879, 0.00003847, -0.00004777, -0.00003534, 0.00001103, -0.00003829, -0.00000799, -0.00003813, -0.00005674, -0.00003403, 0.00018368, 0.00019038, -0.00004008, -0.00008402, -0.00003171, -0.00005063, -0.00004974, -0.00020479, 0.00035711, 0.00035244, -0.00021874, -0.00000649, -0.00006318, 0.00001030, -0.00005331, 0.00000218, -0.00003005, -0.00004664, 0.00003637, -0.00001867, 0.00002040, -0.00003202, 0.00002560, -0.00001426, 0.00000544, 0.00001777, -0.00001147, 0.00001201, -0.00001326, 0.00001817, -0.00000521, 0.00001314, -0.00001024, -0.00000125, -0.00002073, -0.00000225, 0.00000937, },
		{ 0.00001551, -0.00002464, -0.00001146, 0.00000153, -0.00001350, 0.00002220, -0.00000083, 0.00000541, -0.00001251, 0.00002005, -0.00000617, 0.00000610, 0.00000066, 0.00000827, -0.00000019, -0.00001919, -0.00001220, 0.00001976, 0.00003057, -0.00005377, -0.00002671, 0.00000873, -0.00003913, -0.00000383, -0.00004724, -0.00005339, -0.00000699, 0.00019085, 0.00014809, -0.00004395, -0.00007838, -0.00003098, -0.00004946, -0.00006271, -0.00021147, 0.00046899, 0.00027444, -0.00024734, 0.00002189, -0.00007840, 0.00002517, -0.00006792, 0.00001676, -0.00004913, -0.00002520, 0.00002504, -0.00001509, 0.00001839, -0.00003120, 0.00003122, -0.00002385, 0.00001537, 0.00001119, -0.00000727, 0.00000641, -0.00000817, 0.00001685, -0.00000559, 0.00001230, -0.00001161, -0.00000103, -0.00002273, 0.00000388, 0.00000809, },
		{ 0.00001446, -0.00003111, -0.00000385, -0.00000266, -0.00000967, 0.00002297, -0.00000216, 0.00000242, -0.00000927, 0.00002054, -0.00000771, 0.00000683, -0.00000075, 0.00001149, -0.00000624, -0.00001777, -0.00001352, 0.00003016, 0.00001999, -0.00005662, -0.00001864, 0.00000528, -0.00003842, -0.00000082, -0.00005764, -0.00004364, 0.00001706, 0.00018957, 0.00011042, -0.00004419, -0.00007426, -0.00003090, -0.00004643, -0.00008333, -0.00020327, 0.00058447, 0.00016335, -0.00024598, 0.00003674, -0.00008608, 0.00003454, -0.00007844, 0.00002782, -0.00006546, -0.00000312, 0.00000958, -0.00000651, 0.00001212, -0.00002598, 0.00003249, -0.00003088, 0.00002479, 0.00000414, -0.00000279, 0.00000023, -0.00000105, 0.00001273, -0.00000386, 0.00000964, -0.00001118, -0.00000257, -0.00002250, 0.00000866, 0.00000730, },
		{ 0.00001158, -0.00003571, 0.00000379, -0.00000789, -0.00000436, 0.00002248, -0.00000363, -0.00000042, -0.00000467, 0.00001902, -0.00000755, 0.00000563, 0.00000046, 0.00001169, -0.00001003, -0.00001843, -0.00001126, 0.00003762, 0.00000908, -0.00005784, -0.00001014, -0.00000057, -0.00003425, -0.00000175, -0.00006583, -0.00002996, 0.00003755, 0.00018094, 0.00007903, -0.00004338, -0.00006991, -0.00003303, -0.00003978, -0.00011480, -0.00017199, 0.00069161, 0.00002573, -0.00021246, 0.00003269, -0.00008138, 0.00003426, -0.00008124, 0.00003192, -0.00007525, 0.00001541, -0.00000637, 0.00000465, 0.00000294, -0.00001704, 0.00002925, -0.00003449, 0.00003269, -0.00000259, 0.00000120, -0.00000549, 0.00000701, 0.00000640, -0.00000015, 0.00000517, -0.00000892, -0.00000588, -0.00001983, 0.00001164, 0.00000737, },
		{ 0.00000669, -0.00003781, 0.00001066, -0.00001360, 0.00000223, 0.00002053, -0.00000512, -0.00000271, 0.00000066, 0.00001589, -0.00000594, 0.00000279, 0.00000405, 0.00000879, -0.00001122, -0.00002148, -0.00000466, 0.00004042, -0.00000010, -0.00005882, -0.00000042, -0.00000974, -0.00002508, -0.00000916, -0.00006793, -0.00001612, 0.00005552, 0.00016524, 0.00005562, -0.00004422, -0.00006323, -0.00003915, -0.00002781, -0.00016003, -0.00010900, 0.00077539, -0.00012626, -0.00015013, 0.00000822, -0.00006192, 0.00002190, -0.00007398, 0.00002671, -0.00007555, 0.00002660, -0.00001894, 0.00001530, -0.00000693, -0.00000589, 0.00002210, -0.00003437, 0.00003837, -0.00000835, 0.00000403, -0.00000969, 0.00001467, -0.00000112, 0.00000495, -0.00000071, -0.00000513, -0.00001069, -0.00001488, 0.00001266, 0.00000843, },
		{ -0.00000006, -0.00003703, 0.00001598, -0.00001914, 0.00000976, 0.00001699, -0.00000640, -0.00000414, 0.00000603, 0.00001175, -0.00000343, -0.00000102, 0.00000930, 0.00000331, -0.00001016, -0.00002641, 0.00000593, 0.00003797, -0.00000635, -0.00006035, 0.00001063, -0.00002215, -0.00001062, -0.00002439, -0.00006089, -0.00000601, 0.00007276, 0.00014244, 0.00004154, -0.00004903, -0.00005249, -0.00005055, -0.00000974, -0.00022030, -0.00000698, 0.00081960, -0.00027602, -0.00006788, -0.00003361, -0.00002873, -0.00000236, -0.00005644, 0.00001177, -0.00006517, 0.00002805, -0.00002497, 0.00002247, -0.00001502, 0.00000543, 0.00001228, -0.00003078, 0.00004151, -0.00001273, 0.00000527, -0.00001148, 0.00002052, -0.00000846, 0.00001048, -0.00000722, -0.00000046, -0.00001638, -0.00000812, 0.00001191, 0.00001032, },
		{ -0.00000822, -0.00003331, 0.00001914, -0.00002378, 0.00001772, 0.00001189, -0.00000721, -0.00000456, 0.00001078, 0.00000733, -0.00000071, -0.00000493, 0.00001513, -0.00000369, -0.00000780, -0.00003191, 0.00001907, 0.00003093, -0.00000963, -0.00006220, 0.00002204, -0.00003639, 0.00000774, -0.00004671, -0.00004360, -0.00000240, 0.00009079, 0.00011291, 0.00003717, -0.00005897, -0.00003704, -0.00006719, 0.00001325, -0.00029385, 0.00013780, 0.00080943, -0.00040460, 0.00002097, -0.00008535, 0.00001358, -0.00003536, -0.00003099, -0.00001096, -0.00004524, 0.00001942, -0.00002282, 0.00002402, -0.00001923, 0.00001489, 0.00000134, -0.00002438, 0.00004211, -0.00001561, 0.00000481, -0.00001033, 0.00002335, -0.00001420, 0.00001520, -0.00001334, 0.00000420, -0.00002209, -0.00000032, 0.00000983, 0.00001262, },
		{ -0.00001706, -0.00002690, 0.00001972, -0.00002681, 0.00002540, 0.00000546, -0.00000732, -0.00000393, 0.00001435, 0.00000336, 0.00000144, -0.00000795, 0.00002026, -0.00001087, -0.00000549, -0.00003617, 0.00003250, 0.00002114, -0.00001102, -0.00006320, 0.00003184, -0.00004971, 0.00002677, -0.00007308, -0.00001755, -0.00000593, 0.00010985, 0.00007817, 0.00004137, -0.00007354, -0.00001799, -0.00008705, 0.00003731, -0.00037467, 0.00032346, 0.00073450, -0.00049377, 0.00010081, -0.00013651, 0.00005749, -0.00007138, -0.00000229, -0.00003730, -0.00001924, 0.00000270, -0.00001286, 0.00001928, -0.00001835, 0.00002089, -0.00000918, -0.00001605, 0.00004042, -0.00001707, 0.00000283, -0.00000616, 0.00002236, -0.00001709, 0.00001791, -0.00001793, 0.00000783, -0.00002682, 0.00000757, 0.00000710, 0.00001470, },
		{ -0.00002568, -0.00001838, 0.00001758, -0.00002760, 0.00003197, -0.00000182, -0.00000664, -0.00000231, 0.00001632, 0.00000049, 0.00000232, -0.00000917, 0.00002347, -0.00001686, -0.00000457, -0.00003732, 0.00004360, 0.00001107, -0.00001238, -0.00006151, 0.00003743, -0.00005854, 0.00004192, -0.00009840, 0.00001310, -0.00001467, 0.00012832, 0.00004128, 0.00005114, -0.00009026, 0.00000158, -0.00010587, 0.00005561, -0.00045188, 0.00054086, 0.00059166, -0.00052933, 0.00015667, -0.00017560, 0.00009408, -0.00010320, 0.00002355, -0.00006152, 0.00000755, -0.00001801, 0.00000249, 0.00000925, -0.00001242, 0.00002263, -0.00001811, -0.00000669, 0.00003682, -0.00001737, -0.00000014, 0.00000058, 0.00001732, -0.00001629, 0.00001761, -0.00001998, 0.00000949, -0.00002961, 0.00001457, 0.00000443, 0.00001587, },
		{ -0.00003310, -0.00000856, 0.00001286, -0.00002568, 0.00003658, -0.00000929, -0.00000520, 0.00000017, 0.00001647, -0.00000081, 0.00000143, -0.00000794, 0.00002381, -0.00002056, -0.00000608, -0.00003391, 0.00005002, 0.00000331, -0.00001572, -0.00005525, 0.00003629, -0.00005925, 0.00004825, -0.00011644, 0.00004222, -0.00002444, 0.00014283, 0.00000677, 0.00006187, -0.00010491, 0.00001687, -0.00011758, 0.00005892, -0.00051016, 0.00077315, 0.00038687, -0.00050396, 0.00017709, -0.00019256, 0.00011494, -0.00012372, 0.00004033, -0.00007760, 0.00002906, -0.00003735, 0.00001926, -0.00000352, -0.00000281, 0.00002029, -0.00002477, 0.00000299, 0.00003173, -0.00001686, -0.00000339, 0.00000897, 0.00000868, -0.00001160, 0.00001378, -0.00001884, 0.00000847, -0.00002970, 0.00001983, 0.00000248, 0.00001549, },
		{ -0.00003844, 0.00000158, 0.00000603, -0.00002088, 0.00003847, -0.00001615, -0.00000316, 0.00000329, 0.00001478, -0.00000033, -0.00000146, -0.00000399, 0.00002080, -0.00002136, -0.00001039, -0.00002532, 0.00005023, -0.00000019, -0.00002261, -0.00004317, 0.00002679, -0.00004924, 0.00004165, -0.00012128, 0.00006300, -0.00002976, 0.00014888, -0.00002003, 0.00006807, -0.00011238, 0.00002228, -0.00011547, 0.00003707, -0.00053135, 0.00099675, 0.00013579, -0.00041927, 0.00015668, -0.00018113, 0.00011417, -0.00012764, 0.00004327, -0.00008060, 0.00003984, -0.00004983, 0.00003279, -0.00001557, 0.00000817, 0.00001492, -0.00002910, 0.00001248, 0.00002550, -0.00001586, -0.00000611, 0.00001769, -0.00000245, -0.00000352, 0.00000652, -0.00001434, 0.00000451, -0.00002668, 0.00002279, 0.00000170, 0.00001310, },
		{ -0.00004100, 0.00001101, -0.00000218, -0.00001337, 0.00003711, -0.00002162, -0.00000077, 0.00000680, 0.00001148, 0.00000183, -0.00000617, 0.00000249, 0.00001448, -0.00001927, -0.00001715, -0.00001201, 0.00004394, 0.00000139, -0.00003353, -0.00002527, 0.00000888, -0.00002784, 0.00002007, -0.00010874, 0.00006959, -0.00002534, 0.00014211, -0.00003410, 0.00006456, -0.00010792, 0.00001290, -0.00009384, -0.00001882, -0.00049718, 0.00118380, -0.00013740, -0.00028618, 0.00009746, -0.00014043, 0.00008996, -0.00011289, 0.00003034, -0.00006804, 0.00003634, -0.00005117, 0.00003882, -0.00002328, 0.00001770, 0.00000815, -0.00003148, 0.00002153, 0.00001840, -0.00001463, -0.00000753, 0.00002527, -0.00001445, 0.00000676, -0.00000340, -0.00000695, -0.00000214, -0.00002061, 0.00002324, 0.00000221, 0.00000858, },
		{ -0.00004041, 0.00001875, -0.00001083, -0.00000367, 0.00003232, -0.00002505, 0.00000171, 0.00001037, 0.00000704, 0.00000525, -0.00001221, 0.00001087, 0.00000545, -0.00001489, -0.00002521, 0.00000448, 0.00003227, 0.00000748, -0.00004760, -0.00000313, -0.00001546, 0.00000311, -0.00001554, -0.00007772, 0.00005866, -0.00000761, 0.00011965, -0.00003209, 0.00004785, -0.00008864, -0.00001396, -0.00005001, -0.00011377, -0.00039269, 0.00130593, -0.00040287, -0.00012369, 0.00000884, -0.00007553, 0.00004526, -0.00008136, 0.00000304, -0.00004070, 0.00001798, -0.00003939, 0.00003461, -0.00002379, 0.00002323, 0.00000175, -0.00003254, 0.00003001, 0.00001067, -0.00001333, -0.00000706, 0.00003023, -0.00002544, 0.00001758, -0.00001463, 0.00000233, -0.00001071, -0.00001201, 0.00002139, 0.00000382, 0.00000214, },
		{ -0.00003667, 0.00002403, -0.00001885, 0.00000729, 0.00002438, -0.00002606, 0.00000408, 0.00001357, 0.00000211, 0.00000923, -0.00001871, 0.00002011, -0.00000522, -0.00000930, -0.00003292, 0.00002173, 0.00001752, 0.00001616, -0.00006262, 0.00002017, -0.00004233, 0.00003916, -0.00006094, -0.00003093, 0.00003051, 0.00002397, 0.00008137, -0.00001346, 0.00001735, -0.00005471, -0.00005732, 0.00001402, -0.00024641, -0.00020995, 0.00133880, -0.00062925, 0.00004391, -0.00009385, 0.00000329, -0.00001248, -0.00003886, -0.00003357, -0.00000286, -0.00001245, -0.00001545, 0.00001965, -0.00001580, 0.00002305, -0.00000285, -0.00003285, 0.00003776, 0.00000257, -0.00001200, -0.00000441, 0.00003140, -0.00003353, 0.00002699, -0.00002546, 0.00001207, -0.00002002, -0.00000184, 0.00001782, 0.00000602, -0.00000566, },
		{ -0.00003016, 0.00002634, -0.00002519, 0.00001835, 0.00001399, -0.00002462, 0.00000625, 0.00001595, -0.00000250, 0.00001289, -0.00002462, 0.00002894, -0.00001621, -0.00000376, -0.00003843, 0.00003694, 0.00000278, 0.00002460, -0.00007545, 0.00004057, -0.00006644, 0.00007385, -0.00010898, 0.00002519, -0.00001054, 0.00006629, 0.00003055, 0.00001884, -0.00002399, -0.00001014, -0.00011179, 0.00009050, -0.00040734, 0.00004891, 0.00126657, -0.00078851, 0.00019102, -0.00019212, 0.00008236, -0.00007269, 0.00000597, -0.00007169, 0.00003830, -0.00004908, 0.00001667, -0.00000395, 0.00000003, 0.00001682, -0.00000494, -0.00003270, 0.00004452, -0.00000557, -0.00001061, 0.00000037, 0.00002810, -0.00003718, 0.00003318, -0.00003407, 0.00002061, -0.00002862, 0.00000871, 0.00001333, 0.00000808, -0.00001396, },
		{ -0.00002160, 0.00002552, -0.00002895, 0.00002819, 0.00000221, -0.00002103, 0.00000828, 0.00001704, -0.00000596, 0.00001524, -0.00002881, 0.00003602, -0.00002614, 0.00000049, -0.00004014, 0.00004745, -0.00000874, 0.00002967, -0.00008276, 0.00005379, -0.00008214, 0.00009986, -0.00015080, 0.00008142, -0.00005674, 0.00011282, -0.00002624, 0.00005851, -0.00006975, 0.00003745, -0.00016775, 0.00016613, -0.00057870, 0.00036989, 0.00108556, -0.00086060, 0.00029509, -0.00026779, 0.00014733, -0.00012373, 0.00004336, -0.00010231, 0.00007409, -0.00008404, 0.00005063, -0.00003181, 0.00002107, 0.00000565, -0.00000469, -0.00003195, 0.00004980, -0.00001327, -0.00000910, 0.00000683, 0.00002026, -0.00003548, 0.00003472, -0.00003889, 0.00002645, -0.00003511, 0.00001837, 0.00000881, 0.00000924, -0.00002172, },
		{ -0.00001197, 0.00002180, -0.00002953, 0.00003554, -0.00000964, -0.00001589, 0.00001031, 0.00001640, -0.00000750, 0.00001542, -0.00003024, 0.00004013, -0.00003378, 0.00000257, -0.00003703, 0.00005132, -0.00001443, 0.00002877, -0.00008181, 0.00005638, -0.00008463, 0.00011044, -0.00017744, 0.00012742, -0.00009830, 0.00015471, -0.00008011, 0.00009683, -0.00011109, 0.00007789, -0.00021268, 0.00022354, -0.00073547, 0.00072719, 0.00080618, -0.00083699, 0.00034053, -0.00030647, 0.00018604, -0.00015538, 0.00006459, -0.00011718, 0.00009603, -0.00010896, 0.00007898, -0.00005805, 0.00004319, -0.00000808, -0.00000314, -0.00003001, 0.00005293, -0.00001994, -0.00000740, 0.00001423, 0.00000853, -0.00002830, 0.00003090, -0.00003887, 0.00002841, -0.00003832, 0.00002603, 0.00000504, 0.00000882, -0.00002793, },
		{ -0.00000235, 0.00001575, -0.00002672, 0.00003939, -0.00002023, -0.00000997, 0.00001255, 0.00001379, -0.00000661, 0.00001277, -0.00002819, 0.00004037, -0.00003827, 0.00000209, -0.00002899, 0.00004780, -0.00001285, 0.00002043, -0.00007127, 0.00004662, -0.00007118, 0.00010101, -0.00018177, 0.00015376, -0.00012537, 0.00018261, -0.00012154, 0.00012438, -0.00013850, 0.00010046, -0.00023343, 0.00024402, -0.00084844, 0.00108541, 0.00045270, -0.00072253, 0.00032156, -0.00030044, 0.00019109, -0.00016108, 0.00006395, -0.00011068, 0.00009780, -0.00011683, 0.00009473, -0.00007654, 0.00006161, -0.00002123, -0.00000190, -0.00002605, 0.00005314, -0.00002493, -0.00000548, 0.00002156, -0.00000582, -0.00001637, 0.00002184, -0.00003372, 0.00002594, -0.00003755, 0.00003093, 0.00000254, 0.00000643, -0.00003172, },
		{ 0.00000618, 0.00000823, -0.00002075, 0.00003912, -0.00002838, -0.00000407, 0.00001511, 0.00000918, -0.00000309, 0.00000704, -0.00002234, 0.00003632, -0.00003921, -0.00000071, -0.00001687, 0.00003748, -0.00000407, 0.00000481, -0.00005165, 0.00002515, -0.00004196, 0.00007040, -0.00016013, 0.00015402, -0.00013022, 0.00018870, -0.00014238, 0.00013304, -0.00014397, 0.00009630, -0.00021916, 0.00021118, -0.00088855, 0.00140370, 0.00006076, -0.00053500, 0.00024322, -0.00025015, 0.00016145, -0.00013941, 0.00004026, -0.00008134, 0.00007676, -0.00010362, 0.00009287, -0.00008222, 0.00007192, -0.00003063, -0.00000268, -0.00001925, 0.00004981, -0.00002762, -0.00000336, 0.00002773, -0.00002105, -0.00000121, 0.00000853, -0.00002393, 0.00001920, -0.00003271, 0.00003276, 0.00000147, 0.00000205, -0.00003255, },
		{ 0.00001275, 0.00000029, -0.00001234, 0.00003466, -0.00003328, 0.00000117, 0.00001797, 0.00000287, 0.00000281, -0.00000153, -0.00001289, 0.00002807, -0.00003668, -0.00000508, -0.00000239, 0.00002228, 0.00001025, -0.00001625, -0.00002545, -0.00000489, -0.00000039, 0.00002142, -0.00011346, 0.00012627, -0.00010909, 0.00016866, -0.00013782, 0.00011805, -0.00012311, 0.00006082, -0.00016423, 0.00011482, -0.00083200, 0.00164137, -0.00032732, -0.00030253, 0.00012044, -0.00016422, 0.00010277, -0.00009464, -0.00000264, -0.00003247, 0.00003487, -0.00006937, 0.00007163, -0.00007230, 0.00007113, -0.00003386, -0.00000680, -0.00000915, 0.00004264, -0.00002759, -0.00000107, 0.00003167, -0.00003516, 0.00001513, -0.00000730, -0.00001080, 0.00000906, -0.00002431, 0.00003171, 0.00000157, -0.00000393, -0.00003028, },
		{ 0.00001676, -0.00000700, -0.00000255, 0.00002651, -0.00003459, 0.00000537, 0.00002087, -0.00000450, 0.00001039, -0.00001220, -0.00000061, 0.00001628, -0.00003124, -0.00000989, 0.00001216, 0.00000510, 0.00002719, -0.00003949, 0.00000324, -0.00003829, 0.00004722, -0.00003925, -0.00004747, 0.00007386, -0.00006328, 0.00012296, -0.00010770, 0.00007939, -0.00007673, -0.00000453, -0.00007051, -0.00004575, -0.00066512, 0.00176406, -0.00066843, -0.00005920, -0.00002467, -0.00005779, 0.00002634, -0.00003590, -0.00005642, 0.00002817, -0.00002144, -0.00001849, 0.00003308, -0.00004698, 0.00005834, -0.00002983, -0.00001473, 0.00000412, 0.00003184, -0.00002473, 0.00000131, 0.00003253, -0.00004622, 0.00003032, -0.00002349, 0.00000386, -0.00000305, -0.00001341, 0.00002840, 0.00000227, -0.00001072, -0.00002523, },
		{ 0.00001798, -0.00001265, 0.00000731, 0.00001566, -0.00003252, 0.00000848, 0.00002333, -0.00001198, 0.00001855, -0.00002375, 0.00001324, 0.00000207, -0.00002375, -0.00001390, 0.00002437, -0.00001077, 0.00004317, -0.00006091, 0.00002943, -0.00006859, 0.00009274, -0.00010204, 0.00002808, 0.00000502, 0.00000077, 0.00005716, -0.00005702, 0.00002232, -0.00001136, -0.00009155, 0.00005156, -0.00025968, -0.00038818, 0.00174943, -0.00092500, 0.00016042, -0.00016686, 0.00005040, -0.00005311, 0.00002454, -0.00010982, 0.00008973, -0.00008223, 0.00004086, -0.00001705, -0.00000964, 0.00003511, -0.00001914, -0.00002587, 0.00001962, 0.00001827, -0.00001935, 0.00000372, 0.00002979, -0.00005261, 0.00004215, -0.00003774, 0.00001794, -0.00001535, -0.00000139, 0.00002373, 0.00000279, -0.00001729, -0.00001812, },
		{ 0.00001659, -0.00001596, 0.00001588, 0.00000350, -0.00002773, 0.00001074, 0.00002473, -0.00001843, 0.00002591, -0.00003466, 0.00002704, -0.00001311, -0.00001523, -0.00001605, 0.00003216, -0.00002222, 0.00005466, -0.00007657, 0.00004828, -0.00008944, 0.00012765, -0.00015623, 0.00010105, -0.00006846, 0.00007249, -0.00001888, 0.00000481, -0.00004333, 0.00006145, -0.00018607, 0.00018362, -0.00050456, -0.00001720, 0.00159129, -0.00107069, 0.00032680, -0.00028230, 0.00014162, -0.00012039, 0.00007387, -0.00015090, 0.00014034, -0.00013591, 0.00009816, -0.00007035, 0.00003376, 0.00000521, -0.00000400, -0.00003856, 0.00003570, 0.00000338, -0.00001214, 0.00000611, 0.00002335, -0.00005330, 0.00004884, -0.00004803, 0.00002946, -0.00002604, 0.00001025, 0.00001869, 0.00000236, -0.00002252, -0.00000998, },
		{ 0.00001312, -0.00001654, 0.00002195, -0.00000843, -0.00002121, 0.00001262, 0.00002436, -0.00002270, 0.00003103, -0.00004329, 0.00003904, -0.00002763, -0.00000670, -0.00001568, 0.00003420, -0.00002696, 0.00005890, -0.00008344, 0.00005612, -0.00009581, 0.00014475, -0.00019187, 0.00015908, -0.00013345, 0.00013915, -0.00009273, 0.00006562, -0.00010486, 0.00012704, -0.00027008, 0.00030175, -0.00074845, 0.00041669, 0.00130143, -0.00109421, 0.00041999, -0.00035277, 0.00020074, -0.00016292, 0.00010135, -0.00016957, 0.00016950, -0.00017150, 0.00014256, -0.00011725, 0.00007572, -0.00002611, 0.00001218, -0.00005033, 0.00005018, -0.00001094, -0.00000415, 0.00000845, 0.00001356, -0.00004798, 0.00004938, -0.00005290, 0.00003682, -0.00003361, 0.00002019, 0.00001411, 0.00000036, -0.00002540, -0.00000196, },
		{ 0.00000836, -0.00001445, 0.00002467, -0.00001866, -0.00001410, 0.00001463, 0.00002165, -0.00002381, 0.00003265, -0.00004812, 0.00004757, -0.00003994, 0.00000099, -0.00001267, 0.00003017, -0.00002396, 0.00005450, -0.00008006, 0.00005127, -0.00008520, 0.00013966, -0.00020180, 0.00019199, -0.00017796, 0.00018834, -0.00015182, 0.00011290, -0.00014925, 0.00017044, -0.00032501, 0.00038028, -0.00095406, 0.00087052, 0.00090866, -0.00100093, 0.00043254, -0.00036872, 0.00021902, -0.00017317, 0.00010069, -0.00015978, 0.00017031, -0.00018089, 0.00016503, -0.00014891, 0.00010866, -0.00005311, 0.00002554, -0.00005847, 0.00006077, -0.00002269, 0.00000339, 0.00001068, 0.00000124, -0.00003716, 0.00004361, -0.00005173, 0.00003911, -0.00003705, 0.00002747, 0.00001057, -0.00000346, -0.00002527, 0.00000480, },
		{ 0.00000327, -0.00001017, 0.00002366, -0.00002604, -0.00000742, 0.00001721, 0.00001630, -0.00002119, 0.00002994, -0.00004806, 0.00005130, -0.00004872, 0.00000724, -0.00000749, 0.00002086, -0.00001364, 0.00004172, -0.00006691, 0.00003449, -0.00005823, 0.00011187, -0.00018314, 0.00019377, -0.00019351, 0.00021044, -0.00018597, 0.00013641, -0.00016598, 0.00017956, -0.00033549, 0.00039650, -0.00108425, 0.00129491, 0.00045515, -0.00081165, 0.00037010, -0.00033067, 0.00019563, -0.00015019, 0.00007138, -0.00012093, 0.00014105, -0.00016055, 0.00016015, -0.00015895, 0.00012645, -0.00007079, 0.00003257, -0.00006060, 0.00006553, -0.00003010, 0.00000931, 0.00001272, -0.00001245, -0.00002207, 0.00003229, -0.00004477, 0.00003623, -0.00003600, 0.00003163, 0.00000822, -0.00000891, -0.00002187, 0.00000938, },
		{ -0.00000120, -0.00000449, 0.00001909, -0.00002990, -0.00000195, 0.00002052, 0.00000842, -0.00001479, 0.00002264, -0.00004263, 0.00004947, -0.00005306, 0.00001174, -0.00000110, 0.00000799, 0.00000224, 0.00002239, -0.00004634, 0.00000894, -0.00001869, 0.00006498, -0.00013804, 0.00016382, -0.00017686, 0.00020067, -0.00018953, 0.00013046, -0.00014945, 0.00014799, -0.00029291, 0.00033535, -0.00110823, 0.00164089, -0.00000943, -0.00055905, 0.00024976, -0.00024855, 0.00013752, -0.00009962, 0.00001887, -0.00005822, 0.00008572, -0.00011248, 0.00012737, -0.00014478, 0.00012562, -0.00007590, 0.00003097, -0.00005525, 0.00006333, -0.00003198, 0.00001266, 0.00001439, -0.00002609, -0.00000452, 0.00001694, -0.00003314, 0.00002886, -0.00003082, 0.00003275, 0.00000680, -0.00001536, -0.00001552, 0.00001124, },
		{ -0.00000424, 0.00000153, 0.00001168, -0.00003011, 0.00000196, 0.00002438, -0.00000140, -0.00000522, 0.00001125, -0.00003208, 0.00004198, -0.00005256, 0.00001447, 0.00000526, -0.00000609, 0.00002094, -0.00000044, -0.00002208, -0.00002042, 0.00002704, 0.00000621, -0.00007336, 0.00010719, -0.00013073, 0.00016010, -0.00016267, 0.00009527, -0.00010059, 0.00007697, -0.00019794, 0.00019316, -0.00100708, 0.00186704, -0.00043492, -0.00028220, 0.00009653, -0.00013950, 0.00005779, -0.00003249, -0.00004645, 0.00001830, 0.00001342, -0.00004397, 0.00007118, -0.00010815, 0.00010609, -0.00006762, 0.00002015, -0.00004233, 0.00005416, -0.00002800, 0.00001297, 0.00001546, -0.00003819, 0.00001338, -0.00000039, -0.00001854, 0.00001833, -0.00002242, 0.00003135, 0.00000577, -0.00002183, -0.00000699, 0.00001026, },
		{ -0.00000532, 0.00000678, 0.00000253, -0.00002712, 0.00000438, 0.00002825, -0.00001213, 0.00000637, -0.00000306, -0.00001744, 0.00002950, -0.00004739, 0.00001560, 0.00001029, -0.00001884, 0.00003926, -0.00002317, 0.00000145, -0.00004777, 0.00007119, -0.00005488, 0.00000042, 0.00003368, -0.00006328, 0.00009557, -0.00011140, 0.00003717, -0.00002708, -0.00002423, -0.00006145, -0.00002032, -0.00077779, 0.00194566, -0.00077779, -0.00002032, -0.00006145, -0.00002423, -0.00002708, 0.00003717, -0.00011140, 0.00009557, -0.00006328, 0.00003368, 0.00000042, -0.00005488, 0.00007119, -0.00004777, 0.00000145, -0.00002317, 0.00003926, -0.00001884, 0.00001029, 0.00001560, -0.00004739, 0.00002950, -0.00001744, -0.00000306, 0.00000637, -0.00001213, 0.00002825, 0.00000438, -0.00002712, 0.00000253, 0.00000678, },
		{ -0.00000424, 0.00001026, -0.00000699, -0.00002183, 0.00000577, 0.00003135, -0.00002242, 0.00001833, -0.00001854, -0.00000039, 0.00001338, -0.00003819, 0.00001546, 0.00001297, -0.00002800, 0.00005416, -0.00004233, 0.00002015, -0.00006762, 0.00010609, -0.00010815, 0.00007118, -0.00004397, 0.00001342, 0.00001830, -0.00004645, -0.00003249, 0.00005779, -0.00013950, 0.00009653, -0.00028220, -0.00043492, 0.00186704, -0.00100708, 0.00019316, -0.00019794, 0.00007697, -0.00010059, 0.00009527, -0.00016267, 0.00016010, -0.00013073, 0.00010719, -0.00007336, 0.00000621, 0.00002704, -0.00002042, -0.00002208, -0.00000044, 0.00002094, -0.00000609, 0.00000526, 0.00001447, -0.00005256, 0.00004198, -0.00003208, 0.00001125, -0.00000522, -0.00000140, 0.00002438, 0.00000196, -0.00003011, 0.00001168, 0.00000153, },
		{ -0.00000120, 0.00001124, -0.00001552, -0.00001536, 0.00000680, 0.00003275, -0.00003082, 0.00002886, -0.00003314, 0.00001694, -0.00000452, -0.00002609, 0.00001439, 0.00001266, -0.00003198, 0.00006333, -0.00005525, 0.00003097, -0.00007590, 0.00012562, -0.00014478, 0.00012737, -0.00011248, 0.00008572, -0.00005822, 0.00001887, -0.00009962, 0.00013752, -0.00024855, 0.00024976, -0.00055905, -0.00000943, 0.00164089, -0.00110823, 0.00033535, -0.00029291, 0.00014799, -0.00014945, 0.00013046, -0.00018953, 0.00020067, -0.00017686, 0.00016382, -0.00013804, 0.00006498, -0.00001869, 0.00000894, -0.00004634, 0.00002239, 0.00000224, 0.00000799, -0.00000110, 0.00001174, -0.00005306, 0.00004947, -0.00004263, 0.00002264, -0.00001479, 0.00000842, 0.00002052, -0.00000195, -0.00002990, 0.00001909, -0.00000449, },
		{ 0.00000327, 0.00000938, -0.00002187, -0.00000891, 0.00000822, 0.00003163, -0.00003600, 0.00003623, -0.00004477, 0.00003229, -0.00002207, -0.00001245, 0.00001272, 0.00000931, -0.00003010, 0.00006553, -0.00006060, 0.00003257, -0.00007079, 0.00012645, -0.00015895, 0.00016015, -0.00016055, 0.00014105, -0.00012093, 0.00007138, -0.00015019, 0.00019563, -0.00033067, 0.00037010, -0.00081165, 0.00045515, 0.00129491, -0.00108425, 0.00039650, -0.00033549, 0.00017956, -0.00016598, 0.00013641, -0.00018597, 0.00021044, -0.00019351, 0.00019377, -0.00018314, 0.00011187, -0.00005823, 0.00003449, -0.00006691, 0.00004172, -0.00001364, 0.00002086, -0.00000749, 0.00000724, -0.00004872, 0.00005130, -0.00004806, 0.00002994, -0.00002119, 0.00001630, 0.00001721, -0.00000742, -0.00002604, 0.00002366, -0.00001017, },
		{ 0.00000836, 0.00000480, -0.00002527, -0.00000346, 0.00001057, 0.00002747, -0.00003705, 0.00003911, -0.00005173, 0.00004361, -0.00003716, 0.00000124, 0.00001068, 0.00000339, -0.00002269, 0.00006077, -0.00005847, 0.00002554, -0.00005311, 0.00010866, -0.00014891, 0.00016503, -0.00018089, 0.00017031, -0.00015978, 0.00010069, -0.00017317, 0.00021902, -0.00036872, 0.00043254, -0.00100093, 0.00090866, 0.00087052, -0.00095406, 0.00038028, -0.00032501, 0.00017044, -0.00014925, 0.00011290, -0.00015182, 0.00018834, -0.00017796, 0.00019199, -0.00020180, 0.00013966, -0.00008520, 0.00005127, -0.00008006, 0.00005450, -0.00002396, 0.00003017, -0.00001267, 0.00000099, -0.00003994, 0.00004757, -0.00004812, 0.00003265, -0.00002381, 0.00002165, 0.00001463, -0.00001410, -0.00001866, 0.00002467, -0.00001445, },
		{ 0.00001312, -0.00000196, -0.00002540, 0.00000036, 0.00001411, 0.00002019, -0.00003361, 0.00003682, -0.00005290, 0.00004938, -0.00004798, 0.00001356, 0.00000845, -0.00000415, -0.00001094, 0.00005018, -0.00005033, 0.00001218, -0.00002611, 0.00007572, -0.00011725, 0.00014256, -0.00017150, 0.00016950, -0.00016957, 0.00010135, -0.00016292, 0.00020074, -0.00035277, 0.00041999, -0.00109421, 0.00130143, 0.00041669, -0.00074845, 0.00030175, -0.00027008, 0.00012704, -0.00010486, 0.00006562, -0.00009273, 0.00013915, -0.00013345, 0.00015908, -0.00019187, 0.00014475, -0.00009581, 0.00005612, -0.00008344, 0.00005890, -0.00002696, 0.00003420, -0.00001568, -0.00000670, -0.00002763, 0.00003904, -0.00004329, 0.00003103, -0.00002270, 0.00002436, 0.00001262, -0.00002121, -0.00000843, 0.00002195, -0.00001654, },
		{ 0.00001659, -0.00000998, -0.00002252, 0.00000236, 0.00001869, 0.00001025, -0.00002604, 0.00002946, -0.00004803, 0.00004884, -0.00005330, 0.00002335, 0.00000611, -0.00001214, 0.00000338, 0.00003570, -0.00003856, -0.00000400, 0.00000521, 0.00003376, -0.00007035, 0.00009816, -0.00013591, 0.00014034, -0.00015090, 0.00007387, -0.00012039, 0.00014162, -0.00028230, 0.00032680, -0.00107069, 0.00159129, -0.00001720, -0.00050456, 0.00018362, -0.00018607, 0.00006145, -0.00004333, 0.00000481, -0.00001888, 0.00007249, -0.00006846, 0.00010105, -0.00015623, 0.00012765, -0.00008944, 0.00004828, -0.00007657, 0.00005466, -0.00002222, 0.00003216, -0.00001605, -0.00001523, -0.00001311, 0.00002704, -0.00003466, 0.00002591, -0.00001843, 0.00002473, 0.00001074, -0.00002773, 0.00000350, 0.00001588, -0.00001596, },
		{ 0.00001798, -0.00001812, -0.00001729, 0.00000279, 0.00002373, -0.00000139, -0.00001535, 0.00001794, -0.00003774, 0.00004215, -0.00005261, 0.00002979, 0.00000372, -0.00001935, 0.00001827, 0.00001962, -0.00002587, -0.00001914, 0.00003511, -0.00000964, -0.00001705, 0.00004086, -0.00008223, 0.00008973, -0.00010982, 0.00002454, -0.00005311, 0.00005040, -0.00016686, 0.00016042, -0.00092500, 0.00174943, -0.00038818, -0.00025968, 0.00005156, -0.00009155, -0.00001136, 0.00002232, -0.00005702, 0.00005716, 0.00000077, 0.00000502, 0.00002808, -0.00010204, 0.00009274, -0.00006859, 0.00002943, -0.00006091, 0.00004317, -0.00001077, 0.00002437, -0.00001390, -0.00002375, 0.00000207, 0.00001324, -0.00002375, 0.00001855, -0.00001198, 0.00002333, 0.00000848, -0.00003252, 0.00001566, 0.00000731, -0.00001265, },
		{ 0.00001676, -0.00002523, -0.00001072, 0.00000227, 0.00002840, -0.00001341, -0.00000305, 0.00000386, -0.00002349, 0.00003032, -0.00004622, 0.00003253, 0.00000131, -0.00002473, 0.00003184, 0.00000412, -0.00001473, -0.00002983, 0.00005834, -0.00004698, 0.00003308, -0.00001849, -0.00002144, 0.00002817, -0.00005642, -0.00003590, 0.00002634, -0.00005779, -0.00002467, -0.00005920, -0.00066843, 0.00176406, -0.00066512, -0.00004575, -0.00007051, -0.00000453, -0.00007673, 0.00007939, -0.00010770, 0.00012296, -0.00006328, 0.00007386, -0.00004747, -0.00003925, 0.00004722, -0.00003829, 0.00000324, -0.00003949, 0.00002719, 0.00000510, 0.00001216, -0.00000989, -0.00003124, 0.00001628, -0.00000061, -0.00001220, 0.00001039, -0.00000450, 0.00002087, 0.00000537, -0.00003459, 0.00002651, -0.00000255, -0.00000700, },
		{ 0.00001275, -0.00003028, -0.00000393, 0.00000157, 0.00003171, -0.00002431, 0.00000906, -0.00001080, -0.00000730, 0.00001513, -0.00003516, 0.00003167, -0.00000107, -0.00002759, 0.00004264, -0.00000915, -0.00000680, -0.00003386, 0.00007113, -0.00007230, 0.00007163, -0.00006937, 0.00003487, -0.00003247, -0.00000264, -0.00009464, 0.00010277, -0.00016422, 0.00012044, -0.00030253, -0.00032732, 0.00164137, -0.00083200, 0.00011482, -0.00016423, 0.00006082, -0.00012311, 0.00011805, -0.00013782, 0.00016866, -0.00010909, 0.00012627, -0.00011346, 0.00002142, -0.00000039, -0.00000489, -0.00002545, -0.00001625, 0.00001025, 0.00002228, -0.00000239, -0.00000508, -0.00003668, 0.00002807, -0.00001289, -0.00000153, 0.00000281, 0.00000287, 0.00001797, 0.00000117, -0.00003328, 0.00003466, -0.00001234, 0.00000029, },
		{ 0.00000618, -0.00003255, 0.00000205, 0.00000147, 0.00003276, -0.00003271, 0.00001920, -0.00002393, 0.00000853, -0.00000121, -0.00002105, 0.00002773, -0.00000336, -0.00002762, 0.00004981, -0.00001925, -0.00000268, -0.00003063, 0.00007192, -0.00008222, 0.00009287, -0.00010362, 0.00007676, -0.00008134, 0.00004026, -0.00013941, 0.00016145, -0.00025015, 0.00024322, -0.00053500, 0.00006076, 0.00140370, -0.00088855, 0.00021118, -0.00021916, 0.00009630, -0.00014397, 0.00013304, -0.00014238, 0.00018870, -0.00013022, 0.00015402, -0.00016013, 0.00007040, -0.00004196, 0.00002515, -0.00005165, 0.00000481, -0.00000407, 0.00003748, -0.00001687, -0.00000071, -0.00003921, 0.00003632, -0.00002234, 0.00000704, -0.00000309, 0.00000918, 0.00001511, -0.00000407, -0.00002838, 0.00003912, -0.00002075, 0.00000823, },
		{ -0.00000235, -0.00003172, 0.00000643, 0.00000254, 0.00003093, -0.00003755, 0.00002594, -0.00003372, 0.00002184, -0.00001637, -0.00000582, 0.00002156, -0.00000548, -0.00002493, 0.00005314, -0.00002605, -0.00000190, -0.00002123, 0.00006161, -0.00007654, 0.00009473, -0.00011683, 0.00009780, -0.00011068, 0.00006395, -0.00016108, 0.00019109, -0.00030044, 0.00032156, -0.00072253, 0.00045270, 0.00108541, -0.00084844, 0.00024402, -0.00023343, 0.00010046, -0.00013850, 0.00012438, -0.00012154, 0.00018261, -0.00012537, 0.00015376, -0.00018177, 0.00010101, -0.00007118, 0.00004662, -0.00007127, 0.00002043, -0.00001285, 0.00004780, -0.00002899, 0.00000209, -0.00003827, 0.00004037, -0.00002819, 0.00001277, -0.00000661, 0.00001379, 0.00001255, -0.00000997, -0.00002023, 0.00003939, -0.00002672, 0.00001575, },
		{ -0.00001197, -0.00002793, 0.00000882, 0.00000504, 0.00002603, -0.00003832, 0.00002841, -0.00003887, 0.00003090, -0.00002830, 0.00000853, 0.00001423, -0.00000740, -0.00001994, 0.00005293, -0.00003001, -0.00000314, -0.00000808, 0.00004319, -0.00005805, 0.00007898, -0.00010896, 0.00009603, -0.00011718, 0.00006459, -0.00015538, 0.00018604, -0.00030647, 0.00034053, -0.00083699, 0.00080618, 0.00072719, -0.00073547, 0.00022354, -0.00021268, 0.00007789, -0.00011109, 0.00009683, -0.00008011, 0.00015471, -0.00009830, 0.00012742, -0.00017744, 0.00011044, -0.00008463, 0.00005638, -0.00008181, 0.00002877, -0.00001443, 0.00005132, -0.00003703, 0.00000257, -0.00003378, 0.00004013, -0.00003024, 0.00001542, -0.00000750, 0.00001640, 0.00001031, -0.00001589, -0.00000964, 0.00003554, -0.00002953, 0.00002180, },
		{ -0.00002160, -0.00002172, 0.00000924, 0.00000881, 0.00001837, -0.00003511, 0.00002645, -0.00003889, 0.00003472, -0.00003548, 0.00002026, 0.00000683, -0.00000910, -0.00001327, 0.00004980, -0.00003195, -0.00000469, 0.00000565, 0.00002107, -0.00003181, 0.00005063, -0.00008404, 0.00007409, -0.00010231, 0.00004336, -0.00012373, 0.00014733, -0.00026779, 0.00029509, -0.00086060, 0.00108556, 0.00036989, -0.00057870, 0.00016613, -0.00016775, 0.00003745, -0.00006975, 0.00005851, -0.00002624, 0.00011282, -0.00005674, 0.00008142, -0.00015080, 0.00009986, -0.00008214, 0.00005379, -0.00008276, 0.00002967, -0.00000874, 0.00004745, -0.00004014, 0.00000049, -0.00002614, 0.00003602, -0.00002881, 0.00001524, -0.00000596, 0.00001704, 0.00000828, -0.00002103, 0.00000221, 0.00002819, -0.00002895, 0.00002552, },
		{ -0.00003016, -0.00001396, 0.00000808, 0.00001333, 0.00000871, -0.00002862, 0.00002061, -0.00003407, 0.00003318, -0.00003718, 0.00002810, 0.00000037, -0.00001061, -0.00000557, 0.00004452, -0.00003270, -0.00000494, 0.00001682, 0.00000003, -0.00000395, 0.00001667, -0.00004908, 0.00003830, -0.00007169, 0.00000597, -0.00007269, 0.00008236, -0.00019212, 0.00019102, -0.00078851, 0.00126657, 0.00004891, -0.00040734, 0.00009050, -0.00011179, -0.00001014, -0.00002399, 0.00001884, 0.00003055, 0.00006629, -0.00001054, 0.00002519, -0.00010898, 0.00007385, -0.00006644, 0.00004057, -0.00007545, 0.00002460, 0.00000278, 0.00003694, -0.00003843, -0.00000376, -0.00001621, 0.00002894, -0.00002462, 0.00001289, -0.00000250, 0.00001595, 0.00000625, -0.00002462, 0.00001399, 0.00001835, -0.00002519, 0.00002634, },
		{ -0.00003667, -0.00000566, 0.00000602, 0.00001782, -0.00000184, -0.00002002, 0.00001207, -0.00002546, 0.00002699, -0.00003353, 0.00003140, -0.00000441, -0.00001200, 0.00000257, 0.00003776, -0.00003285, -0.00000285, 0.00002305, -0.00001580, 0.00001965, -0.00001545, -0.00001245, -0.00000286, -0.00003357, -0.00003886, -0.00001248, 0.00000329, -0.00009385, 0.00004391, -0.00062925, 0.00133880, -0.00020995, -0.00024641, 0.00001402, -0.00005732, -0.00005471, 0.00001735, -0.00001346, 0.00008137, 0.00002397, 0.00003051, -0.00003093, -0.00006094, 0.00003916, -0.00004233, 0.00002017, -0.00006262, 0.00001616, 0.00001752, 0.00002173, -0.00003292, -0.00000930, -0.00000522, 0.00002011, -0.00001871, 0.00000923, 0.00000211, 0.00001357, 0.00000408, -0.00002606, 0.00002438, 0.00000729, -0.00001885, 0.00002403, },
		{ -0.00004041, 0.00000214, 0.00000382, 0.00002139, -0.00001201, -0.00001071, 0.00000233, -0.00001463, 0.00001758, -0.00002544, 0.00003023, -0.00000706, -0.00001333, 0.00001067, 0.00003001, -0.00003254, 0.00000175, 0.00002323, -0.00002379, 0.00003461, -0.00003939, 0.00001798, -0.00004070, 0.00000304, -0.00008136, 0.00004526, -0.00007553, 0.00000884, -0.00012369, -0.00040287, 0.00130593, -0.00039269, -0.00011377, -0.00005001, -0.00001396, -0.00008864, 0.00004785, -0.00003209, 0.00011965, -0.00000761, 0.00005866, -0.00007772, -0.00001554, 0.00000311, -0.00001546, -0.00000313, -0.00004760, 0.00000748, 0.00003227, 0.00000448, -0.00002521, -0.00001489, 0.00000545, 0.00001087, -0.00001221, 0.00000525, 0.00000704, 0.00001037, 0.00000171, -0.00002505, 0.00003232, -0.00000367, -0.00001083, 0.00001875, },
		{ -0.00004100, 0.00000858, 0.00000221, 0.00002324, -0.00002061, -0.00000214, -0.00000695, -0.00000340, 0.00000676, -0.00001445, 0.00002527, -0.00000753, -0.00001463, 0.00001840, 0.00002153, -0.00003148, 0.00000815, 0.00001770, -0.00002328, 0.00003882, -0.00005117, 0.00003634, -0.00006804, 0.00003034, -0.00011289, 0.00008996, -0.00014043, 0.00009746, -0.00028618, -0.00013740, 0.00118380, -0.00049718, -0.00001882, -0.00009384, 0.00001290, -0.00010792, 0.00006456, -0.00003410, 0.00014211, -0.00002534, 0.00006959, -0.00010874, 0.00002007, -0.00002784, 0.00000888, -0.00002527, -0.00003353, 0.00000139, 0.00004394, -0.00001201, -0.00001715, -0.00001927, 0.00001448, 0.00000249, -0.00000617, 0.00000183, 0.00001148, 0.00000680, -0.00000077, -0.00002162, 0.00003711, -0.00001337, -0.00000218, 0.00001101, },
		{ -0.00003844, 0.00001310, 0.00000170, 0.00002279, -0.00002668, 0.00000451, -0.00001434, 0.00000652, -0.00000352, -0.00000245, 0.00001769, -0.00000611, -0.00001586, 0.00002550, 0.00001248, -0.00002910, 0.00001492, 0.00000817, -0.00001557, 0.00003279, -0.00004983, 0.00003984, -0.00008060, 0.00004327, -0.00012764, 0.00011417, -0.00018113, 0.00015668, -0.00041927, 0.00013579, 0.00099675, -0.00053135, 0.00003707, -0.00011547, 0.00002228, -0.00011238, 0.00006807, -0.00002003, 0.00014888, -0.00002976, 0.00006300, -0.00012128, 0.00004165, -0.00004924, 0.00002679, -0.00004317, -0.00002261, -0.00000019, 0.00005023, -0.00002532, -0.00001039, -0.00002136, 0.00002080, -0.00000399, -0.00000146, -0.00000033, 0.00001478, 0.00000329, -0.00000316, -0.00001615, 0.00003847, -0.00002088, 0.00000603, 0.00000158, },
		{ -0.00003310, 0.00001549, 0.00000248, 0.00001983, -0.00002970, 0.00000847, -0.00001884, 0.00001378, -0.00001160, 0.00000868, 0.00000897, -0.00000339, -0.00001686, 0.00003173, 0.00000299, -0.00002477, 0.00002029, -0.00000281, -0.00000352, 0.00001926, -0.00003735, 0.00002906, -0.00007760, 0.00004033, -0.00012372, 0.00011494, -0.00019256, 0.00017709, -0.00050396, 0.00038687, 0.00077315, -0.00051016, 0.00005892, -0.00011758, 0.00001687, -0.00010491, 0.00006187, 0.00000677, 0.00014283, -0.00002444, 0.00004222, -0.00011644, 0.00004825, -0.00005925, 0.00003629, -0.00005525, -0.00001572, 0.00000331, 0.00005002, -0.00003391, -0.00000608, -0.00002056, 0.00002381, -0.00000794, 0.00000143, -0.00000081, 0.00001647, 0.00000017, -0.00000520, -0.00000929, 0.00003658, -0.00002568, 0.00001286, -0.00000856, },
		{ -0.00002568, 0.00001587, 0.00000443, 0.00001457, -0.00002961, 0.00000949, -0.00001998, 0.00001761, -0.00001629, 0.00001732, 0.00000058, -0.00000014, -0.00001737, 0.00003682, -0.00000669, -0.00001811, 0.00002263, -0.00001242, 0.00000925, 0.00000249, -0.00001801, 0.00000755, -0.00006152, 0.00002355, -0.00010320, 0.00009408, -0.00017560, 0.00015667, -0.00052933, 0.00059166, 0.00054086, -0.00045188, 0.00005561, -0.00010587, 0.00000158, -0.00009026, 0.00005114, 0.00004128, 0.00012832, -0.00001467, 0.00001310, -0.00009840, 0.00004192, -0.00005854, 0.00003743, -0.00006151, -0.00001238, 0.00001107, 0.00004360, -0.00003732, -0.00000457, -0.00001686, 0.00002347, -0.00000917, 0.00000232, 0.00000049, 0.00001632, -0.00000231, -0.00000664, -0.00000182, 0.00003197, -0.00002760, 0.00001758, -0.00001838, },
		{ -0.00001706, 0.00001470, 0.00000710, 0.00000757, -0.00002682, 0.00000783, -0.00001793, 0.00001791, -0.00001709, 0.00002236, -0.00000616, 0.00000283, -0.00001707, 0.00004042, -0.00001605, -0.00000918, 0.00002089, -0.00001835, 0.00001928, -0.00001286, 0.00000270, -0.00001924, -0.00003730, -0.00000229, -0.00007138, 0.00005749, -0.00013651, 0.00010081, -0.00049377, 0.00073450, 0.00032346, -0.00037467, 0.00003731, -0.00008705, -0.00001799, -0.00007354, 0.00004137, 0.00007817, 0.00010985, -0.00000593, -0.00001755, -0.00007308, 0.00002677, -0.00004971, 0.00003184, -0.00006320, -0.00001102, 0.00002114, 0.00003250, -0.00003617, -0.00000549, -0.00001087, 0.00002026, -0.00000795, 0.00000144, 0.00000336, 0.00001435, -0.00000393, -0.00000732, 0.00000546, 0.00002540, -0.00002681, 0.00001972, -0.00002690, },
		{ -0.00000822, 0.00001262, 0.00000983, -0.00000032, -0.00002209, 0.00000420, -0.00001334, 0.00001520, -0.00001420, 0.00002335, -0.00001033, 0.00000481, -0.00001561, 0.00004211, -0.00002438, 0.00000134, 0.00001489, -0.00001923, 0.00002402, -0.00002282, 0.00001942, -0.00004524, -0.00001096, -0.00003099, -0.00003536, 0.00001358, -0.00008535, 0.00002097, -0.00040460, 0.00080943, 0.00013780, -0.00029385, 0.00001325, -0.00006719, -0.00003704, -0.00005897, 0.00003717, 0.00011291, 0.00009079, -0.00000240, -0.00004360, -0.00004671, 0.00000774, -0.00003639, 0.00002204, -0.00006220, -0.00000963, 0.00003093, 0.00001907, -0.00003191, -0.00000780, -0.00000369, 0.00001513, -0.00000493, -0.00000071, 0.00000733, 0.00001078, -0.00000456, -0.00000721, 0.00001189, 0.00001772, -0.00002378, 0.00001914, -0.00003331, },
		{ -0.00000006, 0.00001032, 0.00001191, -0.00000812, -0.00001638, -0.00000046, -0.00000722, 0.00001048, -0.00000846, 0.00002052, -0.00001148, 0.00000527, -0.00001273, 0.00004151, -0.00003078, 0.00001228, 0.00000543, -0.00001502, 0.00002247, -0.00002497, 0.00002805, -0.00006517, 0.00001177, -0.00005644, -0.00000236, -0.00002873, -0.00003361, -0.00006788, -0.00027602, 0.00081960, -0.00000698, -0.00022030, -0.00000974, -0.00005055, -0.00005249, -0.00004903, 0.00004154, 0.00014244, 0.00007276, -0.00000601, -0.00006089, -0.00002439, -0.00001062, -0.00002215, 0.00001063, -0.00006035, -0.00000635, 0.00003797, 0.00000593, -0.00002641, -0.00001016, 0.00000331, 0.00000930, -0.00000102, -0.00000343, 0.00001175, 0.00000603, -0.00000414, -0.00000640, 0.00001699, 0.00000976, -0.00001914, 0.00001598, -0.00003703, },
		{ 0.00000669, 0.00000843, 0.00001266, -0.00001488, -0.00001069, -0.00000513, -0.00000071, 0.00000495, -0.00000112, 0.00001467, -0.00000969, 0.00000403, -0.00000835, 0.00003837, -0.00003437, 0.00002210, -0.00000589, -0.00000693, 0.00001530, -0.00001894, 0.00002660, -0.00007555, 0.00002671, -0.00007398, 0.00002190, -0.00006192, 0.00000822, -0.00015013, -0.00012626, 0.00077539, -0.00010900, -0.00016003, -0.00002781, -0.00003915, -0.00006323, -0.00004422, 0.00005562, 0.00016524, 0.00005552, -0.00001612, -0.00006793, -0.00000916, -0.00002508, -0.00000974, -0.00000042, -0.00005882, -0.00000010, 0.00004042, -0.00000466, -0.00002148, -0.00001122, 0.00000879, 0.00000405, 0.00000279, -0.00000594, 0.00001589, 0.00000066, -0.00000271, -0.00000512, 0.00002053, 0.00000223, -0.00001360, 0.00001066, -0.00003781, },
		{ 0.00001158, 0.00000737, 0.00001164, -0.00001983, -0.00000588, -0.00000892, 0.00000517, -0.00000015, 0.00000640, 0.00000701, -0.00000549, 0.00000120, -0.00000259, 0.00003269, -0.00003449, 0.00002925, -0.00001704, 0.00000294, 0.00000465, -0.00000637, 0.00001541, -0.00007525, 0.00003192, -0.00008124, 0.00003426, -0.00008138, 0.00003269, -0.00021246, 0.00002573, 0.00069161, -0.00017199, -0.00011480, -0.00003978, -0.00003303, -0.00006991, -0.00004338, 0.00007903, 0.00018094, 0.00003755, -0.00002996, -0.00006583, -0.00000175, -0.00003425, -0.00000057, -0.00001014, -0.00005784, 0.00000908, 0.00003762, -0.00001126, -0.00001843, -0.00001003, 0.00001169, 0.00000046, 0.00000563, -0.00000755, 0.00001902, -0.00000467, -0.00000042, -0.00000363, 0.00002248, -0.00000436, -0.00000789, 0.00000379, -0.00003571, },
		{ 0.00001446, 0.00000730, 0.00000866, -0.00002250, -0.00000257, -0.00001118, 0.00000964, -0.00000386, 0.00001273, -0.00000105, 0.00000023, -0.00000279, 0.00000414, 0.00002479, -0.00003088, 0.00003249, -0.00002598, 0.00001212, -0.00000651, 0.00000958, -0.00000312, -0.00006546, 0.00002782, -0.00007844, 0.00003454, -0.00008608, 0.00003674, -0.00024598, 0.00016335, 0.00058447, -0.00020327, -0.00008333, -0.00004643, -0.00003090, -0.00007426, -0.00004419, 0.00011042, 0.00018957, 0.00001706, -0.00004364, -0.00005764, -0.00000082, -0.00003842, 0.00000528, -0.00001864, -0.00005662, 0.00001999, 0.00003016, -0.00001352, -0.00001777, -0.00000624, 0.00001149, -0.00000075, 0.00000683, -0.00000771, 0.00002054, -0.00000927, 0.00000242, -0.00000216, 0.00002297, -0.00000967, -0.00000266, -0.00000385, -0.00003111, },
		{ 0.00001551, 0.00000809, 0.00000388, -0.00002273, -0.00000103, -0.00001161, 0.00001230, -0.00000559, 0.00001685, -0.00000817, 0.00000641, -0.00000727, 0.00001119, 0.00001537, -0.00002385, 0.00003122, -0.00003120, 0.00001839, -0.00001509, 0.00002504, -0.00002520, -0.00004913, 0.00001676, -0.00006792, 0.00002517, -0.00007840, 0.00002189, -0.00024734, 0.00027444, 0.00046899, -0.00021147, -0.00006271, -0.00004946, -0.00003098, -0.00007838, -0.00004395, 0.00014809, 0.00019085, -0.00000699, -0.00005339, -0.00004724, -0.00000383, -0.00003913, 0.00000873, -0.00002671, -0.00005377, 0.00003057, 0.00001976, -0.00001220, -0.00001919, -0.00000019, 0.00000827, 0.00000066, 0.00000610, -0.00000617, 0.00002005, -0.00001251, 0.00000541, -0.00000083, 0.00002220, -0.00001350, 0.00000153, -0.00001146, -0.00002464, },
		{ 0.00001514, 0.00000937, -0.00000225, -0.00002073, -0.00000125, -0.00001024, 0.00001314, -0.00000521, 0.00001817, -0.00001326, 0.00001201, -0.00001147, 0.00001777, 0.00000544, -0.00001426, 0.00002560, -0.00003202, 0.00002040, -0.00001867, 0.00003637, -0.00004664, -0.00003005, 0.00000218, -0.00005331, 0.00001030, -0.00006318, -0.00000649, -0.00021874, 0.00035244, 0.00035711, -0.00020479, -0.00004974, -0.00005063, -0.00003171, -0.00008402, -0.00004008, 0.00019038, 0.00018368, -0.00003403, -0.00005674, -0.00003813, -0.00000799, -0.00003829, 0.00001103, -0.00003534, -0.00004777, 0.00003847, 0.00000879, -0.00000886, -0.00002161, 0.00000710, 0.00000273, 0.00000442, 0.00000351, -0.00000300, 0.00001742, -0.00001388, 0.00000805, 0.00000033, 0.00002040, -0.00001582, 0.00000427, -0.00001821, -0.00001709, },
		{ 0.00001393, 0.00001057, -0.00000904, -0.00001695, -0.00000293, -0.00000734, 0.00001244, -0.00000305, 0.00001662, -0.00001562, 0.00001613, -0.00001465, 0.00002304, -0.00000383, -0.00000346, 0.00001658, -0.00002876, 0.00001793, -0.00001610, 0.00004093, -0.00006382, -0.00001193, -0.00001243, -0.00003842, -0.00000554, -0.00004624, -0.00004075, -0.00016675, 0.00039624, 0.00025685, -0.00018988, -0.00004171, -0.00005117, -0.00003217, -0.00009226, -0.00003039, 0.00023575, 0.00016625, -0.00006161, -0.00005330, -0.00003241, -0.00001122, -0.00003747, 0.00001317, -0.00004514, -0.00003762, 0.00004167, -0.00000029, -0.00000544, -0.00002357, 0.00001420, -0.00000394, 0.00000979, -0.00000050, 0.00000145, 0.00001284, -0.00001310, 0.00000985, 0.00000145, 0.00001773, -0.00001666, 0.00000530, -0.00002339, -0.00000933, },
		{ 0.00001249, 0.00001108, -0.00001568, -0.00001207, -0.00000562, -0.00000337, 0.00001071, 0.00000028, 0.00001263, -0.00001506, 0.00001818, -0.00001623, 0.00002629, -0.00001124, 0.00000693, 0.00000572, -0.00002262, 0.00001197, -0.00000771, 0.00003769, -0.00007437, 0.00000248, -0.00002435, -0.00002626, -0.00001859, -0.00003282, -0.00007278, -0.00010057, 0.00040915, 0.00017231, -0.00017141, -0.00003677, -0.00005164, -0.00003219, -0.00010349, -0.00001298, 0.00028241, 0.00013652, -0.00008573, -0.00004486, -0.00003037, -0.00001277, -0.00003731, 0.00001535, -0.00005584, -0.00002333, 0.00003907, -0.00000557, -0.00000370, -0.00002360, 0.00001959, -0.00001034, 0.00001578, -0.00000525, 0.00000655, 0.00000680, -0.00001015, 0.00001041, 0.00000272, 0.00001433, -0.00001614, 0.00000452, -0.00002647, -0.00000218, },
		{ 0.00001138, 0.00001034, -0.00002135, -0.00000682, -0.00000878, 0.00000118, 0.00000850, 0.00000398, 0.00000698, -0.00001180, 0.00001793, -0.00001589, 0.00002706, -0.00001585, 0.00001536, -0.00000516, -0.00001538, 0.00000443, 0.00000475, 0.00002728, -0.00007757, 0.00001183, -0.00003224, -0.00001832, -0.00002682, -0.00002632, -0.00009583, -0.00002994, 0.00039731, 0.00010451, -0.00015225, -0.00003390, -0.00005205, -0.00003211, -0.00011755, 0.00001395, 0.00032777, 0.00009316, -0.00010178, -0.00003491, -0.00003049, -0.00001328, -0.00003732, 0.00001685, -0.00006621, -0.00000612, 0.00003088, -0.00000615, -0.00000479, -0.00002065, 0.00002198, -0.00001510, 0.00002128, -0.00000990, 0.00001153, 0.00000005, -0.00000536, 0.00000953, 0.00000430, 0.00001032, -0.00001446, 0.00000202, -0.00002713, 0.00000371, },
		},
		{ { -0.00000186, -0.00000291, -0.00000565, 0.00000146, 0.00000962, -0.00000135, 0.00000219, -0.00000491, 0.00000247, 0.00000598, -0.00001191, 0.00000292, -0.00001916, -0.00000435, 0.00001602, -0.00000370, 0.00000811, 0.00001176, -0.00003467, -0.00000876, 0.00003101, -0.00002022, 0.00003063, -0.00002255, -0.00000660, 0.00000299, -0.00010553, -0.00000991, -0.00014163, 0.00004404, 0.00035687, 0.00001717, -0.00024641, 0.00001717, 0.00035687, 0.00004404, -0.00014163, -0.00000991, -0.00010553, 0.00000299, -0.00000660, -0.00002255, 0.00003063, -0.00002022, 0.00003101, -0.00000876, -0.00003467, 0.00001176, 0.00000811, -0.00000370, 0.00001602, -0.00000435, -0.00001916, 0.00000292, -0.00001191, 0.00000598, 0.00000247, -0.00000491, 0.00000219, -0.00000135, 0.00000962, 0.00000146, -0.00000565, -0.00000291, },
		{ -0.00000202, -0.00000278, -0.00000493, 0.00000182, 0.00001018, -0.00000223, 0.00000217, -0.00000565, 0.00000385, 0.00000523, -0.00001243, 0.00000224, -0.00001936, -0.00000138, 0.00001433, -0.00000443, 0.00001121, 0.00001012, -0.00003952, -0.00000158, 0.00002835, -0.00001969, 0.00003005, -0.00002806, 0.00000416, -0.00000866, -0.00009988, -0.00001567, -0.00013358, 0.00007655, 0.00031635, 0.00000296, -0.00024790, 0.00003577, 0.00039759, 0.00000348, -0.00014303, -0.00000709, -0.00010906, 0.00001379, -0.00001824, -0.00001504, 0.00002944, -0.00001945, 0.00003230, -0.00001533, -0.00002905, 0.00001228, 0.00000531, -0.00000267, 0.00001744, -0.00000750, -0.00001867, 0.00000362, -0.00001135, 0.00000651, 0.00000124, -0.00000428, 0.00000239, -0.00000068, 0.00000903, 0.00000117, -0.00000652, -0.00000271, },
		{ -0.00000246, -0.00000228, -0.00000447, 0.00000237, 0.00001055, -0.00000317, 0.00000226, -0.00000642, 0.00000533, 0.00000428, -0.00001292, 0.00000169, -0.00001940, 0.00000152, 0.00001226, -0.00000463, 0.00001432, 0.00000752, -0.00004350, 0.00000604, 0.00002436, -0.00001774, 0.00002743, -0.00003092, 0.00001295, -0.00001979, -0.00009357, -0.00002277, -0.00012039, 0.00010129, 0.00027698, -0.00000758, -0.00025248, 0.00005981, 0.00043694, -0.00004468, -0.00013679, -0.00000856, -0.00010915, 0.00002240, -0.00002956, -0.00000634, 0.00002691, -0.00001760, 0.00003228, -0.00002112, -0.00002284, 0.00001161, 0.00000309, -0.00000156, 0.00001872, -0.00001091, -0.00001775, 0.00000424, -0.00001070, 0.00000679, 0.00000019, -0.00000377, 0.00000284, -0.00000032, 0.00000854, 0.00000081, -0.00000741, -0.00000227, },
		{ -0.00000313, -0.00000139, -0.00000433, 0.00000321, 0.00001058, -0.00000403, 0.00000235, -0.00000714, 0.00000683, 0.00000317, -0.00001340, 0.00000132, -0.00001938, 0.00000444, 0.00000971, -0.00000408, 0.00001712, 0.00000418, -0.00004652, 0.00001385, 0.00001917, -0.00001439, 0.00002270, -0.00003073, 0.00001889, -0.00002919, -0.00008800, -0.00002947, -0.00010391, 0.00011909, 0.00023920, -0.00001485, -0.00026047, 0.00009062, 0.00047280, -0.00009921, -0.00012260, -0.00001519, -0.00010479, 0.00002763, -0.00003935, 0.00000264, 0.00002356, -0.00001500, 0.00003107, -0.00002598, -0.00001624, 0.00000974, 0.00000162, -0.00000055, 0.00001995, -0.00001466, -0.00001626, 0.00000466, -0.00000992, 0.00000680, -0.00000066, -0.00000340, 0.00000355, -0.00000035, 0.00000827, 0.00000025, -0.00000819, -0.00000171, },
		{ -0.00000391, -0.00000017, -0.00000454, 0.00000441, 0.00001016, -0.00000465, 0.00000232, -0.00000772, 0.00000830, 0.00000193, -0.00001385, 0.00000117, -0.00001939, 0.00000742, 0.00000663, -0.00000264, 0.00001930, 0.00000032, -0.00004854, 0.00002161, 0.00001300, -0.00000983, 0.00001601, -0.00002737, 0.00002142, -0.00003591, -0.00008439, -0.00003415, -0.00008617, 0.00013123, 0.00020300, -0.00001899, -0.00027228, 0.00012975, 0.00050251, -0.00015810, -0.00010094, -0.00002727, -0.00009537, 0.00002855, -0.00004657, 0.00001101, 0.00001996, -0.00001201, 0.00002884, -0.00002981, -0.00000951, 0.00000679, 0.00000101, 0.00000022, 0.00002118, -0.00001879, -0.00001411, 0.00000479, -0.00000899, 0.00000655, -0.00000132, -0.00000313, 0.00000448, -0.00000077, 0.00000829, -0.00000059, -0.00000871, -0.00000116, },
		{ -0.00000470, 0.00000131, -0.00000508, 0.00000600, 0.00000920, -0.00000491, 0.00000205, -0.00000808, 0.00000964, 0.00000061, -0.00001426, 0.00000120, -0.00001943, 0.00001043, 0.00000304, -0.00000022, 0.00002063, -0.00000379, -0.00004953, 0.00002904, 0.00000616, -0.00000436, 0.00000772, -0.00002104, 0.00002036, -0.00003936, -0.00008359, -0.00003547, -0.00006910, 0.00013913, 0.00016811, -0.00002005, -0.00028832, 0.00017874, 0.00052312, -0.00021864, -0.00007301, -0.00004448, -0.00008080, 0.00002467, -0.00005041, 0.00001799, 0.00001666, -0.00000902, 0.00002580, -0.00003251, -0.00000293, 0.00000296, 0.00000125, 0.00000070, 0.00002241, -0.00002326, -0.00001126, 0.00000454, -0.00000789, 0.00000606, -0.00000182, -0.00000290, 0.00000556, -0.00000155, 0.00000862, -0.00000177, -0.00000889, -0.00000076, },
		{ -0.00000533, 0.00000291, -0.00000588, 0.00000793, 0.00000767, -0.00000471, 0.00000145, -0.00000811, 0.00001080, -0.00000078, -0.00001458, 0.00000136, -0.00001947, 0.00001342, -0.00000094, 0.00000316, 0.00002092, -0.00000794, -0.00004947, 0.00003585, -0.00000096, 0.00000159, -0.00000160, -0.00001228, 0.00001591, -0.00003941, -0.00008603, -0.00003252, -0.00005432, 0.00014425, 0.00013416, -0.00001805, -0.00030884, 0.00023898, 0.00053150, -0.00027759, -0.00004075, -0.00006582, -0.00006157, 0.00001594, -0.00005039, 0.00002296, 0.00001411, -0.00000636, 0.00002217, -0.00003401, 0.00000316, -0.00000148, 0.00000223, 0.00000093, 0.00002356, -0.00002798, -0.00000772, 0.00000387, -0.00000660, 0.00000535, -0.00000221, -0.00000264, 0.00000667, -0.00000261, 0.00000923, -0.00000331, -0.00000863, -0.00000062, },
		{ -0.00000568, 0.00000450, -0.00000681, 0.00001013, 0.00000558, -0.00000402, 0.00000047, -0.00000777, 0.00001170, -0.00000221, -0.00001474, 0.00000156, -0.00001942, 0.00001624, -0.00000514, 0.00000739, 0.00002009, -0.00001195, -0.00004836, 0.00004174, -0.00000794, 0.00000748, -0.00001124, -0.00000186, 0.00000867, -0.00003638, -0.00009160, -0.00002493, -0.00004303, 0.00014780, 0.00010087, -0.00001321, -0.00033373, 0.00031147, 0.00052465, -0.00033137, -0.00000666, -0.00008972, -0.00003871, 0.00000282, -0.00004642, 0.00002558, 0.00001264, -0.00000430, 0.00001816, -0.00003425, 0.00000845, -0.00000615, 0.00000375, 0.00000101, 0.00002447, -0.00003275, -0.00000358, 0.00000278, -0.00000516, 0.00000447, -0.00000255, -0.00000225, 0.00000767, -0.00000382, 0.00001003, -0.00000516, -0.00000792, -0.00000085, },
		{ -0.00000562, 0.00000592, -0.00000775, 0.00001245, 0.00000304, -0.00000284, -0.00000093, -0.00000702, 0.00001230, -0.00000366, -0.00001467, 0.00000166, -0.00001914, 0.00001870, -0.00000931, 0.00001224, 0.00001818, -0.00001568, -0.00004619, 0.00004641, -0.00001432, 0.00001276, -0.00002036, 0.00000919, -0.00000041, -0.00003100, -0.00009971, -0.00001297, -0.00003584, 0.00015063, 0.00006818, -0.00000606, -0.00036240, 0.00039660, 0.00049997, -0.00037630, 0.00002637, -0.00011412, -0.00001375, -0.00001375, -0.00003884, 0.00002576, 0.00001240, -0.00000303, 0.00001393, -0.00003320, 0.00001260, -0.00001065, 0.00000554, 0.00000113, 0.00002496, -0.00003734, 0.00000099, 0.00000133, -0.00000361, 0.00000348, -0.00000290, -0.00000164, 0.00000841, -0.00000501, 0.00001090, -0.00000722, -0.00000677, -0.00000150, },
		{ -0.00000506, 0.00000701, -0.00000851, 0.00001472, 0.00000017, -0.00000123, -0.00000272, -0.00000586, 0.00001260, -0.00000514, -0.00001430, 0.00000153, -0.00001847, 0.00002059, -0.00001313, 0.00001739, 0.00001532, -0.00001907, -0.00004296, 0.00004959, -0.00001966, 0.00001684, -0.00002812, 0.00001976, -0.00001014, -0.00002438, -0.00010926, 0.00000251, -0.00003271, 0.00015310, 0.00003636, 0.00000248, -0.00039362, 0.00049394, 0.00045554, -0.00040897, 0.00005530, -0.00013668, 0.00001140, -0.00003240, -0.00002836, 0.00002374, 0.00001334, -0.00000260, 0.00000963, -0.00003086, 0.00001531, -0.00001458, 0.00000728, 0.00000150, 0.00002476, -0.00004144, 0.00000578, -0.00000038, -0.00000202, 0.00000245, -0.00000332, -0.00000071, 0.00000875, -0.00000600, 0.00001167, -0.00000937, -0.00000526, -0.00000258, },
		{ -0.00000394, 0.00000765, -0.00000893, 0.00001674, -0.00000284, 0.00000067, -0.00000481, -0.00000433, 0.00001258, -0.00000663, -0.00001356, 0.00000103, -0.00001724, 0.00002167, -0.00001624, 0.00002244, 0.00001175, -0.00002211, -0.00003865, 0.00005105, -0.00002356, 0.00001922, -0.00003370, 0.00002867, -0.00001919, -0.00001790, -0.00011880, 0.00002012, -0.00003297, 0.00015504, 0.00000614, 0.00001102, -0.00042547, 0.00060210, 0.00039033, -0.00042643, 0.00007719, -0.00015493, 0.00003462, -0.00005146, -0.00001604, 0.00002002, 0.00001523, -0.00000293, 0.00000535, -0.00002724, 0.00001635, -0.00001751, 0.00000862, 0.00000237, 0.00002361, -0.00004472, 0.00001050, -0.00000220, -0.00000047, 0.00000144, -0.00000387, 0.00000061, 0.00000854, -0.00000662, 0.00001218, -0.00001144, -0.00000351, -0.00000405, },
		{ -0.00000226, 0.00000773, -0.00000887, 0.00001829, -0.00000576, 0.00000271, -0.00000710, -0.00000251, 0.00001229, -0.00000815, -0.00001241, 0.00000004, -0.00001530, 0.00002170, -0.00001830, 0.00002696, 0.00000778, -0.00002487, -0.00003327, 0.00005062, -0.00002571, 0.00001948, -0.00003640, 0.00003484, -0.00002621, -0.00001303, -0.00012664, 0.00003805, -0.00003540, 0.00015575, -0.00002137, 0.00001769, -0.00045533, 0.00071865, 0.00030446, -0.00042654, 0.00008947, -0.00016655, 0.00005374, -0.00006907, -0.00000318, 0.00001532, 0.00001766, -0.00000385, 0.00000111, -0.00002240, 0.00001555, -0.00001910, 0.00000923, 0.00000399, 0.00002125, -0.00004684, 0.00001484, -0.00000397, 0.00000092, 0.00000054, -0.00000458, 0.00000235, 0.00000770, -0.00000671, 0.00001224, -0.00001323, -0.00000168, -0.00000584, },
		{ -0.00000006, 0.00000719, -0.00000818, 0.00001919, -0.00000837, 0.00000467, -0.00000942, -0.00000051, 0.00001177, -0.00000968, -0.00001081, -0.00000154, -0.00001252, 0.00002048, -0.00001896, 0.00003050, 0.00000377, -0.00002745, -0.00002685, 0.00004819, -0.00002586, 0.00001733, -0.00003571, 0.00003735, -0.00002995, -0.00001122, -0.00013101, 0.00005423, -0.00003834, 0.00015411, -0.00004470, 0.00002025, -0.00047991, 0.00084010, 0.00019926, -0.00040810, 0.00009018, -0.00016955, 0.00006676, -0.00008337, 0.00000877, 0.00001054, 0.00002008, -0.00000506, -0.00000309, -0.00001643, 0.00001282, -0.00001908, 0.00000884, 0.00000654, 0.00001748, -0.00004747, 0.00001850, -0.00000547, 0.00000204, -0.00000021, -0.00000547, 0.00000452, 0.00000615, -0.00000615, 0.00001171, -0.00001456, 0.00000006, -0.00000780, },
		{ 0.00000255, 0.00000605, -0.00000680, 0.00001925, -0.00001042, 0.00000636, -0.00001159, 0.00000153, 0.00001110, -0.00001124, -0.00000877, -0.00000375, -0.00000883, 0.00001789, -0.00001797, 0.00003262, 0.00000013, -0.00002996, -0.00001942, 0.00004374, -0.00002393, 0.00001267, -0.00003137, 0.00003557, -0.00002942, -0.00001375, -0.00013025, 0.00006656, -0.00003989, 0.00014870, -0.00006214, 0.00001618, -0.00049542, 0.00096198, 0.00007738, -0.00037111, 0.00007815, -0.00016250, 0.00007203, -0.00009266, 0.00001838, 0.00000664, 0.00002189, -0.00000621, -0.00000727, -0.00000947, 0.00000820, -0.00001729, 0.00000725, 0.00001016, 0.00001215, -0.00004636, 0.00002116, -0.00000652, 0.00000278, -0.00000073, -0.00000653, 0.00000708, 0.00000391, -0.00000487, 0.00001046, -0.00001525, 0.00000150, -0.00000977, },
		{ 0.00000545, 0.00000434, -0.00000470, 0.00001838, -0.00001174, 0.00000756, -0.00001342, 0.00000346, 0.00001035, -0.00001280, -0.00000632, -0.00000658, -0.00000422, 0.00001388, -0.00001514, 0.00003297, -0.00000277, -0.00003256, -0.00001108, 0.00003733, -0.00001994, 0.00000560, -0.00002338, 0.00002921, -0.00002397, -0.00002153, -0.00012300, 0.00007308, -0.00003809, 0.00013798, -0.00007193, 0.00000291, -0.00049777, 0.00107902, -0.00005733, -0.00031676, 0.00005316, -0.00014466, 0.00006839, -0.00009554, 0.00002434, 0.00000456, 0.00002244, -0.00000692, -0.00001149, -0.00000170, 0.00000182, -0.00001367, 0.00000435, 0.00001488, 0.00000522, -0.00004332, 0.00002258, -0.00000694, 0.00000306, -0.00000100, -0.00000774, 0.00000995, 0.00000103, -0.00000287, 0.00000843, -0.00001519, 0.00000248, -0.00001156, },
		{ 0.00000845, 0.00000218, -0.00000193, 0.00001650, -0.00001215, 0.00000807, -0.00001474, 0.00000513, 0.00000961, -0.00001437, -0.00000354, -0.00000996, 0.00000123, 0.00000850, -0.00001041, 0.00003130, -0.00000460, -0.00003535, -0.00000194, 0.00002914, -0.00001404, -0.00000362, -0.00001205, 0.00001833, -0.00001336, -0.00003507, -0.00010836, 0.00007219, -0.00003120, 0.00012051, -0.00007245, -0.00002199, -0.00048285, 0.00118539, -0.00019992, -0.00024741, 0.00001601, -0.00011611, 0.00005535, -0.00009110, 0.00002560, 0.00000513, 0.00002116, -0.00000682, -0.00001580, 0.00000663, -0.00000608, -0.00000833, 0.00000018, 0.00002063, -0.00000324, -0.00003828, 0.00002256, -0.00000658, 0.00000278, -0.00000099, -0.00000904, 0.00001301, -0.00000236, -0.00000021, 0.00000562, -0.00001428, 0.00000285, -0.00001300, },
		{ 0.00001136, -0.00000028, 0.00000142, 0.00001363, -0.00001156, 0.00000777, -0.00001537, 0.00000638, 0.00000897, -0.00001591, -0.00000054, -0.00001374, 0.00000735, 0.00000192, -0.00000387, 0.00002748, -0.00000512, -0.00003842, 0.00000782, 0.00001943, -0.00000655, -0.00001451, 0.00000204, 0.00000341, 0.00000216, -0.00005433, -0.00008600, 0.00006286, -0.00001785, 0.00009519, -0.00006244, -0.00006049, -0.00044681, 0.00127504, -0.00034459, -0.00016656, -0.00003152, -0.00007781, 0.00003317, -0.00007896, 0.00002148, 0.00000895, 0.00001760, -0.00000558, -0.00002022, 0.00001526, -0.00001512, -0.00000149, -0.00000512, 0.00002722, -0.00001303, -0.00003129, 0.00002102, -0.00000535, 0.00000190, -0.00000071, -0.00001036, 0.00001610, -0.00000611, 0.00000299, 0.00000211, -0.00001251, 0.00000248, -0.00001389, },
		{ 0.00001398, -0.00000284, 0.00000517, 0.00000986, -0.00000995, 0.00000656, -0.00001521, 0.00000712, 0.00000850, -0.00001738, 0.00000255, -0.00001776, 0.00001392, -0.00000559, 0.00000427, 0.00002152, -0.00000417, -0.00004178, 0.00001794, 0.00000856, 0.00000213, -0.00002647, 0.00001806, -0.00001469, 0.00002187, -0.00007870, -0.00005623, 0.00004472, 0.00000274, 0.00006138, -0.00004122, -0.00011392, -0.00038643, 0.00134207, -0.00048504, -0.00007855, -0.00008673, -0.00003151, 0.00000282, -0.00005937, 0.00001176, 0.00001634, 0.00001147, -0.00000299, -0.00002475, 0.00002385, -0.00002488, 0.00000651, -0.00001129, 0.00003435, -0.00002381, -0.00002253, 0.00001795, -0.00000322, 0.00000043, -0.00000017, -0.00001160, 0.00001906, -0.00000998, 0.00000654, -0.00000197, -0.00000993, 0.00000132, -0.00001410, },
		{ 0.00001611, -0.00000530, 0.00000912, 0.00000536, -0.00000739, 0.00000442, -0.00001419, 0.00000724, 0.00000825, -0.00001875, 0.00000557, -0.00002177, 0.00002062, -0.00001366, 0.00001365, 0.00001362, -0.00000175, -0.00004540, 0.00002815, -0.00000305, 0.00001151, -0.00003875, 0.00003497, -0.00003481, 0.00004463, -0.00010705, -0.00002007, 0.00001817, 0.00003063, 0.00001915, -0.00000884, -0.00018268, -0.00029944, 0.00138111, -0.00061479, 0.00001165, -0.00014622, 0.00002027, -0.00003399, -0.00003325, -0.00000333, 0.00002730, 0.00000271, 0.00000106, -0.00002934, 0.00003207, -0.00003483, 0.00001521, -0.00001799, 0.00004162, -0.00003515, -0.00001233, 0.00001347, -0.00000023, -0.00000162, 0.00000057, -0.00001267, 0.00002169, -0.00001375, 0.00001021, -0.00000642, -0.00000666, -0.00000061, -0.00001352, },
		{ 0.00001760, -0.00000744, 0.00001302, 0.00000034, -0.00000401, 0.00000141, -0.00001231, 0.00000671, 0.00000827, -0.00001997, 0.00000835, -0.00002551, 0.00002710, -0.00002183, 0.00002378, 0.00000411, 0.00000205, -0.00004916, 0.00003810, -0.00001490, 0.00002103, -0.00005056, 0.00005162, -0.00005555, 0.00006893, -0.00013775, 0.00002083, -0.00001562, 0.00006513, -0.00003069, 0.00003375, -0.00026609, -0.00018475, 0.00138771, -0.00072763, 0.00009878, -0.00020607, 0.00007447, -0.00007496, -0.00000206, -0.00002302, 0.00004145, -0.00000848, 0.00000652, -0.00003387, 0.00003953, -0.00004443, 0.00002409, -0.00002475, 0.00004856, -0.00004649, -0.00000111, 0.00000780, 0.00000351, -0.00000415, 0.00000146, -0.00001348, 0.00002383, -0.00001716, 0.00001372, -0.00001099, -0.00000288, -0.00000323, -0.00001210, },
		{ 0.00001831, -0.00000906, 0.00001660, -0.00000491, -0.00000000, -0.00000233, -0.00000964, 0.00000554, 0.00000855, -0.00002097, 0.00001071, -0.00002871, 0.00003298, -0.00002961, 0.00003409, -0.00000655, 0.00000697, -0.00005286, 0.00004741, -0.00002649, 0.00003012, -0.00006112, 0.00006681, -0.00007538, 0.00009301, -0.00016879, 0.00006427, -0.00005475, 0.00010473, -0.00008653, 0.00008470, -0.00036228, -0.00004276, 0.00135868, -0.00081797, 0.00017764, -0.00026214, 0.00012766, -0.00011733, 0.00003219, -0.00004613, 0.00005810, -0.00002167, 0.00001318, -0.00003819, 0.00004589, -0.00005311, 0.00003261, -0.00003112, 0.00005466, -0.00005725, 0.00001059, 0.00000127, 0.00000782, -0.00000705, 0.00000240, -0.00001393, 0.00002529, -0.00001996, 0.00001680, -0.00001540, 0.00000117, -0.00000642, -0.00000985, },
		{ 0.00001820, -0.00000998, 0.00001960, -0.00001008, 0.00000435, -0.00000661, -0.00000630, 0.00000379, 0.00000908, -0.00002171, 0.00001250, -0.00003111, 0.00003788, -0.00003647, 0.00004393, -0.00001776, 0.00001269, -0.00005624, 0.00005569, -0.00003728, 0.00003821, -0.00006966, 0.00007939, -0.00009273, 0.00011503, -0.00019796, 0.00010763, -0.00009672, 0.00014723, -0.00014597, 0.00014121, -0.00046813, 0.00012458, 0.00129242, -0.00088125, 0.00024346, -0.00031042, 0.00017633, -0.00015806, 0.00006716, -0.00007110, 0.00007623, -0.00003619, 0.00002071, -0.00004209, 0.00005079, -0.00006033, 0.00004018, -0.00003661, 0.00005941, -0.00006678, 0.00002218, -0.00000574, 0.00001246, -0.00001016, 0.00000332, -0.00001395, 0.00002597, -0.00002194, 0.00001920, -0.00001935, 0.00000522, -0.00000997, -0.00000685, },
		{ 0.00001724, -0.00001008, 0.00002178, -0.00001487, 0.00000875, -0.00001118, -0.00000247, 0.00000158, 0.00000980, -0.00002212, 0.00001358, -0.00003248, 0.00004146, -0.00004190, 0.00005263, -0.00002887, 0.00001876, -0.00005903, 0.00006254, -0.00004677, 0.00004479, -0.00007556, 0.00008837, -0.00010617, 0.00013314, -0.00022298, 0.00014811, -0.00013858, 0.00018986, -0.00020602, 0.00019972, -0.00057938, 0.00031368, 0.00118906, -0.00091423, 0.00029223, -0.00034724, 0.00021710, -0.00019413, 0.00010032, -0.00009611, 0.00009461, -0.00005120, 0.00002861, -0.00004531, 0.00005394, -0.00006563, 0.00004631, -0.00004076, 0.00006234, -0.00007448, 0.00003303, -0.00001277, 0.00001715, -0.00001332, 0.00000412, -0.00001349, 0.00002576, -0.00002294, 0.00002068, -0.00002257, 0.00000898, -0.00001363, -0.00000324, },
		{ 0.00001550, -0.00000930, 0.00002295, -0.00001896, 0.00001287, -0.00001574, 0.00000159, -0.00000094, 0.00001065, -0.00002216, 0.00001384, -0.00003263, 0.00004341, -0.00004546, 0.00005956, -0.00003918, 0.00002470, -0.00006089, 0.00006759, -0.00005450, 0.00004946, -0.00007834, 0.00009294, -0.00011447, 0.00014570, -0.00024170, 0.00018286, -0.00017719, 0.00022950, -0.00026318, 0.00025603, -0.00069076, 0.00051941, 0.00105057, -0.00091522, 0.00032098, -0.00036963, 0.00024705, -0.00022267, 0.00012912, -0.00011923, 0.00011185, -0.00006571, 0.00003633, -0.00004759, 0.00005508, -0.00006860, 0.00005054, -0.00004321, 0.00006306, -0.00007980, 0.00004253, -0.00001936, 0.00002157, -0.00001632, 0.00000471, -0.00001253, 0.00002464, -0.00002285, 0.00002105, -0.00002481, 0.00001217, -0.00001715, 0.00000079, },
		{ 0.00001308, -0.00000764, 0.00002297, -0.00002210, 0.00001641, -0.00001998, 0.00000564, -0.00000358, 0.00001153, -0.00002180, 0.00001325, -0.00003146, 0.00004355, -0.00004679, 0.00006415, -0.00004805, 0.00003000, -0.00006152, 0.00007050, -0.00006008, 0.00005190, -0.00007771, 0.00009259, -0.00011671, 0.00015141, -0.00025228, 0.00020929, -0.00020943, 0.00026288, -0.00031378, 0.00030555, -0.00079628, 0.00073526, 0.00088074, -0.00088423, 0.00032797, -0.00037551, 0.00026388, -0.00024125, 0.00015122, -0.00013856, 0.00012655, -0.00007868, 0.00004323, -0.00004867, 0.00005406, -0.00006897, 0.00005254, -0.00004368, 0.00006129, -0.00008230, 0.00005013, -0.00002505, 0.00002542, -0.00001898, 0.00000501, -0.00001107, 0.00002262, -0.00002163, 0.00002022, -0.00002589, 0.00001453, -0.00002025, 0.00000502, },
		{ 0.00001016, -0.00000515, 0.00002178, -0.00002407, 0.00001907, -0.00002360, 0.00000939, -0.00000614, 0.00001235, -0.00002101, 0.00001177, -0.00002894, 0.00004177, -0.00004567, 0.00006595, -0.00005485, 0.00003418, -0.00006063, 0.00007104, -0.00006324, 0.00005194, -0.00007362, 0.00008714, -0.00011239, 0.00014939, -0.00025335, 0.00022521, -0.00023249, 0.00028690, -0.00035420, 0.00034368, -0.00088952, 0.00095368, 0.00068503, -0.00082297, 0.00031287, -0.00036386, 0.00026615, -0.00024808, 0.00016467, -0.00015241, 0.00013739, -0.00008909, 0.00004870, -0.00004831, 0.00005080, -0.00006660, 0.00005211, -0.00004202, 0.00005691, -0.00008168, 0.00005537, -0.00002942, 0.00002840, -0.00002110, 0.00000495, -0.00000916, 0.00001980, -0.00001931, 0.00001816, -0.00002567, 0.00001587, -0.00002267, 0.00000919, },
		{ 0.00000695, -0.00000198, 0.00001940, -0.00002475, 0.00002064, -0.00002633, 0.00001259, -0.00000840, 0.00001298, -0.00001980, 0.00000948, -0.00002510, 0.00003809, -0.00004202, 0.00006466, -0.00005910, 0.00003683, -0.00005803, 0.00006905, -0.00006381, 0.00004955, -0.00006623, 0.00007674, -0.00010144, 0.00013929, -0.00024416, 0.00022908, -0.00024405, 0.00029886, -0.00038117, 0.00036606, -0.00096406, 0.00116637, 0.00047032, -0.00073475, 0.00027672, -0.00033483, 0.00025338, -0.00024213, 0.00016805, -0.00015940, 0.00014324, -0.00009605, 0.00005215, -0.00004633, 0.00004534, -0.00006148, 0.00004917, -0.00003825, 0.00004995, -0.00007780, 0.00005791, -0.00003213, 0.00003026, -0.00002253, 0.00000449, -0.00000686, 0.00001631, -0.00001603, 0.00001491, -0.00002414, 0.00001603, -0.00002419, 0.00001303, },
		{ 0.00000368, 0.00000169, 0.00001592, -0.00002408, 0.00002095, -0.00002795, 0.00001498, -0.00001017, 0.00001334, -0.00001820, 0.00000645, -0.00002009, 0.00003266, -0.00003591, 0.00006016, -0.00006047, 0.00003763, -0.00005358, 0.00006448, -0.00006173, 0.00004484, -0.00005592, 0.00006188, -0.00008425, 0.00012132, -0.00022461, 0.00022008, -0.00024254, 0.00029673, -0.00039206, 0.00036897, -0.00101386, 0.00136481, 0.00024451, -0.00062431, 0.00022189, -0.00028970, 0.00022609, -0.00022321, 0.00016059, -0.00015860, 0.00014326, -0.00009883, 0.00005311, -0.00004263, 0.00003782, -0.00005377, 0.00004383, -0.00003250, 0.00004061, -0.00007073, 0.00005758, -0.00003294, 0.00003082, -0.00002316, 0.00000360, -0.00000428, 0.00001232, -0.00001196, 0.00001063, -0.00002134, 0.00001497, -0.00002463, 0.00001632, },
		{ 0.00000058, 0.00000563, 0.00001152, -0.00002211, 0.00001993, -0.00002831, 0.00001639, -0.00001127, 0.00001334, -0.00001624, 0.00000284, -0.00001412, 0.00002572, -0.00002758, 0.00005252, -0.00005876, 0.00003638, -0.00004726, 0.00005740, -0.00005709, 0.00003807, -0.00004325, 0.00004337, -0.00006166, 0.00009626, -0.00019532, 0.00019828, -0.00022727, 0.00027933, -0.00038514, 0.00034962, -0.00103369, 0.00154068, 0.00001616, -0.00049749, 0.00015192, -0.00023083, 0.00018581, -0.00019207, 0.00014226, -0.00014961, 0.00013701, -0.00009695, 0.00005125, -0.00003718, 0.00002849, -0.00004375, 0.00003630, -0.00002507, 0.00002927, -0.00006071, 0.00005438, -0.00003174, 0.00002996, -0.00002290, 0.00000228, -0.00000152, 0.00000805, -0.00000734, 0.00000553, -0.00001741, 0.00001271, -0.00002391, 0.00001883, },
		{ -0.00000210, 0.00000958, 0.00000642, -0.00001897, 0.00001760, -0.00002733, 0.00001669, -0.00001157, 0.00001290, -0.00001399, -0.00000117, -0.00000748, 0.00001764, -0.00001741, 0.00004201, -0.00005401, 0.00003304, -0.00003918, 0.00004803, -0.00005011, 0.00002958, -0.00002895, 0.00002224, -0.00003487, 0.00006540, -0.00015762, 0.00016458, -0.00019849, 0.00024646, -0.00035976, 0.00030644, -0.00101952, 0.00168637, -0.00020605, -0.00036086, 0.00007122, -0.00016146, 0.00013494, -0.00015027, 0.00011379, -0.00013260, 0.00012444, -0.00009026, 0.00004641, -0.00003009, 0.00001770, -0.00003184, 0.00002696, -0.00001636, 0.00001644, -0.00004815, 0.00004846, -0.00002853, 0.00002765, -0.00002174, 0.00000057, 0.00000130, 0.00000374, -0.00000245, -0.00000011, -0.00001258, 0.00000936, -0.00002199, 0.00002041, },
		{ -0.00000417, 0.00001327, 0.00000092, -0.00001486, 0.00001407, -0.00002504, 0.00001582, -0.00001100, 0.00001200, -0.00001152, -0.00000538, -0.00000048, 0.00000885, -0.00000594, 0.00002906, -0.00004642, 0.00002771, -0.00002957, 0.00003669, -0.00004110, 0.00001984, -0.00001381, -0.00000029, -0.00000536, 0.00003046, -0.00011343, 0.00012073, -0.00015744, 0.00019900, -0.00031644, 0.00023925, -0.00096883, 0.00179542, -0.00041376, -0.00022134, -0.00001517, -0.00008551, 0.00007661, -0.00010017, 0.00007658, -0.00010832, 0.00010597, -0.00007889, 0.00003865, -0.00002154, 0.00000589, -0.00001853, 0.00001625, -0.00000688, 0.00000274, -0.00003364, 0.00004017, -0.00002345, 0.00002396, -0.00001971, -0.00000149, 0.00000403, -0.00000038, 0.00000239, -0.00000598, -0.00000713, 0.00000512, -0.00001894, 0.00002095, },
		{ -0.00000549, 0.00001644, -0.00000467, -0.00001007, 0.00000954, -0.00002153, 0.00001382, -0.00000952, 0.00001062, -0.00000894, -0.00000954, 0.00000651, -0.00000015, 0.00000623, 0.00001428, -0.00003637, 0.00002063, -0.00001875, 0.00002381, -0.00003048, 0.00000934, 0.00000130, -0.00002291, 0.00002517, -0.00000655, -0.00006515, 0.00006918, -0.00010624, 0.00013887, -0.00025694, 0.00014937, -0.00088086, 0.00186291, -0.00059944, -0.00008575, -0.00010201, -0.00000726, 0.00001441, -0.00004466, 0.00003266, -0.00007803, 0.00008242, -0.00006332, 0.00002821, -0.00001184, -0.00000644, -0.00000440, 0.00000469, 0.00000283, -0.00001111, -0.00001787, 0.00002996, -0.00001677, 0.00001904, -0.00001691, -0.00000382, 0.00000656, -0.00000409, 0.00000690, -0.00001172, -0.00000139, 0.00000026, -0.00001488, 0.00002043, },
		{ -0.00000593, 0.00001889, -0.00001004, -0.00000492, 0.00000428, -0.00001701, 0.00001079, -0.00000719, 0.00000879, -0.00000633, -0.00001346, 0.00001313, -0.00000886, 0.00001843, -0.00000163, -0.00002438, 0.00001218, -0.00000716, 0.00000991, -0.00001875, -0.00000139, 0.00001556, -0.00004432, 0.00005499, -0.00004344, -0.00001550, 0.00001293, -0.00004780, 0.00006891, -0.00018416, 0.00003963, -0.00075673, 0.00188575, -0.00075673, 0.00003963, -0.00018416, 0.00006891, -0.00004780, 0.00001293, -0.00001550, -0.00004344, 0.00005499, -0.00004432, 0.00001556, -0.00000139, -0.00001875, 0.00000991, -0.00000716, 0.00001218, -0.00002438, -0.00000163, 0.00001843, -0.00000886, 0.00001313, -0.00001346, -0.00000633, 0.00000879, -0.00000719, 0.00001079, -0.00001701, 0.00000428, -0.00000492, -0.00001004, 0.00001889, },
		{ -0.00000549, 0.00002043, -0.00001488, 0.00000026, -0.00000139, -0.00001172, 0.00000690, -0.00000409, 0.00000656, -0.00000382, -0.00001691, 0.00001904, -0.00001677, 0.00002996, -0.00001787, -0.00001111, 0.00000283, 0.00000469, -0.00000440, -0.00000644, -0.00001184, 0.00002821, -0.00006332, 0.00008242, -0.00007803, 0.00003266, -0.00004466, 0.00001441, -0.00000726, -0.00010201, -0.00008575, -0.00059944, 0.00186291, -0.00088086, 0.00014937, -0.00025694, 0.00013887, -0.00010624, 0.00006918, -0.00006515, -0.00000655, 0.00002517, -0.00002291, 0.00000130, 0.00000934, -0.00003048, 0.00002381, -0.00001875, 0.00002063, -0.00003637, 0.00001428, 0.00000623, -0.00000015, 0.00000651, -0.00000954, -0.00000894, 0.00001062, -0.00000952, 0.00001382, -0.00002153, 0.00000954, -0.00001007, -0.00000467, 0.00001644, },
		{ -0.00000417, 0.00002095, -0.00001894, 0.00000512, -0.00000713, -0.00000598, 0.00000239, -0.00000038, 0.00000403, -0.00000149, -0.00001971, 0.00002396, -0.00002345, 0.00004017, -0.00003364, 0.00000274, -0.00000688, 0.00001625, -0.00001853, 0.00000589, -0.00002154, 0.00003865, -0.00007889, 0.00010597, -0.00010832, 0.00007658, -0.00010017, 0.00007661, -0.00008551, -0.00001517, -0.00022134, -0.00041376, 0.00179542, -0.00096883, 0.00023925, -0.00031644, 0.00019900, -0.00015744, 0.00012073, -0.00011343, 0.00003046, -0.00000536, -0.00000029, -0.00001381, 0.00001984, -0.00004110, 0.00003669, -0.00002957, 0.00002771, -0.00004642, 0.00002906, -0.00000594, 0.00000885, -0.00000048, -0.00000538, -0.00001152, 0.00001200, -0.00001100, 0.00001582, -0.00002504, 0.00001407, -0.00001486, 0.00000092, 0.00001327, },
		{ -0.00000210, 0.00002041, -0.00002199, 0.00000936, -0.00001258, -0.00000011, -0.00000245, 0.00000374, 0.00000130, 0.00000057, -0.00002174, 0.00002765, -0.00002853, 0.00004846, -0.00004815, 0.00001644, -0.00001636, 0.00002696, -0.00003184, 0.00001770, -0.00003009, 0.00004641, -0.00009026, 0.00012444, -0.00013260, 0.00011379, -0.00015027, 0.00013494, -0.00016146, 0.00007122, -0.00036086, -0.00020605, 0.00168637, -0.00101952, 0.00030644, -0.00035976, 0.00024646, -0.00019849, 0.00016458, -0.00015762, 0.00006540, -0.00003487, 0.00002224, -0.00002895, 0.00002958, -0.00005011, 0.00004803, -0.00003918, 0.00003304, -0.00005401, 0.00004201, -0.00001741, 0.00001764, -0.00000748, -0.00000117, -0.00001399, 0.00001290, -0.00001157, 0.00001669, -0.00002733, 0.00001760, -0.00001897, 0.00000642, 0.00000958, },
		{ 0.00000058, 0.00001883, -0.00002391, 0.00001271, -0.00001741, 0.00000553, -0.00000734, 0.00000805, -0.00000152, 0.00000228, -0.00002290, 0.00002996, -0.00003174, 0.00005438, -0.00006071, 0.00002927, -0.00002507, 0.00003630, -0.00004375, 0.00002849, -0.00003718, 0.00005125, -0.00009695, 0.00013701, -0.00014961, 0.00014226, -0.00019207, 0.00018581, -0.00023083, 0.00015192, -0.00049749, 0.00001616, 0.00154068, -0.00103369, 0.00034962, -0.00038514, 0.00027933, -0.00022727, 0.00019828, -0.00019532, 0.00009626, -0.00006166, 0.00004337, -0.00004325, 0.00003807, -0.00005709, 0.00005740, -0.00004726, 0.00003638, -0.00005876, 0.00005252, -0.00002758, 0.00002572, -0.00001412, 0.00000284, -0.00001624, 0.00001334, -0.00001127, 0.00001639, -0.00002831, 0.00001993, -0.00002211, 0.00001152, 0.00000563, },
		{ 0.00000368, 0.00001632, -0.00002463, 0.00001497, -0.00002134, 0.00001063, -0.00001196, 0.00001232, -0.00000428, 0.00000360, -0.00002316, 0.00003082, -0.00003294, 0.00005758, -0.00007073, 0.00004061, -0.00003250, 0.00004383, -0.00005377, 0.00003782, -0.00004263, 0.00005311, -0.00009883, 0.00014326, -0.00015860, 0.00016059, -0.00022321, 0.00022609, -0.00028970, 0.00022189, -0.00062431, 0.00024451, 0.00136481, -0.00101386, 0.00036897, -0.00039206, 0.00029673, -0.00024254, 0.00022008, -0.00022461, 0.00012132, -0.00008425, 0.00006188, -0.00005592, 0.00004484, -0.00006173, 0.00006448, -0.00005358, 0.00003763, -0.00006047, 0.00006016, -0.00003591, 0.00003266, -0.00002009, 0.00000645, -0.00001820, 0.00001334, -0.00001017, 0.00001498, -0.00002795, 0.00002095, -0.00002408, 0.00001592, 0.00000169, },
		{ 0.00000695, 0.00001303, -0.00002419, 0.00001603, -0.00002414, 0.00001491, -0.00001603, 0.00001631, -0.00000686, 0.00000449, -0.00002253, 0.00003026, -0.00003213, 0.00005791, -0.00007780, 0.00004995, -0.00003825, 0.00004917, -0.00006148, 0.00004534, -0.00004633, 0.00005215, -0.00009605, 0.00014324, -0.00015940, 0.00016805, -0.00024213, 0.00025338, -0.00033483, 0.00027672, -0.00073475, 0.00047032, 0.00116637, -0.00096406, 0.00036606, -0.00038117, 0.00029886, -0.00024405, 0.00022908, -0.00024416, 0.00013929, -0.00010144, 0.00007674, -0.00006623, 0.00004955, -0.00006381, 0.00006905, -0.00005803, 0.00003683, -0.00005910, 0.00006466, -0.00004202, 0.00003809, -0.00002510, 0.00000948, -0.00001980, 0.00001298, -0.00000840, 0.00001259, -0.00002633, 0.00002064, -0.00002475, 0.00001940, -0.00000198, },
		{ 0.00001016, 0.00000919, -0.00002267, 0.00001587, -0.00002567, 0.00001816, -0.00001931, 0.00001980, -0.00000916, 0.00000495, -0.00002110, 0.00002840, -0.00002942, 0.00005537, -0.00008168, 0.00005691, -0.00004202, 0.00005211, -0.00006660, 0.00005080, -0.00004831, 0.00004870, -0.00008909, 0.00013739, -0.00015241, 0.00016467, -0.00024808, 0.00026615, -0.00036386, 0.00031287, -0.00082297, 0.00068503, 0.00095368, -0.00088952, 0.00034368, -0.00035420, 0.00028690, -0.00023249, 0.00022521, -0.00025335, 0.00014939, -0.00011239, 0.00008714, -0.00007362, 0.00005194, -0.00006324, 0.00007104, -0.00006063, 0.00003418, -0.00005485, 0.00006595, -0.00004567, 0.00004177, -0.00002894, 0.00001177, -0.00002101, 0.00001235, -0.00000614, 0.00000939, -0.00002360, 0.00001907, -0.00002407, 0.00002178, -0.00000515, },
		{ 0.00001308, 0.00000502, -0.00002025, 0.00001453, -0.00002589, 0.00002022, -0.00002163, 0.00002262, -0.00001107, 0.00000501, -0.00001898, 0.00002542, -0.00002505, 0.00005013, -0.00008230, 0.00006129, -0.00004368, 0.00005254, -0.00006897, 0.00005406, -0.00004867, 0.00004323, -0.00007868, 0.00012655, -0.00013856, 0.00015122, -0.00024125, 0.00026388, -0.00037551, 0.00032797, -0.00088423, 0.00088074, 0.00073526, -0.00079628, 0.00030555, -0.00031378, 0.00026288, -0.00020943, 0.00020929, -0.00025228, 0.00015141, -0.00011671, 0.00009259, -0.00007771, 0.00005190, -0.00006008, 0.00007050, -0.00006152, 0.00003000, -0.00004805, 0.00006415, -0.00004679, 0.00004355, -0.00003146, 0.00001325, -0.00002180, 0.00001153, -0.00000358, 0.00000564, -0.00001998, 0.00001641, -0.00002210, 0.00002297, -0.00000764, },
		{ 0.00001550, 0.00000079, -0.00001715, 0.00001217, -0.00002481, 0.00002105, -0.00002285, 0.00002464, -0.00001253, 0.00000471, -0.00001632, 0.00002157, -0.00001936, 0.00004253, -0.00007980, 0.00006306, -0.00004321, 0.00005054, -0.00006860, 0.00005508, -0.00004759, 0.00003633, -0.00006571, 0.00011185, -0.00011923, 0.00012912, -0.00022267, 0.00024705, -0.00036963, 0.00032098, -0.00091522, 0.00105057, 0.00051941, -0.00069076, 0.00025603, -0.00026318, 0.00022950, -0.00017719, 0.00018286, -0.00024170, 0.00014570, -0.00011447, 0.00009294, -0.00007834, 0.00004946, -0.00005450, 0.00006759, -0.00006089, 0.00002470, -0.00003918, 0.00005956, -0.00004546, 0.00004341, -0.00003263, 0.00001384, -0.00002216, 0.00001065, -0.00000094, 0.00000159, -0.00001574, 0.00001287, -0.00001896, 0.00002295, -0.00000930, },
		{ 0.00001724, -0.00000324, -0.00001363, 0.00000898, -0.00002257, 0.00002068, -0.00002294, 0.00002576, -0.00001349, 0.00000412, -0.00001332, 0.00001715, -0.00001277, 0.00003303, -0.00007448, 0.00006234, -0.00004076, 0.00004631, -0.00006563, 0.00005394, -0.00004531, 0.00002861, -0.00005120, 0.00009461, -0.00009611, 0.00010032, -0.00019413, 0.00021710, -0.00034724, 0.00029223, -0.00091423, 0.00118906, 0.00031368, -0.00057938, 0.00019972, -0.00020602, 0.00018986, -0.00013858, 0.00014811, -0.00022298, 0.00013314, -0.00010617, 0.00008837, -0.00007556, 0.00004479, -0.00004677, 0.00006254, -0.00005903, 0.00001876, -0.00002887, 0.00005263, -0.00004190, 0.00004146, -0.00003248, 0.00001358, -0.00002212, 0.00000980, 0.00000158, -0.00000247, -0.00001118, 0.00000875, -0.00001487, 0.00002178, -0.00001008, },
		{ 0.00001820, -0.00000685, -0.00000997, 0.00000522, -0.00001935, 0.00001920, -0.00002194, 0.00002597, -0.00001395, 0.00000332, -0.00001016, 0.00001246, -0.00000574, 0.00002218, -0.00006678, 0.00005941, -0.00003661, 0.00004018, -0.00006033, 0.00005079, -0.00004209, 0.00002071, -0.00003619, 0.00007623, -0.00007110, 0.00006716, -0.00015806, 0.00017633, -0.00031042, 0.00024346, -0.00088125, 0.00129242, 0.00012458, -0.00046813, 0.00014121, -0.00014597, 0.00014723, -0.00009672, 0.00010763, -0.00019796, 0.00011503, -0.00009273, 0.00007939, -0.00006966, 0.00003821, -0.00003728, 0.00005569, -0.00005624, 0.00001269, -0.00001776, 0.00004393, -0.00003647, 0.00003788, -0.00003111, 0.00001250, -0.00002171, 0.00000908, 0.00000379, -0.00000630, -0.00000661, 0.00000435, -0.00001008, 0.00001960, -0.00000998, },
		{ 0.00001831, -0.00000985, -0.00000642, 0.00000117, -0.00001540, 0.00001680, -0.00001996, 0.00002529, -0.00001393, 0.00000240, -0.00000705, 0.00000782, 0.00000127, 0.00001059, -0.00005725, 0.00005466, -0.00003112, 0.00003261, -0.00005311, 0.00004589, -0.00003819, 0.00001318, -0.00002167, 0.00005810, -0.00004613, 0.00003219, -0.00011733, 0.00012766, -0.00026214, 0.00017764, -0.00081797, 0.00135868, -0.00004276, -0.00036228, 0.00008470, -0.00008653, 0.00010473, -0.00005475, 0.00006427, -0.00016879, 0.00009301, -0.00007538, 0.00006681, -0.00006112, 0.00003012, -0.00002649, 0.00004741, -0.00005286, 0.00000697, -0.00000655, 0.00003409, -0.00002961, 0.00003298, -0.00002871, 0.00001071, -0.00002097, 0.00000855, 0.00000554, -0.00000964, -0.00000233, -0.00000000, -0.00000491, 0.00001660, -0.00000906, },
		{ 0.00001760, -0.00001210, -0.00000323, -0.00000288, -0.00001099, 0.00001372, -0.00001716, 0.00002383, -0.00001348, 0.00000146, -0.00000415, 0.00000351, 0.00000780, -0.00000111, -0.00004649, 0.00004856, -0.00002475, 0.00002409, -0.00004443, 0.00003953, -0.00003387, 0.00000652, -0.00000848, 0.00004145, -0.00002302, -0.00000206, -0.00007496, 0.00007447, -0.00020607, 0.00009878, -0.00072763, 0.00138771, -0.00018475, -0.00026609, 0.00003375, -0.00003069, 0.00006513, -0.00001562, 0.00002083, -0.00013775, 0.00006893, -0.00005555, 0.00005162, -0.00005056, 0.00002103, -0.00001490, 0.00003810, -0.00004916, 0.00000205, 0.00000411, 0.00002378, -0.00002183, 0.00002710, -0.00002551, 0.00000835, -0.00001997, 0.00000827, 0.00000671, -0.00001231, 0.00000141, -0.00000401, 0.00000034, 0.00001302, -0.00000744, },
		{ 0.00001611, -0.00001352, -0.00000061, -0.00000666, -0.00000642, 0.00001021, -0.00001375, 0.00002169, -0.00001267, 0.00000057, -0.00000162, -0.00000023, 0.00001347, -0.00001233, -0.00003515, 0.00004162, -0.00001799, 0.00001521, -0.00003483, 0.00003207, -0.00002934, 0.00000106, 0.00000271, 0.00002730, -0.00000333, -0.00003325, -0.00003399, 0.00002027, -0.00014622, 0.00001165, -0.00061479, 0.00138111, -0.00029944, -0.00018268, -0.00000884, 0.00001915, 0.00003063, 0.00001817, -0.00002007, -0.00010705, 0.00004463, -0.00003481, 0.00003497, -0.00003875, 0.00001151, -0.00000305, 0.00002815, -0.00004540, -0.00000175, 0.00001362, 0.00001365, -0.00001366, 0.00002062, -0.00002177, 0.00000557, -0.00001875, 0.00000825, 0.00000724, -0.00001419, 0.00000442, -0.00000739, 0.00000536, 0.00000912, -0.00000530, },
		{ 0.00001398, -0.00001410, 0.00000132, -0.00000993, -0.00000197, 0.00000654, -0.00000998, 0.00001906, -0.00001160, -0.00000017, 0.00000043, -0.00000322, 0.00001795, -0.00002253, -0.00002381, 0.00003435, -0.00001129, 0.00000651, -0.00002488, 0.00002385, -0.00002475, -0.00000299, 0.00001147, 0.00001634, 0.00001176, -0.00005937, 0.00000282, -0.00003151, -0.00008673, -0.00007855, -0.00048504, 0.00134207, -0.00038643, -0.00011392, -0.00004122, 0.00006138, 0.00000274, 0.00004472, -0.00005623, -0.00007870, 0.00002187, -0.00001469, 0.00001806, -0.00002647, 0.00000213, 0.00000856, 0.00001794, -0.00004178, -0.00000417, 0.00002152, 0.00000427, -0.00000559, 0.00001392, -0.00001776, 0.00000255, -0.00001738, 0.00000850, 0.00000712, -0.00001521, 0.00000656, -0.00000995, 0.00000986, 0.00000517, -0.00000284, },
		{ 0.00001136, -0.00001389, 0.00000248, -0.00001251, 0.00000211, 0.00000299, -0.00000611, 0.00001610, -0.00001036, -0.00000071, 0.00000190, -0.00000535, 0.00002102, -0.00003129, -0.00001303, 0.00002722, -0.00000512, -0.00000149, -0.00001512, 0.00001526, -0.00002022, -0.00000558, 0.00001760, 0.00000895, 0.00002148, -0.00007896, 0.00003317, -0.00007781, -0.00003152, -0.00016656, -0.00034459, 0.00127504, -0.00044681, -0.00006049, -0.00006244, 0.00009519, -0.00001785, 0.00006286, -0.00008600, -0.00005433, 0.00000216, 0.00000341, 0.00000204, -0.00001451, -0.00000655, 0.00001943, 0.00000782, -0.00003842, -0.00000512, 0.00002748, -0.00000387, 0.00000192, 0.00000735, -0.00001374, -0.00000054, -0.00001591, 0.00000897, 0.00000638, -0.00001537, 0.00000777, -0.00001156, 0.00001363, 0.00000142, -0.00000028, },
		{ 0.00000845, -0.00001300, 0.00000285, -0.00001428, 0.00000562, -0.00000021, -0.00000236, 0.00001301, -0.00000904, -0.00000099, 0.00000278, -0.00000658, 0.00002256, -0.00003828, -0.00000324, 0.00002063, 0.00000018, -0.00000833, -0.00000608, 0.00000663, -0.00001580, -0.00000682, 0.00002116, 0.00000513, 0.00002560, -0.00009110, 0.00005535, -0.00011611, 0.00001601, -0.00024741, -0.00019992, 0.00118539, -0.00048285, -0.00002199, -0.00007245, 0.00012051, -0.00003120, 0.00007219, -0.00010836, -0.00003507, -0.00001336, 0.00001833, -0.00001205, -0.00000362, -0.00001404, 0.00002914, -0.00000194, -0.00003535, -0.00000460, 0.00003130, -0.00001041, 0.00000850, 0.00000123, -0.00000996, -0.00000354, -0.00001437, 0.00000961, 0.00000513, -0.00001474, 0.00000807, -0.00001215, 0.00001650, -0.00000193, 0.00000218, },
		{ 0.00000545, -0.00001156, 0.00000248, -0.00001519, 0.00000843, -0.00000287, 0.00000103, 0.00000995, -0.00000774, -0.00000100, 0.00000306, -0.00000694, 0.00002258, -0.00004332, 0.00000522, 0.00001488, 0.00000435, -0.00001367, 0.00000182, -0.00000170, -0.00001149, -0.00000692, 0.00002244, 0.00000456, 0.00002434, -0.00009554, 0.00006839, -0.00014466, 0.00005316, -0.00031676, -0.00005733, 0.00107902, -0.00049777, 0.00000291, -0.00007193, 0.00013798, -0.00003809, 0.00007308, -0.00012300, -0.00002153, -0.00002397, 0.00002921, -0.00002338, 0.00000560, -0.00001994, 0.00003733, -0.00001108, -0.00003256, -0.00000277, 0.00003297, -0.00001514, 0.00001388, -0.00000422, -0.00000658, -0.00000632, -0.00001280, 0.00001035, 0.00000346, -0.00001342, 0.00000756, -0.00001174, 0.00001838, -0.00000470, 0.00000434, },
		{ 0.00000255, -0.00000977, 0.00000150, -0.00001525, 0.00001046, -0.00000487, 0.00000391, 0.00000708, -0.00000653, -0.00000073, 0.00000278, -0.00000652, 0.00002116, -0.00004636, 0.00001215, 0.00001016, 0.00000725, -0.00001729, 0.00000820, -0.00000947, -0.00000727, -0.00000621, 0.00002189, 0.00000664, 0.00001838, -0.00009266, 0.00007203, -0.00016250, 0.00007815, -0.00037111, 0.00007738, 0.00096198, -0.00049542, 0.00001618, -0.00006214, 0.00014870, -0.00003989, 0.00006656, -0.00013025, -0.00001375, -0.00002942, 0.00003557, -0.00003137, 0.00001267, -0.00002393, 0.00004374, -0.00001942, -0.00002996, 0.00000013, 0.00003262, -0.00001797, 0.00001789, -0.00000883, -0.00000375, -0.00000877, -0.00001124, 0.00001110, 0.00000153, -0.00001159, 0.00000636, -0.00001042, 0.00001925, -0.00000680, 0.00000605, },
		{ -0.00000006, -0.00000780, 0.00000006, -0.00001456, 0.00001171, -0.00000615, 0.00000615, 0.00000452, -0.00000547, -0.00000021, 0.00000204, -0.00000547, 0.00001850, -0.00004747, 0.00001748, 0.00000654, 0.00000884, -0.00001908, 0.00001282, -0.00001643, -0.00000309, -0.00000506, 0.00002008, 0.00001054, 0.00000877, -0.00008337, 0.00006676, -0.00016955, 0.00009018, -0.00040810, 0.00019926, 0.00084010, -0.00047991, 0.00002025, -0.00004470, 0.00015411, -0.00003834, 0.00005423, -0.00013101, -0.00001122, -0.00002995, 0.00003735, -0.00003571, 0.00001733, -0.00002586, 0.00004819, -0.00002685, -0.00002745, 0.00000377, 0.00003050, -0.00001896, 0.00002048, -0.00001252, -0.00000154, -0.00001081, -0.00000968, 0.00001177, -0.00000051, -0.00000942, 0.00000467, -0.00000837, 0.00001919, -0.00000818, 0.00000719, },
		{ -0.00000226, -0.00000584, -0.00000168, -0.00001323, 0.00001224, -0.00000671, 0.00000770, 0.00000235, -0.00000458, 0.00000054, 0.00000092, -0.00000397, 0.00001484, -0.00004684, 0.00002125, 0.00000399, 0.00000923, -0.00001910, 0.00001555, -0.00002240, 0.00000111, -0.00000385, 0.00001766, 0.00001532, -0.00000318, -0.00006907, 0.00005374, -0.00016655, 0.00008947, -0.00042654, 0.00030446, 0.00071865, -0.00045533, 0.00001769, -0.00002137, 0.00015575, -0.00003540, 0.00003805, -0.00012664, -0.00001303, -0.00002621, 0.00003484, -0.00003640, 0.00001948, -0.00002571, 0.00005062, -0.00003327, -0.00002487, 0.00000778, 0.00002696, -0.00001830, 0.00002170, -0.00001530, 0.00000004, -0.00001241, -0.00000815, 0.00001229, -0.00000251, -0.00000710, 0.00000271, -0.00000576, 0.00001829, -0.00000887, 0.00000773, },
		{ -0.00000394, -0.00000405, -0.00000351, -0.00001144, 0.00001218, -0.00000662, 0.00000854, 0.00000061, -0.00000387, 0.00000144, -0.00000047, -0.00000220, 0.00001050, -0.00004472, 0.00002361, 0.00000237, 0.00000862, -0.00001751, 0.00001635, -0.00002724, 0.00000535, -0.00000293, 0.00001523, 0.00002002, -0.00001604, -0.00005146, 0.00003462, -0.00015493, 0.00007719, -0.00042643, 0.00039033, 0.00060210, -0.00042547, 0.00001102, 0.00000614, 0.00015504, -0.00003297, 0.00002012, -0.00011880, -0.00001790, -0.00001919, 0.00002867, -0.00003370, 0.00001922, -0.00002356, 0.00005105, -0.00003865, -0.00002211, 0.00001175, 0.00002244, -0.00001624, 0.00002167, -0.00001724, 0.00000103, -0.00001356, -0.00000663, 0.00001258, -0.00000433, -0.00000481, 0.00000067, -0.00000284, 0.00001674, -0.00000893, 0.00000765, },
		{ -0.00000506, -0.00000258, -0.00000526, -0.00000937, 0.00001167, -0.00000600, 0.00000875, -0.00000071, -0.00000332, 0.00000245, -0.00000202, -0.00000038, 0.00000578, -0.00004144, 0.00002476, 0.00000150, 0.00000728, -0.00001458, 0.00001531, -0.00003086, 0.00000963, -0.00000260, 0.00001334, 0.00002374, -0.00002836, -0.00003240, 0.00001140, -0.00013668, 0.00005530, -0.00040897, 0.00045554, 0.00049394, -0.00039362, 0.00000248, 0.00003636, 0.00015310, -0.00003271, 0.00000251, -0.00010926, -0.00002438, -0.00001014, 0.00001976, -0.00002812, 0.00001684, -0.00001966, 0.00004959, -0.00004296, -0.00001907, 0.00001532, 0.00001739, -0.00001313, 0.00002059, -0.00001847, 0.00000153, -0.00001430, -0.00000514, 0.00001260, -0.00000586, -0.00000272, -0.00000123, 0.00000017, 0.00001472, -0.00000851, 0.00000701, },
		{ -0.00000562, -0.00000150, -0.00000677, -0.00000722, 0.00001090, -0.00000501, 0.00000841, -0.00000164, -0.00000290, 0.00000348, -0.00000361, 0.00000133, 0.00000099, -0.00003734, 0.00002496, 0.00000113, 0.00000554, -0.00001065, 0.00001260, -0.00003320, 0.00001393, -0.00000303, 0.00001240, 0.00002576, -0.00003884, -0.00001375, -0.00001375, -0.00011412, 0.00002637, -0.00037630, 0.00049997, 0.00039660, -0.00036240, -0.00000606, 0.00006818, 0.00015063, -0.00003584, -0.00001297, -0.00009971, -0.00003100, -0.00000041, 0.00000919, -0.00002036, 0.00001276, -0.00001432, 0.00004641, -0.00004619, -0.00001568, 0.00001818, 0.00001224, -0.00000931, 0.00001870, -0.00001914, 0.00000166, -0.00001467, -0.00000366, 0.00001230, -0.00000702, -0.00000093, -0.00000284, 0.00000304, 0.00001245, -0.00000775, 0.00000592, },
		{ -0.00000568, -0.00000085, -0.00000792, -0.00000516, 0.00001003, -0.00000382, 0.00000767, -0.00000225, -0.00000255, 0.00000447, -0.00000516, 0.00000278, -0.00000358, -0.00003275, 0.00002447, 0.00000101, 0.00000375, -0.00000615, 0.00000845, -0.00003425, 0.00001816, -0.00000430, 0.00001264, 0.00002558, -0.00004642, 0.00000282, -0.00003871, -0.00008972, -0.00000666, -0.00033137, 0.00052465, 0.00031147, -0.00033373, -0.00001321, 0.00010087, 0.00014780, -0.00004303, -0.00002493, -0.00009160, -0.00003638, 0.00000867, -0.00000186, -0.00001124, 0.00000748, -0.00000794, 0.00004174, -0.00004836, -0.00001195, 0.00002009, 0.00000739, -0.00000514, 0.00001624, -0.00001942, 0.00000156, -0.00001474, -0.00000221, 0.00001170, -0.00000777, 0.00000047, -0.00000402, 0.00000558, 0.00001013, -0.00000681, 0.00000450, },
		{ -0.00000533, -0.00000062, -0.00000863, -0.00000331, 0.00000923, -0.00000261, 0.00000667, -0.00000264, -0.00000221, 0.00000535, -0.00000660, 0.00000387, -0.00000772, -0.00002798, 0.00002356, 0.00000093, 0.00000223, -0.00000148, 0.00000316, -0.00003401, 0.00002217, -0.00000636, 0.00001411, 0.00002296, -0.00005039, 0.00001594, -0.00006157, -0.00006582, -0.00004075, -0.00027759, 0.00053150, 0.00023898, -0.00030884, -0.00001805, 0.00013416, 0.00014425, -0.00005432, -0.00003252, -0.00008603, -0.00003941, 0.00001591, -0.00001228, -0.00000160, 0.00000159, -0.00000096, 0.00003585, -0.00004947, -0.00000794, 0.00002092, 0.00000316, -0.00000094, 0.00001342, -0.00001947, 0.00000136, -0.00001458, -0.00000078, 0.00001080, -0.00000811, 0.00000145, -0.00000471, 0.00000767, 0.00000793, -0.00000588, 0.00000291, },
		{ -0.00000470, -0.00000076, -0.00000889, -0.00000177, 0.00000862, -0.00000155, 0.00000556, -0.00000290, -0.00000182, 0.00000606, -0.00000789, 0.00000454, -0.00001126, -0.00002326, 0.00002241, 0.00000070, 0.00000125, 0.00000296, -0.00000293, -0.00003251, 0.00002580, -0.00000902, 0.00001666, 0.00001799, -0.00005041, 0.00002467, -0.00008080, -0.00004448, -0.00007301, -0.00021864, 0.00052312, 0.00017874, -0.00028832, -0.00002005, 0.00016811, 0.00013913, -0.00006910, -0.00003547, -0.00008359, -0.00003936, 0.00002036, -0.00002104, 0.00000772, -0.00000436, 0.00000616, 0.00002904, -0.00004953, -0.00000379, 0.00002063, -0.00000022, 0.00000304, 0.00001043, -0.00001943, 0.00000120, -0.00001426, 0.00000061, 0.00000964, -0.00000808, 0.00000205, -0.00000491, 0.00000920, 0.00000600, -0.00000508, 0.00000131, },
		{ -0.00000391, -0.00000116, -0.00000871, -0.00000059, 0.00000829, -0.00000077, 0.00000448, -0.00000313, -0.00000132, 0.00000655, -0.00000899, 0.00000479, -0.00001411, -0.00001879, 0.00002118, 0.00000022, 0.00000101, 0.00000679, -0.00000951, -0.00002981, 0.00002884, -0.00001201, 0.00001996, 0.00001101, -0.00004657, 0.00002855, -0.00009537, -0.00002727, -0.00010094, -0.00015810, 0.00050251, 0.00012975, -0.00027228, -0.00001899, 0.00020300, 0.00013123, -0.00008617, -0.00003415, -0.00008439, -0.00003591, 0.00002142, -0.00002737, 0.00001601, -0.00000983, 0.00001300, 0.00002161, -0.00004854, 0.00000032, 0.00001930, -0.00000264, 0.00000663, 0.00000742, -0.00001939, 0.00000117, -0.00001385, 0.00000193, 0.00000830, -0.00000772, 0.00000232, -0.00000465, 0.00001016, 0.00000441, -0.00000454, -0.00000017, },
		{ -0.00000313, -0.00000171, -0.00000819, 0.00000025, 0.00000827, -0.00000035, 0.00000355, -0.00000340, -0.00000066, 0.00000680, -0.00000992, 0.00000466, -0.00001626, -0.00001466, 0.00001995, -0.00000055, 0.00000162, 0.00000974, -0.00001624, -0.00002598, 0.00003107, -0.00001500, 0.00002356, 0.00000264, -0.00003935, 0.00002763, -0.00010479, -0.00001519, -0.00012260, -0.00009921, 0.00047280, 0.00009062, -0.00026047, -0.00001485, 0.00023920, 0.00011909, -0.00010391, -0.00002947, -0.00008800, -0.00002919, 0.00001889, -0.00003073, 0.00002270, -0.00001439, 0.00001917, 0.00001385, -0.00004652, 0.00000418, 0.00001712, -0.00000408, 0.00000971, 0.00000444, -0.00001938, 0.00000132, -0.00001340, 0.00000317, 0.00000683, -0.00000714, 0.00000235, -0.00000403, 0.00001058, 0.00000321, -0.00000433, -0.00000139, },
		{ -0.00000246, -0.00000227, -0.00000741, 0.00000081, 0.00000854, -0.00000032, 0.00000284, -0.00000377, 0.00000019, 0.00000679, -0.00001070, 0.00000424, -0.00001775, -0.00001091, 0.00001872, -0.00000156, 0.00000309, 0.00001161, -0.00002284, -0.00002112, 0.00003228, -0.00001760, 0.00002691, -0.00000634, -0.00002956, 0.00002240, -0.00010915, -0.00000856, -0.00013679, -0.00004468, 0.00043694, 0.00005981, -0.00025248, -0.00000758, 0.00027698, 0.00010129, -0.00012039, -0.00002277, -0.00009357, -0.00001979, 0.00001295, -0.00003092, 0.00002743, -0.00001774, 0.00002436, 0.00000604, -0.00004350, 0.00000752, 0.00001432, -0.00000463, 0.00001226, 0.00000152, -0.00001940, 0.00000169, -0.00001292, 0.00000428, 0.00000533, -0.00000642, 0.00000226, -0.00000317, 0.00001055, 0.00000237, -0.00000447, -0.00000228, },
		{ -0.00000202, -0.00000271, -0.00000652, 0.00000117, 0.00000903, -0.00000068, 0.00000239, -0.00000428, 0.00000124, 0.00000651, -0.00001135, 0.00000362, -0.00001867, -0.00000750, 0.00001744, -0.00000267, 0.00000531, 0.00001228, -0.00002905, -0.00001533, 0.00003230, -0.00001945, 0.00002944, -0.00001504, -0.00001824, 0.00001379, -0.00010906, -0.00000709, -0.00014303, 0.00000348, 0.00039759, 0.00003577, -0.00024790, 0.00000296, 0.00031635, 0.00007655, -0.00013358, -0.00001567, -0.00009988, -0.00000866, 0.00000416, -0.00002806, 0.00003005, -0.00001969, 0.00002835, -0.00000158, -0.00003952, 0.00001012, 0.00001121, -0.00000443, 0.00001433, -0.00000138, -0.00001936, 0.00000224, -0.00001243, 0.00000523, 0.00000385, -0.00000565, 0.00000217, -0.00000223, 0.00001018, 0.00000182, -0.00000493, -0.00000278, },
		},
		{ { -0.00000390, 0.00000081, 0.00000883, 0.00000110, -0.00000519, -0.00000205, -0.00000600, -0.00000139, -0.00000948, -0.00000692, 0.00001850, 0.00000345, -0.00000089, 0.00000275, -0.00001439, 0.00000545, -0.00001122, -0.00001777, 0.00002848, -0.00000240, -0.00000494, 0.00000157, -0.00003726, 0.00002067, -0.00000991, -0.00003924, 0.00006285, -0.00002999, -0.00004811, -0.00005694, -0.00030133, 0.00012854, 0.00067929, 0.00012854, -0.00030133, -0.00005694, -0.00004811, -0.00002999, 0.00006285, -0.00003924, -0.00000991, 0.00002067, -0.00003726, 0.00000157, -0.00000494, -0.00000240, 0.00002848, -0.00001777, -0.00001122, 0.00000545, -0.00001439, 0.00000275, -0.00000089, 0.00000345, 0.00001850, -0.00000692, -0.00000948, -0.00000139, -0.00000600, -0.00000205, -0.00000519, 0.00000110, 0.00000883, 0.00000081, },
		{ -0.00000390, 0.00000081, 0.00000882, 0.00000110, -0.00000519, -0.00000204, -0.00000598, -0.00000144, -0.00000948, -0.00000686, 0.00001846, 0.00000344, -0.00000089, 0.00000277, -0.00001433, 0.00000534, -0.00001122, -0.00001764, 0.00002839, -0.00000243, -0.00000494, 0.00000160, -0.00003713, 0.00002040, -0.00000991, -0.00003889, 0.00006260, -0.00003008, -0.00004811, -0.00005679, -0.00030058, 0.00012599, 0.00068282, 0.00012599, -0.00030058, -0.00005679, -0.00004811, -0.00003008, 0.00006260, -0.00003889, -0.00000991, 0.00002040, -0.00003713, 0.00000160, -0.00000494, -0.00000243, 0.00002839, -0.00001764, -0.00001122, 0.00000534, -0.00001433, 0.00000277, -0.00000089, 0.00000344, 0.00001846, -0.00000686, -0.00000948, -0.00000144, -0.00000598, -0.00000204, -0.00000519, 0.00000110, 0.00000882, 0.00000081, },
		{ -0.00000390, 0.00000083, 0.00000880, 0.00000109, -0.00000519, -0.00000202, -0.00000591, -0.00000157, -0.00000948, -0.00000669, 0.00001835, 0.00000340, -0.00000089, 0.00000281, -0.00001415, 0.00000500, -0.00001122, -0.00001722, 0.00002812, -0.00000251, -0.00000494, 0.00000171, -0.00003672, 0.00001961, -0.00000991, -0.00003784, 0.00006188, -0.00003033, -0.00004811, -0.00005636, -0.00029835, 0.00011835, 0.00069336, 0.00011835, -0.00029835, -0.00005636, -0.00004811, -0.00003033, 0.00006188, -0.00003784, -0.00000991, 0.00001961, -0.00003672, 0.00000171, -0.00000494, -0.00000251, 0.00002812, -0.00001722, -0.00001122, 0.00000500, -0.00001415, 0.00000281, -0.00000089, 0.00000340, 0.00001835, -0.00000669, -0.00000948, -0.00000157, -0.00000591, -0.00000202, -0.00000519, 0.00000109, 0.00000880, 0.00000083, },
		{ -0.00000390, 0.00000086, 0.00000877, 0.00000107, -0.00000519, -0.00000200, -0.00000580, -0.00000179, -0.00000948, -0.00000639, 0.00001815, 0.00000334, -0.00000089, 0.00000289, -0.00001385, 0.00000443, -0.00001122, -0.00001654, 0.00002768, -0.00000266, -0.00000494, 0.00000189, -0.00003603, 0.00001829, -0.00000991, -0.00003611, 0.00006068, -0.00003076, -0.00004811, -0.00005564, -0.00029466, 0.00010571, 0.00071083, 0.00010571, -0.00029466, -0.00005564, -0.00004811, -0.00003076, 0.00006068, -0.00003611, -0.00000991, 0.00001829, -0.00003603, 0.00000189, -0.00000494, -0.00000266, 0.00002768, -0.00001654, -0.00001122, 0.00000443, -0.00001385, 0.00000289, -0.00000089, 0.00000334, 0.00001815, -0.00000639, -0.00000948, -0.00000179, -0.00000580, -0.00000200, -0.00000519, 0.00000107, 0.00000877, 0.00000086, },
		{ -0.00000390, 0.00000091, 0.00000872, 0.00000105, -0.00000519, -0.00000196, -0.00000565, -0.00000210, -0.00000948, -0.00000599, 0.00001788, 0.00000325, -0.00000089, 0.00000300, -0.00001344, 0.00000365, -0.00001122, -0.00001559, 0.00002706, -0.00000286, -0.00000494, 0.00000214, -0.00003509, 0.00001647, -0.00000991, -0.00003370, 0.00005902, -0.00003135, -0.00004811, -0.00005465, -0.00028953, 0.00008818, 0.00073504, 0.00008818, -0.00028953, -0.00005465, -0.00004811, -0.00003135, 0.00005902, -0.00003370, -0.00000991, 0.00001647, -0.00003509, 0.00000214, -0.00000494, -0.00000286, 0.00002706, -0.00001559, -0.00001122, 0.00000365, -0.00001344, 0.00000300, -0.00000089, 0.00000325, 0.00001788, -0.00000599, -0.00000948, -0.00000210, -0.00000565, -0.00000196, -0.00000519, 0.00000105, 0.00000872, 0.00000091, },
		{ -0.00000390, 0.00000096, 0.00000865, 0.00000102, -0.00000519, -0.00000191, -0.00000546, -0.00000249, -0.00000948, -0.00000547, 0.00001754, 0.00000313, -0.00000089, 0.00000314, -0.00001291, 0.00000266, -0.00001122, -0.00001438, 0.00002628, -0.00000311, -0.00000494, 0.00000245, -0.00003389, 0.00001416, -0.00000991, -0.00003065, 0.00005691, -0.00003210, -0.00004811, -0.00005338, -0.00028303, 0.00006593, 0.00076577, 0.00006593, -0.00028303, -0.00005338, -0.00004811, -0.00003210, 0.00005691, -0.00003065, -0.00000991, 0.00001416, -0.00003389, 0.00000245, -0.00000494, -0.00000311, 0.00002628, -0.00001438, -0.00001122, 0.00000266, -0.00001291, 0.00000314, -0.00000089, 0.00000313, 0.00001754, -0.00000547, -0.00000948, -0.00000249, -0.00000546, -0.00000191, -0.00000519, 0.00000102, 0.00000865, 0.00000096, },
		{ -0.00000390, 0.00000102, 0.00000858, 0.00000099, -0.00000519, -0.00000186, -0.00000522, -0.00000296, -0.00000948, -0.00000485, 0.00001713, 0.00000300, -0.00000089, 0.00000331, -0.00001228, 0.00000147, -0.00001122, -0.00001293, 0.00002534, -0.00000341, -0.00000494, 0.00000283, -0.00003245, 0.00001138, -0.00000991, -0.00002697, 0.00005437, -0.00003300, -0.00004811, -0.00005186, -0.00027522, 0.00003918, 0.00080273, 0.00003918, -0.00027522, -0.00005186, -0.00004811, -0.00003300, 0.00005437, -0.00002697, -0.00000991, 0.00001138, -0.00003245, 0.00000283, -0.00000494, -0.00000341, 0.00002534, -0.00001293, -0.00001122, 0.00000147, -0.00001228, 0.00000331, -0.00000089, 0.00000300, 0.00001713, -0.00000485, -0.00000948, -0.00000296, -0.00000522, -0.00000186, -0.00000519, 0.00000099, 0.00000858, 0.00000102, },
		{ -0.00000390, 0.00000110, 0.00000849, 0.00000095, -0.00000519, -0.00000179, -0.00000495, -0.00000351, -0.00000948, -0.00000413, 0.00001665, 0.00000284, -0.00000089, 0.00000350, -0.00001155, 0.00000009, -0.00001122, -0.00001125, 0.00002426, -0.00000377, -0.00000494, 0.00000326, -0.00003078, 0.00000816, -0.00000991, -0.00002272, 0.00005143, -0.00003404, -0.00004811, -0.00005010, -0.00026616, 0.00000819, 0.00084554, 0.00000819, -0.00026616, -0.00005010, -0.00004811, -0.00003404, 0.00005143, -0.00002272, -0.00000991, 0.00000816, -0.00003078, 0.00000326, -0.00000494, -0.00000377, 0.00002426, -0.00001125, -0.00001122, 0.00000009, -0.00001155, 0.00000350, -0.00000089, 0.00000284, 0.00001665, -0.00000413, -0.00000948, -0.00000351, -0.00000495, -0.00000179, -0.00000519, 0.00000095, 0.00000849, 0.00000110, },
		{ -0.00000390, 0.00000118, 0.00000839, 0.00000091, -0.00000519, -0.00000172, -0.00000465, -0.00000412, -0.00000948, -0.00000332, 0.00001611, 0.00000267, -0.00000089, 0.00000372, -0.00001072, -0.00000146, -0.00001122, -0.00000936, 0.00002303, -0.00000416, -0.00000494, 0.00000375, -0.00002890, 0.00000453, -0.00000991, -0.00001792, 0.00004811, -0.00003521, -0.00004811, -0.00004812, -0.00025595, -0.00002676, 0.00089381, -0.00002676, -0.00025595, -0.00004812, -0.00004811, -0.00003521, 0.00004811, -0.00001792, -0.00000991, 0.00000453, -0.00002890, 0.00000375, -0.00000494, -0.00000416, 0.00002303, -0.00000936, -0.00001122, -0.00000146, -0.00001072, 0.00000372, -0.00000089, 0.00000267, 0.00001611, -0.00000332, -0.00000948, -0.00000412, -0.00000465, -0.00000172, -0.00000519, 0.00000091, 0.00000839, 0.00000118, },
		{ -0.00000390, 0.00000128, 0.00000828, 0.00000086, -0.00000519, -0.00000164, -0.00000431, -0.00000480, -0.00000948, -0.00000242, 0.00001552, 0.00000247, -0.00000089, 0.00000396, -0.00000981, -0.00000318, -0.00001122, -0.00000727, 0.00002168, -0.00000460, -0.00000494, 0.00000429, -0.00002683, 0.00000053, -0.00000991, -0.00001263, 0.00004446, -0.00003651, -0.00004811, -0.00004592, -0.00024469, -0.00006531, 0.00094707, -0.00006531, -0.00024469, -0.00004592, -0.00004811, -0.00003651, 0.00004446, -0.00001263, -0.00000991, 0.00000053, -0.00002683, 0.00000429, -0.00000494, -0.00000460, 0.00002168, -0.00000727, -0.00001122, -0.00000318, -0.00000981, 0.00000396, -0.00000089, 0.00000247, 0.00001552, -0.00000242, -0.00000948, -0.00000480, -0.00000431, -0.00000164, -0.00000519, 0.00000086, 0.00000828, 0.00000128, },
		{ -0.00000390, 0.00000138, 0.00000816, 0.00000081, -0.00000519, -0.00000155, -0.00000395, -0.00000553, -0.00000948, -0.00000145, 0.00001488, 0.00000226, -0.00000089, 0.00000422, -0.00000883, -0.00000504, -0.00001122, -0.00000500, 0.00002021, -0.00000508, -0.00000494, 0.00000488, -0.00002458, -0.00000381, -0.00000991, -0.00000689, 0.00004049, -0.00003792, -0.00004811, -0.00004355, -0.00023247, -0.00010711, 0.00100480, -0.00010711, -0.00023247, -0.00004355, -0.00004811, -0.00003792, 0.00004049, -0.00000689, -0.00000991, -0.00000381, -0.00002458, 0.00000488, -0.00000494, -0.00000508, 0.00002021, -0.00000500, -0.00001122, -0.00000504, -0.00000883, 0.00000422, -0.00000089, 0.00000226, 0.00001488, -0.00000145, -0.00000948, -0.00000553, -0.00000395, -0.00000155, -0.00000519, 0.00000081, 0.00000816, 0.00000138, },
		{ -0.00000390, 0.00000149, 0.00000804, 0.00000075, -0.00000519, -0.00000145, -0.00000356, -0.00000632, -0.00000948, -0.00000041, 0.00001419, 0.00000203, -0.00000089, 0.00000450, -0.00000777, -0.00000703, -0.00001122, -0.00000258, 0.00001865, -0.00000558, -0.00000494, 0.00000551, -0.00002217, -0.00000845, -0.00000991, -0.00000076, 0.00003626, -0.00003942, -0.00004811, -0.00004101, -0.00021943, -0.00015174, 0.00106645, -0.00015174, -0.00021943, -0.00004101, -0.00004811, -0.00003942, 0.00003626, -0.00000076, -0.00000991, -0.00000845, -0.00002217, 0.00000551, -0.00000494, -0.00000558, 0.00001865, -0.00000258, -0.00001122, -0.00000703, -0.00000777, 0.00000450, -0.00000089, 0.00000203, 0.00001419, -0.00000041, -0.00000948, -0.00000632, -0.00000356, -0.00000145, -0.00000519, 0.00000075, 0.00000804, 0.00000149, },
		{ -0.00000390, 0.00000160, 0.00000790, 0.00000069, -0.00000519, -0.00000135, -0.00000314, -0.00000715, -0.00000948, 0.00000068, 0.00001347, 0.00000180, -0.00000089, 0.00000480, -0.00000666, -0.00000912, -0.00001122, -0.00000003, 0.00001700, -0.00000612, -0.00000494, 0.00000617, -0.00001964, -0.00001333, -0.00000991, 0.00000570, 0.00003180, -0.00004100, -0.00004811, -0.00003834, -0.00020569, -0.00019878, 0.00113143, -0.00019878, -0.00020569, -0.00003834, -0.00004811, -0.00004100, 0.00003180, 0.00000570, -0.00000991, -0.00001333, -0.00001964, 0.00000617, -0.00000494, -0.00000612, 0.00001700, -0.00000003, -0.00001122, -0.00000912, -0.00000666, 0.00000480, -0.00000089, 0.00000180, 0.00001347, 0.00000068, -0.00000948, -0.00000715, -0.00000314, -0.00000135, -0.00000519, 0.00000069, 0.00000790, 0.00000160, },
		{ -0.00000390, 0.00000172, 0.00000776, 0.00000063, -0.00000519, -0.00000125, -0.00000272, -0.00000801, -0.00000948, 0.00000182, 0.00001271, 0.00000155, -0.00000089, 0.00000510, -0.00000550, -0.00001131, -0.00001122, 0.00000262, 0.00001528, -0.00000667, -0.00000494, 0.00000686, -0.00001700, -0.00001842, -0.00000991, 0.00001242, 0.00002715, -0.00004265, -0.00004811, -0.00003556, -0.00019138, -0.00024777, 0.00119910, -0.00024777, -0.00019138, -0.00003556, -0.00004811, -0.00004265, 0.00002715, 0.00001242, -0.00000991, -0.00001842, -0.00001700, 0.00000686, -0.00000494, -0.00000667, 0.00001528, 0.00000262, -0.00001122, -0.00001131, -0.00000550, 0.00000510, -0.00000089, 0.00000155, 0.00001271, 0.00000182, -0.00000948, -0.00000801, -0.00000272, -0.00000125, -0.00000519, 0.00000063, 0.00000776, 0.00000172, },
		{ -0.00000390, 0.00000184, 0.00000762, 0.00000057, -0.00000519, -0.00000114, -0.00000228, -0.00000889, -0.00000948, 0.00000299, 0.00001194, 0.00000129, -0.00000089, 0.00000542, -0.00000431, -0.00001355, -0.00001122, 0.00000536, 0.00001351, -0.00000725, -0.00000494, 0.00000757, -0.00001429, -0.00002367, -0.00000991, 0.00001935, 0.00002236, -0.00004435, -0.00004811, -0.00003269, -0.00017663, -0.00029824, 0.00126882, -0.00029824, -0.00017663, -0.00003269, -0.00004811, -0.00004435, 0.00002236, 0.00001935, -0.00000991, -0.00002367, -0.00001429, 0.00000757, -0.00000494, -0.00000725, 0.00001351, 0.00000536, -0.00001122, -0.00001355, -0.00000431, 0.00000542, -0.00000089, 0.00000129, 0.00001194, 0.00000299, -0.00000948, -0.00000889, -0.00000228, -0.00000114, -0.00000519, 0.00000057, 0.00000762, 0.00000184, },
		{ -0.00000390, 0.00000196, 0.00000748, 0.00000050, -0.00000519, -0.00000104, -0.00000183, -0.00000980, -0.00000948, 0.00000419, 0.00001115, 0.00000103, -0.00000089, 0.00000574, -0.00000310, -0.00001584, -0.00001122, 0.00000815, 0.00001171, -0.00000783, -0.00000494, 0.00000829, -0.00001151, -0.00002901, -0.00000991, 0.00002642, 0.00001748, -0.00004608, -0.00004811, -0.00002976, -0.00016159, -0.00034971, 0.00133992, -0.00034971, -0.00016159, -0.00002976, -0.00004811, -0.00004608, 0.00001748, 0.00002642, -0.00000991, -0.00002901, -0.00001151, 0.00000829, -0.00000494, -0.00000783, 0.00001171, 0.00000815, -0.00001122, -0.00001584, -0.00000310, 0.00000574, -0.00000089, 0.00000103, 0.00001115, 0.00000419, -0.00000948, -0.00000980, -0.00000183, -0.00000104, -0.00000519, 0.00000050, 0.00000748, 0.00000196, },
		{ -0.00000390, 0.00000209, 0.00000733, 0.00000044, -0.00000519, -0.00000093, -0.00000137, -0.00001071, -0.00000948, 0.00000540, 0.00001035, 0.00000077, -0.00000089, 0.00000606, -0.00000187, -0.00001816, -0.00001122, 0.00001097, 0.00000989, -0.00000842, -0.00000494, 0.00000903, -0.00000872, -0.00003441, -0.00000991, 0.00003356, 0.00001255, -0.00004783, -0.00004811, -0.00002681, -0.00014640, -0.00040169, 0.00141171, -0.00040169, -0.00014640, -0.00002681, -0.00004811, -0.00004783, 0.00001255, 0.00003356, -0.00000991, -0.00003441, -0.00000872, 0.00000903, -0.00000494, -0.00000842, 0.00000989, 0.00001097, -0.00001122, -0.00001816, -0.00000187, 0.00000606, -0.00000089, 0.00000077, 0.00001035, 0.00000540, -0.00000948, -0.00001071, -0.00000137, -0.00000093, -0.00000519, 0.00000044, 0.00000733, 0.00000209, },
		{ -0.00000390, 0.00000221, 0.00000718, 0.00000037, -0.00000519, -0.00000082, -0.00000092, -0.00001163, -0.00000948, 0.00000660, 0.00000955, 0.00000051, -0.00000089, 0.00000639, -0.00000064, -0.00002047, -0.00001122, 0.00001378, 0.00000806, -0.00000901, -0.00000494, 0.00000976, -0.00000592, -0.00003981, -0.00000991, 0.00004069, 0.00000762, -0.00004958, -0.00004811, -0.00002386, -0.00013122, -0.00045366, 0.00148350, -0.00045366, -0.00013122, -0.00002386, -0.00004811, -0.00004958, 0.00000762, 0.00004069, -0.00000991, -0.00003981, -0.00000592, 0.00000976, -0.00000494, -0.00000901, 0.00000806, 0.00001378, -0.00001122, -0.00002047, -0.00000064, 0.00000639, -0.00000089, 0.00000051, 0.00000955, 0.00000660, -0.00000948, -0.00001163, -0.00000092, -0.00000082, -0.00000519, 0.00000037, 0.00000718, 0.00000221, },
		{ -0.00000390, 0.00000234, 0.00000704, 0.00000031, -0.00000519, -0.00000071, -0.00000047, -0.00001253, -0.00000948, 0.00000780, 0.00000876, 0.00000025, -0.00000089, 0.00000671, 0.00000057, -0.00002276, -0.00001122, 0.00001657, 0.00000626, -0.00000960, -0.00000494, 0.00001048, -0.00000315, -0.00004515, -0.00000991, 0.00004776, 0.00000273, -0.00005131, -0.00004811, -0.00002093, -0.00011618, -0.00050513, 0.00155460, -0.00050513, -0.00011618, -0.00002093, -0.00004811, -0.00005131, 0.00000273, 0.00004776, -0.00000991, -0.00004515, -0.00000315, 0.00001048, -0.00000494, -0.00000960, 0.00000626, 0.00001657, -0.00001122, -0.00002276, 0.00000057, 0.00000671, -0.00000089, 0.00000025, 0.00000876, 0.00000780, -0.00000948, -0.00001253, -0.00000047, -0.00000071, -0.00000519, 0.00000031, 0.00000704, 0.00000234, },
		{ -0.00000390, 0.00000246, 0.00000689, 0.00000025, -0.00000519, -0.00000060, -0.00000003, -0.00001342, -0.00000948, 0.00000897, 0.00000798, -0.00000001, -0.00000089, 0.00000703, 0.00000177, -0.00002501, -0.00001122, 0.00001931, 0.00000449, -0.00001017, -0.00000494, 0.00001119, -0.00000043, -0.00005040, -0.00000991, 0.00005469, -0.00000206, -0.00005301, -0.00004811, -0.00001807, -0.00010143, -0.00055560, 0.00162432, -0.00055560, -0.00010143, -0.00001807, -0.00004811, -0.00005301, -0.00000206, 0.00005469, -0.00000991, -0.00005040, -0.00000043, 0.00001119, -0.00000494, -0.00001017, 0.00000449, 0.00001931, -0.00001122, -0.00002501, 0.00000177, 0.00000703, -0.00000089, -0.00000001, 0.00000798, 0.00000897, -0.00000948, -0.00001342, -0.00000003, -0.00000060, -0.00000519, 0.00000025, 0.00000689, 0.00000246, },
		{ -0.00000390, 0.00000258, 0.00000675, 0.00000018, -0.00000519, -0.00000050, 0.00000040, -0.00001428, -0.00000948, 0.00001011, 0.00000723, -0.00000025, -0.00000089, 0.00000733, 0.00000292, -0.00002719, -0.00001122, 0.00002196, 0.00000277, -0.00001073, -0.00000494, 0.00001188, 0.00000221, -0.00005549, -0.00000991, 0.00006142, -0.00000670, -0.00005466, -0.00004811, -0.00001528, -0.00008712, -0.00060459, 0.00169200, -0.00060459, -0.00008712, -0.00001528, -0.00004811, -0.00005466, -0.00000670, 0.00006142, -0.00000991, -0.00005549, 0.00000221, 0.00001188, -0.00000494, -0.00001073, 0.00000277, 0.00002196, -0.00001122, -0.00002719, 0.00000292, 0.00000733, -0.00000089, -0.00000025, 0.00000723, 0.00001011, -0.00000948, -0.00001428, 0.00000040, -0.00000050, -0.00000519, 0.00000018, 0.00000675, 0.00000258, },
		{ -0.00000390, 0.00000269, 0.00000662, 0.00000013, -0.00000519, -0.00000040, 0.00000081, -0.00001511, -0.00000948, 0.00001120, 0.00000650, -0.00000049, -0.00000089, 0.00000763, 0.00000403, -0.00002929, -0.00001122, 0.00002451, 0.00000112, -0.00001126, -0.00000494, 0.00001254, 0.00000474, -0.00006037, -0.00000991, 0.00006787, -0.00001117, -0.00005624, -0.00004811, -0.00001261, -0.00007337, -0.00065163, 0.00175697, -0.00065163, -0.00007337, -0.00001261, -0.00004811, -0.00005624, -0.00001117, 0.00006787, -0.00000991, -0.00006037, 0.00000474, 0.00001254, -0.00000494, -0.00001126, 0.00000112, 0.00002451, -0.00001122, -0.00002929, 0.00000403, 0.00000763, -0.00000089, -0.00000049, 0.00000650, 0.00001120, -0.00000948, -0.00001511, 0.00000081, -0.00000040, -0.00000519, 0.00000013, 0.00000662, 0.00000269, },
		{ -0.00000390, 0.00000280, 0.00000649, 0.00000007, -0.00000519, -0.00000030, 0.00000120, -0.00001589, -0.00000948, 0.00001224, 0.00000582, -0.00000072, -0.00000089, 0.00000791, 0.00000509, -0.00003128, -0.00001122, 0.00002693, -0.00000044, -0.00001177, -0.00000494, 0.00001317, 0.00000714, -0.00006501, -0.00000991, 0.00007400, -0.00001540, -0.00005774, -0.00004811, -0.00001007, -0.00006033, -0.00069626, 0.00181863, -0.00069626, -0.00006033, -0.00001007, -0.00004811, -0.00005774, -0.00001540, 0.00007400, -0.00000991, -0.00006501, 0.00000714, 0.00001317, -0.00000494, -0.00001177, -0.00000044, 0.00002693, -0.00001122, -0.00003128, 0.00000509, 0.00000791, -0.00000089, -0.00000072, 0.00000582, 0.00001224, -0.00000948, -0.00001589, 0.00000120, -0.00000030, -0.00000519, 0.00000007, 0.00000649, 0.00000280, },
		{ -0.00000390, 0.00000290, 0.00000637, 0.00000002, -0.00000519, -0.00000022, 0.00000156, -0.00001663, -0.00000948, 0.00001321, 0.00000517, -0.00000093, -0.00000089, 0.00000817, 0.00000608, -0.00003314, -0.00001122, 0.00002920, -0.00000191, -0.00001225, -0.00000494, 0.00001376, 0.00000939, -0.00006935, -0.00000991, 0.00007974, -0.00001937, -0.00005915, -0.00004811, -0.00000770, -0.00004812, -0.00073806, 0.00187636, -0.00073806, -0.00004812, -0.00000770, -0.00004811, -0.00005915, -0.00001937, 0.00007974, -0.00000991, -0.00006935, 0.00000939, 0.00001376, -0.00000494, -0.00001225, -0.00000191, 0.00002920, -0.00001122, -0.00003314, 0.00000608, 0.00000817, -0.00000089, -0.00000093, 0.00000517, 0.00001321, -0.00000948, -0.00001663, 0.00000156, -0.00000022, -0.00000519, 0.00000002, 0.00000637, 0.00000290, },
		{ -0.00000390, 0.00000299, 0.00000627, -0.00000003, -0.00000519, -0.00000013, 0.00000190, -0.00001730, -0.00000948, 0.00001411, 0.00000458, -0.00000112, -0.00000089, 0.00000841, 0.00000699, -0.00003485, -0.00001122, 0.00003129, -0.00000326, -0.00001268, -0.00000494, 0.00001430, 0.00001147, -0.00007335, -0.00000991, 0.00008503, -0.00002302, -0.00006045, -0.00004811, -0.00000551, -0.00003685, -0.00077661, 0.00192961, -0.00077661, -0.00003685, -0.00000551, -0.00004811, -0.00006045, -0.00002302, 0.00008503, -0.00000991, -0.00007335, 0.00001147, 0.00001430, -0.00000494, -0.00001268, -0.00000326, 0.00003129, -0.00001122, -0.00003485, 0.00000699, 0.00000841, -0.00000089, -0.00000112, 0.00000458, 0.00001411, -0.00000948, -0.00001730, 0.00000190, -0.00000013, -0.00000519, -0.00000003, 0.00000627, 0.00000299, },
		{ -0.00000390, 0.00000308, 0.00000617, -0.00000007, -0.00000519, -0.00000006, 0.00000220, -0.00001792, -0.00000948, 0.00001492, 0.00000404, -0.00000130, -0.00000089, 0.00000863, 0.00000781, -0.00003641, -0.00001122, 0.00003318, -0.00000449, -0.00001308, -0.00000494, 0.00001479, 0.00001335, -0.00007698, -0.00000991, 0.00008983, -0.00002634, -0.00006162, -0.00004811, -0.00000352, -0.00002664, -0.00081156, 0.00197788, -0.00081156, -0.00002664, -0.00000352, -0.00004811, -0.00006162, -0.00002634, 0.00008983, -0.00000991, -0.00007698, 0.00001335, 0.00001479, -0.00000494, -0.00001308, -0.00000449, 0.00003318, -0.00001122, -0.00003641, 0.00000781, 0.00000863, -0.00000089, -0.00000130, 0.00000404, 0.00001492, -0.00000948, -0.00001792, 0.00000220, -0.00000006, -0.00000519, -0.00000007, 0.00000617, 0.00000308, },
		{ -0.00000390, 0.00000315, 0.00000608, -0.00000011, -0.00000519, 0.00000000, 0.00000247, -0.00001846, -0.00000948, 0.00001564, 0.00000357, -0.00000146, -0.00000089, 0.00000882, 0.00000854, -0.00003779, -0.00001122, 0.00003486, -0.00000557, -0.00001343, -0.00000494, 0.00001522, 0.00001502, -0.00008020, -0.00000991, 0.00009409, -0.00002928, -0.00006267, -0.00004811, -0.00000176, -0.00001759, -0.00084255, 0.00202070, -0.00084255, -0.00001759, -0.00000176, -0.00004811, -0.00006267, -0.00002928, 0.00009409, -0.00000991, -0.00008020, 0.00001502, 0.00001522, -0.00000494, -0.00001343, -0.00000557, 0.00003486, -0.00001122, -0.00003779, 0.00000854, 0.00000882, -0.00000089, -0.00000146, 0.00000357, 0.00001564, -0.00000948, -0.00001846, 0.00000247, 0.00000000, -0.00000519, -0.00000011, 0.00000608, 0.00000315, },
		{ -0.00000390, 0.00000322, 0.00000600, -0.00000015, -0.00000519, 0.00000006, 0.00000271, -0.00001893, -0.00000948, 0.00001626, 0.00000315, -0.00000159, -0.00000089, 0.00000899, 0.00000917, -0.00003898, -0.00001122, 0.00003631, -0.00000651, -0.00001374, -0.00000494, 0.00001560, 0.00001646, -0.00008298, -0.00000991, 0.00009776, -0.00003182, -0.00006357, -0.00004811, -0.00000024, -0.00000977, -0.00086930, 0.00205765, -0.00086930, -0.00000977, -0.00000024, -0.00004811, -0.00006357, -0.00003182, 0.00009776, -0.00000991, -0.00008298, 0.00001646, 0.00001560, -0.00000494, -0.00001374, -0.00000651, 0.00003631, -0.00001122, -0.00003898, 0.00000917, 0.00000899, -0.00000089, -0.00000159, 0.00000315, 0.00001626, -0.00000948, -0.00001893, 0.00000271, 0.00000006, -0.00000519, -0.00000015, 0.00000600, 0.00000322, },
		{ -0.00000390, 0.00000327, 0.00000594, -0.00000017, -0.00000519, 0.00000011, 0.00000290, -0.00001932, -0.00000948, 0.00001678, 0.00000281, -0.00000170, -0.00000089, 0.00000913, 0.00000970, -0.00003997, -0.00001122, 0.00003752, -0.00000729, -0.00001399, -0.00000494, 0.00001591, 0.00001766, -0.00008529, -0.00000991, 0.00010082, -0.00003393, -0.00006432, -0.00004811, 0.00000102, -0.00000327, -0.00089155, 0.00208838, -0.00089155, -0.00000327, 0.00000102, -0.00004811, -0.00006432, -0.00003393, 0.00010082, -0.00000991, -0.00008529, 0.00001766, 0.00001591, -0.00000494, -0.00001399, -0.00000729, 0.00003752, -0.00001122, -0.00003997, 0.00000970, 0.00000913, -0.00000089, -0.00000170, 0.00000281, 0.00001678, -0.00000948, -0.00001932, 0.00000290, 0.00000011, -0.00000519, -0.00000017, 0.00000594, 0.00000327, },
		{ -0.00000390, 0.00000331, 0.00000589, -0.00000020, -0.00000519, 0.00000015, 0.00000306, -0.00001963, -0.00000948, 0.00001719, 0.00000254, -0.00000179, -0.00000089, 0.00000924, 0.00001011, -0.00004075, -0.00001122, 0.00003847, -0.00000791, -0.00001419, -0.00000494, 0.00001616, 0.00001860, -0.00008711, -0.00000991, 0.00010322, -0.00003559, -0.00006490, -0.00004811, 0.00000202, 0.00000185, -0.00090908, 0.00211260, -0.00090908, 0.00000185, 0.00000202, -0.00004811, -0.00006490, -0.00003559, 0.00010322, -0.00000991, -0.00008711, 0.00001860, 0.00001616, -0.00000494, -0.00001419, -0.00000791, 0.00003847, -0.00001122, -0.00004075, 0.00001011, 0.00000924, -0.00000089, -0.00000179, 0.00000254, 0.00001719, -0.00000948, -0.00001963, 0.00000306, 0.00000015, -0.00000519, -0.00000020, 0.00000589, 0.00000331, },
		{ -0.00000390, 0.00000334, 0.00000585, -0.00000021, -0.00000519, 0.00000017, 0.00000317, -0.00001985, -0.00000948, 0.00001748, 0.00000235, -0.00000186, -0.00000089, 0.00000932, 0.00001041, -0.00004131, -0.00001122, 0.00003915, -0.00000835, -0.00001433, -0.00000494, 0.00001634, 0.00001928, -0.00008843, -0.00000991, 0.00010496, -0.00003679, -0.00006533, -0.00004811, 0.00000274, 0.00000555, -0.00092172, 0.00213006, -0.00092172, 0.00000555, 0.00000274, -0.00004811, -0.00006533, -0.00003679, 0.00010496, -0.00000991, -0.00008843, 0.00001928, 0.00001634, -0.00000494, -0.00001433, -0.00000835, 0.00003915, -0.00001122, -0.00004131, 0.00001041, 0.00000932, -0.00000089, -0.00000186, 0.00000235, 0.00001748, -0.00000948, -0.00001985, 0.00000317, 0.00000017, -0.00000519, -0.00000021, 0.00000585, 0.00000334, },
		{ -0.00000390, 0.00000336, 0.00000583, -0.00000022, -0.00000519, 0.00000019, 0.00000323, -0.00001999, -0.00000948, 0.00001766, 0.00000223, -0.00000189, -0.00000089, 0.00000936, 0.00001059, -0.00004165, -0.00001122, 0.00003957, -0.00000862, -0.00001442, -0.00000494, 0.00001645, 0.00001969, -0.00008922, -0.00000991, 0.00010601, -0.00003751, -0.00006559, -0.00004811, 0.00000317, 0.00000778, -0.00092936, 0.00214061, -0.00092936, 0.00000778, 0.00000317, -0.00004811, -0.00006559, -0.00003751, 0.00010601, -0.00000991, -0.00008922, 0.00001969, 0.00001645, -0.00000494, -0.00001442, -0.00000862, 0.00003957, -0.00001122, -0.00004165, 0.00001059, 0.00000936, -0.00000089, -0.00000189, 0.00000223, 0.00001766, -0.00000948, -0.00001999, 0.00000323, 0.00000019, -0.00000519, -0.00000022, 0.00000583, 0.00000336, },
		{ -0.00000390, 0.00000337, 0.00000583, -0.00000023, -0.00000519, 0.00000019, 0.00000326, -0.00002003, -0.00000948, 0.00001772, 0.00000219, -0.00000191, -0.00000089, 0.00000938, 0.00001065, -0.00004177, -0.00001122, 0.00003971, -0.00000871, -0.00001445, -0.00000494, 0.00001648, 0.00001983, -0.00008948, -0.00000991, 0.00010636, -0.00003776, -0.00006567, -0.00004811, 0.00000332, 0.00000852, -0.00093191, 0.00214414, -0.00093191, 0.00000852, 0.00000332, -0.00004811, -0.00006567, -0.00003776, 0.00010636, -0.00000991, -0.00008948, 0.00001983, 0.00001648, -0.00000494, -0.00001445, -0.00000871, 0.00003971, -0.00001122, -0.00004177, 0.00001065, 0.00000938, -0.00000089, -0.00000191, 0.00000219, 0.00001772, -0.00000948, -0.00002003, 0.00000326, 0.00000019, -0.00000519, -0.00000023, 0.00000583, 0.00000337, },
		{ -0.00000390, 0.00000336, 0.00000583, -0.00000022, -0.00000519, 0.00000019, 0.00000323, -0.00001999, -0.00000948, 0.00001766, 0.00000223, -0.00000189, -0.00000089, 0.00000936, 0.00001059, -0.00004165, -0.00001122, 0.00003957, -0.00000862, -0.00001442, -0.00000494, 0.00001645, 0.00001969, -0.00008922, -0.00000991, 0.00010601, -0.00003751, -0.00006559, -0.00004811, 0.00000317, 0.00000778, -0.00092936, 0.00214061, -0.00092936, 0.00000778, 0.00000317, -0.00004811, -0.00006559, -0.00003751, 0.00010601, -0.00000991, -0.00008922, 0.00001969, 0.00001645, -0.00000494, -0.00001442, -0.00000862, 0.00003957, -0.00001122, -0.00004165, 0.00001059, 0.00000936, -0.00000089, -0.00000189, 0.00000223, 0.00001766, -0.00000948, -0.00001999, 0.00000323, 0.00000019, -0.00000519, -0.00000022, 0.00000583, 0.00000336, },
		{ -0.00000390, 0.00000334, 0.00000585, -0.00000021, -0.00000519, 0.00000017, 0.00000317, -0.00001985, -0.00000948, 0.00001748, 0.00000235, -0.00000186, -0.00000089, 0.00000932, 0.00001041, -0.00004131, -0.00001122, 0.00003915, -0.00000835, -0.00001433, -0.00000494, 0.00001634, 0.00001928, -0.00008843, -0.00000991, 0.00010496, -0.00003679, -0.00006533, -0.00004811, 0.00000274, 0.00000555, -0.00092172, 0.00213006, -0.00092172, 0.00000555, 0.00000274, -0.00004811, -0.00006533, -0.00003679, 0.00010496, -0.00000991, -0.00008843, 0.00001928, 0.00001634, -0.00000494, -0.00001433, -0.00000835, 0.00003915, -0.00001122, -0.00004131, 0.00001041, 0.00000932, -0.00000089, -0.00000186, 0.00000235, 0.00001748, -0.00000948, -0.00001985, 0.00000317, 0.00000017, -0.00000519, -0.00000021, 0.00000585, 0.00000334, },
		{ -0.00000390, 0.00000331, 0.00000589, -0.00000020, -0.00000519, 0.00000015, 0.00000306, -0.00001963, -0.00000948, 0.00001719, 0.00000254, -0.00000179, -0.00000089, 0.00000924, 0.00001011, -0.00004075, -0.00001122, 0.00003847, -0.00000791, -0.00001419, -0.00000494, 0.00001616, 0.00001860, -0.00008711, -0.00000991, 0.00010322, -0.00003559, -0.00006490, -0.00004811, 0.00000202, 0.00000185, -0.00090908, 0.00211260, -0.00090908, 0.00000185, 0.00000202, -0.00004811, -0.00006490, -0.00003559, 0.00010322, -0.00000991, -0.00008711, 0.00001860, 0.00001616, -0.00000494, -0.00001419, -0.00000791, 0.00003847, -0.00001122, -0.00004075, 0.00001011, 0.00000924, -0.00000089, -0.00000179, 0.00000254, 0.00001719, -0.00000948, -0.00001963, 0.00000306, 0.00000015, -0.00000519, -0.00000020, 0.00000589, 0.00000331, },
		{ -0.00000390, 0.00000327, 0.00000594, -0.00000017, -0.00000519, 0.00000011, 0.00000290, -0.00001932, -0.00000948, 0.00001678, 0.00000281, -0.00000170, -0.00000089, 0.00000913, 0.00000970, -0.00003997, -0.00001122, 0.00003752, -0.00000729, -0.00001399, -0.00000494, 0.00001591, 0.00001766, -0.00008529, -0.00000991, 0.00010082, -0.00003393, -0.00006432, -0.00004811, 0.00000102, -0.00000327, -0.00089155, 0.00208838, -0.00089155, -0.00000327, 0.00000102, -0.00004811, -0.00006432, -0.00003393, 0.00010082, -0.00000991, -0.00008529, 0.00001766, 0.00001591, -0.00000494, -0.00001399, -0.00000729, 0.00003752, -0.00001122, -0.00003997, 0.00000970, 0.00000913, -0.00000089, -0.00000170, 0.00000281, 0.00001678, -0.00000948, -0.00001932, 0.00000290, 0.00000011, -0.00000519, -0.00000017, 0.00000594, 0.00000327, },
		{ -0.00000390, 0.00000322, 0.00000600, -0.00000015, -0.00000519, 0.00000006, 0.00000271, -0.00001893, -0.00000948, 0.00001626, 0.00000315, -0.00000159, -0.00000089, 0.00000899, 0.00000917, -0.00003898, -0.00001122, 0.00003631, -0.00000651, -0.00001374, -0.00000494, 0.00001560, 0.00001646, -0.00008298, -0.00000991, 0.00009776, -0.00003182, -0.00006357, -0.00004811, -0.00000024, -0.00000977, -0.00086930, 0.00205765, -0.00086930, -0.00000977, -0.00000024, -0.00004811, -0.00006357, -0.00003182, 0.00009776, -0.00000991, -0.00008298, 0.00001646, 0.00001560, -0.00000494, -0.00001374, -0.00000651, 0.00003631, -0.00001122, -0.00003898, 0.00000917, 0.00000899, -0.00000089, -0.00000159, 0.00000315, 0.00001626, -0.00000948, -0.00001893, 0.00000271, 0.00000006, -0.00000519, -0.00000015, 0.00000600, 0.00000322, },
		{ -0.00000390, 0.00000315, 0.00000608, -0.00000011, -0.00000519, 0.00000000, 0.00000247, -0.00001846, -0.00000948, 0.00001564, 0.00000357, -0.00000146, -0.00000089, 0.00000882, 0.00000854, -0.00003779, -0.00001122, 0.00003486, -0.00000557, -0.00001343, -0.00000494, 0.00001522, 0.00001502, -0.00008020, -0.00000991, 0.00009409, -0.00002928, -0.00006267, -0.00004811, -0.00000176, -0.00001759, -0.00084255, 0.00202070, -0.00084255, -0.00001759, -0.00000176, -0.00004811, -0.00006267, -0.00002928, 0.00009409, -0.00000991, -0.00008020, 0.00001502, 0.00001522, -0.00000494, -0.00001343, -0.00000557, 0.00003486, -0.00001122, -0.00003779, 0.00000854, 0.00000882, -0.00000089, -0.00000146, 0.00000357, 0.00001564, -0.00000948, -0.00001846, 0.00000247, 0.00000000, -0.00000519, -0.00000011, 0.00000608, 0.00000315, },
		{ -0.00000390, 0.00000308, 0.00000617, -0.00000007, -0.00000519, -0.00000006, 0.00000220, -0.00001792, -0.00000948, 0.00001492, 0.00000404, -0.00000130, -0.00000089, 0.00000863, 0.00000781, -0.00003641, -0.00001122, 0.00003318, -0.00000449, -0.00001308, -0.00000494, 0.00001479, 0.00001335, -0.00007698, -0.00000991, 0.00008983, -0.00002634, -0.00006162, -0.00004811, -0.00000352, -0.00002664, -0.00081156, 0.00197788, -0.00081156, -0.00002664, -0.00000352, -0.00004811, -0.00006162, -0.00002634, 0.00008983, -0.00000991, -0.00007698, 0.00001335, 0.00001479, -0.00000494, -0.00001308, -0.00000449, 0.00003318, -0.00001122, -0.00003641, 0.00000781, 0.00000863, -0.00000089, -0.00000130, 0.00000404, 0.00001492, -0.00000948, -0.00001792, 0.00000220, -0.00000006, -0.00000519, -0.00000007, 0.00000617, 0.00000308, },
		{ -0.00000390, 0.00000299, 0.00000627, -0.00000003, -0.00000519, -0.00000013, 0.00000190, -0.00001730, -0.00000948, 0.00001411, 0.00000458, -0.00000112, -0.00000089, 0.00000841, 0.00000699, -0.00003485, -0.00001122, 0.00003129, -0.00000326, -0.00001268, -0.00000494, 0.00001430, 0.00001147, -0.00007335, -0.00000991, 0.00008503, -0.00002302, -0.00006045, -0.00004811, -0.00000551, -0.00003685, -0.00077661, 0.00192961, -0.00077661, -0.00003685, -0.00000551, -0.00004811, -0.00006045, -0.00002302, 0.00008503, -0.00000991, -0.00007335, 0.00001147, 0.00001430, -0.00000494, -0.00001268, -0.00000326, 0.00003129, -0.00001122, -0.00003485, 0.00000699, 0.00000841, -0.00000089, -0.00000112, 0.00000458, 0.00001411, -0.00000948, -0.00001730, 0.00000190, -0.00000013, -0.00000519, -0.00000003, 0.00000627, 0.00000299, },
		{ -0.00000390, 0.00000290, 0.00000637, 0.00000002, -0.00000519, -0.00000022, 0.00000156, -0.00001663, -0.00000948, 0.00001321, 0.00000517, -0.00000093, -0.00000089, 0.00000817, 0.00000608, -0.00003314, -0.00001122, 0.00002920, -0.00000191, -0.00001225, -0.00000494, 0.00001376, 0.00000939, -0.00006935, -0.00000991, 0.00007974, -0.00001937, -0.00005915, -0.00004811, -0.00000770, -0.00004812, -0.00073806, 0.00187636, -0.00073806, -0.00004812, -0.00000770, -0.00004811, -0.00005915, -0.00001937, 0.00007974, -0.00000991, -0.00006935, 0.00000939, 0.00001376, -0.00000494, -0.00001225, -0.00000191, 0.00002920, -0.00001122, -0.00003314, 0.00000608, 0.00000817, -0.00000089, -0.00000093, 0.00000517, 0.00001321, -0.00000948, -0.00001663, 0.00000156, -0.00000022, -0.00000519, 0.00000002, 0.00000637, 0.00000290, },
		{ -0.00000390, 0.00000280, 0.00000649, 0.00000007, -0.00000519, -0.00000030, 0.00000120, -0.00001589, -0.00000948, 0.00001224, 0.00000582, -0.00000072, -0.00000089, 0.00000791, 0.00000509, -0.00003128, -0.00001122, 0.00002693, -0.00000044, -0.00001177, -0.00000494, 0.00001317, 0.00000714, -0.00006501, -0.00000991, 0.00007400, -0.00001540, -0.00005774, -0.00004811, -0.00001007, -0.00006033, -0.00069626, 0.00181863, -0.00069626, -0.00006033, -0.00001007, -0.00004811, -0.00005774, -0.00001540, 0.00007400, -0.00000991, -0.00006501, 0.00000714, 0.00001317, -0.00000494, -0.00001177, -0.00000044, 0.00002693, -0.00001122, -0.00003128, 0.00000509, 0.00000791, -0.00000089, -0.00000072, 0.00000582, 0.00001224, -0.00000948, -0.00001589, 0.00000120, -0.00000030, -0.00000519, 0.00000007, 0.00000649, 0.00000280, },
		{ -0.00000390, 0.00000269, 0.00000662, 0.00000013, -0.00000519, -0.00000040, 0.00000081, -0.00001511, -0.00000948, 0.00001120, 0.00000650, -0.00000049, -0.00000089, 0.00000763, 0.00000403, -0.00002929, -0.00001122, 0.00002451, 0.00000112, -0.00001126, -0.00000494, 0.00001254, 0.00000474, -0.00006037, -0.00000991, 0.00006787, -0.00001117, -0.00005624, -0.00004811, -0.00001261, -0.00007337, -0.00065163, 0.00175697, -0.00065163, -0.00007337, -0.00001261, -0.00004811, -0.00005624, -0.00001117, 0.00006787, -0.00000991, -0.00006037, 0.00000474, 0.00001254, -0.00000494, -0.00001126, 0.00000112, 0.00002451, -0.00001122, -0.00002929, 0.00000403, 0.00000763, -0.00000089, -0.00000049, 0.00000650, 0.00001120, -0.00000948, -0.00001511, 0.00000081, -0.00000040, -0.00000519, 0.00000013, 0.00000662, 0.00000269, },
		{ -0.00000390, 0.00000258, 0.00000675, 0.00000018, -0.00000519, -0.00000050, 0.00000040, -0.00001428, -0.00000948, 0.00001011, 0.00000723, -0.00000025, -0.00000089, 0.00000733, 0.00000292, -0.00002719, -0.00001122, 0.00002196, 0.00000277, -0.00001073, -0.00000494, 0.00001188, 0.00000221, -0.00005549, -0.00000991, 0.00006142, -0.00000670, -0.00005466, -0.00004811, -0.00001528, -0.00008712, -0.00060459, 0.00169200, -0.00060459, -0.00008712, -0.00001528, -0.00004811, -0.00005466, -0.00000670, 0.00006142, -0.00000991, -0.00005549, 0.00000221, 0.00001188, -0.00000494, -0.00001073, 0.00000277, 0.00002196, -0.00001122, -0.00002719, 0.00000292, 0.00000733, -0.00000089, -0.00000025, 0.00000723, 0.00001011, -0.00000948, -0.00001428, 0.00000040, -0.00000050, -0.00000519, 0.00000018, 0.00000675, 0.00000258, },
		{ -0.00000390, 0.00000246, 0.00000689, 0.00000025, -0.00000519, -0.00000060, -0.00000003, -0.00001342, -0.00000948, 0.00000897, 0.00000798, -0.00000001, -0.00000089, 0.00000703, 0.00000177, -0.00002501, -0.00001122, 0.00001931, 0.00000449, -0.00001017, -0.00000494, 0.00001119, -0.00000043, -0.00005040, -0.00000991, 0.00005469, -0.00000206, -0.00005301, -0.00004811, -0.00001807, -0.00010143, -0.00055560, 0.00162432, -0.00055560, -0.00010143, -0.00001807, -0.00004811, -0.00005301, -0.00000206, 0.00005469, -0.00000991, -0.00005040, -0.00000043, 0.00001119, -0.00000494, -0.00001017, 0.00000449, 0.00001931, -0.00001122, -0.00002501, 0.00000177, 0.00000703, -0.00000089, -0.00000001, 0.00000798, 0.00000897, -0.00000948, -0.00001342, -0.00000003, -0.00000060, -0.00000519, 0.00000025, 0.00000689, 0.00000246, },
		{ -0.00000390, 0.00000234, 0.00000704, 0.00000031, -0.00000519, -0.00000071, -0.00000047, -0.00001253, -0.00000948, 0.00000780, 0.00000876, 0.00000025, -0.00000089, 0.00000671, 0.00000057, -0.00002276, -0.00001122, 0.00001657, 0.00000626, -0.00000960, -0.00000494, 0.00001048, -0.00000315, -0.00004515, -0.00000991, 0.00004776, 0.00000273, -0.00005131, -0.00004811, -0.00002093, -0.00011618, -0.00050513, 0.00155460, -0.00050513, -0.00011618, -0.00002093, -0.00004811, -0.00005131, 0.00000273, 0.00004776, -0.00000991, -0.00004515, -0.00000315, 0.00001048, -0.00000494, -0.00000960, 0.00000626, 0.00001657, -0.00001122, -0.00002276, 0.00000057, 0.00000671, -0.00000089, 0.00000025, 0.00000876, 0.00000780, -0.00000948, -0.00001253, -0.00000047, -0.00000071, -0.00000519, 0.00000031, 0.00000704, 0.00000234, },
		{ -0.00000390, 0.00000221, 0.00000718, 0.00000037, -0.00000519, -0.00000082, -0.00000092, -0.00001163, -0.00000948, 0.00000660, 0.00000955, 0.00000051, -0.00000089, 0.00000639, -0.00000064, -0.00002047, -0.00001122, 0.00001378, 0.00000806, -0.00000901, -0.00000494, 0.00000976, -0.00000592, -0.00003981, -0.00000991, 0.00004069, 0.00000762, -0.00004958, -0.00004811, -0.00002386, -0.00013122, -0.00045366, 0.00148350, -0.00045366, -0.00013122, -0.00002386, -0.00004811, -0.00004958, 0.00000762, 0.00004069, -0.00000991, -0.00003981, -0.00000592, 0.00000976, -0.00000494, -0.00000901, 0.00000806, 0.00001378, -0.00001122, -0.00002047, -0.00000064, 0.00000639, -0.00000089, 0.00000051, 0.00000955, 0.00000660, -0.00000948, -0.00001163, -0.00000092, -0.00000082, -0.00000519, 0.00000037, 0.00000718, 0.00000221, },
		{ -0.00000390, 0.00000209, 0.00000733, 0.00000044, -0.00000519, -0.00000093, -0.00000137, -0.00001071, -0.00000948, 0.00000540, 0.00001035, 0.00000077, -0.00000089, 0.00000606, -0.00000187, -0.00001816, -0.00001122, 0.00001097, 0.00000989, -0.00000842, -0.00000494, 0.00000903, -0.00000872, -0.00003441, -0.00000991, 0.00003356, 0.00001255, -0.00004783, -0.00004811, -0.00002681, -0.00014640, -0.00040169, 0.00141171, -0.00040169, -0.00014640, -0.00002681, -0.00004811, -0.00004783, 0.00001255, 0.00003356, -0.00000991, -0.00003441, -0.00000872, 0.00000903, -0.00000494, -0.00000842, 0.00000989, 0.00001097, -0.00001122, -0.00001816, -0.00000187, 0.00000606, -0.00000089, 0.00000077, 0.00001035, 0.00000540, -0.00000948, -0.00001071, -0.00000137, -0.00000093, -0.00000519, 0.00000044, 0.00000733, 0.00000209, },
		{ -0.00000390, 0.00000196, 0.00000748, 0.00000050, -0.00000519, -0.00000104, -0.00000183, -0.00000980, -0.00000948, 0.00000419, 0.00001115, 0.00000103, -0.00000089, 0.00000574, -0.00000310, -0.00001584, -0.00001122, 0.00000815, 0.00001171, -0.00000783, -0.00000494, 0.00000829, -0.00001151, -0.00002901, -0.00000991, 0.00002642, 0.00001748, -0.00004608, -0.00004811, -0.00002976, -0.00016159, -0.00034971, 0.00133992, -0.00034971, -0.00016159, -0.00002976, -0.00004811, -0.00004608, 0.00001748, 0.00002642, -0.00000991, -0.00002901, -0.00001151, 0.00000829, -0.00000494, -0.00000783, 0.00001171, 0.00000815, -0.00001122, -0.00001584, -0.00000310, 0.00000574, -0.00000089, 0.00000103, 0.00001115, 0.00000419, -0.00000948, -0.00000980, -0.00000183, -0.00000104, -0.00000519, 0.00000050, 0.00000748, 0.00000196, },
		{ -0.00000390, 0.00000184, 0.00000762, 0.00000057, -0.00000519, -0.00000114, -0.00000228, -0.00000889, -0.00000948, 0.00000299, 0.00001194, 0.00000129, -0.00000089, 0.00000542, -0.00000431, -0.00001355, -0.00001122, 0.00000536, 0.00001351, -0.00000725, -0.00000494, 0.00000757, -0.00001429, -0.00002367, -0.00000991, 0.00001935, 0.00002236, -0.00004435, -0.00004811, -0.00003269, -0.00017663, -0.00029824, 0.00126882, -0.00029824, -0.00017663, -0.00003269, -0.00004811, -0.00004435, 0.00002236, 0.00001935, -0.00000991, -0.00002367, -0.00001429, 0.00000757, -0.00000494, -0.00000725, 0.00001351, 0.00000536, -0.00001122, -0.00001355, -0.00000431, 0.00000542, -0.00000089, 0.00000129, 0.00001194, 0.00000299, -0.00000948, -0.00000889, -0.00000228, -0.00000114, -0.00000519, 0.00000057, 0.00000762, 0.00000184, },
		{ -0.00000390, 0.00000172, 0.00000776, 0.00000063, -0.00000519, -0.00000125, -0.00000272, -0.00000801, -0.00000948, 0.00000182, 0.00001271, 0.00000155, -0.00000089, 0.00000510, -0.00000550, -0.00001131, -0.00001122, 0.00000262, 0.00001528, -0.00000667, -0.00000494, 0.00000686, -0.00001700, -0.00001842, -0.00000991, 0.00001242, 0.00002715, -0.00004265, -0.00004811, -0.00003556, -0.00019138, -0.00024777, 0.00119910, -0.00024777, -0.00019138, -0.00003556, -0.00004811, -0.00004265, 0.00002715, 0.00001242, -0.00000991, -0.00001842, -0.00001700, 0.00000686, -0.00000494, -0.00000667, 0.00001528, 0.00000262, -0.00001122, -0.00001131, -0.00000550, 0.00000510, -0.00000089, 0.00000155, 0.00001271, 0.00000182, -0.00000948, -0.00000801, -0.00000272, -0.00000125, -0.00000519, 0.00000063, 0.00000776, 0.00000172, },
		{ -0.00000390, 0.00000160, 0.00000790, 0.00000069, -0.00000519, -0.00000135, -0.00000314, -0.00000715, -0.00000948, 0.00000068, 0.00001347, 0.00000180, -0.00000089, 0.00000480, -0.00000666, -0.00000912, -0.00001122, -0.00000003, 0.00001700, -0.00000612, -0.00000494, 0.00000617, -0.00001964, -0.00001333, -0.00000991, 0.00000570, 0.00003180, -0.00004100, -0.00004811, -0.00003834, -0.00020569, -0.00019878, 0.00113143, -0.00019878, -0.00020569, -0.00003834, -0.00004811, -0.00004100, 0.00003180, 0.00000570, -0.00000991, -0.00001333, -0.00001964, 0.00000617, -0.00000494, -0.00000612, 0.00001700, -0.00000003, -0.00001122, -0.00000912, -0.00000666, 0.00000480, -0.00000089, 0.00000180, 0.00001347, 0.00000068, -0.00000948, -0.00000715, -0.00000314, -0.00000135, -0.00000519, 0.00000069, 0.00000790, 0.00000160, },
		{ -0.00000390, 0.00000149, 0.00000804, 0.00000075, -0.00000519, -0.00000145, -0.00000356, -0.00000632, -0.00000948, -0.00000041, 0.00001419, 0.00000203, -0.00000089, 0.00000450, -0.00000777, -0.00000703, -0.00001122, -0.00000258, 0.00001865, -0.00000558, -0.00000494, 0.00000551, -0.00002217, -0.00000845, -0.00000991, -0.00000076, 0.00003626, -0.00003942, -0.00004811, -0.00004101, -0.00021943, -0.00015174, 0.00106645, -0.00015174, -0.00021943, -0.00004101, -0.00004811, -0.00003942, 0.00003626, -0.00000076, -0.00000991, -0.00000845, -0.00002217, 0.00000551, -0.00000494, -0.00000558, 0.00001865, -0.00000258, -0.00001122, -0.00000703, -0.00000777, 0.00000450, -0.00000089, 0.00000203, 0.00001419, -0.00000041, -0.00000948, -0.00000632, -0.00000356, -0.00000145, -0.00000519, 0.00000075, 0.00000804, 0.00000149, },
		{ -0.00000390, 0.00000138, 0.00000816, 0.00000081, -0.00000519, -0.00000155, -0.00000395, -0.00000553, -0.00000948, -0.00000145, 0.00001488, 0.00000226, -0.00000089, 0.00000422, -0.00000883, -0.00000504, -0.00001122, -0.00000500, 0.00002021, -0.00000508, -0.00000494, 0.00000488, -0.00002458, -0.00000381, -0.00000991, -0.00000689, 0.00004049, -0.00003792, -0.00004811, -0.00004355, -0.00023247, -0.00010711, 0.00100480, -0.00010711, -0.00023247, -0.00004355, -0.00004811, -0.00003792, 0.00004049, -0.00000689, -0.00000991, -0.00000381, -0.00002458, 0.00000488, -0.00000494, -0.00000508, 0.00002021, -0.00000500, -0.00001122, -0.00000504, -0.00000883, 0.00000422, -0.00000089, 0.00000226, 0.00001488, -0.00000145, -0.00000948, -0.00000553, -0.00000395, -0.00000155, -0.00000519, 0.00000081, 0.00000816, 0.00000138, },
		{ -0.00000390, 0.00000128, 0.00000828, 0.00000086, -0.00000519, -0.00000164, -0.00000431, -0.00000480, -0.00000948, -0.00000242, 0.00001552, 0.00000247, -0.00000089, 0.00000396, -0.00000981, -0.00000318, -0.00001122, -0.00000727, 0.00002168, -0.00000460, -0.00000494, 0.00000429, -0.00002683, 0.00000053, -0.00000991, -0.00001263, 0.00004446, -0.00003651, -0.00004811, -0.00004592, -0.00024469, -0.00006531, 0.00094707, -0.00006531, -0.00024469, -0.00004592, -0.00004811, -0.00003651, 0.00004446, -0.00001263, -0.00000991, 0.00000053, -0.00002683, 0.00000429, -0.00000494, -0.00000460, 0.00002168, -0.00000727, -0.00001122, -0.00000318, -0.00000981, 0.00000396, -0.00000089, 0.00000247, 0.00001552, -0.00000242, -0.00000948, -0.00000480, -0.00000431, -0.00000164, -0.00000519, 0.00000086, 0.00000828, 0.00000128, },
		{ -0.00000390, 0.00000118, 0.00000839, 0.00000091, -0.00000519, -0.00000172, -0.00000465, -0.00000412, -0.00000948, -0.00000332, 0.00001611, 0.00000267, -0.00000089, 0.00000372, -0.00001072, -0.00000146, -0.00001122, -0.00000936, 0.00002303, -0.00000416, -0.00000494, 0.00000375, -0.00002890, 0.00000453, -0.00000991, -0.00001792, 0.00004811, -0.00003521, -0.00004811, -0.00004812, -0.00025595, -0.00002676, 0.00089381, -0.00002676, -0.00025595, -0.00004812, -0.00004811, -0.00003521, 0.00004811, -0.00001792, -0.00000991, 0.00000453, -0.00002890, 0.00000375, -0.00000494, -0.00000416, 0.00002303, -0.00000936, -0.00001122, -0.00000146, -0.00001072, 0.00000372, -0.00000089, 0.00000267, 0.00001611, -0.00000332, -0.00000948, -0.00000412, -0.00000465, -0.00000172, -0.00000519, 0.00000091, 0.00000839, 0.00000118, },
		{ -0.00000390, 0.00000110, 0.00000849, 0.00000095, -0.00000519, -0.00000179, -0.00000495, -0.00000351, -0.00000948, -0.00000413, 0.00001665, 0.00000284, -0.00000089, 0.00000350, -0.00001155, 0.00000009, -0.00001122, -0.00001125, 0.00002426, -0.00000377, -0.00000494, 0.00000326, -0.00003078, 0.00000816, -0.00000991, -0.00002272, 0.00005143, -0.00003404, -0.00004811, -0.00005010, -0.00026616, 0.00000819, 0.00084554, 0.00000819, -0.00026616, -0.00005010, -0.00004811, -0.00003404, 0.00005143, -0.00002272, -0.00000991, 0.00000816, -0.00003078, 0.00000326, -0.00000494, -0.00000377, 0.00002426, -0.00001125, -0.00001122, 0.00000009, -0.00001155, 0.00000350, -0.00000089, 0.00000284, 0.00001665, -0.00000413, -0.00000948, -0.00000351, -0.00000495, -0.00000179, -0.00000519, 0.00000095, 0.00000849, 0.00000110, },
		{ -0.00000390, 0.00000102, 0.00000858, 0.00000099, -0.00000519, -0.00000186, -0.00000522, -0.00000296, -0.00000948, -0.00000485, 0.00001713, 0.00000300, -0.00000089, 0.00000331, -0.00001228, 0.00000147, -0.00001122, -0.00001293, 0.00002534, -0.00000341, -0.00000494, 0.00000283, -0.00003245, 0.00001138, -0.00000991, -0.00002697, 0.00005437, -0.00003300, -0.00004811, -0.00005186, -0.00027522, 0.00003918, 0.00080273, 0.00003918, -0.00027522, -0.00005186, -0.00004811, -0.00003300, 0.00005437, -0.00002697, -0.00000991, 0.00001138, -0.00003245, 0.00000283, -0.00000494, -0.00000341, 0.00002534, -0.00001293, -0.00001122, 0.00000147, -0.00001228, 0.00000331, -0.00000089, 0.00000300, 0.00001713, -0.00000485, -0.00000948, -0.00000296, -0.00000522, -0.00000186, -0.00000519, 0.00000099, 0.00000858, 0.00000102, },
		{ -0.00000390, 0.00000096, 0.00000865, 0.00000102, -0.00000519, -0.00000191, -0.00000546, -0.00000249, -0.00000948, -0.00000547, 0.00001754, 0.00000313, -0.00000089, 0.00000314, -0.00001291, 0.00000266, -0.00001122, -0.00001438, 0.00002628, -0.00000311, -0.00000494, 0.00000245, -0.00003389, 0.00001416, -0.00000991, -0.00003065, 0.00005691, -0.00003210, -0.00004811, -0.00005338, -0.00028303, 0.00006593, 0.00076577, 0.00006593, -0.00028303, -0.00005338, -0.00004811, -0.00003210, 0.00005691, -0.00003065, -0.00000991, 0.00001416, -0.00003389, 0.00000245, -0.00000494, -0.00000311, 0.00002628, -0.00001438, -0.00001122, 0.00000266, -0.00001291, 0.00000314, -0.00000089, 0.00000313, 0.00001754, -0.00000547, -0.00000948, -0.00000249, -0.00000546, -0.00000191, -0.00000519, 0.00000102, 0.00000865, 0.00000096, },
		{ -0.00000390, 0.00000091, 0.00000872, 0.00000105, -0.00000519, -0.00000196, -0.00000565, -0.00000210, -0.00000948, -0.00000599, 0.00001788, 0.00000325, -0.00000089, 0.00000300, -0.00001344, 0.00000365, -0.00001122, -0.00001559, 0.00002706, -0.00000286, -0.00000494, 0.00000214, -0.00003509, 0.00001647, -0.00000991, -0.00003370, 0.00005902, -0.00003135, -0.00004811, -0.00005465, -0.00028953, 0.00008818, 0.00073504, 0.00008818, -0.00028953, -0.00005465, -0.00004811, -0.00003135, 0.00005902, -0.00003370, -0.00000991, 0.00001647, -0.00003509, 0.00000214, -0.00000494, -0.00000286, 0.00002706, -0.00001559, -0.00001122, 0.00000365, -0.00001344, 0.00000300, -0.00000089, 0.00000325, 0.00001788, -0.00000599, -0.00000948, -0.00000210, -0.00000565, -0.00000196, -0.00000519, 0.00000105, 0.00000872, 0.00000091, },
		{ -0.00000390, 0.00000086, 0.00000877, 0.00000107, -0.00000519, -0.00000200, -0.00000580, -0.00000179, -0.00000948, -0.00000639, 0.00001815, 0.00000334, -0.00000089, 0.00000289, -0.00001385, 0.00000443, -0.00001122, -0.00001654, 0.00002768, -0.00000266, -0.00000494, 0.00000189, -0.00003603, 0.00001829, -0.00000991, -0.00003611, 0.00006068, -0.00003076, -0.00004811, -0.00005564, -0.00029466, 0.00010571, 0.00071083, 0.00010571, -0.00029466, -0.00005564, -0.00004811, -0.00003076, 0.00006068, -0.00003611, -0.00000991, 0.00001829, -0.00003603, 0.00000189, -0.00000494, -0.00000266, 0.00002768, -0.00001654, -0.00001122, 0.00000443, -0.00001385, 0.00000289, -0.00000089, 0.00000334, 0.00001815, -0.00000639, -0.00000948, -0.00000179, -0.00000580, -0.00000200, -0.00000519, 0.00000107, 0.00000877, 0.00000086, },
		{ -0.00000390, 0.00000083, 0.00000880, 0.00000109, -0.00000519, -0.00000202, -0.00000591, -0.00000157, -0.00000948, -0.00000669, 0.00001835, 0.00000340, -0.00000089, 0.00000281, -0.00001415, 0.00000500, -0.00001122, -0.00001722, 0.00002812, -0.00000251, -0.00000494, 0.00000171, -0.00003672, 0.00001961, -0.00000991, -0.00003784, 0.00006188, -0.00003033, -0.00004811, -0.00005636, -0.00029835, 0.00011835, 0.00069336, 0.00011835, -0.00029835, -0.00005636, -0.00004811, -0.00003033, 0.00006188, -0.00003784, -0.00000991, 0.00001961, -0.00003672, 0.00000171, -0.00000494, -0.00000251, 0.00002812, -0.00001722, -0.00001122, 0.00000500, -0.00001415, 0.00000281, -0.00000089, 0.00000340, 0.00001835, -0.00000669, -0.00000948, -0.00000157, -0.00000591, -0.00000202, -0.00000519, 0.00000109, 0.00000880, 0.00000083, },
		{ -0.00000390, 0.00000081, 0.00000882, 0.00000110, -0.00000519, -0.00000204, -0.00000598, -0.00000144, -0.00000948, -0.00000686, 0.00001846, 0.00000344, -0.00000089, 0.00000277, -0.00001433, 0.00000534, -0.00001122, -0.00001764, 0.00002839, -0.00000243, -0.00000494, 0.00000160, -0.00003713, 0.00002040, -0.00000991, -0.00003889, 0.00006260, -0.00003008, -0.00004811, -0.00005679, -0.00030058, 0.00012599, 0.00068282, 0.00012599, -0.00030058, -0.00005679, -0.00004811, -0.00003008, 0.00006260, -0.00003889, -0.00000991, 0.00002040, -0.00003713, 0.00000160, -0.00000494, -0.00000243, 0.00002839, -0.00001764, -0.00001122, 0.00000534, -0.00001433, 0.00000277, -0.00000089, 0.00000344, 0.00001846, -0.00000686, -0.00000948, -0.00000144, -0.00000598, -0.00000204, -0.00000519, 0.00000110, 0.00000882, 0.00000081, },
		},
		{ { -0.00000186, -0.00000291, -0.00000565, 0.00000146, 0.00000962, -0.00000135, 0.00000219, -0.00000491, 0.00000247, 0.00000598, -0.00001191, 0.00000292, -0.00001916, -0.00000435, 0.00001602, -0.00000370, 0.00000811, 0.00001176, -0.00003467, -0.00000876, 0.00003101, -0.00002022, 0.00003063, -0.00002255, -0.00000660, 0.00000299, -0.00010553, -0.00000991, -0.00014163, 0.00004404, 0.00035687, 0.00001717, -0.00024641, 0.00001717, 0.00035687, 0.00004404, -0.00014163, -0.00000991, -0.00010553, 0.00000299, -0.00000660, -0.00002255, 0.00003063, -0.00002022, 0.00003101, -0.00000876, -0.00003467, 0.00001176, 0.00000811, -0.00000370, 0.00001602, -0.00000435, -0.00001916, 0.00000292, -0.00001191, 0.00000598, 0.00000247, -0.00000491, 0.00000219, -0.00000135, 0.00000962, 0.00000146, -0.00000565, -0.00000291, },
		{ -0.00000202, -0.00000271, -0.00000652, 0.00000117, 0.00000903, -0.00000068, 0.00000239, -0.00000428, 0.00000124, 0.00000651, -0.00001135, 0.00000362, -0.00001867, -0.00000750, 0.00001744, -0.00000267, 0.00000531, 0.00001228, -0.00002905, -0.00001533, 0.00003230, -0.00001945, 0.00002944, -0.00001504, -0.00001824, 0.00001379, -0.00010906, -0.00000709, -0.00014303, 0.00000348, 0.00039759, 0.00003577, -0.00024790, 0.00000296, 0.00031635, 0.00007655, -0.00013358, -0.00001567, -0.00009988, -0.00000866, 0.00000416, -0.00002806, 0.00003005, -0.00001969, 0.00002835, -0.00000158, -0.00003952, 0.00001012, 0.00001121, -0.00000443, 0.00001433, -0.00000138, -0.00001936, 0.00000224, -0.00001243, 0.00000523, 0.00000385, -0.00000565, 0.00000217, -0.00000223, 0.00001018, 0.00000182, -0.00000493, -0.00000278, },
		{ -0.00000246, -0.00000227, -0.00000741, 0.00000081, 0.00000854, -0.00000032, 0.00000284, -0.00000377, 0.00000019, 0.00000679, -0.00001070, 0.00000424, -0.00001775, -0.00001091, 0.00001872, -0.00000156, 0.00000309, 0.00001161, -0.00002284, -0.00002112, 0.00003228, -0.00001760, 0.00002691, -0.00000634, -0.00002956, 0.00002240, -0.00010915, -0.00000856, -0.00013679, -0.00004468, 0.00043694, 0.00005981, -0.00025248, -0.00000758, 0.00027698, 0.00010129, -0.00012039, -0.00002277, -0.00009357, -0.00001979, 0.00001295, -0.00003092, 0.00002743, -0.00001774, 0.00002436, 0.00000604, -0.00004350, 0.00000752, 0.00001432, -0.00000463, 0.00001226, 0.00000152, -0.00001940, 0.00000169, -0.00001292, 0.00000428, 0.00000533, -0.00000642, 0.00000226, -0.00000317, 0.00001055, 0.00000237, -0.00000447, -0.00000228, },
		{ -0.00000313, -0.00000171, -0.00000819, 0.00000025, 0.00000827, -0.00000035, 0.00000355, -0.00000340, -0.00000066, 0.00000680, -0.00000992, 0.00000466, -0.00001626, -0.00001466, 0.00001995, -0.00000055, 0.00000162, 0.00000974, -0.00001624, -0.00002598, 0.00003107, -0.00001500, 0.00002356, 0.00000264, -0.00003935, 0.00002763, -0.00010479, -0.00001519, -0.00012260, -0.00009921, 0.00047280, 0.00009062, -0.00026047, -0.00001485, 0.00023920, 0.00011909, -0.00010391, -0.00002947, -0.00008800, -0.00002919, 0.00001889, -0.00003073, 0.00002270, -0.00001439, 0.00001917, 0.00001385, -0.00004652, 0.00000418, 0.00001712, -0.00000408, 0.00000971, 0.00000444, -0.00001938, 0.00000132, -0.00001340, 0.00000317, 0.00000683, -0.00000714, 0.00000235, -0.00000403, 0.00001058, 0.00000321, -0.00000433, -0.00000139, },
		{ -0.00000391, -0.00000116, -0.00000871, -0.00000059, 0.00000829, -0.00000077, 0.00000448, -0.00000313, -0.00000132, 0.00000655, -0.00000899, 0.00000479, -0.00001411, -0.00001879, 0.00002118, 0.00000022, 0.00000101, 0.00000679, -0.00000951, -0.00002981, 0.00002884, -0.00001201, 0.00001996, 0.00001101, -0.00004657, 0.00002855, -0.00009537, -0.00002727, -0.00010094, -0.00015810, 0.00050251, 0.00012975, -0.00027228, -0.00001899, 0.00020300, 0.00013123, -0.00008617, -0.00003415, -0.00008439, -0.00003591, 0.00002142, -0.00002737, 0.00001601, -0.00000983, 0.00001300, 0.00002161, -0.00004854, 0.00000032, 0.00001930, -0.00000264, 0.00000663, 0.00000742, -0.00001939, 0.00000117, -0.00001385, 0.00000193, 0.00000830, -0.00000772, 0.00000232, -0.00000465, 0.00001016, 0.00000441, -0.00000454, -0.00000017, },
		{ -0.00000470, -0.00000076, -0.00000889, -0.00000177, 0.00000862, -0.00000155, 0.00000556, -0.00000290, -0.00000182, 0.00000606, -0.00000789, 0.00000454, -0.00001126, -0.00002326, 0.00002241, 0.00000070, 0.00000125, 0.00000296, -0.00000293, -0.00003251, 0.00002580, -0.00000902, 0.00001666, 0.00001799, -0.00005041, 0.00002467, -0.00008080, -0.00004448, -0.00007301, -0.00021864, 0.00052312, 0.00017874, -0.00028832, -0.00002005, 0.00016811, 0.00013913, -0.00006910, -0.00003547, -0.00008359, -0.00003936, 0.00002036, -0.00002104, 0.00000772, -0.00000436, 0.00000616, 0.00002904, -0.00004953, -0.00000379, 0.00002063, -0.00000022, 0.00000304, 0.00001043, -0.00001943, 0.00000120, -0.00001426, 0.00000061, 0.00000964, -0.00000808, 0.00000205, -0.00000491, 0.00000920, 0.00000600, -0.00000508, 0.00000131, },
		{ -0.00000533, -0.00000062, -0.00000863, -0.00000331, 0.00000923, -0.00000261, 0.00000667, -0.00000264, -0.00000221, 0.00000535, -0.00000660, 0.00000387, -0.00000772, -0.00002798, 0.00002356, 0.00000093, 0.00000223, -0.00000148, 0.00000316, -0.00003401, 0.00002217, -0.00000636, 0.00001411, 0.00002296, -0.00005039, 0.00001594, -0.00006157, -0.00006582, -0.00004075, -0.00027759, 0.00053150, 0.00023898, -0.00030884, -0.00001805, 0.00013416, 0.00014425, -0.00005432, -0.00003252, -0.00008603, -0.00003941, 0.00001591, -0.00001228, -0.00000160, 0.00000159, -0.00000096, 0.00003585, -0.00004947, -0.00000794, 0.00002092, 0.00000316, -0.00000094, 0.00001342, -0.00001947, 0.00000136, -0.00001458, -0.00000078, 0.00001080, -0.00000811, 0.00000145, -0.00000471, 0.00000767, 0.00000793, -0.00000588, 0.00000291, },
		{ -0.00000568, -0.00000085, -0.00000792, -0.00000516, 0.00001003, -0.00000382, 0.00000767, -0.00000225, -0.00000255, 0.00000447, -0.00000516, 0.00000278, -0.00000358, -0.00003275, 0.00002447, 0.00000101, 0.00000375, -0.00000615, 0.00000845, -0.00003425, 0.00001816, -0.00000430, 0.00001264, 0.00002558, -0.00004642, 0.00000282, -0.00003871, -0.00008972, -0.00000666, -0.00033137, 0.00052465, 0.00031147, -0.00033373, -0.00001321, 0.00010087, 0.00014780, -0.00004303, -0.00002493, -0.00009160, -0.00003638, 0.00000867, -0.00000186, -0.00001124, 0.00000748, -0.00000794, 0.00004174, -0.00004836, -0.00001195, 0.00002009, 0.00000739, -0.00000514, 0.00001624, -0.00001942, 0.00000156, -0.00001474, -0.00000221, 0.00001170, -0.00000777, 0.00000047, -0.00000402, 0.00000558, 0.00001013, -0.00000681, 0.00000450, },
		{ -0.00000562, -0.00000150, -0.00000677, -0.00000722, 0.00001090, -0.00000501, 0.00000841, -0.00000164, -0.00000290, 0.00000348, -0.00000361, 0.00000133, 0.00000099, -0.00003734, 0.00002496, 0.00000113, 0.00000554, -0.00001065, 0.00001260, -0.00003320, 0.00001393, -0.00000303, 0.00001240, 0.00002576, -0.00003884, -0.00001375, -0.00001375, -0.00011412, 0.00002637, -0.00037630, 0.00049997, 0.00039660, -0.00036240, -0.00000606, 0.00006818, 0.00015063, -0.00003584, -0.00001297, -0.00009971, -0.00003100, -0.00000041, 0.00000919, -0.00002036, 0.00001276, -0.00001432, 0.00004641, -0.00004619, -0.00001568, 0.00001818, 0.00001224, -0.00000931, 0.00001870, -0.00001914, 0.00000166, -0.00001467, -0.00000366, 0.00001230, -0.00000702, -0.00000093, -0.00000284, 0.00000304, 0.00001245, -0.00000775, 0.00000592, },
		{ -0.00000506, -0.00000258, -0.00000526, -0.00000937, 0.00001167, -0.00000600, 0.00000875, -0.00000071, -0.00000332, 0.00000245, -0.00000202, -0.00000038, 0.00000578, -0.00004144, 0.00002476, 0.00000150, 0.00000728, -0.00001458, 0.00001531, -0.00003086, 0.00000963, -0.00000260, 0.00001334, 0.00002374, -0.00002836, -0.00003240, 0.00001140, -0.00013668, 0.00005530, -0.00040897, 0.00045554, 0.00049394, -0.00039362, 0.00000248, 0.00003636, 0.00015310, -0.00003271, 0.00000251, -0.00010926, -0.00002438, -0.00001014, 0.00001976, -0.00002812, 0.00001684, -0.00001966, 0.00004959, -0.00004296, -0.00001907, 0.00001532, 0.00001739, -0.00001313, 0.00002059, -0.00001847, 0.00000153, -0.00001430, -0.00000514, 0.00001260, -0.00000586, -0.00000272, -0.00000123, 0.00000017, 0.00001472, -0.00000851, 0.00000701, },
		{ -0.00000394, -0.00000405, -0.00000351, -0.00001144, 0.00001218, -0.00000662, 0.00000854, 0.00000061, -0.00000387, 0.00000144, -0.00000047, -0.00000220, 0.00001050, -0.00004472, 0.00002361, 0.00000237, 0.00000862, -0.00001751, 0.00001635, -0.00002724, 0.00000535, -0.00000293, 0.00001523, 0.00002002, -0.00001604, -0.00005146, 0.00003462, -0.00015493, 0.00007719, -0.00042643, 0.00039033, 0.00060210, -0.00042547, 0.00001102, 0.00000614, 0.00015504, -0.00003297, 0.00002012, -0.00011880, -0.00001790, -0.00001919, 0.00002867, -0.00003370, 0.00001922, -0.00002356, 0.00005105, -0.00003865, -0.00002211, 0.00001175, 0.00002244, -0.00001624, 0.00002167, -0.00001724, 0.00000103, -0.00001356, -0.00000663, 0.00001258, -0.00000433, -0.00000481, 0.00000067, -0.00000284, 0.00001674, -0.00000893, 0.00000765, },
		{ -0.00000226, -0.00000584, -0.00000168, -0.00001323, 0.00001224, -0.00000671, 0.00000770, 0.00000235, -0.00000458, 0.00000054, 0.00000092, -0.00000397, 0.00001484, -0.00004684, 0.00002125, 0.00000399, 0.00000923, -0.00001910, 0.00001555, -0.00002240, 0.00000111, -0.00000385, 0.00001766, 0.00001532, -0.00000318, -0.00006907, 0.00005374, -0.00016655, 0.00008947, -0.00042654, 0.00030446, 0.00071865, -0.00045533, 0.00001769, -0.00002137, 0.00015575, -0.00003540, 0.00003805, -0.00012664, -0.00001303, -0.00002621, 0.00003484, -0.00003640, 0.00001948, -0.00002571, 0.00005062, -0.00003327, -0.00002487, 0.00000778, 0.00002696, -0.00001830, 0.00002170, -0.00001530, 0.00000004, -0.00001241, -0.00000815, 0.00001229, -0.00000251, -0.00000710, 0.00000271, -0.00000576, 0.00001829, -0.00000887, 0.00000773, },
		{ -0.00000006, -0.00000780, 0.00000006, -0.00001456, 0.00001171, -0.00000615, 0.00000615, 0.00000452, -0.00000547, -0.00000021, 0.00000204, -0.00000547, 0.00001850, -0.00004747, 0.00001748, 0.00000654, 0.00000884, -0.00001908, 0.00001282, -0.00001643, -0.00000309, -0.00000506, 0.00002008, 0.00001054, 0.00000877, -0.00008337, 0.00006676, -0.00016955, 0.00009018, -0.00040810, 0.00019926, 0.00084010, -0.00047991, 0.00002025, -0.00004470, 0.00015411, -0.00003834, 0.00005423, -0.00013101, -0.00001122, -0.00002995, 0.00003735, -0.00003571, 0.00001733, -0.00002586, 0.00004819, -0.00002685, -0.00002745, 0.00000377, 0.00003050, -0.00001896, 0.00002048, -0.00001252, -0.00000154, -0.00001081, -0.00000968, 0.00001177, -0.00000051, -0.00000942, 0.00000467, -0.00000837, 0.00001919, -0.00000818, 0.00000719, },
		{ 0.00000255, -0.00000977, 0.00000150, -0.00001525, 0.00001046, -0.00000487, 0.00000391, 0.00000708, -0.00000653, -0.00000073, 0.00000278, -0.00000652, 0.00002116, -0.00004636, 0.00001215, 0.00001016, 0.00000725, -0.00001729, 0.00000820, -0.00000947, -0.00000727, -0.00000621, 0.00002189, 0.00000664, 0.00001838, -0.00009266, 0.00007203, -0.00016250, 0.00007815, -0.00037111, 0.00007738, 0.00096198, -0.00049542, 0.00001618, -0.00006214, 0.00014870, -0.00003989, 0.00006656, -0.00013025, -0.00001375, -0.00002942, 0.00003557, -0.00003137, 0.00001267, -0.00002393, 0.00004374, -0.00001942, -0.00002996, 0.00000013, 0.00003262, -0.00001797, 0.00001789, -0.00000883, -0.00000375, -0.00000877, -0.00001124, 0.00001110, 0.00000153, -0.00001159, 0.00000636, -0.00001042, 0.00001925, -0.00000680, 0.00000605, },
		{ 0.00000545, -0.00001156, 0.00000248, -0.00001519, 0.00000843, -0.00000287, 0.00000103, 0.00000995, -0.00000774, -0.00000100, 0.00000306, -0.00000694, 0.00002258, -0.00004332, 0.00000522, 0.00001488, 0.00000435, -0.00001367, 0.00000182, -0.00000170, -0.00001149, -0.00000692, 0.00002244, 0.00000456, 0.00002434, -0.00009554, 0.00006839, -0.00014466, 0.00005316, -0.00031676, -0.00005733, 0.00107902, -0.00049777, 0.00000291, -0.00007193, 0.00013798, -0.00003809, 0.00007308, -0.00012300, -0.00002153, -0.00002397, 0.00002921, -0.00002338, 0.00000560, -0.00001994, 0.00003733, -0.00001108, -0.00003256, -0.00000277, 0.00003297, -0.00001514, 0.00001388, -0.00000422, -0.00000658, -0.00000632, -0.00001280, 0.00001035, 0.00000346, -0.00001342, 0.00000756, -0.00001174, 0.00001838, -0.00000470, 0.00000434, },
		{ 0.00000845, -0.00001300, 0.00000285, -0.00001428, 0.00000562, -0.00000021, -0.00000236, 0.00001301, -0.00000904, -0.00000099, 0.00000278, -0.00000658, 0.00002256, -0.00003828, -0.00000324, 0.00002063, 0.00000018, -0.00000833, -0.00000608, 0.00000663, -0.00001580, -0.00000682, 0.00002116, 0.00000513, 0.00002560, -0.00009110, 0.00005535, -0.00011611, 0.00001601, -0.00024741, -0.00019992, 0.00118539, -0.00048285, -0.00002199, -0.00007245, 0.00012051, -0.00003120, 0.00007219, -0.00010836, -0.00003507, -0.00001336, 0.00001833, -0.00001205, -0.00000362, -0.00001404, 0.00002914, -0.00000194, -0.00003535, -0.00000460, 0.00003130, -0.00001041, 0.00000850, 0.00000123, -0.00000996, -0.00000354, -0.00001437, 0.00000961, 0.00000513, -0.00001474, 0.00000807, -0.00001215, 0.00001650, -0.00000193, 0.00000218, },
		{ 0.00001136, -0.00001389, 0.00000248, -0.00001251, 0.00000211, 0.00000299, -0.00000611, 0.00001610, -0.00001036, -0.00000071, 0.00000190, -0.00000535, 0.00002102, -0.00003129, -0.00001303, 0.00002722, -0.00000512, -0.00000149, -0.00001512, 0.00001526, -0.00002022, -0.00000558, 0.00001760, 0.00000895, 0.00002148, -0.00007896, 0.00003317, -0.00007781, -0.00003152, -0.00016656, -0.00034459, 0.00127504, -0.00044681, -0.00006049, -0.00006244, 0.00009519, -0.00001785, 0.00006286, -0.00008600, -0.00005433, 0.00000216, 0.00000341, 0.00000204, -0.00001451, -0.00000655, 0.00001943, 0.00000782, -0.00003842, -0.00000512, 0.00002748, -0.00000387, 0.00000192, 0.00000735, -0.00001374, -0.00000054, -0.00001591, 0.00000897, 0.00000638, -0.00001537, 0.00000777, -0.00001156, 0.00001363, 0.00000142, -0.00000028, },
		{ 0.00001398, -0.00001410, 0.00000132, -0.00000993, -0.00000197, 0.00000654, -0.00000998, 0.00001906, -0.00001160, -0.00000017, 0.00000043, -0.00000322, 0.00001795, -0.00002253, -0.00002381, 0.00003435, -0.00001129, 0.00000651, -0.00002488, 0.00002385, -0.00002475, -0.00000299, 0.00001147, 0.00001634, 0.00001176, -0.00005937, 0.00000282, -0.00003151, -0.00008673, -0.00007855, -0.00048504, 0.00134207, -0.00038643, -0.00011392, -0.00004122, 0.00006138, 0.00000274, 0.00004472, -0.00005623, -0.00007870, 0.00002187, -0.00001469, 0.00001806, -0.00002647, 0.00000213, 0.00000856, 0.00001794, -0.00004178, -0.00000417, 0.00002152, 0.00000427, -0.00000559, 0.00001392, -0.00001776, 0.00000255, -0.00001738, 0.00000850, 0.00000712, -0.00001521, 0.00000656, -0.00000995, 0.00000986, 0.00000517, -0.00000284, },
		{ 0.00001611, -0.00001352, -0.00000061, -0.00000666, -0.00000642, 0.00001021, -0.00001375, 0.00002169, -0.00001267, 0.00000057, -0.00000162, -0.00000023, 0.00001347, -0.00001233, -0.00003515, 0.00004162, -0.00001799, 0.00001521, -0.00003483, 0.00003207, -0.00002934, 0.00000106, 0.00000271, 0.00002730, -0.00000333, -0.00003325, -0.00003399, 0.00002027, -0.00014622, 0.00001165, -0.00061479, 0.00138111, -0.00029944, -0.00018268, -0.00000884, 0.00001915, 0.00003063, 0.00001817, -0.00002007, -0.00010705, 0.00004463, -0.00003481, 0.00003497, -0.00003875, 0.00001151, -0.00000305, 0.00002815, -0.00004540, -0.00000175, 0.00001362, 0.00001365, -0.00001366, 0.00002062, -0.00002177, 0.00000557, -0.00001875, 0.00000825, 0.00000724, -0.00001419, 0.00000442, -0.00000739, 0.00000536, 0.00000912, -0.00000530, },
		{ 0.00001760, -0.00001210, -0.00000323, -0.00000288, -0.00001099, 0.00001372, -0.00001716, 0.00002383, -0.00001348, 0.00000146, -0.00000415, 0.00000351, 0.00000780, -0.00000111, -0.00004649, 0.00004856, -0.00002475, 0.00002409, -0.00004443, 0.00003953, -0.00003387, 0.00000652, -0.00000848, 0.00004145, -0.00002302, -0.00000206, -0.00007496, 0.00007447, -0.00020607, 0.00009878, -0.00072763, 0.00138771, -0.00018475, -0.00026609, 0.00003375, -0.00003069, 0.00006513, -0.00001562, 0.00002083, -0.00013775, 0.00006893, -0.00005555, 0.00005162, -0.00005056, 0.00002103, -0.00001490, 0.00003810, -0.00004916, 0.00000205, 0.00000411, 0.00002378, -0.00002183, 0.00002710, -0.00002551, 0.00000835, -0.00001997, 0.00000827, 0.00000671, -0.00001231, 0.00000141, -0.00000401, 0.00000034, 0.00001302, -0.00000744, },
		{ 0.00001831, -0.00000985, -0.00000642, 0.00000117, -0.00001540, 0.00001680, -0.00001996, 0.00002529, -0.00001393, 0.00000240, -0.00000705, 0.00000782, 0.00000127, 0.00001059, -0.00005725, 0.00005466, -0.00003112, 0.00003261, -0.00005311, 0.00004589, -0.00003819, 0.00001318, -0.00002167, 0.00005810, -0.00004613, 0.00003219, -0.00011733, 0.00012766, -0.00026214, 0.00017764, -0.00081797, 0.00135868, -0.00004276, -0.00036228, 0.00008470, -0.00008653, 0.00010473, -0.00005475, 0.00006427, -0.00016879, 0.00009301, -0.00007538, 0.00006681, -0.00006112, 0.00003012, -0.00002649, 0.00004741, -0.00005286, 0.00000697, -0.00000655, 0.00003409, -0.00002961, 0.00003298, -0.00002871, 0.00001071, -0.00002097, 0.00000855, 0.00000554, -0.00000964, -0.00000233, -0.00000000, -0.00000491, 0.00001660, -0.00000906, },
		{ 0.00001820, -0.00000685, -0.00000997, 0.00000522, -0.00001935, 0.00001920, -0.00002194, 0.00002597, -0.00001395, 0.00000332, -0.00001016, 0.00001246, -0.00000574, 0.00002218, -0.00006678, 0.00005941, -0.00003661, 0.00004018, -0.00006033, 0.00005079, -0.00004209, 0.00002071, -0.00003619, 0.00007623, -0.00007110, 0.00006716, -0.00015806, 0.00017633, -0.00031042, 0.00024346, -0.00088125, 0.00129242, 0.00012458, -0.00046813, 0.00014121, -0.00014597, 0.00014723, -0.00009672, 0.00010763, -0.00019796, 0.00011503, -0.00009273, 0.00007939, -0.00006966, 0.00003821, -0.00003728, 0.00005569, -0.00005624, 0.00001269, -0.00001776, 0.00004393, -0.00003647, 0.00003788, -0.00003111, 0.00001250, -0.00002171, 0.00000908, 0.00000379, -0.00000630, -0.00000661, 0.00000435, -0.00001008, 0.00001960, -0.00000998, },
		{ 0.00001724, -0.00000324, -0.00001363, 0.00000898, -0.00002257, 0.00002068, -0.00002294, 0.00002576, -0.00001349, 0.00000412, -0.00001332, 0.00001715, -0.00001277, 0.00003303, -0.00007448, 0.00006234, -0.00004076, 0.00004631, -0.00006563, 0.00005394, -0.00004531, 0.00002861, -0.00005120, 0.00009461, -0.00009611, 0.00010032, -0.00019413, 0.00021710, -0.00034724, 0.00029223, -0.00091423, 0.00118906, 0.00031368, -0.00057938, 0.00019972, -0.00020602, 0.00018986, -0.00013858, 0.00014811, -0.00022298, 0.00013314, -0.00010617, 0.00008837, -0.00007556, 0.00004479, -0.00004677, 0.00006254, -0.00005903, 0.00001876, -0.00002887, 0.00005263, -0.00004190, 0.00004146, -0.00003248, 0.00001358, -0.00002212, 0.00000980, 0.00000158, -0.00000247, -0.00001118, 0.00000875, -0.00001487, 0.00002178, -0.00001008, },
		{ 0.00001550, 0.00000079, -0.00001715, 0.00001217, -0.00002481, 0.00002105, -0.00002285, 0.00002464, -0.00001253, 0.00000471, -0.00001632, 0.00002157, -0.00001936, 0.00004253, -0.00007980, 0.00006306, -0.00004321, 0.00005054, -0.00006860, 0.00005508, -0.00004759, 0.00003633, -0.00006571, 0.00011185, -0.00011923, 0.00012912, -0.00022267, 0.00024705, -0.00036963, 0.00032098, -0.00091522, 0.00105057, 0.00051941, -0.00069076, 0.00025603, -0.00026318, 0.00022950, -0.00017719, 0.00018286, -0.00024170, 0.00014570, -0.00011447, 0.00009294, -0.00007834, 0.00004946, -0.00005450, 0.00006759, -0.00006089, 0.00002470, -0.00003918, 0.00005956, -0.00004546, 0.00004341, -0.00003263, 0.00001384, -0.00002216, 0.00001065, -0.00000094, 0.00000159, -0.00001574, 0.00001287, -0.00001896, 0.00002295, -0.00000930, },
		{ 0.00001308, 0.00000502, -0.00002025, 0.00001453, -0.00002589, 0.00002022, -0.00002163, 0.00002262, -0.00001107, 0.00000501, -0.00001898, 0.00002542, -0.00002505, 0.00005013, -0.00008230, 0.00006129, -0.00004368, 0.00005254, -0.00006897, 0.00005406, -0.00004867, 0.00004323, -0.00007868, 0.00012655, -0.00013856, 0.00015122, -0.00024125, 0.00026388, -0.00037551, 0.00032797, -0.00088423, 0.00088074, 0.00073526, -0.00079628, 0.00030555, -0.00031378, 0.00026288, -0.00020943, 0.00020929, -0.00025228, 0.00015141, -0.00011671, 0.00009259, -0.00007771, 0.00005190, -0.00006008, 0.00007050, -0.00006152, 0.00003000, -0.00004805, 0.00006415, -0.00004679, 0.00004355, -0.00003146, 0.00001325, -0.00002180, 0.00001153, -0.00000358, 0.00000564, -0.00001998, 0.00001641, -0.00002210, 0.00002297, -0.00000764, },
		{ 0.00001016, 0.00000919, -0.00002267, 0.00001587, -0.00002567, 0.00001816, -0.00001931, 0.00001980, -0.00000916, 0.00000495, -0.00002110, 0.00002840, -0.00002942, 0.00005537, -0.00008168, 0.00005691, -0.00004202, 0.00005211, -0.00006660, 0.00005080, -0.00004831, 0.00004870, -0.00008909, 0.00013739, -0.00015241, 0.00016467, -0.00024808, 0.00026615, -0.00036386, 0.00031287, -0.00082297, 0.00068503, 0.00095368, -0.00088952, 0.00034368, -0.00035420, 0.00028690, -0.00023249, 0.00022521, -0.00025335, 0.00014939, -0.00011239, 0.00008714, -0.00007362, 0.00005194, -0.00006324, 0.00007104, -0.00006063, 0.00003418, -0.00005485, 0.00006595, -0.00004567, 0.00004177, -0.00002894, 0.00001177, -0.00002101, 0.00001235, -0.00000614, 0.00000939, -0.00002360, 0.00001907, -0.00002407, 0.00002178, -0.00000515, },
		{ 0.00000695, 0.00001303, -0.00002419, 0.00001603, -0.00002414, 0.00001491, -0.00001603, 0.00001631, -0.00000686, 0.00000449, -0.00002253, 0.00003026, -0.00003213, 0.00005791, -0.00007780, 0.00004995, -0.00003825, 0.00004917, -0.00006148, 0.00004534, -0.00004633, 0.00005215, -0.00009605, 0.00014324, -0.00015940, 0.00016805, -0.00024213, 0.00025338, -0.00033483, 0.00027672, -0.00073475, 0.00047032, 0.00116637, -0.00096406, 0.00036606, -0.00038117, 0.00029886, -0.00024405, 0.00022908, -0.00024416, 0.00013929, -0.00010144, 0.00007674, -0.00006623, 0.00004955, -0.00006381, 0.00006905, -0.00005803, 0.00003683, -0.00005910, 0.00006466, -0.00004202, 0.00003809, -0.00002510, 0.00000948, -0.00001980, 0.00001298, -0.00000840, 0.00001259, -0.00002633, 0.00002064, -0.00002475, 0.00001940, -0.00000198, },
		{ 0.00000368, 0.00001632, -0.00002463, 0.00001497, -0.00002134, 0.00001063, -0.00001196, 0.00001232, -0.00000428, 0.00000360, -0.00002316, 0.00003082, -0.00003294, 0.00005758, -0.00007073, 0.00004061, -0.00003250, 0.00004383, -0.00005377, 0.00003782, -0.00004263, 0.00005311, -0.00009883, 0.00014326, -0.00015860, 0.00016059, -0.00022321, 0.00022609, -0.00028970, 0.00022189, -0.00062431, 0.00024451, 0.00136481, -0.00101386, 0.00036897, -0.00039206, 0.00029673, -0.00024254, 0.00022008, -0.00022461, 0.00012132, -0.00008425, 0.00006188, -0.00005592, 0.00004484, -0.00006173, 0.00006448, -0.00005358, 0.00003763, -0.00006047, 0.00006016, -0.00003591, 0.00003266, -0.00002009, 0.00000645, -0.00001820, 0.00001334, -0.00001017, 0.00001498, -0.00002795, 0.00002095, -0.00002408, 0.00001592, 0.00000169, },
		{ 0.00000058, 0.00001883, -0.00002391, 0.00001271, -0.00001741, 0.00000553, -0.00000734, 0.00000805, -0.00000152, 0.00000228, -0.00002290, 0.00002996, -0.00003174, 0.00005438, -0.00006071, 0.00002927, -0.00002507, 0.00003630, -0.00004375, 0.00002849, -0.00003718, 0.00005125, -0.00009695, 0.00013701, -0.00014961, 0.00014226, -0.00019207, 0.00018581, -0.00023083, 0.00015192, -0.00049749, 0.00001616, 0.00154068, -0.00103369, 0.00034962, -0.00038514, 0.00027933, -0.00022727, 0.00019828, -0.00019532, 0.00009626, -0.00006166, 0.00004337, -0.00004325, 0.00003807, -0.00005709, 0.00005740, -0.00004726, 0.00003638, -0.00005876, 0.00005252, -0.00002758, 0.00002572, -0.00001412, 0.00000284, -0.00001624, 0.00001334, -0.00001127, 0.00001639, -0.00002831, 0.00001993, -0.00002211, 0.00001152, 0.00000563, },
		{ -0.00000210, 0.00002041, -0.00002199, 0.00000936, -0.00001258, -0.00000011, -0.00000245, 0.00000374, 0.00000130, 0.00000057, -0.00002174, 0.00002765, -0.00002853, 0.00004846, -0.00004815, 0.00001644, -0.00001636, 0.00002696, -0.00003184, 0.00001770, -0.00003009, 0.00004641, -0.00009026, 0.00012444, -0.00013260, 0.00011379, -0.00015027, 0.00013494, -0.00016146, 0.00007122, -0.00036086, -0.00020605, 0.00168637, -0.00101952, 0.00030644, -0.00035976, 0.00024646, -0.00019849, 0.00016458, -0.00015762, 0.00006540, -0.00003487, 0.00002224, -0.00002895, 0.00002958, -0.00005011, 0.00004803, -0.00003918, 0.00003304, -0.00005401, 0.00004201, -0.00001741, 0.00001764, -0.00000748, -0.00000117, -0.00001399, 0.00001290, -0.00001157, 0.00001669, -0.00002733, 0.00001760, -0.00001897, 0.00000642, 0.00000958, },
		{ -0.00000417, 0.00002095, -0.00001894, 0.00000512, -0.00000713, -0.00000598, 0.00000239, -0.00000038, 0.00000403, -0.00000149, -0.00001971, 0.00002396, -0.00002345, 0.00004017, -0.00003364, 0.00000274, -0.00000688, 0.00001625, -0.00001853, 0.00000589, -0.00002154, 0.00003865, -0.00007889, 0.00010597, -0.00010832, 0.00007658, -0.00010017, 0.00007661, -0.00008551, -0.00001517, -0.00022134, -0.00041376, 0.00179542, -0.00096883, 0.00023925, -0.00031644, 0.00019900, -0.00015744, 0.00012073, -0.00011343, 0.00003046, -0.00000536, -0.00000029, -0.00001381, 0.00001984, -0.00004110, 0.00003669, -0.00002957, 0.00002771, -0.00004642, 0.00002906, -0.00000594, 0.00000885, -0.00000048, -0.00000538, -0.00001152, 0.00001200, -0.00001100, 0.00001582, -0.00002504, 0.00001407, -0.00001486, 0.00000092, 0.00001327, },
		{ -0.00000549, 0.00002043, -0.00001488, 0.00000026, -0.00000139, -0.00001172, 0.00000690, -0.00000409, 0.00000656, -0.00000382, -0.00001691, 0.00001904, -0.00001677, 0.00002996, -0.00001787, -0.00001111, 0.00000283, 0.00000469, -0.00000440, -0.00000644, -0.00001184, 0.00002821, -0.00006332, 0.00008242, -0.00007803, 0.00003266, -0.00004466, 0.00001441, -0.00000726, -0.00010201, -0.00008575, -0.00059944, 0.00186291, -0.00088086, 0.00014937, -0.00025694, 0.00013887, -0.00010624, 0.00006918, -0.00006515, -0.00000655, 0.00002517, -0.00002291, 0.00000130, 0.00000934, -0.00003048, 0.00002381, -0.00001875, 0.00002063, -0.00003637, 0.00001428, 0.00000623, -0.00000015, 0.00000651, -0.00000954, -0.00000894, 0.00001062, -0.00000952, 0.00001382, -0.00002153, 0.00000954, -0.00001007, -0.00000467, 0.00001644, },
		{ -0.00000593, 0.00001889, -0.00001004, -0.00000492, 0.00000428, -0.00001701, 0.00001079, -0.00000719, 0.00000879, -0.00000633, -0.00001346, 0.00001313, -0.00000886, 0.00001843, -0.00000163, -0.00002438, 0.00001218, -0.00000716, 0.00000991, -0.00001875, -0.00000139, 0.00001556, -0.00004432, 0.00005499, -0.00004344, -0.00001550, 0.00001293, -0.00004780, 0.00006891, -0.00018416, 0.00003963, -0.00075673, 0.00188575, -0.00075673, 0.00003963, -0.00018416, 0.00006891, -0.00004780, 0.00001293, -0.00001550, -0.00004344, 0.00005499, -0.00004432, 0.00001556, -0.00000139, -0.00001875, 0.00000991, -0.00000716, 0.00001218, -0.00002438, -0.00000163, 0.00001843, -0.00000886, 0.00001313, -0.00001346, -0.00000633, 0.00000879, -0.00000719, 0.00001079, -0.00001701, 0.00000428, -0.00000492, -0.00001004, 0.00001889, },
		{ -0.00000549, 0.00001644, -0.00000467, -0.00001007, 0.00000954, -0.00002153, 0.00001382, -0.00000952, 0.00001062, -0.00000894, -0.00000954, 0.00000651, -0.00000015, 0.00000623, 0.00001428, -0.00003637, 0.00002063, -0.00001875, 0.00002381, -0.00003048, 0.00000934, 0.00000130, -0.00002291, 0.00002517, -0.00000655, -0.00006515, 0.00006918, -0.00010624, 0.00013887, -0.00025694, 0.00014937, -0.00088086, 0.00186291, -0.00059944, -0.00008575, -0.00010201, -0.00000726, 0.00001441, -0.00004466, 0.00003266, -0.00007803, 0.00008242, -0.00006332, 0.00002821, -0.00001184, -0.00000644, -0.00000440, 0.00000469, 0.00000283, -0.00001111, -0.00001787, 0.00002996, -0.00001677, 0.00001904, -0.00001691, -0.00000382, 0.00000656, -0.00000409, 0.00000690, -0.00001172, -0.00000139, 0.00000026, -0.00001488, 0.00002043, },
		{ -0.00000417, 0.00001327, 0.00000092, -0.00001486, 0.00001407, -0.00002504, 0.00001582, -0.00001100, 0.00001200, -0.00001152, -0.00000538, -0.00000048, 0.00000885, -0.00000594, 0.00002906, -0.00004642, 0.00002771, -0.00002957, 0.00003669, -0.00004110, 0.00001984, -0.00001381, -0.00000029, -0.00000536, 0.00003046, -0.00011343, 0.00012073, -0.00015744, 0.00019900, -0.00031644, 0.00023925, -0.00096883, 0.00179542, -0.00041376, -0.00022134, -0.00001517, -0.00008551, 0.00007661, -0.00010017, 0.00007658, -0.00010832, 0.00010597, -0.00007889, 0.00003865, -0.00002154, 0.00000589, -0.00001853, 0.00001625, -0.00000688, 0.00000274, -0.00003364, 0.00004017, -0.00002345, 0.00002396, -0.00001971, -0.00000149, 0.00000403, -0.00000038, 0.00000239, -0.00000598, -0.00000713, 0.00000512, -0.00001894, 0.00002095, },
		{ -0.00000210, 0.00000958, 0.00000642, -0.00001897, 0.00001760, -0.00002733, 0.00001669, -0.00001157, 0.00001290, -0.00001399, -0.00000117, -0.00000748, 0.00001764, -0.00001741, 0.00004201, -0.00005401, 0.00003304, -0.00003918, 0.00004803, -0.00005011, 0.00002958, -0.00002895, 0.00002224, -0.00003487, 0.00006540, -0.00015762, 0.00016458, -0.00019849, 0.00024646, -0.00035976, 0.00030644, -0.00101952, 0.00168637, -0.00020605, -0.00036086, 0.00007122, -0.00016146, 0.00013494, -0.00015027, 0.00011379, -0.00013260, 0.00012444, -0.00009026, 0.00004641, -0.00003009, 0.00001770, -0.00003184, 0.00002696, -0.00001636, 0.00001644, -0.00004815, 0.00004846, -0.00002853, 0.00002765, -0.00002174, 0.00000057, 0.00000130, 0.00000374, -0.00000245, -0.00000011, -0.00001258, 0.00000936, -0.00002199, 0.00002041, },
		{ 0.00000058, 0.00000563, 0.00001152, -0.00002211, 0.00001993, -0.00002831, 0.00001639, -0.00001127, 0.00001334, -0.00001624, 0.00000284, -0.00001412, 0.00002572, -0.00002758, 0.00005252, -0.00005876, 0.00003638, -0.00004726, 0.00005740, -0.00005709, 0.00003807, -0.00004325, 0.00004337, -0.00006166, 0.00009626, -0.00019532, 0.00019828, -0.00022727, 0.00027933, -0.00038514, 0.00034962, -0.00103369, 0.00154068, 0.00001616, -0.00049749, 0.00015192, -0.00023083, 0.00018581, -0.00019207, 0.00014226, -0.00014961, 0.00013701, -0.00009695, 0.00005125, -0.00003718, 0.00002849, -0.00004375, 0.00003630, -0.00002507, 0.00002927, -0.00006071, 0.00005438, -0.00003174, 0.00002996, -0.00002290, 0.00000228, -0.00000152, 0.00000805, -0.00000734, 0.00000553, -0.00001741, 0.00001271, -0.00002391, 0.00001883, },
		{ 0.00000368, 0.00000169, 0.00001592, -0.00002408, 0.00002095, -0.00002795, 0.00001498, -0.00001017, 0.00001334, -0.00001820, 0.00000645, -0.00002009, 0.00003266, -0.00003591, 0.00006016, -0.00006047, 0.00003763, -0.00005358, 0.00006448, -0.00006173, 0.00004484, -0.00005592, 0.00006188, -0.00008425, 0.00012132, -0.00022461, 0.00022008, -0.00024254, 0.00029673, -0.00039206, 0.00036897, -0.00101386, 0.00136481, 0.00024451, -0.00062431, 0.00022189, -0.00028970, 0.00022609, -0.00022321, 0.00016059, -0.00015860, 0.00014326, -0.00009883, 0.00005311, -0.00004263, 0.00003782, -0.00005377, 0.00004383, -0.00003250, 0.00004061, -0.00007073, 0.00005758, -0.00003294, 0.00003082, -0.00002316, 0.00000360, -0.00000428, 0.00001232, -0.00001196, 0.00001063, -0.00002134, 0.00001497, -0.00002463, 0.00001632, },
		{ 0.00000695, -0.00000198, 0.00001940, -0.00002475, 0.00002064, -0.00002633, 0.00001259, -0.00000840, 0.00001298, -0.00001980, 0.00000948, -0.00002510, 0.00003809, -0.00004202, 0.00006466, -0.00005910, 0.00003683, -0.00005803, 0.00006905, -0.00006381, 0.00004955, -0.00006623, 0.00007674, -0.00010144, 0.00013929, -0.00024416, 0.00022908, -0.00024405, 0.00029886, -0.00038117, 0.00036606, -0.00096406, 0.00116637, 0.00047032, -0.00073475, 0.00027672, -0.00033483, 0.00025338, -0.00024213, 0.00016805, -0.00015940, 0.00014324, -0.00009605, 0.00005215, -0.00004633, 0.00004534, -0.00006148, 0.00004917, -0.00003825, 0.00004995, -0.00007780, 0.00005791, -0.00003213, 0.00003026, -0.00002253, 0.00000449, -0.00000686, 0.00001631, -0.00001603, 0.00001491, -0.00002414, 0.00001603, -0.00002419, 0.00001303, },
		{ 0.00001016, -0.00000515, 0.00002178, -0.00002407, 0.00001907, -0.00002360, 0.00000939, -0.00000614, 0.00001235, -0.00002101, 0.00001177, -0.00002894, 0.00004177, -0.00004567, 0.00006595, -0.00005485, 0.00003418, -0.00006063, 0.00007104, -0.00006324, 0.00005194, -0.00007362, 0.00008714, -0.00011239, 0.00014939, -0.00025335, 0.00022521, -0.00023249, 0.00028690, -0.00035420, 0.00034368, -0.00088952, 0.00095368, 0.00068503, -0.00082297, 0.00031287, -0.00036386, 0.00026615, -0.00024808, 0.00016467, -0.00015241, 0.00013739, -0.00008909, 0.00004870, -0.00004831, 0.00005080, -0.00006660, 0.00005211, -0.00004202, 0.00005691, -0.00008168, 0.00005537, -0.00002942, 0.00002840, -0.00002110, 0.00000495, -0.00000916, 0.00001980, -0.00001931, 0.00001816, -0.00002567, 0.00001587, -0.00002267, 0.00000919, },
		{ 0.00001308, -0.00000764, 0.00002297, -0.00002210, 0.00001641, -0.00001998, 0.00000564, -0.00000358, 0.00001153, -0.00002180, 0.00001325, -0.00003146, 0.00004355, -0.00004679, 0.00006415, -0.00004805, 0.00003000, -0.00006152, 0.00007050, -0.00006008, 0.00005190, -0.00007771, 0.00009259, -0.00011671, 0.00015141, -0.00025228, 0.00020929, -0.00020943, 0.00026288, -0.00031378, 0.00030555, -0.00079628, 0.00073526, 0.00088074, -0.00088423, 0.00032797, -0.00037551, 0.00026388, -0.00024125, 0.00015122, -0.00013856, 0.00012655, -0.00007868, 0.00004323, -0.00004867, 0.00005406, -0.00006897, 0.00005254, -0.00004368, 0.00006129, -0.00008230, 0.00005013, -0.00002505, 0.00002542, -0.00001898, 0.00000501, -0.00001107, 0.00002262, -0.00002163, 0.00002022, -0.00002589, 0.00001453, -0.00002025, 0.00000502, },
		{ 0.00001550, -0.00000930, 0.00002295, -0.00001896, 0.00001287, -0.00001574, 0.00000159, -0.00000094, 0.00001065, -0.00002216, 0.00001384, -0.00003263, 0.00004341, -0.00004546, 0.00005956, -0.00003918, 0.00002470, -0.00006089, 0.00006759, -0.00005450, 0.00004946, -0.00007834, 0.00009294, -0.00011447, 0.00014570, -0.00024170, 0.00018286, -0.00017719, 0.00022950, -0.00026318, 0.00025603, -0.00069076, 0.00051941, 0.00105057, -0.00091522, 0.00032098, -0.00036963, 0.00024705, -0.00022267, 0.00012912, -0.00011923, 0.00011185, -0.00006571, 0.00003633, -0.00004759, 0.00005508, -0.00006860, 0.00005054, -0.00004321, 0.00006306, -0.00007980, 0.00004253, -0.00001936, 0.00002157, -0.00001632, 0.00000471, -0.00001253, 0.00002464, -0.00002285, 0.00002105, -0.00002481, 0.00001217, -0.00001715, 0.00000079, },
		{ 0.00001724, -0.00001008, 0.00002178, -0.00001487, 0.00000875, -0.00001118, -0.00000247, 0.00000158, 0.00000980, -0.00002212, 0.00001358, -0.00003248, 0.00004146, -0.00004190, 0.00005263, -0.00002887, 0.00001876, -0.00005903, 0.00006254, -0.00004677, 0.00004479, -0.00007556, 0.00008837, -0.00010617, 0.00013314, -0.00022298, 0.00014811, -0.00013858, 0.00018986, -0.00020602, 0.00019972, -0.00057938, 0.00031368, 0.00118906, -0.00091423, 0.00029223, -0.00034724, 0.00021710, -0.00019413, 0.00010032, -0.00009611, 0.00009461, -0.00005120, 0.00002861, -0.00004531, 0.00005394, -0.00006563, 0.00004631, -0.00004076, 0.00006234, -0.00007448, 0.00003303, -0.00001277, 0.00001715, -0.00001332, 0.00000412, -0.00001349, 0.00002576, -0.00002294, 0.00002068, -0.00002257, 0.00000898, -0.00001363, -0.00000324, },
		{ 0.00001820, -0.00000998, 0.00001960, -0.00001008, 0.00000435, -0.00000661, -0.00000630, 0.00000379, 0.00000908, -0.00002171, 0.00001250, -0.00003111, 0.00003788, -0.00003647, 0.00004393, -0.00001776, 0.00001269, -0.00005624, 0.00005569, -0.00003728, 0.00003821, -0.00006966, 0.00007939, -0.00009273, 0.00011503, -0.00019796, 0.00010763, -0.00009672, 0.00014723, -0.00014597, 0.00014121, -0.00046813, 0.00012458, 0.00129242, -0.00088125, 0.00024346, -0.00031042, 0.00017633, -0.00015806, 0.00006716, -0.00007110, 0.00007623, -0.00003619, 0.00002071, -0.00004209, 0.00005079, -0.00006033, 0.00004018, -0.00003661, 0.00005941, -0.00006678, 0.00002218, -0.00000574, 0.00001246, -0.00001016, 0.00000332, -0.00001395, 0.00002597, -0.00002194, 0.00001920, -0.00001935, 0.00000522, -0.00000997, -0.00000685, },
		{ 0.00001831, -0.00000906, 0.00001660, -0.00000491, -0.00000000, -0.00000233, -0.00000964, 0.00000554, 0.00000855, -0.00002097, 0.00001071, -0.00002871, 0.00003298, -0.00002961, 0.00003409, -0.00000655, 0.00000697, -0.00005286, 0.00004741, -0.00002649, 0.00003012, -0.00006112, 0.00006681, -0.00007538, 0.00009301, -0.00016879, 0.00006427, -0.00005475, 0.00010473, -0.00008653, 0.00008470, -0.00036228, -0.00004276, 0.00135868, -0.00081797, 0.00017764, -0.00026214, 0.00012766, -0.00011733, 0.00003219, -0.00004613, 0.00005810, -0.00002167, 0.00001318, -0.00003819, 0.00004589, -0.00005311, 0.00003261, -0.00003112, 0.00005466, -0.00005725, 0.00001059, 0.00000127, 0.00000782, -0.00000705, 0.00000240, -0.00001393, 0.00002529, -0.00001996, 0.00001680, -0.00001540, 0.00000117, -0.00000642, -0.00000985, },
		{ 0.00001760, -0.00000744, 0.00001302, 0.00000034, -0.00000401, 0.00000141, -0.00001231, 0.00000671, 0.00000827, -0.00001997, 0.00000835, -0.00002551, 0.00002710, -0.00002183, 0.00002378, 0.00000411, 0.00000205, -0.00004916, 0.00003810, -0.00001490, 0.00002103, -0.00005056, 0.00005162, -0.00005555, 0.00006893, -0.00013775, 0.00002083, -0.00001562, 0.00006513, -0.00003069, 0.00003375, -0.00026609, -0.00018475, 0.00138771, -0.00072763, 0.00009878, -0.00020607, 0.00007447, -0.00007496, -0.00000206, -0.00002302, 0.00004145, -0.00000848, 0.00000652, -0.00003387, 0.00003953, -0.00004443, 0.00002409, -0.00002475, 0.00004856, -0.00004649, -0.00000111, 0.00000780, 0.00000351, -0.00000415, 0.00000146, -0.00001348, 0.00002383, -0.00001716, 0.00001372, -0.00001099, -0.00000288, -0.00000323, -0.00001210, },
		{ 0.00001611, -0.00000530, 0.00000912, 0.00000536, -0.00000739, 0.00000442, -0.00001419, 0.00000724, 0.00000825, -0.00001875, 0.00000557, -0.00002177, 0.00002062, -0.00001366, 0.00001365, 0.00001362, -0.00000175, -0.00004540, 0.00002815, -0.00000305, 0.00001151, -0.00003875, 0.00003497, -0.00003481, 0.00004463, -0.00010705, -0.00002007, 0.00001817, 0.00003063, 0.00001915, -0.00000884, -0.00018268, -0.00029944, 0.00138111, -0.00061479, 0.00001165, -0.00014622, 0.00002027, -0.00003399, -0.00003325, -0.00000333, 0.00002730, 0.00000271, 0.00000106, -0.00002934, 0.00003207, -0.00003483, 0.00001521, -0.00001799, 0.00004162, -0.00003515, -0.00001233, 0.00001347, -0.00000023, -0.00000162, 0.00000057, -0.00001267, 0.00002169, -0.00001375, 0.00001021, -0.00000642, -0.00000666, -0.00000061, -0.00001352, },
		{ 0.00001398, -0.00000284, 0.00000517, 0.00000986, -0.00000995, 0.00000656, -0.00001521, 0.00000712, 0.00000850, -0.00001738, 0.00000255, -0.00001776, 0.00001392, -0.00000559, 0.00000427, 0.00002152, -0.00000417, -0.00004178, 0.00001794, 0.00000856, 0.00000213, -0.00002647, 0.00001806, -0.00001469, 0.00002187, -0.00007870, -0.00005623, 0.00004472, 0.00000274, 0.00006138, -0.00004122, -0.00011392, -0.00038643, 0.00134207, -0.00048504, -0.00007855, -0.00008673, -0.00003151, 0.00000282, -0.00005937, 0.00001176, 0.00001634, 0.00001147, -0.00000299, -0.00002475, 0.00002385, -0.00002488, 0.00000651, -0.00001129, 0.00003435, -0.00002381, -0.00002253, 0.00001795, -0.00000322, 0.00000043, -0.00000017, -0.00001160, 0.00001906, -0.00000998, 0.00000654, -0.00000197, -0.00000993, 0.00000132, -0.00001410, },
		{ 0.00001136, -0.00000028, 0.00000142, 0.00001363, -0.00001156, 0.00000777, -0.00001537, 0.00000638, 0.00000897, -0.00001591, -0.00000054, -0.00001374, 0.00000735, 0.00000192, -0.00000387, 0.00002748, -0.00000512, -0.00003842, 0.00000782, 0.00001943, -0.00000655, -0.00001451, 0.00000204, 0.00000341, 0.00000216, -0.00005433, -0.00008600, 0.00006286, -0.00001785, 0.00009519, -0.00006244, -0.00006049, -0.00044681, 0.00127504, -0.00034459, -0.00016656, -0.00003152, -0.00007781, 0.00003317, -0.00007896, 0.00002148, 0.00000895, 0.00001760, -0.00000558, -0.00002022, 0.00001526, -0.00001512, -0.00000149, -0.00000512, 0.00002722, -0.00001303, -0.00003129, 0.00002102, -0.00000535, 0.00000190, -0.00000071, -0.00001036, 0.00001610, -0.00000611, 0.00000299, 0.00000211, -0.00001251, 0.00000248, -0.00001389, },
		{ 0.00000845, 0.00000218, -0.00000193, 0.00001650, -0.00001215, 0.00000807, -0.00001474, 0.00000513, 0.00000961, -0.00001437, -0.00000354, -0.00000996, 0.00000123, 0.00000850, -0.00001041, 0.00003130, -0.00000460, -0.00003535, -0.00000194, 0.00002914, -0.00001404, -0.00000362, -0.00001205, 0.00001833, -0.00001336, -0.00003507, -0.00010836, 0.00007219, -0.00003120, 0.00012051, -0.00007245, -0.00002199, -0.00048285, 0.00118539, -0.00019992, -0.00024741, 0.00001601, -0.00011611, 0.00005535, -0.00009110, 0.00002560, 0.00000513, 0.00002116, -0.00000682, -0.00001580, 0.00000663, -0.00000608, -0.00000833, 0.00000018, 0.00002063, -0.00000324, -0.00003828, 0.00002256, -0.00000658, 0.00000278, -0.00000099, -0.00000904, 0.00001301, -0.00000236, -0.00000021, 0.00000562, -0.00001428, 0.00000285, -0.00001300, },
		{ 0.00000545, 0.00000434, -0.00000470, 0.00001838, -0.00001174, 0.00000756, -0.00001342, 0.00000346, 0.00001035, -0.00001280, -0.00000632, -0.00000658, -0.00000422, 0.00001388, -0.00001514, 0.00003297, -0.00000277, -0.00003256, -0.00001108, 0.00003733, -0.00001994, 0.00000560, -0.00002338, 0.00002921, -0.00002397, -0.00002153, -0.00012300, 0.00007308, -0.00003809, 0.00013798, -0.00007193, 0.00000291, -0.00049777, 0.00107902, -0.00005733, -0.00031676, 0.00005316, -0.00014466, 0.00006839, -0.00009554, 0.00002434, 0.00000456, 0.00002244, -0.00000692, -0.00001149, -0.00000170, 0.00000182, -0.00001367, 0.00000435, 0.00001488, 0.00000522, -0.00004332, 0.00002258, -0.00000694, 0.00000306, -0.00000100, -0.00000774, 0.00000995, 0.00000103, -0.00000287, 0.00000843, -0.00001519, 0.00000248, -0.00001156, },
		{ 0.00000255, 0.00000605, -0.00000680, 0.00001925, -0.00001042, 0.00000636, -0.00001159, 0.00000153, 0.00001110, -0.00001124, -0.00000877, -0.00000375, -0.00000883, 0.00001789, -0.00001797, 0.00003262, 0.00000013, -0.00002996, -0.00001942, 0.00004374, -0.00002393, 0.00001267, -0.00003137, 0.00003557, -0.00002942, -0.00001375, -0.00013025, 0.00006656, -0.00003989, 0.00014870, -0.00006214, 0.00001618, -0.00049542, 0.00096198, 0.00007738, -0.00037111, 0.00007815, -0.00016250, 0.00007203, -0.00009266, 0.00001838, 0.00000664, 0.00002189, -0.00000621, -0.00000727, -0.00000947, 0.00000820, -0.00001729, 0.00000725, 0.00001016, 0.00001215, -0.00004636, 0.00002116, -0.00000652, 0.00000278, -0.00000073, -0.00000653, 0.00000708, 0.00000391, -0.00000487, 0.00001046, -0.00001525, 0.00000150, -0.00000977, },
		{ -0.00000006, 0.00000719, -0.00000818, 0.00001919, -0.00000837, 0.00000467, -0.00000942, -0.00000051, 0.00001177, -0.00000968, -0.00001081, -0.00000154, -0.00001252, 0.00002048, -0.00001896, 0.00003050, 0.00000377, -0.00002745, -0.00002685, 0.00004819, -0.00002586, 0.00001733, -0.00003571, 0.00003735, -0.00002995, -0.00001122, -0.00013101, 0.00005423, -0.00003834, 0.00015411, -0.00004470, 0.00002025, -0.00047991, 0.00084010, 0.00019926, -0.00040810, 0.00009018, -0.00016955, 0.00006676, -0.00008337, 0.00000877, 0.00001054, 0.00002008, -0.00000506, -0.00000309, -0.00001643, 0.00001282, -0.00001908, 0.00000884, 0.00000654, 0.00001748, -0.00004747, 0.00001850, -0.00000547, 0.00000204, -0.00000021, -0.00000547, 0.00000452, 0.00000615, -0.00000615, 0.00001171, -0.00001456, 0.00000006, -0.00000780, },
		{ -0.00000226, 0.00000773, -0.00000887, 0.00001829, -0.00000576, 0.00000271, -0.00000710, -0.00000251, 0.00001229, -0.00000815, -0.00001241, 0.00000004, -0.00001530, 0.00002170, -0.00001830, 0.00002696, 0.00000778, -0.00002487, -0.00003327, 0.00005062, -0.00002571, 0.00001948, -0.00003640, 0.00003484, -0.00002621, -0.00001303, -0.00012664, 0.00003805, -0.00003540, 0.00015575, -0.00002137, 0.00001769, -0.00045533, 0.00071865, 0.00030446, -0.00042654, 0.00008947, -0.00016655, 0.00005374, -0.00006907, -0.00000318, 0.00001532, 0.00001766, -0.00000385, 0.00000111, -0.00002240, 0.00001555, -0.00001910, 0.00000923, 0.00000399, 0.00002125, -0.00004684, 0.00001484, -0.00000397, 0.00000092, 0.00000054, -0.00000458, 0.00000235, 0.00000770, -0.00000671, 0.00001224, -0.00001323, -0.00000168, -0.00000584, },
		{ -0.00000394, 0.00000765, -0.00000893, 0.00001674, -0.00000284, 0.00000067, -0.00000481, -0.00000433, 0.00001258, -0.00000663, -0.00001356, 0.00000103, -0.00001724, 0.00002167, -0.00001624, 0.00002244, 0.00001175, -0.00002211, -0.00003865, 0.00005105, -0.00002356, 0.00001922, -0.00003370, 0.00002867, -0.00001919, -0.00001790, -0.00011880, 0.00002012, -0.00003297, 0.00015504, 0.00000614, 0.00001102, -0.00042547, 0.00060210, 0.00039033, -0.00042643, 0.00007719, -0.00015493, 0.00003462, -0.00005146, -0.00001604, 0.00002002, 0.00001523, -0.00000293, 0.00000535, -0.00002724, 0.00001635, -0.00001751, 0.00000862, 0.00000237, 0.00002361, -0.00004472, 0.00001050, -0.00000220, -0.00000047, 0.00000144, -0.00000387, 0.00000061, 0.00000854, -0.00000662, 0.00001218, -0.00001144, -0.00000351, -0.00000405, },
		{ -0.00000506, 0.00000701, -0.00000851, 0.00001472, 0.00000017, -0.00000123, -0.00000272, -0.00000586, 0.00001260, -0.00000514, -0.00001430, 0.00000153, -0.00001847, 0.00002059, -0.00001313, 0.00001739, 0.00001532, -0.00001907, -0.00004296, 0.00004959, -0.00001966, 0.00001684, -0.00002812, 0.00001976, -0.00001014, -0.00002438, -0.00010926, 0.00000251, -0.00003271, 0.00015310, 0.00003636, 0.00000248, -0.00039362, 0.00049394, 0.00045554, -0.00040897, 0.00005530, -0.00013668, 0.00001140, -0.00003240, -0.00002836, 0.00002374, 0.00001334, -0.00000260, 0.00000963, -0.00003086, 0.00001531, -0.00001458, 0.00000728, 0.00000150, 0.00002476, -0.00004144, 0.00000578, -0.00000038, -0.00000202, 0.00000245, -0.00000332, -0.00000071, 0.00000875, -0.00000600, 0.00001167, -0.00000937, -0.00000526, -0.00000258, },
		{ -0.00000562, 0.00000592, -0.00000775, 0.00001245, 0.00000304, -0.00000284, -0.00000093, -0.00000702, 0.00001230, -0.00000366, -0.00001467, 0.00000166, -0.00001914, 0.00001870, -0.00000931, 0.00001224, 0.00001818, -0.00001568, -0.00004619, 0.00004641, -0.00001432, 0.00001276, -0.00002036, 0.00000919, -0.00000041, -0.00003100, -0.00009971, -0.00001297, -0.00003584, 0.00015063, 0.00006818, -0.00000606, -0.00036240, 0.00039660, 0.00049997, -0.00037630, 0.00002637, -0.00011412, -0.00001375, -0.00001375, -0.00003884, 0.00002576, 0.00001240, -0.00000303, 0.00001393, -0.00003320, 0.00001260, -0.00001065, 0.00000554, 0.00000113, 0.00002496, -0.00003734, 0.00000099, 0.00000133, -0.00000361, 0.00000348, -0.00000290, -0.00000164, 0.00000841, -0.00000501, 0.00001090, -0.00000722, -0.00000677, -0.00000150, },
		{ -0.00000568, 0.00000450, -0.00000681, 0.00001013, 0.00000558, -0.00000402, 0.00000047, -0.00000777, 0.00001170, -0.00000221, -0.00001474, 0.00000156, -0.00001942, 0.00001624, -0.00000514, 0.00000739, 0.00002009, -0.00001195, -0.00004836, 0.00004174, -0.00000794, 0.00000748, -0.00001124, -0.00000186, 0.00000867, -0.00003638, -0.00009160, -0.00002493, -0.00004303, 0.00014780, 0.00010087, -0.00001321, -0.00033373, 0.00031147, 0.00052465, -0.00033137, -0.00000666, -0.00008972, -0.00003871, 0.00000282, -0.00004642, 0.00002558, 0.00001264, -0.00000430, 0.00001816, -0.00003425, 0.00000845, -0.00000615, 0.00000375, 0.00000101, 0.00002447, -0.00003275, -0.00000358, 0.00000278, -0.00000516, 0.00000447, -0.00000255, -0.00000225, 0.00000767, -0.00000382, 0.00001003, -0.00000516, -0.00000792, -0.00000085, },
		{ -0.00000533, 0.00000291, -0.00000588, 0.00000793, 0.00000767, -0.00000471, 0.00000145, -0.00000811, 0.00001080, -0.00000078, -0.00001458, 0.00000136, -0.00001947, 0.00001342, -0.00000094, 0.00000316, 0.00002092, -0.00000794, -0.00004947, 0.00003585, -0.00000096, 0.00000159, -0.00000160, -0.00001228, 0.00001591, -0.00003941, -0.00008603, -0.00003252, -0.00005432, 0.00014425, 0.00013416, -0.00001805, -0.00030884, 0.00023898, 0.00053150, -0.00027759, -0.00004075, -0.00006582, -0.00006157, 0.00001594, -0.00005039, 0.00002296, 0.00001411, -0.00000636, 0.00002217, -0.00003401, 0.00000316, -0.00000148, 0.00000223, 0.00000093, 0.00002356, -0.00002798, -0.00000772, 0.00000387, -0.00000660, 0.00000535, -0.00000221, -0.00000264, 0.00000667, -0.00000261, 0.00000923, -0.00000331, -0.00000863, -0.00000062, },
		{ -0.00000470, 0.00000131, -0.00000508, 0.00000600, 0.00000920, -0.00000491, 0.00000205, -0.00000808, 0.00000964, 0.00000061, -0.00001426, 0.00000120, -0.00001943, 0.00001043, 0.00000304, -0.00000022, 0.00002063, -0.00000379, -0.00004953, 0.00002904, 0.00000616, -0.00000436, 0.00000772, -0.00002104, 0.00002036, -0.00003936, -0.00008359, -0.00003547, -0.00006910, 0.00013913, 0.00016811, -0.00002005, -0.00028832, 0.00017874, 0.00052312, -0.00021864, -0.00007301, -0.00004448, -0.00008080, 0.00002467, -0.00005041, 0.00001799, 0.00001666, -0.00000902, 0.00002580, -0.00003251, -0.00000293, 0.00000296, 0.00000125, 0.00000070, 0.00002241, -0.00002326, -0.00001126, 0.00000454, -0.00000789, 0.00000606, -0.00000182, -0.00000290, 0.00000556, -0.00000155, 0.00000862, -0.00000177, -0.00000889, -0.00000076, },
		{ -0.00000391, -0.00000017, -0.00000454, 0.00000441, 0.00001016, -0.00000465, 0.00000232, -0.00000772, 0.00000830, 0.00000193, -0.00001385, 0.00000117, -0.00001939, 0.00000742, 0.00000663, -0.00000264, 0.00001930, 0.00000032, -0.00004854, 0.00002161, 0.00001300, -0.00000983, 0.00001601, -0.00002737, 0.00002142, -0.00003591, -0.00008439, -0.00003415, -0.00008617, 0.00013123, 0.00020300, -0.00001899, -0.00027228, 0.00012975, 0.00050251, -0.00015810, -0.00010094, -0.00002727, -0.00009537, 0.00002855, -0.00004657, 0.00001101, 0.00001996, -0.00001201, 0.00002884, -0.00002981, -0.00000951, 0.00000679, 0.00000101, 0.00000022, 0.00002118, -0.00001879, -0.00001411, 0.00000479, -0.00000899, 0.00000655, -0.00000132, -0.00000313, 0.00000448, -0.00000077, 0.00000829, -0.00000059, -0.00000871, -0.00000116, },
		{ -0.00000313, -0.00000139, -0.00000433, 0.00000321, 0.00001058, -0.00000403, 0.00000235, -0.00000714, 0.00000683, 0.00000317, -0.00001340, 0.00000132, -0.00001938, 0.00000444, 0.00000971, -0.00000408, 0.00001712, 0.00000418, -0.00004652, 0.00001385, 0.00001917, -0.00001439, 0.00002270, -0.00003073, 0.00001889, -0.00002919, -0.00008800, -0.00002947, -0.00010391, 0.00011909, 0.00023920, -0.00001485, -0.00026047, 0.00009062, 0.00047280, -0.00009921, -0.00012260, -0.00001519, -0.00010479, 0.00002763, -0.00003935, 0.00000264, 0.00002356, -0.00001500, 0.00003107, -0.00002598, -0.00001624, 0.00000974, 0.00000162, -0.00000055, 0.00001995, -0.00001466, -0.00001626, 0.00000466, -0.00000992, 0.00000680, -0.00000066, -0.00000340, 0.00000355, -0.00000035, 0.00000827, 0.00000025, -0.00000819, -0.00000171, },
		{ -0.00000246, -0.00000228, -0.00000447, 0.00000237, 0.00001055, -0.00000317, 0.00000226, -0.00000642, 0.00000533, 0.00000428, -0.00001292, 0.00000169, -0.00001940, 0.00000152, 0.00001226, -0.00000463, 0.00001432, 0.00000752, -0.00004350, 0.00000604, 0.00002436, -0.00001774, 0.00002743, -0.00003092, 0.00001295, -0.00001979, -0.00009357, -0.00002277, -0.00012039, 0.00010129, 0.00027698, -0.00000758, -0.00025248, 0.00005981, 0.00043694, -0.00004468, -0.00013679, -0.00000856, -0.00010915, 0.00002240, -0.00002956, -0.00000634, 0.00002691, -0.00001760, 0.00003228, -0.00002112, -0.00002284, 0.00001161, 0.00000309, -0.00000156, 0.00001872, -0.00001091, -0.00001775, 0.00000424, -0.00001070, 0.00000679, 0.00000019, -0.00000377, 0.00000284, -0.00000032, 0.00000854, 0.00000081, -0.00000741, -0.00000227, },
		{ -0.00000202, -0.00000278, -0.00000493, 0.00000182, 0.00001018, -0.00000223, 0.00000217, -0.00000565, 0.00000385, 0.00000523, -0.00001243, 0.00000224, -0.00001936, -0.00000138, 0.00001433, -0.00000443, 0.00001121, 0.00001012, -0.00003952, -0.00000158, 0.00002835, -0.00001969, 0.00003005, -0.00002806, 0.00000416, -0.00000866, -0.00009988, -0.00001567, -0.00013358, 0.00007655, 0.00031635, 0.00000296, -0.00024790, 0.00003577, 0.00039759, 0.00000348, -0.00014303, -0.00000709, -0.00010906, 0.00001379, -0.00001824, -0.00001504, 0.00002944, -0.00001945, 0.00003230, -0.00001533, -0.00002905, 0.00001228, 0.00000531, -0.00000267, 0.00001744, -0.00000750, -0.00001867, 0.00000362, -0.00001135, 0.00000651, 0.00000124, -0.00000428, 0.00000239, -0.00000068, 0.00000903, 0.00000117, -0.00000652, -0.00000271, },
		},
		{ { 0.00001096, 0.00000793, -0.00002534, -0.00000191, -0.00001188, 0.00000586, 0.00000626, 0.00000726, 0.00000064, -0.00000647, 0.00001556, -0.00001366, 0.00002528, -0.00001713, 0.00002059, -0.00001435, -0.00000893, -0.00000234, 0.00001859, 0.00001173, -0.00007422, 0.00001623, -0.00003617, -0.00001447, -0.00003021, -0.00002757, -0.00010580, 0.00003664, 0.00036791, 0.00005238, -0.00013401, -0.00003250, -0.00005222, -0.00003250, -0.00013401, 0.00005238, 0.00036791, 0.00003664, -0.00010580, -0.00002757, -0.00003021, -0.00001447, -0.00003617, 0.00001623, -0.00007422, 0.00001173, 0.00001859, -0.00000234, -0.00000893, -0.00001435, 0.00002059, -0.00001713, 0.00002528, -0.00001366, 0.00001556, -0.00000647, 0.00000064, 0.00000726, 0.00000626, 0.00000586, -0.00001188, -0.00000191, -0.00002534, 0.00000793, },
		{ 0.00001138, 0.00001034, -0.00002135, -0.00000682, -0.00000878, 0.00000118, 0.00000850, 0.00000398, 0.00000698, -0.00001180, 0.00001793, -0.00001589, 0.00002706, -0.00001585, 0.00001536, -0.00000516, -0.00001538, 0.00000443, 0.00000475, 0.00002728, -0.00007757, 0.00001183, -0.00003224, -0.00001832, -0.00002682, -0.00002632, -0.00009583, -0.00002994, 0.00039731, 0.00010451, -0.00015225, -0.00003390, -0.00005205, -0.00003211, -0.00011755, 0.00001395, 0.00032777, 0.00009316, -0.00010178, -0.00003491, -0.00003049, -0.00001328, -0.00003732, 0.00001685, -0.00006621, -0.00000612, 0.00003088, -0.00000615, -0.00000479, -0.00002065, 0.00002198, -0.00001510, 0.00002128, -0.00000990, 0.00001153, 0.00000005, -0.00000536, 0.00000953, 0.00000430, 0.00001032, -0.00001446, 0.00000202, -0.00002713, 0.00000371, },
		{ 0.00001249, 0.00001108, -0.00001568, -0.00001207, -0.00000562, -0.00000337, 0.00001071, 0.00000028, 0.00001263, -0.00001506, 0.00001818, -0.00001623, 0.00002629, -0.00001124, 0.00000693, 0.00000572, -0.00002262, 0.00001197, -0.00000771, 0.00003769, -0.00007437, 0.00000248, -0.00002435, -0.00002626, -0.00001859, -0.00003282, -0.00007278, -0.00010057, 0.00040915, 0.00017231, -0.00017141, -0.00003677, -0.00005164, -0.00003219, -0.00010349, -0.00001298, 0.00028241, 0.00013652, -0.00008573, -0.00004486, -0.00003037, -0.00001277, -0.00003731, 0.00001535, -0.00005584, -0.00002333, 0.00003907, -0.00000557, -0.00000370, -0.00002360, 0.00001959, -0.00001034, 0.00001578, -0.00000525, 0.00000655, 0.00000680, -0.00001015, 0.00001041, 0.00000272, 0.00001433, -0.00001614, 0.00000452, -0.00002647, -0.00000218, },
		{ 0.00001393, 0.00001057, -0.00000904, -0.00001695, -0.00000293, -0.00000734, 0.00001244, -0.00000305, 0.00001662, -0.00001562, 0.00001613, -0.00001465, 0.00002304, -0.00000383, -0.00000346, 0.00001658, -0.00002876, 0.00001793, -0.00001610, 0.00004093, -0.00006382, -0.00001193, -0.00001243, -0.00003842, -0.00000554, -0.00004624, -0.00004075, -0.00016675, 0.00039624, 0.00025685, -0.00018988, -0.00004171, -0.00005117, -0.00003217, -0.00009226, -0.00003039, 0.00023575, 0.00016625, -0.00006161, -0.00005330, -0.00003241, -0.00001122, -0.00003747, 0.00001317, -0.00004514, -0.00003762, 0.00004167, -0.00000029, -0.00000544, -0.00002357, 0.00001420, -0.00000394, 0.00000979, -0.00000050, 0.00000145, 0.00001284, -0.00001310, 0.00000985, 0.00000145, 0.00001773, -0.00001666, 0.00000530, -0.00002339, -0.00000933, },
		{ 0.00001514, 0.00000937, -0.00000225, -0.00002073, -0.00000125, -0.00001024, 0.00001314, -0.00000521, 0.00001817, -0.00001326, 0.00001201, -0.00001147, 0.00001777, 0.00000544, -0.00001426, 0.00002560, -0.00003202, 0.00002040, -0.00001867, 0.00003637, -0.00004664, -0.00003005, 0.00000218, -0.00005331, 0.00001030, -0.00006318, -0.00000649, -0.00021874, 0.00035244, 0.00035711, -0.00020479, -0.00004974, -0.00005063, -0.00003171, -0.00008402, -0.00004008, 0.00019038, 0.00018368, -0.00003403, -0.00005674, -0.00003813, -0.00000799, -0.00003829, 0.00001103, -0.00003534, -0.00004777, 0.00003847, 0.00000879, -0.00000886, -0.00002161, 0.00000710, 0.00000273, 0.00000442, 0.00000351, -0.00000300, 0.00001742, -0.00001388, 0.00000805, 0.00000033, 0.00002040, -0.00001582, 0.00000427, -0.00001821, -0.00001709, },
		{ 0.00001551, 0.00000809, 0.00000388, -0.00002273, -0.00000103, -0.00001161, 0.00001230, -0.00000559, 0.00001685, -0.00000817, 0.00000641, -0.00000727, 0.00001119, 0.00001537, -0.00002385, 0.00003122, -0.00003120, 0.00001839, -0.00001509, 0.00002504, -0.00002520, -0.00004913, 0.00001676, -0.00006792, 0.00002517, -0.00007840, 0.00002189, -0.00024734, 0.00027444, 0.00046899, -0.00021147, -0.00006271, -0.00004946, -0.00003098, -0.00007838, -0.00004395, 0.00014809, 0.00019085, -0.00000699, -0.00005339, -0.00004724, -0.00000383, -0.00003913, 0.00000873, -0.00002671, -0.00005377, 0.00003057, 0.00001976, -0.00001220, -0.00001919, -0.00000019, 0.00000827, 0.00000066, 0.00000610, -0.00000617, 0.00002005, -0.00001251, 0.00000541, -0.00000083, 0.00002220, -0.00001350, 0.00000153, -0.00001146, -0.00002464, },
		{ 0.00001446, 0.00000730, 0.00000866, -0.00002250, -0.00000257, -0.00001118, 0.00000964, -0.00000386, 0.00001273, -0.00000105, 0.00000023, -0.00000279, 0.00000414, 0.00002479, -0.00003088, 0.00003249, -0.00002598, 0.00001212, -0.00000651, 0.00000958, -0.00000312, -0.00006546, 0.00002782, -0.00007844, 0.00003454, -0.00008608, 0.00003674, -0.00024598, 0.00016335, 0.00058447, -0.00020327, -0.00008333, -0.00004643, -0.00003090, -0.00007426, -0.00004419, 0.00011042, 0.00018957, 0.00001706, -0.00004364, -0.00005764, -0.00000082, -0.00003842, 0.00000528, -0.00001864, -0.00005662, 0.00001999, 0.00003016, -0.00001352, -0.00001777, -0.00000624, 0.00001149, -0.00000075, 0.00000683, -0.00000771, 0.00002054, -0.00000927, 0.00000242, -0.00000216, 0.00002297, -0.00000967, -0.00000266, -0.00000385, -0.00003111, },
		{ 0.00001158, 0.00000737, 0.00001164, -0.00001983, -0.00000588, -0.00000892, 0.00000517, -0.00000015, 0.00000640, 0.00000701, -0.00000549, 0.00000120, -0.00000259, 0.00003269, -0.00003449, 0.00002925, -0.00001704, 0.00000294, 0.00000465, -0.00000637, 0.00001541, -0.00007525, 0.00003192, -0.00008124, 0.00003426, -0.00008138, 0.00003269, -0.00021246, 0.00002573, 0.00069161, -0.00017199, -0.00011480, -0.00003978, -0.00003303, -0.00006991, -0.00004338, 0.00007903, 0.00018094, 0.00003755, -0.00002996, -0.00006583, -0.00000175, -0.00003425, -0.00000057, -0.00001014, -0.00005784, 0.00000908, 0.00003762, -0.00001126, -0.00001843, -0.00001003, 0.00001169, 0.00000046, 0.00000563, -0.00000755, 0.00001902, -0.00000467, -0.00000042, -0.00000363, 0.00002248, -0.00000436, -0.00000789, 0.00000379, -0.00003571, },
		{ 0.00000669, 0.00000843, 0.00001266, -0.00001488, -0.00001069, -0.00000513, -0.00000071, 0.00000495, -0.00000112, 0.00001467, -0.00000969, 0.00000403, -0.00000835, 0.00003837, -0.00003437, 0.00002210, -0.00000589, -0.00000693, 0.00001530, -0.00001894, 0.00002660, -0.00007555, 0.00002671, -0.00007398, 0.00002190, -0.00006192, 0.00000822, -0.00015013, -0.00012626, 0.00077539, -0.00010900, -0.00016003, -0.00002781, -0.00003915, -0.00006323, -0.00004422, 0.00005562, 0.00016524, 0.00005552, -0.00001612, -0.00006793, -0.00000916, -0.00002508, -0.00000974, -0.00000042, -0.00005882, -0.00000010, 0.00004042, -0.00000466, -0.00002148, -0.00001122, 0.00000879, 0.00000405, 0.00000279, -0.00000594, 0.00001589, 0.00000066, -0.00000271, -0.00000512, 0.00002053, 0.00000223, -0.00001360, 0.00001066, -0.00003781, },
		{ -0.00000006, 0.00001032, 0.00001191, -0.00000812, -0.00001638, -0.00000046, -0.00000722, 0.00001048, -0.00000846, 0.00002052, -0.00001148, 0.00000527, -0.00001273, 0.00004151, -0.00003078, 0.00001228, 0.00000543, -0.00001502, 0.00002247, -0.00002497, 0.00002805, -0.00006517, 0.00001177, -0.00005644, -0.00000236, -0.00002873, -0.00003361, -0.00006788, -0.00027602, 0.00081960, -0.00000698, -0.00022030, -0.00000974, -0.00005055, -0.00005249, -0.00004903, 0.00004154, 0.00014244, 0.00007276, -0.00000601, -0.00006089, -0.00002439, -0.00001062, -0.00002215, 0.00001063, -0.00006035, -0.00000635, 0.00003797, 0.00000593, -0.00002641, -0.00001016, 0.00000331, 0.00000930, -0.00000102, -0.00000343, 0.00001175, 0.00000603, -0.00000414, -0.00000640, 0.00001699, 0.00000976, -0.00001914, 0.00001598, -0.00003703, },
		{ -0.00000822, 0.00001262, 0.00000983, -0.00000032, -0.00002209, 0.00000420, -0.00001334, 0.00001520, -0.00001420, 0.00002335, -0.00001033, 0.00000481, -0.00001561, 0.00004211, -0.00002438, 0.00000134, 0.00001489, -0.00001923, 0.00002402, -0.00002282, 0.00001942, -0.00004524, -0.00001096, -0.00003099, -0.00003536, 0.00001358, -0.00008535, 0.00002097, -0.00040460, 0.00080943, 0.00013780, -0.00029385, 0.00001325, -0.00006719, -0.00003704, -0.00005897, 0.00003717, 0.00011291, 0.00009079, -0.00000240, -0.00004360, -0.00004671, 0.00000774, -0.00003639, 0.00002204, -0.00006220, -0.00000963, 0.00003093, 0.00001907, -0.00003191, -0.00000780, -0.00000369, 0.00001513, -0.00000493, -0.00000071, 0.00000733, 0.00001078, -0.00000456, -0.00000721, 0.00001189, 0.00001772, -0.00002378, 0.00001914, -0.00003331, },
		{ -0.00001706, 0.00001470, 0.00000710, 0.00000757, -0.00002682, 0.00000783, -0.00001793, 0.00001791, -0.00001709, 0.00002236, -0.00000616, 0.00000283, -0.00001707, 0.00004042, -0.00001605, -0.00000918, 0.00002089, -0.00001835, 0.00001928, -0.00001286, 0.00000270, -0.00001924, -0.00003730, -0.00000229, -0.00007138, 0.00005749, -0.00013651, 0.00010081, -0.00049377, 0.00073450, 0.00032346, -0.00037467, 0.00003731, -0.00008705, -0.00001799, -0.00007354, 0.00004137, 0.00007817, 0.00010985, -0.00000593, -0.00001755, -0.00007308, 0.00002677, -0.00004971, 0.00003184, -0.00006320, -0.00001102, 0.00002114, 0.00003250, -0.00003617, -0.00000549, -0.00001087, 0.00002026, -0.00000795, 0.00000144, 0.00000336, 0.00001435, -0.00000393, -0.00000732, 0.00000546, 0.00002540, -0.00002681, 0.00001972, -0.00002690, },
		{ -0.00002568, 0.00001587, 0.00000443, 0.00001457, -0.00002961, 0.00000949, -0.00001998, 0.00001761, -0.00001629, 0.00001732, 0.00000058, -0.00000014, -0.00001737, 0.00003682, -0.00000669, -0.00001811, 0.00002263, -0.00001242, 0.00000925, 0.00000249, -0.00001801, 0.00000755, -0.00006152, 0.00002355, -0.00010320, 0.00009408, -0.00017560, 0.00015667, -0.00052933, 0.00059166, 0.00054086, -0.00045188, 0.00005561, -0.00010587, 0.00000158, -0.00009026, 0.00005114, 0.00004128, 0.00012832, -0.00001467, 0.00001310, -0.00009840, 0.00004192, -0.00005854, 0.00003743, -0.00006151, -0.00001238, 0.00001107, 0.00004360, -0.00003732, -0.00000457, -0.00001686, 0.00002347, -0.00000917, 0.00000232, 0.00000049, 0.00001632, -0.00000231, -0.00000664, -0.00000182, 0.00003197, -0.00002760, 0.00001758, -0.00001838, },
		{ -0.00003310, 0.00001549, 0.00000248, 0.00001983, -0.00002970, 0.00000847, -0.00001884, 0.00001378, -0.00001160, 0.00000868, 0.00000897, -0.00000339, -0.00001686, 0.00003173, 0.00000299, -0.00002477, 0.00002029, -0.00000281, -0.00000352, 0.00001926, -0.00003735, 0.00002906, -0.00007760, 0.00004033, -0.00012372, 0.00011494, -0.00019256, 0.00017709, -0.00050396, 0.00038687, 0.00077315, -0.00051016, 0.00005892, -0.00011758, 0.00001687, -0.00010491, 0.00006187, 0.00000677, 0.00014283, -0.00002444, 0.00004222, -0.00011644, 0.00004825, -0.00005925, 0.00003629, -0.00005525, -0.00001572, 0.00000331, 0.00005002, -0.00003391, -0.00000608, -0.00002056, 0.00002381, -0.00000794, 0.00000143, -0.00000081, 0.00001647, 0.00000017, -0.00000520, -0.00000929, 0.00003658, -0.00002568, 0.00001286, -0.00000856, },
		{ -0.00003844, 0.00001310, 0.00000170, 0.00002279, -0.00002668, 0.00000451, -0.00001434, 0.00000652, -0.00000352, -0.00000245, 0.00001769, -0.00000611, -0.00001586, 0.00002550, 0.00001248, -0.00002910, 0.00001492, 0.00000817, -0.00001557, 0.00003279, -0.00004983, 0.00003984, -0.00008060, 0.00004327, -0.00012764, 0.00011417, -0.00018113, 0.00015668, -0.00041927, 0.00013579, 0.00099675, -0.00053135, 0.00003707, -0.00011547, 0.00002228, -0.00011238, 0.00006807, -0.00002003, 0.00014888, -0.00002976, 0.00006300, -0.00012128, 0.00004165, -0.00004924, 0.00002679, -0.00004317, -0.00002261, -0.00000019, 0.00005023, -0.00002532, -0.00001039, -0.00002136, 0.00002080, -0.00000399, -0.00000146, -0.00000033, 0.00001478, 0.00000329, -0.00000316, -0.00001615, 0.00003847, -0.00002088, 0.00000603, 0.00000158, },
		{ -0.00004100, 0.00000858, 0.00000221, 0.00002324, -0.00002061, -0.00000214, -0.00000695, -0.00000340, 0.00000676, -0.00001445, 0.00002527, -0.00000753, -0.00001463, 0.00001840, 0.00002153, -0.00003148, 0.00000815, 0.00001770, -0.00002328, 0.00003882, -0.00005117, 0.00003634, -0.00006804, 0.00003034, -0.00011289, 0.00008996, -0.00014043, 0.00009746, -0.00028618, -0.00013740, 0.00118380, -0.00049718, -0.00001882, -0.00009384, 0.00001290, -0.00010792, 0.00006456, -0.00003410, 0.00014211, -0.00002534, 0.00006959, -0.00010874, 0.00002007, -0.00002784, 0.00000888, -0.00002527, -0.00003353, 0.00000139, 0.00004394, -0.00001201, -0.00001715, -0.00001927, 0.00001448, 0.00000249, -0.00000617, 0.00000183, 0.00001148, 0.00000680, -0.00000077, -0.00002162, 0.00003711, -0.00001337, -0.00000218, 0.00001101, },
		{ -0.00004041, 0.00000214, 0.00000382, 0.00002139, -0.00001201, -0.00001071, 0.00000233, -0.00001463, 0.00001758, -0.00002544, 0.00003023, -0.00000706, -0.00001333, 0.00001067, 0.00003001, -0.00003254, 0.00000175, 0.00002323, -0.00002379, 0.00003461, -0.00003939, 0.00001798, -0.00004070, 0.00000304, -0.00008136, 0.00004526, -0.00007553, 0.00000884, -0.00012369, -0.00040287, 0.00130593, -0.00039269, -0.00011377, -0.00005001, -0.00001396, -0.00008864, 0.00004785, -0.00003209, 0.00011965, -0.00000761, 0.00005866, -0.00007772, -0.00001554, 0.00000311, -0.00001546, -0.00000313, -0.00004760, 0.00000748, 0.00003227, 0.00000448, -0.00002521, -0.00001489, 0.00000545, 0.00001087, -0.00001221, 0.00000525, 0.00000704, 0.00001037, 0.00000171, -0.00002505, 0.00003232, -0.00000367, -0.00001083, 0.00001875, },
		{ -0.00003667, -0.00000566, 0.00000602, 0.00001782, -0.00000184, -0.00002002, 0.00001207, -0.00002546, 0.00002699, -0.00003353, 0.00003140, -0.00000441, -0.00001200, 0.00000257, 0.00003776, -0.00003285, -0.00000285, 0.00002305, -0.00001580, 0.00001965, -0.00001545, -0.00001245, -0.00000286, -0.00003357, -0.00003886, -0.00001248, 0.00000329, -0.00009385, 0.00004391, -0.00062925, 0.00133880, -0.00020995, -0.00024641, 0.00001402, -0.00005732, -0.00005471, 0.00001735, -0.00001346, 0.00008137, 0.00002397, 0.00003051, -0.00003093, -0.00006094, 0.00003916, -0.00004233, 0.00002017, -0.00006262, 0.00001616, 0.00001752, 0.00002173, -0.00003292, -0.00000930, -0.00000522, 0.00002011, -0.00001871, 0.00000923, 0.00000211, 0.00001357, 0.00000408, -0.00002606, 0.00002438, 0.00000729, -0.00001885, 0.00002403, },
		{ -0.00003016, -0.00001396, 0.00000808, 0.00001333, 0.00000871, -0.00002862, 0.00002061, -0.00003407, 0.00003318, -0.00003718, 0.00002810, 0.00000037, -0.00001061, -0.00000557, 0.00004452, -0.00003270, -0.00000494, 0.00001682, 0.00000003, -0.00000395, 0.00001667, -0.00004908, 0.00003830, -0.00007169, 0.00000597, -0.00007269, 0.00008236, -0.00019212, 0.00019102, -0.00078851, 0.00126657, 0.00004891, -0.00040734, 0.00009050, -0.00011179, -0.00001014, -0.00002399, 0.00001884, 0.00003055, 0.00006629, -0.00001054, 0.00002519, -0.00010898, 0.00007385, -0.00006644, 0.00004057, -0.00007545, 0.00002460, 0.00000278, 0.00003694, -0.00003843, -0.00000376, -0.00001621, 0.00002894, -0.00002462, 0.00001289, -0.00000250, 0.00001595, 0.00000625, -0.00002462, 0.00001399, 0.00001835, -0.00002519, 0.00002634, },
		{ -0.00002160, -0.00002172, 0.00000924, 0.00000881, 0.00001837, -0.00003511, 0.00002645, -0.00003889, 0.00003472, -0.00003548, 0.00002026, 0.00000683, -0.00000910, -0.00001327, 0.00004980, -0.00003195, -0.00000469, 0.00000565, 0.00002107, -0.00003181, 0.00005063, -0.00008404, 0.00007409, -0.00010231, 0.00004336, -0.00012373, 0.00014733, -0.00026779, 0.00029509, -0.00086060, 0.00108556, 0.00036989, -0.00057870, 0.00016613, -0.00016775, 0.00003745, -0.00006975, 0.00005851, -0.00002624, 0.00011282, -0.00005674, 0.00008142, -0.00015080, 0.00009986, -0.00008214, 0.00005379, -0.00008276, 0.00002967, -0.00000874, 0.00004745, -0.00004014, 0.00000049, -0.00002614, 0.00003602, -0.00002881, 0.00001524, -0.00000596, 0.00001704, 0.00000828, -0.00002103, 0.00000221, 0.00002819, -0.00002895, 0.00002552, },
		{ -0.00001197, -0.00002793, 0.00000882, 0.00000504, 0.00002603, -0.00003832, 0.00002841, -0.00003887, 0.00003090, -0.00002830, 0.00000853, 0.00001423, -0.00000740, -0.00001994, 0.00005293, -0.00003001, -0.00000314, -0.00000808, 0.00004319, -0.00005805, 0.00007898, -0.00010896, 0.00009603, -0.00011718, 0.00006459, -0.00015538, 0.00018604, -0.00030647, 0.00034053, -0.00083699, 0.00080618, 0.00072719, -0.00073547, 0.00022354, -0.00021268, 0.00007789, -0.00011109, 0.00009683, -0.00008011, 0.00015471, -0.00009830, 0.00012742, -0.00017744, 0.00011044, -0.00008463, 0.00005638, -0.00008181, 0.00002877, -0.00001443, 0.00005132, -0.00003703, 0.00000257, -0.00003378, 0.00004013, -0.00003024, 0.00001542, -0.00000750, 0.00001640, 0.00001031, -0.00001589, -0.00000964, 0.00003554, -0.00002953, 0.00002180, },
		{ -0.00000235, -0.00003172, 0.00000643, 0.00000254, 0.00003093, -0.00003755, 0.00002594, -0.00003372, 0.00002184, -0.00001637, -0.00000582, 0.00002156, -0.00000548, -0.00002493, 0.00005314, -0.00002605, -0.00000190, -0.00002123, 0.00006161, -0.00007654, 0.00009473, -0.00011683, 0.00009780, -0.00011068, 0.00006395, -0.00016108, 0.00019109, -0.00030044, 0.00032156, -0.00072253, 0.00045270, 0.00108541, -0.00084844, 0.00024402, -0.00023343, 0.00010046, -0.00013850, 0.00012438, -0.00012154, 0.00018261, -0.00012537, 0.00015376, -0.00018177, 0.00010101, -0.00007118, 0.00004662, -0.00007127, 0.00002043, -0.00001285, 0.00004780, -0.00002899, 0.00000209, -0.00003827, 0.00004037, -0.00002819, 0.00001277, -0.00000661, 0.00001379, 0.00001255, -0.00000997, -0.00002023, 0.00003939, -0.00002672, 0.00001575, },
		{ 0.00000618, -0.00003255, 0.00000205, 0.00000147, 0.00003276, -0.00003271, 0.00001920, -0.00002393, 0.00000853, -0.00000121, -0.00002105, 0.00002773, -0.00000336, -0.00002762, 0.00004981, -0.00001925, -0.00000268, -0.00003063, 0.00007192, -0.00008222, 0.00009287, -0.00010362, 0.00007676, -0.00008134, 0.00004026, -0.00013941, 0.00016145, -0.00025015, 0.00024322, -0.00053500, 0.00006076, 0.00140370, -0.00088855, 0.00021118, -0.00021916, 0.00009630, -0.00014397, 0.00013304, -0.00014238, 0.00018870, -0.00013022, 0.00015402, -0.00016013, 0.00007040, -0.00004196, 0.00002515, -0.00005165, 0.00000481, -0.00000407, 0.00003748, -0.00001687, -0.00000071, -0.00003921, 0.00003632, -0.00002234, 0.00000704, -0.00000309, 0.00000918, 0.00001511, -0.00000407, -0.00002838, 0.00003912, -0.00002075, 0.00000823, },
		{ 0.00001275, -0.00003028, -0.00000393, 0.00000157, 0.00003171, -0.00002431, 0.00000906, -0.00001080, -0.00000730, 0.00001513, -0.00003516, 0.00003167, -0.00000107, -0.00002759, 0.00004264, -0.00000915, -0.00000680, -0.00003386, 0.00007113, -0.00007230, 0.00007163, -0.00006937, 0.00003487, -0.00003247, -0.00000264, -0.00009464, 0.00010277, -0.00016422, 0.00012044, -0.00030253, -0.00032732, 0.00164137, -0.00083200, 0.00011482, -0.00016423, 0.00006082, -0.00012311, 0.00011805, -0.00013782, 0.00016866, -0.00010909, 0.00012627, -0.00011346, 0.00002142, -0.00000039, -0.00000489, -0.00002545, -0.00001625, 0.00001025, 0.00002228, -0.00000239, -0.00000508, -0.00003668, 0.00002807, -0.00001289, -0.00000153, 0.00000281, 0.00000287, 0.00001797, 0.00000117, -0.00003328, 0.00003466, -0.00001234, 0.00000029, },
		{ 0.00001676, -0.00002523, -0.00001072, 0.00000227, 0.00002840, -0.00001341, -0.00000305, 0.00000386, -0.00002349, 0.00003032, -0.00004622, 0.00003253, 0.00000131, -0.00002473, 0.00003184, 0.00000412, -0.00001473, -0.00002983, 0.00005834, -0.00004698, 0.00003308, -0.00001849, -0.00002144, 0.00002817, -0.00005642, -0.00003590, 0.00002634, -0.00005779, -0.00002467, -0.00005920, -0.00066843, 0.00176406, -0.00066512, -0.00004575, -0.00007051, -0.00000453, -0.00007673, 0.00007939, -0.00010770, 0.00012296, -0.00006328, 0.00007386, -0.00004747, -0.00003925, 0.00004722, -0.00003829, 0.00000324, -0.00003949, 0.00002719, 0.00000510, 0.00001216, -0.00000989, -0.00003124, 0.00001628, -0.00000061, -0.00001220, 0.00001039, -0.00000450, 0.00002087, 0.00000537, -0.00003459, 0.00002651, -0.00000255, -0.00000700, },
		{ 0.00001798, -0.00001812, -0.00001729, 0.00000279, 0.00002373, -0.00000139, -0.00001535, 0.00001794, -0.00003774, 0.00004215, -0.00005261, 0.00002979, 0.00000372, -0.00001935, 0.00001827, 0.00001962, -0.00002587, -0.00001914, 0.00003511, -0.00000964, -0.00001705, 0.00004086, -0.00008223, 0.00008973, -0.00010982, 0.00002454, -0.00005311, 0.00005040, -0.00016686, 0.00016042, -0.00092500, 0.00174943, -0.00038818, -0.00025968, 0.00005156, -0.00009155, -0.00001136, 0.00002232, -0.00005702, 0.00005716, 0.00000077, 0.00000502, 0.00002808, -0.00010204, 0.00009274, -0.00006859, 0.00002943, -0.00006091, 0.00004317, -0.00001077, 0.00002437, -0.00001390, -0.00002375, 0.00000207, 0.00001324, -0.00002375, 0.00001855, -0.00001198, 0.00002333, 0.00000848, -0.00003252, 0.00001566, 0.00000731, -0.00001265, },
		{ 0.00001659, -0.00000998, -0.00002252, 0.00000236, 0.00001869, 0.00001025, -0.00002604, 0.00002946, -0.00004803, 0.00004884, -0.00005330, 0.00002335, 0.00000611, -0.00001214, 0.00000338, 0.00003570, -0.00003856, -0.00000400, 0.00000521, 0.00003376, -0.00007035, 0.00009816, -0.00013591, 0.00014034, -0.00015090, 0.00007387, -0.00012039, 0.00014162, -0.00028230, 0.00032680, -0.00107069, 0.00159129, -0.00001720, -0.00050456, 0.00018362, -0.00018607, 0.00006145, -0.00004333, 0.00000481, -0.00001888, 0.00007249, -0.00006846, 0.00010105, -0.00015623, 0.00012765, -0.00008944, 0.00004828, -0.00007657, 0.00005466, -0.00002222, 0.00003216, -0.00001605, -0.00001523, -0.00001311, 0.00002704, -0.00003466, 0.00002591, -0.00001843, 0.00002473, 0.00001074, -0.00002773, 0.00000350, 0.00001588, -0.00001596, },
		{ 0.00001312, -0.00000196, -0.00002540, 0.00000036, 0.00001411, 0.00002019, -0.00003361, 0.00003682, -0.00005290, 0.00004938, -0.00004798, 0.00001356, 0.00000845, -0.00000415, -0.00001094, 0.00005018, -0.00005033, 0.00001218, -0.00002611, 0.00007572, -0.00011725, 0.00014256, -0.00017150, 0.00016950, -0.00016957, 0.00010135, -0.00016292, 0.00020074, -0.00035277, 0.00041999, -0.00109421, 0.00130143, 0.00041669, -0.00074845, 0.00030175, -0.00027008, 0.00012704, -0.00010486, 0.00006562, -0.00009273, 0.00013915, -0.00013345, 0.00015908, -0.00019187, 0.00014475, -0.00009581, 0.00005612, -0.00008344, 0.00005890, -0.00002696, 0.00003420, -0.00001568, -0.00000670, -0.00002763, 0.00003904, -0.00004329, 0.00003103, -0.00002270, 0.00002436, 0.00001262, -0.00002121, -0.00000843, 0.00002195, -0.00001654, },
		{ 0.00000836, 0.00000480, -0.00002527, -0.00000346, 0.00001057, 0.00002747, -0.00003705, 0.00003911, -0.00005173, 0.00004361, -0.00003716, 0.00000124, 0.00001068, 0.00000339, -0.00002269, 0.00006077, -0.00005847, 0.00002554, -0.00005311, 0.00010866, -0.00014891, 0.00016503, -0.00018089, 0.00017031, -0.00015978, 0.00010069, -0.00017317, 0.00021902, -0.00036872, 0.00043254, -0.00100093, 0.00090866, 0.00087052, -0.00095406, 0.00038028, -0.00032501, 0.00017044, -0.00014925, 0.00011290, -0.00015182, 0.00018834, -0.00017796, 0.00019199, -0.00020180, 0.00013966, -0.00008520, 0.00005127, -0.00008006, 0.00005450, -0.00002396, 0.00003017, -0.00001267, 0.00000099, -0.00003994, 0.00004757, -0.00004812, 0.00003265, -0.00002381, 0.00002165, 0.00001463, -0.00001410, -0.00001866, 0.00002467, -0.00001445, },
		{ 0.00000327, 0.00000938, -0.00002187, -0.00000891, 0.00000822, 0.00003163, -0.00003600, 0.00003623, -0.00004477, 0.00003229, -0.00002207, -0.00001245, 0.00001272, 0.00000931, -0.00003010, 0.00006553, -0.00006060, 0.00003257, -0.00007079, 0.00012645, -0.00015895, 0.00016015, -0.00016055, 0.00014105, -0.00012093, 0.00007138, -0.00015019, 0.00019563, -0.00033067, 0.00037010, -0.00081165, 0.00045515, 0.00129491, -0.00108425, 0.00039650, -0.00033549, 0.00017956, -0.00016598, 0.00013641, -0.00018597, 0.00021044, -0.00019351, 0.00019377, -0.00018314, 0.00011187, -0.00005823, 0.00003449, -0.00006691, 0.00004172, -0.00001364, 0.00002086, -0.00000749, 0.00000724, -0.00004872, 0.00005130, -0.00004806, 0.00002994, -0.00002119, 0.00001630, 0.00001721, -0.00000742, -0.00002604, 0.00002366, -0.00001017, },
		{ -0.00000120, 0.00001124, -0.00001552, -0.00001536, 0.00000680, 0.00003275, -0.00003082, 0.00002886, -0.00003314, 0.00001694, -0.00000452, -0.00002609, 0.00001439, 0.00001266, -0.00003198, 0.00006333, -0.00005525, 0.00003097, -0.00007590, 0.00012562, -0.00014478, 0.00012737, -0.00011248, 0.00008572, -0.00005822, 0.00001887, -0.00009962, 0.00013752, -0.00024855, 0.00024976, -0.00055905, -0.00000943, 0.00164089, -0.00110823, 0.00033535, -0.00029291, 0.00014799, -0.00014945, 0.00013046, -0.00018953, 0.00020067, -0.00017686, 0.00016382, -0.00013804, 0.00006498, -0.00001869, 0.00000894, -0.00004634, 0.00002239, 0.00000224, 0.00000799, -0.00000110, 0.00001174, -0.00005306, 0.00004947, -0.00004263, 0.00002264, -0.00001479, 0.00000842, 0.00002052, -0.00000195, -0.00002990, 0.00001909, -0.00000449, },
		{ -0.00000424, 0.00001026, -0.00000699, -0.00002183, 0.00000577, 0.00003135, -0.00002242, 0.00001833, -0.00001854, -0.00000039, 0.00001338, -0.00003819, 0.00001546, 0.00001297, -0.00002800, 0.00005416, -0.00004233, 0.00002015, -0.00006762, 0.00010609, -0.00010815, 0.00007118, -0.00004397, 0.00001342, 0.00001830, -0.00004645, -0.00003249, 0.00005779, -0.00013950, 0.00009653, -0.00028220, -0.00043492, 0.00186704, -0.00100708, 0.00019316, -0.00019794, 0.00007697, -0.00010059, 0.00009527, -0.00016267, 0.00016010, -0.00013073, 0.00010719, -0.00007336, 0.00000621, 0.00002704, -0.00002042, -0.00002208, -0.00000044, 0.00002094, -0.00000609, 0.00000526, 0.00001447, -0.00005256, 0.00004198, -0.00003208, 0.00001125, -0.00000522, -0.00000140, 0.00002438, 0.00000196, -0.00003011, 0.00001168, 0.00000153, },
		{ -0.00000532, 0.00000678, 0.00000253, -0.00002712, 0.00000438, 0.00002825, -0.00001213, 0.00000637, -0.00000306, -0.00001744, 0.00002950, -0.00004739, 0.00001560, 0.00001029, -0.00001884, 0.00003926, -0.00002317, 0.00000145, -0.00004777, 0.00007119, -0.00005488, 0.00000042, 0.00003368, -0.00006328, 0.00009557, -0.00011140, 0.00003717, -0.00002708, -0.00002423, -0.00006145, -0.00002032, -0.00077779, 0.00194566, -0.00077779, -0.00002032, -0.00006145, -0.00002423, -0.00002708, 0.00003717, -0.00011140, 0.00009557, -0.00006328, 0.00003368, 0.00000042, -0.00005488, 0.00007119, -0.00004777, 0.00000145, -0.00002317, 0.00003926, -0.00001884, 0.00001029, 0.00001560, -0.00004739, 0.00002950, -0.00001744, -0.00000306, 0.00000637, -0.00001213, 0.00002825, 0.00000438, -0.00002712, 0.00000253, 0.00000678, },
		{ -0.00000424, 0.00000153, 0.00001168, -0.00003011, 0.00000196, 0.00002438, -0.00000140, -0.00000522, 0.00001125, -0.00003208, 0.00004198, -0.00005256, 0.00001447, 0.00000526, -0.00000609, 0.00002094, -0.00000044, -0.00002208, -0.00002042, 0.00002704, 0.00000621, -0.00007336, 0.00010719, -0.00013073, 0.00016010, -0.00016267, 0.00009527, -0.00010059, 0.00007697, -0.00019794, 0.00019316, -0.00100708, 0.00186704, -0.00043492, -0.00028220, 0.00009653, -0.00013950, 0.00005779, -0.00003249, -0.00004645, 0.00001830, 0.00001342, -0.00004397, 0.00007118, -0.00010815, 0.00010609, -0.00006762, 0.00002015, -0.00004233, 0.00005416, -0.00002800, 0.00001297, 0.00001546, -0.00003819, 0.00001338, -0.00000039, -0.00001854, 0.00001833, -0.00002242, 0.00003135, 0.00000577, -0.00002183, -0.00000699, 0.00001026, },
		{ -0.00000120, -0.00000449, 0.00001909, -0.00002990, -0.00000195, 0.00002052, 0.00000842, -0.00001479, 0.00002264, -0.00004263, 0.00004947, -0.00005306, 0.00001174, -0.00000110, 0.00000799, 0.00000224, 0.00002239, -0.00004634, 0.00000894, -0.00001869, 0.00006498, -0.00013804, 0.00016382, -0.00017686, 0.00020067, -0.00018953, 0.00013046, -0.00014945, 0.00014799, -0.00029291, 0.00033535, -0.00110823, 0.00164089, -0.00000943, -0.00055905, 0.00024976, -0.00024855, 0.00013752, -0.00009962, 0.00001887, -0.00005822, 0.00008572, -0.00011248, 0.00012737, -0.00014478, 0.00012562, -0.00007590, 0.00003097, -0.00005525, 0.00006333, -0.00003198, 0.00001266, 0.00001439, -0.00002609, -0.00000452, 0.00001694, -0.00003314, 0.00002886, -0.00003082, 0.00003275, 0.00000680, -0.00001536, -0.00001552, 0.00001124, },
		{ 0.00000327, -0.00001017, 0.00002366, -0.00002604, -0.00000742, 0.00001721, 0.00001630, -0.00002119, 0.00002994, -0.00004806, 0.00005130, -0.00004872, 0.00000724, -0.00000749, 0.00002086, -0.00001364, 0.00004172, -0.00006691, 0.00003449, -0.00005823, 0.00011187, -0.00018314, 0.00019377, -0.00019351, 0.00021044, -0.00018597, 0.00013641, -0.00016598, 0.00017956, -0.00033549, 0.00039650, -0.00108425, 0.00129491, 0.00045515, -0.00081165, 0.00037010, -0.00033067, 0.00019563, -0.00015019, 0.00007138, -0.00012093, 0.00014105, -0.00016055, 0.00016015, -0.00015895, 0.00012645, -0.00007079, 0.00003257, -0.00006060, 0.00006553, -0.00003010, 0.00000931, 0.00001272, -0.00001245, -0.00002207, 0.00003229, -0.00004477, 0.00003623, -0.00003600, 0.00003163, 0.00000822, -0.00000891, -0.00002187, 0.00000938, },
		{ 0.00000836, -0.00001445, 0.00002467, -0.00001866, -0.00001410, 0.00001463, 0.00002165, -0.00002381, 0.00003265, -0.00004812, 0.00004757, -0.00003994, 0.00000099, -0.00001267, 0.00003017, -0.00002396, 0.00005450, -0.00008006, 0.00005127, -0.00008520, 0.00013966, -0.00020180, 0.00019199, -0.00017796, 0.00018834, -0.00015182, 0.00011290, -0.00014925, 0.00017044, -0.00032501, 0.00038028, -0.00095406, 0.00087052, 0.00090866, -0.00100093, 0.00043254, -0.00036872, 0.00021902, -0.00017317, 0.00010069, -0.00015978, 0.00017031, -0.00018089, 0.00016503, -0.00014891, 0.00010866, -0.00005311, 0.00002554, -0.00005847, 0.00006077, -0.00002269, 0.00000339, 0.00001068, 0.00000124, -0.00003716, 0.00004361, -0.00005173, 0.00003911, -0.00003705, 0.00002747, 0.00001057, -0.00000346, -0.00002527, 0.00000480, },
		{ 0.00001312, -0.00001654, 0.00002195, -0.00000843, -0.00002121, 0.00001262, 0.00002436, -0.00002270, 0.00003103, -0.00004329, 0.00003904, -0.00002763, -0.00000670, -0.00001568, 0.00003420, -0.00002696, 0.00005890, -0.00008344, 0.00005612, -0.00009581, 0.00014475, -0.00019187, 0.00015908, -0.00013345, 0.00013915, -0.00009273, 0.00006562, -0.00010486, 0.00012704, -0.00027008, 0.00030175, -0.00074845, 0.00041669, 0.00130143, -0.00109421, 0.00041999, -0.00035277, 0.00020074, -0.00016292, 0.00010135, -0.00016957, 0.00016950, -0.00017150, 0.00014256, -0.00011725, 0.00007572, -0.00002611, 0.00001218, -0.00005033, 0.00005018, -0.00001094, -0.00000415, 0.00000845, 0.00001356, -0.00004798, 0.00004938, -0.00005290, 0.00003682, -0.00003361, 0.00002019, 0.00001411, 0.00000036, -0.00002540, -0.00000196, },
		{ 0.00001659, -0.00001596, 0.00001588, 0.00000350, -0.00002773, 0.00001074, 0.00002473, -0.00001843, 0.00002591, -0.00003466, 0.00002704, -0.00001311, -0.00001523, -0.00001605, 0.00003216, -0.00002222, 0.00005466, -0.00007657, 0.00004828, -0.00008944, 0.00012765, -0.00015623, 0.00010105, -0.00006846, 0.00007249, -0.00001888, 0.00000481, -0.00004333, 0.00006145, -0.00018607, 0.00018362, -0.00050456, -0.00001720, 0.00159129, -0.00107069, 0.00032680, -0.00028230, 0.00014162, -0.00012039, 0.00007387, -0.00015090, 0.00014034, -0.00013591, 0.00009816, -0.00007035, 0.00003376, 0.00000521, -0.00000400, -0.00003856, 0.00003570, 0.00000338, -0.00001214, 0.00000611, 0.00002335, -0.00005330, 0.00004884, -0.00004803, 0.00002946, -0.00002604, 0.00001025, 0.00001869, 0.00000236, -0.00002252, -0.00000998, },
		{ 0.00001798, -0.00001265, 0.00000731, 0.00001566, -0.00003252, 0.00000848, 0.00002333, -0.00001198, 0.00001855, -0.00002375, 0.00001324, 0.00000207, -0.00002375, -0.00001390, 0.00002437, -0.00001077, 0.00004317, -0.00006091, 0.00002943, -0.00006859, 0.00009274, -0.00010204, 0.00002808, 0.00000502, 0.00000077, 0.00005716, -0.00005702, 0.00002232, -0.00001136, -0.00009155, 0.00005156, -0.00025968, -0.00038818, 0.00174943, -0.00092500, 0.00016042, -0.00016686, 0.00005040, -0.00005311, 0.00002454, -0.00010982, 0.00008973, -0.00008223, 0.00004086, -0.00001705, -0.00000964, 0.00003511, -0.00001914, -0.00002587, 0.00001962, 0.00001827, -0.00001935, 0.00000372, 0.00002979, -0.00005261, 0.00004215, -0.00003774, 0.00001794, -0.00001535, -0.00000139, 0.00002373, 0.00000279, -0.00001729, -0.00001812, },
		{ 0.00001676, -0.00000700, -0.00000255, 0.00002651, -0.00003459, 0.00000537, 0.00002087, -0.00000450, 0.00001039, -0.00001220, -0.00000061, 0.00001628, -0.00003124, -0.00000989, 0.00001216, 0.00000510, 0.00002719, -0.00003949, 0.00000324, -0.00003829, 0.00004722, -0.00003925, -0.00004747, 0.00007386, -0.00006328, 0.00012296, -0.00010770, 0.00007939, -0.00007673, -0.00000453, -0.00007051, -0.00004575, -0.00066512, 0.00176406, -0.00066843, -0.00005920, -0.00002467, -0.00005779, 0.00002634, -0.00003590, -0.00005642, 0.00002817, -0.00002144, -0.00001849, 0.00003308, -0.00004698, 0.00005834, -0.00002983, -0.00001473, 0.00000412, 0.00003184, -0.00002473, 0.00000131, 0.00003253, -0.00004622, 0.00003032, -0.00002349, 0.00000386, -0.00000305, -0.00001341, 0.00002840, 0.00000227, -0.00001072, -0.00002523, },
		{ 0.00001275, 0.00000029, -0.00001234, 0.00003466, -0.00003328, 0.00000117, 0.00001797, 0.00000287, 0.00000281, -0.00000153, -0.00001289, 0.00002807, -0.00003668, -0.00000508, -0.00000239, 0.00002228, 0.00001025, -0.00001625, -0.00002545, -0.00000489, -0.00000039, 0.00002142, -0.00011346, 0.00012627, -0.00010909, 0.00016866, -0.00013782, 0.00011805, -0.00012311, 0.00006082, -0.00016423, 0.00011482, -0.00083200, 0.00164137, -0.00032732, -0.00030253, 0.00012044, -0.00016422, 0.00010277, -0.00009464, -0.00000264, -0.00003247, 0.00003487, -0.00006937, 0.00007163, -0.00007230, 0.00007113, -0.00003386, -0.00000680, -0.00000915, 0.00004264, -0.00002759, -0.00000107, 0.00003167, -0.00003516, 0.00001513, -0.00000730, -0.00001080, 0.00000906, -0.00002431, 0.00003171, 0.00000157, -0.00000393, -0.00003028, },
		{ 0.00000618, 0.00000823, -0.00002075, 0.00003912, -0.00002838, -0.00000407, 0.00001511, 0.00000918, -0.00000309, 0.00000704, -0.00002234, 0.00003632, -0.00003921, -0.00000071, -0.00001687, 0.00003748, -0.00000407, 0.00000481, -0.00005165, 0.00002515, -0.00004196, 0.00007040, -0.00016013, 0.00015402, -0.00013022, 0.00018870, -0.00014238, 0.00013304, -0.00014397, 0.00009630, -0.00021916, 0.00021118, -0.00088855, 0.00140370, 0.00006076, -0.00053500, 0.00024322, -0.00025015, 0.00016145, -0.00013941, 0.00004026, -0.00008134, 0.00007676, -0.00010362, 0.00009287, -0.00008222, 0.00007192, -0.00003063, -0.00000268, -0.00001925, 0.00004981, -0.00002762, -0.00000336, 0.00002773, -0.00002105, -0.00000121, 0.00000853, -0.00002393, 0.00001920, -0.00003271, 0.00003276, 0.00000147, 0.00000205, -0.00003255, },
		{ -0.00000235, 0.00001575, -0.00002672, 0.00003939, -0.00002023, -0.00000997, 0.00001255, 0.00001379, -0.00000661, 0.00001277, -0.00002819, 0.00004037, -0.00003827, 0.00000209, -0.00002899, 0.00004780, -0.00001285, 0.00002043, -0.00007127, 0.00004662, -0.00007118, 0.00010101, -0.00018177, 0.00015376, -0.00012537, 0.00018261, -0.00012154, 0.00012438, -0.00013850, 0.00010046, -0.00023343, 0.00024402, -0.00084844, 0.00108541, 0.00045270, -0.00072253, 0.00032156, -0.00030044, 0.00019109, -0.00016108, 0.00006395, -0.00011068, 0.00009780, -0.00011683, 0.00009473, -0.00007654, 0.00006161, -0.00002123, -0.00000190, -0.00002605, 0.00005314, -0.00002493, -0.00000548, 0.00002156, -0.00000582, -0.00001637, 0.00002184, -0.00003372, 0.00002594, -0.00003755, 0.00003093, 0.00000254, 0.00000643, -0.00003172, },
		{ -0.00001197, 0.00002180, -0.00002953, 0.00003554, -0.00000964, -0.00001589, 0.00001031, 0.00001640, -0.00000750, 0.00001542, -0.00003024, 0.00004013, -0.00003378, 0.00000257, -0.00003703, 0.00005132, -0.00001443, 0.00002877, -0.00008181, 0.00005638, -0.00008463, 0.00011044, -0.00017744, 0.00012742, -0.00009830, 0.00015471, -0.00008011, 0.00009683, -0.00011109, 0.00007789, -0.00021268, 0.00022354, -0.00073547, 0.00072719, 0.00080618, -0.00083699, 0.00034053, -0.00030647, 0.00018604, -0.00015538, 0.00006459, -0.00011718, 0.00009603, -0.00010896, 0.00007898, -0.00005805, 0.00004319, -0.00000808, -0.00000314, -0.00003001, 0.00005293, -0.00001994, -0.00000740, 0.00001423, 0.00000853, -0.00002830, 0.00003090, -0.00003887, 0.00002841, -0.00003832, 0.00002603, 0.00000504, 0.00000882, -0.00002793, },
		{ -0.00002160, 0.00002552, -0.00002895, 0.00002819, 0.00000221, -0.00002103, 0.00000828, 0.00001704, -0.00000596, 0.00001524, -0.00002881, 0.00003602, -0.00002614, 0.00000049, -0.00004014, 0.00004745, -0.00000874, 0.00002967, -0.00008276, 0.00005379, -0.00008214, 0.00009986, -0.00015080, 0.00008142, -0.00005674, 0.00011282, -0.00002624, 0.00005851, -0.00006975, 0.00003745, -0.00016775, 0.00016613, -0.00057870, 0.00036989, 0.00108556, -0.00086060, 0.00029509, -0.00026779, 0.00014733, -0.00012373, 0.00004336, -0.00010231, 0.00007409, -0.00008404, 0.00005063, -0.00003181, 0.00002107, 0.00000565, -0.00000469, -0.00003195, 0.00004980, -0.00001327, -0.00000910, 0.00000683, 0.00002026, -0.00003548, 0.00003472, -0.00003889, 0.00002645, -0.00003511, 0.00001837, 0.00000881, 0.00000924, -0.00002172, },
		{ -0.00003016, 0.00002634, -0.00002519, 0.00001835, 0.00001399, -0.00002462, 0.00000625, 0.00001595, -0.00000250, 0.00001289, -0.00002462, 0.00002894, -0.00001621, -0.00000376, -0.00003843, 0.00003694, 0.00000278, 0.00002460, -0.00007545, 0.00004057, -0.00006644, 0.00007385, -0.00010898, 0.00002519, -0.00001054, 0.00006629, 0.00003055, 0.00001884, -0.00002399, -0.00001014, -0.00011179, 0.00009050, -0.00040734, 0.00004891, 0.00126657, -0.00078851, 0.00019102, -0.00019212, 0.00008236, -0.00007269, 0.00000597, -0.00007169, 0.00003830, -0.00004908, 0.00001667, -0.00000395, 0.00000003, 0.00001682, -0.00000494, -0.00003270, 0.00004452, -0.00000557, -0.00001061, 0.00000037, 0.00002810, -0.00003718, 0.00003318, -0.00003407, 0.00002061, -0.00002862, 0.00000871, 0.00001333, 0.00000808, -0.00001396, },
		{ -0.00003667, 0.00002403, -0.00001885, 0.00000729, 0.00002438, -0.00002606, 0.00000408, 0.00001357, 0.00000211, 0.00000923, -0.00001871, 0.00002011, -0.00000522, -0.00000930, -0.00003292, 0.00002173, 0.00001752, 0.00001616, -0.00006262, 0.00002017, -0.00004233, 0.00003916, -0.00006094, -0.00003093, 0.00003051, 0.00002397, 0.00008137, -0.00001346, 0.00001735, -0.00005471, -0.00005732, 0.00001402, -0.00024641, -0.00020995, 0.00133880, -0.00062925, 0.00004391, -0.00009385, 0.00000329, -0.00001248, -0.00003886, -0.00003357, -0.00000286, -0.00001245, -0.00001545, 0.00001965, -0.00001580, 0.00002305, -0.00000285, -0.00003285, 0.00003776, 0.00000257, -0.00001200, -0.00000441, 0.00003140, -0.00003353, 0.00002699, -0.00002546, 0.00001207, -0.00002002, -0.00000184, 0.00001782, 0.00000602, -0.00000566, },
		{ -0.00004041, 0.00001875, -0.00001083, -0.00000367, 0.00003232, -0.00002505, 0.00000171, 0.00001037, 0.00000704, 0.00000525, -0.00001221, 0.00001087, 0.00000545, -0.00001489, -0.00002521, 0.00000448, 0.00003227, 0.00000748, -0.00004760, -0.00000313, -0.00001546, 0.00000311, -0.00001554, -0.00007772, 0.00005866, -0.00000761, 0.00011965, -0.00003209, 0.00004785, -0.00008864, -0.00001396, -0.00005001, -0.00011377, -0.00039269, 0.00130593, -0.00040287, -0.00012369, 0.00000884, -0.00007553, 0.00004526, -0.00008136, 0.00000304, -0.00004070, 0.00001798, -0.00003939, 0.00003461, -0.00002379, 0.00002323, 0.00000175, -0.00003254, 0.00003001, 0.00001067, -0.00001333, -0.00000706, 0.00003023, -0.00002544, 0.00001758, -0.00001463, 0.00000233, -0.00001071, -0.00001201, 0.00002139, 0.00000382, 0.00000214, },
		{ -0.00004100, 0.00001101, -0.00000218, -0.00001337, 0.00003711, -0.00002162, -0.00000077, 0.00000680, 0.00001148, 0.00000183, -0.00000617, 0.00000249, 0.00001448, -0.00001927, -0.00001715, -0.00001201, 0.00004394, 0.00000139, -0.00003353, -0.00002527, 0.00000888, -0.00002784, 0.00002007, -0.00010874, 0.00006959, -0.00002534, 0.00014211, -0.00003410, 0.00006456, -0.00010792, 0.00001290, -0.00009384, -0.00001882, -0.00049718, 0.00118380, -0.00013740, -0.00028618, 0.00009746, -0.00014043, 0.00008996, -0.00011289, 0.00003034, -0.00006804, 0.00003634, -0.00005117, 0.00003882, -0.00002328, 0.00001770, 0.00000815, -0.00003148, 0.00002153, 0.00001840, -0.00001463, -0.00000753, 0.00002527, -0.00001445, 0.00000676, -0.00000340, -0.00000695, -0.00000214, -0.00002061, 0.00002324, 0.00000221, 0.00000858, },
		{ -0.00003844, 0.00000158, 0.00000603, -0.00002088, 0.00003847, -0.00001615, -0.00000316, 0.00000329, 0.00001478, -0.00000033, -0.00000146, -0.00000399, 0.00002080, -0.00002136, -0.00001039, -0.00002532, 0.00005023, -0.00000019, -0.00002261, -0.00004317, 0.00002679, -0.00004924, 0.00004165, -0.00012128, 0.00006300, -0.00002976, 0.00014888, -0.00002003, 0.00006807, -0.00011238, 0.00002228, -0.00011547, 0.00003707, -0.00053135, 0.00099675, 0.00013579, -0.00041927, 0.00015668, -0.00018113, 0.00011417, -0.00012764, 0.00004327, -0.00008060, 0.00003984, -0.00004983, 0.00003279, -0.00001557, 0.00000817, 0.00001492, -0.00002910, 0.00001248, 0.00002550, -0.00001586, -0.00000611, 0.00001769, -0.00000245, -0.00000352, 0.00000652, -0.00001434, 0.00000451, -0.00002668, 0.00002279, 0.00000170, 0.00001310, },
		{ -0.00003310, -0.00000856, 0.00001286, -0.00002568, 0.00003658, -0.00000929, -0.00000520, 0.00000017, 0.00001647, -0.00000081, 0.00000143, -0.00000794, 0.00002381, -0.00002056, -0.00000608, -0.00003391, 0.00005002, 0.00000331, -0.00001572, -0.00005525, 0.00003629, -0.00005925, 0.00004825, -0.00011644, 0.00004222, -0.00002444, 0.00014283, 0.00000677, 0.00006187, -0.00010491, 0.00001687, -0.00011758, 0.00005892, -0.00051016, 0.00077315, 0.00038687, -0.00050396, 0.00017709, -0.00019256, 0.00011494, -0.00012372, 0.00004033, -0.00007760, 0.00002906, -0.00003735, 0.00001926, -0.00000352, -0.00000281, 0.00002029, -0.00002477, 0.00000299, 0.00003173, -0.00001686, -0.00000339, 0.00000897, 0.00000868, -0.00001160, 0.00001378, -0.00001884, 0.00000847, -0.00002970, 0.00001983, 0.00000248, 0.00001549, },
		{ -0.00002568, -0.00001838, 0.00001758, -0.00002760, 0.00003197, -0.00000182, -0.00000664, -0.00000231, 0.00001632, 0.00000049, 0.00000232, -0.00000917, 0.00002347, -0.00001686, -0.00000457, -0.00003732, 0.00004360, 0.00001107, -0.00001238, -0.00006151, 0.00003743, -0.00005854, 0.00004192, -0.00009840, 0.00001310, -0.00001467, 0.00012832, 0.00004128, 0.00005114, -0.00009026, 0.00000158, -0.00010587, 0.00005561, -0.00045188, 0.00054086, 0.00059166, -0.00052933, 0.00015667, -0.00017560, 0.00009408, -0.00010320, 0.00002355, -0.00006152, 0.00000755, -0.00001801, 0.00000249, 0.00000925, -0.00001242, 0.00002263, -0.00001811, -0.00000669, 0.00003682, -0.00001737, -0.00000014, 0.00000058, 0.00001732, -0.00001629, 0.00001761, -0.00001998, 0.00000949, -0.00002961, 0.00001457, 0.00000443, 0.00001587, },
		{ -0.00001706, -0.00002690, 0.00001972, -0.00002681, 0.00002540, 0.00000546, -0.00000732, -0.00000393, 0.00001435, 0.00000336, 0.00000144, -0.00000795, 0.00002026, -0.00001087, -0.00000549, -0.00003617, 0.00003250, 0.00002114, -0.00001102, -0.00006320, 0.00003184, -0.00004971, 0.00002677, -0.00007308, -0.00001755, -0.00000593, 0.00010985, 0.00007817, 0.00004137, -0.00007354, -0.00001799, -0.00008705, 0.00003731, -0.00037467, 0.00032346, 0.00073450, -0.00049377, 0.00010081, -0.00013651, 0.00005749, -0.00007138, -0.00000229, -0.00003730, -0.00001924, 0.00000270, -0.00001286, 0.00001928, -0.00001835, 0.00002089, -0.00000918, -0.00001605, 0.00004042, -0.00001707, 0.00000283, -0.00000616, 0.00002236, -0.00001709, 0.00001791, -0.00001793, 0.00000783, -0.00002682, 0.00000757, 0.00000710, 0.00001470, },
		{ -0.00000822, -0.00003331, 0.00001914, -0.00002378, 0.00001772, 0.00001189, -0.00000721, -0.00000456, 0.00001078, 0.00000733, -0.00000071, -0.00000493, 0.00001513, -0.00000369, -0.00000780, -0.00003191, 0.00001907, 0.00003093, -0.00000963, -0.00006220, 0.00002204, -0.00003639, 0.00000774, -0.00004671, -0.00004360, -0.00000240, 0.00009079, 0.00011291, 0.00003717, -0.00005897, -0.00003704, -0.00006719, 0.00001325, -0.00029385, 0.00013780, 0.00080943, -0.00040460, 0.00002097, -0.00008535, 0.00001358, -0.00003536, -0.00003099, -0.00001096, -0.00004524, 0.00001942, -0.00002282, 0.00002402, -0.00001923, 0.00001489, 0.00000134, -0.00002438, 0.00004211, -0.00001561, 0.00000481, -0.00001033, 0.00002335, -0.00001420, 0.00001520, -0.00001334, 0.00000420, -0.00002209, -0.00000032, 0.00000983, 0.00001262, },
		{ -0.00000006, -0.00003703, 0.00001598, -0.00001914, 0.00000976, 0.00001699, -0.00000640, -0.00000414, 0.00000603, 0.00001175, -0.00000343, -0.00000102, 0.00000930, 0.00000331, -0.00001016, -0.00002641, 0.00000593, 0.00003797, -0.00000635, -0.00006035, 0.00001063, -0.00002215, -0.00001062, -0.00002439, -0.00006089, -0.00000601, 0.00007276, 0.00014244, 0.00004154, -0.00004903, -0.00005249, -0.00005055, -0.00000974, -0.00022030, -0.00000698, 0.00081960, -0.00027602, -0.00006788, -0.00003361, -0.00002873, -0.00000236, -0.00005644, 0.00001177, -0.00006517, 0.00002805, -0.00002497, 0.00002247, -0.00001502, 0.00000543, 0.00001228, -0.00003078, 0.00004151, -0.00001273, 0.00000527, -0.00001148, 0.00002052, -0.00000846, 0.00001048, -0.00000722, -0.00000046, -0.00001638, -0.00000812, 0.00001191, 0.00001032, },
		{ 0.00000669, -0.00003781, 0.00001066, -0.00001360, 0.00000223, 0.00002053, -0.00000512, -0.00000271, 0.00000066, 0.00001589, -0.00000594, 0.00000279, 0.00000405, 0.00000879, -0.00001122, -0.00002148, -0.00000466, 0.00004042, -0.00000010, -0.00005882, -0.00000042, -0.00000974, -0.00002508, -0.00000916, -0.00006793, -0.00001612, 0.00005552, 0.00016524, 0.00005562, -0.00004422, -0.00006323, -0.00003915, -0.00002781, -0.00016003, -0.00010900, 0.00077539, -0.00012626, -0.00015013, 0.00000822, -0.00006192, 0.00002190, -0.00007398, 0.00002671, -0.00007555, 0.00002660, -0.00001894, 0.00001530, -0.00000693, -0.00000589, 0.00002210, -0.00003437, 0.00003837, -0.00000835, 0.00000403, -0.00000969, 0.00001467, -0.00000112, 0.00000495, -0.00000071, -0.00000513, -0.00001069, -0.00001488, 0.00001266, 0.00000843, },
		{ 0.00001158, -0.00003571, 0.00000379, -0.00000789, -0.00000436, 0.00002248, -0.00000363, -0.00000042, -0.00000467, 0.00001902, -0.00000755, 0.00000563, 0.00000046, 0.00001169, -0.00001003, -0.00001843, -0.00001126, 0.00003762, 0.00000908, -0.00005784, -0.00001014, -0.00000057, -0.00003425, -0.00000175, -0.00006583, -0.00002996, 0.00003755, 0.00018094, 0.00007903, -0.00004338, -0.00006991, -0.00003303, -0.00003978, -0.00011480, -0.00017199, 0.00069161, 0.00002573, -0.00021246, 0.00003269, -0.00008138, 0.00003426, -0.00008124, 0.00003192, -0.00007525, 0.00001541, -0.00000637, 0.00000465, 0.00000294, -0.00001704, 0.00002925, -0.00003449, 0.00003269, -0.00000259, 0.00000120, -0.00000549, 0.00000701, 0.00000640, -0.00000015, 0.00000517, -0.00000892, -0.00000588, -0.00001983, 0.00001164, 0.00000737, },
		{ 0.00001446, -0.00003111, -0.00000385, -0.00000266, -0.00000967, 0.00002297, -0.00000216, 0.00000242, -0.00000927, 0.00002054, -0.00000771, 0.00000683, -0.00000075, 0.00001149, -0.00000624, -0.00001777, -0.00001352, 0.00003016, 0.00001999, -0.00005662, -0.00001864, 0.00000528, -0.00003842, -0.00000082, -0.00005764, -0.00004364, 0.00001706, 0.00018957, 0.00011042, -0.00004419, -0.00007426, -0.00003090, -0.00004643, -0.00008333, -0.00020327, 0.00058447, 0.00016335, -0.00024598, 0.00003674, -0.00008608, 0.00003454, -0.00007844, 0.00002782, -0.00006546, -0.00000312, 0.00000958, -0.00000651, 0.00001212, -0.00002598, 0.00003249, -0.00003088, 0.00002479, 0.00000414, -0.00000279, 0.00000023, -0.00000105, 0.00001273, -0.00000386, 0.00000964, -0.00001118, -0.00000257, -0.00002250, 0.00000866, 0.00000730, },
		{ 0.00001551, -0.00002464, -0.00001146, 0.00000153, -0.00001350, 0.00002220, -0.00000083, 0.00000541, -0.00001251, 0.00002005, -0.00000617, 0.00000610, 0.00000066, 0.00000827, -0.00000019, -0.00001919, -0.00001220, 0.00001976, 0.00003057, -0.00005377, -0.00002671, 0.00000873, -0.00003913, -0.00000383, -0.00004724, -0.00005339, -0.00000699, 0.00019085, 0.00014809, -0.00004395, -0.00007838, -0.00003098, -0.00004946, -0.00006271, -0.00021147, 0.00046899, 0.00027444, -0.00024734, 0.00002189, -0.00007840, 0.00002517, -0.00006792, 0.00001676, -0.00004913, -0.00002520, 0.00002504, -0.00001509, 0.00001839, -0.00003120, 0.00003122, -0.00002385, 0.00001537, 0.00001119, -0.00000727, 0.00000641, -0.00000817, 0.00001685, -0.00000559, 0.00001230, -0.00001161, -0.00000103, -0.00002273, 0.00000388, 0.00000809, },
		{ 0.00001514, -0.00001709, -0.00001821, 0.00000427, -0.00001582, 0.00002040, 0.00000033, 0.00000805, -0.00001388, 0.00001742, -0.00000300, 0.00000351, 0.00000442, 0.00000273, 0.00000710, -0.00002161, -0.00000886, 0.00000879, 0.00003847, -0.00004777, -0.00003534, 0.00001103, -0.00003829, -0.00000799, -0.00003813, -0.00005674, -0.00003403, 0.00018368, 0.00019038, -0.00004008, -0.00008402, -0.00003171, -0.00005063, -0.00004974, -0.00020479, 0.00035711, 0.00035244, -0.00021874, -0.00000649, -0.00006318, 0.00001030, -0.00005331, 0.00000218, -0.00003005, -0.00004664, 0.00003637, -0.00001867, 0.00002040, -0.00003202, 0.00002560, -0.00001426, 0.00000544, 0.00001777, -0.00001147, 0.00001201, -0.00001326, 0.00001817, -0.00000521, 0.00001314, -0.00001024, -0.00000125, -0.00002073, -0.00000225, 0.00000937, },
		{ 0.00001393, -0.00000933, -0.00002339, 0.00000530, -0.00001666, 0.00001773, 0.00000145, 0.00000985, -0.00001310, 0.00001284, 0.00000145, -0.00000050, 0.00000979, -0.00000394, 0.00001420, -0.00002357, -0.00000544, -0.00000029, 0.00004167, -0.00003762, -0.00004514, 0.00001317, -0.00003747, -0.00001122, -0.00003241, -0.00005330, -0.00006161, 0.00016625, 0.00023575, -0.00003039, -0.00009226, -0.00003217, -0.00005117, -0.00004171, -0.00018988, 0.00025685, 0.00039624, -0.00016675, -0.00004075, -0.00004624, -0.00000554, -0.00003842, -0.00001243, -0.00001193, -0.00006382, 0.00004093, -0.00001610, 0.00001793, -0.00002876, 0.00001658, -0.00000346, -0.00000383, 0.00002304, -0.00001465, 0.00001613, -0.00001562, 0.00001662, -0.00000305, 0.00001244, -0.00000734, -0.00000293, -0.00001695, -0.00000904, 0.00001057, },
		{ 0.00001249, -0.00000218, -0.00002647, 0.00000452, -0.00001614, 0.00001433, 0.00000272, 0.00001041, -0.00001015, 0.00000680, 0.00000655, -0.00000525, 0.00001578, -0.00001034, 0.00001959, -0.00002360, -0.00000370, -0.00000557, 0.00003907, -0.00002333, -0.00005584, 0.00001535, -0.00003731, -0.00001277, -0.00003037, -0.00004486, -0.00008573, 0.00013652, 0.00028241, -0.00001298, -0.00010349, -0.00003219, -0.00005164, -0.00003677, -0.00017141, 0.00017231, 0.00040915, -0.00010057, -0.00007278, -0.00003282, -0.00001859, -0.00002626, -0.00002435, 0.00000248, -0.00007437, 0.00003769, -0.00000771, 0.00001197, -0.00002262, 0.00000572, 0.00000693, -0.00001124, 0.00002629, -0.00001623, 0.00001818, -0.00001506, 0.00001263, 0.00000028, 0.00001071, -0.00000337, -0.00000562, -0.00001207, -0.00001568, 0.00001108, },
		{ 0.00001138, 0.00000371, -0.00002713, 0.00000202, -0.00001446, 0.00001032, 0.00000430, 0.00000953, -0.00000536, 0.00000005, 0.00001153, -0.00000990, 0.00002128, -0.00001510, 0.00002198, -0.00002065, -0.00000479, -0.00000615, 0.00003088, -0.00000612, -0.00006621, 0.00001685, -0.00003732, -0.00001328, -0.00003049, -0.00003491, -0.00010178, 0.00009316, 0.00032777, 0.00001395, -0.00011755, -0.00003211, -0.00005205, -0.00003390, -0.00015225, 0.00010451, 0.00039731, -0.00002994, -0.00009583, -0.00002632, -0.00002682, -0.00001832, -0.00003224, 0.00001183, -0.00007757, 0.00002728, 0.00000475, 0.00000443, -0.00001538, -0.00000516, 0.00001536, -0.00001585, 0.00002706, -0.00001589, 0.00001793, -0.00001180, 0.00000698, 0.00000398, 0.00000850, 0.00000118, -0.00000878, -0.00000682, -0.00002135, 0.00001034, },
		},
		{ { -0.00001314, -0.00001303, 0.00004534, -0.00000816, 0.00000265, -0.00000066, -0.00000454, -0.00000409, 0.00002067, -0.00000132, -0.00003253, 0.00001139, -0.00000249, 0.00002284, -0.00007779, 0.00001459, -0.00000003, -0.00001345, -0.00000956, -0.00001793, 0.00002585, -0.00002909, -0.00001142, -0.00002506, -0.00013670, 0.00005278, 0.00038107, 0.00005074, -0.00014617, -0.00002532, -0.00004261, -0.00000662, 0.00000494, -0.00000662, -0.00004261, -0.00002532, -0.00014617, 0.00005074, 0.00038107, 0.00005278, -0.00013670, -0.00002506, -0.00001142, -0.00002909, 0.00002585, -0.00001793, -0.00000956, -0.00001345, -0.00000003, 0.00001459, -0.00007779, 0.00002284, -0.00000249, 0.00001139, -0.00003253, -0.00000132, 0.00002067, -0.00000409, -0.00000454, -0.00000066, 0.00000265, -0.00000816, 0.00004534, -0.00001303, },
		{ -0.00001399, -0.00001765, 0.00003833, 0.00000342, -0.00000347, 0.00000410, -0.00000633, -0.00000183, 0.00001026, 0.00001469, -0.00004019, 0.00001453, -0.00001238, 0.00003413, -0.00007052, -0.00000659, 0.00001410, -0.00002176, -0.00000040, -0.00002648, 0.00002672, -0.00002210, -0.00001932, -0.00001151, -0.00012291, -0.00003977, 0.00039836, 0.00014094, -0.00016975, -0.00003155, -0.00004372, -0.00001377, 0.00000499, -0.00000168, -0.00003843, -0.00002396, -0.00012013, -0.00001497, 0.00033488, 0.00013466, -0.00012562, -0.00004844, -0.00000248, -0.00003381, 0.00002124, -0.00000690, -0.00001892, -0.00000583, -0.00001197, 0.00002831, -0.00007376, 0.00000439, 0.00000845, 0.00000806, -0.00002326, -0.00001639, 0.00002680, -0.00000296, -0.00000352, -0.00000560, 0.00000887, -0.00001790, 0.00004725, -0.00000440, },
		{ -0.00001612, -0.00001880, 0.00002781, 0.00001505, -0.00000865, 0.00000856, -0.00000906, 0.00000325, -0.00000216, 0.00002844, -0.00004340, 0.00001429, -0.00001688, 0.00003472, -0.00005164, -0.00003295, 0.00002815, -0.00002949, 0.00000787, -0.00003274, 0.00002555, -0.00001596, -0.00002216, -0.00001061, -0.00008973, -0.00012638, 0.00037396, 0.00025356, -0.00018435, -0.00004542, -0.00004117, -0.00002273, 0.00000473, 0.00000160, -0.00003261, -0.00002464, -0.00009611, -0.00005770, 0.00027254, 0.00019563, -0.00009236, -0.00007366, 0.00000139, -0.00003268, 0.00001145, 0.00000643, -0.00002763, -0.00000002, -0.00002033, 0.00003411, -0.00006096, -0.00001637, 0.00001630, 0.00000680, -0.00001423, -0.00002851, 0.00002779, 0.00000101, -0.00000265, -0.00001053, 0.00001407, -0.00002410, 0.00004306, 0.00000793, },
		{ -0.00001837, -0.00001764, 0.00001565, 0.00002519, -0.00001227, 0.00001248, -0.00001222, 0.00000947, -0.00001334, 0.00003641, -0.00003951, 0.00000812, -0.00001342, 0.00002411, -0.00002412, -0.00005992, 0.00003866, -0.00003454, 0.00001375, -0.00003594, 0.00002248, -0.00001112, -0.00002033, -0.00001854, -0.00004988, -0.00018779, 0.00029957, 0.00038041, -0.00018087, -0.00006958, -0.00003509, -0.00003252, 0.00000327, 0.00000415, -0.00002678, -0.00002493, -0.00007681, -0.00008083, 0.00020450, 0.00023220, -0.00004520, -0.00009149, -0.00000437, -0.00002445, -0.00000273, 0.00002020, -0.00003381, 0.00000252, -0.00002407, 0.00003266, -0.00004285, -0.00003502, 0.00001862, 0.00000802, -0.00000588, -0.00003696, 0.00002404, 0.00000658, -0.00000133, -0.00001495, 0.00001700, -0.00002545, 0.00003278, 0.00002264, },
		{ -0.00001907, -0.00001562, 0.00000372, 0.00003253, -0.00001381, 0.00001528, -0.00001463, 0.00001448, -0.00002001, 0.00003600, -0.00002745, -0.00000438, -0.00000252, 0.00000557, 0.00000588, -0.00008119, 0.00004141, -0.00003414, 0.00001477, -0.00003396, 0.00001570, -0.00000515, -0.00001819, -0.00002677, -0.00001885, -0.00020849, 0.00017663, 0.00050498, -0.00014748, -0.00010649, -0.00002588, -0.00004231, 0.00000020, 0.00000641, -0.00002187, -0.00002367, -0.00006287, -0.00008894, 0.00013869, 0.00024556, 0.00000622, -0.00009574, -0.00001979, -0.00001189, -0.00001747, 0.00003056, -0.00003461, 0.00000004, -0.00002252, 0.00002518, -0.00002271, -0.00004870, 0.00001521, 0.00001037, 0.00000238, -0.00004187, 0.00001664, 0.00001236, 0.00000073, -0.00001809, 0.00001660, -0.00002152, 0.00001765, 0.00003743, },
		{ -0.00001643, -0.00001412, -0.00000639, 0.00003612, -0.00001280, 0.00001616, -0.00001487, 0.00001608, -0.00001998, 0.00002667, -0.00000868, -0.00002072, 0.00001196, -0.00001453, 0.00003046, -0.00009035, 0.00003277, -0.00002586, 0.00000848, -0.00002446, 0.00000282, 0.00000536, -0.00002114, -0.00002666, -0.00000848, -0.00018334, 0.00002011, 0.00060250, -0.00007130, -0.00015748, -0.00001426, -0.00005184, -0.00000389, 0.00000764, -0.00001737, -0.00002149, -0.00005285, -0.00008761, 0.00008115, 0.00023882, 0.00005458, -0.00008585, -0.00003968, -0.00000122, -0.00002708, 0.00003309, -0.00002742, -0.00000852, -0.00001584, 0.00001348, -0.00000334, -0.00005621, 0.00000768, 0.00001171, 0.00001119, -0.00004346, 0.00000675, 0.00001731, 0.00000337, -0.00001918, 0.00001248, -0.00001308, 0.00000018, 0.00004938, },
		{ -0.00000914, -0.00001407, -0.00001358, 0.00003545, -0.00000896, 0.00001440, -0.00001184, 0.00001307, -0.00001296, 0.00001042, 0.00001279, -0.00003609, 0.00002415, -0.00002884, 0.00004260, -0.00008335, 0.00001165, -0.00000931, -0.00000583, -0.00000686, -0.00001679, 0.00002178, -0.00003176, -0.00001442, -0.00002142, -0.00012132, -0.00014196, 0.00064395, 0.00005752, -0.00022036, -0.00000225, -0.00006101, -0.00000770, 0.00000622, -0.00001170, -0.00002035, -0.00004387, -0.00008271, 0.00003642, 0.00021527, 0.00009635, -0.00006632, -0.00005589, 0.00000079, -0.00002683, 0.00002521, -0.00001177, -0.00002222, -0.00000583, 0.00000021, 0.00001283, -0.00005754, -0.00000172, 0.00001036, 0.00002034, -0.00004149, -0.00000478, 0.00002084, 0.00000620, -0.00001782, 0.00000520, -0.00000214, -0.00001629, 0.00005554, },
		{ 0.00000304, -0.00001554, -0.00001746, 0.00003060, -0.00000241, 0.00000981, -0.00000540, 0.00000578, -0.00000083, -0.00000856, 0.00003146, -0.00004501, 0.00002843, -0.00003182, 0.00003890, -0.00006057, -0.00001901, 0.00001258, -0.00002564, 0.00001598, -0.00003996, 0.00004113, -0.00004707, 0.00000599, -0.00004921, -0.00004431, -0.00027422, 0.00060389, 0.00023961, -0.00028621, 0.00000520, -0.00006858, -0.00001053, 0.00000101, -0.00000351, -0.00002214, -0.00003324, -0.00007897, 0.00000699, 0.00017839, 0.00013074, -0.00004376, -0.00006128, -0.00000945, -0.00001594, 0.00000830, 0.00000926, -0.00003734, 0.00000384, -0.00001112, 0.00002357, -0.00005313, -0.00001114, 0.00000603, 0.00002839, -0.00003526, -0.00001733, 0.00002266, 0.00000887, -0.00001424, -0.00000372, 0.00000848, -0.00002831, 0.00005380, },
		{ 0.00001885, -0.00001763, -0.00001845, 0.00002232, 0.00000611, 0.00000301, 0.00000344, -0.00000394, 0.00001297, -0.00002514, 0.00004211, -0.00004349, 0.00002190, -0.00002215, 0.00002127, -0.00002763, -0.00005213, 0.00003349, -0.00004507, 0.00003755, -0.00005963, 0.00005605, -0.00005879, 0.00002397, -0.00007536, 0.00001970, -0.00034416, 0.00047039, 0.00046039, -0.00033694, -0.00000096, -0.00007096, -0.00001371, -0.00000713, 0.00000661, -0.00002684, -0.00002030, -0.00007835, -0.00000783, 0.00013294, 0.00015722, -0.00002296, -0.00005360, -0.00002960, 0.00000088, -0.00001164, 0.00002918, -0.00004796, 0.00000833, -0.00001681, 0.00002707, -0.00004356, -0.00001973, -0.00000005, 0.00003303, -0.00002406, -0.00003014, 0.00002243, 0.00001136, -0.00000934, -0.00001218, 0.00001595, -0.00003340, 0.00004368, },
		{ 0.00003551, -0.00001855, -0.00001764, 0.00001203, 0.00001518, -0.00000452, 0.00001265, -0.00001327, 0.00002447, -0.00003487, 0.00004164, -0.00003080, 0.00000581, -0.00000347, -0.00000331, 0.00000588, -0.00007819, 0.00004552, -0.00005662, 0.00004959, -0.00006707, 0.00005745, -0.00005680, 0.00002698, -0.00008245, 0.00004672, -0.00033354, 0.00025341, 0.00068712, -0.00034584, -0.00003290, -0.00006248, -0.00002096, -0.00001468, 0.00001537, -0.00003168, -0.00000741, -0.00007916, -0.00001247, 0.00008583, 0.00017345, -0.00000433, -0.00003705, -0.00005138, 0.00001459, -0.00002586, 0.00004011, -0.00004798, 0.00000346, -0.00001430, 0.00002270, -0.00002976, -0.00002733, -0.00000582, 0.00003216, -0.00000812, -0.00004181, 0.00001956, 0.00001398, -0.00000435, -0.00001819, 0.00001835, -0.00003081, 0.00002678, },
		{ 0.00004927, -0.00001622, -0.00001645, 0.00000154, 0.00002297, -0.00001080, 0.00001979, -0.00001938, 0.00003053, -0.00003536, 0.00003023, -0.00001001, -0.00001463, 0.00001681, -0.00002514, 0.00002954, -0.00008856, 0.00004227, -0.00005422, 0.00004550, -0.00005572, 0.00003905, -0.00003442, 0.00000713, -0.00006041, 0.00002650, -0.00024659, -0.00001249, 0.00087209, -0.00028286, -0.00010178, -0.00003785, -0.00003686, -0.00001658, 0.00001771, -0.00003184, 0.00000073, -0.00007684, -0.00001409, 0.00004556, 0.00017519, 0.00001588, -0.00002050, -0.00006375, 0.00001555, -0.00002654, 0.00003627, -0.00003405, -0.00001209, -0.00000368, 0.00001184, -0.00001352, -0.00003381, -0.00000954, 0.00002496, 0.00001086, -0.00005021, 0.00001327, 0.00001713, -0.00000031, -0.00002061, 0.00001534, -0.00002196, 0.00000663, },
		{ 0.00005644, -0.00000904, -0.00001601, -0.00000745, 0.00002772, -0.00001395, 0.00002284, -0.00002039, 0.00002976, -0.00002698, 0.00001126, 0.00001290, -0.00003212, 0.00003040, -0.00003547, 0.00003606, -0.00007910, 0.00002201, -0.00003640, 0.00002375, -0.00002502, 0.00000145, 0.00000695, -0.00003364, -0.00001228, -0.00003248, -0.00011121, -0.00027353, 0.00096279, -0.00012451, -0.00021089, 0.00000342, -0.00006349, -0.00000913, 0.00000942, -0.00002297, -0.00000045, -0.00006650, -0.00001998, 0.00001983, 0.00015842, 0.00004259, -0.00001309, -0.00005804, -0.00000162, -0.00001095, 0.00001723, -0.00000791, -0.00003515, 0.00001140, -0.00000174, 0.00000219, -0.00003842, -0.00001071, 0.00001267, 0.00002957, -0.00005279, 0.00000304, 0.00002086, 0.00000239, -0.00001954, 0.00000834, -0.00001001, -0.00001221, },
		{ 0.00005459, 0.00000322, -0.00001663, -0.00001381, 0.00002824, -0.00001285, 0.00002091, -0.00001600, 0.00002288, -0.00001250, -0.00000986, 0.00003118, -0.00004015, 0.00003150, -0.00003013, 0.00002436, -0.00005247, -0.00001054, -0.00000795, -0.00001032, 0.00001789, -0.00004618, 0.00005643, -0.00008240, 0.00004550, -0.00010517, 0.00002841, -0.00047007, 0.00091721, 0.00013413, -0.00034844, 0.00005250, -0.00009690, 0.00000667, -0.00000960, -0.00000436, -0.00001232, -0.00004619, -0.00003418, 0.00001232, 0.00012265, 0.00007773, -0.00001919, -0.00003275, -0.00003435, 0.00001608, -0.00001071, 0.00002302, -0.00005825, 0.00002432, -0.00001255, 0.00001362, -0.00003967, -0.00001035, -0.00000167, 0.00004384, -0.00004742, -0.00001084, 0.00002451, 0.00000428, -0.00001632, 0.00000000, 0.00000108, -0.00002543, },
		{ 0.00004362, 0.00001898, -0.00001751, -0.00001733, 0.00002438, -0.00000759, 0.00001453, -0.00000754, 0.00001227, 0.00000394, -0.00002755, 0.00003972, -0.00003559, 0.00001937, -0.00001137, 0.00000048, -0.00001794, -0.00004509, 0.00002114, -0.00004576, 0.00005973, -0.00008835, 0.00009657, -0.00011925, 0.00008889, -0.00015906, 0.00012590, -0.00055445, 0.00071997, 0.00046975, -0.00048342, 0.00008945, -0.00012551, 0.00002328, -0.00003341, 0.00001903, -0.00003082, -0.00001920, -0.00005504, 0.00002035, 0.00007341, 0.00011713, -0.00003523, 0.00000445, -0.00007163, 0.00004289, -0.00003599, 0.00004770, -0.00007185, 0.00002764, -0.00001515, 0.00001742, -0.00003600, -0.00001039, -0.00001417, 0.00004993, -0.00003344, -0.00002680, 0.00002659, 0.00000657, -0.00001287, -0.00000680, 0.00000796, -0.00003045, },
		{ 0.00002602, 0.00003493, -0.00001694, -0.00001862, 0.00001704, 0.00000053, 0.00000554, 0.00000254, 0.00000106, 0.00001836, -0.00003790, 0.00003681, -0.00002005, -0.00000122, 0.00001277, -0.00002449, 0.00001172, -0.00006930, 0.00003929, -0.00006985, 0.00008566, -0.00010871, 0.00010996, -0.00012526, 0.00009618, -0.00016710, 0.00014890, -0.00050726, 0.00039294, 0.00082834, -0.00056825, 0.00008654, -0.00013245, 0.00002816, -0.00005097, 0.00003728, -0.00004696, 0.00000616, -0.00007528, 0.00003500, 0.00002215, 0.00015052, -0.00005034, 0.00003883, -0.00009749, 0.00005512, -0.00004615, 0.00005563, -0.00006805, 0.00001620, -0.00000650, 0.00001226, -0.00002689, -0.00001251, -0.00002160, 0.00004596, -0.00001233, -0.00004192, 0.00002519, 0.00001052, -0.00001087, -0.00001014, 0.00000907, -0.00002724, },
		{ 0.00000634, 0.00004691, -0.00001299, -0.00001878, 0.00000788, 0.00000937, -0.00000354, 0.00001148, -0.00000797, 0.00002807, -0.00003970, 0.00002454, 0.00000077, -0.00002180, 0.00003165, -0.00003871, 0.00002512, -0.00007368, 0.00003824, -0.00007343, 0.00008531, -0.00009681, 0.00008677, -0.00009107, 0.00005790, -0.00011929, 0.00009112, -0.00034591, -0.00000453, 0.00113315, -0.00054953, 0.00001734, -0.00010217, 0.00000877, -0.00005049, 0.00003925, -0.00005019, 0.00001959, -0.00008497, 0.00004423, -0.00001702, 0.00016499, -0.00005083, 0.00005436, -0.00009744, 0.00004194, -0.00003333, 0.00004155, -0.00004441, -0.00001006, 0.00001210, -0.00000001, -0.00001380, -0.00001702, -0.00002263, 0.00003274, 0.00001214, -0.00005253, 0.00001870, 0.00001672, -0.00001085, -0.00000980, 0.00000522, -0.00001845, },
		{ -0.00001018, 0.00005125, -0.00000444, -0.00001870, -0.00000128, 0.00001667, -0.00001030, 0.00001713, -0.00001316, 0.00003227, -0.00003434, 0.00000770, 0.00001939, -0.00003371, 0.00003647, -0.00003454, 0.00001717, -0.00005605, 0.00001714, -0.00005513, 0.00005742, -0.00005324, 0.00003012, -0.00002238, -0.00001756, -0.00002799, -0.00002449, -0.00012152, -0.00038788, 0.00130288, -0.00038528, -0.00013055, -0.00002956, -0.00004019, -0.00002557, 0.00001825, -0.00003363, 0.00001384, -0.00007647, 0.00003805, -0.00003268, 0.00015079, -0.00002674, 0.00004097, -0.00006528, 0.00000174, 0.00000127, 0.00000868, -0.00000604, -0.00004492, 0.00003461, -0.00001428, -0.00000025, -0.00002245, -0.00001840, 0.00001373, 0.00003479, -0.00005521, 0.00000662, 0.00002446, -0.00001191, -0.00000735, -0.00000084, -0.00000838, },
		{ -0.00001942, 0.00004622, 0.00000834, -0.00001853, -0.00000902, 0.00002067, -0.00001320, 0.00001851, -0.00001429, 0.00003179, -0.00002477, -0.00000831, 0.00002956, -0.00003172, 0.00002420, -0.00001216, -0.00000833, -0.00002323, -0.00001620, -0.00002263, 0.00001115, 0.00000947, -0.00004356, 0.00006049, -0.00010541, 0.00007552, -0.00015337, 0.00009605, -0.00066968, 0.00127564, -0.00006370, -0.00034357, 0.00007240, -0.00011086, 0.00001941, -0.00002301, 0.00000121, -0.00001074, -0.00004908, 0.00001347, -0.00002134, 0.00010670, 0.00002239, -0.00000006, -0.00000726, -0.00005546, 0.00004639, -0.00003133, 0.00003522, -0.00007728, 0.00005180, -0.00002356, 0.00000908, -0.00002604, -0.00001187, -0.00000597, 0.00005036, -0.00004801, -0.00000983, 0.00003168, -0.00001194, -0.00000547, -0.00000553, -0.00000142, },
		{ -0.00001982, 0.00003287, 0.00002309, -0.00001747, -0.00001467, 0.00002061, -0.00001189, 0.00001591, -0.00001225, 0.00002823, -0.00001400, -0.00001964, 0.00002875, -0.00001639, -0.00000084, 0.00002006, -0.00003951, 0.00001095, -0.00004783, 0.00001005, -0.00003705, 0.00007080, -0.00010970, 0.00012919, -0.00017309, 0.00015242, -0.00024532, 0.00024024, -0.00078535, 0.00103116, 0.00038373, -0.00057690, 0.00017064, -0.00018084, 0.00006788, -0.00007093, 0.00004287, -0.00004451, -0.00001101, -0.00002325, 0.00001022, 0.00004242, 0.00008548, -0.00005503, 0.00005869, -0.00011025, 0.00008395, -0.00006214, 0.00006468, -0.00009487, 0.00005453, -0.00002165, 0.00001028, -0.00002509, -0.00000660, -0.00002129, 0.00005519, -0.00003133, -0.00002767, 0.00003554, -0.00000861, -0.00000667, -0.00000609, -0.00000034, },
		{ -0.00001305, 0.00001498, 0.00003607, -0.00001412, -0.00001829, 0.00001681, -0.00000722, 0.00001062, -0.00000846, 0.00002313, -0.00000386, -0.00002524, 0.00001896, 0.00000607, -0.00002841, 0.00004855, -0.00006097, 0.00003092, -0.00006322, 0.00002833, -0.00006989, 0.00010985, -0.00014466, 0.00015802, -0.00019271, 0.00017145, -0.00026318, 0.00026974, -0.00071392, 0.00060314, 0.00088091, -0.00075986, 0.00021868, -0.00021873, 0.00009555, -0.00010468, 0.00007310, -0.00007123, 0.00002285, -0.00005849, 0.00004688, -0.00002376, 0.00014289, -0.00010220, 0.00010875, -0.00014059, 0.00009579, -0.00006897, 0.00007045, -0.00008906, 0.00003762, -0.00000594, 0.00000213, -0.00001851, -0.00000508, -0.00002891, 0.00004840, -0.00000816, -0.00004280, 0.00003336, -0.00000048, -0.00001199, -0.00000179, -0.00000511, },
		{ -0.00000337, -0.00000201, 0.00004327, -0.00000720, -0.00002025, 0.00001032, -0.00000076, 0.00000433, -0.00000421, 0.00001728, 0.00000536, -0.00002678, 0.00000562, 0.00002635, -0.00004644, 0.00006002, -0.00006005, 0.00002603, -0.00005371, 0.00002359, -0.00007637, 0.00011353, -0.00013492, 0.00013421, -0.00015253, 0.00012165, -0.00019692, 0.00018220, -0.00048598, 0.00007588, 0.00132096, -0.00081318, 0.00017161, -0.00019589, 0.00007983, -0.00010441, 0.00007412, -0.00007460, 0.00003695, -0.00007694, 0.00007104, -0.00007135, 0.00017362, -0.00012001, 0.00012226, -0.00013052, 0.00007137, -0.00004509, 0.00004865, -0.00005889, 0.00000264, 0.00002081, -0.00001275, -0.00000774, -0.00000764, -0.00002826, 0.00003223, 0.00001655, -0.00005113, 0.00002384, 0.00001203, -0.00002030, 0.00000554, -0.00001296, },
		{ 0.00000393, -0.00001289, 0.00004187, 0.00000351, -0.00002077, 0.00000251, 0.00000565, -0.00000139, -0.00000031, 0.00001073, 0.00001445, -0.00002721, -0.00000510, 0.00003600, -0.00004614, 0.00004722, -0.00003267, -0.00000437, -0.00002106, -0.00000250, -0.00005659, 0.00008189, -0.00008288, 0.00006370, -0.00006257, 0.00001767, -0.00006793, 0.00001465, -0.00017540, -0.00043541, 0.00159276, -0.00067399, 0.00000521, -0.00010060, 0.00001124, -0.00006094, 0.00003728, -0.00004616, 0.00002242, -0.00006884, 0.00007014, -0.00008568, 0.00016385, -0.00009609, 0.00009036, -0.00007745, 0.00001327, 0.00000432, 0.00000590, -0.00001235, -0.00004166, 0.00005061, -0.00002827, 0.00000339, -0.00001239, -0.00002149, 0.00001133, 0.00003722, -0.00004991, 0.00000778, 0.00002638, -0.00002845, 0.00001216, -0.00001944, },
		{ 0.00000462, -0.00001458, 0.00003160, 0.00001660, -0.00001955, -0.00000537, 0.00001048, -0.00000547, 0.00000296, 0.00000327, 0.00002418, -0.00002888, -0.00000912, 0.00003106, -0.00002592, 0.00001253, 0.00001412, -0.00004949, 0.00002219, -0.00003789, -0.00002190, 0.00002830, -0.00000635, -0.00003028, 0.00004812, -0.00010481, 0.00007873, -0.00016987, 0.00012317, -0.00081578, 0.00161338, -0.00032112, -0.00026757, 0.00005119, -0.00009848, 0.00001725, -0.00003076, 0.00000904, -0.00001755, -0.00003545, 0.00004251, -0.00006411, 0.00011333, -0.00003322, 0.00002061, 0.00000521, -0.00006212, 0.00006289, -0.00004224, 0.00003572, -0.00008185, 0.00007269, -0.00003685, 0.00001008, -0.00001611, -0.00001247, -0.00000868, 0.00004921, -0.00003870, -0.00001175, 0.00003854, -0.00003239, 0.00001394, -0.00002031, },
		{ -0.00000267, -0.00000736, 0.00001511, 0.00002907, -0.00001583, -0.00001240, 0.00001275, -0.00000745, 0.00000554, -0.00000492, 0.00003422, -0.00003196, -0.00000639, 0.00001390, 0.00000753, -0.00003226, 0.00006402, -0.00009058, 0.00005708, -0.00006481, 0.00000991, -0.00002603, 0.00006802, -0.00011551, 0.00014187, -0.00020210, 0.00019089, -0.00030475, 0.00032388, -0.00098360, 0.00135570, 0.00020882, -0.00059046, 0.00021415, -0.00021582, 0.00010372, -0.00010772, 0.00007212, -0.00006714, 0.00001007, -0.00000124, -0.00001759, 0.00003648, 0.00005058, -0.00006459, 0.00009162, -0.00012951, 0.00010840, -0.00007670, 0.00006883, -0.00010442, 0.00007736, -0.00003238, 0.00000858, -0.00001587, -0.00000513, -0.00002302, 0.00005019, -0.00001969, -0.00003018, 0.00004427, -0.00002881, 0.00000819, -0.00001344, },
		{ -0.00001581, 0.00000507, -0.00000255, 0.00003718, -0.00000887, -0.00001789, 0.00001209, -0.00000735, 0.00000733, -0.00001289, 0.00004269, -0.00003408, -0.00000086, -0.00000776, 0.00004160, -0.00007014, 0.00009740, -0.00010806, 0.00006581, -0.00006737, 0.00002241, -0.00006071, 0.00011503, -0.00016282, 0.00018635, -0.00023878, 0.00022878, -0.00034322, 0.00037623, -0.00091460, 0.00086167, 0.00082181, -0.00087243, 0.00032388, -0.00029463, 0.00016187, -0.00016256, 0.00011640, -0.00010325, 0.00004724, -0.00004233, 0.00003348, -0.00004241, 0.00012806, -0.00013562, 0.00015226, -0.00016362, 0.00012115, -0.00008221, 0.00007503, -0.00010089, 0.00005994, -0.00001298, -0.00000186, -0.00001056, -0.00000196, -0.00002911, 0.00004074, 0.00000282, -0.00004274, 0.00004074, -0.00001667, -0.00000498, 0.00000006, },
		{ -0.00002980, 0.00001687, -0.00001559, 0.00003771, 0.00000143, -0.00002122, 0.00000858, -0.00000549, 0.00000801, -0.00001892, 0.00004667, -0.00003139, 0.00000147, -0.00002408, 0.00006273, -0.00008547, 0.00009908, -0.00008968, 0.00003968, -0.00003911, 0.00000813, -0.00006465, 0.00012059, -0.00015710, 0.00016684, -0.00020116, 0.00017969, -0.00027422, 0.00027963, -0.00064872, 0.00023531, 0.00138566, -0.00101094, 0.00031752, -0.00029289, 0.00015885, -0.00016753, 0.00011806, -0.00010498, 0.00005685, -0.00006189, 0.00006756, -0.00009825, 0.00017279, -0.00016649, 0.00016481, -0.00014895, 0.00009201, -0.00005371, 0.00005191, -0.00007146, 0.00002306, 0.00001765, -0.00001819, -0.00000157, -0.00000318, -0.00002716, 0.00002409, 0.00002362, -0.00004598, 0.00002773, 0.00000203, -0.00002229, 0.00001598, },
		{ -0.00003869, 0.00002219, -0.00001953, 0.00002941, 0.00001388, -0.00002165, 0.00000270, -0.00000241, 0.00000718, -0.00002124, 0.00004345, -0.00002057, -0.00000444, -0.00002751, 0.00006200, -0.00007036, 0.00006486, -0.00003639, -0.00001616, 0.00001309, -0.00002747, -0.00004060, 0.00008695, -0.00010279, 0.00009163, -0.00010235, 0.00006244, -0.00012509, 0.00008100, -0.00027702, -0.00038374, 0.00176411, -0.00092384, 0.00015861, -0.00019144, 0.00008062, -0.00011085, 0.00006723, -0.00006354, 0.00003002, -0.00004969, 0.00007100, -0.00011474, 0.00016924, -0.00014476, 0.00012281, -0.00008644, 0.00002692, 0.00000086, 0.00000803, -0.00002538, -0.00002361, 0.00005098, -0.00003431, 0.00000754, -0.00000688, -0.00001973, 0.00000500, 0.00003805, -0.00003888, 0.00000802, 0.00002260, -0.00003824, 0.00002846, },
		{ -0.00003787, 0.00001747, -0.00001288, 0.00001382, 0.00002597, -0.00001843, -0.00000474, 0.00000114, 0.00000468, -0.00001879, 0.00003196, -0.00000087, -0.00001985, -0.00001619, 0.00003896, -0.00002833, 0.00000404, 0.00003677, -0.00008332, 0.00007045, -0.00006767, -0.00000374, 0.00003124, -0.00002176, -0.00001133, 0.00002279, -0.00008008, 0.00004972, -0.00014344, 0.00008710, -0.00085848, 0.00185578, -0.00057977, -0.00014301, -0.00000707, -0.00005846, -0.00000463, -0.00002538, 0.00001196, -0.00002641, -0.00000961, 0.00004395, -0.00009071, 0.00011885, -0.00007655, 0.00003879, 0.00000581, -0.00005423, 0.00006282, -0.00003988, 0.00002204, -0.00006647, 0.00007635, -0.00004316, 0.00001255, -0.00001013, -0.00001054, -0.00001171, 0.00004331, -0.00002321, -0.00001334, 0.00003913, -0.00004707, 0.00003217, },
		{ -0.00002609, 0.00000292, 0.00000207, -0.00000490, 0.00003433, -0.00001111, -0.00001257, 0.00000422, 0.00000090, -0.00001199, 0.00001373, 0.00002477, -0.00004108, 0.00000489, 0.00000213, 0.00002660, -0.00006330, 0.00010538, -0.00013639, 0.00010918, -0.00009128, 0.00002515, -0.00002209, 0.00005564, -0.00010527, 0.00013094, -0.00019720, 0.00018922, -0.00031536, 0.00034295, -0.00109284, 0.00162566, -0.00001613, -0.00052457, 0.00020643, -0.00021618, 0.00011634, -0.00012978, 0.00009584, -0.00009101, 0.00004093, 0.00000038, -0.00004022, 0.00003951, 0.00001544, -0.00005983, 0.00009828, -0.00012419, 0.00010932, -0.00007300, 0.00005498, -0.00009267, 0.00008482, -0.00003947, 0.00001043, -0.00001037, -0.00000306, -0.00002250, 0.00003908, -0.00000304, -0.00003053, 0.00004647, -0.00004488, 0.00002452, },
		{ -0.00000613, -0.00001739, 0.00001993, -0.00002105, 0.00003585, -0.00000004, -0.00001922, 0.00000569, -0.00000316, -0.00000278, -0.00000721, 0.00005017, -0.00006046, 0.00002577, -0.00003422, 0.00007494, -0.00011344, 0.00014442, -0.00015252, 0.00011006, -0.00008197, 0.00002915, -0.00005163, 0.00010260, -0.00015863, 0.00018645, -0.00024921, 0.00024804, -0.00038082, 0.00043202, -0.00105761, 0.00111838, 0.00066305, -0.00088184, 0.00037168, -0.00033491, 0.00020548, -0.00020656, 0.00015460, -0.00013545, 0.00007777, -0.00003799, 0.00001350, -0.00004138, 0.00010028, -0.00014035, 0.00016038, -0.00015817, 0.00012193, -0.00007773, 0.00006299, -0.00009484, 0.00007249, -0.00002201, 0.00000079, -0.00000663, 0.00000065, -0.00002597, 0.00002741, 0.00001655, -0.00003883, 0.00004217, -0.00003131, 0.00000681, },
		{ 0.00001600, -0.00003690, 0.00003397, -0.00002932, 0.00002889, 0.00001326, -0.00002291, 0.00000451, -0.00000616, 0.00000589, -0.00002546, 0.00006792, -0.00006924, 0.00003551, -0.00005567, 0.00009912, -0.00012784, 0.00013788, -0.00012076, 0.00006692, -0.00003608, 0.00000306, -0.00004784, 0.00010602, -0.00015654, 0.00017412, -0.00022134, 0.00021181, -0.00032661, 0.00035244, -0.00079583, 0.00044809, 0.00131061, -0.00109692, 0.00041270, -0.00036220, 0.00022186, -0.00022190, 0.00016014, -0.00013616, 0.00008030, -0.00005131, 0.00004788, -0.00009749, 0.00015023, -0.00017666, 0.00017142, -0.00014313, 0.00009396, -0.00005103, 0.00004495, -0.00007368, 0.00004213, 0.00000561, -0.00001366, 0.00000005, 0.00000043, -0.00002290, 0.00001190, 0.00003095, -0.00003612, 0.00002742, -0.00000983, -0.00001595, },
		{ 0.00003313, -0.00004888, 0.00003854, -0.00002661, 0.00001418, 0.00002610, -0.00002194, 0.00000011, -0.00000699, 0.00001119, -0.00003615, 0.00007204, -0.00006118, 0.00002696, -0.00005356, 0.00009029, -0.00010033, 0.00008532, -0.00004726, -0.00000956, 0.00003448, -0.00004363, -0.00001669, 0.00007079, -0.00010545, 0.00010373, -0.00012792, 0.00010040, -0.00018144, 0.00015432, -0.00040623, -0.00023260, 0.00177523, -0.00107449, 0.00028342, -0.00027253, 0.00014750, -0.00016210, 0.00010220, -0.00008514, 0.00004115, -0.00003081, 0.00005035, -0.00011346, 0.00015087, -0.00015848, 0.00012817, -0.00008296, 0.00003338, -0.00000191, 0.00000946, -0.00003770, 0.00000234, 0.00003600, -0.00002797, 0.00000709, -0.00000231, -0.00001564, -0.00000345, 0.00003731, -0.00002353, 0.00000679, 0.00001329, -0.00003673, },
		{ 0.00003957, -0.00004896, 0.00003113, -0.00001328, -0.00000498, 0.00003514, -0.00001536, -0.00000712, -0.00000541, 0.00001149, -0.00003673, 0.00006045, -0.00003550, 0.00000005, -0.00002853, 0.00005182, -0.00003970, 0.00000269, 0.00004591, -0.00009471, 0.00010581, -0.00009024, 0.00002346, 0.00001667, -0.00002924, 0.00000492, -0.00000543, -0.00004148, -0.00000240, -0.00008145, -0.00001063, -0.00077494, 0.00194401, -0.00077494, -0.00001063, -0.00008145, -0.00000240, -0.00004148, -0.00000543, 0.00000492, -0.00002924, 0.00001667, 0.00002346, -0.00009024, 0.00010581, -0.00009471, 0.00004591, 0.00000269, -0.00003970, 0.00005182, -0.00002853, 0.00000005, -0.00003550, 0.00006045, -0.00003673, 0.00001149, -0.00000541, -0.00000712, -0.00001536, 0.00003514, -0.00000498, -0.00001328, 0.00003113, -0.00004896, },
		{ 0.00003313, -0.00003673, 0.00001329, 0.00000679, -0.00002353, 0.00003731, -0.00000345, -0.00001564, -0.00000231, 0.00000709, -0.00002797, 0.00003600, 0.00000234, -0.00003770, 0.00000946, -0.00000191, 0.00003338, -0.00008296, 0.00012817, -0.00015848, 0.00015087, -0.00011346, 0.00005035, -0.00003081, 0.00004115, -0.00008514, 0.00010220, -0.00016210, 0.00014750, -0.00027253, 0.00028342, -0.00107449, 0.00177523, -0.00023260, -0.00040623, 0.00015432, -0.00018144, 0.00010040, -0.00012792, 0.00010373, -0.00010545, 0.00007079, -0.00001669, -0.00004363, 0.00003448, -0.00000956, -0.00004726, 0.00008532, -0.00010033, 0.00009029, -0.00005356, 0.00002696, -0.00006118, 0.00007204, -0.00003615, 0.00001119, -0.00000699, 0.00000011, -0.00002194, 0.00002610, 0.00001418, -0.00002661, 0.00003854, -0.00004888, },
		{ 0.00001600, -0.00001595, -0.00000983, 0.00002742, -0.00003612, 0.00003095, 0.00001190, -0.00002290, 0.00000043, 0.00000005, -0.00001366, 0.00000561, 0.00004213, -0.00007368, 0.00004495, -0.00005103, 0.00009396, -0.00014313, 0.00017142, -0.00017666, 0.00015023, -0.00009749, 0.00004788, -0.00005131, 0.00008030, -0.00013616, 0.00016014, -0.00022190, 0.00022186, -0.00036220, 0.00041270, -0.00109692, 0.00131061, 0.00044809, -0.00079583, 0.00035244, -0.00032661, 0.00021181, -0.00022134, 0.00017412, -0.00015654, 0.00010602, -0.00004784, 0.00000306, -0.00003608, 0.00006692, -0.00012076, 0.00013788, -0.00012784, 0.00009912, -0.00005567, 0.00003551, -0.00006924, 0.00006792, -0.00002546, 0.00000589, -0.00000616, 0.00000451, -0.00002291, 0.00001326, 0.00002889, -0.00002932, 0.00003397, -0.00003690, },
		{ -0.00000613, 0.00000681, -0.00003131, 0.00004217, -0.00003883, 0.00001655, 0.00002741, -0.00002597, 0.00000065, -0.00000663, 0.00000079, -0.00002201, 0.00007249, -0.00009484, 0.00006299, -0.00007773, 0.00012193, -0.00015817, 0.00016038, -0.00014035, 0.00010028, -0.00004138, 0.00001350, -0.00003799, 0.00007777, -0.00013545, 0.00015460, -0.00020656, 0.00020548, -0.00033491, 0.00037168, -0.00088184, 0.00066305, 0.00111838, -0.00105761, 0.00043202, -0.00038082, 0.00024804, -0.00024921, 0.00018645, -0.00015863, 0.00010260, -0.00005163, 0.00002915, -0.00008197, 0.00011006, -0.00015252, 0.00014442, -0.00011344, 0.00007494, -0.00003422, 0.00002577, -0.00006046, 0.00005017, -0.00000721, -0.00000278, -0.00000316, 0.00000569, -0.00001922, -0.00000004, 0.00003585, -0.00002105, 0.00001993, -0.00001739, },
		{ -0.00002609, 0.00002452, -0.00004488, 0.00004647, -0.00003053, -0.00000304, 0.00003908, -0.00002250, -0.00000306, -0.00001037, 0.00001043, -0.00003947, 0.00008482, -0.00009267, 0.00005498, -0.00007300, 0.00010932, -0.00012419, 0.00009828, -0.00005983, 0.00001544, 0.00003951, -0.00004022, 0.00000038, 0.00004093, -0.00009101, 0.00009584, -0.00012978, 0.00011634, -0.00021618, 0.00020643, -0.00052457, -0.00001613, 0.00162566, -0.00109284, 0.00034295, -0.00031536, 0.00018922, -0.00019720, 0.00013094, -0.00010527, 0.00005564, -0.00002209, 0.00002515, -0.00009128, 0.00010918, -0.00013639, 0.00010538, -0.00006330, 0.00002660, 0.00000213, 0.00000489, -0.00004108, 0.00002477, 0.00001373, -0.00001199, 0.00000090, 0.00000422, -0.00001257, -0.00001111, 0.00003433, -0.00000490, 0.00000207, 0.00000292, },
		{ -0.00003787, 0.00003217, -0.00004707, 0.00003913, -0.00001334, -0.00002321, 0.00004331, -0.00001171, -0.00001054, -0.00001013, 0.00001255, -0.00004316, 0.00007635, -0.00006647, 0.00002204, -0.00003988, 0.00006282, -0.00005423, 0.00000581, 0.00003879, -0.00007655, 0.00011885, -0.00009071, 0.00004395, -0.00000961, -0.00002641, 0.00001196, -0.00002538, -0.00000463, -0.00005846, -0.00000707, -0.00014301, -0.00057977, 0.00185578, -0.00085848, 0.00008710, -0.00014344, 0.00004972, -0.00008008, 0.00002279, -0.00001133, -0.00002176, 0.00003124, -0.00000374, -0.00006767, 0.00007045, -0.00008332, 0.00003677, 0.00000404, -0.00002833, 0.00003896, -0.00001619, -0.00001985, -0.00000087, 0.00003196, -0.00001879, 0.00000468, 0.00000114, -0.00000474, -0.00001843, 0.00002597, 0.00001382, -0.00001288, 0.00001747, },
		{ -0.00003869, 0.00002846, -0.00003824, 0.00002260, 0.00000802, -0.00003888, 0.00003805, 0.00000500, -0.00001973, -0.00000688, 0.00000754, -0.00003431, 0.00005098, -0.00002361, -0.00002538, 0.00000803, 0.00000086, 0.00002692, -0.00008644, 0.00012281, -0.00014476, 0.00016924, -0.00011474, 0.00007100, -0.00004969, 0.00003002, -0.00006354, 0.00006723, -0.00011085, 0.00008062, -0.00019144, 0.00015861, -0.00092384, 0.00176411, -0.00038374, -0.00027702, 0.00008100, -0.00012509, 0.00006244, -0.00010235, 0.00009163, -0.00010279, 0.00008695, -0.00004060, -0.00002747, 0.00001309, -0.00001616, -0.00003639, 0.00006486, -0.00007036, 0.00006200, -0.00002751, -0.00000444, -0.00002057, 0.00004345, -0.00002124, 0.00000718, -0.00000241, 0.00000270, -0.00002165, 0.00001388, 0.00002941, -0.00001953, 0.00002219, },
		{ -0.00002980, 0.00001598, -0.00002229, 0.00000203, 0.00002773, -0.00004598, 0.00002362, 0.00002409, -0.00002716, -0.00000318, -0.00000157, -0.00001819, 0.00001765, 0.00002306, -0.00007146, 0.00005191, -0.00005371, 0.00009201, -0.00014895, 0.00016481, -0.00016649, 0.00017279, -0.00009825, 0.00006756, -0.00006189, 0.00005685, -0.00010498, 0.00011806, -0.00016753, 0.00015885, -0.00029289, 0.00031752, -0.00101094, 0.00138566, 0.00023531, -0.00064872, 0.00027963, -0.00027422, 0.00017969, -0.00020116, 0.00016684, -0.00015710, 0.00012059, -0.00006465, 0.00000813, -0.00003911, 0.00003968, -0.00008968, 0.00009908, -0.00008547, 0.00006273, -0.00002408, 0.00000147, -0.00003139, 0.00004667, -0.00001892, 0.00000801, -0.00000549, 0.00000858, -0.00002122, 0.00000143, 0.00003771, -0.00001559, 0.00001687, },
		{ -0.00001581, 0.00000006, -0.00000498, -0.00001667, 0.00004074, -0.00004274, 0.00000282, 0.00004074, -0.00002911, -0.00000196, -0.00001056, -0.00000186, -0.00001298, 0.00005994, -0.00010089, 0.00007503, -0.00008221, 0.00012115, -0.00016362, 0.00015226, -0.00013562, 0.00012806, -0.00004241, 0.00003348, -0.00004233, 0.00004724, -0.00010325, 0.00011640, -0.00016256, 0.00016187, -0.00029463, 0.00032388, -0.00087243, 0.00082181, 0.00086167, -0.00091460, 0.00037623, -0.00034322, 0.00022878, -0.00023878, 0.00018635, -0.00016282, 0.00011503, -0.00006071, 0.00002241, -0.00006737, 0.00006581, -0.00010806, 0.00009740, -0.00007014, 0.00004160, -0.00000776, -0.00000086, -0.00003408, 0.00004269, -0.00001289, 0.00000733, -0.00000735, 0.00001209, -0.00001789, -0.00000887, 0.00003718, -0.00000255, 0.00000507, },
		{ -0.00000267, -0.00001344, 0.00000819, -0.00002881, 0.00004427, -0.00003018, -0.00001969, 0.00005019, -0.00002302, -0.00000513, -0.00001587, 0.00000858, -0.00003238, 0.00007736, -0.00010442, 0.00006883, -0.00007670, 0.00010840, -0.00012951, 0.00009162, -0.00006459, 0.00005058, 0.00003648, -0.00001759, -0.00000124, 0.00001007, -0.00006714, 0.00007212, -0.00010772, 0.00010372, -0.00021582, 0.00021415, -0.00059046, 0.00020882, 0.00135570, -0.00098360, 0.00032388, -0.00030475, 0.00019089, -0.00020210, 0.00014187, -0.00011551, 0.00006802, -0.00002603, 0.00000991, -0.00006481, 0.00005708, -0.00009058, 0.00006402, -0.00003226, 0.00000753, 0.00001390, -0.00000639, -0.00003196, 0.00003422, -0.00000492, 0.00000554, -0.00000745, 0.00001275, -0.00001240, -0.00001583, 0.00002907, 0.00001511, -0.00000736, },
		{ 0.00000462, -0.00002031, 0.00001394, -0.00003239, 0.00003854, -0.00001175, -0.00003870, 0.00004921, -0.00000868, -0.00001247, -0.00001611, 0.00001008, -0.00003685, 0.00007269, -0.00008185, 0.00003572, -0.00004224, 0.00006289, -0.00006212, 0.00000521, 0.00002061, -0.00003322, 0.00011333, -0.00006411, 0.00004251, -0.00003545, -0.00001755, 0.00000904, -0.00003076, 0.00001725, -0.00009848, 0.00005119, -0.00026757, -0.00032112, 0.00161338, -0.00081578, 0.00012317, -0.00016987, 0.00007873, -0.00010481, 0.00004812, -0.00003028, -0.00000635, 0.00002830, -0.00002190, -0.00003789, 0.00002219, -0.00004949, 0.00001412, 0.00001253, -0.00002592, 0.00003106, -0.00000912, -0.00002888, 0.00002418, 0.00000327, 0.00000296, -0.00000547, 0.00001048, -0.00000537, -0.00001955, 0.00001660, 0.00003160, -0.00001458, },
		{ 0.00000393, -0.00001944, 0.00001216, -0.00002845, 0.00002638, 0.00000778, -0.00004991, 0.00003722, 0.00001133, -0.00002149, -0.00001239, 0.00000339, -0.00002827, 0.00005061, -0.00004166, -0.00001235, 0.00000590, 0.00000432, 0.00001327, -0.00007745, 0.00009036, -0.00009609, 0.00016385, -0.00008568, 0.00007014, -0.00006884, 0.00002242, -0.00004616, 0.00003728, -0.00006094, 0.00001124, -0.00010060, 0.00000521, -0.00067399, 0.00159276, -0.00043541, -0.00017540, 0.00001465, -0.00006793, 0.00001767, -0.00006257, 0.00006370, -0.00008288, 0.00008189, -0.00005659, -0.00000250, -0.00002106, -0.00000437, -0.00003267, 0.00004722, -0.00004614, 0.00003600, -0.00000510, -0.00002721, 0.00001445, 0.00001073, -0.00000031, -0.00000139, 0.00000565, 0.00000251, -0.00002077, 0.00000351, 0.00004187, -0.00001289, },
		{ -0.00000337, -0.00001296, 0.00000554, -0.00002030, 0.00001203, 0.00002384, -0.00005113, 0.00001655, 0.00003223, -0.00002826, -0.00000764, -0.00000774, -0.00001275, 0.00002081, 0.00000264, -0.00005889, 0.00004865, -0.00004509, 0.00007137, -0.00013052, 0.00012226, -0.00012001, 0.00017362, -0.00007135, 0.00007104, -0.00007694, 0.00003695, -0.00007460, 0.00007412, -0.00010441, 0.00007983, -0.00019589, 0.00017161, -0.00081318, 0.00132096, 0.00007588, -0.00048598, 0.00018220, -0.00019692, 0.00012165, -0.00015253, 0.00013421, -0.00013492, 0.00011353, -0.00007637, 0.00002359, -0.00005371, 0.00002603, -0.00006005, 0.00006002, -0.00004644, 0.00002635, 0.00000562, -0.00002678, 0.00000536, 0.00001728, -0.00000421, 0.00000433, -0.00000076, 0.00001032, -0.00002025, -0.00000720, 0.00004327, -0.00000201, },
		{ -0.00001305, -0.00000511, -0.00000179, -0.00001199, -0.00000048, 0.00003336, -0.00004280, -0.00000816, 0.00004840, -0.00002891, -0.00000508, -0.00001851, 0.00000213, -0.00000594, 0.00003762, -0.00008906, 0.00007045, -0.00006897, 0.00009579, -0.00014059, 0.00010875, -0.00010220, 0.00014289, -0.00002376, 0.00004688, -0.00005849, 0.00002285, -0.00007123, 0.00007310, -0.00010468, 0.00009555, -0.00021873, 0.00021868, -0.00075986, 0.00088091, 0.00060314, -0.00071392, 0.00026974, -0.00026318, 0.00017145, -0.00019271, 0.00015802, -0.00014466, 0.00010985, -0.00006989, 0.00002833, -0.00006322, 0.00003092, -0.00006097, 0.00004855, -0.00002841, 0.00000607, 0.00001896, -0.00002524, -0.00000386, 0.00002313, -0.00000846, 0.00001062, -0.00000722, 0.00001681, -0.00001829, -0.00001412, 0.00003607, 0.00001498, },
		{ -0.00001982, -0.00000034, -0.00000609, -0.00000667, -0.00000861, 0.00003554, -0.00002767, -0.00003133, 0.00005519, -0.00002129, -0.00000660, -0.00002509, 0.00001028, -0.00002165, 0.00005453, -0.00009487, 0.00006468, -0.00006214, 0.00008395, -0.00011025, 0.00005869, -0.00005503, 0.00008548, 0.00004242, 0.00001022, -0.00002325, -0.00001101, -0.00004451, 0.00004287, -0.00007093, 0.00006788, -0.00018084, 0.00017064, -0.00057690, 0.00038373, 0.00103116, -0.00078535, 0.00024024, -0.00024532, 0.00015242, -0.00017309, 0.00012919, -0.00010970, 0.00007080, -0.00003705, 0.00001005, -0.00004783, 0.00001095, -0.00003951, 0.00002006, -0.00000084, -0.00001639, 0.00002875, -0.00001964, -0.00001400, 0.00002823, -0.00001225, 0.00001591, -0.00001189, 0.00002061, -0.00001467, -0.00001747, 0.00002309, 0.00003287, },
		{ -0.00001942, -0.00000142, -0.00000553, -0.00000547, -0.00001194, 0.00003168, -0.00000983, -0.00004801, 0.00005036, -0.00000597, -0.00001187, -0.00002604, 0.00000908, -0.00002356, 0.00005180, -0.00007728, 0.00003522, -0.00003133, 0.00004639, -0.00005546, -0.00000726, -0.00000006, 0.00002239, 0.00010670, -0.00002134, 0.00001347, -0.00004908, -0.00001074, 0.00000121, -0.00002301, 0.00001941, -0.00011086, 0.00007240, -0.00034357, -0.00006370, 0.00127564, -0.00066968, 0.00009605, -0.00015337, 0.00007552, -0.00010541, 0.00006049, -0.00004356, 0.00000947, 0.00001115, -0.00002263, -0.00001620, -0.00002323, -0.00000833, -0.00001216, 0.00002420, -0.00003172, 0.00002956, -0.00000831, -0.00002477, 0.00003179, -0.00001429, 0.00001851, -0.00001320, 0.00002067, -0.00000902, -0.00001853, 0.00000834, 0.00004622, },
		{ -0.00001018, -0.00000838, -0.00000084, -0.00000735, -0.00001191, 0.00002446, 0.00000662, -0.00005521, 0.00003479, 0.00001373, -0.00001840, -0.00002245, -0.00000025, -0.00001428, 0.00003461, -0.00004492, -0.00000604, 0.00000868, 0.00000127, 0.00000174, -0.00006528, 0.00004097, -0.00002674, 0.00015079, -0.00003268, 0.00003805, -0.00007647, 0.00001384, -0.00003363, 0.00001825, -0.00002557, -0.00004019, -0.00002956, -0.00013055, -0.00038528, 0.00130288, -0.00038788, -0.00012152, -0.00002449, -0.00002799, -0.00001756, -0.00002238, 0.00003012, -0.00005324, 0.00005742, -0.00005513, 0.00001714, -0.00005605, 0.00001717, -0.00003454, 0.00003647, -0.00003371, 0.00001939, 0.00000770, -0.00003434, 0.00003227, -0.00001316, 0.00001713, -0.00001030, 0.00001667, -0.00000128, -0.00001870, -0.00000444, 0.00005125, },
		{ 0.00000634, -0.00001845, 0.00000522, -0.00000980, -0.00001085, 0.00001672, 0.00001870, -0.00005253, 0.00001214, 0.00003274, -0.00002263, -0.00001702, -0.00001380, -0.00000001, 0.00001210, -0.00001006, -0.00004441, 0.00004155, -0.00003333, 0.00004194, -0.00009744, 0.00005436, -0.00005083, 0.00016499, -0.00001702, 0.00004423, -0.00008497, 0.00001959, -0.00005019, 0.00003925, -0.00005049, 0.00000877, -0.00010217, 0.00001734, -0.00054953, 0.00113315, -0.00000453, -0.00034591, 0.00009112, -0.00011929, 0.00005790, -0.00009107, 0.00008677, -0.00009681, 0.00008531, -0.00007343, 0.00003824, -0.00007368, 0.00002512, -0.00003871, 0.00003165, -0.00002180, 0.00000077, 0.00002454, -0.00003970, 0.00002807, -0.00000797, 0.00001148, -0.00000354, 0.00000937, 0.00000788, -0.00001878, -0.00001299, 0.00004691, },
		{ 0.00002602, -0.00002724, 0.00000907, -0.00001014, -0.00001087, 0.00001052, 0.00002519, -0.00004192, -0.00001233, 0.00004596, -0.00002160, -0.00001251, -0.00002689, 0.00001226, -0.00000650, 0.00001620, -0.00006805, 0.00005563, -0.00004615, 0.00005512, -0.00009749, 0.00003883, -0.00005034, 0.00015052, 0.00002215, 0.00003500, -0.00007528, 0.00000616, -0.00004696, 0.00003728, -0.00005097, 0.00002816, -0.00013245, 0.00008654, -0.00056825, 0.00082834, 0.00039294, -0.00050726, 0.00014890, -0.00016710, 0.00009618, -0.00012526, 0.00010996, -0.00010871, 0.00008566, -0.00006985, 0.00003929, -0.00006930, 0.00001172, -0.00002449, 0.00001277, -0.00000122, -0.00002005, 0.00003681, -0.00003790, 0.00001836, 0.00000106, 0.00000254, 0.00000554, 0.00000053, 0.00001704, -0.00001862, -0.00001694, 0.00003493, },
		{ 0.00004362, -0.00003045, 0.00000796, -0.00000680, -0.00001287, 0.00000657, 0.00002659, -0.00002680, -0.00003344, 0.00004993, -0.00001417, -0.00001039, -0.00003600, 0.00001742, -0.00001515, 0.00002764, -0.00007185, 0.00004770, -0.00003599, 0.00004289, -0.00007163, 0.00000445, -0.00003523, 0.00011713, 0.00007341, 0.00002035, -0.00005504, -0.00001920, -0.00003082, 0.00001903, -0.00003341, 0.00002328, -0.00012551, 0.00008945, -0.00048342, 0.00046975, 0.00071997, -0.00055445, 0.00012590, -0.00015906, 0.00008889, -0.00011925, 0.00009657, -0.00008835, 0.00005973, -0.00004576, 0.00002114, -0.00004509, -0.00001794, 0.00000048, -0.00001137, 0.00001937, -0.00003559, 0.00003972, -0.00002755, 0.00000394, 0.00001227, -0.00000754, 0.00001453, -0.00000759, 0.00002438, -0.00001733, -0.00001751, 0.00001898, },
		{ 0.00005459, -0.00002543, 0.00000108, 0.00000000, -0.00001632, 0.00000428, 0.00002451, -0.00001084, -0.00004742, 0.00004384, -0.00000167, -0.00001035, -0.00003967, 0.00001362, -0.00001255, 0.00002432, -0.00005825, 0.00002302, -0.00001071, 0.00001608, -0.00003435, -0.00003275, -0.00001919, 0.00007773, 0.00012265, 0.00001232, -0.00003418, -0.00004619, -0.00001232, -0.00000436, -0.00000960, 0.00000667, -0.00009690, 0.00005250, -0.00034844, 0.00013413, 0.00091721, -0.00047007, 0.00002841, -0.00010517, 0.00004550, -0.00008240, 0.00005643, -0.00004618, 0.00001789, -0.00001032, -0.00000795, -0.00001054, -0.00005247, 0.00002436, -0.00003013, 0.00003150, -0.00004015, 0.00003118, -0.00000986, -0.00001250, 0.00002288, -0.00001600, 0.00002091, -0.00001285, 0.00002824, -0.00001381, -0.00001663, 0.00000322, },
		{ 0.00005644, -0.00001221, -0.00001001, 0.00000834, -0.00001954, 0.00000239, 0.00002086, 0.00000304, -0.00005279, 0.00002957, 0.00001267, -0.00001071, -0.00003842, 0.00000219, -0.00000174, 0.00001140, -0.00003515, -0.00000791, 0.00001723, -0.00001095, -0.00000162, -0.00005804, -0.00001309, 0.00004259, 0.00015842, 0.00001983, -0.00001998, -0.00006650, -0.00000045, -0.00002297, 0.00000942, -0.00000913, -0.00006349, 0.00000342, -0.00021089, -0.00012451, 0.00096279, -0.00027353, -0.00011121, -0.00003248, -0.00001228, -0.00003364, 0.00000695, 0.00000145, -0.00002502, 0.00002375, -0.00003640, 0.00002201, -0.00007910, 0.00003606, -0.00003547, 0.00003040, -0.00003212, 0.00001290, 0.00001126, -0.00002698, 0.00002976, -0.00002039, 0.00002284, -0.00001395, 0.00002772, -0.00000745, -0.00001601, -0.00000904, },
		{ 0.00004927, 0.00000663, -0.00002196, 0.00001534, -0.00002061, -0.00000031, 0.00001713, 0.00001327, -0.00005021, 0.00001086, 0.00002496, -0.00000954, -0.00003381, -0.00001352, 0.00001184, -0.00000368, -0.00001209, -0.00003405, 0.00003627, -0.00002654, 0.00001555, -0.00006375, -0.00002050, 0.00001588, 0.00017519, 0.00004556, -0.00001409, -0.00007684, 0.00000073, -0.00003184, 0.00001771, -0.00001658, -0.00003686, -0.00003785, -0.00010178, -0.00028286, 0.00087209, -0.00001249, -0.00024659, 0.00002650, -0.00006041, 0.00000713, -0.00003442, 0.00003905, -0.00005572, 0.00004550, -0.00005422, 0.00004227, -0.00008856, 0.00002954, -0.00002514, 0.00001681, -0.00001463, -0.00001001, 0.00003023, -0.00003536, 0.00003053, -0.00001938, 0.00001979, -0.00001080, 0.00002297, 0.00000154, -0.00001645, -0.00001622, },
		{ 0.00003551, 0.00002678, -0.00003081, 0.00001835, -0.00001819, -0.00000435, 0.00001398, 0.00001956, -0.00004181, -0.00000812, 0.00003216, -0.00000582, -0.00002733, -0.00002976, 0.00002270, -0.00001430, 0.00000346, -0.00004798, 0.00004011, -0.00002586, 0.00001459, -0.00005138, -0.00003705, -0.00000433, 0.00017345, 0.00008583, -0.00001247, -0.00007916, -0.00000741, -0.00003168, 0.00001537, -0.00001468, -0.00002096, -0.00006248, -0.00003290, -0.00034584, 0.00068712, 0.00025341, -0.00033354, 0.00004672, -0.00008245, 0.00002698, -0.00005680, 0.00005745, -0.00006707, 0.00004959, -0.00005662, 0.00004552, -0.00007819, 0.00000588, -0.00000331, -0.00000347, 0.00000581, -0.00003080, 0.00004164, -0.00003487, 0.00002447, -0.00001327, 0.00001265, -0.00000452, 0.00001518, 0.00001203, -0.00001764, -0.00001855, },
		{ 0.00001885, 0.00004368, -0.00003340, 0.00001595, -0.00001218, -0.00000934, 0.00001136, 0.00002243, -0.00003014, -0.00002406, 0.00003303, -0.00000005, -0.00001973, -0.00004356, 0.00002707, -0.00001681, 0.00000833, -0.00004796, 0.00002918, -0.00001164, 0.00000088, -0.00002960, -0.00005360, -0.00002296, 0.00015722, 0.00013294, -0.00000783, -0.00007835, -0.00002030, -0.00002684, 0.00000661, -0.00000713, -0.00001371, -0.00007096, -0.00000096, -0.00033694, 0.00046039, 0.00047039, -0.00034416, 0.00001970, -0.00007536, 0.00002397, -0.00005879, 0.00005605, -0.00005963, 0.00003755, -0.00004507, 0.00003349, -0.00005213, -0.00002763, 0.00002127, -0.00002215, 0.00002190, -0.00004349, 0.00004211, -0.00002514, 0.00001297, -0.00000394, 0.00000344, 0.00000301, 0.00000611, 0.00002232, -0.00001845, -0.00001763, },
		{ 0.00000304, 0.00005380, -0.00002831, 0.00000848, -0.00000372, -0.00001424, 0.00000887, 0.00002266, -0.00001733, -0.00003526, 0.00002839, 0.00000603, -0.00001114, -0.00005313, 0.00002357, -0.00001112, 0.00000384, -0.00003734, 0.00000926, 0.00000830, -0.00001594, -0.00000945, -0.00006128, -0.00004376, 0.00013074, 0.00017839, 0.00000699, -0.00007897, -0.00003324, -0.00002214, -0.00000351, 0.00000101, -0.00001053, -0.00006858, 0.00000520, -0.00028621, 0.00023961, 0.00060389, -0.00027422, -0.00004431, -0.00004921, 0.00000599, -0.00004707, 0.00004113, -0.00003996, 0.00001598, -0.00002564, 0.00001258, -0.00001901, -0.00006057, 0.00003890, -0.00003182, 0.00002843, -0.00004501, 0.00003146, -0.00000856, -0.00000083, 0.00000578, -0.00000540, 0.00000981, -0.00000241, 0.00003060, -0.00001746, -0.00001554, },
		{ -0.00000914, 0.00005554, -0.00001629, -0.00000214, 0.00000520, -0.00001782, 0.00000620, 0.00002084, -0.00000478, -0.00004149, 0.00002034, 0.00001036, -0.00000172, -0.00005754, 0.00001283, 0.00000021, -0.00000583, -0.00002222, -0.00001177, 0.00002521, -0.00002683, 0.00000079, -0.00005589, -0.00006632, 0.00009635, 0.00021527, 0.00003642, -0.00008271, -0.00004387, -0.00002035, -0.00001170, 0.00000622, -0.00000770, -0.00006101, -0.00000225, -0.00022036, 0.00005752, 0.00064395, -0.00014196, -0.00012132, -0.00002142, -0.00001442, -0.00003176, 0.00002178, -0.00001679, -0.00000686, -0.00000583, -0.00000931, 0.00001165, -0.00008335, 0.00004260, -0.00002884, 0.00002415, -0.00003609, 0.00001279, 0.00001042, -0.00001296, 0.00001307, -0.00001184, 0.00001440, -0.00000896, 0.00003545, -0.00001358, -0.00001407, },
		{ -0.00001643, 0.00004938, 0.00000018, -0.00001308, 0.00001248, -0.00001918, 0.00000337, 0.00001731, 0.00000675, -0.00004346, 0.00001119, 0.00001171, 0.00000768, -0.00005621, -0.00000334, 0.00001348, -0.00001584, -0.00000852, -0.00002742, 0.00003309, -0.00002708, -0.00000122, -0.00003968, -0.00008585, 0.00005458, 0.00023882, 0.00008115, -0.00008761, -0.00005285, -0.00002149, -0.00001737, 0.00000764, -0.00000389, -0.00005184, -0.00001426, -0.00015748, -0.00007130, 0.00060250, 0.00002011, -0.00018334, -0.00000848, -0.00002666, -0.00002114, 0.00000536, 0.00000282, -0.00002446, 0.00000848, -0.00002586, 0.00003277, -0.00009035, 0.00003046, -0.00001453, 0.00001196, -0.00002072, -0.00000868, 0.00002667, -0.00001998, 0.00001608, -0.00001487, 0.00001616, -0.00001280, 0.00003612, -0.00000639, -0.00001412, },
		{ -0.00001907, 0.00003743, 0.00001765, -0.00002152, 0.00001660, -0.00001809, 0.00000073, 0.00001236, 0.00001664, -0.00004187, 0.00000238, 0.00001037, 0.00001521, -0.00004870, -0.00002271, 0.00002518, -0.00002252, 0.00000004, -0.00003461, 0.00003056, -0.00001747, -0.00001189, -0.00001979, -0.00009574, 0.00000622, 0.00024556, 0.00013869, -0.00008894, -0.00006287, -0.00002367, -0.00002187, 0.00000641, 0.00000020, -0.00004231, -0.00002588, -0.00010649, -0.00014748, 0.00050498, 0.00017663, -0.00020849, -0.00001885, -0.00002677, -0.00001819, -0.00000515, 0.00001570, -0.00003396, 0.00001477, -0.00003414, 0.00004141, -0.00008119, 0.00000588, 0.00000557, -0.00000252, -0.00000438, -0.00002745, 0.00003600, -0.00002001, 0.00001448, -0.00001463, 0.00001528, -0.00001381, 0.00003253, 0.00000372, -0.00001562, },
		{ -0.00001837, 0.00002264, 0.00003278, -0.00002545, 0.00001700, -0.00001495, -0.00000133, 0.00000658, 0.00002404, -0.00003696, -0.00000588, 0.00000802, 0.00001862, -0.00003502, -0.00004285, 0.00003266, -0.00002407, 0.00000252, -0.00003381, 0.00002020, -0.00000273, -0.00002445, -0.00000437, -0.00009149, -0.00004520, 0.00023220, 0.00020450, -0.00008083, -0.00007681, -0.00002493, -0.00002678, 0.00000415, 0.00000327, -0.00003252, -0.00003509, -0.00006958, -0.00018087, 0.00038041, 0.00029957, -0.00018779, -0.00004988, -0.00001854, -0.00002033, -0.00001112, 0.00002248, -0.00003594, 0.00001375, -0.00003454, 0.00003866, -0.00005992, -0.00002412, 0.00002411, -0.00001342, 0.00000812, -0.00003951, 0.00003641, -0.00001334, 0.00000947, -0.00001222, 0.00001248, -0.00001227, 0.00002519, 0.00001565, -0.00001764, },
		{ -0.00001612, 0.00000793, 0.00004306, -0.00002410, 0.00001407, -0.00001053, -0.00000265, 0.00000101, 0.00002779, -0.00002851, -0.00001423, 0.00000680, 0.00001630, -0.00001637, -0.00006096, 0.00003411, -0.00002033, -0.00000002, -0.00002763, 0.00000643, 0.00001145, -0.00003268, 0.00000139, -0.00007366, -0.00009236, 0.00019563, 0.00027254, -0.00005770, -0.00009611, -0.00002464, -0.00003261, 0.00000160, 0.00000473, -0.00002273, -0.00004117, -0.00004542, -0.00018435, 0.00025356, 0.00037396, -0.00012638, -0.00008973, -0.00001061, -0.00002216, -0.00001596, 0.00002555, -0.00003274, 0.00000787, -0.00002949, 0.00002815, -0.00003295, -0.00005164, 0.00003472, -0.00001688, 0.00001429, -0.00004340, 0.00002844, -0.00000216, 0.00000325, -0.00000906, 0.00000856, -0.00000865, 0.00001505, 0.00002781, -0.00001880, },
		{ -0.00001399, -0.00000440, 0.00004725, -0.00001790, 0.00000887, -0.00000560, -0.00000352, -0.00000296, 0.00002680, -0.00001639, -0.00002326, 0.00000806, 0.00000845, 0.00000439, -0.00007376, 0.00002831, -0.00001197, -0.00000583, -0.00001892, -0.00000690, 0.00002124, -0.00003381, -0.00000248, -0.00004844, -0.00012562, 0.00013466, 0.00033488, -0.00001497, -0.00012013, -0.00002396, -0.00003843, -0.00000168, 0.00000499, -0.00001377, -0.00004372, -0.00003155, -0.00016975, 0.00014094, 0.00039836, -0.00003977, -0.00012291, -0.00001151, -0.00001932, -0.00002210, 0.00002672, -0.00002648, -0.00000040, -0.00002176, 0.00001410, -0.00000659, -0.00007052, 0.00003413, -0.00001238, 0.00001453, -0.00004019, 0.00001469, 0.00001026, -0.00000183, -0.00000633, 0.00000410, -0.00000347, 0.00000342, 0.00003833, -0.00001765, },
		},
		{ { -0.00000338, -0.00000293, 0.00002005, -0.00000046, -0.00000029, -0.00000076, -0.00001055, 0.00000445, -0.00004299, 0.00000425, 0.00001434, -0.00001497, -0.00001588, -0.00001006, 0.00005244, -0.00000779, -0.00000970, -0.00000566, -0.00003340, -0.00000345, -0.00001368, -0.00001278, -0.00015355, 0.00005426, 0.00036947, 0.00005406, -0.00012865, -0.00002698, -0.00002928, -0.00001304, -0.00000101, -0.00001052, -0.00001601, -0.00001052, -0.00000101, -0.00001304, -0.00002928, -0.00002698, -0.00012865, 0.00005406, 0.00036947, 0.00005426, -0.00015355, -0.00001278, -0.00001368, -0.00000345, -0.00003340, -0.00000566, -0.00000970, -0.00000779, 0.00005244, -0.00001006, -0.00001588, -0.00001497, 0.00001434, 0.00000425, -0.00004299, 0.00000445, -0.00001055, -0.00000076, -0.00000029, -0.00000046, 0.00002005, -0.00000293, },
		{ -0.00000484, -0.00000508, 0.00001398, 0.00001338, -0.00001090, 0.00000979, -0.00002002, 0.00001076, -0.00003679, -0.00001241, 0.00002226, -0.00001478, -0.00001075, -0.00002224, 0.00004672, 0.00001049, -0.00001949, 0.00000322, -0.00003455, -0.00000828, -0.00001427, -0.00000138, -0.00012206, -0.00006975, 0.00037994, 0.00016644, -0.00014858, -0.00003914, -0.00002872, -0.00001992, 0.00000001, -0.00000762, -0.00001687, -0.00001098, -0.00000421, -0.00000730, -0.00002777, -0.00002080, -0.00010399, -0.00002055, 0.00030792, 0.00016603, -0.00014218, -0.00004483, -0.00000590, -0.00000643, -0.00002332, -0.00002021, 0.00000220, -0.00002256, 0.00005028, 0.00000482, -0.00001870, -0.00001578, 0.00000667, 0.00001513, -0.00003892, -0.00000919, 0.00000039, -0.00000873, 0.00000535, -0.00000543, 0.00001571, 0.00000526, },
		{ -0.00000683, -0.00000287, 0.00000207, 0.00002783, -0.00001690, 0.00001313, -0.00001882, 0.00000366, -0.00001903, -0.00003323, 0.00002941, -0.00001538, -0.00000404, -0.00003068, 0.00003519, 0.00002719, -0.00002208, 0.00000275, -0.00002537, -0.00002083, -0.00000700, -0.00000829, -0.00006802, -0.00016717, 0.00031614, 0.00030798, -0.00014957, -0.00006200, -0.00002554, -0.00002693, -0.00000211, -0.00000264, -0.00001842, -0.00001057, -0.00000741, -0.00000399, -0.00002461, -0.00001759, -0.00008133, -0.00006207, 0.00022442, 0.00023862, -0.00008782, -0.00008477, -0.00000042, -0.00001084, -0.00001135, -0.00003253, 0.00000864, -0.00002800, 0.00003818, 0.00002222, -0.00002024, -0.00001536, -0.00000169, 0.00002179, -0.00002976, -0.00002160, 0.00000418, -0.00000772, 0.00000190, -0.00000031, 0.00000351, 0.00001545, },
		{ -0.00000448, -0.00000124, -0.00000770, 0.00003322, -0.00001017, 0.00000361, -0.00000447, -0.00001472, 0.00000403, -0.00005118, 0.00003097, -0.00001330, 0.00000007, -0.00003187, 0.00001783, 0.00004074, -0.00001789, -0.00000379, -0.00001205, -0.00003308, -0.00000074, -0.00001864, -0.00002162, -0.00020468, 0.00017663, 0.00044968, -0.00010961, -0.00010114, -0.00001891, -0.00003425, -0.00000625, 0.00000188, -0.00001804, -0.00001167, -0.00000861, -0.00000362, -0.00002054, -0.00001561, -0.00006266, -0.00008022, 0.00014210, 0.00026405, -0.00000795, -0.00011280, -0.00000728, -0.00001080, -0.00000377, -0.00003650, 0.00000518, -0.00002235, 0.00001775, 0.00003970, -0.00002004, -0.00001348, -0.00001038, 0.00002522, -0.00001962, -0.00002778, -0.00000142, 0.00000081, -0.00000668, 0.00000863, -0.00000818, 0.00002018, },
		{ 0.00000483, -0.00000351, -0.00000994, 0.00002522, 0.00000952, -0.00001505, 0.00001566, -0.00003350, 0.00002021, -0.00005634, 0.00002068, -0.00000369, -0.00000361, -0.00002232, -0.00000561, 0.00005161, -0.00001085, -0.00000933, -0.00000393, -0.00003494, -0.00000597, -0.00001719, -0.00000502, -0.00017340, -0.00000794, 0.00054383, -0.00000545, -0.00015833, -0.00000967, -0.00004240, -0.00001004, 0.00000256, -0.00001315, -0.00001588, -0.00000705, -0.00000566, -0.00001643, -0.00001423, -0.00004745, -0.00008460, 0.00007415, 0.00024955, 0.00007369, -0.00011440, -0.00002818, -0.00000696, -0.00000130, -0.00003205, -0.00000614, -0.00001006, -0.00000439, 0.00005154, -0.00001504, -0.00001265, -0.00001664, 0.00002431, -0.00000927, -0.00002833, -0.00001177, 0.00000923, -0.00001199, 0.00001256, -0.00001097, 0.00001404, },
		{ 0.00001796, -0.00000802, -0.00000548, 0.00000787, 0.00003333, -0.00003110, 0.00002828, -0.00003912, 0.00001852, -0.00004257, -0.00000353, 0.00001451, -0.00001616, -0.00000323, -0.00003159, 0.00005824, -0.00000336, -0.00000944, -0.00000575, -0.00002306, -0.00002459, -0.00000105, -0.00001850, -0.00009688, -0.00018025, 0.00053943, 0.00017286, -0.00022326, -0.00000431, -0.00004963, -0.00001252, -0.00000214, -0.00000359, -0.00002251, -0.00000384, -0.00000869, -0.00001286, -0.00001387, -0.00003399, -0.00008271, 0.00002532, 0.00020907, 0.00013853, -0.00008792, -0.00005431, -0.00000618, 0.00000013, -0.00002396, -0.00001871, 0.00000081, -0.00001982, 0.00005176, -0.00000213, -0.00001568, -0.00001767, 0.00001773, 0.00000210, -0.00002679, -0.00001985, 0.00000990, -0.00000798, 0.00000690, -0.00000250, -0.00000162, },
		{ 0.00002699, -0.00000917, -0.00000006, -0.00000930, 0.00004855, -0.00003207, 0.00002291, -0.00002459, -0.00000271, -0.00001368, -0.00003496, 0.00003483, -0.00003115, 0.00001690, -0.00005089, 0.00005494, 0.00000682, -0.00000631, -0.00001332, -0.00000502, -0.00004587, 0.00001689, -0.00004074, -0.00002045, -0.00027930, 0.00040817, 0.00040478, -0.00026652, -0.00001770, -0.00004966, -0.00001663, -0.00000949, 0.00000683, -0.00002784, -0.00000196, -0.00001062, -0.00001022, -0.00001516, -0.00002086, -0.00007906, -0.00000532, 0.00015671, 0.00017798, -0.00004287, -0.00007250, -0.00001485, 0.00000366, -0.00001661, -0.00002674, 0.00000429, -0.00002369, 0.00003884, 0.00001773, -0.00002184, -0.00001403, 0.00000691, 0.00001371, -0.00002529, -0.00002190, 0.00000092, 0.00000416, -0.00000511, 0.00001168, -0.00001865, },
		{ 0.00002430, -0.00000215, 0.00000071, -0.00001775, 0.00004664, -0.00001334, -0.00000018, 0.00000455, -0.00003261, 0.00001612, -0.00005936, 0.00004501, -0.00003698, 0.00002546, -0.00005229, 0.00003504, 0.00002418, -0.00000641, -0.00001672, 0.00000538, -0.00005310, 0.00001746, -0.00004442, 0.00001407, -0.00027143, 0.00016807, 0.00063126, -0.00024199, -0.00006903, -0.00003387, -0.00002837, -0.00001317, 0.00001125, -0.00002632, -0.00000518, -0.00000925, -0.00000914, -0.00001768, -0.00000861, -0.00007438, -0.00002272, 0.00010440, 0.00019117, 0.00000812, -0.00007462, -0.00003189, 0.00000665, -0.00000972, -0.00002955, 0.00000047, -0.00001812, 0.00001814, 0.00003786, -0.00002568, -0.00001057, -0.00000323, 0.00002205, -0.00002256, -0.00001954, -0.00001267, 0.00001659, -0.00001477, 0.00002187, -0.00002694, },
		{ 0.00000823, 0.00001240, -0.00000389, -0.00001540, 0.00002860, 0.00001811, -0.00002887, 0.00003228, -0.00005299, 0.00002944, -0.00006272, 0.00003464, -0.00002459, 0.00001388, -0.00003052, -0.00000208, 0.00004893, -0.00001334, -0.00000902, -0.00000113, -0.00003660, -0.00000885, -0.00001703, -0.00000648, -0.00017133, -0.00011175, 0.00076720, -0.00010465, -0.00016848, 0.00000042, -0.00005119, -0.00000799, 0.00000421, -0.00001431, -0.00001551, -0.00000351, -0.00001036, -0.00001953, -0.00000015, -0.00006604, -0.00003424, 0.00006180, 0.00018036, 0.00005660, -0.00006264, -0.00004764, 0.00000150, 0.00000066, -0.00003069, -0.00000557, -0.00001037, -0.00000078, 0.00004904, -0.00002025, -0.00001308, -0.00000741, 0.00002331, -0.00001589, -0.00001737, -0.00002299, 0.00002045, -0.00001431, 0.00002107, -0.00002110, },
		{ -0.00001480, 0.00002742, -0.00000890, -0.00000723, 0.00000403, 0.00004736, -0.00004600, 0.00004162, -0.00004942, 0.00001677, -0.00004110, 0.00000374, 0.00000452, -0.00001480, 0.00000722, -0.00004572, 0.00007209, -0.00002198, 0.00000645, -0.00002071, -0.00000316, -0.00005145, 0.00002791, -0.00005813, -0.00003828, -0.00033397, 0.00073555, 0.00016005, -0.00029913, 0.00003964, -0.00007855, 0.00000363, -0.00001280, 0.00000559, -0.00003013, 0.00000463, -0.00001344, -0.00001868, 0.00000117, -0.00005075, -0.00004663, 0.00003500, 0.00014931, 0.00009944, -0.00004515, -0.00005076, -0.00001719, 0.00001625, -0.00003275, -0.00000913, -0.00000694, -0.00001048, 0.00004568, -0.00000289, -0.00002305, -0.00000459, 0.00001731, -0.00000490, -0.00001825, -0.00002517, 0.00001223, -0.00000282, 0.00000994, -0.00000417, },
		{ -0.00003369, 0.00003342, -0.00000759, -0.00000086, -0.00001522, 0.00006044, -0.00003993, 0.00002561, -0.00002113, -0.00001605, -0.00000541, -0.00003439, 0.00003654, -0.00004517, 0.00004213, -0.00007617, 0.00007780, -0.00001997, 0.00001695, -0.00003747, 0.00002610, -0.00008383, 0.00005831, -0.00009516, 0.00005416, -0.00041706, 0.00050987, 0.00050986, -0.00040664, 0.00005043, -0.00009152, 0.00000918, -0.00002909, 0.00002310, -0.00004064, 0.00000945, -0.00001553, -0.00001529, -0.00000517, -0.00002858, -0.00006195, 0.00002385, 0.00010516, 0.00013414, -0.00002875, -0.00003791, -0.00004509, 0.00003106, -0.00003264, -0.00001078, -0.00000836, -0.00001076, 0.00003042, 0.00002105, -0.00003475, -0.00000007, 0.00000909, 0.00000660, -0.00002062, -0.00002061, -0.00000382, 0.00001281, -0.00000372, 0.00001438, },
		{ -0.00003934, 0.00002474, 0.00000298, -0.00000061, -0.00002233, 0.00005242, -0.00001193, -0.00000756, 0.00001724, -0.00004973, 0.00002344, -0.00005957, 0.00005254, -0.00005784, 0.00005361, -0.00007511, 0.00005325, 0.00000301, 0.00000999, -0.00003475, 0.00002976, -0.00008090, 0.00004583, -0.00008085, 0.00005904, -0.00033815, 0.00014171, 0.00084211, -0.00040984, -0.00000830, -0.00006750, -0.00000786, -0.00002979, 0.00002475, -0.00003670, 0.00000391, -0.00001212, -0.00001249, -0.00001504, -0.00000519, -0.00007538, 0.00002060, 0.00006009, 0.00015433, -0.00001155, -0.00001817, -0.00006719, 0.00003241, -0.00002255, -0.00001596, -0.00000934, -0.00000829, 0.00001290, 0.00004034, -0.00003801, -0.00000241, 0.00000592, 0.00001290, -0.00002008, -0.00001498, -0.00001886, 0.00002257, -0.00001050, 0.00002490, },
		{ -0.00003004, 0.00000359, 0.00001894, -0.00000448, -0.00001873, 0.00002948, 0.00002458, -0.00003904, 0.00004398, -0.00006283, 0.00002711, -0.00005736, 0.00004088, -0.00004213, 0.00003246, -0.00003839, 0.00000018, 0.00004436, -0.00001500, -0.00000882, 0.00000229, -0.00003804, -0.00001187, -0.00001381, -0.00001639, -0.00014929, -0.00024806, 0.00102461, -0.00023623, -0.00015808, 0.00000110, -0.00005418, -0.00000725, 0.00000344, -0.00001365, -0.00001492, -0.00000066, -0.00001403, -0.00002169, 0.00001028, -0.00007789, 0.00001324, 0.00002861, 0.00015100, 0.00001484, -0.00000746, -0.00006637, 0.00000962, 0.00000272, -0.00002891, -0.00000439, -0.00001057, 0.00000300, 0.00004507, -0.00002542, -0.00001656, 0.00001132, 0.00001143, -0.00001400, -0.00001316, -0.00002582, 0.00002022, -0.00000616, 0.00002381, },
		{ -0.00001194, -0.00002098, 0.00003155, -0.00000626, -0.00001138, 0.00000399, 0.00005227, -0.00005076, 0.00004370, -0.00004549, 0.00000224, -0.00002954, 0.00000632, -0.00000455, -0.00001151, 0.00001815, -0.00006116, 0.00008515, -0.00004242, 0.00002565, -0.00003949, 0.00002239, -0.00008519, 0.00006814, -0.00011489, 0.00004609, -0.00051709, 0.00095392, 0.00013035, -0.00037104, 0.00008895, -0.00011388, 0.00002855, -0.00003284, 0.00002073, -0.00004060, 0.00001517, -0.00001999, -0.00002075, 0.00001127, -0.00006313, -0.00000698, 0.00002013, 0.00011972, 0.00005591, -0.00001598, -0.00003664, -0.00003498, 0.00003746, -0.00004570, 0.00000536, -0.00001880, 0.00000345, 0.00003396, 0.00000103, -0.00003840, 0.00002081, 0.00000606, -0.00000479, -0.00001513, -0.00002396, 0.00000737, 0.00000555, 0.00001535, },
		{ 0.00000560, -0.00003855, 0.00003287, -0.00000088, -0.00000671, -0.00001355, 0.00006006, -0.00003629, 0.00001716, -0.00000646, -0.00003598, 0.00000542, -0.00003134, 0.00003376, -0.00005377, 0.00006530, -0.00010045, 0.00009894, -0.00004883, 0.00004388, -0.00006600, 0.00006335, -0.00012907, 0.00011087, -0.00016438, 0.00014677, -0.00056833, 0.00061421, 0.00061624, -0.00055896, 0.00013636, -0.00014862, 0.00004984, -0.00006105, 0.00004649, -0.00005757, 0.00002473, -0.00002453, -0.00001442, -0.00000067, -0.00003389, -0.00003834, 0.00003180, 0.00006820, 0.00010497, -0.00003795, 0.00000898, -0.00008202, 0.00006349, -0.00005329, 0.00001074, -0.00002586, 0.00000754, 0.00001601, 0.00002946, -0.00005543, 0.00002348, 0.00000536, 0.00000126, -0.00001640, -0.00001844, -0.00000811, 0.00001578, 0.00000772, },
		{ 0.00001636, -0.00004347, 0.00002095, 0.00001115, -0.00000596, -0.00002011, 0.00004855, -0.00000411, -0.00001917, 0.00003171, -0.00006261, 0.00002341, -0.00004943, 0.00005019, -0.00006975, 0.00007703, -0.00009469, 0.00006870, -0.00001951, 0.00002854, -0.00005502, 0.00005722, -0.00011201, 0.00007983, -0.00012347, 0.00010857, -0.00040073, 0.00010248, 0.00106497, -0.00059783, 0.00007431, -0.00011717, 0.00002622, -0.00005559, 0.00004203, -0.00004931, 0.00001610, -0.00001891, -0.00001046, -0.00001575, -0.00000273, -0.00006796, 0.00004809, 0.00001661, 0.00014244, -0.00005251, 0.00004323, -0.00010334, 0.00005923, -0.00003735, 0.00000093, -0.00002180, 0.00000461, 0.00000447, 0.00004497, -0.00005425, 0.00000946, 0.00001598, -0.00000062, -0.00001268, -0.00001529, -0.00001804, 0.00001711, 0.00000670, },
		{ 0.00002057, -0.00003719, 0.00000086, 0.00002378, -0.00000524, -0.00001974, 0.00002754, 0.00002852, -0.00004362, 0.00004679, -0.00005889, 0.00001082, -0.00003792, 0.00003587, -0.00005068, 0.00004679, -0.00004373, 0.00000088, 0.00003797, -0.00001556, -0.00000879, 0.00000564, -0.00003878, -0.00001467, -0.00000903, -0.00003187, -0.00011338, -0.00040092, 0.00129298, -0.00038594, -0.00013050, -0.00000744, -0.00005115, -0.00000786, 0.00000018, -0.00001115, -0.00001415, 0.00000094, -0.00001578, -0.00002243, 0.00001507, -0.00007951, 0.00004954, -0.00001199, 0.00014652, -0.00003704, 0.00004132, -0.00007952, 0.00001518, 0.00000521, -0.00002621, -0.00000323, -0.00001060, 0.00000698, 0.00003982, -0.00003038, -0.00002155, 0.00003586, -0.00000832, -0.00000469, -0.00001622, -0.00001968, 0.00000880, 0.00001143, },
		{ 0.00002304, -0.00002563, -0.00001905, 0.00002963, 0.00000011, -0.00001821, 0.00000823, 0.00004677, -0.00004288, 0.00003108, -0.00002511, -0.00002525, -0.00000718, 0.00000229, -0.00000967, -0.00000758, 0.00002713, -0.00007420, 0.00009410, -0.00006120, 0.00004494, -0.00005920, 0.00004974, -0.00012041, 0.00011246, -0.00017956, 0.00014316, -0.00070925, 0.00117438, 0.00008947, -0.00043436, 0.00013871, -0.00015228, 0.00006008, -0.00006087, 0.00004106, -0.00005379, 0.00002771, -0.00002895, -0.00001636, 0.00001200, -0.00006491, 0.00002575, -0.00000586, 0.00010872, 0.00001537, -0.00000101, -0.00001559, -0.00005475, 0.00005851, -0.00005801, 0.00002077, -0.00003182, 0.00001916, 0.00001952, 0.00000692, -0.00005723, 0.00005287, -0.00001240, 0.00000118, -0.00001732, -0.00001688, -0.00000311, 0.00001554, },
		{ 0.00002785, -0.00001382, -0.00003269, 0.00002533, 0.00001064, -0.00001762, -0.00000406, 0.00004661, -0.00001972, -0.00000373, 0.00001874, -0.00006039, 0.00001798, -0.00002605, 0.00002685, -0.00005454, 0.00008058, -0.00011785, 0.00011356, -0.00007494, 0.00006993, -0.00009432, 0.00010079, -0.00017406, 0.00016559, -0.00023793, 0.00024458, -0.00071758, 0.00071277, 0.00071567, -0.00070905, 0.00022984, -0.00021449, 0.00010102, -0.00010246, 0.00007526, -0.00007788, 0.00004356, -0.00003846, -0.00000476, -0.00000567, -0.00003160, -0.00001621, 0.00002645, 0.00004299, 0.00008702, -0.00006078, 0.00005706, -0.00011591, 0.00009246, -0.00007171, 0.00003242, -0.00004399, 0.00002708, -0.00000020, 0.00003892, -0.00007839, 0.00005130, -0.00000209, -0.00000223, -0.00001308, -0.00001598, -0.00001076, 0.00001264, },
		{ 0.00003389, -0.00000220, -0.00003937, 0.00001359, 0.00002163, -0.00001501, -0.00001137, 0.00003505, 0.00001022, -0.00003445, 0.00004552, -0.00006859, 0.00001536, -0.00002888, 0.00003708, -0.00006867, 0.00008931, -0.00010602, 0.00007795, -0.00003995, 0.00004613, -0.00007423, 0.00008320, -0.00014100, 0.00011367, -0.00016649, 0.00015699, -0.00045539, 0.00005787, 0.00127804, -0.00078509, 0.00016690, -0.00017612, 0.00006948, -0.00008769, 0.00006148, -0.00006333, 0.00003080, -0.00003001, -0.00000138, -0.00002162, 0.00000082, -0.00005516, 0.00006002, -0.00001940, 0.00014292, -0.00009909, 0.00009645, -0.00013088, 0.00007996, -0.00004904, 0.00001747, -0.00003412, 0.00001760, -0.00000412, 0.00004892, -0.00007070, 0.00002271, 0.00002569, -0.00001585, -0.00000238, -0.00001978, -0.00001076, 0.00000193, },
		{ 0.00003510, 0.00001237, -0.00004220, 0.00000047, 0.00002655, -0.00000612, -0.00001858, 0.00002290, 0.00002980, -0.00004180, 0.00003879, -0.00004030, -0.00001818, -0.00000535, 0.00001899, -0.00004673, 0.00005247, -0.00004476, -0.00000009, 0.00003002, -0.00001486, -0.00000894, 0.00000820, -0.00003827, -0.00001684, -0.00000471, -0.00004870, -0.00007253, -0.00055007, 0.00154424, -0.00053468, -0.00009278, -0.00002276, -0.00004313, -0.00000963, -0.00000521, -0.00000720, -0.00001327, 0.00000099, -0.00001503, -0.00002150, 0.00001344, -0.00006962, 0.00006902, -0.00004766, 0.00015137, -0.00008483, 0.00007544, -0.00008330, 0.00001640, 0.00000839, -0.00002215, -0.00000272, -0.00001045, 0.00001115, 0.00003353, -0.00003499, -0.00002605, 0.00006104, -0.00003086, 0.00000866, -0.00002461, -0.00000673, -0.00001061, },
		{ 0.00002546, 0.00003192, -0.00004333, -0.00000977, 0.00002244, 0.00000927, -0.00002705, 0.00001597, 0.00003175, -0.00002286, 0.00000395, 0.00000975, -0.00006233, 0.00002402, -0.00000838, -0.00000862, -0.00000455, 0.00003232, -0.00008173, 0.00009624, -0.00007512, 0.00006111, -0.00007677, 0.00007470, -0.00015108, 0.00015253, -0.00024089, 0.00023690, -0.00088665, 0.00137371, 0.00004537, -0.00048536, 0.00018352, -0.00018912, 0.00009416, -0.00009345, 0.00006424, -0.00006848, 0.00004128, -0.00003984, -0.00000465, 0.00000158, -0.00005306, 0.00004395, -0.00003067, 0.00010474, -0.00001596, 0.00000083, 0.00000814, -0.00007189, 0.00007452, -0.00006518, 0.00003316, -0.00004299, 0.00003366, 0.00000559, 0.00001162, -0.00007413, 0.00008428, -0.00003256, 0.00001000, -0.00002308, -0.00000613, -0.00001649, },
		{ 0.00000490, 0.00005274, -0.00004025, -0.00001805, 0.00001266, 0.00002518, -0.00003205, 0.00001145, 0.00002216, 0.00000789, -0.00003505, 0.00005083, -0.00008555, 0.00003106, -0.00001911, 0.00001740, -0.00004709, 0.00008410, -0.00012405, 0.00011804, -0.00009407, 0.00009079, -0.00011791, 0.00013234, -0.00021095, 0.00021269, -0.00030406, 0.00033316, -0.00084806, 0.00079677, 0.00079955, -0.00084093, 0.00031808, -0.00027974, 0.00015540, -0.00014773, 0.00010577, -0.00009939, 0.00006415, -0.00005629, 0.00001399, -0.00002120, -0.00001933, -0.00000162, 0.00001564, 0.00002614, 0.00007630, -0.00008732, 0.00009630, -0.00013848, 0.00011074, -0.00008189, 0.00004923, -0.00005869, 0.00004370, -0.00001391, 0.00004517, -0.00009754, 0.00007705, -0.00001018, -0.00000436, -0.00001087, -0.00001365, -0.00001149, },
		{ -0.00001851, 0.00006535, -0.00002683, -0.00002892, 0.00000440, 0.00003281, -0.00002631, 0.00000206, 0.00001403, 0.00002933, -0.00005157, 0.00005601, -0.00006639, 0.00000087, -0.00000039, 0.00001534, -0.00005389, 0.00008672, -0.00010591, 0.00007873, -0.00005539, 0.00005992, -0.00008929, 0.00010333, -0.00016253, 0.00014136, -0.00020385, 0.00019852, -0.00049895, 0.00001659, 0.00145815, -0.00094845, 0.00025455, -0.00023473, 0.00011507, -0.00012141, 0.00008042, -0.00007711, 0.00004640, -0.00004433, 0.00001447, -0.00003267, 0.00000647, -0.00004018, 0.00005850, -0.00004337, 0.00014438, -0.00013751, 0.00013117, -0.00014280, 0.00008910, -0.00005352, 0.00003043, -0.00004377, 0.00002788, -0.00001003, 0.00004951, -0.00008353, 0.00003476, 0.00003378, -0.00002965, 0.00000792, -0.00002637, 0.00000036, },
		{ -0.00003271, 0.00005992, 0.00000123, -0.00004502, 0.00000249, 0.00002720, -0.00000706, -0.00001578, 0.00001605, 0.00002869, -0.00003413, 0.00002057, -0.00000985, -0.00005486, 0.00003596, -0.00000691, -0.00002990, 0.00004705, -0.00004063, -0.00000262, 0.00002096, -0.00001294, -0.00000930, 0.00000934, -0.00003608, -0.00001803, -0.00000280, -0.00006238, -0.00003646, -0.00067435, 0.00174958, -0.00066174, -0.00005412, -0.00004084, -0.00003212, -0.00001210, -0.00001203, -0.00000244, -0.00001233, -0.00000050, -0.00001158, -0.00001899, 0.00000590, -0.00004946, 0.00006981, -0.00006951, 0.00015254, -0.00011747, 0.00008986, -0.00007594, 0.00001240, 0.00001233, -0.00001649, -0.00000334, -0.00001010, 0.00001542, 0.00002654, -0.00003909, -0.00002809, 0.00008146, -0.00005036, 0.00002154, -0.00003481, 0.00000893, },
		{ -0.00002945, 0.00003374, 0.00004018, -0.00006191, 0.00000451, 0.00001203, 0.00001939, -0.00003642, 0.00002531, 0.00001022, 0.00000584, -0.00003372, 0.00005310, -0.00010238, 0.00005973, -0.00002361, -0.00000064, -0.00000396, 0.00003286, -0.00008202, 0.00009024, -0.00008195, 0.00007155, -0.00008958, 0.00009352, -0.00017185, 0.00018060, -0.00028797, 0.00031141, -0.00102596, 0.00152920, 0.00000676, -0.00052137, 0.00021817, -0.00021925, 0.00012563, -0.00012573, 0.00008677, -0.00008260, 0.00005413, -0.00005110, 0.00001320, -0.00001875, -0.00002840, 0.00004440, -0.00004542, 0.00009850, -0.00003427, -0.00000785, 0.00003043, -0.00008280, 0.00008201, -0.00006406, 0.00003954, -0.00004987, 0.00004433, -0.00000496, 0.00001243, -0.00008410, 0.00010785, -0.00004857, 0.00001793, -0.00002968, 0.00000516, },
		{ -0.00001068, -0.00000462, 0.00007682, -0.00006822, 0.00000148, -0.00000190, 0.00003984, -0.00004728, 0.00002952, -0.00000867, 0.00004175, -0.00007178, 0.00008381, -0.00010681, 0.00004334, -0.00001127, 0.00000802, -0.00003337, 0.00007482, -0.00011734, 0.00011084, -0.00010346, 0.00010323, -0.00013318, 0.00015303, -0.00023318, 0.00024363, -0.00035204, 0.00039910, -0.00094435, 0.00085820, 0.00086015, -0.00093954, 0.00038791, -0.00033392, 0.00020415, -0.00018976, 0.00013321, -0.00011888, 0.00008360, -0.00007537, 0.00003951, -0.00004520, 0.00000200, 0.00000243, 0.00000567, 0.00001441, 0.00006901, -0.00010916, 0.00011961, -0.00014460, 0.00011388, -0.00008000, 0.00005761, -0.00006691, 0.00005414, -0.00002172, 0.00004490, -0.00010770, 0.00009495, -0.00001581, -0.00000588, -0.00000976, -0.00001171, },
		{ 0.00001163, -0.00003837, 0.00009400, -0.00005209, -0.00001528, -0.00000420, 0.00004255, -0.00003767, 0.00001712, -0.00001075, 0.00004992, -0.00006779, 0.00006082, -0.00005643, -0.00001638, 0.00003151, -0.00001009, -0.00002837, 0.00006904, -0.00009383, 0.00007051, -0.00006402, 0.00006788, -0.00009797, 0.00011434, -0.00017252, 0.00015802, -0.00023022, 0.00022806, -0.00052884, -0.00001328, 0.00158438, -0.00106628, 0.00032219, -0.00028250, 0.00015503, -0.00015118, 0.00009564, -0.00008850, 0.00006056, -0.00005956, 0.00003503, -0.00004726, 0.00001399, -0.00002588, 0.00004718, -0.00005355, 0.00013864, -0.00015831, 0.00014184, -0.00013597, 0.00008404, -0.00004859, 0.00003765, -0.00004898, 0.00003341, -0.00001129, 0.00004493, -0.00008972, 0.00004338, 0.00003916, -0.00003916, 0.00001526, -0.00003237, },
		{ 0.00002254, -0.00005159, 0.00008051, -0.00001039, -0.00004578, 0.00000666, 0.00002596, -0.00000798, -0.00001209, 0.00000778, 0.00002569, -0.00002232, -0.00000488, 0.00002639, -0.00009137, 0.00007911, -0.00003598, -0.00000327, 0.00003066, -0.00003164, -0.00000637, 0.00001146, -0.00001004, -0.00000984, 0.00000869, -0.00003204, -0.00001801, -0.00000472, -0.00006924, -0.00001358, -0.00075549, 0.00188410, -0.00074839, -0.00002379, -0.00005706, -0.00002058, -0.00001468, -0.00001937, 0.00000231, -0.00001141, -0.00000321, -0.00000562, -0.00001523, -0.00000583, -0.00002259, 0.00005406, -0.00007351, 0.00014177, -0.00012890, 0.00008372, -0.00005813, 0.00000348, 0.00001684, -0.00000930, -0.00000526, -0.00000955, 0.00001958, 0.00001933, -0.00004246, -0.00002626, 0.00009269, -0.00006180, 0.00002963, -0.00004294, },
		{ 0.00001381, -0.00003886, 0.00003919, 0.00004581, -0.00007777, 0.00002050, 0.00000120, 0.00002816, -0.00004439, 0.00003369, -0.00001344, 0.00003751, -0.00007501, 0.00009597, -0.00013692, 0.00009420, -0.00003950, 0.00001331, -0.00000705, 0.00002903, -0.00007505, 0.00007645, -0.00008089, 0.00007479, -0.00009593, 0.00010365, -0.00017910, 0.00019117, -0.00031247, 0.00035400, -0.00110824, 0.00162225, -0.00001868, -0.00054083, 0.00023850, -0.00023797, 0.00014922, -0.00015243, 0.00010474, -0.00009364, 0.00006409, -0.00006043, 0.00003428, -0.00004511, 0.00000474, 0.00002854, -0.00004588, 0.00008448, -0.00003870, -0.00002322, 0.00004788, -0.00008553, 0.00007956, -0.00005352, 0.00003847, -0.00005109, 0.00004950, -0.00001025, 0.00000785, -0.00008382, 0.00011823, -0.00005577, 0.00002169, -0.00003442, },
		{ -0.00001033, -0.00000908, -0.00001273, 0.00009441, -0.00009174, 0.00002118, -0.00001454, 0.00005130, -0.00006001, 0.00004580, -0.00003985, 0.00007515, -0.00010692, 0.00011044, -0.00011963, 0.00005382, -0.00000312, 0.00000160, -0.00001729, 0.00005479, -0.00009903, 0.00009292, -0.00010233, 0.00010568, -0.00013816, 0.00015991, -0.00023681, 0.00025282, -0.00037424, 0.00043242, -0.00099472, 0.00089094, 0.00089164, -0.00099302, 0.00042830, -0.00036755, 0.00023874, -0.00022135, 0.00015260, -0.00013281, 0.00009847, -0.00009198, 0.00006754, -0.00007310, 0.00002857, -0.00000380, 0.00000053, 0.00000520, 0.00006214, -0.00012070, 0.00012393, -0.00013314, 0.00010108, -0.00006523, 0.00005615, -0.00006722, 0.00005665, -0.00002185, 0.00003663, -0.00010582, 0.00010109, -0.00001685, -0.00000738, -0.00000963, },
		{ -0.00003544, 0.00001949, -0.00005205, 0.00011316, -0.00007250, -0.00000200, -0.00000973, 0.00004867, -0.00004663, 0.00003032, -0.00003471, 0.00006796, -0.00008001, 0.00005829, -0.00004086, -0.00003133, 0.00006137, -0.00003265, 0.00000173, 0.00003967, -0.00007205, 0.00005469, -0.00006539, 0.00006915, -0.00009929, 0.00011467, -0.00016898, 0.00016075, -0.00024202, 0.00024195, -0.00054319, -0.00002586, 0.00164143, -0.00112224, 0.00035761, -0.00031031, 0.00018209, -0.00017169, 0.00010411, -0.00009534, 0.00007073, -0.00007272, 0.00005694, -0.00006250, 0.00002212, -0.00001444, 0.00003057, -0.00005109, 0.00012177, -0.00015756, 0.00012842, -0.00011165, 0.00006556, -0.00003442, 0.00003847, -0.00004903, 0.00003335, -0.00000727, 0.00003485, -0.00008770, 0.00004715, 0.00004107, -0.00004242, 0.00001748, },
		{ -0.00004602, 0.00003012, -0.00006223, 0.00009233, -0.00002013, -0.00004481, 0.00001249, 0.00002325, -0.00000883, -0.00000834, -0.00000097, 0.00002150, -0.00000887, -0.00003287, 0.00006035, -0.00011868, 0.00011644, -0.00006064, 0.00002723, 0.00000592, -0.00001925, -0.00001079, 0.00000231, -0.00000660, -0.00001056, 0.00000624, -0.00002631, -0.00001681, -0.00001087, -0.00006735, -0.00000906, -0.00078140, 0.00193092, -0.00078140, -0.00000906, -0.00006735, -0.00001087, -0.00001681, -0.00002631, 0.00000624, -0.00001056, -0.00000660, 0.00000231, -0.00001079, -0.00001925, 0.00000592, 0.00002723, -0.00006064, 0.00011644, -0.00011868, 0.00006035, -0.00003287, -0.00000887, 0.00002150, -0.00000097, -0.00000834, -0.00000883, 0.00002325, 0.00001249, -0.00004481, -0.00002013, 0.00009233, -0.00006223, 0.00003012, },
		{ -0.00003544, 0.00001748, -0.00004242, 0.00004107, 0.00004715, -0.00008770, 0.00003485, -0.00000727, 0.00003335, -0.00004903, 0.00003847, -0.00003442, 0.00006556, -0.00011165, 0.00012842, -0.00015756, 0.00012177, -0.00005109, 0.00003057, -0.00001444, 0.00002212, -0.00006250, 0.00005694, -0.00007272, 0.00007073, -0.00009534, 0.00010411, -0.00017169, 0.00018209, -0.00031031, 0.00035761, -0.00112224, 0.00164143, -0.00002586, -0.00054319, 0.00024195, -0.00024202, 0.00016075, -0.00016898, 0.00011467, -0.00009929, 0.00006915, -0.00006539, 0.00005469, -0.00007205, 0.00003967, 0.00000173, -0.00003265, 0.00006137, -0.00003133, -0.00004086, 0.00005829, -0.00008001, 0.00006796, -0.00003471, 0.00003032, -0.00004663, 0.00004867, -0.00000973, -0.00000200, -0.00007250, 0.00011316, -0.00005205, 0.00001949, },
		{ -0.00001033, -0.00000963, -0.00000738, -0.00001685, 0.00010109, -0.00010582, 0.00003663, -0.00002185, 0.00005665, -0.00006722, 0.00005615, -0.00006523, 0.00010108, -0.00013314, 0.00012393, -0.00012070, 0.00006214, 0.00000520, 0.00000053, -0.00000380, 0.00002857, -0.00007310, 0.00006754, -0.00009198, 0.00009847, -0.00013281, 0.00015260, -0.00022135, 0.00023874, -0.00036755, 0.00042830, -0.00099302, 0.00089164, 0.00089094, -0.00099472, 0.00043242, -0.00037424, 0.00025282, -0.00023681, 0.00015991, -0.00013816, 0.00010568, -0.00010233, 0.00009292, -0.00009903, 0.00005479, -0.00001729, 0.00000160, -0.00000312, 0.00005382, -0.00011963, 0.00011044, -0.00010692, 0.00007515, -0.00003985, 0.00004580, -0.00006001, 0.00005130, -0.00001454, 0.00002118, -0.00009174, 0.00009441, -0.00001273, -0.00000908, },
		{ 0.00001381, -0.00003442, 0.00002169, -0.00005577, 0.00011823, -0.00008382, 0.00000785, -0.00001025, 0.00004950, -0.00005109, 0.00003847, -0.00005352, 0.00007956, -0.00008553, 0.00004788, -0.00002322, -0.00003870, 0.00008448, -0.00004588, 0.00002854, 0.00000474, -0.00004511, 0.00003428, -0.00006043, 0.00006409, -0.00009364, 0.00010474, -0.00015243, 0.00014922, -0.00023797, 0.00023850, -0.00054083, -0.00001868, 0.00162225, -0.00110824, 0.00035400, -0.00031247, 0.00019117, -0.00017910, 0.00010365, -0.00009593, 0.00007479, -0.00008089, 0.00007645, -0.00007505, 0.00002903, -0.00000705, 0.00001331, -0.00003950, 0.00009420, -0.00013692, 0.00009597, -0.00007501, 0.00003751, -0.00001344, 0.00003369, -0.00004439, 0.00002816, 0.00000120, 0.00002050, -0.00007777, 0.00004581, 0.00003919, -0.00003886, },
		{ 0.00002254, -0.00004294, 0.00002963, -0.00006180, 0.00009269, -0.00002626, -0.00004246, 0.00001933, 0.00001958, -0.00000955, -0.00000526, -0.00000930, 0.00001684, 0.00000348, -0.00005813, 0.00008372, -0.00012890, 0.00014177, -0.00007351, 0.00005406, -0.00002259, -0.00000583, -0.00001523, -0.00000562, -0.00000321, -0.00001141, 0.00000231, -0.00001937, -0.00001468, -0.00002058, -0.00005706, -0.00002379, -0.00074839, 0.00188410, -0.00075549, -0.00001358, -0.00006924, -0.00000472, -0.00001801, -0.00003204, 0.00000869, -0.00000984, -0.00001004, 0.00001146, -0.00000637, -0.00003164, 0.00003066, -0.00000327, -0.00003598, 0.00007911, -0.00009137, 0.00002639, -0.00000488, -0.00002232, 0.00002569, 0.00000778, -0.00001209, -0.00000798, 0.00002596, 0.00000666, -0.00004578, -0.00001039, 0.00008051, -0.00005159, },
		{ 0.00001163, -0.00003237, 0.00001526, -0.00003916, 0.00003916, 0.00004338, -0.00008972, 0.00004493, -0.00001129, 0.00003341, -0.00004898, 0.00003765, -0.00004859, 0.00008404, -0.00013597, 0.00014184, -0.00015831, 0.00013864, -0.00005355, 0.00004718, -0.00002588, 0.00001399, -0.00004726, 0.00003503, -0.00005956, 0.00006056, -0.00008850, 0.00009564, -0.00015118, 0.00015503, -0.00028250, 0.00032219, -0.00106628, 0.00158438, -0.00001328, -0.00052884, 0.00022806, -0.00023022, 0.00015802, -0.00017252, 0.00011434, -0.00009797, 0.00006788, -0.00006402, 0.00007051, -0.00009383, 0.00006904, -0.00002837, -0.00001009, 0.00003151, -0.00001638, -0.00005643, 0.00006082, -0.00006779, 0.00004992, -0.00001075, 0.00001712, -0.00003767, 0.00004255, -0.00000420, -0.00001528, -0.00005209, 0.00009400, -0.00003837, },
		{ -0.00001068, -0.00001171, -0.00000976, -0.00000588, -0.00001581, 0.00009495, -0.00010770, 0.00004490, -0.00002172, 0.00005414, -0.00006691, 0.00005761, -0.00008000, 0.00011388, -0.00014460, 0.00011961, -0.00010916, 0.00006901, 0.00001441, 0.00000567, 0.00000243, 0.00000200, -0.00004520, 0.00003951, -0.00007537, 0.00008360, -0.00011888, 0.00013321, -0.00018976, 0.00020415, -0.00033392, 0.00038791, -0.00093954, 0.00086015, 0.00085820, -0.00094435, 0.00039910, -0.00035204, 0.00024363, -0.00023318, 0.00015303, -0.00013318, 0.00010323, -0.00010346, 0.00011084, -0.00011734, 0.00007482, -0.00003337, 0.00000802, -0.00001127, 0.00004334, -0.00010681, 0.00008381, -0.00007178, 0.00004175, -0.00000867, 0.00002952, -0.00004728, 0.00003984, -0.00000190, 0.00000148, -0.00006822, 0.00007682, -0.00000462, },
		{ -0.00002945, 0.00000516, -0.00002968, 0.00001793, -0.00004857, 0.00010785, -0.00008410, 0.00001243, -0.00000496, 0.00004433, -0.00004987, 0.00003954, -0.00006406, 0.00008201, -0.00008280, 0.00003043, -0.00000785, -0.00003427, 0.00009850, -0.00004542, 0.00004440, -0.00002840, -0.00001875, 0.00001320, -0.00005110, 0.00005413, -0.00008260, 0.00008677, -0.00012573, 0.00012563, -0.00021925, 0.00021817, -0.00052137, 0.00000676, 0.00152920, -0.00102596, 0.00031141, -0.00028797, 0.00018060, -0.00017185, 0.00009352, -0.00008958, 0.00007155, -0.00008195, 0.00009024, -0.00008202, 0.00003286, -0.00000396, -0.00000064, -0.00002361, 0.00005973, -0.00010238, 0.00005310, -0.00003372, 0.00000584, 0.00001022, 0.00002531, -0.00003642, 0.00001939, 0.00001203, 0.00000451, -0.00006191, 0.00004018, 0.00003374, },
		{ -0.00003271, 0.00000893, -0.00003481, 0.00002154, -0.00005036, 0.00008146, -0.00002809, -0.00003909, 0.00002654, 0.00001542, -0.00001010, -0.00000334, -0.00001649, 0.00001233, 0.00001240, -0.00007594, 0.00008986, -0.00011747, 0.00015254, -0.00006951, 0.00006981, -0.00004946, 0.00000590, -0.00001899, -0.00001158, -0.00000050, -0.00001233, -0.00000244, -0.00001203, -0.00001210, -0.00003212, -0.00004084, -0.00005412, -0.00066174, 0.00174958, -0.00067435, -0.00003646, -0.00006238, -0.00000280, -0.00001803, -0.00003608, 0.00000934, -0.00000930, -0.00001294, 0.00002096, -0.00000262, -0.00004063, 0.00004705, -0.00002990, -0.00000691, 0.00003596, -0.00005486, -0.00000985, 0.00002057, -0.00003413, 0.00002869, 0.00001605, -0.00001578, -0.00000706, 0.00002720, 0.00000249, -0.00004502, 0.00000123, 0.00005992, },
		{ -0.00001851, 0.00000036, -0.00002637, 0.00000792, -0.00002965, 0.00003378, 0.00003476, -0.00008353, 0.00004951, -0.00001003, 0.00002788, -0.00004377, 0.00003043, -0.00005352, 0.00008910, -0.00014280, 0.00013117, -0.00013751, 0.00014438, -0.00004337, 0.00005850, -0.00004018, 0.00000647, -0.00003267, 0.00001447, -0.00004433, 0.00004640, -0.00007711, 0.00008042, -0.00012141, 0.00011507, -0.00023473, 0.00025455, -0.00094845, 0.00145815, 0.00001659, -0.00049895, 0.00019852, -0.00020385, 0.00014136, -0.00016253, 0.00010333, -0.00008929, 0.00005992, -0.00005539, 0.00007873, -0.00010591, 0.00008672, -0.00005389, 0.00001534, -0.00000039, 0.00000087, -0.00006639, 0.00005601, -0.00005157, 0.00002933, 0.00001403, 0.00000206, -0.00002631, 0.00003281, 0.00000440, -0.00002892, -0.00002683, 0.00006535, },
		{ 0.00000490, -0.00001149, -0.00001365, -0.00001087, -0.00000436, -0.00001018, 0.00007705, -0.00009754, 0.00004517, -0.00001391, 0.00004370, -0.00005869, 0.00004923, -0.00008189, 0.00011074, -0.00013848, 0.00009630, -0.00008732, 0.00007630, 0.00002614, 0.00001564, -0.00000162, -0.00001933, -0.00002120, 0.00001399, -0.00005629, 0.00006415, -0.00009939, 0.00010577, -0.00014773, 0.00015540, -0.00027974, 0.00031808, -0.00084093, 0.00079955, 0.00079677, -0.00084806, 0.00033316, -0.00030406, 0.00021269, -0.00021095, 0.00013234, -0.00011791, 0.00009079, -0.00009407, 0.00011804, -0.00012405, 0.00008410, -0.00004709, 0.00001740, -0.00001911, 0.00003106, -0.00008555, 0.00005083, -0.00003505, 0.00000789, 0.00002216, 0.00001145, -0.00003205, 0.00002518, 0.00001266, -0.00001805, -0.00004025, 0.00005274, },
		{ 0.00002546, -0.00001649, -0.00000613, -0.00002308, 0.00001000, -0.00003256, 0.00008428, -0.00007413, 0.00001162, 0.00000559, 0.00003366, -0.00004299, 0.00003316, -0.00006518, 0.00007452, -0.00007189, 0.00000814, 0.00000083, -0.00001596, 0.00010474, -0.00003067, 0.00004395, -0.00005306, 0.00000158, -0.00000465, -0.00003984, 0.00004128, -0.00006848, 0.00006424, -0.00009345, 0.00009416, -0.00018912, 0.00018352, -0.00048536, 0.00004537, 0.00137371, -0.00088665, 0.00023690, -0.00024089, 0.00015253, -0.00015108, 0.00007470, -0.00007677, 0.00006111, -0.00007512, 0.00009624, -0.00008173, 0.00003232, -0.00000455, -0.00000862, -0.00000838, 0.00002402, -0.00006233, 0.00000975, 0.00000395, -0.00002286, 0.00003175, 0.00001597, -0.00002705, 0.00000927, 0.00002244, -0.00000977, -0.00004333, 0.00003192, },
		{ 0.00003510, -0.00001061, -0.00000673, -0.00002461, 0.00000866, -0.00003086, 0.00006104, -0.00002605, -0.00003499, 0.00003353, 0.00001115, -0.00001045, -0.00000272, -0.00002215, 0.00000839, 0.00001640, -0.00008330, 0.00007544, -0.00008483, 0.00015137, -0.00004766, 0.00006902, -0.00006962, 0.00001344, -0.00002150, -0.00001503, 0.00000099, -0.00001327, -0.00000720, -0.00000521, -0.00000963, -0.00004313, -0.00002276, -0.00009278, -0.00053468, 0.00154424, -0.00055007, -0.00007253, -0.00004870, -0.00000471, -0.00001684, -0.00003827, 0.00000820, -0.00000894, -0.00001486, 0.00003002, -0.00000009, -0.00004476, 0.00005247, -0.00004673, 0.00001899, -0.00000535, -0.00001818, -0.00004030, 0.00003879, -0.00004180, 0.00002980, 0.00002290, -0.00001858, -0.00000612, 0.00002655, 0.00000047, -0.00004220, 0.00001237, },
		{ 0.00003389, 0.00000193, -0.00001076, -0.00001978, -0.00000238, -0.00001585, 0.00002569, 0.00002271, -0.00007070, 0.00004892, -0.00000412, 0.00001760, -0.00003412, 0.00001747, -0.00004904, 0.00007996, -0.00013088, 0.00009645, -0.00009909, 0.00014292, -0.00001940, 0.00006002, -0.00005516, 0.00000082, -0.00002162, -0.00000138, -0.00003001, 0.00003080, -0.00006333, 0.00006148, -0.00008769, 0.00006948, -0.00017612, 0.00016690, -0.00078509, 0.00127804, 0.00005787, -0.00045539, 0.00015699, -0.00016649, 0.00011367, -0.00014100, 0.00008320, -0.00007423, 0.00004613, -0.00003995, 0.00007795, -0.00010602, 0.00008931, -0.00006867, 0.00003708, -0.00002888, 0.00001536, -0.00006859, 0.00004552, -0.00003445, 0.00001022, 0.00003505, -0.00001137, -0.00001501, 0.00002163, 0.00001359, -0.00003937, -0.00000220, },
		{ 0.00002785, 0.00001264, -0.00001076, -0.00001598, -0.00001308, -0.00000223, -0.00000209, 0.00005130, -0.00007839, 0.00003892, -0.00000020, 0.00002708, -0.00004399, 0.00003242, -0.00007171, 0.00009246, -0.00011591, 0.00005706, -0.00006078, 0.00008702, 0.00004299, 0.00002645, -0.00001621, -0.00003160, -0.00000567, -0.00000476, -0.00003846, 0.00004356, -0.00007788, 0.00007526, -0.00010246, 0.00010102, -0.00021449, 0.00022984, -0.00070905, 0.00071567, 0.00071277, -0.00071758, 0.00024458, -0.00023793, 0.00016559, -0.00017406, 0.00010079, -0.00009432, 0.00006993, -0.00007494, 0.00011356, -0.00011785, 0.00008058, -0.00005454, 0.00002685, -0.00002605, 0.00001798, -0.00006039, 0.00001874, -0.00000373, -0.00001972, 0.00004661, -0.00000406, -0.00001762, 0.00001064, 0.00002533, -0.00003269, -0.00001382, },
		{ 0.00002304, 0.00001554, -0.00000311, -0.00001688, -0.00001732, 0.00000118, -0.00001240, 0.00005287, -0.00005723, 0.00000692, 0.00001952, 0.00001916, -0.00003182, 0.00002077, -0.00005801, 0.00005851, -0.00005475, -0.00001559, -0.00000101, 0.00001537, 0.00010872, -0.00000586, 0.00002575, -0.00006491, 0.00001200, -0.00001636, -0.00002895, 0.00002771, -0.00005379, 0.00004106, -0.00006087, 0.00006008, -0.00015228, 0.00013871, -0.00043436, 0.00008947, 0.00117438, -0.00070925, 0.00014316, -0.00017956, 0.00011246, -0.00012041, 0.00004974, -0.00005920, 0.00004494, -0.00006120, 0.00009410, -0.00007420, 0.00002713, -0.00000758, -0.00000967, 0.00000229, -0.00000718, -0.00002525, -0.00002511, 0.00003108, -0.00004288, 0.00004677, 0.00000823, -0.00001821, 0.00000011, 0.00002963, -0.00001905, -0.00002563, },
		{ 0.00002057, 0.00001143, 0.00000880, -0.00001968, -0.00001622, -0.00000469, -0.00000832, 0.00003586, -0.00002155, -0.00003038, 0.00003982, 0.00000698, -0.00001060, -0.00000323, -0.00002621, 0.00000521, 0.00001518, -0.00007952, 0.00004132, -0.00003704, 0.00014652, -0.00001199, 0.00004954, -0.00007951, 0.00001507, -0.00002243, -0.00001578, 0.00000094, -0.00001415, -0.00001115, 0.00000018, -0.00000786, -0.00005115, -0.00000744, -0.00013050, -0.00038594, 0.00129298, -0.00040092, -0.00011338, -0.00003187, -0.00000903, -0.00001467, -0.00003878, 0.00000564, -0.00000879, -0.00001556, 0.00003797, 0.00000088, -0.00004373, 0.00004679, -0.00005068, 0.00003587, -0.00003792, 0.00001082, -0.00005889, 0.00004679, -0.00004362, 0.00002852, 0.00002754, -0.00001974, -0.00000524, 0.00002378, 0.00000086, -0.00003719, },
		{ 0.00001636, 0.00000670, 0.00001711, -0.00001804, -0.00001529, -0.00001268, -0.00000062, 0.00001598, 0.00000946, -0.00005425, 0.00004497, 0.00000447, 0.00000461, -0.00002180, 0.00000093, -0.00003735, 0.00005923, -0.00010334, 0.00004323, -0.00005251, 0.00014244, 0.00001661, 0.00004809, -0.00006796, -0.00000273, -0.00001575, -0.00001046, -0.00001891, 0.00001610, -0.00004931, 0.00004203, -0.00005559, 0.00002622, -0.00011717, 0.00007431, -0.00059783, 0.00106497, 0.00010248, -0.00040073, 0.00010857, -0.00012347, 0.00007983, -0.00011201, 0.00005722, -0.00005502, 0.00002854, -0.00001951, 0.00006870, -0.00009469, 0.00007703, -0.00006975, 0.00005019, -0.00004943, 0.00002341, -0.00006261, 0.00003171, -0.00001917, -0.00000411, 0.00004855, -0.00002011, -0.00000596, 0.00001115, 0.00002095, -0.00004347, },
		{ 0.00000560, 0.00000772, 0.00001578, -0.00000811, -0.00001844, -0.00001640, 0.00000126, 0.00000536, 0.00002348, -0.00005543, 0.00002946, 0.00001601, 0.00000754, -0.00002586, 0.00001074, -0.00005329, 0.00006349, -0.00008202, 0.00000898, -0.00003795, 0.00010497, 0.00006820, 0.00003180, -0.00003834, -0.00003389, -0.00000067, -0.00001442, -0.00002453, 0.00002473, -0.00005757, 0.00004649, -0.00006105, 0.00004984, -0.00014862, 0.00013636, -0.00055896, 0.00061624, 0.00061421, -0.00056833, 0.00014677, -0.00016438, 0.00011087, -0.00012907, 0.00006335, -0.00006600, 0.00004388, -0.00004883, 0.00009894, -0.00010045, 0.00006530, -0.00005377, 0.00003376, -0.00003134, 0.00000542, -0.00003598, -0.00000646, 0.00001716, -0.00003629, 0.00006006, -0.00001355, -0.00000671, -0.00000088, 0.00003287, -0.00003855, },
		{ -0.00001194, 0.00001535, 0.00000555, 0.00000737, -0.00002396, -0.00001513, -0.00000479, 0.00000606, 0.00002081, -0.00003840, 0.00000103, 0.00003396, 0.00000345, -0.00001880, 0.00000536, -0.00004570, 0.00003746, -0.00003498, -0.00003664, -0.00001598, 0.00005591, 0.00011972, 0.00002013, -0.00000698, -0.00006313, 0.00001127, -0.00002075, -0.00001999, 0.00001517, -0.00004060, 0.00002073, -0.00003284, 0.00002855, -0.00011388, 0.00008895, -0.00037104, 0.00013035, 0.00095392, -0.00051709, 0.00004609, -0.00011489, 0.00006814, -0.00008519, 0.00002239, -0.00003949, 0.00002565, -0.00004242, 0.00008515, -0.00006116, 0.00001815, -0.00001151, -0.00000455, 0.00000632, -0.00002954, 0.00000224, -0.00004549, 0.00004370, -0.00005076, 0.00005227, 0.00000399, -0.00001138, -0.00000626, 0.00003155, -0.00002098, },
		{ -0.00003004, 0.00002381, -0.00000616, 0.00002022, -0.00002582, -0.00001316, -0.00001400, 0.00001143, 0.00001132, -0.00001656, -0.00002542, 0.00004507, 0.00000300, -0.00001057, -0.00000439, -0.00002891, 0.00000272, 0.00000962, -0.00006637, -0.00000746, 0.00001484, 0.00015100, 0.00002861, 0.00001324, -0.00007789, 0.00001028, -0.00002169, -0.00001403, -0.00000066, -0.00001492, -0.00001365, 0.00000344, -0.00000725, -0.00005418, 0.00000110, -0.00015808, -0.00023623, 0.00102461, -0.00024806, -0.00014929, -0.00001639, -0.00001381, -0.00001187, -0.00003804, 0.00000229, -0.00000882, -0.00001500, 0.00004436, 0.00000018, -0.00003839, 0.00003246, -0.00004213, 0.00004088, -0.00005736, 0.00002711, -0.00006283, 0.00004398, -0.00003904, 0.00002458, 0.00002948, -0.00001873, -0.00000448, 0.00001894, 0.00000359, },
		{ -0.00003934, 0.00002490, -0.00001050, 0.00002257, -0.00001886, -0.00001498, -0.00002008, 0.00001290, 0.00000592, -0.00000241, -0.00003801, 0.00004034, 0.00001290, -0.00000829, -0.00000934, -0.00001596, -0.00002255, 0.00003241, -0.00006719, -0.00001817, -0.00001155, 0.00015433, 0.00006009, 0.00002060, -0.00007538, -0.00000519, -0.00001504, -0.00001249, -0.00001212, 0.00000391, -0.00003670, 0.00002475, -0.00002979, -0.00000786, -0.00006750, -0.00000830, -0.00040984, 0.00084211, 0.00014171, -0.00033815, 0.00005904, -0.00008085, 0.00004583, -0.00008090, 0.00002976, -0.00003475, 0.00000999, 0.00000301, 0.00005325, -0.00007511, 0.00005361, -0.00005784, 0.00005254, -0.00005957, 0.00002344, -0.00004973, 0.00001724, -0.00000756, -0.00001193, 0.00005242, -0.00002233, -0.00000061, 0.00000298, 0.00002474, },
		{ -0.00003369, 0.00001438, -0.00000372, 0.00001281, -0.00000382, -0.00002061, -0.00002062, 0.00000660, 0.00000909, -0.00000007, -0.00003475, 0.00002105, 0.00003042, -0.00001076, -0.00000836, -0.00001078, -0.00003264, 0.00003106, -0.00004509, -0.00003791, -0.00002875, 0.00013414, 0.00010516, 0.00002385, -0.00006195, -0.00002858, -0.00000517, -0.00001529, -0.00001553, 0.00000945, -0.00004064, 0.00002310, -0.00002909, 0.00000918, -0.00009152, 0.00005043, -0.00040664, 0.00050986, 0.00050987, -0.00041706, 0.00005416, -0.00009516, 0.00005831, -0.00008383, 0.00002610, -0.00003747, 0.00001695, -0.00001997, 0.00007780, -0.00007617, 0.00004213, -0.00004517, 0.00003654, -0.00003439, -0.00000541, -0.00001605, -0.00002113, 0.00002561, -0.00003993, 0.00006044, -0.00001522, -0.00000086, -0.00000759, 0.00003342, },
		{ -0.00001480, -0.00000417, 0.00000994, -0.00000282, 0.00001223, -0.00002517, -0.00001825, -0.00000490, 0.00001731, -0.00000459, -0.00002305, -0.00000289, 0.00004568, -0.00001048, -0.00000694, -0.00000913, -0.00003275, 0.00001625, -0.00001719, -0.00005076, -0.00004515, 0.00009944, 0.00014931, 0.00003500, -0.00004663, -0.00005075, 0.00000117, -0.00001868, -0.00001344, 0.00000463, -0.00003013, 0.00000559, -0.00001280, 0.00000363, -0.00007855, 0.00003964, -0.00029913, 0.00016005, 0.00073555, -0.00033397, -0.00003828, -0.00005813, 0.00002791, -0.00005145, -0.00000316, -0.00002071, 0.00000645, -0.00002198, 0.00007209, -0.00004572, 0.00000722, -0.00001480, 0.00000452, 0.00000374, -0.00004110, 0.00001677, -0.00004942, 0.00004162, -0.00004600, 0.00004736, 0.00000403, -0.00000723, -0.00000890, 0.00002742, },
		{ 0.00000823, -0.00002110, 0.00002107, -0.00001431, 0.00002045, -0.00002299, -0.00001737, -0.00001589, 0.00002331, -0.00000741, -0.00001308, -0.00002025, 0.00004904, -0.00000078, -0.00001037, -0.00000557, -0.00003069, 0.00000066, 0.00000150, -0.00004764, -0.00006264, 0.00005660, 0.00018036, 0.00006180, -0.00003424, -0.00006604, -0.00000015, -0.00001953, -0.00001036, -0.00000351, -0.00001551, -0.00001431, 0.00000421, -0.00000799, -0.00005119, 0.00000042, -0.00016848, -0.00010465, 0.00076720, -0.00011175, -0.00017133, -0.00000648, -0.00001703, -0.00000885, -0.00003660, -0.00000113, -0.00000902, -0.00001334, 0.00004893, -0.00000208, -0.00003052, 0.00001388, -0.00002459, 0.00003464, -0.00006272, 0.00002944, -0.00005299, 0.00003228, -0.00002887, 0.00001811, 0.00002860, -0.00001540, -0.00000389, 0.00001240, },
		{ 0.00002430, -0.00002694, 0.00002187, -0.00001477, 0.00001659, -0.00001267, -0.00001954, -0.00002256, 0.00002205, -0.00000323, -0.00001057, -0.00002568, 0.00003786, 0.00001814, -0.00001812, 0.00000047, -0.00002955, -0.00000972, 0.00000665, -0.00003189, -0.00007462, 0.00000812, 0.00019117, 0.00010440, -0.00002272, -0.00007438, -0.00000861, -0.00001768, -0.00000914, -0.00000925, -0.00000518, -0.00002632, 0.00001125, -0.00001317, -0.00002837, -0.00003387, -0.00006903, -0.00024199, 0.00063126, 0.00016807, -0.00027143, 0.00001407, -0.00004442, 0.00001746, -0.00005310, 0.00000538, -0.00001672, -0.00000641, 0.00002418, 0.00003504, -0.00005229, 0.00002546, -0.00003698, 0.00004501, -0.00005936, 0.00001612, -0.00003261, 0.00000455, -0.00000018, -0.00001334, 0.00004664, -0.00001775, 0.00000071, -0.00000215, },
		{ 0.00002699, -0.00001865, 0.00001168, -0.00000511, 0.00000416, 0.00000092, -0.00002190, -0.00002529, 0.00001371, 0.00000691, -0.00001403, -0.00002184, 0.00001773, 0.00003884, -0.00002369, 0.00000429, -0.00002674, -0.00001661, 0.00000366, -0.00001485, -0.00007250, -0.00004287, 0.00017798, 0.00015671, -0.00000532, -0.00007906, -0.00002086, -0.00001516, -0.00001022, -0.00001062, -0.00000196, -0.00002784, 0.00000683, -0.00000949, -0.00001663, -0.00004966, -0.00001770, -0.00026652, 0.00040478, 0.00040817, -0.00027930, -0.00002045, -0.00004074, 0.00001689, -0.00004587, -0.00000502, -0.00001332, -0.00000631, 0.00000682, 0.00005494, -0.00005089, 0.00001690, -0.00003115, 0.00003483, -0.00003496, -0.00001368, -0.00000271, -0.00002459, 0.00002291, -0.00003207, 0.00004855, -0.00000930, -0.00000006, -0.00000917, },
		{ 0.00001796, -0.00000162, -0.00000250, 0.00000690, -0.00000798, 0.00000990, -0.00001985, -0.00002679, 0.00000210, 0.00001773, -0.00001767, -0.00001568, -0.00000213, 0.00005176, -0.00001982, 0.00000081, -0.00001871, -0.00002396, 0.00000013, -0.00000618, -0.00005431, -0.00008792, 0.00013853, 0.00020907, 0.00002532, -0.00008271, -0.00003399, -0.00001387, -0.00001286, -0.00000869, -0.00000384, -0.00002251, -0.00000359, -0.00000214, -0.00001252, -0.00004963, -0.00000431, -0.00022326, 0.00017286, 0.00053943, -0.00018025, -0.00009688, -0.00001850, -0.00000105, -0.00002459, -0.00002306, -0.00000575, -0.00000944, -0.00000336, 0.00005824, -0.00003159, -0.00000323, -0.00001616, 0.00001451, -0.00000353, -0.00004257, 0.00001852, -0.00003912, 0.00002828, -0.00003110, 0.00003333, 0.00000787, -0.00000548, -0.00000802, },
		{ 0.00000483, 0.00001404, -0.00001097, 0.00001256, -0.00001199, 0.00000923, -0.00001177, -0.00002833, -0.00000927, 0.00002431, -0.00001664, -0.00001265, -0.00001504, 0.00005154, -0.00000439, -0.00001006, -0.00000614, -0.00003205, -0.00000130, -0.00000696, -0.00002818, -0.00011440, 0.00007369, 0.00024955, 0.00007415, -0.00008460, -0.00004745, -0.00001423, -0.00001643, -0.00000566, -0.00000705, -0.00001588, -0.00001315, 0.00000256, -0.00001004, -0.00004240, -0.00000967, -0.00015833, -0.00000545, 0.00054383, -0.00000794, -0.00017340, -0.00000502, -0.00001719, -0.00000597, -0.00003494, -0.00000393, -0.00000933, -0.00001085, 0.00005161, -0.00000561, -0.00002232, -0.00000361, -0.00000369, 0.00002068, -0.00005634, 0.00002021, -0.00003350, 0.00001566, -0.00001505, 0.00000952, 0.00002522, -0.00000994, -0.00000351, },
		{ -0.00000448, 0.00002018, -0.00000818, 0.00000863, -0.00000668, 0.00000081, -0.00000142, -0.00002778, -0.00001962, 0.00002522, -0.00001038, -0.00001348, -0.00002004, 0.00003970, 0.00001775, -0.00002235, 0.00000518, -0.00003650, -0.00000377, -0.00001080, -0.00000728, -0.00011280, -0.00000795, 0.00026405, 0.00014210, -0.00008022, -0.00006266, -0.00001561, -0.00002054, -0.00000362, -0.00000861, -0.00001167, -0.00001804, 0.00000188, -0.00000625, -0.00003425, -0.00001891, -0.00010114, -0.00010961, 0.00044968, 0.00017663, -0.00020468, -0.00002162, -0.00001864, -0.00000074, -0.00003308, -0.00001205, -0.00000379, -0.00001789, 0.00004074, 0.00001783, -0.00003187, 0.00000007, -0.00001330, 0.00003097, -0.00005118, 0.00000403, -0.00001472, -0.00000447, 0.00000361, -0.00001017, 0.00003322, -0.00000770, -0.00000124, },
		{ -0.00000683, 0.00001545, 0.00000351, -0.00000031, 0.00000190, -0.00000772, 0.00000418, -0.00002160, -0.00002976, 0.00002179, -0.00000169, -0.00001536, -0.00002024, 0.00002222, 0.00003818, -0.00002800, 0.00000864, -0.00003253, -0.00001135, -0.00001084, -0.00000042, -0.00008477, -0.00008782, 0.00023862, 0.00022442, -0.00006207, -0.00008133, -0.00001759, -0.00002461, -0.00000399, -0.00000741, -0.00001057, -0.00001842, -0.00000264, -0.00000211, -0.00002693, -0.00002554, -0.00006200, -0.00014957, 0.00030798, 0.00031614, -0.00016717, -0.00006802, -0.00000829, -0.00000700, -0.00002083, -0.00002537, 0.00000275, -0.00002208, 0.00002719, 0.00003519, -0.00003068, -0.00000404, -0.00001538, 0.00002941, -0.00003323, -0.00001903, 0.00000366, -0.00001882, 0.00001313, -0.00001690, 0.00002783, 0.00000207, -0.00000287, },
		{ -0.00000484, 0.00000526, 0.00001571, -0.00000543, 0.00000535, -0.00000873, 0.00000039, -0.00000919, -0.00003892, 0.00001513, 0.00000667, -0.00001578, -0.00001870, 0.00000482, 0.00005028, -0.00002256, 0.00000220, -0.00002021, -0.00002332, -0.00000643, -0.00000590, -0.00004483, -0.00014218, 0.00016603, 0.00030792, -0.00002055, -0.00010399, -0.00002080, -0.00002777, -0.00000730, -0.00000421, -0.00001098, -0.00001687, -0.00000762, 0.00000001, -0.00001992, -0.00002872, -0.00003914, -0.00014858, 0.00016644, 0.00037994, -0.00006975, -0.00012206, -0.00000138, -0.00001427, -0.00000828, -0.00003455, 0.00000322, -0.00001949, 0.00001049, 0.00004672, -0.00002224, -0.00001075, -0.00001478, 0.00002226, -0.00001241, -0.00003679, 0.00001076, -0.00002002, 0.00000979, -0.00001090, 0.00001338, 0.00001398, -0.00000508, },
		},
		{ { 0.00001500, 0.00001542, -0.00008836, 0.00000046, 0.00003180, -0.00001841, 0.00001420, -0.00000280, 0.00002091, -0.00001041, -0.00001177, 0.00000124, 0.00001323, 0.00000152, -0.00002336, -0.00000729, -0.00000410, -0.00001236, 0.00002609, -0.00003499, -0.00015675, 0.00004884, 0.00038831, 0.00005911, -0.00013694, -0.00002357, -0.00005816, 0.00000071, -0.00000845, 0.00000444, 0.00000101, -0.00001427, -0.00001506, -0.00001427, 0.00000101, 0.00000444, -0.00000845, 0.00000071, -0.00005816, -0.00002357, -0.00013694, 0.00005911, 0.00038831, 0.00004884, -0.00015675, -0.00003499, 0.00002609, -0.00001236, -0.00000410, -0.00000729, -0.00002336, 0.00000152, 0.00001323, 0.00000124, -0.00001177, -0.00001041, 0.00002091, -0.00000280, 0.00001420, -0.00001841, 0.00003180, 0.00000046, -0.00008836, 0.00001542, },
		{ 0.00001574, 0.00002778, -0.00006373, -0.00004809, 0.00004966, -0.00001873, 0.00001245, -0.00000293, 0.00001797, 0.00000039, -0.00001589, -0.00000309, 0.00001077, 0.00000979, -0.00001958, -0.00001467, -0.00000067, -0.00001687, 0.00001696, 0.00000030, -0.00013125, -0.00008873, 0.00036606, 0.00021869, -0.00016105, -0.00003435, -0.00005680, -0.00001736, 0.00000005, -0.00000424, 0.00000674, -0.00001229, -0.00001412, -0.00001624, -0.00000490, 0.00000928, -0.00000974, 0.00000467, -0.00004299, -0.00003277, -0.00009230, -0.00005335, 0.00032664, 0.00018145, -0.00012925, -0.00008349, 0.00002502, -0.00000157, -0.00001016, 0.00000022, -0.00002389, -0.00000644, 0.00001161, 0.00000594, -0.00000681, -0.00001641, 0.00001603, 0.00000398, 0.00000809, -0.00000634, 0.00000398, 0.00003884, -0.00008678, -0.00001071, },
		{ 0.00001695, 0.00002528, -0.00002305, -0.00008815, 0.00004546, -0.00000498, 0.00000400, 0.00000209, 0.00001014, 0.00001153, -0.00001631, -0.00000775, 0.00000639, 0.00001492, -0.00001120, -0.00002207, 0.00000128, -0.00001863, 0.00000747, 0.00001375, -0.00007536, -0.00018069, 0.00024565, 0.00038739, -0.00013010, -0.00007601, -0.00003814, -0.00004452, 0.00000879, -0.00001112, 0.00000905, -0.00000921, -0.00001232, -0.00001746, -0.00000992, 0.00000888, -0.00000430, -0.00000045, -0.00002239, -0.00004410, -0.00005537, -0.00010605, 0.00021518, 0.00027123, -0.00005553, -0.00011969, 0.00000161, 0.00001727, -0.00001849, 0.00000724, -0.00002226, -0.00001187, 0.00000584, 0.00001041, -0.00000263, -0.00001608, 0.00000473, 0.00001407, -0.00000099, 0.00000812, -0.00001795, 0.00005297, -0.00005816, -0.00004389, },
		{ 0.00001466, 0.00001387, 0.00001667, -0.00010244, 0.00001508, 0.00001813, -0.00000679, 0.00000946, 0.00000068, 0.00001990, -0.00001279, -0.00001131, 0.00000031, 0.00001679, -0.00000115, -0.00002556, -0.00000129, -0.00001551, -0.00000342, 0.00001638, -0.00002678, -0.00019208, 0.00005299, 0.00050294, -0.00001502, -0.00014395, -0.00001377, -0.00006832, 0.00000835, -0.00001110, 0.00000596, -0.00000439, -0.00001111, -0.00001650, -0.00001474, 0.00000553, 0.00000235, -0.00000517, -0.00000863, -0.00004422, -0.00003938, -0.00010804, 0.00009249, 0.00030151, 0.00003953, -0.00012236, -0.00004147, 0.00003389, -0.00002086, 0.00000779, -0.00001527, -0.00001726, 0.00000033, 0.00001079, 0.00000261, -0.00001273, -0.00000649, 0.00001975, -0.00000441, 0.00001387, -0.00002225, 0.00003953, -0.00001237, -0.00007270, },
		{ 0.00000226, 0.00000474, 0.00003784, -0.00008217, -0.00003489, 0.00004058, -0.00001470, 0.00001656, -0.00000773, 0.00002409, -0.00000706, -0.00001154, -0.00000786, 0.00001709, 0.00000597, -0.00002085, -0.00001134, -0.00000512, -0.00001882, 0.00002056, -0.00000761, -0.00013228, -0.00014125, 0.00049716, 0.00018889, -0.00021169, -0.00000459, -0.00007571, -0.00000646, -0.00000445, -0.00000039, 0.00000037, -0.00001019, -0.00001370, -0.00001895, 0.00000102, 0.00000616, -0.00000446, -0.00000528, -0.00003283, -0.00003951, -0.00008383, -0.00000673, 0.00027158, 0.00013191, -0.00009075, -0.00008265, 0.00002971, -0.00000681, -0.00000409, -0.00000005, -0.00002544, -0.00000062, 0.00000427, 0.00001037, -0.00001027, -0.00001153, 0.00001555, 0.00000238, 0.00000708, -0.00000935, 0.00000967, 0.00003130, -0.00008303, },
		{ -0.00002404, 0.00000622, 0.00003268, -0.00003424, -0.00008604, 0.00004848, -0.00001385, 0.00001934, -0.00001070, 0.00002154, 0.00000107, -0.00000971, -0.00001494, 0.00001426, 0.00001031, -0.00001047, -0.00002379, 0.00000603, -0.00003179, 0.00002274, -0.00000736, -0.00005347, -0.00025205, 0.00033629, 0.00043882, -0.00022837, -0.00003493, -0.00005851, -0.00003425, 0.00000402, -0.00000592, 0.00000274, -0.00000856, -0.00001069, -0.00002076, -0.00000492, 0.00000738, -0.00000064, -0.00000622, -0.00002025, -0.00004064, -0.00006057, -0.00006059, 0.00019347, 0.00020490, -0.00003921, -0.00009854, -0.00000350, 0.00002132, -0.00002297, 0.00001705, -0.00003190, 0.00000015, -0.00000491, 0.00001648, -0.00000741, -0.00001101, 0.00000449, 0.00001407, -0.00000437, 0.00000736, -0.00001588, 0.00005215, -0.00006484, },
		{ -0.00005713, 0.00001483, 0.00001056, 0.00001790, -0.00011301, 0.00002893, 0.00000048, 0.00001278, -0.00000274, 0.00000891, 0.00001382, -0.00001037, -0.00001481, 0.00000382, 0.00001672, -0.00000322, -0.00002714, 0.00000505, -0.00002861, 0.00000882, 0.00000077, -0.00001140, -0.00023545, 0.00006165, 0.00063725, -0.00013163, -0.00011991, -0.00001871, -0.00006808, 0.00000913, -0.00000917, 0.00000342, -0.00000727, -0.00000783, -0.00001965, -0.00001208, 0.00000744, 0.00000191, -0.00000411, -0.00001546, -0.00003256, -0.00005201, -0.00006972, 0.00009376, 0.00024315, 0.00002076, -0.00008503, -0.00004961, 0.00004224, -0.00003033, 0.00002076, -0.00002565, -0.00000625, -0.00000714, 0.00001295, 0.00000107, -0.00001112, -0.00000413, 0.00001943, -0.00000749, 0.00001216, -0.00002055, 0.00004131, -0.00002192, },
		{ -0.00007816, 0.00001548, -0.00000699, 0.00004610, -0.00009736, -0.00001992, 0.00002550, -0.00000327, 0.00001628, -0.00001166, 0.00002924, -0.00001449, -0.00000602, -0.00001383, 0.00002625, -0.00000356, -0.00001628, -0.00001227, -0.00000640, -0.00002312, 0.00002597, -0.00001934, -0.00012584, -0.00021013, 0.00066672, 0.00011335, -0.00024240, 0.00002325, -0.00009435, 0.00000604, -0.00001113, 0.00000508, -0.00000855, -0.00000404, -0.00001732, -0.00001799, 0.00000538, 0.00000276, 0.00000134, -0.00001677, -0.00001925, -0.00005201, -0.00005654, 0.00000873, 0.00023146, 0.00008753, -0.00005842, -0.00007798, 0.00003143, -0.00001080, 0.00000167, -0.00000191, -0.00002344, 0.00000180, -0.00000239, 0.00001580, -0.00001491, -0.00000460, 0.00001264, 0.00000285, 0.00000115, -0.00000516, 0.00001029, 0.00002504, },
		{ -0.00006865, -0.00000518, -0.00000357, 0.00003723, -0.00004340, -0.00008008, 0.00004571, -0.00001863, 0.00003547, -0.00002766, 0.00003663, -0.00001417, 0.00000271, -0.00002799, 0.00002970, -0.00000380, -0.00000150, -0.00003185, 0.00001657, -0.00005095, 0.00004529, -0.00003971, -0.00001538, -0.00034790, 0.00046431, 0.00046277, -0.00033303, 0.00002451, -0.00009211, -0.00001241, -0.00001019, 0.00000656, -0.00001071, -0.00000082, -0.00001444, -0.00002056, -0.00000039, 0.00000448, 0.00000474, -0.00001517, -0.00001268, -0.00004543, -0.00004813, -0.00003393, 0.00016748, 0.00015698, -0.00003409, -0.00007154, -0.00001270, 0.00002732, -0.00002841, 0.00002618, -0.00004015, 0.00001243, -0.00001899, 0.00002691, -0.00001610, -0.00000172, 0.00000023, 0.00001762, -0.00001339, 0.00001232, -0.00001556, 0.00004943, },
		{ -0.00002782, -0.00004281, 0.00001621, 0.00000516, 0.00001880, -0.00011826, 0.00003887, -0.00001749, 0.00003789, -0.00002248, 0.00002312, 0.00000111, -0.00000103, -0.00002478, 0.00001496, 0.00000857, 0.00000027, -0.00003236, 0.00001511, -0.00004630, 0.00002769, -0.00002516, 0.00001727, -0.00029557, 0.00008854, 0.00077560, -0.00028384, -0.00006180, -0.00004398, -0.00005239, -0.00000131, 0.00000098, -0.00000661, -0.00000383, -0.00000819, -0.00002146, -0.00000803, 0.00000707, 0.00000339, -0.00000603, -0.00001782, -0.00002681, -0.00005446, -0.00003439, 0.00007534, 0.00020790, -0.00000669, -0.00004445, -0.00006023, 0.00005161, -0.00004056, 0.00003378, -0.00003681, 0.00000813, -0.00002053, 0.00002081, -0.00000487, -0.00000540, -0.00000514, 0.00002185, -0.00001448, 0.00001271, -0.00001666, 0.00003861, },
		{ 0.00002127, -0.00007319, 0.00002856, -0.00001866, 0.00005155, -0.00010712, -0.00000574, 0.00000617, 0.00001615, 0.00000849, -0.00001110, 0.00002992, -0.00001814, -0.00000319, -0.00001674, 0.00003331, -0.00001427, -0.00000916, -0.00001507, -0.00000688, -0.00002656, 0.00002879, -0.00002916, -0.00011927, -0.00028610, 0.00086897, -0.00001397, -0.00023978, 0.00003762, -0.00010454, 0.00001406, -0.00001537, 0.00000824, -0.00001635, 0.00000342, -0.00002372, -0.00001296, 0.00000647, 0.00000124, 0.00000497, -0.00002588, -0.00000757, -0.00006266, -0.00002019, -0.00000135, 0.00020890, 0.00004292, -0.00002762, -0.00007337, 0.00003287, -0.00001522, 0.00000814, -0.00000624, -0.00001638, -0.00000206, -0.00000446, 0.00001858, -0.00001749, 0.00000116, 0.00001032, 0.00000216, -0.00000608, 0.00000393, 0.00000591, },
		{ 0.00004670, -0.00007028, 0.00001186, -0.00001159, 0.00003811, -0.00004908, -0.00007110, 0.00003631, -0.00001599, 0.00004744, -0.00004457, 0.00005206, -0.00003107, 0.00001811, -0.00004412, 0.00005004, -0.00002587, 0.00001568, -0.00004678, 0.00003306, -0.00007536, 0.00007372, -0.00008159, 0.00003734, -0.00046788, 0.00063627, 0.00044190, -0.00042523, 0.00009054, -0.00013252, 0.00001706, -0.00003133, 0.00002445, -0.00002990, 0.00001381, -0.00002484, -0.00001434, 0.00000058, 0.00000299, 0.00000898, -0.00002384, -0.00000354, -0.00005411, -0.00002150, -0.00002904, 0.00014708, 0.00011815, -0.00003369, -0.00004344, -0.00002284, 0.00003296, -0.00003271, 0.00003223, -0.00004371, 0.00002053, -0.00003172, 0.00003797, -0.00002573, 0.00000909, -0.00000573, 0.00002247, -0.00002607, 0.00002283, -0.00001867, },
		{ 0.00003358, -0.00003045, -0.00003055, 0.00001907, -0.00000233, 0.00001956, -0.00011720, 0.00004213, -0.00003126, 0.00006401, -0.00004726, 0.00004224, -0.00001769, 0.00001527, -0.00004167, 0.00003475, -0.00001014, 0.00001300, -0.00004494, 0.00003190, -0.00007095, 0.00005566, -0.00006463, 0.00006396, -0.00038023, 0.00014861, 0.00089695, -0.00045639, 0.00002590, -0.00009005, -0.00001968, -0.00002489, 0.00002117, -0.00002608, 0.00000877, -0.00001577, -0.00001818, -0.00000578, 0.00000648, 0.00000526, -0.00000958, -0.00001643, -0.00002724, -0.00004381, -0.00001135, 0.00004996, 0.00018721, -0.00003843, -0.00000314, -0.00007204, 0.00006087, -0.00005007, 0.00004473, -0.00004577, 0.00002232, -0.00003690, 0.00003286, -0.00001448, 0.00000394, -0.00000941, 0.00002701, -0.00002596, 0.00001776, -0.00001367, },
		{ -0.00000185, 0.00002014, -0.00006868, 0.00004101, -0.00002953, 0.00005346, -0.00010941, 0.00000443, -0.00001379, 0.00004153, -0.00000764, -0.00000385, 0.00002365, -0.00001498, -0.00000647, -0.00001256, 0.00003307, -0.00002046, -0.00000446, -0.00001566, -0.00001036, -0.00002467, 0.00002328, -0.00002943, -0.00012309, -0.00035416, 0.00108930, -0.00019000, -0.00019611, 0.00002926, -0.00009880, 0.00001149, -0.00001219, 0.00000459, -0.00001843, 0.00000785, -0.00002949, -0.00000577, 0.00000462, 0.00000199, 0.00000545, -0.00003046, -0.00000144, -0.00006504, 0.00001449, -0.00002705, 0.00020083, -0.00000277, 0.00000335, -0.00007067, 0.00003511, -0.00002037, 0.00001489, -0.00001206, -0.00000519, -0.00001249, -0.00000015, 0.00001657, -0.00001592, 0.00000353, 0.00001082, -0.00000238, -0.00001083, 0.00001446, },
		{ -0.00002507, 0.00004492, -0.00007025, 0.00002627, -0.00001700, 0.00003450, -0.00004939, -0.00006128, 0.00002149, -0.00000455, 0.00005115, -0.00005688, 0.00006369, -0.00004540, 0.00003227, -0.00005887, 0.00007028, -0.00005139, 0.00003791, -0.00006634, 0.00005322, -0.00010189, 0.00010381, -0.00012762, 0.00009814, -0.00060492, 0.00085107, 0.00036394, -0.00048637, 0.00014870, -0.00017019, 0.00004727, -0.00005423, 0.00004344, -0.00005096, 0.00003250, -0.00004058, -0.00000172, -0.00000371, 0.00000495, 0.00000965, -0.00002883, 0.00000275, -0.00006066, 0.00001194, -0.00004285, 0.00013661, 0.00008208, -0.00003274, -0.00001836, -0.00003122, 0.00003642, -0.00003456, 0.00003398, -0.00004072, 0.00002119, -0.00003884, 0.00004574, -0.00003288, 0.00001814, -0.00001036, 0.00002539, -0.00003883, 0.00003658, },
		{ -0.00001343, 0.00002810, -0.00003043, -0.00002096, 0.00002368, -0.00001214, 0.00002112, -0.00011015, 0.00003729, -0.00003684, 0.00008561, -0.00007262, 0.00006307, -0.00003980, 0.00003624, -0.00006343, 0.00006137, -0.00003882, 0.00003597, -0.00006647, 0.00005669, -0.00010315, 0.00009295, -0.00011799, 0.00012992, -0.00049334, 0.00025418, 0.00097867, -0.00062794, 0.00013161, -0.00015173, 0.00002726, -0.00006009, 0.00005042, -0.00005442, 0.00003100, -0.00003241, -0.00000690, -0.00000859, 0.00000843, 0.00000495, -0.00001191, -0.00001413, -0.00003097, -0.00002175, -0.00000486, 0.00002761, 0.00017230, -0.00006869, 0.00003501, -0.00008265, 0.00006811, -0.00005698, 0.00005137, -0.00004941, 0.00003202, -0.00005173, 0.00004569, -0.00002515, 0.00001457, -0.00001499, 0.00003304, -0.00004025, 0.00002739, },
		{ 0.00002304, -0.00001071, 0.00002104, -0.00006490, 0.00005242, -0.00003919, 0.00005176, -0.00010300, 0.00000756, -0.00002973, 0.00006849, -0.00003026, 0.00000998, 0.00000992, -0.00000384, -0.00001752, 0.00000103, 0.00002174, -0.00001713, -0.00000718, -0.00000963, -0.00002081, -0.00001308, 0.00000394, -0.00001258, -0.00014864, -0.00039591, 0.00130561, -0.00040184, -0.00011173, -0.00000611, -0.00007271, -0.00000581, 0.00000205, -0.00000932, -0.00001079, 0.00000475, -0.00003022, -0.00000001, 0.00000204, 0.00000357, 0.00000387, -0.00003075, -0.00000171, -0.00005587, 0.00003875, -0.00005724, 0.00020058, -0.00004687, 0.00003344, -0.00006964, 0.00003769, -0.00002528, 0.00002035, -0.00001693, 0.00000712, -0.00002716, 0.00000972, 0.00000979, -0.00001008, 0.00000205, 0.00001466, -0.00001177, -0.00001050, },
		{ 0.00004997, -0.00003308, 0.00004405, -0.00006873, 0.00003730, -0.00001915, 0.00002583, -0.00004242, -0.00005371, 0.00000452, 0.00001438, 0.00004466, -0.00006126, 0.00006799, -0.00005317, 0.00004089, -0.00006742, 0.00008452, -0.00007407, 0.00005969, -0.00008589, 0.00007259, -0.00012532, 0.00012904, -0.00016940, 0.00015600, -0.00074498, 0.00109819, 0.00022539, -0.00050239, 0.00018532, -0.00019486, 0.00007018, -0.00007242, 0.00005793, -0.00006846, 0.00005026, -0.00005786, 0.00001518, -0.00001281, 0.00001072, 0.00000582, -0.00002813, 0.00000291, -0.00005953, 0.00004325, -0.00006786, 0.00013514, 0.00004635, -0.00002849, 0.00000126, -0.00003643, 0.00003705, -0.00003389, 0.00003190, -0.00003158, 0.00001348, -0.00003788, 0.00004746, -0.00003500, 0.00002280, -0.00001108, 0.00002344, -0.00004770, },
		{ 0.00003947, -0.00001595, 0.00002315, -0.00002819, -0.00001468, 0.00003043, -0.00002476, 0.00002516, -0.00009901, 0.00002504, -0.00003207, 0.00009772, -0.00009343, 0.00008122, -0.00006094, 0.00005776, -0.00008588, 0.00009032, -0.00007362, 0.00006681, -0.00009555, 0.00008810, -0.00014080, 0.00013669, -0.00018170, 0.00021135, -0.00063247, 0.00041112, 0.00100122, -0.00077504, 0.00023896, -0.00021907, 0.00008162, -0.00010195, 0.00008475, -0.00008812, 0.00006039, -0.00005720, 0.00001316, -0.00001846, 0.00001510, 0.00000012, -0.00001001, -0.00001462, -0.00003322, 0.00000621, -0.00001373, 0.00001373, 0.00015653, -0.00009235, 0.00006628, -0.00008972, 0.00007166, -0.00005990, 0.00005240, -0.00004572, 0.00003373, -0.00006053, 0.00005524, -0.00003348, 0.00002327, -0.00001892, 0.00003681, -0.00005393, },
		{ -0.00000460, 0.00002752, -0.00001863, 0.00002320, -0.00006207, 0.00006262, -0.00004770, 0.00004739, -0.00008809, 0.00000187, -0.00003724, 0.00009132, -0.00005521, 0.00002789, -0.00000939, 0.00001403, -0.00003554, 0.00002358, -0.00000160, -0.00000165, -0.00002061, 0.00000609, -0.00004089, 0.00001089, -0.00003260, 0.00002653, -0.00020501, -0.00039333, 0.00149159, -0.00062822, 0.00000447, -0.00006640, -0.00002641, -0.00003850, 0.00002827, -0.00003452, 0.00000846, -0.00000892, -0.00002201, 0.00000035, 0.00000189, 0.00000330, 0.00000286, -0.00002915, -0.00000624, -0.00003587, 0.00004786, -0.00008254, 0.00020060, -0.00008491, 0.00005997, -0.00006875, 0.00003916, -0.00002839, 0.00002262, -0.00001798, 0.00001645, -0.00004183, 0.00002213, 0.00000036, -0.00000179, -0.00000178, 0.00002041, -0.00002479, },
		{ -0.00004971, 0.00005943, -0.00004034, 0.00004334, -0.00006565, 0.00004437, -0.00001778, 0.00001225, -0.00002691, -0.00005154, -0.00001000, 0.00003688, 0.00002889, -0.00005660, 0.00006325, -0.00005196, 0.00004098, -0.00006611, 0.00008764, -0.00008758, 0.00007476, -0.00009972, 0.00008571, -0.00013964, 0.00014198, -0.00019712, 0.00019726, -0.00086864, 0.00135784, 0.00003347, -0.00046707, 0.00019089, -0.00019834, 0.00007877, -0.00007996, 0.00006262, -0.00007691, 0.00006141, -0.00007133, 0.00003215, -0.00002416, 0.00001868, -0.00000140, -0.00002195, -0.00000396, -0.00004757, 0.00006463, -0.00009483, 0.00013799, 0.00001204, -0.00002081, 0.00001498, -0.00003857, 0.00003544, -0.00003184, 0.00002803, -0.00001889, -0.00000078, -0.00002902, 0.00004250, -0.00003134, 0.00002212, -0.00000688, 0.00001516, },
		{ -0.00006297, 0.00005050, -0.00001969, 0.00001867, -0.00002380, -0.00001230, 0.00003964, -0.00004060, 0.00003370, -0.00008696, 0.00000837, -0.00001758, 0.00009746, -0.00010539, 0.00009243, -0.00007653, 0.00007506, -0.00010388, 0.00011552, -0.00010825, 0.00010021, -0.00012779, 0.00012197, -0.00017912, 0.00018093, -0.00024826, 0.00029905, -0.00078760, 0.00061598, 0.00095266, -0.00087680, 0.00033029, -0.00027971, 0.00013401, -0.00014314, 0.00011800, -0.00012141, 0.00009182, -0.00008632, 0.00003999, -0.00003513, 0.00002700, -0.00001020, -0.00000196, -0.00002088, -0.00002913, 0.00003264, -0.00003247, 0.00000889, 0.00013672, -0.00010639, 0.00008830, -0.00009193, 0.00007072, -0.00005853, 0.00004804, -0.00003472, 0.00002603, -0.00006032, 0.00005827, -0.00003656, 0.00002718, -0.00001842, 0.00003521, },
		{ -0.00003831, 0.00000508, 0.00002764, -0.00002455, 0.00002615, -0.00006055, 0.00007180, -0.00005572, 0.00004235, -0.00006665, -0.00001225, -0.00003366, 0.00010457, -0.00007718, 0.00004571, -0.00003046, 0.00003522, -0.00005728, 0.00005175, -0.00003448, 0.00002518, -0.00004505, 0.00003199, -0.00007067, 0.00004691, -0.00008597, 0.00008844, -0.00029594, -0.00033303, 0.00162145, -0.00084300, 0.00013656, -0.00014331, 0.00003495, -0.00008325, 0.00006399, -0.00006890, 0.00003809, -0.00003342, -0.00000302, -0.00000756, 0.00000698, -0.00000152, 0.00000554, -0.00002915, -0.00001086, -0.00000944, 0.00004195, -0.00009733, 0.00019480, -0.00011258, 0.00007994, -0.00006611, 0.00003797, -0.00002833, 0.00002036, -0.00001306, 0.00001908, -0.00005184, 0.00003299, -0.00000840, 0.00000588, -0.00000517, 0.00002514, },
		{ 0.00000131, -0.00004401, 0.00006257, -0.00004476, 0.00004195, -0.00006106, 0.00004740, -0.00001340, -0.00000501, -0.00000295, -0.00005692, -0.00001741, 0.00005730, 0.00000747, -0.00004445, 0.00005018, -0.00004167, 0.00003187, -0.00005377, 0.00007695, -0.00008747, 0.00007793, -0.00010273, 0.00008772, -0.00013965, 0.00013632, -0.00020199, 0.00020859, -0.00095444, 0.00160351, -0.00019425, -0.00038378, 0.00016241, -0.00017679, 0.00006910, -0.00007321, 0.00005417, -0.00007247, 0.00006123, -0.00007574, 0.00004416, -0.00003387, 0.00002573, -0.00000916, -0.00001269, -0.00001597, -0.00002512, 0.00007148, -0.00011545, 0.00013908, -0.00001765, -0.00001157, 0.00002394, -0.00003891, 0.00003301, -0.00003016, 0.00002518, -0.00000660, -0.00001763, -0.00001487, 0.00003247, -0.00002326, 0.00001713, 0.00000123, },
		{ 0.00002637, -0.00006416, 0.00005676, -0.00002226, 0.00001385, -0.00001712, -0.00001421, 0.00005125, -0.00005960, 0.00004849, -0.00007762, -0.00000805, 0.00000309, 0.00008491, -0.00010639, 0.00009395, -0.00008313, 0.00008415, -0.00011284, 0.00013098, -0.00013561, 0.00012923, -0.00015690, 0.00015227, -0.00021151, 0.00021767, -0.00030736, 0.00037956, -0.00094173, 0.00085494, 0.00083186, -0.00091925, 0.00039059, -0.00032181, 0.00017484, -0.00017578, 0.00014339, -0.00014766, 0.00011869, -0.00011381, 0.00006903, -0.00005593, 0.00004254, -0.00002498, 0.00001210, -0.00003390, -0.00001563, 0.00005055, -0.00005335, 0.00000983, 0.00011311, -0.00011038, 0.00010060, -0.00008935, 0.00006573, -0.00005376, 0.00004008, -0.00001863, 0.00001019, -0.00005061, 0.00005349, -0.00003307, 0.00002483, -0.00001203, },
		{ 0.00002554, -0.00004824, 0.00001505, 0.00002492, -0.00002862, 0.00002992, -0.00006087, 0.00008035, -0.00006422, 0.00003940, -0.00004222, -0.00003198, -0.00001912, 0.00010469, -0.00009134, 0.00005889, -0.00004865, 0.00005508, -0.00007793, 0.00008001, -0.00007142, 0.00005912, -0.00007738, 0.00006539, -0.00010707, 0.00009104, -0.00015135, 0.00016796, -0.00041768, -0.00021004, 0.00167507, -0.00102005, 0.00026490, -0.00022418, 0.00010232, -0.00013329, 0.00010374, -0.00010742, 0.00007375, -0.00006577, 0.00002559, -0.00002433, 0.00001858, -0.00001242, 0.00001420, -0.00003397, -0.00001087, 0.00001688, 0.00002537, -0.00010049, 0.00018011, -0.00012725, 0.00009134, -0.00006045, 0.00003321, -0.00002455, 0.00001349, -0.00000172, 0.00001303, -0.00005362, 0.00003856, -0.00001322, 0.00000979, -0.00000508, },
		{ 0.00001045, -0.00001521, -0.00003216, 0.00005904, -0.00004532, 0.00003948, -0.00005536, 0.00004689, -0.00000732, -0.00002345, 0.00002753, -0.00006995, -0.00001453, 0.00006987, -0.00001439, -0.00002849, 0.00003181, -0.00002468, 0.00001550, -0.00003218, 0.00005316, -0.00007253, 0.00006676, -0.00009212, 0.00007596, -0.00012277, 0.00010881, -0.00017878, 0.00018040, -0.00098317, 0.00180626, -0.00043251, -0.00026488, 0.00010423, -0.00013181, 0.00004145, -0.00005203, 0.00003241, -0.00005431, 0.00004760, -0.00006759, 0.00004695, -0.00003805, 0.00002841, -0.00001383, -0.00000415, -0.00002906, 0.00000405, 0.00006349, -0.00012452, 0.00013317, -0.00003911, -0.00000340, 0.00003002, -0.00003915, 0.00003128, -0.00003039, 0.00002580, 0.00000122, -0.00003219, 0.00000039, 0.00002064, -0.00001369, 0.00001053, },
		{ -0.00000029, 0.00001055, -0.00005624, 0.00005571, -0.00002150, 0.00000778, -0.00000813, -0.00002029, 0.00006449, -0.00008072, 0.00007014, -0.00007390, -0.00001913, 0.00002429, 0.00006321, -0.00009708, 0.00008544, -0.00007944, 0.00008305, -0.00011014, 0.00013253, -0.00014975, 0.00014727, -0.00017631, 0.00017263, -0.00023120, 0.00023867, -0.00034807, 0.00043769, -0.00107339, 0.00110513, 0.00064949, -0.00089818, 0.00041077, -0.00033697, 0.00019663, -0.00019350, 0.00015537, -0.00016117, 0.00013471, -0.00013318, 0.00009436, -0.00007636, 0.00005831, -0.00004123, 0.00002978, -0.00005206, 0.00000734, 0.00005548, -0.00006900, 0.00001139, 0.00008845, -0.00010613, 0.00010444, -0.00008329, 0.00005806, -0.00004729, 0.00003129, -0.00000128, -0.00001014, -0.00003355, 0.00004199, -0.00002384, 0.00001682, },
		{ 0.00000052, 0.00001918, -0.00005099, 0.00002139, 0.00002194, -0.00003191, 0.00003490, -0.00006355, 0.00008855, -0.00007401, 0.00004136, -0.00001916, -0.00005249, 0.00000312, 0.00009115, -0.00009475, 0.00006396, -0.00005989, 0.00006912, -0.00009252, 0.00010220, -0.00010535, 0.00009350, -0.00011174, 0.00010086, -0.00014431, 0.00013628, -0.00021990, 0.00025439, -0.00055841, -0.00003010, 0.00164226, -0.00113825, 0.00037023, -0.00029490, 0.00016479, -0.00018006, 0.00014034, -0.00014323, 0.00010882, -0.00010021, 0.00005969, -0.00004782, 0.00003567, -0.00002892, 0.00002925, -0.00004514, -0.00000285, 0.00003644, 0.00000496, -0.00009476, 0.00015680, -0.00012869, 0.00009374, -0.00005166, 0.00002494, -0.00001758, 0.00000342, 0.00001439, -0.00000118, -0.00004588, 0.00003677, -0.00001212, 0.00000791, },
		{ 0.00000568, 0.00001721, -0.00003019, -0.00001772, 0.00005052, -0.00004244, 0.00003625, -0.00004938, 0.00004388, -0.00000150, -0.00003955, 0.00006089, -0.00008821, -0.00000092, 0.00007041, -0.00003136, -0.00001336, 0.00001253, -0.00000511, -0.00000411, -0.00000577, 0.00002031, -0.00004530, 0.00004244, -0.00006837, 0.00005103, -0.00009005, 0.00006063, -0.00012775, 0.00010978, -0.00094228, 0.00194005, -0.00065292, -0.00012917, 0.00002690, -0.00007019, 0.00000038, -0.00001991, 0.00000059, -0.00002509, 0.00002182, -0.00004643, 0.00003848, -0.00003419, 0.00002417, -0.00001232, -0.00000014, -0.00003833, 0.00003384, 0.00004441, -0.00012086, 0.00011751, -0.00004985, 0.00000146, 0.00003493, -0.00004071, 0.00003128, -0.00003311, 0.00003101, 0.00000188, -0.00004022, 0.00001248, 0.00001066, -0.00000610, },
		{ 0.00000565, 0.00001426, -0.00000966, -0.00004052, 0.00004685, -0.00001638, 0.00000006, 0.00000269, -0.00002968, 0.00007769, -0.00010165, 0.00009756, -0.00007688, -0.00002096, 0.00003977, 0.00003764, -0.00008065, 0.00006905, -0.00006674, 0.00007236, -0.00009602, 0.00011909, -0.00014756, 0.00014989, -0.00018111, 0.00017812, -0.00023307, 0.00023768, -0.00036150, 0.00045991, -0.00116059, 0.00133816, 0.00042649, -0.00081992, 0.00038952, -0.00032221, 0.00019591, -0.00019306, 0.00015115, -0.00015881, 0.00013574, -0.00013919, 0.00011032, -0.00009150, 0.00007012, -0.00005485, 0.00004708, -0.00007151, 0.00003663, 0.00004670, -0.00007442, 0.00000864, 0.00006645, -0.00009668, 0.00010213, -0.00007575, 0.00004945, -0.00004088, 0.00002444, 0.00001307, -0.00003003, -0.00001324, 0.00002678, -0.00001156, },
		{ 0.00000015, 0.00001179, 0.00000562, -0.00004469, 0.00002107, 0.00002127, -0.00003571, 0.00004145, -0.00006870, 0.00009617, -0.00008519, 0.00005025, -0.00000145, -0.00006820, 0.00002731, 0.00006669, -0.00008726, 0.00005961, -0.00006190, 0.00007434, -0.00009743, 0.00011316, -0.00012949, 0.00012102, -0.00014101, 0.00013158, -0.00017521, 0.00017396, -0.00028050, 0.00033328, -0.00069975, 0.00019036, 0.00152518, -0.00118544, 0.00043774, -0.00034324, 0.00021225, -0.00021529, 0.00016680, -0.00016947, 0.00013613, -0.00012966, 0.00009307, -0.00007357, 0.00005509, -0.00004841, 0.00004879, -0.00006177, 0.00001405, 0.00004455, -0.00001223, -0.00008512, 0.00012784, -0.00011897, 0.00008838, -0.00004080, 0.00001424, -0.00000882, -0.00000742, 0.00003192, -0.00002056, -0.00003003, 0.00002786, -0.00000514, },
		{ -0.00000322, 0.00000535, 0.00001837, -0.00003944, -0.00000510, 0.00004006, -0.00003769, 0.00003316, -0.00004420, 0.00003959, 0.00000195, -0.00004943, 0.00009240, -0.00010708, 0.00002060, 0.00005766, -0.00003970, -0.00000314, -0.00000322, 0.00001241, -0.00002209, 0.00001965, -0.00001522, -0.00001143, 0.00000955, -0.00003536, 0.00001690, -0.00004640, -0.00000227, -0.00005529, 0.00000212, -0.00082925, 0.00198680, -0.00082925, 0.00000212, -0.00005529, -0.00000227, -0.00004640, 0.00001690, -0.00003536, 0.00000955, -0.00001143, -0.00001522, 0.00001965, -0.00002209, 0.00001241, -0.00000322, -0.00000314, -0.00003970, 0.00005766, 0.00002060, -0.00010708, 0.00009240, -0.00004943, 0.00000195, 0.00003959, -0.00004420, 0.00003316, -0.00003769, 0.00004006, -0.00000510, -0.00003944, 0.00001837, 0.00000535, },
		{ 0.00000015, -0.00000514, 0.00002786, -0.00003003, -0.00002056, 0.00003192, -0.00000742, -0.00000882, 0.00001424, -0.00004080, 0.00008838, -0.00011897, 0.00012784, -0.00008512, -0.00001223, 0.00004455, 0.00001405, -0.00006177, 0.00004879, -0.00004841, 0.00005509, -0.00007357, 0.00009307, -0.00012966, 0.00013613, -0.00016947, 0.00016680, -0.00021529, 0.00021225, -0.00034324, 0.00043774, -0.00118544, 0.00152518, 0.00019036, -0.00069975, 0.00033328, -0.00028050, 0.00017396, -0.00017521, 0.00013158, -0.00014101, 0.00012102, -0.00012949, 0.00011316, -0.00009743, 0.00007434, -0.00006190, 0.00005961, -0.00008726, 0.00006669, 0.00002731, -0.00006820, -0.00000145, 0.00005025, -0.00008519, 0.00009617, -0.00006870, 0.00004145, -0.00003571, 0.00002127, 0.00002107, -0.00004469, 0.00000562, 0.00001179, },
		{ 0.00000565, -0.00001156, 0.00002678, -0.00001324, -0.00003003, 0.00001307, 0.00002444, -0.00004088, 0.00004945, -0.00007575, 0.00010213, -0.00009668, 0.00006645, 0.00000864, -0.00007442, 0.00004670, 0.00003663, -0.00007151, 0.00004708, -0.00005485, 0.00007012, -0.00009150, 0.00011032, -0.00013919, 0.00013574, -0.00015881, 0.00015115, -0.00019306, 0.00019591, -0.00032221, 0.00038952, -0.00081992, 0.00042649, 0.00133816, -0.00116059, 0.00045991, -0.00036150, 0.00023768, -0.00023307, 0.00017812, -0.00018111, 0.00014989, -0.00014756, 0.00011909, -0.00009602, 0.00007236, -0.00006674, 0.00006905, -0.00008065, 0.00003764, 0.00003977, -0.00002096, -0.00007688, 0.00009756, -0.00010165, 0.00007769, -0.00002968, 0.00000269, 0.00000006, -0.00001638, 0.00004685, -0.00004052, -0.00000966, 0.00001426, },
		{ 0.00000568, -0.00000610, 0.00001066, 0.00001248, -0.00004022, 0.00000188, 0.00003101, -0.00003311, 0.00003128, -0.00004071, 0.00003493, 0.00000146, -0.00004985, 0.00011751, -0.00012086, 0.00004441, 0.00003384, -0.00003833, -0.00000014, -0.00001232, 0.00002417, -0.00003419, 0.00003848, -0.00004643, 0.00002182, -0.00002509, 0.00000059, -0.00001991, 0.00000038, -0.00007019, 0.00002690, -0.00012917, -0.00065292, 0.00194005, -0.00094228, 0.00010978, -0.00012775, 0.00006063, -0.00009005, 0.00005103, -0.00006837, 0.00004244, -0.00004530, 0.00002031, -0.00000577, -0.00000411, -0.00000511, 0.00001253, -0.00001336, -0.00003136, 0.00007041, -0.00000092, -0.00008821, 0.00006089, -0.00003955, -0.00000150, 0.00004388, -0.00004938, 0.00003625, -0.00004244, 0.00005052, -0.00001772, -0.00003019, 0.00001721, },
		{ 0.00000052, 0.00000791, -0.00001212, 0.00003677, -0.00004588, -0.00000118, 0.00001439, 0.00000342, -0.00001758, 0.00002494, -0.00005166, 0.00009374, -0.00012869, 0.00015680, -0.00009476, 0.00000496, 0.00003644, -0.00000285, -0.00004514, 0.00002925, -0.00002892, 0.00003567, -0.00004782, 0.00005969, -0.00010021, 0.00010882, -0.00014323, 0.00014034, -0.00018006, 0.00016479, -0.00029490, 0.00037023, -0.00113825, 0.00164226, -0.00003010, -0.00055841, 0.00025439, -0.00021990, 0.00013628, -0.00014431, 0.00010086, -0.00011174, 0.00009350, -0.00010535, 0.00010220, -0.00009252, 0.00006912, -0.00005989, 0.00006396, -0.00009475, 0.00009115, 0.00000312, -0.00005249, -0.00001916, 0.00004136, -0.00007401, 0.00008855, -0.00006355, 0.00003490, -0.00003191, 0.00002194, 0.00002139, -0.00005099, 0.00001918, },
		{ -0.00000029, 0.00001682, -0.00002384, 0.00004199, -0.00003355, -0.00001014, -0.00000128, 0.00003129, -0.00004729, 0.00005806, -0.00008329, 0.00010444, -0.00010613, 0.00008845, 0.00001139, -0.00006900, 0.00005548, 0.00000734, -0.00005206, 0.00002978, -0.00004123, 0.00005831, -0.00007636, 0.00009436, -0.00013318, 0.00013471, -0.00016117, 0.00015537, -0.00019350, 0.00019663, -0.00033697, 0.00041077, -0.00089818, 0.00064949, 0.00110513, -0.00107339, 0.00043769, -0.00034807, 0.00023867, -0.00023120, 0.00017263, -0.00017631, 0.00014727, -0.00014975, 0.00013253, -0.00011014, 0.00008305, -0.00007944, 0.00008544, -0.00009708, 0.00006321, 0.00002429, -0.00001913, -0.00007390, 0.00007014, -0.00008072, 0.00006449, -0.00002029, -0.00000813, 0.00000778, -0.00002150, 0.00005571, -0.00005624, 0.00001055, },
		{ 0.00001045, 0.00001053, -0.00001369, 0.00002064, 0.00000039, -0.00003219, 0.00000122, 0.00002580, -0.00003039, 0.00003128, -0.00003915, 0.00003002, -0.00000340, -0.00003911, 0.00013317, -0.00012452, 0.00006349, 0.00000405, -0.00002906, -0.00000415, -0.00001383, 0.00002841, -0.00003805, 0.00004695, -0.00006759, 0.00004760, -0.00005431, 0.00003241, -0.00005203, 0.00004145, -0.00013181, 0.00010423, -0.00026488, -0.00043251, 0.00180626, -0.00098317, 0.00018040, -0.00017878, 0.00010881, -0.00012277, 0.00007596, -0.00009212, 0.00006676, -0.00007253, 0.00005316, -0.00003218, 0.00001550, -0.00002468, 0.00003181, -0.00002849, -0.00001439, 0.00006987, -0.00001453, -0.00006995, 0.00002753, -0.00002345, -0.00000732, 0.00004689, -0.00005536, 0.00003948, -0.00004532, 0.00005904, -0.00003216, -0.00001521, },
		{ 0.00002554, -0.00000508, 0.00000979, -0.00001322, 0.00003856, -0.00005362, 0.00001303, -0.00000172, 0.00001349, -0.00002455, 0.00003321, -0.00006045, 0.00009134, -0.00012725, 0.00018011, -0.00010049, 0.00002537, 0.00001688, -0.00001087, -0.00003397, 0.00001420, -0.00001242, 0.00001858, -0.00002433, 0.00002559, -0.00006577, 0.00007375, -0.00010742, 0.00010374, -0.00013329, 0.00010232, -0.00022418, 0.00026490, -0.00102005, 0.00167507, -0.00021004, -0.00041768, 0.00016796, -0.00015135, 0.00009104, -0.00010707, 0.00006539, -0.00007738, 0.00005912, -0.00007142, 0.00008001, -0.00007793, 0.00005508, -0.00004865, 0.00005889, -0.00009134, 0.00010469, -0.00001912, -0.00003198, -0.00004222, 0.00003940, -0.00006422, 0.00008035, -0.00006087, 0.00002992, -0.00002862, 0.00002492, 0.00001505, -0.00004824, },
		{ 0.00002637, -0.00001203, 0.00002483, -0.00003307, 0.00005349, -0.00005061, 0.00001019, -0.00001863, 0.00004008, -0.00005376, 0.00006573, -0.00008935, 0.00010060, -0.00011038, 0.00011311, 0.00000983, -0.00005335, 0.00005055, -0.00001563, -0.00003390, 0.00001210, -0.00002498, 0.00004254, -0.00005593, 0.00006903, -0.00011381, 0.00011869, -0.00014766, 0.00014339, -0.00017578, 0.00017484, -0.00032181, 0.00039059, -0.00091925, 0.00083186, 0.00085494, -0.00094173, 0.00037956, -0.00030736, 0.00021767, -0.00021151, 0.00015227, -0.00015690, 0.00012923, -0.00013561, 0.00013098, -0.00011284, 0.00008415, -0.00008313, 0.00009395, -0.00010639, 0.00008491, 0.00000309, -0.00000805, -0.00007762, 0.00004849, -0.00005960, 0.00005125, -0.00001421, -0.00001712, 0.00001385, -0.00002226, 0.00005676, -0.00006416, },
		{ 0.00000131, 0.00000123, 0.00001713, -0.00002326, 0.00003247, -0.00001487, -0.00001763, -0.00000660, 0.00002518, -0.00003016, 0.00003301, -0.00003891, 0.00002394, -0.00001157, -0.00001765, 0.00013908, -0.00011545, 0.00007148, -0.00002512, -0.00001597, -0.00001269, -0.00000916, 0.00002573, -0.00003387, 0.00004416, -0.00007574, 0.00006123, -0.00007247, 0.00005417, -0.00007321, 0.00006910, -0.00017679, 0.00016241, -0.00038378, -0.00019425, 0.00160351, -0.00095444, 0.00020859, -0.00020199, 0.00013632, -0.00013965, 0.00008772, -0.00010273, 0.00007793, -0.00008747, 0.00007695, -0.00005377, 0.00003187, -0.00004167, 0.00005018, -0.00004445, 0.00000747, 0.00005730, -0.00001741, -0.00005692, -0.00000295, -0.00000501, -0.00001340, 0.00004740, -0.00006106, 0.00004195, -0.00004476, 0.00006257, -0.00004401, },
		{ -0.00003831, 0.00002514, -0.00000517, 0.00000588, -0.00000840, 0.00003299, -0.00005184, 0.00001908, -0.00001306, 0.00002036, -0.00002833, 0.00003797, -0.00006611, 0.00007994, -0.00011258, 0.00019480, -0.00009733, 0.00004195, -0.00000944, -0.00001086, -0.00002915, 0.00000554, -0.00000152, 0.00000698, -0.00000756, -0.00000302, -0.00003342, 0.00003809, -0.00006890, 0.00006399, -0.00008325, 0.00003495, -0.00014331, 0.00013656, -0.00084300, 0.00162145, -0.00033303, -0.00029594, 0.00008844, -0.00008597, 0.00004691, -0.00007067, 0.00003199, -0.00004505, 0.00002518, -0.00003448, 0.00005175, -0.00005728, 0.00003522, -0.00003046, 0.00004571, -0.00007718, 0.00010457, -0.00003366, -0.00001225, -0.00006665, 0.00004235, -0.00005572, 0.00007180, -0.00006055, 0.00002615, -0.00002455, 0.00002764, 0.00000508, },
		{ -0.00006297, 0.00003521, -0.00001842, 0.00002718, -0.00003656, 0.00005827, -0.00006032, 0.00002603, -0.00003472, 0.00004804, -0.00005853, 0.00007072, -0.00009193, 0.00008830, -0.00010639, 0.00013672, 0.00000889, -0.00003247, 0.00003264, -0.00002913, -0.00002088, -0.00000196, -0.00001020, 0.00002700, -0.00003513, 0.00003999, -0.00008632, 0.00009182, -0.00012141, 0.00011800, -0.00014314, 0.00013401, -0.00027971, 0.00033029, -0.00087680, 0.00095266, 0.00061598, -0.00078760, 0.00029905, -0.00024826, 0.00018093, -0.00017912, 0.00012197, -0.00012779, 0.00010021, -0.00010825, 0.00011552, -0.00010388, 0.00007506, -0.00007653, 0.00009243, -0.00010539, 0.00009746, -0.00001758, 0.00000837, -0.00008696, 0.00003370, -0.00004060, 0.00003964, -0.00001230, -0.00002380, 0.00001867, -0.00001969, 0.00005050, },
		{ -0.00004971, 0.00001516, -0.00000688, 0.00002212, -0.00003134, 0.00004250, -0.00002902, -0.00000078, -0.00001889, 0.00002803, -0.00003184, 0.00003544, -0.00003857, 0.00001498, -0.00002081, 0.00001204, 0.00013799, -0.00009483, 0.00006463, -0.00004757, -0.00000396, -0.00002195, -0.00000140, 0.00001868, -0.00002416, 0.00003215, -0.00007133, 0.00006141, -0.00007691, 0.00006262, -0.00007996, 0.00007877, -0.00019834, 0.00019089, -0.00046707, 0.00003347, 0.00135784, -0.00086864, 0.00019726, -0.00019712, 0.00014198, -0.00013964, 0.00008571, -0.00009972, 0.00007476, -0.00008758, 0.00008764, -0.00006611, 0.00004098, -0.00005196, 0.00006325, -0.00005660, 0.00002889, 0.00003688, -0.00001000, -0.00005154, -0.00002691, 0.00001225, -0.00001778, 0.00004437, -0.00006565, 0.00004334, -0.00004034, 0.00005943, },
		{ -0.00000460, -0.00002479, 0.00002041, -0.00000178, -0.00000179, 0.00000036, 0.00002213, -0.00004183, 0.00001645, -0.00001798, 0.00002262, -0.00002839, 0.00003916, -0.00006875, 0.00005997, -0.00008491, 0.00020060, -0.00008254, 0.00004786, -0.00003587, -0.00000624, -0.00002915, 0.00000286, 0.00000330, 0.00000189, 0.00000035, -0.00002201, -0.00000892, 0.00000846, -0.00003452, 0.00002827, -0.00003850, -0.00002641, -0.00006640, 0.00000447, -0.00062822, 0.00149159, -0.00039333, -0.00020501, 0.00002653, -0.00003260, 0.00001089, -0.00004089, 0.00000609, -0.00002061, -0.00000165, -0.00000160, 0.00002358, -0.00003554, 0.00001403, -0.00000939, 0.00002789, -0.00005521, 0.00009132, -0.00003724, 0.00000187, -0.00008809, 0.00004739, -0.00004770, 0.00006262, -0.00006207, 0.00002320, -0.00001863, 0.00002752, },
		{ 0.00003947, -0.00005393, 0.00003681, -0.00001892, 0.00002327, -0.00003348, 0.00005524, -0.00006053, 0.00003373, -0.00004572, 0.00005240, -0.00005990, 0.00007166, -0.00008972, 0.00006628, -0.00009235, 0.00015653, 0.00001373, -0.00001373, 0.00000621, -0.00003322, -0.00001462, -0.00001001, 0.00000012, 0.00001510, -0.00001846, 0.00001316, -0.00005720, 0.00006039, -0.00008812, 0.00008475, -0.00010195, 0.00008162, -0.00021907, 0.00023896, -0.00077504, 0.00100122, 0.00041112, -0.00063247, 0.00021135, -0.00018170, 0.00013669, -0.00014080, 0.00008810, -0.00009555, 0.00006681, -0.00007362, 0.00009032, -0.00008588, 0.00005776, -0.00006094, 0.00008122, -0.00009343, 0.00009772, -0.00003207, 0.00002504, -0.00009901, 0.00002516, -0.00002476, 0.00003043, -0.00001468, -0.00002819, 0.00002315, -0.00001595, },
		{ 0.00004997, -0.00004770, 0.00002344, -0.00001108, 0.00002280, -0.00003500, 0.00004746, -0.00003788, 0.00001348, -0.00003158, 0.00003190, -0.00003389, 0.00003705, -0.00003643, 0.00000126, -0.00002849, 0.00004635, 0.00013514, -0.00006786, 0.00004325, -0.00005953, 0.00000291, -0.00002813, 0.00000582, 0.00001072, -0.00001281, 0.00001518, -0.00005786, 0.00005026, -0.00006846, 0.00005793, -0.00007242, 0.00007018, -0.00019486, 0.00018532, -0.00050239, 0.00022539, 0.00109819, -0.00074498, 0.00015600, -0.00016940, 0.00012904, -0.00012532, 0.00007259, -0.00008589, 0.00005969, -0.00007407, 0.00008452, -0.00006742, 0.00004089, -0.00005317, 0.00006799, -0.00006126, 0.00004466, 0.00001438, 0.00000452, -0.00005371, -0.00004242, 0.00002583, -0.00001915, 0.00003730, -0.00006873, 0.00004405, -0.00003308, },
		{ 0.00002304, -0.00001050, -0.00001177, 0.00001466, 0.00000205, -0.00001008, 0.00000979, 0.00000972, -0.00002716, 0.00000712, -0.00001693, 0.00002035, -0.00002528, 0.00003769, -0.00006964, 0.00003344, -0.00004687, 0.00020058, -0.00005724, 0.00003875, -0.00005587, -0.00000171, -0.00003075, 0.00000387, 0.00000357, 0.00000204, -0.00000001, -0.00003022, 0.00000475, -0.00001079, -0.00000932, 0.00000205, -0.00000581, -0.00007271, -0.00000611, -0.00011173, -0.00040184, 0.00130561, -0.00039591, -0.00014864, -0.00001258, 0.00000394, -0.00001308, -0.00002081, -0.00000963, -0.00000718, -0.00001713, 0.00002174, 0.00000103, -0.00001752, -0.00000384, 0.00000992, 0.00000998, -0.00003026, 0.00006849, -0.00002973, 0.00000756, -0.00010300, 0.00005176, -0.00003919, 0.00005242, -0.00006490, 0.00002104, -0.00001071, },
		{ -0.00001343, 0.00002739, -0.00004025, 0.00003304, -0.00001499, 0.00001457, -0.00002515, 0.00004569, -0.00005173, 0.00003202, -0.00004941, 0.00005137, -0.00005698, 0.00006811, -0.00008265, 0.00003501, -0.00006869, 0.00017230, 0.00002761, -0.00000486, -0.00002175, -0.00003097, -0.00001413, -0.00001191, 0.00000495, 0.00000843, -0.00000859, -0.00000690, -0.00003241, 0.00003100, -0.00005442, 0.00005042, -0.00006009, 0.00002726, -0.00015173, 0.00013161, -0.00062794, 0.00097867, 0.00025418, -0.00049334, 0.00012992, -0.00011799, 0.00009295, -0.00010315, 0.00005669, -0.00006647, 0.00003597, -0.00003882, 0.00006137, -0.00006343, 0.00003624, -0.00003980, 0.00006307, -0.00007262, 0.00008561, -0.00003684, 0.00003729, -0.00011015, 0.00002112, -0.00001214, 0.00002368, -0.00002096, -0.00003043, 0.00002810, },
		{ -0.00002507, 0.00003658, -0.00003883, 0.00002539, -0.00001036, 0.00001814, -0.00003288, 0.00004574, -0.00003884, 0.00002119, -0.00004072, 0.00003398, -0.00003456, 0.00003642, -0.00003122, -0.00001836, -0.00003274, 0.00008208, 0.00013661, -0.00004285, 0.00001194, -0.00006066, 0.00000275, -0.00002883, 0.00000965, 0.00000495, -0.00000371, -0.00000172, -0.00004058, 0.00003250, -0.00005096, 0.00004344, -0.00005423, 0.00004727, -0.00017019, 0.00014870, -0.00048637, 0.00036394, 0.00085107, -0.00060492, 0.00009814, -0.00012762, 0.00010381, -0.00010189, 0.00005322, -0.00006634, 0.00003791, -0.00005139, 0.00007028, -0.00005887, 0.00003227, -0.00004540, 0.00006369, -0.00005688, 0.00005115, -0.00000455, 0.00002149, -0.00006128, -0.00004939, 0.00003450, -0.00001700, 0.00002627, -0.00007025, 0.00004492, },
		{ -0.00000185, 0.00001446, -0.00001083, -0.00000238, 0.00001082, 0.00000353, -0.00001592, 0.00001657, -0.00000015, -0.00001249, -0.00000519, -0.00001206, 0.00001489, -0.00002037, 0.00003511, -0.00007067, 0.00000335, -0.00000277, 0.00020083, -0.00002705, 0.00001449, -0.00006504, -0.00000144, -0.00003046, 0.00000545, 0.00000199, 0.00000462, -0.00000577, -0.00002949, 0.00000785, -0.00001843, 0.00000459, -0.00001219, 0.00001149, -0.00009880, 0.00002926, -0.00019611, -0.00019000, 0.00108930, -0.00035416, -0.00012309, -0.00002943, 0.00002328, -0.00002467, -0.00001036, -0.00001566, -0.00000446, -0.00002046, 0.00003307, -0.00001256, -0.00000647, -0.00001498, 0.00002365, -0.00000385, -0.00000764, 0.00004153, -0.00001379, 0.00000443, -0.00010941, 0.00005346, -0.00002953, 0.00004101, -0.00006868, 0.00002014, },
		{ 0.00003358, -0.00001367, 0.00001776, -0.00002596, 0.00002701, -0.00000941, 0.00000394, -0.00001448, 0.00003286, -0.00003690, 0.00002232, -0.00004577, 0.00004473, -0.00005007, 0.00006087, -0.00007204, -0.00000314, -0.00003843, 0.00018721, 0.00004996, -0.00001135, -0.00004381, -0.00002724, -0.00001643, -0.00000958, 0.00000526, 0.00000648, -0.00000578, -0.00001818, -0.00001577, 0.00000877, -0.00002608, 0.00002117, -0.00002489, -0.00001968, -0.00009005, 0.00002590, -0.00045639, 0.00089695, 0.00014861, -0.00038023, 0.00006396, -0.00006463, 0.00005566, -0.00007095, 0.00003190, -0.00004494, 0.00001300, -0.00001014, 0.00003475, -0.00004167, 0.00001527, -0.00001769, 0.00004224, -0.00004726, 0.00006401, -0.00003126, 0.00004213, -0.00011720, 0.00001956, -0.00000233, 0.00001907, -0.00003055, -0.00003045, },
		{ 0.00004670, -0.00001867, 0.00002283, -0.00002607, 0.00002247, -0.00000573, 0.00000909, -0.00002573, 0.00003797, -0.00003172, 0.00002053, -0.00004371, 0.00003223, -0.00003271, 0.00003296, -0.00002284, -0.00004344, -0.00003369, 0.00011815, 0.00014708, -0.00002904, -0.00002150, -0.00005411, -0.00000354, -0.00002384, 0.00000898, 0.00000299, 0.00000058, -0.00001434, -0.00002484, 0.00001381, -0.00002990, 0.00002445, -0.00003133, 0.00001706, -0.00013252, 0.00009054, -0.00042523, 0.00044190, 0.00063627, -0.00046788, 0.00003734, -0.00008159, 0.00007372, -0.00007536, 0.00003306, -0.00004678, 0.00001568, -0.00002587, 0.00005004, -0.00004412, 0.00001811, -0.00003107, 0.00005206, -0.00004457, 0.00004744, -0.00001599, 0.00003631, -0.00007110, -0.00004908, 0.00003811, -0.00001159, 0.00001186, -0.00007028, },
		{ 0.00002127, 0.00000591, 0.00000393, -0.00000608, 0.00000216, 0.00001032, 0.00000116, -0.00001749, 0.00001858, -0.00000446, -0.00000206, -0.00001638, -0.00000624, 0.00000814, -0.00001522, 0.00003287, -0.00007337, -0.00002762, 0.00004292, 0.00020890, -0.00000135, -0.00002019, -0.00006266, -0.00000757, -0.00002588, 0.00000497, 0.00000124, 0.00000647, -0.00001296, -0.00002372, 0.00000342, -0.00001635, 0.00000824, -0.00001537, 0.00001406, -0.00010454, 0.00003762, -0.00023978, -0.00001397, 0.00086897, -0.00028610, -0.00011927, -0.00002916, 0.00002879, -0.00002656, -0.00000688, -0.00001507, -0.00000916, -0.00001427, 0.00003331, -0.00001674, -0.00000319, -0.00001814, 0.00002992, -0.00001110, 0.00000849, 0.00001615, 0.00000617, -0.00000574, -0.00010712, 0.00005155, -0.00001866, 0.00002856, -0.00007319, },
		{ -0.00002782, 0.00003861, -0.00001666, 0.00001271, -0.00001448, 0.00002185, -0.00000514, -0.00000540, -0.00000487, 0.00002081, -0.00002053, 0.00000813, -0.00003681, 0.00003378, -0.00004056, 0.00005161, -0.00006023, -0.00004445, -0.00000669, 0.00020790, 0.00007534, -0.00003439, -0.00005446, -0.00002681, -0.00001782, -0.00000603, 0.00000339, 0.00000707, -0.00000803, -0.00002146, -0.00000819, -0.00000383, -0.00000661, 0.00000098, -0.00000131, -0.00005239, -0.00004398, -0.00006180, -0.00028384, 0.00077560, 0.00008854, -0.00029557, 0.00001727, -0.00002516, 0.00002769, -0.00004630, 0.00001511, -0.00003236, 0.00000027, 0.00000857, 0.00001496, -0.00002478, -0.00000103, 0.00000111, 0.00002312, -0.00002248, 0.00003789, -0.00001749, 0.00003887, -0.00011826, 0.00001880, 0.00000516, 0.00001621, -0.00004281, },
		{ -0.00006865, 0.00004943, -0.00001556, 0.00001232, -0.00001339, 0.00001762, 0.00000023, -0.00000172, -0.00001610, 0.00002691, -0.00001899, 0.00001243, -0.00004015, 0.00002618, -0.00002841, 0.00002732, -0.00001270, -0.00007154, -0.00003409, 0.00015698, 0.00016748, -0.00003393, -0.00004813, -0.00004543, -0.00001268, -0.00001517, 0.00000474, 0.00000448, -0.00000039, -0.00002056, -0.00001444, -0.00000082, -0.00001071, 0.00000656, -0.00001019, -0.00001241, -0.00009211, 0.00002451, -0.00033303, 0.00046277, 0.00046431, -0.00034790, -0.00001538, -0.00003971, 0.00004529, -0.00005095, 0.00001657, -0.00003185, -0.00000150, -0.00000380, 0.00002970, -0.00002799, 0.00000271, -0.00001417, 0.00003663, -0.00002766, 0.00003547, -0.00001863, 0.00004571, -0.00008008, -0.00004340, 0.00003723, -0.00000357, -0.00000518, },
		{ -0.00007816, 0.00002504, 0.00001029, -0.00000516, 0.00000115, 0.00000285, 0.00001264, -0.00000460, -0.00001491, 0.00001580, -0.00000239, 0.00000180, -0.00002344, -0.00000191, 0.00000167, -0.00001080, 0.00003143, -0.00007798, -0.00005842, 0.00008753, 0.00023146, 0.00000873, -0.00005654, -0.00005201, -0.00001925, -0.00001677, 0.00000134, 0.00000276, 0.00000538, -0.00001799, -0.00001732, -0.00000404, -0.00000855, 0.00000508, -0.00001113, 0.00000604, -0.00009435, 0.00002325, -0.00024240, 0.00011335, 0.00066672, -0.00021013, -0.00012584, -0.00001934, 0.00002597, -0.00002312, -0.00000640, -0.00001227, -0.00001628, -0.00000356, 0.00002625, -0.00001383, -0.00000602, -0.00001449, 0.00002924, -0.00001166, 0.00001628, -0.00000327, 0.00002550, -0.00001992, -0.00009736, 0.00004610, -0.00000699, 0.00001548, },
		{ -0.00005713, -0.00002192, 0.00004131, -0.00002055, 0.00001216, -0.00000749, 0.00001943, -0.00000413, -0.00001112, 0.00000107, 0.00001295, -0.00000714, -0.00000625, -0.00002565, 0.00002076, -0.00003033, 0.00004224, -0.00004961, -0.00008503, 0.00002076, 0.00024315, 0.00009376, -0.00006972, -0.00005201, -0.00003256, -0.00001546, -0.00000411, 0.00000191, 0.00000744, -0.00001208, -0.00001965, -0.00000783, -0.00000727, 0.00000342, -0.00000917, 0.00000913, -0.00006808, -0.00001871, -0.00011991, -0.00013163, 0.00063725, 0.00006165, -0.00023545, -0.00001140, 0.00000077, 0.00000882, -0.00002861, 0.00000505, -0.00002714, -0.00000322, 0.00001672, 0.00000382, -0.00001481, -0.00001037, 0.00001382, 0.00000891, -0.00000274, 0.00001278, 0.00000048, 0.00002893, -0.00011301, 0.00001790, 0.00001056, 0.00001483, },
		{ -0.00002404, -0.00006484, 0.00005215, -0.00001588, 0.00000736, -0.00000437, 0.00001407, 0.00000449, -0.00001101, -0.00000741, 0.00001648, -0.00000491, 0.00000015, -0.00003190, 0.00001705, -0.00002297, 0.00002132, -0.00000350, -0.00009854, -0.00003921, 0.00020490, 0.00019347, -0.00006059, -0.00006057, -0.00004064, -0.00002025, -0.00000622, -0.00000064, 0.00000738, -0.00000492, -0.00002076, -0.00001069, -0.00000856, 0.00000274, -0.00000592, 0.00000402, -0.00003425, -0.00005851, -0.00003493, -0.00022837, 0.00043882, 0.00033629, -0.00025205, -0.00005347, -0.00000736, 0.00002274, -0.00003179, 0.00000603, -0.00002379, -0.00001047, 0.00001031, 0.00001426, -0.00001494, -0.00000971, 0.00000107, 0.00002154, -0.00001070, 0.00001934, -0.00001385, 0.00004848, -0.00008604, -0.00003424, 0.00003268, 0.00000622, },
		{ 0.00000226, -0.00008303, 0.00003130, 0.00000967, -0.00000935, 0.00000708, 0.00000238, 0.00001555, -0.00001153, -0.00001027, 0.00001037, 0.00000427, -0.00000062, -0.00002544, -0.00000005, -0.00000409, -0.00000681, 0.00002971, -0.00008265, -0.00009075, 0.00013191, 0.00027158, -0.00000673, -0.00008383, -0.00003951, -0.00003283, -0.00000528, -0.00000446, 0.00000616, 0.00000102, -0.00001895, -0.00001370, -0.00001019, 0.00000037, -0.00000039, -0.00000445, -0.00000646, -0.00007571, -0.00000459, -0.00021169, 0.00018889, 0.00049716, -0.00014125, -0.00013228, -0.00000761, 0.00002056, -0.00001882, -0.00000512, -0.00001134, -0.00002085, 0.00000597, 0.00001709, -0.00000786, -0.00001154, -0.00000706, 0.00002409, -0.00000773, 0.00001656, -0.00001470, 0.00004058, -0.00003489, -0.00008217, 0.00003784, 0.00000474, },
		{ 0.00001466, -0.00007270, -0.00001237, 0.00003953, -0.00002225, 0.00001387, -0.00000441, 0.00001975, -0.00000649, -0.00001273, 0.00000261, 0.00001079, 0.00000033, -0.00001726, -0.00001527, 0.00000779, -0.00002086, 0.00003389, -0.00004147, -0.00012236, 0.00003953, 0.00030151, 0.00009249, -0.00010804, -0.00003938, -0.00004422, -0.00000863, -0.00000517, 0.00000235, 0.00000553, -0.00001474, -0.00001650, -0.00001111, -0.00000439, 0.00000596, -0.00001110, 0.00000835, -0.00006832, -0.00001377, -0.00014395, -0.00001502, 0.00050294, 0.00005299, -0.00019208, -0.00002678, 0.00001638, -0.00000342, -0.00001551, -0.00000129, -0.00002556, -0.00000115, 0.00001679, 0.00000031, -0.00001131, -0.00001279, 0.00001990, 0.00000068, 0.00000946, -0.00000679, 0.00001813, 0.00001508, -0.00010244, 0.00001667, 0.00001387, },
		{ 0.00001695, -0.00004389, -0.00005816, 0.00005297, -0.00001795, 0.00000812, -0.00000099, 0.00001407, 0.00000473, -0.00001608, -0.00000263, 0.00001041, 0.00000584, -0.00001187, -0.00002226, 0.00000724, -0.00001849, 0.00001727, 0.00000161, -0.00011969, -0.00005553, 0.00027123, 0.00021518, -0.00010605, -0.00005537, -0.00004410, -0.00002239, -0.00000045, -0.00000430, 0.00000888, -0.00000992, -0.00001746, -0.00001232, -0.00000921, 0.00000905, -0.00001112, 0.00000879, -0.00004452, -0.00003814, -0.00007601, -0.00013010, 0.00038739, 0.00024565, -0.00018069, -0.00007536, 0.00001375, 0.00000747, -0.00001863, 0.00000128, -0.00002207, -0.00001120, 0.00001492, 0.00000639, -0.00000775, -0.00001631, 0.00001153, 0.00001014, 0.00000209, 0.00000400, -0.00000498, 0.00004546, -0.00008815, -0.00002305, 0.00002528, },
		{ 0.00001574, -0.00001071, -0.00008678, 0.00003884, 0.00000398, -0.00000634, 0.00000809, 0.00000398, 0.00001603, -0.00001641, -0.00000681, 0.00000594, 0.00001161, -0.00000644, -0.00002389, 0.00000022, -0.00001016, -0.00000157, 0.00002502, -0.00008349, -0.00012925, 0.00018145, 0.00032664, -0.00005335, -0.00009230, -0.00003277, -0.00004299, 0.00000467, -0.00000974, 0.00000928, -0.00000490, -0.00001624, -0.00001412, -0.00001229, 0.00000674, -0.00000424, 0.00000005, -0.00001736, -0.00005680, -0.00003435, -0.00016105, 0.00021869, 0.00036606, -0.00008873, -0.00013125, 0.00000030, 0.00001696, -0.00001687, -0.00000067, -0.00001467, -0.00001958, 0.00000979, 0.00001077, -0.00000309, -0.00001589, 0.00000039, 0.00001797, -0.00000293, 0.00001245, -0.00001873, 0.00004966, -0.00004809, -0.00006373, 0.00002778, },
		},
		{ { -0.00002711, -0.00002236, 0.00005959, 0.00000686, -0.00007465, 0.00001337, -0.00000370, -0.00000961, -0.00001020, -0.00000587, 0.00004607, -0.00001050, -0.00002448, -0.00000783, 0.00000475, -0.00001161, -0.00002063, -0.00001296, -0.00016410, 0.00005830, 0.00038050, 0.00005300, -0.00014562, -0.00001444, -0.00001936, -0.00001133, -0.00001692, -0.00001080, 0.00002571, -0.00001226, -0.00002040, 0.00000568, 0.00000924, 0.00000568, -0.00002040, -0.00001226, 0.00002571, -0.00001080, -0.00001692, -0.00001133, -0.00001936, -0.00001444, -0.00014562, 0.00005300, 0.00038050, 0.00005830, -0.00016410, -0.00001296, -0.00002063, -0.00001161, 0.00000475, -0.00000783, -0.00002448, -0.00001050, 0.00004607, -0.00000587, -0.00001020, -0.00000961, -0.00000370, 0.00001337, -0.00007465, 0.00000686, 0.00005959, -0.00002236, },
		{ -0.00002614, -0.00003367, 0.00003196, 0.00005002, -0.00006771, -0.00001967, 0.00001444, -0.00001449, -0.00000496, -0.00001920, 0.00003458, 0.00001990, -0.00003256, -0.00001085, -0.00000218, 0.00000212, -0.00002897, 0.00000396, -0.00011301, -0.00011285, 0.00035358, 0.00023450, -0.00015620, -0.00005108, -0.00001235, -0.00002012, -0.00000648, -0.00002469, 0.00001907, 0.00000676, -0.00002601, -0.00000367, 0.00001088, 0.00000711, -0.00000710, -0.00002528, 0.00001789, 0.00000788, -0.00002139, -0.00000996, -0.00002084, -0.00000059, -0.00010083, -0.00006787, 0.00029256, 0.00021925, -0.00013178, -0.00006892, -0.00000180, -0.00002748, 0.00000808, -0.00000570, -0.00001391, -0.00002692, 0.00003190, 0.00002119, -0.00002026, 0.00000187, -0.00002157, 0.00002572, -0.00004898, -0.00004418, 0.00006715, 0.00000035, },
		{ -0.00002217, -0.00003400, -0.00000208, 0.00006961, -0.00003290, -0.00005272, 0.00001641, -0.00000709, -0.00000642, -0.00001960, 0.00001059, 0.00004162, -0.00002142, -0.00002380, -0.00000245, 0.00000137, -0.00001635, -0.00001384, -0.00003477, -0.00020221, 0.00018926, 0.00041552, -0.00008722, -0.00011446, -0.00000273, -0.00003233, 0.00000114, -0.00002660, 0.00000135, 0.00002116, -0.00001877, -0.00001756, 0.00001054, 0.00000767, 0.00000223, -0.00002360, -0.00000167, 0.00002317, -0.00001729, -0.00001204, -0.00002141, 0.00000187, -0.00005517, -0.00011664, 0.00015457, 0.00030059, -0.00002259, -0.00012055, -0.00000475, -0.00002531, -0.00000419, 0.00000434, -0.00001243, -0.00002252, 0.00000098, 0.00004440, -0.00001968, 0.00000359, -0.00002060, 0.00000880, -0.00000412, -0.00007975, 0.00004612, 0.00002943, },
		{ -0.00001319, -0.00002681, -0.00003017, 0.00006207, 0.00001295, -0.00006699, -0.00000127, 0.00000577, -0.00000819, -0.00001491, -0.00001059, 0.00004239, 0.00000588, -0.00003567, -0.00000557, -0.00000184, -0.00000388, -0.00002629, 0.00000231, -0.00016875, -0.00004823, 0.00049222, 0.00009241, -0.00018536, -0.00000548, -0.00003762, -0.00000312, -0.00001550, -0.00001885, 0.00002363, -0.00000280, -0.00002587, 0.00000121, 0.00001257, 0.00000307, -0.00001039, -0.00002143, 0.00002649, -0.00000543, -0.00001329, -0.00002275, -0.00000144, -0.00002111, -0.00011395, 0.00002818, 0.00028635, 0.00011007, -0.00012179, -0.00003961, -0.00001014, -0.00002151, 0.00001146, -0.00001068, -0.00001409, -0.00001996, 0.00004068, 0.00000402, -0.00001210, 0.00000013, -0.00002233, 0.00003095, -0.00007892, -0.00000094, 0.00005455, },
		{ 0.00000224, -0.00001657, -0.00004456, 0.00003414, 0.00005084, -0.00005473, -0.00002959, 0.00001344, -0.00000612, -0.00000899, -0.00002337, 0.00002755, 0.00003031, -0.00002875, -0.00002210, 0.00000376, -0.00000703, -0.00001393, -0.00001231, -0.00006798, -0.00022533, 0.00037757, 0.00035420, -0.00020913, -0.00004361, -0.00002734, -0.00002075, 0.00000036, -0.00003002, 0.00001158, 0.00001430, -0.00002324, -0.00001382, 0.00001536, 0.00000368, 0.00000024, -0.00002662, 0.00001240, 0.00001202, -0.00001388, -0.00002016, -0.00001225, 0.00000341, -0.00008950, -0.00005119, 0.00020219, 0.00021123, -0.00006456, -0.00007923, -0.00000841, -0.00002318, 0.00000115, 0.00000233, -0.00001618, -0.00001757, 0.00001079, 0.00003745, -0.00002779, 0.00001716, -0.00003625, 0.00002920, -0.00003812, -0.00005653, 0.00006234, },
		{ 0.00002414, -0.00000816, -0.00004194, -0.00000321, 0.00006884, -0.00002227, -0.00005261, 0.00000778, 0.00000025, -0.00000406, -0.00002592, 0.00000565, 0.00004049, -0.00000490, -0.00003947, 0.00000497, -0.00001045, -0.00000100, -0.00002918, -0.00000176, -0.00024002, 0.00009286, 0.00057771, -0.00010594, -0.00013246, -0.00000303, -0.00004460, 0.00000969, -0.00002636, -0.00000920, 0.00002375, -0.00001103, -0.00002464, 0.00000723, 0.00001138, 0.00000022, -0.00001486, -0.00001181, 0.00002669, -0.00001234, -0.00001207, -0.00002785, 0.00001841, -0.00006045, -0.00008193, 0.00009455, 0.00024818, 0.00002725, -0.00009172, -0.00003005, -0.00001159, -0.00001732, 0.00001503, -0.00001855, -0.00000863, -0.00001561, 0.00004788, -0.00001626, 0.00000632, -0.00001517, -0.00000838, 0.00001775, -0.00009044, 0.00004074, },
		{ 0.00004741, -0.00000404, -0.00002447, -0.00003814, 0.00006422, 0.00001565, -0.00005661, -0.00001232, 0.00000885, -0.00000312, -0.00001669, -0.00001856, 0.00003904, 0.00001703, -0.00003503, -0.00001570, 0.00000413, -0.00001222, -0.00001140, -0.00001234, -0.00012122, -0.00020322, 0.00059629, 0.00016823, -0.00024463, 0.00001195, -0.00005974, 0.00000468, -0.00001262, -0.00002539, 0.00001760, 0.00000704, -0.00002634, -0.00000871, 0.00001881, -0.00000135, -0.00000156, -0.00002654, 0.00002478, -0.00000218, -0.00000636, -0.00003507, 0.00001482, -0.00002635, -0.00008653, 0.00000968, 0.00021342, 0.00012355, -0.00007082, -0.00005338, -0.00001220, -0.00001963, 0.00000678, -0.00000391, -0.00001449, -0.00001455, 0.00002162, 0.00002264, -0.00002392, 0.00002058, -0.00004528, 0.00004550, -0.00007623, -0.00001091, },
		{ 0.00005696, 0.00000081, -0.00000287, -0.00005716, 0.00003848, 0.00004696, -0.00004014, -0.00003693, 0.00001253, -0.00000404, -0.00000205, -0.00003556, 0.00002693, 0.00002821, -0.00000974, -0.00004487, 0.00002005, -0.00002843, 0.00001441, -0.00004426, -0.00000621, -0.00032176, 0.00033381, 0.00053216, -0.00027979, -0.00003096, -0.00004524, -0.00001977, 0.00000357, -0.00002842, -0.00000234, 0.00002307, -0.00001937, -0.00002126, 0.00001328, 0.00000752, -0.00000116, -0.00001984, 0.00000290, 0.00001686, -0.00000889, -0.00002456, -0.00001038, 0.00001230, -0.00008374, -0.00003211, 0.00012913, 0.00019427, -0.00002550, -0.00005660, -0.00003439, -0.00000497, -0.00001661, 0.00001833, -0.00002683, 0.00000048, -0.00001354, 0.00005178, -0.00003455, 0.00002916, -0.00004057, 0.00001987, -0.00001772, -0.00007018, },
		{ 0.00003350, 0.00001599, 0.00000757, -0.00004901, -0.00000410, 0.00006666, -0.00001393, -0.00004945, 0.00000163, 0.00000107, 0.00000365, -0.00003126, -0.00000084, 0.00003676, 0.00001092, -0.00004747, 0.00000306, -0.00001249, 0.00000113, -0.00002660, 0.00000202, -0.00021542, -0.00008057, 0.00076461, -0.00010506, -0.00016666, 0.00000513, -0.00006099, 0.00001754, -0.00002296, -0.00002050, 0.00002354, -0.00000285, -0.00002614, -0.00000349, 0.00002060, -0.00000646, -0.00000414, -0.00002054, 0.00002900, -0.00001040, -0.00000678, -0.00003842, 0.00003677, -0.00006821, -0.00004792, 0.00004390, 0.00020758, 0.00004188, -0.00004662, -0.00005170, -0.00000412, -0.00002314, 0.00001758, -0.00001777, -0.00000282, -0.00001898, 0.00003706, -0.00000204, -0.00000364, 0.00000502, -0.00003513, 0.00004127, -0.00009630, },
		{ -0.00002425, 0.00004028, 0.00000185, -0.00001701, -0.00004934, 0.00006940, 0.00001232, -0.00004216, -0.00002286, 0.00001248, -0.00000428, -0.00000655, -0.00003618, 0.00004315, 0.00001434, -0.00001549, -0.00004312, 0.00002722, -0.00004094, 0.00002745, -0.00005823, -0.00003153, -0.00037320, 0.00064879, 0.00030938, -0.00033644, 0.00004495, -0.00009102, 0.00001757, -0.00001486, -0.00002368, 0.00000289, 0.00002019, -0.00002593, -0.00001697, 0.00001939, 0.00000101, -0.00000132, -0.00002283, 0.00001761, 0.00000245, -0.00000119, -0.00004281, 0.00002488, -0.00002675, -0.00006622, -0.00000001, 0.00015163, 0.00012429, -0.00003869, -0.00003824, -0.00003486, -0.00000104, -0.00001212, 0.00001430, -0.00002686, 0.00000440, -0.00000685, 0.00004531, -0.00004021, 0.00004266, -0.00006164, 0.00004835, -0.00006233, },
		{ -0.00008252, 0.00005115, -0.00000493, 0.00001520, -0.00007147, 0.00004491, 0.00003927, -0.00002500, -0.00004300, 0.00001490, -0.00001016, 0.00001351, -0.00005040, 0.00002800, 0.00001906, 0.00001535, -0.00006746, 0.00003418, -0.00004770, 0.00003840, -0.00006890, 0.00003814, -0.00035835, 0.00019203, 0.00077041, -0.00035798, -0.00002117, -0.00006161, -0.00001862, 0.00000277, -0.00001969, -0.00002152, 0.00003235, -0.00001661, -0.00002158, 0.00000102, 0.00002033, -0.00001095, -0.00000699, -0.00000901, 0.00002290, -0.00000846, -0.00002143, -0.00001610, 0.00002703, -0.00008580, -0.00000626, 0.00006211, 0.00018457, -0.00001910, -0.00001159, -0.00006561, 0.00001432, -0.00003310, 0.00003119, -0.00003510, 0.00001524, -0.00002890, 0.00005402, -0.00003095, 0.00002764, -0.00002735, -0.00000397, 0.00000984, },
		{ -0.00009167, 0.00002233, 0.00000737, 0.00002146, -0.00005159, -0.00000837, 0.00006721, -0.00001153, -0.00004112, -0.00000522, 0.00000229, 0.00000598, -0.00002461, -0.00001786, 0.00004050, 0.00001339, -0.00003524, -0.00001959, 0.00000711, -0.00002294, 0.00000824, -0.00003923, -0.00012097, -0.00031160, 0.00093583, -0.00005062, -0.00023660, 0.00003356, -0.00008969, 0.00003182, -0.00002411, -0.00002453, 0.00001448, 0.00000921, -0.00002573, -0.00001516, 0.00002691, -0.00000878, 0.00000017, -0.00002185, 0.00002627, -0.00000785, -0.00000102, -0.00004767, 0.00005216, -0.00007483, -0.00001740, 0.00000342, 0.00017346, 0.00004175, -0.00001212, -0.00004986, -0.00001617, -0.00001152, 0.00000422, -0.00000351, -0.00001121, -0.00000527, 0.00001132, 0.00002248, -0.00002448, 0.00003365, -0.00006162, 0.00005708, },
		{ -0.00003648, -0.00004003, 0.00003443, 0.00000209, -0.00000258, -0.00006535, 0.00007923, 0.00000211, -0.00002326, -0.00003550, 0.00002183, -0.00001681, 0.00001712, -0.00006199, 0.00005396, -0.00000146, 0.00001903, -0.00008105, 0.00005951, -0.00008037, 0.00007748, -0.00011680, 0.00006713, -0.00050627, 0.00060154, 0.00053721, -0.00046406, 0.00009346, -0.00013220, 0.00003810, -0.00002635, -0.00000968, -0.00002272, 0.00003841, -0.00002898, -0.00001646, 0.00000726, 0.00001561, -0.00001299, -0.00000915, 0.00000420, 0.00001081, -0.00000177, -0.00004187, 0.00002538, -0.00001909, -0.00005715, 0.00000440, 0.00008931, 0.00013181, -0.00003799, 0.00000468, -0.00007147, 0.00002680, -0.00003720, 0.00003601, -0.00004311, 0.00002685, -0.00003329, 0.00006139, -0.00005004, 0.00005367, -0.00006053, 0.00003373, },
		{ 0.00003605, -0.00008750, 0.00004258, -0.00001292, 0.00003443, -0.00008404, 0.00004987, 0.00003198, -0.00001331, -0.00004544, 0.00001780, -0.00001839, 0.00002716, -0.00005603, 0.00002048, 0.00001436, 0.00003153, -0.00007351, 0.00002979, -0.00004701, 0.00003747, -0.00006152, 0.00003372, -0.00030559, -0.00005123, 0.00104267, -0.00041333, -0.00003481, -0.00006322, -0.00002615, 0.00000483, -0.00001084, -0.00003832, 0.00003621, -0.00001219, -0.00001677, -0.00001701, 0.00003544, -0.00002127, 0.00000405, -0.00001734, 0.00002652, -0.00000983, -0.00001432, -0.00002519, 0.00004386, -0.00009304, 0.00002437, 0.00000026, 0.00018147, -0.00003109, 0.00003223, -0.00008152, 0.00001873, -0.00003489, 0.00002959, -0.00003072, 0.00001638, -0.00002653, 0.00003841, -0.00001252, 0.00000952, 0.00000149, -0.00003539, },
		{ 0.00005573, -0.00007242, 0.00000601, 0.00000096, 0.00002953, -0.00004596, -0.00001960, 0.00007426, -0.00001826, -0.00002507, -0.00001736, 0.00001164, -0.00000612, 0.00000051, -0.00004895, 0.00005624, -0.00000317, 0.00000076, -0.00006402, 0.00004869, -0.00007272, 0.00007310, -0.00011906, 0.00002122, -0.00055092, 0.00103300, 0.00009540, -0.00035994, 0.00009211, -0.00013829, 0.00006017, -0.00003843, -0.00001279, -0.00000964, 0.00002983, -0.00002999, -0.00001810, 0.00002000, 0.00000244, -0.00000911, -0.00001104, 0.00001495, -0.00000031, -0.00000071, -0.00004924, 0.00005883, -0.00007478, 0.00000323, -0.00002007, 0.00013378, 0.00004850, -0.00000730, -0.00001686, -0.00004912, 0.00001670, -0.00002259, 0.00002115, -0.00003022, 0.00001951, -0.00002087, 0.00004843, -0.00004600, 0.00005730, -0.00007444, },
		{ 0.00000400, -0.00000454, -0.00005318, 0.00002826, -0.00000252, 0.00001702, -0.00008399, 0.00009140, -0.00001196, -0.00000297, -0.00004890, 0.00003593, -0.00003706, 0.00004596, -0.00008545, 0.00006225, -0.00002016, 0.00005436, -0.00011276, 0.00008334, -0.00011058, 0.00011477, -0.00015748, 0.00013165, -0.00055133, 0.00042328, 0.00084857, -0.00061009, 0.00014308, -0.00017376, 0.00006054, -0.00004358, 0.00001309, -0.00005456, 0.00005804, -0.00003517, -0.00000580, -0.00001710, 0.00003952, -0.00003193, 0.00000913, -0.00001127, 0.00002090, -0.00000744, -0.00003319, 0.00001786, -0.00000447, -0.00005942, 0.00002342, 0.00002290, 0.00015060, -0.00006187, 0.00005770, -0.00010396, 0.00004824, -0.00005423, 0.00004852, -0.00005120, 0.00003997, -0.00004535, 0.00006073, -0.00004409, 0.00004511, -0.00003830, },
		{ -0.00006049, 0.00004916, -0.00007819, 0.00002813, -0.00001933, 0.00004960, -0.00008959, 0.00004810, 0.00003077, -0.00001045, -0.00003958, 0.00001475, -0.00001861, 0.00002619, -0.00004112, -0.00000440, 0.00002417, 0.00002633, -0.00005111, -0.00000355, -0.00001559, -0.00000002, -0.00000950, -0.00003195, -0.00016393, -0.00036292, 0.00129329, -0.00039972, -0.00009978, -0.00003218, -0.00005566, 0.00002146, -0.00001384, -0.00004087, 0.00002529, 0.00000071, -0.00001460, -0.00002988, 0.00004059, -0.00002040, 0.00000324, -0.00001461, 0.00002198, -0.00000721, -0.00001018, -0.00003015, 0.00005581, -0.00009876, 0.00005206, -0.00004857, 0.00017309, -0.00003925, 0.00005143, -0.00006722, -0.00000063, -0.00001592, 0.00000975, -0.00001102, 0.00000182, -0.00000682, 0.00000841, 0.00001597, -0.00001816, 0.00003417, },
		{ -0.00006415, 0.00003400, -0.00004033, -0.00001272, -0.00000307, 0.00003037, -0.00003147, -0.00003806, 0.00008804, -0.00003372, -0.00000184, -0.00003611, 0.00003167, -0.00003345, 0.00004041, -0.00008773, 0.00007883, -0.00003327, 0.00005137, -0.00011803, 0.00009869, -0.00013232, 0.00014794, -0.00020965, 0.00016569, -0.00073348, 0.00099929, 0.00035676, -0.00053882, 0.00017879, -0.00020586, 0.00010370, -0.00006870, 0.00001734, -0.00004959, 0.00005976, -0.00004273, -0.00000611, -0.00000515, 0.00002975, -0.00003233, 0.00001077, -0.00000420, 0.00001320, -0.00000833, -0.00004054, 0.00005460, -0.00006535, 0.00000972, -0.00002189, 0.00008088, 0.00007261, -0.00003390, 0.00003778, -0.00009052, 0.00004938, -0.00005082, 0.00004482, -0.00004858, 0.00004256, -0.00004588, 0.00006316, -0.00005529, 0.00006260, },
		{ -0.00000247, -0.00002907, 0.00002157, -0.00005546, 0.00001773, -0.00000795, 0.00003637, -0.00009953, 0.00010005, -0.00002345, 0.00001233, -0.00005741, 0.00004762, -0.00005400, 0.00006518, -0.00009186, 0.00005545, -0.00002767, 0.00007291, -0.00011959, 0.00008014, -0.00011133, 0.00011613, -0.00015246, 0.00012795, -0.00047991, 0.00011818, 0.00120547, -0.00073298, 0.00016443, -0.00019552, 0.00007005, -0.00005426, 0.00003254, -0.00008037, 0.00006914, -0.00003787, 0.00000840, -0.00004408, 0.00006113, -0.00004629, 0.00002130, -0.00002044, 0.00002648, -0.00001262, -0.00002306, 0.00000905, 0.00001097, -0.00006745, 0.00005031, -0.00004069, 0.00017075, -0.00008992, 0.00009605, -0.00011294, 0.00004876, -0.00005141, 0.00004241, -0.00004293, 0.00003591, -0.00003718, 0.00004058, -0.00002122, 0.00001856, },
		{ 0.00005436, -0.00006507, 0.00004041, -0.00005165, 0.00000739, -0.00002166, 0.00005665, -0.00008442, 0.00003615, 0.00003924, -0.00001965, -0.00002225, 0.00000052, -0.00000345, 0.00000361, -0.00000253, -0.00004689, 0.00005017, -0.00000272, -0.00000049, -0.00006295, 0.00004273, -0.00006892, 0.00007936, -0.00014472, 0.00003263, -0.00067849, 0.00145045, -0.00027068, -0.00023830, 0.00004416, -0.00011696, 0.00006225, -0.00003900, -0.00001937, -0.00000766, 0.00002805, -0.00002345, -0.00002635, 0.00002498, 0.00000023, -0.00001544, 0.00000310, 0.00000591, 0.00000381, -0.00001483, -0.00002482, 0.00005707, -0.00009659, 0.00006833, -0.00007507, 0.00014715, -0.00002688, 0.00003524, -0.00002258, -0.00003931, 0.00001792, -0.00002127, 0.00001671, -0.00002074, 0.00002032, -0.00002402, 0.00004216, -0.00004136, },
		{ 0.00004429, -0.00003188, -0.00000109, -0.00000494, -0.00002628, -0.00000822, 0.00002678, -0.00001209, -0.00005928, 0.00010391, -0.00005268, 0.00002328, -0.00005701, 0.00005693, -0.00006688, 0.00008130, -0.00011931, 0.00009518, -0.00006251, 0.00009799, -0.00016059, 0.00013499, -0.00017690, 0.00020326, -0.00027347, 0.00026109, -0.00080196, 0.00080323, 0.00072512, -0.00074901, 0.00027308, -0.00027770, 0.00015206, -0.00010710, 0.00005818, -0.00009682, 0.00009206, -0.00006060, 0.00001842, -0.00004339, 0.00006545, -0.00006067, 0.00003511, -0.00002451, 0.00002813, -0.00002049, -0.00002515, 0.00004337, -0.00004960, 0.00000394, -0.00000489, 0.00001785, 0.00010889, -0.00007647, 0.00009184, -0.00012033, 0.00006962, -0.00006548, 0.00005452, -0.00005423, 0.00005152, -0.00005244, 0.00005781, -0.00004466, },
		{ -0.00001394, 0.00003056, -0.00004886, 0.00002898, -0.00003965, -0.00000064, -0.00000958, 0.00004916, -0.00010558, 0.00009887, -0.00002564, 0.00001595, -0.00005424, 0.00004783, -0.00005542, 0.00006137, -0.00007059, 0.00002568, -0.00001444, 0.00006292, -0.00009081, 0.00004075, -0.00007320, 0.00007136, -0.00009093, 0.00004696, -0.00030028, -0.00026773, 0.00153829, -0.00077556, 0.00012174, -0.00017363, 0.00004828, -0.00004218, 0.00003263, -0.00008473, 0.00005931, -0.00002714, 0.00001468, -0.00005903, 0.00006542, -0.00004282, 0.00001662, -0.00001511, 0.00002069, -0.00001022, -0.00001958, 0.00000741, 0.00001925, -0.00007288, 0.00007412, -0.00008895, 0.00017650, -0.00010046, 0.00010104, -0.00008884, 0.00002399, -0.00002694, 0.00001802, -0.00001955, 0.00001664, -0.00001330, 0.00000833, 0.00000963, },
		{ -0.00004674, 0.00004991, -0.00004286, 0.00001026, -0.00001280, -0.00001468, -0.00002131, 0.00005571, -0.00006949, 0.00001544, 0.00005591, -0.00003898, 0.00000465, -0.00002421, 0.00002629, -0.00003671, 0.00005092, -0.00009619, 0.00008320, -0.00004580, 0.00006449, -0.00013146, 0.00010941, -0.00014786, 0.00017864, -0.00026838, 0.00022741, -0.00092134, 0.00145009, -0.00000061, -0.00045147, 0.00016212, -0.00020748, 0.00012665, -0.00008738, 0.00002691, -0.00006150, 0.00006911, -0.00004600, -0.00000205, -0.00001344, 0.00003962, -0.00004906, 0.00003223, -0.00001909, 0.00002237, -0.00002875, -0.00000860, 0.00004720, -0.00008499, 0.00006961, -0.00007589, 0.00010027, 0.00000828, -0.00000908, 0.00003665, -0.00008087, 0.00005126, -0.00004878, 0.00003897, -0.00003816, 0.00004062, -0.00004440, 0.00005272, },
		{ -0.00001698, 0.00000922, 0.00000931, -0.00003365, 0.00001929, -0.00002485, -0.00001987, 0.00002436, 0.00000494, -0.00007565, 0.00011413, -0.00006682, 0.00004174, -0.00007138, 0.00007634, -0.00009137, 0.00010540, -0.00012732, 0.00009134, -0.00007501, 0.00012127, -0.00017155, 0.00013752, -0.00018456, 0.00021381, -0.00028018, 0.00027124, -0.00072937, 0.00045635, 0.00115199, -0.00093915, 0.00033762, -0.00032710, 0.00018465, -0.00013604, 0.00009268, -0.00013461, 0.00011292, -0.00007311, 0.00004435, -0.00008028, 0.00009373, -0.00007930, 0.00004927, -0.00003586, 0.00003621, -0.00002944, -0.00001139, 0.00003361, -0.00003520, -0.00000660, 0.00002074, -0.00004280, 0.00014112, -0.00011148, 0.00012165, -0.00012175, 0.00006538, -0.00005740, 0.00004363, -0.00004184, 0.00004172, -0.00003823, 0.00003313, },
		{ 0.00003357, -0.00003577, 0.00004333, -0.00004219, 0.00001101, -0.00000485, -0.00002581, -0.00000571, 0.00005201, -0.00009915, 0.00008511, -0.00001560, 0.00000507, -0.00003624, 0.00003160, -0.00003535, 0.00003043, -0.00002217, -0.00002408, 0.00001838, 0.00002504, -0.00003017, -0.00002833, -0.00000373, -0.00001034, 0.00001455, -0.00009035, -0.00005712, -0.00065665, 0.00176396, -0.00068433, -0.00001117, -0.00009283, -0.00001664, 0.00000388, 0.00000179, -0.00005704, 0.00002082, 0.00000376, 0.00000351, -0.00005017, 0.00004228, -0.00001461, -0.00000913, 0.00000739, 0.00000087, 0.00000348, -0.00002767, 0.00001810, 0.00001546, -0.00006961, 0.00008580, -0.00011182, 0.00015699, -0.00008242, 0.00006971, -0.00003799, -0.00001681, 0.00000912, -0.00001386, 0.00000830, -0.00000690, 0.00001335, -0.00002152, },
		{ 0.00004113, -0.00002952, 0.00002008, -0.00000048, -0.00002957, 0.00002544, -0.00002816, -0.00002364, 0.00005151, -0.00005090, -0.00000739, 0.00007399, -0.00006113, 0.00003356, -0.00005186, 0.00006088, -0.00008075, 0.00010112, -0.00013427, 0.00010737, -0.00008524, 0.00012186, -0.00018482, 0.00015903, -0.00020831, 0.00025441, -0.00035966, 0.00036229, -0.00102934, 0.00126129, 0.00039742, -0.00070929, 0.00029587, -0.00030784, 0.00019995, -0.00014696, 0.00008616, -0.00012383, 0.00011365, -0.00007563, 0.00003657, -0.00006484, 0.00008561, -0.00008527, 0.00006179, -0.00004423, 0.00004164, -0.00004611, 0.00001255, 0.00003219, -0.00006870, 0.00005950, -0.00005678, 0.00004092, 0.00005368, -0.00006047, 0.00008629, -0.00010518, 0.00006751, -0.00005851, 0.00004383, -0.00003962, 0.00004340, -0.00004366, },
		{ 0.00000017, 0.00001496, -0.00002623, 0.00003968, -0.00004586, 0.00001965, -0.00000185, -0.00004112, 0.00002788, 0.00001297, -0.00008042, 0.00011201, -0.00006898, 0.00004610, -0.00007099, 0.00007926, -0.00009393, 0.00009990, -0.00010274, 0.00006081, -0.00006215, 0.00011060, -0.00014169, 0.00009866, -0.00014727, 0.00017080, -0.00022054, 0.00019009, -0.00053100, 0.00001582, 0.00155633, -0.00104317, 0.00033013, -0.00032499, 0.00017859, -0.00013534, 0.00010129, -0.00014465, 0.00010753, -0.00006775, 0.00005717, -0.00009854, 0.00009813, -0.00007465, 0.00004277, -0.00003000, 0.00003012, -0.00002759, -0.00000765, 0.00003386, -0.00003014, -0.00001305, 0.00004254, -0.00008614, 0.00015227, -0.00011989, 0.00011419, -0.00009066, 0.00003682, -0.00002899, 0.00001647, -0.00001643, 0.00001900, -0.00001165, },
		{ -0.00003493, 0.00003786, -0.00003469, 0.00002659, -0.00000730, -0.00002756, 0.00004061, -0.00005039, 0.00000035, 0.00004684, -0.00008307, 0.00006211, 0.00000334, -0.00001675, -0.00000690, 0.00000195, 0.00000181, -0.00001922, 0.00004012, -0.00007956, 0.00005855, -0.00002738, 0.00004475, -0.00010654, 0.00007486, -0.00010354, 0.00013269, -0.00024113, 0.00018392, -0.00096302, 0.00181334, -0.00043271, -0.00023452, 0.00004183, -0.00012064, 0.00008174, -0.00005920, 0.00000132, -0.00004308, 0.00005241, -0.00002614, -0.00001566, -0.00000681, 0.00003367, -0.00004983, 0.00004123, -0.00002854, 0.00002568, -0.00004566, 0.00003953, 0.00000115, -0.00005765, 0.00008299, -0.00010788, 0.00011317, -0.00004071, 0.00001620, 0.00001982, -0.00005497, 0.00003972, -0.00003737, 0.00002635, -0.00002101, 0.00002840, },
		{ -0.00002341, 0.00001206, 0.00000307, -0.00002290, 0.00004401, -0.00006015, 0.00004705, -0.00002244, -0.00003482, 0.00005072, -0.00003709, -0.00002346, 0.00008451, -0.00007650, 0.00005435, -0.00007128, 0.00008628, -0.00011087, 0.00012883, -0.00014439, 0.00010832, -0.00010431, 0.00015149, -0.00020240, 0.00017097, -0.00022754, 0.00028020, -0.00038647, 0.00039971, -0.00097737, 0.00090207, 0.00086652, -0.00095285, 0.00040183, -0.00038589, 0.00025656, -0.00019554, 0.00013704, -0.00017417, 0.00014463, -0.00009850, 0.00007437, -0.00011083, 0.00011945, -0.00010745, 0.00007808, -0.00005832, 0.00005199, -0.00005770, 0.00002906, 0.00002152, -0.00005617, 0.00004652, -0.00003003, -0.00001488, 0.00008969, -0.00009511, 0.00010576, -0.00009918, 0.00005800, -0.00004470, 0.00002811, -0.00002319, 0.00002769, },
		{ 0.00001247, -0.00002515, 0.00003488, -0.00004433, 0.00004247, -0.00002731, -0.00000734, 0.00004099, -0.00006905, 0.00003765, 0.00000985, -0.00007179, 0.00009604, -0.00005736, 0.00003431, -0.00005321, 0.00006173, -0.00007073, 0.00006403, -0.00004970, 0.00000939, -0.00002749, 0.00006936, -0.00007772, 0.00002772, -0.00007542, 0.00008639, -0.00011059, 0.00004384, -0.00025981, -0.00043059, 0.00184717, -0.00100285, 0.00022156, -0.00025407, 0.00012044, -0.00009265, 0.00007176, -0.00011612, 0.00006823, -0.00003757, 0.00004701, -0.00008691, 0.00007028, -0.00004215, 0.00001356, -0.00000577, 0.00000832, -0.00001241, -0.00001741, 0.00004767, -0.00003776, -0.00001048, 0.00005203, -0.00010289, 0.00013426, -0.00009666, 0.00007350, -0.00003855, -0.00000317, 0.00000653, -0.00001344, 0.00000904, -0.00000349, },
		{ 0.00002389, -0.00002595, 0.00001925, -0.00000987, -0.00001494, 0.00004345, -0.00007080, 0.00008107, -0.00006258, 0.00000162, 0.00004012, -0.00006520, 0.00003827, 0.00002270, -0.00004054, 0.00002430, -0.00003073, 0.00004274, -0.00006990, 0.00009540, -0.00012105, 0.00008862, -0.00007469, 0.00010998, -0.00016787, 0.00013531, -0.00017758, 0.00022680, -0.00035821, 0.00036002, -0.00112191, 0.00165692, -0.00003781, -0.00051401, 0.00020062, -0.00024064, 0.00017297, -0.00013464, 0.00007488, -0.00011665, 0.00010567, -0.00006464, 0.00003419, -0.00006783, 0.00008625, -0.00009062, 0.00007375, -0.00005716, 0.00004772, -0.00006577, 0.00006386, -0.00001602, -0.00004308, 0.00007068, -0.00008523, 0.00005756, 0.00000720, -0.00003567, 0.00006088, -0.00007201, 0.00004993, -0.00004019, 0.00002492, -0.00001722, },
		{ 0.00000273, 0.00000465, -0.00002046, 0.00003503, -0.00005935, 0.00006919, -0.00006425, 0.00004085, 0.00000768, -0.00005703, 0.00005759, -0.00003433, -0.00002638, 0.00008112, -0.00007819, 0.00005939, -0.00007360, 0.00009133, -0.00011453, 0.00012300, -0.00011982, 0.00008126, -0.00009565, 0.00014425, -0.00017670, 0.00013916, -0.00019895, 0.00024844, -0.00034079, 0.00033604, -0.00078619, 0.00043783, 0.00131790, -0.00111037, 0.00043329, -0.00040880, 0.00027024, -0.00020999, 0.00015749, -0.00019227, 0.00014572, -0.00010036, 0.00009484, -0.00013280, 0.00012464, -0.00010291, 0.00007165, -0.00005370, 0.00004631, -0.00005621, 0.00003304, 0.00002297, -0.00005452, 0.00003896, -0.00000823, -0.00005181, 0.00010029, -0.00009826, 0.00008819, -0.00006405, 0.00002711, -0.00001367, -0.00000020, 0.00000271, },
		{ -0.00001277, 0.00002086, -0.00003041, 0.00003186, -0.00003539, 0.00001290, 0.00001834, -0.00005246, 0.00009213, -0.00009395, 0.00004841, -0.00000055, -0.00005453, 0.00007140, -0.00003699, 0.00001149, -0.00002317, 0.00002923, -0.00002963, 0.00001008, 0.00001526, -0.00004642, 0.00001476, 0.00001324, 0.00000038, -0.00005257, 0.00000670, -0.00001172, 0.00001553, -0.00012044, 0.00001262, -0.00079187, 0.00195312, -0.00079187, 0.00001262, -0.00012044, 0.00001553, -0.00001172, 0.00000670, -0.00005257, 0.00000038, 0.00001324, 0.00001476, -0.00004642, 0.00001526, 0.00001008, -0.00002963, 0.00002923, -0.00002317, 0.00001149, -0.00003699, 0.00007140, -0.00005453, -0.00000055, 0.00004841, -0.00009395, 0.00009213, -0.00005246, 0.00001834, 0.00001290, -0.00003539, 0.00003186, -0.00003041, 0.00002086, },
		{ 0.00000273, 0.00000271, -0.00000020, -0.00001367, 0.00002711, -0.00006405, 0.00008819, -0.00009826, 0.00010029, -0.00005181, -0.00000823, 0.00003896, -0.00005452, 0.00002297, 0.00003304, -0.00005621, 0.00004631, -0.00005370, 0.00007165, -0.00010291, 0.00012464, -0.00013280, 0.00009484, -0.00010036, 0.00014572, -0.00019227, 0.00015749, -0.00020999, 0.00027024, -0.00040880, 0.00043329, -0.00111037, 0.00131790, 0.00043783, -0.00078619, 0.00033604, -0.00034079, 0.00024844, -0.00019895, 0.00013916, -0.00017670, 0.00014425, -0.00009565, 0.00008126, -0.00011982, 0.00012300, -0.00011453, 0.00009133, -0.00007360, 0.00005939, -0.00007819, 0.00008112, -0.00002638, -0.00003433, 0.00005759, -0.00005703, 0.00000768, 0.00004085, -0.00006425, 0.00006919, -0.00005935, 0.00003503, -0.00002046, 0.00000465, },
		{ 0.00002389, -0.00001722, 0.00002492, -0.00004019, 0.00004993, -0.00007201, 0.00006088, -0.00003567, 0.00000720, 0.00005756, -0.00008523, 0.00007068, -0.00004308, -0.00001602, 0.00006386, -0.00006577, 0.00004772, -0.00005716, 0.00007375, -0.00009062, 0.00008625, -0.00006783, 0.00003419, -0.00006464, 0.00010567, -0.00011665, 0.00007488, -0.00013464, 0.00017297, -0.00024064, 0.00020062, -0.00051401, -0.00003781, 0.00165692, -0.00112191, 0.00036002, -0.00035821, 0.00022680, -0.00017758, 0.00013531, -0.00016787, 0.00010998, -0.00007469, 0.00008862, -0.00012105, 0.00009540, -0.00006990, 0.00004274, -0.00003073, 0.00002430, -0.00004054, 0.00002270, 0.00003827, -0.00006520, 0.00004012, 0.00000162, -0.00006258, 0.00008107, -0.00007080, 0.00004345, -0.00001494, -0.00000987, 0.00001925, -0.00002595, },
		{ 0.00001247, -0.00000349, 0.00000904, -0.00001344, 0.00000653, -0.00000317, -0.00003855, 0.00007350, -0.00009666, 0.00013426, -0.00010289, 0.00005203, -0.00001048, -0.00003776, 0.00004767, -0.00001741, -0.00001241, 0.00000832, -0.00000577, 0.00001356, -0.00004215, 0.00007028, -0.00008691, 0.00004701, -0.00003757, 0.00006823, -0.00011612, 0.00007176, -0.00009265, 0.00012044, -0.00025407, 0.00022156, -0.00100285, 0.00184717, -0.00043059, -0.00025981, 0.00004384, -0.00011059, 0.00008639, -0.00007542, 0.00002772, -0.00007772, 0.00006936, -0.00002749, 0.00000939, -0.00004970, 0.00006403, -0.00007073, 0.00006173, -0.00005321, 0.00003431, -0.00005736, 0.00009604, -0.00007179, 0.00000985, 0.00003765, -0.00006905, 0.00004099, -0.00000734, -0.00002731, 0.00004247, -0.00004433, 0.00003488, -0.00002515, },
		{ -0.00002341, 0.00002769, -0.00002319, 0.00002811, -0.00004470, 0.00005800, -0.00009918, 0.00010576, -0.00009511, 0.00008969, -0.00001488, -0.00003003, 0.00004652, -0.00005617, 0.00002152, 0.00002906, -0.00005770, 0.00005199, -0.00005832, 0.00007808, -0.00010745, 0.00011945, -0.00011083, 0.00007437, -0.00009850, 0.00014463, -0.00017417, 0.00013704, -0.00019554, 0.00025656, -0.00038589, 0.00040183, -0.00095285, 0.00086652, 0.00090207, -0.00097737, 0.00039971, -0.00038647, 0.00028020, -0.00022754, 0.00017097, -0.00020240, 0.00015149, -0.00010431, 0.00010832, -0.00014439, 0.00012883, -0.00011087, 0.00008628, -0.00007128, 0.00005435, -0.00007650, 0.00008451, -0.00002346, -0.00003709, 0.00005072, -0.00003482, -0.00002244, 0.00004705, -0.00006015, 0.00004401, -0.00002290, 0.00000307, 0.00001206, },
		{ -0.00003493, 0.00002840, -0.00002101, 0.00002635, -0.00003737, 0.00003972, -0.00005497, 0.00001982, 0.00001620, -0.00004071, 0.00011317, -0.00010788, 0.00008299, -0.00005765, 0.00000115, 0.00003953, -0.00004566, 0.00002568, -0.00002854, 0.00004123, -0.00004983, 0.00003367, -0.00000681, -0.00001566, -0.00002614, 0.00005241, -0.00004308, 0.00000132, -0.00005920, 0.00008174, -0.00012064, 0.00004183, -0.00023452, -0.00043271, 0.00181334, -0.00096302, 0.00018392, -0.00024113, 0.00013269, -0.00010354, 0.00007486, -0.00010654, 0.00004475, -0.00002738, 0.00005855, -0.00007956, 0.00004012, -0.00001922, 0.00000181, 0.00000195, -0.00000690, -0.00001675, 0.00000334, 0.00006211, -0.00008307, 0.00004684, 0.00000035, -0.00005039, 0.00004061, -0.00002756, -0.00000730, 0.00002659, -0.00003469, 0.00003786, },
		{ 0.00000017, -0.00001165, 0.00001900, -0.00001643, 0.00001647, -0.00002899, 0.00003682, -0.00009066, 0.00011419, -0.00011989, 0.00015227, -0.00008614, 0.00004254, -0.00001305, -0.00003014, 0.00003386, -0.00000765, -0.00002759, 0.00003012, -0.00003000, 0.00004277, -0.00007465, 0.00009813, -0.00009854, 0.00005717, -0.00006775, 0.00010753, -0.00014465, 0.00010129, -0.00013534, 0.00017859, -0.00032499, 0.00033013, -0.00104317, 0.00155633, 0.00001582, -0.00053100, 0.00019009, -0.00022054, 0.00017080, -0.00014727, 0.00009866, -0.00014169, 0.00011060, -0.00006215, 0.00006081, -0.00010274, 0.00009990, -0.00009393, 0.00007926, -0.00007099, 0.00004610, -0.00006898, 0.00011201, -0.00008042, 0.00001297, 0.00002788, -0.00004112, -0.00000185, 0.00001965, -0.00004586, 0.00003968, -0.00002623, 0.00001496, },
		{ 0.00004113, -0.00004366, 0.00004340, -0.00003962, 0.00004383, -0.00005851, 0.00006751, -0.00010518, 0.00008629, -0.00006047, 0.00005368, 0.00004092, -0.00005678, 0.00005950, -0.00006870, 0.00003219, 0.00001255, -0.00004611, 0.00004164, -0.00004423, 0.00006179, -0.00008527, 0.00008561, -0.00006484, 0.00003657, -0.00007563, 0.00011365, -0.00012383, 0.00008616, -0.00014696, 0.00019995, -0.00030784, 0.00029587, -0.00070929, 0.00039742, 0.00126129, -0.00102934, 0.00036229, -0.00035966, 0.00025441, -0.00020831, 0.00015903, -0.00018482, 0.00012186, -0.00008524, 0.00010737, -0.00013427, 0.00010112, -0.00008075, 0.00006088, -0.00005186, 0.00003356, -0.00006113, 0.00007399, -0.00000739, -0.00005090, 0.00005151, -0.00002364, -0.00002816, 0.00002544, -0.00002957, -0.00000048, 0.00002008, -0.00002952, },
		{ 0.00003357, -0.00002152, 0.00001335, -0.00000690, 0.00000830, -0.00001386, 0.00000912, -0.00001681, -0.00003799, 0.00006971, -0.00008242, 0.00015699, -0.00011182, 0.00008580, -0.00006961, 0.00001546, 0.00001810, -0.00002767, 0.00000348, 0.00000087, 0.00000739, -0.00000913, -0.00001461, 0.00004228, -0.00005017, 0.00000351, 0.00000376, 0.00002082, -0.00005704, 0.00000179, 0.00000388, -0.00001664, -0.00009283, -0.00001117, -0.00068433, 0.00176396, -0.00065665, -0.00005712, -0.00009035, 0.00001455, -0.00001034, -0.00000373, -0.00002833, -0.00003017, 0.00002504, 0.00001838, -0.00002408, -0.00002217, 0.00003043, -0.00003535, 0.00003160, -0.00003624, 0.00000507, -0.00001560, 0.00008511, -0.00009915, 0.00005201, -0.00000571, -0.00002581, -0.00000485, 0.00001101, -0.00004219, 0.00004333, -0.00003577, },
		{ -0.00001698, 0.00003313, -0.00003823, 0.00004172, -0.00004184, 0.00004363, -0.00005740, 0.00006538, -0.00012175, 0.00012165, -0.00011148, 0.00014112, -0.00004280, 0.00002074, -0.00000660, -0.00003520, 0.00003361, -0.00001139, -0.00002944, 0.00003621, -0.00003586, 0.00004927, -0.00007930, 0.00009373, -0.00008028, 0.00004435, -0.00007311, 0.00011292, -0.00013461, 0.00009268, -0.00013604, 0.00018465, -0.00032710, 0.00033762, -0.00093915, 0.00115199, 0.00045635, -0.00072937, 0.00027124, -0.00028018, 0.00021381, -0.00018456, 0.00013752, -0.00017155, 0.00012127, -0.00007501, 0.00009134, -0.00012732, 0.00010540, -0.00009137, 0.00007634, -0.00007138, 0.00004174, -0.00006682, 0.00011413, -0.00007565, 0.00000494, 0.00002436, -0.00001987, -0.00002485, 0.00001929, -0.00003365, 0.00000931, 0.00000922, },
		{ -0.00004674, 0.00005272, -0.00004440, 0.00004062, -0.00003816, 0.00003897, -0.00004878, 0.00005126, -0.00008087, 0.00003665, -0.00000908, 0.00000828, 0.00010027, -0.00007589, 0.00006961, -0.00008499, 0.00004720, -0.00000860, -0.00002875, 0.00002237, -0.00001909, 0.00003223, -0.00004906, 0.00003962, -0.00001344, -0.00000205, -0.00004600, 0.00006911, -0.00006150, 0.00002691, -0.00008738, 0.00012665, -0.00020748, 0.00016212, -0.00045147, -0.00000061, 0.00145009, -0.00092134, 0.00022741, -0.00026838, 0.00017864, -0.00014786, 0.00010941, -0.00013146, 0.00006449, -0.00004580, 0.00008320, -0.00009619, 0.00005092, -0.00003671, 0.00002629, -0.00002421, 0.00000465, -0.00003898, 0.00005591, 0.00001544, -0.00006949, 0.00005571, -0.00002131, -0.00001468, -0.00001280, 0.00001026, -0.00004286, 0.00004991, },
		{ -0.00001394, 0.00000963, 0.00000833, -0.00001330, 0.00001664, -0.00001955, 0.00001802, -0.00002694, 0.00002399, -0.00008884, 0.00010104, -0.00010046, 0.00017650, -0.00008895, 0.00007412, -0.00007288, 0.00001925, 0.00000741, -0.00001958, -0.00001022, 0.00002069, -0.00001511, 0.00001662, -0.00004282, 0.00006542, -0.00005903, 0.00001468, -0.00002714, 0.00005931, -0.00008473, 0.00003263, -0.00004218, 0.00004828, -0.00017363, 0.00012174, -0.00077556, 0.00153829, -0.00026773, -0.00030028, 0.00004696, -0.00009093, 0.00007136, -0.00007320, 0.00004075, -0.00009081, 0.00006292, -0.00001444, 0.00002568, -0.00007059, 0.00006137, -0.00005542, 0.00004783, -0.00005424, 0.00001595, -0.00002564, 0.00009887, -0.00010558, 0.00004916, -0.00000958, -0.00000064, -0.00003965, 0.00002898, -0.00004886, 0.00003056, },
		{ 0.00004429, -0.00004466, 0.00005781, -0.00005244, 0.00005152, -0.00005423, 0.00005452, -0.00006548, 0.00006962, -0.00012033, 0.00009184, -0.00007647, 0.00010889, 0.00001785, -0.00000489, 0.00000394, -0.00004960, 0.00004337, -0.00002515, -0.00002049, 0.00002813, -0.00002451, 0.00003511, -0.00006067, 0.00006545, -0.00004339, 0.00001842, -0.00006060, 0.00009206, -0.00009682, 0.00005818, -0.00010710, 0.00015206, -0.00027770, 0.00027308, -0.00074901, 0.00072512, 0.00080323, -0.00080196, 0.00026109, -0.00027347, 0.00020326, -0.00017690, 0.00013499, -0.00016059, 0.00009799, -0.00006251, 0.00009518, -0.00011931, 0.00008130, -0.00006688, 0.00005693, -0.00005701, 0.00002328, -0.00005268, 0.00010391, -0.00005928, -0.00001209, 0.00002678, -0.00000822, -0.00002628, -0.00000494, -0.00000109, -0.00003188, },
		{ 0.00005436, -0.00004136, 0.00004216, -0.00002402, 0.00002032, -0.00002074, 0.00001671, -0.00002127, 0.00001792, -0.00003931, -0.00002258, 0.00003524, -0.00002688, 0.00014715, -0.00007507, 0.00006833, -0.00009659, 0.00005707, -0.00002482, -0.00001483, 0.00000381, 0.00000591, 0.00000310, -0.00001544, 0.00000023, 0.00002498, -0.00002635, -0.00002345, 0.00002805, -0.00000766, -0.00001937, -0.00003900, 0.00006225, -0.00011696, 0.00004416, -0.00023830, -0.00027068, 0.00145045, -0.00067849, 0.00003263, -0.00014472, 0.00007936, -0.00006892, 0.00004273, -0.00006295, -0.00000049, -0.00000272, 0.00005017, -0.00004689, -0.00000253, 0.00000361, -0.00000345, 0.00000052, -0.00002225, -0.00001965, 0.00003924, 0.00003615, -0.00008442, 0.00005665, -0.00002166, 0.00000739, -0.00005165, 0.00004041, -0.00006507, },
		{ -0.00000247, 0.00001856, -0.00002122, 0.00004058, -0.00003718, 0.00003591, -0.00004293, 0.00004241, -0.00005141, 0.00004876, -0.00011294, 0.00009605, -0.00008992, 0.00017075, -0.00004069, 0.00005031, -0.00006745, 0.00001097, 0.00000905, -0.00002306, -0.00001262, 0.00002648, -0.00002044, 0.00002130, -0.00004629, 0.00006113, -0.00004408, 0.00000840, -0.00003787, 0.00006914, -0.00008037, 0.00003254, -0.00005426, 0.00007005, -0.00019552, 0.00016443, -0.00073298, 0.00120547, 0.00011818, -0.00047991, 0.00012795, -0.00015246, 0.00011613, -0.00011133, 0.00008014, -0.00011959, 0.00007291, -0.00002767, 0.00005545, -0.00009186, 0.00006518, -0.00005400, 0.00004762, -0.00005741, 0.00001233, -0.00002345, 0.00010005, -0.00009953, 0.00003637, -0.00000795, 0.00001773, -0.00005546, 0.00002157, -0.00002907, },
		{ -0.00006415, 0.00006260, -0.00005529, 0.00006316, -0.00004588, 0.00004256, -0.00004858, 0.00004482, -0.00005082, 0.00004938, -0.00009052, 0.00003778, -0.00003390, 0.00007261, 0.00008088, -0.00002189, 0.00000972, -0.00006535, 0.00005460, -0.00004054, -0.00000833, 0.00001320, -0.00000420, 0.00001077, -0.00003233, 0.00002975, -0.00000515, -0.00000611, -0.00004273, 0.00005976, -0.00004959, 0.00001734, -0.00006870, 0.00010370, -0.00020586, 0.00017879, -0.00053882, 0.00035676, 0.00099929, -0.00073348, 0.00016569, -0.00020965, 0.00014794, -0.00013232, 0.00009869, -0.00011803, 0.00005137, -0.00003327, 0.00007883, -0.00008773, 0.00004041, -0.00003345, 0.00003167, -0.00003611, -0.00000184, -0.00003372, 0.00008804, -0.00003806, -0.00003147, 0.00003037, -0.00000307, -0.00001272, -0.00004033, 0.00003400, },
		{ -0.00006049, 0.00003417, -0.00001816, 0.00001597, 0.00000841, -0.00000682, 0.00000182, -0.00001102, 0.00000975, -0.00001592, -0.00000063, -0.00006722, 0.00005143, -0.00003925, 0.00017309, -0.00004857, 0.00005206, -0.00009876, 0.00005581, -0.00003015, -0.00001018, -0.00000721, 0.00002198, -0.00001461, 0.00000324, -0.00002040, 0.00004059, -0.00002988, -0.00001460, 0.00000071, 0.00002529, -0.00004087, -0.00001384, 0.00002146, -0.00005566, -0.00003218, -0.00009978, -0.00039972, 0.00129329, -0.00036292, -0.00016393, -0.00003195, -0.00000950, -0.00000002, -0.00001559, -0.00000355, -0.00005111, 0.00002633, 0.00002417, -0.00000440, -0.00004112, 0.00002619, -0.00001861, 0.00001475, -0.00003958, -0.00001045, 0.00003077, 0.00004810, -0.00008959, 0.00004960, -0.00001933, 0.00002813, -0.00007819, 0.00004916, },
		{ 0.00000400, -0.00003830, 0.00004511, -0.00004409, 0.00006073, -0.00004535, 0.00003997, -0.00005120, 0.00004852, -0.00005423, 0.00004824, -0.00010396, 0.00005770, -0.00006187, 0.00015060, 0.00002290, 0.00002342, -0.00005942, -0.00000447, 0.00001786, -0.00003319, -0.00000744, 0.00002090, -0.00001127, 0.00000913, -0.00003193, 0.00003952, -0.00001710, -0.00000580, -0.00003517, 0.00005804, -0.00005456, 0.00001309, -0.00004358, 0.00006054, -0.00017376, 0.00014308, -0.00061009, 0.00084857, 0.00042328, -0.00055133, 0.00013165, -0.00015748, 0.00011477, -0.00011058, 0.00008334, -0.00011276, 0.00005436, -0.00002016, 0.00006225, -0.00008545, 0.00004596, -0.00003706, 0.00003593, -0.00004890, -0.00000297, -0.00001196, 0.00009140, -0.00008399, 0.00001702, -0.00000252, 0.00002826, -0.00005318, -0.00000454, },
		{ 0.00005573, -0.00007444, 0.00005730, -0.00004600, 0.00004843, -0.00002087, 0.00001951, -0.00003022, 0.00002115, -0.00002259, 0.00001670, -0.00004912, -0.00001686, -0.00000730, 0.00004850, 0.00013378, -0.00002007, 0.00000323, -0.00007478, 0.00005883, -0.00004924, -0.00000071, -0.00000031, 0.00001495, -0.00001104, -0.00000911, 0.00000244, 0.00002000, -0.00001810, -0.00002999, 0.00002983, -0.00000964, -0.00001279, -0.00003843, 0.00006017, -0.00013829, 0.00009211, -0.00035994, 0.00009540, 0.00103300, -0.00055092, 0.00002122, -0.00011906, 0.00007310, -0.00007272, 0.00004869, -0.00006402, 0.00000076, -0.00000317, 0.00005624, -0.00004895, 0.00000051, -0.00000612, 0.00001164, -0.00001736, -0.00002507, -0.00001826, 0.00007426, -0.00001960, -0.00004596, 0.00002953, 0.00000096, 0.00000601, -0.00007242, },
		{ 0.00003605, -0.00003539, 0.00000149, 0.00000952, -0.00001252, 0.00003841, -0.00002653, 0.00001638, -0.00003072, 0.00002959, -0.00003489, 0.00001873, -0.00008152, 0.00003223, -0.00003109, 0.00018147, 0.00000026, 0.00002437, -0.00009304, 0.00004386, -0.00002519, -0.00001432, -0.00000983, 0.00002652, -0.00001734, 0.00000405, -0.00002127, 0.00003544, -0.00001701, -0.00001677, -0.00001219, 0.00003621, -0.00003832, -0.00001084, 0.00000483, -0.00002615, -0.00006322, -0.00003481, -0.00041333, 0.00104267, -0.00005123, -0.00030559, 0.00003372, -0.00006152, 0.00003747, -0.00004701, 0.00002979, -0.00007351, 0.00003153, 0.00001436, 0.00002048, -0.00005603, 0.00002716, -0.00001839, 0.00001780, -0.00004544, -0.00001331, 0.00003198, 0.00004987, -0.00008404, 0.00003443, -0.00001292, 0.00004258, -0.00008750, },
		{ -0.00003648, 0.00003373, -0.00006053, 0.00005367, -0.00005004, 0.00006139, -0.00003329, 0.00002685, -0.00004311, 0.00003601, -0.00003720, 0.00002680, -0.00007147, 0.00000468, -0.00003799, 0.00013181, 0.00008931, 0.00000440, -0.00005715, -0.00001909, 0.00002538, -0.00004187, -0.00000177, 0.00001081, 0.00000420, -0.00000915, -0.00001299, 0.00001561, 0.00000726, -0.00001646, -0.00002898, 0.00003841, -0.00002272, -0.00000968, -0.00002635, 0.00003810, -0.00013220, 0.00009346, -0.00046406, 0.00053721, 0.00060154, -0.00050627, 0.00006713, -0.00011680, 0.00007748, -0.00008037, 0.00005951, -0.00008105, 0.00001903, -0.00000146, 0.00005396, -0.00006199, 0.00001712, -0.00001681, 0.00002183, -0.00003550, -0.00002326, 0.00000211, 0.00007923, -0.00006535, -0.00000258, 0.00000209, 0.00003443, -0.00004003, },
		{ -0.00009167, 0.00005708, -0.00006162, 0.00003365, -0.00002448, 0.00002248, 0.00001132, -0.00000527, -0.00001121, -0.00000351, 0.00000422, -0.00001152, -0.00001617, -0.00004986, -0.00001212, 0.00004175, 0.00017346, 0.00000342, -0.00001740, -0.00007483, 0.00005216, -0.00004767, -0.00000102, -0.00000785, 0.00002627, -0.00002185, 0.00000017, -0.00000878, 0.00002691, -0.00001516, -0.00002573, 0.00000921, 0.00001448, -0.00002453, -0.00002411, 0.00003182, -0.00008969, 0.00003356, -0.00023660, -0.00005062, 0.00093583, -0.00031160, -0.00012097, -0.00003923, 0.00000824, -0.00002294, 0.00000711, -0.00001959, -0.00003524, 0.00001339, 0.00004050, -0.00001786, -0.00002461, 0.00000598, 0.00000229, -0.00000522, -0.00004112, -0.00001153, 0.00006721, -0.00000837, -0.00005159, 0.00002146, 0.00000737, 0.00002233, },
		{ -0.00008252, 0.00000984, -0.00000397, -0.00002735, 0.00002764, -0.00003095, 0.00005402, -0.00002890, 0.00001524, -0.00003510, 0.00003119, -0.00003310, 0.00001432, -0.00006561, -0.00001159, -0.00001910, 0.00018457, 0.00006211, -0.00000626, -0.00008580, 0.00002703, -0.00001610, -0.00002143, -0.00000846, 0.00002290, -0.00000901, -0.00000699, -0.00001095, 0.00002033, 0.00000102, -0.00002158, -0.00001661, 0.00003235, -0.00002152, -0.00001969, 0.00000277, -0.00001862, -0.00006161, -0.00002117, -0.00035798, 0.00077041, 0.00019203, -0.00035835, 0.00003814, -0.00006890, 0.00003840, -0.00004770, 0.00003418, -0.00006746, 0.00001535, 0.00001906, 0.00002800, -0.00005040, 0.00001351, -0.00001016, 0.00001490, -0.00004300, -0.00002500, 0.00003927, 0.00004491, -0.00007147, 0.00001520, -0.00000493, 0.00005115, },
		{ -0.00002425, -0.00006233, 0.00004835, -0.00006164, 0.00004266, -0.00004021, 0.00004531, -0.00000685, 0.00000440, -0.00002686, 0.00001430, -0.00001212, -0.00000104, -0.00003486, -0.00003824, -0.00003869, 0.00012429, 0.00015163, -0.00000001, -0.00006622, -0.00002675, 0.00002488, -0.00004281, -0.00000119, 0.00000245, 0.00001761, -0.00002283, -0.00000132, 0.00000101, 0.00001939, -0.00001697, -0.00002593, 0.00002019, 0.00000289, -0.00002368, -0.00001486, 0.00001757, -0.00009102, 0.00004495, -0.00033644, 0.00030938, 0.00064879, -0.00037320, -0.00003153, -0.00005823, 0.00002745, -0.00004094, 0.00002722, -0.00004312, -0.00001549, 0.00001434, 0.00004315, -0.00003618, -0.00000655, -0.00000428, 0.00001248, -0.00002286, -0.00004216, 0.00001232, 0.00006940, -0.00004934, -0.00001701, 0.00000185, 0.00004028, },
		{ 0.00003350, -0.00009630, 0.00004127, -0.00003513, 0.00000502, -0.00000364, -0.00000204, 0.00003706, -0.00001898, -0.00000282, -0.00001777, 0.00001758, -0.00002314, -0.00000412, -0.00005170, -0.00004662, 0.00004188, 0.00020758, 0.00004390, -0.00004792, -0.00006821, 0.00003677, -0.00003842, -0.00000678, -0.00001040, 0.00002900, -0.00002054, -0.00000414, -0.00000646, 0.00002060, -0.00000349, -0.00002614, -0.00000285, 0.00002354, -0.00002050, -0.00002296, 0.00001754, -0.00006099, 0.00000513, -0.00016666, -0.00010506, 0.00076461, -0.00008057, -0.00021542, 0.00000202, -0.00002660, 0.00000113, -0.00001249, 0.00000306, -0.00004747, 0.00001092, 0.00003676, -0.00000084, -0.00003126, 0.00000365, 0.00000107, 0.00000163, -0.00004945, -0.00001393, 0.00006666, -0.00000410, -0.00004901, 0.00000757, 0.00001599, },
		{ 0.00005696, -0.00007018, -0.00001772, 0.00001987, -0.00004057, 0.00002916, -0.00003455, 0.00005178, -0.00001354, 0.00000048, -0.00002683, 0.00001833, -0.00001661, -0.00000497, -0.00003439, -0.00005660, -0.00002550, 0.00019427, 0.00012913, -0.00003211, -0.00008374, 0.00001230, -0.00001038, -0.00002456, -0.00000889, 0.00001686, 0.00000290, -0.00001984, -0.00000116, 0.00000752, 0.00001328, -0.00002126, -0.00001937, 0.00002307, -0.00000234, -0.00002842, 0.00000357, -0.00001977, -0.00004524, -0.00003096, -0.00027979, 0.00053216, 0.00033381, -0.00032176, -0.00000621, -0.00004426, 0.00001441, -0.00002843, 0.00002005, -0.00004487, -0.00000974, 0.00002821, 0.00002693, -0.00003556, -0.00000205, -0.00000404, 0.00001253, -0.00003693, -0.00004014, 0.00004696, 0.00003848, -0.00005716, -0.00000287, 0.00000081, },
		{ 0.00004741, -0.00001091, -0.00007623, 0.00004550, -0.00004528, 0.00002058, -0.00002392, 0.00002264, 0.00002162, -0.00001455, -0.00001449, -0.00000391, 0.00000678, -0.00001963, -0.00001220, -0.00005338, -0.00007082, 0.00012355, 0.00021342, 0.00000968, -0.00008653, -0.00002635, 0.00001482, -0.00003507, -0.00000636, -0.00000218, 0.00002478, -0.00002654, -0.00000156, -0.00000135, 0.00001881, -0.00000871, -0.00002634, 0.00000704, 0.00001760, -0.00002539, -0.00001262, 0.00000468, -0.00005974, 0.00001195, -0.00024463, 0.00016823, 0.00059629, -0.00020322, -0.00012122, -0.00001234, -0.00001140, -0.00001222, 0.00000413, -0.00001570, -0.00003503, 0.00001703, 0.00003904, -0.00001856, -0.00001669, -0.00000312, 0.00000885, -0.00001232, -0.00005661, 0.00001565, 0.00006422, -0.00003814, -0.00002447, -0.00000404, },
		{ 0.00002414, 0.00004074, -0.00009044, 0.00001775, -0.00000838, -0.00001517, 0.00000632, -0.00001626, 0.00004788, -0.00001561, -0.00000863, -0.00001855, 0.00001503, -0.00001732, -0.00001159, -0.00003005, -0.00009172, 0.00002725, 0.00024818, 0.00009455, -0.00008193, -0.00006045, 0.00001841, -0.00002785, -0.00001207, -0.00001234, 0.00002669, -0.00001181, -0.00001486, 0.00000022, 0.00001138, 0.00000723, -0.00002464, -0.00001103, 0.00002375, -0.00000920, -0.00002636, 0.00000969, -0.00004460, -0.00000303, -0.00013246, -0.00010594, 0.00057771, 0.00009286, -0.00024002, -0.00000176, -0.00002918, -0.00000100, -0.00001045, 0.00000497, -0.00003947, -0.00000490, 0.00004049, 0.00000565, -0.00002592, -0.00000406, 0.00000025, 0.00000778, -0.00005261, -0.00002227, 0.00006884, -0.00000321, -0.00004194, -0.00000816, },
		{ 0.00000224, 0.00006234, -0.00005653, -0.00003812, 0.00002920, -0.00003625, 0.00001716, -0.00002779, 0.00003745, 0.00001079, -0.00001757, -0.00001618, 0.00000233, 0.00000115, -0.00002318, -0.00000841, -0.00007923, -0.00006456, 0.00021123, 0.00020219, -0.00005119, -0.00008950, 0.00000341, -0.00001225, -0.00002016, -0.00001388, 0.00001202, 0.00001240, -0.00002662, 0.00000024, 0.00000368, 0.00001536, -0.00001382, -0.00002324, 0.00001430, 0.00001158, -0.00003002, 0.00000036, -0.00002075, -0.00002734, -0.00004361, -0.00020913, 0.00035420, 0.00037757, -0.00022533, -0.00006798, -0.00001231, -0.00001393, -0.00000703, 0.00000376, -0.00002210, -0.00002875, 0.00003031, 0.00002755, -0.00002337, -0.00000899, -0.00000612, 0.00001344, -0.00002959, -0.00005473, 0.00005084, 0.00003414, -0.00004456, -0.00001657, },
		{ -0.00001319, 0.00005455, -0.00000094, -0.00007892, 0.00003095, -0.00002233, 0.00000013, -0.00001210, 0.00000402, 0.00004068, -0.00001996, -0.00001409, -0.00001068, 0.00001146, -0.00002151, -0.00001014, -0.00003961, -0.00012179, 0.00011007, 0.00028635, 0.00002818, -0.00011395, -0.00002111, -0.00000144, -0.00002275, -0.00001329, -0.00000543, 0.00002649, -0.00002143, -0.00001039, 0.00000307, 0.00001257, 0.00000121, -0.00002587, -0.00000280, 0.00002363, -0.00001885, -0.00001550, -0.00000312, -0.00003762, -0.00000548, -0.00018536, 0.00009241, 0.00049222, -0.00004823, -0.00016875, 0.00000231, -0.00002629, -0.00000388, -0.00000184, -0.00000557, -0.00003567, 0.00000588, 0.00004239, -0.00001059, -0.00001491, -0.00000819, 0.00000577, -0.00000127, -0.00006699, 0.00001295, 0.00006207, -0.00003017, -0.00002681, },
		{ -0.00002217, 0.00002943, 0.00004612, -0.00007975, -0.00000412, 0.00000880, -0.00002060, 0.00000359, -0.00001968, 0.00004440, 0.00000098, -0.00002252, -0.00001243, 0.00000434, -0.00000419, -0.00002531, -0.00000475, -0.00012055, -0.00002259, 0.00030059, 0.00015457, -0.00011664, -0.00005517, 0.00000187, -0.00002141, -0.00001204, -0.00001729, 0.00002317, -0.00000167, -0.00002360, 0.00000223, 0.00000767, 0.00001054, -0.00001756, -0.00001877, 0.00002116, 0.00000135, -0.00002660, 0.00000114, -0.00003233, -0.00000273, -0.00011446, -0.00008722, 0.00041552, 0.00018926, -0.00020221, -0.00003477, -0.00001384, -0.00001635, 0.00000137, -0.00000245, -0.00002380, -0.00002142, 0.00004162, 0.00001059, -0.00001960, -0.00000642, -0.00000709, 0.00001641, -0.00005272, -0.00003290, 0.00006961, -0.00000208, -0.00003400, },
		{ -0.00002614, 0.00000035, 0.00006715, -0.00004418, -0.00004898, 0.00002572, -0.00002157, 0.00000187, -0.00002026, 0.00002119, 0.00003190, -0.00002692, -0.00001391, -0.00000570, 0.00000808, -0.00002748, -0.00000180, -0.00006892, -0.00013178, 0.00021925, 0.00029256, -0.00006787, -0.00010083, -0.00000059, -0.00002084, -0.00000996, -0.00002139, 0.00000788, 0.00001789, -0.00002528, -0.00000710, 0.00000711, 0.00001088, -0.00000367, -0.00002601, 0.00000676, 0.00001907, -0.00002469, -0.00000648, -0.00002012, -0.00001235, -0.00005108, -0.00015620, 0.00023450, 0.00035358, -0.00011285, -0.00011301, 0.00000396, -0.00002897, 0.00000212, -0.00000218, -0.00001085, -0.00003256, 0.00001990, 0.00003458, -0.00001920, -0.00000496, -0.00001449, 0.00001444, -0.00001967, -0.00006771, 0.00005002, 0.00003196, -0.00003367, },
		},
		{ { 0.00001689, 0.00001141, -0.00003608, -0.00000466, 0.00003981, -0.00000825, -0.00002701, -0.00000569, 0.00000719, 0.00001730, -0.00008455, 0.00000688, 0.00000305, -0.00001345, -0.00002783, -0.00002587, -0.00012414, 0.00004013, 0.00037905, 0.00005895, -0.00013506, -0.00001814, -0.00001879, -0.00001923, -0.00001132, -0.00001088, 0.00004573, -0.00000905, -0.00002796, -0.00000846, 0.00001744, -0.00000335, -0.00000068, -0.00000335, 0.00001744, -0.00000846, -0.00002796, -0.00000905, 0.00004573, -0.00001088, -0.00001132, -0.00001923, -0.00001879, -0.00001814, -0.00013506, 0.00005895, 0.00037905, 0.00004013, -0.00012414, -0.00002587, -0.00002783, -0.00001345, 0.00000305, 0.00000688, -0.00008455, 0.00001730, 0.00000719, -0.00000569, -0.00002701, -0.00000825, 0.00003981, -0.00000466, -0.00003608, 0.00001141, },
		{ 0.00001782, 0.00001808, -0.00001612, -0.00003142, 0.00003215, 0.00001744, -0.00002957, -0.00001266, -0.00000371, 0.00003190, -0.00005044, -0.00004974, 0.00002908, -0.00001804, -0.00001562, -0.00003189, -0.00006331, -0.00012947, 0.00031879, 0.00026820, -0.00013510, -0.00006533, -0.00000516, -0.00003060, -0.00000403, -0.00002540, 0.00003145, 0.00002115, -0.00002936, -0.00002026, 0.00001216, 0.00000703, -0.00000451, -0.00000253, 0.00000866, 0.00000720, -0.00002288, -0.00002493, 0.00002956, 0.00002101, -0.00002303, -0.00000625, -0.00002935, -0.00000067, -0.00008862, -0.00007120, 0.00027507, 0.00022759, -0.00010920, -0.00005241, -0.00003291, -0.00000759, -0.00002114, 0.00003128, -0.00006392, -0.00003136, 0.00002648, -0.00000952, -0.00001018, -0.00003019, 0.00002908, 0.00002328, -0.00003597, -0.00000839, },
		{ 0.00001442, 0.00001669, 0.00000742, -0.00004029, 0.00000833, 0.00003361, -0.00001376, -0.00002627, -0.00000640, 0.00001944, 0.00000245, -0.00008504, 0.00001777, -0.00000254, -0.00000988, -0.00003997, -0.00001684, -0.00016429, 0.00009301, 0.00044695, -0.00002391, -0.00013529, -0.00000540, -0.00002877, -0.00001207, -0.00001715, -0.00000214, 0.00004569, -0.00001855, -0.00002691, -0.00000388, 0.00001959, -0.00000744, 0.00000072, -0.00000145, 0.00001551, -0.00001320, -0.00002700, 0.00000026, 0.00004301, -0.00001774, -0.00000630, -0.00002601, -0.00000569, -0.00004086, -0.00010931, 0.00011198, 0.00030954, 0.00000698, -0.00009066, -0.00003548, -0.00001448, -0.00001840, 0.00000905, -0.00000633, -0.00007812, 0.00002806, -0.00000708, 0.00000008, -0.00003170, 0.00000264, 0.00003980, -0.00001571, -0.00003018, },
		{ -0.00000144, 0.00001698, 0.00001780, -0.00002428, -0.00002272, 0.00003588, 0.00000711, -0.00002806, -0.00001608, 0.00000717, 0.00002882, -0.00006388, -0.00003592, 0.00002669, -0.00001628, -0.00002299, -0.00003039, -0.00007888, -0.00014605, 0.00043418, 0.00022240, -0.00018671, -0.00003492, -0.00001833, -0.00002436, -0.00000672, -0.00002392, 0.00003963, 0.00001129, -0.00003212, -0.00001711, 0.00001772, 0.00000303, -0.00000479, -0.00000090, 0.00000930, 0.00000228, -0.00002431, -0.00001948, 0.00003624, 0.00000910, -0.00001732, -0.00001314, -0.00001942, -0.00000925, -0.00009446, -0.00001229, 0.00025748, 0.00015792, -0.00008767, -0.00004580, -0.00003259, 0.00000027, -0.00002508, 0.00003274, -0.00007680, -0.00000857, 0.00001238, -0.00000742, -0.00001221, -0.00002648, 0.00003838, 0.00001168, -0.00003676, },
		{ -0.00002369, 0.00001660, 0.00001608, 0.00000239, -0.00004162, 0.00002085, 0.00002312, -0.00001420, -0.00003010, 0.00000023, 0.00002503, -0.00001150, -0.00008283, 0.00002848, -0.00001032, -0.00000454, -0.00005415, -0.00000537, -0.00021530, 0.00017706, 0.00049573, -0.00012658, -0.00010830, -0.00000862, -0.00002967, -0.00000919, -0.00002066, 0.00000642, 0.00004358, -0.00002794, -0.00002398, 0.00000082, 0.00002081, -0.00001245, 0.00000423, -0.00000185, 0.00001426, -0.00001803, -0.00002423, 0.00000916, 0.00003713, -0.00002105, -0.00000492, -0.00002754, 0.00000293, -0.00006340, -0.00006756, 0.00013518, 0.00024843, -0.00001479, -0.00006089, -0.00004749, 0.00000225, -0.00002711, 0.00001753, -0.00002330, -0.00006338, 0.00003002, -0.00001761, 0.00000653, -0.00003747, 0.00001710, 0.00003343, -0.00002322, },
		{ -0.00003489, 0.00000473, 0.00001545, 0.00001698, -0.00003218, -0.00001045, 0.00003322, 0.00000105, -0.00002924, -0.00001592, 0.00001796, 0.00002062, -0.00007273, -0.00002171, 0.00002226, -0.00001356, -0.00003197, -0.00002590, -0.00010351, -0.00015045, 0.00056823, 0.00012930, -0.00020933, -0.00001092, -0.00003175, -0.00001539, -0.00001186, -0.00001981, 0.00004577, -0.00000008, -0.00003245, -0.00001394, 0.00002391, -0.00000398, -0.00000189, -0.00000149, 0.00001052, -0.00000338, -0.00002391, -0.00001361, 0.00004067, -0.00000270, -0.00001036, -0.00002190, -0.00000377, -0.00002961, -0.00007901, 0.00002881, 0.00023055, 0.00010303, -0.00006361, -0.00004560, -0.00002788, 0.00000488, -0.00002725, 0.00003104, -0.00008482, 0.00001334, -0.00000712, 0.00000215, -0.00002001, -0.00001636, 0.00004214, 0.00000023, },
		{ -0.00002668, -0.00001740, 0.00001803, 0.00001507, -0.00000356, -0.00003919, 0.00003029, 0.00001143, -0.00001192, -0.00003651, 0.00001333, 0.00002214, -0.00002130, -0.00008016, 0.00003999, -0.00002150, 0.00000431, -0.00007229, 0.00000999, -0.00027928, 0.00030280, 0.00050378, -0.00023007, -0.00006158, -0.00002719, -0.00001964, -0.00001295, -0.00001992, 0.00001331, 0.00003982, -0.00003663, -0.00001958, 0.00000609, 0.00001962, -0.00001635, 0.00000807, -0.00000345, 0.00001327, -0.00002242, -0.00002010, 0.00001737, 0.00002830, -0.00002025, -0.00000845, -0.00002076, 0.00000080, -0.00007512, -0.00002359, 0.00013663, 0.00019782, -0.00002866, -0.00003489, -0.00006091, 0.00002184, -0.00003959, 0.00002850, -0.00004335, -0.00004047, 0.00002078, -0.00001967, 0.00000887, -0.00004028, 0.00003164, 0.00002228, },
		{ -0.00000845, -0.00003178, 0.00001041, 0.00001304, 0.00001571, -0.00003893, 0.00000394, 0.00002450, 0.00000068, -0.00003538, -0.00000837, 0.00002307, 0.00001242, -0.00008006, -0.00000469, 0.00001234, -0.00000558, -0.00004737, -0.00001297, -0.00014524, -0.00012783, 0.00070506, -0.00000873, -0.00019242, -0.00000539, -0.00003606, -0.00001018, -0.00001539, -0.00001559, 0.00005085, -0.00001328, -0.00002969, -0.00001112, 0.00003013, -0.00001350, 0.00000448, -0.00000526, 0.00001346, -0.00001034, -0.00002124, -0.00000783, 0.00004235, -0.00001263, -0.00000481, -0.00002821, 0.00001159, -0.00005522, -0.00004600, 0.00004700, 0.00020532, 0.00005634, -0.00003519, -0.00005396, -0.00001569, 0.00000287, -0.00002458, 0.00002342, -0.00008404, 0.00002855, -0.00002620, 0.00001561, -0.00003203, 0.00000001, 0.00003858, },
		{ 0.00000922, -0.00002676, -0.00001150, 0.00001855, 0.00001380, -0.00001016, -0.00003267, 0.00003448, 0.00000199, -0.00001044, -0.00004143, 0.00002888, 0.00001258, -0.00002672, -0.00007686, 0.00005082, -0.00003428, 0.00001506, -0.00009264, 0.00002636, -0.00035240, 0.00047455, 0.00045122, -0.00031063, -0.00001017, -0.00005369, -0.00000198, -0.00002223, -0.00001519, 0.00001833, 0.00003418, -0.00004357, -0.00001442, 0.00001226, 0.00001478, -0.00001697, 0.00000992, -0.00000439, 0.00001144, -0.00002541, -0.00001531, 0.00002453, 0.00001746, -0.00001620, -0.00001589, -0.00000591, -0.00001444, -0.00006880, 0.00000934, 0.00012834, 0.00015404, -0.00003392, -0.00001562, -0.00007204, 0.00004101, -0.00005322, 0.00003958, -0.00006326, -0.00001271, 0.00000146, -0.00001135, 0.00000377, -0.00003645, 0.00004220, },
		{ 0.00002789, -0.00001276, -0.00002872, 0.00001590, 0.00000952, 0.00001416, -0.00004370, 0.00001839, 0.00001153, 0.00000539, -0.00004651, 0.00000797, 0.00001853, 0.00000895, -0.00008914, 0.00001753, -0.00000629, 0.00001197, -0.00007430, 0.00001443, -0.00021322, -0.00005953, 0.00081932, -0.00017658, -0.00013568, -0.00002461, -0.00002464, -0.00001427, -0.00001332, -0.00001390, 0.00005600, -0.00002841, -0.00002344, -0.00000875, 0.00003522, -0.00002397, 0.00001327, -0.00001187, 0.00001827, -0.00001863, -0.00001623, -0.00000255, 0.00004108, -0.00001941, -0.00000284, -0.00002804, 0.00001982, -0.00007654, -0.00000528, 0.00004543, 0.00018629, 0.00001461, -0.00000265, -0.00007085, 0.00000498, -0.00000738, -0.00001536, 0.00000856, -0.00007206, 0.00003220, -0.00003848, 0.00002754, -0.00004442, 0.00002009, },
		{ 0.00004534, -0.00000240, -0.00002497, -0.00000565, 0.00001773, 0.00001261, -0.00001706, -0.00002239, 0.00003220, -0.00000235, -0.00001329, -0.00004012, 0.00004089, 0.00000117, -0.00003056, -0.00007017, 0.00005671, -0.00004351, 0.00002203, -0.00010872, 0.00003493, -0.00042226, 0.00068568, 0.00032647, -0.00034597, 0.00002753, -0.00007614, 0.00001581, -0.00003259, -0.00000911, 0.00002276, 0.00002567, -0.00004724, -0.00000949, 0.00001940, 0.00000575, -0.00001280, 0.00000775, -0.00000269, 0.00000735, -0.00002586, -0.00001071, 0.00003030, 0.00000595, -0.00001078, -0.00002424, 0.00001327, -0.00004009, -0.00004321, 0.00002271, 0.00012157, 0.00011135, -0.00002797, -0.00000715, -0.00007593, 0.00005492, -0.00006373, 0.00004682, -0.00007812, 0.00001399, -0.00002319, 0.00000565, -0.00000938, -0.00002400, },
		{ 0.00003969, 0.00001220, -0.00001246, -0.00002639, 0.00002128, 0.00000487, 0.00001255, -0.00004583, 0.00003063, -0.00000278, 0.00001270, -0.00006010, 0.00003124, 0.00000366, 0.00001255, -0.00010150, 0.00004533, -0.00003442, 0.00004080, -0.00011527, 0.00005946, -0.00031340, 0.00007193, 0.00088052, -0.00034795, -0.00005036, -0.00006637, 0.00000357, -0.00002979, -0.00000365, -0.00001632, 0.00006167, -0.00004498, -0.00001385, -0.00000674, 0.00003781, -0.00003307, 0.00002226, -0.00001965, 0.00002399, -0.00002746, -0.00000934, 0.00000192, 0.00003712, -0.00002271, -0.00000508, -0.00001937, 0.00001585, -0.00008442, 0.00003061, 0.00003333, 0.00017153, -0.00002117, 0.00003060, -0.00009302, 0.00003194, -0.00002460, -0.00000058, -0.00001218, -0.00004959, 0.00002234, -0.00003927, 0.00003255, -0.00005232, },
		{ -0.00000354, 0.00003930, -0.00000968, -0.00002316, 0.00000085, 0.00001493, 0.00001206, -0.00002393, -0.00000930, 0.00002351, -0.00000016, -0.00002267, -0.00002877, 0.00004330, -0.00000586, -0.00003742, -0.00005579, 0.00005182, -0.00004182, 0.00001662, -0.00011026, 0.00002166, -0.00046767, 0.00091598, 0.00013175, -0.00032187, 0.00003534, -0.00008184, 0.00002444, -0.00003771, -0.00000585, 0.00002893, 0.00001282, -0.00004590, -0.00000593, 0.00002721, -0.00000699, -0.00000351, 0.00000055, 0.00000302, -0.00000007, -0.00002279, -0.00000720, 0.00003453, -0.00000490, -0.00000615, -0.00002957, 0.00003071, -0.00006907, -0.00000373, 0.00001518, 0.00012294, 0.00006562, -0.00000921, -0.00001230, -0.00006868, 0.00005933, -0.00006705, 0.00004647, -0.00008304, 0.00003290, -0.00004617, 0.00002639, -0.00002807, },
		{ -0.00005154, 0.00005408, -0.00000495, -0.00000868, -0.00002468, 0.00002615, -0.00000057, 0.00001098, -0.00004497, 0.00003868, -0.00001512, 0.00001901, -0.00007168, 0.00005645, -0.00001835, 0.00002204, -0.00011555, 0.00007572, -0.00006874, 0.00007784, -0.00016739, 0.00011899, -0.00044382, 0.00027633, 0.00086003, -0.00049046, 0.00004273, -0.00011985, 0.00004305, -0.00005420, 0.00001241, -0.00002231, 0.00006699, -0.00006143, -0.00000186, -0.00000481, 0.00003663, -0.00003836, 0.00002867, -0.00002610, 0.00002876, -0.00003534, -0.00000159, 0.00000548, 0.00003113, -0.00002327, -0.00001027, -0.00000319, -0.00000157, -0.00007336, 0.00005061, 0.00002212, 0.00015513, -0.00004680, 0.00005871, -0.00011459, 0.00006011, -0.00004464, 0.00001612, -0.00003453, -0.00002086, 0.00000104, -0.00002727, 0.00002722, },
		{ -0.00004744, 0.00002175, 0.00002456, -0.00001094, -0.00002287, 0.00000864, 0.00000951, 0.00001283, -0.00003055, 0.00000525, 0.00000974, 0.00000816, -0.00003865, -0.00000610, 0.00003205, -0.00000315, -0.00005166, -0.00002984, 0.00003092, -0.00002220, -0.00001001, -0.00008614, -0.00002951, -0.00046160, 0.00113268, -0.00011452, -0.00023709, 0.00000457, -0.00006183, 0.00001620, -0.00003173, -0.00000959, 0.00003920, -0.00000563, -0.00003806, -0.00000477, 0.00003505, -0.00002196, 0.00000979, -0.00001125, 0.00001287, -0.00001110, -0.00001574, -0.00000551, 0.00003731, -0.00001423, -0.00000384, -0.00002856, 0.00004007, -0.00009212, 0.00003928, -0.00000746, 0.00013280, 0.00001670, 0.00002085, -0.00003073, -0.00004932, 0.00005249, -0.00006128, 0.00003685, -0.00007527, 0.00003876, -0.00006035, 0.00004428, },
		{ 0.00001139, -0.00003997, 0.00005938, -0.00001962, -0.00000353, -0.00002275, 0.00002957, -0.00000596, 0.00000932, -0.00004104, 0.00004116, -0.00002262, 0.00002064, -0.00007608, 0.00007667, -0.00004114, 0.00003280, -0.00012654, 0.00010220, -0.00010152, 0.00011460, -0.00022121, 0.00018158, -0.00059118, 0.00054992, 0.00073911, -0.00057314, 0.00011837, -0.00016836, 0.00008277, -0.00008044, 0.00003032, -0.00002905, 0.00006963, -0.00007511, 0.00001074, -0.00000253, 0.00003094, -0.00003797, 0.00003000, -0.00002855, 0.00003031, -0.00004034, 0.00000560, 0.00000834, 0.00002389, -0.00002244, -0.00001563, 0.00001648, -0.00002902, -0.00004364, 0.00004875, 0.00002116, 0.00013071, -0.00005749, 0.00007545, -0.00012891, 0.00008323, -0.00006189, 0.00002957, -0.00005253, 0.00000725, -0.00002584, -0.00000526, },
		{ 0.00005319, -0.00006170, 0.00004691, 0.00000378, -0.00000613, -0.00002482, 0.00001795, 0.00000120, 0.00001548, -0.00003681, 0.00001968, -0.00000679, 0.00002048, -0.00005880, 0.00002584, 0.00000672, 0.00001171, -0.00007524, 0.00000882, -0.00000797, 0.00001890, -0.00006319, -0.00002874, -0.00013139, -0.00037721, 0.00129547, -0.00037937, -0.00010518, -0.00006262, -0.00001437, -0.00001202, -0.00001164, -0.00002274, 0.00005475, -0.00002986, -0.00002310, -0.00000674, 0.00004200, -0.00003694, 0.00002479, -0.00002571, 0.00002560, -0.00002490, -0.00000507, -0.00000601, 0.00003893, -0.00002189, -0.00000401, -0.00001991, 0.00003702, -0.00010092, 0.00007382, -0.00003464, 0.00014586, -0.00003123, 0.00005700, -0.00005824, -0.00002071, 0.00003620, -0.00004775, 0.00001955, -0.00005567, 0.00002982, -0.00006093, },
		{ 0.00002064, -0.00001167, -0.00001844, 0.00005367, -0.00002841, 0.00000054, -0.00001926, 0.00003024, -0.00001005, 0.00000708, -0.00003419, 0.00003760, -0.00002347, 0.00001498, -0.00006903, 0.00008494, -0.00005697, 0.00003824, -0.00012781, 0.00011613, -0.00012206, 0.00013847, -0.00026178, 0.00022811, -0.00073023, 0.00087191, 0.00051554, -0.00057444, 0.00015362, -0.00019408, 0.00010936, -0.00009896, 0.00004335, -0.00003209, 0.00006626, -0.00008264, 0.00002151, 0.00000063, 0.00002071, -0.00003109, 0.00002474, -0.00002506, 0.00002671, -0.00004060, 0.00001062, 0.00001098, 0.00001601, -0.00002161, -0.00001787, 0.00003383, -0.00005930, -0.00000149, 0.00002600, 0.00003461, 0.00009474, -0.00005059, 0.00007665, -0.00013097, 0.00009607, -0.00007140, 0.00003514, -0.00006050, 0.00002726, -0.00005030, },
		{ -0.00004709, 0.00004954, -0.00006605, 0.00006657, -0.00001890, 0.00000318, -0.00002869, 0.00002826, -0.00000963, 0.00002016, -0.00004254, 0.00003249, -0.00002329, 0.00003349, -0.00007872, 0.00006174, -0.00002893, 0.00003704, -0.00010629, 0.00005708, -0.00006166, 0.00007902, -0.00014144, 0.00006176, -0.00028773, -0.00019554, 0.00136461, -0.00062109, 0.00004783, -0.00015301, 0.00005369, -0.00005699, 0.00002107, -0.00004461, 0.00007460, -0.00005829, -0.00000180, -0.00001190, 0.00004710, -0.00004944, 0.00003846, -0.00003989, 0.00003876, -0.00003949, 0.00000792, -0.00000851, 0.00003974, -0.00002839, -0.00000531, -0.00000508, 0.00002093, -0.00009090, 0.00009004, -0.00005481, 0.00015393, -0.00007121, 0.00009153, -0.00008779, 0.00001120, 0.00001542, -0.00003082, -0.00000061, -0.00002900, 0.00000881, },
		{ -0.00006461, 0.00004279, -0.00003614, 0.00000938, 0.00003756, -0.00002940, 0.00000150, -0.00001298, 0.00002683, -0.00001133, 0.00000352, -0.00002471, 0.00002837, -0.00001737, 0.00000136, -0.00004865, 0.00007636, -0.00005906, 0.00003205, -0.00011309, 0.00010931, -0.00011962, 0.00013596, -0.00027205, 0.00023543, -0.00082682, 0.00120510, 0.00020711, -0.00048839, 0.00013432, -0.00018343, 0.00011122, -0.00010062, 0.00004463, -0.00002679, 0.00005362, -0.00008065, 0.00002775, 0.00000519, 0.00000673, -0.00001826, 0.00001287, -0.00001501, 0.00001703, -0.00003495, 0.00001211, 0.00001409, 0.00000765, -0.00002143, -0.00001456, 0.00004312, -0.00008359, 0.00004265, -0.00001021, 0.00006021, 0.00004847, -0.00002750, 0.00006213, -0.00011935, 0.00009653, -0.00007087, 0.00003072, -0.00005529, 0.00003360, },
		{ -0.00001757, -0.00002250, 0.00003349, -0.00005816, 0.00007636, -0.00003893, 0.00001433, -0.00003315, 0.00003824, -0.00002179, 0.00002638, -0.00004746, 0.00004233, -0.00003706, 0.00004360, -0.00009317, 0.00009422, -0.00006760, 0.00006713, -0.00013895, 0.00010731, -0.00012130, 0.00014868, -0.00023478, 0.00017463, -0.00048871, 0.00008723, 0.00131036, -0.00079792, 0.00018917, -0.00024523, 0.00012813, -0.00010940, 0.00006022, -0.00007107, 0.00009529, -0.00008732, 0.00002341, -0.00001954, 0.00004943, -0.00005726, 0.00004785, -0.00005052, 0.00004926, -0.00005217, 0.00002117, -0.00001222, 0.00003988, -0.00003446, -0.00000537, 0.00001196, -0.00000455, -0.00006300, 0.00008335, -0.00005934, 0.00014942, -0.00009635, 0.00011658, -0.00011168, 0.00003926, -0.00000345, -0.00001644, -0.00001692, -0.00000280, },
		{ 0.00002468, -0.00006394, 0.00005445, -0.00005547, 0.00003800, 0.00001397, -0.00002252, -0.00000167, -0.00000321, 0.00001847, -0.00000852, -0.00000217, -0.00001299, 0.00001464, -0.00000547, -0.00001868, -0.00001628, 0.00004975, -0.00004381, 0.00001057, -0.00007905, 0.00007701, -0.00008706, 0.00009698, -0.00023801, 0.00018227, -0.00084449, 0.00150106, -0.00014973, -0.00032721, 0.00005940, -0.00013153, 0.00008236, -0.00007984, 0.00002955, -0.00001006, 0.00002978, -0.00006686, 0.00002717, 0.00001147, -0.00000955, -0.00000129, -0.00000404, 0.00000056, 0.00000176, -0.00002327, 0.00000928, 0.00001833, -0.00000148, -0.00002142, -0.00000519, 0.00004077, -0.00009427, 0.00007733, -0.00004848, 0.00009053, -0.00000220, 0.00000612, 0.00003600, -0.00009686, 0.00008646, -0.00006172, 0.00001797, -0.00003782, },
		{ 0.00001494, -0.00004053, 0.00000565, 0.00000901, -0.00003899, 0.00007417, -0.00005245, 0.00002425, -0.00003626, 0.00004593, -0.00003332, 0.00003292, -0.00005094, 0.00004820, -0.00004598, 0.00004784, -0.00009746, 0.00011558, -0.00010030, 0.00009375, -0.00016473, 0.00014879, -0.00017381, 0.00021263, -0.00032555, 0.00028871, -0.00070940, 0.00045500, 0.00112131, -0.00087757, 0.00028721, -0.00031534, 0.00019098, -0.00015608, 0.00009625, -0.00009531, 0.00011148, -0.00011180, 0.00004859, -0.00002810, 0.00004830, -0.00005893, 0.00005076, -0.00005492, 0.00005426, -0.00006013, 0.00003224, -0.00001583, 0.00003904, -0.00004053, -0.00000175, 0.00002589, -0.00003249, -0.00002362, 0.00005591, -0.00004534, 0.00012851, -0.00010262, 0.00012678, -0.00012408, 0.00005767, -0.00001488, -0.00000984, -0.00002316, },
		{ -0.00001323, 0.00000382, -0.00004801, 0.00005195, -0.00006436, 0.00006155, -0.00001259, -0.00000949, -0.00000870, 0.00000975, 0.00000516, -0.00000098, -0.00001053, 0.00000044, -0.00000190, 0.00000990, -0.00004165, 0.00002354, 0.00000829, -0.00001211, -0.00002553, -0.00002720, 0.00002030, -0.00002401, 0.00001899, -0.00015390, 0.00005620, -0.00075314, 0.00170899, -0.00050588, -0.00011941, -0.00005819, -0.00004421, 0.00002476, -0.00003678, -0.00000239, 0.00001820, -0.00000470, -0.00004098, 0.00001856, 0.00001949, -0.00002627, 0.00001717, -0.00002319, 0.00001916, -0.00001713, -0.00000675, 0.00000226, 0.00002402, -0.00001190, -0.00002004, 0.00000831, 0.00002668, -0.00008744, 0.00009358, -0.00007721, 0.00011575, -0.00004870, 0.00004180, 0.00000557, -0.00006971, 0.00007115, -0.00004874, 0.00000211, },
		{ -0.00001589, 0.00001848, -0.00005198, 0.00002886, -0.00001731, -0.00001253, 0.00006058, -0.00005720, 0.00003036, -0.00003607, 0.00004921, -0.00004184, 0.00003801, -0.00005212, 0.00004940, -0.00004882, 0.00004465, -0.00008879, 0.00011985, -0.00011867, 0.00010851, -0.00017488, 0.00017043, -0.00020501, 0.00025328, -0.00039193, 0.00037591, -0.00091234, 0.00087144, 0.00080910, -0.00084494, 0.00031927, -0.00034321, 0.00022542, -0.00018363, 0.00011878, -0.00010972, 0.00011728, -0.00012601, 0.00006896, -0.00003548, 0.00004333, -0.00005403, 0.00004633, -0.00005169, 0.00005189, -0.00006120, 0.00003887, -0.00001781, 0.00003650, -0.00004624, 0.00000698, 0.00003183, -0.00005477, 0.00001724, 0.00001580, -0.00001629, 0.00009268, -0.00009061, 0.00012108, -0.00012309, 0.00006401, -0.00001632, -0.00001340, },
		{ -0.00000966, 0.00001045, -0.00002154, -0.00002141, 0.00003589, -0.00006039, 0.00007544, -0.00003724, 0.00000674, -0.00001812, 0.00002453, -0.00001196, 0.00001093, -0.00002150, 0.00001480, -0.00001929, 0.00002587, -0.00006300, 0.00006392, -0.00004100, 0.00003060, -0.00007088, 0.00003557, -0.00005302, 0.00006169, -0.00009064, -0.00002594, -0.00014064, -0.00053776, 0.00178643, -0.00080914, 0.00009631, -0.00019444, 0.00006298, -0.00005155, 0.00002202, -0.00004690, 0.00005446, -0.00004623, -0.00000534, 0.00000228, 0.00002877, -0.00004144, 0.00003423, -0.00004125, 0.00003740, -0.00003657, 0.00001225, -0.00000782, 0.00003089, -0.00002395, -0.00001518, 0.00002201, 0.00000456, -0.00006437, 0.00008773, -0.00008824, 0.00012734, -0.00008270, 0.00007090, -0.00002113, -0.00004532, 0.00005741, -0.00003836, },
		{ -0.00002537, 0.00000389, 0.00000648, -0.00004721, 0.00003996, -0.00003826, 0.00001538, 0.00003860, -0.00005290, 0.00003116, -0.00003123, 0.00004637, -0.00004508, 0.00003964, -0.00004993, 0.00004564, -0.00004541, 0.00003422, -0.00006717, 0.00010445, -0.00011728, 0.00010539, -0.00016323, 0.00016408, -0.00020362, 0.00025543, -0.00041350, 0.00040749, -0.00105414, 0.00128462, 0.00040819, -0.00070601, 0.00027734, -0.00031808, 0.00022056, -0.00018232, 0.00011969, -0.00010815, 0.00010805, -0.00012518, 0.00007995, -0.00003940, 0.00003455, -0.00004319, 0.00003515, -0.00004106, 0.00004185, -0.00005440, 0.00003951, -0.00001680, 0.00003120, -0.00005024, 0.00002033, 0.00002704, -0.00006462, 0.00004896, -0.00002552, 0.00001930, 0.00004837, -0.00006562, 0.00010311, -0.00011130, 0.00006015, -0.00000934, },
		{ -0.00003622, -0.00001063, 0.00002483, -0.00004198, 0.00000764, 0.00001103, -0.00004474, 0.00007750, -0.00005589, 0.00002280, -0.00002767, 0.00003885, -0.00003063, 0.00002567, -0.00003420, 0.00002902, -0.00003557, 0.00003967, -0.00007822, 0.00009727, -0.00008889, 0.00007568, -0.00011659, 0.00009818, -0.00012962, 0.00015479, -0.00021484, 0.00012688, -0.00038873, -0.00020465, 0.00170912, -0.00101503, 0.00027976, -0.00032016, 0.00016851, -0.00013083, 0.00008507, -0.00009561, 0.00009207, -0.00008850, 0.00003520, -0.00001954, 0.00003832, -0.00005327, 0.00004734, -0.00005513, 0.00005185, -0.00005308, 0.00003070, -0.00001907, 0.00003795, -0.00003726, -0.00000512, 0.00003110, -0.00001906, -0.00003124, 0.00006241, -0.00007926, 0.00012095, -0.00009904, 0.00008731, -0.00003798, -0.00002969, 0.00005100, },
		{ 0.00000094, -0.00004015, 0.00003068, -0.00001718, -0.00002663, 0.00003541, -0.00004830, 0.00003868, 0.00001280, -0.00004114, 0.00002653, -0.00002148, 0.00003668, -0.00004151, 0.00003617, -0.00004351, 0.00003700, -0.00003648, 0.00001840, -0.00003553, 0.00007095, -0.00009514, 0.00008268, -0.00012840, 0.00012734, -0.00016490, 0.00021098, -0.00037724, 0.00036171, -0.00109470, 0.00163608, -0.00002976, -0.00048690, 0.00016996, -0.00024158, 0.00017469, -0.00014904, 0.00009558, -0.00008797, 0.00008209, -0.00010692, 0.00007827, -0.00003800, 0.00002239, -0.00002791, 0.00001906, -0.00002486, 0.00002551, -0.00004031, 0.00003376, -0.00001207, 0.00002224, -0.00005048, 0.00003576, 0.00001207, -0.00005887, 0.00006360, -0.00005684, 0.00005094, 0.00000480, -0.00003600, 0.00007985, -0.00009481, 0.00005155, },
		{ 0.00005446, -0.00004483, 0.00000287, 0.00002403, -0.00004915, 0.00002985, -0.00001525, -0.00002177, 0.00006839, -0.00006613, 0.00003573, -0.00003493, 0.00005005, -0.00004787, 0.00004075, -0.00004692, 0.00004169, -0.00004901, 0.00004918, -0.00008403, 0.00011706, -0.00012613, 0.00011351, -0.00015251, 0.00014803, -0.00019359, 0.00023583, -0.00033005, 0.00027567, -0.00065282, 0.00021679, 0.00147747, -0.00109612, 0.00039836, -0.00040829, 0.00025044, -0.00019565, 0.00013875, -0.00013818, 0.00012280, -0.00012373, 0.00007400, -0.00004329, 0.00004668, -0.00006043, 0.00005479, -0.00006271, 0.00005986, -0.00006361, 0.00004550, -0.00002917, 0.00004362, -0.00005053, 0.00001057, 0.00003155, -0.00003685, 0.00000271, 0.00002573, -0.00005429, 0.00009787, -0.00009747, 0.00008935, -0.00004286, -0.00002522, },
		{ 0.00004517, 0.00000762, -0.00004998, 0.00005599, -0.00004435, 0.00000414, 0.00001647, -0.00004513, 0.00005283, -0.00001201, -0.00002477, 0.00001771, -0.00000779, 0.00002073, -0.00003085, 0.00002685, -0.00003246, 0.00002403, -0.00002344, 0.00000019, 0.00000088, 0.00002483, -0.00005599, 0.00004364, -0.00007472, 0.00006489, -0.00009281, 0.00012217, -0.00028221, 0.00023078, -0.00100700, 0.00187220, -0.00044567, -0.00022804, 0.00001994, -0.00012742, 0.00009592, -0.00008848, 0.00004911, -0.00005119, 0.00004162, -0.00007226, 0.00006293, -0.00003040, 0.00000774, -0.00001020, 0.00000067, -0.00000602, 0.00000564, -0.00002097, 0.00002251, -0.00000380, 0.00000927, -0.00004483, 0.00004914, -0.00000915, -0.00003908, 0.00005823, -0.00007061, 0.00006970, -0.00002911, -0.00001047, 0.00005912, -0.00008095, },
		{ -0.00002994, 0.00006602, -0.00006231, 0.00002981, 0.00000695, -0.00003858, 0.00003771, -0.00003517, 0.00000235, 0.00005121, -0.00006756, 0.00004365, -0.00003799, 0.00005569, -0.00006056, 0.00005318, -0.00005738, 0.00005119, -0.00005817, 0.00005335, -0.00007914, 0.00011943, -0.00014572, 0.00013607, -0.00017015, 0.00017427, -0.00023052, 0.00028596, -0.00041169, 0.00038770, -0.00088773, 0.00067761, 0.00111775, -0.00104762, 0.00043428, -0.00044056, 0.00029203, -0.00023148, 0.00017093, -0.00016506, 0.00013904, -0.00014450, 0.00010402, -0.00006456, 0.00005219, -0.00006212, 0.00005595, -0.00006323, 0.00006017, -0.00006623, 0.00005425, -0.00003595, 0.00004604, -0.00006154, 0.00003044, 0.00002154, -0.00004301, 0.00002797, -0.00001139, -0.00002207, 0.00006443, -0.00008262, 0.00008010, -0.00003820, },
		{ -0.00007559, 0.00004689, 0.00000472, -0.00004764, 0.00007102, -0.00006503, 0.00003581, -0.00001115, -0.00003043, 0.00005596, -0.00003193, -0.00000701, 0.00000680, 0.00000777, 0.00000051, -0.00001428, 0.00001218, -0.00001726, 0.00000777, -0.00000809, -0.00001705, 0.00003570, -0.00002584, -0.00000731, -0.00000422, -0.00001154, -0.00001209, 0.00000022, 0.00000217, -0.00014143, 0.00002487, -0.00078472, 0.00195541, -0.00078472, 0.00002487, -0.00014143, 0.00000217, 0.00000022, -0.00001209, -0.00001154, -0.00000422, -0.00000731, -0.00002584, 0.00003570, -0.00001705, -0.00000809, 0.00000777, -0.00001726, 0.00001218, -0.00001428, 0.00000051, 0.00000777, 0.00000680, -0.00000701, -0.00003193, 0.00005596, -0.00003043, -0.00001115, 0.00003581, -0.00006503, 0.00007102, -0.00004764, 0.00000472, 0.00004689, },
		{ -0.00002994, -0.00003820, 0.00008010, -0.00008262, 0.00006443, -0.00002207, -0.00001139, 0.00002797, -0.00004301, 0.00002154, 0.00003044, -0.00006154, 0.00004604, -0.00003595, 0.00005425, -0.00006623, 0.00006017, -0.00006323, 0.00005595, -0.00006212, 0.00005219, -0.00006456, 0.00010402, -0.00014450, 0.00013904, -0.00016506, 0.00017093, -0.00023148, 0.00029203, -0.00044056, 0.00043428, -0.00104762, 0.00111775, 0.00067761, -0.00088773, 0.00038770, -0.00041169, 0.00028596, -0.00023052, 0.00017427, -0.00017015, 0.00013607, -0.00014572, 0.00011943, -0.00007914, 0.00005335, -0.00005817, 0.00005119, -0.00005738, 0.00005318, -0.00006056, 0.00005569, -0.00003799, 0.00004365, -0.00006756, 0.00005121, 0.00000235, -0.00003517, 0.00003771, -0.00003858, 0.00000695, 0.00002981, -0.00006231, 0.00006602, },
		{ 0.00004517, -0.00008095, 0.00005912, -0.00001047, -0.00002911, 0.00006970, -0.00007061, 0.00005823, -0.00003908, -0.00000915, 0.00004914, -0.00004483, 0.00000927, -0.00000380, 0.00002251, -0.00002097, 0.00000564, -0.00000602, 0.00000067, -0.00001020, 0.00000774, -0.00003040, 0.00006293, -0.00007226, 0.00004162, -0.00005119, 0.00004911, -0.00008848, 0.00009592, -0.00012742, 0.00001994, -0.00022804, -0.00044567, 0.00187220, -0.00100700, 0.00023078, -0.00028221, 0.00012217, -0.00009281, 0.00006489, -0.00007472, 0.00004364, -0.00005599, 0.00002483, 0.00000088, 0.00000019, -0.00002344, 0.00002403, -0.00003246, 0.00002685, -0.00003085, 0.00002073, -0.00000779, 0.00001771, -0.00002477, -0.00001201, 0.00005283, -0.00004513, 0.00001647, 0.00000414, -0.00004435, 0.00005599, -0.00004998, 0.00000762, },
		{ 0.00005446, -0.00002522, -0.00004286, 0.00008935, -0.00009747, 0.00009787, -0.00005429, 0.00002573, 0.00000271, -0.00003685, 0.00003155, 0.00001057, -0.00005053, 0.00004362, -0.00002917, 0.00004550, -0.00006361, 0.00005986, -0.00006271, 0.00005479, -0.00006043, 0.00004668, -0.00004329, 0.00007400, -0.00012373, 0.00012280, -0.00013818, 0.00013875, -0.00019565, 0.00025044, -0.00040829, 0.00039836, -0.00109612, 0.00147747, 0.00021679, -0.00065282, 0.00027567, -0.00033005, 0.00023583, -0.00019359, 0.00014803, -0.00015251, 0.00011351, -0.00012613, 0.00011706, -0.00008403, 0.00004918, -0.00004901, 0.00004169, -0.00004692, 0.00004075, -0.00004787, 0.00005005, -0.00003493, 0.00003573, -0.00006613, 0.00006839, -0.00002177, -0.00001525, 0.00002985, -0.00004915, 0.00002403, 0.00000287, -0.00004483, },
		{ 0.00000094, 0.00005155, -0.00009481, 0.00007985, -0.00003600, 0.00000480, 0.00005094, -0.00005684, 0.00006360, -0.00005887, 0.00001207, 0.00003576, -0.00005048, 0.00002224, -0.00001207, 0.00003376, -0.00004031, 0.00002551, -0.00002486, 0.00001906, -0.00002791, 0.00002239, -0.00003800, 0.00007827, -0.00010692, 0.00008209, -0.00008797, 0.00009558, -0.00014904, 0.00017469, -0.00024158, 0.00016996, -0.00048690, -0.00002976, 0.00163608, -0.00109470, 0.00036171, -0.00037724, 0.00021098, -0.00016490, 0.00012734, -0.00012840, 0.00008268, -0.00009514, 0.00007095, -0.00003553, 0.00001840, -0.00003648, 0.00003700, -0.00004351, 0.00003617, -0.00004151, 0.00003668, -0.00002148, 0.00002653, -0.00004114, 0.00001280, 0.00003868, -0.00004830, 0.00003541, -0.00002663, -0.00001718, 0.00003068, -0.00004015, },
		{ -0.00003622, 0.00005100, -0.00002969, -0.00003798, 0.00008731, -0.00009904, 0.00012095, -0.00007926, 0.00006241, -0.00003124, -0.00001906, 0.00003110, -0.00000512, -0.00003726, 0.00003795, -0.00001907, 0.00003070, -0.00005308, 0.00005185, -0.00005513, 0.00004734, -0.00005327, 0.00003832, -0.00001954, 0.00003520, -0.00008850, 0.00009207, -0.00009561, 0.00008507, -0.00013083, 0.00016851, -0.00032016, 0.00027976, -0.00101503, 0.00170912, -0.00020465, -0.00038873, 0.00012688, -0.00021484, 0.00015479, -0.00012962, 0.00009818, -0.00011659, 0.00007568, -0.00008889, 0.00009727, -0.00007822, 0.00003967, -0.00003557, 0.00002902, -0.00003420, 0.00002567, -0.00003063, 0.00003885, -0.00002767, 0.00002280, -0.00005589, 0.00007750, -0.00004474, 0.00001103, 0.00000764, -0.00004198, 0.00002483, -0.00001063, },
		{ -0.00002537, -0.00000934, 0.00006015, -0.00011130, 0.00010311, -0.00006562, 0.00004837, 0.00001930, -0.00002552, 0.00004896, -0.00006462, 0.00002704, 0.00002033, -0.00005024, 0.00003120, -0.00001680, 0.00003951, -0.00005440, 0.00004185, -0.00004106, 0.00003515, -0.00004319, 0.00003455, -0.00003940, 0.00007995, -0.00012518, 0.00010805, -0.00010815, 0.00011969, -0.00018232, 0.00022056, -0.00031808, 0.00027734, -0.00070601, 0.00040819, 0.00128462, -0.00105414, 0.00040749, -0.00041350, 0.00025543, -0.00020362, 0.00016408, -0.00016323, 0.00010539, -0.00011728, 0.00010445, -0.00006717, 0.00003422, -0.00004541, 0.00004564, -0.00004993, 0.00003964, -0.00004508, 0.00004637, -0.00003123, 0.00003116, -0.00005290, 0.00003860, 0.00001538, -0.00003826, 0.00003996, -0.00004721, 0.00000648, 0.00000389, },
		{ -0.00000966, -0.00003836, 0.00005741, -0.00004532, -0.00002113, 0.00007090, -0.00008270, 0.00012734, -0.00008824, 0.00008773, -0.00006437, 0.00000456, 0.00002201, -0.00001518, -0.00002395, 0.00003089, -0.00000782, 0.00001225, -0.00003657, 0.00003740, -0.00004125, 0.00003423, -0.00004144, 0.00002877, 0.00000228, -0.00000534, -0.00004623, 0.00005446, -0.00004690, 0.00002202, -0.00005155, 0.00006298, -0.00019444, 0.00009631, -0.00080914, 0.00178643, -0.00053776, -0.00014064, -0.00002594, -0.00009064, 0.00006169, -0.00005302, 0.00003557, -0.00007088, 0.00003060, -0.00004100, 0.00006392, -0.00006300, 0.00002587, -0.00001929, 0.00001480, -0.00002150, 0.00001093, -0.00001196, 0.00002453, -0.00001812, 0.00000674, -0.00003724, 0.00007544, -0.00006039, 0.00003589, -0.00002141, -0.00002154, 0.00001045, },
		{ -0.00001589, -0.00001340, -0.00001632, 0.00006401, -0.00012309, 0.00012108, -0.00009061, 0.00009268, -0.00001629, 0.00001580, 0.00001724, -0.00005477, 0.00003183, 0.00000698, -0.00004624, 0.00003650, -0.00001781, 0.00003887, -0.00006120, 0.00005189, -0.00005169, 0.00004633, -0.00005403, 0.00004333, -0.00003548, 0.00006896, -0.00012601, 0.00011728, -0.00010972, 0.00011878, -0.00018363, 0.00022542, -0.00034321, 0.00031927, -0.00084494, 0.00080910, 0.00087144, -0.00091234, 0.00037591, -0.00039193, 0.00025328, -0.00020501, 0.00017043, -0.00017488, 0.00010851, -0.00011867, 0.00011985, -0.00008879, 0.00004465, -0.00004882, 0.00004940, -0.00005212, 0.00003801, -0.00004184, 0.00004921, -0.00003607, 0.00003036, -0.00005720, 0.00006058, -0.00001253, -0.00001731, 0.00002886, -0.00005198, 0.00001848, },
		{ -0.00001323, 0.00000211, -0.00004874, 0.00007115, -0.00006971, 0.00000557, 0.00004180, -0.00004870, 0.00011575, -0.00007721, 0.00009358, -0.00008744, 0.00002668, 0.00000831, -0.00002004, -0.00001190, 0.00002402, 0.00000226, -0.00000675, -0.00001713, 0.00001916, -0.00002319, 0.00001717, -0.00002627, 0.00001949, 0.00001856, -0.00004098, -0.00000470, 0.00001820, -0.00000239, -0.00003678, 0.00002476, -0.00004421, -0.00005819, -0.00011941, -0.00050588, 0.00170899, -0.00075314, 0.00005620, -0.00015390, 0.00001899, -0.00002401, 0.00002030, -0.00002720, -0.00002553, -0.00001211, 0.00000829, 0.00002354, -0.00004165, 0.00000990, -0.00000190, 0.00000044, -0.00001053, -0.00000098, 0.00000516, 0.00000975, -0.00000870, -0.00000949, -0.00001259, 0.00006155, -0.00006436, 0.00005195, -0.00004801, 0.00000382, },
		{ 0.00001494, -0.00002316, -0.00000984, -0.00001488, 0.00005767, -0.00012408, 0.00012678, -0.00010262, 0.00012851, -0.00004534, 0.00005591, -0.00002362, -0.00003249, 0.00002589, -0.00000175, -0.00004053, 0.00003904, -0.00001583, 0.00003224, -0.00006013, 0.00005426, -0.00005492, 0.00005076, -0.00005893, 0.00004830, -0.00002810, 0.00004859, -0.00011180, 0.00011148, -0.00009531, 0.00009625, -0.00015608, 0.00019098, -0.00031534, 0.00028721, -0.00087757, 0.00112131, 0.00045500, -0.00070940, 0.00028871, -0.00032555, 0.00021263, -0.00017381, 0.00014879, -0.00016473, 0.00009375, -0.00010030, 0.00011558, -0.00009746, 0.00004784, -0.00004598, 0.00004820, -0.00005094, 0.00003292, -0.00003332, 0.00004593, -0.00003626, 0.00002425, -0.00005245, 0.00007417, -0.00003899, 0.00000901, 0.00000565, -0.00004053, },
		{ 0.00002468, -0.00003782, 0.00001797, -0.00006172, 0.00008646, -0.00009686, 0.00003600, 0.00000612, -0.00000220, 0.00009053, -0.00004848, 0.00007733, -0.00009427, 0.00004077, -0.00000519, -0.00002142, -0.00000148, 0.00001833, 0.00000928, -0.00002327, 0.00000176, 0.00000056, -0.00000404, -0.00000129, -0.00000955, 0.00001147, 0.00002717, -0.00006686, 0.00002978, -0.00001006, 0.00002955, -0.00007984, 0.00008236, -0.00013153, 0.00005940, -0.00032721, -0.00014973, 0.00150106, -0.00084449, 0.00018227, -0.00023801, 0.00009698, -0.00008706, 0.00007701, -0.00007905, 0.00001057, -0.00004381, 0.00004975, -0.00001628, -0.00001868, -0.00000547, 0.00001464, -0.00001299, -0.00000217, -0.00000852, 0.00001847, -0.00000321, -0.00000167, -0.00002252, 0.00001397, 0.00003800, -0.00005547, 0.00005445, -0.00006394, },
		{ -0.00001757, -0.00000280, -0.00001692, -0.00001644, -0.00000345, 0.00003926, -0.00011168, 0.00011658, -0.00009635, 0.00014942, -0.00005934, 0.00008335, -0.00006300, -0.00000455, 0.00001196, -0.00000537, -0.00003446, 0.00003988, -0.00001222, 0.00002117, -0.00005217, 0.00004926, -0.00005052, 0.00004785, -0.00005726, 0.00004943, -0.00001954, 0.00002341, -0.00008732, 0.00009529, -0.00007107, 0.00006022, -0.00010940, 0.00012813, -0.00024523, 0.00018917, -0.00079792, 0.00131036, 0.00008723, -0.00048871, 0.00017463, -0.00023478, 0.00014868, -0.00012130, 0.00010731, -0.00013895, 0.00006713, -0.00006760, 0.00009422, -0.00009317, 0.00004360, -0.00003706, 0.00004233, -0.00004746, 0.00002638, -0.00002179, 0.00003824, -0.00003315, 0.00001433, -0.00003893, 0.00007636, -0.00005816, 0.00003349, -0.00002250, },
		{ -0.00006461, 0.00003360, -0.00005529, 0.00003072, -0.00007087, 0.00009653, -0.00011935, 0.00006213, -0.00002750, 0.00004847, 0.00006021, -0.00001021, 0.00004265, -0.00008359, 0.00004312, -0.00001456, -0.00002143, 0.00000765, 0.00001409, 0.00001211, -0.00003495, 0.00001703, -0.00001501, 0.00001287, -0.00001826, 0.00000673, 0.00000519, 0.00002775, -0.00008065, 0.00005362, -0.00002679, 0.00004463, -0.00010062, 0.00011122, -0.00018343, 0.00013432, -0.00048839, 0.00020711, 0.00120510, -0.00082682, 0.00023543, -0.00027205, 0.00013596, -0.00011962, 0.00010931, -0.00011309, 0.00003205, -0.00005906, 0.00007636, -0.00004865, 0.00000136, -0.00001737, 0.00002837, -0.00002471, 0.00000352, -0.00001133, 0.00002683, -0.00001298, 0.00000150, -0.00002940, 0.00003756, 0.00000938, -0.00003614, 0.00004279, },
		{ -0.00004709, 0.00000881, -0.00002900, -0.00000061, -0.00003082, 0.00001542, 0.00001120, -0.00008779, 0.00009153, -0.00007121, 0.00015393, -0.00005481, 0.00009004, -0.00009090, 0.00002093, -0.00000508, -0.00000531, -0.00002839, 0.00003974, -0.00000851, 0.00000792, -0.00003949, 0.00003876, -0.00003989, 0.00003846, -0.00004944, 0.00004710, -0.00001190, -0.00000180, -0.00005829, 0.00007460, -0.00004461, 0.00002107, -0.00005699, 0.00005369, -0.00015301, 0.00004783, -0.00062109, 0.00136461, -0.00019554, -0.00028773, 0.00006176, -0.00014144, 0.00007902, -0.00006166, 0.00005708, -0.00010629, 0.00003704, -0.00002893, 0.00006174, -0.00007872, 0.00003349, -0.00002329, 0.00003249, -0.00004254, 0.00002016, -0.00000963, 0.00002826, -0.00002869, 0.00000318, -0.00001890, 0.00006657, -0.00006605, 0.00004954, },
		{ 0.00002064, -0.00005030, 0.00002726, -0.00006050, 0.00003514, -0.00007140, 0.00009607, -0.00013097, 0.00007665, -0.00005059, 0.00009474, 0.00003461, 0.00002600, -0.00000149, -0.00005930, 0.00003383, -0.00001787, -0.00002161, 0.00001601, 0.00001098, 0.00001062, -0.00004060, 0.00002671, -0.00002506, 0.00002474, -0.00003109, 0.00002071, 0.00000063, 0.00002151, -0.00008264, 0.00006626, -0.00003209, 0.00004335, -0.00009896, 0.00010936, -0.00019408, 0.00015362, -0.00057444, 0.00051554, 0.00087191, -0.00073023, 0.00022811, -0.00026178, 0.00013847, -0.00012206, 0.00011613, -0.00012781, 0.00003824, -0.00005697, 0.00008494, -0.00006903, 0.00001498, -0.00002347, 0.00003760, -0.00003419, 0.00000708, -0.00001005, 0.00003024, -0.00001926, 0.00000054, -0.00002841, 0.00005367, -0.00001844, -0.00001167, },
		{ 0.00005319, -0.00006093, 0.00002982, -0.00005567, 0.00001955, -0.00004775, 0.00003620, -0.00002071, -0.00005824, 0.00005700, -0.00003123, 0.00014586, -0.00003464, 0.00007382, -0.00010092, 0.00003702, -0.00001991, -0.00000401, -0.00002189, 0.00003893, -0.00000601, -0.00000507, -0.00002490, 0.00002560, -0.00002571, 0.00002479, -0.00003694, 0.00004200, -0.00000674, -0.00002310, -0.00002986, 0.00005475, -0.00002274, -0.00001164, -0.00001202, -0.00001437, -0.00006262, -0.00010518, -0.00037937, 0.00129547, -0.00037721, -0.00013139, -0.00002874, -0.00006319, 0.00001890, -0.00000797, 0.00000882, -0.00007524, 0.00001171, 0.00000672, 0.00002584, -0.00005880, 0.00002048, -0.00000679, 0.00001968, -0.00003681, 0.00001548, 0.00000120, 0.00001795, -0.00002482, -0.00000613, 0.00000378, 0.00004691, -0.00006170, },
		{ 0.00001139, -0.00000526, -0.00002584, 0.00000725, -0.00005253, 0.00002957, -0.00006189, 0.00008323, -0.00012891, 0.00007545, -0.00005749, 0.00013071, 0.00002116, 0.00004875, -0.00004364, -0.00002902, 0.00001648, -0.00001563, -0.00002244, 0.00002389, 0.00000834, 0.00000560, -0.00004034, 0.00003031, -0.00002855, 0.00003000, -0.00003797, 0.00003094, -0.00000253, 0.00001074, -0.00007511, 0.00006963, -0.00002905, 0.00003032, -0.00008044, 0.00008277, -0.00016836, 0.00011837, -0.00057314, 0.00073911, 0.00054992, -0.00059118, 0.00018158, -0.00022121, 0.00011460, -0.00010152, 0.00010220, -0.00012654, 0.00003280, -0.00004114, 0.00007667, -0.00007608, 0.00002064, -0.00002262, 0.00004116, -0.00004104, 0.00000932, -0.00000596, 0.00002957, -0.00002275, -0.00000353, -0.00001962, 0.00005938, -0.00003997, },
		{ -0.00004744, 0.00004428, -0.00006035, 0.00003876, -0.00007527, 0.00003685, -0.00006128, 0.00005249, -0.00004932, -0.00003073, 0.00002085, 0.00001670, 0.00013280, -0.00000746, 0.00003928, -0.00009212, 0.00004007, -0.00002856, -0.00000384, -0.00001423, 0.00003731, -0.00000551, -0.00001574, -0.00001110, 0.00001287, -0.00001125, 0.00000979, -0.00002196, 0.00003505, -0.00000477, -0.00003806, -0.00000563, 0.00003920, -0.00000959, -0.00003173, 0.00001620, -0.00006183, 0.00000457, -0.00023709, -0.00011452, 0.00113268, -0.00046160, -0.00002951, -0.00008614, -0.00001001, -0.00002220, 0.00003092, -0.00002984, -0.00005166, -0.00000315, 0.00003205, -0.00000610, -0.00003865, 0.00000816, 0.00000974, 0.00000525, -0.00003055, 0.00001283, 0.00000951, 0.00000864, -0.00002287, -0.00001094, 0.00002456, 0.00002175, },
		{ -0.00005154, 0.00002722, -0.00002727, 0.00000104, -0.00002086, -0.00003453, 0.00001612, -0.00004464, 0.00006011, -0.00011459, 0.00005871, -0.00004680, 0.00015513, 0.00002212, 0.00005061, -0.00007336, -0.00000157, -0.00000319, -0.00001027, -0.00002327, 0.00003113, 0.00000548, -0.00000159, -0.00003534, 0.00002876, -0.00002610, 0.00002867, -0.00003836, 0.00003663, -0.00000481, -0.00000186, -0.00006143, 0.00006699, -0.00002231, 0.00001241, -0.00005420, 0.00004305, -0.00011985, 0.00004273, -0.00049046, 0.00086003, 0.00027633, -0.00044382, 0.00011899, -0.00016739, 0.00007784, -0.00006874, 0.00007572, -0.00011555, 0.00002204, -0.00001835, 0.00005645, -0.00007168, 0.00001901, -0.00001512, 0.00003868, -0.00004497, 0.00001098, -0.00000057, 0.00002615, -0.00002468, -0.00000868, -0.00000495, 0.00005408, },
		{ -0.00000354, -0.00002807, 0.00002639, -0.00004617, 0.00003290, -0.00008304, 0.00004647, -0.00006705, 0.00005933, -0.00006868, -0.00001230, -0.00000921, 0.00006562, 0.00012294, 0.00001518, -0.00000373, -0.00006907, 0.00003071, -0.00002957, -0.00000615, -0.00000490, 0.00003453, -0.00000720, -0.00002279, -0.00000007, 0.00000302, 0.00000055, -0.00000351, -0.00000699, 0.00002721, -0.00000593, -0.00004590, 0.00001282, 0.00002893, -0.00000585, -0.00003771, 0.00002444, -0.00008184, 0.00003534, -0.00032187, 0.00013175, 0.00091598, -0.00046767, 0.00002166, -0.00011026, 0.00001662, -0.00004182, 0.00005182, -0.00005579, -0.00003742, -0.00000586, 0.00004330, -0.00002877, -0.00002267, -0.00000016, 0.00002351, -0.00000930, -0.00002393, 0.00001206, 0.00001493, 0.00000085, -0.00002316, -0.00000968, 0.00003930, },
		{ 0.00003969, -0.00005232, 0.00003255, -0.00003927, 0.00002234, -0.00004959, -0.00001218, -0.00000058, -0.00002460, 0.00003194, -0.00009302, 0.00003060, -0.00002117, 0.00017153, 0.00003333, 0.00003061, -0.00008442, 0.00001585, -0.00001937, -0.00000508, -0.00002271, 0.00003712, 0.00000192, -0.00000934, -0.00002746, 0.00002399, -0.00001965, 0.00002226, -0.00003307, 0.00003781, -0.00000674, -0.00001385, -0.00004498, 0.00006167, -0.00001632, -0.00000365, -0.00002979, 0.00000357, -0.00006637, -0.00005036, -0.00034795, 0.00088052, 0.00007193, -0.00031340, 0.00005946, -0.00011527, 0.00004080, -0.00003442, 0.00004533, -0.00010150, 0.00001255, 0.00000366, 0.00003124, -0.00006010, 0.00001270, -0.00000278, 0.00003063, -0.00004583, 0.00001255, 0.00000487, 0.00002128, -0.00002639, -0.00001246, 0.00001220, },
		{ 0.00004534, -0.00002400, -0.00000938, 0.00000565, -0.00002319, 0.00001399, -0.00007812, 0.00004682, -0.00006373, 0.00005492, -0.00007593, -0.00000715, -0.00002797, 0.00011135, 0.00012157, 0.00002271, -0.00004321, -0.00004009, 0.00001327, -0.00002424, -0.00001078, 0.00000595, 0.00003030, -0.00001071, -0.00002586, 0.00000735, -0.00000269, 0.00000775, -0.00001280, 0.00000575, 0.00001940, -0.00000949, -0.00004724, 0.00002567, 0.00002276, -0.00000911, -0.00003259, 0.00001581, -0.00007614, 0.00002753, -0.00034597, 0.00032647, 0.00068568, -0.00042226, 0.00003493, -0.00010872, 0.00002203, -0.00004351, 0.00005671, -0.00007017, -0.00003056, 0.00000117, 0.00004089, -0.00004012, -0.00001329, -0.00000235, 0.00003220, -0.00002239, -0.00001706, 0.00001261, 0.00001773, -0.00000565, -0.00002497, -0.00000240, },
		{ 0.00002789, 0.00002009, -0.00004442, 0.00002754, -0.00003848, 0.00003220, -0.00007206, 0.00000856, -0.00001536, -0.00000738, 0.00000498, -0.00007085, -0.00000265, 0.00001461, 0.00018629, 0.00004543, -0.00000528, -0.00007654, 0.00001982, -0.00002804, -0.00000284, -0.00001941, 0.00004108, -0.00000255, -0.00001623, -0.00001863, 0.00001827, -0.00001187, 0.00001327, -0.00002397, 0.00003522, -0.00000875, -0.00002344, -0.00002841, 0.00005600, -0.00001390, -0.00001332, -0.00001427, -0.00002464, -0.00002461, -0.00013568, -0.00017658, 0.00081932, -0.00005953, -0.00021322, 0.00001443, -0.00007430, 0.00001197, -0.00000629, 0.00001753, -0.00008914, 0.00000895, 0.00001853, 0.00000797, -0.00004651, 0.00000539, 0.00001153, 0.00001839, -0.00004370, 0.00001416, 0.00000952, 0.00001590, -0.00002872, -0.00001276, },
		{ 0.00000922, 0.00004220, -0.00003645, 0.00000377, -0.00001135, 0.00000146, -0.00001271, -0.00006326, 0.00003958, -0.00005322, 0.00004101, -0.00007204, -0.00001562, -0.00003392, 0.00015404, 0.00012834, 0.00000934, -0.00006880, -0.00001444, -0.00000591, -0.00001589, -0.00001620, 0.00001746, 0.00002453, -0.00001531, -0.00002541, 0.00001144, -0.00000439, 0.00000992, -0.00001697, 0.00001478, 0.00001226, -0.00001442, -0.00004357, 0.00003418, 0.00001833, -0.00001519, -0.00002223, -0.00000198, -0.00005369, -0.00001017, -0.00031063, 0.00045122, 0.00047455, -0.00035240, 0.00002636, -0.00009264, 0.00001506, -0.00003428, 0.00005082, -0.00007686, -0.00002672, 0.00001258, 0.00002888, -0.00004143, -0.00001044, 0.00000199, 0.00003448, -0.00003267, -0.00001016, 0.00001380, 0.00001855, -0.00001150, -0.00002676, },
		{ -0.00000845, 0.00003858, 0.00000001, -0.00003203, 0.00001561, -0.00002620, 0.00002855, -0.00008404, 0.00002342, -0.00002458, 0.00000287, -0.00001569, -0.00005396, -0.00003519, 0.00005634, 0.00020532, 0.00004700, -0.00004600, -0.00005522, 0.00001159, -0.00002821, -0.00000481, -0.00001263, 0.00004235, -0.00000783, -0.00002124, -0.00001034, 0.00001346, -0.00000526, 0.00000448, -0.00001350, 0.00003013, -0.00001112, -0.00002969, -0.00001328, 0.00005085, -0.00001559, -0.00001539, -0.00001018, -0.00003606, -0.00000539, -0.00019242, -0.00000873, 0.00070506, -0.00012783, -0.00014524, -0.00001297, -0.00004737, -0.00000558, 0.00001234, -0.00000469, -0.00008006, 0.00001242, 0.00002307, -0.00000837, -0.00003538, 0.00000068, 0.00002450, 0.00000394, -0.00003893, 0.00001571, 0.00001304, 0.00001041, -0.00003178, },
		{ -0.00002668, 0.00002228, 0.00003164, -0.00004028, 0.00000887, -0.00001967, 0.00002078, -0.00004047, -0.00004335, 0.00002850, -0.00003959, 0.00002184, -0.00006091, -0.00003489, -0.00002866, 0.00019782, 0.00013663, -0.00002359, -0.00007512, 0.00000080, -0.00002076, -0.00000845, -0.00002025, 0.00002830, 0.00001737, -0.00002010, -0.00002242, 0.00001327, -0.00000345, 0.00000807, -0.00001635, 0.00001962, 0.00000609, -0.00001958, -0.00003663, 0.00003982, 0.00001331, -0.00001992, -0.00001295, -0.00001964, -0.00002719, -0.00006158, -0.00023007, 0.00050378, 0.00030280, -0.00027928, 0.00000999, -0.00007229, 0.00000431, -0.00002150, 0.00003999, -0.00008016, -0.00002130, 0.00002214, 0.00001333, -0.00003651, -0.00001192, 0.00001143, 0.00003029, -0.00003919, -0.00000356, 0.00001507, 0.00001803, -0.00001740, },
		{ -0.00003489, 0.00000023, 0.00004214, -0.00001636, -0.00002001, 0.00000215, -0.00000712, 0.00001334, -0.00008482, 0.00003104, -0.00002725, 0.00000488, -0.00002788, -0.00004560, -0.00006361, 0.00010303, 0.00023055, 0.00002881, -0.00007901, -0.00002961, -0.00000377, -0.00002190, -0.00001036, -0.00000270, 0.00004067, -0.00001361, -0.00002391, -0.00000338, 0.00001052, -0.00000149, -0.00000189, -0.00000398, 0.00002391, -0.00001394, -0.00003245, -0.00000008, 0.00004577, -0.00001981, -0.00001186, -0.00001539, -0.00003175, -0.00001092, -0.00020933, 0.00012930, 0.00056823, -0.00015045, -0.00010351, -0.00002590, -0.00003197, -0.00001356, 0.00002226, -0.00002171, -0.00007273, 0.00002062, 0.00001796, -0.00001592, -0.00002924, 0.00000105, 0.00003322, -0.00001045, -0.00003218, 0.00001698, 0.00001545, 0.00000473, },
		{ -0.00002369, -0.00002322, 0.00003343, 0.00001710, -0.00003747, 0.00000653, -0.00001761, 0.00003002, -0.00006338, -0.00002330, 0.00001753, -0.00002711, 0.00000225, -0.00004749, -0.00006089, -0.00001479, 0.00024843, 0.00013518, -0.00006756, -0.00006340, 0.00000293, -0.00002754, -0.00000492, -0.00002105, 0.00003713, 0.00000916, -0.00002423, -0.00001803, 0.00001426, -0.00000185, 0.00000423, -0.00001245, 0.00002081, 0.00000082, -0.00002398, -0.00002794, 0.00004358, 0.00000642, -0.00002066, -0.00000919, -0.00002967, -0.00000862, -0.00010830, -0.00012658, 0.00049573, 0.00017706, -0.00021530, -0.00000537, -0.00005415, -0.00000454, -0.00001032, 0.00002848, -0.00008283, -0.00001150, 0.00002503, 0.00000023, -0.00003010, -0.00001420, 0.00002312, 0.00002085, -0.00004162, 0.00000239, 0.00001608, 0.00001660, },
		{ -0.00000144, -0.00003676, 0.00001168, 0.00003838, -0.00002648, -0.00001221, -0.00000742, 0.00001238, -0.00000857, -0.00007680, 0.00003274, -0.00002508, 0.00000027, -0.00003259, -0.00004580, -0.00008767, 0.00015792, 0.00025748, -0.00001229, -0.00009446, -0.00000925, -0.00001942, -0.00001314, -0.00001732, 0.00000910, 0.00003624, -0.00001948, -0.00002431, 0.00000228, 0.00000930, -0.00000090, -0.00000479, 0.00000303, 0.00001772, -0.00001711, -0.00003212, 0.00001129, 0.00003963, -0.00002392, -0.00000672, -0.00002436, -0.00001833, -0.00003492, -0.00018671, 0.00022240, 0.00043418, -0.00014605, -0.00007888, -0.00003039, -0.00002299, -0.00001628, 0.00002669, -0.00003592, -0.00006388, 0.00002882, 0.00000717, -0.00001608, -0.00002806, 0.00000711, 0.00003588, -0.00002272, -0.00002428, 0.00001780, 0.00001698, },
		{ 0.00001442, -0.00003018, -0.00001571, 0.00003980, 0.00000264, -0.00003170, 0.00000008, -0.00000708, 0.00002806, -0.00007812, -0.00000633, 0.00000905, -0.00001840, -0.00001448, -0.00003548, -0.00009066, 0.00000698, 0.00030954, 0.00011198, -0.00010931, -0.00004086, -0.00000569, -0.00002601, -0.00000630, -0.00001774, 0.00004301, 0.00000026, -0.00002700, -0.00001320, 0.00001551, -0.00000145, 0.00000072, -0.00000744, 0.00001959, -0.00000388, -0.00002691, -0.00001855, 0.00004569, -0.00000214, -0.00001715, -0.00001207, -0.00002877, -0.00000540, -0.00013529, -0.00002391, 0.00044695, 0.00009301, -0.00016429, -0.00001684, -0.00003997, -0.00000988, -0.00000254, 0.00001777, -0.00008504, 0.00000245, 0.00001944, -0.00000640, -0.00002627, -0.00001376, 0.00003361, 0.00000833, -0.00004029, 0.00000742, 0.00001669, },
		{ 0.00001782, -0.00000839, -0.00003597, 0.00002328, 0.00002908, -0.00003019, -0.00001018, -0.00000952, 0.00002648, -0.00003136, -0.00006392, 0.00003128, -0.00002114, -0.00000759, -0.00003291, -0.00005241, -0.00010920, 0.00022759, 0.00027507, -0.00007120, -0.00008862, -0.00000067, -0.00002935, -0.00000625, -0.00002303, 0.00002101, 0.00002956, -0.00002493, -0.00002288, 0.00000720, 0.00000866, -0.00000253, -0.00000451, 0.00000703, 0.00001216, -0.00002026, -0.00002936, 0.00002115, 0.00003145, -0.00002540, -0.00000403, -0.00003060, -0.00000516, -0.00006533, -0.00013510, 0.00026820, 0.00031879, -0.00012947, -0.00006331, -0.00003189, -0.00001562, -0.00001804, 0.00002908, -0.00004974, -0.00005044, 0.00003190, -0.00000371, -0.00001266, -0.00002957, 0.00001744, 0.00003215, -0.00003142, -0.00001612, 0.00001808, },
		},
		{ { -0.00001122, -0.00000616, 0.00000704, 0.00000018, -0.00000291, 0.00000251, -0.00000938, 0.00000687, -0.00000970, -0.00002032, 0.00002842, -0.00001602, -0.00002665, -0.00002792, -0.00014625, 0.00006468, 0.00033770, 0.00006468, -0.00014625, -0.00002792, -0.00002665, -0.00001602, 0.00002842, -0.00002032, -0.00000970, 0.00000687, -0.00000938, 0.00000251, -0.00000291, 0.00000018, 0.00000704, -0.00000616, -0.00001122, -0.00000616, 0.00000704, 0.00000018, -0.00000291, 0.00000251, -0.00000938, 0.00000687, -0.00000970, -0.00002032, 0.00002842, -0.00001602, -0.00002665, -0.00002792, -0.00014625, 0.00006468, 0.00033770, 0.00006468, -0.00014625, -0.00002792, -0.00002665, -0.00001602, 0.00002842, -0.00002032, -0.00000970, 0.00000687, -0.00000938, 0.00000251, -0.00000291, 0.00000018, 0.00000704, -0.00000616, },
		{ -0.00000873, -0.00001166, 0.00000318, 0.00000390, -0.00000338, 0.00000110, -0.00000150, -0.00000698, 0.00000805, -0.00002835, 0.00001122, 0.00001329, -0.00003452, -0.00001327, -0.00008733, -0.00012113, 0.00027527, 0.00027527, -0.00012113, -0.00008733, -0.00001327, -0.00003452, 0.00001329, 0.00001122, -0.00002835, 0.00000805, -0.00000698, -0.00000150, 0.00000110, -0.00000338, 0.00000390, 0.00000318, -0.00001166, -0.00000873, -0.00000116, 0.00000921, -0.00000631, 0.00000371, -0.00000727, -0.00000061, 0.00000527, -0.00002363, 0.00000322, 0.00002217, -0.00004065, -0.00000277, -0.00010072, -0.00007796, 0.00023360, 0.00023360, -0.00007796, -0.00010072, -0.00000277, -0.00004065, 0.00002217, 0.00000322, -0.00002363, 0.00000527, -0.00000061, -0.00000727, 0.00000371, -0.00000631, 0.00000921, -0.00000116, },
		{ -0.00000719, -0.00001122, -0.00000504, 0.00000649, -0.00000157, -0.00000265, 0.00000458, -0.00000927, 0.00000386, -0.00001056, -0.00001778, 0.00002826, -0.00001899, -0.00002776, -0.00002681, -0.00015325, 0.00003233, 0.00042707, 0.00003233, -0.00015325, -0.00002681, -0.00002776, -0.00001899, 0.00002826, -0.00001778, -0.00001056, 0.00000386, -0.00000927, 0.00000458, -0.00000265, -0.00000157, 0.00000649, -0.00000504, -0.00001122, -0.00000719, 0.00000748, 0.00000187, -0.00000318, 0.00000053, -0.00000910, 0.00000906, -0.00000883, -0.00002163, 0.00002770, -0.00001336, -0.00002554, -0.00002846, -0.00013630, 0.00008685, 0.00026239, 0.00008685, -0.00013630, -0.00002846, -0.00002554, -0.00001336, 0.00002770, -0.00002163, -0.00000883, 0.00000906, -0.00000910, 0.00000053, -0.00000318, 0.00000187, 0.00000748, },
		{ -0.00000527, -0.00000583, -0.00001459, 0.00000769, -0.00000184, -0.00000001, -0.00000211, 0.00000533, -0.00001432, 0.00001101, -0.00003328, 0.00002101, 0.00000144, -0.00002568, -0.00002733, -0.00006681, -0.00017323, 0.00032113, 0.00032113, -0.00017323, -0.00006681, -0.00002733, -0.00002568, 0.00000144, 0.00002101, -0.00003328, 0.00001101, -0.00001432, 0.00000533, -0.00000211, -0.00000001, -0.00000184, 0.00000769, -0.00001459, -0.00000583, -0.00000527, 0.00001404, -0.00000875, 0.00000570, -0.00001193, 0.00000472, 0.00000269, -0.00001917, -0.00000291, 0.00002799, -0.00004402, 0.00000406, -0.00010684, -0.00004415, 0.00019651, 0.00019651, -0.00004415, -0.00010684, 0.00000406, -0.00004402, 0.00002799, -0.00000291, -0.00001917, 0.00000269, 0.00000472, -0.00001193, 0.00000570, -0.00000875, 0.00001404, },
		{ 0.00000780, -0.00000810, -0.00001122, -0.00000383, 0.00000583, -0.00000334, -0.00000239, 0.00000672, -0.00000878, 0.00000007, -0.00001142, -0.00001407, 0.00002724, -0.00002226, -0.00002885, -0.00002515, -0.00015728, -0.00000971, 0.00052946, -0.00000971, -0.00015728, -0.00002515, -0.00002885, -0.00002226, 0.00002724, -0.00001407, -0.00001142, 0.00000007, -0.00000878, 0.00000672, -0.00000239, -0.00000334, 0.00000583, -0.00000383, -0.00001122, -0.00000810, 0.00000780, 0.00000349, -0.00000343, -0.00000134, -0.00000843, 0.00001041, -0.00000797, -0.00002174, 0.00002613, -0.00001105, -0.00002445, -0.00002845, -0.00012354, 0.00009880, 0.00020169, 0.00009880, -0.00012354, -0.00002845, -0.00002445, -0.00001105, 0.00002613, -0.00002174, -0.00000797, 0.00001041, -0.00000843, -0.00000134, -0.00000343, 0.00000349, },
		{ 0.00001837, -0.00000912, -0.00000297, -0.00001749, 0.00001234, -0.00000794, 0.00000377, -0.00000588, 0.00001310, -0.00002254, 0.00001412, -0.00003838, 0.00003245, -0.00001316, -0.00001430, -0.00004471, -0.00003956, -0.00023350, 0.00037060, 0.00037060, -0.00023350, -0.00003956, -0.00004471, -0.00001430, -0.00001316, 0.00003245, -0.00003838, 0.00001412, -0.00002254, 0.00001310, -0.00000588, 0.00000377, -0.00000794, 0.00001234, -0.00001749, -0.00000297, -0.00000912, 0.00001837, -0.00001072, 0.00000706, -0.00001547, 0.00000899, 0.00000035, -0.00001500, -0.00000717, 0.00003080, -0.00004467, 0.00000726, -0.00010585, -0.00001975, 0.00016424, 0.00016424, -0.00001975, -0.00010585, 0.00000726, -0.00004467, 0.00003080, -0.00000717, -0.00001500, 0.00000035, 0.00000899, -0.00001547, 0.00000706, -0.00001072, },
		{ 0.00000503, 0.00000800, -0.00000891, -0.00001122, -0.00000255, 0.00000507, -0.00000514, -0.00000214, 0.00000891, -0.00000792, -0.00000442, -0.00001226, -0.00000926, 0.00002540, -0.00002577, -0.00002993, -0.00002299, -0.00015843, -0.00006063, 0.00064326, -0.00006063, -0.00015843, -0.00002299, -0.00002993, -0.00002577, 0.00002540, -0.00000926, -0.00001226, -0.00000442, -0.00000792, 0.00000891, -0.00000214, -0.00000514, 0.00000507, -0.00000255, -0.00001122, -0.00000891, 0.00000800, 0.00000503, -0.00000369, -0.00000308, -0.00000741, 0.00001096, -0.00000713, -0.00002067, 0.00002374, -0.00000909, -0.00002337, -0.00002789, -0.00010821, 0.00010083, 0.00015556, 0.00010083, -0.00010821, -0.00002789, -0.00002337, -0.00000909, 0.00002374, -0.00002067, -0.00000713, 0.00001096, -0.00000741, -0.00000308, -0.00000369, },
		{ -0.00001220, 0.00002216, -0.00001270, -0.00000020, -0.00002032, 0.00001706, -0.00001431, 0.00000789, -0.00001012, 0.00002170, -0.00003150, 0.00001734, -0.00004357, 0.00004531, -0.00003023, -0.00000064, -0.00006505, -0.00000626, -0.00030085, 0.00042296, 0.00042296, -0.00030085, -0.00000626, -0.00006505, -0.00000064, -0.00003023, 0.00004531, -0.00004357, 0.00001734, -0.00003150, 0.00002170, -0.00001012, 0.00000789, -0.00001431, 0.00001706, -0.00002032, -0.00000020, -0.00001270, 0.00002216, -0.00001220, 0.00000782, -0.00001791, 0.00001222, -0.00000176, -0.00001116, -0.00000962, 0.00003073, -0.00004275, 0.00000701, -0.00009815, -0.00000448, 0.00013684, 0.00013684, -0.00000448, -0.00009815, 0.00000701, -0.00004275, 0.00003073, -0.00000962, -0.00001116, -0.00000176, 0.00001222, -0.00001791, 0.00000782, },
		{ -0.00000394, 0.00000648, 0.00000809, -0.00000960, -0.00001122, -0.00000122, 0.00000422, -0.00000692, -0.00000189, 0.00001111, -0.00000673, -0.00000953, -0.00001308, -0.00000350, 0.00002283, -0.00002946, -0.00003097, -0.00002037, -0.00015688, -0.00011921, 0.00076642, -0.00011921, -0.00015688, -0.00002037, -0.00003097, -0.00002946, 0.00002283, -0.00000350, -0.00001308, -0.00000953, -0.00000673, 0.00001111, -0.00000189, -0.00000692, 0.00000422, -0.00000122, -0.00001122, -0.00000960, 0.00000809, 0.00000648, -0.00000394, -0.00000469, -0.00000606, 0.00001075, -0.00000631, -0.00001854, 0.00002064, -0.00000747, -0.00002233, -0.00002684, -0.00009068, 0.00009364, 0.00012349, 0.00009364, -0.00009068, -0.00002684, -0.00002233, -0.00000747, 0.00002064, -0.00001854, -0.00000631, 0.00001075, -0.00000606, -0.00000469, },
		{ 0.00000801, -0.00001322, 0.00002543, -0.00001597, 0.00000247, -0.00002307, 0.00002179, -0.00002085, 0.00001227, -0.00001475, 0.00003095, -0.00004103, 0.00002062, -0.00004880, 0.00005933, -0.00004933, 0.00001494, -0.00008785, 0.00003220, -0.00037388, 0.00047735, 0.00047735, -0.00037388, 0.00003220, -0.00008785, 0.00001494, -0.00004933, 0.00005933, -0.00004880, 0.00002062, -0.00004103, 0.00003095, -0.00001475, 0.00001227, -0.00002085, 0.00002179, -0.00002307, 0.00000247, -0.00001597, 0.00002543, -0.00001322, 0.00000801, -0.00001933, 0.00001446, -0.00000362, -0.00000766, -0.00001038, 0.00002803, -0.00003850, 0.00000360, -0.00008443, 0.00000226, 0.00011423, 0.00011423, 0.00000226, -0.00008443, 0.00000360, -0.00003850, 0.00002803, -0.00001038, -0.00000766, -0.00000362, 0.00001446, -0.00001933, },
		{ -0.00000616, -0.00000417, 0.00000783, 0.00000807, -0.00001018, -0.00001122, 0.00000014, 0.00000331, -0.00000868, -0.00000165, 0.00001331, -0.00000526, -0.00001512, -0.00001386, 0.00000305, 0.00001962, -0.00003328, -0.00003197, -0.00001739, -0.00015293, -0.00018396, 0.00089644, -0.00018396, -0.00015293, -0.00001739, -0.00003197, -0.00003328, 0.00001962, 0.00000305, -0.00001386, -0.00001512, -0.00000526, 0.00001331, -0.00000165, -0.00000868, 0.00000331, 0.00000014, -0.00001122, -0.00001018, 0.00000807, 0.00000783, -0.00000417, -0.00000616, -0.00000443, 0.00000986, -0.00000553, -0.00001548, 0.00001693, -0.00000619, -0.00002133, -0.00002535, -0.00007139, 0.00007823, 0.00010446, 0.00007823, -0.00007139, -0.00002535, -0.00002133, -0.00000619, 0.00001693, -0.00001548, -0.00000553, 0.00000986, -0.00000443, },
		{ -0.00001981, 0.00000771, -0.00001383, 0.00002819, -0.00001894, 0.00000501, -0.00002570, 0.00002647, -0.00002746, 0.00001682, -0.00001966, 0.00004065, -0.00005093, 0.00002391, -0.00005397, 0.00007418, -0.00006998, 0.00003201, -0.00011252, 0.00007468, -0.00045092, 0.00053278, 0.00053278, -0.00045092, 0.00007468, -0.00011252, 0.00003201, -0.00006998, 0.00007418, -0.00005397, 0.00002391, -0.00005093, 0.00004065, -0.00001966, 0.00001682, -0.00002746, 0.00002647, -0.00002570, 0.00000501, -0.00001894, 0.00002819, -0.00001383, 0.00000771, -0.00001981, 0.00001578, -0.00000525, -0.00000451, -0.00000964, 0.00002303, -0.00003225, -0.00000255, -0.00006556, 0.00000135, 0.00009616, 0.00009616, 0.00000135, -0.00006556, -0.00000255, -0.00003225, 0.00002303, -0.00000964, -0.00000451, -0.00000525, 0.00001578, },
		{ -0.00000260, -0.00000749, -0.00000440, 0.00000907, 0.00000797, -0.00001066, -0.00001122, 0.00000150, 0.00000237, -0.00001039, -0.00000143, 0.00001546, -0.00000357, -0.00002105, -0.00001461, 0.00001020, 0.00001591, -0.00003714, -0.00003292, -0.00001413, -0.00014694, -0.00025310, 0.00103046, -0.00025310, -0.00014694, -0.00001413, -0.00003292, -0.00003714, 0.00001591, 0.00001020, -0.00001461, -0.00002105, -0.00000357, 0.00001546, -0.00000143, -0.00001039, 0.00000237, 0.00000150, -0.00001122, -0.00001066, 0.00000797, 0.00000907, -0.00000440, -0.00000749, -0.00000260, 0.00000840, -0.00000478, -0.00001164, 0.00001274, -0.00000522, -0.00002038, -0.00002352, -0.00005084, 0.00005592, 0.00009707, 0.00005592, -0.00005084, -0.00002352, -0.00002038, -0.00000522, 0.00001274, -0.00001164, -0.00000478, 0.00000840, },
		{ 0.00001631, -0.00001949, 0.00000698, -0.00001407, 0.00003047, -0.00002159, 0.00000740, -0.00002818, 0.00003103, -0.00003402, 0.00002145, -0.00002473, 0.00005056, -0.00006099, 0.00002714, -0.00005902, 0.00008950, -0.00009161, 0.00005006, -0.00013837, 0.00011990, -0.00053012, 0.00058819, 0.00058819, -0.00053012, 0.00011990, -0.00013837, 0.00005006, -0.00009161, 0.00008950, -0.00005902, 0.00002714, -0.00006099, 0.00005056, -0.00002473, 0.00002145, -0.00003402, 0.00003103, -0.00002818, 0.00000740, -0.00002159, 0.00003047, -0.00001407, 0.00000698, -0.00001949, 0.00001631, -0.00000664, -0.00000171, -0.00000764, 0.00001617, -0.00002440, -0.00001095, -0.00004260, -0.00000606, 0.00008227, 0.00008227, -0.00000606, -0.00004260, -0.00001095, -0.00002440, 0.00001617, -0.00000764, -0.00000171, -0.00000664, },
		{ 0.00000647, -0.00000062, -0.00000867, -0.00000461, 0.00001019, 0.00000780, -0.00001104, -0.00001122, 0.00000285, 0.00000140, -0.00001202, -0.00000122, 0.00001753, -0.00000173, -0.00002715, -0.00001530, 0.00001770, 0.00001185, -0.00004097, -0.00003381, -0.00001070, -0.00013937, -0.00032463, 0.00116538, -0.00032463, -0.00013937, -0.00001070, -0.00003381, -0.00004097, 0.00001185, 0.00001770, -0.00001530, -0.00002715, -0.00000173, 0.00001753, -0.00000122, -0.00001202, 0.00000140, 0.00000285, -0.00001122, -0.00001104, 0.00000780, 0.00001019, -0.00000461, -0.00000867, -0.00000062, 0.00000647, -0.00000409, -0.00000724, 0.00000823, -0.00000452, -0.00001949, -0.00002142, -0.00002964, 0.00002823, 0.00009955, 0.00002823, -0.00002964, -0.00002142, -0.00001949, -0.00000452, 0.00000823, -0.00000724, -0.00000409, },
		{ -0.00000781, 0.00001618, -0.00001852, 0.00000592, -0.00001401, 0.00003231, -0.00002393, 0.00000961, -0.00003050, 0.00003539, -0.00004040, 0.00002604, -0.00002982, 0.00006046, -0.00007097, 0.00003028, -0.00006387, 0.00010487, -0.00011360, 0.00006855, -0.00016466, 0.00016644, -0.00060946, 0.00064244, 0.00064244, -0.00060946, 0.00016644, -0.00016466, 0.00006855, -0.00011360, 0.00010487, -0.00006387, 0.00003028, -0.00007097, 0.00006046, -0.00002982, 0.00002604, -0.00004040, 0.00003539, -0.00003050, 0.00000961, -0.00002393, 0.00003231, -0.00001401, 0.00000592, -0.00001852, 0.00001618, -0.00000781, 0.00000074, -0.00000463, 0.00000791, -0.00001539, -0.00002101, -0.00001675, -0.00001866, 0.00007209, 0.00007209, -0.00001866, -0.00001675, -0.00002101, -0.00001539, 0.00000791, -0.00000463, 0.00000074, },
		{ -0.00000345, 0.00000423, 0.00000144, -0.00000970, -0.00000480, 0.00001120, 0.00000758, -0.00001134, -0.00001122, 0.00000414, 0.00000044, -0.00001356, -0.00000102, 0.00001950, 0.00000019, -0.00003324, -0.00001595, 0.00002532, 0.00000761, -0.00004469, -0.00003464, -0.00000722, -0.00013069, -0.00039645, 0.00129790, -0.00039645, -0.00013069, -0.00000722, -0.00003464, -0.00004469, 0.00000761, 0.00002532, -0.00001595, -0.00003324, 0.00000019, 0.00001950, -0.00000102, -0.00001356, 0.00000044, 0.00000414, -0.00001122, -0.00001134, 0.00000758, 0.00001120, -0.00000480, -0.00000970, 0.00000144, 0.00000423, -0.00000345, -0.00000248, 0.00000356, -0.00000406, -0.00001867, -0.00001916, -0.00000838, -0.00000315, 0.00010991, -0.00000315, -0.00000838, -0.00001916, -0.00001867, -0.00000406, 0.00000356, -0.00000248, },
		{ 0.00000286, -0.00000877, 0.00001552, -0.00001706, 0.00000464, -0.00001370, 0.00003375, -0.00002596, 0.00001163, -0.00003263, 0.00003948, -0.00004647, 0.00003048, -0.00003481, 0.00007008, -0.00008062, 0.00003324, -0.00006842, 0.00011991, -0.00013529, 0.00008691, -0.00019061, 0.00021280, -0.00068685, 0.00069438, 0.00069438, -0.00068685, 0.00021280, -0.00019061, 0.00008691, -0.00013529, 0.00011991, -0.00006842, 0.00003324, -0.00008062, 0.00007008, -0.00003481, 0.00003048, -0.00004647, 0.00003948, -0.00003263, 0.00001163, -0.00002596, 0.00003375, -0.00001370, 0.00000464, -0.00001706, 0.00001552, -0.00000877, 0.00000286, -0.00000091, -0.00000123, -0.00000570, -0.00003212, 0.00001068, -0.00003501, 0.00006509, 0.00006509, -0.00003501, 0.00001068, -0.00003212, -0.00000570, -0.00000123, -0.00000091, },
		{ 0.00000242, -0.00000286, 0.00000181, 0.00000348, -0.00001060, -0.00000498, 0.00001209, 0.00000732, -0.00001156, -0.00001122, 0.00000536, -0.00000049, -0.00001498, -0.00000085, 0.00002133, 0.00000212, -0.00003916, -0.00001653, 0.00003281, 0.00000334, -0.00004822, -0.00003538, -0.00000378, -0.00012143, -0.00046637, 0.00142471, -0.00046637, -0.00012143, -0.00000378, -0.00003538, -0.00004822, 0.00000334, 0.00003281, -0.00001653, -0.00003916, 0.00000212, 0.00002133, -0.00000085, -0.00001498, -0.00000049, 0.00000536, -0.00001122, -0.00001156, 0.00000732, 0.00001209, -0.00000498, -0.00001060, 0.00000348, 0.00000181, -0.00000286, 0.00000242, -0.00000108, -0.00000380, -0.00001792, -0.00001684, 0.00001228, -0.00003642, 0.00012599, -0.00003642, 0.00001228, -0.00001684, -0.00001792, -0.00000380, -0.00000108, },
		{ 0.00000321, 0.00000467, -0.00000956, 0.00001451, -0.00001528, 0.00000322, -0.00001323, 0.00003485, -0.00002770, 0.00001343, -0.00003454, 0.00004324, -0.00005211, 0.00003466, -0.00003954, 0.00007917, -0.00008972, 0.00003599, -0.00007261, 0.00013418, -0.00015605, 0.00010455, -0.00021545, 0.00025748, -0.00076021, 0.00074285, 0.00074285, -0.00076021, 0.00025748, -0.00021545, 0.00010455, -0.00015605, 0.00013418, -0.00007261, 0.00003599, -0.00008972, 0.00007917, -0.00003954, 0.00003466, -0.00005211, 0.00004324, -0.00003454, 0.00001343, -0.00002770, 0.00003485, -0.00001323, 0.00000322, -0.00001528, 0.00001451, -0.00000956, 0.00000467, 0.00000321, -0.00001072, 0.00000416, -0.00004364, 0.00003838, -0.00005360, 0.00006069, 0.00006069, -0.00005360, 0.00003838, -0.00004364, 0.00000416, -0.00001072, },
		{ -0.00000553, 0.00000723, -0.00000235, -0.00000065, 0.00000545, -0.00001136, -0.00000514, 0.00001286, 0.00000704, -0.00001171, -0.00001122, 0.00000649, -0.00000135, -0.00001626, -0.00000069, 0.00002298, 0.00000399, -0.00004472, -0.00001705, 0.00003991, -0.00000078, -0.00005147, -0.00003604, -0.00000053, -0.00011209, -0.00053224, 0.00154251, -0.00053224, -0.00011209, -0.00000053, -0.00003604, -0.00005147, -0.00000078, 0.00003991, -0.00001705, -0.00004472, 0.00000399, 0.00002298, -0.00000069, -0.00001626, -0.00000135, 0.00000649, -0.00001122, -0.00001171, 0.00000704, 0.00001286, -0.00000514, -0.00001136, 0.00000545, -0.00000065, -0.00000235, 0.00000723, -0.00000553, -0.00000369, -0.00001726, -0.00001457, 0.00003173, -0.00006977, 0.00014559, -0.00006977, 0.00003173, -0.00001457, -0.00001726, -0.00000369, },
		{ -0.00002004, 0.00000743, 0.00000616, -0.00001018, 0.00001328, -0.00001337, 0.00000177, -0.00001266, 0.00003566, -0.00002915, 0.00001501, -0.00003622, 0.00004660, -0.00005720, 0.00003848, -0.00004389, 0.00008749, -0.00009802, 0.00003846, -0.00007636, 0.00014728, -0.00017523, 0.00012091, -0.00023840, 0.00029900, -0.00082750, 0.00078676, 0.00078676, -0.00082750, 0.00029900, -0.00023840, 0.00012091, -0.00017523, 0.00014728, -0.00007636, 0.00003846, -0.00009802, 0.00008749, -0.00004389, 0.00003848, -0.00005720, 0.00004660, -0.00003622, 0.00001501, -0.00002915, 0.00003566, -0.00001266, 0.00000177, -0.00001337, 0.00001328, -0.00001018, 0.00000616, 0.00000743, -0.00002004, 0.00001370, -0.00005493, 0.00006503, -0.00007292, 0.00005830, 0.00005830, -0.00007292, 0.00006503, -0.00005493, 0.00001370, },
		{ -0.00000369, -0.00000962, 0.00001172, -0.00000190, -0.00000301, 0.00000726, -0.00001199, -0.00000527, 0.00001351, 0.00000677, -0.00001182, -0.00001122, 0.00000748, -0.00000213, -0.00001738, -0.00000056, 0.00002444, 0.00000570, -0.00004975, -0.00001749, 0.00004639, -0.00000458, -0.00005436, -0.00003661, 0.00000245, -0.00010319, -0.00059200, 0.00164821, -0.00059200, -0.00010319, 0.00000245, -0.00003661, -0.00005436, -0.00000458, 0.00004639, -0.00001749, -0.00004975, 0.00000570, 0.00002444, -0.00000056, -0.00001738, -0.00000213, 0.00000748, -0.00001122, -0.00001182, 0.00000677, 0.00001351, -0.00000527, -0.00001199, 0.00000726, -0.00000301, -0.00000190, 0.00001172, -0.00000962, -0.00000369, -0.00001669, -0.00001245, 0.00004936, -0.00010146, 0.00016652, -0.00010146, 0.00004936, -0.00001245, -0.00001669, },
		{ 0.00002247, -0.00002867, 0.00001145, 0.00000738, -0.00001066, 0.00001199, -0.00001149, 0.00000039, -0.00001207, 0.00003624, -0.00003034, 0.00001635, -0.00003765, 0.00004950, -0.00006163, 0.00004183, -0.00004773, 0.00009482, -0.00010531, 0.00004060, -0.00007960, 0.00015885, -0.00019223, 0.00013545, -0.00025875, 0.00033599, -0.00088682, 0.00082509, 0.00082509, -0.00088682, 0.00033599, -0.00025875, 0.00013545, -0.00019223, 0.00015885, -0.00007960, 0.00004060, -0.00010531, 0.00009482, -0.00004773, 0.00004183, -0.00006163, 0.00004950, -0.00003765, 0.00001635, -0.00003034, 0.00003624, -0.00001207, 0.00000039, -0.00001149, 0.00001199, -0.00001066, 0.00000738, 0.00001145, -0.00002867, 0.00002247, -0.00006539, 0.00008940, -0.00009155, 0.00005733, 0.00005733, -0.00009155, 0.00008940, -0.00006539, },
		{ -0.00001622, -0.00000375, -0.00001320, 0.00001571, -0.00000153, -0.00000514, 0.00000884, -0.00001249, -0.00000538, 0.00001404, 0.00000653, -0.00001189, -0.00001122, 0.00000833, -0.00000280, -0.00001832, -0.00000044, 0.00002567, 0.00000721, -0.00005411, -0.00001786, 0.00005203, -0.00000792, -0.00005684, -0.00003708, 0.00000503, -0.00009520, -0.00064375, 0.00173895, -0.00064375, -0.00009520, 0.00000503, -0.00003708, -0.00005684, -0.00000792, 0.00005203, -0.00001786, -0.00005411, 0.00000721, 0.00002567, -0.00000044, -0.00001832, -0.00000280, 0.00000833, -0.00001122, -0.00001189, 0.00000653, 0.00001404, -0.00000538, -0.00001249, 0.00000884, -0.00000514, -0.00000153, 0.00001571, -0.00001320, -0.00000375, -0.00001622, -0.00001057, 0.00006461, -0.00012987, 0.00018677, -0.00012987, 0.00006461, -0.00001057, },
		{ -0.00007448, 0.00003003, -0.00003618, 0.00001502, 0.00000832, -0.00001102, 0.00001078, -0.00000978, -0.00000082, -0.00001151, 0.00003664, -0.00003127, 0.00001743, -0.00003881, 0.00005189, -0.00006530, 0.00004463, -0.00005094, 0.00010094, -0.00011140, 0.00004238, -0.00008227, 0.00016854, -0.00020653, 0.00014772, -0.00027588, 0.00036720, -0.00093648, 0.00085692, 0.00085692, -0.00093648, 0.00036720, -0.00027588, 0.00014772, -0.00020653, 0.00016854, -0.00008227, 0.00004238, -0.00011140, 0.00010094, -0.00005094, 0.00004463, -0.00006530, 0.00005189, -0.00003881, 0.00001743, -0.00003127, 0.00003664, -0.00001151, -0.00000082, -0.00000978, 0.00001078, -0.00001102, 0.00000832, 0.00001502, -0.00003618, 0.00003003, -0.00007448, 0.00011037, -0.00010817, 0.00005727, 0.00005727, -0.00010817, 0.00011037, },
		{ -0.00000902, -0.00001584, -0.00000384, -0.00001614, 0.00001900, -0.00000124, -0.00000692, 0.00001013, -0.00001288, -0.00000547, 0.00001445, 0.00000631, -0.00001194, -0.00001122, 0.00000901, -0.00000335, -0.00001906, -0.00000036, 0.00002665, 0.00000845, -0.00005764, -0.00001815, 0.00005662, -0.00001067, -0.00005882, -0.00003746, 0.00000714, -0.00008852, -0.00068583, 0.00181226, -0.00068583, -0.00008852, 0.00000714, -0.00003746, -0.00005882, -0.00001067, 0.00005662, -0.00001815, -0.00005764, 0.00000845, 0.00002665, -0.00000036, -0.00001906, -0.00000335, 0.00000901, -0.00001122, -0.00001194, 0.00000631, 0.00001445, -0.00000547, -0.00001288, 0.00001013, -0.00000692, -0.00000124, 0.00001900, -0.00001614, -0.00000384, -0.00001584, -0.00000902, 0.00007701, -0.00015357, 0.00020454, -0.00015357, 0.00007701, },
		{ 0.00012698, -0.00008173, 0.00003604, -0.00004217, 0.00001790, 0.00000901, -0.00001127, 0.00000977, -0.00000837, -0.00000181, -0.00001104, 0.00003689, -0.00003195, 0.00001825, -0.00003969, 0.00005372, -0.00006812, 0.00004679, -0.00005344, 0.00010570, -0.00011611, 0.00004374, -0.00008431, 0.00017607, -0.00021768, 0.00015728, -0.00028922, 0.00039159, -0.00097505, 0.00088150, 0.00088150, -0.00097505, 0.00039159, -0.00028922, 0.00015728, -0.00021768, 0.00017607, -0.00008431, 0.00004374, -0.00011611, 0.00010570, -0.00005344, 0.00004679, -0.00006812, 0.00005372, -0.00003969, 0.00001825, -0.00003195, 0.00003689, -0.00001104, -0.00000181, -0.00000837, 0.00000977, -0.00001127, 0.00000901, 0.00001790, -0.00004217, 0.00003604, -0.00008173, 0.00012698, -0.00012169, 0.00005766, 0.00005766, -0.00012169, },
		{ 0.00008616, -0.00000786, -0.00001558, -0.00000392, -0.00001832, 0.00002147, -0.00000103, -0.00000827, 0.00001110, -0.00001315, -0.00000553, 0.00001474, 0.00000615, -0.00001196, -0.00001122, 0.00000951, -0.00000375, -0.00001961, -0.00000029, 0.00002736, 0.00000937, -0.00006025, -0.00001837, 0.00006002, -0.00001271, -0.00006028, -0.00003773, 0.00000871, -0.00008349, -0.00071690, 0.00186614, -0.00071690, -0.00008349, 0.00000871, -0.00003773, -0.00006028, -0.00001271, 0.00006002, -0.00001837, -0.00006025, 0.00000937, 0.00002736, -0.00000029, -0.00001961, -0.00000375, 0.00000951, -0.00001122, -0.00001196, 0.00000615, 0.00001474, -0.00000553, -0.00001315, 0.00001110, -0.00000827, -0.00000103, 0.00002147, -0.00001832, -0.00000392, -0.00001558, -0.00000786, 0.00008616, -0.00017138, 0.00021835, -0.00017138, },
		{ -0.00013120, 0.00013849, -0.00008677, 0.00004021, -0.00004634, 0.00001993, 0.00000947, -0.00001143, 0.00000904, -0.00000738, -0.00000250, -0.00001071, 0.00003704, -0.00003240, 0.00001880, -0.00004028, 0.00005496, -0.00007005, 0.00004827, -0.00005515, 0.00010894, -0.00011932, 0.00004467, -0.00008570, 0.00018122, -0.00022531, 0.00016385, -0.00029836, 0.00040832, -0.00100141, 0.00089823, 0.00089823, -0.00100141, 0.00040832, -0.00029836, 0.00016385, -0.00022531, 0.00018122, -0.00008570, 0.00004467, -0.00011932, 0.00010894, -0.00005515, 0.00004827, -0.00007005, 0.00005496, -0.00004028, 0.00001880, -0.00003240, 0.00003704, -0.00001071, -0.00000250, -0.00000738, 0.00000904, -0.00001143, 0.00000947, 0.00001993, -0.00004634, 0.00004021, -0.00008677, 0.00013849, -0.00013120, 0.00005813, 0.00005813, },
		{ -0.00018242, 0.00009177, -0.00000714, -0.00001541, -0.00000399, -0.00001966, 0.00002299, -0.00000090, -0.00000910, 0.00001169, -0.00001331, -0.00000557, 0.00001492, 0.00000605, -0.00001197, -0.00001122, 0.00000981, -0.00000399, -0.00001994, -0.00000025, 0.00002779, 0.00000994, -0.00006185, -0.00001849, 0.00006211, -0.00001397, -0.00006117, -0.00003789, 0.00000967, -0.00008037, -0.00073596, 0.00189907, -0.00073596, -0.00008037, 0.00000967, -0.00003789, -0.00006117, -0.00001397, 0.00006211, -0.00001849, -0.00006185, 0.00000994, 0.00002779, -0.00000025, -0.00001994, -0.00000399, 0.00000981, -0.00001122, -0.00001197, 0.00000605, 0.00001492, -0.00000557, -0.00001331, 0.00001169, -0.00000910, -0.00000090, 0.00002299, -0.00001966, -0.00000399, -0.00001541, -0.00000714, 0.00009177, -0.00018242, 0.00022709, },
		{ 0.00005843, -0.00013611, 0.00014438, -0.00008936, 0.00004235, -0.00004848, 0.00002098, 0.00000969, -0.00001150, 0.00000865, -0.00000686, -0.00000286, -0.00001053, 0.00003711, -0.00003263, 0.00001908, -0.00004058, 0.00005559, -0.00007102, 0.00004902, -0.00005602, 0.00011059, -0.00012095, 0.00004513, -0.00008640, 0.00018383, -0.00022919, 0.00016718, -0.00030300, 0.00041683, -0.00101479, 0.00090670, 0.00090670, -0.00101479, 0.00041683, -0.00030300, 0.00016718, -0.00022919, 0.00018383, -0.00008640, 0.00004513, -0.00012095, 0.00011059, -0.00005602, 0.00004902, -0.00007102, 0.00005559, -0.00004058, 0.00001908, -0.00003263, 0.00003711, -0.00001053, -0.00000286, -0.00000686, 0.00000865, -0.00001150, 0.00000969, 0.00002098, -0.00004848, 0.00004235, -0.00008936, 0.00014438, -0.00013611, 0.00005843, },
		{ 0.00023008, -0.00018617, 0.00009366, -0.00000690, -0.00001536, -0.00000401, -0.00002011, 0.00002351, -0.00000086, -0.00000938, 0.00001189, -0.00001336, -0.00000559, 0.00001498, 0.00000602, -0.00001198, -0.00001122, 0.00000991, -0.00000407, -0.00002005, -0.00000024, 0.00002794, 0.00001013, -0.00006238, -0.00001854, 0.00006282, -0.00001439, -0.00006147, -0.00003794, 0.00000999, -0.00007931, -0.00074238, 0.00191015, -0.00074238, -0.00007931, 0.00000999, -0.00003794, -0.00006147, -0.00001439, 0.00006282, -0.00001854, -0.00006238, 0.00001013, 0.00002794, -0.00000024, -0.00002005, -0.00000407, 0.00000991, -0.00001122, -0.00001198, 0.00000602, 0.00001498, -0.00000559, -0.00001336, 0.00001189, -0.00000938, -0.00000086, 0.00002351, -0.00002011, -0.00000401, -0.00001536, -0.00000690, 0.00009366, -0.00018617, },
		{ 0.00005843, 0.00005843, -0.00013611, 0.00014438, -0.00008936, 0.00004235, -0.00004848, 0.00002098, 0.00000969, -0.00001150, 0.00000865, -0.00000686, -0.00000286, -0.00001053, 0.00003711, -0.00003263, 0.00001908, -0.00004058, 0.00005559, -0.00007102, 0.00004902, -0.00005602, 0.00011059, -0.00012095, 0.00004513, -0.00008640, 0.00018383, -0.00022919, 0.00016718, -0.00030300, 0.00041683, -0.00101479, 0.00090670, 0.00090670, -0.00101479, 0.00041683, -0.00030300, 0.00016718, -0.00022919, 0.00018383, -0.00008640, 0.00004513, -0.00012095, 0.00011059, -0.00005602, 0.00004902, -0.00007102, 0.00005559, -0.00004058, 0.00001908, -0.00003263, 0.00003711, -0.00001053, -0.00000286, -0.00000686, 0.00000865, -0.00001150, 0.00000969, 0.00002098, -0.00004848, 0.00004235, -0.00008936, 0.00014438, -0.00013611, },
		{ -0.00018242, 0.00022709, -0.00018242, 0.00009177, -0.00000714, -0.00001541, -0.00000399, -0.00001966, 0.00002299, -0.00000090, -0.00000910, 0.00001169, -0.00001331, -0.00000557, 0.00001492, 0.00000605, -0.00001197, -0.00001122, 0.00000981, -0.00000399, -0.00001994, -0.00000025, 0.00002779, 0.00000994, -0.00006185, -0.00001849, 0.00006211, -0.00001397, -0.00006117, -0.00003789, 0.00000967, -0.00008037, -0.00073596, 0.00189907, -0.00073596, -0.00008037, 0.00000967, -0.00003789, -0.00006117, -0.00001397, 0.00006211, -0.00001849, -0.00006185, 0.00000994, 0.00002779, -0.00000025, -0.00001994, -0.00000399, 0.00000981, -0.00001122, -0.00001197, 0.00000605, 0.00001492, -0.00000557, -0.00001331, 0.00001169, -0.00000910, -0.00000090, 0.00002299, -0.00001966, -0.00000399, -0.00001541, -0.00000714, 0.00009177, },
		{ -0.00013120, 0.00005813, 0.00005813, -0.00013120, 0.00013849, -0.00008677, 0.00004021, -0.00004634, 0.00001993, 0.00000947, -0.00001143, 0.00000904, -0.00000738, -0.00000250, -0.00001071, 0.00003704, -0.00003240, 0.00001880, -0.00004028, 0.00005496, -0.00007005, 0.00004827, -0.00005515, 0.00010894, -0.00011932, 0.00004467, -0.00008570, 0.00018122, -0.00022531, 0.00016385, -0.00029836, 0.00040832, -0.00100141, 0.00089823, 0.00089823, -0.00100141, 0.00040832, -0.00029836, 0.00016385, -0.00022531, 0.00018122, -0.00008570, 0.00004467, -0.00011932, 0.00010894, -0.00005515, 0.00004827, -0.00007005, 0.00005496, -0.00004028, 0.00001880, -0.00003240, 0.00003704, -0.00001071, -0.00000250, -0.00000738, 0.00000904, -0.00001143, 0.00000947, 0.00001993, -0.00004634, 0.00004021, -0.00008677, 0.00013849, },
		{ 0.00008616, -0.00017138, 0.00021835, -0.00017138, 0.00008616, -0.00000786, -0.00001558, -0.00000392, -0.00001832, 0.00002147, -0.00000103, -0.00000827, 0.00001110, -0.00001315, -0.00000553, 0.00001474, 0.00000615, -0.00001196, -0.00001122, 0.00000951, -0.00000375, -0.00001961, -0.00000029, 0.00002736, 0.00000937, -0.00006025, -0.00001837, 0.00006002, -0.00001271, -0.00006028, -0.00003773, 0.00000871, -0.00008349, -0.00071690, 0.00186614, -0.00071690, -0.00008349, 0.00000871, -0.00003773, -0.00006028, -0.00001271, 0.00006002, -0.00001837, -0.00006025, 0.00000937, 0.00002736, -0.00000029, -0.00001961, -0.00000375, 0.00000951, -0.00001122, -0.00001196, 0.00000615, 0.00001474, -0.00000553, -0.00001315, 0.00001110, -0.00000827, -0.00000103, 0.00002147, -0.00001832, -0.00000392, -0.00001558, -0.00000786, },
		{ 0.00012698, -0.00012169, 0.00005766, 0.00005766, -0.00012169, 0.00012698, -0.00008173, 0.00003604, -0.00004217, 0.00001790, 0.00000901, -0.00001127, 0.00000977, -0.00000837, -0.00000181, -0.00001104, 0.00003689, -0.00003195, 0.00001825, -0.00003969, 0.00005372, -0.00006812, 0.00004679, -0.00005344, 0.00010570, -0.00011611, 0.00004374, -0.00008431, 0.00017607, -0.00021768, 0.00015728, -0.00028922, 0.00039159, -0.00097505, 0.00088150, 0.00088150, -0.00097505, 0.00039159, -0.00028922, 0.00015728, -0.00021768, 0.00017607, -0.00008431, 0.00004374, -0.00011611, 0.00010570, -0.00005344, 0.00004679, -0.00006812, 0.00005372, -0.00003969, 0.00001825, -0.00003195, 0.00003689, -0.00001104, -0.00000181, -0.00000837, 0.00000977, -0.00001127, 0.00000901, 0.00001790, -0.00004217, 0.00003604, -0.00008173, },
		{ -0.00000902, 0.00007701, -0.00015357, 0.00020454, -0.00015357, 0.00007701, -0.00000902, -0.00001584, -0.00000384, -0.00001614, 0.00001900, -0.00000124, -0.00000692, 0.00001013, -0.00001288, -0.00000547, 0.00001445, 0.00000631, -0.00001194, -0.00001122, 0.00000901, -0.00000335, -0.00001906, -0.00000036, 0.00002665, 0.00000845, -0.00005764, -0.00001815, 0.00005662, -0.00001067, -0.00005882, -0.00003746, 0.00000714, -0.00008852, -0.00068583, 0.00181226, -0.00068583, -0.00008852, 0.00000714, -0.00003746, -0.00005882, -0.00001067, 0.00005662, -0.00001815, -0.00005764, 0.00000845, 0.00002665, -0.00000036, -0.00001906, -0.00000335, 0.00000901, -0.00001122, -0.00001194, 0.00000631, 0.00001445, -0.00000547, -0.00001288, 0.00001013, -0.00000692, -0.00000124, 0.00001900, -0.00001614, -0.00000384, -0.00001584, },
		{ -0.00007448, 0.00011037, -0.00010817, 0.00005727, 0.00005727, -0.00010817, 0.00011037, -0.00007448, 0.00003003, -0.00003618, 0.00001502, 0.00000832, -0.00001102, 0.00001078, -0.00000978, -0.00000082, -0.00001151, 0.00003664, -0.00003127, 0.00001743, -0.00003881, 0.00005189, -0.00006530, 0.00004463, -0.00005094, 0.00010094, -0.00011140, 0.00004238, -0.00008227, 0.00016854, -0.00020653, 0.00014772, -0.00027588, 0.00036720, -0.00093648, 0.00085692, 0.00085692, -0.00093648, 0.00036720, -0.00027588, 0.00014772, -0.00020653, 0.00016854, -0.00008227, 0.00004238, -0.00011140, 0.00010094, -0.00005094, 0.00004463, -0.00006530, 0.00005189, -0.00003881, 0.00001743, -0.00003127, 0.00003664, -0.00001151, -0.00000082, -0.00000978, 0.00001078, -0.00001102, 0.00000832, 0.00001502, -0.00003618, 0.00003003, },
		{ -0.00001622, -0.00001057, 0.00006461, -0.00012987, 0.00018677, -0.00012987, 0.00006461, -0.00001057, -0.00001622, -0.00000375, -0.00001320, 0.00001571, -0.00000153, -0.00000514, 0.00000884, -0.00001249, -0.00000538, 0.00001404, 0.00000653, -0.00001189, -0.00001122, 0.00000833, -0.00000280, -0.00001832, -0.00000044, 0.00002567, 0.00000721, -0.00005411, -0.00001786, 0.00005203, -0.00000792, -0.00005684, -0.00003708, 0.00000503, -0.00009520, -0.00064375, 0.00173895, -0.00064375, -0.00009520, 0.00000503, -0.00003708, -0.00005684, -0.00000792, 0.00005203, -0.00001786, -0.00005411, 0.00000721, 0.00002567, -0.00000044, -0.00001832, -0.00000280, 0.00000833, -0.00001122, -0.00001189, 0.00000653, 0.00001404, -0.00000538, -0.00001249, 0.00000884, -0.00000514, -0.00000153, 0.00001571, -0.00001320, -0.00000375, },
		{ 0.00002247, -0.00006539, 0.00008940, -0.00009155, 0.00005733, 0.00005733, -0.00009155, 0.00008940, -0.00006539, 0.00002247, -0.00002867, 0.00001145, 0.00000738, -0.00001066, 0.00001199, -0.00001149, 0.00000039, -0.00001207, 0.00003624, -0.00003034, 0.00001635, -0.00003765, 0.00004950, -0.00006163, 0.00004183, -0.00004773, 0.00009482, -0.00010531, 0.00004060, -0.00007960, 0.00015885, -0.00019223, 0.00013545, -0.00025875, 0.00033599, -0.00088682, 0.00082509, 0.00082509, -0.00088682, 0.00033599, -0.00025875, 0.00013545, -0.00019223, 0.00015885, -0.00007960, 0.00004060, -0.00010531, 0.00009482, -0.00004773, 0.00004183, -0.00006163, 0.00004950, -0.00003765, 0.00001635, -0.00003034, 0.00003624, -0.00001207, 0.00000039, -0.00001149, 0.00001199, -0.00001066, 0.00000738, 0.00001145, -0.00002867, },
		{ -0.00000369, -0.00001669, -0.00001245, 0.00004936, -0.00010146, 0.00016652, -0.00010146, 0.00004936, -0.00001245, -0.00001669, -0.00000369, -0.00000962, 0.00001172, -0.00000190, -0.00000301, 0.00000726, -0.00001199, -0.00000527, 0.00001351, 0.00000677, -0.00001182, -0.00001122, 0.00000748, -0.00000213, -0.00001738, -0.00000056, 0.00002444, 0.00000570, -0.00004975, -0.00001749, 0.00004639, -0.00000458, -0.00005436, -0.00003661, 0.00000245, -0.00010319, -0.00059200, 0.00164821, -0.00059200, -0.00010319, 0.00000245, -0.00003661, -0.00005436, -0.00000458, 0.00004639, -0.00001749, -0.00004975, 0.00000570, 0.00002444, -0.00000056, -0.00001738, -0.00000213, 0.00000748, -0.00001122, -0.00001182, 0.00000677, 0.00001351, -0.00000527, -0.00001199, 0.00000726, -0.00000301, -0.00000190, 0.00001172, -0.00000962, },
		{ -0.00002004, 0.00001370, -0.00005493, 0.00006503, -0.00007292, 0.00005830, 0.00005830, -0.00007292, 0.00006503, -0.00005493, 0.00001370, -0.00002004, 0.00000743, 0.00000616, -0.00001018, 0.00001328, -0.00001337, 0.00000177, -0.00001266, 0.00003566, -0.00002915, 0.00001501, -0.00003622, 0.00004660, -0.00005720, 0.00003848, -0.00004389, 0.00008749, -0.00009802, 0.00003846, -0.00007636, 0.00014728, -0.00017523, 0.00012091, -0.00023840, 0.00029900, -0.00082750, 0.00078676, 0.00078676, -0.00082750, 0.00029900, -0.00023840, 0.00012091, -0.00017523, 0.00014728, -0.00007636, 0.00003846, -0.00009802, 0.00008749, -0.00004389, 0.00003848, -0.00005720, 0.00004660, -0.00003622, 0.00001501, -0.00002915, 0.00003566, -0.00001266, 0.00000177, -0.00001337, 0.00001328, -0.00001018, 0.00000616, 0.00000743, },
		{ -0.00000553, -0.00000369, -0.00001726, -0.00001457, 0.00003173, -0.00006977, 0.00014559, -0.00006977, 0.00003173, -0.00001457, -0.00001726, -0.00000369, -0.00000553, 0.00000723, -0.00000235, -0.00000065, 0.00000545, -0.00001136, -0.00000514, 0.00001286, 0.00000704, -0.00001171, -0.00001122, 0.00000649, -0.00000135, -0.00001626, -0.00000069, 0.00002298, 0.00000399, -0.00004472, -0.00001705, 0.00003991, -0.00000078, -0.00005147, -0.00003604, -0.00000053, -0.00011209, -0.00053224, 0.00154251, -0.00053224, -0.00011209, -0.00000053, -0.00003604, -0.00005147, -0.00000078, 0.00003991, -0.00001705, -0.00004472, 0.00000399, 0.00002298, -0.00000069, -0.00001626, -0.00000135, 0.00000649, -0.00001122, -0.00001171, 0.00000704, 0.00001286, -0.00000514, -0.00001136, 0.00000545, -0.00000065, -0.00000235, 0.00000723, },
		{ 0.00000321, -0.00001072, 0.00000416, -0.00004364, 0.00003838, -0.00005360, 0.00006069, 0.00006069, -0.00005360, 0.00003838, -0.00004364, 0.00000416, -0.00001072, 0.00000321, 0.00000467, -0.00000956, 0.00001451, -0.00001528, 0.00000322, -0.00001323, 0.00003485, -0.00002770, 0.00001343, -0.00003454, 0.00004324, -0.00005211, 0.00003466, -0.00003954, 0.00007917, -0.00008972, 0.00003599, -0.00007261, 0.00013418, -0.00015605, 0.00010455, -0.00021545, 0.00025748, -0.00076021, 0.00074285, 0.00074285, -0.00076021, 0.00025748, -0.00021545, 0.00010455, -0.00015605, 0.00013418, -0.00007261, 0.00003599, -0.00008972, 0.00007917, -0.00003954, 0.00003466, -0.00005211, 0.00004324, -0.00003454, 0.00001343, -0.00002770, 0.00003485, -0.00001323, 0.00000322, -0.00001528, 0.00001451, -0.00000956, 0.00000467, },
		{ 0.00000242, -0.00000108, -0.00000380, -0.00001792, -0.00001684, 0.00001228, -0.00003642, 0.00012599, -0.00003642, 0.00001228, -0.00001684, -0.00001792, -0.00000380, -0.00000108, 0.00000242, -0.00000286, 0.00000181, 0.00000348, -0.00001060, -0.00000498, 0.00001209, 0.00000732, -0.00001156, -0.00001122, 0.00000536, -0.00000049, -0.00001498, -0.00000085, 0.00002133, 0.00000212, -0.00003916, -0.00001653, 0.00003281, 0.00000334, -0.00004822, -0.00003538, -0.00000378, -0.00012143, -0.00046637, 0.00142471, -0.00046637, -0.00012143, -0.00000378, -0.00003538, -0.00004822, 0.00000334, 0.00003281, -0.00001653, -0.00003916, 0.00000212, 0.00002133, -0.00000085, -0.00001498, -0.00000049, 0.00000536, -0.00001122, -0.00001156, 0.00000732, 0.00001209, -0.00000498, -0.00001060, 0.00000348, 0.00000181, -0.00000286, },
		{ 0.00000286, -0.00000091, -0.00000123, -0.00000570, -0.00003212, 0.00001068, -0.00003501, 0.00006509, 0.00006509, -0.00003501, 0.00001068, -0.00003212, -0.00000570, -0.00000123, -0.00000091, 0.00000286, -0.00000877, 0.00001552, -0.00001706, 0.00000464, -0.00001370, 0.00003375, -0.00002596, 0.00001163, -0.00003263, 0.00003948, -0.00004647, 0.00003048, -0.00003481, 0.00007008, -0.00008062, 0.00003324, -0.00006842, 0.00011991, -0.00013529, 0.00008691, -0.00019061, 0.00021280, -0.00068685, 0.00069438, 0.00069438, -0.00068685, 0.00021280, -0.00019061, 0.00008691, -0.00013529, 0.00011991, -0.00006842, 0.00003324, -0.00008062, 0.00007008, -0.00003481, 0.00003048, -0.00004647, 0.00003948, -0.00003263, 0.00001163, -0.00002596, 0.00003375, -0.00001370, 0.00000464, -0.00001706, 0.00001552, -0.00000877, },
		{ -0.00000345, -0.00000248, 0.00000356, -0.00000406, -0.00001867, -0.00001916, -0.00000838, -0.00000315, 0.00010991, -0.00000315, -0.00000838, -0.00001916, -0.00001867, -0.00000406, 0.00000356, -0.00000248, -0.00000345, 0.00000423, 0.00000144, -0.00000970, -0.00000480, 0.00001120, 0.00000758, -0.00001134, -0.00001122, 0.00000414, 0.00000044, -0.00001356, -0.00000102, 0.00001950, 0.00000019, -0.00003324, -0.00001595, 0.00002532, 0.00000761, -0.00004469, -0.00003464, -0.00000722, -0.00013069, -0.00039645, 0.00129790, -0.00039645, -0.00013069, -0.00000722, -0.00003464, -0.00004469, 0.00000761, 0.00002532, -0.00001595, -0.00003324, 0.00000019, 0.00001950, -0.00000102, -0.00001356, 0.00000044, 0.00000414, -0.00001122, -0.00001134, 0.00000758, 0.00001120, -0.00000480, -0.00000970, 0.00000144, 0.00000423, },
		{ -0.00000781, 0.00000074, -0.00000463, 0.00000791, -0.00001539, -0.00002101, -0.00001675, -0.00001866, 0.00007209, 0.00007209, -0.00001866, -0.00001675, -0.00002101, -0.00001539, 0.00000791, -0.00000463, 0.00000074, -0.00000781, 0.00001618, -0.00001852, 0.00000592, -0.00001401, 0.00003231, -0.00002393, 0.00000961, -0.00003050, 0.00003539, -0.00004040, 0.00002604, -0.00002982, 0.00006046, -0.00007097, 0.00003028, -0.00006387, 0.00010487, -0.00011360, 0.00006855, -0.00016466, 0.00016644, -0.00060946, 0.00064244, 0.00064244, -0.00060946, 0.00016644, -0.00016466, 0.00006855, -0.00011360, 0.00010487, -0.00006387, 0.00003028, -0.00007097, 0.00006046, -0.00002982, 0.00002604, -0.00004040, 0.00003539, -0.00003050, 0.00000961, -0.00002393, 0.00003231, -0.00001401, 0.00000592, -0.00001852, 0.00001618, },
		{ 0.00000647, -0.00000409, -0.00000724, 0.00000823, -0.00000452, -0.00001949, -0.00002142, -0.00002964, 0.00002823, 0.00009955, 0.00002823, -0.00002964, -0.00002142, -0.00001949, -0.00000452, 0.00000823, -0.00000724, -0.00000409, 0.00000647, -0.00000062, -0.00000867, -0.00000461, 0.00001019, 0.00000780, -0.00001104, -0.00001122, 0.00000285, 0.00000140, -0.00001202, -0.00000122, 0.00001753, -0.00000173, -0.00002715, -0.00001530, 0.00001770, 0.00001185, -0.00004097, -0.00003381, -0.00001070, -0.00013937, -0.00032463, 0.00116538, -0.00032463, -0.00013937, -0.00001070, -0.00003381, -0.00004097, 0.00001185, 0.00001770, -0.00001530, -0.00002715, -0.00000173, 0.00001753, -0.00000122, -0.00001202, 0.00000140, 0.00000285, -0.00001122, -0.00001104, 0.00000780, 0.00001019, -0.00000461, -0.00000867, -0.00000062, },
		{ 0.00001631, -0.00000664, -0.00000171, -0.00000764, 0.00001617, -0.00002440, -0.00001095, -0.00004260, -0.00000606, 0.00008227, 0.00008227, -0.00000606, -0.00004260, -0.00001095, -0.00002440, 0.00001617, -0.00000764, -0.00000171, -0.00000664, 0.00001631, -0.00001949, 0.00000698, -0.00001407, 0.00003047, -0.00002159, 0.00000740, -0.00002818, 0.00003103, -0.00003402, 0.00002145, -0.00002473, 0.00005056, -0.00006099, 0.00002714, -0.00005902, 0.00008950, -0.00009161, 0.00005006, -0.00013837, 0.00011990, -0.00053012, 0.00058819, 0.00058819, -0.00053012, 0.00011990, -0.00013837, 0.00005006, -0.00009161, 0.00008950, -0.00005902, 0.00002714, -0.00006099, 0.00005056, -0.00002473, 0.00002145, -0.00003402, 0.00003103, -0.00002818, 0.00000740, -0.00002159, 0.00003047, -0.00001407, 0.00000698, -0.00001949, },
		{ -0.00000260, 0.00000840, -0.00000478, -0.00001164, 0.00001274, -0.00000522, -0.00002038, -0.00002352, -0.00005084, 0.00005592, 0.00009707, 0.00005592, -0.00005084, -0.00002352, -0.00002038, -0.00000522, 0.00001274, -0.00001164, -0.00000478, 0.00000840, -0.00000260, -0.00000749, -0.00000440, 0.00000907, 0.00000797, -0.00001066, -0.00001122, 0.00000150, 0.00000237, -0.00001039, -0.00000143, 0.00001546, -0.00000357, -0.00002105, -0.00001461, 0.00001020, 0.00001591, -0.00003714, -0.00003292, -0.00001413, -0.00014694, -0.00025310, 0.00103046, -0.00025310, -0.00014694, -0.00001413, -0.00003292, -0.00003714, 0.00001591, 0.00001020, -0.00001461, -0.00002105, -0.00000357, 0.00001546, -0.00000143, -0.00001039, 0.00000237, 0.00000150, -0.00001122, -0.00001066, 0.00000797, 0.00000907, -0.00000440, -0.00000749, },
		{ -0.00001981, 0.00001578, -0.00000525, -0.00000451, -0.00000964, 0.00002303, -0.00003225, -0.00000255, -0.00006556, 0.00000135, 0.00009616, 0.00009616, 0.00000135, -0.00006556, -0.00000255, -0.00003225, 0.00002303, -0.00000964, -0.00000451, -0.00000525, 0.00001578, -0.00001981, 0.00000771, -0.00001383, 0.00002819, -0.00001894, 0.00000501, -0.00002570, 0.00002647, -0.00002746, 0.00001682, -0.00001966, 0.00004065, -0.00005093, 0.00002391, -0.00005397, 0.00007418, -0.00006998, 0.00003201, -0.00011252, 0.00007468, -0.00045092, 0.00053278, 0.00053278, -0.00045092, 0.00007468, -0.00011252, 0.00003201, -0.00006998, 0.00007418, -0.00005397, 0.00002391, -0.00005093, 0.00004065, -0.00001966, 0.00001682, -0.00002746, 0.00002647, -0.00002570, 0.00000501, -0.00001894, 0.00002819, -0.00001383, 0.00000771, },
		{ -0.00000616, -0.00000443, 0.00000986, -0.00000553, -0.00001548, 0.00001693, -0.00000619, -0.00002133, -0.00002535, -0.00007139, 0.00007823, 0.00010446, 0.00007823, -0.00007139, -0.00002535, -0.00002133, -0.00000619, 0.00001693, -0.00001548, -0.00000553, 0.00000986, -0.00000443, -0.00000616, -0.00000417, 0.00000783, 0.00000807, -0.00001018, -0.00001122, 0.00000014, 0.00000331, -0.00000868, -0.00000165, 0.00001331, -0.00000526, -0.00001512, -0.00001386, 0.00000305, 0.00001962, -0.00003328, -0.00003197, -0.00001739, -0.00015293, -0.00018396, 0.00089644, -0.00018396, -0.00015293, -0.00001739, -0.00003197, -0.00003328, 0.00001962, 0.00000305, -0.00001386, -0.00001512, -0.00000526, 0.00001331, -0.00000165, -0.00000868, 0.00000331, 0.00000014, -0.00001122, -0.00001018, 0.00000807, 0.00000783, -0.00000417, },
		{ 0.00000801, -0.00001933, 0.00001446, -0.00000362, -0.00000766, -0.00001038, 0.00002803, -0.00003850, 0.00000360, -0.00008443, 0.00000226, 0.00011423, 0.00011423, 0.00000226, -0.00008443, 0.00000360, -0.00003850, 0.00002803, -0.00001038, -0.00000766, -0.00000362, 0.00001446, -0.00001933, 0.00000801, -0.00001322, 0.00002543, -0.00001597, 0.00000247, -0.00002307, 0.00002179, -0.00002085, 0.00001227, -0.00001475, 0.00003095, -0.00004103, 0.00002062, -0.00004880, 0.00005933, -0.00004933, 0.00001494, -0.00008785, 0.00003220, -0.00037388, 0.00047735, 0.00047735, -0.00037388, 0.00003220, -0.00008785, 0.00001494, -0.00004933, 0.00005933, -0.00004880, 0.00002062, -0.00004103, 0.00003095, -0.00001475, 0.00001227, -0.00002085, 0.00002179, -0.00002307, 0.00000247, -0.00001597, 0.00002543, -0.00001322, },
		{ -0.00000394, -0.00000469, -0.00000606, 0.00001075, -0.00000631, -0.00001854, 0.00002064, -0.00000747, -0.00002233, -0.00002684, -0.00009068, 0.00009364, 0.00012349, 0.00009364, -0.00009068, -0.00002684, -0.00002233, -0.00000747, 0.00002064, -0.00001854, -0.00000631, 0.00001075, -0.00000606, -0.00000469, -0.00000394, 0.00000648, 0.00000809, -0.00000960, -0.00001122, -0.00000122, 0.00000422, -0.00000692, -0.00000189, 0.00001111, -0.00000673, -0.00000953, -0.00001308, -0.00000350, 0.00002283, -0.00002946, -0.00003097, -0.00002037, -0.00015688, -0.00011921, 0.00076642, -0.00011921, -0.00015688, -0.00002037, -0.00003097, -0.00002946, 0.00002283, -0.00000350, -0.00001308, -0.00000953, -0.00000673, 0.00001111, -0.00000189, -0.00000692, 0.00000422, -0.00000122, -0.00001122, -0.00000960, 0.00000809, 0.00000648, },
		{ -0.00001220, 0.00000782, -0.00001791, 0.00001222, -0.00000176, -0.00001116, -0.00000962, 0.00003073, -0.00004275, 0.00000701, -0.00009815, -0.00000448, 0.00013684, 0.00013684, -0.00000448, -0.00009815, 0.00000701, -0.00004275, 0.00003073, -0.00000962, -0.00001116, -0.00000176, 0.00001222, -0.00001791, 0.00000782, -0.00001220, 0.00002216, -0.00001270, -0.00000020, -0.00002032, 0.00001706, -0.00001431, 0.00000789, -0.00001012, 0.00002170, -0.00003150, 0.00001734, -0.00004357, 0.00004531, -0.00003023, -0.00000064, -0.00006505, -0.00000626, -0.00030085, 0.00042296, 0.00042296, -0.00030085, -0.00000626, -0.00006505, -0.00000064, -0.00003023, 0.00004531, -0.00004357, 0.00001734, -0.00003150, 0.00002170, -0.00001012, 0.00000789, -0.00001431, 0.00001706, -0.00002032, -0.00000020, -0.00001270, 0.00002216, },
		{ 0.00000503, -0.00000369, -0.00000308, -0.00000741, 0.00001096, -0.00000713, -0.00002067, 0.00002374, -0.00000909, -0.00002337, -0.00002789, -0.00010821, 0.00010083, 0.00015556, 0.00010083, -0.00010821, -0.00002789, -0.00002337, -0.00000909, 0.00002374, -0.00002067, -0.00000713, 0.00001096, -0.00000741, -0.00000308, -0.00000369, 0.00000503, 0.00000800, -0.00000891, -0.00001122, -0.00000255, 0.00000507, -0.00000514, -0.00000214, 0.00000891, -0.00000792, -0.00000442, -0.00001226, -0.00000926, 0.00002540, -0.00002577, -0.00002993, -0.00002299, -0.00015843, -0.00006063, 0.00064326, -0.00006063, -0.00015843, -0.00002299, -0.00002993, -0.00002577, 0.00002540, -0.00000926, -0.00001226, -0.00000442, -0.00000792, 0.00000891, -0.00000214, -0.00000514, 0.00000507, -0.00000255, -0.00001122, -0.00000891, 0.00000800, },
		{ 0.00001837, -0.00001072, 0.00000706, -0.00001547, 0.00000899, 0.00000035, -0.00001500, -0.00000717, 0.00003080, -0.00004467, 0.00000726, -0.00010585, -0.00001975, 0.00016424, 0.00016424, -0.00001975, -0.00010585, 0.00000726, -0.00004467, 0.00003080, -0.00000717, -0.00001500, 0.00000035, 0.00000899, -0.00001547, 0.00000706, -0.00001072, 0.00001837, -0.00000912, -0.00000297, -0.00001749, 0.00001234, -0.00000794, 0.00000377, -0.00000588, 0.00001310, -0.00002254, 0.00001412, -0.00003838, 0.00003245, -0.00001316, -0.00001430, -0.00004471, -0.00003956, -0.00023350, 0.00037060, 0.00037060, -0.00023350, -0.00003956, -0.00004471, -0.00001430, -0.00001316, 0.00003245, -0.00003838, 0.00001412, -0.00002254, 0.00001310, -0.00000588, 0.00000377, -0.00000794, 0.00001234, -0.00001749, -0.00000297, -0.00000912, },
		{ 0.00000780, 0.00000349, -0.00000343, -0.00000134, -0.00000843, 0.00001041, -0.00000797, -0.00002174, 0.00002613, -0.00001105, -0.00002445, -0.00002845, -0.00012354, 0.00009880, 0.00020169, 0.00009880, -0.00012354, -0.00002845, -0.00002445, -0.00001105, 0.00002613, -0.00002174, -0.00000797, 0.00001041, -0.00000843, -0.00000134, -0.00000343, 0.00000349, 0.00000780, -0.00000810, -0.00001122, -0.00000383, 0.00000583, -0.00000334, -0.00000239, 0.00000672, -0.00000878, 0.00000007, -0.00001142, -0.00001407, 0.00002724, -0.00002226, -0.00002885, -0.00002515, -0.00015728, -0.00000971, 0.00052946, -0.00000971, -0.00015728, -0.00002515, -0.00002885, -0.00002226, 0.00002724, -0.00001407, -0.00001142, 0.00000007, -0.00000878, 0.00000672, -0.00000239, -0.00000334, 0.00000583, -0.00000383, -0.00001122, -0.00000810, },
		{ -0.00000527, 0.00001404, -0.00000875, 0.00000570, -0.00001193, 0.00000472, 0.00000269, -0.00001917, -0.00000291, 0.00002799, -0.00004402, 0.00000406, -0.00010684, -0.00004415, 0.00019651, 0.00019651, -0.00004415, -0.00010684, 0.00000406, -0.00004402, 0.00002799, -0.00000291, -0.00001917, 0.00000269, 0.00000472, -0.00001193, 0.00000570, -0.00000875, 0.00001404, -0.00000527, -0.00000583, -0.00001459, 0.00000769, -0.00000184, -0.00000001, -0.00000211, 0.00000533, -0.00001432, 0.00001101, -0.00003328, 0.00002101, 0.00000144, -0.00002568, -0.00002733, -0.00006681, -0.00017323, 0.00032113, 0.00032113, -0.00017323, -0.00006681, -0.00002733, -0.00002568, 0.00000144, 0.00002101, -0.00003328, 0.00001101, -0.00001432, 0.00000533, -0.00000211, -0.00000001, -0.00000184, 0.00000769, -0.00001459, -0.00000583, },
		{ -0.00000719, 0.00000748, 0.00000187, -0.00000318, 0.00000053, -0.00000910, 0.00000906, -0.00000883, -0.00002163, 0.00002770, -0.00001336, -0.00002554, -0.00002846, -0.00013630, 0.00008685, 0.00026239, 0.00008685, -0.00013630, -0.00002846, -0.00002554, -0.00001336, 0.00002770, -0.00002163, -0.00000883, 0.00000906, -0.00000910, 0.00000053, -0.00000318, 0.00000187, 0.00000748, -0.00000719, -0.00001122, -0.00000504, 0.00000649, -0.00000157, -0.00000265, 0.00000458, -0.00000927, 0.00000386, -0.00001056, -0.00001778, 0.00002826, -0.00001899, -0.00002776, -0.00002681, -0.00015325, 0.00003233, 0.00042707, 0.00003233, -0.00015325, -0.00002681, -0.00002776, -0.00001899, 0.00002826, -0.00001778, -0.00001056, 0.00000386, -0.00000927, 0.00000458, -0.00000265, -0.00000157, 0.00000649, -0.00000504, -0.00001122, },
		{ -0.00000873, -0.00000116, 0.00000921, -0.00000631, 0.00000371, -0.00000727, -0.00000061, 0.00000527, -0.00002363, 0.00000322, 0.00002217, -0.00004065, -0.00000277, -0.00010072, -0.00007796, 0.00023360, 0.00023360, -0.00007796, -0.00010072, -0.00000277, -0.00004065, 0.00002217, 0.00000322, -0.00002363, 0.00000527, -0.00000061, -0.00000727, 0.00000371, -0.00000631, 0.00000921, -0.00000116, -0.00000873, -0.00001166, 0.00000318, 0.00000390, -0.00000338, 0.00000110, -0.00000150, -0.00000698, 0.00000805, -0.00002835, 0.00001122, 0.00001329, -0.00003452, -0.00001327, -0.00008733, -0.00012113, 0.00027527, 0.00027527, -0.00012113, -0.00008733, -0.00001327, -0.00003452, 0.00001329, 0.00001122, -0.00002835, 0.00000805, -0.00000698, -0.00000150, 0.00000110, -0.00000338, 0.00000390, 0.00000318, -0.00001166, },
		},
		{ { -0.00000068, -0.00000335, 0.00001744, -0.00000846, -0.00002796, -0.00000905, 0.00004573, -0.00001088, -0.00001132, -0.00001923, -0.00001879, -0.00001814, -0.00013506, 0.00005895, 0.00037905, 0.00004013, -0.00012414, -0.00002587, -0.00002783, -0.00001345, 0.00000305, 0.00000688, -0.00008455, 0.00001730, 0.00000719, -0.00000569, -0.00002701, -0.00000825, 0.00003981, -0.00000466, -0.00003608, 0.00001141, 0.00001689, 0.00001141, -0.00003608, -0.00000466, 0.00003981, -0.00000825, -0.00002701, -0.00000569, 0.00000719, 0.00001730, -0.00008455, 0.00000688, 0.00000305, -0.00001345, -0.00002783, -0.00002587, -0.00012414, 0.00004013, 0.00037905, 0.00005895, -0.00013506, -0.00001814, -0.00001879, -0.00001923, -0.00001132, -0.00001088, 0.00004573, -0.00000905, -0.00002796, -0.00000846, 0.00001744, -0.00000335, },
		{ -0.00000253, -0.00000451, 0.00000703, 0.00001216, -0.00002026, -0.00002936, 0.00002115, 0.00003145, -0.00002540, -0.00000403, -0.00003060, -0.00000516, -0.00006533, -0.00013510, 0.00026820, 0.00031879, -0.00012947, -0.00006331, -0.00003189, -0.00001562, -0.00001804, 0.00002908, -0.00004974, -0.00005044, 0.00003190, -0.00000371, -0.00001266, -0.00002957, 0.00001744, 0.00003215, -0.00003142, -0.00001612, 0.00001808, 0.00001782, -0.00000839, -0.00003597, 0.00002328, 0.00002908, -0.00003019, -0.00001018, -0.00000952, 0.00002648, -0.00003136, -0.00006392, 0.00003128, -0.00002114, -0.00000759, -0.00003291, -0.00005241, -0.00010920, 0.00022759, 0.00027507, -0.00007120, -0.00008862, -0.00000067, -0.00002935, -0.00000625, -0.00002303, 0.00002101, 0.00002956, -0.00002493, -0.00002288, 0.00000720, 0.00000866, },
		{ -0.00000145, 0.00000072, -0.00000744, 0.00001959, -0.00000388, -0.00002691, -0.00001855, 0.00004569, -0.00000214, -0.00001715, -0.00001207, -0.00002877, -0.00000540, -0.00013529, -0.00002391, 0.00044695, 0.00009301, -0.00016429, -0.00001684, -0.00003997, -0.00000988, -0.00000254, 0.00001777, -0.00008504, 0.00000245, 0.00001944, -0.00000640, -0.00002627, -0.00001376, 0.00003361, 0.00000833, -0.00004029, 0.00000742, 0.00001669, 0.00001442, -0.00003018, -0.00001571, 0.00003980, 0.00000264, -0.00003170, 0.00000008, -0.00000708, 0.00002806, -0.00007812, -0.00000633, 0.00000905, -0.00001840, -0.00001448, -0.00003548, -0.00009066, 0.00000698, 0.00030954, 0.00011198, -0.00010931, -0.00004086, -0.00000569, -0.00002601, -0.00000630, -0.00001774, 0.00004301, 0.00000026, -0.00002700, -0.00001320, 0.00001551, },
		{ 0.00000930, -0.00000090, -0.00000479, 0.00000303, 0.00001772, -0.00001711, -0.00003212, 0.00001129, 0.00003963, -0.00002392, -0.00000672, -0.00002436, -0.00001833, -0.00003492, -0.00018671, 0.00022240, 0.00043418, -0.00014605, -0.00007888, -0.00003039, -0.00002299, -0.00001628, 0.00002669, -0.00003592, -0.00006388, 0.00002882, 0.00000717, -0.00001608, -0.00002806, 0.00000711, 0.00003588, -0.00002272, -0.00002428, 0.00001780, 0.00001698, -0.00000144, -0.00003676, 0.00001168, 0.00003838, -0.00002648, -0.00001221, -0.00000742, 0.00001238, -0.00000857, -0.00007680, 0.00003274, -0.00002508, 0.00000027, -0.00003259, -0.00004580, -0.00008767, 0.00015792, 0.00025748, -0.00001229, -0.00009446, -0.00000925, -0.00001942, -0.00001314, -0.00001732, 0.00000910, 0.00003624, -0.00001948, -0.00002431, 0.00000228, },
		{ 0.00001426, -0.00000185, 0.00000423, -0.00001245, 0.00002081, 0.00000082, -0.00002398, -0.00002794, 0.00004358, 0.00000642, -0.00002066, -0.00000919, -0.00002967, -0.00000862, -0.00010830, -0.00012658, 0.00049573, 0.00017706, -0.00021530, -0.00000537, -0.00005415, -0.00000454, -0.00001032, 0.00002848, -0.00008283, -0.00001150, 0.00002503, 0.00000023, -0.00003010, -0.00001420, 0.00002312, 0.00002085, -0.00004162, 0.00000239, 0.00001608, 0.00001660, -0.00002369, -0.00002322, 0.00003343, 0.00001710, -0.00003747, 0.00000653, -0.00001761, 0.00003002, -0.00006338, -0.00002330, 0.00001753, -0.00002711, 0.00000225, -0.00004749, -0.00006089, -0.00001479, 0.00024843, 0.00013518, -0.00006756, -0.00006340, 0.00000293, -0.00002754, -0.00000492, -0.00002105, 0.00003713, 0.00000916, -0.00002423, -0.00001803, },
		{ -0.00000338, 0.00001052, -0.00000149, -0.00000189, -0.00000398, 0.00002391, -0.00001394, -0.00003245, -0.00000008, 0.00004577, -0.00001981, -0.00001186, -0.00001539, -0.00003175, -0.00001092, -0.00020933, 0.00012930, 0.00056823, -0.00015045, -0.00010351, -0.00002590, -0.00003197, -0.00001356, 0.00002226, -0.00002171, -0.00007273, 0.00002062, 0.00001796, -0.00001592, -0.00002924, 0.00000105, 0.00003322, -0.00001045, -0.00003218, 0.00001698, 0.00001545, 0.00000473, -0.00003489, 0.00000023, 0.00004214, -0.00001636, -0.00002001, 0.00000215, -0.00000712, 0.00001334, -0.00008482, 0.00003104, -0.00002725, 0.00000488, -0.00002788, -0.00004560, -0.00006361, 0.00010303, 0.00023055, 0.00002881, -0.00007901, -0.00002961, -0.00000377, -0.00002190, -0.00001036, -0.00000270, 0.00004067, -0.00001361, -0.00002391, },
		{ -0.00002242, 0.00001327, -0.00000345, 0.00000807, -0.00001635, 0.00001962, 0.00000609, -0.00001958, -0.00003663, 0.00003982, 0.00001331, -0.00001992, -0.00001295, -0.00001964, -0.00002719, -0.00006158, -0.00023007, 0.00050378, 0.00030280, -0.00027928, 0.00000999, -0.00007229, 0.00000431, -0.00002150, 0.00003999, -0.00008016, -0.00002130, 0.00002214, 0.00001333, -0.00003651, -0.00001192, 0.00001143, 0.00003029, -0.00003919, -0.00000356, 0.00001507, 0.00001803, -0.00001740, -0.00002668, 0.00002228, 0.00003164, -0.00004028, 0.00000887, -0.00001967, 0.00002078, -0.00004047, -0.00004335, 0.00002850, -0.00003959, 0.00002184, -0.00006091, -0.00003489, -0.00002866, 0.00019782, 0.00013663, -0.00002359, -0.00007512, 0.00000080, -0.00002076, -0.00000845, -0.00002025, 0.00002830, 0.00001737, -0.00002010, },
		{ -0.00002124, -0.00001034, 0.00001346, -0.00000526, 0.00000448, -0.00001350, 0.00003013, -0.00001112, -0.00002969, -0.00001328, 0.00005085, -0.00001559, -0.00001539, -0.00001018, -0.00003606, -0.00000539, -0.00019242, -0.00000873, 0.00070506, -0.00012783, -0.00014524, -0.00001297, -0.00004737, -0.00000558, 0.00001234, -0.00000469, -0.00008006, 0.00001242, 0.00002307, -0.00000837, -0.00003538, 0.00000068, 0.00002450, 0.00000394, -0.00003893, 0.00001571, 0.00001304, 0.00001041, -0.00003178, -0.00000845, 0.00003858, 0.00000001, -0.00003203, 0.00001561, -0.00002620, 0.00002855, -0.00008404, 0.00002342, -0.00002458, 0.00000287, -0.00001569, -0.00005396, -0.00003519, 0.00005634, 0.00020532, 0.00004700, -0.00004600, -0.00005522, 0.00001159, -0.00002821, -0.00000481, -0.00001263, 0.00004235, -0.00000783, },
		{ -0.00001531, -0.00002541, 0.00001144, -0.00000439, 0.00000992, -0.00001697, 0.00001478, 0.00001226, -0.00001442, -0.00004357, 0.00003418, 0.00001833, -0.00001519, -0.00002223, -0.00000198, -0.00005369, -0.00001017, -0.00031063, 0.00045122, 0.00047455, -0.00035240, 0.00002636, -0.00009264, 0.00001506, -0.00003428, 0.00005082, -0.00007686, -0.00002672, 0.00001258, 0.00002888, -0.00004143, -0.00001044, 0.00000199, 0.00003448, -0.00003267, -0.00001016, 0.00001380, 0.00001855, -0.00001150, -0.00002676, 0.00000922, 0.00004220, -0.00003645, 0.00000377, -0.00001135, 0.00000146, -0.00001271, -0.00006326, 0.00003958, -0.00005322, 0.00004101, -0.00007204, -0.00001562, -0.00003392, 0.00015404, 0.00012834, 0.00000934, -0.00006880, -0.00001444, -0.00000591, -0.00001589, -0.00001620, 0.00001746, 0.00002453, },
		{ -0.00000255, -0.00001623, -0.00001863, 0.00001827, -0.00001187, 0.00001327, -0.00002397, 0.00003522, -0.00000875, -0.00002344, -0.00002841, 0.00005600, -0.00001390, -0.00001332, -0.00001427, -0.00002464, -0.00002461, -0.00013568, -0.00017658, 0.00081932, -0.00005953, -0.00021322, 0.00001443, -0.00007430, 0.00001197, -0.00000629, 0.00001753, -0.00008914, 0.00000895, 0.00001853, 0.00000797, -0.00004651, 0.00000539, 0.00001153, 0.00001839, -0.00004370, 0.00001416, 0.00000952, 0.00001590, -0.00002872, -0.00001276, 0.00002789, 0.00002009, -0.00004442, 0.00002754, -0.00003848, 0.00003220, -0.00007206, 0.00000856, -0.00001536, -0.00000738, 0.00000498, -0.00007085, -0.00000265, 0.00001461, 0.00018629, 0.00004543, -0.00000528, -0.00007654, 0.00001982, -0.00002804, -0.00000284, -0.00001941, 0.00004108, },
		{ 0.00003030, -0.00001071, -0.00002586, 0.00000735, -0.00000269, 0.00000775, -0.00001280, 0.00000575, 0.00001940, -0.00000949, -0.00004724, 0.00002567, 0.00002276, -0.00000911, -0.00003259, 0.00001581, -0.00007614, 0.00002753, -0.00034597, 0.00032647, 0.00068568, -0.00042226, 0.00003493, -0.00010872, 0.00002203, -0.00004351, 0.00005671, -0.00007017, -0.00003056, 0.00000117, 0.00004089, -0.00004012, -0.00001329, -0.00000235, 0.00003220, -0.00002239, -0.00001706, 0.00001261, 0.00001773, -0.00000565, -0.00002497, -0.00000240, 0.00004534, -0.00002400, -0.00000938, 0.00000565, -0.00002319, 0.00001399, -0.00007812, 0.00004682, -0.00006373, 0.00005492, -0.00007593, -0.00000715, -0.00002797, 0.00011135, 0.00012157, 0.00002271, -0.00004321, -0.00004009, 0.00001327, -0.00002424, -0.00001078, 0.00000595, },
		{ 0.00003712, 0.00000192, -0.00000934, -0.00002746, 0.00002399, -0.00001965, 0.00002226, -0.00003307, 0.00003781, -0.00000674, -0.00001385, -0.00004498, 0.00006167, -0.00001632, -0.00000365, -0.00002979, 0.00000357, -0.00006637, -0.00005036, -0.00034795, 0.00088052, 0.00007193, -0.00031340, 0.00005946, -0.00011527, 0.00004080, -0.00003442, 0.00004533, -0.00010150, 0.00001255, 0.00000366, 0.00003124, -0.00006010, 0.00001270, -0.00000278, 0.00003063, -0.00004583, 0.00001255, 0.00000487, 0.00002128, -0.00002639, -0.00001246, 0.00001220, 0.00003969, -0.00005232, 0.00003255, -0.00003927, 0.00002234, -0.00004959, -0.00001218, -0.00000058, -0.00002460, 0.00003194, -0.00009302, 0.00003060, -0.00002117, 0.00017153, 0.00003333, 0.00003061, -0.00008442, 0.00001585, -0.00001937, -0.00000508, -0.00002271, },
		{ -0.00000490, 0.00003453, -0.00000720, -0.00002279, -0.00000007, 0.00000302, 0.00000055, -0.00000351, -0.00000699, 0.00002721, -0.00000593, -0.00004590, 0.00001282, 0.00002893, -0.00000585, -0.00003771, 0.00002444, -0.00008184, 0.00003534, -0.00032187, 0.00013175, 0.00091598, -0.00046767, 0.00002166, -0.00011026, 0.00001662, -0.00004182, 0.00005182, -0.00005579, -0.00003742, -0.00000586, 0.00004330, -0.00002877, -0.00002267, -0.00000016, 0.00002351, -0.00000930, -0.00002393, 0.00001206, 0.00001493, 0.00000085, -0.00002316, -0.00000968, 0.00003930, -0.00000354, -0.00002807, 0.00002639, -0.00004617, 0.00003290, -0.00008304, 0.00004647, -0.00006705, 0.00005933, -0.00006868, -0.00001230, -0.00000921, 0.00006562, 0.00012294, 0.00001518, -0.00000373, -0.00006907, 0.00003071, -0.00002957, -0.00000615, },
		{ -0.00002327, 0.00003113, 0.00000548, -0.00000159, -0.00003534, 0.00002876, -0.00002610, 0.00002867, -0.00003836, 0.00003663, -0.00000481, -0.00000186, -0.00006143, 0.00006699, -0.00002231, 0.00001241, -0.00005420, 0.00004305, -0.00011985, 0.00004273, -0.00049046, 0.00086003, 0.00027633, -0.00044382, 0.00011899, -0.00016739, 0.00007784, -0.00006874, 0.00007572, -0.00011555, 0.00002204, -0.00001835, 0.00005645, -0.00007168, 0.00001901, -0.00001512, 0.00003868, -0.00004497, 0.00001098, -0.00000057, 0.00002615, -0.00002468, -0.00000868, -0.00000495, 0.00005408, -0.00005154, 0.00002722, -0.00002727, 0.00000104, -0.00002086, -0.00003453, 0.00001612, -0.00004464, 0.00006011, -0.00011459, 0.00005871, -0.00004680, 0.00015513, 0.00002212, 0.00005061, -0.00007336, -0.00000157, -0.00000319, -0.00001027, },
		{ -0.00000384, -0.00001423, 0.00003731, -0.00000551, -0.00001574, -0.00001110, 0.00001287, -0.00001125, 0.00000979, -0.00002196, 0.00003505, -0.00000477, -0.00003806, -0.00000563, 0.00003920, -0.00000959, -0.00003173, 0.00001620, -0.00006183, 0.00000457, -0.00023709, -0.00011452, 0.00113268, -0.00046160, -0.00002951, -0.00008614, -0.00001001, -0.00002220, 0.00003092, -0.00002984, -0.00005166, -0.00000315, 0.00003205, -0.00000610, -0.00003865, 0.00000816, 0.00000974, 0.00000525, -0.00003055, 0.00001283, 0.00000951, 0.00000864, -0.00002287, -0.00001094, 0.00002456, 0.00002175, -0.00004744, 0.00004428, -0.00006035, 0.00003876, -0.00007527, 0.00003685, -0.00006128, 0.00005249, -0.00004932, -0.00003073, 0.00002085, 0.00001670, 0.00013280, -0.00000746, 0.00003928, -0.00009212, 0.00004007, -0.00002856, },
		{ -0.00001563, -0.00002244, 0.00002389, 0.00000834, 0.00000560, -0.00004034, 0.00003031, -0.00002855, 0.00003000, -0.00003797, 0.00003094, -0.00000253, 0.00001074, -0.00007511, 0.00006963, -0.00002905, 0.00003032, -0.00008044, 0.00008277, -0.00016836, 0.00011837, -0.00057314, 0.00073911, 0.00054992, -0.00059118, 0.00018158, -0.00022121, 0.00011460, -0.00010152, 0.00010220, -0.00012654, 0.00003280, -0.00004114, 0.00007667, -0.00007608, 0.00002064, -0.00002262, 0.00004116, -0.00004104, 0.00000932, -0.00000596, 0.00002957, -0.00002275, -0.00000353, -0.00001962, 0.00005938, -0.00003997, 0.00001139, -0.00000526, -0.00002584, 0.00000725, -0.00005253, 0.00002957, -0.00006189, 0.00008323, -0.00012891, 0.00007545, -0.00005749, 0.00013071, 0.00002116, 0.00004875, -0.00004364, -0.00002902, 0.00001648, },
		{ -0.00001991, -0.00000401, -0.00002189, 0.00003893, -0.00000601, -0.00000507, -0.00002490, 0.00002560, -0.00002571, 0.00002479, -0.00003694, 0.00004200, -0.00000674, -0.00002310, -0.00002986, 0.00005475, -0.00002274, -0.00001164, -0.00001202, -0.00001437, -0.00006262, -0.00010518, -0.00037937, 0.00129547, -0.00037721, -0.00013139, -0.00002874, -0.00006319, 0.00001890, -0.00000797, 0.00000882, -0.00007524, 0.00001171, 0.00000672, 0.00002584, -0.00005880, 0.00002048, -0.00000679, 0.00001968, -0.00003681, 0.00001548, 0.00000120, 0.00001795, -0.00002482, -0.00000613, 0.00000378, 0.00004691, -0.00006170, 0.00005319, -0.00006093, 0.00002982, -0.00005567, 0.00001955, -0.00004775, 0.00003620, -0.00002071, -0.00005824, 0.00005700, -0.00003123, 0.00014586, -0.00003464, 0.00007382, -0.00010092, 0.00003702, },
		{ 0.00003383, -0.00001787, -0.00002161, 0.00001601, 0.00001098, 0.00001062, -0.00004060, 0.00002671, -0.00002506, 0.00002474, -0.00003109, 0.00002071, 0.00000063, 0.00002151, -0.00008264, 0.00006626, -0.00003209, 0.00004335, -0.00009896, 0.00010936, -0.00019408, 0.00015362, -0.00057444, 0.00051554, 0.00087191, -0.00073023, 0.00022811, -0.00026178, 0.00013847, -0.00012206, 0.00011613, -0.00012781, 0.00003824, -0.00005697, 0.00008494, -0.00006903, 0.00001498, -0.00002347, 0.00003760, -0.00003419, 0.00000708, -0.00001005, 0.00003024, -0.00001926, 0.00000054, -0.00002841, 0.00005367, -0.00001844, -0.00001167, 0.00002064, -0.00005030, 0.00002726, -0.00006050, 0.00003514, -0.00007140, 0.00009607, -0.00013097, 0.00007665, -0.00005059, 0.00009474, 0.00003461, 0.00002600, -0.00000149, -0.00005930, },
		{ 0.00002093, -0.00000508, -0.00000531, -0.00002839, 0.00003974, -0.00000851, 0.00000792, -0.00003949, 0.00003876, -0.00003989, 0.00003846, -0.00004944, 0.00004710, -0.00001190, -0.00000180, -0.00005829, 0.00007460, -0.00004461, 0.00002107, -0.00005699, 0.00005369, -0.00015301, 0.00004783, -0.00062109, 0.00136461, -0.00019554, -0.00028773, 0.00006176, -0.00014144, 0.00007902, -0.00006166, 0.00005708, -0.00010629, 0.00003704, -0.00002893, 0.00006174, -0.00007872, 0.00003349, -0.00002329, 0.00003249, -0.00004254, 0.00002016, -0.00000963, 0.00002826, -0.00002869, 0.00000318, -0.00001890, 0.00006657, -0.00006605, 0.00004954, -0.00004709, 0.00000881, -0.00002900, -0.00000061, -0.00003082, 0.00001542, 0.00001120, -0.00008779, 0.00009153, -0.00007121, 0.00015393, -0.00005481, 0.00009004, -0.00009090, },
		{ -0.00008359, 0.00004312, -0.00001456, -0.00002143, 0.00000765, 0.00001409, 0.00001211, -0.00003495, 0.00001703, -0.00001501, 0.00001287, -0.00001826, 0.00000673, 0.00000519, 0.00002775, -0.00008065, 0.00005362, -0.00002679, 0.00004463, -0.00010062, 0.00011122, -0.00018343, 0.00013432, -0.00048839, 0.00020711, 0.00120510, -0.00082682, 0.00023543, -0.00027205, 0.00013596, -0.00011962, 0.00010931, -0.00011309, 0.00003205, -0.00005906, 0.00007636, -0.00004865, 0.00000136, -0.00001737, 0.00002837, -0.00002471, 0.00000352, -0.00001133, 0.00002683, -0.00001298, 0.00000150, -0.00002940, 0.00003756, 0.00000938, -0.00003614, 0.00004279, -0.00006461, 0.00003360, -0.00005529, 0.00003072, -0.00007087, 0.00009653, -0.00011935, 0.00006213, -0.00002750, 0.00004847, 0.00006021, -0.00001021, 0.00004265, },
		{ -0.00006300, -0.00000455, 0.00001196, -0.00000537, -0.00003446, 0.00003988, -0.00001222, 0.00002117, -0.00005217, 0.00004926, -0.00005052, 0.00004785, -0.00005726, 0.00004943, -0.00001954, 0.00002341, -0.00008732, 0.00009529, -0.00007107, 0.00006022, -0.00010940, 0.00012813, -0.00024523, 0.00018917, -0.00079792, 0.00131036, 0.00008723, -0.00048871, 0.00017463, -0.00023478, 0.00014868, -0.00012130, 0.00010731, -0.00013895, 0.00006713, -0.00006760, 0.00009422, -0.00009317, 0.00004360, -0.00003706, 0.00004233, -0.00004746, 0.00002638, -0.00002179, 0.00003824, -0.00003315, 0.00001433, -0.00003893, 0.00007636, -0.00005816, 0.00003349, -0.00002250, -0.00001757, -0.00000280, -0.00001692, -0.00001644, -0.00000345, 0.00003926, -0.00011168, 0.00011658, -0.00009635, 0.00014942, -0.00005934, 0.00008335, },
		{ 0.00007733, -0.00009427, 0.00004077, -0.00000519, -0.00002142, -0.00000148, 0.00001833, 0.00000928, -0.00002327, 0.00000176, 0.00000056, -0.00000404, -0.00000129, -0.00000955, 0.00001147, 0.00002717, -0.00006686, 0.00002978, -0.00001006, 0.00002955, -0.00007984, 0.00008236, -0.00013153, 0.00005940, -0.00032721, -0.00014973, 0.00150106, -0.00084449, 0.00018227, -0.00023801, 0.00009698, -0.00008706, 0.00007701, -0.00007905, 0.00001057, -0.00004381, 0.00004975, -0.00001628, -0.00001868, -0.00000547, 0.00001464, -0.00001299, -0.00000217, -0.00000852, 0.00001847, -0.00000321, -0.00000167, -0.00002252, 0.00001397, 0.00003800, -0.00005547, 0.00005445, -0.00006394, 0.00002468, -0.00003782, 0.00001797, -0.00006172, 0.00008646, -0.00009686, 0.00003600, 0.00000612, -0.00000220, 0.00009053, -0.00004848, },
		{ 0.00005591, -0.00002362, -0.00003249, 0.00002589, -0.00000175, -0.00004053, 0.00003904, -0.00001583, 0.00003224, -0.00006013, 0.00005426, -0.00005492, 0.00005076, -0.00005893, 0.00004830, -0.00002810, 0.00004859, -0.00011180, 0.00011148, -0.00009531, 0.00009625, -0.00015608, 0.00019098, -0.00031534, 0.00028721, -0.00087757, 0.00112131, 0.00045500, -0.00070940, 0.00028871, -0.00032555, 0.00021263, -0.00017381, 0.00014879, -0.00016473, 0.00009375, -0.00010030, 0.00011558, -0.00009746, 0.00004784, -0.00004598, 0.00004820, -0.00005094, 0.00003292, -0.00003332, 0.00004593, -0.00003626, 0.00002425, -0.00005245, 0.00007417, -0.00003899, 0.00000901, 0.00000565, -0.00004053, 0.00001494, -0.00002316, -0.00000984, -0.00001488, 0.00005767, -0.00012408, 0.00012678, -0.00010262, 0.00012851, -0.00004534, },
		{ -0.00007721, 0.00009358, -0.00008744, 0.00002668, 0.00000831, -0.00002004, -0.00001190, 0.00002402, 0.00000226, -0.00000675, -0.00001713, 0.00001916, -0.00002319, 0.00001717, -0.00002627, 0.00001949, 0.00001856, -0.00004098, -0.00000470, 0.00001820, -0.00000239, -0.00003678, 0.00002476, -0.00004421, -0.00005819, -0.00011941, -0.00050588, 0.00170899, -0.00075314, 0.00005620, -0.00015390, 0.00001899, -0.00002401, 0.00002030, -0.00002720, -0.00002553, -0.00001211, 0.00000829, 0.00002354, -0.00004165, 0.00000990, -0.00000190, 0.00000044, -0.00001053, -0.00000098, 0.00000516, 0.00000975, -0.00000870, -0.00000949, -0.00001259, 0.00006155, -0.00006436, 0.00005195, -0.00004801, 0.00000382, -0.00001323, 0.00000211, -0.00004874, 0.00007115, -0.00006971, 0.00000557, 0.00004180, -0.00004870, 0.00011575, },
		{ -0.00001629, 0.00001580, 0.00001724, -0.00005477, 0.00003183, 0.00000698, -0.00004624, 0.00003650, -0.00001781, 0.00003887, -0.00006120, 0.00005189, -0.00005169, 0.00004633, -0.00005403, 0.00004333, -0.00003548, 0.00006896, -0.00012601, 0.00011728, -0.00010972, 0.00011878, -0.00018363, 0.00022542, -0.00034321, 0.00031927, -0.00084494, 0.00080910, 0.00087144, -0.00091234, 0.00037591, -0.00039193, 0.00025328, -0.00020501, 0.00017043, -0.00017488, 0.00010851, -0.00011867, 0.00011985, -0.00008879, 0.00004465, -0.00004882, 0.00004940, -0.00005212, 0.00003801, -0.00004184, 0.00004921, -0.00003607, 0.00003036, -0.00005720, 0.00006058, -0.00001253, -0.00001731, 0.00002886, -0.00005198, 0.00001848, -0.00001589, -0.00001340, -0.00001632, 0.00006401, -0.00012309, 0.00012108, -0.00009061, 0.00009268, },
		{ 0.00012734, -0.00008824, 0.00008773, -0.00006437, 0.00000456, 0.00002201, -0.00001518, -0.00002395, 0.00003089, -0.00000782, 0.00001225, -0.00003657, 0.00003740, -0.00004125, 0.00003423, -0.00004144, 0.00002877, 0.00000228, -0.00000534, -0.00004623, 0.00005446, -0.00004690, 0.00002202, -0.00005155, 0.00006298, -0.00019444, 0.00009631, -0.00080914, 0.00178643, -0.00053776, -0.00014064, -0.00002594, -0.00009064, 0.00006169, -0.00005302, 0.00003557, -0.00007088, 0.00003060, -0.00004100, 0.00006392, -0.00006300, 0.00002587, -0.00001929, 0.00001480, -0.00002150, 0.00001093, -0.00001196, 0.00002453, -0.00001812, 0.00000674, -0.00003724, 0.00007544, -0.00006039, 0.00003589, -0.00002141, -0.00002154, 0.00001045, -0.00000966, -0.00003836, 0.00005741, -0.00004532, -0.00002113, 0.00007090, -0.00008270, },
		{ 0.00004837, 0.00001930, -0.00002552, 0.00004896, -0.00006462, 0.00002704, 0.00002033, -0.00005024, 0.00003120, -0.00001680, 0.00003951, -0.00005440, 0.00004185, -0.00004106, 0.00003515, -0.00004319, 0.00003455, -0.00003940, 0.00007995, -0.00012518, 0.00010805, -0.00010815, 0.00011969, -0.00018232, 0.00022056, -0.00031808, 0.00027734, -0.00070601, 0.00040819, 0.00128462, -0.00105414, 0.00040749, -0.00041350, 0.00025543, -0.00020362, 0.00016408, -0.00016323, 0.00010539, -0.00011728, 0.00010445, -0.00006717, 0.00003422, -0.00004541, 0.00004564, -0.00004993, 0.00003964, -0.00004508, 0.00004637, -0.00003123, 0.00003116, -0.00005290, 0.00003860, 0.00001538, -0.00003826, 0.00003996, -0.00004721, 0.00000648, 0.00000389, -0.00002537, -0.00000934, 0.00006015, -0.00011130, 0.00010311, -0.00006562, },
		{ -0.00009904, 0.00012095, -0.00007926, 0.00006241, -0.00003124, -0.00001906, 0.00003110, -0.00000512, -0.00003726, 0.00003795, -0.00001907, 0.00003070, -0.00005308, 0.00005185, -0.00005513, 0.00004734, -0.00005327, 0.00003832, -0.00001954, 0.00003520, -0.00008850, 0.00009207, -0.00009561, 0.00008507, -0.00013083, 0.00016851, -0.00032016, 0.00027976, -0.00101503, 0.00170912, -0.00020465, -0.00038873, 0.00012688, -0.00021484, 0.00015479, -0.00012962, 0.00009818, -0.00011659, 0.00007568, -0.00008889, 0.00009727, -0.00007822, 0.00003967, -0.00003557, 0.00002902, -0.00003420, 0.00002567, -0.00003063, 0.00003885, -0.00002767, 0.00002280, -0.00005589, 0.00007750, -0.00004474, 0.00001103, 0.00000764, -0.00004198, 0.00002483, -0.00001063, -0.00003622, 0.00005100, -0.00002969, -0.00003798, 0.00008731, },
		{ -0.00003600, 0.00000480, 0.00005094, -0.00005684, 0.00006360, -0.00005887, 0.00001207, 0.00003576, -0.00005048, 0.00002224, -0.00001207, 0.00003376, -0.00004031, 0.00002551, -0.00002486, 0.00001906, -0.00002791, 0.00002239, -0.00003800, 0.00007827, -0.00010692, 0.00008209, -0.00008797, 0.00009558, -0.00014904, 0.00017469, -0.00024158, 0.00016996, -0.00048690, -0.00002976, 0.00163608, -0.00109470, 0.00036171, -0.00037724, 0.00021098, -0.00016490, 0.00012734, -0.00012840, 0.00008268, -0.00009514, 0.00007095, -0.00003553, 0.00001840, -0.00003648, 0.00003700, -0.00004351, 0.00003617, -0.00004151, 0.00003668, -0.00002148, 0.00002653, -0.00004114, 0.00001280, 0.00003868, -0.00004830, 0.00003541, -0.00002663, -0.00001718, 0.00003068, -0.00004015, 0.00000094, 0.00005155, -0.00009481, 0.00007985, },
		{ 0.00008935, -0.00009747, 0.00009787, -0.00005429, 0.00002573, 0.00000271, -0.00003685, 0.00003155, 0.00001057, -0.00005053, 0.00004362, -0.00002917, 0.00004550, -0.00006361, 0.00005986, -0.00006271, 0.00005479, -0.00006043, 0.00004668, -0.00004329, 0.00007400, -0.00012373, 0.00012280, -0.00013818, 0.00013875, -0.00019565, 0.00025044, -0.00040829, 0.00039836, -0.00109612, 0.00147747, 0.00021679, -0.00065282, 0.00027567, -0.00033005, 0.00023583, -0.00019359, 0.00014803, -0.00015251, 0.00011351, -0.00012613, 0.00011706, -0.00008403, 0.00004918, -0.00004901, 0.00004169, -0.00004692, 0.00004075, -0.00004787, 0.00005005, -0.00003493, 0.00003573, -0.00006613, 0.00006839, -0.00002177, -0.00001525, 0.00002985, -0.00004915, 0.00002403, 0.00000287, -0.00004483, 0.00005446, -0.00002522, -0.00004286, },
		{ 0.00005912, -0.00001047, -0.00002911, 0.00006970, -0.00007061, 0.00005823, -0.00003908, -0.00000915, 0.00004914, -0.00004483, 0.00000927, -0.00000380, 0.00002251, -0.00002097, 0.00000564, -0.00000602, 0.00000067, -0.00001020, 0.00000774, -0.00003040, 0.00006293, -0.00007226, 0.00004162, -0.00005119, 0.00004911, -0.00008848, 0.00009592, -0.00012742, 0.00001994, -0.00022804, -0.00044567, 0.00187220, -0.00100700, 0.00023078, -0.00028221, 0.00012217, -0.00009281, 0.00006489, -0.00007472, 0.00004364, -0.00005599, 0.00002483, 0.00000088, 0.00000019, -0.00002344, 0.00002403, -0.00003246, 0.00002685, -0.00003085, 0.00002073, -0.00000779, 0.00001771, -0.00002477, -0.00001201, 0.00005283, -0.00004513, 0.00001647, 0.00000414, -0.00004435, 0.00005599, -0.00004998, 0.00000762, 0.00004517, -0.00008095, },
		{ -0.00003820, 0.00008010, -0.00008262, 0.00006443, -0.00002207, -0.00001139, 0.00002797, -0.00004301, 0.00002154, 0.00003044, -0.00006154, 0.00004604, -0.00003595, 0.00005425, -0.00006623, 0.00006017, -0.00006323, 0.00005595, -0.00006212, 0.00005219, -0.00006456, 0.00010402, -0.00014450, 0.00013904, -0.00016506, 0.00017093, -0.00023148, 0.00029203, -0.00044056, 0.00043428, -0.00104762, 0.00111775, 0.00067761, -0.00088773, 0.00038770, -0.00041169, 0.00028596, -0.00023052, 0.00017427, -0.00017015, 0.00013607, -0.00014572, 0.00011943, -0.00007914, 0.00005335, -0.00005817, 0.00005119, -0.00005738, 0.00005318, -0.00006056, 0.00005569, -0.00003799, 0.00004365, -0.00006756, 0.00005121, 0.00000235, -0.00003517, 0.00003771, -0.00003858, 0.00000695, 0.00002981, -0.00006231, 0.00006602, -0.00002994, },
		{ -0.00007559, 0.00004689, 0.00000472, -0.00004764, 0.00007102, -0.00006503, 0.00003581, -0.00001115, -0.00003043, 0.00005596, -0.00003193, -0.00000701, 0.00000680, 0.00000777, 0.00000051, -0.00001428, 0.00001218, -0.00001726, 0.00000777, -0.00000809, -0.00001705, 0.00003570, -0.00002584, -0.00000731, -0.00000422, -0.00001154, -0.00001209, 0.00000022, 0.00000217, -0.00014143, 0.00002487, -0.00078472, 0.00195541, -0.00078472, 0.00002487, -0.00014143, 0.00000217, 0.00000022, -0.00001209, -0.00001154, -0.00000422, -0.00000731, -0.00002584, 0.00003570, -0.00001705, -0.00000809, 0.00000777, -0.00001726, 0.00001218, -0.00001428, 0.00000051, 0.00000777, 0.00000680, -0.00000701, -0.00003193, 0.00005596, -0.00003043, -0.00001115, 0.00003581, -0.00006503, 0.00007102, -0.00004764, 0.00000472, 0.00004689, },
		{ -0.00003820, -0.00002994, 0.00006602, -0.00006231, 0.00002981, 0.00000695, -0.00003858, 0.00003771, -0.00003517, 0.00000235, 0.00005121, -0.00006756, 0.00004365, -0.00003799, 0.00005569, -0.00006056, 0.00005318, -0.00005738, 0.00005119, -0.00005817, 0.00005335, -0.00007914, 0.00011943, -0.00014572, 0.00013607, -0.00017015, 0.00017427, -0.00023052, 0.00028596, -0.00041169, 0.00038770, -0.00088773, 0.00067761, 0.00111775, -0.00104762, 0.00043428, -0.00044056, 0.00029203, -0.00023148, 0.00017093, -0.00016506, 0.00013904, -0.00014450, 0.00010402, -0.00006456, 0.00005219, -0.00006212, 0.00005595, -0.00006323, 0.00006017, -0.00006623, 0.00005425, -0.00003595, 0.00004604, -0.00006154, 0.00003044, 0.00002154, -0.00004301, 0.00002797, -0.00001139, -0.00002207, 0.00006443, -0.00008262, 0.00008010, },
		{ 0.00005912, -0.00008095, 0.00004517, 0.00000762, -0.00004998, 0.00005599, -0.00004435, 0.00000414, 0.00001647, -0.00004513, 0.00005283, -0.00001201, -0.00002477, 0.00001771, -0.00000779, 0.00002073, -0.00003085, 0.00002685, -0.00003246, 0.00002403, -0.00002344, 0.00000019, 0.00000088, 0.00002483, -0.00005599, 0.00004364, -0.00007472, 0.00006489, -0.00009281, 0.00012217, -0.00028221, 0.00023078, -0.00100700, 0.00187220, -0.00044567, -0.00022804, 0.00001994, -0.00012742, 0.00009592, -0.00008848, 0.00004911, -0.00005119, 0.00004162, -0.00007226, 0.00006293, -0.00003040, 0.00000774, -0.00001020, 0.00000067, -0.00000602, 0.00000564, -0.00002097, 0.00002251, -0.00000380, 0.00000927, -0.00004483, 0.00004914, -0.00000915, -0.00003908, 0.00005823, -0.00007061, 0.00006970, -0.00002911, -0.00001047, },
		{ 0.00008935, -0.00004286, -0.00002522, 0.00005446, -0.00004483, 0.00000287, 0.00002403, -0.00004915, 0.00002985, -0.00001525, -0.00002177, 0.00006839, -0.00006613, 0.00003573, -0.00003493, 0.00005005, -0.00004787, 0.00004075, -0.00004692, 0.00004169, -0.00004901, 0.00004918, -0.00008403, 0.00011706, -0.00012613, 0.00011351, -0.00015251, 0.00014803, -0.00019359, 0.00023583, -0.00033005, 0.00027567, -0.00065282, 0.00021679, 0.00147747, -0.00109612, 0.00039836, -0.00040829, 0.00025044, -0.00019565, 0.00013875, -0.00013818, 0.00012280, -0.00012373, 0.00007400, -0.00004329, 0.00004668, -0.00006043, 0.00005479, -0.00006271, 0.00005986, -0.00006361, 0.00004550, -0.00002917, 0.00004362, -0.00005053, 0.00001057, 0.00003155, -0.00003685, 0.00000271, 0.00002573, -0.00005429, 0.00009787, -0.00009747, },
		{ -0.00003600, 0.00007985, -0.00009481, 0.00005155, 0.00000094, -0.00004015, 0.00003068, -0.00001718, -0.00002663, 0.00003541, -0.00004830, 0.00003868, 0.00001280, -0.00004114, 0.00002653, -0.00002148, 0.00003668, -0.00004151, 0.00003617, -0.00004351, 0.00003700, -0.00003648, 0.00001840, -0.00003553, 0.00007095, -0.00009514, 0.00008268, -0.00012840, 0.00012734, -0.00016490, 0.00021098, -0.00037724, 0.00036171, -0.00109470, 0.00163608, -0.00002976, -0.00048690, 0.00016996, -0.00024158, 0.00017469, -0.00014904, 0.00009558, -0.00008797, 0.00008209, -0.00010692, 0.00007827, -0.00003800, 0.00002239, -0.00002791, 0.00001906, -0.00002486, 0.00002551, -0.00004031, 0.00003376, -0.00001207, 0.00002224, -0.00005048, 0.00003576, 0.00001207, -0.00005887, 0.00006360, -0.00005684, 0.00005094, 0.00000480, },
		{ -0.00009904, 0.00008731, -0.00003798, -0.00002969, 0.00005100, -0.00003622, -0.00001063, 0.00002483, -0.00004198, 0.00000764, 0.00001103, -0.00004474, 0.00007750, -0.00005589, 0.00002280, -0.00002767, 0.00003885, -0.00003063, 0.00002567, -0.00003420, 0.00002902, -0.00003557, 0.00003967, -0.00007822, 0.00009727, -0.00008889, 0.00007568, -0.00011659, 0.00009818, -0.00012962, 0.00015479, -0.00021484, 0.00012688, -0.00038873, -0.00020465, 0.00170912, -0.00101503, 0.00027976, -0.00032016, 0.00016851, -0.00013083, 0.00008507, -0.00009561, 0.00009207, -0.00008850, 0.00003520, -0.00001954, 0.00003832, -0.00005327, 0.00004734, -0.00005513, 0.00005185, -0.00005308, 0.00003070, -0.00001907, 0.00003795, -0.00003726, -0.00000512, 0.00003110, -0.00001906, -0.00003124, 0.00006241, -0.00007926, 0.00012095, },
		{ 0.00004837, -0.00006562, 0.00010311, -0.00011130, 0.00006015, -0.00000934, -0.00002537, 0.00000389, 0.00000648, -0.00004721, 0.00003996, -0.00003826, 0.00001538, 0.00003860, -0.00005290, 0.00003116, -0.00003123, 0.00004637, -0.00004508, 0.00003964, -0.00004993, 0.00004564, -0.00004541, 0.00003422, -0.00006717, 0.00010445, -0.00011728, 0.00010539, -0.00016323, 0.00016408, -0.00020362, 0.00025543, -0.00041350, 0.00040749, -0.00105414, 0.00128462, 0.00040819, -0.00070601, 0.00027734, -0.00031808, 0.00022056, -0.00018232, 0.00011969, -0.00010815, 0.00010805, -0.00012518, 0.00007995, -0.00003940, 0.00003455, -0.00004319, 0.00003515, -0.00004106, 0.00004185, -0.00005440, 0.00003951, -0.00001680, 0.00003120, -0.00005024, 0.00002033, 0.00002704, -0.00006462, 0.00004896, -0.00002552, 0.00001930, },
		{ 0.00012734, -0.00008270, 0.00007090, -0.00002113, -0.00004532, 0.00005741, -0.00003836, -0.00000966, 0.00001045, -0.00002154, -0.00002141, 0.00003589, -0.00006039, 0.00007544, -0.00003724, 0.00000674, -0.00001812, 0.00002453, -0.00001196, 0.00001093, -0.00002150, 0.00001480, -0.00001929, 0.00002587, -0.00006300, 0.00006392, -0.00004100, 0.00003060, -0.00007088, 0.00003557, -0.00005302, 0.00006169, -0.00009064, -0.00002594, -0.00014064, -0.00053776, 0.00178643, -0.00080914, 0.00009631, -0.00019444, 0.00006298, -0.00005155, 0.00002202, -0.00004690, 0.00005446, -0.00004623, -0.00000534, 0.00000228, 0.00002877, -0.00004144, 0.00003423, -0.00004125, 0.00003740, -0.00003657, 0.00001225, -0.00000782, 0.00003089, -0.00002395, -0.00001518, 0.00002201, 0.00000456, -0.00006437, 0.00008773, -0.00008824, },
		{ -0.00001629, 0.00009268, -0.00009061, 0.00012108, -0.00012309, 0.00006401, -0.00001632, -0.00001340, -0.00001589, 0.00001848, -0.00005198, 0.00002886, -0.00001731, -0.00001253, 0.00006058, -0.00005720, 0.00003036, -0.00003607, 0.00004921, -0.00004184, 0.00003801, -0.00005212, 0.00004940, -0.00004882, 0.00004465, -0.00008879, 0.00011985, -0.00011867, 0.00010851, -0.00017488, 0.00017043, -0.00020501, 0.00025328, -0.00039193, 0.00037591, -0.00091234, 0.00087144, 0.00080910, -0.00084494, 0.00031927, -0.00034321, 0.00022542, -0.00018363, 0.00011878, -0.00010972, 0.00011728, -0.00012601, 0.00006896, -0.00003548, 0.00004333, -0.00005403, 0.00004633, -0.00005169, 0.00005189, -0.00006120, 0.00003887, -0.00001781, 0.00003650, -0.00004624, 0.00000698, 0.00003183, -0.00005477, 0.00001724, 0.00001580, },
		{ -0.00007721, 0.00011575, -0.00004870, 0.00004180, 0.00000557, -0.00006971, 0.00007115, -0.00004874, 0.00000211, -0.00001323, 0.00000382, -0.00004801, 0.00005195, -0.00006436, 0.00006155, -0.00001259, -0.00000949, -0.00000870, 0.00000975, 0.00000516, -0.00000098, -0.00001053, 0.00000044, -0.00000190, 0.00000990, -0.00004165, 0.00002354, 0.00000829, -0.00001211, -0.00002553, -0.00002720, 0.00002030, -0.00002401, 0.00001899, -0.00015390, 0.00005620, -0.00075314, 0.00170899, -0.00050588, -0.00011941, -0.00005819, -0.00004421, 0.00002476, -0.00003678, -0.00000239, 0.00001820, -0.00000470, -0.00004098, 0.00001856, 0.00001949, -0.00002627, 0.00001717, -0.00002319, 0.00001916, -0.00001713, -0.00000675, 0.00000226, 0.00002402, -0.00001190, -0.00002004, 0.00000831, 0.00002668, -0.00008744, 0.00009358, },
		{ 0.00005591, -0.00004534, 0.00012851, -0.00010262, 0.00012678, -0.00012408, 0.00005767, -0.00001488, -0.00000984, -0.00002316, 0.00001494, -0.00004053, 0.00000565, 0.00000901, -0.00003899, 0.00007417, -0.00005245, 0.00002425, -0.00003626, 0.00004593, -0.00003332, 0.00003292, -0.00005094, 0.00004820, -0.00004598, 0.00004784, -0.00009746, 0.00011558, -0.00010030, 0.00009375, -0.00016473, 0.00014879, -0.00017381, 0.00021263, -0.00032555, 0.00028871, -0.00070940, 0.00045500, 0.00112131, -0.00087757, 0.00028721, -0.00031534, 0.00019098, -0.00015608, 0.00009625, -0.00009531, 0.00011148, -0.00011180, 0.00004859, -0.00002810, 0.00004830, -0.00005893, 0.00005076, -0.00005492, 0.00005426, -0.00006013, 0.00003224, -0.00001583, 0.00003904, -0.00004053, -0.00000175, 0.00002589, -0.00003249, -0.00002362, },
		{ 0.00007733, -0.00004848, 0.00009053, -0.00000220, 0.00000612, 0.00003600, -0.00009686, 0.00008646, -0.00006172, 0.00001797, -0.00003782, 0.00002468, -0.00006394, 0.00005445, -0.00005547, 0.00003800, 0.00001397, -0.00002252, -0.00000167, -0.00000321, 0.00001847, -0.00000852, -0.00000217, -0.00001299, 0.00001464, -0.00000547, -0.00001868, -0.00001628, 0.00004975, -0.00004381, 0.00001057, -0.00007905, 0.00007701, -0.00008706, 0.00009698, -0.00023801, 0.00018227, -0.00084449, 0.00150106, -0.00014973, -0.00032721, 0.00005940, -0.00013153, 0.00008236, -0.00007984, 0.00002955, -0.00001006, 0.00002978, -0.00006686, 0.00002717, 0.00001147, -0.00000955, -0.00000129, -0.00000404, 0.00000056, 0.00000176, -0.00002327, 0.00000928, 0.00001833, -0.00000148, -0.00002142, -0.00000519, 0.00004077, -0.00009427, },
		{ -0.00006300, 0.00008335, -0.00005934, 0.00014942, -0.00009635, 0.00011658, -0.00011168, 0.00003926, -0.00000345, -0.00001644, -0.00001692, -0.00000280, -0.00001757, -0.00002250, 0.00003349, -0.00005816, 0.00007636, -0.00003893, 0.00001433, -0.00003315, 0.00003824, -0.00002179, 0.00002638, -0.00004746, 0.00004233, -0.00003706, 0.00004360, -0.00009317, 0.00009422, -0.00006760, 0.00006713, -0.00013895, 0.00010731, -0.00012130, 0.00014868, -0.00023478, 0.00017463, -0.00048871, 0.00008723, 0.00131036, -0.00079792, 0.00018917, -0.00024523, 0.00012813, -0.00010940, 0.00006022, -0.00007107, 0.00009529, -0.00008732, 0.00002341, -0.00001954, 0.00004943, -0.00005726, 0.00004785, -0.00005052, 0.00004926, -0.00005217, 0.00002117, -0.00001222, 0.00003988, -0.00003446, -0.00000537, 0.00001196, -0.00000455, },
		{ -0.00008359, 0.00004265, -0.00001021, 0.00006021, 0.00004847, -0.00002750, 0.00006213, -0.00011935, 0.00009653, -0.00007087, 0.00003072, -0.00005529, 0.00003360, -0.00006461, 0.00004279, -0.00003614, 0.00000938, 0.00003756, -0.00002940, 0.00000150, -0.00001298, 0.00002683, -0.00001133, 0.00000352, -0.00002471, 0.00002837, -0.00001737, 0.00000136, -0.00004865, 0.00007636, -0.00005906, 0.00003205, -0.00011309, 0.00010931, -0.00011962, 0.00013596, -0.00027205, 0.00023543, -0.00082682, 0.00120510, 0.00020711, -0.00048839, 0.00013432, -0.00018343, 0.00011122, -0.00010062, 0.00004463, -0.00002679, 0.00005362, -0.00008065, 0.00002775, 0.00000519, 0.00000673, -0.00001826, 0.00001287, -0.00001501, 0.00001703, -0.00003495, 0.00001211, 0.00001409, 0.00000765, -0.00002143, -0.00001456, 0.00004312, },
		{ 0.00002093, -0.00009090, 0.00009004, -0.00005481, 0.00015393, -0.00007121, 0.00009153, -0.00008779, 0.00001120, 0.00001542, -0.00003082, -0.00000061, -0.00002900, 0.00000881, -0.00004709, 0.00004954, -0.00006605, 0.00006657, -0.00001890, 0.00000318, -0.00002869, 0.00002826, -0.00000963, 0.00002016, -0.00004254, 0.00003249, -0.00002329, 0.00003349, -0.00007872, 0.00006174, -0.00002893, 0.00003704, -0.00010629, 0.00005708, -0.00006166, 0.00007902, -0.00014144, 0.00006176, -0.00028773, -0.00019554, 0.00136461, -0.00062109, 0.00004783, -0.00015301, 0.00005369, -0.00005699, 0.00002107, -0.00004461, 0.00007460, -0.00005829, -0.00000180, -0.00001190, 0.00004710, -0.00004944, 0.00003846, -0.00003989, 0.00003876, -0.00003949, 0.00000792, -0.00000851, 0.00003974, -0.00002839, -0.00000531, -0.00000508, },
		{ 0.00003383, -0.00005930, -0.00000149, 0.00002600, 0.00003461, 0.00009474, -0.00005059, 0.00007665, -0.00013097, 0.00009607, -0.00007140, 0.00003514, -0.00006050, 0.00002726, -0.00005030, 0.00002064, -0.00001167, -0.00001844, 0.00005367, -0.00002841, 0.00000054, -0.00001926, 0.00003024, -0.00001005, 0.00000708, -0.00003419, 0.00003760, -0.00002347, 0.00001498, -0.00006903, 0.00008494, -0.00005697, 0.00003824, -0.00012781, 0.00011613, -0.00012206, 0.00013847, -0.00026178, 0.00022811, -0.00073023, 0.00087191, 0.00051554, -0.00057444, 0.00015362, -0.00019408, 0.00010936, -0.00009896, 0.00004335, -0.00003209, 0.00006626, -0.00008264, 0.00002151, 0.00000063, 0.00002071, -0.00003109, 0.00002474, -0.00002506, 0.00002671, -0.00004060, 0.00001062, 0.00001098, 0.00001601, -0.00002161, -0.00001787, },
		{ -0.00001991, 0.00003702, -0.00010092, 0.00007382, -0.00003464, 0.00014586, -0.00003123, 0.00005700, -0.00005824, -0.00002071, 0.00003620, -0.00004775, 0.00001955, -0.00005567, 0.00002982, -0.00006093, 0.00005319, -0.00006170, 0.00004691, 0.00000378, -0.00000613, -0.00002482, 0.00001795, 0.00000120, 0.00001548, -0.00003681, 0.00001968, -0.00000679, 0.00002048, -0.00005880, 0.00002584, 0.00000672, 0.00001171, -0.00007524, 0.00000882, -0.00000797, 0.00001890, -0.00006319, -0.00002874, -0.00013139, -0.00037721, 0.00129547, -0.00037937, -0.00010518, -0.00006262, -0.00001437, -0.00001202, -0.00001164, -0.00002274, 0.00005475, -0.00002986, -0.00002310, -0.00000674, 0.00004200, -0.00003694, 0.00002479, -0.00002571, 0.00002560, -0.00002490, -0.00000507, -0.00000601, 0.00003893, -0.00002189, -0.00000401, },
		{ -0.00001563, 0.00001648, -0.00002902, -0.00004364, 0.00004875, 0.00002116, 0.00013071, -0.00005749, 0.00007545, -0.00012891, 0.00008323, -0.00006189, 0.00002957, -0.00005253, 0.00000725, -0.00002584, -0.00000526, 0.00001139, -0.00003997, 0.00005938, -0.00001962, -0.00000353, -0.00002275, 0.00002957, -0.00000596, 0.00000932, -0.00004104, 0.00004116, -0.00002262, 0.00002064, -0.00007608, 0.00007667, -0.00004114, 0.00003280, -0.00012654, 0.00010220, -0.00010152, 0.00011460, -0.00022121, 0.00018158, -0.00059118, 0.00054992, 0.00073911, -0.00057314, 0.00011837, -0.00016836, 0.00008277, -0.00008044, 0.00003032, -0.00002905, 0.00006963, -0.00007511, 0.00001074, -0.00000253, 0.00003094, -0.00003797, 0.00003000, -0.00002855, 0.00003031, -0.00004034, 0.00000560, 0.00000834, 0.00002389, -0.00002244, },
		{ -0.00000384, -0.00002856, 0.00004007, -0.00009212, 0.00003928, -0.00000746, 0.00013280, 0.00001670, 0.00002085, -0.00003073, -0.00004932, 0.00005249, -0.00006128, 0.00003685, -0.00007527, 0.00003876, -0.00006035, 0.00004428, -0.00004744, 0.00002175, 0.00002456, -0.00001094, -0.00002287, 0.00000864, 0.00000951, 0.00001283, -0.00003055, 0.00000525, 0.00000974, 0.00000816, -0.00003865, -0.00000610, 0.00003205, -0.00000315, -0.00005166, -0.00002984, 0.00003092, -0.00002220, -0.00001001, -0.00008614, -0.00002951, -0.00046160, 0.00113268, -0.00011452, -0.00023709, 0.00000457, -0.00006183, 0.00001620, -0.00003173, -0.00000959, 0.00003920, -0.00000563, -0.00003806, -0.00000477, 0.00003505, -0.00002196, 0.00000979, -0.00001125, 0.00001287, -0.00001110, -0.00001574, -0.00000551, 0.00003731, -0.00001423, },
		{ -0.00002327, -0.00001027, -0.00000319, -0.00000157, -0.00007336, 0.00005061, 0.00002212, 0.00015513, -0.00004680, 0.00005871, -0.00011459, 0.00006011, -0.00004464, 0.00001612, -0.00003453, -0.00002086, 0.00000104, -0.00002727, 0.00002722, -0.00005154, 0.00005408, -0.00000495, -0.00000868, -0.00002468, 0.00002615, -0.00000057, 0.00001098, -0.00004497, 0.00003868, -0.00001512, 0.00001901, -0.00007168, 0.00005645, -0.00001835, 0.00002204, -0.00011555, 0.00007572, -0.00006874, 0.00007784, -0.00016739, 0.00011899, -0.00044382, 0.00027633, 0.00086003, -0.00049046, 0.00004273, -0.00011985, 0.00004305, -0.00005420, 0.00001241, -0.00002231, 0.00006699, -0.00006143, -0.00000186, -0.00000481, 0.00003663, -0.00003836, 0.00002867, -0.00002610, 0.00002876, -0.00003534, -0.00000159, 0.00000548, 0.00003113, },
		{ -0.00000490, -0.00000615, -0.00002957, 0.00003071, -0.00006907, -0.00000373, 0.00001518, 0.00012294, 0.00006562, -0.00000921, -0.00001230, -0.00006868, 0.00005933, -0.00006705, 0.00004647, -0.00008304, 0.00003290, -0.00004617, 0.00002639, -0.00002807, -0.00000354, 0.00003930, -0.00000968, -0.00002316, 0.00000085, 0.00001493, 0.00001206, -0.00002393, -0.00000930, 0.00002351, -0.00000016, -0.00002267, -0.00002877, 0.00004330, -0.00000586, -0.00003742, -0.00005579, 0.00005182, -0.00004182, 0.00001662, -0.00011026, 0.00002166, -0.00046767, 0.00091598, 0.00013175, -0.00032187, 0.00003534, -0.00008184, 0.00002444, -0.00003771, -0.00000585, 0.00002893, 0.00001282, -0.00004590, -0.00000593, 0.00002721, -0.00000699, -0.00000351, 0.00000055, 0.00000302, -0.00000007, -0.00002279, -0.00000720, 0.00003453, },
		{ 0.00003712, -0.00002271, -0.00000508, -0.00001937, 0.00001585, -0.00008442, 0.00003061, 0.00003333, 0.00017153, -0.00002117, 0.00003060, -0.00009302, 0.00003194, -0.00002460, -0.00000058, -0.00001218, -0.00004959, 0.00002234, -0.00003927, 0.00003255, -0.00005232, 0.00003969, 0.00001220, -0.00001246, -0.00002639, 0.00002128, 0.00000487, 0.00001255, -0.00004583, 0.00003063, -0.00000278, 0.00001270, -0.00006010, 0.00003124, 0.00000366, 0.00001255, -0.00010150, 0.00004533, -0.00003442, 0.00004080, -0.00011527, 0.00005946, -0.00031340, 0.00007193, 0.00088052, -0.00034795, -0.00005036, -0.00006637, 0.00000357, -0.00002979, -0.00000365, -0.00001632, 0.00006167, -0.00004498, -0.00001385, -0.00000674, 0.00003781, -0.00003307, 0.00002226, -0.00001965, 0.00002399, -0.00002746, -0.00000934, 0.00000192, },
		{ 0.00003030, 0.00000595, -0.00001078, -0.00002424, 0.00001327, -0.00004009, -0.00004321, 0.00002271, 0.00012157, 0.00011135, -0.00002797, -0.00000715, -0.00007593, 0.00005492, -0.00006373, 0.00004682, -0.00007812, 0.00001399, -0.00002319, 0.00000565, -0.00000938, -0.00002400, 0.00004534, -0.00000240, -0.00002497, -0.00000565, 0.00001773, 0.00001261, -0.00001706, -0.00002239, 0.00003220, -0.00000235, -0.00001329, -0.00004012, 0.00004089, 0.00000117, -0.00003056, -0.00007017, 0.00005671, -0.00004351, 0.00002203, -0.00010872, 0.00003493, -0.00042226, 0.00068568, 0.00032647, -0.00034597, 0.00002753, -0.00007614, 0.00001581, -0.00003259, -0.00000911, 0.00002276, 0.00002567, -0.00004724, -0.00000949, 0.00001940, 0.00000575, -0.00001280, 0.00000775, -0.00000269, 0.00000735, -0.00002586, -0.00001071, },
		{ -0.00000255, 0.00004108, -0.00001941, -0.00000284, -0.00002804, 0.00001982, -0.00007654, -0.00000528, 0.00004543, 0.00018629, 0.00001461, -0.00000265, -0.00007085, 0.00000498, -0.00000738, -0.00001536, 0.00000856, -0.00007206, 0.00003220, -0.00003848, 0.00002754, -0.00004442, 0.00002009, 0.00002789, -0.00001276, -0.00002872, 0.00001590, 0.00000952, 0.00001416, -0.00004370, 0.00001839, 0.00001153, 0.00000539, -0.00004651, 0.00000797, 0.00001853, 0.00000895, -0.00008914, 0.00001753, -0.00000629, 0.00001197, -0.00007430, 0.00001443, -0.00021322, -0.00005953, 0.00081932, -0.00017658, -0.00013568, -0.00002461, -0.00002464, -0.00001427, -0.00001332, -0.00001390, 0.00005600, -0.00002841, -0.00002344, -0.00000875, 0.00003522, -0.00002397, 0.00001327, -0.00001187, 0.00001827, -0.00001863, -0.00001623, },
		{ -0.00001531, 0.00002453, 0.00001746, -0.00001620, -0.00001589, -0.00000591, -0.00001444, -0.00006880, 0.00000934, 0.00012834, 0.00015404, -0.00003392, -0.00001562, -0.00007204, 0.00004101, -0.00005322, 0.00003958, -0.00006326, -0.00001271, 0.00000146, -0.00001135, 0.00000377, -0.00003645, 0.00004220, 0.00000922, -0.00002676, -0.00001150, 0.00001855, 0.00001380, -0.00001016, -0.00003267, 0.00003448, 0.00000199, -0.00001044, -0.00004143, 0.00002888, 0.00001258, -0.00002672, -0.00007686, 0.00005082, -0.00003428, 0.00001506, -0.00009264, 0.00002636, -0.00035240, 0.00047455, 0.00045122, -0.00031063, -0.00001017, -0.00005369, -0.00000198, -0.00002223, -0.00001519, 0.00001833, 0.00003418, -0.00004357, -0.00001442, 0.00001226, 0.00001478, -0.00001697, 0.00000992, -0.00000439, 0.00001144, -0.00002541, },
		{ -0.00002124, -0.00000783, 0.00004235, -0.00001263, -0.00000481, -0.00002821, 0.00001159, -0.00005522, -0.00004600, 0.00004700, 0.00020532, 0.00005634, -0.00003519, -0.00005396, -0.00001569, 0.00000287, -0.00002458, 0.00002342, -0.00008404, 0.00002855, -0.00002620, 0.00001561, -0.00003203, 0.00000001, 0.00003858, -0.00000845, -0.00003178, 0.00001041, 0.00001304, 0.00001571, -0.00003893, 0.00000394, 0.00002450, 0.00000068, -0.00003538, -0.00000837, 0.00002307, 0.00001242, -0.00008006, -0.00000469, 0.00001234, -0.00000558, -0.00004737, -0.00001297, -0.00014524, -0.00012783, 0.00070506, -0.00000873, -0.00019242, -0.00000539, -0.00003606, -0.00001018, -0.00001539, -0.00001559, 0.00005085, -0.00001328, -0.00002969, -0.00001112, 0.00003013, -0.00001350, 0.00000448, -0.00000526, 0.00001346, -0.00001034, },
		{ -0.00002242, -0.00002010, 0.00001737, 0.00002830, -0.00002025, -0.00000845, -0.00002076, 0.00000080, -0.00007512, -0.00002359, 0.00013663, 0.00019782, -0.00002866, -0.00003489, -0.00006091, 0.00002184, -0.00003959, 0.00002850, -0.00004335, -0.00004047, 0.00002078, -0.00001967, 0.00000887, -0.00004028, 0.00003164, 0.00002228, -0.00002668, -0.00001740, 0.00001803, 0.00001507, -0.00000356, -0.00003919, 0.00003029, 0.00001143, -0.00001192, -0.00003651, 0.00001333, 0.00002214, -0.00002130, -0.00008016, 0.00003999, -0.00002150, 0.00000431, -0.00007229, 0.00000999, -0.00027928, 0.00030280, 0.00050378, -0.00023007, -0.00006158, -0.00002719, -0.00001964, -0.00001295, -0.00001992, 0.00001331, 0.00003982, -0.00003663, -0.00001958, 0.00000609, 0.00001962, -0.00001635, 0.00000807, -0.00000345, 0.00001327, },
		{ -0.00000338, -0.00002391, -0.00001361, 0.00004067, -0.00000270, -0.00001036, -0.00002190, -0.00000377, -0.00002961, -0.00007901, 0.00002881, 0.00023055, 0.00010303, -0.00006361, -0.00004560, -0.00002788, 0.00000488, -0.00002725, 0.00003104, -0.00008482, 0.00001334, -0.00000712, 0.00000215, -0.00002001, -0.00001636, 0.00004214, 0.00000023, -0.00003489, 0.00000473, 0.00001545, 0.00001698, -0.00003218, -0.00001045, 0.00003322, 0.00000105, -0.00002924, -0.00001592, 0.00001796, 0.00002062, -0.00007273, -0.00002171, 0.00002226, -0.00001356, -0.00003197, -0.00002590, -0.00010351, -0.00015045, 0.00056823, 0.00012930, -0.00020933, -0.00001092, -0.00003175, -0.00001539, -0.00001186, -0.00001981, 0.00004577, -0.00000008, -0.00003245, -0.00001394, 0.00002391, -0.00000398, -0.00000189, -0.00000149, 0.00001052, },
		{ 0.00001426, -0.00001803, -0.00002423, 0.00000916, 0.00003713, -0.00002105, -0.00000492, -0.00002754, 0.00000293, -0.00006340, -0.00006756, 0.00013518, 0.00024843, -0.00001479, -0.00006089, -0.00004749, 0.00000225, -0.00002711, 0.00001753, -0.00002330, -0.00006338, 0.00003002, -0.00001761, 0.00000653, -0.00003747, 0.00001710, 0.00003343, -0.00002322, -0.00002369, 0.00001660, 0.00001608, 0.00000239, -0.00004162, 0.00002085, 0.00002312, -0.00001420, -0.00003010, 0.00000023, 0.00002503, -0.00001150, -0.00008283, 0.00002848, -0.00001032, -0.00000454, -0.00005415, -0.00000537, -0.00021530, 0.00017706, 0.00049573, -0.00012658, -0.00010830, -0.00000862, -0.00002967, -0.00000919, -0.00002066, 0.00000642, 0.00004358, -0.00002794, -0.00002398, 0.00000082, 0.00002081, -0.00001245, 0.00000423, -0.00000185, },
		{ 0.00000930, 0.00000228, -0.00002431, -0.00001948, 0.00003624, 0.00000910, -0.00001732, -0.00001314, -0.00001942, -0.00000925, -0.00009446, -0.00001229, 0.00025748, 0.00015792, -0.00008767, -0.00004580, -0.00003259, 0.00000027, -0.00002508, 0.00003274, -0.00007680, -0.00000857, 0.00001238, -0.00000742, -0.00001221, -0.00002648, 0.00003838, 0.00001168, -0.00003676, -0.00000144, 0.00001698, 0.00001780, -0.00002428, -0.00002272, 0.00003588, 0.00000711, -0.00002806, -0.00001608, 0.00000717, 0.00002882, -0.00006388, -0.00003592, 0.00002669, -0.00001628, -0.00002299, -0.00003039, -0.00007888, -0.00014605, 0.00043418, 0.00022240, -0.00018671, -0.00003492, -0.00001833, -0.00002436, -0.00000672, -0.00002392, 0.00003963, 0.00001129, -0.00003212, -0.00001711, 0.00001772, 0.00000303, -0.00000479, -0.00000090, },
		{ -0.00000145, 0.00001551, -0.00001320, -0.00002700, 0.00000026, 0.00004301, -0.00001774, -0.00000630, -0.00002601, -0.00000569, -0.00004086, -0.00010931, 0.00011198, 0.00030954, 0.00000698, -0.00009066, -0.00003548, -0.00001448, -0.00001840, 0.00000905, -0.00000633, -0.00007812, 0.00002806, -0.00000708, 0.00000008, -0.00003170, 0.00000264, 0.00003980, -0.00001571, -0.00003018, 0.00001442, 0.00001669, 0.00000742, -0.00004029, 0.00000833, 0.00003361, -0.00001376, -0.00002627, -0.00000640, 0.00001944, 0.00000245, -0.00008504, 0.00001777, -0.00000254, -0.00000988, -0.00003997, -0.00001684, -0.00016429, 0.00009301, 0.00044695, -0.00002391, -0.00013529, -0.00000540, -0.00002877, -0.00001207, -0.00001715, -0.00000214, 0.00004569, -0.00001855, -0.00002691, -0.00000388, 0.00001959, -0.00000744, 0.00000072, },
		{ -0.00000253, 0.00000866, 0.00000720, -0.00002288, -0.00002493, 0.00002956, 0.00002101, -0.00002303, -0.00000625, -0.00002935, -0.00000067, -0.00008862, -0.00007120, 0.00027507, 0.00022759, -0.00010920, -0.00005241, -0.00003291, -0.00000759, -0.00002114, 0.00003128, -0.00006392, -0.00003136, 0.00002648, -0.00000952, -0.00001018, -0.00003019, 0.00002908, 0.00002328, -0.00003597, -0.00000839, 0.00001782, 0.00001808, -0.00001612, -0.00003142, 0.00003215, 0.00001744, -0.00002957, -0.00001266, -0.00000371, 0.00003190, -0.00005044, -0.00004974, 0.00002908, -0.00001804, -0.00001562, -0.00003189, -0.00006331, -0.00012947, 0.00031879, 0.00026820, -0.00013510, -0.00006533, -0.00000516, -0.00003060, -0.00000403, -0.00002540, 0.00003145, 0.00002115, -0.00002936, -0.00002026, 0.00001216, 0.00000703, -0.00000451, },
		},
		{ { 0.00000924, 0.00000568, -0.00002040, -0.00001226, 0.00002571, -0.00001080, -0.00001692, -0.00001133, -0.00001936, -0.00001444, -0.00014562, 0.00005300, 0.00038050, 0.00005830, -0.00016410, -0.00001296, -0.00002063, -0.00001161, 0.00000475, -0.00000783, -0.00002448, -0.00001050, 0.00004607, -0.00000587, -0.00001020, -0.00000961, -0.00000370, 0.00001337, -0.00007465, 0.00000686, 0.00005959, -0.00002236, -0.00002711, -0.00002236, 0.00005959, 0.00000686, -0.00007465, 0.00001337, -0.00000370, -0.00000961, -0.00001020, -0.00000587, 0.00004607, -0.00001050, -0.00002448, -0.00000783, 0.00000475, -0.00001161, -0.00002063, -0.00001296, -0.00016410, 0.00005830, 0.00038050, 0.00005300, -0.00014562, -0.00001444, -0.00001936, -0.00001133, -0.00001692, -0.00001080, 0.00002571, -0.00001226, -0.00002040, 0.00000568, },
		{ 0.00000711, 0.00001088, -0.00000367, -0.00002601, 0.00000676, 0.00001907, -0.00002469, -0.00000648, -0.00002012, -0.00001235, -0.00005108, -0.00015620, 0.00023450, 0.00035358, -0.00011285, -0.00011301, 0.00000396, -0.00002897, 0.00000212, -0.00000218, -0.00001085, -0.00003256, 0.00001990, 0.00003458, -0.00001920, -0.00000496, -0.00001449, 0.00001444, -0.00001967, -0.00006771, 0.00005002, 0.00003196, -0.00003367, -0.00002614, 0.00000035, 0.00006715, -0.00004418, -0.00004898, 0.00002572, -0.00002157, 0.00000187, -0.00002026, 0.00002119, 0.00003190, -0.00002692, -0.00001391, -0.00000570, 0.00000808, -0.00002748, -0.00000180, -0.00006892, -0.00013178, 0.00021925, 0.00029256, -0.00006787, -0.00010083, -0.00000059, -0.00002084, -0.00000996, -0.00002139, 0.00000788, 0.00001789, -0.00002528, -0.00000710, },
		{ 0.00000223, 0.00000767, 0.00001054, -0.00001756, -0.00001877, 0.00002116, 0.00000135, -0.00002660, 0.00000114, -0.00003233, -0.00000273, -0.00011446, -0.00008722, 0.00041552, 0.00018926, -0.00020221, -0.00003477, -0.00001384, -0.00001635, 0.00000137, -0.00000245, -0.00002380, -0.00002142, 0.00004162, 0.00001059, -0.00001960, -0.00000642, -0.00000709, 0.00001641, -0.00005272, -0.00003290, 0.00006961, -0.00000208, -0.00003400, -0.00002217, 0.00002943, 0.00004612, -0.00007975, -0.00000412, 0.00000880, -0.00002060, 0.00000359, -0.00001968, 0.00004440, 0.00000098, -0.00002252, -0.00001243, 0.00000434, -0.00000419, -0.00002531, -0.00000475, -0.00012055, -0.00002259, 0.00030059, 0.00015457, -0.00011664, -0.00005517, 0.00000187, -0.00002141, -0.00001204, -0.00001729, 0.00002317, -0.00000167, -0.00002360, },
		{ -0.00001039, 0.00000307, 0.00001257, 0.00000121, -0.00002587, -0.00000280, 0.00002363, -0.00001885, -0.00001550, -0.00000312, -0.00003762, -0.00000548, -0.00018536, 0.00009241, 0.00049222, -0.00004823, -0.00016875, 0.00000231, -0.00002629, -0.00000388, -0.00000184, -0.00000557, -0.00003567, 0.00000588, 0.00004239, -0.00001059, -0.00001491, -0.00000819, 0.00000577, -0.00000127, -0.00006699, 0.00001295, 0.00006207, -0.00003017, -0.00002681, -0.00001319, 0.00005455, -0.00000094, -0.00007892, 0.00003095, -0.00002233, 0.00000013, -0.00001210, 0.00000402, 0.00004068, -0.00001996, -0.00001409, -0.00001068, 0.00001146, -0.00002151, -0.00001014, -0.00003961, -0.00012179, 0.00011007, 0.00028635, 0.00002818, -0.00011395, -0.00002111, -0.00000144, -0.00002275, -0.00001329, -0.00000543, 0.00002649, -0.00002143, },
		{ -0.00002662, 0.00000024, 0.00000368, 0.00001536, -0.00001382, -0.00002324, 0.00001430, 0.00001158, -0.00003002, 0.00000036, -0.00002075, -0.00002734, -0.00004361, -0.00020913, 0.00035420, 0.00037757, -0.00022533, -0.00006798, -0.00001231, -0.00001393, -0.00000703, 0.00000376, -0.00002210, -0.00002875, 0.00003031, 0.00002755, -0.00002337, -0.00000899, -0.00000612, 0.00001344, -0.00002959, -0.00005473, 0.00005084, 0.00003414, -0.00004456, -0.00001657, 0.00000224, 0.00006234, -0.00005653, -0.00003812, 0.00002920, -0.00003625, 0.00001716, -0.00002779, 0.00003745, 0.00001079, -0.00001757, -0.00001618, 0.00000233, 0.00000115, -0.00002318, -0.00000841, -0.00007923, -0.00006456, 0.00021123, 0.00020219, -0.00005119, -0.00008950, 0.00000341, -0.00001225, -0.00002016, -0.00001388, 0.00001202, 0.00001240, },
		{ -0.00001181, -0.00001486, 0.00000022, 0.00001138, 0.00000723, -0.00002464, -0.00001103, 0.00002375, -0.00000920, -0.00002636, 0.00000969, -0.00004460, -0.00000303, -0.00013246, -0.00010594, 0.00057771, 0.00009286, -0.00024002, -0.00000176, -0.00002918, -0.00000100, -0.00001045, 0.00000497, -0.00003947, -0.00000490, 0.00004049, 0.00000565, -0.00002592, -0.00000406, 0.00000025, 0.00000778, -0.00005261, -0.00002227, 0.00006884, -0.00000321, -0.00004194, -0.00000816, 0.00002414, 0.00004074, -0.00009044, 0.00001775, -0.00000838, -0.00001517, 0.00000632, -0.00001626, 0.00004788, -0.00001561, -0.00000863, -0.00001855, 0.00001503, -0.00001732, -0.00001159, -0.00003005, -0.00009172, 0.00002725, 0.00024818, 0.00009455, -0.00008193, -0.00006045, 0.00001841, -0.00002785, -0.00001207, -0.00001234, 0.00002669, },
		{ 0.00002478, -0.00002654, -0.00000156, -0.00000135, 0.00001881, -0.00000871, -0.00002634, 0.00000704, 0.00001760, -0.00002539, -0.00001262, 0.00000468, -0.00005974, 0.00001195, -0.00024463, 0.00016823, 0.00059629, -0.00020322, -0.00012122, -0.00001234, -0.00001140, -0.00001222, 0.00000413, -0.00001570, -0.00003503, 0.00001703, 0.00003904, -0.00001856, -0.00001669, -0.00000312, 0.00000885, -0.00001232, -0.00005661, 0.00001565, 0.00006422, -0.00003814, -0.00002447, -0.00000404, 0.00004741, -0.00001091, -0.00007623, 0.00004550, -0.00004528, 0.00002058, -0.00002392, 0.00002264, 0.00002162, -0.00001455, -0.00001449, -0.00000391, 0.00000678, -0.00001963, -0.00001220, -0.00005338, -0.00007082, 0.00012355, 0.00021342, 0.00000968, -0.00008653, -0.00002635, 0.00001482, -0.00003507, -0.00000636, -0.00000218, },
		{ 0.00001686, 0.00000290, -0.00001984, -0.00000116, 0.00000752, 0.00001328, -0.00002126, -0.00001937, 0.00002307, -0.00000234, -0.00002842, 0.00000357, -0.00001977, -0.00004524, -0.00003096, -0.00027979, 0.00053216, 0.00033381, -0.00032176, -0.00000621, -0.00004426, 0.00001441, -0.00002843, 0.00002005, -0.00004487, -0.00000974, 0.00002821, 0.00002693, -0.00003556, -0.00000205, -0.00000404, 0.00001253, -0.00003693, -0.00004014, 0.00004696, 0.00003848, -0.00005716, -0.00000287, 0.00000081, 0.00005696, -0.00007018, -0.00001772, 0.00001987, -0.00004057, 0.00002916, -0.00003455, 0.00005178, -0.00001354, 0.00000048, -0.00002683, 0.00001833, -0.00001661, -0.00000497, -0.00003439, -0.00005660, -0.00002550, 0.00019427, 0.00012913, -0.00003211, -0.00008374, 0.00001230, -0.00001038, -0.00002456, -0.00000889, },
		{ -0.00001040, 0.00002900, -0.00002054, -0.00000414, -0.00000646, 0.00002060, -0.00000349, -0.00002614, -0.00000285, 0.00002354, -0.00002050, -0.00002296, 0.00001754, -0.00006099, 0.00000513, -0.00016666, -0.00010506, 0.00076461, -0.00008057, -0.00021542, 0.00000202, -0.00002660, 0.00000113, -0.00001249, 0.00000306, -0.00004747, 0.00001092, 0.00003676, -0.00000084, -0.00003126, 0.00000365, 0.00000107, 0.00000163, -0.00004945, -0.00001393, 0.00006666, -0.00000410, -0.00004901, 0.00000757, 0.00001599, 0.00003350, -0.00009630, 0.00004127, -0.00003513, 0.00000502, -0.00000364, -0.00000204, 0.00003706, -0.00001898, -0.00000282, -0.00001777, 0.00001758, -0.00002314, -0.00000412, -0.00005170, -0.00004662, 0.00004188, 0.00020758, 0.00004390, -0.00004792, -0.00006821, 0.00003677, -0.00003842, -0.00000678, },
		{ -0.00000119, 0.00000245, 0.00001761, -0.00002283, -0.00000132, 0.00000101, 0.00001939, -0.00001697, -0.00002593, 0.00002019, 0.00000289, -0.00002368, -0.00001486, 0.00001757, -0.00009102, 0.00004495, -0.00033644, 0.00030938, 0.00064879, -0.00037320, -0.00003153, -0.00005823, 0.00002745, -0.00004094, 0.00002722, -0.00004312, -0.00001549, 0.00001434, 0.00004315, -0.00003618, -0.00000655, -0.00000428, 0.00001248, -0.00002286, -0.00004216, 0.00001232, 0.00006940, -0.00004934, -0.00001701, 0.00000185, 0.00004028, -0.00002425, -0.00006233, 0.00004835, -0.00006164, 0.00004266, -0.00004021, 0.00004531, -0.00000685, 0.00000440, -0.00002686, 0.00001430, -0.00001212, -0.00000104, -0.00003486, -0.00003824, -0.00003869, 0.00012429, 0.00015163, -0.00000001, -0.00006622, -0.00002675, 0.00002488, -0.00004281, },
		{ -0.00002143, -0.00000846, 0.00002290, -0.00000901, -0.00000699, -0.00001095, 0.00002033, 0.00000102, -0.00002158, -0.00001661, 0.00003235, -0.00002152, -0.00001969, 0.00000277, -0.00001862, -0.00006161, -0.00002117, -0.00035798, 0.00077041, 0.00019203, -0.00035835, 0.00003814, -0.00006890, 0.00003840, -0.00004770, 0.00003418, -0.00006746, 0.00001535, 0.00001906, 0.00002800, -0.00005040, 0.00001351, -0.00001016, 0.00001490, -0.00004300, -0.00002500, 0.00003927, 0.00004491, -0.00007147, 0.00001520, -0.00000493, 0.00005115, -0.00008252, 0.00000984, -0.00000397, -0.00002735, 0.00002764, -0.00003095, 0.00005402, -0.00002890, 0.00001524, -0.00003510, 0.00003119, -0.00003310, 0.00001432, -0.00006561, -0.00001159, -0.00001910, 0.00018457, 0.00006211, -0.00000626, -0.00008580, 0.00002703, -0.00001610, },
		{ -0.00004767, -0.00000102, -0.00000785, 0.00002627, -0.00002185, 0.00000017, -0.00000878, 0.00002691, -0.00001516, -0.00002573, 0.00000921, 0.00001448, -0.00002453, -0.00002411, 0.00003182, -0.00008969, 0.00003356, -0.00023660, -0.00005062, 0.00093583, -0.00031160, -0.00012097, -0.00003923, 0.00000824, -0.00002294, 0.00000711, -0.00001959, -0.00003524, 0.00001339, 0.00004050, -0.00001786, -0.00002461, 0.00000598, 0.00000229, -0.00000522, -0.00004112, -0.00001153, 0.00006721, -0.00000837, -0.00005159, 0.00002146, 0.00000737, 0.00002233, -0.00009167, 0.00005708, -0.00006162, 0.00003365, -0.00002448, 0.00002248, 0.00001132, -0.00000527, -0.00001121, -0.00000351, 0.00000422, -0.00001152, -0.00001617, -0.00004986, -0.00001212, 0.00004175, 0.00017346, 0.00000342, -0.00001740, -0.00007483, 0.00005216, },
		{ 0.00002538, -0.00004187, -0.00000177, 0.00001081, 0.00000420, -0.00000915, -0.00001299, 0.00001561, 0.00000726, -0.00001646, -0.00002898, 0.00003841, -0.00002272, -0.00000968, -0.00002635, 0.00003810, -0.00013220, 0.00009346, -0.00046406, 0.00053721, 0.00060154, -0.00050627, 0.00006713, -0.00011680, 0.00007748, -0.00008037, 0.00005951, -0.00008105, 0.00001903, -0.00000146, 0.00005396, -0.00006199, 0.00001712, -0.00001681, 0.00002183, -0.00003550, -0.00002326, 0.00000211, 0.00007923, -0.00006535, -0.00000258, 0.00000209, 0.00003443, -0.00004003, -0.00003648, 0.00003373, -0.00006053, 0.00005367, -0.00005004, 0.00006139, -0.00003329, 0.00002685, -0.00004311, 0.00003601, -0.00003720, 0.00002680, -0.00007147, 0.00000468, -0.00003799, 0.00013181, 0.00008931, 0.00000440, -0.00005715, -0.00001909, },
		{ 0.00004386, -0.00002519, -0.00001432, -0.00000983, 0.00002652, -0.00001734, 0.00000405, -0.00002127, 0.00003544, -0.00001701, -0.00001677, -0.00001219, 0.00003621, -0.00003832, -0.00001084, 0.00000483, -0.00002615, -0.00006322, -0.00003481, -0.00041333, 0.00104267, -0.00005123, -0.00030559, 0.00003372, -0.00006152, 0.00003747, -0.00004701, 0.00002979, -0.00007351, 0.00003153, 0.00001436, 0.00002048, -0.00005603, 0.00002716, -0.00001839, 0.00001780, -0.00004544, -0.00001331, 0.00003198, 0.00004987, -0.00008404, 0.00003443, -0.00001292, 0.00004258, -0.00008750, 0.00003605, -0.00003539, 0.00000149, 0.00000952, -0.00001252, 0.00003841, -0.00002653, 0.00001638, -0.00003072, 0.00002959, -0.00003489, 0.00001873, -0.00008152, 0.00003223, -0.00003109, 0.00018147, 0.00000026, 0.00002437, -0.00009304, },
		{ -0.00007478, 0.00005883, -0.00004924, -0.00000071, -0.00000031, 0.00001495, -0.00001104, -0.00000911, 0.00000244, 0.00002000, -0.00001810, -0.00002999, 0.00002983, -0.00000964, -0.00001279, -0.00003843, 0.00006017, -0.00013829, 0.00009211, -0.00035994, 0.00009540, 0.00103300, -0.00055092, 0.00002122, -0.00011906, 0.00007310, -0.00007272, 0.00004869, -0.00006402, 0.00000076, -0.00000317, 0.00005624, -0.00004895, 0.00000051, -0.00000612, 0.00001164, -0.00001736, -0.00002507, -0.00001826, 0.00007426, -0.00001960, -0.00004596, 0.00002953, 0.00000096, 0.00000601, -0.00007242, 0.00005573, -0.00007444, 0.00005730, -0.00004600, 0.00004843, -0.00002087, 0.00001951, -0.00003022, 0.00002115, -0.00002259, 0.00001670, -0.00004912, -0.00001686, -0.00000730, 0.00004850, 0.00013378, -0.00002007, 0.00000323, },
		{ -0.00005942, -0.00000447, 0.00001786, -0.00003319, -0.00000744, 0.00002090, -0.00001127, 0.00000913, -0.00003193, 0.00003952, -0.00001710, -0.00000580, -0.00003517, 0.00005804, -0.00005456, 0.00001309, -0.00004358, 0.00006054, -0.00017376, 0.00014308, -0.00061009, 0.00084857, 0.00042328, -0.00055133, 0.00013165, -0.00015748, 0.00011477, -0.00011058, 0.00008334, -0.00011276, 0.00005436, -0.00002016, 0.00006225, -0.00008545, 0.00004596, -0.00003706, 0.00003593, -0.00004890, -0.00000297, -0.00001196, 0.00009140, -0.00008399, 0.00001702, -0.00000252, 0.00002826, -0.00005318, -0.00000454, 0.00000400, -0.00003830, 0.00004511, -0.00004409, 0.00006073, -0.00004535, 0.00003997, -0.00005120, 0.00004852, -0.00005423, 0.00004824, -0.00010396, 0.00005770, -0.00006187, 0.00015060, 0.00002290, 0.00002342, },
		{ 0.00005206, -0.00009876, 0.00005581, -0.00003015, -0.00001018, -0.00000721, 0.00002198, -0.00001461, 0.00000324, -0.00002040, 0.00004059, -0.00002988, -0.00001460, 0.00000071, 0.00002529, -0.00004087, -0.00001384, 0.00002146, -0.00005566, -0.00003218, -0.00009978, -0.00039972, 0.00129329, -0.00036292, -0.00016393, -0.00003195, -0.00000950, -0.00000002, -0.00001559, -0.00000355, -0.00005111, 0.00002633, 0.00002417, -0.00000440, -0.00004112, 0.00002619, -0.00001861, 0.00001475, -0.00003958, -0.00001045, 0.00003077, 0.00004810, -0.00008959, 0.00004960, -0.00001933, 0.00002813, -0.00007819, 0.00004916, -0.00006049, 0.00003417, -0.00001816, 0.00001597, 0.00000841, -0.00000682, 0.00000182, -0.00001102, 0.00000975, -0.00001592, -0.00000063, -0.00006722, 0.00005143, -0.00003925, 0.00017309, -0.00004857, },
		{ -0.00002189, 0.00000972, -0.00006535, 0.00005460, -0.00004054, -0.00000833, 0.00001320, -0.00000420, 0.00001077, -0.00003233, 0.00002975, -0.00000515, -0.00000611, -0.00004273, 0.00005976, -0.00004959, 0.00001734, -0.00006870, 0.00010370, -0.00020586, 0.00017879, -0.00053882, 0.00035676, 0.00099929, -0.00073348, 0.00016569, -0.00020965, 0.00014794, -0.00013232, 0.00009869, -0.00011803, 0.00005137, -0.00003327, 0.00007883, -0.00008773, 0.00004041, -0.00003345, 0.00003167, -0.00003611, -0.00000184, -0.00003372, 0.00008804, -0.00003806, -0.00003147, 0.00003037, -0.00000307, -0.00001272, -0.00004033, 0.00003400, -0.00006415, 0.00006260, -0.00005529, 0.00006316, -0.00004588, 0.00004256, -0.00004858, 0.00004482, -0.00005082, 0.00004938, -0.00009052, 0.00003778, -0.00003390, 0.00007261, 0.00008088, },
		{ -0.00004069, 0.00005031, -0.00006745, 0.00001097, 0.00000905, -0.00002306, -0.00001262, 0.00002648, -0.00002044, 0.00002130, -0.00004629, 0.00006113, -0.00004408, 0.00000840, -0.00003787, 0.00006914, -0.00008037, 0.00003254, -0.00005426, 0.00007005, -0.00019552, 0.00016443, -0.00073298, 0.00120547, 0.00011818, -0.00047991, 0.00012795, -0.00015246, 0.00011613, -0.00011133, 0.00008014, -0.00011959, 0.00007291, -0.00002767, 0.00005545, -0.00009186, 0.00006518, -0.00005400, 0.00004762, -0.00005741, 0.00001233, -0.00002345, 0.00010005, -0.00009953, 0.00003637, -0.00000795, 0.00001773, -0.00005546, 0.00002157, -0.00002907, -0.00000247, 0.00001856, -0.00002122, 0.00004058, -0.00003718, 0.00003591, -0.00004293, 0.00004241, -0.00005141, 0.00004876, -0.00011294, 0.00009605, -0.00008992, 0.00017075, },
		{ 0.00014715, -0.00007507, 0.00006833, -0.00009659, 0.00005707, -0.00002482, -0.00001483, 0.00000381, 0.00000591, 0.00000310, -0.00001544, 0.00000023, 0.00002498, -0.00002635, -0.00002345, 0.00002805, -0.00000766, -0.00001937, -0.00003900, 0.00006225, -0.00011696, 0.00004416, -0.00023830, -0.00027068, 0.00145045, -0.00067849, 0.00003263, -0.00014472, 0.00007936, -0.00006892, 0.00004273, -0.00006295, -0.00000049, -0.00000272, 0.00005017, -0.00004689, -0.00000253, 0.00000361, -0.00000345, 0.00000052, -0.00002225, -0.00001965, 0.00003924, 0.00003615, -0.00008442, 0.00005665, -0.00002166, 0.00000739, -0.00005165, 0.00004041, -0.00006507, 0.00005436, -0.00004136, 0.00004216, -0.00002402, 0.00002032, -0.00002074, 0.00001671, -0.00002127, 0.00001792, -0.00003931, -0.00002258, 0.00003524, -0.00002688, },
		{ 0.00010889, 0.00001785, -0.00000489, 0.00000394, -0.00004960, 0.00004337, -0.00002515, -0.00002049, 0.00002813, -0.00002451, 0.00003511, -0.00006067, 0.00006545, -0.00004339, 0.00001842, -0.00006060, 0.00009206, -0.00009682, 0.00005818, -0.00010710, 0.00015206, -0.00027770, 0.00027308, -0.00074901, 0.00072512, 0.00080323, -0.00080196, 0.00026109, -0.00027347, 0.00020326, -0.00017690, 0.00013499, -0.00016059, 0.00009799, -0.00006251, 0.00009518, -0.00011931, 0.00008130, -0.00006688, 0.00005693, -0.00005701, 0.00002328, -0.00005268, 0.00010391, -0.00005928, -0.00001209, 0.00002678, -0.00000822, -0.00002628, -0.00000494, -0.00000109, -0.00003188, 0.00004429, -0.00004466, 0.00005781, -0.00005244, 0.00005152, -0.00005423, 0.00005452, -0.00006548, 0.00006962, -0.00012033, 0.00009184, -0.00007647, },
		{ -0.00010046, 0.00017650, -0.00008895, 0.00007412, -0.00007288, 0.00001925, 0.00000741, -0.00001958, -0.00001022, 0.00002069, -0.00001511, 0.00001662, -0.00004282, 0.00006542, -0.00005903, 0.00001468, -0.00002714, 0.00005931, -0.00008473, 0.00003263, -0.00004218, 0.00004828, -0.00017363, 0.00012174, -0.00077556, 0.00153829, -0.00026773, -0.00030028, 0.00004696, -0.00009093, 0.00007136, -0.00007320, 0.00004075, -0.00009081, 0.00006292, -0.00001444, 0.00002568, -0.00007059, 0.00006137, -0.00005542, 0.00004783, -0.00005424, 0.00001595, -0.00002564, 0.00009887, -0.00010558, 0.00004916, -0.00000958, -0.00000064, -0.00003965, 0.00002898, -0.00004886, 0.00003056, -0.00001394, 0.00000963, 0.00000833, -0.00001330, 0.00001664, -0.00001955, 0.00001802, -0.00002694, 0.00002399, -0.00008884, 0.00010104, },
		{ -0.00000908, 0.00000828, 0.00010027, -0.00007589, 0.00006961, -0.00008499, 0.00004720, -0.00000860, -0.00002875, 0.00002237, -0.00001909, 0.00003223, -0.00004906, 0.00003962, -0.00001344, -0.00000205, -0.00004600, 0.00006911, -0.00006150, 0.00002691, -0.00008738, 0.00012665, -0.00020748, 0.00016212, -0.00045147, -0.00000061, 0.00145009, -0.00092134, 0.00022741, -0.00026838, 0.00017864, -0.00014786, 0.00010941, -0.00013146, 0.00006449, -0.00004580, 0.00008320, -0.00009619, 0.00005092, -0.00003671, 0.00002629, -0.00002421, 0.00000465, -0.00003898, 0.00005591, 0.00001544, -0.00006949, 0.00005571, -0.00002131, -0.00001468, -0.00001280, 0.00001026, -0.00004286, 0.00004991, -0.00004674, 0.00005272, -0.00004440, 0.00004062, -0.00003816, 0.00003897, -0.00004878, 0.00005126, -0.00008087, 0.00003665, },
		{ 0.00012165, -0.00011148, 0.00014112, -0.00004280, 0.00002074, -0.00000660, -0.00003520, 0.00003361, -0.00001139, -0.00002944, 0.00003621, -0.00003586, 0.00004927, -0.00007930, 0.00009373, -0.00008028, 0.00004435, -0.00007311, 0.00011292, -0.00013461, 0.00009268, -0.00013604, 0.00018465, -0.00032710, 0.00033762, -0.00093915, 0.00115199, 0.00045635, -0.00072937, 0.00027124, -0.00028018, 0.00021381, -0.00018456, 0.00013752, -0.00017155, 0.00012127, -0.00007501, 0.00009134, -0.00012732, 0.00010540, -0.00009137, 0.00007634, -0.00007138, 0.00004174, -0.00006682, 0.00011413, -0.00007565, 0.00000494, 0.00002436, -0.00001987, -0.00002485, 0.00001929, -0.00003365, 0.00000931, 0.00000922, -0.00001698, 0.00003313, -0.00003823, 0.00004172, -0.00004184, 0.00004363, -0.00005740, 0.00006538, -0.00012175, },
		{ -0.00003799, 0.00006971, -0.00008242, 0.00015699, -0.00011182, 0.00008580, -0.00006961, 0.00001546, 0.00001810, -0.00002767, 0.00000348, 0.00000087, 0.00000739, -0.00000913, -0.00001461, 0.00004228, -0.00005017, 0.00000351, 0.00000376, 0.00002082, -0.00005704, 0.00000179, 0.00000388, -0.00001664, -0.00009283, -0.00001117, -0.00068433, 0.00176396, -0.00065665, -0.00005712, -0.00009035, 0.00001455, -0.00001034, -0.00000373, -0.00002833, -0.00003017, 0.00002504, 0.00001838, -0.00002408, -0.00002217, 0.00003043, -0.00003535, 0.00003160, -0.00003624, 0.00000507, -0.00001560, 0.00008511, -0.00009915, 0.00005201, -0.00000571, -0.00002581, -0.00000485, 0.00001101, -0.00004219, 0.00004333, -0.00003577, 0.00003357, -0.00002152, 0.00001335, -0.00000690, 0.00000830, -0.00001386, 0.00000912, -0.00001681, },
		{ -0.00010518, 0.00008629, -0.00006047, 0.00005368, 0.00004092, -0.00005678, 0.00005950, -0.00006870, 0.00003219, 0.00001255, -0.00004611, 0.00004164, -0.00004423, 0.00006179, -0.00008527, 0.00008561, -0.00006484, 0.00003657, -0.00007563, 0.00011365, -0.00012383, 0.00008616, -0.00014696, 0.00019995, -0.00030784, 0.00029587, -0.00070929, 0.00039742, 0.00126129, -0.00102934, 0.00036229, -0.00035966, 0.00025441, -0.00020831, 0.00015903, -0.00018482, 0.00012186, -0.00008524, 0.00010737, -0.00013427, 0.00010112, -0.00008075, 0.00006088, -0.00005186, 0.00003356, -0.00006113, 0.00007399, -0.00000739, -0.00005090, 0.00005151, -0.00002364, -0.00002816, 0.00002544, -0.00002957, -0.00000048, 0.00002008, -0.00002952, 0.00004113, -0.00004366, 0.00004340, -0.00003962, 0.00004383, -0.00005851, 0.00006751, },
		{ 0.00003682, -0.00009066, 0.00011419, -0.00011989, 0.00015227, -0.00008614, 0.00004254, -0.00001305, -0.00003014, 0.00003386, -0.00000765, -0.00002759, 0.00003012, -0.00003000, 0.00004277, -0.00007465, 0.00009813, -0.00009854, 0.00005717, -0.00006775, 0.00010753, -0.00014465, 0.00010129, -0.00013534, 0.00017859, -0.00032499, 0.00033013, -0.00104317, 0.00155633, 0.00001582, -0.00053100, 0.00019009, -0.00022054, 0.00017080, -0.00014727, 0.00009866, -0.00014169, 0.00011060, -0.00006215, 0.00006081, -0.00010274, 0.00009990, -0.00009393, 0.00007926, -0.00007099, 0.00004610, -0.00006898, 0.00011201, -0.00008042, 0.00001297, 0.00002788, -0.00004112, -0.00000185, 0.00001965, -0.00004586, 0.00003968, -0.00002623, 0.00001496, 0.00000017, -0.00001165, 0.00001900, -0.00001643, 0.00001647, -0.00002899, },
		{ 0.00003972, -0.00005497, 0.00001982, 0.00001620, -0.00004071, 0.00011317, -0.00010788, 0.00008299, -0.00005765, 0.00000115, 0.00003953, -0.00004566, 0.00002568, -0.00002854, 0.00004123, -0.00004983, 0.00003367, -0.00000681, -0.00001566, -0.00002614, 0.00005241, -0.00004308, 0.00000132, -0.00005920, 0.00008174, -0.00012064, 0.00004183, -0.00023452, -0.00043271, 0.00181334, -0.00096302, 0.00018392, -0.00024113, 0.00013269, -0.00010354, 0.00007486, -0.00010654, 0.00004475, -0.00002738, 0.00005855, -0.00007956, 0.00004012, -0.00001922, 0.00000181, 0.00000195, -0.00000690, -0.00001675, 0.00000334, 0.00006211, -0.00008307, 0.00004684, 0.00000035, -0.00005039, 0.00004061, -0.00002756, -0.00000730, 0.00002659, -0.00003469, 0.00003786, -0.00003493, 0.00002840, -0.00002101, 0.00002635, -0.00003737, },
		{ -0.00004470, 0.00005800, -0.00009918, 0.00010576, -0.00009511, 0.00008969, -0.00001488, -0.00003003, 0.00004652, -0.00005617, 0.00002152, 0.00002906, -0.00005770, 0.00005199, -0.00005832, 0.00007808, -0.00010745, 0.00011945, -0.00011083, 0.00007437, -0.00009850, 0.00014463, -0.00017417, 0.00013704, -0.00019554, 0.00025656, -0.00038589, 0.00040183, -0.00095285, 0.00086652, 0.00090207, -0.00097737, 0.00039971, -0.00038647, 0.00028020, -0.00022754, 0.00017097, -0.00020240, 0.00015149, -0.00010431, 0.00010832, -0.00014439, 0.00012883, -0.00011087, 0.00008628, -0.00007128, 0.00005435, -0.00007650, 0.00008451, -0.00002346, -0.00003709, 0.00005072, -0.00003482, -0.00002244, 0.00004705, -0.00006015, 0.00004401, -0.00002290, 0.00000307, 0.00001206, -0.00002341, 0.00002769, -0.00002319, 0.00002811, },
		{ -0.00001344, 0.00000653, -0.00000317, -0.00003855, 0.00007350, -0.00009666, 0.00013426, -0.00010289, 0.00005203, -0.00001048, -0.00003776, 0.00004767, -0.00001741, -0.00001241, 0.00000832, -0.00000577, 0.00001356, -0.00004215, 0.00007028, -0.00008691, 0.00004701, -0.00003757, 0.00006823, -0.00011612, 0.00007176, -0.00009265, 0.00012044, -0.00025407, 0.00022156, -0.00100285, 0.00184717, -0.00043059, -0.00025981, 0.00004384, -0.00011059, 0.00008639, -0.00007542, 0.00002772, -0.00007772, 0.00006936, -0.00002749, 0.00000939, -0.00004970, 0.00006403, -0.00007073, 0.00006173, -0.00005321, 0.00003431, -0.00005736, 0.00009604, -0.00007179, 0.00000985, 0.00003765, -0.00006905, 0.00004099, -0.00000734, -0.00002731, 0.00004247, -0.00004433, 0.00003488, -0.00002515, 0.00001247, -0.00000349, 0.00000904, },
		{ 0.00002492, -0.00004019, 0.00004993, -0.00007201, 0.00006088, -0.00003567, 0.00000720, 0.00005756, -0.00008523, 0.00007068, -0.00004308, -0.00001602, 0.00006386, -0.00006577, 0.00004772, -0.00005716, 0.00007375, -0.00009062, 0.00008625, -0.00006783, 0.00003419, -0.00006464, 0.00010567, -0.00011665, 0.00007488, -0.00013464, 0.00017297, -0.00024064, 0.00020062, -0.00051401, -0.00003781, 0.00165692, -0.00112191, 0.00036002, -0.00035821, 0.00022680, -0.00017758, 0.00013531, -0.00016787, 0.00010998, -0.00007469, 0.00008862, -0.00012105, 0.00009540, -0.00006990, 0.00004274, -0.00003073, 0.00002430, -0.00004054, 0.00002270, 0.00003827, -0.00006520, 0.00004012, 0.00000162, -0.00006258, 0.00008107, -0.00007080, 0.00004345, -0.00001494, -0.00000987, 0.00001925, -0.00002595, 0.00002389, -0.00001722, },
		{ 0.00000271, -0.00000020, -0.00001367, 0.00002711, -0.00006405, 0.00008819, -0.00009826, 0.00010029, -0.00005181, -0.00000823, 0.00003896, -0.00005452, 0.00002297, 0.00003304, -0.00005621, 0.00004631, -0.00005370, 0.00007165, -0.00010291, 0.00012464, -0.00013280, 0.00009484, -0.00010036, 0.00014572, -0.00019227, 0.00015749, -0.00020999, 0.00027024, -0.00040880, 0.00043329, -0.00111037, 0.00131790, 0.00043783, -0.00078619, 0.00033604, -0.00034079, 0.00024844, -0.00019895, 0.00013916, -0.00017670, 0.00014425, -0.00009565, 0.00008126, -0.00011982, 0.00012300, -0.00011453, 0.00009133, -0.00007360, 0.00005939, -0.00007819, 0.00008112, -0.00002638, -0.00003433, 0.00005759, -0.00005703, 0.00000768, 0.00004085, -0.00006425, 0.00006919, -0.00005935, 0.00003503, -0.00002046, 0.00000465, 0.00000273, },
		{ -0.00001277, 0.00002086, -0.00003041, 0.00003186, -0.00003539, 0.00001290, 0.00001834, -0.00005246, 0.00009213, -0.00009395, 0.00004841, -0.00000055, -0.00005453, 0.00007140, -0.00003699, 0.00001149, -0.00002317, 0.00002923, -0.00002963, 0.00001008, 0.00001526, -0.00004642, 0.00001476, 0.00001324, 0.00000038, -0.00005257, 0.00000670, -0.00001172, 0.00001553, -0.00012044, 0.00001262, -0.00079187, 0.00195312, -0.00079187, 0.00001262, -0.00012044, 0.00001553, -0.00001172, 0.00000670, -0.00005257, 0.00000038, 0.00001324, 0.00001476, -0.00004642, 0.00001526, 0.00001008, -0.00002963, 0.00002923, -0.00002317, 0.00001149, -0.00003699, 0.00007140, -0.00005453, -0.00000055, 0.00004841, -0.00009395, 0.00009213, -0.00005246, 0.00001834, 0.00001290, -0.00003539, 0.00003186, -0.00003041, 0.00002086, },
		{ 0.00000271, 0.00000273, 0.00000465, -0.00002046, 0.00003503, -0.00005935, 0.00006919, -0.00006425, 0.00004085, 0.00000768, -0.00005703, 0.00005759, -0.00003433, -0.00002638, 0.00008112, -0.00007819, 0.00005939, -0.00007360, 0.00009133, -0.00011453, 0.00012300, -0.00011982, 0.00008126, -0.00009565, 0.00014425, -0.00017670, 0.00013916, -0.00019895, 0.00024844, -0.00034079, 0.00033604, -0.00078619, 0.00043783, 0.00131790, -0.00111037, 0.00043329, -0.00040880, 0.00027024, -0.00020999, 0.00015749, -0.00019227, 0.00014572, -0.00010036, 0.00009484, -0.00013280, 0.00012464, -0.00010291, 0.00007165, -0.00005370, 0.00004631, -0.00005621, 0.00003304, 0.00002297, -0.00005452, 0.00003896, -0.00000823, -0.00005181, 0.00010029, -0.00009826, 0.00008819, -0.00006405, 0.00002711, -0.00001367, -0.00000020, },
		{ 0.00002492, -0.00001722, 0.00002389, -0.00002595, 0.00001925, -0.00000987, -0.00001494, 0.00004345, -0.00007080, 0.00008107, -0.00006258, 0.00000162, 0.00004012, -0.00006520, 0.00003827, 0.00002270, -0.00004054, 0.00002430, -0.00003073, 0.00004274, -0.00006990, 0.00009540, -0.00012105, 0.00008862, -0.00007469, 0.00010998, -0.00016787, 0.00013531, -0.00017758, 0.00022680, -0.00035821, 0.00036002, -0.00112191, 0.00165692, -0.00003781, -0.00051401, 0.00020062, -0.00024064, 0.00017297, -0.00013464, 0.00007488, -0.00011665, 0.00010567, -0.00006464, 0.00003419, -0.00006783, 0.00008625, -0.00009062, 0.00007375, -0.00005716, 0.00004772, -0.00006577, 0.00006386, -0.00001602, -0.00004308, 0.00007068, -0.00008523, 0.00005756, 0.00000720, -0.00003567, 0.00006088, -0.00007201, 0.00004993, -0.00004019, },
		{ -0.00001344, 0.00000904, -0.00000349, 0.00001247, -0.00002515, 0.00003488, -0.00004433, 0.00004247, -0.00002731, -0.00000734, 0.00004099, -0.00006905, 0.00003765, 0.00000985, -0.00007179, 0.00009604, -0.00005736, 0.00003431, -0.00005321, 0.00006173, -0.00007073, 0.00006403, -0.00004970, 0.00000939, -0.00002749, 0.00006936, -0.00007772, 0.00002772, -0.00007542, 0.00008639, -0.00011059, 0.00004384, -0.00025981, -0.00043059, 0.00184717, -0.00100285, 0.00022156, -0.00025407, 0.00012044, -0.00009265, 0.00007176, -0.00011612, 0.00006823, -0.00003757, 0.00004701, -0.00008691, 0.00007028, -0.00004215, 0.00001356, -0.00000577, 0.00000832, -0.00001241, -0.00001741, 0.00004767, -0.00003776, -0.00001048, 0.00005203, -0.00010289, 0.00013426, -0.00009666, 0.00007350, -0.00003855, -0.00000317, 0.00000653, },
		{ -0.00004470, 0.00002811, -0.00002319, 0.00002769, -0.00002341, 0.00001206, 0.00000307, -0.00002290, 0.00004401, -0.00006015, 0.00004705, -0.00002244, -0.00003482, 0.00005072, -0.00003709, -0.00002346, 0.00008451, -0.00007650, 0.00005435, -0.00007128, 0.00008628, -0.00011087, 0.00012883, -0.00014439, 0.00010832, -0.00010431, 0.00015149, -0.00020240, 0.00017097, -0.00022754, 0.00028020, -0.00038647, 0.00039971, -0.00097737, 0.00090207, 0.00086652, -0.00095285, 0.00040183, -0.00038589, 0.00025656, -0.00019554, 0.00013704, -0.00017417, 0.00014463, -0.00009850, 0.00007437, -0.00011083, 0.00011945, -0.00010745, 0.00007808, -0.00005832, 0.00005199, -0.00005770, 0.00002906, 0.00002152, -0.00005617, 0.00004652, -0.00003003, -0.00001488, 0.00008969, -0.00009511, 0.00010576, -0.00009918, 0.00005800, },
		{ 0.00003972, -0.00003737, 0.00002635, -0.00002101, 0.00002840, -0.00003493, 0.00003786, -0.00003469, 0.00002659, -0.00000730, -0.00002756, 0.00004061, -0.00005039, 0.00000035, 0.00004684, -0.00008307, 0.00006211, 0.00000334, -0.00001675, -0.00000690, 0.00000195, 0.00000181, -0.00001922, 0.00004012, -0.00007956, 0.00005855, -0.00002738, 0.00004475, -0.00010654, 0.00007486, -0.00010354, 0.00013269, -0.00024113, 0.00018392, -0.00096302, 0.00181334, -0.00043271, -0.00023452, 0.00004183, -0.00012064, 0.00008174, -0.00005920, 0.00000132, -0.00004308, 0.00005241, -0.00002614, -0.00001566, -0.00000681, 0.00003367, -0.00004983, 0.00004123, -0.00002854, 0.00002568, -0.00004566, 0.00003953, 0.00000115, -0.00005765, 0.00008299, -0.00010788, 0.00011317, -0.00004071, 0.00001620, 0.00001982, -0.00005497, },
		{ 0.00003682, -0.00002899, 0.00001647, -0.00001643, 0.00001900, -0.00001165, 0.00000017, 0.00001496, -0.00002623, 0.00003968, -0.00004586, 0.00001965, -0.00000185, -0.00004112, 0.00002788, 0.00001297, -0.00008042, 0.00011201, -0.00006898, 0.00004610, -0.00007099, 0.00007926, -0.00009393, 0.00009990, -0.00010274, 0.00006081, -0.00006215, 0.00011060, -0.00014169, 0.00009866, -0.00014727, 0.00017080, -0.00022054, 0.00019009, -0.00053100, 0.00001582, 0.00155633, -0.00104317, 0.00033013, -0.00032499, 0.00017859, -0.00013534, 0.00010129, -0.00014465, 0.00010753, -0.00006775, 0.00005717, -0.00009854, 0.00009813, -0.00007465, 0.00004277, -0.00003000, 0.00003012, -0.00002759, -0.00000765, 0.00003386, -0.00003014, -0.00001305, 0.00004254, -0.00008614, 0.00015227, -0.00011989, 0.00011419, -0.00009066, },
		{ -0.00010518, 0.00006751, -0.00005851, 0.00004383, -0.00003962, 0.00004340, -0.00004366, 0.00004113, -0.00002952, 0.00002008, -0.00000048, -0.00002957, 0.00002544, -0.00002816, -0.00002364, 0.00005151, -0.00005090, -0.00000739, 0.00007399, -0.00006113, 0.00003356, -0.00005186, 0.00006088, -0.00008075, 0.00010112, -0.00013427, 0.00010737, -0.00008524, 0.00012186, -0.00018482, 0.00015903, -0.00020831, 0.00025441, -0.00035966, 0.00036229, -0.00102934, 0.00126129, 0.00039742, -0.00070929, 0.00029587, -0.00030784, 0.00019995, -0.00014696, 0.00008616, -0.00012383, 0.00011365, -0.00007563, 0.00003657, -0.00006484, 0.00008561, -0.00008527, 0.00006179, -0.00004423, 0.00004164, -0.00004611, 0.00001255, 0.00003219, -0.00006870, 0.00005950, -0.00005678, 0.00004092, 0.00005368, -0.00006047, 0.00008629, },
		{ -0.00003799, -0.00001681, 0.00000912, -0.00001386, 0.00000830, -0.00000690, 0.00001335, -0.00002152, 0.00003357, -0.00003577, 0.00004333, -0.00004219, 0.00001101, -0.00000485, -0.00002581, -0.00000571, 0.00005201, -0.00009915, 0.00008511, -0.00001560, 0.00000507, -0.00003624, 0.00003160, -0.00003535, 0.00003043, -0.00002217, -0.00002408, 0.00001838, 0.00002504, -0.00003017, -0.00002833, -0.00000373, -0.00001034, 0.00001455, -0.00009035, -0.00005712, -0.00065665, 0.00176396, -0.00068433, -0.00001117, -0.00009283, -0.00001664, 0.00000388, 0.00000179, -0.00005704, 0.00002082, 0.00000376, 0.00000351, -0.00005017, 0.00004228, -0.00001461, -0.00000913, 0.00000739, 0.00000087, 0.00000348, -0.00002767, 0.00001810, 0.00001546, -0.00006961, 0.00008580, -0.00011182, 0.00015699, -0.00008242, 0.00006971, },
		{ 0.00012165, -0.00012175, 0.00006538, -0.00005740, 0.00004363, -0.00004184, 0.00004172, -0.00003823, 0.00003313, -0.00001698, 0.00000922, 0.00000931, -0.00003365, 0.00001929, -0.00002485, -0.00001987, 0.00002436, 0.00000494, -0.00007565, 0.00011413, -0.00006682, 0.00004174, -0.00007138, 0.00007634, -0.00009137, 0.00010540, -0.00012732, 0.00009134, -0.00007501, 0.00012127, -0.00017155, 0.00013752, -0.00018456, 0.00021381, -0.00028018, 0.00027124, -0.00072937, 0.00045635, 0.00115199, -0.00093915, 0.00033762, -0.00032710, 0.00018465, -0.00013604, 0.00009268, -0.00013461, 0.00011292, -0.00007311, 0.00004435, -0.00008028, 0.00009373, -0.00007930, 0.00004927, -0.00003586, 0.00003621, -0.00002944, -0.00001139, 0.00003361, -0.00003520, -0.00000660, 0.00002074, -0.00004280, 0.00014112, -0.00011148, },
		{ -0.00000908, 0.00003665, -0.00008087, 0.00005126, -0.00004878, 0.00003897, -0.00003816, 0.00004062, -0.00004440, 0.00005272, -0.00004674, 0.00004991, -0.00004286, 0.00001026, -0.00001280, -0.00001468, -0.00002131, 0.00005571, -0.00006949, 0.00001544, 0.00005591, -0.00003898, 0.00000465, -0.00002421, 0.00002629, -0.00003671, 0.00005092, -0.00009619, 0.00008320, -0.00004580, 0.00006449, -0.00013146, 0.00010941, -0.00014786, 0.00017864, -0.00026838, 0.00022741, -0.00092134, 0.00145009, -0.00000061, -0.00045147, 0.00016212, -0.00020748, 0.00012665, -0.00008738, 0.00002691, -0.00006150, 0.00006911, -0.00004600, -0.00000205, -0.00001344, 0.00003962, -0.00004906, 0.00003223, -0.00001909, 0.00002237, -0.00002875, -0.00000860, 0.00004720, -0.00008499, 0.00006961, -0.00007589, 0.00010027, 0.00000828, },
		{ -0.00010046, 0.00010104, -0.00008884, 0.00002399, -0.00002694, 0.00001802, -0.00001955, 0.00001664, -0.00001330, 0.00000833, 0.00000963, -0.00001394, 0.00003056, -0.00004886, 0.00002898, -0.00003965, -0.00000064, -0.00000958, 0.00004916, -0.00010558, 0.00009887, -0.00002564, 0.00001595, -0.00005424, 0.00004783, -0.00005542, 0.00006137, -0.00007059, 0.00002568, -0.00001444, 0.00006292, -0.00009081, 0.00004075, -0.00007320, 0.00007136, -0.00009093, 0.00004696, -0.00030028, -0.00026773, 0.00153829, -0.00077556, 0.00012174, -0.00017363, 0.00004828, -0.00004218, 0.00003263, -0.00008473, 0.00005931, -0.00002714, 0.00001468, -0.00005903, 0.00006542, -0.00004282, 0.00001662, -0.00001511, 0.00002069, -0.00001022, -0.00001958, 0.00000741, 0.00001925, -0.00007288, 0.00007412, -0.00008895, 0.00017650, },
		{ 0.00010889, -0.00007647, 0.00009184, -0.00012033, 0.00006962, -0.00006548, 0.00005452, -0.00005423, 0.00005152, -0.00005244, 0.00005781, -0.00004466, 0.00004429, -0.00003188, -0.00000109, -0.00000494, -0.00002628, -0.00000822, 0.00002678, -0.00001209, -0.00005928, 0.00010391, -0.00005268, 0.00002328, -0.00005701, 0.00005693, -0.00006688, 0.00008130, -0.00011931, 0.00009518, -0.00006251, 0.00009799, -0.00016059, 0.00013499, -0.00017690, 0.00020326, -0.00027347, 0.00026109, -0.00080196, 0.00080323, 0.00072512, -0.00074901, 0.00027308, -0.00027770, 0.00015206, -0.00010710, 0.00005818, -0.00009682, 0.00009206, -0.00006060, 0.00001842, -0.00004339, 0.00006545, -0.00006067, 0.00003511, -0.00002451, 0.00002813, -0.00002049, -0.00002515, 0.00004337, -0.00004960, 0.00000394, -0.00000489, 0.00001785, },
		{ 0.00014715, -0.00002688, 0.00003524, -0.00002258, -0.00003931, 0.00001792, -0.00002127, 0.00001671, -0.00002074, 0.00002032, -0.00002402, 0.00004216, -0.00004136, 0.00005436, -0.00006507, 0.00004041, -0.00005165, 0.00000739, -0.00002166, 0.00005665, -0.00008442, 0.00003615, 0.00003924, -0.00001965, -0.00002225, 0.00000052, -0.00000345, 0.00000361, -0.00000253, -0.00004689, 0.00005017, -0.00000272, -0.00000049, -0.00006295, 0.00004273, -0.00006892, 0.00007936, -0.00014472, 0.00003263, -0.00067849, 0.00145045, -0.00027068, -0.00023830, 0.00004416, -0.00011696, 0.00006225, -0.00003900, -0.00001937, -0.00000766, 0.00002805, -0.00002345, -0.00002635, 0.00002498, 0.00000023, -0.00001544, 0.00000310, 0.00000591, 0.00000381, -0.00001483, -0.00002482, 0.00005707, -0.00009659, 0.00006833, -0.00007507, },
		{ -0.00004069, 0.00017075, -0.00008992, 0.00009605, -0.00011294, 0.00004876, -0.00005141, 0.00004241, -0.00004293, 0.00003591, -0.00003718, 0.00004058, -0.00002122, 0.00001856, -0.00000247, -0.00002907, 0.00002157, -0.00005546, 0.00001773, -0.00000795, 0.00003637, -0.00009953, 0.00010005, -0.00002345, 0.00001233, -0.00005741, 0.00004762, -0.00005400, 0.00006518, -0.00009186, 0.00005545, -0.00002767, 0.00007291, -0.00011959, 0.00008014, -0.00011133, 0.00011613, -0.00015246, 0.00012795, -0.00047991, 0.00011818, 0.00120547, -0.00073298, 0.00016443, -0.00019552, 0.00007005, -0.00005426, 0.00003254, -0.00008037, 0.00006914, -0.00003787, 0.00000840, -0.00004408, 0.00006113, -0.00004629, 0.00002130, -0.00002044, 0.00002648, -0.00001262, -0.00002306, 0.00000905, 0.00001097, -0.00006745, 0.00005031, },
		{ -0.00002189, 0.00008088, 0.00007261, -0.00003390, 0.00003778, -0.00009052, 0.00004938, -0.00005082, 0.00004482, -0.00004858, 0.00004256, -0.00004588, 0.00006316, -0.00005529, 0.00006260, -0.00006415, 0.00003400, -0.00004033, -0.00001272, -0.00000307, 0.00003037, -0.00003147, -0.00003806, 0.00008804, -0.00003372, -0.00000184, -0.00003611, 0.00003167, -0.00003345, 0.00004041, -0.00008773, 0.00007883, -0.00003327, 0.00005137, -0.00011803, 0.00009869, -0.00013232, 0.00014794, -0.00020965, 0.00016569, -0.00073348, 0.00099929, 0.00035676, -0.00053882, 0.00017879, -0.00020586, 0.00010370, -0.00006870, 0.00001734, -0.00004959, 0.00005976, -0.00004273, -0.00000611, -0.00000515, 0.00002975, -0.00003233, 0.00001077, -0.00000420, 0.00001320, -0.00000833, -0.00004054, 0.00005460, -0.00006535, 0.00000972, },
		{ 0.00005206, -0.00004857, 0.00017309, -0.00003925, 0.00005143, -0.00006722, -0.00000063, -0.00001592, 0.00000975, -0.00001102, 0.00000182, -0.00000682, 0.00000841, 0.00001597, -0.00001816, 0.00003417, -0.00006049, 0.00004916, -0.00007819, 0.00002813, -0.00001933, 0.00004960, -0.00008959, 0.00004810, 0.00003077, -0.00001045, -0.00003958, 0.00001475, -0.00001861, 0.00002619, -0.00004112, -0.00000440, 0.00002417, 0.00002633, -0.00005111, -0.00000355, -0.00001559, -0.00000002, -0.00000950, -0.00003195, -0.00016393, -0.00036292, 0.00129329, -0.00039972, -0.00009978, -0.00003218, -0.00005566, 0.00002146, -0.00001384, -0.00004087, 0.00002529, 0.00000071, -0.00001460, -0.00002988, 0.00004059, -0.00002040, 0.00000324, -0.00001461, 0.00002198, -0.00000721, -0.00001018, -0.00003015, 0.00005581, -0.00009876, },
		{ -0.00005942, 0.00002342, 0.00002290, 0.00015060, -0.00006187, 0.00005770, -0.00010396, 0.00004824, -0.00005423, 0.00004852, -0.00005120, 0.00003997, -0.00004535, 0.00006073, -0.00004409, 0.00004511, -0.00003830, 0.00000400, -0.00000454, -0.00005318, 0.00002826, -0.00000252, 0.00001702, -0.00008399, 0.00009140, -0.00001196, -0.00000297, -0.00004890, 0.00003593, -0.00003706, 0.00004596, -0.00008545, 0.00006225, -0.00002016, 0.00005436, -0.00011276, 0.00008334, -0.00011058, 0.00011477, -0.00015748, 0.00013165, -0.00055133, 0.00042328, 0.00084857, -0.00061009, 0.00014308, -0.00017376, 0.00006054, -0.00004358, 0.00001309, -0.00005456, 0.00005804, -0.00003517, -0.00000580, -0.00001710, 0.00003952, -0.00003193, 0.00000913, -0.00001127, 0.00002090, -0.00000744, -0.00003319, 0.00001786, -0.00000447, },
		{ -0.00007478, 0.00000323, -0.00002007, 0.00013378, 0.00004850, -0.00000730, -0.00001686, -0.00004912, 0.00001670, -0.00002259, 0.00002115, -0.00003022, 0.00001951, -0.00002087, 0.00004843, -0.00004600, 0.00005730, -0.00007444, 0.00005573, -0.00007242, 0.00000601, 0.00000096, 0.00002953, -0.00004596, -0.00001960, 0.00007426, -0.00001826, -0.00002507, -0.00001736, 0.00001164, -0.00000612, 0.00000051, -0.00004895, 0.00005624, -0.00000317, 0.00000076, -0.00006402, 0.00004869, -0.00007272, 0.00007310, -0.00011906, 0.00002122, -0.00055092, 0.00103300, 0.00009540, -0.00035994, 0.00009211, -0.00013829, 0.00006017, -0.00003843, -0.00001279, -0.00000964, 0.00002983, -0.00002999, -0.00001810, 0.00002000, 0.00000244, -0.00000911, -0.00001104, 0.00001495, -0.00000031, -0.00000071, -0.00004924, 0.00005883, },
		{ 0.00004386, -0.00009304, 0.00002437, 0.00000026, 0.00018147, -0.00003109, 0.00003223, -0.00008152, 0.00001873, -0.00003489, 0.00002959, -0.00003072, 0.00001638, -0.00002653, 0.00003841, -0.00001252, 0.00000952, 0.00000149, -0.00003539, 0.00003605, -0.00008750, 0.00004258, -0.00001292, 0.00003443, -0.00008404, 0.00004987, 0.00003198, -0.00001331, -0.00004544, 0.00001780, -0.00001839, 0.00002716, -0.00005603, 0.00002048, 0.00001436, 0.00003153, -0.00007351, 0.00002979, -0.00004701, 0.00003747, -0.00006152, 0.00003372, -0.00030559, -0.00005123, 0.00104267, -0.00041333, -0.00003481, -0.00006322, -0.00002615, 0.00000483, -0.00001084, -0.00003832, 0.00003621, -0.00001219, -0.00001677, -0.00001701, 0.00003544, -0.00002127, 0.00000405, -0.00001734, 0.00002652, -0.00000983, -0.00001432, -0.00002519, },
		{ 0.00002538, -0.00001909, -0.00005715, 0.00000440, 0.00008931, 0.00013181, -0.00003799, 0.00000468, -0.00007147, 0.00002680, -0.00003720, 0.00003601, -0.00004311, 0.00002685, -0.00003329, 0.00006139, -0.00005004, 0.00005367, -0.00006053, 0.00003373, -0.00003648, -0.00004003, 0.00003443, 0.00000209, -0.00000258, -0.00006535, 0.00007923, 0.00000211, -0.00002326, -0.00003550, 0.00002183, -0.00001681, 0.00001712, -0.00006199, 0.00005396, -0.00000146, 0.00001903, -0.00008105, 0.00005951, -0.00008037, 0.00007748, -0.00011680, 0.00006713, -0.00050627, 0.00060154, 0.00053721, -0.00046406, 0.00009346, -0.00013220, 0.00003810, -0.00002635, -0.00000968, -0.00002272, 0.00003841, -0.00002898, -0.00001646, 0.00000726, 0.00001561, -0.00001299, -0.00000915, 0.00000420, 0.00001081, -0.00000177, -0.00004187, },
		{ -0.00004767, 0.00005216, -0.00007483, -0.00001740, 0.00000342, 0.00017346, 0.00004175, -0.00001212, -0.00004986, -0.00001617, -0.00001152, 0.00000422, -0.00000351, -0.00001121, -0.00000527, 0.00001132, 0.00002248, -0.00002448, 0.00003365, -0.00006162, 0.00005708, -0.00009167, 0.00002233, 0.00000737, 0.00002146, -0.00005159, -0.00000837, 0.00006721, -0.00001153, -0.00004112, -0.00000522, 0.00000229, 0.00000598, -0.00002461, -0.00001786, 0.00004050, 0.00001339, -0.00003524, -0.00001959, 0.00000711, -0.00002294, 0.00000824, -0.00003923, -0.00012097, -0.00031160, 0.00093583, -0.00005062, -0.00023660, 0.00003356, -0.00008969, 0.00003182, -0.00002411, -0.00002453, 0.00001448, 0.00000921, -0.00002573, -0.00001516, 0.00002691, -0.00000878, 0.00000017, -0.00002185, 0.00002627, -0.00000785, -0.00000102, },
		{ -0.00002143, -0.00001610, 0.00002703, -0.00008580, -0.00000626, 0.00006211, 0.00018457, -0.00001910, -0.00001159, -0.00006561, 0.00001432, -0.00003310, 0.00003119, -0.00003510, 0.00001524, -0.00002890, 0.00005402, -0.00003095, 0.00002764, -0.00002735, -0.00000397, 0.00000984, -0.00008252, 0.00005115, -0.00000493, 0.00001520, -0.00007147, 0.00004491, 0.00003927, -0.00002500, -0.00004300, 0.00001490, -0.00001016, 0.00001351, -0.00005040, 0.00002800, 0.00001906, 0.00001535, -0.00006746, 0.00003418, -0.00004770, 0.00003840, -0.00006890, 0.00003814, -0.00035835, 0.00019203, 0.00077041, -0.00035798, -0.00002117, -0.00006161, -0.00001862, 0.00000277, -0.00001969, -0.00002152, 0.00003235, -0.00001661, -0.00002158, 0.00000102, 0.00002033, -0.00001095, -0.00000699, -0.00000901, 0.00002290, -0.00000846, },
		{ -0.00000119, -0.00004281, 0.00002488, -0.00002675, -0.00006622, -0.00000001, 0.00015163, 0.00012429, -0.00003869, -0.00003824, -0.00003486, -0.00000104, -0.00001212, 0.00001430, -0.00002686, 0.00000440, -0.00000685, 0.00004531, -0.00004021, 0.00004266, -0.00006164, 0.00004835, -0.00006233, -0.00002425, 0.00004028, 0.00000185, -0.00001701, -0.00004934, 0.00006940, 0.00001232, -0.00004216, -0.00002286, 0.00001248, -0.00000428, -0.00000655, -0.00003618, 0.00004315, 0.00001434, -0.00001549, -0.00004312, 0.00002722, -0.00004094, 0.00002745, -0.00005823, -0.00003153, -0.00037320, 0.00064879, 0.00030938, -0.00033644, 0.00004495, -0.00009102, 0.00001757, -0.00001486, -0.00002368, 0.00000289, 0.00002019, -0.00002593, -0.00001697, 0.00001939, 0.00000101, -0.00000132, -0.00002283, 0.00001761, 0.00000245, },
		{ -0.00001040, -0.00000678, -0.00003842, 0.00003677, -0.00006821, -0.00004792, 0.00004390, 0.00020758, 0.00004188, -0.00004662, -0.00005170, -0.00000412, -0.00002314, 0.00001758, -0.00001777, -0.00000282, -0.00001898, 0.00003706, -0.00000204, -0.00000364, 0.00000502, -0.00003513, 0.00004127, -0.00009630, 0.00003350, 0.00001599, 0.00000757, -0.00004901, -0.00000410, 0.00006666, -0.00001393, -0.00004945, 0.00000163, 0.00000107, 0.00000365, -0.00003126, -0.00000084, 0.00003676, 0.00001092, -0.00004747, 0.00000306, -0.00001249, 0.00000113, -0.00002660, 0.00000202, -0.00021542, -0.00008057, 0.00076461, -0.00010506, -0.00016666, 0.00000513, -0.00006099, 0.00001754, -0.00002296, -0.00002050, 0.00002354, -0.00000285, -0.00002614, -0.00000349, 0.00002060, -0.00000646, -0.00000414, -0.00002054, 0.00002900, },
		{ 0.00001686, -0.00000889, -0.00002456, -0.00001038, 0.00001230, -0.00008374, -0.00003211, 0.00012913, 0.00019427, -0.00002550, -0.00005660, -0.00003439, -0.00000497, -0.00001661, 0.00001833, -0.00002683, 0.00000048, -0.00001354, 0.00005178, -0.00003455, 0.00002916, -0.00004057, 0.00001987, -0.00001772, -0.00007018, 0.00005696, 0.00000081, -0.00000287, -0.00005716, 0.00003848, 0.00004696, -0.00004014, -0.00003693, 0.00001253, -0.00000404, -0.00000205, -0.00003556, 0.00002693, 0.00002821, -0.00000974, -0.00004487, 0.00002005, -0.00002843, 0.00001441, -0.00004426, -0.00000621, -0.00032176, 0.00033381, 0.00053216, -0.00027979, -0.00003096, -0.00004524, -0.00001977, 0.00000357, -0.00002842, -0.00000234, 0.00002307, -0.00001937, -0.00002126, 0.00001328, 0.00000752, -0.00000116, -0.00001984, 0.00000290, },
		{ 0.00002478, -0.00000218, -0.00000636, -0.00003507, 0.00001482, -0.00002635, -0.00008653, 0.00000968, 0.00021342, 0.00012355, -0.00007082, -0.00005338, -0.00001220, -0.00001963, 0.00000678, -0.00000391, -0.00001449, -0.00001455, 0.00002162, 0.00002264, -0.00002392, 0.00002058, -0.00004528, 0.00004550, -0.00007623, -0.00001091, 0.00004741, -0.00000404, -0.00002447, -0.00003814, 0.00006422, 0.00001565, -0.00005661, -0.00001232, 0.00000885, -0.00000312, -0.00001669, -0.00001856, 0.00003904, 0.00001703, -0.00003503, -0.00001570, 0.00000413, -0.00001222, -0.00001140, -0.00001234, -0.00012122, -0.00020322, 0.00059629, 0.00016823, -0.00024463, 0.00001195, -0.00005974, 0.00000468, -0.00001262, -0.00002539, 0.00001760, 0.00000704, -0.00002634, -0.00000871, 0.00001881, -0.00000135, -0.00000156, -0.00002654, },
		{ -0.00001181, 0.00002669, -0.00001234, -0.00001207, -0.00002785, 0.00001841, -0.00006045, -0.00008193, 0.00009455, 0.00024818, 0.00002725, -0.00009172, -0.00003005, -0.00001159, -0.00001732, 0.00001503, -0.00001855, -0.00000863, -0.00001561, 0.00004788, -0.00001626, 0.00000632, -0.00001517, -0.00000838, 0.00001775, -0.00009044, 0.00004074, 0.00002414, -0.00000816, -0.00004194, -0.00000321, 0.00006884, -0.00002227, -0.00005261, 0.00000778, 0.00000025, -0.00000406, -0.00002592, 0.00000565, 0.00004049, -0.00000490, -0.00003947, 0.00000497, -0.00001045, -0.00000100, -0.00002918, -0.00000176, -0.00024002, 0.00009286, 0.00057771, -0.00010594, -0.00013246, -0.00000303, -0.00004460, 0.00000969, -0.00002636, -0.00000920, 0.00002375, -0.00001103, -0.00002464, 0.00000723, 0.00001138, 0.00000022, -0.00001486, },
		{ -0.00002662, 0.00001240, 0.00001202, -0.00001388, -0.00002016, -0.00001225, 0.00000341, -0.00008950, -0.00005119, 0.00020219, 0.00021123, -0.00006456, -0.00007923, -0.00000841, -0.00002318, 0.00000115, 0.00000233, -0.00001618, -0.00001757, 0.00001079, 0.00003745, -0.00002779, 0.00001716, -0.00003625, 0.00002920, -0.00003812, -0.00005653, 0.00006234, 0.00000224, -0.00001657, -0.00004456, 0.00003414, 0.00005084, -0.00005473, -0.00002959, 0.00001344, -0.00000612, -0.00000899, -0.00002337, 0.00002755, 0.00003031, -0.00002875, -0.00002210, 0.00000376, -0.00000703, -0.00001393, -0.00001231, -0.00006798, -0.00022533, 0.00037757, 0.00035420, -0.00020913, -0.00004361, -0.00002734, -0.00002075, 0.00000036, -0.00003002, 0.00001158, 0.00001430, -0.00002324, -0.00001382, 0.00001536, 0.00000368, 0.00000024, },
		{ -0.00001039, -0.00002143, 0.00002649, -0.00000543, -0.00001329, -0.00002275, -0.00000144, -0.00002111, -0.00011395, 0.00002818, 0.00028635, 0.00011007, -0.00012179, -0.00003961, -0.00001014, -0.00002151, 0.00001146, -0.00001068, -0.00001409, -0.00001996, 0.00004068, 0.00000402, -0.00001210, 0.00000013, -0.00002233, 0.00003095, -0.00007892, -0.00000094, 0.00005455, -0.00001319, -0.00002681, -0.00003017, 0.00006207, 0.00001295, -0.00006699, -0.00000127, 0.00000577, -0.00000819, -0.00001491, -0.00001059, 0.00004239, 0.00000588, -0.00003567, -0.00000557, -0.00000184, -0.00000388, -0.00002629, 0.00000231, -0.00016875, -0.00004823, 0.00049222, 0.00009241, -0.00018536, -0.00000548, -0.00003762, -0.00000312, -0.00001550, -0.00001885, 0.00002363, -0.00000280, -0.00002587, 0.00000121, 0.00001257, 0.00000307, },
		{ 0.00000223, -0.00002360, -0.00000167, 0.00002317, -0.00001729, -0.00001204, -0.00002141, 0.00000187, -0.00005517, -0.00011664, 0.00015457, 0.00030059, -0.00002259, -0.00012055, -0.00000475, -0.00002531, -0.00000419, 0.00000434, -0.00001243, -0.00002252, 0.00000098, 0.00004440, -0.00001968, 0.00000359, -0.00002060, 0.00000880, -0.00000412, -0.00007975, 0.00004612, 0.00002943, -0.00002217, -0.00003400, -0.00000208, 0.00006961, -0.00003290, -0.00005272, 0.00001641, -0.00000709, -0.00000642, -0.00001960, 0.00001059, 0.00004162, -0.00002142, -0.00002380, -0.00000245, 0.00000137, -0.00001635, -0.00001384, -0.00003477, -0.00020221, 0.00018926, 0.00041552, -0.00008722, -0.00011446, -0.00000273, -0.00003233, 0.00000114, -0.00002660, 0.00000135, 0.00002116, -0.00001877, -0.00001756, 0.00001054, 0.00000767, },
		{ 0.00000711, -0.00000710, -0.00002528, 0.00001789, 0.00000788, -0.00002139, -0.00000996, -0.00002084, -0.00000059, -0.00010083, -0.00006787, 0.00029256, 0.00021925, -0.00013178, -0.00006892, -0.00000180, -0.00002748, 0.00000808, -0.00000570, -0.00001391, -0.00002692, 0.00003190, 0.00002119, -0.00002026, 0.00000187, -0.00002157, 0.00002572, -0.00004898, -0.00004418, 0.00006715, 0.00000035, -0.00002614, -0.00003367, 0.00003196, 0.00005002, -0.00006771, -0.00001967, 0.00001444, -0.00001449, -0.00000496, -0.00001920, 0.00003458, 0.00001990, -0.00003256, -0.00001085, -0.00000218, 0.00000212, -0.00002897, 0.00000396, -0.00011301, -0.00011285, 0.00035358, 0.00023450, -0.00015620, -0.00005108, -0.00001235, -0.00002012, -0.00000648, -0.00002469, 0.00001907, 0.00000676, -0.00002601, -0.00000367, 0.00001088, },
		},
		{ { -0.00001506, -0.00001427, 0.00000101, 0.00000444, -0.00000845, 0.00000071, -0.00005816, -0.00002357, -0.00013694, 0.00005911, 0.00038831, 0.00004884, -0.00015675, -0.00003499, 0.00002609, -0.00001236, -0.00000410, -0.00000729, -0.00002336, 0.00000152, 0.00001323, 0.00000124, -0.00001177, -0.00001041, 0.00002091, -0.00000280, 0.00001420, -0.00001841, 0.00003180, 0.00000046, -0.00008836, 0.00001542, 0.00001500, 0.00001542, -0.00008836, 0.00000046, 0.00003180, -0.00001841, 0.00001420, -0.00000280, 0.00002091, -0.00001041, -0.00001177, 0.00000124, 0.00001323, 0.00000152, -0.00002336, -0.00000729, -0.00000410, -0.00001236, 0.00002609, -0.00003499, -0.00015675, 0.00004884, 0.00038831, 0.00005911, -0.00013694, -0.00002357, -0.00005816, 0.00000071, -0.00000845, 0.00000444, 0.00000101, -0.00001427, },
		{ -0.00001624, -0.00001412, -0.00001229, 0.00000674, -0.00000424, 0.00000005, -0.00001736, -0.00005680, -0.00003435, -0.00016105, 0.00021869, 0.00036606, -0.00008873, -0.00013125, 0.00000030, 0.00001696, -0.00001687, -0.00000067, -0.00001467, -0.00001958, 0.00000979, 0.00001077, -0.00000309, -0.00001589, 0.00000039, 0.00001797, -0.00000293, 0.00001245, -0.00001873, 0.00004966, -0.00004809, -0.00006373, 0.00002778, 0.00001574, -0.00001071, -0.00008678, 0.00003884, 0.00000398, -0.00000634, 0.00000809, 0.00000398, 0.00001603, -0.00001641, -0.00000681, 0.00000594, 0.00001161, -0.00000644, -0.00002389, 0.00000022, -0.00001016, -0.00000157, 0.00002502, -0.00008349, -0.00012925, 0.00018145, 0.00032664, -0.00005335, -0.00009230, -0.00003277, -0.00004299, 0.00000467, -0.00000974, 0.00000928, -0.00000490, },
		{ -0.00000992, -0.00001746, -0.00001232, -0.00000921, 0.00000905, -0.00001112, 0.00000879, -0.00004452, -0.00003814, -0.00007601, -0.00013010, 0.00038739, 0.00024565, -0.00018069, -0.00007536, 0.00001375, 0.00000747, -0.00001863, 0.00000128, -0.00002207, -0.00001120, 0.00001492, 0.00000639, -0.00000775, -0.00001631, 0.00001153, 0.00001014, 0.00000209, 0.00000400, -0.00000498, 0.00004546, -0.00008815, -0.00002305, 0.00002528, 0.00001695, -0.00004389, -0.00005816, 0.00005297, -0.00001795, 0.00000812, -0.00000099, 0.00001407, 0.00000473, -0.00001608, -0.00000263, 0.00001041, 0.00000584, -0.00001187, -0.00002226, 0.00000724, -0.00001849, 0.00001727, 0.00000161, -0.00011969, -0.00005553, 0.00027123, 0.00021518, -0.00010605, -0.00005537, -0.00004410, -0.00002239, -0.00000045, -0.00000430, 0.00000888, },
		{ 0.00000553, -0.00001474, -0.00001650, -0.00001111, -0.00000439, 0.00000596, -0.00001110, 0.00000835, -0.00006832, -0.00001377, -0.00014395, -0.00001502, 0.00050294, 0.00005299, -0.00019208, -0.00002678, 0.00001638, -0.00000342, -0.00001551, -0.00000129, -0.00002556, -0.00000115, 0.00001679, 0.00000031, -0.00001131, -0.00001279, 0.00001990, 0.00000068, 0.00000946, -0.00000679, 0.00001813, 0.00001508, -0.00010244, 0.00001667, 0.00001387, 0.00001466, -0.00007270, -0.00001237, 0.00003953, -0.00002225, 0.00001387, -0.00000441, 0.00001975, -0.00000649, -0.00001273, 0.00000261, 0.00001079, 0.00000033, -0.00001726, -0.00001527, 0.00000779, -0.00002086, 0.00003389, -0.00004147, -0.00012236, 0.00003953, 0.00030151, 0.00009249, -0.00010804, -0.00003938, -0.00004422, -0.00000863, -0.00000517, 0.00000235, },
		{ 0.00000616, 0.00000102, -0.00001895, -0.00001370, -0.00001019, 0.00000037, -0.00000039, -0.00000445, -0.00000646, -0.00007571, -0.00000459, -0.00021169, 0.00018889, 0.00049716, -0.00014125, -0.00013228, -0.00000761, 0.00002056, -0.00001882, -0.00000512, -0.00001134, -0.00002085, 0.00000597, 0.00001709, -0.00000786, -0.00001154, -0.00000706, 0.00002409, -0.00000773, 0.00001656, -0.00001470, 0.00004058, -0.00003489, -0.00008217, 0.00003784, 0.00000474, 0.00000226, -0.00008303, 0.00003130, 0.00000967, -0.00000935, 0.00000708, 0.00000238, 0.00001555, -0.00001153, -0.00001027, 0.00001037, 0.00000427, -0.00000062, -0.00002544, -0.00000005, -0.00000409, -0.00000681, 0.00002971, -0.00008265, -0.00009075, 0.00013191, 0.00027158, -0.00000673, -0.00008383, -0.00003951, -0.00003283, -0.00000528, -0.00000446, },
		{ -0.00000064, 0.00000738, -0.00000492, -0.00002076, -0.00001069, -0.00000856, 0.00000274, -0.00000592, 0.00000402, -0.00003425, -0.00005851, -0.00003493, -0.00022837, 0.00043882, 0.00033629, -0.00025205, -0.00005347, -0.00000736, 0.00002274, -0.00003179, 0.00000603, -0.00002379, -0.00001047, 0.00001031, 0.00001426, -0.00001494, -0.00000971, 0.00000107, 0.00002154, -0.00001070, 0.00001934, -0.00001385, 0.00004848, -0.00008604, -0.00003424, 0.00003268, 0.00000622, -0.00002404, -0.00006484, 0.00005215, -0.00001588, 0.00000736, -0.00000437, 0.00001407, 0.00000449, -0.00001101, -0.00000741, 0.00001648, -0.00000491, 0.00000015, -0.00003190, 0.00001705, -0.00002297, 0.00002132, -0.00000350, -0.00009854, -0.00003921, 0.00020490, 0.00019347, -0.00006059, -0.00006057, -0.00004064, -0.00002025, -0.00000622, },
		{ -0.00000411, 0.00000191, 0.00000744, -0.00001208, -0.00001965, -0.00000783, -0.00000727, 0.00000342, -0.00000917, 0.00000913, -0.00006808, -0.00001871, -0.00011991, -0.00013163, 0.00063725, 0.00006165, -0.00023545, -0.00001140, 0.00000077, 0.00000882, -0.00002861, 0.00000505, -0.00002714, -0.00000322, 0.00001672, 0.00000382, -0.00001481, -0.00001037, 0.00001382, 0.00000891, -0.00000274, 0.00001278, 0.00000048, 0.00002893, -0.00011301, 0.00001790, 0.00001056, 0.00001483, -0.00005713, -0.00002192, 0.00004131, -0.00002055, 0.00001216, -0.00000749, 0.00001943, -0.00000413, -0.00001112, 0.00000107, 0.00001295, -0.00000714, -0.00000625, -0.00002565, 0.00002076, -0.00003033, 0.00004224, -0.00004961, -0.00008503, 0.00002076, 0.00024315, 0.00009376, -0.00006972, -0.00005201, -0.00003256, -0.00001546, },
		{ -0.00001677, 0.00000134, 0.00000276, 0.00000538, -0.00001799, -0.00001732, -0.00000404, -0.00000855, 0.00000508, -0.00001113, 0.00000604, -0.00009435, 0.00002325, -0.00024240, 0.00011335, 0.00066672, -0.00021013, -0.00012584, -0.00001934, 0.00002597, -0.00002312, -0.00000640, -0.00001227, -0.00001628, -0.00000356, 0.00002625, -0.00001383, -0.00000602, -0.00001449, 0.00002924, -0.00001166, 0.00001628, -0.00000327, 0.00002550, -0.00001992, -0.00009736, 0.00004610, -0.00000699, 0.00001548, -0.00007816, 0.00002504, 0.00001029, -0.00000516, 0.00000115, 0.00000285, 0.00001264, -0.00000460, -0.00001491, 0.00001580, -0.00000239, 0.00000180, -0.00002344, -0.00000191, 0.00000167, -0.00001080, 0.00003143, -0.00007798, -0.00005842, 0.00008753, 0.00023146, 0.00000873, -0.00005654, -0.00005201, -0.00001925, },
		{ -0.00001268, -0.00001517, 0.00000474, 0.00000448, -0.00000039, -0.00002056, -0.00001444, -0.00000082, -0.00001071, 0.00000656, -0.00001019, -0.00001241, -0.00009211, 0.00002451, -0.00033303, 0.00046277, 0.00046431, -0.00034790, -0.00001538, -0.00003971, 0.00004529, -0.00005095, 0.00001657, -0.00003185, -0.00000150, -0.00000380, 0.00002970, -0.00002799, 0.00000271, -0.00001417, 0.00003663, -0.00002766, 0.00003547, -0.00001863, 0.00004571, -0.00008008, -0.00004340, 0.00003723, -0.00000357, -0.00000518, -0.00006865, 0.00004943, -0.00001556, 0.00001232, -0.00001339, 0.00001762, 0.00000023, -0.00000172, -0.00001610, 0.00002691, -0.00001899, 0.00001243, -0.00004015, 0.00002618, -0.00002841, 0.00002732, -0.00001270, -0.00007154, -0.00003409, 0.00015698, 0.00016748, -0.00003393, -0.00004813, -0.00004543, },
		{ -0.00002681, -0.00001782, -0.00000603, 0.00000339, 0.00000707, -0.00000803, -0.00002146, -0.00000819, -0.00000383, -0.00000661, 0.00000098, -0.00000131, -0.00005239, -0.00004398, -0.00006180, -0.00028384, 0.00077560, 0.00008854, -0.00029557, 0.00001727, -0.00002516, 0.00002769, -0.00004630, 0.00001511, -0.00003236, 0.00000027, 0.00000857, 0.00001496, -0.00002478, -0.00000103, 0.00000111, 0.00002312, -0.00002248, 0.00003789, -0.00001749, 0.00003887, -0.00011826, 0.00001880, 0.00000516, 0.00001621, -0.00004281, -0.00002782, 0.00003861, -0.00001666, 0.00001271, -0.00001448, 0.00002185, -0.00000514, -0.00000540, -0.00000487, 0.00002081, -0.00002053, 0.00000813, -0.00003681, 0.00003378, -0.00004056, 0.00005161, -0.00006023, -0.00004445, -0.00000669, 0.00020790, 0.00007534, -0.00003439, -0.00005446, },
		{ -0.00006266, -0.00000757, -0.00002588, 0.00000497, 0.00000124, 0.00000647, -0.00001296, -0.00002372, 0.00000342, -0.00001635, 0.00000824, -0.00001537, 0.00001406, -0.00010454, 0.00003762, -0.00023978, -0.00001397, 0.00086897, -0.00028610, -0.00011927, -0.00002916, 0.00002879, -0.00002656, -0.00000688, -0.00001507, -0.00000916, -0.00001427, 0.00003331, -0.00001674, -0.00000319, -0.00001814, 0.00002992, -0.00001110, 0.00000849, 0.00001615, 0.00000617, -0.00000574, -0.00010712, 0.00005155, -0.00001866, 0.00002856, -0.00007319, 0.00002127, 0.00000591, 0.00000393, -0.00000608, 0.00000216, 0.00001032, 0.00000116, -0.00001749, 0.00001858, -0.00000446, -0.00000206, -0.00001638, -0.00000624, 0.00000814, -0.00001522, 0.00003287, -0.00007337, -0.00002762, 0.00004292, 0.00020890, -0.00000135, -0.00002019, },
		{ -0.00002150, -0.00005411, -0.00000354, -0.00002384, 0.00000898, 0.00000299, 0.00000058, -0.00001434, -0.00002484, 0.00001381, -0.00002990, 0.00002445, -0.00003133, 0.00001706, -0.00013252, 0.00009054, -0.00042523, 0.00044190, 0.00063627, -0.00046788, 0.00003734, -0.00008159, 0.00007372, -0.00007536, 0.00003306, -0.00004678, 0.00001568, -0.00002587, 0.00005004, -0.00004412, 0.00001811, -0.00003107, 0.00005206, -0.00004457, 0.00004744, -0.00001599, 0.00003631, -0.00007110, -0.00004908, 0.00003811, -0.00001159, 0.00001186, -0.00007028, 0.00004670, -0.00001867, 0.00002283, -0.00002607, 0.00002247, -0.00000573, 0.00000909, -0.00002573, 0.00003797, -0.00003172, 0.00002053, -0.00004371, 0.00003223, -0.00003271, 0.00003296, -0.00002284, -0.00004344, -0.00003369, 0.00011815, 0.00014708, -0.00002904, },
		{ -0.00001135, -0.00004381, -0.00002724, -0.00001643, -0.00000958, 0.00000526, 0.00000648, -0.00000578, -0.00001818, -0.00001577, 0.00000877, -0.00002608, 0.00002117, -0.00002489, -0.00001968, -0.00009005, 0.00002590, -0.00045639, 0.00089695, 0.00014861, -0.00038023, 0.00006396, -0.00006463, 0.00005566, -0.00007095, 0.00003190, -0.00004494, 0.00001300, -0.00001014, 0.00003475, -0.00004167, 0.00001527, -0.00001769, 0.00004224, -0.00004726, 0.00006401, -0.00003126, 0.00004213, -0.00011720, 0.00001956, -0.00000233, 0.00001907, -0.00003055, -0.00003045, 0.00003358, -0.00001367, 0.00001776, -0.00002596, 0.00002701, -0.00000941, 0.00000394, -0.00001448, 0.00003286, -0.00003690, 0.00002232, -0.00004577, 0.00004473, -0.00005007, 0.00006087, -0.00007204, -0.00000314, -0.00003843, 0.00018721, 0.00004996, },
		{ -0.00002705, 0.00001449, -0.00006504, -0.00000144, -0.00003046, 0.00000545, 0.00000199, 0.00000462, -0.00000577, -0.00002949, 0.00000785, -0.00001843, 0.00000459, -0.00001219, 0.00001149, -0.00009880, 0.00002926, -0.00019611, -0.00019000, 0.00108930, -0.00035416, -0.00012309, -0.00002943, 0.00002328, -0.00002467, -0.00001036, -0.00001566, -0.00000446, -0.00002046, 0.00003307, -0.00001256, -0.00000647, -0.00001498, 0.00002365, -0.00000385, -0.00000764, 0.00004153, -0.00001379, 0.00000443, -0.00010941, 0.00005346, -0.00002953, 0.00004101, -0.00006868, 0.00002014, -0.00000185, 0.00001446, -0.00001083, -0.00000238, 0.00001082, 0.00000353, -0.00001592, 0.00001657, -0.00000015, -0.00001249, -0.00000519, -0.00001206, 0.00001489, -0.00002037, 0.00003511, -0.00007067, 0.00000335, -0.00000277, 0.00020083, },
		{ 0.00013661, -0.00004285, 0.00001194, -0.00006066, 0.00000275, -0.00002883, 0.00000965, 0.00000495, -0.00000371, -0.00000172, -0.00004058, 0.00003250, -0.00005096, 0.00004344, -0.00005423, 0.00004727, -0.00017019, 0.00014870, -0.00048637, 0.00036394, 0.00085107, -0.00060492, 0.00009814, -0.00012762, 0.00010381, -0.00010189, 0.00005322, -0.00006634, 0.00003791, -0.00005139, 0.00007028, -0.00005887, 0.00003227, -0.00004540, 0.00006369, -0.00005688, 0.00005115, -0.00000455, 0.00002149, -0.00006128, -0.00004939, 0.00003450, -0.00001700, 0.00002627, -0.00007025, 0.00004492, -0.00002507, 0.00003658, -0.00003883, 0.00002539, -0.00001036, 0.00001814, -0.00003288, 0.00004574, -0.00003884, 0.00002119, -0.00004072, 0.00003398, -0.00003456, 0.00003642, -0.00003122, -0.00001836, -0.00003274, 0.00008208, },
		{ 0.00017230, 0.00002761, -0.00000486, -0.00002175, -0.00003097, -0.00001413, -0.00001191, 0.00000495, 0.00000843, -0.00000859, -0.00000690, -0.00003241, 0.00003100, -0.00005442, 0.00005042, -0.00006009, 0.00002726, -0.00015173, 0.00013161, -0.00062794, 0.00097867, 0.00025418, -0.00049334, 0.00012992, -0.00011799, 0.00009295, -0.00010315, 0.00005669, -0.00006647, 0.00003597, -0.00003882, 0.00006137, -0.00006343, 0.00003624, -0.00003980, 0.00006307, -0.00007262, 0.00008561, -0.00003684, 0.00003729, -0.00011015, 0.00002112, -0.00001214, 0.00002368, -0.00002096, -0.00003043, 0.00002810, -0.00001343, 0.00002739, -0.00004025, 0.00003304, -0.00001499, 0.00001457, -0.00002515, 0.00004569, -0.00005173, 0.00003202, -0.00004941, 0.00005137, -0.00005698, 0.00006811, -0.00008265, 0.00003501, -0.00006869, },
		{ -0.00004687, 0.00020058, -0.00005724, 0.00003875, -0.00005587, -0.00000171, -0.00003075, 0.00000387, 0.00000357, 0.00000204, -0.00000001, -0.00003022, 0.00000475, -0.00001079, -0.00000932, 0.00000205, -0.00000581, -0.00007271, -0.00000611, -0.00011173, -0.00040184, 0.00130561, -0.00039591, -0.00014864, -0.00001258, 0.00000394, -0.00001308, -0.00002081, -0.00000963, -0.00000718, -0.00001713, 0.00002174, 0.00000103, -0.00001752, -0.00000384, 0.00000992, 0.00000998, -0.00003026, 0.00006849, -0.00002973, 0.00000756, -0.00010300, 0.00005176, -0.00003919, 0.00005242, -0.00006490, 0.00002104, -0.00001071, 0.00002304, -0.00001050, -0.00001177, 0.00001466, 0.00000205, -0.00001008, 0.00000979, 0.00000972, -0.00002716, 0.00000712, -0.00001693, 0.00002035, -0.00002528, 0.00003769, -0.00006964, 0.00003344, },
		{ -0.00002849, 0.00004635, 0.00013514, -0.00006786, 0.00004325, -0.00005953, 0.00000291, -0.00002813, 0.00000582, 0.00001072, -0.00001281, 0.00001518, -0.00005786, 0.00005026, -0.00006846, 0.00005793, -0.00007242, 0.00007018, -0.00019486, 0.00018532, -0.00050239, 0.00022539, 0.00109819, -0.00074498, 0.00015600, -0.00016940, 0.00012904, -0.00012532, 0.00007259, -0.00008589, 0.00005969, -0.00007407, 0.00008452, -0.00006742, 0.00004089, -0.00005317, 0.00006799, -0.00006126, 0.00004466, 0.00001438, 0.00000452, -0.00005371, -0.00004242, 0.00002583, -0.00001915, 0.00003730, -0.00006873, 0.00004405, -0.00003308, 0.00004997, -0.00004770, 0.00002344, -0.00001108, 0.00002280, -0.00003500, 0.00004746, -0.00003788, 0.00001348, -0.00003158, 0.00003190, -0.00003389, 0.00003705, -0.00003643, 0.00000126, },
		{ 0.00006628, -0.00009235, 0.00015653, 0.00001373, -0.00001373, 0.00000621, -0.00003322, -0.00001462, -0.00001001, 0.00000012, 0.00001510, -0.00001846, 0.00001316, -0.00005720, 0.00006039, -0.00008812, 0.00008475, -0.00010195, 0.00008162, -0.00021907, 0.00023896, -0.00077504, 0.00100122, 0.00041112, -0.00063247, 0.00021135, -0.00018170, 0.00013669, -0.00014080, 0.00008810, -0.00009555, 0.00006681, -0.00007362, 0.00009032, -0.00008588, 0.00005776, -0.00006094, 0.00008122, -0.00009343, 0.00009772, -0.00003207, 0.00002504, -0.00009901, 0.00002516, -0.00002476, 0.00003043, -0.00001468, -0.00002819, 0.00002315, -0.00001595, 0.00003947, -0.00005393, 0.00003681, -0.00001892, 0.00002327, -0.00003348, 0.00005524, -0.00006053, 0.00003373, -0.00004572, 0.00005240, -0.00005990, 0.00007166, -0.00008972, },
		{ -0.00006875, 0.00005997, -0.00008491, 0.00020060, -0.00008254, 0.00004786, -0.00003587, -0.00000624, -0.00002915, 0.00000286, 0.00000330, 0.00000189, 0.00000035, -0.00002201, -0.00000892, 0.00000846, -0.00003452, 0.00002827, -0.00003850, -0.00002641, -0.00006640, 0.00000447, -0.00062822, 0.00149159, -0.00039333, -0.00020501, 0.00002653, -0.00003260, 0.00001089, -0.00004089, 0.00000609, -0.00002061, -0.00000165, -0.00000160, 0.00002358, -0.00003554, 0.00001403, -0.00000939, 0.00002789, -0.00005521, 0.00009132, -0.00003724, 0.00000187, -0.00008809, 0.00004739, -0.00004770, 0.00006262, -0.00006207, 0.00002320, -0.00001863, 0.00002752, -0.00000460, -0.00002479, 0.00002041, -0.00000178, -0.00000179, 0.00000036, 0.00002213, -0.00004183, 0.00001645, -0.00001798, 0.00002262, -0.00002839, 0.00003916, },
		{ -0.00003857, 0.00001498, -0.00002081, 0.00001204, 0.00013799, -0.00009483, 0.00006463, -0.00004757, -0.00000396, -0.00002195, -0.00000140, 0.00001868, -0.00002416, 0.00003215, -0.00007133, 0.00006141, -0.00007691, 0.00006262, -0.00007996, 0.00007877, -0.00019834, 0.00019089, -0.00046707, 0.00003347, 0.00135784, -0.00086864, 0.00019726, -0.00019712, 0.00014198, -0.00013964, 0.00008571, -0.00009972, 0.00007476, -0.00008758, 0.00008764, -0.00006611, 0.00004098, -0.00005196, 0.00006325, -0.00005660, 0.00002889, 0.00003688, -0.00001000, -0.00005154, -0.00002691, 0.00001225, -0.00001778, 0.00004437, -0.00006565, 0.00004334, -0.00004034, 0.00005943, -0.00004971, 0.00001516, -0.00000688, 0.00002212, -0.00003134, 0.00004250, -0.00002902, -0.00000078, -0.00001889, 0.00002803, -0.00003184, 0.00003544, },
		{ 0.00007072, -0.00009193, 0.00008830, -0.00010639, 0.00013672, 0.00000889, -0.00003247, 0.00003264, -0.00002913, -0.00002088, -0.00000196, -0.00001020, 0.00002700, -0.00003513, 0.00003999, -0.00008632, 0.00009182, -0.00012141, 0.00011800, -0.00014314, 0.00013401, -0.00027971, 0.00033029, -0.00087680, 0.00095266, 0.00061598, -0.00078760, 0.00029905, -0.00024826, 0.00018093, -0.00017912, 0.00012197, -0.00012779, 0.00010021, -0.00010825, 0.00011552, -0.00010388, 0.00007506, -0.00007653, 0.00009243, -0.00010539, 0.00009746, -0.00001758, 0.00000837, -0.00008696, 0.00003370, -0.00004060, 0.00003964, -0.00001230, -0.00002380, 0.00001867, -0.00001969, 0.00005050, -0.00006297, 0.00003521, -0.00001842, 0.00002718, -0.00003656, 0.00005827, -0.00006032, 0.00002603, -0.00003472, 0.00004804, -0.00005853, },
		{ -0.00002833, 0.00003797, -0.00006611, 0.00007994, -0.00011258, 0.00019480, -0.00009733, 0.00004195, -0.00000944, -0.00001086, -0.00002915, 0.00000554, -0.00000152, 0.00000698, -0.00000756, -0.00000302, -0.00003342, 0.00003809, -0.00006890, 0.00006399, -0.00008325, 0.00003495, -0.00014331, 0.00013656, -0.00084300, 0.00162145, -0.00033303, -0.00029594, 0.00008844, -0.00008597, 0.00004691, -0.00007067, 0.00003199, -0.00004505, 0.00002518, -0.00003448, 0.00005175, -0.00005728, 0.00003522, -0.00003046, 0.00004571, -0.00007718, 0.00010457, -0.00003366, -0.00001225, -0.00006665, 0.00004235, -0.00005572, 0.00007180, -0.00006055, 0.00002615, -0.00002455, 0.00002764, 0.00000508, -0.00003831, 0.00002514, -0.00000517, 0.00000588, -0.00000840, 0.00003299, -0.00005184, 0.00001908, -0.00001306, 0.00002036, },
		{ -0.00003016, 0.00003301, -0.00003891, 0.00002394, -0.00001157, -0.00001765, 0.00013908, -0.00011545, 0.00007148, -0.00002512, -0.00001597, -0.00001269, -0.00000916, 0.00002573, -0.00003387, 0.00004416, -0.00007574, 0.00006123, -0.00007247, 0.00005417, -0.00007321, 0.00006910, -0.00017679, 0.00016241, -0.00038378, -0.00019425, 0.00160351, -0.00095444, 0.00020859, -0.00020199, 0.00013632, -0.00013965, 0.00008772, -0.00010273, 0.00007793, -0.00008747, 0.00007695, -0.00005377, 0.00003187, -0.00004167, 0.00005018, -0.00004445, 0.00000747, 0.00005730, -0.00001741, -0.00005692, -0.00000295, -0.00000501, -0.00001340, 0.00004740, -0.00006106, 0.00004195, -0.00004476, 0.00006257, -0.00004401, 0.00000131, 0.00000123, 0.00001713, -0.00002326, 0.00003247, -0.00001487, -0.00001763, -0.00000660, 0.00002518, },
		{ 0.00004008, -0.00005376, 0.00006573, -0.00008935, 0.00010060, -0.00011038, 0.00011311, 0.00000983, -0.00005335, 0.00005055, -0.00001563, -0.00003390, 0.00001210, -0.00002498, 0.00004254, -0.00005593, 0.00006903, -0.00011381, 0.00011869, -0.00014766, 0.00014339, -0.00017578, 0.00017484, -0.00032181, 0.00039059, -0.00091925, 0.00083186, 0.00085494, -0.00094173, 0.00037956, -0.00030736, 0.00021767, -0.00021151, 0.00015227, -0.00015690, 0.00012923, -0.00013561, 0.00013098, -0.00011284, 0.00008415, -0.00008313, 0.00009395, -0.00010639, 0.00008491, 0.00000309, -0.00000805, -0.00007762, 0.00004849, -0.00005960, 0.00005125, -0.00001421, -0.00001712, 0.00001385, -0.00002226, 0.00005676, -0.00006416, 0.00002637, -0.00001203, 0.00002483, -0.00003307, 0.00005349, -0.00005061, 0.00001019, -0.00001863, },
		{ -0.00000172, 0.00001349, -0.00002455, 0.00003321, -0.00006045, 0.00009134, -0.00012725, 0.00018011, -0.00010049, 0.00002537, 0.00001688, -0.00001087, -0.00003397, 0.00001420, -0.00001242, 0.00001858, -0.00002433, 0.00002559, -0.00006577, 0.00007375, -0.00010742, 0.00010374, -0.00013329, 0.00010232, -0.00022418, 0.00026490, -0.00102005, 0.00167507, -0.00021004, -0.00041768, 0.00016796, -0.00015135, 0.00009104, -0.00010707, 0.00006539, -0.00007738, 0.00005912, -0.00007142, 0.00008001, -0.00007793, 0.00005508, -0.00004865, 0.00005889, -0.00009134, 0.00010469, -0.00001912, -0.00003198, -0.00004222, 0.00003940, -0.00006422, 0.00008035, -0.00006087, 0.00002992, -0.00002862, 0.00002492, 0.00001505, -0.00004824, 0.00002554, -0.00000508, 0.00000979, -0.00001322, 0.00003856, -0.00005362, 0.00001303, },
		{ 0.00000122, 0.00002580, -0.00003039, 0.00003128, -0.00003915, 0.00003002, -0.00000340, -0.00003911, 0.00013317, -0.00012452, 0.00006349, 0.00000405, -0.00002906, -0.00000415, -0.00001383, 0.00002841, -0.00003805, 0.00004695, -0.00006759, 0.00004760, -0.00005431, 0.00003241, -0.00005203, 0.00004145, -0.00013181, 0.00010423, -0.00026488, -0.00043251, 0.00180626, -0.00098317, 0.00018040, -0.00017878, 0.00010881, -0.00012277, 0.00007596, -0.00009212, 0.00006676, -0.00007253, 0.00005316, -0.00003218, 0.00001550, -0.00002468, 0.00003181, -0.00002849, -0.00001439, 0.00006987, -0.00001453, -0.00006995, 0.00002753, -0.00002345, -0.00000732, 0.00004689, -0.00005536, 0.00003948, -0.00004532, 0.00005904, -0.00003216, -0.00001521, 0.00001045, 0.00001053, -0.00001369, 0.00002064, 0.00000039, -0.00003219, },
		{ -0.00001014, -0.00000128, 0.00003129, -0.00004729, 0.00005806, -0.00008329, 0.00010444, -0.00010613, 0.00008845, 0.00001139, -0.00006900, 0.00005548, 0.00000734, -0.00005206, 0.00002978, -0.00004123, 0.00005831, -0.00007636, 0.00009436, -0.00013318, 0.00013471, -0.00016117, 0.00015537, -0.00019350, 0.00019663, -0.00033697, 0.00041077, -0.00089818, 0.00064949, 0.00110513, -0.00107339, 0.00043769, -0.00034807, 0.00023867, -0.00023120, 0.00017263, -0.00017631, 0.00014727, -0.00014975, 0.00013253, -0.00011014, 0.00008305, -0.00007944, 0.00008544, -0.00009708, 0.00006321, 0.00002429, -0.00001913, -0.00007390, 0.00007014, -0.00008072, 0.00006449, -0.00002029, -0.00000813, 0.00000778, -0.00002150, 0.00005571, -0.00005624, 0.00001055, -0.00000029, 0.00001682, -0.00002384, 0.00004199, -0.00003355, },
		{ -0.00004588, -0.00000118, 0.00001439, 0.00000342, -0.00001758, 0.00002494, -0.00005166, 0.00009374, -0.00012869, 0.00015680, -0.00009476, 0.00000496, 0.00003644, -0.00000285, -0.00004514, 0.00002925, -0.00002892, 0.00003567, -0.00004782, 0.00005969, -0.00010021, 0.00010882, -0.00014323, 0.00014034, -0.00018006, 0.00016479, -0.00029490, 0.00037023, -0.00113825, 0.00164226, -0.00003010, -0.00055841, 0.00025439, -0.00021990, 0.00013628, -0.00014431, 0.00010086, -0.00011174, 0.00009350, -0.00010535, 0.00010220, -0.00009252, 0.00006912, -0.00005989, 0.00006396, -0.00009475, 0.00009115, 0.00000312, -0.00005249, -0.00001916, 0.00004136, -0.00007401, 0.00008855, -0.00006355, 0.00003490, -0.00003191, 0.00002194, 0.00002139, -0.00005099, 0.00001918, 0.00000052, 0.00000791, -0.00001212, 0.00003677, },
		{ 0.00001248, -0.00004022, 0.00000188, 0.00003101, -0.00003311, 0.00003128, -0.00004071, 0.00003493, 0.00000146, -0.00004985, 0.00011751, -0.00012086, 0.00004441, 0.00003384, -0.00003833, -0.00000014, -0.00001232, 0.00002417, -0.00003419, 0.00003848, -0.00004643, 0.00002182, -0.00002509, 0.00000059, -0.00001991, 0.00000038, -0.00007019, 0.00002690, -0.00012917, -0.00065292, 0.00194005, -0.00094228, 0.00010978, -0.00012775, 0.00006063, -0.00009005, 0.00005103, -0.00006837, 0.00004244, -0.00004530, 0.00002031, -0.00000577, -0.00000411, -0.00000511, 0.00001253, -0.00001336, -0.00003136, 0.00007041, -0.00000092, -0.00008821, 0.00006089, -0.00003955, -0.00000150, 0.00004388, -0.00004938, 0.00003625, -0.00004244, 0.00005052, -0.00001772, -0.00003019, 0.00001721, 0.00000568, -0.00000610, 0.00001066, },
		{ 0.00002678, -0.00001324, -0.00003003, 0.00001307, 0.00002444, -0.00004088, 0.00004945, -0.00007575, 0.00010213, -0.00009668, 0.00006645, 0.00000864, -0.00007442, 0.00004670, 0.00003663, -0.00007151, 0.00004708, -0.00005485, 0.00007012, -0.00009150, 0.00011032, -0.00013919, 0.00013574, -0.00015881, 0.00015115, -0.00019306, 0.00019591, -0.00032221, 0.00038952, -0.00081992, 0.00042649, 0.00133816, -0.00116059, 0.00045991, -0.00036150, 0.00023768, -0.00023307, 0.00017812, -0.00018111, 0.00014989, -0.00014756, 0.00011909, -0.00009602, 0.00007236, -0.00006674, 0.00006905, -0.00008065, 0.00003764, 0.00003977, -0.00002096, -0.00007688, 0.00009756, -0.00010165, 0.00007769, -0.00002968, 0.00000269, 0.00000006, -0.00001638, 0.00004685, -0.00004052, -0.00000966, 0.00001426, 0.00000565, -0.00001156, },
		{ -0.00000514, 0.00002786, -0.00003003, -0.00002056, 0.00003192, -0.00000742, -0.00000882, 0.00001424, -0.00004080, 0.00008838, -0.00011897, 0.00012784, -0.00008512, -0.00001223, 0.00004455, 0.00001405, -0.00006177, 0.00004879, -0.00004841, 0.00005509, -0.00007357, 0.00009307, -0.00012966, 0.00013613, -0.00016947, 0.00016680, -0.00021529, 0.00021225, -0.00034324, 0.00043774, -0.00118544, 0.00152518, 0.00019036, -0.00069975, 0.00033328, -0.00028050, 0.00017396, -0.00017521, 0.00013158, -0.00014101, 0.00012102, -0.00012949, 0.00011316, -0.00009743, 0.00007434, -0.00006190, 0.00005961, -0.00008726, 0.00006669, 0.00002731, -0.00006820, -0.00000145, 0.00005025, -0.00008519, 0.00009617, -0.00006870, 0.00004145, -0.00003571, 0.00002127, 0.00002107, -0.00004469, 0.00000562, 0.00001179, 0.00000015, },
		{ -0.00000322, 0.00000535, 0.00001837, -0.00003944, -0.00000510, 0.00004006, -0.00003769, 0.00003316, -0.00004420, 0.00003959, 0.00000195, -0.00004943, 0.00009240, -0.00010708, 0.00002060, 0.00005766, -0.00003970, -0.00000314, -0.00000322, 0.00001241, -0.00002209, 0.00001965, -0.00001522, -0.00001143, 0.00000955, -0.00003536, 0.00001690, -0.00004640, -0.00000227, -0.00005529, 0.00000212, -0.00082925, 0.00198680, -0.00082925, 0.00000212, -0.00005529, -0.00000227, -0.00004640, 0.00001690, -0.00003536, 0.00000955, -0.00001143, -0.00001522, 0.00001965, -0.00002209, 0.00001241, -0.00000322, -0.00000314, -0.00003970, 0.00005766, 0.00002060, -0.00010708, 0.00009240, -0.00004943, 0.00000195, 0.00003959, -0.00004420, 0.00003316, -0.00003769, 0.00004006, -0.00000510, -0.00003944, 0.00001837, 0.00000535, },
		{ -0.00000514, 0.00000015, 0.00001179, 0.00000562, -0.00004469, 0.00002107, 0.00002127, -0.00003571, 0.00004145, -0.00006870, 0.00009617, -0.00008519, 0.00005025, -0.00000145, -0.00006820, 0.00002731, 0.00006669, -0.00008726, 0.00005961, -0.00006190, 0.00007434, -0.00009743, 0.00011316, -0.00012949, 0.00012102, -0.00014101, 0.00013158, -0.00017521, 0.00017396, -0.00028050, 0.00033328, -0.00069975, 0.00019036, 0.00152518, -0.00118544, 0.00043774, -0.00034324, 0.00021225, -0.00021529, 0.00016680, -0.00016947, 0.00013613, -0.00012966, 0.00009307, -0.00007357, 0.00005509, -0.00004841, 0.00004879, -0.00006177, 0.00001405, 0.00004455, -0.00001223, -0.00008512, 0.00012784, -0.00011897, 0.00008838, -0.00004080, 0.00001424, -0.00000882, -0.00000742, 0.00003192, -0.00002056, -0.00003003, 0.00002786, },
		{ 0.00002678, -0.00001156, 0.00000565, 0.00001426, -0.00000966, -0.00004052, 0.00004685, -0.00001638, 0.00000006, 0.00000269, -0.00002968, 0.00007769, -0.00010165, 0.00009756, -0.00007688, -0.00002096, 0.00003977, 0.00003764, -0.00008065, 0.00006905, -0.00006674, 0.00007236, -0.00009602, 0.00011909, -0.00014756, 0.00014989, -0.00018111, 0.00017812, -0.00023307, 0.00023768, -0.00036150, 0.00045991, -0.00116059, 0.00133816, 0.00042649, -0.00081992, 0.00038952, -0.00032221, 0.00019591, -0.00019306, 0.00015115, -0.00015881, 0.00013574, -0.00013919, 0.00011032, -0.00009150, 0.00007012, -0.00005485, 0.00004708, -0.00007151, 0.00003663, 0.00004670, -0.00007442, 0.00000864, 0.00006645, -0.00009668, 0.00010213, -0.00007575, 0.00004945, -0.00004088, 0.00002444, 0.00001307, -0.00003003, -0.00001324, },
		{ 0.00001248, 0.00001066, -0.00000610, 0.00000568, 0.00001721, -0.00003019, -0.00001772, 0.00005052, -0.00004244, 0.00003625, -0.00004938, 0.00004388, -0.00000150, -0.00003955, 0.00006089, -0.00008821, -0.00000092, 0.00007041, -0.00003136, -0.00001336, 0.00001253, -0.00000511, -0.00000411, -0.00000577, 0.00002031, -0.00004530, 0.00004244, -0.00006837, 0.00005103, -0.00009005, 0.00006063, -0.00012775, 0.00010978, -0.00094228, 0.00194005, -0.00065292, -0.00012917, 0.00002690, -0.00007019, 0.00000038, -0.00001991, 0.00000059, -0.00002509, 0.00002182, -0.00004643, 0.00003848, -0.00003419, 0.00002417, -0.00001232, -0.00000014, -0.00003833, 0.00003384, 0.00004441, -0.00012086, 0.00011751, -0.00004985, 0.00000146, 0.00003493, -0.00004071, 0.00003128, -0.00003311, 0.00003101, 0.00000188, -0.00004022, },
		{ -0.00004588, 0.00003677, -0.00001212, 0.00000791, 0.00000052, 0.00001918, -0.00005099, 0.00002139, 0.00002194, -0.00003191, 0.00003490, -0.00006355, 0.00008855, -0.00007401, 0.00004136, -0.00001916, -0.00005249, 0.00000312, 0.00009115, -0.00009475, 0.00006396, -0.00005989, 0.00006912, -0.00009252, 0.00010220, -0.00010535, 0.00009350, -0.00011174, 0.00010086, -0.00014431, 0.00013628, -0.00021990, 0.00025439, -0.00055841, -0.00003010, 0.00164226, -0.00113825, 0.00037023, -0.00029490, 0.00016479, -0.00018006, 0.00014034, -0.00014323, 0.00010882, -0.00010021, 0.00005969, -0.00004782, 0.00003567, -0.00002892, 0.00002925, -0.00004514, -0.00000285, 0.00003644, 0.00000496, -0.00009476, 0.00015680, -0.00012869, 0.00009374, -0.00005166, 0.00002494, -0.00001758, 0.00000342, 0.00001439, -0.00000118, },
		{ -0.00001014, -0.00003355, 0.00004199, -0.00002384, 0.00001682, -0.00000029, 0.00001055, -0.00005624, 0.00005571, -0.00002150, 0.00000778, -0.00000813, -0.00002029, 0.00006449, -0.00008072, 0.00007014, -0.00007390, -0.00001913, 0.00002429, 0.00006321, -0.00009708, 0.00008544, -0.00007944, 0.00008305, -0.00011014, 0.00013253, -0.00014975, 0.00014727, -0.00017631, 0.00017263, -0.00023120, 0.00023867, -0.00034807, 0.00043769, -0.00107339, 0.00110513, 0.00064949, -0.00089818, 0.00041077, -0.00033697, 0.00019663, -0.00019350, 0.00015537, -0.00016117, 0.00013471, -0.00013318, 0.00009436, -0.00007636, 0.00005831, -0.00004123, 0.00002978, -0.00005206, 0.00000734, 0.00005548, -0.00006900, 0.00001139, 0.00008845, -0.00010613, 0.00010444, -0.00008329, 0.00005806, -0.00004729, 0.00003129, -0.00000128, },
		{ 0.00000122, -0.00003219, 0.00000039, 0.00002064, -0.00001369, 0.00001053, 0.00001045, -0.00001521, -0.00003216, 0.00005904, -0.00004532, 0.00003948, -0.00005536, 0.00004689, -0.00000732, -0.00002345, 0.00002753, -0.00006995, -0.00001453, 0.00006987, -0.00001439, -0.00002849, 0.00003181, -0.00002468, 0.00001550, -0.00003218, 0.00005316, -0.00007253, 0.00006676, -0.00009212, 0.00007596, -0.00012277, 0.00010881, -0.00017878, 0.00018040, -0.00098317, 0.00180626, -0.00043251, -0.00026488, 0.00010423, -0.00013181, 0.00004145, -0.00005203, 0.00003241, -0.00005431, 0.00004760, -0.00006759, 0.00004695, -0.00003805, 0.00002841, -0.00001383, -0.00000415, -0.00002906, 0.00000405, 0.00006349, -0.00012452, 0.00013317, -0.00003911, -0.00000340, 0.00003002, -0.00003915, 0.00003128, -0.00003039, 0.00002580, },
		{ -0.00000172, 0.00001303, -0.00005362, 0.00003856, -0.00001322, 0.00000979, -0.00000508, 0.00002554, -0.00004824, 0.00001505, 0.00002492, -0.00002862, 0.00002992, -0.00006087, 0.00008035, -0.00006422, 0.00003940, -0.00004222, -0.00003198, -0.00001912, 0.00010469, -0.00009134, 0.00005889, -0.00004865, 0.00005508, -0.00007793, 0.00008001, -0.00007142, 0.00005912, -0.00007738, 0.00006539, -0.00010707, 0.00009104, -0.00015135, 0.00016796, -0.00041768, -0.00021004, 0.00167507, -0.00102005, 0.00026490, -0.00022418, 0.00010232, -0.00013329, 0.00010374, -0.00010742, 0.00007375, -0.00006577, 0.00002559, -0.00002433, 0.00001858, -0.00001242, 0.00001420, -0.00003397, -0.00001087, 0.00001688, 0.00002537, -0.00010049, 0.00018011, -0.00012725, 0.00009134, -0.00006045, 0.00003321, -0.00002455, 0.00001349, },
		{ 0.00004008, -0.00001863, 0.00001019, -0.00005061, 0.00005349, -0.00003307, 0.00002483, -0.00001203, 0.00002637, -0.00006416, 0.00005676, -0.00002226, 0.00001385, -0.00001712, -0.00001421, 0.00005125, -0.00005960, 0.00004849, -0.00007762, -0.00000805, 0.00000309, 0.00008491, -0.00010639, 0.00009395, -0.00008313, 0.00008415, -0.00011284, 0.00013098, -0.00013561, 0.00012923, -0.00015690, 0.00015227, -0.00021151, 0.00021767, -0.00030736, 0.00037956, -0.00094173, 0.00085494, 0.00083186, -0.00091925, 0.00039059, -0.00032181, 0.00017484, -0.00017578, 0.00014339, -0.00014766, 0.00011869, -0.00011381, 0.00006903, -0.00005593, 0.00004254, -0.00002498, 0.00001210, -0.00003390, -0.00001563, 0.00005055, -0.00005335, 0.00000983, 0.00011311, -0.00011038, 0.00010060, -0.00008935, 0.00006573, -0.00005376, },
		{ -0.00003016, 0.00002518, -0.00000660, -0.00001763, -0.00001487, 0.00003247, -0.00002326, 0.00001713, 0.00000123, 0.00000131, -0.00004401, 0.00006257, -0.00004476, 0.00004195, -0.00006106, 0.00004740, -0.00001340, -0.00000501, -0.00000295, -0.00005692, -0.00001741, 0.00005730, 0.00000747, -0.00004445, 0.00005018, -0.00004167, 0.00003187, -0.00005377, 0.00007695, -0.00008747, 0.00007793, -0.00010273, 0.00008772, -0.00013965, 0.00013632, -0.00020199, 0.00020859, -0.00095444, 0.00160351, -0.00019425, -0.00038378, 0.00016241, -0.00017679, 0.00006910, -0.00007321, 0.00005417, -0.00007247, 0.00006123, -0.00007574, 0.00004416, -0.00003387, 0.00002573, -0.00000916, -0.00001269, -0.00001597, -0.00002512, 0.00007148, -0.00011545, 0.00013908, -0.00001765, -0.00001157, 0.00002394, -0.00003891, 0.00003301, },
		{ -0.00002833, 0.00002036, -0.00001306, 0.00001908, -0.00005184, 0.00003299, -0.00000840, 0.00000588, -0.00000517, 0.00002514, -0.00003831, 0.00000508, 0.00002764, -0.00002455, 0.00002615, -0.00006055, 0.00007180, -0.00005572, 0.00004235, -0.00006665, -0.00001225, -0.00003366, 0.00010457, -0.00007718, 0.00004571, -0.00003046, 0.00003522, -0.00005728, 0.00005175, -0.00003448, 0.00002518, -0.00004505, 0.00003199, -0.00007067, 0.00004691, -0.00008597, 0.00008844, -0.00029594, -0.00033303, 0.00162145, -0.00084300, 0.00013656, -0.00014331, 0.00003495, -0.00008325, 0.00006399, -0.00006890, 0.00003809, -0.00003342, -0.00000302, -0.00000756, 0.00000698, -0.00000152, 0.00000554, -0.00002915, -0.00001086, -0.00000944, 0.00004195, -0.00009733, 0.00019480, -0.00011258, 0.00007994, -0.00006611, 0.00003797, },
		{ 0.00007072, -0.00005853, 0.00004804, -0.00003472, 0.00002603, -0.00006032, 0.00005827, -0.00003656, 0.00002718, -0.00001842, 0.00003521, -0.00006297, 0.00005050, -0.00001969, 0.00001867, -0.00002380, -0.00001230, 0.00003964, -0.00004060, 0.00003370, -0.00008696, 0.00000837, -0.00001758, 0.00009746, -0.00010539, 0.00009243, -0.00007653, 0.00007506, -0.00010388, 0.00011552, -0.00010825, 0.00010021, -0.00012779, 0.00012197, -0.00017912, 0.00018093, -0.00024826, 0.00029905, -0.00078760, 0.00061598, 0.00095266, -0.00087680, 0.00033029, -0.00027971, 0.00013401, -0.00014314, 0.00011800, -0.00012141, 0.00009182, -0.00008632, 0.00003999, -0.00003513, 0.00002700, -0.00001020, -0.00000196, -0.00002088, -0.00002913, 0.00003264, -0.00003247, 0.00000889, 0.00013672, -0.00010639, 0.00008830, -0.00009193, },
		{ -0.00003857, 0.00003544, -0.00003184, 0.00002803, -0.00001889, -0.00000078, -0.00002902, 0.00004250, -0.00003134, 0.00002212, -0.00000688, 0.00001516, -0.00004971, 0.00005943, -0.00004034, 0.00004334, -0.00006565, 0.00004437, -0.00001778, 0.00001225, -0.00002691, -0.00005154, -0.00001000, 0.00003688, 0.00002889, -0.00005660, 0.00006325, -0.00005196, 0.00004098, -0.00006611, 0.00008764, -0.00008758, 0.00007476, -0.00009972, 0.00008571, -0.00013964, 0.00014198, -0.00019712, 0.00019726, -0.00086864, 0.00135784, 0.00003347, -0.00046707, 0.00019089, -0.00019834, 0.00007877, -0.00007996, 0.00006262, -0.00007691, 0.00006141, -0.00007133, 0.00003215, -0.00002416, 0.00001868, -0.00000140, -0.00002195, -0.00000396, -0.00004757, 0.00006463, -0.00009483, 0.00013799, 0.00001204, -0.00002081, 0.00001498, },
		{ -0.00006875, 0.00003916, -0.00002839, 0.00002262, -0.00001798, 0.00001645, -0.00004183, 0.00002213, 0.00000036, -0.00000179, -0.00000178, 0.00002041, -0.00002479, -0.00000460, 0.00002752, -0.00001863, 0.00002320, -0.00006207, 0.00006262, -0.00004770, 0.00004739, -0.00008809, 0.00000187, -0.00003724, 0.00009132, -0.00005521, 0.00002789, -0.00000939, 0.00001403, -0.00003554, 0.00002358, -0.00000160, -0.00000165, -0.00002061, 0.00000609, -0.00004089, 0.00001089, -0.00003260, 0.00002653, -0.00020501, -0.00039333, 0.00149159, -0.00062822, 0.00000447, -0.00006640, -0.00002641, -0.00003850, 0.00002827, -0.00003452, 0.00000846, -0.00000892, -0.00002201, 0.00000035, 0.00000189, 0.00000330, 0.00000286, -0.00002915, -0.00000624, -0.00003587, 0.00004786, -0.00008254, 0.00020060, -0.00008491, 0.00005997, },
		{ 0.00006628, -0.00008972, 0.00007166, -0.00005990, 0.00005240, -0.00004572, 0.00003373, -0.00006053, 0.00005524, -0.00003348, 0.00002327, -0.00001892, 0.00003681, -0.00005393, 0.00003947, -0.00001595, 0.00002315, -0.00002819, -0.00001468, 0.00003043, -0.00002476, 0.00002516, -0.00009901, 0.00002504, -0.00003207, 0.00009772, -0.00009343, 0.00008122, -0.00006094, 0.00005776, -0.00008588, 0.00009032, -0.00007362, 0.00006681, -0.00009555, 0.00008810, -0.00014080, 0.00013669, -0.00018170, 0.00021135, -0.00063247, 0.00041112, 0.00100122, -0.00077504, 0.00023896, -0.00021907, 0.00008162, -0.00010195, 0.00008475, -0.00008812, 0.00006039, -0.00005720, 0.00001316, -0.00001846, 0.00001510, 0.00000012, -0.00001001, -0.00001462, -0.00003322, 0.00000621, -0.00001373, 0.00001373, 0.00015653, -0.00009235, },
		{ -0.00002849, 0.00000126, -0.00003643, 0.00003705, -0.00003389, 0.00003190, -0.00003158, 0.00001348, -0.00003788, 0.00004746, -0.00003500, 0.00002280, -0.00001108, 0.00002344, -0.00004770, 0.00004997, -0.00003308, 0.00004405, -0.00006873, 0.00003730, -0.00001915, 0.00002583, -0.00004242, -0.00005371, 0.00000452, 0.00001438, 0.00004466, -0.00006126, 0.00006799, -0.00005317, 0.00004089, -0.00006742, 0.00008452, -0.00007407, 0.00005969, -0.00008589, 0.00007259, -0.00012532, 0.00012904, -0.00016940, 0.00015600, -0.00074498, 0.00109819, 0.00022539, -0.00050239, 0.00018532, -0.00019486, 0.00007018, -0.00007242, 0.00005793, -0.00006846, 0.00005026, -0.00005786, 0.00001518, -0.00001281, 0.00001072, 0.00000582, -0.00002813, 0.00000291, -0.00005953, 0.00004325, -0.00006786, 0.00013514, 0.00004635, },
		{ -0.00004687, 0.00003344, -0.00006964, 0.00003769, -0.00002528, 0.00002035, -0.00001693, 0.00000712, -0.00002716, 0.00000972, 0.00000979, -0.00001008, 0.00000205, 0.00001466, -0.00001177, -0.00001050, 0.00002304, -0.00001071, 0.00002104, -0.00006490, 0.00005242, -0.00003919, 0.00005176, -0.00010300, 0.00000756, -0.00002973, 0.00006849, -0.00003026, 0.00000998, 0.00000992, -0.00000384, -0.00001752, 0.00000103, 0.00002174, -0.00001713, -0.00000718, -0.00000963, -0.00002081, -0.00001308, 0.00000394, -0.00001258, -0.00014864, -0.00039591, 0.00130561, -0.00040184, -0.00011173, -0.00000611, -0.00007271, -0.00000581, 0.00000205, -0.00000932, -0.00001079, 0.00000475, -0.00003022, -0.00000001, 0.00000204, 0.00000357, 0.00000387, -0.00003075, -0.00000171, -0.00005587, 0.00003875, -0.00005724, 0.00020058, },
		{ 0.00017230, -0.00006869, 0.00003501, -0.00008265, 0.00006811, -0.00005698, 0.00005137, -0.00004941, 0.00003202, -0.00005173, 0.00004569, -0.00002515, 0.00001457, -0.00001499, 0.00003304, -0.00004025, 0.00002739, -0.00001343, 0.00002810, -0.00003043, -0.00002096, 0.00002368, -0.00001214, 0.00002112, -0.00011015, 0.00003729, -0.00003684, 0.00008561, -0.00007262, 0.00006307, -0.00003980, 0.00003624, -0.00006343, 0.00006137, -0.00003882, 0.00003597, -0.00006647, 0.00005669, -0.00010315, 0.00009295, -0.00011799, 0.00012992, -0.00049334, 0.00025418, 0.00097867, -0.00062794, 0.00013161, -0.00015173, 0.00002726, -0.00006009, 0.00005042, -0.00005442, 0.00003100, -0.00003241, -0.00000690, -0.00000859, 0.00000843, 0.00000495, -0.00001191, -0.00001413, -0.00003097, -0.00002175, -0.00000486, 0.00002761, },
		{ 0.00013661, 0.00008208, -0.00003274, -0.00001836, -0.00003122, 0.00003642, -0.00003456, 0.00003398, -0.00004072, 0.00002119, -0.00003884, 0.00004574, -0.00003288, 0.00001814, -0.00001036, 0.00002539, -0.00003883, 0.00003658, -0.00002507, 0.00004492, -0.00007025, 0.00002627, -0.00001700, 0.00003450, -0.00004939, -0.00006128, 0.00002149, -0.00000455, 0.00005115, -0.00005688, 0.00006369, -0.00004540, 0.00003227, -0.00005887, 0.00007028, -0.00005139, 0.00003791, -0.00006634, 0.00005322, -0.00010189, 0.00010381, -0.00012762, 0.00009814, -0.00060492, 0.00085107, 0.00036394, -0.00048637, 0.00014870, -0.00017019, 0.00004727, -0.00005423, 0.00004344, -0.00005096, 0.00003250, -0.00004058, -0.00000172, -0.00000371, 0.00000495, 0.00000965, -0.00002883, 0.00000275, -0.00006066, 0.00001194, -0.00004285, },
		{ -0.00002705, 0.00020083, -0.00000277, 0.00000335, -0.00007067, 0.00003511, -0.00002037, 0.00001489, -0.00001206, -0.00000519, -0.00001249, -0.00000015, 0.00001657, -0.00001592, 0.00000353, 0.00001082, -0.00000238, -0.00001083, 0.00001446, -0.00000185, 0.00002014, -0.00006868, 0.00004101, -0.00002953, 0.00005346, -0.00010941, 0.00000443, -0.00001379, 0.00004153, -0.00000764, -0.00000385, 0.00002365, -0.00001498, -0.00000647, -0.00001256, 0.00003307, -0.00002046, -0.00000446, -0.00001566, -0.00001036, -0.00002467, 0.00002328, -0.00002943, -0.00012309, -0.00035416, 0.00108930, -0.00019000, -0.00019611, 0.00002926, -0.00009880, 0.00001149, -0.00001219, 0.00000459, -0.00001843, 0.00000785, -0.00002949, -0.00000577, 0.00000462, 0.00000199, 0.00000545, -0.00003046, -0.00000144, -0.00006504, 0.00001449, },
		{ -0.00001135, 0.00004996, 0.00018721, -0.00003843, -0.00000314, -0.00007204, 0.00006087, -0.00005007, 0.00004473, -0.00004577, 0.00002232, -0.00003690, 0.00003286, -0.00001448, 0.00000394, -0.00000941, 0.00002701, -0.00002596, 0.00001776, -0.00001367, 0.00003358, -0.00003045, -0.00003055, 0.00001907, -0.00000233, 0.00001956, -0.00011720, 0.00004213, -0.00003126, 0.00006401, -0.00004726, 0.00004224, -0.00001769, 0.00001527, -0.00004167, 0.00003475, -0.00001014, 0.00001300, -0.00004494, 0.00003190, -0.00007095, 0.00005566, -0.00006463, 0.00006396, -0.00038023, 0.00014861, 0.00089695, -0.00045639, 0.00002590, -0.00009005, -0.00001968, -0.00002489, 0.00002117, -0.00002608, 0.00000877, -0.00001577, -0.00001818, -0.00000578, 0.00000648, 0.00000526, -0.00000958, -0.00001643, -0.00002724, -0.00004381, },
		{ -0.00002150, -0.00002904, 0.00014708, 0.00011815, -0.00003369, -0.00004344, -0.00002284, 0.00003296, -0.00003271, 0.00003223, -0.00004371, 0.00002053, -0.00003172, 0.00003797, -0.00002573, 0.00000909, -0.00000573, 0.00002247, -0.00002607, 0.00002283, -0.00001867, 0.00004670, -0.00007028, 0.00001186, -0.00001159, 0.00003811, -0.00004908, -0.00007110, 0.00003631, -0.00001599, 0.00004744, -0.00004457, 0.00005206, -0.00003107, 0.00001811, -0.00004412, 0.00005004, -0.00002587, 0.00001568, -0.00004678, 0.00003306, -0.00007536, 0.00007372, -0.00008159, 0.00003734, -0.00046788, 0.00063627, 0.00044190, -0.00042523, 0.00009054, -0.00013252, 0.00001706, -0.00003133, 0.00002445, -0.00002990, 0.00001381, -0.00002484, -0.00001434, 0.00000058, 0.00000299, 0.00000898, -0.00002384, -0.00000354, -0.00005411, },
		{ -0.00006266, -0.00002019, -0.00000135, 0.00020890, 0.00004292, -0.00002762, -0.00007337, 0.00003287, -0.00001522, 0.00000814, -0.00000624, -0.00001638, -0.00000206, -0.00000446, 0.00001858, -0.00001749, 0.00000116, 0.00001032, 0.00000216, -0.00000608, 0.00000393, 0.00000591, 0.00002127, -0.00007319, 0.00002856, -0.00001866, 0.00005155, -0.00010712, -0.00000574, 0.00000617, 0.00001615, 0.00000849, -0.00001110, 0.00002992, -0.00001814, -0.00000319, -0.00001674, 0.00003331, -0.00001427, -0.00000916, -0.00001507, -0.00000688, -0.00002656, 0.00002879, -0.00002916, -0.00011927, -0.00028610, 0.00086897, -0.00001397, -0.00023978, 0.00003762, -0.00010454, 0.00001406, -0.00001537, 0.00000824, -0.00001635, 0.00000342, -0.00002372, -0.00001296, 0.00000647, 0.00000124, 0.00000497, -0.00002588, -0.00000757, },
		{ -0.00002681, -0.00005446, -0.00003439, 0.00007534, 0.00020790, -0.00000669, -0.00004445, -0.00006023, 0.00005161, -0.00004056, 0.00003378, -0.00003681, 0.00000813, -0.00002053, 0.00002081, -0.00000487, -0.00000540, -0.00000514, 0.00002185, -0.00001448, 0.00001271, -0.00001666, 0.00003861, -0.00002782, -0.00004281, 0.00001621, 0.00000516, 0.00001880, -0.00011826, 0.00003887, -0.00001749, 0.00003789, -0.00002248, 0.00002312, 0.00000111, -0.00000103, -0.00002478, 0.00001496, 0.00000857, 0.00000027, -0.00003236, 0.00001511, -0.00004630, 0.00002769, -0.00002516, 0.00001727, -0.00029557, 0.00008854, 0.00077560, -0.00028384, -0.00006180, -0.00004398, -0.00005239, -0.00000131, 0.00000098, -0.00000661, -0.00000383, -0.00000819, -0.00002146, -0.00000803, 0.00000707, 0.00000339, -0.00000603, -0.00001782, },
		{ -0.00001268, -0.00004543, -0.00004813, -0.00003393, 0.00016748, 0.00015698, -0.00003409, -0.00007154, -0.00001270, 0.00002732, -0.00002841, 0.00002618, -0.00004015, 0.00001243, -0.00001899, 0.00002691, -0.00001610, -0.00000172, 0.00000023, 0.00001762, -0.00001339, 0.00001232, -0.00001556, 0.00004943, -0.00006865, -0.00000518, -0.00000357, 0.00003723, -0.00004340, -0.00008008, 0.00004571, -0.00001863, 0.00003547, -0.00002766, 0.00003663, -0.00001417, 0.00000271, -0.00002799, 0.00002970, -0.00000380, -0.00000150, -0.00003185, 0.00001657, -0.00005095, 0.00004529, -0.00003971, -0.00001538, -0.00034790, 0.00046431, 0.00046277, -0.00033303, 0.00002451, -0.00009211, -0.00001241, -0.00001019, 0.00000656, -0.00001071, -0.00000082, -0.00001444, -0.00002056, -0.00000039, 0.00000448, 0.00000474, -0.00001517, },
		{ -0.00001677, -0.00001925, -0.00005201, -0.00005654, 0.00000873, 0.00023146, 0.00008753, -0.00005842, -0.00007798, 0.00003143, -0.00001080, 0.00000167, -0.00000191, -0.00002344, 0.00000180, -0.00000239, 0.00001580, -0.00001491, -0.00000460, 0.00001264, 0.00000285, 0.00000115, -0.00000516, 0.00001029, 0.00002504, -0.00007816, 0.00001548, -0.00000699, 0.00004610, -0.00009736, -0.00001992, 0.00002550, -0.00000327, 0.00001628, -0.00001166, 0.00002924, -0.00001449, -0.00000602, -0.00001383, 0.00002625, -0.00000356, -0.00001628, -0.00001227, -0.00000640, -0.00002312, 0.00002597, -0.00001934, -0.00012584, -0.00021013, 0.00066672, 0.00011335, -0.00024240, 0.00002325, -0.00009435, 0.00000604, -0.00001113, 0.00000508, -0.00000855, -0.00000404, -0.00001732, -0.00001799, 0.00000538, 0.00000276, 0.00000134, },
		{ -0.00000411, -0.00001546, -0.00003256, -0.00005201, -0.00006972, 0.00009376, 0.00024315, 0.00002076, -0.00008503, -0.00004961, 0.00004224, -0.00003033, 0.00002076, -0.00002565, -0.00000625, -0.00000714, 0.00001295, 0.00000107, -0.00001112, -0.00000413, 0.00001943, -0.00000749, 0.00001216, -0.00002055, 0.00004131, -0.00002192, -0.00005713, 0.00001483, 0.00001056, 0.00001790, -0.00011301, 0.00002893, 0.00000048, 0.00001278, -0.00000274, 0.00000891, 0.00001382, -0.00001037, -0.00001481, 0.00000382, 0.00001672, -0.00000322, -0.00002714, 0.00000505, -0.00002861, 0.00000882, 0.00000077, -0.00001140, -0.00023545, 0.00006165, 0.00063725, -0.00013163, -0.00011991, -0.00001871, -0.00006808, 0.00000913, -0.00000917, 0.00000342, -0.00000727, -0.00000783, -0.00001965, -0.00001208, 0.00000744, 0.00000191, },
		{ -0.00000064, -0.00000622, -0.00002025, -0.00004064, -0.00006057, -0.00006059, 0.00019347, 0.00020490, -0.00003921, -0.00009854, -0.00000350, 0.00002132, -0.00002297, 0.00001705, -0.00003190, 0.00000015, -0.00000491, 0.00001648, -0.00000741, -0.00001101, 0.00000449, 0.00001407, -0.00000437, 0.00000736, -0.00001588, 0.00005215, -0.00006484, -0.00002404, 0.00000622, 0.00003268, -0.00003424, -0.00008604, 0.00004848, -0.00001385, 0.00001934, -0.00001070, 0.00002154, 0.00000107, -0.00000971, -0.00001494, 0.00001426, 0.00001031, -0.00001047, -0.00002379, 0.00000603, -0.00003179, 0.00002274, -0.00000736, -0.00005347, -0.00025205, 0.00033629, 0.00043882, -0.00022837, -0.00003493, -0.00005851, -0.00003425, 0.00000402, -0.00000592, 0.00000274, -0.00000856, -0.00001069, -0.00002076, -0.00000492, 0.00000738, },
		{ 0.00000616, -0.00000446, -0.00000528, -0.00003283, -0.00003951, -0.00008383, -0.00000673, 0.00027158, 0.00013191, -0.00009075, -0.00008265, 0.00002971, -0.00000681, -0.00000409, -0.00000005, -0.00002544, -0.00000062, 0.00000427, 0.00001037, -0.00001027, -0.00001153, 0.00001555, 0.00000238, 0.00000708, -0.00000935, 0.00000967, 0.00003130, -0.00008303, 0.00000226, 0.00000474, 0.00003784, -0.00008217, -0.00003489, 0.00004058, -0.00001470, 0.00001656, -0.00000773, 0.00002409, -0.00000706, -0.00001154, -0.00000786, 0.00001709, 0.00000597, -0.00002085, -0.00001134, -0.00000512, -0.00001882, 0.00002056, -0.00000761, -0.00013228, -0.00014125, 0.00049716, 0.00018889, -0.00021169, -0.00000459, -0.00007571, -0.00000646, -0.00000445, -0.00000039, 0.00000037, -0.00001019, -0.00001370, -0.00001895, 0.00000102, },
		{ 0.00000553, 0.00000235, -0.00000517, -0.00000863, -0.00004422, -0.00003938, -0.00010804, 0.00009249, 0.00030151, 0.00003953, -0.00012236, -0.00004147, 0.00003389, -0.00002086, 0.00000779, -0.00001527, -0.00001726, 0.00000033, 0.00001079, 0.00000261, -0.00001273, -0.00000649, 0.00001975, -0.00000441, 0.00001387, -0.00002225, 0.00003953, -0.00001237, -0.00007270, 0.00001466, 0.00001387, 0.00001667, -0.00010244, 0.00001508, 0.00001813, -0.00000679, 0.00000946, 0.00000068, 0.00001990, -0.00001279, -0.00001131, 0.00000031, 0.00001679, -0.00000115, -0.00002556, -0.00000129, -0.00001551, -0.00000342, 0.00001638, -0.00002678, -0.00019208, 0.00005299, 0.00050294, -0.00001502, -0.00014395, -0.00001377, -0.00006832, 0.00000835, -0.00001110, 0.00000596, -0.00000439, -0.00001111, -0.00001650, -0.00001474, },
		{ -0.00000992, 0.00000888, -0.00000430, -0.00000045, -0.00002239, -0.00004410, -0.00005537, -0.00010605, 0.00021518, 0.00027123, -0.00005553, -0.00011969, 0.00000161, 0.00001727, -0.00001849, 0.00000724, -0.00002226, -0.00001187, 0.00000584, 0.00001041, -0.00000263, -0.00001608, 0.00000473, 0.00001407, -0.00000099, 0.00000812, -0.00001795, 0.00005297, -0.00005816, -0.00004389, 0.00001695, 0.00002528, -0.00002305, -0.00008815, 0.00004546, -0.00000498, 0.00000400, 0.00000209, 0.00001014, 0.00001153, -0.00001631, -0.00000775, 0.00000639, 0.00001492, -0.00001120, -0.00002207, 0.00000128, -0.00001863, 0.00000747, 0.00001375, -0.00007536, -0.00018069, 0.00024565, 0.00038739, -0.00013010, -0.00007601, -0.00003814, -0.00004452, 0.00000879, -0.00001112, 0.00000905, -0.00000921, -0.00001232, -0.00001746, },
		{ -0.00001624, -0.00000490, 0.00000928, -0.00000974, 0.00000467, -0.00004299, -0.00003277, -0.00009230, -0.00005335, 0.00032664, 0.00018145, -0.00012925, -0.00008349, 0.00002502, -0.00000157, -0.00001016, 0.00000022, -0.00002389, -0.00000644, 0.00001161, 0.00000594, -0.00000681, -0.00001641, 0.00001603, 0.00000398, 0.00000809, -0.00000634, 0.00000398, 0.00003884, -0.00008678, -0.00001071, 0.00001574, 0.00002778, -0.00006373, -0.00004809, 0.00004966, -0.00001873, 0.00001245, -0.00000293, 0.00001797, 0.00000039, -0.00001589, -0.00000309, 0.00001077, 0.00000979, -0.00001958, -0.00001467, -0.00000067, -0.00001687, 0.00001696, 0.00000030, -0.00013125, -0.00008873, 0.00036606, 0.00021869, -0.00016105, -0.00003435, -0.00005680, -0.00001736, 0.00000005, -0.00000424, 0.00000674, -0.00001229, -0.00001412, },
		},
		{ { -0.00001601, -0.00001052, -0.00000101, -0.00001304, -0.00002928, -0.00002698, -0.00012865, 0.00005406, 0.00036947, 0.00005426, -0.00015355, -0.00001278, -0.00001368, -0.00000345, -0.00003340, -0.00000566, -0.00000970, -0.00000779, 0.00005244, -0.00001006, -0.00001588, -0.00001497, 0.00001434, 0.00000425, -0.00004299, 0.00000445, -0.00001055, -0.00000076, -0.00000029, -0.00000046, 0.00002005, -0.00000293, -0.00000338, -0.00000293, 0.00002005, -0.00000046, -0.00000029, -0.00000076, -0.00001055, 0.00000445, -0.00004299, 0.00000425, 0.00001434, -0.00001497, -0.00001588, -0.00001006, 0.00005244, -0.00000779, -0.00000970, -0.00000566, -0.00003340, -0.00000345, -0.00001368, -0.00001278, -0.00015355, 0.00005426, 0.00036947, 0.00005406, -0.00012865, -0.00002698, -0.00002928, -0.00001304, -0.00000101, -0.00001052, },
		{ -0.00001098, -0.00001687, -0.00000762, 0.00000001, -0.00001992, -0.00002872, -0.00003914, -0.00014858, 0.00016644, 0.00037994, -0.00006975, -0.00012206, -0.00000138, -0.00001427, -0.00000828, -0.00003455, 0.00000322, -0.00001949, 0.00001049, 0.00004672, -0.00002224, -0.00001075, -0.00001478, 0.00002226, -0.00001241, -0.00003679, 0.00001076, -0.00002002, 0.00000979, -0.00001090, 0.00001338, 0.00001398, -0.00000508, -0.00000484, 0.00000526, 0.00001571, -0.00000543, 0.00000535, -0.00000873, 0.00000039, -0.00000919, -0.00003892, 0.00001513, 0.00000667, -0.00001578, -0.00001870, 0.00000482, 0.00005028, -0.00002256, 0.00000220, -0.00002021, -0.00002332, -0.00000643, -0.00000590, -0.00004483, -0.00014218, 0.00016603, 0.00030792, -0.00002055, -0.00010399, -0.00002080, -0.00002777, -0.00000730, -0.00000421, },
		{ -0.00000741, -0.00001057, -0.00001842, -0.00000264, -0.00000211, -0.00002693, -0.00002554, -0.00006200, -0.00014957, 0.00030798, 0.00031614, -0.00016717, -0.00006802, -0.00000829, -0.00000700, -0.00002083, -0.00002537, 0.00000275, -0.00002208, 0.00002719, 0.00003519, -0.00003068, -0.00000404, -0.00001538, 0.00002941, -0.00003323, -0.00001903, 0.00000366, -0.00001882, 0.00001313, -0.00001690, 0.00002783, 0.00000207, -0.00000287, -0.00000683, 0.00001545, 0.00000351, -0.00000031, 0.00000190, -0.00000772, 0.00000418, -0.00002160, -0.00002976, 0.00002179, -0.00000169, -0.00001536, -0.00002024, 0.00002222, 0.00003818, -0.00002800, 0.00000864, -0.00003253, -0.00001135, -0.00001084, -0.00000042, -0.00008477, -0.00008782, 0.00023862, 0.00022442, -0.00006207, -0.00008133, -0.00001759, -0.00002461, -0.00000399, },
		{ -0.00000362, -0.00000861, -0.00001167, -0.00001804, 0.00000188, -0.00000625, -0.00003425, -0.00001891, -0.00010114, -0.00010961, 0.00044968, 0.00017663, -0.00020468, -0.00002162, -0.00001864, -0.00000074, -0.00003308, -0.00001205, -0.00000379, -0.00001789, 0.00004074, 0.00001783, -0.00003187, 0.00000007, -0.00001330, 0.00003097, -0.00005118, 0.00000403, -0.00001472, -0.00000447, 0.00000361, -0.00001017, 0.00003322, -0.00000770, -0.00000124, -0.00000448, 0.00002018, -0.00000818, 0.00000863, -0.00000668, 0.00000081, -0.00000142, -0.00002778, -0.00001962, 0.00002522, -0.00001038, -0.00001348, -0.00002004, 0.00003970, 0.00001775, -0.00002235, 0.00000518, -0.00003650, -0.00000377, -0.00001080, -0.00000728, -0.00011280, -0.00000795, 0.00026405, 0.00014210, -0.00008022, -0.00006266, -0.00001561, -0.00002054, },
		{ -0.00001643, -0.00000566, -0.00000705, -0.00001588, -0.00001315, 0.00000256, -0.00001004, -0.00004240, -0.00000967, -0.00015833, -0.00000545, 0.00054383, -0.00000794, -0.00017340, -0.00000502, -0.00001719, -0.00000597, -0.00003494, -0.00000393, -0.00000933, -0.00001085, 0.00005161, -0.00000561, -0.00002232, -0.00000361, -0.00000369, 0.00002068, -0.00005634, 0.00002021, -0.00003350, 0.00001566, -0.00001505, 0.00000952, 0.00002522, -0.00000994, -0.00000351, 0.00000483, 0.00001404, -0.00001097, 0.00001256, -0.00001199, 0.00000923, -0.00001177, -0.00002833, -0.00000927, 0.00002431, -0.00001664, -0.00001265, -0.00001504, 0.00005154, -0.00000439, -0.00001006, -0.00000614, -0.00003205, -0.00000130, -0.00000696, -0.00002818, -0.00011440, 0.00007369, 0.00024955, 0.00007415, -0.00008460, -0.00004745, -0.00001423, },
		{ -0.00001387, -0.00001286, -0.00000869, -0.00000384, -0.00002251, -0.00000359, -0.00000214, -0.00001252, -0.00004963, -0.00000431, -0.00022326, 0.00017286, 0.00053943, -0.00018025, -0.00009688, -0.00001850, -0.00000105, -0.00002459, -0.00002306, -0.00000575, -0.00000944, -0.00000336, 0.00005824, -0.00003159, -0.00000323, -0.00001616, 0.00001451, -0.00000353, -0.00004257, 0.00001852, -0.00003912, 0.00002828, -0.00003110, 0.00003333, 0.00000787, -0.00000548, -0.00000802, 0.00001796, -0.00000162, -0.00000250, 0.00000690, -0.00000798, 0.00000990, -0.00001985, -0.00002679, 0.00000210, 0.00001773, -0.00001767, -0.00001568, -0.00000213, 0.00005176, -0.00001982, 0.00000081, -0.00001871, -0.00002396, 0.00000013, -0.00000618, -0.00005431, -0.00008792, 0.00013853, 0.00020907, 0.00002532, -0.00008271, -0.00003399, },
		{ -0.00002086, -0.00001516, -0.00001022, -0.00001062, -0.00000196, -0.00002784, 0.00000683, -0.00000949, -0.00001663, -0.00004966, -0.00001770, -0.00026652, 0.00040478, 0.00040817, -0.00027930, -0.00002045, -0.00004074, 0.00001689, -0.00004587, -0.00000502, -0.00001332, -0.00000631, 0.00000682, 0.00005494, -0.00005089, 0.00001690, -0.00003115, 0.00003483, -0.00003496, -0.00001368, -0.00000271, -0.00002459, 0.00002291, -0.00003207, 0.00004855, -0.00000930, -0.00000006, -0.00000917, 0.00002699, -0.00001865, 0.00001168, -0.00000511, 0.00000416, 0.00000092, -0.00002190, -0.00002529, 0.00001371, 0.00000691, -0.00001403, -0.00002184, 0.00001773, 0.00003884, -0.00002369, 0.00000429, -0.00002674, -0.00001661, 0.00000366, -0.00001485, -0.00007250, -0.00004287, 0.00017798, 0.00015671, -0.00000532, -0.00007906, },
		{ -0.00007438, -0.00000861, -0.00001768, -0.00000914, -0.00000925, -0.00000518, -0.00002632, 0.00001125, -0.00001317, -0.00002837, -0.00003387, -0.00006903, -0.00024199, 0.00063126, 0.00016807, -0.00027143, 0.00001407, -0.00004442, 0.00001746, -0.00005310, 0.00000538, -0.00001672, -0.00000641, 0.00002418, 0.00003504, -0.00005229, 0.00002546, -0.00003698, 0.00004501, -0.00005936, 0.00001612, -0.00003261, 0.00000455, -0.00000018, -0.00001334, 0.00004664, -0.00001775, 0.00000071, -0.00000215, 0.00002430, -0.00002694, 0.00002187, -0.00001477, 0.00001659, -0.00001267, -0.00001954, -0.00002256, 0.00002205, -0.00000323, -0.00001057, -0.00002568, 0.00003786, 0.00001814, -0.00001812, 0.00000047, -0.00002955, -0.00000972, 0.00000665, -0.00003189, -0.00007462, 0.00000812, 0.00019117, 0.00010440, -0.00002272, },
		{ -0.00003424, -0.00006604, -0.00000015, -0.00001953, -0.00001036, -0.00000351, -0.00001551, -0.00001431, 0.00000421, -0.00000799, -0.00005119, 0.00000042, -0.00016848, -0.00010465, 0.00076720, -0.00011175, -0.00017133, -0.00000648, -0.00001703, -0.00000885, -0.00003660, -0.00000113, -0.00000902, -0.00001334, 0.00004893, -0.00000208, -0.00003052, 0.00001388, -0.00002459, 0.00003464, -0.00006272, 0.00002944, -0.00005299, 0.00003228, -0.00002887, 0.00001811, 0.00002860, -0.00001540, -0.00000389, 0.00001240, 0.00000823, -0.00002110, 0.00002107, -0.00001431, 0.00002045, -0.00002299, -0.00001737, -0.00001589, 0.00002331, -0.00000741, -0.00001308, -0.00002025, 0.00004904, -0.00000078, -0.00001037, -0.00000557, -0.00003069, 0.00000066, 0.00000150, -0.00004764, -0.00006264, 0.00005660, 0.00018036, 0.00006180, },
		{ 0.00003500, -0.00004663, -0.00005075, 0.00000117, -0.00001868, -0.00001344, 0.00000463, -0.00003013, 0.00000559, -0.00001280, 0.00000363, -0.00007855, 0.00003964, -0.00029913, 0.00016005, 0.00073555, -0.00033397, -0.00003828, -0.00005813, 0.00002791, -0.00005145, -0.00000316, -0.00002071, 0.00000645, -0.00002198, 0.00007209, -0.00004572, 0.00000722, -0.00001480, 0.00000452, 0.00000374, -0.00004110, 0.00001677, -0.00004942, 0.00004162, -0.00004600, 0.00004736, 0.00000403, -0.00000723, -0.00000890, 0.00002742, -0.00001480, -0.00000417, 0.00000994, -0.00000282, 0.00001223, -0.00002517, -0.00001825, -0.00000490, 0.00001731, -0.00000459, -0.00002305, -0.00000289, 0.00004568, -0.00001048, -0.00000694, -0.00000913, -0.00003275, 0.00001625, -0.00001719, -0.00005076, -0.00004515, 0.00009944, 0.00014931, },
		{ 0.00010516, 0.00002385, -0.00006195, -0.00002858, -0.00000517, -0.00001529, -0.00001553, 0.00000945, -0.00004064, 0.00002310, -0.00002909, 0.00000918, -0.00009152, 0.00005043, -0.00040664, 0.00050986, 0.00050987, -0.00041706, 0.00005416, -0.00009516, 0.00005831, -0.00008383, 0.00002610, -0.00003747, 0.00001695, -0.00001997, 0.00007780, -0.00007617, 0.00004213, -0.00004517, 0.00003654, -0.00003439, -0.00000541, -0.00001605, -0.00002113, 0.00002561, -0.00003993, 0.00006044, -0.00001522, -0.00000086, -0.00000759, 0.00003342, -0.00003369, 0.00001438, -0.00000372, 0.00001281, -0.00000382, -0.00002061, -0.00002062, 0.00000660, 0.00000909, -0.00000007, -0.00003475, 0.00002105, 0.00003042, -0.00001076, -0.00000836, -0.00001078, -0.00003264, 0.00003106, -0.00004509, -0.00003791, -0.00002875, 0.00013414, },
		{ 0.00015433, 0.00006009, 0.00002060, -0.00007538, -0.00000519, -0.00001504, -0.00001249, -0.00001212, 0.00000391, -0.00003670, 0.00002475, -0.00002979, -0.00000786, -0.00006750, -0.00000830, -0.00040984, 0.00084211, 0.00014171, -0.00033815, 0.00005904, -0.00008085, 0.00004583, -0.00008090, 0.00002976, -0.00003475, 0.00000999, 0.00000301, 0.00005325, -0.00007511, 0.00005361, -0.00005784, 0.00005254, -0.00005957, 0.00002344, -0.00004973, 0.00001724, -0.00000756, -0.00001193, 0.00005242, -0.00002233, -0.00000061, 0.00000298, 0.00002474, -0.00003934, 0.00002490, -0.00001050, 0.00002257, -0.00001886, -0.00001498, -0.00002008, 0.00001290, 0.00000592, -0.00000241, -0.00003801, 0.00004034, 0.00001290, -0.00000829, -0.00000934, -0.00001596, -0.00002255, 0.00003241, -0.00006719, -0.00001817, -0.00001155, },
		{ 0.00001484, 0.00015100, 0.00002861, 0.00001324, -0.00007789, 0.00001028, -0.00002169, -0.00001403, -0.00000066, -0.00001492, -0.00001365, 0.00000344, -0.00000725, -0.00005418, 0.00000110, -0.00015808, -0.00023623, 0.00102461, -0.00024806, -0.00014929, -0.00001639, -0.00001381, -0.00001187, -0.00003804, 0.00000229, -0.00000882, -0.00001500, 0.00004436, 0.00000018, -0.00003839, 0.00003246, -0.00004213, 0.00004088, -0.00005736, 0.00002711, -0.00006283, 0.00004398, -0.00003904, 0.00002458, 0.00002948, -0.00001873, -0.00000448, 0.00001894, 0.00000359, -0.00003004, 0.00002381, -0.00000616, 0.00002022, -0.00002582, -0.00001316, -0.00001400, 0.00001143, 0.00001132, -0.00001656, -0.00002542, 0.00004507, 0.00000300, -0.00001057, -0.00000439, -0.00002891, 0.00000272, 0.00000962, -0.00006637, -0.00000746, },
		{ -0.00001598, 0.00005591, 0.00011972, 0.00002013, -0.00000698, -0.00006313, 0.00001127, -0.00002075, -0.00001999, 0.00001517, -0.00004060, 0.00002073, -0.00003284, 0.00002855, -0.00011388, 0.00008895, -0.00037104, 0.00013035, 0.00095392, -0.00051709, 0.00004609, -0.00011489, 0.00006814, -0.00008519, 0.00002239, -0.00003949, 0.00002565, -0.00004242, 0.00008515, -0.00006116, 0.00001815, -0.00001151, -0.00000455, 0.00000632, -0.00002954, 0.00000224, -0.00004549, 0.00004370, -0.00005076, 0.00005227, 0.00000399, -0.00001138, -0.00000626, 0.00003155, -0.00002098, -0.00001194, 0.00001535, 0.00000555, 0.00000737, -0.00002396, -0.00001513, -0.00000479, 0.00000606, 0.00002081, -0.00003840, 0.00000103, 0.00003396, 0.00000345, -0.00001880, 0.00000536, -0.00004570, 0.00003746, -0.00003498, -0.00003664, },
		{ 0.00000898, -0.00003795, 0.00010497, 0.00006820, 0.00003180, -0.00003834, -0.00003389, -0.00000067, -0.00001442, -0.00002453, 0.00002473, -0.00005757, 0.00004649, -0.00006105, 0.00004984, -0.00014862, 0.00013636, -0.00055896, 0.00061624, 0.00061421, -0.00056833, 0.00014677, -0.00016438, 0.00011087, -0.00012907, 0.00006335, -0.00006600, 0.00004388, -0.00004883, 0.00009894, -0.00010045, 0.00006530, -0.00005377, 0.00003376, -0.00003134, 0.00000542, -0.00003598, -0.00000646, 0.00001716, -0.00003629, 0.00006006, -0.00001355, -0.00000671, -0.00000088, 0.00003287, -0.00003855, 0.00000560, 0.00000772, 0.00001578, -0.00000811, -0.00001844, -0.00001640, 0.00000126, 0.00000536, 0.00002348, -0.00005543, 0.00002946, 0.00001601, 0.00000754, -0.00002586, 0.00001074, -0.00005329, 0.00006349, -0.00008202, },
		{ -0.00010334, 0.00004323, -0.00005251, 0.00014244, 0.00001661, 0.00004809, -0.00006796, -0.00000273, -0.00001575, -0.00001046, -0.00001891, 0.00001610, -0.00004931, 0.00004203, -0.00005559, 0.00002622, -0.00011717, 0.00007431, -0.00059783, 0.00106497, 0.00010248, -0.00040073, 0.00010857, -0.00012347, 0.00007983, -0.00011201, 0.00005722, -0.00005502, 0.00002854, -0.00001951, 0.00006870, -0.00009469, 0.00007703, -0.00006975, 0.00005019, -0.00004943, 0.00002341, -0.00006261, 0.00003171, -0.00001917, -0.00000411, 0.00004855, -0.00002011, -0.00000596, 0.00001115, 0.00002095, -0.00004347, 0.00001636, 0.00000670, 0.00001711, -0.00001804, -0.00001529, -0.00001268, -0.00000062, 0.00001598, 0.00000946, -0.00005425, 0.00004497, 0.00000447, 0.00000461, -0.00002180, 0.00000093, -0.00003735, 0.00005923, },
		{ 0.00001518, -0.00007952, 0.00004132, -0.00003704, 0.00014652, -0.00001199, 0.00004954, -0.00007951, 0.00001507, -0.00002243, -0.00001578, 0.00000094, -0.00001415, -0.00001115, 0.00000018, -0.00000786, -0.00005115, -0.00000744, -0.00013050, -0.00038594, 0.00129298, -0.00040092, -0.00011338, -0.00003187, -0.00000903, -0.00001467, -0.00003878, 0.00000564, -0.00000879, -0.00001556, 0.00003797, 0.00000088, -0.00004373, 0.00004679, -0.00005068, 0.00003587, -0.00003792, 0.00001082, -0.00005889, 0.00004679, -0.00004362, 0.00002852, 0.00002754, -0.00001974, -0.00000524, 0.00002378, 0.00000086, -0.00003719, 0.00002057, 0.00001143, 0.00000880, -0.00001968, -0.00001622, -0.00000469, -0.00000832, 0.00003586, -0.00002155, -0.00003038, 0.00003982, 0.00000698, -0.00001060, -0.00000323, -0.00002621, 0.00000521, },
		{ 0.00005851, -0.00005475, -0.00001559, -0.00000101, 0.00001537, 0.00010872, -0.00000586, 0.00002575, -0.00006491, 0.00001200, -0.00001636, -0.00002895, 0.00002771, -0.00005379, 0.00004106, -0.00006087, 0.00006008, -0.00015228, 0.00013871, -0.00043436, 0.00008947, 0.00117438, -0.00070925, 0.00014316, -0.00017956, 0.00011246, -0.00012041, 0.00004974, -0.00005920, 0.00004494, -0.00006120, 0.00009410, -0.00007420, 0.00002713, -0.00000758, -0.00000967, 0.00000229, -0.00000718, -0.00002525, -0.00002511, 0.00003108, -0.00004288, 0.00004677, 0.00000823, -0.00001821, 0.00000011, 0.00002963, -0.00001905, -0.00002563, 0.00002304, 0.00001554, -0.00000311, -0.00001688, -0.00001732, 0.00000118, -0.00001240, 0.00005287, -0.00005723, 0.00000692, 0.00001952, 0.00001916, -0.00003182, 0.00002077, -0.00005801, },
		{ -0.00007171, 0.00009246, -0.00011591, 0.00005706, -0.00006078, 0.00008702, 0.00004299, 0.00002645, -0.00001621, -0.00003160, -0.00000567, -0.00000476, -0.00003846, 0.00004356, -0.00007788, 0.00007526, -0.00010246, 0.00010102, -0.00021449, 0.00022984, -0.00070905, 0.00071567, 0.00071277, -0.00071758, 0.00024458, -0.00023793, 0.00016559, -0.00017406, 0.00010079, -0.00009432, 0.00006993, -0.00007494, 0.00011356, -0.00011785, 0.00008058, -0.00005454, 0.00002685, -0.00002605, 0.00001798, -0.00006039, 0.00001874, -0.00000373, -0.00001972, 0.00004661, -0.00000406, -0.00001762, 0.00001064, 0.00002533, -0.00003269, -0.00001382, 0.00002785, 0.00001264, -0.00001076, -0.00001598, -0.00001308, -0.00000223, -0.00000209, 0.00005130, -0.00007839, 0.00003892, -0.00000020, 0.00002708, -0.00004399, 0.00003242, },
		{ 0.00001747, -0.00004904, 0.00007996, -0.00013088, 0.00009645, -0.00009909, 0.00014292, -0.00001940, 0.00006002, -0.00005516, 0.00000082, -0.00002162, -0.00000138, -0.00003001, 0.00003080, -0.00006333, 0.00006148, -0.00008769, 0.00006948, -0.00017612, 0.00016690, -0.00078509, 0.00127804, 0.00005787, -0.00045539, 0.00015699, -0.00016649, 0.00011367, -0.00014100, 0.00008320, -0.00007423, 0.00004613, -0.00003995, 0.00007795, -0.00010602, 0.00008931, -0.00006867, 0.00003708, -0.00002888, 0.00001536, -0.00006859, 0.00004552, -0.00003445, 0.00001022, 0.00003505, -0.00001137, -0.00001501, 0.00002163, 0.00001359, -0.00003937, -0.00000220, 0.00003389, 0.00000193, -0.00001076, -0.00001978, -0.00000238, -0.00001585, 0.00002569, 0.00002271, -0.00007070, 0.00004892, -0.00000412, 0.00001760, -0.00003412, },
		{ -0.00000272, -0.00002215, 0.00000839, 0.00001640, -0.00008330, 0.00007544, -0.00008483, 0.00015137, -0.00004766, 0.00006902, -0.00006962, 0.00001344, -0.00002150, -0.00001503, 0.00000099, -0.00001327, -0.00000720, -0.00000521, -0.00000963, -0.00004313, -0.00002276, -0.00009278, -0.00053468, 0.00154424, -0.00055007, -0.00007253, -0.00004870, -0.00000471, -0.00001684, -0.00003827, 0.00000820, -0.00000894, -0.00001486, 0.00003002, -0.00000009, -0.00004476, 0.00005247, -0.00004673, 0.00001899, -0.00000535, -0.00001818, -0.00004030, 0.00003879, -0.00004180, 0.00002980, 0.00002290, -0.00001858, -0.00000612, 0.00002655, 0.00000047, -0.00004220, 0.00001237, 0.00003510, -0.00001061, -0.00000673, -0.00002461, 0.00000866, -0.00003086, 0.00006104, -0.00002605, -0.00003499, 0.00003353, 0.00001115, -0.00001045, },
		{ -0.00004299, 0.00003316, -0.00006518, 0.00007452, -0.00007189, 0.00000814, 0.00000083, -0.00001596, 0.00010474, -0.00003067, 0.00004395, -0.00005306, 0.00000158, -0.00000465, -0.00003984, 0.00004128, -0.00006848, 0.00006424, -0.00009345, 0.00009416, -0.00018912, 0.00018352, -0.00048536, 0.00004537, 0.00137371, -0.00088665, 0.00023690, -0.00024089, 0.00015253, -0.00015108, 0.00007470, -0.00007677, 0.00006111, -0.00007512, 0.00009624, -0.00008173, 0.00003232, -0.00000455, -0.00000862, -0.00000838, 0.00002402, -0.00006233, 0.00000975, 0.00000395, -0.00002286, 0.00003175, 0.00001597, -0.00002705, 0.00000927, 0.00002244, -0.00000977, -0.00004333, 0.00003192, 0.00002546, -0.00001649, -0.00000613, -0.00002308, 0.00001000, -0.00003256, 0.00008428, -0.00007413, 0.00001162, 0.00000559, 0.00003366, },
		{ 0.00004370, -0.00005869, 0.00004923, -0.00008189, 0.00011074, -0.00013848, 0.00009630, -0.00008732, 0.00007630, 0.00002614, 0.00001564, -0.00000162, -0.00001933, -0.00002120, 0.00001399, -0.00005629, 0.00006415, -0.00009939, 0.00010577, -0.00014773, 0.00015540, -0.00027974, 0.00031808, -0.00084093, 0.00079955, 0.00079677, -0.00084806, 0.00033316, -0.00030406, 0.00021269, -0.00021095, 0.00013234, -0.00011791, 0.00009079, -0.00009407, 0.00011804, -0.00012405, 0.00008410, -0.00004709, 0.00001740, -0.00001911, 0.00003106, -0.00008555, 0.00005083, -0.00003505, 0.00000789, 0.00002216, 0.00001145, -0.00003205, 0.00002518, 0.00001266, -0.00001805, -0.00004025, 0.00005274, 0.00000490, -0.00001149, -0.00001365, -0.00001087, -0.00000436, -0.00001018, 0.00007705, -0.00009754, 0.00004517, -0.00001391, },
		{ -0.00001003, 0.00002788, -0.00004377, 0.00003043, -0.00005352, 0.00008910, -0.00014280, 0.00013117, -0.00013751, 0.00014438, -0.00004337, 0.00005850, -0.00004018, 0.00000647, -0.00003267, 0.00001447, -0.00004433, 0.00004640, -0.00007711, 0.00008042, -0.00012141, 0.00011507, -0.00023473, 0.00025455, -0.00094845, 0.00145815, 0.00001659, -0.00049895, 0.00019852, -0.00020385, 0.00014136, -0.00016253, 0.00010333, -0.00008929, 0.00005992, -0.00005539, 0.00007873, -0.00010591, 0.00008672, -0.00005389, 0.00001534, -0.00000039, 0.00000087, -0.00006639, 0.00005601, -0.00005157, 0.00002933, 0.00001403, 0.00000206, -0.00002631, 0.00003281, 0.00000440, -0.00002892, -0.00002683, 0.00006535, -0.00001851, 0.00000036, -0.00002637, 0.00000792, -0.00002965, 0.00003378, 0.00003476, -0.00008353, 0.00004951, },
		{ 0.00002654, 0.00001542, -0.00001010, -0.00000334, -0.00001649, 0.00001233, 0.00001240, -0.00007594, 0.00008986, -0.00011747, 0.00015254, -0.00006951, 0.00006981, -0.00004946, 0.00000590, -0.00001899, -0.00001158, -0.00000050, -0.00001233, -0.00000244, -0.00001203, -0.00001210, -0.00003212, -0.00004084, -0.00005412, -0.00066174, 0.00174958, -0.00067435, -0.00003646, -0.00006238, -0.00000280, -0.00001803, -0.00003608, 0.00000934, -0.00000930, -0.00001294, 0.00002096, -0.00000262, -0.00004063, 0.00004705, -0.00002990, -0.00000691, 0.00003596, -0.00005486, -0.00000985, 0.00002057, -0.00003413, 0.00002869, 0.00001605, -0.00001578, -0.00000706, 0.00002720, 0.00000249, -0.00004502, 0.00000123, 0.00005992, -0.00003271, 0.00000893, -0.00003481, 0.00002154, -0.00005036, 0.00008146, -0.00002809, -0.00003909, },
		{ 0.00001243, -0.00000496, 0.00004433, -0.00004987, 0.00003954, -0.00006406, 0.00008201, -0.00008280, 0.00003043, -0.00000785, -0.00003427, 0.00009850, -0.00004542, 0.00004440, -0.00002840, -0.00001875, 0.00001320, -0.00005110, 0.00005413, -0.00008260, 0.00008677, -0.00012573, 0.00012563, -0.00021925, 0.00021817, -0.00052137, 0.00000676, 0.00152920, -0.00102596, 0.00031141, -0.00028797, 0.00018060, -0.00017185, 0.00009352, -0.00008958, 0.00007155, -0.00008195, 0.00009024, -0.00008202, 0.00003286, -0.00000396, -0.00000064, -0.00002361, 0.00005973, -0.00010238, 0.00005310, -0.00003372, 0.00000584, 0.00001022, 0.00002531, -0.00003642, 0.00001939, 0.00001203, 0.00000451, -0.00006191, 0.00004018, 0.00003374, -0.00002945, 0.00000516, -0.00002968, 0.00001793, -0.00004857, 0.00010785, -0.00008410, },
		{ -0.00010770, 0.00004490, -0.00002172, 0.00005414, -0.00006691, 0.00005761, -0.00008000, 0.00011388, -0.00014460, 0.00011961, -0.00010916, 0.00006901, 0.00001441, 0.00000567, 0.00000243, 0.00000200, -0.00004520, 0.00003951, -0.00007537, 0.00008360, -0.00011888, 0.00013321, -0.00018976, 0.00020415, -0.00033392, 0.00038791, -0.00093954, 0.00086015, 0.00085820, -0.00094435, 0.00039910, -0.00035204, 0.00024363, -0.00023318, 0.00015303, -0.00013318, 0.00010323, -0.00010346, 0.00011084, -0.00011734, 0.00007482, -0.00003337, 0.00000802, -0.00001127, 0.00004334, -0.00010681, 0.00008381, -0.00007178, 0.00004175, -0.00000867, 0.00002952, -0.00004728, 0.00003984, -0.00000190, 0.00000148, -0.00006822, 0.00007682, -0.00000462, -0.00001068, -0.00001171, -0.00000976, -0.00000588, -0.00001581, 0.00009495, },
		{ 0.00004338, -0.00008972, 0.00004493, -0.00001129, 0.00003341, -0.00004898, 0.00003765, -0.00004859, 0.00008404, -0.00013597, 0.00014184, -0.00015831, 0.00013864, -0.00005355, 0.00004718, -0.00002588, 0.00001399, -0.00004726, 0.00003503, -0.00005956, 0.00006056, -0.00008850, 0.00009564, -0.00015118, 0.00015503, -0.00028250, 0.00032219, -0.00106628, 0.00158438, -0.00001328, -0.00052884, 0.00022806, -0.00023022, 0.00015802, -0.00017252, 0.00011434, -0.00009797, 0.00006788, -0.00006402, 0.00007051, -0.00009383, 0.00006904, -0.00002837, -0.00001009, 0.00003151, -0.00001638, -0.00005643, 0.00006082, -0.00006779, 0.00004992, -0.00001075, 0.00001712, -0.00003767, 0.00004255, -0.00000420, -0.00001528, -0.00005209, 0.00009400, -0.00003837, 0.00001163, -0.00003237, 0.00001526, -0.00003916, 0.00003916, },
		{ 0.00009269, -0.00002626, -0.00004246, 0.00001933, 0.00001958, -0.00000955, -0.00000526, -0.00000930, 0.00001684, 0.00000348, -0.00005813, 0.00008372, -0.00012890, 0.00014177, -0.00007351, 0.00005406, -0.00002259, -0.00000583, -0.00001523, -0.00000562, -0.00000321, -0.00001141, 0.00000231, -0.00001937, -0.00001468, -0.00002058, -0.00005706, -0.00002379, -0.00074839, 0.00188410, -0.00075549, -0.00001358, -0.00006924, -0.00000472, -0.00001801, -0.00003204, 0.00000869, -0.00000984, -0.00001004, 0.00001146, -0.00000637, -0.00003164, 0.00003066, -0.00000327, -0.00003598, 0.00007911, -0.00009137, 0.00002639, -0.00000488, -0.00002232, 0.00002569, 0.00000778, -0.00001209, -0.00000798, 0.00002596, 0.00000666, -0.00004578, -0.00001039, 0.00008051, -0.00005159, 0.00002254, -0.00004294, 0.00002963, -0.00006180, },
		{ -0.00005577, 0.00011823, -0.00008382, 0.00000785, -0.00001025, 0.00004950, -0.00005109, 0.00003847, -0.00005352, 0.00007956, -0.00008553, 0.00004788, -0.00002322, -0.00003870, 0.00008448, -0.00004588, 0.00002854, 0.00000474, -0.00004511, 0.00003428, -0.00006043, 0.00006409, -0.00009364, 0.00010474, -0.00015243, 0.00014922, -0.00023797, 0.00023850, -0.00054083, -0.00001868, 0.00162225, -0.00110824, 0.00035400, -0.00031247, 0.00019117, -0.00017910, 0.00010365, -0.00009593, 0.00007479, -0.00008089, 0.00007645, -0.00007505, 0.00002903, -0.00000705, 0.00001331, -0.00003950, 0.00009420, -0.00013692, 0.00009597, -0.00007501, 0.00003751, -0.00001344, 0.00003369, -0.00004439, 0.00002816, 0.00000120, 0.00002050, -0.00007777, 0.00004581, 0.00003919, -0.00003886, 0.00001381, -0.00003442, 0.00002169, },
		{ -0.00000738, -0.00001685, 0.00010109, -0.00010582, 0.00003663, -0.00002185, 0.00005665, -0.00006722, 0.00005615, -0.00006523, 0.00010108, -0.00013314, 0.00012393, -0.00012070, 0.00006214, 0.00000520, 0.00000053, -0.00000380, 0.00002857, -0.00007310, 0.00006754, -0.00009198, 0.00009847, -0.00013281, 0.00015260, -0.00022135, 0.00023874, -0.00036755, 0.00042830, -0.00099302, 0.00089164, 0.00089094, -0.00099472, 0.00043242, -0.00037424, 0.00025282, -0.00023681, 0.00015991, -0.00013816, 0.00010568, -0.00010233, 0.00009292, -0.00009903, 0.00005479, -0.00001729, 0.00000160, -0.00000312, 0.00005382, -0.00011963, 0.00011044, -0.00010692, 0.00007515, -0.00003985, 0.00004580, -0.00006001, 0.00005130, -0.00001454, 0.00002118, -0.00009174, 0.00009441, -0.00001273, -0.00000908, -0.00001033, -0.00000963, },
		{ 0.00001748, -0.00004242, 0.00004107, 0.00004715, -0.00008770, 0.00003485, -0.00000727, 0.00003335, -0.00004903, 0.00003847, -0.00003442, 0.00006556, -0.00011165, 0.00012842, -0.00015756, 0.00012177, -0.00005109, 0.00003057, -0.00001444, 0.00002212, -0.00006250, 0.00005694, -0.00007272, 0.00007073, -0.00009534, 0.00010411, -0.00017169, 0.00018209, -0.00031031, 0.00035761, -0.00112224, 0.00164143, -0.00002586, -0.00054319, 0.00024195, -0.00024202, 0.00016075, -0.00016898, 0.00011467, -0.00009929, 0.00006915, -0.00006539, 0.00005469, -0.00007205, 0.00003967, 0.00000173, -0.00003265, 0.00006137, -0.00003133, -0.00004086, 0.00005829, -0.00008001, 0.00006796, -0.00003471, 0.00003032, -0.00004663, 0.00004867, -0.00000973, -0.00000200, -0.00007250, 0.00011316, -0.00005205, 0.00001949, -0.00003544, },
		{ -0.00004602, 0.00003012, -0.00006223, 0.00009233, -0.00002013, -0.00004481, 0.00001249, 0.00002325, -0.00000883, -0.00000834, -0.00000097, 0.00002150, -0.00000887, -0.00003287, 0.00006035, -0.00011868, 0.00011644, -0.00006064, 0.00002723, 0.00000592, -0.00001925, -0.00001079, 0.00000231, -0.00000660, -0.00001056, 0.00000624, -0.00002631, -0.00001681, -0.00001087, -0.00006735, -0.00000906, -0.00078140, 0.00193092, -0.00078140, -0.00000906, -0.00006735, -0.00001087, -0.00001681, -0.00002631, 0.00000624, -0.00001056, -0.00000660, 0.00000231, -0.00001079, -0.00001925, 0.00000592, 0.00002723, -0.00006064, 0.00011644, -0.00011868, 0.00006035, -0.00003287, -0.00000887, 0.00002150, -0.00000097, -0.00000834, -0.00000883, 0.00002325, 0.00001249, -0.00004481, -0.00002013, 0.00009233, -0.00006223, 0.00003012, },
		{ 0.00001748, -0.00003544, 0.00001949, -0.00005205, 0.00011316, -0.00007250, -0.00000200, -0.00000973, 0.00004867, -0.00004663, 0.00003032, -0.00003471, 0.00006796, -0.00008001, 0.00005829, -0.00004086, -0.00003133, 0.00006137, -0.00003265, 0.00000173, 0.00003967, -0.00007205, 0.00005469, -0.00006539, 0.00006915, -0.00009929, 0.00011467, -0.00016898, 0.00016075, -0.00024202, 0.00024195, -0.00054319, -0.00002586, 0.00164143, -0.00112224, 0.00035761, -0.00031031, 0.00018209, -0.00017169, 0.00010411, -0.00009534, 0.00007073, -0.00007272, 0.00005694, -0.00006250, 0.00002212, -0.00001444, 0.00003057, -0.00005109, 0.00012177, -0.00015756, 0.00012842, -0.00011165, 0.00006556, -0.00003442, 0.00003847, -0.00004903, 0.00003335, -0.00000727, 0.00003485, -0.00008770, 0.00004715, 0.00004107, -0.00004242, },
		{ -0.00000738, -0.00000963, -0.00001033, -0.00000908, -0.00001273, 0.00009441, -0.00009174, 0.00002118, -0.00001454, 0.00005130, -0.00006001, 0.00004580, -0.00003985, 0.00007515, -0.00010692, 0.00011044, -0.00011963, 0.00005382, -0.00000312, 0.00000160, -0.00001729, 0.00005479, -0.00009903, 0.00009292, -0.00010233, 0.00010568, -0.00013816, 0.00015991, -0.00023681, 0.00025282, -0.00037424, 0.00043242, -0.00099472, 0.00089094, 0.00089164, -0.00099302, 0.00042830, -0.00036755, 0.00023874, -0.00022135, 0.00015260, -0.00013281, 0.00009847, -0.00009198, 0.00006754, -0.00007310, 0.00002857, -0.00000380, 0.00000053, 0.00000520, 0.00006214, -0.00012070, 0.00012393, -0.00013314, 0.00010108, -0.00006523, 0.00005615, -0.00006722, 0.00005665, -0.00002185, 0.00003663, -0.00010582, 0.00010109, -0.00001685, },
		{ -0.00005577, 0.00002169, -0.00003442, 0.00001381, -0.00003886, 0.00003919, 0.00004581, -0.00007777, 0.00002050, 0.00000120, 0.00002816, -0.00004439, 0.00003369, -0.00001344, 0.00003751, -0.00007501, 0.00009597, -0.00013692, 0.00009420, -0.00003950, 0.00001331, -0.00000705, 0.00002903, -0.00007505, 0.00007645, -0.00008089, 0.00007479, -0.00009593, 0.00010365, -0.00017910, 0.00019117, -0.00031247, 0.00035400, -0.00110824, 0.00162225, -0.00001868, -0.00054083, 0.00023850, -0.00023797, 0.00014922, -0.00015243, 0.00010474, -0.00009364, 0.00006409, -0.00006043, 0.00003428, -0.00004511, 0.00000474, 0.00002854, -0.00004588, 0.00008448, -0.00003870, -0.00002322, 0.00004788, -0.00008553, 0.00007956, -0.00005352, 0.00003847, -0.00005109, 0.00004950, -0.00001025, 0.00000785, -0.00008382, 0.00011823, },
		{ 0.00009269, -0.00006180, 0.00002963, -0.00004294, 0.00002254, -0.00005159, 0.00008051, -0.00001039, -0.00004578, 0.00000666, 0.00002596, -0.00000798, -0.00001209, 0.00000778, 0.00002569, -0.00002232, -0.00000488, 0.00002639, -0.00009137, 0.00007911, -0.00003598, -0.00000327, 0.00003066, -0.00003164, -0.00000637, 0.00001146, -0.00001004, -0.00000984, 0.00000869, -0.00003204, -0.00001801, -0.00000472, -0.00006924, -0.00001358, -0.00075549, 0.00188410, -0.00074839, -0.00002379, -0.00005706, -0.00002058, -0.00001468, -0.00001937, 0.00000231, -0.00001141, -0.00000321, -0.00000562, -0.00001523, -0.00000583, -0.00002259, 0.00005406, -0.00007351, 0.00014177, -0.00012890, 0.00008372, -0.00005813, 0.00000348, 0.00001684, -0.00000930, -0.00000526, -0.00000955, 0.00001958, 0.00001933, -0.00004246, -0.00002626, },
		{ 0.00004338, 0.00003916, -0.00003916, 0.00001526, -0.00003237, 0.00001163, -0.00003837, 0.00009400, -0.00005209, -0.00001528, -0.00000420, 0.00004255, -0.00003767, 0.00001712, -0.00001075, 0.00004992, -0.00006779, 0.00006082, -0.00005643, -0.00001638, 0.00003151, -0.00001009, -0.00002837, 0.00006904, -0.00009383, 0.00007051, -0.00006402, 0.00006788, -0.00009797, 0.00011434, -0.00017252, 0.00015802, -0.00023022, 0.00022806, -0.00052884, -0.00001328, 0.00158438, -0.00106628, 0.00032219, -0.00028250, 0.00015503, -0.00015118, 0.00009564, -0.00008850, 0.00006056, -0.00005956, 0.00003503, -0.00004726, 0.00001399, -0.00002588, 0.00004718, -0.00005355, 0.00013864, -0.00015831, 0.00014184, -0.00013597, 0.00008404, -0.00004859, 0.00003765, -0.00004898, 0.00003341, -0.00001129, 0.00004493, -0.00008972, },
		{ -0.00010770, 0.00009495, -0.00001581, -0.00000588, -0.00000976, -0.00001171, -0.00001068, -0.00000462, 0.00007682, -0.00006822, 0.00000148, -0.00000190, 0.00003984, -0.00004728, 0.00002952, -0.00000867, 0.00004175, -0.00007178, 0.00008381, -0.00010681, 0.00004334, -0.00001127, 0.00000802, -0.00003337, 0.00007482, -0.00011734, 0.00011084, -0.00010346, 0.00010323, -0.00013318, 0.00015303, -0.00023318, 0.00024363, -0.00035204, 0.00039910, -0.00094435, 0.00085820, 0.00086015, -0.00093954, 0.00038791, -0.00033392, 0.00020415, -0.00018976, 0.00013321, -0.00011888, 0.00008360, -0.00007537, 0.00003951, -0.00004520, 0.00000200, 0.00000243, 0.00000567, 0.00001441, 0.00006901, -0.00010916, 0.00011961, -0.00014460, 0.00011388, -0.00008000, 0.00005761, -0.00006691, 0.00005414, -0.00002172, 0.00004490, },
		{ 0.00001243, -0.00008410, 0.00010785, -0.00004857, 0.00001793, -0.00002968, 0.00000516, -0.00002945, 0.00003374, 0.00004018, -0.00006191, 0.00000451, 0.00001203, 0.00001939, -0.00003642, 0.00002531, 0.00001022, 0.00000584, -0.00003372, 0.00005310, -0.00010238, 0.00005973, -0.00002361, -0.00000064, -0.00000396, 0.00003286, -0.00008202, 0.00009024, -0.00008195, 0.00007155, -0.00008958, 0.00009352, -0.00017185, 0.00018060, -0.00028797, 0.00031141, -0.00102596, 0.00152920, 0.00000676, -0.00052137, 0.00021817, -0.00021925, 0.00012563, -0.00012573, 0.00008677, -0.00008260, 0.00005413, -0.00005110, 0.00001320, -0.00001875, -0.00002840, 0.00004440, -0.00004542, 0.00009850, -0.00003427, -0.00000785, 0.00003043, -0.00008280, 0.00008201, -0.00006406, 0.00003954, -0.00004987, 0.00004433, -0.00000496, },
		{ 0.00002654, -0.00003909, -0.00002809, 0.00008146, -0.00005036, 0.00002154, -0.00003481, 0.00000893, -0.00003271, 0.00005992, 0.00000123, -0.00004502, 0.00000249, 0.00002720, -0.00000706, -0.00001578, 0.00001605, 0.00002869, -0.00003413, 0.00002057, -0.00000985, -0.00005486, 0.00003596, -0.00000691, -0.00002990, 0.00004705, -0.00004063, -0.00000262, 0.00002096, -0.00001294, -0.00000930, 0.00000934, -0.00003608, -0.00001803, -0.00000280, -0.00006238, -0.00003646, -0.00067435, 0.00174958, -0.00066174, -0.00005412, -0.00004084, -0.00003212, -0.00001210, -0.00001203, -0.00000244, -0.00001233, -0.00000050, -0.00001158, -0.00001899, 0.00000590, -0.00004946, 0.00006981, -0.00006951, 0.00015254, -0.00011747, 0.00008986, -0.00007594, 0.00001240, 0.00001233, -0.00001649, -0.00000334, -0.00001010, 0.00001542, },
		{ -0.00001003, 0.00004951, -0.00008353, 0.00003476, 0.00003378, -0.00002965, 0.00000792, -0.00002637, 0.00000036, -0.00001851, 0.00006535, -0.00002683, -0.00002892, 0.00000440, 0.00003281, -0.00002631, 0.00000206, 0.00001403, 0.00002933, -0.00005157, 0.00005601, -0.00006639, 0.00000087, -0.00000039, 0.00001534, -0.00005389, 0.00008672, -0.00010591, 0.00007873, -0.00005539, 0.00005992, -0.00008929, 0.00010333, -0.00016253, 0.00014136, -0.00020385, 0.00019852, -0.00049895, 0.00001659, 0.00145815, -0.00094845, 0.00025455, -0.00023473, 0.00011507, -0.00012141, 0.00008042, -0.00007711, 0.00004640, -0.00004433, 0.00001447, -0.00003267, 0.00000647, -0.00004018, 0.00005850, -0.00004337, 0.00014438, -0.00013751, 0.00013117, -0.00014280, 0.00008910, -0.00005352, 0.00003043, -0.00004377, 0.00002788, },
		{ 0.00004370, -0.00001391, 0.00004517, -0.00009754, 0.00007705, -0.00001018, -0.00000436, -0.00001087, -0.00001365, -0.00001149, 0.00000490, 0.00005274, -0.00004025, -0.00001805, 0.00001266, 0.00002518, -0.00003205, 0.00001145, 0.00002216, 0.00000789, -0.00003505, 0.00005083, -0.00008555, 0.00003106, -0.00001911, 0.00001740, -0.00004709, 0.00008410, -0.00012405, 0.00011804, -0.00009407, 0.00009079, -0.00011791, 0.00013234, -0.00021095, 0.00021269, -0.00030406, 0.00033316, -0.00084806, 0.00079677, 0.00079955, -0.00084093, 0.00031808, -0.00027974, 0.00015540, -0.00014773, 0.00010577, -0.00009939, 0.00006415, -0.00005629, 0.00001399, -0.00002120, -0.00001933, -0.00000162, 0.00001564, 0.00002614, 0.00007630, -0.00008732, 0.00009630, -0.00013848, 0.00011074, -0.00008189, 0.00004923, -0.00005869, },
		{ -0.00004299, 0.00003366, 0.00000559, 0.00001162, -0.00007413, 0.00008428, -0.00003256, 0.00001000, -0.00002308, -0.00000613, -0.00001649, 0.00002546, 0.00003192, -0.00004333, -0.00000977, 0.00002244, 0.00000927, -0.00002705, 0.00001597, 0.00003175, -0.00002286, 0.00000395, 0.00000975, -0.00006233, 0.00002402, -0.00000838, -0.00000862, -0.00000455, 0.00003232, -0.00008173, 0.00009624, -0.00007512, 0.00006111, -0.00007677, 0.00007470, -0.00015108, 0.00015253, -0.00024089, 0.00023690, -0.00088665, 0.00137371, 0.00004537, -0.00048536, 0.00018352, -0.00018912, 0.00009416, -0.00009345, 0.00006424, -0.00006848, 0.00004128, -0.00003984, -0.00000465, 0.00000158, -0.00005306, 0.00004395, -0.00003067, 0.00010474, -0.00001596, 0.00000083, 0.00000814, -0.00007189, 0.00007452, -0.00006518, 0.00003316, },
		{ -0.00000272, -0.00001045, 0.00001115, 0.00003353, -0.00003499, -0.00002605, 0.00006104, -0.00003086, 0.00000866, -0.00002461, -0.00000673, -0.00001061, 0.00003510, 0.00001237, -0.00004220, 0.00000047, 0.00002655, -0.00000612, -0.00001858, 0.00002290, 0.00002980, -0.00004180, 0.00003879, -0.00004030, -0.00001818, -0.00000535, 0.00001899, -0.00004673, 0.00005247, -0.00004476, -0.00000009, 0.00003002, -0.00001486, -0.00000894, 0.00000820, -0.00003827, -0.00001684, -0.00000471, -0.00004870, -0.00007253, -0.00055007, 0.00154424, -0.00053468, -0.00009278, -0.00002276, -0.00004313, -0.00000963, -0.00000521, -0.00000720, -0.00001327, 0.00000099, -0.00001503, -0.00002150, 0.00001344, -0.00006962, 0.00006902, -0.00004766, 0.00015137, -0.00008483, 0.00007544, -0.00008330, 0.00001640, 0.00000839, -0.00002215, },
		{ 0.00001747, -0.00003412, 0.00001760, -0.00000412, 0.00004892, -0.00007070, 0.00002271, 0.00002569, -0.00001585, -0.00000238, -0.00001978, -0.00001076, 0.00000193, 0.00003389, -0.00000220, -0.00003937, 0.00001359, 0.00002163, -0.00001501, -0.00001137, 0.00003505, 0.00001022, -0.00003445, 0.00004552, -0.00006859, 0.00001536, -0.00002888, 0.00003708, -0.00006867, 0.00008931, -0.00010602, 0.00007795, -0.00003995, 0.00004613, -0.00007423, 0.00008320, -0.00014100, 0.00011367, -0.00016649, 0.00015699, -0.00045539, 0.00005787, 0.00127804, -0.00078509, 0.00016690, -0.00017612, 0.00006948, -0.00008769, 0.00006148, -0.00006333, 0.00003080, -0.00003001, -0.00000138, -0.00002162, 0.00000082, -0.00005516, 0.00006002, -0.00001940, 0.00014292, -0.00009909, 0.00009645, -0.00013088, 0.00007996, -0.00004904, },
		{ -0.00007171, 0.00003242, -0.00004399, 0.00002708, -0.00000020, 0.00003892, -0.00007839, 0.00005130, -0.00000209, -0.00000223, -0.00001308, -0.00001598, -0.00001076, 0.00001264, 0.00002785, -0.00001382, -0.00003269, 0.00002533, 0.00001064, -0.00001762, -0.00000406, 0.00004661, -0.00001972, -0.00000373, 0.00001874, -0.00006039, 0.00001798, -0.00002605, 0.00002685, -0.00005454, 0.00008058, -0.00011785, 0.00011356, -0.00007494, 0.00006993, -0.00009432, 0.00010079, -0.00017406, 0.00016559, -0.00023793, 0.00024458, -0.00071758, 0.00071277, 0.00071567, -0.00070905, 0.00022984, -0.00021449, 0.00010102, -0.00010246, 0.00007526, -0.00007788, 0.00004356, -0.00003846, -0.00000476, -0.00000567, -0.00003160, -0.00001621, 0.00002645, 0.00004299, 0.00008702, -0.00006078, 0.00005706, -0.00011591, 0.00009246, },
		{ 0.00005851, -0.00005801, 0.00002077, -0.00003182, 0.00001916, 0.00001952, 0.00000692, -0.00005723, 0.00005287, -0.00001240, 0.00000118, -0.00001732, -0.00001688, -0.00000311, 0.00001554, 0.00002304, -0.00002563, -0.00001905, 0.00002963, 0.00000011, -0.00001821, 0.00000823, 0.00004677, -0.00004288, 0.00003108, -0.00002511, -0.00002525, -0.00000718, 0.00000229, -0.00000967, -0.00000758, 0.00002713, -0.00007420, 0.00009410, -0.00006120, 0.00004494, -0.00005920, 0.00004974, -0.00012041, 0.00011246, -0.00017956, 0.00014316, -0.00070925, 0.00117438, 0.00008947, -0.00043436, 0.00013871, -0.00015228, 0.00006008, -0.00006087, 0.00004106, -0.00005379, 0.00002771, -0.00002895, -0.00001636, 0.00001200, -0.00006491, 0.00002575, -0.00000586, 0.00010872, 0.00001537, -0.00000101, -0.00001559, -0.00005475, },
		{ 0.00001518, 0.00000521, -0.00002621, -0.00000323, -0.00001060, 0.00000698, 0.00003982, -0.00003038, -0.00002155, 0.00003586, -0.00000832, -0.00000469, -0.00001622, -0.00001968, 0.00000880, 0.00001143, 0.00002057, -0.00003719, 0.00000086, 0.00002378, -0.00000524, -0.00001974, 0.00002754, 0.00002852, -0.00004362, 0.00004679, -0.00005889, 0.00001082, -0.00003792, 0.00003587, -0.00005068, 0.00004679, -0.00004373, 0.00000088, 0.00003797, -0.00001556, -0.00000879, 0.00000564, -0.00003878, -0.00001467, -0.00000903, -0.00003187, -0.00011338, -0.00040092, 0.00129298, -0.00038594, -0.00013050, -0.00000744, -0.00005115, -0.00000786, 0.00000018, -0.00001115, -0.00001415, 0.00000094, -0.00001578, -0.00002243, 0.00001507, -0.00007951, 0.00004954, -0.00001199, 0.00014652, -0.00003704, 0.00004132, -0.00007952, },
		{ -0.00010334, 0.00005923, -0.00003735, 0.00000093, -0.00002180, 0.00000461, 0.00000447, 0.00004497, -0.00005425, 0.00000946, 0.00001598, -0.00000062, -0.00001268, -0.00001529, -0.00001804, 0.00001711, 0.00000670, 0.00001636, -0.00004347, 0.00002095, 0.00001115, -0.00000596, -0.00002011, 0.00004855, -0.00000411, -0.00001917, 0.00003171, -0.00006261, 0.00002341, -0.00004943, 0.00005019, -0.00006975, 0.00007703, -0.00009469, 0.00006870, -0.00001951, 0.00002854, -0.00005502, 0.00005722, -0.00011201, 0.00007983, -0.00012347, 0.00010857, -0.00040073, 0.00010248, 0.00106497, -0.00059783, 0.00007431, -0.00011717, 0.00002622, -0.00005559, 0.00004203, -0.00004931, 0.00001610, -0.00001891, -0.00001046, -0.00001575, -0.00000273, -0.00006796, 0.00004809, 0.00001661, 0.00014244, -0.00005251, 0.00004323, },
		{ 0.00000898, -0.00008202, 0.00006349, -0.00005329, 0.00001074, -0.00002586, 0.00000754, 0.00001601, 0.00002946, -0.00005543, 0.00002348, 0.00000536, 0.00000126, -0.00001640, -0.00001844, -0.00000811, 0.00001578, 0.00000772, 0.00000560, -0.00003855, 0.00003287, -0.00000088, -0.00000671, -0.00001355, 0.00006006, -0.00003629, 0.00001716, -0.00000646, -0.00003598, 0.00000542, -0.00003134, 0.00003376, -0.00005377, 0.00006530, -0.00010045, 0.00009894, -0.00004883, 0.00004388, -0.00006600, 0.00006335, -0.00012907, 0.00011087, -0.00016438, 0.00014677, -0.00056833, 0.00061421, 0.00061624, -0.00055896, 0.00013636, -0.00014862, 0.00004984, -0.00006105, 0.00004649, -0.00005757, 0.00002473, -0.00002453, -0.00001442, -0.00000067, -0.00003389, -0.00003834, 0.00003180, 0.00006820, 0.00010497, -0.00003795, },
		{ -0.00001598, -0.00003664, -0.00003498, 0.00003746, -0.00004570, 0.00000536, -0.00001880, 0.00000345, 0.00003396, 0.00000103, -0.00003840, 0.00002081, 0.00000606, -0.00000479, -0.00001513, -0.00002396, 0.00000737, 0.00000555, 0.00001535, -0.00001194, -0.00002098, 0.00003155, -0.00000626, -0.00001138, 0.00000399, 0.00005227, -0.00005076, 0.00004370, -0.00004549, 0.00000224, -0.00002954, 0.00000632, -0.00000455, -0.00001151, 0.00001815, -0.00006116, 0.00008515, -0.00004242, 0.00002565, -0.00003949, 0.00002239, -0.00008519, 0.00006814, -0.00011489, 0.00004609, -0.00051709, 0.00095392, 0.00013035, -0.00037104, 0.00008895, -0.00011388, 0.00002855, -0.00003284, 0.00002073, -0.00004060, 0.00001517, -0.00001999, -0.00002075, 0.00001127, -0.00006313, -0.00000698, 0.00002013, 0.00011972, 0.00005591, },
		{ 0.00001484, -0.00000746, -0.00006637, 0.00000962, 0.00000272, -0.00002891, -0.00000439, -0.00001057, 0.00000300, 0.00004507, -0.00002542, -0.00001656, 0.00001132, 0.00001143, -0.00001400, -0.00001316, -0.00002582, 0.00002022, -0.00000616, 0.00002381, -0.00003004, 0.00000359, 0.00001894, -0.00000448, -0.00001873, 0.00002948, 0.00002458, -0.00003904, 0.00004398, -0.00006283, 0.00002711, -0.00005736, 0.00004088, -0.00004213, 0.00003246, -0.00003839, 0.00000018, 0.00004436, -0.00001500, -0.00000882, 0.00000229, -0.00003804, -0.00001187, -0.00001381, -0.00001639, -0.00014929, -0.00024806, 0.00102461, -0.00023623, -0.00015808, 0.00000110, -0.00005418, -0.00000725, 0.00000344, -0.00001365, -0.00001492, -0.00000066, -0.00001403, -0.00002169, 0.00001028, -0.00007789, 0.00001324, 0.00002861, 0.00015100, },
		{ 0.00015433, -0.00001155, -0.00001817, -0.00006719, 0.00003241, -0.00002255, -0.00001596, -0.00000934, -0.00000829, 0.00001290, 0.00004034, -0.00003801, -0.00000241, 0.00000592, 0.00001290, -0.00002008, -0.00001498, -0.00001886, 0.00002257, -0.00001050, 0.00002490, -0.00003934, 0.00002474, 0.00000298, -0.00000061, -0.00002233, 0.00005242, -0.00001193, -0.00000756, 0.00001724, -0.00004973, 0.00002344, -0.00005957, 0.00005254, -0.00005784, 0.00005361, -0.00007511, 0.00005325, 0.00000301, 0.00000999, -0.00003475, 0.00002976, -0.00008090, 0.00004583, -0.00008085, 0.00005904, -0.00033815, 0.00014171, 0.00084211, -0.00040984, -0.00000830, -0.00006750, -0.00000786, -0.00002979, 0.00002475, -0.00003670, 0.00000391, -0.00001212, -0.00001249, -0.00001504, -0.00000519, -0.00007538, 0.00002060, 0.00006009, },
		{ 0.00010516, 0.00013414, -0.00002875, -0.00003791, -0.00004509, 0.00003106, -0.00003264, -0.00001078, -0.00000836, -0.00001076, 0.00003042, 0.00002105, -0.00003475, -0.00000007, 0.00000909, 0.00000660, -0.00002062, -0.00002061, -0.00000382, 0.00001281, -0.00000372, 0.00001438, -0.00003369, 0.00003342, -0.00000759, -0.00000086, -0.00001522, 0.00006044, -0.00003993, 0.00002561, -0.00002113, -0.00001605, -0.00000541, -0.00003439, 0.00003654, -0.00004517, 0.00004213, -0.00007617, 0.00007780, -0.00001997, 0.00001695, -0.00003747, 0.00002610, -0.00008383, 0.00005831, -0.00009516, 0.00005416, -0.00041706, 0.00050987, 0.00050986, -0.00040664, 0.00005043, -0.00009152, 0.00000918, -0.00002909, 0.00002310, -0.00004064, 0.00000945, -0.00001553, -0.00001529, -0.00000517, -0.00002858, -0.00006195, 0.00002385, },
		{ 0.00003500, 0.00014931, 0.00009944, -0.00004515, -0.00005076, -0.00001719, 0.00001625, -0.00003275, -0.00000913, -0.00000694, -0.00001048, 0.00004568, -0.00000289, -0.00002305, -0.00000459, 0.00001731, -0.00000490, -0.00001825, -0.00002517, 0.00001223, -0.00000282, 0.00000994, -0.00000417, -0.00001480, 0.00002742, -0.00000890, -0.00000723, 0.00000403, 0.00004736, -0.00004600, 0.00004162, -0.00004942, 0.00001677, -0.00004110, 0.00000374, 0.00000452, -0.00001480, 0.00000722, -0.00004572, 0.00007209, -0.00002198, 0.00000645, -0.00002071, -0.00000316, -0.00005145, 0.00002791, -0.00005813, -0.00003828, -0.00033397, 0.00073555, 0.00016005, -0.00029913, 0.00003964, -0.00007855, 0.00000363, -0.00001280, 0.00000559, -0.00003013, 0.00000463, -0.00001344, -0.00001868, 0.00000117, -0.00005075, -0.00004663, },
		{ -0.00003424, 0.00006180, 0.00018036, 0.00005660, -0.00006264, -0.00004764, 0.00000150, 0.00000066, -0.00003069, -0.00000557, -0.00001037, -0.00000078, 0.00004904, -0.00002025, -0.00001308, -0.00000741, 0.00002331, -0.00001589, -0.00001737, -0.00002299, 0.00002045, -0.00001431, 0.00002107, -0.00002110, 0.00000823, 0.00001240, -0.00000389, -0.00001540, 0.00002860, 0.00001811, -0.00002887, 0.00003228, -0.00005299, 0.00002944, -0.00006272, 0.00003464, -0.00002459, 0.00001388, -0.00003052, -0.00000208, 0.00004893, -0.00001334, -0.00000902, -0.00000113, -0.00003660, -0.00000885, -0.00001703, -0.00000648, -0.00017133, -0.00011175, 0.00076720, -0.00010465, -0.00016848, 0.00000042, -0.00005119, -0.00000799, 0.00000421, -0.00001431, -0.00001551, -0.00000351, -0.00001036, -0.00001953, -0.00000015, -0.00006604, },
		{ -0.00007438, -0.00002272, 0.00010440, 0.00019117, 0.00000812, -0.00007462, -0.00003189, 0.00000665, -0.00000972, -0.00002955, 0.00000047, -0.00001812, 0.00001814, 0.00003786, -0.00002568, -0.00001057, -0.00000323, 0.00002205, -0.00002256, -0.00001954, -0.00001267, 0.00001659, -0.00001477, 0.00002187, -0.00002694, 0.00002430, -0.00000215, 0.00000071, -0.00001775, 0.00004664, -0.00001334, -0.00000018, 0.00000455, -0.00003261, 0.00001612, -0.00005936, 0.00004501, -0.00003698, 0.00002546, -0.00005229, 0.00003504, 0.00002418, -0.00000641, -0.00001672, 0.00000538, -0.00005310, 0.00001746, -0.00004442, 0.00001407, -0.00027143, 0.00016807, 0.00063126, -0.00024199, -0.00006903, -0.00003387, -0.00002837, -0.00001317, 0.00001125, -0.00002632, -0.00000518, -0.00000925, -0.00000914, -0.00001768, -0.00000861, },
		{ -0.00002086, -0.00007906, -0.00000532, 0.00015671, 0.00017798, -0.00004287, -0.00007250, -0.00001485, 0.00000366, -0.00001661, -0.00002674, 0.00000429, -0.00002369, 0.00003884, 0.00001773, -0.00002184, -0.00001403, 0.00000691, 0.00001371, -0.00002529, -0.00002190, 0.00000092, 0.00000416, -0.00000511, 0.00001168, -0.00001865, 0.00002699, -0.00000917, -0.00000006, -0.00000930, 0.00004855, -0.00003207, 0.00002291, -0.00002459, -0.00000271, -0.00001368, -0.00003496, 0.00003483, -0.00003115, 0.00001690, -0.00005089, 0.00005494, 0.00000682, -0.00000631, -0.00001332, -0.00000502, -0.00004587, 0.00001689, -0.00004074, -0.00002045, -0.00027930, 0.00040817, 0.00040478, -0.00026652, -0.00001770, -0.00004966, -0.00001663, -0.00000949, 0.00000683, -0.00002784, -0.00000196, -0.00001062, -0.00001022, -0.00001516, },
		{ -0.00001387, -0.00003399, -0.00008271, 0.00002532, 0.00020907, 0.00013853, -0.00008792, -0.00005431, -0.00000618, 0.00000013, -0.00002396, -0.00001871, 0.00000081, -0.00001982, 0.00005176, -0.00000213, -0.00001568, -0.00001767, 0.00001773, 0.00000210, -0.00002679, -0.00001985, 0.00000990, -0.00000798, 0.00000690, -0.00000250, -0.00000162, 0.00001796, -0.00000802, -0.00000548, 0.00000787, 0.00003333, -0.00003110, 0.00002828, -0.00003912, 0.00001852, -0.00004257, -0.00000353, 0.00001451, -0.00001616, -0.00000323, -0.00003159, 0.00005824, -0.00000336, -0.00000944, -0.00000575, -0.00002306, -0.00002459, -0.00000105, -0.00001850, -0.00009688, -0.00018025, 0.00053943, 0.00017286, -0.00022326, -0.00000431, -0.00004963, -0.00001252, -0.00000214, -0.00000359, -0.00002251, -0.00000384, -0.00000869, -0.00001286, },
		{ -0.00001643, -0.00001423, -0.00004745, -0.00008460, 0.00007415, 0.00024955, 0.00007369, -0.00011440, -0.00002818, -0.00000696, -0.00000130, -0.00003205, -0.00000614, -0.00001006, -0.00000439, 0.00005154, -0.00001504, -0.00001265, -0.00001664, 0.00002431, -0.00000927, -0.00002833, -0.00001177, 0.00000923, -0.00001199, 0.00001256, -0.00001097, 0.00001404, 0.00000483, -0.00000351, -0.00000994, 0.00002522, 0.00000952, -0.00001505, 0.00001566, -0.00003350, 0.00002021, -0.00005634, 0.00002068, -0.00000369, -0.00000361, -0.00002232, -0.00000561, 0.00005161, -0.00001085, -0.00000933, -0.00000393, -0.00003494, -0.00000597, -0.00001719, -0.00000502, -0.00017340, -0.00000794, 0.00054383, -0.00000545, -0.00015833, -0.00000967, -0.00004240, -0.00001004, 0.00000256, -0.00001315, -0.00001588, -0.00000705, -0.00000566, },
		{ -0.00000362, -0.00002054, -0.00001561, -0.00006266, -0.00008022, 0.00014210, 0.00026405, -0.00000795, -0.00011280, -0.00000728, -0.00001080, -0.00000377, -0.00003650, 0.00000518, -0.00002235, 0.00001775, 0.00003970, -0.00002004, -0.00001348, -0.00001038, 0.00002522, -0.00001962, -0.00002778, -0.00000142, 0.00000081, -0.00000668, 0.00000863, -0.00000818, 0.00002018, -0.00000448, -0.00000124, -0.00000770, 0.00003322, -0.00001017, 0.00000361, -0.00000447, -0.00001472, 0.00000403, -0.00005118, 0.00003097, -0.00001330, 0.00000007, -0.00003187, 0.00001783, 0.00004074, -0.00001789, -0.00000379, -0.00001205, -0.00003308, -0.00000074, -0.00001864, -0.00002162, -0.00020468, 0.00017663, 0.00044968, -0.00010961, -0.00010114, -0.00001891, -0.00003425, -0.00000625, 0.00000188, -0.00001804, -0.00001167, -0.00000861, },
		{ -0.00000741, -0.00000399, -0.00002461, -0.00001759, -0.00008133, -0.00006207, 0.00022442, 0.00023862, -0.00008782, -0.00008477, -0.00000042, -0.00001084, -0.00001135, -0.00003253, 0.00000864, -0.00002800, 0.00003818, 0.00002222, -0.00002024, -0.00001536, -0.00000169, 0.00002179, -0.00002976, -0.00002160, 0.00000418, -0.00000772, 0.00000190, -0.00000031, 0.00000351, 0.00001545, -0.00000683, -0.00000287, 0.00000207, 0.00002783, -0.00001690, 0.00001313, -0.00001882, 0.00000366, -0.00001903, -0.00003323, 0.00002941, -0.00001538, -0.00000404, -0.00003068, 0.00003519, 0.00002719, -0.00002208, 0.00000275, -0.00002537, -0.00002083, -0.00000700, -0.00000829, -0.00006802, -0.00016717, 0.00031614, 0.00030798, -0.00014957, -0.00006200, -0.00002554, -0.00002693, -0.00000211, -0.00000264, -0.00001842, -0.00001057, },
		{ -0.00001098, -0.00000421, -0.00000730, -0.00002777, -0.00002080, -0.00010399, -0.00002055, 0.00030792, 0.00016603, -0.00014218, -0.00004483, -0.00000590, -0.00000643, -0.00002332, -0.00002021, 0.00000220, -0.00002256, 0.00005028, 0.00000482, -0.00001870, -0.00001578, 0.00000667, 0.00001513, -0.00003892, -0.00000919, 0.00000039, -0.00000873, 0.00000535, -0.00000543, 0.00001571, 0.00000526, -0.00000484, -0.00000508, 0.00001398, 0.00001338, -0.00001090, 0.00000979, -0.00002002, 0.00001076, -0.00003679, -0.00001241, 0.00002226, -0.00001478, -0.00001075, -0.00002224, 0.00004672, 0.00001049, -0.00001949, 0.00000322, -0.00003455, -0.00000828, -0.00001427, -0.00000138, -0.00012206, -0.00006975, 0.00037994, 0.00016644, -0.00014858, -0.00003914, -0.00002872, -0.00001992, 0.00000001, -0.00000762, -0.00001687, },
		},
		{ { 0.00000494, -0.00000662, -0.00004261, -0.00002532, -0.00014617, 0.00005074, 0.00038107, 0.00005278, -0.00013670, -0.00002506, -0.00001142, -0.00002909, 0.00002585, -0.00001793, -0.00000956, -0.00001345, -0.00000003, 0.00001459, -0.00007779, 0.00002284, -0.00000249, 0.00001139, -0.00003253, -0.00000132, 0.00002067, -0.00000409, -0.00000454, -0.00000066, 0.00000265, -0.00000816, 0.00004534, -0.00001303, -0.00001314, -0.00001303, 0.00004534, -0.00000816, 0.00000265, -0.00000066, -0.00000454, -0.00000409, 0.00002067, -0.00000132, -0.00003253, 0.00001139, -0.00000249, 0.00002284, -0.00007779, 0.00001459, -0.00000003, -0.00001345, -0.00000956, -0.00001793, 0.00002585, -0.00002909, -0.00001142, -0.00002506, -0.00013670, 0.00005278, 0.00038107, 0.00005074, -0.00014617, -0.00002532, -0.00004261, -0.00000662, },
		{ -0.00000168, 0.00000499, -0.00001377, -0.00004372, -0.00003155, -0.00016975, 0.00014094, 0.00039836, -0.00003977, -0.00012291, -0.00001151, -0.00001932, -0.00002210, 0.00002672, -0.00002648, -0.00000040, -0.00002176, 0.00001410, -0.00000659, -0.00007052, 0.00003413, -0.00001238, 0.00001453, -0.00004019, 0.00001469, 0.00001026, -0.00000183, -0.00000633, 0.00000410, -0.00000347, 0.00000342, 0.00003833, -0.00001765, -0.00001399, -0.00000440, 0.00004725, -0.00001790, 0.00000887, -0.00000560, -0.00000352, -0.00000296, 0.00002680, -0.00001639, -0.00002326, 0.00000806, 0.00000845, 0.00000439, -0.00007376, 0.00002831, -0.00001197, -0.00000583, -0.00001892, -0.00000690, 0.00002124, -0.00003381, -0.00000248, -0.00004844, -0.00012562, 0.00013466, 0.00033488, -0.00001497, -0.00012013, -0.00002396, -0.00003843, },
		{ -0.00003261, 0.00000160, 0.00000473, -0.00002273, -0.00004117, -0.00004542, -0.00018435, 0.00025356, 0.00037396, -0.00012638, -0.00008973, -0.00001061, -0.00002216, -0.00001596, 0.00002555, -0.00003274, 0.00000787, -0.00002949, 0.00002815, -0.00003295, -0.00005164, 0.00003472, -0.00001688, 0.00001429, -0.00004340, 0.00002844, -0.00000216, 0.00000325, -0.00000906, 0.00000856, -0.00000865, 0.00001505, 0.00002781, -0.00001880, -0.00001612, 0.00000793, 0.00004306, -0.00002410, 0.00001407, -0.00001053, -0.00000265, 0.00000101, 0.00002779, -0.00002851, -0.00001423, 0.00000680, 0.00001630, -0.00001637, -0.00006096, 0.00003411, -0.00002033, -0.00000002, -0.00002763, 0.00000643, 0.00001145, -0.00003268, 0.00000139, -0.00007366, -0.00009236, 0.00019563, 0.00027254, -0.00005770, -0.00009611, -0.00002464, },
		{ -0.00002493, -0.00002678, 0.00000415, 0.00000327, -0.00003252, -0.00003509, -0.00006958, -0.00018087, 0.00038041, 0.00029957, -0.00018779, -0.00004988, -0.00001854, -0.00002033, -0.00001112, 0.00002248, -0.00003594, 0.00001375, -0.00003454, 0.00003866, -0.00005992, -0.00002412, 0.00002411, -0.00001342, 0.00000812, -0.00003951, 0.00003641, -0.00001334, 0.00000947, -0.00001222, 0.00001248, -0.00001227, 0.00002519, 0.00001565, -0.00001764, -0.00001837, 0.00002264, 0.00003278, -0.00002545, 0.00001700, -0.00001495, -0.00000133, 0.00000658, 0.00002404, -0.00003696, -0.00000588, 0.00000802, 0.00001862, -0.00003502, -0.00004285, 0.00003266, -0.00002407, 0.00000252, -0.00003381, 0.00002020, -0.00000273, -0.00002445, -0.00000437, -0.00009149, -0.00004520, 0.00023220, 0.00020450, -0.00008083, -0.00007681, },
		{ -0.00006287, -0.00002367, -0.00002187, 0.00000641, 0.00000020, -0.00004231, -0.00002588, -0.00010649, -0.00014748, 0.00050498, 0.00017663, -0.00020849, -0.00001885, -0.00002677, -0.00001819, -0.00000515, 0.00001570, -0.00003396, 0.00001477, -0.00003414, 0.00004141, -0.00008119, 0.00000588, 0.00000557, -0.00000252, -0.00000438, -0.00002745, 0.00003600, -0.00002001, 0.00001448, -0.00001463, 0.00001528, -0.00001381, 0.00003253, 0.00000372, -0.00001562, -0.00001907, 0.00003743, 0.00001765, -0.00002152, 0.00001660, -0.00001809, 0.00000073, 0.00001236, 0.00001664, -0.00004187, 0.00000238, 0.00001037, 0.00001521, -0.00004870, -0.00002271, 0.00002518, -0.00002252, 0.00000004, -0.00003461, 0.00003056, -0.00001747, -0.00001189, -0.00001979, -0.00009574, 0.00000622, 0.00024556, 0.00013869, -0.00008894, },
		{ -0.00008761, -0.00005285, -0.00002149, -0.00001737, 0.00000764, -0.00000389, -0.00005184, -0.00001426, -0.00015748, -0.00007130, 0.00060250, 0.00002011, -0.00018334, -0.00000848, -0.00002666, -0.00002114, 0.00000536, 0.00000282, -0.00002446, 0.00000848, -0.00002586, 0.00003277, -0.00009035, 0.00003046, -0.00001453, 0.00001196, -0.00002072, -0.00000868, 0.00002667, -0.00001998, 0.00001608, -0.00001487, 0.00001616, -0.00001280, 0.00003612, -0.00000639, -0.00001412, -0.00001643, 0.00004938, 0.00000018, -0.00001308, 0.00001248, -0.00001918, 0.00000337, 0.00001731, 0.00000675, -0.00004346, 0.00001119, 0.00001171, 0.00000768, -0.00005621, -0.00000334, 0.00001348, -0.00001584, -0.00000852, -0.00002742, 0.00003309, -0.00002708, -0.00000122, -0.00003968, -0.00008585, 0.00005458, 0.00023882, 0.00008115, },
		{ 0.00003642, -0.00008271, -0.00004387, -0.00002035, -0.00001170, 0.00000622, -0.00000770, -0.00006101, -0.00000225, -0.00022036, 0.00005752, 0.00064395, -0.00014196, -0.00012132, -0.00002142, -0.00001442, -0.00003176, 0.00002178, -0.00001679, -0.00000686, -0.00000583, -0.00000931, 0.00001165, -0.00008335, 0.00004260, -0.00002884, 0.00002415, -0.00003609, 0.00001279, 0.00001042, -0.00001296, 0.00001307, -0.00001184, 0.00001440, -0.00000896, 0.00003545, -0.00001358, -0.00001407, -0.00000914, 0.00005554, -0.00001629, -0.00000214, 0.00000520, -0.00001782, 0.00000620, 0.00002084, -0.00000478, -0.00004149, 0.00002034, 0.00001036, -0.00000172, -0.00005754, 0.00001283, 0.00000021, -0.00000583, -0.00002222, -0.00001177, 0.00002521, -0.00002683, 0.00000079, -0.00005589, -0.00006632, 0.00009635, 0.00021527, },
		{ 0.00017839, 0.00000699, -0.00007897, -0.00003324, -0.00002214, -0.00000351, 0.00000101, -0.00001053, -0.00006858, 0.00000520, -0.00028621, 0.00023961, 0.00060389, -0.00027422, -0.00004431, -0.00004921, 0.00000599, -0.00004707, 0.00004113, -0.00003996, 0.00001598, -0.00002564, 0.00001258, -0.00001901, -0.00006057, 0.00003890, -0.00003182, 0.00002843, -0.00004501, 0.00003146, -0.00000856, -0.00000083, 0.00000578, -0.00000540, 0.00000981, -0.00000241, 0.00003060, -0.00001746, -0.00001554, 0.00000304, 0.00005380, -0.00002831, 0.00000848, -0.00000372, -0.00001424, 0.00000887, 0.00002266, -0.00001733, -0.00003526, 0.00002839, 0.00000603, -0.00001114, -0.00005313, 0.00002357, -0.00001112, 0.00000384, -0.00003734, 0.00000926, 0.00000830, -0.00001594, -0.00000945, -0.00006128, -0.00004376, 0.00013074, },
		{ 0.00015722, 0.00013294, -0.00000783, -0.00007835, -0.00002030, -0.00002684, 0.00000661, -0.00000713, -0.00001371, -0.00007096, -0.00000096, -0.00033694, 0.00046039, 0.00047039, -0.00034416, 0.00001970, -0.00007536, 0.00002397, -0.00005879, 0.00005605, -0.00005963, 0.00003755, -0.00004507, 0.00003349, -0.00005213, -0.00002763, 0.00002127, -0.00002215, 0.00002190, -0.00004349, 0.00004211, -0.00002514, 0.00001297, -0.00000394, 0.00000344, 0.00000301, 0.00000611, 0.00002232, -0.00001845, -0.00001763, 0.00001885, 0.00004368, -0.00003340, 0.00001595, -0.00001218, -0.00000934, 0.00001136, 0.00002243, -0.00003014, -0.00002406, 0.00003303, -0.00000005, -0.00001973, -0.00004356, 0.00002707, -0.00001681, 0.00000833, -0.00004796, 0.00002918, -0.00001164, 0.00000088, -0.00002960, -0.00005360, -0.00002296, },
		{ -0.00000433, 0.00017345, 0.00008583, -0.00001247, -0.00007916, -0.00000741, -0.00003168, 0.00001537, -0.00001468, -0.00002096, -0.00006248, -0.00003290, -0.00034584, 0.00068712, 0.00025341, -0.00033354, 0.00004672, -0.00008245, 0.00002698, -0.00005680, 0.00005745, -0.00006707, 0.00004959, -0.00005662, 0.00004552, -0.00007819, 0.00000588, -0.00000331, -0.00000347, 0.00000581, -0.00003080, 0.00004164, -0.00003487, 0.00002447, -0.00001327, 0.00001265, -0.00000452, 0.00001518, 0.00001203, -0.00001764, -0.00001855, 0.00003551, 0.00002678, -0.00003081, 0.00001835, -0.00001819, -0.00000435, 0.00001398, 0.00001956, -0.00004181, -0.00000812, 0.00003216, -0.00000582, -0.00002733, -0.00002976, 0.00002270, -0.00001430, 0.00000346, -0.00004798, 0.00004011, -0.00002586, 0.00001459, -0.00005138, -0.00003705, },
		{ -0.00002050, 0.00001588, 0.00017519, 0.00004556, -0.00001409, -0.00007684, 0.00000073, -0.00003184, 0.00001771, -0.00001658, -0.00003686, -0.00003785, -0.00010178, -0.00028286, 0.00087209, -0.00001249, -0.00024659, 0.00002650, -0.00006041, 0.00000713, -0.00003442, 0.00003905, -0.00005572, 0.00004550, -0.00005422, 0.00004227, -0.00008856, 0.00002954, -0.00002514, 0.00001681, -0.00001463, -0.00001001, 0.00003023, -0.00003536, 0.00003053, -0.00001938, 0.00001979, -0.00001080, 0.00002297, 0.00000154, -0.00001645, -0.00001622, 0.00004927, 0.00000663, -0.00002196, 0.00001534, -0.00002061, -0.00000031, 0.00001713, 0.00001327, -0.00005021, 0.00001086, 0.00002496, -0.00000954, -0.00003381, -0.00001352, 0.00001184, -0.00000368, -0.00001209, -0.00003405, 0.00003627, -0.00002654, 0.00001555, -0.00006375, },
		{ -0.00005804, -0.00001309, 0.00004259, 0.00015842, 0.00001983, -0.00001998, -0.00006650, -0.00000045, -0.00002297, 0.00000942, -0.00000913, -0.00006349, 0.00000342, -0.00021089, -0.00012451, 0.00096279, -0.00027353, -0.00011121, -0.00003248, -0.00001228, -0.00003364, 0.00000695, 0.00000145, -0.00002502, 0.00002375, -0.00003640, 0.00002201, -0.00007910, 0.00003606, -0.00003547, 0.00003040, -0.00003212, 0.00001290, 0.00001126, -0.00002698, 0.00002976, -0.00002039, 0.00002284, -0.00001395, 0.00002772, -0.00000745, -0.00001601, -0.00000904, 0.00005644, -0.00001221, -0.00001001, 0.00000834, -0.00001954, 0.00000239, 0.00002086, 0.00000304, -0.00005279, 0.00002957, 0.00001267, -0.00001071, -0.00003842, 0.00000219, -0.00000174, 0.00001140, -0.00003515, -0.00000791, 0.00001723, -0.00001095, -0.00000162, },
		{ -0.00003435, -0.00003275, -0.00001919, 0.00007773, 0.00012265, 0.00001232, -0.00003418, -0.00004619, -0.00001232, -0.00000436, -0.00000960, 0.00000667, -0.00009690, 0.00005250, -0.00034844, 0.00013413, 0.00091721, -0.00047007, 0.00002841, -0.00010517, 0.00004550, -0.00008240, 0.00005643, -0.00004618, 0.00001789, -0.00001032, -0.00000795, -0.00001054, -0.00005247, 0.00002436, -0.00003013, 0.00003150, -0.00004015, 0.00003118, -0.00000986, -0.00001250, 0.00002288, -0.00001600, 0.00002091, -0.00001285, 0.00002824, -0.00001381, -0.00001663, 0.00000322, 0.00005459, -0.00002543, 0.00000108, 0.00000000, -0.00001632, 0.00000428, 0.00002451, -0.00001084, -0.00004742, 0.00004384, -0.00000167, -0.00001035, -0.00003967, 0.00001362, -0.00001255, 0.00002432, -0.00005825, 0.00002302, -0.00001071, 0.00001608, },
		{ 0.00004289, -0.00007163, 0.00000445, -0.00003523, 0.00011713, 0.00007341, 0.00002035, -0.00005504, -0.00001920, -0.00003082, 0.00001903, -0.00003341, 0.00002328, -0.00012551, 0.00008945, -0.00048342, 0.00046975, 0.00071997, -0.00055445, 0.00012590, -0.00015906, 0.00008889, -0.00011925, 0.00009657, -0.00008835, 0.00005973, -0.00004576, 0.00002114, -0.00004509, -0.00001794, 0.00000048, -0.00001137, 0.00001937, -0.00003559, 0.00003972, -0.00002755, 0.00000394, 0.00001227, -0.00000754, 0.00001453, -0.00000759, 0.00002438, -0.00001733, -0.00001751, 0.00001898, 0.00004362, -0.00003045, 0.00000796, -0.00000680, -0.00001287, 0.00000657, 0.00002659, -0.00002680, -0.00003344, 0.00004993, -0.00001417, -0.00001039, -0.00003600, 0.00001742, -0.00001515, 0.00002764, -0.00007185, 0.00004770, -0.00003599, },
		{ -0.00004615, 0.00005512, -0.00009749, 0.00003883, -0.00005034, 0.00015052, 0.00002215, 0.00003500, -0.00007528, 0.00000616, -0.00004696, 0.00003728, -0.00005097, 0.00002816, -0.00013245, 0.00008654, -0.00056825, 0.00082834, 0.00039294, -0.00050726, 0.00014890, -0.00016710, 0.00009618, -0.00012526, 0.00010996, -0.00010871, 0.00008566, -0.00006985, 0.00003929, -0.00006930, 0.00001172, -0.00002449, 0.00001277, -0.00000122, -0.00002005, 0.00003681, -0.00003790, 0.00001836, 0.00000106, 0.00000254, 0.00000554, 0.00000053, 0.00001704, -0.00001862, -0.00001694, 0.00003493, 0.00002602, -0.00002724, 0.00000907, -0.00001014, -0.00001087, 0.00001052, 0.00002519, -0.00004192, -0.00001233, 0.00004596, -0.00002160, -0.00001251, -0.00002689, 0.00001226, -0.00000650, 0.00001620, -0.00006805, 0.00005563, },
		{ 0.00004155, -0.00003333, 0.00004194, -0.00009744, 0.00005436, -0.00005083, 0.00016499, -0.00001702, 0.00004423, -0.00008497, 0.00001959, -0.00005019, 0.00003925, -0.00005049, 0.00000877, -0.00010217, 0.00001734, -0.00054953, 0.00113315, -0.00000453, -0.00034591, 0.00009112, -0.00011929, 0.00005790, -0.00009107, 0.00008677, -0.00009681, 0.00008531, -0.00007343, 0.00003824, -0.00007368, 0.00002512, -0.00003871, 0.00003165, -0.00002180, 0.00000077, 0.00002454, -0.00003970, 0.00002807, -0.00000797, 0.00001148, -0.00000354, 0.00000937, 0.00000788, -0.00001878, -0.00001299, 0.00004691, 0.00000634, -0.00001845, 0.00000522, -0.00000980, -0.00001085, 0.00001672, 0.00001870, -0.00005253, 0.00001214, 0.00003274, -0.00002263, -0.00001702, -0.00001380, -0.00000001, 0.00001210, -0.00001006, -0.00004441, },
		{ -0.00000604, 0.00000868, 0.00000127, 0.00000174, -0.00006528, 0.00004097, -0.00002674, 0.00015079, -0.00003268, 0.00003805, -0.00007647, 0.00001384, -0.00003363, 0.00001825, -0.00002557, -0.00004019, -0.00002956, -0.00013055, -0.00038528, 0.00130288, -0.00038788, -0.00012152, -0.00002449, -0.00002799, -0.00001756, -0.00002238, 0.00003012, -0.00005324, 0.00005742, -0.00005513, 0.00001714, -0.00005605, 0.00001717, -0.00003454, 0.00003647, -0.00003371, 0.00001939, 0.00000770, -0.00003434, 0.00003227, -0.00001316, 0.00001713, -0.00001030, 0.00001667, -0.00000128, -0.00001870, -0.00000444, 0.00005125, -0.00001018, -0.00000838, -0.00000084, -0.00000735, -0.00001191, 0.00002446, 0.00000662, -0.00005521, 0.00003479, 0.00001373, -0.00001840, -0.00002245, -0.00000025, -0.00001428, 0.00003461, -0.00004492, },
		{ -0.00007728, 0.00003522, -0.00003133, 0.00004639, -0.00005546, -0.00000726, -0.00000006, 0.00002239, 0.00010670, -0.00002134, 0.00001347, -0.00004908, -0.00001074, 0.00000121, -0.00002301, 0.00001941, -0.00011086, 0.00007240, -0.00034357, -0.00006370, 0.00127564, -0.00066968, 0.00009605, -0.00015337, 0.00007552, -0.00010541, 0.00006049, -0.00004356, 0.00000947, 0.00001115, -0.00002263, -0.00001620, -0.00002323, -0.00000833, -0.00001216, 0.00002420, -0.00003172, 0.00002956, -0.00000831, -0.00002477, 0.00003179, -0.00001429, 0.00001851, -0.00001320, 0.00002067, -0.00000902, -0.00001853, 0.00000834, 0.00004622, -0.00001942, -0.00000142, -0.00000553, -0.00000547, -0.00001194, 0.00003168, -0.00000983, -0.00004801, 0.00005036, -0.00000597, -0.00001187, -0.00002604, 0.00000908, -0.00002356, 0.00005180, },
		{ 0.00005453, -0.00009487, 0.00006468, -0.00006214, 0.00008395, -0.00011025, 0.00005869, -0.00005503, 0.00008548, 0.00004242, 0.00001022, -0.00002325, -0.00001101, -0.00004451, 0.00004287, -0.00007093, 0.00006788, -0.00018084, 0.00017064, -0.00057690, 0.00038373, 0.00103116, -0.00078535, 0.00024024, -0.00024532, 0.00015242, -0.00017309, 0.00012919, -0.00010970, 0.00007080, -0.00003705, 0.00001005, -0.00004783, 0.00001095, -0.00003951, 0.00002006, -0.00000084, -0.00001639, 0.00002875, -0.00001964, -0.00001400, 0.00002823, -0.00001225, 0.00001591, -0.00001189, 0.00002061, -0.00001467, -0.00001747, 0.00002309, 0.00003287, -0.00001982, -0.00000034, -0.00000609, -0.00000667, -0.00000861, 0.00003554, -0.00002767, -0.00003133, 0.00005519, -0.00002129, -0.00000660, -0.00002509, 0.00001028, -0.00002165, },
		{ -0.00000594, 0.00003762, -0.00008906, 0.00007045, -0.00006897, 0.00009579, -0.00014059, 0.00010875, -0.00010220, 0.00014289, -0.00002376, 0.00004688, -0.00005849, 0.00002285, -0.00007123, 0.00007310, -0.00010468, 0.00009555, -0.00021873, 0.00021868, -0.00075986, 0.00088091, 0.00060314, -0.00071392, 0.00026974, -0.00026318, 0.00017145, -0.00019271, 0.00015802, -0.00014466, 0.00010985, -0.00006989, 0.00002833, -0.00006322, 0.00003092, -0.00006097, 0.00004855, -0.00002841, 0.00000607, 0.00001896, -0.00002524, -0.00000386, 0.00002313, -0.00000846, 0.00001062, -0.00000722, 0.00001681, -0.00001829, -0.00001412, 0.00003607, 0.00001498, -0.00001305, -0.00000511, -0.00000179, -0.00001199, -0.00000048, 0.00003336, -0.00004280, -0.00000816, 0.00004840, -0.00002891, -0.00000508, -0.00001851, 0.00000213, },
		{ -0.00001275, 0.00002081, 0.00000264, -0.00005889, 0.00004865, -0.00004509, 0.00007137, -0.00013052, 0.00012226, -0.00012001, 0.00017362, -0.00007135, 0.00007104, -0.00007694, 0.00003695, -0.00007460, 0.00007412, -0.00010441, 0.00007983, -0.00019589, 0.00017161, -0.00081318, 0.00132096, 0.00007588, -0.00048598, 0.00018220, -0.00019692, 0.00012165, -0.00015253, 0.00013421, -0.00013492, 0.00011353, -0.00007637, 0.00002359, -0.00005371, 0.00002603, -0.00006005, 0.00006002, -0.00004644, 0.00002635, 0.00000562, -0.00002678, 0.00000536, 0.00001728, -0.00000421, 0.00000433, -0.00000076, 0.00001032, -0.00002025, -0.00000720, 0.00004327, -0.00000201, -0.00000337, -0.00001296, 0.00000554, -0.00002030, 0.00001203, 0.00002384, -0.00005113, 0.00001655, 0.00003223, -0.00002826, -0.00000764, -0.00000774, },
		{ 0.00000339, -0.00002827, 0.00005061, -0.00004166, -0.00001235, 0.00000590, 0.00000432, 0.00001327, -0.00007745, 0.00009036, -0.00009609, 0.00016385, -0.00008568, 0.00007014, -0.00006884, 0.00002242, -0.00004616, 0.00003728, -0.00006094, 0.00001124, -0.00010060, 0.00000521, -0.00067399, 0.00159276, -0.00043541, -0.00017540, 0.00001465, -0.00006793, 0.00001767, -0.00006257, 0.00006370, -0.00008288, 0.00008189, -0.00005659, -0.00000250, -0.00002106, -0.00000437, -0.00003267, 0.00004722, -0.00004614, 0.00003600, -0.00000510, -0.00002721, 0.00001445, 0.00001073, -0.00000031, -0.00000139, 0.00000565, 0.00000251, -0.00002077, 0.00000351, 0.00004187, -0.00001289, 0.00000393, -0.00001944, 0.00001216, -0.00002845, 0.00002638, 0.00000778, -0.00004991, 0.00003722, 0.00001133, -0.00002149, -0.00001239, },
		{ -0.00001611, 0.00001008, -0.00003685, 0.00007269, -0.00008185, 0.00003572, -0.00004224, 0.00006289, -0.00006212, 0.00000521, 0.00002061, -0.00003322, 0.00011333, -0.00006411, 0.00004251, -0.00003545, -0.00001755, 0.00000904, -0.00003076, 0.00001725, -0.00009848, 0.00005119, -0.00026757, -0.00032112, 0.00161338, -0.00081578, 0.00012317, -0.00016987, 0.00007873, -0.00010481, 0.00004812, -0.00003028, -0.00000635, 0.00002830, -0.00002190, -0.00003789, 0.00002219, -0.00004949, 0.00001412, 0.00001253, -0.00002592, 0.00003106, -0.00000912, -0.00002888, 0.00002418, 0.00000327, 0.00000296, -0.00000547, 0.00001048, -0.00000537, -0.00001955, 0.00001660, 0.00003160, -0.00001458, 0.00000462, -0.00002031, 0.00001394, -0.00003239, 0.00003854, -0.00001175, -0.00003870, 0.00004921, -0.00000868, -0.00001247, },
		{ -0.00000513, -0.00001587, 0.00000858, -0.00003238, 0.00007736, -0.00010442, 0.00006883, -0.00007670, 0.00010840, -0.00012951, 0.00009162, -0.00006459, 0.00005058, 0.00003648, -0.00001759, -0.00000124, 0.00001007, -0.00006714, 0.00007212, -0.00010772, 0.00010372, -0.00021582, 0.00021415, -0.00059046, 0.00020882, 0.00135570, -0.00098360, 0.00032388, -0.00030475, 0.00019089, -0.00020210, 0.00014187, -0.00011551, 0.00006802, -0.00002603, 0.00000991, -0.00006481, 0.00005708, -0.00009058, 0.00006402, -0.00003226, 0.00000753, 0.00001390, -0.00000639, -0.00003196, 0.00003422, -0.00000492, 0.00000554, -0.00000745, 0.00001275, -0.00001240, -0.00001583, 0.00002907, 0.00001511, -0.00000736, -0.00000267, -0.00001344, 0.00000819, -0.00002881, 0.00004427, -0.00003018, -0.00001969, 0.00005019, -0.00002302, },
		{ -0.00002911, -0.00000196, -0.00001056, -0.00000186, -0.00001298, 0.00005994, -0.00010089, 0.00007503, -0.00008221, 0.00012115, -0.00016362, 0.00015226, -0.00013562, 0.00012806, -0.00004241, 0.00003348, -0.00004233, 0.00004724, -0.00010325, 0.00011640, -0.00016256, 0.00016187, -0.00029463, 0.00032388, -0.00087243, 0.00082181, 0.00086167, -0.00091460, 0.00037623, -0.00034322, 0.00022878, -0.00023878, 0.00018635, -0.00016282, 0.00011503, -0.00006071, 0.00002241, -0.00006737, 0.00006581, -0.00010806, 0.00009740, -0.00007014, 0.00004160, -0.00000776, -0.00000086, -0.00003408, 0.00004269, -0.00001289, 0.00000733, -0.00000735, 0.00001209, -0.00001789, -0.00000887, 0.00003718, -0.00000255, 0.00000507, -0.00001581, 0.00000006, -0.00000498, -0.00001667, 0.00004074, -0.00004274, 0.00000282, 0.00004074, },
		{ 0.00002409, -0.00002716, -0.00000318, -0.00000157, -0.00001819, 0.00001765, 0.00002306, -0.00007146, 0.00005191, -0.00005371, 0.00009201, -0.00014895, 0.00016481, -0.00016649, 0.00017279, -0.00009825, 0.00006756, -0.00006189, 0.00005685, -0.00010498, 0.00011806, -0.00016753, 0.00015885, -0.00029289, 0.00031752, -0.00101094, 0.00138566, 0.00023531, -0.00064872, 0.00027963, -0.00027422, 0.00017969, -0.00020116, 0.00016684, -0.00015710, 0.00012059, -0.00006465, 0.00000813, -0.00003911, 0.00003968, -0.00008968, 0.00009908, -0.00008547, 0.00006273, -0.00002408, 0.00000147, -0.00003139, 0.00004667, -0.00001892, 0.00000801, -0.00000549, 0.00000858, -0.00002122, 0.00000143, 0.00003771, -0.00001559, 0.00001687, -0.00002980, 0.00001598, -0.00002229, 0.00000203, 0.00002773, -0.00004598, 0.00002362, },
		{ 0.00003805, 0.00000500, -0.00001973, -0.00000688, 0.00000754, -0.00003431, 0.00005098, -0.00002361, -0.00002538, 0.00000803, 0.00000086, 0.00002692, -0.00008644, 0.00012281, -0.00014476, 0.00016924, -0.00011474, 0.00007100, -0.00004969, 0.00003002, -0.00006354, 0.00006723, -0.00011085, 0.00008062, -0.00019144, 0.00015861, -0.00092384, 0.00176411, -0.00038374, -0.00027702, 0.00008100, -0.00012509, 0.00006244, -0.00010235, 0.00009163, -0.00010279, 0.00008695, -0.00004060, -0.00002747, 0.00001309, -0.00001616, -0.00003639, 0.00006486, -0.00007036, 0.00006200, -0.00002751, -0.00000444, -0.00002057, 0.00004345, -0.00002124, 0.00000718, -0.00000241, 0.00000270, -0.00002165, 0.00001388, 0.00002941, -0.00001953, 0.00002219, -0.00003869, 0.00002846, -0.00003824, 0.00002260, 0.00000802, -0.00003888, },
		{ -0.00002321, 0.00004331, -0.00001171, -0.00001054, -0.00001013, 0.00001255, -0.00004316, 0.00007635, -0.00006647, 0.00002204, -0.00003988, 0.00006282, -0.00005423, 0.00000581, 0.00003879, -0.00007655, 0.00011885, -0.00009071, 0.00004395, -0.00000961, -0.00002641, 0.00001196, -0.00002538, -0.00000463, -0.00005846, -0.00000707, -0.00014301, -0.00057977, 0.00185578, -0.00085848, 0.00008710, -0.00014344, 0.00004972, -0.00008008, 0.00002279, -0.00001133, -0.00002176, 0.00003124, -0.00000374, -0.00006767, 0.00007045, -0.00008332, 0.00003677, 0.00000404, -0.00002833, 0.00003896, -0.00001619, -0.00001985, -0.00000087, 0.00003196, -0.00001879, 0.00000468, 0.00000114, -0.00000474, -0.00001843, 0.00002597, 0.00001382, -0.00001288, 0.00001747, -0.00003787, 0.00003217, -0.00004707, 0.00003913, -0.00001334, },
		{ -0.00003053, -0.00000304, 0.00003908, -0.00002250, -0.00000306, -0.00001037, 0.00001043, -0.00003947, 0.00008482, -0.00009267, 0.00005498, -0.00007300, 0.00010932, -0.00012419, 0.00009828, -0.00005983, 0.00001544, 0.00003951, -0.00004022, 0.00000038, 0.00004093, -0.00009101, 0.00009584, -0.00012978, 0.00011634, -0.00021618, 0.00020643, -0.00052457, -0.00001613, 0.00162566, -0.00109284, 0.00034295, -0.00031536, 0.00018922, -0.00019720, 0.00013094, -0.00010527, 0.00005564, -0.00002209, 0.00002515, -0.00009128, 0.00010918, -0.00013639, 0.00010538, -0.00006330, 0.00002660, 0.00000213, 0.00000489, -0.00004108, 0.00002477, 0.00001373, -0.00001199, 0.00000090, 0.00000422, -0.00001257, -0.00001111, 0.00003433, -0.00000490, 0.00000207, 0.00000292, -0.00002609, 0.00002452, -0.00004488, 0.00004647, },
		{ 0.00004217, -0.00003883, 0.00001655, 0.00002741, -0.00002597, 0.00000065, -0.00000663, 0.00000079, -0.00002201, 0.00007249, -0.00009484, 0.00006299, -0.00007773, 0.00012193, -0.00015817, 0.00016038, -0.00014035, 0.00010028, -0.00004138, 0.00001350, -0.00003799, 0.00007777, -0.00013545, 0.00015460, -0.00020656, 0.00020548, -0.00033491, 0.00037168, -0.00088184, 0.00066305, 0.00111838, -0.00105761, 0.00043202, -0.00038082, 0.00024804, -0.00024921, 0.00018645, -0.00015863, 0.00010260, -0.00005163, 0.00002915, -0.00008197, 0.00011006, -0.00015252, 0.00014442, -0.00011344, 0.00007494, -0.00003422, 0.00002577, -0.00006046, 0.00005017, -0.00000721, -0.00000278, -0.00000316, 0.00000569, -0.00001922, -0.00000004, 0.00003585, -0.00002105, 0.00001993, -0.00001739, -0.00000613, 0.00000681, -0.00003131, },
		{ -0.00000983, 0.00002742, -0.00003612, 0.00003095, 0.00001190, -0.00002290, 0.00000043, 0.00000005, -0.00001366, 0.00000561, 0.00004213, -0.00007368, 0.00004495, -0.00005103, 0.00009396, -0.00014313, 0.00017142, -0.00017666, 0.00015023, -0.00009749, 0.00004788, -0.00005131, 0.00008030, -0.00013616, 0.00016014, -0.00022190, 0.00022186, -0.00036220, 0.00041270, -0.00109692, 0.00131061, 0.00044809, -0.00079583, 0.00035244, -0.00032661, 0.00021181, -0.00022134, 0.00017412, -0.00015654, 0.00010602, -0.00004784, 0.00000306, -0.00003608, 0.00006692, -0.00012076, 0.00013788, -0.00012784, 0.00009912, -0.00005567, 0.00003551, -0.00006924, 0.00006792, -0.00002546, 0.00000589, -0.00000616, 0.00000451, -0.00002291, 0.00001326, 0.00002889, -0.00002932, 0.00003397, -0.00003690, 0.00001600, -0.00001595, },
		{ -0.00003673, 0.00001329, 0.00000679, -0.00002353, 0.00003731, -0.00000345, -0.00001564, -0.00000231, 0.00000709, -0.00002797, 0.00003600, 0.00000234, -0.00003770, 0.00000946, -0.00000191, 0.00003338, -0.00008296, 0.00012817, -0.00015848, 0.00015087, -0.00011346, 0.00005035, -0.00003081, 0.00004115, -0.00008514, 0.00010220, -0.00016210, 0.00014750, -0.00027253, 0.00028342, -0.00107449, 0.00177523, -0.00023260, -0.00040623, 0.00015432, -0.00018144, 0.00010040, -0.00012792, 0.00010373, -0.00010545, 0.00007079, -0.00001669, -0.00004363, 0.00003448, -0.00000956, -0.00004726, 0.00008532, -0.00010033, 0.00009029, -0.00005356, 0.00002696, -0.00006118, 0.00007204, -0.00003615, 0.00001119, -0.00000699, 0.00000011, -0.00002194, 0.00002610, 0.00001418, -0.00002661, 0.00003854, -0.00004888, 0.00003313, },
		{ 0.00003957, -0.00004896, 0.00003113, -0.00001328, -0.00000498, 0.00003514, -0.00001536, -0.00000712, -0.00000541, 0.00001149, -0.00003673, 0.00006045, -0.00003550, 0.00000005, -0.00002853, 0.00005182, -0.00003970, 0.00000269, 0.00004591, -0.00009471, 0.00010581, -0.00009024, 0.00002346, 0.00001667, -0.00002924, 0.00000492, -0.00000543, -0.00004148, -0.00000240, -0.00008145, -0.00001063, -0.00077494, 0.00194401, -0.00077494, -0.00001063, -0.00008145, -0.00000240, -0.00004148, -0.00000543, 0.00000492, -0.00002924, 0.00001667, 0.00002346, -0.00009024, 0.00010581, -0.00009471, 0.00004591, 0.00000269, -0.00003970, 0.00005182, -0.00002853, 0.00000005, -0.00003550, 0.00006045, -0.00003673, 0.00001149, -0.00000541, -0.00000712, -0.00001536, 0.00003514, -0.00000498, -0.00001328, 0.00003113, -0.00004896, },
		{ -0.00003673, 0.00003313, -0.00004888, 0.00003854, -0.00002661, 0.00001418, 0.00002610, -0.00002194, 0.00000011, -0.00000699, 0.00001119, -0.00003615, 0.00007204, -0.00006118, 0.00002696, -0.00005356, 0.00009029, -0.00010033, 0.00008532, -0.00004726, -0.00000956, 0.00003448, -0.00004363, -0.00001669, 0.00007079, -0.00010545, 0.00010373, -0.00012792, 0.00010040, -0.00018144, 0.00015432, -0.00040623, -0.00023260, 0.00177523, -0.00107449, 0.00028342, -0.00027253, 0.00014750, -0.00016210, 0.00010220, -0.00008514, 0.00004115, -0.00003081, 0.00005035, -0.00011346, 0.00015087, -0.00015848, 0.00012817, -0.00008296, 0.00003338, -0.00000191, 0.00000946, -0.00003770, 0.00000234, 0.00003600, -0.00002797, 0.00000709, -0.00000231, -0.00001564, -0.00000345, 0.00003731, -0.00002353, 0.00000679, 0.00001329, },
		{ -0.00000983, -0.00001595, 0.00001600, -0.00003690, 0.00003397, -0.00002932, 0.00002889, 0.00001326, -0.00002291, 0.00000451, -0.00000616, 0.00000589, -0.00002546, 0.00006792, -0.00006924, 0.00003551, -0.00005567, 0.00009912, -0.00012784, 0.00013788, -0.00012076, 0.00006692, -0.00003608, 0.00000306, -0.00004784, 0.00010602, -0.00015654, 0.00017412, -0.00022134, 0.00021181, -0.00032661, 0.00035244, -0.00079583, 0.00044809, 0.00131061, -0.00109692, 0.00041270, -0.00036220, 0.00022186, -0.00022190, 0.00016014, -0.00013616, 0.00008030, -0.00005131, 0.00004788, -0.00009749, 0.00015023, -0.00017666, 0.00017142, -0.00014313, 0.00009396, -0.00005103, 0.00004495, -0.00007368, 0.00004213, 0.00000561, -0.00001366, 0.00000005, 0.00000043, -0.00002290, 0.00001190, 0.00003095, -0.00003612, 0.00002742, },
		{ 0.00004217, -0.00003131, 0.00000681, -0.00000613, -0.00001739, 0.00001993, -0.00002105, 0.00003585, -0.00000004, -0.00001922, 0.00000569, -0.00000316, -0.00000278, -0.00000721, 0.00005017, -0.00006046, 0.00002577, -0.00003422, 0.00007494, -0.00011344, 0.00014442, -0.00015252, 0.00011006, -0.00008197, 0.00002915, -0.00005163, 0.00010260, -0.00015863, 0.00018645, -0.00024921, 0.00024804, -0.00038082, 0.00043202, -0.00105761, 0.00111838, 0.00066305, -0.00088184, 0.00037168, -0.00033491, 0.00020548, -0.00020656, 0.00015460, -0.00013545, 0.00007777, -0.00003799, 0.00001350, -0.00004138, 0.00010028, -0.00014035, 0.00016038, -0.00015817, 0.00012193, -0.00007773, 0.00006299, -0.00009484, 0.00007249, -0.00002201, 0.00000079, -0.00000663, 0.00000065, -0.00002597, 0.00002741, 0.00001655, -0.00003883, },
		{ -0.00003053, 0.00004647, -0.00004488, 0.00002452, -0.00002609, 0.00000292, 0.00000207, -0.00000490, 0.00003433, -0.00001111, -0.00001257, 0.00000422, 0.00000090, -0.00001199, 0.00001373, 0.00002477, -0.00004108, 0.00000489, 0.00000213, 0.00002660, -0.00006330, 0.00010538, -0.00013639, 0.00010918, -0.00009128, 0.00002515, -0.00002209, 0.00005564, -0.00010527, 0.00013094, -0.00019720, 0.00018922, -0.00031536, 0.00034295, -0.00109284, 0.00162566, -0.00001613, -0.00052457, 0.00020643, -0.00021618, 0.00011634, -0.00012978, 0.00009584, -0.00009101, 0.00004093, 0.00000038, -0.00004022, 0.00003951, 0.00001544, -0.00005983, 0.00009828, -0.00012419, 0.00010932, -0.00007300, 0.00005498, -0.00009267, 0.00008482, -0.00003947, 0.00001043, -0.00001037, -0.00000306, -0.00002250, 0.00003908, -0.00000304, },
		{ -0.00002321, -0.00001334, 0.00003913, -0.00004707, 0.00003217, -0.00003787, 0.00001747, -0.00001288, 0.00001382, 0.00002597, -0.00001843, -0.00000474, 0.00000114, 0.00000468, -0.00001879, 0.00003196, -0.00000087, -0.00001985, -0.00001619, 0.00003896, -0.00002833, 0.00000404, 0.00003677, -0.00008332, 0.00007045, -0.00006767, -0.00000374, 0.00003124, -0.00002176, -0.00001133, 0.00002279, -0.00008008, 0.00004972, -0.00014344, 0.00008710, -0.00085848, 0.00185578, -0.00057977, -0.00014301, -0.00000707, -0.00005846, -0.00000463, -0.00002538, 0.00001196, -0.00002641, -0.00000961, 0.00004395, -0.00009071, 0.00011885, -0.00007655, 0.00003879, 0.00000581, -0.00005423, 0.00006282, -0.00003988, 0.00002204, -0.00006647, 0.00007635, -0.00004316, 0.00001255, -0.00001013, -0.00001054, -0.00001171, 0.00004331, },
		{ 0.00003805, -0.00003888, 0.00000802, 0.00002260, -0.00003824, 0.00002846, -0.00003869, 0.00002219, -0.00001953, 0.00002941, 0.00001388, -0.00002165, 0.00000270, -0.00000241, 0.00000718, -0.00002124, 0.00004345, -0.00002057, -0.00000444, -0.00002751, 0.00006200, -0.00007036, 0.00006486, -0.00003639, -0.00001616, 0.00001309, -0.00002747, -0.00004060, 0.00008695, -0.00010279, 0.00009163, -0.00010235, 0.00006244, -0.00012509, 0.00008100, -0.00027702, -0.00038374, 0.00176411, -0.00092384, 0.00015861, -0.00019144, 0.00008062, -0.00011085, 0.00006723, -0.00006354, 0.00003002, -0.00004969, 0.00007100, -0.00011474, 0.00016924, -0.00014476, 0.00012281, -0.00008644, 0.00002692, 0.00000086, 0.00000803, -0.00002538, -0.00002361, 0.00005098, -0.00003431, 0.00000754, -0.00000688, -0.00001973, 0.00000500, },
		{ 0.00002409, 0.00002362, -0.00004598, 0.00002773, 0.00000203, -0.00002229, 0.00001598, -0.00002980, 0.00001687, -0.00001559, 0.00003771, 0.00000143, -0.00002122, 0.00000858, -0.00000549, 0.00000801, -0.00001892, 0.00004667, -0.00003139, 0.00000147, -0.00002408, 0.00006273, -0.00008547, 0.00009908, -0.00008968, 0.00003968, -0.00003911, 0.00000813, -0.00006465, 0.00012059, -0.00015710, 0.00016684, -0.00020116, 0.00017969, -0.00027422, 0.00027963, -0.00064872, 0.00023531, 0.00138566, -0.00101094, 0.00031752, -0.00029289, 0.00015885, -0.00016753, 0.00011806, -0.00010498, 0.00005685, -0.00006189, 0.00006756, -0.00009825, 0.00017279, -0.00016649, 0.00016481, -0.00014895, 0.00009201, -0.00005371, 0.00005191, -0.00007146, 0.00002306, 0.00001765, -0.00001819, -0.00000157, -0.00000318, -0.00002716, },
		{ -0.00002911, 0.00004074, 0.00000282, -0.00004274, 0.00004074, -0.00001667, -0.00000498, 0.00000006, -0.00001581, 0.00000507, -0.00000255, 0.00003718, -0.00000887, -0.00001789, 0.00001209, -0.00000735, 0.00000733, -0.00001289, 0.00004269, -0.00003408, -0.00000086, -0.00000776, 0.00004160, -0.00007014, 0.00009740, -0.00010806, 0.00006581, -0.00006737, 0.00002241, -0.00006071, 0.00011503, -0.00016282, 0.00018635, -0.00023878, 0.00022878, -0.00034322, 0.00037623, -0.00091460, 0.00086167, 0.00082181, -0.00087243, 0.00032388, -0.00029463, 0.00016187, -0.00016256, 0.00011640, -0.00010325, 0.00004724, -0.00004233, 0.00003348, -0.00004241, 0.00012806, -0.00013562, 0.00015226, -0.00016362, 0.00012115, -0.00008221, 0.00007503, -0.00010089, 0.00005994, -0.00001298, -0.00000186, -0.00001056, -0.00000196, },
		{ -0.00000513, -0.00002302, 0.00005019, -0.00001969, -0.00003018, 0.00004427, -0.00002881, 0.00000819, -0.00001344, -0.00000267, -0.00000736, 0.00001511, 0.00002907, -0.00001583, -0.00001240, 0.00001275, -0.00000745, 0.00000554, -0.00000492, 0.00003422, -0.00003196, -0.00000639, 0.00001390, 0.00000753, -0.00003226, 0.00006402, -0.00009058, 0.00005708, -0.00006481, 0.00000991, -0.00002603, 0.00006802, -0.00011551, 0.00014187, -0.00020210, 0.00019089, -0.00030475, 0.00032388, -0.00098360, 0.00135570, 0.00020882, -0.00059046, 0.00021415, -0.00021582, 0.00010372, -0.00010772, 0.00007212, -0.00006714, 0.00001007, -0.00000124, -0.00001759, 0.00003648, 0.00005058, -0.00006459, 0.00009162, -0.00012951, 0.00010840, -0.00007670, 0.00006883, -0.00010442, 0.00007736, -0.00003238, 0.00000858, -0.00001587, },
		{ -0.00001611, -0.00001247, -0.00000868, 0.00004921, -0.00003870, -0.00001175, 0.00003854, -0.00003239, 0.00001394, -0.00002031, 0.00000462, -0.00001458, 0.00003160, 0.00001660, -0.00001955, -0.00000537, 0.00001048, -0.00000547, 0.00000296, 0.00000327, 0.00002418, -0.00002888, -0.00000912, 0.00003106, -0.00002592, 0.00001253, 0.00001412, -0.00004949, 0.00002219, -0.00003789, -0.00002190, 0.00002830, -0.00000635, -0.00003028, 0.00004812, -0.00010481, 0.00007873, -0.00016987, 0.00012317, -0.00081578, 0.00161338, -0.00032112, -0.00026757, 0.00005119, -0.00009848, 0.00001725, -0.00003076, 0.00000904, -0.00001755, -0.00003545, 0.00004251, -0.00006411, 0.00011333, -0.00003322, 0.00002061, 0.00000521, -0.00006212, 0.00006289, -0.00004224, 0.00003572, -0.00008185, 0.00007269, -0.00003685, 0.00001008, },
		{ 0.00000339, -0.00001239, -0.00002149, 0.00001133, 0.00003722, -0.00004991, 0.00000778, 0.00002638, -0.00002845, 0.00001216, -0.00001944, 0.00000393, -0.00001289, 0.00004187, 0.00000351, -0.00002077, 0.00000251, 0.00000565, -0.00000139, -0.00000031, 0.00001073, 0.00001445, -0.00002721, -0.00000510, 0.00003600, -0.00004614, 0.00004722, -0.00003267, -0.00000437, -0.00002106, -0.00000250, -0.00005659, 0.00008189, -0.00008288, 0.00006370, -0.00006257, 0.00001767, -0.00006793, 0.00001465, -0.00017540, -0.00043541, 0.00159276, -0.00067399, 0.00000521, -0.00010060, 0.00001124, -0.00006094, 0.00003728, -0.00004616, 0.00002242, -0.00006884, 0.00007014, -0.00008568, 0.00016385, -0.00009609, 0.00009036, -0.00007745, 0.00001327, 0.00000432, 0.00000590, -0.00001235, -0.00004166, 0.00005061, -0.00002827, },
		{ -0.00001275, -0.00000774, -0.00000764, -0.00002826, 0.00003223, 0.00001655, -0.00005113, 0.00002384, 0.00001203, -0.00002030, 0.00000554, -0.00001296, -0.00000337, -0.00000201, 0.00004327, -0.00000720, -0.00002025, 0.00001032, -0.00000076, 0.00000433, -0.00000421, 0.00001728, 0.00000536, -0.00002678, 0.00000562, 0.00002635, -0.00004644, 0.00006002, -0.00006005, 0.00002603, -0.00005371, 0.00002359, -0.00007637, 0.00011353, -0.00013492, 0.00013421, -0.00015253, 0.00012165, -0.00019692, 0.00018220, -0.00048598, 0.00007588, 0.00132096, -0.00081318, 0.00017161, -0.00019589, 0.00007983, -0.00010441, 0.00007412, -0.00007460, 0.00003695, -0.00007694, 0.00007104, -0.00007135, 0.00017362, -0.00012001, 0.00012226, -0.00013052, 0.00007137, -0.00004509, 0.00004865, -0.00005889, 0.00000264, 0.00002081, },
		{ -0.00000594, 0.00000213, -0.00001851, -0.00000508, -0.00002891, 0.00004840, -0.00000816, -0.00004280, 0.00003336, -0.00000048, -0.00001199, -0.00000179, -0.00000511, -0.00001305, 0.00001498, 0.00003607, -0.00001412, -0.00001829, 0.00001681, -0.00000722, 0.00001062, -0.00000846, 0.00002313, -0.00000386, -0.00002524, 0.00001896, 0.00000607, -0.00002841, 0.00004855, -0.00006097, 0.00003092, -0.00006322, 0.00002833, -0.00006989, 0.00010985, -0.00014466, 0.00015802, -0.00019271, 0.00017145, -0.00026318, 0.00026974, -0.00071392, 0.00060314, 0.00088091, -0.00075986, 0.00021868, -0.00021873, 0.00009555, -0.00010468, 0.00007310, -0.00007123, 0.00002285, -0.00005849, 0.00004688, -0.00002376, 0.00014289, -0.00010220, 0.00010875, -0.00014059, 0.00009579, -0.00006897, 0.00007045, -0.00008906, 0.00003762, },
		{ 0.00005453, -0.00002165, 0.00001028, -0.00002509, -0.00000660, -0.00002129, 0.00005519, -0.00003133, -0.00002767, 0.00003554, -0.00000861, -0.00000667, -0.00000609, -0.00000034, -0.00001982, 0.00003287, 0.00002309, -0.00001747, -0.00001467, 0.00002061, -0.00001189, 0.00001591, -0.00001225, 0.00002823, -0.00001400, -0.00001964, 0.00002875, -0.00001639, -0.00000084, 0.00002006, -0.00003951, 0.00001095, -0.00004783, 0.00001005, -0.00003705, 0.00007080, -0.00010970, 0.00012919, -0.00017309, 0.00015242, -0.00024532, 0.00024024, -0.00078535, 0.00103116, 0.00038373, -0.00057690, 0.00017064, -0.00018084, 0.00006788, -0.00007093, 0.00004287, -0.00004451, -0.00001101, -0.00002325, 0.00001022, 0.00004242, 0.00008548, -0.00005503, 0.00005869, -0.00011025, 0.00008395, -0.00006214, 0.00006468, -0.00009487, },
		{ -0.00007728, 0.00005180, -0.00002356, 0.00000908, -0.00002604, -0.00001187, -0.00000597, 0.00005036, -0.00004801, -0.00000983, 0.00003168, -0.00001194, -0.00000547, -0.00000553, -0.00000142, -0.00001942, 0.00004622, 0.00000834, -0.00001853, -0.00000902, 0.00002067, -0.00001320, 0.00001851, -0.00001429, 0.00003179, -0.00002477, -0.00000831, 0.00002956, -0.00003172, 0.00002420, -0.00001216, -0.00000833, -0.00002323, -0.00001620, -0.00002263, 0.00001115, 0.00000947, -0.00004356, 0.00006049, -0.00010541, 0.00007552, -0.00015337, 0.00009605, -0.00066968, 0.00127564, -0.00006370, -0.00034357, 0.00007240, -0.00011086, 0.00001941, -0.00002301, 0.00000121, -0.00001074, -0.00004908, 0.00001347, -0.00002134, 0.00010670, 0.00002239, -0.00000006, -0.00000726, -0.00005546, 0.00004639, -0.00003133, 0.00003522, },
		{ -0.00000604, -0.00004492, 0.00003461, -0.00001428, -0.00000025, -0.00002245, -0.00001840, 0.00001373, 0.00003479, -0.00005521, 0.00000662, 0.00002446, -0.00001191, -0.00000735, -0.00000084, -0.00000838, -0.00001018, 0.00005125, -0.00000444, -0.00001870, -0.00000128, 0.00001667, -0.00001030, 0.00001713, -0.00001316, 0.00003227, -0.00003434, 0.00000770, 0.00001939, -0.00003371, 0.00003647, -0.00003454, 0.00001717, -0.00005605, 0.00001714, -0.00005513, 0.00005742, -0.00005324, 0.00003012, -0.00002238, -0.00001756, -0.00002799, -0.00002449, -0.00012152, -0.00038788, 0.00130288, -0.00038528, -0.00013055, -0.00002956, -0.00004019, -0.00002557, 0.00001825, -0.00003363, 0.00001384, -0.00007647, 0.00003805, -0.00003268, 0.00015079, -0.00002674, 0.00004097, -0.00006528, 0.00000174, 0.00000127, 0.00000868, },
		{ 0.00004155, -0.00004441, -0.00001006, 0.00001210, -0.00000001, -0.00001380, -0.00001702, -0.00002263, 0.00003274, 0.00001214, -0.00005253, 0.00001870, 0.00001672, -0.00001085, -0.00000980, 0.00000522, -0.00001845, 0.00000634, 0.00004691, -0.00001299, -0.00001878, 0.00000788, 0.00000937, -0.00000354, 0.00001148, -0.00000797, 0.00002807, -0.00003970, 0.00002454, 0.00000077, -0.00002180, 0.00003165, -0.00003871, 0.00002512, -0.00007368, 0.00003824, -0.00007343, 0.00008531, -0.00009681, 0.00008677, -0.00009107, 0.00005790, -0.00011929, 0.00009112, -0.00034591, -0.00000453, 0.00113315, -0.00054953, 0.00001734, -0.00010217, 0.00000877, -0.00005049, 0.00003925, -0.00005019, 0.00001959, -0.00008497, 0.00004423, -0.00001702, 0.00016499, -0.00005083, 0.00005436, -0.00009744, 0.00004194, -0.00003333, },
		{ -0.00004615, 0.00005563, -0.00006805, 0.00001620, -0.00000650, 0.00001226, -0.00002689, -0.00001251, -0.00002160, 0.00004596, -0.00001233, -0.00004192, 0.00002519, 0.00001052, -0.00001087, -0.00001014, 0.00000907, -0.00002724, 0.00002602, 0.00003493, -0.00001694, -0.00001862, 0.00001704, 0.00000053, 0.00000554, 0.00000254, 0.00000106, 0.00001836, -0.00003790, 0.00003681, -0.00002005, -0.00000122, 0.00001277, -0.00002449, 0.00001172, -0.00006930, 0.00003929, -0.00006985, 0.00008566, -0.00010871, 0.00010996, -0.00012526, 0.00009618, -0.00016710, 0.00014890, -0.00050726, 0.00039294, 0.00082834, -0.00056825, 0.00008654, -0.00013245, 0.00002816, -0.00005097, 0.00003728, -0.00004696, 0.00000616, -0.00007528, 0.00003500, 0.00002215, 0.00015052, -0.00005034, 0.00003883, -0.00009749, 0.00005512, },
		{ 0.00004289, -0.00003599, 0.00004770, -0.00007185, 0.00002764, -0.00001515, 0.00001742, -0.00003600, -0.00001039, -0.00001417, 0.00004993, -0.00003344, -0.00002680, 0.00002659, 0.00000657, -0.00001287, -0.00000680, 0.00000796, -0.00003045, 0.00004362, 0.00001898, -0.00001751, -0.00001733, 0.00002438, -0.00000759, 0.00001453, -0.00000754, 0.00001227, 0.00000394, -0.00002755, 0.00003972, -0.00003559, 0.00001937, -0.00001137, 0.00000048, -0.00001794, -0.00004509, 0.00002114, -0.00004576, 0.00005973, -0.00008835, 0.00009657, -0.00011925, 0.00008889, -0.00015906, 0.00012590, -0.00055445, 0.00071997, 0.00046975, -0.00048342, 0.00008945, -0.00012551, 0.00002328, -0.00003341, 0.00001903, -0.00003082, -0.00001920, -0.00005504, 0.00002035, 0.00007341, 0.00011713, -0.00003523, 0.00000445, -0.00007163, },
		{ -0.00003435, 0.00001608, -0.00001071, 0.00002302, -0.00005825, 0.00002432, -0.00001255, 0.00001362, -0.00003967, -0.00001035, -0.00000167, 0.00004384, -0.00004742, -0.00001084, 0.00002451, 0.00000428, -0.00001632, 0.00000000, 0.00000108, -0.00002543, 0.00005459, 0.00000322, -0.00001663, -0.00001381, 0.00002824, -0.00001285, 0.00002091, -0.00001600, 0.00002288, -0.00001250, -0.00000986, 0.00003118, -0.00004015, 0.00003150, -0.00003013, 0.00002436, -0.00005247, -0.00001054, -0.00000795, -0.00001032, 0.00001789, -0.00004618, 0.00005643, -0.00008240, 0.00004550, -0.00010517, 0.00002841, -0.00047007, 0.00091721, 0.00013413, -0.00034844, 0.00005250, -0.00009690, 0.00000667, -0.00000960, -0.00000436, -0.00001232, -0.00004619, -0.00003418, 0.00001232, 0.00012265, 0.00007773, -0.00001919, -0.00003275, },
		{ -0.00005804, -0.00000162, -0.00001095, 0.00001723, -0.00000791, -0.00003515, 0.00001140, -0.00000174, 0.00000219, -0.00003842, -0.00001071, 0.00001267, 0.00002957, -0.00005279, 0.00000304, 0.00002086, 0.00000239, -0.00001954, 0.00000834, -0.00001001, -0.00001221, 0.00005644, -0.00000904, -0.00001601, -0.00000745, 0.00002772, -0.00001395, 0.00002284, -0.00002039, 0.00002976, -0.00002698, 0.00001126, 0.00001290, -0.00003212, 0.00003040, -0.00003547, 0.00003606, -0.00007910, 0.00002201, -0.00003640, 0.00002375, -0.00002502, 0.00000145, 0.00000695, -0.00003364, -0.00001228, -0.00003248, -0.00011121, -0.00027353, 0.00096279, -0.00012451, -0.00021089, 0.00000342, -0.00006349, -0.00000913, 0.00000942, -0.00002297, -0.00000045, -0.00006650, -0.00001998, 0.00001983, 0.00015842, 0.00004259, -0.00001309, },
		{ -0.00002050, -0.00006375, 0.00001555, -0.00002654, 0.00003627, -0.00003405, -0.00001209, -0.00000368, 0.00001184, -0.00001352, -0.00003381, -0.00000954, 0.00002496, 0.00001086, -0.00005021, 0.00001327, 0.00001713, -0.00000031, -0.00002061, 0.00001534, -0.00002196, 0.00000663, 0.00004927, -0.00001622, -0.00001645, 0.00000154, 0.00002297, -0.00001080, 0.00001979, -0.00001938, 0.00003053, -0.00003536, 0.00003023, -0.00001001, -0.00001463, 0.00001681, -0.00002514, 0.00002954, -0.00008856, 0.00004227, -0.00005422, 0.00004550, -0.00005572, 0.00003905, -0.00003442, 0.00000713, -0.00006041, 0.00002650, -0.00024659, -0.00001249, 0.00087209, -0.00028286, -0.00010178, -0.00003785, -0.00003686, -0.00001658, 0.00001771, -0.00003184, 0.00000073, -0.00007684, -0.00001409, 0.00004556, 0.00017519, 0.00001588, },
		{ -0.00000433, -0.00003705, -0.00005138, 0.00001459, -0.00002586, 0.00004011, -0.00004798, 0.00000346, -0.00001430, 0.00002270, -0.00002976, -0.00002733, -0.00000582, 0.00003216, -0.00000812, -0.00004181, 0.00001956, 0.00001398, -0.00000435, -0.00001819, 0.00001835, -0.00003081, 0.00002678, 0.00003551, -0.00001855, -0.00001764, 0.00001203, 0.00001518, -0.00000452, 0.00001265, -0.00001327, 0.00002447, -0.00003487, 0.00004164, -0.00003080, 0.00000581, -0.00000347, -0.00000331, 0.00000588, -0.00007819, 0.00004552, -0.00005662, 0.00004959, -0.00006707, 0.00005745, -0.00005680, 0.00002698, -0.00008245, 0.00004672, -0.00033354, 0.00025341, 0.00068712, -0.00034584, -0.00003290, -0.00006248, -0.00002096, -0.00001468, 0.00001537, -0.00003168, -0.00000741, -0.00007916, -0.00001247, 0.00008583, 0.00017345, },
		{ 0.00015722, -0.00002296, -0.00005360, -0.00002960, 0.00000088, -0.00001164, 0.00002918, -0.00004796, 0.00000833, -0.00001681, 0.00002707, -0.00004356, -0.00001973, -0.00000005, 0.00003303, -0.00002406, -0.00003014, 0.00002243, 0.00001136, -0.00000934, -0.00001218, 0.00001595, -0.00003340, 0.00004368, 0.00001885, -0.00001763, -0.00001845, 0.00002232, 0.00000611, 0.00000301, 0.00000344, -0.00000394, 0.00001297, -0.00002514, 0.00004211, -0.00004349, 0.00002190, -0.00002215, 0.00002127, -0.00002763, -0.00005213, 0.00003349, -0.00004507, 0.00003755, -0.00005963, 0.00005605, -0.00005879, 0.00002397, -0.00007536, 0.00001970, -0.00034416, 0.00047039, 0.00046039, -0.00033694, -0.00000096, -0.00007096, -0.00001371, -0.00000713, 0.00000661, -0.00002684, -0.00002030, -0.00007835, -0.00000783, 0.00013294, },
		{ 0.00017839, 0.00013074, -0.00004376, -0.00006128, -0.00000945, -0.00001594, 0.00000830, 0.00000926, -0.00003734, 0.00000384, -0.00001112, 0.00002357, -0.00005313, -0.00001114, 0.00000603, 0.00002839, -0.00003526, -0.00001733, 0.00002266, 0.00000887, -0.00001424, -0.00000372, 0.00000848, -0.00002831, 0.00005380, 0.00000304, -0.00001554, -0.00001746, 0.00003060, -0.00000241, 0.00000981, -0.00000540, 0.00000578, -0.00000083, -0.00000856, 0.00003146, -0.00004501, 0.00002843, -0.00003182, 0.00003890, -0.00006057, -0.00001901, 0.00001258, -0.00002564, 0.00001598, -0.00003996, 0.00004113, -0.00004707, 0.00000599, -0.00004921, -0.00004431, -0.00027422, 0.00060389, 0.00023961, -0.00028621, 0.00000520, -0.00006858, -0.00001053, 0.00000101, -0.00000351, -0.00002214, -0.00003324, -0.00007897, 0.00000699, },
		{ 0.00003642, 0.00021527, 0.00009635, -0.00006632, -0.00005589, 0.00000079, -0.00002683, 0.00002521, -0.00001177, -0.00002222, -0.00000583, 0.00000021, 0.00001283, -0.00005754, -0.00000172, 0.00001036, 0.00002034, -0.00004149, -0.00000478, 0.00002084, 0.00000620, -0.00001782, 0.00000520, -0.00000214, -0.00001629, 0.00005554, -0.00000914, -0.00001407, -0.00001358, 0.00003545, -0.00000896, 0.00001440, -0.00001184, 0.00001307, -0.00001296, 0.00001042, 0.00001279, -0.00003609, 0.00002415, -0.00002884, 0.00004260, -0.00008335, 0.00001165, -0.00000931, -0.00000583, -0.00000686, -0.00001679, 0.00002178, -0.00003176, -0.00001442, -0.00002142, -0.00012132, -0.00014196, 0.00064395, 0.00005752, -0.00022036, -0.00000225, -0.00006101, -0.00000770, 0.00000622, -0.00001170, -0.00002035, -0.00004387, -0.00008271, },
		{ -0.00008761, 0.00008115, 0.00023882, 0.00005458, -0.00008585, -0.00003968, -0.00000122, -0.00002708, 0.00003309, -0.00002742, -0.00000852, -0.00001584, 0.00001348, -0.00000334, -0.00005621, 0.00000768, 0.00001171, 0.00001119, -0.00004346, 0.00000675, 0.00001731, 0.00000337, -0.00001918, 0.00001248, -0.00001308, 0.00000018, 0.00004938, -0.00001643, -0.00001412, -0.00000639, 0.00003612, -0.00001280, 0.00001616, -0.00001487, 0.00001608, -0.00001998, 0.00002667, -0.00000868, -0.00002072, 0.00001196, -0.00001453, 0.00003046, -0.00009035, 0.00003277, -0.00002586, 0.00000848, -0.00002446, 0.00000282, 0.00000536, -0.00002114, -0.00002666, -0.00000848, -0.00018334, 0.00002011, 0.00060250, -0.00007130, -0.00015748, -0.00001426, -0.00005184, -0.00000389, 0.00000764, -0.00001737, -0.00002149, -0.00005285, },
		{ -0.00006287, -0.00008894, 0.00013869, 0.00024556, 0.00000622, -0.00009574, -0.00001979, -0.00001189, -0.00001747, 0.00003056, -0.00003461, 0.00000004, -0.00002252, 0.00002518, -0.00002271, -0.00004870, 0.00001521, 0.00001037, 0.00000238, -0.00004187, 0.00001664, 0.00001236, 0.00000073, -0.00001809, 0.00001660, -0.00002152, 0.00001765, 0.00003743, -0.00001907, -0.00001562, 0.00000372, 0.00003253, -0.00001381, 0.00001528, -0.00001463, 0.00001448, -0.00002001, 0.00003600, -0.00002745, -0.00000438, -0.00000252, 0.00000557, 0.00000588, -0.00008119, 0.00004141, -0.00003414, 0.00001477, -0.00003396, 0.00001570, -0.00000515, -0.00001819, -0.00002677, -0.00001885, -0.00020849, 0.00017663, 0.00050498, -0.00014748, -0.00010649, -0.00002588, -0.00004231, 0.00000020, 0.00000641, -0.00002187, -0.00002367, },
		{ -0.00002493, -0.00007681, -0.00008083, 0.00020450, 0.00023220, -0.00004520, -0.00009149, -0.00000437, -0.00002445, -0.00000273, 0.00002020, -0.00003381, 0.00000252, -0.00002407, 0.00003266, -0.00004285, -0.00003502, 0.00001862, 0.00000802, -0.00000588, -0.00003696, 0.00002404, 0.00000658, -0.00000133, -0.00001495, 0.00001700, -0.00002545, 0.00003278, 0.00002264, -0.00001837, -0.00001764, 0.00001565, 0.00002519, -0.00001227, 0.00001248, -0.00001222, 0.00000947, -0.00001334, 0.00003641, -0.00003951, 0.00000812, -0.00001342, 0.00002411, -0.00002412, -0.00005992, 0.00003866, -0.00003454, 0.00001375, -0.00003594, 0.00002248, -0.00001112, -0.00002033, -0.00001854, -0.00004988, -0.00018779, 0.00029957, 0.00038041, -0.00018087, -0.00006958, -0.00003509, -0.00003252, 0.00000327, 0.00000415, -0.00002678, },
		{ -0.00003261, -0.00002464, -0.00009611, -0.00005770, 0.00027254, 0.00019563, -0.00009236, -0.00007366, 0.00000139, -0.00003268, 0.00001145, 0.00000643, -0.00002763, -0.00000002, -0.00002033, 0.00003411, -0.00006096, -0.00001637, 0.00001630, 0.00000680, -0.00001423, -0.00002851, 0.00002779, 0.00000101, -0.00000265, -0.00001053, 0.00001407, -0.00002410, 0.00004306, 0.00000793, -0.00001612, -0.00001880, 0.00002781, 0.00001505, -0.00000865, 0.00000856, -0.00000906, 0.00000325, -0.00000216, 0.00002844, -0.00004340, 0.00001429, -0.00001688, 0.00003472, -0.00005164, -0.00003295, 0.00002815, -0.00002949, 0.00000787, -0.00003274, 0.00002555, -0.00001596, -0.00002216, -0.00001061, -0.00008973, -0.00012638, 0.00037396, 0.00025356, -0.00018435, -0.00004542, -0.00004117, -0.00002273, 0.00000473, 0.00000160, },
		{ -0.00000168, -0.00003843, -0.00002396, -0.00012013, -0.00001497, 0.00033488, 0.00013466, -0.00012562, -0.00004844, -0.00000248, -0.00003381, 0.00002124, -0.00000690, -0.00001892, -0.00000583, -0.00001197, 0.00002831, -0.00007376, 0.00000439, 0.00000845, 0.00000806, -0.00002326, -0.00001639, 0.00002680, -0.00000296, -0.00000352, -0.00000560, 0.00000887, -0.00001790, 0.00004725, -0.00000440, -0.00001399, -0.00001765, 0.00003833, 0.00000342, -0.00000347, 0.00000410, -0.00000633, -0.00000183, 0.00001026, 0.00001469, -0.00004019, 0.00001453, -0.00001238, 0.00003413, -0.00007052, -0.00000659, 0.00001410, -0.00002176, -0.00000040, -0.00002648, 0.00002672, -0.00002210, -0.00001932, -0.00001151, -0.00012291, -0.00003977, 0.00039836, 0.00014094, -0.00016975, -0.00003155, -0.00004372, -0.00001377, 0.00000499, },
		},
		{ { -0.00005222, -0.00003250, -0.00013401, 0.00005238, 0.00036791, 0.00003664, -0.00010580, -0.00002757, -0.00003021, -0.00001447, -0.00003617, 0.00001623, -0.00007422, 0.00001173, 0.00001859, -0.00000234, -0.00000893, -0.00001435, 0.00002059, -0.00001713, 0.00002528, -0.00001366, 0.00001556, -0.00000647, 0.00000064, 0.00000726, 0.00000626, 0.00000586, -0.00001188, -0.00000191, -0.00002534, 0.00000793, 0.00001096, 0.00000793, -0.00002534, -0.00000191, -0.00001188, 0.00000586, 0.00000626, 0.00000726, 0.00000064, -0.00000647, 0.00001556, -0.00001366, 0.00002528, -0.00001713, 0.00002059, -0.00001435, -0.00000893, -0.00000234, 0.00001859, 0.00001173, -0.00007422, 0.00001623, -0.00003617, -0.00001447, -0.00003021, -0.00002757, -0.00010580, 0.00003664, 0.00036791, 0.00005238, -0.00013401, -0.00003250, },
		{ -0.00003211, -0.00005205, -0.00003390, -0.00015225, 0.00010451, 0.00039731, -0.00002994, -0.00009583, -0.00002632, -0.00002682, -0.00001832, -0.00003224, 0.00001183, -0.00007757, 0.00002728, 0.00000475, 0.00000443, -0.00001538, -0.00000516, 0.00001536, -0.00001585, 0.00002706, -0.00001589, 0.00001793, -0.00001180, 0.00000698, 0.00000398, 0.00000850, 0.00000118, -0.00000878, -0.00000682, -0.00002135, 0.00001034, 0.00001138, 0.00000371, -0.00002713, 0.00000202, -0.00001446, 0.00001032, 0.00000430, 0.00000953, -0.00000536, 0.00000005, 0.00001153, -0.00000990, 0.00002128, -0.00001510, 0.00002198, -0.00002065, -0.00000479, -0.00000615, 0.00003088, -0.00000612, -0.00006621, 0.00001685, -0.00003732, -0.00001328, -0.00003049, -0.00003491, -0.00010178, 0.00009316, 0.00032777, 0.00001395, -0.00011755, },
		{ -0.00010349, -0.00003219, -0.00005164, -0.00003677, -0.00017141, 0.00017231, 0.00040915, -0.00010057, -0.00007278, -0.00003282, -0.00001859, -0.00002626, -0.00002435, 0.00000248, -0.00007437, 0.00003769, -0.00000771, 0.00001197, -0.00002262, 0.00000572, 0.00000693, -0.00001124, 0.00002629, -0.00001623, 0.00001818, -0.00001506, 0.00001263, 0.00000028, 0.00001071, -0.00000337, -0.00000562, -0.00001207, -0.00001568, 0.00001108, 0.00001249, -0.00000218, -0.00002647, 0.00000452, -0.00001614, 0.00001433, 0.00000272, 0.00001041, -0.00001015, 0.00000680, 0.00000655, -0.00000525, 0.00001578, -0.00001034, 0.00001959, -0.00002360, -0.00000370, -0.00000557, 0.00003907, -0.00002333, -0.00005584, 0.00001535, -0.00003731, -0.00001277, -0.00003037, -0.00004486, -0.00008573, 0.00013652, 0.00028241, -0.00001298, },
		{ -0.00003039, -0.00009226, -0.00003217, -0.00005117, -0.00004171, -0.00018988, 0.00025685, 0.00039624, -0.00016675, -0.00004075, -0.00004624, -0.00000554, -0.00003842, -0.00001243, -0.00001193, -0.00006382, 0.00004093, -0.00001610, 0.00001793, -0.00002876, 0.00001658, -0.00000346, -0.00000383, 0.00002304, -0.00001465, 0.00001613, -0.00001562, 0.00001662, -0.00000305, 0.00001244, -0.00000734, -0.00000293, -0.00001695, -0.00000904, 0.00001057, 0.00001393, -0.00000933, -0.00002339, 0.00000530, -0.00001666, 0.00001773, 0.00000145, 0.00000985, -0.00001310, 0.00001284, 0.00000145, -0.00000050, 0.00000979, -0.00000394, 0.00001420, -0.00002357, -0.00000544, -0.00000029, 0.00004167, -0.00003762, -0.00004514, 0.00001317, -0.00003747, -0.00001122, -0.00003241, -0.00005330, -0.00006161, 0.00016625, 0.00023575, },
		{ 0.00019038, -0.00004008, -0.00008402, -0.00003171, -0.00005063, -0.00004974, -0.00020479, 0.00035711, 0.00035244, -0.00021874, -0.00000649, -0.00006318, 0.00001030, -0.00005331, 0.00000218, -0.00003005, -0.00004664, 0.00003637, -0.00001867, 0.00002040, -0.00003202, 0.00002560, -0.00001426, 0.00000544, 0.00001777, -0.00001147, 0.00001201, -0.00001326, 0.00001817, -0.00000521, 0.00001314, -0.00001024, -0.00000125, -0.00002073, -0.00000225, 0.00000937, 0.00001514, -0.00001709, -0.00001821, 0.00000427, -0.00001582, 0.00002040, 0.00000033, 0.00000805, -0.00001388, 0.00001742, -0.00000300, 0.00000351, 0.00000442, 0.00000273, 0.00000710, -0.00002161, -0.00000886, 0.00000879, 0.00003847, -0.00004777, -0.00003534, 0.00001103, -0.00003829, -0.00000799, -0.00003813, -0.00005674, -0.00003403, 0.00018368, },
		{ 0.00019085, 0.00014809, -0.00004395, -0.00007838, -0.00003098, -0.00004946, -0.00006271, -0.00021147, 0.00046899, 0.00027444, -0.00024734, 0.00002189, -0.00007840, 0.00002517, -0.00006792, 0.00001676, -0.00004913, -0.00002520, 0.00002504, -0.00001509, 0.00001839, -0.00003120, 0.00003122, -0.00002385, 0.00001537, 0.00001119, -0.00000727, 0.00000641, -0.00000817, 0.00001685, -0.00000559, 0.00001230, -0.00001161, -0.00000103, -0.00002273, 0.00000388, 0.00000809, 0.00001551, -0.00002464, -0.00001146, 0.00000153, -0.00001350, 0.00002220, -0.00000083, 0.00000541, -0.00001251, 0.00002005, -0.00000617, 0.00000610, 0.00000066, 0.00000827, -0.00000019, -0.00001919, -0.00001220, 0.00001976, 0.00003057, -0.00005377, -0.00002671, 0.00000873, -0.00003913, -0.00000383, -0.00004724, -0.00005339, -0.00000699, },
		{ 0.00001706, 0.00018957, 0.00011042, -0.00004419, -0.00007426, -0.00003090, -0.00004643, -0.00008333, -0.00020327, 0.00058447, 0.00016335, -0.00024598, 0.00003674, -0.00008608, 0.00003454, -0.00007844, 0.00002782, -0.00006546, -0.00000312, 0.00000958, -0.00000651, 0.00001212, -0.00002598, 0.00003249, -0.00003088, 0.00002479, 0.00000414, -0.00000279, 0.00000023, -0.00000105, 0.00001273, -0.00000386, 0.00000964, -0.00001118, -0.00000257, -0.00002250, 0.00000866, 0.00000730, 0.00001446, -0.00003111, -0.00000385, -0.00000266, -0.00000967, 0.00002297, -0.00000216, 0.00000242, -0.00000927, 0.00002054, -0.00000771, 0.00000683, -0.00000075, 0.00001149, -0.00000624, -0.00001777, -0.00001352, 0.00003016, 0.00001999, -0.00005662, -0.00001864, 0.00000528, -0.00003842, -0.00000082, -0.00005764, -0.00004364, },
		{ -0.00002996, 0.00003755, 0.00018094, 0.00007903, -0.00004338, -0.00006991, -0.00003303, -0.00003978, -0.00011480, -0.00017199, 0.00069161, 0.00002573, -0.00021246, 0.00003269, -0.00008138, 0.00003426, -0.00008124, 0.00003192, -0.00007525, 0.00001541, -0.00000637, 0.00000465, 0.00000294, -0.00001704, 0.00002925, -0.00003449, 0.00003269, -0.00000259, 0.00000120, -0.00000549, 0.00000701, 0.00000640, -0.00000015, 0.00000517, -0.00000892, -0.00000588, -0.00001983, 0.00001164, 0.00000737, 0.00001158, -0.00003571, 0.00000379, -0.00000789, -0.00000436, 0.00002248, -0.00000363, -0.00000042, -0.00000467, 0.00001902, -0.00000755, 0.00000563, 0.00000046, 0.00001169, -0.00001003, -0.00001843, -0.00001126, 0.00003762, 0.00000908, -0.00005784, -0.00001014, -0.00000057, -0.00003425, -0.00000175, -0.00006583, },
		{ -0.00006793, -0.00001612, 0.00005552, 0.00016524, 0.00005562, -0.00004422, -0.00006323, -0.00003915, -0.00002781, -0.00016003, -0.00010900, 0.00077539, -0.00012626, -0.00015013, 0.00000822, -0.00006192, 0.00002190, -0.00007398, 0.00002671, -0.00007555, 0.00002660, -0.00001894, 0.00001530, -0.00000693, -0.00000589, 0.00002210, -0.00003437, 0.00003837, -0.00000835, 0.00000403, -0.00000969, 0.00001467, -0.00000112, 0.00000495, -0.00000071, -0.00000513, -0.00001069, -0.00001488, 0.00001266, 0.00000843, 0.00000669, -0.00003781, 0.00001066, -0.00001360, 0.00000223, 0.00002053, -0.00000512, -0.00000271, 0.00000066, 0.00001589, -0.00000594, 0.00000279, 0.00000405, 0.00000879, -0.00001122, -0.00002148, -0.00000466, 0.00004042, -0.00000010, -0.00005882, -0.00000042, -0.00000974, -0.00002508, -0.00000916, },
		{ -0.00002439, -0.00006089, -0.00000601, 0.00007276, 0.00014244, 0.00004154, -0.00004903, -0.00005249, -0.00005055, -0.00000974, -0.00022030, -0.00000698, 0.00081960, -0.00027602, -0.00006788, -0.00003361, -0.00002873, -0.00000236, -0.00005644, 0.00001177, -0.00006517, 0.00002805, -0.00002497, 0.00002247, -0.00001502, 0.00000543, 0.00001228, -0.00003078, 0.00004151, -0.00001273, 0.00000527, -0.00001148, 0.00002052, -0.00000846, 0.00001048, -0.00000722, -0.00000046, -0.00001638, -0.00000812, 0.00001191, 0.00001032, -0.00000006, -0.00003703, 0.00001598, -0.00001914, 0.00000976, 0.00001699, -0.00000640, -0.00000414, 0.00000603, 0.00001175, -0.00000343, -0.00000102, 0.00000930, 0.00000331, -0.00001016, -0.00002641, 0.00000593, 0.00003797, -0.00000635, -0.00006035, 0.00001063, -0.00002215, -0.00001062, },
		{ 0.00000774, -0.00004671, -0.00004360, -0.00000240, 0.00009079, 0.00011291, 0.00003717, -0.00005897, -0.00003704, -0.00006719, 0.00001325, -0.00029385, 0.00013780, 0.00080943, -0.00040460, 0.00002097, -0.00008535, 0.00001358, -0.00003536, -0.00003099, -0.00001096, -0.00004524, 0.00001942, -0.00002282, 0.00002402, -0.00001923, 0.00001489, 0.00000134, -0.00002438, 0.00004211, -0.00001561, 0.00000481, -0.00001033, 0.00002335, -0.00001420, 0.00001520, -0.00001334, 0.00000420, -0.00002209, -0.00000032, 0.00000983, 0.00001262, -0.00000822, -0.00003331, 0.00001914, -0.00002378, 0.00001772, 0.00001189, -0.00000721, -0.00000456, 0.00001078, 0.00000733, -0.00000071, -0.00000493, 0.00001513, -0.00000369, -0.00000780, -0.00003191, 0.00001907, 0.00003093, -0.00000963, -0.00006220, 0.00002204, -0.00003639, },
		{ -0.00004971, 0.00002677, -0.00007308, -0.00001755, -0.00000593, 0.00010985, 0.00007817, 0.00004137, -0.00007354, -0.00001799, -0.00008705, 0.00003731, -0.00037467, 0.00032346, 0.00073450, -0.00049377, 0.00010081, -0.00013651, 0.00005749, -0.00007138, -0.00000229, -0.00003730, -0.00001924, 0.00000270, -0.00001286, 0.00001928, -0.00001835, 0.00002089, -0.00000918, -0.00001605, 0.00004042, -0.00001707, 0.00000283, -0.00000616, 0.00002236, -0.00001709, 0.00001791, -0.00001793, 0.00000783, -0.00002682, 0.00000757, 0.00000710, 0.00001470, -0.00001706, -0.00002690, 0.00001972, -0.00002681, 0.00002540, 0.00000546, -0.00000732, -0.00000393, 0.00001435, 0.00000336, 0.00000144, -0.00000795, 0.00002026, -0.00001087, -0.00000549, -0.00003617, 0.00003250, 0.00002114, -0.00001102, -0.00006320, 0.00003184, },
		{ 0.00003743, -0.00005854, 0.00004192, -0.00009840, 0.00001310, -0.00001467, 0.00012832, 0.00004128, 0.00005114, -0.00009026, 0.00000158, -0.00010587, 0.00005561, -0.00045188, 0.00054086, 0.00059166, -0.00052933, 0.00015667, -0.00017560, 0.00009408, -0.00010320, 0.00002355, -0.00006152, 0.00000755, -0.00001801, 0.00000249, 0.00000925, -0.00001242, 0.00002263, -0.00001811, -0.00000669, 0.00003682, -0.00001737, -0.00000014, 0.00000058, 0.00001732, -0.00001629, 0.00001761, -0.00001998, 0.00000949, -0.00002961, 0.00001457, 0.00000443, 0.00001587, -0.00002568, -0.00001838, 0.00001758, -0.00002760, 0.00003197, -0.00000182, -0.00000664, -0.00000231, 0.00001632, 0.00000049, 0.00000232, -0.00000917, 0.00002347, -0.00001686, -0.00000457, -0.00003732, 0.00004360, 0.00001107, -0.00001238, -0.00006151, },
		{ -0.00005525, 0.00003629, -0.00005925, 0.00004825, -0.00011644, 0.00004222, -0.00002444, 0.00014283, 0.00000677, 0.00006187, -0.00010491, 0.00001687, -0.00011758, 0.00005892, -0.00051016, 0.00077315, 0.00038687, -0.00050396, 0.00017709, -0.00019256, 0.00011494, -0.00012372, 0.00004033, -0.00007760, 0.00002906, -0.00003735, 0.00001926, -0.00000352, -0.00000281, 0.00002029, -0.00002477, 0.00000299, 0.00003173, -0.00001686, -0.00000339, 0.00000897, 0.00000868, -0.00001160, 0.00001378, -0.00001884, 0.00000847, -0.00002970, 0.00001983, 0.00000248, 0.00001549, -0.00003310, -0.00000856, 0.00001286, -0.00002568, 0.00003658, -0.00000929, -0.00000520, 0.00000017, 0.00001647, -0.00000081, 0.00000143, -0.00000794, 0.00002381, -0.00002056, -0.00000608, -0.00003391, 0.00005002, 0.00000331, -0.00001572, },
		{ -0.00002261, -0.00004317, 0.00002679, -0.00004924, 0.00004165, -0.00012128, 0.00006300, -0.00002976, 0.00014888, -0.00002003, 0.00006807, -0.00011238, 0.00002228, -0.00011547, 0.00003707, -0.00053135, 0.00099675, 0.00013579, -0.00041927, 0.00015668, -0.00018113, 0.00011417, -0.00012764, 0.00004327, -0.00008060, 0.00003984, -0.00004983, 0.00003279, -0.00001557, 0.00000817, 0.00001492, -0.00002910, 0.00001248, 0.00002550, -0.00001586, -0.00000611, 0.00001769, -0.00000245, -0.00000352, 0.00000652, -0.00001434, 0.00000451, -0.00002668, 0.00002279, 0.00000170, 0.00001310, -0.00003844, 0.00000158, 0.00000603, -0.00002088, 0.00003847, -0.00001615, -0.00000316, 0.00000329, 0.00001478, -0.00000033, -0.00000146, -0.00000399, 0.00002080, -0.00002136, -0.00001039, -0.00002532, 0.00005023, -0.00000019, },
		{ 0.00000139, -0.00003353, -0.00002527, 0.00000888, -0.00002784, 0.00002007, -0.00010874, 0.00006959, -0.00002534, 0.00014211, -0.00003410, 0.00006456, -0.00010792, 0.00001290, -0.00009384, -0.00001882, -0.00049718, 0.00118380, -0.00013740, -0.00028618, 0.00009746, -0.00014043, 0.00008996, -0.00011289, 0.00003034, -0.00006804, 0.00003634, -0.00005117, 0.00003882, -0.00002328, 0.00001770, 0.00000815, -0.00003148, 0.00002153, 0.00001840, -0.00001463, -0.00000753, 0.00002527, -0.00001445, 0.00000676, -0.00000340, -0.00000695, -0.00000214, -0.00002061, 0.00002324, 0.00000221, 0.00000858, -0.00004100, 0.00001101, -0.00000218, -0.00001337, 0.00003711, -0.00002162, -0.00000077, 0.00000680, 0.00001148, 0.00000183, -0.00000617, 0.00000249, 0.00001448, -0.00001927, -0.00001715, -0.00001201, 0.00004394, },
		{ 0.00003227, 0.00000748, -0.00004760, -0.00000313, -0.00001546, 0.00000311, -0.00001554, -0.00007772, 0.00005866, -0.00000761, 0.00011965, -0.00003209, 0.00004785, -0.00008864, -0.00001396, -0.00005001, -0.00011377, -0.00039269, 0.00130593, -0.00040287, -0.00012369, 0.00000884, -0.00007553, 0.00004526, -0.00008136, 0.00000304, -0.00004070, 0.00001798, -0.00003939, 0.00003461, -0.00002379, 0.00002323, 0.00000175, -0.00003254, 0.00003001, 0.00001067, -0.00001333, -0.00000706, 0.00003023, -0.00002544, 0.00001758, -0.00001463, 0.00000233, -0.00001071, -0.00001201, 0.00002139, 0.00000382, 0.00000214, -0.00004041, 0.00001875, -0.00001083, -0.00000367, 0.00003232, -0.00002505, 0.00000171, 0.00001037, 0.00000704, 0.00000525, -0.00001221, 0.00001087, 0.00000545, -0.00001489, -0.00002521, 0.00000448, },
		{ 0.00002173, 0.00001752, 0.00001616, -0.00006262, 0.00002017, -0.00004233, 0.00003916, -0.00006094, -0.00003093, 0.00003051, 0.00002397, 0.00008137, -0.00001346, 0.00001735, -0.00005471, -0.00005732, 0.00001402, -0.00024641, -0.00020995, 0.00133880, -0.00062925, 0.00004391, -0.00009385, 0.00000329, -0.00001248, -0.00003886, -0.00003357, -0.00000286, -0.00001245, -0.00001545, 0.00001965, -0.00001580, 0.00002305, -0.00000285, -0.00003285, 0.00003776, 0.00000257, -0.00001200, -0.00000441, 0.00003140, -0.00003353, 0.00002699, -0.00002546, 0.00001207, -0.00002002, -0.00000184, 0.00001782, 0.00000602, -0.00000566, -0.00003667, 0.00002403, -0.00001885, 0.00000729, 0.00002438, -0.00002606, 0.00000408, 0.00001357, 0.00000211, 0.00000923, -0.00001871, 0.00002011, -0.00000522, -0.00000930, -0.00003292, },
		{ -0.00003843, 0.00003694, 0.00000278, 0.00002460, -0.00007545, 0.00004057, -0.00006644, 0.00007385, -0.00010898, 0.00002519, -0.00001054, 0.00006629, 0.00003055, 0.00001884, -0.00002399, -0.00001014, -0.00011179, 0.00009050, -0.00040734, 0.00004891, 0.00126657, -0.00078851, 0.00019102, -0.00019212, 0.00008236, -0.00007269, 0.00000597, -0.00007169, 0.00003830, -0.00004908, 0.00001667, -0.00000395, 0.00000003, 0.00001682, -0.00000494, -0.00003270, 0.00004452, -0.00000557, -0.00001061, 0.00000037, 0.00002810, -0.00003718, 0.00003318, -0.00003407, 0.00002061, -0.00002862, 0.00000871, 0.00001333, 0.00000808, -0.00001396, -0.00003016, 0.00002634, -0.00002519, 0.00001835, 0.00001399, -0.00002462, 0.00000625, 0.00001595, -0.00000250, 0.00001289, -0.00002462, 0.00002894, -0.00001621, -0.00000376, },
		{ 0.00000049, -0.00004014, 0.00004745, -0.00000874, 0.00002967, -0.00008276, 0.00005379, -0.00008214, 0.00009986, -0.00015080, 0.00008142, -0.00005674, 0.00011282, -0.00002624, 0.00005851, -0.00006975, 0.00003745, -0.00016775, 0.00016613, -0.00057870, 0.00036989, 0.00108556, -0.00086060, 0.00029509, -0.00026779, 0.00014733, -0.00012373, 0.00004336, -0.00010231, 0.00007409, -0.00008404, 0.00005063, -0.00003181, 0.00002107, 0.00000565, -0.00000469, -0.00003195, 0.00004980, -0.00001327, -0.00000910, 0.00000683, 0.00002026, -0.00003548, 0.00003472, -0.00003889, 0.00002645, -0.00003511, 0.00001837, 0.00000881, 0.00000924, -0.00002172, -0.00002160, 0.00002552, -0.00002895, 0.00002819, 0.00000221, -0.00002103, 0.00000828, 0.00001704, -0.00000596, 0.00001524, -0.00002881, 0.00003602, -0.00002614, },
		{ -0.00003378, 0.00000257, -0.00003703, 0.00005132, -0.00001443, 0.00002877, -0.00008181, 0.00005638, -0.00008463, 0.00011044, -0.00017744, 0.00012742, -0.00009830, 0.00015471, -0.00008011, 0.00009683, -0.00011109, 0.00007789, -0.00021268, 0.00022354, -0.00073547, 0.00072719, 0.00080618, -0.00083699, 0.00034053, -0.00030647, 0.00018604, -0.00015538, 0.00006459, -0.00011718, 0.00009603, -0.00010896, 0.00007898, -0.00005805, 0.00004319, -0.00000808, -0.00000314, -0.00003001, 0.00005293, -0.00001994, -0.00000740, 0.00001423, 0.00000853, -0.00002830, 0.00003090, -0.00003887, 0.00002841, -0.00003832, 0.00002603, 0.00000504, 0.00000882, -0.00002793, -0.00001197, 0.00002180, -0.00002953, 0.00003554, -0.00000964, -0.00001589, 0.00001031, 0.00001640, -0.00000750, 0.00001542, -0.00003024, 0.00004013, },
		{ 0.00004037, -0.00003827, 0.00000209, -0.00002899, 0.00004780, -0.00001285, 0.00002043, -0.00007127, 0.00004662, -0.00007118, 0.00010101, -0.00018177, 0.00015376, -0.00012537, 0.00018261, -0.00012154, 0.00012438, -0.00013850, 0.00010046, -0.00023343, 0.00024402, -0.00084844, 0.00108541, 0.00045270, -0.00072253, 0.00032156, -0.00030044, 0.00019109, -0.00016108, 0.00006395, -0.00011068, 0.00009780, -0.00011683, 0.00009473, -0.00007654, 0.00006161, -0.00002123, -0.00000190, -0.00002605, 0.00005314, -0.00002493, -0.00000548, 0.00002156, -0.00000582, -0.00001637, 0.00002184, -0.00003372, 0.00002594, -0.00003755, 0.00003093, 0.00000254, 0.00000643, -0.00003172, -0.00000235, 0.00001575, -0.00002672, 0.00003939, -0.00002023, -0.00000997, 0.00001255, 0.00001379, -0.00000661, 0.00001277, -0.00002819, },
		{ -0.00002234, 0.00003632, -0.00003921, -0.00000071, -0.00001687, 0.00003748, -0.00000407, 0.00000481, -0.00005165, 0.00002515, -0.00004196, 0.00007040, -0.00016013, 0.00015402, -0.00013022, 0.00018870, -0.00014238, 0.00013304, -0.00014397, 0.00009630, -0.00021916, 0.00021118, -0.00088855, 0.00140370, 0.00006076, -0.00053500, 0.00024322, -0.00025015, 0.00016145, -0.00013941, 0.00004026, -0.00008134, 0.00007676, -0.00010362, 0.00009287, -0.00008222, 0.00007192, -0.00003063, -0.00000268, -0.00001925, 0.00004981, -0.00002762, -0.00000336, 0.00002773, -0.00002105, -0.00000121, 0.00000853, -0.00002393, 0.00001920, -0.00003271, 0.00003276, 0.00000147, 0.00000205, -0.00003255, 0.00000618, 0.00000823, -0.00002075, 0.00003912, -0.00002838, -0.00000407, 0.00001511, 0.00000918, -0.00000309, 0.00000704, },
		{ -0.00000153, -0.00001289, 0.00002807, -0.00003668, -0.00000508, -0.00000239, 0.00002228, 0.00001025, -0.00001625, -0.00002545, -0.00000489, -0.00000039, 0.00002142, -0.00011346, 0.00012627, -0.00010909, 0.00016866, -0.00013782, 0.00011805, -0.00012311, 0.00006082, -0.00016423, 0.00011482, -0.00083200, 0.00164137, -0.00032732, -0.00030253, 0.00012044, -0.00016422, 0.00010277, -0.00009464, -0.00000264, -0.00003247, 0.00003487, -0.00006937, 0.00007163, -0.00007230, 0.00007113, -0.00003386, -0.00000680, -0.00000915, 0.00004264, -0.00002759, -0.00000107, 0.00003167, -0.00003516, 0.00001513, -0.00000730, -0.00001080, 0.00000906, -0.00002431, 0.00003171, 0.00000157, -0.00000393, -0.00003028, 0.00001275, 0.00000029, -0.00001234, 0.00003466, -0.00003328, 0.00000117, 0.00001797, 0.00000287, 0.00000281, },
		{ 0.00001039, -0.00001220, -0.00000061, 0.00001628, -0.00003124, -0.00000989, 0.00001216, 0.00000510, 0.00002719, -0.00003949, 0.00000324, -0.00003829, 0.00004722, -0.00003925, -0.00004747, 0.00007386, -0.00006328, 0.00012296, -0.00010770, 0.00007939, -0.00007673, -0.00000453, -0.00007051, -0.00004575, -0.00066512, 0.00176406, -0.00066843, -0.00005920, -0.00002467, -0.00005779, 0.00002634, -0.00003590, -0.00005642, 0.00002817, -0.00002144, -0.00001849, 0.00003308, -0.00004698, 0.00005834, -0.00002983, -0.00001473, 0.00000412, 0.00003184, -0.00002473, 0.00000131, 0.00003253, -0.00004622, 0.00003032, -0.00002349, 0.00000386, -0.00000305, -0.00001341, 0.00002840, 0.00000227, -0.00001072, -0.00002523, 0.00001676, -0.00000700, -0.00000255, 0.00002651, -0.00003459, 0.00000537, 0.00002087, -0.00000450, },
		{ -0.00001198, 0.00001855, -0.00002375, 0.00001324, 0.00000207, -0.00002375, -0.00001390, 0.00002437, -0.00001077, 0.00004317, -0.00006091, 0.00002943, -0.00006859, 0.00009274, -0.00010204, 0.00002808, 0.00000502, 0.00000077, 0.00005716, -0.00005702, 0.00002232, -0.00001136, -0.00009155, 0.00005156, -0.00025968, -0.00038818, 0.00174943, -0.00092500, 0.00016042, -0.00016686, 0.00005040, -0.00005311, 0.00002454, -0.00010982, 0.00008973, -0.00008223, 0.00004086, -0.00001705, -0.00000964, 0.00003511, -0.00001914, -0.00002587, 0.00001962, 0.00001827, -0.00001935, 0.00000372, 0.00002979, -0.00005261, 0.00004215, -0.00003774, 0.00001794, -0.00001535, -0.00000139, 0.00002373, 0.00000279, -0.00001729, -0.00001812, 0.00001798, -0.00001265, 0.00000731, 0.00001566, -0.00003252, 0.00000848, 0.00002333, },
		{ 0.00002473, -0.00001843, 0.00002591, -0.00003466, 0.00002704, -0.00001311, -0.00001523, -0.00001605, 0.00003216, -0.00002222, 0.00005466, -0.00007657, 0.00004828, -0.00008944, 0.00012765, -0.00015623, 0.00010105, -0.00006846, 0.00007249, -0.00001888, 0.00000481, -0.00004333, 0.00006145, -0.00018607, 0.00018362, -0.00050456, -0.00001720, 0.00159129, -0.00107069, 0.00032680, -0.00028230, 0.00014162, -0.00012039, 0.00007387, -0.00015090, 0.00014034, -0.00013591, 0.00009816, -0.00007035, 0.00003376, 0.00000521, -0.00000400, -0.00003856, 0.00003570, 0.00000338, -0.00001214, 0.00000611, 0.00002335, -0.00005330, 0.00004884, -0.00004803, 0.00002946, -0.00002604, 0.00001025, 0.00001869, 0.00000236, -0.00002252, -0.00000998, 0.00001659, -0.00001596, 0.00001588, 0.00000350, -0.00002773, 0.00001074, },
		{ 0.00001262, 0.00002436, -0.00002270, 0.00003103, -0.00004329, 0.00003904, -0.00002763, -0.00000670, -0.00001568, 0.00003420, -0.00002696, 0.00005890, -0.00008344, 0.00005612, -0.00009581, 0.00014475, -0.00019187, 0.00015908, -0.00013345, 0.00013915, -0.00009273, 0.00006562, -0.00010486, 0.00012704, -0.00027008, 0.00030175, -0.00074845, 0.00041669, 0.00130143, -0.00109421, 0.00041999, -0.00035277, 0.00020074, -0.00016292, 0.00010135, -0.00016957, 0.00016950, -0.00017150, 0.00014256, -0.00011725, 0.00007572, -0.00002611, 0.00001218, -0.00005033, 0.00005018, -0.00001094, -0.00000415, 0.00000845, 0.00001356, -0.00004798, 0.00004938, -0.00005290, 0.00003682, -0.00003361, 0.00002019, 0.00001411, 0.00000036, -0.00002540, -0.00000196, 0.00001312, -0.00001654, 0.00002195, -0.00000843, -0.00002121, },
		{ -0.00001410, 0.00001463, 0.00002165, -0.00002381, 0.00003265, -0.00004812, 0.00004757, -0.00003994, 0.00000099, -0.00001267, 0.00003017, -0.00002396, 0.00005450, -0.00008006, 0.00005127, -0.00008520, 0.00013966, -0.00020180, 0.00019199, -0.00017796, 0.00018834, -0.00015182, 0.00011290, -0.00014925, 0.00017044, -0.00032501, 0.00038028, -0.00095406, 0.00087052, 0.00090866, -0.00100093, 0.00043254, -0.00036872, 0.00021902, -0.00017317, 0.00010069, -0.00015978, 0.00017031, -0.00018089, 0.00016503, -0.00014891, 0.00010866, -0.00005311, 0.00002554, -0.00005847, 0.00006077, -0.00002269, 0.00000339, 0.00001068, 0.00000124, -0.00003716, 0.00004361, -0.00005173, 0.00003911, -0.00003705, 0.00002747, 0.00001057, -0.00000346, -0.00002527, 0.00000480, 0.00000836, -0.00001445, 0.00002467, -0.00001866, },
		{ -0.00002604, -0.00000742, 0.00001721, 0.00001630, -0.00002119, 0.00002994, -0.00004806, 0.00005130, -0.00004872, 0.00000724, -0.00000749, 0.00002086, -0.00001364, 0.00004172, -0.00006691, 0.00003449, -0.00005823, 0.00011187, -0.00018314, 0.00019377, -0.00019351, 0.00021044, -0.00018597, 0.00013641, -0.00016598, 0.00017956, -0.00033549, 0.00039650, -0.00108425, 0.00129491, 0.00045515, -0.00081165, 0.00037010, -0.00033067, 0.00019563, -0.00015019, 0.00007138, -0.00012093, 0.00014105, -0.00016055, 0.00016015, -0.00015895, 0.00012645, -0.00007079, 0.00003257, -0.00006060, 0.00006553, -0.00003010, 0.00000931, 0.00001272, -0.00001245, -0.00002207, 0.00003229, -0.00004477, 0.00003623, -0.00003600, 0.00003163, 0.00000822, -0.00000891, -0.00002187, 0.00000938, 0.00000327, -0.00001017, 0.00002366, },
		{ 0.00001909, -0.00002990, -0.00000195, 0.00002052, 0.00000842, -0.00001479, 0.00002264, -0.00004263, 0.00004947, -0.00005306, 0.00001174, -0.00000110, 0.00000799, 0.00000224, 0.00002239, -0.00004634, 0.00000894, -0.00001869, 0.00006498, -0.00013804, 0.00016382, -0.00017686, 0.00020067, -0.00018953, 0.00013046, -0.00014945, 0.00014799, -0.00029291, 0.00033535, -0.00110823, 0.00164089, -0.00000943, -0.00055905, 0.00024976, -0.00024855, 0.00013752, -0.00009962, 0.00001887, -0.00005822, 0.00008572, -0.00011248, 0.00012737, -0.00014478, 0.00012562, -0.00007590, 0.00003097, -0.00005525, 0.00006333, -0.00003198, 0.00001266, 0.00001439, -0.00002609, -0.00000452, 0.00001694, -0.00003314, 0.00002886, -0.00003082, 0.00003275, 0.00000680, -0.00001536, -0.00001552, 0.00001124, -0.00000120, -0.00000449, },
		{ 0.00000153, 0.00001168, -0.00003011, 0.00000196, 0.00002438, -0.00000140, -0.00000522, 0.00001125, -0.00003208, 0.00004198, -0.00005256, 0.00001447, 0.00000526, -0.00000609, 0.00002094, -0.00000044, -0.00002208, -0.00002042, 0.00002704, 0.00000621, -0.00007336, 0.00010719, -0.00013073, 0.00016010, -0.00016267, 0.00009527, -0.00010059, 0.00007697, -0.00019794, 0.00019316, -0.00100708, 0.00186704, -0.00043492, -0.00028220, 0.00009653, -0.00013950, 0.00005779, -0.00003249, -0.00004645, 0.00001830, 0.00001342, -0.00004397, 0.00007118, -0.00010815, 0.00010609, -0.00006762, 0.00002015, -0.00004233, 0.00005416, -0.00002800, 0.00001297, 0.00001546, -0.00003819, 0.00001338, -0.00000039, -0.00001854, 0.00001833, -0.00002242, 0.00003135, 0.00000577, -0.00002183, -0.00000699, 0.00001026, -0.00000424, },
		{ -0.00000532, 0.00000678, 0.00000253, -0.00002712, 0.00000438, 0.00002825, -0.00001213, 0.00000637, -0.00000306, -0.00001744, 0.00002950, -0.00004739, 0.00001560, 0.00001029, -0.00001884, 0.00003926, -0.00002317, 0.00000145, -0.00004777, 0.00007119, -0.00005488, 0.00000042, 0.00003368, -0.00006328, 0.00009557, -0.00011140, 0.00003717, -0.00002708, -0.00002423, -0.00006145, -0.00002032, -0.00077779, 0.00194566, -0.00077779, -0.00002032, -0.00006145, -0.00002423, -0.00002708, 0.00003717, -0.00011140, 0.00009557, -0.00006328, 0.00003368, 0.00000042, -0.00005488, 0.00007119, -0.00004777, 0.00000145, -0.00002317, 0.00003926, -0.00001884, 0.00001029, 0.00001560, -0.00004739, 0.00002950, -0.00001744, -0.00000306, 0.00000637, -0.00001213, 0.00002825, 0.00000438, -0.00002712, 0.00000253, 0.00000678, },
		{ 0.00000153, -0.00000424, 0.00001026, -0.00000699, -0.00002183, 0.00000577, 0.00003135, -0.00002242, 0.00001833, -0.00001854, -0.00000039, 0.00001338, -0.00003819, 0.00001546, 0.00001297, -0.00002800, 0.00005416, -0.00004233, 0.00002015, -0.00006762, 0.00010609, -0.00010815, 0.00007118, -0.00004397, 0.00001342, 0.00001830, -0.00004645, -0.00003249, 0.00005779, -0.00013950, 0.00009653, -0.00028220, -0.00043492, 0.00186704, -0.00100708, 0.00019316, -0.00019794, 0.00007697, -0.00010059, 0.00009527, -0.00016267, 0.00016010, -0.00013073, 0.00010719, -0.00007336, 0.00000621, 0.00002704, -0.00002042, -0.00002208, -0.00000044, 0.00002094, -0.00000609, 0.00000526, 0.00001447, -0.00005256, 0.00004198, -0.00003208, 0.00001125, -0.00000522, -0.00000140, 0.00002438, 0.00000196, -0.00003011, 0.00001168, },
		{ 0.00001909, -0.00000449, -0.00000120, 0.00001124, -0.00001552, -0.00001536, 0.00000680, 0.00003275, -0.00003082, 0.00002886, -0.00003314, 0.00001694, -0.00000452, -0.00002609, 0.00001439, 0.00001266, -0.00003198, 0.00006333, -0.00005525, 0.00003097, -0.00007590, 0.00012562, -0.00014478, 0.00012737, -0.00011248, 0.00008572, -0.00005822, 0.00001887, -0.00009962, 0.00013752, -0.00024855, 0.00024976, -0.00055905, -0.00000943, 0.00164089, -0.00110823, 0.00033535, -0.00029291, 0.00014799, -0.00014945, 0.00013046, -0.00018953, 0.00020067, -0.00017686, 0.00016382, -0.00013804, 0.00006498, -0.00001869, 0.00000894, -0.00004634, 0.00002239, 0.00000224, 0.00000799, -0.00000110, 0.00001174, -0.00005306, 0.00004947, -0.00004263, 0.00002264, -0.00001479, 0.00000842, 0.00002052, -0.00000195, -0.00002990, },
		{ -0.00002604, 0.00002366, -0.00001017, 0.00000327, 0.00000938, -0.00002187, -0.00000891, 0.00000822, 0.00003163, -0.00003600, 0.00003623, -0.00004477, 0.00003229, -0.00002207, -0.00001245, 0.00001272, 0.00000931, -0.00003010, 0.00006553, -0.00006060, 0.00003257, -0.00007079, 0.00012645, -0.00015895, 0.00016015, -0.00016055, 0.00014105, -0.00012093, 0.00007138, -0.00015019, 0.00019563, -0.00033067, 0.00037010, -0.00081165, 0.00045515, 0.00129491, -0.00108425, 0.00039650, -0.00033549, 0.00017956, -0.00016598, 0.00013641, -0.00018597, 0.00021044, -0.00019351, 0.00019377, -0.00018314, 0.00011187, -0.00005823, 0.00003449, -0.00006691, 0.00004172, -0.00001364, 0.00002086, -0.00000749, 0.00000724, -0.00004872, 0.00005130, -0.00004806, 0.00002994, -0.00002119, 0.00001630, 0.00001721, -0.00000742, },
		{ -0.00001410, -0.00001866, 0.00002467, -0.00001445, 0.00000836, 0.00000480, -0.00002527, -0.00000346, 0.00001057, 0.00002747, -0.00003705, 0.00003911, -0.00005173, 0.00004361, -0.00003716, 0.00000124, 0.00001068, 0.00000339, -0.00002269, 0.00006077, -0.00005847, 0.00002554, -0.00005311, 0.00010866, -0.00014891, 0.00016503, -0.00018089, 0.00017031, -0.00015978, 0.00010069, -0.00017317, 0.00021902, -0.00036872, 0.00043254, -0.00100093, 0.00090866, 0.00087052, -0.00095406, 0.00038028, -0.00032501, 0.00017044, -0.00014925, 0.00011290, -0.00015182, 0.00018834, -0.00017796, 0.00019199, -0.00020180, 0.00013966, -0.00008520, 0.00005127, -0.00008006, 0.00005450, -0.00002396, 0.00003017, -0.00001267, 0.00000099, -0.00003994, 0.00004757, -0.00004812, 0.00003265, -0.00002381, 0.00002165, 0.00001463, },
		{ 0.00001262, -0.00002121, -0.00000843, 0.00002195, -0.00001654, 0.00001312, -0.00000196, -0.00002540, 0.00000036, 0.00001411, 0.00002019, -0.00003361, 0.00003682, -0.00005290, 0.00004938, -0.00004798, 0.00001356, 0.00000845, -0.00000415, -0.00001094, 0.00005018, -0.00005033, 0.00001218, -0.00002611, 0.00007572, -0.00011725, 0.00014256, -0.00017150, 0.00016950, -0.00016957, 0.00010135, -0.00016292, 0.00020074, -0.00035277, 0.00041999, -0.00109421, 0.00130143, 0.00041669, -0.00074845, 0.00030175, -0.00027008, 0.00012704, -0.00010486, 0.00006562, -0.00009273, 0.00013915, -0.00013345, 0.00015908, -0.00019187, 0.00014475, -0.00009581, 0.00005612, -0.00008344, 0.00005890, -0.00002696, 0.00003420, -0.00001568, -0.00000670, -0.00002763, 0.00003904, -0.00004329, 0.00003103, -0.00002270, 0.00002436, },
		{ 0.00002473, 0.00001074, -0.00002773, 0.00000350, 0.00001588, -0.00001596, 0.00001659, -0.00000998, -0.00002252, 0.00000236, 0.00001869, 0.00001025, -0.00002604, 0.00002946, -0.00004803, 0.00004884, -0.00005330, 0.00002335, 0.00000611, -0.00001214, 0.00000338, 0.00003570, -0.00003856, -0.00000400, 0.00000521, 0.00003376, -0.00007035, 0.00009816, -0.00013591, 0.00014034, -0.00015090, 0.00007387, -0.00012039, 0.00014162, -0.00028230, 0.00032680, -0.00107069, 0.00159129, -0.00001720, -0.00050456, 0.00018362, -0.00018607, 0.00006145, -0.00004333, 0.00000481, -0.00001888, 0.00007249, -0.00006846, 0.00010105, -0.00015623, 0.00012765, -0.00008944, 0.00004828, -0.00007657, 0.00005466, -0.00002222, 0.00003216, -0.00001605, -0.00001523, -0.00001311, 0.00002704, -0.00003466, 0.00002591, -0.00001843, },
		{ -0.00001198, 0.00002333, 0.00000848, -0.00003252, 0.00001566, 0.00000731, -0.00001265, 0.00001798, -0.00001812, -0.00001729, 0.00000279, 0.00002373, -0.00000139, -0.00001535, 0.00001794, -0.00003774, 0.00004215, -0.00005261, 0.00002979, 0.00000372, -0.00001935, 0.00001827, 0.00001962, -0.00002587, -0.00001914, 0.00003511, -0.00000964, -0.00001705, 0.00004086, -0.00008223, 0.00008973, -0.00010982, 0.00002454, -0.00005311, 0.00005040, -0.00016686, 0.00016042, -0.00092500, 0.00174943, -0.00038818, -0.00025968, 0.00005156, -0.00009155, -0.00001136, 0.00002232, -0.00005702, 0.00005716, 0.00000077, 0.00000502, 0.00002808, -0.00010204, 0.00009274, -0.00006859, 0.00002943, -0.00006091, 0.00004317, -0.00001077, 0.00002437, -0.00001390, -0.00002375, 0.00000207, 0.00001324, -0.00002375, 0.00001855, },
		{ 0.00001039, -0.00000450, 0.00002087, 0.00000537, -0.00003459, 0.00002651, -0.00000255, -0.00000700, 0.00001676, -0.00002523, -0.00001072, 0.00000227, 0.00002840, -0.00001341, -0.00000305, 0.00000386, -0.00002349, 0.00003032, -0.00004622, 0.00003253, 0.00000131, -0.00002473, 0.00003184, 0.00000412, -0.00001473, -0.00002983, 0.00005834, -0.00004698, 0.00003308, -0.00001849, -0.00002144, 0.00002817, -0.00005642, -0.00003590, 0.00002634, -0.00005779, -0.00002467, -0.00005920, -0.00066843, 0.00176406, -0.00066512, -0.00004575, -0.00007051, -0.00000453, -0.00007673, 0.00007939, -0.00010770, 0.00012296, -0.00006328, 0.00007386, -0.00004747, -0.00003925, 0.00004722, -0.00003829, 0.00000324, -0.00003949, 0.00002719, 0.00000510, 0.00001216, -0.00000989, -0.00003124, 0.00001628, -0.00000061, -0.00001220, },
		{ -0.00000153, 0.00000281, 0.00000287, 0.00001797, 0.00000117, -0.00003328, 0.00003466, -0.00001234, 0.00000029, 0.00001275, -0.00003028, -0.00000393, 0.00000157, 0.00003171, -0.00002431, 0.00000906, -0.00001080, -0.00000730, 0.00001513, -0.00003516, 0.00003167, -0.00000107, -0.00002759, 0.00004264, -0.00000915, -0.00000680, -0.00003386, 0.00007113, -0.00007230, 0.00007163, -0.00006937, 0.00003487, -0.00003247, -0.00000264, -0.00009464, 0.00010277, -0.00016422, 0.00012044, -0.00030253, -0.00032732, 0.00164137, -0.00083200, 0.00011482, -0.00016423, 0.00006082, -0.00012311, 0.00011805, -0.00013782, 0.00016866, -0.00010909, 0.00012627, -0.00011346, 0.00002142, -0.00000039, -0.00000489, -0.00002545, -0.00001625, 0.00001025, 0.00002228, -0.00000239, -0.00000508, -0.00003668, 0.00002807, -0.00001289, },
		{ -0.00002234, 0.00000704, -0.00000309, 0.00000918, 0.00001511, -0.00000407, -0.00002838, 0.00003912, -0.00002075, 0.00000823, 0.00000618, -0.00003255, 0.00000205, 0.00000147, 0.00003276, -0.00003271, 0.00001920, -0.00002393, 0.00000853, -0.00000121, -0.00002105, 0.00002773, -0.00000336, -0.00002762, 0.00004981, -0.00001925, -0.00000268, -0.00003063, 0.00007192, -0.00008222, 0.00009287, -0.00010362, 0.00007676, -0.00008134, 0.00004026, -0.00013941, 0.00016145, -0.00025015, 0.00024322, -0.00053500, 0.00006076, 0.00140370, -0.00088855, 0.00021118, -0.00021916, 0.00009630, -0.00014397, 0.00013304, -0.00014238, 0.00018870, -0.00013022, 0.00015402, -0.00016013, 0.00007040, -0.00004196, 0.00002515, -0.00005165, 0.00000481, -0.00000407, 0.00003748, -0.00001687, -0.00000071, -0.00003921, 0.00003632, },
		{ 0.00004037, -0.00002819, 0.00001277, -0.00000661, 0.00001379, 0.00001255, -0.00000997, -0.00002023, 0.00003939, -0.00002672, 0.00001575, -0.00000235, -0.00003172, 0.00000643, 0.00000254, 0.00003093, -0.00003755, 0.00002594, -0.00003372, 0.00002184, -0.00001637, -0.00000582, 0.00002156, -0.00000548, -0.00002493, 0.00005314, -0.00002605, -0.00000190, -0.00002123, 0.00006161, -0.00007654, 0.00009473, -0.00011683, 0.00009780, -0.00011068, 0.00006395, -0.00016108, 0.00019109, -0.00030044, 0.00032156, -0.00072253, 0.00045270, 0.00108541, -0.00084844, 0.00024402, -0.00023343, 0.00010046, -0.00013850, 0.00012438, -0.00012154, 0.00018261, -0.00012537, 0.00015376, -0.00018177, 0.00010101, -0.00007118, 0.00004662, -0.00007127, 0.00002043, -0.00001285, 0.00004780, -0.00002899, 0.00000209, -0.00003827, },
		{ -0.00003378, 0.00004013, -0.00003024, 0.00001542, -0.00000750, 0.00001640, 0.00001031, -0.00001589, -0.00000964, 0.00003554, -0.00002953, 0.00002180, -0.00001197, -0.00002793, 0.00000882, 0.00000504, 0.00002603, -0.00003832, 0.00002841, -0.00003887, 0.00003090, -0.00002830, 0.00000853, 0.00001423, -0.00000740, -0.00001994, 0.00005293, -0.00003001, -0.00000314, -0.00000808, 0.00004319, -0.00005805, 0.00007898, -0.00010896, 0.00009603, -0.00011718, 0.00006459, -0.00015538, 0.00018604, -0.00030647, 0.00034053, -0.00083699, 0.00080618, 0.00072719, -0.00073547, 0.00022354, -0.00021268, 0.00007789, -0.00011109, 0.00009683, -0.00008011, 0.00015471, -0.00009830, 0.00012742, -0.00017744, 0.00011044, -0.00008463, 0.00005638, -0.00008181, 0.00002877, -0.00001443, 0.00005132, -0.00003703, 0.00000257, },
		{ 0.00000049, -0.00002614, 0.00003602, -0.00002881, 0.00001524, -0.00000596, 0.00001704, 0.00000828, -0.00002103, 0.00000221, 0.00002819, -0.00002895, 0.00002552, -0.00002160, -0.00002172, 0.00000924, 0.00000881, 0.00001837, -0.00003511, 0.00002645, -0.00003889, 0.00003472, -0.00003548, 0.00002026, 0.00000683, -0.00000910, -0.00001327, 0.00004980, -0.00003195, -0.00000469, 0.00000565, 0.00002107, -0.00003181, 0.00005063, -0.00008404, 0.00007409, -0.00010231, 0.00004336, -0.00012373, 0.00014733, -0.00026779, 0.00029509, -0.00086060, 0.00108556, 0.00036989, -0.00057870, 0.00016613, -0.00016775, 0.00003745, -0.00006975, 0.00005851, -0.00002624, 0.00011282, -0.00005674, 0.00008142, -0.00015080, 0.00009986, -0.00008214, 0.00005379, -0.00008276, 0.00002967, -0.00000874, 0.00004745, -0.00004014, },
		{ -0.00003843, -0.00000376, -0.00001621, 0.00002894, -0.00002462, 0.00001289, -0.00000250, 0.00001595, 0.00000625, -0.00002462, 0.00001399, 0.00001835, -0.00002519, 0.00002634, -0.00003016, -0.00001396, 0.00000808, 0.00001333, 0.00000871, -0.00002862, 0.00002061, -0.00003407, 0.00003318, -0.00003718, 0.00002810, 0.00000037, -0.00001061, -0.00000557, 0.00004452, -0.00003270, -0.00000494, 0.00001682, 0.00000003, -0.00000395, 0.00001667, -0.00004908, 0.00003830, -0.00007169, 0.00000597, -0.00007269, 0.00008236, -0.00019212, 0.00019102, -0.00078851, 0.00126657, 0.00004891, -0.00040734, 0.00009050, -0.00011179, -0.00001014, -0.00002399, 0.00001884, 0.00003055, 0.00006629, -0.00001054, 0.00002519, -0.00010898, 0.00007385, -0.00006644, 0.00004057, -0.00007545, 0.00002460, 0.00000278, 0.00003694, },
		{ 0.00002173, -0.00003292, -0.00000930, -0.00000522, 0.00002011, -0.00001871, 0.00000923, 0.00000211, 0.00001357, 0.00000408, -0.00002606, 0.00002438, 0.00000729, -0.00001885, 0.00002403, -0.00003667, -0.00000566, 0.00000602, 0.00001782, -0.00000184, -0.00002002, 0.00001207, -0.00002546, 0.00002699, -0.00003353, 0.00003140, -0.00000441, -0.00001200, 0.00000257, 0.00003776, -0.00003285, -0.00000285, 0.00002305, -0.00001580, 0.00001965, -0.00001545, -0.00001245, -0.00000286, -0.00003357, -0.00003886, -0.00001248, 0.00000329, -0.00009385, 0.00004391, -0.00062925, 0.00133880, -0.00020995, -0.00024641, 0.00001402, -0.00005732, -0.00005471, 0.00001735, -0.00001346, 0.00008137, 0.00002397, 0.00003051, -0.00003093, -0.00006094, 0.00003916, -0.00004233, 0.00002017, -0.00006262, 0.00001616, 0.00001752, },
		{ 0.00003227, 0.00000448, -0.00002521, -0.00001489, 0.00000545, 0.00001087, -0.00001221, 0.00000525, 0.00000704, 0.00001037, 0.00000171, -0.00002505, 0.00003232, -0.00000367, -0.00001083, 0.00001875, -0.00004041, 0.00000214, 0.00000382, 0.00002139, -0.00001201, -0.00001071, 0.00000233, -0.00001463, 0.00001758, -0.00002544, 0.00003023, -0.00000706, -0.00001333, 0.00001067, 0.00003001, -0.00003254, 0.00000175, 0.00002323, -0.00002379, 0.00003461, -0.00003939, 0.00001798, -0.00004070, 0.00000304, -0.00008136, 0.00004526, -0.00007553, 0.00000884, -0.00012369, -0.00040287, 0.00130593, -0.00039269, -0.00011377, -0.00005001, -0.00001396, -0.00008864, 0.00004785, -0.00003209, 0.00011965, -0.00000761, 0.00005866, -0.00007772, -0.00001554, 0.00000311, -0.00001546, -0.00000313, -0.00004760, 0.00000748, },
		{ 0.00000139, 0.00004394, -0.00001201, -0.00001715, -0.00001927, 0.00001448, 0.00000249, -0.00000617, 0.00000183, 0.00001148, 0.00000680, -0.00000077, -0.00002162, 0.00003711, -0.00001337, -0.00000218, 0.00001101, -0.00004100, 0.00000858, 0.00000221, 0.00002324, -0.00002061, -0.00000214, -0.00000695, -0.00000340, 0.00000676, -0.00001445, 0.00002527, -0.00000753, -0.00001463, 0.00001840, 0.00002153, -0.00003148, 0.00000815, 0.00001770, -0.00002328, 0.00003882, -0.00005117, 0.00003634, -0.00006804, 0.00003034, -0.00011289, 0.00008996, -0.00014043, 0.00009746, -0.00028618, -0.00013740, 0.00118380, -0.00049718, -0.00001882, -0.00009384, 0.00001290, -0.00010792, 0.00006456, -0.00003410, 0.00014211, -0.00002534, 0.00006959, -0.00010874, 0.00002007, -0.00002784, 0.00000888, -0.00002527, -0.00003353, },
		{ -0.00002261, -0.00000019, 0.00005023, -0.00002532, -0.00001039, -0.00002136, 0.00002080, -0.00000399, -0.00000146, -0.00000033, 0.00001478, 0.00000329, -0.00000316, -0.00001615, 0.00003847, -0.00002088, 0.00000603, 0.00000158, -0.00003844, 0.00001310, 0.00000170, 0.00002279, -0.00002668, 0.00000451, -0.00001434, 0.00000652, -0.00000352, -0.00000245, 0.00001769, -0.00000611, -0.00001586, 0.00002550, 0.00001248, -0.00002910, 0.00001492, 0.00000817, -0.00001557, 0.00003279, -0.00004983, 0.00003984, -0.00008060, 0.00004327, -0.00012764, 0.00011417, -0.00018113, 0.00015668, -0.00041927, 0.00013579, 0.00099675, -0.00053135, 0.00003707, -0.00011547, 0.00002228, -0.00011238, 0.00006807, -0.00002003, 0.00014888, -0.00002976, 0.00006300, -0.00012128, 0.00004165, -0.00004924, 0.00002679, -0.00004317, },
		{ -0.00005525, -0.00001572, 0.00000331, 0.00005002, -0.00003391, -0.00000608, -0.00002056, 0.00002381, -0.00000794, 0.00000143, -0.00000081, 0.00001647, 0.00000017, -0.00000520, -0.00000929, 0.00003658, -0.00002568, 0.00001286, -0.00000856, -0.00003310, 0.00001549, 0.00000248, 0.00001983, -0.00002970, 0.00000847, -0.00001884, 0.00001378, -0.00001160, 0.00000868, 0.00000897, -0.00000339, -0.00001686, 0.00003173, 0.00000299, -0.00002477, 0.00002029, -0.00000281, -0.00000352, 0.00001926, -0.00003735, 0.00002906, -0.00007760, 0.00004033, -0.00012372, 0.00011494, -0.00019256, 0.00017709, -0.00050396, 0.00038687, 0.00077315, -0.00051016, 0.00005892, -0.00011758, 0.00001687, -0.00010491, 0.00006187, 0.00000677, 0.00014283, -0.00002444, 0.00004222, -0.00011644, 0.00004825, -0.00005925, 0.00003629, },
		{ 0.00003743, -0.00006151, -0.00001238, 0.00001107, 0.00004360, -0.00003732, -0.00000457, -0.00001686, 0.00002347, -0.00000917, 0.00000232, 0.00000049, 0.00001632, -0.00000231, -0.00000664, -0.00000182, 0.00003197, -0.00002760, 0.00001758, -0.00001838, -0.00002568, 0.00001587, 0.00000443, 0.00001457, -0.00002961, 0.00000949, -0.00001998, 0.00001761, -0.00001629, 0.00001732, 0.00000058, -0.00000014, -0.00001737, 0.00003682, -0.00000669, -0.00001811, 0.00002263, -0.00001242, 0.00000925, 0.00000249, -0.00001801, 0.00000755, -0.00006152, 0.00002355, -0.00010320, 0.00009408, -0.00017560, 0.00015667, -0.00052933, 0.00059166, 0.00054086, -0.00045188, 0.00005561, -0.00010587, 0.00000158, -0.00009026, 0.00005114, 0.00004128, 0.00012832, -0.00001467, 0.00001310, -0.00009840, 0.00004192, -0.00005854, },
		{ -0.00004971, 0.00003184, -0.00006320, -0.00001102, 0.00002114, 0.00003250, -0.00003617, -0.00000549, -0.00001087, 0.00002026, -0.00000795, 0.00000144, 0.00000336, 0.00001435, -0.00000393, -0.00000732, 0.00000546, 0.00002540, -0.00002681, 0.00001972, -0.00002690, -0.00001706, 0.00001470, 0.00000710, 0.00000757, -0.00002682, 0.00000783, -0.00001793, 0.00001791, -0.00001709, 0.00002236, -0.00000616, 0.00000283, -0.00001707, 0.00004042, -0.00001605, -0.00000918, 0.00002089, -0.00001835, 0.00001928, -0.00001286, 0.00000270, -0.00001924, -0.00003730, -0.00000229, -0.00007138, 0.00005749, -0.00013651, 0.00010081, -0.00049377, 0.00073450, 0.00032346, -0.00037467, 0.00003731, -0.00008705, -0.00001799, -0.00007354, 0.00004137, 0.00007817, 0.00010985, -0.00000593, -0.00001755, -0.00007308, 0.00002677, },
		{ 0.00000774, -0.00003639, 0.00002204, -0.00006220, -0.00000963, 0.00003093, 0.00001907, -0.00003191, -0.00000780, -0.00000369, 0.00001513, -0.00000493, -0.00000071, 0.00000733, 0.00001078, -0.00000456, -0.00000721, 0.00001189, 0.00001772, -0.00002378, 0.00001914, -0.00003331, -0.00000822, 0.00001262, 0.00000983, -0.00000032, -0.00002209, 0.00000420, -0.00001334, 0.00001520, -0.00001420, 0.00002335, -0.00001033, 0.00000481, -0.00001561, 0.00004211, -0.00002438, 0.00000134, 0.00001489, -0.00001923, 0.00002402, -0.00002282, 0.00001942, -0.00004524, -0.00001096, -0.00003099, -0.00003536, 0.00001358, -0.00008535, 0.00002097, -0.00040460, 0.00080943, 0.00013780, -0.00029385, 0.00001325, -0.00006719, -0.00003704, -0.00005897, 0.00003717, 0.00011291, 0.00009079, -0.00000240, -0.00004360, -0.00004671, },
		{ -0.00002439, -0.00001062, -0.00002215, 0.00001063, -0.00006035, -0.00000635, 0.00003797, 0.00000593, -0.00002641, -0.00001016, 0.00000331, 0.00000930, -0.00000102, -0.00000343, 0.00001175, 0.00000603, -0.00000414, -0.00000640, 0.00001699, 0.00000976, -0.00001914, 0.00001598, -0.00003703, -0.00000006, 0.00001032, 0.00001191, -0.00000812, -0.00001638, -0.00000046, -0.00000722, 0.00001048, -0.00000846, 0.00002052, -0.00001148, 0.00000527, -0.00001273, 0.00004151, -0.00003078, 0.00001228, 0.00000543, -0.00001502, 0.00002247, -0.00002497, 0.00002805, -0.00006517, 0.00001177, -0.00005644, -0.00000236, -0.00002873, -0.00003361, -0.00006788, -0.00027602, 0.00081960, -0.00000698, -0.00022030, -0.00000974, -0.00005055, -0.00005249, -0.00004903, 0.00004154, 0.00014244, 0.00007276, -0.00000601, -0.00006089, },
		{ -0.00006793, -0.00000916, -0.00002508, -0.00000974, -0.00000042, -0.00005882, -0.00000010, 0.00004042, -0.00000466, -0.00002148, -0.00001122, 0.00000879, 0.00000405, 0.00000279, -0.00000594, 0.00001589, 0.00000066, -0.00000271, -0.00000512, 0.00002053, 0.00000223, -0.00001360, 0.00001066, -0.00003781, 0.00000669, 0.00000843, 0.00001266, -0.00001488, -0.00001069, -0.00000513, -0.00000071, 0.00000495, -0.00000112, 0.00001467, -0.00000969, 0.00000403, -0.00000835, 0.00003837, -0.00003437, 0.00002210, -0.00000589, -0.00000693, 0.00001530, -0.00001894, 0.00002660, -0.00007555, 0.00002671, -0.00007398, 0.00002190, -0.00006192, 0.00000822, -0.00015013, -0.00012626, 0.00077539, -0.00010900, -0.00016003, -0.00002781, -0.00003915, -0.00006323, -0.00004422, 0.00005562, 0.00016524, 0.00005552, -0.00001612, },
		{ -0.00002996, -0.00006583, -0.00000175, -0.00003425, -0.00000057, -0.00001014, -0.00005784, 0.00000908, 0.00003762, -0.00001126, -0.00001843, -0.00001003, 0.00001169, 0.00000046, 0.00000563, -0.00000755, 0.00001902, -0.00000467, -0.00000042, -0.00000363, 0.00002248, -0.00000436, -0.00000789, 0.00000379, -0.00003571, 0.00001158, 0.00000737, 0.00001164, -0.00001983, -0.00000588, -0.00000892, 0.00000517, -0.00000015, 0.00000640, 0.00000701, -0.00000549, 0.00000120, -0.00000259, 0.00003269, -0.00003449, 0.00002925, -0.00001704, 0.00000294, 0.00000465, -0.00000637, 0.00001541, -0.00007525, 0.00003192, -0.00008124, 0.00003426, -0.00008138, 0.00003269, -0.00021246, 0.00002573, 0.00069161, -0.00017199, -0.00011480, -0.00003978, -0.00003303, -0.00006991, -0.00004338, 0.00007903, 0.00018094, 0.00003755, },
		{ 0.00001706, -0.00004364, -0.00005764, -0.00000082, -0.00003842, 0.00000528, -0.00001864, -0.00005662, 0.00001999, 0.00003016, -0.00001352, -0.00001777, -0.00000624, 0.00001149, -0.00000075, 0.00000683, -0.00000771, 0.00002054, -0.00000927, 0.00000242, -0.00000216, 0.00002297, -0.00000967, -0.00000266, -0.00000385, -0.00003111, 0.00001446, 0.00000730, 0.00000866, -0.00002250, -0.00000257, -0.00001118, 0.00000964, -0.00000386, 0.00001273, -0.00000105, 0.00000023, -0.00000279, 0.00000414, 0.00002479, -0.00003088, 0.00003249, -0.00002598, 0.00001212, -0.00000651, 0.00000958, -0.00000312, -0.00006546, 0.00002782, -0.00007844, 0.00003454, -0.00008608, 0.00003674, -0.00024598, 0.00016335, 0.00058447, -0.00020327, -0.00008333, -0.00004643, -0.00003090, -0.00007426, -0.00004419, 0.00011042, 0.00018957, },
		{ 0.00019085, -0.00000699, -0.00005339, -0.00004724, -0.00000383, -0.00003913, 0.00000873, -0.00002671, -0.00005377, 0.00003057, 0.00001976, -0.00001220, -0.00001919, -0.00000019, 0.00000827, 0.00000066, 0.00000610, -0.00000617, 0.00002005, -0.00001251, 0.00000541, -0.00000083, 0.00002220, -0.00001350, 0.00000153, -0.00001146, -0.00002464, 0.00001551, 0.00000809, 0.00000388, -0.00002273, -0.00000103, -0.00001161, 0.00001230, -0.00000559, 0.00001685, -0.00000817, 0.00000641, -0.00000727, 0.00001119, 0.00001537, -0.00002385, 0.00003122, -0.00003120, 0.00001839, -0.00001509, 0.00002504, -0.00002520, -0.00004913, 0.00001676, -0.00006792, 0.00002517, -0.00007840, 0.00002189, -0.00024734, 0.00027444, 0.00046899, -0.00021147, -0.00006271, -0.00004946, -0.00003098, -0.00007838, -0.00004395, 0.00014809, },
		{ 0.00019038, 0.00018368, -0.00003403, -0.00005674, -0.00003813, -0.00000799, -0.00003829, 0.00001103, -0.00003534, -0.00004777, 0.00003847, 0.00000879, -0.00000886, -0.00002161, 0.00000710, 0.00000273, 0.00000442, 0.00000351, -0.00000300, 0.00001742, -0.00001388, 0.00000805, 0.00000033, 0.00002040, -0.00001582, 0.00000427, -0.00001821, -0.00001709, 0.00001514, 0.00000937, -0.00000225, -0.00002073, -0.00000125, -0.00001024, 0.00001314, -0.00000521, 0.00001817, -0.00001326, 0.00001201, -0.00001147, 0.00001777, 0.00000544, -0.00001426, 0.00002560, -0.00003202, 0.00002040, -0.00001867, 0.00003637, -0.00004664, -0.00003005, 0.00000218, -0.00005331, 0.00001030, -0.00006318, -0.00000649, -0.00021874, 0.00035244, 0.00035711, -0.00020479, -0.00004974, -0.00005063, -0.00003171, -0.00008402, -0.00004008, },
		{ -0.00003039, 0.00023575, 0.00016625, -0.00006161, -0.00005330, -0.00003241, -0.00001122, -0.00003747, 0.00001317, -0.00004514, -0.00003762, 0.00004167, -0.00000029, -0.00000544, -0.00002357, 0.00001420, -0.00000394, 0.00000979, -0.00000050, 0.00000145, 0.00001284, -0.00001310, 0.00000985, 0.00000145, 0.00001773, -0.00001666, 0.00000530, -0.00002339, -0.00000933, 0.00001393, 0.00001057, -0.00000904, -0.00001695, -0.00000293, -0.00000734, 0.00001244, -0.00000305, 0.00001662, -0.00001562, 0.00001613, -0.00001465, 0.00002304, -0.00000383, -0.00000346, 0.00001658, -0.00002876, 0.00001793, -0.00001610, 0.00004093, -0.00006382, -0.00001193, -0.00001243, -0.00003842, -0.00000554, -0.00004624, -0.00004075, -0.00016675, 0.00039624, 0.00025685, -0.00018988, -0.00004171, -0.00005117, -0.00003217, -0.00009226, },
		{ -0.00010349, -0.00001298, 0.00028241, 0.00013652, -0.00008573, -0.00004486, -0.00003037, -0.00001277, -0.00003731, 0.00001535, -0.00005584, -0.00002333, 0.00003907, -0.00000557, -0.00000370, -0.00002360, 0.00001959, -0.00001034, 0.00001578, -0.00000525, 0.00000655, 0.00000680, -0.00001015, 0.00001041, 0.00000272, 0.00001433, -0.00001614, 0.00000452, -0.00002647, -0.00000218, 0.00001249, 0.00001108, -0.00001568, -0.00001207, -0.00000562, -0.00000337, 0.00001071, 0.00000028, 0.00001263, -0.00001506, 0.00001818, -0.00001623, 0.00002629, -0.00001124, 0.00000693, 0.00000572, -0.00002262, 0.00001197, -0.00000771, 0.00003769, -0.00007437, 0.00000248, -0.00002435, -0.00002626, -0.00001859, -0.00003282, -0.00007278, -0.00010057, 0.00040915, 0.00017231, -0.00017141, -0.00003677, -0.00005164, -0.00003219, },
		{ -0.00003211, -0.00011755, 0.00001395, 0.00032777, 0.00009316, -0.00010178, -0.00003491, -0.00003049, -0.00001328, -0.00003732, 0.00001685, -0.00006621, -0.00000612, 0.00003088, -0.00000615, -0.00000479, -0.00002065, 0.00002198, -0.00001510, 0.00002128, -0.00000990, 0.00001153, 0.00000005, -0.00000536, 0.00000953, 0.00000430, 0.00001032, -0.00001446, 0.00000202, -0.00002713, 0.00000371, 0.00001138, 0.00001034, -0.00002135, -0.00000682, -0.00000878, 0.00000118, 0.00000850, 0.00000398, 0.00000698, -0.00001180, 0.00001793, -0.00001589, 0.00002706, -0.00001585, 0.00001536, -0.00000516, -0.00001538, 0.00000443, 0.00000475, 0.00002728, -0.00007757, 0.00001183, -0.00003224, -0.00001832, -0.00002682, -0.00002632, -0.00009583, -0.00002994, 0.00039731, 0.00010451, -0.00015225, -0.00003390, -0.00005205, },
		},
		{ { -0.00023878, 0.00000954, 0.00036450, 0.00003641, -0.00013401, -0.00001754, -0.00009790, -0.00000464, 0.00000103, -0.00003018, 0.00003826, -0.00002785, 0.00003864, -0.00001639, -0.00002704, 0.00000413, 0.00001574, -0.00001133, 0.00002365, -0.00001198, -0.00001153, -0.00000471, -0.00000428, -0.00000165, 0.00001010, -0.00001254, 0.00000982, -0.00000898, 0.00001725, -0.00000617, 0.00000198, -0.00001054, 0.00000576, -0.00001054, 0.00000198, -0.00000617, 0.00001725, -0.00000898, 0.00000982, -0.00001254, 0.00001010, -0.00000165, -0.00000428, -0.00000471, -0.00001153, -0.00001198, 0.00002365, -0.00001133, 0.00001574, 0.00000413, -0.00002704, -0.00001639, 0.00003864, -0.00002785, 0.00003826, -0.00003018, 0.00000103, -0.00000464, -0.00009790, -0.00001754, -0.00013401, 0.00003641, 0.00036450, 0.00000954, },
		{ -0.00000464, -0.00024031, 0.00002818, 0.00040518, -0.00000411, -0.00013544, -0.00001469, -0.00010146, 0.00000620, -0.00001065, -0.00002263, 0.00003704, -0.00002704, 0.00003989, -0.00002293, -0.00002146, 0.00000469, 0.00001291, -0.00001026, 0.00002504, -0.00001509, -0.00001108, -0.00000397, -0.00000375, -0.00000108, 0.00000883, -0.00001187, 0.00000999, -0.00000828, 0.00001662, -0.00000643, 0.00000107, -0.00001030, 0.00000557, -0.00001038, 0.00000266, -0.00000577, 0.00001777, -0.00000982, 0.00000976, -0.00001324, 0.00001144, -0.00000237, -0.00000483, -0.00000535, -0.00001176, -0.00000897, 0.00002193, -0.00001202, 0.00001880, 0.00000253, -0.00003193, -0.00000917, 0.00003594, -0.00002728, 0.00003764, -0.00003566, 0.00001176, -0.00001625, -0.00009229, -0.00002327, -0.00012599, 0.00006896, 0.00032394, },
		{ 0.00028446, -0.00001507, -0.00024500, 0.00005232, 0.00044443, -0.00005216, -0.00012930, -0.00001604, -0.00010167, 0.00001492, -0.00002207, -0.00001382, 0.00003439, -0.00002509, 0.00003976, -0.00002860, -0.00001536, 0.00000413, 0.00001057, -0.00000904, 0.00002620, -0.00001839, -0.00001026, -0.00000324, -0.00000321, -0.00000070, 0.00000768, -0.00001125, 0.00001033, -0.00000781, 0.00001603, -0.00000668, 0.00000007, -0.00000976, 0.00000502, -0.00000976, 0.00000302, -0.00000512, 0.00001803, -0.00001065, 0.00000975, -0.00001390, 0.00001281, -0.00000321, -0.00000544, -0.00000579, -0.00001191, -0.00000596, 0.00001974, -0.00001211, 0.00002180, 0.00000004, -0.00003602, -0.00000145, 0.00003184, -0.00002522, 0.00003492, -0.00003841, 0.00002044, -0.00002728, -0.00008609, -0.00003025, -0.00011290, 0.00009381, },
		{ 0.00011179, 0.00024650, -0.00002215, -0.00025317, 0.00008332, 0.00048010, -0.00010651, -0.00011530, -0.00002249, -0.00009749, 0.00002033, -0.00003205, -0.00000466, 0.00003086, -0.00002230, 0.00003837, -0.00003328, -0.00000894, 0.00000244, 0.00000892, -0.00000785, 0.00002725, -0.00002196, -0.00000896, -0.00000264, -0.00000262, -0.00000050, 0.00000665, -0.00001070, 0.00001085, -0.00000765, 0.00001557, -0.00000705, -0.00000089, -0.00000901, 0.00000417, -0.00000869, 0.00000297, -0.00000409, 0.00001788, -0.00001133, 0.00000965, -0.00001444, 0.00001414, -0.00000413, -0.00000610, -0.00000598, -0.00001208, -0.00000286, 0.00001701, -0.00001139, 0.00002442, -0.00000312, -0.00003922, 0.00000655, 0.00002647, -0.00002169, 0.00003001, -0.00003803, 0.00002620, -0.00003649, -0.00008070, -0.00003677, -0.00009661, },
		{ -0.00007913, 0.00012418, 0.00021005, -0.00002604, -0.00026523, 0.00012270, 0.00050956, -0.00016514, -0.00009389, -0.00003432, -0.00008832, 0.00002151, -0.00003953, 0.00000396, 0.00002701, -0.00001906, 0.00003588, -0.00003686, -0.00000246, -0.00000025, 0.00000806, -0.00000683, 0.00002823, -0.00002584, -0.00000706, -0.00000226, -0.00000194, -0.00000049, 0.00000573, -0.00001017, 0.00001153, -0.00000782, 0.00001534, -0.00000763, -0.00000167, -0.00000821, 0.00000313, -0.00000722, 0.00000251, -0.00000264, 0.00001721, -0.00001170, 0.00000937, -0.00001477, 0.00001535, -0.00000511, -0.00000681, -0.00000588, -0.00001234, 0.00000037, 0.00001368, -0.00000969, 0.00002635, -0.00000672, -0.00004149, 0.00001456, 0.00002005, -0.00001688, 0.00002306, -0.00003442, 0.00002847, -0.00004296, -0.00007734, -0.00004120, },
		{ -0.00004220, -0.00006237, 0.00013240, 0.00017484, -0.00002677, -0.00028160, 0.00017201, 0.00052985, -0.00022537, -0.00006628, -0.00005121, -0.00007407, 0.00001794, -0.00004368, 0.00001126, 0.00002339, -0.00001574, 0.00003253, -0.00003924, 0.00000380, -0.00000377, 0.00000797, -0.00000603, 0.00002914, -0.00002999, -0.00000453, -0.00000219, -0.00000116, -0.00000067, 0.00000491, -0.00000963, 0.00001228, -0.00000828, 0.00001535, -0.00000850, -0.00000216, -0.00000748, 0.00000203, -0.00000542, 0.00000165, -0.00000073, 0.00001593, -0.00001164, 0.00000878, -0.00001480, 0.00001637, -0.00000612, -0.00000754, -0.00000553, -0.00001270, 0.00000370, 0.00000977, -0.00000695, 0.00002736, -0.00001052, -0.00004280, 0.00002231, 0.00001289, -0.00001109, 0.00001445, -0.00002777, 0.00002709, -0.00004609, -0.00007686, },
		{ -0.00007968, -0.00003886, -0.00004798, 0.00013790, 0.00014051, -0.00002439, -0.00030250, 0.00023263, 0.00053784, -0.00028394, -0.00003441, -0.00007217, -0.00005522, 0.00000959, -0.00004405, 0.00001662, 0.00002045, -0.00001270, 0.00002851, -0.00004035, 0.00000951, -0.00000782, 0.00000857, -0.00000541, 0.00002990, -0.00003432, -0.00000138, -0.00000248, -0.00000026, -0.00000099, 0.00000413, -0.00000899, 0.00001301, -0.00000896, 0.00001557, -0.00000965, -0.00000229, -0.00000696, 0.00000101, -0.00000343, 0.00000047, 0.00000159, 0.00001401, -0.00001106, 0.00000780, -0.00001446, 0.00001714, -0.00000712, -0.00000823, -0.00000498, -0.00001313, 0.00000707, 0.00000540, -0.00000318, 0.00002726, -0.00001429, -0.00004313, 0.00002951, 0.00000539, -0.00000476, 0.00000474, -0.00001862, 0.00002225, -0.00004576, },
		{ -0.00004228, -0.00008571, -0.00003083, -0.00003713, 0.00014190, 0.00010677, -0.00001911, -0.00032784, 0.00030557, 0.00053054, -0.00033726, -0.00000076, -0.00009562, -0.00003281, -0.00000308, -0.00004052, 0.00001968, 0.00001854, -0.00001020, 0.00002405, -0.00004015, 0.00001435, -0.00001205, 0.00000965, -0.00000488, 0.00003037, -0.00003865, 0.00000231, -0.00000312, 0.00000073, -0.00000142, 0.00000335, -0.00000815, 0.00001357, -0.00000972, 0.00001593, -0.00001105, -0.00000202, -0.00000674, 0.00000022, -0.00000140, -0.00000092, 0.00000423, 0.00001148, -0.00000992, 0.00000636, -0.00001367, 0.00001759, -0.00000810, -0.00000884, -0.00000434, -0.00001352, 0.00001034, 0.00000075, 0.00000149, 0.00002599, -0.00001785, -0.00004246, 0.00003584, -0.00000204, 0.00000158, -0.00000534, -0.00000776, 0.00001457, },
		{ 0.00000499, -0.00003639, -0.00009431, -0.00001837, -0.00003044, 0.00014523, 0.00007357, -0.00001146, -0.00035700, 0.00039120, 0.00050537, -0.00038170, 0.00003177, -0.00011952, -0.00000836, -0.00001915, -0.00003344, 0.00002037, 0.00001780, -0.00000843, 0.00001933, -0.00003860, 0.00001799, -0.00001605, 0.00001094, -0.00000427, 0.00003035, -0.00004273, 0.00000639, -0.00000407, 0.00000178, -0.00000191, 0.00000250, -0.00000703, 0.00001381, -0.00001040, 0.00001629, -0.00001262, -0.00000138, -0.00000689, -0.00000023, 0.00000053, -0.00000235, 0.00000705, 0.00000843, -0.00000823, 0.00000446, -0.00001241, 0.00001770, -0.00000906, -0.00000927, -0.00000373, -0.00001374, 0.00001331, -0.00000392, 0.00000684, 0.00002358, -0.00002108, -0.00004080, 0.00004101, -0.00000892, 0.00000736, -0.00001497, 0.00000380, },
		{ 0.00001492, -0.00000530, -0.00002922, -0.00010442, -0.00000233, -0.00002787, 0.00014826, 0.00004120, -0.00000236, -0.00038878, 0.00048910, 0.00046038, -0.00041381, 0.00006014, -0.00014152, 0.00001624, -0.00003724, -0.00002352, 0.00001890, 0.00001818, -0.00000744, 0.00001447, -0.00003570, 0.00002015, -0.00001942, 0.00001212, -0.00000334, 0.00002960, -0.00004628, 0.00001062, -0.00000522, 0.00000282, -0.00000239, 0.00000152, -0.00000555, 0.00001359, -0.00001084, 0.00001651, -0.00001421, -0.00000042, -0.00000742, -0.00000022, 0.00000217, -0.00000367, 0.00000988, 0.00000501, -0.00000607, 0.00000212, -0.00001070, 0.00001744, -0.00000998, -0.00000946, -0.00000331, -0.00001363, 0.00001575, -0.00000829, 0.00001255, 0.00002016, -0.00002391, -0.00003812, 0.00004475, -0.00001482, 0.00001200, -0.00002328, },
		{ -0.00002946, 0.00002443, -0.00001496, -0.00002214, -0.00011457, 0.00001588, -0.00002873, 0.00015080, 0.00001038, 0.00000678, -0.00042123, 0.00059786, 0.00039457, -0.00043067, 0.00008143, -0.00015917, 0.00003886, -0.00005569, -0.00001180, 0.00001578, 0.00001947, -0.00000717, 0.00000958, -0.00003148, 0.00002059, -0.00002175, 0.00001285, -0.00000187, 0.00002785, -0.00004896, 0.00001474, -0.00000644, 0.00000376, -0.00000279, 0.00000037, -0.00000363, 0.00001278, -0.00001086, 0.00001642, -0.00001568, 0.00000073, -0.00000829, 0.00000030, 0.00000341, -0.00000470, 0.00001250, 0.00000140, -0.00000357, -0.00000058, -0.00000856, 0.00001682, -0.00001087, -0.00000932, -0.00000321, -0.00001301, 0.00001743, -0.00001201, 0.00001821, 0.00001599, -0.00002635, -0.00003441, 0.00004682, -0.00001932, 0.00001498, },
		{ 0.00001588, -0.00003280, 0.00003124, -0.00002261, -0.00001663, -0.00012304, 0.00003445, -0.00003181, 0.00015215, -0.00001777, 0.00001410, -0.00045173, 0.00071506, 0.00030805, -0.00043013, 0.00009307, -0.00017015, 0.00005734, -0.00007266, 0.00000041, 0.00001172, 0.00002126, -0.00000744, 0.00000470, -0.00002600, 0.00001915, -0.00002270, 0.00001283, 0.00000039, 0.00002485, -0.00005044, 0.00001844, -0.00000756, 0.00000451, -0.00000306, -0.00000098, -0.00000125, 0.00001129, -0.00001031, 0.00001584, -0.00001682, 0.00000192, -0.00000943, 0.00000134, 0.00000413, -0.00000527, 0.00001469, -0.00000217, -0.00000089, -0.00000350, -0.00000610, 0.00001589, -0.00001174, -0.00000881, -0.00000356, -0.00001171, 0.00001810, -0.00001470, 0.00002337, 0.00001137, -0.00002847, -0.00002968, 0.00004703, -0.00002211, },
		{ -0.00002294, 0.00001441, -0.00003279, 0.00003443, -0.00002703, -0.00001414, -0.00012809, 0.00005131, -0.00003542, 0.00015119, -0.00004178, 0.00001733, -0.00047699, 0.00083719, 0.00020218, -0.00041102, 0.00009310, -0.00017247, 0.00006968, -0.00008629, 0.00001169, 0.00000762, 0.00002300, -0.00000798, -0.00000017, -0.00001935, 0.00001574, -0.00002200, 0.00001176, 0.00000362, 0.00002040, -0.00005039, 0.00002142, -0.00000839, 0.00000496, -0.00000313, -0.00000255, 0.00000160, 0.00000907, -0.00000907, 0.00001463, -0.00001748, 0.00000297, -0.00001072, 0.00000285, 0.00000427, -0.00000526, 0.00001627, -0.00000545, 0.00000175, -0.00000650, -0.00000343, 0.00001469, -0.00001260, -0.00000789, -0.00000446, -0.00000960, 0.00001756, -0.00001604, 0.00002758, 0.00000669, -0.00003037, -0.00002393, 0.00004527, },
		{ 0.00004152, -0.00002172, 0.00001046, -0.00002916, 0.00003336, -0.00002721, -0.00001596, -0.00012803, 0.00006435, -0.00003767, 0.00014648, -0.00005993, 0.00001397, -0.00049321, 0.00095977, 0.00007959, -0.00037333, 0.00008036, -0.00016472, 0.00007424, -0.00009487, 0.00002059, 0.00000442, 0.00002410, -0.00000843, -0.00000505, -0.00001168, 0.00001041, -0.00001950, 0.00000946, 0.00000794, 0.00001437, -0.00004858, 0.00002338, -0.00000874, 0.00000500, -0.00000295, -0.00000432, 0.00000487, 0.00000613, -0.00000709, 0.00001267, -0.00001747, 0.00000372, -0.00001198, 0.00000477, 0.00000383, -0.00000458, 0.00001704, -0.00000821, 0.00000415, -0.00000937, -0.00000068, 0.00001331, -0.00001345, -0.00000655, -0.00000597, -0.00000662, 0.00001568, -0.00001575, 0.00003041, 0.00000235, -0.00003218, -0.00001721, },
		{ -0.00000959, 0.00003584, -0.00001845, 0.00000411, -0.00002190, 0.00002772, -0.00002248, -0.00002302, -0.00012151, 0.00007159, -0.00003660, 0.00013649, -0.00007045, 0.00000142, -0.00049628, 0.00107753, -0.00005584, -0.00031824, 0.00005465, -0.00014615, 0.00006987, -0.00009703, 0.00002583, 0.00000308, 0.00002393, -0.00000841, -0.00001000, -0.00000319, 0.00000330, -0.00001516, 0.00000584, 0.00001339, 0.00000671, -0.00004481, 0.00002407, -0.00000843, 0.00000454, -0.00000249, -0.00000625, 0.00000846, 0.00000252, -0.00000436, 0.00000991, -0.00001668, 0.00000397, -0.00001305, 0.00000694, 0.00000285, -0.00000321, 0.00001689, -0.00001025, 0.00000607, -0.00001193, 0.00000197, 0.00001184, -0.00001429, -0.00000483, -0.00000807, -0.00000273, 0.00001239, -0.00001365, 0.00003149, -0.00000128, -0.00003405, },
		{ -0.00003610, -0.00000119, 0.00002839, -0.00001330, -0.00000437, -0.00001130, 0.00001758, -0.00001261, -0.00003582, -0.00010761, 0.00007144, -0.00003045, 0.00011977, -0.00007171, -0.00002274, -0.00048210, 0.00118464, -0.00019917, -0.00024816, 0.00001676, -0.00011686, 0.00005610, -0.00009185, 0.00002635, 0.00000439, 0.00002191, -0.00000756, -0.00001505, 0.00000588, -0.00000533, -0.00000908, 0.00000093, 0.00001988, -0.00000249, -0.00003903, 0.00002331, -0.00000733, 0.00000352, -0.00000174, -0.00000829, 0.00001226, -0.00000161, -0.00000095, 0.00000637, -0.00001503, 0.00000359, -0.00001375, 0.00000920, 0.00000143, -0.00000118, 0.00001575, -0.00001140, 0.00000733, -0.00001399, 0.00000438, 0.00001036, -0.00001512, -0.00000280, -0.00001070, 0.00000197, 0.00000776, -0.00000967, 0.00003056, -0.00000385, },
		{ -0.00000512, -0.00003842, 0.00000782, 0.00001943, -0.00000655, -0.00001451, 0.00000204, 0.00000341, 0.00000216, -0.00005433, -0.00008600, 0.00006286, -0.00001785, 0.00009519, -0.00006244, -0.00006049, -0.00044681, 0.00127504, -0.00034459, -0.00016656, -0.00003152, -0.00007781, 0.00003317, -0.00007896, 0.00002148, 0.00000895, 0.00001760, -0.00000558, -0.00002022, 0.00001526, -0.00001512, -0.00000149, -0.00000512, 0.00002722, -0.00001303, -0.00003129, 0.00002102, -0.00000535, 0.00000190, -0.00000071, -0.00001036, 0.00001610, -0.00000611, 0.00000299, 0.00000211, -0.00001251, 0.00000248, -0.00001389, 0.00001136, -0.00000028, 0.00000142, 0.00001363, -0.00001156, 0.00000777, -0.00001537, 0.00000638, 0.00000897, -0.00001591, -0.00000054, -0.00001374, 0.00000735, 0.00000192, -0.00000387, 0.00002748, },
		{ 0.00002227, -0.00000492, -0.00004104, 0.00001719, 0.00000931, 0.00000138, -0.00002572, 0.00001731, -0.00001394, 0.00002112, -0.00007795, -0.00005698, 0.00004547, 0.00000199, 0.00006213, -0.00004197, -0.00011317, -0.00038718, 0.00134282, -0.00048579, -0.00007780, -0.00008748, -0.00003076, 0.00000207, -0.00005863, 0.00001101, 0.00001709, 0.00001072, -0.00000224, -0.00002550, 0.00002460, -0.00002563, 0.00000726, -0.00001204, 0.00003509, -0.00002456, -0.00002179, 0.00001720, -0.00000248, -0.00000032, 0.00000058, -0.00001235, 0.00001981, -0.00001073, 0.00000729, -0.00000272, -0.00000919, 0.00000058, -0.00001336, 0.00001323, -0.00000209, 0.00000442, 0.00001061, -0.00001070, 0.00000730, -0.00001596, 0.00000786, 0.00000775, -0.00001663, 0.00000180, -0.00001701, 0.00001317, -0.00000484, 0.00000352, },
		{ 0.00001216, 0.00001511, -0.00000324, -0.00004392, 0.00002666, -0.00000156, 0.00001002, -0.00003726, 0.00003348, -0.00003332, 0.00004314, -0.00010556, -0.00002156, 0.00001966, 0.00002915, 0.00002064, -0.00001033, -0.00018119, -0.00030093, 0.00138260, -0.00061628, 0.00001314, -0.00014771, 0.00002176, -0.00003548, -0.00003176, -0.00000482, 0.00002879, 0.00000122, 0.00000255, -0.00003083, 0.00003355, -0.00003632, 0.00001670, -0.00001947, 0.00004311, -0.00003664, -0.00001084, 0.00001198, 0.00000126, -0.00000311, 0.00000206, -0.00001416, 0.00002318, -0.00001524, 0.00001169, -0.00000791, -0.00000517, -0.00000209, -0.00001203, 0.00001462, -0.00000381, 0.00000763, 0.00000685, -0.00000888, 0.00000591, -0.00001568, 0.00000872, 0.00000677, -0.00001727, 0.00000408, -0.00002028, 0.00001913, -0.00001217, },
		{ -0.00001962, 0.00002157, 0.00000632, -0.00000017, -0.00004694, 0.00003588, -0.00001269, 0.00001882, -0.00004835, 0.00004940, -0.00005334, 0.00006671, -0.00013553, 0.00001862, -0.00001341, 0.00006291, -0.00002848, 0.00003154, -0.00026388, -0.00018697, 0.00138992, -0.00072985, 0.00010100, -0.00020828, 0.00007668, -0.00007718, 0.00000015, -0.00002523, 0.00004367, -0.00001069, 0.00000873, -0.00003609, 0.00004175, -0.00004664, 0.00002631, -0.00002697, 0.00005077, -0.00004871, 0.00000111, 0.00000559, 0.00000572, -0.00000636, 0.00000367, -0.00001569, 0.00002604, -0.00001937, 0.00001593, -0.00001320, -0.00000067, -0.00000545, -0.00000989, 0.00001538, -0.00000523, 0.00001080, 0.00000256, -0.00000622, 0.00000363, -0.00001453, 0.00000892, 0.00000605, -0.00001775, 0.00000613, -0.00002329, 0.00002489, },
		{ 0.00003006, -0.00002669, 0.00003117, -0.00000363, 0.00000405, -0.00004994, 0.00004449, -0.00002357, 0.00002720, -0.00005820, 0.00006389, -0.00007246, 0.00009010, -0.00016587, 0.00006135, -0.00005183, 0.00010181, -0.00008361, 0.00008178, -0.00035936, -0.00004568, 0.00136160, -0.00082089, 0.00018056, -0.00026506, 0.00013058, -0.00012024, 0.00003511, -0.00004905, 0.00006102, -0.00002459, 0.00001610, -0.00004111, 0.00004881, -0.00005603, 0.00003553, -0.00003404, 0.00005758, -0.00006017, 0.00001351, -0.00000165, 0.00001074, -0.00000997, 0.00000532, -0.00001685, 0.00002821, -0.00002288, 0.00001972, -0.00001832, 0.00000409, -0.00000934, -0.00000693, 0.00001539, -0.00000614, 0.00001368, -0.00000199, -0.00000292, 0.00000059, -0.00001256, 0.00000846, 0.00000563, -0.00001805, 0.00000779, -0.00002579, },
		{ -0.00002751, 0.00003429, -0.00003287, 0.00004033, -0.00001416, 0.00000909, -0.00005265, 0.00005209, -0.00003368, 0.00003461, -0.00006606, 0.00007580, -0.00008914, 0.00011143, -0.00019436, 0.00010404, -0.00009312, 0.00014363, -0.00014238, 0.00013762, -0.00046453, 0.00012098, 0.00129602, -0.00088485, 0.00024706, -0.00031401, 0.00017992, -0.00016166, 0.00007076, -0.00007469, 0.00007983, -0.00003979, 0.00002430, -0.00004568, 0.00005439, -0.00006393, 0.00004378, -0.00004020, 0.00006301, -0.00007037, 0.00002578, -0.00000934, 0.00001606, -0.00001376, 0.00000692, -0.00001754, 0.00002956, -0.00002554, 0.00002280, -0.00002295, 0.00000881, -0.00001356, -0.00000326, 0.00001460, -0.00000638, 0.00001600, -0.00000649, 0.00000075, -0.00000302, -0.00000989, 0.00000739, 0.00000548, -0.00001811, 0.00000890, },
		{ 0.00000934, -0.00002824, 0.00003722, -0.00003766, 0.00004839, -0.00002463, 0.00001452, -0.00005479, 0.00005830, -0.00004253, 0.00004055, -0.00007132, 0.00008413, -0.00010193, 0.00012890, -0.00021874, 0.00014387, -0.00013434, 0.00018562, -0.00020178, 0.00019549, -0.00057514, 0.00030944, 0.00119330, -0.00091847, 0.00029647, -0.00035148, 0.00022134, -0.00019837, 0.00010456, -0.00010035, 0.00009885, -0.00005544, 0.00003285, -0.00004955, 0.00005817, -0.00006986, 0.00005055, -0.00004500, 0.00006658, -0.00007871, 0.00003727, -0.00001701, 0.00002138, -0.00001756, 0.00000836, -0.00001773, 0.00003000, -0.00002718, 0.00002491, -0.00002681, 0.00001322, -0.00001787, 0.00000100, 0.00001300, -0.00000584, 0.00001754, -0.00001063, 0.00000451, -0.00000694, -0.00000671, 0.00000582, 0.00000557, -0.00001788, },
		{ -0.00001732, 0.00000900, -0.00002779, 0.00003857, -0.00004062, 0.00005472, -0.00003434, 0.00001986, -0.00005605, 0.00006275, -0.00004966, 0.00004462, -0.00007350, 0.00008810, -0.00010963, 0.00014086, -0.00023686, 0.00017802, -0.00017235, 0.00022466, -0.00025834, 0.00025119, -0.00068592, 0.00051457, 0.00105541, -0.00092006, 0.00032582, -0.00037447, 0.00025189, -0.00022751, 0.00013396, -0.00012407, 0.00011669, -0.00007055, 0.00004117, -0.00005243, 0.00005992, -0.00007344, 0.00005538, -0.00004805, 0.00006790, -0.00008464, 0.00004737, -0.00002420, 0.00002641, -0.00002116, 0.00000955, -0.00001737, 0.00002948, -0.00002769, 0.00002589, -0.00002965, 0.00001701, -0.00002199, 0.00000563, 0.00001066, -0.00000446, 0.00001811, -0.00001412, 0.00000803, -0.00001090, -0.00000325, 0.00000390, 0.00000581, },
		{ 0.00000614, -0.00001640, 0.00000785, -0.00002607, 0.00003815, -0.00004140, 0.00005875, -0.00004265, 0.00002460, -0.00005612, 0.00006511, -0.00005469, 0.00004650, -0.00007232, 0.00008719, -0.00011131, 0.00014602, -0.00024688, 0.00020390, -0.00020404, 0.00025748, -0.00030839, 0.00030016, -0.00079089, 0.00072987, 0.00088613, -0.00088963, 0.00033337, -0.00038090, 0.00026927, -0.00024665, 0.00015662, -0.00014396, 0.00013195, -0.00008407, 0.00004863, -0.00005406, 0.00005945, -0.00007436, 0.00005794, -0.00004907, 0.00006668, -0.00008769, 0.00005553, -0.00003044, 0.00003081, -0.00002437, 0.00001041, -0.00001647, 0.00002802, -0.00002702, 0.00002562, -0.00003128, 0.00001993, -0.00002564, 0.00001041, 0.00000769, -0.00000224, 0.00001758, -0.00001671, 0.00001101, -0.00001458, 0.00000025, 0.00000181, },
		{ -0.00000024, 0.00000645, -0.00001511, 0.00000588, -0.00002304, 0.00003587, -0.00003977, 0.00006005, -0.00004895, 0.00002828, -0.00005474, 0.00006515, -0.00005735, 0.00004604, -0.00006773, 0.00008124, -0.00010649, 0.00014350, -0.00024746, 0.00021932, -0.00022659, 0.00028100, -0.00034830, 0.00033778, -0.00088363, 0.00094778, 0.00069093, -0.00082887, 0.00031877, -0.00036976, 0.00027205, -0.00025398, 0.00017057, -0.00015831, 0.00014329, -0.00009499, 0.00005460, -0.00005421, 0.00005669, -0.00007249, 0.00005800, -0.00004792, 0.00006281, -0.00008757, 0.00006127, -0.00003532, 0.00003430, -0.00002700, 0.00001085, -0.00001506, 0.00002570, -0.00002521, 0.00002405, -0.00003157, 0.00002176, -0.00002857, 0.00001508, 0.00000427, 0.00000075, 0.00001589, -0.00001818, 0.00001317, -0.00001770, 0.00000350, },
		{ 0.00000624, -0.00000206, 0.00000664, -0.00001346, 0.00000313, -0.00001875, 0.00003175, -0.00003567, 0.00005831, -0.00005276, 0.00003049, -0.00005168, 0.00006271, -0.00005746, 0.00004320, -0.00005989, 0.00007040, -0.00009509, 0.00013295, -0.00023781, 0.00022273, -0.00023771, 0.00029252, -0.00037482, 0.00035971, -0.00095772, 0.00116002, 0.00047666, -0.00074110, 0.00028306, -0.00034118, 0.00025972, -0.00024847, 0.00017439, -0.00016574, 0.00014958, -0.00010239, 0.00005849, -0.00005268, 0.00005168, -0.00006783, 0.00005552, -0.00004459, 0.00005629, -0.00008415, 0.00006425, -0.00003847, 0.00003661, -0.00002888, 0.00001083, -0.00001321, 0.00002265, -0.00002237, 0.00002126, -0.00003048, 0.00002237, -0.00003053, 0.00001938, 0.00000061, 0.00000436, 0.00001306, -0.00001841, 0.00001429, -0.00001999, },
		{ -0.00002122, 0.00000825, -0.00000344, 0.00000661, -0.00001147, -0.00000028, -0.00001336, 0.00002593, -0.00002918, 0.00005343, -0.00005374, 0.00003090, -0.00004685, 0.00005775, -0.00005500, 0.00003811, -0.00004919, 0.00005516, -0.00007753, 0.00011459, -0.00021788, 0.00021335, -0.00023582, 0.00029001, -0.00038533, 0.00036224, -0.00100713, 0.00135808, 0.00025124, -0.00063104, 0.00022862, -0.00029643, 0.00023282, -0.00022994, 0.00016732, -0.00016533, 0.00014999, -0.00010556, 0.00005984, -0.00004936, 0.00004454, -0.00006050, 0.00005056, -0.00003923, 0.00004734, -0.00007746, 0.00006431, -0.00003967, 0.00003755, -0.00002989, 0.00001033, -0.00001101, 0.00001905, -0.00001868, 0.00001736, -0.00002806, 0.00002170, -0.00003136, 0.00002305, -0.00000305, 0.00000842, 0.00000919, -0.00001735, 0.00001422, },
		{ 0.00001288, -0.00002126, 0.00000934, -0.00000422, 0.00000629, -0.00000919, -0.00000421, -0.00000707, 0.00001867, -0.00002053, 0.00004547, -0.00005171, 0.00002933, -0.00004021, 0.00005036, -0.00005004, 0.00003102, -0.00003621, 0.00003632, -0.00005462, 0.00008921, -0.00018827, 0.00019123, -0.00022022, 0.00027228, -0.00037810, 0.00034258, -0.00102664, 0.00153363, 0.00002321, -0.00050454, 0.00015897, -0.00023788, 0.00019286, -0.00019912, 0.00014931, -0.00015665, 0.00014406, -0.00010400, 0.00005830, -0.00004423, 0.00003554, -0.00005080, 0.00004335, -0.00003212, 0.00003632, -0.00006776, 0.00006143, -0.00003879, 0.00003701, -0.00002995, 0.00000933, -0.00000856, 0.00001510, -0.00001439, 0.00001258, -0.00002446, 0.00001975, -0.00003096, 0.00002588, -0.00000646, 0.00001268, 0.00000447, -0.00001506, },
		{ -0.00001167, 0.00001030, -0.00002003, 0.00000939, -0.00000427, 0.00000560, -0.00000668, -0.00000847, -0.00000017, 0.00001034, -0.00001011, 0.00003471, -0.00004671, 0.00002574, -0.00003188, 0.00004073, -0.00004281, 0.00002228, -0.00002165, 0.00001494, -0.00002757, 0.00005810, -0.00015032, 0.00015728, -0.00019119, 0.00023916, -0.00035246, 0.00029914, -0.00101222, 0.00167907, -0.00019875, -0.00036816, 0.00007853, -0.00016876, 0.00014225, -0.00015758, 0.00012109, -0.00013990, 0.00013174, -0.00009756, 0.00005371, -0.00003739, 0.00002500, -0.00003914, 0.00003426, -0.00002366, 0.00002374, -0.00005545, 0.00005576, -0.00003583, 0.00003495, -0.00002904, 0.00000787, -0.00000600, 0.00001104, -0.00000976, 0.00000719, -0.00001988, 0.00001666, -0.00002929, 0.00002771, -0.00000940, 0.00001688, -0.00000088, },
		{ -0.00000656, -0.00000738, 0.00000659, -0.00001755, 0.00000834, -0.00000352, 0.00000451, -0.00000404, -0.00001286, 0.00000700, 0.00000137, 0.00000154, 0.00002158, -0.00003894, 0.00002023, -0.00002208, 0.00002921, -0.00003362, 0.00001236, -0.00000633, -0.00000777, 0.00000212, 0.00002297, -0.00010594, 0.00011324, -0.00014995, 0.00019152, -0.00030896, 0.00023177, -0.00096135, 0.00178794, -0.00040628, -0.00022882, -0.00000769, -0.00009299, 0.00008409, -0.00010765, 0.00008407, -0.00011580, 0.00011345, -0.00008637, 0.00004613, -0.00002902, 0.00001337, -0.00002601, 0.00002373, -0.00001436, 0.00001022, -0.00004112, 0.00004765, -0.00003093, 0.00003145, -0.00002720, 0.00000600, -0.00000345, 0.00000710, -0.00000509, 0.00000151, -0.00001461, 0.00001261, -0.00002642, 0.00002843, -0.00001166, 0.00002075, },
		{ 0.00002404, -0.00001226, -0.00000248, 0.00000195, -0.00001394, 0.00000623, -0.00000193, 0.00000302, -0.00000134, -0.00001714, 0.00001410, -0.00000774, 0.00001383, 0.00000669, -0.00002878, 0.00001304, -0.00001115, 0.00001621, -0.00002289, 0.00000175, 0.00000889, -0.00003050, 0.00003276, -0.00001414, -0.00005755, 0.00006158, -0.00009864, 0.00013127, -0.00024934, 0.00014177, -0.00087327, 0.00185532, -0.00059184, -0.00009334, -0.00009442, -0.00001485, 0.00002201, -0.00005226, 0.00004025, -0.00008562, 0.00009001, -0.00007091, 0.00003581, -0.00001943, 0.00000115, -0.00001200, 0.00001228, -0.00000477, -0.00000352, -0.00002546, 0.00003755, -0.00002437, 0.00002664, -0.00002450, 0.00000378, -0.00000103, 0.00000350, -0.00000070, -0.00000413, -0.00000898, 0.00000785, -0.00002248, 0.00002802, -0.00001308, },
		{ -0.00001356, 0.00002652, -0.00001767, 0.00000271, -0.00000335, -0.00000938, 0.00000316, 0.00000044, 0.00000116, 0.00000130, -0.00002109, 0.00002076, -0.00001649, 0.00002606, -0.00000926, -0.00001675, 0.00000455, 0.00000047, 0.00000229, -0.00001112, -0.00000902, 0.00002319, -0.00005195, 0.00006262, -0.00005107, -0.00000787, 0.00000530, -0.00004017, 0.00006128, -0.00017653, 0.00003200, -0.00074910, 0.00187812, -0.00074910, 0.00003200, -0.00017653, 0.00006128, -0.00004017, 0.00000530, -0.00000787, -0.00005107, 0.00006262, -0.00005195, 0.00002319, -0.00000902, -0.00001112, 0.00000229, 0.00000047, 0.00000455, -0.00001675, -0.00000926, 0.00002606, -0.00001649, 0.00002076, -0.00002109, 0.00000130, 0.00000116, 0.00000044, 0.00000316, -0.00000938, -0.00000335, 0.00000271, -0.00001767, 0.00002652, },
		{ 0.00002404, -0.00001308, 0.00002802, -0.00002248, 0.00000785, -0.00000898, -0.00000413, -0.00000070, 0.00000350, -0.00000103, 0.00000378, -0.00002450, 0.00002664, -0.00002437, 0.00003755, -0.00002546, -0.00000352, -0.00000477, 0.00001228, -0.00001200, 0.00000115, -0.00001943, 0.00003581, -0.00007091, 0.00009001, -0.00008562, 0.00004025, -0.00005226, 0.00002201, -0.00001485, -0.00009442, -0.00009334, -0.00059184, 0.00185532, -0.00087327, 0.00014177, -0.00024934, 0.00013127, -0.00009864, 0.00006158, -0.00005755, -0.00001414, 0.00003276, -0.00003050, 0.00000889, 0.00000175, -0.00002289, 0.00001621, -0.00001115, 0.00001304, -0.00002878, 0.00000669, 0.00001383, -0.00000774, 0.00001410, -0.00001714, -0.00000134, 0.00000302, -0.00000193, 0.00000623, -0.00001394, 0.00000195, -0.00000248, -0.00001226, },
		{ -0.00000656, 0.00002075, -0.00001166, 0.00002843, -0.00002642, 0.00001261, -0.00001461, 0.00000151, -0.00000509, 0.00000710, -0.00000345, 0.00000600, -0.00002720, 0.00003145, -0.00003093, 0.00004765, -0.00004112, 0.00001022, -0.00001436, 0.00002373, -0.00002601, 0.00001337, -0.00002902, 0.00004613, -0.00008637, 0.00011345, -0.00011580, 0.00008407, -0.00010765, 0.00008409, -0.00009299, -0.00000769, -0.00022882, -0.00040628, 0.00178794, -0.00096135, 0.00023177, -0.00030896, 0.00019152, -0.00014995, 0.00011324, -0.00010594, 0.00002297, 0.00000212, -0.00000777, -0.00000633, 0.00001236, -0.00003362, 0.00002921, -0.00002208, 0.00002023, -0.00003894, 0.00002158, 0.00000154, 0.00000137, 0.00000700, -0.00001286, -0.00000404, 0.00000451, -0.00000352, 0.00000834, -0.00001755, 0.00000659, -0.00000738, },
		{ -0.00001167, -0.00000088, 0.00001688, -0.00000940, 0.00002771, -0.00002929, 0.00001666, -0.00001988, 0.00000719, -0.00000976, 0.00001104, -0.00000600, 0.00000787, -0.00002904, 0.00003495, -0.00003583, 0.00005576, -0.00005545, 0.00002374, -0.00002366, 0.00003426, -0.00003914, 0.00002500, -0.00003739, 0.00005371, -0.00009756, 0.00013174, -0.00013990, 0.00012109, -0.00015758, 0.00014225, -0.00016876, 0.00007853, -0.00036816, -0.00019875, 0.00167907, -0.00101222, 0.00029914, -0.00035246, 0.00023916, -0.00019119, 0.00015728, -0.00015032, 0.00005810, -0.00002757, 0.00001494, -0.00002165, 0.00002228, -0.00004281, 0.00004073, -0.00003188, 0.00002574, -0.00004671, 0.00003471, -0.00001011, 0.00001034, -0.00000017, -0.00000847, -0.00000668, 0.00000560, -0.00000427, 0.00000939, -0.00002003, 0.00001030, },
		{ 0.00001288, -0.00001506, 0.00000447, 0.00001268, -0.00000646, 0.00002588, -0.00003096, 0.00001975, -0.00002446, 0.00001258, -0.00001439, 0.00001510, -0.00000856, 0.00000933, -0.00002995, 0.00003701, -0.00003879, 0.00006143, -0.00006776, 0.00003632, -0.00003212, 0.00004335, -0.00005080, 0.00003554, -0.00004423, 0.00005830, -0.00010400, 0.00014406, -0.00015665, 0.00014931, -0.00019912, 0.00019286, -0.00023788, 0.00015897, -0.00050454, 0.00002321, 0.00153363, -0.00102664, 0.00034258, -0.00037810, 0.00027228, -0.00022022, 0.00019123, -0.00018827, 0.00008921, -0.00005462, 0.00003632, -0.00003621, 0.00003102, -0.00005004, 0.00005036, -0.00004021, 0.00002933, -0.00005171, 0.00004547, -0.00002053, 0.00001867, -0.00000707, -0.00000421, -0.00000919, 0.00000629, -0.00000422, 0.00000934, -0.00002126, },
		{ -0.00002122, 0.00001422, -0.00001735, 0.00000919, 0.00000842, -0.00000305, 0.00002305, -0.00003136, 0.00002170, -0.00002806, 0.00001736, -0.00001868, 0.00001905, -0.00001101, 0.00001033, -0.00002989, 0.00003755, -0.00003967, 0.00006431, -0.00007746, 0.00004734, -0.00003923, 0.00005056, -0.00006050, 0.00004454, -0.00004936, 0.00005984, -0.00010556, 0.00014999, -0.00016533, 0.00016732, -0.00022994, 0.00023282, -0.00029643, 0.00022862, -0.00063104, 0.00025124, 0.00135808, -0.00100713, 0.00036224, -0.00038533, 0.00029001, -0.00023582, 0.00021335, -0.00021788, 0.00011459, -0.00007753, 0.00005516, -0.00004919, 0.00003811, -0.00005500, 0.00005775, -0.00004685, 0.00003090, -0.00005374, 0.00005343, -0.00002918, 0.00002593, -0.00001336, -0.00000028, -0.00001147, 0.00000661, -0.00000344, 0.00000825, },
		{ 0.00000624, -0.00001999, 0.00001429, -0.00001841, 0.00001306, 0.00000436, 0.00000061, 0.00001938, -0.00003053, 0.00002237, -0.00003048, 0.00002126, -0.00002237, 0.00002265, -0.00001321, 0.00001083, -0.00002888, 0.00003661, -0.00003847, 0.00006425, -0.00008415, 0.00005629, -0.00004459, 0.00005552, -0.00006783, 0.00005168, -0.00005268, 0.00005849, -0.00010239, 0.00014958, -0.00016574, 0.00017439, -0.00024847, 0.00025972, -0.00034118, 0.00028306, -0.00074110, 0.00047666, 0.00116002, -0.00095772, 0.00035971, -0.00037482, 0.00029252, -0.00023771, 0.00022273, -0.00023781, 0.00013295, -0.00009509, 0.00007040, -0.00005989, 0.00004320, -0.00005746, 0.00006271, -0.00005168, 0.00003049, -0.00005276, 0.00005831, -0.00003567, 0.00003175, -0.00001875, 0.00000313, -0.00001346, 0.00000664, -0.00000206, },
		{ -0.00000024, 0.00000350, -0.00001770, 0.00001317, -0.00001818, 0.00001589, 0.00000075, 0.00000427, 0.00001508, -0.00002857, 0.00002176, -0.00003157, 0.00002405, -0.00002521, 0.00002570, -0.00001506, 0.00001085, -0.00002700, 0.00003430, -0.00003532, 0.00006127, -0.00008757, 0.00006281, -0.00004792, 0.00005800, -0.00007249, 0.00005669, -0.00005421, 0.00005460, -0.00009499, 0.00014329, -0.00015831, 0.00017057, -0.00025398, 0.00027205, -0.00036976, 0.00031877, -0.00082887, 0.00069093, 0.00094778, -0.00088363, 0.00033778, -0.00034830, 0.00028100, -0.00022659, 0.00021932, -0.00024746, 0.00014350, -0.00010649, 0.00008124, -0.00006773, 0.00004604, -0.00005735, 0.00006515, -0.00005474, 0.00002828, -0.00004895, 0.00006005, -0.00003977, 0.00003587, -0.00002304, 0.00000588, -0.00001511, 0.00000645, },
		{ 0.00000614, 0.00000181, 0.00000025, -0.00001458, 0.00001101, -0.00001671, 0.00001758, -0.00000224, 0.00000769, 0.00001041, -0.00002564, 0.00001993, -0.00003128, 0.00002562, -0.00002702, 0.00002802, -0.00001647, 0.00001041, -0.00002437, 0.00003081, -0.00003044, 0.00005553, -0.00008769, 0.00006668, -0.00004907, 0.00005794, -0.00007436, 0.00005945, -0.00005406, 0.00004863, -0.00008407, 0.00013195, -0.00014396, 0.00015662, -0.00024665, 0.00026927, -0.00038090, 0.00033337, -0.00088963, 0.00088613, 0.00072987, -0.00079089, 0.00030016, -0.00030839, 0.00025748, -0.00020404, 0.00020390, -0.00024688, 0.00014602, -0.00011131, 0.00008719, -0.00007232, 0.00004650, -0.00005469, 0.00006511, -0.00005612, 0.00002460, -0.00004265, 0.00005875, -0.00004140, 0.00003815, -0.00002607, 0.00000785, -0.00001640, },
		{ -0.00001732, 0.00000581, 0.00000390, -0.00000325, -0.00001090, 0.00000803, -0.00001412, 0.00001811, -0.00000446, 0.00001066, 0.00000563, -0.00002199, 0.00001701, -0.00002965, 0.00002589, -0.00002769, 0.00002948, -0.00001737, 0.00000955, -0.00002116, 0.00002641, -0.00002420, 0.00004737, -0.00008464, 0.00006790, -0.00004805, 0.00005538, -0.00007344, 0.00005992, -0.00005243, 0.00004117, -0.00007055, 0.00011669, -0.00012407, 0.00013396, -0.00022751, 0.00025189, -0.00037447, 0.00032582, -0.00092006, 0.00105541, 0.00051457, -0.00068592, 0.00025119, -0.00025834, 0.00022466, -0.00017235, 0.00017802, -0.00023686, 0.00014086, -0.00010963, 0.00008810, -0.00007350, 0.00004462, -0.00004966, 0.00006275, -0.00005605, 0.00001986, -0.00003434, 0.00005472, -0.00004062, 0.00003857, -0.00002779, 0.00000900, },
		{ 0.00000934, -0.00001788, 0.00000557, 0.00000582, -0.00000671, -0.00000694, 0.00000451, -0.00001063, 0.00001754, -0.00000584, 0.00001300, 0.00000100, -0.00001787, 0.00001322, -0.00002681, 0.00002491, -0.00002718, 0.00003000, -0.00001773, 0.00000836, -0.00001756, 0.00002138, -0.00001701, 0.00003727, -0.00007871, 0.00006658, -0.00004500, 0.00005055, -0.00006986, 0.00005817, -0.00004955, 0.00003285, -0.00005544, 0.00009885, -0.00010035, 0.00010456, -0.00019837, 0.00022134, -0.00035148, 0.00029647, -0.00091847, 0.00119330, 0.00030944, -0.00057514, 0.00019549, -0.00020178, 0.00018562, -0.00013434, 0.00014387, -0.00021874, 0.00012890, -0.00010193, 0.00008413, -0.00007132, 0.00004055, -0.00004253, 0.00005830, -0.00005479, 0.00001452, -0.00002463, 0.00004839, -0.00003766, 0.00003722, -0.00002824, },
		{ -0.00002751, 0.00000890, -0.00001811, 0.00000548, 0.00000739, -0.00000989, -0.00000302, 0.00000075, -0.00000649, 0.00001600, -0.00000638, 0.00001460, -0.00000326, -0.00001356, 0.00000881, -0.00002295, 0.00002280, -0.00002554, 0.00002956, -0.00001754, 0.00000692, -0.00001376, 0.00001606, -0.00000934, 0.00002578, -0.00007037, 0.00006301, -0.00004020, 0.00004378, -0.00006393, 0.00005439, -0.00004568, 0.00002430, -0.00003979, 0.00007983, -0.00007469, 0.00007076, -0.00016166, 0.00017992, -0.00031401, 0.00024706, -0.00088485, 0.00129602, 0.00012098, -0.00046453, 0.00013762, -0.00014238, 0.00014363, -0.00009312, 0.00010404, -0.00019436, 0.00011143, -0.00008914, 0.00007580, -0.00006606, 0.00003461, -0.00003368, 0.00005209, -0.00005265, 0.00000909, -0.00001416, 0.00004033, -0.00003287, 0.00003429, },
		{ 0.00003006, -0.00002579, 0.00000779, -0.00001805, 0.00000563, 0.00000846, -0.00001256, 0.00000059, -0.00000292, -0.00000199, 0.00001368, -0.00000614, 0.00001539, -0.00000693, -0.00000934, 0.00000409, -0.00001832, 0.00001972, -0.00002288, 0.00002821, -0.00001685, 0.00000532, -0.00000997, 0.00001074, -0.00000165, 0.00001351, -0.00006017, 0.00005758, -0.00003404, 0.00003553, -0.00005603, 0.00004881, -0.00004111, 0.00001610, -0.00002459, 0.00006102, -0.00004905, 0.00003511, -0.00012024, 0.00013058, -0.00026506, 0.00018056, -0.00082089, 0.00136160, -0.00004568, -0.00035936, 0.00008178, -0.00008361, 0.00010181, -0.00005183, 0.00006135, -0.00016587, 0.00009010, -0.00007246, 0.00006389, -0.00005820, 0.00002720, -0.00002357, 0.00004449, -0.00004994, 0.00000405, -0.00000363, 0.00003117, -0.00002669, },
		{ -0.00001962, 0.00002489, -0.00002329, 0.00000613, -0.00001775, 0.00000605, 0.00000892, -0.00001453, 0.00000363, -0.00000622, 0.00000256, 0.00001080, -0.00000523, 0.00001538, -0.00000989, -0.00000545, -0.00000067, -0.00001320, 0.00001593, -0.00001937, 0.00002604, -0.00001569, 0.00000367, -0.00000636, 0.00000572, 0.00000559, 0.00000111, -0.00004871, 0.00005077, -0.00002697, 0.00002631, -0.00004664, 0.00004175, -0.00003609, 0.00000873, -0.00001069, 0.00004367, -0.00002523, 0.00000015, -0.00007718, 0.00007668, -0.00020828, 0.00010100, -0.00072985, 0.00138992, -0.00018697, -0.00026388, 0.00003154, -0.00002848, 0.00006291, -0.00001341, 0.00001862, -0.00013553, 0.00006671, -0.00005334, 0.00004940, -0.00004835, 0.00001882, -0.00001269, 0.00003588, -0.00004694, -0.00000017, 0.00000632, 0.00002157, },
		{ 0.00001216, -0.00001217, 0.00001913, -0.00002028, 0.00000408, -0.00001727, 0.00000677, 0.00000872, -0.00001568, 0.00000591, -0.00000888, 0.00000685, 0.00000763, -0.00000381, 0.00001462, -0.00001203, -0.00000209, -0.00000517, -0.00000791, 0.00001169, -0.00001524, 0.00002318, -0.00001416, 0.00000206, -0.00000311, 0.00000126, 0.00001198, -0.00001084, -0.00003664, 0.00004311, -0.00001947, 0.00001670, -0.00003632, 0.00003355, -0.00003083, 0.00000255, 0.00000122, 0.00002879, -0.00000482, -0.00003176, -0.00003548, 0.00002176, -0.00014771, 0.00001314, -0.00061628, 0.00138260, -0.00030093, -0.00018119, -0.00001033, 0.00002064, 0.00002915, 0.00001966, -0.00002156, -0.00010556, 0.00004314, -0.00003332, 0.00003348, -0.00003726, 0.00001002, -0.00000156, 0.00002666, -0.00004392, -0.00000324, 0.00001511, },
		{ 0.00002227, 0.00000352, -0.00000484, 0.00001317, -0.00001701, 0.00000180, -0.00001663, 0.00000775, 0.00000786, -0.00001596, 0.00000730, -0.00001070, 0.00001061, 0.00000442, -0.00000209, 0.00001323, -0.00001336, 0.00000058, -0.00000919, -0.00000272, 0.00000729, -0.00001073, 0.00001981, -0.00001235, 0.00000058, -0.00000032, -0.00000248, 0.00001720, -0.00002179, -0.00002456, 0.00003509, -0.00001204, 0.00000726, -0.00002563, 0.00002460, -0.00002550, -0.00000224, 0.00001072, 0.00001709, 0.00001101, -0.00005863, 0.00000207, -0.00003076, -0.00008748, -0.00007780, -0.00048579, 0.00134282, -0.00038718, -0.00011317, -0.00004197, 0.00006213, 0.00000199, 0.00004547, -0.00005698, -0.00007795, 0.00002112, -0.00001394, 0.00001731, -0.00002572, 0.00000138, 0.00000931, 0.00001719, -0.00004104, -0.00000492, },
		{ -0.00000512, 0.00002748, -0.00000387, 0.00000192, 0.00000735, -0.00001374, -0.00000054, -0.00001591, 0.00000897, 0.00000638, -0.00001537, 0.00000777, -0.00001156, 0.00001363, 0.00000142, -0.00000028, 0.00001136, -0.00001389, 0.00000248, -0.00001251, 0.00000211, 0.00000299, -0.00000611, 0.00001610, -0.00001036, -0.00000071, 0.00000190, -0.00000535, 0.00002102, -0.00003129, -0.00001303, 0.00002722, -0.00000512, -0.00000149, -0.00001512, 0.00001526, -0.00002022, -0.00000558, 0.00001760, 0.00000895, 0.00002148, -0.00007896, 0.00003317, -0.00007781, -0.00003152, -0.00016656, -0.00034459, 0.00127504, -0.00044681, -0.00006049, -0.00006244, 0.00009519, -0.00001785, 0.00006286, -0.00008600, -0.00005433, 0.00000216, 0.00000341, 0.00000204, -0.00001451, -0.00000655, 0.00001943, 0.00000782, -0.00003842, },
		{ -0.00003610, -0.00000385, 0.00003056, -0.00000967, 0.00000776, 0.00000197, -0.00001070, -0.00000280, -0.00001512, 0.00001036, 0.00000438, -0.00001399, 0.00000733, -0.00001140, 0.00001575, -0.00000118, 0.00000143, 0.00000920, -0.00001375, 0.00000359, -0.00001503, 0.00000637, -0.00000095, -0.00000161, 0.00001226, -0.00000829, -0.00000174, 0.00000352, -0.00000733, 0.00002331, -0.00003903, -0.00000249, 0.00001988, 0.00000093, -0.00000908, -0.00000533, 0.00000588, -0.00001505, -0.00000756, 0.00002191, 0.00000439, 0.00002635, -0.00009185, 0.00005610, -0.00011686, 0.00001676, -0.00024816, -0.00019917, 0.00118464, -0.00048210, -0.00002274, -0.00007171, 0.00011977, -0.00003045, 0.00007144, -0.00010761, -0.00003582, -0.00001261, 0.00001758, -0.00001130, -0.00000437, -0.00001330, 0.00002839, -0.00000119, },
		{ -0.00000959, -0.00003405, -0.00000128, 0.00003149, -0.00001365, 0.00001239, -0.00000273, -0.00000807, -0.00000483, -0.00001429, 0.00001184, 0.00000197, -0.00001193, 0.00000607, -0.00001025, 0.00001689, -0.00000321, 0.00000285, 0.00000694, -0.00001305, 0.00000397, -0.00001668, 0.00000991, -0.00000436, 0.00000252, 0.00000846, -0.00000625, -0.00000249, 0.00000454, -0.00000843, 0.00002407, -0.00004481, 0.00000671, 0.00001339, 0.00000584, -0.00001516, 0.00000330, -0.00000319, -0.00001000, -0.00000841, 0.00002393, 0.00000308, 0.00002583, -0.00009703, 0.00006987, -0.00014615, 0.00005465, -0.00031824, -0.00005584, 0.00107753, -0.00049628, 0.00000142, -0.00007045, 0.00013649, -0.00003660, 0.00007159, -0.00012151, -0.00002302, -0.00002248, 0.00002772, -0.00002190, 0.00000411, -0.00001845, 0.00003584, },
		{ 0.00004152, -0.00001721, -0.00003218, 0.00000235, 0.00003041, -0.00001575, 0.00001568, -0.00000662, -0.00000597, -0.00000655, -0.00001345, 0.00001331, -0.00000068, -0.00000937, 0.00000415, -0.00000821, 0.00001704, -0.00000458, 0.00000383, 0.00000477, -0.00001198, 0.00000372, -0.00001747, 0.00001267, -0.00000709, 0.00000613, 0.00000487, -0.00000432, -0.00000295, 0.00000500, -0.00000874, 0.00002338, -0.00004858, 0.00001437, 0.00000794, 0.00000946, -0.00001950, 0.00001041, -0.00001168, -0.00000505, -0.00000843, 0.00002410, 0.00000442, 0.00002059, -0.00009487, 0.00007424, -0.00016472, 0.00008036, -0.00037333, 0.00007959, 0.00095977, -0.00049321, 0.00001397, -0.00005993, 0.00014648, -0.00003767, 0.00006435, -0.00012803, -0.00001596, -0.00002721, 0.00003336, -0.00002916, 0.00001046, -0.00002172, },
		{ -0.00002294, 0.00004527, -0.00002393, -0.00003037, 0.00000669, 0.00002758, -0.00001604, 0.00001756, -0.00000960, -0.00000446, -0.00000789, -0.00001260, 0.00001469, -0.00000343, -0.00000650, 0.00000175, -0.00000545, 0.00001627, -0.00000526, 0.00000427, 0.00000285, -0.00001072, 0.00000297, -0.00001748, 0.00001463, -0.00000907, 0.00000907, 0.00000160, -0.00000255, -0.00000313, 0.00000496, -0.00000839, 0.00002142, -0.00005039, 0.00002040, 0.00000362, 0.00001176, -0.00002200, 0.00001574, -0.00001935, -0.00000017, -0.00000798, 0.00002300, 0.00000762, 0.00001169, -0.00008629, 0.00006968, -0.00017247, 0.00009310, -0.00041102, 0.00020218, 0.00083719, -0.00047699, 0.00001733, -0.00004178, 0.00015119, -0.00003542, 0.00005131, -0.00012809, -0.00001414, -0.00002703, 0.00003443, -0.00003279, 0.00001441, },
		{ 0.00001588, -0.00002211, 0.00004703, -0.00002968, -0.00002847, 0.00001137, 0.00002337, -0.00001470, 0.00001810, -0.00001171, -0.00000356, -0.00000881, -0.00001174, 0.00001589, -0.00000610, -0.00000350, -0.00000089, -0.00000217, 0.00001469, -0.00000527, 0.00000413, 0.00000134, -0.00000943, 0.00000192, -0.00001682, 0.00001584, -0.00001031, 0.00001129, -0.00000125, -0.00000098, -0.00000306, 0.00000451, -0.00000756, 0.00001844, -0.00005044, 0.00002485, 0.00000039, 0.00001283, -0.00002270, 0.00001915, -0.00002600, 0.00000470, -0.00000744, 0.00002126, 0.00001172, 0.00000041, -0.00007266, 0.00005734, -0.00017015, 0.00009307, -0.00043013, 0.00030805, 0.00071506, -0.00045173, 0.00001410, -0.00001777, 0.00015215, -0.00003181, 0.00003445, -0.00012304, -0.00001663, -0.00002261, 0.00003124, -0.00003280, },
		{ -0.00002946, 0.00001498, -0.00001932, 0.00004682, -0.00003441, -0.00002635, 0.00001599, 0.00001821, -0.00001201, 0.00001743, -0.00001301, -0.00000321, -0.00000932, -0.00001087, 0.00001682, -0.00000856, -0.00000058, -0.00000357, 0.00000140, 0.00001250, -0.00000470, 0.00000341, 0.00000030, -0.00000829, 0.00000073, -0.00001568, 0.00001642, -0.00001086, 0.00001278, -0.00000363, 0.00000037, -0.00000279, 0.00000376, -0.00000644, 0.00001474, -0.00004896, 0.00002785, -0.00000187, 0.00001285, -0.00002175, 0.00002059, -0.00003148, 0.00000958, -0.00000717, 0.00001947, 0.00001578, -0.00001180, -0.00005569, 0.00003886, -0.00015917, 0.00008143, -0.00043067, 0.00039457, 0.00059786, -0.00042123, 0.00000678, 0.00001038, 0.00015080, -0.00002873, 0.00001588, -0.00011457, -0.00002214, -0.00001496, 0.00002443, },
		{ 0.00001492, -0.00002328, 0.00001200, -0.00001482, 0.00004475, -0.00003812, -0.00002391, 0.00002016, 0.00001255, -0.00000829, 0.00001575, -0.00001363, -0.00000331, -0.00000946, -0.00000998, 0.00001744, -0.00001070, 0.00000212, -0.00000607, 0.00000501, 0.00000988, -0.00000367, 0.00000217, -0.00000022, -0.00000742, -0.00000042, -0.00001421, 0.00001651, -0.00001084, 0.00001359, -0.00000555, 0.00000152, -0.00000239, 0.00000282, -0.00000522, 0.00001062, -0.00004628, 0.00002960, -0.00000334, 0.00001212, -0.00001942, 0.00002015, -0.00003570, 0.00001447, -0.00000744, 0.00001818, 0.00001890, -0.00002352, -0.00003724, 0.00001624, -0.00014152, 0.00006014, -0.00041381, 0.00046038, 0.00048910, -0.00038878, -0.00000236, 0.00004120, 0.00014826, -0.00002787, -0.00000233, -0.00010442, -0.00002922, -0.00000530, },
		{ 0.00000499, 0.00000380, -0.00001497, 0.00000736, -0.00000892, 0.00004101, -0.00004080, -0.00002108, 0.00002358, 0.00000684, -0.00000392, 0.00001331, -0.00001374, -0.00000373, -0.00000927, -0.00000906, 0.00001770, -0.00001241, 0.00000446, -0.00000823, 0.00000843, 0.00000705, -0.00000235, 0.00000053, -0.00000023, -0.00000689, -0.00000138, -0.00001262, 0.00001629, -0.00001040, 0.00001381, -0.00000703, 0.00000250, -0.00000191, 0.00000178, -0.00000407, 0.00000639, -0.00004273, 0.00003035, -0.00000427, 0.00001094, -0.00001605, 0.00001799, -0.00003860, 0.00001933, -0.00000843, 0.00001780, 0.00002037, -0.00003344, -0.00001915, -0.00000836, -0.00011952, 0.00003177, -0.00038170, 0.00050537, 0.00039120, -0.00035700, -0.00001146, 0.00007357, 0.00014523, -0.00003044, -0.00001837, -0.00009431, -0.00003639, },
		{ -0.00004228, 0.00001457, -0.00000776, -0.00000534, 0.00000158, -0.00000204, 0.00003584, -0.00004246, -0.00001785, 0.00002599, 0.00000149, 0.00000075, 0.00001034, -0.00001352, -0.00000434, -0.00000884, -0.00000810, 0.00001759, -0.00001367, 0.00000636, -0.00000992, 0.00001148, 0.00000423, -0.00000092, -0.00000140, 0.00000022, -0.00000674, -0.00000202, -0.00001105, 0.00001593, -0.00000972, 0.00001357, -0.00000815, 0.00000335, -0.00000142, 0.00000073, -0.00000312, 0.00000231, -0.00003865, 0.00003037, -0.00000488, 0.00000965, -0.00001205, 0.00001435, -0.00004015, 0.00002405, -0.00001020, 0.00001854, 0.00001968, -0.00004052, -0.00000308, -0.00003281, -0.00009562, -0.00000076, -0.00033726, 0.00053054, 0.00030557, -0.00032784, -0.00001911, 0.00010677, 0.00014190, -0.00003713, -0.00003083, -0.00008571, },
		{ -0.00007968, -0.00004576, 0.00002225, -0.00001862, 0.00000474, -0.00000476, 0.00000539, 0.00002951, -0.00004313, -0.00001429, 0.00002726, -0.00000318, 0.00000540, 0.00000707, -0.00001313, -0.00000498, -0.00000823, -0.00000712, 0.00001714, -0.00001446, 0.00000780, -0.00001106, 0.00001401, 0.00000159, 0.00000047, -0.00000343, 0.00000101, -0.00000696, -0.00000229, -0.00000965, 0.00001557, -0.00000896, 0.00001301, -0.00000899, 0.00000413, -0.00000099, -0.00000026, -0.00000248, -0.00000138, -0.00003432, 0.00002990, -0.00000541, 0.00000857, -0.00000782, 0.00000951, -0.00004035, 0.00002851, -0.00001270, 0.00002045, 0.00001662, -0.00004405, 0.00000959, -0.00005522, -0.00007217, -0.00003441, -0.00028394, 0.00053784, 0.00023263, -0.00030250, -0.00002439, 0.00014051, 0.00013790, -0.00004798, -0.00003886, },
		{ -0.00004220, -0.00007686, -0.00004609, 0.00002709, -0.00002777, 0.00001445, -0.00001109, 0.00001289, 0.00002231, -0.00004280, -0.00001052, 0.00002736, -0.00000695, 0.00000977, 0.00000370, -0.00001270, -0.00000553, -0.00000754, -0.00000612, 0.00001637, -0.00001480, 0.00000878, -0.00001164, 0.00001593, -0.00000073, 0.00000165, -0.00000542, 0.00000203, -0.00000748, -0.00000216, -0.00000850, 0.00001535, -0.00000828, 0.00001228, -0.00000963, 0.00000491, -0.00000067, -0.00000116, -0.00000219, -0.00000453, -0.00002999, 0.00002914, -0.00000603, 0.00000797, -0.00000377, 0.00000380, -0.00003924, 0.00003253, -0.00001574, 0.00002339, 0.00001126, -0.00004368, 0.00001794, -0.00007407, -0.00005121, -0.00006628, -0.00022537, 0.00052985, 0.00017201, -0.00028160, -0.00002677, 0.00017484, 0.00013240, -0.00006237, },
		{ -0.00007913, -0.00004120, -0.00007734, -0.00004296, 0.00002847, -0.00003442, 0.00002306, -0.00001688, 0.00002005, 0.00001456, -0.00004149, -0.00000672, 0.00002635, -0.00000969, 0.00001368, 0.00000037, -0.00001234, -0.00000588, -0.00000681, -0.00000511, 0.00001535, -0.00001477, 0.00000937, -0.00001170, 0.00001721, -0.00000264, 0.00000251, -0.00000722, 0.00000313, -0.00000821, -0.00000167, -0.00000763, 0.00001534, -0.00000782, 0.00001153, -0.00001017, 0.00000573, -0.00000049, -0.00000194, -0.00000226, -0.00000706, -0.00002584, 0.00002823, -0.00000683, 0.00000806, -0.00000025, -0.00000246, -0.00003686, 0.00003588, -0.00001906, 0.00002701, 0.00000396, -0.00003953, 0.00002151, -0.00008832, -0.00003432, -0.00009389, -0.00016514, 0.00050956, 0.00012270, -0.00026523, -0.00002604, 0.00021005, 0.00012418, },
		{ 0.00011179, -0.00009661, -0.00003677, -0.00008070, -0.00003649, 0.00002620, -0.00003803, 0.00003001, -0.00002169, 0.00002647, 0.00000655, -0.00003922, -0.00000312, 0.00002442, -0.00001139, 0.00001701, -0.00000286, -0.00001208, -0.00000598, -0.00000610, -0.00000413, 0.00001414, -0.00001444, 0.00000965, -0.00001133, 0.00001788, -0.00000409, 0.00000297, -0.00000869, 0.00000417, -0.00000901, -0.00000089, -0.00000705, 0.00001557, -0.00000765, 0.00001085, -0.00001070, 0.00000665, -0.00000050, -0.00000262, -0.00000264, -0.00000896, -0.00002196, 0.00002725, -0.00000785, 0.00000892, 0.00000244, -0.00000894, -0.00003328, 0.00003837, -0.00002230, 0.00003086, -0.00000466, -0.00003205, 0.00002033, -0.00009749, -0.00002249, -0.00011530, -0.00010651, 0.00048010, 0.00008332, -0.00025317, -0.00002215, 0.00024650, },
		{ 0.00028446, 0.00009381, -0.00011290, -0.00003025, -0.00008609, -0.00002728, 0.00002044, -0.00003841, 0.00003492, -0.00002522, 0.00003184, -0.00000145, -0.00003602, 0.00000004, 0.00002180, -0.00001211, 0.00001974, -0.00000596, -0.00001191, -0.00000579, -0.00000544, -0.00000321, 0.00001281, -0.00001390, 0.00000975, -0.00001065, 0.00001803, -0.00000512, 0.00000302, -0.00000976, 0.00000502, -0.00000976, 0.00000007, -0.00000668, 0.00001603, -0.00000781, 0.00001033, -0.00001125, 0.00000768, -0.00000070, -0.00000321, -0.00000324, -0.00001026, -0.00001839, 0.00002620, -0.00000904, 0.00001057, 0.00000413, -0.00001536, -0.00002860, 0.00003976, -0.00002509, 0.00003439, -0.00001382, -0.00002207, 0.00001492, -0.00010167, -0.00001604, -0.00012930, -0.00005216, 0.00044443, 0.00005232, -0.00024500, -0.00001507, },
		{ -0.00000464, 0.00032394, 0.00006896, -0.00012599, -0.00002327, -0.00009229, -0.00001625, 0.00001176, -0.00003566, 0.00003764, -0.00002728, 0.00003594, -0.00000917, -0.00003193, 0.00000253, 0.00001880, -0.00001202, 0.00002193, -0.00000897, -0.00001176, -0.00000535, -0.00000483, -0.00000237, 0.00001144, -0.00001324, 0.00000976, -0.00000982, 0.00001777, -0.00000577, 0.00000266, -0.00001038, 0.00000557, -0.00001030, 0.00000107, -0.00000643, 0.00001662, -0.00000828, 0.00000999, -0.00001187, 0.00000883, -0.00000108, -0.00000375, -0.00000397, -0.00001108, -0.00001509, 0.00002504, -0.00001026, 0.00001291, 0.00000469, -0.00002146, -0.00002293, 0.00003989, -0.00002704, 0.00003704, -0.00002263, -0.00001065, 0.00000620, -0.00010146, -0.00001469, -0.00013544, -0.00000411, 0.00040518, 0.00002818, -0.00024031, },
		},
		{ { 0.00064114, 0.00014573, -0.00029265, -0.00005856, -0.00005690, -0.00003176, 0.00006238, -0.00001032, -0.00004806, 0.00003262, -0.00003346, 0.00001130, -0.00002199, 0.00000696, 0.00001020, 0.00002250, -0.00004937, 0.00002174, -0.00002663, 0.00002838, -0.00002962, 0.00002871, -0.00001582, 0.00003768, -0.00004763, 0.00002625, -0.00003605, 0.00003472, -0.00004218, 0.00003771, -0.00003036, 0.00004018, -0.00004205, 0.00004018, -0.00003036, 0.00003771, -0.00004218, 0.00003472, -0.00003605, 0.00002625, -0.00004763, 0.00003768, -0.00001582, 0.00002871, -0.00002962, 0.00002838, -0.00002663, 0.00002174, -0.00004937, 0.00002250, 0.00001020, 0.00000696, -0.00002199, 0.00001130, -0.00003346, 0.00003262, -0.00004806, -0.00001032, 0.00006238, -0.00003176, -0.00005690, -0.00005856, -0.00029265, 0.00014573, },
		{ 0.00014569, 0.00064471, 0.00014079, -0.00029105, -0.00005725, -0.00005758, -0.00003115, 0.00005938, -0.00000709, -0.00004802, 0.00003111, -0.00003419, 0.00001354, -0.00002343, 0.00000869, 0.00000696, 0.00002496, -0.00004933, 0.00002150, -0.00002826, 0.00003051, -0.00003101, 0.00003039, -0.00001803, 0.00003889, -0.00004759, 0.00002655, -0.00003715, 0.00003567, -0.00004272, 0.00003826, -0.00003078, 0.00004025, -0.00004201, 0.00004005, -0.00002988, 0.00003710, -0.00004157, 0.00003371, -0.00003491, 0.00002590, -0.00004759, 0.00003646, -0.00001364, 0.00002705, -0.00002824, 0.00002632, -0.00002506, 0.00002188, -0.00004933, 0.00002023, 0.00001315, 0.00000538, -0.00002067, 0.00000933, -0.00003279, 0.00003379, -0.00004802, -0.00001284, 0.00006461, -0.00003224, -0.00005643, -0.00005928, -0.00029314, },
		{ -0.00029250, 0.00014068, 0.00065536, 0.00013092, -0.00028835, -0.00005534, -0.00005846, -0.00003041, 0.00005563, -0.00000317, -0.00004791, 0.00002926, -0.00003497, 0.00001603, -0.00002500, 0.00001055, 0.00000345, 0.00002757, -0.00004922, 0.00002118, -0.00002992, 0.00003270, -0.00003240, 0.00003208, -0.00002027, 0.00004007, -0.00004748, 0.00002677, -0.00003820, 0.00003657, -0.00004319, 0.00003873, -0.00003113, 0.00004024, -0.00004190, 0.00003984, -0.00002933, 0.00003643, -0.00004091, 0.00003267, -0.00003374, 0.00002549, -0.00004748, 0.00003523, -0.00001152, 0.00002544, -0.00002690, 0.00002437, -0.00002356, 0.00002193, -0.00004922, 0.00001816, 0.00001577, 0.00000396, -0.00001950, 0.00000764, -0.00003218, 0.00003459, -0.00004791, -0.00001462, 0.00006604, -0.00003257, -0.00005618, -0.00005939, },
		{ -0.00005889, -0.00029076, 0.00013074, 0.00067301, 0.00011621, -0.00028458, -0.00005287, -0.00005953, -0.00002954, 0.00005117, 0.00000140, -0.00004773, 0.00002708, -0.00003579, 0.00001876, -0.00002666, 0.00001251, -0.00000029, 0.00003032, -0.00004904, 0.00002077, -0.00003161, 0.00003493, -0.00003379, 0.00003377, -0.00002251, 0.00004120, -0.00004730, 0.00002693, -0.00003919, 0.00003740, -0.00004359, 0.00003913, -0.00003139, 0.00004016, -0.00004172, 0.00003956, -0.00002872, 0.00003571, -0.00004020, 0.00003159, -0.00003256, 0.00002502, -0.00004730, 0.00003401, -0.00000949, 0.00002388, -0.00002560, 0.00002253, -0.00002215, 0.00002190, -0.00004904, 0.00001632, 0.00001804, 0.00000271, -0.00001847, 0.00000627, -0.00003163, 0.00003502, -0.00004773, -0.00001564, 0.00006667, -0.00003275, -0.00005614, },
		{ -0.00005632, -0.00005779, -0.00028791, 0.00011596, 0.00069748, 0.00009681, -0.00027977, -0.00004985, -0.00006079, -0.00002855, 0.00004605, 0.00000657, -0.00004748, 0.00002461, -0.00003664, 0.00002168, -0.00002841, 0.00001457, -0.00000423, 0.00003318, -0.00004879, 0.00002029, -0.00003331, 0.00003717, -0.00003515, 0.00003542, -0.00002473, 0.00004229, -0.00004704, 0.00002703, -0.00004011, 0.00003815, -0.00004392, 0.00003945, -0.00003158, 0.00004000, -0.00004147, 0.00003921, -0.00002805, 0.00003494, -0.00003945, 0.00003050, -0.00003136, 0.00002450, -0.00004704, 0.00003281, -0.00000757, 0.00002239, -0.00002436, 0.00002083, -0.00002083, 0.00002176, -0.00004879, 0.00001472, 0.00001993, 0.00000165, -0.00001762, 0.00000522, -0.00003116, 0.00003507, -0.00004748, -0.00001589, 0.00006648, -0.00003278, },
		{ -0.00003266, -0.00005671, -0.00005610, -0.00028400, 0.00009649, 0.00072853, 0.00007290, -0.00027397, -0.00004632, -0.00006222, -0.00002745, 0.00004031, 0.00001230, -0.00004716, 0.00002187, -0.00003752, 0.00002477, -0.00003022, 0.00001671, -0.00000832, 0.00003611, -0.00004847, 0.00001972, -0.00003500, 0.00003941, -0.00003647, 0.00003704, -0.00002690, 0.00004331, -0.00004672, 0.00002705, -0.00004095, 0.00003882, -0.00004417, 0.00003969, -0.00003169, 0.00003978, -0.00004115, 0.00003880, -0.00002734, 0.00003413, -0.00003866, 0.00002940, -0.00003017, 0.00002394, -0.00004672, 0.00003164, -0.00000576, 0.00002098, -0.00002318, 0.00001928, -0.00001963, 0.00002154, -0.00004847, 0.00001338, 0.00002144, 0.00000079, -0.00001693, 0.00000450, -0.00003077, 0.00003475, -0.00004716, -0.00001537, 0.00006547, },
		{ 0.00006367, -0.00003240, -0.00005732, -0.00005383, -0.00027906, 0.00007252, 0.00076587, 0.00004472, -0.00026723, -0.00004231, -0.00006381, -0.00002626, 0.00003401, 0.00001852, -0.00004677, 0.00001888, -0.00003841, 0.00002800, -0.00003209, 0.00001889, -0.00001253, 0.00003910, -0.00004808, 0.00001908, -0.00003666, 0.00004162, -0.00003775, 0.00003861, -0.00002901, 0.00004427, -0.00004634, 0.00002700, -0.00004170, 0.00003941, -0.00004433, 0.00003985, -0.00003172, 0.00003948, -0.00004076, 0.00003834, -0.00002659, 0.00003329, -0.00003784, 0.00002830, -0.00002900, 0.00002333, -0.00004634, 0.00003050, -0.00000409, 0.00001968, -0.00002209, 0.00001790, -0.00001854, 0.00002123, -0.00004808, 0.00001230, 0.00002254, 0.00000013, -0.00001642, 0.00000411, -0.00003046, 0.00003406, -0.00004677, -0.00001409, },
		{ -0.00001205, 0.00006107, -0.00003199, -0.00005813, -0.00005101, -0.00027313, 0.00004427, 0.00080913, 0.00001253, -0.00025963, -0.00003785, -0.00006554, -0.00002498, 0.00002721, 0.00002518, -0.00004633, 0.00001567, -0.00003932, 0.00003135, -0.00003399, 0.00002111, -0.00001682, 0.00004211, -0.00004764, 0.00001838, -0.00003828, 0.00004378, -0.00003897, 0.00004010, -0.00003103, 0.00004514, -0.00004589, 0.00002688, -0.00004236, 0.00003990, -0.00004441, 0.00003992, -0.00003166, 0.00003912, -0.00004031, 0.00003781, -0.00002581, 0.00003243, -0.00003701, 0.00002721, -0.00002786, 0.00002269, -0.00004589, 0.00002942, -0.00000258, 0.00001849, -0.00002108, 0.00001670, -0.00001759, 0.00002083, -0.00004764, 0.00001151, 0.00002323, -0.00000031, -0.00001610, 0.00000407, -0.00003023, 0.00003300, -0.00004633, },
		{ -0.00004583, -0.00000929, 0.00005772, -0.00003144, -0.00005913, -0.00004766, -0.00026628, 0.00001203, 0.00085790, -0.00002336, -0.00025122, -0.00003299, -0.00006739, -0.00002362, 0.00001997, 0.00003222, -0.00004583, 0.00001227, -0.00004022, 0.00003477, -0.00003590, 0.00002333, -0.00002115, 0.00004512, -0.00004713, 0.00001763, -0.00003985, 0.00004588, -0.00004011, 0.00004151, -0.00003295, 0.00004592, -0.00004539, 0.00002670, -0.00004292, 0.00004029, -0.00004441, 0.00003990, -0.00003152, 0.00003869, -0.00003981, 0.00003724, -0.00002500, 0.00003156, -0.00003615, 0.00002614, -0.00002675, 0.00002203, -0.00004539, 0.00002841, -0.00000124, 0.00001742, -0.00002017, 0.00001568, -0.00001679, 0.00002036, -0.00004713, 0.00001101, 0.00002349, -0.00000054, -0.00001596, 0.00000437, -0.00003009, 0.00003159, },
		{ 0.00002983, -0.00004527, -0.00000581, 0.00005363, -0.00003076, -0.00006033, -0.00004381, -0.00025857, -0.00002391, 0.00091171, -0.00006259, -0.00024211, -0.00002778, -0.00006936, -0.00002221, 0.00001237, 0.00003956, -0.00004527, 0.00000871, -0.00004111, 0.00003825, -0.00003781, 0.00002554, -0.00002547, 0.00004809, -0.00004658, 0.00001682, -0.00004134, 0.00004788, -0.00004117, 0.00004282, -0.00003474, 0.00004661, -0.00004484, 0.00002645, -0.00004337, 0.00004057, -0.00004433, 0.00003980, -0.00003131, 0.00003821, -0.00003926, 0.00003662, -0.00002418, 0.00003068, -0.00003530, 0.00002511, -0.00002570, 0.00002134, -0.00004484, 0.00002746, -0.00000008, 0.00001649, -0.00001937, 0.00001487, -0.00001613, 0.00001980, -0.00004658, 0.00001080, 0.00002333, -0.00000056, -0.00001601, 0.00000500, -0.00003004, },
		{ -0.00003007, 0.00002775, -0.00004467, -0.00000166, 0.00004886, -0.00002994, -0.00006170, -0.00003951, -0.00025007, -0.00006319, 0.00097005, -0.00010480, -0.00023236, -0.00002227, -0.00007141, -0.00002074, 0.00000448, 0.00004714, -0.00004467, 0.00000504, -0.00004198, 0.00004173, -0.00003969, 0.00002772, -0.00002975, 0.00005099, -0.00004598, 0.00001596, -0.00004275, 0.00004978, -0.00004213, 0.00004402, -0.00003639, 0.00004720, -0.00004423, 0.00002613, -0.00004371, 0.00004075, -0.00004416, 0.00003961, -0.00003101, 0.00003767, -0.00003866, 0.00003596, -0.00002334, 0.00002980, -0.00003445, 0.00002412, -0.00002470, 0.00002064, -0.00004423, 0.00002660, 0.00000089, 0.00001569, -0.00001869, 0.00001427, -0.00001563, 0.00001918, -0.00004598, 0.00001089, 0.00002274, -0.00000035, -0.00001625, 0.00000597, },
		{ 0.00000726, -0.00003020, 0.00002537, -0.00004403, 0.00000312, 0.00004344, -0.00002900, -0.00006323, -0.00003479, -0.00024086, -0.00010544, 0.00103234, -0.00014957, -0.00022209, -0.00001650, -0.00007353, -0.00001925, -0.00000363, 0.00005489, -0.00004403, 0.00000127, -0.00004283, 0.00004519, -0.00004154, 0.00002985, -0.00003394, 0.00005381, -0.00004534, 0.00001508, -0.00004406, 0.00005154, -0.00004300, 0.00004510, -0.00003789, 0.00004767, -0.00004359, 0.00002576, -0.00004393, 0.00004083, -0.00004391, 0.00003934, -0.00003064, 0.00003709, -0.00003801, 0.00003527, -0.00002250, 0.00002894, -0.00003361, 0.00002319, -0.00002377, 0.00001993, -0.00004359, 0.00002582, 0.00000166, 0.00001505, -0.00001813, 0.00001389, -0.00001529, 0.00001848, -0.00004534, 0.00001127, 0.00002174, 0.00000007, -0.00001667, },
		{ -0.00001728, 0.00000887, -0.00003041, 0.00002270, -0.00004335, 0.00000849, 0.00003744, -0.00002795, -0.00006492, -0.00002970, -0.00023104, -0.00015024, 0.00109799, -0.00019647, -0.00021137, -0.00001054, -0.00007571, -0.00001773, -0.00001188, 0.00006272, -0.00004335, -0.00000254, -0.00004364, 0.00004860, -0.00004333, 0.00003190, -0.00003800, 0.00005650, -0.00004466, 0.00001416, -0.00004525, 0.00005317, -0.00004375, 0.00004604, -0.00003922, 0.00004803, -0.00004292, 0.00002532, -0.00004404, 0.00004079, -0.00004358, 0.00003899, -0.00003020, 0.00003646, -0.00003734, 0.00003456, -0.00002167, 0.00002809, -0.00003279, 0.00002231, -0.00002292, 0.00001922, -0.00004292, 0.00002515, 0.00000221, 0.00001457, -0.00001770, 0.00001372, -0.00001512, 0.00001773, -0.00004466, 0.00001194, 0.00002033, 0.00000071, },
		{ 0.00000155, -0.00001806, 0.00001077, -0.00003070, 0.00001978, -0.00004265, 0.00001439, 0.00003089, -0.00002680, -0.00006673, -0.00002429, -0.00022070, -0.00019718, 0.00116637, -0.00024506, -0.00020033, -0.00000445, -0.00007791, -0.00001622, -0.00002019, 0.00007056, -0.00004265, -0.00000637, -0.00004441, 0.00005192, -0.00004504, 0.00003386, -0.00004189, 0.00005905, -0.00004395, 0.00001323, -0.00004632, 0.00005463, -0.00004438, 0.00004684, -0.00004036, 0.00004828, -0.00004221, 0.00002484, -0.00004404, 0.00004065, -0.00004318, 0.00003856, -0.00002969, 0.00003579, -0.00003663, 0.00003382, -0.00002086, 0.00002727, -0.00003200, 0.00002151, -0.00002216, 0.00001852, -0.00004221, 0.00002457, 0.00000255, 0.00001424, -0.00001740, 0.00001377, -0.00001511, 0.00001693, -0.00004395, 0.00001290, 0.00001853, },
		{ 0.00001634, 0.00000259, -0.00001900, 0.00001294, -0.00003108, 0.00001663, -0.00004192, 0.00002077, 0.00002388, -0.00002556, -0.00006866, -0.00001861, -0.00020994, -0.00024579, 0.00123682, -0.00029486, -0.00018906, 0.00000173, -0.00008011, -0.00001471, -0.00002847, 0.00007835, -0.00004192, -0.00001017, -0.00004513, 0.00005512, -0.00004665, 0.00003570, -0.00004558, 0.00006143, -0.00004323, 0.00001229, -0.00004726, 0.00005592, -0.00004488, 0.00004749, -0.00004131, 0.00004841, -0.00004148, 0.00002430, -0.00004391, 0.00004040, -0.00004270, 0.00003806, -0.00002911, 0.00003509, -0.00003591, 0.00003307, -0.00002007, 0.00002649, -0.00003124, 0.00002078, -0.00002149, 0.00001782, -0.00004148, 0.00002411, 0.00000267, 0.00001408, -0.00001723, 0.00001404, -0.00001527, 0.00001608, -0.00004323, 0.00001413, },
		{ 0.00001562, 0.00001380, 0.00000381, -0.00002011, 0.00001537, -0.00003154, 0.00001328, -0.00004118, 0.00002757, 0.00001646, -0.00002423, -0.00007069, -0.00001272, -0.00019886, -0.00029560, 0.00130866, -0.00034540, -0.00017767, 0.00000792, -0.00008231, -0.00001323, -0.00003666, 0.00008599, -0.00004118, -0.00001391, -0.00004580, 0.00005817, -0.00004816, 0.00003741, -0.00004902, 0.00006362, -0.00004249, 0.00001136, -0.00004805, 0.00005703, -0.00004526, 0.00004799, -0.00004206, 0.00004841, -0.00004074, 0.00002372, -0.00004367, 0.00004005, -0.00004216, 0.00003749, -0.00002848, 0.00003437, -0.00003516, 0.00003232, -0.00001930, 0.00002575, -0.00003053, 0.00002014, -0.00002092, 0.00001715, -0.00004074, 0.00002376, 0.00000257, 0.00001409, -0.00001720, 0.00001453, -0.00001559, 0.00001520, -0.00004249, },
		{ -0.00004174, 0.00001737, 0.00001093, 0.00000521, -0.00002136, 0.00001804, -0.00003207, 0.00000977, -0.00004043, 0.00003472, 0.00000872, -0.00002284, -0.00007279, -0.00000666, -0.00018757, -0.00034615, 0.00138120, -0.00039619, -0.00016627, 0.00001407, -0.00008447, -0.00001178, -0.00004466, 0.00009342, -0.00004043, -0.00001755, -0.00004640, 0.00006105, -0.00004955, 0.00003897, -0.00005220, 0.00006560, -0.00004174, 0.00001043, -0.00004869, 0.00005793, -0.00004550, 0.00004832, -0.00004259, 0.00004830, -0.00004000, 0.00002311, -0.00004332, 0.00003959, -0.00004155, 0.00003685, -0.00002780, 0.00003364, -0.00003442, 0.00003158, -0.00001858, 0.00002506, -0.00002987, 0.00001959, -0.00002046, 0.00001650, -0.00004000, 0.00002353, 0.00000225, 0.00001426, -0.00001731, 0.00001523, -0.00001608, 0.00001429, },
		{ 0.00001336, -0.00004099, 0.00001934, 0.00000776, 0.00000678, -0.00002275, 0.00002091, -0.00003267, 0.00000612, -0.00003968, 0.00004216, 0.00000071, -0.00002140, -0.00007494, -0.00000051, -0.00017617, -0.00039693, 0.00145373, -0.00044673, -0.00015498, 0.00002013, -0.00008657, -0.00001039, -0.00005241, 0.00010058, -0.00003968, -0.00002107, -0.00004693, 0.00006371, -0.00005080, 0.00004038, -0.00005507, 0.00006735, -0.00004099, 0.00000952, -0.00004918, 0.00005863, -0.00004561, 0.00004849, -0.00004291, 0.00004807, -0.00003925, 0.00002246, -0.00004286, 0.00003904, -0.00004089, 0.00003616, -0.00002708, 0.00003289, -0.00003367, 0.00003084, -0.00001790, 0.00002442, -0.00002926, 0.00001914, -0.00002011, 0.00001589, -0.00003925, 0.00002341, 0.00000172, 0.00001459, -0.00001755, 0.00001614, -0.00001672, },
		{ -0.00001752, 0.00001242, -0.00004025, 0.00002153, 0.00000431, 0.00000849, -0.00002426, 0.00002396, -0.00003333, 0.00000238, -0.00003894, 0.00004980, -0.00000747, -0.00001992, -0.00007714, 0.00000569, -0.00016479, -0.00044747, 0.00152557, -0.00049655, -0.00014390, 0.00002602, -0.00008859, -0.00000907, -0.00005982, 0.00010737, -0.00003894, -0.00002441, -0.00004739, 0.00006614, -0.00005191, 0.00004160, -0.00005761, 0.00006884, -0.00004025, 0.00000863, -0.00004950, 0.00005912, -0.00004558, 0.00004850, -0.00004301, 0.00004772, -0.00003851, 0.00002178, -0.00004229, 0.00003840, -0.00004018, 0.00003542, -0.00002631, 0.00003214, -0.00003293, 0.00003012, -0.00001727, 0.00002385, -0.00002872, 0.00001878, -0.00001987, 0.00001531, -0.00003851, 0.00002342, 0.00000097, 0.00001509, -0.00001793, 0.00001724, },
		{ 0.00001853, -0.00001845, 0.00001148, -0.00003952, 0.00002391, 0.00000062, 0.00001033, -0.00002587, 0.00002717, -0.00003406, -0.00000142, -0.00003822, 0.00005759, -0.00001576, -0.00001841, -0.00007935, 0.00001186, -0.00015351, -0.00049727, 0.00159602, -0.00054516, -0.00013314, 0.00003170, -0.00009052, -0.00000783, -0.00006684, 0.00011376, -0.00003822, -0.00002757, -0.00004776, 0.00006832, -0.00005286, 0.00004264, -0.00005979, 0.00007007, -0.00003952, 0.00000779, -0.00004966, 0.00005940, -0.00004542, 0.00004834, -0.00004289, 0.00004725, -0.00003778, 0.00002109, -0.00004162, 0.00003767, -0.00003942, 0.00003464, -0.00002552, 0.00003139, -0.00003220, 0.00002942, -0.00001669, 0.00002335, -0.00002824, 0.00001853, -0.00001975, 0.00001477, -0.00003778, 0.00002355, 0.00000002, 0.00001574, -0.00001844, },
		{ -0.00001907, 0.00002000, -0.00001952, 0.00001055, -0.00003882, 0.00002647, -0.00000327, 0.00001229, -0.00002759, 0.00003049, -0.00003483, -0.00000524, -0.00003751, 0.00006543, -0.00002406, -0.00001689, -0.00008155, 0.00001796, -0.00014247, -0.00054586, 0.00166440, -0.00059209, -0.00012280, 0.00003711, -0.00009234, -0.00000667, -0.00007338, 0.00011966, -0.00003751, -0.00003049, -0.00004806, 0.00007022, -0.00005363, 0.00004348, -0.00006160, 0.00007103, -0.00003882, 0.00000699, -0.00004965, 0.00005945, -0.00004512, 0.00004801, -0.00004255, 0.00004668, -0.00003708, 0.00002039, -0.00004086, 0.00003687, -0.00003863, 0.00003382, -0.00002470, 0.00003066, -0.00003150, 0.00002875, -0.00001618, 0.00002292, -0.00002784, 0.00001839, -0.00001974, 0.00001429, -0.00003708, 0.00002379, -0.00000112, 0.00001654, },
		{ 0.00001748, -0.00001982, 0.00002162, -0.00002072, 0.00000964, -0.00003814, 0.00002916, -0.00000733, 0.00001434, -0.00002937, 0.00003389, -0.00003564, -0.00000906, -0.00003683, 0.00007326, -0.00003231, -0.00001538, -0.00008372, 0.00002392, -0.00013176, -0.00059277, 0.00173005, -0.00063690, -0.00011298, 0.00004220, -0.00009402, -0.00000562, -0.00007939, 0.00012503, -0.00003683, -0.00003315, -0.00004827, 0.00007182, -0.00005424, 0.00004411, -0.00006301, 0.00007170, -0.00003814, 0.00000623, -0.00004948, 0.00005928, -0.00004468, 0.00004753, -0.00004200, 0.00004600, -0.00003640, 0.00001968, -0.00004001, 0.00003599, -0.00003781, 0.00003297, -0.00002387, 0.00002994, -0.00003082, 0.00002813, -0.00001574, 0.00002257, -0.00002751, 0.00001836, -0.00001985, 0.00001385, -0.00003640, 0.00002415, -0.00000245, },
		{ -0.00000395, 0.00001856, -0.00002068, 0.00002339, -0.00002202, 0.00000875, -0.00003750, 0.00003198, -0.00001152, 0.00001646, -0.00003122, 0.00003736, -0.00003648, -0.00001282, -0.00003619, 0.00008100, -0.00004042, -0.00001388, -0.00008584, 0.00002968, -0.00012148, -0.00063754, 0.00179235, -0.00067914, -0.00010377, 0.00004692, -0.00009555, -0.00000469, -0.00008480, 0.00012981, -0.00003619, -0.00003554, -0.00004840, 0.00007311, -0.00005466, 0.00004454, -0.00006401, 0.00007208, -0.00003750, 0.00000554, -0.00004914, 0.00005889, -0.00004412, 0.00004689, -0.00004123, 0.00004523, -0.00003576, 0.00001897, -0.00003908, 0.00003506, -0.00003697, 0.00003211, -0.00002304, 0.00002925, -0.00003018, 0.00002754, -0.00001537, 0.00002230, -0.00002726, 0.00001843, -0.00002008, 0.00001348, -0.00003576, 0.00002463, },
		{ 0.00002521, -0.00000560, 0.00001976, -0.00002164, 0.00002528, -0.00002343, 0.00000790, -0.00003690, 0.00003488, -0.00001579, 0.00001864, -0.00003310, 0.00004084, -0.00003736, -0.00001650, -0.00003559, 0.00008858, -0.00004832, -0.00001242, -0.00008790, 0.00003519, -0.00011173, -0.00067975, 0.00185068, -0.00071843, -0.00009527, 0.00005122, -0.00009692, -0.00000387, -0.00008958, 0.00013396, -0.00003559, -0.00003762, -0.00004843, 0.00007408, -0.00005490, 0.00004474, -0.00006459, 0.00007217, -0.00003690, 0.00000492, -0.00004864, 0.00005829, -0.00004344, 0.00004609, -0.00004027, 0.00004437, -0.00003516, 0.00001827, -0.00003809, 0.00003407, -0.00003612, 0.00003123, -0.00002220, 0.00002859, -0.00002958, 0.00002700, -0.00001507, 0.00002211, -0.00002709, 0.00001861, -0.00002042, 0.00001316, -0.00003516, },
		{ -0.00003460, 0.00002590, -0.00000740, 0.00002107, -0.00002270, 0.00002729, -0.00002492, 0.00000709, -0.00003634, 0.00003785, -0.00002011, 0.00002086, -0.00003501, 0.00004431, -0.00003825, -0.00002005, -0.00003504, 0.00009593, -0.00005592, -0.00001101, -0.00008986, 0.00004041, -0.00010262, -0.00071898, 0.00190449, -0.00075436, -0.00008756, 0.00005507, -0.00009812, -0.00000318, -0.00009366, 0.00013743, -0.00003504, -0.00003937, -0.00004838, 0.00007472, -0.00005495, 0.00004473, -0.00006476, 0.00007196, -0.00003634, 0.00000436, -0.00004799, 0.00005748, -0.00004264, 0.00004516, -0.00003910, 0.00004342, -0.00003460, 0.00001758, -0.00003703, 0.00003304, -0.00003527, 0.00003035, -0.00002138, 0.00002797, -0.00002902, 0.00002652, -0.00001485, 0.00002201, -0.00002701, 0.00001890, -0.00002087, 0.00001291, },
		{ 0.00001273, -0.00003410, 0.00002669, -0.00000931, 0.00002248, -0.00002385, 0.00002938, -0.00002649, 0.00000633, -0.00003584, 0.00004086, -0.00002444, 0.00002308, -0.00003692, 0.00004774, -0.00003915, -0.00002345, -0.00003453, 0.00010297, -0.00006315, -0.00000965, -0.00009172, 0.00004526, -0.00009421, -0.00075486, 0.00195326, -0.00078661, -0.00008071, 0.00005842, -0.00009913, -0.00000263, -0.00009702, 0.00014020, -0.00003453, -0.00004079, -0.00004824, 0.00007502, -0.00005481, 0.00004450, -0.00006449, 0.00007146, -0.00003584, 0.00000388, -0.00004718, 0.00005647, -0.00004173, 0.00004409, -0.00003776, 0.00004240, -0.00003410, 0.00001691, -0.00003593, 0.00003197, -0.00003441, 0.00002948, -0.00002057, 0.00002740, -0.00002852, 0.00002609, -0.00001472, 0.00002199, -0.00002701, 0.00001929, -0.00002142, },
		{ -0.00002208, 0.00001261, -0.00003365, 0.00002756, -0.00001133, 0.00002397, -0.00002506, 0.00003154, -0.00002811, 0.00000563, -0.00003540, 0.00004387, -0.00002873, 0.00002530, -0.00003882, 0.00005108, -0.00004005, -0.00002666, -0.00003409, 0.00010963, -0.00006995, -0.00000837, -0.00009345, 0.00004972, -0.00008661, -0.00078705, 0.00199653, -0.00081485, -0.00007478, 0.00006124, -0.00009994, -0.00000223, -0.00009961, 0.00014224, -0.00003409, -0.00004184, -0.00004801, 0.00007497, -0.00005449, 0.00004405, -0.00006381, 0.00007066, -0.00003540, 0.00000349, -0.00004623, 0.00005527, -0.00004073, 0.00004290, -0.00003625, 0.00004132, -0.00003365, 0.00001628, -0.00003478, 0.00003089, -0.00003358, 0.00002862, -0.00001979, 0.00002688, -0.00002807, 0.00002573, -0.00001466, 0.00002206, -0.00002709, 0.00001977, },
		{ 0.00002036, -0.00002283, 0.00001256, -0.00003327, 0.00002851, -0.00001344, 0.00002554, -0.00002634, 0.00003375, -0.00002977, 0.00000500, -0.00003501, 0.00004685, -0.00003295, 0.00002748, -0.00004069, 0.00005432, -0.00004095, -0.00002965, -0.00003370, 0.00011585, -0.00007626, -0.00000718, -0.00009503, 0.00005374, -0.00007988, -0.00081524, 0.00203386, -0.00083883, -0.00006984, 0.00006351, -0.00010054, -0.00000196, -0.00010142, 0.00014352, -0.00003370, -0.00004254, -0.00004770, 0.00007459, -0.00005398, 0.00004340, -0.00006271, 0.00006959, -0.00003501, 0.00000318, -0.00004515, 0.00005388, -0.00003963, 0.00004159, -0.00003458, 0.00004019, -0.00003327, 0.00001567, -0.00003361, 0.00002979, -0.00003276, 0.00002778, -0.00001904, 0.00002641, -0.00002769, 0.00002544, -0.00001469, 0.00002222, -0.00002725, },
		{ -0.00002750, 0.00002103, -0.00002367, 0.00001258, -0.00003295, 0.00002954, -0.00001561, 0.00002715, -0.00002767, 0.00003599, -0.00003146, 0.00000443, -0.00003469, 0.00004979, -0.00003704, 0.00002961, -0.00004250, 0.00005741, -0.00004182, -0.00003240, -0.00003338, 0.00012158, -0.00008199, -0.00000608, -0.00009646, 0.00005727, -0.00007407, -0.00083915, 0.00206491, -0.00085830, -0.00006593, 0.00006520, -0.00010093, -0.00000185, -0.00010242, 0.00014404, -0.00003338, -0.00004286, -0.00004731, 0.00007387, -0.00005329, 0.00004253, -0.00006120, 0.00006825, -0.00003469, 0.00000295, -0.00004394, 0.00005233, -0.00003846, 0.00004019, -0.00003277, 0.00003902, -0.00003295, 0.00001511, -0.00003242, 0.00002868, -0.00003197, 0.00002697, -0.00001833, 0.00002600, -0.00002737, 0.00002521, -0.00001480, 0.00002246, },
		{ 0.00002278, -0.00002783, 0.00002178, -0.00002459, 0.00001268, -0.00003270, 0.00003062, -0.00001783, 0.00002881, -0.00002903, 0.00003824, -0.00003316, 0.00000394, -0.00003444, 0.00005265, -0.00004098, 0.00003167, -0.00004425, 0.00006033, -0.00004268, -0.00003487, -0.00003313, 0.00012675, -0.00008712, -0.00000509, -0.00009772, 0.00006028, -0.00006926, -0.00085855, 0.00208938, -0.00087307, -0.00006308, 0.00006630, -0.00010111, -0.00000188, -0.00010261, 0.00014379, -0.00003313, -0.00004280, -0.00004684, 0.00007281, -0.00005244, 0.00004147, -0.00005930, 0.00006665, -0.00003444, 0.00000282, -0.00004263, 0.00005063, -0.00003721, 0.00003870, -0.00003085, 0.00003781, -0.00003270, 0.00001459, -0.00003123, 0.00002759, -0.00003122, 0.00002620, -0.00001766, 0.00002565, -0.00002712, 0.00002505, -0.00001498, },
		{ -0.00001525, 0.00002318, -0.00002823, 0.00002261, -0.00002558, 0.00001284, -0.00003251, 0.00003176, -0.00002007, 0.00003050, -0.00003041, 0.00004046, -0.00003485, 0.00000354, -0.00003426, 0.00005539, -0.00004472, 0.00003364, -0.00004592, 0.00006305, -0.00004350, -0.00003704, -0.00003295, 0.00013132, -0.00009157, -0.00000422, -0.00009880, 0.00006276, -0.00006549, -0.00087325, 0.00210703, -0.00088302, -0.00006134, 0.00006680, -0.00010107, -0.00000206, -0.00010199, 0.00014277, -0.00003295, -0.00004237, -0.00004629, 0.00007144, -0.00005142, 0.00004023, -0.00005703, 0.00006480, -0.00003426, 0.00000278, -0.00004121, 0.00004880, -0.00003591, 0.00003714, -0.00002882, 0.00003659, -0.00003251, 0.00001412, -0.00003004, 0.00002652, -0.00003051, 0.00002548, -0.00001705, 0.00002537, -0.00002693, 0.00002497, },
		{ 0.00002496, -0.00001560, 0.00002366, -0.00002870, 0.00002351, -0.00002663, 0.00001306, -0.00003240, 0.00003294, -0.00002231, 0.00003219, -0.00003181, 0.00004266, -0.00003651, 0.00000321, -0.00003415, 0.00005801, -0.00004822, 0.00003550, -0.00004748, 0.00006554, -0.00004428, -0.00003889, -0.00003284, 0.00013524, -0.00009532, -0.00000347, -0.00009968, 0.00006466, -0.00006279, -0.00088313, 0.00211768, -0.00088803, -0.00006070, 0.00006669, -0.00010082, -0.00000239, -0.00010055, 0.00014099, -0.00003284, -0.00004157, -0.00004568, 0.00006976, -0.00005024, 0.00003881, -0.00005441, 0.00006274, -0.00003415, 0.00000284, -0.00003971, 0.00004684, -0.00003457, 0.00003553, -0.00002670, 0.00003537, -0.00003240, 0.00001371, -0.00002887, 0.00002547, -0.00002985, 0.00002481, -0.00001650, 0.00002516, -0.00002682, },
		{ -0.00002679, 0.00002503, -0.00001601, 0.00002420, -0.00002924, 0.00002447, -0.00002773, 0.00001336, -0.00003237, 0.00003414, -0.00002453, 0.00003387, -0.00003320, 0.00004479, -0.00003814, 0.00000298, -0.00003411, 0.00006046, -0.00005147, 0.00003723, -0.00004892, 0.00006779, -0.00004501, -0.00004041, -0.00003280, 0.00013847, -0.00009832, -0.00000286, -0.00010035, 0.00006598, -0.00006119, -0.00088807, 0.00212125, -0.00088807, -0.00006119, 0.00006598, -0.00010035, -0.00000286, -0.00009832, 0.00013847, -0.00003280, -0.00004041, -0.00004501, 0.00006779, -0.00004892, 0.00003723, -0.00005147, 0.00006046, -0.00003411, 0.00000298, -0.00003814, 0.00004479, -0.00003320, 0.00003387, -0.00002453, 0.00003414, -0.00003237, 0.00001336, -0.00002773, 0.00002447, -0.00002924, 0.00002420, -0.00001601, 0.00002503, },
		{ 0.00002496, -0.00002682, 0.00002516, -0.00001650, 0.00002481, -0.00002985, 0.00002547, -0.00002887, 0.00001371, -0.00003240, 0.00003537, -0.00002670, 0.00003553, -0.00003457, 0.00004684, -0.00003971, 0.00000284, -0.00003415, 0.00006274, -0.00005441, 0.00003881, -0.00005024, 0.00006976, -0.00004568, -0.00004157, -0.00003284, 0.00014099, -0.00010055, -0.00000239, -0.00010082, 0.00006669, -0.00006070, -0.00088803, 0.00211768, -0.00088313, -0.00006279, 0.00006466, -0.00009968, -0.00000347, -0.00009532, 0.00013524, -0.00003284, -0.00003889, -0.00004428, 0.00006554, -0.00004748, 0.00003550, -0.00004822, 0.00005801, -0.00003415, 0.00000321, -0.00003651, 0.00004266, -0.00003181, 0.00003219, -0.00002231, 0.00003294, -0.00003240, 0.00001306, -0.00002663, 0.00002351, -0.00002870, 0.00002366, -0.00001560, },
		{ -0.00001525, 0.00002497, -0.00002693, 0.00002537, -0.00001705, 0.00002548, -0.00003051, 0.00002652, -0.00003004, 0.00001412, -0.00003251, 0.00003659, -0.00002882, 0.00003714, -0.00003591, 0.00004880, -0.00004121, 0.00000278, -0.00003426, 0.00006480, -0.00005703, 0.00004023, -0.00005142, 0.00007144, -0.00004629, -0.00004237, -0.00003295, 0.00014277, -0.00010199, -0.00000206, -0.00010107, 0.00006680, -0.00006134, -0.00088302, 0.00210703, -0.00087325, -0.00006549, 0.00006276, -0.00009880, -0.00000422, -0.00009157, 0.00013132, -0.00003295, -0.00003704, -0.00004350, 0.00006305, -0.00004592, 0.00003364, -0.00004472, 0.00005539, -0.00003426, 0.00000354, -0.00003485, 0.00004046, -0.00003041, 0.00003050, -0.00002007, 0.00003176, -0.00003251, 0.00001284, -0.00002558, 0.00002261, -0.00002823, 0.00002318, },
		{ 0.00002278, -0.00001498, 0.00002505, -0.00002712, 0.00002565, -0.00001766, 0.00002620, -0.00003122, 0.00002759, -0.00003123, 0.00001459, -0.00003270, 0.00003781, -0.00003085, 0.00003870, -0.00003721, 0.00005063, -0.00004263, 0.00000282, -0.00003444, 0.00006665, -0.00005930, 0.00004147, -0.00005244, 0.00007281, -0.00004684, -0.00004280, -0.00003313, 0.00014379, -0.00010261, -0.00000188, -0.00010111, 0.00006630, -0.00006308, -0.00087307, 0.00208938, -0.00085855, -0.00006926, 0.00006028, -0.00009772, -0.00000509, -0.00008712, 0.00012675, -0.00003313, -0.00003487, -0.00004268, 0.00006033, -0.00004425, 0.00003167, -0.00004098, 0.00005265, -0.00003444, 0.00000394, -0.00003316, 0.00003824, -0.00002903, 0.00002881, -0.00001783, 0.00003062, -0.00003270, 0.00001268, -0.00002459, 0.00002178, -0.00002783, },
		{ -0.00002750, 0.00002246, -0.00001480, 0.00002521, -0.00002737, 0.00002600, -0.00001833, 0.00002697, -0.00003197, 0.00002868, -0.00003242, 0.00001511, -0.00003295, 0.00003902, -0.00003277, 0.00004019, -0.00003846, 0.00005233, -0.00004394, 0.00000295, -0.00003469, 0.00006825, -0.00006120, 0.00004253, -0.00005329, 0.00007387, -0.00004731, -0.00004286, -0.00003338, 0.00014404, -0.00010242, -0.00000185, -0.00010093, 0.00006520, -0.00006593, -0.00085830, 0.00206491, -0.00083915, -0.00007407, 0.00005727, -0.00009646, -0.00000608, -0.00008199, 0.00012158, -0.00003338, -0.00003240, -0.00004182, 0.00005741, -0.00004250, 0.00002961, -0.00003704, 0.00004979, -0.00003469, 0.00000443, -0.00003146, 0.00003599, -0.00002767, 0.00002715, -0.00001561, 0.00002954, -0.00003295, 0.00001258, -0.00002367, 0.00002103, },
		{ 0.00002036, -0.00002725, 0.00002222, -0.00001469, 0.00002544, -0.00002769, 0.00002641, -0.00001904, 0.00002778, -0.00003276, 0.00002979, -0.00003361, 0.00001567, -0.00003327, 0.00004019, -0.00003458, 0.00004159, -0.00003963, 0.00005388, -0.00004515, 0.00000318, -0.00003501, 0.00006959, -0.00006271, 0.00004340, -0.00005398, 0.00007459, -0.00004770, -0.00004254, -0.00003370, 0.00014352, -0.00010142, -0.00000196, -0.00010054, 0.00006351, -0.00006984, -0.00083883, 0.00203386, -0.00081524, -0.00007988, 0.00005374, -0.00009503, -0.00000718, -0.00007626, 0.00011585, -0.00003370, -0.00002965, -0.00004095, 0.00005432, -0.00004069, 0.00002748, -0.00003295, 0.00004685, -0.00003501, 0.00000500, -0.00002977, 0.00003375, -0.00002634, 0.00002554, -0.00001344, 0.00002851, -0.00003327, 0.00001256, -0.00002283, },
		{ -0.00002208, 0.00001977, -0.00002709, 0.00002206, -0.00001466, 0.00002573, -0.00002807, 0.00002688, -0.00001979, 0.00002862, -0.00003358, 0.00003089, -0.00003478, 0.00001628, -0.00003365, 0.00004132, -0.00003625, 0.00004290, -0.00004073, 0.00005527, -0.00004623, 0.00000349, -0.00003540, 0.00007066, -0.00006381, 0.00004405, -0.00005449, 0.00007497, -0.00004801, -0.00004184, -0.00003409, 0.00014224, -0.00009961, -0.00000223, -0.00009994, 0.00006124, -0.00007478, -0.00081485, 0.00199653, -0.00078705, -0.00008661, 0.00004972, -0.00009345, -0.00000837, -0.00006995, 0.00010963, -0.00003409, -0.00002666, -0.00004005, 0.00005108, -0.00003882, 0.00002530, -0.00002873, 0.00004387, -0.00003540, 0.00000563, -0.00002811, 0.00003154, -0.00002506, 0.00002397, -0.00001133, 0.00002756, -0.00003365, 0.00001261, },
		{ 0.00001273, -0.00002142, 0.00001929, -0.00002701, 0.00002199, -0.00001472, 0.00002609, -0.00002852, 0.00002740, -0.00002057, 0.00002948, -0.00003441, 0.00003197, -0.00003593, 0.00001691, -0.00003410, 0.00004240, -0.00003776, 0.00004409, -0.00004173, 0.00005647, -0.00004718, 0.00000388, -0.00003584, 0.00007146, -0.00006449, 0.00004450, -0.00005481, 0.00007502, -0.00004824, -0.00004079, -0.00003453, 0.00014020, -0.00009702, -0.00000263, -0.00009913, 0.00005842, -0.00008071, -0.00078661, 0.00195326, -0.00075486, -0.00009421, 0.00004526, -0.00009172, -0.00000965, -0.00006315, 0.00010297, -0.00003453, -0.00002345, -0.00003915, 0.00004774, -0.00003692, 0.00002308, -0.00002444, 0.00004086, -0.00003584, 0.00000633, -0.00002649, 0.00002938, -0.00002385, 0.00002248, -0.00000931, 0.00002669, -0.00003410, },
		{ -0.00003460, 0.00001291, -0.00002087, 0.00001890, -0.00002701, 0.00002201, -0.00001485, 0.00002652, -0.00002902, 0.00002797, -0.00002138, 0.00003035, -0.00003527, 0.00003304, -0.00003703, 0.00001758, -0.00003460, 0.00004342, -0.00003910, 0.00004516, -0.00004264, 0.00005748, -0.00004799, 0.00000436, -0.00003634, 0.00007196, -0.00006476, 0.00004473, -0.00005495, 0.00007472, -0.00004838, -0.00003937, -0.00003504, 0.00013743, -0.00009366, -0.00000318, -0.00009812, 0.00005507, -0.00008756, -0.00075436, 0.00190449, -0.00071898, -0.00010262, 0.00004041, -0.00008986, -0.00001101, -0.00005592, 0.00009593, -0.00003504, -0.00002005, -0.00003825, 0.00004431, -0.00003501, 0.00002086, -0.00002011, 0.00003785, -0.00003634, 0.00000709, -0.00002492, 0.00002729, -0.00002270, 0.00002107, -0.00000740, 0.00002590, },
		{ 0.00002521, -0.00003516, 0.00001316, -0.00002042, 0.00001861, -0.00002709, 0.00002211, -0.00001507, 0.00002700, -0.00002958, 0.00002859, -0.00002220, 0.00003123, -0.00003612, 0.00003407, -0.00003809, 0.00001827, -0.00003516, 0.00004437, -0.00004027, 0.00004609, -0.00004344, 0.00005829, -0.00004864, 0.00000492, -0.00003690, 0.00007217, -0.00006459, 0.00004474, -0.00005490, 0.00007408, -0.00004843, -0.00003762, -0.00003559, 0.00013396, -0.00008958, -0.00000387, -0.00009692, 0.00005122, -0.00009527, -0.00071843, 0.00185068, -0.00067975, -0.00011173, 0.00003519, -0.00008790, -0.00001242, -0.00004832, 0.00008858, -0.00003559, -0.00001650, -0.00003736, 0.00004084, -0.00003310, 0.00001864, -0.00001579, 0.00003488, -0.00003690, 0.00000790, -0.00002343, 0.00002528, -0.00002164, 0.00001976, -0.00000560, },
		{ -0.00000395, 0.00002463, -0.00003576, 0.00001348, -0.00002008, 0.00001843, -0.00002726, 0.00002230, -0.00001537, 0.00002754, -0.00003018, 0.00002925, -0.00002304, 0.00003211, -0.00003697, 0.00003506, -0.00003908, 0.00001897, -0.00003576, 0.00004523, -0.00004123, 0.00004689, -0.00004412, 0.00005889, -0.00004914, 0.00000554, -0.00003750, 0.00007208, -0.00006401, 0.00004454, -0.00005466, 0.00007311, -0.00004840, -0.00003554, -0.00003619, 0.00012981, -0.00008480, -0.00000469, -0.00009555, 0.00004692, -0.00010377, -0.00067914, 0.00179235, -0.00063754, -0.00012148, 0.00002968, -0.00008584, -0.00001388, -0.00004042, 0.00008100, -0.00003619, -0.00001282, -0.00003648, 0.00003736, -0.00003122, 0.00001646, -0.00001152, 0.00003198, -0.00003750, 0.00000875, -0.00002202, 0.00002339, -0.00002068, 0.00001856, },
		{ 0.00001748, -0.00000245, 0.00002415, -0.00003640, 0.00001385, -0.00001985, 0.00001836, -0.00002751, 0.00002257, -0.00001574, 0.00002813, -0.00003082, 0.00002994, -0.00002387, 0.00003297, -0.00003781, 0.00003599, -0.00004001, 0.00001968, -0.00003640, 0.00004600, -0.00004200, 0.00004753, -0.00004468, 0.00005928, -0.00004948, 0.00000623, -0.00003814, 0.00007170, -0.00006301, 0.00004411, -0.00005424, 0.00007182, -0.00004827, -0.00003315, -0.00003683, 0.00012503, -0.00007939, -0.00000562, -0.00009402, 0.00004220, -0.00011298, -0.00063690, 0.00173005, -0.00059277, -0.00013176, 0.00002392, -0.00008372, -0.00001538, -0.00003231, 0.00007326, -0.00003683, -0.00000906, -0.00003564, 0.00003389, -0.00002937, 0.00001434, -0.00000733, 0.00002916, -0.00003814, 0.00000964, -0.00002072, 0.00002162, -0.00001982, },
		{ -0.00001907, 0.00001654, -0.00000112, 0.00002379, -0.00003708, 0.00001429, -0.00001974, 0.00001839, -0.00002784, 0.00002292, -0.00001618, 0.00002875, -0.00003150, 0.00003066, -0.00002470, 0.00003382, -0.00003863, 0.00003687, -0.00004086, 0.00002039, -0.00003708, 0.00004668, -0.00004255, 0.00004801, -0.00004512, 0.00005945, -0.00004965, 0.00000699, -0.00003882, 0.00007103, -0.00006160, 0.00004348, -0.00005363, 0.00007022, -0.00004806, -0.00003049, -0.00003751, 0.00011966, -0.00007338, -0.00000667, -0.00009234, 0.00003711, -0.00012280, -0.00059209, 0.00166440, -0.00054586, -0.00014247, 0.00001796, -0.00008155, -0.00001689, -0.00002406, 0.00006543, -0.00003751, -0.00000524, -0.00003483, 0.00003049, -0.00002759, 0.00001229, -0.00000327, 0.00002647, -0.00003882, 0.00001055, -0.00001952, 0.00002000, },
		{ 0.00001853, -0.00001844, 0.00001574, 0.00000002, 0.00002355, -0.00003778, 0.00001477, -0.00001975, 0.00001853, -0.00002824, 0.00002335, -0.00001669, 0.00002942, -0.00003220, 0.00003139, -0.00002552, 0.00003464, -0.00003942, 0.00003767, -0.00004162, 0.00002109, -0.00003778, 0.00004725, -0.00004289, 0.00004834, -0.00004542, 0.00005940, -0.00004966, 0.00000779, -0.00003952, 0.00007007, -0.00005979, 0.00004264, -0.00005286, 0.00006832, -0.00004776, -0.00002757, -0.00003822, 0.00011376, -0.00006684, -0.00000783, -0.00009052, 0.00003170, -0.00013314, -0.00054516, 0.00159602, -0.00049727, -0.00015351, 0.00001186, -0.00007935, -0.00001841, -0.00001576, 0.00005759, -0.00003822, -0.00000142, -0.00003406, 0.00002717, -0.00002587, 0.00001033, 0.00000062, 0.00002391, -0.00003952, 0.00001148, -0.00001845, },
		{ -0.00001752, 0.00001724, -0.00001793, 0.00001509, 0.00000097, 0.00002342, -0.00003851, 0.00001531, -0.00001987, 0.00001878, -0.00002872, 0.00002385, -0.00001727, 0.00003012, -0.00003293, 0.00003214, -0.00002631, 0.00003542, -0.00004018, 0.00003840, -0.00004229, 0.00002178, -0.00003851, 0.00004772, -0.00004301, 0.00004850, -0.00004558, 0.00005912, -0.00004950, 0.00000863, -0.00004025, 0.00006884, -0.00005761, 0.00004160, -0.00005191, 0.00006614, -0.00004739, -0.00002441, -0.00003894, 0.00010737, -0.00005982, -0.00000907, -0.00008859, 0.00002602, -0.00014390, -0.00049655, 0.00152557, -0.00044747, -0.00016479, 0.00000569, -0.00007714, -0.00001992, -0.00000747, 0.00004980, -0.00003894, 0.00000238, -0.00003333, 0.00002396, -0.00002426, 0.00000849, 0.00000431, 0.00002153, -0.00004025, 0.00001242, },
		{ 0.00001336, -0.00001672, 0.00001614, -0.00001755, 0.00001459, 0.00000172, 0.00002341, -0.00003925, 0.00001589, -0.00002011, 0.00001914, -0.00002926, 0.00002442, -0.00001790, 0.00003084, -0.00003367, 0.00003289, -0.00002708, 0.00003616, -0.00004089, 0.00003904, -0.00004286, 0.00002246, -0.00003925, 0.00004807, -0.00004291, 0.00004849, -0.00004561, 0.00005863, -0.00004918, 0.00000952, -0.00004099, 0.00006735, -0.00005507, 0.00004038, -0.00005080, 0.00006371, -0.00004693, -0.00002107, -0.00003968, 0.00010058, -0.00005241, -0.00001039, -0.00008657, 0.00002013, -0.00015498, -0.00044673, 0.00145373, -0.00039693, -0.00017617, -0.00000051, -0.00007494, -0.00002140, 0.00000071, 0.00004216, -0.00003968, 0.00000612, -0.00003267, 0.00002091, -0.00002275, 0.00000678, 0.00000776, 0.00001934, -0.00004099, },
		{ -0.00004174, 0.00001429, -0.00001608, 0.00001523, -0.00001731, 0.00001426, 0.00000225, 0.00002353, -0.00004000, 0.00001650, -0.00002046, 0.00001959, -0.00002987, 0.00002506, -0.00001858, 0.00003158, -0.00003442, 0.00003364, -0.00002780, 0.00003685, -0.00004155, 0.00003959, -0.00004332, 0.00002311, -0.00004000, 0.00004830, -0.00004259, 0.00004832, -0.00004550, 0.00005793, -0.00004869, 0.00001043, -0.00004174, 0.00006560, -0.00005220, 0.00003897, -0.00004955, 0.00006105, -0.00004640, -0.00001755, -0.00004043, 0.00009342, -0.00004466, -0.00001178, -0.00008447, 0.00001407, -0.00016627, -0.00039619, 0.00138120, -0.00034615, -0.00018757, -0.00000666, -0.00007279, -0.00002284, 0.00000872, 0.00003472, -0.00004043, 0.00000977, -0.00003207, 0.00001804, -0.00002136, 0.00000521, 0.00001093, 0.00001737, },
		{ 0.00001562, -0.00004249, 0.00001520, -0.00001559, 0.00001453, -0.00001720, 0.00001409, 0.00000257, 0.00002376, -0.00004074, 0.00001715, -0.00002092, 0.00002014, -0.00003053, 0.00002575, -0.00001930, 0.00003232, -0.00003516, 0.00003437, -0.00002848, 0.00003749, -0.00004216, 0.00004005, -0.00004367, 0.00002372, -0.00004074, 0.00004841, -0.00004206, 0.00004799, -0.00004526, 0.00005703, -0.00004805, 0.00001136, -0.00004249, 0.00006362, -0.00004902, 0.00003741, -0.00004816, 0.00005817, -0.00004580, -0.00001391, -0.00004118, 0.00008599, -0.00003666, -0.00001323, -0.00008231, 0.00000792, -0.00017767, -0.00034540, 0.00130866, -0.00029560, -0.00019886, -0.00001272, -0.00007069, -0.00002423, 0.00001646, 0.00002757, -0.00004118, 0.00001328, -0.00003154, 0.00001537, -0.00002011, 0.00000381, 0.00001380, },
		{ 0.00001634, 0.00001413, -0.00004323, 0.00001608, -0.00001527, 0.00001404, -0.00001723, 0.00001408, 0.00000267, 0.00002411, -0.00004148, 0.00001782, -0.00002149, 0.00002078, -0.00003124, 0.00002649, -0.00002007, 0.00003307, -0.00003591, 0.00003509, -0.00002911, 0.00003806, -0.00004270, 0.00004040, -0.00004391, 0.00002430, -0.00004148, 0.00004841, -0.00004131, 0.00004749, -0.00004488, 0.00005592, -0.00004726, 0.00001229, -0.00004323, 0.00006143, -0.00004558, 0.00003570, -0.00004665, 0.00005512, -0.00004513, -0.00001017, -0.00004192, 0.00007835, -0.00002847, -0.00001471, -0.00008011, 0.00000173, -0.00018906, -0.00029486, 0.00123682, -0.00024579, -0.00020994, -0.00001861, -0.00006866, -0.00002556, 0.00002388, 0.00002077, -0.00004192, 0.00001663, -0.00003108, 0.00001294, -0.00001900, 0.00000259, },
		{ 0.00000155, 0.00001853, 0.00001290, -0.00004395, 0.00001693, -0.00001511, 0.00001377, -0.00001740, 0.00001424, 0.00000255, 0.00002457, -0.00004221, 0.00001852, -0.00002216, 0.00002151, -0.00003200, 0.00002727, -0.00002086, 0.00003382, -0.00003663, 0.00003579, -0.00002969, 0.00003856, -0.00004318, 0.00004065, -0.00004404, 0.00002484, -0.00004221, 0.00004828, -0.00004036, 0.00004684, -0.00004438, 0.00005463, -0.00004632, 0.00001323, -0.00004395, 0.00005905, -0.00004189, 0.00003386, -0.00004504, 0.00005192, -0.00004441, -0.00000637, -0.00004265, 0.00007056, -0.00002019, -0.00001622, -0.00007791, -0.00000445, -0.00020033, -0.00024506, 0.00116637, -0.00019718, -0.00022070, -0.00002429, -0.00006673, -0.00002680, 0.00003089, 0.00001439, -0.00004265, 0.00001978, -0.00003070, 0.00001077, -0.00001806, },
		{ -0.00001728, 0.00000071, 0.00002033, 0.00001194, -0.00004466, 0.00001773, -0.00001512, 0.00001372, -0.00001770, 0.00001457, 0.00000221, 0.00002515, -0.00004292, 0.00001922, -0.00002292, 0.00002231, -0.00003279, 0.00002809, -0.00002167, 0.00003456, -0.00003734, 0.00003646, -0.00003020, 0.00003899, -0.00004358, 0.00004079, -0.00004404, 0.00002532, -0.00004292, 0.00004803, -0.00003922, 0.00004604, -0.00004375, 0.00005317, -0.00004525, 0.00001416, -0.00004466, 0.00005650, -0.00003800, 0.00003190, -0.00004333, 0.00004860, -0.00004364, -0.00000254, -0.00004335, 0.00006272, -0.00001188, -0.00001773, -0.00007571, -0.00001054, -0.00021137, -0.00019647, 0.00109799, -0.00015024, -0.00023104, -0.00002970, -0.00006492, -0.00002795, 0.00003744, 0.00000849, -0.00004335, 0.00002270, -0.00003041, 0.00000887, },
		{ 0.00000726, -0.00001667, 0.00000007, 0.00002174, 0.00001127, -0.00004534, 0.00001848, -0.00001529, 0.00001389, -0.00001813, 0.00001505, 0.00000166, 0.00002582, -0.00004359, 0.00001993, -0.00002377, 0.00002319, -0.00003361, 0.00002894, -0.00002250, 0.00003527, -0.00003801, 0.00003709, -0.00003064, 0.00003934, -0.00004391, 0.00004083, -0.00004393, 0.00002576, -0.00004359, 0.00004767, -0.00003789, 0.00004510, -0.00004300, 0.00005154, -0.00004406, 0.00001508, -0.00004534, 0.00005381, -0.00003394, 0.00002985, -0.00004154, 0.00004519, -0.00004283, 0.00000127, -0.00004403, 0.00005489, -0.00000363, -0.00001925, -0.00007353, -0.00001650, -0.00022209, -0.00014957, 0.00103234, -0.00010544, -0.00024086, -0.00003479, -0.00006323, -0.00002900, 0.00004344, 0.00000312, -0.00004403, 0.00002537, -0.00003020, },
		{ -0.00003007, 0.00000597, -0.00001625, -0.00000035, 0.00002274, 0.00001089, -0.00004598, 0.00001918, -0.00001563, 0.00001427, -0.00001869, 0.00001569, 0.00000089, 0.00002660, -0.00004423, 0.00002064, -0.00002470, 0.00002412, -0.00003445, 0.00002980, -0.00002334, 0.00003596, -0.00003866, 0.00003767, -0.00003101, 0.00003961, -0.00004416, 0.00004075, -0.00004371, 0.00002613, -0.00004423, 0.00004720, -0.00003639, 0.00004402, -0.00004213, 0.00004978, -0.00004275, 0.00001596, -0.00004598, 0.00005099, -0.00002975, 0.00002772, -0.00003969, 0.00004173, -0.00004198, 0.00000504, -0.00004467, 0.00004714, 0.00000448, -0.00002074, -0.00007141, -0.00002227, -0.00023236, -0.00010480, 0.00097005, -0.00006319, -0.00025007, -0.00003951, -0.00006170, -0.00002994, 0.00004886, -0.00000166, -0.00004467, 0.00002775, },
		{ 0.00002983, -0.00003004, 0.00000500, -0.00001601, -0.00000056, 0.00002333, 0.00001080, -0.00004658, 0.00001980, -0.00001613, 0.00001487, -0.00001937, 0.00001649, -0.00000008, 0.00002746, -0.00004484, 0.00002134, -0.00002570, 0.00002511, -0.00003530, 0.00003068, -0.00002418, 0.00003662, -0.00003926, 0.00003821, -0.00003131, 0.00003980, -0.00004433, 0.00004057, -0.00004337, 0.00002645, -0.00004484, 0.00004661, -0.00003474, 0.00004282, -0.00004117, 0.00004788, -0.00004134, 0.00001682, -0.00004658, 0.00004809, -0.00002547, 0.00002554, -0.00003781, 0.00003825, -0.00004111, 0.00000871, -0.00004527, 0.00003956, 0.00001237, -0.00002221, -0.00006936, -0.00002778, -0.00024211, -0.00006259, 0.00091171, -0.00002391, -0.00025857, -0.00004381, -0.00006033, -0.00003076, 0.00005363, -0.00000581, -0.00004527, },
		{ -0.00004583, 0.00003159, -0.00003009, 0.00000437, -0.00001596, -0.00000054, 0.00002349, 0.00001101, -0.00004713, 0.00002036, -0.00001679, 0.00001568, -0.00002017, 0.00001742, -0.00000124, 0.00002841, -0.00004539, 0.00002203, -0.00002675, 0.00002614, -0.00003615, 0.00003156, -0.00002500, 0.00003724, -0.00003981, 0.00003869, -0.00003152, 0.00003990, -0.00004441, 0.00004029, -0.00004292, 0.00002670, -0.00004539, 0.00004592, -0.00003295, 0.00004151, -0.00004011, 0.00004588, -0.00003985, 0.00001763, -0.00004713, 0.00004512, -0.00002115, 0.00002333, -0.00003590, 0.00003477, -0.00004022, 0.00001227, -0.00004583, 0.00003222, 0.00001997, -0.00002362, -0.00006739, -0.00003299, -0.00025122, -0.00002336, 0.00085790, 0.00001203, -0.00026628, -0.00004766, -0.00005913, -0.00003144, 0.00005772, -0.00000929, },
		{ -0.00001205, -0.00004633, 0.00003300, -0.00003023, 0.00000407, -0.00001610, -0.00000031, 0.00002323, 0.00001151, -0.00004764, 0.00002083, -0.00001759, 0.00001670, -0.00002108, 0.00001849, -0.00000258, 0.00002942, -0.00004589, 0.00002269, -0.00002786, 0.00002721, -0.00003701, 0.00003243, -0.00002581, 0.00003781, -0.00004031, 0.00003912, -0.00003166, 0.00003992, -0.00004441, 0.00003990, -0.00004236, 0.00002688, -0.00004589, 0.00004514, -0.00003103, 0.00004010, -0.00003897, 0.00004378, -0.00003828, 0.00001838, -0.00004764, 0.00004211, -0.00001682, 0.00002111, -0.00003399, 0.00003135, -0.00003932, 0.00001567, -0.00004633, 0.00002518, 0.00002721, -0.00002498, -0.00006554, -0.00003785, -0.00025963, 0.00001253, 0.00080913, 0.00004427, -0.00027313, -0.00005101, -0.00005813, -0.00003199, 0.00006107, },
		{ 0.00006367, -0.00001409, -0.00004677, 0.00003406, -0.00003046, 0.00000411, -0.00001642, 0.00000013, 0.00002254, 0.00001230, -0.00004808, 0.00002123, -0.00001854, 0.00001790, -0.00002209, 0.00001968, -0.00000409, 0.00003050, -0.00004634, 0.00002333, -0.00002900, 0.00002830, -0.00003784, 0.00003329, -0.00002659, 0.00003834, -0.00004076, 0.00003948, -0.00003172, 0.00003985, -0.00004433, 0.00003941, -0.00004170, 0.00002700, -0.00004634, 0.00004427, -0.00002901, 0.00003861, -0.00003775, 0.00004162, -0.00003666, 0.00001908, -0.00004808, 0.00003910, -0.00001253, 0.00001889, -0.00003209, 0.00002800, -0.00003841, 0.00001888, -0.00004677, 0.00001852, 0.00003401, -0.00002626, -0.00006381, -0.00004231, -0.00026723, 0.00004472, 0.00076587, 0.00007252, -0.00027906, -0.00005383, -0.00005732, -0.00003240, },
		{ -0.00003266, 0.00006547, -0.00001537, -0.00004716, 0.00003475, -0.00003077, 0.00000450, -0.00001693, 0.00000079, 0.00002144, 0.00001338, -0.00004847, 0.00002154, -0.00001963, 0.00001928, -0.00002318, 0.00002098, -0.00000576, 0.00003164, -0.00004672, 0.00002394, -0.00003017, 0.00002940, -0.00003866, 0.00003413, -0.00002734, 0.00003880, -0.00004115, 0.00003978, -0.00003169, 0.00003969, -0.00004417, 0.00003882, -0.00004095, 0.00002705, -0.00004672, 0.00004331, -0.00002690, 0.00003704, -0.00003647, 0.00003941, -0.00003500, 0.00001972, -0.00004847, 0.00003611, -0.00000832, 0.00001671, -0.00003022, 0.00002477, -0.00003752, 0.00002187, -0.00004716, 0.00001230, 0.00004031, -0.00002745, -0.00006222, -0.00004632, -0.00027397, 0.00007290, 0.00072853, 0.00009649, -0.00028400, -0.00005610, -0.00005671, },
		{ -0.00005632, -0.00003278, 0.00006648, -0.00001589, -0.00004748, 0.00003507, -0.00003116, 0.00000522, -0.00001762, 0.00000165, 0.00001993, 0.00001472, -0.00004879, 0.00002176, -0.00002083, 0.00002083, -0.00002436, 0.00002239, -0.00000757, 0.00003281, -0.00004704, 0.00002450, -0.00003136, 0.00003050, -0.00003945, 0.00003494, -0.00002805, 0.00003921, -0.00004147, 0.00004000, -0.00003158, 0.00003945, -0.00004392, 0.00003815, -0.00004011, 0.00002703, -0.00004704, 0.00004229, -0.00002473, 0.00003542, -0.00003515, 0.00003717, -0.00003331, 0.00002029, -0.00004879, 0.00003318, -0.00000423, 0.00001457, -0.00002841, 0.00002168, -0.00003664, 0.00002461, -0.00004748, 0.00000657, 0.00004605, -0.00002855, -0.00006079, -0.00004985, -0.00027977, 0.00009681, 0.00069748, 0.00011596, -0.00028791, -0.00005779, },
		{ -0.00005889, -0.00005614, -0.00003275, 0.00006667, -0.00001564, -0.00004773, 0.00003502, -0.00003163, 0.00000627, -0.00001847, 0.00000271, 0.00001804, 0.00001632, -0.00004904, 0.00002190, -0.00002215, 0.00002253, -0.00002560, 0.00002388, -0.00000949, 0.00003401, -0.00004730, 0.00002502, -0.00003256, 0.00003159, -0.00004020, 0.00003571, -0.00002872, 0.00003956, -0.00004172, 0.00004016, -0.00003139, 0.00003913, -0.00004359, 0.00003740, -0.00003919, 0.00002693, -0.00004730, 0.00004120, -0.00002251, 0.00003377, -0.00003379, 0.00003493, -0.00003161, 0.00002077, -0.00004904, 0.00003032, -0.00000029, 0.00001251, -0.00002666, 0.00001876, -0.00003579, 0.00002708, -0.00004773, 0.00000140, 0.00005117, -0.00002954, -0.00005953, -0.00005287, -0.00028458, 0.00011621, 0.00067301, 0.00013074, -0.00029076, },
		{ -0.00029250, -0.00005939, -0.00005618, -0.00003257, 0.00006604, -0.00001462, -0.00004791, 0.00003459, -0.00003218, 0.00000764, -0.00001950, 0.00000396, 0.00001577, 0.00001816, -0.00004922, 0.00002193, -0.00002356, 0.00002437, -0.00002690, 0.00002544, -0.00001152, 0.00003523, -0.00004748, 0.00002549, -0.00003374, 0.00003267, -0.00004091, 0.00003643, -0.00002933, 0.00003984, -0.00004190, 0.00004024, -0.00003113, 0.00003873, -0.00004319, 0.00003657, -0.00003820, 0.00002677, -0.00004748, 0.00004007, -0.00002027, 0.00003208, -0.00003240, 0.00003270, -0.00002992, 0.00002118, -0.00004922, 0.00002757, 0.00000345, 0.00001055, -0.00002500, 0.00001603, -0.00003497, 0.00002926, -0.00004791, -0.00000317, 0.00005563, -0.00003041, -0.00005846, -0.00005534, -0.00028835, 0.00013092, 0.00065536, 0.00014068, },
		{ 0.00014569, -0.00029314, -0.00005928, -0.00005643, -0.00003224, 0.00006461, -0.00001284, -0.00004802, 0.00003379, -0.00003279, 0.00000933, -0.00002067, 0.00000538, 0.00001315, 0.00002023, -0.00004933, 0.00002188, -0.00002506, 0.00002632, -0.00002824, 0.00002705, -0.00001364, 0.00003646, -0.00004759, 0.00002590, -0.00003491, 0.00003371, -0.00004157, 0.00003710, -0.00002988, 0.00004005, -0.00004201, 0.00004025, -0.00003078, 0.00003826, -0.00004272, 0.00003567, -0.00003715, 0.00002655, -0.00004759, 0.00003889, -0.00001803, 0.00003039, -0.00003101, 0.00003051, -0.00002826, 0.00002150, -0.00004933, 0.00002496, 0.00000696, 0.00000869, -0.00002343, 0.00001354, -0.00003419, 0.00003111, -0.00004802, -0.00000709, 0.00005938, -0.00003115, -0.00005758, -0.00005725, -0.00029105, 0.00014079, 0.00064471, },
		},
	}
	;

	float f128_6[64][128][128] = {
		{ { -0.00010526, -0.00000719, 0.00006673, 0.00002995, 0.00005878, 0.00004499, -0.00005119, -0.00003108, -0.00000533, -0.00001332, -0.00000904, 0.00000760, -0.00000973, -0.00000781, -0.00001797, 0.00000984, 0.00000364, -0.00000510, -0.00000765, 0.00001054, 0.00000409, -0.00000256, -0.00000141, -0.00000499, 0.00000244, 0.00000029, -0.00000364, -0.00000247, 0.00000941, 0.00000276, -0.00000577, -0.00000270, 0.00000161, -0.00000623, 0.00000214, 0.00000560, -0.00000444, 0.00000572, -0.00000587, -0.00000601, 0.00000035, 0.00000505, 0.00000051, -0.00000118, 0.00000260, -0.00000031, 0.00000415, -0.00000560, -0.00000288, -0.00000204, 0.00000752, -0.00000706, 0.00000422, -0.00000374, 0.00000345, -0.00000702, 0.00001033, 0.00000129, -0.00000360, -0.00000156, 0.00000049, -0.00000283, 0.00000140, -0.00000012, -0.00000002, -0.00000012, 0.00000140, -0.00000283, 0.00000049, -0.00000156, -0.00000360, 0.00000129, 0.00001033, -0.00000702, 0.00000345, -0.00000374, 0.00000422, -0.00000706, 0.00000752, -0.00000204, -0.00000288, -0.00000560, 0.00000415, -0.00000031, 0.00000260, -0.00000118, 0.00000051, 0.00000505, 0.00000035, -0.00000601, -0.00000587, 0.00000572, -0.00000444, 0.00000560, 0.00000214, -0.00000623, 0.00000161, -0.00000270, -0.00000577, 0.00000276, 0.00000941, -0.00000247, -0.00000364, 0.00000029, 0.00000244, -0.00000499, -0.00000141, -0.00000256, 0.00000409, 0.00001054, -0.00000765, -0.00000510, 0.00000364, 0.00000984, -0.00001797, -0.00000781, -0.00000973, 0.00000760, -0.00000904, -0.00001332, -0.00000533, -0.00003108, -0.00005119, 0.00004499, 0.00005878, 0.00002995, 0.00006673, -0.00000719, },
		{ -0.00000489, -0.00010538, -0.00000924, 0.00006256, 0.00002910, 0.00005349, 0.00004851, -0.00004567, -0.00003237, -0.00000685, -0.00001115, -0.00001141, 0.00000843, -0.00000924, -0.00000693, -0.00001936, 0.00000902, 0.00000540, -0.00000544, -0.00000785, 0.00000887, 0.00000590, -0.00000317, -0.00000139, -0.00000472, 0.00000222, 0.00000038, -0.00000322, -0.00000347, 0.00000919, 0.00000384, -0.00000602, -0.00000281, 0.00000180, -0.00000578, 0.00000053, 0.00000700, -0.00000488, 0.00000562, -0.00000494, -0.00000668, 0.00000030, 0.00000438, 0.00000150, -0.00000200, 0.00000333, -0.00000115, 0.00000481, -0.00000563, -0.00000217, -0.00000365, 0.00000864, -0.00000731, 0.00000433, -0.00000405, 0.00000382, -0.00000725, 0.00000947, 0.00000249, -0.00000376, -0.00000176, 0.00000069, -0.00000286, 0.00000118, 0.00000012, -0.00000002, -0.00000035, 0.00000157, -0.00000272, 0.00000027, -0.00000140, -0.00000337, 0.00000012, 0.00001098, -0.00000656, 0.00000295, -0.00000334, 0.00000401, -0.00000664, 0.00000628, -0.00000049, -0.00000342, -0.00000564, 0.00000349, 0.00000050, 0.00000194, -0.00000044, -0.00000034, 0.00000551, 0.00000055, -0.00000545, -0.00000656, 0.00000550, -0.00000374, 0.00000407, 0.00000362, -0.00000644, 0.00000127, -0.00000253, -0.00000555, 0.00000183, 0.00000938, -0.00000137, -0.00000400, 0.00000015, 0.00000264, -0.00000514, -0.00000165, -0.00000171, 0.00000218, 0.00001202, -0.00000726, -0.00000486, 0.00000218, 0.00001013, -0.00001623, -0.00000877, -0.00001010, 0.00000655, -0.00000664, -0.00001516, -0.00000448, -0.00002894, -0.00005664, 0.00004048, 0.00006420, 0.00003104, 0.00007087, },
		{ 0.00007498, -0.00000233, -0.00010575, -0.00001106, 0.00005838, 0.00002848, 0.00004836, 0.00005118, -0.00004032, -0.00003272, -0.00000895, -0.00000880, -0.00001363, 0.00000894, -0.00000853, -0.00000620, -0.00002032, 0.00000768, 0.00000741, -0.00000580, -0.00000793, 0.00000710, 0.00000756, -0.00000349, -0.00000160, -0.00000436, 0.00000200, 0.00000044, -0.00000278, -0.00000433, 0.00000869, 0.00000507, -0.00000629, -0.00000288, 0.00000186, -0.00000514, -0.00000114, 0.00000820, -0.00000499, 0.00000519, -0.00000377, -0.00000746, 0.00000038, 0.00000352, 0.00000261, -0.00000288, 0.00000409, -0.00000200, 0.00000546, -0.00000573, -0.00000132, -0.00000527, 0.00000959, -0.00000738, 0.00000432, -0.00000426, 0.00000404, -0.00000726, 0.00000841, 0.00000369, -0.00000383, -0.00000198, 0.00000084, -0.00000282, 0.00000090, 0.00000037, -0.00000000, -0.00000056, 0.00000167, -0.00000255, 0.00000002, -0.00000127, -0.00000308, -0.00000098, 0.00001139, -0.00000589, 0.00000235, -0.00000287, 0.00000373, -0.00000609, 0.00000497, 0.00000094, -0.00000377, -0.00000575, 0.00000282, 0.00000127, 0.00000137, 0.00000020, -0.00000104, 0.00000576, 0.00000088, -0.00000501, -0.00000700, 0.00000501, -0.00000283, 0.00000253, 0.00000489, -0.00000639, 0.00000078, -0.00000232, -0.00000537, 0.00000106, 0.00000910, -0.00000021, -0.00000428, -0.00000004, 0.00000281, -0.00000516, -0.00000209, -0.00000067, 0.00000027, 0.00001326, -0.00000665, -0.00000476, 0.00000108, 0.00000993, -0.00001422, -0.00000972, -0.00001041, 0.00000538, -0.00000437, -0.00001658, -0.00000433, -0.00002606, -0.00006175, 0.00003483, 0.00006970, 0.00003242, },
		{ 0.00003414, 0.00007905, 0.00000053, -0.00010638, -0.00001263, 0.00005415, 0.00002807, 0.00004337, 0.00005315, -0.00003531, -0.00003212, -0.00001150, -0.00000644, -0.00001556, 0.00000903, -0.00000757, -0.00000569, -0.00002084, 0.00000586, 0.00000957, -0.00000612, -0.00000793, 0.00000529, 0.00000899, -0.00000350, -0.00000201, -0.00000392, 0.00000178, 0.00000047, -0.00000235, -0.00000500, 0.00000790, 0.00000641, -0.00000654, -0.00000292, 0.00000182, -0.00000437, -0.00000277, 0.00000911, -0.00000471, 0.00000442, -0.00000238, -0.00000832, 0.00000061, 0.00000248, 0.00000381, -0.00000379, 0.00000488, -0.00000284, 0.00000608, -0.00000589, -0.00000036, -0.00000684, 0.00001031, -0.00000724, 0.00000419, -0.00000435, 0.00000409, -0.00000705, 0.00000718, 0.00000485, -0.00000380, -0.00000221, 0.00000095, -0.00000271, 0.00000058, 0.00000062, 0.00000002, -0.00000075, 0.00000171, -0.00000232, -0.00000023, -0.00000118, -0.00000274, -0.00000200, 0.00001157, -0.00000504, 0.00000167, -0.00000238, 0.00000339, -0.00000545, 0.00000364, 0.00000222, -0.00000392, -0.00000593, 0.00000215, 0.00000197, 0.00000089, 0.00000072, -0.00000158, 0.00000582, 0.00000133, -0.00000466, -0.00000723, 0.00000430, -0.00000181, 0.00000104, 0.00000589, -0.00000605, 0.00000016, -0.00000206, -0.00000523, 0.00000044, 0.00000864, 0.00000095, -0.00000445, -0.00000026, 0.00000294, -0.00000506, -0.00000272, 0.00000050, -0.00000156, 0.00001418, -0.00000577, -0.00000484, 0.00000033, 0.00000929, -0.00001206, -0.00001055, -0.00001074, 0.00000418, -0.00000231, -0.00001750, -0.00000485, -0.00002264, -0.00006626, 0.00002796, 0.00007519, },
		{ 0.00008053, 0.00003630, 0.00008303, 0.00000372, -0.00010729, -0.00001395, 0.00004987, 0.00002787, 0.00003853, 0.00005458, -0.00003081, -0.00003062, -0.00001431, -0.00000425, -0.00001704, 0.00000863, -0.00000630, -0.00000543, -0.00002090, 0.00000363, 0.00001178, -0.00000630, -0.00000792, 0.00000353, 0.00001012, -0.00000318, -0.00000262, -0.00000342, 0.00000157, 0.00000048, -0.00000195, -0.00000546, 0.00000684, 0.00000782, -0.00000674, -0.00000296, 0.00000171, -0.00000354, -0.00000426, 0.00000965, -0.00000402, 0.00000330, -0.00000079, -0.00000925, 0.00000096, 0.00000128, 0.00000507, -0.00000470, 0.00000566, -0.00000365, 0.00000666, -0.00000608, 0.00000066, -0.00000830, 0.00001077, -0.00000688, 0.00000392, -0.00000431, 0.00000397, -0.00000662, 0.00000583, 0.00000595, -0.00000367, -0.00000246, 0.00000100, -0.00000253, 0.00000023, 0.00000087, 0.00000006, -0.00000091, 0.00000168, -0.00000204, -0.00000048, -0.00000113, -0.00000238, -0.00000291, 0.00001151, -0.00000404, 0.00000096, -0.00000187, 0.00000301, -0.00000475, 0.00000233, 0.00000330, -0.00000387, -0.00000615, 0.00000149, 0.00000261, 0.00000053, 0.00000111, -0.00000195, 0.00000568, 0.00000188, -0.00000441, -0.00000726, 0.00000343, -0.00000075, -0.00000030, 0.00000658, -0.00000544, -0.00000058, -0.00000178, -0.00000512, -0.00000003, 0.00000802, 0.00000207, -0.00000450, -0.00000052, 0.00000302, -0.00000482, -0.00000352, 0.00000175, -0.00000324, 0.00001475, -0.00000461, -0.00000510, -0.00000007, 0.00000829, -0.00000986, -0.00001118, -0.00001117, 0.00000304, -0.00000053, -0.00001790, -0.00000598, -0.00001888, -0.00006991, 0.00001981, },
		{ 0.00001040, 0.00008554, 0.00003898, 0.00008689, 0.00000730, -0.00010851, -0.00001502, 0.00004552, 0.00002789, 0.00003380, 0.00005560, -0.00002691, -0.00002835, -0.00001719, -0.00000242, -0.00001798, 0.00000769, -0.00000472, -0.00000541, -0.00002053, 0.00000108, 0.00001392, -0.00000628, -0.00000796, 0.00000188, 0.00001092, -0.00000254, -0.00000339, -0.00000289, 0.00000139, 0.00000049, -0.00000162, -0.00000569, 0.00000553, 0.00000925, -0.00000685, -0.00000303, 0.00000158, -0.00000272, -0.00000554, 0.00000975, -0.00000290, 0.00000188, 0.00000095, -0.00001022, 0.00000142, -0.00000004, 0.00000634, -0.00000559, 0.00000641, -0.00000441, 0.00000718, -0.00000628, 0.00000169, -0.00000958, 0.00001093, -0.00000629, 0.00000351, -0.00000413, 0.00000368, -0.00000601, 0.00000439, 0.00000694, -0.00000342, -0.00000270, 0.00000099, -0.00000228, -0.00000015, 0.00000110, 0.00000010, -0.00000103, 0.00000159, -0.00000173, -0.00000072, -0.00000112, -0.00000200, -0.00000369, 0.00001122, -0.00000291, 0.00000024, -0.00000138, 0.00000263, -0.00000401, 0.00000109, 0.00000417, -0.00000362, -0.00000642, 0.00000083, 0.00000315, 0.00000029, 0.00000138, -0.00000216, 0.00000539, 0.00000250, -0.00000423, -0.00000713, 0.00000245, 0.00000028, -0.00000145, 0.00000693, -0.00000457, -0.00000139, -0.00000151, -0.00000502, -0.00000039, 0.00000729, 0.00000310, -0.00000442, -0.00000081, 0.00000304, -0.00000446, -0.00000443, 0.00000300, -0.00000470, 0.00001492, -0.00000317, -0.00000553, -0.00000016, 0.00000703, -0.00000774, -0.00001152, -0.00001174, 0.00000201, 0.00000091, -0.00001779, -0.00000759, -0.00001504, -0.00007245, },
		{ -0.00007370, -0.00000017, 0.00008999, 0.00004230, 0.00009058, 0.00001129, -0.00011006, -0.00001581, 0.00004107, 0.00002813, 0.00002917, 0.00005631, -0.00002363, -0.00002545, -0.00001990, -0.00000110, -0.00001826, 0.00000620, -0.00000286, -0.00000562, -0.00001980, -0.00000167, 0.00001588, -0.00000597, -0.00000809, 0.00000038, 0.00001136, -0.00000159, -0.00000429, -0.00000235, 0.00000123, 0.00000049, -0.00000138, -0.00000568, 0.00000402, 0.00001063, -0.00000681, -0.00000318, 0.00000149, -0.00000199, -0.00000652, 0.00000938, -0.00000136, 0.00000018, 0.00000279, -0.00001119, 0.00000195, -0.00000144, 0.00000759, -0.00000641, 0.00000709, -0.00000509, 0.00000763, -0.00000648, 0.00000269, -0.00001065, 0.00001076, -0.00000547, 0.00000298, -0.00000383, 0.00000323, -0.00000522, 0.00000291, 0.00000781, -0.00000307, -0.00000293, 0.00000091, -0.00000198, -0.00000055, 0.00000132, 0.00000016, -0.00000112, 0.00000145, -0.00000138, -0.00000094, -0.00000115, -0.00000163, -0.00000433, 0.00001071, -0.00000170, -0.00000045, -0.00000093, 0.00000226, -0.00000328, -0.00000005, 0.00000480, -0.00000320, -0.00000670, 0.00000017, 0.00000361, 0.00000016, 0.00000154, -0.00000222, 0.00000495, 0.00000317, -0.00000409, -0.00000687, 0.00000143, 0.00000120, -0.00000234, 0.00000693, -0.00000348, -0.00000222, -0.00000128, -0.00000493, -0.00000065, 0.00000650, 0.00000402, -0.00000422, -0.00000109, 0.00000299, -0.00000398, -0.00000543, 0.00000420, -0.00000587, 0.00001468, -0.00000148, -0.00000612, -0.00000001, 0.00000562, -0.00000583, -0.00001149, -0.00001249, 0.00000110, 0.00000201, -0.00001724, -0.00000953, -0.00001135, },
		{ -0.00000807, -0.00007352, -0.00001176, 0.00009360, 0.00004637, 0.00009409, 0.00001575, -0.00011195, -0.00001634, 0.00003653, 0.00002857, 0.00002465, 0.00005677, -0.00002095, -0.00002213, -0.00002222, -0.00000045, -0.00001785, 0.00000417, -0.00000075, -0.00000599, -0.00001880, -0.00000449, 0.00001751, -0.00000532, -0.00000835, -0.00000092, 0.00001142, -0.00000036, -0.00000527, -0.00000185, 0.00000111, 0.00000051, -0.00000123, -0.00000546, 0.00000236, 0.00001188, -0.00000657, -0.00000345, 0.00000148, -0.00000141, -0.00000713, 0.00000851, 0.00000058, -0.00000173, 0.00000468, -0.00001212, 0.00000254, -0.00000289, 0.00000876, -0.00000713, 0.00000768, -0.00000567, 0.00000798, -0.00000663, 0.00000361, -0.00001145, 0.00001026, -0.00000444, 0.00000234, -0.00000340, 0.00000264, -0.00000431, 0.00000142, 0.00000852, -0.00000261, -0.00000314, 0.00000076, -0.00000162, -0.00000094, 0.00000151, 0.00000022, -0.00000118, 0.00000126, -0.00000102, -0.00000112, -0.00000121, -0.00000127, -0.00000483, 0.00001002, -0.00000045, -0.00000107, -0.00000054, 0.00000191, -0.00000257, -0.00000105, 0.00000520, -0.00000262, -0.00000698, -0.00000049, 0.00000397, 0.00000016, 0.00000158, -0.00000216, 0.00000440, 0.00000384, -0.00000398, -0.00000653, 0.00000042, 0.00000197, -0.00000296, 0.00000661, -0.00000222, -0.00000303, -0.00000111, -0.00000482, -0.00000084, 0.00000568, 0.00000480, -0.00000389, -0.00000137, 0.00000286, -0.00000340, -0.00000647, 0.00000527, -0.00000673, 0.00001402, 0.00000044, -0.00000681, 0.00000032, 0.00000419, -0.00000422, -0.00001106, -0.00001342, 0.00000033, 0.00000280, -0.00001633, -0.00001163, },
		{ -0.00001371, -0.00000539, -0.00007184, -0.00002414, 0.00009609, 0.00005129, 0.00009739, 0.00002069, -0.00011419, -0.00001661, 0.00003189, 0.00002919, 0.00002024, 0.00005703, -0.00001882, -0.00001861, -0.00002394, -0.00000057, -0.00001673, 0.00000166, 0.00000149, -0.00000644, -0.00001762, -0.00000724, 0.00001873, -0.00000428, -0.00000875, -0.00000200, 0.00001111, 0.00000111, -0.00000628, -0.00000142, 0.00000105, 0.00000053, -0.00000118, -0.00000504, 0.00000063, 0.00001293, -0.00000608, -0.00000388, 0.00000160, -0.00000105, -0.00000733, 0.00000714, 0.00000285, -0.00000378, 0.00000654, -0.00001297, 0.00000315, -0.00000434, 0.00000982, -0.00000771, 0.00000815, -0.00000612, 0.00000822, -0.00000672, 0.00000440, -0.00001194, 0.00000941, -0.00000321, 0.00000160, -0.00000286, 0.00000192, -0.00000330, -0.00000003, 0.00000904, -0.00000205, -0.00000331, 0.00000055, -0.00000123, -0.00000133, 0.00000166, 0.00000029, -0.00000120, 0.00000102, -0.00000067, -0.00000127, -0.00000130, -0.00000094, -0.00000518, 0.00000916, 0.00000080, -0.00000160, -0.00000023, 0.00000161, -0.00000190, -0.00000191, 0.00000537, -0.00000191, -0.00000723, -0.00000115, 0.00000423, 0.00000027, 0.00000153, -0.00000199, 0.00000378, 0.00000450, -0.00000386, -0.00000615, -0.00000052, 0.00000254, -0.00000328, 0.00000599, -0.00000085, -0.00000376, -0.00000104, -0.00000468, -0.00000097, 0.00000486, 0.00000543, -0.00000346, -0.00000162, 0.00000265, -0.00000274, -0.00000751, 0.00000618, -0.00000723, 0.00001298, 0.00000252, -0.00000754, 0.00000073, 0.00000284, -0.00000302, -0.00001021, -0.00001450, -0.00000034, 0.00000334, -0.00001519, },
		{ -0.00001395, -0.00001558, -0.00000350, -0.00006866, -0.00003699, 0.00009715, 0.00005712, 0.00010049, 0.00002612, -0.00011676, -0.00001665, 0.00002718, 0.00002995, 0.00001599, 0.00005707, -0.00001711, -0.00001515, -0.00002487, -0.00000151, -0.00001494, -0.00000121, 0.00000378, -0.00000688, -0.00001638, -0.00000979, 0.00001943, -0.00000284, -0.00000930, -0.00000286, 0.00001045, 0.00000276, -0.00000726, -0.00000107, 0.00000104, 0.00000057, -0.00000121, -0.00000447, -0.00000110, 0.00001368, -0.00000528, -0.00000451, 0.00000191, -0.00000095, -0.00000709, 0.00000530, 0.00000540, -0.00000588, 0.00000831, -0.00001370, 0.00000374, -0.00000576, 0.00001073, -0.00000814, 0.00000849, -0.00000644, 0.00000833, -0.00000671, 0.00000502, -0.00001210, 0.00000824, -0.00000182, 0.00000079, -0.00000224, 0.00000111, -0.00000224, -0.00000139, 0.00000938, -0.00000140, -0.00000343, 0.00000027, -0.00000081, -0.00000169, 0.00000178, 0.00000037, -0.00000119, 0.00000076, -0.00000032, -0.00000137, -0.00000142, -0.00000064, -0.00000538, 0.00000817, 0.00000200, -0.00000201, -0.00000000, 0.00000135, -0.00000129, -0.00000260, 0.00000533, -0.00000111, -0.00000743, -0.00000182, 0.00000439, 0.00000049, 0.00000139, -0.00000173, 0.00000311, 0.00000511, -0.00000371, -0.00000575, -0.00000134, 0.00000288, -0.00000331, 0.00000512, 0.00000055, -0.00000436, -0.00000109, -0.00000450, -0.00000107, 0.00000408, 0.00000590, -0.00000296, -0.00000181, 0.00000236, -0.00000202, -0.00000851, 0.00000687, -0.00000736, 0.00001159, 0.00000470, -0.00000825, 0.00000115, 0.00000168, -0.00000226, -0.00000896, -0.00001567, -0.00000096, 0.00000370, },
		{ 0.00000397, -0.00001273, -0.00001710, -0.00000251, -0.00006408, -0.00004996, 0.00009646, 0.00006390, 0.00010344, 0.00003203, -0.00011963, -0.00001651, 0.00002245, 0.00003078, 0.00001195, 0.00005687, -0.00001569, -0.00001195, -0.00002487, -0.00000329, -0.00001258, -0.00000431, 0.00000597, -0.00000720, -0.00001521, -0.00001201, 0.00001955, -0.00000098, -0.00000996, -0.00000350, 0.00000948, 0.00000451, -0.00000816, -0.00000086, 0.00000110, 0.00000061, -0.00000131, -0.00000380, -0.00000273, 0.00001407, -0.00000413, -0.00000536, 0.00000242, -0.00000115, -0.00000641, 0.00000304, 0.00000814, -0.00000796, 0.00000992, -0.00001428, 0.00000429, -0.00000708, 0.00001144, -0.00000838, 0.00000866, -0.00000660, 0.00000829, -0.00000658, 0.00000545, -0.00001191, 0.00000676, -0.00000032, -0.00000005, -0.00000157, 0.00000023, -0.00000116, -0.00000263, 0.00000950, -0.00000068, -0.00000348, -0.00000007, -0.00000037, -0.00000202, 0.00000186, 0.00000045, -0.00000115, 0.00000047, 0.00000001, -0.00000143, -0.00000156, -0.00000038, -0.00000545, 0.00000708, 0.00000313, -0.00000227, 0.00000013, 0.00000115, -0.00000076, -0.00000313, 0.00000508, -0.00000025, -0.00000756, -0.00000248, 0.00000444, 0.00000080, 0.00000120, -0.00000143, 0.00000243, 0.00000564, -0.00000352, -0.00000538, -0.00000201, 0.00000298, -0.00000308, 0.00000407, 0.00000192, -0.00000479, -0.00000128, -0.00000428, -0.00000114, 0.00000333, 0.00000622, -0.00000240, -0.00000194, 0.00000200, -0.00000127, -0.00000941, 0.00000730, -0.00000712, 0.00000990, 0.00000691, -0.00000886, 0.00000147, 0.00000080, -0.00000199, -0.00000737, -0.00001687, -0.00000160, },
		{ -0.00000234, 0.00000426, -0.00001167, -0.00001812, -0.00000247, -0.00005826, -0.00006265, 0.00009375, 0.00007162, 0.00010630, 0.00003838, -0.00012274, -0.00001625, 0.00001775, 0.00003161, 0.00000819, 0.00005637, -0.00001441, -0.00000924, -0.00002385, -0.00000583, -0.00000978, -0.00000749, 0.00000792, -0.00000728, -0.00001421, -0.00001380, 0.00001905, 0.00000124, -0.00001072, -0.00000396, 0.00000825, 0.00000630, -0.00000891, -0.00000081, 0.00000125, 0.00000063, -0.00000145, -0.00000310, -0.00000416, 0.00001403, -0.00000262, -0.00000644, 0.00000317, -0.00000166, -0.00000530, 0.00000041, 0.00001095, -0.00000990, 0.00001130, -0.00001466, 0.00000477, -0.00000828, 0.00001192, -0.00000842, 0.00000867, -0.00000659, 0.00000809, -0.00000630, 0.00000565, -0.00001137, 0.00000501, 0.00000126, -0.00000090, -0.00000085, -0.00000067, -0.00000011, -0.00000372, 0.00000942, 0.00000008, -0.00000347, -0.00000047, 0.00000006, -0.00000230, 0.00000191, 0.00000054, -0.00000109, 0.00000017, 0.00000030, -0.00000143, -0.00000172, -0.00000017, -0.00000540, 0.00000593, 0.00000414, -0.00000239, 0.00000017, 0.00000100, -0.00000030, -0.00000349, 0.00000467, 0.00000062, -0.00000760, -0.00000313, 0.00000439, 0.00000118, 0.00000097, -0.00000109, 0.00000177, 0.00000607, -0.00000325, -0.00000506, -0.00000251, 0.00000284, -0.00000262, 0.00000290, 0.00000319, -0.00000500, -0.00000160, -0.00000403, -0.00000119, 0.00000265, 0.00000640, -0.00000182, -0.00000199, 0.00000158, -0.00000052, -0.00001019, 0.00000746, -0.00000654, 0.00000797, 0.00000905, -0.00000930, 0.00000164, 0.00000025, -0.00000219, -0.00000549, -0.00001800, },
		{ -0.00001894, -0.00000327, 0.00000464, -0.00001087, -0.00001858, -0.00000339, -0.00005144, -0.00007462, 0.00008877, 0.00008019, 0.00010920, 0.00004510, -0.00012602, -0.00001594, 0.00001316, 0.00003236, 0.00000480, 0.00005550, -0.00001311, -0.00000718, -0.00002180, -0.00000904, -0.00000671, -0.00001055, 0.00000949, -0.00000702, -0.00001348, -0.00001508, 0.00001792, 0.00000377, -0.00001149, -0.00000427, 0.00000684, 0.00000803, -0.00000945, -0.00000094, 0.00000150, 0.00000061, -0.00000158, -0.00000245, -0.00000532, 0.00001349, -0.00000072, -0.00000776, 0.00000414, -0.00000248, -0.00000381, -0.00000247, 0.00001374, -0.00001163, 0.00001241, -0.00001482, 0.00000514, -0.00000931, 0.00001215, -0.00000823, 0.00000849, -0.00000640, 0.00000772, -0.00000588, 0.00000561, -0.00001049, 0.00000303, 0.00000286, -0.00000171, -0.00000014, -0.00000158, 0.00000087, -0.00000462, 0.00000915, 0.00000088, -0.00000337, -0.00000091, 0.00000047, -0.00000253, 0.00000190, 0.00000063, -0.00000100, -0.00000013, 0.00000055, -0.00000138, -0.00000188, -0.00000001, -0.00000524, 0.00000475, 0.00000500, -0.00000234, 0.00000013, 0.00000090, 0.00000008, -0.00000370, 0.00000412, 0.00000147, -0.00000754, -0.00000378, 0.00000423, 0.00000163, 0.00000074, -0.00000076, 0.00000117, 0.00000638, -0.00000291, -0.00000481, -0.00000282, 0.00000248, -0.00000199, 0.00000167, 0.00000430, -0.00000498, -0.00000206, -0.00000374, -0.00000122, 0.00000202, 0.00000646, -0.00000125, -0.00000194, 0.00000111, 0.00000020, -0.00001080, 0.00000733, -0.00000564, 0.00000590, 0.00001106, -0.00000952, 0.00000159, 0.00000006, -0.00000282, -0.00000346, },
		{ -0.00000138, -0.00001958, -0.00000446, 0.00000520, -0.00001040, -0.00001842, -0.00000518, -0.00004391, -0.00008544, 0.00008134, 0.00008946, 0.00011231, 0.00005210, -0.00012937, -0.00001569, 0.00000875, 0.00003292, 0.00000185, 0.00005419, -0.00001167, -0.00000590, -0.00001874, -0.00001274, -0.00000358, -0.00001330, 0.00001053, -0.00000632, -0.00001311, -0.00001581, 0.00001620, 0.00000652, -0.00001222, -0.00000450, 0.00000531, 0.00000964, -0.00000973, -0.00000129, 0.00000187, 0.00000054, -0.00000166, -0.00000190, -0.00000611, 0.00001241, 0.00000157, -0.00000930, 0.00000533, -0.00000357, -0.00000200, -0.00000551, 0.00001639, -0.00001305, 0.00001318, -0.00001474, 0.00000539, -0.00001014, 0.00001209, -0.00000782, 0.00000812, -0.00000602, 0.00000718, -0.00000528, 0.00000532, -0.00000930, 0.00000089, 0.00000443, -0.00000245, 0.00000055, -0.00000244, 0.00000176, -0.00000533, 0.00000868, 0.00000168, -0.00000319, -0.00000138, 0.00000085, -0.00000270, 0.00000186, 0.00000073, -0.00000090, -0.00000043, 0.00000075, -0.00000129, -0.00000205, 0.00000010, -0.00000498, 0.00000358, 0.00000570, -0.00000214, 0.00000002, 0.00000084, 0.00000039, -0.00000377, 0.00000347, 0.00000227, -0.00000737, -0.00000440, 0.00000398, 0.00000211, 0.00000051, -0.00000044, 0.00000063, 0.00000656, -0.00000249, -0.00000464, -0.00000295, 0.00000193, -0.00000124, 0.00000047, 0.00000518, -0.00000470, -0.00000264, -0.00000345, -0.00000123, 0.00000145, 0.00000643, -0.00000071, -0.00000181, 0.00000062, 0.00000086, -0.00001123, 0.00000691, -0.00000448, 0.00000376, 0.00001284, -0.00000944, 0.00000129, 0.00000026, -0.00000382, },
		{ -0.00000509, 0.00000061, -0.00001983, -0.00000598, 0.00000600, -0.00001031, -0.00001767, -0.00000770, -0.00003603, -0.00009468, 0.00007137, 0.00009919, 0.00011584, 0.00005927, -0.00013269, -0.00001559, 0.00000463, 0.00003321, -0.00000057, 0.00005239, -0.00000998, -0.00000545, -0.00001480, -0.00001671, -0.00000062, -0.00001556, 0.00001092, -0.00000513, -0.00001312, -0.00001597, 0.00001396, 0.00000938, -0.00001281, -0.00000470, 0.00000375, 0.00001103, -0.00000969, -0.00000187, 0.00000237, 0.00000037, -0.00000164, -0.00000155, -0.00000647, 0.00001077, 0.00000420, -0.00001102, 0.00000671, -0.00000489, 0.00000006, -0.00000860, 0.00001877, -0.00001408, 0.00001359, -0.00001441, 0.00000549, -0.00001075, 0.00001174, -0.00000718, 0.00000757, -0.00000547, 0.00000646, -0.00000452, 0.00000481, -0.00000783, -0.00000135, 0.00000590, -0.00000309, 0.00000118, -0.00000323, 0.00000251, -0.00000583, 0.00000803, 0.00000246, -0.00000291, -0.00000188, 0.00000119, -0.00000281, 0.00000178, 0.00000083, -0.00000079, -0.00000070, 0.00000089, -0.00000114, -0.00000222, 0.00000015, -0.00000466, 0.00000244, 0.00000622, -0.00000180, -0.00000016, 0.00000081, 0.00000063, -0.00000371, 0.00000275, 0.00000298, -0.00000710, -0.00000500, 0.00000364, 0.00000261, 0.00000032, -0.00000017, 0.00000018, 0.00000660, -0.00000199, -0.00000454, -0.00000291, 0.00000124, -0.00000043, -0.00000063, 0.00000581, -0.00000418, -0.00000331, -0.00000318, -0.00000120, 0.00000092, 0.00000633, -0.00000023, -0.00000160, 0.00000013, 0.00000143, -0.00001145, 0.00000622, -0.00000310, 0.00000163, 0.00001433, -0.00000904, 0.00000070, 0.00000080, },
		{ 0.00000164, -0.00000651, 0.00000237, -0.00001959, -0.00000787, 0.00000704, -0.00001061, -0.00001639, -0.00001078, -0.00002816, -0.00010197, 0.00005884, 0.00010907, 0.00012001, 0.00006651, -0.00013588, -0.00001574, 0.00000085, 0.00003314, -0.00000240, 0.00005005, -0.00000796, -0.00000584, -0.00001014, -0.00002071, 0.00000195, -0.00001716, 0.00001058, -0.00000341, -0.00001353, -0.00001561, 0.00001130, 0.00001221, -0.00001318, -0.00000494, 0.00000224, 0.00001212, -0.00000930, -0.00000271, 0.00000301, 0.00000009, -0.00000147, -0.00000145, -0.00000634, 0.00000859, 0.00000713, -0.00001287, 0.00000822, -0.00000639, 0.00000225, -0.00001160, 0.00002080, -0.00001467, 0.00001361, -0.00001383, 0.00000545, -0.00001111, 0.00001109, -0.00000632, 0.00000685, -0.00000475, 0.00000558, -0.00000361, 0.00000407, -0.00000614, -0.00000363, 0.00000722, -0.00000358, 0.00000174, -0.00000393, 0.00000312, -0.00000611, 0.00000725, 0.00000320, -0.00000254, -0.00000238, 0.00000147, -0.00000285, 0.00000166, 0.00000093, -0.00000067, -0.00000095, 0.00000097, -0.00000095, -0.00000238, 0.00000016, -0.00000427, 0.00000136, 0.00000654, -0.00000132, -0.00000036, 0.00000081, 0.00000082, -0.00000355, 0.00000200, 0.00000359, -0.00000671, -0.00000556, 0.00000321, 0.00000310, 0.00000019, 0.00000005, -0.00000017, 0.00000650, -0.00000141, -0.00000452, -0.00000272, 0.00000044, 0.00000036, -0.00000160, 0.00000615, -0.00000343, -0.00000403, -0.00000295, -0.00000112, 0.00000042, 0.00000619, 0.00000018, -0.00000132, -0.00000033, 0.00000189, -0.00001146, 0.00000528, -0.00000159, -0.00000039, 0.00001547, -0.00000830, -0.00000016, },
		{ -0.00000126, 0.00000270, -0.00000796, 0.00000376, -0.00001878, -0.00001013, 0.00000832, -0.00001124, -0.00001469, -0.00001417, -0.00002068, -0.00010699, 0.00004388, 0.00011869, 0.00012510, 0.00007368, -0.00013883, -0.00001622, -0.00000249, 0.00003264, -0.00000360, 0.00004716, -0.00000559, -0.00000701, -0.00000497, -0.00002448, 0.00000391, -0.00001795, 0.00000946, -0.00000118, -0.00001431, -0.00001479, 0.00000835, 0.00001488, -0.00001323, -0.00000529, 0.00000086, 0.00001285, -0.00000852, -0.00000380, 0.00000382, -0.00000034, -0.00000112, -0.00000166, -0.00000569, 0.00000588, 0.00001030, -0.00001479, 0.00000980, -0.00000797, 0.00000449, -0.00001440, 0.00002235, -0.00001476, 0.00001324, -0.00001300, 0.00000524, -0.00001120, 0.00001014, -0.00000527, 0.00000597, -0.00000386, 0.00000456, -0.00000255, 0.00000315, -0.00000428, -0.00000587, 0.00000834, -0.00000392, 0.00000220, -0.00000452, 0.00000356, -0.00000619, 0.00000634, 0.00000386, -0.00000207, -0.00000288, 0.00000168, -0.00000282, 0.00000152, 0.00000103, -0.00000055, -0.00000116, 0.00000099, -0.00000073, -0.00000252, 0.00000013, -0.00000384, 0.00000037, 0.00000668, -0.00000074, -0.00000058, 0.00000080, 0.00000096, -0.00000330, 0.00000125, 0.00000407, -0.00000623, -0.00000607, 0.00000271, 0.00000357, 0.00000013, 0.00000019, -0.00000041, 0.00000627, -0.00000078, -0.00000457, -0.00000240, -0.00000040, 0.00000110, -0.00000236, 0.00000620, -0.00000248, -0.00000476, -0.00000279, -0.00000099, -0.00000005, 0.00000602, 0.00000051, -0.00000098, -0.00000074, 0.00000222, -0.00001126, 0.00000413, -0.00000001, -0.00000224, 0.00001622, -0.00000722, },
		{ -0.00000581, -0.00000255, 0.00000389, -0.00000929, 0.00000468, -0.00001738, -0.00001275, 0.00000981, -0.00001213, -0.00001271, -0.00001761, -0.00001395, -0.00010951, 0.00002670, 0.00012759, 0.00013137, 0.00008071, -0.00014146, -0.00001711, -0.00000536, 0.00003167, -0.00000414, 0.00004374, -0.00000290, -0.00000883, 0.00000044, -0.00002773, 0.00000506, -0.00001785, 0.00000755, 0.00000151, -0.00001539, -0.00001361, 0.00000527, 0.00001724, -0.00001288, -0.00000579, -0.00000032, 0.00001316, -0.00000732, -0.00000516, 0.00000480, -0.00000094, -0.00000052, -0.00000221, -0.00000453, 0.00000271, 0.00001361, -0.00001669, 0.00001138, -0.00000957, 0.00000665, -0.00001687, 0.00002337, -0.00001433, 0.00001249, -0.00001194, 0.00000488, -0.00001104, 0.00000891, -0.00000403, 0.00000495, -0.00000284, 0.00000340, -0.00000138, 0.00000207, -0.00000231, -0.00000799, 0.00000922, -0.00000408, 0.00000255, -0.00000498, 0.00000383, -0.00000607, 0.00000535, 0.00000443, -0.00000151, -0.00000335, 0.00000182, -0.00000273, 0.00000134, 0.00000113, -0.00000044, -0.00000133, 0.00000095, -0.00000048, -0.00000264, 0.00000005, -0.00000339, -0.00000052, 0.00000662, -0.00000007, -0.00000079, 0.00000079, 0.00000105, -0.00000299, 0.00000053, 0.00000441, -0.00000566, -0.00000652, 0.00000214, 0.00000398, 0.00000016, 0.00000026, -0.00000055, 0.00000593, -0.00000010, -0.00000466, -0.00000199, -0.00000124, 0.00000171, -0.00000290, 0.00000596, -0.00000137, -0.00000546, -0.00000274, -0.00000079, -0.00000051, 0.00000585, 0.00000075, -0.00000061, -0.00000107, 0.00000241, -0.00001084, 0.00000281, 0.00000158, -0.00000383, 0.00001654, },
		{ 0.00001644, -0.00000412, -0.00000396, 0.00000511, -0.00001039, 0.00000504, -0.00001538, -0.00001566, 0.00001141, -0.00001318, -0.00001065, -0.00002085, -0.00000830, -0.00010940, 0.00000766, 0.00013525, 0.00013908, 0.00008751, -0.00014371, -0.00001845, -0.00000772, 0.00003021, -0.00000407, 0.00003985, 0.00000005, -0.00001114, 0.00000580, -0.00003021, 0.00000525, -0.00001680, 0.00000490, 0.00000458, -0.00001666, -0.00001221, 0.00000223, 0.00001912, -0.00001204, -0.00000652, -0.00000124, 0.00001299, -0.00000571, -0.00000677, 0.00000594, -0.00000173, 0.00000032, -0.00000313, -0.00000286, -0.00000084, 0.00001695, -0.00001848, 0.00001288, -0.00001107, 0.00000864, -0.00001893, 0.00002377, -0.00001337, 0.00001138, -0.00001069, 0.00000437, -0.00001061, 0.00000742, -0.00000264, 0.00000381, -0.00000171, 0.00000215, -0.00000012, 0.00000088, -0.00000030, -0.00000994, 0.00000981, -0.00000404, 0.00000279, -0.00000530, 0.00000393, -0.00000578, 0.00000431, 0.00000488, -0.00000087, -0.00000377, 0.00000187, -0.00000258, 0.00000115, 0.00000123, -0.00000033, -0.00000146, 0.00000085, -0.00000022, -0.00000273, -0.00000007, -0.00000293, -0.00000130, 0.00000640, 0.00000064, -0.00000098, 0.00000077, 0.00000112, -0.00000265, -0.00000014, 0.00000461, -0.00000502, -0.00000690, 0.00000153, 0.00000434, 0.00000028, 0.00000025, -0.00000059, 0.00000549, 0.00000060, -0.00000478, -0.00000153, -0.00000203, 0.00000216, -0.00000320, 0.00000545, -0.00000017, -0.00000607, -0.00000281, -0.00000051, -0.00000096, 0.00000568, 0.00000092, -0.00000025, -0.00000129, 0.00000244, -0.00001024, 0.00000138, 0.00000309, -0.00000513, },
		{ -0.00000607, 0.00001593, -0.00000221, -0.00000542, 0.00000625, -0.00001112, 0.00000477, -0.00001281, -0.00001877, 0.00001302, -0.00001425, -0.00000869, -0.00002363, -0.00000398, -0.00010664, -0.00001280, 0.00014109, 0.00014845, 0.00009407, -0.00014558, -0.00002025, -0.00000960, 0.00002829, -0.00000342, 0.00003559, 0.00000314, -0.00001372, 0.00001082, -0.00003170, 0.00000439, -0.00001482, 0.00000161, 0.00000788, -0.00001800, -0.00001075, -0.00000059, 0.00002041, -0.00001066, -0.00000749, -0.00000183, 0.00001233, -0.00000368, -0.00000862, 0.00000726, -0.00000273, 0.00000145, -0.00000442, -0.00000073, -0.00000467, 0.00002021, -0.00002007, 0.00001422, -0.00001240, 0.00001034, -0.00002046, 0.00002354, -0.00001190, 0.00000995, -0.00000926, 0.00000372, -0.00000993, 0.00000569, -0.00000114, 0.00000259, -0.00000049, 0.00000082, 0.00000119, -0.00000037, 0.00000168, -0.00001165, 0.00001010, -0.00000382, 0.00000290, -0.00000547, 0.00000386, -0.00000535, 0.00000325, 0.00000519, -0.00000016, -0.00000415, 0.00000183, -0.00000237, 0.00000094, 0.00000133, -0.00000023, -0.00000155, 0.00000071, 0.00000005, -0.00000279, -0.00000022, -0.00000248, -0.00000196, 0.00000602, 0.00000139, -0.00000111, 0.00000072, 0.00000117, -0.00000228, -0.00000074, 0.00000467, -0.00000433, -0.00000719, 0.00000087, 0.00000461, 0.00000050, 0.00000017, -0.00000054, 0.00000497, 0.00000130, -0.00000490, -0.00000105, -0.00000271, 0.00000242, -0.00000325, 0.00000473, 0.00000108, -0.00000654, -0.00000303, -0.00000015, -0.00000140, 0.00000552, 0.00000103, 0.00000008, -0.00000138, 0.00000231, -0.00000948, -0.00000013, 0.00000446, },
		{ 0.00000564, -0.00000665, 0.00001504, -0.00000016, -0.00000683, 0.00000719, -0.00001141, 0.00000386, -0.00000976, -0.00002195, 0.00001451, -0.00001517, -0.00000703, -0.00002571, -0.00000120, -0.00010135, -0.00003413, 0.00014457, 0.00015965, 0.00010042, -0.00014707, -0.00002247, -0.00001104, 0.00002597, -0.00000229, 0.00003108, 0.00000623, -0.00001635, 0.00001523, -0.00003202, 0.00000243, -0.00001198, -0.00000217, 0.00001125, -0.00001923, -0.00000939, -0.00000302, 0.00002098, -0.00000869, -0.00000874, -0.00000208, 0.00001115, -0.00000127, -0.00001069, 0.00000874, -0.00000392, 0.00000284, -0.00000606, 0.00000179, -0.00000865, 0.00002325, -0.00002135, 0.00001530, -0.00001346, 0.00001166, -0.00002141, 0.00002264, -0.00000996, 0.00000826, -0.00000772, 0.00000296, -0.00000902, 0.00000377, 0.00000043, 0.00000132, 0.00000079, -0.00000054, 0.00000251, -0.00000162, 0.00000357, -0.00001306, 0.00001005, -0.00000341, 0.00000291, -0.00000552, 0.00000365, -0.00000479, 0.00000222, 0.00000537, 0.00000060, -0.00000444, 0.00000171, -0.00000211, 0.00000072, 0.00000143, -0.00000015, -0.00000159, 0.00000052, 0.00000032, -0.00000281, -0.00000039, -0.00000204, -0.00000249, 0.00000550, 0.00000212, -0.00000117, 0.00000063, 0.00000121, -0.00000190, -0.00000125, 0.00000460, -0.00000361, -0.00000740, 0.00000019, 0.00000478, 0.00000082, 0.00000004, -0.00000043, 0.00000441, 0.00000196, -0.00000500, -0.00000059, -0.00000325, 0.00000247, -0.00000307, 0.00000382, 0.00000229, -0.00000683, -0.00000340, 0.00000027, -0.00000183, 0.00000537, 0.00000110, 0.00000035, -0.00000133, 0.00000204, -0.00000858, -0.00000164, },
		{ -0.00000311, 0.00000657, -0.00000685, 0.00001382, 0.00000196, -0.00000810, 0.00000784, -0.00001117, 0.00000233, -0.00000633, -0.00002506, 0.00001574, -0.00001581, -0.00000585, -0.00002692, -0.00000004, -0.00009373, -0.00005567, 0.00014510, 0.00017278, 0.00010668, -0.00014828, -0.00002506, -0.00001212, 0.00002333, -0.00000082, 0.00002649, 0.00000913, -0.00001879, 0.00001878, -0.00003106, -0.00000057, -0.00000843, -0.00000624, 0.00001448, -0.00002019, -0.00000830, -0.00000489, 0.00002073, -0.00000612, -0.00001027, -0.00000195, 0.00000947, 0.00000149, -0.00001293, 0.00001035, -0.00000531, 0.00000448, -0.00000800, 0.00000459, -0.00001264, 0.00002593, -0.00002224, 0.00001606, -0.00001418, 0.00001254, -0.00002174, 0.00002111, -0.00000761, 0.00000637, -0.00000609, 0.00000209, -0.00000790, 0.00000170, 0.00000203, 0.00000004, 0.00000208, -0.00000189, 0.00000379, -0.00000283, 0.00000531, -0.00001413, 0.00000969, -0.00000283, 0.00000281, -0.00000543, 0.00000331, -0.00000415, 0.00000123, 0.00000539, 0.00000140, -0.00000465, 0.00000150, -0.00000182, 0.00000051, 0.00000152, -0.00000008, -0.00000159, 0.00000030, 0.00000058, -0.00000280, -0.00000059, -0.00000163, -0.00000291, 0.00000488, 0.00000282, -0.00000115, 0.00000052, 0.00000123, -0.00000152, -0.00000166, 0.00000440, -0.00000290, -0.00000751, -0.00000050, 0.00000484, 0.00000123, -0.00000014, -0.00000027, 0.00000382, 0.00000258, -0.00000506, -0.00000019, -0.00000363, 0.00000231, -0.00000269, 0.00000280, 0.00000342, -0.00000691, -0.00000394, 0.00000074, -0.00000224, 0.00000521, 0.00000116, 0.00000051, -0.00000113, 0.00000164, -0.00000759, },
		{ -0.00000655, -0.00000447, 0.00000723, -0.00000669, 0.00001235, 0.00000404, -0.00000914, 0.00000813, -0.00001039, 0.00000023, -0.00000268, -0.00002793, 0.00001655, -0.00001602, -0.00000529, -0.00002712, -0.00000052, -0.00008413, -0.00007674, 0.00014220, 0.00018782, 0.00011303, -0.00014934, -0.00002789, -0.00001295, 0.00002049, 0.00000086, 0.00002200, 0.00001167, -0.00002080, 0.00002128, -0.00002877, -0.00000453, -0.00000436, -0.00001038, 0.00001737, -0.00002070, -0.00000766, -0.00000606, 0.00001961, -0.00000297, -0.00001206, -0.00000144, 0.00000730, 0.00000456, -0.00001529, 0.00001208, -0.00000687, 0.00000633, -0.00001017, 0.00000757, -0.00001649, 0.00002814, -0.00002264, 0.00001642, -0.00001449, 0.00001291, -0.00002141, 0.00001896, -0.00000491, 0.00000436, -0.00000444, 0.00000116, -0.00000660, -0.00000048, 0.00000361, -0.00000122, 0.00000335, -0.00000320, 0.00000501, -0.00000394, 0.00000683, -0.00001485, 0.00000900, -0.00000210, 0.00000263, -0.00000524, 0.00000286, -0.00000346, 0.00000033, 0.00000527, 0.00000221, -0.00000476, 0.00000120, -0.00000150, 0.00000030, 0.00000160, -0.00000002, -0.00000155, 0.00000006, 0.00000081, -0.00000274, -0.00000081, -0.00000125, -0.00000320, 0.00000416, 0.00000345, -0.00000104, 0.00000037, 0.00000125, -0.00000116, -0.00000198, 0.00000410, -0.00000219, -0.00000753, -0.00000117, 0.00000478, 0.00000172, -0.00000035, -0.00000008, 0.00000323, 0.00000312, -0.00000506, 0.00000014, -0.00000382, 0.00000196, -0.00000215, 0.00000172, 0.00000440, -0.00000675, -0.00000462, 0.00000125, -0.00000261, 0.00000504, 0.00000123, 0.00000055, -0.00000077, 0.00000113, },
		{ 0.00000055, -0.00000550, -0.00000570, 0.00000758, -0.00000620, 0.00001070, 0.00000600, -0.00000988, 0.00000800, -0.00000906, -0.00000232, 0.00000102, -0.00003038, 0.00001683, -0.00001568, -0.00000547, -0.00002626, -0.00000255, -0.00007299, -0.00009662, 0.00013543, 0.00020464, 0.00011975, -0.00015042, -0.00003082, -0.00001367, 0.00001760, 0.00000258, 0.00001777, 0.00001367, -0.00002220, 0.00002258, -0.00002521, -0.00000924, -0.00000001, -0.00001432, 0.00001968, -0.00002059, -0.00000759, -0.00000643, 0.00001759, 0.00000072, -0.00001408, -0.00000058, 0.00000469, 0.00000786, -0.00001773, 0.00001388, -0.00000856, 0.00000834, -0.00001248, 0.00001060, -0.00002004, 0.00002975, -0.00002249, 0.00001634, -0.00001435, 0.00001274, -0.00002045, 0.00001628, -0.00000197, 0.00000230, -0.00000279, 0.00000018, -0.00000515, -0.00000271, 0.00000512, -0.00000242, 0.00000456, -0.00000443, 0.00000610, -0.00000491, 0.00000811, -0.00001518, 0.00000802, -0.00000127, 0.00000238, -0.00000496, 0.00000233, -0.00000276, -0.00000046, 0.00000499, 0.00000303, -0.00000476, 0.00000084, -0.00000116, 0.00000010, 0.00000168, 0.00000002, -0.00000148, -0.00000020, 0.00000100, -0.00000264, -0.00000103, -0.00000091, -0.00000339, 0.00000339, 0.00000400, -0.00000084, 0.00000020, 0.00000126, -0.00000082, -0.00000219, 0.00000371, -0.00000153, -0.00000746, -0.00000183, 0.00000460, 0.00000228, -0.00000056, 0.00000012, 0.00000267, 0.00000357, -0.00000498, 0.00000037, -0.00000383, 0.00000143, -0.00000151, 0.00000065, 0.00000518, -0.00000634, -0.00000544, 0.00000177, -0.00000292, 0.00000483, 0.00000135, 0.00000045, -0.00000028, },
		{ 0.00000035, -0.00000007, -0.00000448, -0.00000674, 0.00000762, -0.00000541, 0.00000895, 0.00000776, -0.00001025, 0.00000742, -0.00000722, -0.00000520, 0.00000460, -0.00003225, 0.00001644, -0.00001469, -0.00000645, -0.00002435, -0.00000594, -0.00006082, -0.00011457, 0.00012448, 0.00022298, 0.00012720, -0.00015175, -0.00003369, -0.00001444, 0.00001479, 0.00000417, 0.00001398, 0.00001498, -0.00002285, 0.00002264, -0.00002049, -0.00001447, 0.00000432, -0.00001779, 0.00002122, -0.00001971, -0.00000823, -0.00000591, 0.00001468, 0.00000488, -0.00001628, 0.00000061, 0.00000170, 0.00001131, -0.00002018, 0.00001569, -0.00001032, 0.00001043, -0.00001484, 0.00001352, -0.00002314, 0.00003065, -0.00002175, 0.00001579, -0.00001373, 0.00001203, -0.00001888, 0.00001315, 0.00000112, 0.00000027, -0.00000121, -0.00000083, -0.00000359, -0.00000493, 0.00000652, -0.00000352, 0.00000569, -0.00000553, 0.00000704, -0.00000570, 0.00000910, -0.00001513, 0.00000677, -0.00000036, 0.00000211, -0.00000460, 0.00000176, -0.00000207, -0.00000113, 0.00000458, 0.00000381, -0.00000464, 0.00000040, -0.00000083, -0.00000007, 0.00000175, 0.00000005, -0.00000139, -0.00000045, 0.00000116, -0.00000249, -0.00000127, -0.00000060, -0.00000349, 0.00000259, 0.00000445, -0.00000054, 0.00000001, 0.00000126, -0.00000050, -0.00000232, 0.00000326, -0.00000093, -0.00000730, -0.00000243, 0.00000430, 0.00000289, -0.00000075, 0.00000030, 0.00000216, 0.00000391, -0.00000483, 0.00000050, -0.00000366, 0.00000076, -0.00000080, -0.00000034, 0.00000572, -0.00000567, -0.00000635, 0.00000225, -0.00000315, 0.00000456, 0.00000155, 0.00000019, },
		{ -0.00000023, 0.00000106, -0.00000069, -0.00000354, -0.00000756, 0.00000736, -0.00000439, 0.00000719, 0.00000922, -0.00001021, 0.00000639, -0.00000494, -0.00000825, 0.00000784, -0.00003339, 0.00001530, -0.00001301, -0.00000822, -0.00002148, -0.00001042, -0.00004821, -0.00012992, 0.00010915, 0.00024244, 0.00013583, -0.00015360, -0.00003630, -0.00001542, 0.00001221, 0.00000549, 0.00001079, 0.00001549, -0.00002264, 0.00002148, -0.00001483, -0.00001991, 0.00000832, -0.00002052, 0.00002177, -0.00001794, -0.00000966, -0.00000449, 0.00001094, 0.00000941, -0.00001858, 0.00000208, -0.00000159, 0.00001483, -0.00002256, 0.00001748, -0.00001209, 0.00001252, -0.00001712, 0.00001621, -0.00002566, 0.00003077, -0.00002038, 0.00001476, -0.00001263, 0.00001078, -0.00001676, 0.00000966, 0.00000425, -0.00000164, 0.00000027, -0.00000183, -0.00000196, -0.00000711, 0.00000776, -0.00000449, 0.00000669, -0.00000649, 0.00000779, -0.00000627, 0.00000979, -0.00001471, 0.00000531, 0.00000057, 0.00000182, -0.00000420, 0.00000116, -0.00000142, -0.00000166, 0.00000405, 0.00000454, -0.00000441, -0.00000008, -0.00000051, -0.00000022, 0.00000181, 0.00000006, -0.00000128, -0.00000069, 0.00000126, -0.00000231, -0.00000150, -0.00000034, -0.00000349, 0.00000178, 0.00000477, -0.00000016, -0.00000019, 0.00000124, -0.00000021, -0.00000235, 0.00000277, -0.00000040, -0.00000705, -0.00000299, 0.00000388, 0.00000352, -0.00000090, 0.00000045, 0.00000171, 0.00000413, -0.00000458, 0.00000052, -0.00000332, -0.00000000, -0.00000010, -0.00000121, 0.00000599, -0.00000478, -0.00000731, 0.00000267, -0.00000328, 0.00000421, 0.00000184, },
		{ 0.00000224, -0.00000081, 0.00000184, -0.00000126, -0.00000272, -0.00000815, 0.00000681, -0.00000320, 0.00000551, 0.00001033, -0.00000974, 0.00000494, -0.00000234, -0.00001129, 0.00001057, -0.00003368, 0.00001338, -0.00001066, -0.00001073, -0.00001781, -0.00001564, -0.00003574, -0.00014207, 0.00008944, 0.00026246, 0.00014615, -0.00015623, -0.00003848, -0.00001676, 0.00000999, 0.00000640, 0.00000830, 0.00001513, -0.00002156, 0.00001921, -0.00000850, -0.00002522, 0.00001167, -0.00002225, 0.00002119, -0.00001519, -0.00001191, -0.00000218, 0.00000647, 0.00001419, -0.00002090, 0.00000376, -0.00000511, 0.00001833, -0.00002480, 0.00001916, -0.00001381, 0.00001451, -0.00001921, 0.00001851, -0.00002746, 0.00003006, -0.00001839, 0.00001327, -0.00001107, 0.00000906, -0.00001418, 0.00000593, 0.00000730, -0.00000338, 0.00000161, -0.00000281, -0.00000030, -0.00000918, 0.00000880, -0.00000530, 0.00000754, -0.00000726, 0.00000833, -0.00000662, 0.00001017, -0.00001395, 0.00000369, 0.00000150, 0.00000154, -0.00000378, 0.00000057, -0.00000084, -0.00000204, 0.00000341, 0.00000520, -0.00000405, -0.00000061, -0.00000021, -0.00000034, 0.00000187, 0.00000007, -0.00000115, -0.00000091, 0.00000131, -0.00000209, -0.00000173, -0.00000012, -0.00000343, 0.00000099, 0.00000497, 0.00000029, -0.00000038, 0.00000120, 0.00000005, -0.00000231, 0.00000226, 0.00000005, -0.00000675, -0.00000347, 0.00000335, 0.00000415, -0.00000099, 0.00000054, 0.00000134, 0.00000424, -0.00000425, 0.00000044, -0.00000284, -0.00000081, 0.00000054, -0.00000191, 0.00000599, -0.00000369, -0.00000829, 0.00000298, -0.00000327, 0.00000377, },
		{ 0.00000323, 0.00000277, -0.00000152, 0.00000263, -0.00000174, -0.00000203, -0.00000850, 0.00000600, -0.00000190, 0.00000399, 0.00001105, -0.00000885, 0.00000314, 0.00000046, -0.00001414, 0.00001261, -0.00003304, 0.00001067, -0.00000768, -0.00001386, -0.00001357, -0.00002122, -0.00002404, -0.00015050, 0.00006549, 0.00028236, 0.00015871, -0.00015992, -0.00004006, -0.00001857, 0.00000824, 0.00000681, 0.00000658, 0.00001389, -0.00001968, 0.00001603, -0.00000184, -0.00003002, 0.00001408, -0.00002277, 0.00001937, -0.00001145, -0.00001497, 0.00000096, 0.00000141, 0.00001906, -0.00002315, 0.00000559, -0.00000875, 0.00002168, -0.00002682, 0.00002068, -0.00001538, 0.00001630, -0.00002099, 0.00002031, -0.00002847, 0.00002848, -0.00001581, 0.00001134, -0.00000908, 0.00000691, -0.00001123, 0.00000209, 0.00001018, -0.00000487, 0.00000279, -0.00000374, 0.00000136, -0.00001109, 0.00000962, -0.00000594, 0.00000821, -0.00000782, 0.00000864, -0.00000672, 0.00001024, -0.00001288, 0.00000195, 0.00000236, 0.00000131, -0.00000336, 0.00000001, -0.00000035, -0.00000227, 0.00000269, 0.00000576, -0.00000357, -0.00000115, 0.00000006, -0.00000043, 0.00000191, 0.00000008, -0.00000103, -0.00000110, 0.00000131, -0.00000184, -0.00000194, 0.00000006, -0.00000330, 0.00000022, 0.00000505, 0.00000080, -0.00000056, 0.00000115, 0.00000028, -0.00000220, 0.00000174, 0.00000041, -0.00000639, -0.00000386, 0.00000273, 0.00000475, -0.00000100, 0.00000058, 0.00000105, 0.00000423, -0.00000384, 0.00000027, -0.00000226, -0.00000161, 0.00000107, -0.00000241, 0.00000572, -0.00000244, -0.00000921, 0.00000314, -0.00000312, },
		{ -0.00000281, 0.00000257, 0.00000341, -0.00000233, 0.00000338, -0.00000209, -0.00000152, -0.00000860, 0.00000498, -0.00000056, 0.00000269, 0.00001134, -0.00000757, 0.00000107, 0.00000330, -0.00001660, 0.00001383, -0.00003142, 0.00000724, -0.00000419, -0.00001743, -0.00000904, -0.00002672, -0.00001365, -0.00015486, 0.00003766, 0.00030132, 0.00017411, -0.00016491, -0.00004093, -0.00002095, 0.00000699, 0.00000667, 0.00000563, 0.00001184, -0.00001711, 0.00001220, 0.00000476, -0.00003395, 0.00001524, -0.00002192, 0.00001624, -0.00000672, -0.00001878, 0.00000482, -0.00000409, 0.00002386, -0.00002522, 0.00000748, -0.00001240, 0.00002480, -0.00002854, 0.00002196, -0.00001673, 0.00001779, -0.00002234, 0.00002148, -0.00002860, 0.00002606, -0.00001269, 0.00000904, -0.00000674, 0.00000441, -0.00000802, -0.00000174, 0.00001278, -0.00000607, 0.00000378, -0.00000461, 0.00000297, -0.00001280, 0.00001017, -0.00000637, 0.00000870, -0.00000817, 0.00000871, -0.00000657, 0.00001003, -0.00001156, 0.00000018, 0.00000312, 0.00000113, -0.00000296, -0.00000050, 0.00000004, -0.00000236, 0.00000192, 0.00000621, -0.00000298, -0.00000170, 0.00000027, -0.00000049, 0.00000195, 0.00000009, -0.00000091, -0.00000125, 0.00000125, -0.00000156, -0.00000215, 0.00000021, -0.00000312, -0.00000050, 0.00000499, 0.00000136, -0.00000070, 0.00000107, 0.00000049, -0.00000204, 0.00000124, 0.00000067, -0.00000599, -0.00000417, 0.00000202, 0.00000530, -0.00000091, 0.00000055, 0.00000085, 0.00000412, -0.00000337, 0.00000002, -0.00000161, -0.00000235, 0.00000146, -0.00000268, 0.00000519, -0.00000110, -0.00001002, 0.00000311, },
		{ 0.00000288, -0.00000233, 0.00000180, 0.00000415, -0.00000321, 0.00000405, -0.00000229, -0.00000118, -0.00000848, 0.00000379, 0.00000073, 0.00000168, 0.00001121, -0.00000596, -0.00000117, 0.00000604, -0.00001853, 0.00001414, -0.00002886, 0.00000319, -0.00000035, -0.00002123, -0.00000451, -0.00003174, -0.00000508, -0.00015493, 0.00000649, 0.00031840, 0.00019292, -0.00017139, -0.00004100, -0.00002392, 0.00000627, 0.00000601, 0.00000540, 0.00000912, -0.00001408, 0.00000805, 0.00001089, -0.00003665, 0.00001494, -0.00001957, 0.00001181, -0.00000109, -0.00002323, 0.00000925, -0.00000983, 0.00002843, -0.00002701, 0.00000935, -0.00001596, 0.00002757, -0.00002988, 0.00002293, -0.00001777, 0.00001888, -0.00002317, 0.00002194, -0.00002782, 0.00002283, -0.00000911, 0.00000645, -0.00000412, 0.00000167, -0.00000469, -0.00000542, 0.00001501, -0.00000694, 0.00000459, -0.00000540, 0.00000450, -0.00001427, 0.00001045, -0.00000661, 0.00000898, -0.00000828, 0.00000854, -0.00000620, 0.00000956, -0.00001004, -0.00000157, 0.00000374, 0.00000104, -0.00000259, -0.00000095, 0.00000032, -0.00000233, 0.00000112, 0.00000654, -0.00000230, -0.00000223, 0.00000043, -0.00000052, 0.00000198, 0.00000011, -0.00000081, -0.00000136, 0.00000114, -0.00000126, -0.00000233, 0.00000031, -0.00000291, -0.00000116, 0.00000482, 0.00000193, -0.00000079, 0.00000096, 0.00000067, -0.00000183, 0.00000078, 0.00000084, -0.00000557, -0.00000437, 0.00000126, 0.00000578, -0.00000071, 0.00000045, 0.00000073, 0.00000391, -0.00000286, -0.00000028, -0.00000094, -0.00000299, 0.00000167, -0.00000274, 0.00000446, 0.00000027, -0.00001067, },
		{ -0.00001110, 0.00000242, -0.00000170, 0.00000094, 0.00000498, -0.00000411, 0.00000459, -0.00000228, -0.00000104, -0.00000814, 0.00000248, 0.00000191, 0.00000099, 0.00001067, -0.00000409, -0.00000345, 0.00000852, -0.00001976, 0.00001347, -0.00002542, -0.00000131, 0.00000364, -0.00002501, -0.00000032, -0.00003585, 0.00000126, -0.00015070, -0.00002730, 0.00033262, 0.00021567, -0.00017948, -0.00004029, -0.00002743, 0.00000599, 0.00000491, 0.00000576, 0.00000590, -0.00001085, 0.00000395, 0.00001614, -0.00003782, 0.00001300, -0.00001570, 0.00000617, 0.00000531, -0.00002817, 0.00001406, -0.00001560, 0.00003260, -0.00002843, 0.00001111, -0.00001932, 0.00002990, -0.00003076, 0.00002352, -0.00001842, 0.00001949, -0.00002338, 0.00002164, -0.00002615, 0.00001887, -0.00000519, 0.00000366, -0.00000131, -0.00000121, -0.00000134, -0.00000885, 0.00001681, -0.00000747, 0.00000521, -0.00000611, 0.00000592, -0.00001547, 0.00001044, -0.00000663, 0.00000906, -0.00000818, 0.00000814, -0.00000561, 0.00000887, -0.00000838, -0.00000325, 0.00000419, 0.00000104, -0.00000226, -0.00000131, 0.00000048, -0.00000217, 0.00000032, 0.00000672, -0.00000153, -0.00000272, 0.00000053, -0.00000051, 0.00000199, 0.00000014, -0.00000072, -0.00000143, 0.00000097, -0.00000096, -0.00000249, 0.00000038, -0.00000266, -0.00000175, 0.00000454, 0.00000251, -0.00000083, 0.00000083, 0.00000083, -0.00000158, 0.00000036, 0.00000093, -0.00000515, -0.00000448, 0.00000045, 0.00000615, -0.00000040, 0.00000029, 0.00000068, 0.00000363, -0.00000232, -0.00000060, -0.00000027, -0.00000349, 0.00000168, -0.00000259, 0.00000357, 0.00000162, },
		{ 0.00000286, -0.00001128, 0.00000173, -0.00000091, 0.00000000, 0.00000586, -0.00000498, 0.00000493, -0.00000207, -0.00000107, -0.00000762, 0.00000111, 0.00000293, 0.00000065, 0.00000976, -0.00000206, -0.00000565, 0.00001059, -0.00002021, 0.00001184, -0.00002123, -0.00000606, 0.00000756, -0.00002850, 0.00000323, -0.00003873, 0.00000508, -0.00014231, -0.00006281, 0.00034294, 0.00024284, -0.00018917, -0.00003891, -0.00003137, 0.00000602, 0.00000349, 0.00000652, 0.00000246, -0.00000773, 0.00000029, 0.00002013, -0.00003719, 0.00000933, -0.00001034, -0.00000056, 0.00001230, -0.00003342, 0.00001905, -0.00002118, 0.00003621, -0.00002939, 0.00001269, -0.00002237, 0.00003169, -0.00003113, 0.00002367, -0.00001861, 0.00001953, -0.00002292, 0.00002055, -0.00002361, 0.00001430, -0.00000105, 0.00000078, 0.00000159, -0.00000412, 0.00000190, -0.00001192, 0.00001813, -0.00000764, 0.00000564, -0.00000673, 0.00000719, -0.00001637, 0.00001013, -0.00000646, 0.00000894, -0.00000785, 0.00000754, -0.00000485, 0.00000801, -0.00000665, -0.00000478, 0.00000443, 0.00000113, -0.00000199, -0.00000159, 0.00000054, -0.00000193, -0.00000046, 0.00000676, -0.00000069, -0.00000317, 0.00000056, -0.00000047, 0.00000200, 0.00000018, -0.00000066, -0.00000144, 0.00000077, -0.00000066, -0.00000262, 0.00000042, -0.00000240, -0.00000228, 0.00000417, 0.00000306, -0.00000081, 0.00000067, 0.00000096, -0.00000132, -0.00000001, 0.00000094, -0.00000474, -0.00000448, -0.00000037, 0.00000640, 0.00000003, 0.00000007, 0.00000069, 0.00000329, -0.00000178, -0.00000092, 0.00000034, -0.00000382, 0.00000150, -0.00000227, 0.00000257, },
		{ 0.00000152, 0.00000394, -0.00001116, 0.00000080, -0.00000001, -0.00000098, 0.00000675, -0.00000576, 0.00000505, -0.00000163, -0.00000126, -0.00000696, -0.00000026, 0.00000374, 0.00000066, 0.00000855, 0.00000003, -0.00000765, 0.00001213, -0.00001981, 0.00000930, -0.00001648, -0.00001083, 0.00001114, -0.00003144, 0.00000584, -0.00004007, 0.00000624, -0.00013013, -0.00009902, 0.00034832, 0.00027475, -0.00020035, -0.00003704, -0.00003553, 0.00000618, 0.00000198, 0.00000744, -0.00000091, -0.00000507, -0.00000254, 0.00002250, -0.00003460, 0.00000393, -0.00000363, -0.00000817, 0.00001964, -0.00003873, 0.00002399, -0.00002635, 0.00003909, -0.00002981, 0.00001402, -0.00002502, 0.00003287, -0.00003092, 0.00002334, -0.00001828, 0.00001896, -0.00002176, 0.00001867, -0.00002028, 0.00000926, 0.00000317, -0.00000207, 0.00000447, -0.00000695, 0.00000492, -0.00001455, 0.00001892, -0.00000746, 0.00000591, -0.00000725, 0.00000830, -0.00001695, 0.00000955, -0.00000610, 0.00000863, -0.00000733, 0.00000676, -0.00000394, 0.00000702, -0.00000491, -0.00000613, 0.00000447, 0.00000131, -0.00000177, -0.00000178, 0.00000050, -0.00000161, -0.00000118, 0.00000665, 0.00000020, -0.00000354, 0.00000053, -0.00000041, 0.00000199, 0.00000024, -0.00000063, -0.00000141, 0.00000053, -0.00000036, -0.00000271, 0.00000042, -0.00000212, -0.00000273, 0.00000372, 0.00000357, -0.00000071, 0.00000049, 0.00000106, -0.00000103, -0.00000033, 0.00000088, -0.00000437, -0.00000440, -0.00000118, 0.00000652, 0.00000057, -0.00000019, 0.00000074, 0.00000292, -0.00000126, -0.00000122, 0.00000088, -0.00000395, 0.00000114, -0.00000180, },
		{ -0.00000125, 0.00000050, 0.00000479, -0.00001074, -0.00000033, 0.00000097, -0.00000196, 0.00000760, -0.00000641, 0.00000490, -0.00000098, -0.00000159, -0.00000619, -0.00000158, 0.00000429, 0.00000100, 0.00000711, 0.00000208, -0.00000932, 0.00001304, -0.00001854, 0.00000597, -0.00001139, -0.00001535, 0.00001416, -0.00003359, 0.00000729, -0.00003970, 0.00000474, -0.00011469, -0.00013478, 0.00034778, 0.00031161, -0.00021274, -0.00003501, -0.00003965, 0.00000618, 0.00000063, 0.00000823, -0.00000387, -0.00000322, -0.00000418, 0.00002300, -0.00002995, -0.00000312, 0.00000424, -0.00001638, 0.00002707, -0.00004389, 0.00002865, -0.00003091, 0.00004114, -0.00002964, 0.00001502, -0.00002717, 0.00003336, -0.00003011, 0.00002248, -0.00001739, 0.00001774, -0.00001989, 0.00001604, -0.00001627, 0.00000389, 0.00000731, -0.00000479, 0.00000724, -0.00000959, 0.00000761, -0.00001666, 0.00001918, -0.00000697, 0.00000603, -0.00000767, 0.00000922, -0.00001722, 0.00000870, -0.00000557, 0.00000816, -0.00000663, 0.00000583, -0.00000293, 0.00000598, -0.00000322, -0.00000725, 0.00000428, 0.00000159, -0.00000160, -0.00000190, 0.00000038, -0.00000125, -0.00000182, 0.00000640, 0.00000111, -0.00000383, 0.00000042, -0.00000033, 0.00000197, 0.00000031, -0.00000063, -0.00000134, 0.00000026, -0.00000009, -0.00000277, 0.00000039, -0.00000182, -0.00000312, 0.00000321, 0.00000403, -0.00000055, 0.00000030, 0.00000113, -0.00000073, -0.00000058, 0.00000077, -0.00000403, -0.00000423, -0.00000196, 0.00000648, 0.00000121, -0.00000047, 0.00000081, 0.00000253, -0.00000078, -0.00000148, 0.00000131, -0.00000389, 0.00000062, },
		{ -0.00000002, -0.00000067, -0.00000043, 0.00000539, -0.00001002, -0.00000164, 0.00000199, -0.00000291, 0.00000838, -0.00000686, 0.00000446, -0.00000013, -0.00000202, -0.00000537, -0.00000280, 0.00000457, 0.00000165, 0.00000551, 0.00000397, -0.00001057, 0.00001328, -0.00001644, 0.00000201, -0.00000621, -0.00001936, 0.00001638, -0.00003475, 0.00000741, -0.00003754, 0.00000073, -0.00009666, -0.00016889, 0.00034044, 0.00035343, -0.00022590, -0.00003324, -0.00004335, 0.00000572, -0.00000027, 0.00000858, -0.00000609, -0.00000252, -0.00000431, 0.00002142, -0.00002328, -0.00001163, 0.00001299, -0.00002490, 0.00003429, -0.00004862, 0.00003279, -0.00003467, 0.00004224, -0.00002884, 0.00001564, -0.00002876, 0.00003312, -0.00002868, 0.00002109, -0.00001592, 0.00001587, -0.00001735, 0.00001275, -0.00001173, -0.00000162, 0.00001124, -0.00000727, 0.00000979, -0.00001197, 0.00000991, -0.00001821, 0.00001892, -0.00000621, 0.00000605, -0.00000800, 0.00000994, -0.00001716, 0.00000760, -0.00000491, 0.00000754, -0.00000579, 0.00000479, -0.00000188, 0.00000492, -0.00000163, -0.00000810, 0.00000387, 0.00000193, -0.00000147, -0.00000194, 0.00000019, -0.00000087, -0.00000237, 0.00000600, 0.00000203, -0.00000403, 0.00000025, -0.00000023, 0.00000194, 0.00000042, -0.00000065, -0.00000123, -0.00000002, 0.00000015, -0.00000279, 0.00000034, -0.00000152, -0.00000344, 0.00000265, 0.00000442, -0.00000031, 0.00000011, 0.00000118, -0.00000043, -0.00000077, 0.00000062, -0.00000375, -0.00000398, -0.00000268, 0.00000629, 0.00000193, -0.00000075, 0.00000089, 0.00000216, -0.00000035, -0.00000167, 0.00000162, -0.00000364, },
		{ -0.00000321, -0.00000074, -0.00000011, -0.00000122, 0.00000569, -0.00000900, -0.00000308, 0.00000300, -0.00000378, 0.00000903, -0.00000708, 0.00000373, 0.00000090, -0.00000252, -0.00000454, -0.00000389, 0.00000457, 0.00000257, 0.00000386, 0.00000560, -0.00001133, 0.00001281, -0.00001360, -0.00000237, -0.00000122, -0.00002260, 0.00001759, -0.00003477, 0.00000611, -0.00003362, -0.00000550, -0.00007689, -0.00020013, 0.00032556, 0.00039999, -0.00023924, -0.00003224, -0.00004623, 0.00000444, -0.00000040, 0.00000817, -0.00000723, -0.00000327, -0.00000269, 0.00001769, -0.00001470, -0.00002133, 0.00002229, -0.00003337, 0.00004100, -0.00005270, 0.00003622, -0.00003746, 0.00004232, -0.00002739, 0.00001586, -0.00002973, 0.00003212, -0.00002661, 0.00001915, -0.00001388, 0.00001337, -0.00001418, 0.00000890, -0.00000681, -0.00000708, 0.00001480, -0.00000940, 0.00001205, -0.00001399, 0.00001174, -0.00001919, 0.00001817, -0.00000521, 0.00000598, -0.00000823, 0.00001045, -0.00001679, 0.00000629, -0.00000414, 0.00000680, -0.00000484, 0.00000369, -0.00000081, 0.00000389, -0.00000020, -0.00000868, 0.00000326, 0.00000233, -0.00000137, -0.00000191, -0.00000005, -0.00000050, -0.00000281, 0.00000548, 0.00000292, -0.00000412, 0.00000001, -0.00000012, 0.00000191, 0.00000054, -0.00000070, -0.00000109, -0.00000030, 0.00000036, -0.00000277, 0.00000025, -0.00000122, -0.00000370, 0.00000207, 0.00000473, -0.00000002, -0.00000008, 0.00000119, -0.00000013, -0.00000091, 0.00000044, -0.00000353, -0.00000369, -0.00000332, 0.00000594, 0.00000272, -0.00000100, 0.00000095, 0.00000181, 0.00000000, -0.00000180, 0.00000179, },
		{ 0.00000184, -0.00000264, -0.00000148, 0.00000036, -0.00000182, 0.00000568, -0.00000772, -0.00000458, 0.00000394, -0.00000452, 0.00000952, -0.00000703, 0.00000270, 0.00000207, -0.00000303, -0.00000375, -0.00000479, 0.00000428, 0.00000369, 0.00000224, 0.00000690, -0.00001155, 0.00001165, -0.00001014, -0.00000693, 0.00000331, -0.00002486, 0.00001767, -0.00003357, 0.00000339, -0.00002810, -0.00001350, -0.00005629, -0.00022730, 0.00030259, 0.00045085, -0.00025200, -0.00003263, -0.00004782, 0.00000197, 0.00000057, 0.00000668, -0.00000700, -0.00000572, 0.00000082, 0.00001184, -0.00000447, -0.00003187, 0.00003176, -0.00004144, 0.00004691, -0.00005589, 0.00003875, -0.00003917, 0.00004136, -0.00002530, 0.00001565, -0.00003003, 0.00003036, -0.00002393, 0.00001669, -0.00001128, 0.00001030, -0.00001048, 0.00000463, -0.00000169, -0.00001232, 0.00001785, -0.00001111, 0.00001394, -0.00001559, 0.00001307, -0.00001959, 0.00001697, -0.00000405, 0.00000586, -0.00000838, 0.00001075, -0.00001612, 0.00000482, -0.00000330, 0.00000598, -0.00000382, 0.00000257, 0.00000021, 0.00000295, 0.00000104, -0.00000897, 0.00000247, 0.00000276, -0.00000129, -0.00000184, -0.00000032, -0.00000015, -0.00000313, 0.00000485, 0.00000376, -0.00000409, -0.00000027, -0.00000001, 0.00000186, 0.00000069, -0.00000078, -0.00000093, -0.00000057, 0.00000052, -0.00000271, 0.00000014, -0.00000091, -0.00000390, 0.00000147, 0.00000495, 0.00000033, -0.00000025, 0.00000116, 0.00000016, -0.00000099, 0.00000026, -0.00000337, -0.00000335, -0.00000386, 0.00000544, 0.00000354, -0.00000121, 0.00000098, 0.00000150, 0.00000028, -0.00000184, },
		{ -0.00000180, 0.00000177, -0.00000195, -0.00000220, 0.00000070, -0.00000220, 0.00000537, -0.00000623, -0.00000609, 0.00000474, -0.00000509, 0.00000980, -0.00000670, 0.00000140, 0.00000331, -0.00000350, -0.00000303, -0.00000550, 0.00000372, 0.00000496, 0.00000074, 0.00000779, -0.00001121, 0.00000988, -0.00000625, -0.00001141, 0.00000711, -0.00002593, 0.00001652, -0.00003115, -0.00000063, -0.00002124, -0.00002272, -0.00003587, -0.00024931, 0.00027123, 0.00050534, -0.00026325, -0.00003508, -0.00004764, -0.00000207, 0.00000293, 0.00000385, -0.00000516, -0.00001002, 0.00000628, 0.00000400, 0.00000710, -0.00004283, 0.00004096, -0.00004873, 0.00005175, -0.00005800, 0.00004024, -0.00003972, 0.00003934, -0.00002261, 0.00001500, -0.00002965, 0.00002783, -0.00002067, 0.00001374, -0.00000817, 0.00000672, -0.00000637, 0.00000009, 0.00000343, -0.00001715, 0.00002029, -0.00001233, 0.00001543, -0.00001675, 0.00001388, -0.00001943, 0.00001540, -0.00000278, 0.00000573, -0.00000843, 0.00001083, -0.00001519, 0.00000321, -0.00000242, 0.00000511, -0.00000276, 0.00000147, 0.00000114, 0.00000213, 0.00000207, -0.00000897, 0.00000153, 0.00000318, -0.00000121, -0.00000173, -0.00000060, 0.00000016, -0.00000332, 0.00000413, 0.00000454, -0.00000395, -0.00000060, 0.00000010, 0.00000180, 0.00000086, -0.00000087, -0.00000076, -0.00000082, 0.00000063, -0.00000260, 0.00000001, -0.00000060, -0.00000405, 0.00000087, 0.00000508, 0.00000072, -0.00000039, 0.00000109, 0.00000045, -0.00000103, 0.00000007, -0.00000327, -0.00000300, -0.00000428, 0.00000480, 0.00000438, -0.00000135, 0.00000097, 0.00000124, 0.00000047, },
		{ 0.00000057, -0.00000169, 0.00000159, -0.00000119, -0.00000282, 0.00000086, -0.00000234, 0.00000478, -0.00000458, -0.00000753, 0.00000536, -0.00000545, 0.00000985, -0.00000607, -0.00000015, 0.00000458, -0.00000388, -0.00000242, -0.00000598, 0.00000291, 0.00000631, -0.00000057, 0.00000822, -0.00001033, 0.00000759, -0.00000211, -0.00001556, 0.00000994, -0.00002571, 0.00001414, -0.00002760, -0.00000577, -0.00001342, -0.00003251, -0.00001662, -0.00026518, 0.00023142, 0.00056250, -0.00027196, -0.00004029, -0.00004521, -0.00000803, 0.00000696, -0.00000055, -0.00000155, -0.00001624, 0.00001361, -0.00000556, 0.00001957, -0.00005373, 0.00004949, -0.00005491, 0.00005526, -0.00005887, 0.00004058, -0.00003906, 0.00003632, -0.00001937, 0.00001393, -0.00002859, 0.00002460, -0.00001691, 0.00001037, -0.00000462, 0.00000274, -0.00000199, -0.00000455, 0.00000838, -0.00002141, 0.00002203, -0.00001302, 0.00001648, -0.00001743, 0.00001418, -0.00001878, 0.00001353, -0.00000148, 0.00000560, -0.00000841, 0.00001070, -0.00001403, 0.00000153, -0.00000155, 0.00000422, -0.00000171, 0.00000044, 0.00000196, 0.00000147, 0.00000286, -0.00000870, 0.00000048, 0.00000358, -0.00000112, -0.00000160, -0.00000086, 0.00000040, -0.00000340, 0.00000334, 0.00000523, -0.00000369, -0.00000096, 0.00000020, 0.00000173, 0.00000106, -0.00000096, -0.00000060, -0.00000104, 0.00000068, -0.00000246, -0.00000014, -0.00000030, -0.00000415, 0.00000028, 0.00000513, 0.00000113, -0.00000050, 0.00000099, 0.00000073, -0.00000102, -0.00000010, -0.00000323, -0.00000264, -0.00000458, 0.00000403, 0.00000519, -0.00000141, 0.00000090, 0.00000104, },
		{ 0.00000090, 0.00000060, -0.00000152, 0.00000133, -0.00000040, -0.00000331, 0.00000082, -0.00000224, 0.00000394, -0.00000286, -0.00000883, 0.00000574, -0.00000557, 0.00000965, -0.00000515, -0.00000188, 0.00000580, -0.00000413, -0.00000196, -0.00000624, 0.00000189, 0.00000766, -0.00000161, 0.00000819, -0.00000896, 0.00000491, 0.00000205, -0.00001911, 0.00001163, -0.00002412, 0.00001059, -0.00002308, -0.00001174, -0.00000508, -0.00004218, 0.00000047, -0.00027414, 0.00018342, 0.00062117, -0.00027697, -0.00004900, -0.00004010, -0.00001618, 0.00001285, -0.00000667, 0.00000391, -0.00002433, 0.00002264, -0.00001648, 0.00003244, -0.00006409, 0.00005691, -0.00005966, 0.00005726, -0.00005840, 0.00003972, -0.00003721, 0.00003237, -0.00001566, 0.00001246, -0.00002686, 0.00002070, -0.00001271, 0.00000665, -0.00000072, -0.00000153, 0.00000252, -0.00000911, 0.00001297, -0.00002496, 0.00002299, -0.00001317, 0.00001710, -0.00001765, 0.00001400, -0.00001767, 0.00001144, -0.00000021, 0.00000552, -0.00000830, 0.00001038, -0.00001266, -0.00000019, -0.00000073, 0.00000335, -0.00000070, -0.00000049, 0.00000262, 0.00000098, 0.00000342, -0.00000819, -0.00000065, 0.00000393, -0.00000101, -0.00000147, -0.00000111, 0.00000057, -0.00000336, 0.00000251, 0.00000580, -0.00000332, -0.00000135, 0.00000028, 0.00000166, 0.00000126, -0.00000106, -0.00000044, -0.00000122, 0.00000068, -0.00000228, -0.00000030, 0.00000000, -0.00000420, -0.00000029, 0.00000510, 0.00000155, -0.00000055, 0.00000085, 0.00000099, -0.00000097, -0.00000026, -0.00000324, -0.00000231, -0.00000474, 0.00000315, 0.00000596, -0.00000135, 0.00000077, },
		{ 0.00000058, 0.00000080, 0.00000057, -0.00000130, 0.00000101, 0.00000036, -0.00000362, 0.00000055, -0.00000192, 0.00000292, -0.00000114, -0.00000991, 0.00000584, -0.00000544, 0.00000920, -0.00000397, -0.00000375, 0.00000691, -0.00000419, -0.00000166, -0.00000627, 0.00000070, 0.00000897, -0.00000234, 0.00000770, -0.00000717, 0.00000199, 0.00000602, -0.00002186, 0.00001204, -0.00002119, 0.00000603, -0.00001785, -0.00001818, 0.00000326, -0.00005101, 0.00001451, -0.00027562, 0.00012777, 0.00067996, -0.00027707, -0.00006187, -0.00003194, -0.00002673, 0.00002070, -0.00001454, 0.00001118, -0.00003415, 0.00003306, -0.00002829, 0.00004517, -0.00007341, 0.00006284, -0.00006271, 0.00005760, -0.00005654, 0.00003766, -0.00003422, 0.00002759, -0.00001157, 0.00001063, -0.00002450, 0.00001623, -0.00000818, 0.00000269, 0.00000341, -0.00000593, 0.00000699, -0.00001343, 0.00001703, -0.00002770, 0.00002316, -0.00001280, 0.00001728, -0.00001743, 0.00001337, -0.00001620, 0.00000922, 0.00000098, 0.00000549, -0.00000811, 0.00000987, -0.00001114, -0.00000188, 0.00000001, 0.00000252, 0.00000023, -0.00000129, 0.00000311, 0.00000067, 0.00000375, -0.00000747, -0.00000180, 0.00000418, -0.00000086, -0.00000134, -0.00000131, 0.00000066, -0.00000322, 0.00000166, 0.00000626, -0.00000285, -0.00000173, 0.00000035, 0.00000157, 0.00000148, -0.00000115, -0.00000030, -0.00000135, 0.00000062, -0.00000208, -0.00000048, 0.00000029, -0.00000421, -0.00000085, 0.00000499, 0.00000197, -0.00000055, 0.00000068, 0.00000123, -0.00000088, -0.00000040, -0.00000329, -0.00000203, -0.00000477, 0.00000220, 0.00000664, -0.00000119, },
		{ -0.00000089, 0.00000035, 0.00000075, 0.00000048, -0.00000105, 0.00000065, 0.00000107, -0.00000372, 0.00000006, -0.00000140, 0.00000177, 0.00000051, -0.00001072, 0.00000564, -0.00000505, 0.00000851, -0.00000256, -0.00000567, 0.00000783, -0.00000404, -0.00000153, -0.00000610, -0.00000061, 0.00001016, -0.00000273, 0.00000677, -0.00000508, -0.00000100, 0.00000956, -0.00002360, 0.00001111, -0.00001702, 0.00000068, -0.00001221, -0.00002466, 0.00001105, -0.00005829, 0.00002473, -0.00026931, 0.00006532, 0.00073728, -0.00027104, -0.00007951, -0.00002047, -0.00003979, 0.00003052, -0.00002410, 0.00002012, -0.00004542, 0.00004447, -0.00004046, 0.00005719, -0.00008121, 0.00006695, -0.00006388, 0.00005621, -0.00005329, 0.00003446, -0.00003019, 0.00002213, -0.00000722, 0.00000850, -0.00002157, 0.00001129, -0.00000342, -0.00000143, 0.00000766, -0.00001033, 0.00001126, -0.00001733, 0.00002044, -0.00002955, 0.00002254, -0.00001194, 0.00001708, -0.00001681, 0.00001236, -0.00001445, 0.00000697, 0.00000202, 0.00000553, -0.00000784, 0.00000919, -0.00000952, -0.00000350, 0.00000064, 0.00000176, 0.00000106, -0.00000192, 0.00000341, 0.00000054, 0.00000387, -0.00000659, -0.00000293, 0.00000433, -0.00000066, -0.00000123, -0.00000147, 0.00000068, -0.00000299, 0.00000082, 0.00000657, -0.00000228, -0.00000211, 0.00000038, 0.00000148, 0.00000170, -0.00000122, -0.00000019, -0.00000143, 0.00000050, -0.00000185, -0.00000066, 0.00000056, -0.00000417, -0.00000139, 0.00000482, 0.00000237, -0.00000049, 0.00000048, 0.00000144, -0.00000077, -0.00000051, -0.00000336, -0.00000180, -0.00000466, 0.00000119, 0.00000722, },
		{ 0.00000767, -0.00000048, 0.00000007, 0.00000074, 0.00000036, -0.00000079, 0.00000029, 0.00000168, -0.00000359, -0.00000063, -0.00000074, 0.00000057, 0.00000200, -0.00001120, 0.00000512, -0.00000441, 0.00000762, -0.00000098, -0.00000756, 0.00000850, -0.00000364, -0.00000158, -0.00000573, -0.00000201, 0.00001118, -0.00000276, 0.00000547, -0.00000279, -0.00000387, 0.00001250, -0.00002422, 0.00000885, -0.00001179, -0.00000516, -0.00000655, -0.00003076, 0.00001774, -0.00006338, 0.00003050, -0.00025521, -0.00000279, 0.00079143, -0.00025769, -0.00010240, -0.00000554, -0.00005534, 0.00004221, -0.00003520, 0.00003048, -0.00005778, 0.00005638, -0.00005241, 0.00006793, -0.00008705, 0.00006897, -0.00006304, 0.00005310, -0.00004873, 0.00003020, -0.00002525, 0.00001613, -0.00000272, 0.00000611, -0.00001814, 0.00000598, 0.00000143, -0.00000556, 0.00001190, -0.00001458, 0.00001517, -0.00002067, 0.00002306, -0.00003046, 0.00002116, -0.00001065, 0.00001654, -0.00001583, 0.00001104, -0.00001250, 0.00000475, 0.00000288, 0.00000564, -0.00000749, 0.00000837, -0.00000782, -0.00000501, 0.00000114, 0.00000110, 0.00000176, -0.00000239, 0.00000352, 0.00000058, 0.00000381, -0.00000558, -0.00000400, 0.00000435, -0.00000041, -0.00000114, -0.00000157, 0.00000061, -0.00000269, 0.00000001, 0.00000674, -0.00000163, -0.00000246, 0.00000039, 0.00000138, 0.00000192, -0.00000126, -0.00000012, -0.00000146, 0.00000034, -0.00000160, -0.00000085, 0.00000082, -0.00000409, -0.00000190, 0.00000460, 0.00000272, -0.00000036, 0.00000026, 0.00000163, -0.00000062, -0.00000059, -0.00000345, -0.00000164, -0.00000443, 0.00000016, },
		{ -0.00000086, 0.00000797, 0.00000006, -0.00000023, 0.00000073, 0.00000022, -0.00000054, -0.00000004, 0.00000216, -0.00000322, -0.00000151, 0.00000000, -0.00000060, 0.00000325, -0.00001132, 0.00000429, -0.00000357, 0.00000655, 0.00000071, -0.00000936, 0.00000887, -0.00000298, -0.00000180, -0.00000520, -0.00000344, 0.00001200, -0.00000244, 0.00000386, -0.00000045, -0.00000647, 0.00001467, -0.00002364, 0.00000537, -0.00000576, -0.00001114, -0.00000125, -0.00003599, 0.00002282, -0.00006572, 0.00003140, -0.00023358, -0.00007514, 0.00084060, -0.00023594, -0.00013085, 0.00001281, -0.00007321, 0.00005551, -0.00004754, 0.00004190, -0.00007077, 0.00006826, -0.00006354, 0.00007685, -0.00009058, 0.00006870, -0.00006016, 0.00004834, -0.00004298, 0.00002505, -0.00001958, 0.00000978, 0.00000181, 0.00000356, -0.00001431, 0.00000044, 0.00000624, -0.00000960, 0.00001599, -0.00001852, 0.00001858, -0.00002334, 0.00002484, -0.00003044, 0.00001909, -0.00000902, 0.00001573, -0.00001458, 0.00000950, -0.00001045, 0.00000267, 0.00000351, 0.00000583, -0.00000705, 0.00000743, -0.00000612, -0.00000636, 0.00000147, 0.00000054, 0.00000233, -0.00000267, 0.00000345, 0.00000078, 0.00000360, -0.00000449, -0.00000496, 0.00000423, -0.00000012, -0.00000108, -0.00000161, 0.00000048, -0.00000233, -0.00000074, 0.00000677, -0.00000092, -0.00000278, 0.00000037, 0.00000127, 0.00000214, -0.00000127, -0.00000009, -0.00000143, 0.00000013, -0.00000135, -0.00000103, 0.00000106, -0.00000397, -0.00000239, 0.00000434, 0.00000304, -0.00000018, 0.00000003, 0.00000177, -0.00000044, -0.00000065, -0.00000354, -0.00000157, -0.00000408, },
		{ -0.00000364, -0.00000184, 0.00000809, 0.00000070, -0.00000054, 0.00000073, 0.00000008, -0.00000031, -0.00000034, 0.00000249, -0.00000261, -0.00000251, 0.00000077, -0.00000168, 0.00000422, -0.00001105, 0.00000316, -0.00000254, 0.00000537, 0.00000244, -0.00001096, 0.00000887, -0.00000206, -0.00000216, -0.00000454, -0.00000485, 0.00001258, -0.00000180, 0.00000205, 0.00000181, -0.00000865, 0.00001595, -0.00002188, 0.00000083, 0.00000075, -0.00001685, 0.00000326, -0.00003992, 0.00002581, -0.00006488, 0.00002722, -0.00020498, -0.00015009, 0.00088297, -0.00020484, -0.00016500, 0.00003440, -0.00009310, 0.00007009, -0.00006074, 0.00005393, -0.00008384, 0.00007950, -0.00007326, 0.00008347, -0.00009153, 0.00006608, -0.00005529, 0.00004207, -0.00003623, 0.00001917, -0.00001337, 0.00000326, 0.00000625, 0.00000090, -0.00001016, -0.00000519, 0.00001089, -0.00001341, 0.00001981, -0.00002202, 0.00002136, -0.00002523, 0.00002572, -0.00002952, 0.00001644, -0.00000715, 0.00001472, -0.00001313, 0.00000781, -0.00000839, 0.00000077, 0.00000391, 0.00000607, -0.00000653, 0.00000639, -0.00000445, -0.00000752, 0.00000163, 0.00000010, 0.00000274, -0.00000277, 0.00000321, 0.00000110, 0.00000327, -0.00000338, -0.00000578, 0.00000395, 0.00000023, -0.00000104, -0.00000160, 0.00000029, -0.00000195, -0.00000141, 0.00000665, -0.00000017, -0.00000304, 0.00000031, 0.00000116, 0.00000233, -0.00000123, -0.00000011, -0.00000136, -0.00000010, -0.00000110, -0.00000119, 0.00000126, -0.00000380, -0.00000286, 0.00000404, 0.00000329, 0.00000006, -0.00000020, 0.00000187, -0.00000024, -0.00000070, -0.00000361, -0.00000159, },
		{ -0.00000169, -0.00000313, -0.00000275, 0.00000805, 0.00000144, -0.00000084, 0.00000071, -0.00000005, -0.00000012, -0.00000057, 0.00000265, -0.00000180, -0.00000360, 0.00000148, -0.00000259, 0.00000485, -0.00001040, 0.00000177, -0.00000140, 0.00000414, 0.00000413, -0.00001231, 0.00000847, -0.00000089, -0.00000265, -0.00000379, -0.00000621, 0.00001291, -0.00000087, 0.00000012, 0.00000388, -0.00001028, 0.00001629, -0.00001900, -0.00000452, 0.00000737, -0.00002188, 0.00000657, -0.00004216, 0.00002636, -0.00006060, 0.00001800, -0.00017025, -0.00022584, 0.00091674, -0.00016366, -0.00020473, 0.00005882, -0.00011454, 0.00008545, -0.00007431, 0.00006604, -0.00009642, 0.00008951, -0.00008100, 0.00008740, -0.00008971, 0.00006110, -0.00004857, 0.00003450, -0.00002869, 0.00001277, -0.00000681, -0.00000324, 0.00001047, -0.00000179, -0.00000580, -0.00001077, 0.00001523, -0.00001689, 0.00002323, -0.00002496, 0.00002340, -0.00002629, 0.00002571, -0.00002778, 0.00001331, -0.00000514, 0.00001360, -0.00001154, 0.00000606, -0.00000639, -0.00000089, 0.00000406, 0.00000636, -0.00000593, 0.00000530, -0.00000285, -0.00000846, 0.00000161, -0.00000022, 0.00000300, -0.00000270, 0.00000282, 0.00000152, 0.00000286, -0.00000229, -0.00000644, 0.00000352, 0.00000062, -0.00000102, -0.00000154, 0.00000005, -0.00000155, -0.00000199, 0.00000639, 0.00000060, -0.00000324, 0.00000022, 0.00000105, 0.00000251, -0.00000115, -0.00000016, -0.00000124, -0.00000036, -0.00000086, -0.00000134, 0.00000143, -0.00000358, -0.00000331, 0.00000373, 0.00000349, 0.00000035, -0.00000041, 0.00000192, -0.00000001, -0.00000072, -0.00000366, },
		{ -0.00000367, -0.00000189, -0.00000258, -0.00000356, 0.00000782, 0.00000224, -0.00000109, 0.00000065, -0.00000015, 0.00000001, -0.00000073, 0.00000264, -0.00000080, -0.00000469, 0.00000208, -0.00000327, 0.00000511, -0.00000939, 0.00000019, -0.00000021, 0.00000291, 0.00000570, -0.00001332, 0.00000766, 0.00000051, -0.00000322, -0.00000298, -0.00000748, 0.00001299, 0.00000030, -0.00000184, 0.00000564, -0.00001128, 0.00001567, -0.00001515, -0.00001036, 0.00001369, -0.00002583, 0.00000832, -0.00004238, 0.00002420, -0.00005275, 0.00000399, -0.00013047, -0.00030045, 0.00094025, -0.00011192, -0.00024968, 0.00008551, -0.00013695, 0.00010102, -0.00008769, 0.00007765, -0.00010791, 0.00009771, -0.00008629, 0.00008834, -0.00008508, 0.00005389, -0.00004021, 0.00002590, -0.00002062, 0.00000609, -0.00000013, -0.00000955, 0.00001437, -0.00000444, -0.00000134, -0.00001615, 0.00001913, -0.00001991, 0.00002614, -0.00002724, 0.00002465, -0.00002649, 0.00002485, -0.00002532, 0.00000986, -0.00000310, 0.00001245, -0.00000990, 0.00000433, -0.00000453, -0.00000227, 0.00000398, 0.00000667, -0.00000524, 0.00000416, -0.00000137, -0.00000916, 0.00000141, -0.00000043, 0.00000311, -0.00000247, 0.00000232, 0.00000200, 0.00000243, -0.00000126, -0.00000690, 0.00000295, 0.00000104, -0.00000102, -0.00000144, -0.00000021, -0.00000117, -0.00000246, 0.00000601, 0.00000138, -0.00000337, 0.00000010, 0.00000094, 0.00000265, -0.00000102, -0.00000026, -0.00000108, -0.00000063, -0.00000065, -0.00000146, 0.00000156, -0.00000331, -0.00000375, 0.00000340, 0.00000363, 0.00000067, -0.00000059, 0.00000191, 0.00000024, -0.00000074, },
		{ -0.00000074, -0.00000363, -0.00000216, -0.00000200, -0.00000424, 0.00000742, 0.00000308, -0.00000129, 0.00000056, -0.00000022, 0.00000010, -0.00000082, 0.00000248, 0.00000033, -0.00000574, 0.00000250, -0.00000368, 0.00000501, -0.00000806, -0.00000151, 0.00000096, 0.00000177, 0.00000709, -0.00001394, 0.00000643, 0.00000209, -0.00000383, -0.00000216, -0.00000862, 0.00001280, 0.00000165, -0.00000371, 0.00000699, -0.00001160, 0.00001416, -0.00001054, -0.00001635, 0.00001929, -0.00002831, 0.00000820, -0.00004033, 0.00001920, -0.00004141, -0.00001431, -0.00008694, -0.00037193, 0.00095201, -0.00004945, -0.00029921, 0.00011368, -0.00015958, 0.00011615, -0.00010026, 0.00008816, -0.00011771, 0.00010357, -0.00008872, 0.00008613, -0.00007771, 0.00004467, -0.00003053, 0.00001659, -0.00001229, -0.00000065, 0.00000647, -0.00001550, 0.00001785, -0.00000698, 0.00000311, -0.00002121, 0.00002249, -0.00002238, 0.00002845, -0.00002877, 0.00002506, -0.00002586, 0.00002321, -0.00002227, 0.00000624, -0.00000116, 0.00001135, -0.00000827, 0.00000267, -0.00000286, -0.00000335, 0.00000366, 0.00000699, -0.00000447, 0.00000303, -0.00000004, -0.00000961, 0.00000105, -0.00000053, 0.00000308, -0.00000211, 0.00000172, 0.00000250, 0.00000199, -0.00000031, -0.00000716, 0.00000225, 0.00000147, -0.00000103, -0.00000131, -0.00000049, -0.00000082, -0.00000282, 0.00000551, 0.00000213, -0.00000342, -0.00000005, 0.00000082, 0.00000276, -0.00000083, -0.00000040, -0.00000090, -0.00000089, -0.00000046, -0.00000155, 0.00000165, -0.00000299, -0.00000416, 0.00000306, 0.00000371, 0.00000101, -0.00000072, 0.00000185, 0.00000051, },
		{ 0.00000080, -0.00000074, -0.00000355, -0.00000250, -0.00000144, -0.00000478, 0.00000686, 0.00000394, -0.00000140, 0.00000041, -0.00000025, 0.00000012, -0.00000083, 0.00000217, 0.00000156, -0.00000667, 0.00000269, -0.00000378, 0.00000454, -0.00000647, -0.00000327, 0.00000204, 0.00000078, 0.00000823, -0.00001413, 0.00000480, 0.00000380, -0.00000443, -0.00000136, -0.00000960, 0.00001237, 0.00000312, -0.00000543, 0.00000788, -0.00001125, 0.00001186, -0.00000545, -0.00002210, 0.00002377, -0.00002899, 0.00000599, -0.00003591, 0.00001138, -0.00002684, -0.00003621, -0.00004114, -0.00043832, 0.00095077, 0.00002360, -0.00035242, 0.00014239, -0.00018163, 0.00013010, -0.00011139, 0.00009698, -0.00012529, 0.00010664, -0.00008802, 0.00008072, -0.00006778, 0.00003374, -0.00001986, 0.00000690, -0.00000401, -0.00000720, 0.00001279, -0.00002094, 0.00002083, -0.00000934, 0.00000743, -0.00002580, 0.00002519, -0.00002423, 0.00003009, -0.00002952, 0.00002463, -0.00002445, 0.00002089, -0.00001879, 0.00000259, 0.00000058, 0.00001035, -0.00000672, 0.00000115, -0.00000142, -0.00000412, 0.00000316, 0.00000727, -0.00000362, 0.00000192, 0.00000111, -0.00000980, 0.00000053, -0.00000055, 0.00000294, -0.00000164, 0.00000108, 0.00000299, 0.00000160, 0.00000050, -0.00000721, 0.00000143, 0.00000189, -0.00000104, -0.00000116, -0.00000077, -0.00000051, -0.00000306, 0.00000491, 0.00000284, -0.00000338, -0.00000023, 0.00000070, 0.00000283, -0.00000059, -0.00000056, -0.00000071, -0.00000115, -0.00000031, -0.00000161, 0.00000169, -0.00000262, -0.00000456, 0.00000271, 0.00000373, 0.00000136, -0.00000080, 0.00000172, },
		{ 0.00000153, 0.00000110, -0.00000074, -0.00000342, -0.00000289, -0.00000092, -0.00000516, 0.00000616, 0.00000478, -0.00000142, 0.00000020, -0.00000023, 0.00000010, -0.00000077, 0.00000175, 0.00000281, -0.00000741, 0.00000259, -0.00000358, 0.00000376, -0.00000469, -0.00000498, 0.00000294, 0.00000000, 0.00000907, -0.00001388, 0.00000281, 0.00000557, -0.00000497, -0.00000062, -0.00001041, 0.00001171, 0.00000466, -0.00000693, 0.00000828, -0.00001027, 0.00000895, -0.00000018, -0.00002721, 0.00002675, -0.00002760, 0.00000154, -0.00002911, 0.00000092, -0.00000947, -0.00006082, 0.00000539, -0.00049772, 0.00093558, 0.00010669, -0.00040814, 0.00017056, -0.00020219, 0.00014214, -0.00012046, 0.00010355, -0.00013015, 0.00010658, -0.00008403, 0.00007223, -0.00005560, 0.00002151, -0.00000861, -0.00000282, 0.00000396, -0.00001336, 0.00001866, -0.00002575, 0.00002325, -0.00001147, 0.00001153, -0.00002981, 0.00002716, -0.00002541, 0.00003103, -0.00002947, 0.00002340, -0.00002233, 0.00001803, -0.00001503, -0.00000093, 0.00000203, 0.00000951, -0.00000528, -0.00000018, -0.00000023, -0.00000459, 0.00000249, 0.00000750, -0.00000270, 0.00000087, 0.00000206, -0.00000975, -0.00000011, -0.00000049, 0.00000269, -0.00000108, 0.00000042, 0.00000343, 0.00000128, 0.00000118, -0.00000706, 0.00000052, 0.00000229, -0.00000104, -0.00000100, -0.00000103, -0.00000027, -0.00000318, 0.00000425, 0.00000349, -0.00000326, -0.00000042, 0.00000058, 0.00000286, -0.00000030, -0.00000074, -0.00000051, -0.00000137, -0.00000021, -0.00000163, 0.00000168, -0.00000221, -0.00000492, 0.00000235, 0.00000371, 0.00000169, -0.00000082, },
		{ -0.00000076, 0.00000128, 0.00000140, -0.00000073, -0.00000324, -0.00000331, -0.00000046, -0.00000538, 0.00000533, 0.00000557, -0.00000132, -0.00000006, -0.00000019, 0.00000003, -0.00000066, 0.00000123, 0.00000404, -0.00000792, 0.00000220, -0.00000309, 0.00000270, -0.00000281, -0.00000656, 0.00000361, -0.00000052, 0.00000956, -0.00001318, 0.00000051, 0.00000734, -0.00000540, 0.00000003, -0.00001103, 0.00001084, 0.00000621, -0.00000815, 0.00000820, -0.00000876, 0.00000564, 0.00000494, -0.00003131, 0.00002792, -0.00002397, -0.00000514, -0.00002005, -0.00001188, 0.00001010, -0.00008713, 0.00005100, -0.00054842, 0.00090587, 0.00019894, -0.00046496, 0.00019698, -0.00022033, 0.00015154, -0.00012686, 0.00010738, -0.00013191, 0.00010318, -0.00007676, 0.00006088, -0.00004157, 0.00000842, 0.00000280, -0.00001220, 0.00001134, -0.00001894, 0.00002393, -0.00002983, 0.00002508, -0.00001333, 0.00001531, -0.00003314, 0.00002834, -0.00002588, 0.00003125, -0.00002863, 0.00002142, -0.00001963, 0.00001477, -0.00001117, -0.00000418, 0.00000311, 0.00000888, -0.00000400, -0.00000131, 0.00000068, -0.00000477, 0.00000171, 0.00000765, -0.00000172, -0.00000010, 0.00000279, -0.00000947, -0.00000085, -0.00000038, 0.00000237, -0.00000048, -0.00000022, 0.00000380, 0.00000104, 0.00000171, -0.00000672, -0.00000044, 0.00000264, -0.00000101, -0.00000084, -0.00000126, -0.00000009, -0.00000320, 0.00000353, 0.00000406, -0.00000305, -0.00000062, 0.00000046, 0.00000284, 0.00000004, -0.00000093, -0.00000032, -0.00000156, -0.00000015, -0.00000162, 0.00000163, -0.00000177, -0.00000526, 0.00000198, 0.00000366, 0.00000199, },
		{ 0.00000226, -0.00000062, 0.00000098, 0.00000170, -0.00000071, -0.00000303, -0.00000375, -0.00000008, -0.00000543, 0.00000440, 0.00000627, -0.00000109, -0.00000037, -0.00000012, -0.00000007, -0.00000050, 0.00000065, 0.00000519, -0.00000814, 0.00000150, -0.00000232, 0.00000144, -0.00000093, -0.00000792, 0.00000397, -0.00000074, 0.00000970, -0.00001205, -0.00000202, 0.00000901, -0.00000567, 0.00000055, -0.00001144, 0.00000977, 0.00000773, -0.00000909, 0.00000769, -0.00000686, 0.00000217, 0.00000956, -0.00003405, 0.00002704, -0.00001805, -0.00001395, -0.00000901, -0.00002656, 0.00003117, -0.00011400, 0.00009411, -0.00058893, 0.00086145, 0.00029907, -0.00052127, 0.00022039, -0.00023512, 0.00015761, -0.00013010, 0.00010808, -0.00013032, 0.00009635, -0.00006635, 0.00004705, -0.00002620, -0.00000504, 0.00001394, -0.00002092, 0.00001791, -0.00002377, 0.00002847, -0.00003311, 0.00002628, -0.00001488, 0.00001868, -0.00003571, 0.00002870, -0.00002564, 0.00003075, -0.00002706, 0.00001881, -0.00001647, 0.00001128, -0.00000738, -0.00000703, 0.00000377, 0.00000847, -0.00000290, -0.00000221, 0.00000133, -0.00000471, 0.00000086, 0.00000769, -0.00000069, -0.00000096, 0.00000329, -0.00000897, -0.00000165, -0.00000025, 0.00000199, 0.00000013, -0.00000080, 0.00000406, 0.00000091, 0.00000208, -0.00000622, -0.00000144, 0.00000293, -0.00000095, -0.00000069, -0.00000145, 0.00000001, -0.00000311, 0.00000278, 0.00000453, -0.00000275, -0.00000083, 0.00000035, 0.00000276, 0.00000041, -0.00000112, -0.00000015, -0.00000171, -0.00000015, -0.00000157, 0.00000153, -0.00000129, -0.00000556, 0.00000160, 0.00000358, },
		{ 0.00000349, 0.00000246, -0.00000040, 0.00000064, 0.00000199, -0.00000068, -0.00000278, -0.00000417, 0.00000018, -0.00000531, 0.00000341, 0.00000687, -0.00000074, -0.00000071, -0.00000004, -0.00000019, -0.00000033, 0.00000005, 0.00000622, -0.00000805, 0.00000051, -0.00000135, 0.00000007, 0.00000086, -0.00000898, 0.00000398, -0.00000064, 0.00000947, -0.00001053, -0.00000469, 0.00001050, -0.00000573, 0.00000092, -0.00001164, 0.00000854, 0.00000919, -0.00000974, 0.00000681, -0.00000473, -0.00000119, 0.00001335, -0.00003512, 0.00002395, -0.00000987, -0.00002467, 0.00000365, -0.00004255, 0.00005290, -0.00014027, 0.00013319, -0.00061804, 0.00080254, 0.00040549, -0.00057529, 0.00023950, -0.00024571, 0.00015976, -0.00012976, 0.00010537, -0.00012524, 0.00008617, -0.00005309, 0.00003120, -0.00001003, -0.00001835, 0.00002440, -0.00002867, 0.00002348, -0.00002774, 0.00003222, -0.00003556, 0.00002686, -0.00001610, 0.00002158, -0.00003746, 0.00002824, -0.00002471, 0.00002959, -0.00002483, 0.00001567, -0.00001302, 0.00000773, -0.00000381, -0.00000938, 0.00000397, 0.00000829, -0.00000197, -0.00000290, 0.00000174, -0.00000444, -0.00000002, 0.00000761, 0.00000036, -0.00000168, 0.00000357, -0.00000830, -0.00000248, -0.00000012, 0.00000158, 0.00000074, -0.00000131, 0.00000420, 0.00000089, 0.00000231, -0.00000557, -0.00000242, 0.00000313, -0.00000086, -0.00000057, -0.00000159, 0.00000004, -0.00000294, 0.00000203, 0.00000489, -0.00000238, -0.00000104, 0.00000024, 0.00000264, 0.00000081, -0.00000129, -0.00000002, -0.00000180, -0.00000019, -0.00000149, 0.00000138, -0.00000080, -0.00000580, 0.00000119, },
		{ 0.00000076, 0.00000340, 0.00000261, -0.00000011, 0.00000028, 0.00000226, -0.00000064, -0.00000251, -0.00000455, 0.00000032, -0.00000505, 0.00000238, 0.00000734, -0.00000027, -0.00000108, 0.00000003, -0.00000031, -0.00000015, -0.00000054, 0.00000707, -0.00000764, -0.00000074, -0.00000023, -0.00000132, 0.00000246, -0.00000967, 0.00000361, -0.00000021, 0.00000891, -0.00000869, -0.00000740, 0.00001173, -0.00000553, 0.00000112, -0.00001161, 0.00000715, 0.00001058, -0.00001013, 0.00000568, -0.00000255, -0.00000417, 0.00001602, -0.00003433, 0.00001863, 0.00000039, -0.00003695, 0.00001742, -0.00005919, 0.00007441, -0.00016477, 0.00016687, -0.00063488, 0.00072981, 0.00051627, -0.00062514, 0.00025306, -0.00025130, 0.00015750, -0.00012556, 0.00009911, -0.00011671, 0.00007285, -0.00003738, 0.00001392, 0.00000633, -0.00003101, 0.00003380, -0.00003520, 0.00002789, -0.00003077, 0.00003512, -0.00003716, 0.00002684, -0.00001697, 0.00002394, -0.00003837, 0.00002696, -0.00002316, 0.00002784, -0.00002204, 0.00001214, -0.00000943, 0.00000427, -0.00000060, -0.00001116, 0.00000371, 0.00000831, -0.00000121, -0.00000336, 0.00000192, -0.00000400, -0.00000087, 0.00000740, 0.00000143, -0.00000224, 0.00000363, -0.00000748, -0.00000331, -0.00000002, 0.00000117, 0.00000130, -0.00000171, 0.00000421, 0.00000098, 0.00000240, -0.00000481, -0.00000337, 0.00000324, -0.00000071, -0.00000048, -0.00000167, 0.00000000, -0.00000270, 0.00000130, 0.00000514, -0.00000194, -0.00000123, 0.00000013, 0.00000247, 0.00000123, -0.00000142, 0.00000007, -0.00000184, -0.00000028, -0.00000138, 0.00000120, -0.00000029, -0.00000599, },
		{ -0.00000610, 0.00000029, 0.00000332, 0.00000268, 0.00000025, -0.00000010, 0.00000249, -0.00000058, -0.00000222, -0.00000489, 0.00000032, -0.00000465, 0.00000136, 0.00000765, 0.00000032, -0.00000145, 0.00000009, -0.00000042, 0.00000001, -0.00000110, 0.00000772, -0.00000690, -0.00000221, 0.00000096, -0.00000265, 0.00000378, -0.00000995, 0.00000287, 0.00000051, 0.00000804, -0.00000662, -0.00001005, 0.00001262, -0.00000505, 0.00000114, -0.00001136, 0.00000562, 0.00001189, -0.00001032, 0.00000442, -0.00000054, -0.00000651, 0.00001730, -0.00003153, 0.00001114, 0.00001244, -0.00005034, 0.00003175, -0.00007577, 0.00009480, -0.00018638, 0.00019395, -0.00063896, 0.00064433, 0.00062924, -0.00066892, 0.00025991, -0.00025127, 0.00015052, -0.00011737, 0.00008932, -0.00010488, 0.00005675, -0.00001977, -0.00000416, 0.00002228, -0.00004254, 0.00004180, -0.00004031, 0.00003107, -0.00003282, 0.00003716, -0.00003795, 0.00002626, -0.00001749, 0.00002573, -0.00003842, 0.00002494, -0.00002104, 0.00002557, -0.00001881, 0.00000839, -0.00000587, 0.00000106, 0.00000214, -0.00001233, 0.00000301, 0.00000851, -0.00000061, -0.00000364, 0.00000192, -0.00000346, -0.00000166, 0.00000706, 0.00000248, -0.00000264, 0.00000349, -0.00000656, -0.00000410, 0.00000004, 0.00000077, 0.00000180, -0.00000199, 0.00000409, 0.00000117, 0.00000238, -0.00000397, -0.00000425, 0.00000323, -0.00000053, -0.00000042, -0.00000170, -0.00000010, -0.00000241, 0.00000061, 0.00000526, -0.00000144, -0.00000141, 0.00000003, 0.00000225, 0.00000166, -0.00000152, 0.00000012, -0.00000183, -0.00000041, -0.00000125, 0.00000100, 0.00000021, },
		{ 0.00000069, -0.00000613, -0.00000022, 0.00000325, 0.00000268, 0.00000067, -0.00000048, 0.00000267, -0.00000050, -0.00000194, -0.00000514, 0.00000018, -0.00000414, 0.00000038, 0.00000779, 0.00000100, -0.00000180, 0.00000011, -0.00000050, 0.00000014, -0.00000159, 0.00000815, -0.00000585, -0.00000382, 0.00000213, -0.00000382, 0.00000476, -0.00000978, 0.00000175, 0.00000150, 0.00000694, -0.00000439, -0.00001252, 0.00001309, -0.00000426, 0.00000095, -0.00001088, 0.00000396, 0.00001313, -0.00001037, 0.00000318, 0.00000110, -0.00000798, 0.00001701, -0.00002669, 0.00000167, 0.00002589, -0.00006431, 0.00004599, -0.00009153, 0.00011317, -0.00020411, 0.00021352, -0.00063016, 0.00054759, 0.00074199, -0.00070472, 0.00025905, -0.00024517, 0.00013866, -0.00010524, 0.00007619, -0.00009009, 0.00003837, -0.00000086, -0.00002237, 0.00003722, -0.00005251, 0.00004815, -0.00004389, 0.00003297, -0.00003391, 0.00003837, -0.00003796, 0.00002516, -0.00001767, 0.00002693, -0.00003763, 0.00002224, -0.00001846, 0.00002290, -0.00001528, 0.00000456, -0.00000248, -0.00000178, 0.00000433, -0.00001286, 0.00000191, 0.00000885, -0.00000012, -0.00000376, 0.00000178, -0.00000284, -0.00000236, 0.00000658, 0.00000349, -0.00000286, 0.00000316, -0.00000556, -0.00000481, 0.00000004, 0.00000040, 0.00000221, -0.00000213, 0.00000384, 0.00000145, 0.00000227, -0.00000309, -0.00000503, 0.00000311, -0.00000029, -0.00000039, -0.00000168, -0.00000026, -0.00000208, -0.00000002, 0.00000526, -0.00000090, -0.00000156, -0.00000006, 0.00000199, 0.00000207, -0.00000155, 0.00000010, -0.00000176, -0.00000058, -0.00000110, 0.00000077, },
		{ 0.00000053, 0.00000114, -0.00000608, -0.00000076, 0.00000320, 0.00000261, 0.00000113, -0.00000083, 0.00000279, -0.00000038, -0.00000168, -0.00000532, -0.00000009, -0.00000355, -0.00000053, 0.00000776, 0.00000177, -0.00000210, 0.00000007, -0.00000053, 0.00000023, -0.00000201, 0.00000835, -0.00000454, -0.00000551, 0.00000320, -0.00000475, 0.00000533, -0.00000916, 0.00000030, 0.00000269, 0.00000567, -0.00000212, -0.00001470, 0.00001307, -0.00000314, 0.00000057, -0.00001017, 0.00000216, 0.00001433, -0.00001037, 0.00000210, 0.00000219, -0.00000839, 0.00001504, -0.00001989, -0.00000947, 0.00004026, -0.00007827, 0.00005948, -0.00010572, 0.00012871, -0.00021709, 0.00022492, -0.00060878, 0.00044143, 0.00085203, -0.00073078, 0.00024969, -0.00023274, 0.00012196, -0.00008938, 0.00006002, -0.00007279, 0.00001829, 0.00001864, -0.00004002, 0.00005059, -0.00006057, 0.00005266, -0.00004587, 0.00003364, -0.00003410, 0.00003881, -0.00003726, 0.00002361, -0.00001752, 0.00002752, -0.00003605, 0.00001896, -0.00001554, 0.00001995, -0.00001158, 0.00000082, 0.00000058, -0.00000414, 0.00000591, -0.00001277, 0.00000047, 0.00000925, 0.00000027, -0.00000376, 0.00000154, -0.00000221, -0.00000293, 0.00000597, 0.00000444, -0.00000290, 0.00000268, -0.00000455, -0.00000542, -0.00000003, 0.00000008, 0.00000253, -0.00000215, 0.00000347, 0.00000179, 0.00000209, -0.00000220, -0.00000569, 0.00000286, -0.00000002, -0.00000039, -0.00000161, -0.00000046, -0.00000174, -0.00000058, 0.00000514, -0.00000033, -0.00000167, -0.00000015, 0.00000170, 0.00000247, -0.00000152, 0.00000004, -0.00000165, -0.00000077, -0.00000095, },
		{ -0.00000079, 0.00000029, 0.00000155, -0.00000592, -0.00000134, 0.00000317, 0.00000249, 0.00000161, -0.00000114, 0.00000284, -0.00000022, -0.00000144, -0.00000539, -0.00000049, -0.00000290, -0.00000134, 0.00000755, 0.00000259, -0.00000233, -0.00000003, -0.00000053, 0.00000028, -0.00000235, 0.00000832, -0.00000301, -0.00000718, 0.00000407, -0.00000537, 0.00000547, -0.00000811, -0.00000142, 0.00000401, 0.00000433, 0.00000008, -0.00001648, 0.00001252, -0.00000170, -0.00000002, -0.00000922, 0.00000022, 0.00001553, -0.00001041, 0.00000133, 0.00000257, -0.00000760, 0.00001137, -0.00001131, -0.00002190, 0.00005498, -0.00009157, 0.00007156, -0.00011765, 0.00014069, -0.00022466, 0.00022780, -0.00057551, 0.00032798, 0.00095678, -0.00074547, 0.00023126, -0.00021398, 0.00010067, -0.00007015, 0.00004130, -0.00005355, -0.00000281, 0.00003802, -0.00005645, 0.00006193, -0.00006645, 0.00005523, -0.00004627, 0.00003314, -0.00003347, 0.00003856, -0.00003595, 0.00002168, -0.00001706, 0.00002753, -0.00003375, 0.00001521, -0.00001238, 0.00001684, -0.00000787, -0.00000270, 0.00000320, -0.00000594, 0.00000686, -0.00001212, -0.00000122, 0.00000968, 0.00000062, -0.00000367, 0.00000123, -0.00000161, -0.00000335, 0.00000525, 0.00000531, -0.00000275, 0.00000207, -0.00000355, -0.00000590, -0.00000019, -0.00000019, 0.00000274, -0.00000204, 0.00000300, 0.00000217, 0.00000188, -0.00000133, -0.00000621, 0.00000250, 0.00000029, -0.00000042, -0.00000149, -0.00000069, -0.00000141, -0.00000104, 0.00000491, 0.00000025, -0.00000175, -0.00000022, 0.00000138, 0.00000282, -0.00000142, -0.00000008, -0.00000150, -0.00000098, },
		{ -0.00000118, -0.00000065, 0.00000005, 0.00000189, -0.00000567, -0.00000196, 0.00000314, 0.00000231, 0.00000210, -0.00000138, 0.00000282, -0.00000003, -0.00000123, -0.00000535, -0.00000100, -0.00000224, -0.00000202, 0.00000717, 0.00000343, -0.00000247, -0.00000020, -0.00000048, 0.00000028, -0.00000259, 0.00000808, -0.00000132, -0.00000875, 0.00000467, -0.00000562, 0.00000517, -0.00000667, -0.00000333, 0.00000537, 0.00000300, 0.00000211, -0.00001778, 0.00001141, 0.00000006, -0.00000079, -0.00000804, -0.00000186, 0.00001675, -0.00001059, 0.00000100, 0.00000211, -0.00000555, 0.00000607, -0.00000124, -0.00003511, 0.00006945, -0.00010360, 0.00008160, -0.00012671, 0.00014855, -0.00022638, 0.00022213, -0.00053139, 0.00020963, 0.00105372, -0.00074744, 0.00020353, -0.00018914, 0.00007523, -0.00004812, 0.00002062, -0.00003304, -0.00002419, 0.00005657, -0.00007106, 0.00007082, -0.00006999, 0.00005586, -0.00004518, 0.00003161, -0.00003214, 0.00003772, -0.00003411, 0.00001944, -0.00001634, 0.00002698, -0.00003081, 0.00001114, -0.00000913, 0.00001372, -0.00000429, -0.00000585, 0.00000528, -0.00000714, 0.00000720, -0.00001097, -0.00000307, 0.00001005, 0.00000096, -0.00000352, 0.00000091, -0.00000106, -0.00000362, 0.00000444, 0.00000606, -0.00000242, 0.00000137, -0.00000261, -0.00000623, -0.00000044, -0.00000040, 0.00000285, -0.00000180, 0.00000246, 0.00000256, 0.00000165, -0.00000051, -0.00000658, 0.00000202, 0.00000063, -0.00000048, -0.00000135, -0.00000092, -0.00000110, -0.00000142, 0.00000458, 0.00000083, -0.00000179, -0.00000029, 0.00000103, 0.00000313, -0.00000123, -0.00000025, -0.00000132, },
		{ -0.00000112, -0.00000138, -0.00000053, -0.00000016, 0.00000217, -0.00000531, -0.00000261, 0.00000313, 0.00000210, 0.00000256, -0.00000155, 0.00000271, 0.00000020, -0.00000106, -0.00000521, -0.00000160, -0.00000159, -0.00000256, 0.00000663, 0.00000427, -0.00000250, -0.00000045, -0.00000038, 0.00000023, -0.00000275, 0.00000766, 0.00000045, -0.00001013, 0.00000494, -0.00000549, 0.00000445, -0.00000491, -0.00000535, 0.00000668, 0.00000180, 0.00000385, -0.00001850, 0.00000973, 0.00000210, -0.00000175, -0.00000663, -0.00000409, 0.00001805, -0.00001098, 0.00000121, 0.00000072, -0.00000224, -0.00000070, 0.00000996, -0.00004858, 0.00008304, -0.00011372, 0.00008905, -0.00013238, 0.00015186, -0.00022204, 0.00020820, -0.00047780, 0.00008893, 0.00114042, -0.00073563, 0.00016654, -0.00015871, 0.00004628, -0.00002394, -0.00000133, -0.00001199, -0.00004513, 0.00007361, -0.00008332, 0.00007700, -0.00007111, 0.00005461, -0.00004275, 0.00002922, -0.00003026, 0.00003640, -0.00003184, 0.00001698, -0.00001538, 0.00002591, -0.00002733, 0.00000688, -0.00000592, 0.00001070, -0.00000095, -0.00000854, 0.00000676, -0.00000773, 0.00000696, -0.00000941, -0.00000498, 0.00001030, 0.00000132, -0.00000336, 0.00000060, -0.00000059, -0.00000373, 0.00000356, 0.00000667, -0.00000193, 0.00000062, -0.00000176, -0.00000641, -0.00000077, -0.00000056, 0.00000286, -0.00000147, 0.00000187, 0.00000295, 0.00000144, 0.00000025, -0.00000680, 0.00000145, 0.00000097, -0.00000054, -0.00000119, -0.00000116, -0.00000083, -0.00000169, 0.00000416, 0.00000139, -0.00000177, -0.00000035, 0.00000068, 0.00000337, -0.00000097, -0.00000046, },
		{ -0.00000070, -0.00000092, -0.00000156, -0.00000044, -0.00000035, 0.00000238, -0.00000486, -0.00000327, 0.00000310, 0.00000187, 0.00000299, -0.00000163, 0.00000251, 0.00000047, -0.00000093, -0.00000497, -0.00000227, -0.00000100, -0.00000292, 0.00000594, 0.00000507, -0.00000241, -0.00000077, -0.00000026, 0.00000016, -0.00000285, 0.00000709, 0.00000224, -0.00001125, 0.00000484, -0.00000496, 0.00000334, -0.00000292, -0.00000735, 0.00000783, 0.00000081, 0.00000520, -0.00001859, 0.00000747, 0.00000440, -0.00000287, -0.00000498, -0.00000648, 0.00001945, -0.00001166, 0.00000203, -0.00000160, 0.00000225, -0.00000869, 0.00002183, -0.00006170, 0.00009513, -0.00012138, 0.00009344, -0.00013431, 0.00015040, -0.00021166, 0.00018659, -0.00041639, -0.00003150, 0.00121469, -0.00070934, 0.00012070, -0.00012344, 0.00001463, 0.00000158, -0.00002378, 0.00000885, -0.00006488, 0.00008851, -0.00009282, 0.00008029, -0.00006988, 0.00005165, -0.00003918, 0.00002616, -0.00002798, 0.00003471, -0.00002924, 0.00001436, -0.00001423, 0.00002437, -0.00002344, 0.00000258, -0.00000288, 0.00000789, 0.00000203, -0.00001067, 0.00000761, -0.00000772, 0.00000622, -0.00000755, -0.00000684, 0.00001038, 0.00000174, -0.00000320, 0.00000033, -0.00000022, -0.00000370, 0.00000266, 0.00000714, -0.00000128, -0.00000015, -0.00000104, -0.00000642, -0.00000117, -0.00000066, 0.00000277, -0.00000105, 0.00000126, 0.00000330, 0.00000126, 0.00000092, -0.00000685, 0.00000079, 0.00000132, -0.00000061, -0.00000101, -0.00000138, -0.00000061, -0.00000186, 0.00000367, 0.00000191, -0.00000171, -0.00000040, 0.00000032, 0.00000354, -0.00000063, },
		{ -0.00000022, -0.00000096, -0.00000073, -0.00000170, -0.00000039, -0.00000050, 0.00000250, -0.00000431, -0.00000393, 0.00000306, 0.00000164, 0.00000335, -0.00000160, 0.00000223, 0.00000076, -0.00000084, -0.00000463, -0.00000297, -0.00000049, -0.00000311, 0.00000514, 0.00000580, -0.00000218, -0.00000114, -0.00000011, 0.00000007, -0.00000288, 0.00000642, 0.00000398, -0.00001204, 0.00000432, -0.00000406, 0.00000193, -0.00000080, -0.00000924, 0.00000872, 0.00000012, 0.00000610, -0.00001801, 0.00000468, 0.00000691, -0.00000413, -0.00000311, -0.00000902, 0.00002099, -0.00001268, 0.00000351, -0.00000484, 0.00000778, -0.00001759, 0.00003387, -0.00007387, 0.00010513, -0.00012613, 0.00009444, -0.00013227, 0.00014415, -0.00019554, 0.00015818, -0.00034904, -0.00014904, 0.00127460, -0.00066828, 0.00006675, -0.00008429, -0.00001878, 0.00002757, -0.00004593, 0.00002873, -0.00008277, 0.00010075, -0.00009924, 0.00008065, -0.00006643, 0.00004722, -0.00003472, 0.00002264, -0.00002546, 0.00003278, -0.00002642, 0.00001165, -0.00001291, 0.00002244, -0.00001926, -0.00000162, -0.00000013, 0.00000541, 0.00000457, -0.00001219, 0.00000782, -0.00000718, 0.00000507, -0.00000551, -0.00000857, 0.00001024, 0.00000222, -0.00000306, 0.00000011, 0.00000005, -0.00000353, 0.00000175, 0.00000745, -0.00000051, -0.00000089, -0.00000046, -0.00000627, -0.00000164, -0.00000072, 0.00000260, -0.00000057, 0.00000065, 0.00000358, 0.00000113, 0.00000150, -0.00000676, 0.00000008, 0.00000164, -0.00000067, -0.00000084, -0.00000157, -0.00000045, -0.00000193, 0.00000313, 0.00000238, -0.00000160, -0.00000044, -0.00000004, 0.00000364, },
		{ 0.00000364, 0.00000026, -0.00000121, -0.00000056, -0.00000181, -0.00000037, -0.00000062, 0.00000254, -0.00000368, -0.00000457, 0.00000297, 0.00000143, 0.00000363, -0.00000147, 0.00000188, 0.00000109, -0.00000078, -0.00000422, -0.00000368, -0.00000009, -0.00000312, 0.00000424, 0.00000644, -0.00000181, -0.00000157, 0.00000004, -0.00000002, -0.00000289, 0.00000567, 0.00000559, -0.00001246, 0.00000341, -0.00000283, 0.00000029, 0.00000132, -0.00001088, 0.00000926, -0.00000018, 0.00000646, -0.00001676, 0.00000140, 0.00000957, -0.00000550, -0.00000103, -0.00001171, 0.00002268, -0.00001408, 0.00000564, -0.00000890, 0.00001413, -0.00002700, 0.00004554, -0.00008453, 0.00011255, -0.00012759, 0.00009185, -0.00012623, 0.00013326, -0.00017419, 0.00012406, -0.00027775, -0.00026116, 0.00131857, -0.00061258, 0.00000575, -0.00004245, -0.00005290, 0.00005312, -0.00006698, 0.00004693, -0.00009821, 0.00010992, -0.00010244, 0.00007817, -0.00006102, 0.00004159, -0.00002965, 0.00001890, -0.00002286, 0.00003070, -0.00002347, 0.00000894, -0.00001149, 0.00002019, -0.00001493, -0.00000557, 0.00000224, 0.00000333, 0.00000659, -0.00001309, 0.00000743, -0.00000618, 0.00000362, -0.00000342, -0.00001005, 0.00000984, 0.00000279, -0.00000296, -0.00000003, 0.00000021, -0.00000325, 0.00000087, 0.00000758, 0.00000036, -0.00000157, -0.00000004, -0.00000598, -0.00000215, -0.00000074, 0.00000237, -0.00000006, 0.00000008, 0.00000379, 0.00000107, 0.00000197, -0.00000653, -0.00000068, 0.00000194, -0.00000072, -0.00000069, -0.00000172, -0.00000035, -0.00000191, 0.00000255, 0.00000279, -0.00000144, -0.00000047, -0.00000038, },
		{ -0.00000069, 0.00000356, 0.00000079, -0.00000145, -0.00000043, -0.00000187, -0.00000039, -0.00000070, 0.00000250, -0.00000298, -0.00000517, 0.00000284, 0.00000126, 0.00000382, -0.00000123, 0.00000145, 0.00000142, -0.00000076, -0.00000374, -0.00000437, 0.00000017, -0.00000296, 0.00000328, 0.00000695, -0.00000131, -0.00000202, 0.00000016, -0.00000010, -0.00000288, 0.00000490, 0.00000703, -0.00001246, 0.00000212, -0.00000136, -0.00000146, 0.00000332, -0.00001218, 0.00000937, -0.00000004, 0.00000626, -0.00001483, -0.00000230, 0.00001230, -0.00000696, 0.00000124, -0.00001453, 0.00002453, -0.00001584, 0.00000836, -0.00001362, 0.00002102, -0.00003651, 0.00005631, -0.00009312, 0.00011697, -0.00012553, 0.00008561, -0.00011630, 0.00011811, -0.00014830, 0.00008549, -0.00020460, -0.00036551, 0.00134543, -0.00054282, -0.00006093, 0.00000076, -0.00008660, 0.00007729, -0.00008616, 0.00006281, -0.00011070, 0.00011574, -0.00010239, 0.00007307, -0.00005395, 0.00003511, -0.00002426, 0.00001514, -0.00002031, 0.00002859, -0.00002046, 0.00000626, -0.00000999, 0.00001769, -0.00001058, -0.00000916, 0.00000413, 0.00000171, 0.00000807, -0.00001336, 0.00000649, -0.00000482, 0.00000198, -0.00000138, -0.00001123, 0.00000917, 0.00000344, -0.00000290, -0.00000011, 0.00000027, -0.00000289, 0.00000005, 0.00000754, 0.00000131, -0.00000216, 0.00000020, -0.00000555, -0.00000268, -0.00000075, 0.00000208, 0.00000046, -0.00000044, 0.00000390, 0.00000107, 0.00000235, -0.00000617, -0.00000146, 0.00000219, -0.00000075, -0.00000056, -0.00000182, -0.00000032, -0.00000182, 0.00000195, 0.00000313, -0.00000123, -0.00000049, },
		{ -0.00000049, -0.00000098, 0.00000339, 0.00000136, -0.00000166, -0.00000035, -0.00000187, -0.00000046, -0.00000074, 0.00000238, -0.00000224, -0.00000571, 0.00000264, 0.00000113, 0.00000392, -0.00000088, 0.00000097, 0.00000176, -0.00000075, -0.00000322, -0.00000500, 0.00000027, -0.00000263, 0.00000229, 0.00000733, -0.00000068, -0.00000247, 0.00000025, -0.00000013, -0.00000288, 0.00000415, 0.00000823, -0.00001205, 0.00000050, 0.00000028, -0.00000320, 0.00000508, -0.00001303, 0.00000899, 0.00000058, 0.00000548, -0.00001228, -0.00000633, 0.00001503, -0.00000847, 0.00000369, -0.00001746, 0.00002651, -0.00001794, 0.00001157, -0.00001882, 0.00002813, -0.00004564, 0.00006564, -0.00009920, 0.00011808, -0.00011988, 0.00007584, -0.00010280, 0.00009923, -0.00011877, 0.00004390, -0.00013166, -0.00045999, 0.00135446, -0.00045999, -0.00013166, 0.00004390, -0.00011877, 0.00009923, -0.00010280, 0.00007584, -0.00011988, 0.00011808, -0.00009920, 0.00006564, -0.00004564, 0.00002813, -0.00001882, 0.00001157, -0.00001794, 0.00002651, -0.00001746, 0.00000369, -0.00000847, 0.00001503, -0.00000633, -0.00001228, 0.00000548, 0.00000058, 0.00000899, -0.00001303, 0.00000508, -0.00000320, 0.00000028, 0.00000050, -0.00001205, 0.00000823, 0.00000415, -0.00000288, -0.00000013, 0.00000025, -0.00000247, -0.00000068, 0.00000733, 0.00000229, -0.00000263, 0.00000027, -0.00000500, -0.00000322, -0.00000075, 0.00000176, 0.00000097, -0.00000088, 0.00000392, 0.00000113, 0.00000264, -0.00000571, -0.00000224, 0.00000238, -0.00000074, -0.00000046, -0.00000187, -0.00000035, -0.00000166, 0.00000136, 0.00000339, -0.00000098, },
		{ -0.00000069, -0.00000049, -0.00000123, 0.00000313, 0.00000195, -0.00000182, -0.00000032, -0.00000182, -0.00000056, -0.00000075, 0.00000219, -0.00000146, -0.00000617, 0.00000235, 0.00000107, 0.00000390, -0.00000044, 0.00000046, 0.00000208, -0.00000075, -0.00000268, -0.00000555, 0.00000020, -0.00000216, 0.00000131, 0.00000754, 0.00000005, -0.00000289, 0.00000027, -0.00000011, -0.00000290, 0.00000344, 0.00000917, -0.00001123, -0.00000138, 0.00000198, -0.00000482, 0.00000649, -0.00001336, 0.00000807, 0.00000171, 0.00000413, -0.00000916, -0.00001058, 0.00001769, -0.00000999, 0.00000626, -0.00002046, 0.00002859, -0.00002031, 0.00001514, -0.00002426, 0.00003511, -0.00005395, 0.00007307, -0.00010239, 0.00011574, -0.00011070, 0.00006281, -0.00008616, 0.00007729, -0.00008660, 0.00000076, -0.00006093, -0.00054282, 0.00134543, -0.00036551, -0.00020460, 0.00008549, -0.00014830, 0.00011811, -0.00011630, 0.00008561, -0.00012553, 0.00011697, -0.00009312, 0.00005631, -0.00003651, 0.00002102, -0.00001362, 0.00000836, -0.00001584, 0.00002453, -0.00001453, 0.00000124, -0.00000696, 0.00001230, -0.00000230, -0.00001483, 0.00000626, -0.00000004, 0.00000937, -0.00001218, 0.00000332, -0.00000146, -0.00000136, 0.00000212, -0.00001246, 0.00000703, 0.00000490, -0.00000288, -0.00000010, 0.00000016, -0.00000202, -0.00000131, 0.00000695, 0.00000328, -0.00000296, 0.00000017, -0.00000437, -0.00000374, -0.00000076, 0.00000142, 0.00000145, -0.00000123, 0.00000382, 0.00000126, 0.00000284, -0.00000517, -0.00000298, 0.00000250, -0.00000070, -0.00000039, -0.00000187, -0.00000043, -0.00000145, 0.00000079, 0.00000356, },
		{ 0.00000364, -0.00000038, -0.00000047, -0.00000144, 0.00000279, 0.00000255, -0.00000191, -0.00000035, -0.00000172, -0.00000069, -0.00000072, 0.00000194, -0.00000068, -0.00000653, 0.00000197, 0.00000107, 0.00000379, 0.00000008, -0.00000006, 0.00000237, -0.00000074, -0.00000215, -0.00000598, -0.00000004, -0.00000157, 0.00000036, 0.00000758, 0.00000087, -0.00000325, 0.00000021, -0.00000003, -0.00000296, 0.00000279, 0.00000984, -0.00001005, -0.00000342, 0.00000362, -0.00000618, 0.00000743, -0.00001309, 0.00000659, 0.00000333, 0.00000224, -0.00000557, -0.00001493, 0.00002019, -0.00001149, 0.00000894, -0.00002347, 0.00003070, -0.00002286, 0.00001890, -0.00002965, 0.00004159, -0.00006102, 0.00007817, -0.00010244, 0.00010992, -0.00009821, 0.00004693, -0.00006698, 0.00005312, -0.00005290, -0.00004245, 0.00000575, -0.00061258, 0.00131857, -0.00026116, -0.00027775, 0.00012406, -0.00017419, 0.00013326, -0.00012623, 0.00009185, -0.00012759, 0.00011255, -0.00008453, 0.00004554, -0.00002700, 0.00001413, -0.00000890, 0.00000564, -0.00001408, 0.00002268, -0.00001171, -0.00000103, -0.00000550, 0.00000957, 0.00000140, -0.00001676, 0.00000646, -0.00000018, 0.00000926, -0.00001088, 0.00000132, 0.00000029, -0.00000283, 0.00000341, -0.00001246, 0.00000559, 0.00000567, -0.00000289, -0.00000002, 0.00000004, -0.00000157, -0.00000181, 0.00000644, 0.00000424, -0.00000312, -0.00000009, -0.00000368, -0.00000422, -0.00000078, 0.00000109, 0.00000188, -0.00000147, 0.00000363, 0.00000143, 0.00000297, -0.00000457, -0.00000368, 0.00000254, -0.00000062, -0.00000037, -0.00000181, -0.00000056, -0.00000121, 0.00000026, },
		{ -0.00000022, 0.00000364, -0.00000004, -0.00000044, -0.00000160, 0.00000238, 0.00000313, -0.00000193, -0.00000045, -0.00000157, -0.00000084, -0.00000067, 0.00000164, 0.00000008, -0.00000676, 0.00000150, 0.00000113, 0.00000358, 0.00000065, -0.00000057, 0.00000260, -0.00000072, -0.00000164, -0.00000627, -0.00000046, -0.00000089, -0.00000051, 0.00000745, 0.00000175, -0.00000353, 0.00000005, 0.00000011, -0.00000306, 0.00000222, 0.00001024, -0.00000857, -0.00000551, 0.00000507, -0.00000718, 0.00000782, -0.00001219, 0.00000457, 0.00000541, -0.00000013, -0.00000162, -0.00001926, 0.00002244, -0.00001291, 0.00001165, -0.00002642, 0.00003278, -0.00002546, 0.00002264, -0.00003472, 0.00004722, -0.00006643, 0.00008065, -0.00009924, 0.00010075, -0.00008277, 0.00002873, -0.00004593, 0.00002757, -0.00001878, -0.00008429, 0.00006675, -0.00066828, 0.00127460, -0.00014904, -0.00034904, 0.00015818, -0.00019554, 0.00014415, -0.00013227, 0.00009444, -0.00012613, 0.00010513, -0.00007387, 0.00003387, -0.00001759, 0.00000778, -0.00000484, 0.00000351, -0.00001268, 0.00002099, -0.00000902, -0.00000311, -0.00000413, 0.00000691, 0.00000468, -0.00001801, 0.00000610, 0.00000012, 0.00000872, -0.00000924, -0.00000080, 0.00000193, -0.00000406, 0.00000432, -0.00001204, 0.00000398, 0.00000642, -0.00000288, 0.00000007, -0.00000011, -0.00000114, -0.00000218, 0.00000580, 0.00000514, -0.00000311, -0.00000049, -0.00000297, -0.00000463, -0.00000084, 0.00000076, 0.00000223, -0.00000160, 0.00000335, 0.00000164, 0.00000306, -0.00000393, -0.00000431, 0.00000250, -0.00000050, -0.00000039, -0.00000170, -0.00000073, -0.00000096, },
		{ -0.00000070, -0.00000063, 0.00000354, 0.00000032, -0.00000040, -0.00000171, 0.00000191, 0.00000367, -0.00000186, -0.00000061, -0.00000138, -0.00000101, -0.00000061, 0.00000132, 0.00000079, -0.00000685, 0.00000092, 0.00000126, 0.00000330, 0.00000126, -0.00000105, 0.00000277, -0.00000066, -0.00000117, -0.00000642, -0.00000104, -0.00000015, -0.00000128, 0.00000714, 0.00000266, -0.00000370, -0.00000022, 0.00000033, -0.00000320, 0.00000174, 0.00001038, -0.00000684, -0.00000755, 0.00000622, -0.00000772, 0.00000761, -0.00001067, 0.00000203, 0.00000789, -0.00000288, 0.00000258, -0.00002344, 0.00002437, -0.00001423, 0.00001436, -0.00002924, 0.00003471, -0.00002798, 0.00002616, -0.00003918, 0.00005165, -0.00006988, 0.00008029, -0.00009282, 0.00008851, -0.00006488, 0.00000885, -0.00002378, 0.00000158, 0.00001463, -0.00012344, 0.00012070, -0.00070934, 0.00121469, -0.00003150, -0.00041639, 0.00018659, -0.00021166, 0.00015040, -0.00013431, 0.00009344, -0.00012138, 0.00009513, -0.00006170, 0.00002183, -0.00000869, 0.00000225, -0.00000160, 0.00000203, -0.00001166, 0.00001945, -0.00000648, -0.00000498, -0.00000287, 0.00000440, 0.00000747, -0.00001859, 0.00000520, 0.00000081, 0.00000783, -0.00000735, -0.00000292, 0.00000334, -0.00000496, 0.00000484, -0.00001125, 0.00000224, 0.00000709, -0.00000285, 0.00000016, -0.00000026, -0.00000077, -0.00000241, 0.00000507, 0.00000594, -0.00000292, -0.00000100, -0.00000227, -0.00000497, -0.00000093, 0.00000047, 0.00000251, -0.00000163, 0.00000299, 0.00000187, 0.00000310, -0.00000327, -0.00000486, 0.00000238, -0.00000035, -0.00000044, -0.00000156, -0.00000092, },
		{ -0.00000112, -0.00000046, -0.00000097, 0.00000337, 0.00000068, -0.00000035, -0.00000177, 0.00000139, 0.00000416, -0.00000169, -0.00000083, -0.00000116, -0.00000119, -0.00000054, 0.00000097, 0.00000145, -0.00000680, 0.00000025, 0.00000144, 0.00000295, 0.00000187, -0.00000147, 0.00000286, -0.00000056, -0.00000077, -0.00000641, -0.00000176, 0.00000062, -0.00000193, 0.00000667, 0.00000356, -0.00000373, -0.00000059, 0.00000060, -0.00000336, 0.00000132, 0.00001030, -0.00000498, -0.00000941, 0.00000696, -0.00000773, 0.00000676, -0.00000854, -0.00000095, 0.00001070, -0.00000592, 0.00000688, -0.00002733, 0.00002591, -0.00001538, 0.00001698, -0.00003184, 0.00003640, -0.00003026, 0.00002922, -0.00004275, 0.00005461, -0.00007111, 0.00007700, -0.00008332, 0.00007361, -0.00004513, -0.00001199, -0.00000133, -0.00002394, 0.00004628, -0.00015871, 0.00016654, -0.00073563, 0.00114042, 0.00008893, -0.00047780, 0.00020820, -0.00022204, 0.00015186, -0.00013238, 0.00008905, -0.00011372, 0.00008304, -0.00004858, 0.00000996, -0.00000070, -0.00000224, 0.00000072, 0.00000121, -0.00001098, 0.00001805, -0.00000409, -0.00000663, -0.00000175, 0.00000210, 0.00000973, -0.00001850, 0.00000385, 0.00000180, 0.00000668, -0.00000535, -0.00000491, 0.00000445, -0.00000549, 0.00000494, -0.00001013, 0.00000045, 0.00000766, -0.00000275, 0.00000023, -0.00000038, -0.00000045, -0.00000250, 0.00000427, 0.00000663, -0.00000256, -0.00000159, -0.00000160, -0.00000521, -0.00000106, 0.00000020, 0.00000271, -0.00000155, 0.00000256, 0.00000210, 0.00000313, -0.00000261, -0.00000531, 0.00000217, -0.00000016, -0.00000053, -0.00000138, },
		{ -0.00000118, -0.00000132, -0.00000025, -0.00000123, 0.00000313, 0.00000103, -0.00000029, -0.00000179, 0.00000083, 0.00000458, -0.00000142, -0.00000110, -0.00000092, -0.00000135, -0.00000048, 0.00000063, 0.00000202, -0.00000658, -0.00000051, 0.00000165, 0.00000256, 0.00000246, -0.00000180, 0.00000285, -0.00000040, -0.00000044, -0.00000623, -0.00000261, 0.00000137, -0.00000242, 0.00000606, 0.00000444, -0.00000362, -0.00000106, 0.00000091, -0.00000352, 0.00000096, 0.00001005, -0.00000307, -0.00001097, 0.00000720, -0.00000714, 0.00000528, -0.00000585, -0.00000429, 0.00001372, -0.00000913, 0.00001114, -0.00003081, 0.00002698, -0.00001634, 0.00001944, -0.00003411, 0.00003772, -0.00003214, 0.00003161, -0.00004518, 0.00005586, -0.00006999, 0.00007082, -0.00007106, 0.00005657, -0.00002419, -0.00003304, 0.00002062, -0.00004812, 0.00007523, -0.00018914, 0.00020353, -0.00074744, 0.00105372, 0.00020963, -0.00053139, 0.00022213, -0.00022638, 0.00014855, -0.00012671, 0.00008160, -0.00010360, 0.00006945, -0.00003511, -0.00000124, 0.00000607, -0.00000555, 0.00000211, 0.00000100, -0.00001059, 0.00001675, -0.00000186, -0.00000804, -0.00000079, 0.00000006, 0.00001141, -0.00001778, 0.00000211, 0.00000300, 0.00000537, -0.00000333, -0.00000667, 0.00000517, -0.00000562, 0.00000467, -0.00000875, -0.00000132, 0.00000808, -0.00000259, 0.00000028, -0.00000048, -0.00000020, -0.00000247, 0.00000343, 0.00000717, -0.00000202, -0.00000224, -0.00000100, -0.00000535, -0.00000123, -0.00000003, 0.00000282, -0.00000138, 0.00000210, 0.00000231, 0.00000314, -0.00000196, -0.00000567, 0.00000189, 0.00000005, -0.00000065, },
		{ -0.00000079, -0.00000098, -0.00000150, -0.00000008, -0.00000142, 0.00000282, 0.00000138, -0.00000022, -0.00000175, 0.00000025, 0.00000491, -0.00000104, -0.00000141, -0.00000069, -0.00000149, -0.00000042, 0.00000029, 0.00000250, -0.00000621, -0.00000133, 0.00000188, 0.00000217, 0.00000300, -0.00000204, 0.00000274, -0.00000019, -0.00000019, -0.00000590, -0.00000355, 0.00000207, -0.00000275, 0.00000531, 0.00000525, -0.00000335, -0.00000161, 0.00000123, -0.00000367, 0.00000062, 0.00000968, -0.00000122, -0.00001212, 0.00000686, -0.00000594, 0.00000320, -0.00000270, -0.00000787, 0.00001684, -0.00001238, 0.00001521, -0.00003375, 0.00002753, -0.00001706, 0.00002168, -0.00003595, 0.00003856, -0.00003347, 0.00003314, -0.00004627, 0.00005523, -0.00006645, 0.00006193, -0.00005645, 0.00003802, -0.00000281, -0.00005355, 0.00004130, -0.00007015, 0.00010067, -0.00021398, 0.00023126, -0.00074547, 0.00095678, 0.00032798, -0.00057551, 0.00022780, -0.00022466, 0.00014069, -0.00011765, 0.00007156, -0.00009157, 0.00005498, -0.00002190, -0.00001131, 0.00001137, -0.00000760, 0.00000257, 0.00000133, -0.00001041, 0.00001553, 0.00000022, -0.00000922, -0.00000002, -0.00000170, 0.00001252, -0.00001648, 0.00000008, 0.00000433, 0.00000401, -0.00000142, -0.00000811, 0.00000547, -0.00000537, 0.00000407, -0.00000718, -0.00000301, 0.00000832, -0.00000235, 0.00000028, -0.00000053, -0.00000003, -0.00000233, 0.00000259, 0.00000755, -0.00000134, -0.00000290, -0.00000049, -0.00000539, -0.00000144, -0.00000022, 0.00000284, -0.00000114, 0.00000161, 0.00000249, 0.00000317, -0.00000134, -0.00000592, 0.00000155, 0.00000029, },
		{ 0.00000053, -0.00000095, -0.00000077, -0.00000165, 0.00000004, -0.00000152, 0.00000247, 0.00000170, -0.00000015, -0.00000167, -0.00000033, 0.00000514, -0.00000058, -0.00000174, -0.00000046, -0.00000161, -0.00000039, -0.00000002, 0.00000286, -0.00000569, -0.00000220, 0.00000209, 0.00000179, 0.00000347, -0.00000215, 0.00000253, 0.00000008, -0.00000003, -0.00000542, -0.00000455, 0.00000268, -0.00000290, 0.00000444, 0.00000597, -0.00000293, -0.00000221, 0.00000154, -0.00000376, 0.00000027, 0.00000925, 0.00000047, -0.00001277, 0.00000591, -0.00000414, 0.00000058, 0.00000082, -0.00001158, 0.00001995, -0.00001554, 0.00001896, -0.00003605, 0.00002752, -0.00001752, 0.00002361, -0.00003726, 0.00003881, -0.00003410, 0.00003364, -0.00004587, 0.00005266, -0.00006057, 0.00005059, -0.00004002, 0.00001864, 0.00001829, -0.00007279, 0.00006002, -0.00008938, 0.00012196, -0.00023274, 0.00024969, -0.00073078, 0.00085203, 0.00044143, -0.00060878, 0.00022492, -0.00021709, 0.00012871, -0.00010572, 0.00005948, -0.00007827, 0.00004026, -0.00000947, -0.00001989, 0.00001504, -0.00000839, 0.00000219, 0.00000210, -0.00001037, 0.00001433, 0.00000216, -0.00001017, 0.00000057, -0.00000314, 0.00001307, -0.00001470, -0.00000212, 0.00000567, 0.00000269, 0.00000030, -0.00000916, 0.00000533, -0.00000475, 0.00000320, -0.00000551, -0.00000454, 0.00000835, -0.00000201, 0.00000023, -0.00000053, 0.00000007, -0.00000210, 0.00000177, 0.00000776, -0.00000053, -0.00000355, -0.00000009, -0.00000532, -0.00000168, -0.00000038, 0.00000279, -0.00000083, 0.00000113, 0.00000261, 0.00000320, -0.00000076, -0.00000608, 0.00000114, },
		{ 0.00000069, 0.00000077, -0.00000110, -0.00000058, -0.00000176, 0.00000010, -0.00000155, 0.00000207, 0.00000199, -0.00000006, -0.00000156, -0.00000090, 0.00000526, -0.00000002, -0.00000208, -0.00000026, -0.00000168, -0.00000039, -0.00000029, 0.00000311, -0.00000503, -0.00000309, 0.00000227, 0.00000145, 0.00000384, -0.00000213, 0.00000221, 0.00000040, 0.00000004, -0.00000481, -0.00000556, 0.00000316, -0.00000286, 0.00000349, 0.00000658, -0.00000236, -0.00000284, 0.00000178, -0.00000376, -0.00000012, 0.00000885, 0.00000191, -0.00001286, 0.00000433, -0.00000178, -0.00000248, 0.00000456, -0.00001528, 0.00002290, -0.00001846, 0.00002224, -0.00003763, 0.00002693, -0.00001767, 0.00002516, -0.00003796, 0.00003837, -0.00003391, 0.00003297, -0.00004389, 0.00004815, -0.00005251, 0.00003722, -0.00002237, -0.00000086, 0.00003837, -0.00009009, 0.00007619, -0.00010524, 0.00013866, -0.00024517, 0.00025905, -0.00070472, 0.00074199, 0.00054759, -0.00063016, 0.00021352, -0.00020411, 0.00011317, -0.00009153, 0.00004599, -0.00006431, 0.00002589, 0.00000167, -0.00002669, 0.00001701, -0.00000798, 0.00000110, 0.00000318, -0.00001037, 0.00001313, 0.00000396, -0.00001088, 0.00000095, -0.00000426, 0.00001309, -0.00001252, -0.00000439, 0.00000694, 0.00000150, 0.00000175, -0.00000978, 0.00000476, -0.00000382, 0.00000213, -0.00000382, -0.00000585, 0.00000815, -0.00000159, 0.00000014, -0.00000050, 0.00000011, -0.00000180, 0.00000100, 0.00000779, 0.00000038, -0.00000414, 0.00000018, -0.00000514, -0.00000194, -0.00000050, 0.00000267, -0.00000048, 0.00000067, 0.00000268, 0.00000325, -0.00000022, -0.00000613, },
		{ -0.00000610, 0.00000021, 0.00000100, -0.00000125, -0.00000041, -0.00000183, 0.00000012, -0.00000152, 0.00000166, 0.00000225, 0.00000003, -0.00000141, -0.00000144, 0.00000526, 0.00000061, -0.00000241, -0.00000010, -0.00000170, -0.00000042, -0.00000053, 0.00000323, -0.00000425, -0.00000397, 0.00000238, 0.00000117, 0.00000409, -0.00000199, 0.00000180, 0.00000077, 0.00000004, -0.00000410, -0.00000656, 0.00000349, -0.00000264, 0.00000248, 0.00000706, -0.00000166, -0.00000346, 0.00000192, -0.00000364, -0.00000061, 0.00000851, 0.00000301, -0.00001233, 0.00000214, 0.00000106, -0.00000587, 0.00000839, -0.00001881, 0.00002557, -0.00002104, 0.00002494, -0.00003842, 0.00002573, -0.00001749, 0.00002626, -0.00003795, 0.00003716, -0.00003282, 0.00003107, -0.00004031, 0.00004180, -0.00004254, 0.00002228, -0.00000416, -0.00001977, 0.00005675, -0.00010488, 0.00008932, -0.00011737, 0.00015052, -0.00025127, 0.00025991, -0.00066892, 0.00062924, 0.00064433, -0.00063896, 0.00019395, -0.00018638, 0.00009480, -0.00007577, 0.00003175, -0.00005034, 0.00001244, 0.00001114, -0.00003153, 0.00001730, -0.00000651, -0.00000054, 0.00000442, -0.00001032, 0.00001189, 0.00000562, -0.00001136, 0.00000114, -0.00000505, 0.00001262, -0.00001005, -0.00000662, 0.00000804, 0.00000051, 0.00000287, -0.00000995, 0.00000378, -0.00000265, 0.00000096, -0.00000221, -0.00000690, 0.00000772, -0.00000110, 0.00000001, -0.00000042, 0.00000009, -0.00000145, 0.00000032, 0.00000765, 0.00000136, -0.00000465, 0.00000032, -0.00000489, -0.00000222, -0.00000058, 0.00000249, -0.00000010, 0.00000025, 0.00000268, 0.00000332, 0.00000029, },
		{ 0.00000076, -0.00000599, -0.00000029, 0.00000120, -0.00000138, -0.00000028, -0.00000184, 0.00000007, -0.00000142, 0.00000123, 0.00000247, 0.00000013, -0.00000123, -0.00000194, 0.00000514, 0.00000130, -0.00000270, 0.00000000, -0.00000167, -0.00000048, -0.00000071, 0.00000324, -0.00000337, -0.00000481, 0.00000240, 0.00000098, 0.00000421, -0.00000171, 0.00000130, 0.00000117, -0.00000002, -0.00000331, -0.00000748, 0.00000363, -0.00000224, 0.00000143, 0.00000740, -0.00000087, -0.00000400, 0.00000192, -0.00000336, -0.00000121, 0.00000831, 0.00000371, -0.00001116, -0.00000060, 0.00000427, -0.00000943, 0.00001214, -0.00002204, 0.00002784, -0.00002316, 0.00002696, -0.00003837, 0.00002394, -0.00001697, 0.00002684, -0.00003716, 0.00003512, -0.00003077, 0.00002789, -0.00003520, 0.00003380, -0.00003101, 0.00000633, 0.00001392, -0.00003738, 0.00007285, -0.00011671, 0.00009911, -0.00012556, 0.00015750, -0.00025130, 0.00025306, -0.00062514, 0.00051627, 0.00072981, -0.00063488, 0.00016687, -0.00016477, 0.00007441, -0.00005919, 0.00001742, -0.00003695, 0.00000039, 0.00001863, -0.00003433, 0.00001602, -0.00000417, -0.00000255, 0.00000568, -0.00001013, 0.00001058, 0.00000715, -0.00001161, 0.00000112, -0.00000553, 0.00001173, -0.00000740, -0.00000869, 0.00000891, -0.00000021, 0.00000361, -0.00000967, 0.00000246, -0.00000132, -0.00000023, -0.00000074, -0.00000764, 0.00000707, -0.00000054, -0.00000015, -0.00000031, 0.00000003, -0.00000108, -0.00000027, 0.00000734, 0.00000238, -0.00000505, 0.00000032, -0.00000455, -0.00000251, -0.00000064, 0.00000226, 0.00000028, -0.00000011, 0.00000261, 0.00000340, },
		{ 0.00000349, 0.00000119, -0.00000580, -0.00000080, 0.00000138, -0.00000149, -0.00000019, -0.00000180, -0.00000002, -0.00000129, 0.00000081, 0.00000264, 0.00000024, -0.00000104, -0.00000238, 0.00000489, 0.00000203, -0.00000294, 0.00000004, -0.00000159, -0.00000057, -0.00000086, 0.00000313, -0.00000242, -0.00000557, 0.00000231, 0.00000089, 0.00000420, -0.00000131, 0.00000074, 0.00000158, -0.00000012, -0.00000248, -0.00000830, 0.00000357, -0.00000168, 0.00000036, 0.00000761, -0.00000002, -0.00000444, 0.00000174, -0.00000290, -0.00000197, 0.00000829, 0.00000397, -0.00000938, -0.00000381, 0.00000773, -0.00001302, 0.00001567, -0.00002483, 0.00002959, -0.00002471, 0.00002824, -0.00003746, 0.00002158, -0.00001610, 0.00002686, -0.00003556, 0.00003222, -0.00002774, 0.00002348, -0.00002867, 0.00002440, -0.00001835, -0.00001003, 0.00003120, -0.00005309, 0.00008617, -0.00012524, 0.00010537, -0.00012976, 0.00015976, -0.00024571, 0.00023950, -0.00057529, 0.00040549, 0.00080254, -0.00061804, 0.00013319, -0.00014027, 0.00005290, -0.00004255, 0.00000365, -0.00002467, -0.00000987, 0.00002395, -0.00003512, 0.00001335, -0.00000119, -0.00000473, 0.00000681, -0.00000974, 0.00000919, 0.00000854, -0.00001164, 0.00000092, -0.00000573, 0.00001050, -0.00000469, -0.00001053, 0.00000947, -0.00000064, 0.00000398, -0.00000898, 0.00000086, 0.00000007, -0.00000135, 0.00000051, -0.00000805, 0.00000622, 0.00000005, -0.00000033, -0.00000019, -0.00000004, -0.00000071, -0.00000074, 0.00000687, 0.00000341, -0.00000531, 0.00000018, -0.00000417, -0.00000278, -0.00000068, 0.00000199, 0.00000064, -0.00000040, 0.00000246, },
		{ 0.00000226, 0.00000358, 0.00000160, -0.00000556, -0.00000129, 0.00000153, -0.00000157, -0.00000015, -0.00000171, -0.00000015, -0.00000112, 0.00000041, 0.00000276, 0.00000035, -0.00000083, -0.00000275, 0.00000453, 0.00000278, -0.00000311, 0.00000001, -0.00000145, -0.00000069, -0.00000095, 0.00000293, -0.00000144, -0.00000622, 0.00000208, 0.00000091, 0.00000406, -0.00000080, 0.00000013, 0.00000199, -0.00000025, -0.00000165, -0.00000897, 0.00000329, -0.00000096, -0.00000069, 0.00000769, 0.00000086, -0.00000471, 0.00000133, -0.00000221, -0.00000290, 0.00000847, 0.00000377, -0.00000703, -0.00000738, 0.00001128, -0.00001647, 0.00001881, -0.00002706, 0.00003075, -0.00002564, 0.00002870, -0.00003571, 0.00001868, -0.00001488, 0.00002628, -0.00003311, 0.00002847, -0.00002377, 0.00001791, -0.00002092, 0.00001394, -0.00000504, -0.00002620, 0.00004705, -0.00006635, 0.00009635, -0.00013032, 0.00010808, -0.00013010, 0.00015761, -0.00023512, 0.00022039, -0.00052127, 0.00029907, 0.00086145, -0.00058893, 0.00009411, -0.00011400, 0.00003117, -0.00002656, -0.00000901, -0.00001395, -0.00001805, 0.00002704, -0.00003405, 0.00000956, 0.00000217, -0.00000686, 0.00000769, -0.00000909, 0.00000773, 0.00000977, -0.00001144, 0.00000055, -0.00000567, 0.00000901, -0.00000202, -0.00001205, 0.00000970, -0.00000074, 0.00000397, -0.00000792, -0.00000093, 0.00000144, -0.00000232, 0.00000150, -0.00000814, 0.00000519, 0.00000065, -0.00000050, -0.00000007, -0.00000012, -0.00000037, -0.00000109, 0.00000627, 0.00000440, -0.00000543, -0.00000008, -0.00000375, -0.00000303, -0.00000071, 0.00000170, 0.00000098, -0.00000062, },
		{ -0.00000076, 0.00000199, 0.00000366, 0.00000198, -0.00000526, -0.00000177, 0.00000163, -0.00000162, -0.00000015, -0.00000156, -0.00000032, -0.00000093, 0.00000004, 0.00000284, 0.00000046, -0.00000062, -0.00000305, 0.00000406, 0.00000353, -0.00000320, -0.00000009, -0.00000126, -0.00000084, -0.00000101, 0.00000264, -0.00000044, -0.00000672, 0.00000171, 0.00000104, 0.00000380, -0.00000022, -0.00000048, 0.00000237, -0.00000038, -0.00000085, -0.00000947, 0.00000279, -0.00000010, -0.00000172, 0.00000765, 0.00000171, -0.00000477, 0.00000068, -0.00000131, -0.00000400, 0.00000888, 0.00000311, -0.00000418, -0.00001117, 0.00001477, -0.00001963, 0.00002142, -0.00002863, 0.00003125, -0.00002588, 0.00002834, -0.00003314, 0.00001531, -0.00001333, 0.00002508, -0.00002983, 0.00002393, -0.00001894, 0.00001134, -0.00001220, 0.00000280, 0.00000842, -0.00004157, 0.00006088, -0.00007676, 0.00010318, -0.00013191, 0.00010738, -0.00012686, 0.00015154, -0.00022033, 0.00019698, -0.00046496, 0.00019894, 0.00090587, -0.00054842, 0.00005100, -0.00008713, 0.00001010, -0.00001188, -0.00002005, -0.00000514, -0.00002397, 0.00002792, -0.00003131, 0.00000494, 0.00000564, -0.00000876, 0.00000820, -0.00000815, 0.00000621, 0.00001084, -0.00001103, 0.00000003, -0.00000540, 0.00000734, 0.00000051, -0.00001318, 0.00000956, -0.00000052, 0.00000361, -0.00000656, -0.00000281, 0.00000270, -0.00000309, 0.00000220, -0.00000792, 0.00000404, 0.00000123, -0.00000066, 0.00000003, -0.00000019, -0.00000006, -0.00000132, 0.00000557, 0.00000533, -0.00000538, -0.00000046, -0.00000331, -0.00000324, -0.00000073, 0.00000140, 0.00000128, },
		{ 0.00000153, -0.00000082, 0.00000169, 0.00000371, 0.00000235, -0.00000492, -0.00000221, 0.00000168, -0.00000163, -0.00000021, -0.00000137, -0.00000051, -0.00000074, -0.00000030, 0.00000286, 0.00000058, -0.00000042, -0.00000326, 0.00000349, 0.00000425, -0.00000318, -0.00000027, -0.00000103, -0.00000100, -0.00000104, 0.00000229, 0.00000052, -0.00000706, 0.00000118, 0.00000128, 0.00000343, 0.00000042, -0.00000108, 0.00000269, -0.00000049, -0.00000011, -0.00000975, 0.00000206, 0.00000087, -0.00000270, 0.00000750, 0.00000249, -0.00000459, -0.00000023, -0.00000018, -0.00000528, 0.00000951, 0.00000203, -0.00000093, -0.00001503, 0.00001803, -0.00002233, 0.00002340, -0.00002947, 0.00003103, -0.00002541, 0.00002716, -0.00002981, 0.00001153, -0.00001147, 0.00002325, -0.00002575, 0.00001866, -0.00001336, 0.00000396, -0.00000282, -0.00000861, 0.00002151, -0.00005560, 0.00007223, -0.00008403, 0.00010658, -0.00013015, 0.00010355, -0.00012046, 0.00014214, -0.00020219, 0.00017056, -0.00040814, 0.00010669, 0.00093558, -0.00049772, 0.00000539, -0.00006082, -0.00000947, 0.00000092, -0.00002911, 0.00000154, -0.00002760, 0.00002675, -0.00002721, -0.00000018, 0.00000895, -0.00001027, 0.00000828, -0.00000693, 0.00000466, 0.00001171, -0.00001041, -0.00000062, -0.00000497, 0.00000557, 0.00000281, -0.00001388, 0.00000907, 0.00000000, 0.00000294, -0.00000498, -0.00000469, 0.00000376, -0.00000358, 0.00000259, -0.00000741, 0.00000281, 0.00000175, -0.00000077, 0.00000010, -0.00000023, 0.00000020, -0.00000142, 0.00000478, 0.00000616, -0.00000516, -0.00000092, -0.00000289, -0.00000342, -0.00000074, 0.00000110, },
		{ 0.00000080, 0.00000172, -0.00000080, 0.00000136, 0.00000373, 0.00000271, -0.00000456, -0.00000262, 0.00000169, -0.00000161, -0.00000031, -0.00000115, -0.00000071, -0.00000056, -0.00000059, 0.00000283, 0.00000070, -0.00000023, -0.00000338, 0.00000284, 0.00000491, -0.00000306, -0.00000051, -0.00000077, -0.00000116, -0.00000104, 0.00000189, 0.00000143, -0.00000721, 0.00000050, 0.00000160, 0.00000299, 0.00000108, -0.00000164, 0.00000294, -0.00000055, 0.00000053, -0.00000980, 0.00000111, 0.00000192, -0.00000362, 0.00000727, 0.00000316, -0.00000412, -0.00000142, 0.00000115, -0.00000672, 0.00001035, 0.00000058, 0.00000259, -0.00001879, 0.00002089, -0.00002445, 0.00002463, -0.00002952, 0.00003009, -0.00002423, 0.00002519, -0.00002580, 0.00000743, -0.00000934, 0.00002083, -0.00002094, 0.00001279, -0.00000720, -0.00000401, 0.00000690, -0.00001986, 0.00003374, -0.00006778, 0.00008072, -0.00008802, 0.00010664, -0.00012529, 0.00009698, -0.00011139, 0.00013010, -0.00018163, 0.00014239, -0.00035242, 0.00002360, 0.00095077, -0.00043832, -0.00004114, -0.00003621, -0.00002684, 0.00001138, -0.00003591, 0.00000599, -0.00002899, 0.00002377, -0.00002210, -0.00000545, 0.00001186, -0.00001125, 0.00000788, -0.00000543, 0.00000312, 0.00001237, -0.00000960, -0.00000136, -0.00000443, 0.00000380, 0.00000480, -0.00001413, 0.00000823, 0.00000078, 0.00000204, -0.00000327, -0.00000647, 0.00000454, -0.00000378, 0.00000269, -0.00000667, 0.00000156, 0.00000217, -0.00000083, 0.00000012, -0.00000025, 0.00000041, -0.00000140, 0.00000394, 0.00000686, -0.00000478, -0.00000144, -0.00000250, -0.00000355, -0.00000074, },
		{ -0.00000074, 0.00000051, 0.00000185, -0.00000072, 0.00000101, 0.00000371, 0.00000306, -0.00000416, -0.00000299, 0.00000165, -0.00000155, -0.00000046, -0.00000089, -0.00000090, -0.00000040, -0.00000083, 0.00000276, 0.00000082, -0.00000005, -0.00000342, 0.00000213, 0.00000551, -0.00000282, -0.00000082, -0.00000049, -0.00000131, -0.00000103, 0.00000147, 0.00000225, -0.00000716, -0.00000031, 0.00000199, 0.00000250, 0.00000172, -0.00000211, 0.00000308, -0.00000053, 0.00000105, -0.00000961, -0.00000004, 0.00000303, -0.00000447, 0.00000699, 0.00000366, -0.00000335, -0.00000286, 0.00000267, -0.00000827, 0.00001135, -0.00000116, 0.00000624, -0.00002227, 0.00002321, -0.00002586, 0.00002506, -0.00002877, 0.00002845, -0.00002238, 0.00002249, -0.00002121, 0.00000311, -0.00000698, 0.00001785, -0.00001550, 0.00000647, -0.00000065, -0.00001229, 0.00001659, -0.00003053, 0.00004467, -0.00007771, 0.00008613, -0.00008872, 0.00010357, -0.00011771, 0.00008816, -0.00010026, 0.00011615, -0.00015958, 0.00011368, -0.00029921, -0.00004945, 0.00095201, -0.00037193, -0.00008694, -0.00001431, -0.00004141, 0.00001920, -0.00004033, 0.00000820, -0.00002831, 0.00001929, -0.00001635, -0.00001054, 0.00001416, -0.00001160, 0.00000699, -0.00000371, 0.00000165, 0.00001280, -0.00000862, -0.00000216, -0.00000383, 0.00000209, 0.00000643, -0.00001394, 0.00000709, 0.00000177, 0.00000096, -0.00000151, -0.00000806, 0.00000501, -0.00000368, 0.00000250, -0.00000574, 0.00000033, 0.00000248, -0.00000082, 0.00000010, -0.00000022, 0.00000056, -0.00000129, 0.00000308, 0.00000742, -0.00000424, -0.00000200, -0.00000216, -0.00000363, },
		{ -0.00000367, -0.00000074, 0.00000024, 0.00000191, -0.00000059, 0.00000067, 0.00000363, 0.00000340, -0.00000375, -0.00000331, 0.00000156, -0.00000146, -0.00000065, -0.00000063, -0.00000108, -0.00000026, -0.00000102, 0.00000265, 0.00000094, 0.00000010, -0.00000337, 0.00000138, 0.00000601, -0.00000246, -0.00000117, -0.00000021, -0.00000144, -0.00000102, 0.00000104, 0.00000295, -0.00000690, -0.00000126, 0.00000243, 0.00000200, 0.00000232, -0.00000247, 0.00000311, -0.00000043, 0.00000141, -0.00000916, -0.00000137, 0.00000416, -0.00000524, 0.00000667, 0.00000398, -0.00000227, -0.00000453, 0.00000433, -0.00000990, 0.00001245, -0.00000310, 0.00000986, -0.00002532, 0.00002485, -0.00002649, 0.00002465, -0.00002724, 0.00002614, -0.00001991, 0.00001913, -0.00001615, -0.00000134, -0.00000444, 0.00001437, -0.00000955, -0.00000013, 0.00000609, -0.00002062, 0.00002590, -0.00004021, 0.00005389, -0.00008508, 0.00008834, -0.00008629, 0.00009771, -0.00010791, 0.00007765, -0.00008769, 0.00010102, -0.00013695, 0.00008551, -0.00024968, -0.00011192, 0.00094025, -0.00030045, -0.00013047, 0.00000399, -0.00005275, 0.00002420, -0.00004238, 0.00000832, -0.00002583, 0.00001369, -0.00001036, -0.00001515, 0.00001567, -0.00001128, 0.00000564, -0.00000184, 0.00000030, 0.00001299, -0.00000748, -0.00000298, -0.00000322, 0.00000051, 0.00000766, -0.00001332, 0.00000570, 0.00000291, -0.00000021, 0.00000019, -0.00000939, 0.00000511, -0.00000327, 0.00000208, -0.00000469, -0.00000080, 0.00000264, -0.00000073, 0.00000001, -0.00000015, 0.00000065, -0.00000109, 0.00000224, 0.00000782, -0.00000356, -0.00000258, -0.00000189, },
		{ -0.00000169, -0.00000366, -0.00000072, -0.00000001, 0.00000192, -0.00000041, 0.00000035, 0.00000349, 0.00000373, -0.00000331, -0.00000358, 0.00000143, -0.00000134, -0.00000086, -0.00000036, -0.00000124, -0.00000016, -0.00000115, 0.00000251, 0.00000105, 0.00000022, -0.00000324, 0.00000060, 0.00000639, -0.00000199, -0.00000155, 0.00000005, -0.00000154, -0.00000102, 0.00000062, 0.00000352, -0.00000644, -0.00000229, 0.00000286, 0.00000152, 0.00000282, -0.00000270, 0.00000300, -0.00000022, 0.00000161, -0.00000846, -0.00000285, 0.00000530, -0.00000593, 0.00000636, 0.00000406, -0.00000089, -0.00000639, 0.00000606, -0.00001154, 0.00001360, -0.00000514, 0.00001331, -0.00002778, 0.00002571, -0.00002629, 0.00002340, -0.00002496, 0.00002323, -0.00001689, 0.00001523, -0.00001077, -0.00000580, -0.00000179, 0.00001047, -0.00000324, -0.00000681, 0.00001277, -0.00002869, 0.00003450, -0.00004857, 0.00006110, -0.00008971, 0.00008740, -0.00008100, 0.00008951, -0.00009642, 0.00006604, -0.00007431, 0.00008545, -0.00011454, 0.00005882, -0.00020473, -0.00016366, 0.00091674, -0.00022584, -0.00017025, 0.00001800, -0.00006060, 0.00002636, -0.00004216, 0.00000657, -0.00002188, 0.00000737, -0.00000452, -0.00001900, 0.00001629, -0.00001028, 0.00000388, 0.00000012, -0.00000087, 0.00001291, -0.00000621, -0.00000379, -0.00000265, -0.00000089, 0.00000847, -0.00001231, 0.00000413, 0.00000414, -0.00000140, 0.00000177, -0.00001040, 0.00000485, -0.00000259, 0.00000148, -0.00000360, -0.00000180, 0.00000265, -0.00000057, -0.00000012, -0.00000005, 0.00000071, -0.00000084, 0.00000144, 0.00000805, -0.00000275, -0.00000313, },
		{ -0.00000364, -0.00000159, -0.00000361, -0.00000070, -0.00000024, 0.00000187, -0.00000020, 0.00000006, 0.00000329, 0.00000404, -0.00000286, -0.00000380, 0.00000126, -0.00000119, -0.00000110, -0.00000010, -0.00000136, -0.00000011, -0.00000123, 0.00000233, 0.00000116, 0.00000031, -0.00000304, -0.00000017, 0.00000665, -0.00000141, -0.00000195, 0.00000029, -0.00000160, -0.00000104, 0.00000023, 0.00000395, -0.00000578, -0.00000338, 0.00000327, 0.00000110, 0.00000321, -0.00000277, 0.00000274, 0.00000010, 0.00000163, -0.00000752, -0.00000445, 0.00000639, -0.00000653, 0.00000607, 0.00000391, 0.00000077, -0.00000839, 0.00000781, -0.00001313, 0.00001472, -0.00000715, 0.00001644, -0.00002952, 0.00002572, -0.00002523, 0.00002136, -0.00002202, 0.00001981, -0.00001341, 0.00001089, -0.00000519, -0.00001016, 0.00000090, 0.00000625, 0.00000326, -0.00001337, 0.00001917, -0.00003623, 0.00004207, -0.00005529, 0.00006608, -0.00009153, 0.00008347, -0.00007326, 0.00007950, -0.00008384, 0.00005393, -0.00006074, 0.00007009, -0.00009310, 0.00003440, -0.00016500, -0.00020484, 0.00088297, -0.00015009, -0.00020498, 0.00002722, -0.00006488, 0.00002581, -0.00003992, 0.00000326, -0.00001685, 0.00000075, 0.00000083, -0.00002188, 0.00001595, -0.00000865, 0.00000181, 0.00000205, -0.00000180, 0.00001258, -0.00000485, -0.00000454, -0.00000216, -0.00000206, 0.00000887, -0.00001096, 0.00000244, 0.00000537, -0.00000254, 0.00000316, -0.00001105, 0.00000422, -0.00000168, 0.00000077, -0.00000251, -0.00000261, 0.00000249, -0.00000034, -0.00000031, 0.00000008, 0.00000073, -0.00000054, 0.00000070, 0.00000809, -0.00000184, },
		{ -0.00000086, -0.00000408, -0.00000157, -0.00000354, -0.00000065, -0.00000044, 0.00000177, 0.00000003, -0.00000018, 0.00000304, 0.00000434, -0.00000239, -0.00000397, 0.00000106, -0.00000103, -0.00000135, 0.00000013, -0.00000143, -0.00000009, -0.00000127, 0.00000214, 0.00000127, 0.00000037, -0.00000278, -0.00000092, 0.00000677, -0.00000074, -0.00000233, 0.00000048, -0.00000161, -0.00000108, -0.00000012, 0.00000423, -0.00000496, -0.00000449, 0.00000360, 0.00000078, 0.00000345, -0.00000267, 0.00000233, 0.00000054, 0.00000147, -0.00000636, -0.00000612, 0.00000743, -0.00000705, 0.00000583, 0.00000351, 0.00000267, -0.00001045, 0.00000950, -0.00001458, 0.00001573, -0.00000902, 0.00001909, -0.00003044, 0.00002484, -0.00002334, 0.00001858, -0.00001852, 0.00001599, -0.00000960, 0.00000624, 0.00000044, -0.00001431, 0.00000356, 0.00000181, 0.00000978, -0.00001958, 0.00002505, -0.00004298, 0.00004834, -0.00006016, 0.00006870, -0.00009058, 0.00007685, -0.00006354, 0.00006826, -0.00007077, 0.00004190, -0.00004754, 0.00005551, -0.00007321, 0.00001281, -0.00013085, -0.00023594, 0.00084060, -0.00007514, -0.00023358, 0.00003140, -0.00006572, 0.00002282, -0.00003599, -0.00000125, -0.00001114, -0.00000576, 0.00000537, -0.00002364, 0.00001467, -0.00000647, -0.00000045, 0.00000386, -0.00000244, 0.00001200, -0.00000344, -0.00000520, -0.00000180, -0.00000298, 0.00000887, -0.00000936, 0.00000071, 0.00000655, -0.00000357, 0.00000429, -0.00001132, 0.00000325, -0.00000060, 0.00000000, -0.00000151, -0.00000322, 0.00000216, -0.00000004, -0.00000054, 0.00000022, 0.00000073, -0.00000023, 0.00000006, 0.00000797, },
		{ 0.00000767, 0.00000016, -0.00000443, -0.00000164, -0.00000345, -0.00000059, -0.00000062, 0.00000163, 0.00000026, -0.00000036, 0.00000272, 0.00000460, -0.00000190, -0.00000409, 0.00000082, -0.00000085, -0.00000160, 0.00000034, -0.00000146, -0.00000012, -0.00000126, 0.00000192, 0.00000138, 0.00000039, -0.00000246, -0.00000163, 0.00000674, 0.00000001, -0.00000269, 0.00000061, -0.00000157, -0.00000114, -0.00000041, 0.00000435, -0.00000400, -0.00000558, 0.00000381, 0.00000058, 0.00000352, -0.00000239, 0.00000176, 0.00000110, 0.00000114, -0.00000501, -0.00000782, 0.00000837, -0.00000749, 0.00000564, 0.00000288, 0.00000475, -0.00001250, 0.00001104, -0.00001583, 0.00001654, -0.00001065, 0.00002116, -0.00003046, 0.00002306, -0.00002067, 0.00001517, -0.00001458, 0.00001190, -0.00000556, 0.00000143, 0.00000598, -0.00001814, 0.00000611, -0.00000272, 0.00001613, -0.00002525, 0.00003020, -0.00004873, 0.00005310, -0.00006304, 0.00006897, -0.00008705, 0.00006793, -0.00005241, 0.00005638, -0.00005778, 0.00003048, -0.00003520, 0.00004221, -0.00005534, -0.00000554, -0.00010240, -0.00025769, 0.00079143, -0.00000279, -0.00025521, 0.00003050, -0.00006338, 0.00001774, -0.00003076, -0.00000655, -0.00000516, -0.00001179, 0.00000885, -0.00002422, 0.00001250, -0.00000387, -0.00000279, 0.00000547, -0.00000276, 0.00001118, -0.00000201, -0.00000573, -0.00000158, -0.00000364, 0.00000850, -0.00000756, -0.00000098, 0.00000762, -0.00000441, 0.00000512, -0.00001120, 0.00000200, 0.00000057, -0.00000074, -0.00000063, -0.00000359, 0.00000168, 0.00000029, -0.00000079, 0.00000036, 0.00000074, 0.00000007, -0.00000048, },
		{ -0.00000089, 0.00000722, 0.00000119, -0.00000466, -0.00000180, -0.00000336, -0.00000051, -0.00000077, 0.00000144, 0.00000048, -0.00000049, 0.00000237, 0.00000482, -0.00000139, -0.00000417, 0.00000056, -0.00000066, -0.00000185, 0.00000050, -0.00000143, -0.00000019, -0.00000122, 0.00000170, 0.00000148, 0.00000038, -0.00000211, -0.00000228, 0.00000657, 0.00000082, -0.00000299, 0.00000068, -0.00000147, -0.00000123, -0.00000066, 0.00000433, -0.00000293, -0.00000659, 0.00000387, 0.00000054, 0.00000341, -0.00000192, 0.00000106, 0.00000176, 0.00000064, -0.00000350, -0.00000952, 0.00000919, -0.00000784, 0.00000553, 0.00000202, 0.00000697, -0.00001445, 0.00001236, -0.00001681, 0.00001708, -0.00001194, 0.00002254, -0.00002955, 0.00002044, -0.00001733, 0.00001126, -0.00001033, 0.00000766, -0.00000143, -0.00000342, 0.00001129, -0.00002157, 0.00000850, -0.00000722, 0.00002213, -0.00003019, 0.00003446, -0.00005329, 0.00005621, -0.00006388, 0.00006695, -0.00008121, 0.00005719, -0.00004046, 0.00004447, -0.00004542, 0.00002012, -0.00002410, 0.00003052, -0.00003979, -0.00002047, -0.00007951, -0.00027104, 0.00073728, 0.00006532, -0.00026931, 0.00002473, -0.00005829, 0.00001105, -0.00002466, -0.00001221, 0.00000068, -0.00001702, 0.00001111, -0.00002360, 0.00000956, -0.00000100, -0.00000508, 0.00000677, -0.00000273, 0.00001016, -0.00000061, -0.00000610, -0.00000153, -0.00000404, 0.00000783, -0.00000567, -0.00000256, 0.00000851, -0.00000505, 0.00000564, -0.00001072, 0.00000051, 0.00000177, -0.00000140, 0.00000006, -0.00000372, 0.00000107, 0.00000065, -0.00000105, 0.00000048, 0.00000075, 0.00000035, },
		{ 0.00000058, -0.00000119, 0.00000664, 0.00000220, -0.00000477, -0.00000203, -0.00000329, -0.00000040, -0.00000088, 0.00000123, 0.00000068, -0.00000055, 0.00000197, 0.00000499, -0.00000085, -0.00000421, 0.00000029, -0.00000048, -0.00000208, 0.00000062, -0.00000135, -0.00000030, -0.00000115, 0.00000148, 0.00000157, 0.00000035, -0.00000173, -0.00000285, 0.00000626, 0.00000166, -0.00000322, 0.00000066, -0.00000131, -0.00000134, -0.00000086, 0.00000418, -0.00000180, -0.00000747, 0.00000375, 0.00000067, 0.00000311, -0.00000129, 0.00000023, 0.00000252, 0.00000001, -0.00000188, -0.00001114, 0.00000987, -0.00000811, 0.00000549, 0.00000098, 0.00000922, -0.00001620, 0.00001337, -0.00001743, 0.00001728, -0.00001280, 0.00002316, -0.00002770, 0.00001703, -0.00001343, 0.00000699, -0.00000593, 0.00000341, 0.00000269, -0.00000818, 0.00001623, -0.00002450, 0.00001063, -0.00001157, 0.00002759, -0.00003422, 0.00003766, -0.00005654, 0.00005760, -0.00006271, 0.00006284, -0.00007341, 0.00004517, -0.00002829, 0.00003306, -0.00003415, 0.00001118, -0.00001454, 0.00002070, -0.00002673, -0.00003194, -0.00006187, -0.00027707, 0.00067996, 0.00012777, -0.00027562, 0.00001451, -0.00005101, 0.00000326, -0.00001818, -0.00001785, 0.00000603, -0.00002119, 0.00001204, -0.00002186, 0.00000602, 0.00000199, -0.00000717, 0.00000770, -0.00000234, 0.00000897, 0.00000070, -0.00000627, -0.00000166, -0.00000419, 0.00000691, -0.00000375, -0.00000397, 0.00000920, -0.00000544, 0.00000584, -0.00000991, -0.00000114, 0.00000292, -0.00000192, 0.00000055, -0.00000362, 0.00000036, 0.00000101, -0.00000130, 0.00000057, 0.00000080, },
		{ 0.00000090, 0.00000077, -0.00000135, 0.00000596, 0.00000315, -0.00000474, -0.00000231, -0.00000324, -0.00000026, -0.00000097, 0.00000099, 0.00000085, -0.00000055, 0.00000155, 0.00000510, -0.00000029, -0.00000420, 0.00000000, -0.00000030, -0.00000228, 0.00000068, -0.00000122, -0.00000044, -0.00000106, 0.00000126, 0.00000166, 0.00000028, -0.00000135, -0.00000332, 0.00000580, 0.00000251, -0.00000336, 0.00000057, -0.00000111, -0.00000147, -0.00000101, 0.00000393, -0.00000065, -0.00000819, 0.00000342, 0.00000098, 0.00000262, -0.00000049, -0.00000070, 0.00000335, -0.00000073, -0.00000019, -0.00001266, 0.00001038, -0.00000830, 0.00000552, -0.00000021, 0.00001144, -0.00001767, 0.00001400, -0.00001765, 0.00001710, -0.00001317, 0.00002299, -0.00002496, 0.00001297, -0.00000911, 0.00000252, -0.00000153, -0.00000072, 0.00000665, -0.00001271, 0.00002070, -0.00002686, 0.00001246, -0.00001566, 0.00003237, -0.00003721, 0.00003972, -0.00005840, 0.00005726, -0.00005966, 0.00005691, -0.00006409, 0.00003244, -0.00001648, 0.00002264, -0.00002433, 0.00000391, -0.00000667, 0.00001285, -0.00001618, -0.00004010, -0.00004900, -0.00027697, 0.00062117, 0.00018342, -0.00027414, 0.00000047, -0.00004218, -0.00000508, -0.00001174, -0.00002308, 0.00001059, -0.00002412, 0.00001163, -0.00001911, 0.00000205, 0.00000491, -0.00000896, 0.00000819, -0.00000161, 0.00000766, 0.00000189, -0.00000624, -0.00000196, -0.00000413, 0.00000580, -0.00000188, -0.00000515, 0.00000965, -0.00000557, 0.00000574, -0.00000883, -0.00000286, 0.00000394, -0.00000224, 0.00000082, -0.00000331, -0.00000040, 0.00000133, -0.00000152, 0.00000060, },
		{ 0.00000057, 0.00000104, 0.00000090, -0.00000141, 0.00000519, 0.00000403, -0.00000458, -0.00000264, -0.00000323, -0.00000010, -0.00000102, 0.00000073, 0.00000099, -0.00000050, 0.00000113, 0.00000513, 0.00000028, -0.00000415, -0.00000030, -0.00000014, -0.00000246, 0.00000068, -0.00000104, -0.00000060, -0.00000096, 0.00000106, 0.00000173, 0.00000020, -0.00000096, -0.00000369, 0.00000523, 0.00000334, -0.00000340, 0.00000040, -0.00000086, -0.00000160, -0.00000112, 0.00000358, 0.00000048, -0.00000870, 0.00000286, 0.00000147, 0.00000196, 0.00000044, -0.00000171, 0.00000422, -0.00000155, 0.00000153, -0.00001403, 0.00001070, -0.00000841, 0.00000560, -0.00000148, 0.00001353, -0.00001878, 0.00001418, -0.00001743, 0.00001648, -0.00001302, 0.00002203, -0.00002141, 0.00000838, -0.00000455, -0.00000199, 0.00000274, -0.00000462, 0.00001037, -0.00001691, 0.00002460, -0.00002859, 0.00001393, -0.00001937, 0.00003632, -0.00003906, 0.00004058, -0.00005887, 0.00005526, -0.00005491, 0.00004949, -0.00005373, 0.00001957, -0.00000556, 0.00001361, -0.00001624, -0.00000155, -0.00000055, 0.00000696, -0.00000803, -0.00004521, -0.00004029, -0.00027196, 0.00056250, 0.00023142, -0.00026518, -0.00001662, -0.00003251, -0.00001342, -0.00000577, -0.00002760, 0.00001414, -0.00002571, 0.00000994, -0.00001556, -0.00000211, 0.00000759, -0.00001033, 0.00000822, -0.00000057, 0.00000631, 0.00000291, -0.00000598, -0.00000242, -0.00000388, 0.00000458, -0.00000015, -0.00000607, 0.00000985, -0.00000545, 0.00000536, -0.00000753, -0.00000458, 0.00000478, -0.00000234, 0.00000086, -0.00000282, -0.00000119, 0.00000159, -0.00000169, },
		{ -0.00000180, 0.00000047, 0.00000124, 0.00000097, -0.00000135, 0.00000438, 0.00000480, -0.00000428, -0.00000300, -0.00000327, 0.00000007, -0.00000103, 0.00000045, 0.00000109, -0.00000039, 0.00000072, 0.00000508, 0.00000087, -0.00000405, -0.00000060, 0.00000001, -0.00000260, 0.00000063, -0.00000082, -0.00000076, -0.00000087, 0.00000086, 0.00000180, 0.00000010, -0.00000060, -0.00000395, 0.00000454, 0.00000413, -0.00000332, 0.00000016, -0.00000060, -0.00000173, -0.00000121, 0.00000318, 0.00000153, -0.00000897, 0.00000207, 0.00000213, 0.00000114, 0.00000147, -0.00000276, 0.00000511, -0.00000242, 0.00000321, -0.00001519, 0.00001083, -0.00000843, 0.00000573, -0.00000278, 0.00001540, -0.00001943, 0.00001388, -0.00001675, 0.00001543, -0.00001233, 0.00002029, -0.00001715, 0.00000343, 0.00000009, -0.00000637, 0.00000672, -0.00000817, 0.00001374, -0.00002067, 0.00002783, -0.00002965, 0.00001500, -0.00002261, 0.00003934, -0.00003972, 0.00004024, -0.00005800, 0.00005175, -0.00004873, 0.00004096, -0.00004283, 0.00000710, 0.00000400, 0.00000628, -0.00001002, -0.00000516, 0.00000385, 0.00000293, -0.00000207, -0.00004764, -0.00003508, -0.00026325, 0.00050534, 0.00027123, -0.00024931, -0.00003587, -0.00002272, -0.00002124, -0.00000063, -0.00003115, 0.00001652, -0.00002593, 0.00000711, -0.00001141, -0.00000625, 0.00000988, -0.00001121, 0.00000779, 0.00000074, 0.00000496, 0.00000372, -0.00000550, -0.00000303, -0.00000350, 0.00000331, 0.00000140, -0.00000670, 0.00000980, -0.00000509, 0.00000474, -0.00000609, -0.00000623, 0.00000537, -0.00000220, 0.00000070, -0.00000220, -0.00000195, 0.00000177, },
		{ 0.00000184, -0.00000184, 0.00000028, 0.00000150, 0.00000098, -0.00000121, 0.00000354, 0.00000544, -0.00000386, -0.00000335, -0.00000337, 0.00000026, -0.00000099, 0.00000016, 0.00000116, -0.00000025, 0.00000033, 0.00000495, 0.00000147, -0.00000390, -0.00000091, 0.00000014, -0.00000271, 0.00000052, -0.00000057, -0.00000093, -0.00000078, 0.00000069, 0.00000186, -0.00000001, -0.00000027, -0.00000409, 0.00000376, 0.00000485, -0.00000313, -0.00000015, -0.00000032, -0.00000184, -0.00000129, 0.00000276, 0.00000247, -0.00000897, 0.00000104, 0.00000295, 0.00000021, 0.00000257, -0.00000382, 0.00000598, -0.00000330, 0.00000482, -0.00001612, 0.00001075, -0.00000838, 0.00000586, -0.00000405, 0.00001697, -0.00001959, 0.00001307, -0.00001559, 0.00001394, -0.00001111, 0.00001785, -0.00001232, -0.00000169, 0.00000463, -0.00001048, 0.00001030, -0.00001128, 0.00001669, -0.00002393, 0.00003036, -0.00003003, 0.00001565, -0.00002530, 0.00004136, -0.00003917, 0.00003875, -0.00005589, 0.00004691, -0.00004144, 0.00003176, -0.00003187, -0.00000447, 0.00001184, 0.00000082, -0.00000572, -0.00000700, 0.00000668, 0.00000057, 0.00000197, -0.00004782, -0.00003263, -0.00025200, 0.00045085, 0.00030259, -0.00022730, -0.00005629, -0.00001350, -0.00002810, 0.00000339, -0.00003357, 0.00001767, -0.00002486, 0.00000331, -0.00000693, -0.00001014, 0.00001165, -0.00001155, 0.00000690, 0.00000224, 0.00000369, 0.00000428, -0.00000479, -0.00000375, -0.00000303, 0.00000207, 0.00000270, -0.00000703, 0.00000952, -0.00000452, 0.00000394, -0.00000458, -0.00000772, 0.00000568, -0.00000182, 0.00000036, -0.00000148, -0.00000264, },
		{ -0.00000321, 0.00000179, -0.00000180, 0.00000000, 0.00000181, 0.00000095, -0.00000100, 0.00000272, 0.00000594, -0.00000332, -0.00000369, -0.00000353, 0.00000044, -0.00000091, -0.00000013, 0.00000119, -0.00000008, -0.00000002, 0.00000473, 0.00000207, -0.00000370, -0.00000122, 0.00000025, -0.00000277, 0.00000036, -0.00000030, -0.00000109, -0.00000070, 0.00000054, 0.00000191, -0.00000012, 0.00000001, -0.00000412, 0.00000292, 0.00000548, -0.00000281, -0.00000050, -0.00000005, -0.00000191, -0.00000137, 0.00000233, 0.00000326, -0.00000868, -0.00000020, 0.00000389, -0.00000081, 0.00000369, -0.00000484, 0.00000680, -0.00000414, 0.00000629, -0.00001679, 0.00001045, -0.00000823, 0.00000598, -0.00000521, 0.00001817, -0.00001919, 0.00001174, -0.00001399, 0.00001205, -0.00000940, 0.00001480, -0.00000708, -0.00000681, 0.00000890, -0.00001418, 0.00001337, -0.00001388, 0.00001915, -0.00002661, 0.00003212, -0.00002973, 0.00001586, -0.00002739, 0.00004232, -0.00003746, 0.00003622, -0.00005270, 0.00004100, -0.00003337, 0.00002229, -0.00002133, -0.00001470, 0.00001769, -0.00000269, -0.00000327, -0.00000723, 0.00000817, -0.00000040, 0.00000444, -0.00004623, -0.00003224, -0.00023924, 0.00039999, 0.00032556, -0.00020013, -0.00007689, -0.00000550, -0.00003362, 0.00000611, -0.00003477, 0.00001759, -0.00002260, -0.00000122, -0.00000237, -0.00001360, 0.00001281, -0.00001133, 0.00000560, 0.00000386, 0.00000257, 0.00000457, -0.00000389, -0.00000454, -0.00000252, 0.00000090, 0.00000373, -0.00000708, 0.00000903, -0.00000378, 0.00000300, -0.00000308, -0.00000900, 0.00000569, -0.00000122, -0.00000011, -0.00000074, },
		{ -0.00000002, -0.00000364, 0.00000162, -0.00000167, -0.00000035, 0.00000216, 0.00000089, -0.00000075, 0.00000193, 0.00000629, -0.00000268, -0.00000398, -0.00000375, 0.00000062, -0.00000077, -0.00000043, 0.00000118, 0.00000011, -0.00000031, 0.00000442, 0.00000265, -0.00000344, -0.00000152, 0.00000034, -0.00000279, 0.00000015, -0.00000002, -0.00000123, -0.00000065, 0.00000042, 0.00000194, -0.00000023, 0.00000025, -0.00000403, 0.00000203, 0.00000600, -0.00000237, -0.00000087, 0.00000019, -0.00000194, -0.00000147, 0.00000193, 0.00000387, -0.00000810, -0.00000163, 0.00000492, -0.00000188, 0.00000479, -0.00000579, 0.00000754, -0.00000491, 0.00000760, -0.00001716, 0.00000994, -0.00000800, 0.00000605, -0.00000621, 0.00001892, -0.00001821, 0.00000991, -0.00001197, 0.00000979, -0.00000727, 0.00001124, -0.00000162, -0.00001173, 0.00001275, -0.00001735, 0.00001587, -0.00001592, 0.00002109, -0.00002868, 0.00003312, -0.00002876, 0.00001564, -0.00002884, 0.00004224, -0.00003467, 0.00003279, -0.00004862, 0.00003429, -0.00002490, 0.00001299, -0.00001163, -0.00002328, 0.00002142, -0.00000431, -0.00000252, -0.00000609, 0.00000858, -0.00000027, 0.00000572, -0.00004335, -0.00003324, -0.00022590, 0.00035343, 0.00034044, -0.00016889, -0.00009666, 0.00000073, -0.00003754, 0.00000741, -0.00003475, 0.00001638, -0.00001936, -0.00000621, 0.00000201, -0.00001644, 0.00001328, -0.00001057, 0.00000397, 0.00000551, 0.00000165, 0.00000457, -0.00000280, -0.00000537, -0.00000202, -0.00000013, 0.00000446, -0.00000686, 0.00000838, -0.00000291, 0.00000199, -0.00000164, -0.00001002, 0.00000539, -0.00000043, -0.00000067, },
		{ -0.00000125, 0.00000062, -0.00000389, 0.00000131, -0.00000148, -0.00000078, 0.00000253, 0.00000081, -0.00000047, 0.00000121, 0.00000648, -0.00000196, -0.00000423, -0.00000403, 0.00000077, -0.00000058, -0.00000073, 0.00000113, 0.00000030, -0.00000055, 0.00000403, 0.00000321, -0.00000312, -0.00000182, 0.00000039, -0.00000277, -0.00000009, 0.00000026, -0.00000134, -0.00000063, 0.00000031, 0.00000197, -0.00000033, 0.00000042, -0.00000383, 0.00000111, 0.00000640, -0.00000182, -0.00000125, 0.00000038, -0.00000190, -0.00000160, 0.00000159, 0.00000428, -0.00000725, -0.00000322, 0.00000598, -0.00000293, 0.00000583, -0.00000663, 0.00000816, -0.00000557, 0.00000870, -0.00001722, 0.00000922, -0.00000767, 0.00000603, -0.00000697, 0.00001918, -0.00001666, 0.00000761, -0.00000959, 0.00000724, -0.00000479, 0.00000731, 0.00000389, -0.00001627, 0.00001604, -0.00001989, 0.00001774, -0.00001739, 0.00002248, -0.00003011, 0.00003336, -0.00002717, 0.00001502, -0.00002964, 0.00004114, -0.00003091, 0.00002865, -0.00004389, 0.00002707, -0.00001638, 0.00000424, -0.00000312, -0.00002995, 0.00002300, -0.00000418, -0.00000322, -0.00000387, 0.00000823, 0.00000063, 0.00000618, -0.00003965, -0.00003501, -0.00021274, 0.00031161, 0.00034778, -0.00013478, -0.00011469, 0.00000474, -0.00003970, 0.00000729, -0.00003359, 0.00001416, -0.00001535, -0.00001139, 0.00000597, -0.00001854, 0.00001304, -0.00000932, 0.00000208, 0.00000711, 0.00000100, 0.00000429, -0.00000158, -0.00000619, -0.00000159, -0.00000098, 0.00000490, -0.00000641, 0.00000760, -0.00000196, 0.00000097, -0.00000033, -0.00001074, 0.00000479, 0.00000050, },
		{ 0.00000152, -0.00000180, 0.00000114, -0.00000395, 0.00000088, -0.00000122, -0.00000126, 0.00000292, 0.00000074, -0.00000019, 0.00000057, 0.00000652, -0.00000118, -0.00000440, -0.00000437, 0.00000088, -0.00000033, -0.00000103, 0.00000106, 0.00000049, -0.00000071, 0.00000357, 0.00000372, -0.00000273, -0.00000212, 0.00000042, -0.00000271, -0.00000036, 0.00000053, -0.00000141, -0.00000063, 0.00000024, 0.00000199, -0.00000041, 0.00000053, -0.00000354, 0.00000020, 0.00000665, -0.00000118, -0.00000161, 0.00000050, -0.00000178, -0.00000177, 0.00000131, 0.00000447, -0.00000613, -0.00000491, 0.00000702, -0.00000394, 0.00000676, -0.00000733, 0.00000863, -0.00000610, 0.00000955, -0.00001695, 0.00000830, -0.00000725, 0.00000591, -0.00000746, 0.00001892, -0.00001455, 0.00000492, -0.00000695, 0.00000447, -0.00000207, 0.00000317, 0.00000926, -0.00002028, 0.00001867, -0.00002176, 0.00001896, -0.00001828, 0.00002334, -0.00003092, 0.00003287, -0.00002502, 0.00001402, -0.00002981, 0.00003909, -0.00002635, 0.00002399, -0.00003873, 0.00001964, -0.00000817, -0.00000363, 0.00000393, -0.00003460, 0.00002250, -0.00000254, -0.00000507, -0.00000091, 0.00000744, 0.00000198, 0.00000618, -0.00003553, -0.00003704, -0.00020035, 0.00027475, 0.00034832, -0.00009902, -0.00013013, 0.00000624, -0.00004007, 0.00000584, -0.00003144, 0.00001114, -0.00001083, -0.00001648, 0.00000930, -0.00001981, 0.00001213, -0.00000765, 0.00000003, 0.00000855, 0.00000066, 0.00000374, -0.00000026, -0.00000696, -0.00000126, -0.00000163, 0.00000505, -0.00000576, 0.00000675, -0.00000098, -0.00000001, 0.00000080, -0.00001116, 0.00000394, },
		{ 0.00000286, 0.00000257, -0.00000227, 0.00000150, -0.00000382, 0.00000034, -0.00000092, -0.00000178, 0.00000329, 0.00000069, 0.00000007, 0.00000003, 0.00000640, -0.00000037, -0.00000448, -0.00000474, 0.00000094, -0.00000001, -0.00000132, 0.00000096, 0.00000067, -0.00000081, 0.00000306, 0.00000417, -0.00000228, -0.00000240, 0.00000042, -0.00000262, -0.00000066, 0.00000077, -0.00000144, -0.00000066, 0.00000018, 0.00000200, -0.00000047, 0.00000056, -0.00000317, -0.00000069, 0.00000676, -0.00000046, -0.00000193, 0.00000054, -0.00000159, -0.00000199, 0.00000113, 0.00000443, -0.00000478, -0.00000665, 0.00000801, -0.00000485, 0.00000754, -0.00000785, 0.00000894, -0.00000646, 0.00001013, -0.00001637, 0.00000719, -0.00000673, 0.00000564, -0.00000764, 0.00001813, -0.00001192, 0.00000190, -0.00000412, 0.00000159, 0.00000078, -0.00000105, 0.00001430, -0.00002361, 0.00002055, -0.00002292, 0.00001953, -0.00001861, 0.00002367, -0.00003113, 0.00003169, -0.00002237, 0.00001269, -0.00002939, 0.00003621, -0.00002118, 0.00001905, -0.00003342, 0.00001230, -0.00000056, -0.00001034, 0.00000933, -0.00003719, 0.00002013, 0.00000029, -0.00000773, 0.00000246, 0.00000652, 0.00000349, 0.00000602, -0.00003137, -0.00003891, -0.00018917, 0.00024284, 0.00034294, -0.00006281, -0.00014231, 0.00000508, -0.00003873, 0.00000323, -0.00002850, 0.00000756, -0.00000606, -0.00002123, 0.00001184, -0.00002021, 0.00001059, -0.00000565, -0.00000206, 0.00000976, 0.00000065, 0.00000293, 0.00000111, -0.00000762, -0.00000107, -0.00000207, 0.00000493, -0.00000498, 0.00000586, 0.00000000, -0.00000091, 0.00000173, -0.00001128, },
		{ -0.00001110, 0.00000162, 0.00000357, -0.00000259, 0.00000168, -0.00000349, -0.00000027, -0.00000060, -0.00000232, 0.00000363, 0.00000068, 0.00000029, -0.00000040, 0.00000615, 0.00000045, -0.00000448, -0.00000515, 0.00000093, 0.00000036, -0.00000158, 0.00000083, 0.00000083, -0.00000083, 0.00000251, 0.00000454, -0.00000175, -0.00000266, 0.00000038, -0.00000249, -0.00000096, 0.00000097, -0.00000143, -0.00000072, 0.00000014, 0.00000199, -0.00000051, 0.00000053, -0.00000272, -0.00000153, 0.00000672, 0.00000032, -0.00000217, 0.00000048, -0.00000131, -0.00000226, 0.00000104, 0.00000419, -0.00000325, -0.00000838, 0.00000887, -0.00000561, 0.00000814, -0.00000818, 0.00000906, -0.00000663, 0.00001044, -0.00001547, 0.00000592, -0.00000611, 0.00000521, -0.00000747, 0.00001681, -0.00000885, -0.00000134, -0.00000121, -0.00000131, 0.00000366, -0.00000519, 0.00001887, -0.00002615, 0.00002164, -0.00002338, 0.00001949, -0.00001842, 0.00002352, -0.00003076, 0.00002990, -0.00001932, 0.00001111, -0.00002843, 0.00003260, -0.00001560, 0.00001406, -0.00002817, 0.00000531, 0.00000617, -0.00001570, 0.00001300, -0.00003782, 0.00001614, 0.00000395, -0.00001085, 0.00000590, 0.00000576, 0.00000491, 0.00000599, -0.00002743, -0.00004029, -0.00017948, 0.00021567, 0.00033262, -0.00002730, -0.00015070, 0.00000126, -0.00003585, -0.00000032, -0.00002501, 0.00000364, -0.00000131, -0.00002542, 0.00001347, -0.00001976, 0.00000852, -0.00000345, -0.00000409, 0.00001067, 0.00000099, 0.00000191, 0.00000248, -0.00000814, -0.00000104, -0.00000228, 0.00000459, -0.00000411, 0.00000498, 0.00000094, -0.00000170, 0.00000242, },
		{ 0.00000288, -0.00001067, 0.00000027, 0.00000446, -0.00000274, 0.00000167, -0.00000299, -0.00000094, -0.00000028, -0.00000286, 0.00000391, 0.00000073, 0.00000045, -0.00000071, 0.00000578, 0.00000126, -0.00000437, -0.00000557, 0.00000084, 0.00000078, -0.00000183, 0.00000067, 0.00000096, -0.00000079, 0.00000193, 0.00000482, -0.00000116, -0.00000291, 0.00000031, -0.00000233, -0.00000126, 0.00000114, -0.00000136, -0.00000081, 0.00000011, 0.00000198, -0.00000052, 0.00000043, -0.00000223, -0.00000230, 0.00000654, 0.00000112, -0.00000233, 0.00000032, -0.00000095, -0.00000259, 0.00000104, 0.00000374, -0.00000157, -0.00001004, 0.00000956, -0.00000620, 0.00000854, -0.00000828, 0.00000898, -0.00000661, 0.00001045, -0.00001427, 0.00000450, -0.00000540, 0.00000459, -0.00000694, 0.00001501, -0.00000542, -0.00000469, 0.00000167, -0.00000412, 0.00000645, -0.00000911, 0.00002283, -0.00002782, 0.00002194, -0.00002317, 0.00001888, -0.00001777, 0.00002293, -0.00002988, 0.00002757, -0.00001596, 0.00000935, -0.00002701, 0.00002843, -0.00000983, 0.00000925, -0.00002323, -0.00000109, 0.00001181, -0.00001957, 0.00001494, -0.00003665, 0.00001089, 0.00000805, -0.00001408, 0.00000912, 0.00000540, 0.00000601, 0.00000627, -0.00002392, -0.00004100, -0.00017139, 0.00019292, 0.00031840, 0.00000649, -0.00015493, -0.00000508, -0.00003174, -0.00000451, -0.00002123, -0.00000035, 0.00000319, -0.00002886, 0.00001414, -0.00001853, 0.00000604, -0.00000117, -0.00000596, 0.00001121, 0.00000168, 0.00000073, 0.00000379, -0.00000848, -0.00000118, -0.00000229, 0.00000405, -0.00000321, 0.00000415, 0.00000180, -0.00000233, },
		{ -0.00000281, 0.00000311, -0.00001002, -0.00000110, 0.00000519, -0.00000268, 0.00000146, -0.00000235, -0.00000161, 0.00000002, -0.00000337, 0.00000412, 0.00000085, 0.00000055, -0.00000091, 0.00000530, 0.00000202, -0.00000417, -0.00000599, 0.00000067, 0.00000124, -0.00000204, 0.00000049, 0.00000107, -0.00000070, 0.00000136, 0.00000499, -0.00000050, -0.00000312, 0.00000021, -0.00000215, -0.00000156, 0.00000125, -0.00000125, -0.00000091, 0.00000009, 0.00000195, -0.00000049, 0.00000027, -0.00000170, -0.00000298, 0.00000621, 0.00000192, -0.00000236, 0.00000004, -0.00000050, -0.00000296, 0.00000113, 0.00000312, 0.00000018, -0.00001156, 0.00001003, -0.00000657, 0.00000871, -0.00000817, 0.00000870, -0.00000637, 0.00001017, -0.00001280, 0.00000297, -0.00000461, 0.00000378, -0.00000607, 0.00001278, -0.00000174, -0.00000802, 0.00000441, -0.00000674, 0.00000904, -0.00001269, 0.00002606, -0.00002860, 0.00002148, -0.00002234, 0.00001779, -0.00001673, 0.00002196, -0.00002854, 0.00002480, -0.00001240, 0.00000748, -0.00002522, 0.00002386, -0.00000409, 0.00000482, -0.00001878, -0.00000672, 0.00001624, -0.00002192, 0.00001524, -0.00003395, 0.00000476, 0.00001220, -0.00001711, 0.00001184, 0.00000563, 0.00000667, 0.00000699, -0.00002095, -0.00004093, -0.00016491, 0.00017411, 0.00030132, 0.00003766, -0.00015486, -0.00001365, -0.00002672, -0.00000904, -0.00001743, -0.00000419, 0.00000724, -0.00003142, 0.00001383, -0.00001660, 0.00000330, 0.00000107, -0.00000757, 0.00001134, 0.00000269, -0.00000056, 0.00000498, -0.00000860, -0.00000152, -0.00000209, 0.00000338, -0.00000233, 0.00000341, 0.00000257, },
		{ 0.00000323, -0.00000312, 0.00000314, -0.00000921, -0.00000244, 0.00000572, -0.00000241, 0.00000107, -0.00000161, -0.00000226, 0.00000027, -0.00000384, 0.00000423, 0.00000105, 0.00000058, -0.00000100, 0.00000475, 0.00000273, -0.00000386, -0.00000639, 0.00000041, 0.00000174, -0.00000220, 0.00000028, 0.00000115, -0.00000056, 0.00000080, 0.00000505, 0.00000022, -0.00000330, 0.00000006, -0.00000194, -0.00000184, 0.00000131, -0.00000110, -0.00000103, 0.00000008, 0.00000191, -0.00000043, 0.00000006, -0.00000115, -0.00000357, 0.00000576, 0.00000269, -0.00000227, -0.00000035, 0.00000001, -0.00000336, 0.00000131, 0.00000236, 0.00000195, -0.00001288, 0.00001024, -0.00000672, 0.00000864, -0.00000782, 0.00000821, -0.00000594, 0.00000962, -0.00001109, 0.00000136, -0.00000374, 0.00000279, -0.00000487, 0.00001018, 0.00000209, -0.00001123, 0.00000691, -0.00000908, 0.00001134, -0.00001581, 0.00002848, -0.00002847, 0.00002031, -0.00002099, 0.00001630, -0.00001538, 0.00002068, -0.00002682, 0.00002168, -0.00000875, 0.00000559, -0.00002315, 0.00001906, 0.00000141, 0.00000096, -0.00001497, -0.00001145, 0.00001937, -0.00002277, 0.00001408, -0.00003002, -0.00000184, 0.00001603, -0.00001968, 0.00001389, 0.00000658, 0.00000681, 0.00000824, -0.00001857, -0.00004006, -0.00015992, 0.00015871, 0.00028236, 0.00006549, -0.00015050, -0.00002404, -0.00002122, -0.00001357, -0.00001386, -0.00000768, 0.00001067, -0.00003304, 0.00001261, -0.00001414, 0.00000046, 0.00000314, -0.00000885, 0.00001105, 0.00000399, -0.00000190, 0.00000600, -0.00000850, -0.00000203, -0.00000174, 0.00000263, -0.00000152, 0.00000277, },
		{ 0.00000224, 0.00000377, -0.00000327, 0.00000298, -0.00000829, -0.00000369, 0.00000599, -0.00000191, 0.00000054, -0.00000081, -0.00000284, 0.00000044, -0.00000425, 0.00000424, 0.00000134, 0.00000054, -0.00000099, 0.00000415, 0.00000335, -0.00000347, -0.00000675, 0.00000005, 0.00000226, -0.00000231, 0.00000005, 0.00000120, -0.00000038, 0.00000029, 0.00000497, 0.00000099, -0.00000343, -0.00000012, -0.00000173, -0.00000209, 0.00000131, -0.00000091, -0.00000115, 0.00000007, 0.00000187, -0.00000034, -0.00000021, -0.00000061, -0.00000405, 0.00000520, 0.00000341, -0.00000204, -0.00000084, 0.00000057, -0.00000378, 0.00000154, 0.00000150, 0.00000369, -0.00001395, 0.00001017, -0.00000662, 0.00000833, -0.00000726, 0.00000754, -0.00000530, 0.00000880, -0.00000918, -0.00000030, -0.00000281, 0.00000161, -0.00000338, 0.00000730, 0.00000593, -0.00001418, 0.00000906, -0.00001107, 0.00001327, -0.00001839, 0.00003006, -0.00002746, 0.00001851, -0.00001921, 0.00001451, -0.00001381, 0.00001916, -0.00002480, 0.00001833, -0.00000511, 0.00000376, -0.00002090, 0.00001419, 0.00000647, -0.00000218, -0.00001191, -0.00001519, 0.00002119, -0.00002225, 0.00001167, -0.00002522, -0.00000850, 0.00001921, -0.00002156, 0.00001513, 0.00000830, 0.00000640, 0.00000999, -0.00001676, -0.00003848, -0.00015623, 0.00014615, 0.00026246, 0.00008944, -0.00014207, -0.00003574, -0.00001564, -0.00001781, -0.00001073, -0.00001066, 0.00001338, -0.00003368, 0.00001057, -0.00001129, -0.00000234, 0.00000494, -0.00000974, 0.00001033, 0.00000551, -0.00000320, 0.00000681, -0.00000815, -0.00000272, -0.00000126, 0.00000184, -0.00000081, },
		{ -0.00000023, 0.00000184, 0.00000421, -0.00000328, 0.00000267, -0.00000731, -0.00000478, 0.00000599, -0.00000121, -0.00000010, -0.00000000, -0.00000332, 0.00000052, -0.00000458, 0.00000413, 0.00000171, 0.00000045, -0.00000090, 0.00000352, 0.00000388, -0.00000299, -0.00000705, -0.00000040, 0.00000277, -0.00000235, -0.00000021, 0.00000124, -0.00000019, -0.00000016, 0.00000477, 0.00000178, -0.00000349, -0.00000034, -0.00000150, -0.00000231, 0.00000126, -0.00000069, -0.00000128, 0.00000006, 0.00000181, -0.00000022, -0.00000051, -0.00000008, -0.00000441, 0.00000454, 0.00000405, -0.00000166, -0.00000142, 0.00000116, -0.00000420, 0.00000182, 0.00000057, 0.00000531, -0.00001471, 0.00000979, -0.00000627, 0.00000779, -0.00000649, 0.00000669, -0.00000449, 0.00000776, -0.00000711, -0.00000196, -0.00000183, 0.00000027, -0.00000164, 0.00000425, 0.00000966, -0.00001676, 0.00001078, -0.00001263, 0.00001476, -0.00002038, 0.00003077, -0.00002566, 0.00001621, -0.00001712, 0.00001252, -0.00001209, 0.00001748, -0.00002256, 0.00001483, -0.00000159, 0.00000208, -0.00001858, 0.00000941, 0.00001094, -0.00000449, -0.00000966, -0.00001794, 0.00002177, -0.00002052, 0.00000832, -0.00001991, -0.00001483, 0.00002148, -0.00002264, 0.00001549, 0.00001079, 0.00000549, 0.00001221, -0.00001542, -0.00003630, -0.00015360, 0.00013583, 0.00024244, 0.00010915, -0.00012992, -0.00004821, -0.00001042, -0.00002148, -0.00000822, -0.00001301, 0.00001530, -0.00003339, 0.00000784, -0.00000825, -0.00000494, 0.00000639, -0.00001021, 0.00000922, 0.00000719, -0.00000439, 0.00000736, -0.00000756, -0.00000354, -0.00000069, 0.00000106, },
		{ 0.00000035, 0.00000019, 0.00000155, 0.00000456, -0.00000315, 0.00000225, -0.00000635, -0.00000567, 0.00000572, -0.00000034, -0.00000080, 0.00000076, -0.00000366, 0.00000050, -0.00000483, 0.00000391, 0.00000216, 0.00000030, -0.00000075, 0.00000289, 0.00000430, -0.00000243, -0.00000730, -0.00000093, 0.00000326, -0.00000232, -0.00000050, 0.00000126, 0.00000001, -0.00000054, 0.00000445, 0.00000259, -0.00000349, -0.00000060, -0.00000127, -0.00000249, 0.00000116, -0.00000045, -0.00000139, 0.00000005, 0.00000175, -0.00000007, -0.00000083, 0.00000040, -0.00000464, 0.00000381, 0.00000458, -0.00000113, -0.00000207, 0.00000176, -0.00000460, 0.00000211, -0.00000036, 0.00000677, -0.00001513, 0.00000910, -0.00000570, 0.00000704, -0.00000553, 0.00000569, -0.00000352, 0.00000652, -0.00000493, -0.00000359, -0.00000083, -0.00000121, 0.00000027, 0.00000112, 0.00001315, -0.00001888, 0.00001203, -0.00001373, 0.00001579, -0.00002175, 0.00003065, -0.00002314, 0.00001352, -0.00001484, 0.00001043, -0.00001032, 0.00001569, -0.00002018, 0.00001131, 0.00000170, 0.00000061, -0.00001628, 0.00000488, 0.00001468, -0.00000591, -0.00000823, -0.00001971, 0.00002122, -0.00001779, 0.00000432, -0.00001447, -0.00002049, 0.00002264, -0.00002285, 0.00001498, 0.00001398, 0.00000417, 0.00001479, -0.00001444, -0.00003369, -0.00015175, 0.00012720, 0.00022298, 0.00012448, -0.00011457, -0.00006082, -0.00000594, -0.00002435, -0.00000645, -0.00001469, 0.00001644, -0.00003225, 0.00000460, -0.00000520, -0.00000722, 0.00000742, -0.00001025, 0.00000776, 0.00000895, -0.00000541, 0.00000762, -0.00000674, -0.00000448, -0.00000007, },
		{ 0.00000055, -0.00000028, 0.00000045, 0.00000135, 0.00000483, -0.00000292, 0.00000177, -0.00000544, -0.00000634, 0.00000518, 0.00000065, -0.00000151, 0.00000143, -0.00000383, 0.00000037, -0.00000498, 0.00000357, 0.00000267, 0.00000012, -0.00000056, 0.00000228, 0.00000460, -0.00000183, -0.00000746, -0.00000153, 0.00000371, -0.00000219, -0.00000082, 0.00000126, 0.00000020, -0.00000084, 0.00000400, 0.00000339, -0.00000339, -0.00000091, -0.00000103, -0.00000264, 0.00000100, -0.00000020, -0.00000148, 0.00000002, 0.00000168, 0.00000010, -0.00000116, 0.00000084, -0.00000476, 0.00000303, 0.00000499, -0.00000046, -0.00000276, 0.00000233, -0.00000496, 0.00000238, -0.00000127, 0.00000802, -0.00001518, 0.00000811, -0.00000491, 0.00000610, -0.00000443, 0.00000456, -0.00000242, 0.00000512, -0.00000271, -0.00000515, 0.00000018, -0.00000279, 0.00000230, -0.00000197, 0.00001628, -0.00002045, 0.00001274, -0.00001435, 0.00001634, -0.00002249, 0.00002975, -0.00002004, 0.00001060, -0.00001248, 0.00000834, -0.00000856, 0.00001388, -0.00001773, 0.00000786, 0.00000469, -0.00000058, -0.00001408, 0.00000072, 0.00001759, -0.00000643, -0.00000759, -0.00002059, 0.00001968, -0.00001432, -0.00000001, -0.00000924, -0.00002521, 0.00002258, -0.00002220, 0.00001367, 0.00001777, 0.00000258, 0.00001760, -0.00001367, -0.00003082, -0.00015042, 0.00011975, 0.00020464, 0.00013543, -0.00009662, -0.00007299, -0.00000255, -0.00002626, -0.00000547, -0.00001568, 0.00001683, -0.00003038, 0.00000102, -0.00000232, -0.00000906, 0.00000800, -0.00000988, 0.00000600, 0.00001070, -0.00000620, 0.00000758, -0.00000570, -0.00000550, },
		{ -0.00000655, 0.00000113, -0.00000077, 0.00000055, 0.00000123, 0.00000504, -0.00000261, 0.00000125, -0.00000462, -0.00000675, 0.00000440, 0.00000172, -0.00000215, 0.00000196, -0.00000382, 0.00000014, -0.00000506, 0.00000312, 0.00000323, -0.00000008, -0.00000035, 0.00000172, 0.00000478, -0.00000117, -0.00000753, -0.00000219, 0.00000410, -0.00000198, -0.00000116, 0.00000125, 0.00000037, -0.00000104, 0.00000345, 0.00000416, -0.00000320, -0.00000125, -0.00000081, -0.00000274, 0.00000081, 0.00000006, -0.00000155, -0.00000002, 0.00000160, 0.00000030, -0.00000150, 0.00000120, -0.00000476, 0.00000221, 0.00000527, 0.00000033, -0.00000346, 0.00000286, -0.00000524, 0.00000263, -0.00000210, 0.00000900, -0.00001485, 0.00000683, -0.00000394, 0.00000501, -0.00000320, 0.00000335, -0.00000122, 0.00000361, -0.00000048, -0.00000660, 0.00000116, -0.00000444, 0.00000436, -0.00000491, 0.00001896, -0.00002141, 0.00001291, -0.00001449, 0.00001642, -0.00002264, 0.00002814, -0.00001649, 0.00000757, -0.00001017, 0.00000633, -0.00000687, 0.00001208, -0.00001529, 0.00000456, 0.00000730, -0.00000144, -0.00001206, -0.00000297, 0.00001961, -0.00000606, -0.00000766, -0.00002070, 0.00001737, -0.00001038, -0.00000436, -0.00000453, -0.00002877, 0.00002128, -0.00002080, 0.00001167, 0.00002200, 0.00000086, 0.00002049, -0.00001295, -0.00002789, -0.00014934, 0.00011303, 0.00018782, 0.00014220, -0.00007674, -0.00008413, -0.00000052, -0.00002712, -0.00000529, -0.00001602, 0.00001655, -0.00002793, -0.00000268, 0.00000023, -0.00001039, 0.00000813, -0.00000914, 0.00000404, 0.00001235, -0.00000669, 0.00000723, -0.00000447, },
		{ -0.00000311, -0.00000759, 0.00000164, -0.00000113, 0.00000051, 0.00000116, 0.00000521, -0.00000224, 0.00000074, -0.00000394, -0.00000691, 0.00000342, 0.00000280, -0.00000269, 0.00000231, -0.00000363, -0.00000019, -0.00000506, 0.00000258, 0.00000382, -0.00000027, -0.00000014, 0.00000123, 0.00000484, -0.00000050, -0.00000751, -0.00000290, 0.00000440, -0.00000166, -0.00000152, 0.00000123, 0.00000052, -0.00000115, 0.00000282, 0.00000488, -0.00000291, -0.00000163, -0.00000059, -0.00000280, 0.00000058, 0.00000030, -0.00000159, -0.00000008, 0.00000152, 0.00000051, -0.00000182, 0.00000150, -0.00000465, 0.00000140, 0.00000539, 0.00000123, -0.00000415, 0.00000331, -0.00000543, 0.00000281, -0.00000283, 0.00000969, -0.00001413, 0.00000531, -0.00000283, 0.00000379, -0.00000189, 0.00000208, 0.00000004, 0.00000203, 0.00000170, -0.00000790, 0.00000209, -0.00000609, 0.00000637, -0.00000761, 0.00002111, -0.00002174, 0.00001254, -0.00001418, 0.00001606, -0.00002224, 0.00002593, -0.00001264, 0.00000459, -0.00000800, 0.00000448, -0.00000531, 0.00001035, -0.00001293, 0.00000149, 0.00000947, -0.00000195, -0.00001027, -0.00000612, 0.00002073, -0.00000489, -0.00000830, -0.00002019, 0.00001448, -0.00000624, -0.00000843, -0.00000057, -0.00003106, 0.00001878, -0.00001879, 0.00000913, 0.00002649, -0.00000082, 0.00002333, -0.00001212, -0.00002506, -0.00014828, 0.00010668, 0.00017278, 0.00014510, -0.00005567, -0.00009373, -0.00000004, -0.00002692, -0.00000585, -0.00001581, 0.00001574, -0.00002506, -0.00000633, 0.00000233, -0.00001117, 0.00000784, -0.00000810, 0.00000196, 0.00001382, -0.00000685, 0.00000657, },
		{ 0.00000564, -0.00000164, -0.00000858, 0.00000204, -0.00000133, 0.00000035, 0.00000110, 0.00000537, -0.00000183, 0.00000027, -0.00000340, -0.00000683, 0.00000229, 0.00000382, -0.00000307, 0.00000247, -0.00000325, -0.00000059, -0.00000500, 0.00000196, 0.00000441, -0.00000043, 0.00000004, 0.00000082, 0.00000478, 0.00000019, -0.00000740, -0.00000361, 0.00000460, -0.00000125, -0.00000190, 0.00000121, 0.00000063, -0.00000117, 0.00000212, 0.00000550, -0.00000249, -0.00000204, -0.00000039, -0.00000281, 0.00000032, 0.00000052, -0.00000159, -0.00000015, 0.00000143, 0.00000072, -0.00000211, 0.00000171, -0.00000444, 0.00000060, 0.00000537, 0.00000222, -0.00000479, 0.00000365, -0.00000552, 0.00000291, -0.00000341, 0.00001005, -0.00001306, 0.00000357, -0.00000162, 0.00000251, -0.00000054, 0.00000079, 0.00000132, 0.00000043, 0.00000377, -0.00000902, 0.00000296, -0.00000772, 0.00000826, -0.00000996, 0.00002264, -0.00002141, 0.00001166, -0.00001346, 0.00001530, -0.00002135, 0.00002325, -0.00000865, 0.00000179, -0.00000606, 0.00000284, -0.00000392, 0.00000874, -0.00001069, -0.00000127, 0.00001115, -0.00000208, -0.00000874, -0.00000869, 0.00002098, -0.00000302, -0.00000939, -0.00001923, 0.00001125, -0.00000217, -0.00001198, 0.00000243, -0.00003202, 0.00001523, -0.00001635, 0.00000623, 0.00003108, -0.00000229, 0.00002597, -0.00001104, -0.00002247, -0.00014707, 0.00010042, 0.00015965, 0.00014457, -0.00003413, -0.00010135, -0.00000120, -0.00002571, -0.00000703, -0.00001517, 0.00001451, -0.00002195, -0.00000976, 0.00000386, -0.00001141, 0.00000719, -0.00000683, -0.00000016, 0.00001504, -0.00000665, },
		{ -0.00000607, 0.00000446, -0.00000013, -0.00000948, 0.00000231, -0.00000138, 0.00000008, 0.00000103, 0.00000552, -0.00000140, -0.00000015, -0.00000303, -0.00000654, 0.00000108, 0.00000473, -0.00000325, 0.00000242, -0.00000271, -0.00000105, -0.00000490, 0.00000130, 0.00000497, -0.00000054, 0.00000017, 0.00000050, 0.00000461, 0.00000087, -0.00000719, -0.00000433, 0.00000467, -0.00000074, -0.00000228, 0.00000117, 0.00000072, -0.00000111, 0.00000139, 0.00000602, -0.00000196, -0.00000248, -0.00000022, -0.00000279, 0.00000005, 0.00000071, -0.00000155, -0.00000023, 0.00000133, 0.00000094, -0.00000237, 0.00000183, -0.00000415, -0.00000016, 0.00000519, 0.00000325, -0.00000535, 0.00000386, -0.00000547, 0.00000290, -0.00000382, 0.00001010, -0.00001165, 0.00000168, -0.00000037, 0.00000119, 0.00000082, -0.00000049, 0.00000259, -0.00000114, 0.00000569, -0.00000993, 0.00000372, -0.00000926, 0.00000995, -0.00001190, 0.00002354, -0.00002046, 0.00001034, -0.00001240, 0.00001422, -0.00002007, 0.00002021, -0.00000467, -0.00000073, -0.00000442, 0.00000145, -0.00000273, 0.00000726, -0.00000862, -0.00000368, 0.00001233, -0.00000183, -0.00000749, -0.00001066, 0.00002041, -0.00000059, -0.00001075, -0.00001800, 0.00000788, 0.00000161, -0.00001482, 0.00000439, -0.00003170, 0.00001082, -0.00001372, 0.00000314, 0.00003559, -0.00000342, 0.00002829, -0.00000960, -0.00002025, -0.00014558, 0.00009407, 0.00014845, 0.00014109, -0.00001280, -0.00010664, -0.00000398, -0.00002363, -0.00000869, -0.00001425, 0.00001302, -0.00001877, -0.00001281, 0.00000477, -0.00001112, 0.00000625, -0.00000542, -0.00000221, 0.00001593, },
		{ 0.00001644, -0.00000513, 0.00000309, 0.00000138, -0.00001024, 0.00000244, -0.00000129, -0.00000025, 0.00000092, 0.00000568, -0.00000096, -0.00000051, -0.00000281, -0.00000607, -0.00000017, 0.00000545, -0.00000320, 0.00000216, -0.00000203, -0.00000153, -0.00000478, 0.00000060, 0.00000549, -0.00000059, 0.00000025, 0.00000028, 0.00000434, 0.00000153, -0.00000690, -0.00000502, 0.00000461, -0.00000014, -0.00000265, 0.00000112, 0.00000077, -0.00000098, 0.00000064, 0.00000640, -0.00000130, -0.00000293, -0.00000007, -0.00000273, -0.00000022, 0.00000085, -0.00000146, -0.00000033, 0.00000123, 0.00000115, -0.00000258, 0.00000187, -0.00000377, -0.00000087, 0.00000488, 0.00000431, -0.00000578, 0.00000393, -0.00000530, 0.00000279, -0.00000404, 0.00000981, -0.00000994, -0.00000030, 0.00000088, -0.00000012, 0.00000215, -0.00000171, 0.00000381, -0.00000264, 0.00000742, -0.00001061, 0.00000437, -0.00001069, 0.00001138, -0.00001337, 0.00002377, -0.00001893, 0.00000864, -0.00001107, 0.00001288, -0.00001848, 0.00001695, -0.00000084, -0.00000286, -0.00000313, 0.00000032, -0.00000173, 0.00000594, -0.00000677, -0.00000571, 0.00001299, -0.00000124, -0.00000652, -0.00001204, 0.00001912, 0.00000223, -0.00001221, -0.00001666, 0.00000458, 0.00000490, -0.00001680, 0.00000525, -0.00003021, 0.00000580, -0.00001114, 0.00000005, 0.00003985, -0.00000407, 0.00003021, -0.00000772, -0.00001845, -0.00014371, 0.00008751, 0.00013908, 0.00013525, 0.00000766, -0.00010940, -0.00000830, -0.00002085, -0.00001065, -0.00001318, 0.00001141, -0.00001566, -0.00001538, 0.00000504, -0.00001039, 0.00000511, -0.00000396, -0.00000412, },
		{ -0.00000581, 0.00001654, -0.00000383, 0.00000158, 0.00000281, -0.00001084, 0.00000241, -0.00000107, -0.00000061, 0.00000075, 0.00000585, -0.00000051, -0.00000079, -0.00000274, -0.00000546, -0.00000137, 0.00000596, -0.00000290, 0.00000171, -0.00000124, -0.00000199, -0.00000466, -0.00000010, 0.00000593, -0.00000055, 0.00000026, 0.00000016, 0.00000398, 0.00000214, -0.00000652, -0.00000566, 0.00000441, 0.00000053, -0.00000299, 0.00000105, 0.00000079, -0.00000079, -0.00000007, 0.00000662, -0.00000052, -0.00000339, 0.00000005, -0.00000264, -0.00000048, 0.00000095, -0.00000133, -0.00000044, 0.00000113, 0.00000134, -0.00000273, 0.00000182, -0.00000335, -0.00000151, 0.00000443, 0.00000535, -0.00000607, 0.00000383, -0.00000498, 0.00000255, -0.00000408, 0.00000922, -0.00000799, -0.00000231, 0.00000207, -0.00000138, 0.00000340, -0.00000284, 0.00000495, -0.00000403, 0.00000891, -0.00001104, 0.00000488, -0.00001194, 0.00001249, -0.00001433, 0.00002337, -0.00001687, 0.00000665, -0.00000957, 0.00001138, -0.00001669, 0.00001361, 0.00000271, -0.00000453, -0.00000221, -0.00000052, -0.00000094, 0.00000480, -0.00000516, -0.00000732, 0.00001316, -0.00000032, -0.00000579, -0.00001288, 0.00001724, 0.00000527, -0.00001361, -0.00001539, 0.00000151, 0.00000755, -0.00001785, 0.00000506, -0.00002773, 0.00000044, -0.00000883, -0.00000290, 0.00004374, -0.00000414, 0.00003167, -0.00000536, -0.00001711, -0.00014146, 0.00008071, 0.00013137, 0.00012759, 0.00002670, -0.00010951, -0.00001395, -0.00001761, -0.00001271, -0.00001213, 0.00000981, -0.00001275, -0.00001738, 0.00000468, -0.00000929, 0.00000389, -0.00000255, },
		{ -0.00000126, -0.00000722, 0.00001622, -0.00000224, -0.00000001, 0.00000413, -0.00001126, 0.00000222, -0.00000074, -0.00000098, 0.00000051, 0.00000602, -0.00000005, -0.00000099, -0.00000279, -0.00000476, -0.00000248, 0.00000620, -0.00000236, 0.00000110, -0.00000040, -0.00000240, -0.00000457, -0.00000078, 0.00000627, -0.00000041, 0.00000019, 0.00000013, 0.00000357, 0.00000271, -0.00000607, -0.00000623, 0.00000407, 0.00000125, -0.00000330, 0.00000096, 0.00000080, -0.00000058, -0.00000074, 0.00000668, 0.00000037, -0.00000384, 0.00000013, -0.00000252, -0.00000073, 0.00000099, -0.00000116, -0.00000055, 0.00000103, 0.00000152, -0.00000282, 0.00000168, -0.00000288, -0.00000207, 0.00000386, 0.00000634, -0.00000619, 0.00000356, -0.00000452, 0.00000220, -0.00000392, 0.00000834, -0.00000587, -0.00000428, 0.00000315, -0.00000255, 0.00000456, -0.00000386, 0.00000597, -0.00000527, 0.00001014, -0.00001120, 0.00000524, -0.00001300, 0.00001324, -0.00001476, 0.00002235, -0.00001440, 0.00000449, -0.00000797, 0.00000980, -0.00001479, 0.00001030, 0.00000588, -0.00000569, -0.00000166, -0.00000112, -0.00000034, 0.00000382, -0.00000380, -0.00000852, 0.00001285, 0.00000086, -0.00000529, -0.00001323, 0.00001488, 0.00000835, -0.00001479, -0.00001431, -0.00000118, 0.00000946, -0.00001795, 0.00000391, -0.00002448, -0.00000497, -0.00000701, -0.00000559, 0.00004716, -0.00000360, 0.00003264, -0.00000249, -0.00001622, -0.00013883, 0.00007368, 0.00012510, 0.00011869, 0.00004388, -0.00010699, -0.00002068, -0.00001417, -0.00001469, -0.00001124, 0.00000832, -0.00001013, -0.00001878, 0.00000376, -0.00000796, 0.00000270, },
		{ 0.00000164, -0.00000016, -0.00000830, 0.00001547, -0.00000039, -0.00000159, 0.00000528, -0.00001146, 0.00000189, -0.00000033, -0.00000132, 0.00000018, 0.00000619, 0.00000042, -0.00000112, -0.00000295, -0.00000403, -0.00000343, 0.00000615, -0.00000160, 0.00000036, 0.00000044, -0.00000272, -0.00000452, -0.00000141, 0.00000650, -0.00000017, 0.00000005, 0.00000019, 0.00000310, 0.00000321, -0.00000556, -0.00000671, 0.00000359, 0.00000200, -0.00000355, 0.00000082, 0.00000081, -0.00000036, -0.00000132, 0.00000654, 0.00000136, -0.00000427, 0.00000016, -0.00000238, -0.00000095, 0.00000097, -0.00000095, -0.00000067, 0.00000093, 0.00000166, -0.00000285, 0.00000147, -0.00000238, -0.00000254, 0.00000320, 0.00000725, -0.00000611, 0.00000312, -0.00000393, 0.00000174, -0.00000358, 0.00000722, -0.00000363, -0.00000614, 0.00000407, -0.00000361, 0.00000558, -0.00000475, 0.00000685, -0.00000632, 0.00001109, -0.00001111, 0.00000545, -0.00001383, 0.00001361, -0.00001467, 0.00002080, -0.00001160, 0.00000225, -0.00000639, 0.00000822, -0.00001287, 0.00000713, 0.00000859, -0.00000634, -0.00000145, -0.00000147, 0.00000009, 0.00000301, -0.00000271, -0.00000930, 0.00001212, 0.00000224, -0.00000494, -0.00001318, 0.00001221, 0.00001130, -0.00001561, -0.00001353, -0.00000341, 0.00001058, -0.00001716, 0.00000195, -0.00002071, -0.00001014, -0.00000584, -0.00000796, 0.00005005, -0.00000240, 0.00003314, 0.00000085, -0.00001574, -0.00013588, 0.00006651, 0.00012001, 0.00010907, 0.00005884, -0.00010197, -0.00002816, -0.00001078, -0.00001639, -0.00001061, 0.00000704, -0.00000787, -0.00001959, 0.00000237, -0.00000651, },
		{ -0.00000509, 0.00000080, 0.00000070, -0.00000904, 0.00001433, 0.00000163, -0.00000310, 0.00000622, -0.00001145, 0.00000143, 0.00000013, -0.00000160, -0.00000023, 0.00000633, 0.00000092, -0.00000120, -0.00000318, -0.00000331, -0.00000418, 0.00000581, -0.00000063, -0.00000043, 0.00000124, -0.00000291, -0.00000454, -0.00000199, 0.00000660, 0.00000018, -0.00000017, 0.00000032, 0.00000261, 0.00000364, -0.00000500, -0.00000710, 0.00000298, 0.00000275, -0.00000371, 0.00000063, 0.00000081, -0.00000016, -0.00000180, 0.00000622, 0.00000244, -0.00000466, 0.00000015, -0.00000222, -0.00000114, 0.00000089, -0.00000070, -0.00000079, 0.00000083, 0.00000178, -0.00000281, 0.00000119, -0.00000188, -0.00000291, 0.00000246, 0.00000803, -0.00000583, 0.00000251, -0.00000323, 0.00000118, -0.00000309, 0.00000590, -0.00000135, -0.00000783, 0.00000481, -0.00000452, 0.00000646, -0.00000547, 0.00000757, -0.00000718, 0.00001174, -0.00001075, 0.00000549, -0.00001441, 0.00001359, -0.00001408, 0.00001877, -0.00000860, 0.00000006, -0.00000489, 0.00000671, -0.00001102, 0.00000420, 0.00001077, -0.00000647, -0.00000155, -0.00000164, 0.00000037, 0.00000237, -0.00000187, -0.00000969, 0.00001103, 0.00000375, -0.00000470, -0.00001281, 0.00000938, 0.00001396, -0.00001597, -0.00001312, -0.00000513, 0.00001092, -0.00001556, -0.00000062, -0.00001671, -0.00001480, -0.00000545, -0.00000998, 0.00005239, -0.00000057, 0.00003321, 0.00000463, -0.00001559, -0.00013269, 0.00005927, 0.00011584, 0.00009919, 0.00007137, -0.00009468, -0.00003603, -0.00000770, -0.00001767, -0.00001031, 0.00000600, -0.00000598, -0.00001983, 0.00000061, },
		{ -0.00000138, -0.00000382, 0.00000026, 0.00000129, -0.00000944, 0.00001284, 0.00000376, -0.00000448, 0.00000691, -0.00001123, 0.00000086, 0.00000062, -0.00000181, -0.00000071, 0.00000643, 0.00000145, -0.00000123, -0.00000345, -0.00000264, -0.00000470, 0.00000518, 0.00000047, -0.00000124, 0.00000193, -0.00000295, -0.00000464, -0.00000249, 0.00000656, 0.00000063, -0.00000044, 0.00000051, 0.00000211, 0.00000398, -0.00000440, -0.00000737, 0.00000227, 0.00000347, -0.00000377, 0.00000039, 0.00000084, 0.00000002, -0.00000214, 0.00000570, 0.00000358, -0.00000498, 0.00000010, -0.00000205, -0.00000129, 0.00000075, -0.00000043, -0.00000090, 0.00000073, 0.00000186, -0.00000270, 0.00000085, -0.00000138, -0.00000319, 0.00000168, 0.00000868, -0.00000533, 0.00000176, -0.00000244, 0.00000055, -0.00000245, 0.00000443, 0.00000089, -0.00000930, 0.00000532, -0.00000528, 0.00000718, -0.00000602, 0.00000812, -0.00000782, 0.00001209, -0.00001014, 0.00000539, -0.00001474, 0.00001318, -0.00001305, 0.00001639, -0.00000551, -0.00000200, -0.00000357, 0.00000533, -0.00000930, 0.00000157, 0.00001241, -0.00000611, -0.00000190, -0.00000166, 0.00000054, 0.00000187, -0.00000129, -0.00000973, 0.00000964, 0.00000531, -0.00000450, -0.00001222, 0.00000652, 0.00001620, -0.00001581, -0.00001311, -0.00000632, 0.00001053, -0.00001330, -0.00000358, -0.00001274, -0.00001874, -0.00000590, -0.00001167, 0.00005419, 0.00000185, 0.00003292, 0.00000875, -0.00001569, -0.00012937, 0.00005210, 0.00011231, 0.00008946, 0.00008134, -0.00008544, -0.00004391, -0.00000518, -0.00001842, -0.00001040, 0.00000520, -0.00000446, -0.00001958, },
		{ -0.00001894, -0.00000346, -0.00000282, 0.00000006, 0.00000159, -0.00000952, 0.00001106, 0.00000590, -0.00000564, 0.00000733, -0.00001080, 0.00000020, 0.00000111, -0.00000194, -0.00000125, 0.00000646, 0.00000202, -0.00000122, -0.00000374, -0.00000206, -0.00000498, 0.00000430, 0.00000167, -0.00000199, 0.00000248, -0.00000282, -0.00000481, -0.00000291, 0.00000638, 0.00000117, -0.00000076, 0.00000074, 0.00000163, 0.00000423, -0.00000378, -0.00000754, 0.00000147, 0.00000412, -0.00000370, 0.00000008, 0.00000090, 0.00000013, -0.00000234, 0.00000500, 0.00000475, -0.00000524, -0.00000001, -0.00000188, -0.00000138, 0.00000055, -0.00000013, -0.00000100, 0.00000063, 0.00000190, -0.00000253, 0.00000047, -0.00000091, -0.00000337, 0.00000088, 0.00000915, -0.00000462, 0.00000087, -0.00000158, -0.00000014, -0.00000171, 0.00000286, 0.00000303, -0.00001049, 0.00000561, -0.00000588, 0.00000772, -0.00000640, 0.00000849, -0.00000823, 0.00001215, -0.00000931, 0.00000514, -0.00001482, 0.00001241, -0.00001163, 0.00001374, -0.00000247, -0.00000381, -0.00000248, 0.00000414, -0.00000776, -0.00000072, 0.00001349, -0.00000532, -0.00000245, -0.00000158, 0.00000061, 0.00000150, -0.00000094, -0.00000945, 0.00000803, 0.00000684, -0.00000427, -0.00001149, 0.00000377, 0.00001792, -0.00001508, -0.00001348, -0.00000702, 0.00000949, -0.00001055, -0.00000671, -0.00000904, -0.00002180, -0.00000718, -0.00001311, 0.00005550, 0.00000480, 0.00003236, 0.00001316, -0.00001594, -0.00012602, 0.00004510, 0.00010920, 0.00008019, 0.00008877, -0.00007462, -0.00005144, -0.00000339, -0.00001858, -0.00001087, 0.00000464, -0.00000327, },
		{ -0.00000234, -0.00001800, -0.00000549, -0.00000219, 0.00000025, 0.00000164, -0.00000930, 0.00000905, 0.00000797, -0.00000654, 0.00000746, -0.00001019, -0.00000052, 0.00000158, -0.00000199, -0.00000182, 0.00000640, 0.00000265, -0.00000119, -0.00000403, -0.00000160, -0.00000500, 0.00000319, 0.00000290, -0.00000262, 0.00000284, -0.00000251, -0.00000506, -0.00000325, 0.00000607, 0.00000177, -0.00000109, 0.00000097, 0.00000118, 0.00000439, -0.00000313, -0.00000760, 0.00000062, 0.00000467, -0.00000349, -0.00000030, 0.00000100, 0.00000017, -0.00000239, 0.00000414, 0.00000593, -0.00000540, -0.00000017, -0.00000172, -0.00000143, 0.00000030, 0.00000017, -0.00000109, 0.00000054, 0.00000191, -0.00000230, 0.00000006, -0.00000047, -0.00000347, 0.00000008, 0.00000942, -0.00000372, -0.00000011, -0.00000067, -0.00000085, -0.00000090, 0.00000126, 0.00000501, -0.00001137, 0.00000565, -0.00000630, 0.00000809, -0.00000659, 0.00000867, -0.00000842, 0.00001192, -0.00000828, 0.00000477, -0.00001466, 0.00001130, -0.00000990, 0.00001095, 0.00000041, -0.00000530, -0.00000166, 0.00000317, -0.00000644, -0.00000262, 0.00001403, -0.00000416, -0.00000310, -0.00000145, 0.00000063, 0.00000125, -0.00000081, -0.00000891, 0.00000630, 0.00000825, -0.00000396, -0.00001072, 0.00000124, 0.00001905, -0.00001380, -0.00001421, -0.00000728, 0.00000792, -0.00000749, -0.00000978, -0.00000583, -0.00002385, -0.00000924, -0.00001441, 0.00005637, 0.00000819, 0.00003161, 0.00001775, -0.00001625, -0.00012274, 0.00003838, 0.00010630, 0.00007162, 0.00009375, -0.00006265, -0.00005826, -0.00000247, -0.00001812, -0.00001167, 0.00000426, },
		{ 0.00000397, -0.00000160, -0.00001687, -0.00000737, -0.00000199, 0.00000080, 0.00000147, -0.00000886, 0.00000691, 0.00000990, -0.00000712, 0.00000730, -0.00000941, -0.00000127, 0.00000200, -0.00000194, -0.00000240, 0.00000622, 0.00000333, -0.00000114, -0.00000428, -0.00000128, -0.00000479, 0.00000192, 0.00000407, -0.00000308, 0.00000298, -0.00000201, -0.00000538, -0.00000352, 0.00000564, 0.00000243, -0.00000143, 0.00000120, 0.00000080, 0.00000444, -0.00000248, -0.00000756, -0.00000025, 0.00000508, -0.00000313, -0.00000076, 0.00000115, 0.00000013, -0.00000227, 0.00000313, 0.00000708, -0.00000545, -0.00000038, -0.00000156, -0.00000143, 0.00000001, 0.00000047, -0.00000115, 0.00000045, 0.00000186, -0.00000202, -0.00000037, -0.00000007, -0.00000348, -0.00000068, 0.00000950, -0.00000263, -0.00000116, 0.00000023, -0.00000157, -0.00000005, -0.00000032, 0.00000676, -0.00001191, 0.00000545, -0.00000658, 0.00000829, -0.00000660, 0.00000866, -0.00000838, 0.00001144, -0.00000708, 0.00000429, -0.00001428, 0.00000992, -0.00000796, 0.00000814, 0.00000304, -0.00000641, -0.00000115, 0.00000242, -0.00000536, -0.00000413, 0.00001407, -0.00000273, -0.00000380, -0.00000131, 0.00000061, 0.00000110, -0.00000086, -0.00000816, 0.00000451, 0.00000948, -0.00000350, -0.00000996, -0.00000098, 0.00001955, -0.00001201, -0.00001521, -0.00000720, 0.00000597, -0.00000431, -0.00001258, -0.00000329, -0.00002487, -0.00001195, -0.00001569, 0.00005687, 0.00001195, 0.00003078, 0.00002245, -0.00001651, -0.00011963, 0.00003203, 0.00010344, 0.00006390, 0.00009646, -0.00004996, -0.00006408, -0.00000251, -0.00001710, -0.00001273, },
		{ -0.00001395, 0.00000370, -0.00000096, -0.00001567, -0.00000896, -0.00000226, 0.00000168, 0.00000115, -0.00000825, 0.00000470, 0.00001159, -0.00000736, 0.00000687, -0.00000851, -0.00000202, 0.00000236, -0.00000181, -0.00000296, 0.00000590, 0.00000408, -0.00000107, -0.00000450, -0.00000109, -0.00000436, 0.00000055, 0.00000512, -0.00000331, 0.00000288, -0.00000134, -0.00000575, -0.00000371, 0.00000511, 0.00000311, -0.00000173, 0.00000139, 0.00000049, 0.00000439, -0.00000182, -0.00000743, -0.00000111, 0.00000533, -0.00000260, -0.00000129, 0.00000135, -0.00000000, -0.00000201, 0.00000200, 0.00000817, -0.00000538, -0.00000064, -0.00000142, -0.00000137, -0.00000032, 0.00000076, -0.00000119, 0.00000037, 0.00000178, -0.00000169, -0.00000081, 0.00000027, -0.00000343, -0.00000140, 0.00000938, -0.00000139, -0.00000224, 0.00000111, -0.00000224, 0.00000079, -0.00000182, 0.00000824, -0.00001210, 0.00000502, -0.00000671, 0.00000833, -0.00000644, 0.00000849, -0.00000814, 0.00001073, -0.00000576, 0.00000374, -0.00001370, 0.00000831, -0.00000588, 0.00000540, 0.00000530, -0.00000709, -0.00000095, 0.00000191, -0.00000451, -0.00000528, 0.00001368, -0.00000110, -0.00000447, -0.00000121, 0.00000057, 0.00000104, -0.00000107, -0.00000726, 0.00000276, 0.00001045, -0.00000286, -0.00000930, -0.00000284, 0.00001943, -0.00000979, -0.00001638, -0.00000688, 0.00000378, -0.00000121, -0.00001494, -0.00000151, -0.00002487, -0.00001515, -0.00001711, 0.00005707, 0.00001599, 0.00002995, 0.00002718, -0.00001665, -0.00011676, 0.00002612, 0.00010049, 0.00005712, 0.00009715, -0.00003699, -0.00006866, -0.00000350, -0.00001558, },
		{ -0.00001371, -0.00001519, 0.00000334, -0.00000034, -0.00001450, -0.00001021, -0.00000302, 0.00000284, 0.00000073, -0.00000754, 0.00000252, 0.00001298, -0.00000723, 0.00000618, -0.00000751, -0.00000274, 0.00000265, -0.00000162, -0.00000346, 0.00000543, 0.00000486, -0.00000097, -0.00000468, -0.00000104, -0.00000376, -0.00000085, 0.00000599, -0.00000328, 0.00000254, -0.00000052, -0.00000615, -0.00000386, 0.00000450, 0.00000378, -0.00000199, 0.00000153, 0.00000027, 0.00000423, -0.00000115, -0.00000723, -0.00000191, 0.00000537, -0.00000191, -0.00000190, 0.00000161, -0.00000023, -0.00000160, 0.00000080, 0.00000916, -0.00000518, -0.00000094, -0.00000130, -0.00000127, -0.00000067, 0.00000102, -0.00000120, 0.00000029, 0.00000166, -0.00000133, -0.00000123, 0.00000055, -0.00000331, -0.00000205, 0.00000904, -0.00000003, -0.00000330, 0.00000192, -0.00000286, 0.00000160, -0.00000321, 0.00000941, -0.00001194, 0.00000440, -0.00000672, 0.00000822, -0.00000612, 0.00000815, -0.00000771, 0.00000982, -0.00000434, 0.00000315, -0.00001297, 0.00000654, -0.00000378, 0.00000285, 0.00000714, -0.00000733, -0.00000105, 0.00000160, -0.00000388, -0.00000608, 0.00001293, 0.00000063, -0.00000504, -0.00000118, 0.00000053, 0.00000105, -0.00000142, -0.00000628, 0.00000111, 0.00001111, -0.00000200, -0.00000875, -0.00000428, 0.00001873, -0.00000724, -0.00001762, -0.00000644, 0.00000149, 0.00000166, -0.00001673, -0.00000057, -0.00002394, -0.00001861, -0.00001882, 0.00005703, 0.00002024, 0.00002919, 0.00003189, -0.00001661, -0.00011419, 0.00002069, 0.00009739, 0.00005129, 0.00009609, -0.00002414, -0.00007184, -0.00000539, },
		{ -0.00000807, -0.00001163, -0.00001633, 0.00000280, 0.00000033, -0.00001342, -0.00001106, -0.00000422, 0.00000419, 0.00000032, -0.00000681, 0.00000044, 0.00001402, -0.00000673, 0.00000527, -0.00000647, -0.00000340, 0.00000286, -0.00000137, -0.00000389, 0.00000480, 0.00000568, -0.00000084, -0.00000482, -0.00000111, -0.00000303, -0.00000222, 0.00000661, -0.00000296, 0.00000197, 0.00000042, -0.00000653, -0.00000398, 0.00000384, 0.00000440, -0.00000216, 0.00000158, 0.00000016, 0.00000397, -0.00000049, -0.00000698, -0.00000262, 0.00000520, -0.00000105, -0.00000257, 0.00000191, -0.00000054, -0.00000107, -0.00000045, 0.00001002, -0.00000483, -0.00000127, -0.00000121, -0.00000112, -0.00000102, 0.00000126, -0.00000118, 0.00000022, 0.00000151, -0.00000094, -0.00000162, 0.00000076, -0.00000314, -0.00000261, 0.00000852, 0.00000142, -0.00000431, 0.00000264, -0.00000340, 0.00000234, -0.00000444, 0.00001026, -0.00001145, 0.00000361, -0.00000663, 0.00000798, -0.00000567, 0.00000768, -0.00000713, 0.00000876, -0.00000289, 0.00000254, -0.00001212, 0.00000468, -0.00000173, 0.00000058, 0.00000851, -0.00000713, -0.00000141, 0.00000148, -0.00000345, -0.00000657, 0.00001188, 0.00000236, -0.00000546, -0.00000123, 0.00000051, 0.00000111, -0.00000185, -0.00000527, -0.00000036, 0.00001142, -0.00000092, -0.00000835, -0.00000532, 0.00001751, -0.00000449, -0.00001880, -0.00000599, -0.00000075, 0.00000417, -0.00001785, -0.00000045, -0.00002222, -0.00002213, -0.00002095, 0.00005677, 0.00002465, 0.00002857, 0.00003653, -0.00001634, -0.00011195, 0.00001575, 0.00009409, 0.00004637, 0.00009360, -0.00001176, -0.00007352, },
		{ -0.00007370, -0.00001135, -0.00000953, -0.00001724, 0.00000201, 0.00000110, -0.00001249, -0.00001149, -0.00000583, 0.00000562, -0.00000001, -0.00000612, -0.00000148, 0.00001468, -0.00000587, 0.00000420, -0.00000543, -0.00000398, 0.00000299, -0.00000109, -0.00000422, 0.00000402, 0.00000650, -0.00000065, -0.00000493, -0.00000128, -0.00000222, -0.00000348, 0.00000693, -0.00000234, 0.00000120, 0.00000143, -0.00000687, -0.00000409, 0.00000317, 0.00000495, -0.00000222, 0.00000154, 0.00000016, 0.00000361, 0.00000017, -0.00000670, -0.00000320, 0.00000480, -0.00000005, -0.00000328, 0.00000226, -0.00000093, -0.00000045, -0.00000170, 0.00001071, -0.00000433, -0.00000163, -0.00000115, -0.00000094, -0.00000138, 0.00000145, -0.00000112, 0.00000016, 0.00000132, -0.00000055, -0.00000198, 0.00000091, -0.00000293, -0.00000307, 0.00000781, 0.00000291, -0.00000522, 0.00000323, -0.00000383, 0.00000298, -0.00000547, 0.00001076, -0.00001065, 0.00000269, -0.00000648, 0.00000763, -0.00000509, 0.00000709, -0.00000641, 0.00000759, -0.00000144, 0.00000195, -0.00001119, 0.00000279, 0.00000018, -0.00000136, 0.00000938, -0.00000652, -0.00000199, 0.00000149, -0.00000318, -0.00000681, 0.00001063, 0.00000402, -0.00000568, -0.00000138, 0.00000049, 0.00000123, -0.00000235, -0.00000429, -0.00000159, 0.00001136, 0.00000038, -0.00000809, -0.00000597, 0.00001588, -0.00000167, -0.00001980, -0.00000562, -0.00000286, 0.00000620, -0.00001826, -0.00000110, -0.00001990, -0.00002545, -0.00002363, 0.00005631, 0.00002917, 0.00002813, 0.00004107, -0.00001581, -0.00011006, 0.00001129, 0.00009058, 0.00004230, 0.00008999, -0.00000017, },
		{ 0.00001040, -0.00007245, -0.00001504, -0.00000759, -0.00001779, 0.00000091, 0.00000201, -0.00001174, -0.00001152, -0.00000774, 0.00000703, -0.00000016, -0.00000553, -0.00000317, 0.00001492, -0.00000470, 0.00000300, -0.00000443, -0.00000446, 0.00000304, -0.00000081, -0.00000442, 0.00000310, 0.00000729, -0.00000039, -0.00000502, -0.00000151, -0.00000139, -0.00000457, 0.00000693, -0.00000145, 0.00000028, 0.00000245, -0.00000713, -0.00000423, 0.00000250, 0.00000539, -0.00000216, 0.00000138, 0.00000029, 0.00000315, 0.00000083, -0.00000642, -0.00000362, 0.00000417, 0.00000109, -0.00000401, 0.00000263, -0.00000138, 0.00000024, -0.00000291, 0.00001122, -0.00000369, -0.00000200, -0.00000112, -0.00000072, -0.00000173, 0.00000159, -0.00000103, 0.00000010, 0.00000110, -0.00000015, -0.00000228, 0.00000099, -0.00000270, -0.00000342, 0.00000694, 0.00000439, -0.00000601, 0.00000368, -0.00000413, 0.00000351, -0.00000629, 0.00001093, -0.00000958, 0.00000169, -0.00000628, 0.00000718, -0.00000441, 0.00000641, -0.00000559, 0.00000634, -0.00000004, 0.00000142, -0.00001022, 0.00000095, 0.00000188, -0.00000290, 0.00000975, -0.00000554, -0.00000272, 0.00000158, -0.00000303, -0.00000685, 0.00000925, 0.00000553, -0.00000569, -0.00000162, 0.00000049, 0.00000139, -0.00000289, -0.00000339, -0.00000254, 0.00001092, 0.00000188, -0.00000796, -0.00000628, 0.00001392, 0.00000108, -0.00002053, -0.00000541, -0.00000472, 0.00000769, -0.00001798, -0.00000242, -0.00001719, -0.00002835, -0.00002691, 0.00005560, 0.00003380, 0.00002789, 0.00004552, -0.00001502, -0.00010851, 0.00000730, 0.00008689, 0.00003898, 0.00008554, },
		{ 0.00008053, 0.00001981, -0.00006991, -0.00001888, -0.00000598, -0.00001790, -0.00000053, 0.00000304, -0.00001117, -0.00001118, -0.00000986, 0.00000829, -0.00000007, -0.00000510, -0.00000461, 0.00001475, -0.00000324, 0.00000175, -0.00000352, -0.00000482, 0.00000302, -0.00000052, -0.00000450, 0.00000207, 0.00000802, -0.00000003, -0.00000512, -0.00000178, -0.00000058, -0.00000544, 0.00000658, -0.00000030, -0.00000075, 0.00000343, -0.00000726, -0.00000441, 0.00000188, 0.00000568, -0.00000195, 0.00000111, 0.00000053, 0.00000261, 0.00000149, -0.00000615, -0.00000387, 0.00000330, 0.00000233, -0.00000475, 0.00000301, -0.00000187, 0.00000096, -0.00000404, 0.00001151, -0.00000291, -0.00000238, -0.00000113, -0.00000048, -0.00000204, 0.00000168, -0.00000091, 0.00000006, 0.00000087, 0.00000023, -0.00000253, 0.00000100, -0.00000246, -0.00000367, 0.00000595, 0.00000583, -0.00000662, 0.00000397, -0.00000431, 0.00000392, -0.00000688, 0.00001077, -0.00000830, 0.00000066, -0.00000608, 0.00000666, -0.00000365, 0.00000566, -0.00000470, 0.00000507, 0.00000128, 0.00000096, -0.00000925, -0.00000079, 0.00000330, -0.00000402, 0.00000965, -0.00000426, -0.00000354, 0.00000171, -0.00000296, -0.00000674, 0.00000782, 0.00000684, -0.00000546, -0.00000195, 0.00000048, 0.00000157, -0.00000342, -0.00000262, -0.00000318, 0.00001012, 0.00000353, -0.00000792, -0.00000630, 0.00001178, 0.00000363, -0.00002090, -0.00000543, -0.00000630, 0.00000863, -0.00001704, -0.00000425, -0.00001431, -0.00003062, -0.00003081, 0.00005458, 0.00003853, 0.00002787, 0.00004987, -0.00001395, -0.00010729, 0.00000372, 0.00008303, 0.00003630, },
		{ 0.00003414, 0.00007519, 0.00002796, -0.00006626, -0.00002264, -0.00000485, -0.00001750, -0.00000231, 0.00000418, -0.00001074, -0.00001055, -0.00001206, 0.00000929, 0.00000033, -0.00000484, -0.00000577, 0.00001418, -0.00000156, 0.00000050, -0.00000272, -0.00000506, 0.00000294, -0.00000026, -0.00000445, 0.00000095, 0.00000864, 0.00000044, -0.00000523, -0.00000206, 0.00000016, -0.00000605, 0.00000589, 0.00000104, -0.00000181, 0.00000430, -0.00000723, -0.00000466, 0.00000133, 0.00000582, -0.00000158, 0.00000072, 0.00000089, 0.00000197, 0.00000215, -0.00000593, -0.00000392, 0.00000222, 0.00000364, -0.00000545, 0.00000339, -0.00000238, 0.00000167, -0.00000504, 0.00001157, -0.00000200, -0.00000274, -0.00000118, -0.00000023, -0.00000232, 0.00000171, -0.00000075, 0.00000002, 0.00000062, 0.00000058, -0.00000271, 0.00000095, -0.00000221, -0.00000380, 0.00000485, 0.00000718, -0.00000705, 0.00000409, -0.00000435, 0.00000419, -0.00000724, 0.00001031, -0.00000684, -0.00000036, -0.00000589, 0.00000608, -0.00000284, 0.00000488, -0.00000379, 0.00000381, 0.00000248, 0.00000061, -0.00000832, -0.00000238, 0.00000442, -0.00000471, 0.00000911, -0.00000277, -0.00000437, 0.00000182, -0.00000292, -0.00000654, 0.00000641, 0.00000790, -0.00000500, -0.00000235, 0.00000047, 0.00000178, -0.00000392, -0.00000201, -0.00000350, 0.00000899, 0.00000529, -0.00000793, -0.00000612, 0.00000957, 0.00000586, -0.00002084, -0.00000569, -0.00000757, 0.00000903, -0.00001556, -0.00000644, -0.00001150, -0.00003212, -0.00003531, 0.00005315, 0.00004337, 0.00002807, 0.00005415, -0.00001263, -0.00010638, 0.00000053, 0.00007905, },
		{ 0.00007498, 0.00003242, 0.00006970, 0.00003483, -0.00006175, -0.00002606, -0.00000433, -0.00001658, -0.00000437, 0.00000538, -0.00001041, -0.00000972, -0.00001422, 0.00000993, 0.00000108, -0.00000476, -0.00000665, 0.00001326, 0.00000027, -0.00000067, -0.00000209, -0.00000516, 0.00000281, -0.00000004, -0.00000428, -0.00000021, 0.00000910, 0.00000106, -0.00000537, -0.00000232, 0.00000078, -0.00000639, 0.00000489, 0.00000253, -0.00000283, 0.00000501, -0.00000700, -0.00000501, 0.00000088, 0.00000576, -0.00000104, 0.00000020, 0.00000137, 0.00000127, 0.00000282, -0.00000575, -0.00000377, 0.00000094, 0.00000497, -0.00000609, 0.00000373, -0.00000287, 0.00000235, -0.00000589, 0.00001139, -0.00000098, -0.00000308, -0.00000127, 0.00000002, -0.00000255, 0.00000167, -0.00000056, -0.00000000, 0.00000037, 0.00000090, -0.00000282, 0.00000084, -0.00000198, -0.00000383, 0.00000369, 0.00000841, -0.00000726, 0.00000404, -0.00000426, 0.00000432, -0.00000738, 0.00000959, -0.00000527, -0.00000132, -0.00000573, 0.00000546, -0.00000200, 0.00000409, -0.00000288, 0.00000261, 0.00000352, 0.00000038, -0.00000746, -0.00000377, 0.00000519, -0.00000499, 0.00000820, -0.00000114, -0.00000514, 0.00000186, -0.00000288, -0.00000629, 0.00000507, 0.00000869, -0.00000433, -0.00000278, 0.00000044, 0.00000200, -0.00000436, -0.00000160, -0.00000349, 0.00000756, 0.00000710, -0.00000793, -0.00000580, 0.00000741, 0.00000768, -0.00002032, -0.00000620, -0.00000853, 0.00000894, -0.00001363, -0.00000880, -0.00000895, -0.00003272, -0.00004032, 0.00005118, 0.00004836, 0.00002848, 0.00005838, -0.00001106, -0.00010575, -0.00000233, },
		{ -0.00000489, 0.00007087, 0.00003104, 0.00006420, 0.00004048, -0.00005664, -0.00002894, -0.00000448, -0.00001516, -0.00000664, 0.00000655, -0.00001010, -0.00000877, -0.00001623, 0.00001013, 0.00000218, -0.00000486, -0.00000726, 0.00001202, 0.00000218, -0.00000171, -0.00000165, -0.00000514, 0.00000264, 0.00000015, -0.00000400, -0.00000137, 0.00000938, 0.00000183, -0.00000555, -0.00000253, 0.00000127, -0.00000644, 0.00000362, 0.00000407, -0.00000374, 0.00000550, -0.00000656, -0.00000545, 0.00000055, 0.00000551, -0.00000034, -0.00000044, 0.00000194, 0.00000050, 0.00000349, -0.00000564, -0.00000342, -0.00000049, 0.00000628, -0.00000664, 0.00000401, -0.00000334, 0.00000295, -0.00000656, 0.00001098, 0.00000012, -0.00000337, -0.00000140, 0.00000027, -0.00000272, 0.00000157, -0.00000035, -0.00000002, 0.00000012, 0.00000118, -0.00000286, 0.00000069, -0.00000176, -0.00000376, 0.00000249, 0.00000947, -0.00000725, 0.00000382, -0.00000405, 0.00000433, -0.00000731, 0.00000864, -0.00000365, -0.00000217, -0.00000563, 0.00000481, -0.00000115, 0.00000333, -0.00000200, 0.00000150, 0.00000438, 0.00000030, -0.00000668, -0.00000494, 0.00000562, -0.00000488, 0.00000700, 0.00000053, -0.00000578, 0.00000180, -0.00000281, -0.00000602, 0.00000384, 0.00000919, -0.00000347, -0.00000322, 0.00000038, 0.00000222, -0.00000472, -0.00000139, -0.00000317, 0.00000590, 0.00000887, -0.00000785, -0.00000544, 0.00000540, 0.00000902, -0.00001936, -0.00000693, -0.00000924, 0.00000843, -0.00001141, -0.00001115, -0.00000685, -0.00003237, -0.00004567, 0.00004851, 0.00005349, 0.00002910, 0.00006256, -0.00000924, -0.00010538, },
		},
		{ { -0.00000652, -0.00000747, -0.00001236, -0.00000850, -0.00004245, -0.00002855, 0.00004942, 0.00004984, 0.00002568, 0.00005689, 0.00003417, -0.00002644, -0.00003379, -0.00000275, 0.00000773, -0.00002330, 0.00000105, -0.00000293, 0.00000220, -0.00001037, -0.00001265, 0.00001316, -0.00001188, 0.00000269, -0.00000570, -0.00000232, 0.00001204, -0.00000435, -0.00001766, -0.00000783, 0.00001541, 0.00000712, -0.00001024, 0.00000637, -0.00000447, 0.00000007, 0.00000471, -0.00001111, 0.00000748, 0.00000275, 0.00000482, -0.00000295, -0.00000009, 0.00000056, -0.00000518, -0.00000523, 0.00000172, 0.00000235, 0.00000488, 0.00000316, -0.00000546, -0.00000131, -0.00000768, 0.00001275, -0.00000594, 0.00000553, 0.00000380, -0.00001056, 0.00000114, -0.00000961, 0.00001107, 0.00000440, -0.00000345, 0.00000065, -0.00000474, 0.00000065, -0.00000345, 0.00000440, 0.00001107, -0.00000961, 0.00000114, -0.00001056, 0.00000380, 0.00000553, -0.00000594, 0.00001275, -0.00000768, -0.00000131, -0.00000546, 0.00000316, 0.00000488, 0.00000235, 0.00000172, -0.00000523, -0.00000518, 0.00000056, -0.00000009, -0.00000295, 0.00000482, 0.00000275, 0.00000748, -0.00001111, 0.00000471, 0.00000007, -0.00000447, 0.00000637, -0.00001024, 0.00000712, 0.00001541, -0.00000783, -0.00001766, -0.00000435, 0.00001204, -0.00000232, -0.00000570, 0.00000269, -0.00001188, 0.00001316, -0.00001265, -0.00001037, 0.00000220, -0.00000293, 0.00000105, -0.00002330, 0.00000773, -0.00000275, -0.00003379, -0.00002644, 0.00003417, 0.00005689, 0.00002568, 0.00004984, 0.00004942, -0.00002855, -0.00004245, -0.00000850, -0.00001236, -0.00000747, },
		{ -0.00000789, -0.00000647, -0.00000721, -0.00001193, -0.00000789, -0.00003847, -0.00003166, 0.00004108, 0.00005177, 0.00002321, 0.00005094, 0.00004055, -0.00001912, -0.00003255, -0.00001000, 0.00001187, -0.00002272, -0.00000249, 0.00000035, -0.00000057, -0.00000623, -0.00001657, 0.00001367, -0.00000960, 0.00000034, -0.00000292, -0.00000581, 0.00001281, -0.00000138, -0.00001768, -0.00001012, 0.00001243, 0.00001164, -0.00001048, 0.00000461, -0.00000255, -0.00000298, 0.00000771, -0.00001205, 0.00000618, 0.00000369, 0.00000441, -0.00000210, -0.00000081, 0.00000110, -0.00000449, -0.00000580, 0.00000040, 0.00000367, 0.00000349, 0.00000478, -0.00000623, 0.00000051, -0.00000981, 0.00001157, -0.00000252, 0.00000117, 0.00000734, -0.00001110, 0.00000186, -0.00001026, 0.00000823, 0.00000743, -0.00000428, 0.00000113, -0.00000444, -0.00000034, -0.00000224, 0.00000150, 0.00001285, -0.00000787, -0.00000017, -0.00000925, -0.00000032, 0.00000942, -0.00000799, 0.00001222, -0.00000451, -0.00000327, -0.00000470, 0.00000153, 0.00000611, 0.00000143, 0.00000238, -0.00000418, -0.00000582, -0.00000016, 0.00000070, -0.00000376, 0.00000507, 0.00000214, 0.00000808, -0.00000936, 0.00000149, 0.00000259, -0.00000559, 0.00000709, -0.00000873, 0.00000236, 0.00001723, -0.00000477, -0.00001764, -0.00000662, 0.00001016, 0.00000141, -0.00000794, 0.00000418, -0.00001279, 0.00001104, -0.00000779, -0.00001450, 0.00000440, -0.00000515, 0.00000281, -0.00002175, 0.00000228, 0.00000365, -0.00003254, -0.00003393, 0.00002627, 0.00006173, 0.00002973, 0.00004723, 0.00005738, -0.00002399, -0.00004682, -0.00000926, -0.00001271, },
		{ -0.00001289, -0.00000849, -0.00000635, -0.00000705, -0.00001151, -0.00000730, -0.00003495, -0.00003351, 0.00003275, 0.00005268, 0.00002236, 0.00004416, 0.00004558, -0.00001246, -0.00002916, -0.00001705, 0.00001386, -0.00001979, -0.00000752, 0.00000416, -0.00000330, -0.00000270, -0.00001904, 0.00001242, -0.00000616, -0.00000251, -0.00000008, -0.00000855, 0.00001224, 0.00000225, -0.00001769, -0.00001170, 0.00000873, 0.00001532, -0.00000917, 0.00000188, -0.00000005, -0.00000619, 0.00001011, -0.00001200, 0.00000421, 0.00000482, 0.00000393, -0.00000124, -0.00000148, 0.00000157, -0.00000395, -0.00000576, -0.00000152, 0.00000517, 0.00000218, 0.00000614, -0.00000674, 0.00000181, -0.00001057, 0.00000876, 0.00000181, -0.00000302, 0.00000984, -0.00001067, 0.00000188, -0.00000985, 0.00000463, 0.00001020, -0.00000445, 0.00000102, -0.00000360, -0.00000164, -0.00000095, -0.00000091, 0.00001341, -0.00000519, -0.00000184, -0.00000748, -0.00000450, 0.00001230, -0.00000840, 0.00001017, -0.00000075, -0.00000494, -0.00000419, 0.00000005, 0.00000700, 0.00000101, 0.00000240, -0.00000283, -0.00000624, -0.00000113, 0.00000152, -0.00000444, 0.00000508, 0.00000191, 0.00000802, -0.00000705, -0.00000156, 0.00000430, -0.00000583, 0.00000690, -0.00000638, -0.00000204, 0.00001760, -0.00000104, -0.00001754, -0.00000826, 0.00000754, 0.00000487, -0.00000926, 0.00000459, -0.00001236, 0.00000767, -0.00000260, -0.00001798, 0.00000554, -0.00000603, 0.00000285, -0.00001863, -0.00000351, 0.00000838, -0.00002886, -0.00004087, 0.00001673, 0.00006511, 0.00003530, 0.00004427, 0.00006464, -0.00001790, -0.00005140, -0.00001032, },
		{ -0.00001182, -0.00001287, -0.00000927, -0.00000621, -0.00000692, -0.00001118, -0.00000666, -0.00003190, -0.00003436, 0.00002482, 0.00005231, 0.00002307, 0.00003685, 0.00004941, -0.00000664, -0.00002432, -0.00002280, 0.00001313, -0.00001466, -0.00001345, 0.00000776, -0.00000530, -0.00000034, -0.00001981, 0.00000957, -0.00000198, -0.00000533, 0.00000229, -0.00001014, 0.00001025, 0.00000633, -0.00001752, -0.00001278, 0.00000483, 0.00001764, -0.00000622, -0.00000159, 0.00000267, -0.00000915, 0.00001159, -0.00001088, 0.00000169, 0.00000603, 0.00000346, -0.00000039, -0.00000215, 0.00000206, -0.00000368, -0.00000510, -0.00000383, 0.00000655, 0.00000124, 0.00000702, -0.00000675, 0.00000233, -0.00000988, 0.00000465, 0.00000640, -0.00000642, 0.00001096, -0.00000920, 0.00000122, -0.00000855, 0.00000069, 0.00001231, -0.00000382, 0.00000034, -0.00000238, -0.00000300, 0.00000011, -0.00000255, 0.00001274, -0.00000186, -0.00000353, -0.00000561, -0.00000827, 0.00001381, -0.00000719, 0.00000700, 0.00000304, -0.00000596, -0.00000409, -0.00000116, 0.00000747, 0.00000110, 0.00000194, -0.00000143, -0.00000627, -0.00000234, 0.00000231, -0.00000491, 0.00000477, 0.00000207, 0.00000744, -0.00000449, -0.00000412, 0.00000507, -0.00000528, 0.00000607, -0.00000366, -0.00000560, 0.00001648, 0.00000307, -0.00001717, -0.00000946, 0.00000455, 0.00000764, -0.00000944, 0.00000395, -0.00001083, 0.00000356, 0.00000227, -0.00002031, 0.00000536, -0.00000557, 0.00000151, -0.00001466, -0.00000876, 0.00001094, -0.00002321, -0.00004646, 0.00000559, 0.00006660, 0.00004227, 0.00004134, 0.00007094, -0.00001024, -0.00005601, },
		{ -0.00006042, -0.00001389, -0.00001262, -0.00001018, -0.00000610, -0.00000676, -0.00001099, -0.00000596, -0.00002921, -0.00003454, 0.00001767, 0.00005049, 0.00002518, 0.00002934, 0.00005205, -0.00000159, -0.00001893, -0.00002631, 0.00000953, -0.00000790, -0.00001941, 0.00001039, -0.00000602, 0.00000056, -0.00001890, 0.00000550, 0.00000233, -0.00000758, 0.00000375, -0.00001035, 0.00000699, 0.00001051, -0.00001690, -0.00001367, 0.00000126, 0.00001822, -0.00000180, -0.00000535, 0.00000518, -0.00001145, 0.00001191, -0.00000872, -0.00000118, 0.00000715, 0.00000305, 0.00000047, -0.00000287, 0.00000266, -0.00000373, -0.00000394, -0.00000624, 0.00000748, 0.00000091, 0.00000725, -0.00000611, 0.00000193, -0.00000787, -0.00000026, 0.00001056, -0.00000850, 0.00001053, -0.00000682, 0.00000006, -0.00000665, -0.00000317, 0.00001345, -0.00000236, -0.00000075, -0.00000102, -0.00000414, 0.00000070, -0.00000333, 0.00001103, 0.00000174, -0.00000488, -0.00000395, -0.00001126, 0.00001377, -0.00000458, 0.00000324, 0.00000631, -0.00000606, -0.00000444, -0.00000208, 0.00000752, 0.00000162, 0.00000120, -0.00000021, -0.00000584, -0.00000372, 0.00000294, -0.00000508, 0.00000411, 0.00000258, 0.00000650, -0.00000197, -0.00000596, 0.00000491, -0.00000416, 0.00000492, -0.00000100, -0.00000801, 0.00001401, 0.00000718, -0.00001631, -0.00001037, 0.00000156, 0.00000941, -0.00000847, 0.00000246, -0.00000858, -0.00000071, 0.00000626, -0.00002117, 0.00000385, -0.00000403, -0.00000069, -0.00001059, -0.00001281, 0.00001124, -0.00001636, -0.00004995, -0.00000684, 0.00006572, 0.00005044, 0.00003881, 0.00007602, -0.00000106, },
		{ 0.00000955, -0.00006437, -0.00001662, -0.00001217, -0.00001117, -0.00000609, -0.00000654, -0.00001091, -0.00000528, -0.00002674, -0.00003439, 0.00001161, 0.00004715, 0.00002842, 0.00002204, 0.00005341, 0.00000302, -0.00001391, -0.00002695, 0.00000338, -0.00000041, -0.00002445, 0.00001137, -0.00000514, -0.00000004, -0.00001663, 0.00000082, 0.00000613, -0.00000877, 0.00000404, -0.00000918, 0.00000283, 0.00001430, -0.00001552, -0.00001468, -0.00000154, 0.00001696, 0.00000370, -0.00000886, 0.00000702, -0.00001277, 0.00001097, -0.00000569, -0.00000419, 0.00000807, 0.00000271, 0.00000138, -0.00000369, 0.00000338, -0.00000403, -0.00000249, -0.00000838, 0.00000763, 0.00000137, 0.00000674, -0.00000479, 0.00000068, -0.00000487, -0.00000530, 0.00001365, -0.00000891, 0.00000864, -0.00000382, -0.00000131, -0.00000449, -0.00000654, 0.00001345, -0.00000018, -0.00000201, 0.00000022, -0.00000485, 0.00000068, -0.00000326, 0.00000858, 0.00000515, -0.00000559, -0.00000276, -0.00001326, 0.00001225, -0.00000101, -0.00000049, 0.00000861, -0.00000515, -0.00000515, -0.00000280, 0.00000722, 0.00000243, 0.00000038, 0.00000065, -0.00000492, -0.00000516, 0.00000330, -0.00000490, 0.00000313, 0.00000333, 0.00000541, 0.00000027, -0.00000695, 0.00000395, -0.00000278, 0.00000378, 0.00000125, -0.00000917, 0.00001053, 0.00001085, -0.00001479, -0.00001115, -0.00000118, 0.00001007, -0.00000652, 0.00000045, -0.00000606, -0.00000463, 0.00000901, -0.00002053, 0.00000127, -0.00000183, -0.00000315, -0.00000704, -0.00001526, 0.00000958, -0.00000922, -0.00005081, -0.00001998, 0.00006194, 0.00005950, 0.00003710, 0.00007972, },
		{ 0.00008187, 0.00002146, -0.00006762, -0.00002008, -0.00001155, -0.00001220, -0.00000617, -0.00000629, -0.00001088, -0.00000472, -0.00002428, -0.00003422, 0.00000684, 0.00004243, 0.00003235, 0.00001543, 0.00005328, 0.00000761, -0.00001005, -0.00002457, -0.00000456, 0.00000669, -0.00002769, 0.00001026, -0.00000261, -0.00000186, -0.00001353, -0.00000376, 0.00000882, -0.00000860, 0.00000309, -0.00000687, -0.00000171, 0.00001716, -0.00001309, -0.00001607, -0.00000332, 0.00001399, 0.00000968, -0.00001153, 0.00000781, -0.00001292, 0.00000882, -0.00000203, -0.00000706, 0.00000866, 0.00000244, 0.00000235, -0.00000459, 0.00000417, -0.00000446, -0.00000106, -0.00000990, 0.00000678, 0.00000267, 0.00000557, -0.00000291, -0.00000122, -0.00000136, -0.00000979, 0.00001513, -0.00000756, 0.00000557, -0.00000061, -0.00000257, -0.00000240, -0.00000911, 0.00001228, 0.00000246, -0.00000316, 0.00000109, -0.00000498, 0.00000005, -0.00000254, 0.00000579, 0.00000796, -0.00000545, -0.00000215, -0.00001423, 0.00000949, 0.00000298, -0.00000363, 0.00000964, -0.00000328, -0.00000603, -0.00000342, 0.00000668, 0.00000336, -0.00000032, 0.00000104, -0.00000360, -0.00000648, 0.00000327, -0.00000435, 0.00000190, 0.00000417, 0.00000436, 0.00000204, -0.00000712, 0.00000246, -0.00000146, 0.00000293, 0.00000292, -0.00000917, 0.00000650, 0.00001366, -0.00001251, -0.00001184, -0.00000350, 0.00000965, -0.00000391, -0.00000166, -0.00000369, -0.00000781, 0.00001033, -0.00001856, -0.00000197, 0.00000051, -0.00000534, -0.00000445, -0.00001610, 0.00000652, -0.00000271, -0.00004887, -0.00003295, 0.00005481, 0.00006895, 0.00003666, },
		{ 0.00003798, 0.00008238, 0.00003448, -0.00006989, -0.00002435, -0.00001080, -0.00001324, -0.00000633, -0.00000609, -0.00001078, -0.00000444, -0.00002167, -0.00003425, 0.00000340, 0.00003662, 0.00003642, 0.00001004, 0.00005144, 0.00001259, -0.00000785, -0.00001948, -0.00001315, 0.00001224, -0.00002846, 0.00000697, 0.00000123, -0.00000433, -0.00001031, -0.00000756, 0.00000997, -0.00000700, 0.00000113, -0.00000389, -0.00000597, 0.00001859, -0.00000943, -0.00001792, -0.00000402, 0.00000972, 0.00001539, -0.00001278, 0.00000732, -0.00001186, 0.00000566, 0.00000192, -0.00000957, 0.00000888, 0.00000223, 0.00000338, -0.00000552, 0.00000489, -0.00000478, 0.00000004, -0.00001050, 0.00000483, 0.00000472, 0.00000392, -0.00000069, -0.00000342, 0.00000212, -0.00001315, 0.00001473, -0.00000462, 0.00000180, 0.00000233, -0.00000342, -0.00000066, -0.00001071, 0.00001007, 0.00000524, -0.00000390, 0.00000144, -0.00000449, -0.00000108, -0.00000142, 0.00000308, 0.00000985, -0.00000438, -0.00000207, -0.00001429, 0.00000588, 0.00000681, -0.00000573, 0.00000933, -0.00000071, -0.00000681, -0.00000410, 0.00000603, 0.00000426, -0.00000075, 0.00000095, -0.00000202, -0.00000750, 0.00000279, -0.00000348, 0.00000053, 0.00000496, 0.00000352, 0.00000325, -0.00000656, 0.00000072, -0.00000049, 0.00000251, 0.00000397, -0.00000825, 0.00000243, 0.00001531, -0.00000952, -0.00001241, -0.00000537, 0.00000833, -0.00000100, -0.00000350, -0.00000179, -0.00001002, 0.00001029, -0.00001562, -0.00000540, 0.00000253, -0.00000690, -0.00000298, -0.00001557, 0.00000278, 0.00000243, -0.00004431, -0.00004473, 0.00004411, 0.00007809, },
		{ 0.00008599, 0.00004150, 0.00008127, 0.00004833, -0.00007089, -0.00002949, -0.00000992, -0.00001432, -0.00000650, -0.00000604, -0.00001048, -0.00000454, -0.00001879, -0.00003460, 0.00000116, 0.00003015, 0.00004004, 0.00000633, 0.00004771, 0.00001821, -0.00000742, -0.00001248, -0.00002107, 0.00001527, -0.00002648, 0.00000183, 0.00000577, -0.00000676, -0.00000766, -0.00001004, 0.00000943, -0.00000419, -0.00000141, -0.00000083, -0.00000932, 0.00001823, -0.00000455, -0.00002016, -0.00000385, 0.00000475, 0.00002007, -0.00001222, 0.00000550, -0.00000975, 0.00000178, 0.00000581, -0.00001150, 0.00000867, 0.00000206, 0.00000444, -0.00000635, 0.00000538, -0.00000476, 0.00000052, -0.00000999, 0.00000185, 0.00000725, 0.00000210, 0.00000153, -0.00000552, 0.00000502, -0.00001493, 0.00001242, -0.00000051, -0.00000207, 0.00000455, -0.00000362, 0.00000057, -0.00001130, 0.00000710, 0.00000777, -0.00000400, 0.00000121, -0.00000348, -0.00000249, -0.00000024, 0.00000081, 0.00001063, -0.00000248, -0.00000236, -0.00001366, 0.00000188, 0.00000997, -0.00000652, 0.00000779, 0.00000220, -0.00000721, -0.00000496, 0.00000533, 0.00000504, -0.00000086, 0.00000044, -0.00000039, -0.00000806, 0.00000184, -0.00000236, -0.00000085, 0.00000555, 0.00000300, 0.00000391, -0.00000548, -0.00000094, -0.00000005, 0.00000255, 0.00000452, -0.00000676, -0.00000123, 0.00001564, -0.00000601, -0.00001273, -0.00000684, 0.00000634, 0.00000181, -0.00000477, -0.00000053, -0.00001125, 0.00000912, -0.00001216, -0.00000855, 0.00000392, -0.00000766, -0.00000254, -0.00001412, -0.00000092, 0.00000576, -0.00003775, -0.00005421, 0.00002991, },
		{ 0.00001270, 0.00009153, 0.00004757, 0.00007869, 0.00006265, -0.00007025, -0.00003559, -0.00000892, -0.00001550, -0.00000658, -0.00000624, -0.00000990, -0.00000510, -0.00001562, -0.00003521, -0.00000015, 0.00002358, 0.00004258, 0.00000468, 0.00004209, 0.00002445, -0.00000845, -0.00000464, -0.00002708, 0.00001519, -0.00002195, -0.00000449, 0.00001019, -0.00000841, -0.00000615, -0.00001095, 0.00000734, -0.00000063, -0.00000394, 0.00000169, -0.00001124, 0.00001596, 0.00000132, -0.00002251, -0.00000318, -0.00000021, 0.00002304, -0.00000965, 0.00000251, -0.00000688, -0.00000241, 0.00000928, -0.00001270, 0.00000806, 0.00000194, 0.00000541, -0.00000692, 0.00000540, -0.00000417, 0.00000019, -0.00000836, -0.00000191, 0.00000988, 0.00000049, 0.00000341, -0.00000714, 0.00000694, -0.00001492, 0.00000846, 0.00000418, -0.00000542, 0.00000570, -0.00000308, 0.00000127, -0.00001101, 0.00000371, 0.00000972, -0.00000336, 0.00000045, -0.00000213, -0.00000392, 0.00000069, -0.00000074, 0.00001028, 0.00000002, -0.00000278, -0.00001263, -0.00000208, 0.00001209, -0.00000595, 0.00000534, 0.00000502, -0.00000698, -0.00000603, 0.00000464, 0.00000563, -0.00000064, -0.00000037, 0.00000110, -0.00000805, 0.00000047, -0.00000115, -0.00000208, 0.00000584, 0.00000282, 0.00000410, -0.00000412, -0.00000227, -0.00000024, 0.00000293, 0.00000478, -0.00000504, -0.00000413, 0.00001469, -0.00000227, -0.00001266, -0.00000799, 0.00000395, 0.00000419, -0.00000531, 0.00000006, -0.00001161, 0.00000717, -0.00000865, -0.00001105, 0.00000450, -0.00000766, -0.00000286, -0.00001225, -0.00000403, 0.00000721, -0.00003008, -0.00006044, },
		{ -0.00006281, -0.00000658, 0.00009352, 0.00005634, 0.00007500, 0.00007695, -0.00006759, -0.00004274, -0.00000776, -0.00001682, -0.00000649, -0.00000676, -0.00000901, -0.00000606, -0.00001227, -0.00003592, -0.00000094, 0.00001749, 0.00004356, 0.00000525, 0.00003479, 0.00003104, -0.00001032, 0.00000278, -0.00003016, 0.00001188, -0.00001551, -0.00001104, 0.00001365, -0.00000868, -0.00000606, -0.00001032, 0.00000414, 0.00000301, -0.00000583, 0.00000312, -0.00001142, 0.00001191, 0.00000772, -0.00002453, -0.00000253, -0.00000445, 0.00002382, -0.00000514, -0.00000127, -0.00000366, -0.00000651, 0.00001204, -0.00001306, 0.00000705, 0.00000189, 0.00000617, -0.00000702, 0.00000478, -0.00000286, -0.00000100, -0.00000579, -0.00000605, 0.00001215, -0.00000051, 0.00000463, -0.00000798, 0.00000762, -0.00001319, 0.00000335, 0.00000874, -0.00000770, 0.00000558, -0.00000185, 0.00000154, -0.00001004, 0.00000028, 0.00001082, -0.00000196, -0.00000067, -0.00000069, -0.00000508, 0.00000112, -0.00000146, 0.00000895, 0.00000278, -0.00000302, -0.00001149, -0.00000562, 0.00001296, -0.00000418, 0.00000241, 0.00000732, -0.00000598, -0.00000728, 0.00000393, 0.00000604, -0.00000016, -0.00000132, 0.00000226, -0.00000743, -0.00000119, -0.00000001, -0.00000303, 0.00000577, 0.00000298, 0.00000395, -0.00000273, -0.00000308, -0.00000100, 0.00000345, 0.00000501, -0.00000342, -0.00000609, 0.00001267, 0.00000132, -0.00001203, -0.00000889, 0.00000142, 0.00000593, -0.00000510, 0.00000009, -0.00001134, 0.00000481, -0.00000548, -0.00001272, 0.00000432, -0.00000709, -0.00000359, -0.00001043, -0.00000623, 0.00000706, -0.00002234, },
		{ -0.00001552, -0.00006112, -0.00002655, 0.00009081, 0.00006764, 0.00007077, 0.00009063, -0.00006249, -0.00005098, -0.00000648, -0.00001830, -0.00000623, -0.00000757, -0.00000787, -0.00000731, -0.00000893, -0.00003646, -0.00000163, 0.00001237, 0.00004266, 0.00000796, 0.00002628, 0.00003746, -0.00001219, 0.00000866, -0.00002974, 0.00000577, -0.00000818, -0.00001678, 0.00001540, -0.00000724, -0.00000739, -0.00000854, 0.00000050, 0.00000601, -0.00000650, 0.00000312, -0.00000983, 0.00000649, 0.00001400, -0.00002568, -0.00000242, -0.00000737, 0.00002220, 0.00000095, -0.00000531, -0.00000055, -0.00001012, 0.00001385, -0.00001257, 0.00000569, 0.00000198, 0.00000656, -0.00000649, 0.00000340, -0.00000081, -0.00000293, -0.00000262, -0.00001005, 0.00001357, -0.00000059, 0.00000499, -0.00000788, 0.00000703, -0.00001003, -0.00000226, 0.00001248, -0.00000852, 0.00000420, -0.00000015, 0.00000156, -0.00000866, -0.00000282, 0.00001092, 0.00000005, -0.00000187, 0.00000056, -0.00000575, 0.00000093, -0.00000138, 0.00000692, 0.00000541, -0.00000283, -0.00001046, -0.00000852, 0.00001255, -0.00000154, -0.00000051, 0.00000875, -0.00000418, -0.00000860, 0.00000313, 0.00000631, 0.00000049, -0.00000225, 0.00000296, -0.00000624, -0.00000297, 0.00000089, -0.00000362, 0.00000533, 0.00000337, 0.00000365, -0.00000149, -0.00000332, -0.00000216, 0.00000385, 0.00000541, -0.00000209, -0.00000710, 0.00000994, 0.00000437, -0.00001076, -0.00000958, -0.00000109, 0.00000693, -0.00000425, -0.00000021, -0.00001070, 0.00000240, -0.00000291, -0.00001354, 0.00000356, -0.00000625, -0.00000439, -0.00000897, -0.00000748, 0.00000585, },
		{ 0.00000426, -0.00001041, -0.00005573, -0.00004557, 0.00008252, 0.00008086, 0.00006681, 0.00010303, -0.00005458, -0.00006029, -0.00000516, -0.00001991, -0.00000584, -0.00000858, -0.00000665, -0.00000863, -0.00000587, -0.00003656, -0.00000262, 0.00000860, 0.00003979, 0.00001245, 0.00001722, 0.00004300, -0.00001319, 0.00001212, -0.00002582, -0.00000220, -0.00000118, -0.00002077, 0.00001500, -0.00000412, -0.00000975, -0.00000624, -0.00000281, 0.00000773, -0.00000562, 0.00000163, -0.00000676, 0.00000037, 0.00001944, -0.00002544, -0.00000329, -0.00000859, 0.00001826, 0.00000801, -0.00000899, 0.00000201, -0.00001295, 0.00001457, -0.00001125, 0.00000402, 0.00000228, 0.00000643, -0.00000522, 0.00000124, 0.00000185, -0.00000533, 0.00000070, -0.00001336, 0.00001376, 0.00000042, 0.00000443, -0.00000687, 0.00000533, -0.00000593, -0.00000764, 0.00001483, -0.00000772, 0.00000178, 0.00000169, 0.00000158, -0.00000716, -0.00000534, 0.00001001, 0.00000242, -0.00000289, 0.00000139, -0.00000580, 0.00000011, -0.00000066, 0.00000456, 0.00000753, -0.00000203, -0.00000965, -0.00001066, 0.00001103, 0.00000154, -0.00000296, 0.00000913, -0.00000170, -0.00000978, 0.00000216, 0.00000649, 0.00000122, -0.00000301, 0.00000313, -0.00000460, -0.00000466, 0.00000143, -0.00000381, 0.00000460, 0.00000388, 0.00000334, -0.00000055, -0.00000301, -0.00000349, 0.00000387, 0.00000614, -0.00000110, -0.00000731, 0.00000692, 0.00000659, -0.00000886, -0.00001003, -0.00000346, 0.00000718, -0.00000294, -0.00000062, -0.00000995, 0.00000016, -0.00000104, -0.00001364, 0.00000249, -0.00000541, -0.00000500, -0.00000798, -0.00000799, },
		{ -0.00000813, 0.00000294, -0.00000743, -0.00004750, -0.00006186, 0.00006824, 0.00009495, 0.00006408, 0.00011352, -0.00004359, -0.00007053, -0.00000398, -0.00002149, -0.00000548, -0.00000960, -0.00000557, -0.00000975, -0.00000335, -0.00003593, -0.00000422, 0.00000634, 0.00003514, 0.00001814, 0.00000842, 0.00004695, -0.00001260, 0.00001274, -0.00001893, -0.00001076, 0.00000428, -0.00002234, 0.00001241, 0.00000029, -0.00001250, -0.00000419, -0.00000505, 0.00000772, -0.00000311, -0.00000111, -0.00000277, -0.00000563, 0.00002329, -0.00002342, -0.00000541, -0.00000800, 0.00001242, 0.00001527, -0.00001168, 0.00000366, -0.00001477, 0.00001417, -0.00000919, 0.00000207, 0.00000288, 0.00000568, -0.00000319, -0.00000155, 0.00000486, -0.00000779, 0.00000364, -0.00001551, 0.00001245, 0.00000255, 0.00000308, -0.00000513, 0.00000287, -0.00000149, -0.00001212, 0.00001541, -0.00000537, -0.00000129, 0.00000331, 0.00000184, -0.00000575, -0.00000711, 0.00000825, 0.00000481, -0.00000344, 0.00000166, -0.00000520, -0.00000122, 0.00000041, 0.00000227, 0.00000884, -0.00000056, -0.00000907, -0.00001209, 0.00000867, 0.00000461, -0.00000458, 0.00000840, 0.00000120, -0.00001059, 0.00000092, 0.00000664, 0.00000193, -0.00000351, 0.00000279, -0.00000267, -0.00000607, 0.00000151, -0.00000364, 0.00000366, 0.00000438, 0.00000314, 0.00000005, -0.00000230, -0.00000471, 0.00000333, 0.00000720, -0.00000039, -0.00000698, 0.00000401, 0.00000777, -0.00000647, -0.00001016, -0.00000566, 0.00000678, -0.00000143, -0.00000092, -0.00000925, -0.00000176, 0.00000019, -0.00001326, 0.00000140, -0.00000480, -0.00000534, -0.00000739, },
		{ -0.00000695, -0.00000835, 0.00000237, -0.00000655, -0.00003771, -0.00007382, 0.00004813, 0.00010842, 0.00006361, 0.00012160, -0.00002945, -0.00008134, -0.00000329, -0.00002280, -0.00000538, -0.00001036, -0.00000489, -0.00001045, -0.00000158, -0.00003444, -0.00000655, 0.00000555, 0.00002912, 0.00002429, 0.00000070, 0.00004872, -0.00001000, 0.00001060, -0.00001010, -0.00001850, 0.00000724, -0.00002127, 0.00000802, 0.00000526, -0.00001483, -0.00000313, -0.00000570, 0.00000584, 0.00000074, -0.00000456, 0.00000141, -0.00001067, 0.00002497, -0.00001942, -0.00000879, -0.00000580, 0.00000537, 0.00002183, -0.00001283, 0.00000419, -0.00001552, 0.00001271, -0.00000653, -0.00000012, 0.00000383, 0.00000427, -0.00000052, -0.00000471, 0.00000783, -0.00000984, 0.00000571, -0.00001612, 0.00000960, 0.00000560, 0.00000121, -0.00000299, 0.00000009, 0.00000267, -0.00001515, 0.00001409, -0.00000181, -0.00000450, 0.00000433, 0.00000248, -0.00000455, -0.00000813, 0.00000587, 0.00000689, -0.00000333, 0.00000133, -0.00000405, -0.00000283, 0.00000153, 0.00000039, 0.00000918, 0.00000149, -0.00000860, -0.00001294, 0.00000581, 0.00000723, -0.00000516, 0.00000671, 0.00000419, -0.00001081, -0.00000061, 0.00000675, 0.00000259, -0.00000373, 0.00000201, -0.00000066, -0.00000704, 0.00000112, -0.00000320, 0.00000266, 0.00000474, 0.00000314, 0.00000033, -0.00000136, -0.00000557, 0.00000214, 0.00000851, 0.00000023, -0.00000641, 0.00000156, 0.00000789, -0.00000383, -0.00000983, -0.00000768, 0.00000586, 0.00000007, -0.00000097, -0.00000866, -0.00000336, 0.00000090, -0.00001267, 0.00000051, -0.00000447, -0.00000548, },
		{ -0.00000561, -0.00000637, -0.00000897, 0.00000275, -0.00000731, -0.00002784, -0.00008020, 0.00002311, 0.00011942, 0.00006629, 0.00012698, -0.00001233, -0.00009219, -0.00000353, -0.00002353, -0.00000583, -0.00001063, -0.00000481, -0.00001054, -0.00000068, -0.00003205, -0.00000956, 0.00000596, 0.00002231, 0.00003011, -0.00000523, 0.00004795, -0.00000534, 0.00000625, -0.00000064, -0.00002411, 0.00000716, -0.00001784, 0.00000260, 0.00000986, -0.00001592, -0.00000365, -0.00000454, 0.00000233, 0.00000537, -0.00000798, 0.00000494, -0.00001400, 0.00002412, -0.00001353, -0.00001318, -0.00000245, -0.00000205, 0.00002689, -0.00001209, 0.00000356, -0.00001525, 0.00001032, -0.00000340, -0.00000247, 0.00000514, 0.00000230, 0.00000256, -0.00000787, 0.00001031, -0.00001105, 0.00000653, -0.00001506, 0.00000539, 0.00000918, -0.00000077, -0.00000084, -0.00000253, 0.00000603, -0.00001642, 0.00001102, 0.00000245, -0.00000730, 0.00000445, 0.00000356, -0.00000358, -0.00000848, 0.00000320, 0.00000835, -0.00000247, 0.00000048, -0.00000257, -0.00000443, 0.00000237, -0.00000081, 0.00000854, 0.00000392, -0.00000806, -0.00001339, 0.00000280, 0.00000910, -0.00000467, 0.00000434, 0.00000690, -0.00001027, -0.00000241, 0.00000681, 0.00000318, -0.00000366, 0.00000091, 0.00000122, -0.00000744, 0.00000029, -0.00000260, 0.00000172, 0.00000489, 0.00000333, 0.00000040, -0.00000043, -0.00000589, 0.00000034, 0.00000984, 0.00000098, -0.00000588, -0.00000022, 0.00000708, -0.00000126, -0.00000895, -0.00000950, 0.00000454, 0.00000137, -0.00000071, -0.00000817, -0.00000473, 0.00000129, -0.00001209, -0.00000008, -0.00000436, },
		{ -0.00000429, -0.00000598, -0.00000540, -0.00001017, 0.00000394, -0.00000894, -0.00001937, -0.00008038, -0.00000523, 0.00012597, 0.00007273, 0.00012976, 0.00000731, -0.00010234, -0.00000523, -0.00002335, -0.00000707, -0.00001020, -0.00000546, -0.00000999, -0.00000060, -0.00002891, -0.00001305, 0.00000717, 0.00001535, 0.00003489, -0.00000890, 0.00004463, 0.00000101, 0.00000061, 0.00000807, -0.00002664, 0.00000405, -0.00001283, -0.00000280, 0.00001318, -0.00001516, -0.00000598, -0.00000174, -0.00000225, 0.00000996, -0.00001057, 0.00000711, -0.00001510, 0.00002073, -0.00000615, -0.00001808, 0.00000138, -0.00000893, 0.00002974, -0.00000933, 0.00000191, -0.00001412, 0.00000721, 0.00000002, -0.00000493, 0.00000675, -0.00000006, 0.00000569, -0.00001059, 0.00001186, -0.00001105, 0.00000588, -0.00001240, 0.00000022, 0.00001279, -0.00000241, 0.00000093, -0.00000460, 0.00000819, -0.00001585, 0.00000662, 0.00000679, -0.00000919, 0.00000353, 0.00000502, -0.00000277, -0.00000833, 0.00000056, 0.00000897, -0.00000090, -0.00000069, -0.00000102, -0.00000575, 0.00000271, -0.00000124, 0.00000705, 0.00000642, -0.00000725, -0.00001364, -0.00000007, 0.00001005, -0.00000324, 0.00000166, 0.00000898, -0.00000890, -0.00000440, 0.00000675, 0.00000371, -0.00000336, -0.00000037, 0.00000281, -0.00000723, -0.00000085, -0.00000200, 0.00000095, 0.00000479, 0.00000367, 0.00000038, 0.00000032, -0.00000557, -0.00000190, 0.00001091, 0.00000209, -0.00000558, -0.00000130, 0.00000563, 0.00000091, -0.00000743, -0.00001108, 0.00000293, 0.00000238, -0.00000018, -0.00000768, -0.00000597, 0.00000153, -0.00001166, -0.00000040, },
		{ -0.00000060, -0.00000400, -0.00000679, -0.00000393, -0.00001189, 0.00000552, -0.00001046, -0.00001352, -0.00007449, -0.00003467, 0.00012620, 0.00008302, 0.00013042, 0.00002874, -0.00011091, -0.00000894, -0.00002197, -0.00000926, -0.00000900, -0.00000682, -0.00000889, -0.00000118, -0.00002525, -0.00001671, 0.00000872, 0.00000885, 0.00003810, -0.00001013, 0.00003907, 0.00000836, -0.00000521, 0.00001480, -0.00002565, -0.00000147, -0.00000736, -0.00000706, 0.00001443, -0.00001224, -0.00000998, 0.00000218, -0.00000703, 0.00001361, -0.00001158, 0.00000737, -0.00001380, 0.00001511, 0.00000206, -0.00002280, 0.00000495, -0.00001443, 0.00002994, -0.00000468, -0.00000045, -0.00001238, 0.00000359, 0.00000353, -0.00000737, 0.00000851, -0.00000250, 0.00000846, -0.00001241, 0.00001214, -0.00000964, 0.00000380, -0.00000846, -0.00000535, 0.00001585, -0.00000329, 0.00000199, -0.00000581, 0.00000901, -0.00001365, 0.00000144, 0.00001056, -0.00000981, 0.00000159, 0.00000665, -0.00000197, -0.00000791, -0.00000175, 0.00000866, 0.00000121, -0.00000192, 0.00000033, -0.00000655, 0.00000241, -0.00000093, 0.00000499, 0.00000867, -0.00000597, -0.00001382, -0.00000261, 0.00001004, -0.00000113, -0.00000092, 0.00001018, -0.00000675, -0.00000639, 0.00000647, 0.00000421, -0.00000288, -0.00000169, 0.00000396, -0.00000644, -0.00000215, -0.00000153, 0.00000040, 0.00000448, 0.00000407, 0.00000042, 0.00000075, -0.00000466, -0.00000433, 0.00001143, 0.00000371, -0.00000556, -0.00000177, 0.00000388, 0.00000237, -0.00000530, -0.00001235, 0.00000108, 0.00000306, 0.00000052, -0.00000708, -0.00000718, 0.00000173, -0.00001141, },
		{ -0.00001126, -0.00000090, -0.00000327, -0.00000817, -0.00000201, -0.00001383, 0.00000689, -0.00001094, -0.00001103, -0.00006344, -0.00006263, 0.00011865, 0.00009663, 0.00012986, 0.00005103, -0.00011699, -0.00001510, -0.00001926, -0.00001242, -0.00000712, -0.00000872, -0.00000747, -0.00000215, -0.00002141, -0.00002019, 0.00001016, 0.00000328, 0.00003944, -0.00000905, 0.00003188, 0.00001585, -0.00001014, 0.00001876, -0.00002135, -0.00000830, -0.00000276, -0.00000924, 0.00001323, -0.00000731, -0.00001510, 0.00000635, -0.00001105, 0.00001548, -0.00001049, 0.00000553, -0.00001029, 0.00000793, 0.00001023, -0.00002659, 0.00000753, -0.00001790, 0.00002738, 0.00000146, -0.00000310, -0.00001031, -0.00000029, 0.00000695, -0.00000965, 0.00001020, -0.00000468, 0.00001044, -0.00001299, 0.00001094, -0.00000685, 0.00000053, -0.00000376, -0.00001066, 0.00001782, -0.00000309, 0.00000215, -0.00000606, 0.00000851, -0.00001020, -0.00000383, 0.00001320, -0.00000900, -0.00000116, 0.00000817, -0.00000098, -0.00000740, -0.00000354, 0.00000745, 0.00000359, -0.00000290, 0.00000124, -0.00000672, 0.00000148, -0.00000004, 0.00000270, 0.00001033, -0.00000413, -0.00001399, -0.00000475, 0.00000919, 0.00000130, -0.00000303, 0.00001035, -0.00000398, -0.00000817, 0.00000590, 0.00000473, -0.00000230, -0.00000292, 0.00000462, -0.00000517, -0.00000344, -0.00000131, 0.00000010, 0.00000401, 0.00000440, 0.00000062, 0.00000081, -0.00000328, -0.00000662, 0.00001115, 0.00000589, -0.00000575, -0.00000188, 0.00000223, 0.00000293, -0.00000271, -0.00001316, -0.00000096, 0.00000342, 0.00000127, -0.00000625, -0.00000842, 0.00000191, },
		{ 0.00000202, -0.00001103, -0.00000151, -0.00000196, -0.00001007, 0.00000011, -0.00001554, 0.00000738, -0.00000967, -0.00001204, -0.00004881, -0.00008646, 0.00010254, 0.00011231, 0.00012928, 0.00007319, -0.00011974, -0.00002398, -0.00001530, -0.00001637, -0.00000480, -0.00001085, -0.00000604, -0.00000319, -0.00001770, -0.00002320, 0.00001119, -0.00000107, 0.00003890, -0.00000608, 0.00002387, 0.00002256, -0.00001335, 0.00001965, -0.00001450, -0.00001502, -0.00000023, -0.00000874, 0.00000965, -0.00000094, -0.00002045, 0.00000976, -0.00001337, 0.00001497, -0.00000712, 0.00000177, -0.00000513, 0.00000010, 0.00001742, -0.00002876, 0.00000854, -0.00001898, 0.00002225, 0.00000847, -0.00000554, -0.00000819, -0.00000420, 0.00001007, -0.00001158, 0.00001156, -0.00000621, 0.00001125, -0.00001210, 0.00000827, -0.00000290, -0.00000347, 0.00000105, -0.00001505, 0.00001828, -0.00000165, 0.00000142, -0.00000542, 0.00000695, -0.00000606, -0.00000853, 0.00001432, -0.00000681, -0.00000437, 0.00000926, 0.00000034, -0.00000696, -0.00000470, 0.00000552, 0.00000589, -0.00000337, 0.00000158, -0.00000622, 0.00000005, 0.00000117, 0.00000053, 0.00001117, -0.00000173, -0.00001411, -0.00000649, 0.00000771, 0.00000368, -0.00000438, 0.00000952, -0.00000085, -0.00000952, 0.00000497, 0.00000526, -0.00000166, -0.00000400, 0.00000477, -0.00000355, -0.00000455, -0.00000140, -0.00000001, 0.00000350, 0.00000458, 0.00000106, 0.00000052, -0.00000164, -0.00000847, 0.00000993, 0.00000851, -0.00000594, -0.00000191, 0.00000099, 0.00000256, 0.00000011, -0.00001335, -0.00000320, 0.00000351, 0.00000193, -0.00000515, -0.00000966, },
		{ -0.00001080, 0.00000194, -0.00001058, -0.00000262, -0.00000007, -0.00001225, 0.00000202, -0.00001647, 0.00000644, -0.00000632, -0.00001605, -0.00003270, -0.00010383, 0.00007803, 0.00012811, 0.00013004, 0.00009431, -0.00011856, -0.00003555, -0.00001040, -0.00002074, -0.00000244, -0.00001283, -0.00000494, -0.00000402, -0.00001436, -0.00002555, 0.00001161, -0.00000415, 0.00003672, -0.00000182, 0.00001586, 0.00002775, -0.00001438, 0.00001775, -0.00000635, -0.00002016, -0.00000067, -0.00000550, 0.00000428, 0.00000589, -0.00002494, 0.00001144, -0.00001330, 0.00001184, -0.00000169, -0.00000338, 0.00000086, -0.00000733, 0.00002275, -0.00002878, 0.00000766, -0.00001761, 0.00001507, 0.00001559, -0.00000731, -0.00000625, -0.00000791, 0.00001268, -0.00001293, 0.00001229, -0.00000674, 0.00001066, -0.00000973, 0.00000437, 0.00000174, -0.00000761, 0.00000531, -0.00001795, 0.00001699, 0.00000098, -0.00000004, -0.00000411, 0.00000471, -0.00000182, -0.00001212, 0.00001375, -0.00000351, -0.00000759, 0.00000961, 0.00000208, -0.00000665, -0.00000526, 0.00000313, 0.00000779, -0.00000315, 0.00000132, -0.00000517, -0.00000168, 0.00000239, -0.00000119, 0.00001107, 0.00000112, -0.00001407, -0.00000793, 0.00000584, 0.00000570, -0.00000484, 0.00000784, 0.00000231, -0.00001024, 0.00000365, 0.00000580, -0.00000098, -0.00000486, 0.00000444, -0.00000175, -0.00000534, -0.00000179, -0.00000000, 0.00000303, 0.00000452, 0.00000174, -0.00000001, -0.00000000, -0.00000961, 0.00000775, 0.00001134, -0.00000587, -0.00000215, 0.00000037, 0.00000136, 0.00000281, -0.00001277, -0.00000558, 0.00000336, 0.00000243, -0.00000380, },
		{ -0.00000228, -0.00001171, 0.00000153, -0.00000976, -0.00000426, 0.00000225, -0.00001436, 0.00000325, -0.00001618, 0.00000376, -0.00000100, -0.00002200, -0.00001736, -0.00011301, 0.00004630, 0.00014160, 0.00013339, 0.00011378, -0.00011316, -0.00004947, -0.00000512, -0.00002503, -0.00000047, -0.00001431, -0.00000442, -0.00000448, -0.00001151, -0.00002721, 0.00001143, -0.00000607, 0.00003329, 0.00000306, 0.00000858, 0.00003098, -0.00001325, 0.00001384, 0.00000164, -0.00002246, -0.00000443, -0.00000005, -0.00000185, 0.00001199, -0.00002752, 0.00001062, -0.00001051, 0.00000628, 0.00000518, -0.00000905, 0.00000665, -0.00001336, 0.00002553, -0.00002639, 0.00000486, -0.00001408, 0.00000660, 0.00002204, -0.00000800, -0.00000464, -0.00001126, 0.00001458, -0.00001351, 0.00001213, -0.00000605, 0.00000862, -0.00000607, -0.00000034, 0.00000651, -0.00001124, 0.00000840, -0.00001902, 0.00001397, 0.00000452, -0.00000190, -0.00000247, 0.00000224, 0.00000198, -0.00001422, 0.00001158, 0.00000045, -0.00001033, 0.00000900, 0.00000422, -0.00000644, -0.00000535, 0.00000061, 0.00000899, -0.00000216, 0.00000054, -0.00000377, -0.00000341, 0.00000334, -0.00000226, 0.00001006, 0.00000417, -0.00001370, -0.00000919, 0.00000384, 0.00000711, -0.00000441, 0.00000558, 0.00000516, -0.00001019, 0.00000201, 0.00000627, -0.00000027, -0.00000549, 0.00000372, 0.00000005, -0.00000571, -0.00000246, 0.00000001, 0.00000271, 0.00000420, 0.00000259, -0.00000061, 0.00000139, -0.00000988, 0.00000476, 0.00001406, -0.00000525, -0.00000280, 0.00000043, -0.00000040, 0.00000505, -0.00001134, -0.00000802, 0.00000297, 0.00000274, },
		{ 0.00000289, -0.00000074, -0.00001224, 0.00000071, -0.00000852, -0.00000634, 0.00000468, -0.00001595, 0.00000339, -0.00001437, -0.00000063, 0.00000567, -0.00002847, -0.00000493, -0.00011321, 0.00000953, 0.00015013, 0.00014017, 0.00013142, -0.00010364, -0.00006508, -0.00000018, -0.00002867, 0.00000069, -0.00001498, -0.00000463, -0.00000454, -0.00000910, -0.00002827, 0.00001077, -0.00000705, 0.00002903, 0.00000797, 0.00000253, 0.00003217, -0.00001041, 0.00000905, 0.00000805, -0.00002113, -0.00001117, 0.00000652, -0.00000740, 0.00001618, -0.00002732, 0.00000691, -0.00000514, -0.00000104, 0.00001253, -0.00001423, 0.00001126, -0.00001715, 0.00002540, -0.00002167, 0.00000042, -0.00000895, -0.00000227, 0.00002705, -0.00000737, -0.00000345, -0.00001408, 0.00001559, -0.00001312, 0.00001089, -0.00000404, 0.00000527, -0.00000153, -0.00000527, 0.00001075, -0.00001376, 0.00000991, -0.00001813, 0.00000947, 0.00000853, -0.00000376, -0.00000088, -0.00000003, 0.00000489, -0.00001468, 0.00000811, 0.00000454, -0.00001218, 0.00000735, 0.00000665, -0.00000620, -0.00000514, -0.00000174, 0.00000929, -0.00000047, -0.00000055, -0.00000227, -0.00000489, 0.00000378, -0.00000258, 0.00000828, 0.00000715, -0.00001284, -0.00001036, 0.00000189, 0.00000781, -0.00000321, 0.00000307, 0.00000741, -0.00000933, 0.00000015, 0.00000661, 0.00000050, -0.00000591, 0.00000271, 0.00000172, -0.00000562, -0.00000328, -0.00000008, 0.00000259, 0.00000365, 0.00000350, -0.00000111, 0.00000233, -0.00000924, 0.00000122, 0.00001627, -0.00000384, -0.00000395, 0.00000105, -0.00000236, 0.00000650, -0.00000906, -0.00001038, 0.00000231, },
		{ 0.00000130, 0.00000294, 0.00000069, -0.00001229, -0.00000055, -0.00000694, -0.00000865, 0.00000686, -0.00001661, 0.00000216, -0.00001106, -0.00000631, 0.00001277, -0.00003390, 0.00000296, -0.00010469, -0.00002927, 0.00015117, 0.00015062, 0.00014759, -0.00009047, -0.00008150, 0.00000363, -0.00003116, 0.00000074, -0.00001475, -0.00000551, -0.00000433, -0.00000699, -0.00002891, 0.00000981, -0.00000737, 0.00002436, 0.00001250, -0.00000212, 0.00003161, -0.00000665, 0.00000458, 0.00001176, -0.00001607, -0.00001993, 0.00001273, -0.00001100, 0.00001747, -0.00002391, 0.00000045, 0.00000217, -0.00000912, 0.00001922, -0.00001790, 0.00001384, -0.00001818, 0.00002234, -0.00001502, -0.00000511, -0.00000298, -0.00001059, 0.00003003, -0.00000532, -0.00000267, -0.00001625, 0.00001555, -0.00001163, 0.00000849, -0.00000080, 0.00000096, 0.00000332, -0.00000977, 0.00001387, -0.00001475, 0.00000963, -0.00001543, 0.00000396, 0.00001248, -0.00000518, 0.00000030, -0.00000175, 0.00000667, -0.00001359, 0.00000381, 0.00000818, -0.00001281, 0.00000472, 0.00000913, -0.00000577, -0.00000483, -0.00000366, 0.00000865, 0.00000174, -0.00000169, -0.00000093, -0.00000591, 0.00000362, -0.00000220, 0.00000601, 0.00000974, -0.00001136, -0.00001151, 0.00000013, 0.00000779, -0.00000149, 0.00000068, 0.00000884, -0.00000772, -0.00000176, 0.00000672, 0.00000132, -0.00000612, 0.00000149, 0.00000313, -0.00000509, -0.00000415, -0.00000037, 0.00000268, 0.00000295, 0.00000433, -0.00000134, 0.00000271, -0.00000780, -0.00000250, 0.00001762, -0.00000152, -0.00000552, 0.00000198, -0.00000411, 0.00000694, -0.00000606, -0.00001244, },
		{ -0.00001395, -0.00000010, 0.00000294, 0.00000190, -0.00001181, -0.00000216, -0.00000518, -0.00001090, 0.00000837, -0.00001605, -0.00000050, -0.00000652, -0.00001258, 0.00001915, -0.00003690, 0.00000539, -0.00008881, -0.00006664, 0.00014278, 0.00016410, 0.00016314, -0.00007437, -0.00009774, 0.00000559, -0.00003218, -0.00000040, -0.00001370, -0.00000684, -0.00000413, -0.00000492, -0.00002935, 0.00000871, -0.00000720, 0.00001950, 0.00001647, -0.00000548, 0.00002987, -0.00000289, 0.00000150, 0.00001221, -0.00000787, -0.00002917, 0.00001694, -0.00001147, 0.00001532, -0.00001739, -0.00000813, 0.00001037, -0.00001670, 0.00002413, -0.00001922, 0.00001385, -0.00001632, 0.00001677, -0.00000716, -0.00001100, 0.00000298, -0.00001751, 0.00003061, -0.00000197, -0.00000221, -0.00001769, 0.00001438, -0.00000903, 0.00000503, 0.00000336, -0.00000377, 0.00000783, -0.00001321, 0.00001540, -0.00001398, 0.00000762, -0.00001132, -0.00000192, 0.00001580, -0.00000578, 0.00000083, -0.00000273, 0.00000726, -0.00001125, -0.00000075, 0.00001089, -0.00001209, 0.00000134, 0.00001136, -0.00000498, -0.00000462, -0.00000498, 0.00000713, 0.00000418, -0.00000258, 0.00000002, -0.00000632, 0.00000286, -0.00000128, 0.00000356, 0.00001164, -0.00000922, -0.00001263, -0.00000142, 0.00000716, 0.00000046, -0.00000130, 0.00000936, -0.00000553, -0.00000350, 0.00000652, 0.00000221, -0.00000615, 0.00000015, 0.00000422, -0.00000420, -0.00000493, -0.00000091, 0.00000293, 0.00000221, 0.00000495, -0.00000119, 0.00000252, -0.00000575, -0.00000600, 0.00001781, 0.00000167, -0.00000729, 0.00000289, -0.00000530, 0.00000628, -0.00000262, },
		{ 0.00000091, -0.00001467, -0.00000193, 0.00000292, 0.00000284, -0.00001087, -0.00000397, -0.00000346, -0.00001277, 0.00000893, -0.00001414, -0.00000435, -0.00000132, -0.00001855, 0.00002371, -0.00003649, 0.00000231, -0.00006782, -0.00009903, 0.00012392, 0.00017903, 0.00017931, -0.00005613, -0.00011290, 0.00000520, -0.00003166, -0.00000257, -0.00001213, -0.00000826, -0.00000423, -0.00000269, -0.00002971, 0.00000752, -0.00000659, 0.00001454, 0.00001993, -0.00000791, 0.00002766, 0.00000003, 0.00000048, 0.00000950, 0.00000220, -0.00003706, 0.00001767, -0.00000810, 0.00000978, -0.00000842, -0.00001772, 0.00001814, -0.00002251, 0.00002629, -0.00001767, 0.00001118, -0.00001189, 0.00000941, 0.00000102, -0.00001641, 0.00000813, -0.00002238, 0.00002868, 0.00000240, -0.00000193, -0.00001835, 0.00001205, -0.00000541, 0.00000075, 0.00000801, -0.00000830, 0.00001135, -0.00001510, 0.00001510, -0.00001148, 0.00000422, -0.00000639, -0.00000750, 0.00001796, -0.00000533, 0.00000061, -0.00000291, 0.00000679, -0.00000809, -0.00000499, 0.00001230, -0.00001010, -0.00000243, 0.00001302, -0.00000368, -0.00000459, -0.00000565, 0.00000493, 0.00000651, -0.00000298, 0.00000045, -0.00000613, 0.00000161, -0.00000007, 0.00000128, 0.00001268, -0.00000648, -0.00001361, -0.00000280, 0.00000609, 0.00000235, -0.00000265, 0.00000900, -0.00000301, -0.00000489, 0.00000596, 0.00000315, -0.00000599, -0.00000125, 0.00000494, -0.00000303, -0.00000550, -0.00000171, 0.00000326, 0.00000157, 0.00000522, -0.00000061, 0.00000183, -0.00000341, -0.00000889, 0.00001672, 0.00000554, -0.00000893, 0.00000343, -0.00000568, 0.00000463, },
		{ 0.00000225, 0.00000410, -0.00001442, -0.00000413, 0.00000284, 0.00000352, -0.00000959, -0.00000575, -0.00000205, -0.00001399, 0.00000836, -0.00001101, -0.00000896, 0.00000380, -0.00002328, 0.00002559, -0.00003226, -0.00000542, -0.00004459, -0.00012333, 0.00009479, 0.00019300, 0.00019748, -0.00003642, -0.00012630, 0.00000228, -0.00002982, -0.00000540, -0.00001047, -0.00000938, -0.00000492, -0.00000021, -0.00002997, 0.00000613, -0.00000543, 0.00000936, 0.00002310, -0.00000986, 0.00002564, 0.00000152, 0.00000167, 0.00000440, 0.00001241, -0.00004178, 0.00001388, -0.00000084, 0.00000155, 0.00000182, -0.00002692, 0.00002413, -0.00002547, 0.00002510, -0.00001316, 0.00000611, -0.00000557, 0.00000127, 0.00000854, -0.00002062, 0.00001183, -0.00002481, 0.00002441, 0.00000739, -0.00000167, -0.00001820, 0.00000863, -0.00000099, -0.00000396, 0.00001258, -0.00001199, 0.00001335, -0.00001515, 0.00001296, -0.00000756, -0.00000005, -0.00000134, -0.00001213, 0.00001863, -0.00000374, -0.00000032, -0.00000243, 0.00000554, -0.00000461, -0.00000840, 0.00001227, -0.00000707, -0.00000617, 0.00001382, -0.00000182, -0.00000479, -0.00000573, 0.00000233, 0.00000841, -0.00000271, 0.00000034, -0.00000542, 0.00000008, 0.00000114, -0.00000056, 0.00001276, -0.00000329, -0.00001432, -0.00000409, 0.00000476, 0.00000392, -0.00000327, 0.00000792, -0.00000043, -0.00000576, 0.00000505, 0.00000407, -0.00000564, -0.00000264, 0.00000531, -0.00000169, -0.00000577, -0.00000273, 0.00000355, 0.00000113, 0.00000509, 0.00000035, 0.00000082, -0.00000107, -0.00001086, 0.00001435, 0.00000975, -0.00001008, 0.00000331, -0.00000517, },
		{ -0.00000385, -0.00000047, 0.00000655, -0.00001312, -0.00000655, 0.00000259, 0.00000404, -0.00000817, -0.00000727, -0.00000114, -0.00001442, 0.00000670, -0.00000700, -0.00001368, 0.00000806, -0.00002603, 0.00002432, -0.00002451, -0.00001626, -0.00002220, -0.00013733, 0.00005695, 0.00020295, 0.00021883, -0.00001550, -0.00013766, -0.00000300, -0.00002717, -0.00000834, -0.00000917, -0.00000987, -0.00000631, 0.00000244, -0.00002995, 0.00000429, -0.00000350, 0.00000377, 0.00002624, -0.00001173, 0.00002424, 0.00000144, 0.00000463, -0.00000177, 0.00002086, -0.00004184, 0.00000523, 0.00000963, -0.00000807, 0.00001181, -0.00003425, 0.00002711, -0.00002485, 0.00002045, -0.00000607, -0.00000062, 0.00000168, -0.00000655, 0.00001450, -0.00002300, 0.00001364, -0.00002469, 0.00001821, 0.00001248, -0.00000127, -0.00001725, 0.00000430, 0.00000385, -0.00000860, 0.00001646, -0.00001425, 0.00001348, -0.00001332, 0.00000928, -0.00000272, -0.00000453, 0.00000316, -0.00001530, 0.00001765, -0.00000116, -0.00000174, -0.00000151, 0.00000387, -0.00000130, -0.00001065, 0.00001084, -0.00000341, -0.00000941, 0.00001358, 0.00000057, -0.00000513, -0.00000537, -0.00000036, 0.00000959, -0.00000172, -0.00000021, -0.00000440, -0.00000148, 0.00000211, -0.00000178, 0.00001195, 0.00000009, -0.00001459, -0.00000537, 0.00000337, 0.00000503, -0.00000318, 0.00000634, 0.00000193, -0.00000600, 0.00000385, 0.00000490, -0.00000508, -0.00000400, 0.00000532, -0.00000031, -0.00000569, -0.00000388, 0.00000368, 0.00000100, 0.00000457, 0.00000161, -0.00000031, 0.00000094, -0.00001173, 0.00001089, 0.00001386, -0.00001036, 0.00000238, },
		{ 0.00000064, -0.00000199, -0.00000308, 0.00000795, -0.00001086, -0.00000896, 0.00000203, 0.00000452, -0.00000678, -0.00000835, -0.00000083, -0.00001406, 0.00000418, -0.00000262, -0.00001783, 0.00001078, -0.00002634, 0.00001999, -0.00001412, -0.00002818, -0.00000346, -0.00014000, 0.00001322, 0.00020560, 0.00024396, 0.00000695, -0.00014709, -0.00001014, -0.00002438, -0.00001088, -0.00000860, -0.00000958, -0.00000828, 0.00000496, -0.00002925, 0.00000161, -0.00000051, -0.00000241, 0.00002954, -0.00001375, 0.00002357, 0.00000009, 0.00000843, -0.00000739, 0.00002579, -0.00003643, -0.00000779, 0.00002189, -0.00001741, 0.00001992, -0.00003840, 0.00002629, -0.00002045, 0.00001274, 0.00000276, -0.00000802, 0.00000875, -0.00001302, 0.00001820, -0.00002325, 0.00001345, -0.00002223, 0.00001069, 0.00001715, -0.00000061, -0.00001557, -0.00000067, 0.00000868, -0.00001264, 0.00001910, -0.00001470, 0.00001165, -0.00000987, 0.00000454, 0.00000237, -0.00000856, 0.00000653, -0.00001673, 0.00001511, 0.00000213, -0.00000336, -0.00000044, 0.00000213, 0.00000145, -0.00001157, 0.00000826, 0.00000040, -0.00001179, 0.00001224, 0.00000333, -0.00000547, -0.00000477, -0.00000281, 0.00000989, -0.00000008, -0.00000102, -0.00000327, -0.00000283, 0.00000264, -0.00000231, 0.00001040, 0.00000336, -0.00001426, -0.00000670, 0.00000202, 0.00000559, -0.00000250, 0.00000455, 0.00000382, -0.00000560, 0.00000248, 0.00000556, -0.00000429, -0.00000528, 0.00000502, 0.00000104, -0.00000528, -0.00000505, 0.00000353, 0.00000119, 0.00000370, 0.00000298, -0.00000131, 0.00000240, -0.00001148, 0.00000668, 0.00001741, -0.00000954, },
		{ -0.00000749, -0.00000173, 0.00000003, -0.00000513, 0.00000815, -0.00000788, -0.00001102, 0.00000100, 0.00000506, -0.00000559, -0.00000892, -0.00000109, -0.00001309, 0.00000117, 0.00000153, -0.00002080, 0.00001155, -0.00002416, 0.00001315, -0.00000248, -0.00003904, 0.00000949, -0.00013175, -0.00003254, 0.00019794, 0.00027254, 0.00003188, -0.00015508, -0.00001845, -0.00002213, -0.00001259, -0.00000889, -0.00000866, -0.00001046, 0.00000684, -0.00002741, -0.00000226, 0.00000372, -0.00000923, 0.00003299, -0.00001584, 0.00002333, -0.00000177, 0.00001181, -0.00001086, 0.00002598, -0.00002556, -0.00002384, 0.00003402, -0.00002459, 0.00002470, -0.00003845, 0.00002140, -0.00001265, 0.00000287, 0.00001220, -0.00001494, 0.00001457, -0.00001728, 0.00001924, -0.00002132, 0.00001139, -0.00001787, 0.00000258, 0.00002090, 0.00000041, -0.00001325, -0.00000592, 0.00001300, -0.00001555, 0.00002008, -0.00001316, 0.00000805, -0.00000526, -0.00000058, 0.00000703, -0.00001153, 0.00000838, -0.00001636, 0.00001130, 0.00000570, -0.00000481, 0.00000049, 0.00000066, 0.00000337, -0.00001121, 0.00000491, 0.00000388, -0.00001302, 0.00000987, 0.00000621, -0.00000563, -0.00000414, -0.00000480, 0.00000926, 0.00000202, -0.00000184, -0.00000228, -0.00000379, 0.00000263, -0.00000220, 0.00000839, 0.00000623, -0.00001323, -0.00000808, 0.00000077, 0.00000563, -0.00000143, 0.00000282, 0.00000510, -0.00000464, 0.00000109, 0.00000596, -0.00000329, -0.00000644, 0.00000443, 0.00000229, -0.00000456, -0.00000611, 0.00000304, 0.00000169, 0.00000262, 0.00000430, -0.00000198, 0.00000315, -0.00001020, 0.00000215, 0.00001997, },
		{ 0.00002121, -0.00000430, -0.00000439, 0.00000179, -0.00000631, 0.00000719, -0.00000455, -0.00001240, -0.00000063, 0.00000574, -0.00000465, -0.00000899, -0.00000177, -0.00001177, -0.00000188, 0.00000493, -0.00002217, 0.00001026, -0.00001987, 0.00000479, 0.00000885, -0.00004697, 0.00001548, -0.00011435, -0.00007591, 0.00017779, 0.00030311, 0.00006089, -0.00016231, -0.00002723, -0.00002091, -0.00001337, -0.00000985, -0.00000751, -0.00001226, 0.00000747, -0.00002396, -0.00000760, 0.00000921, -0.00001648, 0.00003627, -0.00001758, 0.00002290, -0.00000315, 0.00001345, -0.00001089, 0.00002094, -0.00001021, -0.00004087, 0.00004383, -0.00002793, 0.00002512, -0.00003408, 0.00001279, -0.00000237, -0.00000788, 0.00002093, -0.00002025, 0.00001823, -0.00001880, 0.00001756, -0.00001749, 0.00000786, -0.00001223, -0.00000537, 0.00002334, 0.00000181, -0.00001044, -0.00001104, 0.00001631, -0.00001693, 0.00001918, -0.00000973, 0.00000311, -0.00000018, -0.00000539, 0.00001066, -0.00001304, 0.00000859, -0.00001439, 0.00000670, 0.00000905, -0.00000577, 0.00000103, -0.00000035, 0.00000437, -0.00000976, 0.00000128, 0.00000659, -0.00001299, 0.00000670, 0.00000890, -0.00000542, -0.00000366, -0.00000617, 0.00000779, 0.00000430, -0.00000241, -0.00000161, -0.00000426, 0.00000212, -0.00000163, 0.00000623, 0.00000843, -0.00001150, -0.00000946, -0.00000037, 0.00000525, -0.00000020, 0.00000138, 0.00000570, -0.00000327, -0.00000016, 0.00000603, -0.00000210, -0.00000744, 0.00000357, 0.00000337, -0.00000360, -0.00000693, 0.00000217, 0.00000242, 0.00000148, 0.00000537, -0.00000217, 0.00000319, -0.00000815, -0.00000225, },
		{ -0.00000608, 0.00002099, -0.00000021, -0.00000691, 0.00000290, -0.00000644, 0.00000532, -0.00000137, -0.00001279, -0.00000288, 0.00000650, -0.00000390, -0.00000873, -0.00000259, -0.00001046, -0.00000454, 0.00000717, -0.00002179, 0.00000717, -0.00001418, -0.00000389, 0.00001836, -0.00005064, 0.00001448, -0.00009069, -0.00011251, 0.00014425, 0.00033293, 0.00009593, -0.00016936, -0.00003607, -0.00002088, -0.00001348, -0.00001096, -0.00000679, -0.00001298, 0.00000629, -0.00001862, -0.00001439, 0.00001565, -0.00002366, 0.00003879, -0.00001832, 0.00002141, -0.00000296, 0.00001227, -0.00000680, 0.00001111, 0.00000782, -0.00005642, 0.00004922, -0.00002622, 0.00002083, -0.00002566, 0.00000141, 0.00000904, -0.00001806, 0.00002770, -0.00002304, 0.00001919, -0.00001743, 0.00001349, -0.00001230, 0.00000344, -0.00000604, -0.00001246, 0.00002421, 0.00000354, -0.00000732, -0.00001559, 0.00001820, -0.00001653, 0.00001643, -0.00000476, -0.00000250, 0.00000465, -0.00000922, 0.00001280, -0.00001292, 0.00000725, -0.00001125, 0.00000189, 0.00001175, -0.00000600, 0.00000106, -0.00000079, 0.00000451, -0.00000757, -0.00000216, 0.00000824, -0.00001177, 0.00000308, 0.00001110, -0.00000469, -0.00000342, -0.00000688, 0.00000567, 0.00000646, -0.00000251, -0.00000137, -0.00000423, 0.00000121, -0.00000082, 0.00000420, 0.00000980, -0.00000916, -0.00001070, -0.00000147, 0.00000458, 0.00000098, 0.00000037, 0.00000570, -0.00000169, -0.00000110, 0.00000575, -0.00000079, -0.00000821, 0.00000247, 0.00000427, -0.00000249, -0.00000744, 0.00000097, 0.00000324, 0.00000044, 0.00000605, -0.00000184, 0.00000260, -0.00000563, },
		{ -0.00000301, -0.00000900, 0.00001932, 0.00000436, -0.00000885, 0.00000310, -0.00000557, 0.00000293, 0.00000119, -0.00001201, -0.00000565, 0.00000720, -0.00000321, -0.00000837, -0.00000328, -0.00000946, -0.00000650, 0.00000810, -0.00001984, 0.00000286, -0.00000804, -0.00001167, 0.00002487, -0.00004954, 0.00000757, -0.00006441, -0.00013856, 0.00009809, 0.00035817, 0.00013900, -0.00017637, -0.00004498, -0.00002171, -0.00001359, -0.00001144, -0.00000730, -0.00001196, 0.00000291, -0.00001138, -0.00002228, 0.00002239, -0.00002997, 0.00003973, -0.00001722, 0.00001803, -0.00000034, 0.00000766, 0.00000127, -0.00000219, 0.00002607, -0.00006793, 0.00004855, -0.00001895, 0.00001220, -0.00001420, -0.00001129, 0.00001999, -0.00002622, 0.00003148, -0.00002277, 0.00001729, -0.00001345, 0.00000766, -0.00000648, -0.00000122, 0.00000002, -0.00001812, 0.00002341, 0.00000551, -0.00000407, -0.00001918, 0.00001839, -0.00001432, 0.00001208, 0.00000115, -0.00000804, 0.00000852, -0.00001159, 0.00001323, -0.00001124, 0.00000472, -0.00000747, -0.00000257, 0.00001342, -0.00000539, 0.00000059, -0.00000073, 0.00000397, -0.00000505, -0.00000500, 0.00000868, -0.00000959, -0.00000059, 0.00001253, -0.00000339, -0.00000346, -0.00000700, 0.00000317, 0.00000818, -0.00000200, -0.00000154, -0.00000382, 0.00000009, -0.00000002, 0.00000256, 0.00001030, -0.00000638, -0.00001164, -0.00000260, 0.00000375, 0.00000193, -0.00000015, 0.00000520, -0.00000011, -0.00000162, 0.00000517, 0.00000057, -0.00000871, 0.00000116, 0.00000496, -0.00000129, -0.00000758, -0.00000050, 0.00000402, -0.00000035, 0.00000626, -0.00000100, 0.00000155, },
		{ 0.00000027, -0.00000060, -0.00001081, 0.00001640, 0.00000894, -0.00000983, 0.00000228, -0.00000395, 0.00000055, 0.00000271, -0.00001006, -0.00000868, 0.00000760, -0.00000235, -0.00000817, -0.00000356, -0.00000899, -0.00000764, 0.00000778, -0.00001675, -0.00000191, -0.00000242, -0.00001756, 0.00002777, -0.00004398, -0.00000328, -0.00003928, -0.00015155, 0.00004188, 0.00037424, 0.00019161, -0.00018273, -0.00005462, -0.00002256, -0.00001470, -0.00001038, -0.00000974, -0.00000878, -0.00000269, -0.00000267, -0.00003052, 0.00002848, -0.00003441, 0.00003819, -0.00001355, 0.00001211, 0.00000514, -0.00000029, 0.00001238, -0.00001689, 0.00004186, -0.00007325, 0.00004097, -0.00000651, 0.00000039, -0.00000123, -0.00002364, 0.00002892, -0.00003118, 0.00003163, -0.00001932, 0.00001280, -0.00000746, 0.00000092, -0.00000079, -0.00000549, 0.00000533, -0.00002197, 0.00002100, 0.00000761, -0.00000091, -0.00002148, 0.00001677, -0.00001051, 0.00000665, 0.00000730, -0.00001276, 0.00001087, -0.00001224, 0.00001200, -0.00000831, 0.00000150, -0.00000366, -0.00000618, 0.00001387, -0.00000398, -0.00000029, -0.00000028, 0.00000302, -0.00000256, -0.00000695, 0.00000797, -0.00000679, -0.00000389, 0.00001301, -0.00000154, -0.00000368, -0.00000667, 0.00000060, 0.00000922, -0.00000086, -0.00000203, -0.00000318, -0.00000102, 0.00000056, 0.00000144, 0.00000999, -0.00000342, -0.00001211, -0.00000381, 0.00000287, 0.00000257, -0.00000022, 0.00000441, 0.00000126, -0.00000169, 0.00000434, 0.00000186, -0.00000886, -0.00000033, 0.00000543, -0.00000008, -0.00000734, -0.00000211, 0.00000460, -0.00000080, 0.00000598, 0.00000025, },
		{ 0.00000174, -0.00000098, 0.00000132, -0.00001143, 0.00001257, 0.00001303, -0.00000959, 0.00000055, -0.00000198, -0.00000131, 0.00000292, -0.00000712, -0.00001158, 0.00000739, -0.00000109, -0.00000836, -0.00000328, -0.00000908, -0.00000803, 0.00000655, -0.00001314, -0.00000632, 0.00000188, -0.00002098, 0.00002709, -0.00003508, -0.00001559, -0.00001871, -0.00015058, -0.00002014, 0.00037637, 0.00025423, -0.00018683, -0.00006631, -0.00002224, -0.00001793, -0.00000697, -0.00001454, -0.00000341, -0.00001004, 0.00000665, -0.00003803, 0.00003275, -0.00003594, 0.00003341, -0.00000681, 0.00000345, 0.00001333, -0.00001079, 0.00002488, -0.00003052, 0.00005265, -0.00007093, 0.00002660, 0.00000982, -0.00001287, 0.00001139, -0.00003391, 0.00003449, -0.00003217, 0.00002803, -0.00001305, 0.00000638, -0.00000035, -0.00000579, 0.00000400, -0.00000885, 0.00000944, -0.00002381, 0.00001719, 0.00000967, 0.00000197, -0.00002229, 0.00001344, -0.00000552, 0.00000076, 0.00001292, -0.00001599, 0.00001136, -0.00001116, 0.00000942, -0.00000462, -0.00000185, -0.00000035, -0.00000861, 0.00001309, -0.00000196, -0.00000135, 0.00000033, 0.00000193, -0.00000045, -0.00000787, 0.00000630, -0.00000379, -0.00000650, 0.00001247, 0.00000071, -0.00000397, -0.00000611, -0.00000175, 0.00000945, 0.00000082, -0.00000265, -0.00000253, -0.00000191, 0.00000076, 0.00000089, 0.00000906, -0.00000058, -0.00001199, -0.00000513, 0.00000201, 0.00000288, 0.00000004, 0.00000354, 0.00000228, -0.00000133, 0.00000339, 0.00000299, -0.00000866, -0.00000194, 0.00000566, 0.00000109, -0.00000676, -0.00000372, 0.00000485, -0.00000086, 0.00000529, },
		{ 0.00000428, 0.00000328, -0.00000201, 0.00000260, -0.00001099, 0.00000827, 0.00001618, -0.00000806, -0.00000178, -0.00000015, -0.00000221, 0.00000180, -0.00000361, -0.00001386, 0.00000627, 0.00000074, -0.00000905, -0.00000246, -0.00000961, -0.00000795, 0.00000490, -0.00000970, -0.00000969, 0.00000436, -0.00002187, 0.00002349, -0.00002447, -0.00002678, -0.00000515, -0.00013658, -0.00008244, 0.00036031, 0.00032586, -0.00018590, -0.00008186, -0.00001939, -0.00002425, -0.00000077, -0.00002165, 0.00000362, -0.00001822, 0.00001535, -0.00004351, 0.00003403, -0.00003370, 0.00002495, 0.00000302, -0.00000760, 0.00002339, -0.00002236, 0.00003661, -0.00004058, 0.00005649, -0.00006054, 0.00000663, 0.00002804, -0.00002546, 0.00002179, -0.00004066, 0.00003583, -0.00002899, 0.00002106, -0.00000473, -0.00000105, 0.00000688, -0.00001156, 0.00000731, -0.00001093, 0.00001211, -0.00002368, 0.00001230, 0.00001154, 0.00000441, -0.00002154, 0.00000871, 0.00000009, -0.00000486, 0.00001734, -0.00001732, 0.00000993, -0.00000864, 0.00000595, -0.00000076, -0.00000478, 0.00000205, -0.00000971, 0.00001125, 0.00000036, -0.00000236, 0.00000092, 0.00000095, 0.00000108, -0.00000781, 0.00000402, -0.00000100, -0.00000816, 0.00001100, 0.00000313, -0.00000413, -0.00000550, -0.00000367, 0.00000882, 0.00000285, -0.00000317, -0.00000205, -0.00000246, 0.00000054, 0.00000082, 0.00000775, 0.00000188, -0.00001120, -0.00000651, 0.00000116, 0.00000289, 0.00000048, 0.00000275, 0.00000287, -0.00000064, 0.00000247, 0.00000387, -0.00000809, -0.00000360, 0.00000563, 0.00000219, -0.00000588, -0.00000519, 0.00000471, -0.00000056, },
		{ 0.00000003, 0.00000311, 0.00000468, -0.00000264, 0.00000320, -0.00000967, 0.00000395, 0.00001810, -0.00000539, -0.00000428, 0.00000104, -0.00000189, -0.00000043, -0.00000008, -0.00001505, 0.00000406, 0.00000317, -0.00001018, -0.00000126, -0.00001027, -0.00000779, 0.00000334, -0.00000701, -0.00001160, 0.00000495, -0.00002063, 0.00001811, -0.00001399, -0.00003468, 0.00000025, -0.00011227, -0.00013887, 0.00032310, 0.00040363, -0.00017617, -0.00010332, -0.00001286, -0.00003420, 0.00000808, -0.00003040, 0.00001126, -0.00002583, 0.00002195, -0.00004566, 0.00003138, -0.00002723, 0.00001288, 0.00001539, -0.00002008, 0.00003394, -0.00003311, 0.00004532, -0.00004499, 0.00005236, -0.00004282, -0.00001684, 0.00004564, -0.00003523, 0.00002841, -0.00004292, 0.00003265, -0.00002201, 0.00001160, 0.00000458, -0.00000844, 0.00001325, -0.00001564, 0.00000876, -0.00001161, 0.00001327, -0.00002178, 0.00000675, 0.00001305, 0.00000630, -0.00001934, 0.00000304, 0.00000563, -0.00000954, 0.00002004, -0.00001655, 0.00000682, -0.00000517, 0.00000221, 0.00000272, -0.00000685, 0.00000331, -0.00000955, 0.00000867, 0.00000264, -0.00000308, 0.00000131, 0.00000024, 0.00000195, -0.00000692, 0.00000149, 0.00000121, -0.00000878, 0.00000879, 0.00000544, -0.00000400, -0.00000500, -0.00000503, 0.00000745, 0.00000496, -0.00000336, -0.00000189, -0.00000260, -0.00000006, 0.00000108, 0.00000634, 0.00000374, -0.00000977, -0.00000785, 0.00000029, 0.00000269, 0.00000094, 0.00000218, 0.00000305, 0.00000026, 0.00000170, 0.00000445, -0.00000719, -0.00000522, 0.00000531, 0.00000320, -0.00000479, -0.00000641, 0.00000414, },
		{ 0.00000318, 0.00000078, 0.00000195, 0.00000580, -0.00000280, 0.00000315, -0.00000779, 0.00000003, 0.00001864, -0.00000188, -0.00000645, 0.00000122, -0.00000036, -0.00000331, 0.00000281, -0.00001475, 0.00000070, 0.00000608, -0.00001153, -0.00000004, -0.00001065, -0.00000799, 0.00000234, -0.00000543, -0.00001197, 0.00000401, -0.00001812, 0.00001231, -0.00000534, -0.00003792, -0.00000218, -0.00008177, -0.00018340, 0.00026374, 0.00048268, -0.00015321, -0.00013246, -0.00000207, -0.00004756, 0.00001871, -0.00003949, 0.00001795, -0.00003129, 0.00002501, -0.00004345, 0.00002429, -0.00001664, -0.00000208, 0.00002913, -0.00003249, 0.00004320, -0.00004101, 0.00004905, -0.00004241, 0.00004041, -0.00001963, -0.00004098, 0.00005999, -0.00004036, 0.00003020, -0.00004038, 0.00002532, -0.00001212, 0.00000085, 0.00001366, -0.00001474, 0.00001795, -0.00001753, 0.00000826, -0.00001097, 0.00001307, -0.00001844, 0.00000097, 0.00001409, 0.00000758, -0.00001591, -0.00000297, 0.00001044, -0.00001273, 0.00002074, -0.00001382, 0.00000249, -0.00000137, -0.00000123, 0.00000534, -0.00000781, 0.00000342, -0.00000839, 0.00000576, 0.00000453, -0.00000335, 0.00000141, -0.00000011, 0.00000220, -0.00000548, -0.00000089, 0.00000262, -0.00000843, 0.00000614, 0.00000739, -0.00000345, -0.00000470, -0.00000580, 0.00000554, 0.00000686, -0.00000306, -0.00000209, -0.00000237, -0.00000093, 0.00000146, 0.00000508, 0.00000488, -0.00000781, -0.00000900, -0.00000067, 0.00000237, 0.00000131, 0.00000190, 0.00000289, 0.00000118, 0.00000118, 0.00000471, -0.00000604, -0.00000671, 0.00000469, 0.00000410, -0.00000356, -0.00000727, },
		{ -0.00000775, 0.00000191, 0.00000155, 0.00000092, 0.00000655, -0.00000248, 0.00000261, -0.00000564, -0.00000318, 0.00001787, 0.00000200, -0.00000780, 0.00000019, 0.00000216, -0.00000622, 0.00000446, -0.00001276, -0.00000362, 0.00000913, -0.00001273, 0.00000079, -0.00001036, -0.00000888, 0.00000215, -0.00000501, -0.00001111, 0.00000222, -0.00001537, 0.00000740, 0.00000026, -0.00003611, -0.00001073, -0.00005006, -0.00021104, 0.00018370, 0.00055643, -0.00011247, -0.00017026, 0.00001255, -0.00006321, 0.00002944, -0.00004709, 0.00002190, -0.00003298, 0.00002338, -0.00003636, 0.00001291, -0.00000273, -0.00001862, 0.00004258, -0.00004301, 0.00004929, -0.00004427, 0.00004647, -0.00003259, 0.00002197, 0.00000627, -0.00006270, 0.00006874, -0.00003961, 0.00002684, -0.00003340, 0.00001477, -0.00000064, -0.00000981, 0.00002136, -0.00001911, 0.00002042, -0.00001707, 0.00000598, -0.00000928, 0.00001178, -0.00001410, -0.00000461, 0.00001453, 0.00000828, -0.00001164, -0.00000870, 0.00001394, -0.00001410, 0.00001945, -0.00000950, -0.00000241, 0.00000211, -0.00000384, 0.00000683, -0.00000761, 0.00000256, -0.00000660, 0.00000296, 0.00000578, -0.00000309, 0.00000122, -0.00000014, 0.00000196, -0.00000379, -0.00000281, 0.00000312, -0.00000730, 0.00000340, 0.00000872, -0.00000243, -0.00000463, -0.00000607, 0.00000333, 0.00000828, -0.00000216, -0.00000262, -0.00000188, -0.00000187, 0.00000174, 0.00000417, 0.00000529, -0.00000552, -0.00000979, -0.00000178, 0.00000202, 0.00000150, 0.00000187, 0.00000254, 0.00000197, 0.00000097, 0.00000471, -0.00000474, -0.00000795, 0.00000377, 0.00000487, -0.00000226, },
		{ -0.00000095, -0.00000783, 0.00000045, 0.00000219, 0.00000013, 0.00000689, -0.00000175, 0.00000176, -0.00000353, -0.00000550, 0.00001600, 0.00000575, -0.00000798, -0.00000200, 0.00000519, -0.00000846, 0.00000438, -0.00000913, -0.00000849, 0.00001186, -0.00001334, 0.00000081, -0.00000908, -0.00001062, 0.00000278, -0.00000551, -0.00000962, 0.00000047, -0.00001340, 0.00000433, 0.00000234, -0.00003000, -0.00002261, -0.00002218, -0.00021858, 0.00008708, 0.00061698, -0.00005009, -0.00021632, 0.00002935, -0.00007903, 0.00003799, -0.00005107, 0.00002136, -0.00002968, 0.00001650, -0.00002459, -0.00000188, 0.00001305, -0.00003489, 0.00005380, -0.00004981, 0.00005058, -0.00004167, 0.00003717, -0.00001640, -0.00000060, 0.00003160, -0.00007904, 0.00007022, -0.00003260, 0.00001881, -0.00002301, 0.00000238, 0.00001094, -0.00001905, 0.00002670, -0.00002099, 0.00002044, -0.00001438, 0.00000232, -0.00000694, 0.00000973, -0.00000923, -0.00000959, 0.00001431, 0.00000848, -0.00000695, -0.00001354, 0.00001573, -0.00001355, 0.00001644, -0.00000420, -0.00000717, 0.00000468, -0.00000530, 0.00000711, -0.00000642, 0.00000107, -0.00000461, 0.00000062, 0.00000624, -0.00000236, 0.00000083, 0.00000007, 0.00000143, -0.00000216, -0.00000406, 0.00000276, -0.00000567, 0.00000092, 0.00000929, -0.00000098, -0.00000471, -0.00000597, 0.00000112, 0.00000901, -0.00000068, -0.00000336, -0.00000130, -0.00000270, 0.00000176, 0.00000370, 0.00000506, -0.00000316, -0.00001007, -0.00000305, 0.00000166, 0.00000149, 0.00000203, 0.00000214, 0.00000253, 0.00000106, 0.00000452, -0.00000339, -0.00000888, 0.00000256, 0.00000548, },
		{ 0.00000588, 0.00000034, -0.00000755, -0.00000107, 0.00000259, -0.00000037, 0.00000685, -0.00000074, 0.00000079, -0.00000167, -0.00000689, 0.00001338, 0.00000888, -0.00000682, -0.00000507, 0.00000810, -0.00000939, 0.00000237, -0.00000419, -0.00001330, 0.00001373, -0.00001293, -0.00000027, -0.00000671, -0.00001312, 0.00000395, -0.00000639, -0.00000829, -0.00000038, -0.00001294, 0.00000353, 0.00000125, -0.00002121, -0.00003446, -0.00000250, -0.00020516, -0.00001951, 0.00065604, 0.00003629, -0.00026841, 0.00004532, -0.00009207, 0.00004171, -0.00004940, 0.00001499, -0.00002080, 0.00000459, -0.00000915, -0.00001848, 0.00002870, -0.00004879, 0.00006084, -0.00005132, 0.00004603, -0.00003284, 0.00002182, 0.00000423, -0.00002420, 0.00005301, -0.00008765, 0.00006368, -0.00001990, 0.00000732, -0.00001071, -0.00001023, 0.00002114, -0.00002574, 0.00002907, -0.00002018, 0.00001813, -0.00000988, -0.00000215, -0.00000440, 0.00000732, -0.00000425, -0.00001366, 0.00001340, 0.00000829, -0.00000233, -0.00001702, 0.00001562, -0.00001128, 0.00001220, 0.00000139, -0.00001111, 0.00000593, -0.00000549, 0.00000630, -0.00000455, -0.00000066, -0.00000284, -0.00000097, 0.00000591, -0.00000130, 0.00000038, 0.00000039, 0.00000080, -0.00000082, -0.00000456, 0.00000174, -0.00000390, -0.00000104, 0.00000905, 0.00000079, -0.00000481, -0.00000567, -0.00000085, 0.00000896, 0.00000125, -0.00000409, -0.00000081, -0.00000325, 0.00000142, 0.00000367, 0.00000439, -0.00000100, -0.00000974, -0.00000446, 0.00000129, 0.00000133, 0.00000226, 0.00000183, 0.00000279, 0.00000138, 0.00000425, -0.00000211, -0.00000945, 0.00000114, },
		{ -0.00000044, 0.00000605, 0.00000160, -0.00000696, -0.00000251, 0.00000268, -0.00000060, 0.00000653, 0.00000041, -0.00000011, -0.00000020, -0.00000743, 0.00001042, 0.00001103, -0.00000436, -0.00000853, 0.00001022, -0.00000853, -0.00000148, 0.00000142, -0.00001735, 0.00001422, -0.00001119, -0.00000258, -0.00000337, -0.00001604, 0.00000515, -0.00000698, -0.00000786, 0.00000030, -0.00001430, 0.00000475, -0.00000190, -0.00001202, -0.00004296, 0.00000598, -0.00017256, -0.00012749, 0.00066586, 0.00014674, -0.00032221, 0.00005628, -0.00009895, 0.00003806, -0.00004053, 0.00000223, -0.00000665, -0.00001119, 0.00000817, -0.00003474, 0.00004198, -0.00005825, 0.00006212, -0.00004658, 0.00003541, -0.00001841, 0.00000209, 0.00002653, -0.00004549, 0.00006759, -0.00008716, 0.00004949, -0.00000293, -0.00000589, 0.00000175, -0.00002148, 0.00002871, -0.00002915, 0.00002826, -0.00001686, 0.00001391, -0.00000418, -0.00000679, -0.00000213, 0.00000488, 0.00000043, -0.00001660, 0.00001182, 0.00000787, 0.00000181, -0.00001881, 0.00001369, -0.00000770, 0.00000738, 0.00000655, -0.00001369, 0.00000568, -0.00000455, 0.00000474, -0.00000240, -0.00000221, -0.00000158, -0.00000173, 0.00000492, -0.00000014, 0.00000003, 0.00000067, 0.00000026, 0.00000009, -0.00000436, 0.00000032, -0.00000231, -0.00000229, 0.00000809, 0.00000266, -0.00000478, -0.00000537, -0.00000240, 0.00000813, 0.00000343, -0.00000461, -0.00000059, -0.00000344, 0.00000070, 0.00000397, 0.00000351, 0.00000073, -0.00000877, -0.00000592, 0.00000086, 0.00000108, 0.00000245, 0.00000170, 0.00000276, 0.00000181, 0.00000402, -0.00000095, -0.00000964, },
		{ -0.00000949, -0.00000205, 0.00000592, 0.00000279, -0.00000613, -0.00000377, 0.00000243, -0.00000062, 0.00000603, 0.00000158, -0.00000085, 0.00000084, -0.00000732, 0.00000751, 0.00001197, -0.00000090, -0.00001175, 0.00001093, -0.00000571, -0.00000672, 0.00000689, -0.00001991, 0.00001296, -0.00000804, -0.00000596, 0.00000053, -0.00001881, 0.00000575, -0.00000661, -0.00000890, 0.00000273, -0.00001719, 0.00000713, -0.00000548, -0.00000483, -0.00004545, 0.00000215, -0.00012509, -0.00022722, 0.00064041, 0.00027837, -0.00037138, 0.00005726, -0.00009632, 0.00002509, -0.00002385, -0.00001647, 0.00001142, -0.00002886, 0.00002498, -0.00004822, 0.00005070, -0.00006161, 0.00005671, -0.00003550, 0.00001942, 0.00000006, -0.00001955, 0.00004738, -0.00006132, 0.00007331, -0.00007741, 0.00002911, 0.00001617, -0.00001877, 0.00001265, -0.00003001, 0.00003280, -0.00002899, 0.00002448, -0.00001154, 0.00000842, 0.00000198, -0.00001097, -0.00000050, 0.00000268, 0.00000453, -0.00001827, 0.00000961, 0.00000735, 0.00000513, -0.00001881, 0.00001024, -0.00000338, 0.00000267, 0.00001068, -0.00001462, 0.00000400, -0.00000279, 0.00000283, -0.00000040, -0.00000326, -0.00000103, -0.00000171, 0.00000353, 0.00000087, -0.00000008, 0.00000084, -0.00000010, 0.00000052, -0.00000362, -0.00000119, -0.00000116, -0.00000277, 0.00000659, 0.00000442, -0.00000448, -0.00000518, -0.00000343, 0.00000665, 0.00000558, -0.00000471, -0.00000073, -0.00000324, -0.00000032, 0.00000443, 0.00000267, 0.00000187, -0.00000724, -0.00000727, 0.00000027, 0.00000083, 0.00000249, 0.00000179, 0.00000254, 0.00000223, 0.00000390, 0.00000003, },
		{ 0.00000085, -0.00000907, -0.00000359, 0.00000548, 0.00000389, -0.00000511, -0.00000477, 0.00000188, -0.00000054, 0.00000545, 0.00000266, -0.00000134, 0.00000149, -0.00000680, 0.00000498, 0.00001168, 0.00000310, -0.00001409, 0.00000984, -0.00000105, -0.00001264, 0.00001132, -0.00002039, 0.00000981, -0.00000366, -0.00000999, 0.00000441, -0.00002077, 0.00000515, -0.00000480, -0.00001163, 0.00000665, -0.00002083, 0.00000936, -0.00000765, -0.00000171, -0.00004046, -0.00001301, -0.00006914, -0.00030910, 0.00057639, 0.00042510, -0.00040791, 0.00004323, -0.00008159, 0.00000199, 0.00000001, -0.00003946, 0.00003112, -0.00004574, 0.00003862, -0.00005663, 0.00005312, -0.00005790, 0.00004455, -0.00001892, -0.00000030, 0.00002023, -0.00004019, 0.00006370, -0.00006929, 0.00006934, -0.00005948, 0.00000489, 0.00003488, -0.00002932, 0.00002055, -0.00003492, 0.00003307, -0.00002548, 0.00001832, -0.00000496, 0.00000243, 0.00000790, -0.00001415, 0.00000024, 0.00000089, 0.00000783, -0.00001865, 0.00000688, 0.00000687, 0.00000739, -0.00001713, 0.00000577, 0.00000101, -0.00000131, 0.00001335, -0.00001383, 0.00000121, -0.00000067, 0.00000104, 0.00000110, -0.00000361, -0.00000117, -0.00000110, 0.00000206, 0.00000149, 0.00000014, 0.00000085, -0.00000023, 0.00000054, -0.00000255, -0.00000251, -0.00000063, -0.00000256, 0.00000481, 0.00000582, -0.00000381, -0.00000519, -0.00000394, 0.00000473, 0.00000743, -0.00000425, -0.00000123, -0.00000272, -0.00000150, 0.00000486, 0.00000208, 0.00000236, -0.00000529, -0.00000835, -0.00000053, 0.00000064, 0.00000235, 0.00000205, 0.00000224, 0.00000250, 0.00000394, },
		{ 0.00000413, 0.00000156, -0.00000846, -0.00000495, 0.00000471, 0.00000487, -0.00000395, -0.00000547, 0.00000109, -0.00000044, 0.00000488, 0.00000359, -0.00000161, 0.00000185, -0.00000609, 0.00000305, 0.00001036, 0.00000706, -0.00001499, 0.00000681, 0.00000495, -0.00001831, 0.00001385, -0.00001848, 0.00000495, 0.00000149, -0.00001405, 0.00000755, -0.00002130, 0.00000291, -0.00000138, -0.00001581, 0.00001133, -0.00002404, 0.00000995, -0.00000681, -0.00000394, -0.00002798, -0.00003652, -0.00001234, -0.00036479, 0.00047409, 0.00057768, -0.00042290, 0.00000990, -0.00005355, -0.00003053, 0.00002901, -0.00006397, 0.00004943, -0.00005890, 0.00004655, -0.00005814, 0.00004829, -0.00004713, 0.00002662, 0.00000141, -0.00002140, 0.00003934, -0.00005689, 0.00007298, -0.00006806, 0.00005619, -0.00003562, -0.00002022, 0.00005069, -0.00003589, 0.00002454, -0.00003586, 0.00002971, -0.00001922, 0.00001065, 0.00000203, -0.00000332, 0.00001292, -0.00001596, -0.00000001, -0.00000040, 0.00001024, -0.00001780, 0.00000376, 0.00000649, 0.00000854, -0.00001409, 0.00000090, 0.00000484, -0.00000409, 0.00001438, -0.00001153, -0.00000223, 0.00000129, -0.00000025, 0.00000188, -0.00000323, -0.00000187, -0.00000019, 0.00000080, 0.00000157, 0.00000069, 0.00000074, -0.00000017, 0.00000025, -0.00000141, -0.00000341, -0.00000073, -0.00000184, 0.00000302, 0.00000670, -0.00000274, -0.00000538, -0.00000402, 0.00000263, 0.00000873, -0.00000316, -0.00000204, -0.00000201, -0.00000266, 0.00000509, 0.00000186, 0.00000222, -0.00000317, -0.00000898, -0.00000159, 0.00000052, 0.00000204, 0.00000240, 0.00000200, 0.00000254, },
		{ 0.00000233, 0.00000440, 0.00000225, -0.00000777, -0.00000606, 0.00000365, 0.00000569, -0.00000270, -0.00000586, 0.00000016, -0.00000044, 0.00000438, 0.00000436, -0.00000169, 0.00000205, -0.00000538, 0.00000176, 0.00000832, 0.00001040, -0.00001410, 0.00000205, 0.00001153, -0.00002274, 0.00001388, -0.00001419, -0.00000112, 0.00000673, -0.00001742, 0.00000929, -0.00001995, -0.00000108, 0.00000342, -0.00002078, 0.00001565, -0.00002546, 0.00000759, -0.00000192, -0.00001170, -0.00000959, -0.00006387, 0.00003744, -0.00038835, 0.00033779, 0.00072433, -0.00040765, -0.00004533, -0.00001296, -0.00006997, 0.00005975, -0.00008643, 0.00006299, -0.00006553, 0.00004682, -0.00005181, 0.00003629, -0.00003035, 0.00000479, 0.00002300, -0.00004112, 0.00005461, -0.00006721, 0.00007364, -0.00005760, 0.00003565, -0.00000887, -0.00004313, 0.00006143, -0.00003744, 0.00002431, -0.00003303, 0.00002338, -0.00001116, 0.00000246, 0.00000856, -0.00000816, 0.00001659, -0.00001621, -0.00000123, -0.00000124, 0.00001173, -0.00001587, 0.00000042, 0.00000623, 0.00000863, -0.00001014, -0.00000373, 0.00000758, -0.00000541, 0.00001385, -0.00000812, -0.00000574, 0.00000262, -0.00000079, 0.00000188, -0.00000225, -0.00000288, 0.00000070, 0.00000003, 0.00000106, 0.00000147, 0.00000059, 0.00000000, -0.00000018, -0.00000042, -0.00000376, -0.00000141, -0.00000086, 0.00000150, 0.00000694, -0.00000133, -0.00000568, -0.00000382, 0.00000059, 0.00000931, -0.00000150, -0.00000299, -0.00000129, -0.00000363, 0.00000497, 0.00000207, 0.00000161, -0.00000111, -0.00000903, -0.00000289, 0.00000044, 0.00000162, 0.00000273, 0.00000193, },
		{ 0.00000212, 0.00000190, 0.00000462, 0.00000300, -0.00000707, -0.00000686, 0.00000236, 0.00000626, -0.00000137, -0.00000597, -0.00000081, -0.00000060, 0.00000395, 0.00000502, -0.00000165, 0.00000218, -0.00000479, 0.00000106, 0.00000600, 0.00001259, -0.00001130, -0.00000396, 0.00001776, -0.00002506, 0.00001110, -0.00000795, -0.00000759, 0.00001123, -0.00001939, 0.00000914, -0.00001659, -0.00000657, 0.00000897, -0.00002552, 0.00001837, -0.00002388, 0.00000143, 0.00000718, -0.00002398, 0.00001183, -0.00008971, 0.00007316, -0.00037715, 0.00017576, 0.00085169, -0.00035481, -0.00012252, 0.00003711, -0.00011215, 0.00008782, -0.00010287, 0.00006866, -0.00006350, 0.00003845, -0.00003782, 0.00001829, -0.00000957, -0.00001833, 0.00004302, -0.00005674, 0.00006371, -0.00006954, 0.00006535, -0.00003922, 0.00001055, 0.00001732, -0.00006102, 0.00006566, -0.00003370, 0.00002024, -0.00002718, 0.00001504, -0.00000238, -0.00000529, 0.00001393, -0.00001162, 0.00001863, -0.00001490, -0.00000322, -0.00000170, 0.00001235, -0.00001305, -0.00000293, 0.00000606, 0.00000786, -0.00000581, -0.00000753, 0.00000891, -0.00000527, 0.00001206, -0.00000417, -0.00000875, 0.00000299, -0.00000048, 0.00000122, -0.00000093, -0.00000387, 0.00000126, -0.00000010, 0.00000001, 0.00000232, 0.00000053, 0.00000019, -0.00000060, 0.00000027, -0.00000354, -0.00000247, 0.00000010, 0.00000044, 0.00000655, 0.00000032, -0.00000597, -0.00000351, -0.00000114, 0.00000911, 0.00000057, -0.00000388, -0.00000073, -0.00000427, 0.00000444, 0.00000266, 0.00000071, 0.00000062, -0.00000844, -0.00000435, 0.00000033, 0.00000119, 0.00000290, },
		{ 0.00000284, 0.00000257, 0.00000136, 0.00000467, 0.00000388, -0.00000639, -0.00000738, 0.00000094, 0.00000653, 0.00000002, -0.00000582, -0.00000171, -0.00000096, 0.00000356, 0.00000561, -0.00000158, 0.00000231, -0.00000434, 0.00000078, 0.00000383, 0.00001331, -0.00000682, -0.00001045, 0.00002268, -0.00002465, 0.00000565, -0.00000053, -0.00001354, 0.00001422, -0.00001944, 0.00000691, -0.00001146, -0.00001292, 0.00001432, -0.00002889, 0.00001831, -0.00001852, -0.00000863, 0.00001967, -0.00003865, 0.00003251, -0.00010870, 0.00008970, -0.00033239, -0.00000034, 0.00094620, -0.00025975, -0.00021854, 0.00009124, -0.00015152, 0.00010832, -0.00010956, 0.00006405, -0.00005179, 0.00002171, -0.00001759, -0.00000346, 0.00001244, -0.00003976, 0.00005867, -0.00006596, 0.00006511, -0.00006341, 0.00004902, -0.00001535, -0.00001566, 0.00003961, -0.00007179, 0.00006283, -0.00002519, 0.00001326, -0.00001938, 0.00000587, 0.00000603, -0.00001176, 0.00001762, -0.00001344, 0.00001901, -0.00001222, -0.00000571, -0.00000194, 0.00001223, -0.00000961, -0.00000608, 0.00000588, 0.00000650, -0.00000164, -0.00001008, 0.00000873, -0.00000389, 0.00000949, -0.00000026, -0.00001079, 0.00000225, 0.00000056, 0.00000015, 0.00000044, -0.00000457, 0.00000131, 0.00000041, -0.00000139, 0.00000303, 0.00000068, 0.00000030, -0.00000090, 0.00000057, -0.00000284, -0.00000368, 0.00000081, -0.00000003, 0.00000562, 0.00000200, -0.00000612, -0.00000324, -0.00000242, 0.00000816, 0.00000285, -0.00000450, -0.00000047, -0.00000453, 0.00000355, 0.00000348, -0.00000023, 0.00000185, -0.00000722, -0.00000583, 0.00000007, 0.00000085, },
		{ 0.00000068, 0.00000250, 0.00000320, 0.00000088, 0.00000441, 0.00000492, -0.00000574, -0.00000765, -0.00000049, 0.00000645, 0.00000143, -0.00000545, -0.00000249, -0.00000150, 0.00000313, 0.00000620, -0.00000150, 0.00000248, -0.00000398, 0.00000067, 0.00000219, 0.00001244, -0.00000110, -0.00001654, 0.00002547, -0.00002128, -0.00000189, 0.00000702, -0.00001797, 0.00001508, -0.00001735, 0.00000279, -0.00000516, -0.00001921, 0.00001839, -0.00002979, 0.00001469, -0.00000924, -0.00002183, 0.00003380, -0.00005280, 0.00004844, -0.00011640, 0.00008457, -0.00025917, -0.00017645, 0.00099570, -0.00012160, -0.00032686, 0.00014205, -0.00018179, 0.00011658, -0.00010368, 0.00004803, -0.00003076, -0.00000175, 0.00000638, -0.00002600, 0.00003258, -0.00005660, 0.00006764, -0.00006732, 0.00005837, -0.00004955, 0.00002675, 0.00001081, -0.00003943, 0.00005526, -0.00007431, 0.00005339, -0.00001317, 0.00000470, -0.00001087, -0.00000300, 0.00001318, -0.00001636, 0.00001939, -0.00001360, 0.00001785, -0.00000848, -0.00000839, -0.00000211, 0.00001148, -0.00000581, -0.00000881, 0.00000559, 0.00000487, 0.00000190, -0.00001111, 0.00000717, -0.00000168, 0.00000669, 0.00000307, -0.00001157, 0.00000044, 0.00000206, -0.00000099, 0.00000153, -0.00000473, 0.00000075, 0.00000147, -0.00000290, 0.00000340, 0.00000112, 0.00000031, -0.00000102, 0.00000050, -0.00000182, -0.00000480, 0.00000107, 0.00000008, 0.00000432, 0.00000354, -0.00000600, -0.00000313, -0.00000318, 0.00000660, 0.00000506, -0.00000467, -0.00000058, -0.00000442, 0.00000239, 0.00000437, -0.00000100, 0.00000245, -0.00000550, -0.00000713, -0.00000045, },
		{ -0.00000131, 0.00000069, 0.00000194, 0.00000388, 0.00000062, 0.00000375, 0.00000607, -0.00000505, -0.00000775, -0.00000180, 0.00000597, 0.00000282, -0.00000488, -0.00000309, -0.00000217, 0.00000260, 0.00000683, -0.00000144, 0.00000263, -0.00000361, 0.00000050, 0.00000135, 0.00001010, 0.00000516, -0.00002135, 0.00002557, -0.00001515, -0.00001058, 0.00001354, -0.00002008, 0.00001353, -0.00001327, -0.00000265, 0.00000137, -0.00002434, 0.00002015, -0.00002745, 0.00000729, 0.00000332, -0.00003661, 0.00004714, -0.00006325, 0.00005609, -0.00011004, 0.00005838, -0.00016593, -0.00033763, 0.00099096, 0.00005595, -0.00043765, 0.00018087, -0.00019687, 0.00010899, -0.00008386, 0.00002111, -0.00000233, -0.00002908, 0.00003074, -0.00004599, 0.00004783, -0.00006651, 0.00006849, -0.00006047, 0.00004421, -0.00002983, 0.00000145, 0.00003573, -0.00005760, 0.00006248, -0.00006862, 0.00003871, 0.00000065, -0.00000396, -0.00000286, -0.00001061, 0.00001841, -0.00001882, 0.00001925, -0.00001229, 0.00001546, -0.00000408, -0.00001093, -0.00000234, 0.00001026, -0.00000193, -0.00001093, 0.00000507, 0.00000327, 0.00000445, -0.00001062, 0.00000458, 0.00000084, 0.00000418, 0.00000543, -0.00001102, -0.00000219, 0.00000364, -0.00000186, 0.00000209, -0.00000427, -0.00000035, 0.00000285, -0.00000422, 0.00000324, 0.00000188, 0.00000021, -0.00000095, 0.00000012, -0.00000069, -0.00000561, 0.00000081, 0.00000067, 0.00000288, 0.00000474, -0.00000554, -0.00000324, -0.00000344, 0.00000465, 0.00000693, -0.00000426, -0.00000103, -0.00000402, 0.00000111, 0.00000515, -0.00000143, 0.00000242, -0.00000349, -0.00000805, },
		{ -0.00000841, -0.00000252, 0.00000084, 0.00000126, 0.00000444, 0.00000073, 0.00000269, 0.00000724, -0.00000424, -0.00000778, -0.00000290, 0.00000511, 0.00000412, -0.00000414, -0.00000351, -0.00000290, 0.00000189, 0.00000752, -0.00000136, 0.00000270, -0.00000310, 0.00000007, 0.00000140, 0.00000664, 0.00001118, -0.00002414, 0.00002282, -0.00000689, -0.00001925, 0.00001796, -0.00001939, 0.00000969, -0.00000778, -0.00000852, 0.00000709, -0.00002731, 0.00001888, -0.00002162, -0.00000341, 0.00001779, -0.00005078, 0.00005700, -0.00006710, 0.00005305, -0.00008901, 0.00001477, -0.00006359, -0.00046966, 0.00092710, 0.00026433, -0.00053859, 0.00019886, -0.00019178, 0.00008367, -0.00005071, -0.00001444, 0.00003021, -0.00005651, 0.00005180, -0.00006024, 0.00005579, -0.00006802, 0.00006090, -0.00004620, 0.00002448, -0.00000697, -0.00002355, 0.00005613, -0.00006786, 0.00006071, -0.00005589, 0.00002085, 0.00001429, -0.00001131, 0.00000371, -0.00001631, 0.00002140, -0.00001914, 0.00001745, -0.00000982, 0.00001220, 0.00000057, -0.00001303, -0.00000275, 0.00000873, 0.00000176, -0.00001228, 0.00000421, 0.00000193, 0.00000585, -0.00000880, 0.00000146, 0.00000313, 0.00000239, 0.00000662, -0.00000927, -0.00000521, 0.00000488, -0.00000217, 0.00000200, -0.00000322, -0.00000182, 0.00000427, -0.00000509, 0.00000246, 0.00000289, 0.00000007, -0.00000076, -0.00000043, 0.00000037, -0.00000596, 0.00000003, 0.00000157, 0.00000153, 0.00000547, -0.00000472, -0.00000355, -0.00000328, 0.00000256, 0.00000824, -0.00000325, -0.00000173, -0.00000349, -0.00000012, 0.00000566, -0.00000144, 0.00000186, -0.00000143, },
		{ 0.00000041, -0.00000811, -0.00000403, 0.00000104, 0.00000059, 0.00000473, 0.00000130, 0.00000130, 0.00000827, -0.00000320, -0.00000784, -0.00000370, 0.00000392, 0.00000524, -0.00000321, -0.00000375, -0.00000360, 0.00000094, 0.00000822, -0.00000117, 0.00000259, -0.00000234, -0.00000072, 0.00000227, 0.00000257, 0.00001619, -0.00002444, 0.00001744, 0.00000253, -0.00002666, 0.00001946, -0.00001583, 0.00000411, -0.00000175, -0.00001378, 0.00001095, -0.00002738, 0.00001430, -0.00001268, -0.00001623, 0.00003220, -0.00006190, 0.00006085, -0.00006230, 0.00003854, -0.00005511, -0.00004000, 0.00003582, -0.00056072, 0.00080455, 0.00049046, -0.00061593, 0.00018824, -0.00016367, 0.00004109, -0.00000689, -0.00005474, 0.00006252, -0.00007987, 0.00006605, -0.00006624, 0.00005503, -0.00006089, 0.00004577, -0.00002643, 0.00000182, 0.00001590, -0.00004499, 0.00006941, -0.00006906, 0.00005066, -0.00003821, 0.00000225, 0.00002589, -0.00001629, 0.00000822, -0.00001980, 0.00002216, -0.00001759, 0.00001438, -0.00000660, 0.00000848, 0.00000507, -0.00001445, -0.00000341, 0.00000702, 0.00000503, -0.00001274, 0.00000298, 0.00000101, 0.00000608, -0.00000601, -0.00000166, 0.00000474, 0.00000159, 0.00000665, -0.00000665, -0.00000816, 0.00000541, -0.00000175, 0.00000126, -0.00000172, -0.00000340, 0.00000545, -0.00000531, 0.00000108, 0.00000403, -0.00000001, -0.00000051, -0.00000103, 0.00000117, -0.00000579, -0.00000115, 0.00000256, 0.00000048, 0.00000567, -0.00000360, -0.00000400, -0.00000286, 0.00000055, 0.00000887, -0.00000172, -0.00000253, -0.00000298, -0.00000116, 0.00000580, -0.00000104, 0.00000094, },
		{ -0.00000013, 0.00000180, -0.00000712, -0.00000569, 0.00000111, 0.00000010, 0.00000460, 0.00000234, -0.00000028, 0.00000896, -0.00000184, -0.00000796, -0.00000419, 0.00000249, 0.00000609, -0.00000211, -0.00000383, -0.00000419, -0.00000024, 0.00000886, -0.00000080, 0.00000222, -0.00000128, -0.00000188, 0.00000374, -0.00000153, 0.00001957, -0.00002213, 0.00001009, 0.00001192, -0.00003168, 0.00001764, -0.00000985, -0.00000228, 0.00000375, -0.00001743, 0.00001219, -0.00002423, 0.00000674, -0.00000162, -0.00002946, 0.00004430, -0.00006764, 0.00005691, -0.00004808, 0.00001362, -0.00001227, -0.00009795, 0.00012062, -0.00060291, 0.00062948, 0.00071766, -0.00065608, 0.00014361, -0.00011256, -0.00001567, 0.00004293, -0.00009469, 0.00008983, -0.00009522, 0.00007075, -0.00006258, 0.00004541, -0.00004611, 0.00002510, -0.00000387, -0.00002069, 0.00003567, -0.00006023, 0.00007402, -0.00006138, 0.00003414, -0.00001827, -0.00001467, 0.00003400, -0.00001830, 0.00001052, -0.00002115, 0.00002097, -0.00001461, 0.00001052, -0.00000306, 0.00000468, 0.00000909, -0.00001506, -0.00000433, 0.00000526, 0.00000770, -0.00001229, 0.00000139, 0.00000055, 0.00000529, -0.00000275, -0.00000423, 0.00000534, 0.00000181, 0.00000574, -0.00000361, -0.00001057, 0.00000498, -0.00000060, 0.00000003, -0.00000003, -0.00000481, 0.00000612, -0.00000480, -0.00000077, 0.00000511, 0.00000007, -0.00000028, -0.00000154, 0.00000162, -0.00000514, -0.00000255, 0.00000341, -0.00000013, 0.00000536, -0.00000230, -0.00000448, -0.00000233, -0.00000115, 0.00000875, 0.00000018, -0.00000321, -0.00000265, -0.00000191, 0.00000558, -0.00000033, },
		{ 0.00000055, -0.00000110, 0.00000259, -0.00000553, -0.00000729, 0.00000090, -0.00000008, 0.00000400, 0.00000375, -0.00000180, 0.00000913, -0.00000011, -0.00000815, -0.00000441, 0.00000095, 0.00000657, -0.00000084, -0.00000378, -0.00000462, -0.00000163, 0.00000934, -0.00000014, 0.00000153, 0.00000007, -0.00000332, 0.00000549, -0.00000509, 0.00002089, -0.00001743, 0.00000166, 0.00002008, -0.00003355, 0.00001262, -0.00000234, -0.00000833, 0.00000768, -0.00001867, 0.00001046, -0.00001812, -0.00000289, 0.00001003, -0.00004109, 0.00005193, -0.00006629, 0.00004444, -0.00002521, -0.00001881, 0.00003388, -0.00015049, 0.00018100, -0.00059318, 0.00041347, 0.00092709, -0.00064732, 0.00006313, -0.00004183, -0.00008106, 0.00009262, -0.00012865, 0.00010759, -0.00009946, 0.00006441, -0.00004925, 0.00002812, -0.00002586, 0.00000174, 0.00001836, -0.00004001, 0.00004974, -0.00006754, 0.00006968, -0.00004620, 0.00001382, 0.00000107, -0.00002784, 0.00003774, -0.00001730, 0.00001084, -0.00002068, 0.00001829, -0.00001073, 0.00000636, 0.00000043, 0.00000112, 0.00001237, -0.00001477, -0.00000550, 0.00000354, 0.00000963, -0.00001097, -0.00000046, 0.00000050, 0.00000377, 0.00000049, -0.00000584, 0.00000482, 0.00000290, 0.00000426, -0.00000064, -0.00001203, 0.00000350, 0.00000114, -0.00000144, 0.00000155, -0.00000578, 0.00000615, -0.00000358, -0.00000290, 0.00000592, 0.00000043, -0.00000013, -0.00000189, 0.00000169, -0.00000411, -0.00000395, 0.00000393, -0.00000024, 0.00000465, -0.00000095, -0.00000488, -0.00000183, -0.00000243, 0.00000794, 0.00000223, -0.00000360, -0.00000258, -0.00000234, 0.00000504, },
		{ 0.00000431, 0.00000141, -0.00000180, 0.00000274, -0.00000352, -0.00000858, 0.00000024, 0.00000012, 0.00000297, 0.00000535, -0.00000302, 0.00000863, 0.00000197, -0.00000835, -0.00000443, -0.00000053, 0.00000659, 0.00000057, -0.00000359, -0.00000486, -0.00000316, 0.00000955, 0.00000087, 0.00000054, 0.00000165, -0.00000484, 0.00000715, -0.00000763, 0.00002005, -0.00001092, -0.00000674, 0.00002598, -0.00003195, 0.00000505, 0.00000548, -0.00001283, 0.00000919, -0.00001711, 0.00000592, -0.00000987, -0.00001321, 0.00002048, -0.00004915, 0.00005342, -0.00005706, 0.00002402, 0.00000402, -0.00005446, 0.00007694, -0.00018964, 0.00021026, -0.00053392, 0.00017266, 0.00109969, -0.00058147, -0.00005070, 0.00004182, -0.00014753, 0.00013528, -0.00015124, 0.00011226, -0.00009092, 0.00004709, -0.00002773, 0.00000559, -0.00000313, -0.00002107, 0.00003716, -0.00005360, 0.00005641, -0.00006639, 0.00005736, -0.00002593, -0.00000723, 0.00001717, -0.00003587, 0.00003698, -0.00001379, 0.00000972, -0.00001891, 0.00001467, -0.00000646, 0.00000230, 0.00000359, -0.00000196, 0.00001477, -0.00001360, -0.00000685, 0.00000194, 0.00001077, -0.00000890, -0.00000244, 0.00000073, 0.00000184, 0.00000323, -0.00000626, 0.00000327, 0.00000455, 0.00000265, 0.00000182, -0.00001232, 0.00000106, 0.00000319, -0.00000283, 0.00000274, -0.00000614, 0.00000550, -0.00000181, -0.00000503, 0.00000626, 0.00000111, -0.00000008, -0.00000205, 0.00000141, -0.00000288, -0.00000516, 0.00000401, 0.00000011, 0.00000368, 0.00000029, -0.00000511, -0.00000147, -0.00000322, 0.00000660, 0.00000417, -0.00000354, -0.00000281, -0.00000252, },
		{ -0.00000254, 0.00000355, 0.00000211, -0.00000212, 0.00000231, -0.00000136, -0.00000932, -0.00000094, 0.00000068, 0.00000163, 0.00000691, -0.00000369, 0.00000738, 0.00000429, -0.00000846, -0.00000439, -0.00000181, 0.00000612, 0.00000205, -0.00000328, -0.00000490, -0.00000472, 0.00000936, 0.00000226, -0.00000069, 0.00000330, -0.00000624, 0.00000840, -0.00000887, 0.00001719, -0.00000341, -0.00001405, 0.00002894, -0.00002707, -0.00000399, 0.00001226, -0.00001480, 0.00000785, -0.00001291, -0.00000069, -0.00000077, -0.00002257, 0.00002796, -0.00005206, 0.00004789, -0.00004032, -0.00000240, 0.00003606, -0.00008819, 0.00011061, -0.00020921, 0.00020565, -0.00043268, -0.00007384, 0.00121824, -0.00045534, -0.00019068, 0.00012884, -0.00020642, 0.00016422, -0.00015815, 0.00010184, -0.00006967, 0.00002049, -0.00000079, -0.00001888, 0.00001866, -0.00004016, 0.00005000, -0.00005981, 0.00005510, -0.00005746, 0.00003912, -0.00000355, -0.00002603, 0.00002807, -0.00003822, 0.00003226, -0.00000866, 0.00000788, -0.00001640, 0.00001061, -0.00000224, -0.00000135, 0.00000620, -0.00000442, 0.00001621, -0.00001164, -0.00000829, 0.00000051, 0.00001111, -0.00000626, -0.00000435, 0.00000106, -0.00000014, 0.00000511, -0.00000542, 0.00000096, 0.00000634, 0.00000133, 0.00000343, -0.00001139, -0.00000207, 0.00000516, -0.00000382, 0.00000333, -0.00000582, 0.00000424, 0.00000027, -0.00000687, 0.00000600, 0.00000208, -0.00000010, -0.00000204, 0.00000087, -0.00000162, -0.00000602, 0.00000362, 0.00000081, 0.00000266, 0.00000128, -0.00000510, -0.00000128, -0.00000354, 0.00000491, 0.00000578, -0.00000295, -0.00000328, },
		{ -0.00000386, -0.00000256, 0.00000289, 0.00000252, -0.00000203, 0.00000144, 0.00000066, -0.00000933, -0.00000264, 0.00000148, 0.00000019, 0.00000815, -0.00000363, 0.00000542, 0.00000666, -0.00000833, -0.00000439, -0.00000276, 0.00000516, 0.00000352, -0.00000280, -0.00000479, -0.00000619, 0.00000870, 0.00000399, -0.00000206, 0.00000485, -0.00000728, 0.00000896, -0.00000869, 0.00001275, 0.00000420, -0.00001939, 0.00002871, -0.00001964, -0.00001313, 0.00001674, -0.00001361, 0.00000376, -0.00000674, -0.00000818, 0.00000765, -0.00002930, 0.00003102, -0.00004896, 0.00003551, -0.00001766, -0.00003177, 0.00006662, -0.00011483, 0.00012969, -0.00020568, 0.00016876, -0.00030138, -0.00030589, 0.00126935, -0.00027167, -0.00034524, 0.00020779, -0.00024911, 0.00017398, -0.00014694, 0.00007639, -0.00003767, -0.00001217, 0.00002782, -0.00004161, 0.00003623, -0.00005296, 0.00005525, -0.00005812, 0.00004642, -0.00004252, 0.00001769, 0.00001780, -0.00004007, 0.00003265, -0.00003524, 0.00002475, -0.00000301, 0.00000603, -0.00001364, 0.00000654, 0.00000162, -0.00000439, 0.00000817, -0.00000618, 0.00001669, -0.00000902, -0.00000971, -0.00000076, 0.00001073, -0.00000329, -0.00000598, 0.00000129, -0.00000187, 0.00000592, -0.00000349, -0.00000170, 0.00000783, 0.00000065, 0.00000404, -0.00000940, -0.00000550, 0.00000668, -0.00000417, 0.00000324, -0.00000488, 0.00000258, 0.00000238, -0.00000819, 0.00000507, 0.00000328, -0.00000011, -0.00000190, 0.00000019, -0.00000050, -0.00000644, 0.00000281, 0.00000169, 0.00000175, 0.00000193, -0.00000486, -0.00000128, -0.00000349, 0.00000309, 0.00000687, -0.00000186, },
		{ -0.00000037, -0.00000439, -0.00000271, 0.00000244, 0.00000262, -0.00000161, 0.00000038, 0.00000227, -0.00000851, -0.00000475, 0.00000235, -0.00000111, 0.00000884, -0.00000274, 0.00000290, 0.00000885, -0.00000780, -0.00000455, -0.00000331, 0.00000378, 0.00000484, -0.00000213, -0.00000455, -0.00000748, 0.00000751, 0.00000599, -0.00000339, 0.00000611, -0.00000779, 0.00000871, -0.00000721, 0.00000732, 0.00001101, -0.00002220, 0.00002553, -0.00001077, -0.00002096, 0.00001798, -0.00000919, -0.00000243, 0.00000027, -0.00001512, 0.00001382, -0.00003208, 0.00002888, -0.00003990, 0.00001754, 0.00000830, -0.00006032, 0.00009125, -0.00012998, 0.00013093, -0.00017871, 0.00010522, -0.00015493, -0.00050446, 0.00124521, -0.00003945, -0.00049926, 0.00026671, -0.00026825, 0.00016123, -0.00011752, 0.00003811, 0.00000145, -0.00004663, 0.00005393, -0.00005905, 0.00004691, -0.00005781, 0.00005242, -0.00004916, 0.00003202, -0.00002408, -0.00000388, 0.00003533, -0.00004770, 0.00003086, -0.00002809, 0.00001594, 0.00000204, 0.00000474, -0.00001095, 0.00000272, 0.00000492, -0.00000674, 0.00000948, -0.00000723, 0.00001628, -0.00000590, -0.00001101, -0.00000185, 0.00000975, -0.00000026, -0.00000714, 0.00000129, -0.00000312, 0.00000564, -0.00000078, -0.00000424, 0.00000863, 0.00000082, 0.00000368, -0.00000665, -0.00000876, 0.00000741, -0.00000375, 0.00000250, -0.00000347, 0.00000077, 0.00000423, -0.00000880, 0.00000353, 0.00000454, -0.00000001, -0.00000170, -0.00000051, 0.00000035, -0.00000640, 0.00000171, 0.00000257, 0.00000110, 0.00000224, -0.00000444, -0.00000140, -0.00000317, 0.00000135, 0.00000735, },
		{ 0.00000721, 0.00000135, -0.00000467, -0.00000310, 0.00000223, 0.00000245, -0.00000100, -0.00000065, 0.00000324, -0.00000688, -0.00000704, 0.00000303, -0.00000204, 0.00000880, -0.00000105, 0.00000008, 0.00001057, -0.00000675, -0.00000493, -0.00000344, 0.00000210, 0.00000589, -0.00000124, -0.00000423, -0.00000847, 0.00000581, 0.00000811, -0.00000450, 0.00000693, -0.00000764, 0.00000762, -0.00000470, 0.00000157, 0.00001629, -0.00002228, 0.00002010, -0.00000176, -0.00002625, 0.00001556, -0.00000208, -0.00000955, 0.00000668, -0.00002001, 0.00001647, -0.00003015, 0.00002153, -0.00002589, -0.00000382, 0.00003424, -0.00008410, 0.00010612, -0.00013070, 0.00011349, -0.00013126, 0.00002396, -0.00000947, -0.00065370, 0.00114466, 0.00022656, -0.00063564, 0.00029471, -0.00025893, 0.00012547, -0.00007235, -0.00000881, 0.00004279, -0.00007812, 0.00007357, -0.00006839, 0.00004909, -0.00005433, 0.00004226, -0.00003465, 0.00001431, -0.00000498, -0.00002274, 0.00004698, -0.00004832, 0.00002358, -0.00001846, 0.00000745, 0.00000558, 0.00000431, -0.00000846, -0.00000073, 0.00000760, -0.00000840, 0.00001015, -0.00000760, 0.00001508, -0.00000248, -0.00001207, -0.00000281, 0.00000831, 0.00000257, -0.00000766, 0.00000094, -0.00000380, 0.00000438, 0.00000229, -0.00000621, 0.00000850, 0.00000185, 0.00000254, -0.00000355, -0.00001143, 0.00000712, -0.00000258, 0.00000127, -0.00000186, -0.00000092, 0.00000560, -0.00000864, 0.00000151, 0.00000568, 0.00000028, -0.00000150, -0.00000115, 0.00000088, -0.00000597, 0.00000049, 0.00000328, 0.00000079, 0.00000224, -0.00000392, -0.00000158, -0.00000272, -0.00000015, },
		{ -0.00000132, 0.00000652, 0.00000308, -0.00000455, -0.00000378, 0.00000222, 0.00000212, -0.00000039, -0.00000140, 0.00000348, -0.00000461, -0.00000924, 0.00000326, -0.00000239, 0.00000799, 0.00000126, -0.00000272, 0.00001156, -0.00000511, -0.00000552, -0.00000322, 0.00000029, 0.00000653, -0.00000011, -0.00000387, -0.00000912, 0.00000367, 0.00001018, -0.00000523, 0.00000718, -0.00000682, 0.00000582, -0.00000152, -0.00000386, 0.00001958, -0.00001989, 0.00001340, 0.00000606, -0.00002818, 0.00000970, 0.00000661, -0.00001614, 0.00001105, -0.00002165, 0.00001486, -0.00002351, 0.00000990, -0.00000881, -0.00002572, 0.00005665, -0.00009963, 0.00010856, -0.00011603, 0.00007920, -0.00006913, -0.00006404, 0.00011955, -0.00074270, 0.00097364, 0.00050694, -0.00073706, 0.00028350, -0.00021968, 0.00006929, -0.00001629, -0.00005872, 0.00008087, -0.00010210, 0.00008359, -0.00006797, 0.00004246, -0.00004334, 0.00002657, -0.00001703, -0.00000394, 0.00001202, -0.00003666, 0.00005165, -0.00004239, 0.00001250, -0.00000832, 0.00000066, 0.00000708, 0.00000472, -0.00000612, -0.00000380, 0.00000968, -0.00000942, 0.00001022, -0.00000734, 0.00001321, 0.00000105, -0.00001278, -0.00000367, 0.00000661, 0.00000498, -0.00000748, 0.00000023, -0.00000391, 0.00000239, 0.00000526, -0.00000725, 0.00000735, 0.00000360, 0.00000095, -0.00000055, -0.00001315, 0.00000576, -0.00000082, -0.00000019, -0.00000031, -0.00000226, 0.00000636, -0.00000775, -0.00000077, 0.00000650, 0.00000083, -0.00000133, -0.00000166, 0.00000108, -0.00000526, -0.00000070, 0.00000371, 0.00000080, 0.00000204, -0.00000341, -0.00000175, -0.00000223, },
		{ -0.00000178, -0.00000211, 0.00000544, 0.00000460, -0.00000392, -0.00000471, 0.00000232, 0.00000180, 0.00000005, -0.00000174, 0.00000299, -0.00000195, -0.00001100, 0.00000284, -0.00000210, 0.00000648, 0.00000393, -0.00000513, 0.00001163, -0.00000292, -0.00000624, -0.00000274, -0.00000145, 0.00000664, 0.00000125, -0.00000348, -0.00000940, 0.00000119, 0.00001201, -0.00000544, 0.00000681, -0.00000538, 0.00000348, 0.00000191, -0.00000843, 0.00002072, -0.00001559, 0.00000659, 0.00001162, -0.00002649, 0.00000123, 0.00001536, -0.00002064, 0.00001216, -0.00001934, 0.00000899, -0.00001299, -0.00000434, 0.00000886, -0.00004505, 0.00007234, -0.00010442, 0.00009759, -0.00008720, 0.00003227, -0.00000015, -0.00014718, 0.00021914, -0.00076661, 0.00074479, 0.00077947, -0.00078810, 0.00022883, -0.00015294, -0.00000169, 0.00004396, -0.00010520, 0.00011041, -0.00011489, 0.00008213, -0.00005758, 0.00002806, -0.00002682, 0.00000792, 0.00000089, -0.00002005, 0.00002469, -0.00004428, 0.00004937, -0.00003133, -0.00000023, 0.00000042, -0.00000343, 0.00000648, 0.00000567, -0.00000372, -0.00000655, 0.00001120, -0.00000985, 0.00000975, -0.00000651, 0.00001085, 0.00000447, -0.00001305, -0.00000448, 0.00000481, 0.00000679, -0.00000659, -0.00000077, -0.00000357, 0.00000002, 0.00000768, -0.00000715, 0.00000528, 0.00000577, -0.00000068, 0.00000194, -0.00001370, 0.00000343, 0.00000125, -0.00000160, 0.00000093, -0.00000307, 0.00000647, -0.00000630, -0.00000305, 0.00000683, 0.00000166, -0.00000117, -0.00000203, 0.00000101, -0.00000443, -0.00000169, 0.00000380, 0.00000109, 0.00000176, -0.00000301, -0.00000186, },
		{ -0.00000191, -0.00000139, -0.00000258, 0.00000413, 0.00000571, -0.00000279, -0.00000576, 0.00000235, 0.00000163, 0.00000017, -0.00000159, 0.00000191, 0.00000076, -0.00001203, 0.00000164, -0.00000118, 0.00000446, 0.00000660, -0.00000681, 0.00001067, -0.00000030, -0.00000696, -0.00000217, -0.00000293, 0.00000615, 0.00000277, -0.00000304, -0.00000934, -0.00000147, 0.00001342, -0.00000502, 0.00000585, -0.00000345, 0.00000088, 0.00000519, -0.00001180, 0.00001986, -0.00001020, 0.00000074, 0.00001428, -0.00002155, -0.00000849, 0.00002248, -0.00002171, 0.00000932, -0.00001307, -0.00000038, -0.00000016, -0.00001895, 0.00002441, -0.00005899, 0.00007899, -0.00009740, 0.00007409, -0.00004750, -0.00002127, 0.00006692, -0.00021483, 0.00028028, -0.00072715, 0.00047625, 0.00102130, -0.00077717, 0.00013141, -0.00006509, -0.00007945, 0.00010071, -0.00014202, 0.00012713, -0.00011432, 0.00006897, -0.00003851, 0.00000813, -0.00000753, -0.00001076, 0.00001640, -0.00003185, 0.00003163, -0.00004529, 0.00004114, -0.00001721, -0.00001238, 0.00000627, -0.00000442, 0.00000417, 0.00000664, -0.00000104, -0.00000903, 0.00001219, -0.00000974, 0.00000879, -0.00000521, 0.00000815, 0.00000760, -0.00001282, -0.00000528, 0.00000306, 0.00000790, -0.00000507, -0.00000193, -0.00000295, -0.00000240, 0.00000924, -0.00000588, 0.00000254, 0.00000798, -0.00000199, 0.00000362, -0.00001304, 0.00000038, 0.00000328, -0.00000267, 0.00000167, -0.00000331, 0.00000600, -0.00000449, -0.00000506, 0.00000658, 0.00000271, -0.00000097, -0.00000229, 0.00000074, -0.00000360, -0.00000240, 0.00000358, 0.00000153, 0.00000154, -0.00000276, },
		{ -0.00000270, -0.00000193, -0.00000106, -0.00000277, 0.00000279, 0.00000630, -0.00000121, -0.00000677, 0.00000216, 0.00000176, -0.00000008, -0.00000102, 0.00000048, 0.00000316, -0.00001210, -0.00000033, 0.00000020, 0.00000222, 0.00000892, -0.00000752, 0.00000873, 0.00000253, -0.00000748, -0.00000166, -0.00000399, 0.00000505, 0.00000435, -0.00000252, -0.00000899, -0.00000415, 0.00001424, -0.00000396, 0.00000439, -0.00000120, -0.00000176, 0.00000804, -0.00001383, 0.00001738, -0.00000462, -0.00000330, 0.00001391, -0.00001431, -0.00001777, 0.00002640, -0.00001849, 0.00000249, -0.00000359, -0.00001178, 0.00001293, -0.00003150, 0.00003536, -0.00006543, 0.00007547, -0.00007917, 0.00004075, -0.00000181, -0.00007442, 0.00012363, -0.00025869, 0.00029883, -0.00063227, 0.00018983, 0.00121140, -0.00069817, -0.00000268, 0.00003410, -0.00015452, 0.00014634, -0.00016402, 0.00012835, -0.00010001, 0.00004556, -0.00001333, -0.00001427, 0.00001141, -0.00002667, 0.00002731, -0.00003798, 0.00003245, -0.00004038, 0.00002877, -0.00000239, -0.00002195, 0.00000835, -0.00000253, 0.00000090, 0.00000702, 0.00000211, -0.00001122, 0.00001265, -0.00000910, 0.00000739, -0.00000351, 0.00000531, 0.00001028, -0.00001204, -0.00000608, 0.00000146, 0.00000828, -0.00000309, -0.00000306, -0.00000226, -0.00000455, 0.00000973, -0.00000359, -0.00000048, 0.00000985, -0.00000266, 0.00000436, -0.00001132, -0.00000300, 0.00000490, -0.00000318, 0.00000185, -0.00000302, 0.00000513, -0.00000259, -0.00000661, 0.00000572, 0.00000387, -0.00000067, -0.00000247, 0.00000037, -0.00000291, -0.00000281, 0.00000312, 0.00000198, 0.00000149, },
		{ 0.00000166, -0.00000277, -0.00000197, -0.00000074, -0.00000280, 0.00000157, 0.00000634, 0.00000065, -0.00000752, 0.00000160, 0.00000224, -0.00000068, -0.00000016, -0.00000101, 0.00000494, -0.00001112, -0.00000293, 0.00000177, 0.00000011, 0.00001056, -0.00000714, 0.00000599, 0.00000526, -0.00000763, -0.00000137, -0.00000452, 0.00000342, 0.00000586, -0.00000185, -0.00000841, -0.00000669, 0.00001436, -0.00000228, 0.00000256, 0.00000116, -0.00000421, 0.00001025, -0.00001458, 0.00001384, 0.00000033, -0.00000508, 0.00001099, -0.00000618, -0.00002491, 0.00002597, -0.00001083, -0.00000758, 0.00000768, -0.00002326, 0.00002398, -0.00003973, 0.00003985, -0.00006332, 0.00006216, -0.00005192, 0.00000169, 0.00004405, -0.00012022, 0.00016299, -0.00027378, 0.00027582, -0.00049512, -0.00009121, 0.00133282, -0.00055155, -0.00016234, 0.00013250, -0.00021722, 0.00017436, -0.00016790, 0.00011350, -0.00007352, 0.00001485, 0.00001441, -0.00003570, 0.00002704, -0.00003753, 0.00003227, -0.00003806, 0.00002772, -0.00003108, 0.00001450, 0.00001085, -0.00002753, 0.00000649, 0.00000147, -0.00000236, 0.00000627, 0.00000577, -0.00001301, 0.00001252, -0.00000795, 0.00000559, -0.00000156, 0.00000250, 0.00001236, -0.00001073, -0.00000690, 0.00000011, 0.00000799, -0.00000086, -0.00000397, -0.00000169, -0.00000620, 0.00000914, -0.00000058, -0.00000336, 0.00001102, -0.00000253, 0.00000416, -0.00000880, -0.00000631, 0.00000583, -0.00000304, 0.00000150, -0.00000235, 0.00000408, -0.00000083, -0.00000754, 0.00000434, 0.00000498, -0.00000017, -0.00000261, -0.00000003, -0.00000241, -0.00000294, 0.00000257, 0.00000231, },
		{ 0.00000243, 0.00000204, -0.00000289, -0.00000210, -0.00000040, -0.00000272, 0.00000057, 0.00000585, 0.00000257, -0.00000782, 0.00000060, 0.00000305, -0.00000150, 0.00000076, -0.00000227, 0.00000586, -0.00000912, -0.00000587, 0.00000320, -0.00000156, 0.00001128, -0.00000569, 0.00000275, 0.00000757, -0.00000723, -0.00000137, -0.00000450, 0.00000139, 0.00000713, -0.00000099, -0.00000769, -0.00000892, 0.00001374, -0.00000008, 0.00000052, 0.00000348, -0.00000630, 0.00001174, -0.00001422, 0.00000980, 0.00000406, -0.00000462, 0.00000642, 0.00000126, -0.00002846, 0.00002068, 0.00000063, -0.00001942, 0.00001882, -0.00003268, 0.00003094, -0.00004199, 0.00003699, -0.00005291, 0.00004082, -0.00001914, -0.00003811, 0.00008419, -0.00015274, 0.00018043, -0.00025903, 0.00021714, -0.00033239, -0.00034464, 0.00137456, -0.00034464, -0.00033239, 0.00021714, -0.00025903, 0.00018043, -0.00015274, 0.00008419, -0.00003811, -0.00001914, 0.00004082, -0.00005291, 0.00003699, -0.00004199, 0.00003094, -0.00003268, 0.00001882, -0.00001942, 0.00000063, 0.00002068, -0.00002846, 0.00000126, 0.00000642, -0.00000462, 0.00000406, 0.00000980, -0.00001422, 0.00001174, -0.00000630, 0.00000348, 0.00000052, -0.00000008, 0.00001374, -0.00000892, -0.00000769, -0.00000099, 0.00000713, 0.00000139, -0.00000450, -0.00000137, -0.00000723, 0.00000757, 0.00000275, -0.00000569, 0.00001128, -0.00000156, 0.00000320, -0.00000587, -0.00000912, 0.00000586, -0.00000227, 0.00000076, -0.00000150, 0.00000305, 0.00000060, -0.00000782, 0.00000257, 0.00000585, 0.00000057, -0.00000272, -0.00000040, -0.00000210, -0.00000289, 0.00000204, },
		{ 0.00000166, 0.00000231, 0.00000257, -0.00000294, -0.00000241, -0.00000003, -0.00000261, -0.00000017, 0.00000498, 0.00000434, -0.00000754, -0.00000083, 0.00000408, -0.00000235, 0.00000150, -0.00000304, 0.00000583, -0.00000631, -0.00000880, 0.00000416, -0.00000253, 0.00001102, -0.00000336, -0.00000058, 0.00000914, -0.00000620, -0.00000169, -0.00000397, -0.00000086, 0.00000799, 0.00000011, -0.00000690, -0.00001073, 0.00001236, 0.00000250, -0.00000156, 0.00000559, -0.00000795, 0.00001252, -0.00001301, 0.00000577, 0.00000627, -0.00000236, 0.00000147, 0.00000649, -0.00002753, 0.00001085, 0.00001450, -0.00003108, 0.00002772, -0.00003806, 0.00003227, -0.00003753, 0.00002704, -0.00003570, 0.00001441, 0.00001485, -0.00007352, 0.00011350, -0.00016790, 0.00017436, -0.00021722, 0.00013250, -0.00016234, -0.00055155, 0.00133282, -0.00009121, -0.00049512, 0.00027582, -0.00027378, 0.00016299, -0.00012022, 0.00004405, 0.00000169, -0.00005192, 0.00006216, -0.00006332, 0.00003985, -0.00003973, 0.00002398, -0.00002326, 0.00000768, -0.00000758, -0.00001083, 0.00002597, -0.00002491, -0.00000618, 0.00001099, -0.00000508, 0.00000033, 0.00001384, -0.00001458, 0.00001025, -0.00000421, 0.00000116, 0.00000256, -0.00000228, 0.00001436, -0.00000669, -0.00000841, -0.00000185, 0.00000586, 0.00000342, -0.00000452, -0.00000137, -0.00000763, 0.00000526, 0.00000599, -0.00000714, 0.00001056, 0.00000011, 0.00000177, -0.00000293, -0.00001112, 0.00000494, -0.00000101, -0.00000016, -0.00000068, 0.00000224, 0.00000160, -0.00000752, 0.00000065, 0.00000634, 0.00000157, -0.00000280, -0.00000074, -0.00000197, -0.00000277, },
		{ -0.00000270, 0.00000149, 0.00000198, 0.00000312, -0.00000281, -0.00000291, 0.00000037, -0.00000247, -0.00000067, 0.00000387, 0.00000572, -0.00000661, -0.00000259, 0.00000513, -0.00000302, 0.00000185, -0.00000318, 0.00000490, -0.00000300, -0.00001132, 0.00000436, -0.00000266, 0.00000985, -0.00000048, -0.00000359, 0.00000973, -0.00000455, -0.00000226, -0.00000306, -0.00000309, 0.00000828, 0.00000146, -0.00000608, -0.00001204, 0.00001028, 0.00000531, -0.00000351, 0.00000739, -0.00000910, 0.00001265, -0.00001122, 0.00000211, 0.00000702, 0.00000090, -0.00000253, 0.00000835, -0.00002195, -0.00000239, 0.00002877, -0.00004038, 0.00003245, -0.00003798, 0.00002731, -0.00002667, 0.00001141, -0.00001427, -0.00001333, 0.00004556, -0.00010001, 0.00012835, -0.00016402, 0.00014634, -0.00015452, 0.00003410, -0.00000268, -0.00069817, 0.00121140, 0.00018983, -0.00063227, 0.00029883, -0.00025869, 0.00012363, -0.00007442, -0.00000181, 0.00004075, -0.00007917, 0.00007547, -0.00006543, 0.00003536, -0.00003150, 0.00001293, -0.00001178, -0.00000359, 0.00000249, -0.00001849, 0.00002640, -0.00001777, -0.00001431, 0.00001391, -0.00000330, -0.00000462, 0.00001738, -0.00001383, 0.00000804, -0.00000176, -0.00000120, 0.00000439, -0.00000396, 0.00001424, -0.00000415, -0.00000899, -0.00000252, 0.00000435, 0.00000505, -0.00000399, -0.00000166, -0.00000748, 0.00000253, 0.00000873, -0.00000752, 0.00000892, 0.00000222, 0.00000020, -0.00000033, -0.00001210, 0.00000316, 0.00000048, -0.00000102, -0.00000008, 0.00000176, 0.00000216, -0.00000677, -0.00000121, 0.00000630, 0.00000279, -0.00000277, -0.00000106, -0.00000193, },
		{ -0.00000191, -0.00000276, 0.00000154, 0.00000153, 0.00000358, -0.00000240, -0.00000360, 0.00000074, -0.00000229, -0.00000097, 0.00000271, 0.00000658, -0.00000506, -0.00000449, 0.00000600, -0.00000331, 0.00000167, -0.00000267, 0.00000328, 0.00000038, -0.00001304, 0.00000362, -0.00000199, 0.00000798, 0.00000254, -0.00000588, 0.00000924, -0.00000240, -0.00000295, -0.00000193, -0.00000507, 0.00000790, 0.00000306, -0.00000528, -0.00001282, 0.00000760, 0.00000815, -0.00000521, 0.00000879, -0.00000974, 0.00001219, -0.00000903, -0.00000104, 0.00000664, 0.00000417, -0.00000442, 0.00000627, -0.00001238, -0.00001721, 0.00004114, -0.00004529, 0.00003163, -0.00003185, 0.00001640, -0.00001076, -0.00000753, 0.00000813, -0.00003851, 0.00006897, -0.00011432, 0.00012713, -0.00014202, 0.00010071, -0.00007945, -0.00006509, 0.00013141, -0.00077717, 0.00102130, 0.00047625, -0.00072715, 0.00028028, -0.00021483, 0.00006692, -0.00002127, -0.00004750, 0.00007409, -0.00009740, 0.00007899, -0.00005899, 0.00002441, -0.00001895, -0.00000016, -0.00000038, -0.00001307, 0.00000932, -0.00002171, 0.00002248, -0.00000849, -0.00002155, 0.00001428, 0.00000074, -0.00001020, 0.00001986, -0.00001180, 0.00000519, 0.00000088, -0.00000345, 0.00000585, -0.00000502, 0.00001342, -0.00000147, -0.00000934, -0.00000304, 0.00000277, 0.00000615, -0.00000293, -0.00000217, -0.00000696, -0.00000030, 0.00001067, -0.00000681, 0.00000660, 0.00000446, -0.00000118, 0.00000164, -0.00001203, 0.00000076, 0.00000191, -0.00000159, 0.00000017, 0.00000163, 0.00000235, -0.00000576, -0.00000279, 0.00000571, 0.00000413, -0.00000258, -0.00000139, },
		{ -0.00000178, -0.00000186, -0.00000301, 0.00000176, 0.00000109, 0.00000380, -0.00000169, -0.00000443, 0.00000101, -0.00000203, -0.00000117, 0.00000166, 0.00000683, -0.00000305, -0.00000630, 0.00000647, -0.00000307, 0.00000093, -0.00000160, 0.00000125, 0.00000343, -0.00001370, 0.00000194, -0.00000068, 0.00000577, 0.00000528, -0.00000715, 0.00000768, 0.00000002, -0.00000357, -0.00000077, -0.00000659, 0.00000679, 0.00000481, -0.00000448, -0.00001305, 0.00000447, 0.00001085, -0.00000651, 0.00000975, -0.00000985, 0.00001120, -0.00000655, -0.00000372, 0.00000567, 0.00000648, -0.00000343, 0.00000042, -0.00000023, -0.00003133, 0.00004937, -0.00004428, 0.00002469, -0.00002005, 0.00000089, 0.00000792, -0.00002682, 0.00002806, -0.00005758, 0.00008213, -0.00011489, 0.00011041, -0.00010520, 0.00004396, -0.00000169, -0.00015294, 0.00022883, -0.00078810, 0.00077947, 0.00074479, -0.00076661, 0.00021914, -0.00014718, -0.00000015, 0.00003227, -0.00008720, 0.00009759, -0.00010442, 0.00007234, -0.00004505, 0.00000886, -0.00000434, -0.00001299, 0.00000899, -0.00001934, 0.00001216, -0.00002064, 0.00001536, 0.00000123, -0.00002649, 0.00001162, 0.00000659, -0.00001559, 0.00002072, -0.00000843, 0.00000191, 0.00000348, -0.00000538, 0.00000681, -0.00000544, 0.00001201, 0.00000119, -0.00000940, -0.00000348, 0.00000125, 0.00000664, -0.00000145, -0.00000274, -0.00000624, -0.00000292, 0.00001163, -0.00000513, 0.00000393, 0.00000648, -0.00000210, 0.00000284, -0.00001100, -0.00000195, 0.00000299, -0.00000174, 0.00000005, 0.00000180, 0.00000232, -0.00000471, -0.00000392, 0.00000460, 0.00000544, -0.00000211, },
		{ -0.00000132, -0.00000223, -0.00000175, -0.00000341, 0.00000204, 0.00000080, 0.00000371, -0.00000070, -0.00000526, 0.00000108, -0.00000166, -0.00000133, 0.00000083, 0.00000650, -0.00000077, -0.00000775, 0.00000636, -0.00000226, -0.00000031, -0.00000019, -0.00000082, 0.00000576, -0.00001315, -0.00000055, 0.00000095, 0.00000360, 0.00000735, -0.00000725, 0.00000526, 0.00000239, -0.00000391, 0.00000023, -0.00000748, 0.00000498, 0.00000661, -0.00000367, -0.00001278, 0.00000105, 0.00001321, -0.00000734, 0.00001022, -0.00000942, 0.00000968, -0.00000380, -0.00000612, 0.00000472, 0.00000708, 0.00000066, -0.00000832, 0.00001250, -0.00004239, 0.00005165, -0.00003666, 0.00001202, -0.00000394, -0.00001703, 0.00002657, -0.00004334, 0.00004246, -0.00006797, 0.00008359, -0.00010210, 0.00008087, -0.00005872, -0.00001629, 0.00006929, -0.00021968, 0.00028350, -0.00073706, 0.00050694, 0.00097364, -0.00074270, 0.00011955, -0.00006404, -0.00006913, 0.00007920, -0.00011603, 0.00010856, -0.00009963, 0.00005665, -0.00002572, -0.00000881, 0.00000990, -0.00002351, 0.00001486, -0.00002165, 0.00001105, -0.00001614, 0.00000661, 0.00000970, -0.00002818, 0.00000606, 0.00001340, -0.00001989, 0.00001958, -0.00000386, -0.00000152, 0.00000582, -0.00000682, 0.00000718, -0.00000523, 0.00001018, 0.00000367, -0.00000912, -0.00000387, -0.00000011, 0.00000653, 0.00000029, -0.00000322, -0.00000552, -0.00000511, 0.00001156, -0.00000272, 0.00000126, 0.00000799, -0.00000239, 0.00000326, -0.00000924, -0.00000461, 0.00000348, -0.00000140, -0.00000039, 0.00000212, 0.00000222, -0.00000378, -0.00000455, 0.00000308, 0.00000652, },
		{ 0.00000721, -0.00000015, -0.00000272, -0.00000158, -0.00000392, 0.00000224, 0.00000079, 0.00000328, 0.00000049, -0.00000597, 0.00000088, -0.00000115, -0.00000150, 0.00000028, 0.00000568, 0.00000151, -0.00000864, 0.00000560, -0.00000092, -0.00000186, 0.00000127, -0.00000258, 0.00000712, -0.00001143, -0.00000355, 0.00000254, 0.00000185, 0.00000850, -0.00000621, 0.00000229, 0.00000438, -0.00000380, 0.00000094, -0.00000766, 0.00000257, 0.00000831, -0.00000281, -0.00001207, -0.00000248, 0.00001508, -0.00000760, 0.00001015, -0.00000840, 0.00000760, -0.00000073, -0.00000846, 0.00000431, 0.00000558, 0.00000745, -0.00001846, 0.00002358, -0.00004832, 0.00004698, -0.00002274, -0.00000498, 0.00001431, -0.00003465, 0.00004226, -0.00005433, 0.00004909, -0.00006839, 0.00007357, -0.00007812, 0.00004279, -0.00000881, -0.00007235, 0.00012547, -0.00025893, 0.00029471, -0.00063564, 0.00022656, 0.00114466, -0.00065370, -0.00000947, 0.00002396, -0.00013126, 0.00011349, -0.00013070, 0.00010612, -0.00008410, 0.00003424, -0.00000382, -0.00002589, 0.00002153, -0.00003015, 0.00001647, -0.00002001, 0.00000668, -0.00000955, -0.00000208, 0.00001556, -0.00002625, -0.00000176, 0.00002010, -0.00002228, 0.00001629, 0.00000157, -0.00000470, 0.00000762, -0.00000764, 0.00000693, -0.00000450, 0.00000811, 0.00000581, -0.00000847, -0.00000423, -0.00000124, 0.00000589, 0.00000210, -0.00000344, -0.00000493, -0.00000675, 0.00001057, 0.00000008, -0.00000105, 0.00000880, -0.00000204, 0.00000303, -0.00000704, -0.00000688, 0.00000324, -0.00000065, -0.00000100, 0.00000245, 0.00000223, -0.00000310, -0.00000467, 0.00000135, },
		{ -0.00000037, 0.00000735, 0.00000135, -0.00000317, -0.00000140, -0.00000444, 0.00000224, 0.00000110, 0.00000257, 0.00000171, -0.00000640, 0.00000035, -0.00000051, -0.00000170, -0.00000001, 0.00000454, 0.00000353, -0.00000880, 0.00000423, 0.00000077, -0.00000347, 0.00000250, -0.00000375, 0.00000741, -0.00000876, -0.00000665, 0.00000368, 0.00000082, 0.00000863, -0.00000424, -0.00000078, 0.00000564, -0.00000312, 0.00000129, -0.00000714, -0.00000026, 0.00000975, -0.00000185, -0.00001101, -0.00000590, 0.00001628, -0.00000723, 0.00000948, -0.00000674, 0.00000492, 0.00000272, -0.00001095, 0.00000474, 0.00000204, 0.00001594, -0.00002809, 0.00003086, -0.00004770, 0.00003533, -0.00000388, -0.00002408, 0.00003202, -0.00004916, 0.00005242, -0.00005781, 0.00004691, -0.00005905, 0.00005393, -0.00004663, 0.00000145, 0.00003811, -0.00011752, 0.00016123, -0.00026825, 0.00026671, -0.00049926, -0.00003945, 0.00124521, -0.00050446, -0.00015493, 0.00010522, -0.00017871, 0.00013093, -0.00012998, 0.00009125, -0.00006032, 0.00000830, 0.00001754, -0.00003990, 0.00002888, -0.00003208, 0.00001382, -0.00001512, 0.00000027, -0.00000243, -0.00000919, 0.00001798, -0.00002096, -0.00001077, 0.00002553, -0.00002220, 0.00001101, 0.00000732, -0.00000721, 0.00000871, -0.00000779, 0.00000611, -0.00000339, 0.00000599, 0.00000751, -0.00000748, -0.00000455, -0.00000213, 0.00000484, 0.00000378, -0.00000331, -0.00000455, -0.00000780, 0.00000885, 0.00000290, -0.00000274, 0.00000884, -0.00000111, 0.00000235, -0.00000475, -0.00000851, 0.00000227, 0.00000038, -0.00000161, 0.00000262, 0.00000244, -0.00000271, -0.00000439, },
		{ -0.00000386, -0.00000186, 0.00000687, 0.00000309, -0.00000349, -0.00000128, -0.00000486, 0.00000193, 0.00000175, 0.00000169, 0.00000281, -0.00000644, -0.00000050, 0.00000019, -0.00000190, -0.00000011, 0.00000328, 0.00000507, -0.00000819, 0.00000238, 0.00000258, -0.00000488, 0.00000324, -0.00000417, 0.00000668, -0.00000550, -0.00000940, 0.00000404, 0.00000065, 0.00000783, -0.00000170, -0.00000349, 0.00000592, -0.00000187, 0.00000129, -0.00000598, -0.00000329, 0.00001073, -0.00000076, -0.00000971, -0.00000902, 0.00001669, -0.00000618, 0.00000817, -0.00000439, 0.00000162, 0.00000654, -0.00001364, 0.00000603, -0.00000301, 0.00002475, -0.00003524, 0.00003265, -0.00004007, 0.00001780, 0.00001769, -0.00004252, 0.00004642, -0.00005812, 0.00005525, -0.00005296, 0.00003623, -0.00004161, 0.00002782, -0.00001217, -0.00003767, 0.00007639, -0.00014694, 0.00017398, -0.00024911, 0.00020779, -0.00034524, -0.00027167, 0.00126935, -0.00030589, -0.00030138, 0.00016876, -0.00020568, 0.00012969, -0.00011483, 0.00006662, -0.00003177, -0.00001766, 0.00003551, -0.00004896, 0.00003102, -0.00002930, 0.00000765, -0.00000818, -0.00000674, 0.00000376, -0.00001361, 0.00001674, -0.00001313, -0.00001964, 0.00002871, -0.00001939, 0.00000420, 0.00001275, -0.00000869, 0.00000896, -0.00000728, 0.00000485, -0.00000206, 0.00000399, 0.00000870, -0.00000619, -0.00000479, -0.00000280, 0.00000352, 0.00000516, -0.00000276, -0.00000439, -0.00000833, 0.00000666, 0.00000542, -0.00000363, 0.00000815, 0.00000019, 0.00000148, -0.00000264, -0.00000933, 0.00000066, 0.00000144, -0.00000203, 0.00000252, 0.00000289, -0.00000256, },
		{ -0.00000254, -0.00000328, -0.00000295, 0.00000578, 0.00000491, -0.00000354, -0.00000128, -0.00000510, 0.00000128, 0.00000266, 0.00000081, 0.00000362, -0.00000602, -0.00000162, 0.00000087, -0.00000204, -0.00000010, 0.00000208, 0.00000600, -0.00000687, 0.00000027, 0.00000424, -0.00000582, 0.00000333, -0.00000382, 0.00000516, -0.00000207, -0.00001139, 0.00000343, 0.00000133, 0.00000634, 0.00000096, -0.00000542, 0.00000511, -0.00000014, 0.00000106, -0.00000435, -0.00000626, 0.00001111, 0.00000051, -0.00000829, -0.00001164, 0.00001621, -0.00000442, 0.00000620, -0.00000135, -0.00000224, 0.00001061, -0.00001640, 0.00000788, -0.00000866, 0.00003226, -0.00003822, 0.00002807, -0.00002603, -0.00000355, 0.00003912, -0.00005746, 0.00005510, -0.00005981, 0.00005000, -0.00004016, 0.00001866, -0.00001888, -0.00000079, 0.00002049, -0.00006967, 0.00010184, -0.00015815, 0.00016422, -0.00020642, 0.00012884, -0.00019068, -0.00045534, 0.00121824, -0.00007384, -0.00043268, 0.00020565, -0.00020921, 0.00011061, -0.00008819, 0.00003606, -0.00000240, -0.00004032, 0.00004789, -0.00005206, 0.00002796, -0.00002257, -0.00000077, -0.00000069, -0.00001291, 0.00000785, -0.00001480, 0.00001226, -0.00000399, -0.00002707, 0.00002894, -0.00001405, -0.00000341, 0.00001719, -0.00000887, 0.00000840, -0.00000624, 0.00000330, -0.00000069, 0.00000226, 0.00000936, -0.00000472, -0.00000490, -0.00000328, 0.00000205, 0.00000612, -0.00000181, -0.00000439, -0.00000846, 0.00000429, 0.00000738, -0.00000369, 0.00000691, 0.00000163, 0.00000068, -0.00000094, -0.00000932, -0.00000136, 0.00000231, -0.00000212, 0.00000211, 0.00000355, },
		{ 0.00000431, -0.00000252, -0.00000281, -0.00000354, 0.00000417, 0.00000660, -0.00000322, -0.00000147, -0.00000511, 0.00000029, 0.00000368, 0.00000011, 0.00000401, -0.00000516, -0.00000288, 0.00000141, -0.00000205, -0.00000008, 0.00000111, 0.00000626, -0.00000503, -0.00000181, 0.00000550, -0.00000614, 0.00000274, -0.00000283, 0.00000319, 0.00000106, -0.00001232, 0.00000182, 0.00000265, 0.00000455, 0.00000327, -0.00000626, 0.00000323, 0.00000184, 0.00000073, -0.00000244, -0.00000890, 0.00001077, 0.00000194, -0.00000685, -0.00001360, 0.00001477, -0.00000196, 0.00000359, 0.00000230, -0.00000646, 0.00001467, -0.00001891, 0.00000972, -0.00001379, 0.00003698, -0.00003587, 0.00001717, -0.00000723, -0.00002593, 0.00005736, -0.00006639, 0.00005641, -0.00005360, 0.00003716, -0.00002107, -0.00000313, 0.00000559, -0.00002773, 0.00004709, -0.00009092, 0.00011226, -0.00015124, 0.00013528, -0.00014753, 0.00004182, -0.00005070, -0.00058147, 0.00109969, 0.00017266, -0.00053392, 0.00021026, -0.00018964, 0.00007694, -0.00005446, 0.00000402, 0.00002402, -0.00005706, 0.00005342, -0.00004915, 0.00002048, -0.00001321, -0.00000987, 0.00000592, -0.00001711, 0.00000919, -0.00001283, 0.00000548, 0.00000505, -0.00003195, 0.00002598, -0.00000674, -0.00001092, 0.00002005, -0.00000763, 0.00000715, -0.00000484, 0.00000165, 0.00000054, 0.00000087, 0.00000955, -0.00000316, -0.00000486, -0.00000359, 0.00000057, 0.00000659, -0.00000053, -0.00000443, -0.00000835, 0.00000197, 0.00000863, -0.00000302, 0.00000535, 0.00000297, 0.00000012, 0.00000024, -0.00000858, -0.00000352, 0.00000274, -0.00000180, 0.00000141, },
		{ 0.00000055, 0.00000504, -0.00000234, -0.00000258, -0.00000360, 0.00000223, 0.00000794, -0.00000243, -0.00000183, -0.00000488, -0.00000095, 0.00000465, -0.00000024, 0.00000393, -0.00000395, -0.00000411, 0.00000169, -0.00000189, -0.00000013, 0.00000043, 0.00000592, -0.00000290, -0.00000358, 0.00000615, -0.00000578, 0.00000155, -0.00000144, 0.00000114, 0.00000350, -0.00001203, -0.00000064, 0.00000426, 0.00000290, 0.00000482, -0.00000584, 0.00000049, 0.00000377, 0.00000050, -0.00000046, -0.00001097, 0.00000963, 0.00000354, -0.00000550, -0.00001477, 0.00001237, 0.00000112, 0.00000043, 0.00000636, -0.00001073, 0.00001829, -0.00002068, 0.00001084, -0.00001730, 0.00003774, -0.00002784, 0.00000107, 0.00001382, -0.00004620, 0.00006968, -0.00006754, 0.00004974, -0.00004001, 0.00001836, 0.00000174, -0.00002586, 0.00002812, -0.00004925, 0.00006441, -0.00009946, 0.00010759, -0.00012865, 0.00009262, -0.00008106, -0.00004183, 0.00006313, -0.00064732, 0.00092709, 0.00041347, -0.00059318, 0.00018100, -0.00015049, 0.00003388, -0.00001881, -0.00002521, 0.00004444, -0.00006629, 0.00005193, -0.00004109, 0.00001003, -0.00000289, -0.00001812, 0.00001046, -0.00001867, 0.00000768, -0.00000833, -0.00000234, 0.00001262, -0.00003355, 0.00002008, 0.00000166, -0.00001743, 0.00002089, -0.00000509, 0.00000549, -0.00000332, 0.00000007, 0.00000153, -0.00000014, 0.00000934, -0.00000163, -0.00000462, -0.00000378, -0.00000084, 0.00000657, 0.00000095, -0.00000441, -0.00000815, -0.00000011, 0.00000913, -0.00000180, 0.00000375, 0.00000400, -0.00000008, 0.00000090, -0.00000729, -0.00000553, 0.00000259, -0.00000110, },
		{ -0.00000013, -0.00000033, 0.00000558, -0.00000191, -0.00000265, -0.00000321, 0.00000018, 0.00000875, -0.00000115, -0.00000233, -0.00000448, -0.00000230, 0.00000536, -0.00000013, 0.00000341, -0.00000255, -0.00000514, 0.00000162, -0.00000154, -0.00000028, 0.00000007, 0.00000511, -0.00000077, -0.00000480, 0.00000612, -0.00000481, -0.00000003, 0.00000003, -0.00000060, 0.00000498, -0.00001057, -0.00000361, 0.00000574, 0.00000181, 0.00000534, -0.00000423, -0.00000275, 0.00000529, 0.00000055, 0.00000139, -0.00001229, 0.00000770, 0.00000526, -0.00000433, -0.00001506, 0.00000909, 0.00000468, -0.00000306, 0.00001052, -0.00001461, 0.00002097, -0.00002115, 0.00001052, -0.00001830, 0.00003400, -0.00001467, -0.00001827, 0.00003414, -0.00006138, 0.00007402, -0.00006023, 0.00003567, -0.00002069, -0.00000387, 0.00002510, -0.00004611, 0.00004541, -0.00006258, 0.00007075, -0.00009522, 0.00008983, -0.00009469, 0.00004293, -0.00001567, -0.00011256, 0.00014361, -0.00065608, 0.00071766, 0.00062948, -0.00060291, 0.00012062, -0.00009795, -0.00001227, 0.00001362, -0.00004808, 0.00005691, -0.00006764, 0.00004430, -0.00002946, -0.00000162, 0.00000674, -0.00002423, 0.00001219, -0.00001743, 0.00000375, -0.00000228, -0.00000985, 0.00001764, -0.00003168, 0.00001192, 0.00001009, -0.00002213, 0.00001957, -0.00000153, 0.00000374, -0.00000188, -0.00000128, 0.00000222, -0.00000080, 0.00000886, -0.00000024, -0.00000419, -0.00000383, -0.00000211, 0.00000609, 0.00000249, -0.00000419, -0.00000796, -0.00000184, 0.00000896, -0.00000028, 0.00000234, 0.00000460, 0.00000010, 0.00000111, -0.00000569, -0.00000712, 0.00000180, },
		{ 0.00000041, 0.00000094, -0.00000104, 0.00000580, -0.00000116, -0.00000298, -0.00000253, -0.00000172, 0.00000887, 0.00000055, -0.00000286, -0.00000400, -0.00000360, 0.00000567, 0.00000048, 0.00000256, -0.00000115, -0.00000579, 0.00000117, -0.00000103, -0.00000051, -0.00000001, 0.00000403, 0.00000108, -0.00000531, 0.00000545, -0.00000340, -0.00000172, 0.00000126, -0.00000175, 0.00000541, -0.00000816, -0.00000665, 0.00000665, 0.00000159, 0.00000474, -0.00000166, -0.00000601, 0.00000608, 0.00000101, 0.00000298, -0.00001274, 0.00000503, 0.00000702, -0.00000341, -0.00001445, 0.00000507, 0.00000848, -0.00000660, 0.00001438, -0.00001759, 0.00002216, -0.00001980, 0.00000822, -0.00001629, 0.00002589, 0.00000225, -0.00003821, 0.00005066, -0.00006906, 0.00006941, -0.00004499, 0.00001590, 0.00000182, -0.00002643, 0.00004577, -0.00006089, 0.00005503, -0.00006624, 0.00006605, -0.00007987, 0.00006252, -0.00005474, -0.00000689, 0.00004109, -0.00016367, 0.00018824, -0.00061593, 0.00049046, 0.00080455, -0.00056072, 0.00003582, -0.00004000, -0.00005511, 0.00003854, -0.00006230, 0.00006085, -0.00006190, 0.00003220, -0.00001623, -0.00001268, 0.00001430, -0.00002738, 0.00001095, -0.00001378, -0.00000175, 0.00000411, -0.00001583, 0.00001946, -0.00002666, 0.00000253, 0.00001744, -0.00002444, 0.00001619, 0.00000257, 0.00000227, -0.00000072, -0.00000234, 0.00000259, -0.00000117, 0.00000822, 0.00000094, -0.00000360, -0.00000375, -0.00000321, 0.00000524, 0.00000392, -0.00000370, -0.00000784, -0.00000320, 0.00000827, 0.00000130, 0.00000130, 0.00000473, 0.00000059, 0.00000104, -0.00000403, -0.00000811, },
		{ -0.00000841, -0.00000143, 0.00000186, -0.00000144, 0.00000566, -0.00000012, -0.00000349, -0.00000173, -0.00000325, 0.00000824, 0.00000256, -0.00000328, -0.00000355, -0.00000472, 0.00000547, 0.00000153, 0.00000157, 0.00000003, -0.00000596, 0.00000037, -0.00000043, -0.00000076, 0.00000007, 0.00000289, 0.00000246, -0.00000509, 0.00000427, -0.00000182, -0.00000322, 0.00000200, -0.00000217, 0.00000488, -0.00000521, -0.00000927, 0.00000662, 0.00000239, 0.00000313, 0.00000146, -0.00000880, 0.00000585, 0.00000193, 0.00000421, -0.00001228, 0.00000176, 0.00000873, -0.00000275, -0.00001303, 0.00000057, 0.00001220, -0.00000982, 0.00001745, -0.00001914, 0.00002140, -0.00001631, 0.00000371, -0.00001131, 0.00001429, 0.00002085, -0.00005589, 0.00006071, -0.00006786, 0.00005613, -0.00002355, -0.00000697, 0.00002448, -0.00004620, 0.00006090, -0.00006802, 0.00005579, -0.00006024, 0.00005180, -0.00005651, 0.00003021, -0.00001444, -0.00005071, 0.00008367, -0.00019178, 0.00019886, -0.00053859, 0.00026433, 0.00092710, -0.00046966, -0.00006359, 0.00001477, -0.00008901, 0.00005305, -0.00006710, 0.00005700, -0.00005078, 0.00001779, -0.00000341, -0.00002162, 0.00001888, -0.00002731, 0.00000709, -0.00000852, -0.00000778, 0.00000969, -0.00001939, 0.00001796, -0.00001925, -0.00000689, 0.00002282, -0.00002414, 0.00001118, 0.00000664, 0.00000140, 0.00000007, -0.00000310, 0.00000270, -0.00000136, 0.00000752, 0.00000189, -0.00000290, -0.00000351, -0.00000414, 0.00000412, 0.00000511, -0.00000290, -0.00000778, -0.00000424, 0.00000724, 0.00000269, 0.00000073, 0.00000444, 0.00000126, 0.00000084, -0.00000252, },
		{ -0.00000131, -0.00000805, -0.00000349, 0.00000242, -0.00000143, 0.00000515, 0.00000111, -0.00000402, -0.00000103, -0.00000426, 0.00000693, 0.00000465, -0.00000344, -0.00000324, -0.00000554, 0.00000474, 0.00000288, 0.00000067, 0.00000081, -0.00000561, -0.00000069, 0.00000012, -0.00000095, 0.00000021, 0.00000188, 0.00000324, -0.00000422, 0.00000285, -0.00000035, -0.00000427, 0.00000209, -0.00000186, 0.00000364, -0.00000219, -0.00001102, 0.00000543, 0.00000418, 0.00000084, 0.00000458, -0.00001062, 0.00000445, 0.00000327, 0.00000507, -0.00001093, -0.00000193, 0.00001026, -0.00000234, -0.00001093, -0.00000408, 0.00001546, -0.00001229, 0.00001925, -0.00001882, 0.00001841, -0.00001061, -0.00000286, -0.00000396, 0.00000065, 0.00003871, -0.00006862, 0.00006248, -0.00005760, 0.00003573, 0.00000145, -0.00002983, 0.00004421, -0.00006047, 0.00006849, -0.00006651, 0.00004783, -0.00004599, 0.00003074, -0.00002908, -0.00000233, 0.00002111, -0.00008386, 0.00010899, -0.00019687, 0.00018087, -0.00043765, 0.00005595, 0.00099096, -0.00033763, -0.00016593, 0.00005838, -0.00011004, 0.00005609, -0.00006325, 0.00004714, -0.00003661, 0.00000332, 0.00000729, -0.00002745, 0.00002015, -0.00002434, 0.00000137, -0.00000265, -0.00001327, 0.00001353, -0.00002008, 0.00001354, -0.00001058, -0.00001515, 0.00002557, -0.00002135, 0.00000516, 0.00001010, 0.00000135, 0.00000050, -0.00000361, 0.00000263, -0.00000144, 0.00000683, 0.00000260, -0.00000217, -0.00000309, -0.00000488, 0.00000282, 0.00000597, -0.00000180, -0.00000775, -0.00000505, 0.00000607, 0.00000375, 0.00000062, 0.00000388, 0.00000194, 0.00000069, },
		{ 0.00000068, -0.00000045, -0.00000713, -0.00000550, 0.00000245, -0.00000100, 0.00000437, 0.00000239, -0.00000442, -0.00000058, -0.00000467, 0.00000506, 0.00000660, -0.00000318, -0.00000313, -0.00000600, 0.00000354, 0.00000432, 0.00000008, 0.00000107, -0.00000480, -0.00000182, 0.00000050, -0.00000102, 0.00000031, 0.00000112, 0.00000340, -0.00000290, 0.00000147, 0.00000075, -0.00000473, 0.00000153, -0.00000099, 0.00000206, 0.00000044, -0.00001157, 0.00000307, 0.00000669, -0.00000168, 0.00000717, -0.00001111, 0.00000190, 0.00000487, 0.00000559, -0.00000881, -0.00000581, 0.00001148, -0.00000211, -0.00000839, -0.00000848, 0.00001785, -0.00001360, 0.00001939, -0.00001636, 0.00001318, -0.00000300, -0.00001087, 0.00000470, -0.00001317, 0.00005339, -0.00007431, 0.00005526, -0.00003943, 0.00001081, 0.00002675, -0.00004955, 0.00005837, -0.00006732, 0.00006764, -0.00005660, 0.00003258, -0.00002600, 0.00000638, -0.00000175, -0.00003076, 0.00004803, -0.00010368, 0.00011658, -0.00018179, 0.00014205, -0.00032686, -0.00012160, 0.00099570, -0.00017645, -0.00025917, 0.00008457, -0.00011640, 0.00004844, -0.00005280, 0.00003380, -0.00002183, -0.00000924, 0.00001469, -0.00002979, 0.00001839, -0.00001921, -0.00000516, 0.00000279, -0.00001735, 0.00001508, -0.00001797, 0.00000702, -0.00000189, -0.00002128, 0.00002547, -0.00001654, -0.00000110, 0.00001244, 0.00000219, 0.00000067, -0.00000398, 0.00000248, -0.00000150, 0.00000620, 0.00000313, -0.00000150, -0.00000249, -0.00000545, 0.00000143, 0.00000645, -0.00000049, -0.00000765, -0.00000574, 0.00000492, 0.00000441, 0.00000088, 0.00000320, 0.00000250, },
		{ 0.00000284, 0.00000085, 0.00000007, -0.00000583, -0.00000722, 0.00000185, -0.00000023, 0.00000348, 0.00000355, -0.00000453, -0.00000047, -0.00000450, 0.00000285, 0.00000816, -0.00000242, -0.00000324, -0.00000612, 0.00000200, 0.00000562, -0.00000003, 0.00000081, -0.00000368, -0.00000284, 0.00000057, -0.00000090, 0.00000030, 0.00000068, 0.00000303, -0.00000139, 0.00000041, 0.00000131, -0.00000457, 0.00000044, 0.00000015, 0.00000056, 0.00000225, -0.00001079, -0.00000026, 0.00000949, -0.00000389, 0.00000873, -0.00001008, -0.00000164, 0.00000650, 0.00000588, -0.00000608, -0.00000961, 0.00001223, -0.00000194, -0.00000571, -0.00001222, 0.00001901, -0.00001344, 0.00001762, -0.00001176, 0.00000603, 0.00000587, -0.00001938, 0.00001326, -0.00002519, 0.00006283, -0.00007179, 0.00003961, -0.00001566, -0.00001535, 0.00004902, -0.00006341, 0.00006511, -0.00006596, 0.00005867, -0.00003976, 0.00001244, -0.00000346, -0.00001759, 0.00002171, -0.00005179, 0.00006405, -0.00010956, 0.00010832, -0.00015152, 0.00009124, -0.00021854, -0.00025975, 0.00094620, -0.00000034, -0.00033239, 0.00008970, -0.00010870, 0.00003251, -0.00003865, 0.00001967, -0.00000863, -0.00001852, 0.00001831, -0.00002889, 0.00001432, -0.00001292, -0.00001146, 0.00000691, -0.00001944, 0.00001422, -0.00001354, -0.00000053, 0.00000565, -0.00002465, 0.00002268, -0.00001045, -0.00000682, 0.00001331, 0.00000383, 0.00000078, -0.00000434, 0.00000231, -0.00000158, 0.00000561, 0.00000356, -0.00000096, -0.00000171, -0.00000582, 0.00000002, 0.00000653, 0.00000094, -0.00000738, -0.00000639, 0.00000388, 0.00000467, 0.00000136, 0.00000257, },
		{ 0.00000212, 0.00000290, 0.00000119, 0.00000033, -0.00000435, -0.00000844, 0.00000062, 0.00000071, 0.00000266, 0.00000444, -0.00000427, -0.00000073, -0.00000388, 0.00000057, 0.00000911, -0.00000114, -0.00000351, -0.00000597, 0.00000032, 0.00000655, 0.00000044, 0.00000010, -0.00000247, -0.00000354, 0.00000027, -0.00000060, 0.00000019, 0.00000053, 0.00000232, 0.00000001, -0.00000010, 0.00000126, -0.00000387, -0.00000093, 0.00000122, -0.00000048, 0.00000299, -0.00000875, -0.00000417, 0.00001206, -0.00000527, 0.00000891, -0.00000753, -0.00000581, 0.00000786, 0.00000606, -0.00000293, -0.00001305, 0.00001235, -0.00000170, -0.00000322, -0.00001490, 0.00001863, -0.00001162, 0.00001393, -0.00000529, -0.00000238, 0.00001504, -0.00002718, 0.00002024, -0.00003370, 0.00006566, -0.00006102, 0.00001732, 0.00001055, -0.00003922, 0.00006535, -0.00006954, 0.00006371, -0.00005674, 0.00004302, -0.00001833, -0.00000957, 0.00001829, -0.00003782, 0.00003845, -0.00006350, 0.00006866, -0.00010287, 0.00008782, -0.00011215, 0.00003711, -0.00012252, -0.00035481, 0.00085169, 0.00017576, -0.00037715, 0.00007316, -0.00008971, 0.00001183, -0.00002398, 0.00000718, 0.00000143, -0.00002388, 0.00001837, -0.00002552, 0.00000897, -0.00000657, -0.00001659, 0.00000914, -0.00001939, 0.00001123, -0.00000759, -0.00000795, 0.00001110, -0.00002506, 0.00001776, -0.00000396, -0.00001130, 0.00001259, 0.00000600, 0.00000106, -0.00000479, 0.00000218, -0.00000165, 0.00000502, 0.00000395, -0.00000060, -0.00000081, -0.00000597, -0.00000137, 0.00000626, 0.00000236, -0.00000686, -0.00000707, 0.00000300, 0.00000462, 0.00000190, },
		{ 0.00000233, 0.00000193, 0.00000273, 0.00000162, 0.00000044, -0.00000289, -0.00000903, -0.00000111, 0.00000161, 0.00000207, 0.00000497, -0.00000363, -0.00000129, -0.00000299, -0.00000150, 0.00000931, 0.00000059, -0.00000382, -0.00000568, -0.00000133, 0.00000694, 0.00000150, -0.00000086, -0.00000141, -0.00000376, -0.00000042, -0.00000018, 0.00000000, 0.00000059, 0.00000147, 0.00000106, 0.00000003, 0.00000070, -0.00000288, -0.00000225, 0.00000188, -0.00000079, 0.00000262, -0.00000574, -0.00000812, 0.00001385, -0.00000541, 0.00000758, -0.00000373, -0.00001014, 0.00000863, 0.00000623, 0.00000042, -0.00001587, 0.00001173, -0.00000124, -0.00000123, -0.00001621, 0.00001659, -0.00000816, 0.00000856, 0.00000246, -0.00001116, 0.00002338, -0.00003303, 0.00002431, -0.00003744, 0.00006143, -0.00004313, -0.00000887, 0.00003565, -0.00005760, 0.00007364, -0.00006721, 0.00005461, -0.00004112, 0.00002300, 0.00000479, -0.00003035, 0.00003629, -0.00005181, 0.00004682, -0.00006553, 0.00006299, -0.00008643, 0.00005975, -0.00006997, -0.00001296, -0.00004533, -0.00040765, 0.00072433, 0.00033779, -0.00038835, 0.00003744, -0.00006387, -0.00000959, -0.00001170, -0.00000192, 0.00000759, -0.00002546, 0.00001565, -0.00002078, 0.00000342, -0.00000108, -0.00001995, 0.00000929, -0.00001742, 0.00000673, -0.00000112, -0.00001419, 0.00001388, -0.00002274, 0.00001153, 0.00000205, -0.00001410, 0.00001040, 0.00000832, 0.00000176, -0.00000538, 0.00000205, -0.00000169, 0.00000436, 0.00000438, -0.00000044, 0.00000016, -0.00000586, -0.00000270, 0.00000569, 0.00000365, -0.00000606, -0.00000777, 0.00000225, 0.00000440, },
		{ 0.00000413, 0.00000254, 0.00000200, 0.00000240, 0.00000204, 0.00000052, -0.00000159, -0.00000898, -0.00000317, 0.00000222, 0.00000186, 0.00000509, -0.00000266, -0.00000201, -0.00000204, -0.00000316, 0.00000873, 0.00000263, -0.00000402, -0.00000538, -0.00000274, 0.00000670, 0.00000302, -0.00000184, -0.00000073, -0.00000341, -0.00000141, 0.00000025, -0.00000017, 0.00000074, 0.00000069, 0.00000157, 0.00000080, -0.00000019, -0.00000187, -0.00000323, 0.00000188, -0.00000025, 0.00000129, -0.00000223, -0.00001153, 0.00001438, -0.00000409, 0.00000484, 0.00000090, -0.00001409, 0.00000854, 0.00000649, 0.00000376, -0.00001780, 0.00001024, -0.00000040, -0.00000001, -0.00001596, 0.00001292, -0.00000332, 0.00000203, 0.00001065, -0.00001922, 0.00002971, -0.00003586, 0.00002454, -0.00003589, 0.00005069, -0.00002022, -0.00003562, 0.00005619, -0.00006806, 0.00007298, -0.00005689, 0.00003934, -0.00002140, 0.00000141, 0.00002662, -0.00004713, 0.00004829, -0.00005814, 0.00004655, -0.00005890, 0.00004943, -0.00006397, 0.00002901, -0.00003053, -0.00005355, 0.00000990, -0.00042290, 0.00057768, 0.00047409, -0.00036479, -0.00001234, -0.00003652, -0.00002798, -0.00000394, -0.00000681, 0.00000995, -0.00002404, 0.00001133, -0.00001581, -0.00000138, 0.00000291, -0.00002130, 0.00000755, -0.00001405, 0.00000149, 0.00000495, -0.00001848, 0.00001385, -0.00001831, 0.00000495, 0.00000681, -0.00001499, 0.00000706, 0.00001036, 0.00000305, -0.00000609, 0.00000185, -0.00000161, 0.00000359, 0.00000488, -0.00000044, 0.00000109, -0.00000547, -0.00000395, 0.00000487, 0.00000471, -0.00000495, -0.00000846, 0.00000156, },
		{ 0.00000085, 0.00000394, 0.00000250, 0.00000224, 0.00000205, 0.00000235, 0.00000064, -0.00000053, -0.00000835, -0.00000529, 0.00000236, 0.00000208, 0.00000486, -0.00000150, -0.00000272, -0.00000123, -0.00000425, 0.00000743, 0.00000473, -0.00000394, -0.00000519, -0.00000381, 0.00000582, 0.00000481, -0.00000256, -0.00000063, -0.00000251, -0.00000255, 0.00000054, -0.00000023, 0.00000085, 0.00000014, 0.00000149, 0.00000206, -0.00000110, -0.00000117, -0.00000361, 0.00000110, 0.00000104, -0.00000067, 0.00000121, -0.00001383, 0.00001335, -0.00000131, 0.00000101, 0.00000577, -0.00001713, 0.00000739, 0.00000687, 0.00000688, -0.00001865, 0.00000783, 0.00000089, 0.00000024, -0.00001415, 0.00000790, 0.00000243, -0.00000496, 0.00001832, -0.00002548, 0.00003307, -0.00003492, 0.00002055, -0.00002932, 0.00003488, 0.00000489, -0.00005948, 0.00006934, -0.00006929, 0.00006370, -0.00004019, 0.00002023, -0.00000030, -0.00001892, 0.00004455, -0.00005790, 0.00005312, -0.00005663, 0.00003862, -0.00004574, 0.00003112, -0.00003946, 0.00000001, 0.00000199, -0.00008159, 0.00004323, -0.00040791, 0.00042510, 0.00057639, -0.00030910, -0.00006914, -0.00001301, -0.00004046, -0.00000171, -0.00000765, 0.00000936, -0.00002083, 0.00000665, -0.00001163, -0.00000480, 0.00000515, -0.00002077, 0.00000441, -0.00000999, -0.00000366, 0.00000981, -0.00002039, 0.00001132, -0.00001264, -0.00000105, 0.00000984, -0.00001409, 0.00000310, 0.00001168, 0.00000498, -0.00000680, 0.00000149, -0.00000134, 0.00000266, 0.00000545, -0.00000054, 0.00000188, -0.00000477, -0.00000511, 0.00000389, 0.00000548, -0.00000359, -0.00000907, },
		{ -0.00000949, 0.00000003, 0.00000390, 0.00000223, 0.00000254, 0.00000179, 0.00000249, 0.00000083, 0.00000027, -0.00000727, -0.00000724, 0.00000187, 0.00000267, 0.00000443, -0.00000032, -0.00000324, -0.00000073, -0.00000471, 0.00000558, 0.00000665, -0.00000343, -0.00000518, -0.00000448, 0.00000442, 0.00000659, -0.00000277, -0.00000116, -0.00000119, -0.00000362, 0.00000052, -0.00000010, 0.00000084, -0.00000008, 0.00000087, 0.00000353, -0.00000171, -0.00000103, -0.00000326, -0.00000040, 0.00000283, -0.00000279, 0.00000400, -0.00001462, 0.00001068, 0.00000267, -0.00000338, 0.00001024, -0.00001881, 0.00000513, 0.00000735, 0.00000961, -0.00001827, 0.00000453, 0.00000268, -0.00000050, -0.00001097, 0.00000198, 0.00000842, -0.00001154, 0.00002448, -0.00002899, 0.00003280, -0.00003001, 0.00001265, -0.00001877, 0.00001617, 0.00002911, -0.00007741, 0.00007331, -0.00006132, 0.00004738, -0.00001955, 0.00000006, 0.00001942, -0.00003550, 0.00005671, -0.00006161, 0.00005070, -0.00004822, 0.00002498, -0.00002886, 0.00001142, -0.00001647, -0.00002385, 0.00002509, -0.00009632, 0.00005726, -0.00037138, 0.00027837, 0.00064041, -0.00022722, -0.00012509, 0.00000215, -0.00004545, -0.00000483, -0.00000548, 0.00000713, -0.00001719, 0.00000273, -0.00000890, -0.00000661, 0.00000575, -0.00001881, 0.00000053, -0.00000596, -0.00000804, 0.00001296, -0.00001991, 0.00000689, -0.00000672, -0.00000571, 0.00001093, -0.00001175, -0.00000090, 0.00001197, 0.00000751, -0.00000732, 0.00000084, -0.00000085, 0.00000158, 0.00000603, -0.00000062, 0.00000243, -0.00000377, -0.00000613, 0.00000279, 0.00000592, -0.00000205, },
		{ -0.00000044, -0.00000964, -0.00000095, 0.00000402, 0.00000181, 0.00000276, 0.00000170, 0.00000245, 0.00000108, 0.00000086, -0.00000592, -0.00000877, 0.00000073, 0.00000351, 0.00000397, 0.00000070, -0.00000344, -0.00000059, -0.00000461, 0.00000343, 0.00000813, -0.00000240, -0.00000537, -0.00000478, 0.00000266, 0.00000809, -0.00000229, -0.00000231, 0.00000032, -0.00000436, 0.00000009, 0.00000026, 0.00000067, 0.00000003, -0.00000014, 0.00000492, -0.00000173, -0.00000158, -0.00000221, -0.00000240, 0.00000474, -0.00000455, 0.00000568, -0.00001369, 0.00000655, 0.00000738, -0.00000770, 0.00001369, -0.00001881, 0.00000181, 0.00000787, 0.00001182, -0.00001660, 0.00000043, 0.00000488, -0.00000213, -0.00000679, -0.00000418, 0.00001391, -0.00001686, 0.00002826, -0.00002915, 0.00002871, -0.00002148, 0.00000175, -0.00000589, -0.00000293, 0.00004949, -0.00008716, 0.00006759, -0.00004549, 0.00002653, 0.00000209, -0.00001841, 0.00003541, -0.00004658, 0.00006212, -0.00005825, 0.00004198, -0.00003474, 0.00000817, -0.00001119, -0.00000665, 0.00000223, -0.00004053, 0.00003806, -0.00009895, 0.00005628, -0.00032221, 0.00014674, 0.00066586, -0.00012749, -0.00017256, 0.00000598, -0.00004296, -0.00001202, -0.00000190, 0.00000475, -0.00001430, 0.00000030, -0.00000786, -0.00000698, 0.00000515, -0.00001604, -0.00000337, -0.00000258, -0.00001119, 0.00001422, -0.00001735, 0.00000142, -0.00000148, -0.00000853, 0.00001022, -0.00000853, -0.00000436, 0.00001103, 0.00001042, -0.00000743, -0.00000020, -0.00000011, 0.00000041, 0.00000653, -0.00000060, 0.00000268, -0.00000251, -0.00000696, 0.00000160, 0.00000605, },
		{ 0.00000588, 0.00000114, -0.00000945, -0.00000211, 0.00000425, 0.00000138, 0.00000279, 0.00000183, 0.00000226, 0.00000133, 0.00000129, -0.00000446, -0.00000974, -0.00000100, 0.00000439, 0.00000367, 0.00000142, -0.00000325, -0.00000081, -0.00000409, 0.00000125, 0.00000896, -0.00000085, -0.00000567, -0.00000481, 0.00000079, 0.00000905, -0.00000104, -0.00000390, 0.00000174, -0.00000456, -0.00000082, 0.00000080, 0.00000039, 0.00000038, -0.00000130, 0.00000591, -0.00000097, -0.00000284, -0.00000066, -0.00000455, 0.00000630, -0.00000549, 0.00000593, -0.00001111, 0.00000139, 0.00001220, -0.00001128, 0.00001562, -0.00001702, -0.00000233, 0.00000829, 0.00001340, -0.00001366, -0.00000425, 0.00000732, -0.00000440, -0.00000215, -0.00000988, 0.00001813, -0.00002018, 0.00002907, -0.00002574, 0.00002114, -0.00001023, -0.00001071, 0.00000732, -0.00001990, 0.00006368, -0.00008765, 0.00005301, -0.00002420, 0.00000423, 0.00002182, -0.00003284, 0.00004603, -0.00005132, 0.00006084, -0.00004879, 0.00002870, -0.00001848, -0.00000915, 0.00000459, -0.00002080, 0.00001499, -0.00004940, 0.00004171, -0.00009207, 0.00004532, -0.00026841, 0.00003629, 0.00065604, -0.00001951, -0.00020516, -0.00000250, -0.00003446, -0.00002121, 0.00000125, 0.00000353, -0.00001294, -0.00000038, -0.00000829, -0.00000639, 0.00000395, -0.00001312, -0.00000671, -0.00000027, -0.00001293, 0.00001373, -0.00001330, -0.00000419, 0.00000237, -0.00000939, 0.00000810, -0.00000507, -0.00000682, 0.00000888, 0.00001338, -0.00000689, -0.00000167, 0.00000079, -0.00000074, 0.00000685, -0.00000037, 0.00000259, -0.00000107, -0.00000755, 0.00000034, },
		{ -0.00000095, 0.00000548, 0.00000256, -0.00000888, -0.00000339, 0.00000452, 0.00000106, 0.00000253, 0.00000214, 0.00000203, 0.00000149, 0.00000166, -0.00000305, -0.00001007, -0.00000316, 0.00000506, 0.00000370, 0.00000176, -0.00000270, -0.00000130, -0.00000336, -0.00000068, 0.00000901, 0.00000112, -0.00000597, -0.00000471, -0.00000098, 0.00000929, 0.00000092, -0.00000567, 0.00000276, -0.00000406, -0.00000216, 0.00000143, 0.00000007, 0.00000083, -0.00000236, 0.00000624, 0.00000062, -0.00000461, 0.00000107, -0.00000642, 0.00000711, -0.00000530, 0.00000468, -0.00000717, -0.00000420, 0.00001644, -0.00001355, 0.00001573, -0.00001354, -0.00000695, 0.00000848, 0.00001431, -0.00000959, -0.00000923, 0.00000973, -0.00000694, 0.00000232, -0.00001438, 0.00002044, -0.00002099, 0.00002670, -0.00001905, 0.00001094, 0.00000238, -0.00002301, 0.00001881, -0.00003260, 0.00007022, -0.00007904, 0.00003160, -0.00000060, -0.00001640, 0.00003717, -0.00004167, 0.00005058, -0.00004981, 0.00005380, -0.00003489, 0.00001305, -0.00000188, -0.00002459, 0.00001650, -0.00002968, 0.00002136, -0.00005107, 0.00003799, -0.00007903, 0.00002935, -0.00021632, -0.00005009, 0.00061698, 0.00008708, -0.00021858, -0.00002218, -0.00002261, -0.00003000, 0.00000234, 0.00000433, -0.00001340, 0.00000047, -0.00000962, -0.00000551, 0.00000278, -0.00001062, -0.00000908, 0.00000081, -0.00001334, 0.00001186, -0.00000849, -0.00000913, 0.00000438, -0.00000846, 0.00000519, -0.00000200, -0.00000798, 0.00000575, 0.00001600, -0.00000550, -0.00000353, 0.00000176, -0.00000175, 0.00000689, 0.00000013, 0.00000219, 0.00000045, -0.00000783, },
		{ -0.00000775, -0.00000226, 0.00000487, 0.00000377, -0.00000795, -0.00000474, 0.00000471, 0.00000097, 0.00000197, 0.00000254, 0.00000187, 0.00000150, 0.00000202, -0.00000178, -0.00000979, -0.00000552, 0.00000529, 0.00000417, 0.00000174, -0.00000187, -0.00000188, -0.00000262, -0.00000216, 0.00000828, 0.00000333, -0.00000607, -0.00000463, -0.00000243, 0.00000872, 0.00000340, -0.00000730, 0.00000312, -0.00000281, -0.00000379, 0.00000196, -0.00000014, 0.00000122, -0.00000309, 0.00000578, 0.00000296, -0.00000660, 0.00000256, -0.00000761, 0.00000683, -0.00000384, 0.00000211, -0.00000241, -0.00000950, 0.00001945, -0.00001410, 0.00001394, -0.00000870, -0.00001164, 0.00000828, 0.00001453, -0.00000461, -0.00001410, 0.00001178, -0.00000928, 0.00000598, -0.00001707, 0.00002042, -0.00001911, 0.00002136, -0.00000981, -0.00000064, 0.00001477, -0.00003340, 0.00002684, -0.00003961, 0.00006874, -0.00006270, 0.00000627, 0.00002197, -0.00003259, 0.00004647, -0.00004427, 0.00004929, -0.00004301, 0.00004258, -0.00001862, -0.00000273, 0.00001291, -0.00003636, 0.00002338, -0.00003298, 0.00002190, -0.00004709, 0.00002944, -0.00006321, 0.00001255, -0.00017026, -0.00011247, 0.00055643, 0.00018370, -0.00021104, -0.00005006, -0.00001073, -0.00003611, 0.00000026, 0.00000740, -0.00001537, 0.00000222, -0.00001111, -0.00000501, 0.00000215, -0.00000888, -0.00001036, 0.00000079, -0.00001273, 0.00000913, -0.00000362, -0.00001276, 0.00000446, -0.00000622, 0.00000216, 0.00000019, -0.00000780, 0.00000200, 0.00001787, -0.00000318, -0.00000564, 0.00000261, -0.00000248, 0.00000655, 0.00000092, 0.00000155, 0.00000191, },
		{ 0.00000318, -0.00000727, -0.00000356, 0.00000410, 0.00000469, -0.00000671, -0.00000604, 0.00000471, 0.00000118, 0.00000118, 0.00000289, 0.00000190, 0.00000131, 0.00000237, -0.00000067, -0.00000900, -0.00000781, 0.00000488, 0.00000508, 0.00000146, -0.00000093, -0.00000237, -0.00000209, -0.00000306, 0.00000686, 0.00000554, -0.00000580, -0.00000470, -0.00000345, 0.00000739, 0.00000614, -0.00000843, 0.00000262, -0.00000089, -0.00000548, 0.00000220, -0.00000011, 0.00000141, -0.00000335, 0.00000453, 0.00000576, -0.00000839, 0.00000342, -0.00000781, 0.00000534, -0.00000123, -0.00000137, 0.00000249, -0.00001382, 0.00002074, -0.00001273, 0.00001044, -0.00000297, -0.00001591, 0.00000758, 0.00001409, 0.00000097, -0.00001844, 0.00001307, -0.00001097, 0.00000826, -0.00001753, 0.00001795, -0.00001474, 0.00001366, 0.00000085, -0.00001212, 0.00002532, -0.00004038, 0.00003020, -0.00004036, 0.00005999, -0.00004098, -0.00001963, 0.00004041, -0.00004241, 0.00004905, -0.00004101, 0.00004320, -0.00003249, 0.00002913, -0.00000208, -0.00001664, 0.00002429, -0.00004345, 0.00002501, -0.00003129, 0.00001795, -0.00003949, 0.00001871, -0.00004756, -0.00000207, -0.00013246, -0.00015321, 0.00048268, 0.00026374, -0.00018340, -0.00008177, -0.00000218, -0.00003792, -0.00000534, 0.00001231, -0.00001812, 0.00000401, -0.00001197, -0.00000543, 0.00000234, -0.00000799, -0.00001065, -0.00000004, -0.00001153, 0.00000608, 0.00000070, -0.00001475, 0.00000281, -0.00000331, -0.00000036, 0.00000122, -0.00000645, -0.00000188, 0.00001864, 0.00000003, -0.00000779, 0.00000315, -0.00000280, 0.00000580, 0.00000195, 0.00000078, },
		{ 0.00000003, 0.00000414, -0.00000641, -0.00000479, 0.00000320, 0.00000531, -0.00000522, -0.00000719, 0.00000445, 0.00000170, 0.00000026, 0.00000305, 0.00000218, 0.00000094, 0.00000269, 0.00000029, -0.00000785, -0.00000977, 0.00000374, 0.00000634, 0.00000108, -0.00000006, -0.00000260, -0.00000189, -0.00000336, 0.00000496, 0.00000745, -0.00000503, -0.00000500, -0.00000400, 0.00000544, 0.00000879, -0.00000878, 0.00000121, 0.00000149, -0.00000692, 0.00000195, 0.00000024, 0.00000131, -0.00000308, 0.00000264, 0.00000867, -0.00000955, 0.00000331, -0.00000685, 0.00000272, 0.00000221, -0.00000517, 0.00000682, -0.00001655, 0.00002004, -0.00000954, 0.00000563, 0.00000304, -0.00001934, 0.00000630, 0.00001305, 0.00000675, -0.00002178, 0.00001327, -0.00001161, 0.00000876, -0.00001564, 0.00001325, -0.00000844, 0.00000458, 0.00001160, -0.00002201, 0.00003265, -0.00004292, 0.00002841, -0.00003523, 0.00004564, -0.00001684, -0.00004282, 0.00005236, -0.00004499, 0.00004532, -0.00003311, 0.00003394, -0.00002008, 0.00001539, 0.00001288, -0.00002723, 0.00003138, -0.00004566, 0.00002195, -0.00002583, 0.00001126, -0.00003040, 0.00000808, -0.00003420, -0.00001286, -0.00010332, -0.00017617, 0.00040363, 0.00032310, -0.00013887, -0.00011227, 0.00000025, -0.00003468, -0.00001399, 0.00001811, -0.00002063, 0.00000495, -0.00001160, -0.00000701, 0.00000334, -0.00000779, -0.00001027, -0.00000126, -0.00001018, 0.00000317, 0.00000406, -0.00001505, -0.00000008, -0.00000043, -0.00000189, 0.00000104, -0.00000428, -0.00000539, 0.00001810, 0.00000395, -0.00000967, 0.00000320, -0.00000264, 0.00000468, 0.00000311, },
		{ 0.00000428, -0.00000056, 0.00000471, -0.00000519, -0.00000588, 0.00000219, 0.00000563, -0.00000360, -0.00000809, 0.00000387, 0.00000247, -0.00000064, 0.00000287, 0.00000275, 0.00000048, 0.00000289, 0.00000116, -0.00000651, -0.00001120, 0.00000188, 0.00000775, 0.00000082, 0.00000054, -0.00000246, -0.00000205, -0.00000317, 0.00000285, 0.00000882, -0.00000367, -0.00000550, -0.00000413, 0.00000313, 0.00001100, -0.00000816, -0.00000100, 0.00000402, -0.00000781, 0.00000108, 0.00000095, 0.00000092, -0.00000236, 0.00000036, 0.00001125, -0.00000971, 0.00000205, -0.00000478, -0.00000076, 0.00000595, -0.00000864, 0.00000993, -0.00001732, 0.00001734, -0.00000486, 0.00000009, 0.00000871, -0.00002154, 0.00000441, 0.00001154, 0.00001230, -0.00002368, 0.00001211, -0.00001093, 0.00000731, -0.00001156, 0.00000688, -0.00000105, -0.00000473, 0.00002106, -0.00002899, 0.00003583, -0.00004066, 0.00002179, -0.00002546, 0.00002804, 0.00000663, -0.00006054, 0.00005649, -0.00004058, 0.00003661, -0.00002236, 0.00002339, -0.00000760, 0.00000302, 0.00002495, -0.00003370, 0.00003403, -0.00004351, 0.00001535, -0.00001822, 0.00000362, -0.00002165, -0.00000077, -0.00002425, -0.00001939, -0.00008186, -0.00018590, 0.00032586, 0.00036031, -0.00008244, -0.00013658, -0.00000515, -0.00002678, -0.00002447, 0.00002349, -0.00002187, 0.00000436, -0.00000969, -0.00000970, 0.00000490, -0.00000795, -0.00000961, -0.00000246, -0.00000905, 0.00000074, 0.00000627, -0.00001386, -0.00000361, 0.00000180, -0.00000221, -0.00000015, -0.00000178, -0.00000806, 0.00001618, 0.00000827, -0.00001099, 0.00000260, -0.00000201, 0.00000328, },
		{ 0.00000174, 0.00000529, -0.00000086, 0.00000485, -0.00000372, -0.00000676, 0.00000109, 0.00000566, -0.00000194, -0.00000866, 0.00000299, 0.00000339, -0.00000133, 0.00000228, 0.00000354, 0.00000004, 0.00000288, 0.00000201, -0.00000513, -0.00001199, -0.00000058, 0.00000906, 0.00000089, 0.00000076, -0.00000191, -0.00000253, -0.00000265, 0.00000082, 0.00000945, -0.00000175, -0.00000611, -0.00000397, 0.00000071, 0.00001247, -0.00000650, -0.00000379, 0.00000630, -0.00000787, -0.00000045, 0.00000193, 0.00000033, -0.00000135, -0.00000196, 0.00001309, -0.00000861, -0.00000035, -0.00000185, -0.00000462, 0.00000942, -0.00001116, 0.00001136, -0.00001599, 0.00001292, 0.00000076, -0.00000552, 0.00001344, -0.00002229, 0.00000197, 0.00000967, 0.00001719, -0.00002381, 0.00000944, -0.00000885, 0.00000400, -0.00000579, -0.00000035, 0.00000638, -0.00001305, 0.00002803, -0.00003217, 0.00003449, -0.00003391, 0.00001139, -0.00001287, 0.00000982, 0.00002660, -0.00007093, 0.00005265, -0.00003052, 0.00002488, -0.00001079, 0.00001333, 0.00000345, -0.00000681, 0.00003341, -0.00003594, 0.00003275, -0.00003803, 0.00000665, -0.00001004, -0.00000341, -0.00001454, -0.00000697, -0.00001793, -0.00002224, -0.00006631, -0.00018683, 0.00025423, 0.00037637, -0.00002014, -0.00015058, -0.00001871, -0.00001559, -0.00003508, 0.00002709, -0.00002098, 0.00000188, -0.00000632, -0.00001314, 0.00000655, -0.00000803, -0.00000908, -0.00000328, -0.00000836, -0.00000109, 0.00000739, -0.00001158, -0.00000712, 0.00000292, -0.00000131, -0.00000198, 0.00000055, -0.00000959, 0.00001303, 0.00001257, -0.00001143, 0.00000132, -0.00000098, },
		{ 0.00000027, 0.00000025, 0.00000598, -0.00000080, 0.00000460, -0.00000211, -0.00000734, -0.00000008, 0.00000543, -0.00000033, -0.00000886, 0.00000186, 0.00000434, -0.00000169, 0.00000126, 0.00000441, -0.00000022, 0.00000257, 0.00000287, -0.00000381, -0.00001211, -0.00000342, 0.00000999, 0.00000144, 0.00000056, -0.00000102, -0.00000318, -0.00000203, -0.00000086, 0.00000922, 0.00000060, -0.00000667, -0.00000368, -0.00000154, 0.00001301, -0.00000389, -0.00000679, 0.00000797, -0.00000695, -0.00000256, 0.00000302, -0.00000028, -0.00000029, -0.00000398, 0.00001387, -0.00000618, -0.00000366, 0.00000150, -0.00000831, 0.00001200, -0.00001224, 0.00001087, -0.00001276, 0.00000730, 0.00000665, -0.00001051, 0.00001677, -0.00002148, -0.00000091, 0.00000761, 0.00002100, -0.00002197, 0.00000533, -0.00000549, -0.00000079, 0.00000092, -0.00000746, 0.00001280, -0.00001932, 0.00003163, -0.00003118, 0.00002892, -0.00002364, -0.00000123, 0.00000039, -0.00000651, 0.00004097, -0.00007325, 0.00004186, -0.00001689, 0.00001238, -0.00000029, 0.00000514, 0.00001211, -0.00001355, 0.00003819, -0.00003441, 0.00002848, -0.00003052, -0.00000267, -0.00000269, -0.00000878, -0.00000974, -0.00001038, -0.00001470, -0.00002256, -0.00005462, -0.00018273, 0.00019161, 0.00037424, 0.00004188, -0.00015155, -0.00003928, -0.00000328, -0.00004398, 0.00002777, -0.00001756, -0.00000242, -0.00000191, -0.00001675, 0.00000778, -0.00000764, -0.00000899, -0.00000356, -0.00000817, -0.00000235, 0.00000760, -0.00000868, -0.00001006, 0.00000271, 0.00000055, -0.00000395, 0.00000228, -0.00000983, 0.00000894, 0.00001640, -0.00001081, -0.00000060, },
		{ -0.00000301, 0.00000155, -0.00000100, 0.00000626, -0.00000035, 0.00000402, -0.00000050, -0.00000758, -0.00000129, 0.00000496, 0.00000116, -0.00000871, 0.00000057, 0.00000517, -0.00000162, -0.00000011, 0.00000520, -0.00000015, 0.00000193, 0.00000375, -0.00000260, -0.00001164, -0.00000638, 0.00001030, 0.00000256, -0.00000002, 0.00000009, -0.00000382, -0.00000154, -0.00000200, 0.00000818, 0.00000317, -0.00000700, -0.00000346, -0.00000339, 0.00001253, -0.00000059, -0.00000959, 0.00000868, -0.00000500, -0.00000505, 0.00000397, -0.00000073, 0.00000059, -0.00000539, 0.00001342, -0.00000257, -0.00000747, 0.00000472, -0.00001124, 0.00001323, -0.00001159, 0.00000852, -0.00000804, 0.00000115, 0.00001208, -0.00001432, 0.00001839, -0.00001918, -0.00000407, 0.00000551, 0.00002341, -0.00001812, 0.00000002, -0.00000122, -0.00000648, 0.00000766, -0.00001345, 0.00001729, -0.00002277, 0.00003148, -0.00002622, 0.00001999, -0.00001129, -0.00001420, 0.00001220, -0.00001895, 0.00004855, -0.00006793, 0.00002607, -0.00000219, 0.00000127, 0.00000766, -0.00000034, 0.00001803, -0.00001722, 0.00003973, -0.00002997, 0.00002239, -0.00002228, -0.00001138, 0.00000291, -0.00001196, -0.00000730, -0.00001144, -0.00001359, -0.00002171, -0.00004498, -0.00017637, 0.00013900, 0.00035817, 0.00009809, -0.00013856, -0.00006441, 0.00000757, -0.00004954, 0.00002487, -0.00001167, -0.00000804, 0.00000286, -0.00001984, 0.00000810, -0.00000650, -0.00000946, -0.00000328, -0.00000837, -0.00000321, 0.00000720, -0.00000565, -0.00001201, 0.00000119, 0.00000293, -0.00000557, 0.00000310, -0.00000885, 0.00000436, 0.00001932, -0.00000900, },
		{ -0.00000608, -0.00000563, 0.00000260, -0.00000184, 0.00000605, 0.00000044, 0.00000324, 0.00000097, -0.00000744, -0.00000249, 0.00000427, 0.00000247, -0.00000821, -0.00000079, 0.00000575, -0.00000110, -0.00000169, 0.00000570, 0.00000037, 0.00000098, 0.00000458, -0.00000147, -0.00001070, -0.00000916, 0.00000980, 0.00000420, -0.00000082, 0.00000121, -0.00000423, -0.00000137, -0.00000251, 0.00000646, 0.00000567, -0.00000688, -0.00000342, -0.00000469, 0.00001110, 0.00000308, -0.00001177, 0.00000824, -0.00000216, -0.00000757, 0.00000451, -0.00000079, 0.00000106, -0.00000600, 0.00001175, 0.00000189, -0.00001125, 0.00000725, -0.00001292, 0.00001280, -0.00000922, 0.00000465, -0.00000250, -0.00000476, 0.00001643, -0.00001653, 0.00001820, -0.00001559, -0.00000732, 0.00000354, 0.00002421, -0.00001246, -0.00000604, 0.00000344, -0.00001230, 0.00001349, -0.00001743, 0.00001919, -0.00002304, 0.00002770, -0.00001806, 0.00000904, 0.00000141, -0.00002566, 0.00002083, -0.00002622, 0.00004922, -0.00005642, 0.00000782, 0.00001111, -0.00000680, 0.00001227, -0.00000296, 0.00002141, -0.00001832, 0.00003879, -0.00002366, 0.00001565, -0.00001439, -0.00001862, 0.00000629, -0.00001298, -0.00000679, -0.00001096, -0.00001348, -0.00002088, -0.00003607, -0.00016936, 0.00009593, 0.00033293, 0.00014425, -0.00011251, -0.00009069, 0.00001448, -0.00005064, 0.00001836, -0.00000389, -0.00001418, 0.00000717, -0.00002179, 0.00000717, -0.00000454, -0.00001046, -0.00000259, -0.00000873, -0.00000390, 0.00000650, -0.00000288, -0.00001279, -0.00000137, 0.00000532, -0.00000644, 0.00000290, -0.00000691, -0.00000021, 0.00002099, },
		{ 0.00002121, -0.00000225, -0.00000815, 0.00000319, -0.00000217, 0.00000537, 0.00000148, 0.00000242, 0.00000217, -0.00000693, -0.00000360, 0.00000337, 0.00000357, -0.00000744, -0.00000210, 0.00000603, -0.00000016, -0.00000327, 0.00000570, 0.00000138, -0.00000020, 0.00000525, -0.00000037, -0.00000946, -0.00001150, 0.00000843, 0.00000623, -0.00000163, 0.00000212, -0.00000426, -0.00000161, -0.00000241, 0.00000430, 0.00000779, -0.00000617, -0.00000366, -0.00000542, 0.00000890, 0.00000670, -0.00001299, 0.00000659, 0.00000128, -0.00000976, 0.00000437, -0.00000035, 0.00000103, -0.00000577, 0.00000905, 0.00000670, -0.00001439, 0.00000859, -0.00001304, 0.00001066, -0.00000539, -0.00000018, 0.00000311, -0.00000973, 0.00001918, -0.00001693, 0.00001631, -0.00001104, -0.00001044, 0.00000181, 0.00002334, -0.00000537, -0.00001223, 0.00000786, -0.00001749, 0.00001756, -0.00001880, 0.00001823, -0.00002025, 0.00002093, -0.00000788, -0.00000237, 0.00001279, -0.00003408, 0.00002512, -0.00002793, 0.00004383, -0.00004087, -0.00001021, 0.00002094, -0.00001089, 0.00001345, -0.00000315, 0.00002290, -0.00001758, 0.00003627, -0.00001648, 0.00000921, -0.00000760, -0.00002396, 0.00000747, -0.00001226, -0.00000751, -0.00000985, -0.00001337, -0.00002091, -0.00002723, -0.00016231, 0.00006089, 0.00030311, 0.00017779, -0.00007591, -0.00011435, 0.00001548, -0.00004697, 0.00000885, 0.00000479, -0.00001987, 0.00001026, -0.00002217, 0.00000493, -0.00000188, -0.00001177, -0.00000177, -0.00000899, -0.00000465, 0.00000574, -0.00000063, -0.00001240, -0.00000455, 0.00000719, -0.00000631, 0.00000179, -0.00000439, -0.00000430, },
		{ -0.00000749, 0.00001997, 0.00000215, -0.00001020, 0.00000315, -0.00000198, 0.00000430, 0.00000262, 0.00000169, 0.00000304, -0.00000611, -0.00000456, 0.00000229, 0.00000443, -0.00000644, -0.00000329, 0.00000596, 0.00000109, -0.00000464, 0.00000510, 0.00000282, -0.00000143, 0.00000563, 0.00000077, -0.00000808, -0.00001323, 0.00000623, 0.00000839, -0.00000220, 0.00000263, -0.00000379, -0.00000228, -0.00000184, 0.00000202, 0.00000926, -0.00000480, -0.00000414, -0.00000563, 0.00000621, 0.00000987, -0.00001302, 0.00000388, 0.00000491, -0.00001121, 0.00000337, 0.00000066, 0.00000049, -0.00000481, 0.00000570, 0.00001130, -0.00001636, 0.00000838, -0.00001153, 0.00000703, -0.00000058, -0.00000526, 0.00000805, -0.00001316, 0.00002008, -0.00001555, 0.00001300, -0.00000592, -0.00001325, 0.00000041, 0.00002090, 0.00000258, -0.00001787, 0.00001139, -0.00002132, 0.00001924, -0.00001728, 0.00001457, -0.00001494, 0.00001220, 0.00000287, -0.00001265, 0.00002140, -0.00003845, 0.00002470, -0.00002459, 0.00003402, -0.00002384, -0.00002556, 0.00002598, -0.00001086, 0.00001181, -0.00000177, 0.00002333, -0.00001584, 0.00003299, -0.00000923, 0.00000372, -0.00000226, -0.00002741, 0.00000684, -0.00001046, -0.00000866, -0.00000889, -0.00001259, -0.00002213, -0.00001845, -0.00015508, 0.00003188, 0.00027254, 0.00019794, -0.00003254, -0.00013175, 0.00000949, -0.00003904, -0.00000248, 0.00001315, -0.00002416, 0.00001155, -0.00002080, 0.00000153, 0.00000117, -0.00001309, -0.00000109, -0.00000892, -0.00000559, 0.00000506, 0.00000100, -0.00001102, -0.00000788, 0.00000815, -0.00000513, 0.00000003, -0.00000173, },
		{ 0.00000064, -0.00000954, 0.00001741, 0.00000668, -0.00001148, 0.00000240, -0.00000131, 0.00000298, 0.00000370, 0.00000119, 0.00000353, -0.00000505, -0.00000528, 0.00000104, 0.00000502, -0.00000528, -0.00000429, 0.00000556, 0.00000248, -0.00000560, 0.00000382, 0.00000455, -0.00000250, 0.00000559, 0.00000202, -0.00000670, -0.00001426, 0.00000336, 0.00001040, -0.00000231, 0.00000264, -0.00000283, -0.00000327, -0.00000102, -0.00000008, 0.00000989, -0.00000281, -0.00000477, -0.00000547, 0.00000333, 0.00001224, -0.00001179, 0.00000040, 0.00000826, -0.00001157, 0.00000145, 0.00000213, -0.00000044, -0.00000336, 0.00000213, 0.00001511, -0.00001673, 0.00000653, -0.00000856, 0.00000237, 0.00000454, -0.00000987, 0.00001165, -0.00001470, 0.00001910, -0.00001264, 0.00000868, -0.00000067, -0.00001557, -0.00000061, 0.00001715, 0.00001069, -0.00002223, 0.00001345, -0.00002325, 0.00001820, -0.00001302, 0.00000875, -0.00000802, 0.00000276, 0.00001274, -0.00002045, 0.00002629, -0.00003840, 0.00001992, -0.00001741, 0.00002189, -0.00000779, -0.00003643, 0.00002579, -0.00000739, 0.00000843, 0.00000009, 0.00002357, -0.00001375, 0.00002954, -0.00000241, -0.00000051, 0.00000161, -0.00002925, 0.00000496, -0.00000828, -0.00000958, -0.00000860, -0.00001088, -0.00002438, -0.00001014, -0.00014709, 0.00000695, 0.00024396, 0.00020560, 0.00001322, -0.00014000, -0.00000346, -0.00002818, -0.00001412, 0.00001999, -0.00002634, 0.00001078, -0.00001783, -0.00000262, 0.00000418, -0.00001406, -0.00000083, -0.00000835, -0.00000678, 0.00000452, 0.00000203, -0.00000896, -0.00001086, 0.00000795, -0.00000308, -0.00000199, },
		{ -0.00000385, 0.00000238, -0.00001036, 0.00001386, 0.00001089, -0.00001173, 0.00000094, -0.00000031, 0.00000161, 0.00000457, 0.00000100, 0.00000368, -0.00000388, -0.00000569, -0.00000031, 0.00000532, -0.00000400, -0.00000508, 0.00000490, 0.00000385, -0.00000600, 0.00000193, 0.00000634, -0.00000318, 0.00000503, 0.00000337, -0.00000537, -0.00001459, 0.00000009, 0.00001195, -0.00000178, 0.00000211, -0.00000148, -0.00000440, -0.00000021, -0.00000172, 0.00000959, -0.00000036, -0.00000537, -0.00000513, 0.00000057, 0.00001358, -0.00000941, -0.00000341, 0.00001084, -0.00001065, -0.00000130, 0.00000387, -0.00000151, -0.00000174, -0.00000116, 0.00001765, -0.00001530, 0.00000316, -0.00000453, -0.00000272, 0.00000928, -0.00001332, 0.00001348, -0.00001425, 0.00001646, -0.00000860, 0.00000385, 0.00000430, -0.00001725, -0.00000127, 0.00001248, 0.00001821, -0.00002469, 0.00001364, -0.00002300, 0.00001450, -0.00000655, 0.00000168, -0.00000062, -0.00000607, 0.00002045, -0.00002485, 0.00002711, -0.00003425, 0.00001181, -0.00000807, 0.00000963, 0.00000523, -0.00004184, 0.00002086, -0.00000177, 0.00000463, 0.00000144, 0.00002424, -0.00001173, 0.00002624, 0.00000377, -0.00000350, 0.00000429, -0.00002995, 0.00000244, -0.00000631, -0.00000987, -0.00000917, -0.00000834, -0.00002717, -0.00000300, -0.00013766, -0.00001550, 0.00021883, 0.00020295, 0.00005695, -0.00013733, -0.00002220, -0.00001626, -0.00002451, 0.00002432, -0.00002603, 0.00000806, -0.00001368, -0.00000700, 0.00000670, -0.00001442, -0.00000114, -0.00000727, -0.00000817, 0.00000404, 0.00000259, -0.00000655, -0.00001312, 0.00000655, -0.00000047, },
		{ 0.00000225, -0.00000517, 0.00000331, -0.00001008, 0.00000975, 0.00001435, -0.00001086, -0.00000107, 0.00000082, 0.00000035, 0.00000509, 0.00000113, 0.00000355, -0.00000273, -0.00000577, -0.00000169, 0.00000531, -0.00000264, -0.00000564, 0.00000407, 0.00000505, -0.00000576, -0.00000043, 0.00000792, -0.00000327, 0.00000392, 0.00000476, -0.00000409, -0.00001432, -0.00000329, 0.00001276, -0.00000056, 0.00000114, 0.00000008, -0.00000542, 0.00000034, -0.00000271, 0.00000841, 0.00000233, -0.00000573, -0.00000479, -0.00000182, 0.00001382, -0.00000617, -0.00000707, 0.00001227, -0.00000840, -0.00000461, 0.00000554, -0.00000243, -0.00000032, -0.00000374, 0.00001863, -0.00001213, -0.00000134, -0.00000005, -0.00000756, 0.00001296, -0.00001515, 0.00001335, -0.00001199, 0.00001258, -0.00000396, -0.00000099, 0.00000863, -0.00001820, -0.00000167, 0.00000739, 0.00002441, -0.00002481, 0.00001183, -0.00002062, 0.00000854, 0.00000127, -0.00000557, 0.00000611, -0.00001316, 0.00002510, -0.00002547, 0.00002413, -0.00002692, 0.00000182, 0.00000155, -0.00000084, 0.00001388, -0.00004178, 0.00001241, 0.00000440, 0.00000167, 0.00000152, 0.00002564, -0.00000986, 0.00002310, 0.00000936, -0.00000543, 0.00000613, -0.00002997, -0.00000021, -0.00000492, -0.00000938, -0.00001047, -0.00000540, -0.00002982, 0.00000228, -0.00012630, -0.00003642, 0.00019748, 0.00019300, 0.00009479, -0.00012333, -0.00004459, -0.00000542, -0.00003226, 0.00002559, -0.00002328, 0.00000380, -0.00000896, -0.00001101, 0.00000836, -0.00001399, -0.00000205, -0.00000575, -0.00000959, 0.00000352, 0.00000284, -0.00000413, -0.00001442, 0.00000410, },
		{ 0.00000091, 0.00000463, -0.00000568, 0.00000343, -0.00000893, 0.00000554, 0.00001672, -0.00000889, -0.00000341, 0.00000183, -0.00000061, 0.00000522, 0.00000157, 0.00000326, -0.00000171, -0.00000550, -0.00000303, 0.00000494, -0.00000125, -0.00000599, 0.00000315, 0.00000596, -0.00000489, -0.00000301, 0.00000900, -0.00000265, 0.00000235, 0.00000609, -0.00000280, -0.00001361, -0.00000648, 0.00001268, 0.00000128, -0.00000007, 0.00000161, -0.00000613, 0.00000045, -0.00000298, 0.00000651, 0.00000493, -0.00000565, -0.00000459, -0.00000368, 0.00001302, -0.00000243, -0.00001010, 0.00001230, -0.00000499, -0.00000809, 0.00000679, -0.00000291, 0.00000061, -0.00000533, 0.00001796, -0.00000750, -0.00000639, 0.00000422, -0.00001148, 0.00001510, -0.00001510, 0.00001135, -0.00000830, 0.00000801, 0.00000075, -0.00000541, 0.00001205, -0.00001835, -0.00000193, 0.00000240, 0.00002868, -0.00002238, 0.00000813, -0.00001641, 0.00000102, 0.00000941, -0.00001189, 0.00001118, -0.00001767, 0.00002629, -0.00002251, 0.00001814, -0.00001772, -0.00000842, 0.00000978, -0.00000810, 0.00001767, -0.00003706, 0.00000220, 0.00000950, 0.00000048, 0.00000003, 0.00002766, -0.00000791, 0.00001993, 0.00001454, -0.00000659, 0.00000752, -0.00002971, -0.00000269, -0.00000423, -0.00000826, -0.00001213, -0.00000257, -0.00003166, 0.00000520, -0.00011290, -0.00005613, 0.00017931, 0.00017903, 0.00012392, -0.00009903, -0.00006782, 0.00000231, -0.00003649, 0.00002371, -0.00001855, -0.00000132, -0.00000435, -0.00001414, 0.00000893, -0.00001277, -0.00000346, -0.00000397, -0.00001087, 0.00000284, 0.00000292, -0.00000193, -0.00001467, },
		{ -0.00001395, -0.00000262, 0.00000628, -0.00000530, 0.00000289, -0.00000729, 0.00000167, 0.00001781, -0.00000600, -0.00000575, 0.00000252, -0.00000119, 0.00000495, 0.00000221, 0.00000293, -0.00000091, -0.00000493, -0.00000420, 0.00000422, 0.00000015, -0.00000615, 0.00000221, 0.00000652, -0.00000350, -0.00000553, 0.00000936, -0.00000130, 0.00000046, 0.00000716, -0.00000142, -0.00001263, -0.00000922, 0.00001164, 0.00000356, -0.00000128, 0.00000286, -0.00000632, 0.00000002, -0.00000258, 0.00000418, 0.00000713, -0.00000498, -0.00000462, -0.00000498, 0.00001136, 0.00000134, -0.00001209, 0.00001089, -0.00000075, -0.00001125, 0.00000726, -0.00000273, 0.00000083, -0.00000578, 0.00001580, -0.00000192, -0.00001132, 0.00000762, -0.00001398, 0.00001540, -0.00001321, 0.00000783, -0.00000377, 0.00000336, 0.00000503, -0.00000903, 0.00001438, -0.00001769, -0.00000221, -0.00000197, 0.00003061, -0.00001751, 0.00000298, -0.00001100, -0.00000716, 0.00001677, -0.00001632, 0.00001385, -0.00001922, 0.00002413, -0.00001670, 0.00001037, -0.00000813, -0.00001739, 0.00001532, -0.00001147, 0.00001694, -0.00002917, -0.00000787, 0.00001221, 0.00000150, -0.00000289, 0.00002987, -0.00000548, 0.00001647, 0.00001950, -0.00000720, 0.00000871, -0.00002935, -0.00000492, -0.00000413, -0.00000684, -0.00001370, -0.00000040, -0.00003218, 0.00000559, -0.00009774, -0.00007437, 0.00016314, 0.00016410, 0.00014278, -0.00006664, -0.00008881, 0.00000539, -0.00003690, 0.00001915, -0.00001258, -0.00000652, -0.00000050, -0.00001605, 0.00000837, -0.00001090, -0.00000518, -0.00000216, -0.00001181, 0.00000190, 0.00000294, -0.00000010, },
		{ 0.00000130, -0.00001244, -0.00000606, 0.00000694, -0.00000411, 0.00000198, -0.00000552, -0.00000152, 0.00001762, -0.00000250, -0.00000780, 0.00000271, -0.00000134, 0.00000433, 0.00000295, 0.00000268, -0.00000037, -0.00000415, -0.00000509, 0.00000313, 0.00000149, -0.00000612, 0.00000132, 0.00000672, -0.00000176, -0.00000772, 0.00000884, 0.00000068, -0.00000149, 0.00000779, 0.00000013, -0.00001151, -0.00001136, 0.00000974, 0.00000601, -0.00000220, 0.00000362, -0.00000591, -0.00000093, -0.00000169, 0.00000174, 0.00000865, -0.00000366, -0.00000483, -0.00000577, 0.00000913, 0.00000472, -0.00001281, 0.00000818, 0.00000381, -0.00001359, 0.00000667, -0.00000175, 0.00000030, -0.00000518, 0.00001248, 0.00000396, -0.00001543, 0.00000963, -0.00001475, 0.00001387, -0.00000977, 0.00000332, 0.00000096, -0.00000080, 0.00000849, -0.00001163, 0.00001555, -0.00001625, -0.00000267, -0.00000532, 0.00003003, -0.00001059, -0.00000298, -0.00000511, -0.00001502, 0.00002234, -0.00001818, 0.00001384, -0.00001790, 0.00001922, -0.00000912, 0.00000217, 0.00000045, -0.00002391, 0.00001747, -0.00001100, 0.00001273, -0.00001993, -0.00001607, 0.00001176, 0.00000458, -0.00000665, 0.00003161, -0.00000212, 0.00001250, 0.00002436, -0.00000737, 0.00000981, -0.00002891, -0.00000699, -0.00000433, -0.00000551, -0.00001475, 0.00000074, -0.00003116, 0.00000363, -0.00008150, -0.00009047, 0.00014759, 0.00015062, 0.00015117, -0.00002927, -0.00010469, 0.00000296, -0.00003390, 0.00001277, -0.00000631, -0.00001106, 0.00000216, -0.00001661, 0.00000686, -0.00000865, -0.00000694, -0.00000055, -0.00001229, 0.00000069, 0.00000294, },
		{ 0.00000289, 0.00000231, -0.00001038, -0.00000906, 0.00000650, -0.00000236, 0.00000105, -0.00000395, -0.00000384, 0.00001627, 0.00000122, -0.00000924, 0.00000233, -0.00000111, 0.00000350, 0.00000365, 0.00000259, -0.00000008, -0.00000328, -0.00000562, 0.00000172, 0.00000271, -0.00000591, 0.00000050, 0.00000661, 0.00000015, -0.00000933, 0.00000741, 0.00000307, -0.00000321, 0.00000781, 0.00000189, -0.00001036, -0.00001284, 0.00000715, 0.00000828, -0.00000258, 0.00000378, -0.00000489, -0.00000227, -0.00000055, -0.00000047, 0.00000929, -0.00000174, -0.00000514, -0.00000620, 0.00000665, 0.00000735, -0.00001218, 0.00000454, 0.00000811, -0.00001468, 0.00000489, -0.00000003, -0.00000088, -0.00000376, 0.00000853, 0.00000947, -0.00001813, 0.00000991, -0.00001376, 0.00001075, -0.00000527, -0.00000153, 0.00000527, -0.00000404, 0.00001089, -0.00001312, 0.00001559, -0.00001408, -0.00000345, -0.00000737, 0.00002705, -0.00000227, -0.00000895, 0.00000042, -0.00002167, 0.00002540, -0.00001715, 0.00001126, -0.00001423, 0.00001253, -0.00000104, -0.00000514, 0.00000691, -0.00002732, 0.00001618, -0.00000740, 0.00000652, -0.00001117, -0.00002113, 0.00000805, 0.00000905, -0.00001041, 0.00003217, 0.00000253, 0.00000797, 0.00002903, -0.00000705, 0.00001077, -0.00002827, -0.00000910, -0.00000454, -0.00000463, -0.00001498, 0.00000069, -0.00002867, -0.00000018, -0.00006508, -0.00010364, 0.00013142, 0.00014017, 0.00015013, 0.00000953, -0.00011321, -0.00000493, -0.00002847, 0.00000567, -0.00000063, -0.00001437, 0.00000339, -0.00001595, 0.00000468, -0.00000634, -0.00000852, 0.00000071, -0.00001224, -0.00000074, },
		{ -0.00000228, 0.00000274, 0.00000297, -0.00000802, -0.00001134, 0.00000505, -0.00000040, 0.00000043, -0.00000280, -0.00000525, 0.00001406, 0.00000476, -0.00000988, 0.00000139, -0.00000061, 0.00000259, 0.00000420, 0.00000271, 0.00000001, -0.00000246, -0.00000571, 0.00000005, 0.00000372, -0.00000549, -0.00000027, 0.00000627, 0.00000201, -0.00001019, 0.00000516, 0.00000558, -0.00000441, 0.00000711, 0.00000384, -0.00000919, -0.00001370, 0.00000417, 0.00001006, -0.00000226, 0.00000334, -0.00000341, -0.00000377, 0.00000054, -0.00000216, 0.00000899, 0.00000061, -0.00000535, -0.00000644, 0.00000422, 0.00000900, -0.00001033, 0.00000045, 0.00001158, -0.00001422, 0.00000198, 0.00000224, -0.00000247, -0.00000190, 0.00000452, 0.00001397, -0.00001902, 0.00000840, -0.00001124, 0.00000651, -0.00000034, -0.00000607, 0.00000862, -0.00000605, 0.00001213, -0.00001351, 0.00001458, -0.00001126, -0.00000464, -0.00000800, 0.00002204, 0.00000660, -0.00001408, 0.00000486, -0.00002639, 0.00002553, -0.00001336, 0.00000665, -0.00000905, 0.00000518, 0.00000628, -0.00001051, 0.00001062, -0.00002752, 0.00001199, -0.00000185, -0.00000005, -0.00000443, -0.00002246, 0.00000164, 0.00001384, -0.00001325, 0.00003098, 0.00000858, 0.00000306, 0.00003329, -0.00000607, 0.00001143, -0.00002721, -0.00001151, -0.00000448, -0.00000442, -0.00001431, -0.00000047, -0.00002503, -0.00000512, -0.00004947, -0.00011316, 0.00011378, 0.00013339, 0.00014160, 0.00004630, -0.00011301, -0.00001736, -0.00002200, -0.00000100, 0.00000376, -0.00001618, 0.00000325, -0.00001436, 0.00000225, -0.00000426, -0.00000976, 0.00000153, -0.00001171, },
		{ -0.00001080, -0.00000380, 0.00000243, 0.00000336, -0.00000558, -0.00001277, 0.00000281, 0.00000136, 0.00000037, -0.00000215, -0.00000587, 0.00001134, 0.00000775, -0.00000961, -0.00000000, -0.00000001, 0.00000174, 0.00000452, 0.00000303, -0.00000000, -0.00000179, -0.00000534, -0.00000175, 0.00000444, -0.00000486, -0.00000098, 0.00000580, 0.00000365, -0.00001024, 0.00000231, 0.00000784, -0.00000484, 0.00000570, 0.00000584, -0.00000793, -0.00001407, 0.00000112, 0.00001107, -0.00000119, 0.00000239, -0.00000168, -0.00000517, 0.00000132, -0.00000315, 0.00000779, 0.00000313, -0.00000526, -0.00000665, 0.00000208, 0.00000961, -0.00000759, -0.00000351, 0.00001375, -0.00001212, -0.00000182, 0.00000471, -0.00000411, -0.00000004, 0.00000098, 0.00001699, -0.00001795, 0.00000531, -0.00000761, 0.00000174, 0.00000437, -0.00000973, 0.00001066, -0.00000674, 0.00001229, -0.00001293, 0.00001268, -0.00000791, -0.00000625, -0.00000731, 0.00001559, 0.00001507, -0.00001761, 0.00000766, -0.00002878, 0.00002275, -0.00000733, 0.00000086, -0.00000338, -0.00000169, 0.00001184, -0.00001330, 0.00001144, -0.00002494, 0.00000589, 0.00000428, -0.00000550, -0.00000067, -0.00002016, -0.00000635, 0.00001775, -0.00001438, 0.00002775, 0.00001586, -0.00000182, 0.00003672, -0.00000415, 0.00001161, -0.00002555, -0.00001436, -0.00000402, -0.00000494, -0.00001283, -0.00000244, -0.00002074, -0.00001040, -0.00003555, -0.00011856, 0.00009431, 0.00013004, 0.00012811, 0.00007803, -0.00010383, -0.00003270, -0.00001605, -0.00000632, 0.00000644, -0.00001647, 0.00000202, -0.00001225, -0.00000007, -0.00000262, -0.00001058, 0.00000194, },
		{ 0.00000202, -0.00000966, -0.00000515, 0.00000193, 0.00000351, -0.00000320, -0.00001335, 0.00000011, 0.00000256, 0.00000099, -0.00000191, -0.00000594, 0.00000851, 0.00000993, -0.00000847, -0.00000164, 0.00000052, 0.00000106, 0.00000458, 0.00000350, -0.00000001, -0.00000140, -0.00000455, -0.00000355, 0.00000477, -0.00000400, -0.00000166, 0.00000526, 0.00000497, -0.00000952, -0.00000085, 0.00000952, -0.00000438, 0.00000368, 0.00000771, -0.00000649, -0.00001411, -0.00000173, 0.00001117, 0.00000053, 0.00000117, 0.00000005, -0.00000622, 0.00000158, -0.00000337, 0.00000589, 0.00000552, -0.00000470, -0.00000696, 0.00000034, 0.00000926, -0.00000437, -0.00000681, 0.00001432, -0.00000853, -0.00000606, 0.00000695, -0.00000542, 0.00000142, -0.00000165, 0.00001828, -0.00001505, 0.00000105, -0.00000347, -0.00000290, 0.00000827, -0.00001210, 0.00001125, -0.00000621, 0.00001156, -0.00001158, 0.00001007, -0.00000420, -0.00000819, -0.00000554, 0.00000847, 0.00002225, -0.00001898, 0.00000854, -0.00002876, 0.00001742, 0.00000010, -0.00000513, 0.00000177, -0.00000712, 0.00001497, -0.00001337, 0.00000976, -0.00002045, -0.00000094, 0.00000965, -0.00000874, -0.00000023, -0.00001502, -0.00001450, 0.00001965, -0.00001335, 0.00002256, 0.00002387, -0.00000608, 0.00003890, -0.00000107, 0.00001119, -0.00002320, -0.00001770, -0.00000319, -0.00000604, -0.00001085, -0.00000480, -0.00001637, -0.00001530, -0.00002398, -0.00011974, 0.00007319, 0.00012928, 0.00011231, 0.00010254, -0.00008646, -0.00004881, -0.00001204, -0.00000967, 0.00000738, -0.00001554, 0.00000011, -0.00001007, -0.00000196, -0.00000151, -0.00001103, },
		{ -0.00001126, 0.00000191, -0.00000842, -0.00000625, 0.00000127, 0.00000342, -0.00000096, -0.00001316, -0.00000271, 0.00000293, 0.00000223, -0.00000188, -0.00000575, 0.00000589, 0.00001115, -0.00000662, -0.00000328, 0.00000081, 0.00000062, 0.00000440, 0.00000401, 0.00000010, -0.00000131, -0.00000344, -0.00000517, 0.00000462, -0.00000292, -0.00000230, 0.00000473, 0.00000590, -0.00000817, -0.00000398, 0.00001035, -0.00000303, 0.00000130, 0.00000919, -0.00000475, -0.00001399, -0.00000413, 0.00001033, 0.00000270, -0.00000004, 0.00000148, -0.00000672, 0.00000124, -0.00000290, 0.00000359, 0.00000745, -0.00000354, -0.00000740, -0.00000098, 0.00000817, -0.00000116, -0.00000900, 0.00001320, -0.00000383, -0.00001020, 0.00000851, -0.00000606, 0.00000215, -0.00000309, 0.00001782, -0.00001066, -0.00000376, 0.00000053, -0.00000685, 0.00001094, -0.00001299, 0.00001044, -0.00000468, 0.00001020, -0.00000965, 0.00000695, -0.00000029, -0.00001031, -0.00000310, 0.00000146, 0.00002738, -0.00001790, 0.00000753, -0.00002659, 0.00001023, 0.00000793, -0.00001029, 0.00000553, -0.00001049, 0.00001548, -0.00001105, 0.00000635, -0.00001510, -0.00000731, 0.00001323, -0.00000924, -0.00000276, -0.00000830, -0.00002135, 0.00001876, -0.00001014, 0.00001585, 0.00003188, -0.00000905, 0.00003944, 0.00000328, 0.00001016, -0.00002019, -0.00002141, -0.00000215, -0.00000747, -0.00000872, -0.00000712, -0.00001242, -0.00001926, -0.00001510, -0.00011699, 0.00005103, 0.00012986, 0.00009663, 0.00011865, -0.00006263, -0.00006344, -0.00001103, -0.00001094, 0.00000689, -0.00001383, -0.00000201, -0.00000817, -0.00000327, -0.00000090, },
		{ -0.00000060, -0.00001141, 0.00000173, -0.00000718, -0.00000708, 0.00000052, 0.00000306, 0.00000108, -0.00001235, -0.00000530, 0.00000237, 0.00000388, -0.00000177, -0.00000556, 0.00000371, 0.00001143, -0.00000433, -0.00000466, 0.00000075, 0.00000042, 0.00000407, 0.00000448, 0.00000040, -0.00000153, -0.00000215, -0.00000644, 0.00000396, -0.00000169, -0.00000288, 0.00000421, 0.00000647, -0.00000639, -0.00000675, 0.00001018, -0.00000092, -0.00000113, 0.00001004, -0.00000261, -0.00001382, -0.00000597, 0.00000867, 0.00000499, -0.00000093, 0.00000241, -0.00000655, 0.00000033, -0.00000192, 0.00000121, 0.00000866, -0.00000175, -0.00000791, -0.00000197, 0.00000665, 0.00000159, -0.00000981, 0.00001056, 0.00000144, -0.00001365, 0.00000901, -0.00000581, 0.00000199, -0.00000329, 0.00001585, -0.00000535, -0.00000846, 0.00000380, -0.00000964, 0.00001214, -0.00001241, 0.00000846, -0.00000250, 0.00000851, -0.00000737, 0.00000353, 0.00000359, -0.00001238, -0.00000045, -0.00000468, 0.00002994, -0.00001443, 0.00000495, -0.00002280, 0.00000206, 0.00001511, -0.00001380, 0.00000737, -0.00001158, 0.00001361, -0.00000703, 0.00000218, -0.00000998, -0.00001224, 0.00001443, -0.00000706, -0.00000736, -0.00000147, -0.00002565, 0.00001480, -0.00000521, 0.00000836, 0.00003907, -0.00001013, 0.00003810, 0.00000885, 0.00000872, -0.00001671, -0.00002525, -0.00000118, -0.00000889, -0.00000682, -0.00000900, -0.00000926, -0.00002197, -0.00000894, -0.00011091, 0.00002874, 0.00013042, 0.00008302, 0.00012620, -0.00003467, -0.00007449, -0.00001352, -0.00001046, 0.00000552, -0.00001189, -0.00000393, -0.00000679, -0.00000400, },
		{ -0.00000429, -0.00000040, -0.00001166, 0.00000153, -0.00000597, -0.00000768, -0.00000018, 0.00000238, 0.00000293, -0.00001108, -0.00000743, 0.00000091, 0.00000563, -0.00000130, -0.00000558, 0.00000209, 0.00001091, -0.00000190, -0.00000557, 0.00000032, 0.00000038, 0.00000367, 0.00000479, 0.00000095, -0.00000200, -0.00000085, -0.00000723, 0.00000281, -0.00000037, -0.00000336, 0.00000371, 0.00000675, -0.00000440, -0.00000890, 0.00000898, 0.00000166, -0.00000324, 0.00001005, -0.00000007, -0.00001364, -0.00000725, 0.00000642, 0.00000705, -0.00000124, 0.00000271, -0.00000575, -0.00000102, -0.00000069, -0.00000090, 0.00000897, 0.00000056, -0.00000833, -0.00000277, 0.00000502, 0.00000353, -0.00000919, 0.00000679, 0.00000662, -0.00001585, 0.00000819, -0.00000460, 0.00000093, -0.00000241, 0.00001279, 0.00000022, -0.00001240, 0.00000588, -0.00001105, 0.00001186, -0.00001059, 0.00000569, -0.00000006, 0.00000675, -0.00000493, 0.00000002, 0.00000721, -0.00001412, 0.00000191, -0.00000933, 0.00002974, -0.00000893, 0.00000138, -0.00001808, -0.00000615, 0.00002073, -0.00001510, 0.00000711, -0.00001057, 0.00000996, -0.00000225, -0.00000174, -0.00000598, -0.00001516, 0.00001318, -0.00000280, -0.00001283, 0.00000405, -0.00002664, 0.00000807, 0.00000061, 0.00000101, 0.00004463, -0.00000890, 0.00003489, 0.00001535, 0.00000717, -0.00001305, -0.00002891, -0.00000060, -0.00000999, -0.00000546, -0.00001020, -0.00000707, -0.00002335, -0.00000523, -0.00010234, 0.00000731, 0.00012976, 0.00007273, 0.00012597, -0.00000523, -0.00008038, -0.00001937, -0.00000894, 0.00000394, -0.00001017, -0.00000540, -0.00000598, },
		{ -0.00000561, -0.00000436, -0.00000008, -0.00001209, 0.00000129, -0.00000473, -0.00000817, -0.00000071, 0.00000137, 0.00000454, -0.00000950, -0.00000895, -0.00000126, 0.00000708, -0.00000022, -0.00000588, 0.00000098, 0.00000984, 0.00000034, -0.00000589, -0.00000043, 0.00000040, 0.00000333, 0.00000489, 0.00000172, -0.00000260, 0.00000029, -0.00000744, 0.00000122, 0.00000091, -0.00000366, 0.00000318, 0.00000681, -0.00000241, -0.00001027, 0.00000690, 0.00000434, -0.00000467, 0.00000910, 0.00000280, -0.00001339, -0.00000806, 0.00000392, 0.00000854, -0.00000081, 0.00000237, -0.00000443, -0.00000257, 0.00000048, -0.00000247, 0.00000835, 0.00000320, -0.00000848, -0.00000358, 0.00000356, 0.00000445, -0.00000730, 0.00000245, 0.00001102, -0.00001642, 0.00000603, -0.00000253, -0.00000084, -0.00000077, 0.00000918, 0.00000539, -0.00001506, 0.00000653, -0.00001105, 0.00001031, -0.00000787, 0.00000256, 0.00000230, 0.00000514, -0.00000247, -0.00000340, 0.00001032, -0.00001525, 0.00000356, -0.00001209, 0.00002689, -0.00000205, -0.00000245, -0.00001318, -0.00001353, 0.00002412, -0.00001400, 0.00000494, -0.00000798, 0.00000537, 0.00000233, -0.00000454, -0.00000365, -0.00001592, 0.00000986, 0.00000260, -0.00001784, 0.00000716, -0.00002411, -0.00000064, 0.00000625, -0.00000534, 0.00004795, -0.00000523, 0.00003011, 0.00002231, 0.00000596, -0.00000956, -0.00003205, -0.00000068, -0.00001054, -0.00000481, -0.00001063, -0.00000583, -0.00002353, -0.00000353, -0.00009219, -0.00001233, 0.00012698, 0.00006629, 0.00011942, 0.00002311, -0.00008020, -0.00002784, -0.00000731, 0.00000275, -0.00000897, -0.00000637, },
		{ -0.00000695, -0.00000548, -0.00000447, 0.00000051, -0.00001267, 0.00000090, -0.00000336, -0.00000866, -0.00000097, 0.00000007, 0.00000586, -0.00000768, -0.00000983, -0.00000383, 0.00000789, 0.00000156, -0.00000641, 0.00000023, 0.00000851, 0.00000214, -0.00000557, -0.00000136, 0.00000033, 0.00000314, 0.00000474, 0.00000266, -0.00000320, 0.00000112, -0.00000704, -0.00000066, 0.00000201, -0.00000373, 0.00000259, 0.00000675, -0.00000061, -0.00001081, 0.00000419, 0.00000671, -0.00000516, 0.00000723, 0.00000581, -0.00001294, -0.00000860, 0.00000149, 0.00000918, 0.00000039, 0.00000153, -0.00000283, -0.00000405, 0.00000133, -0.00000333, 0.00000689, 0.00000587, -0.00000813, -0.00000455, 0.00000248, 0.00000433, -0.00000450, -0.00000181, 0.00001409, -0.00001515, 0.00000267, 0.00000009, -0.00000299, 0.00000121, 0.00000560, 0.00000960, -0.00001612, 0.00000571, -0.00000984, 0.00000783, -0.00000471, -0.00000052, 0.00000427, 0.00000383, -0.00000012, -0.00000653, 0.00001271, -0.00001552, 0.00000419, -0.00001283, 0.00002183, 0.00000537, -0.00000580, -0.00000879, -0.00001942, 0.00002497, -0.00001067, 0.00000141, -0.00000456, 0.00000074, 0.00000584, -0.00000570, -0.00000313, -0.00001483, 0.00000526, 0.00000802, -0.00002127, 0.00000724, -0.00001850, -0.00001010, 0.00001060, -0.00001000, 0.00004872, 0.00000070, 0.00002429, 0.00002912, 0.00000555, -0.00000655, -0.00003444, -0.00000158, -0.00001045, -0.00000489, -0.00001036, -0.00000538, -0.00002280, -0.00000329, -0.00008134, -0.00002945, 0.00012160, 0.00006361, 0.00010842, 0.00004813, -0.00007382, -0.00003771, -0.00000655, 0.00000237, -0.00000835, },
		{ -0.00000813, -0.00000739, -0.00000534, -0.00000480, 0.00000140, -0.00001326, 0.00000019, -0.00000176, -0.00000925, -0.00000092, -0.00000143, 0.00000678, -0.00000566, -0.00001016, -0.00000647, 0.00000777, 0.00000401, -0.00000698, -0.00000039, 0.00000720, 0.00000333, -0.00000471, -0.00000230, 0.00000005, 0.00000314, 0.00000438, 0.00000366, -0.00000364, 0.00000151, -0.00000607, -0.00000267, 0.00000279, -0.00000351, 0.00000193, 0.00000664, 0.00000092, -0.00001059, 0.00000120, 0.00000840, -0.00000458, 0.00000461, 0.00000867, -0.00001209, -0.00000907, -0.00000056, 0.00000884, 0.00000227, 0.00000041, -0.00000122, -0.00000520, 0.00000166, -0.00000344, 0.00000481, 0.00000825, -0.00000711, -0.00000575, 0.00000184, 0.00000331, -0.00000129, -0.00000537, 0.00001541, -0.00001212, -0.00000149, 0.00000287, -0.00000513, 0.00000308, 0.00000255, 0.00001245, -0.00001551, 0.00000364, -0.00000779, 0.00000486, -0.00000155, -0.00000319, 0.00000568, 0.00000288, 0.00000207, -0.00000919, 0.00001417, -0.00001477, 0.00000366, -0.00001168, 0.00001527, 0.00001242, -0.00000800, -0.00000541, -0.00002342, 0.00002329, -0.00000563, -0.00000277, -0.00000111, -0.00000311, 0.00000772, -0.00000505, -0.00000419, -0.00001250, 0.00000029, 0.00001241, -0.00002234, 0.00000428, -0.00001076, -0.00001893, 0.00001274, -0.00001260, 0.00004695, 0.00000842, 0.00001814, 0.00003514, 0.00000634, -0.00000422, -0.00003593, -0.00000335, -0.00000975, -0.00000557, -0.00000960, -0.00000548, -0.00002149, -0.00000398, -0.00007053, -0.00004359, 0.00011352, 0.00006408, 0.00009495, 0.00006824, -0.00006186, -0.00004750, -0.00000743, 0.00000294, },
		{ 0.00000426, -0.00000799, -0.00000798, -0.00000500, -0.00000541, 0.00000249, -0.00001364, -0.00000104, 0.00000016, -0.00000995, -0.00000062, -0.00000294, 0.00000718, -0.00000346, -0.00001003, -0.00000886, 0.00000659, 0.00000692, -0.00000731, -0.00000110, 0.00000614, 0.00000387, -0.00000349, -0.00000301, -0.00000055, 0.00000334, 0.00000388, 0.00000460, -0.00000381, 0.00000143, -0.00000466, -0.00000460, 0.00000313, -0.00000301, 0.00000122, 0.00000649, 0.00000216, -0.00000978, -0.00000170, 0.00000913, -0.00000296, 0.00000154, 0.00001103, -0.00001066, -0.00000965, -0.00000203, 0.00000753, 0.00000456, -0.00000066, 0.00000011, -0.00000580, 0.00000139, -0.00000289, 0.00000242, 0.00001001, -0.00000534, -0.00000716, 0.00000158, 0.00000169, 0.00000178, -0.00000772, 0.00001483, -0.00000764, -0.00000593, 0.00000533, -0.00000687, 0.00000443, 0.00000042, 0.00001376, -0.00001336, 0.00000070, -0.00000533, 0.00000185, 0.00000124, -0.00000522, 0.00000643, 0.00000228, 0.00000402, -0.00001125, 0.00001457, -0.00001295, 0.00000201, -0.00000899, 0.00000801, 0.00001826, -0.00000859, -0.00000329, -0.00002544, 0.00001944, 0.00000037, -0.00000676, 0.00000163, -0.00000562, 0.00000773, -0.00000281, -0.00000624, -0.00000975, -0.00000412, 0.00001500, -0.00002077, -0.00000118, -0.00000220, -0.00002582, 0.00001212, -0.00001319, 0.00004300, 0.00001722, 0.00001245, 0.00003979, 0.00000860, -0.00000262, -0.00003656, -0.00000587, -0.00000863, -0.00000665, -0.00000858, -0.00000584, -0.00001991, -0.00000516, -0.00006029, -0.00005458, 0.00010303, 0.00006681, 0.00008086, 0.00008252, -0.00004557, -0.00005573, -0.00001041, },
		{ -0.00001552, 0.00000585, -0.00000748, -0.00000897, -0.00000439, -0.00000625, 0.00000356, -0.00001354, -0.00000291, 0.00000240, -0.00001070, -0.00000021, -0.00000425, 0.00000693, -0.00000109, -0.00000958, -0.00001076, 0.00000437, 0.00000994, -0.00000710, -0.00000209, 0.00000541, 0.00000385, -0.00000216, -0.00000332, -0.00000149, 0.00000365, 0.00000337, 0.00000533, -0.00000362, 0.00000089, -0.00000297, -0.00000624, 0.00000296, -0.00000225, 0.00000049, 0.00000631, 0.00000313, -0.00000860, -0.00000418, 0.00000875, -0.00000051, -0.00000154, 0.00001255, -0.00000852, -0.00001046, -0.00000283, 0.00000541, 0.00000692, -0.00000138, 0.00000093, -0.00000575, 0.00000056, -0.00000187, 0.00000005, 0.00001092, -0.00000282, -0.00000866, 0.00000156, -0.00000015, 0.00000420, -0.00000852, 0.00001248, -0.00000226, -0.00001003, 0.00000703, -0.00000788, 0.00000499, -0.00000059, 0.00001357, -0.00001005, -0.00000262, -0.00000293, -0.00000081, 0.00000340, -0.00000649, 0.00000656, 0.00000198, 0.00000569, -0.00001257, 0.00001385, -0.00001012, -0.00000055, -0.00000531, 0.00000095, 0.00002220, -0.00000737, -0.00000242, -0.00002568, 0.00001400, 0.00000649, -0.00000983, 0.00000312, -0.00000650, 0.00000601, 0.00000050, -0.00000854, -0.00000739, -0.00000724, 0.00001540, -0.00001678, -0.00000818, 0.00000577, -0.00002974, 0.00000866, -0.00001219, 0.00003746, 0.00002628, 0.00000796, 0.00004266, 0.00001237, -0.00000163, -0.00003646, -0.00000893, -0.00000731, -0.00000787, -0.00000757, -0.00000623, -0.00001830, -0.00000648, -0.00005098, -0.00006249, 0.00009063, 0.00007077, 0.00006764, 0.00009081, -0.00002655, -0.00006112, },
		{ -0.00006281, -0.00002234, 0.00000706, -0.00000623, -0.00001043, -0.00000359, -0.00000709, 0.00000432, -0.00001272, -0.00000548, 0.00000481, -0.00001134, 0.00000009, -0.00000510, 0.00000593, 0.00000142, -0.00000889, -0.00001203, 0.00000132, 0.00001267, -0.00000609, -0.00000342, 0.00000501, 0.00000345, -0.00000100, -0.00000308, -0.00000273, 0.00000395, 0.00000298, 0.00000577, -0.00000303, -0.00000001, -0.00000119, -0.00000743, 0.00000226, -0.00000132, -0.00000016, 0.00000604, 0.00000393, -0.00000728, -0.00000598, 0.00000732, 0.00000241, -0.00000418, 0.00001296, -0.00000562, -0.00001149, -0.00000302, 0.00000278, 0.00000895, -0.00000146, 0.00000112, -0.00000508, -0.00000069, -0.00000067, -0.00000196, 0.00001082, 0.00000028, -0.00001004, 0.00000154, -0.00000185, 0.00000558, -0.00000770, 0.00000874, 0.00000335, -0.00001319, 0.00000762, -0.00000798, 0.00000463, -0.00000051, 0.00001215, -0.00000605, -0.00000579, -0.00000100, -0.00000286, 0.00000478, -0.00000702, 0.00000617, 0.00000189, 0.00000705, -0.00001306, 0.00001204, -0.00000651, -0.00000366, -0.00000127, -0.00000514, 0.00002382, -0.00000445, -0.00000253, -0.00002453, 0.00000772, 0.00001191, -0.00001142, 0.00000312, -0.00000583, 0.00000301, 0.00000414, -0.00001032, -0.00000606, -0.00000868, 0.00001365, -0.00001104, -0.00001551, 0.00001188, -0.00003016, 0.00000278, -0.00001032, 0.00003104, 0.00003479, 0.00000525, 0.00004356, 0.00001749, -0.00000094, -0.00003592, -0.00001227, -0.00000606, -0.00000901, -0.00000676, -0.00000649, -0.00001682, -0.00000776, -0.00004274, -0.00006759, 0.00007695, 0.00007500, 0.00005634, 0.00009352, -0.00000658, },
		{ 0.00001270, -0.00006044, -0.00003008, 0.00000721, -0.00000403, -0.00001225, -0.00000286, -0.00000766, 0.00000450, -0.00001105, -0.00000865, 0.00000717, -0.00001161, 0.00000006, -0.00000531, 0.00000419, 0.00000395, -0.00000799, -0.00001266, -0.00000227, 0.00001469, -0.00000413, -0.00000504, 0.00000478, 0.00000293, -0.00000024, -0.00000227, -0.00000412, 0.00000410, 0.00000282, 0.00000584, -0.00000208, -0.00000115, 0.00000047, -0.00000805, 0.00000110, -0.00000037, -0.00000064, 0.00000563, 0.00000464, -0.00000603, -0.00000698, 0.00000502, 0.00000534, -0.00000595, 0.00001209, -0.00000208, -0.00001263, -0.00000278, 0.00000002, 0.00001028, -0.00000074, 0.00000069, -0.00000392, -0.00000213, 0.00000045, -0.00000336, 0.00000972, 0.00000371, -0.00001101, 0.00000127, -0.00000308, 0.00000570, -0.00000542, 0.00000418, 0.00000846, -0.00001492, 0.00000694, -0.00000714, 0.00000341, 0.00000049, 0.00000988, -0.00000191, -0.00000836, 0.00000019, -0.00000417, 0.00000540, -0.00000692, 0.00000541, 0.00000194, 0.00000806, -0.00001270, 0.00000928, -0.00000241, -0.00000688, 0.00000251, -0.00000965, 0.00002304, -0.00000021, -0.00000318, -0.00002251, 0.00000132, 0.00001596, -0.00001124, 0.00000169, -0.00000394, -0.00000063, 0.00000734, -0.00001095, -0.00000615, -0.00000841, 0.00001019, -0.00000449, -0.00002195, 0.00001519, -0.00002708, -0.00000464, -0.00000845, 0.00002445, 0.00004209, 0.00000468, 0.00004258, 0.00002358, -0.00000015, -0.00003521, -0.00001562, -0.00000510, -0.00000990, -0.00000624, -0.00000658, -0.00001550, -0.00000892, -0.00003559, -0.00007025, 0.00006265, 0.00007869, 0.00004757, 0.00009153, },
		{ 0.00008599, 0.00002991, -0.00005421, -0.00003775, 0.00000576, -0.00000092, -0.00001412, -0.00000254, -0.00000766, 0.00000392, -0.00000855, -0.00001216, 0.00000912, -0.00001125, -0.00000053, -0.00000477, 0.00000181, 0.00000634, -0.00000684, -0.00001273, -0.00000601, 0.00001564, -0.00000123, -0.00000676, 0.00000452, 0.00000255, -0.00000005, -0.00000094, -0.00000548, 0.00000391, 0.00000300, 0.00000555, -0.00000085, -0.00000236, 0.00000184, -0.00000806, -0.00000039, 0.00000044, -0.00000086, 0.00000504, 0.00000533, -0.00000496, -0.00000721, 0.00000220, 0.00000779, -0.00000652, 0.00000997, 0.00000188, -0.00001366, -0.00000236, -0.00000248, 0.00001063, 0.00000081, -0.00000024, -0.00000249, -0.00000348, 0.00000121, -0.00000400, 0.00000777, 0.00000710, -0.00001130, 0.00000057, -0.00000362, 0.00000455, -0.00000207, -0.00000051, 0.00001242, -0.00001493, 0.00000502, -0.00000552, 0.00000153, 0.00000210, 0.00000725, 0.00000185, -0.00000999, 0.00000052, -0.00000476, 0.00000538, -0.00000635, 0.00000444, 0.00000206, 0.00000867, -0.00001150, 0.00000581, 0.00000178, -0.00000975, 0.00000550, -0.00001222, 0.00002007, 0.00000475, -0.00000385, -0.00002016, -0.00000455, 0.00001823, -0.00000932, -0.00000083, -0.00000141, -0.00000419, 0.00000943, -0.00001004, -0.00000766, -0.00000676, 0.00000577, 0.00000183, -0.00002648, 0.00001527, -0.00002107, -0.00001248, -0.00000742, 0.00001821, 0.00004771, 0.00000633, 0.00004004, 0.00003015, 0.00000116, -0.00003460, -0.00001879, -0.00000454, -0.00001048, -0.00000604, -0.00000650, -0.00001432, -0.00000992, -0.00002949, -0.00007089, 0.00004833, 0.00008127, 0.00004150, },
		{ 0.00003798, 0.00007809, 0.00004411, -0.00004473, -0.00004431, 0.00000243, 0.00000278, -0.00001557, -0.00000298, -0.00000690, 0.00000253, -0.00000540, -0.00001562, 0.00001029, -0.00001002, -0.00000179, -0.00000350, -0.00000100, 0.00000833, -0.00000537, -0.00001241, -0.00000952, 0.00001531, 0.00000243, -0.00000825, 0.00000397, 0.00000251, -0.00000049, 0.00000072, -0.00000656, 0.00000325, 0.00000352, 0.00000496, 0.00000053, -0.00000348, 0.00000279, -0.00000750, -0.00000202, 0.00000095, -0.00000075, 0.00000426, 0.00000603, -0.00000410, -0.00000681, -0.00000071, 0.00000933, -0.00000573, 0.00000681, 0.00000588, -0.00001429, -0.00000207, -0.00000438, 0.00000985, 0.00000308, -0.00000142, -0.00000108, -0.00000449, 0.00000144, -0.00000390, 0.00000524, 0.00001007, -0.00001071, -0.00000066, -0.00000342, 0.00000233, 0.00000180, -0.00000462, 0.00001473, -0.00001315, 0.00000212, -0.00000342, -0.00000069, 0.00000392, 0.00000472, 0.00000483, -0.00001050, 0.00000004, -0.00000478, 0.00000489, -0.00000552, 0.00000338, 0.00000223, 0.00000888, -0.00000957, 0.00000192, 0.00000566, -0.00001186, 0.00000732, -0.00001278, 0.00001539, 0.00000972, -0.00000402, -0.00001792, -0.00000943, 0.00001859, -0.00000597, -0.00000389, 0.00000113, -0.00000700, 0.00000997, -0.00000756, -0.00001031, -0.00000433, 0.00000123, 0.00000697, -0.00002846, 0.00001224, -0.00001315, -0.00001948, -0.00000785, 0.00001259, 0.00005144, 0.00001004, 0.00003642, 0.00003662, 0.00000340, -0.00003425, -0.00002167, -0.00000444, -0.00001078, -0.00000609, -0.00000633, -0.00001324, -0.00001080, -0.00002435, -0.00006989, 0.00003448, 0.00008238, },
		{ 0.00008187, 0.00003666, 0.00006895, 0.00005481, -0.00003295, -0.00004887, -0.00000271, 0.00000652, -0.00001610, -0.00000445, -0.00000534, 0.00000051, -0.00000197, -0.00001856, 0.00001033, -0.00000781, -0.00000369, -0.00000166, -0.00000391, 0.00000965, -0.00000350, -0.00001184, -0.00001251, 0.00001366, 0.00000650, -0.00000917, 0.00000292, 0.00000293, -0.00000146, 0.00000246, -0.00000712, 0.00000204, 0.00000436, 0.00000417, 0.00000190, -0.00000435, 0.00000327, -0.00000648, -0.00000360, 0.00000104, -0.00000032, 0.00000336, 0.00000668, -0.00000342, -0.00000603, -0.00000328, 0.00000964, -0.00000363, 0.00000298, 0.00000949, -0.00001423, -0.00000215, -0.00000545, 0.00000796, 0.00000579, -0.00000254, 0.00000005, -0.00000498, 0.00000109, -0.00000316, 0.00000246, 0.00001228, -0.00000911, -0.00000240, -0.00000257, -0.00000061, 0.00000557, -0.00000756, 0.00001513, -0.00000979, -0.00000136, -0.00000122, -0.00000291, 0.00000557, 0.00000267, 0.00000678, -0.00000990, -0.00000106, -0.00000446, 0.00000417, -0.00000459, 0.00000235, 0.00000244, 0.00000866, -0.00000706, -0.00000203, 0.00000882, -0.00001292, 0.00000781, -0.00001153, 0.00000968, 0.00001399, -0.00000332, -0.00001607, -0.00001309, 0.00001716, -0.00000171, -0.00000687, 0.00000309, -0.00000860, 0.00000882, -0.00000376, -0.00001353, -0.00000186, -0.00000261, 0.00001026, -0.00002769, 0.00000669, -0.00000456, -0.00002457, -0.00001005, 0.00000761, 0.00005328, 0.00001543, 0.00003235, 0.00004243, 0.00000684, -0.00003422, -0.00002428, -0.00000472, -0.00001088, -0.00000629, -0.00000617, -0.00001220, -0.00001155, -0.00002008, -0.00006762, 0.00002146, },
		{ 0.00000955, 0.00007972, 0.00003710, 0.00005950, 0.00006194, -0.00001998, -0.00005081, -0.00000922, 0.00000958, -0.00001526, -0.00000704, -0.00000315, -0.00000183, 0.00000127, -0.00002053, 0.00000901, -0.00000463, -0.00000606, 0.00000045, -0.00000652, 0.00001007, -0.00000118, -0.00001115, -0.00001479, 0.00001085, 0.00001053, -0.00000917, 0.00000125, 0.00000378, -0.00000278, 0.00000395, -0.00000695, 0.00000027, 0.00000541, 0.00000333, 0.00000313, -0.00000490, 0.00000330, -0.00000516, -0.00000492, 0.00000065, 0.00000038, 0.00000243, 0.00000722, -0.00000280, -0.00000515, -0.00000515, 0.00000861, -0.00000049, -0.00000101, 0.00001225, -0.00001326, -0.00000276, -0.00000559, 0.00000515, 0.00000858, -0.00000326, 0.00000068, -0.00000485, 0.00000022, -0.00000201, -0.00000018, 0.00001345, -0.00000654, -0.00000449, -0.00000131, -0.00000382, 0.00000864, -0.00000891, 0.00001365, -0.00000530, -0.00000487, 0.00000068, -0.00000479, 0.00000674, 0.00000137, 0.00000763, -0.00000838, -0.00000249, -0.00000403, 0.00000338, -0.00000369, 0.00000138, 0.00000271, 0.00000807, -0.00000419, -0.00000569, 0.00001097, -0.00001277, 0.00000702, -0.00000886, 0.00000370, 0.00001696, -0.00000154, -0.00001468, -0.00001552, 0.00001430, 0.00000283, -0.00000918, 0.00000404, -0.00000877, 0.00000613, 0.00000082, -0.00001663, -0.00000004, -0.00000514, 0.00001137, -0.00002445, -0.00000041, 0.00000338, -0.00002695, -0.00001391, 0.00000302, 0.00005341, 0.00002204, 0.00002842, 0.00004715, 0.00001161, -0.00003439, -0.00002674, -0.00000528, -0.00001091, -0.00000654, -0.00000609, -0.00001117, -0.00001217, -0.00001662, -0.00006437, },
		{ -0.00006042, -0.00000106, 0.00007602, 0.00003881, 0.00005044, 0.00006572, -0.00000684, -0.00004995, -0.00001636, 0.00001124, -0.00001281, -0.00001059, -0.00000069, -0.00000403, 0.00000385, -0.00002117, 0.00000626, -0.00000071, -0.00000858, 0.00000246, -0.00000847, 0.00000941, 0.00000156, -0.00001037, -0.00001631, 0.00000718, 0.00001401, -0.00000801, -0.00000100, 0.00000492, -0.00000416, 0.00000491, -0.00000596, -0.00000197, 0.00000650, 0.00000258, 0.00000411, -0.00000508, 0.00000294, -0.00000372, -0.00000584, -0.00000021, 0.00000120, 0.00000162, 0.00000752, -0.00000208, -0.00000444, -0.00000606, 0.00000631, 0.00000324, -0.00000458, 0.00001377, -0.00001126, -0.00000395, -0.00000488, 0.00000174, 0.00001103, -0.00000333, 0.00000070, -0.00000414, -0.00000102, -0.00000075, -0.00000236, 0.00001345, -0.00000317, -0.00000665, 0.00000006, -0.00000682, 0.00001053, -0.00000850, 0.00001056, -0.00000026, -0.00000787, 0.00000193, -0.00000611, 0.00000725, 0.00000091, 0.00000748, -0.00000624, -0.00000394, -0.00000373, 0.00000266, -0.00000287, 0.00000047, 0.00000305, 0.00000715, -0.00000118, -0.00000872, 0.00001191, -0.00001145, 0.00000518, -0.00000535, -0.00000180, 0.00001822, 0.00000126, -0.00001367, -0.00001690, 0.00001051, 0.00000699, -0.00001035, 0.00000375, -0.00000758, 0.00000233, 0.00000550, -0.00001890, 0.00000056, -0.00000602, 0.00001039, -0.00001941, -0.00000790, 0.00000953, -0.00002631, -0.00001893, -0.00000159, 0.00005205, 0.00002934, 0.00002518, 0.00005049, 0.00001767, -0.00003454, -0.00002921, -0.00000596, -0.00001099, -0.00000676, -0.00000610, -0.00001018, -0.00001262, -0.00001389, },
		{ -0.00001182, -0.00005601, -0.00001024, 0.00007094, 0.00004134, 0.00004227, 0.00006660, 0.00000559, -0.00004646, -0.00002321, 0.00001094, -0.00000876, -0.00001466, 0.00000151, -0.00000557, 0.00000536, -0.00002031, 0.00000227, 0.00000356, -0.00001083, 0.00000395, -0.00000944, 0.00000764, 0.00000455, -0.00000946, -0.00001717, 0.00000307, 0.00001648, -0.00000560, -0.00000366, 0.00000607, -0.00000528, 0.00000507, -0.00000412, -0.00000449, 0.00000744, 0.00000207, 0.00000477, -0.00000491, 0.00000231, -0.00000234, -0.00000627, -0.00000143, 0.00000194, 0.00000110, 0.00000747, -0.00000116, -0.00000409, -0.00000596, 0.00000304, 0.00000700, -0.00000719, 0.00001381, -0.00000827, -0.00000561, -0.00000353, -0.00000186, 0.00001274, -0.00000255, 0.00000011, -0.00000300, -0.00000238, 0.00000034, -0.00000382, 0.00001231, 0.00000069, -0.00000855, 0.00000122, -0.00000920, 0.00001096, -0.00000642, 0.00000640, 0.00000465, -0.00000988, 0.00000233, -0.00000675, 0.00000702, 0.00000124, 0.00000655, -0.00000383, -0.00000510, -0.00000368, 0.00000206, -0.00000215, -0.00000039, 0.00000346, 0.00000603, 0.00000169, -0.00001088, 0.00001159, -0.00000915, 0.00000267, -0.00000159, -0.00000622, 0.00001764, 0.00000483, -0.00001278, -0.00001752, 0.00000633, 0.00001025, -0.00001014, 0.00000229, -0.00000533, -0.00000198, 0.00000957, -0.00001981, -0.00000034, -0.00000530, 0.00000776, -0.00001345, -0.00001466, 0.00001313, -0.00002280, -0.00002432, -0.00000664, 0.00004941, 0.00003685, 0.00002307, 0.00005231, 0.00002482, -0.00003436, -0.00003190, -0.00000666, -0.00001118, -0.00000692, -0.00000621, -0.00000927, -0.00001287, },
		{ -0.00001289, -0.00001032, -0.00005140, -0.00001790, 0.00006464, 0.00004427, 0.00003530, 0.00006511, 0.00001673, -0.00004087, -0.00002886, 0.00000838, -0.00000351, -0.00001863, 0.00000285, -0.00000603, 0.00000554, -0.00001798, -0.00000260, 0.00000767, -0.00001236, 0.00000459, -0.00000926, 0.00000487, 0.00000754, -0.00000826, -0.00001754, -0.00000104, 0.00001760, -0.00000204, -0.00000638, 0.00000690, -0.00000583, 0.00000430, -0.00000156, -0.00000705, 0.00000802, 0.00000191, 0.00000508, -0.00000444, 0.00000152, -0.00000113, -0.00000624, -0.00000283, 0.00000240, 0.00000101, 0.00000700, 0.00000005, -0.00000419, -0.00000494, -0.00000075, 0.00001017, -0.00000840, 0.00001230, -0.00000450, -0.00000748, -0.00000184, -0.00000519, 0.00001341, -0.00000091, -0.00000095, -0.00000164, -0.00000360, 0.00000102, -0.00000445, 0.00001020, 0.00000463, -0.00000985, 0.00000188, -0.00001067, 0.00000984, -0.00000302, 0.00000181, 0.00000876, -0.00001057, 0.00000181, -0.00000674, 0.00000614, 0.00000218, 0.00000517, -0.00000152, -0.00000576, -0.00000395, 0.00000157, -0.00000148, -0.00000124, 0.00000393, 0.00000482, 0.00000421, -0.00001200, 0.00001011, -0.00000619, -0.00000005, 0.00000188, -0.00000917, 0.00001532, 0.00000873, -0.00001170, -0.00001769, 0.00000225, 0.00001224, -0.00000855, -0.00000008, -0.00000251, -0.00000616, 0.00001242, -0.00001904, -0.00000270, -0.00000330, 0.00000416, -0.00000752, -0.00001979, 0.00001386, -0.00001705, -0.00002916, -0.00001246, 0.00004558, 0.00004416, 0.00002236, 0.00005268, 0.00003275, -0.00003351, -0.00003495, -0.00000730, -0.00001151, -0.00000705, -0.00000635, -0.00000849, },
		{ -0.00000789, -0.00001271, -0.00000926, -0.00004682, -0.00002399, 0.00005738, 0.00004723, 0.00002973, 0.00006173, 0.00002627, -0.00003393, -0.00003254, 0.00000365, 0.00000228, -0.00002175, 0.00000281, -0.00000515, 0.00000440, -0.00001450, -0.00000779, 0.00001104, -0.00001279, 0.00000418, -0.00000794, 0.00000141, 0.00001016, -0.00000662, -0.00001764, -0.00000477, 0.00001723, 0.00000236, -0.00000873, 0.00000709, -0.00000559, 0.00000259, 0.00000149, -0.00000936, 0.00000808, 0.00000214, 0.00000507, -0.00000376, 0.00000070, -0.00000016, -0.00000582, -0.00000418, 0.00000238, 0.00000143, 0.00000611, 0.00000153, -0.00000470, -0.00000327, -0.00000451, 0.00001222, -0.00000799, 0.00000942, -0.00000032, -0.00000925, -0.00000017, -0.00000787, 0.00001285, 0.00000150, -0.00000224, -0.00000034, -0.00000444, 0.00000113, -0.00000428, 0.00000743, 0.00000823, -0.00001026, 0.00000186, -0.00001110, 0.00000734, 0.00000117, -0.00000252, 0.00001157, -0.00000981, 0.00000051, -0.00000623, 0.00000478, 0.00000349, 0.00000367, 0.00000040, -0.00000580, -0.00000449, 0.00000110, -0.00000081, -0.00000210, 0.00000441, 0.00000369, 0.00000618, -0.00001205, 0.00000771, -0.00000298, -0.00000255, 0.00000461, -0.00001048, 0.00001164, 0.00001243, -0.00001012, -0.00001768, -0.00000138, 0.00001281, -0.00000581, -0.00000292, 0.00000034, -0.00000960, 0.00001367, -0.00001657, -0.00000623, -0.00000057, 0.00000035, -0.00000249, -0.00002272, 0.00001187, -0.00001000, -0.00003255, -0.00001912, 0.00004055, 0.00005094, 0.00002321, 0.00005177, 0.00004108, -0.00003166, -0.00003847, -0.00000789, -0.00001193, -0.00000721, -0.00000647, },
		},
		{ { -0.00000147, -0.00000170, -0.00000406, -0.00000507, -0.00000327, -0.00000268, -0.00000918, -0.00000436, -0.00001399, -0.00003987, -0.00002009, 0.00002956, 0.00005443, 0.00001737, 0.00002752, 0.00005260, 0.00000452, -0.00000508, -0.00002595, -0.00001384, 0.00000662, -0.00002397, 0.00000507, -0.00000274, -0.00000248, 0.00000942, -0.00001672, 0.00000798, 0.00001054, 0.00000017, -0.00000803, -0.00001283, 0.00001098, 0.00000280, -0.00000831, -0.00000175, -0.00000279, 0.00000736, 0.00000771, -0.00000768, -0.00000027, -0.00000560, -0.00000643, 0.00000455, -0.00000389, 0.00000717, -0.00000432, -0.00000888, 0.00001385, -0.00000309, -0.00001586, 0.00000222, 0.00001239, 0.00000679, 0.00000262, -0.00001810, -0.00001143, 0.00001077, 0.00000824, 0.00001053, -0.00001220, -0.00000433, 0.00000442, -0.00000499, 0.00000768, -0.00000499, 0.00000442, -0.00000433, -0.00001220, 0.00001053, 0.00000824, 0.00001077, -0.00001143, -0.00001810, 0.00000262, 0.00000679, 0.00001239, 0.00000222, -0.00001586, -0.00000309, 0.00001385, -0.00000888, -0.00000432, 0.00000717, -0.00000389, 0.00000455, -0.00000643, -0.00000560, -0.00000027, -0.00000768, 0.00000771, 0.00000736, -0.00000279, -0.00000175, -0.00000831, 0.00000280, 0.00001098, -0.00001283, -0.00000803, 0.00000017, 0.00001054, 0.00000798, -0.00001672, 0.00000942, -0.00000248, -0.00000274, 0.00000507, -0.00002397, 0.00000662, -0.00001384, -0.00002595, -0.00000508, 0.00000452, 0.00005260, 0.00002752, 0.00001737, 0.00005443, 0.00002956, -0.00002009, -0.00003987, -0.00001399, -0.00000436, -0.00000918, -0.00000268, -0.00000327, -0.00000507, -0.00000406, -0.00000170, },
		{ -0.00000216, -0.00000140, -0.00000158, -0.00000311, -0.00000536, -0.00000412, -0.00000127, -0.00000888, -0.00000516, -0.00000957, -0.00003728, -0.00002367, 0.00001849, 0.00005316, 0.00002379, 0.00001556, 0.00005447, 0.00001227, -0.00000241, -0.00001791, -0.00002397, 0.00000987, -0.00002116, -0.00000234, 0.00000372, -0.00000765, 0.00001148, -0.00001382, 0.00000141, 0.00001399, 0.00000071, -0.00000629, -0.00001367, 0.00000681, 0.00000699, -0.00000817, -0.00000344, -0.00000140, 0.00000349, 0.00001031, -0.00000519, -0.00000330, -0.00000186, -0.00000948, 0.00000458, -0.00000245, 0.00000385, 0.00000131, -0.00001229, 0.00001060, 0.00000362, -0.00001604, -0.00000241, 0.00001125, 0.00000712, 0.00000653, -0.00001410, -0.00001784, 0.00000968, 0.00000815, 0.00001160, -0.00000759, -0.00000955, 0.00000675, -0.00000577, 0.00000666, -0.00000242, 0.00000095, 0.00000055, -0.00001468, 0.00000752, 0.00000929, 0.00001045, -0.00000515, -0.00001946, -0.00000207, 0.00000551, 0.00001300, 0.00000648, -0.00001361, -0.00000941, 0.00001380, -0.00000294, -0.00000970, 0.00000847, -0.00000298, 0.00000236, -0.00000229, -0.00000883, 0.00000105, -0.00000753, 0.00000304, 0.00001106, -0.00000277, -0.00000188, -0.00000644, -0.00000212, 0.00001365, -0.00000984, -0.00001112, 0.00000086, 0.00000639, 0.00001295, -0.00001614, 0.00000405, 0.00000422, -0.00000839, 0.00000916, -0.00002196, -0.00000057, -0.00000359, -0.00002960, -0.00001037, -0.00000182, 0.00004682, 0.00003953, 0.00001461, 0.00005184, 0.00004124, -0.00001436, -0.00004237, -0.00001851, -0.00000443, -0.00000895, -0.00000404, -0.00000312, -0.00000429, -0.00000485, },
		{ -0.00000515, -0.00000300, -0.00000126, -0.00000163, -0.00000230, -0.00000500, -0.00000540, -0.00000029, -0.00000769, -0.00000677, -0.00000544, -0.00003436, -0.00002607, 0.00000933, 0.00004789, 0.00003235, 0.00000582, 0.00005143, 0.00002182, -0.00000231, -0.00000770, -0.00003083, 0.00000768, -0.00001420, -0.00001098, 0.00000861, -0.00000966, 0.00000996, -0.00000853, -0.00000500, 0.00001536, 0.00000303, -0.00000576, -0.00001322, 0.00000256, 0.00000930, -0.00000582, -0.00000621, 0.00000009, 0.00000091, 0.00001014, -0.00000058, -0.00000670, 0.00000090, -0.00001030, 0.00000218, 0.00000072, -0.00000007, 0.00000550, -0.00001239, 0.00000494, 0.00000924, -0.00001380, -0.00000698, 0.00000950, 0.00000691, 0.00000892, -0.00000796, -0.00002305, 0.00000639, 0.00000926, 0.00001093, -0.00000188, -0.00001373, 0.00000698, -0.00000444, 0.00000394, 0.00000105, -0.00000242, 0.00000395, -0.00001461, 0.00000298, 0.00001062, 0.00000966, 0.00000000, -0.00001835, -0.00000659, 0.00000308, 0.00001297, 0.00001025, -0.00000996, -0.00001411, 0.00001046, 0.00000398, -0.00001311, 0.00000699, 0.00000003, -0.00000094, 0.00000148, -0.00001028, 0.00000009, -0.00000519, -0.00000224, 0.00001303, -0.00000057, -0.00000365, -0.00000358, -0.00000625, 0.00001375, -0.00000467, -0.00001483, 0.00000160, 0.00000306, 0.00001513, -0.00001209, -0.00000322, 0.00001025, -0.00001121, 0.00000906, -0.00001607, -0.00000910, 0.00000392, -0.00002806, -0.00001704, -0.00000765, 0.00003830, 0.00004970, 0.00001615, 0.00004641, 0.00005196, -0.00000581, -0.00004460, -0.00002322, -0.00000516, -0.00000861, -0.00000498, -0.00000363, -0.00000340, },
		{ -0.00000281, -0.00000481, -0.00000407, -0.00000124, -0.00000164, -0.00000185, -0.00000408, -0.00000662, -0.00000022, -0.00000557, -0.00000884, -0.00000211, -0.00003071, -0.00002816, 0.00000277, 0.00003938, 0.00004093, 0.00000031, 0.00004321, 0.00003257, -0.00000345, 0.00000172, -0.00003208, 0.00000030, -0.00000526, -0.00001821, 0.00001007, -0.00000806, 0.00000583, -0.00000271, -0.00000952, 0.00001396, 0.00000677, -0.00000555, -0.00001263, -0.00000064, 0.00000935, -0.00000199, -0.00000871, 0.00000031, 0.00000047, 0.00000747, 0.00000476, -0.00000888, 0.00000151, -0.00000858, -0.00000203, 0.00000431, -0.00000303, 0.00000711, -0.00000940, -0.00000157, 0.00001249, -0.00000936, -0.00001083, 0.00000700, 0.00000654, 0.00000954, -0.00000072, -0.00002572, 0.00000068, 0.00001116, 0.00000932, 0.00000363, -0.00001570, 0.00000477, -0.00000144, 0.00000045, 0.00000427, -0.00000460, 0.00000525, -0.00001228, -0.00000214, 0.00001136, 0.00000914, 0.00000361, -0.00001547, -0.00001005, -0.00000035, 0.00001199, 0.00001352, -0.00000563, -0.00001649, 0.00000465, 0.00001005, -0.00001342, 0.00000290, 0.00000405, -0.00000392, 0.00000363, -0.00000936, -0.00000278, -0.00000189, -0.00000651, 0.00001229, 0.00000364, -0.00000610, -0.00000113, -0.00000836, 0.00001103, 0.00000183, -0.00001783, 0.00000112, 0.00000156, 0.00001434, -0.00000569, -0.00001031, 0.00001363, -0.00001026, 0.00000536, -0.00000838, -0.00001625, 0.00000694, -0.00002218, -0.00002311, -0.00001375, 0.00002798, 0.00005671, 0.00002178, 0.00003964, 0.00006026, 0.00000572, -0.00004595, -0.00002840, -0.00000629, -0.00000850, -0.00000541, -0.00000455, },
		{ -0.00000547, -0.00000285, -0.00000392, -0.00000508, -0.00000154, -0.00000150, -0.00000175, -0.00000292, -0.00000729, -0.00000125, -0.00000281, -0.00001072, -0.00000012, -0.00002595, -0.00003044, -0.00000128, 0.00002917, 0.00004725, 0.00000030, 0.00003065, 0.00004282, -0.00000366, 0.00000773, -0.00002702, -0.00001026, 0.00000272, -0.00002180, 0.00000746, -0.00000374, 0.00000093, 0.00000177, -0.00001117, 0.00001012, 0.00001076, -0.00000447, -0.00001285, -0.00000236, 0.00000770, 0.00000195, -0.00000958, -0.00000150, 0.00000202, 0.00000354, 0.00000908, -0.00000856, -0.00000038, -0.00000489, -0.00000672, 0.00000679, -0.00000383, 0.00000590, -0.00000443, -0.00000718, 0.00001264, -0.00000349, -0.00001320, 0.00000371, 0.00000620, 0.00000859, 0.00000627, -0.00002496, -0.00000697, 0.00001290, 0.00000781, 0.00000778, -0.00001493, 0.00000051, 0.00000225, -0.00000264, 0.00000623, -0.00000502, 0.00000453, -0.00000854, -0.00000677, 0.00001087, 0.00000925, 0.00000578, -0.00001174, -0.00001188, -0.00000423, 0.00000972, 0.00001629, -0.00000114, -0.00001649, -0.00000219, 0.00001366, -0.00001035, -0.00000273, 0.00000762, -0.00000533, 0.00000350, -0.00000638, -0.00000650, 0.00000087, -0.00000852, 0.00000875, 0.00000881, -0.00000780, -0.00000029, -0.00000795, 0.00000615, 0.00000815, -0.00001878, -0.00000135, 0.00000202, 0.00001138, 0.00000128, -0.00001529, 0.00001330, -0.00000591, -0.00000019, -0.00000131, -0.00002006, 0.00000534, -0.00001404, -0.00002666, -0.00002026, 0.00001637, 0.00005978, 0.00003061, 0.00003325, 0.00006512, 0.00001978, -0.00004552, -0.00003434, -0.00000768, -0.00000875, -0.00000546, },
		{ -0.00000546, -0.00000602, -0.00000359, -0.00000274, -0.00000572, -0.00000225, -0.00000127, -0.00000182, -0.00000191, -0.00000709, -0.00000321, -0.00000005, -0.00001168, 0.00000010, -0.00002009, -0.00003285, -0.00000363, 0.00001916, 0.00004952, 0.00000584, 0.00001580, 0.00005023, -0.00000088, 0.00000893, -0.00001691, -0.00002080, 0.00000704, -0.00002078, 0.00000160, 0.00000139, -0.00000277, 0.00000370, -0.00001005, 0.00000510, 0.00001351, -0.00000163, -0.00001408, -0.00000308, 0.00000557, 0.00000455, -0.00000801, -0.00000513, 0.00000448, 0.00000008, 0.00001087, -0.00000529, -0.00000416, -0.00000059, -0.00001030, 0.00000698, -0.00000211, 0.00000260, 0.00000085, -0.00001044, 0.00000979, 0.00000258, -0.00001350, -0.00000017, 0.00000579, 0.00000667, 0.00001174, -0.00002055, -0.00001534, 0.00001326, 0.00000734, 0.00000989, -0.00001162, -0.00000472, 0.00000536, -0.00000433, 0.00000642, -0.00000374, 0.00000248, -0.00000451, -0.00001001, 0.00000894, 0.00000985, 0.00000700, -0.00000803, -0.00001200, -0.00000777, 0.00000598, 0.00001836, 0.00000328, -0.00001459, -0.00000854, 0.00001391, -0.00000468, -0.00000827, 0.00000935, -0.00000451, 0.00000131, -0.00000235, -0.00000965, 0.00000189, -0.00000790, 0.00000328, 0.00001337, -0.00000747, -0.00000161, -0.00000548, 0.00000053, 0.00001268, -0.00001680, -0.00000574, 0.00000368, 0.00000770, 0.00000702, -0.00001694, 0.00000940, 0.00000031, -0.00000542, 0.00000322, -0.00001991, 0.00000043, -0.00000617, -0.00002656, -0.00002649, 0.00000375, 0.00005851, 0.00004137, 0.00002877, 0.00006614, 0.00003536, -0.00004229, -0.00004112, -0.00000937, -0.00000929, },
		{ -0.00000988, -0.00000573, -0.00000599, -0.00000491, -0.00000167, -0.00000577, -0.00000327, -0.00000113, -0.00000183, -0.00000133, -0.00000601, -0.00000558, 0.00000196, -0.00001118, -0.00000145, -0.00001356, -0.00003479, -0.00000547, 0.00001107, 0.00004693, 0.00001561, 0.00000153, 0.00005252, 0.00000603, 0.00000571, -0.00000458, -0.00002796, 0.00000620, -0.00001582, -0.00000550, 0.00000516, -0.00000393, 0.00000305, -0.00000732, 0.00000061, 0.00001385, 0.00000306, -0.00001562, -0.00000395, 0.00000433, 0.00000491, -0.00000419, -0.00000944, 0.00000622, -0.00000135, 0.00000945, 0.00000030, -0.00000837, 0.00000266, -0.00001149, 0.00000445, 0.00000155, -0.00000139, 0.00000478, -0.00001064, 0.00000481, 0.00000752, -0.00001151, -0.00000411, 0.00000505, 0.00000446, 0.00001483, -0.00001316, -0.00002276, 0.00001118, 0.00000832, 0.00001005, -0.00000669, -0.00000955, 0.00000680, -0.00000416, 0.00000498, -0.00000139, 0.00000004, -0.00000119, -0.00001141, 0.00000583, 0.00001050, 0.00000784, -0.00000489, -0.00001082, -0.00001017, 0.00000094, 0.00001932, 0.00000764, -0.00001149, -0.00001319, 0.00001083, 0.00000208, -0.00001208, 0.00000848, -0.00000163, -0.00000202, 0.00000137, -0.00001105, 0.00000066, -0.00000516, -0.00000262, 0.00001579, -0.00000453, -0.00000469, -0.00000216, -0.00000418, 0.00001426, -0.00001194, -0.00001108, 0.00000523, 0.00000477, 0.00001036, -0.00001515, 0.00000320, 0.00000639, -0.00000848, 0.00000442, -0.00001653, -0.00000566, -0.00000068, -0.00002290, -0.00003120, -0.00000949, 0.00005266, 0.00005255, 0.00002742, 0.00006352, 0.00005109, -0.00003541, -0.00004851, -0.00001164, },
		{ -0.00001493, -0.00001027, -0.00000652, -0.00000536, -0.00000650, -0.00000104, -0.00000521, -0.00000434, -0.00000131, -0.00000163, -0.00000125, -0.00000438, -0.00000765, 0.00000259, -0.00000909, -0.00000430, -0.00000722, -0.00003542, -0.00000792, 0.00000593, 0.00003990, 0.00002725, -0.00000914, 0.00004831, 0.00001668, 0.00000011, 0.00000651, -0.00002943, 0.00000053, -0.00000899, -0.00001141, 0.00000604, -0.00000242, 0.00000088, -0.00000469, -0.00000194, 0.00001150, 0.00000868, -0.00001612, -0.00000620, 0.00000478, 0.00000311, 0.00000069, -0.00001271, 0.00000572, 0.00000004, 0.00000530, 0.00000662, -0.00001122, 0.00000353, -0.00000980, -0.00000026, 0.00000585, -0.00000446, 0.00000623, -0.00000799, -0.00000087, 0.00001023, -0.00000758, -0.00000735, 0.00000367, 0.00000253, 0.00001531, -0.00000413, -0.00002756, 0.00000620, 0.00001045, 0.00000895, -0.00000141, -0.00001274, 0.00000605, -0.00000232, 0.00000258, 0.00000111, -0.00000187, 0.00000084, -0.00001108, 0.00000219, 0.00001066, 0.00000869, -0.00000245, -0.00000902, -0.00001086, -0.00000481, 0.00001861, 0.00001201, -0.00000787, -0.00001552, 0.00000531, 0.00000813, -0.00001300, 0.00000508, 0.00000238, -0.00000519, 0.00000355, -0.00001014, -0.00000245, -0.00000155, -0.00000732, 0.00001515, 0.00000071, -0.00000833, 0.00000051, -0.00000671, 0.00001264, -0.00000518, -0.00001588, 0.00000535, 0.00000355, 0.00001104, -0.00001085, -0.00000347, 0.00001050, -0.00000849, 0.00000273, -0.00001157, -0.00001084, 0.00000142, -0.00001691, -0.00003307, -0.00002253, 0.00004218, 0.00006250, 0.00002996, 0.00005807, 0.00006544, -0.00002434, -0.00005588, },
		{ -0.00006230, -0.00001972, -0.00001028, -0.00000795, -0.00000429, -0.00000801, -0.00000107, -0.00000421, -0.00000517, -0.00000193, -0.00000126, -0.00000147, -0.00000269, -0.00000879, 0.00000160, -0.00000580, -0.00000755, -0.00000208, -0.00003402, -0.00001166, 0.00000381, 0.00002995, 0.00003788, -0.00001392, 0.00003774, 0.00002910, -0.00000492, 0.00001343, -0.00002467, -0.00000792, -0.00000303, -0.00001425, 0.00000374, 0.00000065, -0.00000107, -0.00000365, -0.00000201, 0.00000728, 0.00001358, -0.00001417, -0.00001049, 0.00000680, 0.00000025, 0.00000485, -0.00001333, 0.00000225, 0.00000386, -0.00000008, 0.00001168, -0.00001121, 0.00000149, -0.00000573, -0.00000580, 0.00000916, -0.00000534, 0.00000493, -0.00000353, -0.00000566, 0.00001020, -0.00000259, -0.00000916, 0.00000154, 0.00000110, 0.00001355, 0.00000488, -0.00002854, -0.00000130, 0.00001279, 0.00000767, 0.00000300, -0.00001356, 0.00000328, 0.00000043, 0.00000018, 0.00000293, -0.00000273, 0.00000152, -0.00000956, -0.00000126, 0.00000998, 0.00000963, -0.00000048, -0.00000730, -0.00000979, -0.00001030, 0.00001575, 0.00001629, -0.00000413, -0.00001559, -0.00000122, 0.00001189, -0.00001077, 0.00000008, 0.00000620, -0.00000699, 0.00000355, -0.00000718, -0.00000639, 0.00000149, -0.00000962, 0.00001144, 0.00000704, -0.00001101, 0.00000128, -0.00000656, 0.00000852, 0.00000184, -0.00001863, 0.00000326, 0.00000413, 0.00000976, -0.00000554, -0.00000884, 0.00001161, -0.00000571, -0.00000047, -0.00000688, -0.00001374, 0.00000035, -0.00001049, -0.00003139, -0.00003404, 0.00002731, 0.00006946, 0.00003662, 0.00005107, 0.00007692, -0.00000903, },
		{ 0.00001005, -0.00006655, -0.00002640, -0.00000987, -0.00000994, -0.00000309, -0.00000909, -0.00000181, -0.00000305, -0.00000554, -0.00000289, -0.00000095, -0.00000166, -0.00000144, -0.00000866, -0.00000074, -0.00000212, -0.00001017, 0.00000104, -0.00003030, -0.00001661, 0.00000391, 0.00001920, 0.00004495, -0.00001190, 0.00002262, 0.00004032, -0.00000659, 0.00001476, -0.00001517, -0.00001605, -0.00000023, -0.00001349, -0.00000066, 0.00000343, -0.00000121, -0.00000482, -0.00000026, 0.00000289, 0.00001599, -0.00000899, -0.00001647, 0.00000917, -0.00000189, 0.00000656, -0.00001048, -0.00000375, 0.00000864, -0.00000460, 0.00001370, -0.00000774, -0.00000291, -0.00000064, -0.00001048, 0.00001013, -0.00000355, 0.00000152, 0.00000122, -0.00000835, 0.00000767, 0.00000230, -0.00000904, -0.00000109, 0.00000005, 0.00001037, 0.00001230, -0.00002533, -0.00001012, 0.00001403, 0.00000718, 0.00000577, -0.00001204, -0.00000068, 0.00000304, -0.00000138, 0.00000363, -0.00000252, 0.00000123, -0.00000756, -0.00000393, 0.00000845, 0.00001045, 0.00000138, -0.00000614, -0.00000739, -0.00001446, 0.00001060, 0.00002011, -0.00000030, -0.00001394, -0.00000724, 0.00001251, -0.00000602, -0.00000508, 0.00000857, -0.00000675, 0.00000148, -0.00000311, -0.00000980, 0.00000275, -0.00000915, 0.00000559, 0.00001278, -0.00001139, -0.00000035, -0.00000424, 0.00000335, 0.00000742, -0.00001838, -0.00000095, 0.00000578, 0.00000773, -0.00000078, -0.00001180, 0.00000977, -0.00000134, -0.00000353, -0.00000382, -0.00001407, -0.00000263, -0.00000545, -0.00002639, -0.00004236, 0.00000892, 0.00007163, 0.00004701, 0.00004418, 0.00008431, },
		{ 0.00008679, 0.00003185, -0.00006727, -0.00003518, -0.00000920, -0.00001227, -0.00000212, -0.00000951, -0.00000316, -0.00000198, -0.00000543, -0.00000392, -0.00000101, -0.00000154, -0.00000090, -0.00000736, -0.00000373, 0.00000096, -0.00001132, 0.00000182, -0.00002460, -0.00002207, 0.00000484, 0.00000980, 0.00004688, -0.00000379, 0.00000597, 0.00004728, -0.00000327, 0.00001108, -0.00000389, -0.00002090, -0.00000157, -0.00001011, -0.00000510, 0.00000410, 0.00000116, -0.00000758, 0.00000167, 0.00000026, 0.00001480, -0.00000095, -0.00002264, 0.00001005, -0.00000163, 0.00000490, -0.00000460, -0.00001070, 0.00001231, -0.00000638, 0.00001185, -0.00000138, -0.00000820, 0.00000371, -0.00001281, 0.00000814, 0.00000047, -0.00000268, 0.00000472, -0.00000837, 0.00000352, 0.00000593, -0.00000696, -0.00000366, -0.00000097, 0.00000665, 0.00001705, -0.00001850, -0.00001850, 0.00001301, 0.00000801, 0.00000682, -0.00000885, -0.00000470, 0.00000456, -0.00000166, 0.00000323, -0.00000160, 0.00000057, -0.00000573, -0.00000560, 0.00000636, 0.00001078, 0.00000343, -0.00000560, -0.00000446, -0.00001638, 0.00000352, 0.00002276, 0.00000380, -0.00001135, -0.00001161, 0.00001000, -0.00000015, -0.00000892, 0.00000872, -0.00000450, -0.00000186, 0.00000078, -0.00001154, 0.00000174, -0.00000644, -0.00000086, 0.00001633, -0.00000884, -0.00000396, -0.00000095, -0.00000121, 0.00001032, -0.00001509, -0.00000634, 0.00000727, 0.00000621, 0.00000246, -0.00001215, 0.00000591, 0.00000306, -0.00000512, -0.00000281, -0.00001254, -0.00000594, -0.00000290, -0.00001933, -0.00004603, -0.00001132, 0.00006737, 0.00005996, 0.00003928, },
		{ 0.00003825, 0.00008424, 0.00005475, -0.00006313, -0.00004599, -0.00000863, -0.00001461, -0.00000175, -0.00000911, -0.00000498, -0.00000117, -0.00000498, -0.00000467, -0.00000166, -0.00000103, -0.00000099, -0.00000538, -0.00000648, 0.00000262, -0.00001066, 0.00000053, -0.00001782, -0.00002694, 0.00000518, 0.00000325, 0.00004344, 0.00000828, -0.00000873, 0.00004786, 0.00000491, 0.00000464, 0.00000580, -0.00002070, -0.00000625, -0.00000627, -0.00000752, 0.00000177, 0.00000538, -0.00001022, 0.00000187, 0.00000069, 0.00001013, 0.00000831, -0.00002682, 0.00000767, 0.00000194, 0.00000023, 0.00000268, -0.00001634, 0.00001292, -0.00000441, 0.00000653, 0.00000624, -0.00001242, 0.00000573, -0.00001205, 0.00000351, 0.00000550, -0.00000609, 0.00000594, -0.00000598, -0.00000096, 0.00000751, -0.00000347, -0.00000542, -0.00000229, 0.00000311, 0.00001878, -0.00000944, -0.00002464, 0.00000913, 0.00000996, 0.00000672, -0.00000505, -0.00000769, 0.00000451, -0.00000077, 0.00000223, -0.00000058, 0.00000007, -0.00000437, -0.00000638, 0.00000417, 0.00001037, 0.00000568, -0.00000538, -0.00000194, -0.00001563, -0.00000454, 0.00002341, 0.00000836, -0.00000846, -0.00001377, 0.00000521, 0.00000524, -0.00001040, 0.00000663, -0.00000096, -0.00000532, 0.00000332, -0.00001101, -0.00000121, -0.00000268, -0.00000626, 0.00001672, -0.00000367, -0.00000832, 0.00000178, -0.00000385, 0.00001016, -0.00000964, -0.00001148, 0.00000742, 0.00000593, 0.00000393, -0.00001054, 0.00000141, 0.00000613, -0.00000471, -0.00000339, -0.00001036, -0.00000825, -0.00000290, -0.00001207, -0.00004439, -0.00003096, 0.00005564, 0.00007341, },
		{ 0.00008451, 0.00004257, 0.00007737, 0.00007666, -0.00005295, -0.00005834, -0.00000877, -0.00001649, -0.00000237, -0.00000783, -0.00000709, -0.00000068, -0.00000445, -0.00000488, -0.00000288, -0.00000036, -0.00000135, -0.00000340, -0.00000821, 0.00000252, -0.00000843, -0.00000200, -0.00001111, -0.00003013, 0.00000393, 0.00000013, 0.00003567, 0.00002139, -0.00001858, 0.00004157, 0.00001607, -0.00000148, 0.00001135, -0.00001555, -0.00001197, -0.00000440, -0.00000673, -0.00000302, 0.00000965, -0.00001070, -0.00000104, 0.00000423, 0.00000349, 0.00001631, -0.00002688, 0.00000112, 0.00000834, -0.00000579, 0.00000895, -0.00001853, 0.00000940, 0.00000099, -0.00000062, 0.00001289, -0.00001381, 0.00000461, -0.00000840, -0.00000255, 0.00000993, -0.00000743, 0.00000460, -0.00000212, -0.00000450, 0.00000682, 0.00000050, -0.00000571, -0.00000399, 0.00000007, 0.00001785, 0.00000010, -0.00002718, 0.00000262, 0.00001217, 0.00000637, -0.00000168, -0.00000901, 0.00000299, 0.00000072, 0.00000125, 0.00000006, -0.00000004, -0.00000347, -0.00000668, 0.00000235, 0.00000916, 0.00000788, -0.00000496, -0.00000057, -0.00001243, -0.00001225, 0.00002135, 0.00001333, -0.00000555, -0.00001384, -0.00000054, 0.00000872, -0.00000922, 0.00000299, 0.00000277, -0.00000776, 0.00000380, -0.00000841, -0.00000514, 0.00000068, -0.00000936, 0.00001385, 0.00000293, -0.00001190, 0.00000269, -0.00000404, 0.00000752, -0.00000351, -0.00001493, 0.00000556, 0.00000692, 0.00000418, -0.00000803, -0.00000245, 0.00000720, -0.00000263, -0.00000450, -0.00000869, -0.00000903, -0.00000460, -0.00000644, -0.00003790, -0.00004713, 0.00003650, },
		{ 0.00001149, 0.00008994, 0.00005281, 0.00006793, 0.00009519, -0.00003604, -0.00007120, -0.00001049, -0.00001737, -0.00000428, -0.00000571, -0.00000929, -0.00000051, -0.00000404, -0.00000451, -0.00000436, 0.00000006, -0.00000151, -0.00000199, -0.00000855, 0.00000087, -0.00000538, -0.00000475, -0.00000555, -0.00003099, 0.00000076, 0.00000003, 0.00002545, 0.00003267, -0.00002195, 0.00002969, 0.00002724, -0.00000450, 0.00001186, -0.00000739, -0.00001585, -0.00000606, -0.00000306, -0.00000838, 0.00001164, -0.00000749, -0.00000711, 0.00000947, -0.00000264, 0.00002056, -0.00002164, -0.00000903, 0.00001576, -0.00001068, 0.00001189, -0.00001606, 0.00000208, 0.00000821, -0.00000725, 0.00001648, -0.00001143, 0.00000059, -0.00000304, -0.00000840, 0.00001228, -0.00000610, 0.00000125, 0.00000188, -0.00000618, 0.00000433, 0.00000383, -0.00000424, -0.00000578, -0.00000256, 0.00001505, 0.00000843, -0.00002560, -0.00000542, 0.00001347, 0.00000662, 0.00000062, -0.00000864, 0.00000060, 0.00000206, 0.00000085, 0.00000014, 0.00000017, -0.00000272, -0.00000688, 0.00000109, 0.00000743, 0.00000958, -0.00000386, -0.00000064, -0.00000761, -0.00001818, 0.00001633, 0.00001823, -0.00000255, -0.00001241, -0.00000585, 0.00000952, -0.00000591, -0.00000103, 0.00000560, -0.00000848, 0.00000223, -0.00000454, -0.00000876, 0.00000248, -0.00000973, 0.00000857, 0.00000933, -0.00001332, 0.00000125, -0.00000215, 0.00000367, 0.00000170, -0.00001579, 0.00000180, 0.00000850, 0.00000408, -0.00000570, -0.00000489, 0.00000637, 0.00000018, -0.00000508, -0.00000812, -0.00000854, -0.00000666, -0.00000358, -0.00002827, -0.00005717, },
		{ -0.00005948, -0.00001631, 0.00008658, 0.00006814, 0.00005858, 0.00010807, -0.00001250, -0.00008292, -0.00001481, -0.00001679, -0.00000763, -0.00000298, -0.00001131, -0.00000072, -0.00000377, -0.00000381, -0.00000559, -0.00000021, -0.00000115, -0.00000140, -0.00000762, -0.00000165, -0.00000245, -0.00000680, -0.00000177, -0.00002945, -0.00000393, 0.00000187, 0.00001490, 0.00004004, -0.00001883, 0.00001485, 0.00003541, -0.00000297, 0.00000839, 0.00000075, -0.00001559, -0.00001122, 0.00000156, -0.00001174, 0.00000956, -0.00000046, -0.00001484, 0.00001389, -0.00000564, 0.00001950, -0.00001149, -0.00002069, 0.00002147, -0.00001200, 0.00001011, -0.00000916, -0.00000731, 0.00001480, -0.00001105, 0.00001577, -0.00000550, -0.00000510, 0.00000233, -0.00001240, 0.00001171, -0.00000235, -0.00000296, 0.00000476, -0.00000570, 0.00000095, 0.00000560, -0.00000125, -0.00000703, -0.00000505, 0.00001131, 0.00001439, -0.00002035, -0.00001333, 0.00001274, 0.00000790, 0.00000177, -0.00000711, -0.00000179, 0.00000261, 0.00000121, -0.00000018, 0.00000034, -0.00000182, -0.00000722, 0.00000030, 0.00000566, 0.00001033, -0.00000185, -0.00000192, -0.00000239, -0.00002121, 0.00000871, 0.00002223, 0.00000085, -0.00001021, -0.00000968, 0.00000768, -0.00000158, -0.00000424, 0.00000685, -0.00000737, -0.00000079, -0.00000059, -0.00001094, 0.00000214, -0.00000775, 0.00000230, 0.00001392, -0.00001188, -0.00000221, 0.00000074, 0.00000009, 0.00000482, -0.00001394, -0.00000304, 0.00000968, 0.00000451, -0.00000413, -0.00000587, 0.00000435, 0.00000265, -0.00000445, -0.00000854, -0.00000763, -0.00000791, -0.00000359, -0.00001796, },
		{ -0.00000948, -0.00005407, -0.00004268, 0.00007243, 0.00008602, 0.00005250, 0.00011373, 0.00001649, -0.00009126, -0.00002276, -0.00001449, -0.00001228, -0.00000011, -0.00001276, -0.00000146, -0.00000350, -0.00000318, -0.00000608, -0.00000141, -0.00000026, -0.00000148, -0.00000594, -0.00000416, -0.00000044, -0.00000767, 0.00000016, -0.00002599, -0.00000933, 0.00000439, 0.00000573, 0.00004257, -0.00001062, 0.00000017, 0.00003847, 0.00000275, 0.00000341, 0.00000595, -0.00001050, -0.00001799, 0.00000435, -0.00001098, 0.00000301, 0.00000885, -0.00002152, 0.00001480, -0.00000378, 0.00001322, 0.00000146, -0.00003078, 0.00002275, -0.00000826, 0.00000377, 0.00000038, -0.00001611, 0.00001829, -0.00001051, 0.00001079, 0.00000256, -0.00001055, 0.00000600, -0.00001350, 0.00000821, 0.00000283, -0.00000664, 0.00000567, -0.00000345, -0.00000217, 0.00000537, 0.00000249, -0.00000702, -0.00000763, 0.00000731, 0.00001755, -0.00001270, -0.00001939, 0.00000942, 0.00000999, 0.00000228, -0.00000522, -0.00000341, 0.00000212, 0.00000209, -0.00000046, 0.00000011, -0.00000060, -0.00000763, -0.00000038, 0.00000433, 0.00000994, 0.00000090, -0.00000377, 0.00000199, -0.00002079, -0.00000044, 0.00002433, 0.00000493, -0.00000783, -0.00001159, 0.00000391, 0.00000242, -0.00000581, 0.00000637, -0.00000488, -0.00000429, 0.00000233, -0.00001107, -0.00000013, -0.00000448, -0.00000336, 0.00001565, -0.00000776, -0.00000664, 0.00000328, -0.00000203, 0.00000544, -0.00001011, -0.00000770, 0.00000951, 0.00000586, -0.00000336, -0.00000585, 0.00000209, 0.00000399, -0.00000262, -0.00000925, -0.00000724, -0.00000780, -0.00000552, },
		{ -0.00000785, -0.00000460, -0.00004275, -0.00006312, 0.00004743, 0.00010232, 0.00005267, 0.00011184, 0.00004856, -0.00009366, -0.00003503, -0.00001068, -0.00001763, 0.00000217, -0.00001315, -0.00000293, -0.00000299, -0.00000305, -0.00000558, -0.00000338, 0.00000080, -0.00000178, -0.00000419, -0.00000593, 0.00000026, -0.00000738, 0.00000063, -0.00002141, -0.00001454, 0.00000651, -0.00000101, 0.00004047, 0.00000043, -0.00001163, 0.00003592, 0.00001076, -0.00000018, 0.00000657, -0.00000193, -0.00002326, 0.00000277, -0.00000543, -0.00000651, 0.00001750, -0.00002418, 0.00001037, 0.00000294, 0.00000365, 0.00001392, -0.00003612, 0.00001792, 0.00000037, -0.00000535, 0.00000975, -0.00002160, 0.00001703, -0.00000552, 0.00000293, 0.00001059, -0.00001385, 0.00000682, -0.00001150, 0.00000264, 0.00000801, -0.00000859, 0.00000441, -0.00000028, -0.00000409, 0.00000335, 0.00000587, -0.00000524, -0.00001023, 0.00000338, 0.00001815, -0.00000429, -0.00002230, 0.00000376, 0.00001209, 0.00000289, -0.00000374, -0.00000390, 0.00000082, 0.00000297, -0.00000023, -0.00000068, 0.00000082, -0.00000781, -0.00000136, 0.00000376, 0.00000854, 0.00000389, -0.00000535, 0.00000458, -0.00001718, -0.00000962, 0.00002364, 0.00000968, -0.00000551, -0.00001169, -0.00000065, 0.00000499, -0.00000550, 0.00000464, -0.00000181, -0.00000728, 0.00000349, -0.00000918, -0.00000354, -0.00000122, -0.00000719, 0.00001430, -0.00000193, -0.00001064, 0.00000431, -0.00000216, 0.00000397, -0.00000551, -0.00001100, 0.00000756, 0.00000795, -0.00000291, -0.00000556, 0.00000034, 0.00000401, -0.00000015, -0.00000941, -0.00000786, -0.00000661, },
		{ -0.00000521, -0.00000914, -0.00000375, -0.00002874, -0.00007404, 0.00001399, 0.00011199, 0.00006092, 0.00010377, 0.00008034, -0.00008770, -0.00005158, -0.00000628, -0.00002264, 0.00000293, -0.00001201, -0.00000529, -0.00000209, -0.00000356, -0.00000422, -0.00000562, 0.00000147, -0.00000178, -0.00000293, -0.00000660, -0.00000020, -0.00000639, 0.00000025, -0.00001647, -0.00001888, 0.00000761, -0.00000504, 0.00003476, 0.00001199, -0.00001896, 0.00002879, 0.00001836, -0.00000027, 0.00000296, 0.00000710, -0.00002387, -0.00000424, 0.00000357, -0.00001594, 0.00002216, -0.00002072, 0.00000058, 0.00001262, -0.00000596, 0.00002243, -0.00003459, 0.00000704, 0.00001195, -0.00001428, 0.00001595, -0.00002185, 0.00001077, 0.00000261, -0.00000549, 0.00001630, -0.00001370, 0.00000455, -0.00000712, -0.00000359, 0.00001175, -0.00000817, 0.00000144, 0.00000273, -0.00000429, 0.00000028, 0.00000781, -0.00000162, -0.00001238, -0.00000056, 0.00001689, 0.00000331, -0.00002156, -0.00000321, 0.00001311, 0.00000424, -0.00000303, -0.00000340, -0.00000069, 0.00000325, 0.00000075, -0.00000190, 0.00000204, -0.00000735, -0.00000293, 0.00000392, 0.00000662, 0.00000643, -0.00000593, 0.00000500, -0.00001135, -0.00001719, 0.00001977, 0.00001467, -0.00000313, -0.00001051, -0.00000484, 0.00000553, -0.00000370, 0.00000242, 0.00000101, -0.00000904, 0.00000272, -0.00000591, -0.00000695, 0.00000091, -0.00000863, 0.00001046, 0.00000417, -0.00001295, 0.00000330, -0.00000053, 0.00000140, -0.00000146, -0.00001219, 0.00000401, 0.00001015, -0.00000212, -0.00000556, -0.00000062, 0.00000305, 0.00000213, -0.00000841, -0.00000940, },
		{ -0.00001114, -0.00000463, -0.00000862, -0.00000595, -0.00001576, -0.00007377, -0.00002314, 0.00011026, 0.00007697, 0.00009263, 0.00010803, -0.00007178, -0.00007122, -0.00000291, -0.00002599, 0.00000132, -0.00000915, -0.00000842, -0.00000090, -0.00000456, -0.00000250, -0.00000741, 0.00000120, -0.00000119, -0.00000234, -0.00000629, -0.00000133, -0.00000528, -0.00000046, -0.00001168, -0.00002204, 0.00000751, -0.00000657, 0.00002672, 0.00002224, -0.00002149, 0.00001917, 0.00002318, 0.00000358, -0.00000267, 0.00001300, -0.00001795, -0.00001555, 0.00001284, -0.00002163, 0.00002035, -0.00001094, -0.00001253, 0.00002190, -0.00001202, 0.00002446, -0.00002593, -0.00000785, 0.00002323, -0.00001978, 0.00001681, -0.00001645, 0.00000090, 0.00001145, -0.00001191, 0.00001804, -0.00000981, -0.00000006, -0.00000174, -0.00000893, 0.00001306, -0.00000543, -0.00000232, 0.00000465, -0.00000286, -0.00000275, 0.00000761, 0.00000326, -0.00001333, -0.00000471, 0.00001455, 0.00000902, -0.00001761, -0.00000997, 0.00001210, 0.00000655, -0.00000294, -0.00000249, -0.00000169, 0.00000257, 0.00000238, -0.00000309, 0.00000259, -0.00000598, -0.00000506, 0.00000441, 0.00000486, 0.00000794, -0.00000517, 0.00000351, -0.00000468, -0.00002181, 0.00001299, 0.00001906, -0.00000034, -0.00000869, -0.00000785, 0.00000412, -0.00000125, 0.00000055, 0.00000303, -0.00000930, 0.00000039, -0.00000226, -0.00000933, 0.00000130, -0.00000787, 0.00000531, 0.00000909, -0.00001284, 0.00000043, 0.00000200, -0.00000108, 0.00000109, -0.00001120, -0.00000038, 0.00001160, -0.00000051, -0.00000602, -0.00000096, 0.00000173, 0.00000353, -0.00000621, },
		{ -0.00000329, -0.00001219, -0.00000550, -0.00000652, -0.00000923, -0.00000693, -0.00006322, -0.00005772, 0.00009397, 0.00009798, 0.00008268, 0.00012830, -0.00004574, -0.00009153, -0.00000272, -0.00002641, -0.00000311, -0.00000480, -0.00001185, 0.00000017, -0.00000556, -0.00000111, -0.00000814, -0.00000020, -0.00000007, -0.00000224, -0.00000543, -0.00000253, -0.00000445, -0.00000127, -0.00000726, -0.00002397, 0.00000625, -0.00000602, 0.00001754, 0.00003011, -0.00001997, 0.00000934, 0.00002402, 0.00001002, -0.00000706, 0.00001306, -0.00000594, -0.00002785, 0.00001836, -0.00002071, 0.00001158, 0.00000310, -0.00002532, 0.00002700, -0.00001191, 0.00001942, -0.00001199, -0.00002322, 0.00003062, -0.00001948, 0.00001174, -0.00000675, -0.00001002, 0.00001826, -0.00001436, 0.00001526, -0.00000303, -0.00000551, 0.00000306, -0.00001215, 0.00001162, -0.00000104, -0.00000576, 0.00000496, -0.00000041, -0.00000471, 0.00000519, 0.00000828, -0.00001228, -0.00000917, 0.00001165, 0.00001249, -0.00001161, -0.00001494, 0.00000867, 0.00000943, -0.00000289, -0.00000188, -0.00000171, 0.00000099, 0.00000415, -0.00000367, 0.00000210, -0.00000374, -0.00000737, 0.00000465, 0.00000384, 0.00000813, -0.00000319, 0.00000088, 0.00000135, -0.00002280, 0.00000429, 0.00002183, 0.00000313, -0.00000672, -0.00000933, 0.00000135, 0.00000091, -0.00000045, 0.00000406, -0.00000825, -0.00000279, 0.00000080, -0.00001005, -0.00000003, -0.00000570, 0.00000020, 0.00001184, -0.00001033, -0.00000348, 0.00000437, -0.00000253, 0.00000178, -0.00000858, -0.00000459, 0.00001165, 0.00000201, -0.00000670, -0.00000116, 0.00000069, 0.00000379, },
		{ 0.00000312, -0.00000047, -0.00001186, -0.00000775, -0.00000394, -0.00001145, -0.00000381, -0.00004578, -0.00008343, 0.00006290, 0.00011869, 0.00007834, 0.00013925, -0.00001126, -0.00010901, -0.00000785, -0.00002315, -0.00001011, 0.00000017, -0.00001476, 0.00000051, -0.00000601, -0.00000062, -0.00000752, -0.00000252, 0.00000114, -0.00000220, -0.00000451, -0.00000338, -0.00000400, -0.00000224, -0.00000321, -0.00002469, 0.00000387, -0.00000374, 0.00000810, 0.00003521, -0.00001558, 0.00000094, 0.00002126, 0.00001641, -0.00000723, 0.00000660, 0.00000924, -0.00003664, 0.00001676, -0.00001236, -0.00000221, 0.00001752, -0.00003362, 0.00002511, -0.00000493, 0.00000898, 0.00000367, -0.00003508, 0.00003136, -0.00001264, 0.00000204, 0.00000456, -0.00001895, 0.00002080, -0.00001201, 0.00000874, 0.00000490, -0.00001005, 0.00000598, -0.00001266, 0.00000780, 0.00000390, -0.00000796, 0.00000366, 0.00000220, -0.00000493, 0.00000120, 0.00001210, -0.00000879, -0.00001364, 0.00000835, 0.00001398, -0.00000506, -0.00001701, 0.00000319, 0.00001197, -0.00000213, -0.00000211, -0.00000068, -0.00000100, 0.00000538, -0.00000316, 0.00000054, -0.00000103, -0.00000926, 0.00000405, 0.00000391, 0.00000714, -0.00000054, -0.00000190, 0.00000559, -0.00002026, -0.00000486, 0.00002209, 0.00000728, -0.00000474, -0.00000942, -0.00000188, 0.00000202, -0.00000048, 0.00000430, -0.00000632, -0.00000599, 0.00000251, -0.00000899, -0.00000251, -0.00000316, -0.00000375, 0.00001205, -0.00000612, -0.00000733, 0.00000565, -0.00000248, 0.00000088, -0.00000528, -0.00000770, 0.00001006, 0.00000507, -0.00000710, -0.00000166, 0.00000021, },
		{ 0.00000021, 0.00000207, 0.00000156, -0.00001000, -0.00001055, -0.00000232, -0.00001113, -0.00000597, -0.00002639, -0.00009557, 0.00002032, 0.00013240, 0.00008273, 0.00014122, 0.00002823, -0.00011974, -0.00001978, -0.00001649, -0.00001857, 0.00000443, -0.00001618, -0.00000044, -0.00000553, -0.00000125, -0.00000578, -0.00000515, 0.00000187, -0.00000184, -0.00000381, -0.00000380, -0.00000368, -0.00000356, 0.00000038, -0.00002411, 0.00000034, -0.00000004, -0.00000092, 0.00003745, -0.00000936, -0.00000543, 0.00001657, 0.00001994, -0.00000180, -0.00000448, 0.00002293, -0.00003773, 0.00000677, 0.00000158, -0.00001696, 0.00002775, -0.00003420, 0.00001546, 0.00000721, -0.00000337, 0.00001683, -0.00004015, 0.00002441, -0.00000056, -0.00000947, 0.00001420, -0.00002334, 0.00001806, -0.00000550, 0.00000030, 0.00001189, -0.00001222, 0.00000633, -0.00001059, 0.00000256, 0.00000824, -0.00000838, 0.00000120, 0.00000410, -0.00000335, -0.00000320, 0.00001354, -0.00000302, -0.00001741, 0.00000444, 0.00001412, 0.00000067, -0.00001588, -0.00000323, 0.00001309, -0.00000015, -0.00000329, 0.00000097, -0.00000265, 0.00000549, -0.00000143, -0.00000171, 0.00000149, -0.00001009, 0.00000227, 0.00000494, 0.00000552, 0.00000203, -0.00000395, 0.00000749, -0.00001507, -0.00001285, 0.00001940, 0.00001168, -0.00000262, -0.00000854, -0.00000476, 0.00000179, 0.00000017, 0.00000417, -0.00000401, -0.00000857, 0.00000257, -0.00000657, -0.00000526, -0.00000121, -0.00000594, 0.00001006, -0.00000129, -0.00001010, 0.00000537, -0.00000109, -0.00000090, -0.00000227, -0.00000918, 0.00000710, 0.00000802, -0.00000674, -0.00000266, },
		{ -0.00000401, 0.00000033, 0.00000118, 0.00000246, -0.00000714, -0.00001273, -0.00000276, -0.00000804, -0.00001117, -0.00001022, -0.00009242, -0.00002729, 0.00013262, 0.00009638, 0.00013690, 0.00006818, -0.00012030, -0.00003866, -0.00000789, -0.00002666, 0.00000651, -0.00001539, -0.00000286, -0.00000412, -0.00000276, -0.00000352, -0.00000734, 0.00000167, -0.00000108, -0.00000330, -0.00000403, -0.00000318, -0.00000523, 0.00000319, -0.00002196, -0.00000444, 0.00000468, -0.00000880, 0.00003673, -0.00000188, -0.00001016, 0.00001206, 0.00001884, 0.00000833, -0.00001615, 0.00003025, -0.00002893, -0.00000995, 0.00001688, -0.00002777, 0.00003013, -0.00002601, -0.00000016, 0.00002085, -0.00001337, 0.00002395, -0.00003699, 0.00001098, 0.00001373, -0.00001927, 0.00001930, -0.00002192, 0.00001058, 0.00000331, -0.00000768, 0.00001611, -0.00001127, 0.00000425, -0.00000675, -0.00000290, 0.00001103, -0.00000697, -0.00000169, 0.00000475, -0.00000051, -0.00000667, 0.00001200, 0.00000415, -0.00001946, -0.00000034, 0.00001355, 0.00000474, -0.00001210, -0.00000907, 0.00001193, 0.00000307, -0.00000496, 0.00000247, -0.00000323, 0.00000421, 0.00000119, -0.00000397, 0.00000310, -0.00000949, -0.00000060, 0.00000645, 0.00000399, 0.00000385, -0.00000475, 0.00000717, -0.00000856, -0.00001833, 0.00001397, 0.00001557, -0.00000008, -0.00000716, -0.00000672, 0.00000038, 0.00000093, 0.00000404, -0.00000170, -0.00001013, 0.00000109, -0.00000352, -0.00000742, -0.00000044, -0.00000638, 0.00000669, 0.00000303, -0.00001118, 0.00000360, 0.00000108, -0.00000273, -0.00000023, -0.00000900, 0.00000341, 0.00001022, -0.00000539, },
		{ -0.00000320, -0.00000533, 0.00000016, 0.00000075, 0.00000244, -0.00000419, -0.00001325, -0.00000551, -0.00000332, -0.00001628, -0.00000108, -0.00007584, -0.00007153, 0.00011503, 0.00011650, 0.00013080, 0.00010409, -0.00010869, -0.00006290, 0.00000010, -0.00003232, 0.00000525, -0.00001224, -0.00000642, -0.00000226, -0.00000455, -0.00000151, -0.00000849, 0.00000049, -0.00000017, -0.00000269, -0.00000435, -0.00000234, -0.00000695, 0.00000468, -0.00001796, -0.00001031, 0.00000971, -0.00001462, 0.00003278, 0.00000670, -0.00001404, 0.00000934, 0.00001322, 0.00002005, -0.00002362, 0.00002791, -0.00001119, -0.00002896, 0.00002814, -0.00003051, 0.00002324, -0.00001075, -0.00001779, 0.00003145, -0.00001738, 0.00002331, -0.00002647, -0.00000585, 0.00002638, -0.00002421, 0.00001836, -0.00001511, 0.00000030, 0.00001191, -0.00001306, 0.00001651, -0.00000738, 0.00000056, -0.00000226, -0.00000745, 0.00001176, -0.00000411, -0.00000427, 0.00000403, 0.00000267, -0.00000813, 0.00000769, 0.00001124, -0.00001882, -0.00000604, 0.00001259, 0.00000702, -0.00000690, -0.00001287, 0.00000825, 0.00000695, -0.00000634, 0.00000301, -0.00000233, 0.00000180, 0.00000399, -0.00000543, 0.00000330, -0.00000746, -0.00000406, 0.00000767, 0.00000324, 0.00000454, -0.00000428, 0.00000527, -0.00000217, -0.00002055, 0.00000666, 0.00001804, 0.00000306, -0.00000556, -0.00000761, -0.00000171, 0.00000125, 0.00000410, 0.00000042, -0.00001056, -0.00000150, -0.00000066, -0.00000841, -0.00000096, -0.00000554, 0.00000295, 0.00000601, -0.00001048, 0.00000086, 0.00000331, -0.00000394, 0.00000056, -0.00000757, -0.00000026, 0.00001123, },
		{ 0.00001098, -0.00000058, -0.00000619, -0.00000047, 0.00000067, 0.00000205, -0.00000205, -0.00001169, -0.00000978, 0.00000099, -0.00001840, -0.00000033, -0.00005091, -0.00010412, 0.00007909, 0.00013714, 0.00012790, 0.00013277, -0.00008502, -0.00008926, 0.00000446, -0.00003384, 0.00000030, -0.00000733, -0.00001026, -0.00000068, -0.00000595, -0.00000031, -0.00000847, -0.00000131, 0.00000040, -0.00000169, -0.00000487, -0.00000137, -0.00000809, 0.00000425, -0.00001213, -0.00001667, 0.00001392, -0.00001729, 0.00002533, 0.00001623, -0.00001757, 0.00000868, 0.00000509, 0.00002915, -0.00002304, 0.00001559, 0.00001133, -0.00004421, 0.00003067, -0.00002343, 0.00000863, 0.00000746, -0.00003245, 0.00003508, -0.00001360, 0.00001558, -0.00001154, -0.00002193, 0.00003381, -0.00002252, 0.00001164, -0.00000484, -0.00001008, 0.00001788, -0.00001445, 0.00001312, -0.00000152, -0.00000353, 0.00000173, -0.00001028, 0.00001042, -0.00000047, -0.00000599, 0.00000227, 0.00000517, -0.00000712, 0.00000166, 0.00001658, -0.00001496, -0.00001223, 0.00001113, 0.00000793, -0.00000171, -0.00001370, 0.00000259, 0.00001040, -0.00000654, 0.00000210, -0.00000009, -0.00000105, 0.00000613, -0.00000551, 0.00000203, -0.00000447, -0.00000733, 0.00000791, 0.00000362, 0.00000419, -0.00000295, 0.00000270, 0.00000292, -0.00001949, -0.00000124, 0.00001833, 0.00000669, -0.00000380, -0.00000758, -0.00000385, 0.00000081, 0.00000428, 0.00000235, -0.00000993, -0.00000463, 0.00000132, -0.00000800, -0.00000244, -0.00000413, -0.00000032, 0.00000730, -0.00000842, -0.00000212, 0.00000497, -0.00000422, 0.00000026, -0.00000548, -0.00000331, },
		{ -0.00000544, 0.00000974, 0.00000198, -0.00000636, -0.00000142, 0.00000055, 0.00000192, -0.00000121, -0.00000851, -0.00001396, 0.00000283, -0.00001602, -0.00000656, -0.00002453, -0.00011910, 0.00002880, 0.00015048, 0.00013189, 0.00015335, -0.00005155, -0.00011351, 0.00000251, -0.00003059, -0.00000759, -0.00000199, -0.00001331, -0.00000008, -0.00000654, -0.00000008, -0.00000757, -0.00000309, 0.00000022, -0.00000030, -0.00000532, -0.00000083, -0.00000794, 0.00000168, -0.00000507, -0.00002231, 0.00001587, -0.00001588, 0.00001457, 0.00002610, -0.00002039, 0.00000896, -0.00000228, 0.00003181, -0.00001311, -0.00000359, 0.00003227, -0.00005004, 0.00002216, -0.00000797, -0.00000944, 0.00002327, -0.00003973, 0.00002979, -0.00000280, 0.00000364, 0.00000362, -0.00003333, 0.00003387, -0.00001433, 0.00000109, 0.00000598, -0.00001784, 0.00001961, -0.00001164, 0.00000694, 0.00000479, -0.00000679, 0.00000438, -0.00001106, 0.00000740, 0.00000321, -0.00000659, 0.00000006, 0.00000625, -0.00000393, -0.00000453, 0.00001877, -0.00000810, -0.00001797, 0.00000874, 0.00000821, 0.00000222, -0.00001146, -0.00000382, 0.00001223, -0.00000501, -0.00000025, 0.00000285, -0.00000339, 0.00000686, -0.00000402, -0.00000034, -0.00000131, -0.00000963, 0.00000672, 0.00000506, 0.00000323, -0.00000136, 0.00000031, 0.00000607, -0.00001583, -0.00000831, 0.00001612, 0.00001036, -0.00000170, -0.00000694, -0.00000555, -0.00000041, 0.00000432, 0.00000410, -0.00000841, -0.00000773, 0.00000201, -0.00000639, -0.00000431, -0.00000286, -0.00000259, 0.00000702, -0.00000565, -0.00000471, 0.00000574, -0.00000362, -0.00000074, -0.00000331, },
		{ -0.00000144, -0.00000663, 0.00000790, 0.00000414, -0.00000592, -0.00000230, -0.00000002, 0.00000237, -0.00000147, -0.00000487, -0.00001632, 0.00000098, -0.00000950, -0.00001615, -0.00000349, -0.00011452, -0.00002785, 0.00014897, 0.00014358, 0.00016757, -0.00001203, -0.00013147, -0.00000720, -0.00002330, -0.00001677, 0.00000219, -0.00001468, -0.00000072, -0.00000636, -0.00000051, -0.00000641, -0.00000430, -0.00000078, 0.00000104, -0.00000510, -0.00000134, -0.00000611, -0.00000262, 0.00000192, -0.00002569, 0.00001430, -0.00001013, 0.00000150, 0.00003486, -0.00002123, 0.00000808, -0.00000572, 0.00002617, 0.00000422, -0.00002378, 0.00004500, -0.00004311, 0.00000387, 0.00001139, -0.00002528, 0.00003185, -0.00003729, 0.00001632, 0.00001194, -0.00000830, 0.00001495, -0.00003746, 0.00002641, -0.00000164, -0.00001022, 0.00001440, -0.00002111, 0.00001670, -0.00000556, -0.00000033, 0.00001001, -0.00000830, 0.00000530, -0.00000994, 0.00000336, 0.00000631, -0.00000613, -0.00000197, 0.00000564, 0.00000050, -0.00000919, 0.00001716, 0.00000069, -0.00002199, 0.00000494, 0.00000849, 0.00000428, -0.00000689, -0.00000938, 0.00001155, -0.00000176, -0.00000341, 0.00000551, -0.00000438, 0.00000590, -0.00000128, -0.00000304, 0.00000118, -0.00001041, 0.00000411, 0.00000707, 0.00000232, -0.00000013, -0.00000134, 0.00000721, -0.00001070, -0.00001338, 0.00001166, 0.00001336, 0.00000091, -0.00000594, -0.00000657, -0.00000212, 0.00000391, 0.00000565, -0.00000616, -0.00001032, 0.00000129, -0.00000404, -0.00000593, -0.00000220, -0.00000375, 0.00000563, -0.00000287, -0.00000649, 0.00000563, -0.00000239, -0.00000202, },
		{ -0.00000329, -0.00000001, -0.00000707, 0.00000579, 0.00000582, -0.00000521, -0.00000269, -0.00000122, 0.00000322, -0.00000209, -0.00000217, -0.00001569, -0.00000438, -0.00000101, -0.00002459, 0.00000751, -0.00009305, -0.00008059, 0.00012765, 0.00016000, 0.00017909, 0.00002954, -0.00014005, -0.00002440, -0.00001401, -0.00002511, 0.00000389, -0.00001409, -0.00000231, -0.00000586, -0.00000107, -0.00000549, -0.00000476, -0.00000213, 0.00000160, -0.00000360, -0.00000319, -0.00000289, -0.00000746, 0.00000714, -0.00002540, 0.00000864, -0.00000083, -0.00001185, 0.00004041, -0.00001839, 0.00000404, -0.00000345, 0.00001321, 0.00002408, -0.00003821, 0.00004489, -0.00002392, -0.00001948, 0.00002842, -0.00003360, 0.00003046, -0.00002567, -0.00000198, 0.00002609, -0.00001612, 0.00001971, -0.00003383, 0.00001332, 0.00001228, -0.00001907, 0.00001827, -0.00001928, 0.00001003, 0.00000204, -0.00000687, 0.00001296, -0.00000770, 0.00000458, -0.00000737, -0.00000095, 0.00000837, -0.00000483, -0.00000336, 0.00000365, 0.00000484, -0.00001110, 0.00001207, 0.00000970, -0.00002305, -0.00000044, 0.00000900, 0.00000455, -0.00000135, -0.00001262, 0.00000813, 0.00000252, -0.00000639, 0.00000693, -0.00000354, 0.00000350, 0.00000195, -0.00000517, 0.00000236, -0.00000954, 0.00000054, 0.00000892, 0.00000208, 0.00000041, -0.00000210, 0.00000678, -0.00000529, -0.00001581, 0.00000574, 0.00001493, 0.00000406, -0.00000466, -0.00000692, -0.00000389, 0.00000286, 0.00000685, -0.00000336, -0.00001201, -0.00000070, -0.00000157, -0.00000683, -0.00000230, -0.00000401, 0.00000370, -0.00000061, -0.00000739, 0.00000486, -0.00000084, },
		{ 0.00000081, -0.00000442, 0.00000098, -0.00000692, 0.00000356, 0.00000719, -0.00000460, -0.00000242, -0.00000279, 0.00000385, -0.00000208, -0.00000140, -0.00001201, -0.00001161, 0.00000632, -0.00002807, 0.00000709, -0.00006132, -0.00011948, 0.00008612, 0.00017477, 0.00019198, 0.00007027, -0.00013803, -0.00004712, -0.00000544, -0.00003078, 0.00000261, -0.00001206, -0.00000408, -0.00000565, -0.00000137, -0.00000496, -0.00000485, -0.00000302, 0.00000063, -0.00000065, -0.00000602, 0.00000066, -0.00001115, 0.00000897, -0.00002078, -0.00000046, 0.00000998, -0.00002272, 0.00004057, -0.00001059, -0.00000398, 0.00000401, -0.00000334, 0.00004020, -0.00004142, 0.00003102, 0.00000307, -0.00004093, 0.00003717, -0.00003133, 0.00001950, -0.00000826, -0.00002020, 0.00003516, -0.00001707, 0.00001731, -0.00002410, -0.00000208, 0.00002390, -0.00002304, 0.00001678, -0.00001315, 0.00000143, 0.00000916, -0.00001117, 0.00001308, -0.00000522, 0.00000269, -0.00000401, -0.00000484, 0.00000918, -0.00000303, -0.00000397, 0.00000097, 0.00000784, -0.00000979, 0.00000475, 0.00001696, -0.00002041, -0.00000703, 0.00000945, 0.00000372, 0.00000361, -0.00001266, 0.00000258, 0.00000665, -0.00000816, 0.00000649, -0.00000100, 0.00000039, 0.00000472, -0.00000603, 0.00000201, -0.00000738, -0.00000323, 0.00000989, 0.00000283, 0.00000028, -0.00000220, 0.00000545, -0.00000063, -0.00001559, -0.00000053, 0.00001452, 0.00000747, -0.00000299, -0.00000679, -0.00000535, 0.00000124, 0.00000745, -0.00000020, -0.00001256, -0.00000364, 0.00000045, -0.00000676, -0.00000305, -0.00000375, 0.00000177, 0.00000091, -0.00000758, 0.00000369, },
		{ 0.00000232, 0.00000249, -0.00000536, 0.00000154, -0.00000623, 0.00000115, 0.00000848, -0.00000428, -0.00000172, -0.00000409, 0.00000350, -0.00000074, -0.00000275, -0.00000642, -0.00001821, 0.00000964, -0.00002480, -0.00000251, -0.00002803, -0.00013756, 0.00002927, 0.00017964, 0.00020877, 0.00010940, -0.00012605, -0.00007233, -0.00000015, -0.00003291, -0.00000113, -0.00000971, -0.00000521, -0.00000605, -0.00000149, -0.00000440, -0.00000533, -0.00000262, -0.00000209, 0.00000322, -0.00000870, 0.00000302, -0.00001207, 0.00000658, -0.00001244, -0.00001107, 0.00001942, -0.00002833, 0.00003392, 0.00000209, -0.00001495, 0.00001384, -0.00001817, 0.00004696, -0.00003119, 0.00000678, 0.00003046, -0.00005345, 0.00003398, -0.00001877, 0.00000246, 0.00000975, -0.00003336, 0.00003614, -0.00001064, 0.00000939, -0.00001148, -0.00001605, 0.00003042, -0.00002112, 0.00001067, -0.00000461, -0.00000693, 0.00001409, -0.00001242, 0.00001055, -0.00000150, 0.00000028, -0.00000053, -0.00000777, 0.00000870, -0.00000098, -0.00000393, -0.00000153, 0.00000872, -0.00000573, -0.00000292, 0.00002082, -0.00001416, -0.00001384, 0.00000918, 0.00000272, 0.00000678, -0.00000949, -0.00000379, 0.00000934, -0.00000801, 0.00000417, 0.00000255, -0.00000240, 0.00000623, -0.00000532, 0.00000040, -0.00000464, -0.00000641, 0.00000950, 0.00000454, -0.00000018, -0.00000203, 0.00000387, 0.00000271, -0.00001326, -0.00000600, 0.00001203, 0.00001060, -0.00000074, -0.00000636, -0.00000632, -0.00000069, 0.00000723, 0.00000300, -0.00001179, -0.00000706, 0.00000152, -0.00000571, -0.00000414, -0.00000336, 0.00000016, 0.00000169, -0.00000733, },
		{ -0.00000685, 0.00000080, 0.00000418, -0.00000605, 0.00000157, -0.00000493, -0.00000152, 0.00000977, -0.00000406, -0.00000118, -0.00000437, 0.00000166, 0.00000196, -0.00000541, -0.00000088, -0.00002173, 0.00000742, -0.00001560, -0.00001640, -0.00000146, -0.00013280, -0.00003349, 0.00016703, 0.00022874, 0.00014842, -0.00010586, -0.00009700, 0.00000032, -0.00003185, -0.00000600, -0.00000823, -0.00000530, -0.00000676, -0.00000203, -0.00000307, -0.00000694, -0.00000065, -0.00000596, 0.00000668, -0.00000966, 0.00000271, -0.00000940, 0.00000048, -0.00000236, -0.00002035, 0.00002454, -0.00002684, 0.00002070, 0.00001778, -0.00002601, 0.00002184, -0.00002615, 0.00004136, -0.00000960, -0.00002088, 0.00005014, -0.00005229, 0.00001891, 0.00000034, -0.00001517, 0.00002305, -0.00003803, 0.00002845, 0.00000127, -0.00000079, 0.00000034, -0.00002555, 0.00003055, -0.00001397, 0.00000186, 0.00000401, -0.00001311, 0.00001582, -0.00001062, 0.00000612, 0.00000255, -0.00000199, 0.00000250, -0.00000939, 0.00000705, 0.00000109, -0.00000354, -0.00000317, 0.00000734, -0.00000016, -0.00000900, 0.00002043, -0.00000530, -0.00001946, 0.00000747, 0.00000233, 0.00000757, -0.00000398, -0.00000935, 0.00000959, -0.00000584, 0.00000056, 0.00000606, -0.00000395, 0.00000609, -0.00000325, -0.00000182, -0.00000212, -0.00000837, 0.00000768, 0.00000678, -0.00000046, -0.00000200, 0.00000246, 0.00000462, -0.00000969, -0.00000979, 0.00000785, 0.00001274, 0.00000216, -0.00000566, -0.00000683, -0.00000255, 0.00000613, 0.00000583, -0.00000969, -0.00001040, 0.00000132, -0.00000392, -0.00000518, -0.00000316, -0.00000101, 0.00000191, },
		{ 0.00000177, -0.00000627, -0.00000090, 0.00000586, -0.00000630, 0.00000089, -0.00000292, -0.00000443, 0.00001083, -0.00000343, -0.00000148, -0.00000317, -0.00000155, 0.00000527, -0.00000792, 0.00000260, -0.00002073, 0.00000016, -0.00000353, -0.00002881, 0.00001283, -0.00010869, -0.00009048, 0.00013277, 0.00024724, 0.00019010, -0.00007900, -0.00011907, -0.00000415, -0.00002901, -0.00001050, -0.00000828, -0.00000476, -0.00000688, -0.00000388, -0.00000035, -0.00000983, 0.00000228, -0.00000954, 0.00000797, -0.00000753, -0.00000088, -0.00000359, -0.00000741, 0.00000655, -0.00002538, 0.00002335, -0.00001819, 0.00000310, 0.00003299, -0.00003332, 0.00002391, -0.00002422, 0.00002429, 0.00001746, -0.00004352, 0.00005587, -0.00003667, -0.00000406, 0.00001994, -0.00002765, 0.00002788, -0.00003334, 0.00001412, 0.00001503, -0.00000949, 0.00000832, -0.00002903, 0.00002471, -0.00000355, -0.00000717, 0.00001061, -0.00001594, 0.00001423, -0.00000648, 0.00000088, 0.00000608, -0.00000361, 0.00000467, -0.00000953, 0.00000444, 0.00000307, -0.00000313, -0.00000361, 0.00000429, 0.00000531, -0.00001200, 0.00001601, 0.00000439, -0.00002249, 0.00000389, 0.00000289, 0.00000621, 0.00000240, -0.00001255, 0.00000709, -0.00000225, -0.00000331, 0.00000847, -0.00000369, 0.00000442, -0.00000047, -0.00000385, -0.00000047, -0.00000889, 0.00000476, 0.00000893, -0.00000012, -0.00000228, 0.00000139, 0.00000535, -0.00000582, -0.00001147, 0.00000276, 0.00001328, 0.00000556, -0.00000454, -0.00000705, -0.00000404, 0.00000431, 0.00000783, -0.00000640, -0.00001308, -0.00000025, -0.00000180, -0.00000582, -0.00000328, -0.00000176, },
		{ -0.00000222, 0.00000145, -0.00000560, -0.00000283, 0.00000734, -0.00000584, -0.00000068, -0.00000029, -0.00000722, 0.00001113, -0.00000179, -0.00000307, -0.00000060, -0.00000533, 0.00000789, -0.00000870, 0.00000275, -0.00001535, -0.00000983, 0.00000731, -0.00003500, 0.00001341, -0.00007314, -0.00013087, 0.00007807, 0.00025635, 0.00023663, -0.00004551, -0.00013792, -0.00001254, -0.00002607, -0.00001375, -0.00000947, -0.00000483, -0.00000521, -0.00000765, 0.00000369, -0.00001324, 0.00000468, -0.00001092, 0.00000562, -0.00000189, -0.00000706, 0.00000344, -0.00001423, 0.00001134, -0.00002426, 0.00001568, -0.00000441, -0.00001498, 0.00004350, -0.00003342, 0.00001773, -0.00001261, 0.00000039, 0.00004198, -0.00005384, 0.00004538, -0.00001033, -0.00002812, 0.00003359, -0.00003091, 0.00002328, -0.00002128, -0.00000291, 0.00002650, -0.00001366, 0.00001096, -0.00002672, 0.00001480, 0.00000749, -0.00001411, 0.00001382, -0.00001519, 0.00001000, -0.00000113, -0.00000406, 0.00000842, -0.00000428, 0.00000575, -0.00000822, 0.00000119, 0.00000480, -0.00000284, -0.00000299, 0.00000059, 0.00000916, -0.00001135, 0.00000881, 0.00001287, -0.00002195, -0.00000140, 0.00000411, 0.00000352, 0.00000799, -0.00001250, 0.00000236, 0.00000164, -0.00000634, 0.00000906, -0.00000164, 0.00000188, 0.00000220, -0.00000500, 0.00000001, -0.00000815, 0.00000135, 0.00001036, 0.00000108, -0.00000281, 0.00000053, 0.00000530, -0.00000236, -0.00001117, -0.00000223, 0.00001191, 0.00000899, -0.00000276, -0.00000717, -0.00000504, 0.00000216, 0.00000864, -0.00000234, -0.00001451, -0.00000307, 0.00000015, -0.00000580, -0.00000373, },
		{ -0.00000440, -0.00000254, 0.00000103, -0.00000480, -0.00000492, 0.00000832, -0.00000436, -0.00000313, 0.00000261, -0.00000931, 0.00001005, 0.00000128, -0.00000585, 0.00000257, -0.00000837, 0.00000850, -0.00000682, -0.00000043, -0.00000733, -0.00001920, 0.00001326, -0.00003287, 0.00000310, -0.00003617, -0.00014766, 0.00001002, 0.00024711, 0.00028746, -0.00000318, -0.00015402, -0.00002354, -0.00002404, -0.00001604, -0.00001041, -0.00000709, -0.00000101, -0.00001309, 0.00000795, -0.00001551, 0.00000463, -0.00000854, -0.00000074, 0.00000624, -0.00001372, 0.00000890, -0.00001708, 0.00001011, -0.00001695, 0.00000348, 0.00001068, -0.00002899, 0.00004577, -0.00002469, 0.00000379, 0.00000509, -0.00002328, 0.00005631, -0.00004806, 0.00002127, 0.00001944, -0.00004572, 0.00003661, -0.00002394, 0.00001131, -0.00000593, -0.00001807, 0.00003232, -0.00001195, 0.00000859, -0.00002034, 0.00000344, 0.00001664, -0.00001744, 0.00001332, -0.00001150, 0.00000431, 0.00000421, -0.00000780, 0.00000920, -0.00000390, 0.00000567, -0.00000567, -0.00000228, 0.00000609, -0.00000265, -0.00000182, -0.00000261, 0.00001045, -0.00000749, 0.00000075, 0.00001832, -0.00001768, -0.00000759, 0.00000529, 0.00000063, 0.00001149, -0.00000913, -0.00000339, 0.00000457, -0.00000772, 0.00000766, 0.00000161, -0.00000065, 0.00000398, -0.00000493, -0.00000057, -0.00000665, -0.00000187, 0.00001067, 0.00000306, -0.00000330, -0.00000033, 0.00000488, 0.00000028, -0.00000937, -0.00000621, 0.00000875, 0.00001179, -0.00000012, -0.00000718, -0.00000566, 0.00000016, 0.00000814, 0.00000187, -0.00001427, -0.00000677, 0.00000141, -0.00000500, },
		{ -0.00000349, -0.00000508, -0.00000284, 0.00000057, -0.00000387, -0.00000698, 0.00000842, -0.00000173, -0.00000614, 0.00000510, -0.00000999, 0.00000718, 0.00000564, -0.00000907, 0.00000506, -0.00000940, 0.00000635, -0.00000238, -0.00000563, 0.00000065, -0.00002483, 0.00001245, -0.00002344, -0.00001234, -0.00000697, -0.00013975, -0.00005968, 0.00021255, 0.00033771, 0.00005197, -0.00016764, -0.00003676, -0.00002257, -0.00001890, -0.00000923, -0.00001257, 0.00000536, -0.00001867, 0.00001040, -0.00001461, 0.00000058, -0.00000202, -0.00000989, 0.00001438, -0.00001790, 0.00001003, -0.00001417, 0.00000293, -0.00000556, -0.00000957, 0.00002263, -0.00003513, 0.00003825, -0.00000822, -0.00001447, 0.00002297, -0.00003951, 0.00005569, -0.00002730, -0.00000953, 0.00004396, -0.00005105, 0.00002770, -0.00000910, -0.00000363, 0.00000790, -0.00002761, 0.00003094, -0.00000494, 0.00000307, -0.00001242, -0.00000678, 0.00002214, -0.00001673, 0.00000972, -0.00000608, -0.00000154, 0.00000850, -0.00000979, 0.00000835, -0.00000257, 0.00000454, -0.00000226, -0.00000546, 0.00000668, -0.00000233, -0.00000072, -0.00000443, 0.00000904, -0.00000171, -0.00000610, 0.00001970, -0.00001044, -0.00001341, 0.00000556, -0.00000146, 0.00001224, -0.00000329, -0.00000856, 0.00000551, -0.00000722, 0.00000472, 0.00000515, -0.00000227, 0.00000447, -0.00000374, -0.00000176, -0.00000500, -0.00000437, 0.00000975, 0.00000544, -0.00000339, -0.00000133, 0.00000432, 0.00000205, -0.00000680, -0.00000860, 0.00000438, 0.00001324, 0.00000332, -0.00000687, -0.00000618, -0.00000130, 0.00000652, 0.00000547, -0.00001216, -0.00001073, 0.00000150, },
		{ 0.00000010, -0.00000152, -0.00000552, -0.00000323, 0.00000011, -0.00000286, -0.00000873, 0.00000734, 0.00000189, -0.00000908, 0.00000642, -0.00000872, 0.00000257, 0.00001056, -0.00001151, 0.00000562, -0.00000766, 0.00000165, 0.00000339, -0.00001069, 0.00000596, -0.00002491, 0.00000544, -0.00001030, -0.00002632, 0.00000875, -0.00011221, -0.00011772, 0.00015078, 0.00037794, 0.00012391, -0.00017722, -0.00005359, -0.00001993, -0.00002439, -0.00000468, -0.00002091, 0.00001197, -0.00002186, 0.00000876, -0.00000907, -0.00000767, 0.00000726, -0.00001904, 0.00001923, -0.00001687, 0.00000530, -0.00000573, -0.00000792, 0.00000610, -0.00001921, 0.00002774, -0.00003169, 0.00002220, 0.00001223, -0.00003169, 0.00003477, -0.00004325, 0.00003983, 0.00000265, -0.00003791, 0.00005585, -0.00004195, 0.00000946, 0.00000863, -0.00001638, 0.00001628, -0.00002969, 0.00002302, 0.00000509, -0.00000296, -0.00000539, -0.00001406, 0.00002340, -0.00001255, 0.00000428, -0.00000030, -0.00000644, 0.00001109, -0.00000985, 0.00000609, -0.00000053, 0.00000263, 0.00000145, -0.00000780, 0.00000633, -0.00000164, -0.00000024, -0.00000452, 0.00000558, 0.00000423, -0.00001011, 0.00001700, -0.00000178, -0.00001745, 0.00000427, -0.00000224, 0.00001040, 0.00000358, -0.00001167, 0.00000408, -0.00000521, 0.00000109, 0.00000798, -0.00000243, 0.00000370, -0.00000188, -0.00000297, -0.00000371, -0.00000591, 0.00000782, 0.00000773, -0.00000279, -0.00000250, 0.00000365, 0.00000311, -0.00000411, -0.00000925, -0.00000036, 0.00001286, 0.00000721, -0.00000588, -0.00000688, -0.00000210, 0.00000430, 0.00000777, -0.00000838, -0.00001413, },
		{ -0.00001613, -0.00000282, 0.00000047, -0.00000546, -0.00000379, -0.00000032, -0.00000189, -0.00000990, 0.00000502, 0.00000598, -0.00001111, 0.00000591, -0.00000546, -0.00000309, 0.00001480, -0.00001186, 0.00000348, -0.00000338, -0.00000433, 0.00000858, -0.00001347, 0.00000699, -0.00001965, -0.00000503, 0.00000186, -0.00003378, 0.00001009, -0.00007467, -0.00015287, 0.00006685, 0.00039574, 0.00021398, -0.00017791, -0.00007714, -0.00001392, -0.00003382, 0.00000285, -0.00002990, 0.00001564, -0.00001986, 0.00000147, 0.00000108, -0.00001846, 0.00001626, -0.00002455, 0.00001785, -0.00000932, -0.00000469, 0.00000575, -0.00001850, 0.00001374, -0.00002203, 0.00002442, -0.00001980, 0.00000157, 0.00003116, -0.00004230, 0.00003602, -0.00003336, 0.00001314, 0.00003301, -0.00005527, 0.00005143, -0.00002078, -0.00001241, 0.00002336, -0.00002273, 0.00001735, -0.00002489, 0.00001105, 0.00001506, -0.00000704, -0.00000086, -0.00001771, 0.00002087, -0.00000620, -0.00000156, 0.00000464, -0.00000965, 0.00001174, -0.00000815, 0.00000282, 0.00000185, 0.00000037, 0.00000483, -0.00000881, 0.00000491, -0.00000044, -0.00000059, -0.00000312, 0.00000128, 0.00000870, -0.00001055, 0.00001126, 0.00000640, -0.00001866, 0.00000130, -0.00000176, 0.00000677, 0.00000983, -0.00001180, 0.00000063, -0.00000261, -0.00000225, 0.00000937, -0.00000107, 0.00000215, 0.00000003, -0.00000371, -0.00000304, -0.00000653, 0.00000531, 0.00000947, -0.00000141, -0.00000366, 0.00000280, 0.00000371, -0.00000176, -0.00000840, -0.00000456, 0.00001056, 0.00001088, -0.00000383, -0.00000784, -0.00000241, 0.00000216, 0.00000836, -0.00000357, },
		{ 0.00000134, -0.00001607, -0.00000691, 0.00000188, -0.00000467, -0.00000451, -0.00000074, -0.00000104, -0.00001031, 0.00000168, 0.00000980, -0.00001146, 0.00000335, -0.00000074, -0.00000858, 0.00001700, -0.00000932, -0.00000123, 0.00000227, -0.00000974, 0.00001130, -0.00001268, 0.00000374, -0.00001106, -0.00001521, 0.00000909, -0.00003272, 0.00000077, -0.00003833, -0.00015935, -0.00002730, 0.00037903, 0.00031816, -0.00016132, -0.00011104, -0.00000344, -0.00004646, 0.00001069, -0.00003570, 0.00001290, -0.00001091, -0.00001115, 0.00001370, -0.00002839, 0.00002109, -0.00002321, 0.00000893, 0.00000376, -0.00001706, 0.00001616, -0.00002445, 0.00001411, -0.00001686, 0.00001389, -0.00000322, -0.00001809, 0.00004311, -0.00004247, 0.00002561, -0.00001300, -0.00001663, 0.00005477, -0.00005630, 0.00003193, 0.00000628, -0.00003092, 0.00003024, -0.00002083, 0.00001182, -0.00001571, -0.00000158, 0.00002215, -0.00000769, 0.00000073, -0.00001813, 0.00001573, 0.00000082, -0.00000655, 0.00000799, -0.00001091, 0.00001053, -0.00000506, -0.00000097, 0.00000415, -0.00000176, 0.00000726, -0.00000820, 0.00000251, 0.00000117, -0.00000164, -0.00000093, -0.00000254, 0.00001060, -0.00000769, 0.00000423, 0.00001235, -0.00001666, -0.00000287, -0.00000058, 0.00000251, 0.00001409, -0.00000879, -0.00000383, -0.00000047, -0.00000453, 0.00000907, 0.00000141, 0.00000051, 0.00000141, -0.00000372, -0.00000293, -0.00000652, 0.00000268, 0.00001035, 0.00000060, -0.00000452, 0.00000168, 0.00000400, 0.00000005, -0.00000659, -0.00000751, 0.00000670, 0.00001351, -0.00000056, -0.00000884, -0.00000270, 0.00000074, 0.00000726, },
		{ 0.00000499, 0.00000530, -0.00001369, -0.00001145, 0.00000206, -0.00000311, -0.00000528, -0.00000118, -0.00000035, -0.00001000, -0.00000221, 0.00001254, -0.00000970, -0.00000091, 0.00000436, -0.00001255, 0.00001619, -0.00000388, -0.00000739, 0.00000745, -0.00001271, 0.00001043, -0.00000839, -0.00000222, -0.00000223, -0.00002167, 0.00000955, -0.00002462, -0.00001263, -0.00001255, -0.00013875, -0.00011514, 0.00032014, 0.00042544, -0.00011689, -0.00015711, 0.00000968, -0.00005875, 0.00001426, -0.00003413, 0.00000151, 0.00000452, -0.00002640, 0.00002479, -0.00003319, 0.00001835, -0.00001364, -0.00000626, 0.00001890, -0.00002731, 0.00002106, -0.00002259, 0.00000633, -0.00000536, -0.00000006, 0.00001272, -0.00003136, 0.00004449, -0.00003151, 0.00000627, 0.00001137, -0.00004082, 0.00006143, -0.00004063, 0.00000321, 0.00003108, -0.00004011, 0.00002715, -0.00001173, 0.00000254, -0.00000569, -0.00001171, 0.00002464, -0.00000474, 0.00000004, -0.00001642, 0.00000942, 0.00000719, -0.00000989, 0.00000943, -0.00001028, 0.00000784, -0.00000112, -0.00000467, 0.00000587, -0.00000324, 0.00000828, -0.00000597, -0.00000051, 0.00000288, -0.00000295, 0.00000115, -0.00000482, 0.00000968, -0.00000271, -0.00000217, 0.00001496, -0.00001190, -0.00000727, 0.00000051, -0.00000128, 0.00001560, -0.00000335, -0.00000794, 0.00000033, -0.00000543, 0.00000733, 0.00000429, -0.00000056, 0.00000197, -0.00000306, -0.00000314, -0.00000624, 0.00000030, 0.00001030, 0.00000291, -0.00000482, 0.00000029, 0.00000404, 0.00000136, -0.00000440, -0.00000888, 0.00000204, 0.00001441, 0.00000373, -0.00000937, -0.00000347, 0.00000035, },
		{ 0.00000088, 0.00000243, 0.00000742, -0.00000929, -0.00001538, 0.00000056, -0.00000094, -0.00000585, -0.00000176, 0.00000021, -0.00000910, -0.00000607, 0.00001358, -0.00000591, -0.00000597, 0.00000852, -0.00001389, 0.00001218, 0.00000346, -0.00001321, 0.00001031, -0.00001215, 0.00000613, -0.00000202, -0.00000839, 0.00000396, -0.00002264, 0.00000413, -0.00001358, -0.00002327, -0.00000206, -0.00009968, -0.00017972, 0.00021953, 0.00051825, -0.00003506, -0.00021266, 0.00001965, -0.00006475, 0.00000848, -0.00002241, -0.00001806, 0.00002305, -0.00003943, 0.00002952, -0.00002930, 0.00000671, 0.00000271, -0.00002376, 0.00003103, -0.00003078, 0.00001740, -0.00001230, -0.00000748, 0.00000829, -0.00001230, 0.00002292, -0.00003472, 0.00003481, -0.00001230, -0.00001630, 0.00003204, -0.00005252, 0.00005106, -0.00001297, -0.00002608, 0.00004614, -0.00003703, 0.00001534, 0.00000110, -0.00000662, 0.00000196, -0.00001737, 0.00002233, 0.00000072, -0.00000161, -0.00001383, 0.00000314, 0.00001207, -0.00001127, 0.00000906, -0.00000812, 0.00000416, 0.00000305, -0.00000769, 0.00000659, -0.00000369, 0.00000771, -0.00000248, -0.00000356, 0.00000421, -0.00000397, 0.00000235, -0.00000508, 0.00000652, 0.00000276, -0.00000635, 0.00001405, -0.00000556, -0.00001078, 0.00000079, -0.00000386, 0.00001434, 0.00000327, -0.00001040, -0.00000059, -0.00000518, 0.00000476, 0.00000679, -0.00000061, 0.00000175, -0.00000204, -0.00000332, -0.00000594, -0.00000167, 0.00000949, 0.00000515, -0.00000441, -0.00000122, 0.00000377, 0.00000230, -0.00000228, -0.00000876, -0.00000252, 0.00001323, 0.00000843, -0.00000880, -0.00000504, },
		{ -0.00000737, 0.00000175, 0.00000055, 0.00000732, -0.00000375, -0.00001764, -0.00000275, 0.00000134, -0.00000588, -0.00000262, 0.00000069, -0.00000784, -0.00000942, 0.00001269, -0.00000071, -0.00001061, 0.00001060, -0.00001217, 0.00000569, 0.00001102, -0.00001682, 0.00000967, -0.00000814, -0.00000020, 0.00000422, -0.00001239, 0.00000587, -0.00001853, -0.00000412, -0.00000439, -0.00002660, -0.00000579, -0.00005524, -0.00020841, 0.00008772, 0.00057533, 0.00008854, -0.00026842, 0.00001711, -0.00005789, -0.00000999, -0.00000101, -0.00004183, 0.00003879, -0.00004456, 0.00002401, -0.00001550, -0.00001208, 0.00002146, -0.00003788, 0.00003501, -0.00002446, 0.00000489, 0.00000390, -0.00002268, 0.00001874, -0.00001811, 0.00002429, -0.00002778, 0.00001705, 0.00000962, -0.00003528, 0.00004262, -0.00004875, 0.00002685, 0.00001836, -0.00004714, 0.00004700, -0.00002270, -0.00000101, 0.00001303, -0.00001222, 0.00000535, -0.00001828, 0.00001642, 0.00000692, -0.00000284, -0.00001132, -0.00000236, 0.00001509, -0.00001076, 0.00000720, -0.00000487, 0.00000005, 0.00000682, -0.00000951, 0.00000599, -0.00000297, 0.00000575, 0.00000160, -0.00000594, 0.00000474, -0.00000425, 0.00000223, -0.00000351, 0.00000230, 0.00000711, -0.00000752, 0.00001037, 0.00000082, -0.00001245, -0.00000003, -0.00000504, 0.00001095, 0.00000955, -0.00001037, -0.00000297, -0.00000439, 0.00000209, 0.00000834, 0.00000039, 0.00000107, -0.00000106, -0.00000324, -0.00000573, -0.00000323, 0.00000815, 0.00000700, -0.00000332, -0.00000260, 0.00000311, 0.00000297, -0.00000049, -0.00000754, -0.00000615, 0.00001009, 0.00001263, -0.00000663, },
		{ -0.00000272, -0.00000993, 0.00000216, 0.00000007, 0.00000525, 0.00000167, -0.00001740, -0.00000747, 0.00000300, -0.00000504, -0.00000380, 0.00000109, -0.00000637, -0.00001194, 0.00001005, 0.00000493, -0.00001368, 0.00000999, -0.00000780, -0.00000184, 0.00001689, -0.00001689, 0.00000548, -0.00000198, -0.00000643, 0.00000819, -0.00001284, 0.00000367, -0.00001168, -0.00001133, -0.00000060, -0.00002186, -0.00001774, -0.00001903, -0.00019688, -0.00005555, 0.00057682, 0.00024910, -0.00030833, -0.00000859, -0.00003388, -0.00004050, 0.00002535, -0.00006273, 0.00004484, -0.00003731, 0.00000711, 0.00000601, -0.00003313, 0.00003637, -0.00004294, 0.00002762, -0.00000838, -0.00001343, 0.00002090, -0.00003349, 0.00002132, -0.00001497, 0.00001682, -0.00001349, -0.00000334, 0.00002780, -0.00004497, 0.00004000, -0.00003130, -0.00000394, 0.00004404, -0.00005384, 0.00003361, -0.00000178, -0.00001626, 0.00002000, -0.00001243, 0.00000442, -0.00001563, 0.00000892, 0.00001206, -0.00000275, -0.00000930, -0.00000680, 0.00001626, -0.00000867, 0.00000427, -0.00000107, -0.00000387, 0.00000959, -0.00000974, 0.00000402, -0.00000122, 0.00000289, 0.00000545, -0.00000704, 0.00000419, -0.00000361, 0.00000085, -0.00000083, -0.00000163, 0.00000925, -0.00000579, 0.00000531, 0.00000581, -0.00001187, -0.00000182, -0.00000508, 0.00000645, 0.00001421, -0.00000768, -0.00000602, -0.00000383, -0.00000011, 0.00000870, 0.00000217, 0.00000043, -0.00000043, -0.00000282, -0.00000554, -0.00000452, 0.00000654, 0.00000832, -0.00000176, -0.00000358, 0.00000206, 0.00000339, 0.00000096, -0.00000575, -0.00000836, 0.00000562, 0.00001538, },
		{ 0.00001597, 0.00000253, -0.00001188, 0.00000136, 0.00000120, 0.00000210, 0.00000563, -0.00001440, -0.00001268, 0.00000323, -0.00000316, -0.00000519, 0.00000129, -0.00000472, -0.00001356, 0.00000617, 0.00000998, -0.00001442, 0.00000685, -0.00000194, -0.00000862, 0.00001960, -0.00001314, -0.00000110, 0.00000426, -0.00001051, 0.00000870, -0.00000993, -0.00000084, -0.00000520, -0.00001447, -0.00000304, -0.00001214, -0.00002951, -0.00000088, -0.00015086, -0.00018510, 0.00051016, 0.00043030, -0.00031181, -0.00006553, 0.00000629, -0.00007703, 0.00004813, -0.00007243, 0.00003575, -0.00001651, -0.00001835, 0.00002958, -0.00004954, 0.00004126, -0.00003536, 0.00000907, 0.00001379, -0.00003185, 0.00003252, -0.00003506, 0.00001381, -0.00000362, 0.00000378, 0.00000285, -0.00002013, 0.00003703, -0.00004272, 0.00002532, -0.00000601, -0.00003233, 0.00005662, -0.00004448, 0.00001037, 0.00001898, -0.00002526, 0.00001988, -0.00000758, 0.00000076, -0.00001146, 0.00000186, 0.00001497, -0.00000117, -0.00000765, -0.00001022, 0.00001577, -0.00000538, 0.00000073, 0.00000276, -0.00000702, 0.00001089, -0.00000822, 0.00000090, 0.00000110, -0.00000013, 0.00000826, -0.00000650, 0.00000264, -0.00000220, -0.00000135, 0.00000196, -0.00000415, 0.00000883, -0.00000210, 0.00000047, 0.00000842, -0.00000927, -0.00000398, -0.00000455, 0.00000188, 0.00001645, -0.00000289, -0.00000867, -0.00000409, -0.00000161, 0.00000803, 0.00000420, 0.00000025, -0.00000026, -0.00000220, -0.00000521, -0.00000563, 0.00000479, 0.00000911, -0.00000000, -0.00000399, 0.00000076, 0.00000346, 0.00000215, -0.00000384, -0.00000905, 0.00000071, },
		{ -0.00000366, 0.00001418, 0.00000826, -0.00001230, -0.00000103, 0.00000351, -0.00000089, 0.00000716, -0.00000913, -0.00001709, 0.00000141, -0.00000044, -0.00000641, 0.00000104, -0.00000287, -0.00001437, 0.00000166, 0.00001361, -0.00001268, 0.00000199, 0.00000393, -0.00001315, 0.00001856, -0.00000643, -0.00000814, 0.00000853, -0.00001121, 0.00000597, -0.00000521, -0.00000513, -0.00000170, -0.00001258, -0.00000956, -0.00000286, -0.00003361, -0.00000382, -0.00008502, -0.00027626, 0.00037529, 0.00060499, -0.00025864, -0.00015463, 0.00005425, -0.00010865, 0.00005712, -0.00006417, 0.00000994, 0.00001440, -0.00004568, 0.00004743, -0.00005460, 0.00003229, -0.00001539, -0.00001646, 0.00003555, -0.00004365, 0.00003367, -0.00002528, -0.00000240, 0.00001194, -0.00000947, 0.00001540, -0.00002848, 0.00003505, -0.00002973, 0.00000347, 0.00001917, -0.00005037, 0.00005278, -0.00002235, -0.00001536, 0.00003294, -0.00002512, 0.00001317, 0.00000012, -0.00000325, -0.00000777, -0.00000340, 0.00001535, 0.00000141, -0.00000600, -0.00001273, 0.00001384, -0.00000127, -0.00000295, 0.00000604, -0.00000888, 0.00001044, -0.00000513, -0.00000284, 0.00000333, -0.00000258, 0.00000949, -0.00000439, 0.00000045, -0.00000044, -0.00000370, 0.00000393, -0.00000468, 0.00000632, 0.00000217, -0.00000282, 0.00000845, -0.00000546, -0.00000579, -0.00000400, -0.00000196, 0.00001613, 0.00000297, -0.00000993, -0.00000538, -0.00000257, 0.00000671, 0.00000598, 0.00000073, -0.00000042, -0.00000162, -0.00000460, -0.00000656, 0.00000292, 0.00000946, 0.00000171, -0.00000382, -0.00000054, 0.00000307, 0.00000317, -0.00000210, -0.00000852, },
		{ -0.00000727, -0.00000676, 0.00001035, 0.00001330, -0.00001052, -0.00000479, 0.00000602, -0.00000249, 0.00000597, -0.00000279, -0.00001935, -0.00000261, 0.00000252, -0.00000694, 0.00000009, -0.00000080, -0.00001445, -0.00000297, 0.00001542, -0.00000894, -0.00000334, 0.00000842, -0.00001464, 0.00001423, 0.00000151, -0.00001357, 0.00000953, -0.00000867, 0.00000147, -0.00000094, -0.00000706, -0.00000210, -0.00000718, -0.00001601, 0.00000069, -0.00002642, -0.00002322, -0.00001903, -0.00031150, 0.00018720, 0.00073943, -0.00013538, -0.00026612, 0.00009445, -0.00012208, 0.00004388, -0.00003581, -0.00002859, 0.00004737, -0.00006586, 0.00005211, -0.00004422, 0.00000979, 0.00001238, -0.00004168, 0.00004943, -0.00004345, 0.00002234, -0.00000591, -0.00002256, 0.00002575, -0.00001729, 0.00001985, -0.00002648, 0.00002327, -0.00001055, -0.00001866, 0.00003681, -0.00005358, 0.00003421, 0.00000549, -0.00003564, 0.00003579, -0.00001614, 0.00000276, 0.00000750, -0.00000540, -0.00000577, -0.00000648, 0.00001371, 0.00000416, -0.00000400, -0.00001434, 0.00001064, 0.00000326, -0.00000625, 0.00000826, -0.00000911, 0.00000828, -0.00000096, -0.00000645, 0.00000478, -0.00000388, 0.00000897, -0.00000112, -0.00000176, 0.00000117, -0.00000556, 0.00000447, -0.00000330, 0.00000272, 0.00000566, -0.00000386, 0.00000643, -0.00000151, -0.00000661, -0.00000376, -0.00000467, 0.00001367, 0.00000865, -0.00000917, -0.00000742, -0.00000338, 0.00000520, 0.00000714, 0.00000183, -0.00000061, -0.00000133, -0.00000372, -0.00000722, 0.00000091, 0.00000948, 0.00000325, -0.00000322, -0.00000155, 0.00000218, 0.00000399, -0.00000056, },
		{ 0.00000089, -0.00000582, -0.00000827, 0.00000534, 0.00001655, -0.00000642, -0.00000911, 0.00000755, -0.00000190, 0.00000262, 0.00000302, -0.00001856, -0.00000830, 0.00000474, -0.00000624, -0.00000168, 0.00000136, -0.00001380, -0.00000740, 0.00001543, -0.00000404, -0.00000796, 0.00001070, -0.00001320, 0.00000792, 0.00000873, -0.00001594, 0.00000718, -0.00000423, -0.00000267, 0.00000104, -0.00000585, -0.00000531, -0.00000153, -0.00001810, -0.00000433, -0.00000959, -0.00004857, 0.00002811, -0.00028563, -0.00002539, 0.00080055, 0.00005830, -0.00037836, 0.00010714, -0.00010597, 0.00000550, 0.00000816, -0.00007035, 0.00007165, -0.00007032, 0.00003922, -0.00001895, -0.00002118, 0.00003994, -0.00005824, 0.00004949, -0.00002929, 0.00000068, 0.00001765, -0.00003990, 0.00003187, -0.00001591, 0.00001493, -0.00001585, 0.00000625, 0.00000849, -0.00003452, 0.00004223, -0.00004214, 0.00000688, 0.00003048, -0.00004446, 0.00002689, -0.00000169, -0.00000723, 0.00001173, -0.00000455, -0.00000549, -0.00000788, 0.00001102, 0.00000631, -0.00000158, -0.00001488, 0.00000635, 0.00000773, -0.00000865, 0.00000900, -0.00000762, 0.00000477, 0.00000353, -0.00000915, 0.00000494, -0.00000383, 0.00000697, 0.00000258, -0.00000337, 0.00000220, -0.00000649, 0.00000344, -0.00000062, -0.00000075, 0.00000740, -0.00000272, 0.00000338, 0.00000155, -0.00000617, -0.00000388, -0.00000626, 0.00000985, 0.00001305, -0.00000632, -0.00000957, -0.00000452, 0.00000378, 0.00000761, 0.00000326, -0.00000050, -0.00000141, -0.00000272, -0.00000742, -0.00000128, 0.00000918, 0.00000463, -0.00000241, -0.00000207, 0.00000090, 0.00000447, },
		{ 0.00000440, 0.00000239, -0.00000451, -0.00000835, 0.00000021, 0.00001725, -0.00000053, -0.00001280, 0.00000712, 0.00000089, -0.00000162, 0.00000671, -0.00001452, -0.00001445, 0.00000517, -0.00000403, -0.00000406, 0.00000324, -0.00001226, -0.00001151, 0.00001394, 0.00000112, -0.00001106, 0.00001068, -0.00000971, 0.00000127, 0.00001368, -0.00001489, 0.00000263, 0.00000010, -0.00000466, 0.00000006, -0.00000249, -0.00000872, 0.00000085, -0.00001336, -0.00001683, 0.00001048, -0.00006727, 0.00004374, -0.00020798, -0.00022532, 0.00076457, 0.00030551, -0.00046040, 0.00007384, -0.00005591, -0.00005269, 0.00005647, -0.00010245, 0.00007694, -0.00005408, 0.00000962, 0.00001544, -0.00005179, 0.00005812, -0.00005952, 0.00003363, -0.00000377, -0.00002528, 0.00003791, -0.00004786, 0.00002660, -0.00000488, 0.00000297, -0.00000122, -0.00000995, 0.00002150, -0.00003996, 0.00003486, -0.00002052, -0.00002093, 0.00004526, -0.00003957, 0.00000943, 0.00001302, -0.00001297, 0.00001137, -0.00000106, -0.00000604, -0.00000857, 0.00000821, 0.00000746, 0.00000096, -0.00001412, 0.00000127, 0.00001158, -0.00000966, 0.00000804, -0.00000462, 0.00000059, 0.00000749, -0.00001029, 0.00000360, -0.00000258, 0.00000408, 0.00000590, -0.00000391, 0.00000245, -0.00000638, 0.00000115, 0.00000240, -0.00000304, 0.00000708, -0.00000013, 0.00000049, 0.00000307, -0.00000464, -0.00000409, -0.00000697, 0.00000553, 0.00001548, -0.00000187, -0.00001100, -0.00000627, 0.00000250, 0.00000754, 0.00000466, 0.00000011, -0.00000176, -0.00000184, -0.00000703, -0.00000355, 0.00000849, 0.00000594, -0.00000164, -0.00000203, -0.00000050, },
		{ -0.00000164, 0.00000361, 0.00000395, -0.00000341, -0.00000749, -0.00000400, 0.00001527, 0.00000603, -0.00001465, 0.00000430, 0.00000511, -0.00000510, 0.00000725, -0.00000801, -0.00001943, 0.00000299, -0.00000051, -0.00000648, 0.00000428, -0.00000967, -0.00001522, 0.00001128, 0.00000587, -0.00001237, 0.00000890, -0.00000536, -0.00000433, 0.00001572, -0.00001119, -0.00000231, 0.00000254, -0.00000382, -0.00000294, 0.00000078, -0.00000945, -0.00000218, -0.00000245, -0.00003197, 0.00002512, -0.00006918, 0.00002538, -0.00010063, -0.00037533, 0.00062470, 0.00057213, -0.00047827, -0.00001580, 0.00002181, -0.00011680, 0.00009351, -0.00011236, 0.00005718, -0.00001828, -0.00002993, 0.00004910, -0.00007199, 0.00005961, -0.00004308, 0.00000478, 0.00002630, -0.00004731, 0.00004770, -0.00004243, 0.00001000, 0.00001253, -0.00001096, 0.00001152, -0.00001990, 0.00002497, -0.00003451, 0.00001810, 0.00000411, -0.00004134, 0.00004589, -0.00002311, -0.00001074, 0.00002279, -0.00001233, 0.00000693, 0.00000352, -0.00000607, -0.00000937, 0.00000577, 0.00000773, 0.00000312, -0.00001197, -0.00000406, 0.00001419, -0.00000893, 0.00000548, -0.00000070, -0.00000343, 0.00001012, -0.00000955, 0.00000098, -0.00000067, 0.00000107, 0.00000817, -0.00000318, 0.00000201, -0.00000540, -0.00000181, 0.00000483, -0.00000357, 0.00000506, 0.00000284, -0.00000129, 0.00000295, -0.00000257, -0.00000405, -0.00000713, 0.00000144, 0.00001576, 0.00000331, -0.00001101, -0.00000859, 0.00000115, 0.00000717, 0.00000572, 0.00000122, -0.00000211, -0.00000137, -0.00000605, -0.00000567, 0.00000724, 0.00000728, -0.00000099, -0.00000160, },
		{ -0.00000106, -0.00000220, 0.00000211, 0.00000540, -0.00000235, -0.00000631, -0.00000664, 0.00001114, 0.00001190, -0.00001382, -0.00000052, 0.00000938, -0.00000637, 0.00000456, -0.00000064, -0.00002163, -0.00000195, 0.00000349, -0.00000808, 0.00000394, -0.00000599, -0.00001831, 0.00000763, 0.00000989, -0.00001217, 0.00000621, -0.00000121, -0.00000818, 0.00001520, -0.00000635, -0.00000588, 0.00000230, -0.00000094, -0.00000587, 0.00000143, -0.00000575, -0.00001020, 0.00001083, -0.00004288, 0.00002681, -0.00005042, -0.00001846, 0.00000720, -0.00044733, 0.00039545, 0.00081167, -0.00040385, -0.00015791, 0.00010900, -0.00016701, 0.00010381, -0.00009241, 0.00001348, 0.00002906, -0.00006809, 0.00007085, -0.00007380, 0.00004169, -0.00001206, -0.00002948, 0.00005177, -0.00005761, 0.00004269, -0.00002381, -0.00001381, 0.00003027, -0.00002080, 0.00001734, -0.00002064, 0.00001889, -0.00002125, -0.00000206, 0.00002435, -0.00004919, 0.00003292, -0.00000072, -0.00002705, 0.00002419, -0.00000567, 0.00000058, 0.00000710, -0.00000454, -0.00001057, 0.00000367, 0.00000755, 0.00000442, -0.00000862, -0.00000895, 0.00001499, -0.00000642, 0.00000178, 0.00000334, -0.00000645, 0.00001094, -0.00000704, -0.00000237, 0.00000124, -0.00000138, 0.00000902, -0.00000137, 0.00000119, -0.00000389, -0.00000478, 0.00000597, -0.00000233, 0.00000218, 0.00000514, -0.00000150, 0.00000164, -0.00000065, -0.00000349, -0.00000696, -0.00000197, 0.00001416, 0.00000822, -0.00000924, -0.00001107, -0.00000057, 0.00000675, 0.00000634, 0.00000263, -0.00000214, -0.00000144, -0.00000465, -0.00000733, 0.00000531, 0.00000862, -0.00000039, },
		{ 0.00000036, -0.00000073, -0.00000201, 0.00000014, 0.00000642, -0.00000107, -0.00000533, -0.00000755, 0.00000587, 0.00001582, -0.00001019, -0.00000626, 0.00001216, -0.00000465, -0.00000041, 0.00000559, -0.00002002, -0.00000893, 0.00000670, -0.00000799, 0.00000197, -0.00000158, -0.00002035, 0.00000301, 0.00001314, -0.00001091, 0.00000333, 0.00000219, -0.00001035, 0.00001313, -0.00000190, -0.00000715, -0.00000009, 0.00000208, -0.00000638, -0.00000202, 0.00000196, -0.00002013, 0.00002086, -0.00004336, 0.00001239, -0.00001507, -0.00007096, 0.00008696, -0.00042995, 0.00011190, 0.00097451, -0.00022434, -0.00033124, 0.00017835, -0.00018304, 0.00007743, -0.00004336, -0.00004490, 0.00007462, -0.00009213, 0.00007189, -0.00005425, 0.00000767, 0.00002534, -0.00005899, 0.00006399, -0.00005150, 0.00002308, 0.00000338, -0.00003776, 0.00004157, -0.00002177, 0.00001394, -0.00001268, 0.00000652, -0.00000558, -0.00001919, 0.00003484, -0.00004345, 0.00001093, 0.00002036, -0.00003443, 0.00001673, 0.00000435, -0.00000472, 0.00000802, -0.00000124, -0.00001175, 0.00000147, 0.00000743, 0.00000467, -0.00000461, -0.00001261, 0.00001371, -0.00000244, -0.00000234, 0.00000666, -0.00000787, 0.00000987, -0.00000327, -0.00000569, 0.00000248, -0.00000283, 0.00000846, 0.00000109, 0.00000039, -0.00000223, -0.00000718, 0.00000551, 0.00000017, -0.00000055, 0.00000606, -0.00000027, -0.00000012, 0.00000051, -0.00000237, -0.00000652, -0.00000454, 0.00001124, 0.00001200, -0.00000581, -0.00001309, -0.00000294, 0.00000630, 0.00000658, 0.00000402, -0.00000161, -0.00000198, -0.00000316, -0.00000821, 0.00000272, 0.00000980, },
		{ 0.00001051, 0.00000150, -0.00000081, -0.00000114, -0.00000189, 0.00000667, 0.00000062, -0.00000478, -0.00000707, 0.00000067, 0.00001700, -0.00000440, -0.00001143, 0.00001228, -0.00000016, -0.00000592, 0.00000884, -0.00001456, -0.00001637, 0.00000767, -0.00000573, -0.00000138, 0.00000282, -0.00002071, -0.00000265, 0.00001565, -0.00000896, 0.00000059, 0.00000493, -0.00001145, 0.00001061, 0.00000123, -0.00000638, -0.00000296, 0.00000302, -0.00000306, -0.00000913, 0.00001100, -0.00002733, 0.00002242, -0.00003056, -0.00001529, 0.00002599, -0.00011239, 0.00011872, -0.00033215, -0.00017661, 0.00101961, 0.00005078, -0.00049862, 0.00019988, -0.00015099, 0.00001460, 0.00002420, -0.00010225, 0.00010336, -0.00009203, 0.00004906, -0.00001705, -0.00003350, 0.00005795, -0.00007409, 0.00005778, -0.00002929, -0.00000608, 0.00003131, -0.00005408, 0.00004129, -0.00001215, 0.00000257, 0.00000027, -0.00000682, 0.00000685, -0.00002858, 0.00003380, -0.00002720, -0.00001309, 0.00003366, -0.00003086, 0.00000301, 0.00001375, -0.00000643, 0.00000578, 0.00000303, -0.00001206, -0.00000129, 0.00000763, 0.00000408, -0.00000071, -0.00001442, 0.00001042, 0.00000233, -0.00000598, 0.00000854, -0.00000744, 0.00000726, 0.00000100, -0.00000821, 0.00000265, -0.00000319, 0.00000680, 0.00000364, -0.00000002, -0.00000072, -0.00000866, 0.00000357, 0.00000314, -0.00000227, 0.00000541, 0.00000179, -0.00000150, 0.00000062, -0.00000092, -0.00000576, -0.00000627, 0.00000767, 0.00001412, -0.00000123, -0.00001396, -0.00000599, 0.00000568, 0.00000667, 0.00000508, -0.00000046, -0.00000273, -0.00000199, -0.00000814, -0.00000032, },
		{ -0.00000342, 0.00001040, 0.00000318, -0.00000129, 0.00000010, -0.00000345, 0.00000590, 0.00000271, -0.00000456, -0.00000585, -0.00000341, 0.00001538, 0.00000229, -0.00001459, 0.00000935, 0.00000592, -0.00000993, 0.00000808, -0.00000640, -0.00002221, 0.00000536, -0.00000149, -0.00000524, 0.00000617, -0.00001884, -0.00000921, 0.00001725, -0.00000636, -0.00000216, 0.00000742, -0.00001211, 0.00000828, 0.00000308, -0.00000483, -0.00000432, 0.00000047, 0.00000375, -0.00001740, 0.00001726, -0.00002736, 0.00001294, -0.00000651, -0.00004784, 0.00005864, -0.00012654, 0.00009673, -0.00018157, -0.00041708, 0.00092562, 0.00038749, -0.00061365, 0.00014988, -0.00006958, -0.00007227, 0.00009163, -0.00014056, 0.00010332, -0.00006426, 0.00000665, 0.00002796, -0.00006967, 0.00007514, -0.00006889, 0.00003339, 0.00000346, -0.00003626, 0.00005132, -0.00005686, 0.00002783, 0.00000590, -0.00001238, 0.00001276, -0.00001582, 0.00001206, -0.00002863, 0.00002320, -0.00000634, -0.00003207, 0.00003546, -0.00001797, -0.00001227, 0.00001869, -0.00000344, 0.00000132, 0.00000680, -0.00001066, -0.00000473, 0.00000799, 0.00000318, 0.00000231, -0.00001410, 0.00000566, 0.00000697, -0.00000835, 0.00000860, -0.00000538, 0.00000377, 0.00000492, -0.00000938, 0.00000166, -0.00000268, 0.00000456, 0.00000575, 0.00000015, 0.00000053, -0.00000909, 0.00000056, 0.00000572, -0.00000247, 0.00000355, 0.00000385, -0.00000193, -0.00000022, 0.00000044, -0.00000461, -0.00000726, 0.00000404, 0.00001445, 0.00000371, -0.00001321, -0.00000944, 0.00000456, 0.00000684, 0.00000566, 0.00000114, -0.00000331, -0.00000145, -0.00000717, },
		{ -0.00000557, -0.00000606, 0.00000921, 0.00000534, -0.00000195, 0.00000126, -0.00000415, 0.00000412, 0.00000494, -0.00000435, -0.00000459, -0.00000577, 0.00001158, 0.00000838, -0.00001483, 0.00000392, 0.00001180, -0.00001080, 0.00000352, 0.00000236, -0.00002452, -0.00000044, 0.00000370, -0.00000836, 0.00000746, -0.00001455, -0.00001608, 0.00001742, -0.00000283, -0.00000527, 0.00001004, -0.00001255, 0.00000600, 0.00000454, -0.00000415, -0.00000281, -0.00000538, 0.00001178, -0.00002305, 0.00001690, -0.00001808, -0.00000606, 0.00002204, -0.00007391, 0.00007041, -0.00010608, 0.00003100, -0.00001771, -0.00056525, 0.00069836, 0.00073132, -0.00063144, 0.00001992, 0.00004645, -0.00016028, 0.00013706, -0.00014514, 0.00007002, -0.00001384, -0.00004413, 0.00006733, -0.00008915, 0.00007026, -0.00004345, -0.00000308, 0.00003737, -0.00005804, 0.00005678, -0.00004415, 0.00000394, 0.00002693, -0.00002492, 0.00001956, -0.00001703, 0.00000907, -0.00002108, 0.00000773, 0.00001255, -0.00004104, 0.00002585, -0.00000033, -0.00002395, 0.00001690, 0.00000341, -0.00000336, 0.00000855, -0.00000723, -0.00000846, 0.00000800, 0.00000261, 0.00000391, -0.00001186, 0.00000031, 0.00001056, -0.00000892, 0.00000691, -0.00000227, 0.00000024, 0.00000779, -0.00000900, -0.00000026, -0.00000173, 0.00000227, 0.00000707, 0.00000086, 0.00000151, -0.00000855, -0.00000295, 0.00000721, -0.00000115, 0.00000117, 0.00000518, -0.00000124, -0.00000157, 0.00000128, -0.00000308, -0.00000760, 0.00000076, 0.00001322, 0.00000817, -0.00001070, -0.00001273, 0.00000263, 0.00000716, 0.00000581, 0.00000287, -0.00000334, -0.00000163, },
		{ -0.00000241, -0.00000381, -0.00000781, 0.00000691, 0.00000770, -0.00000239, 0.00000190, -0.00000378, 0.00000160, 0.00000686, -0.00000376, -0.00000384, -0.00000634, 0.00000669, 0.00001258, -0.00001205, -0.00000265, 0.00001564, -0.00000783, -0.00000344, 0.00000925, -0.00002211, -0.00000880, 0.00000816, -0.00000941, 0.00000614, -0.00000830, -0.00002223, 0.00001543, 0.00000196, -0.00000893, 0.00001267, -0.00001228, 0.00000298, 0.00000681, -0.00000545, 0.00000155, -0.00001257, 0.00001754, -0.00002252, 0.00000805, -0.00000083, -0.00002900, 0.00004564, -0.00008311, 0.00005509, -0.00005521, -0.00005580, 0.00011820, -0.00059653, 0.00037208, 0.00101755, -0.00052113, -0.00017695, 0.00016754, -0.00022169, 0.00014208, -0.00010995, 0.00000903, 0.00004623, -0.00008811, 0.00008819, -0.00008460, 0.00004315, -0.00000422, -0.00004137, 0.00006209, -0.00006417, 0.00004533, -0.00001881, -0.00002398, 0.00004396, -0.00002963, 0.00001759, -0.00001007, 0.00000005, -0.00001012, -0.00000702, 0.00002429, -0.00003853, 0.00000840, 0.00001611, -0.00002812, 0.00000852, 0.00001161, -0.00000592, 0.00000738, -0.00000225, -0.00001158, 0.00000704, 0.00000285, 0.00000403, -0.00000834, -0.00000463, 0.00001240, -0.00000762, 0.00000396, 0.00000110, -0.00000260, 0.00000922, -0.00000722, -0.00000261, -0.00000081, 0.00000033, 0.00000749, 0.00000196, 0.00000232, -0.00000721, -0.00000637, 0.00000723, 0.00000130, -0.00000091, 0.00000533, 0.00000033, -0.00000284, 0.00000132, -0.00000134, -0.00000733, -0.00000193, 0.00001086, 0.00001146, -0.00000668, -0.00001520, -0.00000028, 0.00000752, 0.00000576, 0.00000434, -0.00000262, },
		{ -0.00000116, -0.00000341, -0.00000239, -0.00000840, 0.00000373, 0.00000980, -0.00000222, 0.00000177, -0.00000245, -0.00000113, 0.00000801, -0.00000249, -0.00000377, -0.00000558, 0.00000193, 0.00001414, -0.00000701, -0.00000862, 0.00001614, -0.00000156, -0.00001050, 0.00001216, -0.00001512, -0.00001774, 0.00001004, -0.00000759, 0.00000243, -0.00000126, -0.00002630, 0.00001067, 0.00000801, -0.00001279, 0.00001455, -0.00001033, -0.00000171, 0.00001058, -0.00000867, 0.00000711, -0.00001796, 0.00001767, -0.00001405, -0.00000801, 0.00001953, -0.00004792, 0.00005542, -0.00006986, 0.00001527, 0.00001147, -0.00013519, 0.00019399, -0.00051234, 0.00000359, 0.00118545, -0.00027659, -0.00040520, 0.00025547, -0.00023202, 0.00009835, -0.00004080, -0.00006460, 0.00009825, -0.00011071, 0.00008244, -0.00005573, 0.00000064, 0.00003767, -0.00007001, 0.00006960, -0.00005200, 0.00001983, 0.00001221, -0.00004784, 0.00005080, -0.00002361, 0.00000706, 0.00000232, -0.00001067, -0.00000065, -0.00001651, 0.00002661, -0.00002663, -0.00001112, 0.00002608, -0.00002350, -0.00000392, 0.00001786, -0.00000472, 0.00000353, 0.00000303, -0.00001306, 0.00000471, 0.00000395, 0.00000303, -0.00000447, -0.00000826, 0.00001221, -0.00000480, 0.00000049, 0.00000398, -0.00000427, 0.00000921, -0.00000447, -0.00000487, -0.00000028, -0.00000107, 0.00000712, 0.00000317, 0.00000301, -0.00000530, -0.00000921, 0.00000572, 0.00000421, -0.00000202, 0.00000430, 0.00000223, -0.00000350, 0.00000055, 0.00000030, -0.00000648, -0.00000393, 0.00000791, 0.00001321, -0.00000176, -0.00001620, -0.00000405, 0.00000759, 0.00000580, 0.00000527, },
		{ 0.00000558, 0.00000077, -0.00000418, -0.00000167, -0.00000784, 0.00000015, 0.00001111, -0.00000116, 0.00000086, -0.00000054, -0.00000350, 0.00000805, -0.00000055, -0.00000422, -0.00000419, -0.00000171, 0.00001310, -0.00000106, -0.00001244, 0.00001302, 0.00000639, -0.00001521, 0.00001006, -0.00000508, -0.00002473, 0.00000796, -0.00000301, -0.00000259, 0.00000493, -0.00002706, 0.00000301, 0.00001466, -0.00001586, 0.00001447, -0.00000571, -0.00000842, 0.00001548, -0.00001236, 0.00001114, -0.00001843, 0.00001041, 0.00000124, -0.00002680, 0.00003599, -0.00005519, 0.00004635, -0.00003578, -0.00003796, 0.00007348, -0.00018149, 0.00019487, -0.00033994, -0.00033966, 0.00119307, 0.00007785, -0.00061230, 0.00027348, -0.00017866, 0.00001183, 0.00004497, -0.00013006, 0.00012510, -0.00010274, 0.00004978, -0.00000990, -0.00004526, 0.00006966, -0.00008001, 0.00005684, -0.00002450, -0.00001240, 0.00004001, -0.00006038, 0.00004421, -0.00000763, -0.00000859, 0.00001534, -0.00001826, 0.00000351, -0.00001872, 0.00002053, -0.00001004, -0.00002679, 0.00002661, -0.00001178, -0.00001637, 0.00001932, 0.00000051, -0.00000161, 0.00000702, -0.00001219, 0.00000111, 0.00000551, 0.00000161, -0.00000115, -0.00001008, 0.00001021, -0.00000118, -0.00000271, 0.00000579, -0.00000464, 0.00000804, -0.00000131, -0.00000656, -0.00000030, -0.00000194, 0.00000617, 0.00000426, 0.00000361, -0.00000301, -0.00001111, 0.00000291, 0.00000684, -0.00000181, 0.00000245, 0.00000382, -0.00000326, -0.00000078, 0.00000150, -0.00000514, -0.00000521, 0.00000484, 0.00001335, 0.00000329, -0.00001537, -0.00000826, 0.00000697, 0.00000616, },
		{ 0.00000687, 0.00000542, 0.00000274, -0.00000431, -0.00000178, -0.00000644, -0.00000323, 0.00001125, 0.00000080, -0.00000055, 0.00000141, -0.00000499, 0.00000690, 0.00000184, -0.00000478, -0.00000288, -0.00000373, 0.00001014, 0.00000428, -0.00001322, 0.00000711, 0.00001375, -0.00001573, 0.00000345, 0.00000537, -0.00002743, 0.00000162, 0.00000312, -0.00000718, 0.00000865, -0.00002388, -0.00000681, 0.00002056, -0.00001671, 0.00001126, 0.00000191, -0.00001645, 0.00001991, -0.00001410, 0.00001090, -0.00001227, -0.00000335, 0.00001927, -0.00004191, 0.00004184, -0.00004642, 0.00001943, 0.00001029, -0.00008824, 0.00011100, -0.00017958, 0.00012696, -0.00012564, -0.00059482, 0.00102873, 0.00048902, -0.00073991, 0.00019765, -0.00006692, -0.00009703, 0.00012259, -0.00016677, 0.00011579, -0.00006393, -0.00000154, 0.00003978, -0.00008087, 0.00008172, -0.00006775, 0.00002695, 0.00001046, -0.00004183, 0.00005648, -0.00005761, 0.00002508, 0.00001404, -0.00002390, 0.00002391, -0.00001919, 0.00000098, -0.00001453, 0.00000960, 0.00000571, -0.00003437, 0.00001788, 0.00000303, -0.00002465, 0.00001469, 0.00000841, -0.00000595, 0.00000838, -0.00000891, -0.00000308, 0.00000684, 0.00000051, 0.00000100, -0.00001006, 0.00000700, 0.00000245, -0.00000499, 0.00000632, -0.00000391, 0.00000618, 0.00000173, -0.00000744, -0.00000084, -0.00000243, 0.00000489, 0.00000510, 0.00000409, -0.00000054, -0.00001187, -0.00000078, 0.00000857, -0.00000027, 0.00000039, 0.00000460, -0.00000218, -0.00000222, 0.00000200, -0.00000349, -0.00000577, 0.00000202, 0.00001211, 0.00000771, -0.00001269, -0.00001226, 0.00000529, },
		{ 0.00000243, 0.00000773, 0.00000510, 0.00000435, -0.00000358, -0.00000256, -0.00000469, -0.00000586, 0.00001004, 0.00000339, -0.00000200, 0.00000288, -0.00000534, 0.00000480, 0.00000421, -0.00000504, -0.00000215, -0.00000418, 0.00000634, 0.00000781, -0.00001101, 0.00000008, 0.00001839, -0.00001147, -0.00000572, 0.00001324, -0.00002449, -0.00000789, 0.00000875, -0.00000955, 0.00000887, -0.00001716, -0.00001713, 0.00002390, -0.00001405, 0.00000448, 0.00001171, -0.00002385, 0.00002147, -0.00001155, 0.00000489, -0.00000018, -0.00001994, 0.00003410, -0.00004717, 0.00003320, -0.00002240, -0.00001811, 0.00005480, -0.00011913, 0.00011131, -0.00012932, 0.00001491, 0.00007700, -0.00071828, 0.00071589, 0.00088424, -0.00073848, 0.00002645, 0.00007851, -0.00019776, 0.00016691, -0.00016098, 0.00006954, -0.00000369, -0.00005685, 0.00007831, -0.00009514, 0.00006960, -0.00003634, -0.00001161, 0.00004258, -0.00005968, 0.00005687, -0.00004003, -0.00000175, 0.00003498, -0.00003329, 0.00002448, -0.00001254, -0.00000686, -0.00000711, -0.00000159, 0.00001612, -0.00003249, 0.00000301, 0.00001598, -0.00002595, 0.00000479, 0.00001635, -0.00000744, 0.00000659, -0.00000401, -0.00000679, 0.00000719, 0.00000024, 0.00000180, -0.00000862, 0.00000339, 0.00000536, -0.00000605, 0.00000568, -0.00000249, 0.00000411, 0.00000428, -0.00000746, -0.00000173, -0.00000268, 0.00000345, 0.00000565, 0.00000439, 0.00000188, -0.00001141, -0.00000482, 0.00000895, 0.00000224, -0.00000123, 0.00000435, -0.00000058, -0.00000331, 0.00000173, -0.00000181, -0.00000566, -0.00000029, 0.00000988, 0.00001091, -0.00000851, -0.00001532, },
		{ -0.00001683, -0.00000146, 0.00000833, 0.00000498, 0.00000531, -0.00000204, -0.00000365, -0.00000312, -0.00000737, 0.00000765, 0.00000614, -0.00000301, 0.00000354, -0.00000459, 0.00000220, 0.00000612, -0.00000469, -0.00000212, -0.00000353, 0.00000283, 0.00000898, -0.00000676, -0.00000611, 0.00001896, -0.00000335, -0.00001459, 0.00001610, -0.00001618, -0.00001819, 0.00001158, -0.00000848, 0.00000565, -0.00000838, -0.00002575, 0.00002302, -0.00000732, -0.00000513, 0.00002161, -0.00002798, 0.00001792, -0.00000353, -0.00000622, 0.00001461, -0.00003392, 0.00003990, -0.00003894, 0.00001084, 0.00001061, -0.00005484, 0.00008372, -0.00011952, 0.00007306, -0.00004550, -0.00010563, 0.00022079, -0.00069583, 0.00030952, 0.00118654, -0.00058141, -0.00021492, 0.00021803, -0.00025821, 0.00016037, -0.00011085, -0.00000300, 0.00006107, -0.00009933, 0.00009350, -0.00008321, 0.00003635, 0.00000522, -0.00004759, 0.00006230, -0.00006071, 0.00004132, -0.00001259, -0.00002896, 0.00004872, -0.00003299, 0.00001634, -0.00000018, -0.00001653, -0.00000049, -0.00000912, 0.00001903, -0.00002282, -0.00001316, 0.00002282, -0.00001973, -0.00000770, 0.00002138, -0.00000491, 0.00000228, 0.00000107, -0.00000893, 0.00000618, 0.00000089, 0.00000150, -0.00000646, 0.00000013, 0.00000716, -0.00000590, 0.00000424, -0.00000083, 0.00000221, 0.00000617, -0.00000672, -0.00000276, -0.00000282, 0.00000194, 0.00000594, 0.00000449, 0.00000402, -0.00000978, -0.00000867, 0.00000783, 0.00000518, -0.00000189, 0.00000320, 0.00000103, -0.00000374, 0.00000086, -0.00000040, -0.00000497, -0.00000194, 0.00000713, 0.00001258, -0.00000345, },
		{ 0.00000172, -0.00001645, -0.00000591, 0.00000823, 0.00000529, 0.00000556, 0.00000001, -0.00000456, -0.00000214, -0.00000770, 0.00000448, 0.00000849, -0.00000320, 0.00000328, -0.00000303, -0.00000038, 0.00000722, -0.00000368, -0.00000262, -0.00000245, 0.00000037, 0.00000803, -0.00000194, -0.00000983, 0.00001537, 0.00000648, -0.00002025, 0.00001294, -0.00000441, -0.00002629, 0.00000989, -0.00000395, 0.00000023, 0.00000024, -0.00003051, 0.00001711, 0.00000274, -0.00001551, 0.00002876, -0.00002635, 0.00000822, 0.00000909, -0.00001943, 0.00002716, -0.00003984, 0.00003316, -0.00001784, -0.00001958, 0.00004263, -0.00007868, 0.00008721, -0.00008746, 0.00000696, 0.00004685, -0.00019847, 0.00027653, -0.00054563, -0.00011540, 0.00133239, -0.00027488, -0.00047714, 0.00030679, -0.00025443, 0.00009934, -0.00002806, -0.00008231, 0.00011089, -0.00011528, 0.00007997, -0.00004817, -0.00000850, 0.00004472, -0.00007043, 0.00006381, -0.00004484, 0.00001469, 0.00001693, -0.00004907, 0.00005083, -0.00002235, 0.00000177, 0.00001387, -0.00002380, 0.00000196, -0.00001104, 0.00001503, -0.00000923, -0.00002570, 0.00002142, -0.00000796, -0.00001902, 0.00002125, 0.00000140, -0.00000294, 0.00000474, -0.00000885, 0.00000391, 0.00000211, 0.00000066, -0.00000431, -0.00000226, 0.00000778, -0.00000484, 0.00000243, 0.00000069, 0.00000065, 0.00000740, -0.00000540, -0.00000377, -0.00000288, 0.00000040, 0.00000598, 0.00000446, 0.00000565, -0.00000715, -0.00001181, 0.00000527, 0.00000792, -0.00000136, 0.00000154, 0.00000219, -0.00000342, -0.00000029, 0.00000048, -0.00000391, -0.00000286, 0.00000431, 0.00001267, },
		{ 0.00001140, 0.00000628, -0.00001418, -0.00001024, 0.00000703, 0.00000606, 0.00000529, 0.00000215, -0.00000488, -0.00000193, -0.00000705, 0.00000108, 0.00000998, -0.00000245, 0.00000230, -0.00000113, -0.00000248, 0.00000741, -0.00000218, -0.00000332, -0.00000151, -0.00000077, 0.00000577, 0.00000194, -0.00001033, 0.00000878, 0.00001530, -0.00002067, 0.00000452, 0.00000765, -0.00002943, 0.00000325, 0.00000272, -0.00000525, 0.00000636, -0.00003009, 0.00000670, 0.00001408, -0.00002379, 0.00003041, -0.00001773, -0.00000661, 0.00002327, -0.00003022, 0.00003248, -0.00003430, 0.00001425, 0.00001097, -0.00004898, 0.00006304, -0.00008099, 0.00006302, -0.00003116, -0.00006743, 0.00012026, -0.00023722, 0.00023963, -0.00031298, -0.00048053, 0.00128718, 0.00013989, -0.00069662, 0.00030728, -0.00017926, -0.00000326, 0.00006468, -0.00014536, 0.00012967, -0.00009860, 0.00004107, -0.00000025, -0.00005180, 0.00007055, -0.00007356, 0.00004687, -0.00001712, -0.00001495, 0.00004035, -0.00005670, 0.00004034, -0.00000408, -0.00001471, 0.00002495, -0.00002532, -0.00000120, -0.00000782, 0.00000692, 0.00000368, -0.00003117, 0.00001232, 0.00000558, -0.00002564, 0.00001527, 0.00000983, -0.00000696, 0.00000578, -0.00000659, 0.00000102, 0.00000330, -0.00000014, -0.00000268, -0.00000365, 0.00000746, -0.00000329, 0.00000064, 0.00000189, -0.00000053, 0.00000804, -0.00000368, -0.00000469, -0.00000284, -0.00000112, 0.00000576, 0.00000442, 0.00000662, -0.00000384, -0.00001381, 0.00000157, 0.00000988, 0.00000032, -0.00000011, 0.00000259, -0.00000255, -0.00000134, 0.00000074, -0.00000274, -0.00000306, 0.00000180, },
		{ -0.00000012, 0.00000914, 0.00000968, -0.00001036, -0.00001373, 0.00000460, 0.00000708, 0.00000480, 0.00000395, -0.00000438, -0.00000238, -0.00000582, -0.00000201, 0.00001035, -0.00000088, 0.00000094, 0.00000070, -0.00000385, 0.00000680, -0.00000050, -0.00000390, -0.00000101, -0.00000082, 0.00000329, 0.00000387, -0.00000792, 0.00000124, 0.00002066, -0.00001538, -0.00000668, 0.00001652, -0.00002614, -0.00000707, 0.00000911, -0.00000849, 0.00000842, -0.00002457, -0.00000621, 0.00002373, -0.00002714, 0.00002502, -0.00000292, -0.00002342, 0.00003454, -0.00003397, 0.00002738, -0.00001743, -0.00001223, 0.00003912, -0.00006765, 0.00006417, -0.00005965, 0.00001761, 0.00003341, -0.00012766, 0.00015317, -0.00021241, 0.00012991, -0.00005846, -0.00072204, 0.00105448, 0.00059319, -0.00081043, 0.00020134, -0.00004659, -0.00012038, 0.00013986, -0.00017269, 0.00010997, -0.00005305, -0.00001192, 0.00004627, -0.00008083, 0.00007527, -0.00005641, 0.00001681, 0.00001394, -0.00003898, 0.00005157, -0.00005023, 0.00002006, 0.00001671, -0.00002792, 0.00002934, -0.00001979, -0.00000908, -0.00000190, -0.00000156, 0.00001214, -0.00002861, -0.00000151, 0.00001655, -0.00002544, 0.00000468, 0.00001773, -0.00000794, 0.00000389, -0.00000297, -0.00000154, 0.00000390, -0.00000051, -0.00000176, -0.00000420, 0.00000660, -0.00000161, -0.00000092, 0.00000272, -0.00000138, 0.00000815, -0.00000164, -0.00000549, -0.00000269, -0.00000254, 0.00000519, 0.00000449, 0.00000686, -0.00000025, -0.00001437, -0.00000285, 0.00001063, 0.00000286, -0.00000125, 0.00000220, -0.00000147, -0.00000199, 0.00000048, -0.00000174, -0.00000267, },
		{ -0.00000188, -0.00000125, 0.00000634, 0.00001157, -0.00000558, -0.00001579, 0.00000109, 0.00000795, 0.00000447, 0.00000513, -0.00000310, -0.00000321, -0.00000447, -0.00000438, 0.00000958, 0.00000119, -0.00000043, 0.00000213, -0.00000445, 0.00000567, 0.00000108, -0.00000419, -0.00000096, -0.00000032, 0.00000154, 0.00000361, -0.00000383, -0.00000502, 0.00002118, -0.00000567, -0.00001725, 0.00001948, -0.00001680, -0.00001833, 0.00001250, -0.00000790, 0.00000618, -0.00001554, -0.00001863, 0.00002874, -0.00002381, 0.00001299, 0.00001508, -0.00003760, 0.00003840, -0.00002778, 0.00001191, 0.00000672, -0.00003861, 0.00005757, -0.00006852, 0.00004413, -0.00002007, -0.00003568, 0.00008733, -0.00015541, 0.00013646, -0.00013352, -0.00001522, 0.00015745, -0.00080474, 0.00067635, 0.00100107, -0.00077260, -0.00000194, 0.00011019, -0.00021795, 0.00017342, -0.00015462, 0.00005572, 0.00000851, -0.00006315, 0.00007752, -0.00008732, 0.00005789, -0.00002450, -0.00001721, 0.00003914, -0.00005086, 0.00004837, -0.00003211, -0.00000444, 0.00003413, -0.00003374, 0.00002557, -0.00000833, -0.00001888, 0.00000336, -0.00000712, 0.00001436, -0.00001959, -0.00001578, 0.00002146, -0.00001845, -0.00000774, 0.00002237, -0.00000499, -0.00000017, 0.00000065, -0.00000287, 0.00000364, -0.00000038, -0.00000142, -0.00000427, 0.00000556, -0.00000001, -0.00000217, 0.00000324, -0.00000195, 0.00000775, 0.00000065, -0.00000613, -0.00000252, -0.00000368, 0.00000417, 0.00000476, 0.00000649, 0.00000313, -0.00001336, -0.00000742, 0.00000990, 0.00000581, -0.00000152, 0.00000121, -0.00000058, -0.00000211, -0.00000006, -0.00000115, },
		{ -0.00000105, -0.00000100, -0.00000153, 0.00000348, 0.00001190, -0.00000055, -0.00001608, -0.00000307, 0.00000823, 0.00000452, 0.00000562, -0.00000126, -0.00000402, -0.00000336, -0.00000587, 0.00000789, 0.00000338, -0.00000148, 0.00000304, -0.00000441, 0.00000430, 0.00000241, -0.00000424, -0.00000114, 0.00000016, 0.00000099, 0.00000171, 0.00000023, -0.00000831, 0.00001698, 0.00000581, -0.00002388, 0.00001554, -0.00000371, -0.00002709, 0.00001086, -0.00000328, 0.00000097, -0.00000573, -0.00002742, 0.00002717, -0.00001397, -0.00000310, 0.00003175, -0.00004449, 0.00003201, -0.00001179, -0.00001024, 0.00003116, -0.00005645, 0.00005960, -0.00004977, 0.00000778, 0.00002622, -0.00008071, 0.00011464, -0.00014209, 0.00007593, -0.00002535, -0.00015194, 0.00028917, -0.00072921, 0.00022435, 0.00128397, -0.00056755, -0.00026526, 0.00024603, -0.00026478, 0.00015245, -0.00009480, -0.00001847, 0.00006795, -0.00009713, 0.00008433, -0.00006991, 0.00002408, 0.00001236, -0.00004518, 0.00005144, -0.00004807, 0.00003284, -0.00000794, -0.00002662, 0.00004341, -0.00003043, 0.00001476, 0.00000590, -0.00002701, 0.00000508, -0.00000807, 0.00001082, -0.00000751, -0.00002630, 0.00001881, -0.00000683, -0.00001845, 0.00002187, 0.00000143, -0.00000479, 0.00000286, -0.00000254, 0.00000266, -0.00000000, -0.00000137, -0.00000420, 0.00000452, 0.00000146, -0.00000317, 0.00000349, -0.00000221, 0.00000681, 0.00000313, -0.00000654, -0.00000246, -0.00000436, 0.00000267, 0.00000518, 0.00000576, 0.00000585, -0.00001088, -0.00001152, 0.00000766, 0.00000863, -0.00000075, -0.00000001, -0.00000017, -0.00000181, -0.00000055, },
		{ -0.00000075, -0.00000135, -0.00000035, -0.00000105, 0.00000100, 0.00001082, 0.00000405, -0.00001457, -0.00000728, 0.00000758, 0.00000501, 0.00000555, 0.00000078, -0.00000447, -0.00000269, -0.00000653, 0.00000563, 0.00000532, -0.00000208, 0.00000345, -0.00000393, 0.00000288, 0.00000351, -0.00000419, -0.00000131, 0.00000027, 0.00000154, -0.00000073, 0.00000275, -0.00000803, 0.00000958, 0.00001590, -0.00002439, 0.00000579, 0.00000946, -0.00003027, 0.00000370, 0.00000389, -0.00000471, 0.00000187, -0.00003037, 0.00001891, 0.00000016, -0.00001908, 0.00004234, -0.00004094, 0.00001543, 0.00001047, -0.00003265, 0.00004808, -0.00005933, 0.00004326, -0.00001574, -0.00003449, 0.00006510, -0.00010359, 0.00010737, -0.00009149, -0.00000970, 0.00007992, -0.00024242, 0.00031614, -0.00052994, -0.00021664, 0.00138512, -0.00021664, -0.00052994, 0.00031614, -0.00024242, 0.00007992, -0.00000970, -0.00009149, 0.00010737, -0.00010359, 0.00006510, -0.00003449, -0.00001574, 0.00004326, -0.00005933, 0.00004808, -0.00003265, 0.00001047, 0.00001543, -0.00004094, 0.00004234, -0.00001908, 0.00000016, 0.00001891, -0.00003037, 0.00000187, -0.00000471, 0.00000389, 0.00000370, -0.00003027, 0.00000946, 0.00000579, -0.00002439, 0.00001590, 0.00000958, -0.00000803, 0.00000275, -0.00000073, 0.00000154, 0.00000027, -0.00000131, -0.00000419, 0.00000351, 0.00000288, -0.00000393, 0.00000345, -0.00000208, 0.00000532, 0.00000563, -0.00000653, -0.00000269, -0.00000447, 0.00000078, 0.00000555, 0.00000501, 0.00000758, -0.00000728, -0.00001457, 0.00000405, 0.00001082, 0.00000100, -0.00000105, -0.00000035, -0.00000135, },
		{ -0.00000105, -0.00000055, -0.00000181, -0.00000017, -0.00000001, -0.00000075, 0.00000863, 0.00000766, -0.00001152, -0.00001088, 0.00000585, 0.00000576, 0.00000518, 0.00000267, -0.00000436, -0.00000246, -0.00000654, 0.00000313, 0.00000681, -0.00000221, 0.00000349, -0.00000317, 0.00000146, 0.00000452, -0.00000420, -0.00000137, -0.00000000, 0.00000266, -0.00000254, 0.00000286, -0.00000479, 0.00000143, 0.00002187, -0.00001845, -0.00000683, 0.00001881, -0.00002630, -0.00000751, 0.00001082, -0.00000807, 0.00000508, -0.00002701, 0.00000590, 0.00001476, -0.00003043, 0.00004341, -0.00002662, -0.00000794, 0.00003284, -0.00004807, 0.00005144, -0.00004518, 0.00001236, 0.00002408, -0.00006991, 0.00008433, -0.00009713, 0.00006795, -0.00001847, -0.00009480, 0.00015245, -0.00026478, 0.00024603, -0.00026526, -0.00056755, 0.00128397, 0.00022435, -0.00072921, 0.00028917, -0.00015194, -0.00002535, 0.00007593, -0.00014209, 0.00011464, -0.00008071, 0.00002622, 0.00000778, -0.00004977, 0.00005960, -0.00005645, 0.00003116, -0.00001024, -0.00001179, 0.00003201, -0.00004449, 0.00003175, -0.00000310, -0.00001397, 0.00002717, -0.00002742, -0.00000573, 0.00000097, -0.00000328, 0.00001086, -0.00002709, -0.00000371, 0.00001554, -0.00002388, 0.00000581, 0.00001698, -0.00000831, 0.00000023, 0.00000171, 0.00000099, 0.00000016, -0.00000114, -0.00000424, 0.00000241, 0.00000430, -0.00000441, 0.00000304, -0.00000148, 0.00000338, 0.00000789, -0.00000587, -0.00000336, -0.00000402, -0.00000126, 0.00000562, 0.00000452, 0.00000823, -0.00000307, -0.00001608, -0.00000055, 0.00001190, 0.00000348, -0.00000153, -0.00000100, },
		{ -0.00000188, -0.00000115, -0.00000006, -0.00000211, -0.00000058, 0.00000121, -0.00000152, 0.00000581, 0.00000990, -0.00000742, -0.00001336, 0.00000313, 0.00000649, 0.00000476, 0.00000417, -0.00000368, -0.00000252, -0.00000613, 0.00000065, 0.00000775, -0.00000195, 0.00000324, -0.00000217, -0.00000001, 0.00000556, -0.00000427, -0.00000142, -0.00000038, 0.00000364, -0.00000287, 0.00000065, -0.00000017, -0.00000499, 0.00002237, -0.00000774, -0.00001845, 0.00002146, -0.00001578, -0.00001959, 0.00001436, -0.00000712, 0.00000336, -0.00001888, -0.00000833, 0.00002557, -0.00003374, 0.00003413, -0.00000444, -0.00003211, 0.00004837, -0.00005086, 0.00003914, -0.00001721, -0.00002450, 0.00005789, -0.00008732, 0.00007752, -0.00006315, 0.00000851, 0.00005572, -0.00015462, 0.00017342, -0.00021795, 0.00011019, -0.00000194, -0.00077260, 0.00100107, 0.00067635, -0.00080474, 0.00015745, -0.00001522, -0.00013352, 0.00013646, -0.00015541, 0.00008733, -0.00003568, -0.00002007, 0.00004413, -0.00006852, 0.00005757, -0.00003861, 0.00000672, 0.00001191, -0.00002778, 0.00003840, -0.00003760, 0.00001508, 0.00001299, -0.00002381, 0.00002874, -0.00001863, -0.00001554, 0.00000618, -0.00000790, 0.00001250, -0.00001833, -0.00001680, 0.00001948, -0.00001725, -0.00000567, 0.00002118, -0.00000502, -0.00000383, 0.00000361, 0.00000154, -0.00000032, -0.00000096, -0.00000419, 0.00000108, 0.00000567, -0.00000445, 0.00000213, -0.00000043, 0.00000119, 0.00000958, -0.00000438, -0.00000447, -0.00000321, -0.00000310, 0.00000513, 0.00000447, 0.00000795, 0.00000109, -0.00001579, -0.00000558, 0.00001157, 0.00000634, -0.00000125, },
		{ -0.00000012, -0.00000267, -0.00000174, 0.00000048, -0.00000199, -0.00000147, 0.00000220, -0.00000125, 0.00000286, 0.00001063, -0.00000285, -0.00001437, -0.00000025, 0.00000686, 0.00000449, 0.00000519, -0.00000254, -0.00000269, -0.00000549, -0.00000164, 0.00000815, -0.00000138, 0.00000272, -0.00000092, -0.00000161, 0.00000660, -0.00000420, -0.00000176, -0.00000051, 0.00000390, -0.00000154, -0.00000297, 0.00000389, -0.00000794, 0.00001773, 0.00000468, -0.00002544, 0.00001655, -0.00000151, -0.00002861, 0.00001214, -0.00000156, -0.00000190, -0.00000908, -0.00001979, 0.00002934, -0.00002792, 0.00001671, 0.00002006, -0.00005023, 0.00005157, -0.00003898, 0.00001394, 0.00001681, -0.00005641, 0.00007527, -0.00008083, 0.00004627, -0.00001192, -0.00005305, 0.00010997, -0.00017269, 0.00013986, -0.00012038, -0.00004659, 0.00020134, -0.00081043, 0.00059319, 0.00105448, -0.00072204, -0.00005846, 0.00012991, -0.00021241, 0.00015317, -0.00012766, 0.00003341, 0.00001761, -0.00005965, 0.00006417, -0.00006765, 0.00003912, -0.00001223, -0.00001743, 0.00002738, -0.00003397, 0.00003454, -0.00002342, -0.00000292, 0.00002502, -0.00002714, 0.00002373, -0.00000621, -0.00002457, 0.00000842, -0.00000849, 0.00000911, -0.00000707, -0.00002614, 0.00001652, -0.00000668, -0.00001538, 0.00002066, 0.00000124, -0.00000792, 0.00000387, 0.00000329, -0.00000082, -0.00000101, -0.00000390, -0.00000050, 0.00000680, -0.00000385, 0.00000070, 0.00000094, -0.00000088, 0.00001035, -0.00000201, -0.00000582, -0.00000238, -0.00000438, 0.00000395, 0.00000480, 0.00000708, 0.00000460, -0.00001373, -0.00001036, 0.00000968, 0.00000914, },
		{ 0.00001140, 0.00000180, -0.00000306, -0.00000274, 0.00000074, -0.00000134, -0.00000255, 0.00000259, -0.00000011, 0.00000032, 0.00000988, 0.00000157, -0.00001381, -0.00000384, 0.00000662, 0.00000442, 0.00000576, -0.00000112, -0.00000284, -0.00000469, -0.00000368, 0.00000804, -0.00000053, 0.00000189, 0.00000064, -0.00000329, 0.00000746, -0.00000365, -0.00000268, -0.00000014, 0.00000330, 0.00000102, -0.00000659, 0.00000578, -0.00000696, 0.00000983, 0.00001527, -0.00002564, 0.00000558, 0.00001232, -0.00003117, 0.00000368, 0.00000692, -0.00000782, -0.00000120, -0.00002532, 0.00002495, -0.00001471, -0.00000408, 0.00004034, -0.00005670, 0.00004035, -0.00001495, -0.00001712, 0.00004687, -0.00007356, 0.00007055, -0.00005180, -0.00000025, 0.00004107, -0.00009860, 0.00012967, -0.00014536, 0.00006468, -0.00000326, -0.00017926, 0.00030728, -0.00069662, 0.00013989, 0.00128718, -0.00048053, -0.00031298, 0.00023963, -0.00023722, 0.00012026, -0.00006743, -0.00003116, 0.00006302, -0.00008099, 0.00006304, -0.00004898, 0.00001097, 0.00001425, -0.00003430, 0.00003248, -0.00003022, 0.00002327, -0.00000661, -0.00001773, 0.00003041, -0.00002379, 0.00001408, 0.00000670, -0.00003009, 0.00000636, -0.00000525, 0.00000272, 0.00000325, -0.00002943, 0.00000765, 0.00000452, -0.00002067, 0.00001530, 0.00000878, -0.00001033, 0.00000194, 0.00000577, -0.00000077, -0.00000151, -0.00000332, -0.00000218, 0.00000741, -0.00000248, -0.00000113, 0.00000230, -0.00000245, 0.00000998, 0.00000108, -0.00000705, -0.00000193, -0.00000488, 0.00000215, 0.00000529, 0.00000606, 0.00000703, -0.00001024, -0.00001418, 0.00000628, },
		{ 0.00000172, 0.00001267, 0.00000431, -0.00000286, -0.00000391, 0.00000048, -0.00000029, -0.00000342, 0.00000219, 0.00000154, -0.00000136, 0.00000792, 0.00000527, -0.00001181, -0.00000715, 0.00000565, 0.00000446, 0.00000598, 0.00000040, -0.00000288, -0.00000377, -0.00000540, 0.00000740, 0.00000065, 0.00000069, 0.00000243, -0.00000484, 0.00000778, -0.00000226, -0.00000431, 0.00000066, 0.00000211, 0.00000391, -0.00000885, 0.00000474, -0.00000294, 0.00000140, 0.00002125, -0.00001902, -0.00000796, 0.00002142, -0.00002570, -0.00000923, 0.00001503, -0.00001104, 0.00000196, -0.00002380, 0.00001387, 0.00000177, -0.00002235, 0.00005083, -0.00004907, 0.00001693, 0.00001469, -0.00004484, 0.00006381, -0.00007043, 0.00004472, -0.00000850, -0.00004817, 0.00007997, -0.00011528, 0.00011089, -0.00008231, -0.00002806, 0.00009934, -0.00025443, 0.00030679, -0.00047714, -0.00027488, 0.00133239, -0.00011540, -0.00054563, 0.00027653, -0.00019847, 0.00004685, 0.00000696, -0.00008746, 0.00008721, -0.00007868, 0.00004263, -0.00001958, -0.00001784, 0.00003316, -0.00003984, 0.00002716, -0.00001943, 0.00000909, 0.00000822, -0.00002635, 0.00002876, -0.00001551, 0.00000274, 0.00001711, -0.00003051, 0.00000024, 0.00000023, -0.00000395, 0.00000989, -0.00002629, -0.00000441, 0.00001294, -0.00002025, 0.00000648, 0.00001537, -0.00000983, -0.00000194, 0.00000803, 0.00000037, -0.00000245, -0.00000262, -0.00000368, 0.00000722, -0.00000038, -0.00000303, 0.00000328, -0.00000320, 0.00000849, 0.00000448, -0.00000770, -0.00000214, -0.00000456, 0.00000001, 0.00000556, 0.00000529, 0.00000823, -0.00000591, -0.00001645, },
		{ -0.00001683, -0.00000345, 0.00001258, 0.00000713, -0.00000194, -0.00000497, -0.00000040, 0.00000086, -0.00000374, 0.00000103, 0.00000320, -0.00000189, 0.00000518, 0.00000783, -0.00000867, -0.00000978, 0.00000402, 0.00000449, 0.00000594, 0.00000194, -0.00000282, -0.00000276, -0.00000672, 0.00000617, 0.00000221, -0.00000083, 0.00000424, -0.00000590, 0.00000716, 0.00000013, -0.00000646, 0.00000150, 0.00000089, 0.00000618, -0.00000893, 0.00000107, 0.00000228, -0.00000491, 0.00002138, -0.00000770, -0.00001973, 0.00002282, -0.00001316, -0.00002282, 0.00001903, -0.00000912, -0.00000049, -0.00001653, -0.00000018, 0.00001634, -0.00003299, 0.00004872, -0.00002896, -0.00001259, 0.00004132, -0.00006071, 0.00006230, -0.00004759, 0.00000522, 0.00003635, -0.00008321, 0.00009350, -0.00009933, 0.00006107, -0.00000300, -0.00011085, 0.00016037, -0.00025821, 0.00021803, -0.00021492, -0.00058141, 0.00118654, 0.00030952, -0.00069583, 0.00022079, -0.00010563, -0.00004550, 0.00007306, -0.00011952, 0.00008372, -0.00005484, 0.00001061, 0.00001084, -0.00003894, 0.00003990, -0.00003392, 0.00001461, -0.00000622, -0.00000353, 0.00001792, -0.00002798, 0.00002161, -0.00000513, -0.00000732, 0.00002302, -0.00002575, -0.00000838, 0.00000565, -0.00000848, 0.00001158, -0.00001819, -0.00001618, 0.00001610, -0.00001459, -0.00000335, 0.00001896, -0.00000611, -0.00000676, 0.00000898, 0.00000283, -0.00000353, -0.00000212, -0.00000469, 0.00000612, 0.00000220, -0.00000459, 0.00000354, -0.00000301, 0.00000614, 0.00000765, -0.00000737, -0.00000312, -0.00000365, -0.00000204, 0.00000531, 0.00000498, 0.00000833, -0.00000146, },
		{ 0.00000243, -0.00001532, -0.00000851, 0.00001091, 0.00000988, -0.00000029, -0.00000566, -0.00000181, 0.00000173, -0.00000331, -0.00000058, 0.00000435, -0.00000123, 0.00000224, 0.00000895, -0.00000482, -0.00001141, 0.00000188, 0.00000439, 0.00000565, 0.00000345, -0.00000268, -0.00000173, -0.00000746, 0.00000428, 0.00000411, -0.00000249, 0.00000568, -0.00000605, 0.00000536, 0.00000339, -0.00000862, 0.00000180, 0.00000024, 0.00000719, -0.00000679, -0.00000401, 0.00000659, -0.00000744, 0.00001635, 0.00000479, -0.00002595, 0.00001598, 0.00000301, -0.00003249, 0.00001612, -0.00000159, -0.00000711, -0.00000686, -0.00001254, 0.00002448, -0.00003329, 0.00003498, -0.00000175, -0.00004003, 0.00005687, -0.00005968, 0.00004258, -0.00001161, -0.00003634, 0.00006960, -0.00009514, 0.00007831, -0.00005685, -0.00000369, 0.00006954, -0.00016098, 0.00016691, -0.00019776, 0.00007851, 0.00002645, -0.00073848, 0.00088424, 0.00071589, -0.00071828, 0.00007700, 0.00001491, -0.00012932, 0.00011131, -0.00011913, 0.00005480, -0.00001811, -0.00002240, 0.00003320, -0.00004717, 0.00003410, -0.00001994, -0.00000018, 0.00000489, -0.00001155, 0.00002147, -0.00002385, 0.00001171, 0.00000448, -0.00001405, 0.00002390, -0.00001713, -0.00001716, 0.00000887, -0.00000955, 0.00000875, -0.00000789, -0.00002449, 0.00001324, -0.00000572, -0.00001147, 0.00001839, 0.00000008, -0.00001101, 0.00000781, 0.00000634, -0.00000418, -0.00000215, -0.00000504, 0.00000421, 0.00000480, -0.00000534, 0.00000288, -0.00000200, 0.00000339, 0.00001004, -0.00000586, -0.00000469, -0.00000256, -0.00000358, 0.00000435, 0.00000510, 0.00000773, },
		{ 0.00000687, 0.00000529, -0.00001226, -0.00001269, 0.00000771, 0.00001211, 0.00000202, -0.00000577, -0.00000349, 0.00000200, -0.00000222, -0.00000218, 0.00000460, 0.00000039, -0.00000027, 0.00000857, -0.00000078, -0.00001187, -0.00000054, 0.00000409, 0.00000510, 0.00000489, -0.00000243, -0.00000084, -0.00000744, 0.00000173, 0.00000618, -0.00000391, 0.00000632, -0.00000499, 0.00000245, 0.00000700, -0.00001006, 0.00000100, 0.00000051, 0.00000684, -0.00000308, -0.00000891, 0.00000838, -0.00000595, 0.00000841, 0.00001469, -0.00002465, 0.00000303, 0.00001788, -0.00003437, 0.00000571, 0.00000960, -0.00001453, 0.00000098, -0.00001919, 0.00002391, -0.00002390, 0.00001404, 0.00002508, -0.00005761, 0.00005648, -0.00004183, 0.00001046, 0.00002695, -0.00006775, 0.00008172, -0.00008087, 0.00003978, -0.00000154, -0.00006393, 0.00011579, -0.00016677, 0.00012259, -0.00009703, -0.00006692, 0.00019765, -0.00073991, 0.00048902, 0.00102873, -0.00059482, -0.00012564, 0.00012696, -0.00017958, 0.00011100, -0.00008824, 0.00001029, 0.00001943, -0.00004642, 0.00004184, -0.00004191, 0.00001927, -0.00000335, -0.00001227, 0.00001090, -0.00001410, 0.00001991, -0.00001645, 0.00000191, 0.00001126, -0.00001671, 0.00002056, -0.00000681, -0.00002388, 0.00000865, -0.00000718, 0.00000312, 0.00000162, -0.00002743, 0.00000537, 0.00000345, -0.00001573, 0.00001375, 0.00000711, -0.00001322, 0.00000428, 0.00001014, -0.00000373, -0.00000288, -0.00000478, 0.00000184, 0.00000690, -0.00000499, 0.00000141, -0.00000055, 0.00000080, 0.00001125, -0.00000323, -0.00000644, -0.00000178, -0.00000431, 0.00000274, 0.00000542, },
		{ 0.00000558, 0.00000616, 0.00000697, -0.00000826, -0.00001537, 0.00000329, 0.00001335, 0.00000484, -0.00000521, -0.00000514, 0.00000150, -0.00000078, -0.00000326, 0.00000382, 0.00000245, -0.00000181, 0.00000684, 0.00000291, -0.00001111, -0.00000301, 0.00000361, 0.00000426, 0.00000617, -0.00000194, -0.00000030, -0.00000656, -0.00000131, 0.00000804, -0.00000464, 0.00000579, -0.00000271, -0.00000118, 0.00001021, -0.00001008, -0.00000115, 0.00000161, 0.00000551, 0.00000111, -0.00001219, 0.00000702, -0.00000161, 0.00000051, 0.00001932, -0.00001637, -0.00001178, 0.00002661, -0.00002679, -0.00001004, 0.00002053, -0.00001872, 0.00000351, -0.00001826, 0.00001534, -0.00000859, -0.00000763, 0.00004421, -0.00006038, 0.00004001, -0.00001240, -0.00002450, 0.00005684, -0.00008001, 0.00006966, -0.00004526, -0.00000990, 0.00004978, -0.00010274, 0.00012510, -0.00013006, 0.00004497, 0.00001183, -0.00017866, 0.00027348, -0.00061230, 0.00007785, 0.00119307, -0.00033966, -0.00033994, 0.00019487, -0.00018149, 0.00007348, -0.00003796, -0.00003578, 0.00004635, -0.00005519, 0.00003599, -0.00002680, 0.00000124, 0.00001041, -0.00001843, 0.00001114, -0.00001236, 0.00001548, -0.00000842, -0.00000571, 0.00001447, -0.00001586, 0.00001466, 0.00000301, -0.00002706, 0.00000493, -0.00000259, -0.00000301, 0.00000796, -0.00002473, -0.00000508, 0.00001006, -0.00001521, 0.00000639, 0.00001302, -0.00001244, -0.00000106, 0.00001310, -0.00000171, -0.00000419, -0.00000422, -0.00000055, 0.00000805, -0.00000350, -0.00000054, 0.00000086, -0.00000116, 0.00001111, 0.00000015, -0.00000784, -0.00000167, -0.00000418, 0.00000077, },
		{ -0.00000116, 0.00000527, 0.00000580, 0.00000759, -0.00000405, -0.00001620, -0.00000176, 0.00001321, 0.00000791, -0.00000393, -0.00000648, 0.00000030, 0.00000055, -0.00000350, 0.00000223, 0.00000430, -0.00000202, 0.00000421, 0.00000572, -0.00000921, -0.00000530, 0.00000301, 0.00000317, 0.00000712, -0.00000107, -0.00000028, -0.00000487, -0.00000447, 0.00000921, -0.00000427, 0.00000398, 0.00000049, -0.00000480, 0.00001221, -0.00000826, -0.00000447, 0.00000303, 0.00000395, 0.00000471, -0.00001306, 0.00000303, 0.00000353, -0.00000472, 0.00001786, -0.00000392, -0.00002350, 0.00002608, -0.00001112, -0.00002663, 0.00002661, -0.00001651, -0.00000065, -0.00001067, 0.00000232, 0.00000706, -0.00002361, 0.00005080, -0.00004784, 0.00001221, 0.00001983, -0.00005200, 0.00006960, -0.00007001, 0.00003767, 0.00000064, -0.00005573, 0.00008244, -0.00011071, 0.00009825, -0.00006460, -0.00004080, 0.00009835, -0.00023202, 0.00025547, -0.00040520, -0.00027659, 0.00118545, 0.00000359, -0.00051234, 0.00019399, -0.00013519, 0.00001147, 0.00001527, -0.00006986, 0.00005542, -0.00004792, 0.00001953, -0.00000801, -0.00001405, 0.00001767, -0.00001796, 0.00000711, -0.00000867, 0.00001058, -0.00000171, -0.00001033, 0.00001455, -0.00001279, 0.00000801, 0.00001067, -0.00002630, -0.00000126, 0.00000243, -0.00000759, 0.00001004, -0.00001774, -0.00001512, 0.00001216, -0.00001050, -0.00000156, 0.00001614, -0.00000862, -0.00000701, 0.00001414, 0.00000193, -0.00000558, -0.00000377, -0.00000249, 0.00000801, -0.00000113, -0.00000245, 0.00000177, -0.00000222, 0.00000980, 0.00000373, -0.00000840, -0.00000239, -0.00000341, },
		{ -0.00000241, -0.00000262, 0.00000434, 0.00000576, 0.00000752, -0.00000028, -0.00001520, -0.00000668, 0.00001146, 0.00001086, -0.00000193, -0.00000733, -0.00000134, 0.00000132, -0.00000284, 0.00000033, 0.00000533, -0.00000091, 0.00000130, 0.00000723, -0.00000637, -0.00000721, 0.00000232, 0.00000196, 0.00000749, 0.00000033, -0.00000081, -0.00000261, -0.00000722, 0.00000922, -0.00000260, 0.00000110, 0.00000396, -0.00000762, 0.00001240, -0.00000463, -0.00000834, 0.00000403, 0.00000285, 0.00000704, -0.00001158, -0.00000225, 0.00000738, -0.00000592, 0.00001161, 0.00000852, -0.00002812, 0.00001611, 0.00000840, -0.00003853, 0.00002429, -0.00000702, -0.00001012, 0.00000005, -0.00001007, 0.00001759, -0.00002963, 0.00004396, -0.00002398, -0.00001881, 0.00004533, -0.00006417, 0.00006209, -0.00004137, -0.00000422, 0.00004315, -0.00008460, 0.00008819, -0.00008811, 0.00004623, 0.00000903, -0.00010995, 0.00014208, -0.00022169, 0.00016754, -0.00017695, -0.00052113, 0.00101755, 0.00037208, -0.00059653, 0.00011820, -0.00005580, -0.00005521, 0.00005509, -0.00008311, 0.00004564, -0.00002900, -0.00000083, 0.00000805, -0.00002252, 0.00001754, -0.00001257, 0.00000155, -0.00000545, 0.00000681, 0.00000298, -0.00001228, 0.00001267, -0.00000893, 0.00000196, 0.00001543, -0.00002223, -0.00000830, 0.00000614, -0.00000941, 0.00000816, -0.00000880, -0.00002211, 0.00000925, -0.00000344, -0.00000783, 0.00001564, -0.00000265, -0.00001205, 0.00001258, 0.00000669, -0.00000634, -0.00000384, -0.00000376, 0.00000686, 0.00000160, -0.00000378, 0.00000190, -0.00000239, 0.00000770, 0.00000691, -0.00000781, -0.00000381, },
		{ -0.00000557, -0.00000163, -0.00000334, 0.00000287, 0.00000581, 0.00000716, 0.00000263, -0.00001273, -0.00001070, 0.00000817, 0.00001322, 0.00000076, -0.00000760, -0.00000308, 0.00000128, -0.00000157, -0.00000124, 0.00000518, 0.00000117, -0.00000115, 0.00000721, -0.00000295, -0.00000855, 0.00000151, 0.00000086, 0.00000707, 0.00000227, -0.00000173, -0.00000026, -0.00000900, 0.00000779, 0.00000024, -0.00000227, 0.00000691, -0.00000892, 0.00001056, 0.00000031, -0.00001186, 0.00000391, 0.00000261, 0.00000800, -0.00000846, -0.00000723, 0.00000855, -0.00000336, 0.00000341, 0.00001690, -0.00002395, -0.00000033, 0.00002585, -0.00004104, 0.00001255, 0.00000773, -0.00002108, 0.00000907, -0.00001703, 0.00001956, -0.00002492, 0.00002693, 0.00000394, -0.00004415, 0.00005678, -0.00005804, 0.00003737, -0.00000308, -0.00004345, 0.00007026, -0.00008915, 0.00006733, -0.00004413, -0.00001384, 0.00007002, -0.00014514, 0.00013706, -0.00016028, 0.00004645, 0.00001992, -0.00063144, 0.00073132, 0.00069836, -0.00056525, -0.00001771, 0.00003100, -0.00010608, 0.00007041, -0.00007391, 0.00002204, -0.00000606, -0.00001808, 0.00001690, -0.00002305, 0.00001178, -0.00000538, -0.00000281, -0.00000415, 0.00000454, 0.00000600, -0.00001255, 0.00001004, -0.00000527, -0.00000283, 0.00001742, -0.00001608, -0.00001455, 0.00000746, -0.00000836, 0.00000370, -0.00000044, -0.00002452, 0.00000236, 0.00000352, -0.00001080, 0.00001180, 0.00000392, -0.00001483, 0.00000838, 0.00001158, -0.00000577, -0.00000459, -0.00000435, 0.00000494, 0.00000412, -0.00000415, 0.00000126, -0.00000195, 0.00000534, 0.00000921, -0.00000606, },
		{ -0.00000342, -0.00000717, -0.00000145, -0.00000331, 0.00000114, 0.00000566, 0.00000684, 0.00000456, -0.00000944, -0.00001321, 0.00000371, 0.00001445, 0.00000404, -0.00000726, -0.00000461, 0.00000044, -0.00000022, -0.00000193, 0.00000385, 0.00000355, -0.00000247, 0.00000572, 0.00000056, -0.00000909, 0.00000053, 0.00000015, 0.00000575, 0.00000456, -0.00000268, 0.00000166, -0.00000938, 0.00000492, 0.00000377, -0.00000538, 0.00000860, -0.00000835, 0.00000697, 0.00000566, -0.00001410, 0.00000231, 0.00000318, 0.00000799, -0.00000473, -0.00001066, 0.00000680, 0.00000132, -0.00000344, 0.00001869, -0.00001227, -0.00001797, 0.00003546, -0.00003207, -0.00000634, 0.00002320, -0.00002863, 0.00001206, -0.00001582, 0.00001276, -0.00001238, 0.00000590, 0.00002783, -0.00005686, 0.00005132, -0.00003626, 0.00000346, 0.00003339, -0.00006889, 0.00007514, -0.00006967, 0.00002796, 0.00000665, -0.00006426, 0.00010332, -0.00014056, 0.00009163, -0.00007227, -0.00006958, 0.00014988, -0.00061365, 0.00038749, 0.00092562, -0.00041708, -0.00018157, 0.00009673, -0.00012654, 0.00005864, -0.00004784, -0.00000651, 0.00001294, -0.00002736, 0.00001726, -0.00001740, 0.00000375, 0.00000047, -0.00000432, -0.00000483, 0.00000308, 0.00000828, -0.00001211, 0.00000742, -0.00000216, -0.00000636, 0.00001725, -0.00000921, -0.00001884, 0.00000617, -0.00000524, -0.00000149, 0.00000536, -0.00002221, -0.00000640, 0.00000808, -0.00000993, 0.00000592, 0.00000935, -0.00001459, 0.00000229, 0.00001538, -0.00000341, -0.00000585, -0.00000456, 0.00000271, 0.00000590, -0.00000345, 0.00000010, -0.00000129, 0.00000318, 0.00001040, },
		{ 0.00001051, -0.00000032, -0.00000814, -0.00000199, -0.00000273, -0.00000046, 0.00000508, 0.00000667, 0.00000568, -0.00000599, -0.00001396, -0.00000123, 0.00001412, 0.00000767, -0.00000627, -0.00000576, -0.00000092, 0.00000062, -0.00000150, 0.00000179, 0.00000541, -0.00000227, 0.00000314, 0.00000357, -0.00000866, -0.00000072, -0.00000002, 0.00000364, 0.00000680, -0.00000319, 0.00000265, -0.00000821, 0.00000100, 0.00000726, -0.00000744, 0.00000854, -0.00000598, 0.00000233, 0.00001042, -0.00001442, -0.00000071, 0.00000408, 0.00000763, -0.00000129, -0.00001206, 0.00000303, 0.00000578, -0.00000643, 0.00001375, 0.00000301, -0.00003086, 0.00003366, -0.00001309, -0.00002720, 0.00003380, -0.00002858, 0.00000685, -0.00000682, 0.00000027, 0.00000257, -0.00001215, 0.00004129, -0.00005408, 0.00003131, -0.00000608, -0.00002929, 0.00005778, -0.00007409, 0.00005795, -0.00003350, -0.00001705, 0.00004906, -0.00009203, 0.00010336, -0.00010225, 0.00002420, 0.00001460, -0.00015099, 0.00019988, -0.00049862, 0.00005078, 0.00101961, -0.00017661, -0.00033215, 0.00011872, -0.00011239, 0.00002599, -0.00001529, -0.00003056, 0.00002242, -0.00002733, 0.00001100, -0.00000913, -0.00000306, 0.00000302, -0.00000296, -0.00000638, 0.00000123, 0.00001061, -0.00001145, 0.00000493, 0.00000059, -0.00000896, 0.00001565, -0.00000265, -0.00002071, 0.00000282, -0.00000138, -0.00000573, 0.00000767, -0.00001637, -0.00001456, 0.00000884, -0.00000592, -0.00000016, 0.00001228, -0.00001143, -0.00000440, 0.00001700, 0.00000067, -0.00000707, -0.00000478, 0.00000062, 0.00000667, -0.00000189, -0.00000114, -0.00000081, 0.00000150, },
		{ 0.00000036, 0.00000980, 0.00000272, -0.00000821, -0.00000316, -0.00000198, -0.00000161, 0.00000402, 0.00000658, 0.00000630, -0.00000294, -0.00001309, -0.00000581, 0.00001200, 0.00001124, -0.00000454, -0.00000652, -0.00000237, 0.00000051, -0.00000012, -0.00000027, 0.00000606, -0.00000055, 0.00000017, 0.00000551, -0.00000718, -0.00000223, 0.00000039, 0.00000109, 0.00000846, -0.00000283, 0.00000248, -0.00000569, -0.00000327, 0.00000987, -0.00000787, 0.00000666, -0.00000234, -0.00000244, 0.00001371, -0.00001261, -0.00000461, 0.00000467, 0.00000743, 0.00000147, -0.00001175, -0.00000124, 0.00000802, -0.00000472, 0.00000435, 0.00001673, -0.00003443, 0.00002036, 0.00001093, -0.00004345, 0.00003484, -0.00001919, -0.00000558, 0.00000652, -0.00001268, 0.00001394, -0.00002177, 0.00004157, -0.00003776, 0.00000338, 0.00002308, -0.00005150, 0.00006399, -0.00005899, 0.00002534, 0.00000767, -0.00005425, 0.00007189, -0.00009213, 0.00007462, -0.00004490, -0.00004336, 0.00007743, -0.00018304, 0.00017835, -0.00033124, -0.00022434, 0.00097451, 0.00011190, -0.00042995, 0.00008696, -0.00007096, -0.00001507, 0.00001239, -0.00004336, 0.00002086, -0.00002013, 0.00000196, -0.00000202, -0.00000638, 0.00000208, -0.00000009, -0.00000715, -0.00000190, 0.00001313, -0.00001035, 0.00000219, 0.00000333, -0.00001091, 0.00001314, 0.00000301, -0.00002035, -0.00000158, 0.00000197, -0.00000799, 0.00000670, -0.00000893, -0.00002002, 0.00000559, -0.00000041, -0.00000465, 0.00001216, -0.00000626, -0.00001019, 0.00001582, 0.00000587, -0.00000755, -0.00000533, -0.00000107, 0.00000642, 0.00000014, -0.00000201, -0.00000073, },
		{ -0.00000106, -0.00000039, 0.00000862, 0.00000531, -0.00000733, -0.00000465, -0.00000144, -0.00000214, 0.00000263, 0.00000634, 0.00000675, -0.00000057, -0.00001107, -0.00000924, 0.00000822, 0.00001416, -0.00000197, -0.00000696, -0.00000349, -0.00000065, 0.00000164, -0.00000150, 0.00000514, 0.00000218, -0.00000233, 0.00000597, -0.00000478, -0.00000389, 0.00000119, -0.00000137, 0.00000902, -0.00000138, 0.00000124, -0.00000237, -0.00000704, 0.00001094, -0.00000645, 0.00000334, 0.00000178, -0.00000642, 0.00001499, -0.00000895, -0.00000862, 0.00000442, 0.00000755, 0.00000367, -0.00001057, -0.00000454, 0.00000710, 0.00000058, -0.00000567, 0.00002419, -0.00002705, -0.00000072, 0.00003292, -0.00004919, 0.00002435, -0.00000206, -0.00002125, 0.00001889, -0.00002064, 0.00001734, -0.00002080, 0.00003027, -0.00001381, -0.00002381, 0.00004269, -0.00005761, 0.00005177, -0.00002948, -0.00001206, 0.00004169, -0.00007380, 0.00007085, -0.00006809, 0.00002906, 0.00001348, -0.00009241, 0.00010381, -0.00016701, 0.00010900, -0.00015791, -0.00040385, 0.00081167, 0.00039545, -0.00044733, 0.00000720, -0.00001846, -0.00005042, 0.00002681, -0.00004288, 0.00001083, -0.00001020, -0.00000575, 0.00000143, -0.00000587, -0.00000094, 0.00000230, -0.00000588, -0.00000635, 0.00001520, -0.00000818, -0.00000121, 0.00000621, -0.00001217, 0.00000989, 0.00000763, -0.00001831, -0.00000599, 0.00000394, -0.00000808, 0.00000349, -0.00000195, -0.00002163, -0.00000064, 0.00000456, -0.00000637, 0.00000938, -0.00000052, -0.00001382, 0.00001190, 0.00001114, -0.00000664, -0.00000631, -0.00000235, 0.00000540, 0.00000211, -0.00000220, },
		{ -0.00000164, -0.00000160, -0.00000099, 0.00000728, 0.00000724, -0.00000567, -0.00000605, -0.00000137, -0.00000211, 0.00000122, 0.00000572, 0.00000717, 0.00000115, -0.00000859, -0.00001101, 0.00000331, 0.00001576, 0.00000144, -0.00000713, -0.00000405, -0.00000257, 0.00000295, -0.00000129, 0.00000284, 0.00000506, -0.00000357, 0.00000483, -0.00000181, -0.00000540, 0.00000201, -0.00000318, 0.00000817, 0.00000107, -0.00000067, 0.00000098, -0.00000955, 0.00001012, -0.00000343, -0.00000070, 0.00000548, -0.00000893, 0.00001419, -0.00000406, -0.00001197, 0.00000312, 0.00000773, 0.00000577, -0.00000937, -0.00000607, 0.00000352, 0.00000693, -0.00001233, 0.00002279, -0.00001074, -0.00002311, 0.00004589, -0.00004134, 0.00000411, 0.00001810, -0.00003451, 0.00002497, -0.00001990, 0.00001152, -0.00001096, 0.00001253, 0.00001000, -0.00004243, 0.00004770, -0.00004731, 0.00002630, 0.00000478, -0.00004308, 0.00005961, -0.00007199, 0.00004910, -0.00002993, -0.00001828, 0.00005718, -0.00011236, 0.00009351, -0.00011680, 0.00002181, -0.00001580, -0.00047827, 0.00057213, 0.00062470, -0.00037533, -0.00010063, 0.00002538, -0.00006918, 0.00002512, -0.00003197, -0.00000245, -0.00000218, -0.00000945, 0.00000078, -0.00000294, -0.00000382, 0.00000254, -0.00000231, -0.00001119, 0.00001572, -0.00000433, -0.00000536, 0.00000890, -0.00001237, 0.00000587, 0.00001128, -0.00001522, -0.00000967, 0.00000428, -0.00000648, -0.00000051, 0.00000299, -0.00001943, -0.00000801, 0.00000725, -0.00000510, 0.00000511, 0.00000430, -0.00001465, 0.00000603, 0.00001527, -0.00000400, -0.00000749, -0.00000341, 0.00000395, 0.00000361, },
		{ 0.00000440, -0.00000050, -0.00000203, -0.00000164, 0.00000594, 0.00000849, -0.00000355, -0.00000703, -0.00000184, -0.00000176, 0.00000011, 0.00000466, 0.00000754, 0.00000250, -0.00000627, -0.00001100, -0.00000187, 0.00001548, 0.00000553, -0.00000697, -0.00000409, -0.00000464, 0.00000307, 0.00000049, -0.00000013, 0.00000708, -0.00000304, 0.00000240, 0.00000115, -0.00000638, 0.00000245, -0.00000391, 0.00000590, 0.00000408, -0.00000258, 0.00000360, -0.00001029, 0.00000749, 0.00000059, -0.00000462, 0.00000804, -0.00000966, 0.00001158, 0.00000127, -0.00001412, 0.00000096, 0.00000746, 0.00000821, -0.00000857, -0.00000604, -0.00000106, 0.00001137, -0.00001297, 0.00001302, 0.00000943, -0.00003957, 0.00004526, -0.00002093, -0.00002052, 0.00003486, -0.00003996, 0.00002150, -0.00000995, -0.00000122, 0.00000297, -0.00000488, 0.00002660, -0.00004786, 0.00003791, -0.00002528, -0.00000377, 0.00003363, -0.00005952, 0.00005812, -0.00005179, 0.00001544, 0.00000962, -0.00005408, 0.00007694, -0.00010245, 0.00005647, -0.00005269, -0.00005591, 0.00007384, -0.00046040, 0.00030551, 0.00076457, -0.00022532, -0.00020798, 0.00004374, -0.00006727, 0.00001048, -0.00001683, -0.00001336, 0.00000085, -0.00000872, -0.00000249, 0.00000006, -0.00000466, 0.00000010, 0.00000263, -0.00001489, 0.00001368, 0.00000127, -0.00000971, 0.00001068, -0.00001106, 0.00000112, 0.00001394, -0.00001151, -0.00001226, 0.00000324, -0.00000406, -0.00000403, 0.00000517, -0.00001445, -0.00001452, 0.00000671, -0.00000162, 0.00000089, 0.00000712, -0.00001280, -0.00000053, 0.00001725, 0.00000021, -0.00000835, -0.00000451, 0.00000239, },
		{ 0.00000089, 0.00000447, 0.00000090, -0.00000207, -0.00000241, 0.00000463, 0.00000918, -0.00000128, -0.00000742, -0.00000272, -0.00000141, -0.00000050, 0.00000326, 0.00000761, 0.00000378, -0.00000452, -0.00000957, -0.00000632, 0.00001305, 0.00000985, -0.00000626, -0.00000388, -0.00000617, 0.00000155, 0.00000338, -0.00000272, 0.00000740, -0.00000075, -0.00000062, 0.00000344, -0.00000649, 0.00000220, -0.00000337, 0.00000258, 0.00000697, -0.00000383, 0.00000494, -0.00000915, 0.00000353, 0.00000477, -0.00000762, 0.00000900, -0.00000865, 0.00000773, 0.00000635, -0.00001488, -0.00000158, 0.00000631, 0.00001102, -0.00000788, -0.00000549, -0.00000455, 0.00001173, -0.00000723, -0.00000169, 0.00002689, -0.00004446, 0.00003048, 0.00000688, -0.00004214, 0.00004223, -0.00003452, 0.00000849, 0.00000625, -0.00001585, 0.00001493, -0.00001591, 0.00003187, -0.00003990, 0.00001765, 0.00000068, -0.00002929, 0.00004949, -0.00005824, 0.00003994, -0.00002118, -0.00001895, 0.00003922, -0.00007032, 0.00007165, -0.00007035, 0.00000816, 0.00000550, -0.00010597, 0.00010714, -0.00037836, 0.00005830, 0.00080055, -0.00002539, -0.00028563, 0.00002811, -0.00004857, -0.00000959, -0.00000433, -0.00001810, -0.00000153, -0.00000531, -0.00000585, 0.00000104, -0.00000267, -0.00000423, 0.00000718, -0.00001594, 0.00000873, 0.00000792, -0.00001320, 0.00001070, -0.00000796, -0.00000404, 0.00001543, -0.00000740, -0.00001380, 0.00000136, -0.00000168, -0.00000624, 0.00000474, -0.00000830, -0.00001856, 0.00000302, 0.00000262, -0.00000190, 0.00000755, -0.00000911, -0.00000642, 0.00001655, 0.00000534, -0.00000827, -0.00000582, },
		{ -0.00000727, -0.00000056, 0.00000399, 0.00000218, -0.00000155, -0.00000322, 0.00000325, 0.00000948, 0.00000091, -0.00000722, -0.00000372, -0.00000133, -0.00000061, 0.00000183, 0.00000714, 0.00000520, -0.00000338, -0.00000742, -0.00000917, 0.00000865, 0.00001367, -0.00000467, -0.00000376, -0.00000661, -0.00000151, 0.00000643, -0.00000386, 0.00000566, 0.00000272, -0.00000330, 0.00000447, -0.00000556, 0.00000117, -0.00000176, -0.00000112, 0.00000897, -0.00000388, 0.00000478, -0.00000645, -0.00000096, 0.00000828, -0.00000911, 0.00000826, -0.00000625, 0.00000326, 0.00001064, -0.00001434, -0.00000400, 0.00000416, 0.00001371, -0.00000648, -0.00000577, -0.00000540, 0.00000750, 0.00000276, -0.00001614, 0.00003579, -0.00003564, 0.00000549, 0.00003421, -0.00005358, 0.00003681, -0.00001866, -0.00001055, 0.00002327, -0.00002648, 0.00001985, -0.00001729, 0.00002575, -0.00002256, -0.00000591, 0.00002234, -0.00004345, 0.00004943, -0.00004168, 0.00001238, 0.00000979, -0.00004422, 0.00005211, -0.00006586, 0.00004737, -0.00002859, -0.00003581, 0.00004388, -0.00012208, 0.00009445, -0.00026612, -0.00013538, 0.00073943, 0.00018720, -0.00031150, -0.00001903, -0.00002322, -0.00002642, 0.00000069, -0.00001601, -0.00000718, -0.00000210, -0.00000706, -0.00000094, 0.00000147, -0.00000867, 0.00000953, -0.00001357, 0.00000151, 0.00001423, -0.00001464, 0.00000842, -0.00000334, -0.00000894, 0.00001542, -0.00000297, -0.00001445, -0.00000080, 0.00000009, -0.00000694, 0.00000252, -0.00000261, -0.00001935, -0.00000279, 0.00000597, -0.00000249, 0.00000602, -0.00000479, -0.00001052, 0.00001330, 0.00001035, -0.00000676, },
		{ -0.00000366, -0.00000852, -0.00000210, 0.00000317, 0.00000307, -0.00000054, -0.00000382, 0.00000171, 0.00000946, 0.00000292, -0.00000656, -0.00000460, -0.00000162, -0.00000042, 0.00000073, 0.00000598, 0.00000671, -0.00000257, -0.00000538, -0.00000993, 0.00000297, 0.00001613, -0.00000196, -0.00000400, -0.00000579, -0.00000546, 0.00000845, -0.00000282, 0.00000217, 0.00000632, -0.00000468, 0.00000393, -0.00000370, -0.00000044, 0.00000045, -0.00000439, 0.00000949, -0.00000258, 0.00000333, -0.00000284, -0.00000513, 0.00001044, -0.00000888, 0.00000604, -0.00000295, -0.00000127, 0.00001384, -0.00001273, -0.00000600, 0.00000141, 0.00001535, -0.00000340, -0.00000777, -0.00000325, 0.00000012, 0.00001317, -0.00002512, 0.00003294, -0.00001536, -0.00002235, 0.00005278, -0.00005037, 0.00001917, 0.00000347, -0.00002973, 0.00003505, -0.00002848, 0.00001540, -0.00000947, 0.00001194, -0.00000240, -0.00002528, 0.00003367, -0.00004365, 0.00003555, -0.00001646, -0.00001539, 0.00003229, -0.00005460, 0.00004743, -0.00004568, 0.00001440, 0.00000994, -0.00006417, 0.00005712, -0.00010865, 0.00005425, -0.00015463, -0.00025864, 0.00060499, 0.00037529, -0.00027626, -0.00008502, -0.00000382, -0.00003361, -0.00000286, -0.00000956, -0.00001258, -0.00000170, -0.00000513, -0.00000521, 0.00000597, -0.00001121, 0.00000853, -0.00000814, -0.00000643, 0.00001856, -0.00001315, 0.00000393, 0.00000199, -0.00001268, 0.00001361, 0.00000166, -0.00001437, -0.00000287, 0.00000104, -0.00000641, -0.00000044, 0.00000141, -0.00001709, -0.00000913, 0.00000716, -0.00000089, 0.00000351, -0.00000103, -0.00001230, 0.00000826, 0.00001418, },
		{ 0.00001597, 0.00000071, -0.00000905, -0.00000384, 0.00000215, 0.00000346, 0.00000076, -0.00000399, -0.00000000, 0.00000911, 0.00000479, -0.00000563, -0.00000521, -0.00000220, -0.00000026, 0.00000025, 0.00000420, 0.00000803, -0.00000161, -0.00000409, -0.00000867, -0.00000289, 0.00001645, 0.00000188, -0.00000455, -0.00000398, -0.00000927, 0.00000842, 0.00000047, -0.00000210, 0.00000883, -0.00000415, 0.00000196, -0.00000135, -0.00000220, 0.00000264, -0.00000650, 0.00000826, -0.00000013, 0.00000110, 0.00000090, -0.00000822, 0.00001089, -0.00000702, 0.00000276, 0.00000073, -0.00000538, 0.00001577, -0.00001022, -0.00000765, -0.00000117, 0.00001497, 0.00000186, -0.00001146, 0.00000076, -0.00000758, 0.00001988, -0.00002526, 0.00001898, 0.00001037, -0.00004448, 0.00005662, -0.00003233, -0.00000601, 0.00002532, -0.00004272, 0.00003703, -0.00002013, 0.00000285, 0.00000378, -0.00000362, 0.00001381, -0.00003506, 0.00003252, -0.00003185, 0.00001379, 0.00000907, -0.00003536, 0.00004126, -0.00004954, 0.00002958, -0.00001835, -0.00001651, 0.00003575, -0.00007243, 0.00004813, -0.00007703, 0.00000629, -0.00006553, -0.00031181, 0.00043030, 0.00051016, -0.00018510, -0.00015086, -0.00000088, -0.00002951, -0.00001214, -0.00000304, -0.00001447, -0.00000520, -0.00000084, -0.00000993, 0.00000870, -0.00001051, 0.00000426, -0.00000110, -0.00001314, 0.00001960, -0.00000862, -0.00000194, 0.00000685, -0.00001442, 0.00000998, 0.00000617, -0.00001356, -0.00000472, 0.00000129, -0.00000519, -0.00000316, 0.00000323, -0.00001268, -0.00001440, 0.00000563, 0.00000210, 0.00000120, 0.00000136, -0.00001188, 0.00000253, },
		{ -0.00000272, 0.00001538, 0.00000562, -0.00000836, -0.00000575, 0.00000096, 0.00000339, 0.00000206, -0.00000358, -0.00000176, 0.00000832, 0.00000654, -0.00000452, -0.00000554, -0.00000282, -0.00000043, 0.00000043, 0.00000217, 0.00000870, -0.00000011, -0.00000383, -0.00000602, -0.00000768, 0.00001421, 0.00000645, -0.00000508, -0.00000182, -0.00001187, 0.00000581, 0.00000531, -0.00000579, 0.00000925, -0.00000163, -0.00000083, 0.00000085, -0.00000361, 0.00000419, -0.00000704, 0.00000545, 0.00000289, -0.00000122, 0.00000402, -0.00000974, 0.00000959, -0.00000387, -0.00000107, 0.00000427, -0.00000867, 0.00001626, -0.00000680, -0.00000930, -0.00000275, 0.00001206, 0.00000892, -0.00001563, 0.00000442, -0.00001243, 0.00002000, -0.00001626, -0.00000178, 0.00003361, -0.00005384, 0.00004404, -0.00000394, -0.00003130, 0.00004000, -0.00004497, 0.00002780, -0.00000334, -0.00001349, 0.00001682, -0.00001497, 0.00002132, -0.00003349, 0.00002090, -0.00001343, -0.00000838, 0.00002762, -0.00004294, 0.00003637, -0.00003313, 0.00000601, 0.00000711, -0.00003731, 0.00004484, -0.00006273, 0.00002535, -0.00004050, -0.00003388, -0.00000859, -0.00030833, 0.00024910, 0.00057682, -0.00005555, -0.00019688, -0.00001903, -0.00001774, -0.00002186, -0.00000060, -0.00001133, -0.00001168, 0.00000367, -0.00001284, 0.00000819, -0.00000643, -0.00000198, 0.00000548, -0.00001689, 0.00001689, -0.00000184, -0.00000780, 0.00000999, -0.00001368, 0.00000493, 0.00001005, -0.00001194, -0.00000637, 0.00000109, -0.00000380, -0.00000504, 0.00000300, -0.00000747, -0.00001740, 0.00000167, 0.00000525, 0.00000007, 0.00000216, -0.00000993, },
		{ -0.00000737, -0.00000663, 0.00001263, 0.00001009, -0.00000615, -0.00000754, -0.00000049, 0.00000297, 0.00000311, -0.00000260, -0.00000332, 0.00000700, 0.00000815, -0.00000323, -0.00000573, -0.00000324, -0.00000106, 0.00000107, 0.00000039, 0.00000834, 0.00000209, -0.00000439, -0.00000297, -0.00001037, 0.00000955, 0.00001095, -0.00000504, -0.00000003, -0.00001245, 0.00000082, 0.00001037, -0.00000752, 0.00000711, 0.00000230, -0.00000351, 0.00000223, -0.00000425, 0.00000474, -0.00000594, 0.00000160, 0.00000575, -0.00000297, 0.00000599, -0.00000951, 0.00000682, 0.00000005, -0.00000487, 0.00000720, -0.00001076, 0.00001509, -0.00000236, -0.00001132, -0.00000284, 0.00000692, 0.00001642, -0.00001828, 0.00000535, -0.00001222, 0.00001303, -0.00000101, -0.00002270, 0.00004700, -0.00004714, 0.00001836, 0.00002685, -0.00004875, 0.00004262, -0.00003528, 0.00000962, 0.00001705, -0.00002778, 0.00002429, -0.00001811, 0.00001874, -0.00002268, 0.00000390, 0.00000489, -0.00002446, 0.00003501, -0.00003788, 0.00002146, -0.00001208, -0.00001550, 0.00002401, -0.00004456, 0.00003879, -0.00004183, -0.00000101, -0.00000999, -0.00005789, 0.00001711, -0.00026842, 0.00008854, 0.00057533, 0.00008772, -0.00020841, -0.00005524, -0.00000579, -0.00002660, -0.00000439, -0.00000412, -0.00001853, 0.00000587, -0.00001239, 0.00000422, -0.00000020, -0.00000814, 0.00000967, -0.00001682, 0.00001102, 0.00000569, -0.00001217, 0.00001060, -0.00001061, -0.00000071, 0.00001269, -0.00000942, -0.00000784, 0.00000069, -0.00000262, -0.00000588, 0.00000134, -0.00000275, -0.00001764, -0.00000375, 0.00000732, 0.00000055, 0.00000175, },
		{ 0.00000088, -0.00000504, -0.00000880, 0.00000843, 0.00001323, -0.00000252, -0.00000876, -0.00000228, 0.00000230, 0.00000377, -0.00000122, -0.00000441, 0.00000515, 0.00000949, -0.00000167, -0.00000594, -0.00000332, -0.00000204, 0.00000175, -0.00000061, 0.00000679, 0.00000476, -0.00000518, -0.00000059, -0.00001040, 0.00000327, 0.00001434, -0.00000386, 0.00000079, -0.00001078, -0.00000556, 0.00001405, -0.00000635, 0.00000276, 0.00000652, -0.00000508, 0.00000235, -0.00000397, 0.00000421, -0.00000356, -0.00000248, 0.00000771, -0.00000369, 0.00000659, -0.00000769, 0.00000305, 0.00000416, -0.00000812, 0.00000906, -0.00001127, 0.00001207, 0.00000314, -0.00001383, -0.00000161, 0.00000072, 0.00002233, -0.00001737, 0.00000196, -0.00000662, 0.00000110, 0.00001534, -0.00003703, 0.00004614, -0.00002608, -0.00001297, 0.00005106, -0.00005252, 0.00003204, -0.00001630, -0.00001230, 0.00003481, -0.00003472, 0.00002292, -0.00001230, 0.00000829, -0.00000748, -0.00001230, 0.00001740, -0.00003078, 0.00003103, -0.00002376, 0.00000271, 0.00000671, -0.00002930, 0.00002952, -0.00003943, 0.00002305, -0.00001806, -0.00002241, 0.00000848, -0.00006475, 0.00001965, -0.00021266, -0.00003506, 0.00051825, 0.00021953, -0.00017972, -0.00009968, -0.00000206, -0.00002327, -0.00001358, 0.00000413, -0.00002264, 0.00000396, -0.00000839, -0.00000202, 0.00000613, -0.00001215, 0.00001031, -0.00001321, 0.00000346, 0.00001218, -0.00001389, 0.00000852, -0.00000597, -0.00000591, 0.00001358, -0.00000607, -0.00000910, 0.00000021, -0.00000176, -0.00000585, -0.00000094, 0.00000056, -0.00001538, -0.00000929, 0.00000742, 0.00000243, },
		{ 0.00000499, 0.00000035, -0.00000347, -0.00000937, 0.00000373, 0.00001441, 0.00000204, -0.00000888, -0.00000440, 0.00000136, 0.00000404, 0.00000029, -0.00000482, 0.00000291, 0.00001030, 0.00000030, -0.00000624, -0.00000314, -0.00000306, 0.00000197, -0.00000056, 0.00000429, 0.00000733, -0.00000543, 0.00000033, -0.00000794, -0.00000335, 0.00001560, -0.00000128, 0.00000051, -0.00000727, -0.00001190, 0.00001496, -0.00000217, -0.00000271, 0.00000968, -0.00000482, 0.00000115, -0.00000295, 0.00000288, -0.00000051, -0.00000597, 0.00000828, -0.00000324, 0.00000587, -0.00000467, -0.00000112, 0.00000784, -0.00001028, 0.00000943, -0.00000989, 0.00000719, 0.00000942, -0.00001642, 0.00000004, -0.00000474, 0.00002464, -0.00001171, -0.00000569, 0.00000254, -0.00001173, 0.00002715, -0.00004011, 0.00003108, 0.00000321, -0.00004063, 0.00006143, -0.00004082, 0.00001137, 0.00000627, -0.00003151, 0.00004449, -0.00003136, 0.00001272, -0.00000006, -0.00000536, 0.00000633, -0.00002259, 0.00002106, -0.00002731, 0.00001890, -0.00000626, -0.00001364, 0.00001835, -0.00003319, 0.00002479, -0.00002640, 0.00000452, 0.00000151, -0.00003413, 0.00001426, -0.00005875, 0.00000968, -0.00015711, -0.00011689, 0.00042544, 0.00032014, -0.00011514, -0.00013875, -0.00001255, -0.00001263, -0.00002462, 0.00000955, -0.00002167, -0.00000223, -0.00000222, -0.00000839, 0.00001043, -0.00001271, 0.00000745, -0.00000739, -0.00000388, 0.00001619, -0.00001255, 0.00000436, -0.00000091, -0.00000970, 0.00001254, -0.00000221, -0.00001000, -0.00000035, -0.00000118, -0.00000528, -0.00000311, 0.00000206, -0.00001145, -0.00001369, 0.00000530, },
		{ 0.00000134, 0.00000726, 0.00000074, -0.00000270, -0.00000884, -0.00000056, 0.00001351, 0.00000670, -0.00000751, -0.00000659, 0.00000005, 0.00000400, 0.00000168, -0.00000452, 0.00000060, 0.00001035, 0.00000268, -0.00000652, -0.00000293, -0.00000372, 0.00000141, 0.00000051, 0.00000141, 0.00000907, -0.00000453, -0.00000047, -0.00000383, -0.00000879, 0.00001409, 0.00000251, -0.00000058, -0.00000287, -0.00001666, 0.00001235, 0.00000423, -0.00000769, 0.00001060, -0.00000254, -0.00000093, -0.00000164, 0.00000117, 0.00000251, -0.00000820, 0.00000726, -0.00000176, 0.00000415, -0.00000097, -0.00000506, 0.00001053, -0.00001091, 0.00000799, -0.00000655, 0.00000082, 0.00001573, -0.00001813, 0.00000073, -0.00000769, 0.00002215, -0.00000158, -0.00001571, 0.00001182, -0.00002083, 0.00003024, -0.00003092, 0.00000628, 0.00003193, -0.00005630, 0.00005477, -0.00001663, -0.00001300, 0.00002561, -0.00004247, 0.00004311, -0.00001809, -0.00000322, 0.00001389, -0.00001686, 0.00001411, -0.00002445, 0.00001616, -0.00001706, 0.00000376, 0.00000893, -0.00002321, 0.00002109, -0.00002839, 0.00001370, -0.00001115, -0.00001091, 0.00001290, -0.00003570, 0.00001069, -0.00004646, -0.00000344, -0.00011104, -0.00016132, 0.00031816, 0.00037903, -0.00002730, -0.00015935, -0.00003833, 0.00000077, -0.00003272, 0.00000909, -0.00001521, -0.00001106, 0.00000374, -0.00001268, 0.00001130, -0.00000974, 0.00000227, -0.00000123, -0.00000932, 0.00001700, -0.00000858, -0.00000074, 0.00000335, -0.00001146, 0.00000980, 0.00000168, -0.00001031, -0.00000104, -0.00000074, -0.00000451, -0.00000467, 0.00000188, -0.00000691, -0.00001607, },
		{ -0.00001613, -0.00000357, 0.00000836, 0.00000216, -0.00000241, -0.00000784, -0.00000383, 0.00001088, 0.00001056, -0.00000456, -0.00000840, -0.00000176, 0.00000371, 0.00000280, -0.00000366, -0.00000141, 0.00000947, 0.00000531, -0.00000653, -0.00000304, -0.00000371, 0.00000003, 0.00000215, -0.00000107, 0.00000937, -0.00000225, -0.00000261, 0.00000063, -0.00001180, 0.00000983, 0.00000677, -0.00000176, 0.00000130, -0.00001866, 0.00000640, 0.00001126, -0.00001055, 0.00000870, 0.00000128, -0.00000312, -0.00000059, -0.00000044, 0.00000491, -0.00000881, 0.00000483, 0.00000037, 0.00000185, 0.00000282, -0.00000815, 0.00001174, -0.00000965, 0.00000464, -0.00000156, -0.00000620, 0.00002087, -0.00001771, -0.00000086, -0.00000704, 0.00001506, 0.00001105, -0.00002489, 0.00001735, -0.00002273, 0.00002336, -0.00001241, -0.00002078, 0.00005143, -0.00005527, 0.00003301, 0.00001314, -0.00003336, 0.00003602, -0.00004230, 0.00003116, 0.00000157, -0.00001980, 0.00002442, -0.00002203, 0.00001374, -0.00001850, 0.00000575, -0.00000469, -0.00000932, 0.00001785, -0.00002455, 0.00001626, -0.00001846, 0.00000108, 0.00000147, -0.00001986, 0.00001564, -0.00002990, 0.00000285, -0.00003382, -0.00001392, -0.00007714, -0.00017791, 0.00021398, 0.00039574, 0.00006685, -0.00015287, -0.00007467, 0.00001009, -0.00003378, 0.00000186, -0.00000503, -0.00001965, 0.00000699, -0.00001347, 0.00000858, -0.00000433, -0.00000338, 0.00000348, -0.00001186, 0.00001480, -0.00000309, -0.00000546, 0.00000591, -0.00001111, 0.00000598, 0.00000502, -0.00000990, -0.00000189, -0.00000032, -0.00000379, -0.00000546, 0.00000047, -0.00000282, },
		{ 0.00000010, -0.00001413, -0.00000838, 0.00000777, 0.00000430, -0.00000210, -0.00000688, -0.00000588, 0.00000721, 0.00001286, -0.00000036, -0.00000925, -0.00000411, 0.00000311, 0.00000365, -0.00000250, -0.00000279, 0.00000773, 0.00000782, -0.00000591, -0.00000371, -0.00000297, -0.00000188, 0.00000370, -0.00000243, 0.00000798, 0.00000109, -0.00000521, 0.00000408, -0.00001167, 0.00000358, 0.00001040, -0.00000224, 0.00000427, -0.00001745, -0.00000178, 0.00001700, -0.00001011, 0.00000423, 0.00000558, -0.00000452, -0.00000024, -0.00000164, 0.00000633, -0.00000780, 0.00000145, 0.00000263, -0.00000053, 0.00000609, -0.00000985, 0.00001109, -0.00000644, -0.00000030, 0.00000428, -0.00001255, 0.00002340, -0.00001406, -0.00000539, -0.00000296, 0.00000509, 0.00002302, -0.00002969, 0.00001628, -0.00001638, 0.00000863, 0.00000946, -0.00004195, 0.00005585, -0.00003791, 0.00000265, 0.00003983, -0.00004325, 0.00003477, -0.00003169, 0.00001223, 0.00002220, -0.00003169, 0.00002774, -0.00001921, 0.00000610, -0.00000792, -0.00000573, 0.00000530, -0.00001687, 0.00001923, -0.00001904, 0.00000726, -0.00000767, -0.00000907, 0.00000876, -0.00002186, 0.00001197, -0.00002091, -0.00000468, -0.00002439, -0.00001993, -0.00005359, -0.00017722, 0.00012391, 0.00037794, 0.00015078, -0.00011772, -0.00011221, 0.00000875, -0.00002632, -0.00001030, 0.00000544, -0.00002491, 0.00000596, -0.00001069, 0.00000339, 0.00000165, -0.00000766, 0.00000562, -0.00001151, 0.00001056, 0.00000257, -0.00000872, 0.00000642, -0.00000908, 0.00000189, 0.00000734, -0.00000873, -0.00000286, 0.00000011, -0.00000323, -0.00000552, -0.00000152, },
		{ -0.00000349, 0.00000150, -0.00001073, -0.00001216, 0.00000547, 0.00000652, -0.00000130, -0.00000618, -0.00000687, 0.00000332, 0.00001324, 0.00000438, -0.00000860, -0.00000680, 0.00000205, 0.00000432, -0.00000133, -0.00000339, 0.00000544, 0.00000975, -0.00000437, -0.00000500, -0.00000176, -0.00000374, 0.00000447, -0.00000227, 0.00000515, 0.00000472, -0.00000722, 0.00000551, -0.00000856, -0.00000329, 0.00001224, -0.00000146, 0.00000556, -0.00001341, -0.00001044, 0.00001970, -0.00000610, -0.00000171, 0.00000904, -0.00000443, -0.00000072, -0.00000233, 0.00000668, -0.00000546, -0.00000226, 0.00000454, -0.00000257, 0.00000835, -0.00000979, 0.00000850, -0.00000154, -0.00000608, 0.00000972, -0.00001673, 0.00002214, -0.00000678, -0.00001242, 0.00000307, -0.00000494, 0.00003094, -0.00002761, 0.00000790, -0.00000363, -0.00000910, 0.00002770, -0.00005105, 0.00004396, -0.00000953, -0.00002730, 0.00005569, -0.00003951, 0.00002297, -0.00001447, -0.00000822, 0.00003825, -0.00003513, 0.00002263, -0.00000957, -0.00000556, 0.00000293, -0.00001417, 0.00001003, -0.00001790, 0.00001438, -0.00000989, -0.00000202, 0.00000058, -0.00001461, 0.00001040, -0.00001867, 0.00000536, -0.00001257, -0.00000923, -0.00001890, -0.00002257, -0.00003676, -0.00016764, 0.00005197, 0.00033771, 0.00021255, -0.00005968, -0.00013975, -0.00000697, -0.00001234, -0.00002344, 0.00001245, -0.00002483, 0.00000065, -0.00000563, -0.00000238, 0.00000635, -0.00000940, 0.00000506, -0.00000907, 0.00000564, 0.00000718, -0.00000999, 0.00000510, -0.00000614, -0.00000173, 0.00000842, -0.00000698, -0.00000387, 0.00000057, -0.00000284, -0.00000508, },
		{ -0.00000440, -0.00000500, 0.00000141, -0.00000677, -0.00001427, 0.00000187, 0.00000814, 0.00000016, -0.00000566, -0.00000718, -0.00000012, 0.00001179, 0.00000875, -0.00000621, -0.00000937, 0.00000028, 0.00000488, -0.00000033, -0.00000330, 0.00000306, 0.00001067, -0.00000187, -0.00000665, -0.00000057, -0.00000493, 0.00000398, -0.00000065, 0.00000161, 0.00000766, -0.00000772, 0.00000457, -0.00000339, -0.00000913, 0.00001149, 0.00000063, 0.00000529, -0.00000759, -0.00001768, 0.00001832, 0.00000075, -0.00000749, 0.00001045, -0.00000261, -0.00000182, -0.00000265, 0.00000609, -0.00000228, -0.00000567, 0.00000567, -0.00000390, 0.00000920, -0.00000780, 0.00000421, 0.00000431, -0.00001150, 0.00001332, -0.00001744, 0.00001664, 0.00000344, -0.00002034, 0.00000859, -0.00001195, 0.00003232, -0.00001807, -0.00000593, 0.00001131, -0.00002394, 0.00003661, -0.00004572, 0.00001944, 0.00002127, -0.00004806, 0.00005631, -0.00002328, 0.00000509, 0.00000379, -0.00002469, 0.00004577, -0.00002899, 0.00001068, 0.00000348, -0.00001695, 0.00001011, -0.00001708, 0.00000890, -0.00001372, 0.00000624, -0.00000074, -0.00000854, 0.00000463, -0.00001551, 0.00000795, -0.00001309, -0.00000101, -0.00000709, -0.00001041, -0.00001604, -0.00002404, -0.00002354, -0.00015402, -0.00000318, 0.00028746, 0.00024711, 0.00001002, -0.00014766, -0.00003617, 0.00000310, -0.00003287, 0.00001326, -0.00001920, -0.00000733, -0.00000043, -0.00000682, 0.00000850, -0.00000837, 0.00000257, -0.00000585, 0.00000128, 0.00001005, -0.00000931, 0.00000261, -0.00000313, -0.00000436, 0.00000832, -0.00000492, -0.00000480, 0.00000103, -0.00000254, },
		{ -0.00000222, -0.00000373, -0.00000580, 0.00000015, -0.00000307, -0.00001451, -0.00000234, 0.00000864, 0.00000216, -0.00000504, -0.00000717, -0.00000276, 0.00000899, 0.00001191, -0.00000223, -0.00001117, -0.00000236, 0.00000530, 0.00000053, -0.00000281, 0.00000108, 0.00001036, 0.00000135, -0.00000815, 0.00000001, -0.00000500, 0.00000220, 0.00000188, -0.00000164, 0.00000906, -0.00000634, 0.00000164, 0.00000236, -0.00001250, 0.00000799, 0.00000352, 0.00000411, -0.00000140, -0.00002195, 0.00001287, 0.00000881, -0.00001135, 0.00000916, 0.00000059, -0.00000299, -0.00000284, 0.00000480, 0.00000119, -0.00000822, 0.00000575, -0.00000428, 0.00000842, -0.00000406, -0.00000113, 0.00001000, -0.00001519, 0.00001382, -0.00001411, 0.00000749, 0.00001480, -0.00002672, 0.00001096, -0.00001366, 0.00002650, -0.00000291, -0.00002128, 0.00002328, -0.00003091, 0.00003359, -0.00002812, -0.00001033, 0.00004538, -0.00005384, 0.00004198, 0.00000039, -0.00001261, 0.00001773, -0.00003342, 0.00004350, -0.00001498, -0.00000441, 0.00001568, -0.00002426, 0.00001134, -0.00001423, 0.00000344, -0.00000706, -0.00000189, 0.00000562, -0.00001092, 0.00000468, -0.00001324, 0.00000369, -0.00000765, -0.00000521, -0.00000483, -0.00000947, -0.00001375, -0.00002607, -0.00001254, -0.00013792, -0.00004551, 0.00023663, 0.00025635, 0.00007807, -0.00013087, -0.00007314, 0.00001341, -0.00003500, 0.00000731, -0.00000983, -0.00001535, 0.00000275, -0.00000870, 0.00000789, -0.00000533, -0.00000060, -0.00000307, -0.00000179, 0.00001113, -0.00000722, -0.00000029, -0.00000068, -0.00000584, 0.00000734, -0.00000283, -0.00000560, 0.00000145, },
		{ 0.00000177, -0.00000176, -0.00000328, -0.00000582, -0.00000180, -0.00000025, -0.00001308, -0.00000640, 0.00000783, 0.00000431, -0.00000404, -0.00000705, -0.00000454, 0.00000556, 0.00001328, 0.00000276, -0.00001147, -0.00000582, 0.00000535, 0.00000139, -0.00000228, -0.00000012, 0.00000893, 0.00000476, -0.00000889, -0.00000047, -0.00000385, -0.00000047, 0.00000442, -0.00000369, 0.00000847, -0.00000331, -0.00000225, 0.00000709, -0.00001255, 0.00000240, 0.00000621, 0.00000289, 0.00000389, -0.00002249, 0.00000439, 0.00001601, -0.00001200, 0.00000531, 0.00000429, -0.00000361, -0.00000313, 0.00000307, 0.00000444, -0.00000953, 0.00000467, -0.00000361, 0.00000608, 0.00000088, -0.00000648, 0.00001423, -0.00001594, 0.00001061, -0.00000717, -0.00000355, 0.00002471, -0.00002903, 0.00000832, -0.00000949, 0.00001503, 0.00001412, -0.00003334, 0.00002788, -0.00002765, 0.00001994, -0.00000406, -0.00003667, 0.00005587, -0.00004352, 0.00001746, 0.00002429, -0.00002422, 0.00002391, -0.00003332, 0.00003299, 0.00000310, -0.00001819, 0.00002335, -0.00002538, 0.00000655, -0.00000741, -0.00000359, -0.00000088, -0.00000753, 0.00000797, -0.00000954, 0.00000228, -0.00000983, -0.00000035, -0.00000388, -0.00000688, -0.00000476, -0.00000828, -0.00001050, -0.00002901, -0.00000415, -0.00011907, -0.00007900, 0.00019010, 0.00024724, 0.00013277, -0.00009048, -0.00010869, 0.00001283, -0.00002881, -0.00000353, 0.00000016, -0.00002073, 0.00000260, -0.00000792, 0.00000527, -0.00000155, -0.00000317, -0.00000148, -0.00000343, 0.00001083, -0.00000443, -0.00000292, 0.00000089, -0.00000630, 0.00000586, -0.00000090, -0.00000627, },
		{ -0.00000685, 0.00000191, -0.00000101, -0.00000316, -0.00000518, -0.00000392, 0.00000132, -0.00001040, -0.00000969, 0.00000583, 0.00000613, -0.00000255, -0.00000683, -0.00000566, 0.00000216, 0.00001274, 0.00000785, -0.00000979, -0.00000969, 0.00000462, 0.00000246, -0.00000200, -0.00000046, 0.00000678, 0.00000768, -0.00000837, -0.00000212, -0.00000182, -0.00000325, 0.00000609, -0.00000395, 0.00000606, 0.00000056, -0.00000584, 0.00000959, -0.00000935, -0.00000398, 0.00000757, 0.00000233, 0.00000747, -0.00001946, -0.00000530, 0.00002043, -0.00000900, -0.00000016, 0.00000734, -0.00000317, -0.00000354, 0.00000109, 0.00000705, -0.00000939, 0.00000250, -0.00000199, 0.00000255, 0.00000612, -0.00001062, 0.00001582, -0.00001311, 0.00000401, 0.00000186, -0.00001397, 0.00003055, -0.00002555, 0.00000034, -0.00000079, 0.00000127, 0.00002845, -0.00003803, 0.00002305, -0.00001517, 0.00000034, 0.00001891, -0.00005229, 0.00005014, -0.00002088, -0.00000960, 0.00004136, -0.00002615, 0.00002184, -0.00002601, 0.00001778, 0.00002070, -0.00002684, 0.00002454, -0.00002035, -0.00000236, 0.00000048, -0.00000940, 0.00000271, -0.00000966, 0.00000668, -0.00000596, -0.00000065, -0.00000694, -0.00000307, -0.00000203, -0.00000676, -0.00000530, -0.00000823, -0.00000600, -0.00003185, 0.00000032, -0.00009700, -0.00010586, 0.00014842, 0.00022874, 0.00016703, -0.00003349, -0.00013280, -0.00000146, -0.00001640, -0.00001560, 0.00000742, -0.00002173, -0.00000088, -0.00000541, 0.00000196, 0.00000166, -0.00000437, -0.00000118, -0.00000406, 0.00000977, -0.00000152, -0.00000493, 0.00000157, -0.00000605, 0.00000418, 0.00000080, },
		{ 0.00000232, -0.00000733, 0.00000169, 0.00000016, -0.00000336, -0.00000414, -0.00000571, 0.00000152, -0.00000706, -0.00001179, 0.00000300, 0.00000723, -0.00000069, -0.00000632, -0.00000636, -0.00000074, 0.00001060, 0.00001203, -0.00000600, -0.00001326, 0.00000271, 0.00000387, -0.00000203, -0.00000018, 0.00000454, 0.00000950, -0.00000641, -0.00000464, 0.00000040, -0.00000532, 0.00000623, -0.00000240, 0.00000255, 0.00000417, -0.00000801, 0.00000934, -0.00000379, -0.00000949, 0.00000678, 0.00000272, 0.00000918, -0.00001384, -0.00001416, 0.00002082, -0.00000292, -0.00000573, 0.00000872, -0.00000153, -0.00000393, -0.00000098, 0.00000870, -0.00000777, -0.00000053, 0.00000028, -0.00000150, 0.00001055, -0.00001242, 0.00001409, -0.00000693, -0.00000461, 0.00001067, -0.00002112, 0.00003042, -0.00001605, -0.00001148, 0.00000939, -0.00001064, 0.00003614, -0.00003336, 0.00000975, 0.00000246, -0.00001877, 0.00003398, -0.00005345, 0.00003046, 0.00000678, -0.00003119, 0.00004696, -0.00001817, 0.00001384, -0.00001495, 0.00000209, 0.00003392, -0.00002833, 0.00001942, -0.00001107, -0.00001244, 0.00000658, -0.00001207, 0.00000302, -0.00000870, 0.00000322, -0.00000209, -0.00000262, -0.00000533, -0.00000440, -0.00000149, -0.00000605, -0.00000521, -0.00000971, -0.00000113, -0.00003291, -0.00000015, -0.00007233, -0.00012605, 0.00010940, 0.00020877, 0.00017964, 0.00002927, -0.00013756, -0.00002803, -0.00000251, -0.00002480, 0.00000964, -0.00001821, -0.00000642, -0.00000275, -0.00000074, 0.00000350, -0.00000409, -0.00000172, -0.00000428, 0.00000848, 0.00000115, -0.00000623, 0.00000154, -0.00000536, 0.00000249, },
		{ 0.00000081, 0.00000369, -0.00000758, 0.00000091, 0.00000177, -0.00000375, -0.00000305, -0.00000676, 0.00000045, -0.00000364, -0.00001256, -0.00000020, 0.00000745, 0.00000124, -0.00000535, -0.00000679, -0.00000299, 0.00000747, 0.00001452, -0.00000053, -0.00001559, -0.00000063, 0.00000545, -0.00000220, 0.00000028, 0.00000283, 0.00000989, -0.00000323, -0.00000738, 0.00000201, -0.00000603, 0.00000472, 0.00000039, -0.00000100, 0.00000649, -0.00000816, 0.00000665, 0.00000258, -0.00001266, 0.00000361, 0.00000372, 0.00000945, -0.00000703, -0.00002041, 0.00001696, 0.00000475, -0.00000979, 0.00000784, 0.00000097, -0.00000397, -0.00000303, 0.00000918, -0.00000484, -0.00000401, 0.00000269, -0.00000522, 0.00001308, -0.00001117, 0.00000916, 0.00000143, -0.00001315, 0.00001678, -0.00002304, 0.00002390, -0.00000208, -0.00002410, 0.00001731, -0.00001707, 0.00003516, -0.00002020, -0.00000826, 0.00001950, -0.00003133, 0.00003717, -0.00004093, 0.00000307, 0.00003102, -0.00004142, 0.00004020, -0.00000334, 0.00000401, -0.00000398, -0.00001059, 0.00004057, -0.00002272, 0.00000998, -0.00000046, -0.00002078, 0.00000897, -0.00001115, 0.00000066, -0.00000602, -0.00000065, 0.00000063, -0.00000302, -0.00000485, -0.00000496, -0.00000137, -0.00000565, -0.00000408, -0.00001206, 0.00000261, -0.00003078, -0.00000544, -0.00004712, -0.00013803, 0.00007027, 0.00019198, 0.00017477, 0.00008612, -0.00011948, -0.00006132, 0.00000709, -0.00002807, 0.00000632, -0.00001161, -0.00001201, -0.00000140, -0.00000208, 0.00000385, -0.00000279, -0.00000242, -0.00000460, 0.00000719, 0.00000356, -0.00000692, 0.00000098, -0.00000442, },
		{ -0.00000329, -0.00000084, 0.00000486, -0.00000739, -0.00000061, 0.00000370, -0.00000401, -0.00000230, -0.00000683, -0.00000157, -0.00000070, -0.00001201, -0.00000336, 0.00000685, 0.00000286, -0.00000389, -0.00000692, -0.00000466, 0.00000406, 0.00001493, 0.00000574, -0.00001581, -0.00000529, 0.00000678, -0.00000210, 0.00000041, 0.00000208, 0.00000892, 0.00000054, -0.00000954, 0.00000236, -0.00000517, 0.00000195, 0.00000350, -0.00000354, 0.00000693, -0.00000639, 0.00000252, 0.00000813, -0.00001262, -0.00000135, 0.00000455, 0.00000900, -0.00000044, -0.00002305, 0.00000970, 0.00001207, -0.00001110, 0.00000484, 0.00000365, -0.00000336, -0.00000483, 0.00000837, -0.00000095, -0.00000737, 0.00000458, -0.00000770, 0.00001296, -0.00000687, 0.00000204, 0.00001003, -0.00001928, 0.00001827, -0.00001907, 0.00001228, 0.00001332, -0.00003383, 0.00001971, -0.00001612, 0.00002609, -0.00000198, -0.00002567, 0.00003046, -0.00003360, 0.00002842, -0.00001948, -0.00002392, 0.00004489, -0.00003821, 0.00002408, 0.00001321, -0.00000345, 0.00000404, -0.00001839, 0.00004041, -0.00001185, -0.00000083, 0.00000864, -0.00002540, 0.00000714, -0.00000746, -0.00000289, -0.00000319, -0.00000360, 0.00000160, -0.00000213, -0.00000476, -0.00000549, -0.00000107, -0.00000586, -0.00000231, -0.00001409, 0.00000389, -0.00002511, -0.00001401, -0.00002440, -0.00014005, 0.00002954, 0.00017909, 0.00016000, 0.00012765, -0.00008059, -0.00009305, 0.00000751, -0.00002459, -0.00000101, -0.00000438, -0.00001569, -0.00000217, -0.00000209, 0.00000322, -0.00000122, -0.00000269, -0.00000521, 0.00000582, 0.00000579, -0.00000707, -0.00000001, },
		{ -0.00000144, -0.00000202, -0.00000239, 0.00000563, -0.00000649, -0.00000287, 0.00000563, -0.00000375, -0.00000220, -0.00000593, -0.00000404, 0.00000129, -0.00001032, -0.00000616, 0.00000565, 0.00000391, -0.00000212, -0.00000657, -0.00000594, 0.00000091, 0.00001336, 0.00001166, -0.00001338, -0.00001070, 0.00000721, -0.00000134, -0.00000013, 0.00000232, 0.00000707, 0.00000411, -0.00001041, 0.00000118, -0.00000304, -0.00000128, 0.00000590, -0.00000438, 0.00000551, -0.00000341, -0.00000176, 0.00001155, -0.00000938, -0.00000689, 0.00000428, 0.00000849, 0.00000494, -0.00002199, 0.00000069, 0.00001716, -0.00000919, 0.00000050, 0.00000564, -0.00000197, -0.00000613, 0.00000631, 0.00000336, -0.00000994, 0.00000530, -0.00000830, 0.00001001, -0.00000033, -0.00000556, 0.00001670, -0.00002111, 0.00001440, -0.00001022, -0.00000164, 0.00002641, -0.00003746, 0.00001495, -0.00000830, 0.00001194, 0.00001632, -0.00003729, 0.00003185, -0.00002528, 0.00001139, 0.00000387, -0.00004311, 0.00004500, -0.00002378, 0.00000422, 0.00002617, -0.00000572, 0.00000808, -0.00002123, 0.00003486, 0.00000150, -0.00001013, 0.00001430, -0.00002569, 0.00000192, -0.00000262, -0.00000611, -0.00000134, -0.00000510, 0.00000104, -0.00000078, -0.00000430, -0.00000641, -0.00000051, -0.00000636, -0.00000072, -0.00001468, 0.00000219, -0.00001677, -0.00002330, -0.00000720, -0.00013147, -0.00001203, 0.00016757, 0.00014358, 0.00014897, -0.00002785, -0.00011452, -0.00000349, -0.00001615, -0.00000950, 0.00000098, -0.00001632, -0.00000487, -0.00000147, 0.00000237, -0.00000002, -0.00000230, -0.00000592, 0.00000414, 0.00000790, -0.00000663, },
		{ -0.00000544, -0.00000331, -0.00000074, -0.00000362, 0.00000574, -0.00000471, -0.00000565, 0.00000702, -0.00000259, -0.00000286, -0.00000431, -0.00000639, 0.00000201, -0.00000773, -0.00000841, 0.00000410, 0.00000432, -0.00000041, -0.00000555, -0.00000694, -0.00000170, 0.00001036, 0.00001612, -0.00000831, -0.00001583, 0.00000607, 0.00000031, -0.00000136, 0.00000323, 0.00000506, 0.00000672, -0.00000963, -0.00000131, -0.00000034, -0.00000402, 0.00000686, -0.00000339, 0.00000285, -0.00000025, -0.00000501, 0.00001223, -0.00000382, -0.00001146, 0.00000222, 0.00000821, 0.00000874, -0.00001797, -0.00000810, 0.00001877, -0.00000453, -0.00000393, 0.00000625, 0.00000006, -0.00000659, 0.00000321, 0.00000740, -0.00001106, 0.00000438, -0.00000679, 0.00000479, 0.00000694, -0.00001164, 0.00001961, -0.00001784, 0.00000598, 0.00000109, -0.00001433, 0.00003387, -0.00003333, 0.00000362, 0.00000364, -0.00000280, 0.00002979, -0.00003973, 0.00002327, -0.00000944, -0.00000797, 0.00002216, -0.00005004, 0.00003227, -0.00000359, -0.00001311, 0.00003181, -0.00000228, 0.00000896, -0.00002039, 0.00002610, 0.00001457, -0.00001588, 0.00001587, -0.00002231, -0.00000507, 0.00000168, -0.00000794, -0.00000083, -0.00000532, -0.00000030, 0.00000022, -0.00000309, -0.00000757, -0.00000008, -0.00000654, -0.00000008, -0.00001331, -0.00000199, -0.00000759, -0.00003059, 0.00000251, -0.00011351, -0.00005155, 0.00015335, 0.00013189, 0.00015048, 0.00002880, -0.00011910, -0.00002453, -0.00000656, -0.00001602, 0.00000283, -0.00001396, -0.00000851, -0.00000121, 0.00000192, 0.00000055, -0.00000142, -0.00000636, 0.00000198, 0.00000974, },
		{ 0.00001098, -0.00000331, -0.00000548, 0.00000026, -0.00000422, 0.00000497, -0.00000212, -0.00000842, 0.00000730, -0.00000032, -0.00000413, -0.00000244, -0.00000800, 0.00000132, -0.00000463, -0.00000993, 0.00000235, 0.00000428, 0.00000081, -0.00000385, -0.00000758, -0.00000380, 0.00000669, 0.00001833, -0.00000124, -0.00001949, 0.00000292, 0.00000270, -0.00000295, 0.00000419, 0.00000362, 0.00000791, -0.00000733, -0.00000447, 0.00000203, -0.00000551, 0.00000613, -0.00000105, -0.00000009, 0.00000210, -0.00000654, 0.00001040, 0.00000259, -0.00001370, -0.00000171, 0.00000793, 0.00001113, -0.00001223, -0.00001496, 0.00001658, 0.00000166, -0.00000712, 0.00000517, 0.00000227, -0.00000599, -0.00000047, 0.00001042, -0.00001028, 0.00000173, -0.00000353, -0.00000152, 0.00001312, -0.00001445, 0.00001788, -0.00001008, -0.00000484, 0.00001164, -0.00002252, 0.00003381, -0.00002193, -0.00001154, 0.00001558, -0.00001360, 0.00003508, -0.00003245, 0.00000746, 0.00000863, -0.00002343, 0.00003067, -0.00004421, 0.00001133, 0.00001559, -0.00002304, 0.00002915, 0.00000509, 0.00000868, -0.00001757, 0.00001623, 0.00002533, -0.00001729, 0.00001392, -0.00001667, -0.00001213, 0.00000425, -0.00000809, -0.00000137, -0.00000487, -0.00000169, 0.00000040, -0.00000131, -0.00000847, -0.00000031, -0.00000595, -0.00000068, -0.00001026, -0.00000733, 0.00000030, -0.00003384, 0.00000446, -0.00008926, -0.00008502, 0.00013277, 0.00012790, 0.00013714, 0.00007909, -0.00010412, -0.00005091, -0.00000033, -0.00001840, 0.00000099, -0.00000978, -0.00001169, -0.00000205, 0.00000205, 0.00000067, -0.00000047, -0.00000619, -0.00000058, },
		{ -0.00000320, 0.00001123, -0.00000026, -0.00000757, 0.00000056, -0.00000394, 0.00000331, 0.00000086, -0.00001048, 0.00000601, 0.00000295, -0.00000554, -0.00000096, -0.00000841, -0.00000066, -0.00000150, -0.00001056, 0.00000042, 0.00000410, 0.00000125, -0.00000171, -0.00000761, -0.00000556, 0.00000306, 0.00001804, 0.00000666, -0.00002055, -0.00000217, 0.00000527, -0.00000428, 0.00000454, 0.00000324, 0.00000767, -0.00000406, -0.00000746, 0.00000330, -0.00000543, 0.00000399, 0.00000180, -0.00000233, 0.00000301, -0.00000634, 0.00000695, 0.00000825, -0.00001287, -0.00000690, 0.00000702, 0.00001259, -0.00000604, -0.00001882, 0.00001124, 0.00000769, -0.00000813, 0.00000267, 0.00000403, -0.00000427, -0.00000411, 0.00001176, -0.00000745, -0.00000226, 0.00000056, -0.00000738, 0.00001651, -0.00001306, 0.00001191, 0.00000030, -0.00001511, 0.00001836, -0.00002421, 0.00002638, -0.00000585, -0.00002647, 0.00002331, -0.00001738, 0.00003145, -0.00001779, -0.00001075, 0.00002324, -0.00003051, 0.00002814, -0.00002896, -0.00001119, 0.00002791, -0.00002362, 0.00002005, 0.00001322, 0.00000934, -0.00001404, 0.00000670, 0.00003278, -0.00001462, 0.00000971, -0.00001031, -0.00001796, 0.00000468, -0.00000695, -0.00000234, -0.00000435, -0.00000269, -0.00000017, 0.00000049, -0.00000849, -0.00000151, -0.00000455, -0.00000226, -0.00000642, -0.00001224, 0.00000525, -0.00003232, 0.00000010, -0.00006290, -0.00010869, 0.00010409, 0.00013080, 0.00011650, 0.00011503, -0.00007153, -0.00007584, -0.00000108, -0.00001628, -0.00000332, -0.00000551, -0.00001325, -0.00000419, 0.00000244, 0.00000075, 0.00000016, -0.00000533, },
		{ -0.00000401, -0.00000539, 0.00001022, 0.00000341, -0.00000900, -0.00000023, -0.00000273, 0.00000108, 0.00000360, -0.00001118, 0.00000303, 0.00000669, -0.00000638, -0.00000044, -0.00000742, -0.00000352, 0.00000109, -0.00001013, -0.00000170, 0.00000404, 0.00000093, 0.00000038, -0.00000672, -0.00000716, -0.00000008, 0.00001557, 0.00001397, -0.00001833, -0.00000856, 0.00000717, -0.00000475, 0.00000385, 0.00000399, 0.00000645, -0.00000060, -0.00000949, 0.00000310, -0.00000397, 0.00000119, 0.00000421, -0.00000323, 0.00000247, -0.00000496, 0.00000307, 0.00001193, -0.00000907, -0.00001210, 0.00000474, 0.00001355, -0.00000034, -0.00001946, 0.00000415, 0.00001200, -0.00000667, -0.00000051, 0.00000475, -0.00000169, -0.00000697, 0.00001103, -0.00000290, -0.00000675, 0.00000425, -0.00001127, 0.00001611, -0.00000768, 0.00000331, 0.00001058, -0.00002192, 0.00001930, -0.00001927, 0.00001373, 0.00001098, -0.00003699, 0.00002395, -0.00001337, 0.00002085, -0.00000016, -0.00002601, 0.00003013, -0.00002777, 0.00001688, -0.00000995, -0.00002893, 0.00003025, -0.00001615, 0.00000833, 0.00001884, 0.00001206, -0.00001016, -0.00000188, 0.00003673, -0.00000880, 0.00000468, -0.00000444, -0.00002196, 0.00000319, -0.00000523, -0.00000318, -0.00000403, -0.00000330, -0.00000108, 0.00000167, -0.00000734, -0.00000352, -0.00000276, -0.00000412, -0.00000286, -0.00001539, 0.00000651, -0.00002666, -0.00000789, -0.00003866, -0.00012030, 0.00006818, 0.00013690, 0.00009638, 0.00013262, -0.00002729, -0.00009242, -0.00001022, -0.00001117, -0.00000804, -0.00000276, -0.00001273, -0.00000714, 0.00000246, 0.00000118, 0.00000033, },
		{ 0.00000021, -0.00000266, -0.00000674, 0.00000802, 0.00000710, -0.00000918, -0.00000227, -0.00000090, -0.00000109, 0.00000537, -0.00001010, -0.00000129, 0.00001006, -0.00000594, -0.00000121, -0.00000526, -0.00000657, 0.00000257, -0.00000857, -0.00000401, 0.00000417, 0.00000017, 0.00000179, -0.00000476, -0.00000854, -0.00000262, 0.00001168, 0.00001940, -0.00001285, -0.00001507, 0.00000749, -0.00000395, 0.00000203, 0.00000552, 0.00000494, 0.00000227, -0.00001009, 0.00000149, -0.00000171, -0.00000143, 0.00000549, -0.00000265, 0.00000097, -0.00000329, -0.00000015, 0.00001309, -0.00000323, -0.00001588, 0.00000067, 0.00001412, 0.00000444, -0.00001741, -0.00000302, 0.00001354, -0.00000320, -0.00000335, 0.00000410, 0.00000120, -0.00000838, 0.00000824, 0.00000256, -0.00001059, 0.00000633, -0.00001222, 0.00001189, 0.00000030, -0.00000550, 0.00001806, -0.00002334, 0.00001420, -0.00000947, -0.00000056, 0.00002441, -0.00004015, 0.00001683, -0.00000337, 0.00000721, 0.00001546, -0.00003420, 0.00002775, -0.00001696, 0.00000158, 0.00000677, -0.00003773, 0.00002293, -0.00000448, -0.00000180, 0.00001994, 0.00001657, -0.00000543, -0.00000936, 0.00003745, -0.00000092, -0.00000004, 0.00000034, -0.00002411, 0.00000038, -0.00000356, -0.00000368, -0.00000380, -0.00000381, -0.00000184, 0.00000187, -0.00000515, -0.00000578, -0.00000125, -0.00000553, -0.00000044, -0.00001618, 0.00000443, -0.00001857, -0.00001649, -0.00001978, -0.00011974, 0.00002823, 0.00014122, 0.00008273, 0.00013240, 0.00002032, -0.00009557, -0.00002639, -0.00000597, -0.00001113, -0.00000232, -0.00001055, -0.00001000, 0.00000156, 0.00000207, },
		{ 0.00000312, 0.00000021, -0.00000166, -0.00000710, 0.00000507, 0.00001006, -0.00000770, -0.00000528, 0.00000088, -0.00000248, 0.00000565, -0.00000733, -0.00000612, 0.00001205, -0.00000375, -0.00000316, -0.00000251, -0.00000899, 0.00000251, -0.00000599, -0.00000632, 0.00000430, -0.00000048, 0.00000202, -0.00000188, -0.00000942, -0.00000474, 0.00000728, 0.00002209, -0.00000486, -0.00002026, 0.00000559, -0.00000190, -0.00000054, 0.00000714, 0.00000391, 0.00000405, -0.00000926, -0.00000103, 0.00000054, -0.00000316, 0.00000538, -0.00000100, -0.00000068, -0.00000211, -0.00000213, 0.00001197, 0.00000319, -0.00001701, -0.00000506, 0.00001398, 0.00000835, -0.00001364, -0.00000879, 0.00001210, 0.00000120, -0.00000493, 0.00000220, 0.00000366, -0.00000796, 0.00000390, 0.00000780, -0.00001266, 0.00000598, -0.00001005, 0.00000490, 0.00000874, -0.00001201, 0.00002080, -0.00001895, 0.00000456, 0.00000204, -0.00001264, 0.00003136, -0.00003508, 0.00000367, 0.00000898, -0.00000493, 0.00002511, -0.00003362, 0.00001752, -0.00000221, -0.00001236, 0.00001676, -0.00003664, 0.00000924, 0.00000660, -0.00000723, 0.00001641, 0.00002126, 0.00000094, -0.00001558, 0.00003521, 0.00000810, -0.00000374, 0.00000387, -0.00002469, -0.00000321, -0.00000224, -0.00000400, -0.00000338, -0.00000451, -0.00000220, 0.00000114, -0.00000252, -0.00000752, -0.00000062, -0.00000601, 0.00000051, -0.00001476, 0.00000017, -0.00001011, -0.00002315, -0.00000785, -0.00010901, -0.00001126, 0.00013925, 0.00007834, 0.00011869, 0.00006290, -0.00008343, -0.00004578, -0.00000381, -0.00001145, -0.00000394, -0.00000775, -0.00001186, -0.00000047, },
		{ -0.00000329, 0.00000379, 0.00000069, -0.00000116, -0.00000670, 0.00000201, 0.00001165, -0.00000459, -0.00000858, 0.00000178, -0.00000253, 0.00000437, -0.00000348, -0.00001033, 0.00001184, 0.00000020, -0.00000570, -0.00000003, -0.00001005, 0.00000080, -0.00000279, -0.00000825, 0.00000406, -0.00000045, 0.00000091, 0.00000135, -0.00000933, -0.00000672, 0.00000313, 0.00002183, 0.00000429, -0.00002280, 0.00000135, 0.00000088, -0.00000319, 0.00000813, 0.00000384, 0.00000465, -0.00000737, -0.00000374, 0.00000210, -0.00000367, 0.00000415, 0.00000099, -0.00000171, -0.00000188, -0.00000289, 0.00000943, 0.00000867, -0.00001494, -0.00001161, 0.00001249, 0.00001165, -0.00000917, -0.00001228, 0.00000828, 0.00000519, -0.00000471, -0.00000041, 0.00000496, -0.00000576, -0.00000104, 0.00001162, -0.00001215, 0.00000306, -0.00000551, -0.00000303, 0.00001526, -0.00001436, 0.00001826, -0.00001002, -0.00000675, 0.00001174, -0.00001948, 0.00003062, -0.00002322, -0.00001199, 0.00001942, -0.00001191, 0.00002700, -0.00002532, 0.00000310, 0.00001158, -0.00002071, 0.00001836, -0.00002785, -0.00000594, 0.00001306, -0.00000706, 0.00001002, 0.00002402, 0.00000934, -0.00001997, 0.00003011, 0.00001754, -0.00000602, 0.00000625, -0.00002397, -0.00000726, -0.00000127, -0.00000445, -0.00000253, -0.00000543, -0.00000224, -0.00000007, -0.00000020, -0.00000814, -0.00000111, -0.00000556, 0.00000017, -0.00001185, -0.00000480, -0.00000311, -0.00002641, -0.00000272, -0.00009153, -0.00004574, 0.00012830, 0.00008268, 0.00009798, 0.00009397, -0.00005772, -0.00006322, -0.00000693, -0.00000923, -0.00000652, -0.00000550, -0.00001219, },
		{ -0.00001114, -0.00000621, 0.00000353, 0.00000173, -0.00000096, -0.00000602, -0.00000051, 0.00001160, -0.00000038, -0.00001120, 0.00000109, -0.00000108, 0.00000200, 0.00000043, -0.00001284, 0.00000909, 0.00000531, -0.00000787, 0.00000130, -0.00000933, -0.00000226, 0.00000039, -0.00000930, 0.00000303, 0.00000055, -0.00000125, 0.00000412, -0.00000785, -0.00000869, -0.00000034, 0.00001906, 0.00001299, -0.00002181, -0.00000468, 0.00000351, -0.00000517, 0.00000794, 0.00000486, 0.00000441, -0.00000506, -0.00000598, 0.00000259, -0.00000309, 0.00000238, 0.00000257, -0.00000169, -0.00000249, -0.00000294, 0.00000655, 0.00001210, -0.00000997, -0.00001761, 0.00000902, 0.00001455, -0.00000471, -0.00001333, 0.00000326, 0.00000761, -0.00000275, -0.00000286, 0.00000465, -0.00000232, -0.00000543, 0.00001306, -0.00000893, -0.00000174, -0.00000006, -0.00000981, 0.00001804, -0.00001191, 0.00001145, 0.00000090, -0.00001645, 0.00001681, -0.00001978, 0.00002323, -0.00000785, -0.00002593, 0.00002446, -0.00001202, 0.00002190, -0.00001253, -0.00001094, 0.00002035, -0.00002163, 0.00001284, -0.00001555, -0.00001795, 0.00001300, -0.00000267, 0.00000358, 0.00002318, 0.00001917, -0.00002149, 0.00002224, 0.00002672, -0.00000657, 0.00000751, -0.00002204, -0.00001168, -0.00000046, -0.00000528, -0.00000133, -0.00000629, -0.00000234, -0.00000119, 0.00000120, -0.00000741, -0.00000250, -0.00000456, -0.00000090, -0.00000842, -0.00000915, 0.00000132, -0.00002599, -0.00000291, -0.00007122, -0.00007178, 0.00010803, 0.00009263, 0.00007697, 0.00011026, -0.00002314, -0.00007377, -0.00001576, -0.00000595, -0.00000862, -0.00000463, },
		{ -0.00000521, -0.00000940, -0.00000841, 0.00000213, 0.00000305, -0.00000062, -0.00000556, -0.00000212, 0.00001015, 0.00000401, -0.00001219, -0.00000146, 0.00000140, -0.00000053, 0.00000330, -0.00001295, 0.00000417, 0.00001046, -0.00000863, 0.00000091, -0.00000695, -0.00000591, 0.00000272, -0.00000904, 0.00000101, 0.00000242, -0.00000370, 0.00000553, -0.00000484, -0.00001051, -0.00000313, 0.00001467, 0.00001977, -0.00001719, -0.00001135, 0.00000500, -0.00000593, 0.00000643, 0.00000662, 0.00000392, -0.00000293, -0.00000735, 0.00000204, -0.00000190, 0.00000075, 0.00000325, -0.00000069, -0.00000340, -0.00000303, 0.00000424, 0.00001311, -0.00000321, -0.00002156, 0.00000331, 0.00001689, -0.00000056, -0.00001238, -0.00000162, 0.00000781, 0.00000028, -0.00000429, 0.00000273, 0.00000144, -0.00000817, 0.00001175, -0.00000359, -0.00000712, 0.00000455, -0.00001370, 0.00001630, -0.00000549, 0.00000261, 0.00001077, -0.00002185, 0.00001595, -0.00001428, 0.00001195, 0.00000704, -0.00003459, 0.00002243, -0.00000596, 0.00001262, 0.00000058, -0.00002072, 0.00002216, -0.00001594, 0.00000357, -0.00000424, -0.00002387, 0.00000710, 0.00000296, -0.00000027, 0.00001836, 0.00002879, -0.00001896, 0.00001199, 0.00003476, -0.00000504, 0.00000761, -0.00001888, -0.00001647, 0.00000025, -0.00000639, -0.00000020, -0.00000660, -0.00000293, -0.00000178, 0.00000147, -0.00000562, -0.00000422, -0.00000356, -0.00000209, -0.00000529, -0.00001201, 0.00000293, -0.00002264, -0.00000628, -0.00005158, -0.00008770, 0.00008034, 0.00010377, 0.00006092, 0.00011199, 0.00001399, -0.00007404, -0.00002874, -0.00000375, -0.00000914, },
		{ -0.00000785, -0.00000661, -0.00000786, -0.00000941, -0.00000015, 0.00000401, 0.00000034, -0.00000556, -0.00000291, 0.00000795, 0.00000756, -0.00001100, -0.00000551, 0.00000397, -0.00000216, 0.00000431, -0.00001064, -0.00000193, 0.00001430, -0.00000719, -0.00000122, -0.00000354, -0.00000918, 0.00000349, -0.00000728, -0.00000181, 0.00000464, -0.00000550, 0.00000499, -0.00000065, -0.00001169, -0.00000551, 0.00000968, 0.00002364, -0.00000962, -0.00001718, 0.00000458, -0.00000535, 0.00000389, 0.00000854, 0.00000376, -0.00000136, -0.00000781, 0.00000082, -0.00000068, -0.00000023, 0.00000297, 0.00000082, -0.00000390, -0.00000374, 0.00000289, 0.00001209, 0.00000376, -0.00002230, -0.00000429, 0.00001815, 0.00000338, -0.00001023, -0.00000524, 0.00000587, 0.00000335, -0.00000409, -0.00000028, 0.00000441, -0.00000859, 0.00000801, 0.00000264, -0.00001150, 0.00000682, -0.00001385, 0.00001059, 0.00000293, -0.00000552, 0.00001703, -0.00002160, 0.00000975, -0.00000535, 0.00000037, 0.00001792, -0.00003612, 0.00001392, 0.00000365, 0.00000294, 0.00001037, -0.00002418, 0.00001750, -0.00000651, -0.00000543, 0.00000277, -0.00002326, -0.00000193, 0.00000657, -0.00000018, 0.00001076, 0.00003592, -0.00001163, 0.00000043, 0.00004047, -0.00000101, 0.00000651, -0.00001454, -0.00002141, 0.00000063, -0.00000738, 0.00000026, -0.00000593, -0.00000419, -0.00000178, 0.00000080, -0.00000338, -0.00000558, -0.00000305, -0.00000299, -0.00000293, -0.00001315, 0.00000217, -0.00001763, -0.00001068, -0.00003503, -0.00009366, 0.00004856, 0.00011184, 0.00005267, 0.00010232, 0.00004743, -0.00006312, -0.00004275, -0.00000460, },
		{ -0.00000948, -0.00000552, -0.00000780, -0.00000724, -0.00000925, -0.00000262, 0.00000399, 0.00000209, -0.00000585, -0.00000336, 0.00000586, 0.00000951, -0.00000770, -0.00001011, 0.00000544, -0.00000203, 0.00000328, -0.00000664, -0.00000776, 0.00001565, -0.00000336, -0.00000448, -0.00000013, -0.00001107, 0.00000233, -0.00000429, -0.00000488, 0.00000637, -0.00000581, 0.00000242, 0.00000391, -0.00001159, -0.00000783, 0.00000493, 0.00002433, -0.00000044, -0.00002079, 0.00000199, -0.00000377, 0.00000090, 0.00000994, 0.00000433, -0.00000038, -0.00000763, -0.00000060, 0.00000011, -0.00000046, 0.00000209, 0.00000212, -0.00000341, -0.00000522, 0.00000228, 0.00000999, 0.00000942, -0.00001939, -0.00001270, 0.00001755, 0.00000731, -0.00000763, -0.00000702, 0.00000249, 0.00000537, -0.00000217, -0.00000345, 0.00000567, -0.00000664, 0.00000283, 0.00000821, -0.00001350, 0.00000600, -0.00001055, 0.00000256, 0.00001079, -0.00001051, 0.00001829, -0.00001611, 0.00000038, 0.00000377, -0.00000826, 0.00002275, -0.00003078, 0.00000146, 0.00001322, -0.00000378, 0.00001480, -0.00002152, 0.00000885, 0.00000301, -0.00001098, 0.00000435, -0.00001799, -0.00001050, 0.00000595, 0.00000341, 0.00000275, 0.00003847, 0.00000017, -0.00001062, 0.00004257, 0.00000573, 0.00000439, -0.00000933, -0.00002599, 0.00000016, -0.00000767, -0.00000044, -0.00000416, -0.00000594, -0.00000148, -0.00000026, -0.00000141, -0.00000608, -0.00000318, -0.00000350, -0.00000146, -0.00001276, -0.00000011, -0.00001228, -0.00001449, -0.00002276, -0.00009126, 0.00001649, 0.00011373, 0.00005250, 0.00008602, 0.00007243, -0.00004268, -0.00005407, },
		{ -0.00005948, -0.00001796, -0.00000359, -0.00000791, -0.00000763, -0.00000854, -0.00000445, 0.00000265, 0.00000435, -0.00000587, -0.00000413, 0.00000451, 0.00000968, -0.00000304, -0.00001394, 0.00000482, 0.00000009, 0.00000074, -0.00000221, -0.00001188, 0.00001392, 0.00000230, -0.00000775, 0.00000214, -0.00001094, -0.00000059, -0.00000079, -0.00000737, 0.00000685, -0.00000424, -0.00000158, 0.00000768, -0.00000968, -0.00001021, 0.00000085, 0.00002223, 0.00000871, -0.00002121, -0.00000239, -0.00000192, -0.00000185, 0.00001033, 0.00000566, 0.00000030, -0.00000722, -0.00000182, 0.00000034, -0.00000018, 0.00000121, 0.00000261, -0.00000179, -0.00000711, 0.00000177, 0.00000790, 0.00001274, -0.00001333, -0.00002035, 0.00001439, 0.00001131, -0.00000505, -0.00000703, -0.00000125, 0.00000560, 0.00000095, -0.00000570, 0.00000476, -0.00000296, -0.00000235, 0.00001171, -0.00001240, 0.00000233, -0.00000510, -0.00000550, 0.00001577, -0.00001105, 0.00001480, -0.00000731, -0.00000916, 0.00001011, -0.00001200, 0.00002147, -0.00002069, -0.00001149, 0.00001950, -0.00000564, 0.00001389, -0.00001484, -0.00000046, 0.00000956, -0.00001174, 0.00000156, -0.00001122, -0.00001559, 0.00000075, 0.00000839, -0.00000297, 0.00003541, 0.00001485, -0.00001883, 0.00004004, 0.00001490, 0.00000187, -0.00000393, -0.00002945, -0.00000177, -0.00000680, -0.00000245, -0.00000165, -0.00000762, -0.00000140, -0.00000115, -0.00000021, -0.00000559, -0.00000381, -0.00000377, -0.00000072, -0.00001131, -0.00000298, -0.00000763, -0.00001679, -0.00001481, -0.00008292, -0.00001250, 0.00010807, 0.00005858, 0.00006814, 0.00008658, -0.00001631, },
		{ 0.00001149, -0.00005717, -0.00002827, -0.00000358, -0.00000666, -0.00000854, -0.00000812, -0.00000508, 0.00000018, 0.00000637, -0.00000489, -0.00000570, 0.00000408, 0.00000850, 0.00000180, -0.00001579, 0.00000170, 0.00000367, -0.00000215, 0.00000125, -0.00001332, 0.00000933, 0.00000857, -0.00000973, 0.00000248, -0.00000876, -0.00000454, 0.00000223, -0.00000848, 0.00000560, -0.00000103, -0.00000591, 0.00000952, -0.00000585, -0.00001241, -0.00000255, 0.00001823, 0.00001633, -0.00001818, -0.00000761, -0.00000064, -0.00000386, 0.00000958, 0.00000743, 0.00000109, -0.00000688, -0.00000272, 0.00000017, 0.00000014, 0.00000085, 0.00000206, 0.00000060, -0.00000864, 0.00000062, 0.00000662, 0.00001347, -0.00000542, -0.00002560, 0.00000843, 0.00001505, -0.00000256, -0.00000578, -0.00000424, 0.00000383, 0.00000433, -0.00000618, 0.00000188, 0.00000125, -0.00000610, 0.00001228, -0.00000840, -0.00000304, 0.00000059, -0.00001143, 0.00001648, -0.00000725, 0.00000821, 0.00000208, -0.00001606, 0.00001189, -0.00001068, 0.00001576, -0.00000903, -0.00002164, 0.00002056, -0.00000264, 0.00000947, -0.00000711, -0.00000749, 0.00001164, -0.00000838, -0.00000306, -0.00000606, -0.00001585, -0.00000739, 0.00001186, -0.00000450, 0.00002724, 0.00002969, -0.00002195, 0.00003267, 0.00002545, 0.00000003, 0.00000076, -0.00003099, -0.00000555, -0.00000475, -0.00000538, 0.00000087, -0.00000855, -0.00000199, -0.00000151, 0.00000006, -0.00000436, -0.00000451, -0.00000404, -0.00000051, -0.00000929, -0.00000571, -0.00000428, -0.00001737, -0.00001049, -0.00007120, -0.00003604, 0.00009519, 0.00006793, 0.00005281, 0.00008994, },
		{ 0.00008451, 0.00003650, -0.00004713, -0.00003790, -0.00000644, -0.00000460, -0.00000903, -0.00000869, -0.00000450, -0.00000263, 0.00000720, -0.00000245, -0.00000803, 0.00000418, 0.00000692, 0.00000556, -0.00001493, -0.00000351, 0.00000752, -0.00000404, 0.00000269, -0.00001190, 0.00000293, 0.00001385, -0.00000936, 0.00000068, -0.00000514, -0.00000841, 0.00000380, -0.00000776, 0.00000277, 0.00000299, -0.00000922, 0.00000872, -0.00000054, -0.00001384, -0.00000555, 0.00001333, 0.00002135, -0.00001225, -0.00001243, -0.00000057, -0.00000496, 0.00000788, 0.00000916, 0.00000235, -0.00000668, -0.00000347, -0.00000004, 0.00000006, 0.00000125, 0.00000072, 0.00000299, -0.00000901, -0.00000168, 0.00000637, 0.00001217, 0.00000262, -0.00002718, 0.00000010, 0.00001785, 0.00000007, -0.00000399, -0.00000571, 0.00000050, 0.00000682, -0.00000450, -0.00000212, 0.00000460, -0.00000743, 0.00000993, -0.00000255, -0.00000840, 0.00000461, -0.00001381, 0.00001289, -0.00000062, 0.00000099, 0.00000940, -0.00001853, 0.00000895, -0.00000579, 0.00000834, 0.00000112, -0.00002688, 0.00001631, 0.00000349, 0.00000423, -0.00000104, -0.00001070, 0.00000965, -0.00000302, -0.00000673, -0.00000440, -0.00001197, -0.00001555, 0.00001135, -0.00000148, 0.00001607, 0.00004157, -0.00001858, 0.00002139, 0.00003567, 0.00000013, 0.00000393, -0.00003013, -0.00001111, -0.00000200, -0.00000843, 0.00000252, -0.00000821, -0.00000340, -0.00000135, -0.00000036, -0.00000288, -0.00000488, -0.00000445, -0.00000068, -0.00000709, -0.00000783, -0.00000237, -0.00001649, -0.00000877, -0.00005834, -0.00005295, 0.00007666, 0.00007737, 0.00004257, },
		{ 0.00003825, 0.00007341, 0.00005564, -0.00003096, -0.00004439, -0.00001207, -0.00000290, -0.00000825, -0.00001036, -0.00000339, -0.00000471, 0.00000613, 0.00000141, -0.00001054, 0.00000393, 0.00000593, 0.00000742, -0.00001148, -0.00000964, 0.00001016, -0.00000385, 0.00000178, -0.00000832, -0.00000367, 0.00001672, -0.00000626, -0.00000268, -0.00000121, -0.00001101, 0.00000332, -0.00000532, -0.00000096, 0.00000663, -0.00001040, 0.00000524, 0.00000521, -0.00001377, -0.00000846, 0.00000836, 0.00002341, -0.00000454, -0.00001563, -0.00000194, -0.00000538, 0.00000568, 0.00001037, 0.00000417, -0.00000638, -0.00000437, 0.00000007, -0.00000058, 0.00000223, -0.00000077, 0.00000451, -0.00000769, -0.00000505, 0.00000672, 0.00000996, 0.00000913, -0.00002464, -0.00000944, 0.00001878, 0.00000311, -0.00000229, -0.00000542, -0.00000347, 0.00000751, -0.00000096, -0.00000598, 0.00000594, -0.00000609, 0.00000550, 0.00000351, -0.00001205, 0.00000573, -0.00001242, 0.00000624, 0.00000653, -0.00000441, 0.00001292, -0.00001634, 0.00000268, 0.00000023, 0.00000194, 0.00000767, -0.00002682, 0.00000831, 0.00001013, 0.00000069, 0.00000187, -0.00001022, 0.00000538, 0.00000177, -0.00000752, -0.00000627, -0.00000625, -0.00002070, 0.00000580, 0.00000464, 0.00000491, 0.00004786, -0.00000873, 0.00000828, 0.00004344, 0.00000325, 0.00000518, -0.00002694, -0.00001782, 0.00000053, -0.00001066, 0.00000262, -0.00000648, -0.00000538, -0.00000099, -0.00000103, -0.00000166, -0.00000467, -0.00000498, -0.00000117, -0.00000498, -0.00000911, -0.00000175, -0.00001461, -0.00000863, -0.00004599, -0.00006313, 0.00005475, 0.00008424, },
		{ 0.00008679, 0.00003928, 0.00005996, 0.00006737, -0.00001132, -0.00004603, -0.00001933, -0.00000290, -0.00000594, -0.00001254, -0.00000281, -0.00000512, 0.00000306, 0.00000591, -0.00001215, 0.00000246, 0.00000621, 0.00000727, -0.00000634, -0.00001509, 0.00001032, -0.00000121, -0.00000095, -0.00000396, -0.00000884, 0.00001633, -0.00000086, -0.00000644, 0.00000174, -0.00001154, 0.00000078, -0.00000186, -0.00000450, 0.00000872, -0.00000892, -0.00000015, 0.00001000, -0.00001161, -0.00001135, 0.00000380, 0.00002276, 0.00000352, -0.00001638, -0.00000446, -0.00000560, 0.00000343, 0.00001078, 0.00000636, -0.00000560, -0.00000573, 0.00000057, -0.00000160, 0.00000323, -0.00000166, 0.00000456, -0.00000470, -0.00000885, 0.00000682, 0.00000801, 0.00001301, -0.00001850, -0.00001850, 0.00001705, 0.00000665, -0.00000097, -0.00000366, -0.00000696, 0.00000593, 0.00000352, -0.00000837, 0.00000472, -0.00000268, 0.00000047, 0.00000814, -0.00001281, 0.00000371, -0.00000820, -0.00000138, 0.00001185, -0.00000638, 0.00001231, -0.00001070, -0.00000460, 0.00000490, -0.00000163, 0.00001005, -0.00002264, -0.00000095, 0.00001480, 0.00000026, 0.00000167, -0.00000758, 0.00000116, 0.00000410, -0.00000510, -0.00001011, -0.00000157, -0.00002090, -0.00000389, 0.00001108, -0.00000327, 0.00004728, 0.00000597, -0.00000379, 0.00004688, 0.00000980, 0.00000484, -0.00002207, -0.00002460, 0.00000182, -0.00001132, 0.00000096, -0.00000373, -0.00000736, -0.00000090, -0.00000154, -0.00000101, -0.00000392, -0.00000543, -0.00000198, -0.00000316, -0.00000951, -0.00000212, -0.00001227, -0.00000920, -0.00003518, -0.00006727, 0.00003185, },
		{ 0.00001005, 0.00008431, 0.00004418, 0.00004701, 0.00007163, 0.00000892, -0.00004236, -0.00002639, -0.00000545, -0.00000263, -0.00001407, -0.00000382, -0.00000353, -0.00000134, 0.00000977, -0.00001180, -0.00000078, 0.00000773, 0.00000578, -0.00000095, -0.00001838, 0.00000742, 0.00000335, -0.00000424, -0.00000035, -0.00001139, 0.00001278, 0.00000559, -0.00000915, 0.00000275, -0.00000980, -0.00000311, 0.00000148, -0.00000675, 0.00000857, -0.00000508, -0.00000602, 0.00001251, -0.00000724, -0.00001394, -0.00000030, 0.00002011, 0.00001060, -0.00001446, -0.00000739, -0.00000614, 0.00000138, 0.00001045, 0.00000845, -0.00000393, -0.00000756, 0.00000123, -0.00000252, 0.00000363, -0.00000138, 0.00000304, -0.00000068, -0.00001204, 0.00000577, 0.00000718, 0.00001403, -0.00001012, -0.00002533, 0.00001230, 0.00001037, 0.00000005, -0.00000109, -0.00000904, 0.00000230, 0.00000767, -0.00000835, 0.00000122, 0.00000152, -0.00000355, 0.00001013, -0.00001048, -0.00000064, -0.00000291, -0.00000774, 0.00001370, -0.00000460, 0.00000864, -0.00000375, -0.00001048, 0.00000656, -0.00000189, 0.00000917, -0.00001647, -0.00000899, 0.00001599, 0.00000289, -0.00000026, -0.00000482, -0.00000121, 0.00000343, -0.00000066, -0.00001349, -0.00000023, -0.00001605, -0.00001517, 0.00001476, -0.00000659, 0.00004032, 0.00002262, -0.00001190, 0.00004495, 0.00001920, 0.00000391, -0.00001661, -0.00003030, 0.00000104, -0.00001017, -0.00000212, -0.00000074, -0.00000866, -0.00000144, -0.00000166, -0.00000095, -0.00000289, -0.00000554, -0.00000305, -0.00000181, -0.00000909, -0.00000309, -0.00000994, -0.00000987, -0.00002640, -0.00006655, },
		{ -0.00006230, -0.00000903, 0.00007692, 0.00005107, 0.00003662, 0.00006946, 0.00002731, -0.00003404, -0.00003139, -0.00001049, 0.00000035, -0.00001374, -0.00000688, -0.00000047, -0.00000571, 0.00001161, -0.00000884, -0.00000554, 0.00000976, 0.00000413, 0.00000326, -0.00001863, 0.00000184, 0.00000852, -0.00000656, 0.00000128, -0.00001101, 0.00000704, 0.00001144, -0.00000962, 0.00000149, -0.00000639, -0.00000718, 0.00000355, -0.00000699, 0.00000620, 0.00000008, -0.00001077, 0.00001189, -0.00000122, -0.00001559, -0.00000413, 0.00001629, 0.00001575, -0.00001030, -0.00000979, -0.00000730, -0.00000048, 0.00000963, 0.00000998, -0.00000126, -0.00000956, 0.00000152, -0.00000273, 0.00000293, 0.00000018, 0.00000043, 0.00000328, -0.00001356, 0.00000300, 0.00000767, 0.00001279, -0.00000130, -0.00002854, 0.00000488, 0.00001355, 0.00000110, 0.00000154, -0.00000916, -0.00000259, 0.00001020, -0.00000566, -0.00000353, 0.00000493, -0.00000534, 0.00000916, -0.00000580, -0.00000573, 0.00000149, -0.00001121, 0.00001168, -0.00000008, 0.00000386, 0.00000225, -0.00001333, 0.00000485, 0.00000025, 0.00000680, -0.00001049, -0.00001417, 0.00001358, 0.00000728, -0.00000201, -0.00000365, -0.00000107, 0.00000065, 0.00000374, -0.00001425, -0.00000303, -0.00000792, -0.00002467, 0.00001343, -0.00000492, 0.00002910, 0.00003774, -0.00001392, 0.00003788, 0.00002995, 0.00000381, -0.00001166, -0.00003402, -0.00000208, -0.00000755, -0.00000580, 0.00000160, -0.00000879, -0.00000269, -0.00000147, -0.00000126, -0.00000193, -0.00000517, -0.00000421, -0.00000107, -0.00000801, -0.00000429, -0.00000795, -0.00001028, -0.00001972, },
		{ -0.00001493, -0.00005588, -0.00002434, 0.00006544, 0.00005807, 0.00002996, 0.00006250, 0.00004218, -0.00002253, -0.00003307, -0.00001691, 0.00000142, -0.00001084, -0.00001157, 0.00000273, -0.00000849, 0.00001050, -0.00000347, -0.00001085, 0.00001104, 0.00000355, 0.00000535, -0.00001588, -0.00000518, 0.00001264, -0.00000671, 0.00000051, -0.00000833, 0.00000071, 0.00001515, -0.00000732, -0.00000155, -0.00000245, -0.00001014, 0.00000355, -0.00000519, 0.00000238, 0.00000508, -0.00001300, 0.00000813, 0.00000531, -0.00001552, -0.00000787, 0.00001201, 0.00001861, -0.00000481, -0.00001086, -0.00000902, -0.00000245, 0.00000869, 0.00001066, 0.00000219, -0.00001108, 0.00000084, -0.00000187, 0.00000111, 0.00000258, -0.00000232, 0.00000605, -0.00001274, -0.00000141, 0.00000895, 0.00001045, 0.00000620, -0.00002756, -0.00000413, 0.00001531, 0.00000253, 0.00000367, -0.00000735, -0.00000758, 0.00001023, -0.00000087, -0.00000799, 0.00000623, -0.00000446, 0.00000585, -0.00000026, -0.00000980, 0.00000353, -0.00001122, 0.00000662, 0.00000530, 0.00000004, 0.00000572, -0.00001271, 0.00000069, 0.00000311, 0.00000478, -0.00000620, -0.00001612, 0.00000868, 0.00001150, -0.00000194, -0.00000469, 0.00000088, -0.00000242, 0.00000604, -0.00001141, -0.00000899, 0.00000053, -0.00002943, 0.00000651, 0.00000011, 0.00001668, 0.00004831, -0.00000914, 0.00002725, 0.00003990, 0.00000593, -0.00000792, -0.00003542, -0.00000722, -0.00000430, -0.00000909, 0.00000259, -0.00000765, -0.00000438, -0.00000125, -0.00000163, -0.00000131, -0.00000434, -0.00000521, -0.00000104, -0.00000650, -0.00000536, -0.00000652, -0.00001027, },
		{ -0.00000988, -0.00001164, -0.00004851, -0.00003541, 0.00005109, 0.00006352, 0.00002742, 0.00005255, 0.00005266, -0.00000949, -0.00003120, -0.00002290, -0.00000068, -0.00000566, -0.00001653, 0.00000442, -0.00000848, 0.00000639, 0.00000320, -0.00001515, 0.00001036, 0.00000477, 0.00000523, -0.00001108, -0.00001194, 0.00001426, -0.00000418, -0.00000216, -0.00000469, -0.00000453, 0.00001579, -0.00000262, -0.00000516, 0.00000066, -0.00001105, 0.00000137, -0.00000202, -0.00000163, 0.00000848, -0.00001208, 0.00000208, 0.00001083, -0.00001319, -0.00001149, 0.00000764, 0.00001932, 0.00000094, -0.00001017, -0.00001082, -0.00000489, 0.00000784, 0.00001050, 0.00000583, -0.00001141, -0.00000119, 0.00000004, -0.00000139, 0.00000498, -0.00000416, 0.00000680, -0.00000955, -0.00000669, 0.00001005, 0.00000832, 0.00001118, -0.00002276, -0.00001316, 0.00001483, 0.00000446, 0.00000505, -0.00000411, -0.00001151, 0.00000752, 0.00000481, -0.00001064, 0.00000478, -0.00000139, 0.00000155, 0.00000445, -0.00001149, 0.00000266, -0.00000837, 0.00000030, 0.00000945, -0.00000135, 0.00000622, -0.00000944, -0.00000419, 0.00000491, 0.00000433, -0.00000395, -0.00001562, 0.00000306, 0.00001385, 0.00000061, -0.00000732, 0.00000305, -0.00000393, 0.00000516, -0.00000550, -0.00001582, 0.00000620, -0.00002796, -0.00000458, 0.00000571, 0.00000603, 0.00005252, 0.00000153, 0.00001561, 0.00004693, 0.00001107, -0.00000547, -0.00003479, -0.00001356, -0.00000145, -0.00001118, 0.00000196, -0.00000558, -0.00000601, -0.00000133, -0.00000183, -0.00000113, -0.00000327, -0.00000577, -0.00000167, -0.00000491, -0.00000599, -0.00000573, },
		{ -0.00000546, -0.00000929, -0.00000937, -0.00004112, -0.00004229, 0.00003536, 0.00006614, 0.00002877, 0.00004137, 0.00005851, 0.00000375, -0.00002649, -0.00002656, -0.00000617, 0.00000043, -0.00001991, 0.00000322, -0.00000542, 0.00000031, 0.00000940, -0.00001694, 0.00000702, 0.00000770, 0.00000368, -0.00000574, -0.00001680, 0.00001268, 0.00000053, -0.00000548, -0.00000161, -0.00000747, 0.00001337, 0.00000328, -0.00000790, 0.00000189, -0.00000965, -0.00000235, 0.00000131, -0.00000451, 0.00000935, -0.00000827, -0.00000468, 0.00001391, -0.00000854, -0.00001459, 0.00000328, 0.00001836, 0.00000598, -0.00000777, -0.00001200, -0.00000803, 0.00000700, 0.00000985, 0.00000894, -0.00001001, -0.00000451, 0.00000248, -0.00000374, 0.00000642, -0.00000433, 0.00000536, -0.00000472, -0.00001162, 0.00000989, 0.00000734, 0.00001326, -0.00001534, -0.00002055, 0.00001174, 0.00000667, 0.00000579, -0.00000017, -0.00001350, 0.00000258, 0.00000979, -0.00001044, 0.00000085, 0.00000260, -0.00000211, 0.00000698, -0.00001030, -0.00000059, -0.00000416, -0.00000529, 0.00001087, 0.00000008, 0.00000448, -0.00000513, -0.00000801, 0.00000455, 0.00000557, -0.00000308, -0.00001408, -0.00000163, 0.00001351, 0.00000510, -0.00001005, 0.00000370, -0.00000277, 0.00000139, 0.00000160, -0.00002078, 0.00000704, -0.00002080, -0.00001691, 0.00000893, -0.00000088, 0.00005023, 0.00001580, 0.00000584, 0.00004952, 0.00001916, -0.00000363, -0.00003285, -0.00002009, 0.00000010, -0.00001168, -0.00000005, -0.00000321, -0.00000709, -0.00000191, -0.00000182, -0.00000127, -0.00000225, -0.00000572, -0.00000274, -0.00000359, -0.00000602, },
		{ -0.00000547, -0.00000546, -0.00000875, -0.00000768, -0.00003434, -0.00004552, 0.00001978, 0.00006512, 0.00003325, 0.00003061, 0.00005978, 0.00001637, -0.00002026, -0.00002666, -0.00001404, 0.00000534, -0.00002006, -0.00000131, -0.00000019, -0.00000591, 0.00001330, -0.00001529, 0.00000128, 0.00001138, 0.00000202, -0.00000135, -0.00001878, 0.00000815, 0.00000615, -0.00000795, -0.00000029, -0.00000780, 0.00000881, 0.00000875, -0.00000852, 0.00000087, -0.00000650, -0.00000638, 0.00000350, -0.00000533, 0.00000762, -0.00000273, -0.00001035, 0.00001366, -0.00000219, -0.00001649, -0.00000114, 0.00001629, 0.00000972, -0.00000423, -0.00001188, -0.00001174, 0.00000578, 0.00000925, 0.00001087, -0.00000677, -0.00000854, 0.00000453, -0.00000502, 0.00000623, -0.00000264, 0.00000225, 0.00000051, -0.00001493, 0.00000778, 0.00000781, 0.00001290, -0.00000697, -0.00002496, 0.00000627, 0.00000859, 0.00000620, 0.00000371, -0.00001320, -0.00000349, 0.00001264, -0.00000718, -0.00000443, 0.00000590, -0.00000383, 0.00000679, -0.00000672, -0.00000489, -0.00000038, -0.00000856, 0.00000908, 0.00000354, 0.00000202, -0.00000150, -0.00000958, 0.00000195, 0.00000770, -0.00000236, -0.00001285, -0.00000447, 0.00001076, 0.00001012, -0.00001117, 0.00000177, 0.00000093, -0.00000374, 0.00000746, -0.00002180, 0.00000272, -0.00001026, -0.00002702, 0.00000773, -0.00000366, 0.00004282, 0.00003065, 0.00000030, 0.00004725, 0.00002917, -0.00000128, -0.00003044, -0.00002595, -0.00000012, -0.00001072, -0.00000281, -0.00000125, -0.00000729, -0.00000292, -0.00000175, -0.00000150, -0.00000154, -0.00000508, -0.00000392, -0.00000285, },
		{ -0.00000281, -0.00000455, -0.00000541, -0.00000850, -0.00000629, -0.00002840, -0.00004595, 0.00000572, 0.00006026, 0.00003964, 0.00002178, 0.00005671, 0.00002798, -0.00001375, -0.00002311, -0.00002218, 0.00000694, -0.00001625, -0.00000838, 0.00000536, -0.00001026, 0.00001363, -0.00001031, -0.00000569, 0.00001434, 0.00000156, 0.00000112, -0.00001783, 0.00000183, 0.00001103, -0.00000836, -0.00000113, -0.00000610, 0.00000364, 0.00001229, -0.00000651, -0.00000189, -0.00000278, -0.00000936, 0.00000363, -0.00000392, 0.00000405, 0.00000290, -0.00001342, 0.00001005, 0.00000465, -0.00001649, -0.00000563, 0.00001352, 0.00001199, -0.00000035, -0.00001005, -0.00001547, 0.00000361, 0.00000914, 0.00001136, -0.00000214, -0.00001228, 0.00000525, -0.00000460, 0.00000427, 0.00000045, -0.00000144, 0.00000477, -0.00001570, 0.00000363, 0.00000932, 0.00001116, 0.00000068, -0.00002572, -0.00000072, 0.00000954, 0.00000654, 0.00000700, -0.00001083, -0.00000936, 0.00001249, -0.00000157, -0.00000940, 0.00000711, -0.00000303, 0.00000431, -0.00000203, -0.00000858, 0.00000151, -0.00000888, 0.00000476, 0.00000747, 0.00000047, 0.00000031, -0.00000871, -0.00000199, 0.00000935, -0.00000064, -0.00001263, -0.00000555, 0.00000677, 0.00001396, -0.00000952, -0.00000271, 0.00000583, -0.00000806, 0.00001007, -0.00001821, -0.00000526, 0.00000030, -0.00003208, 0.00000172, -0.00000345, 0.00003257, 0.00004321, 0.00000031, 0.00004093, 0.00003938, 0.00000277, -0.00002816, -0.00003071, -0.00000211, -0.00000884, -0.00000557, -0.00000022, -0.00000662, -0.00000408, -0.00000185, -0.00000164, -0.00000124, -0.00000407, -0.00000481, },
		{ -0.00000515, -0.00000340, -0.00000363, -0.00000498, -0.00000861, -0.00000516, -0.00002322, -0.00004460, -0.00000581, 0.00005196, 0.00004641, 0.00001615, 0.00004970, 0.00003830, -0.00000765, -0.00001704, -0.00002806, 0.00000392, -0.00000910, -0.00001607, 0.00000906, -0.00001121, 0.00001025, -0.00000322, -0.00001209, 0.00001513, 0.00000306, 0.00000160, -0.00001483, -0.00000467, 0.00001375, -0.00000625, -0.00000358, -0.00000365, -0.00000057, 0.00001303, -0.00000224, -0.00000519, 0.00000009, -0.00001028, 0.00000148, -0.00000094, 0.00000003, 0.00000699, -0.00001311, 0.00000398, 0.00001046, -0.00001411, -0.00000996, 0.00001025, 0.00001297, 0.00000308, -0.00000659, -0.00001835, 0.00000000, 0.00000966, 0.00001062, 0.00000298, -0.00001461, 0.00000395, -0.00000242, 0.00000105, 0.00000394, -0.00000444, 0.00000698, -0.00001373, -0.00000188, 0.00001093, 0.00000926, 0.00000639, -0.00002305, -0.00000796, 0.00000892, 0.00000691, 0.00000950, -0.00000698, -0.00001380, 0.00000924, 0.00000494, -0.00001239, 0.00000550, -0.00000007, 0.00000072, 0.00000218, -0.00001030, 0.00000090, -0.00000670, -0.00000058, 0.00001014, 0.00000091, 0.00000009, -0.00000621, -0.00000582, 0.00000930, 0.00000256, -0.00001322, -0.00000576, 0.00000303, 0.00001536, -0.00000500, -0.00000853, 0.00000996, -0.00000966, 0.00000861, -0.00001098, -0.00001420, 0.00000768, -0.00003083, -0.00000770, -0.00000231, 0.00002182, 0.00005143, 0.00000582, 0.00003235, 0.00004789, 0.00000933, -0.00002607, -0.00003436, -0.00000544, -0.00000677, -0.00000769, -0.00000029, -0.00000540, -0.00000500, -0.00000230, -0.00000163, -0.00000126, -0.00000300, },
		{ -0.00000216, -0.00000485, -0.00000429, -0.00000312, -0.00000404, -0.00000895, -0.00000443, -0.00001851, -0.00004237, -0.00001436, 0.00004124, 0.00005184, 0.00001461, 0.00003953, 0.00004682, -0.00000182, -0.00001037, -0.00002960, -0.00000359, -0.00000057, -0.00002196, 0.00000916, -0.00000839, 0.00000422, 0.00000405, -0.00001614, 0.00001295, 0.00000639, 0.00000086, -0.00001112, -0.00000984, 0.00001365, -0.00000212, -0.00000644, -0.00000188, -0.00000277, 0.00001106, 0.00000304, -0.00000753, 0.00000105, -0.00000883, -0.00000229, 0.00000236, -0.00000298, 0.00000847, -0.00000970, -0.00000294, 0.00001380, -0.00000941, -0.00001361, 0.00000648, 0.00001300, 0.00000551, -0.00000207, -0.00001946, -0.00000515, 0.00001045, 0.00000929, 0.00000752, -0.00001468, 0.00000055, 0.00000095, -0.00000242, 0.00000666, -0.00000577, 0.00000675, -0.00000955, -0.00000759, 0.00001160, 0.00000815, 0.00000968, -0.00001784, -0.00001410, 0.00000653, 0.00000712, 0.00001125, -0.00000241, -0.00001604, 0.00000362, 0.00001060, -0.00001229, 0.00000131, 0.00000385, -0.00000245, 0.00000458, -0.00000948, -0.00000186, -0.00000330, -0.00000519, 0.00001031, 0.00000349, -0.00000140, -0.00000344, -0.00000817, 0.00000699, 0.00000681, -0.00001367, -0.00000629, 0.00000071, 0.00001399, 0.00000141, -0.00001382, 0.00001148, -0.00000765, 0.00000372, -0.00000234, -0.00002116, 0.00000987, -0.00002397, -0.00001791, -0.00000241, 0.00001227, 0.00005447, 0.00001556, 0.00002379, 0.00005316, 0.00001849, -0.00002367, -0.00003728, -0.00000957, -0.00000516, -0.00000888, -0.00000127, -0.00000412, -0.00000536, -0.00000311, -0.00000158, -0.00000140, },
		},
		{ { -0.00000301, -0.00000454, 0.00000310, 0.00000673, -0.00000549, -0.00000441, -0.00000131, -0.00000431, -0.00000026, -0.00000182, -0.00000019, -0.00000432, -0.00000800, -0.00000293, -0.00002656, -0.00002911, -0.00000247, 0.00001693, 0.00004350, 0.00002057, 0.00001103, 0.00004418, 0.00001356, 0.00000460, 0.00000433, -0.00003565, -0.00000800, 0.00000872, -0.00001108, -0.00000133, -0.00000434, -0.00000201, -0.00000090, -0.00000589, 0.00000356, 0.00000327, 0.00000204, -0.00000742, -0.00000502, 0.00000469, -0.00000476, 0.00000707, 0.00000278, -0.00000832, 0.00000480, -0.00000593, -0.00000138, 0.00000096, -0.00001072, 0.00001235, 0.00001433, 0.00000135, -0.00001238, -0.00002370, 0.00001035, 0.00001765, -0.00000078, -0.00000250, -0.00001852, 0.00000220, 0.00001977, -0.00001204, -0.00000293, 0.00000768, -0.00000770, 0.00000768, -0.00000293, -0.00001204, 0.00001977, 0.00000220, -0.00001852, -0.00000250, -0.00000078, 0.00001765, 0.00001035, -0.00002370, -0.00001238, 0.00000135, 0.00001433, 0.00001235, -0.00001072, 0.00000096, -0.00000138, -0.00000593, 0.00000480, -0.00000832, 0.00000278, 0.00000707, -0.00000476, 0.00000469, -0.00000502, -0.00000742, 0.00000204, 0.00000327, 0.00000356, -0.00000589, -0.00000090, -0.00000201, -0.00000434, -0.00000133, -0.00001108, 0.00000872, -0.00000800, -0.00003565, 0.00000433, 0.00000460, 0.00001356, 0.00004418, 0.00001103, 0.00002057, 0.00004350, 0.00001693, -0.00000247, -0.00002911, -0.00002656, -0.00000293, -0.00000800, -0.00000432, -0.00000019, -0.00000182, -0.00000026, -0.00000431, -0.00000131, -0.00000441, -0.00000549, 0.00000673, 0.00000310, -0.00000454, },
		{ -0.00000363, -0.00000351, -0.00000408, -0.00000046, 0.00000763, -0.00000166, -0.00000622, -0.00000148, -0.00000400, -0.00000093, -0.00000133, -0.00000065, -0.00000209, -0.00000874, -0.00000068, -0.00002009, -0.00003321, -0.00000535, 0.00000831, 0.00003619, 0.00003179, 0.00000308, 0.00003797, 0.00002558, -0.00000099, 0.00001346, -0.00002496, -0.00002306, 0.00001025, -0.00000595, -0.00000520, -0.00000263, -0.00000315, -0.00000021, -0.00000563, 0.00000137, 0.00000256, 0.00000421, -0.00000457, -0.00000936, 0.00000630, -0.00000398, 0.00000244, 0.00000839, -0.00000959, 0.00000330, -0.00000278, -0.00000559, 0.00000525, -0.00001281, 0.00000544, 0.00001918, 0.00000317, -0.00000595, -0.00002582, -0.00000125, 0.00002324, 0.00000109, -0.00000045, -0.00001384, -0.00001071, 0.00002206, -0.00000177, -0.00001163, 0.00000968, -0.00000507, 0.00000155, 0.00000542, -0.00001528, 0.00001006, 0.00001391, -0.00001678, -0.00000805, -0.00000058, 0.00001121, 0.00001708, -0.00001619, -0.00001860, -0.00000118, 0.00001006, 0.00001525, -0.00000558, -0.00000298, 0.00000115, -0.00000700, 0.00000408, -0.00000521, -0.00000244, 0.00000884, -0.00000283, 0.00000221, -0.00000189, -0.00000797, -0.00000049, 0.00000221, 0.00000617, -0.00000484, -0.00000308, 0.00000047, -0.00000676, 0.00000123, -0.00001104, 0.00000071, 0.00000588, -0.00003542, -0.00001149, 0.00000893, 0.00000567, 0.00004225, 0.00002377, 0.00001205, 0.00004371, 0.00002866, 0.00000174, -0.00002517, -0.00003104, -0.00000670, -0.00000712, -0.00000653, -0.00000011, -0.00000220, 0.00000015, -0.00000402, -0.00000227, -0.00000190, -0.00000764, 0.00000341, 0.00000654, },
		{ 0.00000832, -0.00000094, -0.00000453, -0.00000331, -0.00000286, 0.00000614, 0.00000242, -0.00000635, -0.00000263, -0.00000352, -0.00000164, -0.00000078, -0.00000137, -0.00000014, -0.00000878, -0.00000058, -0.00001195, -0.00003605, -0.00000951, 0.00000465, 0.00002407, 0.00004044, 0.00000342, 0.00002490, 0.00003719, -0.00000208, 0.00001308, -0.00000808, -0.00003222, 0.00000353, 0.00000143, -0.00000717, -0.00000305, -0.00000312, -0.00000000, -0.00000538, 0.00000025, 0.00000120, 0.00000430, 0.00000062, -0.00001309, 0.00000497, 0.00000015, -0.00000366, 0.00001211, -0.00000752, -0.00000078, 0.00000192, -0.00000986, 0.00000728, -0.00001004, -0.00000453, 0.00002235, 0.00000639, -0.00000181, -0.00002196, -0.00001454, 0.00002418, 0.00000701, -0.00000200, -0.00000619, -0.00001922, 0.00001532, 0.00001136, -0.00001563, 0.00000570, 0.00000115, -0.00000465, 0.00000921, -0.00001115, -0.00000211, 0.00001931, -0.00000848, -0.00001416, -0.00000136, 0.00000667, 0.00001847, -0.00000620, -0.00002178, -0.00000550, 0.00000692, 0.00001514, 0.00000003, -0.00000489, 0.00000187, -0.00000660, 0.00000218, -0.00000176, -0.00000593, 0.00000788, 0.00000016, 0.00000050, -0.00000027, -0.00000748, -0.00000196, -0.00000025, 0.00000751, -0.00000147, -0.00000642, 0.00000281, -0.00000756, 0.00000031, -0.00000628, -0.00000877, 0.00001254, -0.00002468, -0.00002742, 0.00000703, 0.00000306, 0.00003439, 0.00003628, 0.00000987, 0.00003767, 0.00003991, 0.00000931, -0.00002138, -0.00003408, -0.00001129, -0.00000644, -0.00000859, -0.00000053, -0.00000234, 0.00000002, -0.00000297, -0.00000391, 0.00000005, -0.00000740, -0.00000128, },
		{ -0.00000562, 0.00000743, 0.00000294, -0.00000498, -0.00000320, -0.00000369, 0.00000325, 0.00000531, -0.00000461, -0.00000407, -0.00000332, -0.00000232, -0.00000021, -0.00000209, 0.00000110, -0.00000760, -0.00000259, -0.00000360, -0.00003557, -0.00001655, 0.00000505, 0.00001161, 0.00004205, 0.00001204, 0.00000974, 0.00004286, 0.00000445, 0.00000557, 0.00000747, -0.00003109, -0.00000870, 0.00000634, -0.00000525, -0.00000488, -0.00000338, 0.00000081, -0.00000549, -0.00000057, 0.00000092, 0.00000165, 0.00000636, -0.00001359, -0.00000040, 0.00000635, -0.00000841, 0.00001172, -0.00000174, -0.00000636, 0.00000624, -0.00001173, 0.00000501, -0.00000277, -0.00001435, 0.00002086, 0.00001237, -0.00000051, -0.00001430, -0.00002489, 0.00001795, 0.00001618, -0.00000438, -0.00000016, -0.00002013, 0.00000222, 0.00002133, -0.00001192, -0.00000263, 0.00000705, -0.00000715, 0.00000723, -0.00000294, -0.00001108, 0.00001645, 0.00000297, -0.00001685, -0.00000492, 0.00000436, 0.00001645, 0.00000280, -0.00002049, -0.00001091, 0.00000398, 0.00001372, 0.00000437, -0.00000470, 0.00000172, -0.00000560, -0.00000007, 0.00000116, -0.00000743, 0.00000525, 0.00000277, -0.00000004, 0.00000065, -0.00000714, -0.00000225, -0.00000253, 0.00000615, 0.00000378, -0.00000911, 0.00000297, -0.00000532, -0.00000372, -0.00000014, -0.00001447, 0.00000998, -0.00000872, -0.00003650, -0.00000243, 0.00000332, 0.00002445, 0.00004443, 0.00001472, 0.00002870, 0.00004716, 0.00002048, -0.00001642, -0.00003631, -0.00001639, -0.00000603, -0.00001034, -0.00000162, -0.00000211, -0.00000073, -0.00000144, -0.00000536, 0.00000052, -0.00000510, },
		{ -0.00000196, -0.00000806, 0.00000395, 0.00000654, -0.00000384, -0.00000400, -0.00000346, 0.00000032, 0.00000618, -0.00000162, -0.00000494, -0.00000357, -0.00000304, 0.00000026, -0.00000243, 0.00000125, -0.00000515, -0.00000579, 0.00000288, -0.00003034, -0.00002590, 0.00000635, 0.00000338, 0.00003515, 0.00002506, -0.00000128, 0.00003916, 0.00001719, -0.00000301, 0.00001535, -0.00002029, -0.00002052, 0.00000520, -0.00000036, -0.00000576, -0.00000525, 0.00000211, -0.00000500, -0.00000246, 0.00000253, -0.00000239, 0.00000969, -0.00000912, -0.00000867, 0.00001147, -0.00000886, 0.00000643, 0.00000619, -0.00001092, 0.00000770, -0.00000938, -0.00000152, 0.00000605, -0.00001985, 0.00001296, 0.00002029, -0.00000017, -0.00000663, -0.00002865, 0.00000513, 0.00002496, -0.00000374, 0.00000121, -0.00001416, -0.00001157, 0.00002326, -0.00000134, -0.00001089, 0.00000878, -0.00000469, 0.00000159, 0.00000452, -0.00001338, 0.00000743, 0.00001241, -0.00001367, -0.00001050, 0.00000255, 0.00001370, 0.00000867, -0.00001550, -0.00001545, 0.00000016, 0.00001203, 0.00000711, -0.00000348, 0.00000169, -0.00000435, -0.00000241, 0.00000324, -0.00000725, 0.00000198, 0.00000440, 0.00000008, 0.00000177, -0.00000715, -0.00000246, -0.00000304, 0.00000213, 0.00000885, -0.00000893, -0.00000007, -0.00000072, -0.00000824, 0.00000344, -0.00001381, 0.00000090, 0.00000521, -0.00003504, -0.00001632, 0.00000234, 0.00001553, 0.00004646, 0.00002455, 0.00002094, 0.00004850, 0.00003354, -0.00000862, -0.00003774, -0.00002217, -0.00000595, -0.00001151, -0.00000356, -0.00000152, -0.00000187, -0.00000004, -0.00000584, -0.00000068, },
		{ -0.00000290, 0.00000055, -0.00000788, -0.00000085, 0.00000826, -0.00000086, -0.00000507, -0.00000318, -0.00000162, 0.00000512, 0.00000149, -0.00000464, -0.00000406, -0.00000388, 0.00000039, -0.00000210, 0.00000031, -0.00000201, -0.00000865, 0.00000571, -0.00002062, -0.00003492, 0.00000470, 0.00000173, 0.00002214, 0.00003642, -0.00000349, 0.00002687, 0.00003084, -0.00000632, 0.00001358, -0.00000511, -0.00002604, -0.00000209, 0.00000407, -0.00000350, -0.00000837, 0.00000234, -0.00000256, -0.00000588, 0.00000497, -0.00000506, 0.00000823, -0.00000039, -0.00001639, 0.00001192, -0.00000371, -0.00000204, 0.00001301, -0.00001163, 0.00000474, -0.00000306, -0.00000972, 0.00001208, -0.00001802, -0.00000021, 0.00002689, 0.00000224, -0.00000213, -0.00002508, -0.00001045, 0.00002866, 0.00000224, -0.00000181, -0.00000538, -0.00002030, 0.00001601, 0.00001162, -0.00001429, 0.00000493, 0.00000085, -0.00000379, 0.00000757, -0.00000929, -0.00000291, 0.00001568, -0.00000533, -0.00001513, -0.00000088, 0.00001170, 0.00001142, -0.00000910, -0.00001734, -0.00000470, 0.00000995, 0.00000893, -0.00000230, 0.00000196, -0.00000256, -0.00000477, 0.00000432, -0.00000572, -0.00000131, 0.00000512, 0.00000018, 0.00000320, -0.00000666, -0.00000386, -0.00000146, -0.00000277, 0.00001113, -0.00000481, -0.00000527, 0.00000370, -0.00001016, 0.00000236, -0.00000799, -0.00000917, 0.00001158, -0.00002431, -0.00002881, -0.00000290, 0.00000847, 0.00004296, 0.00003589, 0.00001758, 0.00004435, 0.00004560, 0.00000303, -0.00003742, -0.00002888, -0.00000660, -0.00001173, -0.00000630, -0.00000091, -0.00000293, 0.00000063, -0.00000507, },
		{ -0.00000340, -0.00000508, 0.00000140, -0.00000557, -0.00000517, 0.00000719, 0.00000313, -0.00000526, -0.00000363, -0.00000235, 0.00000303, 0.00000361, -0.00000316, -0.00000436, -0.00000479, -0.00000012, -0.00000105, -0.00000123, 0.00000079, -0.00000974, 0.00000439, -0.00000862, -0.00004003, -0.00000225, 0.00000561, 0.00000813, 0.00004067, 0.00000383, 0.00001089, 0.00003910, -0.00000126, 0.00000529, 0.00000747, -0.00002270, -0.00001179, 0.00000423, 0.00000187, -0.00001040, -0.00000038, 0.00000205, -0.00000948, 0.00000578, -0.00000390, 0.00000200, 0.00000917, -0.00001927, 0.00000582, 0.00000538, -0.00000982, 0.00001516, -0.00000699, -0.00000195, 0.00000463, -0.00001556, 0.00001184, -0.00000898, -0.00001445, 0.00002798, 0.00000874, -0.00000145, -0.00001675, -0.00002337, 0.00002416, 0.00001268, -0.00000590, 0.00000128, -0.00002100, 0.00000266, 0.00002104, -0.00001032, -0.00000255, 0.00000572, -0.00000567, 0.00000561, -0.00000221, -0.00000968, 0.00001187, 0.00000451, -0.00001560, -0.00000652, 0.00001001, 0.00001260, -0.00000353, -0.00001615, -0.00000945, 0.00000674, 0.00001045, -0.00000142, 0.00000189, 0.00000014, -0.00000680, 0.00000405, -0.00000309, -0.00000427, 0.00000516, 0.00000012, 0.00000421, -0.00000467, -0.00000670, 0.00000093, -0.00000602, 0.00000913, 0.00000214, -0.00000990, 0.00000517, -0.00000795, -0.00000249, -0.00000093, -0.00001506, 0.00000911, -0.00000965, -0.00003455, -0.00001233, 0.00000200, 0.00003557, 0.00004550, 0.00001991, 0.00003700, 0.00005387, 0.00001810, -0.00003382, -0.00003631, -0.00000870, -0.00001078, -0.00000946, -0.00000079, -0.00000344, 0.00000030, },
		{ -0.00000080, -0.00000155, -0.00000631, 0.00000045, -0.00000245, -0.00000747, 0.00000366, 0.00000652, -0.00000364, -0.00000481, -0.00000246, 0.00000100, 0.00000423, -0.00000102, -0.00000407, -0.00000552, -0.00000140, 0.00000035, -0.00000258, 0.00000226, -0.00000842, -0.00000003, 0.00000235, -0.00003855, -0.00001406, 0.00001115, -0.00000150, 0.00003563, 0.00001708, -0.00000214, 0.00003801, 0.00001054, -0.00000335, 0.00001255, -0.00001263, -0.00001847, -0.00000134, 0.00000749, -0.00000842, -0.00000640, 0.00000689, -0.00001055, 0.00000276, 0.00000136, -0.00000603, 0.00001482, -0.00001465, -0.00000536, 0.00001409, -0.00001260, 0.00001085, 0.00000182, -0.00000933, 0.00000982, -0.00001573, 0.00000472, 0.00000357, -0.00002411, 0.00002087, 0.00001859, -0.00000236, -0.00000807, -0.00002931, 0.00001180, 0.00002356, -0.00000677, 0.00000274, -0.00001471, -0.00001105, 0.00002253, -0.00000028, -0.00000942, 0.00000669, -0.00000339, 0.00000093, 0.00000355, -0.00001045, 0.00000363, 0.00001126, -0.00001063, -0.00001272, 0.00000692, 0.00001376, 0.00000035, -0.00001307, -0.00001250, 0.00000209, 0.00001151, -0.00000023, 0.00000070, 0.00000350, -0.00000760, 0.00000210, 0.00000026, -0.00000650, 0.00000445, 0.00000033, 0.00000394, -0.00000097, -0.00000986, 0.00000200, -0.00000587, 0.00000349, 0.00000899, -0.00001104, 0.00000247, -0.00000264, -0.00000795, 0.00000328, -0.00001458, 0.00000080, 0.00000250, -0.00003147, -0.00002282, -0.00000563, 0.00002565, 0.00005112, 0.00002732, 0.00002960, 0.00005668, 0.00003477, -0.00002550, -0.00004349, -0.00001308, -0.00000893, -0.00001232, -0.00000167, -0.00000327, },
		{ -0.00000266, -0.00000216, -0.00000023, -0.00000626, -0.00000159, -0.00000000, -0.00000725, -0.00000088, 0.00000774, -0.00000024, -0.00000586, -0.00000285, -0.00000023, 0.00000353, 0.00000094, -0.00000305, -0.00000573, -0.00000328, 0.00000142, -0.00000297, 0.00000200, -0.00000518, -0.00000537, 0.00000926, -0.00003009, -0.00002749, 0.00001356, -0.00000368, 0.00002322, 0.00003011, -0.00000705, 0.00002799, 0.00002379, -0.00000668, 0.00000967, -0.00000133, -0.00001827, -0.00001029, 0.00000902, -0.00000126, -0.00001332, 0.00000857, -0.00000689, -0.00000404, 0.00000803, -0.00001122, 0.00001313, -0.00000346, -0.00001692, 0.00001735, -0.00000807, 0.00000147, 0.00001106, -0.00001339, 0.00000956, -0.00000961, -0.00000632, 0.00001392, -0.00002489, 0.00000639, 0.00002806, -0.00000119, -0.00000280, -0.00002714, -0.00000449, 0.00002975, -0.00000178, -0.00000063, -0.00000563, -0.00001965, 0.00001541, 0.00001136, -0.00001153, 0.00000301, 0.00000100, -0.00000298, 0.00000526, -0.00000627, -0.00000445, 0.00001192, -0.00000195, -0.00001635, 0.00000117, 0.00001501, 0.00000331, -0.00000997, -0.00001298, -0.00000321, 0.00001117, 0.00000201, -0.00000165, 0.00000637, -0.00000621, -0.00000135, 0.00000347, -0.00000735, 0.00000269, 0.00000129, 0.00000217, 0.00000340, -0.00001139, 0.00000017, -0.00000253, -0.00000315, 0.00001253, -0.00000744, -0.00000315, 0.00000282, -0.00001067, 0.00000269, -0.00000921, -0.00000815, 0.00000783, -0.00002163, -0.00003006, -0.00001485, 0.00001370, 0.00005153, 0.00003792, 0.00002505, 0.00005397, 0.00005042, -0.00001185, -0.00004870, -0.00002021, -0.00000698, -0.00001412, -0.00000369, },
		{ -0.00000655, -0.00000213, -0.00000330, 0.00000026, -0.00000527, -0.00000369, 0.00000091, -0.00000521, -0.00000459, 0.00000616, 0.00000381, -0.00000561, -0.00000403, -0.00000072, 0.00000229, 0.00000207, -0.00000154, -0.00000516, -0.00000527, 0.00000151, -0.00000214, 0.00000039, -0.00000132, -0.00000929, 0.00001061, -0.00001682, -0.00003798, 0.00000954, 0.00000103, 0.00000845, 0.00003716, -0.00000249, 0.00001347, 0.00003254, -0.00000251, 0.00000262, 0.00000567, -0.00001138, -0.00001749, 0.00000361, 0.00000875, -0.00001676, 0.00000436, 0.00000114, -0.00001165, 0.00001173, -0.00000994, 0.00000428, 0.00000964, -0.00002301, 0.00001219, 0.00000237, -0.00000859, 0.00001616, -0.00001134, 0.00000358, 0.00000013, -0.00001612, 0.00001707, -0.00001618, -0.00001088, 0.00003213, 0.00000473, -0.00000192, -0.00001933, -0.00001903, 0.00002769, 0.00000832, -0.00000545, 0.00000137, -0.00002052, 0.00000293, 0.00001927, -0.00000722, -0.00000308, 0.00000421, -0.00000373, 0.00000308, -0.00000058, -0.00000846, 0.00000678, 0.00000645, -0.00001488, -0.00000665, 0.00001492, 0.00000678, -0.00000798, -0.00001148, -0.00000754, 0.00000861, 0.00000527, -0.00000416, 0.00000726, -0.00000234, -0.00000525, 0.00000535, -0.00000628, -0.00000017, 0.00000294, -0.00000031, 0.00000664, -0.00000978, -0.00000443, 0.00000187, -0.00000759, 0.00001102, -0.00000034, -0.00000865, 0.00000538, -0.00000910, -0.00000167, -0.00000273, -0.00001334, 0.00000587, -0.00000985, -0.00003108, -0.00002404, -0.00000002, 0.00004603, 0.00004910, 0.00002539, 0.00004720, 0.00006237, 0.00000658, -0.00004995, -0.00002981, -0.00000611, -0.00001442, },
		{ -0.00001330, -0.00000963, -0.00000208, -0.00000405, 0.00000004, -0.00000392, -0.00000513, 0.00000038, -0.00000276, -0.00000616, 0.00000244, 0.00000676, -0.00000332, -0.00000570, -0.00000112, 0.00000131, 0.00000219, -0.00000007, -0.00000380, -0.00000674, 0.00000028, -0.00000043, -0.00000160, 0.00000166, -0.00001034, 0.00000697, -0.00000263, -0.00004169, -0.00000118, 0.00000891, -0.00000311, 0.00003541, 0.00000873, 0.00000047, 0.00003333, 0.00000706, -0.00000315, 0.00000598, -0.00000200, -0.00001806, -0.00000768, 0.00001630, -0.00001282, -0.00000544, 0.00000996, -0.00001547, 0.00000903, -0.00000202, -0.00000750, 0.00001824, -0.00001975, -0.00000021, 0.00001382, -0.00001402, 0.00001416, -0.00000344, -0.00000522, 0.00000878, -0.00001992, 0.00001137, -0.00000169, -0.00002440, 0.00002723, 0.00001503, -0.00000331, -0.00001041, -0.00002712, 0.00001733, 0.00001970, -0.00000747, 0.00000338, -0.00001484, -0.00000958, 0.00001990, 0.00000175, -0.00000779, 0.00000390, -0.00000142, -0.00000052, 0.00000303, -0.00000723, -0.00000086, 0.00001055, -0.00000821, -0.00001386, 0.00001160, 0.00001134, -0.00000669, -0.00000969, -0.00000947, 0.00000402, 0.00000845, -0.00000528, 0.00000530, 0.00000305, -0.00000788, 0.00000487, -0.00000333, -0.00000355, 0.00000449, -0.00000211, 0.00000727, -0.00000502, -0.00000990, 0.00000447, -0.00000782, 0.00000525, 0.00000711, -0.00001090, 0.00000369, -0.00000426, -0.00000675, 0.00000124, -0.00001319, -0.00000044, -0.00000097, -0.00002587, -0.00003035, -0.00001453, 0.00003431, 0.00005789, 0.00003144, 0.00003891, 0.00006856, 0.00002810, -0.00004540, -0.00004084, -0.00000750, },
		{ -0.00001194, -0.00001134, -0.00001239, -0.00000262, -0.00000453, -0.00000061, -0.00000267, -0.00000579, -0.00000084, -0.00000115, -0.00000552, -0.00000167, 0.00000715, 0.00000065, -0.00000678, -0.00000219, 0.00000094, 0.00000166, 0.00000082, -0.00000195, -0.00000709, -0.00000207, 0.00000132, -0.00000289, 0.00000279, -0.00000839, 0.00000056, 0.00000846, -0.00003712, -0.00001575, 0.00001500, -0.00000785, 0.00002588, 0.00002133, -0.00000627, 0.00002656, 0.00001736, -0.00000385, 0.00000149, 0.00000420, -0.00001065, -0.00001958, 0.00001580, -0.00000110, -0.00001655, 0.00001419, -0.00001205, -0.00000004, 0.00000864, -0.00001586, 0.00001767, -0.00000775, -0.00001454, 0.00002011, -0.00001140, 0.00000563, 0.00000678, -0.00001217, 0.00001205, -0.00001586, -0.00000057, 0.00001210, -0.00002875, 0.00001353, 0.00002607, -0.00000315, -0.00000437, -0.00002720, 0.00000218, 0.00002737, -0.00000404, 0.00000070, -0.00000652, -0.00001737, 0.00001324, 0.00001118, -0.00000797, 0.00000022, 0.00000171, -0.00000263, 0.00000294, -0.00000266, -0.00000669, 0.00000856, 0.00000092, -0.00001712, 0.00000439, 0.00001589, -0.00000451, -0.00000925, -0.00000881, -0.00000119, 0.00000994, -0.00000397, 0.00000095, 0.00000796, -0.00000765, 0.00000197, 0.00000054, -0.00000631, 0.00000485, -0.00000214, 0.00000507, 0.00000125, -0.00001353, 0.00000330, -0.00000414, -0.00000191, 0.00001151, -0.00000850, -0.00000113, 0.00000103, -0.00000950, 0.00000117, -0.00000922, -0.00000690, 0.00000244, -0.00001723, -0.00003147, -0.00002750, 0.00001691, 0.00006107, 0.00004251, 0.00003232, 0.00006802, 0.00005005, -0.00003382, -0.00005157, },
		{ -0.00005983, -0.00001980, -0.00000927, -0.00001458, -0.00000358, -0.00000486, -0.00000158, -0.00000155, -0.00000603, -0.00000194, -0.00000080, -0.00000374, -0.00000435, 0.00000470, 0.00000481, -0.00000607, -0.00000415, 0.00000085, 0.00000117, 0.00000090, -0.00000018, -0.00000609, -0.00000482, 0.00000214, -0.00000283, 0.00000200, -0.00000456, -0.00000577, 0.00001382, -0.00002563, -0.00002956, 0.00001541, -0.00000539, 0.00001251, 0.00003018, -0.00000528, 0.00001580, 0.00002397, 0.00000071, -0.00000317, 0.00000367, 0.00000118, -0.00002512, 0.00000509, 0.00001392, -0.00002244, 0.00000988, -0.00000176, -0.00001131, 0.00001590, -0.00001583, 0.00000774, 0.00000771, -0.00002381, 0.00001715, -0.00000140, -0.00000525, 0.00001428, -0.00001339, 0.00000845, -0.00000601, -0.00001316, 0.00001907, -0.00002232, -0.00000466, 0.00003257, 0.00000156, -0.00000262, -0.00002121, -0.00001244, 0.00002776, 0.00000439, -0.00000366, 0.00000015, -0.00001842, 0.00000247, 0.00001667, -0.00000303, -0.00000433, 0.00000295, -0.00000190, 0.00000021, 0.00000175, -0.00000784, 0.00000197, 0.00000819, -0.00001441, -0.00000509, 0.00001801, 0.00000007, -0.00001043, -0.00000696, -0.00000500, 0.00000870, -0.00000048, -0.00000400, 0.00001023, -0.00000408, -0.00000218, 0.00000378, -0.00000726, 0.00000335, -0.00000026, 0.00000120, 0.00000654, -0.00001330, -0.00000154, 0.00000107, -0.00000708, 0.00001104, -0.00000252, -0.00000627, 0.00000398, -0.00000859, -0.00000194, -0.00000450, -0.00001042, 0.00000090, -0.00000906, -0.00002718, -0.00003594, -0.00000407, 0.00005578, 0.00005612, 0.00003064, 0.00006134, 0.00006916, -0.00001488, },
		{ 0.00001057, -0.00006307, -0.00003098, -0.00000785, -0.00001617, -0.00000493, -0.00000486, -0.00000306, -0.00000042, -0.00000616, -0.00000271, -0.00000117, -0.00000232, -0.00000468, 0.00000059, 0.00000726, -0.00000307, -0.00000635, 0.00000029, 0.00000120, 0.00000037, 0.00000090, -0.00000394, -0.00000701, 0.00000144, -0.00000149, 0.00000005, -0.00000054, -0.00000970, 0.00001299, -0.00001076, -0.00003819, 0.00000884, 0.00000167, 0.00000004, 0.00003236, 0.00000172, 0.00000546, 0.00002501, 0.00000760, -0.00000380, -0.00000263, 0.00001064, -0.00001976, -0.00001233, 0.00002440, -0.00001792, -0.00000249, 0.00001058, -0.00001848, 0.00001492, -0.00000679, -0.00000669, 0.00001909, -0.00002299, 0.00000541, 0.00001120, -0.00001274, 0.00001516, -0.00000795, 0.00000011, 0.00000474, -0.00002061, 0.00001638, -0.00000825, -0.00002065, 0.00003049, 0.00001095, -0.00000350, -0.00001322, -0.00002197, 0.00002043, 0.00001457, -0.00000600, 0.00000270, -0.00001399, -0.00000780, 0.00001590, 0.00000474, -0.00000649, 0.00000103, 0.00000073, -0.00000251, 0.00000319, -0.00000443, -0.00000531, 0.00001007, -0.00000659, -0.00001334, 0.00001547, 0.00000701, -0.00001179, -0.00000605, -0.00000607, 0.00000507, 0.00000364, -0.00000736, 0.00000871, 0.00000171, -0.00000558, 0.00000496, -0.00000594, 0.00000028, 0.00000252, -0.00000238, 0.00000881, -0.00000900, -0.00000796, 0.00000473, -0.00000815, 0.00000634, 0.00000420, -0.00000892, 0.00000336, -0.00000492, -0.00000561, -0.00000168, -0.00001031, -0.00000295, -0.00000422, -0.00001967, -0.00003760, -0.00002473, 0.00004055, 0.00006798, 0.00003620, 0.00005092, 0.00008205, },
		{ 0.00008605, 0.00004003, -0.00005857, -0.00004487, -0.00000788, -0.00001702, -0.00000698, -0.00000413, -0.00000517, 0.00000060, -0.00000601, -0.00000354, -0.00000133, -0.00000217, -0.00000315, -0.00000314, 0.00000669, 0.00000148, -0.00000749, -0.00000136, 0.00000171, -0.00000014, 0.00000091, -0.00000134, -0.00000778, -0.00000078, 0.00000041, -0.00000190, 0.00000218, -0.00001020, 0.00000750, 0.00000319, -0.00003902, -0.00000312, 0.00000931, -0.00000796, 0.00002780, 0.00001122, -0.00000132, 0.00002155, 0.00001302, 0.00000061, -0.00000978, 0.00001166, -0.00000469, -0.00002813, 0.00002319, -0.00000287, -0.00001726, 0.00001787, -0.00001667, 0.00000521, 0.00000676, -0.00001807, 0.00002054, -0.00001196, -0.00000995, 0.00001984, -0.00001259, 0.00000894, 0.00000146, -0.00000838, 0.00001131, -0.00001987, 0.00000575, 0.00000710, -0.00002843, 0.00001927, 0.00002189, -0.00000359, -0.00000699, -0.00002455, 0.00000812, 0.00002227, -0.00000405, 0.00000140, -0.00000736, -0.00001397, 0.00000963, 0.00001149, -0.00000431, -0.00000291, 0.00000278, -0.00000291, 0.00000118, 0.00000081, -0.00000923, 0.00000584, 0.00000277, -0.00001674, 0.00000797, 0.00001426, -0.00001099, -0.00000756, -0.00000469, 0.00000078, 0.00000639, -0.00000773, 0.00000397, 0.00000741, -0.00000636, 0.00000366, -0.00000297, -0.00000322, 0.00000472, -0.00000398, 0.00000748, -0.00000230, -0.00001300, 0.00000467, -0.00000525, -0.00000004, 0.00000863, -0.00000778, 0.00000004, -0.00000075, -0.00000754, -0.00000159, -0.00000799, -0.00000625, -0.00000323, -0.00001232, -0.00003251, -0.00004024, 0.00001660, 0.00007268, 0.00004915, 0.00004097, },
		{ 0.00003656, 0.00008050, 0.00006921, -0.00004389, -0.00005996, -0.00001055, -0.00001662, -0.00001034, -0.00000246, -0.00000749, 0.00000090, -0.00000502, -0.00000502, -0.00000080, -0.00000307, -0.00000136, -0.00000468, 0.00000327, 0.00000574, -0.00000639, -0.00000398, 0.00000204, -0.00000003, 0.00000000, 0.00000080, -0.00000672, -0.00000382, 0.00000184, -0.00000285, 0.00000283, -0.00000765, 0.00000012, 0.00001270, -0.00003203, -0.00001692, 0.00001379, -0.00001013, 0.00001855, 0.00001983, -0.00000364, 0.00001616, 0.00001494, 0.00000736, -0.00001196, 0.00000283, 0.00001297, -0.00003324, 0.00000840, 0.00001654, -0.00002603, 0.00001485, -0.00000580, -0.00000854, 0.00001743, -0.00002017, 0.00001116, 0.00000411, -0.00002149, 0.00001959, -0.00000452, -0.00000102, 0.00001002, -0.00001252, 0.00001105, -0.00001193, -0.00000759, 0.00001714, -0.00002555, 0.00000245, 0.00002957, -0.00000005, -0.00000415, -0.00002140, -0.00000479, 0.00002438, 0.00000192, -0.00000143, -0.00000185, -0.00001487, 0.00000099, 0.00001399, 0.00000157, -0.00000607, 0.00000214, -0.00000065, -0.00000236, 0.00000432, -0.00000790, -0.00000199, 0.00000911, -0.00001364, -0.00000221, 0.00001863, -0.00000636, -0.00001119, -0.00000273, -0.00000207, 0.00000658, -0.00000529, -0.00000198, 0.00001060, -0.00000386, 0.00000076, 0.00000030, -0.00000577, 0.00000514, -0.00000306, 0.00000360, 0.00000412, -0.00001440, 0.00000078, -0.00000050, -0.00000506, 0.00000913, -0.00000363, -0.00000379, 0.00000181, -0.00000693, -0.00000331, -0.00000556, -0.00000754, -0.00000458, -0.00000784, -0.00002329, -0.00004672, -0.00001158, 0.00006542, 0.00006641, },
		{ 0.00008162, 0.00004180, 0.00006783, 0.00009257, -0.00001797, -0.00007323, -0.00001756, -0.00001433, -0.00001526, -0.00000047, -0.00000890, -0.00000044, -0.00000277, -0.00000714, -0.00000003, -0.00000382, -0.00000094, -0.00000354, -0.00000135, 0.00000769, -0.00000281, -0.00000655, 0.00000136, 0.00000091, -0.00000118, 0.00000165, -0.00000411, -0.00000655, 0.00000189, -0.00000237, 0.00000161, -0.00000350, -0.00000628, 0.00001598, -0.00001948, -0.00002850, 0.00001300, -0.00000720, 0.00000745, 0.00002536, -0.00000225, 0.00001096, 0.00001411, 0.00001205, -0.00000632, -0.00001126, 0.00002356, -0.00002314, -0.00001444, 0.00003024, -0.00002227, 0.00000164, 0.00000878, -0.00001881, 0.00001906, -0.00001163, -0.00000426, 0.00001726, -0.00002341, 0.00001002, 0.00000745, -0.00000927, 0.00001340, -0.00001033, 0.00000491, -0.00000107, -0.00001761, 0.00001814, -0.00001409, -0.00001389, 0.00002998, 0.00000758, -0.00000369, -0.00001562, -0.00001445, 0.00002033, 0.00000952, -0.00000308, 0.00000083, -0.00001187, -0.00000639, 0.00001132, 0.00000828, -0.00000575, -0.00000139, 0.00000248, -0.00000459, 0.00000398, -0.00000251, -0.00000903, 0.00000937, -0.00000555, -0.00001103, 0.00001758, 0.00000171, -0.00001477, -0.00000255, -0.00000230, 0.00000453, -0.00000160, -0.00000676, 0.00001000, 0.00000099, -0.00000197, 0.00000247, -0.00000652, 0.00000355, -0.00000032, -0.00000079, 0.00000796, -0.00001161, -0.00000516, 0.00000331, -0.00000680, 0.00000609, 0.00000135, -0.00000603, 0.00000191, -0.00000454, -0.00000511, -0.00000440, -0.00000710, -0.00000607, -0.00000691, -0.00001402, -0.00004316, -0.00003714, 0.00004431, },
		{ 0.00001219, 0.00008681, 0.00005743, 0.00005375, 0.00010485, 0.00001756, -0.00008005, -0.00003053, -0.00001025, -0.00002086, 0.00000016, -0.00000797, -0.00000387, 0.00000031, -0.00000900, -0.00000020, -0.00000320, -0.00000245, -0.00000088, -0.00000478, 0.00000625, 0.00000208, -0.00000760, -0.00000078, 0.00000219, -0.00000172, 0.00000091, -0.00000090, -0.00000782, 0.00000022, -0.00000077, -0.00000057, 0.00000043, -0.00000963, 0.00001330, -0.00000504, -0.00003460, 0.00000675, -0.00000121, -0.00000295, 0.00002687, 0.00000171, 0.00000649, 0.00001300, 0.00001174, 0.00000492, -0.00002221, 0.00002034, -0.00000118, -0.00003428, 0.00002957, -0.00000568, -0.00001541, 0.00001885, -0.00001952, 0.00001036, 0.00000312, -0.00001767, 0.00002088, -0.00001474, -0.00000444, 0.00001707, -0.00001126, 0.00001020, -0.00000329, -0.00000328, 0.00000754, -0.00002039, 0.00001064, 0.00000044, -0.00002393, 0.00002210, 0.00001696, -0.00000307, -0.00001022, -0.00001907, 0.00001199, 0.00001569, -0.00000203, 0.00000105, -0.00000751, -0.00001024, 0.00000515, 0.00001236, -0.00000120, -0.00000574, 0.00000380, -0.00000370, 0.00000029, 0.00000348, -0.00001154, 0.00000364, 0.00000347, -0.00001476, 0.00001094, 0.00001061, -0.00001549, -0.00000546, -0.00000051, 0.00000191, 0.00000139, -0.00000881, 0.00000601, 0.00000606, -0.00000289, 0.00000294, -0.00000552, 0.00000061, 0.00000278, -0.00000377, 0.00000824, -0.00000594, -0.00001050, 0.00000417, -0.00000522, 0.00000145, 0.00000492, -0.00000576, 0.00000019, -0.00000190, -0.00000573, -0.00000456, -0.00000619, -0.00000637, -0.00000812, -0.00000826, -0.00003219, -0.00005322, },
		{ -0.00005587, -0.00002343, 0.00007549, 0.00007913, 0.00004578, 0.00010348, 0.00005755, -0.00007504, -0.00004967, -0.00000609, -0.00002492, -0.00000263, -0.00000406, -0.00000857, 0.00000275, -0.00000916, -0.00000231, -0.00000096, -0.00000496, 0.00000121, -0.00000533, 0.00000204, 0.00000618, -0.00000606, -0.00000383, 0.00000280, -0.00000100, -0.00000094, 0.00000159, -0.00000692, -0.00000274, 0.00000104, -0.00000244, 0.00000261, -0.00000924, 0.00000665, 0.00000743, -0.00003361, -0.00000345, 0.00000538, -0.00001066, 0.00002409, 0.00000754, 0.00000217, 0.00001348, 0.00000720, 0.00001557, -0.00002269, 0.00000373, 0.00002226, -0.00004008, 0.00001254, 0.00001656, -0.00002666, 0.00001799, -0.00000975, -0.00000413, 0.00001619, -0.00002206, 0.00001330, 0.00000007, -0.00001671, 0.00001918, -0.00000588, 0.00000262, 0.00000464, -0.00000911, 0.00001051, -0.00001576, -0.00000117, 0.00001183, -0.00002474, 0.00000849, 0.00002422, -0.00000010, -0.00000667, -0.00001915, 0.00000239, 0.00001830, 0.00000140, 0.00000039, -0.00000388, -0.00001047, -0.00000143, 0.00001178, 0.00000577, -0.00000782, 0.00000172, -0.00000022, -0.00000409, 0.00000651, -0.00000838, -0.00000487, 0.00000888, -0.00001202, 0.00000126, 0.00001674, -0.00001157, -0.00001074, 0.00000126, 0.00000061, 0.00000255, -0.00000811, 0.00000043, 0.00000918, -0.00000136, 0.00000217, -0.00000356, -0.00000252, 0.00000486, -0.00000436, 0.00000558, 0.00000031, -0.00001308, 0.00000178, -0.00000187, -0.00000252, 0.00000587, -0.00000352, -0.00000191, -0.00000024, -0.00000501, -0.00000513, -0.00000588, -0.00000556, -0.00000923, -0.00000723, -0.00001909, },
		{ -0.00000935, -0.00004610, -0.00005264, 0.00004613, 0.00009786, 0.00005013, 0.00009062, 0.00009400, -0.00005410, -0.00007233, -0.00000555, -0.00002472, -0.00000991, 0.00000171, -0.00001236, 0.00000265, -0.00000667, -0.00000621, 0.00000173, -0.00000659, 0.00000097, -0.00000294, -0.00000284, 0.00000748, -0.00000211, -0.00000641, 0.00000179, 0.00000086, -0.00000276, 0.00000225, -0.00000401, -0.00000585, 0.00000190, -0.00000297, 0.00000242, -0.00000590, -0.00000100, 0.00001490, -0.00002589, -0.00001494, 0.00001015, -0.00001428, 0.00001702, 0.00001472, -0.00000219, 0.00001508, 0.00000245, 0.00001926, -0.00001096, -0.00001820, 0.00003479, -0.00002702, -0.00001402, 0.00003290, -0.00002451, 0.00000570, 0.00000530, -0.00001647, 0.00002058, -0.00001525, -0.00000133, 0.00001341, -0.00002083, 0.00001256, 0.00000396, -0.00000481, 0.00000937, -0.00000994, 0.00000759, -0.00000682, -0.00001183, 0.00001600, -0.00001733, -0.00000596, 0.00002605, 0.00000565, -0.00000456, -0.00001656, -0.00000582, 0.00001701, 0.00000565, 0.00000031, -0.00000166, -0.00000865, -0.00000576, 0.00000700, 0.00001171, -0.00000546, -0.00000306, 0.00000337, -0.00000619, 0.00000503, -0.00000160, -0.00001158, 0.00000805, -0.00000462, -0.00000731, 0.00001743, -0.00000356, -0.00001588, 0.00000070, 0.00000145, 0.00000218, -0.00000592, -0.00000459, 0.00000917, 0.00000193, 0.00000131, -0.00000154, -0.00000492, 0.00000517, -0.00000271, 0.00000160, 0.00000497, -0.00001215, -0.00000265, 0.00000117, -0.00000439, 0.00000447, -0.00000069, -0.00000321, 0.00000014, -0.00000362, -0.00000523, -0.00000628, -0.00000477, -0.00000870, -0.00000944, },
		{ -0.00001192, -0.00000597, -0.00002952, -0.00006708, 0.00000408, 0.00010272, 0.00006807, 0.00007351, 0.00011850, -0.00001696, -0.00009230, -0.00001311, -0.00001877, -0.00002049, 0.00000633, -0.00001259, -0.00000104, -0.00000205, -0.00001033, 0.00000292, -0.00000578, -0.00000185, 0.00000067, -0.00000586, 0.00000529, 0.00000267, -0.00000692, -0.00000094, 0.00000282, -0.00000324, 0.00000071, -0.00000022, -0.00000763, 0.00000096, -0.00000185, 0.00000031, -0.00000148, -0.00000669, 0.00001608, -0.00001365, -0.00002458, 0.00001115, -0.00001306, 0.00000631, 0.00002200, -0.00000526, 0.00001510, 0.00000206, 0.00001360, 0.00000761, -0.00003365, 0.00002880, -0.00000001, -0.00003709, 0.00003351, -0.00000826, -0.00001171, 0.00001707, -0.00001981, 0.00001412, -0.00000108, -0.00001524, 0.00001858, -0.00001515, 0.00000058, 0.00001296, -0.00000791, 0.00000897, -0.00000610, 0.00000142, 0.00000193, -0.00001723, 0.00001248, -0.00000584, -0.00001637, 0.00002140, 0.00001257, -0.00000246, -0.00001313, -0.00001135, 0.00001294, 0.00000907, 0.00000126, -0.00000023, -0.00000661, -0.00000696, 0.00000061, 0.00001354, 0.00000104, -0.00000773, 0.00000431, -0.00000464, 0.00000027, 0.00000494, -0.00001309, 0.00000165, 0.00000330, -0.00001119, 0.00001250, 0.00000579, -0.00001788, -0.00000321, 0.00000359, 0.00000164, -0.00000375, -0.00000767, 0.00000631, 0.00000536, 0.00000139, 0.00000007, -0.00000616, 0.00000376, 0.00000015, -0.00000191, 0.00000685, -0.00000845, -0.00000716, 0.00000238, -0.00000406, 0.00000192, 0.00000143, -0.00000332, -0.00000027, -0.00000238, -0.00000460, -0.00000671, -0.00000495, -0.00000668, },
		{ -0.00000462, -0.00001217, -0.00000827, -0.00001388, -0.00006373, -0.00003942, 0.00008574, 0.00009374, 0.00006220, 0.00012595, 0.00003120, -0.00010112, -0.00003170, -0.00000863, -0.00003054, 0.00000603, -0.00000782, -0.00000742, 0.00000258, -0.00001238, 0.00000110, -0.00000244, -0.00000574, 0.00000303, -0.00000559, 0.00000075, 0.00000598, -0.00000462, -0.00000431, 0.00000343, -0.00000171, -0.00000223, 0.00000271, -0.00000695, -0.00000180, 0.00000030, -0.00000231, 0.00000179, -0.00000855, 0.00001169, -0.00000037, -0.00002945, 0.00000732, -0.00000726, -0.00000619, 0.00002685, -0.00000449, 0.00001074, 0.00000771, 0.00000202, 0.00002326, -0.00003360, 0.00000606, 0.00002807, -0.00004410, 0.00001610, 0.00001471, -0.00002424, 0.00001854, -0.00001239, 0.00000066, 0.00001283, -0.00002129, 0.00001334, -0.00000310, -0.00001071, 0.00001639, -0.00000511, 0.00000450, -0.00000027, -0.00000438, 0.00000701, -0.00001626, 0.00000406, 0.00000467, -0.00001994, 0.00001196, 0.00001803, 0.00000077, -0.00000976, -0.00001426, 0.00000768, 0.00001086, 0.00000263, 0.00000132, -0.00000519, -0.00000605, -0.00000431, 0.00001028, 0.00000891, -0.00000905, 0.00000141, -0.00000055, -0.00000470, 0.00000781, -0.00000890, -0.00000679, 0.00000753, -0.00000923, 0.00000442, 0.00001285, -0.00001505, -0.00000948, 0.00000501, 0.00000231, -0.00000240, -0.00000869, 0.00000197, 0.00000736, 0.00000271, 0.00000141, -0.00000630, 0.00000118, 0.00000300, -0.00000378, 0.00000605, -0.00000357, -0.00001009, 0.00000139, -0.00000244, -0.00000045, 0.00000229, -0.00000263, -0.00000079, -0.00000162, -0.00000368, -0.00000643, -0.00000612, },
		{ -0.00000733, -0.00000403, -0.00000974, -0.00001265, -0.00000531, -0.00004643, -0.00007132, 0.00004638, 0.00011511, 0.00006503, 0.00011759, 0.00008045, -0.00009109, -0.00006004, 0.00000052, -0.00003507, -0.00000154, 0.00000078, -0.00001354, 0.00000444, -0.00001080, -0.00000361, 0.00000177, -0.00000831, 0.00000244, -0.00000249, -0.00000371, 0.00000613, -0.00000030, -0.00000644, 0.00000167, 0.00000133, -0.00000483, 0.00000316, -0.00000367, -0.00000534, 0.00000205, -0.00000374, 0.00000237, -0.00000647, 0.00000434, 0.00001013, -0.00002781, -0.00000094, 0.00000124, -0.00001717, 0.00002602, 0.00000205, 0.00000172, 0.00001651, -0.00000799, 0.00002738, -0.00001704, -0.00002252, 0.00004266, -0.00002991, -0.00001189, 0.00003229, -0.00002384, 0.00000852, 0.00000133, -0.00001220, 0.00001933, -0.00001676, 0.00000093, 0.00000874, -0.00001588, 0.00001265, 0.00000181, -0.00000082, 0.00000430, -0.00000714, 0.00000733, -0.00001068, -0.00000506, 0.00001053, -0.00001690, 0.00000114, 0.00001988, 0.00000512, -0.00000635, -0.00001520, 0.00000236, 0.00001125, 0.00000351, 0.00000324, -0.00000396, -0.00000489, -0.00000596, 0.00000360, 0.00001431, -0.00000532, -0.00000397, 0.00000322, -0.00000690, 0.00000588, -0.00000150, -0.00001285, 0.00000594, -0.00000332, -0.00000296, 0.00001505, -0.00000797, -0.00001547, 0.00000372, 0.00000450, -0.00000153, -0.00000842, -0.00000229, 0.00000717, 0.00000467, 0.00000292, -0.00000551, -0.00000189, 0.00000487, -0.00000364, 0.00000352, 0.00000082, -0.00001069, -0.00000113, -0.00000073, -0.00000189, 0.00000208, -0.00000178, -0.00000112, -0.00000114, -0.00000301, -0.00000535, },
		{ -0.00000408, -0.00000752, -0.00000526, -0.00000626, -0.00001503, -0.00000552, -0.00002423, -0.00008232, -0.00000674, 0.00011840, 0.00008377, 0.00010154, 0.00011951, -0.00005912, -0.00009143, 0.00000147, -0.00003054, -0.00001541, 0.00000923, -0.00001573, 0.00000179, -0.00000595, -0.00000926, 0.00000437, -0.00000783, -0.00000096, 0.00000130, -0.00000575, 0.00000309, 0.00000387, -0.00000571, -0.00000210, 0.00000414, -0.00000522, 0.00000058, 0.00000098, -0.00000782, 0.00000192, -0.00000292, 0.00000018, -0.00000220, -0.00000246, 0.00001500, -0.00001985, -0.00001141, 0.00000918, -0.00002281, 0.00001748, 0.00001355, -0.00000851, 0.00002242, -0.00000938, 0.00001768, 0.00000804, -0.00004230, 0.00003535, -0.00000029, -0.00003650, 0.00003414, -0.00000935, -0.00000731, 0.00001325, -0.00001752, 0.00001561, -0.00000459, -0.00001184, 0.00001428, -0.00001299, 0.00000401, 0.00000888, -0.00000371, 0.00000570, -0.00000635, 0.00000419, -0.00000372, -0.00001135, 0.00001069, -0.00000987, -0.00000761, 0.00001757, 0.00000946, -0.00000252, -0.00001462, -0.00000263, 0.00001091, 0.00000360, 0.00000480, -0.00000193, -0.00000472, -0.00000467, -0.00000316, 0.00001451, 0.00000238, -0.00000863, 0.00000396, -0.00000530, 0.00000080, 0.00000516, -0.00001362, -0.00000041, 0.00000276, -0.00000650, 0.00001221, 0.00000072, -0.00001837, -0.00000090, 0.00000713, -0.00000013, -0.00000775, -0.00000552, 0.00000509, 0.00000622, 0.00000485, -0.00000382, -0.00000482, 0.00000523, -0.00000189, 0.00000055, 0.00000365, -0.00000917, -0.00000411, 0.00000023, -0.00000236, 0.00000136, -0.00000112, -0.00000136, -0.00000062, -0.00000269, },
		{ -0.00000235, -0.00000336, -0.00000636, -0.00000722, -0.00000416, -0.00001335, -0.00001139, -0.00000710, -0.00007109, -0.00005826, 0.00009454, 0.00011105, 0.00008965, 0.00014063, -0.00000919, -0.00011513, -0.00001168, -0.00001756, -0.00003082, 0.00001240, -0.00001171, -0.00000465, -0.00000021, -0.00001309, 0.00000355, -0.00000439, -0.00000528, 0.00000332, -0.00000454, -0.00000129, 0.00000551, -0.00000190, -0.00000602, 0.00000459, -0.00000253, -0.00000390, 0.00000466, -0.00000744, -0.00000078, -0.00000012, -0.00000325, 0.00000146, -0.00000578, 0.00001355, -0.00000808, -0.00002047, 0.00001288, -0.00002059, 0.00000221, 0.00002565, -0.00001411, 0.00001998, -0.00000016, -0.00000021, 0.00002876, -0.00004273, 0.00000878, 0.00003036, -0.00004445, 0.00001780, 0.00001221, -0.00001947, 0.00001646, -0.00001296, 0.00000458, 0.00000850, -0.00001841, 0.00001114, -0.00000447, -0.00000490, 0.00001234, -0.00000265, 0.00000416, -0.00000338, 0.00000003, 0.00000173, -0.00001328, 0.00000654, -0.00000226, -0.00001229, 0.00001223, 0.00001237, 0.00000164, -0.00001249, -0.00000728, 0.00001016, 0.00000353, 0.00000489, 0.00000138, -0.00000533, -0.00000242, -0.00000696, 0.00000937, 0.00001065, -0.00000936, 0.00000087, -0.00000134, -0.00000426, 0.00000796, -0.00000907, -0.00000805, 0.00000541, -0.00000538, 0.00000651, 0.00000776, -0.00001671, -0.00000761, 0.00000828, 0.00000253, -0.00000695, -0.00000763, 0.00000205, 0.00000653, 0.00000690, -0.00000117, -0.00000705, 0.00000395, 0.00000070, -0.00000179, 0.00000460, -0.00000636, -0.00000656, 0.00000020, -0.00000221, 0.00000055, -0.00000063, -0.00000177, -0.00000006, },
		{ 0.00000015, -0.00000152, -0.00000338, -0.00000463, -0.00000820, -0.00000483, -0.00000869, -0.00001720, -0.00000120, -0.00004510, -0.00009211, 0.00004449, 0.00013218, 0.00009166, 0.00014356, 0.00004833, -0.00012034, -0.00003985, -0.00000179, -0.00004105, 0.00000672, -0.00000240, -0.00001171, 0.00000325, -0.00001319, -0.00000060, 0.00000012, -0.00000806, 0.00000231, -0.00000130, -0.00000435, 0.00000354, 0.00000320, -0.00000747, 0.00000154, 0.00000226, -0.00000778, 0.00000506, -0.00000357, -0.00000514, 0.00000289, -0.00000548, 0.00000210, -0.00000464, 0.00000769, 0.00000342, -0.00002451, 0.00001013, -0.00001104, -0.00001519, 0.00003204, -0.00001033, 0.00000820, 0.00001495, -0.00001569, 0.00003438, -0.00002318, -0.00002423, 0.00004626, -0.00003048, -0.00000930, 0.00002911, -0.00002019, 0.00000933, -0.00000196, -0.00000717, 0.00001590, -0.00001595, 0.00000167, 0.00000453, -0.00000974, 0.00001073, 0.00000140, 0.00000156, -0.00000029, -0.00000308, 0.00000429, -0.00001146, 0.00000069, 0.00000330, -0.00001283, 0.00000589, 0.00001305, 0.00000537, -0.00000867, -0.00001130, 0.00000861, 0.00000434, 0.00000300, 0.00000516, -0.00000532, -0.00000135, -0.00000667, 0.00000143, 0.00001555, -0.00000494, -0.00000435, 0.00000213, -0.00000657, 0.00000615, -0.00000185, -0.00001306, 0.00000317, -0.00000137, 0.00000115, 0.00001095, -0.00001110, -0.00001388, 0.00000644, 0.00000616, -0.00000553, -0.00000900, -0.00000104, 0.00000542, 0.00000838, 0.00000236, -0.00000800, 0.00000126, 0.00000322, -0.00000284, 0.00000395, -0.00000317, -0.00000797, -0.00000065, -0.00000181, -0.00000024, -0.00000015, -0.00000231, },
		{ -0.00000263, -0.00000039, -0.00000021, -0.00000355, -0.00000354, -0.00000717, -0.00000752, -0.00000431, -0.00001822, -0.00000619, -0.00001706, -0.00009852, -0.00001953, 0.00013123, 0.00010940, 0.00013607, 0.00010071, -0.00010091, -0.00007720, 0.00000791, -0.00004079, -0.00000744, 0.00000797, -0.00001534, 0.00000243, -0.00000976, -0.00000611, 0.00000319, -0.00000788, -0.00000096, 0.00000152, -0.00000429, -0.00000094, 0.00000655, -0.00000471, -0.00000408, 0.00000646, -0.00000831, 0.00000121, 0.00000248, -0.00000882, 0.00000375, -0.00000457, -0.00000088, -0.00000060, 0.00000120, 0.00001063, -0.00002166, 0.00000158, 0.00000192, -0.00002815, 0.00002771, 0.00000309, -0.00000796, 0.00002695, -0.00001934, 0.00002240, 0.00000652, -0.00004719, 0.00003861, -0.00000074, -0.00003329, 0.00003144, -0.00000819, -0.00000339, 0.00000853, -0.00001331, 0.00001466, -0.00000654, -0.00000865, 0.00000912, -0.00000881, 0.00000544, 0.00000588, -0.00000002, 0.00000153, -0.00000431, 0.00000423, -0.00000765, -0.00000450, 0.00000580, -0.00001055, 0.00000032, 0.00001181, 0.00000767, -0.00000355, -0.00001382, 0.00000551, 0.00000659, -0.00000007, 0.00000765, -0.00000328, -0.00000230, -0.00000351, -0.00000557, 0.00001474, 0.00000304, -0.00000847, 0.00000268, -0.00000540, 0.00000148, 0.00000440, -0.00001315, -0.00000267, 0.00000236, -0.00000148, 0.00001016, -0.00000385, -0.00001729, 0.00000147, 0.00000941, -0.00000280, -0.00000986, -0.00000376, 0.00000336, 0.00000863, 0.00000625, -0.00000718, -0.00000229, 0.00000484, -0.00000243, 0.00000235, -0.00000035, -0.00000823, -0.00000199, -0.00000135, -0.00000104, 0.00000032, },
		{ 0.00000049, -0.00000233, -0.00000165, 0.00000092, -0.00000296, -0.00000380, -0.00000467, -0.00000993, -0.00000329, -0.00001344, -0.00001610, 0.00000101, -0.00007852, -0.00007814, 0.00009890, 0.00013417, 0.00013009, 0.00013853, -0.00005845, -0.00011254, 0.00000344, -0.00002935, -0.00002527, 0.00001379, -0.00001299, -0.00000215, -0.00000500, -0.00001039, 0.00000331, -0.00000527, -0.00000431, 0.00000189, -0.00000146, -0.00000501, 0.00000563, 0.00000168, -0.00000925, 0.00000698, -0.00000422, -0.00000550, 0.00000766, -0.00000908, 0.00000096, -0.00000048, -0.00000567, 0.00000296, -0.00000219, 0.00001160, -0.00001309, -0.00000911, 0.00001249, -0.00003125, 0.00001219, 0.00002074, -0.00001991, 0.00002775, -0.00000845, -0.00000004, 0.00003123, -0.00004853, 0.00001056, 0.00002994, -0.00004117, 0.00001688, 0.00001024, -0.00001350, 0.00001240, -0.00001137, 0.00000674, 0.00000440, -0.00001445, 0.00000728, -0.00000376, -0.00000035, 0.00000832, 0.00000044, 0.00000197, -0.00000406, 0.00000267, -0.00000361, -0.00000778, 0.00000565, -0.00000712, -0.00000367, 0.00000968, 0.00000803, 0.00000170, -0.00001377, 0.00000061, 0.00000969, -0.00000247, 0.00000720, 0.00000099, -0.00000430, -0.00000011, -0.00000871, 0.00000871, 0.00001099, -0.00000862, -0.00000012, -0.00000221, -0.00000312, 0.00000704, -0.00000872, -0.00000898, 0.00000313, -0.00000095, 0.00000715, 0.00000236, -0.00001672, -0.00000520, 0.00001055, 0.00000134, -0.00000992, -0.00000618, 0.00000101, 0.00000745, 0.00000963, -0.00000437, -0.00000582, 0.00000491, -0.00000082, 0.00000053, 0.00000162, -0.00000745, -0.00000356, -0.00000092, -0.00000176, },
		{ -0.00000213, 0.00000009, -0.00000141, -0.00000297, 0.00000100, -0.00000120, -0.00000491, -0.00000244, -0.00000977, -0.00000636, -0.00000595, -0.00002307, 0.00000346, -0.00004349, -0.00011280, 0.00003828, 0.00014962, 0.00013496, 0.00016003, -0.00000171, -0.00013364, -0.00001836, -0.00001172, -0.00003960, 0.00001105, -0.00000540, -0.00000762, -0.00000166, -0.00001191, 0.00000085, -0.00000221, -0.00000573, -0.00000033, 0.00000175, -0.00000568, 0.00000021, 0.00000853, -0.00001026, 0.00000227, 0.00000306, -0.00001154, 0.00000855, -0.00000452, -0.00000482, 0.00000454, -0.00000892, 0.00000282, -0.00000084, 0.00000755, -0.00000270, -0.00001702, 0.00001597, -0.00002324, -0.00000941, 0.00003379, -0.00002019, 0.00001533, 0.00001112, -0.00002007, 0.00003867, -0.00002782, -0.00002416, 0.00004567, -0.00002805, -0.00000749, 0.00002459, -0.00001427, 0.00000805, -0.00000376, -0.00000264, 0.00001137, -0.00001329, 0.00000072, 0.00000176, -0.00000373, 0.00000778, 0.00000238, 0.00000186, -0.00000315, 0.00000071, -0.00000040, -0.00000903, 0.00000388, -0.00000369, -0.00000624, 0.00000762, 0.00000694, 0.00000548, -0.00001067, -0.00000531, 0.00001198, -0.00000229, 0.00000349, 0.00000598, -0.00000529, 0.00000126, -0.00000736, 0.00000042, 0.00001522, -0.00000408, -0.00000449, 0.00000045, -0.00000538, 0.00000567, -0.00000228, -0.00001259, 0.00000021, 0.00000114, 0.00000425, 0.00000591, -0.00001283, -0.00001130, 0.00000857, 0.00000608, -0.00000854, -0.00000840, -0.00000124, 0.00000519, 0.00001161, 0.00000008, -0.00000825, 0.00000322, 0.00000135, -0.00000083, 0.00000256, -0.00000589, -0.00000511, -0.00000066, },
		{ -0.00000084, -0.00000192, -0.00000086, -0.00000041, -0.00000350, -0.00000036, 0.00000111, -0.00000546, -0.00000219, -0.00000651, -0.00001128, -0.00000074, -0.00002204, -0.00000690, -0.00000977, -0.00011398, -0.00003480, 0.00013927, 0.00015121, 0.00017132, 0.00005775, -0.00013231, -0.00005363, 0.00000336, -0.00004453, -0.00000026, 0.00000360, -0.00001055, -0.00000115, -0.00001113, -0.00000237, -0.00000060, -0.00000488, -0.00000323, 0.00000244, -0.00000204, -0.00000690, 0.00001145, -0.00000500, -0.00000612, 0.00000964, -0.00001294, 0.00000345, 0.00000360, -0.00001057, 0.00000697, -0.00000770, -0.00000201, 0.00000366, 0.00000235, 0.00000482, -0.00001856, 0.00001123, -0.00000802, -0.00002844, 0.00003457, -0.00000714, -0.00000465, 0.00002817, -0.00002632, 0.00002602, 0.00000406, -0.00004833, 0.00003793, -0.00000018, -0.00002874, 0.00002617, -0.00000472, -0.00000078, 0.00000411, -0.00000833, 0.00001177, -0.00000664, -0.00000653, 0.00000435, -0.00000367, 0.00000519, 0.00000440, 0.00000210, -0.00000214, -0.00000109, 0.00000170, -0.00000872, 0.00000136, -0.00000067, -0.00000790, 0.00000590, 0.00000566, 0.00000661, -0.00000523, -0.00001043, 0.00001167, 0.00000118, -0.00000198, 0.00000922, -0.00000360, 0.00000003, -0.00000326, -0.00000638, 0.00001389, 0.00000332, -0.00000752, 0.00000070, -0.00000485, 0.00000195, 0.00000308, -0.00001201, -0.00000496, 0.00000243, 0.00000300, 0.00000681, -0.00000751, -0.00001491, 0.00000376, 0.00000984, -0.00000530, -0.00001016, -0.00000341, 0.00000252, 0.00001167, 0.00000521, -0.00000866, 0.00000012, 0.00000321, -0.00000128, 0.00000256, -0.00000389, -0.00000635, },
		{ -0.00000699, -0.00000166, -0.00000118, -0.00000186, -0.00000008, -0.00000281, -0.00000256, 0.00000259, -0.00000412, -0.00000423, -0.00000189, -0.00001429, -0.00000137, -0.00001347, -0.00002105, 0.00000904, -0.00008551, -0.00009771, 0.00009531, 0.00016870, 0.00018190, 0.00011121, -0.00010740, -0.00009298, 0.00000745, -0.00003853, -0.00001600, 0.00000917, -0.00000925, -0.00000257, -0.00000987, -0.00000465, -0.00000079, -0.00000335, -0.00000422, -0.00000059, 0.00000378, -0.00001114, 0.00000745, 0.00000514, -0.00001372, 0.00001100, -0.00000761, -0.00000611, 0.00001132, -0.00001226, 0.00000418, -0.00000158, -0.00000924, 0.00000725, 0.00000033, 0.00000664, -0.00001366, 0.00000147, 0.00000728, -0.00003706, 0.00002095, 0.00001353, -0.00002164, 0.00003256, -0.00001539, 0.00000126, 0.00003083, -0.00005023, 0.00001028, 0.00002807, -0.00003512, 0.00001334, 0.00000964, -0.00000756, 0.00000748, -0.00000818, 0.00000688, 0.00000145, -0.00001055, 0.00000280, -0.00000137, 0.00000246, 0.00000530, 0.00000292, -0.00000102, -0.00000259, 0.00000278, -0.00000733, -0.00000144, 0.00000191, -0.00000881, 0.00000399, 0.00000544, 0.00000510, 0.00000065, -0.00001277, 0.00000794, 0.00000687, -0.00000648, 0.00000869, 0.00000081, -0.00000244, 0.00000076, -0.00000910, 0.00000790, 0.00001019, -0.00000701, -0.00000154, -0.00000286, -0.00000174, 0.00000539, -0.00000794, -0.00000973, 0.00000124, 0.00000332, 0.00000634, -0.00000266, -0.00001533, -0.00000235, 0.00001118, -0.00000056, -0.00001078, -0.00000566, -0.00000005, 0.00000997, 0.00000981, -0.00000667, -0.00000348, 0.00000398, -0.00000075, 0.00000197, -0.00000187, },
		{ -0.00000021, -0.00000683, -0.00000303, -0.00000039, -0.00000222, -0.00000077, -0.00000134, -0.00000430, 0.00000201, -0.00000084, -0.00000708, 0.00000115, -0.00001265, -0.00000757, -0.00000270, -0.00002919, 0.00000818, -0.00004258, -0.00013072, 0.00002413, 0.00017070, 0.00019763, 0.00015661, -0.00006397, -0.00012563, -0.00000360, -0.00002527, -0.00003015, 0.00000835, -0.00000496, -0.00000354, -0.00000960, -0.00000591, -0.00000134, -0.00000338, -0.00000217, -0.00000563, 0.00000751, -0.00000890, -0.00000270, 0.00001517, -0.00001518, 0.00000478, 0.00000276, -0.00001549, 0.00001378, -0.00000731, -0.00000355, 0.00000649, -0.00001431, 0.00000584, 0.00000347, 0.00000358, -0.00000580, -0.00000774, 0.00001601, -0.00003225, -0.00000207, 0.00003139, -0.00002610, 0.00002116, 0.00000638, -0.00002146, 0.00003951, -0.00002972, -0.00002348, 0.00004178, -0.00002274, -0.00000723, 0.00002000, -0.00000722, 0.00000523, -0.00000385, 0.00000048, 0.00000691, -0.00000973, -0.00000151, 0.00000079, 0.00000110, 0.00000491, 0.00000383, 0.00000044, -0.00000371, 0.00000287, -0.00000511, -0.00000430, 0.00000389, -0.00000863, 0.00000111, 0.00000667, 0.00000235, 0.00000469, -0.00001133, 0.00000167, 0.00001231, -0.00000747, 0.00000412, 0.00000607, -0.00000384, 0.00000251, -0.00000750, 0.00000023, 0.00001344, -0.00000272, -0.00000459, -0.00000137, -0.00000373, 0.00000458, -0.00000262, -0.00001185, -0.00000238, 0.00000382, 0.00000596, 0.00000071, -0.00001326, -0.00000786, 0.00000954, 0.00000447, -0.00000959, -0.00000788, -0.00000240, 0.00000709, 0.00001284, -0.00000267, -0.00000640, 0.00000323, 0.00000040, 0.00000129, },
		{ 0.00000092, 0.00000094, -0.00000595, -0.00000456, -0.00000014, -0.00000161, -0.00000211, -0.00000019, -0.00000439, -0.00000070, 0.00000291, -0.00000830, 0.00000045, -0.00000656, -0.00001532, 0.00000355, -0.00002620, -0.00000694, -0.00000444, -0.00012579, -0.00005489, 0.00014266, 0.00021519, 0.00019768, -0.00000898, -0.00014381, -0.00002822, -0.00001142, -0.00003818, 0.00000174, -0.00000099, -0.00000227, -0.00000987, -0.00000780, -0.00000036, -0.00000577, 0.00000131, -0.00000882, 0.00000543, -0.00000020, -0.00001409, 0.00001876, -0.00000757, -0.00000707, 0.00001292, -0.00001900, 0.00000802, 0.00000331, -0.00001241, 0.00001153, -0.00001302, -0.00000184, 0.00000984, -0.00000025, 0.00000002, -0.00001172, 0.00001555, -0.00001739, -0.00002473, 0.00003681, -0.00001503, -0.00000024, 0.00002658, -0.00002973, 0.00002723, 0.00000195, -0.00004656, 0.00003354, 0.00000205, -0.00002412, 0.00001959, 0.00000046, 0.00000030, 0.00000086, -0.00000374, 0.00000790, -0.00000526, -0.00000577, 0.00000093, 0.00000126, 0.00000405, 0.00000414, 0.00000224, -0.00000416, 0.00000191, -0.00000232, -0.00000683, 0.00000474, -0.00000680, -0.00000292, 0.00000849, 0.00000044, 0.00000542, -0.00000677, -0.00000489, 0.00001491, -0.00000398, -0.00000265, 0.00000946, -0.00000252, 0.00000167, -0.00000345, -0.00000576, 0.00001188, 0.00000344, -0.00000617, -0.00000157, -0.00000387, 0.00000203, 0.00000160, -0.00001062, -0.00000689, 0.00000296, 0.00000625, 0.00000281, -0.00001006, -0.00001142, 0.00000555, 0.00000826, -0.00000644, -0.00000952, -0.00000467, 0.00000370, 0.00001386, 0.00000236, -0.00000765, 0.00000120, 0.00000149, },
		{ 0.00000189, 0.00000095, 0.00000170, -0.00000466, -0.00000572, -0.00000080, -0.00000033, -0.00000310, -0.00000041, -0.00000265, -0.00000418, 0.00000488, -0.00000621, -0.00000377, 0.00000081, -0.00001938, 0.00000142, -0.00001414, -0.00002461, 0.00001460, -0.00009006, -0.00011719, 0.00008112, 0.00022189, 0.00023878, 0.00005336, -0.00014462, -0.00006087, -0.00000301, -0.00003934, -0.00000723, -0.00000017, 0.00000055, -0.00000829, -0.00001207, 0.00000241, -0.00000869, 0.00000262, -0.00000669, -0.00000272, 0.00001044, -0.00001980, 0.00001196, 0.00000710, -0.00001844, 0.00001645, -0.00001324, -0.00000466, 0.00001471, -0.00001675, 0.00000936, -0.00000447, -0.00001284, 0.00001440, -0.00000006, 0.00000062, -0.00000954, 0.00000842, -0.00000023, -0.00003754, 0.00002626, 0.00000607, -0.00002012, 0.00003387, -0.00001969, 0.00000227, 0.00002844, -0.00004802, 0.00000741, 0.00002596, -0.00002763, 0.00000792, 0.00001060, -0.00000258, 0.00000286, -0.00000438, 0.00000537, -0.00000003, -0.00000756, -0.00000117, 0.00000183, 0.00000372, 0.00000367, 0.00000388, -0.00000354, 0.00000004, 0.00000055, -0.00000838, 0.00000388, -0.00000326, -0.00000731, 0.00000923, 0.00000096, 0.00000304, -0.00000123, -0.00000922, 0.00001335, 0.00000281, -0.00000849, 0.00000886, 0.00000140, -0.00000032, 0.00000043, -0.00000806, 0.00000666, 0.00000863, -0.00000489, -0.00000315, -0.00000322, -0.00000051, 0.00000347, -0.00000701, -0.00001023, 0.00000009, 0.00000667, 0.00000446, -0.00000688, -0.00001277, 0.00000063, 0.00000973, -0.00000207, -0.00000983, -0.00000687, 0.00000020, 0.00001300, 0.00000727, -0.00000686, -0.00000134, },
		{ -0.00000345, 0.00000139, 0.00000107, 0.00000235, -0.00000330, -0.00000622, -0.00000218, 0.00000076, -0.00000281, -0.00000215, -0.00000022, -0.00000634, 0.00000354, -0.00000120, -0.00000883, 0.00000522, -0.00001661, -0.00000773, -0.00000042, -0.00003364, 0.00001120, -0.00004202, -0.00014396, -0.00000220, 0.00020167, 0.00027821, 0.00012348, -0.00012814, -0.00009567, -0.00000273, -0.00003621, -0.00001509, -0.00000254, 0.00000176, -0.00000256, -0.00001834, 0.00000442, -0.00000850, -0.00000134, 0.00000089, -0.00001276, 0.00001612, -0.00001496, -0.00000375, 0.00002186, -0.00002188, 0.00000961, 0.00000030, -0.00001847, 0.00002020, -0.00001230, -0.00000040, 0.00000768, -0.00002105, 0.00001203, 0.00000627, -0.00000314, -0.00000445, 0.00000036, 0.00001150, -0.00003619, 0.00000423, 0.00002614, -0.00002775, 0.00002434, 0.00000225, -0.00002092, 0.00003711, -0.00002837, -0.00002327, 0.00003597, -0.00001544, -0.00000862, 0.00001628, -0.00000029, 0.00000183, -0.00000264, 0.00000186, 0.00000339, -0.00000629, -0.00000415, 0.00000136, 0.00000429, 0.00000297, 0.00000469, -0.00000175, -0.00000224, 0.00000273, -0.00000828, 0.00000113, 0.00000121, -0.00001048, 0.00000741, 0.00000409, -0.00000065, 0.00000272, -0.00000987, 0.00000822, 0.00001014, -0.00001059, 0.00000412, 0.00000601, -0.00000134, 0.00000237, -0.00000677, 0.00000046, 0.00001062, -0.00000109, -0.00000468, -0.00000301, -0.00000210, 0.00000318, -0.00000287, -0.00001105, -0.00000409, 0.00000609, 0.00000624, -0.00000404, -0.00001252, -0.00000380, 0.00000877, 0.00000220, -0.00000836, -0.00000862, -0.00000334, 0.00001069, 0.00001119, -0.00000437, },
		{ -0.00000091, -0.00000451, 0.00000035, 0.00000078, 0.00000305, -0.00000195, -0.00000615, -0.00000364, 0.00000081, -0.00000105, -0.00000440, 0.00000110, -0.00000571, -0.00000075, 0.00000428, -0.00001124, 0.00000405, -0.00000793, -0.00001822, 0.00000687, -0.00002928, -0.00000676, -0.00000251, -0.00013069, -0.00008394, 0.00014490, 0.00030478, 0.00020335, -0.00009325, -0.00012911, -0.00001015, -0.00003194, -0.00002108, -0.00000503, -0.00000199, 0.00000687, -0.00002297, 0.00000153, -0.00000251, -0.00001015, 0.00000963, -0.00001791, 0.00001159, -0.00000032, -0.00002109, 0.00002790, -0.00001295, -0.00000576, 0.00001506, -0.00002571, 0.00001501, 0.00000067, -0.00001323, 0.00001667, -0.00002065, 0.00000104, 0.00001619, -0.00000663, -0.00000152, -0.00000353, 0.00001425, -0.00002343, -0.00001917, 0.00003457, -0.00001927, 0.00000357, 0.00002338, -0.00002990, 0.00002548, 0.00000123, -0.00004313, 0.00002645, 0.00000589, -0.00002036, 0.00001299, 0.00000628, 0.00000020, -0.00000090, -0.00000043, 0.00000409, -0.00000318, -0.00000608, -0.00000074, 0.00000507, 0.00000294, 0.00000431, 0.00000076, -0.00000405, 0.00000346, -0.00000632, -0.00000283, 0.00000509, -0.00001092, 0.00000273, 0.00000839, -0.00000314, 0.00000343, -0.00000720, 0.00000173, 0.00001510, -0.00000782, -0.00000287, 0.00000871, -0.00000010, 0.00000229, -0.00000361, -0.00000413, 0.00000897, 0.00000344, -0.00000470, -0.00000366, -0.00000283, 0.00000180, 0.00000021, -0.00000936, -0.00000802, 0.00000369, 0.00000791, -0.00000122, -0.00001149, -0.00000702, 0.00000620, 0.00000515, -0.00000542, -0.00000923, -0.00000681, 0.00000725, 0.00001378, },
		{ 0.00001501, 0.00000288, -0.00000453, -0.00000053, -0.00000017, 0.00000354, -0.00000038, -0.00000587, -0.00000458, -0.00000041, 0.00000131, -0.00000552, 0.00000001, -0.00000255, -0.00000572, 0.00000706, -0.00000891, -0.00000194, 0.00000212, -0.00002358, 0.00000386, -0.00001512, -0.00002573, 0.00001501, -0.00008882, -0.00013908, 0.00005625, 0.00030076, 0.00029001, -0.00003489, -0.00015918, -0.00002467, -0.00002720, -0.00002769, -0.00000405, -0.00001086, 0.00001539, -0.00002051, -0.00000890, 0.00000805, -0.00001896, 0.00001277, -0.00001302, -0.00000278, 0.00001700, -0.00003018, 0.00001953, 0.00000626, -0.00002201, 0.00002242, -0.00002100, -0.00000030, 0.00001653, -0.00002161, 0.00001634, -0.00000995, -0.00001481, 0.00002335, -0.00000443, -0.00000397, -0.00000210, 0.00001011, -0.00000707, -0.00003382, 0.00002700, 0.00000028, -0.00001682, 0.00003212, -0.00002087, 0.00000188, 0.00002530, -0.00004290, 0.00000223, 0.00002444, -0.00002009, 0.00000176, 0.00001269, 0.00000106, -0.00000060, -0.00000100, 0.00000300, -0.00000023, -0.00000583, -0.00000380, 0.00000486, 0.00000408, 0.00000315, 0.00000304, -0.00000458, 0.00000249, -0.00000300, -0.00000669, 0.00000680, -0.00000808, -0.00000359, 0.00001148, -0.00000258, 0.00000106, -0.00000314, -0.00000358, 0.00001601, -0.00000125, -0.00000903, 0.00000772, 0.00000305, 0.00000152, -0.00000060, -0.00000587, 0.00000500, 0.00000661, -0.00000271, -0.00000456, -0.00000341, 0.00000034, 0.00000167, -0.00000628, -0.00001023, -0.00000029, 0.00000859, 0.00000196, -0.00001005, -0.00000915, 0.00000316, 0.00000625, -0.00000206, -0.00000813, -0.00000974, 0.00000275, },
		{ -0.00000269, 0.00001472, 0.00000675, -0.00000393, -0.00000074, -0.00000144, 0.00000324, 0.00000164, -0.00000558, -0.00000485, -0.00000230, 0.00000284, -0.00000433, -0.00000325, 0.00000118, -0.00000848, 0.00000533, -0.00000264, -0.00000906, 0.00000821, -0.00002040, -0.00000706, -0.00000007, -0.00003430, 0.00000969, -0.00003968, -0.00015256, -0.00004391, 0.00025086, 0.00036987, 0.00005429, -0.00018125, -0.00004862, -0.00001973, -0.00003795, 0.00000071, -0.00002013, 0.00001604, -0.00000759, -0.00002466, 0.00001710, -0.00002064, 0.00000542, 0.00000138, -0.00001998, 0.00002672, -0.00002412, -0.00000161, 0.00002693, -0.00002924, 0.00001655, -0.00000504, -0.00001914, 0.00002658, -0.00001913, 0.00000520, 0.00000681, -0.00002768, 0.00002122, 0.00000573, -0.00001046, 0.00000130, 0.00000474, 0.00000491, -0.00003482, 0.00000748, 0.00001995, -0.00002591, 0.00002441, -0.00000019, -0.00001977, 0.00003244, -0.00002408, -0.00002413, 0.00002962, -0.00000743, -0.00001113, 0.00001381, 0.00000556, -0.00000112, -0.00000089, 0.00000173, 0.00000119, -0.00000374, -0.00000626, 0.00000280, 0.00000598, 0.00000222, 0.00000421, -0.00000354, 0.00000028, 0.00000051, -0.00000898, 0.00000555, -0.00000283, -0.00000928, 0.00001135, 0.00000120, -0.00000261, -0.00000008, -0.00000608, 0.00001308, 0.00000654, -0.00001170, 0.00000309, 0.00000630, 0.00000162, 0.00000117, -0.00000518, 0.00000081, 0.00000724, 0.00000043, -0.00000463, -0.00000429, -0.00000088, 0.00000187, -0.00000322, -0.00001018, -0.00000468, 0.00000740, 0.00000532, -0.00000796, -0.00001065, 0.00000041, 0.00000593, 0.00000049, -0.00000546, -0.00001123, },
		{ -0.00001041, -0.00000847, 0.00001247, 0.00001072, -0.00000307, -0.00000039, -0.00000219, 0.00000164, 0.00000398, -0.00000502, -0.00000489, -0.00000382, 0.00000248, -0.00000100, -0.00000687, 0.00000290, -0.00000734, -0.00000015, 0.00000426, -0.00001290, 0.00000739, -0.00001031, -0.00001905, 0.00000734, -0.00002895, -0.00000828, -0.00000381, -0.00012635, -0.00012686, 0.00015304, 0.00041871, 0.00017691, -0.00018355, -0.00008722, -0.00000768, -0.00005111, 0.00000406, -0.00002195, 0.00000405, 0.00001318, -0.00003791, 0.00001649, -0.00001048, -0.00001126, 0.00001798, -0.00002953, 0.00002119, -0.00000372, -0.00002606, 0.00003742, -0.00002083, -0.00000145, 0.00001472, -0.00003163, 0.00002374, -0.00000475, -0.00001184, 0.00002098, -0.00002976, 0.00000747, 0.00002052, -0.00001543, 0.00000137, 0.00000314, 0.00000877, -0.00002461, -0.00001418, 0.00002915, -0.00001962, 0.00000556, 0.00001989, -0.00002788, 0.00002111, 0.00000239, -0.00003927, 0.00001808, 0.00001066, -0.00001778, 0.00000740, 0.00001166, -0.00000033, -0.00000137, 0.00000124, 0.00000113, -0.00000125, -0.00000679, -0.00000085, 0.00000740, 0.00000251, 0.00000400, -0.00000141, -0.00000215, 0.00000299, -0.00000886, 0.00000175, 0.00000288, -0.00001213, 0.00000740, 0.00000662, -0.00000509, 0.00000046, -0.00000579, 0.00000806, 0.00001279, -0.00000983, -0.00000332, 0.00000756, 0.00000324, 0.00000194, -0.00000343, -0.00000200, 0.00000563, 0.00000318, -0.00000325, -0.00000522, -0.00000212, 0.00000146, -0.00000102, -0.00000840, -0.00000803, 0.00000420, 0.00000799, -0.00000483, -0.00001174, -0.00000205, 0.00000510, 0.00000158, -0.00000223, },
		{ 0.00000010, -0.00000707, -0.00001326, 0.00000769, 0.00001459, -0.00000179, -0.00000023, -0.00000156, -0.00000112, 0.00000582, -0.00000357, -0.00000531, -0.00000423, 0.00000024, 0.00000290, -0.00000848, 0.00000111, -0.00000294, -0.00000633, 0.00000795, -0.00001110, 0.00000081, 0.00000127, -0.00002506, 0.00000387, -0.00001482, -0.00002474, 0.00000849, -0.00007805, -0.00016813, 0.00002453, 0.00040967, 0.00032204, -0.00014686, -0.00014358, 0.00000513, -0.00006010, -0.00000269, -0.00001070, -0.00001836, 0.00003224, -0.00003849, 0.00000127, 0.00000935, -0.00002901, 0.00002591, -0.00002353, 0.00000072, 0.00002151, -0.00004066, 0.00002959, 0.00000186, -0.00002294, 0.00002740, -0.00002994, 0.00000713, 0.00001558, -0.00002540, 0.00002411, -0.00001809, -0.00001356, 0.00003205, -0.00001252, -0.00000501, 0.00000616, 0.00000663, -0.00001075, -0.00002804, 0.00002348, -0.00000266, -0.00001332, 0.00002842, -0.00001936, -0.00000035, 0.00002248, -0.00003622, -0.00000434, 0.00002370, -0.00001352, -0.00000392, 0.00001513, 0.00000352, -0.00000252, 0.00000119, 0.00000062, 0.00000021, -0.00000515, -0.00000474, 0.00000703, 0.00000428, 0.00000300, 0.00000081, -0.00000376, 0.00000367, -0.00000646, -0.00000311, 0.00000687, -0.00001105, 0.00000084, 0.00001112, -0.00000460, -0.00000146, -0.00000413, 0.00000308, 0.00001581, -0.00000431, -0.00000883, 0.00000566, 0.00000564, 0.00000278, -0.00000180, -0.00000305, 0.00000314, 0.00000432, -0.00000082, -0.00000540, -0.00000366, 0.00000080, 0.00000023, -0.00000598, -0.00000947, -0.00000013, 0.00000887, -0.00000072, -0.00001205, -0.00000465, 0.00000447, 0.00000150, },
		{ 0.00000130, 0.00000047, -0.00000221, -0.00001534, 0.00000040, 0.00001746, 0.00000067, -0.00000107, 0.00000057, -0.00000399, 0.00000603, -0.00000076, -0.00000628, -0.00000366, -0.00000267, 0.00000526, -0.00000666, -0.00000357, 0.00000196, -0.00000971, 0.00000641, -0.00000473, -0.00000721, 0.00000851, -0.00002203, -0.00000701, -0.00000139, -0.00002983, 0.00000056, -0.00003177, -0.00015941, -0.00010131, 0.00032669, 0.00046053, -0.00005138, -0.00021061, 0.00000603, -0.00005399, -0.00002564, 0.00001130, -0.00004050, 0.00003688, -0.00002025, -0.00002531, 0.00002894, -0.00003612, 0.00001723, -0.00000237, -0.00002510, 0.00003752, -0.00003564, 0.00000436, 0.00002873, -0.00003556, 0.00002459, -0.00001320, -0.00001640, 0.00003129, -0.00002661, 0.00001303, 0.00000288, -0.00003227, 0.00003207, 0.00000089, -0.00001581, 0.00000984, 0.00000412, -0.00000092, -0.00002947, 0.00000679, 0.00001462, -0.00002198, 0.00002184, -0.00000054, -0.00001904, 0.00002678, -0.00001785, -0.00002598, 0.00002372, -0.00000001, -0.00001387, 0.00001240, 0.00000990, -0.00000289, 0.00000057, 0.00000070, 0.00000018, -0.00000239, -0.00000717, 0.00000433, 0.00000673, 0.00000232, 0.00000224, -0.00000404, 0.00000263, -0.00000280, -0.00000721, 0.00000772, -0.00000660, -0.00000596, 0.00001252, -0.00000095, -0.00000429, -0.00000287, -0.00000045, 0.00001547, 0.00000272, -0.00001132, 0.00000110, 0.00000721, 0.00000450, -0.00000055, -0.00000304, 0.00000116, 0.00000373, 0.00000146, -0.00000419, -0.00000525, -0.00000028, 0.00000091, -0.00000383, -0.00000909, -0.00000420, 0.00000745, 0.00000356, -0.00001086, -0.00000767, 0.00000396, },
		{ 0.00000284, 0.00000201, -0.00000106, 0.00000222, -0.00001347, -0.00000822, 0.00001778, 0.00000503, -0.00000304, 0.00000322, -0.00000537, 0.00000382, 0.00000309, -0.00000713, -0.00000308, -0.00000464, 0.00000472, -0.00000189, -0.00000857, 0.00000420, -0.00000846, 0.00000074, 0.00000245, -0.00001193, 0.00000858, -0.00001224, -0.00001800, 0.00000373, -0.00002273, -0.00001473, -0.00000592, -0.00011281, -0.00018759, 0.00017685, 0.00054964, 0.00011049, -0.00026507, -0.00002327, -0.00002569, -0.00006126, 0.00003166, -0.00004746, 0.00001887, 0.00001308, -0.00005096, 0.00003488, -0.00002444, -0.00000713, 0.00002383, -0.00004162, 0.00003339, -0.00001095, -0.00002731, 0.00004535, -0.00002997, 0.00000575, 0.00001097, -0.00003510, 0.00003273, -0.00001254, -0.00000749, 0.00002286, -0.00003863, 0.00001717, 0.00002089, -0.00002426, 0.00000834, 0.00000602, 0.00000165, -0.00002102, -0.00001159, 0.00002242, -0.00001685, 0.00000536, 0.00001714, -0.00002494, 0.00001511, 0.00000518, -0.00003577, 0.00000984, 0.00001531, -0.00001613, 0.00000329, 0.00001575, -0.00000052, -0.00000114, 0.00000142, -0.00000065, -0.00000010, -0.00000716, 0.00000005, 0.00000842, 0.00000288, 0.00000261, -0.00000318, 0.00000061, 0.00000074, -0.00000916, 0.00000536, -0.00000061, -0.00001053, 0.00000999, 0.00000441, -0.00000600, -0.00000310, -0.00000237, 0.00001289, 0.00000906, -0.00001008, -0.00000436, 0.00000653, 0.00000689, 0.00000097, -0.00000288, 0.00000023, 0.00000235, 0.00000243, -0.00000183, -0.00000611, -0.00000201, 0.00000124, -0.00000219, -0.00000769, -0.00000687, 0.00000415, 0.00000670, -0.00000779, -0.00001070, },
		{ -0.00001268, 0.00000032, 0.00000385, -0.00000321, 0.00000418, -0.00000780, -0.00001583, 0.00001403, 0.00001116, -0.00000508, 0.00000468, -0.00000400, -0.00000047, 0.00000666, -0.00000667, -0.00000356, -0.00000472, 0.00000156, 0.00000357, -0.00001090, 0.00000213, -0.00000362, -0.00000560, 0.00000642, -0.00001112, 0.00000302, -0.00000141, -0.00002269, -0.00000076, -0.00001058, -0.00002380, -0.00000478, -0.00005504, -0.00021027, -0.00000595, 0.00054790, 0.00032138, -0.00026977, -0.00009612, 0.00001881, -0.00009365, 0.00003274, -0.00002904, -0.00001843, 0.00004675, -0.00005947, 0.00001826, 0.00000418, -0.00003582, 0.00003962, -0.00003726, 0.00000836, 0.00002221, -0.00004906, 0.00004063, -0.00000592, -0.00002025, 0.00002978, -0.00003842, 0.00001689, 0.00001125, -0.00002661, 0.00003098, -0.00002796, -0.00000818, 0.00003807, -0.00002273, -0.00000152, 0.00001246, -0.00000039, -0.00001009, -0.00002256, 0.00001706, -0.00000257, -0.00001074, 0.00002406, -0.00001616, -0.00000408, 0.00002060, -0.00002931, -0.00001108, 0.00002339, -0.00000839, -0.00000824, 0.00001705, 0.00000529, -0.00000306, 0.00000191, -0.00000111, 0.00000063, -0.00000499, -0.00000417, 0.00000808, 0.00000478, 0.00000241, -0.00000185, -0.00000146, 0.00000308, -0.00000854, 0.00000093, 0.00000460, -0.00001143, 0.00000446, 0.00000918, -0.00000521, -0.00000463, -0.00000348, 0.00000943, 0.00001333, -0.00000591, -0.00000859, 0.00000330, 0.00000875, 0.00000347, -0.00000282, -0.00000014, 0.00000145, 0.00000178, 0.00000058, -0.00000544, -0.00000421, 0.00000099, -0.00000084, -0.00000613, -0.00000783, 0.00000023, 0.00000765, -0.00000333, },
		{ 0.00000122, -0.00001255, -0.00000369, 0.00000592, -0.00000418, 0.00000267, -0.00000018, -0.00001965, 0.00000585, 0.00001748, -0.00000527, 0.00000343, 0.00000008, -0.00000515, 0.00000815, -0.00000401, -0.00000537, -0.00000327, -0.00000245, 0.00000699, -0.00000895, -0.00000326, 0.00000164, -0.00000898, 0.00000539, -0.00000613, -0.00000377, 0.00000516, -0.00001938, -0.00000996, -0.00000259, -0.00002055, -0.00001827, -0.00001390, -0.00017016, -0.00017032, 0.00043496, 0.00053464, -0.00018667, -0.00020707, 0.00005625, -0.00010009, 0.00000213, 0.00001149, -0.00005829, 0.00006094, -0.00004002, -0.00001812, 0.00003643, -0.00005201, 0.00003286, -0.00001105, -0.00002619, 0.00004620, -0.00004776, 0.00001375, 0.00002619, -0.00003908, 0.00003202, -0.00002320, -0.00000995, 0.00003284, -0.00003299, 0.00002196, -0.00000427, -0.00003299, 0.00004229, -0.00000783, -0.00001711, 0.00001855, -0.00000111, -0.00000376, -0.00002234, 0.00000258, 0.00001128, -0.00001747, 0.00001770, 0.00000080, -0.00001920, 0.00002129, -0.00001098, -0.00002822, 0.00001871, 0.00000587, -0.00001591, 0.00001153, 0.00001277, -0.00000320, 0.00000114, -0.00000054, -0.00000015, -0.00000205, -0.00000666, 0.00000541, 0.00000715, 0.00000254, -0.00000071, -0.00000295, 0.00000380, -0.00000591, -0.00000380, 0.00000731, -0.00000874, -0.00000198, 0.00001137, -0.00000189, -0.00000628, -0.00000470, 0.00000595, 0.00001531, -0.00000038, -0.00001022, -0.00000145, 0.00000865, 0.00000694, -0.00000222, -0.00000082, 0.00000160, 0.00000036, 0.00000173, -0.00000320, -0.00000601, -0.00000026, 0.00000040, -0.00000475, -0.00000762, -0.00000297, 0.00000625, },
		{ 0.00000341, 0.00000436, -0.00000998, -0.00000823, 0.00000670, -0.00000270, -0.00000157, 0.00000635, -0.00001783, -0.00000522, 0.00002128, -0.00000190, -0.00000066, 0.00000517, -0.00000783, 0.00000629, 0.00000060, -0.00000757, -0.00000175, -0.00000516, 0.00000676, -0.00000346, -0.00000894, 0.00000394, -0.00000785, 0.00000080, -0.00000065, -0.00000766, 0.00000571, -0.00001141, -0.00001669, -0.00000369, -0.00000904, -0.00002966, -0.00000439, -0.00009345, -0.00026705, 0.00022669, 0.00068423, 0.00000255, -0.00032295, 0.00005222, -0.00006407, -0.00005640, 0.00005474, -0.00007718, 0.00004172, 0.00000432, -0.00005822, 0.00005312, -0.00004276, 0.00000291, 0.00002495, -0.00005162, 0.00004629, -0.00002145, -0.00002385, 0.00004971, -0.00003851, 0.00001481, 0.00000370, -0.00003478, 0.00004004, -0.00002136, -0.00000031, 0.00002108, -0.00004507, 0.00002864, 0.00001637, -0.00003045, 0.00001759, 0.00000408, -0.00000435, -0.00001416, -0.00001216, 0.00001617, -0.00001232, 0.00000339, 0.00001553, -0.00002214, 0.00000874, 0.00000880, -0.00003286, 0.00000273, 0.00001886, -0.00001482, 0.00000059, 0.00001819, 0.00000018, -0.00000087, 0.00000065, -0.00000140, 0.00000007, -0.00000668, 0.00000131, 0.00000865, 0.00000364, 0.00000008, -0.00000372, 0.00000307, -0.00000235, -0.00000727, 0.00000692, -0.00000387, -0.00000710, 0.00001021, 0.00000273, -0.00000667, -0.00000638, 0.00000257, 0.00001545, 0.00000506, -0.00000913, -0.00000594, 0.00000592, 0.00001026, -0.00000012, -0.00000218, 0.00000232, -0.00000045, 0.00000098, -0.00000037, -0.00000634, -0.00000255, 0.00000123, -0.00000335, -0.00000705, -0.00000478, },
		{ -0.00000545, 0.00000052, 0.00000522, -0.00000579, -0.00001171, 0.00000503, 0.00000105, -0.00000632, 0.00000903, -0.00001061, -0.00001589, 0.00001995, 0.00000524, -0.00000587, 0.00000858, -0.00000669, 0.00000139, 0.00000552, -0.00000848, -0.00000171, -0.00000538, 0.00000331, 0.00000287, -0.00001167, 0.00000187, -0.00000361, -0.00000395, 0.00000196, -0.00000739, 0.00000252, -0.00000427, -0.00001666, -0.00001147, 0.00000050, -0.00002666, -0.00002198, -0.00001967, -0.00027072, -0.00002367, 0.00070808, 0.00027800, -0.00038686, -0.00002185, 0.00001010, -0.00011884, 0.00007238, -0.00005772, -0.00000853, 0.00005457, -0.00007926, 0.00003989, -0.00000802, -0.00003687, 0.00005089, -0.00005202, 0.00001951, 0.00001815, -0.00005332, 0.00005047, -0.00001586, -0.00001386, 0.00002832, -0.00004449, 0.00002733, 0.00000391, -0.00002426, 0.00003509, -0.00003768, 0.00000105, 0.00003935, -0.00003248, 0.00000597, 0.00001464, -0.00000832, -0.00000541, -0.00001912, 0.00000962, -0.00000017, -0.00000950, 0.00002006, -0.00001249, -0.00000849, 0.00001962, -0.00002309, -0.00001681, 0.00002283, -0.00000461, -0.00001079, 0.00001785, 0.00000695, -0.00000270, 0.00000133, -0.00000198, 0.00000054, -0.00000474, -0.00000262, 0.00000827, 0.00000560, 0.00000087, -0.00000392, 0.00000142, 0.00000101, -0.00000857, 0.00000402, 0.00000123, -0.00000941, 0.00000632, 0.00000683, -0.00000508, -0.00000804, -0.00000100, 0.00001426, 0.00000962, -0.00000612, -0.00000866, 0.00000117, 0.00001178, 0.00000386, -0.00000354, 0.00000242, 0.00000029, -0.00000102, 0.00000147, -0.00000467, -0.00000512, 0.00000101, -0.00000163, -0.00000653, },
		{ -0.00000584, -0.00000575, -0.00000139, 0.00000407, -0.00000157, -0.00001280, 0.00000090, 0.00000551, -0.00000897, 0.00000670, -0.00000055, -0.00002233, 0.00001245, 0.00001421, -0.00000925, 0.00000805, -0.00000169, -0.00000450, 0.00000843, -0.00000676, -0.00000363, -0.00000357, -0.00000122, 0.00000701, -0.00000992, -0.00000323, 0.00000048, -0.00000584, 0.00000096, -0.00000493, -0.00000041, -0.00000148, -0.00001110, -0.00001847, -0.00000058, -0.00000900, -0.00004664, 0.00001741, -0.00019222, -0.00024191, 0.00057626, 0.00057616, -0.00033785, -0.00016755, 0.00009046, -0.00014903, 0.00004256, -0.00000143, -0.00006828, 0.00008332, -0.00006414, -0.00000283, 0.00003680, -0.00006476, 0.00004974, -0.00002389, -0.00002220, 0.00005059, -0.00005813, 0.00002508, 0.00001929, -0.00003852, 0.00003696, -0.00003323, -0.00000048, 0.00003038, -0.00003650, 0.00003015, -0.00001370, -0.00002880, 0.00004926, -0.00001865, -0.00001347, 0.00002467, -0.00000913, -0.00000262, -0.00001579, -0.00000375, 0.00001016, -0.00001352, 0.00001318, 0.00000295, -0.00002011, 0.00001669, -0.00000470, -0.00003001, 0.00001451, 0.00000981, -0.00001660, 0.00001056, 0.00001457, -0.00000221, 0.00000066, -0.00000148, -0.00000035, -0.00000223, -0.00000496, 0.00000592, 0.00000761, 0.00000214, -0.00000372, -0.00000067, 0.00000337, -0.00000765, -0.00000011, 0.00000489, -0.00000868, 0.00000128, 0.00000895, -0.00000182, -0.00000871, -0.00000484, 0.00001185, 0.00001315, -0.00000229, -0.00000907, -0.00000396, 0.00001025, 0.00000888, -0.00000348, 0.00000101, 0.00000231, -0.00000271, 0.00000117, -0.00000163, -0.00000666, -0.00000072, 0.00000021, },
		{ 0.00000136, -0.00000445, -0.00000627, -0.00000223, 0.00000213, 0.00000118, -0.00001123, -0.00000439, 0.00000849, -0.00000793, 0.00000052, 0.00000845, -0.00002195, 0.00000032, 0.00002146, -0.00000809, 0.00000316, 0.00000508, -0.00000844, 0.00000765, -0.00000246, -0.00000648, -0.00000149, -0.00000443, 0.00000736, -0.00000473, -0.00000820, 0.00000157, -0.00000433, -0.00000161, -0.00000350, -0.00000064, -0.00000233, -0.00000540, -0.00001789, -0.00001284, 0.00001179, -0.00005567, 0.00000622, -0.00007499, -0.00036142, 0.00031057, 0.00080490, -0.00014058, -0.00034573, 0.00012564, -0.00011738, -0.00003417, 0.00006694, -0.00010502, 0.00006924, -0.00001308, -0.00005681, 0.00006713, -0.00006229, 0.00001787, 0.00002021, -0.00005709, 0.00005733, -0.00003352, -0.00001582, 0.00004915, -0.00004442, 0.00002388, -0.00000616, -0.00003010, 0.00004390, -0.00002928, 0.00000856, 0.00001534, -0.00004704, 0.00003942, 0.00000746, -0.00003192, 0.00002643, -0.00000208, -0.00000715, -0.00000629, -0.00001543, 0.00001155, -0.00000747, 0.00000055, 0.00001483, -0.00002002, 0.00000309, 0.00001224, -0.00003028, -0.00000279, 0.00002064, -0.00001319, -0.00000110, 0.00001908, 0.00000191, -0.00000098, -0.00000049, -0.00000156, -0.00000052, -0.00000523, 0.00000250, 0.00000864, 0.00000407, -0.00000302, -0.00000280, 0.00000432, -0.00000506, -0.00000405, 0.00000621, -0.00000570, -0.00000321, 0.00000860, 0.00000195, -0.00000769, -0.00000845, 0.00000804, 0.00001567, 0.00000170, -0.00000768, -0.00000769, 0.00000570, 0.00001301, -0.00000084, -0.00000164, 0.00000421, -0.00000255, -0.00000118, 0.00000114, -0.00000605, -0.00000359, },
		{ -0.00000629, 0.00000087, -0.00000222, -0.00000686, -0.00000272, 0.00000067, 0.00000192, -0.00000801, -0.00000894, 0.00000846, -0.00000345, -0.00000661, 0.00001287, -0.00001475, -0.00001259, 0.00002336, -0.00000161, -0.00000409, 0.00001029, -0.00000815, 0.00000329, 0.00000274, -0.00000836, -0.00000089, -0.00000509, 0.00000454, 0.00000108, -0.00001016, -0.00000088, -0.00000150, -0.00000268, -0.00000481, 0.00000088, -0.00000291, -0.00000467, -0.00000886, -0.00002799, 0.00002035, -0.00003764, -0.00003775, 0.00002445, -0.00035171, -0.00001599, 0.00087613, 0.00018836, -0.00048081, 0.00006812, -0.00002197, -0.00012823, 0.00010830, -0.00009225, 0.00001114, 0.00005264, -0.00009333, 0.00006230, -0.00002583, -0.00003097, 0.00005699, -0.00006495, 0.00003232, 0.00000961, -0.00005211, 0.00005687, -0.00002598, -0.00000474, 0.00002272, -0.00004652, 0.00003633, -0.00000510, -0.00001831, 0.00003499, -0.00004490, 0.00001254, 0.00003497, -0.00003903, 0.00001551, 0.00001190, -0.00001447, 0.00000169, -0.00001842, 0.00000292, 0.00000333, -0.00000935, 0.00001685, -0.00000933, -0.00001252, 0.00001906, -0.00001792, -0.00002077, 0.00002140, -0.00000175, -0.00001170, 0.00001732, 0.00000890, -0.00000198, -0.00000006, -0.00000215, -0.00000016, -0.00000399, -0.00000066, 0.00000807, 0.00000630, -0.00000159, -0.00000466, 0.00000380, -0.00000166, -0.00000670, 0.00000522, -0.00000183, -0.00000598, 0.00000626, 0.00000497, -0.00000500, -0.00001085, 0.00000284, 0.00001688, 0.00000575, -0.00000545, -0.00000904, -0.00000041, 0.00001423, 0.00000434, -0.00000403, 0.00000438, -0.00000013, -0.00000407, 0.00000188, -0.00000326, },
		{ 0.00000033, -0.00000716, -0.00000152, 0.00000016, -0.00000675, -0.00000369, 0.00000024, 0.00000130, -0.00000477, -0.00001130, 0.00000532, 0.00000250, -0.00001145, 0.00001123, -0.00000354, -0.00002172, 0.00001815, 0.00000832, -0.00001013, 0.00001109, -0.00000343, -0.00000254, 0.00000639, -0.00000775, -0.00000233, -0.00000375, 0.00000071, 0.00000479, -0.00000832, -0.00000487, -0.00000040, -0.00000078, -0.00000748, 0.00000074, 0.00000035, -0.00000953, 0.00000257, -0.00003426, 0.00000814, 0.00000039, -0.00008138, 0.00006292, -0.00023261, -0.00030303, 0.00074408, 0.00057225, -0.00048649, -0.00009629, 0.00010050, -0.00019047, 0.00008920, -0.00002674, -0.00006697, 0.00009891, -0.00008816, 0.00001857, 0.00002896, -0.00007127, 0.00006455, -0.00003877, -0.00001345, 0.00004936, -0.00006438, 0.00003609, 0.00000917, -0.00003356, 0.00003790, -0.00004116, 0.00001032, 0.00002412, -0.00003590, 0.00003554, -0.00002347, -0.00001995, 0.00005086, -0.00002888, -0.00000559, 0.00002619, -0.00001728, 0.00000202, -0.00001150, -0.00001040, 0.00001069, -0.00001062, 0.00000919, 0.00000493, -0.00002116, 0.00001311, 0.00000022, -0.00003056, 0.00001071, 0.00001196, -0.00001576, 0.00000908, 0.00001579, -0.00000041, -0.00000066, -0.00000195, -0.00000080, -0.00000241, -0.00000257, 0.00000609, 0.00000806, 0.00000068, -0.00000582, 0.00000196, 0.00000165, -0.00000751, 0.00000258, 0.00000165, -0.00000668, 0.00000299, 0.00000640, -0.00000143, -0.00001118, -0.00000314, 0.00001613, 0.00000995, -0.00000308, -0.00000823, -0.00000594, 0.00001155, 0.00001048, -0.00000418, 0.00000215, 0.00000341, -0.00000543, -0.00000012, },
		{ -0.00000391, 0.00000258, -0.00000534, -0.00000491, 0.00000139, -0.00000526, -0.00000533, 0.00000032, 0.00000061, -0.00000274, -0.00001128, 0.00000041, 0.00000741, -0.00001203, 0.00000468, 0.00000732, -0.00002383, 0.00000703, 0.00001778, -0.00001161, 0.00000677, 0.00000359, -0.00000682, 0.00000669, -0.00000459, -0.00000480, -0.00000213, -0.00000198, 0.00000538, -0.00000427, -0.00000739, -0.00000256, 0.00000264, -0.00000801, -0.00000386, 0.00000725, -0.00001516, 0.00000739, -0.00002466, -0.00002004, 0.00003689, -0.00009294, 0.00003063, -0.00006642, -0.00046146, 0.00043124, 0.00089298, -0.00030500, -0.00032824, 0.00018343, -0.00017608, 0.00000332, 0.00006437, -0.00012480, 0.00009621, -0.00003664, -0.00004589, 0.00007349, -0.00007911, 0.00003521, 0.00001001, -0.00005657, 0.00006405, -0.00004472, -0.00000461, 0.00004340, -0.00004612, 0.00003089, -0.00001677, -0.00002164, 0.00004321, -0.00003430, 0.00001720, 0.00000651, -0.00004343, 0.00004679, -0.00000394, -0.00002790, 0.00003214, -0.00001051, -0.00000595, 0.00000030, -0.00002009, 0.00000888, -0.00000336, -0.00000218, 0.00001439, -0.00001858, -0.00000120, 0.00001460, -0.00002752, -0.00000678, 0.00002050, -0.00001086, -0.00000230, 0.00001882, 0.00000443, -0.00000161, -0.00000158, -0.00000160, -0.00000142, -0.00000306, 0.00000354, 0.00000867, 0.00000351, -0.00000583, -0.00000082, 0.00000405, -0.00000642, -0.00000075, 0.00000386, -0.00000563, -0.00000014, 0.00000618, 0.00000184, -0.00000919, -0.00000864, 0.00001286, 0.00001405, -0.00000058, -0.00000634, -0.00000912, 0.00000563, 0.00001503, -0.00000092, -0.00000155, 0.00000594, -0.00000389, },
		{ 0.00000023, -0.00000721, 0.00000181, -0.00000146, -0.00000743, 0.00000042, -0.00000257, -0.00000687, -0.00000014, 0.00000082, -0.00000204, -0.00000984, -0.00000441, 0.00000950, -0.00000845, -0.00000366, 0.00001389, -0.00001850, -0.00000624, 0.00002268, -0.00000722, -0.00000076, 0.00000947, -0.00000724, 0.00000372, -0.00000045, -0.00000653, -0.00000172, -0.00000278, 0.00000389, -0.00000067, -0.00000667, -0.00000670, 0.00000407, -0.00000371, -0.00001213, 0.00001313, -0.00001460, 0.00000028, -0.00000253, -0.00004791, 0.00004885, -0.00006023, -0.00004536, 0.00006985, -0.00045055, 0.00002778, 0.00103303, 0.00005726, -0.00053748, 0.00015897, -0.00007261, -0.00011807, 0.00013294, -0.00012609, 0.00003754, 0.00004027, -0.00009810, 0.00008101, -0.00004613, -0.00001870, 0.00005627, -0.00007329, 0.00004415, -0.00000186, -0.00004532, 0.00005820, -0.00003403, 0.00000525, 0.00001376, -0.00004371, 0.00004185, -0.00001375, -0.00000985, 0.00003024, -0.00004745, 0.00002376, 0.00002549, -0.00004033, 0.00002437, 0.00000501, -0.00001692, 0.00000896, -0.00002007, -0.00000192, 0.00000666, -0.00000964, 0.00001435, -0.00000716, -0.00001533, 0.00001824, -0.00001369, -0.00002272, 0.00001877, 0.00000067, -0.00001141, 0.00001566, 0.00001119, -0.00000130, -0.00000172, -0.00000201, -0.00000127, -0.00000268, 0.00000134, 0.00000795, 0.00000622, -0.00000442, -0.00000388, 0.00000491, -0.00000385, -0.00000378, 0.00000446, -0.00000353, -0.00000246, 0.00000484, 0.00000392, -0.00000553, -0.00001222, 0.00000714, 0.00001720, 0.00000255, -0.00000451, -0.00000945, -0.00000142, 0.00001582, 0.00000522, -0.00000449, 0.00000571, },
		{ 0.00000249, 0.00000490, -0.00000769, -0.00000200, 0.00000236, -0.00000734, -0.00000266, 0.00000009, -0.00000711, -0.00000184, 0.00000183, -0.00000177, -0.00000829, -0.00000786, 0.00000847, -0.00000255, -0.00001039, 0.00001432, -0.00000824, -0.00001697, 0.00002076, 0.00000152, -0.00000792, 0.00001132, -0.00000359, -0.00000060, 0.00000251, -0.00000633, -0.00000271, -0.00000253, 0.00000229, 0.00000072, -0.00000352, -0.00000953, 0.00000068, 0.00000466, -0.00001924, 0.00001185, -0.00000408, -0.00001589, 0.00001949, -0.00005721, 0.00002621, 0.00000224, -0.00011638, 0.00011941, -0.00029559, -0.00033888, 0.00092464, 0.00051664, -0.00061301, -0.00000424, 0.00008374, -0.00021543, 0.00013344, -0.00006026, -0.00005370, 0.00010385, -0.00010709, 0.00004227, 0.00001389, -0.00006978, 0.00007414, -0.00005273, -0.00000162, 0.00004248, -0.00006491, 0.00004434, -0.00000216, -0.00002505, 0.00003427, -0.00004510, 0.00002022, 0.00001535, -0.00003100, 0.00003652, -0.00003113, -0.00000809, 0.00004615, -0.00003575, 0.00000449, 0.00002249, -0.00002298, 0.00000851, -0.00001011, -0.00001574, 0.00001189, -0.00000866, 0.00000615, 0.00000598, -0.00002162, 0.00001028, 0.00000352, -0.00002947, 0.00000692, 0.00001276, -0.00001363, 0.00000700, 0.00001676, 0.00000168, -0.00000238, -0.00000212, -0.00000158, -0.00000223, 0.00000001, 0.00000636, 0.00000806, -0.00000172, -0.00000637, 0.00000397, -0.00000052, -0.00000572, 0.00000352, -0.00000111, -0.00000372, 0.00000307, 0.00000452, -0.00000146, -0.00001292, -0.00000000, 0.00001817, 0.00000667, -0.00000315, -0.00000772, -0.00000714, 0.00001219, 0.00001198, -0.00000438, },
		{ -0.00000030, -0.00000202, 0.00000750, -0.00000446, -0.00000694, 0.00000367, -0.00000420, -0.00000632, 0.00000106, -0.00000544, -0.00000440, 0.00000256, -0.00000079, -0.00000732, -0.00000979, 0.00000532, 0.00000329, -0.00001332, 0.00000942, 0.00000282, -0.00002173, 0.00001269, 0.00001101, -0.00001134, 0.00000838, 0.00000212, -0.00000363, 0.00000298, -0.00000445, -0.00000396, -0.00000271, 0.00000184, 0.00000029, -0.00000069, -0.00000813, -0.00000712, 0.00001247, -0.00001915, 0.00000049, 0.00001306, -0.00003067, 0.00002686, -0.00003928, -0.00002095, 0.00006115, -0.00013840, 0.00007090, -0.00007683, -0.00055327, 0.00058411, 0.00093269, -0.00047137, -0.00027311, 0.00021532, -0.00022885, 0.00004984, 0.00004680, -0.00013205, 0.00011699, -0.00006191, -0.00002699, 0.00007050, -0.00008965, 0.00005147, -0.00000374, -0.00005001, 0.00006477, -0.00005250, 0.00000749, 0.00003343, -0.00004305, 0.00003409, -0.00002586, -0.00001102, 0.00003801, -0.00003501, 0.00002345, -0.00000343, -0.00003446, 0.00004856, -0.00001505, -0.00001919, 0.00003285, -0.00001856, -0.00000135, 0.00000407, -0.00002450, 0.00000774, -0.00000048, -0.00000411, 0.00001352, -0.00001738, -0.00000402, 0.00001541, -0.00002416, -0.00000960, 0.00001875, -0.00000777, -0.00000338, 0.00001792, 0.00000729, -0.00000255, -0.00000246, -0.00000185, -0.00000216, -0.00000060, 0.00000463, 0.00000862, 0.00000166, -0.00000748, 0.00000141, 0.00000259, -0.00000605, 0.00000149, 0.00000101, -0.00000402, 0.00000141, 0.00000401, 0.00000176, -0.00001076, -0.00000684, 0.00001604, 0.00001145, -0.00000181, -0.00000544, -0.00000988, 0.00000537, 0.00001645, },
		{ 0.00001654, 0.00000661, -0.00000503, 0.00000641, 0.00000126, -0.00001013, 0.00000114, 0.00000055, -0.00000828, -0.00000060, -0.00000248, -0.00000647, 0.00000184, 0.00000123, -0.00000656, -0.00001091, 0.00000139, 0.00000738, -0.00001222, 0.00000183, 0.00001090, -0.00001972, 0.00000161, 0.00001744, -0.00000961, 0.00000233, 0.00000690, -0.00000366, 0.00000140, -0.00000242, -0.00000411, -0.00000397, 0.00000208, 0.00000007, -0.00000074, -0.00000246, -0.00001518, 0.00001375, -0.00000891, -0.00001736, 0.00002723, -0.00003227, 0.00001262, -0.00000147, -0.00006698, 0.00008314, -0.00009549, -0.00003852, 0.00010631, -0.00055875, 0.00011319, 0.00115803, -0.00010223, -0.00054913, 0.00023544, -0.00013233, -0.00008934, 0.00014120, -0.00015238, 0.00006548, 0.00001945, -0.00009205, 0.00009207, -0.00006484, -0.00000241, 0.00004872, -0.00007519, 0.00005234, -0.00001381, -0.00003417, 0.00005392, -0.00003820, 0.00001387, 0.00000324, -0.00003645, 0.00004263, -0.00001992, -0.00000082, 0.00002168, -0.00004415, 0.00003187, 0.00001292, -0.00003572, 0.00002990, -0.00000398, -0.00001505, 0.00001430, -0.00002284, -0.00000457, 0.00000895, -0.00000966, 0.00001211, -0.00000593, -0.00001654, 0.00001662, -0.00001005, -0.00002290, 0.00001501, 0.00000300, -0.00001045, 0.00001330, 0.00001361, -0.00000079, -0.00000328, -0.00000197, -0.00000238, -0.00000102, 0.00000333, 0.00000808, 0.00000482, -0.00000678, -0.00000203, 0.00000455, -0.00000474, -0.00000098, 0.00000239, -0.00000353, 0.00000005, 0.00000304, 0.00000340, -0.00000673, -0.00001159, 0.00001083, 0.00001572, 0.00000040, -0.00000379, -0.00000953, -0.00000208, },
		{ -0.00000760, 0.00001204, 0.00001354, -0.00000422, 0.00000210, 0.00000662, -0.00000928, -0.00000434, 0.00000414, -0.00000693, -0.00000425, 0.00000002, -0.00000667, -0.00000063, 0.00000347, -0.00000497, -0.00001184, -0.00000255, 0.00000922, -0.00000830, -0.00000550, 0.00001409, -0.00001275, -0.00000853, 0.00001867, -0.00000388, -0.00000378, 0.00000851, -0.00000099, -0.00000034, -0.00000182, -0.00000282, -0.00000543, 0.00000131, 0.00000195, -0.00000391, 0.00000411, -0.00001790, 0.00000535, 0.00000819, -0.00003234, 0.00002794, -0.00001545, -0.00001709, 0.00003589, -0.00008489, 0.00005412, -0.00000812, -0.00014374, 0.00017942, -0.00038040, -0.00033616, 0.00109848, 0.00041179, -0.00070068, 0.00009325, 0.00004373, -0.00021861, 0.00016682, -0.00009540, -0.00003094, 0.00009668, -0.00011676, 0.00006355, -0.00000548, -0.00006043, 0.00007643, -0.00006277, 0.00001069, 0.00003131, -0.00005933, 0.00004792, -0.00001232, -0.00001479, 0.00002670, -0.00004404, 0.00002705, 0.00000613, -0.00002283, 0.00003254, -0.00003444, 0.00000417, 0.00003578, -0.00003724, 0.00001399, 0.00001452, -0.00002451, 0.00001470, -0.00001121, -0.00001879, 0.00001276, -0.00000721, 0.00000396, 0.00000585, -0.00002095, 0.00000770, 0.00000541, -0.00002681, 0.00000296, 0.00001276, -0.00001073, 0.00000451, 0.00001759, 0.00000367, -0.00000404, -0.00000230, -0.00000250, -0.00000166, 0.00000250, 0.00000704, 0.00000704, -0.00000447, -0.00000532, 0.00000473, -0.00000219, -0.00000314, 0.00000275, -0.00000249, -0.00000102, 0.00000214, 0.00000353, -0.00000236, -0.00001319, 0.00000369, 0.00001798, 0.00000402, -0.00000300, -0.00000733, },
		{ -0.00000494, -0.00000983, 0.00000472, 0.00001745, 0.00000087, -0.00000298, 0.00000877, -0.00000424, -0.00000998, 0.00000407, -0.00000251, -0.00000775, 0.00000024, -0.00000466, -0.00000379, 0.00000440, -0.00000190, -0.00001240, -0.00000652, 0.00000914, -0.00000317, -0.00001064, 0.00001264, -0.00000394, -0.00001487, 0.00001499, 0.00000308, -0.00000736, 0.00000674, 0.00000237, -0.00000035, -0.00000288, -0.00000136, -0.00000536, -0.00000177, 0.00000579, -0.00000749, 0.00000673, -0.00001202, -0.00001011, 0.00002338, -0.00003441, 0.00001092, 0.00001369, -0.00004532, 0.00005077, -0.00006212, -0.00001198, 0.00007822, -0.00018566, 0.00012494, -0.00011271, -0.00062060, 0.00075675, 0.00091475, -0.00061853, -0.00019030, 0.00021670, -0.00026542, 0.00009664, 0.00001775, -0.00012519, 0.00012704, -0.00008368, -0.00000372, 0.00005864, -0.00009165, 0.00006325, -0.00001810, -0.00003884, 0.00005918, -0.00005500, 0.00001791, 0.00002124, -0.00003585, 0.00003266, -0.00003139, -0.00000054, 0.00002952, -0.00003111, 0.00002558, -0.00001197, -0.00002181, 0.00004381, -0.00002305, -0.00000797, 0.00002815, -0.00002380, 0.00000499, 0.00000450, -0.00002727, 0.00000737, 0.00000125, -0.00000501, 0.00001175, -0.00001591, -0.00000568, 0.00001467, -0.00002004, -0.00001165, 0.00001599, -0.00000424, -0.00000446, 0.00001673, 0.00001001, -0.00000346, -0.00000324, -0.00000236, -0.00000249, 0.00000176, 0.00000610, 0.00000807, -0.00000124, -0.00000746, 0.00000307, 0.00000075, -0.00000430, 0.00000210, -0.00000116, -0.00000182, 0.00000151, 0.00000275, 0.00000104, -0.00001170, -0.00000352, 0.00001712, 0.00000871, -0.00000235, },
		{ -0.00000079, -0.00000349, -0.00000906, -0.00000262, 0.00001654, 0.00000840, -0.00000564, 0.00000651, 0.00000280, -0.00001248, -0.00000032, 0.00000268, -0.00000863, -0.00000236, -0.00000167, -0.00000583, 0.00000299, 0.00000210, -0.00001155, -0.00001078, 0.00000748, 0.00000203, -0.00001301, 0.00000814, 0.00000392, -0.00001661, 0.00000857, 0.00000859, -0.00000767, 0.00000328, 0.00000418, 0.00000183, -0.00000408, -0.00000168, -0.00000271, -0.00000652, 0.00000904, -0.00000752, 0.00000244, 0.00000057, -0.00002476, 0.00002699, -0.00001933, -0.00001779, 0.00003987, -0.00005372, 0.00003262, -0.00000873, -0.00008003, 0.00011778, -0.00014101, -0.00001348, 0.00012305, -0.00066258, 0.00023821, 0.00123280, -0.00027084, -0.00051621, 0.00028532, -0.00018939, -0.00004756, 0.00013173, -0.00016578, 0.00008902, -0.00000560, -0.00007618, 0.00009321, -0.00007807, 0.00001449, 0.00003598, -0.00007025, 0.00005494, -0.00002357, -0.00002091, 0.00004482, -0.00003766, 0.00001921, -0.00000653, -0.00002630, 0.00003853, -0.00002204, 0.00000653, 0.00001127, -0.00003521, 0.00003465, 0.00000013, -0.00002614, 0.00003038, -0.00001246, -0.00000963, 0.00001633, -0.00002530, -0.00000538, 0.00000989, -0.00000900, 0.00000963, -0.00000525, -0.00001622, 0.00001398, -0.00000667, -0.00002182, 0.00001056, 0.00000528, -0.00000926, 0.00001072, 0.00001579, -0.00000033, -0.00000460, -0.00000227, -0.00000313, 0.00000066, 0.00000548, 0.00000824, 0.00000200, -0.00000792, 0.00000013, 0.00000313, -0.00000412, 0.00000071, 0.00000012, -0.00000229, 0.00000102, 0.00000178, 0.00000281, -0.00000817, -0.00000896, 0.00001309, 0.00001324, },
		{ 0.00001608, 0.00000226, -0.00000301, -0.00000676, -0.00000755, 0.00001120, 0.00001516, -0.00000378, 0.00000112, 0.00000834, -0.00000995, -0.00000715, 0.00000544, -0.00000571, -0.00000643, 0.00000011, -0.00000540, -0.00000036, 0.00000532, -0.00000829, -0.00001492, 0.00000405, 0.00000681, -0.00001286, 0.00000230, 0.00000942, -0.00001476, 0.00000184, 0.00001147, -0.00000580, 0.00000033, 0.00000357, 0.00000464, -0.00000313, -0.00000477, 0.00000151, -0.00001022, 0.00000833, -0.00000177, -0.00000722, 0.00001311, -0.00002956, 0.00001456, 0.00000752, -0.00004377, 0.00004681, -0.00003381, -0.00001044, 0.00004728, -0.00011269, 0.00008947, -0.00003299, -0.00015571, 0.00023222, -0.00048013, -0.00028658, 0.00124372, 0.00026904, -0.00073921, 0.00017984, -0.00001074, -0.00019968, 0.00018318, -0.00012492, -0.00000363, 0.00007880, -0.00011503, 0.00007811, -0.00002508, -0.00004526, 0.00007100, -0.00006675, 0.00002068, 0.00001825, -0.00004867, 0.00004604, -0.00001926, -0.00000507, 0.00001685, -0.00003822, 0.00002936, -0.00000131, -0.00001333, 0.00002439, -0.00003212, 0.00001392, 0.00002186, -0.00003277, 0.00002032, 0.00000443, -0.00002149, 0.00001866, -0.00001364, -0.00001936, 0.00001269, -0.00000578, 0.00000219, 0.00000474, -0.00001900, 0.00000500, 0.00000630, -0.00002300, -0.00000106, 0.00001237, -0.00000762, 0.00000199, 0.00001816, 0.00000544, -0.00000534, -0.00000278, -0.00000324, -0.00000087, 0.00000491, 0.00000812, 0.00000461, -0.00000675, -0.00000316, 0.00000418, -0.00000270, -0.00000087, 0.00000099, -0.00000233, 0.00000049, 0.00000111, 0.00000308, -0.00000408, -0.00001156, 0.00000695, },
		{ 0.00000042, 0.00001620, 0.00000641, -0.00000264, -0.00000461, -0.00000913, 0.00000366, 0.00001813, 0.00000239, -0.00000422, 0.00000955, -0.00000315, -0.00001283, 0.00000353, -0.00000012, -0.00000928, -0.00000100, -0.00000279, -0.00000382, 0.00000595, -0.00000274, -0.00001752, -0.00000157, 0.00001066, -0.00001044, -0.00000384, 0.00001238, -0.00001082, -0.00000383, 0.00001226, -0.00000355, -0.00000103, 0.00000168, 0.00000561, 0.00000116, -0.00000935, 0.00000430, -0.00000972, 0.00000207, 0.00000809, -0.00001646, 0.00001758, -0.00001999, -0.00000948, 0.00003276, -0.00005173, 0.00002724, 0.00000683, -0.00005502, 0.00007505, -0.00009106, 0.00000834, 0.00008223, -0.00022621, 0.00018605, -0.00017559, -0.00064824, 0.00093031, 0.00083892, -0.00072774, -0.00009492, 0.00018954, -0.00027907, 0.00013472, -0.00001645, -0.00010599, 0.00012420, -0.00009732, 0.00001917, 0.00004052, -0.00008482, 0.00006811, -0.00002995, -0.00002562, 0.00004842, -0.00005154, 0.00002450, 0.00000930, -0.00002623, 0.00002702, -0.00003217, 0.00000756, 0.00001981, -0.00002351, 0.00002293, -0.00001676, -0.00000819, 0.00003326, -0.00002594, 0.00000294, 0.00001916, -0.00002471, 0.00001094, 0.00000213, -0.00002764, 0.00000699, 0.00000222, -0.00000514, 0.00000908, -0.00001381, -0.00000667, 0.00001275, -0.00001537, -0.00001318, 0.00001284, -0.00000072, -0.00000545, 0.00001541, 0.00001227, -0.00000401, -0.00000414, -0.00000295, -0.00000240, 0.00000388, 0.00000808, 0.00000649, -0.00000452, -0.00000588, 0.00000364, -0.00000059, -0.00000207, 0.00000125, -0.00000195, -0.00000013, 0.00000079, 0.00000245, -0.00000065, -0.00001133, },
		{ -0.00000913, -0.00000493, 0.00001355, 0.00001051, -0.00000137, -0.00000350, -0.00000804, -0.00000318, 0.00001606, 0.00001037, -0.00000616, 0.00000593, 0.00000474, -0.00001398, -0.00000277, 0.00000499, -0.00000849, -0.00000488, 0.00000006, -0.00000522, 0.00000337, 0.00000340, -0.00001677, -0.00000911, 0.00001253, -0.00000570, -0.00000969, 0.00001301, -0.00000572, -0.00000842, 0.00001215, -0.00000177, -0.00000157, 0.00000078, 0.00000318, 0.00000753, -0.00001206, 0.00000252, -0.00000389, -0.00000771, 0.00001673, -0.00001841, 0.00000981, 0.00000046, -0.00003310, 0.00004193, -0.00003463, -0.00001156, 0.00004738, -0.00007514, 0.00005830, -0.00002548, -0.00008246, 0.00014580, -0.00019079, 0.00002923, 0.00011158, -0.00074534, 0.00039246, 0.00124540, -0.00042672, -0.00044713, 0.00030209, -0.00023263, -0.00000125, 0.00010728, -0.00016391, 0.00010311, -0.00002959, -0.00005378, 0.00008440, -0.00008314, 0.00002840, 0.00002091, -0.00005968, 0.00005136, -0.00002901, -0.00000822, 0.00003279, -0.00003280, 0.00002023, -0.00001342, -0.00001550, 0.00003069, -0.00001968, 0.00001035, 0.00000162, -0.00002231, 0.00003115, -0.00000997, -0.00001386, 0.00002555, -0.00001807, -0.00000245, 0.00001480, -0.00002625, -0.00000505, 0.00000966, -0.00000768, 0.00000668, -0.00000462, -0.00001484, 0.00001052, -0.00000343, -0.00002000, 0.00000604, 0.00000733, -0.00000803, 0.00000827, 0.00001740, 0.00000029, -0.00000576, -0.00000287, -0.00000339, 0.00000212, 0.00000801, 0.00000792, -0.00000187, -0.00000747, 0.00000181, 0.00000145, -0.00000246, 0.00000091, -0.00000131, -0.00000074, 0.00000065, 0.00000155, 0.00000154, },
		{ 0.00000252, -0.00000608, -0.00000819, 0.00000905, 0.00001326, 0.00000123, -0.00000324, -0.00000589, -0.00000725, 0.00000994, 0.00001664, -0.00000305, -0.00000035, 0.00000981, -0.00000944, -0.00001049, 0.00000622, -0.00000365, -0.00000928, 0.00000061, -0.00000356, -0.00000106, 0.00000745, -0.00001173, -0.00001690, 0.00001083, 0.00000122, -0.00001443, 0.00001107, 0.00000043, -0.00001236, 0.00001153, 0.00000026, -0.00000309, 0.00000232, -0.00000170, 0.00001245, -0.00000947, -0.00000467, 0.00000483, -0.00001581, 0.00001812, -0.00000962, -0.00000728, 0.00002125, -0.00004287, 0.00002795, 0.00000111, -0.00005035, 0.00006490, -0.00005712, 0.00000500, 0.00004971, -0.00013488, 0.00012701, -0.00007109, -0.00014692, 0.00026605, -0.00058209, -0.00018949, 0.00134032, 0.00010680, -0.00072671, 0.00024183, -0.00006797, -0.00016320, 0.00018017, -0.00014287, 0.00002221, 0.00005407, -0.00010234, 0.00008315, -0.00004081, -0.00002760, 0.00005922, -0.00006395, 0.00002616, 0.00000607, -0.00003508, 0.00003922, -0.00002189, 0.00000205, 0.00000701, -0.00002903, 0.00002687, -0.00000529, -0.00000483, 0.00001402, -0.00002431, 0.00001882, 0.00000736, -0.00002337, 0.00002183, -0.00000513, -0.00001489, 0.00001924, -0.00001603, -0.00001787, 0.00001156, -0.00000417, 0.00000045, 0.00000316, -0.00001605, 0.00000206, 0.00000661, -0.00001867, -0.00000475, 0.00001178, -0.00000473, -0.00000021, 0.00001825, 0.00000701, -0.00000623, -0.00000365, -0.00000359, -0.00000009, 0.00000743, 0.00000922, 0.00000084, -0.00000788, -0.00000070, 0.00000281, -0.00000199, 0.00000025, -0.00000068, -0.00000117, 0.00000050, 0.00000077, },
		{ 0.00000017, 0.00000264, -0.00000309, -0.00000935, 0.00000402, 0.00001388, 0.00000470, -0.00000293, -0.00000420, -0.00000806, 0.00000240, 0.00001846, 0.00000413, -0.00000553, 0.00000962, -0.00000108, -0.00001544, 0.00000198, 0.00000300, -0.00001102, -0.00000242, 0.00000001, -0.00000463, 0.00000737, -0.00000348, -0.00002203, 0.00000445, 0.00000908, -0.00001650, 0.00000598, 0.00000752, -0.00001531, 0.00000941, 0.00000415, -0.00000662, 0.00000547, -0.00000576, 0.00001215, -0.00000064, -0.00001437, 0.00001112, -0.00001676, 0.00000954, 0.00000703, -0.00002444, 0.00003030, -0.00003180, -0.00000389, 0.00003784, -0.00006788, 0.00004741, -0.00000810, -0.00005633, 0.00009465, -0.00012160, 0.00003847, 0.00006957, -0.00025126, 0.00024377, -0.00026052, -0.00062561, 0.00108274, 0.00071443, -0.00078647, -0.00000379, 0.00014158, -0.00026817, 0.00015708, -0.00004842, -0.00007901, 0.00010923, -0.00010014, 0.00003711, 0.00002010, -0.00007083, 0.00006526, -0.00003685, -0.00001323, 0.00003480, -0.00004289, 0.00002617, -0.00000021, -0.00001642, 0.00001867, -0.00002824, 0.00001168, 0.00001115, -0.00001415, 0.00001617, -0.00001641, 0.00000341, 0.00001904, -0.00002313, 0.00001091, 0.00000815, -0.00002111, 0.00001468, -0.00000181, -0.00002561, 0.00000617, 0.00000281, -0.00000496, 0.00000586, -0.00001105, -0.00000739, 0.00001021, -0.00001059, -0.00001419, 0.00000978, 0.00000230, -0.00000613, 0.00001395, 0.00001398, -0.00000408, -0.00000531, -0.00000339, -0.00000208, 0.00000588, 0.00001038, 0.00000361, -0.00000732, -0.00000331, 0.00000318, -0.00000092, -0.00000036, -0.00000026, -0.00000136, 0.00000029, },
		{ 0.00000010, -0.00000030, 0.00000222, -0.00000059, -0.00000893, -0.00000042, 0.00001245, 0.00000802, -0.00000173, -0.00000354, -0.00000668, -0.00000373, 0.00001521, 0.00001227, -0.00000637, 0.00000455, 0.00000706, -0.00001453, -0.00000625, 0.00000753, -0.00000802, -0.00000792, 0.00000267, -0.00000491, 0.00000323, 0.00000488, -0.00002171, -0.00000593, 0.00001522, -0.00001413, -0.00000223, 0.00001433, -0.00001565, 0.00000416, 0.00001057, -0.00001093, 0.00000716, -0.00000539, 0.00000535, 0.00001140, -0.00002085, 0.00000979, -0.00000853, -0.00000593, 0.00002312, -0.00003100, 0.00002110, -0.00000411, -0.00003773, 0.00005572, -0.00005287, 0.00000163, 0.00004686, -0.00009196, 0.00008558, -0.00005025, -0.00007173, 0.00016045, -0.00023625, 0.00008415, 0.00006829, -0.00079070, 0.00055842, 0.00119301, -0.00054976, -0.00035627, 0.00028605, -0.00025421, 0.00004034, 0.00007373, -0.00014782, 0.00010495, -0.00004758, -0.00002951, 0.00006788, -0.00007930, 0.00003658, 0.00000673, -0.00004589, 0.00004245, -0.00002916, 0.00000156, 0.00002027, -0.00002514, 0.00001709, -0.00001609, -0.00000638, 0.00002113, -0.00001364, 0.00000983, -0.00000485, -0.00000818, 0.00002204, -0.00001533, -0.00000177, 0.00001665, -0.00001944, 0.00000438, 0.00001039, -0.00002513, -0.00000428, 0.00000867, -0.00000605, 0.00000334, -0.00000369, -0.00001295, 0.00000668, -0.00000045, -0.00001783, 0.00000204, 0.00000886, -0.00000680, 0.00000608, 0.00001828, 0.00000121, -0.00000690, -0.00000359, -0.00000324, 0.00000338, 0.00001093, 0.00000660, -0.00000596, -0.00000564, 0.00000262, 0.00000033, -0.00000067, -0.00000007, -0.00000138, },
		{ -0.00000136, -0.00000000, -0.00000061, 0.00000143, 0.00000131, -0.00000757, -0.00000373, 0.00000973, 0.00001023, 0.00000056, -0.00000352, -0.00000477, -0.00000682, 0.00000850, 0.00001774, -0.00000190, -0.00000229, 0.00001095, -0.00000766, -0.00001437, 0.00000658, -0.00000094, -0.00001262, 0.00000166, -0.00000150, -0.00000247, 0.00000966, -0.00001509, -0.00001730, 0.00001635, -0.00000660, -0.00001185, 0.00001832, -0.00001135, -0.00000476, 0.00001784, -0.00001264, 0.00000404, 0.00000070, -0.00000520, 0.00002060, -0.00001868, -0.00000065, 0.00000537, -0.00002033, 0.00002915, -0.00002125, -0.00000287, 0.00002618, -0.00005526, 0.00004379, -0.00001113, -0.00004963, 0.00007816, -0.00008155, 0.00002775, 0.00004114, -0.00014609, 0.00015965, -0.00011722, -0.00011630, 0.00027101, -0.00066988, -0.00005304, 0.00137421, -0.00005304, -0.00066988, 0.00027101, -0.00011630, -0.00011722, 0.00015965, -0.00014609, 0.00004114, 0.00002775, -0.00008155, 0.00007816, -0.00004963, -0.00001113, 0.00004379, -0.00005526, 0.00002618, -0.00000287, -0.00002125, 0.00002915, -0.00002033, 0.00000537, -0.00000065, -0.00001868, 0.00002060, -0.00000520, 0.00000070, 0.00000404, -0.00001264, 0.00001784, -0.00000476, -0.00001135, 0.00001832, -0.00001185, -0.00000660, 0.00001635, -0.00001730, -0.00001509, 0.00000966, -0.00000247, -0.00000150, 0.00000166, -0.00001262, -0.00000094, 0.00000658, -0.00001437, -0.00000766, 0.00001095, -0.00000229, -0.00000190, 0.00001774, 0.00000850, -0.00000682, -0.00000477, -0.00000352, 0.00000056, 0.00001023, 0.00000973, -0.00000373, -0.00000757, 0.00000131, 0.00000143, -0.00000061, -0.00000000, },
		{ 0.00000010, -0.00000138, -0.00000007, -0.00000067, 0.00000033, 0.00000262, -0.00000564, -0.00000596, 0.00000660, 0.00001093, 0.00000338, -0.00000324, -0.00000359, -0.00000690, 0.00000121, 0.00001828, 0.00000608, -0.00000680, 0.00000886, 0.00000204, -0.00001783, -0.00000045, 0.00000668, -0.00001295, -0.00000369, 0.00000334, -0.00000605, 0.00000867, -0.00000428, -0.00002513, 0.00001039, 0.00000438, -0.00001944, 0.00001665, -0.00000177, -0.00001533, 0.00002204, -0.00000818, -0.00000485, 0.00000983, -0.00001364, 0.00002113, -0.00000638, -0.00001609, 0.00001709, -0.00002514, 0.00002027, 0.00000156, -0.00002916, 0.00004245, -0.00004589, 0.00000673, 0.00003658, -0.00007930, 0.00006788, -0.00002951, -0.00004758, 0.00010495, -0.00014782, 0.00007373, 0.00004034, -0.00025421, 0.00028605, -0.00035627, -0.00054976, 0.00119301, 0.00055842, -0.00079070, 0.00006829, 0.00008415, -0.00023625, 0.00016045, -0.00007173, -0.00005025, 0.00008558, -0.00009196, 0.00004686, 0.00000163, -0.00005287, 0.00005572, -0.00003773, -0.00000411, 0.00002110, -0.00003100, 0.00002312, -0.00000593, -0.00000853, 0.00000979, -0.00002085, 0.00001140, 0.00000535, -0.00000539, 0.00000716, -0.00001093, 0.00001057, 0.00000416, -0.00001565, 0.00001433, -0.00000223, -0.00001413, 0.00001522, -0.00000593, -0.00002171, 0.00000488, 0.00000323, -0.00000491, 0.00000267, -0.00000792, -0.00000802, 0.00000753, -0.00000625, -0.00001453, 0.00000706, 0.00000455, -0.00000637, 0.00001227, 0.00001521, -0.00000373, -0.00000668, -0.00000354, -0.00000173, 0.00000802, 0.00001245, -0.00000042, -0.00000893, -0.00000059, 0.00000222, -0.00000030, },
		{ 0.00000017, 0.00000029, -0.00000136, -0.00000026, -0.00000036, -0.00000092, 0.00000318, -0.00000331, -0.00000732, 0.00000361, 0.00001038, 0.00000588, -0.00000208, -0.00000339, -0.00000531, -0.00000408, 0.00001398, 0.00001395, -0.00000613, 0.00000230, 0.00000978, -0.00001419, -0.00001059, 0.00001021, -0.00000739, -0.00001105, 0.00000586, -0.00000496, 0.00000281, 0.00000617, -0.00002561, -0.00000181, 0.00001468, -0.00002111, 0.00000815, 0.00001091, -0.00002313, 0.00001904, 0.00000341, -0.00001641, 0.00001617, -0.00001415, 0.00001115, 0.00001168, -0.00002824, 0.00001867, -0.00001642, -0.00000021, 0.00002617, -0.00004289, 0.00003480, -0.00001323, -0.00003685, 0.00006526, -0.00007083, 0.00002010, 0.00003711, -0.00010014, 0.00010923, -0.00007901, -0.00004842, 0.00015708, -0.00026817, 0.00014158, -0.00000379, -0.00078647, 0.00071443, 0.00108274, -0.00062561, -0.00026052, 0.00024377, -0.00025126, 0.00006957, 0.00003847, -0.00012160, 0.00009465, -0.00005633, -0.00000810, 0.00004741, -0.00006788, 0.00003784, -0.00000389, -0.00003180, 0.00003030, -0.00002444, 0.00000703, 0.00000954, -0.00001676, 0.00001112, -0.00001437, -0.00000064, 0.00001215, -0.00000576, 0.00000547, -0.00000662, 0.00000415, 0.00000941, -0.00001531, 0.00000752, 0.00000598, -0.00001650, 0.00000908, 0.00000445, -0.00002203, -0.00000348, 0.00000737, -0.00000463, 0.00000001, -0.00000242, -0.00001102, 0.00000300, 0.00000198, -0.00001544, -0.00000108, 0.00000962, -0.00000553, 0.00000413, 0.00001846, 0.00000240, -0.00000806, -0.00000420, -0.00000293, 0.00000470, 0.00001388, 0.00000402, -0.00000935, -0.00000309, 0.00000264, },
		{ 0.00000252, 0.00000077, 0.00000050, -0.00000117, -0.00000068, 0.00000025, -0.00000199, 0.00000281, -0.00000070, -0.00000788, 0.00000084, 0.00000922, 0.00000743, -0.00000009, -0.00000359, -0.00000365, -0.00000623, 0.00000701, 0.00001825, -0.00000021, -0.00000473, 0.00001178, -0.00000475, -0.00001867, 0.00000661, 0.00000206, -0.00001605, 0.00000316, 0.00000045, -0.00000417, 0.00001156, -0.00001787, -0.00001603, 0.00001924, -0.00001489, -0.00000513, 0.00002183, -0.00002337, 0.00000736, 0.00001882, -0.00002431, 0.00001402, -0.00000483, -0.00000529, 0.00002687, -0.00002903, 0.00000701, 0.00000205, -0.00002189, 0.00003922, -0.00003508, 0.00000607, 0.00002616, -0.00006395, 0.00005922, -0.00002760, -0.00004081, 0.00008315, -0.00010234, 0.00005407, 0.00002221, -0.00014287, 0.00018017, -0.00016320, -0.00006797, 0.00024183, -0.00072671, 0.00010680, 0.00134032, -0.00018949, -0.00058209, 0.00026605, -0.00014692, -0.00007109, 0.00012701, -0.00013488, 0.00004971, 0.00000500, -0.00005712, 0.00006490, -0.00005035, 0.00000111, 0.00002795, -0.00004287, 0.00002125, -0.00000728, -0.00000962, 0.00001812, -0.00001581, 0.00000483, -0.00000467, -0.00000947, 0.00001245, -0.00000170, 0.00000232, -0.00000309, 0.00000026, 0.00001153, -0.00001236, 0.00000043, 0.00001107, -0.00001443, 0.00000122, 0.00001083, -0.00001690, -0.00001173, 0.00000745, -0.00000106, -0.00000356, 0.00000061, -0.00000928, -0.00000365, 0.00000622, -0.00001049, -0.00000944, 0.00000981, -0.00000035, -0.00000305, 0.00001664, 0.00000994, -0.00000725, -0.00000589, -0.00000324, 0.00000123, 0.00001326, 0.00000905, -0.00000819, -0.00000608, },
		{ -0.00000913, 0.00000154, 0.00000155, 0.00000065, -0.00000074, -0.00000131, 0.00000091, -0.00000246, 0.00000145, 0.00000181, -0.00000747, -0.00000187, 0.00000792, 0.00000801, 0.00000212, -0.00000339, -0.00000287, -0.00000576, 0.00000029, 0.00001740, 0.00000827, -0.00000803, 0.00000733, 0.00000604, -0.00002000, -0.00000343, 0.00001052, -0.00001484, -0.00000462, 0.00000668, -0.00000768, 0.00000966, -0.00000505, -0.00002625, 0.00001480, -0.00000245, -0.00001807, 0.00002555, -0.00001386, -0.00000997, 0.00003115, -0.00002231, 0.00000162, 0.00001035, -0.00001968, 0.00003069, -0.00001550, -0.00001342, 0.00002023, -0.00003280, 0.00003279, -0.00000822, -0.00002901, 0.00005136, -0.00005968, 0.00002091, 0.00002840, -0.00008314, 0.00008440, -0.00005378, -0.00002959, 0.00010311, -0.00016391, 0.00010728, -0.00000125, -0.00023263, 0.00030209, -0.00044713, -0.00042672, 0.00124540, 0.00039246, -0.00074534, 0.00011158, 0.00002923, -0.00019079, 0.00014580, -0.00008246, -0.00002548, 0.00005830, -0.00007514, 0.00004738, -0.00001156, -0.00003463, 0.00004193, -0.00003310, 0.00000046, 0.00000981, -0.00001841, 0.00001673, -0.00000771, -0.00000389, 0.00000252, -0.00001206, 0.00000753, 0.00000318, 0.00000078, -0.00000157, -0.00000177, 0.00001215, -0.00000842, -0.00000572, 0.00001301, -0.00000969, -0.00000570, 0.00001253, -0.00000911, -0.00001677, 0.00000340, 0.00000337, -0.00000522, 0.00000006, -0.00000488, -0.00000849, 0.00000499, -0.00000277, -0.00001398, 0.00000474, 0.00000593, -0.00000616, 0.00001037, 0.00001606, -0.00000318, -0.00000804, -0.00000350, -0.00000137, 0.00001051, 0.00001355, -0.00000493, },
		{ 0.00000042, -0.00001133, -0.00000065, 0.00000245, 0.00000079, -0.00000013, -0.00000195, 0.00000125, -0.00000207, -0.00000059, 0.00000364, -0.00000588, -0.00000452, 0.00000649, 0.00000808, 0.00000388, -0.00000240, -0.00000295, -0.00000414, -0.00000401, 0.00001227, 0.00001541, -0.00000545, -0.00000072, 0.00001284, -0.00001318, -0.00001537, 0.00001275, -0.00000667, -0.00001381, 0.00000908, -0.00000514, 0.00000222, 0.00000699, -0.00002764, 0.00000213, 0.00001094, -0.00002471, 0.00001916, 0.00000294, -0.00002594, 0.00003326, -0.00000819, -0.00001676, 0.00002293, -0.00002351, 0.00001981, 0.00000756, -0.00003217, 0.00002702, -0.00002623, 0.00000930, 0.00002450, -0.00005154, 0.00004842, -0.00002562, -0.00002995, 0.00006811, -0.00008482, 0.00004052, 0.00001917, -0.00009732, 0.00012420, -0.00010599, -0.00001645, 0.00013472, -0.00027907, 0.00018954, -0.00009492, -0.00072774, 0.00083892, 0.00093031, -0.00064824, -0.00017559, 0.00018605, -0.00022621, 0.00008223, 0.00000834, -0.00009106, 0.00007505, -0.00005502, 0.00000683, 0.00002724, -0.00005173, 0.00003276, -0.00000948, -0.00001999, 0.00001758, -0.00001646, 0.00000809, 0.00000207, -0.00000972, 0.00000430, -0.00000935, 0.00000116, 0.00000561, 0.00000168, -0.00000103, -0.00000355, 0.00001226, -0.00000383, -0.00001082, 0.00001238, -0.00000384, -0.00001044, 0.00001066, -0.00000157, -0.00001752, -0.00000274, 0.00000595, -0.00000382, -0.00000279, -0.00000100, -0.00000928, -0.00000012, 0.00000353, -0.00001283, -0.00000315, 0.00000955, -0.00000422, 0.00000239, 0.00001813, 0.00000366, -0.00000913, -0.00000461, -0.00000264, 0.00000641, 0.00001620, },
		{ 0.00001608, 0.00000695, -0.00001156, -0.00000408, 0.00000308, 0.00000111, 0.00000049, -0.00000233, 0.00000099, -0.00000087, -0.00000270, 0.00000418, -0.00000316, -0.00000675, 0.00000461, 0.00000812, 0.00000491, -0.00000087, -0.00000324, -0.00000278, -0.00000534, 0.00000544, 0.00001816, 0.00000199, -0.00000762, 0.00001237, -0.00000106, -0.00002300, 0.00000630, 0.00000500, -0.00001900, 0.00000474, 0.00000219, -0.00000578, 0.00001269, -0.00001936, -0.00001364, 0.00001866, -0.00002149, 0.00000443, 0.00002032, -0.00003277, 0.00002186, 0.00001392, -0.00003212, 0.00002439, -0.00001333, -0.00000131, 0.00002936, -0.00003822, 0.00001685, -0.00000507, -0.00001926, 0.00004604, -0.00004867, 0.00001825, 0.00002068, -0.00006675, 0.00007100, -0.00004526, -0.00002508, 0.00007811, -0.00011503, 0.00007880, -0.00000363, -0.00012492, 0.00018318, -0.00019968, -0.00001074, 0.00017984, -0.00073921, 0.00026904, 0.00124372, -0.00028658, -0.00048013, 0.00023222, -0.00015571, -0.00003299, 0.00008947, -0.00011269, 0.00004728, -0.00001044, -0.00003381, 0.00004681, -0.00004377, 0.00000752, 0.00001456, -0.00002956, 0.00001311, -0.00000722, -0.00000177, 0.00000833, -0.00001022, 0.00000151, -0.00000477, -0.00000313, 0.00000464, 0.00000357, 0.00000033, -0.00000580, 0.00001147, 0.00000184, -0.00001476, 0.00000942, 0.00000230, -0.00001286, 0.00000681, 0.00000405, -0.00001492, -0.00000829, 0.00000532, -0.00000036, -0.00000540, 0.00000011, -0.00000643, -0.00000571, 0.00000544, -0.00000715, -0.00000995, 0.00000834, 0.00000112, -0.00000378, 0.00001516, 0.00001120, -0.00000755, -0.00000676, -0.00000301, 0.00000226, },
		{ -0.00000079, 0.00001324, 0.00001309, -0.00000896, -0.00000817, 0.00000281, 0.00000178, 0.00000102, -0.00000229, 0.00000012, 0.00000071, -0.00000412, 0.00000313, 0.00000013, -0.00000792, 0.00000200, 0.00000824, 0.00000548, 0.00000066, -0.00000313, -0.00000227, -0.00000460, -0.00000033, 0.00001579, 0.00001072, -0.00000926, 0.00000528, 0.00001056, -0.00002182, -0.00000667, 0.00001398, -0.00001622, -0.00000525, 0.00000963, -0.00000900, 0.00000989, -0.00000538, -0.00002530, 0.00001633, -0.00000963, -0.00001246, 0.00003038, -0.00002614, 0.00000013, 0.00003465, -0.00003521, 0.00001127, 0.00000653, -0.00002204, 0.00003853, -0.00002630, -0.00000653, 0.00001921, -0.00003766, 0.00004482, -0.00002091, -0.00002357, 0.00005494, -0.00007025, 0.00003598, 0.00001449, -0.00007807, 0.00009321, -0.00007618, -0.00000560, 0.00008902, -0.00016578, 0.00013173, -0.00004756, -0.00018939, 0.00028532, -0.00051621, -0.00027084, 0.00123280, 0.00023821, -0.00066258, 0.00012305, -0.00001348, -0.00014101, 0.00011778, -0.00008003, -0.00000873, 0.00003262, -0.00005372, 0.00003987, -0.00001779, -0.00001933, 0.00002699, -0.00002476, 0.00000057, 0.00000244, -0.00000752, 0.00000904, -0.00000652, -0.00000271, -0.00000168, -0.00000408, 0.00000183, 0.00000418, 0.00000328, -0.00000767, 0.00000859, 0.00000857, -0.00001661, 0.00000392, 0.00000814, -0.00001301, 0.00000203, 0.00000748, -0.00001078, -0.00001155, 0.00000210, 0.00000299, -0.00000583, -0.00000167, -0.00000236, -0.00000863, 0.00000268, -0.00000032, -0.00001248, 0.00000280, 0.00000651, -0.00000564, 0.00000840, 0.00001654, -0.00000262, -0.00000906, -0.00000349, },
		{ -0.00000494, -0.00000235, 0.00000871, 0.00001712, -0.00000352, -0.00001170, 0.00000104, 0.00000275, 0.00000151, -0.00000182, -0.00000116, 0.00000210, -0.00000430, 0.00000075, 0.00000307, -0.00000746, -0.00000124, 0.00000807, 0.00000610, 0.00000176, -0.00000249, -0.00000236, -0.00000324, -0.00000346, 0.00001001, 0.00001673, -0.00000446, -0.00000424, 0.00001599, -0.00001165, -0.00002004, 0.00001467, -0.00000568, -0.00001591, 0.00001175, -0.00000501, 0.00000125, 0.00000737, -0.00002727, 0.00000450, 0.00000499, -0.00002380, 0.00002815, -0.00000797, -0.00002305, 0.00004381, -0.00002181, -0.00001197, 0.00002558, -0.00003111, 0.00002952, -0.00000054, -0.00003139, 0.00003266, -0.00003585, 0.00002124, 0.00001791, -0.00005500, 0.00005918, -0.00003884, -0.00001810, 0.00006325, -0.00009165, 0.00005864, -0.00000372, -0.00008368, 0.00012704, -0.00012519, 0.00001775, 0.00009664, -0.00026542, 0.00021670, -0.00019030, -0.00061853, 0.00091475, 0.00075675, -0.00062060, -0.00011271, 0.00012494, -0.00018566, 0.00007822, -0.00001198, -0.00006212, 0.00005077, -0.00004532, 0.00001369, 0.00001092, -0.00003441, 0.00002338, -0.00001011, -0.00001202, 0.00000673, -0.00000749, 0.00000579, -0.00000177, -0.00000536, -0.00000136, -0.00000288, -0.00000035, 0.00000237, 0.00000674, -0.00000736, 0.00000308, 0.00001499, -0.00001487, -0.00000394, 0.00001264, -0.00001064, -0.00000317, 0.00000914, -0.00000652, -0.00001240, -0.00000190, 0.00000440, -0.00000379, -0.00000466, 0.00000024, -0.00000775, -0.00000251, 0.00000407, -0.00000998, -0.00000424, 0.00000877, -0.00000298, 0.00000087, 0.00001745, 0.00000472, -0.00000983, },
		{ -0.00000760, -0.00000733, -0.00000300, 0.00000402, 0.00001798, 0.00000369, -0.00001319, -0.00000236, 0.00000353, 0.00000214, -0.00000102, -0.00000249, 0.00000275, -0.00000314, -0.00000219, 0.00000473, -0.00000532, -0.00000447, 0.00000704, 0.00000704, 0.00000250, -0.00000166, -0.00000250, -0.00000230, -0.00000404, 0.00000367, 0.00001759, 0.00000451, -0.00001073, 0.00001276, 0.00000296, -0.00002681, 0.00000541, 0.00000770, -0.00002095, 0.00000585, 0.00000396, -0.00000721, 0.00001276, -0.00001879, -0.00001121, 0.00001470, -0.00002451, 0.00001452, 0.00001399, -0.00003724, 0.00003578, 0.00000417, -0.00003444, 0.00003254, -0.00002283, 0.00000613, 0.00002705, -0.00004404, 0.00002670, -0.00001479, -0.00001232, 0.00004792, -0.00005933, 0.00003131, 0.00001069, -0.00006277, 0.00007643, -0.00006043, -0.00000548, 0.00006355, -0.00011676, 0.00009668, -0.00003094, -0.00009540, 0.00016682, -0.00021861, 0.00004373, 0.00009325, -0.00070068, 0.00041179, 0.00109848, -0.00033616, -0.00038040, 0.00017942, -0.00014374, -0.00000812, 0.00005412, -0.00008489, 0.00003589, -0.00001709, -0.00001545, 0.00002794, -0.00003234, 0.00000819, 0.00000535, -0.00001790, 0.00000411, -0.00000391, 0.00000195, 0.00000131, -0.00000543, -0.00000282, -0.00000182, -0.00000034, -0.00000099, 0.00000851, -0.00000378, -0.00000388, 0.00001867, -0.00000853, -0.00001275, 0.00001409, -0.00000550, -0.00000830, 0.00000922, -0.00000255, -0.00001184, -0.00000497, 0.00000347, -0.00000063, -0.00000667, 0.00000002, -0.00000425, -0.00000693, 0.00000414, -0.00000434, -0.00000928, 0.00000662, 0.00000210, -0.00000422, 0.00001354, 0.00001204, },
		{ 0.00001654, -0.00000208, -0.00000953, -0.00000379, 0.00000040, 0.00001572, 0.00001083, -0.00001159, -0.00000673, 0.00000340, 0.00000304, 0.00000005, -0.00000353, 0.00000239, -0.00000098, -0.00000474, 0.00000455, -0.00000203, -0.00000678, 0.00000482, 0.00000808, 0.00000333, -0.00000102, -0.00000238, -0.00000197, -0.00000328, -0.00000079, 0.00001361, 0.00001330, -0.00001045, 0.00000300, 0.00001501, -0.00002290, -0.00001005, 0.00001662, -0.00001654, -0.00000593, 0.00001211, -0.00000966, 0.00000895, -0.00000457, -0.00002284, 0.00001430, -0.00001505, -0.00000398, 0.00002990, -0.00003572, 0.00001292, 0.00003187, -0.00004415, 0.00002168, -0.00000082, -0.00001992, 0.00004263, -0.00003645, 0.00000324, 0.00001387, -0.00003820, 0.00005392, -0.00003417, -0.00001381, 0.00005234, -0.00007519, 0.00004872, -0.00000241, -0.00006484, 0.00009207, -0.00009205, 0.00001945, 0.00006548, -0.00015238, 0.00014120, -0.00008934, -0.00013233, 0.00023544, -0.00054913, -0.00010223, 0.00115803, 0.00011319, -0.00055875, 0.00010631, -0.00003852, -0.00009549, 0.00008314, -0.00006698, -0.00000147, 0.00001262, -0.00003227, 0.00002723, -0.00001736, -0.00000891, 0.00001375, -0.00001518, -0.00000246, -0.00000074, 0.00000007, 0.00000208, -0.00000397, -0.00000411, -0.00000242, 0.00000140, -0.00000366, 0.00000690, 0.00000233, -0.00000961, 0.00001744, 0.00000161, -0.00001972, 0.00001090, 0.00000183, -0.00001222, 0.00000738, 0.00000139, -0.00001091, -0.00000656, 0.00000123, 0.00000184, -0.00000647, -0.00000248, -0.00000060, -0.00000828, 0.00000055, 0.00000114, -0.00001013, 0.00000126, 0.00000641, -0.00000503, 0.00000661, },
		{ -0.00000030, 0.00001645, 0.00000537, -0.00000988, -0.00000544, -0.00000181, 0.00001145, 0.00001604, -0.00000684, -0.00001076, 0.00000176, 0.00000401, 0.00000141, -0.00000402, 0.00000101, 0.00000149, -0.00000605, 0.00000259, 0.00000141, -0.00000748, 0.00000166, 0.00000862, 0.00000463, -0.00000060, -0.00000216, -0.00000185, -0.00000246, -0.00000255, 0.00000729, 0.00001792, -0.00000338, -0.00000777, 0.00001875, -0.00000960, -0.00002416, 0.00001541, -0.00000402, -0.00001738, 0.00001352, -0.00000411, -0.00000048, 0.00000774, -0.00002450, 0.00000407, -0.00000135, -0.00001856, 0.00003285, -0.00001919, -0.00001505, 0.00004856, -0.00003446, -0.00000343, 0.00002345, -0.00003501, 0.00003801, -0.00001102, -0.00002586, 0.00003409, -0.00004305, 0.00003343, 0.00000749, -0.00005250, 0.00006477, -0.00005001, -0.00000374, 0.00005147, -0.00008965, 0.00007050, -0.00002699, -0.00006191, 0.00011699, -0.00013205, 0.00004680, 0.00004984, -0.00022885, 0.00021532, -0.00027311, -0.00047137, 0.00093269, 0.00058411, -0.00055327, -0.00007683, 0.00007090, -0.00013840, 0.00006115, -0.00002095, -0.00003928, 0.00002686, -0.00003067, 0.00001306, 0.00000049, -0.00001915, 0.00001247, -0.00000712, -0.00000813, -0.00000069, 0.00000029, 0.00000184, -0.00000271, -0.00000396, -0.00000445, 0.00000298, -0.00000363, 0.00000212, 0.00000838, -0.00001134, 0.00001101, 0.00001269, -0.00002173, 0.00000282, 0.00000942, -0.00001332, 0.00000329, 0.00000532, -0.00000979, -0.00000732, -0.00000079, 0.00000256, -0.00000440, -0.00000544, 0.00000106, -0.00000632, -0.00000420, 0.00000367, -0.00000694, -0.00000446, 0.00000750, -0.00000202, },
		{ 0.00000249, -0.00000438, 0.00001198, 0.00001219, -0.00000714, -0.00000772, -0.00000315, 0.00000667, 0.00001817, -0.00000000, -0.00001292, -0.00000146, 0.00000452, 0.00000307, -0.00000372, -0.00000111, 0.00000352, -0.00000572, -0.00000052, 0.00000397, -0.00000637, -0.00000172, 0.00000806, 0.00000636, 0.00000001, -0.00000223, -0.00000158, -0.00000212, -0.00000238, 0.00000168, 0.00001676, 0.00000700, -0.00001363, 0.00001276, 0.00000692, -0.00002947, 0.00000352, 0.00001028, -0.00002162, 0.00000598, 0.00000615, -0.00000866, 0.00001189, -0.00001574, -0.00001011, 0.00000851, -0.00002298, 0.00002249, 0.00000449, -0.00003575, 0.00004615, -0.00000809, -0.00003113, 0.00003652, -0.00003100, 0.00001535, 0.00002022, -0.00004510, 0.00003427, -0.00002505, -0.00000216, 0.00004434, -0.00006491, 0.00004248, -0.00000162, -0.00005273, 0.00007414, -0.00006978, 0.00001389, 0.00004227, -0.00010709, 0.00010385, -0.00005370, -0.00006026, 0.00013344, -0.00021543, 0.00008374, -0.00000424, -0.00061301, 0.00051664, 0.00092464, -0.00033888, -0.00029559, 0.00011941, -0.00011638, 0.00000224, 0.00002621, -0.00005721, 0.00001949, -0.00001589, -0.00000408, 0.00001185, -0.00001924, 0.00000466, 0.00000068, -0.00000953, -0.00000352, 0.00000072, 0.00000229, -0.00000253, -0.00000271, -0.00000633, 0.00000251, -0.00000060, -0.00000359, 0.00001132, -0.00000792, 0.00000152, 0.00002076, -0.00001697, -0.00000824, 0.00001432, -0.00001039, -0.00000255, 0.00000847, -0.00000786, -0.00000829, -0.00000177, 0.00000183, -0.00000184, -0.00000711, 0.00000009, -0.00000266, -0.00000734, 0.00000236, -0.00000200, -0.00000769, 0.00000490, },
		{ 0.00000023, 0.00000571, -0.00000449, 0.00000522, 0.00001582, -0.00000142, -0.00000945, -0.00000451, 0.00000255, 0.00001720, 0.00000714, -0.00001222, -0.00000553, 0.00000392, 0.00000484, -0.00000246, -0.00000353, 0.00000446, -0.00000378, -0.00000385, 0.00000491, -0.00000388, -0.00000442, 0.00000622, 0.00000795, 0.00000134, -0.00000268, -0.00000127, -0.00000201, -0.00000172, -0.00000130, 0.00001119, 0.00001566, -0.00001141, 0.00000067, 0.00001877, -0.00002272, -0.00001369, 0.00001824, -0.00001533, -0.00000716, 0.00001435, -0.00000964, 0.00000666, -0.00000192, -0.00002007, 0.00000896, -0.00001692, 0.00000501, 0.00002437, -0.00004033, 0.00002549, 0.00002376, -0.00004745, 0.00003024, -0.00000985, -0.00001375, 0.00004185, -0.00004371, 0.00001376, 0.00000525, -0.00003403, 0.00005820, -0.00004532, -0.00000186, 0.00004415, -0.00007329, 0.00005627, -0.00001870, -0.00004613, 0.00008101, -0.00009810, 0.00004027, 0.00003754, -0.00012609, 0.00013294, -0.00011807, -0.00007261, 0.00015897, -0.00053748, 0.00005726, 0.00103303, 0.00002778, -0.00045055, 0.00006985, -0.00004536, -0.00006023, 0.00004885, -0.00004791, -0.00000253, 0.00000028, -0.00001460, 0.00001313, -0.00001213, -0.00000371, 0.00000407, -0.00000670, -0.00000667, -0.00000067, 0.00000389, -0.00000278, -0.00000172, -0.00000653, -0.00000045, 0.00000372, -0.00000724, 0.00000947, -0.00000076, -0.00000722, 0.00002268, -0.00000624, -0.00001850, 0.00001389, -0.00000366, -0.00000845, 0.00000950, -0.00000441, -0.00000984, -0.00000204, 0.00000082, -0.00000014, -0.00000687, -0.00000257, 0.00000042, -0.00000743, -0.00000146, 0.00000181, -0.00000721, },
		{ -0.00000391, -0.00000389, 0.00000594, -0.00000155, -0.00000092, 0.00001503, 0.00000563, -0.00000912, -0.00000634, -0.00000058, 0.00001405, 0.00001286, -0.00000864, -0.00000919, 0.00000184, 0.00000618, -0.00000014, -0.00000563, 0.00000386, -0.00000075, -0.00000642, 0.00000405, -0.00000082, -0.00000583, 0.00000351, 0.00000867, 0.00000354, -0.00000306, -0.00000142, -0.00000160, -0.00000158, -0.00000161, 0.00000443, 0.00001882, -0.00000230, -0.00001086, 0.00002050, -0.00000678, -0.00002752, 0.00001460, -0.00000120, -0.00001858, 0.00001439, -0.00000218, -0.00000336, 0.00000888, -0.00002009, 0.00000030, -0.00000595, -0.00001051, 0.00003214, -0.00002790, -0.00000394, 0.00004679, -0.00004343, 0.00000651, 0.00001720, -0.00003430, 0.00004321, -0.00002164, -0.00001677, 0.00003089, -0.00004612, 0.00004340, -0.00000461, -0.00004472, 0.00006405, -0.00005657, 0.00001001, 0.00003521, -0.00007911, 0.00007349, -0.00004589, -0.00003664, 0.00009621, -0.00012480, 0.00006437, 0.00000332, -0.00017608, 0.00018343, -0.00032824, -0.00030500, 0.00089298, 0.00043124, -0.00046146, -0.00006642, 0.00003063, -0.00009294, 0.00003689, -0.00002004, -0.00002466, 0.00000739, -0.00001516, 0.00000725, -0.00000386, -0.00000801, 0.00000264, -0.00000256, -0.00000739, -0.00000427, 0.00000538, -0.00000198, -0.00000213, -0.00000480, -0.00000459, 0.00000669, -0.00000682, 0.00000359, 0.00000677, -0.00001161, 0.00001778, 0.00000703, -0.00002383, 0.00000732, 0.00000468, -0.00001203, 0.00000741, 0.00000041, -0.00001128, -0.00000274, 0.00000061, 0.00000032, -0.00000533, -0.00000526, 0.00000139, -0.00000491, -0.00000534, 0.00000258, },
		{ 0.00000033, -0.00000012, -0.00000543, 0.00000341, 0.00000215, -0.00000418, 0.00001048, 0.00001155, -0.00000594, -0.00000823, -0.00000308, 0.00000995, 0.00001613, -0.00000314, -0.00001118, -0.00000143, 0.00000640, 0.00000299, -0.00000668, 0.00000165, 0.00000258, -0.00000751, 0.00000165, 0.00000196, -0.00000582, 0.00000068, 0.00000806, 0.00000609, -0.00000257, -0.00000241, -0.00000080, -0.00000195, -0.00000066, -0.00000041, 0.00001579, 0.00000908, -0.00001576, 0.00001196, 0.00001071, -0.00003056, 0.00000022, 0.00001311, -0.00002116, 0.00000493, 0.00000919, -0.00001062, 0.00001069, -0.00001040, -0.00001150, 0.00000202, -0.00001728, 0.00002619, -0.00000559, -0.00002888, 0.00005086, -0.00001995, -0.00002347, 0.00003554, -0.00003590, 0.00002412, 0.00001032, -0.00004116, 0.00003790, -0.00003356, 0.00000917, 0.00003609, -0.00006438, 0.00004936, -0.00001345, -0.00003877, 0.00006455, -0.00007127, 0.00002896, 0.00001857, -0.00008816, 0.00009891, -0.00006697, -0.00002674, 0.00008920, -0.00019047, 0.00010050, -0.00009629, -0.00048649, 0.00057225, 0.00074408, -0.00030303, -0.00023261, 0.00006292, -0.00008138, 0.00000039, 0.00000814, -0.00003426, 0.00000257, -0.00000953, 0.00000035, 0.00000074, -0.00000748, -0.00000078, -0.00000040, -0.00000487, -0.00000832, 0.00000479, 0.00000071, -0.00000375, -0.00000233, -0.00000775, 0.00000639, -0.00000254, -0.00000343, 0.00001109, -0.00001013, 0.00000832, 0.00001815, -0.00002172, -0.00000354, 0.00001123, -0.00001145, 0.00000250, 0.00000532, -0.00001130, -0.00000477, 0.00000130, 0.00000024, -0.00000369, -0.00000675, 0.00000016, -0.00000152, -0.00000716, },
		{ -0.00000629, -0.00000326, 0.00000188, -0.00000407, -0.00000013, 0.00000438, -0.00000403, 0.00000434, 0.00001423, -0.00000041, -0.00000904, -0.00000545, 0.00000575, 0.00001688, 0.00000284, -0.00001085, -0.00000500, 0.00000497, 0.00000626, -0.00000598, -0.00000183, 0.00000522, -0.00000670, -0.00000166, 0.00000380, -0.00000466, -0.00000159, 0.00000630, 0.00000807, -0.00000066, -0.00000399, -0.00000016, -0.00000215, -0.00000006, -0.00000198, 0.00000890, 0.00001732, -0.00001170, -0.00000175, 0.00002140, -0.00002077, -0.00001792, 0.00001906, -0.00001252, -0.00000933, 0.00001685, -0.00000935, 0.00000333, 0.00000292, -0.00001842, 0.00000169, -0.00001447, 0.00001190, 0.00001551, -0.00003903, 0.00003497, 0.00001254, -0.00004490, 0.00003499, -0.00001831, -0.00000510, 0.00003633, -0.00004652, 0.00002272, -0.00000474, -0.00002598, 0.00005687, -0.00005211, 0.00000961, 0.00003232, -0.00006495, 0.00005699, -0.00003097, -0.00002583, 0.00006230, -0.00009333, 0.00005264, 0.00001114, -0.00009225, 0.00010830, -0.00012823, -0.00002197, 0.00006812, -0.00048081, 0.00018836, 0.00087613, -0.00001599, -0.00035171, 0.00002445, -0.00003775, -0.00003764, 0.00002035, -0.00002799, -0.00000886, -0.00000467, -0.00000291, 0.00000088, -0.00000481, -0.00000268, -0.00000150, -0.00000088, -0.00001016, 0.00000108, 0.00000454, -0.00000509, -0.00000089, -0.00000836, 0.00000274, 0.00000329, -0.00000815, 0.00001029, -0.00000409, -0.00000161, 0.00002336, -0.00001259, -0.00001475, 0.00001287, -0.00000661, -0.00000345, 0.00000846, -0.00000894, -0.00000801, 0.00000192, 0.00000067, -0.00000272, -0.00000686, -0.00000222, 0.00000087, },
		{ 0.00000136, -0.00000359, -0.00000605, 0.00000114, -0.00000118, -0.00000255, 0.00000421, -0.00000164, -0.00000084, 0.00001301, 0.00000570, -0.00000769, -0.00000768, 0.00000170, 0.00001567, 0.00000804, -0.00000845, -0.00000769, 0.00000195, 0.00000860, -0.00000321, -0.00000570, 0.00000621, -0.00000405, -0.00000506, 0.00000432, -0.00000280, -0.00000302, 0.00000407, 0.00000864, 0.00000250, -0.00000523, -0.00000052, -0.00000156, -0.00000049, -0.00000098, 0.00000191, 0.00001908, -0.00000110, -0.00001319, 0.00002064, -0.00000279, -0.00003028, 0.00001224, 0.00000309, -0.00002002, 0.00001483, 0.00000055, -0.00000747, 0.00001155, -0.00001543, -0.00000629, -0.00000715, -0.00000208, 0.00002643, -0.00003192, 0.00000746, 0.00003942, -0.00004704, 0.00001534, 0.00000856, -0.00002928, 0.00004390, -0.00003010, -0.00000616, 0.00002388, -0.00004442, 0.00004915, -0.00001582, -0.00003352, 0.00005733, -0.00005709, 0.00002021, 0.00001787, -0.00006229, 0.00006713, -0.00005681, -0.00001308, 0.00006924, -0.00010502, 0.00006694, -0.00003417, -0.00011738, 0.00012564, -0.00034573, -0.00014058, 0.00080490, 0.00031057, -0.00036142, -0.00007499, 0.00000622, -0.00005567, 0.00001179, -0.00001284, -0.00001789, -0.00000540, -0.00000233, -0.00000064, -0.00000350, -0.00000161, -0.00000433, 0.00000157, -0.00000820, -0.00000473, 0.00000736, -0.00000443, -0.00000149, -0.00000648, -0.00000246, 0.00000765, -0.00000844, 0.00000508, 0.00000316, -0.00000809, 0.00002146, 0.00000032, -0.00002195, 0.00000845, 0.00000052, -0.00000793, 0.00000849, -0.00000439, -0.00001123, 0.00000118, 0.00000213, -0.00000223, -0.00000627, -0.00000445, },
		{ -0.00000584, 0.00000021, -0.00000072, -0.00000666, -0.00000163, 0.00000117, -0.00000271, 0.00000231, 0.00000101, -0.00000348, 0.00000888, 0.00001025, -0.00000396, -0.00000907, -0.00000229, 0.00001315, 0.00001185, -0.00000484, -0.00000871, -0.00000182, 0.00000895, 0.00000128, -0.00000868, 0.00000489, -0.00000011, -0.00000765, 0.00000337, -0.00000067, -0.00000372, 0.00000214, 0.00000761, 0.00000592, -0.00000496, -0.00000223, -0.00000035, -0.00000148, 0.00000066, -0.00000221, 0.00001457, 0.00001056, -0.00001660, 0.00000981, 0.00001451, -0.00003001, -0.00000470, 0.00001669, -0.00002011, 0.00000295, 0.00001318, -0.00001352, 0.00001016, -0.00000375, -0.00001579, -0.00000262, -0.00000913, 0.00002467, -0.00001347, -0.00001865, 0.00004926, -0.00002880, -0.00001370, 0.00003015, -0.00003650, 0.00003038, -0.00000048, -0.00003323, 0.00003696, -0.00003852, 0.00001929, 0.00002508, -0.00005813, 0.00005059, -0.00002220, -0.00002389, 0.00004974, -0.00006476, 0.00003680, -0.00000283, -0.00006414, 0.00008332, -0.00006828, -0.00000143, 0.00004256, -0.00014903, 0.00009046, -0.00016755, -0.00033785, 0.00057616, 0.00057626, -0.00024191, -0.00019222, 0.00001741, -0.00004664, -0.00000900, -0.00000058, -0.00001847, -0.00001110, -0.00000148, -0.00000041, -0.00000493, 0.00000096, -0.00000584, 0.00000048, -0.00000323, -0.00000992, 0.00000701, -0.00000122, -0.00000357, -0.00000363, -0.00000676, 0.00000843, -0.00000450, -0.00000169, 0.00000805, -0.00000925, 0.00001421, 0.00001245, -0.00002233, -0.00000055, 0.00000670, -0.00000897, 0.00000551, 0.00000090, -0.00001280, -0.00000157, 0.00000407, -0.00000139, -0.00000575, },
		{ -0.00000545, -0.00000653, -0.00000163, 0.00000101, -0.00000512, -0.00000467, 0.00000147, -0.00000102, 0.00000029, 0.00000242, -0.00000354, 0.00000386, 0.00001178, 0.00000117, -0.00000866, -0.00000612, 0.00000962, 0.00001426, -0.00000100, -0.00000804, -0.00000508, 0.00000683, 0.00000632, -0.00000941, 0.00000123, 0.00000402, -0.00000857, 0.00000101, 0.00000142, -0.00000392, 0.00000087, 0.00000560, 0.00000827, -0.00000262, -0.00000474, 0.00000054, -0.00000198, 0.00000133, -0.00000270, 0.00000695, 0.00001785, -0.00001079, -0.00000461, 0.00002283, -0.00001681, -0.00002309, 0.00001962, -0.00000849, -0.00001249, 0.00002006, -0.00000950, -0.00000017, 0.00000962, -0.00001912, -0.00000541, -0.00000832, 0.00001464, 0.00000597, -0.00003248, 0.00003935, 0.00000105, -0.00003768, 0.00003509, -0.00002426, 0.00000391, 0.00002733, -0.00004449, 0.00002832, -0.00001386, -0.00001586, 0.00005047, -0.00005332, 0.00001815, 0.00001951, -0.00005202, 0.00005089, -0.00003687, -0.00000802, 0.00003989, -0.00007926, 0.00005457, -0.00000853, -0.00005772, 0.00007238, -0.00011884, 0.00001010, -0.00002185, -0.00038686, 0.00027800, 0.00070808, -0.00002367, -0.00027072, -0.00001967, -0.00002198, -0.00002666, 0.00000050, -0.00001147, -0.00001666, -0.00000427, 0.00000252, -0.00000739, 0.00000196, -0.00000395, -0.00000361, 0.00000187, -0.00001167, 0.00000287, 0.00000331, -0.00000538, -0.00000171, -0.00000848, 0.00000552, 0.00000139, -0.00000669, 0.00000858, -0.00000587, 0.00000524, 0.00001995, -0.00001589, -0.00001061, 0.00000903, -0.00000632, 0.00000105, 0.00000503, -0.00001171, -0.00000579, 0.00000522, 0.00000052, },
		{ 0.00000341, -0.00000478, -0.00000705, -0.00000335, 0.00000123, -0.00000255, -0.00000634, -0.00000037, 0.00000098, -0.00000045, 0.00000232, -0.00000218, -0.00000012, 0.00001026, 0.00000592, -0.00000594, -0.00000913, 0.00000506, 0.00001545, 0.00000257, -0.00000638, -0.00000667, 0.00000273, 0.00001021, -0.00000710, -0.00000387, 0.00000692, -0.00000727, -0.00000235, 0.00000307, -0.00000372, 0.00000008, 0.00000364, 0.00000865, 0.00000131, -0.00000668, 0.00000007, -0.00000140, 0.00000065, -0.00000087, 0.00000018, 0.00001819, 0.00000059, -0.00001482, 0.00001886, 0.00000273, -0.00003286, 0.00000880, 0.00000874, -0.00002214, 0.00001553, 0.00000339, -0.00001232, 0.00001617, -0.00001216, -0.00001416, -0.00000435, 0.00000408, 0.00001759, -0.00003045, 0.00001637, 0.00002864, -0.00004507, 0.00002108, -0.00000031, -0.00002136, 0.00004004, -0.00003478, 0.00000370, 0.00001481, -0.00003851, 0.00004971, -0.00002385, -0.00002145, 0.00004629, -0.00005162, 0.00002495, 0.00000291, -0.00004276, 0.00005312, -0.00005822, 0.00000432, 0.00004172, -0.00007718, 0.00005474, -0.00005640, -0.00006407, 0.00005222, -0.00032295, 0.00000255, 0.00068423, 0.00022669, -0.00026705, -0.00009345, -0.00000439, -0.00002966, -0.00000904, -0.00000369, -0.00001669, -0.00001141, 0.00000571, -0.00000766, -0.00000065, 0.00000080, -0.00000785, 0.00000394, -0.00000894, -0.00000346, 0.00000676, -0.00000516, -0.00000175, -0.00000757, 0.00000060, 0.00000629, -0.00000783, 0.00000517, -0.00000066, -0.00000190, 0.00002128, -0.00000522, -0.00001783, 0.00000635, -0.00000157, -0.00000270, 0.00000670, -0.00000823, -0.00000998, 0.00000436, },
		{ 0.00000122, 0.00000625, -0.00000297, -0.00000762, -0.00000475, 0.00000040, -0.00000026, -0.00000601, -0.00000320, 0.00000173, 0.00000036, 0.00000160, -0.00000082, -0.00000222, 0.00000694, 0.00000865, -0.00000145, -0.00001022, -0.00000038, 0.00001531, 0.00000595, -0.00000470, -0.00000628, -0.00000189, 0.00001137, -0.00000198, -0.00000874, 0.00000731, -0.00000380, -0.00000591, 0.00000380, -0.00000295, -0.00000071, 0.00000254, 0.00000715, 0.00000541, -0.00000666, -0.00000205, -0.00000015, -0.00000054, 0.00000114, -0.00000320, 0.00001277, 0.00001153, -0.00001591, 0.00000587, 0.00001871, -0.00002822, -0.00001098, 0.00002129, -0.00001920, 0.00000080, 0.00001770, -0.00001747, 0.00001128, 0.00000258, -0.00002234, -0.00000376, -0.00000111, 0.00001855, -0.00001711, -0.00000783, 0.00004229, -0.00003299, -0.00000427, 0.00002196, -0.00003299, 0.00003284, -0.00000995, -0.00002320, 0.00003202, -0.00003908, 0.00002619, 0.00001375, -0.00004776, 0.00004620, -0.00002619, -0.00001105, 0.00003286, -0.00005201, 0.00003643, -0.00001812, -0.00004002, 0.00006094, -0.00005829, 0.00001149, 0.00000213, -0.00010009, 0.00005625, -0.00020707, -0.00018667, 0.00053464, 0.00043496, -0.00017032, -0.00017016, -0.00001390, -0.00001827, -0.00002055, -0.00000259, -0.00000996, -0.00001938, 0.00000516, -0.00000377, -0.00000613, 0.00000539, -0.00000898, 0.00000164, -0.00000326, -0.00000895, 0.00000699, -0.00000245, -0.00000327, -0.00000537, -0.00000401, 0.00000815, -0.00000515, 0.00000008, 0.00000343, -0.00000527, 0.00001748, 0.00000585, -0.00001965, -0.00000018, 0.00000267, -0.00000418, 0.00000592, -0.00000369, -0.00001255, },
		{ -0.00001268, -0.00000333, 0.00000765, 0.00000023, -0.00000783, -0.00000613, -0.00000084, 0.00000099, -0.00000421, -0.00000544, 0.00000058, 0.00000178, 0.00000145, -0.00000014, -0.00000282, 0.00000347, 0.00000875, 0.00000330, -0.00000859, -0.00000591, 0.00001333, 0.00000943, -0.00000348, -0.00000463, -0.00000521, 0.00000918, 0.00000446, -0.00001143, 0.00000460, 0.00000093, -0.00000854, 0.00000308, -0.00000146, -0.00000185, 0.00000241, 0.00000478, 0.00000808, -0.00000417, -0.00000499, 0.00000063, -0.00000111, 0.00000191, -0.00000306, 0.00000529, 0.00001705, -0.00000824, -0.00000839, 0.00002339, -0.00001108, -0.00002931, 0.00002060, -0.00000408, -0.00001616, 0.00002406, -0.00001074, -0.00000257, 0.00001706, -0.00002256, -0.00001009, -0.00000039, 0.00001246, -0.00000152, -0.00002273, 0.00003807, -0.00000818, -0.00002796, 0.00003098, -0.00002661, 0.00001125, 0.00001689, -0.00003842, 0.00002978, -0.00002025, -0.00000592, 0.00004063, -0.00004906, 0.00002221, 0.00000836, -0.00003726, 0.00003962, -0.00003582, 0.00000418, 0.00001826, -0.00005947, 0.00004675, -0.00001843, -0.00002904, 0.00003274, -0.00009365, 0.00001881, -0.00009612, -0.00026977, 0.00032138, 0.00054790, -0.00000595, -0.00021027, -0.00005504, -0.00000478, -0.00002380, -0.00001058, -0.00000076, -0.00002269, -0.00000141, 0.00000302, -0.00001112, 0.00000642, -0.00000560, -0.00000362, 0.00000213, -0.00001090, 0.00000357, 0.00000156, -0.00000472, -0.00000356, -0.00000667, 0.00000666, -0.00000047, -0.00000400, 0.00000468, -0.00000508, 0.00001116, 0.00001403, -0.00001583, -0.00000780, 0.00000418, -0.00000321, 0.00000385, 0.00000032, },
		{ 0.00000284, -0.00001070, -0.00000779, 0.00000670, 0.00000415, -0.00000687, -0.00000769, -0.00000219, 0.00000124, -0.00000201, -0.00000611, -0.00000183, 0.00000243, 0.00000235, 0.00000023, -0.00000288, 0.00000097, 0.00000689, 0.00000653, -0.00000436, -0.00001008, 0.00000906, 0.00001289, -0.00000237, -0.00000310, -0.00000600, 0.00000441, 0.00000999, -0.00001053, -0.00000061, 0.00000536, -0.00000916, 0.00000074, 0.00000061, -0.00000318, 0.00000261, 0.00000288, 0.00000842, 0.00000005, -0.00000716, -0.00000010, -0.00000065, 0.00000142, -0.00000114, -0.00000052, 0.00001575, 0.00000329, -0.00001613, 0.00001531, 0.00000984, -0.00003577, 0.00000518, 0.00001511, -0.00002494, 0.00001714, 0.00000536, -0.00001685, 0.00002242, -0.00001159, -0.00002102, 0.00000165, 0.00000602, 0.00000834, -0.00002426, 0.00002089, 0.00001717, -0.00003863, 0.00002286, -0.00000749, -0.00001254, 0.00003273, -0.00003510, 0.00001097, 0.00000575, -0.00002997, 0.00004535, -0.00002731, -0.00001095, 0.00003339, -0.00004162, 0.00002383, -0.00000713, -0.00002444, 0.00003488, -0.00005096, 0.00001308, 0.00001887, -0.00004746, 0.00003166, -0.00006126, -0.00002569, -0.00002327, -0.00026507, 0.00011049, 0.00054964, 0.00017685, -0.00018759, -0.00011281, -0.00000592, -0.00001473, -0.00002273, 0.00000373, -0.00001800, -0.00001224, 0.00000858, -0.00001193, 0.00000245, 0.00000074, -0.00000846, 0.00000420, -0.00000857, -0.00000189, 0.00000472, -0.00000464, -0.00000308, -0.00000713, 0.00000309, 0.00000382, -0.00000537, 0.00000322, -0.00000304, 0.00000503, 0.00001778, -0.00000822, -0.00001347, 0.00000222, -0.00000106, 0.00000201, },
		{ 0.00000130, 0.00000396, -0.00000767, -0.00001086, 0.00000356, 0.00000745, -0.00000420, -0.00000909, -0.00000383, 0.00000091, -0.00000028, -0.00000525, -0.00000419, 0.00000146, 0.00000373, 0.00000116, -0.00000304, -0.00000055, 0.00000450, 0.00000721, 0.00000110, -0.00001132, 0.00000272, 0.00001547, -0.00000045, -0.00000287, -0.00000429, -0.00000095, 0.00001252, -0.00000596, -0.00000660, 0.00000772, -0.00000721, -0.00000280, 0.00000263, -0.00000404, 0.00000224, 0.00000232, 0.00000673, 0.00000433, -0.00000717, -0.00000239, 0.00000018, 0.00000070, 0.00000057, -0.00000289, 0.00000990, 0.00001240, -0.00001387, -0.00000001, 0.00002372, -0.00002598, -0.00001785, 0.00002678, -0.00001904, -0.00000054, 0.00002184, -0.00002198, 0.00001462, 0.00000679, -0.00002947, -0.00000092, 0.00000412, 0.00000984, -0.00001581, 0.00000089, 0.00003207, -0.00003227, 0.00000288, 0.00001303, -0.00002661, 0.00003129, -0.00001640, -0.00001320, 0.00002459, -0.00003556, 0.00002873, 0.00000436, -0.00003564, 0.00003752, -0.00002510, -0.00000237, 0.00001723, -0.00003612, 0.00002894, -0.00002531, -0.00002025, 0.00003688, -0.00004050, 0.00001130, -0.00002564, -0.00005399, 0.00000603, -0.00021061, -0.00005138, 0.00046053, 0.00032669, -0.00010131, -0.00015941, -0.00003177, 0.00000056, -0.00002983, -0.00000139, -0.00000701, -0.00002203, 0.00000851, -0.00000721, -0.00000473, 0.00000641, -0.00000971, 0.00000196, -0.00000357, -0.00000666, 0.00000526, -0.00000267, -0.00000366, -0.00000628, -0.00000076, 0.00000603, -0.00000399, 0.00000057, -0.00000107, 0.00000067, 0.00001746, 0.00000040, -0.00001534, -0.00000221, 0.00000047, },
		{ 0.00000010, 0.00000150, 0.00000447, -0.00000465, -0.00001205, -0.00000072, 0.00000887, -0.00000013, -0.00000947, -0.00000598, 0.00000023, 0.00000080, -0.00000366, -0.00000540, -0.00000082, 0.00000432, 0.00000314, -0.00000305, -0.00000180, 0.00000278, 0.00000564, 0.00000566, -0.00000883, -0.00000431, 0.00001581, 0.00000308, -0.00000413, -0.00000146, -0.00000460, 0.00001112, 0.00000084, -0.00001105, 0.00000687, -0.00000311, -0.00000646, 0.00000367, -0.00000376, 0.00000081, 0.00000300, 0.00000428, 0.00000703, -0.00000474, -0.00000515, 0.00000021, 0.00000062, 0.00000119, -0.00000252, 0.00000352, 0.00001513, -0.00000392, -0.00001352, 0.00002370, -0.00000434, -0.00003622, 0.00002248, -0.00000035, -0.00001936, 0.00002842, -0.00001332, -0.00000266, 0.00002348, -0.00002804, -0.00001075, 0.00000663, 0.00000616, -0.00000501, -0.00001252, 0.00003205, -0.00001356, -0.00001809, 0.00002411, -0.00002540, 0.00001558, 0.00000713, -0.00002994, 0.00002740, -0.00002294, 0.00000186, 0.00002959, -0.00004066, 0.00002151, 0.00000072, -0.00002353, 0.00002591, -0.00002901, 0.00000935, 0.00000127, -0.00003849, 0.00003224, -0.00001836, -0.00001070, -0.00000269, -0.00006010, 0.00000513, -0.00014358, -0.00014686, 0.00032204, 0.00040967, 0.00002453, -0.00016813, -0.00007805, 0.00000849, -0.00002474, -0.00001482, 0.00000387, -0.00002506, 0.00000127, 0.00000081, -0.00001110, 0.00000795, -0.00000633, -0.00000294, 0.00000111, -0.00000848, 0.00000290, 0.00000024, -0.00000423, -0.00000531, -0.00000357, 0.00000582, -0.00000112, -0.00000156, -0.00000023, -0.00000179, 0.00001459, 0.00000769, -0.00001326, -0.00000707, },
		{ -0.00001041, -0.00000223, 0.00000158, 0.00000510, -0.00000205, -0.00001174, -0.00000483, 0.00000799, 0.00000420, -0.00000803, -0.00000840, -0.00000102, 0.00000146, -0.00000212, -0.00000522, -0.00000325, 0.00000318, 0.00000563, -0.00000200, -0.00000343, 0.00000194, 0.00000324, 0.00000756, -0.00000332, -0.00000983, 0.00001279, 0.00000806, -0.00000579, 0.00000046, -0.00000509, 0.00000662, 0.00000740, -0.00001213, 0.00000288, 0.00000175, -0.00000886, 0.00000299, -0.00000215, -0.00000141, 0.00000400, 0.00000251, 0.00000740, -0.00000085, -0.00000679, -0.00000125, 0.00000113, 0.00000124, -0.00000137, -0.00000033, 0.00001166, 0.00000740, -0.00001778, 0.00001066, 0.00001808, -0.00003927, 0.00000239, 0.00002111, -0.00002788, 0.00001989, 0.00000556, -0.00001962, 0.00002915, -0.00001418, -0.00002461, 0.00000877, 0.00000314, 0.00000137, -0.00001543, 0.00002052, 0.00000747, -0.00002976, 0.00002098, -0.00001184, -0.00000475, 0.00002374, -0.00003163, 0.00001472, -0.00000145, -0.00002083, 0.00003742, -0.00002606, -0.00000372, 0.00002119, -0.00002953, 0.00001798, -0.00001126, -0.00001048, 0.00001649, -0.00003791, 0.00001318, 0.00000405, -0.00002195, 0.00000406, -0.00005111, -0.00000768, -0.00008722, -0.00018355, 0.00017691, 0.00041871, 0.00015304, -0.00012686, -0.00012635, -0.00000381, -0.00000828, -0.00002895, 0.00000734, -0.00001905, -0.00001031, 0.00000739, -0.00001290, 0.00000426, -0.00000015, -0.00000734, 0.00000290, -0.00000687, -0.00000100, 0.00000248, -0.00000382, -0.00000489, -0.00000502, 0.00000398, 0.00000164, -0.00000219, -0.00000039, -0.00000307, 0.00001072, 0.00001247, -0.00000847, },
		{ -0.00000269, -0.00001123, -0.00000546, 0.00000049, 0.00000593, 0.00000041, -0.00001065, -0.00000796, 0.00000532, 0.00000740, -0.00000468, -0.00001018, -0.00000322, 0.00000187, -0.00000088, -0.00000429, -0.00000463, 0.00000043, 0.00000724, 0.00000081, -0.00000518, 0.00000117, 0.00000162, 0.00000630, 0.00000309, -0.00001170, 0.00000654, 0.00001308, -0.00000608, -0.00000008, -0.00000261, 0.00000120, 0.00001135, -0.00000928, -0.00000283, 0.00000555, -0.00000898, 0.00000051, 0.00000028, -0.00000354, 0.00000421, 0.00000222, 0.00000598, 0.00000280, -0.00000626, -0.00000374, 0.00000119, 0.00000173, -0.00000089, -0.00000112, 0.00000556, 0.00001381, -0.00001113, -0.00000743, 0.00002962, -0.00002413, -0.00002408, 0.00003244, -0.00001977, -0.00000019, 0.00002441, -0.00002591, 0.00001995, 0.00000748, -0.00003482, 0.00000491, 0.00000474, 0.00000130, -0.00001046, 0.00000573, 0.00002122, -0.00002768, 0.00000681, 0.00000520, -0.00001913, 0.00002658, -0.00001914, -0.00000504, 0.00001655, -0.00002924, 0.00002693, -0.00000161, -0.00002412, 0.00002672, -0.00001998, 0.00000138, 0.00000542, -0.00002064, 0.00001710, -0.00002466, -0.00000759, 0.00001604, -0.00002013, 0.00000071, -0.00003795, -0.00001973, -0.00004862, -0.00018125, 0.00005429, 0.00036987, 0.00025086, -0.00004391, -0.00015256, -0.00003968, 0.00000969, -0.00003430, -0.00000007, -0.00000706, -0.00002040, 0.00000821, -0.00000906, -0.00000264, 0.00000533, -0.00000848, 0.00000118, -0.00000325, -0.00000433, 0.00000284, -0.00000230, -0.00000485, -0.00000558, 0.00000164, 0.00000324, -0.00000144, -0.00000074, -0.00000393, 0.00000675, 0.00001472, },
		{ 0.00001501, 0.00000275, -0.00000974, -0.00000813, -0.00000206, 0.00000625, 0.00000316, -0.00000915, -0.00001005, 0.00000196, 0.00000859, -0.00000029, -0.00001023, -0.00000628, 0.00000167, 0.00000034, -0.00000341, -0.00000456, -0.00000271, 0.00000661, 0.00000500, -0.00000587, -0.00000060, 0.00000152, 0.00000305, 0.00000772, -0.00000903, -0.00000125, 0.00001601, -0.00000358, -0.00000314, 0.00000106, -0.00000258, 0.00001148, -0.00000359, -0.00000808, 0.00000680, -0.00000669, -0.00000300, 0.00000249, -0.00000458, 0.00000304, 0.00000315, 0.00000408, 0.00000486, -0.00000380, -0.00000583, -0.00000023, 0.00000300, -0.00000100, -0.00000060, 0.00000106, 0.00001269, 0.00000176, -0.00002009, 0.00002444, 0.00000223, -0.00004290, 0.00002530, 0.00000188, -0.00002087, 0.00003212, -0.00001682, 0.00000028, 0.00002700, -0.00003382, -0.00000707, 0.00001011, -0.00000210, -0.00000397, -0.00000443, 0.00002335, -0.00001481, -0.00000995, 0.00001634, -0.00002161, 0.00001653, -0.00000030, -0.00002100, 0.00002242, -0.00002201, 0.00000626, 0.00001953, -0.00003018, 0.00001700, -0.00000278, -0.00001302, 0.00001277, -0.00001896, 0.00000805, -0.00000890, -0.00002051, 0.00001539, -0.00001086, -0.00000405, -0.00002769, -0.00002720, -0.00002467, -0.00015918, -0.00003489, 0.00029001, 0.00030076, 0.00005625, -0.00013908, -0.00008882, 0.00001501, -0.00002573, -0.00001512, 0.00000386, -0.00002358, 0.00000212, -0.00000194, -0.00000891, 0.00000706, -0.00000572, -0.00000255, 0.00000001, -0.00000552, 0.00000131, -0.00000041, -0.00000458, -0.00000587, -0.00000038, 0.00000354, -0.00000017, -0.00000053, -0.00000453, 0.00000288, },
		{ -0.00000091, 0.00001378, 0.00000725, -0.00000681, -0.00000923, -0.00000542, 0.00000515, 0.00000620, -0.00000702, -0.00001149, -0.00000122, 0.00000791, 0.00000369, -0.00000802, -0.00000936, 0.00000021, 0.00000180, -0.00000283, -0.00000366, -0.00000470, 0.00000344, 0.00000897, -0.00000413, -0.00000361, 0.00000229, -0.00000010, 0.00000871, -0.00000287, -0.00000782, 0.00001510, 0.00000173, -0.00000720, 0.00000343, -0.00000314, 0.00000839, 0.00000273, -0.00001092, 0.00000509, -0.00000283, -0.00000632, 0.00000346, -0.00000405, 0.00000076, 0.00000431, 0.00000294, 0.00000507, -0.00000074, -0.00000608, -0.00000318, 0.00000409, -0.00000043, -0.00000090, 0.00000020, 0.00000628, 0.00001299, -0.00002036, 0.00000589, 0.00002645, -0.00004313, 0.00000123, 0.00002548, -0.00002990, 0.00002338, 0.00000357, -0.00001927, 0.00003457, -0.00001917, -0.00002343, 0.00001425, -0.00000353, -0.00000152, -0.00000663, 0.00001619, 0.00000104, -0.00002065, 0.00001667, -0.00001323, 0.00000067, 0.00001501, -0.00002571, 0.00001506, -0.00000576, -0.00001295, 0.00002790, -0.00002109, -0.00000032, 0.00001159, -0.00001791, 0.00000963, -0.00001015, -0.00000251, 0.00000153, -0.00002297, 0.00000687, -0.00000199, -0.00000503, -0.00002108, -0.00003194, -0.00001015, -0.00012911, -0.00009325, 0.00020335, 0.00030478, 0.00014490, -0.00008394, -0.00013069, -0.00000251, -0.00000676, -0.00002928, 0.00000687, -0.00001822, -0.00000793, 0.00000405, -0.00001124, 0.00000428, -0.00000075, -0.00000571, 0.00000110, -0.00000440, -0.00000105, 0.00000081, -0.00000364, -0.00000615, -0.00000195, 0.00000305, 0.00000078, 0.00000035, -0.00000451, },
		{ -0.00000345, -0.00000437, 0.00001119, 0.00001069, -0.00000334, -0.00000862, -0.00000836, 0.00000220, 0.00000877, -0.00000380, -0.00001252, -0.00000404, 0.00000624, 0.00000609, -0.00000409, -0.00001105, -0.00000287, 0.00000318, -0.00000210, -0.00000301, -0.00000468, -0.00000109, 0.00001062, 0.00000046, -0.00000677, 0.00000237, -0.00000134, 0.00000601, 0.00000412, -0.00001059, 0.00001014, 0.00000822, -0.00000987, 0.00000272, -0.00000065, 0.00000409, 0.00000741, -0.00001048, 0.00000121, 0.00000113, -0.00000828, 0.00000273, -0.00000224, -0.00000175, 0.00000469, 0.00000297, 0.00000429, 0.00000136, -0.00000415, -0.00000629, 0.00000339, 0.00000186, -0.00000264, 0.00000183, -0.00000029, 0.00001628, -0.00000862, -0.00001544, 0.00003597, -0.00002327, -0.00002837, 0.00003711, -0.00002092, 0.00000225, 0.00002434, -0.00002775, 0.00002614, 0.00000423, -0.00003619, 0.00001150, 0.00000036, -0.00000445, -0.00000314, 0.00000627, 0.00001203, -0.00002105, 0.00000768, -0.00000040, -0.00001230, 0.00002020, -0.00001847, 0.00000030, 0.00000961, -0.00002188, 0.00002186, -0.00000375, -0.00001496, 0.00001612, -0.00001276, 0.00000089, -0.00000134, -0.00000850, 0.00000442, -0.00001834, -0.00000256, 0.00000176, -0.00000254, -0.00001509, -0.00003621, -0.00000273, -0.00009567, -0.00012814, 0.00012348, 0.00027821, 0.00020167, -0.00000220, -0.00014396, -0.00004202, 0.00001120, -0.00003364, -0.00000042, -0.00000773, -0.00001661, 0.00000522, -0.00000883, -0.00000120, 0.00000354, -0.00000634, -0.00000022, -0.00000215, -0.00000281, 0.00000076, -0.00000218, -0.00000622, -0.00000330, 0.00000235, 0.00000107, 0.00000139, },
		{ 0.00000189, -0.00000134, -0.00000686, 0.00000727, 0.00001300, 0.00000020, -0.00000687, -0.00000983, -0.00000207, 0.00000973, 0.00000063, -0.00001277, -0.00000688, 0.00000446, 0.00000667, 0.00000009, -0.00001023, -0.00000701, 0.00000347, -0.00000051, -0.00000322, -0.00000315, -0.00000489, 0.00000863, 0.00000666, -0.00000806, 0.00000043, -0.00000032, 0.00000140, 0.00000886, -0.00000849, 0.00000281, 0.00001335, -0.00000922, -0.00000123, 0.00000304, 0.00000096, 0.00000923, -0.00000731, -0.00000326, 0.00000388, -0.00000838, 0.00000055, 0.00000004, -0.00000354, 0.00000388, 0.00000367, 0.00000372, 0.00000183, -0.00000117, -0.00000756, -0.00000003, 0.00000537, -0.00000438, 0.00000286, -0.00000258, 0.00001060, 0.00000792, -0.00002763, 0.00002596, 0.00000741, -0.00004802, 0.00002844, 0.00000227, -0.00001969, 0.00003387, -0.00002012, 0.00000607, 0.00002626, -0.00003754, -0.00000023, 0.00000842, -0.00000954, 0.00000062, -0.00000006, 0.00001440, -0.00001284, -0.00000447, 0.00000936, -0.00001675, 0.00001471, -0.00000466, -0.00001324, 0.00001645, -0.00001844, 0.00000710, 0.00001196, -0.00001980, 0.00001044, -0.00000272, -0.00000669, 0.00000262, -0.00000869, 0.00000241, -0.00001207, -0.00000829, 0.00000055, -0.00000017, -0.00000723, -0.00003934, -0.00000301, -0.00006087, -0.00014462, 0.00005336, 0.00023878, 0.00022189, 0.00008112, -0.00011719, -0.00009006, 0.00001460, -0.00002461, -0.00001414, 0.00000142, -0.00001938, 0.00000081, -0.00000377, -0.00000621, 0.00000488, -0.00000418, -0.00000265, -0.00000041, -0.00000310, -0.00000033, -0.00000080, -0.00000572, -0.00000466, 0.00000170, 0.00000095, },
		{ 0.00000092, 0.00000149, 0.00000120, -0.00000765, 0.00000236, 0.00001386, 0.00000370, -0.00000467, -0.00000952, -0.00000644, 0.00000826, 0.00000555, -0.00001142, -0.00001006, 0.00000281, 0.00000625, 0.00000296, -0.00000689, -0.00001062, 0.00000160, 0.00000203, -0.00000387, -0.00000157, -0.00000617, 0.00000344, 0.00001188, -0.00000576, -0.00000345, 0.00000167, -0.00000252, 0.00000946, -0.00000265, -0.00000398, 0.00001491, -0.00000489, -0.00000677, 0.00000542, 0.00000044, 0.00000849, -0.00000292, -0.00000680, 0.00000474, -0.00000683, -0.00000232, 0.00000191, -0.00000416, 0.00000224, 0.00000414, 0.00000405, 0.00000126, 0.00000093, -0.00000577, -0.00000526, 0.00000790, -0.00000374, 0.00000086, 0.00000030, 0.00000046, 0.00001959, -0.00002412, 0.00000205, 0.00003354, -0.00004656, 0.00000195, 0.00002723, -0.00002973, 0.00002658, -0.00000024, -0.00001503, 0.00003681, -0.00002473, -0.00001739, 0.00001555, -0.00001172, 0.00000002, -0.00000025, 0.00000984, -0.00000184, -0.00001302, 0.00001153, -0.00001241, 0.00000331, 0.00000802, -0.00001900, 0.00001292, -0.00000707, -0.00000757, 0.00001876, -0.00001409, -0.00000020, 0.00000543, -0.00000882, 0.00000131, -0.00000577, -0.00000036, -0.00000780, -0.00000987, -0.00000227, -0.00000099, 0.00000174, -0.00003818, -0.00001142, -0.00002822, -0.00014381, -0.00000898, 0.00019768, 0.00021519, 0.00014266, -0.00005489, -0.00012579, -0.00000444, -0.00000694, -0.00002620, 0.00000355, -0.00001532, -0.00000656, 0.00000045, -0.00000830, 0.00000291, -0.00000070, -0.00000439, -0.00000019, -0.00000211, -0.00000161, -0.00000014, -0.00000456, -0.00000595, 0.00000094, },
		{ -0.00000021, 0.00000129, 0.00000040, 0.00000323, -0.00000640, -0.00000267, 0.00001284, 0.00000709, -0.00000240, -0.00000788, -0.00000959, 0.00000447, 0.00000954, -0.00000786, -0.00001326, 0.00000071, 0.00000596, 0.00000382, -0.00000238, -0.00001185, -0.00000262, 0.00000458, -0.00000373, -0.00000137, -0.00000459, -0.00000272, 0.00001344, 0.00000023, -0.00000750, 0.00000251, -0.00000384, 0.00000607, 0.00000412, -0.00000747, 0.00001231, 0.00000167, -0.00001133, 0.00000469, 0.00000235, 0.00000667, 0.00000111, -0.00000863, 0.00000389, -0.00000430, -0.00000511, 0.00000287, -0.00000371, 0.00000044, 0.00000383, 0.00000491, 0.00000110, 0.00000079, -0.00000151, -0.00000973, 0.00000691, 0.00000048, -0.00000385, 0.00000523, -0.00000722, 0.00002000, -0.00000723, -0.00002274, 0.00004178, -0.00002348, -0.00002972, 0.00003951, -0.00002146, 0.00000638, 0.00002116, -0.00002610, 0.00003139, -0.00000207, -0.00003225, 0.00001601, -0.00000774, -0.00000580, 0.00000358, 0.00000347, 0.00000584, -0.00001431, 0.00000649, -0.00000355, -0.00000731, 0.00001378, -0.00001549, 0.00000276, 0.00000478, -0.00001518, 0.00001517, -0.00000270, -0.00000890, 0.00000751, -0.00000563, -0.00000217, -0.00000338, -0.00000134, -0.00000591, -0.00000960, -0.00000354, -0.00000496, 0.00000835, -0.00003015, -0.00002527, -0.00000360, -0.00012563, -0.00006397, 0.00015661, 0.00019763, 0.00017070, 0.00002413, -0.00013072, -0.00004258, 0.00000818, -0.00002919, -0.00000270, -0.00000757, -0.00001265, 0.00000115, -0.00000708, -0.00000084, 0.00000201, -0.00000430, -0.00000134, -0.00000077, -0.00000222, -0.00000039, -0.00000303, -0.00000683, },
		{ -0.00000699, -0.00000187, 0.00000197, -0.00000075, 0.00000398, -0.00000348, -0.00000667, 0.00000981, 0.00000997, -0.00000005, -0.00000566, -0.00001078, -0.00000056, 0.00001118, -0.00000235, -0.00001533, -0.00000266, 0.00000634, 0.00000332, 0.00000124, -0.00000973, -0.00000794, 0.00000539, -0.00000174, -0.00000286, -0.00000154, -0.00000701, 0.00001019, 0.00000790, -0.00000910, 0.00000076, -0.00000244, 0.00000081, 0.00000869, -0.00000648, 0.00000687, 0.00000794, -0.00001277, 0.00000065, 0.00000510, 0.00000544, 0.00000399, -0.00000881, 0.00000191, -0.00000144, -0.00000733, 0.00000278, -0.00000259, -0.00000102, 0.00000292, 0.00000530, 0.00000246, -0.00000137, 0.00000280, -0.00001055, 0.00000145, 0.00000688, -0.00000818, 0.00000748, -0.00000756, 0.00000964, 0.00001334, -0.00003512, 0.00002807, 0.00001028, -0.00005023, 0.00003083, 0.00000126, -0.00001539, 0.00003256, -0.00002164, 0.00001353, 0.00002095, -0.00003706, 0.00000728, 0.00000147, -0.00001366, 0.00000664, 0.00000033, 0.00000725, -0.00000924, -0.00000158, 0.00000418, -0.00001226, 0.00001132, -0.00000611, -0.00000761, 0.00001100, -0.00001372, 0.00000514, 0.00000745, -0.00001114, 0.00000378, -0.00000059, -0.00000422, -0.00000335, -0.00000079, -0.00000465, -0.00000987, -0.00000257, -0.00000925, 0.00000917, -0.00001600, -0.00003853, 0.00000745, -0.00009298, -0.00010740, 0.00011121, 0.00018190, 0.00016870, 0.00009531, -0.00009771, -0.00008551, 0.00000904, -0.00002105, -0.00001347, -0.00000137, -0.00001429, -0.00000189, -0.00000423, -0.00000412, 0.00000259, -0.00000256, -0.00000281, -0.00000008, -0.00000186, -0.00000118, -0.00000166, },
		{ -0.00000084, -0.00000635, -0.00000389, 0.00000256, -0.00000128, 0.00000321, 0.00000012, -0.00000866, 0.00000521, 0.00001167, 0.00000252, -0.00000341, -0.00001016, -0.00000530, 0.00000984, 0.00000376, -0.00001491, -0.00000751, 0.00000681, 0.00000300, 0.00000243, -0.00000496, -0.00001201, 0.00000308, 0.00000195, -0.00000485, 0.00000070, -0.00000752, 0.00000332, 0.00001389, -0.00000638, -0.00000326, 0.00000003, -0.00000360, 0.00000922, -0.00000198, 0.00000118, 0.00001167, -0.00001043, -0.00000523, 0.00000661, 0.00000566, 0.00000590, -0.00000790, -0.00000067, 0.00000136, -0.00000872, 0.00000170, -0.00000109, -0.00000214, 0.00000210, 0.00000440, 0.00000519, -0.00000367, 0.00000435, -0.00000653, -0.00000664, 0.00001177, -0.00000833, 0.00000411, -0.00000078, -0.00000472, 0.00002617, -0.00002874, -0.00000018, 0.00003793, -0.00004833, 0.00000406, 0.00002602, -0.00002632, 0.00002817, -0.00000465, -0.00000714, 0.00003457, -0.00002844, -0.00000802, 0.00001123, -0.00001856, 0.00000482, 0.00000235, 0.00000366, -0.00000201, -0.00000770, 0.00000697, -0.00001057, 0.00000360, 0.00000345, -0.00001294, 0.00000964, -0.00000612, -0.00000500, 0.00001145, -0.00000690, -0.00000204, 0.00000244, -0.00000323, -0.00000488, -0.00000060, -0.00000237, -0.00001113, -0.00000115, -0.00001055, 0.00000360, -0.00000026, -0.00004453, 0.00000336, -0.00005363, -0.00013231, 0.00005775, 0.00017132, 0.00015121, 0.00013927, -0.00003480, -0.00011398, -0.00000977, -0.00000690, -0.00002204, -0.00000074, -0.00001128, -0.00000651, -0.00000219, -0.00000546, 0.00000111, -0.00000036, -0.00000350, -0.00000041, -0.00000086, -0.00000192, },
		{ -0.00000213, -0.00000066, -0.00000511, -0.00000589, 0.00000256, -0.00000083, 0.00000135, 0.00000322, -0.00000825, 0.00000008, 0.00001161, 0.00000519, -0.00000124, -0.00000840, -0.00000854, 0.00000608, 0.00000857, -0.00001130, -0.00001283, 0.00000591, 0.00000425, 0.00000114, 0.00000021, -0.00001259, -0.00000228, 0.00000567, -0.00000538, 0.00000045, -0.00000449, -0.00000408, 0.00001522, 0.00000042, -0.00000736, 0.00000126, -0.00000529, 0.00000598, 0.00000349, -0.00000229, 0.00001198, -0.00000531, -0.00001067, 0.00000548, 0.00000694, 0.00000762, -0.00000624, -0.00000369, 0.00000388, -0.00000903, -0.00000040, 0.00000071, -0.00000315, 0.00000186, 0.00000238, 0.00000778, -0.00000373, 0.00000176, 0.00000072, -0.00001329, 0.00001137, -0.00000264, -0.00000376, 0.00000805, -0.00001427, 0.00002459, -0.00000749, -0.00002805, 0.00004567, -0.00002416, -0.00002782, 0.00003867, -0.00002007, 0.00001112, 0.00001533, -0.00002019, 0.00003379, -0.00000941, -0.00002324, 0.00001597, -0.00001702, -0.00000270, 0.00000755, -0.00000084, 0.00000282, -0.00000892, 0.00000454, -0.00000482, -0.00000452, 0.00000855, -0.00001154, 0.00000306, 0.00000227, -0.00001026, 0.00000853, 0.00000021, -0.00000568, 0.00000175, -0.00000033, -0.00000573, -0.00000221, 0.00000085, -0.00001191, -0.00000166, -0.00000762, -0.00000540, 0.00001105, -0.00003960, -0.00001172, -0.00001836, -0.00013364, -0.00000171, 0.00016003, 0.00013496, 0.00014962, 0.00003828, -0.00011280, -0.00004349, 0.00000346, -0.00002307, -0.00000595, -0.00000636, -0.00000977, -0.00000244, -0.00000491, -0.00000120, 0.00000100, -0.00000297, -0.00000141, 0.00000009, },
		{ 0.00000049, -0.00000176, -0.00000092, -0.00000356, -0.00000745, 0.00000162, 0.00000053, -0.00000082, 0.00000491, -0.00000582, -0.00000437, 0.00000963, 0.00000745, 0.00000101, -0.00000618, -0.00000992, 0.00000134, 0.00001055, -0.00000520, -0.00001672, 0.00000236, 0.00000715, -0.00000095, 0.00000313, -0.00000898, -0.00000872, 0.00000704, -0.00000312, -0.00000221, -0.00000012, -0.00000862, 0.00001099, 0.00000871, -0.00000871, -0.00000011, -0.00000430, 0.00000099, 0.00000720, -0.00000247, 0.00000969, 0.00000061, -0.00001377, 0.00000170, 0.00000803, 0.00000968, -0.00000367, -0.00000712, 0.00000565, -0.00000778, -0.00000361, 0.00000267, -0.00000406, 0.00000197, 0.00000044, 0.00000832, -0.00000035, -0.00000376, 0.00000728, -0.00001445, 0.00000440, 0.00000674, -0.00001137, 0.00001240, -0.00001350, 0.00001024, 0.00001688, -0.00004117, 0.00002994, 0.00001056, -0.00004853, 0.00003123, -0.00000004, -0.00000845, 0.00002775, -0.00001991, 0.00002074, 0.00001219, -0.00003125, 0.00001249, -0.00000911, -0.00001309, 0.00001160, -0.00000219, 0.00000296, -0.00000567, -0.00000048, 0.00000096, -0.00000908, 0.00000766, -0.00000550, -0.00000422, 0.00000698, -0.00000925, 0.00000168, 0.00000563, -0.00000501, -0.00000146, 0.00000189, -0.00000431, -0.00000527, 0.00000331, -0.00001039, -0.00000500, -0.00000215, -0.00001299, 0.00001379, -0.00002527, -0.00002935, 0.00000344, -0.00011254, -0.00005845, 0.00013853, 0.00013009, 0.00013417, 0.00009890, -0.00007814, -0.00007852, 0.00000101, -0.00001610, -0.00001344, -0.00000329, -0.00000993, -0.00000467, -0.00000380, -0.00000296, 0.00000092, -0.00000165, -0.00000233, },
		{ -0.00000263, 0.00000032, -0.00000104, -0.00000135, -0.00000199, -0.00000823, -0.00000035, 0.00000235, -0.00000243, 0.00000484, -0.00000229, -0.00000718, 0.00000625, 0.00000863, 0.00000336, -0.00000376, -0.00000986, -0.00000280, 0.00000941, 0.00000147, -0.00001729, -0.00000385, 0.00001016, -0.00000148, 0.00000236, -0.00000267, -0.00001315, 0.00000440, 0.00000148, -0.00000540, 0.00000268, -0.00000847, 0.00000304, 0.00001474, -0.00000557, -0.00000351, -0.00000230, -0.00000328, 0.00000765, -0.00000007, 0.00000659, 0.00000551, -0.00001382, -0.00000355, 0.00000767, 0.00001181, 0.00000032, -0.00001055, 0.00000580, -0.00000450, -0.00000765, 0.00000423, -0.00000431, 0.00000153, -0.00000002, 0.00000588, 0.00000544, -0.00000881, 0.00000912, -0.00000865, -0.00000654, 0.00001466, -0.00001331, 0.00000853, -0.00000339, -0.00000819, 0.00003144, -0.00003329, -0.00000074, 0.00003861, -0.00004719, 0.00000652, 0.00002240, -0.00001934, 0.00002695, -0.00000796, 0.00000309, 0.00002771, -0.00002815, 0.00000192, 0.00000158, -0.00002166, 0.00001063, 0.00000120, -0.00000060, -0.00000088, -0.00000457, 0.00000375, -0.00000882, 0.00000248, 0.00000121, -0.00000831, 0.00000646, -0.00000408, -0.00000471, 0.00000655, -0.00000094, -0.00000429, 0.00000152, -0.00000096, -0.00000788, 0.00000319, -0.00000611, -0.00000976, 0.00000243, -0.00001534, 0.00000797, -0.00000744, -0.00004079, 0.00000791, -0.00007720, -0.00010091, 0.00010071, 0.00013607, 0.00010940, 0.00013123, -0.00001953, -0.00009852, -0.00001706, -0.00000619, -0.00001822, -0.00000431, -0.00000752, -0.00000717, -0.00000354, -0.00000355, -0.00000021, -0.00000039, },
		{ 0.00000015, -0.00000231, -0.00000015, -0.00000024, -0.00000181, -0.00000065, -0.00000797, -0.00000317, 0.00000395, -0.00000284, 0.00000322, 0.00000126, -0.00000800, 0.00000236, 0.00000838, 0.00000542, -0.00000104, -0.00000900, -0.00000553, 0.00000616, 0.00000644, -0.00001388, -0.00001110, 0.00001095, 0.00000115, -0.00000137, 0.00000317, -0.00001306, -0.00000185, 0.00000615, -0.00000657, 0.00000213, -0.00000435, -0.00000494, 0.00001555, 0.00000143, -0.00000667, -0.00000135, -0.00000532, 0.00000516, 0.00000300, 0.00000434, 0.00000861, -0.00001130, -0.00000867, 0.00000537, 0.00001305, 0.00000589, -0.00001283, 0.00000330, 0.00000069, -0.00001146, 0.00000429, -0.00000308, -0.00000029, 0.00000156, 0.00000140, 0.00001073, -0.00000974, 0.00000453, 0.00000167, -0.00001595, 0.00001590, -0.00000717, -0.00000196, 0.00000933, -0.00002019, 0.00002911, -0.00000930, -0.00003048, 0.00004626, -0.00002423, -0.00002318, 0.00003438, -0.00001569, 0.00001495, 0.00000820, -0.00001033, 0.00003204, -0.00001519, -0.00001104, 0.00001013, -0.00002451, 0.00000342, 0.00000769, -0.00000464, 0.00000210, -0.00000548, 0.00000289, -0.00000514, -0.00000357, 0.00000506, -0.00000778, 0.00000226, 0.00000154, -0.00000747, 0.00000320, 0.00000354, -0.00000435, -0.00000130, 0.00000231, -0.00000806, 0.00000012, -0.00000060, -0.00001319, 0.00000325, -0.00001171, -0.00000240, 0.00000672, -0.00004105, -0.00000179, -0.00003985, -0.00012034, 0.00004833, 0.00014356, 0.00009166, 0.00013218, 0.00004449, -0.00009211, -0.00004510, -0.00000120, -0.00001720, -0.00000869, -0.00000483, -0.00000820, -0.00000463, -0.00000338, -0.00000152, },
		{ -0.00000235, -0.00000006, -0.00000177, -0.00000063, 0.00000055, -0.00000221, 0.00000020, -0.00000656, -0.00000636, 0.00000460, -0.00000179, 0.00000070, 0.00000395, -0.00000705, -0.00000117, 0.00000690, 0.00000653, 0.00000205, -0.00000763, -0.00000695, 0.00000253, 0.00000828, -0.00000761, -0.00001671, 0.00000776, 0.00000651, -0.00000538, 0.00000541, -0.00000805, -0.00000907, 0.00000796, -0.00000426, -0.00000134, 0.00000087, -0.00000936, 0.00001065, 0.00000937, -0.00000696, -0.00000242, -0.00000533, 0.00000138, 0.00000489, 0.00000353, 0.00001016, -0.00000728, -0.00001249, 0.00000164, 0.00001237, 0.00001223, -0.00001229, -0.00000226, 0.00000654, -0.00001328, 0.00000173, 0.00000003, -0.00000338, 0.00000416, -0.00000265, 0.00001234, -0.00000490, -0.00000447, 0.00001114, -0.00001841, 0.00000850, 0.00000458, -0.00001296, 0.00001646, -0.00001947, 0.00001221, 0.00001780, -0.00004445, 0.00003036, 0.00000878, -0.00004273, 0.00002876, -0.00000021, -0.00000016, 0.00001998, -0.00001411, 0.00002565, 0.00000221, -0.00002059, 0.00001288, -0.00002047, -0.00000808, 0.00001355, -0.00000578, 0.00000146, -0.00000325, -0.00000012, -0.00000078, -0.00000744, 0.00000466, -0.00000390, -0.00000253, 0.00000459, -0.00000602, -0.00000190, 0.00000551, -0.00000129, -0.00000454, 0.00000332, -0.00000528, -0.00000439, 0.00000355, -0.00001309, -0.00000021, -0.00000465, -0.00001171, 0.00001240, -0.00003082, -0.00001756, -0.00001168, -0.00011513, -0.00000919, 0.00014063, 0.00008965, 0.00011105, 0.00009454, -0.00005826, -0.00007109, -0.00000710, -0.00001139, -0.00001335, -0.00000416, -0.00000722, -0.00000636, -0.00000336, },
		{ -0.00000408, -0.00000269, -0.00000062, -0.00000136, -0.00000112, 0.00000136, -0.00000236, 0.00000023, -0.00000411, -0.00000917, 0.00000365, 0.00000055, -0.00000189, 0.00000523, -0.00000482, -0.00000382, 0.00000485, 0.00000622, 0.00000509, -0.00000552, -0.00000775, -0.00000013, 0.00000713, -0.00000090, -0.00001837, 0.00000072, 0.00001221, -0.00000650, 0.00000276, -0.00000041, -0.00001362, 0.00000516, 0.00000080, -0.00000530, 0.00000396, -0.00000863, 0.00000238, 0.00001451, -0.00000316, -0.00000467, -0.00000472, -0.00000193, 0.00000480, 0.00000360, 0.00001091, -0.00000263, -0.00001462, -0.00000252, 0.00000946, 0.00001757, -0.00000761, -0.00000987, 0.00001069, -0.00001135, -0.00000372, 0.00000419, -0.00000635, 0.00000570, -0.00000371, 0.00000888, 0.00000401, -0.00001299, 0.00001428, -0.00001184, -0.00000459, 0.00001561, -0.00001752, 0.00001325, -0.00000731, -0.00000935, 0.00003414, -0.00003650, -0.00000029, 0.00003535, -0.00004230, 0.00000804, 0.00001768, -0.00000938, 0.00002242, -0.00000851, 0.00001355, 0.00001748, -0.00002281, 0.00000918, -0.00001141, -0.00001985, 0.00001500, -0.00000246, -0.00000220, 0.00000018, -0.00000292, 0.00000192, -0.00000782, 0.00000098, 0.00000058, -0.00000522, 0.00000414, -0.00000210, -0.00000571, 0.00000387, 0.00000309, -0.00000575, 0.00000130, -0.00000096, -0.00000783, 0.00000437, -0.00000926, -0.00000595, 0.00000179, -0.00001573, 0.00000923, -0.00001541, -0.00003054, 0.00000147, -0.00009143, -0.00005912, 0.00011951, 0.00010154, 0.00008377, 0.00011840, -0.00000674, -0.00008232, -0.00002423, -0.00000552, -0.00001503, -0.00000626, -0.00000526, -0.00000752, },
		{ -0.00000733, -0.00000535, -0.00000301, -0.00000114, -0.00000112, -0.00000178, 0.00000208, -0.00000189, -0.00000073, -0.00000113, -0.00001069, 0.00000082, 0.00000352, -0.00000364, 0.00000487, -0.00000189, -0.00000551, 0.00000292, 0.00000467, 0.00000717, -0.00000229, -0.00000842, -0.00000153, 0.00000450, 0.00000372, -0.00001547, -0.00000797, 0.00001505, -0.00000296, -0.00000332, 0.00000594, -0.00001285, -0.00000150, 0.00000588, -0.00000690, 0.00000322, -0.00000397, -0.00000532, 0.00001431, 0.00000360, -0.00000596, -0.00000489, -0.00000396, 0.00000324, 0.00000351, 0.00001125, 0.00000236, -0.00001520, -0.00000635, 0.00000512, 0.00001988, 0.00000114, -0.00001690, 0.00001053, -0.00000506, -0.00001068, 0.00000733, -0.00000714, 0.00000430, -0.00000082, 0.00000181, 0.00001265, -0.00001588, 0.00000874, 0.00000093, -0.00001676, 0.00001933, -0.00001220, 0.00000133, 0.00000852, -0.00002384, 0.00003229, -0.00001189, -0.00002991, 0.00004266, -0.00002252, -0.00001704, 0.00002738, -0.00000799, 0.00001651, 0.00000172, 0.00000205, 0.00002602, -0.00001717, 0.00000124, -0.00000094, -0.00002781, 0.00001013, 0.00000434, -0.00000647, 0.00000237, -0.00000374, 0.00000205, -0.00000534, -0.00000367, 0.00000316, -0.00000483, 0.00000133, 0.00000167, -0.00000644, -0.00000030, 0.00000613, -0.00000371, -0.00000249, 0.00000244, -0.00000831, 0.00000177, -0.00000361, -0.00001080, 0.00000444, -0.00001354, 0.00000078, -0.00000154, -0.00003507, 0.00000052, -0.00006004, -0.00009109, 0.00008045, 0.00011759, 0.00006503, 0.00011511, 0.00004638, -0.00007132, -0.00004643, -0.00000531, -0.00001265, -0.00000974, -0.00000403, },
		{ -0.00000462, -0.00000612, -0.00000643, -0.00000368, -0.00000162, -0.00000079, -0.00000263, 0.00000229, -0.00000045, -0.00000244, 0.00000139, -0.00001009, -0.00000357, 0.00000605, -0.00000378, 0.00000300, 0.00000118, -0.00000630, 0.00000141, 0.00000271, 0.00000736, 0.00000197, -0.00000869, -0.00000240, 0.00000231, 0.00000501, -0.00000948, -0.00001505, 0.00001285, 0.00000442, -0.00000923, 0.00000753, -0.00000679, -0.00000890, 0.00000781, -0.00000470, -0.00000055, 0.00000141, -0.00000905, 0.00000891, 0.00001028, -0.00000431, -0.00000605, -0.00000519, 0.00000132, 0.00000263, 0.00001086, 0.00000768, -0.00001426, -0.00000976, 0.00000077, 0.00001803, 0.00001196, -0.00001994, 0.00000467, 0.00000406, -0.00001626, 0.00000701, -0.00000438, -0.00000027, 0.00000450, -0.00000511, 0.00001639, -0.00001071, -0.00000310, 0.00001334, -0.00002129, 0.00001283, 0.00000066, -0.00001239, 0.00001854, -0.00002424, 0.00001471, 0.00001610, -0.00004410, 0.00002807, 0.00000606, -0.00003360, 0.00002326, 0.00000202, 0.00000771, 0.00001074, -0.00000449, 0.00002685, -0.00000619, -0.00000726, 0.00000732, -0.00002945, -0.00000037, 0.00001169, -0.00000855, 0.00000179, -0.00000231, 0.00000030, -0.00000180, -0.00000695, 0.00000271, -0.00000223, -0.00000171, 0.00000343, -0.00000431, -0.00000462, 0.00000598, 0.00000075, -0.00000559, 0.00000303, -0.00000574, -0.00000244, 0.00000110, -0.00001238, 0.00000258, -0.00000742, -0.00000782, 0.00000603, -0.00003054, -0.00000863, -0.00003170, -0.00010112, 0.00003120, 0.00012595, 0.00006220, 0.00009374, 0.00008574, -0.00003942, -0.00006373, -0.00001388, -0.00000827, -0.00001217, },
		{ -0.00001192, -0.00000668, -0.00000495, -0.00000671, -0.00000460, -0.00000238, -0.00000027, -0.00000332, 0.00000143, 0.00000192, -0.00000406, 0.00000238, -0.00000716, -0.00000845, 0.00000685, -0.00000191, 0.00000015, 0.00000376, -0.00000616, 0.00000007, 0.00000139, 0.00000536, 0.00000631, -0.00000767, -0.00000375, 0.00000164, 0.00000359, -0.00000321, -0.00001788, 0.00000579, 0.00001250, -0.00001119, 0.00000330, 0.00000165, -0.00001309, 0.00000494, 0.00000027, -0.00000464, 0.00000431, -0.00000773, 0.00000104, 0.00001354, 0.00000061, -0.00000696, -0.00000661, -0.00000023, 0.00000126, 0.00000907, 0.00001294, -0.00001135, -0.00001313, -0.00000246, 0.00001257, 0.00002140, -0.00001637, -0.00000584, 0.00001248, -0.00001723, 0.00000193, 0.00000142, -0.00000610, 0.00000897, -0.00000791, 0.00001296, 0.00000058, -0.00001515, 0.00001858, -0.00001524, -0.00000108, 0.00001412, -0.00001981, 0.00001707, -0.00001171, -0.00000826, 0.00003351, -0.00003709, -0.00000001, 0.00002880, -0.00003365, 0.00000761, 0.00001360, 0.00000206, 0.00001510, -0.00000526, 0.00002200, 0.00000631, -0.00001306, 0.00001115, -0.00002458, -0.00001365, 0.00001608, -0.00000669, -0.00000148, 0.00000031, -0.00000185, 0.00000096, -0.00000763, -0.00000022, 0.00000071, -0.00000324, 0.00000282, -0.00000094, -0.00000692, 0.00000267, 0.00000529, -0.00000586, 0.00000067, -0.00000185, -0.00000578, 0.00000292, -0.00001033, -0.00000205, -0.00000104, -0.00001259, 0.00000633, -0.00002049, -0.00001877, -0.00001311, -0.00009230, -0.00001696, 0.00011850, 0.00007351, 0.00006807, 0.00010272, 0.00000408, -0.00006708, -0.00002952, -0.00000597, },
		{ -0.00000935, -0.00000944, -0.00000870, -0.00000477, -0.00000628, -0.00000523, -0.00000362, 0.00000014, -0.00000321, -0.00000069, 0.00000447, -0.00000439, 0.00000117, -0.00000265, -0.00001215, 0.00000497, 0.00000160, -0.00000271, 0.00000517, -0.00000492, -0.00000154, 0.00000131, 0.00000193, 0.00000917, -0.00000459, -0.00000592, 0.00000218, 0.00000145, 0.00000070, -0.00001588, -0.00000356, 0.00001743, -0.00000731, -0.00000462, 0.00000805, -0.00001158, -0.00000160, 0.00000503, -0.00000619, 0.00000337, -0.00000306, -0.00000546, 0.00001171, 0.00000700, -0.00000576, -0.00000865, -0.00000166, 0.00000031, 0.00000565, 0.00001701, -0.00000582, -0.00001656, -0.00000456, 0.00000565, 0.00002605, -0.00000596, -0.00001733, 0.00001600, -0.00001183, -0.00000682, 0.00000759, -0.00000994, 0.00000937, -0.00000481, 0.00000396, 0.00001256, -0.00002083, 0.00001341, -0.00000133, -0.00001525, 0.00002058, -0.00001647, 0.00000530, 0.00000570, -0.00002451, 0.00003290, -0.00001402, -0.00002702, 0.00003479, -0.00001820, -0.00001096, 0.00001926, 0.00000245, 0.00001508, -0.00000219, 0.00001472, 0.00001702, -0.00001428, 0.00001015, -0.00001494, -0.00002589, 0.00001490, -0.00000100, -0.00000590, 0.00000242, -0.00000297, 0.00000190, -0.00000585, -0.00000401, 0.00000225, -0.00000276, 0.00000086, 0.00000179, -0.00000641, -0.00000211, 0.00000748, -0.00000284, -0.00000294, 0.00000097, -0.00000659, 0.00000173, -0.00000621, -0.00000667, 0.00000265, -0.00001236, 0.00000171, -0.00000991, -0.00002472, -0.00000555, -0.00007233, -0.00005410, 0.00009400, 0.00009062, 0.00005013, 0.00009786, 0.00004613, -0.00005264, -0.00004610, },
		{ -0.00005587, -0.00001909, -0.00000723, -0.00000923, -0.00000556, -0.00000588, -0.00000513, -0.00000501, -0.00000024, -0.00000191, -0.00000352, 0.00000587, -0.00000252, -0.00000187, 0.00000178, -0.00001308, 0.00000031, 0.00000558, -0.00000436, 0.00000486, -0.00000252, -0.00000356, 0.00000217, -0.00000136, 0.00000918, 0.00000043, -0.00000811, 0.00000255, 0.00000061, 0.00000126, -0.00001074, -0.00001157, 0.00001674, 0.00000126, -0.00001202, 0.00000888, -0.00000487, -0.00000838, 0.00000651, -0.00000409, -0.00000022, 0.00000172, -0.00000782, 0.00000577, 0.00001178, -0.00000143, -0.00001047, -0.00000388, 0.00000039, 0.00000140, 0.00001830, 0.00000239, -0.00001915, -0.00000667, -0.00000010, 0.00002422, 0.00000849, -0.00002474, 0.00001183, -0.00000117, -0.00001576, 0.00001051, -0.00000911, 0.00000464, 0.00000262, -0.00000588, 0.00001918, -0.00001671, 0.00000007, 0.00001330, -0.00002206, 0.00001619, -0.00000413, -0.00000975, 0.00001799, -0.00002666, 0.00001656, 0.00001254, -0.00004008, 0.00002226, 0.00000373, -0.00002269, 0.00001557, 0.00000720, 0.00001348, 0.00000217, 0.00000754, 0.00002409, -0.00001066, 0.00000538, -0.00000345, -0.00003361, 0.00000743, 0.00000665, -0.00000924, 0.00000261, -0.00000244, 0.00000104, -0.00000274, -0.00000692, 0.00000159, -0.00000094, -0.00000100, 0.00000280, -0.00000383, -0.00000606, 0.00000618, 0.00000204, -0.00000533, 0.00000121, -0.00000496, -0.00000096, -0.00000231, -0.00000916, 0.00000275, -0.00000857, -0.00000406, -0.00000263, -0.00002492, -0.00000609, -0.00004967, -0.00007504, 0.00005755, 0.00010348, 0.00004578, 0.00007913, 0.00007549, -0.00002343, },
		{ 0.00001219, -0.00005322, -0.00003219, -0.00000826, -0.00000812, -0.00000637, -0.00000619, -0.00000456, -0.00000573, -0.00000190, 0.00000019, -0.00000576, 0.00000492, 0.00000145, -0.00000522, 0.00000417, -0.00001050, -0.00000594, 0.00000824, -0.00000377, 0.00000278, 0.00000061, -0.00000552, 0.00000294, -0.00000289, 0.00000606, 0.00000601, -0.00000881, 0.00000139, 0.00000191, -0.00000051, -0.00000546, -0.00001549, 0.00001061, 0.00001094, -0.00001476, 0.00000347, 0.00000364, -0.00001154, 0.00000348, 0.00000029, -0.00000370, 0.00000380, -0.00000574, -0.00000120, 0.00001236, 0.00000515, -0.00001024, -0.00000751, 0.00000105, -0.00000203, 0.00001569, 0.00001199, -0.00001907, -0.00001022, -0.00000307, 0.00001696, 0.00002210, -0.00002393, 0.00000044, 0.00001064, -0.00002039, 0.00000754, -0.00000328, -0.00000329, 0.00001020, -0.00001126, 0.00001707, -0.00000444, -0.00001474, 0.00002088, -0.00001767, 0.00000312, 0.00001036, -0.00001952, 0.00001885, -0.00001541, -0.00000568, 0.00002957, -0.00003428, -0.00000118, 0.00002034, -0.00002221, 0.00000492, 0.00001174, 0.00001300, 0.00000649, 0.00000171, 0.00002687, -0.00000295, -0.00000121, 0.00000675, -0.00003460, -0.00000504, 0.00001330, -0.00000963, 0.00000043, -0.00000057, -0.00000077, 0.00000022, -0.00000782, -0.00000090, 0.00000091, -0.00000172, 0.00000219, -0.00000078, -0.00000760, 0.00000208, 0.00000625, -0.00000478, -0.00000088, -0.00000245, -0.00000320, -0.00000020, -0.00000900, 0.00000031, -0.00000387, -0.00000797, 0.00000016, -0.00002086, -0.00001025, -0.00003053, -0.00008005, 0.00001756, 0.00010485, 0.00005375, 0.00005743, 0.00008681, },
		{ 0.00008162, 0.00004431, -0.00003714, -0.00004316, -0.00001402, -0.00000691, -0.00000607, -0.00000710, -0.00000440, -0.00000511, -0.00000454, 0.00000191, -0.00000603, 0.00000135, 0.00000609, -0.00000680, 0.00000331, -0.00000516, -0.00001161, 0.00000796, -0.00000079, -0.00000032, 0.00000355, -0.00000652, 0.00000247, -0.00000197, 0.00000099, 0.00001000, -0.00000676, -0.00000160, 0.00000453, -0.00000230, -0.00000255, -0.00001477, 0.00000171, 0.00001758, -0.00001103, -0.00000555, 0.00000937, -0.00000903, -0.00000251, 0.00000398, -0.00000459, 0.00000248, -0.00000139, -0.00000575, 0.00000828, 0.00001132, -0.00000639, -0.00001187, 0.00000083, -0.00000308, 0.00000952, 0.00002033, -0.00001445, -0.00001562, -0.00000369, 0.00000758, 0.00002998, -0.00001389, -0.00001409, 0.00001814, -0.00001761, -0.00000107, 0.00000491, -0.00001033, 0.00001340, -0.00000927, 0.00000745, 0.00001002, -0.00002341, 0.00001726, -0.00000426, -0.00001163, 0.00001906, -0.00001881, 0.00000878, 0.00000164, -0.00002227, 0.00003024, -0.00001444, -0.00002314, 0.00002356, -0.00001126, -0.00000632, 0.00001205, 0.00001411, 0.00001096, -0.00000225, 0.00002536, 0.00000745, -0.00000720, 0.00001300, -0.00002850, -0.00001948, 0.00001598, -0.00000628, -0.00000350, 0.00000161, -0.00000237, 0.00000189, -0.00000655, -0.00000411, 0.00000165, -0.00000118, 0.00000091, 0.00000136, -0.00000655, -0.00000281, 0.00000769, -0.00000135, -0.00000354, -0.00000094, -0.00000382, -0.00000003, -0.00000714, -0.00000277, -0.00000044, -0.00000890, -0.00000047, -0.00001526, -0.00001433, -0.00001756, -0.00007323, -0.00001797, 0.00009257, 0.00006783, 0.00004180, },
		{ 0.00003656, 0.00006641, 0.00006542, -0.00001158, -0.00004672, -0.00002329, -0.00000784, -0.00000458, -0.00000754, -0.00000556, -0.00000331, -0.00000693, 0.00000181, -0.00000379, -0.00000363, 0.00000913, -0.00000506, -0.00000050, 0.00000078, -0.00001440, 0.00000412, 0.00000360, -0.00000306, 0.00000514, -0.00000577, 0.00000030, 0.00000076, -0.00000386, 0.00001060, -0.00000198, -0.00000529, 0.00000658, -0.00000207, -0.00000273, -0.00001119, -0.00000636, 0.00001863, -0.00000221, -0.00001364, 0.00000911, -0.00000199, -0.00000790, 0.00000432, -0.00000236, -0.00000065, 0.00000214, -0.00000607, 0.00000157, 0.00001399, 0.00000099, -0.00001487, -0.00000185, -0.00000143, 0.00000192, 0.00002438, -0.00000479, -0.00002140, -0.00000415, -0.00000005, 0.00002957, 0.00000245, -0.00002555, 0.00001714, -0.00000759, -0.00001193, 0.00001105, -0.00001252, 0.00001002, -0.00000102, -0.00000452, 0.00001959, -0.00002149, 0.00000411, 0.00001116, -0.00002017, 0.00001743, -0.00000854, -0.00000580, 0.00001485, -0.00002603, 0.00001654, 0.00000840, -0.00003324, 0.00001297, 0.00000283, -0.00001196, 0.00000736, 0.00001494, 0.00001616, -0.00000364, 0.00001983, 0.00001855, -0.00001013, 0.00001379, -0.00001692, -0.00003203, 0.00001270, 0.00000012, -0.00000765, 0.00000283, -0.00000285, 0.00000184, -0.00000382, -0.00000672, 0.00000080, 0.00000000, -0.00000003, 0.00000204, -0.00000398, -0.00000639, 0.00000574, 0.00000327, -0.00000468, -0.00000136, -0.00000307, -0.00000080, -0.00000502, -0.00000502, 0.00000090, -0.00000749, -0.00000246, -0.00001034, -0.00001662, -0.00001055, -0.00005996, -0.00004389, 0.00006921, 0.00008050, },
		{ 0.00008605, 0.00004097, 0.00004915, 0.00007268, 0.00001660, -0.00004024, -0.00003251, -0.00001232, -0.00000323, -0.00000625, -0.00000799, -0.00000159, -0.00000754, -0.00000075, 0.00000004, -0.00000778, 0.00000863, -0.00000004, -0.00000525, 0.00000467, -0.00001300, -0.00000230, 0.00000748, -0.00000398, 0.00000472, -0.00000322, -0.00000297, 0.00000366, -0.00000636, 0.00000741, 0.00000397, -0.00000773, 0.00000639, 0.00000078, -0.00000469, -0.00000756, -0.00001099, 0.00001426, 0.00000797, -0.00001674, 0.00000277, 0.00000584, -0.00000923, 0.00000081, 0.00000118, -0.00000291, 0.00000278, -0.00000291, -0.00000431, 0.00001149, 0.00000963, -0.00001397, -0.00000736, 0.00000140, -0.00000405, 0.00002227, 0.00000812, -0.00002455, -0.00000699, -0.00000359, 0.00002189, 0.00001927, -0.00002843, 0.00000710, 0.00000575, -0.00001987, 0.00001131, -0.00000838, 0.00000146, 0.00000894, -0.00001259, 0.00001984, -0.00000995, -0.00001196, 0.00002054, -0.00001807, 0.00000676, 0.00000521, -0.00001667, 0.00001787, -0.00001726, -0.00000287, 0.00002319, -0.00002813, -0.00000469, 0.00001166, -0.00000978, 0.00000061, 0.00001302, 0.00002155, -0.00000132, 0.00001122, 0.00002780, -0.00000796, 0.00000931, -0.00000312, -0.00003902, 0.00000319, 0.00000750, -0.00001020, 0.00000218, -0.00000190, 0.00000041, -0.00000078, -0.00000778, -0.00000134, 0.00000091, -0.00000014, 0.00000171, -0.00000136, -0.00000749, 0.00000148, 0.00000669, -0.00000314, -0.00000315, -0.00000217, -0.00000133, -0.00000354, -0.00000601, 0.00000060, -0.00000517, -0.00000413, -0.00000698, -0.00001702, -0.00000788, -0.00004487, -0.00005857, 0.00004003, },
		{ 0.00001057, 0.00008205, 0.00005092, 0.00003620, 0.00006798, 0.00004055, -0.00002473, -0.00003760, -0.00001967, -0.00000422, -0.00000295, -0.00001031, -0.00000168, -0.00000561, -0.00000492, 0.00000336, -0.00000892, 0.00000420, 0.00000634, -0.00000815, 0.00000473, -0.00000796, -0.00000900, 0.00000881, -0.00000238, 0.00000252, 0.00000028, -0.00000594, 0.00000496, -0.00000558, 0.00000171, 0.00000871, -0.00000736, 0.00000364, 0.00000507, -0.00000607, -0.00000605, -0.00001179, 0.00000701, 0.00001547, -0.00001334, -0.00000659, 0.00001007, -0.00000531, -0.00000443, 0.00000319, -0.00000251, 0.00000073, 0.00000103, -0.00000649, 0.00000474, 0.00001590, -0.00000780, -0.00001399, 0.00000270, -0.00000600, 0.00001457, 0.00002043, -0.00002197, -0.00001322, -0.00000350, 0.00001095, 0.00003049, -0.00002065, -0.00000825, 0.00001638, -0.00002061, 0.00000474, 0.00000011, -0.00000795, 0.00001516, -0.00001274, 0.00001120, 0.00000541, -0.00002299, 0.00001909, -0.00000669, -0.00000679, 0.00001492, -0.00001848, 0.00001058, -0.00000249, -0.00001792, 0.00002440, -0.00001233, -0.00001976, 0.00001064, -0.00000263, -0.00000380, 0.00000760, 0.00002501, 0.00000546, 0.00000172, 0.00003236, 0.00000004, 0.00000167, 0.00000884, -0.00003819, -0.00001076, 0.00001299, -0.00000970, -0.00000054, 0.00000005, -0.00000149, 0.00000144, -0.00000701, -0.00000394, 0.00000090, 0.00000037, 0.00000120, 0.00000029, -0.00000635, -0.00000307, 0.00000726, 0.00000059, -0.00000468, -0.00000232, -0.00000117, -0.00000271, -0.00000616, -0.00000042, -0.00000306, -0.00000486, -0.00000493, -0.00001617, -0.00000785, -0.00003098, -0.00006307, },
		{ -0.00005983, -0.00001488, 0.00006916, 0.00006134, 0.00003064, 0.00005612, 0.00005578, -0.00000407, -0.00003594, -0.00002718, -0.00000906, 0.00000090, -0.00001042, -0.00000450, -0.00000194, -0.00000859, 0.00000398, -0.00000627, -0.00000252, 0.00001104, -0.00000708, 0.00000107, -0.00000154, -0.00001330, 0.00000654, 0.00000120, -0.00000026, 0.00000335, -0.00000726, 0.00000378, -0.00000218, -0.00000408, 0.00001023, -0.00000400, -0.00000048, 0.00000870, -0.00000500, -0.00000696, -0.00001043, 0.00000007, 0.00001801, -0.00000509, -0.00001441, 0.00000819, 0.00000197, -0.00000784, 0.00000175, 0.00000021, -0.00000190, 0.00000295, -0.00000433, -0.00000303, 0.00001667, 0.00000247, -0.00001842, 0.00000015, -0.00000366, 0.00000439, 0.00002776, -0.00001244, -0.00002121, -0.00000262, 0.00000156, 0.00003257, -0.00000466, -0.00002232, 0.00001907, -0.00001316, -0.00000601, 0.00000845, -0.00001339, 0.00001428, -0.00000525, -0.00000140, 0.00001715, -0.00002381, 0.00000771, 0.00000774, -0.00001583, 0.00001590, -0.00001131, -0.00000176, 0.00000988, -0.00002244, 0.00001392, 0.00000509, -0.00002512, 0.00000118, 0.00000367, -0.00000317, 0.00000071, 0.00002397, 0.00001580, -0.00000528, 0.00003018, 0.00001251, -0.00000539, 0.00001541, -0.00002956, -0.00002563, 0.00001382, -0.00000577, -0.00000456, 0.00000200, -0.00000283, 0.00000214, -0.00000482, -0.00000609, -0.00000018, 0.00000090, 0.00000117, 0.00000085, -0.00000415, -0.00000607, 0.00000481, 0.00000470, -0.00000435, -0.00000374, -0.00000080, -0.00000194, -0.00000603, -0.00000155, -0.00000158, -0.00000486, -0.00000358, -0.00001458, -0.00000927, -0.00001980, },
		{ -0.00001194, -0.00005157, -0.00003382, 0.00005005, 0.00006802, 0.00003232, 0.00004251, 0.00006107, 0.00001691, -0.00002750, -0.00003147, -0.00001723, 0.00000244, -0.00000690, -0.00000922, 0.00000117, -0.00000950, 0.00000103, -0.00000113, -0.00000850, 0.00001151, -0.00000191, -0.00000414, 0.00000330, -0.00001353, 0.00000125, 0.00000507, -0.00000214, 0.00000485, -0.00000631, 0.00000054, 0.00000197, -0.00000765, 0.00000796, 0.00000095, -0.00000397, 0.00000994, -0.00000119, -0.00000881, -0.00000925, -0.00000451, 0.00001589, 0.00000439, -0.00001712, 0.00000092, 0.00000856, -0.00000669, -0.00000266, 0.00000294, -0.00000263, 0.00000171, 0.00000022, -0.00000797, 0.00001118, 0.00001324, -0.00001737, -0.00000652, 0.00000070, -0.00000404, 0.00002737, 0.00000218, -0.00002720, -0.00000437, -0.00000315, 0.00002607, 0.00001353, -0.00002875, 0.00001210, -0.00000057, -0.00001586, 0.00001205, -0.00001217, 0.00000678, 0.00000563, -0.00001140, 0.00002011, -0.00001454, -0.00000775, 0.00001767, -0.00001586, 0.00000864, -0.00000004, -0.00001205, 0.00001419, -0.00001655, -0.00000110, 0.00001580, -0.00001958, -0.00001065, 0.00000420, 0.00000149, -0.00000385, 0.00001736, 0.00002656, -0.00000627, 0.00002133, 0.00002588, -0.00000785, 0.00001500, -0.00001575, -0.00003712, 0.00000846, 0.00000056, -0.00000839, 0.00000279, -0.00000289, 0.00000132, -0.00000207, -0.00000709, -0.00000195, 0.00000082, 0.00000166, 0.00000094, -0.00000219, -0.00000678, 0.00000065, 0.00000715, -0.00000167, -0.00000552, -0.00000115, -0.00000084, -0.00000579, -0.00000267, -0.00000061, -0.00000453, -0.00000262, -0.00001239, -0.00001134, },
		{ -0.00001330, -0.00000750, -0.00004084, -0.00004540, 0.00002810, 0.00006856, 0.00003891, 0.00003144, 0.00005789, 0.00003431, -0.00001453, -0.00003035, -0.00002587, -0.00000097, -0.00000044, -0.00001319, 0.00000124, -0.00000675, -0.00000426, 0.00000369, -0.00001090, 0.00000711, 0.00000525, -0.00000782, 0.00000447, -0.00000990, -0.00000502, 0.00000727, -0.00000211, 0.00000449, -0.00000355, -0.00000333, 0.00000487, -0.00000788, 0.00000305, 0.00000530, -0.00000528, 0.00000845, 0.00000402, -0.00000947, -0.00000969, -0.00000669, 0.00001134, 0.00001160, -0.00001386, -0.00000821, 0.00001055, -0.00000086, -0.00000723, 0.00000303, -0.00000052, -0.00000142, 0.00000390, -0.00000779, 0.00000175, 0.00001990, -0.00000958, -0.00001484, 0.00000338, -0.00000747, 0.00001970, 0.00001733, -0.00002712, -0.00001041, -0.00000331, 0.00001503, 0.00002723, -0.00002440, -0.00000169, 0.00001137, -0.00001992, 0.00000878, -0.00000522, -0.00000344, 0.00001416, -0.00001402, 0.00001382, -0.00000021, -0.00001975, 0.00001824, -0.00000750, -0.00000202, 0.00000903, -0.00001547, 0.00000996, -0.00000544, -0.00001282, 0.00001630, -0.00000768, -0.00001806, -0.00000200, 0.00000598, -0.00000315, 0.00000706, 0.00003333, 0.00000047, 0.00000873, 0.00003541, -0.00000311, 0.00000891, -0.00000118, -0.00004169, -0.00000263, 0.00000697, -0.00001034, 0.00000166, -0.00000160, -0.00000043, 0.00000028, -0.00000674, -0.00000380, -0.00000007, 0.00000219, 0.00000131, -0.00000112, -0.00000570, -0.00000332, 0.00000676, 0.00000244, -0.00000616, -0.00000276, 0.00000038, -0.00000513, -0.00000392, 0.00000004, -0.00000405, -0.00000208, -0.00000963, },
		{ -0.00000655, -0.00001442, -0.00000611, -0.00002981, -0.00004995, 0.00000658, 0.00006237, 0.00004720, 0.00002539, 0.00004910, 0.00004603, -0.00000002, -0.00002404, -0.00003108, -0.00000985, 0.00000587, -0.00001334, -0.00000273, -0.00000167, -0.00000910, 0.00000538, -0.00000865, -0.00000034, 0.00001102, -0.00000759, 0.00000187, -0.00000443, -0.00000978, 0.00000664, -0.00000031, 0.00000294, -0.00000017, -0.00000628, 0.00000535, -0.00000525, -0.00000234, 0.00000726, -0.00000416, 0.00000527, 0.00000861, -0.00000754, -0.00001148, -0.00000798, 0.00000678, 0.00001492, -0.00000665, -0.00001488, 0.00000645, 0.00000678, -0.00000846, -0.00000058, 0.00000308, -0.00000373, 0.00000421, -0.00000308, -0.00000722, 0.00001927, 0.00000293, -0.00002052, 0.00000137, -0.00000545, 0.00000832, 0.00002769, -0.00001903, -0.00001933, -0.00000192, 0.00000473, 0.00003213, -0.00001088, -0.00001618, 0.00001707, -0.00001612, 0.00000013, 0.00000358, -0.00001134, 0.00001616, -0.00000859, 0.00000237, 0.00001219, -0.00002301, 0.00000964, 0.00000428, -0.00000994, 0.00001173, -0.00001165, 0.00000114, 0.00000436, -0.00001676, 0.00000875, 0.00000361, -0.00001749, -0.00001138, 0.00000567, 0.00000262, -0.00000251, 0.00003254, 0.00001347, -0.00000249, 0.00003716, 0.00000845, 0.00000103, 0.00000954, -0.00003798, -0.00001682, 0.00001061, -0.00000929, -0.00000132, 0.00000039, -0.00000214, 0.00000151, -0.00000527, -0.00000516, -0.00000154, 0.00000207, 0.00000229, -0.00000072, -0.00000403, -0.00000561, 0.00000381, 0.00000616, -0.00000459, -0.00000521, 0.00000091, -0.00000369, -0.00000527, 0.00000026, -0.00000330, -0.00000213, },
		{ -0.00000266, -0.00000369, -0.00001412, -0.00000698, -0.00002021, -0.00004870, -0.00001185, 0.00005042, 0.00005397, 0.00002505, 0.00003792, 0.00005153, 0.00001370, -0.00001485, -0.00003006, -0.00002163, 0.00000783, -0.00000815, -0.00000921, 0.00000269, -0.00001067, 0.00000282, -0.00000315, -0.00000744, 0.00001253, -0.00000315, -0.00000253, 0.00000017, -0.00001139, 0.00000340, 0.00000217, 0.00000129, 0.00000269, -0.00000735, 0.00000347, -0.00000135, -0.00000621, 0.00000637, -0.00000165, 0.00000201, 0.00001117, -0.00000321, -0.00001298, -0.00000997, 0.00000331, 0.00001501, 0.00000117, -0.00001635, -0.00000195, 0.00001192, -0.00000445, -0.00000627, 0.00000526, -0.00000298, 0.00000100, 0.00000301, -0.00001153, 0.00001136, 0.00001541, -0.00001965, -0.00000563, -0.00000063, -0.00000178, 0.00002975, -0.00000449, -0.00002714, -0.00000280, -0.00000119, 0.00002806, 0.00000639, -0.00002489, 0.00001392, -0.00000632, -0.00000961, 0.00000956, -0.00001339, 0.00001106, 0.00000147, -0.00000807, 0.00001735, -0.00001692, -0.00000346, 0.00001313, -0.00001122, 0.00000803, -0.00000404, -0.00000689, 0.00000857, -0.00001332, -0.00000126, 0.00000902, -0.00001029, -0.00001827, -0.00000133, 0.00000967, -0.00000668, 0.00002379, 0.00002799, -0.00000705, 0.00003011, 0.00002322, -0.00000368, 0.00001356, -0.00002749, -0.00003009, 0.00000926, -0.00000537, -0.00000518, 0.00000200, -0.00000297, 0.00000142, -0.00000328, -0.00000573, -0.00000305, 0.00000094, 0.00000353, -0.00000023, -0.00000285, -0.00000586, -0.00000024, 0.00000774, -0.00000088, -0.00000725, -0.00000000, -0.00000159, -0.00000626, -0.00000023, -0.00000216, },
		{ -0.00000080, -0.00000327, -0.00000167, -0.00001232, -0.00000893, -0.00001308, -0.00004349, -0.00002550, 0.00003477, 0.00005668, 0.00002960, 0.00002732, 0.00005112, 0.00002565, -0.00000563, -0.00002282, -0.00003147, 0.00000250, 0.00000080, -0.00001458, 0.00000328, -0.00000795, -0.00000264, 0.00000247, -0.00001104, 0.00000899, 0.00000349, -0.00000587, 0.00000200, -0.00000986, -0.00000097, 0.00000394, 0.00000033, 0.00000445, -0.00000650, 0.00000026, 0.00000210, -0.00000760, 0.00000350, 0.00000070, -0.00000023, 0.00001151, 0.00000209, -0.00001250, -0.00001307, 0.00000035, 0.00001376, 0.00000692, -0.00001272, -0.00001063, 0.00001126, 0.00000363, -0.00001045, 0.00000355, 0.00000093, -0.00000339, 0.00000669, -0.00000942, -0.00000028, 0.00002253, -0.00001105, -0.00001471, 0.00000274, -0.00000677, 0.00002356, 0.00001180, -0.00002931, -0.00000807, -0.00000236, 0.00001859, 0.00002087, -0.00002411, 0.00000357, 0.00000472, -0.00001573, 0.00000982, -0.00000933, 0.00000182, 0.00001085, -0.00001260, 0.00001409, -0.00000536, -0.00001465, 0.00001482, -0.00000603, 0.00000136, 0.00000276, -0.00001055, 0.00000689, -0.00000640, -0.00000842, 0.00000749, -0.00000134, -0.00001847, -0.00001263, 0.00001255, -0.00000335, 0.00001054, 0.00003801, -0.00000214, 0.00001708, 0.00003563, -0.00000150, 0.00001115, -0.00001406, -0.00003855, 0.00000235, -0.00000003, -0.00000842, 0.00000226, -0.00000258, 0.00000035, -0.00000140, -0.00000552, -0.00000407, -0.00000102, 0.00000423, 0.00000100, -0.00000246, -0.00000481, -0.00000364, 0.00000652, 0.00000366, -0.00000747, -0.00000245, 0.00000045, -0.00000631, -0.00000155, },
		{ -0.00000340, 0.00000030, -0.00000344, -0.00000079, -0.00000946, -0.00001078, -0.00000870, -0.00003631, -0.00003382, 0.00001810, 0.00005387, 0.00003700, 0.00001991, 0.00004550, 0.00003557, 0.00000200, -0.00001233, -0.00003455, -0.00000965, 0.00000911, -0.00001506, -0.00000093, -0.00000249, -0.00000795, 0.00000517, -0.00000990, 0.00000214, 0.00000913, -0.00000602, 0.00000093, -0.00000670, -0.00000467, 0.00000421, 0.00000012, 0.00000516, -0.00000427, -0.00000309, 0.00000405, -0.00000680, 0.00000014, 0.00000189, -0.00000142, 0.00001045, 0.00000674, -0.00000945, -0.00001615, -0.00000353, 0.00001260, 0.00001001, -0.00000652, -0.00001560, 0.00000451, 0.00001187, -0.00000968, -0.00000221, 0.00000561, -0.00000567, 0.00000572, -0.00000255, -0.00001032, 0.00002104, 0.00000266, -0.00002100, 0.00000128, -0.00000590, 0.00001268, 0.00002416, -0.00002337, -0.00001675, -0.00000145, 0.00000874, 0.00002798, -0.00001445, -0.00000898, 0.00001184, -0.00001556, 0.00000463, -0.00000195, -0.00000699, 0.00001516, -0.00000982, 0.00000538, 0.00000582, -0.00001927, 0.00000917, 0.00000200, -0.00000390, 0.00000578, -0.00000948, 0.00000205, -0.00000038, -0.00001040, 0.00000187, 0.00000423, -0.00001179, -0.00002270, 0.00000747, 0.00000529, -0.00000126, 0.00003910, 0.00001089, 0.00000383, 0.00004067, 0.00000813, 0.00000561, -0.00000225, -0.00004003, -0.00000862, 0.00000439, -0.00000974, 0.00000079, -0.00000123, -0.00000105, -0.00000012, -0.00000479, -0.00000436, -0.00000316, 0.00000361, 0.00000303, -0.00000235, -0.00000363, -0.00000526, 0.00000313, 0.00000719, -0.00000517, -0.00000557, 0.00000140, -0.00000508, },
		{ -0.00000290, -0.00000507, 0.00000063, -0.00000293, -0.00000091, -0.00000630, -0.00001173, -0.00000660, -0.00002888, -0.00003742, 0.00000303, 0.00004560, 0.00004435, 0.00001758, 0.00003589, 0.00004296, 0.00000847, -0.00000290, -0.00002881, -0.00002431, 0.00001158, -0.00000917, -0.00000799, 0.00000236, -0.00001016, 0.00000370, -0.00000527, -0.00000481, 0.00001113, -0.00000277, -0.00000146, -0.00000386, -0.00000666, 0.00000320, 0.00000018, 0.00000512, -0.00000131, -0.00000572, 0.00000432, -0.00000477, -0.00000256, 0.00000196, -0.00000230, 0.00000893, 0.00000995, -0.00000470, -0.00001734, -0.00000910, 0.00001142, 0.00001170, -0.00000088, -0.00001513, -0.00000533, 0.00001568, -0.00000291, -0.00000929, 0.00000757, -0.00000379, 0.00000085, 0.00000493, -0.00001429, 0.00001162, 0.00001601, -0.00002030, -0.00000538, -0.00000181, 0.00000224, 0.00002866, -0.00001045, -0.00002508, -0.00000213, 0.00000224, 0.00002689, -0.00000021, -0.00001802, 0.00001208, -0.00000972, -0.00000306, 0.00000474, -0.00001163, 0.00001301, -0.00000204, -0.00000371, 0.00001192, -0.00001639, -0.00000039, 0.00000823, -0.00000506, 0.00000497, -0.00000588, -0.00000256, 0.00000234, -0.00000837, -0.00000350, 0.00000407, -0.00000209, -0.00002604, -0.00000511, 0.00001358, -0.00000632, 0.00003084, 0.00002687, -0.00000349, 0.00003642, 0.00002214, 0.00000173, 0.00000470, -0.00003492, -0.00002062, 0.00000571, -0.00000865, -0.00000201, 0.00000031, -0.00000210, 0.00000039, -0.00000388, -0.00000406, -0.00000464, 0.00000149, 0.00000512, -0.00000162, -0.00000318, -0.00000507, -0.00000086, 0.00000826, -0.00000085, -0.00000788, 0.00000055, },
		{ -0.00000196, -0.00000068, -0.00000584, -0.00000004, -0.00000187, -0.00000152, -0.00000356, -0.00001151, -0.00000595, -0.00002217, -0.00003774, -0.00000862, 0.00003354, 0.00004850, 0.00002094, 0.00002455, 0.00004646, 0.00001553, 0.00000234, -0.00001632, -0.00003504, 0.00000521, 0.00000090, -0.00001381, 0.00000344, -0.00000824, -0.00000072, -0.00000007, -0.00000893, 0.00000885, 0.00000213, -0.00000304, -0.00000246, -0.00000715, 0.00000177, 0.00000008, 0.00000440, 0.00000198, -0.00000725, 0.00000324, -0.00000241, -0.00000435, 0.00000169, -0.00000348, 0.00000711, 0.00001203, 0.00000016, -0.00001545, -0.00001550, 0.00000867, 0.00001370, 0.00000255, -0.00001050, -0.00001367, 0.00001241, 0.00000743, -0.00001338, 0.00000452, 0.00000159, -0.00000469, 0.00000878, -0.00001089, -0.00000134, 0.00002326, -0.00001157, -0.00001416, 0.00000121, -0.00000374, 0.00002496, 0.00000513, -0.00002865, -0.00000663, -0.00000017, 0.00002029, 0.00001296, -0.00001985, 0.00000605, -0.00000152, -0.00000938, 0.00000770, -0.00001092, 0.00000619, 0.00000643, -0.00000886, 0.00001147, -0.00000867, -0.00000912, 0.00000969, -0.00000239, 0.00000253, -0.00000246, -0.00000500, 0.00000211, -0.00000525, -0.00000576, -0.00000036, 0.00000520, -0.00002052, -0.00002029, 0.00001535, -0.00000301, 0.00001719, 0.00003916, -0.00000128, 0.00002506, 0.00003515, 0.00000338, 0.00000635, -0.00002590, -0.00003034, 0.00000288, -0.00000579, -0.00000515, 0.00000125, -0.00000243, 0.00000026, -0.00000304, -0.00000357, -0.00000494, -0.00000162, 0.00000618, 0.00000032, -0.00000346, -0.00000400, -0.00000384, 0.00000654, 0.00000395, -0.00000806, },
		{ -0.00000562, -0.00000510, 0.00000052, -0.00000536, -0.00000144, -0.00000073, -0.00000211, -0.00000162, -0.00001034, -0.00000603, -0.00001639, -0.00003631, -0.00001642, 0.00002048, 0.00004716, 0.00002870, 0.00001472, 0.00004443, 0.00002445, 0.00000332, -0.00000243, -0.00003650, -0.00000872, 0.00000998, -0.00001447, -0.00000014, -0.00000372, -0.00000532, 0.00000297, -0.00000911, 0.00000378, 0.00000615, -0.00000253, -0.00000225, -0.00000714, 0.00000065, -0.00000004, 0.00000277, 0.00000525, -0.00000743, 0.00000116, -0.00000007, -0.00000560, 0.00000172, -0.00000470, 0.00000437, 0.00001372, 0.00000398, -0.00001091, -0.00002049, 0.00000280, 0.00001645, 0.00000436, -0.00000492, -0.00001685, 0.00000297, 0.00001645, -0.00001108, -0.00000294, 0.00000723, -0.00000715, 0.00000705, -0.00000263, -0.00001192, 0.00002133, 0.00000222, -0.00002013, -0.00000016, -0.00000438, 0.00001618, 0.00001795, -0.00002489, -0.00001430, -0.00000051, 0.00001237, 0.00002086, -0.00001435, -0.00000277, 0.00000501, -0.00001173, 0.00000624, -0.00000636, -0.00000174, 0.00001172, -0.00000841, 0.00000635, -0.00000040, -0.00001359, 0.00000636, 0.00000165, 0.00000092, -0.00000057, -0.00000549, 0.00000081, -0.00000338, -0.00000488, -0.00000525, 0.00000634, -0.00000870, -0.00003109, 0.00000747, 0.00000557, 0.00000445, 0.00004286, 0.00000974, 0.00001204, 0.00004205, 0.00001161, 0.00000505, -0.00001655, -0.00003557, -0.00000360, -0.00000259, -0.00000760, 0.00000110, -0.00000209, -0.00000021, -0.00000232, -0.00000332, -0.00000407, -0.00000461, 0.00000531, 0.00000325, -0.00000369, -0.00000320, -0.00000498, 0.00000294, 0.00000743, },
		{ 0.00000832, -0.00000128, -0.00000740, 0.00000005, -0.00000391, -0.00000297, 0.00000002, -0.00000234, -0.00000053, -0.00000859, -0.00000644, -0.00001129, -0.00003408, -0.00002138, 0.00000931, 0.00003991, 0.00003767, 0.00000987, 0.00003628, 0.00003439, 0.00000306, 0.00000703, -0.00002742, -0.00002468, 0.00001254, -0.00000877, -0.00000628, 0.00000031, -0.00000756, 0.00000281, -0.00000642, -0.00000147, 0.00000751, -0.00000025, -0.00000196, -0.00000748, -0.00000027, 0.00000050, 0.00000016, 0.00000788, -0.00000593, -0.00000176, 0.00000218, -0.00000660, 0.00000187, -0.00000489, 0.00000003, 0.00001514, 0.00000692, -0.00000550, -0.00002178, -0.00000620, 0.00001847, 0.00000667, -0.00000136, -0.00001416, -0.00000848, 0.00001931, -0.00000211, -0.00001115, 0.00000921, -0.00000465, 0.00000115, 0.00000570, -0.00001563, 0.00001136, 0.00001532, -0.00001922, -0.00000619, -0.00000200, 0.00000701, 0.00002418, -0.00001454, -0.00002196, -0.00000181, 0.00000639, 0.00002235, -0.00000453, -0.00001004, 0.00000728, -0.00000986, 0.00000192, -0.00000078, -0.00000752, 0.00001211, -0.00000366, 0.00000015, 0.00000497, -0.00001309, 0.00000062, 0.00000430, 0.00000120, 0.00000025, -0.00000538, -0.00000000, -0.00000312, -0.00000305, -0.00000717, 0.00000143, 0.00000353, -0.00003222, -0.00000808, 0.00001308, -0.00000208, 0.00003719, 0.00002490, 0.00000342, 0.00004044, 0.00002407, 0.00000465, -0.00000951, -0.00003605, -0.00001195, -0.00000058, -0.00000878, -0.00000014, -0.00000137, -0.00000078, -0.00000164, -0.00000352, -0.00000263, -0.00000635, 0.00000242, 0.00000614, -0.00000286, -0.00000331, -0.00000453, -0.00000094, },
		{ -0.00000363, 0.00000654, 0.00000341, -0.00000764, -0.00000190, -0.00000227, -0.00000402, 0.00000015, -0.00000220, -0.00000011, -0.00000653, -0.00000712, -0.00000670, -0.00003104, -0.00002517, 0.00000174, 0.00002866, 0.00004371, 0.00001205, 0.00002377, 0.00004225, 0.00000567, 0.00000893, -0.00001149, -0.00003542, 0.00000588, 0.00000071, -0.00001104, 0.00000123, -0.00000676, 0.00000047, -0.00000308, -0.00000484, 0.00000617, 0.00000221, -0.00000049, -0.00000797, -0.00000189, 0.00000221, -0.00000283, 0.00000884, -0.00000244, -0.00000521, 0.00000408, -0.00000700, 0.00000115, -0.00000298, -0.00000558, 0.00001525, 0.00001006, -0.00000118, -0.00001860, -0.00001619, 0.00001708, 0.00001121, -0.00000058, -0.00000805, -0.00001678, 0.00001391, 0.00001006, -0.00001528, 0.00000542, 0.00000155, -0.00000507, 0.00000968, -0.00001163, -0.00000177, 0.00002206, -0.00001071, -0.00001384, -0.00000045, 0.00000109, 0.00002324, -0.00000125, -0.00002582, -0.00000595, 0.00000317, 0.00001918, 0.00000544, -0.00001281, 0.00000525, -0.00000559, -0.00000278, 0.00000330, -0.00000959, 0.00000839, 0.00000244, -0.00000398, 0.00000630, -0.00000936, -0.00000457, 0.00000421, 0.00000256, 0.00000137, -0.00000563, -0.00000021, -0.00000315, -0.00000263, -0.00000520, -0.00000595, 0.00001025, -0.00002306, -0.00002496, 0.00001346, -0.00000099, 0.00002558, 0.00003797, 0.00000308, 0.00003179, 0.00003619, 0.00000831, -0.00000535, -0.00003321, -0.00002009, -0.00000068, -0.00000874, -0.00000209, -0.00000065, -0.00000133, -0.00000093, -0.00000400, -0.00000148, -0.00000622, -0.00000166, 0.00000763, -0.00000046, -0.00000408, -0.00000351, },
		},
		{ { -0.00000230, -0.00000115, -0.00000412, -0.00000680, 0.00000001, 0.00000347, -0.00000095, -0.00000223, -0.00000166, -0.00000222, -0.00000065, -0.00000225, -0.00000367, 0.00000409, 0.00000257, -0.00000576, 0.00000256, -0.00000319, -0.00001964, -0.00000444, -0.00000486, -0.00002321, -0.00000359, 0.00000175, 0.00000039, 0.00003940, 0.00002505, -0.00001656, 0.00002063, 0.00003929, 0.00000244, 0.00000587, 0.00001200, -0.00002259, -0.00002858, 0.00000781, 0.00001582, -0.00001130, -0.00000845, 0.00000331, -0.00001072, 0.00000230, 0.00001396, -0.00000864, -0.00000368, 0.00000072, -0.00000924, 0.00001261, 0.00000701, -0.00001659, 0.00000374, 0.00000991, 0.00000072, 0.00000493, -0.00000674, -0.00000759, 0.00000154, 0.00000069, 0.00000940, -0.00000144, -0.00001350, 0.00000578, 0.00000457, -0.00000327, 0.00000299, -0.00000327, 0.00000457, 0.00000578, -0.00001350, -0.00000144, 0.00000940, 0.00000069, 0.00000154, -0.00000759, -0.00000674, 0.00000493, 0.00000072, 0.00000991, 0.00000374, -0.00001659, 0.00000701, 0.00001261, -0.00000924, 0.00000072, -0.00000368, -0.00000864, 0.00001396, 0.00000230, -0.00001072, 0.00000331, -0.00000845, -0.00001130, 0.00001582, 0.00000781, -0.00002858, -0.00002259, 0.00001200, 0.00000587, 0.00000244, 0.00003929, 0.00002063, -0.00001656, 0.00002505, 0.00003940, 0.00000039, 0.00000175, -0.00000359, -0.00002321, -0.00000486, -0.00000444, -0.00001964, -0.00000319, 0.00000256, -0.00000576, 0.00000257, 0.00000409, -0.00000367, -0.00000225, -0.00000065, -0.00000222, -0.00000166, -0.00000223, -0.00000095, 0.00000347, 0.00000001, -0.00000680, -0.00000412, -0.00000115, },
		{ -0.00000123, -0.00000184, -0.00000183, -0.00000243, -0.00000677, -0.00000311, 0.00000346, 0.00000095, -0.00000222, -0.00000165, -0.00000242, -0.00000124, -0.00000075, -0.00000463, 0.00000068, 0.00000598, -0.00000425, -0.00000167, 0.00000365, -0.00001573, -0.00001272, 0.00000025, -0.00001970, -0.00001218, 0.00000722, -0.00000561, 0.00002093, 0.00004079, -0.00000689, -0.00000324, 0.00004002, 0.00001765, -0.00000273, 0.00001418, -0.00000206, -0.00003272, -0.00001319, 0.00002145, 0.00000294, -0.00001726, 0.00000251, -0.00000376, -0.00000990, 0.00001514, 0.00000250, -0.00001078, 0.00000322, -0.00000760, 0.00000082, 0.00001590, -0.00001066, -0.00000805, 0.00001242, 0.00000343, 0.00000316, -0.00000103, -0.00000988, -0.00000134, 0.00000077, 0.00000676, 0.00000488, -0.00001293, -0.00000111, 0.00000793, -0.00000292, 0.00000228, -0.00000153, -0.00000046, 0.00001028, -0.00000900, -0.00000942, 0.00001019, 0.00000248, 0.00000112, -0.00000203, -0.00001106, 0.00000266, 0.00000301, 0.00000345, 0.00001127, -0.00001155, -0.00000681, 0.00001757, -0.00000171, -0.00000611, 0.00000231, -0.00001068, 0.00000343, 0.00001209, -0.00000803, -0.00000341, -0.00000007, -0.00001324, 0.00000106, 0.00001676, -0.00001013, -0.00003397, -0.00000457, 0.00001479, -0.00000175, 0.00002460, 0.00004027, -0.00000747, 0.00000252, 0.00004537, 0.00001745, -0.00000280, 0.00000053, -0.00002052, -0.00001359, 0.00000081, -0.00001661, -0.00001207, 0.00000323, -0.00000369, -0.00000174, 0.00000509, -0.00000098, -0.00000332, -0.00000087, -0.00000175, -0.00000175, -0.00000187, -0.00000224, 0.00000204, 0.00000258, -0.00000511, -0.00000610, },
		{ -0.00000726, -0.00000255, -0.00000087, -0.00000237, -0.00000174, -0.00000543, -0.00000563, 0.00000182, 0.00000275, -0.00000151, -0.00000177, -0.00000223, -0.00000220, 0.00000005, -0.00000356, -0.00000314, 0.00000629, 0.00000030, -0.00000567, 0.00000509, -0.00000692, -0.00001848, -0.00000126, -0.00001166, -0.00002002, 0.00000673, -0.00000057, 0.00000170, 0.00004039, 0.00001417, -0.00001687, 0.00002467, 0.00003307, -0.00000042, 0.00000500, 0.00001262, -0.00001915, -0.00003174, 0.00000928, 0.00002115, -0.00001509, -0.00000826, 0.00000555, -0.00001392, 0.00000439, 0.00001512, -0.00001031, -0.00000211, -0.00000051, -0.00000830, 0.00001373, 0.00000225, -0.00001547, 0.00000684, 0.00000935, 0.00000162, 0.00000203, -0.00000789, -0.00000547, 0.00000029, 0.00000434, 0.00000836, -0.00000867, -0.00000769, 0.00000835, -0.00000061, 0.00000027, 0.00000159, -0.00000468, 0.00000992, -0.00000026, -0.00001553, 0.00000637, 0.00000708, -0.00000127, 0.00000261, -0.00001008, -0.00000400, 0.00000662, -0.00000024, 0.00001029, -0.00000010, -0.00001561, 0.00001123, 0.00001023, -0.00000914, 0.00000137, -0.00000497, -0.00000704, 0.00001192, 0.00000183, -0.00000850, 0.00000044, -0.00000650, -0.00000956, 0.00001030, 0.00000853, -0.00002738, -0.00002578, 0.00001179, 0.00000558, 0.00000781, 0.00004426, 0.00001536, -0.00001202, 0.00003435, 0.00003698, 0.00000126, -0.00000073, -0.00001454, -0.00002090, -0.00000031, -0.00000903, -0.00001804, -0.00000113, -0.00000026, -0.00000429, 0.00000327, 0.00000189, -0.00000299, -0.00000166, -0.00000142, -0.00000177, -0.00000148, -0.00000264, -0.00000004, 0.00000365, -0.00000211, },
		{ 0.00000108, -0.00000672, -0.00000479, -0.00000029, -0.00000208, -0.00000198, -0.00000372, -0.00000675, -0.00000094, 0.00000362, -0.00000009, -0.00000185, -0.00000186, -0.00000291, -0.00000033, -0.00000150, -0.00000533, 0.00000332, 0.00000514, -0.00000621, 0.00000150, 0.00000181, -0.00001777, -0.00000785, -0.00000406, -0.00002221, -0.00000094, 0.00000837, -0.00000796, 0.00002546, 0.00003271, -0.00001307, 0.00000244, 0.00003692, 0.00001283, -0.00000352, 0.00001276, 0.00000177, -0.00003353, -0.00001552, 0.00002786, 0.00000218, -0.00002027, 0.00000614, -0.00000603, -0.00001046, 0.00001834, 0.00000063, -0.00001017, 0.00000347, -0.00000845, 0.00000324, 0.00001260, -0.00001273, -0.00000412, 0.00001262, 0.00000388, 0.00000199, -0.00000404, -0.00000783, -0.00000179, 0.00000230, 0.00000939, -0.00000258, -0.00001206, 0.00000558, 0.00000310, -0.00000229, 0.00000424, -0.00000546, 0.00000446, 0.00000887, -0.00001562, -0.00000260, 0.00001201, -0.00000170, 0.00000273, -0.00000399, -0.00001067, 0.00000612, 0.00000192, 0.00000393, 0.00000895, -0.00001370, -0.00000200, 0.00001738, -0.00000359, -0.00000451, 0.00000099, -0.00000998, 0.00000322, 0.00000995, -0.00000572, -0.00000476, -0.00000079, -0.00000960, -0.00000271, 0.00001444, -0.00000772, -0.00003635, -0.00000587, 0.00001393, 0.00000067, 0.00003248, 0.00003784, -0.00000861, 0.00001317, 0.00004708, 0.00001596, -0.00000256, -0.00000988, -0.00002363, -0.00000682, -0.00000192, -0.00001804, -0.00000850, 0.00000120, -0.00000398, 0.00000010, 0.00000322, -0.00000132, -0.00000221, -0.00000150, -0.00000175, -0.00000124, -0.00000232, -0.00000188, 0.00000296, },
		{ 0.00000102, 0.00000324, -0.00000436, -0.00000690, -0.00000091, -0.00000102, -0.00000248, -0.00000252, -0.00000638, -0.00000386, 0.00000308, 0.00000164, -0.00000158, -0.00000154, -0.00000303, -0.00000155, 0.00000015, -0.00000507, -0.00000112, 0.00000739, -0.00000275, -0.00000366, 0.00000601, -0.00001092, -0.00001467, -0.00000133, -0.00001802, -0.00001150, 0.00001301, -0.00000537, 0.00000598, 0.00003805, 0.00000374, -0.00001329, 0.00002596, 0.00002718, -0.00000146, 0.00000265, 0.00001454, -0.00001664, -0.00003606, 0.00001346, 0.00002481, -0.00001983, -0.00000582, 0.00000615, -0.00001710, 0.00000781, 0.00001502, -0.00001157, -0.00000065, -0.00000189, -0.00000656, 0.00001313, -0.00000196, -0.00001273, 0.00000882, 0.00000906, 0.00000183, -0.00000138, -0.00000742, -0.00000436, -0.00000006, 0.00000842, 0.00000379, -0.00001306, -0.00000001, 0.00000706, -0.00000374, 0.00000419, -0.00000199, -0.00000301, 0.00001328, -0.00000811, -0.00001309, 0.00001246, 0.00000268, -0.00000114, 0.00000259, -0.00001202, -0.00000022, 0.00000687, -0.00000079, 0.00000993, -0.00000387, -0.00001264, 0.00001403, 0.00000742, -0.00000817, 0.00000127, -0.00000566, -0.00000604, 0.00000967, 0.00000274, -0.00000787, -0.00000187, -0.00000328, -0.00001023, 0.00000598, 0.00001030, -0.00002811, -0.00002790, 0.00001088, 0.00000506, 0.00001551, 0.00004696, 0.00001077, -0.00000414, 0.00004118, 0.00003496, 0.00000193, -0.00000840, -0.00002254, -0.00001491, 0.00000062, -0.00001285, -0.00001479, -0.00000118, -0.00000210, -0.00000230, 0.00000246, 0.00000058, -0.00000182, -0.00000179, -0.00000185, -0.00000118, -0.00000165, -0.00000285, },
		{ -0.00000281, -0.00000123, 0.00000359, -0.00000095, -0.00000771, -0.00000290, 0.00000015, -0.00000249, -0.00000217, -0.00000503, -0.00000602, 0.00000116, 0.00000304, -0.00000079, -0.00000132, -0.00000258, -0.00000288, 0.00000050, -0.00000306, -0.00000467, 0.00000599, 0.00000253, -0.00000653, 0.00000455, -0.00000191, -0.00001703, -0.00000418, -0.00001087, -0.00001920, 0.00000962, 0.00000439, -0.00000733, 0.00002887, 0.00002200, -0.00001472, 0.00000743, 0.00003186, 0.00001061, -0.00000527, 0.00001160, 0.00000625, -0.00003666, -0.00001577, 0.00003466, -0.00000112, -0.00002119, 0.00000959, -0.00000976, -0.00000924, 0.00002089, -0.00000171, -0.00000903, 0.00000309, -0.00000838, 0.00000474, 0.00000844, -0.00001272, -0.00000061, 0.00001238, 0.00000443, -0.00000043, -0.00000568, -0.00000581, -0.00000262, 0.00000553, 0.00000898, -0.00001000, -0.00000716, 0.00000917, -0.00000215, 0.00000077, 0.00000345, -0.00000764, 0.00001012, 0.00000372, -0.00001885, 0.00000540, 0.00001024, -0.00000430, 0.00000446, -0.00000671, -0.00000851, 0.00000854, 0.00000051, 0.00000449, 0.00000570, -0.00001390, 0.00000269, 0.00001567, -0.00000445, -0.00000332, -0.00000030, -0.00000910, 0.00000207, 0.00000918, -0.00000397, -0.00000699, 0.00000049, -0.00000754, -0.00000703, 0.00001426, -0.00000693, -0.00003871, -0.00000668, 0.00001168, 0.00000535, 0.00003988, 0.00003387, -0.00000607, 0.00002327, 0.00004744, 0.00001562, -0.00000706, -0.00002062, -0.00002125, -0.00000216, -0.00000615, -0.00001685, -0.00000640, -0.00000116, -0.00000288, 0.00000057, 0.00000142, -0.00000064, -0.00000172, -0.00000214, -0.00000138, -0.00000100, },
		{ -0.00000070, -0.00000200, -0.00000286, 0.00000213, 0.00000220, -0.00000650, -0.00000560, 0.00000047, -0.00000164, -0.00000247, -0.00000347, -0.00000686, -0.00000163, 0.00000356, 0.00000043, -0.00000115, -0.00000180, -0.00000366, -0.00000046, -0.00000073, -0.00000592, 0.00000202, 0.00000659, -0.00000535, -0.00000027, 0.00000458, -0.00001348, -0.00000961, -0.00000526, -0.00002097, 0.00000009, 0.00001339, -0.00000943, 0.00001217, 0.00003142, -0.00000387, -0.00000742, 0.00002398, 0.00002309, -0.00000194, -0.00000075, 0.00001809, -0.00001615, -0.00004017, 0.00001992, 0.00002577, -0.00002407, -0.00000177, 0.00000463, -0.00001924, 0.00001186, 0.00001368, -0.00001218, 0.00000039, -0.00000291, -0.00000467, 0.00001086, -0.00000469, -0.00000921, 0.00000977, 0.00000919, 0.00000088, -0.00000442, -0.00000587, -0.00000433, 0.00000124, 0.00001125, -0.00000323, -0.00001327, 0.00000710, 0.00000288, -0.00000389, 0.00000665, -0.00000616, 0.00000142, 0.00001312, -0.00001534, -0.00000696, 0.00001522, -0.00000201, 0.00000060, 0.00000111, -0.00001254, 0.00000370, 0.00000594, -0.00000080, 0.00000849, -0.00000667, -0.00000838, 0.00001501, 0.00000496, -0.00000683, 0.00000060, -0.00000575, -0.00000597, 0.00000800, 0.00000431, -0.00000885, -0.00000270, -0.00000006, -0.00001279, 0.00000370, 0.00001173, -0.00003003, -0.00002884, 0.00000852, 0.00000550, 0.00002466, 0.00004730, 0.00000844, 0.00000549, 0.00004592, 0.00003404, -0.00000045, -0.00001934, -0.00002522, -0.00000824, -0.00000168, -0.00001438, -0.00001155, -0.00000273, -0.00000242, -0.00000094, 0.00000087, 0.00000043, -0.00000090, -0.00000229, -0.00000189, },
		{ -0.00000249, -0.00000095, -0.00000097, -0.00000326, -0.00000038, 0.00000382, -0.00000342, -0.00000780, -0.00000071, -0.00000016, -0.00000283, -0.00000241, -0.00000628, -0.00000453, 0.00000284, 0.00000186, -0.00000093, -0.00000098, -0.00000358, -0.00000214, 0.00000067, -0.00000485, -0.00000237, 0.00000751, -0.00000107, -0.00000477, 0.00000610, -0.00000617, -0.00001349, -0.00000363, -0.00001757, -0.00001072, 0.00001604, -0.00000231, -0.00000280, 0.00002884, 0.00001075, -0.00001181, 0.00000984, 0.00002646, 0.00001055, -0.00000860, 0.00001176, 0.00001024, -0.00004153, -0.00001316, 0.00004028, -0.00000605, -0.00001960, 0.00001170, -0.00001402, -0.00000633, 0.00002226, -0.00000400, -0.00000768, 0.00000242, -0.00000755, 0.00000476, 0.00000447, -0.00001095, 0.00000195, 0.00001222, 0.00000449, -0.00000357, -0.00000577, -0.00000452, -0.00000284, 0.00000943, 0.00000511, -0.00001510, 0.00000013, 0.00000892, -0.00000582, 0.00000467, 0.00000025, -0.00000688, 0.00001438, -0.00000376, -0.00001767, 0.00001220, 0.00000594, -0.00000468, 0.00000525, -0.00000926, -0.00000499, 0.00000954, -0.00000067, 0.00000458, 0.00000218, -0.00001209, 0.00000629, 0.00001320, -0.00000435, -0.00000281, -0.00000120, -0.00000855, 0.00000056, 0.00000972, -0.00000352, -0.00000901, 0.00000313, -0.00000759, -0.00001053, 0.00001524, -0.00000794, -0.00004013, -0.00000802, 0.00000875, 0.00001240, 0.00004574, 0.00002999, -0.00000041, 0.00003201, 0.00004793, 0.00001410, -0.00001651, -0.00002805, -0.00001524, -0.00000105, -0.00000967, -0.00001406, -0.00000626, -0.00000263, -0.00000141, -0.00000026, 0.00000050, 0.00000033, -0.00000181, },
		{ -0.00000057, -0.00000269, -0.00000172, -0.00000036, -0.00000242, -0.00000271, 0.00000322, 0.00000046, -0.00000820, -0.00000340, 0.00000119, -0.00000255, -0.00000224, -0.00000466, -0.00000668, 0.00000078, 0.00000315, -0.00000049, -0.00000047, -0.00000262, -0.00000373, 0.00000057, -0.00000251, -0.00000531, 0.00000525, 0.00000383, -0.00000631, 0.00000324, 0.00000109, -0.00001332, -0.00000543, -0.00001195, -0.00001850, 0.00001119, 0.00000820, -0.00000999, 0.00001785, 0.00002111, -0.00000690, -0.00000151, 0.00001939, 0.00002161, -0.00000316, -0.00000418, 0.00002285, -0.00001819, -0.00004262, 0.00002750, 0.00002384, -0.00002653, 0.00000274, 0.00000121, -0.00001966, 0.00001572, 0.00001145, -0.00001214, 0.00000095, -0.00000326, -0.00000342, 0.00000757, -0.00000551, -0.00000579, 0.00001012, 0.00000948, -0.00000130, -0.00000635, -0.00000406, -0.00000481, 0.00000419, 0.00001138, -0.00001071, -0.00000932, 0.00001156, -0.00000227, -0.00000139, 0.00000654, -0.00000905, 0.00000712, 0.00000912, -0.00001970, 0.00000094, 0.00001424, -0.00000537, 0.00000285, -0.00000148, -0.00001126, 0.00000689, 0.00000438, -0.00000053, 0.00000607, -0.00000780, -0.00000384, 0.00001436, 0.00000335, -0.00000567, -0.00000025, -0.00000554, -0.00000676, 0.00000746, 0.00000560, -0.00001083, -0.00000173, 0.00000197, -0.00001610, 0.00000342, 0.00001179, -0.00003207, -0.00002913, 0.00000456, 0.00000776, 0.00003404, 0.00004612, 0.00000904, 0.00001543, 0.00004984, 0.00003317, -0.00000791, -0.00003005, -0.00002235, -0.00000371, -0.00000544, -0.00001341, -0.00000972, -0.00000452, -0.00000155, -0.00000086, -0.00000047, 0.00000106, },
		{ 0.00000063, 0.00000091, -0.00000207, -0.00000254, -0.00000063, -0.00000100, -0.00000372, 0.00000070, 0.00000350, -0.00000612, -0.00000671, 0.00000132, -0.00000121, -0.00000281, -0.00000286, -0.00000729, -0.00000234, 0.00000375, 0.00000043, -0.00000048, -0.00000116, -0.00000443, -0.00000091, -0.00000021, -0.00000596, 0.00000116, 0.00000703, -0.00000436, -0.00000142, 0.00000531, -0.00000924, -0.00000841, -0.00000693, -0.00002121, 0.00000113, 0.00001630, -0.00000834, 0.00000462, 0.00002398, 0.00000163, -0.00000571, 0.00000875, 0.00002240, 0.00001149, -0.00001330, 0.00001381, 0.00001255, -0.00004692, -0.00000781, 0.00004345, -0.00001118, -0.00001590, 0.00001173, -0.00001776, -0.00000226, 0.00002227, -0.00000589, -0.00000643, 0.00000189, -0.00000647, 0.00000324, 0.00000163, -0.00000817, 0.00000351, 0.00001237, 0.00000355, -0.00000660, -0.00000465, -0.00000419, -0.00000168, 0.00001244, -0.00000143, -0.00001633, 0.00000747, 0.00000550, -0.00000673, 0.00000757, -0.00000400, -0.00000318, 0.00001548, -0.00001131, -0.00001253, 0.00001596, 0.00000089, -0.00000314, 0.00000452, -0.00001068, -0.00000091, 0.00000917, -0.00000137, 0.00000380, -0.00000073, -0.00000878, 0.00000827, 0.00001077, -0.00000372, -0.00000289, -0.00000158, -0.00000872, -0.00000059, 0.00001101, -0.00000453, -0.00000986, 0.00000602, -0.00000947, -0.00001229, 0.00001610, -0.00001020, -0.00004033, -0.00001072, 0.00000615, 0.00002118, 0.00004986, 0.00002762, 0.00000716, 0.00003976, 0.00004884, 0.00000854, -0.00002878, -0.00002979, -0.00000869, -0.00000312, -0.00001096, -0.00001136, -0.00000745, -0.00000269, -0.00000066, -0.00000158, },
		{ -0.00000188, -0.00000083, 0.00000173, -0.00000067, -0.00000276, -0.00000162, -0.00000006, -0.00000295, -0.00000237, 0.00000418, -0.00000201, -0.00000900, -0.00000059, 0.00000090, -0.00000333, -0.00000188, -0.00000608, -0.00000563, 0.00000293, 0.00000197, -0.00000081, 0.00000005, -0.00000377, -0.00000304, 0.00000089, -0.00000453, -0.00000294, 0.00000734, -0.00000010, -0.00000514, 0.00000545, -0.00000319, -0.00001030, -0.00000411, -0.00001895, -0.00001043, 0.00001828, -0.00000017, -0.00000588, 0.00002036, 0.00000917, -0.00000487, 0.00000224, 0.00001370, 0.00002262, -0.00000600, -0.00000650, 0.00002779, -0.00002250, -0.00004238, 0.00003463, 0.00001966, -0.00002650, 0.00000643, -0.00000336, -0.00001821, 0.00001871, 0.00000881, -0.00001168, 0.00000124, -0.00000299, -0.00000337, 0.00000420, -0.00000458, -0.00000309, 0.00001038, 0.00000941, -0.00000431, -0.00000682, -0.00000272, -0.00000493, 0.00000796, 0.00000820, -0.00001639, -0.00000257, 0.00001234, -0.00000636, 0.00000235, 0.00000387, -0.00000925, 0.00001203, 0.00000235, -0.00001989, 0.00000823, 0.00001038, -0.00000651, 0.00000442, -0.00000433, -0.00000836, 0.00000878, 0.00000277, -0.00000046, 0.00000318, -0.00000704, 0.00000001, 0.00001273, 0.00000265, -0.00000503, -0.00000086, -0.00000552, -0.00000798, 0.00000812, 0.00000584, -0.00001282, 0.00000060, 0.00000215, -0.00001886, 0.00000435, 0.00001014, -0.00003327, -0.00002969, -0.00000054, 0.00001210, 0.00004271, 0.00004476, 0.00001223, 0.00002508, 0.00005407, 0.00002980, -0.00002037, -0.00003674, -0.00001592, -0.00000281, -0.00000840, -0.00001099, -0.00000974, -0.00000537, -0.00000054, },
		{ -0.00000176, -0.00000122, -0.00000238, 0.00000131, 0.00000086, -0.00000204, -0.00000258, -0.00000036, -0.00000108, -0.00000421, 0.00000205, 0.00000240, -0.00000863, -0.00000437, 0.00000258, -0.00000272, -0.00000230, -0.00000370, -0.00000772, 0.00000027, 0.00000366, -0.00000078, 0.00000030, -0.00000194, -0.00000467, 0.00000023, -0.00000204, -0.00000554, 0.00000486, 0.00000447, -0.00000611, 0.00000229, 0.00000238, -0.00000978, -0.00000386, -0.00001323, -0.00001964, 0.00001288, 0.00001054, -0.00001067, 0.00001241, 0.00001454, -0.00000316, 0.00000132, 0.00000443, 0.00002083, 0.00001204, -0.00001854, 0.00001761, 0.00001240, -0.00005134, -0.00000063, 0.00004366, -0.00001512, -0.00001107, 0.00000960, -0.00002013, 0.00000219, 0.00002108, -0.00000725, -0.00000537, 0.00000180, -0.00000584, 0.00000067, 0.00000047, -0.00000522, 0.00000443, 0.00001267, 0.00000141, -0.00000873, -0.00000303, -0.00000444, 0.00000104, 0.00001310, -0.00000886, -0.00001338, 0.00001267, 0.00000061, -0.00000491, 0.00000814, -0.00000743, 0.00000226, 0.00001279, -0.00001660, -0.00000510, 0.00001600, -0.00000317, -0.00000068, 0.00000241, -0.00001041, 0.00000282, 0.00000790, -0.00000171, 0.00000213, -0.00000230, -0.00000488, 0.00000869, 0.00000893, -0.00000308, -0.00000321, -0.00000167, -0.00000966, -0.00000080, 0.00001220, -0.00000658, -0.00000911, 0.00000801, -0.00001221, -0.00001215, 0.00001576, -0.00001271, -0.00003966, -0.00001510, 0.00000472, 0.00003071, 0.00005285, 0.00002735, 0.00001555, 0.00004765, 0.00004875, -0.00000293, -0.00004005, -0.00002569, -0.00000457, -0.00000649, -0.00000978, -0.00001016, -0.00000869, },
		{ -0.00001095, -0.00000478, -0.00000050, -0.00000312, -0.00000015, 0.00000176, -0.00000070, -0.00000261, -0.00000170, 0.00000036, -0.00000370, -0.00000168, 0.00000475, -0.00000506, -0.00000844, 0.00000222, -0.00000045, -0.00000357, -0.00000168, -0.00000748, -0.00000369, 0.00000434, 0.00000040, -0.00000049, 0.00000004, -0.00000475, -0.00000186, 0.00000009, -0.00000583, 0.00000073, 0.00000741, -0.00000389, -0.00000224, 0.00000545, -0.00000657, -0.00000553, -0.00000656, -0.00002332, 0.00000133, 0.00001913, -0.00000826, 0.00000209, 0.00001783, -0.00000158, 0.00000122, 0.00000244, 0.00000873, 0.00002517, -0.00001058, -0.00000692, 0.00003171, -0.00002808, -0.00003922, 0.00003991, 0.00001450, -0.00002405, 0.00000838, -0.00000806, -0.00001519, 0.00002048, 0.00000615, -0.00001100, 0.00000159, -0.00000253, -0.00000469, 0.00000169, -0.00000252, -0.00000126, 0.00001084, 0.00000843, -0.00000741, -0.00000601, -0.00000220, -0.00000390, 0.00001115, 0.00000222, -0.00001858, 0.00000479, 0.00000979, -0.00000800, 0.00000554, -0.00000020, -0.00000646, 0.00001427, -0.00000513, -0.00001600, 0.00001291, 0.00000528, -0.00000549, 0.00000456, -0.00000645, -0.00000453, 0.00000927, 0.00000147, -0.00000099, 0.00000060, -0.00000475, 0.00000253, 0.00001086, 0.00000252, -0.00000491, -0.00000106, -0.00000609, -0.00000894, 0.00000958, 0.00000473, -0.00001388, 0.00000335, 0.00000060, -0.00002010, 0.00000534, 0.00000717, -0.00003322, -0.00003134, -0.00000593, 0.00001808, 0.00005045, 0.00004439, 0.00001706, 0.00003500, 0.00005848, 0.00002096, -0.00003536, -0.00003695, -0.00000924, -0.00000514, -0.00000896, -0.00000888, },
		{ -0.00000732, -0.00001087, -0.00000874, -0.00000096, -0.00000280, -0.00000196, 0.00000172, 0.00000047, -0.00000155, -0.00000288, 0.00000007, -0.00000135, -0.00000461, 0.00000357, 0.00000030, -0.00001032, -0.00000109, 0.00000260, -0.00000412, -0.00000144, -0.00000503, -0.00000725, 0.00000278, 0.00000266, -0.00000137, 0.00000092, -0.00000305, -0.00000413, 0.00000058, -0.00000410, -0.00000328, 0.00000744, 0.00000061, -0.00000576, 0.00000492, -0.00000153, -0.00000751, -0.00000185, -0.00002022, -0.00001247, 0.00002095, 0.00000102, -0.00000779, 0.00001760, 0.00000234, -0.00000261, 0.00000685, -0.00000181, 0.00002202, 0.00001110, -0.00002318, 0.00002237, 0.00000975, -0.00005362, 0.00000694, 0.00004124, -0.00001692, -0.00000635, 0.00000583, -0.00002070, 0.00000631, 0.00001906, -0.00000814, -0.00000439, 0.00000211, -0.00000624, -0.00000210, 0.00000096, -0.00000277, 0.00000516, 0.00001266, -0.00000163, -0.00000950, -0.00000164, -0.00000453, 0.00000469, 0.00001063, -0.00001500, -0.00000734, 0.00001446, -0.00000375, -0.00000151, 0.00000616, -0.00000853, 0.00000750, 0.00000705, -0.00001811, 0.00000233, 0.00001289, -0.00000515, 0.00000147, -0.00000037, -0.00000840, 0.00000552, 0.00000626, -0.00000205, -0.00000007, -0.00000222, -0.00000134, 0.00000803, 0.00000784, -0.00000280, -0.00000339, -0.00000188, -0.00001104, 0.00000015, 0.00001253, -0.00000882, -0.00000701, 0.00000837, -0.00001460, -0.00001074, 0.00001378, -0.00001455, -0.00003889, -0.00002085, 0.00000472, 0.00004023, 0.00005582, 0.00002891, 0.00002455, 0.00005656, 0.00004472, -0.00001997, -0.00004625, -0.00001797, -0.00000444, -0.00000880, },
		{ -0.00000864, -0.00000694, -0.00000855, -0.00001193, -0.00000328, -0.00000196, -0.00000347, 0.00000093, 0.00000101, -0.00000019, -0.00000263, -0.00000179, 0.00000083, -0.00000473, -0.00000047, 0.00000453, -0.00000825, -0.00000629, 0.00000423, -0.00000247, -0.00000305, -0.00000202, -0.00000844, -0.00000111, 0.00000468, -0.00000098, 0.00000019, -0.00000059, -0.00000503, -0.00000083, -0.00000169, -0.00000551, 0.00000459, 0.00000533, -0.00000627, 0.00000119, 0.00000334, -0.00000768, -0.00000100, -0.00001208, -0.00002284, 0.00001347, 0.00001364, -0.00001278, 0.00001145, 0.00001004, -0.00000856, 0.00000925, -0.00000105, 0.00000593, 0.00002789, -0.00001631, -0.00000526, 0.00003365, -0.00003355, -0.00003375, 0.00004252, 0.00000975, -0.00001993, 0.00000822, -0.00001194, -0.00001123, 0.00002097, 0.00000371, -0.00001021, 0.00000225, -0.00000250, -0.00000698, 0.00000064, -0.00000015, -0.00000005, 0.00001142, 0.00000630, -0.00000979, -0.00000451, -0.00000228, -0.00000145, 0.00001234, -0.00000500, -0.00001670, 0.00001057, 0.00000536, -0.00000696, 0.00000676, -0.00000394, -0.00000158, 0.00001293, -0.00001101, -0.00000941, 0.00001403, 0.00000074, -0.00000317, 0.00000317, -0.00000710, -0.00000063, 0.00000865, 0.00000047, -0.00000225, -0.00000095, -0.00000170, 0.00000367, 0.00000934, 0.00000247, -0.00000501, -0.00000099, -0.00000739, -0.00000907, 0.00001109, 0.00000256, -0.00001351, 0.00000542, -0.00000190, -0.00001955, 0.00000532, 0.00000373, -0.00003210, -0.00003445, -0.00001095, 0.00002496, 0.00005779, 0.00004545, 0.00002281, 0.00004638, 0.00006112, 0.00000478, -0.00004855, -0.00003065, -0.00000554, },
		{ -0.00001035, -0.00000780, -0.00000831, -0.00000545, -0.00001270, -0.00000716, -0.00000147, -0.00000437, -0.00000037, 0.00000115, 0.00000036, -0.00000086, -0.00000351, 0.00000082, -0.00000206, -0.00000460, 0.00000495, -0.00000275, -0.00001052, 0.00000242, 0.00000124, -0.00000481, -0.00000069, -0.00000658, -0.00000569, 0.00000457, 0.00000119, -0.00000126, 0.00000095, -0.00000389, -0.00000308, -0.00000029, -0.00000533, 0.00000031, 0.00000796, -0.00000322, -0.00000366, 0.00000555, -0.00000476, -0.00000343, -0.00000338, -0.00002505, -0.00000135, 0.00002294, -0.00000879, -0.00000024, 0.00001792, -0.00000979, 0.00000273, 0.00000912, -0.00000814, 0.00002532, 0.00000821, -0.00002625, 0.00002695, 0.00000529, -0.00005320, 0.00001347, 0.00003717, -0.00001634, -0.00000280, 0.00000129, -0.00001955, 0.00000957, 0.00001656, -0.00000866, -0.00000324, 0.00000247, -0.00000784, -0.00000417, 0.00000257, -0.00000104, 0.00000602, 0.00001181, -0.00000492, -0.00000896, -0.00000092, -0.00000370, 0.00000803, 0.00000535, -0.00001806, -0.00000016, 0.00001288, -0.00000605, 0.00000179, 0.00000250, -0.00000672, 0.00001061, 0.00000005, -0.00001558, 0.00000767, 0.00000811, -0.00000487, 0.00000239, -0.00000283, -0.00000517, 0.00000693, 0.00000467, -0.00000278, -0.00000212, -0.00000071, 0.00000121, 0.00000699, 0.00000723, -0.00000294, -0.00000321, -0.00000264, -0.00001225, 0.00000193, 0.00001161, -0.00001034, -0.00000437, 0.00000709, -0.00001569, -0.00000910, 0.00001041, -0.00001524, -0.00003874, -0.00002738, 0.00000586, 0.00004963, 0.00005968, 0.00003150, 0.00003502, 0.00006597, 0.00003341, -0.00003958, -0.00004471, },
		{ -0.00005534, -0.00002038, -0.00000652, -0.00001068, -0.00000366, -0.00001040, -0.00001136, -0.00000215, -0.00000452, -0.00000223, 0.00000140, -0.00000009, 0.00000097, -0.00000316, -0.00000155, 0.00000107, -0.00000583, 0.00000127, 0.00000310, -0.00001077, -0.00000284, 0.00000477, -0.00000443, -0.00000195, -0.00000308, -0.00000843, 0.00000151, 0.00000397, -0.00000169, 0.00000065, -0.00000153, -0.00000443, -0.00000075, -0.00000364, -0.00000325, 0.00000728, 0.00000191, -0.00000663, 0.00000372, 0.00000021, -0.00000613, 0.00000133, -0.00001859, -0.00001691, 0.00002243, 0.00000383, -0.00001191, 0.00001908, -0.00000154, -0.00001027, 0.00001671, -0.00000725, 0.00000595, 0.00002952, -0.00002217, -0.00000202, 0.00003326, -0.00003766, -0.00002714, 0.00004247, 0.00000652, -0.00001523, 0.00000617, -0.00001441, -0.00000702, 0.00002038, 0.00000156, -0.00000920, 0.00000316, -0.00000343, -0.00000948, 0.00000110, 0.00000190, 0.00000096, 0.00001172, 0.00000318, -0.00001093, -0.00000302, -0.00000237, 0.00000198, 0.00001067, -0.00001142, -0.00001153, 0.00001331, 0.00000094, -0.00000409, 0.00000553, -0.00000576, 0.00000361, 0.00000840, -0.00001359, -0.00000226, 0.00001187, -0.00000197, -0.00000078, 0.00000079, -0.00000606, 0.00000260, 0.00000741, -0.00000045, -0.00000397, -0.00000110, 0.00000124, 0.00000379, 0.00000836, 0.00000214, -0.00000497, -0.00000103, -0.00000915, -0.00000808, 0.00001190, 0.00000005, -0.00001184, 0.00000610, -0.00000426, -0.00001769, 0.00000380, 0.00000068, -0.00003053, -0.00003886, -0.00001536, 0.00003217, 0.00006568, 0.00004757, 0.00002966, 0.00006000, 0.00005842, -0.00001816, },
		{ 0.00001274, -0.00005677, -0.00003546, -0.00000643, -0.00001234, -0.00000487, -0.00000581, -0.00001408, -0.00000459, -0.00000384, -0.00000467, 0.00000155, -0.00000047, 0.00000111, -0.00000049, -0.00000416, 0.00000180, -0.00000333, -0.00000384, 0.00000559, -0.00000630, -0.00000883, 0.00000515, -0.00000103, -0.00000448, -0.00000059, -0.00000780, -0.00000328, 0.00000518, -0.00000009, -0.00000081, 0.00000027, -0.00000387, -0.00000232, -0.00000222, -0.00000475, 0.00000404, 0.00000632, -0.00000571, -0.00000100, 0.00000420, -0.00000594, 0.00000060, -0.00000796, -0.00002545, 0.00001060, 0.00001847, -0.00001548, 0.00000972, 0.00001264, -0.00001904, 0.00001103, 0.00000748, -0.00001292, 0.00002952, 0.00000365, -0.00002732, 0.00003036, 0.00000015, -0.00005036, 0.00001798, 0.00003276, -0.00001381, -0.00000102, -0.00000309, -0.00001709, 0.00001175, 0.00001385, -0.00000883, -0.00000177, 0.00000231, -0.00001032, -0.00000493, 0.00000458, 0.00000012, 0.00000700, 0.00000980, -0.00000768, -0.00000755, -0.00000080, -0.00000160, 0.00000971, -0.00000145, -0.00001728, 0.00000598, 0.00000911, -0.00000582, 0.00000350, -0.00000126, -0.00000264, 0.00001049, -0.00000600, -0.00001010, 0.00000974, 0.00000344, -0.00000300, 0.00000176, -0.00000415, -0.00000150, 0.00000717, 0.00000323, -0.00000403, -0.00000336, 0.00000159, 0.00000259, 0.00000611, 0.00000669, -0.00000325, -0.00000275, -0.00000411, -0.00001271, 0.00000391, 0.00000962, -0.00001063, -0.00000217, 0.00000481, -0.00001515, -0.00000821, 0.00000635, -0.00001484, -0.00003958, -0.00003416, 0.00000751, 0.00005950, 0.00006466, 0.00003468, 0.00004843, 0.00007331, },
		{ 0.00007504, 0.00004645, -0.00004432, -0.00005269, -0.00001040, -0.00001160, -0.00000933, -0.00000126, -0.00001351, -0.00000877, -0.00000275, -0.00000692, 0.00000044, 0.00000051, -0.00000071, 0.00000242, -0.00000420, -0.00000074, 0.00000067, -0.00000654, 0.00000315, 0.00000032, -0.00001172, 0.00000112, 0.00000362, -0.00000553, -0.00000090, -0.00000463, -0.00000713, 0.00000339, 0.00000272, -0.00000161, 0.00000036, -0.00000207, -0.00000330, -0.00000213, -0.00000457, 0.00000038, 0.00000781, -0.00000147, -0.00000533, 0.00000440, -0.00000233, -0.00000284, 0.00000034, -0.00002334, -0.00000693, 0.00002565, -0.00000693, -0.00000579, 0.00002235, -0.00001425, -0.00000674, 0.00002153, -0.00001451, 0.00000853, 0.00002933, -0.00002716, 0.00000189, 0.00003090, -0.00003968, -0.00002075, 0.00004044, 0.00000532, -0.00001104, 0.00000288, -0.00001528, -0.00000315, 0.00001895, -0.00000027, -0.00000781, 0.00000399, -0.00000550, -0.00001133, 0.00000266, 0.00000332, 0.00000210, 0.00001121, -0.00000036, -0.00001071, -0.00000202, -0.00000177, 0.00000534, 0.00000622, -0.00001522, -0.00000481, 0.00001276, -0.00000187, -0.00000094, 0.00000246, -0.00000488, 0.00000725, 0.00000221, -0.00001237, 0.00000335, 0.00000772, -0.00000244, 0.00000067, -0.00000167, -0.00000363, 0.00000474, 0.00000594, -0.00000162, -0.00000558, 0.00000001, 0.00000344, 0.00000347, 0.00000774, 0.00000147, -0.00000454, -0.00000158, -0.00001084, -0.00000620, 0.00001159, -0.00000197, -0.00000954, 0.00000528, -0.00000557, -0.00001543, 0.00000090, -0.00000150, -0.00002909, -0.00004413, -0.00001951, 0.00003977, 0.00007487, 0.00004996, 0.00003897, },
		{ 0.00003905, 0.00006569, 0.00007422, -0.00001659, -0.00006616, -0.00002142, -0.00000822, -0.00001505, 0.00000002, -0.00000885, -0.00001347, -0.00000254, -0.00000755, -0.00000290, 0.00000259, -0.00000252, 0.00000279, -0.00000089, -0.00000432, 0.00000257, -0.00000491, -0.00000218, 0.00000484, -0.00000942, -0.00000541, 0.00000618, -0.00000332, -0.00000332, -0.00000150, -0.00000797, -0.00000071, 0.00000457, -0.00000055, -0.00000070, -0.00000073, -0.00000273, -0.00000277, -0.00000428, -0.00000202, 0.00000643, 0.00000335, -0.00000623, 0.00000079, 0.00000194, -0.00000466, 0.00000262, -0.00001366, -0.00002068, 0.00001988, 0.00000929, -0.00001691, 0.00001894, 0.00000279, -0.00002347, 0.00001987, 0.00000249, -0.00001514, 0.00003333, -0.00000180, -0.00002654, 0.00003203, -0.00000447, -0.00004595, 0.00002015, 0.00002914, -0.00001021, -0.00000107, -0.00000662, -0.00001388, 0.00001284, 0.00001110, -0.00000854, -0.00000007, 0.00000115, -0.00001299, -0.00000425, 0.00000633, 0.00000112, 0.00000774, 0.00000670, -0.00000930, -0.00000592, -0.00000083, 0.00000148, 0.00000882, -0.00000789, -0.00001314, 0.00000951, 0.00000489, -0.00000365, 0.00000296, -0.00000352, 0.00000221, 0.00000720, -0.00000931, -0.00000360, 0.00000857, 0.00000027, -0.00000069, -0.00000005, -0.00000393, 0.00000187, 0.00000659, 0.00000182, -0.00000561, -0.00000344, 0.00000393, 0.00000308, 0.00000560, 0.00000588, -0.00000336, -0.00000236, -0.00000611, -0.00001212, 0.00000538, 0.00000716, -0.00000971, -0.00000111, 0.00000251, -0.00001335, -0.00000858, 0.00000236, -0.00001383, -0.00004128, -0.00004115, 0.00000928, 0.00007076, 0.00007015, },
		{ 0.00008531, 0.00005232, 0.00005268, 0.00008829, 0.00002264, -0.00006783, -0.00004053, -0.00000493, -0.00001816, -0.00000436, -0.00000163, -0.00001602, -0.00000493, -0.00000568, -0.00000770, 0.00000351, -0.00000181, -0.00000016, 0.00000342, -0.00000534, 0.00000053, -0.00000059, -0.00000627, 0.00000444, -0.00000337, -0.00001045, 0.00000433, 0.00000125, -0.00000529, -0.00000066, -0.00000596, -0.00000471, 0.00000388, 0.00000175, -0.00000115, -0.00000088, -0.00000144, -0.00000254, -0.00000478, -0.00000336, 0.00000404, 0.00000607, -0.00000342, -0.00000356, 0.00000326, -0.00000289, 0.00000040, -0.00000355, -0.00002405, 0.00000421, 0.00002253, -0.00001516, 0.00000358, 0.00002008, -0.00002518, 0.00000064, 0.00002262, -0.00002109, 0.00001272, 0.00002724, -0.00003061, 0.00000552, 0.00002741, -0.00003956, -0.00001561, 0.00003750, 0.00000600, -0.00000808, -0.00000083, -0.00001474, 0.00000005, 0.00001692, -0.00000168, -0.00000594, 0.00000418, -0.00000842, -0.00001193, 0.00000463, 0.00000424, 0.00000339, 0.00000953, -0.00000356, -0.00000947, -0.00000159, -0.00000009, 0.00000737, 0.00000007, -0.00001549, 0.00000140, 0.00000983, -0.00000239, 0.00000102, -0.00000104, -0.00000165, 0.00000807, -0.00000360, -0.00000810, 0.00000608, 0.00000327, -0.00000117, 0.00000068, -0.00000334, -0.00000046, 0.00000577, 0.00000441, -0.00000315, -0.00000653, 0.00000188, 0.00000469, 0.00000323, 0.00000710, 0.00000068, -0.00000380, -0.00000285, -0.00001190, -0.00000398, 0.00001024, -0.00000299, -0.00000748, 0.00000351, -0.00000549, -0.00001365, -0.00000276, -0.00000283, -0.00002804, -0.00004991, -0.00002398, 0.00004852, },
		{ 0.00001145, 0.00008404, 0.00007524, 0.00004653, 0.00008581, 0.00006384, -0.00005048, -0.00006404, -0.00000753, -0.00001514, -0.00001363, 0.00000388, -0.00001348, -0.00001031, -0.00000259, -0.00001112, 0.00000070, 0.00000185, -0.00000398, 0.00000487, -0.00000207, -0.00000364, 0.00000251, -0.00000615, 0.00000001, 0.00000233, -0.00001091, -0.00000116, 0.00000513, -0.00000456, -0.00000210, -0.00000323, -0.00000655, 0.00000106, 0.00000331, 0.00000002, -0.00000163, -0.00000120, -0.00000086, -0.00000509, -0.00000508, 0.00000223, 0.00000637, 0.00000058, -0.00000516, 0.00000074, 0.00000038, -0.00000177, 0.00000151, -0.00001781, -0.00001191, 0.00002389, -0.00000139, -0.00001286, 0.00002460, -0.00000927, -0.00002260, 0.00002699, -0.00000446, -0.00001467, 0.00003580, -0.00000721, -0.00002454, 0.00003200, -0.00000778, -0.00004113, 0.00002037, 0.00002697, -0.00000652, -0.00000255, -0.00000892, -0.00001048, 0.00001294, 0.00000845, -0.00000761, 0.00000150, -0.00000119, -0.00001501, -0.00000250, 0.00000751, 0.00000228, 0.00000773, 0.00000300, -0.00000955, -0.00000459, -0.00000040, 0.00000465, 0.00000529, -0.00001219, -0.00000714, 0.00000996, 0.00000183, -0.00000090, 0.00000056, -0.00000336, 0.00000601, 0.00000200, -0.00000915, 0.00000187, 0.00000520, -0.00000081, 0.00000098, -0.00000226, -0.00000230, 0.00000443, 0.00000556, 0.00000021, -0.00000707, -0.00000241, 0.00000571, 0.00000318, 0.00000532, 0.00000474, -0.00000307, -0.00000244, -0.00000815, -0.00001060, 0.00000590, 0.00000498, -0.00000817, -0.00000132, 0.00000100, -0.00001109, -0.00001011, -0.00000113, -0.00001265, -0.00004345, -0.00004881, },
		{ -0.00005635, -0.00002907, 0.00005943, 0.00009611, 0.00005563, 0.00007205, 0.00009443, -0.00001241, -0.00008224, -0.00002219, -0.00000644, -0.00002265, 0.00000252, -0.00000532, -0.00001612, -0.00000161, -0.00001035, -0.00000591, 0.00000550, -0.00000474, 0.00000166, 0.00000337, -0.00000586, 0.00000164, -0.00000248, -0.00000469, 0.00000417, -0.00000694, -0.00000695, 0.00000560, -0.00000125, -0.00000386, -0.00000180, -0.00000612, -0.00000192, 0.00000286, 0.00000187, -0.00000114, -0.00000262, 0.00000074, -0.00000335, -0.00000766, 0.00000043, 0.00000606, 0.00000300, -0.00000327, -0.00000284, 0.00000138, -0.00000106, 0.00000162, -0.00000854, -0.00001987, 0.00001309, 0.00001455, -0.00001851, 0.00001316, 0.00001283, -0.00003209, 0.00000963, 0.00002007, -0.00002574, 0.00001735, 0.00002373, -0.00003241, 0.00000823, 0.00002376, -0.00003786, -0.00001220, 0.00003470, 0.00000795, -0.00000663, -0.00000427, -0.00001318, 0.00000238, 0.00001447, -0.00000249, -0.00000371, 0.00000327, -0.00001153, -0.00001111, 0.00000637, 0.00000500, 0.00000453, 0.00000665, -0.00000575, -0.00000776, -0.00000143, 0.00000251, 0.00000716, -0.00000609, -0.00001244, 0.00000548, 0.00000609, -0.00000098, 0.00000099, -0.00000344, 0.00000265, 0.00000589, -0.00000723, -0.00000249, 0.00000572, -0.00000003, 0.00000085, -0.00000058, -0.00000370, 0.00000273, 0.00000592, 0.00000279, -0.00000488, -0.00000647, 0.00000383, 0.00000521, 0.00000325, 0.00000616, 0.00000011, -0.00000307, -0.00000469, -0.00001200, -0.00000208, 0.00000838, -0.00000300, -0.00000631, 0.00000165, -0.00000434, -0.00001281, -0.00000656, -0.00000369, -0.00002719, },
		{ -0.00001140, -0.00004589, -0.00005782, 0.00001496, 0.00009909, 0.00007964, 0.00005973, 0.00010500, 0.00003883, -0.00008205, -0.00005010, 0.00000114, -0.00002418, -0.00000770, 0.00000443, -0.00001754, -0.00000546, -0.00000554, -0.00001275, 0.00000486, -0.00000077, -0.00000392, 0.00000641, -0.00000345, -0.00000220, 0.00000119, -0.00000621, 0.00000183, -0.00000165, -0.00000979, 0.00000246, 0.00000244, -0.00000402, -0.00000188, -0.00000496, -0.00000357, 0.00000113, 0.00000239, 0.00000119, -0.00000394, -0.00000014, 0.00000066, -0.00000915, -0.00000326, 0.00000648, 0.00000374, -0.00000049, -0.00000379, -0.00000113, 0.00000096, 0.00000092, -0.00000243, -0.00001815, -0.00000187, 0.00002181, -0.00001014, -0.00000503, 0.00002513, -0.00002069, -0.00001735, 0.00003096, -0.00001173, -0.00001208, 0.00003656, -0.00001186, -0.00002207, 0.00003081, -0.00000953, -0.00003686, 0.00001946, 0.00002636, -0.00000351, -0.00000482, -0.00000998, -0.00000729, 0.00001218, 0.00000609, -0.00000596, 0.00000238, -0.00000445, -0.00001578, -0.00000033, 0.00000819, 0.00000367, 0.00000653, -0.00000058, -0.00000861, -0.00000376, 0.00000089, 0.00000682, -0.00000003, -0.00001332, -0.00000118, 0.00000797, 0.00000074, 0.00000103, -0.00000252, -0.00000089, 0.00000744, -0.00000329, -0.00000600, 0.00000490, 0.00000120, -0.00000006, 0.00000140, -0.00000401, 0.00000027, 0.00000604, 0.00000426, -0.00000165, -0.00000791, -0.00000072, 0.00000672, 0.00000332, 0.00000494, 0.00000349, -0.00000240, -0.00000320, -0.00000970, -0.00000866, 0.00000553, 0.00000355, -0.00000677, -0.00000230, 0.00000057, -0.00000914, -0.00001227, -0.00000422, },
		{ -0.00001019, -0.00000444, -0.00002610, -0.00006414, -0.00003435, 0.00007315, 0.00010609, 0.00006245, 0.00009614, 0.00008742, -0.00005413, -0.00008290, -0.00000339, -0.00001447, -0.00002225, 0.00000839, -0.00001118, -0.00001304, -0.00000074, -0.00001500, -0.00000177, 0.00000523, -0.00000697, 0.00000407, 0.00000220, -0.00000515, 0.00000173, -0.00000421, -0.00000203, 0.00000166, -0.00000891, -0.00000213, 0.00000426, -0.00000248, -0.00000217, -0.00000427, -0.00000427, 0.00000001, 0.00000087, 0.00000347, -0.00000256, -0.00000383, 0.00000411, -0.00000659, -0.00000912, 0.00000615, 0.00000515, 0.00000031, -0.00000144, -0.00000422, 0.00000048, 0.00000231, -0.00000033, -0.00001201, -0.00001182, 0.00001708, 0.00000472, -0.00001647, 0.00002006, 0.00000260, -0.00003406, 0.00001799, 0.00001490, -0.00002797, 0.00002138, 0.00001952, -0.00003287, 0.00000987, 0.00002073, -0.00003545, -0.00001041, 0.00003277, 0.00001036, -0.00000653, -0.00000697, -0.00001103, 0.00000380, 0.00001180, -0.00000249, -0.00000154, 0.00000105, -0.00001402, -0.00000919, 0.00000753, 0.00000598, 0.00000500, 0.00000297, -0.00000658, -0.00000614, -0.00000102, 0.00000533, 0.00000451, -0.00001058, -0.00000734, 0.00000677, 0.00000307, 0.00000122, -0.00000084, -0.00000373, 0.00000635, 0.00000167, -0.00000780, 0.00000254, 0.00000310, -0.00000145, 0.00000253, -0.00000247, -0.00000269, 0.00000542, 0.00000544, 0.00000095, -0.00000643, -0.00000545, 0.00000533, 0.00000541, 0.00000338, 0.00000489, -0.00000003, -0.00000273, -0.00000666, -0.00001123, -0.00000090, 0.00000666, -0.00000252, -0.00000611, 0.00000041, -0.00000286, -0.00001277, },
		{ -0.00000785, -0.00001452, -0.00000729, -0.00000972, -0.00004892, -0.00006851, 0.00002105, 0.00011467, 0.00008464, 0.00008055, 0.00011686, -0.00000060, -0.00010353, -0.00002794, 0.00000142, -0.00003116, 0.00000088, 0.00000087, -0.00001881, -0.00000100, -0.00001076, -0.00001124, 0.00000777, -0.00000426, -0.00000217, 0.00000653, -0.00000404, -0.00000075, -0.00000125, -0.00000442, 0.00000200, -0.00000599, -0.00000580, 0.00000385, -0.00000062, -0.00000175, -0.00000356, -0.00000544, 0.00000016, -0.00000085, 0.00000292, 0.00000192, -0.00000723, 0.00000310, 0.00000038, -0.00001383, 0.00000182, 0.00000877, -0.00000043, 0.00000093, -0.00000380, -0.00000364, 0.00000412, 0.00000138, -0.00000740, -0.00001375, 0.00000613, 0.00001494, -0.00001448, 0.00000375, 0.00002067, -0.00002908, -0.00000962, 0.00003145, -0.00001795, -0.00000835, 0.00003576, -0.00001541, -0.00001973, 0.00002918, -0.00001002, -0.00003370, 0.00001830, 0.00002693, -0.00000157, -0.00000728, -0.00000998, -0.00000459, 0.00001070, 0.00000432, -0.00000376, 0.00000206, -0.00000798, -0.00001508, 0.00000163, 0.00000872, 0.00000503, 0.00000404, -0.00000332, -0.00000698, -0.00000326, 0.00000298, 0.00000708, -0.00000565, -0.00001131, 0.00000312, 0.00000490, 0.00000148, 0.00000126, -0.00000484, 0.00000283, 0.00000612, -0.00000691, -0.00000131, 0.00000507, -0.00000198, 0.00000162, 0.00000056, -0.00000472, 0.00000315, 0.00000676, 0.00000270, -0.00000361, -0.00000785, 0.00000106, 0.00000715, 0.00000364, 0.00000421, 0.00000248, -0.00000169, -0.00000453, -0.00001042, -0.00000689, 0.00000472, 0.00000286, -0.00000608, -0.00000328, 0.00000090, },
		{ 0.00000000, -0.00000172, -0.00001299, -0.00001386, -0.00000500, -0.00002460, -0.00007416, -0.00003863, 0.00008921, 0.00011482, 0.00007627, 0.00012069, 0.00006274, -0.00009451, -0.00006893, 0.00000971, -0.00002567, -0.00001627, 0.00001047, -0.00001663, -0.00000768, -0.00000341, -0.00001744, 0.00000322, 0.00000255, -0.00000727, 0.00000581, 0.00000063, -0.00000323, 0.00000005, -0.00000458, 0.00000068, -0.00000306, -0.00000786, 0.00000230, 0.00000063, -0.00000136, -0.00000160, -0.00000718, -0.00000022, -0.00000012, -0.00000086, 0.00000617, -0.00000599, -0.00000321, 0.00000749, -0.00001234, -0.00000745, 0.00001208, 0.00000105, -0.00000026, 0.00000062, -0.00000769, 0.00000210, 0.00000532, -0.00000521, -0.00001167, -0.00000253, 0.00001527, -0.00000361, -0.00001044, 0.00002246, -0.00000797, -0.00003155, 0.00002408, 0.00000852, -0.00002798, 0.00002426, 0.00001530, -0.00003249, 0.00001069, 0.00001861, -0.00003317, -0.00000965, 0.00003195, 0.00001254, -0.00000735, -0.00000876, -0.00000869, 0.00000430, 0.00000918, -0.00000155, -0.00000003, -0.00000221, -0.00001523, -0.00000679, 0.00000815, 0.00000728, 0.00000432, -0.00000084, -0.00000610, -0.00000488, -0.00000001, 0.00000739, 0.00000009, -0.00001231, -0.00000190, 0.00000568, 0.00000168, 0.00000289, -0.00000348, -0.00000185, 0.00000811, -0.00000297, -0.00000554, 0.00000555, -0.00000039, -0.00000106, 0.00000316, -0.00000420, -0.00000066, 0.00000735, 0.00000449, -0.00000110, -0.00000738, -0.00000386, 0.00000623, 0.00000565, 0.00000333, 0.00000351, 0.00000022, -0.00000301, -0.00000824, -0.00001005, -0.00000042, 0.00000551, -0.00000225, -0.00000640, },
		{ -0.00000612, -0.00000360, 0.00000124, -0.00000698, -0.00001670, -0.00001063, -0.00000704, -0.00005348, -0.00008034, 0.00003044, 0.00012917, 0.00009347, 0.00010838, 0.00011387, -0.00004958, -0.00010892, -0.00000419, -0.00000654, -0.00003252, 0.00000867, -0.00000559, -0.00001616, 0.00000067, -0.00001621, -0.00000673, 0.00000788, -0.00000694, 0.00000045, 0.00000504, -0.00000292, -0.00000049, -0.00000388, -0.00000081, -0.00000061, -0.00000892, 0.00000026, 0.00000194, -0.00000227, 0.00000126, -0.00000732, -0.00000344, 0.00000319, -0.00000440, 0.00000558, 0.00000094, -0.00001033, 0.00000833, -0.00000297, -0.00001720, 0.00000959, 0.00000749, -0.00000420, 0.00000444, -0.00000641, -0.00000491, 0.00000898, -0.00000166, -0.00001078, -0.00000554, 0.00000944, 0.00000610, -0.00001369, 0.00001061, 0.00001284, -0.00003320, -0.00000151, 0.00002905, -0.00002238, -0.00000442, 0.00003386, -0.00001793, -0.00001775, 0.00002773, -0.00000988, -0.00003166, 0.00001756, 0.00002808, -0.00000072, -0.00000946, -0.00000922, -0.00000256, 0.00000870, 0.00000342, -0.00000144, 0.00000032, -0.00001094, -0.00001321, 0.00000299, 0.00000947, 0.00000585, 0.00000055, -0.00000477, -0.00000516, -0.00000273, 0.00000533, 0.00000517, -0.00001003, -0.00000717, 0.00000499, 0.00000219, 0.00000309, -0.00000018, -0.00000536, 0.00000631, 0.00000277, -0.00000789, 0.00000314, 0.00000301, -0.00000350, 0.00000323, -0.00000108, -0.00000419, 0.00000581, 0.00000675, 0.00000086, -0.00000532, -0.00000697, 0.00000248, 0.00000736, 0.00000397, 0.00000314, 0.00000190, -0.00000131, -0.00000599, -0.00001041, -0.00000567, 0.00000406, 0.00000242, },
		{ 0.00000487, -0.00000268, -0.00000640, 0.00000001, -0.00000117, -0.00001298, -0.00001807, -0.00000480, -0.00002315, -0.00008684, -0.00004039, 0.00010601, 0.00012378, 0.00009993, 0.00013817, 0.00001984, -0.00012418, -0.00004407, 0.00001292, -0.00003505, -0.00000666, 0.00000750, -0.00001891, -0.00000245, -0.00000906, -0.00001606, 0.00000677, -0.00000140, -0.00000528, 0.00000565, -0.00000002, -0.00000093, -0.00000347, -0.00000251, 0.00000182, -0.00000894, -0.00000309, 0.00000431, -0.00000390, 0.00000234, -0.00000350, -0.00000899, 0.00000533, -0.00000305, -0.00000088, 0.00000881, -0.00001120, 0.00000025, 0.00000909, -0.00001949, -0.00000167, 0.00001592, -0.00000516, 0.00000233, 0.00000099, -0.00001200, 0.00000681, 0.00000578, -0.00001120, -0.00000648, 0.00000475, 0.00000878, -0.00000789, -0.00000303, 0.00002016, -0.00001646, -0.00002606, 0.00002722, 0.00000227, -0.00002645, 0.00002586, 0.00001144, -0.00003172, 0.00001120, 0.00001726, -0.00003152, -0.00000924, 0.00003206, 0.00001404, -0.00000862, -0.00000966, -0.00000645, 0.00000392, 0.00000699, 0.00000016, 0.00000027, -0.00000589, -0.00001492, -0.00000457, 0.00000858, 0.00000869, 0.00000227, -0.00000404, -0.00000468, -0.00000397, 0.00000163, 0.00000790, -0.00000484, -0.00001116, 0.00000232, 0.00000335, 0.00000193, 0.00000315, -0.00000562, 0.00000137, 0.00000743, -0.00000638, -0.00000168, 0.00000598, -0.00000335, 0.00000039, 0.00000259, -0.00000516, 0.00000190, 0.00000847, 0.00000312, -0.00000316, -0.00000750, -0.00000221, 0.00000671, 0.00000601, 0.00000288, 0.00000238, 0.00000053, -0.00000379, -0.00000912, -0.00000901, -0.00000019, },
		{ -0.00000494, 0.00000393, 0.00000153, -0.00000633, -0.00000318, 0.00000095, -0.00000591, -0.00001917, -0.00001407, -0.00000311, -0.00006071, -0.00009186, 0.00004253, 0.00014188, 0.00011072, 0.00013866, 0.00008992, -0.00009897, -0.00009639, 0.00001365, -0.00001910, -0.00002702, 0.00001216, -0.00001198, -0.00001043, -0.00000199, -0.00001918, -0.00000095, 0.00000482, -0.00000716, 0.00000233, 0.00000273, 0.00000014, -0.00000261, -0.00000517, 0.00000362, -0.00000639, -0.00000850, 0.00000652, -0.00000322, -0.00000093, 0.00000327, -0.00001235, 0.00000164, 0.00000389, -0.00000815, 0.00001019, -0.00000275, -0.00001160, 0.00001434, -0.00000980, -0.00001714, 0.00001822, 0.00000212, -0.00000545, 0.00000844, -0.00001174, -0.00000316, 0.00001352, -0.00000803, -0.00001012, 0.00000391, 0.00000705, -0.00000141, -0.00000892, 0.00001360, 0.00000405, -0.00003313, 0.00000526, 0.00002488, -0.00002490, -0.00000093, 0.00003138, -0.00001967, -0.00001602, 0.00002671, -0.00000976, -0.00003037, 0.00001754, 0.00002922, -0.00000073, -0.00001110, -0.00000799, -0.00000135, 0.00000648, 0.00000356, 0.00000037, -0.00000264, -0.00001266, -0.00001075, 0.00000375, 0.00001063, 0.00000560, -0.00000335, -0.00000483, -0.00000352, -0.00000183, 0.00000714, 0.00000160, -0.00001207, -0.00000243, 0.00000465, 0.00000076, 0.00000440, -0.00000247, -0.00000388, 0.00000823, -0.00000123, -0.00000629, 0.00000594, -0.00000007, -0.00000332, 0.00000402, -0.00000281, -0.00000260, 0.00000794, 0.00000611, -0.00000119, -0.00000639, -0.00000563, 0.00000344, 0.00000763, 0.00000402, 0.00000199, 0.00000171, -0.00000150, -0.00000709, -0.00001009, },
		{ -0.00000940, -0.00000847, 0.00000038, 0.00000418, -0.00000374, -0.00000551, -0.00000038, -0.00000081, -0.00001267, -0.00002315, -0.00000306, -0.00002291, -0.00010161, -0.00003877, 0.00012135, 0.00013693, 0.00013351, 0.00013926, -0.00003640, -0.00013532, -0.00001590, 0.00000581, -0.00003808, 0.00000257, 0.00000066, -0.00001636, -0.00000035, -0.00001537, -0.00001078, 0.00000685, -0.00000449, -0.00000190, 0.00000310, 0.00000163, 0.00000019, -0.00000793, 0.00000240, -0.00000027, -0.00001395, 0.00000486, 0.00000228, -0.00000715, 0.00000765, -0.00000848, -0.00000799, 0.00001074, -0.00000850, 0.00000193, 0.00000974, -0.00001701, 0.00000673, 0.00000720, -0.00002593, 0.00000790, 0.00001561, -0.00001130, 0.00000779, -0.00000187, -0.00001502, 0.00001378, 0.00000220, -0.00001559, 0.00000271, 0.00000729, 0.00000033, -0.00000725, 0.00000295, 0.00001445, -0.00002146, -0.00001949, 0.00002763, -0.00000298, -0.00002413, 0.00002638, 0.00000805, -0.00003074, 0.00001179, 0.00001622, -0.00003057, -0.00000861, 0.00003266, 0.00001475, -0.00000993, -0.00000981, -0.00000458, 0.00000281, 0.00000560, 0.00000221, -0.00000093, -0.00000915, -0.00001331, -0.00000296, 0.00000921, 0.00000973, -0.00000097, -0.00000607, -0.00000278, -0.00000318, 0.00000350, 0.00000655, -0.00000887, -0.00000793, 0.00000450, 0.00000110, 0.00000305, 0.00000192, -0.00000625, 0.00000461, 0.00000480, -0.00000757, 0.00000220, 0.00000427, -0.00000490, 0.00000200, 0.00000114, -0.00000504, 0.00000453, 0.00000881, 0.00000135, -0.00000487, -0.00000691, -0.00000088, 0.00000709, 0.00000626, 0.00000208, 0.00000172, 0.00000050, -0.00000464, },
		{ -0.00000212, -0.00000753, -0.00001004, -0.00000422, 0.00000420, -0.00000025, -0.00000587, -0.00000256, -0.00000010, -0.00000409, -0.00002261, -0.00001681, 0.00000157, -0.00007125, -0.00010124, 0.00005524, 0.00015429, 0.00014054, 0.00016273, 0.00004313, -0.00013746, -0.00006954, 0.00002001, -0.00003038, -0.00001660, 0.00000948, -0.00001478, -0.00000438, -0.00000863, -0.00001747, 0.00000332, -0.00000016, -0.00000419, 0.00000171, 0.00000128, 0.00000433, -0.00000759, -0.00000349, 0.00000697, -0.00001453, -0.00000345, 0.00001047, -0.00001026, 0.00000407, 0.00000240, -0.00001688, 0.00000933, 0.00000103, -0.00001028, 0.00001557, -0.00000928, -0.00000954, 0.00001910, -0.00001921, -0.00001216, 0.00002504, -0.00000636, -0.00000277, 0.00001069, -0.00001852, 0.00000246, 0.00001530, -0.00001594, -0.00000425, 0.00001130, -0.00000002, -0.00000501, -0.00000260, 0.00001227, -0.00000339, -0.00003001, 0.00000982, 0.00002013, -0.00002588, 0.00000183, 0.00002866, -0.00002089, -0.00001423, 0.00002602, -0.00001006, -0.00002931, 0.00001816, 0.00002996, -0.00000125, -0.00001212, -0.00000658, -0.00000101, 0.00000445, 0.00000467, 0.00000103, -0.00000621, -0.00001282, -0.00000840, 0.00000422, 0.00001205, 0.00000396, -0.00000691, -0.00000371, -0.00000217, -0.00000055, 0.00000770, -0.00000257, -0.00001157, 0.00000149, 0.00000304, 0.00000072, 0.00000456, -0.00000445, -0.00000106, 0.00000803, -0.00000436, -0.00000311, 0.00000647, -0.00000286, -0.00000198, 0.00000370, -0.00000403, -0.00000030, 0.00000938, 0.00000485, -0.00000317, -0.00000669, -0.00000431, 0.00000415, 0.00000797, 0.00000358, 0.00000117, 0.00000153, },
		{ 0.00000143, -0.00000011, -0.00000504, -0.00000955, -0.00000831, 0.00000162, 0.00000271, -0.00000477, -0.00000372, -0.00000184, 0.00000014, -0.00001261, -0.00002886, 0.00000066, -0.00002535, -0.00011677, -0.00003444, 0.00013396, 0.00016014, 0.00017249, 0.00011571, -0.00009577, -0.00012465, 0.00000650, -0.00000782, -0.00003271, 0.00000700, -0.00000667, -0.00000958, -0.00000360, -0.00001901, -0.00000346, 0.00000268, -0.00000446, 0.00000086, -0.00000110, 0.00000628, -0.00000161, -0.00001148, 0.00000935, -0.00000661, -0.00001564, 0.00001410, -0.00000427, -0.00000696, 0.00001257, -0.00001597, -0.00000300, 0.00001392, -0.00001513, 0.00000779, 0.00000749, -0.00002151, 0.00001519, 0.00000089, -0.00002910, 0.00001946, 0.00001019, -0.00001484, 0.00001446, -0.00000836, -0.00001463, 0.00002085, -0.00000544, -0.00001570, 0.00001336, 0.00000397, -0.00000678, -0.00000268, 0.00000548, 0.00000749, -0.00002275, -0.00001348, 0.00002613, -0.00000691, -0.00002163, 0.00002613, 0.00000505, -0.00002948, 0.00001263, 0.00001503, -0.00003006, -0.00000749, 0.00003330, 0.00001478, -0.00001102, -0.00000940, -0.00000332, 0.00000125, 0.00000528, 0.00000395, -0.00000351, -0.00001123, -0.00001100, -0.00000202, 0.00001030, 0.00000984, -0.00000489, -0.00000668, -0.00000082, -0.00000229, 0.00000495, 0.00000373, -0.00001099, -0.00000394, 0.00000471, -0.00000019, 0.00000401, -0.00000015, -0.00000508, 0.00000664, 0.00000140, -0.00000647, 0.00000475, 0.00000149, -0.00000485, 0.00000299, -0.00000062, -0.00000389, 0.00000689, 0.00000834, -0.00000065, -0.00000589, -0.00000603, 0.00000013, 0.00000761, 0.00000606, 0.00000128, },
		{ 0.00000277, 0.00000091, 0.00000091, -0.00000270, -0.00000741, -0.00001057, -0.00000289, 0.00000421, -0.00000279, -0.00000414, -0.00000269, -0.00000136, -0.00000153, -0.00002751, -0.00001779, 0.00000558, -0.00008455, -0.00010730, 0.00006561, 0.00017104, 0.00018421, 0.00016922, -0.00002207, -0.00015492, -0.00003684, 0.00001346, -0.00003453, -0.00000585, 0.00000150, -0.00001158, -0.00000173, -0.00001676, -0.00001055, 0.00000330, -0.00000449, 0.00000114, -0.00000241, 0.00000268, 0.00000779, -0.00001473, 0.00000219, 0.00000697, -0.00002293, 0.00000622, 0.00000981, -0.00001672, 0.00001205, -0.00000259, -0.00001867, 0.00001800, -0.00000559, -0.00000910, 0.00001983, -0.00001763, -0.00000346, 0.00002042, -0.00002904, -0.00000214, 0.00002748, -0.00001563, 0.00000387, 0.00000921, -0.00002442, 0.00001124, 0.00001248, -0.00002250, 0.00000615, 0.00001377, -0.00000911, -0.00000370, 0.00000248, 0.00000767, -0.00000798, -0.00002548, 0.00001211, 0.00001565, -0.00002585, 0.00000387, 0.00002583, -0.00002165, -0.00001213, 0.00002532, -0.00001083, -0.00002801, 0.00001913, 0.00003014, -0.00000198, -0.00001253, -0.00000526, -0.00000149, 0.00000308, 0.00000633, 0.00000017, -0.00000953, -0.00001157, -0.00000664, 0.00000484, 0.00001329, 0.00000099, -0.00000949, -0.00000182, -0.00000105, 0.00000085, 0.00000675, -0.00000611, -0.00000915, 0.00000376, 0.00000130, 0.00000148, 0.00000342, -0.00000517, 0.00000195, 0.00000604, -0.00000566, 0.00000024, 0.00000502, -0.00000446, -0.00000031, 0.00000247, -0.00000439, 0.00000233, 0.00001003, 0.00000303, -0.00000467, -0.00000640, -0.00000331, 0.00000498, 0.00000805, },
		{ 0.00000819, 0.00000522, 0.00000095, 0.00000110, -0.00000115, -0.00000468, -0.00001014, -0.00000798, 0.00000332, 0.00000001, -0.00000480, -0.00000174, -0.00000445, 0.00000235, -0.00001374, -0.00003421, 0.00000584, -0.00003140, -0.00013002, -0.00003001, 0.00014510, 0.00019935, 0.00020722, 0.00006366, -0.00014483, -0.00009519, 0.00001565, -0.00002112, -0.00002055, 0.00000375, -0.00001009, -0.00000157, -0.00001234, -0.00001696, 0.00000256, -0.00000460, -0.00000000, 0.00000060, -0.00000495, 0.00001333, -0.00000747, -0.00001187, 0.00001632, -0.00001687, -0.00001195, 0.00002173, -0.00001421, -0.00000195, 0.00001467, -0.00002416, 0.00000600, 0.00001303, -0.00002080, 0.00001574, 0.00000137, -0.00002270, 0.00002352, -0.00000938, -0.00002690, 0.00003010, 0.00000032, -0.00001382, 0.00001980, -0.00001732, -0.00000958, 0.00002528, -0.00001543, -0.00001008, 0.00002205, -0.00000432, -0.00000989, 0.00000348, 0.00000397, 0.00000151, -0.00002122, -0.00000899, 0.00002368, -0.00000964, -0.00001926, 0.00002531, 0.00000238, -0.00002774, 0.00001359, 0.00001339, -0.00002956, -0.00000592, 0.00003367, 0.00001432, -0.00001175, -0.00000863, -0.00000282, -0.00000034, 0.00000603, 0.00000469, -0.00000689, -0.00001176, -0.00000867, -0.00000147, 0.00001179, 0.00000860, -0.00000878, -0.00000593, 0.00000097, -0.00000128, 0.00000543, 0.00000034, -0.00001095, -0.00000047, 0.00000372, -0.00000035, 0.00000404, -0.00000202, -0.00000263, 0.00000687, -0.00000141, -0.00000390, 0.00000534, -0.00000120, -0.00000363, 0.00000305, -0.00000214, -0.00000191, 0.00000875, 0.00000705, -0.00000251, -0.00000611, -0.00000532, 0.00000110, },
		{ -0.00000256, 0.00000611, 0.00000747, 0.00000203, 0.00000109, -0.00000040, -0.00000269, -0.00000728, -0.00001153, -0.00000061, 0.00000315, -0.00000526, -0.00000118, -0.00000435, -0.00000182, 0.00000104, -0.00003361, -0.00001632, 0.00000708, -0.00009858, -0.00011071, 0.00007141, 0.00019945, 0.00023782, 0.00014716, -0.00009466, -0.00014592, -0.00000924, -0.00000236, -0.00002827, -0.00000035, -0.00000804, -0.00000241, -0.00000564, -0.00002237, -0.00000024, -0.00000245, -0.00000513, 0.00000650, -0.00000957, 0.00000830, 0.00000816, -0.00002207, 0.00001154, 0.00000184, -0.00002866, 0.00001886, 0.00000332, -0.00001975, 0.00002092, -0.00001165, -0.00001585, 0.00002520, -0.00001485, -0.00000379, 0.00002094, -0.00002581, 0.00000602, 0.00001674, -0.00003641, 0.00001139, 0.00002395, -0.00002274, 0.00001289, 0.00000317, -0.00002689, 0.00002103, 0.00000469, -0.00002475, 0.00001858, 0.00000997, -0.00001685, 0.00000164, 0.00000427, 0.00000179, -0.00000935, -0.00002105, 0.00001269, 0.00001187, -0.00002528, 0.00000529, 0.00002290, -0.00002183, -0.00000970, 0.00002421, -0.00001184, -0.00002628, 0.00002010, 0.00002980, -0.00000271, -0.00001240, -0.00000432, -0.00000258, 0.00000272, 0.00000790, -0.00000216, -0.00001183, -0.00000946, -0.00000560, 0.00000593, 0.00001378, -0.00000294, -0.00001069, 0.00000043, -0.00000007, 0.00000187, 0.00000461, -0.00000813, -0.00000593, 0.00000436, 0.00000024, 0.00000214, 0.00000153, -0.00000431, 0.00000404, 0.00000327, -0.00000500, 0.00000254, 0.00000249, -0.00000463, 0.00000099, 0.00000077, -0.00000382, 0.00000499, 0.00000976, 0.00000090, -0.00000534, -0.00000600, },
		{ -0.00000583, -0.00000503, 0.00000245, 0.00000839, 0.00000390, 0.00000136, 0.00000020, -0.00000212, -0.00000370, -0.00001154, -0.00000678, 0.00000473, -0.00000373, -0.00000314, -0.00000071, -0.00000748, 0.00000582, -0.00001674, -0.00003779, 0.00001085, -0.00004055, -0.00013969, -0.00002949, 0.00015982, 0.00025772, 0.00022416, -0.00001366, -0.00017051, -0.00005605, 0.00000857, -0.00002664, -0.00000672, -0.00000690, -0.00000606, 0.00000313, -0.00002408, -0.00000793, 0.00000359, -0.00001201, 0.00000842, -0.00000447, -0.00000582, 0.00002082, -0.00001718, -0.00000729, 0.00002052, -0.00002917, -0.00000212, 0.00002472, -0.00002496, 0.00000772, 0.00001191, -0.00003040, 0.00001762, 0.00000720, -0.00002326, 0.00002377, -0.00000805, -0.00001923, 0.00002914, -0.00002164, -0.00001856, 0.00003671, -0.00001170, -0.00000773, 0.00002134, -0.00002633, -0.00000023, 0.00002480, -0.00002478, 0.00000088, 0.00002568, -0.00001492, -0.00000785, 0.00000847, -0.00000070, -0.00000199, -0.00001828, -0.00000623, 0.00002105, -0.00001149, -0.00001713, 0.00002401, 0.00000009, -0.00002536, 0.00001432, 0.00001130, -0.00002873, -0.00000411, 0.00003362, 0.00001360, -0.00001204, -0.00000776, -0.00000314, -0.00000146, 0.00000752, 0.00000396, -0.00001025, -0.00001080, -0.00000688, -0.00000086, 0.00001330, 0.00000593, -0.00001195, -0.00000411, 0.00000250, -0.00000036, 0.00000475, -0.00000259, -0.00000925, 0.00000172, 0.00000251, 0.00000013, 0.00000297, -0.00000282, 0.00000010, 0.00000553, -0.00000274, -0.00000109, 0.00000415, -0.00000293, -0.00000194, 0.00000222, -0.00000308, 0.00000065, 0.00000988, 0.00000498, -0.00000375, },
		{ -0.00000107, -0.00000513, -0.00000599, -0.00000164, 0.00000738, 0.00000599, 0.00000191, 0.00000124, -0.00000228, -0.00000172, -0.00000779, -0.00001229, 0.00000237, 0.00000046, -0.00000655, 0.00000216, -0.00000761, -0.00000062, 0.00000241, -0.00003972, -0.00001285, 0.00000474, -0.00011043, -0.00011463, 0.00007320, 0.00024772, 0.00029115, 0.00008978, -0.00015895, -0.00011213, 0.00000336, -0.00001965, -0.00001249, -0.00000432, -0.00001376, 0.00000971, -0.00001711, -0.00002138, 0.00000965, -0.00001319, 0.00000011, 0.00000874, -0.00001767, 0.00001812, 0.00000345, -0.00002679, 0.00002274, -0.00000892, -0.00002884, 0.00003138, -0.00000846, -0.00001716, 0.00002775, -0.00002318, -0.00000766, 0.00002841, -0.00002451, 0.00000515, 0.00001752, -0.00003140, 0.00001696, 0.00000787, -0.00003860, 0.00002548, 0.00001460, -0.00002515, 0.00002158, -0.00000663, -0.00002418, 0.00002883, -0.00000659, -0.00002082, 0.00002951, 0.00000063, -0.00002041, 0.00000874, 0.00000205, -0.00000323, -0.00000817, -0.00001769, 0.00001231, 0.00000879, -0.00002443, 0.00000616, 0.00001993, -0.00002124, -0.00000712, 0.00002250, -0.00001272, -0.00002415, 0.00002084, 0.00002905, -0.00000328, -0.00001185, -0.00000399, -0.00000389, 0.00000346, 0.00000866, -0.00000548, -0.00001263, -0.00000718, -0.00000506, 0.00000753, 0.00001303, -0.00000718, -0.00001042, 0.00000273, 0.00000073, 0.00000213, 0.00000203, -0.00000840, -0.00000303, 0.00000394, 0.00000004, 0.00000203, -0.00000020, -0.00000228, 0.00000464, 0.00000090, -0.00000305, 0.00000318, -0.00000007, -0.00000368, 0.00000153, -0.00000096, -0.00000238, 0.00000742, 0.00000841, },
		{ 0.00001001, 0.00000241, -0.00000394, -0.00000561, -0.00000500, 0.00000439, 0.00000769, 0.00000263, 0.00000241, -0.00000160, -0.00000235, -0.00000270, -0.00001356, -0.00000431, 0.00000487, -0.00000757, 0.00000018, -0.00000159, -0.00000967, 0.00000964, -0.00002144, -0.00003857, 0.00001335, -0.00005047, -0.00014620, -0.00003650, 0.00018695, 0.00033277, 0.00020900, -0.00010503, -0.00016432, -0.00002016, -0.00001179, -0.00001927, 0.00000192, -0.00002115, 0.00000637, 0.00000014, -0.00003422, 0.00000669, -0.00000285, -0.00001567, 0.00001835, -0.00001451, -0.00000140, 0.00002579, -0.00002895, 0.00000277, 0.00001929, -0.00004032, 0.00001278, 0.00002091, -0.00003314, 0.00002020, 0.00000357, -0.00003207, 0.00002927, -0.00000298, -0.00002093, 0.00002942, -0.00001891, -0.00001109, 0.00002982, -0.00003289, -0.00000511, 0.00003702, -0.00002265, 0.00000219, 0.00001763, -0.00003247, 0.00001140, 0.00001853, -0.00003022, 0.00001482, 0.00002258, -0.00002440, -0.00000139, 0.00001011, -0.00000665, -0.00000254, -0.00001531, -0.00000484, 0.00001864, -0.00001276, -0.00001527, 0.00002225, -0.00000163, -0.00002239, 0.00001449, 0.00000899, -0.00002742, -0.00000233, 0.00003316, 0.00001275, -0.00001189, -0.00000706, -0.00000414, -0.00000169, 0.00000912, 0.00000172, -0.00001276, -0.00000881, -0.00000583, 0.00000017, 0.00001429, 0.00000206, -0.00001388, -0.00000158, 0.00000370, 0.00000013, 0.00000322, -0.00000432, -0.00000685, 0.00000260, 0.00000176, 0.00000049, 0.00000129, -0.00000226, 0.00000210, 0.00000348, -0.00000245, 0.00000087, 0.00000197, -0.00000343, -0.00000041, 0.00000082, -0.00000327, 0.00000356, },
		{ -0.00000018, 0.00000930, 0.00000599, -0.00000228, -0.00000439, -0.00000663, -0.00000005, 0.00000824, 0.00000379, 0.00000280, 0.00000071, -0.00000413, 0.00000005, -0.00000936, -0.00001200, 0.00000517, -0.00000364, -0.00000578, 0.00000450, -0.00001160, 0.00000231, 0.00000158, -0.00004404, -0.00000926, -0.00000113, -0.00011759, -0.00012429, 0.00007600, 0.00032173, 0.00032855, -0.00000193, -0.00019842, -0.00006282, -0.00000439, -0.00002910, 0.00000750, -0.00001867, -0.00001064, 0.00001949, -0.00003408, -0.00001133, 0.00001443, -0.00002579, 0.00001150, 0.00000506, -0.00002346, 0.00003035, -0.00000700, -0.00002617, 0.00003268, -0.00002397, -0.00002172, 0.00004012, -0.00002314, -0.00000825, 0.00002972, -0.00003435, 0.00000506, 0.00002596, -0.00003183, 0.00001589, 0.00000951, -0.00003238, 0.00002658, -0.00000471, -0.00003408, 0.00003660, 0.00000142, -0.00002166, 0.00002700, -0.00001784, -0.00001611, 0.00003187, -0.00001838, -0.00001069, 0.00003553, -0.00001170, -0.00001841, 0.00001476, -0.00000343, -0.00000585, -0.00000573, -0.00001567, 0.00001162, 0.00000627, -0.00002344, 0.00000648, 0.00001705, -0.00001976, -0.00000477, 0.00002024, -0.00001320, -0.00002181, 0.00002123, 0.00002803, -0.00000359, -0.00001107, -0.00000441, -0.00000490, 0.00000506, 0.00000808, -0.00000901, -0.00001187, -0.00000534, -0.00000461, 0.00000935, 0.00001082, -0.00001103, -0.00000879, 0.00000482, 0.00000117, 0.00000156, -0.00000014, -0.00000737, -0.00000108, 0.00000331, 0.00000026, 0.00000101, -0.00000099, 0.00000001, 0.00000393, -0.00000028, -0.00000090, 0.00000230, -0.00000189, -0.00000218, 0.00000133, -0.00000241, },
		{ -0.00000083, -0.00000266, 0.00000659, 0.00000886, -0.00000008, -0.00000305, -0.00000606, -0.00000470, 0.00000669, 0.00000575, 0.00000212, 0.00000353, -0.00000418, -0.00000141, -0.00000245, -0.00001547, -0.00000083, 0.00000305, -0.00001014, 0.00000413, -0.00000451, -0.00000988, 0.00001226, -0.00002644, -0.00003682, 0.00001140, -0.00005739, -0.00015306, -0.00005193, 0.00023730, 0.00041453, 0.00015109, -0.00019271, -0.00012560, -0.00000041, -0.00003703, 0.00000146, -0.00000070, -0.00003268, 0.00002459, -0.00001287, -0.00003748, 0.00002285, -0.00001682, -0.00001220, 0.00002596, -0.00002768, 0.00000921, 0.00002532, -0.00003964, 0.00001702, 0.00001109, -0.00004672, 0.00002984, 0.00000995, -0.00003565, 0.00003231, -0.00000932, -0.00002749, 0.00003791, -0.00001546, -0.00001360, 0.00003064, -0.00002862, 0.00000026, 0.00002462, -0.00003992, 0.00001073, 0.00003045, -0.00002941, 0.00001325, 0.00000897, -0.00003342, 0.00002221, 0.00000754, -0.00002937, 0.00002815, 0.00001318, -0.00002965, 0.00000724, 0.00000752, -0.00001176, -0.00000071, -0.00001325, -0.00000420, 0.00001653, -0.00001365, -0.00001369, 0.00002010, -0.00000253, -0.00001909, 0.00001392, 0.00000682, -0.00002568, -0.00000076, 0.00003236, 0.00001192, -0.00001133, -0.00000682, -0.00000550, -0.00000083, 0.00001014, -0.00000168, -0.00001388, -0.00000648, -0.00000539, 0.00000178, 0.00001423, -0.00000247, -0.00001429, 0.00000129, 0.00000443, -0.00000003, 0.00000150, -0.00000472, -0.00000469, 0.00000273, 0.00000160, 0.00000015, -0.00000018, -0.00000071, 0.00000289, 0.00000176, -0.00000114, 0.00000144, -0.00000028, -0.00000295, 0.00000062, },
		{ -0.00000067, 0.00000055, -0.00000340, 0.00000271, 0.00001020, 0.00000280, -0.00000220, -0.00000382, -0.00000769, 0.00000249, 0.00000799, 0.00000151, 0.00000461, -0.00000094, -0.00000519, 0.00000197, -0.00001180, -0.00001016, 0.00000660, -0.00000805, -0.00000296, 0.00000461, -0.00001473, 0.00000591, -0.00000143, -0.00004502, -0.00000850, -0.00000794, -0.00011981, -0.00014525, 0.00008977, 0.00042156, 0.00033206, -0.00011755, -0.00020051, -0.00001272, -0.00003106, -0.00002317, 0.00002356, -0.00004004, 0.00000378, 0.00002114, -0.00005161, 0.00000728, 0.00001025, -0.00003559, 0.00002687, -0.00000575, -0.00002362, 0.00004172, -0.00002196, -0.00001864, 0.00003790, -0.00004024, -0.00000742, 0.00004193, -0.00003707, 0.00000571, 0.00002512, -0.00004193, 0.00001990, 0.00001764, -0.00003450, 0.00002580, -0.00000174, -0.00002794, 0.00003216, -0.00001812, -0.00002315, 0.00004182, -0.00001231, -0.00001295, 0.00002707, -0.00002732, -0.00000433, 0.00002870, -0.00002718, 0.00000359, 0.00003456, -0.00002347, -0.00001139, 0.00001736, -0.00001035, -0.00000553, -0.00000333, -0.00001472, 0.00001096, 0.00000410, -0.00002234, 0.00000622, 0.00001453, -0.00001745, -0.00000298, 0.00001767, -0.00001315, -0.00001948, 0.00002129, 0.00002681, -0.00000356, -0.00001032, -0.00000553, -0.00000517, 0.00000699, 0.00000598, -0.00001193, -0.00000996, -0.00000423, -0.00000381, 0.00001091, 0.00000724, -0.00001386, -0.00000609, 0.00000647, 0.00000100, 0.00000051, -0.00000132, -0.00000587, -0.00000006, 0.00000302, 0.00000024, -0.00000043, -0.00000061, 0.00000168, 0.00000269, -0.00000018, 0.00000049, 0.00000061, -0.00000271, },
		{ -0.00000194, -0.00000194, 0.00000111, -0.00000250, -0.00000120, 0.00000936, 0.00000630, -0.00000172, -0.00000148, -0.00000747, -0.00000346, 0.00000867, 0.00000275, 0.00000287, 0.00000397, -0.00000691, 0.00000058, -0.00000348, -0.00001626, 0.00000266, -0.00000055, -0.00001036, 0.00000704, -0.00000844, -0.00000786, 0.00001223, -0.00002942, -0.00003461, 0.00000481, -0.00005760, -0.00016678, -0.00007161, 0.00032030, 0.00048584, 0.00004778, -0.00025640, -0.00006275, -0.00000543, -0.00005585, 0.00003067, -0.00001731, -0.00003476, 0.00004290, -0.00003480, -0.00002986, 0.00003491, -0.00003550, -0.00000017, 0.00002732, -0.00004036, 0.00002484, 0.00001762, -0.00004572, 0.00003245, -0.00000369, -0.00004551, 0.00004502, -0.00000639, -0.00003074, 0.00004056, -0.00002410, -0.00001668, 0.00004098, -0.00002731, -0.00000257, 0.00002652, -0.00003460, 0.00001214, 0.00001430, -0.00004045, 0.00002528, 0.00001842, -0.00003005, 0.00002221, -0.00000264, -0.00002833, 0.00002902, -0.00000541, -0.00002166, 0.00003717, -0.00000016, -0.00002898, 0.00001513, 0.00000133, -0.00001442, 0.00000232, -0.00001240, -0.00000378, 0.00001463, -0.00001420, -0.00001249, 0.00001774, -0.00000246, -0.00001586, 0.00001268, 0.00000504, -0.00002370, 0.00000052, 0.00003129, 0.00001120, -0.00001047, -0.00000722, -0.00000673, 0.00000097, 0.00000997, -0.00000551, -0.00001345, -0.00000445, -0.00000519, 0.00000379, 0.00001275, -0.00000698, -0.00001313, 0.00000413, 0.00000445, -0.00000076, 0.00000030, -0.00000423, -0.00000324, 0.00000278, 0.00000160, -0.00000093, -0.00000076, 0.00000100, 0.00000266, 0.00000103, 0.00000021, 0.00000079, },
		{ 0.00000071, -0.00000110, -0.00000274, 0.00000058, -0.00000060, -0.00000378, 0.00000609, 0.00000967, -0.00000052, -0.00000074, -0.00000416, -0.00000849, 0.00000574, 0.00000619, 0.00000000, 0.00000673, -0.00000363, -0.00000492, 0.00000310, -0.00001428, -0.00000698, 0.00000578, -0.00001117, 0.00000077, 0.00000220, -0.00001545, 0.00000818, -0.00000510, -0.00004253, -0.00001344, -0.00001098, -0.00012048, -0.00018068, 0.00012800, 0.00053876, 0.00028617, -0.00023808, -0.00016085, 0.00002194, -0.00006834, 0.00000088, 0.00002745, -0.00006148, 0.00002799, 0.00001055, -0.00006330, 0.00003068, -0.00000428, -0.00003790, 0.00004200, -0.00002254, -0.00001601, 0.00004844, -0.00003838, -0.00000450, 0.00003573, -0.00005369, 0.00001175, 0.00003529, -0.00004636, 0.00002171, 0.00001408, -0.00004330, 0.00003356, 0.00000494, -0.00003140, 0.00003222, -0.00001373, -0.00001886, 0.00003195, -0.00002893, -0.00000803, 0.00003973, -0.00002305, -0.00000144, 0.00002137, -0.00003222, 0.00000813, 0.00001977, -0.00003010, 0.00001852, 0.00002656, -0.00003130, -0.00000148, 0.00001544, -0.00001656, -0.00000274, -0.00000190, -0.00001431, 0.00001040, 0.00000221, -0.00002122, 0.00000539, 0.00001267, -0.00001460, -0.00000195, 0.00001514, -0.00001264, -0.00001731, 0.00002106, 0.00002549, -0.00000317, -0.00000987, -0.00000710, -0.00000442, 0.00000856, 0.00000263, -0.00001360, -0.00000751, -0.00000382, -0.00000243, 0.00001166, 0.00000268, -0.00001524, -0.00000268, 0.00000738, 0.00000016, -0.00000036, -0.00000154, -0.00000466, 0.00000054, 0.00000309, -0.00000050, -0.00000151, 0.00000055, 0.00000234, 0.00000183, 0.00000061, },
		{ 0.00000130, 0.00000077, -0.00000042, -0.00000284, -0.00000083, 0.00000114, -0.00000404, 0.00000119, 0.00001131, 0.00000259, -0.00000186, -0.00000008, -0.00000958, -0.00000088, 0.00000947, -0.00000059, 0.00000486, 0.00000313, -0.00000889, 0.00000287, -0.00000563, -0.00001523, 0.00000472, -0.00000471, -0.00000808, 0.00000733, -0.00001142, -0.00000480, 0.00000921, -0.00002812, -0.00003561, -0.00000380, -0.00004949, -0.00019521, -0.00008520, 0.00043986, 0.00052511, -0.00009285, -0.00027857, 0.00000831, -0.00003652, -0.00005779, 0.00005928, -0.00004577, -0.00002285, 0.00005412, -0.00006097, -0.00001035, 0.00003752, -0.00005376, 0.00001906, 0.00001985, -0.00004831, 0.00004235, 0.00000288, -0.00004440, 0.00004510, -0.00002255, -0.00003562, 0.00005428, -0.00002458, -0.00001872, 0.00004220, -0.00003719, -0.00000144, 0.00003732, -0.00003556, 0.00000963, 0.00001768, -0.00003525, 0.00002155, 0.00000128, -0.00003393, 0.00003507, 0.00000400, -0.00002452, 0.00002638, -0.00001404, -0.00001832, 0.00002974, -0.00001670, -0.00000866, 0.00003934, -0.00001387, -0.00002261, 0.00001972, -0.00000672, -0.00001395, 0.00000523, -0.00001252, -0.00000327, 0.00001281, -0.00001445, -0.00001178, 0.00001546, -0.00000147, -0.00001305, 0.00001104, 0.00000377, -0.00002168, 0.00000155, 0.00002998, 0.00001069, -0.00000952, -0.00000831, -0.00000739, 0.00000328, 0.00000837, -0.00000898, -0.00001175, -0.00000308, -0.00000478, 0.00000579, 0.00000982, -0.00001080, -0.00001055, 0.00000652, 0.00000358, -0.00000158, -0.00000001, -0.00000364, -0.00000237, 0.00000318, 0.00000112, -0.00000221, -0.00000028, 0.00000208, 0.00000216, },
		{ 0.00000224, 0.00000191, 0.00000118, -0.00000001, -0.00000227, -0.00000239, 0.00000155, -0.00000203, -0.00000336, 0.00000960, 0.00000744, -0.00000304, 0.00000167, -0.00000581, -0.00000818, 0.00000891, 0.00000293, -0.00000005, 0.00000824, -0.00000668, -0.00000338, 0.00000278, -0.00001558, -0.00000361, 0.00000279, -0.00001154, 0.00000337, -0.00000158, -0.00001321, 0.00000705, -0.00000652, -0.00003865, -0.00002519, -0.00000490, -0.00012705, -0.00022841, 0.00020424, 0.00065582, 0.00018305, -0.00033962, -0.00008553, 0.00002831, -0.00010332, 0.00003952, 0.00001391, -0.00007460, 0.00005646, -0.00001285, -0.00006336, 0.00005266, -0.00002722, -0.00002973, 0.00005193, -0.00004186, -0.00000065, 0.00004744, -0.00005220, 0.00001380, 0.00002530, -0.00006047, 0.00003180, 0.00002098, -0.00004815, 0.00003584, -0.00000131, -0.00003740, 0.00004279, -0.00000933, -0.00002305, 0.00003343, -0.00002359, -0.00000736, 0.00002587, -0.00003425, 0.00000776, 0.00003100, -0.00002818, 0.00000956, 0.00001137, -0.00003104, 0.00001785, 0.00000743, -0.00002595, 0.00003021, 0.00001359, -0.00003312, 0.00000834, 0.00000941, -0.00002030, 0.00000137, -0.00000177, -0.00001394, 0.00000984, 0.00000059, -0.00002022, 0.00000417, 0.00001166, -0.00001162, -0.00000160, 0.00001291, -0.00001186, -0.00001534, 0.00002059, 0.00002415, -0.00000247, -0.00000994, -0.00000871, -0.00000269, 0.00000917, -0.00000138, -0.00001380, -0.00000516, -0.00000380, -0.00000054, 0.00001115, -0.00000226, -0.00001494, 0.00000099, 0.00000722, -0.00000108, -0.00000047, -0.00000133, -0.00000406, 0.00000135, 0.00000301, -0.00000189, -0.00000163, 0.00000169, },
		{ 0.00000079, 0.00000222, 0.00000222, 0.00000193, 0.00000026, -0.00000145, -0.00000320, 0.00000022, 0.00000079, -0.00000528, 0.00000446, 0.00001181, -0.00000156, -0.00000028, 0.00000026, -0.00001158, 0.00000281, 0.00000843, -0.00000325, 0.00000745, 0.00000102, -0.00000941, 0.00000436, -0.00000798, -0.00001252, 0.00000418, -0.00000746, -0.00000456, 0.00000443, -0.00001133, -0.00000319, 0.00000475, -0.00002165, -0.00004393, -0.00000892, -0.00003554, -0.00024461, -0.00007707, 0.00059075, 0.00051141, -0.00025192, -0.00024965, 0.00006811, -0.00008662, -0.00003688, 0.00007685, -0.00007925, 0.00000344, 0.00005209, -0.00008421, 0.00001821, 0.00002750, -0.00006570, 0.00004144, 0.00000343, -0.00004801, 0.00005739, -0.00001647, -0.00003469, 0.00005125, -0.00004146, -0.00001832, 0.00005491, -0.00004026, -0.00000203, 0.00003625, -0.00004520, 0.00001502, 0.00002756, -0.00003821, 0.00002007, 0.00000619, -0.00003049, 0.00002609, -0.00001110, -0.00002184, 0.00003792, -0.00000913, -0.00001469, 0.00002459, -0.00002205, -0.00000607, 0.00002415, -0.00002308, 0.00000639, 0.00003408, -0.00002437, -0.00001248, 0.00001961, -0.00001441, -0.00001073, 0.00000708, -0.00001312, -0.00000263, 0.00001103, -0.00001447, -0.00001158, 0.00001358, 0.00000017, -0.00001088, 0.00000932, 0.00000292, -0.00001974, 0.00000238, 0.00002848, 0.00001046, -0.00000872, -0.00000990, -0.00000714, 0.00000546, 0.00000547, -0.00001144, -0.00000932, -0.00000245, -0.00000388, 0.00000726, 0.00000566, -0.00001336, -0.00000683, 0.00000797, 0.00000188, -0.00000180, 0.00000038, -0.00000358, -0.00000151, 0.00000379, -0.00000022, -0.00000293, },
		{ -0.00000331, -0.00000072, 0.00000206, 0.00000214, 0.00000280, 0.00000078, -0.00000098, -0.00000283, -0.00000197, 0.00000229, -0.00000367, -0.00000187, 0.00001259, 0.00000365, -0.00000391, 0.00000405, -0.00000852, -0.00000621, 0.00001093, -0.00000102, 0.00000145, 0.00000840, -0.00000903, -0.00000137, 0.00000105, -0.00001475, -0.00000177, -0.00000106, -0.00000880, 0.00000274, -0.00000431, -0.00000918, 0.00000154, -0.00000241, -0.00003772, -0.00004168, 0.00001397, -0.00014964, -0.00027903, 0.00032746, 0.00074836, 0.00002649, -0.00039909, 0.00000963, 0.00000104, -0.00012238, 0.00008359, -0.00001714, -0.00007237, 0.00008108, -0.00004539, -0.00004879, 0.00006640, -0.00005360, -0.00001102, 0.00005251, -0.00005886, 0.00001996, 0.00003754, -0.00005953, 0.00003227, 0.00000806, -0.00005822, 0.00004815, 0.00000194, -0.00004157, 0.00004450, -0.00001756, -0.00002515, 0.00004542, -0.00002190, -0.00001149, 0.00002918, -0.00002908, 0.00000358, 0.00001563, -0.00003272, 0.00002048, 0.00001818, -0.00002685, 0.00001694, -0.00000003, -0.00002415, 0.00002221, -0.00000480, -0.00001568, 0.00003568, -0.00000092, -0.00002872, 0.00001528, 0.00000090, -0.00002073, 0.00000546, -0.00000274, -0.00001333, 0.00000917, -0.00000073, -0.00001946, 0.00000283, 0.00001149, -0.00000887, -0.00000170, 0.00001108, -0.00001101, -0.00001351, 0.00001988, 0.00002289, -0.00000159, -0.00001062, -0.00000992, -0.00000034, 0.00000849, -0.00000528, -0.00001267, -0.00000336, -0.00000375, 0.00000156, 0.00000920, -0.00000690, -0.00001290, 0.00000433, 0.00000579, -0.00000208, 0.00000040, -0.00000147, -0.00000377, 0.00000266, 0.00000215, },
		{ 0.00000399, -0.00000223, -0.00000249, 0.00000161, 0.00000177, 0.00000330, 0.00000199, -0.00000098, -0.00000187, -0.00000339, 0.00000124, 0.00000002, -0.00000589, 0.00000829, 0.00001038, -0.00000508, 0.00000258, -0.00000103, -0.00001243, 0.00000676, 0.00000556, -0.00000433, 0.00000956, -0.00000166, -0.00000843, 0.00000416, -0.00000904, -0.00000938, 0.00000089, -0.00000693, -0.00000225, -0.00000054, -0.00000704, -0.00000595, 0.00000228, -0.00001170, -0.00006219, -0.00000340, -0.00002334, -0.00031614, -0.00002952, 0.00075694, 0.00043035, -0.00040290, -0.00017151, 0.00009899, -0.00014211, 0.00000595, 0.00007505, -0.00010808, 0.00003968, 0.00003392, -0.00009704, 0.00004970, 0.00000519, -0.00006657, 0.00006131, -0.00001924, -0.00003797, 0.00006573, -0.00003633, -0.00001797, 0.00004867, -0.00005595, 0.00000297, 0.00004639, -0.00004962, 0.00001549, 0.00002385, -0.00004598, 0.00002905, 0.00001398, -0.00003483, 0.00002634, -0.00000514, -0.00002186, 0.00002487, -0.00001959, -0.00000725, 0.00003365, -0.00001781, -0.00000367, 0.00001764, -0.00002454, 0.00000498, 0.00001400, -0.00002273, 0.00001958, 0.00002296, -0.00002915, -0.00000152, 0.00001491, -0.00001983, -0.00000584, 0.00000744, -0.00001367, -0.00000199, 0.00000933, -0.00001438, -0.00001179, 0.00001233, 0.00000209, -0.00000931, 0.00000778, 0.00000230, -0.00001791, 0.00000302, 0.00002684, 0.00001052, -0.00000831, -0.00001167, -0.00000596, 0.00000694, 0.00000177, -0.00001256, -0.00000678, -0.00000233, -0.00000242, 0.00000772, 0.00000075, -0.00001422, -0.00000248, 0.00000802, -0.00000021, -0.00000087, 0.00000078, -0.00000418, -0.00000004, },
		{ -0.00000307, 0.00000417, 0.00000017, -0.00000390, 0.00000068, 0.00000142, 0.00000290, 0.00000377, -0.00000073, -0.00000149, -0.00000309, -0.00000151, 0.00000276, -0.00000535, 0.00000080, 0.00001424, -0.00000098, -0.00000275, 0.00000530, -0.00001138, -0.00000269, 0.00001092, -0.00000442, 0.00000383, 0.00000705, -0.00000978, -0.00000022, -0.00000106, -0.00001175, -0.00000306, -0.00000343, -0.00000419, -0.00000219, -0.00000314, -0.00000630, -0.00000712, 0.00000931, -0.00004526, -0.00005707, 0.00004509, -0.00019815, -0.00031596, 0.00049749, 0.00079007, -0.00016684, -0.00040040, 0.00010272, -0.00005572, -0.00011553, 0.00012088, -0.00006003, -0.00005174, 0.00009364, -0.00007988, -0.00002065, 0.00006671, -0.00007671, 0.00001502, 0.00004196, -0.00006875, 0.00004147, 0.00002003, -0.00005799, 0.00004651, -0.00001249, -0.00004688, 0.00005703, -0.00001749, -0.00002804, 0.00004555, -0.00003080, -0.00000924, 0.00004104, -0.00003001, 0.00000041, 0.00002080, -0.00002927, 0.00001122, 0.00000420, -0.00002495, 0.00002734, 0.00000484, -0.00002021, 0.00001885, -0.00000953, -0.00001375, 0.00002028, -0.00001347, -0.00000208, 0.00003378, -0.00001328, -0.00001971, 0.00001762, -0.00000786, -0.00001804, 0.00000849, -0.00000430, -0.00001245, 0.00000836, -0.00000177, -0.00001904, 0.00000168, 0.00001195, -0.00000659, -0.00000193, 0.00000958, -0.00001020, -0.00001179, 0.00001892, 0.00002179, -0.00000074, -0.00001182, -0.00001041, 0.00000211, 0.00000656, -0.00000845, -0.00001067, -0.00000225, -0.00000335, 0.00000343, 0.00000588, -0.00001058, -0.00000932, 0.00000667, 0.00000322, -0.00000204, 0.00000185, -0.00000254, },
		{ 0.00000031, -0.00000493, 0.00000223, 0.00000305, -0.00000416, -0.00000090, 0.00000142, 0.00000160, 0.00000517, 0.00000080, -0.00000217, -0.00000174, -0.00000365, 0.00000231, -0.00000125, -0.00000543, 0.00001197, 0.00000701, -0.00000691, 0.00000566, -0.00000377, -0.00001134, 0.00000966, 0.00000165, -0.00000358, 0.00001022, -0.00000367, -0.00000682, 0.00000215, -0.00000750, -0.00000786, -0.00000383, -0.00000226, -0.00000393, -0.00000475, 0.00000056, -0.00001507, 0.00000626, -0.00000294, -0.00008963, 0.00001909, -0.00002474, -0.00040298, 0.00007225, 0.00091290, 0.00027933, -0.00051736, -0.00005791, 0.00009022, -0.00018659, 0.00006273, 0.00005063, -0.00012278, 0.00007760, 0.00000138, -0.00009413, 0.00007632, -0.00002530, -0.00005443, 0.00007315, -0.00004329, -0.00001944, 0.00006468, -0.00005210, 0.00000231, 0.00003743, -0.00006249, 0.00002361, 0.00003064, -0.00005053, 0.00002969, 0.00000799, -0.00003935, 0.00003767, -0.00000018, -0.00002668, 0.00002732, -0.00001369, -0.00001197, 0.00001875, -0.00002217, 0.00000614, 0.00002410, -0.00002044, 0.00000522, 0.00000793, -0.00002117, 0.00001183, 0.00000246, -0.00001591, 0.00002762, 0.00000911, -0.00002755, 0.00000731, 0.00000704, -0.00002191, -0.00000064, 0.00000645, -0.00001388, -0.00000149, 0.00000780, -0.00001431, -0.00001217, 0.00001174, 0.00000391, -0.00000813, 0.00000653, 0.00000174, -0.00001612, 0.00000348, 0.00002514, 0.00001079, -0.00000839, -0.00001322, -0.00000409, 0.00000735, -0.00000208, -0.00001239, -0.00000460, -0.00000236, -0.00000059, 0.00000686, -0.00000429, -0.00001317, 0.00000182, 0.00000638, -0.00000183, 0.00000129, },
		{ -0.00000038, 0.00000298, -0.00000451, -0.00000130, 0.00000508, -0.00000266, -0.00000306, 0.00000161, 0.00000025, 0.00000491, 0.00000382, -0.00000288, -0.00000119, -0.00000346, -0.00000086, 0.00000277, -0.00000668, 0.00000445, 0.00001382, -0.00000525, 0.00000006, 0.00000467, -0.00001312, 0.00000144, 0.00000870, -0.00000614, 0.00000584, 0.00000492, -0.00000870, -0.00000115, -0.00000178, -0.00000781, -0.00000802, -0.00000180, -0.00000053, -0.00001057, 0.00000397, -0.00000850, -0.00001521, 0.00002796, -0.00006594, -0.00006265, 0.00008198, -0.00027815, -0.00031815, 0.00070208, 0.00075924, -0.00036962, -0.00033918, 0.00017130, -0.00012917, -0.00007979, 0.00013989, -0.00010477, -0.00001500, 0.00008859, -0.00010762, 0.00001585, 0.00005183, -0.00009016, 0.00004271, 0.00002161, -0.00006841, 0.00005890, -0.00000157, -0.00004750, 0.00005308, -0.00003177, -0.00002881, 0.00005655, -0.00003288, -0.00001092, 0.00003897, -0.00003798, 0.00000672, 0.00003111, -0.00003230, 0.00000985, 0.00001071, -0.00002488, 0.00001395, -0.00000517, -0.00001335, 0.00002735, -0.00000558, -0.00001095, 0.00001533, -0.00001455, -0.00000306, 0.00001313, -0.00001624, 0.00001112, 0.00002549, -0.00002064, -0.00000889, 0.00001510, -0.00001482, -0.00001321, 0.00000993, -0.00000593, -0.00001141, 0.00000744, -0.00000260, -0.00001889, 0.00000094, 0.00001269, -0.00000475, -0.00000204, 0.00000823, -0.00000943, -0.00001013, 0.00001775, 0.00002089, -0.00000012, -0.00001327, -0.00001007, 0.00000412, 0.00000376, -0.00001049, -0.00000835, -0.00000171, -0.00000243, 0.00000456, 0.00000150, -0.00001271, -0.00000466, 0.00000734, 0.00000015, },
		{ 0.00000320, -0.00000202, 0.00000409, -0.00000160, -0.00000495, 0.00000486, 0.00000061, -0.00000504, 0.00000118, 0.00000003, 0.00000259, 0.00000694, -0.00000172, -0.00000256, -0.00000149, -0.00000389, 0.00000332, -0.00000292, -0.00000357, 0.00001452, 0.00000237, -0.00000645, 0.00000763, -0.00000697, -0.00000829, 0.00001036, -0.00000187, -0.00000165, 0.00000907, -0.00000380, -0.00000614, -0.00000062, -0.00000296, -0.00001013, -0.00000726, 0.00000551, -0.00001147, -0.00000444, 0.00000828, -0.00003016, 0.00002040, -0.00000219, -0.00012097, 0.00006102, -0.00005312, -0.00048947, 0.00023472, 0.00102805, 0.00007023, -0.00057190, 0.00006840, 0.00004124, -0.00020576, 0.00012054, 0.00000680, -0.00011714, 0.00010750, -0.00003930, -0.00007416, 0.00009099, -0.00005721, -0.00003094, 0.00007331, -0.00006314, 0.00000389, 0.00005401, -0.00006011, 0.00002162, 0.00001999, -0.00005959, 0.00003912, 0.00001150, -0.00004320, 0.00003746, -0.00000743, -0.00002712, 0.00003943, -0.00001176, -0.00001621, 0.00002348, -0.00001787, -0.00000362, 0.00001008, -0.00001862, 0.00001517, 0.00001246, -0.00001740, 0.00000964, -0.00000137, -0.00001347, 0.00001295, -0.00000698, -0.00000483, 0.00002883, -0.00000379, -0.00002073, 0.00001195, -0.00000183, -0.00002060, 0.00000371, 0.00000458, -0.00001366, -0.00000120, 0.00000645, -0.00001436, -0.00001247, 0.00001164, 0.00000542, -0.00000705, 0.00000550, 0.00000113, -0.00001431, 0.00000371, 0.00002347, 0.00001111, -0.00000893, -0.00001422, -0.00000198, 0.00000665, -0.00000546, -0.00001127, -0.00000300, -0.00000220, 0.00000123, 0.00000459, -0.00000875, -0.00001022, 0.00000519, },
		{ 0.00000031, 0.00000593, -0.00000240, 0.00000293, 0.00000274, -0.00000666, 0.00000172, 0.00000452, -0.00000544, -0.00000081, 0.00000126, -0.00000047, 0.00000782, 0.00000225, -0.00000474, -0.00000026, -0.00000411, 0.00000010, 0.00000208, -0.00000704, 0.00000839, 0.00001112, -0.00000786, 0.00000338, 0.00000218, -0.00001269, 0.00000467, 0.00000509, -0.00000570, 0.00000621, 0.00000346, -0.00000665, -0.00000457, 0.00000072, -0.00000537, -0.00001550, 0.00000538, -0.00000144, -0.00001931, 0.00001684, -0.00001915, -0.00001718, 0.00004919, -0.00010118, -0.00004924, 0.00011230, -0.00038722, -0.00026502, 0.00091697, 0.00064508, -0.00054931, -0.00022470, 0.00019844, -0.00020126, -0.00002035, 0.00013352, -0.00013996, 0.00003058, 0.00006500, -0.00012107, 0.00005273, 0.00002412, -0.00008993, 0.00006557, -0.00000443, -0.00005739, 0.00006820, -0.00002254, -0.00003044, 0.00005057, -0.00004546, -0.00000818, 0.00004736, -0.00004103, 0.00000561, 0.00002685, -0.00003780, 0.00001923, 0.00001842, -0.00002908, 0.00001502, 0.00000160, -0.00001791, 0.00001184, -0.00001002, -0.00000130, 0.00002157, -0.00001092, -0.00000229, 0.00000818, -0.00001410, 0.00000480, 0.00000342, -0.00001272, 0.00002048, 0.00001352, -0.00002184, 0.00000075, 0.00000884, -0.00001864, -0.00000759, 0.00000978, -0.00000725, -0.00001040, 0.00000650, -0.00000332, -0.00001884, 0.00000065, 0.00001340, -0.00000318, -0.00000194, 0.00000686, -0.00000864, -0.00000856, 0.00001640, 0.00002016, 0.00000014, -0.00001464, -0.00000907, 0.00000532, 0.00000064, -0.00001137, -0.00000617, -0.00000146, -0.00000109, 0.00000457, -0.00000340, -0.00001285, },
		{ -0.00001189, -0.00000573, 0.00000674, -0.00000077, -0.00000008, 0.00000639, -0.00000483, -0.00000340, 0.00000685, -0.00000299, -0.00000416, 0.00000265, -0.00000199, 0.00000517, 0.00000749, -0.00000490, -0.00000185, -0.00000160, -0.00000380, 0.00000391, -0.00000456, -0.00000033, 0.00001496, -0.00000224, -0.00000394, 0.00000746, -0.00000913, -0.00000423, 0.00000845, -0.00000355, 0.00000001, 0.00000679, -0.00000153, -0.00000805, -0.00000211, 0.00000347, -0.00001737, -0.00000616, 0.00001331, -0.00002466, 0.00000364, 0.00001123, -0.00004773, 0.00004575, -0.00001638, -0.00014671, 0.00011852, -0.00011909, -0.00055287, 0.00045181, 0.00107349, -0.00017154, -0.00055424, 0.00018123, -0.00003799, -0.00019171, 0.00016529, -0.00004746, -0.00009034, 0.00012122, -0.00007912, -0.00004048, 0.00008953, -0.00008305, -0.00000099, 0.00006124, -0.00007415, 0.00002689, 0.00003605, -0.00005876, 0.00003568, 0.00000049, -0.00004829, 0.00004647, -0.00000650, -0.00003005, 0.00003759, -0.00001881, -0.00001250, 0.00003479, -0.00001862, -0.00000624, 0.00001660, -0.00001755, 0.00000117, 0.00000186, -0.00001058, 0.00001821, 0.00000215, -0.00001079, 0.00000895, -0.00000744, -0.00000427, 0.00000871, -0.00001168, 0.00000714, 0.00002357, -0.00001270, -0.00001120, 0.00001177, -0.00000956, -0.00001670, 0.00000651, 0.00000240, -0.00001312, -0.00000111, 0.00000528, -0.00001451, -0.00001250, 0.00001174, 0.00000663, -0.00000585, 0.00000452, 0.00000045, -0.00001244, 0.00000371, 0.00002187, 0.00001131, -0.00000973, -0.00001458, -0.00000009, 0.00000512, -0.00000799, -0.00000968, -0.00000194, -0.00000164, 0.00000260, 0.00000107, },
		{ 0.00000324, -0.00000811, -0.00001083, 0.00000456, 0.00000252, -0.00000322, 0.00000710, 0.00000048, -0.00000786, 0.00000543, 0.00000202, -0.00000701, 0.00000206, -0.00000058, 0.00000027, 0.00001060, -0.00000096, -0.00000549, 0.00000069, -0.00000483, 0.00000126, 0.00000063, -0.00000585, 0.00001148, 0.00000671, -0.00000793, 0.00000566, -0.00000113, -0.00000982, 0.00000549, 0.00000168, -0.00000367, 0.00000427, 0.00000426, -0.00000543, -0.00000959, 0.00000738, -0.00000724, -0.00002257, 0.00001763, -0.00000988, -0.00002380, 0.00003250, -0.00003931, -0.00000743, 0.00006544, -0.00014741, -0.00001060, 0.00012007, -0.00051282, -0.00014255, 0.00110948, 0.00045204, -0.00067546, -0.00007808, 0.00017727, -0.00025385, 0.00005074, 0.00010131, -0.00015624, 0.00007480, 0.00002705, -0.00011620, 0.00008158, -0.00001044, -0.00007561, 0.00007830, -0.00003040, -0.00003814, 0.00006747, -0.00003833, -0.00001088, 0.00004006, -0.00005084, 0.00001034, 0.00003234, -0.00004089, 0.00001778, 0.00001268, -0.00003101, 0.00002601, 0.00000617, -0.00002213, 0.00001554, -0.00000454, -0.00001091, 0.00000651, -0.00000947, 0.00000797, 0.00001262, -0.00001084, 0.00000312, 0.00000023, -0.00000905, 0.00000784, -0.00000556, -0.00000445, 0.00002390, 0.00000132, -0.00001757, 0.00000690, 0.00000084, -0.00001901, -0.00000242, 0.00000845, -0.00000809, -0.00000953, 0.00000561, -0.00000401, -0.00001871, 0.00000069, 0.00001388, -0.00000164, -0.00000170, 0.00000535, -0.00000775, -0.00000712, 0.00001496, 0.00001950, 0.00000006, -0.00001565, -0.00000776, 0.00000562, -0.00000226, -0.00001133, -0.00000441, -0.00000117, 0.00000042, },
		{ -0.00000067, 0.00000308, -0.00000330, -0.00001311, -0.00000051, 0.00000587, -0.00000430, 0.00000402, 0.00000689, -0.00000846, -0.00000006, 0.00000710, -0.00000671, -0.00000153, 0.00000257, -0.00000349, 0.00000890, 0.00000601, -0.00000767, -0.00000036, -0.00000225, -0.00000314, 0.00000365, -0.00000542, 0.00000365, 0.00001280, -0.00000526, -0.00000057, 0.00000499, -0.00000883, -0.00000089, 0.00000466, -0.00000257, -0.00000025, 0.00000504, 0.00000242, -0.00001346, 0.00000012, 0.00000915, -0.00002891, 0.00000204, 0.00001675, -0.00004057, 0.00002059, 0.00000445, -0.00006157, 0.00007924, -0.00004988, -0.00015527, 0.00018046, -0.00022601, -0.00056797, 0.00070344, 0.00102934, -0.00041161, -0.00046664, 0.00025748, -0.00012972, -0.00014524, 0.00018590, -0.00009996, -0.00004742, 0.00011470, -0.00010891, -0.00000022, 0.00007202, -0.00009686, 0.00002874, 0.00003968, -0.00007407, 0.00004460, 0.00001506, -0.00004897, 0.00004178, -0.00001654, -0.00003179, 0.00004494, -0.00001944, -0.00001488, 0.00003108, -0.00002390, 0.00000096, 0.00002579, -0.00002017, 0.00000098, 0.00000910, -0.00001404, 0.00000181, -0.00000333, -0.00000093, 0.00001559, -0.00000430, -0.00000353, 0.00000428, -0.00000880, 0.00000339, 0.00000118, -0.00001067, 0.00001659, 0.00001402, -0.00001603, -0.00000189, 0.00000752, -0.00001460, -0.00001148, 0.00000762, 0.00000035, -0.00001244, -0.00000116, 0.00000425, -0.00001463, -0.00001223, 0.00001178, 0.00000768, -0.00000444, 0.00000342, -0.00000021, -0.00001052, 0.00000351, 0.00002033, 0.00001129, -0.00001052, -0.00001439, 0.00000124, 0.00000319, -0.00000956, -0.00000805, -0.00000122, },
		{ -0.00000309, -0.00000061, 0.00000171, 0.00000057, -0.00001184, -0.00000687, 0.00000698, -0.00000208, -0.00000141, 0.00001076, -0.00000366, -0.00000705, 0.00000861, -0.00000206, -0.00000650, 0.00000429, -0.00000328, 0.00000298, 0.00001191, -0.00000500, -0.00000478, 0.00000104, -0.00000515, 0.00000212, -0.00000115, -0.00000290, 0.00001246, 0.00000192, -0.00000550, 0.00000552, -0.00000352, -0.00000541, 0.00000316, 0.00000034, -0.00000166, 0.00000055, 0.00000831, -0.00000724, -0.00001394, 0.00001752, -0.00001550, -0.00002523, 0.00003201, -0.00002675, -0.00001936, 0.00004545, -0.00006652, 0.00001744, 0.00006770, -0.00019561, 0.00005336, 0.00008991, -0.00063297, 0.00005124, 0.00124496, 0.00020052, -0.00072694, 0.00007267, 0.00011266, -0.00027374, 0.00011810, 0.00004974, -0.00014917, 0.00010754, -0.00001706, -0.00009377, 0.00009602, -0.00004411, -0.00005052, 0.00007834, -0.00005067, -0.00001519, 0.00005753, -0.00004599, 0.00000663, 0.00002473, -0.00004755, 0.00002293, 0.00001565, -0.00003372, 0.00002340, 0.00000022, -0.00002007, 0.00002650, -0.00000302, -0.00001396, 0.00001244, -0.00000694, -0.00000602, 0.00000048, -0.00000443, 0.00001242, 0.00000366, -0.00000678, 0.00000412, -0.00000570, -0.00000169, 0.00000581, -0.00001102, 0.00000565, 0.00002120, -0.00000797, -0.00000995, 0.00000855, -0.00000676, -0.00001647, 0.00000144, 0.00000648, -0.00000851, -0.00000886, 0.00000480, -0.00000463, -0.00001837, 0.00000081, 0.00001414, 0.00000009, -0.00000147, 0.00000372, -0.00000671, -0.00000583, 0.00001346, 0.00001877, -0.00000020, -0.00001619, -0.00000655, 0.00000521, -0.00000455, -0.00001076, },
		{ -0.00001033, -0.00000664, -0.00000058, 0.00000056, 0.00000236, -0.00000787, -0.00001200, 0.00000431, 0.00000254, -0.00000592, 0.00000938, 0.00000478, -0.00001107, 0.00000443, 0.00000498, -0.00000916, 0.00000197, 0.00000064, -0.00000328, 0.00001257, 0.00000251, -0.00000900, 0.00000114, -0.00000313, -0.00000199, 0.00000229, -0.00000462, 0.00000715, 0.00000840, -0.00000548, 0.00000188, 0.00000119, -0.00000551, -0.00000019, 0.00000075, 0.00000070, -0.00000353, 0.00000586, 0.00000609, -0.00002186, 0.00000782, 0.00001058, -0.00004189, 0.00001782, 0.00001156, -0.00005401, 0.00004429, -0.00001514, -0.00006432, 0.00011349, -0.00010196, -0.00013650, 0.00022995, -0.00036664, -0.00051344, 0.00095784, 0.00089055, -0.00061399, -0.00032571, 0.00028299, -0.00021332, -0.00007562, 0.00017752, -0.00013888, 0.00000226, 0.00008926, -0.00012208, 0.00003777, 0.00004272, -0.00009597, 0.00005182, 0.00001374, -0.00006365, 0.00005367, -0.00000419, -0.00003384, 0.00003956, -0.00002751, -0.00001444, 0.00003621, -0.00002521, -0.00000161, 0.00002068, -0.00002241, 0.00001040, 0.00001534, -0.00001736, 0.00000439, 0.00000308, -0.00000941, -0.00000079, -0.00000418, 0.00000729, 0.00000926, -0.00000599, 0.00000167, -0.00000200, -0.00000571, 0.00000725, -0.00000670, -0.00000487, 0.00002117, 0.00000332, -0.00001399, 0.00000476, 0.00000093, -0.00001634, -0.00000622, 0.00000732, -0.00000132, -0.00001177, -0.00000129, 0.00000333, -0.00001460, -0.00001177, 0.00001161, 0.00000879, -0.00000290, 0.00000209, -0.00000076, -0.00000859, 0.00000314, 0.00001879, 0.00001107, -0.00001106, -0.00001393, 0.00000191, 0.00000134, },
		{ 0.00000445, -0.00000611, -0.00001001, -0.00000211, 0.00000032, 0.00000234, -0.00000320, -0.00001378, -0.00000171, 0.00000672, -0.00000645, 0.00000311, 0.00001232, -0.00000859, -0.00000387, 0.00000989, -0.00000655, -0.00000384, 0.00000473, -0.00000552, 0.00000718, 0.00001083, -0.00000857, -0.00000306, 0.00000056, -0.00000463, 0.00000202, -0.00000240, 0.00000134, 0.00001050, -0.00000141, -0.00000171, 0.00000248, -0.00000318, -0.00000139, -0.00000182, 0.00000235, -0.00000175, -0.00000322, 0.00001526, -0.00001371, -0.00001459, 0.00002853, -0.00003035, -0.00001975, 0.00004356, -0.00004987, -0.00000312, 0.00004917, -0.00009465, 0.00005660, 0.00004847, -0.00023293, 0.00013465, 0.00001211, -0.00071999, 0.00030331, 0.00129475, -0.00007653, -0.00069707, 0.00019954, 0.00001970, -0.00025621, 0.00016738, -0.00000977, -0.00012044, 0.00012176, -0.00005758, -0.00005907, 0.00009355, -0.00006958, -0.00002067, 0.00006652, -0.00006140, 0.00000623, 0.00004154, -0.00004492, 0.00001853, 0.00000880, -0.00003755, 0.00002773, 0.00000142, -0.00002256, 0.00002235, -0.00000768, -0.00000825, 0.00002188, -0.00000781, -0.00000692, 0.00000762, -0.00000619, -0.00000428, -0.00000380, 0.00000283, 0.00001176, -0.00000264, -0.00000123, 0.00000123, -0.00000783, 0.00000517, 0.00000018, -0.00001160, 0.00001441, 0.00001397, -0.00001247, -0.00000171, 0.00000606, -0.00001225, -0.00001215, 0.00000368, 0.00000437, -0.00000864, -0.00000838, 0.00000409, -0.00000511, -0.00001780, 0.00000081, 0.00001431, 0.00000205, -0.00000141, 0.00000201, -0.00000552, -0.00000469, 0.00001189, 0.00001791, -0.00000044, -0.00001633, -0.00000571, },
		{ -0.00001351, 0.00000205, -0.00000010, -0.00001058, -0.00000557, 0.00000022, 0.00000175, 0.00000028, -0.00001180, -0.00000860, 0.00000739, -0.00000245, -0.00000450, 0.00001436, 0.00000014, -0.00001132, 0.00000867, 0.00000077, -0.00000926, 0.00000491, -0.00000246, -0.00000071, 0.00001465, -0.00000207, -0.00000853, 0.00000183, -0.00000356, -0.00000095, 0.00000014, -0.00000164, 0.00000861, 0.00000310, -0.00000261, 0.00000162, -0.00000199, 0.00000015, -0.00000336, -0.00000097, 0.00000458, -0.00000972, 0.00001090, 0.00000675, -0.00003101, 0.00002105, 0.00000462, -0.00005178, 0.00003888, -0.00000472, -0.00005890, 0.00006929, -0.00004693, -0.00004995, 0.00013827, -0.00016563, -0.00008548, 0.00024808, -0.00052248, -0.00037827, 0.00117726, 0.00066942, -0.00074911, -0.00015837, 0.00025573, -0.00027088, 0.00000230, 0.00014273, -0.00015619, 0.00004773, 0.00005109, -0.00011649, 0.00006554, 0.00000870, -0.00008161, 0.00006388, -0.00001066, -0.00004625, 0.00005330, -0.00001782, -0.00001757, 0.00003094, -0.00003084, -0.00000032, 0.00002367, -0.00002389, 0.00000688, 0.00000985, -0.00001587, 0.00001445, 0.00000618, -0.00001216, 0.00000430, -0.00000026, -0.00000569, -0.00000469, -0.00000098, 0.00001185, 0.00000202, -0.00000374, 0.00000331, -0.00000724, 0.00000009, 0.00000660, -0.00001220, 0.00000335, 0.00002024, -0.00000550, -0.00000825, 0.00000742, -0.00000594, -0.00001506, -0.00000188, 0.00000609, -0.00000255, -0.00001122, -0.00000144, 0.00000259, -0.00001436, -0.00001133, 0.00001124, 0.00001014, -0.00000137, 0.00000055, -0.00000117, -0.00000669, 0.00000263, 0.00001717, 0.00001077, -0.00001126, },
		{ -0.00000047, -0.00001629, -0.00000533, 0.00000372, -0.00000703, -0.00000936, -0.00000125, 0.00000154, 0.00000194, -0.00000752, -0.00001333, 0.00000342, 0.00000356, -0.00000887, 0.00000941, 0.00001069, -0.00001260, 0.00000105, 0.00000841, -0.00000985, 0.00000005, 0.00000294, -0.00000584, 0.00001153, 0.00000728, -0.00001043, -0.00000130, -0.00000037, -0.00000326, 0.00000044, -0.00000196, 0.00000559, 0.00000564, -0.00000166, 0.00000143, -0.00000264, 0.00000095, -0.00000013, -0.00000760, 0.00000775, -0.00000570, -0.00000450, 0.00002318, -0.00002498, -0.00000883, 0.00003644, -0.00004946, -0.00000424, 0.00004673, -0.00007403, 0.00002438, 0.00003834, -0.00011521, 0.00010413, 0.00000475, -0.00024596, 0.00021760, -0.00011300, -0.00074673, 0.00058617, 0.00124332, -0.00034033, -0.00059509, 0.00028119, -0.00008069, -0.00020593, 0.00018924, -0.00006425, -0.00007727, 0.00011545, -0.00008604, -0.00002030, 0.00007611, -0.00008214, 0.00000715, 0.00004667, -0.00006154, 0.00002179, 0.00002390, -0.00003692, 0.00002316, -0.00000382, -0.00002441, 0.00002518, -0.00000749, -0.00001109, 0.00001646, -0.00000984, 0.00000140, 0.00001447, -0.00000836, -0.00000251, 0.00000306, -0.00000381, -0.00000537, -0.00000479, 0.00000961, 0.00000734, -0.00000500, 0.00000323, -0.00000367, -0.00000590, 0.00000920, -0.00000656, -0.00000766, 0.00001936, 0.00000494, -0.00001196, 0.00000467, 0.00000085, -0.00001476, -0.00000731, 0.00000439, 0.00000246, -0.00000866, -0.00000804, 0.00000354, -0.00000536, -0.00001716, 0.00000057, 0.00001460, 0.00000418, -0.00000162, 0.00000033, -0.00000420, -0.00000368, 0.00001022, 0.00001693, },
		{ 0.00001540, 0.00001055, -0.00001121, -0.00001337, 0.00000195, -0.00000101, -0.00001061, -0.00000478, 0.00000133, 0.00000251, -0.00000309, -0.00001416, -0.00000356, 0.00000763, -0.00000756, 0.00000023, 0.00001708, -0.00000582, -0.00000860, 0.00001098, -0.00000426, -0.00000679, 0.00000587, -0.00000525, 0.00000379, 0.00001394, -0.00000620, -0.00000671, 0.00000133, -0.00000286, -0.00000092, -0.00000187, 0.00000329, 0.00000677, -0.00000106, 0.00000237, -0.00000227, -0.00000226, 0.00000611, -0.00001031, 0.00000142, 0.00000654, -0.00001713, 0.00002042, 0.00000192, -0.00003743, 0.00003768, -0.00001017, -0.00005368, 0.00006021, -0.00003131, -0.00005039, 0.00008804, -0.00008592, -0.00001626, 0.00014328, -0.00022839, -0.00000546, 0.00021927, -0.00066629, -0.00016643, 0.00132604, 0.00039393, -0.00080042, 0.00000473, 0.00018564, -0.00029180, 0.00007235, 0.00009056, -0.00014976, 0.00007972, 0.00000932, -0.00009489, 0.00007825, -0.00002227, -0.00005836, 0.00006378, -0.00002859, -0.00002665, 0.00004521, -0.00002392, -0.00000408, 0.00001936, -0.00002725, 0.00000787, 0.00001123, -0.00001756, 0.00000957, 0.00000171, -0.00000702, 0.00001343, 0.00000012, -0.00000674, 0.00000202, -0.00000095, -0.00000412, -0.00000772, 0.00000459, 0.00001206, -0.00000356, 0.00000043, 0.00000141, -0.00000952, 0.00000613, 0.00000234, -0.00001370, 0.00001114, 0.00001488, -0.00001039, -0.00000128, 0.00000610, -0.00001131, -0.00001167, 0.00000101, 0.00000444, -0.00000344, -0.00001084, -0.00000153, 0.00000209, -0.00001395, -0.00001109, 0.00001083, 0.00001176, -0.00000000, -0.00000117, -0.00000140, -0.00000482, 0.00000197, },
		{ -0.00000281, 0.00000837, 0.00001596, -0.00000024, -0.00001634, -0.00000529, 0.00000328, -0.00000751, -0.00000892, -0.00000030, 0.00000287, 0.00000019, -0.00001156, -0.00001031, 0.00000701, -0.00000179, -0.00000791, 0.00001564, 0.00000584, -0.00001399, 0.00000612, 0.00000429, -0.00001057, 0.00000353, -0.00000042, -0.00000341, 0.00001421, 0.00000228, -0.00001004, -0.00000059, -0.00000085, -0.00000166, -0.00000253, 0.00000107, 0.00000804, -0.00000089, 0.00000191, 0.00000170, -0.00000777, 0.00000855, -0.00000351, -0.00001144, 0.00001496, -0.00001398, -0.00000089, 0.00002896, -0.00003923, 0.00000451, 0.00003700, -0.00006819, 0.00002021, 0.00003727, -0.00009220, 0.00005864, 0.00001101, -0.00011977, 0.00014982, -0.00005992, -0.00022493, 0.00028191, -0.00027280, -0.00069362, 0.00086250, 0.00109267, -0.00055466, -0.00044362, 0.00030768, -0.00016749, -0.00013517, 0.00018164, -0.00010286, -0.00003001, 0.00009196, -0.00009761, 0.00001399, 0.00004924, -0.00008079, 0.00002734, 0.00002423, -0.00005288, 0.00002928, 0.00000882, -0.00002536, 0.00002116, -0.00001069, -0.00001205, 0.00001767, -0.00001019, -0.00000244, 0.00000865, -0.00000690, 0.00000696, 0.00000687, -0.00000600, -0.00000100, 0.00000012, -0.00000153, -0.00000789, -0.00000238, 0.00001372, 0.00000150, -0.00000374, 0.00000492, -0.00000824, -0.00000117, 0.00000937, -0.00001186, -0.00000101, 0.00001956, -0.00000305, -0.00000770, 0.00000768, -0.00000519, -0.00001431, -0.00000297, 0.00000396, 0.00000090, -0.00000872, -0.00000779, 0.00000323, -0.00000538, -0.00001662, 0.00000017, 0.00001515, 0.00000634, -0.00000210, -0.00000128, -0.00000279, },
		{ -0.00000148, -0.00000297, 0.00000111, 0.00001356, 0.00001057, -0.00001108, -0.00001356, 0.00000190, -0.00000143, -0.00001066, -0.00000412, 0.00000277, 0.00000239, -0.00000733, -0.00001414, 0.00000189, 0.00000456, -0.00001079, 0.00000736, 0.00001610, -0.00001113, -0.00000349, 0.00001017, -0.00000826, -0.00000263, 0.00000420, -0.00000600, 0.00000874, 0.00001024, -0.00000835, -0.00000477, 0.00000003, -0.00000081, -0.00000277, -0.00000233, 0.00000898, 0.00000126, -0.00000196, 0.00000732, -0.00000934, 0.00000206, 0.00000956, -0.00001927, 0.00000858, 0.00000412, -0.00002293, 0.00003292, -0.00000954, -0.00003746, 0.00005360, -0.00003253, -0.00004406, 0.00007533, -0.00006378, -0.00002676, 0.00009311, -0.00012354, 0.00003348, 0.00012156, -0.00027501, 0.00009132, 0.00013647, -0.00076771, 0.00010204, 0.00137871, 0.00010204, -0.00076771, 0.00013647, 0.00009132, -0.00027501, 0.00012156, 0.00003348, -0.00012354, 0.00009311, -0.00002676, -0.00006378, 0.00007533, -0.00004406, -0.00003253, 0.00005360, -0.00003746, -0.00000954, 0.00003292, -0.00002293, 0.00000412, 0.00000858, -0.00001927, 0.00000956, 0.00000206, -0.00000934, 0.00000732, -0.00000196, 0.00000126, 0.00000898, -0.00000233, -0.00000277, -0.00000081, 0.00000003, -0.00000477, -0.00000835, 0.00001024, 0.00000874, -0.00000600, 0.00000420, -0.00000263, -0.00000826, 0.00001017, -0.00000349, -0.00001113, 0.00001610, 0.00000736, -0.00001079, 0.00000456, 0.00000189, -0.00001414, -0.00000733, 0.00000239, 0.00000277, -0.00000412, -0.00001066, -0.00000143, 0.00000190, -0.00001356, -0.00001108, 0.00001057, 0.00001356, 0.00000111, -0.00000297, },
		{ -0.00000281, -0.00000279, -0.00000128, -0.00000210, 0.00000634, 0.00001515, 0.00000017, -0.00001662, -0.00000538, 0.00000323, -0.00000779, -0.00000872, 0.00000090, 0.00000396, -0.00000297, -0.00001431, -0.00000519, 0.00000768, -0.00000770, -0.00000305, 0.00001956, -0.00000101, -0.00001186, 0.00000937, -0.00000117, -0.00000824, 0.00000492, -0.00000374, 0.00000150, 0.00001372, -0.00000238, -0.00000789, -0.00000153, 0.00000012, -0.00000100, -0.00000600, 0.00000687, 0.00000696, -0.00000690, 0.00000865, -0.00000244, -0.00001019, 0.00001767, -0.00001205, -0.00001069, 0.00002116, -0.00002536, 0.00000882, 0.00002928, -0.00005288, 0.00002423, 0.00002734, -0.00008079, 0.00004924, 0.00001399, -0.00009761, 0.00009196, -0.00003001, -0.00010286, 0.00018164, -0.00013517, -0.00016749, 0.00030768, -0.00044362, -0.00055466, 0.00109267, 0.00086250, -0.00069362, -0.00027280, 0.00028191, -0.00022493, -0.00005992, 0.00014982, -0.00011977, 0.00001101, 0.00005864, -0.00009220, 0.00003727, 0.00002021, -0.00006819, 0.00003700, 0.00000451, -0.00003923, 0.00002896, -0.00000089, -0.00001398, 0.00001496, -0.00001144, -0.00000351, 0.00000855, -0.00000777, 0.00000170, 0.00000191, -0.00000089, 0.00000804, 0.00000107, -0.00000253, -0.00000166, -0.00000085, -0.00000059, -0.00001004, 0.00000228, 0.00001421, -0.00000341, -0.00000042, 0.00000353, -0.00001057, 0.00000429, 0.00000612, -0.00001399, 0.00000584, 0.00001564, -0.00000791, -0.00000179, 0.00000701, -0.00001031, -0.00001156, 0.00000019, 0.00000287, -0.00000030, -0.00000892, -0.00000751, 0.00000328, -0.00000529, -0.00001634, -0.00000024, 0.00001596, 0.00000837, },
		{ 0.00001540, 0.00000197, -0.00000482, -0.00000140, -0.00000117, -0.00000000, 0.00001176, 0.00001083, -0.00001109, -0.00001395, 0.00000209, -0.00000153, -0.00001084, -0.00000344, 0.00000444, 0.00000101, -0.00001167, -0.00001131, 0.00000610, -0.00000128, -0.00001039, 0.00001488, 0.00001114, -0.00001370, 0.00000234, 0.00000613, -0.00000952, 0.00000141, 0.00000043, -0.00000356, 0.00001206, 0.00000459, -0.00000772, -0.00000412, -0.00000095, 0.00000202, -0.00000674, 0.00000012, 0.00001343, -0.00000702, 0.00000171, 0.00000957, -0.00001756, 0.00001123, 0.00000787, -0.00002725, 0.00001936, -0.00000408, -0.00002392, 0.00004521, -0.00002665, -0.00002859, 0.00006378, -0.00005836, -0.00002227, 0.00007825, -0.00009489, 0.00000932, 0.00007972, -0.00014976, 0.00009056, 0.00007235, -0.00029180, 0.00018564, 0.00000473, -0.00080042, 0.00039393, 0.00132604, -0.00016643, -0.00066629, 0.00021927, -0.00000546, -0.00022839, 0.00014328, -0.00001626, -0.00008592, 0.00008804, -0.00005039, -0.00003131, 0.00006021, -0.00005368, -0.00001017, 0.00003768, -0.00003743, 0.00000192, 0.00002042, -0.00001713, 0.00000654, 0.00000142, -0.00001031, 0.00000611, -0.00000226, -0.00000227, 0.00000237, -0.00000106, 0.00000677, 0.00000329, -0.00000187, -0.00000092, -0.00000286, 0.00000133, -0.00000671, -0.00000620, 0.00001394, 0.00000379, -0.00000525, 0.00000587, -0.00000679, -0.00000426, 0.00001098, -0.00000860, -0.00000582, 0.00001708, 0.00000023, -0.00000756, 0.00000763, -0.00000356, -0.00001416, -0.00000309, 0.00000251, 0.00000133, -0.00000478, -0.00001061, -0.00000101, 0.00000195, -0.00001337, -0.00001121, 0.00001055, },
		{ -0.00000047, 0.00001693, 0.00001022, -0.00000368, -0.00000420, 0.00000033, -0.00000162, 0.00000418, 0.00001460, 0.00000057, -0.00001716, -0.00000536, 0.00000354, -0.00000804, -0.00000866, 0.00000246, 0.00000439, -0.00000731, -0.00001476, 0.00000085, 0.00000467, -0.00001196, 0.00000494, 0.00001936, -0.00000766, -0.00000656, 0.00000920, -0.00000590, -0.00000367, 0.00000323, -0.00000500, 0.00000734, 0.00000961, -0.00000479, -0.00000537, -0.00000381, 0.00000306, -0.00000251, -0.00000836, 0.00001447, 0.00000140, -0.00000984, 0.00001646, -0.00001109, -0.00000749, 0.00002518, -0.00002441, -0.00000382, 0.00002316, -0.00003692, 0.00002390, 0.00002179, -0.00006154, 0.00004667, 0.00000715, -0.00008214, 0.00007611, -0.00002030, -0.00008604, 0.00011545, -0.00007727, -0.00006425, 0.00018924, -0.00020593, -0.00008069, 0.00028119, -0.00059509, -0.00034033, 0.00124332, 0.00058617, -0.00074673, -0.00011300, 0.00021760, -0.00024596, 0.00000475, 0.00010413, -0.00011521, 0.00003834, 0.00002438, -0.00007403, 0.00004673, -0.00000424, -0.00004946, 0.00003644, -0.00000883, -0.00002498, 0.00002318, -0.00000450, -0.00000570, 0.00000775, -0.00000760, -0.00000013, 0.00000095, -0.00000264, 0.00000143, -0.00000166, 0.00000564, 0.00000559, -0.00000196, 0.00000044, -0.00000326, -0.00000037, -0.00000130, -0.00001043, 0.00000728, 0.00001153, -0.00000584, 0.00000294, 0.00000005, -0.00000985, 0.00000841, 0.00000105, -0.00001260, 0.00001069, 0.00000941, -0.00000887, 0.00000356, 0.00000342, -0.00001333, -0.00000752, 0.00000194, 0.00000154, -0.00000125, -0.00000936, -0.00000703, 0.00000372, -0.00000533, -0.00001629, },
		{ -0.00001351, -0.00001126, 0.00001077, 0.00001717, 0.00000263, -0.00000669, -0.00000117, 0.00000055, -0.00000137, 0.00001014, 0.00001124, -0.00001133, -0.00001436, 0.00000259, -0.00000144, -0.00001122, -0.00000255, 0.00000609, -0.00000188, -0.00001506, -0.00000594, 0.00000742, -0.00000825, -0.00000550, 0.00002024, 0.00000335, -0.00001220, 0.00000660, 0.00000009, -0.00000724, 0.00000331, -0.00000374, 0.00000202, 0.00001185, -0.00000098, -0.00000469, -0.00000569, -0.00000026, 0.00000430, -0.00001216, 0.00000618, 0.00001445, -0.00001587, 0.00000985, 0.00000688, -0.00002389, 0.00002367, -0.00000032, -0.00003084, 0.00003094, -0.00001757, -0.00001782, 0.00005330, -0.00004625, -0.00001066, 0.00006388, -0.00008161, 0.00000870, 0.00006554, -0.00011649, 0.00005109, 0.00004773, -0.00015619, 0.00014273, 0.00000230, -0.00027088, 0.00025573, -0.00015837, -0.00074911, 0.00066942, 0.00117726, -0.00037827, -0.00052248, 0.00024808, -0.00008548, -0.00016563, 0.00013827, -0.00004995, -0.00004693, 0.00006929, -0.00005890, -0.00000472, 0.00003888, -0.00005178, 0.00000462, 0.00002105, -0.00003101, 0.00000675, 0.00001090, -0.00000972, 0.00000458, -0.00000097, -0.00000336, 0.00000015, -0.00000199, 0.00000162, -0.00000261, 0.00000310, 0.00000861, -0.00000164, 0.00000014, -0.00000095, -0.00000356, 0.00000183, -0.00000853, -0.00000207, 0.00001465, -0.00000071, -0.00000246, 0.00000491, -0.00000926, 0.00000077, 0.00000867, -0.00001132, 0.00000014, 0.00001436, -0.00000450, -0.00000245, 0.00000739, -0.00000860, -0.00001180, 0.00000028, 0.00000175, 0.00000022, -0.00000557, -0.00001058, -0.00000010, 0.00000205, },
		{ 0.00000445, -0.00000571, -0.00001633, -0.00000044, 0.00001791, 0.00001189, -0.00000469, -0.00000552, 0.00000201, -0.00000141, 0.00000205, 0.00001431, 0.00000081, -0.00001780, -0.00000511, 0.00000409, -0.00000838, -0.00000864, 0.00000437, 0.00000368, -0.00001215, -0.00001225, 0.00000606, -0.00000171, -0.00001247, 0.00001397, 0.00001441, -0.00001160, 0.00000018, 0.00000517, -0.00000783, 0.00000123, -0.00000123, -0.00000264, 0.00001176, 0.00000283, -0.00000380, -0.00000428, -0.00000619, 0.00000762, -0.00000692, -0.00000781, 0.00002188, -0.00000825, -0.00000768, 0.00002235, -0.00002256, 0.00000142, 0.00002773, -0.00003755, 0.00000880, 0.00001853, -0.00004492, 0.00004154, 0.00000623, -0.00006140, 0.00006652, -0.00002067, -0.00006958, 0.00009355, -0.00005907, -0.00005758, 0.00012176, -0.00012044, -0.00000977, 0.00016738, -0.00025621, 0.00001970, 0.00019954, -0.00069707, -0.00007653, 0.00129475, 0.00030331, -0.00071999, 0.00001211, 0.00013465, -0.00023293, 0.00004847, 0.00005660, -0.00009465, 0.00004917, -0.00000312, -0.00004987, 0.00004356, -0.00001975, -0.00003035, 0.00002853, -0.00001459, -0.00001371, 0.00001526, -0.00000322, -0.00000175, 0.00000235, -0.00000182, -0.00000139, -0.00000318, 0.00000248, -0.00000171, -0.00000141, 0.00001050, 0.00000134, -0.00000240, 0.00000202, -0.00000463, 0.00000056, -0.00000306, -0.00000857, 0.00001083, 0.00000718, -0.00000552, 0.00000473, -0.00000384, -0.00000655, 0.00000989, -0.00000387, -0.00000859, 0.00001232, 0.00000311, -0.00000645, 0.00000672, -0.00000171, -0.00001378, -0.00000320, 0.00000234, 0.00000032, -0.00000211, -0.00001001, -0.00000611, },
		{ -0.00001033, 0.00000134, 0.00000191, -0.00001393, -0.00001106, 0.00001107, 0.00001879, 0.00000314, -0.00000859, -0.00000076, 0.00000209, -0.00000290, 0.00000879, 0.00001161, -0.00001177, -0.00001460, 0.00000333, -0.00000129, -0.00001177, -0.00000132, 0.00000732, -0.00000622, -0.00001634, 0.00000093, 0.00000476, -0.00001399, 0.00000332, 0.00002117, -0.00000487, -0.00000670, 0.00000725, -0.00000571, -0.00000200, 0.00000167, -0.00000599, 0.00000926, 0.00000729, -0.00000418, -0.00000079, -0.00000941, 0.00000308, 0.00000439, -0.00001736, 0.00001534, 0.00001040, -0.00002241, 0.00002068, -0.00000161, -0.00002521, 0.00003621, -0.00001444, -0.00002751, 0.00003956, -0.00003384, -0.00000419, 0.00005367, -0.00006365, 0.00001374, 0.00005182, -0.00009597, 0.00004272, 0.00003777, -0.00012208, 0.00008926, 0.00000226, -0.00013888, 0.00017752, -0.00007562, -0.00021332, 0.00028299, -0.00032571, -0.00061399, 0.00089055, 0.00095784, -0.00051344, -0.00036664, 0.00022995, -0.00013650, -0.00010196, 0.00011349, -0.00006432, -0.00001514, 0.00004429, -0.00005401, 0.00001156, 0.00001782, -0.00004189, 0.00001058, 0.00000782, -0.00002186, 0.00000609, 0.00000586, -0.00000353, 0.00000070, 0.00000075, -0.00000019, -0.00000551, 0.00000119, 0.00000188, -0.00000548, 0.00000840, 0.00000715, -0.00000462, 0.00000229, -0.00000199, -0.00000313, 0.00000114, -0.00000900, 0.00000251, 0.00001257, -0.00000328, 0.00000064, 0.00000197, -0.00000916, 0.00000498, 0.00000443, -0.00001107, 0.00000478, 0.00000938, -0.00000592, 0.00000254, 0.00000431, -0.00001200, -0.00000787, 0.00000236, 0.00000056, -0.00000058, -0.00000664, },
		{ -0.00000309, -0.00001076, -0.00000455, 0.00000521, -0.00000655, -0.00001619, -0.00000020, 0.00001877, 0.00001346, -0.00000583, -0.00000671, 0.00000372, -0.00000147, 0.00000009, 0.00001414, 0.00000081, -0.00001837, -0.00000463, 0.00000480, -0.00000886, -0.00000851, 0.00000648, 0.00000144, -0.00001647, -0.00000676, 0.00000855, -0.00000995, -0.00000797, 0.00002120, 0.00000565, -0.00001102, 0.00000581, -0.00000169, -0.00000570, 0.00000412, -0.00000678, 0.00000366, 0.00001242, -0.00000443, 0.00000048, -0.00000602, -0.00000694, 0.00001244, -0.00001396, -0.00000302, 0.00002650, -0.00002007, 0.00000022, 0.00002340, -0.00003372, 0.00001565, 0.00002293, -0.00004755, 0.00002473, 0.00000663, -0.00004599, 0.00005753, -0.00001519, -0.00005067, 0.00007834, -0.00005052, -0.00004411, 0.00009602, -0.00009377, -0.00001706, 0.00010754, -0.00014917, 0.00004974, 0.00011810, -0.00027374, 0.00011266, 0.00007267, -0.00072694, 0.00020052, 0.00124496, 0.00005124, -0.00063297, 0.00008991, 0.00005336, -0.00019561, 0.00006770, 0.00001744, -0.00006652, 0.00004545, -0.00001936, -0.00002675, 0.00003201, -0.00002523, -0.00001550, 0.00001752, -0.00001394, -0.00000724, 0.00000831, 0.00000055, -0.00000166, 0.00000034, 0.00000316, -0.00000541, -0.00000352, 0.00000552, -0.00000550, 0.00000192, 0.00001246, -0.00000290, -0.00000115, 0.00000212, -0.00000515, 0.00000104, -0.00000478, -0.00000500, 0.00001191, 0.00000298, -0.00000328, 0.00000429, -0.00000650, -0.00000206, 0.00000861, -0.00000705, -0.00000366, 0.00001076, -0.00000141, -0.00000208, 0.00000698, -0.00000687, -0.00001184, 0.00000057, 0.00000171, -0.00000061, },
		{ -0.00000067, -0.00000122, -0.00000805, -0.00000956, 0.00000319, 0.00000124, -0.00001439, -0.00001052, 0.00001129, 0.00002033, 0.00000351, -0.00001052, -0.00000021, 0.00000342, -0.00000444, 0.00000768, 0.00001178, -0.00001223, -0.00001463, 0.00000425, -0.00000116, -0.00001244, 0.00000035, 0.00000762, -0.00001148, -0.00001460, 0.00000752, -0.00000189, -0.00001603, 0.00001402, 0.00001659, -0.00001067, 0.00000118, 0.00000339, -0.00000880, 0.00000428, -0.00000353, -0.00000430, 0.00001559, -0.00000093, -0.00000333, 0.00000181, -0.00001404, 0.00000910, 0.00000098, -0.00002017, 0.00002579, 0.00000096, -0.00002390, 0.00003108, -0.00001488, -0.00001944, 0.00004494, -0.00003179, -0.00001654, 0.00004178, -0.00004897, 0.00001506, 0.00004460, -0.00007407, 0.00003968, 0.00002874, -0.00009686, 0.00007202, -0.00000022, -0.00010891, 0.00011470, -0.00004742, -0.00009996, 0.00018590, -0.00014524, -0.00012972, 0.00025748, -0.00046664, -0.00041161, 0.00102934, 0.00070344, -0.00056797, -0.00022601, 0.00018046, -0.00015527, -0.00004988, 0.00007924, -0.00006157, 0.00000445, 0.00002059, -0.00004057, 0.00001675, 0.00000204, -0.00002891, 0.00000915, 0.00000012, -0.00001346, 0.00000242, 0.00000504, -0.00000025, -0.00000257, 0.00000466, -0.00000089, -0.00000883, 0.00000499, -0.00000057, -0.00000526, 0.00001280, 0.00000365, -0.00000542, 0.00000365, -0.00000314, -0.00000225, -0.00000036, -0.00000767, 0.00000601, 0.00000890, -0.00000349, 0.00000257, -0.00000153, -0.00000671, 0.00000710, -0.00000006, -0.00000846, 0.00000689, 0.00000402, -0.00000430, 0.00000587, -0.00000051, -0.00001311, -0.00000330, 0.00000308, },
		{ 0.00000324, 0.00000042, -0.00000117, -0.00000441, -0.00001133, -0.00000226, 0.00000562, -0.00000776, -0.00001565, 0.00000006, 0.00001950, 0.00001496, -0.00000712, -0.00000775, 0.00000535, -0.00000170, -0.00000164, 0.00001388, 0.00000069, -0.00001871, -0.00000401, 0.00000561, -0.00000953, -0.00000809, 0.00000845, -0.00000242, -0.00001901, 0.00000084, 0.00000690, -0.00001757, 0.00000132, 0.00002390, -0.00000445, -0.00000556, 0.00000784, -0.00000905, 0.00000023, 0.00000312, -0.00001084, 0.00001262, 0.00000797, -0.00000947, 0.00000651, -0.00001091, -0.00000454, 0.00001554, -0.00002213, 0.00000617, 0.00002601, -0.00003101, 0.00001268, 0.00001778, -0.00004089, 0.00003234, 0.00001034, -0.00005084, 0.00004006, -0.00001088, -0.00003833, 0.00006747, -0.00003814, -0.00003040, 0.00007830, -0.00007561, -0.00001044, 0.00008158, -0.00011620, 0.00002705, 0.00007480, -0.00015624, 0.00010131, 0.00005074, -0.00025385, 0.00017727, -0.00007808, -0.00067546, 0.00045204, 0.00110948, -0.00014255, -0.00051282, 0.00012007, -0.00001060, -0.00014741, 0.00006544, -0.00000743, -0.00003931, 0.00003250, -0.00002380, -0.00000988, 0.00001763, -0.00002257, -0.00000724, 0.00000738, -0.00000959, -0.00000543, 0.00000426, 0.00000427, -0.00000367, 0.00000168, 0.00000549, -0.00000982, -0.00000113, 0.00000566, -0.00000793, 0.00000671, 0.00001148, -0.00000585, 0.00000063, 0.00000126, -0.00000483, 0.00000069, -0.00000549, -0.00000096, 0.00001060, 0.00000027, -0.00000058, 0.00000206, -0.00000701, 0.00000202, 0.00000543, -0.00000786, 0.00000048, 0.00000710, -0.00000322, 0.00000252, 0.00000456, -0.00001083, -0.00000811, },
		{ -0.00001189, 0.00000107, 0.00000260, -0.00000164, -0.00000194, -0.00000968, -0.00000799, 0.00000512, -0.00000009, -0.00001458, -0.00000973, 0.00001131, 0.00002187, 0.00000371, -0.00001244, 0.00000045, 0.00000452, -0.00000585, 0.00000663, 0.00001174, -0.00001250, -0.00001451, 0.00000528, -0.00000111, -0.00001312, 0.00000240, 0.00000651, -0.00001670, -0.00000956, 0.00001177, -0.00001120, -0.00001270, 0.00002357, 0.00000714, -0.00001168, 0.00000871, -0.00000427, -0.00000744, 0.00000895, -0.00001079, 0.00000215, 0.00001821, -0.00001058, 0.00000186, 0.00000117, -0.00001755, 0.00001660, -0.00000624, -0.00001862, 0.00003479, -0.00001250, -0.00001881, 0.00003759, -0.00003005, -0.00000650, 0.00004647, -0.00004829, 0.00000049, 0.00003568, -0.00005876, 0.00003605, 0.00002689, -0.00007415, 0.00006124, -0.00000099, -0.00008305, 0.00008953, -0.00004048, -0.00007912, 0.00012122, -0.00009034, -0.00004746, 0.00016529, -0.00019171, -0.00003799, 0.00018123, -0.00055424, -0.00017154, 0.00107349, 0.00045181, -0.00055287, -0.00011909, 0.00011852, -0.00014671, -0.00001638, 0.00004575, -0.00004773, 0.00001123, 0.00000364, -0.00002466, 0.00001331, -0.00000616, -0.00001737, 0.00000347, -0.00000211, -0.00000805, -0.00000153, 0.00000679, 0.00000001, -0.00000355, 0.00000845, -0.00000423, -0.00000913, 0.00000746, -0.00000394, -0.00000224, 0.00001496, -0.00000033, -0.00000456, 0.00000391, -0.00000380, -0.00000160, -0.00000185, -0.00000490, 0.00000749, 0.00000517, -0.00000199, 0.00000265, -0.00000416, -0.00000299, 0.00000685, -0.00000340, -0.00000483, 0.00000639, -0.00000008, -0.00000077, 0.00000674, -0.00000573, },
		{ 0.00000031, -0.00001285, -0.00000340, 0.00000457, -0.00000109, -0.00000146, -0.00000617, -0.00001137, 0.00000064, 0.00000532, -0.00000907, -0.00001464, 0.00000014, 0.00002016, 0.00001640, -0.00000856, -0.00000864, 0.00000686, -0.00000194, -0.00000318, 0.00001340, 0.00000065, -0.00001884, -0.00000332, 0.00000650, -0.00001040, -0.00000725, 0.00000978, -0.00000759, -0.00001864, 0.00000884, 0.00000075, -0.00002184, 0.00001352, 0.00002048, -0.00001272, 0.00000342, 0.00000480, -0.00001410, 0.00000818, -0.00000229, -0.00001092, 0.00002157, -0.00000130, -0.00001002, 0.00001184, -0.00001791, 0.00000160, 0.00001502, -0.00002908, 0.00001842, 0.00001923, -0.00003780, 0.00002685, 0.00000561, -0.00004103, 0.00004736, -0.00000818, -0.00004546, 0.00005057, -0.00003044, -0.00002254, 0.00006820, -0.00005739, -0.00000443, 0.00006557, -0.00008993, 0.00002412, 0.00005273, -0.00012107, 0.00006500, 0.00003058, -0.00013996, 0.00013352, -0.00002035, -0.00020126, 0.00019844, -0.00022470, -0.00054931, 0.00064508, 0.00091697, -0.00026502, -0.00038722, 0.00011230, -0.00004924, -0.00010118, 0.00004919, -0.00001718, -0.00001915, 0.00001684, -0.00001931, -0.00000144, 0.00000538, -0.00001550, -0.00000537, 0.00000072, -0.00000457, -0.00000665, 0.00000346, 0.00000621, -0.00000570, 0.00000509, 0.00000467, -0.00001269, 0.00000218, 0.00000338, -0.00000786, 0.00001112, 0.00000839, -0.00000704, 0.00000208, 0.00000010, -0.00000411, -0.00000026, -0.00000474, 0.00000225, 0.00000782, -0.00000047, 0.00000126, -0.00000081, -0.00000544, 0.00000452, 0.00000172, -0.00000666, 0.00000274, 0.00000293, -0.00000240, 0.00000593, },
		{ 0.00000320, 0.00000519, -0.00001022, -0.00000875, 0.00000459, 0.00000123, -0.00000220, -0.00000300, -0.00001127, -0.00000546, 0.00000665, -0.00000198, -0.00001422, -0.00000893, 0.00001111, 0.00002347, 0.00000371, -0.00001431, 0.00000113, 0.00000550, -0.00000705, 0.00000542, 0.00001164, -0.00001247, -0.00001436, 0.00000645, -0.00000120, -0.00001366, 0.00000458, 0.00000371, -0.00002060, -0.00000183, 0.00001195, -0.00002073, -0.00000379, 0.00002883, -0.00000483, -0.00000698, 0.00001295, -0.00001347, -0.00000137, 0.00000964, -0.00001740, 0.00001246, 0.00001517, -0.00001862, 0.00001008, -0.00000362, -0.00001787, 0.00002348, -0.00001621, -0.00001176, 0.00003943, -0.00002712, -0.00000743, 0.00003746, -0.00004320, 0.00001150, 0.00003912, -0.00005959, 0.00001999, 0.00002162, -0.00006011, 0.00005401, 0.00000389, -0.00006314, 0.00007331, -0.00003094, -0.00005721, 0.00009099, -0.00007416, -0.00003930, 0.00010750, -0.00011714, 0.00000680, 0.00012054, -0.00020576, 0.00004124, 0.00006840, -0.00057190, 0.00007023, 0.00102805, 0.00023472, -0.00048947, -0.00005312, 0.00006102, -0.00012097, -0.00000219, 0.00002040, -0.00003016, 0.00000828, -0.00000444, -0.00001147, 0.00000551, -0.00000726, -0.00001013, -0.00000296, -0.00000062, -0.00000614, -0.00000380, 0.00000907, -0.00000165, -0.00000187, 0.00001036, -0.00000829, -0.00000697, 0.00000763, -0.00000645, 0.00000237, 0.00001452, -0.00000357, -0.00000292, 0.00000332, -0.00000389, -0.00000149, -0.00000256, -0.00000172, 0.00000694, 0.00000259, 0.00000003, 0.00000118, -0.00000504, 0.00000061, 0.00000486, -0.00000495, -0.00000160, 0.00000409, -0.00000202, },
		{ -0.00000038, 0.00000015, 0.00000734, -0.00000466, -0.00001271, 0.00000150, 0.00000456, -0.00000243, -0.00000171, -0.00000835, -0.00001049, 0.00000376, 0.00000412, -0.00001007, -0.00001327, -0.00000012, 0.00002089, 0.00001775, -0.00001013, -0.00000943, 0.00000823, -0.00000204, -0.00000475, 0.00001269, 0.00000094, -0.00001889, -0.00000260, 0.00000744, -0.00001141, -0.00000593, 0.00000993, -0.00001321, -0.00001482, 0.00001510, -0.00000889, -0.00002064, 0.00002549, 0.00001112, -0.00001624, 0.00001313, -0.00000306, -0.00001455, 0.00001533, -0.00001095, -0.00000558, 0.00002735, -0.00001335, -0.00000517, 0.00001395, -0.00002488, 0.00001071, 0.00000985, -0.00003230, 0.00003111, 0.00000672, -0.00003798, 0.00003897, -0.00001092, -0.00003288, 0.00005655, -0.00002881, -0.00003177, 0.00005308, -0.00004750, -0.00000157, 0.00005890, -0.00006841, 0.00002161, 0.00004271, -0.00009016, 0.00005183, 0.00001585, -0.00010762, 0.00008859, -0.00001500, -0.00010477, 0.00013989, -0.00007979, -0.00012917, 0.00017130, -0.00033918, -0.00036962, 0.00075924, 0.00070208, -0.00031815, -0.00027815, 0.00008198, -0.00006265, -0.00006594, 0.00002796, -0.00001521, -0.00000850, 0.00000397, -0.00001057, -0.00000053, -0.00000180, -0.00000802, -0.00000781, -0.00000178, -0.00000115, -0.00000870, 0.00000492, 0.00000584, -0.00000614, 0.00000870, 0.00000144, -0.00001312, 0.00000467, 0.00000006, -0.00000525, 0.00001382, 0.00000445, -0.00000668, 0.00000277, -0.00000086, -0.00000346, -0.00000119, -0.00000288, 0.00000382, 0.00000491, 0.00000025, 0.00000161, -0.00000306, -0.00000266, 0.00000508, -0.00000130, -0.00000451, 0.00000298, },
		{ 0.00000031, 0.00000129, -0.00000183, 0.00000638, 0.00000182, -0.00001317, -0.00000429, 0.00000686, -0.00000059, -0.00000236, -0.00000460, -0.00001239, -0.00000208, 0.00000735, -0.00000409, -0.00001322, -0.00000839, 0.00001079, 0.00002514, 0.00000348, -0.00001612, 0.00000174, 0.00000653, -0.00000813, 0.00000391, 0.00001174, -0.00001217, -0.00001431, 0.00000780, -0.00000149, -0.00001388, 0.00000645, -0.00000064, -0.00002191, 0.00000704, 0.00000731, -0.00002755, 0.00000911, 0.00002762, -0.00001591, 0.00000246, 0.00001183, -0.00002117, 0.00000793, 0.00000522, -0.00002044, 0.00002410, 0.00000614, -0.00002217, 0.00001875, -0.00001197, -0.00001369, 0.00002732, -0.00002668, -0.00000018, 0.00003767, -0.00003935, 0.00000799, 0.00002969, -0.00005053, 0.00003064, 0.00002361, -0.00006249, 0.00003743, 0.00000231, -0.00005210, 0.00006468, -0.00001944, -0.00004329, 0.00007315, -0.00005443, -0.00002530, 0.00007632, -0.00009413, 0.00000138, 0.00007760, -0.00012278, 0.00005063, 0.00006273, -0.00018659, 0.00009022, -0.00005791, -0.00051736, 0.00027933, 0.00091290, 0.00007225, -0.00040298, -0.00002474, 0.00001909, -0.00008963, -0.00000294, 0.00000626, -0.00001507, 0.00000056, -0.00000475, -0.00000393, -0.00000226, -0.00000383, -0.00000786, -0.00000750, 0.00000215, -0.00000682, -0.00000367, 0.00001022, -0.00000358, 0.00000165, 0.00000966, -0.00001134, -0.00000377, 0.00000566, -0.00000691, 0.00000701, 0.00001197, -0.00000543, -0.00000125, 0.00000231, -0.00000365, -0.00000174, -0.00000217, 0.00000080, 0.00000517, 0.00000160, 0.00000142, -0.00000090, -0.00000416, 0.00000305, 0.00000223, -0.00000493, },
		{ -0.00000307, -0.00000254, 0.00000185, -0.00000204, 0.00000322, 0.00000667, -0.00000932, -0.00001058, 0.00000588, 0.00000343, -0.00000335, -0.00000225, -0.00001067, -0.00000845, 0.00000656, 0.00000211, -0.00001041, -0.00001182, -0.00000074, 0.00002179, 0.00001892, -0.00001179, -0.00001020, 0.00000958, -0.00000193, -0.00000659, 0.00001195, 0.00000168, -0.00001904, -0.00000177, 0.00000836, -0.00001245, -0.00000430, 0.00000849, -0.00001804, -0.00000786, 0.00001762, -0.00001971, -0.00001328, 0.00003378, -0.00000208, -0.00001347, 0.00002028, -0.00001375, -0.00000953, 0.00001885, -0.00002021, 0.00000484, 0.00002734, -0.00002495, 0.00000420, 0.00001122, -0.00002927, 0.00002080, 0.00000041, -0.00003001, 0.00004104, -0.00000924, -0.00003080, 0.00004555, -0.00002804, -0.00001749, 0.00005703, -0.00004688, -0.00001249, 0.00004651, -0.00005799, 0.00002003, 0.00004147, -0.00006875, 0.00004196, 0.00001502, -0.00007671, 0.00006671, -0.00002065, -0.00007988, 0.00009364, -0.00005174, -0.00006003, 0.00012088, -0.00011553, -0.00005572, 0.00010272, -0.00040040, -0.00016684, 0.00079007, 0.00049749, -0.00031596, -0.00019815, 0.00004509, -0.00005707, -0.00004526, 0.00000931, -0.00000712, -0.00000630, -0.00000314, -0.00000219, -0.00000419, -0.00000343, -0.00000306, -0.00001175, -0.00000106, -0.00000022, -0.00000978, 0.00000705, 0.00000383, -0.00000442, 0.00001092, -0.00000269, -0.00001138, 0.00000530, -0.00000275, -0.00000098, 0.00001424, 0.00000080, -0.00000535, 0.00000276, -0.00000151, -0.00000309, -0.00000149, -0.00000073, 0.00000377, 0.00000290, 0.00000142, 0.00000068, -0.00000390, 0.00000017, 0.00000417, },
		{ 0.00000399, -0.00000004, -0.00000418, 0.00000078, -0.00000087, -0.00000021, 0.00000802, -0.00000248, -0.00001422, 0.00000075, 0.00000772, -0.00000242, -0.00000233, -0.00000678, -0.00001256, 0.00000177, 0.00000694, -0.00000596, -0.00001167, -0.00000831, 0.00001052, 0.00002684, 0.00000302, -0.00001791, 0.00000230, 0.00000778, -0.00000931, 0.00000209, 0.00001233, -0.00001179, -0.00001438, 0.00000933, -0.00000199, -0.00001367, 0.00000744, -0.00000584, -0.00001983, 0.00001491, -0.00000152, -0.00002915, 0.00002296, 0.00001958, -0.00002273, 0.00001400, 0.00000498, -0.00002454, 0.00001764, -0.00000367, -0.00001781, 0.00003365, -0.00000725, -0.00001959, 0.00002487, -0.00002186, -0.00000514, 0.00002634, -0.00003483, 0.00001398, 0.00002905, -0.00004598, 0.00002385, 0.00001549, -0.00004962, 0.00004639, 0.00000297, -0.00005595, 0.00004867, -0.00001797, -0.00003633, 0.00006573, -0.00003797, -0.00001924, 0.00006131, -0.00006657, 0.00000519, 0.00004970, -0.00009704, 0.00003392, 0.00003968, -0.00010808, 0.00007505, 0.00000595, -0.00014211, 0.00009899, -0.00017151, -0.00040290, 0.00043035, 0.00075694, -0.00002952, -0.00031614, -0.00002334, -0.00000340, -0.00006219, -0.00001170, 0.00000228, -0.00000595, -0.00000704, -0.00000054, -0.00000225, -0.00000693, 0.00000089, -0.00000938, -0.00000904, 0.00000416, -0.00000843, -0.00000166, 0.00000956, -0.00000433, 0.00000556, 0.00000676, -0.00001243, -0.00000103, 0.00000258, -0.00000508, 0.00001038, 0.00000829, -0.00000589, 0.00000002, 0.00000124, -0.00000339, -0.00000187, -0.00000098, 0.00000199, 0.00000330, 0.00000177, 0.00000161, -0.00000249, -0.00000223, },
		{ -0.00000331, 0.00000215, 0.00000266, -0.00000377, -0.00000147, 0.00000040, -0.00000208, 0.00000579, 0.00000433, -0.00001290, -0.00000690, 0.00000920, 0.00000156, -0.00000375, -0.00000336, -0.00001267, -0.00000528, 0.00000849, -0.00000034, -0.00000992, -0.00001062, -0.00000159, 0.00002289, 0.00001988, -0.00001351, -0.00001101, 0.00001108, -0.00000170, -0.00000887, 0.00001149, 0.00000283, -0.00001946, -0.00000073, 0.00000917, -0.00001333, -0.00000274, 0.00000546, -0.00002073, 0.00000090, 0.00001528, -0.00002872, -0.00000092, 0.00003568, -0.00001568, -0.00000480, 0.00002221, -0.00002415, -0.00000003, 0.00001694, -0.00002685, 0.00001818, 0.00002048, -0.00003272, 0.00001563, 0.00000358, -0.00002908, 0.00002918, -0.00001149, -0.00002190, 0.00004542, -0.00002515, -0.00001756, 0.00004450, -0.00004157, 0.00000194, 0.00004815, -0.00005822, 0.00000806, 0.00003227, -0.00005953, 0.00003754, 0.00001996, -0.00005886, 0.00005251, -0.00001102, -0.00005360, 0.00006640, -0.00004879, -0.00004539, 0.00008108, -0.00007237, -0.00001714, 0.00008359, -0.00012238, 0.00000104, 0.00000963, -0.00039909, 0.00002649, 0.00074836, 0.00032746, -0.00027903, -0.00014964, 0.00001397, -0.00004168, -0.00003772, -0.00000241, 0.00000154, -0.00000918, -0.00000431, 0.00000274, -0.00000880, -0.00000106, -0.00000177, -0.00001475, 0.00000105, -0.00000137, -0.00000903, 0.00000840, 0.00000145, -0.00000102, 0.00001093, -0.00000621, -0.00000852, 0.00000405, -0.00000391, 0.00000365, 0.00001259, -0.00000187, -0.00000367, 0.00000229, -0.00000197, -0.00000283, -0.00000098, 0.00000078, 0.00000280, 0.00000214, 0.00000206, -0.00000072, },
		{ 0.00000079, -0.00000293, -0.00000022, 0.00000379, -0.00000151, -0.00000358, 0.00000038, -0.00000180, 0.00000188, 0.00000797, -0.00000683, -0.00001336, 0.00000566, 0.00000726, -0.00000388, -0.00000245, -0.00000932, -0.00001144, 0.00000547, 0.00000546, -0.00000714, -0.00000990, -0.00000872, 0.00001046, 0.00002848, 0.00000238, -0.00001974, 0.00000292, 0.00000932, -0.00001088, 0.00000017, 0.00001358, -0.00001158, -0.00001447, 0.00001103, -0.00000263, -0.00001312, 0.00000708, -0.00001073, -0.00001441, 0.00001961, -0.00001248, -0.00002437, 0.00003408, 0.00000639, -0.00002308, 0.00002415, -0.00000607, -0.00002205, 0.00002459, -0.00001469, -0.00000913, 0.00003792, -0.00002184, -0.00001110, 0.00002609, -0.00003049, 0.00000619, 0.00002007, -0.00003821, 0.00002756, 0.00001502, -0.00004520, 0.00003625, -0.00000203, -0.00004026, 0.00005491, -0.00001832, -0.00004146, 0.00005125, -0.00003469, -0.00001647, 0.00005739, -0.00004801, 0.00000343, 0.00004144, -0.00006570, 0.00002750, 0.00001821, -0.00008421, 0.00005209, 0.00000344, -0.00007925, 0.00007685, -0.00003688, -0.00008662, 0.00006811, -0.00024965, -0.00025192, 0.00051141, 0.00059075, -0.00007707, -0.00024461, -0.00003554, -0.00000892, -0.00004393, -0.00002165, 0.00000475, -0.00000319, -0.00001133, 0.00000443, -0.00000456, -0.00000746, 0.00000418, -0.00001252, -0.00000798, 0.00000436, -0.00000941, 0.00000102, 0.00000745, -0.00000325, 0.00000843, 0.00000281, -0.00001158, 0.00000026, -0.00000028, -0.00000156, 0.00001181, 0.00000446, -0.00000528, 0.00000079, 0.00000022, -0.00000320, -0.00000145, 0.00000026, 0.00000193, 0.00000222, 0.00000222, },
		{ 0.00000224, 0.00000169, -0.00000163, -0.00000189, 0.00000301, 0.00000135, -0.00000406, -0.00000133, -0.00000047, -0.00000108, 0.00000722, 0.00000099, -0.00001494, -0.00000226, 0.00001115, -0.00000054, -0.00000380, -0.00000516, -0.00001380, -0.00000138, 0.00000917, -0.00000269, -0.00000871, -0.00000994, -0.00000247, 0.00002415, 0.00002059, -0.00001534, -0.00001186, 0.00001291, -0.00000160, -0.00001162, 0.00001166, 0.00000417, -0.00002022, 0.00000059, 0.00000984, -0.00001394, -0.00000177, 0.00000137, -0.00002030, 0.00000941, 0.00000834, -0.00003312, 0.00001359, 0.00003021, -0.00002595, 0.00000743, 0.00001785, -0.00003104, 0.00001137, 0.00000956, -0.00002818, 0.00003100, 0.00000776, -0.00003425, 0.00002587, -0.00000736, -0.00002359, 0.00003343, -0.00002305, -0.00000933, 0.00004279, -0.00003740, -0.00000131, 0.00003584, -0.00004815, 0.00002098, 0.00003180, -0.00006047, 0.00002530, 0.00001380, -0.00005220, 0.00004744, -0.00000065, -0.00004186, 0.00005193, -0.00002973, -0.00002722, 0.00005266, -0.00006336, -0.00001285, 0.00005646, -0.00007460, 0.00001391, 0.00003952, -0.00010332, 0.00002831, -0.00008553, -0.00033962, 0.00018305, 0.00065582, 0.00020424, -0.00022841, -0.00012705, -0.00000490, -0.00002519, -0.00003865, -0.00000652, 0.00000705, -0.00001321, -0.00000158, 0.00000337, -0.00001154, 0.00000279, -0.00000361, -0.00001558, 0.00000278, -0.00000338, -0.00000668, 0.00000824, -0.00000005, 0.00000293, 0.00000891, -0.00000818, -0.00000581, 0.00000167, -0.00000304, 0.00000744, 0.00000960, -0.00000336, -0.00000203, 0.00000155, -0.00000239, -0.00000227, -0.00000001, 0.00000118, 0.00000191, },
		{ 0.00000130, 0.00000216, 0.00000208, -0.00000028, -0.00000221, 0.00000112, 0.00000318, -0.00000237, -0.00000364, -0.00000001, -0.00000158, 0.00000358, 0.00000652, -0.00001055, -0.00001080, 0.00000982, 0.00000579, -0.00000478, -0.00000308, -0.00001175, -0.00000898, 0.00000837, 0.00000328, -0.00000739, -0.00000831, -0.00000952, 0.00001069, 0.00002998, 0.00000155, -0.00002168, 0.00000377, 0.00001104, -0.00001305, -0.00000147, 0.00001546, -0.00001178, -0.00001445, 0.00001281, -0.00000327, -0.00001252, 0.00000523, -0.00001395, -0.00000672, 0.00001972, -0.00002261, -0.00001387, 0.00003934, -0.00000866, -0.00001670, 0.00002974, -0.00001832, -0.00001404, 0.00002638, -0.00002452, 0.00000400, 0.00003507, -0.00003393, 0.00000128, 0.00002155, -0.00003525, 0.00001768, 0.00000963, -0.00003556, 0.00003732, -0.00000144, -0.00003719, 0.00004220, -0.00001872, -0.00002458, 0.00005428, -0.00003562, -0.00002255, 0.00004510, -0.00004440, 0.00000288, 0.00004235, -0.00004831, 0.00001985, 0.00001906, -0.00005376, 0.00003752, -0.00001035, -0.00006097, 0.00005412, -0.00002285, -0.00004577, 0.00005928, -0.00005779, -0.00003652, 0.00000831, -0.00027857, -0.00009285, 0.00052511, 0.00043986, -0.00008520, -0.00019521, -0.00004949, -0.00000380, -0.00003561, -0.00002812, 0.00000921, -0.00000480, -0.00001142, 0.00000733, -0.00000808, -0.00000471, 0.00000472, -0.00001523, -0.00000563, 0.00000287, -0.00000889, 0.00000313, 0.00000486, -0.00000059, 0.00000947, -0.00000088, -0.00000958, -0.00000008, -0.00000186, 0.00000259, 0.00001131, 0.00000119, -0.00000404, 0.00000114, -0.00000083, -0.00000284, -0.00000042, 0.00000077, },
		{ 0.00000071, 0.00000061, 0.00000183, 0.00000234, 0.00000055, -0.00000151, -0.00000050, 0.00000309, 0.00000054, -0.00000466, -0.00000154, -0.00000036, 0.00000016, 0.00000738, -0.00000268, -0.00001524, 0.00000268, 0.00001166, -0.00000243, -0.00000382, -0.00000751, -0.00001360, 0.00000263, 0.00000856, -0.00000442, -0.00000710, -0.00000987, -0.00000317, 0.00002549, 0.00002106, -0.00001731, -0.00001264, 0.00001514, -0.00000195, -0.00001460, 0.00001267, 0.00000539, -0.00002122, 0.00000221, 0.00001040, -0.00001431, -0.00000190, -0.00000274, -0.00001656, 0.00001544, -0.00000148, -0.00003130, 0.00002656, 0.00001852, -0.00003010, 0.00001977, 0.00000813, -0.00003222, 0.00002137, -0.00000144, -0.00002305, 0.00003973, -0.00000803, -0.00002893, 0.00003195, -0.00001886, -0.00001373, 0.00003222, -0.00003140, 0.00000494, 0.00003356, -0.00004330, 0.00001408, 0.00002171, -0.00004636, 0.00003529, 0.00001175, -0.00005369, 0.00003573, -0.00000450, -0.00003838, 0.00004844, -0.00001601, -0.00002254, 0.00004200, -0.00003790, -0.00000428, 0.00003068, -0.00006330, 0.00001055, 0.00002799, -0.00006148, 0.00002745, 0.00000088, -0.00006834, 0.00002194, -0.00016085, -0.00023808, 0.00028617, 0.00053876, 0.00012800, -0.00018068, -0.00012048, -0.00001098, -0.00001344, -0.00004253, -0.00000510, 0.00000818, -0.00001545, 0.00000220, 0.00000077, -0.00001117, 0.00000578, -0.00000698, -0.00001428, 0.00000310, -0.00000492, -0.00000363, 0.00000673, 0.00000000, 0.00000619, 0.00000574, -0.00000849, -0.00000416, -0.00000074, -0.00000052, 0.00000967, 0.00000609, -0.00000378, -0.00000060, 0.00000058, -0.00000274, -0.00000110, },
		{ -0.00000194, 0.00000079, 0.00000021, 0.00000103, 0.00000266, 0.00000100, -0.00000076, -0.00000093, 0.00000160, 0.00000278, -0.00000324, -0.00000423, 0.00000030, -0.00000076, 0.00000445, 0.00000413, -0.00001313, -0.00000698, 0.00001275, 0.00000379, -0.00000519, -0.00000445, -0.00001345, -0.00000551, 0.00000997, 0.00000097, -0.00000673, -0.00000722, -0.00001047, 0.00001120, 0.00003129, 0.00000052, -0.00002370, 0.00000504, 0.00001268, -0.00001586, -0.00000246, 0.00001774, -0.00001249, -0.00001420, 0.00001463, -0.00000378, -0.00001240, 0.00000232, -0.00001442, 0.00000133, 0.00001513, -0.00002898, -0.00000016, 0.00003717, -0.00002166, -0.00000541, 0.00002902, -0.00002833, -0.00000264, 0.00002221, -0.00003005, 0.00001842, 0.00002528, -0.00004045, 0.00001430, 0.00001214, -0.00003460, 0.00002652, -0.00000257, -0.00002731, 0.00004098, -0.00001668, -0.00002410, 0.00004056, -0.00003074, -0.00000639, 0.00004502, -0.00004551, -0.00000369, 0.00003245, -0.00004572, 0.00001762, 0.00002484, -0.00004036, 0.00002732, -0.00000017, -0.00003550, 0.00003491, -0.00002986, -0.00003480, 0.00004290, -0.00003476, -0.00001731, 0.00003067, -0.00005585, -0.00000543, -0.00006275, -0.00025640, 0.00004778, 0.00048584, 0.00032030, -0.00007161, -0.00016678, -0.00005760, 0.00000481, -0.00003461, -0.00002942, 0.00001223, -0.00000786, -0.00000844, 0.00000704, -0.00001036, -0.00000055, 0.00000266, -0.00001626, -0.00000348, 0.00000058, -0.00000691, 0.00000397, 0.00000287, 0.00000275, 0.00000867, -0.00000346, -0.00000747, -0.00000148, -0.00000172, 0.00000630, 0.00000936, -0.00000120, -0.00000250, 0.00000111, -0.00000194, },
		{ -0.00000067, -0.00000271, 0.00000061, 0.00000049, -0.00000018, 0.00000269, 0.00000168, -0.00000061, -0.00000043, 0.00000024, 0.00000302, -0.00000006, -0.00000587, -0.00000132, 0.00000051, 0.00000100, 0.00000647, -0.00000609, -0.00001386, 0.00000724, 0.00001091, -0.00000381, -0.00000423, -0.00000996, -0.00001193, 0.00000598, 0.00000699, -0.00000517, -0.00000553, -0.00001032, -0.00000356, 0.00002681, 0.00002129, -0.00001948, -0.00001315, 0.00001767, -0.00000298, -0.00001745, 0.00001453, 0.00000622, -0.00002234, 0.00000410, 0.00001096, -0.00001472, -0.00000333, -0.00000553, -0.00001035, 0.00001736, -0.00001139, -0.00002347, 0.00003456, 0.00000359, -0.00002718, 0.00002870, -0.00000433, -0.00002732, 0.00002707, -0.00001295, -0.00001231, 0.00004182, -0.00002315, -0.00001812, 0.00003216, -0.00002794, -0.00000174, 0.00002580, -0.00003450, 0.00001764, 0.00001990, -0.00004193, 0.00002512, 0.00000571, -0.00003707, 0.00004193, -0.00000742, -0.00004024, 0.00003790, -0.00001864, -0.00002196, 0.00004172, -0.00002362, -0.00000575, 0.00002687, -0.00003559, 0.00001025, 0.00000728, -0.00005161, 0.00002114, 0.00000378, -0.00004004, 0.00002356, -0.00002317, -0.00003106, -0.00001272, -0.00020051, -0.00011755, 0.00033206, 0.00042156, 0.00008977, -0.00014525, -0.00011981, -0.00000794, -0.00000850, -0.00004502, -0.00000143, 0.00000591, -0.00001473, 0.00000461, -0.00000296, -0.00000805, 0.00000660, -0.00001016, -0.00001180, 0.00000197, -0.00000519, -0.00000094, 0.00000461, 0.00000151, 0.00000799, 0.00000249, -0.00000769, -0.00000382, -0.00000220, 0.00000280, 0.00001020, 0.00000271, -0.00000340, 0.00000055, },
		{ -0.00000083, 0.00000062, -0.00000295, -0.00000028, 0.00000144, -0.00000114, 0.00000176, 0.00000289, -0.00000071, -0.00000018, 0.00000015, 0.00000160, 0.00000273, -0.00000469, -0.00000472, 0.00000150, -0.00000003, 0.00000443, 0.00000129, -0.00001429, -0.00000247, 0.00001423, 0.00000178, -0.00000539, -0.00000648, -0.00001388, -0.00000168, 0.00001014, -0.00000083, -0.00000550, -0.00000682, -0.00001133, 0.00001192, 0.00003236, -0.00000076, -0.00002568, 0.00000682, 0.00001392, -0.00001909, -0.00000253, 0.00002010, -0.00001369, -0.00001365, 0.00001653, -0.00000420, -0.00001325, -0.00000071, -0.00001176, 0.00000752, 0.00000724, -0.00002965, 0.00001318, 0.00002815, -0.00002937, 0.00000754, 0.00002221, -0.00003342, 0.00000897, 0.00001325, -0.00002941, 0.00003045, 0.00001073, -0.00003992, 0.00002462, 0.00000026, -0.00002862, 0.00003064, -0.00001360, -0.00001546, 0.00003791, -0.00002749, -0.00000932, 0.00003231, -0.00003565, 0.00000995, 0.00002984, -0.00004672, 0.00001109, 0.00001702, -0.00003964, 0.00002532, 0.00000921, -0.00002768, 0.00002596, -0.00001220, -0.00001682, 0.00002285, -0.00003748, -0.00001287, 0.00002459, -0.00003268, -0.00000070, 0.00000146, -0.00003703, -0.00000041, -0.00012560, -0.00019271, 0.00015109, 0.00041453, 0.00023730, -0.00005193, -0.00015306, -0.00005739, 0.00001140, -0.00003682, -0.00002644, 0.00001226, -0.00000988, -0.00000451, 0.00000413, -0.00001014, 0.00000305, -0.00000083, -0.00001547, -0.00000245, -0.00000141, -0.00000418, 0.00000353, 0.00000212, 0.00000575, 0.00000669, -0.00000470, -0.00000606, -0.00000305, -0.00000008, 0.00000886, 0.00000659, -0.00000266, },
		{ -0.00000018, -0.00000241, 0.00000133, -0.00000218, -0.00000189, 0.00000230, -0.00000090, -0.00000028, 0.00000393, 0.00000001, -0.00000099, 0.00000101, 0.00000026, 0.00000331, -0.00000108, -0.00000737, -0.00000014, 0.00000156, 0.00000117, 0.00000482, -0.00000879, -0.00001103, 0.00001082, 0.00000935, -0.00000461, -0.00000534, -0.00001187, -0.00000901, 0.00000808, 0.00000506, -0.00000490, -0.00000441, -0.00001107, -0.00000359, 0.00002803, 0.00002123, -0.00002181, -0.00001320, 0.00002024, -0.00000477, -0.00001976, 0.00001705, 0.00000648, -0.00002344, 0.00000627, 0.00001162, -0.00001567, -0.00000573, -0.00000585, -0.00000343, 0.00001476, -0.00001841, -0.00001170, 0.00003553, -0.00001069, -0.00001838, 0.00003187, -0.00001611, -0.00001784, 0.00002700, -0.00002166, 0.00000142, 0.00003660, -0.00003408, -0.00000471, 0.00002658, -0.00003238, 0.00000951, 0.00001589, -0.00003183, 0.00002596, 0.00000506, -0.00003435, 0.00002972, -0.00000825, -0.00002314, 0.00004012, -0.00002172, -0.00002397, 0.00003268, -0.00002617, -0.00000700, 0.00003035, -0.00002346, 0.00000506, 0.00001150, -0.00002579, 0.00001443, -0.00001133, -0.00003408, 0.00001949, -0.00001064, -0.00001867, 0.00000750, -0.00002910, -0.00000439, -0.00006282, -0.00019842, -0.00000193, 0.00032855, 0.00032173, 0.00007600, -0.00012429, -0.00011759, -0.00000113, -0.00000926, -0.00004404, 0.00000158, 0.00000231, -0.00001160, 0.00000450, -0.00000578, -0.00000364, 0.00000517, -0.00001200, -0.00000936, 0.00000005, -0.00000413, 0.00000071, 0.00000280, 0.00000379, 0.00000824, -0.00000005, -0.00000663, -0.00000439, -0.00000228, 0.00000599, 0.00000930, },
		{ 0.00001001, 0.00000356, -0.00000327, 0.00000082, -0.00000041, -0.00000343, 0.00000197, 0.00000087, -0.00000245, 0.00000348, 0.00000210, -0.00000226, 0.00000129, 0.00000049, 0.00000176, 0.00000260, -0.00000685, -0.00000432, 0.00000322, 0.00000013, 0.00000370, -0.00000158, -0.00001388, 0.00000206, 0.00001429, 0.00000017, -0.00000583, -0.00000881, -0.00001276, 0.00000172, 0.00000912, -0.00000169, -0.00000414, -0.00000706, -0.00001189, 0.00001275, 0.00003316, -0.00000233, -0.00002742, 0.00000899, 0.00001449, -0.00002239, -0.00000163, 0.00002225, -0.00001527, -0.00001276, 0.00001864, -0.00000484, -0.00001531, -0.00000254, -0.00000665, 0.00001011, -0.00000139, -0.00002440, 0.00002258, 0.00001482, -0.00003022, 0.00001853, 0.00001140, -0.00003247, 0.00001763, 0.00000219, -0.00002265, 0.00003702, -0.00000511, -0.00003289, 0.00002982, -0.00001109, -0.00001891, 0.00002942, -0.00002093, -0.00000298, 0.00002927, -0.00003207, 0.00000357, 0.00002020, -0.00003314, 0.00002091, 0.00001278, -0.00004032, 0.00001929, 0.00000277, -0.00002895, 0.00002579, -0.00000140, -0.00001451, 0.00001835, -0.00001567, -0.00000285, 0.00000669, -0.00003422, 0.00000014, 0.00000637, -0.00002115, 0.00000192, -0.00001927, -0.00001179, -0.00002016, -0.00016432, -0.00010503, 0.00020900, 0.00033277, 0.00018695, -0.00003650, -0.00014620, -0.00005047, 0.00001335, -0.00003857, -0.00002144, 0.00000964, -0.00000967, -0.00000159, 0.00000018, -0.00000757, 0.00000487, -0.00000431, -0.00001356, -0.00000270, -0.00000235, -0.00000160, 0.00000241, 0.00000263, 0.00000769, 0.00000439, -0.00000500, -0.00000561, -0.00000394, 0.00000241, },
		{ -0.00000107, 0.00000841, 0.00000742, -0.00000238, -0.00000096, 0.00000153, -0.00000368, -0.00000007, 0.00000318, -0.00000305, 0.00000090, 0.00000464, -0.00000228, -0.00000020, 0.00000203, 0.00000004, 0.00000394, -0.00000303, -0.00000840, 0.00000203, 0.00000213, 0.00000073, 0.00000273, -0.00001042, -0.00000718, 0.00001303, 0.00000753, -0.00000506, -0.00000718, -0.00001263, -0.00000548, 0.00000866, 0.00000346, -0.00000389, -0.00000399, -0.00001185, -0.00000328, 0.00002905, 0.00002084, -0.00002415, -0.00001272, 0.00002250, -0.00000712, -0.00002124, 0.00001993, 0.00000616, -0.00002443, 0.00000879, 0.00001231, -0.00001769, -0.00000817, -0.00000323, 0.00000205, 0.00000874, -0.00002041, 0.00000063, 0.00002951, -0.00002082, -0.00000659, 0.00002883, -0.00002418, -0.00000663, 0.00002158, -0.00002515, 0.00001460, 0.00002548, -0.00003860, 0.00000787, 0.00001696, -0.00003140, 0.00001752, 0.00000515, -0.00002451, 0.00002841, -0.00000766, -0.00002318, 0.00002775, -0.00001716, -0.00000846, 0.00003138, -0.00002884, -0.00000892, 0.00002274, -0.00002679, 0.00000345, 0.00001812, -0.00001767, 0.00000874, 0.00000011, -0.00001319, 0.00000965, -0.00002138, -0.00001711, 0.00000971, -0.00001376, -0.00000432, -0.00001249, -0.00001965, 0.00000336, -0.00011213, -0.00015895, 0.00008978, 0.00029115, 0.00024772, 0.00007320, -0.00011463, -0.00011043, 0.00000474, -0.00001285, -0.00003972, 0.00000241, -0.00000062, -0.00000761, 0.00000216, -0.00000655, 0.00000046, 0.00000237, -0.00001229, -0.00000779, -0.00000172, -0.00000228, 0.00000124, 0.00000191, 0.00000599, 0.00000738, -0.00000164, -0.00000599, -0.00000513, },
		{ -0.00000583, -0.00000375, 0.00000498, 0.00000988, 0.00000065, -0.00000308, 0.00000222, -0.00000194, -0.00000293, 0.00000415, -0.00000109, -0.00000274, 0.00000553, 0.00000010, -0.00000282, 0.00000297, 0.00000013, 0.00000251, 0.00000172, -0.00000925, -0.00000259, 0.00000475, -0.00000036, 0.00000250, -0.00000411, -0.00001195, 0.00000593, 0.00001330, -0.00000086, -0.00000688, -0.00001080, -0.00001025, 0.00000396, 0.00000752, -0.00000146, -0.00000314, -0.00000776, -0.00001204, 0.00001360, 0.00003362, -0.00000411, -0.00002873, 0.00001130, 0.00001432, -0.00002536, 0.00000009, 0.00002401, -0.00001713, -0.00001149, 0.00002105, -0.00000623, -0.00001828, -0.00000199, -0.00000070, 0.00000847, -0.00000785, -0.00001492, 0.00002568, 0.00000088, -0.00002478, 0.00002480, -0.00000023, -0.00002633, 0.00002134, -0.00000773, -0.00001170, 0.00003671, -0.00001856, -0.00002164, 0.00002914, -0.00001923, -0.00000805, 0.00002377, -0.00002326, 0.00000720, 0.00001762, -0.00003040, 0.00001191, 0.00000772, -0.00002496, 0.00002472, -0.00000212, -0.00002917, 0.00002052, -0.00000729, -0.00001718, 0.00002082, -0.00000582, -0.00000447, 0.00000842, -0.00001201, 0.00000359, -0.00000793, -0.00002408, 0.00000313, -0.00000606, -0.00000690, -0.00000672, -0.00002664, 0.00000857, -0.00005605, -0.00017051, -0.00001366, 0.00022416, 0.00025772, 0.00015982, -0.00002949, -0.00013969, -0.00004055, 0.00001085, -0.00003779, -0.00001674, 0.00000582, -0.00000748, -0.00000071, -0.00000314, -0.00000373, 0.00000473, -0.00000678, -0.00001154, -0.00000370, -0.00000212, 0.00000020, 0.00000136, 0.00000390, 0.00000839, 0.00000245, -0.00000503, },
		{ -0.00000256, -0.00000600, -0.00000534, 0.00000090, 0.00000976, 0.00000499, -0.00000382, 0.00000077, 0.00000099, -0.00000463, 0.00000249, 0.00000254, -0.00000500, 0.00000327, 0.00000404, -0.00000431, 0.00000153, 0.00000214, 0.00000024, 0.00000436, -0.00000593, -0.00000813, 0.00000461, 0.00000187, -0.00000007, 0.00000043, -0.00001069, -0.00000294, 0.00001378, 0.00000593, -0.00000560, -0.00000946, -0.00001183, -0.00000216, 0.00000790, 0.00000272, -0.00000258, -0.00000432, -0.00001240, -0.00000271, 0.00002980, 0.00002010, -0.00002628, -0.00001184, 0.00002421, -0.00000970, -0.00002183, 0.00002290, 0.00000529, -0.00002528, 0.00001187, 0.00001269, -0.00002105, -0.00000935, 0.00000179, 0.00000427, 0.00000164, -0.00001685, 0.00000997, 0.00001858, -0.00002475, 0.00000469, 0.00002103, -0.00002689, 0.00000317, 0.00001289, -0.00002274, 0.00002395, 0.00001139, -0.00003641, 0.00001674, 0.00000602, -0.00002581, 0.00002094, -0.00000379, -0.00001485, 0.00002520, -0.00001585, -0.00001165, 0.00002092, -0.00001975, 0.00000332, 0.00001886, -0.00002866, 0.00000184, 0.00001154, -0.00002207, 0.00000816, 0.00000830, -0.00000957, 0.00000650, -0.00000513, -0.00000245, -0.00000024, -0.00002237, -0.00000564, -0.00000241, -0.00000804, -0.00000035, -0.00002827, -0.00000236, -0.00000924, -0.00014592, -0.00009466, 0.00014716, 0.00023782, 0.00019945, 0.00007141, -0.00011071, -0.00009858, 0.00000708, -0.00001632, -0.00003361, 0.00000104, -0.00000182, -0.00000435, -0.00000118, -0.00000526, 0.00000315, -0.00000061, -0.00001153, -0.00000728, -0.00000269, -0.00000040, 0.00000109, 0.00000203, 0.00000747, 0.00000611, },
		{ 0.00000819, 0.00000110, -0.00000532, -0.00000611, -0.00000251, 0.00000705, 0.00000875, -0.00000191, -0.00000214, 0.00000305, -0.00000363, -0.00000120, 0.00000534, -0.00000390, -0.00000141, 0.00000687, -0.00000263, -0.00000202, 0.00000404, -0.00000035, 0.00000372, -0.00000047, -0.00001095, 0.00000034, 0.00000543, -0.00000128, 0.00000097, -0.00000593, -0.00000878, 0.00000860, 0.00001179, -0.00000147, -0.00000867, -0.00001176, -0.00000689, 0.00000469, 0.00000603, -0.00000034, -0.00000282, -0.00000863, -0.00001175, 0.00001432, 0.00003367, -0.00000592, -0.00002956, 0.00001339, 0.00001359, -0.00002774, 0.00000238, 0.00002531, -0.00001926, -0.00000964, 0.00002368, -0.00000899, -0.00002122, 0.00000151, 0.00000397, 0.00000348, -0.00000989, -0.00000432, 0.00002205, -0.00001008, -0.00001543, 0.00002528, -0.00000958, -0.00001732, 0.00001980, -0.00001382, 0.00000032, 0.00003010, -0.00002690, -0.00000938, 0.00002352, -0.00002270, 0.00000137, 0.00001574, -0.00002080, 0.00001303, 0.00000600, -0.00002416, 0.00001467, -0.00000195, -0.00001421, 0.00002173, -0.00001195, -0.00001687, 0.00001632, -0.00001187, -0.00000747, 0.00001333, -0.00000495, 0.00000060, -0.00000000, -0.00000460, 0.00000256, -0.00001696, -0.00001234, -0.00000157, -0.00001009, 0.00000375, -0.00002055, -0.00002112, 0.00001565, -0.00009519, -0.00014483, 0.00006366, 0.00020722, 0.00019935, 0.00014510, -0.00003001, -0.00013002, -0.00003140, 0.00000584, -0.00003421, -0.00001374, 0.00000235, -0.00000445, -0.00000174, -0.00000480, 0.00000001, 0.00000332, -0.00000798, -0.00001014, -0.00000468, -0.00000115, 0.00000110, 0.00000095, 0.00000522, },
		{ 0.00000277, 0.00000805, 0.00000498, -0.00000331, -0.00000640, -0.00000467, 0.00000303, 0.00001003, 0.00000233, -0.00000439, 0.00000247, -0.00000031, -0.00000446, 0.00000502, 0.00000024, -0.00000566, 0.00000604, 0.00000195, -0.00000517, 0.00000342, 0.00000148, 0.00000130, 0.00000376, -0.00000915, -0.00000611, 0.00000675, 0.00000085, -0.00000105, -0.00000182, -0.00000949, 0.00000099, 0.00001329, 0.00000484, -0.00000664, -0.00001157, -0.00000953, 0.00000017, 0.00000633, 0.00000308, -0.00000149, -0.00000526, -0.00001253, -0.00000198, 0.00003014, 0.00001913, -0.00002801, -0.00001083, 0.00002532, -0.00001213, -0.00002165, 0.00002583, 0.00000387, -0.00002585, 0.00001565, 0.00001211, -0.00002548, -0.00000798, 0.00000767, 0.00000248, -0.00000370, -0.00000911, 0.00001377, 0.00000615, -0.00002250, 0.00001248, 0.00001124, -0.00002442, 0.00000921, 0.00000387, -0.00001563, 0.00002748, -0.00000214, -0.00002904, 0.00002042, -0.00000346, -0.00001763, 0.00001983, -0.00000910, -0.00000559, 0.00001800, -0.00001867, -0.00000259, 0.00001205, -0.00001672, 0.00000981, 0.00000622, -0.00002293, 0.00000697, 0.00000219, -0.00001473, 0.00000779, 0.00000268, -0.00000241, 0.00000114, -0.00000449, 0.00000330, -0.00001055, -0.00001676, -0.00000173, -0.00001158, 0.00000150, -0.00000585, -0.00003453, 0.00001346, -0.00003684, -0.00015492, -0.00002207, 0.00016922, 0.00018421, 0.00017104, 0.00006561, -0.00010730, -0.00008455, 0.00000558, -0.00001779, -0.00002751, -0.00000153, -0.00000136, -0.00000269, -0.00000414, -0.00000279, 0.00000421, -0.00000289, -0.00001057, -0.00000741, -0.00000270, 0.00000091, 0.00000091, },
		{ 0.00000143, 0.00000128, 0.00000606, 0.00000761, 0.00000013, -0.00000603, -0.00000589, -0.00000065, 0.00000834, 0.00000689, -0.00000389, -0.00000062, 0.00000299, -0.00000485, 0.00000149, 0.00000475, -0.00000647, 0.00000140, 0.00000664, -0.00000508, -0.00000015, 0.00000401, -0.00000019, 0.00000471, -0.00000394, -0.00001099, 0.00000373, 0.00000495, -0.00000229, -0.00000082, -0.00000668, -0.00000489, 0.00000984, 0.00001030, -0.00000202, -0.00001100, -0.00001123, -0.00000351, 0.00000395, 0.00000528, 0.00000125, -0.00000332, -0.00000940, -0.00001102, 0.00001478, 0.00003330, -0.00000749, -0.00003006, 0.00001503, 0.00001263, -0.00002948, 0.00000505, 0.00002613, -0.00002163, -0.00000691, 0.00002613, -0.00001348, -0.00002275, 0.00000749, 0.00000548, -0.00000268, -0.00000678, 0.00000397, 0.00001336, -0.00001570, -0.00000544, 0.00002085, -0.00001463, -0.00000836, 0.00001446, -0.00001484, 0.00001019, 0.00001946, -0.00002910, 0.00000089, 0.00001519, -0.00002151, 0.00000749, 0.00000779, -0.00001513, 0.00001392, -0.00000300, -0.00001597, 0.00001257, -0.00000696, -0.00000427, 0.00001410, -0.00001564, -0.00000661, 0.00000935, -0.00001148, -0.00000161, 0.00000628, -0.00000110, 0.00000086, -0.00000446, 0.00000268, -0.00000346, -0.00001901, -0.00000360, -0.00000958, -0.00000667, 0.00000700, -0.00003271, -0.00000782, 0.00000650, -0.00012465, -0.00009577, 0.00011571, 0.00017249, 0.00016014, 0.00013396, -0.00003444, -0.00011677, -0.00002535, 0.00000066, -0.00002886, -0.00001261, 0.00000014, -0.00000184, -0.00000372, -0.00000477, 0.00000271, 0.00000162, -0.00000831, -0.00000955, -0.00000504, -0.00000011, },
		{ -0.00000212, 0.00000153, 0.00000117, 0.00000358, 0.00000797, 0.00000415, -0.00000431, -0.00000669, -0.00000317, 0.00000485, 0.00000938, -0.00000030, -0.00000403, 0.00000370, -0.00000198, -0.00000286, 0.00000647, -0.00000311, -0.00000436, 0.00000803, -0.00000106, -0.00000445, 0.00000456, 0.00000072, 0.00000304, 0.00000149, -0.00001157, -0.00000257, 0.00000770, -0.00000055, -0.00000217, -0.00000371, -0.00000691, 0.00000396, 0.00001205, 0.00000422, -0.00000840, -0.00001282, -0.00000621, 0.00000103, 0.00000467, 0.00000445, -0.00000101, -0.00000658, -0.00001212, -0.00000125, 0.00002996, 0.00001816, -0.00002931, -0.00001006, 0.00002602, -0.00001423, -0.00002089, 0.00002866, 0.00000183, -0.00002588, 0.00002013, 0.00000982, -0.00003001, -0.00000339, 0.00001227, -0.00000260, -0.00000501, -0.00000002, 0.00001130, -0.00000425, -0.00001594, 0.00001530, 0.00000246, -0.00001852, 0.00001069, -0.00000277, -0.00000636, 0.00002504, -0.00001216, -0.00001921, 0.00001910, -0.00000954, -0.00000928, 0.00001557, -0.00001028, 0.00000103, 0.00000933, -0.00001688, 0.00000240, 0.00000407, -0.00001026, 0.00001047, -0.00000345, -0.00001453, 0.00000697, -0.00000349, -0.00000759, 0.00000433, 0.00000128, 0.00000171, -0.00000419, -0.00000016, 0.00000332, -0.00001747, -0.00000863, -0.00000438, -0.00001478, 0.00000948, -0.00001660, -0.00003038, 0.00002001, -0.00006954, -0.00013746, 0.00004313, 0.00016273, 0.00014054, 0.00015429, 0.00005524, -0.00010124, -0.00007125, 0.00000157, -0.00001681, -0.00002261, -0.00000409, -0.00000010, -0.00000256, -0.00000587, -0.00000025, 0.00000420, -0.00000422, -0.00001004, -0.00000753, },
		{ -0.00000940, -0.00000464, 0.00000050, 0.00000172, 0.00000208, 0.00000626, 0.00000709, -0.00000088, -0.00000691, -0.00000487, 0.00000135, 0.00000881, 0.00000453, -0.00000504, 0.00000114, 0.00000200, -0.00000490, 0.00000427, 0.00000220, -0.00000757, 0.00000480, 0.00000461, -0.00000625, 0.00000192, 0.00000305, 0.00000110, 0.00000450, -0.00000793, -0.00000887, 0.00000655, 0.00000350, -0.00000318, -0.00000278, -0.00000607, -0.00000097, 0.00000973, 0.00000921, -0.00000296, -0.00001331, -0.00000915, -0.00000093, 0.00000221, 0.00000560, 0.00000281, -0.00000458, -0.00000981, -0.00000993, 0.00001475, 0.00003266, -0.00000861, -0.00003057, 0.00001622, 0.00001179, -0.00003074, 0.00000805, 0.00002638, -0.00002413, -0.00000298, 0.00002763, -0.00001949, -0.00002146, 0.00001445, 0.00000295, -0.00000725, 0.00000033, 0.00000729, 0.00000271, -0.00001559, 0.00000220, 0.00001378, -0.00001502, -0.00000187, 0.00000779, -0.00001130, 0.00001561, 0.00000790, -0.00002593, 0.00000720, 0.00000673, -0.00001701, 0.00000974, 0.00000193, -0.00000850, 0.00001074, -0.00000799, -0.00000848, 0.00000765, -0.00000715, 0.00000228, 0.00000486, -0.00001395, -0.00000027, 0.00000240, -0.00000793, 0.00000019, 0.00000163, 0.00000310, -0.00000190, -0.00000449, 0.00000685, -0.00001078, -0.00001537, -0.00000035, -0.00001636, 0.00000066, 0.00000257, -0.00003808, 0.00000581, -0.00001590, -0.00013532, -0.00003640, 0.00013926, 0.00013351, 0.00013693, 0.00012135, -0.00003877, -0.00010161, -0.00002291, -0.00000306, -0.00002315, -0.00001267, -0.00000081, -0.00000038, -0.00000551, -0.00000374, 0.00000418, 0.00000038, -0.00000847, },
		{ -0.00000494, -0.00001009, -0.00000709, -0.00000150, 0.00000171, 0.00000199, 0.00000402, 0.00000763, 0.00000344, -0.00000563, -0.00000639, -0.00000119, 0.00000611, 0.00000794, -0.00000260, -0.00000281, 0.00000402, -0.00000332, -0.00000007, 0.00000594, -0.00000629, -0.00000123, 0.00000823, -0.00000388, -0.00000247, 0.00000440, 0.00000076, 0.00000465, -0.00000243, -0.00001207, 0.00000160, 0.00000714, -0.00000183, -0.00000352, -0.00000483, -0.00000335, 0.00000560, 0.00001063, 0.00000375, -0.00001075, -0.00001266, -0.00000264, 0.00000037, 0.00000356, 0.00000648, -0.00000135, -0.00000799, -0.00001110, -0.00000073, 0.00002922, 0.00001754, -0.00003037, -0.00000976, 0.00002671, -0.00001602, -0.00001967, 0.00003138, -0.00000093, -0.00002490, 0.00002488, 0.00000526, -0.00003313, 0.00000405, 0.00001360, -0.00000892, -0.00000141, 0.00000705, 0.00000391, -0.00001012, -0.00000803, 0.00001352, -0.00000316, -0.00001174, 0.00000844, -0.00000545, 0.00000212, 0.00001822, -0.00001714, -0.00000980, 0.00001434, -0.00001160, -0.00000275, 0.00001019, -0.00000815, 0.00000389, 0.00000164, -0.00001235, 0.00000327, -0.00000093, -0.00000322, 0.00000652, -0.00000850, -0.00000639, 0.00000362, -0.00000517, -0.00000261, 0.00000014, 0.00000273, 0.00000233, -0.00000716, 0.00000482, -0.00000095, -0.00001918, -0.00000199, -0.00001043, -0.00001198, 0.00001216, -0.00002702, -0.00001910, 0.00001365, -0.00009639, -0.00009897, 0.00008992, 0.00013866, 0.00011072, 0.00014188, 0.00004253, -0.00009186, -0.00006071, -0.00000311, -0.00001407, -0.00001917, -0.00000591, 0.00000095, -0.00000318, -0.00000633, 0.00000153, 0.00000393, },
		{ 0.00000487, -0.00000019, -0.00000901, -0.00000912, -0.00000379, 0.00000053, 0.00000238, 0.00000288, 0.00000601, 0.00000671, -0.00000221, -0.00000750, -0.00000316, 0.00000312, 0.00000847, 0.00000190, -0.00000516, 0.00000259, 0.00000039, -0.00000335, 0.00000598, -0.00000168, -0.00000638, 0.00000743, 0.00000137, -0.00000562, 0.00000315, 0.00000193, 0.00000335, 0.00000232, -0.00001116, -0.00000484, 0.00000790, 0.00000163, -0.00000397, -0.00000468, -0.00000404, 0.00000227, 0.00000869, 0.00000858, -0.00000457, -0.00001492, -0.00000589, 0.00000027, 0.00000016, 0.00000699, 0.00000392, -0.00000645, -0.00000966, -0.00000862, 0.00001404, 0.00003206, -0.00000924, -0.00003152, 0.00001726, 0.00001120, -0.00003172, 0.00001144, 0.00002586, -0.00002645, 0.00000227, 0.00002722, -0.00002606, -0.00001646, 0.00002016, -0.00000303, -0.00000789, 0.00000878, 0.00000475, -0.00000648, -0.00001120, 0.00000578, 0.00000681, -0.00001200, 0.00000099, 0.00000233, -0.00000516, 0.00001592, -0.00000167, -0.00001949, 0.00000909, 0.00000025, -0.00001120, 0.00000881, -0.00000088, -0.00000305, 0.00000533, -0.00000899, -0.00000350, 0.00000234, -0.00000390, 0.00000431, -0.00000309, -0.00000894, 0.00000182, -0.00000251, -0.00000347, -0.00000093, -0.00000002, 0.00000565, -0.00000528, -0.00000140, 0.00000677, -0.00001606, -0.00000906, -0.00000245, -0.00001891, 0.00000750, -0.00000666, -0.00003505, 0.00001292, -0.00004407, -0.00012418, 0.00001984, 0.00013817, 0.00009993, 0.00012378, 0.00010601, -0.00004039, -0.00008684, -0.00002315, -0.00000480, -0.00001807, -0.00001298, -0.00000117, 0.00000001, -0.00000640, -0.00000268, },
		{ -0.00000612, 0.00000242, 0.00000406, -0.00000567, -0.00001041, -0.00000599, -0.00000131, 0.00000190, 0.00000314, 0.00000397, 0.00000736, 0.00000248, -0.00000697, -0.00000532, 0.00000086, 0.00000675, 0.00000581, -0.00000419, -0.00000108, 0.00000323, -0.00000350, 0.00000301, 0.00000314, -0.00000789, 0.00000277, 0.00000631, -0.00000536, -0.00000018, 0.00000309, 0.00000219, 0.00000499, -0.00000717, -0.00001003, 0.00000517, 0.00000533, -0.00000273, -0.00000516, -0.00000477, 0.00000055, 0.00000585, 0.00000947, 0.00000299, -0.00001321, -0.00001094, 0.00000032, -0.00000144, 0.00000342, 0.00000870, -0.00000256, -0.00000922, -0.00000946, -0.00000072, 0.00002808, 0.00001756, -0.00003166, -0.00000988, 0.00002773, -0.00001775, -0.00001793, 0.00003386, -0.00000442, -0.00002238, 0.00002905, -0.00000151, -0.00003320, 0.00001284, 0.00001061, -0.00001369, 0.00000610, 0.00000944, -0.00000554, -0.00001078, -0.00000166, 0.00000898, -0.00000491, -0.00000641, 0.00000444, -0.00000420, 0.00000749, 0.00000959, -0.00001720, -0.00000297, 0.00000833, -0.00001033, 0.00000094, 0.00000558, -0.00000440, 0.00000319, -0.00000344, -0.00000732, 0.00000126, -0.00000227, 0.00000194, 0.00000026, -0.00000892, -0.00000061, -0.00000081, -0.00000388, -0.00000049, -0.00000292, 0.00000504, 0.00000045, -0.00000694, 0.00000788, -0.00000673, -0.00001621, 0.00000067, -0.00001616, -0.00000559, 0.00000867, -0.00003252, -0.00000654, -0.00000419, -0.00010892, -0.00004958, 0.00011387, 0.00010838, 0.00009347, 0.00012917, 0.00003044, -0.00008034, -0.00005348, -0.00000704, -0.00001063, -0.00001670, -0.00000698, 0.00000124, -0.00000360, },
		{ 0.00000000, -0.00000640, -0.00000225, 0.00000551, -0.00000042, -0.00001005, -0.00000824, -0.00000301, 0.00000022, 0.00000351, 0.00000333, 0.00000565, 0.00000623, -0.00000386, -0.00000738, -0.00000110, 0.00000449, 0.00000735, -0.00000066, -0.00000420, 0.00000316, -0.00000106, -0.00000039, 0.00000555, -0.00000554, -0.00000297, 0.00000811, -0.00000185, -0.00000348, 0.00000289, 0.00000168, 0.00000568, -0.00000190, -0.00001231, 0.00000009, 0.00000739, -0.00000001, -0.00000488, -0.00000610, -0.00000084, 0.00000432, 0.00000728, 0.00000815, -0.00000679, -0.00001523, -0.00000221, -0.00000003, -0.00000155, 0.00000918, 0.00000430, -0.00000869, -0.00000876, -0.00000735, 0.00001254, 0.00003195, -0.00000965, -0.00003317, 0.00001861, 0.00001069, -0.00003249, 0.00001530, 0.00002426, -0.00002798, 0.00000852, 0.00002408, -0.00003155, -0.00000797, 0.00002246, -0.00001044, -0.00000361, 0.00001527, -0.00000253, -0.00001167, -0.00000521, 0.00000532, 0.00000210, -0.00000769, 0.00000062, -0.00000026, 0.00000105, 0.00001208, -0.00000745, -0.00001234, 0.00000749, -0.00000321, -0.00000599, 0.00000617, -0.00000086, -0.00000012, -0.00000022, -0.00000718, -0.00000160, -0.00000136, 0.00000063, 0.00000230, -0.00000786, -0.00000306, 0.00000068, -0.00000458, 0.00000005, -0.00000323, 0.00000063, 0.00000581, -0.00000727, 0.00000255, 0.00000322, -0.00001744, -0.00000341, -0.00000768, -0.00001663, 0.00001047, -0.00001627, -0.00002567, 0.00000971, -0.00006893, -0.00009451, 0.00006274, 0.00012069, 0.00007627, 0.00011482, 0.00008921, -0.00003863, -0.00007416, -0.00002460, -0.00000500, -0.00001386, -0.00001299, -0.00000172, },
		{ -0.00000785, 0.00000090, -0.00000328, -0.00000608, 0.00000286, 0.00000472, -0.00000689, -0.00001042, -0.00000453, -0.00000169, 0.00000248, 0.00000421, 0.00000364, 0.00000715, 0.00000106, -0.00000785, -0.00000361, 0.00000270, 0.00000676, 0.00000315, -0.00000472, 0.00000056, 0.00000162, -0.00000198, 0.00000507, -0.00000131, -0.00000691, 0.00000612, 0.00000283, -0.00000484, 0.00000126, 0.00000148, 0.00000490, 0.00000312, -0.00001131, -0.00000565, 0.00000708, 0.00000298, -0.00000326, -0.00000698, -0.00000332, 0.00000404, 0.00000503, 0.00000872, 0.00000163, -0.00001508, -0.00000798, 0.00000206, -0.00000376, 0.00000432, 0.00001070, -0.00000459, -0.00000998, -0.00000728, -0.00000157, 0.00002693, 0.00001830, -0.00003370, -0.00001002, 0.00002918, -0.00001973, -0.00001541, 0.00003576, -0.00000835, -0.00001795, 0.00003145, -0.00000962, -0.00002908, 0.00002067, 0.00000375, -0.00001448, 0.00001494, 0.00000613, -0.00001375, -0.00000740, 0.00000138, 0.00000412, -0.00000364, -0.00000380, 0.00000093, -0.00000043, 0.00000877, 0.00000182, -0.00001383, 0.00000038, 0.00000310, -0.00000723, 0.00000192, 0.00000292, -0.00000085, 0.00000016, -0.00000544, -0.00000356, -0.00000175, -0.00000062, 0.00000385, -0.00000580, -0.00000599, 0.00000200, -0.00000442, -0.00000125, -0.00000075, -0.00000404, 0.00000653, -0.00000217, -0.00000426, 0.00000777, -0.00001124, -0.00001076, -0.00000100, -0.00001881, 0.00000087, 0.00000088, -0.00003116, 0.00000142, -0.00002794, -0.00010353, -0.00000060, 0.00011686, 0.00008055, 0.00008464, 0.00011467, 0.00002105, -0.00006851, -0.00004892, -0.00000972, -0.00000729, -0.00001452, },
		{ -0.00001019, -0.00001277, -0.00000286, 0.00000041, -0.00000611, -0.00000252, 0.00000666, -0.00000090, -0.00001123, -0.00000666, -0.00000273, -0.00000003, 0.00000489, 0.00000338, 0.00000541, 0.00000533, -0.00000545, -0.00000643, 0.00000095, 0.00000544, 0.00000542, -0.00000269, -0.00000247, 0.00000253, -0.00000145, 0.00000310, 0.00000254, -0.00000780, 0.00000167, 0.00000635, -0.00000373, -0.00000084, 0.00000122, 0.00000307, 0.00000677, -0.00000734, -0.00001058, 0.00000451, 0.00000533, -0.00000102, -0.00000614, -0.00000658, 0.00000297, 0.00000500, 0.00000598, 0.00000753, -0.00000919, -0.00001402, 0.00000105, -0.00000154, -0.00000249, 0.00001180, 0.00000380, -0.00001103, -0.00000697, -0.00000653, 0.00001036, 0.00003277, -0.00001041, -0.00003545, 0.00002073, 0.00000987, -0.00003287, 0.00001952, 0.00002138, -0.00002797, 0.00001490, 0.00001799, -0.00003406, 0.00000260, 0.00002006, -0.00001647, 0.00000472, 0.00001708, -0.00001182, -0.00001201, -0.00000033, 0.00000231, 0.00000048, -0.00000422, -0.00000144, 0.00000031, 0.00000515, 0.00000615, -0.00000912, -0.00000659, 0.00000411, -0.00000383, -0.00000256, 0.00000347, 0.00000087, 0.00000001, -0.00000427, -0.00000427, -0.00000217, -0.00000248, 0.00000426, -0.00000213, -0.00000891, 0.00000166, -0.00000203, -0.00000421, 0.00000173, -0.00000515, 0.00000220, 0.00000407, -0.00000697, 0.00000523, -0.00000177, -0.00001500, -0.00000074, -0.00001304, -0.00001118, 0.00000839, -0.00002225, -0.00001447, -0.00000339, -0.00008290, -0.00005413, 0.00008742, 0.00009614, 0.00006245, 0.00010609, 0.00007315, -0.00003435, -0.00006414, -0.00002610, -0.00000444, },
		{ -0.00001140, -0.00000422, -0.00001227, -0.00000914, 0.00000057, -0.00000230, -0.00000677, 0.00000355, 0.00000553, -0.00000866, -0.00000970, -0.00000320, -0.00000240, 0.00000349, 0.00000494, 0.00000332, 0.00000672, -0.00000072, -0.00000791, -0.00000165, 0.00000426, 0.00000604, 0.00000027, -0.00000401, 0.00000140, -0.00000006, 0.00000120, 0.00000490, -0.00000600, -0.00000329, 0.00000744, -0.00000089, -0.00000252, 0.00000103, 0.00000074, 0.00000797, -0.00000118, -0.00001332, -0.00000003, 0.00000682, 0.00000089, -0.00000376, -0.00000861, -0.00000058, 0.00000653, 0.00000367, 0.00000819, -0.00000033, -0.00001578, -0.00000445, 0.00000238, -0.00000596, 0.00000609, 0.00001218, -0.00000729, -0.00000998, -0.00000482, -0.00000351, 0.00002636, 0.00001946, -0.00003686, -0.00000953, 0.00003081, -0.00002207, -0.00001186, 0.00003656, -0.00001208, -0.00001173, 0.00003096, -0.00001735, -0.00002069, 0.00002513, -0.00000503, -0.00001014, 0.00002181, -0.00000187, -0.00001815, -0.00000243, 0.00000092, 0.00000096, -0.00000113, -0.00000379, -0.00000049, 0.00000374, 0.00000648, -0.00000326, -0.00000915, 0.00000066, -0.00000014, -0.00000394, 0.00000119, 0.00000239, 0.00000113, -0.00000357, -0.00000496, -0.00000188, -0.00000402, 0.00000244, 0.00000246, -0.00000979, -0.00000165, 0.00000183, -0.00000621, 0.00000119, -0.00000220, -0.00000345, 0.00000641, -0.00000392, -0.00000077, 0.00000486, -0.00001275, -0.00000554, -0.00000546, -0.00001754, 0.00000443, -0.00000770, -0.00002418, 0.00000114, -0.00005010, -0.00008205, 0.00003883, 0.00010500, 0.00005973, 0.00007964, 0.00009909, 0.00001496, -0.00005782, -0.00004589, },
		{ -0.00005635, -0.00002719, -0.00000369, -0.00000656, -0.00001281, -0.00000434, 0.00000165, -0.00000631, -0.00000300, 0.00000838, -0.00000208, -0.00001200, -0.00000469, -0.00000307, 0.00000011, 0.00000616, 0.00000325, 0.00000521, 0.00000383, -0.00000647, -0.00000488, 0.00000279, 0.00000592, 0.00000273, -0.00000370, -0.00000058, 0.00000085, -0.00000003, 0.00000572, -0.00000249, -0.00000723, 0.00000589, 0.00000265, -0.00000344, 0.00000099, -0.00000098, 0.00000609, 0.00000548, -0.00001244, -0.00000609, 0.00000716, 0.00000251, -0.00000143, -0.00000776, -0.00000575, 0.00000665, 0.00000453, 0.00000500, 0.00000637, -0.00001111, -0.00001153, 0.00000327, -0.00000371, -0.00000249, 0.00001447, 0.00000238, -0.00001318, -0.00000427, -0.00000663, 0.00000795, 0.00003470, -0.00001220, -0.00003786, 0.00002376, 0.00000823, -0.00003241, 0.00002373, 0.00001735, -0.00002574, 0.00002007, 0.00000963, -0.00003209, 0.00001283, 0.00001316, -0.00001851, 0.00001455, 0.00001309, -0.00001987, -0.00000854, 0.00000162, -0.00000106, 0.00000138, -0.00000284, -0.00000327, 0.00000300, 0.00000606, 0.00000043, -0.00000766, -0.00000335, 0.00000074, -0.00000262, -0.00000114, 0.00000187, 0.00000286, -0.00000192, -0.00000612, -0.00000180, -0.00000386, -0.00000125, 0.00000560, -0.00000695, -0.00000694, 0.00000417, -0.00000469, -0.00000248, 0.00000164, -0.00000586, 0.00000337, 0.00000166, -0.00000474, 0.00000550, -0.00000591, -0.00001035, -0.00000161, -0.00001612, -0.00000532, 0.00000252, -0.00002265, -0.00000644, -0.00002219, -0.00008224, -0.00001241, 0.00009443, 0.00007205, 0.00005563, 0.00009611, 0.00005943, -0.00002907, },
		{ 0.00001145, -0.00004881, -0.00004345, -0.00001265, -0.00000113, -0.00001011, -0.00001109, 0.00000100, -0.00000132, -0.00000817, 0.00000498, 0.00000590, -0.00001060, -0.00000815, -0.00000244, -0.00000307, 0.00000474, 0.00000532, 0.00000318, 0.00000571, -0.00000241, -0.00000707, 0.00000021, 0.00000556, 0.00000443, -0.00000230, -0.00000226, 0.00000098, -0.00000081, 0.00000520, 0.00000187, -0.00000915, 0.00000200, 0.00000601, -0.00000336, 0.00000056, -0.00000090, 0.00000183, 0.00000996, -0.00000714, -0.00001219, 0.00000529, 0.00000465, -0.00000040, -0.00000459, -0.00000955, 0.00000300, 0.00000773, 0.00000228, 0.00000751, -0.00000250, -0.00001501, -0.00000119, 0.00000150, -0.00000761, 0.00000845, 0.00001294, -0.00001048, -0.00000892, -0.00000255, -0.00000652, 0.00002697, 0.00002037, -0.00004113, -0.00000778, 0.00003200, -0.00002454, -0.00000721, 0.00003580, -0.00001467, -0.00000446, 0.00002699, -0.00002260, -0.00000927, 0.00002460, -0.00001286, -0.00000139, 0.00002389, -0.00001191, -0.00001781, 0.00000151, -0.00000177, 0.00000038, 0.00000074, -0.00000516, 0.00000058, 0.00000637, 0.00000223, -0.00000508, -0.00000509, -0.00000086, -0.00000120, -0.00000163, 0.00000002, 0.00000331, 0.00000106, -0.00000655, -0.00000323, -0.00000210, -0.00000456, 0.00000513, -0.00000116, -0.00001091, 0.00000233, 0.00000001, -0.00000615, 0.00000251, -0.00000364, -0.00000207, 0.00000487, -0.00000398, 0.00000185, 0.00000070, -0.00001112, -0.00000259, -0.00001031, -0.00001348, 0.00000388, -0.00001363, -0.00001514, -0.00000753, -0.00006404, -0.00005048, 0.00006384, 0.00008581, 0.00004653, 0.00007524, 0.00008404, },
		{ 0.00008531, 0.00004852, -0.00002398, -0.00004991, -0.00002804, -0.00000283, -0.00000276, -0.00001365, -0.00000549, 0.00000351, -0.00000748, -0.00000299, 0.00001024, -0.00000398, -0.00001190, -0.00000285, -0.00000380, 0.00000068, 0.00000710, 0.00000323, 0.00000469, 0.00000188, -0.00000653, -0.00000315, 0.00000441, 0.00000577, -0.00000046, -0.00000334, 0.00000068, -0.00000117, 0.00000327, 0.00000608, -0.00000810, -0.00000360, 0.00000807, -0.00000165, -0.00000104, 0.00000102, -0.00000239, 0.00000983, 0.00000140, -0.00001549, 0.00000007, 0.00000737, -0.00000009, -0.00000159, -0.00000947, -0.00000356, 0.00000953, 0.00000339, 0.00000424, 0.00000463, -0.00001193, -0.00000842, 0.00000418, -0.00000594, -0.00000168, 0.00001692, 0.00000005, -0.00001474, -0.00000083, -0.00000808, 0.00000600, 0.00003750, -0.00001561, -0.00003956, 0.00002741, 0.00000552, -0.00003061, 0.00002724, 0.00001272, -0.00002109, 0.00002262, 0.00000064, -0.00002518, 0.00002008, 0.00000358, -0.00001516, 0.00002253, 0.00000421, -0.00002405, -0.00000355, 0.00000040, -0.00000289, 0.00000326, -0.00000356, -0.00000342, 0.00000607, 0.00000404, -0.00000336, -0.00000478, -0.00000254, -0.00000144, -0.00000088, -0.00000115, 0.00000175, 0.00000388, -0.00000471, -0.00000596, -0.00000066, -0.00000529, 0.00000125, 0.00000433, -0.00001045, -0.00000337, 0.00000444, -0.00000627, -0.00000059, 0.00000053, -0.00000534, 0.00000342, -0.00000016, -0.00000181, 0.00000351, -0.00000770, -0.00000568, -0.00000493, -0.00001602, -0.00000163, -0.00000436, -0.00001816, -0.00000493, -0.00004053, -0.00006783, 0.00002264, 0.00008829, 0.00005268, 0.00005232, },
		{ 0.00003905, 0.00007015, 0.00007076, 0.00000928, -0.00004115, -0.00004128, -0.00001383, 0.00000236, -0.00000858, -0.00001335, 0.00000251, -0.00000111, -0.00000971, 0.00000716, 0.00000538, -0.00001212, -0.00000611, -0.00000236, -0.00000336, 0.00000588, 0.00000560, 0.00000308, 0.00000393, -0.00000344, -0.00000561, 0.00000182, 0.00000659, 0.00000187, -0.00000393, -0.00000005, -0.00000069, 0.00000027, 0.00000857, -0.00000360, -0.00000931, 0.00000720, 0.00000221, -0.00000352, 0.00000296, -0.00000365, 0.00000489, 0.00000951, -0.00001314, -0.00000789, 0.00000882, 0.00000148, -0.00000083, -0.00000592, -0.00000930, 0.00000670, 0.00000774, 0.00000112, 0.00000633, -0.00000425, -0.00001299, 0.00000115, -0.00000007, -0.00000854, 0.00001110, 0.00001284, -0.00001388, -0.00000662, -0.00000107, -0.00001021, 0.00002914, 0.00002015, -0.00004595, -0.00000447, 0.00003203, -0.00002654, -0.00000180, 0.00003333, -0.00001514, 0.00000249, 0.00001987, -0.00002347, 0.00000279, 0.00001894, -0.00001691, 0.00000929, 0.00001988, -0.00002068, -0.00001366, 0.00000262, -0.00000466, 0.00000194, 0.00000079, -0.00000623, 0.00000335, 0.00000643, -0.00000202, -0.00000428, -0.00000277, -0.00000273, -0.00000073, -0.00000070, -0.00000055, 0.00000457, -0.00000071, -0.00000797, -0.00000150, -0.00000332, -0.00000332, 0.00000618, -0.00000541, -0.00000942, 0.00000484, -0.00000218, -0.00000491, 0.00000257, -0.00000432, -0.00000089, 0.00000279, -0.00000252, 0.00000259, -0.00000290, -0.00000755, -0.00000254, -0.00001347, -0.00000885, 0.00000002, -0.00001505, -0.00000822, -0.00002142, -0.00006616, -0.00001659, 0.00007422, 0.00006569, },
		{ 0.00007504, 0.00003897, 0.00004996, 0.00007487, 0.00003977, -0.00001951, -0.00004413, -0.00002909, -0.00000150, 0.00000090, -0.00001543, -0.00000557, 0.00000528, -0.00000954, -0.00000197, 0.00001159, -0.00000620, -0.00001084, -0.00000158, -0.00000454, 0.00000147, 0.00000774, 0.00000347, 0.00000344, 0.00000001, -0.00000558, -0.00000162, 0.00000594, 0.00000474, -0.00000363, -0.00000167, 0.00000067, -0.00000244, 0.00000772, 0.00000335, -0.00001237, 0.00000221, 0.00000725, -0.00000488, 0.00000246, -0.00000094, -0.00000187, 0.00001276, -0.00000481, -0.00001522, 0.00000622, 0.00000534, -0.00000177, -0.00000202, -0.00001071, -0.00000036, 0.00001121, 0.00000210, 0.00000332, 0.00000266, -0.00001133, -0.00000550, 0.00000399, -0.00000781, -0.00000027, 0.00001895, -0.00000315, -0.00001528, 0.00000288, -0.00001104, 0.00000532, 0.00004044, -0.00002075, -0.00003968, 0.00003090, 0.00000189, -0.00002716, 0.00002933, 0.00000853, -0.00001451, 0.00002153, -0.00000674, -0.00001425, 0.00002235, -0.00000579, -0.00000693, 0.00002565, -0.00000693, -0.00002334, 0.00000034, -0.00000284, -0.00000233, 0.00000440, -0.00000533, -0.00000147, 0.00000781, 0.00000038, -0.00000457, -0.00000213, -0.00000330, -0.00000207, 0.00000036, -0.00000161, 0.00000272, 0.00000339, -0.00000713, -0.00000463, -0.00000090, -0.00000553, 0.00000362, 0.00000112, -0.00001172, 0.00000032, 0.00000315, -0.00000654, 0.00000067, -0.00000074, -0.00000420, 0.00000242, -0.00000071, 0.00000051, 0.00000044, -0.00000692, -0.00000275, -0.00000877, -0.00001351, -0.00000126, -0.00000933, -0.00001160, -0.00001040, -0.00005269, -0.00004432, 0.00004645, },
		{ 0.00001274, 0.00007331, 0.00004843, 0.00003468, 0.00006466, 0.00005950, 0.00000751, -0.00003416, -0.00003958, -0.00001484, 0.00000635, -0.00000821, -0.00001515, 0.00000481, -0.00000217, -0.00001063, 0.00000962, 0.00000391, -0.00001271, -0.00000411, -0.00000275, -0.00000325, 0.00000669, 0.00000611, 0.00000259, 0.00000159, -0.00000336, -0.00000403, 0.00000323, 0.00000717, -0.00000150, -0.00000415, 0.00000176, -0.00000300, 0.00000344, 0.00000974, -0.00001010, -0.00000600, 0.00001049, -0.00000264, -0.00000126, 0.00000350, -0.00000582, 0.00000911, 0.00000598, -0.00001728, -0.00000145, 0.00000971, -0.00000160, -0.00000080, -0.00000755, -0.00000768, 0.00000980, 0.00000700, 0.00000012, 0.00000458, -0.00000493, -0.00001032, 0.00000231, -0.00000177, -0.00000883, 0.00001385, 0.00001175, -0.00001709, -0.00000309, -0.00000102, -0.00001381, 0.00003276, 0.00001798, -0.00005036, 0.00000015, 0.00003036, -0.00002732, 0.00000365, 0.00002952, -0.00001292, 0.00000748, 0.00001103, -0.00001904, 0.00001264, 0.00000972, -0.00001548, 0.00001847, 0.00001060, -0.00002545, -0.00000796, 0.00000060, -0.00000594, 0.00000420, -0.00000100, -0.00000571, 0.00000632, 0.00000404, -0.00000475, -0.00000222, -0.00000232, -0.00000387, 0.00000027, -0.00000081, -0.00000009, 0.00000518, -0.00000328, -0.00000780, -0.00000059, -0.00000448, -0.00000103, 0.00000515, -0.00000883, -0.00000630, 0.00000559, -0.00000384, -0.00000333, 0.00000180, -0.00000416, -0.00000049, 0.00000111, -0.00000047, 0.00000155, -0.00000467, -0.00000384, -0.00000459, -0.00001408, -0.00000581, -0.00000487, -0.00001234, -0.00000643, -0.00003546, -0.00005677, },
		{ -0.00005534, -0.00001816, 0.00005842, 0.00006000, 0.00002966, 0.00004757, 0.00006568, 0.00003217, -0.00001536, -0.00003886, -0.00003053, 0.00000068, 0.00000380, -0.00001769, -0.00000426, 0.00000610, -0.00001184, 0.00000005, 0.00001190, -0.00000808, -0.00000915, -0.00000103, -0.00000497, 0.00000214, 0.00000836, 0.00000379, 0.00000124, -0.00000110, -0.00000397, -0.00000045, 0.00000741, 0.00000260, -0.00000606, 0.00000079, -0.00000078, -0.00000197, 0.00001187, -0.00000226, -0.00001359, 0.00000840, 0.00000361, -0.00000576, 0.00000553, -0.00000409, 0.00000094, 0.00001331, -0.00001153, -0.00001142, 0.00001067, 0.00000198, -0.00000237, -0.00000302, -0.00001093, 0.00000318, 0.00001172, 0.00000096, 0.00000190, 0.00000110, -0.00000948, -0.00000343, 0.00000316, -0.00000920, 0.00000156, 0.00002038, -0.00000702, -0.00001441, 0.00000617, -0.00001523, 0.00000652, 0.00004247, -0.00002714, -0.00003766, 0.00003326, -0.00000202, -0.00002217, 0.00002952, 0.00000595, -0.00000725, 0.00001671, -0.00001027, -0.00000154, 0.00001908, -0.00001191, 0.00000383, 0.00002243, -0.00001691, -0.00001859, 0.00000133, -0.00000613, 0.00000021, 0.00000372, -0.00000663, 0.00000191, 0.00000728, -0.00000325, -0.00000364, -0.00000075, -0.00000443, -0.00000153, 0.00000065, -0.00000169, 0.00000397, 0.00000151, -0.00000843, -0.00000308, -0.00000195, -0.00000443, 0.00000477, -0.00000284, -0.00001077, 0.00000310, 0.00000127, -0.00000583, 0.00000107, -0.00000155, -0.00000316, 0.00000097, -0.00000009, 0.00000140, -0.00000223, -0.00000452, -0.00000215, -0.00001136, -0.00001040, -0.00000366, -0.00001068, -0.00000652, -0.00002038, },
		{ -0.00001035, -0.00004471, -0.00003958, 0.00003341, 0.00006597, 0.00003502, 0.00003150, 0.00005968, 0.00004963, 0.00000586, -0.00002738, -0.00003874, -0.00001524, 0.00001041, -0.00000910, -0.00001569, 0.00000709, -0.00000437, -0.00001034, 0.00001161, 0.00000193, -0.00001225, -0.00000264, -0.00000321, -0.00000294, 0.00000723, 0.00000699, 0.00000121, -0.00000071, -0.00000212, -0.00000278, 0.00000467, 0.00000693, -0.00000517, -0.00000283, 0.00000239, -0.00000487, 0.00000811, 0.00000767, -0.00001558, 0.00000005, 0.00001061, -0.00000672, 0.00000250, 0.00000179, -0.00000605, 0.00001288, -0.00000016, -0.00001806, 0.00000535, 0.00000803, -0.00000370, -0.00000092, -0.00000896, -0.00000492, 0.00001181, 0.00000602, -0.00000104, 0.00000257, -0.00000417, -0.00000784, 0.00000247, -0.00000324, -0.00000866, 0.00001656, 0.00000957, -0.00001955, 0.00000129, -0.00000280, -0.00001634, 0.00003717, 0.00001347, -0.00005320, 0.00000529, 0.00002695, -0.00002625, 0.00000821, 0.00002532, -0.00000814, 0.00000912, 0.00000273, -0.00000979, 0.00001792, -0.00000024, -0.00000879, 0.00002294, -0.00000135, -0.00002505, -0.00000338, -0.00000343, -0.00000476, 0.00000555, -0.00000366, -0.00000322, 0.00000796, 0.00000031, -0.00000533, -0.00000029, -0.00000308, -0.00000389, 0.00000095, -0.00000126, 0.00000119, 0.00000457, -0.00000569, -0.00000658, -0.00000069, -0.00000481, 0.00000124, 0.00000242, -0.00001052, -0.00000275, 0.00000495, -0.00000460, -0.00000206, 0.00000082, -0.00000351, -0.00000086, 0.00000036, 0.00000115, -0.00000037, -0.00000437, -0.00000147, -0.00000716, -0.00001270, -0.00000545, -0.00000831, -0.00000780, },
		{ -0.00000864, -0.00000554, -0.00003065, -0.00004855, 0.00000478, 0.00006112, 0.00004638, 0.00002281, 0.00004545, 0.00005779, 0.00002496, -0.00001095, -0.00003445, -0.00003210, 0.00000373, 0.00000532, -0.00001955, -0.00000190, 0.00000542, -0.00001351, 0.00000256, 0.00001109, -0.00000907, -0.00000739, -0.00000099, -0.00000501, 0.00000247, 0.00000934, 0.00000367, -0.00000170, -0.00000095, -0.00000225, 0.00000047, 0.00000865, -0.00000063, -0.00000710, 0.00000317, -0.00000317, 0.00000074, 0.00001403, -0.00000941, -0.00001101, 0.00001293, -0.00000158, -0.00000394, 0.00000676, -0.00000696, 0.00000536, 0.00001057, -0.00001670, -0.00000500, 0.00001234, -0.00000145, -0.00000228, -0.00000451, -0.00000979, 0.00000630, 0.00001142, -0.00000005, -0.00000015, 0.00000064, -0.00000698, -0.00000250, 0.00000225, -0.00001021, 0.00000371, 0.00002097, -0.00001123, -0.00001194, 0.00000822, -0.00001993, 0.00000975, 0.00004252, -0.00003375, -0.00003355, 0.00003365, -0.00000526, -0.00001631, 0.00002789, 0.00000593, -0.00000105, 0.00000925, -0.00000856, 0.00001004, 0.00001145, -0.00001278, 0.00001364, 0.00001347, -0.00002284, -0.00001208, -0.00000100, -0.00000768, 0.00000334, 0.00000119, -0.00000627, 0.00000533, 0.00000459, -0.00000551, -0.00000169, -0.00000083, -0.00000503, -0.00000059, 0.00000019, -0.00000098, 0.00000468, -0.00000111, -0.00000844, -0.00000202, -0.00000305, -0.00000247, 0.00000423, -0.00000629, -0.00000825, 0.00000453, -0.00000047, -0.00000473, 0.00000083, -0.00000179, -0.00000263, -0.00000019, 0.00000101, 0.00000093, -0.00000347, -0.00000196, -0.00000328, -0.00001193, -0.00000855, -0.00000694, },
		{ -0.00000732, -0.00000880, -0.00000444, -0.00001797, -0.00004625, -0.00001997, 0.00004472, 0.00005656, 0.00002455, 0.00002891, 0.00005582, 0.00004023, 0.00000472, -0.00002085, -0.00003889, -0.00001455, 0.00001378, -0.00001074, -0.00001460, 0.00000837, -0.00000701, -0.00000882, 0.00001253, 0.00000015, -0.00001104, -0.00000188, -0.00000339, -0.00000280, 0.00000784, 0.00000803, -0.00000134, -0.00000222, -0.00000007, -0.00000205, 0.00000626, 0.00000552, -0.00000840, -0.00000037, 0.00000147, -0.00000515, 0.00001289, 0.00000233, -0.00001811, 0.00000705, 0.00000750, -0.00000853, 0.00000616, -0.00000151, -0.00000375, 0.00001446, -0.00000734, -0.00001500, 0.00001063, 0.00000469, -0.00000453, -0.00000164, -0.00000950, -0.00000163, 0.00001266, 0.00000516, -0.00000277, 0.00000096, -0.00000210, -0.00000624, 0.00000211, -0.00000439, -0.00000814, 0.00001906, 0.00000631, -0.00002070, 0.00000583, -0.00000635, -0.00001692, 0.00004124, 0.00000694, -0.00005362, 0.00000975, 0.00002237, -0.00002318, 0.00001110, 0.00002202, -0.00000181, 0.00000685, -0.00000261, 0.00000234, 0.00001760, -0.00000779, 0.00000102, 0.00002095, -0.00001247, -0.00002022, -0.00000185, -0.00000751, -0.00000153, 0.00000492, -0.00000576, 0.00000061, 0.00000744, -0.00000328, -0.00000410, 0.00000058, -0.00000413, -0.00000305, 0.00000092, -0.00000137, 0.00000266, 0.00000278, -0.00000725, -0.00000503, -0.00000144, -0.00000412, 0.00000260, -0.00000109, -0.00001032, 0.00000030, 0.00000357, -0.00000461, -0.00000135, 0.00000007, -0.00000288, -0.00000155, 0.00000047, 0.00000172, -0.00000196, -0.00000280, -0.00000096, -0.00000874, -0.00001087, },
		{ -0.00001095, -0.00000888, -0.00000896, -0.00000514, -0.00000924, -0.00003695, -0.00003536, 0.00002096, 0.00005848, 0.00003500, 0.00001706, 0.00004439, 0.00005045, 0.00001808, -0.00000593, -0.00003134, -0.00003322, 0.00000717, 0.00000534, -0.00002010, 0.00000060, 0.00000335, -0.00001388, 0.00000473, 0.00000958, -0.00000894, -0.00000609, -0.00000106, -0.00000491, 0.00000252, 0.00001086, 0.00000253, -0.00000475, 0.00000060, -0.00000099, 0.00000147, 0.00000927, -0.00000453, -0.00000645, 0.00000456, -0.00000549, 0.00000528, 0.00001291, -0.00001600, -0.00000513, 0.00001427, -0.00000646, -0.00000020, 0.00000554, -0.00000800, 0.00000979, 0.00000479, -0.00001858, 0.00000222, 0.00001115, -0.00000390, -0.00000220, -0.00000601, -0.00000741, 0.00000843, 0.00001084, -0.00000126, -0.00000252, 0.00000169, -0.00000469, -0.00000253, 0.00000159, -0.00001100, 0.00000615, 0.00002048, -0.00001519, -0.00000806, 0.00000838, -0.00002405, 0.00001450, 0.00003991, -0.00003922, -0.00002808, 0.00003171, -0.00000692, -0.00001058, 0.00002517, 0.00000873, 0.00000244, 0.00000122, -0.00000158, 0.00001783, 0.00000209, -0.00000826, 0.00001913, 0.00000133, -0.00002332, -0.00000656, -0.00000553, -0.00000657, 0.00000545, -0.00000224, -0.00000389, 0.00000741, 0.00000073, -0.00000583, 0.00000009, -0.00000186, -0.00000475, 0.00000004, -0.00000049, 0.00000040, 0.00000434, -0.00000369, -0.00000748, -0.00000168, -0.00000357, -0.00000045, 0.00000222, -0.00000844, -0.00000506, 0.00000475, -0.00000168, -0.00000370, 0.00000036, -0.00000170, -0.00000261, -0.00000070, 0.00000176, -0.00000015, -0.00000312, -0.00000050, -0.00000478, },
		{ -0.00000176, -0.00000869, -0.00001016, -0.00000978, -0.00000649, -0.00000457, -0.00002569, -0.00004005, -0.00000293, 0.00004875, 0.00004765, 0.00001555, 0.00002735, 0.00005285, 0.00003071, 0.00000472, -0.00001510, -0.00003966, -0.00001271, 0.00001576, -0.00001215, -0.00001221, 0.00000801, -0.00000911, -0.00000658, 0.00001220, -0.00000080, -0.00000966, -0.00000167, -0.00000321, -0.00000308, 0.00000893, 0.00000869, -0.00000488, -0.00000230, 0.00000213, -0.00000171, 0.00000790, 0.00000282, -0.00001041, 0.00000241, -0.00000068, -0.00000317, 0.00001600, -0.00000510, -0.00001660, 0.00001279, 0.00000226, -0.00000743, 0.00000814, -0.00000491, 0.00000061, 0.00001267, -0.00001338, -0.00000886, 0.00001310, 0.00000104, -0.00000444, -0.00000303, -0.00000873, 0.00000141, 0.00001267, 0.00000443, -0.00000522, 0.00000047, 0.00000067, -0.00000584, 0.00000180, -0.00000537, -0.00000725, 0.00002108, 0.00000219, -0.00002013, 0.00000960, -0.00001107, -0.00001512, 0.00004366, -0.00000063, -0.00005134, 0.00001240, 0.00001761, -0.00001854, 0.00001204, 0.00002083, 0.00000443, 0.00000132, -0.00000316, 0.00001454, 0.00001241, -0.00001067, 0.00001054, 0.00001288, -0.00001964, -0.00001323, -0.00000386, -0.00000978, 0.00000238, 0.00000229, -0.00000611, 0.00000447, 0.00000486, -0.00000554, -0.00000204, 0.00000023, -0.00000467, -0.00000194, 0.00000030, -0.00000078, 0.00000366, 0.00000027, -0.00000772, -0.00000370, -0.00000230, -0.00000272, 0.00000258, -0.00000437, -0.00000863, 0.00000240, 0.00000205, -0.00000421, -0.00000108, -0.00000036, -0.00000258, -0.00000204, 0.00000086, 0.00000131, -0.00000238, -0.00000122, },
		{ -0.00000188, -0.00000054, -0.00000537, -0.00000974, -0.00001099, -0.00000840, -0.00000281, -0.00001592, -0.00003674, -0.00002037, 0.00002980, 0.00005407, 0.00002508, 0.00001223, 0.00004476, 0.00004271, 0.00001210, -0.00000054, -0.00002969, -0.00003327, 0.00001014, 0.00000435, -0.00001886, 0.00000215, 0.00000060, -0.00001282, 0.00000584, 0.00000812, -0.00000798, -0.00000552, -0.00000086, -0.00000503, 0.00000265, 0.00001273, 0.00000001, -0.00000704, 0.00000318, -0.00000046, 0.00000277, 0.00000878, -0.00000836, -0.00000433, 0.00000442, -0.00000651, 0.00001038, 0.00000823, -0.00001989, 0.00000235, 0.00001203, -0.00000925, 0.00000387, 0.00000235, -0.00000636, 0.00001234, -0.00000257, -0.00001639, 0.00000820, 0.00000796, -0.00000493, -0.00000272, -0.00000682, -0.00000431, 0.00000941, 0.00001038, -0.00000309, -0.00000458, 0.00000420, -0.00000337, -0.00000299, 0.00000124, -0.00001168, 0.00000881, 0.00001871, -0.00001821, -0.00000336, 0.00000643, -0.00002650, 0.00001966, 0.00003463, -0.00004238, -0.00002250, 0.00002779, -0.00000650, -0.00000600, 0.00002262, 0.00001370, 0.00000224, -0.00000487, 0.00000917, 0.00002036, -0.00000588, -0.00000017, 0.00001828, -0.00001043, -0.00001895, -0.00000411, -0.00001030, -0.00000319, 0.00000545, -0.00000514, -0.00000010, 0.00000734, -0.00000294, -0.00000453, 0.00000089, -0.00000304, -0.00000377, 0.00000005, -0.00000081, 0.00000197, 0.00000293, -0.00000563, -0.00000608, -0.00000188, -0.00000333, 0.00000090, -0.00000059, -0.00000900, -0.00000201, 0.00000418, -0.00000237, -0.00000295, -0.00000006, -0.00000162, -0.00000276, -0.00000067, 0.00000173, -0.00000083, },
		{ 0.00000063, -0.00000158, -0.00000066, -0.00000269, -0.00000745, -0.00001136, -0.00001096, -0.00000312, -0.00000869, -0.00002979, -0.00002878, 0.00000854, 0.00004884, 0.00003976, 0.00000716, 0.00002762, 0.00004986, 0.00002118, 0.00000615, -0.00001072, -0.00004033, -0.00001020, 0.00001610, -0.00001229, -0.00000947, 0.00000602, -0.00000986, -0.00000453, 0.00001101, -0.00000059, -0.00000872, -0.00000158, -0.00000289, -0.00000372, 0.00001077, 0.00000827, -0.00000878, -0.00000073, 0.00000380, -0.00000137, 0.00000917, -0.00000091, -0.00001068, 0.00000452, -0.00000314, 0.00000089, 0.00001596, -0.00001253, -0.00001131, 0.00001548, -0.00000318, -0.00000400, 0.00000757, -0.00000673, 0.00000550, 0.00000747, -0.00001633, -0.00000143, 0.00001244, -0.00000168, -0.00000419, -0.00000465, -0.00000660, 0.00000355, 0.00001237, 0.00000351, -0.00000817, 0.00000163, 0.00000324, -0.00000647, 0.00000189, -0.00000643, -0.00000589, 0.00002227, -0.00000226, -0.00001776, 0.00001173, -0.00001590, -0.00001118, 0.00004345, -0.00000781, -0.00004692, 0.00001255, 0.00001381, -0.00001330, 0.00001149, 0.00002240, 0.00000875, -0.00000571, 0.00000163, 0.00002398, 0.00000462, -0.00000834, 0.00001630, 0.00000113, -0.00002121, -0.00000693, -0.00000841, -0.00000924, 0.00000531, -0.00000142, -0.00000436, 0.00000703, 0.00000116, -0.00000596, -0.00000021, -0.00000091, -0.00000443, -0.00000116, -0.00000048, 0.00000043, 0.00000375, -0.00000234, -0.00000729, -0.00000286, -0.00000281, -0.00000121, 0.00000132, -0.00000671, -0.00000612, 0.00000350, 0.00000070, -0.00000372, -0.00000100, -0.00000063, -0.00000254, -0.00000207, 0.00000091, },
		{ -0.00000057, 0.00000106, -0.00000047, -0.00000086, -0.00000155, -0.00000452, -0.00000972, -0.00001341, -0.00000544, -0.00000371, -0.00002235, -0.00003005, -0.00000791, 0.00003317, 0.00004984, 0.00001543, 0.00000904, 0.00004612, 0.00003404, 0.00000776, 0.00000456, -0.00002913, -0.00003207, 0.00001179, 0.00000342, -0.00001610, 0.00000197, -0.00000173, -0.00001083, 0.00000560, 0.00000746, -0.00000676, -0.00000554, -0.00000025, -0.00000567, 0.00000335, 0.00001436, -0.00000384, -0.00000780, 0.00000607, -0.00000053, 0.00000438, 0.00000689, -0.00001126, -0.00000148, 0.00000285, -0.00000537, 0.00001424, 0.00000094, -0.00001970, 0.00000912, 0.00000712, -0.00000905, 0.00000654, -0.00000139, -0.00000227, 0.00001156, -0.00000932, -0.00001071, 0.00001138, 0.00000419, -0.00000481, -0.00000406, -0.00000635, -0.00000130, 0.00000948, 0.00001012, -0.00000579, -0.00000551, 0.00000757, -0.00000342, -0.00000326, 0.00000095, -0.00001214, 0.00001145, 0.00001572, -0.00001966, 0.00000121, 0.00000274, -0.00002653, 0.00002384, 0.00002750, -0.00004262, -0.00001819, 0.00002285, -0.00000418, -0.00000316, 0.00002161, 0.00001939, -0.00000151, -0.00000690, 0.00002111, 0.00001785, -0.00000999, 0.00000820, 0.00001119, -0.00001850, -0.00001195, -0.00000543, -0.00001332, 0.00000109, 0.00000324, -0.00000631, 0.00000383, 0.00000525, -0.00000531, -0.00000251, 0.00000057, -0.00000373, -0.00000262, -0.00000047, -0.00000049, 0.00000315, 0.00000078, -0.00000668, -0.00000466, -0.00000224, -0.00000255, 0.00000119, -0.00000340, -0.00000820, 0.00000046, 0.00000322, -0.00000271, -0.00000242, -0.00000036, -0.00000172, -0.00000269, },
		{ -0.00000249, -0.00000181, 0.00000033, 0.00000050, -0.00000026, -0.00000141, -0.00000263, -0.00000626, -0.00001406, -0.00000967, -0.00000105, -0.00001524, -0.00002805, -0.00001651, 0.00001410, 0.00004793, 0.00003201, -0.00000041, 0.00002999, 0.00004574, 0.00001240, 0.00000875, -0.00000802, -0.00004013, -0.00000794, 0.00001524, -0.00001053, -0.00000759, 0.00000313, -0.00000901, -0.00000352, 0.00000972, 0.00000056, -0.00000855, -0.00000120, -0.00000281, -0.00000435, 0.00001320, 0.00000629, -0.00001209, 0.00000218, 0.00000458, -0.00000067, 0.00000954, -0.00000499, -0.00000926, 0.00000525, -0.00000468, 0.00000594, 0.00001220, -0.00001767, -0.00000376, 0.00001438, -0.00000688, 0.00000025, 0.00000467, -0.00000582, 0.00000892, 0.00000013, -0.00001510, 0.00000511, 0.00000943, -0.00000284, -0.00000452, -0.00000577, -0.00000357, 0.00000449, 0.00001222, 0.00000195, -0.00001095, 0.00000447, 0.00000476, -0.00000755, 0.00000242, -0.00000768, -0.00000400, 0.00002226, -0.00000633, -0.00001402, 0.00001170, -0.00001960, -0.00000605, 0.00004028, -0.00001316, -0.00004153, 0.00001024, 0.00001176, -0.00000860, 0.00001055, 0.00002646, 0.00000984, -0.00001181, 0.00001075, 0.00002884, -0.00000280, -0.00000231, 0.00001604, -0.00001072, -0.00001757, -0.00000363, -0.00001349, -0.00000617, 0.00000610, -0.00000477, -0.00000107, 0.00000751, -0.00000237, -0.00000485, 0.00000067, -0.00000214, -0.00000358, -0.00000098, -0.00000093, 0.00000186, 0.00000284, -0.00000453, -0.00000628, -0.00000241, -0.00000283, -0.00000016, -0.00000071, -0.00000780, -0.00000342, 0.00000382, -0.00000038, -0.00000326, -0.00000097, -0.00000095, },
		{ -0.00000070, -0.00000189, -0.00000229, -0.00000090, 0.00000043, 0.00000087, -0.00000094, -0.00000242, -0.00000273, -0.00001155, -0.00001438, -0.00000168, -0.00000824, -0.00002522, -0.00001934, -0.00000045, 0.00003404, 0.00004592, 0.00000549, 0.00000844, 0.00004730, 0.00002466, 0.00000550, 0.00000852, -0.00002884, -0.00003003, 0.00001173, 0.00000370, -0.00001279, -0.00000006, -0.00000270, -0.00000885, 0.00000431, 0.00000800, -0.00000597, -0.00000575, 0.00000060, -0.00000683, 0.00000496, 0.00001501, -0.00000838, -0.00000667, 0.00000849, -0.00000080, 0.00000594, 0.00000370, -0.00001254, 0.00000111, 0.00000060, -0.00000201, 0.00001522, -0.00000696, -0.00001534, 0.00001312, 0.00000142, -0.00000616, 0.00000665, -0.00000389, 0.00000288, 0.00000710, -0.00001327, -0.00000323, 0.00001125, 0.00000124, -0.00000433, -0.00000587, -0.00000442, 0.00000088, 0.00000919, 0.00000977, -0.00000921, -0.00000469, 0.00001086, -0.00000467, -0.00000291, 0.00000039, -0.00001218, 0.00001368, 0.00001186, -0.00001924, 0.00000463, -0.00000177, -0.00002407, 0.00002577, 0.00001992, -0.00004017, -0.00001615, 0.00001809, -0.00000075, -0.00000194, 0.00002309, 0.00002398, -0.00000742, -0.00000387, 0.00003142, 0.00001217, -0.00000943, 0.00001339, 0.00000009, -0.00002097, -0.00000526, -0.00000961, -0.00001348, 0.00000458, -0.00000027, -0.00000535, 0.00000659, 0.00000202, -0.00000592, -0.00000073, -0.00000046, -0.00000366, -0.00000180, -0.00000115, 0.00000043, 0.00000356, -0.00000163, -0.00000686, -0.00000347, -0.00000247, -0.00000164, 0.00000047, -0.00000560, -0.00000650, 0.00000220, 0.00000213, -0.00000286, -0.00000200, },
		{ -0.00000281, -0.00000100, -0.00000138, -0.00000214, -0.00000172, -0.00000064, 0.00000142, 0.00000057, -0.00000288, -0.00000116, -0.00000640, -0.00001685, -0.00000615, -0.00000216, -0.00002125, -0.00002062, -0.00000706, 0.00001562, 0.00004744, 0.00002327, -0.00000607, 0.00003387, 0.00003988, 0.00000535, 0.00001168, -0.00000668, -0.00003871, -0.00000693, 0.00001426, -0.00000703, -0.00000754, 0.00000049, -0.00000699, -0.00000397, 0.00000918, 0.00000207, -0.00000910, -0.00000030, -0.00000332, -0.00000445, 0.00001567, 0.00000269, -0.00001390, 0.00000570, 0.00000449, 0.00000051, 0.00000854, -0.00000851, -0.00000671, 0.00000446, -0.00000430, 0.00001024, 0.00000540, -0.00001885, 0.00000372, 0.00001012, -0.00000764, 0.00000345, 0.00000077, -0.00000215, 0.00000917, -0.00000716, -0.00001000, 0.00000898, 0.00000553, -0.00000262, -0.00000581, -0.00000568, -0.00000043, 0.00000443, 0.00001238, -0.00000061, -0.00001272, 0.00000844, 0.00000474, -0.00000838, 0.00000309, -0.00000903, -0.00000171, 0.00002089, -0.00000924, -0.00000976, 0.00000959, -0.00002119, -0.00000112, 0.00003466, -0.00001577, -0.00003666, 0.00000625, 0.00001160, -0.00000527, 0.00001061, 0.00003186, 0.00000743, -0.00001472, 0.00002200, 0.00002887, -0.00000733, 0.00000439, 0.00000962, -0.00001920, -0.00001087, -0.00000418, -0.00001703, -0.00000191, 0.00000455, -0.00000653, 0.00000253, 0.00000599, -0.00000467, -0.00000306, 0.00000050, -0.00000288, -0.00000258, -0.00000132, -0.00000079, 0.00000304, 0.00000116, -0.00000602, -0.00000503, -0.00000217, -0.00000249, 0.00000015, -0.00000290, -0.00000771, -0.00000095, 0.00000359, -0.00000123, },
		{ 0.00000102, -0.00000285, -0.00000165, -0.00000118, -0.00000185, -0.00000179, -0.00000182, 0.00000058, 0.00000246, -0.00000230, -0.00000210, -0.00000118, -0.00001479, -0.00001285, 0.00000062, -0.00001491, -0.00002254, -0.00000840, 0.00000193, 0.00003496, 0.00004118, -0.00000414, 0.00001077, 0.00004696, 0.00001551, 0.00000506, 0.00001088, -0.00002790, -0.00002811, 0.00001030, 0.00000598, -0.00001023, -0.00000328, -0.00000187, -0.00000787, 0.00000274, 0.00000967, -0.00000604, -0.00000566, 0.00000127, -0.00000817, 0.00000742, 0.00001403, -0.00001264, -0.00000387, 0.00000993, -0.00000079, 0.00000687, -0.00000022, -0.00001202, 0.00000259, -0.00000114, 0.00000268, 0.00001246, -0.00001309, -0.00000811, 0.00001328, -0.00000301, -0.00000199, 0.00000419, -0.00000374, 0.00000706, -0.00000001, -0.00001306, 0.00000379, 0.00000842, -0.00000006, -0.00000436, -0.00000742, -0.00000138, 0.00000183, 0.00000906, 0.00000882, -0.00001273, -0.00000196, 0.00001313, -0.00000656, -0.00000189, -0.00000065, -0.00001157, 0.00001502, 0.00000781, -0.00001710, 0.00000615, -0.00000582, -0.00001983, 0.00002481, 0.00001346, -0.00003606, -0.00001664, 0.00001454, 0.00000265, -0.00000146, 0.00002718, 0.00002596, -0.00001329, 0.00000374, 0.00003805, 0.00000598, -0.00000537, 0.00001301, -0.00001150, -0.00001802, -0.00000133, -0.00001467, -0.00001092, 0.00000601, -0.00000366, -0.00000275, 0.00000739, -0.00000112, -0.00000507, 0.00000015, -0.00000155, -0.00000303, -0.00000154, -0.00000158, 0.00000164, 0.00000308, -0.00000386, -0.00000638, -0.00000252, -0.00000248, -0.00000102, -0.00000091, -0.00000690, -0.00000436, 0.00000324, },
		{ 0.00000108, 0.00000296, -0.00000188, -0.00000232, -0.00000124, -0.00000175, -0.00000150, -0.00000221, -0.00000132, 0.00000322, 0.00000010, -0.00000398, 0.00000120, -0.00000850, -0.00001804, -0.00000192, -0.00000682, -0.00002363, -0.00000988, -0.00000256, 0.00001596, 0.00004708, 0.00001317, -0.00000861, 0.00003784, 0.00003248, 0.00000067, 0.00001393, -0.00000587, -0.00003635, -0.00000772, 0.00001444, -0.00000271, -0.00000960, -0.00000079, -0.00000476, -0.00000572, 0.00000995, 0.00000322, -0.00000998, 0.00000099, -0.00000451, -0.00000359, 0.00001738, -0.00000200, -0.00001370, 0.00000895, 0.00000393, 0.00000192, 0.00000612, -0.00001067, -0.00000399, 0.00000273, -0.00000170, 0.00001201, -0.00000260, -0.00001562, 0.00000887, 0.00000446, -0.00000546, 0.00000424, -0.00000229, 0.00000310, 0.00000558, -0.00001206, -0.00000258, 0.00000939, 0.00000230, -0.00000179, -0.00000783, -0.00000404, 0.00000199, 0.00000388, 0.00001262, -0.00000412, -0.00001273, 0.00001260, 0.00000324, -0.00000845, 0.00000347, -0.00001017, 0.00000063, 0.00001834, -0.00001046, -0.00000603, 0.00000614, -0.00002027, 0.00000218, 0.00002786, -0.00001552, -0.00003353, 0.00000177, 0.00001276, -0.00000352, 0.00001283, 0.00003692, 0.00000244, -0.00001307, 0.00003271, 0.00002546, -0.00000796, 0.00000837, -0.00000094, -0.00002221, -0.00000406, -0.00000785, -0.00001777, 0.00000181, 0.00000150, -0.00000621, 0.00000514, 0.00000332, -0.00000533, -0.00000150, -0.00000033, -0.00000291, -0.00000186, -0.00000185, -0.00000009, 0.00000362, -0.00000094, -0.00000675, -0.00000372, -0.00000198, -0.00000208, -0.00000029, -0.00000479, -0.00000672, },
		{ -0.00000726, -0.00000211, 0.00000365, -0.00000004, -0.00000264, -0.00000148, -0.00000177, -0.00000142, -0.00000166, -0.00000299, 0.00000189, 0.00000327, -0.00000429, -0.00000026, -0.00000113, -0.00001804, -0.00000903, -0.00000031, -0.00002090, -0.00001454, -0.00000073, 0.00000126, 0.00003698, 0.00003435, -0.00001202, 0.00001536, 0.00004426, 0.00000781, 0.00000558, 0.00001179, -0.00002578, -0.00002738, 0.00000853, 0.00001030, -0.00000956, -0.00000650, 0.00000044, -0.00000850, 0.00000183, 0.00001192, -0.00000704, -0.00000497, 0.00000137, -0.00000914, 0.00001023, 0.00001123, -0.00001561, -0.00000010, 0.00001029, -0.00000024, 0.00000662, -0.00000400, -0.00001008, 0.00000261, -0.00000127, 0.00000708, 0.00000637, -0.00001553, -0.00000026, 0.00000992, -0.00000468, 0.00000159, 0.00000027, -0.00000061, 0.00000835, -0.00000769, -0.00000867, 0.00000836, 0.00000434, 0.00000029, -0.00000547, -0.00000789, 0.00000203, 0.00000162, 0.00000935, 0.00000684, -0.00001547, 0.00000225, 0.00001373, -0.00000830, -0.00000051, -0.00000211, -0.00001031, 0.00001512, 0.00000439, -0.00001392, 0.00000555, -0.00000826, -0.00001509, 0.00002115, 0.00000928, -0.00003174, -0.00001915, 0.00001262, 0.00000500, -0.00000042, 0.00003307, 0.00002467, -0.00001687, 0.00001417, 0.00004039, 0.00000170, -0.00000057, 0.00000673, -0.00002002, -0.00001166, -0.00000126, -0.00001848, -0.00000692, 0.00000509, -0.00000567, 0.00000030, 0.00000629, -0.00000314, -0.00000356, 0.00000005, -0.00000220, -0.00000223, -0.00000177, -0.00000151, 0.00000275, 0.00000182, -0.00000563, -0.00000543, -0.00000174, -0.00000237, -0.00000087, -0.00000255, },
		{ -0.00000123, -0.00000610, -0.00000511, 0.00000258, 0.00000204, -0.00000224, -0.00000187, -0.00000175, -0.00000175, -0.00000087, -0.00000332, -0.00000098, 0.00000509, -0.00000174, -0.00000369, 0.00000323, -0.00001207, -0.00001661, 0.00000081, -0.00001359, -0.00002052, 0.00000053, -0.00000280, 0.00001745, 0.00004537, 0.00000252, -0.00000747, 0.00004027, 0.00002460, -0.00000175, 0.00001479, -0.00000457, -0.00003397, -0.00001013, 0.00001676, 0.00000106, -0.00001324, -0.00000007, -0.00000341, -0.00000803, 0.00001209, 0.00000343, -0.00001068, 0.00000231, -0.00000611, -0.00000171, 0.00001757, -0.00000681, -0.00001155, 0.00001127, 0.00000345, 0.00000301, 0.00000266, -0.00001106, -0.00000203, 0.00000112, 0.00000248, 0.00001019, -0.00000942, -0.00000900, 0.00001028, -0.00000046, -0.00000153, 0.00000228, -0.00000292, 0.00000793, -0.00000111, -0.00001293, 0.00000488, 0.00000676, 0.00000077, -0.00000134, -0.00000988, -0.00000103, 0.00000316, 0.00000343, 0.00001242, -0.00000805, -0.00001066, 0.00001590, 0.00000082, -0.00000760, 0.00000322, -0.00001078, 0.00000250, 0.00001514, -0.00000990, -0.00000376, 0.00000251, -0.00001726, 0.00000294, 0.00002145, -0.00001319, -0.00003272, -0.00000206, 0.00001418, -0.00000273, 0.00001765, 0.00004002, -0.00000324, -0.00000689, 0.00004079, 0.00002093, -0.00000561, 0.00000722, -0.00001218, -0.00001970, 0.00000025, -0.00001272, -0.00001573, 0.00000365, -0.00000167, -0.00000425, 0.00000598, 0.00000068, -0.00000463, -0.00000075, -0.00000124, -0.00000242, -0.00000165, -0.00000222, 0.00000095, 0.00000346, -0.00000311, -0.00000677, -0.00000243, -0.00000183, -0.00000184, },
		},
		{ { 0.00000476, 0.00000500, 0.00000062, -0.00000868, -0.00000992, -0.00000043, 0.00000452, -0.00000040, -0.00000239, 0.00000093, -0.00000088, -0.00000414, -0.00000035, 0.00000401, 0.00000272, -0.00000256, -0.00000595, 0.00000123, 0.00000898, 0.00000169, -0.00000566, -0.00000399, -0.00000556, -0.00000148, 0.00000402, -0.00000827, -0.00000741, 0.00000936, -0.00000968, -0.00002670, 0.00000371, 0.00000826, -0.00001237, 0.00002109, 0.00004261, -0.00000613, -0.00002511, 0.00001196, 0.00002847, 0.00001701, -0.00000155, -0.00000966, 0.00000713, 0.00000119, -0.00002276, 0.00000444, 0.00002427, -0.00001490, -0.00001576, 0.00001600, -0.00000635, -0.00001760, 0.00001465, 0.00000609, -0.00001550, 0.00000362, 0.00000744, -0.00000820, 0.00000000, 0.00000641, -0.00000163, -0.00000158, 0.00000033, -0.00000091, -0.00000066, -0.00000091, 0.00000033, -0.00000158, -0.00000163, 0.00000641, 0.00000000, -0.00000820, 0.00000744, 0.00000362, -0.00001550, 0.00000609, 0.00001465, -0.00001760, -0.00000635, 0.00001600, -0.00001576, -0.00001490, 0.00002427, 0.00000444, -0.00002276, 0.00000119, 0.00000713, -0.00000966, -0.00000155, 0.00001701, 0.00002847, 0.00001196, -0.00002511, -0.00000613, 0.00004261, 0.00002109, -0.00001237, 0.00000826, 0.00000371, -0.00002670, -0.00000968, 0.00000936, -0.00000741, -0.00000827, 0.00000402, -0.00000148, -0.00000556, -0.00000399, -0.00000566, 0.00000169, 0.00000898, 0.00000123, -0.00000595, -0.00000256, 0.00000272, 0.00000401, -0.00000035, -0.00000414, -0.00000088, 0.00000093, -0.00000239, -0.00000040, 0.00000452, -0.00000043, -0.00000992, -0.00000868, 0.00000062, 0.00000500, },
		{ 0.00000319, 0.00000521, 0.00000473, 0.00000417, -0.00000261, -0.00001100, -0.00000778, 0.00000297, 0.00000441, -0.00000190, -0.00000180, 0.00000104, -0.00000223, -0.00000366, 0.00000141, 0.00000458, 0.00000115, -0.00000531, -0.00000396, 0.00000652, 0.00000693, -0.00000322, -0.00000512, -0.00000479, -0.00000463, 0.00000377, -0.00000109, -0.00001211, 0.00000358, 0.00000471, -0.00002471, -0.00001294, 0.00001516, -0.00000546, -0.00000307, 0.00003909, 0.00002109, -0.00002387, -0.00000939, 0.00002027, 0.00002454, 0.00001250, -0.00001175, -0.00000640, 0.00001572, -0.00000974, -0.00002078, 0.00002080, 0.00001179, -0.00002556, 0.00000082, 0.00001323, -0.00001880, -0.00000441, 0.00001841, -0.00000641, -0.00001037, 0.00000928, 0.00000033, -0.00000686, 0.00000386, 0.00000358, -0.00000149, -0.00000116, -0.00000052, 0.00000016, -0.00000166, -0.00000055, 0.00000137, -0.00000395, 0.00000107, 0.00000764, -0.00000625, -0.00000394, 0.00001156, -0.00000674, -0.00001161, 0.00001611, 0.00000135, -0.00002020, 0.00000807, 0.00000600, -0.00002347, 0.00000306, 0.00002316, -0.00000996, -0.00001677, 0.00000346, 0.00000009, -0.00000531, 0.00000434, 0.00002317, 0.00003028, -0.00000670, -0.00002698, 0.00002348, 0.00004359, -0.00000085, -0.00000545, 0.00001175, -0.00001589, -0.00002357, 0.00000521, 0.00000183, -0.00001198, -0.00000038, 0.00000235, -0.00000550, -0.00000404, -0.00000494, -0.00000354, 0.00000688, 0.00000647, -0.00000346, -0.00000492, -0.00000030, 0.00000339, 0.00000320, -0.00000218, -0.00000387, 0.00000051, 0.00000036, -0.00000261, 0.00000088, 0.00000357, -0.00000374, -0.00001061, -0.00000555, },
		{ -0.00000988, -0.00000208, 0.00000488, 0.00000502, 0.00000462, 0.00000256, -0.00000616, -0.00001207, -0.00000427, 0.00000585, 0.00000303, -0.00000320, -0.00000084, 0.00000071, -0.00000334, -0.00000266, 0.00000322, 0.00000454, -0.00000183, -0.00000657, 0.00000105, 0.00000888, 0.00000182, -0.00000520, -0.00000465, -0.00000594, 0.00000021, 0.00000460, -0.00000947, -0.00000640, 0.00001062, -0.00001117, -0.00002601, 0.00000748, 0.00000957, -0.00001326, 0.00001862, 0.00003771, -0.00000629, -0.00002187, 0.00000531, 0.00002050, 0.00002385, 0.00000272, -0.00002024, 0.00000720, 0.00001567, -0.00002342, -0.00000663, 0.00002805, -0.00000835, -0.00002140, 0.00001456, -0.00000040, -0.00002009, 0.00001043, 0.00001135, -0.00001319, -0.00000136, 0.00000821, -0.00000446, -0.00000308, 0.00000447, 0.00000223, -0.00000125, -0.00000212, 0.00000061, -0.00000011, -0.00000300, 0.00000197, -0.00000031, -0.00000540, 0.00000632, 0.00000404, -0.00001032, 0.00000455, 0.00000811, -0.00001489, -0.00000011, 0.00001617, -0.00001278, -0.00001115, 0.00001407, -0.00000893, -0.00001780, 0.00001690, 0.00001263, -0.00001647, -0.00001140, 0.00000079, -0.00000007, -0.00000120, 0.00000891, 0.00003282, 0.00002103, -0.00002550, -0.00000785, 0.00004631, 0.00002539, -0.00001065, 0.00000611, -0.00000056, -0.00002752, -0.00000809, 0.00000853, -0.00000860, -0.00000756, 0.00000374, -0.00000289, -0.00000562, -0.00000339, -0.00000566, 0.00000151, 0.00000877, 0.00000134, -0.00000520, -0.00000281, 0.00000093, 0.00000385, 0.00000185, -0.00000380, -0.00000270, 0.00000157, -0.00000060, -0.00000251, 0.00000177, 0.00000185, -0.00000655, },
		{ -0.00000039, -0.00000856, -0.00000786, 0.00000122, 0.00000563, 0.00000465, 0.00000427, 0.00000013, -0.00000960, -0.00001144, 0.00000025, 0.00000741, 0.00000057, -0.00000381, 0.00000035, -0.00000010, -0.00000418, -0.00000105, 0.00000503, 0.00000295, -0.00000531, -0.00000443, 0.00000639, 0.00000669, -0.00000273, -0.00000491, -0.00000582, -0.00000407, 0.00000561, -0.00000174, -0.00001281, 0.00000538, 0.00000468, -0.00002586, -0.00001031, 0.00001785, -0.00000584, -0.00000376, 0.00003563, 0.00001652, -0.00002123, -0.00000822, 0.00001048, 0.00002158, 0.00002212, -0.00001392, -0.00001701, 0.00002329, 0.00000224, -0.00002835, 0.00001350, 0.00001966, -0.00002345, -0.00000564, 0.00001589, -0.00001466, -0.00000966, 0.00001755, -0.00000021, -0.00001163, 0.00000509, 0.00000366, -0.00000530, -0.00000033, 0.00000427, 0.00000211, -0.00000277, -0.00000174, 0.00000222, -0.00000265, -0.00000174, 0.00000396, -0.00000474, -0.00000254, 0.00001001, -0.00000365, -0.00000790, 0.00001137, -0.00000185, -0.00001453, 0.00001124, 0.00000602, -0.00001851, 0.00000190, 0.00000846, -0.00001807, -0.00000415, 0.00002049, 0.00000122, -0.00001770, -0.00001048, 0.00000075, 0.00000307, -0.00000051, 0.00001857, 0.00003857, -0.00000033, -0.00002899, 0.00002343, 0.00004825, 0.00000332, -0.00000553, 0.00000742, -0.00001903, -0.00002252, 0.00000590, 0.00000043, -0.00001224, -0.00000000, 0.00000140, -0.00000634, -0.00000377, -0.00000455, -0.00000350, 0.00000660, 0.00000619, -0.00000297, -0.00000429, -0.00000131, 0.00000181, 0.00000408, -0.00000020, -0.00000461, -0.00000086, 0.00000192, -0.00000181, -0.00000213, 0.00000224, },
		{ -0.00000139, 0.00000221, -0.00000295, -0.00000940, -0.00000483, 0.00000381, 0.00000560, 0.00000430, 0.00000347, -0.00000308, -0.00001234, -0.00000879, 0.00000500, 0.00000702, -0.00000237, -0.00000328, 0.00000145, -0.00000153, -0.00000452, 0.00000142, 0.00000601, -0.00000057, -0.00000695, 0.00000079, 0.00000856, 0.00000175, -0.00000443, -0.00000506, -0.00000676, 0.00000179, 0.00000566, -0.00001087, -0.00000593, 0.00001219, -0.00001223, -0.00002557, 0.00001073, 0.00001021, -0.00001386, 0.00001795, 0.00003268, -0.00000889, -0.00001723, 0.00000237, 0.00001025, 0.00002725, 0.00001100, -0.00002868, 0.00000098, 0.00002871, -0.00001773, -0.00001792, 0.00002680, 0.00000026, -0.00002403, 0.00001010, 0.00000501, -0.00001954, 0.00000470, 0.00001431, -0.00000841, -0.00000543, 0.00000654, -0.00000016, -0.00000434, 0.00000100, 0.00000515, 0.00000074, -0.00000462, 0.00000104, 0.00000136, -0.00000495, 0.00000237, 0.00000199, -0.00000812, 0.00000429, 0.00000789, -0.00001025, 0.00000043, 0.00001087, -0.00001169, -0.00000569, 0.00001455, -0.00000702, -0.00001352, 0.00000937, -0.00000288, -0.00001659, 0.00000817, 0.00001626, -0.00000660, -0.00001944, -0.00001010, 0.00000573, 0.00000287, 0.00000182, 0.00003334, 0.00003073, -0.00002225, -0.00001030, 0.00004801, 0.00003106, -0.00000757, 0.00000314, -0.00000545, -0.00002832, -0.00000661, 0.00000802, -0.00000960, -0.00000742, 0.00000342, -0.00000381, -0.00000596, -0.00000313, -0.00000527, 0.00000150, 0.00000825, 0.00000129, -0.00000439, -0.00000269, -0.00000059, 0.00000295, 0.00000351, -0.00000253, -0.00000416, 0.00000108, 0.00000135, -0.00000304, },
		{ -0.00000013, -0.00000393, -0.00000029, 0.00000148, -0.00000550, -0.00000881, -0.00000141, 0.00000539, 0.00000513, 0.00000397, 0.00000200, -0.00000683, -0.00001366, -0.00000417, 0.00000880, 0.00000449, -0.00000478, -0.00000155, 0.00000184, -0.00000351, -0.00000371, 0.00000444, 0.00000486, -0.00000475, -0.00000486, 0.00000615, 0.00000621, -0.00000234, -0.00000428, -0.00000679, -0.00000403, 0.00000759, -0.00000193, -0.00001398, 0.00000680, 0.00000523, -0.00002677, -0.00000820, 0.00001998, -0.00000661, -0.00000353, 0.00003372, 0.00001061, -0.00002037, -0.00000392, 0.00000293, 0.00001440, 0.00003068, -0.00001077, -0.00002885, 0.00002432, 0.00001661, -0.00003042, 0.00000275, 0.00002434, -0.00001740, -0.00001142, 0.00001524, -0.00000895, -0.00001291, 0.00001383, 0.00000544, -0.00001003, 0.00000012, 0.00000506, -0.00000201, -0.00000378, 0.00000294, 0.00000594, -0.00000288, -0.00000383, 0.00000383, -0.00000251, -0.00000362, 0.00000581, -0.00000386, -0.00000625, 0.00001035, 0.00000013, -0.00001041, 0.00000889, 0.00000288, -0.00001457, 0.00000506, 0.00000830, -0.00001430, -0.00000359, 0.00000762, -0.00001107, -0.00000810, 0.00001403, 0.00001003, -0.00001302, -0.00002281, -0.00000451, 0.00001048, -0.00000143, 0.00001261, 0.00004388, 0.00000864, -0.00002853, 0.00002139, 0.00005212, 0.00000945, -0.00000526, 0.00000204, -0.00002232, -0.00002149, 0.00000650, -0.00000051, -0.00001260, -0.00000019, 0.00000075, -0.00000686, -0.00000397, -0.00000424, -0.00000303, 0.00000626, 0.00000562, -0.00000255, -0.00000352, -0.00000185, 0.00000018, 0.00000406, 0.00000179, -0.00000430, -0.00000253, 0.00000248, },
		{ -0.00000022, 0.00000279, -0.00000224, -0.00000402, 0.00000091, -0.00000014, -0.00000742, -0.00000683, 0.00000172, 0.00000601, 0.00000451, 0.00000351, -0.00000031, -0.00001065, -0.00001283, 0.00000180, 0.00001042, 0.00000037, -0.00000559, 0.00000089, 0.00000084, -0.00000539, -0.00000118, 0.00000670, 0.00000118, -0.00000694, 0.00000044, 0.00000813, 0.00000143, -0.00000362, -0.00000499, -0.00000776, 0.00000287, 0.00000715, -0.00001198, -0.00000619, 0.00001372, -0.00001253, -0.00002529, 0.00001331, 0.00001044, -0.00001450, 0.00001864, 0.00002851, -0.00001370, -0.00001292, 0.00000353, -0.00000023, 0.00002587, 0.00002135, -0.00003244, -0.00000989, 0.00003718, -0.00000663, -0.00002649, 0.00002035, 0.00000862, -0.00002259, 0.00000392, 0.00000833, -0.00001609, -0.00000094, 0.00001425, -0.00000232, -0.00000739, 0.00000286, 0.00000345, -0.00000348, -0.00000304, 0.00000635, 0.00000394, -0.00000633, 0.00000022, 0.00000318, -0.00000613, 0.00000151, 0.00000451, -0.00000920, 0.00000096, 0.00001049, -0.00000808, -0.00000366, 0.00001125, -0.00000693, -0.00000929, 0.00001023, -0.00000191, -0.00001293, 0.00000329, 0.00000079, -0.00001193, 0.00000049, 0.00001499, 0.00000401, -0.00002162, -0.00002271, 0.00000605, 0.00000854, -0.00000257, 0.00003090, 0.00003933, -0.00001545, -0.00001208, 0.00004748, 0.00003734, -0.00000280, -0.00000041, -0.00001111, -0.00002900, -0.00000533, 0.00000763, -0.00001015, -0.00000771, 0.00000275, -0.00000421, -0.00000628, -0.00000335, -0.00000468, 0.00000179, 0.00000753, 0.00000105, -0.00000362, -0.00000230, -0.00000170, 0.00000158, 0.00000439, -0.00000068, -0.00000486, },
		{ -0.00000310, -0.00000397, 0.00000203, 0.00000169, -0.00000432, -0.00000305, 0.00000157, -0.00000241, -0.00000810, -0.00000399, 0.00000409, 0.00000592, 0.00000385, 0.00000274, -0.00000351, -0.00001381, -0.00000932, 0.00000781, 0.00000900, -0.00000397, -0.00000419, 0.00000287, -0.00000170, -0.00000593, 0.00000269, 0.00000648, -0.00000350, -0.00000515, 0.00000576, 0.00000565, -0.00000222, -0.00000343, -0.00000731, -0.00000446, 0.00000925, -0.00000142, -0.00001526, 0.00000739, 0.00000627, -0.00002707, -0.00000673, 0.00002155, -0.00000757, -0.00000300, 0.00003341, 0.00000453, -0.00002198, 0.00000201, -0.00000083, 0.00000427, 0.00003576, -0.00000280, -0.00003866, 0.00001857, 0.00003004, -0.00002603, -0.00000890, 0.00002419, -0.00000882, -0.00001467, 0.00001147, -0.00000329, -0.00001329, 0.00000830, 0.00000912, -0.00000591, -0.00000424, 0.00000411, 0.00000202, -0.00000556, 0.00000000, 0.00000898, -0.00000132, -0.00000610, 0.00000442, -0.00000139, -0.00000550, 0.00000643, -0.00000178, -0.00000906, 0.00000867, 0.00000399, -0.00001081, 0.00000496, 0.00000618, -0.00001194, -0.00000088, 0.00000760, -0.00000881, -0.00000702, 0.00000391, -0.00000444, -0.00000811, 0.00000589, 0.00001463, -0.00000461, -0.00003085, -0.00001412, 0.00001426, 0.00000076, 0.00000730, 0.00004592, 0.00001867, -0.00002484, 0.00001838, 0.00005449, 0.00001713, -0.00000418, -0.00000436, -0.00002586, -0.00002038, 0.00000683, -0.00000108, -0.00001275, -0.00000099, 0.00000011, -0.00000689, -0.00000446, -0.00000422, -0.00000223, 0.00000604, 0.00000484, -0.00000228, -0.00000273, -0.00000198, -0.00000127, 0.00000327, 0.00000342, },
		{ 0.00000443, 0.00000133, -0.00000467, -0.00000184, 0.00000337, -0.00000070, -0.00000549, -0.00000135, 0.00000116, -0.00000461, -0.00000739, -0.00000100, 0.00000564, 0.00000535, 0.00000318, 0.00000148, -0.00000750, -0.00001528, -0.00000326, 0.00001205, 0.00000470, -0.00000677, -0.00000099, 0.00000304, -0.00000492, -0.00000409, 0.00000621, 0.00000322, -0.00000630, -0.00000006, 0.00000764, 0.00000099, -0.00000308, -0.00000445, -0.00000850, 0.00000320, 0.00000874, -0.00001235, -0.00000710, 0.00001471, -0.00001190, -0.00002493, 0.00001498, 0.00001047, -0.00001524, 0.00002001, 0.00002585, -0.00001989, -0.00001041, 0.00000815, -0.00000869, 0.00001961, 0.00003110, -0.00003026, -0.00002253, 0.00003901, 0.00000748, -0.00002993, 0.00000997, 0.00001434, -0.00001712, -0.00000211, 0.00000859, -0.00001076, -0.00000497, 0.00001128, 0.00000354, -0.00000665, -0.00000180, 0.00000533, -0.00000084, -0.00000653, 0.00000556, 0.00000752, -0.00000661, -0.00000155, 0.00000468, -0.00000620, -0.00000018, 0.00000642, -0.00000853, -0.00000267, 0.00001116, -0.00000451, -0.00000656, 0.00000929, -0.00000198, -0.00001020, 0.00000432, 0.00000118, -0.00000976, -0.00000253, 0.00000136, -0.00000542, -0.00000429, 0.00000987, 0.00001264, -0.00001825, -0.00003383, 0.00000078, 0.00001310, -0.00000375, 0.00002710, 0.00004568, -0.00000596, -0.00001194, 0.00004511, 0.00004336, 0.00000365, -0.00000413, -0.00001761, -0.00002956, -0.00000421, 0.00000711, -0.00001019, -0.00000815, 0.00000155, -0.00000428, -0.00000635, -0.00000399, -0.00000414, 0.00000240, 0.00000681, 0.00000062, -0.00000300, -0.00000175, -0.00000232, 0.00000002, },
		{ -0.00000237, 0.00000192, 0.00000445, -0.00000130, -0.00000481, 0.00000087, 0.00000304, -0.00000357, -0.00000517, 0.00000020, -0.00000025, -0.00000597, -0.00000571, 0.00000171, 0.00000647, 0.00000436, 0.00000257, -0.00000056, -0.00001185, -0.00001392, 0.00000425, 0.00001286, -0.00000108, -0.00000668, 0.00000248, 0.00000076, -0.00000707, -0.00000008, 0.00000730, -0.00000153, -0.00000515, 0.00000512, 0.00000517, -0.00000238, -0.00000271, -0.00000713, -0.00000506, 0.00001013, -0.00000027, -0.00001612, 0.00000693, 0.00000739, -0.00002638, -0.00000587, 0.00002241, -0.00000844, -0.00000258, 0.00003424, -0.00000059, -0.00002597, 0.00000774, -0.00000019, -0.00000666, 0.00003571, 0.00000806, -0.00004367, 0.00000754, 0.00003931, -0.00001587, -0.00001854, 0.00001882, 0.00000010, -0.00001414, 0.00000568, 0.00000072, -0.00001078, 0.00000248, 0.00000998, -0.00000029, -0.00000675, 0.00000111, 0.00000557, -0.00000529, -0.00000367, 0.00001041, 0.00000155, -0.00000771, 0.00000377, 0.00000024, -0.00000672, 0.00000573, 0.00000070, -0.00001030, 0.00000554, 0.00000683, -0.00000908, 0.00000075, 0.00000727, -0.00000755, -0.00000521, 0.00000420, -0.00000360, -0.00000771, -0.00000143, 0.00000016, -0.00000468, -0.00000155, 0.00001452, 0.00000481, -0.00003320, -0.00002536, 0.00001283, 0.00000409, 0.00000407, 0.00004544, 0.00002826, -0.00001795, 0.00001567, 0.00005516, 0.00002561, -0.00000202, -0.00001149, -0.00002977, -0.00001913, 0.00000683, -0.00000145, -0.00001251, -0.00000225, -0.00000076, -0.00000647, -0.00000500, -0.00000456, -0.00000129, 0.00000604, 0.00000401, -0.00000222, -0.00000206, -0.00000175, },
		{ -0.00000117, -0.00000244, -0.00000150, 0.00000373, 0.00000315, -0.00000368, -0.00000331, 0.00000308, 0.00000088, -0.00000567, -0.00000363, 0.00000077, -0.00000194, -0.00000628, -0.00000363, 0.00000412, 0.00000651, 0.00000311, 0.00000207, -0.00000379, -0.00001549, -0.00000907, 0.00001107, 0.00000964, -0.00000598, -0.00000370, 0.00000423, -0.00000314, -0.00000659, 0.00000432, 0.00000513, -0.00000482, -0.00000067, 0.00000703, 0.00000066, -0.00000302, -0.00000368, -0.00000854, 0.00000282, 0.00000990, -0.00001167, -0.00000826, 0.00001465, -0.00001046, -0.00002404, 0.00001559, 0.00001031, -0.00001588, 0.00002136, 0.00002480, -0.00002620, -0.00001059, 0.00001474, -0.00001332, 0.00000963, 0.00003757, -0.00002248, -0.00003356, 0.00003374, 0.00002135, -0.00002705, -0.00000199, 0.00001564, -0.00000880, -0.00000610, 0.00000577, -0.00000508, -0.00000639, 0.00000633, 0.00000774, -0.00000339, -0.00000610, 0.00000498, 0.00000271, -0.00000852, 0.00000308, 0.00001047, -0.00000524, -0.00000364, 0.00000527, -0.00000522, -0.00000203, 0.00000715, -0.00000653, -0.00000562, 0.00000980, -0.00000061, -0.00000753, 0.00000572, 0.00000185, -0.00000848, -0.00000166, 0.00000155, -0.00000511, -0.00000663, -0.00000094, 0.00000104, -0.00000534, 0.00000288, 0.00001736, -0.00001107, -0.00004081, -0.00000858, 0.00001409, -0.00000254, 0.00002366, 0.00004953, 0.00000488, -0.00000916, 0.00004186, 0.00004849, 0.00001132, -0.00000760, -0.00002485, -0.00003010, -0.00000317, 0.00000636, -0.00000979, -0.00000852, -0.00000016, -0.00000426, -0.00000609, -0.00000485, -0.00000380, 0.00000317, 0.00000624, 0.00000011, -0.00000264, },
		{ -0.00000231, -0.00000166, -0.00000126, -0.00000299, 0.00000023, 0.00000479, 0.00000075, -0.00000492, -0.00000058, 0.00000360, -0.00000225, -0.00000611, -0.00000192, 0.00000034, -0.00000314, -0.00000593, -0.00000131, 0.00000618, 0.00000554, 0.00000206, 0.00000134, -0.00000834, -0.00001680, -0.00000129, 0.00001475, 0.00000355, -0.00000792, 0.00000057, 0.00000307, -0.00000667, -0.00000335, 0.00000682, 0.00000103, -0.00000458, 0.00000412, 0.00000485, -0.00000264, -0.00000250, -0.00000625, -0.00000533, 0.00000991, 0.00000115, -0.00001598, 0.00000547, 0.00000803, -0.00002448, -0.00000531, 0.00002230, -0.00000899, -0.00000238, 0.00003551, -0.00000399, -0.00003151, 0.00001158, 0.00000437, -0.00001596, 0.00003017, 0.00001903, -0.00004238, -0.00000596, 0.00004216, -0.00000210, -0.00002363, 0.00000920, 0.00000697, -0.00000967, -0.00000039, 0.00000200, -0.00000623, -0.00000207, 0.00000800, 0.00000538, -0.00000673, -0.00000301, 0.00000760, -0.00000318, -0.00000698, 0.00000989, 0.00000493, -0.00000801, 0.00000210, 0.00000170, -0.00000688, 0.00000408, 0.00000274, -0.00000981, 0.00000187, 0.00000791, -0.00000582, -0.00000254, 0.00000609, -0.00000270, -0.00000708, -0.00000091, -0.00000005, -0.00000586, -0.00000681, 0.00000184, 0.00000077, -0.00000640, 0.00001054, 0.00001254, -0.00003007, -0.00003513, 0.00000615, 0.00000602, 0.00000324, 0.00004380, 0.00003642, -0.00000858, 0.00001442, 0.00005453, 0.00003411, 0.00000113, -0.00001895, -0.00003408, -0.00001779, 0.00000655, -0.00000179, -0.00001183, -0.00000373, -0.00000197, -0.00000579, -0.00000539, -0.00000514, -0.00000046, 0.00000621, 0.00000331, },
		{ 0.00000587, -0.00000027, -0.00000259, -0.00000076, -0.00000209, -0.00000278, 0.00000242, 0.00000457, -0.00000214, -0.00000431, 0.00000218, 0.00000197, -0.00000484, -0.00000514, -0.00000098, -0.00000032, -0.00000378, -0.00000532, 0.00000147, 0.00000739, 0.00000371, 0.00000174, -0.00000049, -0.00001346, -0.00001425, 0.00000731, 0.00001383, -0.00000286, -0.00000628, 0.00000378, -0.00000051, -0.00000793, 0.00000110, 0.00000634, -0.00000234, -0.00000123, 0.00000614, 0.00000068, -0.00000345, -0.00000313, -0.00000757, 0.00000207, 0.00001006, -0.00000996, -0.00000905, 0.00001322, -0.00000862, -0.00002220, 0.00001519, 0.00000979, -0.00001612, 0.00002223, 0.00002498, -0.00003141, -0.00001350, 0.00002138, -0.00001327, -0.00000190, 0.00003883, -0.00001093, -0.00004004, 0.00002266, 0.00003172, -0.00001824, -0.00001265, 0.00001187, 0.00000033, -0.00000670, 0.00000079, -0.00000066, -0.00000502, 0.00000083, 0.00000935, 0.00000150, -0.00000880, 0.00000256, 0.00000599, -0.00000852, -0.00000028, 0.00001196, -0.00000250, -0.00000525, 0.00000467, -0.00000361, -0.00000338, 0.00000656, -0.00000396, -0.00000715, 0.00000688, 0.00000253, -0.00000643, 0.00000165, 0.00000371, -0.00000490, -0.00000635, -0.00000068, -0.00000045, -0.00000824, -0.00000505, 0.00000575, -0.00000297, -0.00000370, 0.00001753, -0.00000267, -0.00004225, -0.00001939, 0.00001012, -0.00000087, 0.00002170, 0.00005156, 0.00001580, -0.00000375, 0.00003889, 0.00005259, 0.00001951, -0.00001062, -0.00003253, -0.00003073, -0.00000222, 0.00000534, -0.00000913, -0.00000867, -0.00000222, -0.00000434, -0.00000557, -0.00000571, -0.00000367, 0.00000389, },
		{ 0.00000017, 0.00000635, 0.00000292, -0.00000241, -0.00000167, -0.00000067, -0.00000306, -0.00000157, 0.00000445, 0.00000276, -0.00000438, -0.00000190, 0.00000340, -0.00000098, -0.00000581, -0.00000391, -0.00000074, -0.00000055, -0.00000446, -0.00000419, 0.00000466, 0.00000703, 0.00000187, 0.00000201, -0.00000428, -0.00001726, -0.00000761, 0.00001376, 0.00000887, -0.00000694, -0.00000237, 0.00000430, -0.00000466, -0.00000637, 0.00000467, 0.00000378, -0.00000311, 0.00000269, 0.00000463, -0.00000268, -0.00000309, -0.00000498, -0.00000471, 0.00000861, 0.00000223, -0.00001447, 0.00000344, 0.00000759, -0.00002145, -0.00000460, 0.00002101, -0.00000919, -0.00000220, 0.00003647, -0.00000546, -0.00003733, 0.00001250, 0.00001143, -0.00002161, 0.00002016, 0.00002723, -0.00003489, -0.00001851, 0.00003788, 0.00001216, -0.00002274, -0.00000251, 0.00000990, -0.00000221, -0.00000485, 0.00000034, -0.00000104, -0.00000428, 0.00000393, 0.00000966, -0.00000423, -0.00000700, 0.00000751, -0.00000004, -0.00000896, 0.00000762, 0.00000787, -0.00000672, -0.00000002, 0.00000240, -0.00000598, 0.00000209, 0.00000373, -0.00000791, -0.00000132, 0.00000705, -0.00000203, -0.00000410, 0.00000324, 0.00000132, -0.00000645, -0.00000635, 0.00000104, -0.00000240, -0.00001081, 0.00000068, 0.00000642, -0.00000771, 0.00000454, 0.00001663, -0.00002310, -0.00004086, -0.00000424, 0.00000432, 0.00000385, 0.00004243, 0.00004304, 0.00000213, 0.00001524, 0.00005349, 0.00004205, 0.00000479, -0.00002635, -0.00003864, -0.00001649, 0.00000604, -0.00000219, -0.00001085, -0.00000523, -0.00000345, -0.00000508, -0.00000562, -0.00000577, },
		{ -0.00000643, -0.00000360, 0.00000435, 0.00000560, -0.00000032, -0.00000276, -0.00000071, -0.00000135, -0.00000365, 0.00000067, 0.00000546, -0.00000033, -0.00000472, 0.00000105, 0.00000231, -0.00000355, -0.00000543, -0.00000333, -0.00000030, -0.00000068, -0.00000548, -0.00000180, 0.00000733, 0.00000501, 0.00000114, 0.00000169, -0.00000983, -0.00001751, 0.00000139, 0.00001583, 0.00000226, -0.00000750, 0.00000152, 0.00000213, -0.00000751, -0.00000297, 0.00000623, 0.00000106, -0.00000145, 0.00000473, 0.00000118, -0.00000413, -0.00000330, -0.00000556, 0.00000155, 0.00000883, -0.00000763, -0.00000880, 0.00001045, -0.00000701, -0.00001910, 0.00001404, 0.00000860, -0.00001570, 0.00002246, 0.00002564, -0.00003464, -0.00001838, 0.00002627, -0.00000880, -0.00001245, 0.00003424, 0.00000158, -0.00004020, 0.00000845, 0.00003613, -0.00000539, -0.00001942, 0.00000373, 0.00000792, -0.00000353, -0.00000476, 0.00000136, -0.00000152, -0.00000371, 0.00000819, 0.00000664, -0.00000915, -0.00000111, 0.00000795, -0.00000668, -0.00000350, 0.00001157, 0.00000090, -0.00000569, 0.00000303, -0.00000200, -0.00000378, 0.00000495, -0.00000165, -0.00000702, 0.00000329, 0.00000412, -0.00000375, -0.00000174, 0.00000343, -0.00000068, -0.00000886, -0.00000469, 0.00000288, -0.00000738, -0.00000948, 0.00000773, 0.00000154, -0.00000797, 0.00001387, 0.00000428, -0.00003834, -0.00002861, 0.00000133, -0.00000116, 0.00002126, 0.00005299, 0.00002607, 0.00000365, 0.00003722, 0.00005606, 0.00002743, -0.00001334, -0.00004021, -0.00003146, -0.00000145, 0.00000412, -0.00000832, -0.00000863, -0.00000440, -0.00000457, -0.00000502, },
		{ -0.00000449, -0.00000582, -0.00000621, 0.00000064, 0.00000622, 0.00000282, -0.00000227, -0.00000213, -0.00000022, -0.00000258, -0.00000327, 0.00000353, 0.00000450, -0.00000334, -0.00000282, 0.00000265, -0.00000024, -0.00000457, -0.00000502, -0.00000307, 0.00000089, -0.00000168, -0.00000606, 0.00000200, 0.00000813, 0.00000243, 0.00000180, -0.00000075, -0.00001532, -0.00001316, 0.00000965, 0.00001339, -0.00000326, -0.00000533, 0.00000367, -0.00000145, -0.00000826, 0.00000084, 0.00000609, -0.00000048, 0.00000094, 0.00000424, -0.00000215, -0.00000449, -0.00000385, -0.00000281, 0.00000660, 0.00000231, -0.00001156, 0.00000153, 0.00000565, -0.00001770, -0.00000320, 0.00001854, -0.00000922, -0.00000170, 0.00003659, -0.00000534, -0.00004212, 0.00001027, 0.00001904, -0.00002246, 0.00000777, 0.00003046, -0.00002289, -0.00002702, 0.00002748, 0.00002364, -0.00001598, -0.00001358, 0.00000801, 0.00000639, -0.00000628, -0.00000353, 0.00000323, -0.00000382, -0.00000091, 0.00001157, -0.00000002, -0.00000965, 0.00000539, 0.00000306, -0.00000909, 0.00000433, 0.00000953, -0.00000404, -0.00000182, 0.00000202, -0.00000438, 0.00000045, 0.00000346, -0.00000532, -0.00000328, 0.00000462, 0.00000122, -0.00000372, -0.00000024, 0.00000360, -0.00000395, -0.00001076, -0.00000032, 0.00000130, -0.00001255, -0.00000242, 0.00001059, -0.00000566, -0.00000130, 0.00001640, -0.00001482, -0.00004120, -0.00001566, -0.00000217, 0.00000392, 0.00004213, 0.00004880, 0.00001316, 0.00001812, 0.00005313, 0.00004933, 0.00000823, -0.00003352, -0.00004309, -0.00001534, 0.00000524, -0.00000260, -0.00000968, -0.00000668, -0.00000501, },
		{ -0.00000654, -0.00000484, -0.00000465, -0.00000704, -0.00000335, 0.00000447, 0.00000518, 0.00000010, -0.00000295, -0.00000118, -0.00000041, -0.00000400, -0.00000131, 0.00000577, 0.00000152, -0.00000455, 0.00000003, 0.00000192, -0.00000238, -0.00000450, -0.00000542, -0.00000191, 0.00000205, -0.00000396, -0.00000482, 0.00000596, 0.00000653, 0.00000101, 0.00000258, -0.00000574, -0.00001809, -0.00000536, 0.00001455, 0.00000830, -0.00000626, -0.00000213, 0.00000372, -0.00000512, -0.00000711, 0.00000431, 0.00000506, -0.00000091, 0.00000265, 0.00000212, -0.00000464, -0.00000454, -0.00000285, 0.00000193, 0.00000616, -0.00000534, -0.00000701, 0.00000675, -0.00000633, -0.00001474, 0.00001267, 0.00000647, -0.00001456, 0.00002225, 0.00002599, -0.00003555, -0.00002398, 0.00002812, -0.00000121, -0.00001974, 0.00002458, 0.00001203, -0.00003388, -0.00000553, 0.00003364, 0.00000857, -0.00002068, -0.00000688, 0.00001203, 0.00000267, -0.00000906, 0.00000060, 0.00000282, -0.00000616, 0.00000491, 0.00001067, -0.00000710, -0.00000487, 0.00000796, -0.00000368, -0.00000560, 0.00000943, 0.00000404, -0.00000465, 0.00000087, -0.00000100, -0.00000317, 0.00000289, -0.00000023, -0.00000549, -0.00000001, 0.00000389, -0.00000043, -0.00000361, 0.00000148, 0.00000292, -0.00000900, -0.00000918, 0.00000403, -0.00000485, -0.00001284, 0.00000693, 0.00000649, -0.00000892, 0.00000811, 0.00000788, -0.00003070, -0.00003377, -0.00001060, -0.00000544, 0.00002114, 0.00005496, 0.00003572, 0.00001210, 0.00003741, 0.00005974, 0.00003458, -0.00001626, -0.00004753, -0.00003209, -0.00000102, 0.00000273, -0.00000738, -0.00000853, },
		{ -0.00000841, -0.00000817, -0.00000645, -0.00000400, -0.00000622, -0.00000639, 0.00000116, 0.00000564, 0.00000285, -0.00000166, -0.00000291, -0.00000014, -0.00000162, -0.00000465, 0.00000212, 0.00000582, -0.00000202, -0.00000322, 0.00000169, -0.00000023, -0.00000294, -0.00000491, -0.00000586, 0.00000071, 0.00000156, -0.00000632, -0.00000122, 0.00000812, 0.00000373, 0.00000146, 0.00000149, -0.00001173, -0.00001650, 0.00000308, 0.00001548, 0.00000282, -0.00000696, 0.00000086, 0.00000208, -0.00000825, -0.00000430, 0.00000724, 0.00000337, -0.00000081, 0.00000337, -0.00000082, -0.00000644, -0.00000348, 0.00000046, 0.00000454, 0.00000086, -0.00000761, 0.00000053, 0.00000212, -0.00001391, -0.00000068, 0.00001512, -0.00000945, -0.00000058, 0.00003566, -0.00000437, -0.00004485, 0.00000552, 0.00002525, -0.00001854, -0.00000439, 0.00002776, -0.00000913, -0.00002948, 0.00001342, 0.00002979, -0.00000492, -0.00002139, 0.00000175, 0.00001390, -0.00000415, -0.00000818, 0.00000537, -0.00000120, -0.00000504, 0.00001083, 0.00000465, -0.00001014, 0.00000191, 0.00000507, -0.00000742, 0.00000099, 0.00000944, -0.00000061, -0.00000262, 0.00000064, -0.00000266, -0.00000039, 0.00000215, -0.00000284, -0.00000369, 0.00000143, 0.00000311, -0.00000179, -0.00000325, 0.00000389, -0.00000069, -0.00001323, -0.00000346, 0.00000398, -0.00001197, -0.00000609, 0.00001227, -0.00000139, -0.00000511, 0.00001247, -0.00000781, -0.00003642, -0.00002513, -0.00001305, 0.00000108, 0.00004258, 0.00005480, 0.00002403, 0.00002250, 0.00005433, 0.00005636, 0.00001074, -0.00004066, -0.00004700, -0.00001435, 0.00000403, -0.00000295, },
		{ 0.00000108, -0.00000622, -0.00000849, -0.00000865, -0.00000493, -0.00000455, -0.00000775, -0.00000272, 0.00000439, 0.00000436, 0.00000094, -0.00000286, -0.00000218, 0.00000044, -0.00000375, -0.00000327, 0.00000549, 0.00000319, -0.00000391, -0.00000066, 0.00000103, -0.00000171, -0.00000272, -0.00000658, -0.00000462, 0.00000351, -0.00000124, -0.00000667, 0.00000340, 0.00000749, 0.00000181, 0.00000262, -0.00000238, -0.00001599, -0.00001121, 0.00000971, 0.00001360, -0.00000192, -0.00000613, 0.00000332, -0.00000109, -0.00001032, 0.00000037, 0.00000903, 0.00000078, -0.00000010, 0.00000318, -0.00000449, -0.00000695, -0.00000005, 0.00000373, 0.00000239, -0.00000390, -0.00000354, 0.00000283, -0.00000716, -0.00000947, 0.00001170, 0.00000329, -0.00001287, 0.00002198, 0.00002540, -0.00003443, -0.00002889, 0.00002655, 0.00000755, -0.00002237, 0.00001181, 0.00001797, -0.00002253, -0.00001610, 0.00002494, 0.00002048, -0.00001624, -0.00001739, 0.00001160, 0.00001025, -0.00001066, -0.00000237, 0.00000652, -0.00000610, 0.00000069, 0.00001256, -0.00000329, -0.00000752, 0.00000600, -0.00000055, -0.00000602, 0.00000619, 0.00000608, -0.00000226, -0.00000109, -0.00000095, -0.00000183, 0.00000103, 0.00000001, -0.00000321, -0.00000225, 0.00000214, 0.00000253, -0.00000368, -0.00000119, 0.00000502, -0.00000731, -0.00001290, 0.00000290, -0.00000186, -0.00001424, 0.00000412, 0.00001024, -0.00000667, 0.00000235, 0.00000741, -0.00002184, -0.00003362, -0.00002287, -0.00001459, 0.00001919, 0.00005791, 0.00004554, 0.00002084, 0.00003938, 0.00006461, 0.00004090, -0.00002008, -0.00005434, -0.00003224, -0.00000106, },
		{ -0.00001326, 0.00000219, -0.00000324, -0.00000700, -0.00000982, -0.00000767, -0.00000345, -0.00000704, -0.00000642, 0.00000200, 0.00000458, 0.00000270, -0.00000045, -0.00000374, -0.00000063, -0.00000042, -0.00000550, 0.00000042, 0.00000660, -0.00000058, -0.00000310, 0.00000077, -0.00000087, -0.00000144, -0.00000367, -0.00000825, -0.00000100, 0.00000420, -0.00000488, -0.00000401, 0.00000659, 0.00000519, 0.00000184, 0.00000228, -0.00000774, -0.00001677, -0.00000448, 0.00001388, 0.00001021, -0.00000598, -0.00000379, 0.00000479, -0.00000597, -0.00000996, 0.00000658, 0.00000822, -0.00000209, 0.00000171, 0.00000131, -0.00000842, -0.00000437, 0.00000479, 0.00000324, -0.00000229, -0.00000335, 0.00000106, -0.00000265, -0.00001087, 0.00000307, 0.00001128, -0.00001025, 0.00000132, 0.00003387, -0.00000342, -0.00004507, -0.00000052, 0.00002859, -0.00001095, -0.00001389, 0.00001964, 0.00000332, -0.00002544, -0.00000105, 0.00002940, 0.00000778, -0.00002420, -0.00000728, 0.00001833, 0.00000101, -0.00001188, 0.00000494, 0.00000242, -0.00000734, 0.00000797, 0.00000844, -0.00000836, -0.00000182, 0.00000532, -0.00000453, -0.00000143, 0.00000766, 0.00000270, -0.00000206, -0.00000130, -0.00000142, -0.00000027, 0.00000033, -0.00000114, -0.00000275, -0.00000161, 0.00000333, 0.00000088, -0.00000493, 0.00000256, 0.00000216, -0.00001353, -0.00000721, 0.00000480, -0.00000972, -0.00000902, 0.00001132, 0.00000347, -0.00000587, 0.00000650, -0.00000398, -0.00002819, -0.00003021, -0.00002636, -0.00000653, 0.00004223, 0.00006197, 0.00003507, 0.00002761, 0.00005750, 0.00006394, 0.00001186, -0.00004828, -0.00005001, },
		{ -0.00006090, -0.00003145, -0.00000153, -0.00000102, -0.00000477, -0.00000853, -0.00001085, -0.00000466, -0.00000464, -0.00000882, -0.00000169, 0.00000440, 0.00000295, 0.00000195, -0.00000222, -0.00000356, 0.00000088, -0.00000299, -0.00000491, 0.00000467, 0.00000451, -0.00000289, -0.00000093, -0.00000014, -0.00000166, -0.00000064, -0.00000657, -0.00000756, 0.00000331, 0.00000172, -0.00000670, 0.00000032, 0.00000705, 0.00000349, 0.00000269, -0.00000061, -0.00001207, -0.00001454, 0.00000194, 0.00001635, 0.00000535, -0.00000902, 0.00000043, 0.00000375, -0.00001168, -0.00000533, 0.00001217, 0.00000388, -0.00000325, 0.00000396, -0.00000332, -0.00001027, 0.00000207, 0.00000716, -0.00000175, -0.00000397, 0.00000132, -0.00000041, -0.00000970, -0.00000394, 0.00001166, -0.00000080, -0.00001103, 0.00002208, 0.00002360, -0.00003200, -0.00003195, 0.00002200, 0.00001543, -0.00002007, -0.00000149, 0.00001815, -0.00000875, -0.00002104, 0.00001223, 0.00002771, -0.00000736, -0.00002527, 0.00000683, 0.00001712, -0.00000892, -0.00000627, 0.00000839, -0.00000389, -0.00000308, 0.00001198, 0.00000113, -0.00000827, 0.00000267, 0.00000172, -0.00000471, 0.00000276, 0.00000651, 0.00000091, -0.00000220, -0.00000185, -0.00000031, -0.00000011, -0.00000077, -0.00000094, -0.00000306, -0.00000043, 0.00000431, -0.00000224, -0.00000354, 0.00000530, -0.00000478, -0.00001495, 0.00000009, 0.00000039, -0.00001354, 0.00000055, 0.00001179, -0.00000225, -0.00000159, 0.00000352, -0.00001434, -0.00002855, -0.00003242, -0.00002789, 0.00001301, 0.00006110, 0.00005662, 0.00002975, 0.00004258, 0.00007139, 0.00004685, -0.00002550, },
		{ 0.00001164, -0.00005712, -0.00005206, -0.00001165, -0.00000040, -0.00000359, -0.00000530, -0.00001174, -0.00000873, -0.00000258, -0.00000829, -0.00000657, 0.00000335, 0.00000329, 0.00000206, 0.00000123, -0.00000424, -0.00000172, 0.00000067, -0.00000569, -0.00000130, 0.00000671, 0.00000081, -0.00000245, 0.00000001, -0.00000197, -0.00000035, -0.00000176, -0.00000964, -0.00000359, 0.00000536, -0.00000230, -0.00000533, 0.00000367, 0.00000590, 0.00000348, 0.00000213, -0.00000474, -0.00001393, -0.00001113, 0.00000816, 0.00001715, -0.00000154, -0.00000929, 0.00000557, -0.00000158, -0.00001509, 0.00000371, 0.00001351, -0.00000239, -0.00000087, 0.00000396, -0.00000982, -0.00000675, 0.00000951, 0.00000339, -0.00000686, 0.00000035, 0.00000345, -0.00000795, -0.00000927, 0.00000785, 0.00000765, -0.00001188, 0.00000388, 0.00003165, -0.00000323, -0.00004294, -0.00000637, 0.00002841, -0.00000157, -0.00001902, 0.00000783, 0.00001182, -0.00001607, -0.00001274, 0.00002289, 0.00001909, -0.00002156, -0.00001672, 0.00001853, 0.00000775, -0.00001321, 0.00000233, 0.00000565, -0.00000732, 0.00000405, 0.00001036, -0.00000484, -0.00000468, 0.00000373, -0.00000137, -0.00000235, 0.00000475, 0.00000505, -0.00000021, -0.00000312, -0.00000103, 0.00000057, -0.00000137, -0.00000052, -0.00000103, -0.00000372, 0.00000208, 0.00000332, -0.00000495, 0.00000045, 0.00000374, -0.00001210, -0.00001040, 0.00000360, -0.00000689, -0.00001033, 0.00000843, 0.00000737, -0.00000368, 0.00000056, -0.00000398, -0.00001909, -0.00002980, -0.00003911, -0.00001954, 0.00003864, 0.00007037, 0.00004724, 0.00003299, 0.00006245, 0.00007292, },
		{ 0.00008027, 0.00005321, -0.00003284, -0.00006782, -0.00002940, -0.00000212, -0.00000381, -0.00000304, -0.00000850, -0.00001331, -0.00000400, -0.00000465, -0.00001042, -0.00000046, 0.00000481, 0.00000101, 0.00000272, -0.00000092, -0.00000494, 0.00000067, -0.00000198, -0.00000594, 0.00000345, 0.00000524, -0.00000157, -0.00000075, -0.00000141, -0.00000210, 0.00000129, -0.00000562, -0.00000985, 0.00000164, 0.00000377, -0.00000486, -0.00000214, 0.00000472, 0.00000542, 0.00000377, -0.00000062, -0.00000773, -0.00001430, -0.00000722, 0.00001508, 0.00001449, -0.00000934, -0.00000479, 0.00000818, -0.00001060, -0.00001211, 0.00001368, 0.00000830, -0.00000634, 0.00000397, -0.00000101, -0.00001396, 0.00000286, 0.00001200, -0.00000536, -0.00000588, 0.00000687, -0.00000220, -0.00001376, 0.00000100, 0.00001285, -0.00000538, -0.00000949, 0.00002279, 0.00002071, -0.00002919, -0.00003254, 0.00001563, 0.00002079, -0.00001371, -0.00001274, 0.00001273, 0.00000439, -0.00001964, -0.00000144, 0.00002887, 0.00000359, -0.00002867, -0.00000092, 0.00002133, -0.00000419, -0.00000950, 0.00000789, -0.00000059, -0.00000527, 0.00000933, 0.00000491, -0.00000690, -0.00000104, 0.00000242, -0.00000219, 0.00000007, 0.00000534, 0.00000406, -0.00000205, -0.00000330, 0.00000083, -0.00000037, -0.00000210, 0.00000065, -0.00000256, -0.00000293, 0.00000457, -0.00000002, -0.00000477, 0.00000406, -0.00000249, -0.00001509, -0.00000338, 0.00000112, -0.00001132, -0.00000253, 0.00001097, 0.00000276, -0.00000273, -0.00000216, -0.00001005, -0.00002040, -0.00003686, -0.00004302, 0.00000081, 0.00006250, 0.00006968, 0.00003949, 0.00004641, },
		{ 0.00003891, 0.00006858, 0.00008383, 0.00001064, -0.00006779, -0.00005350, -0.00000901, -0.00000363, -0.00000430, -0.00000320, -0.00001386, -0.00000987, -0.00000124, -0.00000978, -0.00000716, 0.00000516, 0.00000214, 0.00000076, 0.00000302, -0.00000407, -0.00000317, 0.00000123, -0.00000528, -0.00000265, 0.00000611, 0.00000194, -0.00000127, -0.00000055, -0.00000335, 0.00000033, 0.00000025, -0.00000989, -0.00000608, 0.00000465, 0.00000029, -0.00000447, 0.00000016, 0.00000485, 0.00000635, 0.00000207, -0.00000373, -0.00000887, -0.00001501, -0.00000113, 0.00002163, 0.00000636, -0.00001412, 0.00000369, 0.00000418, -0.00001862, -0.00000132, 0.00001844, -0.00000128, -0.00000426, 0.00000664, -0.00001010, -0.00001044, 0.00001377, 0.00000536, -0.00001217, 0.00000265, 0.00000753, -0.00001288, -0.00000950, 0.00001306, 0.00000488, -0.00001434, 0.00000675, 0.00002949, -0.00000421, -0.00003915, -0.00001075, 0.00002499, 0.00000752, -0.00001923, -0.00000517, 0.00001484, -0.00000374, -0.00001931, 0.00001214, 0.00002641, -0.00001444, -0.00002411, 0.00001449, 0.00001413, -0.00001142, -0.00000132, 0.00000731, -0.00000524, 0.00000039, 0.00001002, -0.00000062, -0.00000586, 0.00000078, 0.00000109, -0.00000165, 0.00000155, 0.00000594, 0.00000245, -0.00000419, -0.00000152, 0.00000165, -0.00000246, -0.00000091, 0.00000076, -0.00000452, 0.00000001, 0.00000476, -0.00000356, -0.00000148, 0.00000371, -0.00000983, -0.00001220, 0.00000092, -0.00000458, -0.00000987, 0.00000475, 0.00000932, 0.00000046, -0.00000357, -0.00000713, -0.00001172, -0.00002446, -0.00004806, -0.00003681, 0.00002923, 0.00007872, 0.00006173, },
		{ 0.00008439, 0.00005140, 0.00005070, 0.00009083, 0.00006073, -0.00004190, -0.00007585, -0.00002611, -0.00000230, -0.00000746, -0.00000124, -0.00000820, -0.00001612, -0.00000314, -0.00000429, -0.00001252, 0.00000058, 0.00000581, -0.00000110, 0.00000286, 0.00000086, -0.00000581, -0.00000024, -0.00000116, -0.00000612, 0.00000213, 0.00000516, -0.00000004, -0.00000004, -0.00000279, -0.00000281, 0.00000288, -0.00000429, -0.00001093, -0.00000070, 0.00000394, -0.00000182, -0.00000279, 0.00000062, 0.00000649, 0.00000657, -0.00000188, -0.00000456, -0.00001035, -0.00001565, 0.00000910, 0.00002368, -0.00000602, -0.00001143, 0.00001096, -0.00000700, -0.00001890, 0.00001297, 0.00001365, -0.00000876, 0.00000292, 0.00000225, -0.00001728, 0.00000191, 0.00001761, -0.00000762, -0.00000952, 0.00001219, -0.00000205, -0.00001872, 0.00000458, 0.00001520, -0.00000982, -0.00000863, 0.00002415, 0.00001715, -0.00002687, -0.00003067, 0.00000892, 0.00002273, -0.00000500, -0.00002004, 0.00000319, 0.00001418, -0.00001275, -0.00001296, 0.00002412, 0.00001378, -0.00002698, -0.00000952, 0.00002166, 0.00000227, -0.00001065, 0.00000531, 0.00000252, -0.00000534, 0.00000559, 0.00000704, -0.00000386, -0.00000406, 0.00000141, 0.00000068, -0.00000127, 0.00000301, 0.00000639, -0.00000066, -0.00000472, 0.00000121, 0.00000008, -0.00000339, 0.00000124, -0.00000127, -0.00000465, 0.00000345, 0.00000208, -0.00000454, 0.00000205, -0.00000124, -0.00001368, -0.00000643, 0.00000018, -0.00000855, -0.00000430, 0.00000841, 0.00000688, -0.00000117, -0.00000757, -0.00000956, -0.00001187, -0.00003522, -0.00005661, -0.00001773, 0.00005923, },
		{ 0.00001224, 0.00008441, 0.00007915, 0.00004655, 0.00007532, 0.00009655, 0.00000970, -0.00008050, -0.00005506, -0.00000503, -0.00000711, -0.00000571, -0.00000075, -0.00001593, -0.00001132, 0.00000052, -0.00001113, -0.00000837, 0.00000710, 0.00000156, -0.00000107, 0.00000438, -0.00000310, -0.00000450, 0.00000090, -0.00000454, -0.00000320, 0.00000485, 0.00000250, 0.00000053, -0.00000089, -0.00000499, 0.00000084, 0.00000194, -0.00000939, -0.00000750, 0.00000240, 0.00000160, -0.00000137, -0.00000266, 0.00000141, 0.00000973, 0.00000328, -0.00000502, -0.00000318, -0.00001482, -0.00001240, 0.00002190, 0.00001679, -0.00001666, -0.00000092, 0.00001013, -0.00001973, -0.00000797, 0.00002211, 0.00000138, -0.00000804, 0.00000876, -0.00000896, -0.00001492, 0.00001673, 0.00000907, -0.00001727, 0.00000304, 0.00001271, -0.00001654, -0.00001153, 0.00001794, 0.00000339, -0.00001734, 0.00000946, 0.00002782, -0.00000634, -0.00003464, -0.00001286, 0.00001931, 0.00001451, -0.00001510, -0.00001677, 0.00001223, 0.00000859, -0.00001982, -0.00000004, 0.00002823, -0.00000491, -0.00002762, 0.00000734, 0.00001825, -0.00000675, -0.00000450, 0.00000683, -0.00000203, -0.00000191, 0.00000772, 0.00000311, -0.00000510, -0.00000261, 0.00000217, 0.00000023, -0.00000107, 0.00000529, 0.00000516, -0.00000410, -0.00000259, 0.00000244, -0.00000272, -0.00000187, 0.00000198, -0.00000409, -0.00000209, 0.00000484, -0.00000143, -0.00000247, 0.00000235, -0.00000769, -0.00001239, -0.00000231, -0.00000351, -0.00000816, 0.00000146, 0.00000918, 0.00000502, -0.00000499, -0.00001173, -0.00000791, -0.00001628, -0.00005080, -0.00005554, },
		{ -0.00006519, -0.00004102, 0.00004839, 0.00009891, 0.00006691, 0.00005618, 0.00010227, 0.00006972, -0.00005192, -0.00008561, -0.00002209, -0.00000146, -0.00001180, 0.00000019, -0.00000742, -0.00001910, -0.00000244, -0.00000340, -0.00001478, 0.00000100, 0.00000727, -0.00000285, 0.00000216, 0.00000261, -0.00000574, -0.00000159, -0.00000111, -0.00000543, 0.00000118, 0.00000403, 0.00000148, 0.00000137, -0.00000435, -0.00000366, 0.00000431, -0.00000324, -0.00001077, -0.00000254, 0.00000184, 0.00000115, -0.00000076, -0.00000440, 0.00000549, 0.00001113, -0.00000318, -0.00000380, -0.00000359, -0.00002123, -0.00000117, 0.00003075, 0.00000122, -0.00001803, 0.00001112, -0.00000146, -0.00002454, 0.00000985, 0.00001919, -0.00000994, 0.00000075, 0.00000597, -0.00001953, -0.00000076, 0.00002309, -0.00000794, -0.00001429, 0.00001635, 0.00000005, -0.00002370, 0.00000628, 0.00001830, -0.00001350, -0.00000860, 0.00002591, 0.00001346, -0.00002559, -0.00002689, 0.00000324, 0.00002127, 0.00000402, -0.00002254, -0.00000816, 0.00001887, -0.00000249, -0.00001994, 0.00001505, 0.00002069, -0.00002088, -0.00001666, 0.00001793, 0.00000864, -0.00000895, 0.00000166, 0.00000429, -0.00000350, 0.00000198, 0.00000712, -0.00000010, -0.00000560, -0.00000090, 0.00000300, -0.00000111, 0.00000032, 0.00000738, 0.00000154, -0.00000549, 0.00000077, 0.00000083, -0.00000417, 0.00000088, 0.00000015, -0.00000521, 0.00000155, 0.00000335, -0.00000307, 0.00000016, -0.00000133, -0.00001155, -0.00000830, -0.00000195, -0.00000625, -0.00000462, 0.00000522, 0.00000918, 0.00000206, -0.00001097, -0.00001210, -0.00000566, -0.00002831, },
		{ -0.00000818, -0.00004649, -0.00007186, -0.00001241, 0.00008390, 0.00009885, 0.00005780, 0.00008277, 0.00011029, 0.00000956, -0.00009481, -0.00005765, 0.00000018, -0.00001012, -0.00000801, 0.00000174, -0.00001757, -0.00001314, 0.00000235, -0.00001194, -0.00001008, 0.00000860, 0.00000176, -0.00000295, 0.00000478, -0.00000156, -0.00000506, -0.00000043, -0.00000404, -0.00000257, 0.00000321, 0.00000207, 0.00000296, -0.00000085, -0.00000704, 0.00000146, 0.00000325, -0.00000889, -0.00000730, -0.00000042, 0.00000050, 0.00000325, -0.00000337, -0.00000439, 0.00001237, 0.00000667, -0.00000806, 0.00000132, -0.00001024, -0.00002340, 0.00001700, 0.00002798, -0.00001559, -0.00000756, 0.00001493, -0.00001810, -0.00001529, 0.00002375, 0.00000538, -0.00001160, 0.00000984, -0.00000645, -0.00001941, 0.00001782, 0.00001400, -0.00002123, 0.00000147, 0.00001809, -0.00001832, -0.00001487, 0.00002168, 0.00000326, -0.00002039, 0.00001157, 0.00002680, -0.00000931, -0.00003035, -0.00001249, 0.00001277, 0.00001832, -0.00000811, -0.00002491, 0.00000516, 0.00001819, -0.00001483, -0.00001069, 0.00002449, 0.00000442, -0.00002647, -0.00000101, 0.00001886, -0.00000029, -0.00000586, 0.00000440, 0.00000108, -0.00000230, 0.00000436, 0.00000540, -0.00000278, -0.00000546, 0.00000167, 0.00000254, -0.00000247, 0.00000350, 0.00000718, -0.00000286, -0.00000369, 0.00000258, -0.00000230, -0.00000288, 0.00000231, -0.00000290, -0.00000353, 0.00000372, 0.00000065, -0.00000224, 0.00000036, -0.00000643, -0.00001131, -0.00000511, -0.00000380, -0.00000610, -0.00000068, 0.00000757, 0.00000856, -0.00000383, -0.00001570, -0.00000817, },
		{ -0.00001583, -0.00000360, -0.00001853, -0.00006619, -0.00006565, 0.00002807, 0.00010983, 0.00008671, 0.00006467, 0.00011395, 0.00007970, -0.00006187, -0.00009716, -0.00001829, 0.00000075, -0.00001627, 0.00000077, -0.00000623, -0.00002185, -0.00000225, -0.00000212, -0.00001665, 0.00000060, 0.00000873, -0.00000367, 0.00000081, 0.00000362, -0.00000451, -0.00000276, -0.00000233, -0.00000415, 0.00000119, 0.00000180, 0.00000261, 0.00000374, -0.00000612, -0.00000478, 0.00000597, -0.00000289, -0.00000995, -0.00000278, -0.00000212, 0.00000251, 0.00000411, -0.00000887, 0.00000157, 0.00001672, -0.00000312, -0.00000595, 0.00000429, -0.00002237, -0.00001422, 0.00003370, 0.00001166, -0.00002286, 0.00000835, 0.00000496, -0.00002805, 0.00000462, 0.00002401, -0.00000955, -0.00000227, 0.00000952, -0.00002022, -0.00000472, 0.00002750, -0.00000623, -0.00001929, 0.00001869, 0.00000365, -0.00002778, 0.00000598, 0.00002145, -0.00001591, -0.00000929, 0.00002767, 0.00001017, -0.00002553, -0.00002211, -0.00000042, 0.00001720, 0.00001147, -0.00002052, -0.00001878, 0.00001803, 0.00000836, -0.00002127, 0.00000417, 0.00002276, -0.00001218, -0.00002051, 0.00001107, 0.00001311, -0.00000450, -0.00000166, 0.00000410, -0.00000056, -0.00000043, 0.00000538, 0.00000328, -0.00000537, -0.00000371, 0.00000409, 0.00000022, -0.00000192, 0.00000692, 0.00000387, -0.00000523, -0.00000023, 0.00000139, -0.00000423, -0.00000006, 0.00000112, -0.00000467, -0.00000041, 0.00000347, -0.00000096, -0.00000093, -0.00000249, -0.00000960, -0.00000883, -0.00000439, -0.00000506, -0.00000400, 0.00000251, 0.00000954, 0.00000544, -0.00001150, },
		{ -0.00000122, -0.00001728, -0.00001154, -0.00000306, -0.00003623, -0.00008185, -0.00004262, 0.00007373, 0.00011833, 0.00007455, 0.00009213, 0.00012388, 0.00001032, -0.00010968, -0.00006192, 0.00000592, -0.00001192, -0.00001102, 0.00000365, -0.00001849, -0.00001509, 0.00000371, -0.00001200, -0.00001179, 0.00000912, 0.00000258, -0.00000413, 0.00000392, -0.00000010, -0.00000420, -0.00000248, -0.00000451, -0.00000060, 0.00000175, 0.00000025, 0.00000588, -0.00000009, -0.00000975, 0.00000230, 0.00000453, -0.00000918, -0.00000553, -0.00000241, -0.00000346, 0.00000793, -0.00000087, -0.00001168, 0.00001281, 0.00001247, -0.00001156, 0.00000310, -0.00000207, -0.00003176, 0.00000709, 0.00003765, -0.00001035, -0.00001487, 0.00001745, -0.00001405, -0.00002213, 0.00002296, 0.00001014, -0.00001431, 0.00000962, -0.00000296, -0.00002310, 0.00001685, 0.00001929, -0.00002331, -0.00000165, 0.00002271, -0.00001801, -0.00001875, 0.00002369, 0.00000423, -0.00002291, 0.00001283, 0.00002635, -0.00001257, -0.00002699, -0.00001005, 0.00000681, 0.00001869, -0.00000021, -0.00002852, -0.00000427, 0.00002319, -0.00000620, -0.00001743, 0.00001649, 0.00001116, -0.00002114, -0.00000839, 0.00001566, 0.00000627, -0.00000458, 0.00000092, 0.00000300, -0.00000091, 0.00000108, 0.00000583, 0.00000027, -0.00000700, -0.00000009, 0.00000444, -0.00000249, 0.00000126, 0.00000801, -0.00000081, -0.00000425, 0.00000199, -0.00000157, -0.00000349, 0.00000178, -0.00000154, -0.00000398, 0.00000193, 0.00000201, -0.00000099, -0.00000145, -0.00000628, -0.00000975, -0.00000683, -0.00000495, -0.00000463, -0.00000159, 0.00000557, 0.00001020, },
		{ 0.00000864, 0.00000754, -0.00000942, -0.00001856, -0.00000611, -0.00000920, -0.00005887, -0.00008704, -0.00000164, 0.00011277, 0.00010989, 0.00007880, 0.00012607, 0.00008938, -0.00007094, -0.00010982, -0.00001584, 0.00000417, -0.00001997, 0.00000018, -0.00000507, -0.00002378, -0.00000267, -0.00000097, -0.00001754, -0.00000045, 0.00000941, -0.00000325, -0.00000056, 0.00000315, -0.00000237, -0.00000290, -0.00000507, -0.00000292, 0.00000277, -0.00000134, 0.00000278, 0.00000729, -0.00000793, -0.00000651, 0.00000828, -0.00000321, -0.00000945, -0.00000077, -0.00000679, 0.00000092, 0.00001107, -0.00001112, -0.00000542, 0.00002195, -0.00000051, -0.00001056, 0.00001123, -0.00001845, -0.00002773, 0.00003123, 0.00002366, -0.00002442, 0.00000306, 0.00001093, -0.00002886, -0.00000195, 0.00002725, -0.00000755, -0.00000564, 0.00001226, -0.00001923, -0.00000923, 0.00003009, -0.00000284, -0.00002355, 0.00001895, 0.00000797, -0.00003025, 0.00000399, 0.00002393, -0.00001687, -0.00001035, 0.00002902, 0.00000757, -0.00002647, -0.00001729, -0.00000169, 0.00001179, 0.00001609, -0.00001523, -0.00002656, 0.00001256, 0.00001715, -0.00001727, -0.00000572, 0.00001975, -0.00000325, -0.00002019, 0.00000279, 0.00001442, 0.00000173, -0.00000334, 0.00000208, 0.00000233, -0.00000109, 0.00000264, 0.00000540, -0.00000368, -0.00000611, 0.00000375, 0.00000203, -0.00000312, 0.00000535, 0.00000566, -0.00000392, -0.00000130, 0.00000141, -0.00000368, -0.00000110, 0.00000132, -0.00000345, -0.00000181, 0.00000258, 0.00000102, -0.00000097, -0.00000406, -0.00000848, -0.00000839, -0.00000624, -0.00000501, -0.00000331, 0.00000095, },
		{ -0.00000180, 0.00000425, 0.00000993, 0.00000128, -0.00001569, -0.00001599, -0.00000282, -0.00002289, -0.00008268, -0.00007423, 0.00005165, 0.00013331, 0.00009780, 0.00010597, 0.00013638, 0.00001113, -0.00012373, -0.00006788, 0.00001094, -0.00001198, -0.00001404, 0.00000426, -0.00001869, -0.00001656, 0.00000407, -0.00001152, -0.00001272, 0.00000830, 0.00000337, -0.00000384, 0.00000200, 0.00000035, -0.00000157, -0.00000454, -0.00000669, 0.00000252, 0.00000123, -0.00000325, 0.00000888, 0.00000185, -0.00001331, 0.00000314, 0.00000640, -0.00001078, -0.00000295, -0.00000225, -0.00000989, 0.00001090, 0.00000518, -0.00001886, 0.00000990, 0.00002002, -0.00001387, 0.00000138, 0.00000798, -0.00003552, -0.00000650, 0.00004358, -0.00000137, -0.00002114, 0.00001708, -0.00000843, -0.00002738, 0.00001977, 0.00001484, -0.00001566, 0.00000817, 0.00000088, -0.00002532, 0.00001410, 0.00002395, -0.00002323, -0.00000551, 0.00002581, -0.00001590, -0.00002229, 0.00002373, 0.00000577, -0.00002443, 0.00001320, 0.00002619, -0.00001560, -0.00002489, -0.00000632, 0.00000250, 0.00001620, 0.00000676, -0.00002760, -0.00001366, 0.00002295, 0.00000352, -0.00001905, 0.00000650, 0.00001372, -0.00001317, -0.00001302, 0.00000938, 0.00001122, -0.00000069, -0.00000231, 0.00000311, 0.00000152, -0.00000113, 0.00000458, 0.00000303, -0.00000696, -0.00000237, 0.00000528, -0.00000140, -0.00000066, 0.00000752, 0.00000143, -0.00000393, 0.00000091, -0.00000095, -0.00000351, 0.00000073, -0.00000057, -0.00000351, 0.00000014, 0.00000238, 0.00000069, -0.00000248, -0.00000693, -0.00000855, -0.00000735, -0.00000611, -0.00000429, },
		{ -0.00000558, -0.00000343, 0.00000055, 0.00000758, 0.00000755, -0.00000597, -0.00001849, -0.00001198, -0.00000352, -0.00004468, -0.00010062, -0.00003814, 0.00010341, 0.00013340, 0.00010125, 0.00014027, 0.00009701, -0.00007909, -0.00012213, -0.00001559, 0.00000789, -0.00002184, -0.00000143, -0.00000477, -0.00002439, -0.00000340, -0.00000075, -0.00001714, -0.00000148, 0.00000850, -0.00000175, -0.00000096, 0.00000074, -0.00000006, -0.00000111, -0.00000914, -0.00000267, 0.00000630, -0.00000480, 0.00000131, 0.00001203, -0.00000931, -0.00000936, 0.00001144, -0.00000367, -0.00001030, 0.00000342, -0.00001051, -0.00000426, 0.00001856, -0.00000975, -0.00001470, 0.00002513, 0.00000514, -0.00001630, 0.00001538, -0.00001007, -0.00003912, 0.00002319, 0.00003497, -0.00002185, -0.00000359, 0.00001523, -0.00002699, -0.00000879, 0.00002832, -0.00000434, -0.00000871, 0.00001377, -0.00001677, -0.00001342, 0.00003051, 0.00000147, -0.00002623, 0.00001739, 0.00001208, -0.00003079, 0.00000097, 0.00002511, -0.00001651, -0.00001130, 0.00002964, 0.00000573, -0.00002798, -0.00001322, -0.00000077, 0.00000643, 0.00001742, -0.00000841, -0.00003025, 0.00000428, 0.00002193, -0.00000956, -0.00001232, 0.00001272, 0.00000366, -0.00001600, -0.00000490, 0.00001215, 0.00000821, -0.00000255, -0.00000096, 0.00000412, -0.00000008, -0.00000006, 0.00000582, -0.00000128, -0.00000739, 0.00000224, 0.00000355, -0.00000309, 0.00000330, 0.00000641, -0.00000188, -0.00000192, 0.00000084, -0.00000286, -0.00000183, 0.00000082, -0.00000214, -0.00000235, 0.00000122, 0.00000229, -0.00000023, -0.00000526, -0.00000834, -0.00000776, -0.00000693, },
		{ -0.00000714, -0.00000643, -0.00000502, -0.00000220, 0.00000420, 0.00000851, 0.00000221, -0.00001135, -0.00001911, -0.00000776, -0.00001029, -0.00007360, -0.00010175, 0.00001773, 0.00013827, 0.00012665, 0.00012787, 0.00014793, 0.00001008, -0.00013587, -0.00007458, 0.00001374, -0.00001044, -0.00001587, 0.00000303, -0.00001871, -0.00001666, 0.00000317, -0.00001122, -0.00001201, 0.00000634, 0.00000311, -0.00000156, -0.00000022, -0.00000124, 0.00000266, -0.00000553, -0.00001106, 0.00000610, 0.00000254, -0.00000841, 0.00001125, 0.00000549, -0.00001759, 0.00000341, 0.00000954, -0.00001364, -0.00000080, 0.00000093, -0.00001748, 0.00001050, 0.00001415, -0.00002403, 0.00000321, 0.00002790, -0.00001351, -0.00000358, 0.00001766, -0.00003375, -0.00002149, 0.00004421, 0.00000995, -0.00002476, 0.00001398, -0.00000247, -0.00003028, 0.00001473, 0.00001862, -0.00001542, 0.00000586, 0.00000438, -0.00002576, 0.00001022, 0.00002714, -0.00002115, -0.00000918, 0.00002699, -0.00001264, -0.00002473, 0.00002198, 0.00000719, -0.00002469, 0.00001291, 0.00002596, -0.00001800, -0.00002399, -0.00000222, 0.00000038, 0.00001193, 0.00001141, -0.00002314, -0.00002090, 0.00001810, 0.00001177, -0.00001570, -0.00000291, 0.00001175, -0.00000469, -0.00001395, 0.00000156, 0.00001328, 0.00000497, -0.00000402, 0.00000147, 0.00000392, -0.00000172, 0.00000240, 0.00000469, -0.00000559, -0.00000436, 0.00000487, 0.00000018, -0.00000173, 0.00000602, 0.00000321, -0.00000270, -0.00000020, -0.00000077, -0.00000305, -0.00000037, -0.00000023, -0.00000253, -0.00000108, 0.00000194, 0.00000207, -0.00000256, -0.00000771, -0.00000827, },
		{ -0.00000881, -0.00000774, -0.00000648, -0.00000587, -0.00000480, 0.00000070, 0.00000742, 0.00000557, -0.00000288, -0.00001488, -0.00001888, -0.00000356, -0.00002704, -0.00010305, -0.00007641, 0.00007883, 0.00015192, 0.00013364, 0.00015989, 0.00010109, -0.00008752, -0.00013221, -0.00001758, 0.00001031, -0.00002110, -0.00000323, -0.00000637, -0.00002355, -0.00000360, -0.00000230, -0.00001566, -0.00000142, 0.00000547, 0.00000005, 0.00000069, -0.00000352, 0.00000121, 0.00000321, -0.00001377, -0.00000442, 0.00001176, -0.00000767, -0.00000238, 0.00001763, -0.00000953, -0.00001376, 0.00001517, -0.00000336, -0.00001321, 0.00000890, -0.00001164, -0.00001275, 0.00002455, -0.00000410, -0.00002465, 0.00002478, 0.00001338, -0.00002136, 0.00001562, 0.00000114, -0.00004613, 0.00001062, 0.00004326, -0.00001527, -0.00001003, 0.00001704, -0.00002306, -0.00001480, 0.00002703, -0.00000055, -0.00001088, 0.00001390, -0.00001339, -0.00001653, 0.00002885, 0.00000575, -0.00002689, 0.00001464, 0.00001517, -0.00002949, -0.00000227, 0.00002470, -0.00001525, -0.00001173, 0.00002943, 0.00000451, -0.00002955, -0.00001036, 0.00000172, 0.00000224, 0.00001579, -0.00000188, -0.00002962, -0.00000454, 0.00002191, -0.00000044, -0.00001437, 0.00000368, 0.00000701, -0.00000924, -0.00001028, 0.00000689, 0.00001329, 0.00000076, -0.00000382, 0.00000420, 0.00000191, -0.00000177, 0.00000474, 0.00000091, -0.00000728, 0.00000022, 0.00000421, -0.00000209, 0.00000146, 0.00000602, 0.00000032, -0.00000178, -0.00000009, -0.00000218, -0.00000207, -0.00000006, -0.00000121, -0.00000214, 0.00000001, 0.00000264, 0.00000071, -0.00000557, },
		{ -0.00000199, -0.00000789, -0.00000896, -0.00000700, -0.00000545, -0.00000619, -0.00000365, 0.00000532, 0.00000714, 0.00000081, -0.00000570, -0.00001896, -0.00001642, -0.00000214, -0.00005639, -0.00011963, -0.00002501, 0.00012771, 0.00015755, 0.00016160, 0.00016041, 0.00000464, -0.00014598, -0.00008008, 0.00001313, -0.00000837, -0.00001505, -0.00000005, -0.00001962, -0.00001458, 0.00000133, -0.00001221, -0.00000912, 0.00000405, 0.00000064, 0.00000265, -0.00000151, -0.00000561, 0.00000768, -0.00000425, -0.00001763, 0.00000905, 0.00000635, -0.00001472, 0.00001204, 0.00001111, -0.00002207, 0.00000230, 0.00001431, -0.00001708, -0.00000046, 0.00000715, -0.00002435, 0.00000578, 0.00002445, -0.00002543, -0.00000671, 0.00003424, -0.00000964, -0.00001056, 0.00002484, -0.00002671, -0.00003520, 0.00003899, 0.00002149, -0.00002466, 0.00000900, 0.00000261, -0.00003062, 0.00000872, 0.00002077, -0.00001375, 0.00000325, 0.00000700, -0.00002446, 0.00000603, 0.00002834, -0.00001769, -0.00001179, 0.00002632, -0.00000907, -0.00002566, 0.00001896, 0.00000794, -0.00002375, 0.00001231, 0.00002539, -0.00001959, -0.00002399, 0.00000144, 0.00000046, 0.00000718, 0.00001313, -0.00001671, -0.00002462, 0.00001025, 0.00001662, -0.00000873, -0.00000954, 0.00000610, 0.00000220, -0.00001133, -0.00000597, 0.00001197, 0.00001099, -0.00000341, -0.00000117, 0.00000531, -0.00000075, 0.00000024, 0.00000486, -0.00000357, -0.00000542, 0.00000344, 0.00000156, -0.00000177, 0.00000408, 0.00000410, -0.00000088, -0.00000087, -0.00000107, -0.00000244, -0.00000117, -0.00000047, -0.00000160, -0.00000149, 0.00000123, 0.00000264, },
		{ 0.00000232, 0.00000113, -0.00000491, -0.00000917, -0.00000880, -0.00000545, -0.00000507, -0.00000724, 0.00000041, 0.00000872, 0.00000254, -0.00000170, -0.00000833, -0.00002403, -0.00000960, -0.00001049, -0.00009342, -0.00010980, 0.00003880, 0.00015862, 0.00017539, 0.00018964, 0.00010136, -0.00009880, -0.00013847, -0.00002069, 0.00000942, -0.00001772, -0.00000373, -0.00001080, -0.00002179, -0.00000204, -0.00000612, -0.00001396, 0.00000096, 0.00000037, 0.00000084, 0.00000523, -0.00000881, 0.00000008, 0.00001011, -0.00001770, -0.00000901, 0.00001854, -0.00000880, -0.00000854, 0.00002330, -0.00000777, -0.00001976, 0.00001875, -0.00000129, -0.00001824, 0.00001416, -0.00000901, -0.00002322, 0.00002705, 0.00000541, -0.00003316, 0.00002011, 0.00002286, -0.00002398, 0.00001189, 0.00001287, -0.00004746, -0.00000436, 0.00004672, -0.00000578, -0.00001464, 0.00001623, -0.00001808, -0.00001911, 0.00002367, 0.00000302, -0.00001179, 0.00001286, -0.00000976, -0.00001807, 0.00002559, 0.00000912, -0.00002555, 0.00001156, 0.00001675, -0.00002687, -0.00000500, 0.00002275, -0.00001366, -0.00001138, 0.00002851, 0.00000364, -0.00003077, -0.00000880, 0.00000495, -0.00000011, 0.00001213, 0.00000295, -0.00002544, -0.00001184, 0.00001751, 0.00000765, -0.00001187, -0.00000501, 0.00000633, -0.00000181, -0.00001238, -0.00000002, 0.00001573, 0.00000584, -0.00000529, 0.00000261, 0.00000389, -0.00000200, 0.00000282, 0.00000214, -0.00000605, -0.00000158, 0.00000383, -0.00000065, 0.00000032, 0.00000477, 0.00000208, -0.00000086, -0.00000095, -0.00000191, -0.00000197, -0.00000088, -0.00000088, -0.00000161, -0.00000052, },
		{ -0.00000127, 0.00000089, 0.00000226, -0.00000144, -0.00000700, -0.00001014, -0.00000774, -0.00000330, -0.00000683, -0.00000651, 0.00000685, 0.00000696, -0.00000268, -0.00000074, -0.00001478, -0.00002601, -0.00000088, -0.00003510, -0.00012378, -0.00007064, 0.00009762, 0.00018422, 0.00020985, 0.00017787, -0.00000759, -0.00015545, -0.00008184, 0.00000880, -0.00000776, -0.00001036, -0.00000413, -0.00002290, -0.00000989, -0.00000041, -0.00001573, -0.00000414, 0.00000282, -0.00000491, 0.00000797, -0.00000038, -0.00001290, 0.00001211, 0.00000027, -0.00002569, 0.00001005, 0.00001297, -0.00002119, 0.00001032, 0.00001856, -0.00002584, -0.00000093, 0.00002053, -0.00001993, -0.00000292, 0.00001548, -0.00002849, -0.00000320, 0.00003387, -0.00002203, -0.00001831, 0.00003724, -0.00000240, -0.00001770, 0.00002803, -0.00001585, -0.00004516, 0.00002865, 0.00003095, -0.00002065, 0.00000351, 0.00000589, -0.00002870, 0.00000274, 0.00002093, -0.00001109, 0.00000093, 0.00000841, -0.00002183, 0.00000236, 0.00002747, -0.00001371, -0.00001279, 0.00002428, -0.00000599, -0.00002504, 0.00001539, 0.00000765, -0.00002193, 0.00001176, 0.00002432, -0.00002040, -0.00002444, 0.00000411, 0.00000224, 0.00000307, 0.00001203, -0.00001000, -0.00002443, 0.00000144, 0.00001716, -0.00000024, -0.00001213, -0.00000147, 0.00000603, -0.00000623, -0.00001156, 0.00000775, 0.00001585, -0.00000039, -0.00000371, 0.00000510, 0.00000112, -0.00000101, 0.00000372, -0.00000175, -0.00000534, 0.00000159, 0.00000223, -0.00000101, 0.00000233, 0.00000401, 0.00000099, -0.00000083, -0.00000160, -0.00000206, -0.00000153, -0.00000092, -0.00000114, },
		{ -0.00000103, -0.00000130, -0.00000022, 0.00000191, 0.00000058, -0.00000368, -0.00000867, -0.00001089, -0.00000474, -0.00000277, -0.00000996, -0.00000120, 0.00001122, -0.00000015, -0.00000319, -0.00000059, -0.00002522, -0.00001986, 0.00000053, -0.00007371, -0.00013152, -0.00001344, 0.00014639, 0.00022278, 0.00023474, 0.00009964, -0.00011647, -0.00014038, -0.00002248, 0.00000344, -0.00001275, -0.00000097, -0.00001840, -0.00002042, 0.00000258, -0.00001201, -0.00001351, 0.00000659, -0.00000593, -0.00000094, 0.00001285, -0.00001366, -0.00000459, 0.00001886, -0.00001941, -0.00001674, 0.00002549, -0.00000710, -0.00001691, 0.00002789, -0.00000337, -0.00002693, 0.00002108, 0.00000321, -0.00002467, 0.00001749, -0.00000237, -0.00003359, 0.00002471, 0.00001732, -0.00003816, 0.00001135, 0.00003165, -0.00002295, 0.00000521, 0.00002277, -0.00004298, -0.00001912, 0.00004449, 0.00000475, -0.00001623, 0.00001338, -0.00001317, -0.00002130, 0.00001888, 0.00000565, -0.00001137, 0.00001105, -0.00000652, -0.00001792, 0.00002151, 0.00001096, -0.00002263, 0.00000899, 0.00001673, -0.00002365, -0.00000673, 0.00001967, -0.00001226, -0.00001024, 0.00002714, 0.00000288, -0.00003138, -0.00000834, 0.00000810, -0.00000047, 0.00000757, 0.00000534, -0.00001905, -0.00001622, 0.00001009, 0.00001284, -0.00000597, -0.00001130, 0.00000231, 0.00000438, -0.00001120, -0.00000689, 0.00001502, 0.00001141, -0.00000460, 0.00000004, 0.00000494, -0.00000083, 0.00000098, 0.00000209, -0.00000432, -0.00000260, 0.00000264, 0.00000059, 0.00000004, 0.00000320, 0.00000301, 0.00000050, -0.00000130, -0.00000205, -0.00000185, -0.00000132, },
		{ -0.00000163, -0.00000117, -0.00000133, -0.00000089, 0.00000137, 0.00000128, -0.00000161, -0.00000504, -0.00001100, -0.00000978, -0.00000071, -0.00000607, -0.00001047, 0.00000764, 0.00000848, -0.00000709, 0.00000180, -0.00000735, -0.00003324, -0.00000684, -0.00001520, -0.00011278, -0.00011119, 0.00004704, 0.00019829, 0.00027285, 0.00020625, -0.00002771, -0.00016744, -0.00007743, 0.00000188, -0.00001123, -0.00000135, -0.00000745, -0.00003000, -0.00000298, -0.00000046, -0.00002272, 0.00000203, 0.00000431, -0.00001378, 0.00001283, 0.00000444, -0.00002239, 0.00001422, 0.00000845, -0.00003391, 0.00000786, 0.00002202, -0.00002646, 0.00000542, 0.00002714, -0.00002774, -0.00000667, 0.00002743, -0.00002084, -0.00000849, 0.00002424, -0.00002836, -0.00001524, 0.00004011, -0.00001387, -0.00002942, 0.00003573, 0.00000707, -0.00002302, 0.00002680, -0.00000339, -0.00004970, 0.00001501, 0.00003637, -0.00001352, -0.00000098, 0.00000703, -0.00002531, -0.00000231, 0.00001914, -0.00000810, -0.00000063, 0.00000861, -0.00001846, -0.00000020, 0.00002487, -0.00001006, -0.00001205, 0.00002160, -0.00000393, -0.00002323, 0.00001195, 0.00000631, -0.00001968, 0.00001150, 0.00002280, -0.00002062, -0.00002493, 0.00000560, 0.00000502, 0.00000036, 0.00000885, -0.00000442, -0.00002086, -0.00000632, 0.00001363, 0.00000750, -0.00001056, -0.00000883, 0.00000629, -0.00000033, -0.00001425, 0.00000179, 0.00001832, 0.00000441, -0.00000501, 0.00000333, 0.00000297, -0.00000091, 0.00000192, -0.00000079, -0.00000436, -0.00000001, 0.00000205, 0.00000004, 0.00000122, 0.00000322, 0.00000237, -0.00000009, -0.00000193, -0.00000214, },
		{ -0.00000235, -0.00000209, -0.00000129, -0.00000127, -0.00000162, 0.00000067, 0.00000205, -0.00000099, -0.00000263, -0.00000688, -0.00001339, -0.00000521, 0.00000074, -0.00001183, -0.00000448, 0.00001396, -0.00000122, -0.00000697, 0.00000600, -0.00002158, -0.00003101, 0.00000297, -0.00004868, -0.00013624, -0.00007114, 0.00010966, 0.00026871, 0.00029943, 0.00010046, -0.00014415, -0.00013922, -0.00001961, -0.00000849, -0.00000852, 0.00000694, -0.00002850, -0.00002136, 0.00001115, -0.00001892, -0.00001607, 0.00001570, -0.00001177, -0.00000668, 0.00002284, -0.00001621, -0.00001334, 0.00002798, -0.00001746, -0.00002717, 0.00003104, -0.00000188, -0.00002657, 0.00003011, 0.00000386, -0.00003426, 0.00002104, 0.00001022, -0.00003113, 0.00001748, 0.00000749, -0.00004157, 0.00001724, 0.00002935, -0.00003821, -0.00000020, 0.00003771, -0.00001805, -0.00000265, 0.00002904, -0.00003385, -0.00003105, 0.00003693, 0.00001413, -0.00001439, 0.00000959, -0.00000925, -0.00002143, 0.00001352, 0.00000686, -0.00000988, 0.00000902, -0.00000408, -0.00001631, 0.00001739, 0.00001102, -0.00001886, 0.00000753, 0.00001544, -0.00002056, -0.00000733, 0.00001605, -0.00001143, -0.00000854, 0.00002560, 0.00000207, -0.00003134, -0.00000863, 0.00001062, 0.00000083, 0.00000322, 0.00000523, -0.00001202, -0.00001713, 0.00000155, 0.00001415, 0.00000145, -0.00001397, -0.00000350, 0.00000795, -0.00000764, -0.00001220, 0.00001148, 0.00001602, -0.00000161, -0.00000248, 0.00000450, 0.00000112, 0.00000001, 0.00000096, -0.00000285, -0.00000267, 0.00000120, 0.00000118, 0.00000037, 0.00000190, 0.00000307, 0.00000179, -0.00000090, },
		{ 0.00000100, -0.00000165, -0.00000266, -0.00000182, -0.00000090, -0.00000196, -0.00000091, 0.00000276, 0.00000035, -0.00000295, -0.00000257, -0.00001072, -0.00001295, 0.00000127, -0.00000354, -0.00001452, 0.00000669, 0.00001134, -0.00001074, 0.00000106, 0.00000121, -0.00003544, -0.00001801, -0.00000215, -0.00008865, -0.00013822, -0.00002096, 0.00019073, 0.00034718, 0.00025245, -0.00005443, -0.00018658, -0.00006554, -0.00000484, -0.00002140, 0.00001115, -0.00000752, -0.00004184, 0.00000477, 0.00000307, -0.00003335, 0.00000761, 0.00001002, -0.00002521, 0.00001525, 0.00001359, -0.00003251, 0.00001236, 0.00001981, -0.00004050, 0.00000172, 0.00003243, -0.00002907, -0.00000274, 0.00003562, -0.00002672, -0.00001472, 0.00003369, -0.00001873, -0.00001657, 0.00003140, -0.00002333, -0.00002822, 0.00004140, -0.00000218, -0.00003773, 0.00002954, 0.00001689, -0.00002499, 0.00002184, 0.00000821, -0.00004834, 0.00000057, 0.00003665, -0.00000483, -0.00000328, 0.00000633, -0.00002138, -0.00000589, 0.00001583, -0.00000544, -0.00000122, 0.00000787, -0.00001498, -0.00000136, 0.00002115, -0.00000744, -0.00000984, 0.00001905, -0.00000310, -0.00002081, 0.00000914, 0.00000420, -0.00001749, 0.00001158, 0.00002099, -0.00002043, -0.00002517, 0.00000599, 0.00000805, -0.00000071, 0.00000458, -0.00000081, -0.00001506, -0.00001162, 0.00000720, 0.00001270, -0.00000583, -0.00001409, 0.00000352, 0.00000465, -0.00001396, -0.00000439, 0.00001787, 0.00000978, -0.00000431, 0.00000063, 0.00000392, 0.00000042, 0.00000033, -0.00000095, -0.00000310, -0.00000088, 0.00000127, 0.00000081, 0.00000085, 0.00000227, 0.00000290, },
		{ 0.00000256, 0.00000249, 0.00000018, -0.00000235, -0.00000290, -0.00000102, -0.00000115, -0.00000267, 0.00000159, 0.00000319, -0.00000314, -0.00000254, -0.00000391, -0.00001532, -0.00000727, 0.00000457, -0.00001187, -0.00000896, 0.00001554, -0.00000009, -0.00001160, 0.00000933, -0.00001400, -0.00003909, -0.00000381, -0.00002487, -0.00012166, -0.00012487, 0.00004612, 0.00030326, 0.00038538, 0.00011085, -0.00018429, -0.00013852, -0.00000857, -0.00002555, -0.00000828, 0.00002115, -0.00003927, -0.00002672, 0.00002371, -0.00002489, -0.00002316, 0.00002751, -0.00001502, -0.00001708, 0.00003354, -0.00001464, -0.00002575, 0.00003544, -0.00001097, -0.00003907, 0.00003355, 0.00000691, -0.00003608, 0.00002885, 0.00001346, -0.00004038, 0.00001780, 0.00001903, -0.00003596, 0.00001346, 0.00001883, -0.00004518, 0.00000556, 0.00003900, -0.00003293, -0.00001242, 0.00003954, -0.00001010, -0.00000962, 0.00003085, -0.00002210, -0.00003824, 0.00002554, 0.00002040, -0.00000959, 0.00000616, -0.00000681, -0.00002002, 0.00000841, 0.00000653, -0.00000779, 0.00000725, -0.00000263, -0.00001370, 0.00001389, 0.00000949, -0.00001506, 0.00000743, 0.00001346, -0.00001816, -0.00000702, 0.00001249, -0.00001124, -0.00000662, 0.00002410, 0.00000118, -0.00003072, -0.00000927, 0.00001221, 0.00000320, -0.00000016, 0.00000315, -0.00000573, -0.00001491, -0.00000628, 0.00001170, 0.00000832, -0.00001285, -0.00000923, 0.00000839, -0.00000318, -0.00001501, 0.00000615, 0.00001854, 0.00000307, -0.00000386, 0.00000262, 0.00000297, 0.00000029, -0.00000059, -0.00000226, -0.00000208, 0.00000011, 0.00000105, 0.00000081, 0.00000125, },
		{ 0.00000094, 0.00000168, 0.00000259, 0.00000191, -0.00000054, -0.00000330, -0.00000249, -0.00000009, -0.00000258, -0.00000173, 0.00000471, 0.00000018, -0.00000542, -0.00000046, -0.00000886, -0.00001644, 0.00000166, 0.00000036, -0.00001732, 0.00000361, 0.00001432, -0.00001208, -0.00000251, 0.00000806, -0.00003144, -0.00003046, 0.00000006, -0.00005675, -0.00014469, -0.00009906, 0.00015378, 0.00042534, 0.00032281, -0.00008326, -0.00021800, -0.00004687, -0.00000722, -0.00003999, 0.00002477, -0.00000167, -0.00005823, 0.00001110, 0.00001176, -0.00004669, 0.00001036, 0.00002070, -0.00003747, 0.00001325, 0.00002668, -0.00004108, 0.00000545, 0.00003298, -0.00004358, -0.00000836, 0.00004254, -0.00002782, -0.00001353, 0.00004243, -0.00002211, -0.00002417, 0.00003784, -0.00001317, -0.00002574, 0.00003514, -0.00001392, -0.00003964, 0.00003703, 0.00001087, -0.00004145, 0.00001964, 0.00002494, -0.00002301, 0.00001473, 0.00001692, -0.00004185, -0.00001208, 0.00003184, 0.00000344, -0.00000283, 0.00000461, -0.00001776, -0.00000787, 0.00001166, -0.00000362, -0.00000091, 0.00000660, -0.00001189, -0.00000119, 0.00001706, -0.00000616, -0.00000675, 0.00001722, -0.00000331, -0.00001842, 0.00000716, 0.00000176, -0.00001567, 0.00001187, 0.00001910, -0.00001999, -0.00002503, 0.00000554, 0.00001075, -0.00000029, 0.00000024, 0.00000063, -0.00000845, -0.00001383, -0.00000045, 0.00001439, 0.00000040, -0.00001612, -0.00000092, 0.00000745, -0.00001144, -0.00000937, 0.00001476, 0.00001439, -0.00000148, -0.00000199, 0.00000346, 0.00000236, -0.00000034, -0.00000196, -0.00000227, -0.00000093, 0.00000044, 0.00000091, },
		{ 0.00000048, 0.00000084, 0.00000128, 0.00000201, 0.00000228, 0.00000156, -0.00000163, -0.00000419, -0.00000093, -0.00000037, -0.00000415, 0.00000191, 0.00000535, -0.00000507, -0.00000385, -0.00000046, -0.00001576, -0.00001065, 0.00000742, -0.00000968, -0.00001342, 0.00001474, 0.00000279, -0.00001499, 0.00000830, -0.00000500, -0.00004068, -0.00001767, -0.00000907, -0.00008943, -0.00016423, -0.00004195, 0.00031521, 0.00049022, 0.00013954, -0.00023680, -0.00014382, 0.00001321, -0.00004494, -0.00001562, 0.00004140, -0.00004781, -0.00003822, 0.00003906, -0.00002747, -0.00003556, 0.00004017, -0.00001356, -0.00003178, 0.00004263, -0.00000771, -0.00004038, 0.00003912, 0.00000008, -0.00005052, 0.00003174, 0.00001848, -0.00004369, 0.00002353, 0.00002430, -0.00004383, 0.00001122, 0.00002828, -0.00003765, 0.00000576, 0.00002942, -0.00004337, -0.00000834, 0.00004413, -0.00002316, -0.00002292, 0.00003656, -0.00000078, -0.00001395, 0.00002853, -0.00001013, -0.00003989, 0.00001257, 0.00002233, -0.00000312, 0.00000419, -0.00000574, -0.00001777, 0.00000412, 0.00000486, -0.00000564, 0.00000602, -0.00000207, -0.00001068, 0.00001142, 0.00000685, -0.00001188, 0.00000853, 0.00001145, -0.00001668, -0.00000622, 0.00000946, -0.00001152, -0.00000488, 0.00002272, 0.00000030, -0.00002966, -0.00000998, 0.00001285, 0.00000599, -0.00000220, -0.00000005, -0.00000104, -0.00001051, -0.00001198, 0.00000646, 0.00001294, -0.00000879, -0.00001315, 0.00000616, 0.00000067, -0.00001519, 0.00000042, 0.00001841, 0.00000835, -0.00000338, -0.00000007, 0.00000386, 0.00000167, -0.00000173, -0.00000269, -0.00000145, -0.00000020, },
		{ -0.00000050, 0.00000013, 0.00000033, 0.00000100, 0.00000182, 0.00000179, 0.00000211, 0.00000128, -0.00000355, -0.00000384, 0.00000095, -0.00000268, -0.00000338, 0.00000654, 0.00000124, -0.00000843, 0.00000039, -0.00000564, -0.00001908, -0.00000002, 0.00000451, -0.00001767, -0.00000105, 0.00001570, -0.00001034, -0.00000714, 0.00001054, -0.00002212, -0.00003933, -0.00000985, -0.00002479, -0.00012685, -0.00017645, 0.00008306, 0.00049614, 0.00042114, -0.00010646, -0.00026597, -0.00002481, -0.00000051, -0.00006700, 0.00003574, 0.00001221, -0.00007758, 0.00001319, 0.00002632, -0.00006072, 0.00000800, 0.00003601, -0.00004805, 0.00000548, 0.00004222, -0.00004575, -0.00000659, 0.00004582, -0.00004170, -0.00002145, 0.00005042, -0.00002218, -0.00002559, 0.00004605, -0.00001399, -0.00003355, 0.00003862, -0.00000458, -0.00003411, 0.00003434, -0.00000174, -0.00004716, 0.00002751, 0.00002271, -0.00003974, 0.00000792, 0.00002946, -0.00001753, 0.00000750, 0.00002159, -0.00003200, -0.00002092, 0.00002308, 0.00000943, 0.00000010, 0.00000292, -0.00001493, -0.00000851, 0.00000728, -0.00000287, 0.00000002, 0.00000523, -0.00000948, -0.00000003, 0.00001325, -0.00000621, -0.00000346, 0.00001642, -0.00000413, -0.00001655, 0.00000591, -0.00000049, -0.00001435, 0.00001212, 0.00001729, -0.00001937, -0.00002450, 0.00000459, 0.00001277, 0.00000114, -0.00000340, 0.00000026, -0.00000231, -0.00001312, -0.00000760, 0.00001257, 0.00000622, -0.00001474, -0.00000535, 0.00000761, -0.00000796, -0.00001228, 0.00000991, 0.00001713, 0.00000293, -0.00000350, 0.00000166, 0.00000403, 0.00000025, -0.00000315, -0.00000241, },
		{ -0.00000384, -0.00000147, 0.00000028, 0.00000004, 0.00000013, 0.00000169, 0.00000195, 0.00000122, 0.00000266, 0.00000001, -0.00000558, -0.00000145, 0.00000103, -0.00000552, 0.00000123, 0.00000805, -0.00000596, -0.00000647, 0.00000239, -0.00001419, -0.00001436, 0.00000811, -0.00000570, -0.00001621, 0.00001101, 0.00000595, -0.00001518, 0.00000336, 0.00000196, -0.00003407, -0.00003409, -0.00000664, -0.00004526, -0.00017997, -0.00014719, 0.00029412, 0.00060677, 0.00019523, -0.00029797, -0.00016183, 0.00004616, -0.00006197, -0.00003347, 0.00006565, -0.00005064, -0.00005650, 0.00005482, -0.00002426, -0.00005283, 0.00005102, -0.00000588, -0.00004919, 0.00004756, 0.00000476, -0.00005489, 0.00003734, 0.00001463, -0.00005932, 0.00002506, 0.00003127, -0.00004775, 0.00001436, 0.00003471, -0.00004349, 0.00000191, 0.00003626, -0.00003531, -0.00000432, 0.00003708, -0.00003626, -0.00002186, 0.00004357, -0.00001083, -0.00002963, 0.00002934, 0.00000783, -0.00001469, 0.00002338, -0.00000011, -0.00003639, 0.00000045, 0.00001970, 0.00000333, 0.00000427, -0.00000547, -0.00001537, 0.00000091, 0.00000229, -0.00000389, 0.00000542, -0.00000212, -0.00000776, 0.00001004, 0.00000371, -0.00000971, 0.00001040, 0.00000995, -0.00001605, -0.00000544, 0.00000717, -0.00001195, -0.00000357, 0.00002138, -0.00000047, -0.00002830, -0.00001055, 0.00001272, 0.00000863, -0.00000296, -0.00000347, 0.00000172, -0.00000519, -0.00001488, -0.00000007, 0.00001438, -0.00000329, -0.00001424, 0.00000244, 0.00000279, -0.00001338, -0.00000443, 0.00001584, 0.00001294, -0.00000091, -0.00000259, 0.00000333, 0.00000350, -0.00000178, },
		{ 0.00000187, -0.00000371, -0.00000360, -0.00000023, 0.00000069, -0.00000057, 0.00000050, 0.00000263, 0.00000107, 0.00000134, 0.00000335, -0.00000298, -0.00000574, 0.00000172, -0.00000196, -0.00000547, 0.00000754, 0.00000348, -0.00001102, -0.00000055, -0.00000189, -0.00001973, -0.00000343, 0.00000733, -0.00001500, -0.00000584, 0.00001388, -0.00000594, -0.00001029, 0.00000713, -0.00001035, -0.00004020, -0.00002993, -0.00000117, -0.00008538, -0.00024061, -0.00002077, 0.00054616, 0.00054701, -0.00011376, -0.00033231, -0.00000539, 0.00001940, -0.00010009, 0.00003956, 0.00003486, -0.00009687, 0.00000842, 0.00004614, -0.00007251, -0.00000103, 0.00005426, -0.00005421, -0.00000834, 0.00005772, -0.00004461, -0.00002267, 0.00005583, -0.00003425, -0.00003574, 0.00005429, -0.00001243, -0.00003708, 0.00004538, -0.00000321, -0.00004113, 0.00003539, 0.00000579, -0.00003982, 0.00002893, 0.00001092, -0.00004927, 0.00001456, 0.00003095, -0.00003302, -0.00000326, 0.00002951, -0.00000995, 0.00000199, 0.00002217, -0.00002104, -0.00002488, 0.00001232, 0.00001188, 0.00000452, 0.00000222, -0.00001293, -0.00000831, 0.00000325, -0.00000317, 0.00000116, 0.00000410, -0.00000780, 0.00000167, 0.00001017, -0.00000725, -0.00000060, 0.00001658, -0.00000501, -0.00001545, 0.00000508, -0.00000219, -0.00001342, 0.00001204, 0.00001561, -0.00001857, -0.00002366, 0.00000345, 0.00001401, 0.00000305, -0.00000594, -0.00000125, 0.00000248, -0.00001027, -0.00001294, 0.00000807, 0.00001010, -0.00001068, -0.00000825, 0.00000552, -0.00000487, -0.00001296, 0.00000457, 0.00001748, 0.00000787, -0.00000320, -0.00000085, 0.00000462, },
		{ 0.00000150, 0.00000486, -0.00000050, -0.00000496, -0.00000259, 0.00000112, 0.00000028, -0.00000125, 0.00000193, 0.00000268, -0.00000026, 0.00000295, 0.00000239, -0.00000649, -0.00000278, 0.00000266, -0.00000616, -0.00000045, 0.00001046, -0.00000529, -0.00000967, 0.00000358, -0.00001087, -0.00001696, 0.00000601, -0.00000130, -0.00001578, 0.00000489, 0.00000725, -0.00001100, -0.00000348, 0.00000345, -0.00001988, -0.00004720, -0.00002000, 0.00000207, -0.00016613, -0.00025781, 0.00023275, 0.00072319, 0.00028455, -0.00036011, -0.00019891, 0.00008776, -0.00007080, -0.00006306, 0.00009005, -0.00004459, -0.00008061, 0.00006768, -0.00001363, -0.00007313, 0.00005711, 0.00000835, -0.00006669, 0.00004623, 0.00002168, -0.00006656, 0.00002932, 0.00003066, -0.00006351, 0.00001396, 0.00004324, -0.00004716, 0.00000238, 0.00004286, -0.00003891, -0.00000875, 0.00004132, -0.00002897, -0.00001485, 0.00004024, -0.00002516, -0.00003247, 0.00003740, 0.00000157, -0.00003133, 0.00001944, 0.00001384, -0.00001194, 0.00001723, 0.00000659, -0.00002919, -0.00000876, 0.00001338, 0.00000809, 0.00000627, -0.00000520, -0.00001325, -0.00000131, -0.00000068, -0.00000284, 0.00000533, -0.00000244, -0.00000528, 0.00000955, 0.00000069, -0.00000864, 0.00001249, 0.00000926, -0.00001596, -0.00000511, 0.00000558, -0.00001214, -0.00000279, 0.00001992, -0.00000103, -0.00002671, -0.00001090, 0.00001209, 0.00001074, -0.00000275, -0.00000639, 0.00000276, -0.00000011, -0.00001503, -0.00000633, 0.00001259, 0.00000194, -0.00001234, -0.00000127, 0.00000277, -0.00001068, -0.00000759, 0.00001167, 0.00001584, 0.00000300, -0.00000395, },
		{ -0.00000101, -0.00000308, 0.00000370, 0.00000381, -0.00000297, -0.00000540, -0.00000082, 0.00000204, -0.00000119, -0.00000083, 0.00000374, 0.00000097, -0.00000030, 0.00000503, -0.00000137, -0.00000771, 0.00000170, -0.00000048, -0.00000729, 0.00000722, 0.00000630, -0.00001220, -0.00000313, 0.00000105, -0.00001776, -0.00000739, 0.00000732, -0.00000978, -0.00000871, 0.00000799, -0.00000060, -0.00000933, -0.00000198, -0.00000032, -0.00003041, -0.00005522, 0.00000692, -0.00002578, -0.00027962, -0.00015120, 0.00056201, 0.00069458, -0.00009405, -0.00041496, 0.00000346, 0.00005467, -0.00013456, 0.00003199, 0.00006503, -0.00011207, -0.00000493, 0.00006905, -0.00007892, -0.00001707, 0.00007267, -0.00005364, -0.00002709, 0.00007024, -0.00003675, -0.00004063, 0.00006077, -0.00002175, -0.00004892, 0.00005301, 0.00000017, -0.00004604, 0.00004008, 0.00000868, -0.00004537, 0.00002839, 0.00001615, -0.00004152, 0.00001991, 0.00002163, -0.00004565, 0.00000067, 0.00003401, -0.00002284, -0.00001171, 0.00002525, -0.00000216, -0.00000063, 0.00001964, -0.00001108, -0.00002406, 0.00000179, 0.00001045, 0.00000901, 0.00000301, -0.00001137, -0.00000773, -0.00000016, -0.00000425, 0.00000215, 0.00000335, -0.00000671, 0.00000342, 0.00000799, -0.00000878, 0.00000145, 0.00001740, -0.00000552, -0.00001509, 0.00000429, -0.00000314, -0.00001264, 0.00001148, 0.00001401, -0.00001756, -0.00002259, 0.00000233, 0.00001455, 0.00000490, -0.00000731, -0.00000312, 0.00000555, -0.00000635, -0.00001575, 0.00000224, 0.00001115, -0.00000534, -0.00000873, 0.00000224, -0.00000313, -0.00001196, -0.00000009, 0.00001564, 0.00001214, },
		{ 0.00001654, 0.00000736, -0.00000349, -0.00000098, 0.00000498, 0.00000182, -0.00000521, -0.00000483, 0.00000161, 0.00000152, -0.00000287, 0.00000145, 0.00000411, -0.00000153, 0.00000214, 0.00000498, -0.00000640, -0.00000477, 0.00000389, -0.00000571, -0.00000261, 0.00001170, -0.00000311, -0.00001226, 0.00000242, -0.00000681, -0.00001700, 0.00000145, 0.00000161, -0.00001129, -0.00000195, 0.00000456, -0.00000271, -0.00000910, -0.00000279, -0.00000117, -0.00005126, -0.00004786, 0.00004238, -0.00012166, -0.00035929, 0.00012914, 0.00082433, 0.00040992, -0.00041206, -0.00025933, 0.00013220, -0.00006564, -0.00010315, 0.00010945, -0.00002762, -0.00010786, 0.00007415, 0.00000468, -0.00009343, 0.00005598, 0.00002798, -0.00008107, 0.00003757, 0.00004075, -0.00007286, 0.00001558, 0.00004553, -0.00006182, -0.00000013, 0.00005228, -0.00004165, -0.00001066, 0.00004720, -0.00003052, -0.00001894, 0.00004234, -0.00001958, -0.00002371, 0.00003836, -0.00001226, -0.00003826, 0.00002692, 0.00001154, -0.00002801, 0.00000905, 0.00001606, -0.00000675, 0.00001192, 0.00000963, -0.00002027, -0.00001388, 0.00000500, 0.00000997, 0.00000941, -0.00000419, -0.00001147, -0.00000280, -0.00000363, -0.00000254, 0.00000549, -0.00000274, -0.00000338, 0.00000962, -0.00000177, -0.00000844, 0.00001425, 0.00000937, -0.00001599, -0.00000541, 0.00000447, -0.00001181, -0.00000246, 0.00001815, -0.00000139, -0.00002494, -0.00001103, 0.00001122, 0.00001218, -0.00000203, -0.00000839, 0.00000262, 0.00000388, -0.00001306, -0.00001108, 0.00000830, 0.00000548, -0.00000816, -0.00000364, 0.00000098, -0.00000827, -0.00000892, 0.00000701, },
		{ -0.00000336, 0.00001234, 0.00001478, 0.00000255, -0.00000406, 0.00000144, 0.00000513, -0.00000075, -0.00000703, -0.00000270, 0.00000367, -0.00000094, -0.00000287, 0.00000452, 0.00000179, -0.00000229, 0.00000572, 0.00000107, -0.00000907, 0.00000064, 0.00000127, -0.00000812, 0.00000561, 0.00000868, -0.00001131, -0.00000648, 0.00000184, -0.00001336, -0.00001013, 0.00000358, -0.00000353, -0.00000765, -0.00000158, 0.00000310, -0.00000253, -0.00001465, 0.00000361, -0.00000996, -0.00007882, -0.00000470, 0.00004243, -0.00028464, -0.00029622, 0.00053299, 0.00085240, -0.00003757, -0.00050730, -0.00000667, 0.00010435, -0.00016386, 0.00001021, 0.00009932, -0.00011901, -0.00002696, 0.00009163, -0.00007729, -0.00003888, 0.00008781, -0.00004512, -0.00004831, 0.00007699, -0.00002261, -0.00005758, 0.00005915, -0.00000584, -0.00005863, 0.00004640, 0.00001371, -0.00005084, 0.00003073, 0.00001972, -0.00004532, 0.00001869, 0.00002459, -0.00003876, 0.00000912, 0.00002847, -0.00003720, -0.00001153, 0.00003152, -0.00001148, -0.00001593, 0.00001791, 0.00000393, -0.00000012, 0.00001558, -0.00000362, -0.00001956, -0.00000657, 0.00000577, 0.00001211, 0.00000519, -0.00000969, -0.00000710, -0.00000290, -0.00000580, 0.00000272, 0.00000298, -0.00000598, 0.00000489, 0.00000662, -0.00001027, 0.00000256, 0.00001843, -0.00000538, -0.00001524, 0.00000321, -0.00000339, -0.00001173, 0.00001038, 0.00001236, -0.00001634, -0.00002134, 0.00000137, 0.00001459, 0.00000634, -0.00000774, -0.00000470, 0.00000697, -0.00000242, -0.00001603, -0.00000352, 0.00000932, -0.00000024, -0.00000668, -0.00000089, -0.00000308, -0.00001027, },
		{ -0.00000697, -0.00000891, 0.00000293, 0.00001523, 0.00001103, -0.00000116, -0.00000310, 0.00000346, 0.00000438, -0.00000395, -0.00000766, 0.00000102, 0.00000360, -0.00000404, -0.00000000, 0.00000576, -0.00000199, 0.00000041, 0.00000703, -0.00000508, -0.00000684, 0.00000407, -0.00000426, -0.00000438, 0.00001116, -0.00000016, -0.00001301, -0.00000097, -0.00000360, -0.00001360, -0.00000424, 0.00000109, -0.00000305, -0.00000709, -0.00000350, 0.00000790, -0.00001014, -0.00001711, 0.00001721, -0.00004227, -0.00008522, 0.00006628, -0.00005103, -0.00043687, -0.00001215, 0.00089397, 0.00056795, -0.00044075, -0.00034351, 0.00017099, -0.00004231, -0.00014967, 0.00011845, 0.00000026, -0.00013410, 0.00007135, 0.00002922, -0.00011002, 0.00004630, 0.00005043, -0.00008930, 0.00002198, 0.00005885, -0.00007210, -0.00000214, 0.00005660, -0.00005410, -0.00001501, 0.00005670, -0.00003194, -0.00002268, 0.00004684, -0.00001959, -0.00002684, 0.00003908, -0.00000884, -0.00002918, 0.00003204, -0.00000004, -0.00003844, 0.00001438, 0.00001724, -0.00002078, 0.00000035, 0.00001432, -0.00000079, 0.00000873, 0.00000965, -0.00001165, -0.00001477, -0.00000348, 0.00000860, 0.00001248, -0.00000208, -0.00000978, -0.00000386, -0.00000628, -0.00000291, 0.00000564, -0.00000282, -0.00000204, 0.00000985, -0.00000346, -0.00000871, 0.00001534, 0.00001005, -0.00001578, -0.00000634, 0.00000354, -0.00001085, -0.00000237, 0.00001600, -0.00000162, -0.00002302, -0.00001095, 0.00001034, 0.00001294, -0.00000124, -0.00000934, 0.00000194, 0.00000639, -0.00000990, -0.00001364, 0.00000270, 0.00000655, -0.00000294, -0.00000390, -0.00000159, },
		{ -0.00000271, -0.00000431, -0.00000888, -0.00000512, 0.00000864, 0.00001536, 0.00000650, -0.00000318, -0.00000145, 0.00000503, 0.00000255, -0.00000756, -0.00000580, 0.00000478, 0.00000047, -0.00000505, 0.00000438, 0.00000337, -0.00000391, 0.00000513, 0.00000370, -0.00000918, -0.00000129, 0.00000249, -0.00000745, 0.00000333, 0.00000951, -0.00000837, -0.00000920, -0.00000044, -0.00000767, -0.00000984, -0.00000371, 0.00000146, -0.00000136, -0.00001310, 0.00000237, 0.00001034, -0.00002727, -0.00000194, 0.00001826, -0.00009353, -0.00003605, 0.00010746, -0.00025194, -0.00043988, 0.00045357, 0.00100448, 0.00006158, -0.00059831, -0.00004267, 0.00016377, -0.00018087, -0.00002612, 0.00013248, -0.00011431, -0.00005579, 0.00010978, -0.00006625, -0.00006367, 0.00009642, -0.00002900, -0.00006868, 0.00007608, -0.00000401, -0.00007055, 0.00005076, 0.00001098, -0.00006306, 0.00003537, 0.00002594, -0.00005064, 0.00001878, 0.00002805, -0.00004089, 0.00000801, 0.00002960, -0.00003210, -0.00000128, 0.00003043, -0.00002585, -0.00001990, 0.00002436, -0.00000140, -0.00001552, 0.00000939, 0.00000702, 0.00000276, 0.00001169, 0.00000077, -0.00001310, -0.00001156, -0.00000079, 0.00001281, 0.00000811, -0.00000734, -0.00000643, -0.00000510, -0.00000752, 0.00000275, 0.00000287, -0.00000536, 0.00000588, 0.00000581, -0.00001134, 0.00000290, 0.00001923, -0.00000461, -0.00001557, 0.00000170, -0.00000313, -0.00001049, 0.00000886, 0.00001057, -0.00001497, -0.00001997, 0.00000064, 0.00001435, 0.00000718, -0.00000758, -0.00000558, 0.00000717, 0.00000075, -0.00001427, -0.00000807, 0.00000525, 0.00000334, -0.00000277, },
		{ 0.00000191, -0.00000203, -0.00000376, -0.00000700, -0.00000843, 0.00000006, 0.00001261, 0.00001311, 0.00000252, -0.00000389, 0.00000051, 0.00000618, -0.00000105, -0.00001015, -0.00000107, 0.00000589, -0.00000409, -0.00000223, 0.00000690, -0.00000132, -0.00000165, 0.00000783, -0.00000273, -0.00000816, 0.00000285, -0.00000238, -0.00000482, 0.00000884, 0.00000233, -0.00001104, -0.00000564, -0.00000297, -0.00000679, -0.00000899, -0.00000420, 0.00000736, -0.00000791, -0.00001696, 0.00001784, -0.00000395, -0.00003756, 0.00002981, -0.00001904, -0.00012444, 0.00006724, 0.00003625, -0.00047833, -0.00018005, 0.00091772, 0.00074888, -0.00043352, -0.00044697, 0.00019425, 0.00000039, -0.00019611, 0.00011260, 0.00003667, -0.00015448, 0.00005790, 0.00005686, -0.00011938, 0.00002842, 0.00007214, -0.00008926, 0.00000137, 0.00007275, -0.00006391, -0.00002109, 0.00006177, -0.00004142, -0.00002818, 0.00005571, -0.00001963, -0.00003173, 0.00004181, -0.00000796, -0.00003105, 0.00003220, 0.00000126, -0.00003035, 0.00002284, 0.00000930, -0.00003349, 0.00000233, 0.00001791, -0.00001164, -0.00000501, 0.00000947, 0.00000418, 0.00000810, 0.00000795, -0.00000480, -0.00001227, -0.00001029, 0.00000447, 0.00001430, 0.00000102, -0.00000775, -0.00000463, -0.00000856, -0.00000376, 0.00000558, -0.00000263, -0.00000111, 0.00000992, -0.00000441, -0.00000904, 0.00001564, 0.00001093, -0.00001512, -0.00000768, 0.00000253, -0.00000932, -0.00000230, 0.00001348, -0.00000183, -0.00002102, -0.00001069, 0.00000961, 0.00001316, -0.00000071, -0.00000938, 0.00000128, 0.00000741, -0.00000646, -0.00001395, -0.00000292, 0.00000509, },
		{ 0.00000002, 0.00000470, 0.00000184, -0.00000260, -0.00000592, -0.00000849, -0.00000584, 0.00000547, 0.00001406, 0.00000973, -0.00000039, -0.00000393, 0.00000313, 0.00000603, -0.00000634, -0.00000947, 0.00000454, 0.00000284, -0.00000660, 0.00000303, 0.00000513, -0.00000445, 0.00000343, 0.00000564, -0.00000770, -0.00000344, 0.00000235, -0.00000533, 0.00000145, 0.00000797, -0.00000420, -0.00000970, -0.00000583, -0.00000259, -0.00000520, -0.00001318, 0.00000271, 0.00001022, -0.00002377, -0.00000492, 0.00002767, -0.00003561, -0.00001824, 0.00004898, -0.00009354, -0.00008279, 0.00015755, -0.00018394, -0.00056486, 0.00032515, 0.00113247, 0.00020353, -0.00067393, -0.00010799, 0.00022480, -0.00017938, -0.00007429, 0.00015838, -0.00009644, -0.00008767, 0.00011963, -0.00004624, -0.00008761, 0.00009614, -0.00000726, -0.00008468, 0.00006701, 0.00001612, -0.00007718, 0.00003670, 0.00002592, -0.00006137, 0.00002175, 0.00003482, -0.00004557, 0.00000619, 0.00003232, -0.00003293, -0.00000171, 0.00003046, -0.00002296, -0.00000930, 0.00002767, -0.00001398, -0.00002327, 0.00001442, 0.00000543, -0.00001122, 0.00000174, 0.00000668, 0.00000665, 0.00000924, 0.00000249, -0.00000648, -0.00001293, -0.00000753, 0.00001078, 0.00001077, -0.00000413, -0.00000549, -0.00000686, -0.00000923, 0.00000226, 0.00000286, -0.00000468, 0.00000636, 0.00000527, -0.00001177, 0.00000275, 0.00001951, -0.00000339, -0.00001577, -0.00000019, -0.00000261, -0.00000886, 0.00000712, 0.00000858, -0.00001355, -0.00001855, 0.00000019, 0.00001403, 0.00000738, -0.00000719, -0.00000561, 0.00000667, 0.00000274, -0.00001127, -0.00001077, },
		{ -0.00001241, -0.00000755, 0.00000170, 0.00000527, 0.00000131, -0.00000453, -0.00000793, -0.00000837, -0.00000157, 0.00000967, 0.00001320, 0.00000653, -0.00000275, -0.00000307, 0.00000637, 0.00000302, -0.00001130, -0.00000454, 0.00000747, -0.00000274, -0.00000459, 0.00000683, 0.00000029, -0.00000317, 0.00000702, -0.00000008, -0.00000794, 0.00000036, -0.00000105, -0.00000325, 0.00000543, 0.00000294, -0.00000622, -0.00000991, -0.00000619, 0.00000224, -0.00001047, -0.00001441, 0.00001733, -0.00000244, -0.00003423, 0.00002361, 0.00001049, -0.00006026, 0.00003190, 0.00001623, -0.00015682, 0.00004312, 0.00012732, -0.00047644, -0.00035825, 0.00088591, 0.00093718, -0.00038071, -0.00056003, 0.00019270, 0.00005973, -0.00023455, 0.00008969, 0.00007690, -0.00016449, 0.00003448, 0.00008335, -0.00011898, 0.00000445, 0.00008931, -0.00008031, -0.00002114, 0.00007983, -0.00004942, -0.00003817, 0.00006007, -0.00002586, -0.00003744, 0.00004957, -0.00000686, -0.00003645, 0.00003309, 0.00000237, -0.00003100, 0.00002315, 0.00000897, -0.00002733, 0.00001285, 0.00001442, -0.00002496, -0.00000698, 0.00001400, -0.00000287, -0.00000636, 0.00000312, 0.00000683, 0.00000954, 0.00000602, -0.00000038, -0.00000784, -0.00001431, -0.00000123, 0.00001400, 0.00000447, -0.00000504, -0.00000506, -0.00001048, -0.00000493, 0.00000518, -0.00000220, -0.00000042, 0.00000963, -0.00000479, -0.00000908, 0.00001527, 0.00001166, -0.00001403, -0.00000915, 0.00000137, -0.00000740, -0.00000210, 0.00001074, -0.00000210, -0.00001904, -0.00001030, 0.00000915, 0.00001301, -0.00000064, -0.00000878, 0.00000103, 0.00000726, -0.00000348, },
		{ 0.00000351, -0.00000784, -0.00001153, -0.00000523, 0.00000383, 0.00000586, -0.00000064, -0.00000684, -0.00000918, -0.00000634, 0.00000335, 0.00001160, 0.00001138, 0.00000370, -0.00000500, -0.00000029, 0.00000860, -0.00000325, -0.00001266, 0.00000251, 0.00000541, -0.00000674, 0.00000072, 0.00000621, -0.00000352, 0.00000131, 0.00000605, -0.00000485, -0.00000480, 0.00000032, -0.00000242, 0.00000118, 0.00000394, -0.00000024, -0.00000677, -0.00001336, -0.00000026, 0.00000407, -0.00002248, -0.00000169, 0.00002566, -0.00003038, -0.00001950, 0.00004626, -0.00003595, -0.00004400, 0.00007575, -0.00007598, -0.00013706, 0.00018341, -0.00008907, -0.00065548, 0.00015653, 0.00121866, 0.00038144, -0.00071926, -0.00020108, 0.00027697, -0.00015581, -0.00012858, 0.00017123, -0.00006634, -0.00011762, 0.00011850, -0.00001959, -0.00010658, 0.00008626, 0.00001678, -0.00009337, 0.00005092, 0.00003438, -0.00007630, 0.00001928, 0.00003646, -0.00005393, 0.00000792, 0.00003903, -0.00003674, -0.00000494, 0.00003202, -0.00002299, -0.00000880, 0.00002738, -0.00001322, -0.00001366, 0.00002139, -0.00000387, -0.00002168, 0.00000403, 0.00000798, -0.00000465, -0.00000337, 0.00000349, 0.00000995, 0.00000876, 0.00000257, -0.00000109, -0.00001133, -0.00001290, 0.00000646, 0.00001214, -0.00000027, -0.00000399, -0.00000818, -0.00001083, 0.00000129, 0.00000281, -0.00000383, 0.00000640, 0.00000470, -0.00001155, 0.00000249, 0.00001917, -0.00000204, -0.00001570, -0.00000228, -0.00000202, -0.00000693, 0.00000535, 0.00000648, -0.00001220, -0.00001718, -0.00000000, 0.00001381, 0.00000703, -0.00000687, -0.00000488, 0.00000596, },
		{ 0.00000134, 0.00000637, -0.00000133, -0.00000970, -0.00001060, -0.00000265, 0.00000654, 0.00000499, -0.00000344, -0.00000886, -0.00000923, -0.00000245, 0.00000724, 0.00001157, 0.00000978, 0.00000031, -0.00000617, 0.00000456, 0.00000722, -0.00001038, -0.00000865, 0.00000740, -0.00000021, -0.00000613, 0.00000521, 0.00000217, -0.00000330, 0.00000476, 0.00000183, -0.00000567, -0.00000268, -0.00000135, 0.00000042, 0.00000219, 0.00000056, 0.00000064, -0.00001183, -0.00001364, 0.00001140, -0.00000682, -0.00002833, 0.00002383, 0.00000999, -0.00005236, 0.00002216, 0.00003214, -0.00008023, 0.00002083, 0.00005837, -0.00017460, -0.00000325, 0.00020833, -0.00043050, -0.00052765, 0.00079575, 0.00111358, -0.00027814, -0.00066890, 0.00015986, 0.00012895, -0.00025725, 0.00005060, 0.00011472, -0.00016108, 0.00000390, 0.00010420, -0.00010804, -0.00002207, 0.00009873, -0.00006357, -0.00004187, 0.00007870, -0.00003107, -0.00005056, 0.00005188, -0.00001013, -0.00004137, 0.00003959, 0.00000421, -0.00003641, 0.00002236, 0.00000975, -0.00002705, 0.00001379, 0.00001317, -0.00002117, 0.00000411, 0.00001503, -0.00001500, -0.00001213, 0.00000700, 0.00000358, -0.00000411, -0.00000287, 0.00000661, 0.00001194, 0.00000506, 0.00000181, -0.00000304, -0.00001527, -0.00000703, 0.00001137, 0.00000739, -0.00000160, -0.00000490, -0.00001202, -0.00000631, 0.00000439, -0.00000160, 0.00000021, 0.00000892, -0.00000485, -0.00000864, 0.00001448, 0.00001199, -0.00001268, -0.00001051, 0.00000009, -0.00000532, -0.00000174, 0.00000797, -0.00000244, -0.00001722, -0.00000986, 0.00000902, 0.00001269, -0.00000109, -0.00000785, },
		{ -0.00000682, -0.00000357, 0.00000537, 0.00000328, -0.00000463, -0.00001068, -0.00000958, 0.00000129, 0.00000831, 0.00000246, -0.00000621, -0.00001043, -0.00000738, 0.00000225, 0.00000890, 0.00001114, 0.00000806, -0.00000404, -0.00000436, 0.00000930, 0.00000114, -0.00001423, -0.00000118, 0.00000711, -0.00000509, -0.00000182, 0.00000577, -0.00000131, -0.00000020, 0.00000440, -0.00000138, -0.00000435, -0.00000358, 0.00000013, 0.00000350, -0.00000193, 0.00000172, 0.00000000, -0.00002117, -0.00000248, 0.00001695, -0.00002879, -0.00001250, 0.00004222, -0.00003014, -0.00004029, 0.00006199, -0.00002610, -0.00007549, 0.00009241, -0.00004164, -0.00018906, 0.00018008, 0.00001967, -0.00070052, -0.00003700, 0.00124910, 0.00058174, -0.00072130, -0.00031479, 0.00030938, -0.00011041, -0.00018113, 0.00016701, -0.00002757, -0.00014042, 0.00010564, 0.00000981, -0.00011710, 0.00006790, 0.00003916, -0.00009317, 0.00003037, 0.00004766, -0.00006821, 0.00000148, 0.00004100, -0.00004228, -0.00000377, 0.00003819, -0.00002596, -0.00001294, 0.00002760, -0.00001301, -0.00001229, 0.00002142, -0.00000476, -0.00001406, 0.00001341, 0.00000290, -0.00001629, -0.00000459, 0.00000634, 0.00000215, -0.00000512, -0.00000124, 0.00001142, 0.00000996, 0.00000229, 0.00000243, -0.00000799, -0.00001590, 0.00000093, 0.00001156, 0.00000361, -0.00000170, -0.00000891, -0.00001229, -0.00000010, 0.00000259, -0.00000280, 0.00000615, 0.00000391, -0.00001082, 0.00000240, 0.00001830, -0.00000085, -0.00001538, -0.00000432, -0.00000145, -0.00000487, 0.00000368, 0.00000441, -0.00001099, -0.00001599, 0.00000004, 0.00001382, 0.00000626, },
		{ 0.00001237, -0.00000203, -0.00000685, 0.00000219, 0.00000512, -0.00000011, -0.00000651, -0.00001196, -0.00000693, 0.00000578, 0.00000787, -0.00000069, -0.00000880, -0.00001099, -0.00000334, 0.00000576, 0.00000893, 0.00001146, 0.00000465, -0.00000773, 0.00000109, 0.00001033, -0.00000728, -0.00001217, 0.00000513, 0.00000265, -0.00000594, 0.00000227, 0.00000329, -0.00000153, 0.00000180, 0.00000195, -0.00000146, -0.00000511, -0.00000408, 0.00000554, 0.00000066, -0.00000522, 0.00000789, -0.00000971, -0.00002449, 0.00001808, 0.00000277, -0.00004349, 0.00002422, 0.00002841, -0.00006761, 0.00001177, 0.00005796, -0.00009259, -0.00000331, 0.00010027, -0.00017271, -0.00006465, 0.00026710, -0.00034664, -0.00066956, 0.00065247, 0.00125794, -0.00012868, -0.00075757, 0.00009387, 0.00019806, -0.00025841, -0.00000047, 0.00014356, -0.00014353, -0.00002930, 0.00011563, -0.00008779, -0.00004685, 0.00009860, -0.00004175, -0.00005737, 0.00006961, -0.00001209, -0.00005640, 0.00003886, 0.00000313, -0.00003969, 0.00002781, 0.00001188, -0.00003214, 0.00001163, 0.00001323, -0.00002036, 0.00000588, 0.00001364, -0.00001351, -0.00000186, 0.00001194, -0.00000583, -0.00001282, -0.00000103, 0.00000652, 0.00000046, -0.00000693, 0.00000389, 0.00001391, 0.00000558, 0.00000261, 0.00000090, -0.00001373, -0.00001158, 0.00000689, 0.00000893, 0.00000225, -0.00000391, -0.00001306, -0.00000784, 0.00000317, -0.00000093, 0.00000088, 0.00000782, -0.00000484, -0.00000770, 0.00001357, 0.00001183, -0.00001133, -0.00001165, -0.00000114, -0.00000323, -0.00000131, 0.00000536, -0.00000275, -0.00001567, -0.00000949, 0.00000925, },
		{ 0.00000022, 0.00001417, 0.00000526, -0.00000713, -0.00000194, 0.00000507, 0.00000238, -0.00000198, -0.00000877, -0.00001251, -0.00000201, 0.00000874, 0.00000562, -0.00000373, -0.00001134, -0.00000934, 0.00000162, 0.00000677, 0.00000931, 0.00001158, -0.00000109, -0.00000786, 0.00000775, 0.00000567, -0.00001334, -0.00000568, 0.00000684, -0.00000190, -0.00000347, 0.00000333, 0.00000145, -0.00000006, 0.00000077, 0.00000152, -0.00000135, -0.00000864, 0.00000095, 0.00000874, -0.00000818, 0.00000016, 0.00000996, -0.00002691, -0.00001009, 0.00003124, -0.00002960, -0.00002893, 0.00005640, -0.00002149, -0.00006448, 0.00007087, -0.00000609, -0.00010730, 0.00009456, 0.00000508, -0.00022919, 0.00014794, 0.00012706, -0.00069530, -0.00023587, 0.00121634, 0.00078571, -0.00067177, -0.00043684, 0.00031294, -0.00004780, -0.00022338, 0.00014455, 0.00001423, -0.00015178, 0.00008260, 0.00003728, -0.00011719, 0.00004393, 0.00005613, -0.00008420, 0.00000887, 0.00005385, -0.00005467, -0.00001373, 0.00003919, -0.00002873, -0.00001152, 0.00003304, -0.00001532, -0.00001689, 0.00002034, -0.00000475, -0.00001207, 0.00001421, 0.00000109, -0.00001115, 0.00000573, 0.00000571, -0.00000894, -0.00000993, 0.00000165, 0.00000734, -0.00000366, -0.00000587, 0.00001049, 0.00001189, 0.00000268, 0.00000418, -0.00000427, -0.00001633, -0.00000452, 0.00000896, 0.00000668, 0.00000129, -0.00000878, -0.00001347, -0.00000187, 0.00000212, -0.00000160, 0.00000575, 0.00000281, -0.00000981, 0.00000265, 0.00001713, -0.00000004, -0.00001496, -0.00000613, -0.00000084, -0.00000287, 0.00000213, 0.00000256, -0.00000990, -0.00001513, },
		{ -0.00001443, -0.00000934, 0.00000979, 0.00001221, -0.00000333, -0.00000597, 0.00000345, 0.00000384, 0.00000031, -0.00000340, -0.00001191, -0.00001035, 0.00000356, 0.00000909, 0.00000294, -0.00000703, -0.00001302, -0.00000495, 0.00000508, 0.00000622, 0.00001128, 0.00000917, -0.00000710, -0.00000308, 0.00001127, -0.00000263, -0.00001379, 0.00000080, 0.00000459, -0.00000352, -0.00000118, 0.00000263, 0.00000205, -0.00000069, -0.00000047, 0.00000394, -0.00000562, -0.00000945, 0.00001082, 0.00000221, -0.00001388, 0.00001343, -0.00000270, -0.00003681, 0.00001991, 0.00001769, -0.00005669, 0.00001693, 0.00005022, -0.00007636, -0.00000735, 0.00008354, -0.00009376, -0.00003750, 0.00013448, -0.00014994, -0.00013092, 0.00029544, -0.00023674, -0.00076883, 0.00046886, 0.00135263, 0.00005738, -0.00081069, -0.00000137, 0.00025554, -0.00023570, -0.00005654, 0.00015796, -0.00011383, -0.00005974, 0.00011555, -0.00006134, -0.00006582, 0.00008896, -0.00001857, -0.00006527, 0.00005435, 0.00000425, -0.00005521, 0.00002358, 0.00001189, -0.00003328, 0.00001648, 0.00001529, -0.00002503, 0.00000272, 0.00001277, -0.00001262, 0.00000066, 0.00001104, -0.00000614, -0.00000443, 0.00000666, 0.00000088, -0.00000984, -0.00000804, 0.00000581, 0.00000559, -0.00000816, -0.00000025, 0.00001434, 0.00000737, 0.00000312, 0.00000339, -0.00001075, -0.00001406, 0.00000157, 0.00000857, 0.00000583, -0.00000198, -0.00001332, -0.00000945, 0.00000150, -0.00000030, 0.00000169, 0.00000644, -0.00000496, -0.00000637, 0.00001280, 0.00001129, -0.00001024, -0.00001258, -0.00000212, -0.00000121, -0.00000095, 0.00000298, -0.00000289, },
		{ 0.00000106, -0.00000883, -0.00001468, 0.00000038, 0.00001488, 0.00000420, -0.00000779, -0.00000022, 0.00000508, 0.00000112, -0.00000003, -0.00000640, -0.00001396, -0.00000518, 0.00000728, 0.00000777, 0.00000025, -0.00001097, -0.00001204, 0.00000068, 0.00000552, 0.00000668, 0.00001343, 0.00000312, -0.00000963, 0.00000429, 0.00000893, -0.00000982, -0.00000961, 0.00000390, 0.00000193, -0.00000313, -0.00000095, 0.00000356, 0.00000246, -0.00000410, 0.00000251, 0.00000432, -0.00001357, -0.00000112, 0.00001638, -0.00001297, -0.00000576, 0.00002142, -0.00002836, -0.00002266, 0.00004403, -0.00002335, -0.00004866, 0.00006482, -0.00000465, -0.00008809, 0.00007000, 0.00002168, -0.00013359, 0.00008067, 0.00005706, -0.00025004, 0.00009259, 0.00021848, -0.00064251, -0.00041939, 0.00112106, 0.00097222, -0.00056910, -0.00055158, 0.00028252, 0.00002354, -0.00024789, 0.00010609, 0.00005240, -0.00014939, 0.00005297, 0.00005840, -0.00010685, 0.00001833, 0.00006501, -0.00006831, -0.00001001, 0.00005232, -0.00003844, -0.00002405, 0.00003200, -0.00001583, -0.00001450, 0.00002511, -0.00000664, -0.00001680, 0.00001199, 0.00000056, -0.00000889, 0.00000744, 0.00000377, -0.00000624, -0.00000007, 0.00000499, -0.00000166, -0.00001144, -0.00000435, 0.00000969, -0.00000002, -0.00000895, 0.00000741, 0.00001339, 0.00000422, 0.00000491, -0.00000122, -0.00001472, -0.00000870, 0.00000485, 0.00000819, 0.00000452, -0.00000755, -0.00001415, -0.00000401, 0.00000129, -0.00000026, 0.00000535, 0.00000141, -0.00000873, 0.00000328, 0.00001592, 0.00000035, -0.00001466, -0.00000762, -0.00000005, -0.00000104, 0.00000059, },
		{ -0.00000082, 0.00000083, -0.00000271, -0.00001345, -0.00000957, 0.00001053, 0.00001234, -0.00000485, -0.00000533, 0.00000493, 0.00000267, 0.00000018, -0.00000069, -0.00001092, -0.00001250, 0.00000070, 0.00000833, 0.00000636, -0.00000341, -0.00001436, -0.00000753, 0.00000456, 0.00000420, 0.00000962, 0.00001268, -0.00000435, -0.00000661, 0.00000958, 0.00000227, -0.00001251, -0.00000471, 0.00000430, 0.00000100, -0.00000386, -0.00000051, 0.00000667, -0.00000140, -0.00000554, 0.00000921, -0.00000316, -0.00001687, 0.00001459, 0.00000755, -0.00002393, 0.00001523, 0.00000891, -0.00004797, 0.00001535, 0.00003590, -0.00006474, 0.00000206, 0.00007173, -0.00007605, -0.00003300, 0.00010408, -0.00008241, -0.00007647, 0.00015488, -0.00010912, -0.00019112, 0.00029060, -0.00011625, -0.00081639, 0.00026341, 0.00138562, 0.00026341, -0.00081639, -0.00011625, 0.00029060, -0.00019112, -0.00010912, 0.00015488, -0.00007647, -0.00008241, 0.00010408, -0.00003300, -0.00007605, 0.00007173, 0.00000206, -0.00006474, 0.00003590, 0.00001535, -0.00004797, 0.00000891, 0.00001523, -0.00002393, 0.00000755, 0.00001459, -0.00001687, -0.00000316, 0.00000921, -0.00000554, -0.00000140, 0.00000667, -0.00000051, -0.00000386, 0.00000100, 0.00000430, -0.00000471, -0.00001251, 0.00000227, 0.00000958, -0.00000661, -0.00000435, 0.00001268, 0.00000962, 0.00000420, 0.00000456, -0.00000753, -0.00001436, -0.00000341, 0.00000636, 0.00000833, 0.00000070, -0.00001250, -0.00001092, -0.00000069, 0.00000018, 0.00000267, 0.00000493, -0.00000533, -0.00000485, 0.00001234, 0.00001053, -0.00000957, -0.00001345, -0.00000271, 0.00000083, },
		{ 0.00000106, 0.00000059, -0.00000104, -0.00000005, -0.00000762, -0.00001466, 0.00000035, 0.00001592, 0.00000328, -0.00000873, 0.00000141, 0.00000535, -0.00000026, 0.00000129, -0.00000401, -0.00001415, -0.00000755, 0.00000452, 0.00000819, 0.00000485, -0.00000870, -0.00001472, -0.00000122, 0.00000491, 0.00000422, 0.00001339, 0.00000741, -0.00000895, -0.00000002, 0.00000969, -0.00000435, -0.00001144, -0.00000166, 0.00000499, -0.00000007, -0.00000624, 0.00000377, 0.00000744, -0.00000889, 0.00000056, 0.00001199, -0.00001680, -0.00000664, 0.00002511, -0.00001450, -0.00001583, 0.00003200, -0.00002405, -0.00003844, 0.00005232, -0.00001001, -0.00006831, 0.00006501, 0.00001833, -0.00010685, 0.00005840, 0.00005297, -0.00014939, 0.00005240, 0.00010609, -0.00024789, 0.00002354, 0.00028252, -0.00055158, -0.00056910, 0.00097222, 0.00112106, -0.00041939, -0.00064251, 0.00021848, 0.00009259, -0.00025004, 0.00005706, 0.00008067, -0.00013359, 0.00002168, 0.00007000, -0.00008809, -0.00000465, 0.00006482, -0.00004866, -0.00002335, 0.00004403, -0.00002266, -0.00002836, 0.00002142, -0.00000576, -0.00001297, 0.00001638, -0.00000112, -0.00001357, 0.00000432, 0.00000251, -0.00000410, 0.00000246, 0.00000356, -0.00000095, -0.00000313, 0.00000193, 0.00000390, -0.00000961, -0.00000982, 0.00000893, 0.00000429, -0.00000963, 0.00000312, 0.00001343, 0.00000668, 0.00000552, 0.00000068, -0.00001204, -0.00001097, 0.00000025, 0.00000777, 0.00000728, -0.00000518, -0.00001396, -0.00000640, -0.00000003, 0.00000112, 0.00000508, -0.00000022, -0.00000779, 0.00000420, 0.00001488, 0.00000038, -0.00001468, -0.00000883, },
		{ -0.00001443, -0.00000289, 0.00000298, -0.00000095, -0.00000121, -0.00000212, -0.00001258, -0.00001024, 0.00001129, 0.00001280, -0.00000637, -0.00000496, 0.00000644, 0.00000169, -0.00000030, 0.00000150, -0.00000945, -0.00001332, -0.00000198, 0.00000583, 0.00000857, 0.00000157, -0.00001406, -0.00001075, 0.00000339, 0.00000312, 0.00000737, 0.00001434, -0.00000025, -0.00000816, 0.00000559, 0.00000581, -0.00000804, -0.00000984, 0.00000088, 0.00000666, -0.00000443, -0.00000614, 0.00001104, 0.00000066, -0.00001262, 0.00001277, 0.00000272, -0.00002503, 0.00001529, 0.00001648, -0.00003328, 0.00001189, 0.00002358, -0.00005521, 0.00000425, 0.00005435, -0.00006527, -0.00001857, 0.00008896, -0.00006582, -0.00006134, 0.00011555, -0.00005974, -0.00011383, 0.00015796, -0.00005654, -0.00023570, 0.00025554, -0.00000137, -0.00081069, 0.00005738, 0.00135263, 0.00046886, -0.00076883, -0.00023674, 0.00029544, -0.00013092, -0.00014994, 0.00013448, -0.00003750, -0.00009376, 0.00008354, -0.00000735, -0.00007636, 0.00005022, 0.00001693, -0.00005669, 0.00001769, 0.00001991, -0.00003681, -0.00000270, 0.00001343, -0.00001388, 0.00000221, 0.00001082, -0.00000945, -0.00000562, 0.00000394, -0.00000047, -0.00000069, 0.00000205, 0.00000263, -0.00000118, -0.00000352, 0.00000459, 0.00000080, -0.00001379, -0.00000263, 0.00001127, -0.00000308, -0.00000710, 0.00000917, 0.00001128, 0.00000622, 0.00000508, -0.00000495, -0.00001302, -0.00000703, 0.00000294, 0.00000909, 0.00000356, -0.00001035, -0.00001191, -0.00000340, 0.00000031, 0.00000384, 0.00000345, -0.00000597, -0.00000333, 0.00001221, 0.00000979, -0.00000934, },
		{ 0.00000022, -0.00001513, -0.00000990, 0.00000256, 0.00000213, -0.00000287, -0.00000084, -0.00000613, -0.00001496, -0.00000004, 0.00001713, 0.00000265, -0.00000981, 0.00000281, 0.00000575, -0.00000160, 0.00000212, -0.00000187, -0.00001347, -0.00000878, 0.00000129, 0.00000668, 0.00000896, -0.00000452, -0.00001633, -0.00000427, 0.00000418, 0.00000268, 0.00001189, 0.00001049, -0.00000587, -0.00000366, 0.00000734, 0.00000165, -0.00000993, -0.00000894, 0.00000571, 0.00000573, -0.00001115, 0.00000109, 0.00001421, -0.00001207, -0.00000475, 0.00002034, -0.00001689, -0.00001532, 0.00003304, -0.00001152, -0.00002873, 0.00003919, -0.00001373, -0.00005467, 0.00005385, 0.00000887, -0.00008420, 0.00005613, 0.00004393, -0.00011719, 0.00003728, 0.00008260, -0.00015178, 0.00001423, 0.00014455, -0.00022338, -0.00004780, 0.00031294, -0.00043684, -0.00067177, 0.00078571, 0.00121634, -0.00023587, -0.00069530, 0.00012706, 0.00014794, -0.00022919, 0.00000508, 0.00009456, -0.00010730, -0.00000609, 0.00007087, -0.00006448, -0.00002149, 0.00005640, -0.00002893, -0.00002960, 0.00003124, -0.00001009, -0.00002691, 0.00000996, 0.00000016, -0.00000818, 0.00000874, 0.00000095, -0.00000864, -0.00000135, 0.00000152, 0.00000077, -0.00000006, 0.00000145, 0.00000333, -0.00000347, -0.00000190, 0.00000684, -0.00000568, -0.00001334, 0.00000567, 0.00000775, -0.00000786, -0.00000109, 0.00001158, 0.00000931, 0.00000677, 0.00000162, -0.00000934, -0.00001134, -0.00000373, 0.00000562, 0.00000874, -0.00000201, -0.00001251, -0.00000877, -0.00000198, 0.00000238, 0.00000507, -0.00000194, -0.00000713, 0.00000526, 0.00001417, },
		{ 0.00001237, 0.00000925, -0.00000949, -0.00001567, -0.00000275, 0.00000536, -0.00000131, -0.00000323, -0.00000114, -0.00001165, -0.00001133, 0.00001183, 0.00001357, -0.00000770, -0.00000484, 0.00000782, 0.00000088, -0.00000093, 0.00000317, -0.00000784, -0.00001306, -0.00000391, 0.00000225, 0.00000893, 0.00000689, -0.00001158, -0.00001373, 0.00000090, 0.00000261, 0.00000558, 0.00001391, 0.00000389, -0.00000693, 0.00000046, 0.00000652, -0.00000103, -0.00001282, -0.00000583, 0.00001194, -0.00000186, -0.00001351, 0.00001364, 0.00000588, -0.00002036, 0.00001323, 0.00001163, -0.00003214, 0.00001188, 0.00002781, -0.00003969, 0.00000313, 0.00003886, -0.00005640, -0.00001209, 0.00006961, -0.00005737, -0.00004175, 0.00009860, -0.00004685, -0.00008779, 0.00011563, -0.00002930, -0.00014353, 0.00014356, -0.00000047, -0.00025841, 0.00019806, 0.00009387, -0.00075757, -0.00012868, 0.00125794, 0.00065247, -0.00066956, -0.00034664, 0.00026710, -0.00006465, -0.00017271, 0.00010027, -0.00000331, -0.00009259, 0.00005796, 0.00001177, -0.00006761, 0.00002841, 0.00002422, -0.00004349, 0.00000277, 0.00001808, -0.00002449, -0.00000971, 0.00000789, -0.00000522, 0.00000066, 0.00000554, -0.00000408, -0.00000511, -0.00000146, 0.00000195, 0.00000180, -0.00000153, 0.00000329, 0.00000227, -0.00000594, 0.00000265, 0.00000513, -0.00001217, -0.00000728, 0.00001033, 0.00000109, -0.00000773, 0.00000465, 0.00001146, 0.00000893, 0.00000576, -0.00000334, -0.00001099, -0.00000880, -0.00000069, 0.00000787, 0.00000578, -0.00000693, -0.00001196, -0.00000651, -0.00000011, 0.00000512, 0.00000219, -0.00000685, -0.00000203, },
		{ -0.00000682, 0.00000626, 0.00001382, 0.00000004, -0.00001599, -0.00001099, 0.00000441, 0.00000368, -0.00000487, -0.00000145, -0.00000432, -0.00001538, -0.00000085, 0.00001830, 0.00000240, -0.00001082, 0.00000391, 0.00000615, -0.00000280, 0.00000259, -0.00000010, -0.00001229, -0.00000891, -0.00000170, 0.00000361, 0.00001156, 0.00000093, -0.00001590, -0.00000799, 0.00000243, 0.00000229, 0.00000996, 0.00001142, -0.00000124, -0.00000512, 0.00000215, 0.00000634, -0.00000459, -0.00001629, 0.00000290, 0.00001341, -0.00001406, -0.00000476, 0.00002142, -0.00001229, -0.00001301, 0.00002760, -0.00001294, -0.00002596, 0.00003819, -0.00000377, -0.00004228, 0.00004100, 0.00000148, -0.00006821, 0.00004766, 0.00003037, -0.00009317, 0.00003916, 0.00006790, -0.00011710, 0.00000981, 0.00010564, -0.00014042, -0.00002757, 0.00016701, -0.00018113, -0.00011041, 0.00030938, -0.00031479, -0.00072130, 0.00058174, 0.00124910, -0.00003700, -0.00070052, 0.00001967, 0.00018008, -0.00018906, -0.00004164, 0.00009241, -0.00007549, -0.00002610, 0.00006199, -0.00004029, -0.00003014, 0.00004222, -0.00001250, -0.00002879, 0.00001695, -0.00000248, -0.00002117, 0.00000000, 0.00000172, -0.00000193, 0.00000350, 0.00000013, -0.00000358, -0.00000435, -0.00000138, 0.00000440, -0.00000020, -0.00000131, 0.00000577, -0.00000182, -0.00000509, 0.00000711, -0.00000118, -0.00001423, 0.00000114, 0.00000930, -0.00000436, -0.00000404, 0.00000806, 0.00001114, 0.00000890, 0.00000225, -0.00000738, -0.00001043, -0.00000621, 0.00000246, 0.00000831, 0.00000129, -0.00000958, -0.00001068, -0.00000463, 0.00000328, 0.00000537, -0.00000357, },
		{ 0.00000134, -0.00000785, -0.00000109, 0.00001269, 0.00000902, -0.00000986, -0.00001722, -0.00000244, 0.00000797, -0.00000174, -0.00000532, 0.00000009, -0.00001051, -0.00001268, 0.00001199, 0.00001448, -0.00000864, -0.00000485, 0.00000892, 0.00000021, -0.00000160, 0.00000439, -0.00000631, -0.00001202, -0.00000490, -0.00000160, 0.00000739, 0.00001137, -0.00000703, -0.00001527, -0.00000304, 0.00000181, 0.00000506, 0.00001194, 0.00000661, -0.00000287, -0.00000411, 0.00000358, 0.00000700, -0.00001213, -0.00001500, 0.00001503, 0.00000411, -0.00002117, 0.00001317, 0.00001379, -0.00002705, 0.00000975, 0.00002236, -0.00003641, 0.00000421, 0.00003959, -0.00004137, -0.00001013, 0.00005188, -0.00005056, -0.00003107, 0.00007870, -0.00004187, -0.00006357, 0.00009873, -0.00002207, -0.00010804, 0.00010420, 0.00000390, -0.00016108, 0.00011472, 0.00005060, -0.00025725, 0.00012895, 0.00015986, -0.00066890, -0.00027814, 0.00111358, 0.00079575, -0.00052765, -0.00043050, 0.00020833, -0.00000325, -0.00017460, 0.00005837, 0.00002083, -0.00008023, 0.00003214, 0.00002216, -0.00005236, 0.00000999, 0.00002383, -0.00002833, -0.00000682, 0.00001140, -0.00001364, -0.00001183, 0.00000064, 0.00000056, 0.00000219, 0.00000042, -0.00000135, -0.00000268, -0.00000567, 0.00000183, 0.00000476, -0.00000330, 0.00000217, 0.00000521, -0.00000613, -0.00000021, 0.00000740, -0.00000865, -0.00001038, 0.00000722, 0.00000456, -0.00000617, 0.00000031, 0.00000978, 0.00001157, 0.00000724, -0.00000245, -0.00000923, -0.00000886, -0.00000344, 0.00000499, 0.00000654, -0.00000265, -0.00001060, -0.00000970, -0.00000133, 0.00000637, },
		{ 0.00000351, 0.00000596, -0.00000488, -0.00000687, 0.00000703, 0.00001381, -0.00000000, -0.00001718, -0.00001220, 0.00000648, 0.00000535, -0.00000693, -0.00000202, -0.00000228, -0.00001570, -0.00000204, 0.00001917, 0.00000249, -0.00001155, 0.00000470, 0.00000640, -0.00000383, 0.00000281, 0.00000129, -0.00001083, -0.00000818, -0.00000399, -0.00000027, 0.00001214, 0.00000646, -0.00001290, -0.00001133, -0.00000109, 0.00000257, 0.00000876, 0.00000995, 0.00000349, -0.00000337, -0.00000465, 0.00000798, 0.00000403, -0.00002168, -0.00000387, 0.00002139, -0.00001366, -0.00001322, 0.00002738, -0.00000880, -0.00002299, 0.00003202, -0.00000494, -0.00003674, 0.00003903, 0.00000792, -0.00005393, 0.00003646, 0.00001928, -0.00007630, 0.00003438, 0.00005092, -0.00009337, 0.00001678, 0.00008626, -0.00010658, -0.00001959, 0.00011850, -0.00011762, -0.00006634, 0.00017123, -0.00012858, -0.00015581, 0.00027697, -0.00020108, -0.00071926, 0.00038144, 0.00121866, 0.00015653, -0.00065548, -0.00008907, 0.00018341, -0.00013706, -0.00007598, 0.00007575, -0.00004400, -0.00003595, 0.00004626, -0.00001950, -0.00003038, 0.00002566, -0.00000169, -0.00002248, 0.00000407, -0.00000026, -0.00001336, -0.00000677, -0.00000024, 0.00000394, 0.00000118, -0.00000242, 0.00000032, -0.00000480, -0.00000485, 0.00000605, 0.00000131, -0.00000352, 0.00000621, 0.00000072, -0.00000674, 0.00000541, 0.00000251, -0.00001266, -0.00000325, 0.00000860, -0.00000029, -0.00000500, 0.00000370, 0.00001138, 0.00001160, 0.00000335, -0.00000634, -0.00000918, -0.00000684, -0.00000064, 0.00000586, 0.00000383, -0.00000523, -0.00001153, -0.00000784, },
		{ -0.00001241, -0.00000348, 0.00000726, 0.00000103, -0.00000878, -0.00000064, 0.00001301, 0.00000915, -0.00001030, -0.00001904, -0.00000210, 0.00001074, -0.00000210, -0.00000740, 0.00000137, -0.00000915, -0.00001403, 0.00001166, 0.00001527, -0.00000908, -0.00000479, 0.00000963, -0.00000042, -0.00000220, 0.00000518, -0.00000493, -0.00001048, -0.00000506, -0.00000504, 0.00000447, 0.00001400, -0.00000123, -0.00001431, -0.00000784, -0.00000038, 0.00000602, 0.00000954, 0.00000683, 0.00000312, -0.00000636, -0.00000287, 0.00001400, -0.00000698, -0.00002496, 0.00001442, 0.00001285, -0.00002733, 0.00000897, 0.00002315, -0.00003100, 0.00000237, 0.00003309, -0.00003645, -0.00000686, 0.00004957, -0.00003744, -0.00002586, 0.00006007, -0.00003817, -0.00004942, 0.00007983, -0.00002114, -0.00008031, 0.00008931, 0.00000445, -0.00011898, 0.00008335, 0.00003448, -0.00016449, 0.00007690, 0.00008969, -0.00023455, 0.00005973, 0.00019270, -0.00056003, -0.00038071, 0.00093718, 0.00088591, -0.00035825, -0.00047644, 0.00012732, 0.00004312, -0.00015682, 0.00001623, 0.00003190, -0.00006026, 0.00001049, 0.00002361, -0.00003423, -0.00000244, 0.00001733, -0.00001441, -0.00001047, 0.00000224, -0.00000619, -0.00000991, -0.00000622, 0.00000294, 0.00000543, -0.00000325, -0.00000105, 0.00000036, -0.00000794, -0.00000008, 0.00000702, -0.00000317, 0.00000029, 0.00000683, -0.00000459, -0.00000274, 0.00000747, -0.00000454, -0.00001130, 0.00000302, 0.00000637, -0.00000307, -0.00000275, 0.00000653, 0.00001320, 0.00000967, -0.00000157, -0.00000837, -0.00000793, -0.00000453, 0.00000131, 0.00000527, 0.00000170, -0.00000755, },
		{ 0.00000002, -0.00001077, -0.00001127, 0.00000274, 0.00000667, -0.00000561, -0.00000719, 0.00000738, 0.00001403, 0.00000019, -0.00001855, -0.00001355, 0.00000858, 0.00000712, -0.00000886, -0.00000261, -0.00000019, -0.00001577, -0.00000339, 0.00001951, 0.00000275, -0.00001177, 0.00000527, 0.00000636, -0.00000468, 0.00000286, 0.00000226, -0.00000923, -0.00000686, -0.00000549, -0.00000413, 0.00001077, 0.00001078, -0.00000753, -0.00001293, -0.00000648, 0.00000249, 0.00000924, 0.00000665, 0.00000668, 0.00000174, -0.00001122, 0.00000543, 0.00001442, -0.00002327, -0.00001398, 0.00002767, -0.00000930, -0.00002296, 0.00003046, -0.00000171, -0.00003293, 0.00003232, 0.00000619, -0.00004557, 0.00003482, 0.00002175, -0.00006137, 0.00002592, 0.00003670, -0.00007718, 0.00001612, 0.00006701, -0.00008468, -0.00000726, 0.00009614, -0.00008761, -0.00004624, 0.00011963, -0.00008767, -0.00009644, 0.00015838, -0.00007429, -0.00017938, 0.00022480, -0.00010799, -0.00067393, 0.00020353, 0.00113247, 0.00032515, -0.00056486, -0.00018394, 0.00015755, -0.00008279, -0.00009354, 0.00004898, -0.00001824, -0.00003561, 0.00002767, -0.00000492, -0.00002377, 0.00001022, 0.00000271, -0.00001318, -0.00000520, -0.00000259, -0.00000583, -0.00000970, -0.00000420, 0.00000797, 0.00000145, -0.00000533, 0.00000235, -0.00000344, -0.00000770, 0.00000564, 0.00000343, -0.00000445, 0.00000513, 0.00000303, -0.00000660, 0.00000284, 0.00000454, -0.00000947, -0.00000634, 0.00000603, 0.00000313, -0.00000393, -0.00000039, 0.00000973, 0.00001406, 0.00000547, -0.00000584, -0.00000849, -0.00000592, -0.00000260, 0.00000184, 0.00000470, },
		{ 0.00000191, 0.00000509, -0.00000292, -0.00001395, -0.00000646, 0.00000741, 0.00000128, -0.00000938, -0.00000071, 0.00001316, 0.00000961, -0.00001069, -0.00002102, -0.00000183, 0.00001348, -0.00000230, -0.00000932, 0.00000253, -0.00000768, -0.00001512, 0.00001093, 0.00001564, -0.00000904, -0.00000441, 0.00000992, -0.00000111, -0.00000263, 0.00000558, -0.00000376, -0.00000856, -0.00000463, -0.00000775, 0.00000102, 0.00001430, 0.00000447, -0.00001029, -0.00001227, -0.00000480, 0.00000795, 0.00000810, 0.00000418, 0.00000947, -0.00000501, -0.00001164, 0.00001791, 0.00000233, -0.00003349, 0.00000930, 0.00002284, -0.00003035, 0.00000126, 0.00003220, -0.00003105, -0.00000796, 0.00004181, -0.00003173, -0.00001963, 0.00005571, -0.00002818, -0.00004142, 0.00006177, -0.00002109, -0.00006391, 0.00007275, 0.00000137, -0.00008926, 0.00007214, 0.00002842, -0.00011938, 0.00005686, 0.00005790, -0.00015448, 0.00003667, 0.00011260, -0.00019611, 0.00000039, 0.00019425, -0.00044697, -0.00043352, 0.00074888, 0.00091772, -0.00018005, -0.00047833, 0.00003625, 0.00006724, -0.00012444, -0.00001904, 0.00002981, -0.00003756, -0.00000395, 0.00001784, -0.00001696, -0.00000791, 0.00000736, -0.00000420, -0.00000899, -0.00000679, -0.00000297, -0.00000564, -0.00001104, 0.00000233, 0.00000884, -0.00000482, -0.00000238, 0.00000285, -0.00000816, -0.00000273, 0.00000783, -0.00000165, -0.00000132, 0.00000690, -0.00000223, -0.00000409, 0.00000589, -0.00000107, -0.00001015, -0.00000105, 0.00000618, 0.00000051, -0.00000389, 0.00000252, 0.00001311, 0.00001261, 0.00000006, -0.00000843, -0.00000700, -0.00000376, -0.00000203, },
		{ -0.00000271, -0.00000277, 0.00000334, 0.00000525, -0.00000807, -0.00001427, 0.00000075, 0.00000717, -0.00000558, -0.00000758, 0.00000718, 0.00001435, 0.00000064, -0.00001997, -0.00001497, 0.00001057, 0.00000886, -0.00001049, -0.00000313, 0.00000170, -0.00001557, -0.00000461, 0.00001923, 0.00000290, -0.00001134, 0.00000581, 0.00000588, -0.00000536, 0.00000287, 0.00000275, -0.00000752, -0.00000510, -0.00000643, -0.00000734, 0.00000811, 0.00001281, -0.00000079, -0.00001156, -0.00001310, 0.00000077, 0.00001169, 0.00000276, 0.00000702, 0.00000939, -0.00001552, -0.00000140, 0.00002436, -0.00001990, -0.00002585, 0.00003043, -0.00000128, -0.00003210, 0.00002960, 0.00000801, -0.00004089, 0.00002805, 0.00001878, -0.00005064, 0.00002594, 0.00003537, -0.00006306, 0.00001098, 0.00005076, -0.00007055, -0.00000401, 0.00007608, -0.00006868, -0.00002900, 0.00009642, -0.00006367, -0.00006625, 0.00010978, -0.00005579, -0.00011431, 0.00013248, -0.00002612, -0.00018087, 0.00016377, -0.00004267, -0.00059831, 0.00006158, 0.00100448, 0.00045357, -0.00043988, -0.00025194, 0.00010746, -0.00003605, -0.00009353, 0.00001826, -0.00000194, -0.00002727, 0.00001034, 0.00000237, -0.00001310, -0.00000136, 0.00000146, -0.00000371, -0.00000984, -0.00000767, -0.00000044, -0.00000920, -0.00000837, 0.00000951, 0.00000333, -0.00000745, 0.00000249, -0.00000129, -0.00000918, 0.00000370, 0.00000513, -0.00000391, 0.00000337, 0.00000438, -0.00000505, 0.00000047, 0.00000478, -0.00000580, -0.00000756, 0.00000255, 0.00000503, -0.00000145, -0.00000318, 0.00000650, 0.00001536, 0.00000864, -0.00000512, -0.00000888, -0.00000431, },
		{ -0.00000697, -0.00000159, -0.00000390, -0.00000294, 0.00000655, 0.00000270, -0.00001364, -0.00000990, 0.00000639, 0.00000194, -0.00000934, -0.00000124, 0.00001294, 0.00001034, -0.00001095, -0.00002302, -0.00000162, 0.00001600, -0.00000237, -0.00001085, 0.00000354, -0.00000634, -0.00001578, 0.00001005, 0.00001534, -0.00000871, -0.00000346, 0.00000985, -0.00000204, -0.00000282, 0.00000564, -0.00000291, -0.00000628, -0.00000386, -0.00000978, -0.00000208, 0.00001248, 0.00000860, -0.00000348, -0.00001477, -0.00001165, 0.00000965, 0.00000873, -0.00000079, 0.00001432, 0.00000035, -0.00002078, 0.00001724, 0.00001438, -0.00003844, -0.00000004, 0.00003204, -0.00002918, -0.00000884, 0.00003908, -0.00002684, -0.00001959, 0.00004684, -0.00002268, -0.00003194, 0.00005670, -0.00001501, -0.00005410, 0.00005660, -0.00000214, -0.00007210, 0.00005885, 0.00002198, -0.00008930, 0.00005043, 0.00004630, -0.00011002, 0.00002922, 0.00007135, -0.00013410, 0.00000026, 0.00011845, -0.00014967, -0.00004231, 0.00017099, -0.00034351, -0.00044075, 0.00056795, 0.00089397, -0.00001215, -0.00043687, -0.00005103, 0.00006628, -0.00008522, -0.00004227, 0.00001721, -0.00001711, -0.00001014, 0.00000790, -0.00000350, -0.00000709, -0.00000305, 0.00000109, -0.00000424, -0.00001360, -0.00000360, -0.00000097, -0.00001301, -0.00000016, 0.00001116, -0.00000438, -0.00000426, 0.00000407, -0.00000684, -0.00000508, 0.00000703, 0.00000041, -0.00000199, 0.00000576, -0.00000000, -0.00000404, 0.00000360, 0.00000102, -0.00000766, -0.00000395, 0.00000438, 0.00000346, -0.00000310, -0.00000116, 0.00001103, 0.00001523, 0.00000293, -0.00000891, },
		{ -0.00000336, -0.00001027, -0.00000308, -0.00000089, -0.00000668, -0.00000024, 0.00000932, -0.00000352, -0.00001603, -0.00000242, 0.00000697, -0.00000470, -0.00000774, 0.00000634, 0.00001459, 0.00000137, -0.00002134, -0.00001634, 0.00001236, 0.00001038, -0.00001173, -0.00000339, 0.00000321, -0.00001524, -0.00000538, 0.00001843, 0.00000256, -0.00001027, 0.00000662, 0.00000489, -0.00000598, 0.00000298, 0.00000272, -0.00000580, -0.00000290, -0.00000710, -0.00000969, 0.00000519, 0.00001211, 0.00000577, -0.00000657, -0.00001956, -0.00000362, 0.00001558, -0.00000012, 0.00000393, 0.00001791, -0.00001593, -0.00001148, 0.00003152, -0.00001153, -0.00003720, 0.00002847, 0.00000912, -0.00003876, 0.00002459, 0.00001869, -0.00004532, 0.00001972, 0.00003073, -0.00005084, 0.00001371, 0.00004640, -0.00005863, -0.00000584, 0.00005915, -0.00005758, -0.00002261, 0.00007699, -0.00004831, -0.00004512, 0.00008781, -0.00003888, -0.00007729, 0.00009163, -0.00002696, -0.00011901, 0.00009932, 0.00001021, -0.00016386, 0.00010435, -0.00000667, -0.00050730, -0.00003757, 0.00085240, 0.00053299, -0.00029622, -0.00028464, 0.00004243, -0.00000470, -0.00007882, -0.00000996, 0.00000361, -0.00001465, -0.00000253, 0.00000310, -0.00000158, -0.00000765, -0.00000353, 0.00000358, -0.00001013, -0.00001336, 0.00000184, -0.00000648, -0.00001131, 0.00000868, 0.00000561, -0.00000812, 0.00000127, 0.00000064, -0.00000907, 0.00000107, 0.00000572, -0.00000229, 0.00000179, 0.00000452, -0.00000287, -0.00000094, 0.00000367, -0.00000270, -0.00000703, -0.00000075, 0.00000513, 0.00000144, -0.00000406, 0.00000255, 0.00001478, 0.00001234, },
		{ 0.00001654, 0.00000701, -0.00000892, -0.00000827, 0.00000098, -0.00000364, -0.00000816, 0.00000548, 0.00000830, -0.00001108, -0.00001306, 0.00000388, 0.00000262, -0.00000839, -0.00000203, 0.00001218, 0.00001122, -0.00001103, -0.00002494, -0.00000139, 0.00001815, -0.00000246, -0.00001181, 0.00000447, -0.00000541, -0.00001599, 0.00000937, 0.00001425, -0.00000844, -0.00000177, 0.00000962, -0.00000338, -0.00000274, 0.00000549, -0.00000254, -0.00000363, -0.00000280, -0.00001147, -0.00000419, 0.00000941, 0.00000997, 0.00000500, -0.00001388, -0.00002027, 0.00000963, 0.00001192, -0.00000675, 0.00001606, 0.00000905, -0.00002801, 0.00001154, 0.00002692, -0.00003826, -0.00001226, 0.00003836, -0.00002371, -0.00001958, 0.00004234, -0.00001894, -0.00003052, 0.00004720, -0.00001066, -0.00004165, 0.00005228, -0.00000013, -0.00006182, 0.00004553, 0.00001558, -0.00007286, 0.00004075, 0.00003757, -0.00008107, 0.00002798, 0.00005598, -0.00009343, 0.00000468, 0.00007415, -0.00010786, -0.00002762, 0.00010945, -0.00010315, -0.00006564, 0.00013220, -0.00025933, -0.00041206, 0.00040992, 0.00082433, 0.00012914, -0.00035929, -0.00012166, 0.00004238, -0.00004786, -0.00005126, -0.00000117, -0.00000279, -0.00000910, -0.00000271, 0.00000456, -0.00000195, -0.00001129, 0.00000161, 0.00000145, -0.00001700, -0.00000681, 0.00000242, -0.00001226, -0.00000311, 0.00001170, -0.00000261, -0.00000571, 0.00000389, -0.00000477, -0.00000640, 0.00000498, 0.00000214, -0.00000153, 0.00000411, 0.00000145, -0.00000287, 0.00000152, 0.00000161, -0.00000483, -0.00000521, 0.00000182, 0.00000498, -0.00000098, -0.00000349, 0.00000736, },
		{ -0.00000101, 0.00001214, 0.00001564, -0.00000009, -0.00001196, -0.00000313, 0.00000224, -0.00000873, -0.00000534, 0.00001115, 0.00000224, -0.00001575, -0.00000635, 0.00000555, -0.00000312, -0.00000731, 0.00000490, 0.00001455, 0.00000233, -0.00002259, -0.00001756, 0.00001401, 0.00001148, -0.00001264, -0.00000314, 0.00000429, -0.00001509, -0.00000552, 0.00001740, 0.00000145, -0.00000878, 0.00000799, 0.00000342, -0.00000671, 0.00000335, 0.00000215, -0.00000425, -0.00000016, -0.00000773, -0.00001137, 0.00000301, 0.00000901, 0.00001045, 0.00000179, -0.00002406, -0.00001108, 0.00001964, -0.00000063, -0.00000216, 0.00002525, -0.00001171, -0.00002284, 0.00003401, 0.00000067, -0.00004565, 0.00002163, 0.00001991, -0.00004152, 0.00001615, 0.00002839, -0.00004537, 0.00000868, 0.00004008, -0.00004604, 0.00000017, 0.00005301, -0.00004892, -0.00002175, 0.00006077, -0.00004063, -0.00003675, 0.00007024, -0.00002709, -0.00005364, 0.00007267, -0.00001707, -0.00007892, 0.00006905, -0.00000493, -0.00011207, 0.00006503, 0.00003199, -0.00013456, 0.00005467, 0.00000346, -0.00041496, -0.00009405, 0.00069458, 0.00056201, -0.00015120, -0.00027962, -0.00002578, 0.00000692, -0.00005522, -0.00003041, -0.00000032, -0.00000198, -0.00000933, -0.00000060, 0.00000799, -0.00000871, -0.00000978, 0.00000732, -0.00000739, -0.00001776, 0.00000105, -0.00000313, -0.00001220, 0.00000630, 0.00000722, -0.00000729, -0.00000048, 0.00000170, -0.00000771, -0.00000137, 0.00000503, -0.00000030, 0.00000097, 0.00000374, -0.00000083, -0.00000119, 0.00000204, -0.00000082, -0.00000540, -0.00000297, 0.00000381, 0.00000370, -0.00000308, },
		{ 0.00000150, -0.00000395, 0.00000300, 0.00001584, 0.00001167, -0.00000759, -0.00001068, 0.00000277, -0.00000127, -0.00001234, 0.00000194, 0.00001259, -0.00000633, -0.00001503, -0.00000011, 0.00000276, -0.00000639, -0.00000275, 0.00001074, 0.00001209, -0.00001090, -0.00002671, -0.00000103, 0.00001992, -0.00000279, -0.00001214, 0.00000558, -0.00000511, -0.00001596, 0.00000926, 0.00001249, -0.00000864, 0.00000069, 0.00000955, -0.00000528, -0.00000244, 0.00000533, -0.00000284, -0.00000068, -0.00000131, -0.00001325, -0.00000520, 0.00000627, 0.00000809, 0.00001338, -0.00000876, -0.00002919, 0.00000659, 0.00001723, -0.00001194, 0.00001384, 0.00001944, -0.00003133, 0.00000157, 0.00003740, -0.00003247, -0.00002516, 0.00004024, -0.00001485, -0.00002897, 0.00004132, -0.00000875, -0.00003891, 0.00004286, 0.00000238, -0.00004716, 0.00004324, 0.00001396, -0.00006351, 0.00003066, 0.00002932, -0.00006656, 0.00002168, 0.00004623, -0.00006669, 0.00000835, 0.00005711, -0.00007313, -0.00001363, 0.00006768, -0.00008061, -0.00004459, 0.00009005, -0.00006306, -0.00007080, 0.00008776, -0.00019891, -0.00036011, 0.00028455, 0.00072319, 0.00023275, -0.00025781, -0.00016613, 0.00000207, -0.00002000, -0.00004720, -0.00001988, 0.00000345, -0.00000348, -0.00001100, 0.00000725, 0.00000489, -0.00001578, -0.00000130, 0.00000601, -0.00001696, -0.00001087, 0.00000358, -0.00000967, -0.00000529, 0.00001046, -0.00000045, -0.00000616, 0.00000266, -0.00000278, -0.00000649, 0.00000239, 0.00000295, -0.00000026, 0.00000268, 0.00000193, -0.00000125, 0.00000028, 0.00000112, -0.00000259, -0.00000496, -0.00000050, 0.00000486, },
		{ 0.00000187, 0.00000462, -0.00000085, -0.00000320, 0.00000787, 0.00001748, 0.00000457, -0.00001296, -0.00000487, 0.00000552, -0.00000825, -0.00001068, 0.00001010, 0.00000807, -0.00001294, -0.00001027, 0.00000248, -0.00000125, -0.00000594, 0.00000305, 0.00001401, 0.00000345, -0.00002366, -0.00001857, 0.00001561, 0.00001204, -0.00001342, -0.00000219, 0.00000508, -0.00001545, -0.00000501, 0.00001658, -0.00000060, -0.00000725, 0.00001017, 0.00000167, -0.00000780, 0.00000410, 0.00000116, -0.00000317, 0.00000325, -0.00000831, -0.00001293, 0.00000222, 0.00000452, 0.00001188, 0.00001232, -0.00002488, -0.00002104, 0.00002217, 0.00000199, -0.00000995, 0.00002951, -0.00000326, -0.00003302, 0.00003095, 0.00001456, -0.00004927, 0.00001092, 0.00002893, -0.00003982, 0.00000579, 0.00003539, -0.00004113, -0.00000321, 0.00004538, -0.00003708, -0.00001243, 0.00005429, -0.00003574, -0.00003425, 0.00005583, -0.00002267, -0.00004461, 0.00005772, -0.00000834, -0.00005421, 0.00005426, -0.00000103, -0.00007251, 0.00004614, 0.00000842, -0.00009687, 0.00003486, 0.00003956, -0.00010009, 0.00001940, -0.00000539, -0.00033231, -0.00011376, 0.00054701, 0.00054616, -0.00002077, -0.00024061, -0.00008538, -0.00000117, -0.00002993, -0.00004020, -0.00001035, 0.00000713, -0.00001029, -0.00000594, 0.00001388, -0.00000584, -0.00001500, 0.00000733, -0.00000343, -0.00001973, -0.00000189, -0.00000055, -0.00001102, 0.00000348, 0.00000754, -0.00000547, -0.00000196, 0.00000172, -0.00000574, -0.00000298, 0.00000335, 0.00000134, 0.00000107, 0.00000263, 0.00000050, -0.00000057, 0.00000069, -0.00000023, -0.00000360, -0.00000371, },
		{ -0.00000384, -0.00000178, 0.00000350, 0.00000333, -0.00000259, -0.00000091, 0.00001294, 0.00001584, -0.00000443, -0.00001338, 0.00000279, 0.00000244, -0.00001424, -0.00000329, 0.00001438, -0.00000007, -0.00001488, -0.00000519, 0.00000172, -0.00000347, -0.00000296, 0.00000863, 0.00001272, -0.00001055, -0.00002830, -0.00000047, 0.00002138, -0.00000357, -0.00001195, 0.00000717, -0.00000544, -0.00001605, 0.00000995, 0.00001040, -0.00000971, 0.00000371, 0.00001004, -0.00000776, -0.00000212, 0.00000542, -0.00000389, 0.00000229, 0.00000091, -0.00001537, -0.00000547, 0.00000427, 0.00000333, 0.00001970, 0.00000045, -0.00003639, -0.00000011, 0.00002338, -0.00001469, 0.00000783, 0.00002934, -0.00002963, -0.00001083, 0.00004357, -0.00002186, -0.00003626, 0.00003708, -0.00000432, -0.00003531, 0.00003626, 0.00000191, -0.00004349, 0.00003471, 0.00001436, -0.00004775, 0.00003127, 0.00002506, -0.00005932, 0.00001463, 0.00003734, -0.00005489, 0.00000476, 0.00004756, -0.00004919, -0.00000588, 0.00005102, -0.00005283, -0.00002426, 0.00005482, -0.00005650, -0.00005064, 0.00006565, -0.00003347, -0.00006197, 0.00004616, -0.00016183, -0.00029797, 0.00019523, 0.00060677, 0.00029412, -0.00014719, -0.00017997, -0.00004526, -0.00000664, -0.00003409, -0.00003407, 0.00000196, 0.00000336, -0.00001518, 0.00000595, 0.00001101, -0.00001621, -0.00000570, 0.00000811, -0.00001436, -0.00001419, 0.00000239, -0.00000647, -0.00000596, 0.00000805, 0.00000123, -0.00000552, 0.00000103, -0.00000145, -0.00000558, 0.00000001, 0.00000266, 0.00000122, 0.00000195, 0.00000169, 0.00000013, 0.00000004, 0.00000028, -0.00000147, },
		{ -0.00000050, -0.00000241, -0.00000315, 0.00000025, 0.00000403, 0.00000166, -0.00000350, 0.00000293, 0.00001713, 0.00000991, -0.00001228, -0.00000796, 0.00000761, -0.00000535, -0.00001474, 0.00000622, 0.00001257, -0.00000760, -0.00001312, -0.00000231, 0.00000026, -0.00000340, 0.00000114, 0.00001277, 0.00000459, -0.00002450, -0.00001937, 0.00001729, 0.00001212, -0.00001435, -0.00000049, 0.00000591, -0.00001655, -0.00000413, 0.00001642, -0.00000346, -0.00000621, 0.00001325, -0.00000003, -0.00000948, 0.00000523, 0.00000002, -0.00000287, 0.00000728, -0.00000851, -0.00001493, 0.00000292, 0.00000010, 0.00000943, 0.00002308, -0.00002092, -0.00003200, 0.00002159, 0.00000750, -0.00001753, 0.00002946, 0.00000792, -0.00003974, 0.00002271, 0.00002751, -0.00004716, -0.00000174, 0.00003434, -0.00003411, -0.00000458, 0.00003862, -0.00003355, -0.00001399, 0.00004605, -0.00002559, -0.00002218, 0.00005042, -0.00002145, -0.00004170, 0.00004582, -0.00000659, -0.00004575, 0.00004222, 0.00000548, -0.00004805, 0.00003601, 0.00000800, -0.00006072, 0.00002632, 0.00001319, -0.00007758, 0.00001221, 0.00003574, -0.00006700, -0.00000051, -0.00002481, -0.00026597, -0.00010646, 0.00042114, 0.00049614, 0.00008306, -0.00017645, -0.00012685, -0.00002479, -0.00000985, -0.00003933, -0.00002212, 0.00001054, -0.00000714, -0.00001034, 0.00001570, -0.00000105, -0.00001767, 0.00000451, -0.00000002, -0.00001908, -0.00000564, 0.00000039, -0.00000843, 0.00000124, 0.00000654, -0.00000338, -0.00000268, 0.00000095, -0.00000384, -0.00000355, 0.00000128, 0.00000211, 0.00000179, 0.00000182, 0.00000100, 0.00000033, 0.00000013, },
		{ 0.00000048, -0.00000020, -0.00000145, -0.00000269, -0.00000173, 0.00000167, 0.00000386, -0.00000007, -0.00000338, 0.00000835, 0.00001841, 0.00000042, -0.00001519, 0.00000067, 0.00000616, -0.00001315, -0.00000879, 0.00001294, 0.00000646, -0.00001198, -0.00001051, -0.00000104, -0.00000005, -0.00000220, 0.00000599, 0.00001285, -0.00000998, -0.00002966, 0.00000030, 0.00002272, -0.00000488, -0.00001152, 0.00000946, -0.00000622, -0.00001668, 0.00001145, 0.00000853, -0.00001188, 0.00000685, 0.00001142, -0.00001068, -0.00000207, 0.00000602, -0.00000564, 0.00000486, 0.00000412, -0.00001777, -0.00000574, 0.00000419, -0.00000312, 0.00002233, 0.00001257, -0.00003989, -0.00001013, 0.00002853, -0.00001395, -0.00000078, 0.00003656, -0.00002292, -0.00002316, 0.00004413, -0.00000834, -0.00004337, 0.00002942, 0.00000576, -0.00003765, 0.00002828, 0.00001122, -0.00004383, 0.00002430, 0.00002353, -0.00004369, 0.00001848, 0.00003174, -0.00005052, 0.00000008, 0.00003912, -0.00004038, -0.00000771, 0.00004263, -0.00003178, -0.00001356, 0.00004017, -0.00003556, -0.00002747, 0.00003906, -0.00003822, -0.00004781, 0.00004140, -0.00001562, -0.00004494, 0.00001321, -0.00014382, -0.00023680, 0.00013954, 0.00049022, 0.00031521, -0.00004195, -0.00016423, -0.00008943, -0.00000907, -0.00001767, -0.00004068, -0.00000500, 0.00000830, -0.00001499, 0.00000279, 0.00001474, -0.00001342, -0.00000968, 0.00000742, -0.00001065, -0.00001576, -0.00000046, -0.00000385, -0.00000507, 0.00000535, 0.00000191, -0.00000415, -0.00000037, -0.00000093, -0.00000419, -0.00000163, 0.00000156, 0.00000228, 0.00000201, 0.00000128, 0.00000084, },
		{ 0.00000094, 0.00000091, 0.00000044, -0.00000093, -0.00000227, -0.00000196, -0.00000034, 0.00000236, 0.00000346, -0.00000199, -0.00000148, 0.00001439, 0.00001476, -0.00000937, -0.00001144, 0.00000745, -0.00000092, -0.00001612, 0.00000040, 0.00001439, -0.00000045, -0.00001383, -0.00000845, 0.00000063, 0.00000024, -0.00000029, 0.00001075, 0.00000554, -0.00002503, -0.00001999, 0.00001910, 0.00001187, -0.00001567, 0.00000176, 0.00000716, -0.00001842, -0.00000331, 0.00001722, -0.00000675, -0.00000616, 0.00001706, -0.00000119, -0.00001189, 0.00000660, -0.00000091, -0.00000362, 0.00001166, -0.00000787, -0.00001776, 0.00000461, -0.00000283, 0.00000344, 0.00003184, -0.00001208, -0.00004185, 0.00001692, 0.00001473, -0.00002301, 0.00002494, 0.00001964, -0.00004145, 0.00001087, 0.00003703, -0.00003964, -0.00001392, 0.00003514, -0.00002574, -0.00001317, 0.00003784, -0.00002417, -0.00002211, 0.00004243, -0.00001353, -0.00002782, 0.00004254, -0.00000836, -0.00004358, 0.00003298, 0.00000545, -0.00004108, 0.00002668, 0.00001325, -0.00003747, 0.00002070, 0.00001036, -0.00004669, 0.00001176, 0.00001110, -0.00005823, -0.00000167, 0.00002477, -0.00003999, -0.00000722, -0.00004687, -0.00021800, -0.00008326, 0.00032281, 0.00042534, 0.00015378, -0.00009906, -0.00014469, -0.00005675, 0.00000006, -0.00003046, -0.00003144, 0.00000806, -0.00000251, -0.00001208, 0.00001432, 0.00000361, -0.00001732, 0.00000036, 0.00000166, -0.00001644, -0.00000886, -0.00000046, -0.00000542, 0.00000018, 0.00000471, -0.00000173, -0.00000258, -0.00000009, -0.00000249, -0.00000330, -0.00000054, 0.00000191, 0.00000259, 0.00000168, },
		{ 0.00000256, 0.00000125, 0.00000081, 0.00000105, 0.00000011, -0.00000208, -0.00000226, -0.00000059, 0.00000029, 0.00000297, 0.00000262, -0.00000386, 0.00000307, 0.00001854, 0.00000615, -0.00001501, -0.00000318, 0.00000839, -0.00000923, -0.00001285, 0.00000832, 0.00001170, -0.00000628, -0.00001491, -0.00000573, 0.00000315, -0.00000016, 0.00000320, 0.00001221, -0.00000927, -0.00003072, 0.00000118, 0.00002410, -0.00000662, -0.00001124, 0.00001249, -0.00000702, -0.00001816, 0.00001346, 0.00000743, -0.00001506, 0.00000949, 0.00001389, -0.00001370, -0.00000263, 0.00000725, -0.00000779, 0.00000653, 0.00000841, -0.00002002, -0.00000681, 0.00000616, -0.00000959, 0.00002040, 0.00002554, -0.00003824, -0.00002210, 0.00003085, -0.00000962, -0.00001010, 0.00003954, -0.00001242, -0.00003293, 0.00003900, 0.00000556, -0.00004518, 0.00001883, 0.00001346, -0.00003596, 0.00001903, 0.00001780, -0.00004038, 0.00001346, 0.00002885, -0.00003608, 0.00000691, 0.00003355, -0.00003907, -0.00001097, 0.00003544, -0.00002575, -0.00001464, 0.00003354, -0.00001708, -0.00001502, 0.00002751, -0.00002316, -0.00002489, 0.00002371, -0.00002672, -0.00003927, 0.00002115, -0.00000828, -0.00002555, -0.00000857, -0.00013852, -0.00018429, 0.00011085, 0.00038538, 0.00030326, 0.00004612, -0.00012487, -0.00012166, -0.00002487, -0.00000381, -0.00003909, -0.00001400, 0.00000933, -0.00001160, -0.00000009, 0.00001554, -0.00000896, -0.00001187, 0.00000457, -0.00000727, -0.00001532, -0.00000391, -0.00000254, -0.00000314, 0.00000319, 0.00000159, -0.00000267, -0.00000115, -0.00000102, -0.00000290, -0.00000235, 0.00000018, 0.00000249, },
		{ 0.00000100, 0.00000290, 0.00000227, 0.00000085, 0.00000081, 0.00000127, -0.00000088, -0.00000310, -0.00000095, 0.00000033, 0.00000042, 0.00000392, 0.00000063, -0.00000431, 0.00000978, 0.00001787, -0.00000439, -0.00001396, 0.00000465, 0.00000352, -0.00001409, -0.00000583, 0.00001270, 0.00000720, -0.00001162, -0.00001506, -0.00000081, 0.00000458, -0.00000071, 0.00000805, 0.00000599, -0.00002517, -0.00002043, 0.00002099, 0.00001158, -0.00001749, 0.00000420, 0.00000914, -0.00002081, -0.00000310, 0.00001905, -0.00000984, -0.00000744, 0.00002115, -0.00000136, -0.00001498, 0.00000787, -0.00000122, -0.00000544, 0.00001583, -0.00000589, -0.00002138, 0.00000633, -0.00000328, -0.00000483, 0.00003665, 0.00000057, -0.00004834, 0.00000821, 0.00002184, -0.00002499, 0.00001689, 0.00002954, -0.00003773, -0.00000218, 0.00004140, -0.00002822, -0.00002333, 0.00003140, -0.00001657, -0.00001873, 0.00003369, -0.00001472, -0.00002672, 0.00003562, -0.00000274, -0.00002907, 0.00003243, 0.00000172, -0.00004050, 0.00001981, 0.00001236, -0.00003251, 0.00001359, 0.00001525, -0.00002521, 0.00001002, 0.00000761, -0.00003335, 0.00000307, 0.00000477, -0.00004184, -0.00000752, 0.00001115, -0.00002140, -0.00000484, -0.00006554, -0.00018658, -0.00005443, 0.00025245, 0.00034718, 0.00019073, -0.00002096, -0.00013822, -0.00008865, -0.00000215, -0.00001801, -0.00003544, 0.00000121, 0.00000106, -0.00001074, 0.00001134, 0.00000669, -0.00001452, -0.00000354, 0.00000127, -0.00001295, -0.00001072, -0.00000257, -0.00000295, 0.00000035, 0.00000276, -0.00000091, -0.00000196, -0.00000090, -0.00000182, -0.00000266, -0.00000165, },
		{ -0.00000235, -0.00000090, 0.00000179, 0.00000307, 0.00000190, 0.00000037, 0.00000118, 0.00000120, -0.00000267, -0.00000285, 0.00000096, 0.00000001, 0.00000112, 0.00000450, -0.00000248, -0.00000161, 0.00001602, 0.00001148, -0.00001220, -0.00000764, 0.00000795, -0.00000350, -0.00001397, 0.00000145, 0.00001415, 0.00000155, -0.00001713, -0.00001202, 0.00000523, 0.00000322, 0.00000083, 0.00001062, -0.00000863, -0.00003134, 0.00000207, 0.00002560, -0.00000854, -0.00001143, 0.00001605, -0.00000733, -0.00002056, 0.00001544, 0.00000753, -0.00001886, 0.00001102, 0.00001739, -0.00001631, -0.00000408, 0.00000902, -0.00000988, 0.00000686, 0.00001352, -0.00002143, -0.00000925, 0.00000959, -0.00001439, 0.00001413, 0.00003693, -0.00003105, -0.00003385, 0.00002904, -0.00000265, -0.00001805, 0.00003771, -0.00000020, -0.00003821, 0.00002935, 0.00001724, -0.00004157, 0.00000749, 0.00001748, -0.00003113, 0.00001022, 0.00002104, -0.00003426, 0.00000386, 0.00003011, -0.00002657, -0.00000188, 0.00003104, -0.00002717, -0.00001746, 0.00002798, -0.00001334, -0.00001621, 0.00002284, -0.00000668, -0.00001177, 0.00001570, -0.00001607, -0.00001892, 0.00001115, -0.00002136, -0.00002850, 0.00000694, -0.00000852, -0.00000849, -0.00001961, -0.00013922, -0.00014415, 0.00010046, 0.00029943, 0.00026871, 0.00010966, -0.00007114, -0.00013624, -0.00004868, 0.00000297, -0.00003101, -0.00002158, 0.00000600, -0.00000697, -0.00000122, 0.00001396, -0.00000448, -0.00001183, 0.00000074, -0.00000521, -0.00001339, -0.00000688, -0.00000263, -0.00000099, 0.00000205, 0.00000067, -0.00000162, -0.00000127, -0.00000129, -0.00000209, },
		{ -0.00000163, -0.00000214, -0.00000193, -0.00000009, 0.00000237, 0.00000322, 0.00000122, 0.00000004, 0.00000205, -0.00000001, -0.00000436, -0.00000079, 0.00000192, -0.00000091, 0.00000297, 0.00000333, -0.00000501, 0.00000441, 0.00001832, 0.00000179, -0.00001425, -0.00000033, 0.00000629, -0.00000883, -0.00001056, 0.00000750, 0.00001363, -0.00000632, -0.00002086, -0.00000442, 0.00000885, 0.00000036, 0.00000502, 0.00000560, -0.00002493, -0.00002062, 0.00002280, 0.00001150, -0.00001968, 0.00000631, 0.00001195, -0.00002323, -0.00000393, 0.00002160, -0.00001205, -0.00001006, 0.00002487, -0.00000020, -0.00001846, 0.00000861, -0.00000063, -0.00000810, 0.00001914, -0.00000231, -0.00002531, 0.00000703, -0.00000098, -0.00001352, 0.00003637, 0.00001501, -0.00004970, -0.00000339, 0.00002680, -0.00002302, 0.00000707, 0.00003573, -0.00002942, -0.00001387, 0.00004011, -0.00001524, -0.00002836, 0.00002424, -0.00000849, -0.00002084, 0.00002743, -0.00000667, -0.00002774, 0.00002714, 0.00000542, -0.00002646, 0.00002202, 0.00000786, -0.00003391, 0.00000845, 0.00001422, -0.00002239, 0.00000444, 0.00001283, -0.00001378, 0.00000431, 0.00000203, -0.00002272, -0.00000046, -0.00000298, -0.00003000, -0.00000745, -0.00000135, -0.00001123, 0.00000188, -0.00007743, -0.00016744, -0.00002771, 0.00020625, 0.00027285, 0.00019829, 0.00004704, -0.00011119, -0.00011278, -0.00001520, -0.00000684, -0.00003324, -0.00000735, 0.00000180, -0.00000709, 0.00000848, 0.00000764, -0.00001047, -0.00000607, -0.00000071, -0.00000978, -0.00001100, -0.00000504, -0.00000161, 0.00000128, 0.00000137, -0.00000089, -0.00000133, -0.00000117, },
		{ -0.00000103, -0.00000132, -0.00000185, -0.00000205, -0.00000130, 0.00000050, 0.00000301, 0.00000320, 0.00000004, 0.00000059, 0.00000264, -0.00000260, -0.00000432, 0.00000209, 0.00000098, -0.00000083, 0.00000494, 0.00000004, -0.00000460, 0.00001141, 0.00001502, -0.00000689, -0.00001120, 0.00000438, 0.00000231, -0.00001130, -0.00000597, 0.00001284, 0.00001009, -0.00001622, -0.00001905, 0.00000534, 0.00000757, -0.00000047, 0.00000810, -0.00000834, -0.00003138, 0.00000288, 0.00002714, -0.00001024, -0.00001226, 0.00001967, -0.00000673, -0.00002365, 0.00001673, 0.00000899, -0.00002263, 0.00001096, 0.00002151, -0.00001792, -0.00000652, 0.00001105, -0.00001137, 0.00000565, 0.00001888, -0.00002130, -0.00001317, 0.00001338, -0.00001623, 0.00000475, 0.00004449, -0.00001912, -0.00004298, 0.00002277, 0.00000521, -0.00002295, 0.00003165, 0.00001135, -0.00003816, 0.00001732, 0.00002471, -0.00003359, -0.00000237, 0.00001749, -0.00002467, 0.00000321, 0.00002108, -0.00002693, -0.00000337, 0.00002789, -0.00001691, -0.00000710, 0.00002549, -0.00001674, -0.00001941, 0.00001886, -0.00000459, -0.00001366, 0.00001285, -0.00000094, -0.00000593, 0.00000659, -0.00001351, -0.00001201, 0.00000258, -0.00002042, -0.00001840, -0.00000097, -0.00001275, 0.00000344, -0.00002248, -0.00014038, -0.00011647, 0.00009964, 0.00023474, 0.00022278, 0.00014639, -0.00001344, -0.00013152, -0.00007371, 0.00000053, -0.00001986, -0.00002522, -0.00000059, -0.00000319, -0.00000015, 0.00001122, -0.00000120, -0.00000996, -0.00000277, -0.00000474, -0.00001089, -0.00000867, -0.00000368, 0.00000058, 0.00000191, -0.00000022, -0.00000130, },
		{ -0.00000127, -0.00000114, -0.00000092, -0.00000153, -0.00000206, -0.00000160, -0.00000083, 0.00000099, 0.00000401, 0.00000233, -0.00000101, 0.00000223, 0.00000159, -0.00000534, -0.00000175, 0.00000372, -0.00000101, 0.00000112, 0.00000510, -0.00000371, -0.00000039, 0.00001585, 0.00000775, -0.00001156, -0.00000623, 0.00000603, -0.00000147, -0.00001213, -0.00000024, 0.00001716, 0.00000144, -0.00002443, -0.00001000, 0.00001203, 0.00000307, 0.00000224, 0.00000411, -0.00002444, -0.00002040, 0.00002432, 0.00001176, -0.00002193, 0.00000765, 0.00001539, -0.00002504, -0.00000599, 0.00002428, -0.00001279, -0.00001371, 0.00002747, 0.00000236, -0.00002183, 0.00000841, 0.00000093, -0.00001109, 0.00002093, 0.00000274, -0.00002870, 0.00000589, 0.00000351, -0.00002065, 0.00003095, 0.00002865, -0.00004516, -0.00001585, 0.00002803, -0.00001770, -0.00000240, 0.00003724, -0.00001831, -0.00002203, 0.00003387, -0.00000320, -0.00002849, 0.00001548, -0.00000292, -0.00001993, 0.00002053, -0.00000093, -0.00002584, 0.00001856, 0.00001032, -0.00002119, 0.00001297, 0.00001005, -0.00002569, 0.00000027, 0.00001211, -0.00001290, -0.00000038, 0.00000797, -0.00000491, 0.00000282, -0.00000414, -0.00001573, -0.00000041, -0.00000989, -0.00002290, -0.00000413, -0.00001036, -0.00000776, 0.00000880, -0.00008184, -0.00015545, -0.00000759, 0.00017787, 0.00020985, 0.00018422, 0.00009762, -0.00007064, -0.00012378, -0.00003510, -0.00000088, -0.00002601, -0.00001478, -0.00000074, -0.00000268, 0.00000696, 0.00000685, -0.00000651, -0.00000683, -0.00000330, -0.00000774, -0.00001014, -0.00000700, -0.00000144, 0.00000226, 0.00000089, },
		{ 0.00000232, -0.00000052, -0.00000161, -0.00000088, -0.00000088, -0.00000197, -0.00000191, -0.00000095, -0.00000086, 0.00000208, 0.00000477, 0.00000032, -0.00000065, 0.00000383, -0.00000158, -0.00000605, 0.00000214, 0.00000282, -0.00000200, 0.00000389, 0.00000261, -0.00000529, 0.00000584, 0.00001573, -0.00000002, -0.00001238, -0.00000181, 0.00000633, -0.00000501, -0.00001187, 0.00000765, 0.00001751, -0.00001184, -0.00002544, 0.00000295, 0.00001213, -0.00000011, 0.00000495, -0.00000880, -0.00003077, 0.00000364, 0.00002851, -0.00001138, -0.00001366, 0.00002275, -0.00000500, -0.00002687, 0.00001675, 0.00001156, -0.00002555, 0.00000912, 0.00002559, -0.00001807, -0.00000976, 0.00001286, -0.00001179, 0.00000302, 0.00002367, -0.00001911, -0.00001808, 0.00001623, -0.00001464, -0.00000578, 0.00004672, -0.00000436, -0.00004746, 0.00001287, 0.00001189, -0.00002398, 0.00002286, 0.00002011, -0.00003316, 0.00000541, 0.00002705, -0.00002322, -0.00000901, 0.00001416, -0.00001824, -0.00000129, 0.00001875, -0.00001976, -0.00000777, 0.00002330, -0.00000854, -0.00000880, 0.00001854, -0.00000901, -0.00001770, 0.00001011, 0.00000008, -0.00000881, 0.00000523, 0.00000084, 0.00000037, 0.00000096, -0.00001396, -0.00000612, -0.00000204, -0.00002179, -0.00001080, -0.00000373, -0.00001772, 0.00000942, -0.00002069, -0.00013847, -0.00009880, 0.00010136, 0.00018964, 0.00017539, 0.00015862, 0.00003880, -0.00010980, -0.00009342, -0.00001049, -0.00000960, -0.00002403, -0.00000833, -0.00000170, 0.00000254, 0.00000872, 0.00000041, -0.00000724, -0.00000507, -0.00000545, -0.00000880, -0.00000917, -0.00000491, 0.00000113, },
		{ -0.00000199, 0.00000264, 0.00000123, -0.00000149, -0.00000160, -0.00000047, -0.00000117, -0.00000244, -0.00000107, -0.00000087, -0.00000088, 0.00000410, 0.00000408, -0.00000177, 0.00000156, 0.00000344, -0.00000542, -0.00000357, 0.00000486, 0.00000024, -0.00000075, 0.00000531, -0.00000117, -0.00000341, 0.00001099, 0.00001197, -0.00000597, -0.00001133, 0.00000220, 0.00000610, -0.00000954, -0.00000873, 0.00001662, 0.00001025, -0.00002462, -0.00001671, 0.00001313, 0.00000718, 0.00000046, 0.00000144, -0.00002399, -0.00001959, 0.00002539, 0.00001231, -0.00002375, 0.00000794, 0.00001896, -0.00002566, -0.00000907, 0.00002632, -0.00001179, -0.00001769, 0.00002834, 0.00000603, -0.00002446, 0.00000700, 0.00000325, -0.00001375, 0.00002077, 0.00000872, -0.00003062, 0.00000261, 0.00000900, -0.00002466, 0.00002149, 0.00003899, -0.00003520, -0.00002671, 0.00002484, -0.00001056, -0.00000964, 0.00003424, -0.00000671, -0.00002543, 0.00002445, 0.00000578, -0.00002435, 0.00000715, -0.00000046, -0.00001708, 0.00001431, 0.00000230, -0.00002207, 0.00001111, 0.00001204, -0.00001472, 0.00000635, 0.00000905, -0.00001763, -0.00000425, 0.00000768, -0.00000561, -0.00000151, 0.00000265, 0.00000064, 0.00000405, -0.00000912, -0.00001221, 0.00000133, -0.00001458, -0.00001962, -0.00000005, -0.00001505, -0.00000837, 0.00001313, -0.00008008, -0.00014598, 0.00000464, 0.00016041, 0.00016160, 0.00015755, 0.00012771, -0.00002501, -0.00011963, -0.00005639, -0.00000214, -0.00001642, -0.00001896, -0.00000570, 0.00000081, 0.00000714, 0.00000532, -0.00000365, -0.00000619, -0.00000545, -0.00000700, -0.00000896, -0.00000789, },
		{ -0.00000881, -0.00000557, 0.00000071, 0.00000264, 0.00000001, -0.00000214, -0.00000121, -0.00000006, -0.00000207, -0.00000218, -0.00000009, -0.00000178, 0.00000032, 0.00000602, 0.00000146, -0.00000209, 0.00000421, 0.00000022, -0.00000728, 0.00000091, 0.00000474, -0.00000177, 0.00000191, 0.00000420, -0.00000382, 0.00000076, 0.00001329, 0.00000689, -0.00001028, -0.00000924, 0.00000701, 0.00000368, -0.00001437, -0.00000044, 0.00002191, -0.00000454, -0.00002962, -0.00000188, 0.00001579, 0.00000224, 0.00000172, -0.00001036, -0.00002955, 0.00000451, 0.00002943, -0.00001173, -0.00001525, 0.00002470, -0.00000227, -0.00002949, 0.00001517, 0.00001464, -0.00002689, 0.00000575, 0.00002885, -0.00001653, -0.00001339, 0.00001390, -0.00001088, -0.00000055, 0.00002703, -0.00001480, -0.00002306, 0.00001704, -0.00001003, -0.00001527, 0.00004326, 0.00001062, -0.00004613, 0.00000114, 0.00001562, -0.00002136, 0.00001338, 0.00002478, -0.00002465, -0.00000410, 0.00002455, -0.00001275, -0.00001164, 0.00000890, -0.00001321, -0.00000336, 0.00001517, -0.00001376, -0.00000953, 0.00001763, -0.00000238, -0.00000767, 0.00001176, -0.00000442, -0.00001377, 0.00000321, 0.00000121, -0.00000352, 0.00000069, 0.00000005, 0.00000547, -0.00000142, -0.00001566, -0.00000230, -0.00000360, -0.00002355, -0.00000637, -0.00000323, -0.00002110, 0.00001031, -0.00001758, -0.00013221, -0.00008752, 0.00010109, 0.00015989, 0.00013364, 0.00015192, 0.00007883, -0.00007641, -0.00010305, -0.00002704, -0.00000356, -0.00001888, -0.00001488, -0.00000288, 0.00000557, 0.00000742, 0.00000070, -0.00000480, -0.00000587, -0.00000648, -0.00000774, },
		{ -0.00000714, -0.00000827, -0.00000771, -0.00000256, 0.00000207, 0.00000194, -0.00000108, -0.00000253, -0.00000023, -0.00000037, -0.00000305, -0.00000077, -0.00000020, -0.00000270, 0.00000321, 0.00000602, -0.00000173, 0.00000018, 0.00000487, -0.00000436, -0.00000559, 0.00000469, 0.00000240, -0.00000172, 0.00000392, 0.00000147, -0.00000402, 0.00000497, 0.00001328, 0.00000156, -0.00001395, -0.00000469, 0.00001175, -0.00000291, -0.00001570, 0.00001177, 0.00001810, -0.00002090, -0.00002314, 0.00001141, 0.00001193, 0.00000038, -0.00000222, -0.00002399, -0.00001800, 0.00002596, 0.00001291, -0.00002469, 0.00000719, 0.00002198, -0.00002473, -0.00001264, 0.00002699, -0.00000918, -0.00002115, 0.00002714, 0.00001022, -0.00002576, 0.00000438, 0.00000586, -0.00001542, 0.00001862, 0.00001473, -0.00003028, -0.00000247, 0.00001398, -0.00002476, 0.00000995, 0.00004421, -0.00002149, -0.00003375, 0.00001766, -0.00000358, -0.00001351, 0.00002790, 0.00000321, -0.00002403, 0.00001415, 0.00001050, -0.00001748, 0.00000093, -0.00000080, -0.00001364, 0.00000954, 0.00000341, -0.00001759, 0.00000549, 0.00001125, -0.00000841, 0.00000254, 0.00000610, -0.00001106, -0.00000553, 0.00000266, -0.00000124, -0.00000022, -0.00000156, 0.00000311, 0.00000634, -0.00001201, -0.00001122, 0.00000317, -0.00001666, -0.00001871, 0.00000303, -0.00001587, -0.00001044, 0.00001374, -0.00007458, -0.00013587, 0.00001008, 0.00014793, 0.00012787, 0.00012665, 0.00013827, 0.00001773, -0.00010175, -0.00007360, -0.00001029, -0.00000776, -0.00001911, -0.00001135, 0.00000221, 0.00000851, 0.00000420, -0.00000220, -0.00000502, -0.00000643, },
		{ -0.00000558, -0.00000693, -0.00000776, -0.00000834, -0.00000526, -0.00000023, 0.00000229, 0.00000122, -0.00000235, -0.00000214, 0.00000082, -0.00000183, -0.00000286, 0.00000084, -0.00000192, -0.00000188, 0.00000641, 0.00000330, -0.00000309, 0.00000355, 0.00000224, -0.00000739, -0.00000128, 0.00000582, -0.00000006, -0.00000008, 0.00000412, -0.00000096, -0.00000255, 0.00000821, 0.00001215, -0.00000490, -0.00001600, 0.00000366, 0.00001272, -0.00001232, -0.00000956, 0.00002193, 0.00000428, -0.00003025, -0.00000841, 0.00001742, 0.00000643, -0.00000077, -0.00001322, -0.00002798, 0.00000573, 0.00002964, -0.00001130, -0.00001651, 0.00002511, 0.00000097, -0.00003079, 0.00001208, 0.00001739, -0.00002623, 0.00000147, 0.00003051, -0.00001342, -0.00001677, 0.00001377, -0.00000871, -0.00000434, 0.00002832, -0.00000879, -0.00002699, 0.00001523, -0.00000359, -0.00002185, 0.00003497, 0.00002319, -0.00003912, -0.00001007, 0.00001538, -0.00001630, 0.00000514, 0.00002513, -0.00001470, -0.00000975, 0.00001856, -0.00000426, -0.00001051, 0.00000342, -0.00001030, -0.00000367, 0.00001144, -0.00000936, -0.00000931, 0.00001203, 0.00000131, -0.00000480, 0.00000630, -0.00000267, -0.00000914, -0.00000111, -0.00000006, 0.00000074, -0.00000096, -0.00000175, 0.00000850, -0.00000148, -0.00001714, -0.00000075, -0.00000340, -0.00002439, -0.00000477, -0.00000143, -0.00002184, 0.00000789, -0.00001559, -0.00012213, -0.00007909, 0.00009701, 0.00014027, 0.00010125, 0.00013340, 0.00010341, -0.00003814, -0.00010062, -0.00004468, -0.00000352, -0.00001198, -0.00001849, -0.00000597, 0.00000755, 0.00000758, 0.00000055, -0.00000343, },
		{ -0.00000180, -0.00000429, -0.00000611, -0.00000735, -0.00000855, -0.00000693, -0.00000248, 0.00000069, 0.00000238, 0.00000014, -0.00000351, -0.00000057, 0.00000073, -0.00000351, -0.00000095, 0.00000091, -0.00000393, 0.00000143, 0.00000752, -0.00000066, -0.00000140, 0.00000528, -0.00000237, -0.00000696, 0.00000303, 0.00000458, -0.00000113, 0.00000152, 0.00000311, -0.00000231, -0.00000069, 0.00001122, 0.00000938, -0.00001302, -0.00001317, 0.00001372, 0.00000650, -0.00001905, 0.00000352, 0.00002295, -0.00001366, -0.00002760, 0.00000676, 0.00001620, 0.00000250, -0.00000632, -0.00002489, -0.00001560, 0.00002619, 0.00001320, -0.00002443, 0.00000577, 0.00002373, -0.00002229, -0.00001590, 0.00002581, -0.00000551, -0.00002323, 0.00002395, 0.00001410, -0.00002532, 0.00000088, 0.00000817, -0.00001566, 0.00001484, 0.00001977, -0.00002738, -0.00000843, 0.00001708, -0.00002114, -0.00000137, 0.00004358, -0.00000650, -0.00003552, 0.00000798, 0.00000138, -0.00001387, 0.00002002, 0.00000990, -0.00001886, 0.00000518, 0.00001090, -0.00000989, -0.00000225, -0.00000295, -0.00001078, 0.00000640, 0.00000314, -0.00001331, 0.00000185, 0.00000888, -0.00000325, 0.00000123, 0.00000252, -0.00000669, -0.00000454, -0.00000157, 0.00000035, 0.00000200, -0.00000384, 0.00000337, 0.00000830, -0.00001272, -0.00001152, 0.00000407, -0.00001656, -0.00001869, 0.00000426, -0.00001404, -0.00001198, 0.00001094, -0.00006788, -0.00012373, 0.00001113, 0.00013638, 0.00010597, 0.00009780, 0.00013331, 0.00005165, -0.00007423, -0.00008268, -0.00002289, -0.00000282, -0.00001599, -0.00001569, 0.00000128, 0.00000993, 0.00000425, },
		{ 0.00000864, 0.00000095, -0.00000331, -0.00000501, -0.00000624, -0.00000839, -0.00000848, -0.00000406, -0.00000097, 0.00000102, 0.00000258, -0.00000181, -0.00000345, 0.00000132, -0.00000110, -0.00000368, 0.00000141, -0.00000130, -0.00000392, 0.00000566, 0.00000535, -0.00000312, 0.00000203, 0.00000375, -0.00000611, -0.00000368, 0.00000540, 0.00000264, -0.00000109, 0.00000233, 0.00000208, -0.00000334, 0.00000173, 0.00001442, 0.00000279, -0.00002019, -0.00000325, 0.00001975, -0.00000572, -0.00001727, 0.00001715, 0.00001256, -0.00002656, -0.00001523, 0.00001609, 0.00001179, -0.00000169, -0.00001729, -0.00002647, 0.00000757, 0.00002902, -0.00001035, -0.00001687, 0.00002393, 0.00000399, -0.00003025, 0.00000797, 0.00001895, -0.00002355, -0.00000284, 0.00003009, -0.00000923, -0.00001923, 0.00001226, -0.00000564, -0.00000755, 0.00002725, -0.00000195, -0.00002886, 0.00001093, 0.00000306, -0.00002442, 0.00002366, 0.00003123, -0.00002773, -0.00001845, 0.00001123, -0.00001056, -0.00000051, 0.00002195, -0.00000542, -0.00001112, 0.00001107, 0.00000092, -0.00000679, -0.00000077, -0.00000945, -0.00000321, 0.00000828, -0.00000651, -0.00000793, 0.00000729, 0.00000278, -0.00000134, 0.00000277, -0.00000292, -0.00000507, -0.00000290, -0.00000237, 0.00000315, -0.00000056, -0.00000325, 0.00000941, -0.00000045, -0.00001754, -0.00000097, -0.00000267, -0.00002378, -0.00000507, 0.00000018, -0.00001997, 0.00000417, -0.00001584, -0.00010982, -0.00007094, 0.00008938, 0.00012607, 0.00007880, 0.00010989, 0.00011277, -0.00000164, -0.00008704, -0.00005887, -0.00000920, -0.00000611, -0.00001856, -0.00000942, 0.00000754, },
		{ -0.00000122, 0.00001020, 0.00000557, -0.00000159, -0.00000463, -0.00000495, -0.00000683, -0.00000975, -0.00000628, -0.00000145, -0.00000099, 0.00000201, 0.00000193, -0.00000398, -0.00000154, 0.00000178, -0.00000349, -0.00000157, 0.00000199, -0.00000425, -0.00000081, 0.00000801, 0.00000126, -0.00000249, 0.00000444, -0.00000009, -0.00000700, 0.00000027, 0.00000583, 0.00000108, -0.00000091, 0.00000300, 0.00000092, -0.00000458, 0.00000627, 0.00001566, -0.00000839, -0.00002114, 0.00001116, 0.00001649, -0.00001743, -0.00000620, 0.00002319, -0.00000427, -0.00002852, -0.00000021, 0.00001869, 0.00000681, -0.00001005, -0.00002699, -0.00001257, 0.00002635, 0.00001283, -0.00002291, 0.00000423, 0.00002369, -0.00001875, -0.00001801, 0.00002271, -0.00000165, -0.00002331, 0.00001929, 0.00001685, -0.00002310, -0.00000296, 0.00000962, -0.00001431, 0.00001014, 0.00002296, -0.00002213, -0.00001405, 0.00001745, -0.00001487, -0.00001035, 0.00003765, 0.00000709, -0.00003176, -0.00000207, 0.00000310, -0.00001156, 0.00001247, 0.00001281, -0.00001168, -0.00000087, 0.00000793, -0.00000346, -0.00000241, -0.00000553, -0.00000918, 0.00000453, 0.00000230, -0.00000975, -0.00000009, 0.00000588, 0.00000025, 0.00000175, -0.00000060, -0.00000451, -0.00000248, -0.00000420, -0.00000010, 0.00000392, -0.00000413, 0.00000258, 0.00000912, -0.00001179, -0.00001200, 0.00000371, -0.00001509, -0.00001849, 0.00000365, -0.00001102, -0.00001192, 0.00000592, -0.00006192, -0.00010968, 0.00001032, 0.00012388, 0.00009213, 0.00007455, 0.00011833, 0.00007373, -0.00004262, -0.00008185, -0.00003623, -0.00000306, -0.00001154, -0.00001728, },
		{ -0.00001583, -0.00001150, 0.00000544, 0.00000954, 0.00000251, -0.00000400, -0.00000506, -0.00000439, -0.00000883, -0.00000960, -0.00000249, -0.00000093, -0.00000096, 0.00000347, -0.00000041, -0.00000467, 0.00000112, -0.00000006, -0.00000423, 0.00000139, -0.00000023, -0.00000523, 0.00000387, 0.00000692, -0.00000192, 0.00000022, 0.00000409, -0.00000371, -0.00000537, 0.00000328, 0.00000538, -0.00000043, -0.00000056, 0.00000410, -0.00000166, -0.00000450, 0.00001311, 0.00001107, -0.00002051, -0.00001218, 0.00002276, 0.00000417, -0.00002127, 0.00000836, 0.00001803, -0.00001878, -0.00002052, 0.00001147, 0.00001720, -0.00000042, -0.00002211, -0.00002553, 0.00001017, 0.00002767, -0.00000929, -0.00001591, 0.00002145, 0.00000598, -0.00002778, 0.00000365, 0.00001869, -0.00001929, -0.00000623, 0.00002750, -0.00000472, -0.00002022, 0.00000952, -0.00000227, -0.00000955, 0.00002401, 0.00000462, -0.00002805, 0.00000496, 0.00000835, -0.00002286, 0.00001166, 0.00003370, -0.00001422, -0.00002237, 0.00000429, -0.00000595, -0.00000312, 0.00001672, 0.00000157, -0.00000887, 0.00000411, 0.00000251, -0.00000212, -0.00000278, -0.00000995, -0.00000289, 0.00000597, -0.00000478, -0.00000612, 0.00000374, 0.00000261, 0.00000180, 0.00000119, -0.00000415, -0.00000233, -0.00000276, -0.00000451, 0.00000362, 0.00000081, -0.00000367, 0.00000873, 0.00000060, -0.00001665, -0.00000212, -0.00000225, -0.00002185, -0.00000623, 0.00000077, -0.00001627, 0.00000075, -0.00001829, -0.00009716, -0.00006187, 0.00007970, 0.00011395, 0.00006467, 0.00008671, 0.00010983, 0.00002807, -0.00006565, -0.00006619, -0.00001853, -0.00000360, },
		{ -0.00000818, -0.00000817, -0.00001570, -0.00000383, 0.00000856, 0.00000757, -0.00000068, -0.00000610, -0.00000380, -0.00000511, -0.00001131, -0.00000643, 0.00000036, -0.00000224, 0.00000065, 0.00000372, -0.00000353, -0.00000290, 0.00000231, -0.00000288, -0.00000230, 0.00000258, -0.00000369, -0.00000286, 0.00000718, 0.00000350, -0.00000247, 0.00000254, 0.00000167, -0.00000546, -0.00000278, 0.00000540, 0.00000436, -0.00000230, 0.00000108, 0.00000440, -0.00000586, -0.00000029, 0.00001886, -0.00000101, -0.00002647, 0.00000442, 0.00002449, -0.00001069, -0.00001483, 0.00001819, 0.00000516, -0.00002491, -0.00000811, 0.00001832, 0.00001277, -0.00001249, -0.00003035, -0.00000931, 0.00002680, 0.00001157, -0.00002039, 0.00000326, 0.00002168, -0.00001487, -0.00001832, 0.00001809, 0.00000147, -0.00002123, 0.00001400, 0.00001782, -0.00001941, -0.00000645, 0.00000984, -0.00001160, 0.00000538, 0.00002375, -0.00001529, -0.00001810, 0.00001493, -0.00000756, -0.00001559, 0.00002798, 0.00001700, -0.00002340, -0.00001024, 0.00000132, -0.00000806, 0.00000667, 0.00001237, -0.00000439, -0.00000337, 0.00000325, 0.00000050, -0.00000042, -0.00000730, -0.00000889, 0.00000325, 0.00000146, -0.00000704, -0.00000085, 0.00000296, 0.00000207, 0.00000321, -0.00000257, -0.00000404, -0.00000043, -0.00000506, -0.00000156, 0.00000478, -0.00000295, 0.00000176, 0.00000860, -0.00001008, -0.00001194, 0.00000235, -0.00001314, -0.00001757, 0.00000174, -0.00000801, -0.00001012, 0.00000018, -0.00005765, -0.00009481, 0.00000956, 0.00011029, 0.00008277, 0.00005780, 0.00009885, 0.00008390, -0.00001241, -0.00007186, -0.00004649, },
		{ -0.00006519, -0.00002831, -0.00000566, -0.00001210, -0.00001097, 0.00000206, 0.00000918, 0.00000522, -0.00000462, -0.00000625, -0.00000195, -0.00000830, -0.00001155, -0.00000133, 0.00000016, -0.00000307, 0.00000335, 0.00000155, -0.00000521, 0.00000015, 0.00000088, -0.00000417, 0.00000083, 0.00000077, -0.00000549, 0.00000154, 0.00000738, 0.00000032, -0.00000111, 0.00000300, -0.00000090, -0.00000560, -0.00000010, 0.00000712, 0.00000198, -0.00000350, 0.00000429, 0.00000166, -0.00000895, 0.00000864, 0.00001793, -0.00001666, -0.00002088, 0.00002069, 0.00001505, -0.00001994, -0.00000249, 0.00001887, -0.00000816, -0.00002254, 0.00000402, 0.00002127, 0.00000324, -0.00002689, -0.00002559, 0.00001346, 0.00002591, -0.00000860, -0.00001350, 0.00001830, 0.00000628, -0.00002370, 0.00000005, 0.00001635, -0.00001429, -0.00000794, 0.00002309, -0.00000076, -0.00001953, 0.00000597, 0.00000075, -0.00000994, 0.00001919, 0.00000985, -0.00002454, -0.00000146, 0.00001112, -0.00001803, 0.00000122, 0.00003075, -0.00000117, -0.00002123, -0.00000359, -0.00000380, -0.00000318, 0.00001113, 0.00000549, -0.00000440, -0.00000076, 0.00000115, 0.00000184, -0.00000254, -0.00001077, -0.00000324, 0.00000431, -0.00000366, -0.00000435, 0.00000137, 0.00000148, 0.00000403, 0.00000118, -0.00000543, -0.00000111, -0.00000159, -0.00000574, 0.00000261, 0.00000216, -0.00000285, 0.00000727, 0.00000100, -0.00001478, -0.00000340, -0.00000244, -0.00001910, -0.00000742, 0.00000019, -0.00001180, -0.00000146, -0.00002209, -0.00008561, -0.00005192, 0.00006972, 0.00010227, 0.00005618, 0.00006691, 0.00009891, 0.00004839, -0.00004102, },
		{ 0.00001224, -0.00005554, -0.00005080, -0.00001628, -0.00000791, -0.00001173, -0.00000499, 0.00000502, 0.00000918, 0.00000146, -0.00000816, -0.00000351, -0.00000231, -0.00001239, -0.00000769, 0.00000235, -0.00000247, -0.00000143, 0.00000484, -0.00000209, -0.00000409, 0.00000198, -0.00000187, -0.00000272, 0.00000244, -0.00000259, -0.00000410, 0.00000516, 0.00000529, -0.00000107, 0.00000023, 0.00000217, -0.00000261, -0.00000510, 0.00000311, 0.00000772, -0.00000191, -0.00000203, 0.00000683, -0.00000450, -0.00000675, 0.00001825, 0.00000734, -0.00002762, -0.00000491, 0.00002823, -0.00000004, -0.00001982, 0.00000859, 0.00001223, -0.00001677, -0.00001510, 0.00001451, 0.00001931, -0.00001286, -0.00003464, -0.00000634, 0.00002782, 0.00000946, -0.00001734, 0.00000339, 0.00001794, -0.00001153, -0.00001654, 0.00001271, 0.00000304, -0.00001727, 0.00000907, 0.00001673, -0.00001492, -0.00000896, 0.00000876, -0.00000804, 0.00000138, 0.00002211, -0.00000797, -0.00001973, 0.00001013, -0.00000092, -0.00001666, 0.00001679, 0.00002190, -0.00001240, -0.00001482, -0.00000318, -0.00000502, 0.00000328, 0.00000973, 0.00000141, -0.00000266, -0.00000137, 0.00000160, 0.00000240, -0.00000750, -0.00000939, 0.00000194, 0.00000084, -0.00000499, -0.00000089, 0.00000053, 0.00000250, 0.00000485, -0.00000320, -0.00000454, 0.00000090, -0.00000450, -0.00000310, 0.00000438, -0.00000107, 0.00000156, 0.00000710, -0.00000837, -0.00001113, 0.00000052, -0.00001132, -0.00001593, -0.00000075, -0.00000571, -0.00000711, -0.00000503, -0.00005506, -0.00008050, 0.00000970, 0.00009655, 0.00007532, 0.00004655, 0.00007915, 0.00008441, },
		{ 0.00008439, 0.00005923, -0.00001773, -0.00005661, -0.00003522, -0.00001187, -0.00000956, -0.00000757, -0.00000117, 0.00000688, 0.00000841, -0.00000430, -0.00000855, 0.00000018, -0.00000643, -0.00001368, -0.00000124, 0.00000205, -0.00000454, 0.00000208, 0.00000345, -0.00000465, -0.00000127, 0.00000124, -0.00000339, 0.00000008, 0.00000121, -0.00000472, -0.00000066, 0.00000639, 0.00000301, -0.00000127, 0.00000068, 0.00000141, -0.00000406, -0.00000386, 0.00000704, 0.00000559, -0.00000534, 0.00000252, 0.00000531, -0.00001065, 0.00000227, 0.00002166, -0.00000952, -0.00002698, 0.00001378, 0.00002412, -0.00001296, -0.00001275, 0.00001418, 0.00000319, -0.00002004, -0.00000500, 0.00002273, 0.00000892, -0.00003067, -0.00002687, 0.00001715, 0.00002415, -0.00000863, -0.00000982, 0.00001520, 0.00000458, -0.00001872, -0.00000205, 0.00001219, -0.00000952, -0.00000762, 0.00001761, 0.00000191, -0.00001728, 0.00000225, 0.00000292, -0.00000876, 0.00001365, 0.00001297, -0.00001890, -0.00000700, 0.00001096, -0.00001143, -0.00000602, 0.00002368, 0.00000910, -0.00001565, -0.00001035, -0.00000456, -0.00000188, 0.00000657, 0.00000649, 0.00000062, -0.00000279, -0.00000182, 0.00000394, -0.00000070, -0.00001093, -0.00000429, 0.00000288, -0.00000281, -0.00000279, -0.00000004, -0.00000004, 0.00000516, 0.00000213, -0.00000612, -0.00000116, -0.00000024, -0.00000581, 0.00000086, 0.00000286, -0.00000110, 0.00000581, 0.00000058, -0.00001252, -0.00000429, -0.00000314, -0.00001612, -0.00000820, -0.00000124, -0.00000746, -0.00000230, -0.00002611, -0.00007585, -0.00004190, 0.00006073, 0.00009083, 0.00005070, 0.00005140, },
		{ 0.00003891, 0.00006173, 0.00007872, 0.00002923, -0.00003681, -0.00004806, -0.00002446, -0.00001172, -0.00000713, -0.00000357, 0.00000046, 0.00000932, 0.00000475, -0.00000987, -0.00000458, 0.00000092, -0.00001220, -0.00000983, 0.00000371, -0.00000148, -0.00000356, 0.00000476, 0.00000001, -0.00000452, 0.00000076, -0.00000091, -0.00000246, 0.00000165, -0.00000152, -0.00000419, 0.00000245, 0.00000594, 0.00000155, -0.00000165, 0.00000109, 0.00000078, -0.00000586, -0.00000062, 0.00001002, 0.00000039, -0.00000524, 0.00000731, -0.00000132, -0.00001142, 0.00001413, 0.00001449, -0.00002411, -0.00001444, 0.00002641, 0.00001214, -0.00001931, -0.00000374, 0.00001484, -0.00000517, -0.00001923, 0.00000752, 0.00002499, -0.00001075, -0.00003915, -0.00000421, 0.00002949, 0.00000675, -0.00001434, 0.00000488, 0.00001306, -0.00000950, -0.00001288, 0.00000753, 0.00000265, -0.00001217, 0.00000536, 0.00001377, -0.00001044, -0.00001010, 0.00000664, -0.00000426, -0.00000128, 0.00001844, -0.00000132, -0.00001862, 0.00000418, 0.00000369, -0.00001412, 0.00000636, 0.00002163, -0.00000113, -0.00001501, -0.00000887, -0.00000373, 0.00000207, 0.00000635, 0.00000485, 0.00000016, -0.00000447, 0.00000029, 0.00000465, -0.00000608, -0.00000989, 0.00000025, 0.00000033, -0.00000335, -0.00000055, -0.00000127, 0.00000194, 0.00000611, -0.00000265, -0.00000528, 0.00000123, -0.00000317, -0.00000407, 0.00000302, 0.00000076, 0.00000214, 0.00000516, -0.00000716, -0.00000978, -0.00000124, -0.00000987, -0.00001386, -0.00000320, -0.00000430, -0.00000363, -0.00000901, -0.00005350, -0.00006779, 0.00001064, 0.00008383, 0.00006858, },
		{ 0.00008027, 0.00004641, 0.00003949, 0.00006968, 0.00006250, 0.00000081, -0.00004302, -0.00003686, -0.00002040, -0.00001005, -0.00000216, -0.00000273, 0.00000276, 0.00001097, -0.00000253, -0.00001132, 0.00000112, -0.00000338, -0.00001509, -0.00000249, 0.00000406, -0.00000477, -0.00000002, 0.00000457, -0.00000293, -0.00000256, 0.00000065, -0.00000210, -0.00000037, 0.00000083, -0.00000330, -0.00000205, 0.00000406, 0.00000534, 0.00000007, -0.00000219, 0.00000242, -0.00000104, -0.00000690, 0.00000491, 0.00000933, -0.00000527, -0.00000059, 0.00000789, -0.00000950, -0.00000419, 0.00002133, -0.00000092, -0.00002867, 0.00000359, 0.00002887, -0.00000144, -0.00001964, 0.00000439, 0.00001273, -0.00001274, -0.00001371, 0.00002079, 0.00001563, -0.00003254, -0.00002919, 0.00002071, 0.00002279, -0.00000949, -0.00000538, 0.00001285, 0.00000100, -0.00001376, -0.00000220, 0.00000687, -0.00000588, -0.00000536, 0.00001200, 0.00000286, -0.00001396, -0.00000101, 0.00000397, -0.00000634, 0.00000830, 0.00001368, -0.00001211, -0.00001060, 0.00000818, -0.00000479, -0.00000934, 0.00001449, 0.00001508, -0.00000722, -0.00001430, -0.00000773, -0.00000062, 0.00000377, 0.00000542, 0.00000472, -0.00000214, -0.00000486, 0.00000377, 0.00000164, -0.00000985, -0.00000562, 0.00000129, -0.00000210, -0.00000141, -0.00000075, -0.00000157, 0.00000524, 0.00000345, -0.00000594, -0.00000198, 0.00000067, -0.00000494, -0.00000092, 0.00000272, 0.00000101, 0.00000481, -0.00000046, -0.00001042, -0.00000465, -0.00000400, -0.00001331, -0.00000850, -0.00000304, -0.00000381, -0.00000212, -0.00002940, -0.00006782, -0.00003284, 0.00005321, },
		{ 0.00001164, 0.00007292, 0.00006245, 0.00003299, 0.00004724, 0.00007037, 0.00003864, -0.00001954, -0.00003911, -0.00002980, -0.00001909, -0.00000398, 0.00000056, -0.00000368, 0.00000737, 0.00000843, -0.00001033, -0.00000689, 0.00000360, -0.00001040, -0.00001210, 0.00000374, 0.00000045, -0.00000495, 0.00000332, 0.00000208, -0.00000372, -0.00000103, -0.00000052, -0.00000137, 0.00000057, -0.00000103, -0.00000312, -0.00000021, 0.00000505, 0.00000475, -0.00000235, -0.00000137, 0.00000373, -0.00000468, -0.00000484, 0.00001036, 0.00000405, -0.00000732, 0.00000565, 0.00000233, -0.00001321, 0.00000775, 0.00001853, -0.00001672, -0.00002156, 0.00001909, 0.00002289, -0.00001274, -0.00001607, 0.00001182, 0.00000783, -0.00001902, -0.00000157, 0.00002841, -0.00000637, -0.00004294, -0.00000323, 0.00003165, 0.00000388, -0.00001188, 0.00000765, 0.00000785, -0.00000927, -0.00000795, 0.00000345, 0.00000035, -0.00000686, 0.00000339, 0.00000951, -0.00000675, -0.00000982, 0.00000396, -0.00000087, -0.00000239, 0.00001351, 0.00000371, -0.00001509, -0.00000158, 0.00000557, -0.00000929, -0.00000154, 0.00001715, 0.00000816, -0.00001113, -0.00001393, -0.00000474, 0.00000213, 0.00000348, 0.00000590, 0.00000367, -0.00000533, -0.00000230, 0.00000536, -0.00000359, -0.00000964, -0.00000176, -0.00000035, -0.00000197, 0.00000001, -0.00000245, 0.00000081, 0.00000671, -0.00000130, -0.00000569, 0.00000067, -0.00000172, -0.00000424, 0.00000123, 0.00000206, 0.00000329, 0.00000335, -0.00000657, -0.00000829, -0.00000258, -0.00000873, -0.00001174, -0.00000530, -0.00000359, -0.00000040, -0.00001165, -0.00005206, -0.00005712, },
		{ -0.00006090, -0.00002550, 0.00004685, 0.00007139, 0.00004258, 0.00002975, 0.00005662, 0.00006110, 0.00001301, -0.00002789, -0.00003242, -0.00002855, -0.00001434, 0.00000352, -0.00000159, -0.00000225, 0.00001179, 0.00000055, -0.00001354, 0.00000039, 0.00000009, -0.00001495, -0.00000478, 0.00000530, -0.00000354, -0.00000224, 0.00000431, -0.00000043, -0.00000306, -0.00000094, -0.00000077, -0.00000011, -0.00000031, -0.00000185, -0.00000220, 0.00000091, 0.00000651, 0.00000276, -0.00000471, 0.00000172, 0.00000267, -0.00000827, 0.00000113, 0.00001198, -0.00000308, -0.00000389, 0.00000839, -0.00000627, -0.00000892, 0.00001712, 0.00000683, -0.00002527, -0.00000736, 0.00002771, 0.00001223, -0.00002104, -0.00000875, 0.00001815, -0.00000149, -0.00002007, 0.00001543, 0.00002200, -0.00003195, -0.00003200, 0.00002360, 0.00002208, -0.00001103, -0.00000080, 0.00001166, -0.00000394, -0.00000970, -0.00000041, 0.00000132, -0.00000397, -0.00000175, 0.00000716, 0.00000207, -0.00001027, -0.00000332, 0.00000396, -0.00000325, 0.00000388, 0.00001217, -0.00000533, -0.00001168, 0.00000375, 0.00000043, -0.00000902, 0.00000535, 0.00001635, 0.00000194, -0.00001454, -0.00001207, -0.00000061, 0.00000269, 0.00000349, 0.00000705, 0.00000032, -0.00000670, 0.00000172, 0.00000331, -0.00000756, -0.00000657, -0.00000064, -0.00000166, -0.00000014, -0.00000093, -0.00000289, 0.00000451, 0.00000467, -0.00000491, -0.00000299, 0.00000088, -0.00000356, -0.00000222, 0.00000195, 0.00000295, 0.00000440, -0.00000169, -0.00000882, -0.00000464, -0.00000466, -0.00001085, -0.00000853, -0.00000477, -0.00000102, -0.00000153, -0.00003145, },
		{ -0.00001326, -0.00005001, -0.00004828, 0.00001186, 0.00006394, 0.00005750, 0.00002761, 0.00003507, 0.00006197, 0.00004223, -0.00000653, -0.00002636, -0.00003021, -0.00002819, -0.00000398, 0.00000650, -0.00000587, 0.00000347, 0.00001132, -0.00000902, -0.00000972, 0.00000480, -0.00000721, -0.00001353, 0.00000216, 0.00000256, -0.00000493, 0.00000088, 0.00000333, -0.00000161, -0.00000275, -0.00000114, 0.00000033, -0.00000027, -0.00000142, -0.00000130, -0.00000206, 0.00000270, 0.00000766, -0.00000143, -0.00000453, 0.00000532, -0.00000182, -0.00000836, 0.00000844, 0.00000797, -0.00000734, 0.00000242, 0.00000494, -0.00001188, 0.00000101, 0.00001833, -0.00000728, -0.00002420, 0.00000778, 0.00002940, -0.00000105, -0.00002544, 0.00000332, 0.00001964, -0.00001389, -0.00001095, 0.00002859, -0.00000052, -0.00004507, -0.00000342, 0.00003387, 0.00000132, -0.00001025, 0.00001128, 0.00000307, -0.00001087, -0.00000265, 0.00000106, -0.00000335, -0.00000229, 0.00000324, 0.00000479, -0.00000437, -0.00000842, 0.00000131, 0.00000171, -0.00000209, 0.00000822, 0.00000658, -0.00000996, -0.00000597, 0.00000479, -0.00000379, -0.00000598, 0.00001021, 0.00001388, -0.00000448, -0.00001677, -0.00000774, 0.00000228, 0.00000184, 0.00000519, 0.00000659, -0.00000401, -0.00000488, 0.00000420, -0.00000100, -0.00000825, -0.00000367, -0.00000144, -0.00000087, 0.00000077, -0.00000310, -0.00000058, 0.00000660, 0.00000042, -0.00000550, -0.00000042, -0.00000063, -0.00000374, -0.00000045, 0.00000270, 0.00000458, 0.00000200, -0.00000642, -0.00000704, -0.00000345, -0.00000767, -0.00000982, -0.00000700, -0.00000324, 0.00000219, },
		{ 0.00000108, -0.00000106, -0.00003224, -0.00005434, -0.00002008, 0.00004090, 0.00006461, 0.00003938, 0.00002084, 0.00004554, 0.00005791, 0.00001919, -0.00001459, -0.00002287, -0.00003362, -0.00002184, 0.00000741, 0.00000235, -0.00000667, 0.00001024, 0.00000412, -0.00001424, -0.00000186, 0.00000290, -0.00001290, -0.00000731, 0.00000502, -0.00000119, -0.00000368, 0.00000253, 0.00000214, -0.00000225, -0.00000321, 0.00000001, 0.00000103, -0.00000183, -0.00000095, -0.00000109, -0.00000226, 0.00000608, 0.00000619, -0.00000602, -0.00000055, 0.00000600, -0.00000752, -0.00000329, 0.00001256, 0.00000069, -0.00000610, 0.00000652, -0.00000237, -0.00001066, 0.00001025, 0.00001160, -0.00001739, -0.00001624, 0.00002048, 0.00002494, -0.00001610, -0.00002253, 0.00001797, 0.00001181, -0.00002237, 0.00000755, 0.00002655, -0.00002889, -0.00003443, 0.00002540, 0.00002198, -0.00001287, 0.00000329, 0.00001170, -0.00000947, -0.00000716, 0.00000283, -0.00000354, -0.00000390, 0.00000239, 0.00000373, -0.00000005, -0.00000695, -0.00000449, 0.00000318, -0.00000010, 0.00000078, 0.00000903, 0.00000037, -0.00001032, -0.00000109, 0.00000332, -0.00000613, -0.00000192, 0.00001360, 0.00000971, -0.00001121, -0.00001599, -0.00000238, 0.00000262, 0.00000181, 0.00000749, 0.00000340, -0.00000667, -0.00000124, 0.00000351, -0.00000462, -0.00000658, -0.00000272, -0.00000171, 0.00000103, -0.00000066, -0.00000391, 0.00000319, 0.00000549, -0.00000327, -0.00000375, 0.00000044, -0.00000218, -0.00000286, 0.00000094, 0.00000436, 0.00000439, -0.00000272, -0.00000775, -0.00000455, -0.00000493, -0.00000865, -0.00000849, -0.00000622, },
		{ -0.00000841, -0.00000295, 0.00000403, -0.00001435, -0.00004700, -0.00004066, 0.00001074, 0.00005636, 0.00005433, 0.00002250, 0.00002403, 0.00005480, 0.00004258, 0.00000108, -0.00001305, -0.00002513, -0.00003642, -0.00000781, 0.00001247, -0.00000511, -0.00000139, 0.00001227, -0.00000609, -0.00001197, 0.00000398, -0.00000346, -0.00001323, -0.00000069, 0.00000389, -0.00000325, -0.00000179, 0.00000311, 0.00000143, -0.00000369, -0.00000284, 0.00000215, -0.00000039, -0.00000266, 0.00000064, -0.00000262, -0.00000061, 0.00000944, 0.00000099, -0.00000742, 0.00000507, 0.00000191, -0.00001014, 0.00000465, 0.00001083, -0.00000504, -0.00000120, 0.00000537, -0.00000818, -0.00000415, 0.00001390, 0.00000175, -0.00002139, -0.00000492, 0.00002979, 0.00001342, -0.00002948, -0.00000913, 0.00002776, -0.00000439, -0.00001854, 0.00002525, 0.00000552, -0.00004485, -0.00000437, 0.00003566, -0.00000058, -0.00000945, 0.00001512, -0.00000068, -0.00001391, 0.00000212, 0.00000053, -0.00000761, 0.00000086, 0.00000454, 0.00000046, -0.00000348, -0.00000644, -0.00000082, 0.00000337, -0.00000081, 0.00000337, 0.00000724, -0.00000430, -0.00000825, 0.00000208, 0.00000086, -0.00000696, 0.00000282, 0.00001548, 0.00000308, -0.00001650, -0.00001173, 0.00000149, 0.00000146, 0.00000373, 0.00000812, -0.00000122, -0.00000632, 0.00000156, 0.00000071, -0.00000586, -0.00000491, -0.00000294, -0.00000023, 0.00000169, -0.00000322, -0.00000202, 0.00000582, 0.00000212, -0.00000465, -0.00000162, -0.00000014, -0.00000291, -0.00000166, 0.00000285, 0.00000564, 0.00000116, -0.00000639, -0.00000622, -0.00000400, -0.00000645, -0.00000817, },
		{ -0.00000654, -0.00000853, -0.00000738, 0.00000273, -0.00000102, -0.00003209, -0.00004753, -0.00001626, 0.00003458, 0.00005974, 0.00003741, 0.00001210, 0.00003572, 0.00005496, 0.00002114, -0.00000544, -0.00001060, -0.00003377, -0.00003070, 0.00000788, 0.00000811, -0.00000892, 0.00000649, 0.00000693, -0.00001284, -0.00000485, 0.00000403, -0.00000918, -0.00000900, 0.00000292, 0.00000148, -0.00000361, -0.00000043, 0.00000389, -0.00000001, -0.00000549, -0.00000023, 0.00000289, -0.00000317, -0.00000100, 0.00000087, -0.00000465, 0.00000404, 0.00000943, -0.00000560, -0.00000368, 0.00000796, -0.00000487, -0.00000710, 0.00001067, 0.00000491, -0.00000616, 0.00000282, 0.00000060, -0.00000906, 0.00000267, 0.00001203, -0.00000688, -0.00002068, 0.00000857, 0.00003364, -0.00000553, -0.00003388, 0.00001203, 0.00002458, -0.00001974, -0.00000121, 0.00002812, -0.00002398, -0.00003555, 0.00002599, 0.00002225, -0.00001456, 0.00000647, 0.00001267, -0.00001474, -0.00000633, 0.00000675, -0.00000701, -0.00000534, 0.00000616, 0.00000193, -0.00000285, -0.00000454, -0.00000464, 0.00000212, 0.00000265, -0.00000091, 0.00000506, 0.00000431, -0.00000711, -0.00000512, 0.00000372, -0.00000213, -0.00000626, 0.00000830, 0.00001455, -0.00000536, -0.00001809, -0.00000574, 0.00000258, 0.00000101, 0.00000653, 0.00000596, -0.00000482, -0.00000396, 0.00000205, -0.00000191, -0.00000542, -0.00000450, -0.00000238, 0.00000192, 0.00000003, -0.00000455, 0.00000152, 0.00000577, -0.00000131, -0.00000400, -0.00000041, -0.00000118, -0.00000295, 0.00000010, 0.00000518, 0.00000447, -0.00000335, -0.00000704, -0.00000465, -0.00000484, },
		{ -0.00000449, -0.00000501, -0.00000668, -0.00000968, -0.00000260, 0.00000524, -0.00001534, -0.00004309, -0.00003352, 0.00000823, 0.00004933, 0.00005313, 0.00001812, 0.00001316, 0.00004880, 0.00004213, 0.00000392, -0.00000217, -0.00001566, -0.00004120, -0.00001482, 0.00001640, -0.00000130, -0.00000566, 0.00001059, -0.00000242, -0.00001255, 0.00000130, -0.00000032, -0.00001076, -0.00000395, 0.00000360, -0.00000024, -0.00000372, 0.00000122, 0.00000462, -0.00000328, -0.00000532, 0.00000346, 0.00000045, -0.00000438, 0.00000202, -0.00000182, -0.00000404, 0.00000953, 0.00000433, -0.00000909, 0.00000306, 0.00000539, -0.00000965, -0.00000002, 0.00001157, -0.00000091, -0.00000382, 0.00000323, -0.00000353, -0.00000628, 0.00000639, 0.00000801, -0.00001358, -0.00001598, 0.00002364, 0.00002748, -0.00002702, -0.00002289, 0.00003046, 0.00000777, -0.00002246, 0.00001904, 0.00001027, -0.00004212, -0.00000534, 0.00003659, -0.00000170, -0.00000922, 0.00001854, -0.00000320, -0.00001770, 0.00000565, 0.00000153, -0.00001156, 0.00000231, 0.00000660, -0.00000281, -0.00000385, -0.00000449, -0.00000215, 0.00000424, 0.00000094, -0.00000048, 0.00000609, 0.00000084, -0.00000826, -0.00000145, 0.00000367, -0.00000533, -0.00000326, 0.00001339, 0.00000965, -0.00001316, -0.00001532, -0.00000075, 0.00000180, 0.00000243, 0.00000813, 0.00000200, -0.00000606, -0.00000168, 0.00000089, -0.00000307, -0.00000502, -0.00000457, -0.00000024, 0.00000265, -0.00000282, -0.00000334, 0.00000450, 0.00000353, -0.00000327, -0.00000258, -0.00000022, -0.00000213, -0.00000227, 0.00000282, 0.00000622, 0.00000064, -0.00000621, -0.00000582, },
		{ -0.00000643, -0.00000502, -0.00000457, -0.00000440, -0.00000863, -0.00000832, 0.00000412, -0.00000145, -0.00003146, -0.00004021, -0.00001334, 0.00002743, 0.00005606, 0.00003722, 0.00000365, 0.00002607, 0.00005299, 0.00002126, -0.00000116, 0.00000133, -0.00002861, -0.00003834, 0.00000428, 0.00001387, -0.00000797, 0.00000154, 0.00000773, -0.00000948, -0.00000738, 0.00000288, -0.00000469, -0.00000886, -0.00000068, 0.00000343, -0.00000174, -0.00000375, 0.00000412, 0.00000329, -0.00000702, -0.00000165, 0.00000495, -0.00000378, -0.00000200, 0.00000303, -0.00000569, 0.00000090, 0.00001157, -0.00000350, -0.00000668, 0.00000795, -0.00000111, -0.00000915, 0.00000664, 0.00000819, -0.00000371, -0.00000152, 0.00000136, -0.00000476, -0.00000353, 0.00000792, 0.00000373, -0.00001942, -0.00000539, 0.00003613, 0.00000845, -0.00004020, 0.00000158, 0.00003424, -0.00001245, -0.00000880, 0.00002627, -0.00001838, -0.00003464, 0.00002564, 0.00002246, -0.00001570, 0.00000860, 0.00001404, -0.00001910, -0.00000701, 0.00001045, -0.00000880, -0.00000763, 0.00000883, 0.00000155, -0.00000556, -0.00000330, -0.00000413, 0.00000118, 0.00000473, -0.00000145, 0.00000106, 0.00000623, -0.00000297, -0.00000751, 0.00000213, 0.00000152, -0.00000750, 0.00000226, 0.00001583, 0.00000139, -0.00001751, -0.00000983, 0.00000169, 0.00000114, 0.00000501, 0.00000733, -0.00000180, -0.00000548, -0.00000068, -0.00000030, -0.00000333, -0.00000543, -0.00000355, 0.00000231, 0.00000105, -0.00000472, -0.00000033, 0.00000546, 0.00000067, -0.00000365, -0.00000135, -0.00000071, -0.00000276, -0.00000032, 0.00000560, 0.00000435, -0.00000360, },
		{ 0.00000017, -0.00000577, -0.00000562, -0.00000508, -0.00000345, -0.00000523, -0.00001085, -0.00000219, 0.00000604, -0.00001649, -0.00003864, -0.00002635, 0.00000479, 0.00004205, 0.00005349, 0.00001524, 0.00000213, 0.00004304, 0.00004243, 0.00000385, 0.00000432, -0.00000424, -0.00004086, -0.00002310, 0.00001663, 0.00000454, -0.00000771, 0.00000642, 0.00000068, -0.00001081, -0.00000240, 0.00000104, -0.00000635, -0.00000645, 0.00000132, 0.00000324, -0.00000410, -0.00000203, 0.00000705, -0.00000132, -0.00000791, 0.00000373, 0.00000209, -0.00000598, 0.00000240, -0.00000002, -0.00000672, 0.00000787, 0.00000762, -0.00000896, -0.00000004, 0.00000751, -0.00000700, -0.00000423, 0.00000966, 0.00000393, -0.00000428, -0.00000104, 0.00000034, -0.00000485, -0.00000221, 0.00000990, -0.00000251, -0.00002274, 0.00001216, 0.00003788, -0.00001851, -0.00003489, 0.00002723, 0.00002016, -0.00002161, 0.00001143, 0.00001250, -0.00003733, -0.00000546, 0.00003647, -0.00000220, -0.00000919, 0.00002101, -0.00000460, -0.00002145, 0.00000759, 0.00000344, -0.00001447, 0.00000223, 0.00000861, -0.00000471, -0.00000498, -0.00000309, -0.00000268, 0.00000463, 0.00000269, -0.00000311, 0.00000378, 0.00000467, -0.00000637, -0.00000466, 0.00000430, -0.00000237, -0.00000694, 0.00000887, 0.00001376, -0.00000761, -0.00001726, -0.00000428, 0.00000201, 0.00000187, 0.00000703, 0.00000466, -0.00000419, -0.00000446, -0.00000055, -0.00000074, -0.00000391, -0.00000581, -0.00000098, 0.00000340, -0.00000190, -0.00000438, 0.00000276, 0.00000445, -0.00000157, -0.00000306, -0.00000067, -0.00000167, -0.00000241, 0.00000292, 0.00000635, },
		{ 0.00000587, 0.00000389, -0.00000367, -0.00000571, -0.00000557, -0.00000434, -0.00000222, -0.00000867, -0.00000913, 0.00000534, -0.00000222, -0.00003073, -0.00003253, -0.00001062, 0.00001951, 0.00005259, 0.00003889, -0.00000375, 0.00001580, 0.00005156, 0.00002170, -0.00000087, 0.00001012, -0.00001939, -0.00004225, -0.00000267, 0.00001753, -0.00000370, -0.00000297, 0.00000575, -0.00000505, -0.00000824, -0.00000045, -0.00000068, -0.00000635, -0.00000490, 0.00000371, 0.00000165, -0.00000643, 0.00000253, 0.00000688, -0.00000715, -0.00000396, 0.00000656, -0.00000338, -0.00000361, 0.00000467, -0.00000525, -0.00000250, 0.00001196, -0.00000028, -0.00000852, 0.00000599, 0.00000256, -0.00000880, 0.00000150, 0.00000935, 0.00000083, -0.00000502, -0.00000066, 0.00000079, -0.00000670, 0.00000033, 0.00001187, -0.00001265, -0.00001824, 0.00003172, 0.00002266, -0.00004004, -0.00001093, 0.00003883, -0.00000190, -0.00001327, 0.00002138, -0.00001350, -0.00003141, 0.00002498, 0.00002223, -0.00001612, 0.00000979, 0.00001519, -0.00002220, -0.00000862, 0.00001322, -0.00000905, -0.00000996, 0.00001006, 0.00000207, -0.00000757, -0.00000313, -0.00000345, 0.00000068, 0.00000614, -0.00000123, -0.00000234, 0.00000634, 0.00000110, -0.00000793, -0.00000051, 0.00000378, -0.00000628, -0.00000286, 0.00001383, 0.00000731, -0.00001425, -0.00001346, -0.00000049, 0.00000174, 0.00000371, 0.00000739, 0.00000147, -0.00000532, -0.00000378, -0.00000032, -0.00000098, -0.00000514, -0.00000484, 0.00000197, 0.00000218, -0.00000431, -0.00000214, 0.00000457, 0.00000242, -0.00000278, -0.00000209, -0.00000076, -0.00000259, -0.00000027, },
		{ -0.00000231, 0.00000331, 0.00000621, -0.00000046, -0.00000514, -0.00000539, -0.00000579, -0.00000197, -0.00000373, -0.00001183, -0.00000179, 0.00000655, -0.00001779, -0.00003408, -0.00001895, 0.00000113, 0.00003411, 0.00005453, 0.00001442, -0.00000858, 0.00003642, 0.00004380, 0.00000324, 0.00000602, 0.00000615, -0.00003513, -0.00003007, 0.00001254, 0.00001054, -0.00000640, 0.00000077, 0.00000184, -0.00000681, -0.00000586, -0.00000005, -0.00000091, -0.00000708, -0.00000270, 0.00000609, -0.00000254, -0.00000582, 0.00000791, 0.00000187, -0.00000981, 0.00000274, 0.00000408, -0.00000688, 0.00000170, 0.00000210, -0.00000801, 0.00000493, 0.00000989, -0.00000698, -0.00000318, 0.00000760, -0.00000301, -0.00000673, 0.00000538, 0.00000800, -0.00000207, -0.00000623, 0.00000200, -0.00000039, -0.00000967, 0.00000697, 0.00000920, -0.00002363, -0.00000210, 0.00004216, -0.00000596, -0.00004238, 0.00001903, 0.00003017, -0.00001596, 0.00000437, 0.00001158, -0.00003151, -0.00000399, 0.00003551, -0.00000238, -0.00000899, 0.00002230, -0.00000531, -0.00002448, 0.00000803, 0.00000547, -0.00001598, 0.00000115, 0.00000991, -0.00000533, -0.00000625, -0.00000250, -0.00000264, 0.00000485, 0.00000412, -0.00000458, 0.00000103, 0.00000682, -0.00000335, -0.00000667, 0.00000307, 0.00000057, -0.00000792, 0.00000355, 0.00001475, -0.00000129, -0.00001680, -0.00000834, 0.00000134, 0.00000206, 0.00000554, 0.00000618, -0.00000131, -0.00000593, -0.00000314, 0.00000034, -0.00000192, -0.00000611, -0.00000225, 0.00000360, -0.00000058, -0.00000492, 0.00000075, 0.00000479, 0.00000023, -0.00000299, -0.00000126, -0.00000166, },
		{ -0.00000117, -0.00000264, 0.00000011, 0.00000624, 0.00000317, -0.00000380, -0.00000485, -0.00000609, -0.00000426, -0.00000016, -0.00000852, -0.00000979, 0.00000636, -0.00000317, -0.00003010, -0.00002485, -0.00000760, 0.00001132, 0.00004849, 0.00004186, -0.00000916, 0.00000488, 0.00004953, 0.00002366, -0.00000254, 0.00001409, -0.00000858, -0.00004081, -0.00001107, 0.00001736, 0.00000288, -0.00000534, 0.00000104, -0.00000094, -0.00000663, -0.00000511, 0.00000155, -0.00000166, -0.00000848, 0.00000185, 0.00000572, -0.00000753, -0.00000061, 0.00000980, -0.00000562, -0.00000653, 0.00000715, -0.00000203, -0.00000522, 0.00000527, -0.00000364, -0.00000524, 0.00001047, 0.00000308, -0.00000852, 0.00000271, 0.00000498, -0.00000610, -0.00000339, 0.00000774, 0.00000633, -0.00000639, -0.00000508, 0.00000577, -0.00000610, -0.00000880, 0.00001564, -0.00000199, -0.00002705, 0.00002135, 0.00003374, -0.00003356, -0.00002248, 0.00003757, 0.00000963, -0.00001332, 0.00001474, -0.00001059, -0.00002620, 0.00002480, 0.00002136, -0.00001588, 0.00001031, 0.00001559, -0.00002404, -0.00001046, 0.00001465, -0.00000826, -0.00001167, 0.00000990, 0.00000282, -0.00000854, -0.00000368, -0.00000302, 0.00000066, 0.00000703, -0.00000067, -0.00000482, 0.00000513, 0.00000432, -0.00000659, -0.00000314, 0.00000423, -0.00000370, -0.00000598, 0.00000964, 0.00001107, -0.00000907, -0.00001549, -0.00000379, 0.00000207, 0.00000311, 0.00000651, 0.00000412, -0.00000363, -0.00000628, -0.00000194, 0.00000077, -0.00000363, -0.00000567, 0.00000088, 0.00000308, -0.00000331, -0.00000368, 0.00000315, 0.00000373, -0.00000150, -0.00000244, },
		{ -0.00000237, -0.00000175, -0.00000206, -0.00000222, 0.00000401, 0.00000604, -0.00000129, -0.00000456, -0.00000500, -0.00000647, -0.00000076, -0.00000225, -0.00001251, -0.00000145, 0.00000683, -0.00001913, -0.00002977, -0.00001149, -0.00000202, 0.00002561, 0.00005516, 0.00001567, -0.00001795, 0.00002826, 0.00004544, 0.00000407, 0.00000409, 0.00001283, -0.00002536, -0.00003320, 0.00000481, 0.00001452, -0.00000155, -0.00000468, 0.00000016, -0.00000143, -0.00000771, -0.00000360, 0.00000420, -0.00000521, -0.00000755, 0.00000727, 0.00000075, -0.00000908, 0.00000683, 0.00000554, -0.00001030, 0.00000070, 0.00000573, -0.00000672, 0.00000024, 0.00000377, -0.00000771, 0.00000155, 0.00001041, -0.00000367, -0.00000529, 0.00000557, 0.00000111, -0.00000675, -0.00000029, 0.00000998, 0.00000248, -0.00001078, 0.00000072, 0.00000568, -0.00001414, 0.00000010, 0.00001882, -0.00001854, -0.00001587, 0.00003931, 0.00000754, -0.00004367, 0.00000806, 0.00003571, -0.00000666, -0.00000019, 0.00000774, -0.00002597, -0.00000059, 0.00003424, -0.00000258, -0.00000844, 0.00002241, -0.00000587, -0.00002638, 0.00000739, 0.00000693, -0.00001612, -0.00000027, 0.00001013, -0.00000506, -0.00000713, -0.00000271, -0.00000238, 0.00000517, 0.00000512, -0.00000515, -0.00000153, 0.00000730, -0.00000008, -0.00000707, 0.00000076, 0.00000248, -0.00000668, -0.00000108, 0.00001286, 0.00000425, -0.00001392, -0.00001185, -0.00000056, 0.00000257, 0.00000436, 0.00000647, 0.00000171, -0.00000571, -0.00000597, -0.00000025, 0.00000020, -0.00000517, -0.00000357, 0.00000304, 0.00000087, -0.00000481, -0.00000130, 0.00000445, 0.00000192, },
		{ 0.00000443, 0.00000002, -0.00000232, -0.00000175, -0.00000300, 0.00000062, 0.00000681, 0.00000240, -0.00000414, -0.00000399, -0.00000635, -0.00000428, 0.00000155, -0.00000815, -0.00001019, 0.00000711, -0.00000421, -0.00002956, -0.00001761, -0.00000413, 0.00000365, 0.00004336, 0.00004511, -0.00001194, -0.00000596, 0.00004568, 0.00002710, -0.00000375, 0.00001310, 0.00000078, -0.00003383, -0.00001825, 0.00001264, 0.00000987, -0.00000429, -0.00000542, 0.00000136, -0.00000253, -0.00000976, 0.00000118, 0.00000432, -0.00001020, -0.00000198, 0.00000929, -0.00000656, -0.00000451, 0.00001116, -0.00000267, -0.00000853, 0.00000642, -0.00000018, -0.00000620, 0.00000468, -0.00000155, -0.00000661, 0.00000752, 0.00000556, -0.00000653, -0.00000084, 0.00000533, -0.00000180, -0.00000665, 0.00000354, 0.00001128, -0.00000497, -0.00001076, 0.00000859, -0.00000211, -0.00001712, 0.00001434, 0.00000997, -0.00002993, 0.00000748, 0.00003901, -0.00002253, -0.00003026, 0.00003110, 0.00001961, -0.00000869, 0.00000815, -0.00001041, -0.00001989, 0.00002585, 0.00002001, -0.00001524, 0.00001047, 0.00001498, -0.00002493, -0.00001190, 0.00001471, -0.00000710, -0.00001235, 0.00000874, 0.00000320, -0.00000850, -0.00000445, -0.00000308, 0.00000099, 0.00000764, -0.00000006, -0.00000630, 0.00000322, 0.00000621, -0.00000409, -0.00000492, 0.00000304, -0.00000099, -0.00000677, 0.00000470, 0.00001205, -0.00000326, -0.00001528, -0.00000750, 0.00000148, 0.00000318, 0.00000535, 0.00000564, -0.00000100, -0.00000739, -0.00000461, 0.00000116, -0.00000135, -0.00000549, -0.00000070, 0.00000337, -0.00000184, -0.00000467, 0.00000133, },
		{ -0.00000310, 0.00000342, 0.00000327, -0.00000127, -0.00000198, -0.00000273, -0.00000228, 0.00000484, 0.00000604, -0.00000223, -0.00000422, -0.00000446, -0.00000689, 0.00000011, -0.00000099, -0.00001275, -0.00000108, 0.00000683, -0.00002038, -0.00002586, -0.00000436, -0.00000418, 0.00001713, 0.00005449, 0.00001838, -0.00002484, 0.00001867, 0.00004592, 0.00000730, 0.00000076, 0.00001426, -0.00001412, -0.00003085, -0.00000461, 0.00001463, 0.00000589, -0.00000811, -0.00000444, 0.00000391, -0.00000702, -0.00000881, 0.00000760, -0.00000088, -0.00001194, 0.00000618, 0.00000496, -0.00001081, 0.00000399, 0.00000867, -0.00000906, -0.00000178, 0.00000643, -0.00000550, -0.00000139, 0.00000442, -0.00000610, -0.00000132, 0.00000898, 0.00000000, -0.00000556, 0.00000202, 0.00000411, -0.00000424, -0.00000591, 0.00000912, 0.00000830, -0.00001329, -0.00000329, 0.00001147, -0.00001467, -0.00000882, 0.00002419, -0.00000890, -0.00002603, 0.00003004, 0.00001857, -0.00003866, -0.00000280, 0.00003576, 0.00000427, -0.00000083, 0.00000201, -0.00002198, 0.00000453, 0.00003341, -0.00000300, -0.00000757, 0.00002155, -0.00000673, -0.00002707, 0.00000627, 0.00000739, -0.00001526, -0.00000142, 0.00000925, -0.00000446, -0.00000731, -0.00000343, -0.00000222, 0.00000565, 0.00000576, -0.00000515, -0.00000350, 0.00000648, 0.00000269, -0.00000593, -0.00000170, 0.00000287, -0.00000419, -0.00000397, 0.00000900, 0.00000781, -0.00000932, -0.00001381, -0.00000351, 0.00000274, 0.00000385, 0.00000592, 0.00000409, -0.00000399, -0.00000810, -0.00000241, 0.00000157, -0.00000305, -0.00000432, 0.00000169, 0.00000203, -0.00000397, },
		{ -0.00000022, -0.00000486, -0.00000068, 0.00000439, 0.00000158, -0.00000170, -0.00000230, -0.00000362, 0.00000105, 0.00000753, 0.00000179, -0.00000468, -0.00000335, -0.00000628, -0.00000421, 0.00000275, -0.00000771, -0.00001015, 0.00000763, -0.00000533, -0.00002900, -0.00001111, -0.00000041, -0.00000280, 0.00003734, 0.00004748, -0.00001208, -0.00001545, 0.00003933, 0.00003090, -0.00000257, 0.00000854, 0.00000605, -0.00002271, -0.00002162, 0.00000401, 0.00001499, 0.00000049, -0.00001193, 0.00000079, 0.00000329, -0.00001293, -0.00000191, 0.00001023, -0.00000929, -0.00000693, 0.00001125, -0.00000366, -0.00000808, 0.00001049, 0.00000096, -0.00000920, 0.00000451, 0.00000151, -0.00000613, 0.00000318, 0.00000022, -0.00000633, 0.00000394, 0.00000635, -0.00000304, -0.00000348, 0.00000345, 0.00000286, -0.00000739, -0.00000232, 0.00001425, -0.00000094, -0.00001609, 0.00000833, 0.00000392, -0.00002259, 0.00000862, 0.00002035, -0.00002649, -0.00000663, 0.00003718, -0.00000989, -0.00003244, 0.00002135, 0.00002587, -0.00000023, 0.00000353, -0.00001292, -0.00001370, 0.00002851, 0.00001864, -0.00001450, 0.00001044, 0.00001331, -0.00002529, -0.00001253, 0.00001372, -0.00000619, -0.00001198, 0.00000715, 0.00000287, -0.00000776, -0.00000499, -0.00000362, 0.00000143, 0.00000813, 0.00000044, -0.00000694, 0.00000118, 0.00000670, -0.00000118, -0.00000539, 0.00000084, 0.00000089, -0.00000559, 0.00000037, 0.00001042, 0.00000180, -0.00001283, -0.00001065, -0.00000031, 0.00000351, 0.00000451, 0.00000601, 0.00000172, -0.00000683, -0.00000742, -0.00000014, 0.00000091, -0.00000402, -0.00000224, 0.00000279, },
		{ -0.00000013, 0.00000248, -0.00000253, -0.00000430, 0.00000179, 0.00000406, 0.00000018, -0.00000185, -0.00000352, -0.00000255, 0.00000562, 0.00000626, -0.00000303, -0.00000424, -0.00000397, -0.00000686, 0.00000075, -0.00000019, -0.00001260, -0.00000051, 0.00000650, -0.00002149, -0.00002232, 0.00000204, -0.00000526, 0.00000945, 0.00005212, 0.00002139, -0.00002853, 0.00000864, 0.00004388, 0.00001261, -0.00000143, 0.00001048, -0.00000451, -0.00002281, -0.00001302, 0.00001003, 0.00001403, -0.00000810, -0.00001107, 0.00000762, -0.00000359, -0.00001430, 0.00000830, 0.00000506, -0.00001457, 0.00000288, 0.00000889, -0.00001041, 0.00000013, 0.00001035, -0.00000625, -0.00000386, 0.00000581, -0.00000362, -0.00000251, 0.00000383, -0.00000383, -0.00000288, 0.00000594, 0.00000294, -0.00000378, -0.00000201, 0.00000506, 0.00000012, -0.00001003, 0.00000544, 0.00001383, -0.00001291, -0.00000895, 0.00001524, -0.00001142, -0.00001740, 0.00002434, 0.00000275, -0.00003042, 0.00001661, 0.00002432, -0.00002885, -0.00001077, 0.00003068, 0.00001440, 0.00000293, -0.00000392, -0.00002037, 0.00001061, 0.00003372, -0.00000353, -0.00000661, 0.00001998, -0.00000820, -0.00002677, 0.00000523, 0.00000680, -0.00001398, -0.00000193, 0.00000759, -0.00000403, -0.00000679, -0.00000428, -0.00000234, 0.00000621, 0.00000615, -0.00000486, -0.00000475, 0.00000486, 0.00000444, -0.00000371, -0.00000351, 0.00000184, -0.00000155, -0.00000478, 0.00000449, 0.00000880, -0.00000417, -0.00001366, -0.00000683, 0.00000200, 0.00000397, 0.00000513, 0.00000539, -0.00000141, -0.00000881, -0.00000550, 0.00000148, -0.00000029, -0.00000393, },
		{ -0.00000139, -0.00000304, 0.00000135, 0.00000108, -0.00000416, -0.00000253, 0.00000351, 0.00000295, -0.00000059, -0.00000269, -0.00000439, 0.00000129, 0.00000825, 0.00000150, -0.00000527, -0.00000313, -0.00000596, -0.00000381, 0.00000342, -0.00000742, -0.00000960, 0.00000802, -0.00000661, -0.00002832, -0.00000545, 0.00000314, -0.00000757, 0.00003106, 0.00004801, -0.00001030, -0.00002225, 0.00003073, 0.00003334, 0.00000182, 0.00000287, 0.00000573, -0.00001010, -0.00001944, -0.00000660, 0.00001626, 0.00000817, -0.00001659, -0.00000288, 0.00000937, -0.00001352, -0.00000702, 0.00001455, -0.00000569, -0.00001169, 0.00001087, 0.00000043, -0.00001025, 0.00000789, 0.00000429, -0.00000812, 0.00000199, 0.00000237, -0.00000495, 0.00000136, 0.00000104, -0.00000462, 0.00000074, 0.00000515, 0.00000100, -0.00000434, -0.00000016, 0.00000654, -0.00000543, -0.00000841, 0.00001431, 0.00000470, -0.00001954, 0.00000501, 0.00001010, -0.00002403, 0.00000026, 0.00002680, -0.00001792, -0.00001773, 0.00002871, 0.00000098, -0.00002868, 0.00001100, 0.00002725, 0.00001025, 0.00000237, -0.00001723, -0.00000889, 0.00003268, 0.00001795, -0.00001386, 0.00001021, 0.00001073, -0.00002557, -0.00001223, 0.00001219, -0.00000593, -0.00001087, 0.00000566, 0.00000179, -0.00000676, -0.00000506, -0.00000443, 0.00000175, 0.00000856, 0.00000079, -0.00000695, -0.00000057, 0.00000601, 0.00000142, -0.00000452, -0.00000153, 0.00000145, -0.00000328, -0.00000237, 0.00000702, 0.00000500, -0.00000879, -0.00001234, -0.00000308, 0.00000347, 0.00000430, 0.00000560, 0.00000381, -0.00000483, -0.00000940, -0.00000295, 0.00000221, },
		{ -0.00000039, 0.00000224, -0.00000213, -0.00000181, 0.00000192, -0.00000086, -0.00000461, -0.00000020, 0.00000408, 0.00000181, -0.00000131, -0.00000429, -0.00000297, 0.00000619, 0.00000660, -0.00000350, -0.00000455, -0.00000377, -0.00000634, 0.00000140, -0.00000000, -0.00001224, 0.00000043, 0.00000590, -0.00002252, -0.00001903, 0.00000742, -0.00000553, 0.00000332, 0.00004825, 0.00002343, -0.00002899, -0.00000033, 0.00003857, 0.00001857, -0.00000051, 0.00000307, 0.00000075, -0.00001048, -0.00001770, 0.00000122, 0.00002049, -0.00000415, -0.00001807, 0.00000846, 0.00000190, -0.00001851, 0.00000602, 0.00001124, -0.00001453, -0.00000185, 0.00001137, -0.00000790, -0.00000365, 0.00001001, -0.00000254, -0.00000474, 0.00000396, -0.00000174, -0.00000265, 0.00000222, -0.00000174, -0.00000277, 0.00000211, 0.00000427, -0.00000033, -0.00000530, 0.00000366, 0.00000509, -0.00001163, -0.00000021, 0.00001755, -0.00000966, -0.00001466, 0.00001589, -0.00000564, -0.00002345, 0.00001966, 0.00001350, -0.00002835, 0.00000224, 0.00002329, -0.00001701, -0.00001392, 0.00002212, 0.00002158, 0.00001048, -0.00000822, -0.00002123, 0.00001652, 0.00003563, -0.00000376, -0.00000584, 0.00001785, -0.00001031, -0.00002586, 0.00000468, 0.00000538, -0.00001281, -0.00000174, 0.00000561, -0.00000407, -0.00000582, -0.00000491, -0.00000273, 0.00000669, 0.00000639, -0.00000443, -0.00000531, 0.00000295, 0.00000503, -0.00000105, -0.00000418, -0.00000010, 0.00000035, -0.00000381, 0.00000057, 0.00000741, 0.00000025, -0.00001144, -0.00000960, 0.00000013, 0.00000427, 0.00000465, 0.00000563, 0.00000122, -0.00000786, -0.00000856, },
		{ -0.00000988, -0.00000655, 0.00000185, 0.00000177, -0.00000251, -0.00000060, 0.00000157, -0.00000270, -0.00000380, 0.00000185, 0.00000385, 0.00000093, -0.00000281, -0.00000520, 0.00000134, 0.00000877, 0.00000151, -0.00000566, -0.00000339, -0.00000562, -0.00000289, 0.00000374, -0.00000756, -0.00000860, 0.00000853, -0.00000809, -0.00002752, -0.00000056, 0.00000611, -0.00001065, 0.00002539, 0.00004631, -0.00000785, -0.00002550, 0.00002103, 0.00003282, 0.00000891, -0.00000120, -0.00000007, 0.00000079, -0.00001140, -0.00001647, 0.00001263, 0.00001690, -0.00001780, -0.00000893, 0.00001407, -0.00001115, -0.00001278, 0.00001617, -0.00000011, -0.00001489, 0.00000811, 0.00000455, -0.00001032, 0.00000404, 0.00000632, -0.00000540, -0.00000031, 0.00000197, -0.00000300, -0.00000011, 0.00000061, -0.00000212, -0.00000125, 0.00000223, 0.00000447, -0.00000308, -0.00000446, 0.00000821, -0.00000136, -0.00001319, 0.00001135, 0.00001043, -0.00002009, -0.00000040, 0.00001456, -0.00002140, -0.00000835, 0.00002805, -0.00000663, -0.00002342, 0.00001567, 0.00000720, -0.00002024, 0.00000272, 0.00002385, 0.00002050, 0.00000531, -0.00002187, -0.00000629, 0.00003771, 0.00001862, -0.00001326, 0.00000957, 0.00000748, -0.00002601, -0.00001117, 0.00001062, -0.00000640, -0.00000947, 0.00000460, 0.00000021, -0.00000594, -0.00000465, -0.00000520, 0.00000182, 0.00000888, 0.00000105, -0.00000657, -0.00000183, 0.00000454, 0.00000322, -0.00000266, -0.00000334, 0.00000071, -0.00000084, -0.00000320, 0.00000303, 0.00000585, -0.00000427, -0.00001207, -0.00000616, 0.00000256, 0.00000462, 0.00000502, 0.00000488, -0.00000208, },
		{ 0.00000319, -0.00000555, -0.00001061, -0.00000374, 0.00000357, 0.00000088, -0.00000261, 0.00000036, 0.00000051, -0.00000387, -0.00000218, 0.00000320, 0.00000339, -0.00000030, -0.00000492, -0.00000346, 0.00000647, 0.00000688, -0.00000354, -0.00000494, -0.00000404, -0.00000550, 0.00000235, -0.00000038, -0.00001198, 0.00000183, 0.00000521, -0.00002357, -0.00001589, 0.00001175, -0.00000545, -0.00000085, 0.00004359, 0.00002348, -0.00002698, -0.00000670, 0.00003028, 0.00002317, 0.00000434, -0.00000531, 0.00000009, 0.00000346, -0.00001677, -0.00000996, 0.00002316, 0.00000306, -0.00002347, 0.00000600, 0.00000807, -0.00002020, 0.00000135, 0.00001611, -0.00001161, -0.00000674, 0.00001156, -0.00000394, -0.00000625, 0.00000764, 0.00000107, -0.00000395, 0.00000137, -0.00000055, -0.00000166, 0.00000016, -0.00000052, -0.00000116, -0.00000149, 0.00000358, 0.00000386, -0.00000686, 0.00000033, 0.00000928, -0.00001037, -0.00000641, 0.00001841, -0.00000441, -0.00001880, 0.00001323, 0.00000082, -0.00002556, 0.00001179, 0.00002080, -0.00002078, -0.00000974, 0.00001572, -0.00000640, -0.00001175, 0.00001250, 0.00002454, 0.00002027, -0.00000939, -0.00002387, 0.00002109, 0.00003909, -0.00000307, -0.00000546, 0.00001516, -0.00001294, -0.00002471, 0.00000471, 0.00000358, -0.00001211, -0.00000109, 0.00000377, -0.00000463, -0.00000479, -0.00000512, -0.00000322, 0.00000693, 0.00000652, -0.00000396, -0.00000531, 0.00000115, 0.00000458, 0.00000141, -0.00000366, -0.00000223, 0.00000104, -0.00000180, -0.00000190, 0.00000441, 0.00000297, -0.00000778, -0.00001100, -0.00000261, 0.00000417, 0.00000473, 0.00000521, },
		},
		{ { -0.00000364, -0.00000414, -0.00000084, 0.00000667, 0.00000638, -0.00000261, -0.00000424, 0.00000341, 0.00000421, -0.00000187, -0.00000111, 0.00000225, -0.00000201, -0.00000526, -0.00000051, 0.00000275, 0.00000052, -0.00000204, -0.00000496, -0.00000641, -0.00000095, 0.00000616, 0.00000712, 0.00000469, 0.00000140, -0.00000307, -0.00000348, 0.00000085, 0.00000053, -0.00000757, -0.00001338, -0.00000541, 0.00001028, 0.00000814, -0.00001237, -0.00001046, 0.00001554, 0.00000925, -0.00002429, -0.00001774, 0.00001094, -0.00000204, -0.00001614, 0.00001690, 0.00003142, -0.00000108, -0.00000705, 0.00001285, -0.00000026, -0.00000646, 0.00002410, 0.00001339, -0.00003348, -0.00001290, 0.00003815, 0.00001026, -0.00003760, -0.00000930, 0.00002841, 0.00000455, -0.00002035, -0.00000221, 0.00001351, 0.00000115, -0.00000992, 0.00000115, 0.00001351, -0.00000221, -0.00002035, 0.00000455, 0.00002841, -0.00000930, -0.00003760, 0.00001026, 0.00003815, -0.00001290, -0.00003348, 0.00001339, 0.00002410, -0.00000646, -0.00000026, 0.00001285, -0.00000705, -0.00000108, 0.00003142, 0.00001690, -0.00001614, -0.00000204, 0.00001094, -0.00001774, -0.00002429, 0.00000925, 0.00001554, -0.00001046, -0.00001237, 0.00000814, 0.00001028, -0.00000541, -0.00001338, -0.00000757, 0.00000053, 0.00000085, -0.00000348, -0.00000307, 0.00000140, 0.00000469, 0.00000712, 0.00000616, -0.00000095, -0.00000641, -0.00000496, -0.00000204, 0.00000052, 0.00000275, -0.00000051, -0.00000526, -0.00000201, 0.00000225, -0.00000111, -0.00000187, 0.00000421, 0.00000341, -0.00000424, -0.00000261, 0.00000638, 0.00000667, -0.00000084, -0.00000414, },
		{ -0.00000285, -0.00000403, -0.00000372, -0.00000378, 0.00000195, 0.00000867, 0.00000386, -0.00000556, -0.00000282, 0.00000500, 0.00000222, -0.00000257, 0.00000123, 0.00000195, -0.00000489, -0.00000499, 0.00000206, 0.00000325, -0.00000045, -0.00000277, -0.00000547, -0.00000580, 0.00000049, 0.00000657, 0.00000685, 0.00000442, 0.00000042, -0.00000366, -0.00000260, 0.00000130, -0.00000082, -0.00000976, -0.00001375, -0.00000130, 0.00001372, 0.00000325, -0.00001649, -0.00000256, 0.00001981, -0.00000227, -0.00002814, -0.00000403, 0.00001347, -0.00001335, -0.00001150, 0.00002870, 0.00002290, -0.00001144, 0.00000010, 0.00001450, -0.00000865, -0.00000161, 0.00002991, 0.00000085, -0.00003855, 0.00000156, 0.00004009, -0.00000325, -0.00003631, 0.00000141, 0.00002574, -0.00000168, -0.00001725, 0.00000099, 0.00001211, -0.00000034, -0.00001032, 0.00000349, 0.00001493, -0.00000765, -0.00002169, 0.00001324, 0.00002714, -0.00002124, -0.00003327, 0.00002269, 0.00002998, -0.00002385, -0.00002314, 0.00002110, 0.00001607, -0.00000601, 0.00000653, 0.00000683, -0.00000871, 0.00001224, 0.00003235, 0.00000433, -0.00001337, 0.00000658, 0.00000114, -0.00002697, -0.00001432, 0.00001629, 0.00000737, -0.00001449, -0.00000631, 0.00001013, 0.00000586, -0.00000797, -0.00001231, -0.00000551, 0.00000153, 0.00000004, -0.00000407, -0.00000222, 0.00000205, 0.00000494, 0.00000755, 0.00000542, -0.00000249, -0.00000669, -0.00000427, -0.00000120, 0.00000111, 0.00000140, -0.00000259, -0.00000410, 0.00000039, 0.00000112, -0.00000282, -0.00000018, 0.00000528, 0.00000142, -0.00000425, 0.00000064, 0.00000734, 0.00000391, },
		{ 0.00000683, 0.00000100, -0.00000402, -0.00000372, -0.00000392, -0.00000267, 0.00000523, 0.00000932, 0.00000005, -0.00000738, -0.00000027, 0.00000553, -0.00000019, -0.00000190, 0.00000343, 0.00000006, -0.00000737, -0.00000319, 0.00000443, 0.00000283, -0.00000149, -0.00000332, -0.00000577, -0.00000496, 0.00000170, 0.00000683, 0.00000677, 0.00000392, -0.00000075, -0.00000383, -0.00000161, 0.00000134, -0.00000250, -0.00001196, -0.00001287, 0.00000395, 0.00001486, -0.00000353, -0.00001676, 0.00000707, 0.00001808, -0.00001469, -0.00002404, 0.00000939, 0.00000775, -0.00002233, -0.00000007, 0.00003520, 0.00000917, -0.00001547, 0.00000943, 0.00001046, -0.00001511, 0.00000755, 0.00003059, -0.00001373, -0.00003671, 0.00001592, 0.00003571, -0.00001466, -0.00003071, 0.00000891, 0.00002122, -0.00000494, -0.00001424, 0.00000260, 0.00001157, -0.00000209, -0.00001104, 0.00000735, 0.00001491, -0.00001484, -0.00001954, 0.00002252, 0.00002091, -0.00003157, -0.00002375, 0.00003091, 0.00001768, -0.00002883, -0.00001074, 0.00002310, 0.00000926, -0.00000148, 0.00000908, -0.00000032, -0.00000362, 0.00002406, 0.00002616, -0.00000456, -0.00000554, 0.00000868, -0.00001227, -0.00002882, -0.00000201, 0.00001717, -0.00000155, -0.00001414, -0.00000055, 0.00000934, 0.00000175, -0.00000916, -0.00001097, -0.00000355, 0.00000211, -0.00000101, -0.00000426, -0.00000133, 0.00000240, 0.00000539, 0.00000797, 0.00000426, -0.00000397, -0.00000662, -0.00000342, -0.00000036, 0.00000108, -0.00000048, -0.00000363, -0.00000198, 0.00000167, -0.00000087, -0.00000345, 0.00000191, 0.00000534, -0.00000032, -0.00000303, 0.00000348, },
		{ -0.00000101, 0.00000538, 0.00000516, -0.00000162, -0.00000443, -0.00000343, -0.00000402, -0.00000062, 0.00000845, 0.00000810, -0.00000434, -0.00000746, 0.00000274, 0.00000466, -0.00000227, 0.00000009, 0.00000460, -0.00000311, -0.00000860, -0.00000022, 0.00000601, 0.00000170, -0.00000236, -0.00000373, -0.00000587, -0.00000401, 0.00000266, 0.00000711, 0.00000676, 0.00000304, -0.00000192, -0.00000354, -0.00000071, 0.00000094, -0.00000453, -0.00001383, -0.00001031, 0.00000931, 0.00001289, -0.00001022, -0.00001247, 0.00001529, 0.00001031, -0.00002372, -0.00001299, 0.00001772, -0.00000405, -0.00002477, 0.00001469, 0.00003363, -0.00000543, -0.00001216, 0.00001684, 0.00000181, -0.00001676, 0.00001816, 0.00002493, -0.00002667, -0.00002851, 0.00002682, 0.00002685, -0.00002191, -0.00002326, 0.00001277, 0.00001701, -0.00000607, -0.00001247, 0.00000390, 0.00001172, -0.00000497, -0.00001094, 0.00001245, 0.00001208, -0.00002215, -0.00001318, 0.00002983, 0.00001035, -0.00003760, -0.00001128, 0.00003303, 0.00000464, -0.00002740, 0.00000022, 0.00002070, 0.00000614, 0.00000427, 0.00000674, -0.00000477, 0.00000679, 0.00003056, 0.00001589, -0.00000713, 0.00000299, 0.00000307, -0.00002439, -0.00002312, 0.00000845, 0.00001255, -0.00000829, -0.00001049, 0.00000322, 0.00000675, -0.00000128, -0.00000949, -0.00000953, -0.00000169, 0.00000218, -0.00000211, -0.00000405, -0.00000062, 0.00000261, 0.00000612, 0.00000817, 0.00000270, -0.00000519, -0.00000620, -0.00000246, 0.00000028, 0.00000032, -0.00000237, -0.00000340, 0.00000036, 0.00000159, -0.00000299, -0.00000292, 0.00000382, 0.00000461, -0.00000134, },
		{ 0.00000351, -0.00000148, 0.00000125, 0.00000614, 0.00000278, -0.00000364, -0.00000426, -0.00000331, -0.00000375, 0.00000239, 0.00001089, 0.00000489, -0.00000833, -0.00000558, 0.00000528, 0.00000248, -0.00000324, 0.00000286, 0.00000411, -0.00000676, -0.00000807, 0.00000322, 0.00000646, 0.00000023, -0.00000295, -0.00000403, -0.00000577, -0.00000308, 0.00000344, 0.00000753, 0.00000662, 0.00000174, -0.00000283, -0.00000284, -0.00000006, 0.00000007, -0.00000684, -0.00001483, -0.00000609, 0.00001342, 0.00000802, -0.00001459, -0.00000456, 0.00001912, -0.00000127, -0.00002601, 0.00000142, 0.00001788, -0.00001743, -0.00001890, 0.00002786, 0.00002408, -0.00001627, -0.00000327, 0.00001918, -0.00000844, -0.00001257, 0.00002666, 0.00001397, -0.00003477, -0.00001642, 0.00003229, 0.00001645, -0.00002468, -0.00001639, 0.00001403, 0.00001444, -0.00000664, -0.00001196, 0.00000604, 0.00001157, -0.00000900, -0.00000880, 0.00001752, 0.00000590, -0.00002735, -0.00000336, 0.00003287, -0.00000236, -0.00003795, 0.00000091, 0.00002918, -0.00000572, -0.00002119, 0.00000730, 0.00001671, 0.00000730, 0.00000803, 0.00000118, -0.00000374, 0.00001903, 0.00003007, 0.00000584, -0.00000359, 0.00000768, -0.00000841, -0.00003101, -0.00001235, 0.00001394, 0.00000487, -0.00001118, -0.00000556, 0.00000437, 0.00000366, -0.00000309, -0.00000941, -0.00000794, -0.00000003, 0.00000173, -0.00000301, -0.00000356, -0.00000020, 0.00000291, 0.00000709, 0.00000793, 0.00000089, -0.00000604, -0.00000549, -0.00000149, 0.00000049, -0.00000105, -0.00000372, -0.00000202, 0.00000221, 0.00000030, -0.00000461, -0.00000150, 0.00000517, },
		{ 0.00000039, 0.00000584, 0.00000246, -0.00000083, 0.00000327, 0.00000576, 0.00000010, -0.00000491, -0.00000376, -0.00000337, -0.00000274, 0.00000606, 0.00001171, 0.00000005, -0.00001082, -0.00000214, 0.00000648, -0.00000039, -0.00000260, 0.00000549, 0.00000176, -0.00000986, -0.00000575, 0.00000629, 0.00000579, -0.00000119, -0.00000328, -0.00000426, -0.00000551, -0.00000225, 0.00000419, 0.00000813, 0.00000612, 0.00000014, -0.00000325, -0.00000188, 0.00000020, -0.00000128, -0.00000915, -0.00001441, -0.00000083, 0.00001506, 0.00000154, -0.00001498, 0.00000452, 0.00001695, -0.00001282, -0.00002046, 0.00001423, 0.00000974, -0.00002720, -0.00000613, 0.00003489, 0.00000947, -0.00002018, 0.00000754, 0.00001538, -0.00001666, -0.00000373, 0.00003019, 0.00000056, -0.00003656, -0.00000384, 0.00003236, 0.00000726, -0.00002424, -0.00001155, 0.00001453, 0.00001353, -0.00000809, -0.00001175, 0.00000936, 0.00000993, -0.00001323, -0.00000409, 0.00002074, -0.00000287, -0.00002856, 0.00000781, 0.00003058, -0.00001428, -0.00003308, 0.00000981, 0.00002140, -0.00001122, -0.00001318, 0.00001001, 0.00001414, 0.00001124, 0.00000771, -0.00000433, 0.00000339, 0.00002897, 0.00002363, -0.00000023, 0.00000318, 0.00000563, -0.00002165, -0.00003007, -0.00000053, 0.00001349, -0.00000274, -0.00001019, -0.00000140, 0.00000332, 0.00000113, -0.00000403, -0.00000915, -0.00000614, 0.00000124, 0.00000086, -0.00000356, -0.00000304, -0.00000005, 0.00000352, 0.00000810, 0.00000714, -0.00000095, -0.00000643, -0.00000458, -0.00000065, 0.00000013, -0.00000273, -0.00000412, 0.00000006, 0.00000305, -0.00000171, -0.00000533, },
		{ -0.00000384, -0.00000506, 0.00000232, 0.00000594, 0.00000174, 0.00000035, 0.00000470, 0.00000441, -0.00000248, -0.00000538, -0.00000320, -0.00000346, -0.00000068, 0.00000978, 0.00001029, -0.00000553, -0.00001104, 0.00000197, 0.00000586, -0.00000300, -0.00000039, 0.00000697, -0.00000203, -0.00001147, -0.00000213, 0.00000835, 0.00000434, -0.00000231, -0.00000346, -0.00000439, -0.00000516, -0.00000157, 0.00000507, 0.00000877, 0.00000506, -0.00000153, -0.00000306, -0.00000087, -0.00000009, -0.00000306, -0.00001099, -0.00001232, 0.00000431, 0.00001370, -0.00000449, -0.00001114, 0.00001167, 0.00000934, -0.00002027, -0.00000880, 0.00002094, -0.00000376, -0.00002951, 0.00000953, 0.00003341, -0.00000555, -0.00001655, 0.00001614, 0.00000680, -0.00002010, 0.00000683, 0.00002772, -0.00001181, -0.00003264, 0.00000626, 0.00002875, 0.00000087, -0.00002270, -0.00000873, 0.00001593, 0.00001317, -0.00001104, -0.00001047, 0.00001319, 0.00000610, -0.00001617, 0.00000265, 0.00002059, -0.00001234, -0.00002503, 0.00001764, 0.00002358, -0.00002279, -0.00002510, 0.00001376, 0.00001278, -0.00001168, -0.00000638, 0.00000984, 0.00001486, 0.00001512, 0.00000329, -0.00000634, 0.00001476, 0.00003344, 0.00001442, -0.00000065, 0.00000890, -0.00000327, -0.00003195, -0.00002233, 0.00000843, 0.00000835, -0.00000776, -0.00000672, 0.00000069, 0.00000118, -0.00000039, -0.00000458, -0.00000869, -0.00000417, 0.00000194, -0.00000017, -0.00000374, -0.00000270, 0.00000003, 0.00000453, 0.00000889, 0.00000583, -0.00000260, -0.00000640, -0.00000355, -0.00000008, -0.00000084, -0.00000427, -0.00000343, 0.00000221, 0.00000268, },
		{ 0.00000383, 0.00000128, -0.00000556, -0.00000396, 0.00000398, 0.00000569, 0.00000147, 0.00000173, 0.00000534, 0.00000232, -0.00000463, -0.00000517, -0.00000279, -0.00000325, 0.00000248, 0.00001263, 0.00000645, -0.00001057, -0.00000876, 0.00000553, 0.00000354, -0.00000439, 0.00000274, 0.00000655, -0.00000631, -0.00001103, 0.00000193, 0.00000907, 0.00000257, -0.00000304, -0.00000358, -0.00000440, -0.00000482, -0.00000099, 0.00000624, 0.00000924, 0.00000335, -0.00000295, -0.00000223, -0.00000010, -0.00000102, -0.00000502, -0.00001181, -0.00000881, 0.00000802, 0.00000989, -0.00000804, -0.00000439, 0.00001430, -0.00000109, -0.00002087, 0.00000496, 0.00001923, -0.00001774, -0.00002327, 0.00002295, 0.00002402, -0.00001643, -0.00000745, 0.00001955, -0.00000347, -0.00001790, 0.00001572, 0.00002036, -0.00002032, -0.00002534, 0.00001235, 0.00002397, -0.00000277, -0.00002190, -0.00000666, 0.00001891, 0.00001176, -0.00001492, -0.00000718, 0.00001618, 0.00000036, -0.00001646, 0.00000986, 0.00001654, -0.00002015, -0.00001755, 0.00002383, 0.00001397, -0.00002662, -0.00001696, 0.00001301, 0.00000622, -0.00000876, -0.00000256, 0.00000942, 0.00001883, 0.00001618, -0.00000316, -0.00000267, 0.00002683, 0.00003152, 0.00000635, 0.00000352, 0.00000976, -0.00001635, -0.00003584, -0.00001084, 0.00001210, 0.00000119, -0.00000907, -0.00000279, 0.00000053, -0.00000088, -0.00000106, -0.00000507, -0.00000785, -0.00000222, 0.00000202, -0.00000111, -0.00000369, -0.00000264, 0.00000029, 0.00000587, 0.00000921, 0.00000415, -0.00000390, -0.00000602, -0.00000253, 0.00000004, -0.00000228, -0.00000522, -0.00000179, },
		{ -0.00000527, 0.00000037, 0.00000450, -0.00000075, -0.00000652, -0.00000233, 0.00000524, 0.00000530, 0.00000161, 0.00000305, 0.00000514, -0.00000019, -0.00000609, -0.00000449, -0.00000262, -0.00000230, 0.00000647, 0.00001361, 0.00000067, -0.00001374, -0.00000451, 0.00000746, 0.00000028, -0.00000396, 0.00000574, 0.00000408, -0.00000989, -0.00000863, 0.00000548, 0.00000857, 0.00000094, -0.00000351, -0.00000367, -0.00000426, -0.00000463, -0.00000035, 0.00000775, 0.00000921, 0.00000108, -0.00000375, -0.00000089, 0.00000011, -0.00000257, -0.00000672, -0.00001120, -0.00000474, 0.00000940, 0.00000514, -0.00000793, 0.00000269, 0.00001145, -0.00001048, -0.00001430, 0.00001597, 0.00000980, -0.00002719, -0.00001053, 0.00002987, 0.00001003, -0.00002033, 0.00000333, 0.00001700, -0.00001197, -0.00001138, 0.00002044, 0.00001077, -0.00002386, -0.00001759, 0.00001473, 0.00002016, -0.00000504, -0.00002258, -0.00000364, 0.00002286, 0.00000813, -0.00001837, -0.00000188, 0.00001698, -0.00000608, -0.00001356, 0.00001558, 0.00000937, -0.00002439, -0.00000816, 0.00002538, 0.00000443, -0.00002626, -0.00001120, 0.00000948, 0.00000331, -0.00000510, -0.00000163, 0.00001122, 0.00002415, 0.00001296, -0.00000837, 0.00000665, 0.00003585, 0.00002474, 0.00000233, 0.00000906, 0.00000395, -0.00002924, -0.00003226, 0.00000034, 0.00001028, -0.00000505, -0.00000717, -0.00000013, -0.00000116, -0.00000211, -0.00000133, -0.00000552, -0.00000654, -0.00000059, 0.00000159, -0.00000177, -0.00000363, -0.00000280, 0.00000096, 0.00000730, 0.00000894, 0.00000234, -0.00000479, -0.00000539, -0.00000164, -0.00000039, -0.00000389, },
		{ -0.00000137, -0.00000529, -0.00000432, 0.00000249, 0.00000408, -0.00000289, -0.00000662, -0.00000051, 0.00000609, 0.00000493, 0.00000207, 0.00000407, 0.00000409, -0.00000279, -0.00000672, -0.00000365, -0.00000257, -0.00000024, 0.00001053, 0.00001199, -0.00000593, -0.00001413, 0.00000054, 0.00000718, -0.00000277, -0.00000178, 0.00000745, 0.00000010, -0.00001177, -0.00000492, 0.00000786, 0.00000733, -0.00000033, -0.00000389, -0.00000365, -0.00000404, -0.00000468, 0.00000058, 0.00000949, 0.00000836, -0.00000150, -0.00000364, 0.00000067, -0.00000053, -0.00000451, -0.00000755, -0.00000922, -0.00000128, 0.00000835, 0.00000140, -0.00000442, 0.00000725, 0.00000427, -0.00001522, -0.00000295, 0.00002034, -0.00000384, -0.00002886, 0.00000434, 0.00002844, -0.00000391, -0.00001706, 0.00001189, 0.00000999, -0.00001627, -0.00000326, 0.00002028, 0.00000200, -0.00002323, -0.00001171, 0.00001510, 0.00001817, -0.00000778, -0.00002408, 0.00000159, 0.00002622, 0.00000212, -0.00001984, 0.00000442, 0.00001487, -0.00001150, -0.00000806, 0.00001823, 0.00000096, -0.00002435, 0.00000063, 0.00002289, -0.00000280, -0.00002364, -0.00000897, 0.00000583, 0.00000377, -0.00000305, -0.00000179, 0.00001639, 0.00002801, 0.00000608, -0.00000922, 0.00001937, 0.00003935, 0.00001626, 0.00000303, 0.00001204, -0.00000754, -0.00003770, -0.00002288, 0.00000779, 0.00000472, -0.00000836, -0.00000369, 0.00000045, -0.00000324, -0.00000241, -0.00000165, -0.00000572, -0.00000485, 0.00000047, 0.00000092, -0.00000214, -0.00000378, -0.00000297, 0.00000211, 0.00000854, 0.00000812, 0.00000061, -0.00000531, -0.00000459, -0.00000100, },
		{ -0.00000372, -0.00000077, -0.00000277, -0.00000607, -0.00000255, 0.00000404, 0.00000273, -0.00000467, -0.00000595, 0.00000127, 0.00000660, 0.00000466, 0.00000272, 0.00000462, 0.00000229, -0.00000511, -0.00000656, -0.00000294, -0.00000227, 0.00000306, 0.00001360, 0.00000759, -0.00001180, -0.00001155, 0.00000494, 0.00000489, -0.00000448, 0.00000139, 0.00000711, -0.00000427, -0.00001145, -0.00000091, 0.00000884, 0.00000591, -0.00000129, -0.00000431, -0.00000340, -0.00000388, -0.00000493, 0.00000205, 0.00001119, 0.00000648, -0.00000393, -0.00000247, 0.00000195, -0.00000219, -0.00000632, -0.00000703, -0.00000649, 0.00000050, 0.00000582, 0.00000029, 0.00000074, 0.00000734, -0.00000428, -0.00001338, 0.00000898, 0.00001663, -0.00001674, -0.00002242, 0.00001637, 0.00001979, -0.00001357, -0.00000890, 0.00001563, 0.00000148, -0.00001584, 0.00000350, 0.00001647, -0.00000380, -0.00002060, -0.00000853, 0.00001559, 0.00001736, -0.00001228, -0.00002479, 0.00000913, 0.00002727, -0.00000527, -0.00001841, 0.00001011, 0.00001022, -0.00001447, -0.00000153, 0.00001728, -0.00000650, -0.00002076, 0.00000684, 0.00001822, -0.00000684, -0.00002120, -0.00000972, 0.00000425, 0.00000579, -0.00000355, -0.00000052, 0.00002418, 0.00002800, -0.00000200, -0.00000412, 0.00003201, 0.00003707, 0.00000941, 0.00000673, 0.00000953, -0.00002136, -0.00003915, -0.00001117, 0.00000986, -0.00000178, -0.00000826, -0.00000052, -0.00000081, -0.00000474, -0.00000216, -0.00000218, -0.00000540, -0.00000309, 0.00000091, 0.00000029, -0.00000238, -0.00000422, -0.00000286, 0.00000362, 0.00000936, 0.00000691, -0.00000091, -0.00000551, },
		{ -0.00000220, -0.00000544, -0.00000284, -0.00000105, -0.00000432, -0.00000597, -0.00000037, 0.00000466, 0.00000085, -0.00000579, -0.00000475, 0.00000283, 0.00000685, 0.00000454, 0.00000346, 0.00000453, -0.00000008, -0.00000680, -0.00000582, -0.00000250, -0.00000121, 0.00000726, 0.00001453, 0.00000104, -0.00001541, -0.00000670, 0.00000741, 0.00000152, -0.00000423, 0.00000434, 0.00000462, -0.00000773, -0.00000916, 0.00000237, 0.00000872, 0.00000475, -0.00000218, -0.00000471, -0.00000282, -0.00000404, -0.00000522, 0.00000424, 0.00001240, 0.00000353, -0.00000564, -0.00000033, 0.00000239, -0.00000476, -0.00000728, -0.00000505, -0.00000408, 0.00000015, 0.00000350, 0.00000232, 0.00000496, 0.00000280, -0.00001054, -0.00000553, 0.00001700, 0.00000638, -0.00002438, -0.00001052, 0.00002183, 0.00000745, -0.00001664, 0.00000047, 0.00001414, -0.00000548, -0.00001205, 0.00000698, 0.00001141, -0.00000612, -0.00001840, -0.00000725, 0.00001762, 0.00001607, -0.00001856, -0.00002291, 0.00001780, 0.00002489, -0.00001233, -0.00001418, 0.00001372, 0.00000430, -0.00001442, 0.00000419, 0.00001346, -0.00001133, -0.00001543, 0.00000975, 0.00001357, -0.00000836, -0.00002079, -0.00001155, 0.00000553, 0.00000704, -0.00000580, 0.00000415, 0.00003227, 0.00002326, -0.00000800, 0.00000637, 0.00004128, 0.00003080, 0.00000615, 0.00001009, 0.00000094, -0.00003316, -0.00003363, -0.00000101, 0.00000711, -0.00000665, -0.00000571, 0.00000103, -0.00000297, -0.00000524, -0.00000188, -0.00000274, -0.00000451, -0.00000159, 0.00000088, -0.00000014, -0.00000272, -0.00000482, -0.00000229, 0.00000525, 0.00000967, 0.00000552, },
		{ 0.00000955, 0.00000408, -0.00000330, -0.00000514, -0.00000210, -0.00000189, -0.00000564, -0.00000491, 0.00000168, 0.00000426, -0.00000110, -0.00000613, -0.00000327, 0.00000409, 0.00000691, 0.00000460, 0.00000412, 0.00000366, -0.00000269, -0.00000762, -0.00000484, -0.00000221, 0.00000104, 0.00001149, 0.00001251, -0.00000630, -0.00001576, -0.00000102, 0.00000736, -0.00000159, -0.00000214, 0.00000580, 0.00000073, -0.00000922, -0.00000577, 0.00000431, 0.00000813, 0.00000397, -0.00000328, -0.00000487, -0.00000197, -0.00000475, -0.00000519, 0.00000713, 0.00001259, -0.00000021, -0.00000608, 0.00000236, 0.00000144, -0.00000772, -0.00000667, -0.00000210, -0.00000318, -0.00000182, 0.00000314, 0.00000655, 0.00000582, -0.00000455, -0.00001157, 0.00000537, 0.00001808, -0.00000649, -0.00002433, 0.00000236, 0.00001959, -0.00000412, -0.00001342, 0.00000757, 0.00000905, -0.00000900, -0.00000738, 0.00000702, 0.00000758, -0.00000614, -0.00001816, -0.00000604, 0.00002129, 0.00001258, -0.00002537, -0.00001743, 0.00002567, 0.00001916, -0.00001736, -0.00000829, 0.00001449, -0.00000125, -0.00001189, 0.00000768, 0.00000842, -0.00001302, -0.00001043, 0.00001003, 0.00001055, -0.00000916, -0.00002290, -0.00001223, 0.00000880, 0.00000583, -0.00000762, 0.00001269, 0.00003780, 0.00001494, -0.00000915, 0.00001981, 0.00004536, 0.00002343, 0.00000627, 0.00000982, -0.00001173, -0.00003939, -0.00002358, 0.00000483, 0.00000170, -0.00000849, -0.00000243, 0.00000062, -0.00000496, -0.00000496, -0.00000192, -0.00000301, -0.00000320, -0.00000060, 0.00000066, -0.00000041, -0.00000333, -0.00000533, -0.00000124, 0.00000674, },
		{ 0.00000015, 0.00000794, 0.00000914, 0.00000262, -0.00000424, -0.00000457, -0.00000165, -0.00000319, -0.00000635, -0.00000309, 0.00000310, 0.00000302, -0.00000267, -0.00000578, -0.00000175, 0.00000502, 0.00000688, 0.00000487, 0.00000451, 0.00000199, -0.00000511, -0.00000755, -0.00000397, -0.00000169, 0.00000462, 0.00001453, 0.00000742, -0.00001262, -0.00001275, 0.00000378, 0.00000510, -0.00000318, 0.00000086, 0.00000501, -0.00000325, -0.00000840, -0.00000249, 0.00000488, 0.00000778, 0.00000331, -0.00000471, -0.00000446, -0.00000115, -0.00000608, -0.00000439, 0.00001044, 0.00001129, -0.00000416, -0.00000486, 0.00000487, -0.00000115, -0.00001021, -0.00000408, 0.00000075, -0.00000453, -0.00000394, 0.00000580, 0.00001080, 0.00000216, -0.00001149, -0.00000642, 0.00001506, 0.00001178, -0.00001718, -0.00001715, 0.00001162, 0.00001133, -0.00001125, -0.00000635, 0.00001039, 0.00000307, -0.00000900, -0.00000419, 0.00000503, 0.00000634, -0.00000596, -0.00001992, -0.00000306, 0.00002533, 0.00000597, -0.00003066, -0.00000858, 0.00003081, 0.00001126, -0.00001939, -0.00000235, 0.00001274, -0.00000508, -0.00000823, 0.00000855, 0.00000397, -0.00001233, -0.00000721, 0.00000932, 0.00000945, -0.00001117, -0.00002649, -0.00001028, 0.00001203, 0.00000200, -0.00000653, 0.00002372, 0.00003867, 0.00000573, -0.00000432, 0.00003297, 0.00004442, 0.00001747, 0.00000758, 0.00000414, -0.00002471, -0.00003858, -0.00001267, 0.00000560, -0.00000367, -0.00000744, -0.00000005, -0.00000114, -0.00000607, -0.00000440, -0.00000225, -0.00000272, -0.00000182, -0.00000012, 0.00000048, -0.00000071, -0.00000420, -0.00000550, },
		{ -0.00000516, -0.00000522, 0.00000162, 0.00000883, 0.00000857, 0.00000109, -0.00000496, -0.00000379, -0.00000169, -0.00000470, -0.00000615, -0.00000092, 0.00000358, 0.00000135, -0.00000358, -0.00000494, -0.00000040, 0.00000563, 0.00000688, 0.00000535, 0.00000434, -0.00000036, -0.00000689, -0.00000681, -0.00000340, -0.00000035, 0.00000913, 0.00001510, 0.00000007, -0.00001621, -0.00000733, 0.00000631, 0.00000177, -0.00000262, 0.00000333, 0.00000206, -0.00000586, -0.00000573, -0.00000040, 0.00000465, 0.00000804, 0.00000227, -0.00000627, -0.00000326, -0.00000081, -0.00000787, -0.00000240, 0.00001356, 0.00000826, -0.00000748, -0.00000199, 0.00000628, -0.00000519, -0.00001113, 0.00000026, 0.00000208, -0.00000800, -0.00000435, 0.00001119, 0.00001247, -0.00000528, -0.00001462, 0.00000343, 0.00001959, 0.00000049, -0.00002188, -0.00000599, 0.00001432, 0.00000072, -0.00001236, 0.00000111, 0.00000899, -0.00000129, -0.00000699, -0.00000366, 0.00000327, 0.00000749, -0.00000758, -0.00002231, 0.00000260, 0.00002773, -0.00000331, -0.00003251, 0.00000213, 0.00003208, 0.00000300, -0.00001849, 0.00000223, 0.00000959, -0.00000674, -0.00000499, 0.00000752, 0.00000132, -0.00001081, -0.00000600, 0.00000933, 0.00000927, -0.00001554, -0.00002958, -0.00000569, 0.00001310, -0.00000296, -0.00000094, 0.00003459, 0.00003444, -0.00000118, 0.00000565, 0.00004312, 0.00004022, 0.00001390, 0.00000706, -0.00000629, -0.00003402, -0.00003183, -0.00000436, 0.00000251, -0.00000693, -0.00000478, 0.00000059, -0.00000322, -0.00000624, -0.00000407, -0.00000257, -0.00000189, -0.00000071, 0.00000004, 0.00000042, -0.00000125, },
		{ 0.00000035, -0.00000211, -0.00000592, -0.00000455, 0.00000299, 0.00000956, 0.00000783, -0.00000060, -0.00000534, -0.00000290, -0.00000234, -0.00000602, -0.00000498, 0.00000104, 0.00000303, -0.00000027, -0.00000375, -0.00000388, 0.00000066, 0.00000598, 0.00000707, 0.00000589, 0.00000338, -0.00000305, -0.00000773, -0.00000579, -0.00000303, 0.00000231, 0.00001357, 0.00001233, -0.00000791, -0.00001604, -0.00000122, 0.00000597, -0.00000105, -0.00000019, 0.00000393, -0.00000207, -0.00000606, -0.00000242, -0.00000002, 0.00000454, 0.00000883, 0.00000040, -0.00000747, -0.00000131, -0.00000143, -0.00000961, 0.00000098, 0.00001571, 0.00000370, -0.00000928, 0.00000204, 0.00000573, -0.00000989, -0.00000956, 0.00000537, 0.00000069, -0.00001236, -0.00000146, 0.00001761, 0.00000962, -0.00001395, -0.00001178, 0.00001443, 0.00001690, -0.00001155, -0.00001915, 0.00000464, 0.00001015, -0.00000811, -0.00000826, 0.00000605, 0.00000512, -0.00000280, -0.00000520, -0.00000537, 0.00000361, 0.00000939, -0.00001185, -0.00002328, 0.00001040, 0.00002672, -0.00001355, -0.00002995, 0.00001257, 0.00002958, -0.00000394, -0.00001567, 0.00000479, 0.00000642, -0.00000674, -0.00000317, 0.00000599, 0.00000059, -0.00001005, -0.00000588, 0.00001097, 0.00000824, -0.00002205, -0.00003028, 0.00000007, 0.00001086, -0.00000654, 0.00000910, 0.00004245, 0.00002655, -0.00000340, 0.00001843, 0.00004899, 0.00003501, 0.00001172, 0.00000238, -0.00001863, -0.00003710, -0.00002218, -0.00000053, -0.00000203, -0.00000739, -0.00000218, -0.00000036, -0.00000473, -0.00000588, -0.00000413, -0.00000251, -0.00000077, -0.00000003, 0.00000013, },
		{ 0.00000033, 0.00000031, 0.00000006, -0.00000316, -0.00000630, -0.00000372, 0.00000424, 0.00001027, 0.00000679, -0.00000239, -0.00000523, -0.00000216, -0.00000355, -0.00000672, -0.00000303, 0.00000227, 0.00000171, -0.00000141, -0.00000328, -0.00000286, 0.00000136, 0.00000621, 0.00000757, 0.00000622, 0.00000148, -0.00000556, -0.00000755, -0.00000494, -0.00000241, 0.00000644, 0.00001655, 0.00000621, -0.00001445, -0.00001215, 0.00000361, 0.00000329, -0.00000199, 0.00000292, 0.00000193, -0.00000576, -0.00000368, 0.00000007, -0.00000109, 0.00000541, 0.00000956, -0.00000242, -0.00000769, 0.00000097, -0.00000332, -0.00001059, 0.00000553, 0.00001607, -0.00000175, -0.00000886, 0.00000624, 0.00000277, -0.00001396, -0.00000512, 0.00000960, -0.00000388, -0.00001559, 0.00000519, 0.00002237, 0.00000192, -0.00002036, -0.00000302, 0.00002237, 0.00000763, -0.00001985, -0.00001045, 0.00001069, 0.00000138, -0.00001217, -0.00000160, 0.00000711, 0.00000127, -0.00000189, -0.00000539, -0.00000759, 0.00000672, 0.00000987, -0.00001814, -0.00002110, 0.00001858, 0.00002157, -0.00002247, -0.00002335, 0.00002079, 0.00002442, -0.00000861, -0.00001230, 0.00000560, 0.00000420, -0.00000621, -0.00000284, 0.00000529, 0.00000085, -0.00001085, -0.00000533, 0.00001391, 0.00000480, -0.00002918, -0.00002771, 0.00000478, 0.00000578, -0.00000630, 0.00002180, 0.00004545, 0.00001768, -0.00000010, 0.00003130, 0.00005097, 0.00003025, 0.00000864, -0.00000681, -0.00002903, -0.00003375, -0.00001324, -0.00000105, -0.00000565, -0.00000573, -0.00000085, -0.00000205, -0.00000532, -0.00000556, -0.00000437, -0.00000193, 0.00000030, },
		{ -0.00000094, 0.00000106, 0.00000061, 0.00000049, -0.00000062, -0.00000414, -0.00000632, -0.00000294, 0.00000554, 0.00001096, 0.00000525, -0.00000408, -0.00000458, -0.00000187, -0.00000510, -0.00000646, -0.00000077, 0.00000245, 0.00000007, -0.00000178, -0.00000241, -0.00000210, 0.00000168, 0.00000654, 0.00000843, 0.00000595, -0.00000126, -0.00000733, -0.00000659, -0.00000450, -0.00000089, 0.00001155, 0.00001667, -0.00000226, -0.00001767, -0.00000577, 0.00000564, -0.00000033, -0.00000043, 0.00000504, -0.00000239, -0.00000735, 0.00000047, 0.00000056, -0.00000262, 0.00000764, 0.00000938, -0.00000584, -0.00000644, 0.00000285, -0.00000640, -0.00001006, 0.00001059, 0.00001410, -0.00000700, -0.00000602, 0.00000935, -0.00000240, -0.00001591, 0.00000168, 0.00001117, -0.00001090, -0.00001553, 0.00001451, 0.00002284, -0.00000897, -0.00002141, 0.00000924, 0.00002389, -0.00000506, -0.00002139, 0.00000056, 0.00001009, -0.00000820, -0.00001065, 0.00000443, 0.00000486, -0.00000061, -0.00000026, -0.00000807, -0.00000820, 0.00001172, 0.00000726, -0.00002456, -0.00001519, 0.00002489, 0.00001298, -0.00002809, -0.00001422, 0.00002566, 0.00001823, -0.00001105, -0.00000948, 0.00000556, 0.00000313, -0.00000625, -0.00000319, 0.00000607, 0.00000071, -0.00001291, -0.00000306, 0.00001677, -0.00000161, -0.00003489, -0.00002249, 0.00000650, -0.00000017, -0.00000093, 0.00003441, 0.00004340, 0.00001069, 0.00000787, 0.00004221, 0.00005039, 0.00002580, 0.00000234, -0.00001861, -0.00003424, -0.00002609, -0.00000772, -0.00000412, -0.00000696, -0.00000342, -0.00000109, -0.00000352, -0.00000524, -0.00000561, -0.00000442, },
		{ -0.00000594, -0.00000400, 0.00000010, 0.00000153, 0.00000100, 0.00000045, -0.00000161, -0.00000477, -0.00000619, -0.00000228, 0.00000711, 0.00001143, 0.00000310, -0.00000535, -0.00000354, -0.00000227, -0.00000654, -0.00000514, 0.00000119, 0.00000157, -0.00000136, -0.00000134, -0.00000144, -0.00000178, 0.00000173, 0.00000724, 0.00000947, 0.00000468, -0.00000443, -0.00000795, -0.00000535, -0.00000439, 0.00000213, 0.00001655, 0.00001299, -0.00001117, -0.00001651, 0.00000104, 0.00000434, -0.00000301, 0.00000315, 0.00000464, -0.00000770, -0.00000577, 0.00000477, -0.00000139, -0.00000332, 0.00001090, 0.00000755, -0.00000906, -0.00000364, 0.00000348, -0.00001012, -0.00000750, 0.00001512, 0.00000976, -0.00001081, -0.00000122, 0.00001013, -0.00000878, -0.00001451, 0.00000937, 0.00000875, -0.00001846, -0.00001073, 0.00002402, 0.00001749, -0.00001982, -0.00001564, 0.00002100, 0.00001790, -0.00001669, -0.00001576, 0.00000945, 0.00000339, -0.00001484, -0.00000495, 0.00000747, 0.00000125, 0.00000009, 0.00000007, -0.00001224, -0.00000573, 0.00001658, 0.00000118, -0.00002889, -0.00000647, 0.00002752, 0.00000276, -0.00002947, -0.00000459, 0.00002722, 0.00001244, -0.00001204, -0.00000763, 0.00000566, 0.00000260, -0.00000737, -0.00000298, 0.00000798, -0.00000100, -0.00001501, 0.00000128, 0.00001775, -0.00001020, -0.00003754, -0.00001646, 0.00000450, -0.00000433, 0.00000919, 0.00004426, 0.00003773, 0.00000756, 0.00001861, 0.00005043, 0.00004849, 0.00002008, -0.00000809, -0.00002951, -0.00003302, -0.00001756, -0.00000649, -0.00000733, -0.00000600, -0.00000187, -0.00000232, -0.00000419, -0.00000504, },
		{ -0.00000413, -0.00000517, -0.00000620, -0.00000314, 0.00000088, 0.00000195, 0.00000149, -0.00000008, -0.00000264, -0.00000498, -0.00000622, -0.00000153, 0.00000909, 0.00001129, 0.00000046, -0.00000587, -0.00000245, -0.00000340, -0.00000735, -0.00000299, 0.00000230, -0.00000008, -0.00000212, -0.00000026, -0.00000070, -0.00000197, 0.00000174, 0.00000853, 0.00001031, 0.00000217, -0.00000740, -0.00000727, -0.00000437, -0.00000411, 0.00000682, 0.00001986, 0.00000558, -0.00001817, -0.00001117, 0.00000602, 0.00000043, -0.00000321, 0.00000726, 0.00000100, -0.00001200, -0.00000102, 0.00000738, -0.00000520, -0.00000203, 0.00001421, 0.00000381, -0.00001104, 0.00000022, 0.00000221, -0.00001351, -0.00000288, 0.00001798, 0.00000366, -0.00001212, 0.00000441, 0.00000786, -0.00001473, -0.00000933, 0.00001579, 0.00000209, -0.00002396, -0.00000121, 0.00003056, 0.00000674, -0.00002688, -0.00000404, 0.00002807, 0.00000603, -0.00002312, -0.00000527, 0.00001276, -0.00000657, -0.00001626, 0.00000211, 0.00000682, -0.00000145, 0.00000245, -0.00000223, -0.00001590, -0.00000008, 0.00001907, -0.00000727, -0.00002945, 0.00000310, 0.00002584, -0.00000681, -0.00002694, 0.00000378, 0.00002638, 0.00000775, -0.00001258, -0.00000637, 0.00000643, 0.00000164, -0.00000924, -0.00000128, 0.00000986, -0.00000454, -0.00001564, 0.00000687, 0.00001557, -0.00001917, -0.00003670, -0.00001192, -0.00000033, -0.00000438, 0.00002219, 0.00004973, 0.00003079, 0.00000891, 0.00003023, 0.00005626, 0.00004535, 0.00001102, -0.00002139, -0.00003594, -0.00002658, -0.00001138, -0.00000847, -0.00000883, -0.00000392, -0.00000174, -0.00000358, },
		{ -0.00000274, -0.00000415, -0.00000387, -0.00000569, -0.00000601, -0.00000216, 0.00000133, 0.00000262, 0.00000181, -0.00000115, -0.00000331, -0.00000497, -0.00000660, -0.00000033, 0.00001134, 0.00001017, -0.00000227, -0.00000551, -0.00000177, -0.00000502, -0.00000712, -0.00000054, 0.00000222, -0.00000195, -0.00000194, 0.00000117, -0.00000044, -0.00000268, 0.00000203, 0.00001047, 0.00001039, -0.00000150, -0.00000943, -0.00000559, -0.00000404, -0.00000294, 0.00001271, 0.00001993, -0.00000439, -0.00002117, -0.00000316, 0.00000747, -0.00000431, -0.00000030, 0.00000988, -0.00000541, -0.00001335, 0.00000573, 0.00000689, -0.00000943, 0.00000171, 0.00001627, -0.00000140, -0.00001091, 0.00000417, -0.00000113, -0.00001547, 0.00000321, 0.00001822, -0.00000301, -0.00001043, 0.00000930, 0.00000263, -0.00001837, -0.00000104, 0.00001878, -0.00000761, -0.00002494, 0.00001125, 0.00003140, -0.00000698, -0.00002728, 0.00001021, 0.00002770, -0.00000791, -0.00002214, 0.00000601, 0.00000927, -0.00001593, -0.00001236, 0.00000757, 0.00000352, -0.00000174, 0.00000460, -0.00000707, -0.00001702, 0.00000739, 0.00001773, -0.00001596, -0.00002587, 0.00001124, 0.00002055, -0.00001393, -0.00002184, 0.00000997, 0.00002438, 0.00000406, -0.00001332, -0.00000496, 0.00000766, -0.00000049, -0.00001083, 0.00000207, 0.00001037, -0.00000917, -0.00001381, 0.00001204, 0.00001010, -0.00002649, -0.00003338, -0.00001058, -0.00000585, 0.00000074, 0.00003550, 0.00005074, 0.00002486, 0.00001411, 0.00004153, 0.00006026, 0.00003970, -0.00000252, -0.00003438, -0.00003580, -0.00001802, -0.00000928, -0.00001142, -0.00000808, -0.00000221, },
		{ -0.00000587, -0.00000181, -0.00000394, -0.00000390, -0.00000398, -0.00000623, -0.00000527, -0.00000150, 0.00000173, 0.00000363, 0.00000155, -0.00000249, -0.00000338, -0.00000517, -0.00000722, 0.00000171, 0.00001339, 0.00000792, -0.00000454, -0.00000445, -0.00000186, -0.00000660, -0.00000573, 0.00000152, 0.00000100, -0.00000344, -0.00000084, 0.00000256, -0.00000085, -0.00000374, 0.00000302, 0.00001287, 0.00000913, -0.00000587, -0.00000993, -0.00000349, -0.00000436, -0.00000017, 0.00001861, 0.00001574, -0.00001475, -0.00001905, 0.00000509, 0.00000488, -0.00000770, 0.00000509, 0.00000923, -0.00001292, -0.00001051, 0.00001239, 0.00000288, -0.00001228, 0.00000742, 0.00001592, -0.00000699, -0.00000834, 0.00000697, -0.00000607, -0.00001505, 0.00000962, 0.00001547, -0.00000874, -0.00000602, 0.00001191, -0.00000450, -0.00001819, 0.00000854, 0.00001683, -0.00001790, -0.00002001, 0.00002354, 0.00002535, -0.00001987, -0.00002021, 0.00002272, 0.00001966, -0.00001932, -0.00001427, 0.00001379, 0.00000033, -0.00002132, -0.00000507, 0.00000958, -0.00000033, 0.00000046, 0.00000467, -0.00001307, -0.00001444, 0.00001443, 0.00001248, -0.00002262, -0.00001917, 0.00001625, 0.00001337, -0.00001779, -0.00001584, 0.00001401, 0.00002220, 0.00000076, -0.00001426, -0.00000273, 0.00000856, -0.00000386, -0.00001101, 0.00000632, 0.00000865, -0.00001343, -0.00000955, 0.00001508, 0.00000242, -0.00003082, -0.00002956, -0.00001280, -0.00000946, 0.00001053, 0.00004682, 0.00004846, 0.00002142, 0.00002208, 0.00005221, 0.00006224, 0.00002954, -0.00001968, -0.00004313, -0.00002946, -0.00001091, -0.00001091, -0.00001317, },
		{ -0.00001431, -0.00001261, -0.00000371, -0.00000272, -0.00000444, -0.00000335, -0.00000466, -0.00000627, -0.00000431, -0.00000134, 0.00000250, 0.00000473, 0.00000046, -0.00000360, -0.00000291, -0.00000599, -0.00000761, 0.00000470, 0.00001460, 0.00000477, -0.00000585, -0.00000316, -0.00000284, -0.00000752, -0.00000346, 0.00000259, -0.00000096, -0.00000407, 0.00000094, 0.00000360, -0.00000203, -0.00000486, 0.00000506, 0.00001521, 0.00000611, -0.00001012, -0.00000863, -0.00000171, -0.00000487, 0.00000451, 0.00002284, 0.00000734, -0.00002288, -0.00001210, 0.00001097, -0.00000085, -0.00000791, 0.00001115, 0.00000447, -0.00001914, -0.00000356, 0.00001666, -0.00000386, -0.00001219, 0.00001379, 0.00001261, -0.00001152, -0.00000371, 0.00000754, -0.00001153, -0.00001189, 0.00001492, 0.00001014, -0.00001215, -0.00000003, 0.00001121, -0.00001167, -0.00001362, 0.00001690, 0.00000980, -0.00002574, -0.00000954, 0.00003211, 0.00001337, -0.00002801, -0.00000737, 0.00002946, 0.00000636, -0.00002447, -0.00000250, 0.00001521, -0.00001078, -0.00002107, 0.00000266, 0.00000796, -0.00000263, 0.00000389, 0.00000172, -0.00001809, -0.00000846, 0.00001889, 0.00000462, -0.00002575, -0.00001127, 0.00001762, 0.00000623, -0.00001869, -0.00001028, 0.00001662, 0.00002015, -0.00000279, -0.00001484, 0.00000040, 0.00000820, -0.00000777, -0.00000909, 0.00001013, 0.00000477, -0.00001590, -0.00000397, 0.00001496, -0.00000560, -0.00003207, -0.00002738, -0.00001751, -0.00000915, 0.00002332, 0.00005486, 0.00004457, 0.00002092, 0.00003198, 0.00006228, 0.00006071, 0.00001342, -0.00003738, -0.00004472, -0.00001964, -0.00000780, },
		{ -0.00001027, -0.00000916, -0.00001697, -0.00001016, -0.00000287, -0.00000403, -0.00000390, -0.00000325, -0.00000557, -0.00000554, -0.00000370, -0.00000149, 0.00000398, 0.00000539, -0.00000138, -0.00000395, -0.00000237, -0.00000759, -0.00000715, 0.00000834, 0.00001442, 0.00000131, -0.00000601, -0.00000226, -0.00000438, -0.00000731, -0.00000096, 0.00000236, -0.00000302, -0.00000365, 0.00000304, 0.00000399, -0.00000399, -0.00000558, 0.00000828, 0.00001672, 0.00000132, -0.00001328, -0.00000578, -0.00000085, -0.00000478, 0.00001073, 0.00002370, -0.00000404, -0.00002644, -0.00000204, 0.00001255, -0.00000763, -0.00000416, 0.00001551, -0.00000387, -0.00002171, 0.00000607, 0.00001679, -0.00001148, -0.00000845, 0.00001904, 0.00000671, -0.00001366, 0.00000187, 0.00000535, -0.00001610, -0.00000633, 0.00001784, 0.00000335, -0.00001240, 0.00000586, 0.00000710, -0.00001677, -0.00000536, 0.00002158, -0.00000084, -0.00002844, 0.00000424, 0.00003419, -0.00000154, -0.00002878, 0.00000754, 0.00002824, -0.00000794, -0.00002189, 0.00000883, 0.00000980, -0.00002020, -0.00001572, 0.00000811, 0.00000413, -0.00000227, 0.00000660, -0.00000382, -0.00002023, -0.00000064, 0.00001950, -0.00000371, -0.00002510, -0.00000420, 0.00001604, 0.00000060, -0.00001774, -0.00000571, 0.00001865, 0.00001792, -0.00000678, -0.00001430, 0.00000386, 0.00000608, -0.00001103, -0.00000523, 0.00001221, -0.00000027, -0.00001581, 0.00000127, 0.00001182, -0.00001216, -0.00003144, -0.00002819, -0.00002263, -0.00000410, 0.00003701, 0.00005953, 0.00004057, 0.00002325, 0.00004358, 0.00007115, 0.00005326, -0.00000824, -0.00005120, -0.00003857, },
		{ -0.00005711, -0.00002688, -0.00000482, -0.00001336, -0.00001717, -0.00000731, -0.00000367, -0.00000459, -0.00000279, -0.00000406, -0.00000600, -0.00000426, -0.00000392, -0.00000136, 0.00000610, 0.00000503, -0.00000345, -0.00000334, -0.00000242, -0.00000966, -0.00000531, 0.00001189, 0.00001267, -0.00000170, -0.00000529, -0.00000221, -0.00000585, -0.00000588, 0.00000100, 0.00000099, -0.00000451, -0.00000228, 0.00000507, 0.00000356, -0.00000661, -0.00000534, 0.00001253, 0.00001656, -0.00000477, -0.00001446, -0.00000206, -0.00000114, -0.00000327, 0.00001737, 0.00002004, -0.00001606, -0.00002417, 0.00000838, 0.00000920, -0.00001290, 0.00000284, 0.00001604, -0.00001388, -0.00001914, 0.00001589, 0.00001217, -0.00001768, -0.00000151, 0.00002151, -0.00000055, -0.00001266, 0.00000688, 0.00000062, -0.00001849, 0.00000054, 0.00001763, -0.00000329, -0.00000953, 0.00000991, 0.00000058, -0.00001809, 0.00000464, 0.00002096, -0.00001233, -0.00002463, 0.00001782, 0.00002883, -0.00001525, -0.00002187, 0.00001995, 0.00001952, -0.00001870, -0.00001286, 0.00001569, -0.00000046, -0.00002493, -0.00000756, 0.00000984, 0.00000022, 0.00000049, 0.00000691, -0.00001034, -0.00001863, 0.00000679, 0.00001639, -0.00001047, -0.00002168, 0.00000066, 0.00001293, -0.00000304, -0.00001614, -0.00000192, 0.00002054, 0.00001492, -0.00001080, -0.00001222, 0.00000667, 0.00000240, -0.00001255, -0.00000043, 0.00001192, -0.00000500, -0.00001341, 0.00000465, 0.00000683, -0.00001631, -0.00003078, -0.00003206, -0.00002593, 0.00000524, 0.00004991, 0.00006149, 0.00003730, 0.00002834, 0.00005693, 0.00007688, 0.00003768, -0.00003247, },
		{ 0.00001352, -0.00005425, -0.00005319, -0.00001392, -0.00000485, -0.00001763, -0.00001477, -0.00000577, -0.00000531, -0.00000372, -0.00000209, -0.00000558, -0.00000535, -0.00000314, -0.00000500, -0.00000029, 0.00000834, 0.00000339, -0.00000499, -0.00000202, -0.00000360, -0.00001143, -0.00000199, 0.00001446, 0.00000971, -0.00000363, -0.00000430, -0.00000308, -0.00000652, -0.00000366, 0.00000178, -0.00000095, -0.00000501, -0.00000033, 0.00000676, 0.00000217, -0.00000956, -0.00000359, 0.00001722, 0.00001402, -0.00001121, -0.00001314, 0.00000159, -0.00000231, 0.00000019, 0.00002278, 0.00001174, -0.00002592, -0.00001629, 0.00001621, 0.00000188, -0.00001449, 0.00001114, 0.00001163, -0.00002283, -0.00001136, 0.00002314, 0.00000365, -0.00002041, 0.00000714, 0.00002026, -0.00000740, -0.00000869, 0.00000984, -0.00000563, -0.00001791, 0.00000724, 0.00001439, -0.00000821, -0.00000452, 0.00001089, -0.00000651, -0.00001492, 0.00001372, 0.00001493, -0.00002139, -0.00001491, 0.00002747, 0.00001739, -0.00002386, -0.00000945, 0.00002614, 0.00000625, -0.00002268, -0.00000086, 0.00001579, -0.00001201, -0.00002388, 0.00000044, 0.00000808, -0.00000185, 0.00000420, 0.00000419, -0.00001578, -0.00001380, 0.00001188, 0.00001089, -0.00001443, -0.00001718, 0.00000304, 0.00000980, -0.00000508, -0.00001469, 0.00000169, 0.00002217, 0.00001077, -0.00001406, -0.00000878, 0.00000793, -0.00000192, -0.00001182, 0.00000399, 0.00000951, -0.00000818, -0.00000984, 0.00000546, 0.00000166, -0.00001825, -0.00003186, -0.00003787, -0.00002584, 0.00001765, 0.00006114, 0.00006153, 0.00003517, 0.00003664, 0.00007148, 0.00007628, },
		{ 0.00008514, 0.00006586, -0.00001682, -0.00006809, -0.00004070, -0.00000427, -0.00000939, -0.00001941, -0.00001120, -0.00000640, -0.00000633, -0.00000167, -0.00000269, -0.00000696, -0.00000353, -0.00000302, -0.00000636, 0.00000208, 0.00000985, 0.00000073, -0.00000537, -0.00000075, -0.00000599, -0.00001203, 0.00000230, 0.00001538, 0.00000635, -0.00000432, -0.00000381, -0.00000441, -0.00000591, -0.00000146, 0.00000119, -0.00000267, -0.00000449, 0.00000180, 0.00000791, -0.00000028, -0.00001235, 0.00000003, 0.00002139, 0.00000885, -0.00001672, -0.00000936, 0.00000422, -0.00000364, 0.00000549, 0.00002517, -0.00000006, -0.00003106, -0.00000459, 0.00001921, -0.00000717, -0.00001136, 0.00001806, 0.00000270, -0.00002797, 0.00000009, 0.00002562, -0.00000671, -0.00001859, 0.00001535, 0.00001541, -0.00001213, -0.00000277, 0.00000981, -0.00001192, -0.00001438, 0.00001225, 0.00000902, -0.00001035, 0.00000106, 0.00000852, -0.00001201, -0.00000792, 0.00001931, 0.00000505, -0.00002527, -0.00000174, 0.00003050, 0.00000315, -0.00002504, 0.00000449, 0.00002444, -0.00000713, -0.00001909, 0.00000976, 0.00000933, -0.00002112, -0.00001805, 0.00000581, 0.00000443, -0.00000129, 0.00000701, -0.00000092, -0.00001849, -0.00000734, 0.00001370, 0.00000491, -0.00001555, -0.00001318, 0.00000362, 0.00000755, -0.00000642, -0.00001351, 0.00000566, 0.00002296, 0.00000560, -0.00001577, -0.00000476, 0.00000724, -0.00000570, -0.00000918, 0.00000690, 0.00000601, -0.00000929, -0.00000657, 0.00000405, -0.00000233, -0.00001918, -0.00003556, -0.00004389, -0.00002179, 0.00003179, 0.00007048, 0.00006015, 0.00003473, 0.00004889, },
		{ 0.00003715, 0.00006535, 0.00009402, 0.00004353, -0.00004802, -0.00007016, -0.00002388, -0.00000101, -0.00001550, -0.00001769, -0.00000860, -0.00000870, -0.00000547, 0.00000042, -0.00000478, -0.00000715, -0.00000119, -0.00000440, -0.00000706, 0.00000548, 0.00000987, -0.00000214, -0.00000443, -0.00000041, -0.00000904, -0.00001082, 0.00000655, 0.00001452, 0.00000353, -0.00000420, -0.00000431, -0.00000538, -0.00000408, -0.00000021, -0.00000037, -0.00000350, -0.00000329, 0.00000382, 0.00000844, -0.00000381, -0.00001430, 0.00000554, 0.00002391, 0.00000136, -0.00002002, -0.00000377, 0.00000521, -0.00000424, 0.00001182, 0.00002324, -0.00001311, -0.00002998, 0.00000804, 0.00001649, -0.00001509, -0.00000402, 0.00002111, -0.00000883, -0.00002745, 0.00001255, 0.00002244, -0.00001623, -0.00001242, 0.00002106, 0.00000811, -0.00001357, 0.00000348, 0.00000667, -0.00001669, -0.00000870, 0.00001456, 0.00000300, -0.00000947, 0.00000544, 0.00000360, -0.00001418, 0.00000101, 0.00001982, -0.00000582, -0.00002272, 0.00001122, 0.00002620, -0.00000971, -0.00001890, 0.00001552, 0.00001586, -0.00001635, -0.00000979, 0.00001534, -0.00000128, -0.00002523, -0.00000992, 0.00000748, 0.00000097, 0.00000144, 0.00000747, -0.00000683, -0.00001792, -0.00000124, 0.00001253, 0.00000010, -0.00001479, -0.00001060, 0.00000358, 0.00000627, -0.00000788, -0.00001211, 0.00001004, 0.00002226, 0.00000008, -0.00001558, -0.00000121, 0.00000494, -0.00000794, -0.00000558, 0.00000784, 0.00000268, -0.00000872, -0.00000471, 0.00000153, -0.00000461, -0.00002064, -0.00004170, -0.00004847, -0.00001404, 0.00004676, 0.00007793, 0.00005754, },
		{ 0.00008323, 0.00005416, 0.00004423, 0.00008457, 0.00009317, 0.00001025, -0.00007313, -0.00005995, -0.00000844, -0.00000440, -0.00001972, -0.00001347, -0.00000864, -0.00001101, -0.00000248, 0.00000112, -0.00000762, -0.00000549, 0.00000052, -0.00000705, -0.00000620, 0.00000906, 0.00000819, -0.00000423, -0.00000267, -0.00000163, -0.00001169, -0.00000782, 0.00000967, 0.00001246, 0.00000190, -0.00000416, -0.00000569, -0.00000510, -0.00000171, -0.00000051, -0.00000202, -0.00000316, -0.00000196, 0.00000563, 0.00000819, -0.00000830, -0.00001460, 0.00001246, 0.00002365, -0.00000747, -0.00002014, 0.00000252, 0.00000444, -0.00000334, 0.00001774, 0.00001662, -0.00002457, -0.00002267, 0.00001843, 0.00000883, -0.00001936, 0.00000560, 0.00001886, -0.00002010, -0.00002094, 0.00002293, 0.00001424, -0.00002244, -0.00000337, 0.00002288, 0.00000023, -0.00001151, 0.00000833, 0.00000116, -0.00001884, -0.00000227, 0.00001394, -0.00000212, -0.00000630, 0.00000736, -0.00000218, -0.00001229, 0.00000929, 0.00001517, -0.00001441, -0.00001448, 0.00002034, 0.00001618, -0.00001751, -0.00000790, 0.00002033, 0.00000368, -0.00001876, 0.00000151, 0.00001421, -0.00001240, -0.00002371, -0.00000235, 0.00000597, -0.00000064, 0.00000487, 0.00000522, -0.00001169, -0.00001470, 0.00000295, 0.00000965, -0.00000274, -0.00001353, -0.00000937, 0.00000392, 0.00000539, -0.00000975, -0.00000988, 0.00001433, 0.00001979, -0.00000485, -0.00001380, 0.00000104, 0.00000192, -0.00000833, -0.00000221, 0.00000718, 0.00000042, -0.00000749, -0.00000456, -0.00000083, -0.00000567, -0.00002382, -0.00004930, -0.00005064, -0.00000313, 0.00006214, },
		{ 0.00001059, 0.00007756, 0.00008575, 0.00005137, 0.00005762, 0.00010271, 0.00007825, -0.00002924, -0.00008585, -0.00004085, 0.00000062, -0.00001189, -0.00001973, -0.00000918, -0.00001157, -0.00001137, 0.00000167, -0.00000049, -0.00000981, -0.00000221, 0.00000048, -0.00001002, -0.00000345, 0.00001166, 0.00000530, -0.00000481, -0.00000115, -0.00000432, -0.00001284, -0.00000382, 0.00001095, 0.00001028, 0.00000143, -0.00000510, -0.00000720, -0.00000312, 0.00000004, -0.00000232, -0.00000276, -0.00000196, -0.00000096, 0.00000738, 0.00000692, -0.00001339, -0.00001255, 0.00001980, 0.00001991, -0.00001614, -0.00001670, 0.00000820, 0.00000230, -0.00000063, 0.00002160, 0.00000614, -0.00003180, -0.00001074, 0.00002400, -0.00000158, -0.00001856, 0.00001474, 0.00001138, -0.00002812, -0.00000975, 0.00002866, 0.00000306, -0.00002385, 0.00000632, 0.00002056, -0.00000628, -0.00000673, 0.00001051, -0.00000532, -0.00001801, 0.00000343, 0.00001096, -0.00000518, -0.00000219, 0.00000642, -0.00000683, -0.00000691, 0.00001447, 0.00000696, -0.00001816, -0.00000310, 0.00002315, 0.00000380, -0.00001834, 0.00000395, 0.00001784, -0.00000776, -0.00001421, 0.00001063, 0.00000708, -0.00002058, -0.00001791, 0.00000250, 0.00000290, 0.00000013, 0.00000735, 0.00000101, -0.00001423, -0.00001033, 0.00000467, 0.00000661, -0.00000384, -0.00001287, -0.00000873, 0.00000503, 0.00000409, -0.00001174, -0.00000645, 0.00001777, 0.00001582, -0.00000841, -0.00001120, 0.00000170, -0.00000077, -0.00000722, 0.00000012, 0.00000578, -0.00000056, -0.00000666, -0.00000564, -0.00000225, -0.00000654, -0.00002909, -0.00005719, -0.00005007, },
		{ -0.00006457, -0.00004669, 0.00002718, 0.00009223, 0.00008510, 0.00005169, 0.00007746, 0.00011377, 0.00004769, -0.00006713, -0.00008291, -0.00001920, 0.00000098, -0.00001926, -0.00001548, -0.00000750, -0.00001589, -0.00000860, 0.00000510, -0.00000420, -0.00000996, 0.00000154, -0.00000176, -0.00001195, 0.00000067, 0.00001242, 0.00000232, -0.00000390, -0.00000096, -0.00000761, -0.00001186, -0.00000014, 0.00001050, 0.00000910, 0.00000131, -0.00000744, -0.00000760, 0.00000021, 0.00000000, -0.00000483, -0.00000186, -0.00000067, -0.00000045, 0.00000921, 0.00000428, -0.00001840, -0.00000770, 0.00002617, 0.00001261, -0.00002294, -0.00001008, 0.00001211, -0.00000039, 0.00000358, 0.00002204, -0.00000631, -0.00003305, 0.00000305, 0.00002353, -0.00001181, -0.00001284, 0.00002071, 0.00000034, -0.00003071, 0.00000350, 0.00002847, -0.00000832, -0.00002028, 0.00001434, 0.00001497, -0.00000993, -0.00000073, 0.00000954, -0.00001119, -0.00001469, 0.00000724, 0.00000684, -0.00000580, 0.00000130, 0.00000326, -0.00000874, 0.00000016, 0.00001513, -0.00000205, -0.00001604, 0.00000786, 0.00001929, -0.00000696, -0.00001267, 0.00001259, 0.00000950, -0.00001466, -0.00000498, 0.00001443, -0.00000334, -0.00002371, -0.00001035, 0.00000388, 0.00000019, 0.00000270, 0.00000776, -0.00000370, -0.00001420, -0.00000643, 0.00000439, 0.00000460, -0.00000413, -0.00001320, -0.00000771, 0.00000660, 0.00000184, -0.00001313, -0.00000201, 0.00001965, 0.00001109, -0.00001031, -0.00000872, 0.00000111, -0.00000243, -0.00000544, 0.00000122, 0.00000453, -0.00000075, -0.00000681, -0.00000703, -0.00000270, -0.00000826, -0.00003608, },
		{ -0.00001129, -0.00004415, -0.00007113, -0.00004020, 0.00004670, 0.00010472, 0.00008190, 0.00005824, 0.00010112, 0.00011112, 0.00000431, -0.00009464, -0.00006562, -0.00000200, -0.00000610, -0.00002261, -0.00000920, -0.00000986, -0.00001909, -0.00000303, 0.00000599, -0.00000862, -0.00000747, 0.00000414, -0.00000567, -0.00001179, 0.00000493, 0.00001126, 0.00000046, -0.00000235, -0.00000268, -0.00001010, -0.00000904, 0.00000196, 0.00000930, 0.00000948, 0.00000028, -0.00001076, -0.00000579, 0.00000372, -0.00000242, -0.00000665, 0.00000068, -0.00000022, -0.00000008, 0.00001113, -0.00000005, -0.00002234, -0.00000014, 0.00003000, 0.00000249, -0.00002628, -0.00000138, 0.00001349, -0.00000270, 0.00000838, 0.00001840, -0.00001816, -0.00002797, 0.00001551, 0.00001740, -0.00001912, -0.00000382, 0.00002173, -0.00001154, -0.00002718, 0.00001569, 0.00002266, -0.00001722, -0.00001294, 0.00001897, 0.00000785, -0.00001015, 0.00000473, 0.00000583, -0.00001520, -0.00000995, 0.00000871, 0.00000292, -0.00000444, 0.00000303, -0.00000067, -0.00000731, 0.00000659, 0.00001134, -0.00000884, -0.00000895, 0.00001504, 0.00001051, -0.00001282, -0.00000315, 0.00001526, -0.00000129, -0.00001503, 0.00000518, 0.00001184, -0.00001351, -0.00002162, -0.00000368, 0.00000250, -0.00000080, 0.00000572, 0.00000597, -0.00000741, -0.00001235, -0.00000409, 0.00000331, 0.00000393, -0.00000471, -0.00001412, -0.00000570, 0.00000785, -0.00000132, -0.00001327, 0.00000279, 0.00001972, 0.00000650, -0.00001082, -0.00000699, 0.00000001, -0.00000290, -0.00000380, 0.00000147, 0.00000387, -0.00000090, -0.00000784, -0.00000801, -0.00000280, },
		{ -0.00000844, -0.00000330, -0.00001546, -0.00005293, -0.00007682, -0.00002976, 0.00006862, 0.00011340, 0.00007846, 0.00007327, 0.00012269, 0.00008996, -0.00004450, -0.00010502, -0.00003978, 0.00000566, -0.00001636, -0.00002011, -0.00000435, -0.00001561, -0.00001883, 0.00000348, 0.00000353, -0.00001176, -0.00000298, 0.00000427, -0.00000980, -0.00000934, 0.00000792, 0.00000903, 0.00000038, -0.00000156, -0.00000593, -0.00001048, -0.00000561, 0.00000186, 0.00000886, 0.00001099, -0.00000284, -0.00001373, -0.00000133, 0.00000568, -0.00000688, -0.00000634, 0.00000406, -0.00000127, 0.00000082, 0.00001273, -0.00000614, -0.00002407, 0.00000935, 0.00002996, -0.00000891, -0.00002516, 0.00000782, 0.00001217, -0.00000388, 0.00001253, 0.00001103, -0.00002688, -0.00001774, 0.00002390, 0.00000748, -0.00002174, 0.00000597, 0.00001745, -0.00002131, -0.00001844, 0.00002409, 0.00001298, -0.00002180, -0.00000390, 0.00001958, 0.00000117, -0.00000737, 0.00000822, 0.00000049, -0.00001672, -0.00000511, 0.00000811, 0.00000026, -0.00000217, 0.00000263, -0.00000367, -0.00000312, 0.00001026, 0.00000471, -0.00001117, 0.00000053, 0.00001643, 0.00000013, -0.00001239, 0.00000642, 0.00001148, -0.00001047, -0.00000931, 0.00001245, 0.00000406, -0.00002034, -0.00001587, 0.00000030, -0.00000007, 0.00000027, 0.00000782, 0.00000277, -0.00000929, -0.00000999, -0.00000346, 0.00000262, 0.00000411, -0.00000623, -0.00001476, -0.00000271, 0.00000801, -0.00000482, -0.00001185, 0.00000714, 0.00001824, 0.00000271, -0.00001057, -0.00000616, -0.00000089, -0.00000257, -0.00000273, 0.00000149, 0.00000367, -0.00000156, -0.00000916, },
		{ -0.00000271, -0.00001022, -0.00000880, -0.00000449, -0.00002031, -0.00006262, -0.00008118, -0.00001430, 0.00009134, 0.00011740, 0.00007865, 0.00009644, 0.00013403, 0.00004978, -0.00008835, -0.00009608, -0.00001378, 0.00000259, -0.00002445, -0.00001294, -0.00000390, -0.00002220, -0.00001420, 0.00000833, -0.00000148, -0.00001201, 0.00000177, 0.00000166, -0.00001248, -0.00000551, 0.00000880, 0.00000722, 0.00000173, -0.00000284, -0.00000930, -0.00000824, -0.00000335, 0.00000010, 0.00001055, 0.00001221, -0.00000841, -0.00001451, 0.00000504, 0.00000449, -0.00001192, -0.00000305, 0.00000680, -0.00000377, 0.00000297, 0.00001325, -0.00001360, -0.00002254, 0.00001934, 0.00002534, -0.00001955, -0.00001941, 0.00001576, 0.00000860, -0.00000356, 0.00001479, 0.00000125, -0.00003062, -0.00000471, 0.00002663, -0.00000354, -0.00001933, 0.00001386, 0.00000907, -0.00002665, -0.00000676, 0.00002712, 0.00000201, -0.00002152, 0.00000454, 0.00001670, -0.00000348, -0.00000282, 0.00000901, -0.00000506, -0.00001589, -0.00000126, 0.00000632, -0.00000074, -0.00000023, 0.00000064, -0.00000446, 0.00000220, 0.00001005, -0.00000218, -0.00000839, 0.00000911, 0.00001201, -0.00000821, -0.00000652, 0.00001244, 0.00000308, -0.00001485, -0.00000009, 0.00001428, -0.00000606, -0.00002224, -0.00000893, 0.00000113, -0.00000217, 0.00000273, 0.00000823, -0.00000059, -0.00000939, -0.00000834, -0.00000385, 0.00000300, 0.00000419, -0.00000865, -0.00001431, 0.00000065, 0.00000670, -0.00000787, -0.00000912, 0.00001038, 0.00001585, -0.00000009, -0.00001022, -0.00000592, -0.00000123, -0.00000200, -0.00000213, 0.00000168, 0.00000343, },
		{ 0.00000196, 0.00000274, -0.00000385, -0.00001095, -0.00000972, -0.00000601, -0.00002566, -0.00007375, -0.00008294, 0.00000675, 0.00011239, 0.00011773, 0.00008664, 0.00012342, 0.00012713, -0.00000425, -0.00011693, -0.00007138, 0.00000435, -0.00000813, -0.00002631, -0.00000468, -0.00000874, -0.00002628, -0.00000636, 0.00000947, -0.00000694, -0.00000899, 0.00000474, -0.00000268, -0.00001259, -0.00000197, 0.00000783, 0.00000723, 0.00000311, -0.00000660, -0.00001091, -0.00000416, -0.00000380, -0.00000167, 0.00001480, 0.00001118, -0.00001554, -0.00001150, 0.00001146, -0.00000059, -0.00001540, 0.00000302, 0.00000732, -0.00000690, 0.00000674, 0.00001170, -0.00002145, -0.00001715, 0.00002794, 0.00001632, -0.00002725, -0.00000983, 0.00002093, 0.00000368, -0.00000189, 0.00001434, -0.00000897, -0.00002869, 0.00000823, 0.00002361, -0.00001296, -0.00001296, 0.00001791, -0.00000113, -0.00002653, 0.00000500, 0.00002468, -0.00000761, -0.00001720, 0.00001054, 0.00001173, -0.00000527, 0.00000191, 0.00000720, -0.00000958, -0.00001347, 0.00000108, 0.00000437, -0.00000038, 0.00000048, -0.00000172, -0.00000265, 0.00000661, 0.00000627, -0.00000668, -0.00000162, 0.00001384, 0.00000372, -0.00001179, 0.00000207, 0.00001284, -0.00000656, -0.00001312, 0.00000898, 0.00001021, -0.00001521, -0.00001943, -0.00000316, -0.00000037, -0.00000271, 0.00000545, 0.00000700, -0.00000305, -0.00000851, -0.00000800, -0.00000421, 0.00000438, 0.00000333, -0.00001129, -0.00001240, 0.00000349, 0.00000415, -0.00000980, -0.00000569, 0.00001230, 0.00001320, -0.00000209, -0.00001008, -0.00000579, -0.00000111, -0.00000155, -0.00000163, },
		{ -0.00000119, -0.00000096, 0.00000187, 0.00000168, -0.00000444, -0.00001196, -0.00001147, -0.00000709, -0.00003215, -0.00008653, -0.00008010, 0.00003268, 0.00012933, 0.00011780, 0.00010485, 0.00014601, 0.00009724, -0.00006172, -0.00012352, -0.00003929, 0.00000992, -0.00002054, -0.00002102, 0.00000034, -0.00001714, -0.00002534, 0.00000188, 0.00000644, -0.00001029, -0.00000395, 0.00000455, -0.00000675, -0.00001032, -0.00000035, 0.00000643, 0.00000954, 0.00000257, -0.00001192, -0.00000917, -0.00000018, -0.00000739, -0.00000123, 0.00002065, 0.00000627, -0.00002211, -0.00000417, 0.00001535, -0.00000887, -0.00001519, 0.00001040, 0.00000456, -0.00000919, 0.00001178, 0.00000727, -0.00002818, -0.00000810, 0.00003316, 0.00000415, -0.00003021, 0.00000194, 0.00002238, -0.00000150, 0.00000056, 0.00001104, -0.00001751, -0.00002170, 0.00001841, 0.00001613, -0.00001878, -0.00000470, 0.00001738, -0.00001054, -0.00002144, 0.00001421, 0.00001810, -0.00001395, -0.00001054, 0.00001324, 0.00000642, -0.00000433, 0.00000546, 0.00000356, -0.00001237, -0.00001046, 0.00000198, 0.00000308, 0.00000052, -0.00000024, -0.00000310, 0.00000110, 0.00000839, 0.00000056, -0.00000702, 0.00000660, 0.00001324, -0.00000527, -0.00000974, 0.00000986, 0.00000769, -0.00001380, -0.00000627, 0.00001455, 0.00000181, -0.00002074, -0.00001359, -0.00000001, -0.00000277, -0.00000154, 0.00000737, 0.00000490, -0.00000415, -0.00000777, -0.00000874, -0.00000366, 0.00000604, 0.00000123, -0.00001327, -0.00000932, 0.00000511, 0.00000095, -0.00001037, -0.00000224, 0.00001309, 0.00001070, -0.00000370, -0.00001011, -0.00000546, -0.00000084, },
		{ -0.00000500, -0.00000068, -0.00000058, -0.00000029, 0.00000101, 0.00000090, -0.00000441, -0.00001417, -0.00001344, -0.00000720, -0.00004109, -0.00010013, -0.00007063, 0.00006119, 0.00014124, 0.00012272, 0.00013176, 0.00015393, 0.00004564, -0.00010972, -0.00010765, -0.00001019, 0.00000329, -0.00002837, -0.00001120, -0.00000085, -0.00002533, -0.00001906, 0.00000726, 0.00000081, -0.00000986, 0.00000068, 0.00000134, -0.00000856, -0.00000719, -0.00000135, 0.00000665, 0.00001322, -0.00000145, -0.00001662, -0.00000374, 0.00000119, -0.00001301, 0.00000326, 0.00002579, -0.00000298, -0.00002535, 0.00000641, 0.00001440, -0.00001822, -0.00001002, 0.00001673, -0.00000147, -0.00000902, 0.00001694, -0.00000032, -0.00003208, 0.00000347, 0.00003349, -0.00000904, -0.00002750, 0.00001371, 0.00001992, -0.00000589, 0.00000299, 0.00000548, -0.00002265, -0.00001136, 0.00002405, 0.00000637, -0.00002021, 0.00000314, 0.00001292, -0.00001696, -0.00001311, 0.00001921, 0.00000954, -0.00001623, -0.00000359, 0.00001283, 0.00000230, -0.00000156, 0.00000699, -0.00000076, -0.00001333, -0.00000772, 0.00000202, 0.00000275, 0.00000107, -0.00000187, -0.00000258, 0.00000527, 0.00000684, -0.00000469, -0.00000296, 0.00001326, 0.00000767, -0.00001171, -0.00000327, 0.00001382, -0.00000090, -0.00001609, 0.00000293, 0.00001471, -0.00000805, -0.00002153, -0.00000706, 0.00000031, -0.00000465, 0.00000071, 0.00000801, 0.00000289, -0.00000423, -0.00000795, -0.00000976, -0.00000192, 0.00000708, -0.00000179, -0.00001388, -0.00000579, 0.00000524, -0.00000215, -0.00000978, 0.00000078, 0.00001314, 0.00000841, -0.00000515, -0.00001005, },
		{ -0.00000642, -0.00000973, -0.00000469, -0.00000049, 0.00000057, -0.00000023, -0.00000056, 0.00000123, -0.00000450, -0.00001816, -0.00001433, -0.00000677, -0.00005388, -0.00011240, -0.00005356, 0.00008903, 0.00014991, 0.00013731, 0.00016076, 0.00013819, -0.00001943, -0.00013676, -0.00007564, 0.00000730, -0.00001036, -0.00002783, -0.00000189, -0.00000824, -0.00002926, -0.00000969, 0.00000769, -0.00000440, -0.00000581, 0.00000243, -0.00000313, -0.00000712, -0.00000549, -0.00000409, 0.00001016, 0.00001602, -0.00000917, -0.00001801, 0.00000396, -0.00000206, -0.00001807, 0.00001223, 0.00002728, -0.00001522, -0.00002288, 0.00001754, 0.00000748, -0.00002550, -0.00000020, 0.00001945, -0.00000943, -0.00000533, 0.00002049, -0.00001040, -0.00003168, 0.00001560, 0.00002832, -0.00002064, -0.00001933, 0.00002323, 0.00001413, -0.00000878, 0.00000462, -0.00000117, -0.00002348, 0.00000001, 0.00002458, -0.00000327, -0.00001772, 0.00000869, 0.00000617, -0.00001924, -0.00000389, 0.00001964, 0.00000131, -0.00001486, 0.00000196, 0.00001033, 0.00000023, 0.00000177, 0.00000639, -0.00000473, -0.00001291, -0.00000574, 0.00000192, 0.00000312, 0.00000068, -0.00000338, -0.00000004, 0.00000812, 0.00000254, -0.00000724, 0.00000413, 0.00001591, -0.00000076, -0.00001348, 0.00000488, 0.00001252, -0.00000968, -0.00001279, 0.00001108, 0.00000952, -0.00001632, -0.00001811, -0.00000190, -0.00000138, -0.00000519, 0.00000314, 0.00000760, 0.00000165, -0.00000407, -0.00000922, -0.00001014, 0.00000062, 0.00000688, -0.00000494, -0.00001298, -0.00000262, 0.00000412, -0.00000463, -0.00000844, 0.00000325, 0.00001277, 0.00000618, },
		{ 0.00001202, 0.00000392, -0.00000725, -0.00000930, -0.00000478, 0.00000023, 0.00000206, -0.00000148, -0.00000205, 0.00000309, -0.00000610, -0.00002356, -0.00001273, -0.00000742, -0.00007101, -0.00012036, -0.00002987, 0.00011351, 0.00016010, 0.00016337, 0.00018089, 0.00009492, -0.00008419, -0.00013689, -0.00003870, 0.00000943, -0.00002331, -0.00001934, 0.00000189, -0.00001866, -0.00002655, -0.00000116, 0.00000343, -0.00000613, -0.00000030, 0.00000006, -0.00000609, -0.00000326, -0.00000721, -0.00000625, 0.00001724, 0.00001515, -0.00001898, -0.00001402, 0.00001101, -0.00001035, -0.00001930, 0.00002419, 0.00002265, -0.00002745, -0.00001378, 0.00002565, -0.00000462, -0.00002760, 0.00001223, 0.00001679, -0.00001697, 0.00000188, 0.00002061, -0.00002138, -0.00002631, 0.00002586, 0.00001827, -0.00002813, -0.00000712, 0.00002869, 0.00000623, -0.00000986, 0.00000496, -0.00000748, -0.00002009, 0.00001010, 0.00002065, -0.00001084, -0.00001270, 0.00001102, -0.00000081, -0.00001749, 0.00000398, 0.00001636, -0.00000480, -0.00001112, 0.00000517, 0.00000713, 0.00000028, 0.00000447, 0.00000416, -0.00000766, -0.00001179, -0.00000449, 0.00000219, 0.00000357, -0.00000068, -0.00000373, 0.00000371, 0.00000836, -0.00000281, -0.00000582, 0.00001176, 0.00001356, -0.00000905, -0.00001021, 0.00001155, 0.00000653, -0.00001548, -0.00000532, 0.00001533, 0.00000083, -0.00002084, -0.00001221, 0.00000077, -0.00000379, -0.00000434, 0.00000503, 0.00000676, 0.00000124, -0.00000452, -0.00001110, -0.00000931, 0.00000318, 0.00000537, -0.00000740, -0.00001092, -0.00000040, 0.00000223, -0.00000633, -0.00000672, 0.00000527, },
		{ -0.00000477, 0.00000694, 0.00001077, 0.00000180, -0.00000748, -0.00000921, -0.00000511, 0.00000202, 0.00000313, -0.00000430, -0.00000216, 0.00000593, -0.00001055, -0.00002879, -0.00000811, -0.00001155, -0.00009135, -0.00012138, -0.00000245, 0.00013435, 0.00017808, 0.00019728, 0.00017971, 0.00002814, -0.00013315, -0.00011216, -0.00000904, -0.00000123, -0.00002868, -0.00000743, -0.00000243, -0.00002699, -0.00001790, 0.00000265, -0.00000278, -0.00000286, 0.00000332, -0.00000557, -0.00000512, 0.00000053, -0.00001279, -0.00000487, 0.00002618, 0.00000850, -0.00002764, -0.00000435, 0.00001381, -0.00002196, -0.00001406, 0.00003584, 0.00001108, -0.00003575, 0.00000072, 0.00002737, -0.00001912, -0.00002256, 0.00002394, 0.00000848, -0.00002142, 0.00001131, 0.00001611, -0.00003104, -0.00001643, 0.00003195, 0.00000520, -0.00002975, 0.00000681, 0.00002910, -0.00000224, -0.00000927, 0.00000389, -0.00001211, -0.00001347, 0.00001718, 0.00001378, -0.00001527, -0.00000689, 0.00001027, -0.00000623, -0.00001290, 0.00000899, 0.00001102, -0.00000795, -0.00000666, 0.00000603, 0.00000453, 0.00000184, 0.00000578, 0.00000114, -0.00000938, -0.00001059, -0.00000362, 0.00000297, 0.00000347, -0.00000243, -0.00000233, 0.00000726, 0.00000569, -0.00000705, -0.00000059, 0.00001721, 0.00000699, -0.00001441, -0.00000338, 0.00001428, -0.00000190, -0.00001640, 0.00000354, 0.00001436, -0.00000859, -0.00002090, -0.00000598, 0.00000088, -0.00000577, -0.00000259, 0.00000613, 0.00000614, 0.00000120, -0.00000599, -0.00001277, -0.00000731, 0.00000497, 0.00000302, -0.00000868, -0.00000835, 0.00000072, 0.00000008, -0.00000733, },
		{ -0.00000207, -0.00000778, -0.00000258, 0.00000813, 0.00000894, 0.00000023, -0.00000726, -0.00000993, -0.00000500, 0.00000496, 0.00000263, -0.00000802, 0.00000022, 0.00000813, -0.00001831, -0.00003162, -0.00000144, -0.00002127, -0.00011206, -0.00011451, 0.00002512, 0.00015490, 0.00020882, 0.00022913, 0.00014777, -0.00004992, -0.00015413, -0.00007238, 0.00000457, -0.00001697, -0.00002347, 0.00000133, -0.00001346, -0.00002855, -0.00000704, -0.00000008, -0.00000678, 0.00000438, 0.00000204, -0.00001154, 0.00000053, 0.00000092, -0.00002055, 0.00000228, 0.00003356, -0.00000410, -0.00003130, 0.00000898, 0.00000954, -0.00003321, -0.00000158, 0.00004298, -0.00000589, -0.00003664, 0.00001727, 0.00002085, -0.00003187, -0.00001045, 0.00003126, -0.00000390, -0.00002068, 0.00002060, 0.00000694, -0.00003704, -0.00000371, 0.00003238, -0.00000816, -0.00002496, 0.00001968, 0.00002451, -0.00000974, -0.00000749, 0.00000173, -0.00001419, -0.00000520, 0.00002038, 0.00000581, -0.00001648, -0.00000184, 0.00000743, -0.00000903, -0.00000719, 0.00001069, 0.00000546, -0.00000831, -0.00000286, 0.00000524, 0.00000328, 0.00000393, 0.00000552, -0.00000187, -0.00001013, -0.00000962, -0.00000266, 0.00000398, 0.00000250, -0.00000368, 0.00000066, 0.00000915, 0.00000094, -0.00000831, 0.00000685, 0.00001852, -0.00000164, -0.00001522, 0.00000436, 0.00001220, -0.00000974, -0.00001237, 0.00001068, 0.00000867, -0.00001603, -0.00001724, -0.00000116, -0.00000079, -0.00000670, -0.00000061, 0.00000671, 0.00000604, 0.00000081, -0.00000832, -0.00001352, -0.00000465, 0.00000560, 0.00000056, -0.00000875, -0.00000589, 0.00000090, },
		{ -0.00000388, 0.00000039, -0.00000416, -0.00000764, -0.00000012, 0.00000854, 0.00000677, -0.00000036, -0.00000724, -0.00001153, -0.00000351, 0.00000837, -0.00000036, -0.00001111, 0.00000531, 0.00000755, -0.00002835, -0.00003011, 0.00000478, -0.00003721, -0.00012965, -0.00010134, 0.00005066, 0.00018183, 0.00025258, 0.00024449, 0.00008298, -0.00012139, -0.00014276, -0.00003199, 0.00000013, -0.00002823, -0.00001010, 0.00000131, -0.00002607, -0.00002142, 0.00000067, -0.00000801, -0.00000489, 0.00001206, -0.00000515, -0.00001395, 0.00000919, -0.00000464, -0.00002692, 0.00001541, 0.00003536, -0.00002037, -0.00002703, 0.00002190, -0.00000253, -0.00003957, 0.00001636, 0.00004200, -0.00002445, -0.00002845, 0.00003120, 0.00000668, -0.00003863, 0.00000636, 0.00003139, -0.00001725, -0.00001408, 0.00002698, -0.00000563, -0.00003772, 0.00000933, 0.00002696, -0.00001884, -0.00001465, 0.00002894, 0.00001598, -0.00001511, -0.00000515, -0.00000091, -0.00001343, 0.00000299, 0.00001979, -0.00000159, -0.00001518, 0.00000154, 0.00000391, -0.00000915, -0.00000207, 0.00000961, 0.00000112, -0.00000679, -0.00000054, 0.00000385, 0.00000349, 0.00000564, 0.00000401, -0.00000432, -0.00001034, -0.00000885, -0.00000135, 0.00000471, 0.00000081, -0.00000364, 0.00000434, 0.00000847, -0.00000416, -0.00000578, 0.00001410, 0.00001518, -0.00000952, -0.00001154, 0.00001022, 0.00000624, -0.00001436, -0.00000507, 0.00001381, 0.00000025, -0.00001977, -0.00001154, 0.00000136, -0.00000315, -0.00000653, 0.00000120, 0.00000720, 0.00000625, -0.00000044, -0.00001089, -0.00001307, -0.00000207, 0.00000520, -0.00000143, -0.00000795, },
		{ -0.00000269, -0.00000672, -0.00000236, -0.00000067, -0.00000618, -0.00000670, 0.00000237, 0.00000785, 0.00000486, 0.00000005, -0.00000824, -0.00001338, 0.00000004, 0.00001080, -0.00000582, -0.00001165, 0.00001189, 0.00000248, -0.00003819, -0.00002375, 0.00000739, -0.00005782, -0.00014155, -0.00008538, 0.00007563, 0.00022277, 0.00030229, 0.00022850, -0.00000669, -0.00016803, -0.00010463, -0.00000506, -0.00001684, -0.00002771, 0.00000449, -0.00000926, -0.00003351, -0.00000778, 0.00000064, -0.00001672, 0.00000417, 0.00001549, -0.00001641, -0.00000966, 0.00001696, -0.00001657, -0.00002758, 0.00003207, 0.00002851, -0.00003585, -0.00001417, 0.00002949, -0.00002033, -0.00003721, 0.00003556, 0.00003129, -0.00003955, -0.00001216, 0.00003798, -0.00001202, -0.00003646, 0.00002373, 0.00002347, -0.00002783, -0.00000269, 0.00002815, -0.00001914, -0.00003262, 0.00001988, 0.00001694, -0.00002438, -0.00000100, 0.00003282, 0.00000529, -0.00001771, -0.00000285, -0.00000328, -0.00001020, 0.00000964, 0.00001623, -0.00000730, -0.00001245, 0.00000309, 0.00000094, -0.00000732, 0.00000131, 0.00000694, -0.00000130, -0.00000459, 0.00000024, 0.00000282, 0.00000473, 0.00000635, 0.00000183, -0.00000605, -0.00001037, -0.00000804, 0.00000030, 0.00000472, -0.00000097, -0.00000199, 0.00000739, 0.00000526, -0.00000767, 0.00000001, 0.00001877, 0.00000823, -0.00001429, -0.00000501, 0.00001230, -0.00000132, -0.00001445, 0.00000282, 0.00001214, -0.00000833, -0.00001944, -0.00000570, 0.00000160, -0.00000531, -0.00000558, 0.00000276, 0.00000790, 0.00000625, -0.00000262, -0.00001300, -0.00001164, -0.00000009, 0.00000422, },
		{ 0.00000116, 0.00000315, -0.00000333, -0.00000532, -0.00000127, -0.00000236, -0.00000793, -0.00000476, 0.00000435, 0.00000608, 0.00000397, 0.00000090, -0.00001084, -0.00001425, 0.00000549, 0.00001062, -0.00001249, -0.00000832, 0.00001753, -0.00000730, -0.00004487, -0.00001409, 0.00000401, -0.00007986, -0.00014769, -0.00007024, 0.00010613, 0.00028250, 0.00034343, 0.00017150, -0.00010371, -0.00017725, -0.00005402, -0.00000006, -0.00003544, -0.00001375, 0.00001177, -0.00002700, -0.00003035, 0.00000659, -0.00000858, -0.00002045, 0.00001817, 0.00001081, -0.00002726, 0.00000205, 0.00001908, -0.00003234, -0.00001928, 0.00004750, 0.00001249, -0.00004527, 0.00000487, 0.00002772, -0.00003925, -0.00002460, 0.00005059, 0.00001220, -0.00004645, 0.00000849, 0.00003476, -0.00003029, -0.00002487, 0.00003691, 0.00000895, -0.00003243, 0.00001081, 0.00002308, -0.00003069, -0.00002274, 0.00002570, 0.00000480, -0.00002351, 0.00001304, 0.00003075, -0.00000548, -0.00001754, -0.00000102, -0.00000474, -0.00000535, 0.00001382, 0.00001093, -0.00001087, -0.00000937, 0.00000326, -0.00000072, -0.00000470, 0.00000258, 0.00000400, -0.00000186, -0.00000276, -0.00000002, 0.00000273, 0.00000627, 0.00000592, -0.00000042, -0.00000719, -0.00001040, -0.00000690, 0.00000195, 0.00000388, -0.00000207, 0.00000091, 0.00000863, 0.00000054, -0.00000811, 0.00000736, 0.00001934, -0.00000014, -0.00001481, 0.00000192, 0.00001020, -0.00000782, -0.00001040, 0.00000861, 0.00000651, -0.00001477, -0.00001589, -0.00000117, 0.00000020, -0.00000675, -0.00000420, 0.00000427, 0.00000877, 0.00000556, -0.00000535, -0.00001421, -0.00000978, },
		{ -0.00001453, -0.00000797, 0.00000192, 0.00000226, -0.00000359, -0.00000378, -0.00000068, -0.00000476, -0.00000893, -0.00000191, 0.00000513, 0.00000381, 0.00000463, 0.00000108, -0.00001486, -0.00001277, 0.00001158, 0.00000679, -0.00001813, -0.00000120, 0.00001949, -0.00002010, -0.00004622, -0.00000427, -0.00000560, -0.00009982, -0.00015099, -0.00005670, 0.00015182, 0.00035876, 0.00035684, 0.00007401, -0.00018510, -0.00014703, -0.00000936, -0.00001656, -0.00004385, 0.00000856, 0.00000547, -0.00004424, -0.00001530, 0.00001439, -0.00002416, -0.00001453, 0.00003180, -0.00000295, -0.00003216, 0.00001857, 0.00001188, -0.00004691, -0.00000147, 0.00005611, -0.00001005, -0.00004440, 0.00002505, 0.00001508, -0.00005343, -0.00000336, 0.00005648, -0.00001106, -0.00004241, 0.00002803, 0.00002155, -0.00004283, -0.00000616, 0.00004203, -0.00000865, -0.00002942, 0.00002292, 0.00001243, -0.00003772, -0.00001034, 0.00002576, -0.00000639, -0.00001650, 0.00002445, 0.00002343, -0.00001444, -0.00001513, 0.00000018, -0.00000489, 0.00000004, 0.00001525, 0.00000514, -0.00001246, -0.00000666, 0.00000284, -0.00000098, -0.00000243, 0.00000212, 0.00000184, -0.00000122, -0.00000186, -0.00000056, 0.00000362, 0.00000743, 0.00000457, -0.00000235, -0.00000800, -0.00001036, -0.00000532, 0.00000317, 0.00000247, -0.00000192, 0.00000410, 0.00000746, -0.00000402, -0.00000506, 0.00001398, 0.00001566, -0.00000743, -0.00001144, 0.00000689, 0.00000505, -0.00001120, -0.00000398, 0.00001058, -0.00000115, -0.00001776, -0.00001066, 0.00000134, -0.00000198, -0.00000738, -0.00000256, 0.00000589, 0.00000943, 0.00000398, -0.00000803, },
		{ 0.00000177, -0.00001021, -0.00001433, -0.00000641, 0.00000254, 0.00000147, -0.00000358, -0.00000201, -0.00000090, -0.00000773, -0.00000860, 0.00000131, 0.00000426, 0.00000205, 0.00000667, -0.00000056, -0.00001918, -0.00000822, 0.00001627, -0.00000036, -0.00002035, 0.00000783, 0.00001597, -0.00003278, -0.00004211, 0.00000228, -0.00001917, -0.00011615, -0.00015615, -0.00004031, 0.00022244, 0.00043972, 0.00032441, -0.00005076, -0.00022916, -0.00008757, 0.00001306, -0.00004516, -0.00003431, 0.00002909, -0.00001520, -0.00005200, 0.00000759, 0.00000998, -0.00003955, 0.00000221, 0.00003843, -0.00002289, -0.00002672, 0.00003444, -0.00000543, -0.00005428, 0.00002297, 0.00005344, -0.00003359, -0.00003178, 0.00004009, -0.00000647, -0.00005766, 0.00002184, 0.00005053, -0.00003256, -0.00002779, 0.00004084, 0.00000139, -0.00004567, 0.00001510, 0.00003736, -0.00002480, -0.00001934, 0.00003037, -0.00000160, -0.00003883, 0.00000169, 0.00002059, -0.00001382, -0.00000514, 0.00003092, 0.00001262, -0.00002027, -0.00001135, 0.00000085, -0.00000366, 0.00000492, 0.00001422, -0.00000015, -0.00001257, -0.00000458, 0.00000252, -0.00000030, -0.00000125, 0.00000076, 0.00000094, -0.00000024, -0.00000191, -0.00000070, 0.00000515, 0.00000776, 0.00000274, -0.00000384, -0.00000869, -0.00001003, -0.00000343, 0.00000365, 0.00000110, -0.00000041, 0.00000643, 0.00000420, -0.00000671, 0.00000070, 0.00001784, 0.00000898, -0.00001167, -0.00000581, 0.00000851, -0.00000102, -0.00001069, 0.00000237, 0.00000839, -0.00000853, -0.00001718, -0.00000534, 0.00000184, -0.00000426, -0.00000724, -0.00000063, 0.00000757, 0.00000946, },
		{ 0.00000897, 0.00000865, -0.00000058, -0.00001180, -0.00001399, -0.00000490, 0.00000317, 0.00000056, -0.00000314, -0.00000009, -0.00000236, -0.00001072, -0.00000656, 0.00000391, 0.00000188, 0.00000188, 0.00000913, -0.00000463, -0.00002205, -0.00000107, 0.00001756, -0.00000897, -0.00001783, 0.00001572, 0.00000720, -0.00004200, -0.00003479, 0.00000368, -0.00003289, -0.00013079, -0.00016684, -0.00001066, 0.00032274, 0.00050451, 0.00023599, -0.00017850, -0.00022273, -0.00001882, 0.00000511, -0.00007085, -0.00000678, 0.00003672, -0.00004440, -0.00004372, 0.00002990, -0.00000790, -0.00004676, 0.00002632, 0.00003286, -0.00004276, -0.00000983, 0.00004301, -0.00002981, -0.00004968, 0.00004812, 0.00003804, -0.00005140, -0.00000964, 0.00004450, -0.00003182, -0.00004926, 0.00004459, 0.00003341, -0.00004642, -0.00000619, 0.00004304, -0.00002039, -0.00003751, 0.00003339, 0.00002391, -0.00003531, -0.00000475, 0.00003115, -0.00001587, -0.00003418, 0.00001069, 0.00001213, -0.00001571, 0.00000776, 0.00003132, 0.00000066, -0.00002247, -0.00000711, 0.00000128, -0.00000135, 0.00000849, 0.00001139, -0.00000437, -0.00001180, -0.00000302, 0.00000271, 0.00000058, -0.00000133, -0.00000056, 0.00000118, 0.00000036, -0.00000252, -0.00000005, 0.00000675, 0.00000718, 0.00000086, -0.00000498, -0.00000932, -0.00000923, -0.00000158, 0.00000339, 0.00000040, 0.00000198, 0.00000696, -0.00000003, -0.00000643, 0.00000741, 0.00001783, 0.00000137, -0.00001212, -0.00000010, 0.00000673, -0.00000581, -0.00000698, 0.00000649, 0.00000306, -0.00001378, -0.00001388, -0.00000115, 0.00000078, -0.00000618, -0.00000639, 0.00000160, },
		{ -0.00000480, 0.00000397, 0.00000967, 0.00000723, -0.00000263, -0.00001311, -0.00001367, -0.00000311, 0.00000365, -0.00000056, -0.00000194, 0.00000154, -0.00000530, -0.00001282, -0.00000302, 0.00000488, -0.00000108, 0.00000384, 0.00001046, -0.00001067, -0.00002183, 0.00000690, 0.00001447, -0.00001625, -0.00001114, 0.00001947, -0.00000441, -0.00004596, -0.00002798, 0.00000092, -0.00004365, -0.00014932, -0.00018198, 0.00004656, 0.00044752, 0.00052755, 0.00009511, -0.00027960, -0.00016648, 0.00003566, -0.00002988, -0.00007875, 0.00003040, 0.00002405, -0.00007132, -0.00001855, 0.00004179, -0.00003488, -0.00003958, 0.00005041, 0.00001365, -0.00005493, 0.00001532, 0.00003897, -0.00005490, -0.00003157, 0.00006676, 0.00001242, -0.00005773, 0.00001644, 0.00003564, -0.00005408, -0.00002920, 0.00005871, 0.00000919, -0.00004877, 0.00001655, 0.00003376, -0.00003777, -0.00002019, 0.00004393, 0.00000516, -0.00003759, 0.00001053, 0.00002507, -0.00002730, -0.00002544, 0.00001504, 0.00000318, -0.00001179, 0.00001908, 0.00002598, -0.00001016, -0.00002135, -0.00000314, 0.00000180, 0.00000154, 0.00001032, 0.00000761, -0.00000731, -0.00001061, -0.00000166, 0.00000339, 0.00000097, -0.00000233, -0.00000117, 0.00000204, 0.00000023, -0.00000309, 0.00000136, 0.00000789, 0.00000588, -0.00000079, -0.00000594, -0.00000978, -0.00000796, -0.00000011, 0.00000271, 0.00000076, 0.00000433, 0.00000542, -0.00000367, -0.00000310, 0.00001302, 0.00001411, -0.00000498, -0.00000933, 0.00000380, 0.00000277, -0.00000784, -0.00000182, 0.00000719, -0.00000360, -0.00001597, -0.00000918, 0.00000135, -0.00000123, -0.00000748, },
		{ -0.00000367, -0.00000794, -0.00000248, 0.00000606, 0.00000944, 0.00000569, -0.00000428, -0.00001456, -0.00001311, -0.00000089, 0.00000354, -0.00000163, 0.00000023, 0.00000214, -0.00000945, -0.00001308, 0.00000115, 0.00000370, -0.00000324, 0.00000754, 0.00000927, -0.00001707, -0.00001784, 0.00001317, 0.00000777, -0.00001964, -0.00000282, 0.00001763, -0.00001511, -0.00004555, -0.00002491, -0.00000212, -0.00005169, -0.00017877, -0.00019303, 0.00014463, 0.00057910, 0.00048596, -0.00007890, -0.00032761, -0.00007601, 0.00005646, -0.00007786, -0.00006014, 0.00006318, -0.00000891, -0.00008398, 0.00001752, 0.00003590, -0.00006197, -0.00001639, 0.00006550, -0.00001559, -0.00005320, 0.00004180, 0.00002054, -0.00007291, -0.00000270, 0.00007276, -0.00001737, -0.00004985, 0.00003912, 0.00001475, -0.00006677, -0.00000203, 0.00006030, -0.00001581, -0.00003904, 0.00003414, 0.00001556, -0.00004595, 0.00000174, 0.00004414, -0.00001401, -0.00003144, 0.00002262, 0.00001381, -0.00003376, -0.00001522, 0.00001458, -0.00000347, -0.00000339, 0.00002628, 0.00001647, -0.00001807, -0.00001779, 0.00000014, 0.00000266, 0.00000435, 0.00001039, 0.00000363, -0.00000909, -0.00000924, -0.00000022, 0.00000421, 0.00000050, -0.00000361, -0.00000083, 0.00000290, -0.00000057, -0.00000315, 0.00000319, 0.00000826, 0.00000421, -0.00000213, -0.00000684, -0.00000990, -0.00000639, 0.00000075, 0.00000214, 0.00000212, 0.00000568, 0.00000231, -0.00000533, 0.00000231, 0.00001585, 0.00000799, -0.00000853, -0.00000484, 0.00000497, -0.00000157, -0.00000682, 0.00000276, 0.00000452, -0.00000967, -0.00001514, -0.00000438, 0.00000207, },
		{ -0.00000044, 0.00000125, -0.00000607, -0.00000732, 0.00000037, 0.00000732, 0.00000849, 0.00000456, -0.00000590, -0.00001632, -0.00001179, 0.00000153, 0.00000251, -0.00000204, 0.00000316, 0.00000091, -0.00001391, -0.00001102, 0.00000462, 0.00000069, -0.00000334, 0.00001159, 0.00000505, -0.00002168, -0.00001100, 0.00001563, -0.00000005, -0.00001819, 0.00000356, 0.00001137, -0.00002170, -0.00004427, -0.00002618, -0.00000047, -0.00006228, -0.00022363, -0.00018367, 0.00028929, 0.00068867, 0.00036800, -0.00025359, -0.00030790, 0.00002193, 0.00003520, -0.00011871, -0.00001634, 0.00007694, -0.00005386, -0.00007400, 0.00005314, 0.00001060, -0.00007846, 0.00001848, 0.00006436, -0.00004696, -0.00003540, 0.00006102, -0.00000936, -0.00007724, 0.00003040, 0.00006327, -0.00004350, -0.00002923, 0.00005161, -0.00001314, -0.00006594, 0.00002543, 0.00004902, -0.00003503, -0.00002027, 0.00004184, -0.00000637, -0.00004278, 0.00002249, 0.00003435, -0.00002880, -0.00001897, 0.00002873, 0.00000026, -0.00003465, -0.00000615, 0.00001065, -0.00000588, 0.00000699, 0.00002794, 0.00000509, -0.00002221, -0.00001292, 0.00000267, 0.00000390, 0.00000650, 0.00000899, 0.00000003, -0.00000992, -0.00000774, 0.00000139, 0.00000472, -0.00000080, -0.00000452, 0.00000029, 0.00000324, -0.00000167, -0.00000245, 0.00000497, 0.00000783, 0.00000250, -0.00000323, -0.00000766, -0.00000957, -0.00000482, 0.00000109, 0.00000215, 0.00000394, 0.00000538, -0.00000123, -0.00000425, 0.00000812, 0.00001517, 0.00000136, -0.00000886, -0.00000043, 0.00000363, -0.00000460, -0.00000362, 0.00000514, -0.00000043, -0.00001375, -0.00001207, },
		{ -0.00001524, -0.00000780, 0.00000212, -0.00000077, -0.00000795, -0.00000545, 0.00000319, 0.00000739, 0.00000744, 0.00000401, -0.00000807, -0.00001809, -0.00000931, 0.00000349, 0.00000065, -0.00000109, 0.00000603, -0.00000250, -0.00001738, -0.00000697, 0.00000614, -0.00000290, -0.00000090, 0.00001413, -0.00000133, -0.00002279, -0.00000363, 0.00001367, -0.00000580, -0.00001335, 0.00000513, 0.00000423, -0.00002341, -0.00004669, -0.00002841, 0.00000881, -0.00008518, -0.00028125, -0.00013305, 0.00047289, 0.00074229, 0.00017956, -0.00039107, -0.00022332, 0.00009729, -0.00002210, -0.00013347, 0.00004097, 0.00006230, -0.00009642, -0.00004030, 0.00007537, -0.00002861, -0.00007591, 0.00005572, 0.00004443, -0.00007064, -0.00000462, 0.00006565, -0.00004364, -0.00006483, 0.00005923, 0.00003995, -0.00005876, -0.00000132, 0.00004993, -0.00004073, -0.00005153, 0.00004625, 0.00002830, -0.00004367, 0.00000194, 0.00003794, -0.00002582, -0.00002948, 0.00003678, 0.00001766, -0.00003593, -0.00000388, 0.00002789, -0.00001238, -0.00003092, -0.00000013, 0.00000553, -0.00000352, 0.00001649, 0.00002413, -0.00000574, -0.00002264, -0.00000779, 0.00000460, 0.00000536, 0.00000759, 0.00000665, -0.00000286, -0.00001007, -0.00000607, 0.00000305, 0.00000453, -0.00000257, -0.00000463, 0.00000173, 0.00000289, -0.00000259, -0.00000108, 0.00000629, 0.00000677, 0.00000097, -0.00000420, -0.00000831, -0.00000881, -0.00000351, 0.00000120, 0.00000296, 0.00000540, 0.00000340, -0.00000383, -0.00000064, 0.00001252, 0.00001138, -0.00000394, -0.00000661, 0.00000248, 0.00000090, -0.00000541, 0.00000019, 0.00000467, -0.00000608, },
		{ 0.00000170, -0.00001095, -0.00001427, -0.00000333, 0.00000306, -0.00000357, -0.00000878, -0.00000244, 0.00000520, 0.00000635, 0.00000702, 0.00000366, -0.00001122, -0.00001913, -0.00000572, 0.00000421, -0.00000140, 0.00000154, 0.00000769, -0.00000768, -0.00001865, -0.00000212, 0.00000522, -0.00000543, 0.00000332, 0.00001372, -0.00000780, -0.00002007, 0.00000146, 0.00000884, -0.00000712, -0.00000861, 0.00000143, 0.00000065, -0.00002303, -0.00005579, -0.00002490, 0.00002336, -0.00013143, -0.00033863, -0.00002234, 0.00067168, 0.00071027, -0.00005383, -0.00045835, -0.00009493, 0.00012612, -0.00009650, -0.00011129, 0.00009334, 0.00001926, -0.00012106, 0.00000968, 0.00007437, -0.00007047, -0.00005150, 0.00008390, 0.00000926, -0.00007826, 0.00003132, 0.00005230, -0.00007308, -0.00003754, 0.00007608, 0.00000866, -0.00005901, 0.00002614, 0.00003423, -0.00006074, -0.00002739, 0.00005553, 0.00000419, -0.00004025, 0.00002117, 0.00002425, -0.00003755, -0.00001011, 0.00004142, -0.00000117, -0.00003448, 0.00000979, 0.00002110, -0.00002162, -0.00002462, 0.00000230, 0.00000165, 0.00000265, 0.00002262, 0.00001619, -0.00001413, -0.00002015, -0.00000310, 0.00000618, 0.00000670, 0.00000748, 0.00000395, -0.00000492, -0.00000972, -0.00000419, 0.00000448, 0.00000351, -0.00000430, -0.00000387, 0.00000298, 0.00000196, -0.00000297, 0.00000065, 0.00000691, 0.00000537, -0.00000032, -0.00000510, -0.00000866, -0.00000780, -0.00000258, 0.00000151, 0.00000437, 0.00000576, 0.00000040, -0.00000438, 0.00000441, 0.00001422, 0.00000576, -0.00000678, -0.00000310, 0.00000338, -0.00000178, -0.00000411, 0.00000303, },
		{ -0.00000166, 0.00000389, -0.00000272, -0.00001408, -0.00001147, 0.00000057, 0.00000238, -0.00000659, -0.00000807, 0.00000115, 0.00000572, 0.00000487, 0.00000770, 0.00000270, -0.00001525, -0.00001860, -0.00000170, 0.00000328, -0.00000259, 0.00000544, 0.00000705, -0.00001342, -0.00001722, 0.00000196, 0.00000247, -0.00000560, 0.00000754, 0.00001019, -0.00001195, -0.00001510, 0.00000232, 0.00000441, -0.00000394, -0.00000808, -0.00000475, 0.00000356, -0.00002624, -0.00007033, -0.00000833, 0.00003377, -0.00020795, -0.00037255, 0.00015718, 0.00084805, 0.00057724, -0.00029097, -0.00043752, 0.00004314, 0.00009845, -0.00016214, -0.00005393, 0.00012190, -0.00004195, -0.00011665, 0.00006188, 0.00004730, -0.00010145, -0.00000959, 0.00009335, -0.00003221, -0.00006590, 0.00006223, 0.00002299, -0.00008929, -0.00000175, 0.00007651, -0.00002231, -0.00004452, 0.00004552, 0.00000880, -0.00006817, -0.00000005, 0.00005193, -0.00001665, -0.00002694, 0.00003227, 0.00000541, -0.00003888, 0.00000974, 0.00003612, -0.00001725, -0.00002598, 0.00001887, 0.00001078, -0.00002635, -0.00001805, 0.00000194, 0.00000078, 0.00001055, 0.00002402, 0.00000625, -0.00001906, -0.00001588, 0.00000082, 0.00000759, 0.00000752, 0.00000632, 0.00000139, -0.00000622, -0.00000898, -0.00000219, 0.00000537, 0.00000180, -0.00000549, -0.00000245, 0.00000369, 0.00000079, -0.00000266, 0.00000233, 0.00000681, 0.00000390, -0.00000141, -0.00000591, -0.00000866, -0.00000677, -0.00000192, 0.00000232, 0.00000581, 0.00000464, -0.00000250, -0.00000250, 0.00000933, 0.00001282, -0.00000001, -0.00000699, 0.00000028, 0.00000265, -0.00000331, },
		{ 0.00000123, -0.00000331, 0.00000067, 0.00000257, -0.00000732, -0.00001514, -0.00000758, 0.00000338, 0.00000021, -0.00000910, -0.00000558, 0.00000438, 0.00000452, 0.00000397, 0.00000929, 0.00000025, -0.00001937, -0.00001605, 0.00000164, 0.00000098, -0.00000194, 0.00000947, 0.00000373, -0.00001812, -0.00001362, 0.00000397, -0.00000057, -0.00000326, 0.00000969, 0.00000500, -0.00001224, -0.00001089, -0.00000082, 0.00000401, 0.00000086, -0.00001430, -0.00000804, 0.00001214, -0.00003927, -0.00008367, 0.00002504, 0.00002470, -0.00031176, -0.00035402, 0.00039700, 0.00095819, 0.00034970, -0.00048439, -0.00033213, 0.00015329, 0.00002190, -0.00019484, 0.00002424, 0.00011421, -0.00010370, -0.00008087, 0.00010012, -0.00000016, -0.00011042, 0.00003927, 0.00007967, -0.00006838, -0.00003576, 0.00007886, -0.00001528, -0.00008747, 0.00003354, 0.00006082, -0.00004486, -0.00001990, 0.00005173, -0.00001937, -0.00006184, 0.00002345, 0.00003791, -0.00002902, -0.00000869, 0.00003297, -0.00001274, -0.00003037, 0.00002481, 0.00002339, -0.00002697, -0.00001369, 0.00002188, -0.00000016, -0.00002692, -0.00001297, 0.00000052, 0.00000345, 0.00001765, 0.00002067, -0.00000343, -0.00002045, -0.00001092, 0.00000400, 0.00000878, 0.00000752, 0.00000449, -0.00000067, -0.00000691, -0.00000794, -0.00000022, 0.00000549, -0.00000025, -0.00000587, -0.00000078, 0.00000374, -0.00000023, -0.00000175, 0.00000362, 0.00000616, 0.00000254, -0.00000237, -0.00000656, -0.00000834, -0.00000588, -0.00000126, 0.00000367, 0.00000660, 0.00000226, -0.00000415, 0.00000137, 0.00001261, 0.00000888, -0.00000448, -0.00000519, 0.00000260, },
		{ -0.00000245, 0.00000369, 0.00000009, -0.00000221, 0.00000181, -0.00000035, -0.00001108, -0.00001430, -0.00000331, 0.00000470, -0.00000310, -0.00001025, -0.00000160, 0.00000614, 0.00000209, 0.00000461, 0.00001092, -0.00000409, -0.00002238, -0.00001178, 0.00000326, -0.00000167, 0.00000088, 0.00001214, -0.00000166, -0.00002046, -0.00000933, 0.00000365, -0.00000207, 0.00000033, 0.00000852, 0.00000083, -0.00000908, -0.00001057, -0.00000517, 0.00000960, 0.00000226, -0.00002628, -0.00000274, 0.00002089, -0.00006535, -0.00008499, 0.00007130, -0.00002066, -0.00042620, -0.00025650, 0.00066845, 0.00096347, 0.00005804, -0.00059276, -0.00016754, 0.00020544, -0.00008047, -0.00018030, 0.00010066, 0.00006898, -0.00014670, -0.00002156, 0.00011159, -0.00005526, -0.00009256, 0.00008161, 0.00004544, -0.00008876, 0.00000440, 0.00007594, -0.00005265, -0.00006801, 0.00005956, 0.00003397, -0.00005363, 0.00000746, 0.00004373, -0.00004287, -0.00004456, 0.00003776, 0.00001872, -0.00003078, 0.00000855, 0.00002436, -0.00002505, -0.00001545, 0.00003171, 0.00000754, -0.00002901, -0.00000140, 0.00001923, -0.00000928, -0.00002475, -0.00001004, -0.00000007, 0.00000885, 0.00002186, 0.00001375, -0.00001106, -0.00001900, -0.00000606, 0.00000661, 0.00000955, 0.00000660, 0.00000248, -0.00000209, -0.00000717, -0.00000666, 0.00000149, 0.00000482, -0.00000220, -0.00000542, 0.00000074, 0.00000326, -0.00000081, -0.00000052, 0.00000435, 0.00000518, 0.00000135, -0.00000323, -0.00000697, -0.00000786, -0.00000516, -0.00000029, 0.00000526, 0.00000622, -0.00000063, -0.00000378, 0.00000609, 0.00001332, 0.00000362, -0.00000687, },
		{ -0.00000155, -0.00000717, 0.00000028, 0.00000396, -0.00000023, -0.00000093, 0.00000133, -0.00000394, -0.00001345, -0.00001196, 0.00000073, 0.00000424, -0.00000685, -0.00000933, 0.00000293, 0.00000568, -0.00000042, 0.00000709, 0.00001131, -0.00000988, -0.00002318, -0.00000686, 0.00000273, -0.00000327, 0.00000526, 0.00001222, -0.00000760, -0.00002011, -0.00000609, 0.00000204, -0.00000095, 0.00000284, 0.00000451, 0.00000033, -0.00000515, -0.00001567, -0.00000620, 0.00001991, -0.00000481, -0.00003886, 0.00001410, 0.00002088, -0.00010162, -0.00006312, 0.00011694, -0.00011366, -0.00052131, -0.00006588, 0.00092558, 0.00084236, -0.00024843, -0.00059263, 0.00001531, 0.00018616, -0.00017720, -0.00011911, 0.00015195, -0.00000289, -0.00015648, 0.00004542, 0.00009129, -0.00010213, -0.00005126, 0.00010544, -0.00000038, -0.00008741, 0.00004364, 0.00005393, -0.00007953, -0.00003633, 0.00007050, 0.00000383, -0.00004760, 0.00002981, 0.00002468, -0.00005621, -0.00002200, 0.00004078, 0.00000053, -0.00002328, 0.00001981, 0.00001029, -0.00002871, 0.00000092, 0.00002979, -0.00000687, -0.00002436, 0.00000776, 0.00001278, -0.00001527, -0.00002164, -0.00000878, 0.00000154, 0.00001521, 0.00002203, 0.00000518, -0.00001569, -0.00001570, -0.00000169, 0.00000874, 0.00000962, 0.00000492, 0.00000076, -0.00000292, -0.00000716, -0.00000523, 0.00000272, 0.00000352, -0.00000368, -0.00000435, 0.00000183, 0.00000252, -0.00000085, 0.00000071, 0.00000450, 0.00000408, 0.00000032, -0.00000397, -0.00000713, -0.00000738, -0.00000448, 0.00000113, 0.00000653, 0.00000454, -0.00000300, -0.00000129, 0.00001024, 0.00001135, },
		{ 0.00001263, 0.00000733, -0.00000556, -0.00000594, 0.00000250, 0.00000410, 0.00000015, -0.00000038, -0.00000053, -0.00000728, -0.00001432, -0.00000858, 0.00000393, 0.00000198, -0.00001002, -0.00000612, 0.00000665, 0.00000298, -0.00000160, 0.00001085, 0.00000938, -0.00001591, -0.00002134, -0.00000276, 0.00000058, -0.00000273, 0.00000970, 0.00000948, -0.00001231, -0.00001798, -0.00000502, 0.00000114, 0.00000237, 0.00000203, 0.00000011, 0.00000465, -0.00000448, -0.00002476, 0.00000044, 0.00003003, -0.00002272, -0.00004403, 0.00004007, 0.00000309, -0.00013785, -0.00001197, 0.00014053, -0.00025316, -0.00055935, 0.00021115, 0.00111433, 0.00059920, -0.00051250, -0.00048572, 0.00017098, 0.00010257, -0.00023839, -0.00002676, 0.00016174, -0.00008159, -0.00012837, 0.00010124, 0.00004394, -0.00012703, 0.00000284, 0.00010412, -0.00004531, -0.00006488, 0.00007134, 0.00001885, -0.00008958, -0.00000102, 0.00006520, -0.00002139, -0.00003013, 0.00004152, 0.00000064, -0.00005735, -0.00000062, 0.00003404, -0.00001169, -0.00001057, 0.00002265, -0.00000412, -0.00002387, 0.00001393, 0.00002104, -0.00001641, -0.00001574, 0.00001217, 0.00000494, -0.00001811, -0.00001895, -0.00000792, 0.00000562, 0.00002046, 0.00001829, -0.00000306, -0.00001728, -0.00001155, 0.00000214, 0.00001033, 0.00000876, 0.00000289, -0.00000037, -0.00000335, -0.00000698, -0.00000377, 0.00000335, 0.00000189, -0.00000446, -0.00000296, 0.00000239, 0.00000183, -0.00000044, 0.00000166, 0.00000418, 0.00000301, -0.00000057, -0.00000454, -0.00000710, -0.00000697, -0.00000359, 0.00000289, 0.00000694, 0.00000197, -0.00000391, 0.00000269, },
		{ -0.00000288, 0.00000702, 0.00001263, 0.00000239, -0.00000788, -0.00000392, 0.00000429, 0.00000458, 0.00000059, -0.00000108, -0.00000309, -0.00000982, -0.00001381, -0.00000461, 0.00000573, -0.00000175, -0.00001149, -0.00000117, 0.00000822, -0.00000102, -0.00000039, 0.00001453, 0.00000482, -0.00002058, -0.00001742, -0.00000067, -0.00000182, 0.00000017, 0.00001249, 0.00000486, -0.00001456, -0.00001591, -0.00000586, 0.00000290, 0.00000574, -0.00000300, -0.00000113, 0.00001231, -0.00001078, -0.00003330, 0.00001646, 0.00003271, -0.00004897, -0.00003391, 0.00006660, -0.00003713, -0.00015836, 0.00006450, 0.00011831, -0.00042083, -0.00050469, 0.00054353, 0.00118579, 0.00026651, -0.00068337, -0.00029929, 0.00026225, -0.00001991, -0.00024505, 0.00007128, 0.00012611, -0.00014470, -0.00006923, 0.00013014, -0.00001749, -0.00012272, 0.00005526, 0.00007852, -0.00007737, -0.00002794, 0.00008056, -0.00001979, -0.00008157, 0.00002896, 0.00004721, -0.00003572, -0.00000760, 0.00004064, -0.00002156, -0.00004806, 0.00001441, 0.00002175, -0.00001562, 0.00000224, 0.00001769, -0.00001430, -0.00001334, 0.00002043, 0.00000911, -0.00001964, -0.00000643, 0.00001188, -0.00000216, -0.00001873, -0.00001722, -0.00000609, 0.00001136, 0.00002292, 0.00001176, -0.00000947, -0.00001640, -0.00000723, 0.00000547, 0.00001117, 0.00000701, 0.00000101, -0.00000084, -0.00000362, -0.00000669, -0.00000240, 0.00000335, 0.00000025, -0.00000454, -0.00000160, 0.00000252, 0.00000140, 0.00000022, 0.00000221, 0.00000357, 0.00000204, -0.00000134, -0.00000492, -0.00000701, -0.00000659, -0.00000232, 0.00000462, 0.00000616, -0.00000070, },
		{ 0.00000428, -0.00000255, -0.00000006, 0.00001049, 0.00001034, -0.00000236, -0.00000861, -0.00000164, 0.00000601, 0.00000539, 0.00000030, -0.00000293, -0.00000557, -0.00001137, -0.00001212, -0.00000056, 0.00000561, -0.00000607, -0.00001039, 0.00000421, 0.00000688, -0.00000470, 0.00000333, 0.00001645, -0.00000161, -0.00002265, -0.00001289, -0.00000087, -0.00000284, 0.00000444, 0.00001248, 0.00000026, -0.00001446, -0.00001576, -0.00000682, 0.00000806, 0.00000606, -0.00001079, 0.00000392, 0.00001912, -0.00002527, -0.00003486, 0.00003917, 0.00002120, -0.00007571, -0.00000467, 0.00008079, -0.00009622, -0.00014690, 0.00014990, 0.00003262, -0.00058162, -0.00033553, 0.00087911, 0.00110997, -0.00010058, -0.00072996, -0.00007914, 0.00027025, -0.00014560, -0.00019479, 0.00014778, 0.00005489, -0.00017441, 0.00000479, 0.00012449, -0.00007597, -0.00009072, 0.00009233, 0.00003654, -0.00008881, 0.00001279, 0.00007004, -0.00005211, -0.00005939, 0.00004707, 0.00002329, -0.00003711, 0.00001291, 0.00002918, -0.00003660, -0.00003283, 0.00002075, 0.00000910, -0.00001205, 0.00001085, 0.00000806, -0.00001772, -0.00000134, 0.00001976, -0.00000200, -0.00001727, 0.00000080, 0.00000825, -0.00000732, -0.00001836, -0.00001606, -0.00000246, 0.00001721, 0.00002188, 0.00000415, -0.00001333, -0.00001392, -0.00000305, 0.00000828, 0.00001098, 0.00000464, -0.00000026, -0.00000080, -0.00000393, -0.00000630, -0.00000127, 0.00000283, -0.00000109, -0.00000406, -0.00000048, 0.00000240, 0.00000131, 0.00000089, 0.00000233, 0.00000283, 0.00000119, -0.00000195, -0.00000512, -0.00000696, -0.00000609, -0.00000062, 0.00000581, },
		{ 0.00000128, 0.00000604, 0.00000181, -0.00000285, 0.00000381, 0.00001216, 0.00000643, -0.00000615, -0.00000817, 0.00000077, 0.00000804, 0.00000601, -0.00000123, -0.00000533, -0.00000747, -0.00001198, -0.00000942, 0.00000289, 0.00000340, -0.00000967, -0.00000661, 0.00000829, 0.00000293, -0.00000635, 0.00000857, 0.00001545, -0.00000827, -0.00002185, -0.00000944, -0.00000245, -0.00000141, 0.00000821, 0.00000979, -0.00000239, -0.00001357, -0.00001834, -0.00000523, 0.00001531, 0.00000068, -0.00001756, 0.00001609, 0.00001943, -0.00004523, -0.00002378, 0.00006129, -0.00000741, -0.00009159, 0.00004037, 0.00006974, -0.00016110, -0.00009330, 0.00021742, -0.00011961, -0.00069000, -0.00005374, 0.00115488, 0.00088591, -0.00043662, -0.00064931, 0.00012269, 0.00019921, -0.00023944, -0.00010208, 0.00018231, -0.00003166, -0.00016303, 0.00007348, 0.00008697, -0.00011588, -0.00004055, 0.00010546, -0.00000967, -0.00007833, 0.00004634, 0.00004418, -0.00007093, -0.00003046, 0.00005101, 0.00000096, -0.00002766, 0.00002590, 0.00001195, -0.00004213, -0.00001712, 0.00001924, 0.00000037, -0.00000416, 0.00001309, -0.00000201, -0.00001448, 0.00000809, 0.00001354, -0.00000935, -0.00001152, 0.00000445, 0.00000321, -0.00001038, -0.00001794, -0.00001454, 0.00000293, 0.00002148, 0.00001765, -0.00000288, -0.00001469, -0.00001061, 0.00000090, 0.00001037, 0.00000963, 0.00000217, -0.00000067, -0.00000055, -0.00000439, -0.00000582, -0.00000044, 0.00000199, -0.00000197, -0.00000324, 0.00000032, 0.00000224, 0.00000149, 0.00000138, 0.00000212, 0.00000207, 0.00000045, -0.00000238, -0.00000523, -0.00000696, -0.00000527, },
		{ -0.00000695, -0.00000402, 0.00000300, 0.00000515, -0.00000045, -0.00000136, 0.00000766, 0.00001167, 0.00000186, -0.00000869, -0.00000695, 0.00000354, 0.00001033, 0.00000566, -0.00000398, -0.00000756, -0.00000862, -0.00001169, -0.00000595, 0.00000495, -0.00000047, -0.00001119, -0.00000098, 0.00000956, -0.00000225, -0.00000491, 0.00001350, 0.00001142, -0.00001330, -0.00001914, -0.00000821, -0.00000364, 0.00000230, 0.00000958, 0.00000598, -0.00000213, -0.00001432, -0.00002260, 0.00000118, 0.00002127, -0.00001099, -0.00001837, 0.00003275, 0.00000846, -0.00006402, 0.00000198, 0.00007278, -0.00004962, -0.00008568, 0.00008999, 0.00002599, -0.00021140, 0.00000043, 0.00023637, -0.00032124, -0.00070123, 0.00031070, 0.00131186, 0.00054459, -0.00068162, -0.00046723, 0.00026182, 0.00007408, -0.00027729, 0.00000681, 0.00016738, -0.00010949, -0.00011515, 0.00011899, 0.00002935, -0.00012778, 0.00001363, 0.00009346, -0.00004773, -0.00005099, 0.00006481, 0.00001126, -0.00007369, -0.00000301, 0.00004287, -0.00001400, -0.00001239, 0.00002907, -0.00000534, -0.00003913, -0.00000532, 0.00001314, -0.00000241, 0.00000391, 0.00000948, -0.00000879, -0.00000692, 0.00001236, 0.00000489, -0.00001173, -0.00000513, 0.00000450, -0.00000157, -0.00001193, -0.00001773, -0.00001176, 0.00000916, 0.00002297, 0.00001140, -0.00000821, -0.00001409, -0.00000694, 0.00000459, 0.00001141, 0.00000727, 0.00000019, -0.00000027, -0.00000045, -0.00000499, -0.00000524, 0.00000002, 0.00000104, -0.00000234, -0.00000231, 0.00000083, 0.00000219, 0.00000179, 0.00000159, 0.00000169, 0.00000134, -0.00000015, -0.00000262, -0.00000535, },
		{ -0.00000272, -0.00000556, -0.00000676, -0.00000240, 0.00000409, 0.00000337, -0.00000171, 0.00000158, 0.00001045, 0.00000925, -0.00000253, -0.00001008, -0.00000512, 0.00000692, 0.00001239, 0.00000375, -0.00000737, -0.00000905, -0.00000915, -0.00001049, -0.00000221, 0.00000495, -0.00000491, -0.00000972, 0.00000476, 0.00000738, -0.00000660, -0.00000057, 0.00001621, 0.00000554, -0.00001551, -0.00001635, -0.00000911, -0.00000264, 0.00000669, 0.00000760, 0.00000349, 0.00000023, -0.00001885, -0.00002564, 0.00001287, 0.00002162, -0.00002644, -0.00000915, 0.00004787, -0.00001518, -0.00007285, 0.00003875, 0.00006441, -0.00009518, -0.00005213, 0.00012733, -0.00004781, -0.00022540, 0.00011720, 0.00018188, -0.00053502, -0.00058465, 0.00070301, 0.00131045, 0.00014320, -0.00079586, -0.00023109, 0.00031290, -0.00006770, -0.00025238, 0.00010246, 0.00011015, -0.00015832, -0.00004555, 0.00013123, -0.00003167, -0.00011095, 0.00005770, 0.00006239, -0.00006884, -0.00001611, 0.00006551, -0.00001955, -0.00006272, 0.00001654, 0.00002793, -0.00001909, 0.00000279, 0.00002355, -0.00001808, -0.00003099, 0.00000048, 0.00000653, 0.00000013, 0.00000875, 0.00000258, -0.00001034, 0.00000145, 0.00001104, -0.00000292, -0.00000977, -0.00000036, 0.00000199, -0.00000512, -0.00001276, -0.00001735, -0.00000733, 0.00001483, 0.00002138, 0.00000461, -0.00001140, -0.00001221, -0.00000310, 0.00000781, 0.00001111, 0.00000435, -0.00000082, 0.00000063, -0.00000076, -0.00000561, -0.00000461, 0.00000015, 0.00000018, -0.00000227, -0.00000143, 0.00000116, 0.00000228, 0.00000206, 0.00000153, 0.00000115, 0.00000067, -0.00000058, },
		{ 0.00000008, -0.00000079, -0.00000276, -0.00000585, -0.00000624, -0.00000063, 0.00000426, 0.00000126, -0.00000147, 0.00000519, 0.00001150, 0.00000557, -0.00000621, -0.00001054, -0.00000249, 0.00001091, 0.00001337, 0.00000019, -0.00001056, -0.00000966, -0.00000918, -0.00000829, 0.00000102, 0.00000269, -0.00000833, -0.00000534, 0.00000858, 0.00000246, -0.00000822, 0.00000520, 0.00001556, -0.00000015, -0.00001500, -0.00001535, -0.00001064, 0.00000142, 0.00000936, 0.00000303, 0.00000454, 0.00000218, -0.00002768, -0.00002371, 0.00002753, 0.00001291, -0.00004026, 0.00001102, 0.00005376, -0.00004785, -0.00006401, 0.00007709, 0.00003178, -0.00012939, 0.00000612, 0.00013509, -0.00013724, -0.00018773, 0.00022681, 0.00004480, -0.00071020, -0.00033490, 0.00105439, 0.00114184, -0.00024774, -0.00076930, 0.00000312, 0.00027497, -0.00018772, -0.00017605, 0.00016131, 0.00002926, -0.00016759, 0.00002593, 0.00011034, -0.00007994, -0.00007288, 0.00008190, 0.00002300, -0.00007014, 0.00001607, 0.00005114, -0.00004112, -0.00004379, 0.00002536, 0.00001242, -0.00001547, 0.00001317, 0.00001297, -0.00002417, -0.00002194, 0.00000089, 0.00000270, 0.00000543, 0.00000882, -0.00000417, -0.00000694, 0.00000738, 0.00000566, -0.00000754, -0.00000537, 0.00000160, -0.00000157, -0.00000730, -0.00001342, -0.00001609, -0.00000164, 0.00001861, 0.00001725, -0.00000141, -0.00001262, -0.00000956, 0.00000085, 0.00001018, 0.00000940, 0.00000155, -0.00000068, 0.00000157, -0.00000156, -0.00000612, -0.00000398, 0.00000001, -0.00000042, -0.00000188, -0.00000069, 0.00000143, 0.00000246, 0.00000219, 0.00000126, 0.00000056, },
		{ 0.00000088, -0.00000005, -0.00000040, -0.00000078, -0.00000284, -0.00000617, -0.00000533, 0.00000091, 0.00000347, -0.00000046, 0.00000034, 0.00000850, 0.00001064, 0.00000139, -0.00000906, -0.00001015, 0.00000121, 0.00001502, 0.00001248, -0.00000450, -0.00001283, -0.00000960, -0.00000864, -0.00000524, 0.00000276, -0.00000125, -0.00000920, 0.00000070, 0.00000891, -0.00000329, -0.00000619, 0.00001012, 0.00001186, -0.00000362, -0.00001328, -0.00001708, -0.00001048, 0.00000776, 0.00000810, -0.00000167, 0.00000982, 0.00000028, -0.00003885, -0.00001383, 0.00004033, -0.00000554, -0.00004571, 0.00003866, 0.00004393, -0.00008100, -0.00003430, 0.00010410, -0.00002170, -0.00013774, 0.00007638, 0.00010200, -0.00021865, -0.00009655, 0.00029311, -0.00016172, -0.00079473, 0.00002271, 0.00129788, 0.00083173, -0.00056125, -0.00062292, 0.00018557, 0.00016961, -0.00025753, -0.00007281, 0.00017227, -0.00005161, -0.00013891, 0.00008086, 0.00006581, -0.00010473, -0.00002614, 0.00008335, -0.00001313, -0.00005486, 0.00003768, 0.00002823, -0.00005029, -0.00002377, 0.00002447, 0.00000130, -0.00000685, 0.00001664, 0.00000175, -0.00002423, -0.00001537, -0.00000160, 0.00000305, 0.00001026, 0.00000475, -0.00000792, -0.00000072, 0.00000899, -0.00000113, -0.00000815, -0.00000079, 0.00000080, -0.00000490, -0.00000848, -0.00001392, -0.00001343, 0.00000433, 0.00001976, 0.00001173, -0.00000591, -0.00001237, -0.00000637, 0.00000475, 0.00001124, 0.00000657, -0.00000043, 0.00000042, 0.00000209, -0.00000275, -0.00000639, -0.00000344, -0.00000028, -0.00000071, -0.00000131, -0.00000008, 0.00000169, 0.00000263, 0.00000214, },
		{ 0.00000270, 0.00000195, 0.00000043, -0.00000068, -0.00000069, -0.00000058, -0.00000306, -0.00000639, -0.00000410, 0.00000188, 0.00000201, -0.00000113, 0.00000328, 0.00001067, 0.00000820, -0.00000270, -0.00001111, -0.00000874, 0.00000608, 0.00001832, 0.00000935, -0.00000937, -0.00001389, -0.00000917, -0.00000730, -0.00000188, 0.00000220, -0.00000550, -0.00000667, 0.00000625, 0.00000541, -0.00000743, -0.00000117, 0.00001212, 0.00000684, -0.00000387, -0.00001270, -0.00002082, -0.00000657, 0.00001393, 0.00000220, -0.00000317, 0.00001771, -0.00000819, -0.00004813, 0.00000450, 0.00004531, -0.00003043, -0.00003725, 0.00006605, 0.00001604, -0.00010336, 0.00001255, 0.00010761, -0.00008437, -0.00011142, 0.00013883, 0.00002814, -0.00026560, 0.00003281, 0.00028466, -0.00040171, -0.00074962, 0.00043499, 0.00138495, 0.00043499, -0.00074962, -0.00040171, 0.00028466, 0.00003281, -0.00026560, 0.00002814, 0.00013883, -0.00011142, -0.00008437, 0.00010761, 0.00001255, -0.00010336, 0.00001604, 0.00006605, -0.00003725, -0.00003043, 0.00004531, 0.00000450, -0.00004813, -0.00000819, 0.00001771, -0.00000317, 0.00000220, 0.00001393, -0.00000657, -0.00002082, -0.00001270, -0.00000387, 0.00000684, 0.00001212, -0.00000117, -0.00000743, 0.00000541, 0.00000625, -0.00000667, -0.00000550, 0.00000220, -0.00000188, -0.00000730, -0.00000917, -0.00001389, -0.00000937, 0.00000935, 0.00001832, 0.00000608, -0.00000874, -0.00001111, -0.00000270, 0.00000820, 0.00001067, 0.00000328, -0.00000113, 0.00000201, 0.00000188, -0.00000410, -0.00000639, -0.00000306, -0.00000058, -0.00000069, -0.00000068, 0.00000043, 0.00000195, },
		{ 0.00000088, 0.00000214, 0.00000263, 0.00000169, -0.00000008, -0.00000131, -0.00000071, -0.00000028, -0.00000344, -0.00000639, -0.00000275, 0.00000209, 0.00000042, -0.00000043, 0.00000657, 0.00001124, 0.00000475, -0.00000637, -0.00001237, -0.00000591, 0.00001173, 0.00001976, 0.00000433, -0.00001343, -0.00001392, -0.00000848, -0.00000490, 0.00000080, -0.00000079, -0.00000815, -0.00000113, 0.00000899, -0.00000072, -0.00000792, 0.00000475, 0.00001026, 0.00000305, -0.00000160, -0.00001537, -0.00002423, 0.00000175, 0.00001664, -0.00000685, 0.00000130, 0.00002447, -0.00002377, -0.00005029, 0.00002823, 0.00003768, -0.00005486, -0.00001313, 0.00008335, -0.00002614, -0.00010473, 0.00006581, 0.00008086, -0.00013891, -0.00005161, 0.00017227, -0.00007281, -0.00025753, 0.00016961, 0.00018557, -0.00062292, -0.00056125, 0.00083173, 0.00129788, 0.00002271, -0.00079473, -0.00016172, 0.00029311, -0.00009655, -0.00021865, 0.00010200, 0.00007638, -0.00013774, -0.00002170, 0.00010410, -0.00003430, -0.00008100, 0.00004393, 0.00003866, -0.00004571, -0.00000554, 0.00004033, -0.00001383, -0.00003885, 0.00000028, 0.00000982, -0.00000167, 0.00000810, 0.00000776, -0.00001048, -0.00001708, -0.00001328, -0.00000362, 0.00001186, 0.00001012, -0.00000619, -0.00000329, 0.00000891, 0.00000070, -0.00000920, -0.00000125, 0.00000276, -0.00000524, -0.00000864, -0.00000960, -0.00001283, -0.00000450, 0.00001248, 0.00001502, 0.00000121, -0.00001015, -0.00000906, 0.00000139, 0.00001064, 0.00000850, 0.00000034, -0.00000046, 0.00000347, 0.00000091, -0.00000533, -0.00000617, -0.00000284, -0.00000078, -0.00000040, -0.00000005, },
		{ 0.00000008, 0.00000056, 0.00000126, 0.00000219, 0.00000246, 0.00000143, -0.00000069, -0.00000188, -0.00000042, 0.00000001, -0.00000398, -0.00000612, -0.00000156, 0.00000157, -0.00000068, 0.00000155, 0.00000940, 0.00001018, 0.00000085, -0.00000956, -0.00001262, -0.00000141, 0.00001725, 0.00001861, -0.00000164, -0.00001609, -0.00001342, -0.00000730, -0.00000157, 0.00000160, -0.00000537, -0.00000754, 0.00000566, 0.00000738, -0.00000694, -0.00000417, 0.00000882, 0.00000543, 0.00000270, 0.00000089, -0.00002194, -0.00002417, 0.00001297, 0.00001317, -0.00001547, 0.00001242, 0.00002536, -0.00004379, -0.00004112, 0.00005114, 0.00001607, -0.00007014, 0.00002300, 0.00008190, -0.00007288, -0.00007994, 0.00011034, 0.00002593, -0.00016759, 0.00002926, 0.00016131, -0.00017605, -0.00018772, 0.00027497, 0.00000312, -0.00076930, -0.00024774, 0.00114184, 0.00105439, -0.00033490, -0.00071020, 0.00004480, 0.00022681, -0.00018773, -0.00013724, 0.00013509, 0.00000612, -0.00012939, 0.00003178, 0.00007709, -0.00006401, -0.00004785, 0.00005376, 0.00001102, -0.00004026, 0.00001291, 0.00002753, -0.00002371, -0.00002768, 0.00000218, 0.00000454, 0.00000303, 0.00000936, 0.00000142, -0.00001064, -0.00001535, -0.00001500, -0.00000015, 0.00001556, 0.00000520, -0.00000822, 0.00000246, 0.00000858, -0.00000534, -0.00000833, 0.00000269, 0.00000102, -0.00000829, -0.00000918, -0.00000966, -0.00001056, 0.00000019, 0.00001337, 0.00001091, -0.00000249, -0.00001054, -0.00000621, 0.00000557, 0.00001150, 0.00000519, -0.00000147, 0.00000126, 0.00000426, -0.00000063, -0.00000624, -0.00000585, -0.00000276, -0.00000079, },
		{ -0.00000272, -0.00000058, 0.00000067, 0.00000115, 0.00000153, 0.00000206, 0.00000228, 0.00000116, -0.00000143, -0.00000227, 0.00000018, 0.00000015, -0.00000461, -0.00000561, -0.00000076, 0.00000063, -0.00000082, 0.00000435, 0.00001111, 0.00000781, -0.00000310, -0.00001221, -0.00001140, 0.00000461, 0.00002138, 0.00001483, -0.00000733, -0.00001735, -0.00001276, -0.00000512, 0.00000199, -0.00000036, -0.00000977, -0.00000292, 0.00001104, 0.00000145, -0.00001034, 0.00000258, 0.00000875, 0.00000013, 0.00000653, 0.00000048, -0.00003099, -0.00001808, 0.00002355, 0.00000279, -0.00001909, 0.00002793, 0.00001654, -0.00006272, -0.00001955, 0.00006551, -0.00001611, -0.00006884, 0.00006239, 0.00005770, -0.00011095, -0.00003167, 0.00013123, -0.00004555, -0.00015832, 0.00011015, 0.00010246, -0.00025238, -0.00006770, 0.00031290, -0.00023109, -0.00079586, 0.00014320, 0.00131045, 0.00070301, -0.00058465, -0.00053502, 0.00018188, 0.00011720, -0.00022540, -0.00004781, 0.00012733, -0.00005213, -0.00009518, 0.00006441, 0.00003875, -0.00007285, -0.00001518, 0.00004787, -0.00000915, -0.00002644, 0.00002162, 0.00001287, -0.00002564, -0.00001885, 0.00000023, 0.00000349, 0.00000760, 0.00000669, -0.00000264, -0.00000911, -0.00001635, -0.00001551, 0.00000554, 0.00001621, -0.00000057, -0.00000660, 0.00000738, 0.00000476, -0.00000972, -0.00000491, 0.00000495, -0.00000221, -0.00001049, -0.00000915, -0.00000905, -0.00000737, 0.00000375, 0.00001239, 0.00000692, -0.00000512, -0.00001008, -0.00000253, 0.00000925, 0.00001045, 0.00000158, -0.00000171, 0.00000337, 0.00000409, -0.00000240, -0.00000676, -0.00000556, },
		{ -0.00000695, -0.00000535, -0.00000262, -0.00000015, 0.00000134, 0.00000169, 0.00000159, 0.00000179, 0.00000219, 0.00000083, -0.00000231, -0.00000234, 0.00000104, 0.00000002, -0.00000524, -0.00000499, -0.00000045, -0.00000027, 0.00000019, 0.00000727, 0.00001141, 0.00000459, -0.00000694, -0.00001409, -0.00000821, 0.00001140, 0.00002297, 0.00000916, -0.00001176, -0.00001773, -0.00001193, -0.00000157, 0.00000450, -0.00000513, -0.00001173, 0.00000489, 0.00001236, -0.00000692, -0.00000879, 0.00000948, 0.00000391, -0.00000241, 0.00001314, -0.00000532, -0.00003913, -0.00000534, 0.00002907, -0.00001239, -0.00001400, 0.00004287, -0.00000301, -0.00007369, 0.00001126, 0.00006481, -0.00005099, -0.00004773, 0.00009346, 0.00001363, -0.00012778, 0.00002935, 0.00011899, -0.00011515, -0.00010949, 0.00016738, 0.00000681, -0.00027729, 0.00007408, 0.00026182, -0.00046723, -0.00068162, 0.00054459, 0.00131186, 0.00031070, -0.00070123, -0.00032124, 0.00023637, 0.00000043, -0.00021140, 0.00002599, 0.00008999, -0.00008568, -0.00004962, 0.00007278, 0.00000198, -0.00006402, 0.00000846, 0.00003275, -0.00001837, -0.00001099, 0.00002127, 0.00000118, -0.00002260, -0.00001432, -0.00000213, 0.00000598, 0.00000958, 0.00000230, -0.00000364, -0.00000821, -0.00001914, -0.00001330, 0.00001142, 0.00001350, -0.00000491, -0.00000225, 0.00000956, -0.00000098, -0.00001119, -0.00000047, 0.00000495, -0.00000595, -0.00001169, -0.00000862, -0.00000756, -0.00000398, 0.00000566, 0.00001033, 0.00000354, -0.00000695, -0.00000869, 0.00000186, 0.00001167, 0.00000766, -0.00000136, -0.00000045, 0.00000515, 0.00000300, -0.00000402, },
		{ 0.00000128, -0.00000527, -0.00000696, -0.00000523, -0.00000238, 0.00000045, 0.00000207, 0.00000212, 0.00000138, 0.00000149, 0.00000224, 0.00000032, -0.00000324, -0.00000197, 0.00000199, -0.00000044, -0.00000582, -0.00000439, -0.00000055, -0.00000067, 0.00000217, 0.00000963, 0.00001037, 0.00000090, -0.00001061, -0.00001469, -0.00000288, 0.00001765, 0.00002148, 0.00000293, -0.00001454, -0.00001794, -0.00001038, 0.00000321, 0.00000445, -0.00001152, -0.00000935, 0.00001354, 0.00000809, -0.00001448, -0.00000201, 0.00001309, -0.00000416, 0.00000037, 0.00001924, -0.00001712, -0.00004213, 0.00001195, 0.00002590, -0.00002766, 0.00000096, 0.00005101, -0.00003046, -0.00007093, 0.00004418, 0.00004634, -0.00007833, -0.00000967, 0.00010546, -0.00004055, -0.00011588, 0.00008697, 0.00007348, -0.00016303, -0.00003166, 0.00018231, -0.00010208, -0.00023944, 0.00019921, 0.00012269, -0.00064931, -0.00043662, 0.00088591, 0.00115488, -0.00005374, -0.00069000, -0.00011961, 0.00021742, -0.00009330, -0.00016110, 0.00006974, 0.00004037, -0.00009159, -0.00000741, 0.00006129, -0.00002378, -0.00004523, 0.00001943, 0.00001609, -0.00001756, 0.00000068, 0.00001531, -0.00000523, -0.00001834, -0.00001357, -0.00000239, 0.00000979, 0.00000821, -0.00000141, -0.00000245, -0.00000944, -0.00002185, -0.00000827, 0.00001545, 0.00000857, -0.00000635, 0.00000293, 0.00000829, -0.00000661, -0.00000967, 0.00000340, 0.00000289, -0.00000942, -0.00001198, -0.00000747, -0.00000533, -0.00000123, 0.00000601, 0.00000804, 0.00000077, -0.00000817, -0.00000615, 0.00000643, 0.00001216, 0.00000381, -0.00000285, 0.00000181, 0.00000604, },
		{ 0.00000428, 0.00000581, -0.00000062, -0.00000609, -0.00000696, -0.00000512, -0.00000195, 0.00000119, 0.00000283, 0.00000233, 0.00000089, 0.00000131, 0.00000240, -0.00000048, -0.00000406, -0.00000109, 0.00000283, -0.00000127, -0.00000630, -0.00000393, -0.00000080, -0.00000026, 0.00000464, 0.00001098, 0.00000828, -0.00000305, -0.00001392, -0.00001333, 0.00000415, 0.00002188, 0.00001721, -0.00000246, -0.00001606, -0.00001836, -0.00000732, 0.00000825, 0.00000080, -0.00001727, -0.00000200, 0.00001976, -0.00000134, -0.00001772, 0.00000806, 0.00001085, -0.00001205, 0.00000910, 0.00002075, -0.00003283, -0.00003660, 0.00002918, 0.00001291, -0.00003711, 0.00002329, 0.00004707, -0.00005939, -0.00005211, 0.00007004, 0.00001279, -0.00008881, 0.00003654, 0.00009233, -0.00009072, -0.00007597, 0.00012449, 0.00000479, -0.00017441, 0.00005489, 0.00014778, -0.00019479, -0.00014560, 0.00027025, -0.00007914, -0.00072996, -0.00010058, 0.00110997, 0.00087911, -0.00033553, -0.00058162, 0.00003262, 0.00014990, -0.00014690, -0.00009622, 0.00008079, -0.00000467, -0.00007571, 0.00002120, 0.00003917, -0.00003486, -0.00002527, 0.00001912, 0.00000392, -0.00001079, 0.00000606, 0.00000806, -0.00000682, -0.00001576, -0.00001446, 0.00000026, 0.00001248, 0.00000444, -0.00000284, -0.00000087, -0.00001289, -0.00002265, -0.00000161, 0.00001645, 0.00000333, -0.00000470, 0.00000688, 0.00000421, -0.00001039, -0.00000607, 0.00000561, -0.00000056, -0.00001212, -0.00001137, -0.00000557, -0.00000293, 0.00000030, 0.00000539, 0.00000601, -0.00000164, -0.00000861, -0.00000236, 0.00001034, 0.00001049, -0.00000006, -0.00000255, },
		{ -0.00000288, -0.00000070, 0.00000616, 0.00000462, -0.00000232, -0.00000659, -0.00000701, -0.00000492, -0.00000134, 0.00000204, 0.00000357, 0.00000221, 0.00000022, 0.00000140, 0.00000252, -0.00000160, -0.00000454, 0.00000025, 0.00000335, -0.00000240, -0.00000669, -0.00000362, -0.00000084, 0.00000101, 0.00000701, 0.00001117, 0.00000547, -0.00000723, -0.00001640, -0.00000947, 0.00001176, 0.00002292, 0.00001136, -0.00000609, -0.00001722, -0.00001873, -0.00000216, 0.00001188, -0.00000643, -0.00001964, 0.00000911, 0.00002043, -0.00001334, -0.00001430, 0.00001769, 0.00000224, -0.00001562, 0.00002175, 0.00001441, -0.00004806, -0.00002156, 0.00004064, -0.00000760, -0.00003572, 0.00004721, 0.00002896, -0.00008157, -0.00001979, 0.00008056, -0.00002794, -0.00007737, 0.00007852, 0.00005526, -0.00012272, -0.00001749, 0.00013014, -0.00006923, -0.00014470, 0.00012611, 0.00007128, -0.00024505, -0.00001991, 0.00026225, -0.00029929, -0.00068337, 0.00026651, 0.00118579, 0.00054353, -0.00050469, -0.00042083, 0.00011831, 0.00006450, -0.00015836, -0.00003713, 0.00006660, -0.00003391, -0.00004897, 0.00003271, 0.00001646, -0.00003330, -0.00001078, 0.00001231, -0.00000113, -0.00000300, 0.00000574, 0.00000290, -0.00000586, -0.00001591, -0.00001456, 0.00000486, 0.00001249, 0.00000017, -0.00000182, -0.00000067, -0.00001742, -0.00002058, 0.00000482, 0.00001453, -0.00000039, -0.00000102, 0.00000822, -0.00000117, -0.00001149, -0.00000175, 0.00000573, -0.00000461, -0.00001381, -0.00000982, -0.00000309, -0.00000108, 0.00000059, 0.00000458, 0.00000429, -0.00000392, -0.00000788, 0.00000239, 0.00001263, 0.00000702, },
		{ 0.00001263, 0.00000269, -0.00000391, 0.00000197, 0.00000694, 0.00000289, -0.00000359, -0.00000697, -0.00000710, -0.00000454, -0.00000057, 0.00000301, 0.00000418, 0.00000166, -0.00000044, 0.00000183, 0.00000239, -0.00000296, -0.00000446, 0.00000189, 0.00000335, -0.00000377, -0.00000698, -0.00000335, -0.00000037, 0.00000289, 0.00000876, 0.00001033, 0.00000214, -0.00001155, -0.00001728, -0.00000306, 0.00001829, 0.00002046, 0.00000562, -0.00000792, -0.00001895, -0.00001811, 0.00000494, 0.00001217, -0.00001574, -0.00001641, 0.00002104, 0.00001393, -0.00002387, -0.00000412, 0.00002265, -0.00001057, -0.00001169, 0.00003404, -0.00000062, -0.00005735, 0.00000064, 0.00004152, -0.00003013, -0.00002139, 0.00006520, -0.00000102, -0.00008958, 0.00001885, 0.00007134, -0.00006488, -0.00004531, 0.00010412, 0.00000284, -0.00012703, 0.00004394, 0.00010124, -0.00012837, -0.00008159, 0.00016174, -0.00002676, -0.00023839, 0.00010257, 0.00017098, -0.00048572, -0.00051250, 0.00059920, 0.00111433, 0.00021115, -0.00055935, -0.00025316, 0.00014053, -0.00001197, -0.00013785, 0.00000309, 0.00004007, -0.00004403, -0.00002272, 0.00003003, 0.00000044, -0.00002476, -0.00000448, 0.00000465, 0.00000011, 0.00000203, 0.00000237, 0.00000114, -0.00000502, -0.00001798, -0.00001231, 0.00000948, 0.00000970, -0.00000273, 0.00000058, -0.00000276, -0.00002134, -0.00001591, 0.00000938, 0.00001085, -0.00000160, 0.00000298, 0.00000665, -0.00000612, -0.00001002, 0.00000198, 0.00000393, -0.00000858, -0.00001432, -0.00000728, -0.00000053, -0.00000038, 0.00000015, 0.00000410, 0.00000250, -0.00000594, -0.00000556, 0.00000733, },
		{ -0.00000155, 0.00001135, 0.00001024, -0.00000129, -0.00000300, 0.00000454, 0.00000653, 0.00000113, -0.00000448, -0.00000738, -0.00000713, -0.00000397, 0.00000032, 0.00000408, 0.00000450, 0.00000071, -0.00000085, 0.00000252, 0.00000183, -0.00000435, -0.00000368, 0.00000352, 0.00000272, -0.00000523, -0.00000716, -0.00000292, 0.00000076, 0.00000492, 0.00000962, 0.00000874, -0.00000169, -0.00001570, -0.00001569, 0.00000518, 0.00002203, 0.00001521, 0.00000154, -0.00000878, -0.00002164, -0.00001527, 0.00001278, 0.00000776, -0.00002436, -0.00000687, 0.00002979, 0.00000092, -0.00002871, 0.00001029, 0.00001981, -0.00002328, 0.00000053, 0.00004078, -0.00002200, -0.00005621, 0.00002468, 0.00002981, -0.00004760, 0.00000383, 0.00007050, -0.00003633, -0.00007953, 0.00005393, 0.00004364, -0.00008741, -0.00000038, 0.00010544, -0.00005126, -0.00010213, 0.00009129, 0.00004542, -0.00015648, -0.00000289, 0.00015195, -0.00011911, -0.00017720, 0.00018616, 0.00001531, -0.00059263, -0.00024843, 0.00084236, 0.00092558, -0.00006588, -0.00052131, -0.00011366, 0.00011694, -0.00006312, -0.00010162, 0.00002088, 0.00001410, -0.00003886, -0.00000481, 0.00001991, -0.00000620, -0.00001567, -0.00000515, 0.00000033, 0.00000451, 0.00000284, -0.00000095, 0.00000204, -0.00000609, -0.00002011, -0.00000760, 0.00001222, 0.00000526, -0.00000327, 0.00000273, -0.00000686, -0.00002318, -0.00000988, 0.00001131, 0.00000709, -0.00000042, 0.00000568, 0.00000293, -0.00000933, -0.00000685, 0.00000424, 0.00000073, -0.00001196, -0.00001345, -0.00000394, 0.00000133, -0.00000093, -0.00000023, 0.00000396, 0.00000028, -0.00000717, },
		{ -0.00000245, -0.00000687, 0.00000362, 0.00001332, 0.00000609, -0.00000378, -0.00000063, 0.00000622, 0.00000526, -0.00000029, -0.00000516, -0.00000786, -0.00000697, -0.00000323, 0.00000135, 0.00000518, 0.00000435, -0.00000052, -0.00000081, 0.00000326, 0.00000074, -0.00000542, -0.00000220, 0.00000482, 0.00000149, -0.00000666, -0.00000717, -0.00000209, 0.00000248, 0.00000660, 0.00000955, 0.00000661, -0.00000606, -0.00001900, -0.00001106, 0.00001375, 0.00002186, 0.00000885, -0.00000007, -0.00001004, -0.00002475, -0.00000928, 0.00001923, -0.00000140, -0.00002901, 0.00000754, 0.00003171, -0.00001545, -0.00002505, 0.00002436, 0.00000855, -0.00003078, 0.00001872, 0.00003776, -0.00004456, -0.00004287, 0.00004373, 0.00000746, -0.00005363, 0.00003397, 0.00005956, -0.00006801, -0.00005265, 0.00007594, 0.00000440, -0.00008876, 0.00004544, 0.00008161, -0.00009256, -0.00005526, 0.00011159, -0.00002156, -0.00014670, 0.00006898, 0.00010066, -0.00018030, -0.00008047, 0.00020544, -0.00016754, -0.00059276, 0.00005804, 0.00096347, 0.00066845, -0.00025650, -0.00042620, -0.00002066, 0.00007130, -0.00008499, -0.00006535, 0.00002089, -0.00000274, -0.00002628, 0.00000226, 0.00000960, -0.00000517, -0.00001057, -0.00000908, 0.00000083, 0.00000852, 0.00000033, -0.00000207, 0.00000365, -0.00000933, -0.00002046, -0.00000166, 0.00001214, 0.00000088, -0.00000167, 0.00000326, -0.00001178, -0.00002238, -0.00000409, 0.00001092, 0.00000461, 0.00000209, 0.00000614, -0.00000160, -0.00001025, -0.00000310, 0.00000470, -0.00000331, -0.00001430, -0.00001108, -0.00000035, 0.00000181, -0.00000221, 0.00000009, 0.00000369, },
		{ 0.00000123, 0.00000260, -0.00000519, -0.00000448, 0.00000888, 0.00001261, 0.00000137, -0.00000415, 0.00000226, 0.00000660, 0.00000367, -0.00000126, -0.00000588, -0.00000834, -0.00000656, -0.00000237, 0.00000254, 0.00000616, 0.00000362, -0.00000175, -0.00000023, 0.00000374, -0.00000078, -0.00000587, -0.00000025, 0.00000549, -0.00000022, -0.00000794, -0.00000691, -0.00000067, 0.00000449, 0.00000752, 0.00000878, 0.00000400, -0.00001092, -0.00002045, -0.00000343, 0.00002067, 0.00001765, 0.00000345, 0.00000052, -0.00001297, -0.00002692, -0.00000016, 0.00002188, -0.00001369, -0.00002697, 0.00002339, 0.00002481, -0.00003037, -0.00001274, 0.00003297, -0.00000869, -0.00002902, 0.00003791, 0.00002345, -0.00006184, -0.00001937, 0.00005173, -0.00001990, -0.00004486, 0.00006082, 0.00003354, -0.00008747, -0.00001528, 0.00007886, -0.00003576, -0.00006838, 0.00007967, 0.00003927, -0.00011042, -0.00000016, 0.00010012, -0.00008087, -0.00010370, 0.00011421, 0.00002424, -0.00019484, 0.00002190, 0.00015329, -0.00033213, -0.00048439, 0.00034970, 0.00095819, 0.00039700, -0.00035402, -0.00031176, 0.00002470, 0.00002504, -0.00008367, -0.00003927, 0.00001214, -0.00000804, -0.00001430, 0.00000086, 0.00000401, -0.00000082, -0.00001089, -0.00001224, 0.00000500, 0.00000969, -0.00000326, -0.00000057, 0.00000397, -0.00001362, -0.00001812, 0.00000373, 0.00000947, -0.00000194, 0.00000098, 0.00000164, -0.00001605, -0.00001937, 0.00000025, 0.00000929, 0.00000397, 0.00000452, 0.00000438, -0.00000558, -0.00000910, 0.00000021, 0.00000338, -0.00000758, -0.00001514, -0.00000732, 0.00000257, 0.00000067, -0.00000331, },
		{ -0.00000166, -0.00000331, 0.00000265, 0.00000028, -0.00000699, -0.00000001, 0.00001282, 0.00000933, -0.00000250, -0.00000250, 0.00000464, 0.00000581, 0.00000232, -0.00000192, -0.00000677, -0.00000866, -0.00000591, -0.00000141, 0.00000390, 0.00000681, 0.00000233, -0.00000266, 0.00000079, 0.00000369, -0.00000245, -0.00000549, 0.00000180, 0.00000537, -0.00000219, -0.00000898, -0.00000622, 0.00000139, 0.00000632, 0.00000752, 0.00000759, 0.00000082, -0.00001588, -0.00001906, 0.00000625, 0.00002402, 0.00001055, 0.00000078, 0.00000194, -0.00001805, -0.00002635, 0.00001078, 0.00001887, -0.00002598, -0.00001725, 0.00003612, 0.00000974, -0.00003888, 0.00000541, 0.00003227, -0.00002694, -0.00001665, 0.00005193, -0.00000005, -0.00006817, 0.00000880, 0.00004552, -0.00004452, -0.00002231, 0.00007651, -0.00000175, -0.00008929, 0.00002299, 0.00006223, -0.00006590, -0.00003221, 0.00009335, -0.00000959, -0.00010145, 0.00004730, 0.00006188, -0.00011665, -0.00004195, 0.00012190, -0.00005393, -0.00016214, 0.00009845, 0.00004314, -0.00043752, -0.00029097, 0.00057724, 0.00084805, 0.00015718, -0.00037255, -0.00020795, 0.00003377, -0.00000833, -0.00007033, -0.00002624, 0.00000356, -0.00000475, -0.00000808, -0.00000394, 0.00000441, 0.00000232, -0.00001510, -0.00001195, 0.00001019, 0.00000754, -0.00000560, 0.00000247, 0.00000196, -0.00001722, -0.00001342, 0.00000705, 0.00000544, -0.00000259, 0.00000328, -0.00000170, -0.00001860, -0.00001525, 0.00000270, 0.00000770, 0.00000487, 0.00000572, 0.00000115, -0.00000807, -0.00000659, 0.00000238, 0.00000057, -0.00001147, -0.00001408, -0.00000272, 0.00000389, },
		{ 0.00000170, 0.00000303, -0.00000411, -0.00000178, 0.00000338, -0.00000310, -0.00000678, 0.00000576, 0.00001422, 0.00000441, -0.00000438, 0.00000040, 0.00000576, 0.00000437, 0.00000151, -0.00000258, -0.00000780, -0.00000866, -0.00000510, -0.00000032, 0.00000537, 0.00000691, 0.00000065, -0.00000297, 0.00000196, 0.00000298, -0.00000387, -0.00000430, 0.00000351, 0.00000448, -0.00000419, -0.00000972, -0.00000492, 0.00000395, 0.00000748, 0.00000670, 0.00000618, -0.00000310, -0.00002015, -0.00001413, 0.00001619, 0.00002262, 0.00000265, 0.00000165, 0.00000230, -0.00002462, -0.00002162, 0.00002110, 0.00000979, -0.00003448, -0.00000117, 0.00004142, -0.00001011, -0.00003755, 0.00002425, 0.00002117, -0.00004025, 0.00000419, 0.00005553, -0.00002739, -0.00006074, 0.00003423, 0.00002614, -0.00005901, 0.00000866, 0.00007608, -0.00003754, -0.00007308, 0.00005230, 0.00003132, -0.00007826, 0.00000926, 0.00008390, -0.00005150, -0.00007047, 0.00007437, 0.00000968, -0.00012106, 0.00001926, 0.00009334, -0.00011129, -0.00009650, 0.00012612, -0.00009493, -0.00045835, -0.00005383, 0.00071027, 0.00067168, -0.00002234, -0.00033863, -0.00013143, 0.00002336, -0.00002490, -0.00005579, -0.00002303, 0.00000065, 0.00000143, -0.00000861, -0.00000712, 0.00000884, 0.00000146, -0.00002007, -0.00000780, 0.00001372, 0.00000332, -0.00000543, 0.00000522, -0.00000212, -0.00001865, -0.00000768, 0.00000769, 0.00000154, -0.00000140, 0.00000421, -0.00000572, -0.00001913, -0.00001122, 0.00000366, 0.00000702, 0.00000635, 0.00000520, -0.00000244, -0.00000878, -0.00000357, 0.00000306, -0.00000333, -0.00001427, -0.00001095, },
		{ -0.00001524, -0.00000608, 0.00000467, 0.00000019, -0.00000541, 0.00000090, 0.00000248, -0.00000661, -0.00000394, 0.00001138, 0.00001252, -0.00000064, -0.00000383, 0.00000340, 0.00000540, 0.00000296, 0.00000120, -0.00000351, -0.00000881, -0.00000831, -0.00000420, 0.00000097, 0.00000677, 0.00000629, -0.00000108, -0.00000259, 0.00000289, 0.00000173, -0.00000463, -0.00000257, 0.00000453, 0.00000305, -0.00000607, -0.00001007, -0.00000286, 0.00000665, 0.00000759, 0.00000536, 0.00000460, -0.00000779, -0.00002264, -0.00000574, 0.00002413, 0.00001649, -0.00000352, 0.00000553, -0.00000013, -0.00003092, -0.00001238, 0.00002789, -0.00000388, -0.00003593, 0.00001766, 0.00003678, -0.00002948, -0.00002582, 0.00003794, 0.00000194, -0.00004367, 0.00002830, 0.00004625, -0.00005153, -0.00004073, 0.00004993, -0.00000132, -0.00005876, 0.00003995, 0.00005923, -0.00006483, -0.00004364, 0.00006565, -0.00000462, -0.00007064, 0.00004443, 0.00005572, -0.00007591, -0.00002861, 0.00007537, -0.00004030, -0.00009642, 0.00006230, 0.00004097, -0.00013347, -0.00002210, 0.00009729, -0.00022332, -0.00039107, 0.00017956, 0.00074229, 0.00047289, -0.00013305, -0.00028125, -0.00008518, 0.00000881, -0.00002841, -0.00004669, -0.00002341, 0.00000423, 0.00000513, -0.00001335, -0.00000580, 0.00001367, -0.00000363, -0.00002279, -0.00000133, 0.00001413, -0.00000090, -0.00000290, 0.00000614, -0.00000697, -0.00001738, -0.00000250, 0.00000603, -0.00000109, 0.00000065, 0.00000349, -0.00000931, -0.00001809, -0.00000807, 0.00000401, 0.00000744, 0.00000739, 0.00000319, -0.00000545, -0.00000795, -0.00000077, 0.00000212, -0.00000780, },
		{ -0.00000044, -0.00001207, -0.00001375, -0.00000043, 0.00000514, -0.00000362, -0.00000460, 0.00000363, -0.00000043, -0.00000886, 0.00000136, 0.00001517, 0.00000812, -0.00000425, -0.00000123, 0.00000538, 0.00000394, 0.00000215, 0.00000109, -0.00000482, -0.00000957, -0.00000766, -0.00000323, 0.00000250, 0.00000783, 0.00000497, -0.00000245, -0.00000167, 0.00000324, 0.00000029, -0.00000452, -0.00000080, 0.00000472, 0.00000139, -0.00000774, -0.00000992, 0.00000003, 0.00000899, 0.00000650, 0.00000390, 0.00000267, -0.00001292, -0.00002221, 0.00000509, 0.00002794, 0.00000699, -0.00000588, 0.00001065, -0.00000615, -0.00003465, 0.00000026, 0.00002873, -0.00001897, -0.00002880, 0.00003435, 0.00002249, -0.00004278, -0.00000637, 0.00004184, -0.00002027, -0.00003503, 0.00004902, 0.00002543, -0.00006594, -0.00001314, 0.00005161, -0.00002923, -0.00004350, 0.00006327, 0.00003040, -0.00007724, -0.00000936, 0.00006102, -0.00003540, -0.00004696, 0.00006436, 0.00001848, -0.00007846, 0.00001060, 0.00005314, -0.00007400, -0.00005386, 0.00007694, -0.00001634, -0.00011871, 0.00003520, 0.00002193, -0.00030790, -0.00025359, 0.00036800, 0.00068867, 0.00028929, -0.00018367, -0.00022363, -0.00006228, -0.00000047, -0.00002618, -0.00004427, -0.00002170, 0.00001137, 0.00000356, -0.00001819, -0.00000005, 0.00001563, -0.00001100, -0.00002168, 0.00000505, 0.00001159, -0.00000334, 0.00000069, 0.00000462, -0.00001102, -0.00001391, 0.00000091, 0.00000316, -0.00000204, 0.00000251, 0.00000153, -0.00001179, -0.00001632, -0.00000590, 0.00000456, 0.00000849, 0.00000732, 0.00000037, -0.00000732, -0.00000607, 0.00000125, },
		{ -0.00000367, 0.00000207, -0.00000438, -0.00001514, -0.00000967, 0.00000452, 0.00000276, -0.00000682, -0.00000157, 0.00000497, -0.00000484, -0.00000853, 0.00000799, 0.00001585, 0.00000231, -0.00000533, 0.00000231, 0.00000568, 0.00000212, 0.00000214, 0.00000075, -0.00000639, -0.00000990, -0.00000684, -0.00000213, 0.00000421, 0.00000826, 0.00000319, -0.00000315, -0.00000057, 0.00000290, -0.00000083, -0.00000361, 0.00000050, 0.00000421, -0.00000022, -0.00000924, -0.00000909, 0.00000363, 0.00001039, 0.00000435, 0.00000266, 0.00000014, -0.00001779, -0.00001807, 0.00001647, 0.00002628, -0.00000339, -0.00000347, 0.00001458, -0.00001522, -0.00003376, 0.00001381, 0.00002262, -0.00003144, -0.00001401, 0.00004414, 0.00000174, -0.00004595, 0.00001556, 0.00003414, -0.00003904, -0.00001581, 0.00006030, -0.00000203, -0.00006677, 0.00001475, 0.00003912, -0.00004985, -0.00001737, 0.00007276, -0.00000270, -0.00007291, 0.00002054, 0.00004180, -0.00005320, -0.00001559, 0.00006550, -0.00001639, -0.00006197, 0.00003590, 0.00001752, -0.00008398, -0.00000891, 0.00006318, -0.00006014, -0.00007786, 0.00005646, -0.00007601, -0.00032761, -0.00007890, 0.00048596, 0.00057910, 0.00014463, -0.00019303, -0.00017877, -0.00005169, -0.00000212, -0.00002491, -0.00004555, -0.00001511, 0.00001763, -0.00000282, -0.00001964, 0.00000777, 0.00001317, -0.00001784, -0.00001707, 0.00000927, 0.00000754, -0.00000324, 0.00000370, 0.00000115, -0.00001308, -0.00000945, 0.00000214, 0.00000023, -0.00000163, 0.00000354, -0.00000089, -0.00001311, -0.00001456, -0.00000428, 0.00000569, 0.00000944, 0.00000606, -0.00000248, -0.00000794, },
		{ -0.00000480, -0.00000748, -0.00000123, 0.00000135, -0.00000918, -0.00001597, -0.00000360, 0.00000719, -0.00000182, -0.00000784, 0.00000277, 0.00000380, -0.00000933, -0.00000498, 0.00001411, 0.00001302, -0.00000310, -0.00000367, 0.00000542, 0.00000433, 0.00000076, 0.00000271, -0.00000011, -0.00000796, -0.00000978, -0.00000594, -0.00000079, 0.00000588, 0.00000789, 0.00000136, -0.00000309, 0.00000023, 0.00000204, -0.00000117, -0.00000233, 0.00000097, 0.00000339, -0.00000166, -0.00001061, -0.00000731, 0.00000761, 0.00001032, 0.00000154, 0.00000180, -0.00000314, -0.00002135, -0.00001016, 0.00002598, 0.00001908, -0.00001179, 0.00000318, 0.00001504, -0.00002544, -0.00002730, 0.00002507, 0.00001053, -0.00003759, 0.00000516, 0.00004393, -0.00002019, -0.00003777, 0.00003376, 0.00001655, -0.00004877, 0.00000919, 0.00005871, -0.00002920, -0.00005408, 0.00003564, 0.00001644, -0.00005773, 0.00001242, 0.00006676, -0.00003157, -0.00005490, 0.00003897, 0.00001532, -0.00005493, 0.00001365, 0.00005041, -0.00003958, -0.00003488, 0.00004179, -0.00001855, -0.00007132, 0.00002405, 0.00003040, -0.00007875, -0.00002988, 0.00003566, -0.00016648, -0.00027960, 0.00009511, 0.00052755, 0.00044752, 0.00004656, -0.00018198, -0.00014932, -0.00004365, 0.00000092, -0.00002798, -0.00004596, -0.00000441, 0.00001947, -0.00001114, -0.00001625, 0.00001447, 0.00000690, -0.00002183, -0.00001067, 0.00001046, 0.00000384, -0.00000108, 0.00000488, -0.00000302, -0.00001282, -0.00000530, 0.00000154, -0.00000194, -0.00000056, 0.00000365, -0.00000311, -0.00001367, -0.00001311, -0.00000263, 0.00000723, 0.00000967, 0.00000397, },
		{ 0.00000897, 0.00000160, -0.00000639, -0.00000618, 0.00000078, -0.00000115, -0.00001388, -0.00001378, 0.00000306, 0.00000649, -0.00000698, -0.00000581, 0.00000673, -0.00000010, -0.00001212, 0.00000137, 0.00001783, 0.00000741, -0.00000643, -0.00000003, 0.00000696, 0.00000198, 0.00000040, 0.00000339, -0.00000158, -0.00000923, -0.00000932, -0.00000498, 0.00000086, 0.00000718, 0.00000675, -0.00000005, -0.00000252, 0.00000036, 0.00000118, -0.00000056, -0.00000133, 0.00000058, 0.00000271, -0.00000302, -0.00001180, -0.00000437, 0.00001139, 0.00000849, -0.00000135, 0.00000128, -0.00000711, -0.00002247, 0.00000066, 0.00003132, 0.00000776, -0.00001571, 0.00001213, 0.00001069, -0.00003418, -0.00001587, 0.00003115, -0.00000475, -0.00003531, 0.00002391, 0.00003339, -0.00003751, -0.00002039, 0.00004304, -0.00000619, -0.00004642, 0.00003341, 0.00004459, -0.00004926, -0.00003182, 0.00004450, -0.00000964, -0.00005140, 0.00003804, 0.00004812, -0.00004968, -0.00002981, 0.00004301, -0.00000983, -0.00004276, 0.00003286, 0.00002632, -0.00004676, -0.00000790, 0.00002990, -0.00004372, -0.00004440, 0.00003672, -0.00000678, -0.00007085, 0.00000511, -0.00001882, -0.00022273, -0.00017850, 0.00023599, 0.00050451, 0.00032274, -0.00001066, -0.00016684, -0.00013079, -0.00003289, 0.00000368, -0.00003479, -0.00004200, 0.00000720, 0.00001572, -0.00001783, -0.00000897, 0.00001756, -0.00000107, -0.00002205, -0.00000463, 0.00000913, 0.00000188, 0.00000188, 0.00000391, -0.00000656, -0.00001072, -0.00000236, -0.00000009, -0.00000314, 0.00000056, 0.00000317, -0.00000490, -0.00001399, -0.00001180, -0.00000058, 0.00000865, },
		{ 0.00000177, 0.00000946, 0.00000757, -0.00000063, -0.00000724, -0.00000426, 0.00000184, -0.00000534, -0.00001718, -0.00000853, 0.00000839, 0.00000237, -0.00001069, -0.00000102, 0.00000851, -0.00000581, -0.00001167, 0.00000898, 0.00001784, 0.00000070, -0.00000671, 0.00000420, 0.00000643, -0.00000041, 0.00000110, 0.00000365, -0.00000343, -0.00001003, -0.00000869, -0.00000384, 0.00000274, 0.00000776, 0.00000515, -0.00000070, -0.00000191, -0.00000024, 0.00000094, 0.00000076, -0.00000125, -0.00000030, 0.00000252, -0.00000458, -0.00001257, -0.00000015, 0.00001422, 0.00000492, -0.00000366, 0.00000085, -0.00001135, -0.00002027, 0.00001262, 0.00003092, -0.00000514, -0.00001382, 0.00002059, 0.00000169, -0.00003883, -0.00000160, 0.00003037, -0.00001934, -0.00002480, 0.00003736, 0.00001510, -0.00004567, 0.00000139, 0.00004084, -0.00002779, -0.00003256, 0.00005053, 0.00002184, -0.00005766, -0.00000647, 0.00004009, -0.00003178, -0.00003359, 0.00005344, 0.00002297, -0.00005428, -0.00000543, 0.00003444, -0.00002672, -0.00002289, 0.00003843, 0.00000221, -0.00003955, 0.00000998, 0.00000759, -0.00005200, -0.00001520, 0.00002909, -0.00003431, -0.00004516, 0.00001306, -0.00008757, -0.00022916, -0.00005076, 0.00032441, 0.00043972, 0.00022244, -0.00004031, -0.00015615, -0.00011615, -0.00001917, 0.00000228, -0.00004211, -0.00003278, 0.00001597, 0.00000783, -0.00002035, -0.00000036, 0.00001627, -0.00000822, -0.00001918, -0.00000056, 0.00000667, 0.00000205, 0.00000426, 0.00000131, -0.00000860, -0.00000773, -0.00000090, -0.00000201, -0.00000358, 0.00000147, 0.00000254, -0.00000641, -0.00001433, -0.00001021, },
		{ -0.00001453, -0.00000803, 0.00000398, 0.00000943, 0.00000589, -0.00000256, -0.00000738, -0.00000198, 0.00000134, -0.00001066, -0.00001776, -0.00000115, 0.00001058, -0.00000398, -0.00001120, 0.00000505, 0.00000689, -0.00001144, -0.00000743, 0.00001566, 0.00001398, -0.00000506, -0.00000402, 0.00000746, 0.00000410, -0.00000192, 0.00000247, 0.00000317, -0.00000532, -0.00001036, -0.00000800, -0.00000235, 0.00000457, 0.00000743, 0.00000362, -0.00000056, -0.00000186, -0.00000122, 0.00000184, 0.00000212, -0.00000243, -0.00000098, 0.00000284, -0.00000666, -0.00001246, 0.00000514, 0.00001525, 0.00000004, -0.00000489, 0.00000018, -0.00001513, -0.00001444, 0.00002343, 0.00002445, -0.00001650, -0.00000639, 0.00002576, -0.00001034, -0.00003772, 0.00001243, 0.00002292, -0.00002942, -0.00000865, 0.00004203, -0.00000616, -0.00004283, 0.00002155, 0.00002803, -0.00004241, -0.00001106, 0.00005648, -0.00000336, -0.00005343, 0.00001508, 0.00002505, -0.00004440, -0.00001005, 0.00005611, -0.00000147, -0.00004691, 0.00001188, 0.00001857, -0.00003216, -0.00000295, 0.00003180, -0.00001453, -0.00002416, 0.00001439, -0.00001530, -0.00004424, 0.00000547, 0.00000856, -0.00004385, -0.00001656, -0.00000936, -0.00014703, -0.00018510, 0.00007401, 0.00035684, 0.00035876, 0.00015182, -0.00005670, -0.00015099, -0.00009982, -0.00000560, -0.00000427, -0.00004622, -0.00002010, 0.00001949, -0.00000120, -0.00001813, 0.00000679, 0.00001158, -0.00001277, -0.00001486, 0.00000108, 0.00000463, 0.00000381, 0.00000513, -0.00000191, -0.00000893, -0.00000476, -0.00000068, -0.00000378, -0.00000359, 0.00000226, 0.00000192, -0.00000797, },
		{ 0.00000116, -0.00000978, -0.00001421, -0.00000535, 0.00000556, 0.00000877, 0.00000427, -0.00000420, -0.00000675, 0.00000020, -0.00000117, -0.00001589, -0.00001477, 0.00000651, 0.00000861, -0.00001040, -0.00000782, 0.00001020, 0.00000192, -0.00001481, -0.00000014, 0.00001934, 0.00000736, -0.00000811, 0.00000054, 0.00000863, 0.00000091, -0.00000207, 0.00000388, 0.00000195, -0.00000690, -0.00001040, -0.00000719, -0.00000042, 0.00000592, 0.00000627, 0.00000273, -0.00000002, -0.00000276, -0.00000186, 0.00000400, 0.00000258, -0.00000470, -0.00000072, 0.00000326, -0.00000937, -0.00001087, 0.00001093, 0.00001382, -0.00000535, -0.00000474, -0.00000102, -0.00001754, -0.00000548, 0.00003075, 0.00001304, -0.00002351, 0.00000480, 0.00002570, -0.00002274, -0.00003069, 0.00002308, 0.00001081, -0.00003243, 0.00000895, 0.00003691, -0.00002487, -0.00003029, 0.00003476, 0.00000849, -0.00004645, 0.00001220, 0.00005059, -0.00002460, -0.00003925, 0.00002772, 0.00000487, -0.00004527, 0.00001249, 0.00004750, -0.00001928, -0.00003234, 0.00001908, 0.00000205, -0.00002726, 0.00001081, 0.00001817, -0.00002045, -0.00000858, 0.00000659, -0.00003035, -0.00002700, 0.00001177, -0.00001375, -0.00003544, -0.00000006, -0.00005402, -0.00017725, -0.00010371, 0.00017150, 0.00034343, 0.00028250, 0.00010613, -0.00007024, -0.00014769, -0.00007986, 0.00000401, -0.00001409, -0.00004487, -0.00000730, 0.00001753, -0.00000832, -0.00001249, 0.00001062, 0.00000549, -0.00001425, -0.00001084, 0.00000090, 0.00000397, 0.00000608, 0.00000435, -0.00000476, -0.00000793, -0.00000236, -0.00000127, -0.00000532, -0.00000333, 0.00000315, },
		{ -0.00000269, 0.00000422, -0.00000009, -0.00001164, -0.00001300, -0.00000262, 0.00000625, 0.00000790, 0.00000276, -0.00000558, -0.00000531, 0.00000160, -0.00000570, -0.00001944, -0.00000833, 0.00001214, 0.00000282, -0.00001445, -0.00000132, 0.00001230, -0.00000501, -0.00001429, 0.00000823, 0.00001877, 0.00000001, -0.00000767, 0.00000526, 0.00000739, -0.00000199, -0.00000097, 0.00000472, 0.00000030, -0.00000804, -0.00001037, -0.00000605, 0.00000183, 0.00000635, 0.00000473, 0.00000282, 0.00000024, -0.00000459, -0.00000130, 0.00000694, 0.00000131, -0.00000732, 0.00000094, 0.00000309, -0.00001245, -0.00000730, 0.00001623, 0.00000964, -0.00001020, -0.00000328, -0.00000285, -0.00001771, 0.00000529, 0.00003282, -0.00000100, -0.00002438, 0.00001694, 0.00001988, -0.00003262, -0.00001914, 0.00002815, -0.00000269, -0.00002783, 0.00002347, 0.00002373, -0.00003646, -0.00001202, 0.00003798, -0.00001216, -0.00003955, 0.00003129, 0.00003556, -0.00003721, -0.00002033, 0.00002949, -0.00001417, -0.00003585, 0.00002851, 0.00003207, -0.00002758, -0.00001657, 0.00001696, -0.00000966, -0.00001641, 0.00001549, 0.00000417, -0.00001672, 0.00000064, -0.00000778, -0.00003351, -0.00000926, 0.00000449, -0.00002771, -0.00001684, -0.00000506, -0.00010463, -0.00016803, -0.00000669, 0.00022850, 0.00030229, 0.00022277, 0.00007563, -0.00008538, -0.00014155, -0.00005782, 0.00000739, -0.00002375, -0.00003819, 0.00000248, 0.00001189, -0.00001165, -0.00000582, 0.00001080, 0.00000004, -0.00001338, -0.00000824, 0.00000005, 0.00000486, 0.00000785, 0.00000237, -0.00000670, -0.00000618, -0.00000067, -0.00000236, -0.00000672, },
		{ -0.00000388, -0.00000795, -0.00000143, 0.00000520, -0.00000207, -0.00001307, -0.00001089, -0.00000044, 0.00000625, 0.00000720, 0.00000120, -0.00000653, -0.00000315, 0.00000136, -0.00001154, -0.00001977, 0.00000025, 0.00001381, -0.00000507, -0.00001436, 0.00000624, 0.00001022, -0.00001154, -0.00000952, 0.00001518, 0.00001410, -0.00000578, -0.00000416, 0.00000847, 0.00000434, -0.00000364, 0.00000081, 0.00000471, -0.00000135, -0.00000885, -0.00001034, -0.00000432, 0.00000401, 0.00000564, 0.00000349, 0.00000385, -0.00000054, -0.00000679, 0.00000112, 0.00000961, -0.00000207, -0.00000915, 0.00000391, 0.00000154, -0.00001518, -0.00000159, 0.00001979, 0.00000299, -0.00001343, -0.00000091, -0.00000515, -0.00001511, 0.00001598, 0.00002894, -0.00001465, -0.00001884, 0.00002696, 0.00000933, -0.00003772, -0.00000563, 0.00002698, -0.00001408, -0.00001725, 0.00003139, 0.00000636, -0.00003863, 0.00000668, 0.00003120, -0.00002845, -0.00002445, 0.00004200, 0.00001636, -0.00003957, -0.00000253, 0.00002190, -0.00002703, -0.00002037, 0.00003536, 0.00001541, -0.00002692, -0.00000464, 0.00000919, -0.00001395, -0.00000515, 0.00001206, -0.00000489, -0.00000801, 0.00000067, -0.00002142, -0.00002607, 0.00000131, -0.00001010, -0.00002823, 0.00000013, -0.00003199, -0.00014276, -0.00012139, 0.00008298, 0.00024449, 0.00025258, 0.00018183, 0.00005066, -0.00010134, -0.00012965, -0.00003721, 0.00000478, -0.00003011, -0.00002835, 0.00000755, 0.00000531, -0.00001111, -0.00000036, 0.00000837, -0.00000351, -0.00001153, -0.00000724, -0.00000036, 0.00000677, 0.00000854, -0.00000012, -0.00000764, -0.00000416, 0.00000039, },
		{ -0.00000207, 0.00000090, -0.00000589, -0.00000875, 0.00000056, 0.00000560, -0.00000465, -0.00001352, -0.00000832, 0.00000081, 0.00000604, 0.00000671, -0.00000061, -0.00000670, -0.00000079, -0.00000116, -0.00001724, -0.00001603, 0.00000867, 0.00001068, -0.00001237, -0.00000974, 0.00001220, 0.00000436, -0.00001522, -0.00000164, 0.00001852, 0.00000685, -0.00000831, 0.00000094, 0.00000915, 0.00000066, -0.00000368, 0.00000250, 0.00000398, -0.00000266, -0.00000962, -0.00001013, -0.00000187, 0.00000552, 0.00000393, 0.00000328, 0.00000524, -0.00000286, -0.00000831, 0.00000546, 0.00001069, -0.00000719, -0.00000903, 0.00000743, -0.00000184, -0.00001648, 0.00000581, 0.00002038, -0.00000520, -0.00001419, 0.00000173, -0.00000749, -0.00000974, 0.00002451, 0.00001968, -0.00002496, -0.00000816, 0.00003238, -0.00000371, -0.00003704, 0.00000694, 0.00002060, -0.00002068, -0.00000390, 0.00003126, -0.00001045, -0.00003187, 0.00002085, 0.00001727, -0.00003664, -0.00000589, 0.00004298, -0.00000158, -0.00003321, 0.00000954, 0.00000898, -0.00003130, -0.00000410, 0.00003356, 0.00000228, -0.00002055, 0.00000092, 0.00000053, -0.00001154, 0.00000204, 0.00000438, -0.00000678, -0.00000008, -0.00000704, -0.00002855, -0.00001346, 0.00000133, -0.00002347, -0.00001697, 0.00000457, -0.00007238, -0.00015413, -0.00004992, 0.00014777, 0.00022913, 0.00020882, 0.00015490, 0.00002512, -0.00011451, -0.00011206, -0.00002127, -0.00000144, -0.00003162, -0.00001831, 0.00000813, 0.00000022, -0.00000802, 0.00000263, 0.00000496, -0.00000500, -0.00000993, -0.00000726, 0.00000023, 0.00000894, 0.00000813, -0.00000258, -0.00000778, },
		{ -0.00000477, -0.00000733, 0.00000008, 0.00000072, -0.00000835, -0.00000868, 0.00000302, 0.00000497, -0.00000731, -0.00001277, -0.00000599, 0.00000120, 0.00000614, 0.00000613, -0.00000259, -0.00000577, 0.00000088, -0.00000598, -0.00002090, -0.00000859, 0.00001436, 0.00000354, -0.00001640, -0.00000190, 0.00001428, -0.00000338, -0.00001441, 0.00000699, 0.00001721, -0.00000059, -0.00000705, 0.00000569, 0.00000726, -0.00000233, -0.00000243, 0.00000347, 0.00000297, -0.00000362, -0.00001059, -0.00000938, 0.00000114, 0.00000578, 0.00000184, 0.00000453, 0.00000603, -0.00000666, -0.00000795, 0.00001102, 0.00000899, -0.00001290, -0.00000623, 0.00001027, -0.00000689, -0.00001527, 0.00001378, 0.00001718, -0.00001347, -0.00001211, 0.00000389, -0.00000927, -0.00000224, 0.00002910, 0.00000681, -0.00002975, 0.00000520, 0.00003195, -0.00001643, -0.00003104, 0.00001611, 0.00001131, -0.00002142, 0.00000848, 0.00002394, -0.00002256, -0.00001912, 0.00002737, 0.00000072, -0.00003575, 0.00001108, 0.00003584, -0.00001406, -0.00002196, 0.00001381, -0.00000435, -0.00002764, 0.00000850, 0.00002618, -0.00000487, -0.00001279, 0.00000053, -0.00000512, -0.00000557, 0.00000332, -0.00000286, -0.00000278, 0.00000265, -0.00001790, -0.00002699, -0.00000243, -0.00000743, -0.00002868, -0.00000123, -0.00000904, -0.00011216, -0.00013315, 0.00002814, 0.00017971, 0.00019728, 0.00017808, 0.00013435, -0.00000245, -0.00012138, -0.00009135, -0.00001155, -0.00000811, -0.00002879, -0.00001055, 0.00000593, -0.00000216, -0.00000430, 0.00000313, 0.00000202, -0.00000511, -0.00000921, -0.00000748, 0.00000180, 0.00001077, 0.00000694, },
		{ 0.00001202, 0.00000527, -0.00000672, -0.00000633, 0.00000223, -0.00000040, -0.00001092, -0.00000740, 0.00000537, 0.00000318, -0.00000931, -0.00001110, -0.00000452, 0.00000124, 0.00000676, 0.00000503, -0.00000434, -0.00000379, 0.00000077, -0.00001221, -0.00002084, 0.00000083, 0.00001533, -0.00000532, -0.00001548, 0.00000653, 0.00001155, -0.00001021, -0.00000905, 0.00001356, 0.00001176, -0.00000582, -0.00000281, 0.00000836, 0.00000371, -0.00000373, -0.00000068, 0.00000357, 0.00000219, -0.00000449, -0.00001179, -0.00000766, 0.00000416, 0.00000447, 0.00000028, 0.00000713, 0.00000517, -0.00001112, -0.00000480, 0.00001636, 0.00000398, -0.00001749, -0.00000081, 0.00001102, -0.00001270, -0.00001084, 0.00002065, 0.00001010, -0.00002009, -0.00000748, 0.00000496, -0.00000986, 0.00000623, 0.00002869, -0.00000712, -0.00002813, 0.00001827, 0.00002586, -0.00002631, -0.00002138, 0.00002061, 0.00000188, -0.00001697, 0.00001679, 0.00001223, -0.00002760, -0.00000462, 0.00002565, -0.00001378, -0.00002745, 0.00002265, 0.00002419, -0.00001930, -0.00001035, 0.00001101, -0.00001402, -0.00001898, 0.00001515, 0.00001724, -0.00000625, -0.00000721, -0.00000326, -0.00000609, 0.00000006, -0.00000030, -0.00000613, 0.00000343, -0.00000116, -0.00002655, -0.00001866, 0.00000189, -0.00001934, -0.00002331, 0.00000943, -0.00003870, -0.00013689, -0.00008419, 0.00009492, 0.00018089, 0.00016337, 0.00016010, 0.00011351, -0.00002987, -0.00012036, -0.00007101, -0.00000742, -0.00001273, -0.00002356, -0.00000610, 0.00000309, -0.00000205, -0.00000148, 0.00000206, 0.00000023, -0.00000478, -0.00000930, -0.00000725, 0.00000392, },
		{ -0.00000642, 0.00000618, 0.00001277, 0.00000325, -0.00000844, -0.00000463, 0.00000412, -0.00000262, -0.00001298, -0.00000494, 0.00000688, 0.00000062, -0.00001014, -0.00000922, -0.00000407, 0.00000165, 0.00000760, 0.00000314, -0.00000519, -0.00000138, -0.00000190, -0.00001811, -0.00001632, 0.00000952, 0.00001108, -0.00001279, -0.00000968, 0.00001252, 0.00000488, -0.00001348, -0.00000076, 0.00001591, 0.00000413, -0.00000724, 0.00000254, 0.00000812, -0.00000004, -0.00000338, 0.00000068, 0.00000312, 0.00000192, -0.00000574, -0.00001291, -0.00000473, 0.00000639, 0.00000177, 0.00000023, 0.00001033, 0.00000196, -0.00001486, 0.00000131, 0.00001964, -0.00000389, -0.00001924, 0.00000617, 0.00000869, -0.00001772, -0.00000327, 0.00002458, 0.00000001, -0.00002348, -0.00000117, 0.00000462, -0.00000878, 0.00001413, 0.00002323, -0.00001933, -0.00002064, 0.00002832, 0.00001560, -0.00003168, -0.00001040, 0.00002049, -0.00000533, -0.00000943, 0.00001945, -0.00000020, -0.00002550, 0.00000748, 0.00001754, -0.00002288, -0.00001522, 0.00002728, 0.00001223, -0.00001807, -0.00000206, 0.00000396, -0.00001801, -0.00000917, 0.00001602, 0.00001016, -0.00000409, -0.00000549, -0.00000712, -0.00000313, 0.00000243, -0.00000581, -0.00000440, 0.00000769, -0.00000969, -0.00002926, -0.00000824, -0.00000189, -0.00002783, -0.00001036, 0.00000730, -0.00007564, -0.00013676, -0.00001943, 0.00013819, 0.00016076, 0.00013731, 0.00014991, 0.00008903, -0.00005356, -0.00011240, -0.00005388, -0.00000677, -0.00001433, -0.00001816, -0.00000450, 0.00000123, -0.00000056, -0.00000023, 0.00000057, -0.00000049, -0.00000469, -0.00000973, },
		{ -0.00000500, -0.00001005, -0.00000515, 0.00000841, 0.00001314, 0.00000078, -0.00000978, -0.00000215, 0.00000524, -0.00000579, -0.00001388, -0.00000179, 0.00000708, -0.00000192, -0.00000976, -0.00000795, -0.00000423, 0.00000289, 0.00000801, 0.00000071, -0.00000465, 0.00000031, -0.00000706, -0.00002153, -0.00000805, 0.00001471, 0.00000293, -0.00001609, -0.00000090, 0.00001382, -0.00000327, -0.00001171, 0.00000767, 0.00001326, -0.00000296, -0.00000469, 0.00000684, 0.00000527, -0.00000258, -0.00000187, 0.00000107, 0.00000275, 0.00000202, -0.00000772, -0.00001333, -0.00000076, 0.00000699, -0.00000156, 0.00000230, 0.00001283, -0.00000359, -0.00001623, 0.00000954, 0.00001921, -0.00001311, -0.00001696, 0.00001292, 0.00000314, -0.00002021, 0.00000637, 0.00002405, -0.00001136, -0.00002265, 0.00000548, 0.00000299, -0.00000589, 0.00001992, 0.00001371, -0.00002750, -0.00000904, 0.00003349, 0.00000347, -0.00003208, -0.00000032, 0.00001694, -0.00000902, -0.00000147, 0.00001673, -0.00001002, -0.00001822, 0.00001440, 0.00000641, -0.00002535, -0.00000298, 0.00002579, 0.00000326, -0.00001301, 0.00000119, -0.00000374, -0.00001662, -0.00000145, 0.00001322, 0.00000665, -0.00000135, -0.00000719, -0.00000856, 0.00000134, 0.00000068, -0.00000986, 0.00000081, 0.00000726, -0.00001906, -0.00002533, -0.00000085, -0.00001120, -0.00002837, 0.00000329, -0.00001019, -0.00010765, -0.00010972, 0.00004564, 0.00015393, 0.00013176, 0.00012272, 0.00014124, 0.00006119, -0.00007063, -0.00010013, -0.00004109, -0.00000720, -0.00001344, -0.00001417, -0.00000441, 0.00000090, 0.00000101, -0.00000029, -0.00000058, -0.00000068, },
		{ -0.00000119, -0.00000084, -0.00000546, -0.00001011, -0.00000370, 0.00001070, 0.00001309, -0.00000224, -0.00001037, 0.00000095, 0.00000511, -0.00000932, -0.00001327, 0.00000123, 0.00000604, -0.00000366, -0.00000874, -0.00000777, -0.00000415, 0.00000490, 0.00000737, -0.00000154, -0.00000277, -0.00000001, -0.00001359, -0.00002074, 0.00000181, 0.00001455, -0.00000627, -0.00001380, 0.00000769, 0.00000986, -0.00000974, -0.00000527, 0.00001324, 0.00000660, -0.00000702, 0.00000056, 0.00000839, 0.00000110, -0.00000310, -0.00000024, 0.00000052, 0.00000308, 0.00000198, -0.00001046, -0.00001237, 0.00000356, 0.00000546, -0.00000433, 0.00000642, 0.00001324, -0.00001054, -0.00001395, 0.00001810, 0.00001421, -0.00002144, -0.00001054, 0.00001738, -0.00000470, -0.00001878, 0.00001613, 0.00001841, -0.00002170, -0.00001751, 0.00001104, 0.00000056, -0.00000150, 0.00002238, 0.00000194, -0.00003021, 0.00000415, 0.00003316, -0.00000810, -0.00002818, 0.00000727, 0.00001178, -0.00000919, 0.00000456, 0.00001040, -0.00001519, -0.00000887, 0.00001535, -0.00000417, -0.00002211, 0.00000627, 0.00002065, -0.00000123, -0.00000739, -0.00000018, -0.00000917, -0.00001192, 0.00000257, 0.00000954, 0.00000643, -0.00000035, -0.00001032, -0.00000675, 0.00000455, -0.00000395, -0.00001029, 0.00000644, 0.00000188, -0.00002534, -0.00001714, 0.00000034, -0.00002102, -0.00002054, 0.00000992, -0.00003929, -0.00012352, -0.00006172, 0.00009724, 0.00014601, 0.00010485, 0.00011780, 0.00012933, 0.00003268, -0.00008010, -0.00008653, -0.00003215, -0.00000709, -0.00001147, -0.00001196, -0.00000444, 0.00000168, 0.00000187, -0.00000096, },
		{ 0.00000196, -0.00000163, -0.00000155, -0.00000111, -0.00000579, -0.00001008, -0.00000209, 0.00001320, 0.00001230, -0.00000569, -0.00000980, 0.00000415, 0.00000349, -0.00001240, -0.00001129, 0.00000333, 0.00000438, -0.00000421, -0.00000800, -0.00000851, -0.00000305, 0.00000700, 0.00000545, -0.00000271, -0.00000037, -0.00000316, -0.00001943, -0.00001521, 0.00001021, 0.00000898, -0.00001312, -0.00000656, 0.00001284, 0.00000207, -0.00001179, 0.00000372, 0.00001384, -0.00000162, -0.00000668, 0.00000627, 0.00000661, -0.00000265, -0.00000172, 0.00000048, -0.00000038, 0.00000437, 0.00000108, -0.00001347, -0.00000958, 0.00000720, 0.00000191, -0.00000527, 0.00001173, 0.00001054, -0.00001720, -0.00000761, 0.00002468, 0.00000500, -0.00002653, -0.00000113, 0.00001791, -0.00001296, -0.00001296, 0.00002361, 0.00000823, -0.00002869, -0.00000897, 0.00001434, -0.00000189, 0.00000368, 0.00002093, -0.00000983, -0.00002725, 0.00001632, 0.00002794, -0.00001715, -0.00002145, 0.00001170, 0.00000674, -0.00000690, 0.00000732, 0.00000302, -0.00001540, -0.00000059, 0.00001146, -0.00001150, -0.00001554, 0.00001118, 0.00001480, -0.00000167, -0.00000380, -0.00000416, -0.00001091, -0.00000660, 0.00000311, 0.00000723, 0.00000783, -0.00000197, -0.00001259, -0.00000268, 0.00000474, -0.00000899, -0.00000694, 0.00000947, -0.00000636, -0.00002628, -0.00000874, -0.00000468, -0.00002631, -0.00000813, 0.00000435, -0.00007138, -0.00011693, -0.00000425, 0.00012713, 0.00012342, 0.00008664, 0.00011773, 0.00011239, 0.00000675, -0.00008294, -0.00007375, -0.00002566, -0.00000601, -0.00000972, -0.00001095, -0.00000385, 0.00000274, },
		{ -0.00000271, 0.00000343, 0.00000168, -0.00000213, -0.00000200, -0.00000123, -0.00000592, -0.00001022, -0.00000009, 0.00001585, 0.00001038, -0.00000912, -0.00000787, 0.00000670, 0.00000065, -0.00001431, -0.00000865, 0.00000419, 0.00000300, -0.00000385, -0.00000834, -0.00000939, -0.00000059, 0.00000823, 0.00000273, -0.00000217, 0.00000113, -0.00000893, -0.00002224, -0.00000606, 0.00001428, -0.00000009, -0.00001485, 0.00000308, 0.00001244, -0.00000652, -0.00000821, 0.00001201, 0.00000911, -0.00000839, -0.00000218, 0.00001005, 0.00000220, -0.00000446, 0.00000064, -0.00000023, -0.00000074, 0.00000632, -0.00000126, -0.00001589, -0.00000506, 0.00000901, -0.00000282, -0.00000348, 0.00001670, 0.00000454, -0.00002152, 0.00000201, 0.00002712, -0.00000676, -0.00002665, 0.00000907, 0.00001386, -0.00001933, -0.00000354, 0.00002663, -0.00000471, -0.00003062, 0.00000125, 0.00001479, -0.00000356, 0.00000860, 0.00001576, -0.00001941, -0.00001955, 0.00002534, 0.00001934, -0.00002254, -0.00001360, 0.00001325, 0.00000297, -0.00000377, 0.00000680, -0.00000305, -0.00001192, 0.00000449, 0.00000504, -0.00001451, -0.00000841, 0.00001221, 0.00001055, 0.00000010, -0.00000335, -0.00000824, -0.00000930, -0.00000284, 0.00000173, 0.00000722, 0.00000880, -0.00000551, -0.00001248, 0.00000166, 0.00000177, -0.00001201, -0.00000148, 0.00000833, -0.00001420, -0.00002220, -0.00000390, -0.00001294, -0.00002445, 0.00000259, -0.00001378, -0.00009608, -0.00008835, 0.00004978, 0.00013403, 0.00009644, 0.00007865, 0.00011740, 0.00009134, -0.00001430, -0.00008118, -0.00006262, -0.00002031, -0.00000449, -0.00000880, -0.00001022, },
		{ -0.00000844, -0.00000916, -0.00000156, 0.00000367, 0.00000149, -0.00000273, -0.00000257, -0.00000089, -0.00000616, -0.00001057, 0.00000271, 0.00001824, 0.00000714, -0.00001185, -0.00000482, 0.00000801, -0.00000271, -0.00001476, -0.00000623, 0.00000411, 0.00000262, -0.00000346, -0.00000999, -0.00000929, 0.00000277, 0.00000782, 0.00000027, -0.00000007, 0.00000030, -0.00001587, -0.00002034, 0.00000406, 0.00001245, -0.00000931, -0.00001047, 0.00001148, 0.00000642, -0.00001239, 0.00000013, 0.00001643, 0.00000053, -0.00001117, 0.00000471, 0.00001026, -0.00000312, -0.00000367, 0.00000263, -0.00000217, 0.00000026, 0.00000811, -0.00000511, -0.00001672, 0.00000049, 0.00000822, -0.00000737, 0.00000117, 0.00001958, -0.00000390, -0.00002180, 0.00001298, 0.00002409, -0.00001844, -0.00002131, 0.00001745, 0.00000597, -0.00002174, 0.00000748, 0.00002390, -0.00001774, -0.00002688, 0.00001103, 0.00001253, -0.00000388, 0.00001217, 0.00000782, -0.00002516, -0.00000891, 0.00002996, 0.00000935, -0.00002407, -0.00000614, 0.00001273, 0.00000082, -0.00000127, 0.00000406, -0.00000634, -0.00000688, 0.00000568, -0.00000133, -0.00001373, -0.00000284, 0.00001099, 0.00000886, 0.00000186, -0.00000561, -0.00001048, -0.00000593, -0.00000156, 0.00000038, 0.00000903, 0.00000792, -0.00000934, -0.00000980, 0.00000427, -0.00000298, -0.00001176, 0.00000353, 0.00000348, -0.00001883, -0.00001561, -0.00000435, -0.00002011, -0.00001636, 0.00000566, -0.00003978, -0.00010502, -0.00004450, 0.00008996, 0.00012269, 0.00007327, 0.00007846, 0.00011340, 0.00006862, -0.00002976, -0.00007682, -0.00005293, -0.00001546, -0.00000330, },
		{ -0.00001129, -0.00000280, -0.00000801, -0.00000784, -0.00000090, 0.00000387, 0.00000147, -0.00000380, -0.00000290, 0.00000001, -0.00000699, -0.00001082, 0.00000650, 0.00001972, 0.00000279, -0.00001327, -0.00000132, 0.00000785, -0.00000570, -0.00001412, -0.00000471, 0.00000393, 0.00000331, -0.00000409, -0.00001235, -0.00000741, 0.00000597, 0.00000572, -0.00000080, 0.00000250, -0.00000368, -0.00002162, -0.00001351, 0.00001184, 0.00000518, -0.00001503, -0.00000129, 0.00001526, -0.00000315, -0.00001282, 0.00001051, 0.00001504, -0.00000895, -0.00000884, 0.00001134, 0.00000659, -0.00000731, -0.00000067, 0.00000303, -0.00000444, 0.00000292, 0.00000871, -0.00000995, -0.00001520, 0.00000583, 0.00000473, -0.00001015, 0.00000785, 0.00001897, -0.00001294, -0.00001722, 0.00002266, 0.00001569, -0.00002718, -0.00001154, 0.00002173, -0.00000382, -0.00001912, 0.00001740, 0.00001551, -0.00002797, -0.00001816, 0.00001840, 0.00000838, -0.00000270, 0.00001349, -0.00000138, -0.00002628, 0.00000249, 0.00003000, -0.00000014, -0.00002234, -0.00000005, 0.00001113, -0.00000008, -0.00000022, 0.00000068, -0.00000665, -0.00000242, 0.00000372, -0.00000579, -0.00001076, 0.00000028, 0.00000948, 0.00000930, 0.00000196, -0.00000904, -0.00001010, -0.00000268, -0.00000235, 0.00000046, 0.00001126, 0.00000493, -0.00001179, -0.00000567, 0.00000414, -0.00000747, -0.00000862, 0.00000599, -0.00000303, -0.00001909, -0.00000986, -0.00000920, -0.00002261, -0.00000610, -0.00000200, -0.00006562, -0.00009464, 0.00000431, 0.00011112, 0.00010112, 0.00005824, 0.00008190, 0.00010472, 0.00004670, -0.00004020, -0.00007113, -0.00004415, },
		{ -0.00006457, -0.00003608, -0.00000826, -0.00000270, -0.00000703, -0.00000681, -0.00000075, 0.00000453, 0.00000122, -0.00000544, -0.00000243, 0.00000111, -0.00000872, -0.00001031, 0.00001109, 0.00001965, -0.00000201, -0.00001313, 0.00000184, 0.00000660, -0.00000771, -0.00001320, -0.00000413, 0.00000460, 0.00000439, -0.00000643, -0.00001420, -0.00000370, 0.00000776, 0.00000270, 0.00000019, 0.00000388, -0.00001035, -0.00002371, -0.00000334, 0.00001443, -0.00000498, -0.00001466, 0.00000950, 0.00001259, -0.00001267, -0.00000696, 0.00001929, 0.00000786, -0.00001604, -0.00000205, 0.00001513, 0.00000016, -0.00000874, 0.00000326, 0.00000130, -0.00000580, 0.00000684, 0.00000724, -0.00001469, -0.00001119, 0.00000954, -0.00000073, -0.00000993, 0.00001497, 0.00001434, -0.00002028, -0.00000832, 0.00002847, 0.00000350, -0.00003071, 0.00000034, 0.00002071, -0.00001284, -0.00001181, 0.00002353, 0.00000305, -0.00003305, -0.00000631, 0.00002204, 0.00000358, -0.00000039, 0.00001211, -0.00001008, -0.00002294, 0.00001261, 0.00002617, -0.00000770, -0.00001840, 0.00000428, 0.00000921, -0.00000045, -0.00000067, -0.00000186, -0.00000483, 0.00000000, 0.00000021, -0.00000760, -0.00000744, 0.00000131, 0.00000910, 0.00001050, -0.00000014, -0.00001186, -0.00000761, -0.00000096, -0.00000390, 0.00000232, 0.00001242, 0.00000067, -0.00001195, -0.00000176, 0.00000154, -0.00000996, -0.00000420, 0.00000510, -0.00000860, -0.00001589, -0.00000750, -0.00001548, -0.00001926, 0.00000098, -0.00001920, -0.00008291, -0.00006713, 0.00004769, 0.00011377, 0.00007746, 0.00005169, 0.00008510, 0.00009223, 0.00002718, -0.00004669, },
		{ 0.00001059, -0.00005007, -0.00005719, -0.00002909, -0.00000654, -0.00000225, -0.00000564, -0.00000666, -0.00000056, 0.00000578, 0.00000012, -0.00000722, -0.00000077, 0.00000170, -0.00001120, -0.00000841, 0.00001582, 0.00001777, -0.00000645, -0.00001174, 0.00000409, 0.00000503, -0.00000873, -0.00001287, -0.00000384, 0.00000661, 0.00000467, -0.00001033, -0.00001423, 0.00000101, 0.00000735, 0.00000013, 0.00000290, 0.00000250, -0.00001791, -0.00002058, 0.00000708, 0.00001063, -0.00001421, -0.00000776, 0.00001784, 0.00000395, -0.00001834, 0.00000380, 0.00002315, -0.00000310, -0.00001816, 0.00000696, 0.00001447, -0.00000691, -0.00000683, 0.00000642, -0.00000219, -0.00000518, 0.00001096, 0.00000343, -0.00001801, -0.00000532, 0.00001051, -0.00000673, -0.00000628, 0.00002056, 0.00000632, -0.00002385, 0.00000306, 0.00002866, -0.00000975, -0.00002812, 0.00001138, 0.00001474, -0.00001856, -0.00000158, 0.00002400, -0.00001074, -0.00003180, 0.00000614, 0.00002160, -0.00000063, 0.00000230, 0.00000820, -0.00001670, -0.00001614, 0.00001991, 0.00001980, -0.00001255, -0.00001339, 0.00000692, 0.00000738, -0.00000096, -0.00000196, -0.00000276, -0.00000232, 0.00000004, -0.00000312, -0.00000720, -0.00000510, 0.00000143, 0.00001028, 0.00001095, -0.00000382, -0.00001284, -0.00000432, -0.00000115, -0.00000481, 0.00000530, 0.00001166, -0.00000345, -0.00001002, 0.00000048, -0.00000221, -0.00000981, -0.00000049, 0.00000167, -0.00001137, -0.00001157, -0.00000918, -0.00001973, -0.00001189, 0.00000062, -0.00004085, -0.00008585, -0.00002924, 0.00007825, 0.00010271, 0.00005762, 0.00005137, 0.00008575, 0.00007756, },
		{ 0.00008323, 0.00006214, -0.00000313, -0.00005064, -0.00004930, -0.00002382, -0.00000567, -0.00000083, -0.00000456, -0.00000749, 0.00000042, 0.00000718, -0.00000221, -0.00000833, 0.00000192, 0.00000104, -0.00001380, -0.00000485, 0.00001979, 0.00001433, -0.00000988, -0.00000975, 0.00000539, 0.00000392, -0.00000937, -0.00001353, -0.00000274, 0.00000965, 0.00000295, -0.00001470, -0.00001169, 0.00000522, 0.00000487, -0.00000064, 0.00000597, -0.00000235, -0.00002371, -0.00001240, 0.00001421, 0.00000151, -0.00001876, 0.00000368, 0.00002033, -0.00000790, -0.00001751, 0.00001618, 0.00002034, -0.00001448, -0.00001441, 0.00001517, 0.00000929, -0.00001229, -0.00000218, 0.00000736, -0.00000630, -0.00000212, 0.00001394, -0.00000227, -0.00001884, 0.00000116, 0.00000833, -0.00001151, 0.00000023, 0.00002288, -0.00000337, -0.00002244, 0.00001424, 0.00002293, -0.00002094, -0.00002010, 0.00001886, 0.00000560, -0.00001936, 0.00000883, 0.00001843, -0.00002267, -0.00002457, 0.00001662, 0.00001774, -0.00000334, 0.00000444, 0.00000252, -0.00002014, -0.00000747, 0.00002365, 0.00001246, -0.00001460, -0.00000830, 0.00000819, 0.00000563, -0.00000196, -0.00000316, -0.00000202, -0.00000051, -0.00000171, -0.00000510, -0.00000569, -0.00000416, 0.00000190, 0.00001246, 0.00000967, -0.00000782, -0.00001169, -0.00000163, -0.00000267, -0.00000423, 0.00000819, 0.00000906, -0.00000620, -0.00000705, 0.00000052, -0.00000549, -0.00000762, 0.00000112, -0.00000248, -0.00001101, -0.00000864, -0.00001347, -0.00001972, -0.00000440, -0.00000844, -0.00005995, -0.00007313, 0.00001025, 0.00009317, 0.00008457, 0.00004423, 0.00005416, },
		{ 0.00003715, 0.00005754, 0.00007793, 0.00004676, -0.00001404, -0.00004847, -0.00004170, -0.00002064, -0.00000461, 0.00000153, -0.00000471, -0.00000872, 0.00000268, 0.00000784, -0.00000558, -0.00000794, 0.00000494, -0.00000121, -0.00001558, 0.00000008, 0.00002226, 0.00001004, -0.00001211, -0.00000788, 0.00000627, 0.00000358, -0.00001060, -0.00001479, 0.00000010, 0.00001253, -0.00000124, -0.00001792, -0.00000683, 0.00000747, 0.00000144, 0.00000097, 0.00000748, -0.00000992, -0.00002523, -0.00000128, 0.00001534, -0.00000979, -0.00001635, 0.00001586, 0.00001552, -0.00001890, -0.00000971, 0.00002620, 0.00001122, -0.00002272, -0.00000582, 0.00001982, 0.00000101, -0.00001418, 0.00000360, 0.00000544, -0.00000947, 0.00000300, 0.00001456, -0.00000870, -0.00001669, 0.00000667, 0.00000348, -0.00001357, 0.00000811, 0.00002106, -0.00001242, -0.00001623, 0.00002244, 0.00001255, -0.00002745, -0.00000883, 0.00002111, -0.00000402, -0.00001509, 0.00001649, 0.00000804, -0.00002998, -0.00001311, 0.00002324, 0.00001182, -0.00000424, 0.00000521, -0.00000377, -0.00002002, 0.00000136, 0.00002391, 0.00000554, -0.00001430, -0.00000381, 0.00000844, 0.00000382, -0.00000329, -0.00000350, -0.00000037, -0.00000021, -0.00000408, -0.00000538, -0.00000431, -0.00000420, 0.00000353, 0.00001452, 0.00000655, -0.00001082, -0.00000904, -0.00000041, -0.00000443, -0.00000214, 0.00000987, 0.00000548, -0.00000706, -0.00000440, -0.00000119, -0.00000715, -0.00000478, 0.00000042, -0.00000547, -0.00000870, -0.00000860, -0.00001769, -0.00001550, -0.00000101, -0.00002388, -0.00007016, -0.00004802, 0.00004353, 0.00009402, 0.00006535, },
		{ 0.00008514, 0.00004889, 0.00003473, 0.00006015, 0.00007048, 0.00003179, -0.00002179, -0.00004389, -0.00003556, -0.00001918, -0.00000233, 0.00000405, -0.00000657, -0.00000929, 0.00000601, 0.00000690, -0.00000918, -0.00000570, 0.00000724, -0.00000476, -0.00001577, 0.00000560, 0.00002296, 0.00000566, -0.00001351, -0.00000642, 0.00000755, 0.00000362, -0.00001318, -0.00001555, 0.00000491, 0.00001370, -0.00000734, -0.00001849, -0.00000092, 0.00000701, -0.00000129, 0.00000443, 0.00000581, -0.00001805, -0.00002112, 0.00000933, 0.00000976, -0.00001909, -0.00000713, 0.00002444, 0.00000449, -0.00002504, 0.00000315, 0.00003050, -0.00000174, -0.00002527, 0.00000505, 0.00001931, -0.00000792, -0.00001201, 0.00000852, 0.00000106, -0.00001035, 0.00000902, 0.00001225, -0.00001438, -0.00001192, 0.00000981, -0.00000277, -0.00001213, 0.00001541, 0.00001535, -0.00001859, -0.00000671, 0.00002562, 0.00000009, -0.00002797, 0.00000270, 0.00001806, -0.00001136, -0.00000717, 0.00001921, -0.00000459, -0.00003106, -0.00000006, 0.00002517, 0.00000549, -0.00000364, 0.00000422, -0.00000936, -0.00001672, 0.00000885, 0.00002139, 0.00000003, -0.00001235, -0.00000028, 0.00000791, 0.00000180, -0.00000449, -0.00000267, 0.00000119, -0.00000146, -0.00000591, -0.00000441, -0.00000381, -0.00000432, 0.00000635, 0.00001538, 0.00000230, -0.00001203, -0.00000599, -0.00000075, -0.00000537, 0.00000073, 0.00000985, 0.00000208, -0.00000636, -0.00000302, -0.00000353, -0.00000696, -0.00000269, -0.00000167, -0.00000633, -0.00000640, -0.00001120, -0.00001941, -0.00000939, -0.00000427, -0.00004070, -0.00006809, -0.00001682, 0.00006586, },
		{ 0.00001352, 0.00007628, 0.00007148, 0.00003664, 0.00003517, 0.00006153, 0.00006114, 0.00001765, -0.00002584, -0.00003787, -0.00003186, -0.00001825, 0.00000166, 0.00000546, -0.00000984, -0.00000818, 0.00000951, 0.00000399, -0.00001182, -0.00000192, 0.00000793, -0.00000878, -0.00001406, 0.00001077, 0.00002217, 0.00000169, -0.00001469, -0.00000508, 0.00000980, 0.00000304, -0.00001718, -0.00001443, 0.00001089, 0.00001188, -0.00001380, -0.00001578, 0.00000419, 0.00000420, -0.00000185, 0.00000808, 0.00000044, -0.00002388, -0.00001201, 0.00001579, -0.00000086, -0.00002268, 0.00000625, 0.00002614, -0.00000945, -0.00002386, 0.00001739, 0.00002747, -0.00001491, -0.00002139, 0.00001493, 0.00001372, -0.00001492, -0.00000651, 0.00001089, -0.00000452, -0.00000821, 0.00001439, 0.00000724, -0.00001791, -0.00000563, 0.00000984, -0.00000869, -0.00000740, 0.00002026, 0.00000714, -0.00002041, 0.00000365, 0.00002314, -0.00001136, -0.00002283, 0.00001163, 0.00001114, -0.00001449, 0.00000188, 0.00001621, -0.00001629, -0.00002592, 0.00001174, 0.00002278, 0.00000019, -0.00000231, 0.00000159, -0.00001314, -0.00001121, 0.00001402, 0.00001722, -0.00000359, -0.00000956, 0.00000217, 0.00000676, -0.00000033, -0.00000501, -0.00000095, 0.00000178, -0.00000366, -0.00000652, -0.00000308, -0.00000430, -0.00000363, 0.00000971, 0.00001446, -0.00000199, -0.00001143, -0.00000360, -0.00000202, -0.00000499, 0.00000339, 0.00000834, -0.00000029, -0.00000500, -0.00000314, -0.00000535, -0.00000558, -0.00000209, -0.00000372, -0.00000531, -0.00000577, -0.00001477, -0.00001763, -0.00000485, -0.00001392, -0.00005319, -0.00005425, },
		{ -0.00005711, -0.00003247, 0.00003768, 0.00007688, 0.00005693, 0.00002834, 0.00003730, 0.00006149, 0.00004991, 0.00000524, -0.00002593, -0.00003206, -0.00003078, -0.00001631, 0.00000683, 0.00000465, -0.00001341, -0.00000500, 0.00001192, -0.00000043, -0.00001255, 0.00000240, 0.00000667, -0.00001222, -0.00001080, 0.00001492, 0.00002054, -0.00000192, -0.00001614, -0.00000304, 0.00001293, 0.00000066, -0.00002168, -0.00001047, 0.00001639, 0.00000679, -0.00001863, -0.00001034, 0.00000691, 0.00000049, 0.00000022, 0.00000984, -0.00000756, -0.00002493, -0.00000046, 0.00001569, -0.00001286, -0.00001870, 0.00001952, 0.00001995, -0.00002187, -0.00001525, 0.00002883, 0.00001782, -0.00002463, -0.00001233, 0.00002096, 0.00000464, -0.00001809, 0.00000058, 0.00000991, -0.00000953, -0.00000329, 0.00001763, 0.00000054, -0.00001849, 0.00000062, 0.00000688, -0.00001266, -0.00000055, 0.00002151, -0.00000151, -0.00001768, 0.00001217, 0.00001589, -0.00001914, -0.00001388, 0.00001604, 0.00000284, -0.00001290, 0.00000920, 0.00000838, -0.00002417, -0.00001606, 0.00002004, 0.00001737, -0.00000327, -0.00000114, -0.00000206, -0.00001446, -0.00000477, 0.00001656, 0.00001253, -0.00000534, -0.00000661, 0.00000356, 0.00000507, -0.00000228, -0.00000451, 0.00000099, 0.00000100, -0.00000588, -0.00000585, -0.00000221, -0.00000529, -0.00000170, 0.00001267, 0.00001189, -0.00000531, -0.00000966, -0.00000242, -0.00000334, -0.00000345, 0.00000503, 0.00000610, -0.00000136, -0.00000392, -0.00000426, -0.00000600, -0.00000406, -0.00000279, -0.00000459, -0.00000367, -0.00000731, -0.00001717, -0.00001336, -0.00000482, -0.00002688, },
		{ -0.00001027, -0.00003857, -0.00005120, -0.00000824, 0.00005326, 0.00007115, 0.00004358, 0.00002325, 0.00004057, 0.00005953, 0.00003701, -0.00000410, -0.00002263, -0.00002819, -0.00003144, -0.00001216, 0.00001182, 0.00000127, -0.00001581, -0.00000027, 0.00001221, -0.00000523, -0.00001103, 0.00000608, 0.00000386, -0.00001430, -0.00000678, 0.00001792, 0.00001865, -0.00000571, -0.00001774, 0.00000060, 0.00001604, -0.00000420, -0.00002510, -0.00000371, 0.00001950, -0.00000064, -0.00002023, -0.00000382, 0.00000660, -0.00000227, 0.00000413, 0.00000811, -0.00001572, -0.00002020, 0.00000980, 0.00000883, -0.00002189, -0.00000794, 0.00002824, 0.00000754, -0.00002878, -0.00000154, 0.00003419, 0.00000424, -0.00002844, -0.00000084, 0.00002158, -0.00000536, -0.00001677, 0.00000710, 0.00000586, -0.00001240, 0.00000335, 0.00001784, -0.00000633, -0.00001610, 0.00000535, 0.00000187, -0.00001366, 0.00000671, 0.00001904, -0.00000845, -0.00001148, 0.00001679, 0.00000607, -0.00002171, -0.00000387, 0.00001551, -0.00000416, -0.00000763, 0.00001255, -0.00000204, -0.00002644, -0.00000404, 0.00002370, 0.00001073, -0.00000478, -0.00000085, -0.00000578, -0.00001328, 0.00000132, 0.00001672, 0.00000828, -0.00000558, -0.00000399, 0.00000399, 0.00000304, -0.00000365, -0.00000302, 0.00000236, -0.00000096, -0.00000731, -0.00000438, -0.00000226, -0.00000601, 0.00000131, 0.00001442, 0.00000834, -0.00000715, -0.00000759, -0.00000237, -0.00000395, -0.00000138, 0.00000539, 0.00000398, -0.00000149, -0.00000370, -0.00000554, -0.00000557, -0.00000325, -0.00000390, -0.00000403, -0.00000287, -0.00001016, -0.00001697, -0.00000916, },
		{ -0.00001431, -0.00000780, -0.00001964, -0.00004472, -0.00003738, 0.00001342, 0.00006071, 0.00006228, 0.00003198, 0.00002092, 0.00004457, 0.00005486, 0.00002332, -0.00000915, -0.00001751, -0.00002738, -0.00003207, -0.00000560, 0.00001496, -0.00000397, -0.00001590, 0.00000477, 0.00001013, -0.00000909, -0.00000777, 0.00000820, 0.00000040, -0.00001484, -0.00000279, 0.00002015, 0.00001662, -0.00001028, -0.00001869, 0.00000623, 0.00001762, -0.00001127, -0.00002575, 0.00000462, 0.00001889, -0.00000846, -0.00001809, 0.00000172, 0.00000389, -0.00000263, 0.00000796, 0.00000266, -0.00002107, -0.00001078, 0.00001521, -0.00000250, -0.00002447, 0.00000636, 0.00002946, -0.00000737, -0.00002801, 0.00001337, 0.00003211, -0.00000954, -0.00002574, 0.00000980, 0.00001690, -0.00001362, -0.00001167, 0.00001121, -0.00000003, -0.00001215, 0.00001014, 0.00001492, -0.00001189, -0.00001153, 0.00000754, -0.00000371, -0.00001152, 0.00001261, 0.00001379, -0.00001219, -0.00000386, 0.00001666, -0.00000356, -0.00001914, 0.00000447, 0.00001115, -0.00000791, -0.00000085, 0.00001097, -0.00001210, -0.00002288, 0.00000734, 0.00002284, 0.00000451, -0.00000487, -0.00000171, -0.00000863, -0.00001012, 0.00000611, 0.00001521, 0.00000506, -0.00000486, -0.00000203, 0.00000360, 0.00000094, -0.00000407, -0.00000096, 0.00000259, -0.00000346, -0.00000752, -0.00000284, -0.00000316, -0.00000585, 0.00000477, 0.00001460, 0.00000470, -0.00000761, -0.00000599, -0.00000291, -0.00000360, 0.00000046, 0.00000473, 0.00000250, -0.00000134, -0.00000431, -0.00000627, -0.00000466, -0.00000335, -0.00000444, -0.00000272, -0.00000371, -0.00001261, },
		{ -0.00000587, -0.00001317, -0.00001091, -0.00001091, -0.00002946, -0.00004313, -0.00001968, 0.00002954, 0.00006224, 0.00005221, 0.00002208, 0.00002142, 0.00004846, 0.00004682, 0.00001053, -0.00000946, -0.00001280, -0.00002956, -0.00003082, 0.00000242, 0.00001508, -0.00000955, -0.00001343, 0.00000865, 0.00000632, -0.00001101, -0.00000386, 0.00000856, -0.00000273, -0.00001426, 0.00000076, 0.00002220, 0.00001401, -0.00001584, -0.00001779, 0.00001337, 0.00001625, -0.00001917, -0.00002262, 0.00001248, 0.00001443, -0.00001444, -0.00001307, 0.00000467, 0.00000046, -0.00000033, 0.00000958, -0.00000507, -0.00002132, 0.00000033, 0.00001379, -0.00001427, -0.00001932, 0.00001966, 0.00002272, -0.00002021, -0.00001987, 0.00002535, 0.00002354, -0.00002001, -0.00001790, 0.00001683, 0.00000854, -0.00001819, -0.00000450, 0.00001191, -0.00000602, -0.00000874, 0.00001547, 0.00000962, -0.00001505, -0.00000607, 0.00000697, -0.00000834, -0.00000699, 0.00001592, 0.00000742, -0.00001228, 0.00000288, 0.00001239, -0.00001051, -0.00001292, 0.00000923, 0.00000509, -0.00000770, 0.00000488, 0.00000509, -0.00001905, -0.00001475, 0.00001574, 0.00001861, -0.00000017, -0.00000436, -0.00000349, -0.00000993, -0.00000587, 0.00000913, 0.00001287, 0.00000302, -0.00000374, -0.00000085, 0.00000256, -0.00000084, -0.00000344, 0.00000100, 0.00000152, -0.00000573, -0.00000660, -0.00000186, -0.00000445, -0.00000454, 0.00000792, 0.00001339, 0.00000171, -0.00000722, -0.00000517, -0.00000338, -0.00000249, 0.00000155, 0.00000363, 0.00000173, -0.00000150, -0.00000527, -0.00000623, -0.00000398, -0.00000390, -0.00000394, -0.00000181, },
		{ -0.00000274, -0.00000221, -0.00000808, -0.00001142, -0.00000928, -0.00001802, -0.00003580, -0.00003438, -0.00000252, 0.00003970, 0.00006026, 0.00004153, 0.00001411, 0.00002486, 0.00005074, 0.00003550, 0.00000074, -0.00000585, -0.00001058, -0.00003338, -0.00002649, 0.00001010, 0.00001204, -0.00001381, -0.00000917, 0.00001037, 0.00000207, -0.00001083, -0.00000049, 0.00000766, -0.00000496, -0.00001332, 0.00000406, 0.00002438, 0.00000997, -0.00002184, -0.00001393, 0.00002055, 0.00001124, -0.00002587, -0.00001596, 0.00001773, 0.00000739, -0.00001702, -0.00000707, 0.00000460, -0.00000174, 0.00000352, 0.00000757, -0.00001236, -0.00001593, 0.00000927, 0.00000601, -0.00002214, -0.00000791, 0.00002770, 0.00001021, -0.00002728, -0.00000698, 0.00003140, 0.00001125, -0.00002494, -0.00000761, 0.00001878, -0.00000104, -0.00001837, 0.00000263, 0.00000930, -0.00001043, -0.00000301, 0.00001822, 0.00000321, -0.00001547, -0.00000113, 0.00000417, -0.00001091, -0.00000140, 0.00001627, 0.00000171, -0.00000943, 0.00000689, 0.00000573, -0.00001335, -0.00000541, 0.00000988, -0.00000030, -0.00000431, 0.00000747, -0.00000316, -0.00002117, -0.00000439, 0.00001993, 0.00001271, -0.00000294, -0.00000404, -0.00000559, -0.00000943, -0.00000150, 0.00001039, 0.00001047, 0.00000203, -0.00000268, -0.00000044, 0.00000117, -0.00000194, -0.00000195, 0.00000222, -0.00000054, -0.00000712, -0.00000502, -0.00000177, -0.00000551, -0.00000227, 0.00001017, 0.00001134, -0.00000033, -0.00000660, -0.00000497, -0.00000331, -0.00000115, 0.00000181, 0.00000262, 0.00000133, -0.00000216, -0.00000601, -0.00000569, -0.00000387, -0.00000415, },
		{ -0.00000413, -0.00000358, -0.00000174, -0.00000392, -0.00000883, -0.00000847, -0.00001138, -0.00002658, -0.00003594, -0.00002139, 0.00001102, 0.00004535, 0.00005626, 0.00003023, 0.00000891, 0.00003079, 0.00004973, 0.00002219, -0.00000438, -0.00000033, -0.00001192, -0.00003670, -0.00001917, 0.00001557, 0.00000687, -0.00001564, -0.00000454, 0.00000986, -0.00000128, -0.00000924, 0.00000164, 0.00000643, -0.00000637, -0.00001258, 0.00000775, 0.00002638, 0.00000378, -0.00002694, -0.00000681, 0.00002584, 0.00000310, -0.00002945, -0.00000727, 0.00001907, -0.00000008, -0.00001590, -0.00000223, 0.00000245, -0.00000145, 0.00000682, 0.00000211, -0.00001626, -0.00000657, 0.00001276, -0.00000527, -0.00002312, 0.00000603, 0.00002807, -0.00000404, -0.00002688, 0.00000674, 0.00003056, -0.00000121, -0.00002396, 0.00000209, 0.00001579, -0.00000933, -0.00001473, 0.00000786, 0.00000441, -0.00001212, 0.00000366, 0.00001798, -0.00000288, -0.00001351, 0.00000221, 0.00000022, -0.00001104, 0.00000381, 0.00001421, -0.00000203, -0.00000520, 0.00000738, -0.00000102, -0.00001200, 0.00000100, 0.00000726, -0.00000321, 0.00000043, 0.00000602, -0.00001117, -0.00001817, 0.00000558, 0.00001986, 0.00000682, -0.00000411, -0.00000437, -0.00000727, -0.00000740, 0.00000217, 0.00001031, 0.00000853, 0.00000174, -0.00000197, -0.00000070, -0.00000026, -0.00000212, -0.00000008, 0.00000230, -0.00000299, -0.00000735, -0.00000340, -0.00000245, -0.00000587, 0.00000046, 0.00001129, 0.00000909, -0.00000153, -0.00000622, -0.00000498, -0.00000264, -0.00000008, 0.00000149, 0.00000195, 0.00000088, -0.00000314, -0.00000620, -0.00000517, },
		{ -0.00000594, -0.00000504, -0.00000419, -0.00000232, -0.00000187, -0.00000600, -0.00000733, -0.00000649, -0.00001756, -0.00003302, -0.00002951, -0.00000809, 0.00002008, 0.00004849, 0.00005043, 0.00001861, 0.00000756, 0.00003773, 0.00004426, 0.00000919, -0.00000433, 0.00000450, -0.00001646, -0.00003754, -0.00001020, 0.00001775, 0.00000128, -0.00001501, -0.00000100, 0.00000798, -0.00000298, -0.00000737, 0.00000260, 0.00000566, -0.00000763, -0.00001204, 0.00001244, 0.00002722, -0.00000459, -0.00002947, 0.00000276, 0.00002752, -0.00000647, -0.00002889, 0.00000118, 0.00001658, -0.00000573, -0.00001224, 0.00000007, 0.00000009, 0.00000125, 0.00000747, -0.00000495, -0.00001484, 0.00000339, 0.00000945, -0.00001576, -0.00001669, 0.00001790, 0.00002100, -0.00001564, -0.00001982, 0.00001749, 0.00002402, -0.00001073, -0.00001846, 0.00000875, 0.00000937, -0.00001451, -0.00000878, 0.00001013, -0.00000122, -0.00001081, 0.00000976, 0.00001512, -0.00000750, -0.00001012, 0.00000348, -0.00000364, -0.00000906, 0.00000755, 0.00001090, -0.00000332, -0.00000139, 0.00000477, -0.00000577, -0.00000770, 0.00000464, 0.00000315, -0.00000301, 0.00000434, 0.00000104, -0.00001651, -0.00001117, 0.00001299, 0.00001655, 0.00000213, -0.00000439, -0.00000535, -0.00000795, -0.00000443, 0.00000468, 0.00000947, 0.00000724, 0.00000173, -0.00000178, -0.00000144, -0.00000134, -0.00000136, 0.00000157, 0.00000119, -0.00000514, -0.00000654, -0.00000227, -0.00000354, -0.00000535, 0.00000310, 0.00001143, 0.00000711, -0.00000228, -0.00000619, -0.00000477, -0.00000161, 0.00000045, 0.00000100, 0.00000153, 0.00000010, -0.00000400, },
		{ -0.00000094, -0.00000442, -0.00000561, -0.00000524, -0.00000352, -0.00000109, -0.00000342, -0.00000696, -0.00000412, -0.00000772, -0.00002609, -0.00003424, -0.00001861, 0.00000234, 0.00002580, 0.00005039, 0.00004221, 0.00000787, 0.00001069, 0.00004340, 0.00003441, -0.00000093, -0.00000017, 0.00000650, -0.00002249, -0.00003489, -0.00000161, 0.00001677, -0.00000306, -0.00001291, 0.00000071, 0.00000607, -0.00000319, -0.00000625, 0.00000313, 0.00000556, -0.00000948, -0.00001105, 0.00001823, 0.00002566, -0.00001422, -0.00002809, 0.00001298, 0.00002489, -0.00001519, -0.00002456, 0.00000726, 0.00001172, -0.00000820, -0.00000807, -0.00000026, -0.00000061, 0.00000486, 0.00000443, -0.00001065, -0.00000820, 0.00001009, 0.00000056, -0.00002139, -0.00000506, 0.00002389, 0.00000924, -0.00002141, -0.00000897, 0.00002284, 0.00001451, -0.00001553, -0.00001090, 0.00001117, 0.00000168, -0.00001591, -0.00000240, 0.00000935, -0.00000602, -0.00000700, 0.00001410, 0.00001059, -0.00001006, -0.00000640, 0.00000285, -0.00000644, -0.00000584, 0.00000938, 0.00000764, -0.00000262, 0.00000056, 0.00000047, -0.00000735, -0.00000239, 0.00000504, -0.00000043, -0.00000033, 0.00000564, -0.00000577, -0.00001767, -0.00000226, 0.00001667, 0.00001155, -0.00000089, -0.00000450, -0.00000659, -0.00000733, -0.00000126, 0.00000595, 0.00000843, 0.00000654, 0.00000168, -0.00000210, -0.00000241, -0.00000178, 0.00000007, 0.00000245, -0.00000077, -0.00000646, -0.00000510, -0.00000187, -0.00000458, -0.00000408, 0.00000525, 0.00001096, 0.00000554, -0.00000294, -0.00000632, -0.00000414, -0.00000062, 0.00000049, 0.00000061, 0.00000106, },
		{ 0.00000033, 0.00000030, -0.00000193, -0.00000437, -0.00000556, -0.00000532, -0.00000205, -0.00000085, -0.00000573, -0.00000565, -0.00000105, -0.00001324, -0.00003375, -0.00002903, -0.00000681, 0.00000864, 0.00003025, 0.00005097, 0.00003130, -0.00000010, 0.00001768, 0.00004545, 0.00002180, -0.00000630, 0.00000578, 0.00000478, -0.00002771, -0.00002918, 0.00000480, 0.00001391, -0.00000533, -0.00001085, 0.00000085, 0.00000529, -0.00000284, -0.00000621, 0.00000420, 0.00000560, -0.00001230, -0.00000861, 0.00002442, 0.00002079, -0.00002335, -0.00002247, 0.00002157, 0.00001858, -0.00002110, -0.00001814, 0.00000987, 0.00000672, -0.00000759, -0.00000539, -0.00000189, 0.00000127, 0.00000711, -0.00000160, -0.00001217, 0.00000138, 0.00001069, -0.00001045, -0.00001985, 0.00000763, 0.00002237, -0.00000302, -0.00002036, 0.00000192, 0.00002237, 0.00000519, -0.00001559, -0.00000388, 0.00000960, -0.00000512, -0.00001396, 0.00000277, 0.00000624, -0.00000886, -0.00000175, 0.00001607, 0.00000553, -0.00001059, -0.00000332, 0.00000097, -0.00000769, -0.00000242, 0.00000956, 0.00000541, -0.00000109, 0.00000007, -0.00000368, -0.00000576, 0.00000193, 0.00000292, -0.00000199, 0.00000329, 0.00000361, -0.00001215, -0.00001445, 0.00000621, 0.00001655, 0.00000644, -0.00000241, -0.00000494, -0.00000755, -0.00000556, 0.00000148, 0.00000622, 0.00000757, 0.00000621, 0.00000136, -0.00000286, -0.00000328, -0.00000141, 0.00000171, 0.00000227, -0.00000303, -0.00000672, -0.00000355, -0.00000216, -0.00000523, -0.00000239, 0.00000679, 0.00001027, 0.00000424, -0.00000372, -0.00000630, -0.00000316, 0.00000006, 0.00000031, },
		{ 0.00000035, 0.00000013, -0.00000003, -0.00000077, -0.00000251, -0.00000413, -0.00000588, -0.00000473, -0.00000036, -0.00000218, -0.00000739, -0.00000203, -0.00000053, -0.00002218, -0.00003710, -0.00001863, 0.00000238, 0.00001172, 0.00003501, 0.00004899, 0.00001843, -0.00000340, 0.00002655, 0.00004245, 0.00000910, -0.00000654, 0.00001086, 0.00000007, -0.00003028, -0.00002205, 0.00000824, 0.00001097, -0.00000588, -0.00001005, 0.00000059, 0.00000599, -0.00000317, -0.00000674, 0.00000642, 0.00000479, -0.00001567, -0.00000394, 0.00002958, 0.00001257, -0.00002995, -0.00001355, 0.00002672, 0.00001040, -0.00002328, -0.00001185, 0.00000939, 0.00000361, -0.00000537, -0.00000520, -0.00000280, 0.00000512, 0.00000605, -0.00000826, -0.00000811, 0.00001015, 0.00000464, -0.00001915, -0.00001155, 0.00001690, 0.00001443, -0.00001178, -0.00001395, 0.00000962, 0.00001761, -0.00000146, -0.00001236, 0.00000069, 0.00000537, -0.00000956, -0.00000989, 0.00000573, 0.00000204, -0.00000928, 0.00000370, 0.00001571, 0.00000098, -0.00000961, -0.00000143, -0.00000131, -0.00000747, 0.00000040, 0.00000883, 0.00000454, -0.00000002, -0.00000242, -0.00000606, -0.00000207, 0.00000393, -0.00000019, -0.00000105, 0.00000597, -0.00000122, -0.00001604, -0.00000791, 0.00001233, 0.00001357, 0.00000231, -0.00000303, -0.00000579, -0.00000773, -0.00000305, 0.00000338, 0.00000589, 0.00000707, 0.00000598, 0.00000066, -0.00000388, -0.00000375, -0.00000027, 0.00000303, 0.00000104, -0.00000498, -0.00000602, -0.00000234, -0.00000290, -0.00000534, -0.00000060, 0.00000783, 0.00000956, 0.00000299, -0.00000455, -0.00000592, -0.00000211, },
		{ -0.00000516, -0.00000125, 0.00000042, 0.00000004, -0.00000071, -0.00000189, -0.00000257, -0.00000407, -0.00000624, -0.00000322, 0.00000059, -0.00000478, -0.00000693, 0.00000251, -0.00000436, -0.00003183, -0.00003402, -0.00000629, 0.00000706, 0.00001390, 0.00004022, 0.00004312, 0.00000565, -0.00000118, 0.00003444, 0.00003459, -0.00000094, -0.00000296, 0.00001310, -0.00000569, -0.00002958, -0.00001554, 0.00000927, 0.00000933, -0.00000600, -0.00001081, 0.00000132, 0.00000752, -0.00000499, -0.00000674, 0.00000959, 0.00000223, -0.00001849, 0.00000300, 0.00003208, 0.00000213, -0.00003251, -0.00000331, 0.00002773, 0.00000260, -0.00002231, -0.00000758, 0.00000749, 0.00000327, -0.00000366, -0.00000699, -0.00000129, 0.00000899, 0.00000111, -0.00001236, 0.00000072, 0.00001432, -0.00000599, -0.00002188, 0.00000049, 0.00001959, 0.00000343, -0.00001462, -0.00000528, 0.00001247, 0.00001119, -0.00000435, -0.00000800, 0.00000208, 0.00000026, -0.00001113, -0.00000519, 0.00000628, -0.00000199, -0.00000748, 0.00000826, 0.00001356, -0.00000240, -0.00000787, -0.00000081, -0.00000326, -0.00000627, 0.00000227, 0.00000804, 0.00000465, -0.00000040, -0.00000573, -0.00000586, 0.00000206, 0.00000333, -0.00000262, 0.00000177, 0.00000631, -0.00000733, -0.00001621, 0.00000007, 0.00001510, 0.00000913, -0.00000035, -0.00000340, -0.00000681, -0.00000689, -0.00000036, 0.00000434, 0.00000535, 0.00000688, 0.00000563, -0.00000040, -0.00000494, -0.00000358, 0.00000135, 0.00000358, -0.00000092, -0.00000615, -0.00000470, -0.00000169, -0.00000379, -0.00000496, 0.00000109, 0.00000857, 0.00000883, 0.00000162, -0.00000522, },
		{ 0.00000015, -0.00000550, -0.00000420, -0.00000071, 0.00000048, -0.00000012, -0.00000182, -0.00000272, -0.00000225, -0.00000440, -0.00000607, -0.00000114, -0.00000005, -0.00000744, -0.00000367, 0.00000560, -0.00001267, -0.00003858, -0.00002471, 0.00000414, 0.00000758, 0.00001747, 0.00004442, 0.00003297, -0.00000432, 0.00000573, 0.00003867, 0.00002372, -0.00000653, 0.00000200, 0.00001203, -0.00001028, -0.00002649, -0.00001117, 0.00000945, 0.00000932, -0.00000721, -0.00001233, 0.00000397, 0.00000855, -0.00000823, -0.00000508, 0.00001274, -0.00000235, -0.00001939, 0.00001126, 0.00003081, -0.00000858, -0.00003066, 0.00000597, 0.00002533, -0.00000306, -0.00001992, -0.00000596, 0.00000634, 0.00000503, -0.00000419, -0.00000900, 0.00000307, 0.00001039, -0.00000635, -0.00001125, 0.00001133, 0.00001162, -0.00001715, -0.00001718, 0.00001178, 0.00001506, -0.00000642, -0.00001149, 0.00000216, 0.00001080, 0.00000580, -0.00000394, -0.00000453, 0.00000075, -0.00000408, -0.00001021, -0.00000115, 0.00000487, -0.00000486, -0.00000416, 0.00001129, 0.00001044, -0.00000439, -0.00000608, -0.00000115, -0.00000446, -0.00000471, 0.00000331, 0.00000778, 0.00000488, -0.00000249, -0.00000840, -0.00000325, 0.00000501, 0.00000086, -0.00000318, 0.00000510, 0.00000378, -0.00001275, -0.00001262, 0.00000742, 0.00001453, 0.00000462, -0.00000169, -0.00000397, -0.00000755, -0.00000511, 0.00000199, 0.00000451, 0.00000487, 0.00000688, 0.00000502, -0.00000175, -0.00000578, -0.00000267, 0.00000302, 0.00000310, -0.00000309, -0.00000635, -0.00000319, -0.00000165, -0.00000457, -0.00000424, 0.00000262, 0.00000914, 0.00000794, },
		{ 0.00000955, 0.00000674, -0.00000124, -0.00000533, -0.00000333, -0.00000041, 0.00000066, -0.00000060, -0.00000320, -0.00000301, -0.00000192, -0.00000496, -0.00000496, 0.00000062, -0.00000243, -0.00000849, 0.00000170, 0.00000483, -0.00002358, -0.00003939, -0.00001173, 0.00000982, 0.00000627, 0.00002343, 0.00004536, 0.00001981, -0.00000915, 0.00001494, 0.00003780, 0.00001269, -0.00000762, 0.00000583, 0.00000880, -0.00001223, -0.00002290, -0.00000916, 0.00001055, 0.00001003, -0.00001043, -0.00001302, 0.00000842, 0.00000768, -0.00001189, -0.00000125, 0.00001449, -0.00000829, -0.00001736, 0.00001916, 0.00002567, -0.00001743, -0.00002537, 0.00001258, 0.00002129, -0.00000604, -0.00001816, -0.00000614, 0.00000758, 0.00000702, -0.00000738, -0.00000900, 0.00000905, 0.00000757, -0.00001342, -0.00000412, 0.00001959, 0.00000236, -0.00002433, -0.00000649, 0.00001808, 0.00000537, -0.00001157, -0.00000455, 0.00000582, 0.00000655, 0.00000314, -0.00000182, -0.00000318, -0.00000210, -0.00000667, -0.00000772, 0.00000144, 0.00000236, -0.00000608, -0.00000021, 0.00001259, 0.00000713, -0.00000519, -0.00000475, -0.00000197, -0.00000487, -0.00000328, 0.00000397, 0.00000813, 0.00000431, -0.00000577, -0.00000922, 0.00000073, 0.00000580, -0.00000214, -0.00000159, 0.00000736, -0.00000102, -0.00001576, -0.00000630, 0.00001251, 0.00001149, 0.00000104, -0.00000221, -0.00000484, -0.00000762, -0.00000269, 0.00000366, 0.00000412, 0.00000460, 0.00000691, 0.00000409, -0.00000327, -0.00000613, -0.00000110, 0.00000426, 0.00000168, -0.00000491, -0.00000564, -0.00000189, -0.00000210, -0.00000514, -0.00000330, 0.00000408, },
		{ -0.00000220, 0.00000552, 0.00000967, 0.00000525, -0.00000229, -0.00000482, -0.00000272, -0.00000014, 0.00000088, -0.00000159, -0.00000451, -0.00000274, -0.00000188, -0.00000524, -0.00000297, 0.00000103, -0.00000571, -0.00000665, 0.00000711, -0.00000101, -0.00003363, -0.00003316, 0.00000094, 0.00001009, 0.00000615, 0.00003080, 0.00004128, 0.00000637, -0.00000800, 0.00002326, 0.00003227, 0.00000415, -0.00000580, 0.00000704, 0.00000553, -0.00001155, -0.00002079, -0.00000836, 0.00001357, 0.00000975, -0.00001543, -0.00001133, 0.00001346, 0.00000419, -0.00001442, 0.00000430, 0.00001372, -0.00001418, -0.00001233, 0.00002489, 0.00001780, -0.00002291, -0.00001856, 0.00001607, 0.00001762, -0.00000725, -0.00001840, -0.00000612, 0.00001141, 0.00000698, -0.00001205, -0.00000548, 0.00001414, 0.00000047, -0.00001664, 0.00000745, 0.00002183, -0.00001052, -0.00002438, 0.00000638, 0.00001700, -0.00000553, -0.00001054, 0.00000280, 0.00000496, 0.00000232, 0.00000350, 0.00000015, -0.00000408, -0.00000505, -0.00000728, -0.00000476, 0.00000239, -0.00000033, -0.00000564, 0.00000353, 0.00001240, 0.00000424, -0.00000522, -0.00000404, -0.00000282, -0.00000471, -0.00000218, 0.00000475, 0.00000872, 0.00000237, -0.00000916, -0.00000773, 0.00000462, 0.00000434, -0.00000423, 0.00000152, 0.00000741, -0.00000670, -0.00001541, 0.00000104, 0.00001453, 0.00000726, -0.00000121, -0.00000250, -0.00000582, -0.00000680, -0.00000008, 0.00000453, 0.00000346, 0.00000454, 0.00000685, 0.00000283, -0.00000475, -0.00000579, 0.00000085, 0.00000466, -0.00000037, -0.00000597, -0.00000432, -0.00000105, -0.00000284, -0.00000544, },
		{ -0.00000372, -0.00000551, -0.00000091, 0.00000691, 0.00000936, 0.00000362, -0.00000286, -0.00000422, -0.00000238, 0.00000029, 0.00000091, -0.00000309, -0.00000540, -0.00000218, -0.00000216, -0.00000474, -0.00000081, -0.00000052, -0.00000826, -0.00000178, 0.00000986, -0.00001117, -0.00003915, -0.00002136, 0.00000953, 0.00000673, 0.00000941, 0.00003707, 0.00003201, -0.00000412, -0.00000200, 0.00002800, 0.00002418, -0.00000052, -0.00000355, 0.00000579, 0.00000425, -0.00000972, -0.00002120, -0.00000684, 0.00001822, 0.00000684, -0.00002076, -0.00000650, 0.00001728, -0.00000153, -0.00001447, 0.00001022, 0.00001011, -0.00001841, -0.00000527, 0.00002727, 0.00000913, -0.00002479, -0.00001228, 0.00001736, 0.00001559, -0.00000853, -0.00002060, -0.00000380, 0.00001647, 0.00000350, -0.00001584, 0.00000148, 0.00001563, -0.00000890, -0.00001357, 0.00001979, 0.00001637, -0.00002242, -0.00001674, 0.00001663, 0.00000898, -0.00001338, -0.00000428, 0.00000734, 0.00000074, 0.00000029, 0.00000582, 0.00000050, -0.00000649, -0.00000703, -0.00000632, -0.00000219, 0.00000195, -0.00000247, -0.00000393, 0.00000648, 0.00001119, 0.00000205, -0.00000493, -0.00000388, -0.00000340, -0.00000431, -0.00000129, 0.00000591, 0.00000884, -0.00000091, -0.00001145, -0.00000427, 0.00000711, 0.00000139, -0.00000448, 0.00000489, 0.00000494, -0.00001155, -0.00001180, 0.00000759, 0.00001360, 0.00000306, -0.00000227, -0.00000294, -0.00000656, -0.00000511, 0.00000229, 0.00000462, 0.00000272, 0.00000466, 0.00000660, 0.00000127, -0.00000595, -0.00000467, 0.00000273, 0.00000404, -0.00000255, -0.00000607, -0.00000277, -0.00000077, },
		{ -0.00000137, -0.00000100, -0.00000459, -0.00000531, 0.00000061, 0.00000812, 0.00000854, 0.00000211, -0.00000297, -0.00000378, -0.00000214, 0.00000092, 0.00000047, -0.00000485, -0.00000572, -0.00000165, -0.00000241, -0.00000324, 0.00000045, -0.00000369, -0.00000836, 0.00000472, 0.00000779, -0.00002288, -0.00003770, -0.00000754, 0.00001204, 0.00000303, 0.00001626, 0.00003935, 0.00001937, -0.00000922, 0.00000608, 0.00002801, 0.00001639, -0.00000179, -0.00000305, 0.00000377, 0.00000583, -0.00000897, -0.00002364, -0.00000280, 0.00002289, 0.00000063, -0.00002435, 0.00000096, 0.00001823, -0.00000806, -0.00001150, 0.00001487, 0.00000442, -0.00001984, 0.00000212, 0.00002622, 0.00000159, -0.00002408, -0.00000778, 0.00001817, 0.00001510, -0.00001171, -0.00002323, 0.00000200, 0.00002028, -0.00000326, -0.00001627, 0.00000999, 0.00001189, -0.00001706, -0.00000391, 0.00002844, 0.00000434, -0.00002886, -0.00000384, 0.00002034, -0.00000295, -0.00001522, 0.00000427, 0.00000725, -0.00000442, 0.00000140, 0.00000835, -0.00000128, -0.00000922, -0.00000755, -0.00000451, -0.00000053, 0.00000067, -0.00000364, -0.00000150, 0.00000836, 0.00000949, 0.00000058, -0.00000468, -0.00000404, -0.00000365, -0.00000389, -0.00000033, 0.00000733, 0.00000786, -0.00000492, -0.00001177, 0.00000010, 0.00000745, -0.00000178, -0.00000277, 0.00000718, 0.00000054, -0.00001413, -0.00000593, 0.00001199, 0.00001053, -0.00000024, -0.00000257, -0.00000365, -0.00000672, -0.00000279, 0.00000409, 0.00000407, 0.00000207, 0.00000493, 0.00000609, -0.00000051, -0.00000662, -0.00000289, 0.00000408, 0.00000249, -0.00000432, -0.00000529, },
		{ -0.00000527, -0.00000389, -0.00000039, -0.00000164, -0.00000539, -0.00000479, 0.00000234, 0.00000894, 0.00000730, 0.00000096, -0.00000280, -0.00000363, -0.00000177, 0.00000159, -0.00000059, -0.00000654, -0.00000552, -0.00000133, -0.00000211, -0.00000116, -0.00000013, -0.00000717, -0.00000505, 0.00001028, 0.00000034, -0.00003226, -0.00002924, 0.00000395, 0.00000906, 0.00000233, 0.00002474, 0.00003585, 0.00000665, -0.00000837, 0.00001296, 0.00002415, 0.00001122, -0.00000163, -0.00000510, 0.00000331, 0.00000948, -0.00001120, -0.00002626, 0.00000443, 0.00002538, -0.00000816, -0.00002439, 0.00000937, 0.00001558, -0.00001356, -0.00000608, 0.00001698, -0.00000188, -0.00001837, 0.00000813, 0.00002286, -0.00000364, -0.00002258, -0.00000504, 0.00002016, 0.00001473, -0.00001759, -0.00002386, 0.00001077, 0.00002044, -0.00001138, -0.00001197, 0.00001700, 0.00000333, -0.00002033, 0.00001003, 0.00002987, -0.00001053, -0.00002719, 0.00000980, 0.00001597, -0.00001430, -0.00001048, 0.00001145, 0.00000269, -0.00000793, 0.00000514, 0.00000940, -0.00000474, -0.00001120, -0.00000672, -0.00000257, 0.00000011, -0.00000089, -0.00000375, 0.00000108, 0.00000921, 0.00000775, -0.00000035, -0.00000463, -0.00000426, -0.00000367, -0.00000351, 0.00000094, 0.00000857, 0.00000548, -0.00000863, -0.00000989, 0.00000408, 0.00000574, -0.00000396, 0.00000028, 0.00000746, -0.00000451, -0.00001374, 0.00000067, 0.00001361, 0.00000647, -0.00000230, -0.00000262, -0.00000449, -0.00000609, -0.00000019, 0.00000514, 0.00000305, 0.00000161, 0.00000530, 0.00000524, -0.00000233, -0.00000652, -0.00000075, 0.00000450, 0.00000037, },
		{ 0.00000383, -0.00000179, -0.00000522, -0.00000228, 0.00000004, -0.00000253, -0.00000602, -0.00000390, 0.00000415, 0.00000921, 0.00000587, 0.00000029, -0.00000264, -0.00000369, -0.00000111, 0.00000202, -0.00000222, -0.00000785, -0.00000507, -0.00000106, -0.00000088, 0.00000053, -0.00000279, -0.00000907, 0.00000119, 0.00001210, -0.00001084, -0.00003584, -0.00001635, 0.00000976, 0.00000352, 0.00000635, 0.00003152, 0.00002683, -0.00000267, -0.00000316, 0.00001618, 0.00001883, 0.00000942, -0.00000256, -0.00000876, 0.00000622, 0.00001301, -0.00001696, -0.00002662, 0.00001397, 0.00002383, -0.00001755, -0.00002015, 0.00001654, 0.00000986, -0.00001646, 0.00000036, 0.00001618, -0.00000718, -0.00001492, 0.00001176, 0.00001891, -0.00000666, -0.00002190, -0.00000277, 0.00002397, 0.00001235, -0.00002534, -0.00002032, 0.00002036, 0.00001572, -0.00001790, -0.00000347, 0.00001955, -0.00000745, -0.00001643, 0.00002402, 0.00002295, -0.00002327, -0.00001774, 0.00001923, 0.00000496, -0.00002087, -0.00000109, 0.00001430, -0.00000439, -0.00000804, 0.00000989, 0.00000802, -0.00000881, -0.00001181, -0.00000502, -0.00000102, -0.00000010, -0.00000223, -0.00000295, 0.00000335, 0.00000924, 0.00000624, -0.00000099, -0.00000482, -0.00000440, -0.00000358, -0.00000304, 0.00000257, 0.00000907, 0.00000193, -0.00001103, -0.00000631, 0.00000655, 0.00000274, -0.00000439, 0.00000354, 0.00000553, -0.00000876, -0.00001057, 0.00000645, 0.00001263, 0.00000248, -0.00000325, -0.00000279, -0.00000517, -0.00000463, 0.00000232, 0.00000534, 0.00000173, 0.00000147, 0.00000569, 0.00000398, -0.00000396, -0.00000556, 0.00000128, },
		{ -0.00000384, 0.00000268, 0.00000221, -0.00000343, -0.00000427, -0.00000084, -0.00000008, -0.00000355, -0.00000640, -0.00000260, 0.00000583, 0.00000889, 0.00000453, 0.00000003, -0.00000270, -0.00000374, -0.00000017, 0.00000194, -0.00000417, -0.00000869, -0.00000458, -0.00000039, 0.00000118, 0.00000069, -0.00000672, -0.00000776, 0.00000835, 0.00000843, -0.00002233, -0.00003195, -0.00000327, 0.00000890, -0.00000065, 0.00001442, 0.00003344, 0.00001476, -0.00000634, 0.00000329, 0.00001512, 0.00001486, 0.00000984, -0.00000638, -0.00001168, 0.00001278, 0.00001376, -0.00002510, -0.00002279, 0.00002358, 0.00001764, -0.00002503, -0.00001234, 0.00002059, 0.00000265, -0.00001617, 0.00000610, 0.00001319, -0.00001047, -0.00001104, 0.00001317, 0.00001593, -0.00000873, -0.00002270, 0.00000087, 0.00002875, 0.00000626, -0.00003264, -0.00001181, 0.00002772, 0.00000683, -0.00002010, 0.00000680, 0.00001614, -0.00001655, -0.00000555, 0.00003341, 0.00000953, -0.00002951, -0.00000376, 0.00002094, -0.00000880, -0.00002027, 0.00000934, 0.00001167, -0.00001114, -0.00000449, 0.00001370, 0.00000431, -0.00001232, -0.00001099, -0.00000306, -0.00000009, -0.00000087, -0.00000306, -0.00000153, 0.00000506, 0.00000877, 0.00000507, -0.00000157, -0.00000516, -0.00000439, -0.00000346, -0.00000231, 0.00000434, 0.00000835, -0.00000213, -0.00001147, -0.00000203, 0.00000697, -0.00000039, -0.00000300, 0.00000586, 0.00000197, -0.00001104, -0.00000553, 0.00001029, 0.00000978, -0.00000068, -0.00000346, -0.00000320, -0.00000538, -0.00000248, 0.00000441, 0.00000470, 0.00000035, 0.00000174, 0.00000594, 0.00000232, -0.00000506, },
		{ 0.00000039, -0.00000533, -0.00000171, 0.00000305, 0.00000006, -0.00000412, -0.00000273, 0.00000013, -0.00000065, -0.00000458, -0.00000643, -0.00000095, 0.00000714, 0.00000810, 0.00000352, -0.00000005, -0.00000304, -0.00000356, 0.00000086, 0.00000124, -0.00000614, -0.00000915, -0.00000403, 0.00000113, 0.00000332, -0.00000140, -0.00001019, -0.00000274, 0.00001349, -0.00000053, -0.00003007, -0.00002165, 0.00000563, 0.00000318, -0.00000023, 0.00002363, 0.00002897, 0.00000339, -0.00000433, 0.00000771, 0.00001124, 0.00001414, 0.00001001, -0.00001318, -0.00001122, 0.00002140, 0.00000981, -0.00003308, -0.00001428, 0.00003058, 0.00000781, -0.00002856, -0.00000287, 0.00002074, -0.00000409, -0.00001323, 0.00000993, 0.00000936, -0.00001175, -0.00000809, 0.00001353, 0.00001453, -0.00001155, -0.00002424, 0.00000726, 0.00003236, -0.00000384, -0.00003656, 0.00000056, 0.00003019, -0.00000373, -0.00001666, 0.00001538, 0.00000754, -0.00002018, 0.00000947, 0.00003489, -0.00000613, -0.00002720, 0.00000974, 0.00001423, -0.00002046, -0.00001282, 0.00001695, 0.00000452, -0.00001498, 0.00000154, 0.00001506, -0.00000083, -0.00001441, -0.00000915, -0.00000128, 0.00000020, -0.00000188, -0.00000325, 0.00000014, 0.00000612, 0.00000813, 0.00000419, -0.00000225, -0.00000551, -0.00000426, -0.00000328, -0.00000119, 0.00000579, 0.00000629, -0.00000575, -0.00000986, 0.00000176, 0.00000549, -0.00000260, -0.00000039, 0.00000648, -0.00000214, -0.00001082, 0.00000005, 0.00001171, 0.00000606, -0.00000274, -0.00000337, -0.00000376, -0.00000491, 0.00000010, 0.00000576, 0.00000327, -0.00000083, 0.00000246, 0.00000584, },
		{ 0.00000351, 0.00000517, -0.00000150, -0.00000461, 0.00000030, 0.00000221, -0.00000202, -0.00000372, -0.00000105, 0.00000049, -0.00000149, -0.00000549, -0.00000604, 0.00000089, 0.00000793, 0.00000709, 0.00000291, -0.00000020, -0.00000356, -0.00000301, 0.00000173, -0.00000003, -0.00000794, -0.00000941, -0.00000309, 0.00000366, 0.00000437, -0.00000556, -0.00001118, 0.00000487, 0.00001394, -0.00001235, -0.00003101, -0.00000841, 0.00000768, -0.00000359, 0.00000584, 0.00003007, 0.00001903, -0.00000374, 0.00000118, 0.00000803, 0.00000730, 0.00001671, 0.00000730, -0.00002119, -0.00000572, 0.00002918, 0.00000091, -0.00003795, -0.00000236, 0.00003287, -0.00000336, -0.00002735, 0.00000590, 0.00001752, -0.00000880, -0.00000900, 0.00001157, 0.00000604, -0.00001196, -0.00000664, 0.00001444, 0.00001403, -0.00001639, -0.00002468, 0.00001645, 0.00003229, -0.00001642, -0.00003477, 0.00001397, 0.00002666, -0.00001257, -0.00000844, 0.00001918, -0.00000327, -0.00001627, 0.00002408, 0.00002786, -0.00001890, -0.00001743, 0.00001788, 0.00000142, -0.00002601, -0.00000127, 0.00001912, -0.00000456, -0.00001459, 0.00000802, 0.00001342, -0.00000609, -0.00001483, -0.00000684, 0.00000007, -0.00000006, -0.00000284, -0.00000283, 0.00000174, 0.00000662, 0.00000753, 0.00000344, -0.00000308, -0.00000577, -0.00000403, -0.00000295, 0.00000023, 0.00000646, 0.00000322, -0.00000807, -0.00000676, 0.00000411, 0.00000286, -0.00000324, 0.00000248, 0.00000528, -0.00000558, -0.00000833, 0.00000489, 0.00001089, 0.00000239, -0.00000375, -0.00000331, -0.00000426, -0.00000364, 0.00000278, 0.00000614, 0.00000125, -0.00000148, },
		{ -0.00000101, -0.00000134, 0.00000461, 0.00000382, -0.00000292, -0.00000299, 0.00000159, 0.00000036, -0.00000340, -0.00000237, 0.00000032, 0.00000028, -0.00000246, -0.00000620, -0.00000519, 0.00000270, 0.00000817, 0.00000612, 0.00000261, -0.00000062, -0.00000405, -0.00000211, 0.00000218, -0.00000169, -0.00000953, -0.00000949, -0.00000128, 0.00000675, 0.00000322, -0.00001049, -0.00000829, 0.00001255, 0.00000845, -0.00002312, -0.00002439, 0.00000307, 0.00000299, -0.00000713, 0.00001589, 0.00003056, 0.00000679, -0.00000477, 0.00000674, 0.00000427, 0.00000614, 0.00002070, 0.00000022, -0.00002740, 0.00000464, 0.00003303, -0.00001128, -0.00003760, 0.00001035, 0.00002983, -0.00001318, -0.00002215, 0.00001208, 0.00001245, -0.00001094, -0.00000497, 0.00001172, 0.00000390, -0.00001247, -0.00000607, 0.00001701, 0.00001277, -0.00002326, -0.00002191, 0.00002685, 0.00002682, -0.00002851, -0.00002667, 0.00002493, 0.00001816, -0.00001676, 0.00000181, 0.00001684, -0.00001216, -0.00000543, 0.00003363, 0.00001469, -0.00002477, -0.00000405, 0.00001772, -0.00001299, -0.00002372, 0.00001031, 0.00001529, -0.00001247, -0.00001022, 0.00001289, 0.00000931, -0.00001031, -0.00001383, -0.00000453, 0.00000094, -0.00000071, -0.00000354, -0.00000192, 0.00000304, 0.00000676, 0.00000711, 0.00000266, -0.00000401, -0.00000587, -0.00000373, -0.00000236, 0.00000170, 0.00000601, -0.00000022, -0.00000860, -0.00000311, 0.00000460, 0.00000009, -0.00000227, 0.00000466, 0.00000274, -0.00000746, -0.00000434, 0.00000810, 0.00000845, -0.00000062, -0.00000402, -0.00000343, -0.00000443, -0.00000162, 0.00000516, 0.00000538, },
		{ 0.00000683, 0.00000348, -0.00000303, -0.00000032, 0.00000534, 0.00000191, -0.00000345, -0.00000087, 0.00000167, -0.00000198, -0.00000363, -0.00000048, 0.00000108, -0.00000036, -0.00000342, -0.00000662, -0.00000397, 0.00000426, 0.00000797, 0.00000539, 0.00000240, -0.00000133, -0.00000426, -0.00000101, 0.00000211, -0.00000355, -0.00001097, -0.00000916, 0.00000175, 0.00000934, -0.00000055, -0.00001414, -0.00000155, 0.00001717, -0.00000201, -0.00002882, -0.00001227, 0.00000868, -0.00000554, -0.00000456, 0.00002616, 0.00002406, -0.00000362, -0.00000032, 0.00000908, -0.00000148, 0.00000926, 0.00002310, -0.00001074, -0.00002883, 0.00001768, 0.00003091, -0.00002375, -0.00003157, 0.00002091, 0.00002252, -0.00001954, -0.00001484, 0.00001491, 0.00000735, -0.00001104, -0.00000209, 0.00001157, 0.00000260, -0.00001424, -0.00000494, 0.00002122, 0.00000891, -0.00003071, -0.00001466, 0.00003571, 0.00001592, -0.00003671, -0.00001373, 0.00003059, 0.00000755, -0.00001511, 0.00001046, 0.00000943, -0.00001547, 0.00000917, 0.00003520, -0.00000007, -0.00002233, 0.00000775, 0.00000939, -0.00002404, -0.00001469, 0.00001808, 0.00000707, -0.00001676, -0.00000353, 0.00001486, 0.00000395, -0.00001287, -0.00001196, -0.00000250, 0.00000134, -0.00000161, -0.00000383, -0.00000075, 0.00000392, 0.00000677, 0.00000683, 0.00000170, -0.00000496, -0.00000577, -0.00000332, -0.00000149, 0.00000283, 0.00000443, -0.00000319, -0.00000737, 0.00000006, 0.00000343, -0.00000190, -0.00000019, 0.00000553, -0.00000027, -0.00000738, 0.00000005, 0.00000932, 0.00000523, -0.00000267, -0.00000392, -0.00000372, -0.00000402, 0.00000100, },
		{ -0.00000285, 0.00000391, 0.00000734, 0.00000064, -0.00000425, 0.00000142, 0.00000528, -0.00000018, -0.00000282, 0.00000112, 0.00000039, -0.00000410, -0.00000259, 0.00000140, 0.00000111, -0.00000120, -0.00000427, -0.00000669, -0.00000249, 0.00000542, 0.00000755, 0.00000494, 0.00000205, -0.00000222, -0.00000407, 0.00000004, 0.00000153, -0.00000551, -0.00001231, -0.00000797, 0.00000586, 0.00001013, -0.00000631, -0.00001449, 0.00000737, 0.00001629, -0.00001432, -0.00002697, 0.00000114, 0.00000658, -0.00001337, 0.00000433, 0.00003235, 0.00001224, -0.00000871, 0.00000683, 0.00000653, -0.00000601, 0.00001607, 0.00002110, -0.00002314, -0.00002385, 0.00002998, 0.00002269, -0.00003327, -0.00002124, 0.00002714, 0.00001324, -0.00002169, -0.00000765, 0.00001493, 0.00000349, -0.00001032, -0.00000034, 0.00001211, 0.00000099, -0.00001725, -0.00000168, 0.00002574, 0.00000141, -0.00003631, -0.00000325, 0.00004009, 0.00000156, -0.00003855, 0.00000085, 0.00002991, -0.00000161, -0.00000865, 0.00001450, 0.00000010, -0.00001144, 0.00002290, 0.00002870, -0.00001150, -0.00001335, 0.00001347, -0.00000403, -0.00002814, -0.00000227, 0.00001981, -0.00000256, -0.00001649, 0.00000325, 0.00001372, -0.00000130, -0.00001375, -0.00000976, -0.00000082, 0.00000130, -0.00000260, -0.00000366, 0.00000042, 0.00000442, 0.00000685, 0.00000657, 0.00000049, -0.00000580, -0.00000547, -0.00000277, -0.00000045, 0.00000325, 0.00000206, -0.00000499, -0.00000489, 0.00000195, 0.00000123, -0.00000257, 0.00000222, 0.00000500, -0.00000282, -0.00000556, 0.00000386, 0.00000867, 0.00000195, -0.00000378, -0.00000372, -0.00000403, },
		},
		{ { -0.00000114, -0.00000146, -0.00000097, 0.00000198, 0.00000477, 0.00000290, -0.00000241, -0.00000486, -0.00000280, -0.00000069, 0.00000019, 0.00000224, 0.00000343, 0.00000019, -0.00000392, -0.00000367, -0.00000120, -0.00000008, 0.00000156, 0.00000394, 0.00000252, -0.00000283, -0.00000639, -0.00000558, -0.00000178, 0.00000302, 0.00000412, -0.00000083, -0.00000316, 0.00000373, 0.00000790, -0.00000133, -0.00000976, -0.00000272, 0.00000601, 0.00000082, -0.00000583, -0.00000035, 0.00000577, 0.00000273, 0.00000074, 0.00000389, 0.00000209, -0.00000391, -0.00000535, -0.00000593, -0.00000775, -0.00000110, 0.00000755, -0.00000070, -0.00001192, 0.00000141, 0.00001764, 0.00000145, -0.00001773, 0.00000163, 0.00002221, -0.00000029, -0.00001828, 0.00001254, 0.00002846, -0.00001872, -0.00004593, 0.00000987, 0.00005621, 0.00000987, -0.00004593, -0.00001872, 0.00002846, 0.00001254, -0.00001828, -0.00000029, 0.00002221, 0.00000163, -0.00001773, 0.00000145, 0.00001764, 0.00000141, -0.00001192, -0.00000070, 0.00000755, -0.00000110, -0.00000775, -0.00000593, -0.00000535, -0.00000391, 0.00000209, 0.00000389, 0.00000074, 0.00000273, 0.00000577, -0.00000035, -0.00000583, 0.00000082, 0.00000601, -0.00000272, -0.00000976, -0.00000133, 0.00000790, 0.00000373, -0.00000316, -0.00000083, 0.00000412, 0.00000302, -0.00000178, -0.00000558, -0.00000639, -0.00000283, 0.00000252, 0.00000394, 0.00000156, -0.00000008, -0.00000120, -0.00000367, -0.00000392, 0.00000019, 0.00000343, 0.00000224, 0.00000019, -0.00000069, -0.00000280, -0.00000486, -0.00000241, 0.00000290, 0.00000477, 0.00000198, -0.00000097, -0.00000146, },
		{ -0.00000101, -0.00000162, -0.00000114, -0.00000131, -0.00000088, 0.00000204, 0.00000476, 0.00000259, -0.00000272, -0.00000457, -0.00000225, -0.00000056, 0.00000008, 0.00000221, 0.00000312, -0.00000037, -0.00000397, -0.00000323, -0.00000092, 0.00000034, 0.00000229, 0.00000371, 0.00000085, -0.00000399, -0.00000573, -0.00000442, -0.00000145, 0.00000289, 0.00000386, -0.00000117, -0.00000266, 0.00000496, 0.00000777, -0.00000311, -0.00001033, -0.00000116, 0.00000694, -0.00000023, -0.00000684, -0.00000012, 0.00000650, 0.00000357, 0.00000119, 0.00000294, 0.00000090, -0.00000335, -0.00000460, -0.00000699, -0.00000800, 0.00000095, 0.00000722, -0.00000428, -0.00001089, 0.00000832, 0.00001912, -0.00000580, -0.00002068, 0.00000809, 0.00002459, -0.00000826, -0.00002347, 0.00001746, 0.00003269, -0.00002157, -0.00004814, 0.00001130, 0.00005545, 0.00000860, -0.00004214, -0.00001611, 0.00002281, 0.00000823, -0.00001156, 0.00000696, 0.00001823, -0.00000411, -0.00001322, 0.00000800, 0.00001454, -0.00000521, -0.00001172, 0.00000273, 0.00000704, -0.00000305, -0.00000700, -0.00000495, -0.00000630, -0.00000429, 0.00000340, 0.00000468, 0.00000017, 0.00000183, 0.00000502, -0.00000041, -0.00000463, 0.00000167, 0.00000481, -0.00000408, -0.00000884, 0.00000029, 0.00000765, 0.00000248, -0.00000341, -0.00000038, 0.00000434, 0.00000303, -0.00000228, -0.00000673, -0.00000678, -0.00000150, 0.00000400, 0.00000400, 0.00000086, -0.00000048, -0.00000152, -0.00000400, -0.00000377, 0.00000066, 0.00000367, 0.00000237, 0.00000032, -0.00000101, -0.00000342, -0.00000495, -0.00000199, 0.00000312, 0.00000468, 0.00000194, },
		{ 0.00000451, 0.00000194, -0.00000103, -0.00000176, -0.00000114, -0.00000115, -0.00000075, 0.00000210, 0.00000465, 0.00000221, -0.00000286, -0.00000411, -0.00000181, -0.00000061, 0.00000003, 0.00000227, 0.00000268, -0.00000100, -0.00000389, -0.00000272, -0.00000066, 0.00000081, 0.00000303, 0.00000328, -0.00000094, -0.00000493, -0.00000481, -0.00000332, -0.00000130, 0.00000269, 0.00000359, -0.00000137, -0.00000193, 0.00000611, 0.00000725, -0.00000495, -0.00001051, 0.00000054, 0.00000754, -0.00000142, -0.00000762, 0.00000028, 0.00000719, 0.00000433, 0.00000150, 0.00000189, -0.00000013, -0.00000268, -0.00000409, -0.00000800, -0.00000770, 0.00000290, 0.00000605, -0.00000769, -0.00000861, 0.00001507, 0.00001886, -0.00001325, -0.00002189, 0.00001478, 0.00002519, -0.00001648, -0.00002695, 0.00002254, 0.00003533, -0.00002431, -0.00004870, 0.00001254, 0.00005325, 0.00000780, -0.00003692, -0.00001407, 0.00001601, 0.00000491, -0.00000362, 0.00001307, 0.00001299, -0.00000871, -0.00000748, 0.00001341, 0.00001004, -0.00001114, -0.00001041, 0.00000575, 0.00000578, -0.00000472, -0.00000583, -0.00000416, -0.00000739, -0.00000442, 0.00000477, 0.00000526, -0.00000051, 0.00000092, 0.00000427, -0.00000028, -0.00000329, 0.00000229, 0.00000342, -0.00000515, -0.00000765, 0.00000166, 0.00000705, 0.00000131, -0.00000342, 0.00000014, 0.00000451, 0.00000290, -0.00000295, -0.00000782, -0.00000687, -0.00000008, 0.00000527, 0.00000390, 0.00000023, -0.00000085, -0.00000184, -0.00000423, -0.00000356, 0.00000102, 0.00000387, 0.00000261, 0.00000040, -0.00000154, -0.00000405, -0.00000483, -0.00000148, 0.00000321, },
		{ -0.00000094, 0.00000318, 0.00000429, 0.00000198, -0.00000102, -0.00000191, -0.00000115, -0.00000098, -0.00000058, 0.00000217, 0.00000440, 0.00000178, -0.00000281, -0.00000354, -0.00000155, -0.00000080, 0.00000009, 0.00000238, 0.00000210, -0.00000167, -0.00000365, -0.00000214, -0.00000045, 0.00000131, 0.00000375, 0.00000263, -0.00000280, -0.00000562, -0.00000370, -0.00000232, -0.00000131, 0.00000241, 0.00000331, -0.00000142, -0.00000102, 0.00000710, 0.00000634, -0.00000674, -0.00001028, 0.00000227, 0.00000778, -0.00000268, -0.00000812, 0.00000081, 0.00000780, 0.00000500, 0.00000165, 0.00000078, -0.00000097, -0.00000199, -0.00000383, -0.00000884, -0.00000689, 0.00000454, 0.00000408, -0.00001062, -0.00000516, 0.00002123, 0.00001691, -0.00002036, -0.00002131, 0.00002120, 0.00002397, -0.00002445, -0.00002864, 0.00002732, 0.00003635, -0.00002656, -0.00004768, 0.00001328, 0.00004976, 0.00000775, -0.00003055, -0.00001289, 0.00000842, 0.00000286, 0.00000515, 0.00001775, 0.00000689, -0.00001189, -0.00000094, 0.00001731, 0.00000448, -0.00001602, -0.00000820, 0.00000810, 0.00000393, -0.00000596, -0.00000434, -0.00000365, -0.00000850, -0.00000426, 0.00000613, 0.00000560, -0.00000127, -0.00000000, 0.00000354, 0.00000003, -0.00000188, 0.00000265, 0.00000193, -0.00000590, -0.00000628, 0.00000272, 0.00000617, 0.00000026, -0.00000320, 0.00000071, 0.00000461, 0.00000262, -0.00000375, -0.00000878, -0.00000665, 0.00000138, 0.00000629, 0.00000370, -0.00000033, -0.00000120, -0.00000217, -0.00000435, -0.00000331, 0.00000128, 0.00000406, 0.00000293, 0.00000038, -0.00000225, -0.00000460, -0.00000450, },
		{ -0.00000504, -0.00000396, -0.00000042, 0.00000302, 0.00000405, 0.00000204, -0.00000100, -0.00000204, -0.00000116, -0.00000081, -0.00000037, 0.00000220, 0.00000403, 0.00000137, -0.00000256, -0.00000291, -0.00000149, -0.00000108, 0.00000028, 0.00000251, 0.00000136, -0.00000234, -0.00000324, -0.00000152, -0.00000029, 0.00000185, 0.00000440, 0.00000176, -0.00000466, -0.00000601, -0.00000244, -0.00000146, -0.00000147, 0.00000210, 0.00000307, -0.00000130, 0.00000005, 0.00000785, 0.00000509, -0.00000838, -0.00000965, 0.00000394, 0.00000764, -0.00000395, -0.00000832, 0.00000145, 0.00000832, 0.00000555, 0.00000167, -0.00000032, -0.00000162, -0.00000137, -0.00000380, -0.00000938, -0.00000563, 0.00000570, 0.00000145, -0.00001281, -0.00000073, 0.00002637, 0.00001342, -0.00002664, -0.00001905, 0.00002684, 0.00002101, -0.00003167, -0.00002865, 0.00003138, 0.00003585, -0.00002800, -0.00004524, 0.00001327, 0.00004526, 0.00000866, -0.00002338, -0.00001272, 0.00000050, 0.00000229, 0.00001427, 0.00002081, 0.00000041, -0.00001345, 0.00000591, 0.00001949, -0.00000173, -0.00001962, -0.00000536, 0.00000962, 0.00000166, -0.00000668, -0.00000268, -0.00000345, -0.00000955, -0.00000380, 0.00000740, 0.00000570, -0.00000207, -0.00000090, 0.00000285, 0.00000050, -0.00000047, 0.00000275, 0.00000041, -0.00000631, -0.00000482, 0.00000342, 0.00000508, -0.00000060, -0.00000278, 0.00000130, 0.00000463, 0.00000217, -0.00000468, -0.00000955, -0.00000615, 0.00000280, 0.00000705, 0.00000342, -0.00000079, -0.00000154, -0.00000248, -0.00000438, -0.00000306, 0.00000146, 0.00000429, 0.00000330, 0.00000022, -0.00000312, },
		{ -0.00000012, -0.00000408, -0.00000530, -0.00000325, 0.00000004, 0.00000275, 0.00000380, 0.00000213, -0.00000098, -0.00000217, -0.00000116, -0.00000062, -0.00000011, 0.00000220, 0.00000355, 0.00000100, -0.00000211, -0.00000228, -0.00000164, -0.00000140, 0.00000064, 0.00000260, 0.00000043, -0.00000295, -0.00000264, -0.00000090, -0.00000017, 0.00000243, 0.00000495, 0.00000067, -0.00000647, -0.00000610, -0.00000108, -0.00000078, -0.00000176, 0.00000178, 0.00000288, -0.00000100, 0.00000120, 0.00000831, 0.00000356, -0.00000978, -0.00000865, 0.00000544, 0.00000715, -0.00000515, -0.00000822, 0.00000215, 0.00000873, 0.00000596, 0.00000155, -0.00000134, -0.00000209, -0.00000088, -0.00000395, -0.00000952, -0.00000404, 0.00000621, -0.00000165, -0.00001401, 0.00000441, 0.00003016, 0.00000869, -0.00003163, -0.00001535, 0.00003125, 0.00001654, -0.00003770, -0.00002716, 0.00003432, 0.00003400, -0.00002837, -0.00004166, 0.00001232, 0.00004008, 0.00001065, -0.00001582, -0.00001364, -0.00000730, 0.00000324, 0.00002325, 0.00002221, -0.00000593, -0.00001339, 0.00001252, 0.00001984, -0.00000812, -0.00002179, -0.00000222, 0.00001020, -0.00000079, -0.00000683, -0.00000101, -0.00000359, -0.00001043, -0.00000305, 0.00000852, 0.00000554, -0.00000287, -0.00000176, 0.00000223, 0.00000112, 0.00000089, 0.00000262, -0.00000103, -0.00000638, -0.00000338, 0.00000376, 0.00000386, -0.00000124, -0.00000218, 0.00000189, 0.00000455, 0.00000154, -0.00000567, -0.00001008, -0.00000538, 0.00000411, 0.00000754, 0.00000313, -0.00000115, -0.00000187, -0.00000276, -0.00000432, -0.00000284, 0.00000156, 0.00000458, 0.00000369, },
		{ 0.00000493, 0.00000404, -0.00000066, -0.00000507, -0.00000535, -0.00000242, 0.00000040, 0.00000239, 0.00000356, 0.00000223, -0.00000095, -0.00000227, -0.00000117, -0.00000042, 0.00000019, 0.00000214, 0.00000295, 0.00000072, -0.00000146, -0.00000171, -0.00000201, -0.00000168, 0.00000115, 0.00000259, -0.00000066, -0.00000345, -0.00000186, -0.00000029, -0.00000010, 0.00000303, 0.00000538, -0.00000060, -0.00000815, -0.00000588, 0.00000030, -0.00000030, -0.00000215, 0.00000146, 0.00000277, -0.00000052, 0.00000237, 0.00000845, 0.00000183, -0.00001084, -0.00000735, 0.00000669, 0.00000632, -0.00000621, -0.00000783, 0.00000290, 0.00000902, 0.00000622, 0.00000133, -0.00000226, -0.00000243, -0.00000060, -0.00000420, -0.00000917, -0.00000226, 0.00000598, -0.00000498, -0.00001410, 0.00000991, 0.00003233, 0.00000310, -0.00003500, -0.00001058, 0.00003406, 0.00001090, -0.00004219, -0.00002450, 0.00003585, 0.00003111, -0.00002748, -0.00003725, 0.00001037, 0.00003459, 0.00001374, -0.00000829, -0.00001560, -0.00001449, 0.00000565, 0.00003161, 0.00002205, -0.00001167, -0.00001183, 0.00001839, 0.00001841, -0.00001422, -0.00002251, 0.00000090, 0.00000983, -0.00000320, -0.00000642, 0.00000052, -0.00000402, -0.00001103, -0.00000203, 0.00000941, 0.00000515, -0.00000365, -0.00000256, 0.00000169, 0.00000184, 0.00000213, 0.00000230, -0.00000231, -0.00000616, -0.00000206, 0.00000373, 0.00000260, -0.00000166, -0.00000145, 0.00000247, 0.00000436, 0.00000071, -0.00000670, -0.00001032, -0.00000440, 0.00000524, 0.00000779, 0.00000284, -0.00000142, -0.00000220, -0.00000301, -0.00000420, -0.00000265, 0.00000164, },
		{ -0.00000251, 0.00000171, 0.00000534, 0.00000429, -0.00000140, -0.00000603, -0.00000516, -0.00000153, 0.00000064, 0.00000197, 0.00000334, 0.00000234, -0.00000091, -0.00000234, -0.00000118, -0.00000021, 0.00000052, 0.00000201, 0.00000228, 0.00000056, -0.00000066, -0.00000126, -0.00000257, -0.00000188, 0.00000180, 0.00000244, -0.00000190, -0.00000380, -0.00000091, 0.00000026, -0.00000008, 0.00000365, 0.00000566, -0.00000203, -0.00000964, -0.00000538, 0.00000166, -0.00000004, -0.00000262, 0.00000119, 0.00000275, 0.00000011, 0.00000348, 0.00000824, -0.00000000, -0.00001153, -0.00000582, 0.00000763, 0.00000523, -0.00000708, -0.00000717, 0.00000365, 0.00000917, 0.00000632, 0.00000105, -0.00000302, -0.00000268, -0.00000056, -0.00000446, -0.00000832, -0.00000048, 0.00000497, -0.00000826, -0.00001299, 0.00001539, 0.00003275, -0.00000290, -0.00003653, -0.00000517, 0.00003502, 0.00000454, -0.00004488, -0.00002103, 0.00003579, 0.00002749, -0.00002529, -0.00003239, 0.00000742, 0.00002920, 0.00001783, -0.00000118, -0.00001845, -0.00002065, 0.00000934, 0.00003893, 0.00002057, -0.00001638, -0.00000904, 0.00002303, 0.00001541, -0.00001960, -0.00002186, 0.00000368, 0.00000859, -0.00000537, -0.00000551, 0.00000179, -0.00000469, -0.00001128, -0.00000082, 0.00001001, 0.00000456, -0.00000437, -0.00000329, 0.00000124, 0.00000262, 0.00000322, 0.00000184, -0.00000338, -0.00000570, -0.00000094, 0.00000338, 0.00000137, -0.00000184, -0.00000063, 0.00000301, 0.00000404, -0.00000030, -0.00000769, -0.00001024, -0.00000326, 0.00000615, 0.00000782, 0.00000261, -0.00000159, -0.00000252, -0.00000322, -0.00000404, },
		{ -0.00000338, -0.00000385, -0.00000240, 0.00000182, 0.00000580, 0.00000438, -0.00000232, -0.00000688, -0.00000475, -0.00000062, 0.00000076, 0.00000152, 0.00000315, 0.00000242, -0.00000087, -0.00000238, -0.00000119, 0.00000001, 0.00000086, 0.00000181, 0.00000154, 0.00000054, 0.00000027, -0.00000096, -0.00000330, -0.00000195, 0.00000255, 0.00000211, -0.00000324, -0.00000395, 0.00000019, 0.00000074, -0.00000009, 0.00000427, 0.00000576, -0.00000357, -0.00001090, -0.00000461, 0.00000293, 0.00000001, -0.00000313, 0.00000097, 0.00000285, 0.00000086, 0.00000446, 0.00000769, -0.00000183, -0.00001180, -0.00000417, 0.00000820, 0.00000393, -0.00000770, -0.00000628, 0.00000436, 0.00000918, 0.00000626, 0.00000072, -0.00000360, -0.00000290, -0.00000077, -0.00000461, -0.00000696, 0.00000113, 0.00000319, -0.00001123, -0.00001072, 0.00002045, 0.00003141, -0.00000883, -0.00003612, 0.00000039, 0.00003404, -0.00000208, -0.00004566, -0.00001720, 0.00003407, 0.00002353, -0.00002182, -0.00002742, 0.00000361, 0.00002427, 0.00002272, 0.00000513, -0.00002191, -0.00002542, 0.00001399, 0.00004487, 0.00001809, -0.00001975, -0.00000539, 0.00002607, 0.00001112, -0.00002388, -0.00002004, 0.00000586, 0.00000659, -0.00000711, -0.00000422, 0.00000267, -0.00000551, -0.00001112, 0.00000051, 0.00001029, 0.00000379, -0.00000501, -0.00000393, 0.00000089, 0.00000342, 0.00000412, 0.00000131, -0.00000417, -0.00000508, -0.00000009, 0.00000276, 0.00000025, -0.00000181, 0.00000026, 0.00000350, 0.00000357, -0.00000149, -0.00000859, -0.00000983, -0.00000202, 0.00000680, 0.00000768, 0.00000246, -0.00000168, -0.00000284, },
		{ -0.00000169, -0.00000315, -0.00000351, -0.00000367, -0.00000232, 0.00000199, 0.00000625, 0.00000427, -0.00000337, -0.00000754, -0.00000412, 0.00000023, 0.00000075, 0.00000109, 0.00000298, 0.00000246, -0.00000080, -0.00000237, -0.00000122, 0.00000024, 0.00000122, 0.00000152, 0.00000078, 0.00000068, 0.00000128, -0.00000087, -0.00000413, -0.00000183, 0.00000336, 0.00000155, -0.00000465, -0.00000385, 0.00000141, 0.00000112, -0.00000012, 0.00000488, 0.00000567, -0.00000519, -0.00001189, -0.00000361, 0.00000407, -0.00000015, -0.00000367, 0.00000084, 0.00000307, 0.00000169, 0.00000525, 0.00000684, -0.00000354, -0.00001165, -0.00000250, 0.00000838, 0.00000251, -0.00000805, -0.00000519, 0.00000502, 0.00000905, 0.00000606, 0.00000039, -0.00000400, -0.00000314, -0.00000122, -0.00000456, -0.00000515, 0.00000239, 0.00000075, -0.00001360, -0.00000742, 0.00002470, 0.00002842, -0.00001421, -0.00003384, 0.00000559, 0.00003115, -0.00000847, -0.00004456, -0.00001342, 0.00003074, 0.00001957, -0.00001725, -0.00002270, -0.00000083, 0.00002011, 0.00002815, 0.00001038, -0.00002565, -0.00002855, 0.00001926, 0.00004922, 0.00001501, -0.00002158, -0.00000135, 0.00002726, 0.00000597, -0.00002678, -0.00001730, 0.00000723, 0.00000405, -0.00000831, -0.00000270, 0.00000311, -0.00000636, -0.00001052, 0.00000188, 0.00001023, 0.00000290, -0.00000554, -0.00000447, 0.00000062, 0.00000419, 0.00000483, 0.00000078, -0.00000467, -0.00000438, 0.00000045, 0.00000195, -0.00000071, -0.00000159, 0.00000116, 0.00000393, 0.00000295, -0.00000283, -0.00000934, -0.00000909, -0.00000076, 0.00000715, 0.00000739, 0.00000242, },
		{ 0.00000702, 0.00000248, -0.00000165, -0.00000344, -0.00000360, -0.00000350, -0.00000224, 0.00000224, 0.00000666, 0.00000391, -0.00000449, -0.00000796, -0.00000332, 0.00000097, 0.00000064, 0.00000070, 0.00000281, 0.00000246, -0.00000070, -0.00000232, -0.00000127, 0.00000047, 0.00000158, 0.00000115, 0.00000002, 0.00000100, 0.00000231, -0.00000099, -0.00000502, -0.00000152, 0.00000418, 0.00000075, -0.00000605, -0.00000350, 0.00000271, 0.00000137, -0.00000018, 0.00000546, 0.00000538, -0.00000684, -0.00001257, -0.00000243, 0.00000503, -0.00000050, -0.00000419, 0.00000082, 0.00000342, 0.00000254, 0.00000578, 0.00000573, -0.00000504, -0.00001112, -0.00000091, 0.00000817, 0.00000105, -0.00000811, -0.00000396, 0.00000560, 0.00000877, 0.00000572, 0.00000010, -0.00000423, -0.00000347, -0.00000186, -0.00000422, -0.00000298, 0.00000316, -0.00000220, -0.00001517, -0.00000331, 0.00002782, 0.00002401, -0.00001862, -0.00002989, 0.00000998, 0.00002651, -0.00001417, -0.00004173, -0.00001010, 0.00002598, 0.00001592, -0.00001181, -0.00001850, -0.00000560, 0.00001697, 0.00003378, 0.00001438, -0.00002930, -0.00002989, 0.00002474, 0.00005190, 0.00001179, -0.00002183, 0.00000257, 0.00002648, 0.00000041, -0.00002813, -0.00001399, 0.00000766, 0.00000118, -0.00000888, -0.00000111, 0.00000308, -0.00000712, -0.00000948, 0.00000319, 0.00000983, 0.00000194, -0.00000596, -0.00000490, 0.00000043, 0.00000487, 0.00000534, 0.00000032, -0.00000488, -0.00000371, 0.00000065, 0.00000104, -0.00000146, -0.00000123, 0.00000207, 0.00000430, 0.00000215, -0.00000428, -0.00000987, -0.00000803, 0.00000044, 0.00000722, },
		{ 0.00000151, 0.00000700, 0.00000661, 0.00000265, -0.00000157, -0.00000371, -0.00000367, -0.00000337, -0.00000214, 0.00000259, 0.00000696, 0.00000329, -0.00000562, -0.00000808, -0.00000242, 0.00000155, 0.00000046, 0.00000037, 0.00000261, 0.00000241, -0.00000054, -0.00000222, -0.00000134, 0.00000070, 0.00000195, 0.00000072, -0.00000071, 0.00000147, 0.00000331, -0.00000134, -0.00000590, -0.00000099, 0.00000494, -0.00000030, -0.00000740, -0.00000286, 0.00000404, 0.00000150, -0.00000024, 0.00000598, 0.00000490, -0.00000845, -0.00001294, -0.00000112, 0.00000577, -0.00000103, -0.00000466, 0.00000094, 0.00000389, 0.00000335, 0.00000603, 0.00000445, -0.00000626, -0.00001025, 0.00000051, 0.00000761, -0.00000037, -0.00000788, -0.00000264, 0.00000609, 0.00000835, 0.00000526, -0.00000012, -0.00000431, -0.00000391, -0.00000263, -0.00000352, -0.00000061, 0.00000331, -0.00000547, -0.00001576, 0.00000134, 0.00002956, 0.00001851, -0.00002169, -0.00002458, 0.00001316, 0.00002044, -0.00001878, -0.00003743, -0.00000756, 0.00002005, 0.00001284, -0.00000584, -0.00001501, -0.00001036, 0.00001498, 0.00003925, 0.00001708, -0.00003247, -0.00002944, 0.00003001, 0.00005296, 0.00000883, -0.00002062, 0.00000586, 0.00002378, -0.00000508, -0.00002788, -0.00001044, 0.00000712, -0.00000177, -0.00000883, 0.00000039, 0.00000259, -0.00000769, -0.00000805, 0.00000438, 0.00000910, 0.00000094, -0.00000627, -0.00000523, 0.00000030, 0.00000543, 0.00000569, 0.00000001, -0.00000483, -0.00000314, 0.00000053, 0.00000010, -0.00000200, -0.00000076, 0.00000295, 0.00000457, 0.00000116, -0.00000580, -0.00001013, -0.00000671, },
		{ -0.00001009, -0.00000518, 0.00000240, 0.00000654, 0.00000621, 0.00000293, -0.00000145, -0.00000395, -0.00000374, -0.00000327, -0.00000196, 0.00000302, 0.00000709, 0.00000239, -0.00000665, -0.00000789, -0.00000147, 0.00000195, 0.00000025, 0.00000012, 0.00000237, 0.00000231, -0.00000030, -0.00000207, -0.00000146, 0.00000093, 0.00000231, 0.00000021, -0.00000139, 0.00000209, 0.00000423, -0.00000190, -0.00000671, -0.00000026, 0.00000558, -0.00000160, -0.00000862, -0.00000194, 0.00000535, 0.00000148, -0.00000031, 0.00000644, 0.00000424, -0.00000998, -0.00001298, 0.00000027, 0.00000629, -0.00000171, -0.00000505, 0.00000121, 0.00000445, 0.00000405, 0.00000597, 0.00000308, -0.00000712, -0.00000912, 0.00000167, 0.00000673, -0.00000167, -0.00000739, -0.00000128, 0.00000646, 0.00000780, 0.00000471, -0.00000025, -0.00000429, -0.00000450, -0.00000343, -0.00000243, 0.00000183, 0.00000279, -0.00000881, -0.00001529, 0.00000619, 0.00002974, 0.00001232, -0.00002318, -0.00001832, 0.00001485, 0.00001331, -0.00002202, -0.00003203, -0.00000605, 0.00001331, 0.00001046, 0.00000030, -0.00001234, -0.00001473, 0.00001419, 0.00004421, 0.00001852, -0.00003478, -0.00002731, 0.00003473, 0.00005257, 0.00000653, -0.00001819, 0.00000809, 0.00001935, -0.00001004, -0.00002613, -0.00000701, 0.00000565, -0.00000456, -0.00000821, 0.00000164, 0.00000173, -0.00000796, -0.00000629, 0.00000535, 0.00000810, -0.00000005, -0.00000645, -0.00000543, 0.00000022, 0.00000583, 0.00000590, -0.00000010, -0.00000457, -0.00000277, 0.00000013, -0.00000078, -0.00000231, -0.00000022, 0.00000379, 0.00000475, -0.00000001, -0.00000733, },
		{ -0.00000135, -0.00000880, -0.00000971, -0.00000351, 0.00000305, 0.00000587, 0.00000586, 0.00000329, -0.00000131, -0.00000416, -0.00000381, -0.00000319, -0.00000169, 0.00000352, 0.00000698, 0.00000125, -0.00000751, -0.00000739, -0.00000056, 0.00000214, 0.00000007, -0.00000005, 0.00000205, 0.00000217, 0.00000004, -0.00000190, -0.00000164, 0.00000117, 0.00000267, -0.00000034, -0.00000198, 0.00000284, 0.00000503, -0.00000265, -0.00000741, 0.00000065, 0.00000604, -0.00000311, -0.00000965, -0.00000075, 0.00000661, 0.00000132, -0.00000037, 0.00000680, 0.00000342, -0.00001139, -0.00001271, 0.00000167, 0.00000655, -0.00000251, -0.00000534, 0.00000163, 0.00000508, 0.00000459, 0.00000561, 0.00000172, -0.00000759, -0.00000782, 0.00000253, 0.00000561, -0.00000279, -0.00000667, 0.00000007, 0.00000673, 0.00000713, 0.00000409, -0.00000029, -0.00000419, -0.00000524, -0.00000418, -0.00000094, 0.00000416, 0.00000159, -0.00001197, -0.00001374, 0.00001089, 0.00002833, 0.00000591, -0.00002298, -0.00001160, 0.00001487, 0.00000560, -0.00002369, -0.00002593, -0.00000571, 0.00000612, 0.00000885, 0.00000622, -0.00001049, -0.00001837, 0.00001454, 0.00004834, 0.00001886, -0.00003593, -0.00002372, 0.00003861, 0.00005104, 0.00000514, -0.00001493, 0.00000887, 0.00001352, -0.00001407, -0.00002306, -0.00000399, 0.00000340, -0.00000698, -0.00000713, 0.00000252, 0.00000057, -0.00000787, -0.00000429, 0.00000606, 0.00000687, -0.00000099, -0.00000652, -0.00000553, 0.00000015, 0.00000605, 0.00000602, 0.00000001, -0.00000416, -0.00000263, -0.00000049, -0.00000153, -0.00000242, 0.00000034, 0.00000457, 0.00000479, },
		{ 0.00000529, 0.00000469, -0.00000285, -0.00001013, -0.00000899, -0.00000178, 0.00000344, 0.00000504, 0.00000558, 0.00000371, -0.00000115, -0.00000432, -0.00000392, -0.00000314, -0.00000128, 0.00000404, 0.00000657, -0.00000008, -0.00000809, -0.00000658, 0.00000024, 0.00000215, -0.00000004, -0.00000014, 0.00000162, 0.00000200, 0.00000049, -0.00000171, -0.00000188, 0.00000143, 0.00000302, -0.00000093, -0.00000249, 0.00000368, 0.00000567, -0.00000355, -0.00000793, 0.00000170, 0.00000627, -0.00000479, -0.00001044, 0.00000068, 0.00000776, 0.00000103, -0.00000043, 0.00000707, 0.00000246, -0.00001262, -0.00001215, 0.00000305, 0.00000655, -0.00000340, -0.00000549, 0.00000223, 0.00000573, 0.00000491, 0.00000497, 0.00000046, -0.00000767, -0.00000645, 0.00000303, 0.00000433, -0.00000369, -0.00000576, 0.00000138, 0.00000689, 0.00000635, 0.00000344, -0.00000022, -0.00000407, -0.00000612, -0.00000478, 0.00000089, 0.00000620, -0.00000024, -0.00001472, -0.00001119, 0.00001508, 0.00002540, -0.00000029, -0.00002110, -0.00000490, 0.00001320, -0.00000221, -0.00002375, -0.00001957, -0.00000656, -0.00000112, 0.00000797, 0.00001156, -0.00000939, -0.00002096, 0.00001588, 0.00005141, 0.00001833, -0.00003569, -0.00001896, 0.00004145, 0.00004871, 0.00000484, -0.00001128, 0.00000798, 0.00000671, -0.00001688, -0.00001897, -0.00000164, 0.00000057, -0.00000889, -0.00000574, 0.00000293, -0.00000074, -0.00000737, -0.00000216, 0.00000649, 0.00000548, -0.00000187, -0.00000648, -0.00000551, 0.00000008, 0.00000607, 0.00000612, 0.00000037, -0.00000369, -0.00000278, -0.00000127, -0.00000209, -0.00000236, 0.00000090, },
		{ -0.00000216, 0.00000143, 0.00000592, 0.00000442, -0.00000448, -0.00001126, -0.00000795, -0.00000009, 0.00000354, 0.00000414, 0.00000539, 0.00000416, -0.00000097, -0.00000445, -0.00000406, -0.00000308, -0.00000072, 0.00000453, 0.00000581, -0.00000156, -0.00000833, -0.00000553, 0.00000084, 0.00000198, -0.00000003, -0.00000019, 0.00000106, 0.00000184, 0.00000108, -0.00000153, -0.00000220, 0.00000170, 0.00000338, -0.00000156, -0.00000290, 0.00000458, 0.00000613, -0.00000456, -0.00000825, 0.00000285, 0.00000624, -0.00000659, -0.00001093, 0.00000232, 0.00000877, 0.00000061, -0.00000047, 0.00000722, 0.00000140, -0.00001365, -0.00001132, 0.00000435, 0.00000630, -0.00000433, -0.00000547, 0.00000298, 0.00000636, 0.00000496, 0.00000411, -0.00000060, -0.00000736, -0.00000510, 0.00000317, 0.00000297, -0.00000434, -0.00000472, 0.00000260, 0.00000693, 0.00000548, 0.00000277, -0.00000006, -0.00000398, -0.00000709, -0.00000514, 0.00000300, 0.00000780, -0.00000258, -0.00001681, -0.00000780, 0.00001845, 0.00002113, -0.00000584, -0.00001766, 0.00000130, 0.00000994, -0.00000964, -0.00002227, -0.00001339, -0.00000850, -0.00000806, 0.00000767, 0.00001601, -0.00000886, -0.00002226, 0.00001800, 0.00005324, 0.00001724, -0.00003396, -0.00001339, 0.00004319, 0.00004600, 0.00000563, -0.00000777, 0.00000530, -0.00000058, -0.00001826, -0.00001422, -0.00000010, -0.00000261, -0.00001019, -0.00000422, 0.00000286, -0.00000207, -0.00000647, -0.00000002, 0.00000662, 0.00000399, -0.00000266, -0.00000634, -0.00000539, -0.00000004, 0.00000592, 0.00000623, 0.00000096, -0.00000323, -0.00000321, -0.00000212, -0.00000244, },
		{ -0.00000296, -0.00000256, -0.00000186, 0.00000192, 0.00000646, 0.00000395, -0.00000617, -0.00001210, -0.00000663, 0.00000149, 0.00000338, 0.00000321, 0.00000529, 0.00000462, -0.00000076, -0.00000454, -0.00000426, -0.00000300, 0.00000001, 0.00000494, 0.00000467, -0.00000308, -0.00000815, -0.00000429, 0.00000118, 0.00000169, 0.00000015, -0.00000022, 0.00000035, 0.00000172, 0.00000181, -0.00000139, -0.00000260, 0.00000200, 0.00000373, -0.00000220, -0.00000321, 0.00000550, 0.00000640, -0.00000564, -0.00000835, 0.00000403, 0.00000592, -0.00000844, -0.00001109, 0.00000411, 0.00000959, 0.00000008, -0.00000051, 0.00000726, 0.00000027, -0.00001444, -0.00001027, 0.00000552, 0.00000582, -0.00000526, -0.00000527, 0.00000387, 0.00000691, 0.00000471, 0.00000308, -0.00000140, -0.00000673, -0.00000388, 0.00000298, 0.00000161, -0.00000472, -0.00000360, 0.00000371, 0.00000688, 0.00000455, 0.00000212, 0.00000018, -0.00000394, -0.00000809, -0.00000519, 0.00000527, 0.00000883, -0.00000527, -0.00001809, -0.00000380, 0.00002073, 0.00001578, -0.00001037, -0.00001295, 0.00000658, 0.00000530, -0.00001626, -0.00001945, -0.00000776, -0.00001135, -0.00001438, 0.00000775, 0.00001931, -0.00000868, -0.00002212, 0.00002063, 0.00005381, 0.00001591, -0.00003073, -0.00000736, 0.00004390, 0.00004330, 0.00000737, -0.00000490, 0.00000091, -0.00000782, -0.00001817, -0.00000919, 0.00000055, -0.00000586, -0.00001085, -0.00000276, 0.00000230, -0.00000329, -0.00000519, 0.00000204, 0.00000646, 0.00000247, -0.00000336, -0.00000612, -0.00000517, -0.00000022, 0.00000560, 0.00000641, 0.00000176, -0.00000287, -0.00000392, },
		{ -0.00000268, -0.00000484, -0.00000369, -0.00000246, -0.00000152, 0.00000236, 0.00000687, 0.00000326, -0.00000788, -0.00001260, -0.00000509, 0.00000285, 0.00000299, 0.00000232, 0.00000526, 0.00000505, -0.00000052, -0.00000460, -0.00000453, -0.00000284, 0.00000092, 0.00000518, 0.00000313, -0.00000454, -0.00000753, -0.00000296, 0.00000121, 0.00000134, 0.00000051, -0.00000029, -0.00000052, 0.00000168, 0.00000267, -0.00000132, -0.00000308, 0.00000235, 0.00000409, -0.00000286, -0.00000343, 0.00000641, 0.00000647, -0.00000671, -0.00000822, 0.00000518, 0.00000531, -0.00001027, -0.00001090, 0.00000600, 0.00001021, -0.00000055, -0.00000053, 0.00000718, -0.00000090, -0.00001499, -0.00000903, 0.00000654, 0.00000511, -0.00000616, -0.00000486, 0.00000488, 0.00000731, 0.00000415, 0.00000198, -0.00000187, -0.00000582, -0.00000286, 0.00000249, 0.00000035, -0.00000485, -0.00000245, 0.00000470, 0.00000674, 0.00000358, 0.00000149, 0.00000045, -0.00000400, -0.00000906, -0.00000489, 0.00000755, 0.00000922, -0.00000813, -0.00001843, 0.00000055, 0.00002175, 0.00000972, -0.00001360, -0.00000731, 0.00001059, -0.00000037, -0.00002173, -0.00001556, -0.00000298, -0.00001486, -0.00001985, 0.00000795, 0.00002130, -0.00000861, -0.00002049, 0.00002353, 0.00005315, 0.00001466, -0.00002612, -0.00000121, 0.00004375, 0.00004096, 0.00000980, -0.00000313, -0.00000497, -0.00001450, -0.00001668, -0.00000428, 0.00000034, -0.00000891, -0.00001092, -0.00000153, 0.00000133, -0.00000427, -0.00000360, 0.00000391, 0.00000607, 0.00000098, -0.00000398, -0.00000582, -0.00000488, -0.00000047, 0.00000515, 0.00000670, 0.00000270, },
		{ 0.00000712, 0.00000371, -0.00000270, -0.00000592, -0.00000426, -0.00000216, -0.00000115, 0.00000276, 0.00000714, 0.00000235, -0.00000951, -0.00001271, -0.00000340, 0.00000394, 0.00000241, 0.00000154, 0.00000528, 0.00000544, -0.00000022, -0.00000463, -0.00000487, -0.00000259, 0.00000198, 0.00000518, 0.00000121, -0.00000585, -0.00000643, -0.00000164, 0.00000091, 0.00000098, 0.00000108, -0.00000044, -0.00000155, 0.00000178, 0.00000366, -0.00000136, -0.00000363, 0.00000276, 0.00000445, -0.00000352, -0.00000356, 0.00000729, 0.00000635, -0.00000775, -0.00000786, 0.00000625, 0.00000441, -0.00001201, -0.00001036, 0.00000794, 0.00001060, -0.00000125, -0.00000055, 0.00000698, -0.00000206, -0.00001528, -0.00000765, 0.00000736, 0.00000421, -0.00000699, -0.00000425, 0.00000596, 0.00000753, 0.00000330, 0.00000087, -0.00000199, -0.00000473, -0.00000210, 0.00000176, -0.00000077, -0.00000476, -0.00000133, 0.00000556, 0.00000651, 0.00000258, 0.00000092, 0.00000073, -0.00000414, -0.00000990, -0.00000421, 0.00000972, 0.00000892, -0.00001093, -0.00001779, 0.00000493, 0.00002143, 0.00000334, -0.00001538, -0.00000116, 0.00001311, -0.00000670, -0.00002580, -0.00001095, 0.00000073, -0.00001871, -0.00002434, 0.00000798, 0.00002194, -0.00000840, -0.00001743, 0.00002643, 0.00005144, 0.00001378, -0.00002032, 0.00000479, 0.00004301, 0.00003922, 0.00001252, -0.00000284, -0.00001198, -0.00002019, -0.00001400, 0.00000016, -0.00000060, -0.00001154, -0.00001049, -0.00000068, 0.00000005, -0.00000491, -0.00000181, 0.00000551, 0.00000549, -0.00000044, -0.00000452, -0.00000545, -0.00000452, -0.00000082, 0.00000461, },
		{ -0.00000126, 0.00000404, 0.00000768, 0.00000470, -0.00000297, -0.00000707, -0.00000462, -0.00000172, -0.00000078, 0.00000311, 0.00000721, 0.00000121, -0.00001099, -0.00001241, -0.00000166, 0.00000471, 0.00000172, 0.00000089, 0.00000532, 0.00000577, 0.00000015, -0.00000465, -0.00000528, -0.00000220, 0.00000317, 0.00000487, -0.00000106, -0.00000688, -0.00000488, -0.00000043, 0.00000027, 0.00000071, 0.00000184, -0.00000074, -0.00000270, 0.00000204, 0.00000476, -0.00000153, -0.00000424, 0.00000326, 0.00000481, -0.00000419, -0.00000361, 0.00000809, 0.00000607, -0.00000870, -0.00000731, 0.00000717, 0.00000325, -0.00001360, -0.00000946, 0.00000985, 0.00001075, -0.00000200, -0.00000056, 0.00000668, -0.00000318, -0.00001531, -0.00000619, 0.00000797, 0.00000315, -0.00000770, -0.00000343, 0.00000707, 0.00000750, 0.00000218, -0.00000014, -0.00000174, -0.00000355, -0.00000165, 0.00000086, -0.00000168, -0.00000447, -0.00000027, 0.00000628, 0.00000622, 0.00000159, 0.00000040, 0.00000099, -0.00000438, -0.00001054, -0.00000318, 0.00001159, 0.00000795, -0.00001346, -0.00001619, 0.00000903, 0.00001977, -0.00000294, -0.00001564, 0.00000505, 0.00001403, -0.00001323, -0.00002833, -0.00000602, 0.00000326, -0.00002258, -0.00002780, 0.00000757, 0.00002124, -0.00000784, -0.00001308, 0.00002913, 0.00004892, 0.00001348, -0.00001363, 0.00001046, 0.00004205, 0.00003822, 0.00001504, -0.00000427, -0.00001966, -0.00002453, -0.00001044, 0.00000382, -0.00000207, -0.00001354, -0.00000973, -0.00000031, -0.00000140, -0.00000515, 0.00000007, 0.00000679, 0.00000479, -0.00000174, -0.00000500, -0.00000503, -0.00000413, },
		{ -0.00000457, -0.00000372, -0.00000177, 0.00000350, 0.00000835, 0.00000559, -0.00000349, -0.00000819, -0.00000475, -0.00000119, -0.00000043, 0.00000342, 0.00000707, -0.00000013, -0.00001223, -0.00001169, 0.00000001, 0.00000513, 0.00000100, 0.00000042, 0.00000532, 0.00000603, 0.00000060, -0.00000467, -0.00000575, -0.00000165, 0.00000445, 0.00000418, -0.00000360, -0.00000754, -0.00000292, 0.00000056, -0.00000069, 0.00000059, 0.00000277, -0.00000122, -0.00000394, 0.00000251, 0.00000592, -0.00000187, -0.00000487, 0.00000386, 0.00000516, -0.00000488, -0.00000361, 0.00000882, 0.00000565, -0.00000952, -0.00000659, 0.00000791, 0.00000187, -0.00001497, -0.00000825, 0.00001169, 0.00001067, -0.00000279, -0.00000057, 0.00000629, -0.00000422, -0.00001510, -0.00000470, 0.00000836, 0.00000197, -0.00000827, -0.00000241, 0.00000816, 0.00000720, 0.00000086, -0.00000097, -0.00000115, -0.00000238, -0.00000152, -0.00000013, -0.00000235, -0.00000404, 0.00000069, 0.00000687, 0.00000588, 0.00000061, -0.00000008, 0.00000117, -0.00000467, -0.00001088, -0.00000186, 0.00001304, 0.00000638, -0.00001551, -0.00001374, 0.00001255, 0.00001691, -0.00000873, -0.00001446, 0.00001086, 0.00001337, -0.00001952, -0.00002931, -0.00000115, 0.00000462, -0.00002615, -0.00003028, 0.00000647, 0.00001932, -0.00000674, -0.00000765, 0.00003152, 0.00004590, 0.00001390, -0.00000637, 0.00001573, 0.00004121, 0.00003795, 0.00001684, -0.00000750, -0.00002745, -0.00002730, -0.00000637, 0.00000652, -0.00000382, -0.00001482, -0.00000882, -0.00000048, -0.00000286, -0.00000497, 0.00000192, 0.00000773, 0.00000404, -0.00000291, -0.00000542, },
		{ -0.00000393, -0.00000580, -0.00000407, -0.00000333, -0.00000234, 0.00000302, 0.00000909, 0.00000629, -0.00000421, -0.00000918, -0.00000464, -0.00000062, -0.00000009, 0.00000369, 0.00000665, -0.00000164, -0.00001312, -0.00001058, 0.00000151, 0.00000519, 0.00000034, 0.00000013, 0.00000523, 0.00000622, 0.00000116, -0.00000471, -0.00000626, -0.00000091, 0.00000575, 0.00000306, -0.00000634, -0.00000772, -0.00000061, 0.00000124, -0.00000192, 0.00000070, 0.00000383, -0.00000193, -0.00000523, 0.00000322, 0.00000710, -0.00000240, -0.00000550, 0.00000457, 0.00000548, -0.00000558, -0.00000356, 0.00000946, 0.00000513, -0.00001020, -0.00000575, 0.00000842, 0.00000032, -0.00001606, -0.00000675, 0.00001340, 0.00001035, -0.00000359, -0.00000058, 0.00000583, -0.00000516, -0.00001467, -0.00000321, 0.00000853, 0.00000071, -0.00000866, -0.00000121, 0.00000916, 0.00000660, -0.00000061, -0.00000156, -0.00000026, -0.00000131, -0.00000168, -0.00000112, -0.00000277, -0.00000351, 0.00000153, 0.00000733, 0.00000549, -0.00000032, -0.00000051, 0.00000127, -0.00000498, -0.00001087, -0.00000031, 0.00001394, 0.00000432, -0.00001692, -0.00001059, 0.00001525, 0.00001303, -0.00001366, -0.00001201, 0.00001587, 0.00001127, -0.00002518, -0.00002885, 0.00000331, 0.00000489, -0.00002917, -0.00003190, 0.00000451, 0.00001637, -0.00000503, -0.00000141, 0.00003354, 0.00004271, 0.00001507, 0.00000112, 0.00002063, 0.00004087, 0.00003824, 0.00001740, -0.00001244, -0.00003479, -0.00002841, -0.00000223, 0.00000815, -0.00000558, -0.00001531, -0.00000796, -0.00000117, -0.00000419, -0.00000439, 0.00000361, 0.00000833, 0.00000330, },
		{ 0.00000863, 0.00000264, -0.00000481, -0.00000614, -0.00000354, -0.00000296, -0.00000294, 0.00000266, 0.00000985, 0.00000671, -0.00000509, -0.00000994, -0.00000432, -0.00000006, 0.00000025, 0.00000390, 0.00000592, -0.00000328, -0.00001357, -0.00000912, 0.00000273, 0.00000492, -0.00000017, 0.00000001, 0.00000502, 0.00000637, 0.00000183, -0.00000479, -0.00000679, 0.00000005, 0.00000700, 0.00000148, -0.00000916, -0.00000737, 0.00000194, 0.00000153, -0.00000335, 0.00000108, 0.00000496, -0.00000291, -0.00000651, 0.00000418, 0.00000826, -0.00000312, -0.00000609, 0.00000540, 0.00000576, -0.00000631, -0.00000346, 0.00001000, 0.00000452, -0.00001072, -0.00000483, 0.00000868, -0.00000133, -0.00001683, -0.00000501, 0.00001492, 0.00000981, -0.00000438, -0.00000059, 0.00000531, -0.00000598, -0.00001403, -0.00000177, 0.00000847, -0.00000059, -0.00000884, 0.00000013, 0.00001002, 0.00000571, -0.00000214, -0.00000186, 0.00000086, -0.00000042, -0.00000212, -0.00000203, -0.00000294, -0.00000294, 0.00000224, 0.00000768, 0.00000507, -0.00000120, -0.00000091, 0.00000125, -0.00000526, -0.00001046, 0.00000134, 0.00001419, 0.00000192, -0.00001755, -0.00000698, 0.00001694, 0.00000842, -0.00001747, -0.00000858, 0.00001975, 0.00000797, -0.00002985, -0.00002715, 0.00000705, 0.00000425, -0.00003141, -0.00003283, 0.00000159, 0.00001260, -0.00000270, 0.00000535, 0.00003524, 0.00003971, 0.00001695, 0.00000855, 0.00002532, 0.00004130, 0.00003880, 0.00001626, -0.00001886, -0.00004111, -0.00002795, 0.00000155, 0.00000872, -0.00000705, -0.00001506, -0.00000733, -0.00000233, -0.00000523, -0.00000347, 0.00000504, },
		{ -0.00000230, 0.00000613, 0.00000867, 0.00000209, -0.00000555, -0.00000645, -0.00000302, -0.00000264, -0.00000350, 0.00000245, 0.00001054, 0.00000681, -0.00000603, -0.00001041, -0.00000385, 0.00000044, 0.00000062, 0.00000404, 0.00000484, -0.00000496, -0.00001352, -0.00000740, 0.00000356, 0.00000438, -0.00000046, 0.00000002, 0.00000463, 0.00000650, 0.00000262, -0.00000492, -0.00000731, 0.00000121, 0.00000813, -0.00000057, -0.00001195, -0.00000643, 0.00000461, 0.00000136, -0.00000490, 0.00000179, 0.00000608, -0.00000416, -0.00000770, 0.00000542, 0.00000932, -0.00000405, -0.00000660, 0.00000635, 0.00000598, -0.00000706, -0.00000333, 0.00001045, 0.00000388, -0.00001107, -0.00000389, 0.00000869, -0.00000303, -0.00001726, -0.00000309, 0.00001622, 0.00000907, -0.00000514, -0.00000060, 0.00000475, -0.00000666, -0.00001323, -0.00000043, 0.00000820, -0.00000187, -0.00000880, 0.00000157, 0.00001069, 0.00000454, -0.00000364, -0.00000184, 0.00000212, 0.00000024, -0.00000276, -0.00000281, -0.00000289, -0.00000238, 0.00000280, 0.00000792, 0.00000462, -0.00000202, -0.00000131, 0.00000111, -0.00000543, -0.00000963, 0.00000298, 0.00001376, -0.00000064, -0.00001733, -0.00000315, 0.00001751, 0.00000341, -0.00001995, -0.00000450, 0.00002224, 0.00000378, -0.00003328, -0.00002447, 0.00000987, 0.00000292, -0.00003276, -0.00003331, -0.00000232, 0.00000825, 0.00000018, 0.00001234, 0.00003674, 0.00003719, 0.00001939, 0.00001567, 0.00003002, 0.00004270, 0.00003920, 0.00001307, -0.00002633, -0.00004595, -0.00002612, 0.00000460, 0.00000834, -0.00000801, -0.00001418, -0.00000711, -0.00000385, -0.00000587, },
		{ -0.00000557, -0.00000605, -0.00000101, 0.00000682, 0.00000851, 0.00000169, -0.00000617, -0.00000671, -0.00000251, -0.00000239, -0.00000399, 0.00000239, 0.00001106, 0.00000655, -0.00000693, -0.00001053, -0.00000328, 0.00000084, 0.00000104, 0.00000406, 0.00000338, -0.00000659, -0.00001290, -0.00000552, 0.00000392, 0.00000365, -0.00000044, 0.00000011, 0.00000403, 0.00000664, 0.00000353, -0.00000514, -0.00000780, 0.00000257, 0.00000906, -0.00000308, -0.00001457, -0.00000490, 0.00000729, 0.00000070, -0.00000645, 0.00000283, 0.00000711, -0.00000569, -0.00000873, 0.00000690, 0.00001023, -0.00000517, -0.00000698, 0.00000741, 0.00000611, -0.00000785, -0.00000317, 0.00001081, 0.00000322, -0.00001126, -0.00000298, 0.00000845, -0.00000471, -0.00001733, -0.00000105, 0.00001726, 0.00000815, -0.00000586, -0.00000061, 0.00000417, -0.00000719, -0.00001229, 0.00000079, 0.00000773, -0.00000310, -0.00000854, 0.00000307, 0.00001110, 0.00000312, -0.00000503, -0.00000148, 0.00000344, 0.00000061, -0.00000354, -0.00000339, -0.00000265, -0.00000185, 0.00000324, 0.00000806, 0.00000415, -0.00000278, -0.00000172, 0.00000086, -0.00000544, -0.00000839, 0.00000447, 0.00001263, -0.00000318, -0.00001627, 0.00000063, 0.00001695, -0.00000164, -0.00002103, -0.00000016, 0.00002321, -0.00000092, -0.00003532, -0.00002115, 0.00001163, 0.00000116, -0.00003317, -0.00003354, -0.00000714, 0.00000354, 0.00000343, 0.00001931, 0.00003825, 0.00003540, 0.00002220, 0.00002235, 0.00003502, 0.00004512, 0.00003893, 0.00000762, -0.00003433, -0.00004893, -0.00002325, 0.00000660, 0.00000723, -0.00000824, -0.00001283, -0.00000741, },
		{ -0.00001122, -0.00000827, -0.00000732, -0.00000573, 0.00000029, 0.00000711, 0.00000823, 0.00000143, -0.00000668, -0.00000692, -0.00000203, -0.00000220, -0.00000433, 0.00000250, 0.00001131, 0.00000592, -0.00000768, -0.00001027, -0.00000272, 0.00000114, 0.00000156, 0.00000393, 0.00000153, -0.00000808, -0.00001167, -0.00000359, 0.00000376, 0.00000282, -0.00000007, 0.00000022, 0.00000320, 0.00000683, 0.00000456, -0.00000544, -0.00000822, 0.00000411, 0.00000970, -0.00000600, -0.00001690, -0.00000278, 0.00000982, -0.00000044, -0.00000792, 0.00000420, 0.00000797, -0.00000746, -0.00000953, 0.00000861, 0.00001093, -0.00000646, -0.00000719, 0.00000857, 0.00000612, -0.00000867, -0.00000298, 0.00001110, 0.00000256, -0.00001131, -0.00000212, 0.00000797, -0.00000629, -0.00001705, 0.00000104, 0.00001802, 0.00000708, -0.00000653, -0.00000062, 0.00000360, -0.00000756, -0.00001125, 0.00000187, 0.00000710, -0.00000422, -0.00000804, 0.00000455, 0.00001122, 0.00000152, -0.00000622, -0.00000083, 0.00000471, 0.00000069, -0.00000438, -0.00000375, -0.00000228, -0.00000139, 0.00000356, 0.00000812, 0.00000367, -0.00000346, -0.00000216, 0.00000052, -0.00000521, -0.00000679, 0.00000569, 0.00001085, -0.00000550, -0.00001443, 0.00000410, 0.00001531, -0.00000639, -0.00002072, 0.00000405, 0.00002266, -0.00000575, -0.00003592, -0.00001749, 0.00001231, -0.00000077, -0.00003270, -0.00003373, -0.00001274, -0.00000135, 0.00000685, 0.00002608, 0.00004001, 0.00003451, 0.00002515, 0.00002855, 0.00004062, 0.00004846, 0.00003745, -0.00000012, -0.00004227, -0.00004983, -0.00001977, 0.00000733, 0.00000565, -0.00000763, },
		{ 0.00000390, -0.00000616, -0.00000958, -0.00000968, -0.00000894, -0.00000495, 0.00000146, 0.00000699, 0.00000788, 0.00000129, -0.00000709, -0.00000707, -0.00000163, -0.00000208, -0.00000448, 0.00000272, 0.00001119, 0.00000495, -0.00000814, -0.00000964, -0.00000224, 0.00000135, 0.00000219, 0.00000360, -0.00000071, -0.00000930, -0.00000983, -0.00000177, 0.00000305, 0.00000200, 0.00000067, 0.00000026, 0.00000212, 0.00000713, 0.00000569, -0.00000584, -0.00000854, 0.00000579, 0.00000999, -0.00000927, -0.00001882, -0.00000014, 0.00001209, -0.00000205, -0.00000918, 0.00000586, 0.00000858, -0.00000945, -0.00001003, 0.00001049, 0.00001136, -0.00000787, -0.00000719, 0.00000980, 0.00000598, -0.00000950, -0.00000274, 0.00001134, 0.00000193, -0.00001124, -0.00000137, 0.00000730, -0.00000773, -0.00001642, 0.00000312, 0.00001848, 0.00000589, -0.00000712, -0.00000061, 0.00000304, -0.00000779, -0.00001015, 0.00000277, 0.00000634, -0.00000521, -0.00000734, 0.00000597, 0.00001104, -0.00000019, -0.00000714, 0.00000009, 0.00000584, 0.00000049, -0.00000521, -0.00000387, -0.00000182, -0.00000101, 0.00000377, 0.00000809, 0.00000319, -0.00000408, -0.00000264, 0.00000013, -0.00000470, -0.00000489, 0.00000651, 0.00000851, -0.00000744, -0.00001192, 0.00000703, 0.00001275, -0.00001052, -0.00001918, 0.00000777, 0.00002067, -0.00001035, -0.00003516, -0.00001384, 0.00001198, -0.00000265, -0.00003145, -0.00003404, -0.00001892, -0.00000630, 0.00001017, 0.00003253, 0.00004229, 0.00003461, 0.00002806, 0.00003436, 0.00004710, 0.00005243, 0.00003421, -0.00000997, -0.00004949, -0.00004858, -0.00001614, 0.00000669, },
		{ -0.00001283, 0.00000474, 0.00000231, -0.00000387, -0.00000812, -0.00001157, -0.00001026, -0.00000378, 0.00000240, 0.00000654, 0.00000752, 0.00000127, -0.00000738, -0.00000714, -0.00000133, -0.00000202, -0.00000436, 0.00000301, 0.00001061, 0.00000371, -0.00000820, -0.00000867, -0.00000193, 0.00000149, 0.00000296, 0.00000299, -0.00000331, -0.00001015, -0.00000740, -0.00000018, 0.00000180, 0.00000133, 0.00000179, 0.00000017, 0.00000082, 0.00000756, 0.00000690, -0.00000636, -0.00000873, 0.00000756, 0.00000986, -0.00001279, -0.00002020, 0.00000294, 0.00001396, -0.00000405, -0.00001014, 0.00000777, 0.00000887, -0.00001157, -0.00001018, 0.00001249, 0.00001148, -0.00000936, -0.00000695, 0.00001106, 0.00000568, -0.00001035, -0.00000245, 0.00001152, 0.00000133, -0.00001109, -0.00000074, 0.00000646, -0.00000897, -0.00001548, 0.00000514, 0.00001863, 0.00000462, -0.00000763, -0.00000060, 0.00000252, -0.00000787, -0.00000901, 0.00000350, 0.00000547, -0.00000603, -0.00000645, 0.00000726, 0.00001054, -0.00000194, -0.00000773, 0.00000120, 0.00000675, 0.00000004, -0.00000593, -0.00000377, -0.00000132, -0.00000073, 0.00000389, 0.00000798, 0.00000269, -0.00000465, -0.00000317, -0.00000026, -0.00000387, -0.00000279, 0.00000685, 0.00000573, -0.00000887, -0.00000892, 0.00000925, 0.00000949, -0.00001378, -0.00001660, 0.00001069, 0.00001744, -0.00001440, -0.00003318, -0.00001045, 0.00001077, -0.00000429, -0.00002961, -0.00003456, -0.00002544, -0.00001123, 0.00001310, 0.00003861, 0.00004533, 0.00003568, 0.00003074, 0.00003999, 0.00005463, 0.00005660, 0.00002878, -0.00002158, -0.00005540, -0.00004531, },
		{ -0.00005945, -0.00004030, -0.00001029, 0.00000162, 0.00000114, -0.00000093, -0.00000704, -0.00001379, -0.00001115, -0.00000233, 0.00000303, 0.00000583, 0.00000717, 0.00000130, -0.00000754, -0.00000712, -0.00000118, -0.00000197, -0.00000394, 0.00000330, 0.00000948, 0.00000230, -0.00000776, -0.00000744, -0.00000186, 0.00000160, 0.00000387, 0.00000204, -0.00000621, -0.00001049, -0.00000446, 0.00000102, 0.00000004, 0.00000092, 0.00000325, -0.00000016, -0.00000067, 0.00000819, 0.00000815, -0.00000698, -0.00000877, 0.00000935, 0.00000928, -0.00001645, -0.00002098, 0.00000636, 0.00001534, -0.00000637, -0.00001073, 0.00000984, 0.00000878, -0.00001376, -0.00000994, 0.00001452, 0.00001126, -0.00001087, -0.00000644, 0.00001231, 0.00000519, -0.00001119, -0.00000210, 0.00001165, 0.00000077, -0.00001086, -0.00000025, 0.00000551, -0.00000996, -0.00001427, 0.00000702, 0.00001849, 0.00000329, -0.00000805, -0.00000057, 0.00000204, -0.00000782, -0.00000788, 0.00000404, 0.00000453, -0.00000664, -0.00000540, 0.00000835, 0.00000974, -0.00000364, -0.00000798, 0.00000243, 0.00000738, -0.00000061, -0.00000649, -0.00000346, -0.00000083, -0.00000054, 0.00000395, 0.00000780, 0.00000219, -0.00000518, -0.00000374, -0.00000061, -0.00000270, -0.00000060, 0.00000666, 0.00000268, -0.00000972, -0.00000563, 0.00001063, 0.00000580, -0.00001598, -0.00001329, 0.00001259, 0.00001324, -0.00001765, -0.00003023, -0.00000753, 0.00000889, -0.00000559, -0.00002738, -0.00003533, -0.00003204, -0.00001618, 0.00001539, 0.00004435, 0.00004935, 0.00003764, 0.00003312, 0.00004572, 0.00006328, 0.00006041, 0.00002080, -0.00003440, },
		{ 0.00001014, -0.00004777, -0.00006125, -0.00003398, -0.00000885, -0.00000239, 0.00000062, 0.00000246, -0.00000645, -0.00001615, -0.00001155, -0.00000077, 0.00000332, 0.00000494, 0.00000683, 0.00000135, -0.00000755, -0.00000701, -0.00000120, -0.00000191, -0.00000318, 0.00000349, 0.00000776, 0.00000084, -0.00000673, -0.00000605, -0.00000210, 0.00000176, 0.00000492, 0.00000070, -0.00000935, -0.00001024, -0.00000109, 0.00000172, -0.00000214, 0.00000088, 0.00000500, -0.00000078, -0.00000230, 0.00000904, 0.00000941, -0.00000770, -0.00000864, 0.00001111, 0.00000823, -0.00002012, -0.00002110, 0.00000997, 0.00001613, -0.00000890, -0.00001088, 0.00001198, 0.00000829, -0.00001592, -0.00000928, 0.00001652, 0.00001068, -0.00001232, -0.00000566, 0.00001350, 0.00000450, -0.00001199, -0.00000167, 0.00001175, 0.00000023, -0.00001060, 0.00000009, 0.00000450, -0.00001069, -0.00001282, 0.00000873, 0.00001805, 0.00000194, -0.00000836, -0.00000052, 0.00000161, -0.00000766, -0.00000677, 0.00000440, 0.00000357, -0.00000704, -0.00000425, 0.00000920, 0.00000868, -0.00000519, -0.00000786, 0.00000367, 0.00000769, -0.00000138, -0.00000684, -0.00000299, -0.00000039, -0.00000042, 0.00000396, 0.00000753, 0.00000167, -0.00000568, -0.00000435, -0.00000084, -0.00000120, 0.00000155, 0.00000591, -0.00000049, -0.00000994, -0.00000228, 0.00001113, 0.00000196, -0.00001703, -0.00000957, 0.00001332, 0.00000841, -0.00001994, -0.00002658, -0.00000522, 0.00000655, -0.00000649, -0.00002499, -0.00003629, -0.00003849, -0.00002121, 0.00001681, 0.00004985, 0.00005447, 0.00004036, 0.00003524, 0.00005188, 0.00007293, 0.00006319, },
		{ 0.00008332, 0.00006423, -0.00000317, -0.00006089, -0.00006057, -0.00002688, -0.00000876, -0.00000694, 0.00000089, 0.00000602, -0.00000642, -0.00001844, -0.00001143, 0.00000077, 0.00000328, 0.00000398, 0.00000647, 0.00000139, -0.00000735, -0.00000681, -0.00000145, -0.00000178, -0.00000206, 0.00000347, 0.00000542, -0.00000054, -0.00000508, -0.00000460, -0.00000269, 0.00000202, 0.00000607, -0.00000111, -0.00001261, -0.00000931, 0.00000257, 0.00000180, -0.00000463, 0.00000131, 0.00000694, -0.00000177, -0.00000400, 0.00001015, 0.00001064, -0.00000852, -0.00000833, 0.00001275, 0.00000672, -0.00002367, -0.00002052, 0.00001364, 0.00001631, -0.00001151, -0.00001055, 0.00001409, 0.00000737, -0.00001795, -0.00000820, 0.00001838, 0.00000975, -0.00001365, -0.00000462, 0.00001456, 0.00000362, -0.00001273, -0.00000116, 0.00001181, -0.00000028, -0.00001033, 0.00000029, 0.00000346, -0.00001113, -0.00001119, 0.00001021, 0.00001735, 0.00000061, -0.00000857, -0.00000044, 0.00000124, -0.00000739, -0.00000572, 0.00000458, 0.00000261, -0.00000722, -0.00000303, 0.00000977, 0.00000739, -0.00000652, -0.00000740, 0.00000483, 0.00000766, -0.00000221, -0.00000695, -0.00000240, -0.00000002, -0.00000036, 0.00000392, 0.00000719, 0.00000114, -0.00000616, -0.00000498, -0.00000090, 0.00000059, 0.00000354, 0.00000463, -0.00000359, -0.00000955, 0.00000091, 0.00001078, -0.00000171, -0.00001695, -0.00000579, 0.00001286, 0.00000329, -0.00002122, -0.00002254, -0.00000355, 0.00000399, -0.00000704, -0.00002263, -0.00003735, -0.00004459, -0.00002647, 0.00001721, 0.00005522, 0.00006074, 0.00004368, 0.00003726, 0.00005883, },
		{ 0.00003948, 0.00006687, 0.00009393, 0.00006286, -0.00001883, -0.00007295, -0.00005739, -0.00001963, -0.00001011, -0.00001163, 0.00000196, 0.00000944, -0.00000694, -0.00002043, -0.00001084, 0.00000210, 0.00000294, 0.00000303, 0.00000605, 0.00000140, -0.00000690, -0.00000655, -0.00000194, -0.00000153, -0.00000058, 0.00000313, 0.00000248, -0.00000167, -0.00000279, -0.00000324, -0.00000361, 0.00000246, 0.00000726, -0.00000340, -0.00001587, -0.00000765, 0.00000638, 0.00000119, -0.00000730, 0.00000228, 0.00000898, -0.00000316, -0.00000570, 0.00001152, 0.00001178, -0.00000942, -0.00000785, 0.00001422, 0.00000476, -0.00002697, -0.00001925, 0.00001720, 0.00001584, -0.00001408, -0.00000974, 0.00001606, 0.00000604, -0.00001975, -0.00000673, 0.00002003, 0.00000849, -0.00001479, -0.00000332, 0.00001545, 0.00000255, -0.00001336, -0.00000056, 0.00001182, -0.00000078, -0.00001006, 0.00000036, 0.00000245, -0.00001128, -0.00000943, 0.00001144, 0.00001639, -0.00000069, -0.00000865, -0.00000035, 0.00000091, -0.00000704, -0.00000474, 0.00000460, 0.00000170, -0.00000717, -0.00000182, 0.00001002, 0.00000596, -0.00000755, -0.00000664, 0.00000583, 0.00000730, -0.00000300, -0.00000680, -0.00000174, 0.00000024, -0.00000034, 0.00000385, 0.00000677, 0.00000058, -0.00000663, -0.00000557, -0.00000073, 0.00000261, 0.00000524, 0.00000288, -0.00000646, -0.00000864, 0.00000372, 0.00000968, -0.00000494, -0.00001583, -0.00000227, 0.00001127, -0.00000177, -0.00002151, -0.00001839, -0.00000250, 0.00000141, -0.00000732, -0.00002046, -0.00003836, -0.00005022, -0.00003210, 0.00001650, 0.00006062, 0.00006812, 0.00004748, },
		{ 0.00007644, 0.00005170, 0.00004233, 0.00007619, 0.00010412, 0.00005847, -0.00003632, -0.00008316, -0.00005187, -0.00001282, -0.00001284, -0.00001603, 0.00000376, 0.00001244, -0.00000789, -0.00002192, -0.00000989, 0.00000311, 0.00000242, 0.00000217, 0.00000549, 0.00000135, -0.00000614, -0.00000625, -0.00000271, -0.00000109, 0.00000122, 0.00000236, -0.00000100, -0.00000239, 0.00000010, -0.00000211, -0.00000486, 0.00000315, 0.00000843, -0.00000621, -0.00001899, -0.00000523, 0.00001015, -0.00000014, -0.00000999, 0.00000381, 0.00001098, -0.00000496, -0.00000730, 0.00001315, 0.00001280, -0.00001037, -0.00000720, 0.00001543, 0.00000242, -0.00002988, -0.00001735, 0.00002051, 0.00001476, -0.00001647, -0.00000848, 0.00001778, 0.00000435, -0.00002124, -0.00000491, 0.00002138, 0.00000693, -0.00001567, -0.00000182, 0.00001612, 0.00000131, -0.00001386, 0.00000011, 0.00001177, -0.00000128, -0.00000980, 0.00000032, 0.00000151, -0.00001115, -0.00000760, 0.00001239, 0.00001522, -0.00000192, -0.00000862, -0.00000023, 0.00000064, -0.00000663, -0.00000385, 0.00000448, 0.00000087, -0.00000691, -0.00000066, 0.00000995, 0.00000444, -0.00000823, -0.00000564, 0.00000659, 0.00000665, -0.00000370, -0.00000640, -0.00000108, 0.00000041, -0.00000033, 0.00000375, 0.00000625, -0.00000000, -0.00000710, -0.00000610, -0.00000030, 0.00000478, 0.00000656, 0.00000076, -0.00000896, -0.00000733, 0.00000599, 0.00000801, -0.00000751, -0.00001385, 0.00000068, 0.00000873, -0.00000647, -0.00002094, -0.00001441, -0.00000197, -0.00000100, -0.00000746, -0.00001859, -0.00003913, -0.00005529, -0.00003828, 0.00001470, 0.00006615, },
		{ 0.00001187, 0.00007186, 0.00008548, 0.00005635, 0.00004632, 0.00008685, 0.00011303, 0.00005063, -0.00005492, -0.00009082, -0.00004441, -0.00000705, -0.00001671, -0.00001977, 0.00000610, 0.00001477, -0.00000909, -0.00002274, -0.00000875, 0.00000369, 0.00000184, 0.00000143, 0.00000469, 0.00000128, -0.00000502, -0.00000598, -0.00000375, -0.00000039, 0.00000330, 0.00000104, -0.00000494, -0.00000256, 0.00000352, -0.00000132, -0.00000639, 0.00000414, 0.00000950, -0.00000953, -0.00002182, -0.00000207, 0.00001372, -0.00000220, -0.00001253, 0.00000592, 0.00001281, -0.00000718, -0.00000871, 0.00001503, 0.00001363, -0.00001136, -0.00000641, 0.00001634, -0.00000024, -0.00003229, -0.00001488, 0.00002343, 0.00001311, -0.00001857, -0.00000681, 0.00001916, 0.00000235, -0.00002233, -0.00000280, 0.00002237, 0.00000515, -0.00001623, -0.00000016, 0.00001651, -0.00000006, -0.00001419, 0.00000087, 0.00001164, -0.00000178, -0.00000957, 0.00000020, 0.00000068, -0.00001074, -0.00000576, 0.00001304, 0.00001386, -0.00000305, -0.00000847, -0.00000010, 0.00000041, -0.00000617, -0.00000305, 0.00000424, 0.00000014, -0.00000647, 0.00000040, 0.00000957, 0.00000294, -0.00000853, -0.00000448, 0.00000705, 0.00000575, -0.00000423, -0.00000579, -0.00000046, 0.00000049, -0.00000031, 0.00000362, 0.00000565, -0.00000063, -0.00000756, -0.00000650, 0.00000042, 0.00000698, 0.00000742, -0.00000161, -0.00001098, -0.00000575, 0.00000760, 0.00000596, -0.00000925, -0.00001126, 0.00000287, 0.00000545, -0.00001058, -0.00001969, -0.00001078, -0.00000180, -0.00000314, -0.00000765, -0.00001706, -0.00003950, -0.00005981, -0.00004514, },
		{ -0.00006385, -0.00005275, 0.00000816, 0.00007773, 0.00009496, 0.00006162, 0.00005203, 0.00009871, 0.00011973, 0.00003908, -0.00007372, -0.00009539, -0.00003557, -0.00000279, -0.00002139, -0.00002252, 0.00000869, 0.00001622, -0.00001030, -0.00002279, -0.00000763, 0.00000378, 0.00000135, 0.00000081, 0.00000357, 0.00000123, -0.00000348, -0.00000579, -0.00000508, 0.00000063, 0.00000557, -0.00000091, -0.00000921, -0.00000204, 0.00000737, -0.00000099, -0.00000810, 0.00000547, 0.00001036, -0.00001331, -0.00002421, 0.00000177, 0.00001689, -0.00000494, -0.00001477, 0.00000856, 0.00001435, -0.00000976, -0.00000984, 0.00001712, 0.00001425, -0.00001235, -0.00000550, 0.00001690, -0.00000311, -0.00003411, -0.00001195, 0.00002583, 0.00001098, -0.00002026, -0.00000481, 0.00002012, 0.00000013, -0.00002295, -0.00000050, 0.00002293, 0.00000321, -0.00001645, 0.00000159, 0.00001660, -0.00000153, -0.00001432, 0.00000168, 0.00001143, -0.00000230, -0.00000936, 0.00000001, -0.00000001, -0.00001008, -0.00000395, 0.00001337, 0.00001235, -0.00000405, -0.00000820, 0.00000004, 0.00000022, -0.00000567, -0.00000236, 0.00000390, -0.00000047, -0.00000587, 0.00000129, 0.00000891, 0.00000152, -0.00000845, -0.00000324, 0.00000717, 0.00000467, -0.00000454, -0.00000500, 0.00000009, 0.00000049, -0.00000027, 0.00000345, 0.00000495, -0.00000130, -0.00000800, -0.00000672, 0.00000142, 0.00000911, 0.00000778, -0.00000407, -0.00001247, -0.00000411, 0.00000849, 0.00000378, -0.00001006, -0.00000834, 0.00000414, 0.00000174, -0.00001391, -0.00001799, -0.00000765, -0.00000182, -0.00000496, -0.00000803, -0.00001585, -0.00003930, },
		{ -0.00001487, -0.00003843, -0.00006753, -0.00006115, 0.00000375, 0.00008362, 0.00010459, 0.00006783, 0.00006007, 0.00011138, 0.00012320, 0.00002384, -0.00009166, -0.00009650, -0.00002608, -0.00000040, -0.00002639, -0.00002408, 0.00001117, 0.00001670, -0.00001123, -0.00002203, -0.00000676, 0.00000341, 0.00000112, 0.00000027, 0.00000204, 0.00000128, -0.00000149, -0.00000575, -0.00000666, 0.00000202, 0.00000793, -0.00000356, -0.00001365, -0.00000075, 0.00001149, -0.00000122, -0.00000990, 0.00000715, 0.00001092, -0.00001748, -0.00002602, 0.00000620, 0.00001950, -0.00000828, -0.00001653, 0.00001165, 0.00001546, -0.00001265, -0.00001059, 0.00001937, 0.00001460, -0.00001332, -0.00000452, 0.00001708, -0.00000610, -0.00003526, -0.00000868, 0.00002761, 0.00000847, -0.00002147, -0.00000258, 0.00002060, -0.00000222, -0.00002306, 0.00000191, 0.00002304, 0.00000120, -0.00001629, 0.00000337, 0.00001635, -0.00000303, -0.00001423, 0.00000252, 0.00001111, -0.00000283, -0.00000917, -0.00000020, -0.00000056, -0.00000920, -0.00000223, 0.00001340, 0.00001073, -0.00000491, -0.00000783, 0.00000019, 0.00000007, -0.00000516, -0.00000178, 0.00000349, -0.00000094, -0.00000516, 0.00000199, 0.00000800, 0.00000026, -0.00000800, -0.00000203, 0.00000694, 0.00000350, -0.00000462, -0.00000408, 0.00000053, 0.00000043, -0.00000019, 0.00000323, 0.00000416, -0.00000203, -0.00000841, -0.00000671, 0.00000270, 0.00001104, 0.00000763, -0.00000649, -0.00001341, -0.00000256, 0.00000865, 0.00000171, -0.00000995, -0.00000539, 0.00000446, -0.00000211, -0.00001639, -0.00001609, -0.00000507, -0.00000186, -0.00000649, -0.00000875, },
		{ -0.00000780, -0.00000991, -0.00001398, -0.00003686, -0.00007100, -0.00007024, -0.00000116, 0.00008931, 0.00011410, 0.00007545, 0.00007095, 0.00012426, 0.00012248, 0.00000523, -0.00010765, -0.00009407, -0.00001674, -0.00000005, -0.00003117, -0.00002436, 0.00001314, 0.00001621, -0.00001157, -0.00002053, -0.00000637, 0.00000265, 0.00000129, -0.00000027, 0.00000004, 0.00000151, 0.00000095, -0.00000595, -0.00000846, 0.00000384, 0.00001026, -0.00000694, -0.00001807, 0.00000140, 0.00001572, -0.00000207, -0.00001168, 0.00000916, 0.00001109, -0.00002196, -0.00002713, 0.00001107, 0.00002140, -0.00001209, -0.00001769, 0.00001510, 0.00001606, -0.00001574, -0.00001091, 0.00002171, 0.00001467, -0.00001424, -0.00000349, 0.00001687, -0.00000908, -0.00003572, -0.00000522, 0.00002872, 0.00000572, -0.00002214, -0.00000024, 0.00002057, -0.00000457, -0.00002264, 0.00000432, 0.00002268, -0.00000080, -0.00001575, 0.00000510, 0.00001577, -0.00000452, -0.00001391, 0.00000337, 0.00001068, -0.00000337, -0.00000898, -0.00000043, -0.00000093, -0.00000814, -0.00000063, 0.00001313, 0.00000904, -0.00000561, -0.00000735, 0.00000032, -0.00000006, -0.00000463, -0.00000130, 0.00000302, -0.00000128, -0.00000438, 0.00000246, 0.00000690, -0.00000076, -0.00000722, -0.00000091, 0.00000637, 0.00000231, -0.00000445, -0.00000310, 0.00000085, 0.00000035, -0.00000008, 0.00000297, 0.00000328, -0.00000281, -0.00000876, -0.00000642, 0.00000419, 0.00001267, 0.00000700, -0.00000869, -0.00001382, -0.00000130, 0.00000816, -0.00000001, -0.00000899, -0.00000271, 0.00000387, -0.00000580, -0.00001802, -0.00001423, -0.00000302, -0.00000176, },
		{ -0.00000144, -0.00000141, -0.00000901, -0.00001155, -0.00001306, -0.00003462, -0.00007441, -0.00007985, -0.00000644, 0.00009443, 0.00012330, 0.00008513, 0.00008508, 0.00013650, 0.00011666, -0.00001619, -0.00012061, -0.00008826, -0.00000840, -0.00000171, -0.00003517, -0.00002341, 0.00001421, 0.00001486, -0.00001103, -0.00001842, -0.00000667, 0.00000163, 0.00000200, -0.00000095, -0.00000249, 0.00000205, 0.00000384, -0.00000648, -0.00001040, 0.00000610, 0.00001243, -0.00001104, -0.00002227, 0.00000440, 0.00001986, -0.00000355, -0.00001331, 0.00001146, 0.00001077, -0.00002660, -0.00002744, 0.00001622, 0.00002247, -0.00001621, -0.00001814, 0.00001875, 0.00001604, -0.00001893, -0.00001074, 0.00002406, 0.00001444, -0.00001508, -0.00000246, 0.00001628, -0.00001194, -0.00003547, -0.00000172, 0.00002913, 0.00000285, -0.00002225, 0.00000211, 0.00002002, -0.00000683, -0.00002171, 0.00000663, 0.00002186, -0.00000270, -0.00001485, 0.00000670, 0.00001487, -0.00000593, -0.00001334, 0.00000418, 0.00001011, -0.00000391, -0.00000879, -0.00000063, -0.00000113, -0.00000693, 0.00000080, 0.00001256, 0.00000731, -0.00000614, -0.00000679, 0.00000044, -0.00000017, -0.00000411, -0.00000091, 0.00000253, -0.00000147, -0.00000358, 0.00000270, 0.00000569, -0.00000149, -0.00000617, 0.00000002, 0.00000549, 0.00000118, -0.00000404, -0.00000211, 0.00000105, 0.00000027, 0.00000007, 0.00000265, 0.00000230, -0.00000365, -0.00000902, -0.00000580, 0.00000584, 0.00001388, 0.00000596, -0.00001057, -0.00001378, -0.00000045, 0.00000715, -0.00000119, -0.00000733, -0.00000055, 0.00000250, -0.00000907, -0.00001890, -0.00001261, },
		{ -0.00001916, -0.00001139, -0.00000021, -0.00000077, -0.00001025, -0.00001361, -0.00001195, -0.00003185, -0.00007786, -0.00008974, -0.00001200, 0.00009854, 0.00013216, 0.00009762, 0.00010268, 0.00014703, 0.00010503, -0.00003952, -0.00012955, -0.00007950, -0.00000187, -0.00000516, -0.00003783, -0.00002145, 0.00001401, 0.00001287, -0.00000936, -0.00001593, -0.00000782, 0.00000056, 0.00000331, -0.00000192, -0.00000554, 0.00000302, 0.00000711, -0.00000741, -0.00001242, 0.00000881, 0.00001430, -0.00001582, -0.00002605, 0.00000824, 0.00002372, -0.00000564, -0.00001467, 0.00001398, 0.00000992, -0.00003126, -0.00002689, 0.00002146, 0.00002265, -0.00002045, -0.00001781, 0.00002245, 0.00001538, -0.00002209, -0.00001004, 0.00002634, 0.00001390, -0.00001582, -0.00000148, 0.00001534, -0.00001457, -0.00003454, 0.00000167, 0.00002882, 0.00000001, -0.00002179, 0.00000434, 0.00001899, -0.00000887, -0.00002028, 0.00000874, 0.00002062, -0.00000441, -0.00001361, 0.00000810, 0.00001367, -0.00000719, -0.00001255, 0.00000492, 0.00000941, -0.00000444, -0.00000858, -0.00000081, -0.00000116, -0.00000561, 0.00000203, 0.00001173, 0.00000560, -0.00000648, -0.00000616, 0.00000054, -0.00000025, -0.00000358, -0.00000062, 0.00000204, -0.00000155, -0.00000280, 0.00000270, 0.00000445, -0.00000192, -0.00000493, 0.00000071, 0.00000436, 0.00000016, -0.00000344, -0.00000116, 0.00000115, 0.00000020, 0.00000025, 0.00000226, 0.00000124, -0.00000454, -0.00000913, -0.00000485, 0.00000755, 0.00001459, 0.00000460, -0.00001200, -0.00001340, -0.00000012, 0.00000578, -0.00000170, -0.00000518, 0.00000090, 0.00000054, -0.00001170, },
		{ -0.00000176, -0.00001356, -0.00001900, -0.00001064, 0.00000083, 0.00000016, -0.00001165, -0.00001601, -0.00001054, -0.00002874, -0.00008142, -0.00009957, -0.00001785, 0.00010115, 0.00014077, 0.00011377, 0.00012374, 0.00015463, 0.00008712, -0.00006366, -0.00013364, -0.00006846, 0.00000214, -0.00001000, -0.00003869, -0.00001881, 0.00001223, 0.00001055, -0.00000638, -0.00001335, -0.00000993, -0.00000035, 0.00000526, -0.00000334, -0.00000906, 0.00000455, 0.00001070, -0.00000882, -0.00001442, 0.00001195, 0.00001572, -0.00002120, -0.00002922, 0.00001284, 0.00002711, -0.00000830, -0.00001566, 0.00001664, 0.00000848, -0.00003577, -0.00002546, 0.00002658, 0.00002190, -0.00002461, -0.00001668, 0.00002603, 0.00001404, -0.00002507, -0.00000882, 0.00002847, 0.00001305, -0.00001645, -0.00000058, 0.00001409, -0.00001687, -0.00003298, 0.00000482, 0.00002786, -0.00000267, -0.00002082, 0.00000634, 0.00001751, -0.00001061, -0.00001843, 0.00001056, 0.00001901, -0.00000587, -0.00001209, 0.00000924, 0.00001220, -0.00000826, -0.00001155, 0.00000555, 0.00000858, -0.00000494, -0.00000834, -0.00000093, -0.00000102, -0.00000423, 0.00000304, 0.00001066, 0.00000393, -0.00000663, -0.00000548, 0.00000060, -0.00000032, -0.00000307, -0.00000040, 0.00000156, -0.00000152, -0.00000209, 0.00000249, 0.00000323, -0.00000202, -0.00000359, 0.00000113, 0.00000306, -0.00000068, -0.00000267, -0.00000030, 0.00000116, 0.00000018, 0.00000043, 0.00000181, 0.00000009, -0.00000546, -0.00000907, -0.00000354, 0.00000924, 0.00001476, 0.00000302, -0.00001292, -0.00001281, -0.00000036, 0.00000424, -0.00000147, -0.00000280, 0.00000153, },
		{ -0.00000044, 0.00000131, -0.00000415, -0.00001460, -0.00001860, -0.00001037, 0.00000179, 0.00000127, -0.00001331, -0.00001856, -0.00000879, -0.00002553, -0.00008506, -0.00010899, -0.00002414, 0.00010175, 0.00014946, 0.00013443, 0.00014794, 0.00015798, 0.00006278, -0.00008729, -0.00013230, -0.00005603, 0.00000308, -0.00001564, -0.00003738, -0.00001594, 0.00000870, 0.00000829, -0.00000201, -0.00001101, -0.00001303, -0.00000083, 0.00000781, -0.00000539, -0.00001295, 0.00000676, 0.00001449, -0.00001075, -0.00001628, 0.00001546, 0.00001658, -0.00002704, -0.00003159, 0.00001807, 0.00002987, -0.00001142, -0.00001619, 0.00001931, 0.00000647, -0.00003996, -0.00002317, 0.00003138, 0.00002024, -0.00002850, -0.00001477, 0.00002931, 0.00001207, -0.00002774, -0.00000710, 0.00003034, 0.00001191, -0.00001695, 0.00000020, 0.00001258, -0.00001875, -0.00003088, 0.00000760, 0.00002629, -0.00000507, -0.00001938, 0.00000802, 0.00001567, -0.00001195, -0.00001623, 0.00001201, 0.00001711, -0.00000701, -0.00001036, 0.00001006, 0.00001054, -0.00000909, -0.00001038, 0.00000604, 0.00000763, -0.00000539, -0.00000805, -0.00000100, -0.00000072, -0.00000282, 0.00000381, 0.00000938, 0.00000235, -0.00000660, -0.00000477, 0.00000062, -0.00000036, -0.00000257, -0.00000026, 0.00000110, -0.00000142, -0.00000149, 0.00000210, 0.00000213, -0.00000182, -0.00000224, 0.00000125, 0.00000167, -0.00000131, -0.00000181, 0.00000046, 0.00000111, 0.00000020, 0.00000061, 0.00000128, -0.00000115, -0.00000641, -0.00000877, -0.00000190, 0.00001079, 0.00001435, 0.00000137, -0.00001330, -0.00001215, -0.00000117, 0.00000274, -0.00000050, },
		{ 0.00000147, 0.00000112, 0.00000165, 0.00000028, -0.00000637, -0.00001485, -0.00001817, -0.00001053, 0.00000279, 0.00000245, -0.00001528, -0.00002106, -0.00000671, -0.00002252, -0.00008862, -0.00011763, -0.00003116, 0.00009985, 0.00015877, 0.00016043, 0.00017466, 0.00015574, 0.00003221, -0.00010899, -0.00012525, -0.00004324, 0.00000059, -0.00002140, -0.00003371, -0.00001337, 0.00000334, 0.00000654, 0.00000373, -0.00000926, -0.00001703, -0.00000063, 0.00001085, -0.00000823, -0.00001708, 0.00000974, 0.00001832, -0.00001322, -0.00001791, 0.00001928, 0.00001676, -0.00003319, -0.00003301, 0.00002377, 0.00003185, -0.00001489, -0.00001619, 0.00002188, 0.00000389, -0.00004366, -0.00002010, 0.00003564, 0.00001775, -0.00003192, -0.00001214, 0.00003212, 0.00000952, -0.00002997, -0.00000493, 0.00003189, 0.00001052, -0.00001730, 0.00000084, 0.00001089, -0.00002014, -0.00002834, 0.00000990, 0.00002423, -0.00000711, -0.00001758, 0.00000929, 0.00001357, -0.00001284, -0.00001377, 0.00001305, 0.00001498, -0.00000779, -0.00000850, 0.00001053, 0.00000874, -0.00000964, -0.00000908, 0.00000635, 0.00000659, -0.00000577, -0.00000772, -0.00000099, -0.00000029, -0.00000142, 0.00000433, 0.00000793, 0.00000089, -0.00000640, -0.00000405, 0.00000059, -0.00000038, -0.00000208, -0.00000017, 0.00000068, -0.00000127, -0.00000101, 0.00000158, 0.00000121, -0.00000136, -0.00000097, 0.00000110, 0.00000027, -0.00000174, -0.00000091, 0.00000109, 0.00000104, 0.00000028, 0.00000077, 0.00000068, -0.00000245, -0.00000734, -0.00000822, 0.00000003, 0.00001209, 0.00001339, -0.00000024, -0.00001315, -0.00001155, -0.00000248, },
		{ -0.00001114, -0.00000419, 0.00000059, 0.00000325, 0.00000324, -0.00000141, -0.00000819, -0.00001442, -0.00001785, -0.00001098, 0.00000389, 0.00000349, -0.00001753, -0.00002327, -0.00000443, -0.00001997, -0.00009187, -0.00012517, -0.00003933, 0.00009509, 0.00016942, 0.00019244, 0.00020297, 0.00014665, -0.00000403, -0.00012729, -0.00011254, -0.00003123, -0.00000543, -0.00002649, -0.00002766, -0.00001165, -0.00000375, 0.00000572, 0.00001069, -0.00000845, -0.00002179, 0.00000048, 0.00001422, -0.00001197, -0.00002124, 0.00001354, 0.00002205, -0.00001624, -0.00001918, 0.00002330, 0.00001618, -0.00003945, -0.00003338, 0.00002974, 0.00003293, -0.00001855, -0.00001563, 0.00002421, 0.00000082, -0.00004673, -0.00001634, 0.00003920, 0.00001452, -0.00003469, -0.00000890, 0.00003432, 0.00000648, -0.00003164, -0.00000238, 0.00003303, 0.00000890, -0.00001751, 0.00000131, 0.00000910, -0.00002101, -0.00002549, 0.00001166, 0.00002177, -0.00000869, -0.00001551, 0.00001013, 0.00001130, -0.00001323, -0.00001116, 0.00001363, 0.00001274, -0.00000819, -0.00000658, 0.00001064, 0.00000688, -0.00000990, -0.00000770, 0.00000646, 0.00000547, -0.00000606, -0.00000734, -0.00000092, 0.00000027, -0.00000008, 0.00000459, 0.00000635, -0.00000042, -0.00000604, -0.00000336, 0.00000053, -0.00000038, -0.00000161, -0.00000013, 0.00000029, -0.00000111, -0.00000068, 0.00000098, 0.00000050, -0.00000068, 0.00000015, 0.00000071, -0.00000106, -0.00000197, -0.00000003, 0.00000159, 0.00000097, 0.00000039, 0.00000088, 0.00000002, -0.00000381, -0.00000823, -0.00000736, 0.00000219, 0.00001304, 0.00001193, -0.00000167, -0.00001253, },
		{ -0.00000283, -0.00001152, -0.00001100, -0.00000614, 0.00000022, 0.00000568, 0.00000418, -0.00000356, -0.00000947, -0.00001349, -0.00001775, -0.00001156, 0.00000508, 0.00000422, -0.00001995, -0.00002496, -0.00000217, -0.00001811, -0.00009445, -0.00013138, -0.00004922, 0.00008724, 0.00018236, 0.00023093, 0.00023161, 0.00012962, -0.00004499, -0.00014082, -0.00009455, -0.00002112, -0.00001482, -0.00003016, -0.00001941, -0.00001134, -0.00001235, 0.00000624, 0.00001863, -0.00000888, -0.00002706, 0.00000267, 0.00001767, -0.00001671, -0.00002523, 0.00001818, 0.00002547, -0.00001976, -0.00002001, 0.00002739, 0.00001478, -0.00004562, -0.00003263, 0.00003577, 0.00003306, -0.00002226, -0.00001451, 0.00002619, -0.00000264, -0.00004901, -0.00001205, 0.00004190, 0.00001072, -0.00003667, -0.00000520, 0.00003579, 0.00000309, -0.00003266, 0.00000046, 0.00003371, 0.00000710, -0.00001757, 0.00000162, 0.00000728, -0.00002133, -0.00002244, 0.00001283, 0.00001906, -0.00000980, -0.00001330, 0.00001049, 0.00000899, -0.00001312, -0.00000850, 0.00001375, 0.00001046, -0.00000822, -0.00000470, 0.00001039, 0.00000503, -0.00000984, -0.00000630, 0.00000636, 0.00000431, -0.00000625, -0.00000690, -0.00000078, 0.00000092, 0.00000118, 0.00000459, 0.00000469, -0.00000155, -0.00000555, -0.00000270, 0.00000043, -0.00000035, -0.00000116, -0.00000013, -0.00000006, -0.00000096, -0.00000049, 0.00000039, 0.00000005, 0.00000013, 0.00000106, 0.00000012, -0.00000225, -0.00000203, 0.00000077, 0.00000199, 0.00000094, 0.00000053, 0.00000093, -0.00000070, -0.00000522, -0.00000903, -0.00000618, 0.00000450, 0.00001357, 0.00001003, },
		{ 0.00001362, 0.00000781, -0.00000366, -0.00001025, -0.00001118, -0.00000814, 0.00000041, 0.00000817, 0.00000438, -0.00000594, -0.00001010, -0.00001225, -0.00001788, -0.00001210, 0.00000628, 0.00000443, -0.00002235, -0.00002596, -0.00000022, -0.00001712, -0.00009592, -0.00013617, -0.00006146, 0.00007629, 0.00019866, 0.00027611, 0.00025909, 0.00010382, -0.00008936, -0.00014833, -0.00007201, -0.00001396, -0.00002717, -0.00003167, -0.00000932, -0.00001293, -0.00002208, 0.00000846, 0.00002717, -0.00001079, -0.00003253, 0.00000608, 0.00002095, -0.00002244, -0.00002879, 0.00002362, 0.00002842, -0.00002372, -0.00002030, 0.00003141, 0.00001255, -0.00005145, -0.00003075, 0.00004161, 0.00003220, -0.00002583, -0.00001287, 0.00002771, -0.00000638, -0.00005042, -0.00000741, 0.00004362, 0.00000654, -0.00003777, -0.00000120, 0.00003644, -0.00000050, -0.00003296, 0.00000346, 0.00003388, 0.00000518, -0.00001748, 0.00000177, 0.00000552, -0.00002112, -0.00001934, 0.00001341, 0.00001621, -0.00001040, -0.00001104, 0.00001039, 0.00000674, -0.00001252, -0.00000591, 0.00001344, 0.00000825, -0.00000790, -0.00000293, 0.00000980, 0.00000327, -0.00000950, -0.00000493, 0.00000605, 0.00000314, -0.00000632, -0.00000641, -0.00000057, 0.00000166, 0.00000231, 0.00000434, 0.00000299, -0.00000250, -0.00000495, -0.00000210, 0.00000029, -0.00000029, -0.00000073, -0.00000016, -0.00000039, -0.00000087, -0.00000043, -0.00000015, -0.00000015, 0.00000100, 0.00000172, -0.00000058, -0.00000325, -0.00000197, 0.00000145, 0.00000230, 0.00000097, 0.00000068, 0.00000092, -0.00000148, -0.00000665, -0.00000970, -0.00000468, 0.00000687, },
		{ -0.00000286, 0.00000917, 0.00001314, 0.00000539, -0.00000409, -0.00000885, -0.00001167, -0.00001001, 0.00000113, 0.00001047, 0.00000383, -0.00000828, -0.00001010, -0.00001093, -0.00001819, -0.00001242, 0.00000732, 0.00000400, -0.00002447, -0.00002614, 0.00000103, -0.00001703, -0.00009576, -0.00013961, -0.00007677, 0.00006250, 0.00021949, 0.00032782, 0.00028371, 0.00006875, -0.00013554, -0.00014884, -0.00004596, -0.00001066, -0.00004182, -0.00003042, 0.00000208, -0.00001679, -0.00003245, 0.00001262, 0.00003587, -0.00001431, -0.00003784, 0.00001074, 0.00002376, -0.00002912, -0.00003167, 0.00002976, 0.00003071, -0.00002799, -0.00001998, 0.00003521, 0.00000952, -0.00005674, -0.00002777, 0.00004705, 0.00003040, -0.00002912, -0.00001078, 0.00002867, -0.00001025, -0.00005089, -0.00000260, 0.00004431, 0.00000216, -0.00003792, 0.00000289, 0.00003624, -0.00000414, -0.00003253, 0.00000651, 0.00003353, 0.00000320, -0.00001726, 0.00000176, 0.00000389, -0.00002042, -0.00001630, 0.00001340, 0.00001336, -0.00001053, -0.00000887, 0.00000985, 0.00000465, -0.00001147, -0.00000349, 0.00001272, 0.00000618, -0.00000727, -0.00000135, 0.00000891, 0.00000167, -0.00000888, -0.00000366, 0.00000554, 0.00000201, -0.00000628, -0.00000589, -0.00000028, 0.00000245, 0.00000329, 0.00000387, 0.00000131, -0.00000324, -0.00000427, -0.00000158, 0.00000014, -0.00000019, -0.00000032, -0.00000021, -0.00000071, -0.00000085, -0.00000048, -0.00000057, -0.00000009, 0.00000184, 0.00000210, -0.00000132, -0.00000403, -0.00000183, 0.00000198, 0.00000256, 0.00000106, 0.00000081, 0.00000082, -0.00000230, -0.00000808, -0.00001019, },
		{ -0.00000948, -0.00001045, -0.00000076, 0.00001129, 0.00001215, 0.00000291, -0.00000415, -0.00000744, -0.00001242, -0.00001156, 0.00000231, 0.00001232, 0.00000263, -0.00001033, -0.00000953, -0.00000973, -0.00001857, -0.00001238, 0.00000800, 0.00000284, -0.00002599, -0.00002550, 0.00000118, -0.00001775, -0.00009348, -0.00014195, -0.00009583, 0.00004640, 0.00024601, 0.00038554, 0.00030369, 0.00002435, -0.00018174, -0.00014169, -0.00001767, -0.00001190, -0.00005792, -0.00002597, 0.00001411, -0.00002318, -0.00004286, 0.00001882, 0.00004423, -0.00001947, -0.00004258, 0.00001658, 0.00002579, -0.00003661, -0.00003365, 0.00003645, 0.00003218, -0.00003246, -0.00001903, 0.00003862, 0.00000576, -0.00006127, -0.00002378, 0.00005186, 0.00002772, -0.00003198, -0.00000832, 0.00002904, -0.00001409, -0.00005040, 0.00000215, 0.00004396, -0.00000218, -0.00003712, 0.00000690, 0.00003517, -0.00000765, -0.00003135, 0.00000948, 0.00003265, 0.00000120, -0.00001692, 0.00000162, 0.00000245, -0.00001927, -0.00001344, 0.00001285, 0.00001061, -0.00001022, -0.00000686, 0.00000895, 0.00000283, -0.00001004, -0.00000131, 0.00001165, 0.00000432, -0.00000638, -0.00000001, 0.00000778, 0.00000028, -0.00000803, -0.00000252, 0.00000484, 0.00000094, -0.00000612, -0.00000532, 0.00000007, 0.00000327, 0.00000410, 0.00000319, -0.00000031, -0.00000378, -0.00000355, -0.00000113, -0.00000002, -0.00000005, 0.00000006, -0.00000029, -0.00000103, -0.00000092, -0.00000059, -0.00000082, 0.00000018, 0.00000258, 0.00000223, -0.00000203, -0.00000457, -0.00000169, 0.00000235, 0.00000280, 0.00000122, 0.00000090, 0.00000064, -0.00000316, },
		{ 0.00000038, -0.00000405, -0.00001080, -0.00001043, 0.00000155, 0.00001312, 0.00001068, 0.00000050, -0.00000387, -0.00000615, -0.00001333, -0.00001263, 0.00000379, 0.00001353, 0.00000092, -0.00001187, -0.00000855, -0.00000882, -0.00001880, -0.00001190, 0.00000804, 0.00000095, -0.00002656, -0.00002414, -0.00000019, -0.00001905, -0.00008859, -0.00014366, -0.00011925, 0.00002882, 0.00027933, 0.00044831, 0.00031722, -0.00002901, -0.00022600, -0.00012659, 0.00001141, -0.00001809, -0.00007447, -0.00001806, 0.00002602, -0.00003217, -0.00005269, 0.00002706, 0.00005174, -0.00002620, -0.00004638, 0.00002346, 0.00002677, -0.00004471, -0.00003450, 0.00004350, 0.00003271, -0.00003695, -0.00001744, 0.00004151, 0.00000138, -0.00006484, -0.00001892, 0.00005584, 0.00002427, -0.00003429, -0.00000563, 0.00002877, -0.00001774, -0.00004897, 0.00000665, 0.00004260, -0.00000629, -0.00003541, 0.00001062, 0.00003330, -0.00001087, -0.00002946, 0.00001225, 0.00003125, -0.00000074, -0.00001646, 0.00000138, 0.00000125, -0.00001776, -0.00001085, 0.00001185, 0.00000808, -0.00000954, -0.00000511, 0.00000775, 0.00000135, -0.00000831, 0.00000055, 0.00001030, 0.00000273, -0.00000531, 0.00000105, 0.00000648, -0.00000086, -0.00000701, -0.00000156, 0.00000400, -0.00000003, -0.00000585, -0.00000473, 0.00000048, 0.00000409, 0.00000471, 0.00000235, -0.00000183, -0.00000413, -0.00000283, -0.00000078, -0.00000016, 0.00000012, 0.00000041, -0.00000039, -0.00000137, -0.00000109, -0.00000073, -0.00000086, 0.00000062, 0.00000313, 0.00000212, -0.00000263, -0.00000489, -0.00000160, 0.00000256, 0.00000305, 0.00000143, 0.00000093, },
		{ 0.00000169, 0.00000088, 0.00000005, -0.00000497, -0.00001202, -0.00001009, 0.00000400, 0.00001456, 0.00000880, -0.00000173, -0.00000331, -0.00000509, -0.00001425, -0.00001310, 0.00000535, 0.00001392, -0.00000106, -0.00001274, -0.00000739, -0.00000830, -0.00001867, -0.00001100, 0.00000716, -0.00000154, -0.00002583, -0.00002229, -0.00000347, -0.00002055, -0.00008068, -0.00014540, -0.00014749, 0.00001086, 0.00032034, 0.00051478, 0.00032262, -0.00009049, -0.00026636, -0.00010368, 0.00003972, -0.00002933, -0.00009041, -0.00000671, 0.00003699, -0.00004363, -0.00006130, 0.00003714, 0.00005789, -0.00003427, -0.00004887, 0.00003114, 0.00002646, -0.00005315, -0.00003408, 0.00005066, 0.00003222, -0.00004129, -0.00001524, 0.00004372, -0.00000348, -0.00006731, -0.00001338, 0.00005884, 0.00002020, -0.00003595, -0.00000281, 0.00002788, -0.00002103, -0.00004669, 0.00001071, 0.00004031, -0.00000997, -0.00003289, 0.00001388, 0.00003071, -0.00001365, -0.00002695, 0.00001471, 0.00002939, -0.00000259, -0.00001590, 0.00000107, 0.00000032, -0.00001597, -0.00000860, 0.00001047, 0.00000583, -0.00000857, -0.00000367, 0.00000634, 0.00000024, -0.00000636, 0.00000205, 0.00000876, 0.00000143, -0.00000412, 0.00000180, 0.00000507, -0.00000171, -0.00000588, -0.00000081, 0.00000306, -0.00000087, -0.00000548, -0.00000411, 0.00000096, 0.00000488, 0.00000510, 0.00000137, -0.00000321, -0.00000430, -0.00000213, -0.00000050, -0.00000028, 0.00000032, 0.00000072, -0.00000051, -0.00000173, -0.00000135, -0.00000084, -0.00000067, 0.00000116, 0.00000345, 0.00000184, -0.00000305, -0.00000501, -0.00000160, 0.00000264, 0.00000334, },
		{ 0.00000261, 0.00000369, 0.00000195, 0.00000074, -0.00000037, -0.00000589, -0.00001306, -0.00000942, 0.00000650, 0.00001551, 0.00000660, -0.00000367, -0.00000260, -0.00000431, -0.00001501, -0.00001292, 0.00000676, 0.00001343, -0.00000303, -0.00001285, -0.00000632, -0.00000819, -0.00001791, -0.00000975, 0.00000509, -0.00000443, -0.00002351, -0.00002029, -0.00000895, -0.00002170, -0.00006952, -0.00014800, -0.00018081, -0.00000610, 0.00036971, 0.00058323, 0.00031839, -0.00015877, -0.00030096, -0.00007354, 0.00006573, -0.00004539, -0.00010465, 0.00000785, 0.00004624, -0.00005728, -0.00006807, 0.00004874, 0.00006223, -0.00004337, -0.00004976, 0.00003929, 0.00002469, -0.00006161, -0.00003229, 0.00005765, 0.00003067, -0.00004530, -0.00001250, 0.00004515, -0.00000864, -0.00006854, -0.00000737, 0.00006074, 0.00001567, -0.00003692, -0.00000002, 0.00002642, -0.00002382, -0.00004365, 0.00001415, 0.00003723, -0.00001305, -0.00002967, 0.00001653, 0.00002752, -0.00001586, -0.00002391, 0.00001674, 0.00002710, -0.00000428, -0.00001525, 0.00000072, -0.00000031, -0.00001399, -0.00000674, 0.00000882, 0.00000391, -0.00000741, -0.00000258, 0.00000483, -0.00000045, -0.00000430, 0.00000316, 0.00000710, 0.00000044, -0.00000291, 0.00000224, 0.00000364, -0.00000227, -0.00000469, -0.00000028, 0.00000205, -0.00000158, -0.00000504, -0.00000348, 0.00000149, 0.00000561, 0.00000527, 0.00000030, -0.00000441, -0.00000431, -0.00000148, -0.00000030, -0.00000036, 0.00000055, 0.00000098, -0.00000066, -0.00000214, -0.00000168, -0.00000087, -0.00000024, 0.00000172, 0.00000350, 0.00000145, -0.00000325, -0.00000499, -0.00000173, },
		{ -0.00000488, -0.00000201, 0.00000251, 0.00000411, 0.00000220, 0.00000051, -0.00000085, -0.00000681, -0.00001388, -0.00000839, 0.00000892, 0.00001592, 0.00000422, -0.00000525, -0.00000185, -0.00000385, -0.00001542, -0.00001210, 0.00000774, 0.00001205, -0.00000472, -0.00001222, -0.00000561, -0.00000842, -0.00001627, -0.00000835, 0.00000157, -0.00000738, -0.00001936, -0.00001861, -0.00001680, -0.00002189, -0.00005505, -0.00015241, -0.00021916, -0.00002053, 0.00042772, 0.00065162, 0.00030341, -0.00023211, -0.00032814, -0.00003712, 0.00008801, -0.00006570, -0.00011619, 0.00002515, 0.00005306, -0.00007264, -0.00007251, 0.00006141, 0.00006441, -0.00005309, -0.00004884, 0.00004753, 0.00002138, -0.00006972, -0.00002910, 0.00006418, 0.00002807, -0.00004880, -0.00000931, 0.00004571, -0.00001389, -0.00006849, -0.00000112, 0.00006146, 0.00001088, -0.00003718, 0.00000263, 0.00002445, -0.00002600, -0.00004000, 0.00001686, 0.00003351, -0.00001542, -0.00002594, 0.00001847, 0.00002389, -0.00001741, -0.00002046, 0.00001828, 0.00002447, -0.00000579, -0.00001453, 0.00000038, -0.00000066, -0.00001193, -0.00000529, 0.00000700, 0.00000236, -0.00000616, -0.00000185, 0.00000331, -0.00000074, -0.00000222, 0.00000388, 0.00000542, -0.00000025, -0.00000173, 0.00000239, 0.00000226, -0.00000256, -0.00000351, 0.00000003, 0.00000102, -0.00000214, -0.00000454, -0.00000282, 0.00000206, 0.00000625, 0.00000522, -0.00000082, -0.00000541, -0.00000421, -0.00000092, -0.00000016, -0.00000039, 0.00000079, 0.00000120, -0.00000085, -0.00000260, -0.00000206, -0.00000078, 0.00000040, 0.00000224, 0.00000329, 0.00000101, -0.00000321, },
		{ 0.00000062, -0.00000293, -0.00000474, -0.00000243, 0.00000238, 0.00000457, 0.00000241, 0.00000019, -0.00000140, -0.00000771, -0.00001441, -0.00000702, 0.00001115, 0.00001576, 0.00000180, -0.00000643, -0.00000120, -0.00000368, -0.00001530, -0.00001070, 0.00000805, 0.00000986, -0.00000581, -0.00001094, -0.00000553, -0.00000884, -0.00001356, -0.00000708, -0.00000353, -0.00000999, -0.00001328, -0.00001779, -0.00002703, -0.00002040, -0.00003742, -0.00015966, -0.00026220, -0.00003076, 0.00049426, 0.00071767, 0.00027693, -0.00030845, -0.00034653, 0.00000424, 0.00010530, -0.00008945, -0.00012417, 0.00004450, 0.00005688, -0.00008907, -0.00007421, 0.00007457, 0.00006418, -0.00006293, -0.00004600, 0.00005544, 0.00001653, -0.00007713, -0.00002456, 0.00006997, 0.00002451, -0.00005163, -0.00000582, 0.00004536, -0.00001903, -0.00006713, 0.00000513, 0.00006099, 0.00000602, -0.00003674, 0.00000502, 0.00002208, -0.00002748, -0.00003591, 0.00001874, 0.00002936, -0.00001699, -0.00002187, 0.00001962, 0.00001998, -0.00001824, -0.00001675, 0.00001927, 0.00002158, -0.00000708, -0.00001375, 0.00000006, -0.00000073, -0.00000987, -0.00000425, 0.00000512, 0.00000117, -0.00000490, -0.00000145, 0.00000188, -0.00000064, -0.00000023, 0.00000423, 0.00000378, -0.00000067, -0.00000065, 0.00000228, 0.00000099, -0.00000259, -0.00000240, 0.00000013, 0.00000003, -0.00000254, -0.00000399, -0.00000215, 0.00000268, 0.00000677, 0.00000496, -0.00000192, -0.00000619, -0.00000401, -0.00000046, -0.00000006, -0.00000035, 0.00000103, 0.00000135, -0.00000107, -0.00000312, -0.00000246, -0.00000052, 0.00000121, 0.00000262, 0.00000281, },
		{ 0.00000283, 0.00000212, 0.00000033, -0.00000243, -0.00000463, -0.00000299, 0.00000227, 0.00000506, 0.00000255, -0.00000020, -0.00000202, -0.00000858, -0.00001460, -0.00000533, 0.00001303, 0.00001502, -0.00000053, -0.00000721, -0.00000080, -0.00000372, -0.00001448, -0.00000888, 0.00000743, 0.00000705, -0.00000605, -0.00000921, -0.00000629, -0.00000920, -0.00000962, -0.00000626, -0.00001029, -0.00001180, -0.00000530, -0.00001840, -0.00003948, -0.00001650, -0.00001706, -0.00017074, -0.00030920, -0.00003511, 0.00056874, 0.00077899, 0.00023872, -0.00038545, -0.00035519, 0.00004889, 0.00011665, -0.00011553, -0.00012792, 0.00006504, 0.00005729, -0.00010585, -0.00007294, 0.00008761, 0.00006143, -0.00007239, -0.00004128, 0.00006259, 0.00001024, -0.00008347, -0.00001882, 0.00007475, 0.00002012, -0.00005364, -0.00000215, 0.00004407, -0.00002384, -0.00006450, 0.00001113, 0.00005935, 0.00000125, -0.00003565, 0.00000705, 0.00001943, -0.00002821, -0.00003155, 0.00001974, 0.00002496, -0.00001772, -0.00001765, 0.00001997, 0.00001599, -0.00001832, -0.00001292, 0.00001967, 0.00001852, -0.00000812, -0.00001291, -0.00000019, -0.00000056, -0.00000789, -0.00000359, 0.00000327, 0.00000031, -0.00000372, -0.00000136, 0.00000062, -0.00000019, 0.00000161, 0.00000425, 0.00000225, -0.00000085, 0.00000027, 0.00000196, -0.00000011, -0.00000241, -0.00000141, 0.00000006, -0.00000090, -0.00000281, -0.00000343, -0.00000146, 0.00000332, 0.00000714, 0.00000450, -0.00000297, -0.00000675, -0.00000376, -0.00000011, 0.00000001, -0.00000025, 0.00000126, 0.00000143, -0.00000134, -0.00000368, -0.00000283, -0.00000007, 0.00000215, },
		{ 0.00000057, 0.00000313, 0.00000280, 0.00000128, 0.00000021, -0.00000176, -0.00000459, -0.00000364, 0.00000219, 0.00000556, 0.00000262, -0.00000066, -0.00000271, -0.00000937, -0.00001438, -0.00000339, 0.00001446, 0.00001374, -0.00000260, -0.00000764, -0.00000075, -0.00000386, -0.00001283, -0.00000684, 0.00000573, 0.00000387, -0.00000524, -0.00000729, -0.00000806, -0.00000922, -0.00000442, -0.00000630, -0.00001864, -0.00001231, 0.00000441, -0.00002102, -0.00005376, -0.00000954, 0.00000535, -0.00018660, -0.00035909, -0.00003199, 0.00065008, 0.00083316, 0.00018907, -0.00046061, -0.00035360, 0.00009502, 0.00012140, -0.00014271, -0.00012704, 0.00008582, 0.00005410, -0.00012216, -0.00006865, 0.00009986, 0.00005622, -0.00008093, -0.00003479, 0.00006857, 0.00000273, -0.00008841, -0.00001207, 0.00007828, 0.00001507, -0.00005475, 0.00000151, 0.00004189, -0.00002812, -0.00006069, 0.00001666, 0.00005663, -0.00000324, -0.00003400, 0.00000866, 0.00001663, -0.00002817, -0.00002713, 0.00001988, 0.00002053, -0.00001763, -0.00001349, 0.00001952, 0.00001208, -0.00001768, -0.00000914, 0.00001949, 0.00001537, -0.00000891, -0.00001202, -0.00000035, -0.00000019, -0.00000608, -0.00000327, 0.00000153, -0.00000025, -0.00000269, -0.00000151, -0.00000040, 0.00000054, 0.00000322, 0.00000399, 0.00000088, -0.00000085, 0.00000099, 0.00000149, -0.00000100, -0.00000206, -0.00000058, -0.00000016, -0.00000171, -0.00000296, -0.00000286, -0.00000074, 0.00000396, 0.00000734, 0.00000387, -0.00000392, -0.00000710, -0.00000350, 0.00000012, 0.00000009, -0.00000008, 0.00000147, 0.00000144, -0.00000166, -0.00000427, -0.00000312, },
		{ -0.00000486, -0.00000329, 0.00000140, 0.00000409, 0.00000253, 0.00000036, 0.00000028, -0.00000098, -0.00000465, -0.00000433, 0.00000216, 0.00000601, 0.00000261, -0.00000116, -0.00000349, -0.00001007, -0.00001370, -0.00000127, 0.00001531, 0.00001199, -0.00000428, -0.00000779, -0.00000116, -0.00000395, -0.00001026, -0.00000483, 0.00000282, 0.00000063, -0.00000326, -0.00000551, -0.00001089, -0.00000854, 0.00000199, -0.00000758, -0.00002837, -0.00001106, 0.00001549, -0.00002617, -0.00006931, 0.00000108, 0.00002887, -0.00020800, -0.00041044, -0.00002000, 0.00073675, 0.00087791, 0.00012882, -0.00053145, -0.00034175, 0.00014067, 0.00011930, -0.00016963, -0.00012139, 0.00010581, 0.00004731, -0.00013719, -0.00006145, 0.00011068, 0.00004873, -0.00008808, -0.00002680, 0.00007304, -0.00000573, -0.00009168, -0.00000461, 0.00008037, 0.00000957, -0.00005489, 0.00000500, 0.00003889, -0.00003167, -0.00005585, 0.00002151, 0.00005294, -0.00000732, -0.00003188, 0.00000979, 0.00001383, -0.00002739, -0.00002281, 0.00001920, 0.00001624, -0.00001677, -0.00000957, 0.00001836, 0.00000843, -0.00001637, -0.00000554, 0.00001875, 0.00001224, -0.00000944, -0.00001110, -0.00000042, 0.00000033, -0.00000447, -0.00000322, -0.00000003, -0.00000056, -0.00000186, -0.00000184, -0.00000114, 0.00000147, 0.00000455, 0.00000350, -0.00000029, -0.00000073, 0.00000150, 0.00000092, -0.00000168, -0.00000161, 0.00000007, -0.00000048, -0.00000240, -0.00000301, -0.00000229, 0.00000001, 0.00000458, 0.00000735, 0.00000310, -0.00000473, -0.00000724, -0.00000327, 0.00000024, 0.00000019, 0.00000015, 0.00000163, 0.00000136, -0.00000203, },
		{ 0.00000119, -0.00000246, -0.00000543, -0.00000329, 0.00000239, 0.00000493, 0.00000200, -0.00000056, 0.00000057, -0.00000017, -0.00000482, -0.00000501, 0.00000217, 0.00000637, 0.00000252, -0.00000169, -0.00000435, -0.00001062, -0.00001252, 0.00000094, 0.00001547, 0.00000988, -0.00000546, -0.00000776, -0.00000209, -0.00000381, -0.00000675, -0.00000315, -0.00000132, -0.00000232, -0.00000009, -0.00000420, -0.00001474, -0.00000683, 0.00000942, -0.00001048, -0.00003914, -0.00000762, 0.00002743, -0.00003427, -0.00008539, 0.00001573, 0.00005237, -0.00023543, -0.00046155, 0.00000194, 0.00082675, 0.00091119, 0.00005935, -0.00059556, -0.00032012, 0.00018393, 0.00011046, -0.00019492, -0.00011113, 0.00012401, 0.00003720, -0.00015019, -0.00005167, 0.00011947, 0.00003931, -0.00009342, -0.00001766, 0.00007570, -0.00001476, -0.00009310, 0.00000326, 0.00008091, 0.00000387, -0.00005405, 0.00000815, 0.00003520, -0.00003435, -0.00005015, 0.00002552, 0.00004842, -0.00001086, -0.00002942, 0.00001045, 0.00001114, -0.00002594, -0.00001875, 0.00001779, 0.00001228, -0.00001526, -0.00000604, 0.00001658, 0.00000518, -0.00001447, -0.00000225, 0.00001749, 0.00000920, -0.00000971, -0.00001015, -0.00000039, 0.00000093, -0.00000311, -0.00000339, -0.00000138, -0.00000071, -0.00000126, -0.00000228, -0.00000156, 0.00000254, 0.00000558, 0.00000285, -0.00000124, -0.00000053, 0.00000180, 0.00000033, -0.00000212, -0.00000112, 0.00000052, -0.00000085, -0.00000294, -0.00000298, -0.00000174, 0.00000080, 0.00000516, 0.00000716, 0.00000224, -0.00000536, -0.00000722, -0.00000309, 0.00000027, 0.00000033, 0.00000042, 0.00000172, },
		{ 0.00000073, 0.00000175, 0.00000093, -0.00000293, -0.00000593, -0.00000310, 0.00000350, 0.00000559, 0.00000125, -0.00000137, 0.00000106, 0.00000060, -0.00000510, -0.00000561, 0.00000223, 0.00000661, 0.00000238, -0.00000223, -0.00000531, -0.00001097, -0.00001081, 0.00000312, 0.00001487, 0.00000755, -0.00000605, -0.00000770, -0.00000356, -0.00000324, -0.00000237, -0.00000207, -0.00000662, -0.00000465, 0.00000420, -0.00000373, -0.00001947, -0.00000377, 0.00001758, -0.00001531, -0.00005051, -0.00000169, 0.00003960, -0.00004557, -0.00010111, 0.00003459, 0.00007454, -0.00026909, -0.00051046, 0.00003455, 0.00091774, 0.00093135, -0.00001748, -0.00065079, -0.00028969, 0.00022296, 0.00009539, -0.00021732, -0.00009670, 0.00013950, 0.00002421, -0.00016045, -0.00003976, 0.00012577, 0.00002841, -0.00009662, -0.00000781, 0.00007637, -0.00002395, -0.00009256, 0.00001116, 0.00007986, -0.00000179, -0.00005226, 0.00001083, 0.00003095, -0.00003603, -0.00004380, 0.00002855, 0.00004327, -0.00001380, -0.00002672, 0.00001064, 0.00000869, -0.00002391, -0.00001511, 0.00001577, 0.00000877, -0.00001321, -0.00000303, 0.00001431, 0.00000244, -0.00001212, 0.00000062, 0.00001580, 0.00000633, -0.00000974, -0.00000918, -0.00000026, 0.00000158, -0.00000201, -0.00000369, -0.00000248, -0.00000075, -0.00000091, -0.00000273, -0.00000167, 0.00000366, 0.00000629, 0.00000212, -0.00000198, -0.00000032, 0.00000190, -0.00000023, -0.00000234, -0.00000063, 0.00000079, -0.00000123, -0.00000334, -0.00000290, -0.00000119, 0.00000161, 0.00000567, 0.00000676, 0.00000132, -0.00000579, -0.00000705, -0.00000299, 0.00000021, 0.00000053, },
		{ 0.00000011, 0.00000078, 0.00000106, 0.00000169, 0.00000057, -0.00000343, -0.00000631, -0.00000267, 0.00000465, 0.00000600, 0.00000034, -0.00000202, 0.00000170, 0.00000125, -0.00000544, -0.00000608, 0.00000227, 0.00000669, 0.00000222, -0.00000279, -0.00000638, -0.00001106, -0.00000856, 0.00000513, 0.00001346, 0.00000515, -0.00000601, -0.00000773, -0.00000553, -0.00000208, 0.00000278, -0.00000188, -0.00001291, -0.00000599, 0.00000943, -0.00000440, -0.00002481, 0.00000090, 0.00002607, -0.00002229, -0.00006188, 0.00000690, 0.00005126, -0.00006014, -0.00011549, 0.00005757, 0.00009399, -0.00030881, -0.00055509, 0.00007805, 0.00100709, 0.00093718, -0.00009939, -0.00069534, -0.00025184, 0.00025616, 0.00007494, -0.00023570, -0.00007880, 0.00015151, 0.00000899, -0.00016746, -0.00002632, 0.00012920, 0.00001656, -0.00009750, 0.00000227, 0.00007499, -0.00003285, -0.00009006, 0.00001873, 0.00007724, -0.00000715, -0.00004961, 0.00001289, 0.00002634, -0.00003665, -0.00003703, 0.00003053, 0.00003767, -0.00001608, -0.00002393, 0.00001042, 0.00000658, -0.00002143, -0.00001196, 0.00001331, 0.00000582, -0.00001080, -0.00000061, 0.00001172, 0.00000030, -0.00000943, 0.00000299, 0.00001375, 0.00000371, -0.00000954, -0.00000819, -0.00000004, 0.00000220, -0.00000118, -0.00000407, -0.00000333, -0.00000075, -0.00000079, -0.00000314, -0.00000147, 0.00000474, 0.00000669, 0.00000135, -0.00000253, -0.00000015, 0.00000183, -0.00000071, -0.00000238, -0.00000021, 0.00000089, -0.00000156, -0.00000359, -0.00000280, -0.00000065, 0.00000246, 0.00000608, 0.00000616, 0.00000040, -0.00000601, -0.00000678, -0.00000298, },
		{ -0.00000645, -0.00000308, -0.00000002, 0.00000110, 0.00000137, 0.00000152, 0.00000011, -0.00000395, -0.00000654, -0.00000202, 0.00000576, 0.00000610, -0.00000067, -0.00000242, 0.00000245, 0.00000173, -0.00000580, -0.00000639, 0.00000226, 0.00000659, 0.00000208, -0.00000336, -0.00000756, -0.00001084, -0.00000578, 0.00000682, 0.00001124, 0.00000284, -0.00000531, -0.00000801, -0.00000792, -0.00000014, 0.00000851, -0.00000281, -0.00001996, -0.00000609, 0.00001531, -0.00000648, -0.00003043, 0.00000734, 0.00003440, -0.00003151, -0.00007264, 0.00001815, 0.00006162, -0.00007782, -0.00012755, 0.00008429, 0.00010932, -0.00035399, -0.00059331, 0.00013214, 0.00109204, 0.00092805, -0.00018381, -0.00072783, -0.00020832, 0.00028225, 0.00005022, -0.00024914, -0.00005831, 0.00015945, -0.00000768, -0.00017085, -0.00001205, 0.00012958, 0.00000437, -0.00009599, 0.00001207, 0.00007160, -0.00004103, -0.00008570, 0.00002561, 0.00007315, -0.00001199, -0.00004623, 0.00001427, 0.00002154, -0.00003617, -0.00003009, 0.00003144, 0.00003181, -0.00001769, -0.00002113, 0.00000985, 0.00000486, -0.00001863, -0.00000939, 0.00001055, 0.00000348, -0.00000817, 0.00000115, 0.00000896, -0.00000120, -0.00000657, 0.00000481, 0.00001144, 0.00000138, -0.00000915, -0.00000719, 0.00000024, 0.00000277, -0.00000060, -0.00000445, -0.00000395, -0.00000078, -0.00000087, -0.00000343, -0.00000100, 0.00000574, 0.00000680, 0.00000060, -0.00000290, -0.00000004, 0.00000162, -0.00000107, -0.00000226, 0.00000011, 0.00000085, -0.00000182, -0.00000373, -0.00000269, -0.00000010, 0.00000332, 0.00000636, 0.00000538, -0.00000047, -0.00000601, },
		{ -0.00000125, -0.00000582, -0.00000609, -0.00000327, -0.00000016, 0.00000148, 0.00000166, 0.00000126, -0.00000045, -0.00000445, -0.00000655, -0.00000115, 0.00000676, 0.00000588, -0.00000167, -0.00000254, 0.00000321, 0.00000200, -0.00000610, -0.00000651, 0.00000212, 0.00000630, 0.00000200, -0.00000395, -0.00000885, -0.00001022, -0.00000252, 0.00000806, 0.00000823, 0.00000080, -0.00000401, -0.00000867, -0.00001063, 0.00000269, 0.00001455, -0.00000505, -0.00002743, -0.00000471, 0.00002148, -0.00001012, -0.00003591, 0.00001561, 0.00004205, -0.00004290, -0.00008210, 0.00003188, 0.00006990, -0.00009823, -0.00013630, 0.00011405, 0.00011922, -0.00040365, -0.00062308, 0.00019599, 0.00116978, 0.00090391, -0.00026801, -0.00074743, -0.00016110, 0.00030031, 0.00002260, -0.00025704, -0.00003628, 0.00016296, -0.00002493, -0.00017044, 0.00000233, 0.00012688, -0.00000757, -0.00009218, 0.00002110, 0.00006636, -0.00004809, -0.00007966, 0.00003148, 0.00006778, -0.00001611, -0.00004226, 0.00001490, 0.00001677, -0.00003463, -0.00002322, 0.00003127, 0.00002590, -0.00001866, -0.00001844, 0.00000900, 0.00000358, -0.00001565, -0.00000742, 0.00000766, 0.00000176, -0.00000551, 0.00000227, 0.00000619, -0.00000206, -0.00000367, 0.00000606, 0.00000899, -0.00000063, -0.00000859, -0.00000619, 0.00000058, 0.00000323, -0.00000023, -0.00000478, -0.00000437, -0.00000087, -0.00000112, -0.00000356, -0.00000030, 0.00000658, 0.00000666, -0.00000009, -0.00000312, -0.00000002, 0.00000133, -0.00000129, -0.00000204, 0.00000031, 0.00000072, -0.00000198, -0.00000376, -0.00000258, 0.00000046, 0.00000419, 0.00000649, 0.00000444, },
		{ 0.00000644, 0.00000339, -0.00000190, -0.00000545, -0.00000575, -0.00000354, -0.00000026, 0.00000190, 0.00000189, 0.00000089, -0.00000110, -0.00000492, -0.00000631, -0.00000008, 0.00000754, 0.00000534, -0.00000257, -0.00000235, 0.00000388, 0.00000203, -0.00000625, -0.00000643, 0.00000180, 0.00000583, 0.00000202, -0.00000460, -0.00001025, -0.00000914, 0.00000116, 0.00000870, 0.00000451, -0.00000081, -0.00000216, -0.00000982, -0.00001348, 0.00000651, 0.00002064, -0.00000870, -0.00003494, -0.00000174, 0.00002755, -0.00001539, -0.00004080, 0.00002563, 0.00004848, -0.00005624, -0.00008963, 0.00004770, 0.00007535, -0.00012075, -0.00014091, 0.00014589, 0.00012254, -0.00045641, -0.00064258, 0.00026822, 0.00123764, 0.00086535, -0.00034918, -0.00075384, -0.00011231, 0.00030982, -0.00000646, -0.00025908, -0.00001382, 0.00016193, -0.00004187, -0.00016629, 0.00001608, 0.00012124, -0.00001867, -0.00008627, 0.00002891, 0.00005955, -0.00005366, -0.00007225, 0.00003607, 0.00006134, -0.00001935, -0.00003790, 0.00001477, 0.00001222, -0.00003208, -0.00001665, 0.00003011, 0.00002010, -0.00001903, -0.00001593, 0.00000798, 0.00000275, -0.00001265, -0.00000603, 0.00000481, 0.00000064, -0.00000296, 0.00000279, 0.00000358, -0.00000233, -0.00000089, 0.00000674, 0.00000648, -0.00000229, -0.00000790, -0.00000520, 0.00000093, 0.00000356, -0.00000005, -0.00000503, -0.00000462, -0.00000107, -0.00000147, -0.00000349, 0.00000056, 0.00000723, 0.00000631, -0.00000069, -0.00000325, -0.00000011, 0.00000100, -0.00000136, -0.00000177, 0.00000037, 0.00000054, -0.00000202, -0.00000372, -0.00000249, 0.00000105, 0.00000503, },
		{ 0.00000168, 0.00000583, 0.00000620, 0.00000226, -0.00000239, -0.00000494, -0.00000545, -0.00000386, -0.00000033, 0.00000233, 0.00000206, 0.00000041, -0.00000183, -0.00000531, -0.00000578, 0.00000113, 0.00000802, 0.00000449, -0.00000326, -0.00000188, 0.00000437, 0.00000185, -0.00000618, -0.00000618, 0.00000121, 0.00000522, 0.00000219, -0.00000533, -0.00001171, -0.00000754, 0.00000517, 0.00000861, 0.00000019, -0.00000185, 0.00000011, -0.00001155, -0.00001629, 0.00001132, 0.00002643, -0.00001378, -0.00004210, 0.00000284, 0.00003307, -0.00002221, -0.00004464, 0.00003718, 0.00005316, -0.00007117, -0.00009464, 0.00006507, 0.00007737, -0.00014456, -0.00014070, 0.00017858, 0.00011840, -0.00051055, -0.00065029, 0.00034698, 0.00129321, 0.00081354, -0.00042466, -0.00074732, -0.00006409, 0.00031072, -0.00003546, -0.00025527, 0.00000795, 0.00015652, -0.00005762, -0.00015864, 0.00002853, 0.00011295, -0.00002842, -0.00007860, 0.00003514, 0.00005153, -0.00005750, -0.00006380, 0.00003919, 0.00005414, -0.00002162, -0.00003334, 0.00001392, 0.00000808, -0.00002866, -0.00001058, 0.00002803, 0.00001458, -0.00001887, -0.00001365, 0.00000687, 0.00000234, -0.00000974, -0.00000518, 0.00000213, 0.00000006, -0.00000066, 0.00000278, 0.00000126, -0.00000206, 0.00000166, 0.00000689, 0.00000403, -0.00000360, -0.00000710, -0.00000424, 0.00000128, 0.00000374, -0.00000000, -0.00000516, -0.00000476, -0.00000139, -0.00000187, -0.00000319, 0.00000152, 0.00000765, 0.00000580, -0.00000119, -0.00000331, -0.00000029, 0.00000067, -0.00000129, -0.00000150, 0.00000029, 0.00000035, -0.00000195, -0.00000365, -0.00000241, },
		{ -0.00000355, -0.00000232, 0.00000234, 0.00000654, 0.00000577, 0.00000110, -0.00000270, -0.00000433, -0.00000521, -0.00000421, -0.00000035, 0.00000276, 0.00000214, -0.00000016, -0.00000264, -0.00000557, -0.00000493, 0.00000242, 0.00000812, 0.00000341, -0.00000363, -0.00000117, 0.00000458, 0.00000148, -0.00000580, -0.00000580, 0.00000031, 0.00000452, 0.00000254, -0.00000619, -0.00001321, -0.00000536, 0.00000938, 0.00000770, -0.00000460, -0.00000219, 0.00000267, -0.00001389, -0.00001886, 0.00001708, 0.00003162, -0.00002020, -0.00004847, 0.00000892, 0.00003762, -0.00003041, -0.00004702, 0.00004993, 0.00005565, -0.00008715, -0.00009667, 0.00008327, 0.00007550, -0.00016867, -0.00013525, 0.00021071, 0.00010628, -0.00056409, -0.00064511, 0.00042998, 0.00133445, 0.00075020, -0.00049198, -0.00072869, -0.00001847, 0.00030337, -0.00006294, -0.00024593, 0.00002798, 0.00014711, -0.00007141, -0.00014792, 0.00003907, 0.00010246, -0.00003640, -0.00006961, 0.00003953, 0.00004273, -0.00005942, -0.00005471, 0.00004071, 0.00004646, -0.00002288, -0.00002879, 0.00001243, 0.00000451, -0.00002452, -0.00000518, 0.00002519, 0.00000946, -0.00001826, -0.00001164, 0.00000576, 0.00000232, -0.00000704, -0.00000479, -0.00000026, -0.00000008, 0.00000128, 0.00000234, -0.00000068, -0.00000135, 0.00000386, 0.00000658, 0.00000171, -0.00000457, -0.00000623, -0.00000330, 0.00000159, 0.00000378, -0.00000005, -0.00000517, -0.00000483, -0.00000183, -0.00000225, -0.00000267, 0.00000252, 0.00000785, 0.00000517, -0.00000158, -0.00000334, -0.00000055, 0.00000040, -0.00000110, -0.00000127, 0.00000010, 0.00000019, -0.00000177, },
		{ 0.00000010, -0.00000151, -0.00000347, -0.00000222, 0.00000304, 0.00000713, 0.00000516, -0.00000002, -0.00000282, -0.00000367, -0.00000504, -0.00000455, -0.00000031, 0.00000315, 0.00000213, -0.00000083, -0.00000350, -0.00000568, -0.00000374, 0.00000370, 0.00000777, 0.00000218, -0.00000361, -0.00000028, 0.00000443, 0.00000100, -0.00000506, -0.00000538, -0.00000095, 0.00000379, 0.00000308, -0.00000721, -0.00001469, -0.00000258, 0.00001365, 0.00000588, -0.00000966, -0.00000174, 0.00000536, -0.00001686, -0.00002097, 0.00002368, 0.00003588, -0.00002780, -0.00005369, 0.00001629, 0.00004080, -0.00003968, -0.00004758, 0.00006341, 0.00005560, -0.00010356, -0.00009539, 0.00010148, 0.00006950, -0.00019199, -0.00012441, 0.00024077, 0.00008602, -0.00061485, -0.00062644, 0.00051461, 0.00135983, 0.00067754, -0.00054905, -0.00069922, 0.00002270, 0.00028850, -0.00008760, -0.00023169, 0.00004537, 0.00013428, -0.00008258, -0.00013469, 0.00004727, 0.00009030, -0.00004233, -0.00005978, 0.00004192, 0.00003362, -0.00005937, -0.00004542, 0.00004061, 0.00003862, -0.00002313, -0.00002442, 0.00001040, 0.00000163, -0.00001983, -0.00000058, 0.00002173, 0.00000486, -0.00001730, -0.00000990, 0.00000472, 0.00000261, -0.00000464, -0.00000477, -0.00000229, 0.00000011, 0.00000279, 0.00000160, -0.00000215, -0.00000033, 0.00000565, 0.00000588, -0.00000039, -0.00000521, -0.00000532, -0.00000242, 0.00000185, 0.00000369, -0.00000014, -0.00000506, -0.00000487, -0.00000237, -0.00000257, -0.00000195, 0.00000350, 0.00000781, 0.00000448, -0.00000187, -0.00000336, -0.00000086, 0.00000020, -0.00000082, -0.00000112, -0.00000018, },
		{ -0.00000107, -0.00000050, 0.00000010, -0.00000119, -0.00000340, -0.00000208, 0.00000376, 0.00000756, 0.00000439, -0.00000105, -0.00000277, -0.00000300, -0.00000493, -0.00000484, -0.00000024, 0.00000346, 0.00000203, -0.00000159, -0.00000440, -0.00000557, -0.00000222, 0.00000489, 0.00000696, 0.00000089, -0.00000313, 0.00000068, 0.00000384, 0.00000051, -0.00000389, -0.00000500, -0.00000260, 0.00000313, 0.00000381, -0.00000844, -0.00001608, 0.00000083, 0.00001783, 0.00000312, -0.00001481, -0.00000047, 0.00000799, -0.00002040, -0.00002245, 0.00003093, 0.00003895, -0.00003632, -0.00005740, 0.00002466, 0.00004231, -0.00004963, -0.00004606, 0.00007707, 0.00005283, -0.00011966, -0.00009067, 0.00011878, 0.00005937, -0.00021336, -0.00010836, 0.00026719, 0.00005792, -0.00066062, -0.00059424, 0.00059808, 0.00136840, 0.00059808, -0.00059424, -0.00066062, 0.00005792, 0.00026719, -0.00010836, -0.00021336, 0.00005937, 0.00011878, -0.00009067, -0.00011966, 0.00005283, 0.00007707, -0.00004606, -0.00004963, 0.00004231, 0.00002466, -0.00005740, -0.00003632, 0.00003895, 0.00003093, -0.00002245, -0.00002040, 0.00000799, -0.00000047, -0.00001481, 0.00000312, 0.00001783, 0.00000083, -0.00001608, -0.00000844, 0.00000381, 0.00000313, -0.00000260, -0.00000500, -0.00000389, 0.00000051, 0.00000384, 0.00000068, -0.00000313, 0.00000089, 0.00000696, 0.00000489, -0.00000222, -0.00000557, -0.00000440, -0.00000159, 0.00000203, 0.00000346, -0.00000024, -0.00000484, -0.00000493, -0.00000300, -0.00000277, -0.00000105, 0.00000439, 0.00000756, 0.00000376, -0.00000208, -0.00000340, -0.00000119, 0.00000010, -0.00000050, },
		{ 0.00000010, -0.00000018, -0.00000112, -0.00000082, 0.00000020, -0.00000086, -0.00000336, -0.00000187, 0.00000448, 0.00000781, 0.00000350, -0.00000195, -0.00000257, -0.00000237, -0.00000487, -0.00000506, -0.00000014, 0.00000369, 0.00000185, -0.00000242, -0.00000532, -0.00000521, -0.00000039, 0.00000588, 0.00000565, -0.00000033, -0.00000215, 0.00000160, 0.00000279, 0.00000011, -0.00000229, -0.00000477, -0.00000464, 0.00000261, 0.00000472, -0.00000990, -0.00001730, 0.00000486, 0.00002173, -0.00000058, -0.00001983, 0.00000163, 0.00001040, -0.00002442, -0.00002313, 0.00003862, 0.00004061, -0.00004542, -0.00005937, 0.00003362, 0.00004192, -0.00005978, -0.00004233, 0.00009030, 0.00004727, -0.00013469, -0.00008258, 0.00013428, 0.00004537, -0.00023169, -0.00008760, 0.00028850, 0.00002270, -0.00069922, -0.00054905, 0.00067754, 0.00135983, 0.00051461, -0.00062644, -0.00061485, 0.00008602, 0.00024077, -0.00012441, -0.00019199, 0.00006950, 0.00010148, -0.00009539, -0.00010356, 0.00005560, 0.00006341, -0.00004758, -0.00003968, 0.00004080, 0.00001629, -0.00005369, -0.00002780, 0.00003588, 0.00002368, -0.00002097, -0.00001686, 0.00000536, -0.00000174, -0.00000966, 0.00000588, 0.00001365, -0.00000258, -0.00001469, -0.00000721, 0.00000308, 0.00000379, -0.00000095, -0.00000538, -0.00000506, 0.00000100, 0.00000443, -0.00000028, -0.00000361, 0.00000218, 0.00000777, 0.00000370, -0.00000374, -0.00000568, -0.00000350, -0.00000083, 0.00000213, 0.00000315, -0.00000031, -0.00000455, -0.00000504, -0.00000367, -0.00000282, -0.00000002, 0.00000516, 0.00000713, 0.00000304, -0.00000222, -0.00000347, -0.00000151, },
		{ -0.00000355, -0.00000177, 0.00000019, 0.00000010, -0.00000127, -0.00000110, 0.00000040, -0.00000055, -0.00000334, -0.00000158, 0.00000517, 0.00000785, 0.00000252, -0.00000267, -0.00000225, -0.00000183, -0.00000483, -0.00000517, -0.00000005, 0.00000378, 0.00000159, -0.00000330, -0.00000623, -0.00000457, 0.00000171, 0.00000658, 0.00000386, -0.00000135, -0.00000068, 0.00000234, 0.00000128, -0.00000008, -0.00000026, -0.00000479, -0.00000704, 0.00000232, 0.00000576, -0.00001164, -0.00001826, 0.00000946, 0.00002519, -0.00000518, -0.00002452, 0.00000451, 0.00001243, -0.00002879, -0.00002288, 0.00004646, 0.00004071, -0.00005471, -0.00005942, 0.00004273, 0.00003953, -0.00006961, -0.00003640, 0.00010246, 0.00003907, -0.00014792, -0.00007141, 0.00014711, 0.00002798, -0.00024593, -0.00006294, 0.00030337, -0.00001847, -0.00072869, -0.00049198, 0.00075020, 0.00133445, 0.00042998, -0.00064511, -0.00056409, 0.00010628, 0.00021071, -0.00013525, -0.00016867, 0.00007550, 0.00008327, -0.00009667, -0.00008715, 0.00005565, 0.00004993, -0.00004702, -0.00003041, 0.00003762, 0.00000892, -0.00004847, -0.00002020, 0.00003162, 0.00001708, -0.00001886, -0.00001389, 0.00000267, -0.00000219, -0.00000460, 0.00000770, 0.00000938, -0.00000536, -0.00001321, -0.00000619, 0.00000254, 0.00000452, 0.00000031, -0.00000580, -0.00000580, 0.00000148, 0.00000458, -0.00000117, -0.00000363, 0.00000341, 0.00000812, 0.00000242, -0.00000493, -0.00000557, -0.00000264, -0.00000016, 0.00000214, 0.00000276, -0.00000035, -0.00000421, -0.00000521, -0.00000433, -0.00000270, 0.00000110, 0.00000577, 0.00000654, 0.00000234, -0.00000232, },
		{ 0.00000168, -0.00000241, -0.00000365, -0.00000195, 0.00000035, 0.00000029, -0.00000150, -0.00000129, 0.00000067, -0.00000029, -0.00000331, -0.00000119, 0.00000580, 0.00000765, 0.00000152, -0.00000319, -0.00000187, -0.00000139, -0.00000476, -0.00000516, -0.00000000, 0.00000374, 0.00000128, -0.00000424, -0.00000710, -0.00000360, 0.00000403, 0.00000689, 0.00000166, -0.00000206, 0.00000126, 0.00000278, -0.00000066, 0.00000006, 0.00000213, -0.00000518, -0.00000974, 0.00000234, 0.00000687, -0.00001365, -0.00001887, 0.00001458, 0.00002803, -0.00001058, -0.00002866, 0.00000808, 0.00001392, -0.00003334, -0.00002162, 0.00005414, 0.00003919, -0.00006380, -0.00005750, 0.00005153, 0.00003514, -0.00007860, -0.00002842, 0.00011295, 0.00002853, -0.00015864, -0.00005762, 0.00015652, 0.00000795, -0.00025527, -0.00003546, 0.00031072, -0.00006409, -0.00074732, -0.00042466, 0.00081354, 0.00129321, 0.00034698, -0.00065029, -0.00051055, 0.00011840, 0.00017858, -0.00014070, -0.00014456, 0.00007737, 0.00006507, -0.00009464, -0.00007117, 0.00005316, 0.00003718, -0.00004464, -0.00002221, 0.00003307, 0.00000284, -0.00004210, -0.00001378, 0.00002643, 0.00001132, -0.00001629, -0.00001155, 0.00000011, -0.00000185, 0.00000019, 0.00000861, 0.00000517, -0.00000754, -0.00001171, -0.00000533, 0.00000219, 0.00000522, 0.00000121, -0.00000618, -0.00000618, 0.00000185, 0.00000437, -0.00000188, -0.00000326, 0.00000449, 0.00000802, 0.00000113, -0.00000578, -0.00000531, -0.00000183, 0.00000041, 0.00000206, 0.00000233, -0.00000033, -0.00000386, -0.00000545, -0.00000494, -0.00000239, 0.00000226, 0.00000620, 0.00000583, },
		{ 0.00000644, 0.00000503, 0.00000105, -0.00000249, -0.00000372, -0.00000202, 0.00000054, 0.00000037, -0.00000177, -0.00000136, 0.00000100, -0.00000011, -0.00000325, -0.00000069, 0.00000631, 0.00000723, 0.00000056, -0.00000349, -0.00000147, -0.00000107, -0.00000462, -0.00000503, -0.00000005, 0.00000356, 0.00000093, -0.00000520, -0.00000790, -0.00000229, 0.00000648, 0.00000674, -0.00000089, -0.00000233, 0.00000358, 0.00000279, -0.00000296, 0.00000064, 0.00000481, -0.00000603, -0.00001265, 0.00000275, 0.00000798, -0.00001593, -0.00001903, 0.00002010, 0.00003011, -0.00001665, -0.00003208, 0.00001222, 0.00001477, -0.00003790, -0.00001935, 0.00006134, 0.00003607, -0.00007225, -0.00005366, 0.00005955, 0.00002891, -0.00008627, -0.00001867, 0.00012124, 0.00001608, -0.00016629, -0.00004187, 0.00016193, -0.00001382, -0.00025908, -0.00000646, 0.00030982, -0.00011231, -0.00075384, -0.00034918, 0.00086535, 0.00123764, 0.00026822, -0.00064258, -0.00045641, 0.00012254, 0.00014589, -0.00014091, -0.00012075, 0.00007535, 0.00004770, -0.00008963, -0.00005624, 0.00004848, 0.00002563, -0.00004080, -0.00001539, 0.00002755, -0.00000174, -0.00003494, -0.00000870, 0.00002064, 0.00000651, -0.00001348, -0.00000982, -0.00000216, -0.00000081, 0.00000451, 0.00000870, 0.00000116, -0.00000914, -0.00001025, -0.00000460, 0.00000202, 0.00000583, 0.00000180, -0.00000643, -0.00000625, 0.00000203, 0.00000388, -0.00000235, -0.00000257, 0.00000534, 0.00000754, -0.00000008, -0.00000631, -0.00000492, -0.00000110, 0.00000089, 0.00000189, 0.00000190, -0.00000026, -0.00000354, -0.00000575, -0.00000545, -0.00000190, 0.00000339, },
		{ -0.00000125, 0.00000444, 0.00000649, 0.00000419, 0.00000046, -0.00000258, -0.00000376, -0.00000198, 0.00000072, 0.00000031, -0.00000204, -0.00000129, 0.00000133, -0.00000002, -0.00000312, -0.00000009, 0.00000666, 0.00000658, -0.00000030, -0.00000356, -0.00000112, -0.00000087, -0.00000437, -0.00000478, -0.00000023, 0.00000323, 0.00000058, -0.00000619, -0.00000859, -0.00000063, 0.00000899, 0.00000606, -0.00000367, -0.00000206, 0.00000619, 0.00000227, -0.00000551, 0.00000176, 0.00000766, -0.00000742, -0.00001565, 0.00000358, 0.00000900, -0.00001844, -0.00001866, 0.00002590, 0.00003127, -0.00002322, -0.00003463, 0.00001677, 0.00001490, -0.00004226, -0.00001611, 0.00006778, 0.00003148, -0.00007966, -0.00004809, 0.00006636, 0.00002110, -0.00009218, -0.00000757, 0.00012688, 0.00000233, -0.00017044, -0.00002493, 0.00016296, -0.00003628, -0.00025704, 0.00002260, 0.00030031, -0.00016110, -0.00074743, -0.00026801, 0.00090391, 0.00116978, 0.00019599, -0.00062308, -0.00040365, 0.00011922, 0.00011405, -0.00013630, -0.00009823, 0.00006990, 0.00003188, -0.00008210, -0.00004290, 0.00004205, 0.00001561, -0.00003591, -0.00001012, 0.00002148, -0.00000471, -0.00002743, -0.00000505, 0.00001455, 0.00000269, -0.00001063, -0.00000867, -0.00000401, 0.00000080, 0.00000823, 0.00000806, -0.00000252, -0.00001022, -0.00000885, -0.00000395, 0.00000200, 0.00000630, 0.00000212, -0.00000651, -0.00000610, 0.00000200, 0.00000321, -0.00000254, -0.00000167, 0.00000588, 0.00000676, -0.00000115, -0.00000655, -0.00000445, -0.00000045, 0.00000126, 0.00000166, 0.00000148, -0.00000016, -0.00000327, -0.00000609, -0.00000582, },
		{ -0.00000645, -0.00000601, -0.00000047, 0.00000538, 0.00000636, 0.00000332, -0.00000010, -0.00000269, -0.00000373, -0.00000182, 0.00000085, 0.00000011, -0.00000226, -0.00000107, 0.00000162, -0.00000004, -0.00000290, 0.00000060, 0.00000680, 0.00000574, -0.00000100, -0.00000343, -0.00000087, -0.00000078, -0.00000395, -0.00000445, -0.00000060, 0.00000277, 0.00000024, -0.00000719, -0.00000915, 0.00000138, 0.00001144, 0.00000481, -0.00000657, -0.00000120, 0.00000896, 0.00000115, -0.00000817, 0.00000348, 0.00001055, -0.00000939, -0.00001863, 0.00000486, 0.00000985, -0.00002113, -0.00001769, 0.00003181, 0.00003144, -0.00003009, -0.00003617, 0.00002154, 0.00001427, -0.00004623, -0.00001199, 0.00007315, 0.00002561, -0.00008570, -0.00004103, 0.00007160, 0.00001207, -0.00009599, 0.00000437, 0.00012958, -0.00001205, -0.00017085, -0.00000768, 0.00015945, -0.00005831, -0.00024914, 0.00005022, 0.00028225, -0.00020832, -0.00072783, -0.00018381, 0.00092805, 0.00109204, 0.00013214, -0.00059331, -0.00035399, 0.00010932, 0.00008429, -0.00012755, -0.00007782, 0.00006162, 0.00001815, -0.00007264, -0.00003151, 0.00003440, 0.00000734, -0.00003043, -0.00000648, 0.00001531, -0.00000609, -0.00001996, -0.00000281, 0.00000851, -0.00000014, -0.00000792, -0.00000801, -0.00000531, 0.00000284, 0.00001124, 0.00000682, -0.00000578, -0.00001084, -0.00000756, -0.00000336, 0.00000208, 0.00000659, 0.00000226, -0.00000639, -0.00000580, 0.00000173, 0.00000245, -0.00000242, -0.00000067, 0.00000610, 0.00000576, -0.00000202, -0.00000654, -0.00000395, 0.00000011, 0.00000152, 0.00000137, 0.00000110, -0.00000002, -0.00000308, },
		{ 0.00000011, -0.00000298, -0.00000678, -0.00000601, 0.00000040, 0.00000616, 0.00000608, 0.00000246, -0.00000065, -0.00000280, -0.00000359, -0.00000156, 0.00000089, -0.00000021, -0.00000238, -0.00000071, 0.00000183, -0.00000015, -0.00000253, 0.00000135, 0.00000669, 0.00000474, -0.00000147, -0.00000314, -0.00000079, -0.00000075, -0.00000333, -0.00000407, -0.00000118, 0.00000220, -0.00000004, -0.00000819, -0.00000954, 0.00000371, 0.00001375, 0.00000299, -0.00000943, 0.00000030, 0.00001172, -0.00000061, -0.00001080, 0.00000582, 0.00001331, -0.00001196, -0.00002143, 0.00000658, 0.00001042, -0.00002393, -0.00001608, 0.00003767, 0.00003053, -0.00003703, -0.00003665, 0.00002634, 0.00001289, -0.00004961, -0.00000715, 0.00007724, 0.00001873, -0.00009006, -0.00003285, 0.00007499, 0.00000227, -0.00009750, 0.00001656, 0.00012920, -0.00002632, -0.00016746, 0.00000899, 0.00015151, -0.00007880, -0.00023570, 0.00007494, 0.00025616, -0.00025184, -0.00069534, -0.00009939, 0.00093718, 0.00100709, 0.00007805, -0.00055509, -0.00030881, 0.00009399, 0.00005757, -0.00011549, -0.00006014, 0.00005126, 0.00000690, -0.00006188, -0.00002229, 0.00002607, 0.00000090, -0.00002481, -0.00000440, 0.00000943, -0.00000599, -0.00001291, -0.00000188, 0.00000278, -0.00000208, -0.00000553, -0.00000773, -0.00000601, 0.00000515, 0.00001346, 0.00000513, -0.00000856, -0.00001106, -0.00000638, -0.00000279, 0.00000222, 0.00000669, 0.00000227, -0.00000608, -0.00000544, 0.00000125, 0.00000170, -0.00000202, 0.00000034, 0.00000600, 0.00000465, -0.00000267, -0.00000631, -0.00000343, 0.00000057, 0.00000169, 0.00000106, 0.00000078, },
		{ 0.00000073, 0.00000053, 0.00000021, -0.00000299, -0.00000705, -0.00000579, 0.00000132, 0.00000676, 0.00000567, 0.00000161, -0.00000119, -0.00000290, -0.00000334, -0.00000123, 0.00000079, -0.00000063, -0.00000234, -0.00000023, 0.00000190, -0.00000032, -0.00000198, 0.00000212, 0.00000629, 0.00000366, -0.00000167, -0.00000273, -0.00000091, -0.00000075, -0.00000248, -0.00000369, -0.00000201, 0.00000158, -0.00000026, -0.00000918, -0.00000974, 0.00000633, 0.00001580, 0.00000062, -0.00001212, 0.00000244, 0.00001431, -0.00000303, -0.00001321, 0.00000877, 0.00001577, -0.00001511, -0.00002391, 0.00000869, 0.00001064, -0.00002672, -0.00001380, 0.00004327, 0.00002855, -0.00004380, -0.00003603, 0.00003095, 0.00001083, -0.00005226, -0.00000179, 0.00007986, 0.00001116, -0.00009256, -0.00002395, 0.00007637, -0.00000781, -0.00009662, 0.00002841, 0.00012577, -0.00003976, -0.00016045, 0.00002421, 0.00013950, -0.00009670, -0.00021732, 0.00009539, 0.00022296, -0.00028969, -0.00065079, -0.00001748, 0.00093135, 0.00091774, 0.00003455, -0.00051046, -0.00026909, 0.00007454, 0.00003459, -0.00010111, -0.00004557, 0.00003960, -0.00000169, -0.00005051, -0.00001531, 0.00001758, -0.00000377, -0.00001947, -0.00000373, 0.00000420, -0.00000465, -0.00000662, -0.00000207, -0.00000237, -0.00000324, -0.00000356, -0.00000770, -0.00000605, 0.00000755, 0.00001487, 0.00000312, -0.00001081, -0.00001097, -0.00000531, -0.00000223, 0.00000238, 0.00000661, 0.00000223, -0.00000561, -0.00000510, 0.00000060, 0.00000106, -0.00000137, 0.00000125, 0.00000559, 0.00000350, -0.00000310, -0.00000593, -0.00000293, 0.00000093, 0.00000175, },
		{ 0.00000119, 0.00000172, 0.00000042, 0.00000033, 0.00000027, -0.00000309, -0.00000722, -0.00000536, 0.00000224, 0.00000716, 0.00000516, 0.00000080, -0.00000174, -0.00000298, -0.00000294, -0.00000085, 0.00000052, -0.00000112, -0.00000212, 0.00000033, 0.00000180, -0.00000053, -0.00000124, 0.00000285, 0.00000558, 0.00000254, -0.00000156, -0.00000228, -0.00000126, -0.00000071, -0.00000138, -0.00000339, -0.00000311, 0.00000093, -0.00000039, -0.00001015, -0.00000971, 0.00000920, 0.00001749, -0.00000225, -0.00001447, 0.00000518, 0.00001658, -0.00000604, -0.00001526, 0.00001228, 0.00001779, -0.00001875, -0.00002594, 0.00001114, 0.00001045, -0.00002942, -0.00001086, 0.00004842, 0.00002552, -0.00005015, -0.00003435, 0.00003520, 0.00000815, -0.00005405, 0.00000387, 0.00008091, 0.00000326, -0.00009310, -0.00001476, 0.00007570, -0.00001766, -0.00009342, 0.00003931, 0.00011947, -0.00005167, -0.00015019, 0.00003720, 0.00012401, -0.00011113, -0.00019492, 0.00011046, 0.00018393, -0.00032012, -0.00059556, 0.00005935, 0.00091119, 0.00082675, 0.00000194, -0.00046155, -0.00023543, 0.00005237, 0.00001573, -0.00008539, -0.00003427, 0.00002743, -0.00000762, -0.00003914, -0.00001048, 0.00000942, -0.00000683, -0.00001474, -0.00000420, -0.00000009, -0.00000232, -0.00000132, -0.00000315, -0.00000675, -0.00000381, -0.00000209, -0.00000776, -0.00000546, 0.00000988, 0.00001547, 0.00000094, -0.00001252, -0.00001062, -0.00000435, -0.00000169, 0.00000252, 0.00000637, 0.00000217, -0.00000501, -0.00000482, -0.00000017, 0.00000057, -0.00000056, 0.00000200, 0.00000493, 0.00000239, -0.00000329, -0.00000543, -0.00000246, },
		{ -0.00000486, -0.00000203, 0.00000136, 0.00000163, 0.00000015, 0.00000019, 0.00000024, -0.00000327, -0.00000724, -0.00000473, 0.00000310, 0.00000735, 0.00000458, 0.00000001, -0.00000229, -0.00000301, -0.00000240, -0.00000048, 0.00000007, -0.00000161, -0.00000168, 0.00000092, 0.00000150, -0.00000073, -0.00000029, 0.00000350, 0.00000455, 0.00000147, -0.00000114, -0.00000184, -0.00000186, -0.00000056, -0.00000003, -0.00000322, -0.00000447, 0.00000033, -0.00000042, -0.00001110, -0.00000944, 0.00001224, 0.00001875, -0.00000554, -0.00001637, 0.00000843, 0.00001836, -0.00000957, -0.00001677, 0.00001624, 0.00001920, -0.00002281, -0.00002739, 0.00001383, 0.00000979, -0.00003188, -0.00000732, 0.00005294, 0.00002151, -0.00005585, -0.00003167, 0.00003889, 0.00000500, -0.00005489, 0.00000957, 0.00008037, -0.00000461, -0.00009168, -0.00000573, 0.00007304, -0.00002680, -0.00008808, 0.00004873, 0.00011068, -0.00006145, -0.00013719, 0.00004731, 0.00010581, -0.00012139, -0.00016963, 0.00011930, 0.00014067, -0.00034175, -0.00053145, 0.00012882, 0.00087791, 0.00073675, -0.00002000, -0.00041044, -0.00020800, 0.00002887, 0.00000108, -0.00006931, -0.00002617, 0.00001549, -0.00001106, -0.00002837, -0.00000758, 0.00000199, -0.00000854, -0.00001089, -0.00000551, -0.00000326, 0.00000063, 0.00000282, -0.00000483, -0.00001026, -0.00000395, -0.00000116, -0.00000779, -0.00000428, 0.00001199, 0.00001531, -0.00000127, -0.00001370, -0.00001007, -0.00000349, -0.00000116, 0.00000261, 0.00000601, 0.00000216, -0.00000433, -0.00000465, -0.00000098, 0.00000028, 0.00000036, 0.00000253, 0.00000409, 0.00000140, -0.00000329, },
		{ 0.00000057, -0.00000312, -0.00000427, -0.00000166, 0.00000144, 0.00000147, -0.00000008, 0.00000009, 0.00000012, -0.00000350, -0.00000710, -0.00000392, 0.00000387, 0.00000734, 0.00000396, -0.00000074, -0.00000286, -0.00000296, -0.00000171, -0.00000016, -0.00000058, -0.00000206, -0.00000100, 0.00000149, 0.00000099, -0.00000085, 0.00000088, 0.00000399, 0.00000322, 0.00000054, -0.00000040, -0.00000151, -0.00000269, -0.00000025, 0.00000153, -0.00000327, -0.00000608, -0.00000019, -0.00000035, -0.00001202, -0.00000891, 0.00001537, 0.00001949, -0.00000914, -0.00001768, 0.00001208, 0.00001952, -0.00001349, -0.00001763, 0.00002053, 0.00001988, -0.00002713, -0.00002817, 0.00001663, 0.00000866, -0.00003400, -0.00000324, 0.00005663, 0.00001666, -0.00006069, -0.00002812, 0.00004189, 0.00000151, -0.00005475, 0.00001507, 0.00007828, -0.00001207, -0.00008841, 0.00000273, 0.00006857, -0.00003479, -0.00008093, 0.00005622, 0.00009986, -0.00006865, -0.00012216, 0.00005410, 0.00008582, -0.00012704, -0.00014271, 0.00012140, 0.00009502, -0.00035360, -0.00046061, 0.00018907, 0.00083316, 0.00065008, -0.00003199, -0.00035909, -0.00018660, 0.00000535, -0.00000954, -0.00005376, -0.00002102, 0.00000441, -0.00001231, -0.00001864, -0.00000630, -0.00000442, -0.00000922, -0.00000806, -0.00000729, -0.00000524, 0.00000387, 0.00000573, -0.00000684, -0.00001283, -0.00000386, -0.00000075, -0.00000764, -0.00000260, 0.00001374, 0.00001446, -0.00000339, -0.00001438, -0.00000937, -0.00000271, -0.00000066, 0.00000262, 0.00000556, 0.00000219, -0.00000364, -0.00000459, -0.00000176, 0.00000021, 0.00000128, 0.00000280, 0.00000313, },
		{ 0.00000283, 0.00000215, -0.00000007, -0.00000283, -0.00000368, -0.00000134, 0.00000143, 0.00000126, -0.00000025, 0.00000001, -0.00000011, -0.00000376, -0.00000675, -0.00000297, 0.00000450, 0.00000714, 0.00000332, -0.00000146, -0.00000343, -0.00000281, -0.00000090, 0.00000006, -0.00000141, -0.00000241, -0.00000011, 0.00000196, 0.00000027, -0.00000085, 0.00000225, 0.00000425, 0.00000161, -0.00000019, 0.00000062, -0.00000136, -0.00000372, 0.00000031, 0.00000327, -0.00000359, -0.00000789, -0.00000056, -0.00000019, -0.00001291, -0.00000812, 0.00001852, 0.00001967, -0.00001292, -0.00001832, 0.00001599, 0.00001997, -0.00001765, -0.00001772, 0.00002496, 0.00001974, -0.00003155, -0.00002821, 0.00001943, 0.00000705, -0.00003565, 0.00000125, 0.00005935, 0.00001113, -0.00006450, -0.00002384, 0.00004407, -0.00000215, -0.00005364, 0.00002012, 0.00007475, -0.00001882, -0.00008347, 0.00001024, 0.00006259, -0.00004128, -0.00007239, 0.00006143, 0.00008761, -0.00007294, -0.00010585, 0.00005729, 0.00006504, -0.00012792, -0.00011553, 0.00011665, 0.00004889, -0.00035519, -0.00038545, 0.00023872, 0.00077899, 0.00056874, -0.00003511, -0.00030920, -0.00017074, -0.00001706, -0.00001650, -0.00003948, -0.00001840, -0.00000530, -0.00001180, -0.00001029, -0.00000626, -0.00000962, -0.00000920, -0.00000629, -0.00000921, -0.00000605, 0.00000705, 0.00000743, -0.00000888, -0.00001448, -0.00000372, -0.00000080, -0.00000721, -0.00000053, 0.00001502, 0.00001303, -0.00000533, -0.00001460, -0.00000858, -0.00000202, -0.00000020, 0.00000255, 0.00000506, 0.00000227, -0.00000299, -0.00000463, -0.00000243, 0.00000033, 0.00000212, },
		{ 0.00000062, 0.00000281, 0.00000262, 0.00000121, -0.00000052, -0.00000246, -0.00000312, -0.00000107, 0.00000135, 0.00000103, -0.00000035, -0.00000006, -0.00000046, -0.00000401, -0.00000619, -0.00000192, 0.00000496, 0.00000677, 0.00000268, -0.00000215, -0.00000399, -0.00000254, 0.00000003, 0.00000013, -0.00000240, -0.00000259, 0.00000099, 0.00000228, -0.00000065, -0.00000067, 0.00000378, 0.00000423, -0.00000023, -0.00000064, 0.00000188, -0.00000145, -0.00000490, 0.00000117, 0.00000512, -0.00000425, -0.00000987, -0.00000073, 0.00000006, -0.00001375, -0.00000708, 0.00002158, 0.00001927, -0.00001675, -0.00001824, 0.00001998, 0.00001962, -0.00002187, -0.00001699, 0.00002936, 0.00001874, -0.00003591, -0.00002748, 0.00002208, 0.00000502, -0.00003674, 0.00000602, 0.00006099, 0.00000513, -0.00006713, -0.00001903, 0.00004536, -0.00000582, -0.00005163, 0.00002451, 0.00006997, -0.00002456, -0.00007713, 0.00001653, 0.00005544, -0.00004600, -0.00006293, 0.00006418, 0.00007457, -0.00007421, -0.00008907, 0.00005688, 0.00004450, -0.00012417, -0.00008945, 0.00010530, 0.00000424, -0.00034653, -0.00030845, 0.00027693, 0.00071767, 0.00049426, -0.00003076, -0.00026220, -0.00015966, -0.00003742, -0.00002040, -0.00002703, -0.00001779, -0.00001328, -0.00000999, -0.00000353, -0.00000708, -0.00001356, -0.00000884, -0.00000553, -0.00001094, -0.00000581, 0.00000986, 0.00000805, -0.00001070, -0.00001530, -0.00000368, -0.00000120, -0.00000643, 0.00000180, 0.00001576, 0.00001115, -0.00000702, -0.00001441, -0.00000771, -0.00000140, 0.00000019, 0.00000241, 0.00000457, 0.00000238, -0.00000243, -0.00000474, -0.00000293, },
		{ -0.00000488, -0.00000321, 0.00000101, 0.00000329, 0.00000224, 0.00000040, -0.00000078, -0.00000206, -0.00000260, -0.00000085, 0.00000120, 0.00000079, -0.00000039, -0.00000016, -0.00000092, -0.00000421, -0.00000541, -0.00000082, 0.00000522, 0.00000625, 0.00000206, -0.00000282, -0.00000454, -0.00000214, 0.00000102, 0.00000003, -0.00000351, -0.00000256, 0.00000226, 0.00000239, -0.00000173, -0.00000025, 0.00000542, 0.00000388, -0.00000222, -0.00000074, 0.00000331, -0.00000185, -0.00000616, 0.00000236, 0.00000700, -0.00000529, -0.00001193, -0.00000066, 0.00000038, -0.00001453, -0.00000579, 0.00002447, 0.00001828, -0.00002046, -0.00001741, 0.00002389, 0.00001847, -0.00002594, -0.00001542, 0.00003351, 0.00001686, -0.00004000, -0.00002600, 0.00002445, 0.00000263, -0.00003718, 0.00001088, 0.00006146, -0.00000112, -0.00006849, -0.00001389, 0.00004571, -0.00000931, -0.00004880, 0.00002807, 0.00006418, -0.00002910, -0.00006972, 0.00002138, 0.00004753, -0.00004884, -0.00005309, 0.00006441, 0.00006141, -0.00007251, -0.00007264, 0.00005306, 0.00002515, -0.00011619, -0.00006570, 0.00008801, -0.00003712, -0.00032814, -0.00023211, 0.00030341, 0.00065162, 0.00042772, -0.00002053, -0.00021916, -0.00015241, -0.00005505, -0.00002189, -0.00001680, -0.00001861, -0.00001936, -0.00000738, 0.00000157, -0.00000835, -0.00001627, -0.00000842, -0.00000561, -0.00001222, -0.00000472, 0.00001205, 0.00000774, -0.00001210, -0.00001542, -0.00000385, -0.00000185, -0.00000525, 0.00000422, 0.00001592, 0.00000892, -0.00000839, -0.00001388, -0.00000681, -0.00000085, 0.00000051, 0.00000220, 0.00000411, 0.00000251, -0.00000201, },
		{ 0.00000261, -0.00000173, -0.00000499, -0.00000325, 0.00000145, 0.00000350, 0.00000172, -0.00000024, -0.00000087, -0.00000168, -0.00000214, -0.00000066, 0.00000098, 0.00000055, -0.00000036, -0.00000030, -0.00000148, -0.00000431, -0.00000441, 0.00000030, 0.00000527, 0.00000561, 0.00000149, -0.00000348, -0.00000504, -0.00000158, 0.00000205, -0.00000028, -0.00000469, -0.00000227, 0.00000364, 0.00000224, -0.00000291, 0.00000044, 0.00000710, 0.00000316, -0.00000430, -0.00000045, 0.00000483, -0.00000258, -0.00000741, 0.00000391, 0.00000882, -0.00000674, -0.00001399, -0.00000031, 0.00000072, -0.00001525, -0.00000428, 0.00002710, 0.00001674, -0.00002391, -0.00001586, 0.00002752, 0.00001653, -0.00002967, -0.00001305, 0.00003723, 0.00001415, -0.00004365, -0.00002382, 0.00002642, -0.00000002, -0.00003692, 0.00001567, 0.00006074, -0.00000737, -0.00006854, -0.00000864, 0.00004515, -0.00001250, -0.00004530, 0.00003067, 0.00005765, -0.00003229, -0.00006161, 0.00002469, 0.00003929, -0.00004976, -0.00004337, 0.00006223, 0.00004874, -0.00006807, -0.00005728, 0.00004624, 0.00000785, -0.00010465, -0.00004539, 0.00006573, -0.00007354, -0.00030096, -0.00015877, 0.00031839, 0.00058323, 0.00036971, -0.00000610, -0.00018081, -0.00014800, -0.00006952, -0.00002170, -0.00000895, -0.00002029, -0.00002351, -0.00000443, 0.00000509, -0.00000975, -0.00001791, -0.00000819, -0.00000632, -0.00001285, -0.00000303, 0.00001343, 0.00000676, -0.00001292, -0.00001501, -0.00000431, -0.00000260, -0.00000367, 0.00000660, 0.00001551, 0.00000650, -0.00000942, -0.00001306, -0.00000589, -0.00000037, 0.00000074, 0.00000195, 0.00000369, },
		{ 0.00000169, 0.00000334, 0.00000264, -0.00000160, -0.00000501, -0.00000305, 0.00000184, 0.00000345, 0.00000116, -0.00000067, -0.00000084, -0.00000135, -0.00000173, -0.00000051, 0.00000072, 0.00000032, -0.00000028, -0.00000050, -0.00000213, -0.00000430, -0.00000321, 0.00000137, 0.00000510, 0.00000488, 0.00000096, -0.00000411, -0.00000548, -0.00000087, 0.00000306, -0.00000081, -0.00000588, -0.00000171, 0.00000507, 0.00000180, -0.00000412, 0.00000143, 0.00000876, 0.00000205, -0.00000636, 0.00000024, 0.00000634, -0.00000367, -0.00000857, 0.00000583, 0.00001047, -0.00000860, -0.00001597, 0.00000032, 0.00000107, -0.00001590, -0.00000259, 0.00002939, 0.00001471, -0.00002695, -0.00001365, 0.00003071, 0.00001388, -0.00003289, -0.00000997, 0.00004031, 0.00001071, -0.00004669, -0.00002103, 0.00002788, -0.00000281, -0.00003595, 0.00002020, 0.00005884, -0.00001338, -0.00006731, -0.00000348, 0.00004372, -0.00001524, -0.00004129, 0.00003222, 0.00005066, -0.00003408, -0.00005315, 0.00002646, 0.00003114, -0.00004887, -0.00003427, 0.00005789, 0.00003714, -0.00006130, -0.00004363, 0.00003699, -0.00000671, -0.00009041, -0.00002933, 0.00003972, -0.00010368, -0.00026636, -0.00009049, 0.00032262, 0.00051478, 0.00032034, 0.00001086, -0.00014749, -0.00014540, -0.00008068, -0.00002055, -0.00000347, -0.00002229, -0.00002583, -0.00000154, 0.00000716, -0.00001100, -0.00001867, -0.00000830, -0.00000739, -0.00001274, -0.00000106, 0.00001392, 0.00000535, -0.00001310, -0.00001425, -0.00000509, -0.00000331, -0.00000173, 0.00000880, 0.00001456, 0.00000400, -0.00001009, -0.00001202, -0.00000497, 0.00000005, 0.00000088, },
		{ 0.00000038, 0.00000093, 0.00000143, 0.00000305, 0.00000256, -0.00000160, -0.00000489, -0.00000263, 0.00000212, 0.00000313, 0.00000062, -0.00000086, -0.00000073, -0.00000109, -0.00000137, -0.00000039, 0.00000041, 0.00000012, -0.00000016, -0.00000078, -0.00000283, -0.00000413, -0.00000183, 0.00000235, 0.00000471, 0.00000409, 0.00000048, -0.00000473, -0.00000585, -0.00000003, 0.00000400, -0.00000156, -0.00000701, -0.00000086, 0.00000648, 0.00000105, -0.00000531, 0.00000273, 0.00001030, 0.00000055, -0.00000831, 0.00000135, 0.00000775, -0.00000511, -0.00000954, 0.00000808, 0.00001185, -0.00001085, -0.00001776, 0.00000125, 0.00000138, -0.00001646, -0.00000074, 0.00003125, 0.00001225, -0.00002946, -0.00001087, 0.00003330, 0.00001062, -0.00003541, -0.00000629, 0.00004260, 0.00000665, -0.00004897, -0.00001774, 0.00002877, -0.00000563, -0.00003429, 0.00002427, 0.00005584, -0.00001892, -0.00006484, 0.00000138, 0.00004151, -0.00001744, -0.00003695, 0.00003271, 0.00004350, -0.00003450, -0.00004471, 0.00002677, 0.00002346, -0.00004638, -0.00002620, 0.00005174, 0.00002706, -0.00005269, -0.00003217, 0.00002602, -0.00001806, -0.00007447, -0.00001809, 0.00001141, -0.00012659, -0.00022600, -0.00002901, 0.00031722, 0.00044831, 0.00027933, 0.00002882, -0.00011925, -0.00014366, -0.00008859, -0.00001905, -0.00000019, -0.00002414, -0.00002656, 0.00000095, 0.00000804, -0.00001190, -0.00001880, -0.00000882, -0.00000855, -0.00001187, 0.00000092, 0.00001353, 0.00000379, -0.00001263, -0.00001333, -0.00000615, -0.00000387, 0.00000050, 0.00001068, 0.00001312, 0.00000155, -0.00001043, -0.00001080, -0.00000405, },
		{ -0.00000948, -0.00000316, 0.00000064, 0.00000090, 0.00000122, 0.00000280, 0.00000235, -0.00000169, -0.00000457, -0.00000203, 0.00000223, 0.00000258, 0.00000018, -0.00000082, -0.00000059, -0.00000092, -0.00000103, -0.00000029, 0.00000006, -0.00000005, -0.00000002, -0.00000113, -0.00000355, -0.00000378, -0.00000031, 0.00000319, 0.00000410, 0.00000327, 0.00000007, -0.00000532, -0.00000612, 0.00000094, 0.00000484, -0.00000252, -0.00000803, 0.00000028, 0.00000778, -0.00000001, -0.00000638, 0.00000432, 0.00001165, -0.00000131, -0.00001004, 0.00000283, 0.00000895, -0.00000686, -0.00001022, 0.00001061, 0.00001285, -0.00001344, -0.00001927, 0.00000245, 0.00000162, -0.00001692, 0.00000120, 0.00003265, 0.00000948, -0.00003135, -0.00000765, 0.00003517, 0.00000690, -0.00003712, -0.00000218, 0.00004396, 0.00000215, -0.00005040, -0.00001409, 0.00002904, -0.00000832, -0.00003198, 0.00002772, 0.00005186, -0.00002378, -0.00006127, 0.00000576, 0.00003862, -0.00001903, -0.00003246, 0.00003218, 0.00003645, -0.00003365, -0.00003661, 0.00002579, 0.00001658, -0.00004258, -0.00001947, 0.00004423, 0.00001882, -0.00004286, -0.00002318, 0.00001411, -0.00002597, -0.00005792, -0.00001190, -0.00001767, -0.00014169, -0.00018174, 0.00002435, 0.00030369, 0.00038554, 0.00024601, 0.00004640, -0.00009583, -0.00014195, -0.00009348, -0.00001775, 0.00000118, -0.00002550, -0.00002599, 0.00000284, 0.00000800, -0.00001238, -0.00001857, -0.00000973, -0.00000953, -0.00001033, 0.00000263, 0.00001232, 0.00000231, -0.00001156, -0.00001242, -0.00000744, -0.00000415, 0.00000291, 0.00001215, 0.00001129, -0.00000076, -0.00001045, },
		{ -0.00000286, -0.00001019, -0.00000808, -0.00000230, 0.00000082, 0.00000081, 0.00000106, 0.00000256, 0.00000198, -0.00000183, -0.00000403, -0.00000132, 0.00000210, 0.00000184, -0.00000009, -0.00000057, -0.00000048, -0.00000085, -0.00000071, -0.00000021, -0.00000032, -0.00000019, 0.00000014, -0.00000158, -0.00000427, -0.00000324, 0.00000131, 0.00000387, 0.00000329, 0.00000245, -0.00000028, -0.00000589, -0.00000628, 0.00000201, 0.00000554, -0.00000366, -0.00000888, 0.00000167, 0.00000891, -0.00000135, -0.00000727, 0.00000618, 0.00001272, -0.00000349, -0.00001147, 0.00000465, 0.00000985, -0.00000887, -0.00001053, 0.00001336, 0.00001340, -0.00001630, -0.00002042, 0.00000389, 0.00000176, -0.00001726, 0.00000320, 0.00003353, 0.00000651, -0.00003253, -0.00000414, 0.00003624, 0.00000289, -0.00003792, 0.00000216, 0.00004431, -0.00000260, -0.00005089, -0.00001025, 0.00002867, -0.00001078, -0.00002912, 0.00003040, 0.00004705, -0.00002777, -0.00005674, 0.00000952, 0.00003521, -0.00001998, -0.00002799, 0.00003071, 0.00002976, -0.00003167, -0.00002912, 0.00002376, 0.00001074, -0.00003784, -0.00001431, 0.00003587, 0.00001262, -0.00003245, -0.00001679, 0.00000208, -0.00003042, -0.00004182, -0.00001066, -0.00004596, -0.00014884, -0.00013554, 0.00006875, 0.00028371, 0.00032782, 0.00021949, 0.00006250, -0.00007677, -0.00013961, -0.00009576, -0.00001703, 0.00000103, -0.00002614, -0.00002447, 0.00000400, 0.00000732, -0.00001242, -0.00001819, -0.00001093, -0.00001010, -0.00000828, 0.00000383, 0.00001047, 0.00000113, -0.00001001, -0.00001167, -0.00000885, -0.00000409, 0.00000539, 0.00001314, 0.00000917, },
		{ 0.00001362, 0.00000687, -0.00000468, -0.00000970, -0.00000665, -0.00000148, 0.00000092, 0.00000068, 0.00000097, 0.00000230, 0.00000145, -0.00000197, -0.00000325, -0.00000058, 0.00000172, 0.00000100, -0.00000015, -0.00000015, -0.00000043, -0.00000087, -0.00000039, -0.00000016, -0.00000073, -0.00000029, 0.00000029, -0.00000210, -0.00000495, -0.00000250, 0.00000299, 0.00000434, 0.00000231, 0.00000166, -0.00000057, -0.00000641, -0.00000632, 0.00000314, 0.00000605, -0.00000493, -0.00000950, 0.00000327, 0.00000980, -0.00000293, -0.00000790, 0.00000825, 0.00001344, -0.00000591, -0.00001252, 0.00000674, 0.00001039, -0.00001104, -0.00001040, 0.00001621, 0.00001341, -0.00001934, -0.00002112, 0.00000552, 0.00000177, -0.00001748, 0.00000518, 0.00003388, 0.00000346, -0.00003296, -0.00000050, 0.00003644, -0.00000120, -0.00003777, 0.00000654, 0.00004362, -0.00000741, -0.00005042, -0.00000638, 0.00002771, -0.00001287, -0.00002583, 0.00003220, 0.00004161, -0.00003075, -0.00005145, 0.00001255, 0.00003141, -0.00002030, -0.00002372, 0.00002842, 0.00002362, -0.00002879, -0.00002244, 0.00002095, 0.00000608, -0.00003253, -0.00001079, 0.00002717, 0.00000846, -0.00002208, -0.00001293, -0.00000932, -0.00003167, -0.00002717, -0.00001396, -0.00007201, -0.00014833, -0.00008936, 0.00010382, 0.00025909, 0.00027611, 0.00019866, 0.00007629, -0.00006146, -0.00013617, -0.00009592, -0.00001712, -0.00000022, -0.00002596, -0.00002235, 0.00000443, 0.00000628, -0.00001210, -0.00001788, -0.00001225, -0.00001010, -0.00000594, 0.00000438, 0.00000817, 0.00000041, -0.00000814, -0.00001118, -0.00001025, -0.00000366, 0.00000781, },
		{ -0.00000283, 0.00001003, 0.00001357, 0.00000450, -0.00000618, -0.00000903, -0.00000522, -0.00000070, 0.00000093, 0.00000053, 0.00000094, 0.00000199, 0.00000077, -0.00000203, -0.00000225, 0.00000012, 0.00000106, 0.00000013, 0.00000005, 0.00000039, -0.00000049, -0.00000096, -0.00000006, -0.00000013, -0.00000116, -0.00000035, 0.00000043, -0.00000270, -0.00000555, -0.00000155, 0.00000469, 0.00000459, 0.00000118, 0.00000092, -0.00000078, -0.00000690, -0.00000625, 0.00000431, 0.00000636, -0.00000630, -0.00000984, 0.00000503, 0.00001039, -0.00000470, -0.00000822, 0.00001046, 0.00001375, -0.00000850, -0.00001312, 0.00000899, 0.00001049, -0.00001330, -0.00000980, 0.00001906, 0.00001283, -0.00002244, -0.00002133, 0.00000728, 0.00000162, -0.00001757, 0.00000710, 0.00003371, 0.00000046, -0.00003266, 0.00000309, 0.00003579, -0.00000520, -0.00003667, 0.00001072, 0.00004190, -0.00001205, -0.00004901, -0.00000264, 0.00002619, -0.00001451, -0.00002226, 0.00003306, 0.00003577, -0.00003263, -0.00004562, 0.00001478, 0.00002739, -0.00002001, -0.00001976, 0.00002547, 0.00001818, -0.00002523, -0.00001671, 0.00001767, 0.00000267, -0.00002706, -0.00000888, 0.00001863, 0.00000624, -0.00001235, -0.00001134, -0.00001941, -0.00003016, -0.00001482, -0.00002112, -0.00009455, -0.00014082, -0.00004499, 0.00012962, 0.00023161, 0.00023093, 0.00018236, 0.00008724, -0.00004922, -0.00013138, -0.00009445, -0.00001811, -0.00000217, -0.00002496, -0.00001995, 0.00000422, 0.00000508, -0.00001156, -0.00001775, -0.00001349, -0.00000947, -0.00000356, 0.00000418, 0.00000568, 0.00000022, -0.00000614, -0.00001100, -0.00001152, },
		{ -0.00001114, -0.00001253, -0.00000167, 0.00001193, 0.00001304, 0.00000219, -0.00000736, -0.00000823, -0.00000381, 0.00000002, 0.00000088, 0.00000039, 0.00000097, 0.00000159, -0.00000003, -0.00000197, -0.00000106, 0.00000071, 0.00000015, -0.00000068, 0.00000050, 0.00000098, -0.00000068, -0.00000111, 0.00000029, -0.00000013, -0.00000161, -0.00000038, 0.00000053, -0.00000336, -0.00000604, -0.00000042, 0.00000635, 0.00000459, -0.00000008, 0.00000027, -0.00000092, -0.00000734, -0.00000606, 0.00000547, 0.00000646, -0.00000770, -0.00000990, 0.00000688, 0.00001064, -0.00000658, -0.00000819, 0.00001274, 0.00001363, -0.00001116, -0.00001323, 0.00001130, 0.00001013, -0.00001551, -0.00000869, 0.00002177, 0.00001166, -0.00002549, -0.00002101, 0.00000910, 0.00000131, -0.00001751, 0.00000890, 0.00003303, -0.00000238, -0.00003164, 0.00000648, 0.00003432, -0.00000890, -0.00003469, 0.00001452, 0.00003920, -0.00001634, -0.00004673, 0.00000082, 0.00002421, -0.00001563, -0.00001855, 0.00003293, 0.00002974, -0.00003338, -0.00003945, 0.00001618, 0.00002330, -0.00001918, -0.00001624, 0.00002205, 0.00001354, -0.00002124, -0.00001197, 0.00001422, 0.00000048, -0.00002179, -0.00000845, 0.00001069, 0.00000572, -0.00000375, -0.00001165, -0.00002766, -0.00002649, -0.00000543, -0.00003123, -0.00011254, -0.00012729, -0.00000403, 0.00014665, 0.00020297, 0.00019244, 0.00016942, 0.00009509, -0.00003933, -0.00012517, -0.00009187, -0.00001997, -0.00000443, -0.00002327, -0.00001753, 0.00000349, 0.00000389, -0.00001098, -0.00001785, -0.00001442, -0.00000819, -0.00000141, 0.00000324, 0.00000325, 0.00000059, -0.00000419, },
		{ 0.00000147, -0.00000248, -0.00001155, -0.00001315, -0.00000024, 0.00001339, 0.00001209, 0.00000003, -0.00000822, -0.00000734, -0.00000245, 0.00000068, 0.00000077, 0.00000028, 0.00000104, 0.00000109, -0.00000091, -0.00000174, 0.00000027, 0.00000110, -0.00000097, -0.00000136, 0.00000121, 0.00000158, -0.00000101, -0.00000127, 0.00000068, -0.00000017, -0.00000208, -0.00000038, 0.00000059, -0.00000405, -0.00000640, 0.00000089, 0.00000793, 0.00000433, -0.00000142, -0.00000029, -0.00000099, -0.00000772, -0.00000577, 0.00000659, 0.00000635, -0.00000908, -0.00000964, 0.00000874, 0.00001053, -0.00000850, -0.00000779, 0.00001498, 0.00001305, -0.00001377, -0.00001284, 0.00001357, 0.00000929, -0.00001758, -0.00000711, 0.00002423, 0.00000990, -0.00002834, -0.00002014, 0.00001089, 0.00000084, -0.00001730, 0.00001052, 0.00003189, -0.00000493, -0.00002997, 0.00000952, 0.00003212, -0.00001214, -0.00003192, 0.00001775, 0.00003564, -0.00002010, -0.00004366, 0.00000389, 0.00002188, -0.00001619, -0.00001489, 0.00003185, 0.00002377, -0.00003301, -0.00003319, 0.00001676, 0.00001928, -0.00001791, -0.00001322, 0.00001832, 0.00000974, -0.00001708, -0.00000823, 0.00001085, -0.00000063, -0.00001703, -0.00000926, 0.00000373, 0.00000654, 0.00000334, -0.00001337, -0.00003371, -0.00002140, 0.00000059, -0.00004324, -0.00012525, -0.00010899, 0.00003221, 0.00015574, 0.00017466, 0.00016043, 0.00015877, 0.00009985, -0.00003116, -0.00011763, -0.00008862, -0.00002252, -0.00000671, -0.00002106, -0.00001528, 0.00000245, 0.00000279, -0.00001053, -0.00001817, -0.00001485, -0.00000637, 0.00000028, 0.00000165, 0.00000112, },
		{ -0.00000044, -0.00000050, 0.00000274, -0.00000117, -0.00001215, -0.00001330, 0.00000137, 0.00001435, 0.00001079, -0.00000190, -0.00000877, -0.00000641, -0.00000115, 0.00000128, 0.00000061, 0.00000020, 0.00000111, 0.00000046, -0.00000181, -0.00000131, 0.00000167, 0.00000125, -0.00000224, -0.00000182, 0.00000213, 0.00000210, -0.00000149, -0.00000142, 0.00000110, -0.00000026, -0.00000257, -0.00000036, 0.00000062, -0.00000477, -0.00000660, 0.00000235, 0.00000938, 0.00000381, -0.00000282, -0.00000072, -0.00000100, -0.00000805, -0.00000539, 0.00000763, 0.00000604, -0.00001038, -0.00000909, 0.00001054, 0.00001006, -0.00001036, -0.00000701, 0.00001711, 0.00001201, -0.00001623, -0.00001195, 0.00001567, 0.00000802, -0.00001938, -0.00000507, 0.00002629, 0.00000760, -0.00003088, -0.00001875, 0.00001258, 0.00000020, -0.00001695, 0.00001191, 0.00003034, -0.00000710, -0.00002774, 0.00001207, 0.00002931, -0.00001477, -0.00002850, 0.00002024, 0.00003138, -0.00002317, -0.00003996, 0.00000647, 0.00001931, -0.00001619, -0.00001142, 0.00002987, 0.00001807, -0.00003159, -0.00002704, 0.00001658, 0.00001546, -0.00001628, -0.00001075, 0.00001449, 0.00000676, -0.00001295, -0.00000539, 0.00000781, -0.00000083, -0.00001303, -0.00001101, -0.00000201, 0.00000829, 0.00000870, -0.00001594, -0.00003738, -0.00001564, 0.00000308, -0.00005603, -0.00013230, -0.00008729, 0.00006278, 0.00015798, 0.00014794, 0.00013443, 0.00014946, 0.00010175, -0.00002414, -0.00010899, -0.00008506, -0.00002553, -0.00000879, -0.00001856, -0.00001331, 0.00000127, 0.00000179, -0.00001037, -0.00001860, -0.00001460, -0.00000415, 0.00000131, },
		{ -0.00000176, 0.00000153, -0.00000280, -0.00000147, 0.00000424, -0.00000036, -0.00001281, -0.00001292, 0.00000302, 0.00001476, 0.00000924, -0.00000354, -0.00000907, -0.00000546, 0.00000009, 0.00000181, 0.00000043, 0.00000018, 0.00000116, -0.00000030, -0.00000267, -0.00000068, 0.00000306, 0.00000113, -0.00000359, -0.00000202, 0.00000323, 0.00000249, -0.00000209, -0.00000152, 0.00000156, -0.00000040, -0.00000307, -0.00000032, 0.00000060, -0.00000548, -0.00000663, 0.00000393, 0.00001066, 0.00000304, -0.00000423, -0.00000102, -0.00000093, -0.00000834, -0.00000494, 0.00000858, 0.00000555, -0.00001155, -0.00000826, 0.00001220, 0.00000924, -0.00001209, -0.00000587, 0.00001901, 0.00001056, -0.00001843, -0.00001061, 0.00001751, 0.00000634, -0.00002082, -0.00000267, 0.00002786, 0.00000482, -0.00003298, -0.00001687, 0.00001409, -0.00000058, -0.00001645, 0.00001305, 0.00002847, -0.00000882, -0.00002507, 0.00001404, 0.00002603, -0.00001668, -0.00002461, 0.00002190, 0.00002658, -0.00002546, -0.00003577, 0.00000848, 0.00001664, -0.00001566, -0.00000830, 0.00002711, 0.00001284, -0.00002922, -0.00002120, 0.00001572, 0.00001195, -0.00001442, -0.00000882, 0.00001070, 0.00000455, -0.00000906, -0.00000334, 0.00000526, -0.00000035, -0.00000993, -0.00001335, -0.00000638, 0.00001055, 0.00001223, -0.00001881, -0.00003869, -0.00001000, 0.00000214, -0.00006846, -0.00013364, -0.00006366, 0.00008712, 0.00015463, 0.00012374, 0.00011377, 0.00014077, 0.00010115, -0.00001785, -0.00009957, -0.00008142, -0.00002874, -0.00001054, -0.00001601, -0.00001165, 0.00000016, 0.00000083, -0.00001064, -0.00001900, -0.00001356, },
		{ -0.00001916, -0.00001170, 0.00000054, 0.00000090, -0.00000518, -0.00000170, 0.00000578, -0.00000012, -0.00001340, -0.00001200, 0.00000460, 0.00001459, 0.00000755, -0.00000485, -0.00000913, -0.00000454, 0.00000124, 0.00000226, 0.00000025, 0.00000020, 0.00000115, -0.00000116, -0.00000344, 0.00000016, 0.00000436, 0.00000071, -0.00000493, -0.00000192, 0.00000445, 0.00000270, -0.00000280, -0.00000155, 0.00000204, -0.00000062, -0.00000358, -0.00000025, 0.00000054, -0.00000616, -0.00000648, 0.00000560, 0.00001173, 0.00000203, -0.00000561, -0.00000116, -0.00000081, -0.00000858, -0.00000444, 0.00000941, 0.00000492, -0.00001255, -0.00000719, 0.00001367, 0.00000810, -0.00001361, -0.00000441, 0.00002062, 0.00000874, -0.00002028, -0.00000887, 0.00001899, 0.00000434, -0.00002179, 0.00000001, 0.00002882, 0.00000167, -0.00003454, -0.00001457, 0.00001534, -0.00000148, -0.00001582, 0.00001390, 0.00002634, -0.00001004, -0.00002209, 0.00001538, 0.00002245, -0.00001781, -0.00002045, 0.00002265, 0.00002146, -0.00002689, -0.00003126, 0.00000992, 0.00001398, -0.00001467, -0.00000564, 0.00002372, 0.00000824, -0.00002605, -0.00001582, 0.00001430, 0.00000881, -0.00001242, -0.00000741, 0.00000711, 0.00000302, -0.00000554, -0.00000192, 0.00000331, 0.00000056, -0.00000782, -0.00001593, -0.00000936, 0.00001287, 0.00001401, -0.00002145, -0.00003783, -0.00000516, -0.00000187, -0.00007950, -0.00012955, -0.00003952, 0.00010503, 0.00014703, 0.00010268, 0.00009762, 0.00013216, 0.00009854, -0.00001200, -0.00008974, -0.00007786, -0.00003185, -0.00001195, -0.00001361, -0.00001025, -0.00000077, -0.00000021, -0.00001139, },
		{ -0.00000144, -0.00001261, -0.00001890, -0.00000907, 0.00000250, -0.00000055, -0.00000733, -0.00000119, 0.00000715, -0.00000045, -0.00001378, -0.00001057, 0.00000596, 0.00001388, 0.00000584, -0.00000580, -0.00000902, -0.00000365, 0.00000230, 0.00000265, 0.00000007, 0.00000027, 0.00000105, -0.00000211, -0.00000404, 0.00000118, 0.00000549, 0.00000002, -0.00000617, -0.00000149, 0.00000569, 0.00000270, -0.00000358, -0.00000147, 0.00000253, -0.00000091, -0.00000411, -0.00000017, 0.00000044, -0.00000679, -0.00000614, 0.00000731, 0.00001256, 0.00000080, -0.00000693, -0.00000113, -0.00000063, -0.00000879, -0.00000391, 0.00001011, 0.00000418, -0.00001334, -0.00000593, 0.00001487, 0.00000670, -0.00001485, -0.00000270, 0.00002186, 0.00000663, -0.00002171, -0.00000683, 0.00002002, 0.00000211, -0.00002225, 0.00000285, 0.00002913, -0.00000172, -0.00003547, -0.00001194, 0.00001628, -0.00000246, -0.00001508, 0.00001444, 0.00002406, -0.00001074, -0.00001893, 0.00001604, 0.00001875, -0.00001814, -0.00001621, 0.00002247, 0.00001622, -0.00002744, -0.00002660, 0.00001077, 0.00001146, -0.00001331, -0.00000355, 0.00001986, 0.00000440, -0.00002227, -0.00001104, 0.00001243, 0.00000610, -0.00001040, -0.00000648, 0.00000384, 0.00000205, -0.00000249, -0.00000095, 0.00000200, 0.00000163, -0.00000667, -0.00001842, -0.00001103, 0.00001486, 0.00001421, -0.00002341, -0.00003517, -0.00000171, -0.00000840, -0.00008826, -0.00012061, -0.00001619, 0.00011666, 0.00013650, 0.00008508, 0.00008513, 0.00012330, 0.00009443, -0.00000644, -0.00007985, -0.00007441, -0.00003462, -0.00001306, -0.00001155, -0.00000901, -0.00000141, },
		{ -0.00000780, -0.00000176, -0.00000302, -0.00001423, -0.00001802, -0.00000580, 0.00000387, -0.00000271, -0.00000899, -0.00000001, 0.00000816, -0.00000130, -0.00001382, -0.00000869, 0.00000700, 0.00001267, 0.00000419, -0.00000642, -0.00000876, -0.00000281, 0.00000328, 0.00000297, -0.00000008, 0.00000035, 0.00000085, -0.00000310, -0.00000445, 0.00000231, 0.00000637, -0.00000091, -0.00000722, -0.00000076, 0.00000690, 0.00000246, -0.00000438, -0.00000128, 0.00000302, -0.00000130, -0.00000463, -0.00000006, 0.00000032, -0.00000735, -0.00000561, 0.00000904, 0.00001313, -0.00000063, -0.00000814, -0.00000093, -0.00000043, -0.00000898, -0.00000337, 0.00001068, 0.00000337, -0.00001391, -0.00000452, 0.00001577, 0.00000510, -0.00001575, -0.00000080, 0.00002268, 0.00000432, -0.00002264, -0.00000457, 0.00002057, -0.00000024, -0.00002214, 0.00000572, 0.00002872, -0.00000522, -0.00003572, -0.00000908, 0.00001687, -0.00000349, -0.00001424, 0.00001467, 0.00002171, -0.00001091, -0.00001574, 0.00001606, 0.00001510, -0.00001769, -0.00001209, 0.00002140, 0.00001107, -0.00002713, -0.00002196, 0.00001109, 0.00000916, -0.00001168, -0.00000207, 0.00001572, 0.00000140, -0.00001807, -0.00000694, 0.00001026, 0.00000384, -0.00000846, -0.00000595, 0.00000095, 0.00000151, 0.00000004, -0.00000027, 0.00000129, 0.00000265, -0.00000637, -0.00002053, -0.00001157, 0.00001621, 0.00001314, -0.00002436, -0.00003117, -0.00000005, -0.00001674, -0.00009407, -0.00010765, 0.00000523, 0.00012248, 0.00012426, 0.00007095, 0.00007545, 0.00011410, 0.00008931, -0.00000116, -0.00007024, -0.00007100, -0.00003686, -0.00001398, -0.00000991, },
		{ -0.00001487, -0.00000875, -0.00000649, -0.00000186, -0.00000507, -0.00001609, -0.00001639, -0.00000211, 0.00000446, -0.00000539, -0.00000995, 0.00000171, 0.00000865, -0.00000256, -0.00001341, -0.00000649, 0.00000763, 0.00001104, 0.00000270, -0.00000671, -0.00000841, -0.00000203, 0.00000416, 0.00000323, -0.00000019, 0.00000043, 0.00000053, -0.00000408, -0.00000462, 0.00000350, 0.00000694, -0.00000203, -0.00000800, 0.00000026, 0.00000800, 0.00000199, -0.00000516, -0.00000094, 0.00000349, -0.00000178, -0.00000516, 0.00000007, 0.00000019, -0.00000783, -0.00000491, 0.00001073, 0.00001340, -0.00000223, -0.00000920, -0.00000056, -0.00000020, -0.00000917, -0.00000283, 0.00001111, 0.00000252, -0.00001423, -0.00000303, 0.00001635, 0.00000337, -0.00001629, 0.00000120, 0.00002304, 0.00000191, -0.00002306, -0.00000222, 0.00002060, -0.00000258, -0.00002147, 0.00000847, 0.00002761, -0.00000868, -0.00003526, -0.00000610, 0.00001708, -0.00000452, -0.00001332, 0.00001460, 0.00001937, -0.00001059, -0.00001265, 0.00001546, 0.00001165, -0.00001653, -0.00000828, 0.00001950, 0.00000620, -0.00002602, -0.00001748, 0.00001092, 0.00000715, -0.00000990, -0.00000122, 0.00001149, -0.00000075, -0.00001365, -0.00000356, 0.00000793, 0.00000202, -0.00000666, -0.00000575, -0.00000149, 0.00000128, 0.00000204, 0.00000027, 0.00000112, 0.00000341, -0.00000676, -0.00002203, -0.00001123, 0.00001670, 0.00001117, -0.00002408, -0.00002639, -0.00000040, -0.00002608, -0.00009650, -0.00009166, 0.00002384, 0.00012320, 0.00011138, 0.00006007, 0.00006783, 0.00010459, 0.00008362, 0.00000375, -0.00006115, -0.00006753, -0.00003843, },
		{ -0.00006385, -0.00003930, -0.00001585, -0.00000803, -0.00000496, -0.00000182, -0.00000765, -0.00001799, -0.00001391, 0.00000174, 0.00000414, -0.00000834, -0.00001006, 0.00000378, 0.00000849, -0.00000411, -0.00001247, -0.00000407, 0.00000778, 0.00000911, 0.00000142, -0.00000672, -0.00000800, -0.00000130, 0.00000495, 0.00000345, -0.00000027, 0.00000049, 0.00000009, -0.00000500, -0.00000454, 0.00000467, 0.00000717, -0.00000324, -0.00000845, 0.00000152, 0.00000891, 0.00000129, -0.00000587, -0.00000047, 0.00000390, -0.00000236, -0.00000567, 0.00000022, 0.00000004, -0.00000820, -0.00000405, 0.00001235, 0.00001337, -0.00000395, -0.00001008, -0.00000001, 0.00000001, -0.00000936, -0.00000230, 0.00001143, 0.00000168, -0.00001432, -0.00000153, 0.00001660, 0.00000159, -0.00001645, 0.00000321, 0.00002293, -0.00000050, -0.00002295, 0.00000013, 0.00002012, -0.00000481, -0.00002026, 0.00001098, 0.00002583, -0.00001195, -0.00003411, -0.00000311, 0.00001690, -0.00000550, -0.00001235, 0.00001425, 0.00001712, -0.00000984, -0.00000976, 0.00001435, 0.00000856, -0.00001477, -0.00000494, 0.00001689, 0.00000177, -0.00002421, -0.00001331, 0.00001036, 0.00000547, -0.00000810, -0.00000099, 0.00000737, -0.00000204, -0.00000921, -0.00000091, 0.00000557, 0.00000063, -0.00000508, -0.00000579, -0.00000348, 0.00000123, 0.00000357, 0.00000081, 0.00000135, 0.00000378, -0.00000763, -0.00002279, -0.00001030, 0.00001622, 0.00000869, -0.00002252, -0.00002139, -0.00000279, -0.00003557, -0.00009539, -0.00007372, 0.00003908, 0.00011973, 0.00009871, 0.00005203, 0.00006162, 0.00009496, 0.00007773, 0.00000816, -0.00005275, },
		{ 0.00001187, -0.00004514, -0.00005981, -0.00003950, -0.00001706, -0.00000765, -0.00000314, -0.00000180, -0.00001078, -0.00001969, -0.00001058, 0.00000545, 0.00000287, -0.00001126, -0.00000925, 0.00000596, 0.00000760, -0.00000575, -0.00001098, -0.00000161, 0.00000742, 0.00000698, 0.00000042, -0.00000650, -0.00000756, -0.00000063, 0.00000565, 0.00000362, -0.00000031, 0.00000049, -0.00000046, -0.00000579, -0.00000423, 0.00000575, 0.00000705, -0.00000448, -0.00000853, 0.00000294, 0.00000957, 0.00000040, -0.00000647, 0.00000014, 0.00000424, -0.00000305, -0.00000617, 0.00000041, -0.00000010, -0.00000847, -0.00000305, 0.00001386, 0.00001304, -0.00000576, -0.00001074, 0.00000068, 0.00000020, -0.00000957, -0.00000178, 0.00001164, 0.00000087, -0.00001419, -0.00000006, 0.00001651, -0.00000016, -0.00001623, 0.00000515, 0.00002237, -0.00000280, -0.00002233, 0.00000235, 0.00001916, -0.00000681, -0.00001857, 0.00001311, 0.00002343, -0.00001488, -0.00003229, -0.00000024, 0.00001634, -0.00000641, -0.00001136, 0.00001363, 0.00001503, -0.00000871, -0.00000718, 0.00001281, 0.00000592, -0.00001253, -0.00000220, 0.00001372, -0.00000207, -0.00002182, -0.00000953, 0.00000950, 0.00000414, -0.00000639, -0.00000132, 0.00000352, -0.00000256, -0.00000494, 0.00000104, 0.00000330, -0.00000039, -0.00000375, -0.00000598, -0.00000502, 0.00000128, 0.00000469, 0.00000143, 0.00000184, 0.00000369, -0.00000875, -0.00002274, -0.00000909, 0.00001477, 0.00000610, -0.00001977, -0.00001671, -0.00000705, -0.00004441, -0.00009082, -0.00005492, 0.00005063, 0.00011303, 0.00008685, 0.00004632, 0.00005635, 0.00008548, 0.00007186, },
		{ 0.00007644, 0.00006615, 0.00001470, -0.00003828, -0.00005529, -0.00003913, -0.00001859, -0.00000746, -0.00000100, -0.00000197, -0.00001441, -0.00002094, -0.00000647, 0.00000873, 0.00000068, -0.00001385, -0.00000751, 0.00000801, 0.00000599, -0.00000733, -0.00000896, 0.00000076, 0.00000656, 0.00000478, -0.00000030, -0.00000610, -0.00000710, -0.00000000, 0.00000625, 0.00000375, -0.00000033, 0.00000041, -0.00000108, -0.00000640, -0.00000370, 0.00000665, 0.00000659, -0.00000564, -0.00000823, 0.00000444, 0.00000995, -0.00000066, -0.00000691, 0.00000087, 0.00000448, -0.00000385, -0.00000663, 0.00000064, -0.00000023, -0.00000862, -0.00000192, 0.00001522, 0.00001239, -0.00000760, -0.00001115, 0.00000151, 0.00000032, -0.00000980, -0.00000128, 0.00001177, 0.00000011, -0.00001386, 0.00000131, 0.00001612, -0.00000182, -0.00001567, 0.00000693, 0.00002138, -0.00000491, -0.00002124, 0.00000435, 0.00001778, -0.00000848, -0.00001647, 0.00001476, 0.00002051, -0.00001735, -0.00002988, 0.00000242, 0.00001543, -0.00000720, -0.00001037, 0.00001280, 0.00001315, -0.00000730, -0.00000496, 0.00001098, 0.00000381, -0.00000999, -0.00000014, 0.00001015, -0.00000523, -0.00001899, -0.00000621, 0.00000843, 0.00000315, -0.00000486, -0.00000211, 0.00000010, -0.00000239, -0.00000100, 0.00000236, 0.00000122, -0.00000109, -0.00000271, -0.00000625, -0.00000614, 0.00000135, 0.00000549, 0.00000217, 0.00000242, 0.00000311, -0.00000989, -0.00002192, -0.00000789, 0.00001244, 0.00000376, -0.00001603, -0.00001284, -0.00001282, -0.00005187, -0.00008316, -0.00003632, 0.00005847, 0.00010412, 0.00007619, 0.00004233, 0.00005170, },
		{ 0.00003948, 0.00004748, 0.00006812, 0.00006062, 0.00001650, -0.00003210, -0.00005022, -0.00003836, -0.00002046, -0.00000732, 0.00000141, -0.00000250, -0.00001839, -0.00002151, -0.00000177, 0.00001127, -0.00000227, -0.00001583, -0.00000494, 0.00000968, 0.00000372, -0.00000864, -0.00000646, 0.00000288, 0.00000524, 0.00000261, -0.00000073, -0.00000557, -0.00000663, 0.00000058, 0.00000677, 0.00000385, -0.00000034, 0.00000024, -0.00000174, -0.00000680, -0.00000300, 0.00000730, 0.00000583, -0.00000664, -0.00000755, 0.00000596, 0.00001002, -0.00000182, -0.00000717, 0.00000170, 0.00000460, -0.00000474, -0.00000704, 0.00000091, -0.00000035, -0.00000865, -0.00000069, 0.00001639, 0.00001144, -0.00000943, -0.00001128, 0.00000245, 0.00000036, -0.00001006, -0.00000078, 0.00001182, -0.00000056, -0.00001336, 0.00000255, 0.00001545, -0.00000332, -0.00001479, 0.00000849, 0.00002003, -0.00000673, -0.00001975, 0.00000604, 0.00001606, -0.00000974, -0.00001408, 0.00001584, 0.00001720, -0.00001925, -0.00002697, 0.00000476, 0.00001422, -0.00000785, -0.00000942, 0.00001178, 0.00001152, -0.00000570, -0.00000316, 0.00000898, 0.00000228, -0.00000730, 0.00000119, 0.00000638, -0.00000765, -0.00001587, -0.00000340, 0.00000726, 0.00000246, -0.00000361, -0.00000324, -0.00000279, -0.00000167, 0.00000248, 0.00000313, -0.00000058, -0.00000153, -0.00000194, -0.00000655, -0.00000690, 0.00000140, 0.00000605, 0.00000303, 0.00000294, 0.00000210, -0.00001084, -0.00002043, -0.00000694, 0.00000944, 0.00000196, -0.00001163, -0.00001011, -0.00001963, -0.00005739, -0.00007295, -0.00001883, 0.00006286, 0.00009393, 0.00006687, },
		{ 0.00008332, 0.00005883, 0.00003726, 0.00004368, 0.00006074, 0.00005522, 0.00001721, -0.00002647, -0.00004459, -0.00003735, -0.00002263, -0.00000704, 0.00000399, -0.00000355, -0.00002254, -0.00002122, 0.00000329, 0.00001286, -0.00000579, -0.00001695, -0.00000171, 0.00001078, 0.00000091, -0.00000955, -0.00000359, 0.00000463, 0.00000354, 0.00000059, -0.00000090, -0.00000498, -0.00000616, 0.00000114, 0.00000719, 0.00000392, -0.00000036, -0.00000002, -0.00000240, -0.00000695, -0.00000221, 0.00000766, 0.00000483, -0.00000740, -0.00000652, 0.00000739, 0.00000977, -0.00000303, -0.00000722, 0.00000261, 0.00000458, -0.00000572, -0.00000739, 0.00000124, -0.00000044, -0.00000857, 0.00000061, 0.00001735, 0.00001021, -0.00001119, -0.00001113, 0.00000346, 0.00000029, -0.00001033, -0.00000028, 0.00001181, -0.00000116, -0.00001273, 0.00000362, 0.00001456, -0.00000462, -0.00001365, 0.00000975, 0.00001838, -0.00000820, -0.00001795, 0.00000737, 0.00001409, -0.00001055, -0.00001151, 0.00001631, 0.00001364, -0.00002052, -0.00002367, 0.00000672, 0.00001275, -0.00000833, -0.00000852, 0.00001064, 0.00001015, -0.00000400, -0.00000177, 0.00000694, 0.00000131, -0.00000463, 0.00000180, 0.00000257, -0.00000931, -0.00001261, -0.00000111, 0.00000607, 0.00000202, -0.00000269, -0.00000460, -0.00000508, -0.00000054, 0.00000542, 0.00000347, -0.00000206, -0.00000178, -0.00000145, -0.00000681, -0.00000735, 0.00000139, 0.00000647, 0.00000398, 0.00000328, 0.00000077, -0.00001143, -0.00001844, -0.00000642, 0.00000602, 0.00000089, -0.00000694, -0.00000876, -0.00002688, -0.00006057, -0.00006089, -0.00000317, 0.00006423, },
		{ 0.00001014, 0.00006319, 0.00007293, 0.00005188, 0.00003524, 0.00004036, 0.00005447, 0.00004985, 0.00001681, -0.00002121, -0.00003849, -0.00003629, -0.00002499, -0.00000649, 0.00000655, -0.00000522, -0.00002658, -0.00001994, 0.00000841, 0.00001332, -0.00000957, -0.00001703, 0.00000196, 0.00001113, -0.00000228, -0.00000994, -0.00000049, 0.00000591, 0.00000155, -0.00000120, -0.00000084, -0.00000435, -0.00000568, 0.00000167, 0.00000753, 0.00000396, -0.00000042, -0.00000039, -0.00000299, -0.00000684, -0.00000138, 0.00000769, 0.00000367, -0.00000786, -0.00000519, 0.00000868, 0.00000920, -0.00000425, -0.00000704, 0.00000357, 0.00000440, -0.00000677, -0.00000766, 0.00000161, -0.00000052, -0.00000836, 0.00000194, 0.00001805, 0.00000873, -0.00001282, -0.00001069, 0.00000450, 0.00000009, -0.00001060, 0.00000023, 0.00001175, -0.00000167, -0.00001199, 0.00000450, 0.00001350, -0.00000566, -0.00001232, 0.00001068, 0.00001652, -0.00000928, -0.00001592, 0.00000829, 0.00001198, -0.00001088, -0.00000890, 0.00001613, 0.00000997, -0.00002110, -0.00002012, 0.00000823, 0.00001111, -0.00000864, -0.00000770, 0.00000941, 0.00000904, -0.00000230, -0.00000078, 0.00000500, 0.00000088, -0.00000214, 0.00000172, -0.00000109, -0.00001024, -0.00000935, 0.00000070, 0.00000492, 0.00000176, -0.00000210, -0.00000605, -0.00000673, 0.00000084, 0.00000776, 0.00000349, -0.00000318, -0.00000191, -0.00000120, -0.00000701, -0.00000755, 0.00000135, 0.00000683, 0.00000494, 0.00000332, -0.00000077, -0.00001155, -0.00001615, -0.00000645, 0.00000246, 0.00000062, -0.00000239, -0.00000885, -0.00003398, -0.00006125, -0.00004777, },
		{ -0.00005945, -0.00003440, 0.00002080, 0.00006041, 0.00006328, 0.00004572, 0.00003312, 0.00003764, 0.00004935, 0.00004435, 0.00001539, -0.00001618, -0.00003204, -0.00003533, -0.00002738, -0.00000559, 0.00000889, -0.00000753, -0.00003023, -0.00001765, 0.00001324, 0.00001259, -0.00001329, -0.00001598, 0.00000580, 0.00001063, -0.00000563, -0.00000972, 0.00000268, 0.00000666, -0.00000060, -0.00000270, -0.00000061, -0.00000374, -0.00000518, 0.00000219, 0.00000780, 0.00000395, -0.00000054, -0.00000083, -0.00000346, -0.00000649, -0.00000061, 0.00000738, 0.00000243, -0.00000798, -0.00000364, 0.00000974, 0.00000835, -0.00000540, -0.00000664, 0.00000453, 0.00000404, -0.00000788, -0.00000782, 0.00000204, -0.00000057, -0.00000805, 0.00000329, 0.00001849, 0.00000702, -0.00001427, -0.00000996, 0.00000551, -0.00000025, -0.00001086, 0.00000077, 0.00001165, -0.00000210, -0.00001119, 0.00000519, 0.00001231, -0.00000644, -0.00001087, 0.00001126, 0.00001452, -0.00000994, -0.00001376, 0.00000878, 0.00000984, -0.00001073, -0.00000637, 0.00001534, 0.00000636, -0.00002098, -0.00001645, 0.00000928, 0.00000935, -0.00000877, -0.00000698, 0.00000815, 0.00000819, -0.00000067, -0.00000016, 0.00000325, 0.00000092, 0.00000004, 0.00000102, -0.00000446, -0.00001049, -0.00000621, 0.00000204, 0.00000387, 0.00000160, -0.00000186, -0.00000744, -0.00000776, 0.00000230, 0.00000948, 0.00000330, -0.00000394, -0.00000197, -0.00000118, -0.00000712, -0.00000754, 0.00000130, 0.00000717, 0.00000583, 0.00000303, -0.00000233, -0.00001115, -0.00001379, -0.00000704, -0.00000093, 0.00000114, 0.00000162, -0.00001029, -0.00004030, },
		{ -0.00001283, -0.00004531, -0.00005540, -0.00002158, 0.00002878, 0.00005660, 0.00005463, 0.00003999, 0.00003074, 0.00003568, 0.00004533, 0.00003861, 0.00001310, -0.00001123, -0.00002544, -0.00003456, -0.00002961, -0.00000429, 0.00001077, -0.00001045, -0.00003318, -0.00001440, 0.00001744, 0.00001069, -0.00001660, -0.00001378, 0.00000949, 0.00000925, -0.00000892, -0.00000887, 0.00000573, 0.00000685, -0.00000279, -0.00000387, -0.00000026, -0.00000317, -0.00000465, 0.00000269, 0.00000798, 0.00000389, -0.00000073, -0.00000132, -0.00000377, -0.00000593, 0.00000004, 0.00000675, 0.00000120, -0.00000773, -0.00000194, 0.00001054, 0.00000726, -0.00000645, -0.00000603, 0.00000547, 0.00000350, -0.00000901, -0.00000787, 0.00000252, -0.00000060, -0.00000763, 0.00000462, 0.00001863, 0.00000514, -0.00001548, -0.00000897, 0.00000646, -0.00000074, -0.00001109, 0.00000133, 0.00001152, -0.00000245, -0.00001035, 0.00000568, 0.00001106, -0.00000695, -0.00000936, 0.00001148, 0.00001249, -0.00001018, -0.00001157, 0.00000887, 0.00000777, -0.00001014, -0.00000405, 0.00001396, 0.00000294, -0.00002020, -0.00001279, 0.00000986, 0.00000756, -0.00000873, -0.00000636, 0.00000690, 0.00000756, 0.00000082, 0.00000017, 0.00000179, 0.00000133, 0.00000180, -0.00000018, -0.00000740, -0.00001015, -0.00000331, 0.00000299, 0.00000296, 0.00000149, -0.00000193, -0.00000867, -0.00000820, 0.00000371, 0.00001061, 0.00000301, -0.00000436, -0.00000202, -0.00000133, -0.00000714, -0.00000738, 0.00000127, 0.00000752, 0.00000654, 0.00000240, -0.00000378, -0.00001026, -0.00001157, -0.00000812, -0.00000387, 0.00000231, 0.00000474, },
		{ 0.00000390, 0.00000669, -0.00001614, -0.00004858, -0.00004949, -0.00000997, 0.00003421, 0.00005243, 0.00004710, 0.00003436, 0.00002806, 0.00003461, 0.00004229, 0.00003253, 0.00001017, -0.00000630, -0.00001892, -0.00003404, -0.00003145, -0.00000265, 0.00001198, -0.00001384, -0.00003516, -0.00001035, 0.00002067, 0.00000777, -0.00001918, -0.00001052, 0.00001275, 0.00000703, -0.00001192, -0.00000744, 0.00000851, 0.00000651, -0.00000489, -0.00000470, 0.00000013, -0.00000264, -0.00000408, 0.00000319, 0.00000809, 0.00000377, -0.00000101, -0.00000182, -0.00000387, -0.00000521, 0.00000049, 0.00000584, 0.00000009, -0.00000714, -0.00000019, 0.00001104, 0.00000597, -0.00000734, -0.00000521, 0.00000634, 0.00000277, -0.00001015, -0.00000779, 0.00000304, -0.00000061, -0.00000712, 0.00000589, 0.00001848, 0.00000312, -0.00001642, -0.00000773, 0.00000730, -0.00000137, -0.00001124, 0.00000193, 0.00001134, -0.00000274, -0.00000950, 0.00000598, 0.00000980, -0.00000719, -0.00000787, 0.00001136, 0.00001049, -0.00001003, -0.00000945, 0.00000858, 0.00000586, -0.00000918, -0.00000205, 0.00001209, -0.00000014, -0.00001882, -0.00000927, 0.00000999, 0.00000579, -0.00000854, -0.00000584, 0.00000569, 0.00000713, 0.00000212, 0.00000026, 0.00000067, 0.00000200, 0.00000305, -0.00000177, -0.00000983, -0.00000930, -0.00000071, 0.00000360, 0.00000219, 0.00000135, -0.00000224, -0.00000964, -0.00000814, 0.00000495, 0.00001119, 0.00000272, -0.00000448, -0.00000208, -0.00000163, -0.00000707, -0.00000709, 0.00000129, 0.00000788, 0.00000699, 0.00000146, -0.00000495, -0.00000894, -0.00000968, -0.00000958, -0.00000616, },
		{ -0.00001122, -0.00000763, 0.00000565, 0.00000733, -0.00001977, -0.00004983, -0.00004227, -0.00000012, 0.00003745, 0.00004846, 0.00004062, 0.00002855, 0.00002515, 0.00003451, 0.00004001, 0.00002608, 0.00000685, -0.00000135, -0.00001274, -0.00003373, -0.00003270, -0.00000077, 0.00001231, -0.00001749, -0.00003592, -0.00000575, 0.00002266, 0.00000405, -0.00002072, -0.00000639, 0.00001531, 0.00000410, -0.00001443, -0.00000550, 0.00001085, 0.00000569, -0.00000679, -0.00000521, 0.00000052, -0.00000216, -0.00000346, 0.00000367, 0.00000812, 0.00000356, -0.00000139, -0.00000228, -0.00000375, -0.00000438, 0.00000069, 0.00000471, -0.00000083, -0.00000622, 0.00000152, 0.00001122, 0.00000455, -0.00000804, -0.00000422, 0.00000710, 0.00000187, -0.00001125, -0.00000756, 0.00000360, -0.00000062, -0.00000653, 0.00000708, 0.00001802, 0.00000104, -0.00001705, -0.00000629, 0.00000797, -0.00000212, -0.00001131, 0.00000256, 0.00001110, -0.00000298, -0.00000867, 0.00000612, 0.00000857, -0.00000719, -0.00000646, 0.00001093, 0.00000861, -0.00000953, -0.00000746, 0.00000797, 0.00000420, -0.00000792, -0.00000044, 0.00000982, -0.00000278, -0.00001690, -0.00000600, 0.00000970, 0.00000411, -0.00000822, -0.00000544, 0.00000456, 0.00000683, 0.00000320, 0.00000022, -0.00000007, 0.00000282, 0.00000376, -0.00000359, -0.00001167, -0.00000808, 0.00000153, 0.00000393, 0.00000156, 0.00000114, -0.00000272, -0.00001027, -0.00000768, 0.00000592, 0.00001131, 0.00000250, -0.00000433, -0.00000220, -0.00000203, -0.00000692, -0.00000668, 0.00000143, 0.00000823, 0.00000711, 0.00000029, -0.00000573, -0.00000732, -0.00000827, },
		{ -0.00000557, -0.00000741, -0.00001283, -0.00000824, 0.00000723, 0.00000660, -0.00002325, -0.00004893, -0.00003433, 0.00000762, 0.00003893, 0.00004512, 0.00003502, 0.00002235, 0.00002220, 0.00003540, 0.00003825, 0.00001931, 0.00000343, 0.00000354, -0.00000714, -0.00003354, -0.00003317, 0.00000116, 0.00001163, -0.00002115, -0.00003532, -0.00000092, 0.00002321, -0.00000016, -0.00002103, -0.00000164, 0.00001695, 0.00000063, -0.00001627, -0.00000318, 0.00001263, 0.00000447, -0.00000839, -0.00000544, 0.00000086, -0.00000172, -0.00000278, 0.00000415, 0.00000806, 0.00000324, -0.00000185, -0.00000265, -0.00000339, -0.00000354, 0.00000061, 0.00000344, -0.00000148, -0.00000503, 0.00000312, 0.00001110, 0.00000307, -0.00000854, -0.00000310, 0.00000773, 0.00000079, -0.00001229, -0.00000719, 0.00000417, -0.00000061, -0.00000586, 0.00000815, 0.00001726, -0.00000105, -0.00001733, -0.00000471, 0.00000845, -0.00000298, -0.00001126, 0.00000322, 0.00001081, -0.00000317, -0.00000785, 0.00000611, 0.00000741, -0.00000698, -0.00000517, 0.00001023, 0.00000690, -0.00000873, -0.00000569, 0.00000711, 0.00000283, -0.00000645, 0.00000070, 0.00000729, -0.00000490, -0.00001457, -0.00000308, 0.00000906, 0.00000257, -0.00000780, -0.00000514, 0.00000353, 0.00000664, 0.00000403, 0.00000011, -0.00000044, 0.00000365, 0.00000392, -0.00000552, -0.00001290, -0.00000659, 0.00000338, 0.00000406, 0.00000104, 0.00000084, -0.00000328, -0.00001053, -0.00000693, 0.00000655, 0.00001106, 0.00000239, -0.00000399, -0.00000239, -0.00000251, -0.00000671, -0.00000617, 0.00000169, 0.00000851, 0.00000682, -0.00000101, -0.00000605, },
		{ -0.00000230, -0.00000587, -0.00000385, -0.00000711, -0.00001418, -0.00000801, 0.00000834, 0.00000460, -0.00002612, -0.00004595, -0.00002633, 0.00001307, 0.00003920, 0.00004270, 0.00003002, 0.00001567, 0.00001939, 0.00003719, 0.00003674, 0.00001234, 0.00000018, 0.00000825, -0.00000232, -0.00003331, -0.00003276, 0.00000292, 0.00000987, -0.00002447, -0.00003328, 0.00000378, 0.00002224, -0.00000450, -0.00001995, 0.00000341, 0.00001751, -0.00000315, -0.00001733, -0.00000064, 0.00001376, 0.00000298, -0.00000963, -0.00000543, 0.00000111, -0.00000131, -0.00000202, 0.00000462, 0.00000792, 0.00000280, -0.00000238, -0.00000289, -0.00000281, -0.00000276, 0.00000024, 0.00000212, -0.00000184, -0.00000364, 0.00000454, 0.00001069, 0.00000157, -0.00000880, -0.00000187, 0.00000820, -0.00000043, -0.00001323, -0.00000666, 0.00000475, -0.00000060, -0.00000514, 0.00000907, 0.00001622, -0.00000309, -0.00001726, -0.00000303, 0.00000869, -0.00000389, -0.00001107, 0.00000388, 0.00001045, -0.00000333, -0.00000706, 0.00000598, 0.00000635, -0.00000660, -0.00000405, 0.00000932, 0.00000542, -0.00000770, -0.00000416, 0.00000608, 0.00000179, -0.00000490, 0.00000136, 0.00000461, -0.00000643, -0.00001195, -0.00000057, 0.00000813, 0.00000121, -0.00000731, -0.00000492, 0.00000262, 0.00000650, 0.00000463, 0.00000002, -0.00000046, 0.00000438, 0.00000356, -0.00000740, -0.00001352, -0.00000496, 0.00000484, 0.00000404, 0.00000062, 0.00000044, -0.00000385, -0.00001041, -0.00000603, 0.00000681, 0.00001054, 0.00000245, -0.00000350, -0.00000264, -0.00000302, -0.00000645, -0.00000555, 0.00000209, 0.00000867, 0.00000613, },
		{ 0.00000863, 0.00000504, -0.00000347, -0.00000523, -0.00000233, -0.00000733, -0.00001506, -0.00000705, 0.00000872, 0.00000155, -0.00002795, -0.00004111, -0.00001886, 0.00001626, 0.00003880, 0.00004130, 0.00002532, 0.00000855, 0.00001695, 0.00003971, 0.00003524, 0.00000535, -0.00000270, 0.00001260, 0.00000159, -0.00003283, -0.00003141, 0.00000425, 0.00000705, -0.00002715, -0.00002985, 0.00000797, 0.00001975, -0.00000858, -0.00001747, 0.00000842, 0.00001694, -0.00000698, -0.00001755, 0.00000192, 0.00001419, 0.00000134, -0.00001046, -0.00000526, 0.00000125, -0.00000091, -0.00000120, 0.00000507, 0.00000768, 0.00000224, -0.00000294, -0.00000294, -0.00000203, -0.00000212, -0.00000042, 0.00000086, -0.00000186, -0.00000214, 0.00000571, 0.00001002, 0.00000013, -0.00000884, -0.00000059, 0.00000847, -0.00000177, -0.00001403, -0.00000598, 0.00000531, -0.00000059, -0.00000438, 0.00000981, 0.00001492, -0.00000501, -0.00001683, -0.00000133, 0.00000868, -0.00000483, -0.00001072, 0.00000452, 0.00001000, -0.00000346, -0.00000631, 0.00000576, 0.00000540, -0.00000609, -0.00000312, 0.00000826, 0.00000418, -0.00000651, -0.00000291, 0.00000496, 0.00000108, -0.00000335, 0.00000153, 0.00000194, -0.00000737, -0.00000916, 0.00000148, 0.00000700, 0.00000005, -0.00000679, -0.00000479, 0.00000183, 0.00000637, 0.00000502, 0.00000001, -0.00000017, 0.00000492, 0.00000273, -0.00000912, -0.00001357, -0.00000328, 0.00000592, 0.00000390, 0.00000025, -0.00000006, -0.00000432, -0.00000994, -0.00000509, 0.00000671, 0.00000985, 0.00000266, -0.00000294, -0.00000296, -0.00000354, -0.00000614, -0.00000481, 0.00000264, },
		{ -0.00000393, 0.00000330, 0.00000833, 0.00000361, -0.00000439, -0.00000419, -0.00000117, -0.00000796, -0.00001531, -0.00000558, 0.00000815, -0.00000223, -0.00002841, -0.00003479, -0.00001244, 0.00001740, 0.00003824, 0.00004087, 0.00002063, 0.00000112, 0.00001507, 0.00004271, 0.00003354, -0.00000141, -0.00000503, 0.00001637, 0.00000451, -0.00003190, -0.00002917, 0.00000489, 0.00000331, -0.00002885, -0.00002518, 0.00001127, 0.00001587, -0.00001201, -0.00001366, 0.00001303, 0.00001525, -0.00001059, -0.00001692, 0.00000432, 0.00001394, -0.00000031, -0.00001087, -0.00000498, 0.00000127, -0.00000051, -0.00000032, 0.00000549, 0.00000733, 0.00000153, -0.00000351, -0.00000277, -0.00000112, -0.00000168, -0.00000131, -0.00000026, -0.00000156, -0.00000061, 0.00000660, 0.00000916, -0.00000121, -0.00000866, 0.00000071, 0.00000853, -0.00000321, -0.00001467, -0.00000516, 0.00000583, -0.00000058, -0.00000359, 0.00001035, 0.00001340, -0.00000675, -0.00001606, 0.00000032, 0.00000842, -0.00000575, -0.00001020, 0.00000513, 0.00000946, -0.00000356, -0.00000558, 0.00000548, 0.00000457, -0.00000550, -0.00000240, 0.00000710, 0.00000322, -0.00000523, -0.00000193, 0.00000383, 0.00000070, -0.00000192, 0.00000124, -0.00000061, -0.00000772, -0.00000634, 0.00000306, 0.00000575, -0.00000091, -0.00000626, -0.00000471, 0.00000116, 0.00000622, 0.00000523, 0.00000013, 0.00000034, 0.00000519, 0.00000151, -0.00001058, -0.00001312, -0.00000164, 0.00000665, 0.00000369, -0.00000009, -0.00000062, -0.00000464, -0.00000918, -0.00000421, 0.00000629, 0.00000909, 0.00000302, -0.00000234, -0.00000333, -0.00000407, -0.00000580, },
		{ -0.00000457, -0.00000542, -0.00000291, 0.00000404, 0.00000773, 0.00000192, -0.00000497, -0.00000286, -0.00000048, -0.00000882, -0.00001482, -0.00000382, 0.00000652, -0.00000637, -0.00002730, -0.00002745, -0.00000750, 0.00001684, 0.00003795, 0.00004121, 0.00001573, -0.00000637, 0.00001390, 0.00004590, 0.00003152, -0.00000765, -0.00000674, 0.00001932, 0.00000647, -0.00003028, -0.00002615, 0.00000462, -0.00000115, -0.00002931, -0.00001952, 0.00001337, 0.00001086, -0.00001446, -0.00000873, 0.00001691, 0.00001255, -0.00001374, -0.00001551, 0.00000638, 0.00001304, -0.00000186, -0.00001088, -0.00000467, 0.00000117, -0.00000008, 0.00000061, 0.00000588, 0.00000687, 0.00000069, -0.00000404, -0.00000235, -0.00000013, -0.00000152, -0.00000238, -0.00000115, -0.00000097, 0.00000086, 0.00000720, 0.00000816, -0.00000241, -0.00000827, 0.00000197, 0.00000836, -0.00000470, -0.00001510, -0.00000422, 0.00000629, -0.00000057, -0.00000279, 0.00001067, 0.00001169, -0.00000825, -0.00001497, 0.00000187, 0.00000791, -0.00000659, -0.00000952, 0.00000565, 0.00000882, -0.00000361, -0.00000488, 0.00000516, 0.00000386, -0.00000487, -0.00000187, 0.00000592, 0.00000251, -0.00000394, -0.00000122, 0.00000277, 0.00000059, -0.00000069, 0.00000056, -0.00000292, -0.00000754, -0.00000360, 0.00000418, 0.00000445, -0.00000165, -0.00000575, -0.00000467, 0.00000060, 0.00000603, 0.00000532, 0.00000042, 0.00000100, 0.00000513, 0.00000001, -0.00001169, -0.00001223, -0.00000013, 0.00000707, 0.00000342, -0.00000043, -0.00000119, -0.00000475, -0.00000819, -0.00000349, 0.00000559, 0.00000835, 0.00000350, -0.00000177, -0.00000372, },
		{ -0.00000126, -0.00000413, -0.00000503, -0.00000500, -0.00000174, 0.00000479, 0.00000679, 0.00000007, -0.00000515, -0.00000140, -0.00000031, -0.00000973, -0.00001354, -0.00000207, 0.00000382, -0.00001044, -0.00002453, -0.00001966, -0.00000427, 0.00001504, 0.00003822, 0.00004205, 0.00001046, -0.00001363, 0.00001348, 0.00004892, 0.00002913, -0.00001308, -0.00000784, 0.00002124, 0.00000757, -0.00002780, -0.00002258, 0.00000326, -0.00000602, -0.00002833, -0.00001323, 0.00001403, 0.00000505, -0.00001564, -0.00000294, 0.00001977, 0.00000903, -0.00001619, -0.00001346, 0.00000795, 0.00001159, -0.00000318, -0.00001054, -0.00000438, 0.00000099, 0.00000040, 0.00000159, 0.00000622, 0.00000628, -0.00000027, -0.00000447, -0.00000168, 0.00000086, -0.00000165, -0.00000355, -0.00000174, -0.00000014, 0.00000218, 0.00000750, 0.00000707, -0.00000343, -0.00000770, 0.00000315, 0.00000797, -0.00000619, -0.00001531, -0.00000318, 0.00000668, -0.00000056, -0.00000200, 0.00001075, 0.00000985, -0.00000946, -0.00001360, 0.00000325, 0.00000717, -0.00000731, -0.00000870, 0.00000607, 0.00000809, -0.00000361, -0.00000419, 0.00000481, 0.00000326, -0.00000424, -0.00000153, 0.00000476, 0.00000204, -0.00000270, -0.00000074, 0.00000184, 0.00000071, 0.00000027, -0.00000043, -0.00000488, -0.00000688, -0.00000106, 0.00000487, 0.00000317, -0.00000220, -0.00000528, -0.00000465, 0.00000015, 0.00000577, 0.00000532, 0.00000089, 0.00000172, 0.00000471, -0.00000166, -0.00001241, -0.00001099, 0.00000121, 0.00000721, 0.00000311, -0.00000078, -0.00000172, -0.00000462, -0.00000707, -0.00000297, 0.00000470, 0.00000768, 0.00000404, },
		{ 0.00000712, 0.00000461, -0.00000082, -0.00000452, -0.00000545, -0.00000452, -0.00000044, 0.00000549, 0.00000551, -0.00000181, -0.00000491, 0.00000005, -0.00000068, -0.00001049, -0.00001154, -0.00000060, 0.00000016, -0.00001400, -0.00002019, -0.00001198, -0.00000284, 0.00001252, 0.00003922, 0.00004301, 0.00000479, -0.00002032, 0.00001378, 0.00005144, 0.00002643, -0.00001743, -0.00000840, 0.00002194, 0.00000798, -0.00002434, -0.00001871, 0.00000073, -0.00001095, -0.00002580, -0.00000670, 0.00001311, -0.00000116, -0.00001538, 0.00000334, 0.00002143, 0.00000493, -0.00001779, -0.00001093, 0.00000892, 0.00000972, -0.00000421, -0.00000990, -0.00000414, 0.00000073, 0.00000092, 0.00000258, 0.00000651, 0.00000556, -0.00000133, -0.00000476, -0.00000077, 0.00000176, -0.00000210, -0.00000473, -0.00000199, 0.00000087, 0.00000330, 0.00000753, 0.00000596, -0.00000425, -0.00000699, 0.00000421, 0.00000736, -0.00000765, -0.00001528, -0.00000206, 0.00000698, -0.00000055, -0.00000125, 0.00001060, 0.00000794, -0.00001036, -0.00001201, 0.00000441, 0.00000625, -0.00000786, -0.00000775, 0.00000635, 0.00000729, -0.00000356, -0.00000352, 0.00000445, 0.00000276, -0.00000363, -0.00000136, 0.00000366, 0.00000178, -0.00000155, -0.00000044, 0.00000108, 0.00000098, 0.00000091, -0.00000164, -0.00000643, -0.00000585, 0.00000121, 0.00000518, 0.00000198, -0.00000259, -0.00000487, -0.00000463, -0.00000022, 0.00000544, 0.00000528, 0.00000154, 0.00000241, 0.00000394, -0.00000340, -0.00001271, -0.00000951, 0.00000235, 0.00000714, 0.00000276, -0.00000115, -0.00000216, -0.00000426, -0.00000592, -0.00000270, 0.00000371, },
		{ -0.00000268, 0.00000270, 0.00000670, 0.00000515, -0.00000047, -0.00000488, -0.00000582, -0.00000398, 0.00000098, 0.00000607, 0.00000391, -0.00000360, -0.00000427, 0.00000133, -0.00000153, -0.00001092, -0.00000891, 0.00000034, -0.00000428, -0.00001668, -0.00001450, -0.00000497, -0.00000313, 0.00000980, 0.00004096, 0.00004375, -0.00000121, -0.00002612, 0.00001466, 0.00005315, 0.00002353, -0.00002049, -0.00000861, 0.00002130, 0.00000795, -0.00001985, -0.00001486, -0.00000298, -0.00001556, -0.00002173, -0.00000037, 0.00001059, -0.00000731, -0.00001360, 0.00000972, 0.00002175, 0.00000055, -0.00001843, -0.00000813, 0.00000922, 0.00000755, -0.00000489, -0.00000906, -0.00000400, 0.00000045, 0.00000149, 0.00000358, 0.00000674, 0.00000470, -0.00000245, -0.00000485, 0.00000035, 0.00000249, -0.00000286, -0.00000582, -0.00000187, 0.00000198, 0.00000415, 0.00000731, 0.00000488, -0.00000486, -0.00000616, 0.00000511, 0.00000654, -0.00000903, -0.00001499, -0.00000090, 0.00000718, -0.00000053, -0.00000055, 0.00001021, 0.00000600, -0.00001090, -0.00001027, 0.00000531, 0.00000518, -0.00000822, -0.00000671, 0.00000647, 0.00000641, -0.00000343, -0.00000286, 0.00000409, 0.00000235, -0.00000308, -0.00000132, 0.00000267, 0.00000168, -0.00000052, -0.00000029, 0.00000051, 0.00000134, 0.00000121, -0.00000296, -0.00000753, -0.00000454, 0.00000313, 0.00000518, 0.00000092, -0.00000284, -0.00000453, -0.00000460, -0.00000052, 0.00000505, 0.00000526, 0.00000232, 0.00000299, 0.00000285, -0.00000509, -0.00001260, -0.00000788, 0.00000326, 0.00000687, 0.00000236, -0.00000152, -0.00000246, -0.00000369, -0.00000484, },
		{ -0.00000296, -0.00000392, -0.00000287, 0.00000176, 0.00000641, 0.00000560, -0.00000022, -0.00000517, -0.00000612, -0.00000336, 0.00000247, 0.00000646, 0.00000204, -0.00000519, -0.00000329, 0.00000230, -0.00000276, -0.00001085, -0.00000586, 0.00000055, -0.00000919, -0.00001817, -0.00000782, 0.00000091, -0.00000490, 0.00000737, 0.00004330, 0.00004390, -0.00000736, -0.00003073, 0.00001591, 0.00005381, 0.00002063, -0.00002212, -0.00000868, 0.00001931, 0.00000775, -0.00001438, -0.00001135, -0.00000776, -0.00001945, -0.00001626, 0.00000530, 0.00000658, -0.00001295, -0.00001037, 0.00001578, 0.00002073, -0.00000380, -0.00001809, -0.00000527, 0.00000883, 0.00000527, -0.00000519, -0.00000809, -0.00000394, 0.00000018, 0.00000212, 0.00000455, 0.00000688, 0.00000371, -0.00000360, -0.00000472, 0.00000161, 0.00000298, -0.00000388, -0.00000673, -0.00000140, 0.00000308, 0.00000471, 0.00000691, 0.00000387, -0.00000527, -0.00000526, 0.00000582, 0.00000552, -0.00001027, -0.00001444, 0.00000027, 0.00000726, -0.00000051, 0.00000008, 0.00000959, 0.00000411, -0.00001109, -0.00000844, 0.00000592, 0.00000403, -0.00000835, -0.00000564, 0.00000640, 0.00000550, -0.00000321, -0.00000220, 0.00000373, 0.00000200, -0.00000260, -0.00000139, 0.00000181, 0.00000172, 0.00000035, -0.00000022, 0.00000015, 0.00000169, 0.00000118, -0.00000429, -0.00000815, -0.00000308, 0.00000467, 0.00000494, 0.00000001, -0.00000300, -0.00000426, -0.00000454, -0.00000076, 0.00000462, 0.00000529, 0.00000321, 0.00000338, 0.00000149, -0.00000663, -0.00001210, -0.00000617, 0.00000395, 0.00000646, 0.00000192, -0.00000186, -0.00000256, },
		{ -0.00000216, -0.00000244, -0.00000212, -0.00000321, -0.00000323, 0.00000096, 0.00000623, 0.00000592, -0.00000004, -0.00000539, -0.00000634, -0.00000266, 0.00000399, 0.00000662, -0.00000002, -0.00000647, -0.00000207, 0.00000286, -0.00000422, -0.00001019, -0.00000261, -0.00000010, -0.00001422, -0.00001826, -0.00000058, 0.00000530, -0.00000777, 0.00000563, 0.00004600, 0.00004319, -0.00001339, -0.00003396, 0.00001724, 0.00005324, 0.00001800, -0.00002226, -0.00000886, 0.00001601, 0.00000767, -0.00000806, -0.00000850, -0.00001339, -0.00002227, -0.00000964, 0.00000994, 0.00000130, -0.00001766, -0.00000584, 0.00002113, 0.00001845, -0.00000780, -0.00001681, -0.00000258, 0.00000780, 0.00000300, -0.00000514, -0.00000709, -0.00000398, -0.00000006, 0.00000277, 0.00000548, 0.00000693, 0.00000260, -0.00000472, -0.00000434, 0.00000297, 0.00000317, -0.00000510, -0.00000736, -0.00000060, 0.00000411, 0.00000496, 0.00000636, 0.00000298, -0.00000547, -0.00000433, 0.00000630, 0.00000435, -0.00001132, -0.00001365, 0.00000140, 0.00000722, -0.00000047, 0.00000061, 0.00000877, 0.00000232, -0.00001093, -0.00000659, 0.00000624, 0.00000285, -0.00000825, -0.00000456, 0.00000613, 0.00000458, -0.00000290, -0.00000156, 0.00000338, 0.00000170, -0.00000220, -0.00000153, 0.00000108, 0.00000184, 0.00000106, -0.00000019, -0.00000003, 0.00000198, 0.00000084, -0.00000553, -0.00000833, -0.00000156, 0.00000581, 0.00000453, -0.00000072, -0.00000308, -0.00000406, -0.00000445, -0.00000097, 0.00000416, 0.00000539, 0.00000414, 0.00000354, -0.00000009, -0.00000795, -0.00001126, -0.00000448, 0.00000442, 0.00000592, 0.00000143, },
		{ 0.00000529, 0.00000090, -0.00000236, -0.00000209, -0.00000127, -0.00000278, -0.00000369, 0.00000037, 0.00000612, 0.00000607, 0.00000008, -0.00000551, -0.00000648, -0.00000187, 0.00000548, 0.00000649, -0.00000216, -0.00000737, -0.00000074, 0.00000293, -0.00000574, -0.00000889, 0.00000057, -0.00000164, -0.00001897, -0.00001688, 0.00000671, 0.00000798, -0.00001128, 0.00000484, 0.00004871, 0.00004145, -0.00001896, -0.00003569, 0.00001833, 0.00005141, 0.00001588, -0.00002096, -0.00000939, 0.00001156, 0.00000797, -0.00000112, -0.00000656, -0.00001957, -0.00002375, -0.00000221, 0.00001320, -0.00000490, -0.00002110, -0.00000029, 0.00002540, 0.00001508, -0.00001119, -0.00001472, -0.00000024, 0.00000620, 0.00000089, -0.00000478, -0.00000612, -0.00000407, -0.00000022, 0.00000344, 0.00000635, 0.00000689, 0.00000138, -0.00000576, -0.00000369, 0.00000433, 0.00000303, -0.00000645, -0.00000767, 0.00000046, 0.00000497, 0.00000491, 0.00000573, 0.00000223, -0.00000549, -0.00000340, 0.00000655, 0.00000305, -0.00001215, -0.00001262, 0.00000246, 0.00000707, -0.00000043, 0.00000103, 0.00000776, 0.00000068, -0.00001044, -0.00000479, 0.00000627, 0.00000170, -0.00000793, -0.00000355, 0.00000567, 0.00000368, -0.00000249, -0.00000093, 0.00000302, 0.00000143, -0.00000188, -0.00000171, 0.00000049, 0.00000200, 0.00000162, -0.00000014, -0.00000004, 0.00000215, 0.00000024, -0.00000658, -0.00000809, -0.00000008, 0.00000657, 0.00000404, -0.00000128, -0.00000314, -0.00000392, -0.00000432, -0.00000115, 0.00000371, 0.00000558, 0.00000504, 0.00000344, -0.00000178, -0.00000899, -0.00001013, -0.00000285, 0.00000469, },
		{ -0.00000135, 0.00000479, 0.00000457, 0.00000034, -0.00000242, -0.00000153, -0.00000049, -0.00000263, -0.00000416, 0.00000001, 0.00000602, 0.00000605, 0.00000015, -0.00000553, -0.00000652, -0.00000099, 0.00000687, 0.00000606, -0.00000429, -0.00000787, 0.00000057, 0.00000252, -0.00000713, -0.00000698, 0.00000340, -0.00000399, -0.00002306, -0.00001407, 0.00001352, 0.00000887, -0.00001493, 0.00000514, 0.00005104, 0.00003861, -0.00002372, -0.00003593, 0.00001886, 0.00004834, 0.00001454, -0.00001837, -0.00001049, 0.00000622, 0.00000885, 0.00000612, -0.00000571, -0.00002593, -0.00002369, 0.00000560, 0.00001487, -0.00001160, -0.00002298, 0.00000591, 0.00002833, 0.00001089, -0.00001374, -0.00001197, 0.00000159, 0.00000416, -0.00000094, -0.00000418, -0.00000524, -0.00000419, -0.00000029, 0.00000409, 0.00000713, 0.00000673, 0.00000007, -0.00000667, -0.00000279, 0.00000561, 0.00000253, -0.00000782, -0.00000759, 0.00000172, 0.00000561, 0.00000459, 0.00000508, 0.00000163, -0.00000534, -0.00000251, 0.00000655, 0.00000167, -0.00001271, -0.00001139, 0.00000342, 0.00000680, -0.00000037, 0.00000132, 0.00000661, -0.00000075, -0.00000965, -0.00000311, 0.00000604, 0.00000065, -0.00000741, -0.00000265, 0.00000503, 0.00000284, -0.00000198, -0.00000034, 0.00000267, 0.00000117, -0.00000164, -0.00000190, 0.00000004, 0.00000217, 0.00000205, -0.00000005, 0.00000007, 0.00000214, -0.00000056, -0.00000739, -0.00000751, 0.00000125, 0.00000698, 0.00000352, -0.00000169, -0.00000319, -0.00000381, -0.00000416, -0.00000131, 0.00000329, 0.00000586, 0.00000587, 0.00000305, -0.00000351, -0.00000971, -0.00000880, },
		{ -0.00001009, -0.00000733, -0.00000001, 0.00000475, 0.00000379, -0.00000022, -0.00000231, -0.00000078, 0.00000013, -0.00000277, -0.00000457, -0.00000010, 0.00000590, 0.00000583, 0.00000022, -0.00000543, -0.00000645, -0.00000005, 0.00000810, 0.00000535, -0.00000629, -0.00000796, 0.00000173, 0.00000164, -0.00000821, -0.00000456, 0.00000565, -0.00000701, -0.00002613, -0.00001004, 0.00001935, 0.00000809, -0.00001819, 0.00000653, 0.00005257, 0.00003473, -0.00002731, -0.00003478, 0.00001852, 0.00004421, 0.00001419, -0.00001473, -0.00001234, 0.00000030, 0.00001046, 0.00001331, -0.00000605, -0.00003203, -0.00002202, 0.00001331, 0.00001485, -0.00001832, -0.00002318, 0.00001232, 0.00002974, 0.00000619, -0.00001529, -0.00000881, 0.00000279, 0.00000183, -0.00000243, -0.00000343, -0.00000450, -0.00000429, -0.00000025, 0.00000471, 0.00000780, 0.00000646, -0.00000128, -0.00000739, -0.00000167, 0.00000673, 0.00000167, -0.00000912, -0.00000712, 0.00000308, 0.00000597, 0.00000405, 0.00000445, 0.00000121, -0.00000505, -0.00000171, 0.00000629, 0.00000027, -0.00001298, -0.00000998, 0.00000424, 0.00000644, -0.00000031, 0.00000148, 0.00000535, -0.00000194, -0.00000862, -0.00000160, 0.00000558, -0.00000026, -0.00000671, -0.00000190, 0.00000423, 0.00000209, -0.00000139, 0.00000021, 0.00000231, 0.00000093, -0.00000146, -0.00000207, -0.00000030, 0.00000231, 0.00000237, 0.00000012, 0.00000025, 0.00000195, -0.00000147, -0.00000789, -0.00000665, 0.00000239, 0.00000709, 0.00000302, -0.00000196, -0.00000327, -0.00000374, -0.00000395, -0.00000145, 0.00000293, 0.00000621, 0.00000654, 0.00000240, -0.00000518, },
		{ 0.00000151, -0.00000671, -0.00001013, -0.00000580, 0.00000116, 0.00000457, 0.00000295, -0.00000076, -0.00000200, 0.00000010, 0.00000053, -0.00000314, -0.00000483, 0.00000001, 0.00000569, 0.00000543, 0.00000030, -0.00000523, -0.00000627, 0.00000094, 0.00000910, 0.00000438, -0.00000805, -0.00000769, 0.00000259, 0.00000039, -0.00000883, -0.00000177, 0.00000712, -0.00001044, -0.00002788, -0.00000508, 0.00002378, 0.00000586, -0.00002062, 0.00000883, 0.00005296, 0.00003001, -0.00002944, -0.00003247, 0.00001708, 0.00003925, 0.00001498, -0.00001036, -0.00001501, -0.00000584, 0.00001284, 0.00002005, -0.00000756, -0.00003743, -0.00001878, 0.00002044, 0.00001316, -0.00002458, -0.00002169, 0.00001851, 0.00002956, 0.00000134, -0.00001576, -0.00000547, 0.00000331, -0.00000061, -0.00000352, -0.00000263, -0.00000391, -0.00000431, -0.00000012, 0.00000526, 0.00000835, 0.00000609, -0.00000264, -0.00000788, -0.00000037, 0.00000761, 0.00000051, -0.00001025, -0.00000626, 0.00000445, 0.00000603, 0.00000335, 0.00000389, 0.00000094, -0.00000466, -0.00000103, 0.00000577, -0.00000112, -0.00001294, -0.00000845, 0.00000490, 0.00000598, -0.00000024, 0.00000150, 0.00000404, -0.00000286, -0.00000740, -0.00000030, 0.00000494, -0.00000099, -0.00000590, -0.00000134, 0.00000331, 0.00000147, -0.00000071, 0.00000072, 0.00000195, 0.00000070, -0.00000134, -0.00000222, -0.00000054, 0.00000241, 0.00000261, 0.00000037, 0.00000046, 0.00000155, -0.00000242, -0.00000808, -0.00000562, 0.00000329, 0.00000696, 0.00000259, -0.00000214, -0.00000337, -0.00000367, -0.00000371, -0.00000157, 0.00000265, 0.00000661, 0.00000700, },
		{ 0.00000702, 0.00000722, 0.00000044, -0.00000803, -0.00000987, -0.00000428, 0.00000215, 0.00000430, 0.00000207, -0.00000123, -0.00000146, 0.00000104, 0.00000065, -0.00000371, -0.00000488, 0.00000032, 0.00000534, 0.00000487, 0.00000043, -0.00000490, -0.00000596, 0.00000194, 0.00000983, 0.00000319, -0.00000948, -0.00000712, 0.00000308, -0.00000111, -0.00000888, 0.00000118, 0.00000766, -0.00001399, -0.00002813, 0.00000041, 0.00002648, 0.00000257, -0.00002183, 0.00001179, 0.00005190, 0.00002474, -0.00002989, -0.00002930, 0.00001438, 0.00003378, 0.00001697, -0.00000560, -0.00001850, -0.00001181, 0.00001592, 0.00002598, -0.00001010, -0.00004173, -0.00001417, 0.00002651, 0.00000998, -0.00002989, -0.00001862, 0.00002401, 0.00002782, -0.00000331, -0.00001517, -0.00000220, 0.00000316, -0.00000298, -0.00000422, -0.00000186, -0.00000347, -0.00000423, 0.00000010, 0.00000572, 0.00000877, 0.00000560, -0.00000396, -0.00000811, 0.00000105, 0.00000817, -0.00000091, -0.00001112, -0.00000504, 0.00000573, 0.00000578, 0.00000254, 0.00000342, 0.00000082, -0.00000419, -0.00000050, 0.00000503, -0.00000243, -0.00001257, -0.00000684, 0.00000538, 0.00000546, -0.00000018, 0.00000137, 0.00000271, -0.00000350, -0.00000605, 0.00000075, 0.00000418, -0.00000152, -0.00000502, -0.00000099, 0.00000231, 0.00000100, 0.00000002, 0.00000115, 0.00000158, 0.00000047, -0.00000127, -0.00000232, -0.00000070, 0.00000246, 0.00000281, 0.00000070, 0.00000064, 0.00000097, -0.00000332, -0.00000796, -0.00000449, 0.00000391, 0.00000666, 0.00000224, -0.00000224, -0.00000350, -0.00000360, -0.00000344, -0.00000165, 0.00000248, },
		{ -0.00000169, 0.00000242, 0.00000739, 0.00000715, -0.00000076, -0.00000909, -0.00000934, -0.00000283, 0.00000295, 0.00000393, 0.00000116, -0.00000159, -0.00000071, 0.00000195, 0.00000045, -0.00000438, -0.00000467, 0.00000078, 0.00000483, 0.00000419, 0.00000062, -0.00000447, -0.00000554, 0.00000290, 0.00001023, 0.00000188, -0.00001052, -0.00000636, 0.00000311, -0.00000270, -0.00000831, 0.00000405, 0.00000723, -0.00001730, -0.00002678, 0.00000597, 0.00002726, -0.00000135, -0.00002158, 0.00001501, 0.00004922, 0.00001926, -0.00002855, -0.00002565, 0.00001038, 0.00002815, 0.00002011, -0.00000083, -0.00002270, -0.00001725, 0.00001957, 0.00003074, -0.00001342, -0.00004456, -0.00000847, 0.00003115, 0.00000559, -0.00003384, -0.00001421, 0.00002842, 0.00002470, -0.00000742, -0.00001360, 0.00000075, 0.00000239, -0.00000515, -0.00000456, -0.00000122, -0.00000314, -0.00000400, 0.00000039, 0.00000606, 0.00000905, 0.00000502, -0.00000519, -0.00000805, 0.00000251, 0.00000838, -0.00000250, -0.00001165, -0.00000354, 0.00000684, 0.00000525, 0.00000169, 0.00000307, 0.00000084, -0.00000367, -0.00000015, 0.00000407, -0.00000361, -0.00001189, -0.00000519, 0.00000567, 0.00000488, -0.00000012, 0.00000112, 0.00000141, -0.00000385, -0.00000465, 0.00000155, 0.00000336, -0.00000183, -0.00000413, -0.00000087, 0.00000128, 0.00000068, 0.00000078, 0.00000152, 0.00000122, 0.00000024, -0.00000122, -0.00000237, -0.00000080, 0.00000246, 0.00000298, 0.00000109, 0.00000075, 0.00000023, -0.00000412, -0.00000754, -0.00000337, 0.00000427, 0.00000625, 0.00000199, -0.00000232, -0.00000367, -0.00000351, -0.00000315, },
		{ -0.00000338, -0.00000284, -0.00000168, 0.00000246, 0.00000768, 0.00000680, -0.00000202, -0.00000983, -0.00000859, -0.00000149, 0.00000357, 0.00000350, 0.00000026, -0.00000181, 0.00000025, 0.00000276, -0.00000009, -0.00000508, -0.00000417, 0.00000131, 0.00000412, 0.00000342, 0.00000089, -0.00000393, -0.00000501, 0.00000379, 0.00001029, 0.00000051, -0.00001112, -0.00000551, 0.00000267, -0.00000422, -0.00000711, 0.00000659, 0.00000586, -0.00002004, -0.00002388, 0.00001112, 0.00002607, -0.00000539, -0.00001975, 0.00001809, 0.00004487, 0.00001399, -0.00002542, -0.00002191, 0.00000513, 0.00002272, 0.00002427, 0.00000361, -0.00002742, -0.00002182, 0.00002353, 0.00003407, -0.00001720, -0.00004566, -0.00000208, 0.00003404, 0.00000039, -0.00003612, -0.00000883, 0.00003141, 0.00002045, -0.00001072, -0.00001123, 0.00000319, 0.00000113, -0.00000696, -0.00000461, -0.00000077, -0.00000290, -0.00000360, 0.00000072, 0.00000626, 0.00000918, 0.00000436, -0.00000628, -0.00000770, 0.00000393, 0.00000820, -0.00000417, -0.00001180, -0.00000183, 0.00000769, 0.00000446, 0.00000086, 0.00000285, 0.00000097, -0.00000313, 0.00000001, 0.00000293, -0.00000461, -0.00001090, -0.00000357, 0.00000576, 0.00000427, -0.00000009, 0.00000074, 0.00000019, -0.00000395, -0.00000324, 0.00000211, 0.00000255, -0.00000195, -0.00000330, -0.00000096, 0.00000027, 0.00000054, 0.00000154, 0.00000181, 0.00000086, 0.00000001, -0.00000119, -0.00000238, -0.00000087, 0.00000242, 0.00000315, 0.00000152, 0.00000076, -0.00000062, -0.00000475, -0.00000688, -0.00000232, 0.00000438, 0.00000580, 0.00000182, -0.00000240, -0.00000385, },
		{ -0.00000251, -0.00000404, -0.00000322, -0.00000252, -0.00000159, 0.00000261, 0.00000782, 0.00000615, -0.00000326, -0.00001024, -0.00000769, -0.00000030, 0.00000404, 0.00000301, -0.00000063, -0.00000184, 0.00000137, 0.00000338, -0.00000094, -0.00000570, -0.00000338, 0.00000184, 0.00000322, 0.00000262, 0.00000124, -0.00000329, -0.00000437, 0.00000456, 0.00001001, -0.00000082, -0.00001128, -0.00000469, 0.00000179, -0.00000551, -0.00000537, 0.00000859, 0.00000368, -0.00002186, -0.00001960, 0.00001541, 0.00002303, -0.00000904, -0.00001638, 0.00002057, 0.00003893, 0.00000934, -0.00002065, -0.00001845, -0.00000118, 0.00001783, 0.00002920, 0.00000742, -0.00003239, -0.00002529, 0.00002749, 0.00003579, -0.00002103, -0.00004488, 0.00000454, 0.00003502, -0.00000517, -0.00003653, -0.00000290, 0.00003275, 0.00001539, -0.00001299, -0.00000826, 0.00000497, -0.00000048, -0.00000832, -0.00000446, -0.00000056, -0.00000268, -0.00000302, 0.00000105, 0.00000632, 0.00000917, 0.00000365, -0.00000717, -0.00000708, 0.00000523, 0.00000763, -0.00000582, -0.00001153, -0.00000000, 0.00000824, 0.00000348, 0.00000011, 0.00000275, 0.00000119, -0.00000262, -0.00000004, 0.00000166, -0.00000538, -0.00000964, -0.00000203, 0.00000566, 0.00000365, -0.00000008, 0.00000026, -0.00000091, -0.00000380, -0.00000190, 0.00000244, 0.00000180, -0.00000188, -0.00000257, -0.00000126, -0.00000066, 0.00000056, 0.00000228, 0.00000201, 0.00000052, -0.00000021, -0.00000118, -0.00000234, -0.00000091, 0.00000234, 0.00000334, 0.00000197, 0.00000064, -0.00000153, -0.00000516, -0.00000603, -0.00000140, 0.00000429, 0.00000534, 0.00000171, },
		{ 0.00000493, 0.00000164, -0.00000265, -0.00000420, -0.00000301, -0.00000220, -0.00000142, 0.00000284, 0.00000779, 0.00000524, -0.00000440, -0.00001032, -0.00000670, 0.00000071, 0.00000436, 0.00000247, -0.00000145, -0.00000166, 0.00000260, 0.00000373, -0.00000206, -0.00000616, -0.00000231, 0.00000230, 0.00000213, 0.00000184, 0.00000169, -0.00000256, -0.00000365, 0.00000515, 0.00000941, -0.00000203, -0.00001103, -0.00000402, 0.00000052, -0.00000642, -0.00000320, 0.00000983, 0.00000090, -0.00002251, -0.00001422, 0.00001841, 0.00001839, -0.00001183, -0.00001167, 0.00002205, 0.00003161, 0.00000565, -0.00001449, -0.00001560, -0.00000829, 0.00001374, 0.00003459, 0.00001037, -0.00003725, -0.00002748, 0.00003111, 0.00003585, -0.00002450, -0.00004219, 0.00001090, 0.00003406, -0.00001058, -0.00003500, 0.00000310, 0.00003233, 0.00000991, -0.00001410, -0.00000498, 0.00000598, -0.00000226, -0.00000917, -0.00000420, -0.00000060, -0.00000243, -0.00000226, 0.00000133, 0.00000622, 0.00000902, 0.00000290, -0.00000783, -0.00000621, 0.00000632, 0.00000669, -0.00000735, -0.00001084, 0.00000183, 0.00000845, 0.00000237, -0.00000052, 0.00000277, 0.00000146, -0.00000215, -0.00000030, 0.00000030, -0.00000588, -0.00000815, -0.00000060, 0.00000538, 0.00000303, -0.00000010, -0.00000029, -0.00000186, -0.00000345, -0.00000066, 0.00000259, 0.00000115, -0.00000168, -0.00000201, -0.00000171, -0.00000146, 0.00000072, 0.00000295, 0.00000214, 0.00000019, -0.00000042, -0.00000117, -0.00000227, -0.00000095, 0.00000223, 0.00000356, 0.00000239, 0.00000040, -0.00000242, -0.00000535, -0.00000507, -0.00000066, 0.00000404, },
		{ -0.00000012, 0.00000369, 0.00000458, 0.00000156, -0.00000284, -0.00000432, -0.00000276, -0.00000187, -0.00000115, 0.00000313, 0.00000754, 0.00000411, -0.00000538, -0.00001008, -0.00000567, 0.00000154, 0.00000455, 0.00000189, -0.00000218, -0.00000124, 0.00000386, 0.00000376, -0.00000338, -0.00000638, -0.00000103, 0.00000262, 0.00000089, 0.00000112, 0.00000223, -0.00000176, -0.00000287, 0.00000554, 0.00000852, -0.00000305, -0.00001043, -0.00000359, -0.00000101, -0.00000683, -0.00000079, 0.00001020, -0.00000222, -0.00002179, -0.00000812, 0.00001984, 0.00001252, -0.00001339, -0.00000593, 0.00002221, 0.00002325, 0.00000324, -0.00000730, -0.00001364, -0.00001582, 0.00001065, 0.00004008, 0.00001232, -0.00004166, -0.00002837, 0.00003400, 0.00003432, -0.00002716, -0.00003770, 0.00001654, 0.00003125, -0.00001535, -0.00003163, 0.00000869, 0.00003016, 0.00000441, -0.00001401, -0.00000165, 0.00000621, -0.00000404, -0.00000952, -0.00000395, -0.00000088, -0.00000209, -0.00000134, 0.00000155, 0.00000596, 0.00000873, 0.00000215, -0.00000822, -0.00000515, 0.00000715, 0.00000544, -0.00000865, -0.00000978, 0.00000356, 0.00000831, 0.00000120, -0.00000100, 0.00000288, 0.00000178, -0.00000176, -0.00000078, -0.00000108, -0.00000610, -0.00000647, 0.00000067, 0.00000495, 0.00000243, -0.00000017, -0.00000090, -0.00000264, -0.00000295, 0.00000043, 0.00000260, 0.00000064, -0.00000140, -0.00000164, -0.00000228, -0.00000211, 0.00000100, 0.00000355, 0.00000220, -0.00000011, -0.00000062, -0.00000116, -0.00000217, -0.00000098, 0.00000213, 0.00000380, 0.00000275, 0.00000004, -0.00000325, -0.00000530, -0.00000408, },
		{ -0.00000504, -0.00000312, 0.00000022, 0.00000330, 0.00000429, 0.00000146, -0.00000306, -0.00000438, -0.00000248, -0.00000154, -0.00000079, 0.00000342, 0.00000705, 0.00000280, -0.00000615, -0.00000955, -0.00000468, 0.00000217, 0.00000463, 0.00000130, -0.00000278, -0.00000060, 0.00000508, 0.00000342, -0.00000482, -0.00000631, 0.00000041, 0.00000275, -0.00000047, 0.00000050, 0.00000285, -0.00000090, -0.00000207, 0.00000570, 0.00000740, -0.00000380, -0.00000955, -0.00000345, -0.00000268, -0.00000668, 0.00000166, 0.00000962, -0.00000536, -0.00001962, -0.00000173, 0.00001949, 0.00000591, -0.00001345, 0.00000041, 0.00002081, 0.00001427, 0.00000229, 0.00000050, -0.00001272, -0.00002338, 0.00000866, 0.00004526, 0.00001327, -0.00004524, -0.00002800, 0.00003585, 0.00003138, -0.00002865, -0.00003167, 0.00002101, 0.00002684, -0.00001905, -0.00002664, 0.00001342, 0.00002637, -0.00000073, -0.00001281, 0.00000145, 0.00000570, -0.00000563, -0.00000938, -0.00000380, -0.00000137, -0.00000162, -0.00000032, 0.00000167, 0.00000555, 0.00000832, 0.00000145, -0.00000832, -0.00000395, 0.00000764, 0.00000394, -0.00000965, -0.00000838, 0.00000509, 0.00000785, 0.00000005, -0.00000130, 0.00000307, 0.00000210, -0.00000147, -0.00000146, -0.00000244, -0.00000601, -0.00000466, 0.00000176, 0.00000440, 0.00000185, -0.00000029, -0.00000152, -0.00000324, -0.00000234, 0.00000136, 0.00000251, 0.00000028, -0.00000108, -0.00000149, -0.00000291, -0.00000256, 0.00000137, 0.00000403, 0.00000220, -0.00000037, -0.00000081, -0.00000116, -0.00000204, -0.00000100, 0.00000204, 0.00000405, 0.00000302, -0.00000042, -0.00000396, },
		{ -0.00000094, -0.00000450, -0.00000460, -0.00000225, 0.00000038, 0.00000293, 0.00000406, 0.00000128, -0.00000331, -0.00000435, -0.00000217, -0.00000120, -0.00000033, 0.00000370, 0.00000629, 0.00000138, -0.00000665, -0.00000878, -0.00000375, 0.00000262, 0.00000461, 0.00000071, -0.00000320, 0.00000026, 0.00000617, 0.00000272, -0.00000628, -0.00000590, 0.00000193, 0.00000265, -0.00000188, 0.00000003, 0.00000354, -0.00000000, -0.00000127, 0.00000560, 0.00000613, -0.00000426, -0.00000850, -0.00000365, -0.00000434, -0.00000596, 0.00000393, 0.00000810, -0.00000820, -0.00001602, 0.00000448, 0.00001731, -0.00000094, -0.00001189, 0.00000689, 0.00001775, 0.00000515, 0.00000286, 0.00000842, -0.00001289, -0.00003055, 0.00000775, 0.00004976, 0.00001328, -0.00004768, -0.00002656, 0.00003635, 0.00002732, -0.00002864, -0.00002445, 0.00002397, 0.00002120, -0.00002131, -0.00002036, 0.00001691, 0.00002123, -0.00000516, -0.00001062, 0.00000408, 0.00000454, -0.00000689, -0.00000884, -0.00000383, -0.00000199, -0.00000097, 0.00000078, 0.00000165, 0.00000500, 0.00000780, 0.00000081, -0.00000812, -0.00000268, 0.00000778, 0.00000227, -0.00001028, -0.00000674, 0.00000634, 0.00000710, -0.00000102, -0.00000142, 0.00000331, 0.00000241, -0.00000131, -0.00000232, -0.00000370, -0.00000562, -0.00000280, 0.00000263, 0.00000375, 0.00000131, -0.00000045, -0.00000214, -0.00000365, -0.00000167, 0.00000210, 0.00000238, 0.00000009, -0.00000080, -0.00000155, -0.00000354, -0.00000281, 0.00000178, 0.00000440, 0.00000217, -0.00000058, -0.00000098, -0.00000115, -0.00000191, -0.00000102, 0.00000198, 0.00000429, 0.00000318, },
		{ 0.00000451, 0.00000321, -0.00000148, -0.00000483, -0.00000405, -0.00000154, 0.00000040, 0.00000261, 0.00000387, 0.00000102, -0.00000356, -0.00000423, -0.00000184, -0.00000085, 0.00000023, 0.00000390, 0.00000527, -0.00000008, -0.00000687, -0.00000782, -0.00000295, 0.00000290, 0.00000451, 0.00000014, -0.00000342, 0.00000131, 0.00000705, 0.00000166, -0.00000765, -0.00000515, 0.00000342, 0.00000229, -0.00000329, -0.00000028, 0.00000427, 0.00000092, -0.00000051, 0.00000526, 0.00000477, -0.00000442, -0.00000739, -0.00000416, -0.00000583, -0.00000472, 0.00000578, 0.00000575, -0.00001041, -0.00001114, 0.00001004, 0.00001341, -0.00000748, -0.00000871, 0.00001299, 0.00001307, -0.00000362, 0.00000491, 0.00001601, -0.00001407, -0.00003692, 0.00000780, 0.00005325, 0.00001254, -0.00004870, -0.00002431, 0.00003533, 0.00002254, -0.00002695, -0.00001648, 0.00002519, 0.00001478, -0.00002189, -0.00001325, 0.00001886, 0.00001507, -0.00000861, -0.00000769, 0.00000605, 0.00000290, -0.00000770, -0.00000800, -0.00000409, -0.00000268, -0.00000013, 0.00000189, 0.00000150, 0.00000433, 0.00000719, 0.00000028, -0.00000762, -0.00000142, 0.00000754, 0.00000054, -0.00001051, -0.00000495, 0.00000725, 0.00000611, -0.00000193, -0.00000137, 0.00000359, 0.00000269, -0.00000130, -0.00000332, -0.00000481, -0.00000493, -0.00000094, 0.00000328, 0.00000303, 0.00000081, -0.00000066, -0.00000272, -0.00000389, -0.00000100, 0.00000268, 0.00000227, 0.00000003, -0.00000061, -0.00000181, -0.00000411, -0.00000286, 0.00000221, 0.00000465, 0.00000210, -0.00000075, -0.00000115, -0.00000114, -0.00000176, -0.00000103, 0.00000194, },
		{ -0.00000101, 0.00000194, 0.00000468, 0.00000312, -0.00000199, -0.00000495, -0.00000342, -0.00000101, 0.00000032, 0.00000237, 0.00000367, 0.00000066, -0.00000377, -0.00000400, -0.00000152, -0.00000048, 0.00000086, 0.00000400, 0.00000400, -0.00000150, -0.00000678, -0.00000673, -0.00000228, 0.00000303, 0.00000434, -0.00000038, -0.00000341, 0.00000248, 0.00000765, 0.00000029, -0.00000884, -0.00000408, 0.00000481, 0.00000167, -0.00000463, -0.00000041, 0.00000502, 0.00000183, 0.00000017, 0.00000468, 0.00000340, -0.00000429, -0.00000630, -0.00000495, -0.00000700, -0.00000305, 0.00000704, 0.00000273, -0.00001172, -0.00000521, 0.00001454, 0.00000800, -0.00001322, -0.00000411, 0.00001823, 0.00000696, -0.00001156, 0.00000823, 0.00002281, -0.00001611, -0.00004214, 0.00000860, 0.00005545, 0.00001130, -0.00004814, -0.00002157, 0.00003269, 0.00001746, -0.00002347, -0.00000826, 0.00002459, 0.00000809, -0.00002068, -0.00000580, 0.00001912, 0.00000832, -0.00001089, -0.00000428, 0.00000722, 0.00000095, -0.00000800, -0.00000699, -0.00000460, -0.00000335, 0.00000090, 0.00000294, 0.00000119, 0.00000357, 0.00000650, -0.00000012, -0.00000684, -0.00000023, 0.00000694, -0.00000116, -0.00001033, -0.00000311, 0.00000777, 0.00000496, -0.00000266, -0.00000117, 0.00000386, 0.00000289, -0.00000145, -0.00000442, -0.00000573, -0.00000399, 0.00000085, 0.00000371, 0.00000229, 0.00000034, -0.00000092, -0.00000323, -0.00000397, -0.00000037, 0.00000312, 0.00000221, 0.00000008, -0.00000056, -0.00000225, -0.00000457, -0.00000272, 0.00000259, 0.00000476, 0.00000204, -0.00000088, -0.00000131, -0.00000114, -0.00000162, },
		},
		{ { -0.00000025, 0.00000007, -0.00000003, -0.00000134, -0.00000164, 0.00000119, 0.00000457, 0.00000347, -0.00000241, -0.00000717, -0.00000623, -0.00000214, -0.00000034, -0.00000179, -0.00000299, -0.00000160, 0.00000134, 0.00000389, 0.00000443, 0.00000159, -0.00000326, -0.00000473, 0.00000065, 0.00000776, 0.00000782, 0.00000057, -0.00000476, -0.00000265, 0.00000104, -0.00000106, -0.00000614, -0.00000511, 0.00000437, 0.00001415, 0.00001385, 0.00000122, -0.00001363, -0.00001459, 0.00000306, 0.00002098, 0.00001456, -0.00001347, -0.00002979, -0.00001254, 0.00001643, 0.00001995, -0.00000321, -0.00001989, -0.00001254, 0.00000053, 0.00000086, -0.00000275, 0.00000243, 0.00000740, 0.00000050, -0.00000749, -0.00000254, 0.00000382, -0.00000580, -0.00001791, -0.00000575, 0.00002071, 0.00002156, -0.00000996, -0.00003037, -0.00000996, 0.00002156, 0.00002071, -0.00000575, -0.00001791, -0.00000580, 0.00000382, -0.00000254, -0.00000749, 0.00000050, 0.00000740, 0.00000243, -0.00000275, 0.00000086, 0.00000053, -0.00001254, -0.00001989, -0.00000321, 0.00001995, 0.00001643, -0.00001254, -0.00002979, -0.00001347, 0.00001456, 0.00002098, 0.00000306, -0.00001459, -0.00001363, 0.00000122, 0.00001385, 0.00001415, 0.00000437, -0.00000511, -0.00000614, -0.00000106, 0.00000104, -0.00000265, -0.00000476, 0.00000057, 0.00000782, 0.00000776, 0.00000065, -0.00000473, -0.00000326, 0.00000159, 0.00000443, 0.00000389, 0.00000134, -0.00000160, -0.00000299, -0.00000179, -0.00000034, -0.00000214, -0.00000623, -0.00000717, -0.00000241, 0.00000347, 0.00000457, 0.00000119, -0.00000164, -0.00000134, -0.00000003, 0.00000007, },
		{ 0.00000013, 0.00000036, -0.00000021, -0.00000015, -0.00000017, -0.00000143, -0.00000209, 0.00000066, 0.00000500, 0.00000521, -0.00000060, -0.00000689, -0.00000761, -0.00000371, -0.00000077, -0.00000113, -0.00000231, -0.00000188, 0.00000015, 0.00000284, 0.00000466, 0.00000299, -0.00000226, -0.00000554, -0.00000139, 0.00000664, 0.00000883, 0.00000252, -0.00000399, -0.00000342, 0.00000067, 0.00000048, -0.00000416, -0.00000603, -0.00000019, 0.00001016, 0.00001554, 0.00000776, -0.00000966, -0.00001876, -0.00000506, 0.00001907, 0.00002245, -0.00000419, -0.00002977, -0.00002143, 0.00000957, 0.00002297, 0.00000526, -0.00001601, -0.00001624, -0.00000450, 0.00000052, 0.00000011, 0.00000339, 0.00000564, 0.00000006, -0.00000428, 0.00000125, 0.00000393, -0.00000892, -0.00001984, -0.00000410, 0.00002330, 0.00002136, -0.00001253, -0.00003153, -0.00000805, 0.00002362, 0.00001972, -0.00000844, -0.00001815, -0.00000282, 0.00000587, -0.00000482, -0.00001173, -0.00000085, 0.00000929, 0.00000311, -0.00000528, -0.00000069, 0.00000409, -0.00000728, -0.00002064, -0.00001100, 0.00001389, 0.00001999, -0.00000310, -0.00002589, -0.00002004, 0.00000513, 0.00001915, 0.00000962, -0.00000836, -0.00001478, -0.00000503, 0.00000996, 0.00001624, 0.00000925, -0.00000256, -0.00000721, -0.00000308, 0.00000067, -0.00000183, -0.00000490, -0.00000116, 0.00000626, 0.00000818, 0.00000259, -0.00000337, -0.00000370, 0.00000021, 0.00000374, 0.00000453, 0.00000257, -0.00000092, -0.00000336, -0.00000261, -0.00000041, -0.00000091, -0.00000464, -0.00000684, -0.00000376, 0.00000174, 0.00000383, 0.00000141, -0.00000132, -0.00000126, },
		{ -0.00000118, -0.00000125, 0.00000029, 0.00000069, -0.00000009, -0.00000031, -0.00000025, -0.00000147, -0.00000258, -0.00000015, 0.00000501, 0.00000679, 0.00000158, -0.00000595, -0.00000856, -0.00000544, -0.00000171, -0.00000079, -0.00000148, -0.00000172, -0.00000084, 0.00000151, 0.00000436, 0.00000417, -0.00000087, -0.00000570, -0.00000322, 0.00000497, 0.00000911, 0.00000438, -0.00000261, -0.00000378, -0.00000025, 0.00000120, -0.00000187, -0.00000534, -0.00000364, 0.00000500, 0.00001469, 0.00001340, -0.00000347, -0.00002003, -0.00001332, 0.00001365, 0.00002742, 0.00000634, -0.00002564, -0.00002811, 0.00000055, 0.00002222, 0.00001284, -0.00000948, -0.00001735, -0.00000997, -0.00000202, 0.00000231, 0.00000585, 0.00000488, -0.00000167, -0.00000282, 0.00000557, 0.00000641, -0.00001104, -0.00002348, -0.00000438, 0.00002652, 0.00002339, -0.00001458, -0.00003469, -0.00000782, 0.00002669, 0.00002088, -0.00001102, -0.00002049, -0.00000106, 0.00000937, -0.00000497, -0.00001602, -0.00000406, 0.00001037, 0.00000509, -0.00000671, -0.00000344, 0.00000563, -0.00000169, -0.00001846, -0.00001681, 0.00000600, 0.00001981, 0.00000521, -0.00001900, -0.00002295, -0.00000418, 0.00001401, 0.00001348, -0.00000132, -0.00001303, -0.00000991, 0.00000459, 0.00001605, 0.00001356, 0.00000130, -0.00000696, -0.00000506, -0.00000046, -0.00000131, -0.00000457, -0.00000246, 0.00000441, 0.00000788, 0.00000418, -0.00000166, -0.00000351, -0.00000093, 0.00000274, 0.00000469, 0.00000366, 0.00000005, -0.00000335, -0.00000342, -0.00000089, -0.00000008, -0.00000303, -0.00000604, -0.00000459, 0.00000017, 0.00000292, 0.00000138, },
		{ 0.00000196, 0.00000115, -0.00000122, -0.00000137, 0.00000038, 0.00000105, 0.00000010, -0.00000042, -0.00000029, -0.00000141, -0.00000302, -0.00000117, 0.00000455, 0.00000800, 0.00000393, -0.00000434, -0.00000890, -0.00000713, -0.00000313, -0.00000091, -0.00000064, -0.00000116, -0.00000147, 0.00000007, 0.00000356, 0.00000497, 0.00000069, -0.00000521, -0.00000460, 0.00000301, 0.00000859, 0.00000583, -0.00000077, -0.00000347, -0.00000134, 0.00000097, 0.00000012, -0.00000335, -0.00000543, -0.00000039, 0.00001143, 0.00001714, 0.00000391, -0.00001806, -0.00002026, 0.00000562, 0.00002858, 0.00001638, -0.00001794, -0.00003127, -0.00000904, 0.00001772, 0.00001806, -0.00000136, -0.00001530, -0.00001462, -0.00000651, 0.00000295, 0.00000921, 0.00000575, -0.00000381, -0.00000340, 0.00000930, 0.00001091, -0.00001117, -0.00002788, -0.00000702, 0.00002919, 0.00002744, -0.00001504, -0.00003907, -0.00000990, 0.00002964, 0.00002416, -0.00001240, -0.00002430, -0.00000126, 0.00001331, -0.00000285, -0.00001933, -0.00000869, 0.00000996, 0.00000769, -0.00000660, -0.00000647, 0.00000509, 0.00000309, -0.00001405, -0.00001980, -0.00000222, 0.00001617, 0.00001100, -0.00001054, -0.00002199, -0.00001182, 0.00000661, 0.00001405, 0.00000519, -0.00000888, -0.00001259, -0.00000125, 0.00001360, 0.00001651, 0.00000586, -0.00000521, -0.00000648, -0.00000213, -0.00000133, -0.00000402, -0.00000323, 0.00000256, 0.00000694, 0.00000520, 0.00000015, -0.00000274, -0.00000165, 0.00000161, 0.00000437, 0.00000445, 0.00000119, -0.00000294, -0.00000407, -0.00000164, 0.00000029, -0.00000156, -0.00000492, -0.00000494, -0.00000112, },
		{ -0.00000488, -0.00000211, 0.00000108, 0.00000085, -0.00000139, -0.00000167, 0.00000034, 0.00000139, 0.00000036, -0.00000047, -0.00000028, -0.00000123, -0.00000331, -0.00000228, 0.00000362, 0.00000869, 0.00000623, -0.00000214, -0.00000849, -0.00000853, -0.00000490, -0.00000159, -0.00000001, -0.00000030, -0.00000166, -0.00000128, 0.00000237, 0.00000525, 0.00000218, -0.00000421, -0.00000535, 0.00000108, 0.00000736, 0.00000659, 0.00000122, -0.00000238, -0.00000217, -0.00000009, 0.00000129, -0.00000064, -0.00000535, -0.00000508, 0.00000632, 0.00001831, 0.00001125, -0.00001314, -0.00002466, -0.00000368, 0.00002574, 0.00002425, -0.00000790, -0.00003023, -0.00001742, 0.00001022, 0.00001983, 0.00000686, -0.00001015, -0.00001723, -0.00001218, 0.00000139, 0.00001254, 0.00000844, -0.00000535, -0.00000574, 0.00001149, 0.00001659, -0.00000876, -0.00003186, -0.00001190, 0.00003020, 0.00003276, -0.00001322, -0.00004352, -0.00001434, 0.00003131, 0.00002900, -0.00001178, -0.00002859, -0.00000368, 0.00001657, 0.00000113, -0.00002083, -0.00001394, 0.00000772, 0.00001004, -0.00000500, -0.00000890, 0.00000290, 0.00000622, -0.00000852, -0.00001977, -0.00000932, 0.00000995, 0.00001345, -0.00000213, -0.00001763, -0.00001659, -0.00000166, 0.00001134, 0.00000993, -0.00000323, -0.00001267, -0.00000649, 0.00000936, 0.00001754, 0.00001036, -0.00000209, -0.00000686, -0.00000397, -0.00000198, -0.00000356, -0.00000353, 0.00000093, 0.00000556, 0.00000556, 0.00000180, -0.00000150, -0.00000181, 0.00000056, 0.00000367, 0.00000483, 0.00000233, -0.00000218, -0.00000444, -0.00000254, 0.00000024, -0.00000033, -0.00000362, },
		{ 0.00000058, -0.00000227, -0.00000450, -0.00000280, 0.00000033, 0.00000058, -0.00000162, -0.00000211, 0.00000011, 0.00000165, 0.00000067, -0.00000047, -0.00000026, -0.00000095, -0.00000340, -0.00000335, 0.00000231, 0.00000875, 0.00000822, 0.00000045, -0.00000727, -0.00000940, -0.00000683, -0.00000285, 0.00000025, 0.00000072, -0.00000140, -0.00000238, 0.00000095, 0.00000501, 0.00000336, -0.00000290, -0.00000540, -0.00000050, 0.00000565, 0.00000648, 0.00000299, -0.00000057, -0.00000236, -0.00000165, 0.00000136, 0.00000210, -0.00000355, -0.00000829, 0.00000023, 0.00001673, 0.00001734, -0.00000606, -0.00002580, -0.00001268, 0.00001942, 0.00002868, 0.00000277, -0.00002511, -0.00002297, 0.00000112, 0.00001769, 0.00001357, -0.00000263, -0.00001687, -0.00001787, -0.00000250, 0.00001479, 0.00001263, -0.00000541, -0.00000914, 0.00001151, 0.00002234, -0.00000381, -0.00003424, -0.00001832, 0.00002877, 0.00003823, -0.00000896, -0.00004685, -0.00002063, 0.00003083, 0.00003440, -0.00000883, -0.00003224, -0.00000802, 0.00001822, 0.00000617, -0.00002012, -0.00001884, 0.00000378, 0.00001136, -0.00000233, -0.00001011, -0.00000020, 0.00000732, -0.00000301, -0.00001707, -0.00001422, 0.00000247, 0.00001238, 0.00000472, -0.00001095, -0.00001788, -0.00000928, 0.00000598, 0.00001204, 0.00000273, -0.00001028, -0.00001021, 0.00000409, 0.00001645, 0.00001400, 0.00000199, -0.00000591, -0.00000551, -0.00000318, -0.00000344, -0.00000353, -0.00000032, 0.00000397, 0.00000528, 0.00000305, -0.00000002, -0.00000139, -0.00000018, 0.00000271, 0.00000475, 0.00000330, -0.00000114, -0.00000446, -0.00000346, -0.00000016, },
		{ -0.00000426, -0.00000084, 0.00000117, -0.00000097, -0.00000389, -0.00000324, -0.00000026, 0.00000040, -0.00000180, -0.00000264, -0.00000031, 0.00000176, 0.00000098, -0.00000042, -0.00000023, -0.00000061, -0.00000327, -0.00000423, 0.00000078, 0.00000816, 0.00000968, 0.00000319, -0.00000526, -0.00000953, -0.00000867, -0.00000462, -0.00000001, 0.00000171, -0.00000074, -0.00000310, -0.00000052, 0.00000431, 0.00000409, -0.00000155, -0.00000484, -0.00000148, 0.00000378, 0.00000549, 0.00000415, 0.00000170, -0.00000164, -0.00000325, 0.00000028, 0.00000421, -0.00000050, -0.00000955, -0.00000581, 0.00001269, 0.00002122, 0.00000198, -0.00002355, -0.00001991, 0.00001077, 0.00002902, 0.00001226, -0.00001681, -0.00002461, -0.00000782, 0.00001197, 0.00001736, 0.00000597, -0.00001317, -0.00002222, -0.00000827, 0.00001506, 0.00001754, -0.00000347, -0.00001258, 0.00000929, 0.00002701, 0.00000311, -0.00003413, -0.00002518, 0.00002465, 0.00004262, -0.00000269, -0.00004810, -0.00002782, 0.00002777, 0.00003921, -0.00000379, -0.00003427, -0.00001353, 0.00001772, 0.00001123, -0.00001725, -0.00002248, -0.00000132, 0.00001110, 0.00000069, -0.00000984, -0.00000335, 0.00000649, 0.00000146, -0.00001254, -0.00001633, -0.00000486, 0.00000831, 0.00000890, -0.00000339, -0.00001576, -0.00001492, -0.00000093, 0.00001120, 0.00000780, -0.00000601, -0.00001179, -0.00000126, 0.00001341, 0.00001613, 0.00000640, -0.00000363, -0.00000628, -0.00000465, -0.00000380, -0.00000347, -0.00000118, 0.00000241, 0.00000447, 0.00000376, 0.00000147, -0.00000045, -0.00000047, 0.00000170, 0.00000423, 0.00000397, 0.00000004, -0.00000413, },
		{ 0.00000122, -0.00000346, -0.00000486, -0.00000169, 0.00000142, 0.00000022, -0.00000314, -0.00000349, -0.00000073, 0.00000035, -0.00000184, -0.00000317, -0.00000092, 0.00000168, 0.00000123, -0.00000033, -0.00000020, -0.00000024, -0.00000293, -0.00000484, -0.00000081, 0.00000699, 0.00001043, 0.00000578, -0.00000261, -0.00000876, -0.00001013, -0.00000676, -0.00000088, 0.00000247, 0.00000021, -0.00000340, -0.00000185, 0.00000330, 0.00000431, -0.00000043, -0.00000387, -0.00000170, 0.00000213, 0.00000378, 0.00000441, 0.00000404, 0.00000007, -0.00000437, -0.00000172, 0.00000521, 0.00000311, -0.00000876, -0.00001086, 0.00000688, 0.00002232, 0.00000970, -0.00001840, -0.00002425, 0.00000133, 0.00002538, 0.00001897, -0.00000684, -0.00002198, -0.00001481, 0.00000375, 0.00001736, 0.00001403, -0.00000643, -0.00002404, -0.00001497, 0.00001281, 0.00002211, 0.00000052, -0.00001499, 0.00000528, 0.00002967, 0.00001102, -0.00003114, -0.00003122, 0.00001817, 0.00004489, 0.00000468, -0.00004673, -0.00003471, 0.00002229, 0.00004236, 0.00000261, -0.00003406, -0.00001917, 0.00001503, 0.00001530, -0.00001277, -0.00002422, -0.00000676, 0.00000910, 0.00000328, -0.00000824, -0.00000580, 0.00000424, 0.00000423, -0.00000724, -0.00001562, -0.00001078, 0.00000225, 0.00000987, 0.00000358, -0.00001091, -0.00001772, -0.00000805, 0.00000764, 0.00001101, -0.00000077, -0.00001106, -0.00000575, 0.00000896, 0.00001637, 0.00001040, -0.00000026, -0.00000595, -0.00000601, -0.00000464, -0.00000357, -0.00000173, 0.00000105, 0.00000331, 0.00000385, 0.00000272, 0.00000086, -0.00000021, 0.00000085, 0.00000337, 0.00000425, },
		{ 0.00000230, 0.00000407, 0.00000227, -0.00000253, -0.00000519, -0.00000264, 0.00000138, 0.00000126, -0.00000229, -0.00000360, -0.00000113, 0.00000041, -0.00000167, -0.00000357, -0.00000165, 0.00000135, 0.00000137, -0.00000023, -0.00000018, 0.00000010, -0.00000244, -0.00000510, -0.00000225, 0.00000538, 0.00001035, 0.00000791, 0.00000045, -0.00000704, -0.00001096, -0.00000904, -0.00000234, 0.00000284, 0.00000127, -0.00000328, -0.00000290, 0.00000218, 0.00000404, 0.00000025, -0.00000275, -0.00000119, 0.00000106, 0.00000166, 0.00000361, 0.00000600, 0.00000264, -0.00000459, -0.00000420, 0.00000484, 0.00000655, -0.00000619, -0.00001419, 0.00000029, 0.00002063, 0.00001590, -0.00001133, -0.00002518, -0.00000729, 0.00001860, 0.00002185, 0.00000295, -0.00001558, -0.00001839, -0.00000536, 0.00001341, 0.00001994, 0.00000237, -0.00002253, -0.00002129, 0.00000811, 0.00002525, 0.00000606, -0.00001552, 0.00000035, 0.00002986, 0.00001870, -0.00002546, -0.00003530, 0.00001020, 0.00004439, 0.00001194, -0.00004279, -0.00004016, 0.00001504, 0.00004313, 0.00000933, -0.00003148, -0.00002389, 0.00001059, 0.00001762, -0.00000753, -0.00002383, -0.00001163, 0.00000563, 0.00000479, -0.00000579, -0.00000706, 0.00000129, 0.00000508, -0.00000224, -0.00001263, -0.00001440, -0.00000452, 0.00000774, 0.00000868, -0.00000447, -0.00001737, -0.00001408, 0.00000212, 0.00001175, 0.00000434, -0.00000827, -0.00000864, 0.00000389, 0.00001469, 0.00001332, 0.00000372, -0.00000439, -0.00000683, -0.00000578, -0.00000400, -0.00000214, -0.00000002, 0.00000202, 0.00000338, 0.00000354, 0.00000232, 0.00000061, 0.00000034, },
		{ 0.00000191, 0.00000032, 0.00000120, 0.00000348, 0.00000305, -0.00000141, -0.00000522, -0.00000359, 0.00000106, 0.00000213, -0.00000136, -0.00000359, -0.00000153, 0.00000051, -0.00000130, -0.00000373, -0.00000237, 0.00000081, 0.00000131, -0.00000016, -0.00000015, 0.00000039, -0.00000187, -0.00000503, -0.00000337, 0.00000356, 0.00000948, 0.00000931, 0.00000361, -0.00000444, -0.00001092, -0.00001119, -0.00000432, 0.00000270, 0.00000228, -0.00000285, -0.00000359, 0.00000115, 0.00000344, 0.00000038, -0.00000183, -0.00000009, 0.00000081, -0.00000047, 0.00000179, 0.00000714, 0.00000571, -0.00000365, -0.00000662, 0.00000315, 0.00000913, -0.00000242, -0.00001542, -0.00000605, 0.00001657, 0.00001973, -0.00000362, -0.00002282, -0.00001369, 0.00001006, 0.00002056, 0.00001076, -0.00000664, -0.00001776, -0.00001349, 0.00000618, 0.00002244, 0.00001174, -0.00001753, -0.00002584, 0.00000156, 0.00002608, 0.00001221, -0.00001376, -0.00000437, 0.00002764, 0.00002500, -0.00001781, -0.00003665, 0.00000192, 0.00004104, 0.00001789, -0.00003686, -0.00004333, 0.00000707, 0.00004127, 0.00001526, -0.00002693, -0.00002686, 0.00000523, 0.00001784, -0.00000250, -0.00002154, -0.00001518, 0.00000129, 0.00000484, -0.00000317, -0.00000700, -0.00000158, 0.00000422, 0.00000162, -0.00000820, -0.00001539, -0.00001069, 0.00000319, 0.00001109, 0.00000219, -0.00001418, -0.00001798, -0.00000427, 0.00000996, 0.00000835, -0.00000409, -0.00000955, -0.00000094, 0.00001139, 0.00001468, 0.00000767, -0.00000171, -0.00000676, -0.00000690, -0.00000482, -0.00000261, -0.00000082, 0.00000079, 0.00000248, 0.00000381, 0.00000371, },
		{ 0.00000349, 0.00000479, 0.00000354, 0.00000087, 0.00000027, 0.00000253, 0.00000347, -0.00000022, -0.00000494, -0.00000446, 0.00000050, 0.00000280, -0.00000037, -0.00000346, -0.00000196, 0.00000054, -0.00000076, -0.00000357, -0.00000295, 0.00000010, 0.00000102, -0.00000018, -0.00000013, 0.00000062, -0.00000132, -0.00000469, -0.00000406, 0.00000178, 0.00000792, 0.00000979, 0.00000650, -0.00000112, -0.00000988, -0.00001294, -0.00000667, 0.00000200, 0.00000306, -0.00000225, -0.00000393, 0.00000038, 0.00000273, -0.00000004, -0.00000137, 0.00000132, 0.00000151, -0.00000216, -0.00000083, 0.00000715, 0.00000883, -0.00000152, -0.00000843, 0.00000045, 0.00001041, 0.00000181, -0.00001457, -0.00001123, 0.00001098, 0.00002080, 0.00000343, -0.00001788, -0.00001697, 0.00000143, 0.00001559, 0.00001515, 0.00000307, -0.00001297, -0.00001892, -0.00000300, 0.00002088, 0.00002000, -0.00000960, -0.00002749, -0.00000571, 0.00002422, 0.00001784, -0.00000981, -0.00000782, 0.00002365, 0.00002911, -0.00000929, -0.00003504, -0.00000543, 0.00003533, 0.00002160, -0.00002994, -0.00004387, -0.00000045, 0.00003711, 0.00001947, -0.00002121, -0.00002765, -0.00000006, 0.00001608, 0.00000144, -0.00001795, -0.00001697, -0.00000315, 0.00000344, -0.00000102, -0.00000587, -0.00000376, 0.00000218, 0.00000383, -0.00000336, -0.00001390, -0.00001524, -0.00000274, 0.00001057, 0.00000778, -0.00000898, -0.00001920, -0.00001033, 0.00000606, 0.00001050, 0.00000059, -0.00000845, -0.00000475, 0.00000708, 0.00001427, 0.00001093, 0.00000175, -0.00000558, -0.00000765, -0.00000592, -0.00000331, -0.00000146, -0.00000026, 0.00000130, },
		{ -0.00000107, 0.00000006, 0.00000263, 0.00000539, 0.00000529, 0.00000201, -0.00000030, 0.00000135, 0.00000348, 0.00000092, -0.00000437, -0.00000519, -0.00000025, 0.00000323, 0.00000065, -0.00000316, -0.00000241, 0.00000042, -0.00000017, -0.00000307, -0.00000323, -0.00000065, 0.00000046, -0.00000036, -0.00000012, 0.00000081, -0.00000085, -0.00000418, -0.00000428, 0.00000029, 0.00000591, 0.00000924, 0.00000877, 0.00000263, -0.00000777, -0.00001402, -0.00000916, 0.00000076, 0.00000350, -0.00000163, -0.00000401, -0.00000005, 0.00000214, -0.00000088, -0.00000160, 0.00000265, 0.00000308, -0.00000303, -0.00000386, 0.00000588, 0.00001147, 0.00000162, -0.00000918, -0.00000276, 0.00001019, 0.00000574, -0.00001202, -0.00001464, 0.00000484, 0.00001928, 0.00000874, -0.00001150, -0.00001688, -0.00000571, 0.00000808, 0.00001538, 0.00001163, -0.00000487, -0.00002043, -0.00001231, 0.00001543, 0.00002559, 0.00000007, -0.00002561, -0.00001232, 0.00001985, 0.00002186, -0.00000431, -0.00000922, 0.00001886, 0.00003071, -0.00000110, -0.00003080, -0.00001080, 0.00002820, 0.00002258, -0.00002322, -0.00004197, -0.00000652, 0.00003140, 0.00002141, -0.00001534, -0.00002631, -0.00000437, 0.00001290, 0.00000374, -0.00001391, -0.00001694, -0.00000692, 0.00000093, 0.00000014, -0.00000413, -0.00000492, -0.00000036, 0.00000434, 0.00000097, -0.00001055, -0.00001752, -0.00000884, 0.00000745, 0.00001131, -0.00000291, -0.00001771, -0.00001502, 0.00000084, 0.00001050, 0.00000487, -0.00000575, -0.00000701, 0.00000249, 0.00001220, 0.00001296, 0.00000543, -0.00000330, -0.00000764, -0.00000707, -0.00000432, -0.00000209, },
		{ -0.00000562, -0.00000289, -0.00000170, -0.00000110, 0.00000136, 0.00000539, 0.00000691, 0.00000364, -0.00000036, 0.00000013, 0.00000307, 0.00000188, -0.00000356, -0.00000570, -0.00000116, 0.00000339, 0.00000167, -0.00000266, -0.00000283, 0.00000008, 0.00000031, -0.00000230, -0.00000309, -0.00000128, -0.00000032, -0.00000078, -0.00000015, 0.00000097, -0.00000051, -0.00000365, -0.00000406, -0.00000069, 0.00000373, 0.00000772, 0.00001011, 0.00000646, -0.00000469, -0.00001420, -0.00001154, -0.00000092, 0.00000353, -0.00000116, -0.00000399, -0.00000013, 0.00000189, -0.00000190, -0.00000257, 0.00000354, 0.00000529, -0.00000280, -0.00000682, 0.00000341, 0.00001318, 0.00000537, -0.00000862, -0.00000589, 0.00000860, 0.00000875, -0.00000841, -0.00001604, -0.00000084, 0.00001574, 0.00001165, -0.00000500, -0.00001383, -0.00001015, -0.00000035, 0.00001153, 0.00001733, 0.00000496, -0.00001756, -0.00001988, 0.00000705, 0.00002742, 0.00000991, -0.00002024, -0.00001690, 0.00001377, 0.00002349, 0.00000173, -0.00000827, 0.00001441, 0.00003001, 0.00000569, -0.00002475, -0.00001359, 0.00002082, 0.00002090, -0.00001779, -0.00003832, -0.00001046, 0.00002516, 0.00002102, -0.00001031, -0.00002334, -0.00000705, 0.00000910, 0.00000423, -0.00001026, -0.00001544, -0.00000947, -0.00000211, 0.00000013, -0.00000237, -0.00000504, -0.00000276, 0.00000341, 0.00000413, -0.00000617, -0.00001740, -0.00001399, 0.00000250, 0.00001228, 0.00000283, -0.00001402, -0.00001761, -0.00000468, 0.00000847, 0.00000796, -0.00000212, -0.00000752, -0.00000162, 0.00000888, 0.00001343, 0.00000875, -0.00000016, -0.00000663, -0.00000794, },
		{ -0.00000454, -0.00000818, -0.00000704, -0.00000396, -0.00000226, -0.00000207, -0.00000015, 0.00000477, 0.00000820, 0.00000561, 0.00000017, -0.00000097, 0.00000228, 0.00000257, -0.00000259, -0.00000594, -0.00000214, 0.00000325, 0.00000261, -0.00000194, -0.00000312, -0.00000048, 0.00000053, -0.00000140, -0.00000249, -0.00000161, -0.00000122, -0.00000148, -0.00000030, 0.00000112, -0.00000030, -0.00000323, -0.00000352, -0.00000103, 0.00000171, 0.00000539, 0.00001030, 0.00000997, -0.00000080, -0.00001335, -0.00001356, -0.00000288, 0.00000316, -0.00000093, -0.00000403, 0.00000004, 0.00000213, -0.00000280, -0.00000422, 0.00000365, 0.00000776, -0.00000138, -0.00000921, -0.00000000, 0.00001361, 0.00000923, -0.00000676, -0.00000839, 0.00000600, 0.00001045, -0.00000447, -0.00001556, -0.00000530, 0.00001107, 0.00001201, 0.00000039, -0.00000875, -0.00001126, -0.00000799, 0.00000446, 0.00001900, 0.00001459, -0.00001076, -0.00002415, -0.00000270, 0.00002514, 0.00001821, -0.00001215, -0.00001845, 0.00000717, 0.00002247, 0.00000715, -0.00000522, 0.00001131, 0.00002775, 0.00001042, -0.00001801, -0.00001371, 0.00001434, 0.00001716, -0.00001440, -0.00003391, -0.00001211, 0.00001943, 0.00001874, -0.00000686, -0.00001956, -0.00000788, 0.00000557, 0.00000319, -0.00000766, -0.00001310, -0.00001053, -0.00000500, -0.00000095, -0.00000109, -0.00000437, -0.00000455, 0.00000154, 0.00000579, -0.00000168, -0.00001518, -0.00001739, -0.00000324, 0.00001070, 0.00000719, -0.00000902, -0.00001784, -0.00000952, 0.00000494, 0.00000940, 0.00000167, -0.00000642, -0.00000464, 0.00000493, 0.00001231, 0.00001112, 0.00000341, },
		{ 0.00001215, 0.00000686, -0.00000149, -0.00000749, -0.00000829, -0.00000532, -0.00000288, -0.00000283, -0.00000171, 0.00000357, 0.00000896, 0.00000773, 0.00000131, -0.00000175, 0.00000122, 0.00000288, -0.00000155, -0.00000587, -0.00000309, 0.00000278, 0.00000336, -0.00000100, -0.00000315, -0.00000118, 0.00000036, -0.00000056, -0.00000145, -0.00000145, -0.00000210, -0.00000249, -0.00000062, 0.00000130, -0.00000020, -0.00000305, -0.00000285, -0.00000070, 0.00000018, 0.00000252, 0.00000925, 0.00001278, 0.00000361, -0.00001144, -0.00001499, -0.00000493, 0.00000248, -0.00000100, -0.00000431, 0.00000028, 0.00000291, -0.00000327, -0.00000633, 0.00000277, 0.00001005, 0.00000112, -0.00001060, -0.00000392, 0.00001263, 0.00001264, -0.00000382, -0.00000984, 0.00000290, 0.00001076, -0.00000089, -0.00001368, -0.00000808, 0.00000626, 0.00001018, 0.00000379, -0.00000292, -0.00000912, -0.00001335, -0.00000437, 0.00001629, 0.00002209, -0.00000125, -0.00002419, -0.00001195, 0.00001918, 0.00002356, -0.00000263, -0.00001652, 0.00000142, 0.00001915, 0.00001093, -0.00000079, 0.00001020, 0.00002495, 0.00001293, -0.00001168, -0.00001165, 0.00000961, 0.00001231, -0.00001335, -0.00002981, -0.00001181, 0.00001505, 0.00001536, -0.00000531, -0.00001587, -0.00000706, 0.00000302, 0.00000121, -0.00000648, -0.00001065, -0.00001023, -0.00000715, -0.00000270, -0.00000061, -0.00000335, -0.00000554, -0.00000067, 0.00000596, 0.00000218, -0.00001149, -0.00001864, -0.00000872, 0.00000705, 0.00000949, -0.00000375, -0.00001594, -0.00001289, 0.00000071, 0.00000910, 0.00000487, -0.00000419, -0.00000623, 0.00000104, 0.00000984, },
		{ -0.00000217, 0.00000649, 0.00001165, 0.00000962, 0.00000218, -0.00000573, -0.00000906, -0.00000688, -0.00000370, -0.00000343, -0.00000319, 0.00000193, 0.00000908, 0.00000975, 0.00000299, -0.00000204, 0.00000005, 0.00000279, -0.00000058, -0.00000548, -0.00000391, 0.00000203, 0.00000381, 0.00000007, -0.00000282, -0.00000187, -0.00000025, -0.00000003, -0.00000010, -0.00000068, -0.00000273, -0.00000374, -0.00000120, 0.00000149, -0.00000013, -0.00000315, -0.00000225, 0.00000023, -0.00000059, -0.00000051, 0.00000698, 0.00001454, 0.00000817, -0.00000851, -0.00001566, -0.00000688, 0.00000159, -0.00000133, -0.00000495, 0.00000034, 0.00000413, -0.00000306, -0.00000855, 0.00000085, 0.00001171, 0.00000442, -0.00001071, -0.00000782, 0.00001031, 0.00001515, -0.00000018, -0.00001004, -0.00000013, 0.00000985, 0.00000182, -0.00001103, -0.00000911, 0.00000218, 0.00000689, 0.00000479, 0.00000233, -0.00000447, -0.00001548, -0.00001319, 0.00000967, 0.00002594, 0.00000915, -0.00001992, -0.00001891, 0.00001074, 0.00002511, 0.00000669, -0.00001134, -0.00000221, 0.00001438, 0.00001245, 0.00000396, 0.00001123, 0.00002267, 0.00001360, -0.00000666, -0.00000826, 0.00000703, 0.00000741, -0.00001440, -0.00002693, -0.00001027, 0.00001243, 0.00001179, -0.00000553, -0.00001307, -0.00000515, 0.00000188, -0.00000098, -0.00000667, -0.00000877, -0.00000898, -0.00000821, -0.00000460, -0.00000099, -0.00000243, -0.00000577, -0.00000273, 0.00000495, 0.00000489, -0.00000712, -0.00001777, -0.00001303, 0.00000214, 0.00000950, 0.00000079, -0.00001253, -0.00001439, -0.00000337, 0.00000738, 0.00000694, -0.00000144, -0.00000638, },
		{ 0.00000114, -0.00000535, -0.00000427, 0.00000289, 0.00000974, 0.00001122, 0.00000598, -0.00000292, -0.00000902, -0.00000843, -0.00000478, -0.00000396, -0.00000449, 0.00000001, 0.00000851, 0.00001143, 0.00000508, -0.00000172, -0.00000106, 0.00000229, 0.00000018, -0.00000484, -0.00000446, 0.00000107, 0.00000384, 0.00000114, -0.00000206, -0.00000235, -0.00000125, -0.00000002, 0.00000134, 0.00000073, -0.00000291, -0.00000515, -0.00000210, 0.00000166, -0.00000001, -0.00000356, -0.00000195, 0.00000159, -0.00000043, -0.00000330, 0.00000370, 0.00001499, 0.00001249, -0.00000474, -0.00001545, -0.00000854, 0.00000066, -0.00000180, -0.00000599, -0.00000001, 0.00000560, -0.00000200, -0.00001049, -0.00000199, 0.00001235, 0.00000808, -0.00000944, -0.00001118, 0.00000694, 0.00001644, 0.00000363, -0.00000902, -0.00000261, 0.00000811, 0.00000340, -0.00000829, -0.00000870, -0.00000059, 0.00000308, 0.00000352, 0.00000593, 0.00000154, -0.00001408, -0.00002028, 0.00000039, 0.00002531, 0.00001844, -0.00001214, -0.00002225, 0.00000149, 0.00002277, 0.00001424, -0.00000381, -0.00000289, 0.00000939, 0.00001160, 0.00000797, 0.00001403, 0.00002178, 0.00001321, -0.00000340, -0.00000457, 0.00000647, 0.00000342, -0.00001689, -0.00002577, -0.00000844, 0.00001155, 0.00000890, -0.00000702, -0.00001166, -0.00000293, 0.00000216, -0.00000265, -0.00000786, -0.00000792, -0.00000736, -0.00000811, -0.00000612, -0.00000201, -0.00000195, -0.00000550, -0.00000432, 0.00000320, 0.00000626, -0.00000283, -0.00001518, -0.00001562, -0.00000308, 0.00000745, 0.00000384, -0.00000847, -0.00001405, -0.00000654, 0.00000481, 0.00000768, },
		{ 0.00000211, 0.00000721, 0.00000302, -0.00000364, -0.00000510, -0.00000035, 0.00000677, 0.00001138, 0.00000936, 0.00000072, -0.00000796, -0.00000967, -0.00000609, -0.00000454, -0.00000558, -0.00000202, 0.00000731, 0.00001257, 0.00000735, -0.00000077, -0.00000188, 0.00000148, 0.00000062, -0.00000402, -0.00000465, 0.00000004, 0.00000341, 0.00000201, -0.00000090, -0.00000239, -0.00000249, -0.00000068, 0.00000256, 0.00000273, -0.00000243, -0.00000653, -0.00000336, 0.00000174, 0.00000023, -0.00000420, -0.00000214, 0.00000311, 0.00000071, -0.00000544, -0.00000030, 0.00001401, 0.00001618, -0.00000036, -0.00001434, -0.00000978, -0.00000017, -0.00000227, -0.00000737, -0.00000096, 0.00000702, -0.00000011, -0.00001176, -0.00000545, 0.00001174, 0.00001160, -0.00000691, -0.00001357, 0.00000295, 0.00001637, 0.00000714, -0.00000702, -0.00000420, 0.00000602, 0.00000388, -0.00000603, -0.00000738, -0.00000183, -0.00000034, 0.00000055, 0.00000721, 0.00000752, -0.00000959, -0.00002434, -0.00000984, 0.00002024, 0.00002480, -0.00000237, -0.00002139, -0.00000675, 0.00001723, 0.00001882, 0.00000473, -0.00000038, 0.00000547, 0.00000885, 0.00001037, 0.00001781, 0.00002276, 0.00001276, -0.00000185, -0.00000154, 0.00000736, 0.00000094, -0.00001993, -0.00002640, -0.00000719, 0.00001195, 0.00000726, -0.00000900, -0.00001173, -0.00000115, 0.00000353, -0.00000332, -0.00000949, -0.00000822, -0.00000600, -0.00000710, -0.00000687, -0.00000331, -0.00000205, -0.00000506, -0.00000533, 0.00000117, 0.00000639, 0.00000080, -0.00001147, -0.00001629, -0.00000775, 0.00000392, 0.00000502, -0.00000466, -0.00001228, -0.00000835, },
		{ -0.00000974, -0.00000868, -0.00000003, 0.00000596, 0.00000391, -0.00000184, -0.00000479, -0.00000274, 0.00000327, 0.00001006, 0.00001180, 0.00000477, -0.00000580, -0.00001030, -0.00000751, -0.00000528, -0.00000650, -0.00000400, 0.00000559, 0.00001303, 0.00000955, 0.00000077, -0.00000223, 0.00000048, 0.00000061, -0.00000320, -0.00000442, -0.00000086, 0.00000251, 0.00000246, 0.00000054, -0.00000182, -0.00000371, -0.00000207, 0.00000326, 0.00000512, -0.00000114, -0.00000766, -0.00000496, 0.00000166, 0.00000066, -0.00000496, -0.00000291, 0.00000447, 0.00000274, -0.00000659, -0.00000461, 0.00001160, 0.00001888, 0.00000430, -0.00001237, -0.00001051, -0.00000077, -0.00000256, -0.00000894, -0.00000264, 0.00000805, 0.00000249, -0.00001204, -0.00000913, 0.00000982, 0.00001448, -0.00000340, -0.00001468, -0.00000117, 0.00001502, 0.00000991, -0.00000440, -0.00000479, 0.00000405, 0.00000351, -0.00000457, -0.00000576, -0.00000167, -0.00000265, -0.00000324, 0.00000606, 0.00001219, -0.00000304, -0.00002469, -0.00001915, 0.00001162, 0.00002698, 0.00000744, -0.00001665, -0.00001242, 0.00000979, 0.00001987, 0.00001274, 0.00000492, 0.00000367, 0.00000511, 0.00001070, 0.00002156, 0.00002555, 0.00001319, -0.00000151, 0.00000017, 0.00000883, 0.00000011, -0.00002258, -0.00002843, -0.00000717, 0.00001289, 0.00000705, -0.00001068, -0.00001299, -0.00000035, 0.00000540, -0.00000280, -0.00001091, -0.00000949, -0.00000535, -0.00000560, -0.00000669, -0.00000447, -0.00000267, -0.00000474, -0.00000585, -0.00000079, 0.00000553, 0.00000343, -0.00000730, -0.00001521, -0.00001121, -0.00000032, 0.00000437, -0.00000184, },
		{ 0.00000227, -0.00000045, -0.00000722, -0.00000780, -0.00000116, 0.00000451, 0.00000383, -0.00000045, -0.00000371, -0.00000402, -0.00000017, 0.00000750, 0.00001293, 0.00000872, -0.00000266, -0.00001003, -0.00000880, -0.00000621, -0.00000733, -0.00000582, 0.00000352, 0.00001275, 0.00001142, 0.00000277, -0.00000195, -0.00000048, 0.00000014, -0.00000255, -0.00000378, -0.00000143, 0.00000126, 0.00000229, 0.00000202, -0.00000054, -0.00000458, -0.00000409, 0.00000314, 0.00000763, 0.00000102, -0.00000829, -0.00000682, 0.00000132, 0.00000130, -0.00000568, -0.00000431, 0.00000535, 0.00000545, -0.00000652, -0.00000875, 0.00000793, 0.00002030, 0.00000890, -0.00000968, -0.00001068, -0.00000108, -0.00000248, -0.00001046, -0.00000502, 0.00000836, 0.00000550, -0.00001116, -0.00001255, 0.00000672, 0.00001631, 0.00000064, -0.00001441, -0.00000491, 0.00001266, 0.00001169, -0.00000160, -0.00000445, 0.00000258, 0.00000269, -0.00000399, -0.00000443, -0.00000058, -0.00000351, -0.00000688, 0.00000289, 0.00001465, 0.00000420, -0.00002140, -0.00002594, 0.00000099, 0.00002454, 0.00001536, -0.00000911, -0.00001454, 0.00000212, 0.00001755, 0.00001888, 0.00001203, 0.00000463, 0.00000155, 0.00000905, 0.00002435, 0.00002969, 0.00001518, -0.00000153, 0.00000036, 0.00000996, 0.00000060, -0.00002415, -0.00003115, -0.00000864, 0.00001355, 0.00000803, -0.00001145, -0.00001486, -0.00000080, 0.00000711, -0.00000124, -0.00001160, -0.00001131, -0.00000566, -0.00000412, -0.00000568, -0.00000511, -0.00000359, -0.00000472, -0.00000607, -0.00000250, 0.00000402, 0.00000495, -0.00000326, -0.00001279, -0.00001314, -0.00000448, },
		{ -0.00001354, -0.00000792, -0.00000065, -0.00000057, -0.00000541, -0.00000626, -0.00000111, 0.00000347, 0.00000308, 0.00000017, -0.00000236, -0.00000419, -0.00000302, 0.00000413, 0.00001258, 0.00001202, 0.00000116, -0.00000871, -0.00000966, -0.00000728, -0.00000817, -0.00000745, 0.00000131, 0.00001178, 0.00001272, 0.00000498, -0.00000095, -0.00000117, -0.00000075, -0.00000228, -0.00000285, -0.00000148, -0.00000013, 0.00000139, 0.00000324, 0.00000143, -0.00000478, -0.00000651, 0.00000199, 0.00000986, 0.00000398, -0.00000819, -0.00000880, 0.00000064, 0.00000213, -0.00000617, -0.00000626, 0.00000547, 0.00000849, -0.00000511, -0.00001225, 0.00000328, 0.00002029, 0.00001310, -0.00000643, -0.00001032, -0.00000108, -0.00000193, -0.00001168, -0.00000797, 0.00000770, 0.00000852, -0.00000908, -0.00001525, 0.00000275, 0.00001681, 0.00000472, -0.00001285, -0.00000788, 0.00000965, 0.00001239, 0.00000098, -0.00000343, 0.00000179, 0.00000183, -0.00000413, -0.00000377, 0.00000085, -0.00000293, -0.00000955, -0.00000151, 0.00001450, 0.00001071, -0.00001525, -0.00002911, -0.00000977, 0.00001794, 0.00001980, -0.00000049, -0.00001291, -0.00000419, 0.00001274, 0.00002224, 0.00001966, 0.00000835, -0.00000071, 0.00000597, 0.00002550, 0.00003435, 0.00001898, -0.00000101, -0.00000073, 0.00001000, 0.00000176, -0.00002432, -0.00003373, -0.00001142, 0.00001327, 0.00000966, -0.00001099, -0.00001667, -0.00000239, 0.00000810, 0.00000094, -0.00001131, -0.00001316, -0.00000687, -0.00000311, -0.00000414, -0.00000507, -0.00000447, -0.00000500, -0.00000619, -0.00000390, 0.00000218, 0.00000547, 0.00000025, -0.00000954, },
		{ 0.00000301, -0.00000599, -0.00001264, -0.00001026, -0.00000373, -0.00000194, -0.00000475, -0.00000473, -0.00000005, 0.00000330, 0.00000215, -0.00000002, -0.00000123, -0.00000350, -0.00000493, 0.00000050, 0.00001085, 0.00001423, 0.00000518, -0.00000635, -0.00000982, -0.00000834, -0.00000911, -0.00000889, -0.00000088, 0.00001023, 0.00001326, 0.00000712, 0.00000072, -0.00000131, -0.00000190, -0.00000257, -0.00000183, -0.00000088, -0.00000138, -0.00000025, 0.00000385, 0.00000390, -0.00000402, -0.00000896, -0.00000024, 0.00001142, 0.00000753, -0.00000717, -0.00001070, -0.00000043, 0.00000310, -0.00000626, -0.00000859, 0.00000463, 0.00001146, -0.00000243, -0.00001469, -0.00000194, 0.00001878, 0.00001657, -0.00000284, -0.00000948, -0.00000080, -0.00000084, -0.00001231, -0.00001121, 0.00000592, 0.00001112, -0.00000600, -0.00001684, -0.00000164, 0.00001591, 0.00000833, -0.00001027, -0.00000981, 0.00000643, 0.00001209, 0.00000304, -0.00000206, 0.00000167, 0.00000128, -0.00000465, -0.00000391, 0.00000203, -0.00000129, -0.00001074, -0.00000614, 0.00001192, 0.00001531, -0.00000751, -0.00002834, -0.00001883, 0.00000840, 0.00001991, 0.00000731, -0.00000820, -0.00000789, 0.00000675, 0.00002257, 0.00002648, 0.00001430, -0.00000085, 0.00000239, 0.00002476, 0.00003863, 0.00002441, 0.00000086, -0.00000245, 0.00000853, 0.00000272, -0.00002318, -0.00003544, -0.00001501, 0.00001164, 0.00001123, -0.00000934, -0.00001778, -0.00000469, 0.00000803, 0.00000319, -0.00001005, -0.00001457, -0.00000870, -0.00000281, -0.00000246, -0.00000436, -0.00000506, -0.00000544, -0.00000633, -0.00000507, 0.00000019, 0.00000517, },
		{ -0.00000185, 0.00000426, 0.00000500, -0.00000251, -0.00001082, -0.00001140, -0.00000637, -0.00000406, -0.00000536, -0.00000384, 0.00000155, 0.00000418, 0.00000160, -0.00000083, -0.00000070, -0.00000238, -0.00000577, -0.00000283, 0.00000806, 0.00001504, 0.00000889, -0.00000313, -0.00000903, -0.00000914, -0.00001018, -0.00001024, -0.00000289, 0.00000833, 0.00001297, 0.00000884, 0.00000293, -0.00000068, -0.00000307, -0.00000352, -0.00000099, 0.00000038, -0.00000215, -0.00000249, 0.00000356, 0.00000655, -0.00000216, -0.00001101, -0.00000346, 0.00001194, 0.00001136, -0.00000514, -0.00001228, -0.00000190, 0.00000408, -0.00000582, -0.00001104, 0.00000273, 0.00001392, 0.00000130, -0.00001575, -0.00000725, 0.00001588, 0.00001903, 0.00000085, -0.00000826, -0.00000033, 0.00000074, -0.00001213, -0.00001441, 0.00000306, 0.00001288, -0.00000222, -0.00001711, -0.00000597, 0.00001371, 0.00001107, -0.00000705, -0.00001059, 0.00000339, 0.00001101, 0.00000441, -0.00000066, 0.00000207, 0.00000125, -0.00000515, -0.00000473, 0.00000251, 0.00000082, -0.00001032, -0.00001005, 0.00000756, 0.00001729, 0.00000032, -0.00002404, -0.00002477, -0.00000230, 0.00001568, 0.00001255, -0.00000175, -0.00000845, 0.00000108, 0.00002027, 0.00003144, 0.00002150, 0.00000150, -0.00000065, 0.00002236, 0.00004175, 0.00003089, 0.00000455, -0.00000398, 0.00000558, 0.00000271, -0.00002122, -0.00003582, -0.00001869, 0.00000866, 0.00001208, -0.00000688, -0.00001781, -0.00000716, 0.00000689, 0.00000499, -0.00000809, -0.00001524, -0.00001076, -0.00000327, -0.00000100, -0.00000320, -0.00000521, -0.00000583, -0.00000648, -0.00000608, },
		{ -0.00000653, -0.00000697, -0.00000394, 0.00000285, 0.00000628, 0.00000067, -0.00000850, -0.00001147, -0.00000820, -0.00000633, -0.00000707, -0.00000405, 0.00000308, 0.00000601, 0.00000188, -0.00000181, -0.00000090, -0.00000130, -0.00000568, -0.00000542, 0.00000471, 0.00001442, 0.00001175, 0.00000058, -0.00000718, -0.00000940, -0.00001131, -0.00001159, -0.00000467, 0.00000630, 0.00001188, 0.00000986, 0.00000540, 0.00000085, -0.00000393, -0.00000512, -0.00000066, 0.00000219, -0.00000214, -0.00000506, 0.00000214, 0.00000895, 0.00000080, -0.00001220, -0.00000739, 0.00001115, 0.00001506, -0.00000211, -0.00001329, -0.00000371, 0.00000495, -0.00000476, -0.00001332, -0.00000019, 0.00001548, 0.00000573, -0.00001524, -0.00001216, 0.00001180, 0.00002030, 0.00000438, -0.00000674, 0.00000023, 0.00000269, -0.00001101, -0.00001716, -0.00000070, 0.00001346, 0.00000180, -0.00001600, -0.00000972, 0.00001049, 0.00001266, -0.00000363, -0.00001029, 0.00000085, 0.00000945, 0.00000508, 0.00000050, 0.00000274, 0.00000175, -0.00000529, -0.00000591, 0.00000209, 0.00000277, -0.00000856, -0.00001254, 0.00000239, 0.00001654, 0.00000686, -0.00001728, -0.00002683, -0.00001229, 0.00000797, 0.00001405, 0.00000470, -0.00000612, -0.00000300, 0.00001630, 0.00003399, 0.00002880, 0.00000619, -0.00000226, 0.00001892, 0.00004322, 0.00003764, 0.00001014, -0.00000450, 0.00000155, 0.00000118, -0.00001914, -0.00003479, -0.00002174, 0.00000466, 0.00001173, -0.00000420, -0.00001667, -0.00000923, 0.00000493, 0.00000601, -0.00000585, -0.00001511, -0.00001265, -0.00000429, 0.00000000, -0.00000186, -0.00000493, -0.00000599, },
		{ -0.00000440, -0.00000583, -0.00000631, -0.00000766, -0.00000608, 0.00000101, 0.00000693, 0.00000347, -0.00000597, -0.00001078, -0.00000909, -0.00000821, -0.00000941, -0.00000550, 0.00000391, 0.00000836, 0.00000323, -0.00000245, -0.00000170, -0.00000068, -0.00000502, -0.00000703, 0.00000136, 0.00001256, 0.00001337, 0.00000428, -0.00000432, -0.00000883, -0.00001236, -0.00001304, -0.00000621, 0.00000439, 0.00001016, 0.00000991, 0.00000775, 0.00000328, -0.00000412, -0.00000720, -0.00000110, 0.00000428, -0.00000112, -0.00000754, -0.00000046, 0.00001061, 0.00000465, -0.00001215, -0.00001160, 0.00000892, 0.00001817, 0.00000177, -0.00001351, -0.00000574, 0.00000556, -0.00000310, -0.00001512, -0.00000393, 0.00001579, 0.00001039, -0.00001315, -0.00001620, 0.00000689, 0.00002027, 0.00000753, -0.00000505, 0.00000077, 0.00000485, -0.00000894, -0.00001909, -0.00000503, 0.00001269, 0.00000556, -0.00001365, -0.00001251, 0.00000664, 0.00001299, -0.00000043, -0.00000913, -0.00000099, 0.00000772, 0.00000517, 0.00000125, 0.00000338, 0.00000263, -0.00000486, -0.00000703, 0.00000083, 0.00000400, -0.00000601, -0.00001330, -0.00000254, 0.00001349, 0.00001111, -0.00000951, -0.00002504, -0.00001988, -0.00000171, 0.00001144, 0.00000943, -0.00000186, -0.00000475, 0.00001189, 0.00003417, 0.00003518, 0.00001258, -0.00000185, 0.00001529, 0.00004298, 0.00004388, 0.00001733, -0.00000337, -0.00000288, -0.00000208, -0.00001773, -0.00003261, -0.00002363, 0.00000021, 0.00001003, -0.00000189, -0.00001458, -0.00001048, 0.00000259, 0.00000615, -0.00000377, -0.00001437, -0.00001412, -0.00000557, 0.00000053, -0.00000065, },
		{ 0.00000073, 0.00000030, -0.00000389, -0.00000544, -0.00000569, -0.00000798, -0.00000821, -0.00000128, 0.00000697, 0.00000588, -0.00000343, -0.00000962, -0.00000916, -0.00000935, -0.00001182, -0.00000802, 0.00000357, 0.00001063, 0.00000557, -0.00000226, -0.00000272, -0.00000072, -0.00000425, -0.00000770, -0.00000147, 0.00000986, 0.00001355, 0.00000741, -0.00000071, -0.00000721, -0.00001309, -0.00001466, -0.00000760, 0.00000282, 0.00000808, 0.00000888, 0.00000955, 0.00000648, -0.00000331, -0.00000948, -0.00000255, 0.00000627, 0.00000099, -0.00000946, -0.00000409, 0.00001108, 0.00000901, -0.00001058, -0.00001555, 0.00000533, 0.00002027, 0.00000621, -0.00001280, -0.00000783, 0.00000580, -0.00000092, -0.00001615, -0.00000820, 0.00001467, 0.00001475, -0.00000963, -0.00001896, 0.00000155, 0.00001898, 0.00001008, -0.00000329, 0.00000121, 0.00000698, -0.00000601, -0.00001990, -0.00000949, 0.00001056, 0.00000859, -0.00001039, -0.00001406, 0.00000265, 0.00001213, 0.00000218, -0.00000739, -0.00000207, 0.00000609, 0.00000486, 0.00000159, 0.00000379, 0.00000365, -0.00000383, -0.00000772, -0.00000094, 0.00000421, -0.00000337, -0.00001242, -0.00000635, 0.00000901, 0.00001260, -0.00000221, -0.00002015, -0.00002398, -0.00001151, 0.00000519, 0.00001109, 0.00000284, -0.00000410, 0.00000821, 0.00003257, 0.00003993, 0.00001981, 0.00000070, 0.00001232, 0.00004136, 0.00004901, 0.00002554, -0.00000030, -0.00000695, -0.00000690, -0.00001760, -0.00002985, -0.00002411, -0.00000406, 0.00000716, -0.00000042, -0.00001198, -0.00001076, 0.00000034, 0.00000557, -0.00000213, -0.00001333, -0.00001510, -0.00000682, },
		{ -0.00001576, -0.00000783, 0.00000087, 0.00000099, -0.00000363, -0.00000504, -0.00000465, -0.00000772, -0.00001016, -0.00000400, 0.00000637, 0.00000791, -0.00000093, -0.00000824, -0.00000865, -0.00000964, -0.00001378, -0.00001117, 0.00000190, 0.00001217, 0.00000852, -0.00000096, -0.00000345, -0.00000137, -0.00000382, -0.00000768, -0.00000342, 0.00000690, 0.00001230, 0.00000946, 0.00000325, -0.00000444, -0.00001320, -0.00001642, -0.00000899, 0.00000171, 0.00000598, 0.00000681, 0.00001038, 0.00001010, -0.00000124, -0.00001152, -0.00000504, 0.00000772, 0.00000411, -0.00001035, -0.00000843, 0.00001004, 0.00001334, -0.00000746, -0.00001868, 0.00000061, 0.00002100, 0.00001084, -0.00001109, -0.00000978, 0.00000556, 0.00000163, -0.00001619, -0.00001260, 0.00001209, 0.00001831, -0.00000502, -0.00002018, -0.00000374, 0.00001652, 0.00001188, -0.00000158, 0.00000147, 0.00000887, -0.00000245, -0.00001940, -0.00001359, 0.00000724, 0.00001050, -0.00000663, -0.00001429, -0.00000103, 0.00001029, 0.00000396, -0.00000541, -0.00000245, 0.00000474, 0.00000437, 0.00000160, 0.00000383, 0.00000453, -0.00000236, -0.00000770, -0.00000280, 0.00000338, -0.00000126, -0.00001038, -0.00000852, 0.00000414, 0.00001150, 0.00000336, -0.00001346, -0.00002424, -0.00001967, -0.00000347, 0.00000904, 0.00000638, -0.00000173, 0.00000604, 0.00003012, 0.00004288, 0.00002699, 0.00000511, 0.00001065, 0.00003895, 0.00005278, 0.00003412, 0.00000468, -0.00000999, -0.00001282, -0.00001911, -0.00002716, -0.00002329, -0.00000761, 0.00000350, -0.00000004, -0.00000936, -0.00001018, -0.00000146, 0.00000459, -0.00000104, -0.00001233, },
		{ -0.00000035, -0.00001159, -0.00001637, -0.00000863, 0.00000120, 0.00000164, -0.00000372, -0.00000494, -0.00000334, -0.00000673, -0.00001166, -0.00000707, 0.00000502, 0.00000949, 0.00000155, -0.00000675, -0.00000785, -0.00000920, -0.00001494, -0.00001439, -0.00000097, 0.00001247, 0.00001146, 0.00000145, -0.00000335, -0.00000233, -0.00000402, -0.00000742, -0.00000437, 0.00000425, 0.00000994, 0.00001005, 0.00000701, -0.00000059, -0.00001239, -0.00001821, -0.00001055, 0.00000106, 0.00000423, 0.00000389, 0.00000988, 0.00001368, 0.00000215, -0.00001283, -0.00000847, 0.00000817, 0.00000793, -0.00000985, -0.00001297, 0.00000732, 0.00001703, -0.00000295, -0.00002049, -0.00000478, 0.00002017, 0.00001522, -0.00000842, -0.00001138, 0.00000482, 0.00000434, -0.00001513, -0.00001668, 0.00000817, 0.00002062, 0.00000023, -0.00001975, -0.00000852, 0.00001315, 0.00001280, -0.00000001, 0.00000154, 0.00001031, 0.00000144, -0.00001757, -0.00001688, 0.00000309, 0.00001106, -0.00000287, -0.00001327, -0.00000402, 0.00000780, 0.00000481, -0.00000351, -0.00000228, 0.00000377, 0.00000388, 0.00000142, 0.00000351, 0.00000504, -0.00000071, -0.00000691, -0.00000427, 0.00000177, -0.00000016, -0.00000784, -0.00000893, -0.00000010, 0.00000846, 0.00000645, -0.00000646, -0.00002108, -0.00002483, -0.00001284, 0.00000344, 0.00000737, 0.00000114, 0.00000559, 0.00002778, 0.00004431, 0.00003345, 0.00001081, 0.00001059, 0.00003641, 0.00005523, 0.00004249, 0.00001126, -0.00001157, -0.00001923, -0.00002231, -0.00002513, -0.00002152, -0.00001013, -0.00000043, -0.00000069, -0.00000714, -0.00000907, -0.00000267, 0.00000359, },
		{ -0.00000338, 0.00000285, 0.00000023, -0.00001114, -0.00001725, -0.00000942, 0.00000188, 0.00000256, -0.00000409, -0.00000541, -0.00000207, -0.00000497, -0.00001240, -0.00001025, 0.00000286, 0.00001048, 0.00000400, -0.00000513, -0.00000695, -0.00000831, -0.00001519, -0.00001710, -0.00000458, 0.00001127, 0.00001369, 0.00000463, -0.00000204, -0.00000306, -0.00000494, -0.00000741, -0.00000447, 0.00000243, 0.00000695, 0.00000901, 0.00000997, 0.00000409, -0.00001040, -0.00001981, -0.00001243, 0.00000079, 0.00000316, 0.00000050, 0.00000785, 0.00001665, 0.00000675, -0.00001293, -0.00001251, 0.00000728, 0.00001198, -0.00000775, -0.00001709, 0.00000301, 0.00001948, 0.00000256, -0.00002062, -0.00001030, 0.00001774, 0.00001891, -0.00000496, -0.00001244, 0.00000361, 0.00000694, -0.00001297, -0.00002001, 0.00000322, 0.00002137, 0.00000558, -0.00001774, -0.00001237, 0.00000916, 0.00001283, 0.00000132, 0.00000141, 0.00001117, 0.00000531, -0.00001454, -0.00001898, -0.00000144, 0.00001022, 0.00000040, -0.00001128, -0.00000605, 0.00000504, 0.00000477, -0.00000194, -0.00000176, 0.00000318, 0.00000352, 0.00000124, 0.00000296, 0.00000506, 0.00000078, -0.00000548, -0.00000501, -0.00000019, -0.00000019, -0.00000554, -0.00000788, -0.00000299, 0.00000448, 0.00000690, -0.00000056, -0.00001556, -0.00002632, -0.00002112, -0.00000472, 0.00000502, 0.00000305, 0.00000639, 0.00002625, 0.00004489, 0.00003893, 0.00001715, 0.00001209, 0.00003431, 0.00005669, 0.00005030, 0.00001898, -0.00001153, -0.00002555, -0.00002694, -0.00002417, -0.00001928, -0.00001161, -0.00000419, -0.00000209, -0.00000552, -0.00000781, },
		{ -0.00000447, -0.00000672, -0.00000388, 0.00000244, 0.00000106, -0.00001078, -0.00001856, -0.00001054, 0.00000283, 0.00000405, -0.00000442, -0.00000658, -0.00000127, -0.00000264, -0.00001210, -0.00001318, -0.00000007, 0.00001065, 0.00000632, -0.00000329, -0.00000602, -0.00000728, -0.00001468, -0.00001890, -0.00000831, 0.00000868, 0.00001459, 0.00000796, 0.00000060, -0.00000297, -0.00000636, -0.00000810, -0.00000409, 0.00000174, 0.00000398, 0.00000642, 0.00001160, 0.00000917, -0.00000708, -0.00002093, -0.00001472, 0.00000067, 0.00000300, -0.00000288, 0.00000427, 0.00001841, 0.00001220, -0.00001141, -0.00001664, 0.00000483, 0.00001569, -0.00000406, -0.00002018, -0.00000257, 0.00002021, 0.00000849, -0.00001888, -0.00001534, 0.00001389, 0.00002153, -0.00000098, -0.00001281, 0.00000201, 0.00000921, -0.00000983, -0.00002220, -0.00000231, 0.00002042, 0.00001046, -0.00001438, -0.00001499, 0.00000494, 0.00001199, 0.00000232, 0.00000112, 0.00001139, 0.00000883, -0.00001056, -0.00001965, -0.00000584, 0.00000811, 0.00000279, -0.00000868, -0.00000700, 0.00000240, 0.00000398, -0.00000087, -0.00000111, 0.00000289, 0.00000334, 0.00000118, 0.00000234, 0.00000462, 0.00000184, -0.00000370, -0.00000487, -0.00000197, -0.00000123, -0.00000404, -0.00000600, -0.00000419, 0.00000061, 0.00000514, 0.00000326, -0.00000911, -0.00002426, -0.00002683, -0.00001392, -0.00000068, 0.00000269, 0.00000739, 0.00002572, 0.00004532, 0.00004357, 0.00002363, 0.00001484, 0.00003299, 0.00005759, 0.00005748, 0.00002743, -0.00000996, -0.00003138, -0.00003263, -0.00002441, -0.00001704, -0.00001226, -0.00000752, -0.00000386, },
		{ -0.00001041, -0.00000570, -0.00000374, -0.00000593, -0.00000451, 0.00000221, 0.00000236, -0.00001014, -0.00002028, -0.00001235, 0.00000374, 0.00000629, -0.00000430, -0.00000834, -0.00000133, -0.00000013, -0.00001065, -0.00001540, -0.00000354, 0.00000980, 0.00000828, -0.00000116, -0.00000494, -0.00000633, -0.00001377, -0.00001967, -0.00001149, 0.00000514, 0.00001377, 0.00001066, 0.00000431, -0.00000154, -0.00000779, -0.00000972, -0.00000377, 0.00000224, 0.00000166, 0.00000264, 0.00001147, 0.00001408, -0.00000242, -0.00002122, -0.00001741, 0.00000043, 0.00000383, -0.00000571, -0.00000064, 0.00001845, 0.00001796, -0.00000806, -0.00002025, 0.00000088, 0.00001843, 0.00000095, -0.00002169, -0.00000887, 0.00001894, 0.00001417, -0.00001537, -0.00001933, 0.00000897, 0.00002281, 0.00000320, -0.00001242, 0.00000015, 0.00001093, -0.00000596, -0.00002300, -0.00000791, 0.00001781, 0.00001436, -0.00001008, -0.00001619, 0.00000086, 0.00001042, 0.00000294, 0.00000070, 0.00001096, 0.00001170, -0.00000601, -0.00001882, -0.00000961, 0.00000504, 0.00000405, -0.00000590, -0.00000692, 0.00000018, 0.00000269, -0.00000037, -0.00000050, 0.00000279, 0.00000331, 0.00000132, 0.00000185, 0.00000385, 0.00000229, -0.00000194, -0.00000390, -0.00000313, -0.00000286, -0.00000364, -0.00000401, -0.00000383, -0.00000227, 0.00000202, 0.00000456, -0.00000320, -0.00001949, -0.00002904, -0.00002237, -0.00000896, -0.00000081, 0.00000726, 0.00002576, 0.00004615, 0.00004784, 0.00003006, 0.00001846, 0.00003247, 0.00005832, 0.00006416, 0.00003634, -0.00000709, -0.00003653, -0.00003899, -0.00002575, -0.00001512, -0.00001241, },
		{ -0.00001359, -0.00001239, -0.00001305, -0.00000743, -0.00000299, -0.00000530, -0.00000557, 0.00000178, 0.00000415, -0.00000880, -0.00002211, -0.00001509, 0.00000415, 0.00000921, -0.00000329, -0.00001032, -0.00000247, 0.00000201, -0.00000820, -0.00001648, -0.00000713, 0.00000784, 0.00000955, 0.00000119, -0.00000350, -0.00000550, -0.00001286, -0.00001957, -0.00001361, 0.00000138, 0.00001122, 0.00001192, 0.00000853, 0.00000156, -0.00000857, -0.00001224, -0.00000409, 0.00000370, 0.00000055, -0.00000176, 0.00000936, 0.00001818, 0.00000339, -0.00002034, -0.00002036, -0.00000026, 0.00000558, -0.00000746, -0.00000641, 0.00001638, 0.00002330, -0.00000290, -0.00002269, -0.00000430, 0.00001968, 0.00000678, -0.00002125, -0.00001521, 0.00001563, 0.00001889, -0.00001040, -0.00002180, 0.00000346, 0.00002263, 0.00000720, -0.00001128, -0.00000178, 0.00001193, -0.00000168, -0.00002229, -0.00001303, 0.00001379, 0.00001687, -0.00000533, -0.00001596, -0.00000271, 0.00000829, 0.00000313, 0.00000020, 0.00000998, 0.00001372, -0.00000131, -0.00001659, -0.00001233, 0.00000142, 0.00000408, -0.00000337, -0.00000599, -0.00000139, 0.00000115, -0.00000042, -0.00000010, 0.00000277, 0.00000336, 0.00000164, 0.00000164, 0.00000298, 0.00000214, -0.00000056, -0.00000238, -0.00000340, -0.00000456, -0.00000434, -0.00000258, -0.00000239, -0.00000365, -0.00000141, 0.00000352, 0.00000102, -0.00001331, -0.00002765, -0.00002850, -0.00001837, -0.00000753, 0.00000474, 0.00002538, 0.00004742, 0.00005228, 0.00003661, 0.00002268, 0.00003260, 0.00005906, 0.00007059, 0.00004567, -0.00000313, -0.00004099, -0.00004576, -0.00002796, },
		{ -0.00005288, -0.00003082, -0.00001233, -0.00001236, -0.00001576, -0.00000917, -0.00000195, -0.00000450, -0.00000709, 0.00000073, 0.00000617, -0.00000647, -0.00002355, -0.00001873, 0.00000355, 0.00001243, -0.00000107, -0.00001198, -0.00000467, 0.00000322, -0.00000515, -0.00001613, -0.00001028, 0.00000490, 0.00000973, 0.00000354, -0.00000149, -0.00000462, -0.00001228, -0.00001906, -0.00001443, -0.00000181, 0.00000735, 0.00001116, 0.00001239, 0.00000629, -0.00000801, -0.00001534, -0.00000553, 0.00000565, 0.00000100, -0.00000607, 0.00000535, 0.00002083, 0.00000996, -0.00001802, -0.00002332, -0.00000167, 0.00000796, -0.00000772, -0.00001243, 0.00001211, 0.00002745, 0.00000375, -0.00002336, -0.00001020, 0.00001907, 0.00001277, -0.00001874, -0.00002084, 0.00001052, 0.00002207, -0.00000449, -0.00002249, -0.00000207, 0.00002101, 0.00001067, -0.00000948, -0.00000361, 0.00001216, 0.00000266, -0.00002011, -0.00001717, 0.00000877, 0.00001777, -0.00000065, -0.00001443, -0.00000547, 0.00000586, 0.00000291, -0.00000037, 0.00000856, 0.00001477, 0.00000313, -0.00001323, -0.00001373, -0.00000225, 0.00000297, -0.00000145, -0.00000450, -0.00000219, -0.00000035, -0.00000092, 0.00000004, 0.00000273, 0.00000341, 0.00000208, 0.00000175, 0.00000225, 0.00000152, 0.00000020, -0.00000069, -0.00000274, -0.00000580, -0.00000581, -0.00000215, -0.00000058, -0.00000345, -0.00000420, 0.00000084, 0.00000287, -0.00000721, -0.00002326, -0.00003123, -0.00002717, -0.00001676, -0.00000100, 0.00002328, 0.00004859, 0.00005724, 0.00004376, 0.00002752, 0.00003314, 0.00005976, 0.00007693, 0.00005555, 0.00000185, -0.00004489, },
		{ 0.00000800, -0.00004830, -0.00006046, -0.00003427, -0.00001110, -0.00001223, -0.00001881, -0.00001124, -0.00000054, -0.00000311, -0.00000886, -0.00000129, 0.00000794, -0.00000310, -0.00002402, -0.00002293, 0.00000154, 0.00001536, 0.00000236, -0.00001267, -0.00000757, 0.00000308, -0.00000216, -0.00001436, -0.00001234, 0.00000137, 0.00000854, 0.00000544, 0.00000116, -0.00000334, -0.00001215, -0.00001869, -0.00001409, -0.00000367, 0.00000291, 0.00000813, 0.00001494, 0.00001221, -0.00000553, -0.00001839, -0.00000834, 0.00000746, 0.00000306, -0.00000955, -0.00000023, 0.00002152, 0.00001676, -0.00001408, -0.00002591, -0.00000398, 0.00001056, -0.00000627, -0.00001795, 0.00000581, 0.00002970, 0.00001132, -0.00002188, -0.00001610, 0.00001646, 0.00001821, -0.00001432, -0.00002508, 0.00000409, 0.00002327, 0.00000171, -0.00002134, -0.00000706, 0.00001819, 0.00001333, -0.00000719, -0.00000516, 0.00001161, 0.00000666, -0.00001667, -0.00001995, 0.00000328, 0.00001700, 0.00000343, -0.00001190, -0.00000722, 0.00000339, 0.00000231, -0.00000098, 0.00000684, 0.00001486, 0.00000693, -0.00000911, -0.00001369, -0.00000550, 0.00000095, -0.00000038, -0.00000280, -0.00000224, -0.00000159, -0.00000173, -0.00000013, 0.00000260, 0.00000339, 0.00000252, 0.00000217, 0.00000186, 0.00000068, 0.00000027, 0.00000074, -0.00000138, -0.00000621, -0.00000755, -0.00000283, 0.00000089, -0.00000203, -0.00000567, -0.00000251, 0.00000231, -0.00000248, -0.00001709, -0.00003023, -0.00003368, -0.00002710, -0.00001008, 0.00001818, 0.00004855, 0.00006258, 0.00005201, 0.00003337, 0.00003404, 0.00006020, 0.00008316, 0.00006616, },
		{ 0.00008903, 0.00007760, 0.00001561, -0.00005111, -0.00006870, -0.00003843, -0.00000970, -0.00001172, -0.00002227, -0.00001407, 0.00000101, -0.00000076, -0.00001039, -0.00000437, 0.00000884, 0.00000101, -0.00002305, -0.00002706, -0.00000196, 0.00001730, 0.00000668, -0.00001189, -0.00001050, 0.00000144, 0.00000000, -0.00001151, -0.00001281, -0.00000211, 0.00000590, 0.00000632, 0.00000433, -0.00000123, -0.00001229, -0.00001900, -0.00001308, -0.00000376, -0.00000112, 0.00000304, 0.00001527, 0.00001854, -0.00000084, -0.00002061, -0.00001250, 0.00000843, 0.00000653, -0.00001154, -0.00000682, 0.00001990, 0.00002311, -0.00000857, -0.00002770, -0.00000723, 0.00001288, -0.00000310, -0.00002221, -0.00000206, 0.00002947, 0.00001901, -0.00001810, -0.00002124, 0.00001203, 0.00002239, -0.00000841, -0.00002739, -0.00000298, 0.00002232, 0.00000748, -0.00001853, -0.00001103, 0.00001450, 0.00001500, -0.00000464, -0.00000632, 0.00001038, 0.00001003, -0.00001229, -0.00002111, -0.00000210, 0.00001474, 0.00000650, -0.00000875, -0.00000791, 0.00000115, 0.00000143, -0.00000163, 0.00000497, 0.00001408, 0.00000984, -0.00000466, -0.00001229, -0.00000792, -0.00000160, -0.00000027, -0.00000122, -0.00000166, -0.00000239, -0.00000266, -0.00000057, 0.00000234, 0.00000325, 0.00000284, 0.00000277, 0.00000190, -0.00000009, -0.00000026, 0.00000160, 0.00000032, -0.00000565, -0.00000899, -0.00000441, 0.00000149, -0.00000001, -0.00000559, -0.00000553, -0.00000012, 0.00000004, -0.00001062, -0.00002600, -0.00003670, -0.00003673, -0.00002175, 0.00000922, 0.00004584, 0.00006753, 0.00006163, 0.00004089, 0.00003557, 0.00006014, },
		{ 0.00003839, 0.00005958, 0.00009409, 0.00008970, 0.00002502, -0.00005297, -0.00007772, -0.00004366, -0.00000813, -0.00001042, -0.00002593, -0.00001803, 0.00000222, 0.00000266, -0.00001107, -0.00000826, 0.00000834, 0.00000526, -0.00002046, -0.00003036, -0.00000666, 0.00001762, 0.00001119, -0.00000942, -0.00001264, -0.00000143, 0.00000066, -0.00000827, -0.00001141, -0.00000470, 0.00000204, 0.00000563, 0.00000756, 0.00000206, -0.00001226, -0.00002035, -0.00001214, -0.00000202, -0.00000371, -0.00000345, 0.00001277, 0.00002422, 0.00000598, -0.00002120, -0.00001763, 0.00000800, 0.00001089, -0.00001157, -0.00001365, 0.00001591, 0.00002832, -0.00000170, -0.00002826, -0.00001128, 0.00001440, 0.00000151, -0.00002457, -0.00001078, 0.00002648, 0.00002593, -0.00001222, -0.00002489, 0.00000621, 0.00002477, -0.00000166, -0.00002750, -0.00000991, 0.00001932, 0.00001219, -0.00001448, -0.00001364, 0.00001038, 0.00001560, -0.00000206, -0.00000699, 0.00000862, 0.00001250, -0.00000740, -0.00002061, -0.00000683, 0.00001131, 0.00000827, -0.00000546, -0.00000759, -0.00000063, 0.00000039, -0.00000229, 0.00000309, 0.00001259, 0.00001168, -0.00000032, -0.00000974, -0.00000923, -0.00000428, -0.00000107, -0.00000006, -0.00000067, -0.00000267, -0.00000355, -0.00000123, 0.00000194, 0.00000299, 0.00000298, 0.00000339, 0.00000235, -0.00000053, -0.00000115, 0.00000171, 0.00000191, -0.00000425, -0.00000968, -0.00000645, 0.00000096, 0.00000189, -0.00000415, -0.00000742, -0.00000353, 0.00000011, -0.00000525, -0.00001966, -0.00003575, -0.00004384, -0.00003450, -0.00000370, 0.00003899, 0.00007070, 0.00007237, 0.00005077, },
		{ 0.00008321, 0.00006345, 0.00004353, 0.00005892, 0.00009790, 0.00010198, 0.00003641, -0.00005335, -0.00008739, -0.00005040, -0.00000668, -0.00000796, -0.00002929, -0.00002328, 0.00000244, 0.00000692, -0.00001032, -0.00001237, 0.00000619, 0.00000885, -0.00001646, -0.00003211, -0.00001190, 0.00001600, 0.00001500, -0.00000546, -0.00001320, -0.00000486, -0.00000061, -0.00000559, -0.00000830, -0.00000559, -0.00000241, 0.00000294, 0.00001018, 0.00000663, -0.00001141, -0.00002273, -0.00001212, 0.00000115, -0.00000404, -0.00001028, 0.00000727, 0.00002811, 0.00001441, -0.00001948, -0.00002308, 0.00000577, 0.00001547, -0.00000944, -0.00001989, 0.00000975, 0.00003172, 0.00000610, -0.00002722, -0.00001579, 0.00001468, 0.00000709, -0.00002458, -0.00001944, 0.00002078, 0.00003117, -0.00000474, -0.00002646, -0.00000033, 0.00002505, 0.00000517, -0.00002541, -0.00001596, 0.00001464, 0.00001532, -0.00000972, -0.00001475, 0.00000628, 0.00001520, 0.00000032, -0.00000718, 0.00000655, 0.00001395, -0.00000246, -0.00001858, -0.00001044, 0.00000719, 0.00000865, -0.00000245, -0.00000646, -0.00000180, -0.00000064, -0.00000296, 0.00000130, 0.00001059, 0.00001244, 0.00000350, -0.00000639, -0.00000931, -0.00000666, -0.00000260, 0.00000046, 0.00000049, -0.00000244, -0.00000425, -0.00000204, 0.00000141, 0.00000263, 0.00000289, 0.00000386, 0.00000309, -0.00000047, -0.00000205, 0.00000112, 0.00000298, -0.00000235, -0.00000938, -0.00000835, -0.00000058, 0.00000303, -0.00000192, -0.00000781, -0.00000689, -0.00000188, -0.00000199, -0.00001274, -0.00003121, -0.00004703, -0.00004631, -0.00001975, 0.00002699, 0.00007025, },
		{ 0.00000964, 0.00006431, 0.00009240, 0.00007872, 0.00005210, 0.00005908, 0.00010029, 0.00011373, 0.00004964, -0.00005164, -0.00009729, -0.00005898, -0.00000591, -0.00000424, -0.00003171, -0.00002959, 0.00000104, 0.00001141, -0.00000786, -0.00001586, 0.00000255, 0.00001098, -0.00001170, -0.00003185, -0.00001670, 0.00001255, 0.00001719, -0.00000068, -0.00001164, -0.00000785, -0.00000378, -0.00000447, -0.00000410, -0.00000415, -0.00000654, -0.00000181, 0.00001133, 0.00001224, -0.00000905, -0.00002575, -0.00001372, 0.00000496, -0.00000165, -0.00001620, -0.00000090, 0.00002918, 0.00002355, -0.00001507, -0.00002801, 0.00000166, 0.00001948, -0.00000524, -0.00002474, 0.00000189, 0.00003279, 0.00001422, -0.00002436, -0.00002031, 0.00001344, 0.00001296, -0.00002209, -0.00002708, 0.00001274, 0.00003399, 0.00000357, -0.00002564, -0.00000679, 0.00002323, 0.00001134, -0.00002140, -0.00002050, 0.00000885, 0.00001660, -0.00000487, -0.00001440, 0.00000263, 0.00001398, 0.00000232, -0.00000691, 0.00000437, 0.00001436, 0.00000210, -0.00001531, -0.00001263, 0.00000291, 0.00000773, -0.00000011, -0.00000482, -0.00000230, -0.00000152, -0.00000358, -0.00000031, 0.00000831, 0.00001219, 0.00000651, -0.00000265, -0.00000819, -0.00000839, -0.00000456, 0.00000027, 0.00000156, -0.00000177, -0.00000465, -0.00000289, 0.00000077, 0.00000221, 0.00000261, 0.00000408, 0.00000392, 0.00000009, -0.00000266, 0.00000004, 0.00000329, -0.00000041, -0.00000811, -0.00000961, -0.00000275, 0.00000309, 0.00000039, -0.00000677, -0.00000933, -0.00000511, -0.00000123, -0.00000677, -0.00002423, -0.00004567, -0.00005509, -0.00003721, },
		{ -0.00005913, -0.00005371, -0.00001222, 0.00005143, 0.00009759, 0.00009551, 0.00006493, 0.00006149, 0.00010159, 0.00012418, 0.00006414, -0.00004738, -0.00010666, -0.00006947, -0.00000653, 0.00000048, -0.00003254, -0.00003635, -0.00000232, 0.00001527, -0.00000379, -0.00001790, -0.00000195, 0.00001113, -0.00000716, -0.00002962, -0.00002005, 0.00000793, 0.00001710, 0.00000386, -0.00000789, -0.00000927, -0.00000830, -0.00000574, 0.00000018, -0.00000013, -0.00000921, -0.00000828, 0.00001020, 0.00001826, -0.00000463, -0.00002867, -0.00001736, 0.00000831, 0.00000343, -0.00001997, -0.00001089, 0.00002672, 0.00003222, -0.00000799, -0.00003152, -0.00000406, 0.00002220, 0.00000061, -0.00002754, -0.00000696, 0.00003124, 0.00002196, -0.00001965, -0.00002428, 0.00001061, 0.00001840, -0.00001725, -0.00003284, 0.00000310, 0.00003390, 0.00001179, -0.00002245, -0.00001237, 0.00001958, 0.00001618, -0.00001600, -0.00002312, 0.00000264, 0.00001599, -0.00000053, -0.00001285, -0.00000024, 0.00001222, 0.00000381, -0.00000629, 0.00000228, 0.00001381, 0.00000591, -0.00001122, -0.00001326, -0.00000099, 0.00000576, 0.00000131, -0.00000301, -0.00000216, -0.00000212, -0.00000412, -0.00000171, 0.00000594, 0.00001108, 0.00000853, 0.00000107, -0.00000606, -0.00000922, -0.00000661, -0.00000060, 0.00000231, -0.00000080, -0.00000467, -0.00000370, 0.00000003, 0.00000176, 0.00000221, 0.00000397, 0.00000461, 0.00000107, -0.00000275, -0.00000119, 0.00000279, 0.00000112, -0.00000617, -0.00000989, -0.00000499, 0.00000204, 0.00000212, -0.00000478, -0.00001035, -0.00000851, -0.00000270, -0.00000293, -0.00001644, -0.00004005, },
		{ -0.00000959, -0.00003138, -0.00005751, -0.00006666, -0.00003662, 0.00003105, 0.00009624, 0.00011174, 0.00008217, 0.00006779, 0.00010286, 0.00013280, 0.00007898, -0.00004054, -0.00011462, -0.00008148, -0.00000920, 0.00000554, -0.00003141, -0.00004267, -0.00000757, 0.00001757, 0.00000124, -0.00001790, -0.00000633, 0.00000925, -0.00000393, -0.00002601, -0.00002114, 0.00000321, 0.00001452, 0.00000693, -0.00000246, -0.00000817, -0.00001312, -0.00000981, 0.00000324, 0.00000616, -0.00000935, -0.00001563, 0.00000619, 0.00002372, 0.00000204, -0.00003050, -0.00002299, 0.00001009, 0.00001059, -0.00002057, -0.00002144, 0.00002051, 0.00003915, 0.00000132, -0.00003283, -0.00001085, 0.00002307, 0.00000748, -0.00002787, -0.00001595, 0.00002704, 0.00002856, -0.00001330, -0.00002714, 0.00000636, 0.00002270, -0.00001054, -0.00003603, -0.00000721, 0.00003075, 0.00001893, -0.00001723, -0.00001638, 0.00001464, 0.00001925, -0.00000988, -0.00002366, -0.00000327, 0.00001371, 0.00000280, -0.00001049, -0.00000216, 0.00001021, 0.00000476, -0.00000543, 0.00000045, 0.00001250, 0.00000873, -0.00000678, -0.00001241, -0.00000407, 0.00000314, 0.00000171, -0.00000137, -0.00000152, -0.00000235, -0.00000450, -0.00000286, 0.00000366, 0.00000935, 0.00000949, 0.00000438, -0.00000320, -0.00000901, -0.00000842, -0.00000203, 0.00000258, 0.00000029, -0.00000433, -0.00000438, -0.00000079, 0.00000129, 0.00000177, 0.00000359, 0.00000501, 0.00000223, -0.00000224, -0.00000222, 0.00000164, 0.00000191, -0.00000400, -0.00000913, -0.00000672, 0.00000024, 0.00000282, -0.00000255, -0.00000992, -0.00001114, -0.00000559, -0.00000178, },
		{ -0.00000878, -0.00000316, -0.00000513, -0.00002152, -0.00005038, -0.00007381, -0.00006073, 0.00000387, 0.00008616, 0.00012448, 0.00010292, 0.00007947, 0.00010584, 0.00013969, 0.00009299, -0.00003162, -0.00012040, -0.00009420, -0.00001428, 0.00000998, -0.00002842, -0.00004761, -0.00001409, 0.00001764, 0.00000621, -0.00001573, -0.00000942, 0.00000587, -0.00000291, -0.00002209, -0.00001966, -0.00000026, 0.00000989, 0.00000737, 0.00000358, -0.00000396, -0.00001683, -0.00001649, 0.00000377, 0.00001385, -0.00000615, -0.00002263, -0.00000087, 0.00002747, 0.00001071, -0.00003025, -0.00003007, 0.00000931, 0.00001880, -0.00001746, -0.00003109, 0.00001086, 0.00004319, 0.00001200, -0.00003140, -0.00001792, 0.00002179, 0.00001458, -0.00002559, -0.00002417, 0.00002045, 0.00003332, -0.00000573, -0.00002841, 0.00000108, 0.00002531, -0.00000266, -0.00003627, -0.00001711, 0.00002479, 0.00002412, -0.00001062, -0.00001838, 0.00000913, 0.00002037, -0.00000377, -0.00002224, -0.00000824, 0.00001019, 0.00000480, -0.00000779, -0.00000310, 0.00000824, 0.00000522, -0.00000446, -0.00000102, 0.00001065, 0.00001044, -0.00000246, -0.00001031, -0.00000600, 0.00000033, 0.00000115, -0.00000019, -0.00000060, -0.00000219, -0.00000467, -0.00000373, 0.00000160, 0.00000724, 0.00000945, 0.00000699, 0.00000002, -0.00000780, -0.00000967, -0.00000379, 0.00000230, 0.00000131, -0.00000365, -0.00000485, -0.00000166, 0.00000080, 0.00000138, 0.00000302, 0.00000500, 0.00000333, -0.00000119, -0.00000275, 0.00000018, 0.00000183, -0.00000209, -0.00000755, -0.00000753, -0.00000177, 0.00000242, -0.00000079, -0.00000847, -0.00001236, },
		{ -0.00000673, -0.00001206, -0.00001117, -0.00000622, -0.00000386, -0.00001255, -0.00003898, -0.00007374, -0.00008129, -0.00002805, 0.00006612, 0.00013040, 0.00012499, 0.00009732, 0.00011279, 0.00014581, 0.00010521, -0.00002175, -0.00012363, -0.00010649, -0.00002158, 0.00001280, -0.00002423, -0.00005043, -0.00002081, 0.00001530, 0.00000986, -0.00001193, -0.00001022, 0.00000211, -0.00000454, -0.00001930, -0.00001595, -0.00000119, 0.00000430, 0.00000448, 0.00000872, 0.00000326, -0.00001795, -0.00002494, 0.00000073, 0.00002158, 0.00000071, -0.00002784, -0.00001058, 0.00002836, 0.00002064, -0.00002708, -0.00003762, 0.00000542, 0.00002669, -0.00001065, -0.00003837, -0.00000135, 0.00004349, 0.00002291, -0.00002706, -0.00002439, 0.00001835, 0.00002108, -0.00002089, -0.00003081, 0.00001200, 0.00003573, 0.00000244, -0.00002778, -0.00000464, 0.00002588, 0.00000552, -0.00003357, -0.00002557, 0.00001664, 0.00002674, -0.00000348, -0.00001820, 0.00000381, 0.00001966, 0.00000165, -0.00001924, -0.00001181, 0.00000597, 0.00000536, -0.00000524, -0.00000319, 0.00000657, 0.00000529, -0.00000348, -0.00000208, 0.00000852, 0.00001107, 0.00000134, -0.00000734, -0.00000666, -0.00000223, -0.00000017, 0.00000035, 0.00000036, -0.00000170, -0.00000459, -0.00000432, -0.00000015, 0.00000497, 0.00000855, 0.00000870, 0.00000322, -0.00000574, -0.00001017, -0.00000560, 0.00000148, 0.00000210, -0.00000272, -0.00000506, -0.00000254, 0.00000026, 0.00000107, 0.00000240, 0.00000461, 0.00000413, 0.00000020, -0.00000262, -0.00000122, 0.00000098, -0.00000085, -0.00000559, -0.00000729, -0.00000340, 0.00000120, -0.00000002, },
		{ -0.00000028, -0.00000039, -0.00000548, -0.00001066, -0.00001198, -0.00000972, -0.00000575, -0.00000636, -0.00002542, -0.00006624, -0.00009525, -0.00006144, 0.00003629, 0.00012628, 0.00014500, 0.00012093, 0.00012599, 0.00015314, 0.00011527, -0.00001245, -0.00012469, -0.00011718, -0.00003027, 0.00001328, -0.00001996, -0.00005094, -0.00002643, 0.00001104, 0.00001108, -0.00000764, -0.00000820, -0.00000053, -0.00000854, -0.00001905, -0.00001107, 0.00000131, -0.00000067, -0.00000182, 0.00001128, 0.00001268, -0.00001524, -0.00003371, -0.00000638, 0.00002766, 0.00001088, -0.00002985, -0.00002199, 0.00002553, 0.00003063, -0.00002059, -0.00004439, -0.00000163, 0.00003284, -0.00000077, -0.00004209, -0.00001483, 0.00003966, 0.00003276, -0.00002004, -0.00002940, 0.00001308, 0.00002626, -0.00001423, -0.00003519, 0.00000244, 0.00003546, 0.00001048, -0.00002516, -0.00001013, 0.00002437, 0.00001311, -0.00002825, -0.00003174, 0.00000717, 0.00002648, 0.00000325, -0.00001602, -0.00000065, 0.00001747, 0.00000589, -0.00001520, -0.00001372, 0.00000168, 0.00000458, -0.00000325, -0.00000269, 0.00000534, 0.00000513, -0.00000258, -0.00000274, 0.00000636, 0.00001075, 0.00000432, -0.00000392, -0.00000607, -0.00000414, -0.00000193, 0.00000017, 0.00000112, -0.00000098, -0.00000425, -0.00000462, -0.00000153, 0.00000275, 0.00000701, 0.00000945, 0.00000606, -0.00000307, -0.00000980, -0.00000719, 0.00000022, 0.00000252, -0.00000166, -0.00000498, -0.00000337, -0.00000036, 0.00000085, 0.00000184, 0.00000393, 0.00000450, 0.00000164, -0.00000185, -0.00000221, -0.00000034, -0.00000048, -0.00000376, -0.00000617, -0.00000414, },
		{ -0.00000459, -0.00000381, -0.00000135, -0.00000164, -0.00000537, -0.00000893, -0.00001098, -0.00001231, -0.00000995, -0.00000414, -0.00001229, -0.00005243, -0.00010036, -0.00009228, -0.00000146, 0.00010981, 0.00015871, 0.00014832, 0.00014719, 0.00016451, 0.00012382, -0.00000532, -0.00012478, -0.00012551, -0.00003897, 0.00001128, -0.00001704, -0.00004967, -0.00002975, 0.00000603, 0.00000923, -0.00000447, -0.00000359, -0.00000051, -0.00001385, -0.00002239, -0.00000669, 0.00000744, -0.00000317, -0.00001074, 0.00000983, 0.00002279, -0.00000803, -0.00004094, -0.00001730, 0.00003041, 0.00002331, -0.00002756, -0.00003366, 0.00001859, 0.00003926, -0.00001090, -0.00004904, -0.00001132, 0.00003602, 0.00001104, -0.00004150, -0.00002804, 0.00003187, 0.00004037, -0.00001099, -0.00003229, 0.00000654, 0.00002955, -0.00000628, -0.00003691, -0.00000734, 0.00003250, 0.00001760, -0.00002072, -0.00001474, 0.00002100, 0.00001934, -0.00002098, -0.00003506, -0.00000257, 0.00002342, 0.00000872, -0.00001231, -0.00000373, 0.00001436, 0.00000866, -0.00001072, -0.00001400, -0.00000212, 0.00000276, -0.00000210, -0.00000191, 0.00000458, 0.00000490, -0.00000178, -0.00000304, 0.00000436, 0.00000970, 0.00000631, -0.00000052, -0.00000445, -0.00000515, -0.00000375, -0.00000064, 0.00000150, -0.00000019, -0.00000368, -0.00000464, -0.00000254, 0.00000075, 0.00000507, 0.00000927, 0.00000827, -0.00000011, -0.00000859, -0.00000832, -0.00000130, 0.00000252, -0.00000061, -0.00000461, -0.00000410, -0.00000107, 0.00000066, 0.00000145, 0.00000311, 0.00000439, 0.00000286, -0.00000062, -0.00000255, -0.00000173, -0.00000092, -0.00000252, },
		{ -0.00000191, -0.00000215, -0.00000311, -0.00000252, -0.00000144, -0.00000318, -0.00000663, -0.00000783, -0.00000852, -0.00001295, -0.00001502, -0.00000617, -0.00000208, -0.00003464, -0.00009569, -0.00011645, -0.00004358, 0.00008021, 0.00016168, 0.00017587, 0.00017689, 0.00018319, 0.00013287, -0.00000146, -0.00012586, -0.00013147, -0.00004602, 0.00000753, -0.00001672, -0.00004790, -0.00003007, 0.00000200, 0.00000446, -0.00000414, 0.00000253, 0.00000338, -0.00001871, -0.00002963, -0.00000480, 0.00001649, -0.00000155, -0.00002072, 0.00000347, 0.00003155, 0.00000357, -0.00004473, -0.00003100, 0.00002849, 0.00003639, -0.00002047, -0.00004385, 0.00000781, 0.00004508, 0.00000128, -0.00005044, -0.00002257, 0.00003540, 0.00002331, -0.00003648, -0.00003940, 0.00002081, 0.00004476, -0.00000082, -0.00003266, -0.00000053, 0.00003065, 0.00000214, -0.00003584, -0.00001644, 0.00002712, 0.00002313, -0.00001487, -0.00001792, 0.00001626, 0.00002363, -0.00001262, -0.00003533, -0.00001151, 0.00001801, 0.00001224, -0.00000777, -0.00000518, 0.00001095, 0.00000993, -0.00000640, -0.00001284, -0.00000500, 0.00000032, -0.00000190, -0.00000121, 0.00000423, 0.00000472, -0.00000107, -0.00000304, 0.00000266, 0.00000817, 0.00000730, 0.00000250, -0.00000210, -0.00000515, -0.00000528, -0.00000190, 0.00000139, 0.00000049, -0.00000295, -0.00000441, -0.00000317, -0.00000092, 0.00000296, 0.00000829, 0.00000968, 0.00000283, -0.00000666, -0.00000882, -0.00000288, 0.00000210, 0.00000032, -0.00000401, -0.00000467, -0.00000187, 0.00000044, 0.00000123, 0.00000232, 0.00000389, 0.00000367, 0.00000079, -0.00000223, -0.00000280, },
		{ -0.00000139, -0.00000327, -0.00000303, -0.00000264, -0.00000226, -0.00000071, -0.00000026, -0.00000422, -0.00000905, -0.00000818, -0.00000551, -0.00001115, -0.00001928, -0.00001168, 0.00000336, -0.00001598, -0.00008199, -0.00013051, -0.00008532, 0.00003873, 0.00015010, 0.00019852, 0.00021380, 0.00021223, 0.00014573, -0.00000089, -0.00013034, -0.00013619, -0.00004993, 0.00000363, -0.00001969, -0.00004749, -0.00002754, 0.00000079, -0.00000212, -0.00000803, 0.00000829, 0.00001158, -0.00002096, -0.00004012, -0.00000726, 0.00002686, 0.00000529, -0.00002959, -0.00000779, 0.00003683, 0.00001851, -0.00004347, -0.00004572, 0.00002118, 0.00004810, -0.00000878, -0.00005084, -0.00000593, 0.00004690, 0.00001473, -0.00004787, -0.00003400, 0.00003072, 0.00003443, -0.00002748, -0.00004759, 0.00000764, 0.00004538, 0.00000934, -0.00003045, -0.00000735, 0.00002954, 0.00001024, -0.00003217, -0.00002402, 0.00001986, 0.00002652, -0.00000821, -0.00001930, 0.00001081, 0.00002569, -0.00000411, -0.00003273, -0.00001874, 0.00001099, 0.00001343, -0.00000317, -0.00000505, 0.00000784, 0.00000992, -0.00000272, -0.00001065, -0.00000670, -0.00000227, -0.00000260, -0.00000086, 0.00000413, 0.00000467, -0.00000042, -0.00000282, 0.00000133, 0.00000644, 0.00000739, 0.00000485, 0.00000057, -0.00000420, -0.00000625, -0.00000336, 0.00000078, 0.00000092, -0.00000217, -0.00000396, -0.00000344, -0.00000220, 0.00000089, 0.00000671, 0.00001020, 0.00000543, -0.00000422, -0.00000860, -0.00000429, 0.00000132, 0.00000101, -0.00000324, -0.00000503, -0.00000273, 0.00000013, 0.00000115, 0.00000169, 0.00000316, 0.00000400, 0.00000208, },
		{ 0.00000389, 0.00000301, -0.00000031, -0.00000307, -0.00000384, -0.00000373, -0.00000239, 0.00000098, 0.00000213, -0.00000404, -0.00001197, -0.00001041, -0.00000310, -0.00000719, -0.00002125, -0.00001907, 0.00000323, 0.00000027, -0.00006158, -0.00013245, -0.00012133, -0.00001120, 0.00012170, 0.00021042, 0.00025445, 0.00025364, 0.00016675, -0.00000205, -0.00014047, -0.00014196, -0.00004995, 0.00000186, -0.00002567, -0.00005051, -0.00002337, 0.00000395, -0.00000852, -0.00001673, 0.00001132, 0.00002349, -0.00001851, -0.00005222, -0.00001539, 0.00003623, 0.00001751, -0.00003496, -0.00002297, 0.00003674, 0.00003494, -0.00003621, -0.00005923, 0.00000859, 0.00005642, 0.00000648, -0.00005322, -0.00002113, 0.00004401, 0.00002788, -0.00004119, -0.00004407, 0.00002234, 0.00004297, -0.00001554, -0.00005168, -0.00000624, 0.00004215, 0.00001841, -0.00002597, -0.00001319, 0.00002647, 0.00001725, -0.00002636, -0.00002946, 0.00001146, 0.00002750, -0.00000148, -0.00001881, 0.00000537, 0.00002555, 0.00000365, -0.00002777, -0.00002359, 0.00000329, 0.00001224, 0.00000069, -0.00000365, 0.00000549, 0.00000900, 0.00000002, -0.00000791, -0.00000716, -0.00000456, -0.00000400, -0.00000108, 0.00000406, 0.00000476, 0.00000021, -0.00000242, 0.00000040, 0.00000471, 0.00000677, 0.00000638, 0.00000321, -0.00000249, -0.00000646, -0.00000472, -0.00000021, 0.00000100, -0.00000146, -0.00000337, -0.00000342, -0.00000305, -0.00000096, 0.00000474, 0.00000988, 0.00000747, -0.00000152, -0.00000768, -0.00000536, 0.00000031, 0.00000141, -0.00000239, -0.00000515, -0.00000358, -0.00000033, 0.00000112, 0.00000127, 0.00000240, },
		{ 0.00000105, 0.00000175, 0.00000351, 0.00000347, 0.00000067, -0.00000232, -0.00000400, -0.00000497, -0.00000358, 0.00000189, 0.00000526, -0.00000217, -0.00001445, -0.00001440, -0.00000242, -0.00000199, -0.00002005, -0.00002624, -0.00000202, 0.00001133, -0.00003804, -0.00012215, -0.00014664, -0.00006424, 0.00007645, 0.00020587, 0.00029327, 0.00030749, 0.00020065, -0.00000155, -0.00015761, -0.00015208, -0.00004660, 0.00000475, -0.00003316, -0.00005865, -0.00001974, 0.00001218, -0.00001217, -0.00002971, 0.00000924, 0.00003739, -0.00000981, -0.00006343, -0.00002952, 0.00004195, 0.00003417, -0.00003467, -0.00004012, 0.00003010, 0.00005041, -0.00002296, -0.00006923, -0.00000824, 0.00005965, 0.00002364, -0.00005014, -0.00003596, 0.00003638, 0.00003911, -0.00003089, -0.00005140, 0.00001117, 0.00004785, -0.00000205, -0.00005128, -0.00001934, 0.00003547, 0.00002542, -0.00001979, -0.00001750, 0.00002190, 0.00002262, -0.00001907, -0.00003234, 0.00000277, 0.00002605, 0.00000460, -0.00001659, 0.00000059, 0.00002353, 0.00000995, -0.00002123, -0.00002573, -0.00000412, 0.00000898, 0.00000322, -0.00000148, 0.00000416, 0.00000763, 0.00000173, -0.00000508, -0.00000652, -0.00000618, -0.00000576, -0.00000193, 0.00000380, 0.00000492, 0.00000087, -0.00000189, -0.00000016, 0.00000317, 0.00000567, 0.00000707, 0.00000549, -0.00000029, -0.00000588, -0.00000573, -0.00000143, 0.00000071, -0.00000093, -0.00000274, -0.00000315, -0.00000349, -0.00000247, 0.00000262, 0.00000882, 0.00000879, 0.00000115, -0.00000616, -0.00000594, -0.00000077, 0.00000151, -0.00000156, -0.00000505, -0.00000434, -0.00000092, 0.00000105, },
		{ -0.00000160, 0.00000086, 0.00000093, 0.00000132, 0.00000308, 0.00000349, 0.00000128, -0.00000132, -0.00000338, -0.00000583, -0.00000554, 0.00000155, 0.00000839, 0.00000141, -0.00001553, -0.00001947, -0.00000421, 0.00000302, -0.00001566, -0.00003115, -0.00001074, 0.00001554, -0.00001555, -0.00010156, -0.00015759, -0.00011381, 0.00001706, 0.00018038, 0.00032306, 0.00037132, 0.00025159, 0.00000580, -0.00018147, -0.00017029, -0.00004196, 0.00001439, -0.00003942, -0.00007268, -0.00001957, 0.00002498, -0.00001039, -0.00004520, 0.00000022, 0.00005058, 0.00000566, -0.00007083, -0.00004876, 0.00004156, 0.00005324, -0.00002724, -0.00005657, 0.00001673, 0.00006230, -0.00000473, -0.00007370, -0.00002746, 0.00005674, 0.00004056, -0.00004155, -0.00004850, 0.00002467, 0.00004696, -0.00001800, -0.00005498, -0.00000145, 0.00004853, 0.00001142, -0.00004662, -0.00003031, 0.00002618, 0.00002970, -0.00001269, -0.00001996, 0.00001643, 0.00002602, -0.00001106, -0.00003258, -0.00000534, 0.00002247, 0.00000939, -0.00001305, -0.00000298, 0.00002017, 0.00001434, -0.00001399, -0.00002522, -0.00001035, 0.00000424, 0.00000402, 0.00000083, 0.00000388, 0.00000627, 0.00000249, -0.00000257, -0.00000506, -0.00000692, -0.00000753, -0.00000335, 0.00000315, 0.00000503, 0.00000156, -0.00000125, -0.00000038, 0.00000192, 0.00000433, 0.00000701, 0.00000718, 0.00000207, -0.00000460, -0.00000619, -0.00000264, 0.00000009, -0.00000066, -0.00000216, -0.00000273, -0.00000356, -0.00000356, 0.00000054, 0.00000719, 0.00000933, 0.00000356, -0.00000420, -0.00000597, -0.00000176, 0.00000135, -0.00000085, -0.00000476, -0.00000495, },
		{ -0.00000438, -0.00000537, -0.00000228, 0.00000051, 0.00000079, 0.00000109, 0.00000279, 0.00000326, 0.00000136, -0.00000046, -0.00000210, -0.00000588, -0.00000779, -0.00000021, 0.00001070, 0.00000627, -0.00001449, -0.00002453, -0.00000860, 0.00000633, -0.00000896, -0.00003228, -0.00002045, 0.00001273, 0.00000194, -0.00007452, -0.00015260, -0.00015306, -0.00005108, 0.00013173, 0.00033590, 0.00043982, 0.00032206, 0.00002643, -0.00020951, -0.00019998, -0.00003979, 0.00003174, -0.00004089, -0.00009194, -0.00002589, 0.00004039, -0.00000105, -0.00006026, -0.00001651, 0.00005980, 0.00002710, -0.00007158, -0.00007100, 0.00003328, 0.00007183, -0.00001233, -0.00006939, -0.00000237, 0.00006828, 0.00001651, -0.00007133, -0.00004671, 0.00004751, 0.00005502, -0.00002819, -0.00005709, 0.00001014, 0.00005042, -0.00000400, -0.00005435, -0.00001400, 0.00004509, 0.00002338, -0.00003842, -0.00003815, 0.00001538, 0.00003097, -0.00000555, -0.00002054, 0.00001069, 0.00002734, -0.00000313, -0.00003036, -0.00001214, 0.00001727, 0.00001245, -0.00000877, -0.00000507, 0.00001613, 0.00001667, -0.00000693, -0.00002241, -0.00001476, -0.00000121, 0.00000303, 0.00000266, 0.00000444, 0.00000529, 0.00000255, -0.00000065, -0.00000313, -0.00000672, -0.00000894, -0.00000515, 0.00000202, 0.00000493, 0.00000223, -0.00000052, -0.00000032, 0.00000101, 0.00000297, 0.00000636, 0.00000818, 0.00000429, -0.00000281, -0.00000604, -0.00000365, -0.00000073, -0.00000068, -0.00000172, -0.00000226, -0.00000334, -0.00000420, -0.00000133, 0.00000518, 0.00000910, 0.00000550, -0.00000203, -0.00000546, -0.00000252, 0.00000102, -0.00000030, },
		{ 0.00000063, 0.00000007, -0.00000397, -0.00000561, -0.00000289, 0.00000004, 0.00000051, 0.00000096, 0.00000276, 0.00000302, 0.00000094, -0.00000010, -0.00000048, -0.00000489, -0.00000966, -0.00000319, 0.00001148, 0.00001156, -0.00001106, -0.00002831, -0.00001499, 0.00000674, -0.00000156, -0.00002911, -0.00002841, 0.00000432, 0.00001147, -0.00004608, -0.00013263, -0.00017606, -0.00012035, 0.00006085, 0.00032436, 0.00050500, 0.00041190, 0.00006709, -0.00023670, -0.00024322, -0.00004512, 0.00005600, -0.00003374, -0.00011410, -0.00004121, 0.00005517, 0.00001685, -0.00007126, -0.00004022, 0.00006179, 0.00005226, -0.00006358, -0.00009309, 0.00001652, 0.00008666, 0.00000914, -0.00007589, -0.00002514, 0.00006676, 0.00003813, -0.00006180, -0.00006342, 0.00003274, 0.00006506, -0.00001156, -0.00006059, -0.00000548, 0.00004910, 0.00000950, -0.00004964, -0.00002503, 0.00003815, 0.00003263, -0.00002782, -0.00004228, 0.00000430, 0.00002932, 0.00000084, -0.00001942, 0.00000529, 0.00002674, 0.00000402, -0.00002613, -0.00001704, 0.00001116, 0.00001356, -0.00000440, -0.00000562, 0.00001208, 0.00001709, -0.00000079, -0.00001794, -0.00001698, -0.00000655, 0.00000046, 0.00000353, 0.00000547, 0.00000491, 0.00000225, 0.00000057, -0.00000111, -0.00000569, -0.00000973, -0.00000706, 0.00000039, 0.00000448, 0.00000280, 0.00000028, -0.00000003, 0.00000045, 0.00000173, 0.00000533, 0.00000849, 0.00000613, -0.00000076, -0.00000529, -0.00000429, -0.00000159, -0.00000094, -0.00000148, -0.00000182, -0.00000292, -0.00000441, -0.00000286, 0.00000300, 0.00000821, 0.00000684, 0.00000014, -0.00000450, -0.00000294, },
		{ -0.00000322, -0.00000300, 0.00000031, 0.00000028, -0.00000363, -0.00000571, -0.00000334, -0.00000047, 0.00000005, 0.00000078, 0.00000298, 0.00000302, 0.00000021, -0.00000044, 0.00000100, -0.00000294, -0.00001056, -0.00000680, 0.00001039, 0.00001625, -0.00000559, -0.00002971, -0.00002216, 0.00000368, 0.00000457, -0.00002223, -0.00003225, -0.00000696, 0.00001175, -0.00002166, -0.00010126, -0.00017890, -0.00018194, -0.00002784, 0.00028294, 0.00055698, 0.00051753, 0.00013364, -0.00025569, -0.00029978, -0.00006356, 0.00008414, -0.00001477, -0.00013520, -0.00006671, 0.00006516, 0.00004266, -0.00007447, -0.00006862, 0.00005402, 0.00007775, -0.00004595, -0.00011138, -0.00000786, 0.00009463, 0.00003500, -0.00007413, -0.00004866, 0.00005720, 0.00005723, -0.00004587, -0.00007531, 0.00001408, 0.00006930, 0.00000636, -0.00005857, -0.00002034, 0.00004328, 0.00002100, -0.00004158, -0.00003337, 0.00002876, 0.00003843, -0.00001615, -0.00004265, -0.00000589, 0.00002522, 0.00000585, -0.00001702, 0.00000067, 0.00002454, 0.00000982, -0.00002050, -0.00001976, 0.00000489, 0.00001280, -0.00000054, -0.00000483, 0.00000859, 0.00001600, 0.00000391, -0.00001259, -0.00001698, -0.00001101, -0.00000323, 0.00000313, 0.00000649, 0.00000514, 0.00000192, 0.00000115, 0.00000067, -0.00000405, -0.00000974, -0.00000878, -0.00000163, 0.00000355, 0.00000314, 0.00000110, 0.00000046, 0.00000022, 0.00000074, 0.00000413, 0.00000822, 0.00000745, 0.00000130, -0.00000407, -0.00000446, -0.00000232, -0.00000135, -0.00000147, -0.00000153, -0.00000239, -0.00000425, -0.00000397, 0.00000083, 0.00000677, 0.00000755, 0.00000213, },
		{ 0.00000762, 0.00000378, -0.00000179, -0.00000270, 0.00000015, 0.00000043, -0.00000338, -0.00000576, -0.00000361, -0.00000090, -0.00000058, 0.00000040, 0.00000333, 0.00000341, -0.00000051, -0.00000149, 0.00000188, -0.00000040, -0.00001014, -0.00001027, 0.00000751, 0.00001934, 0.00000103, -0.00002812, -0.00002852, -0.00000249, 0.00000758, -0.00001333, -0.00003056, -0.00001761, 0.00000349, -0.00000599, -0.00006414, -0.00016056, -0.00022707, -0.00012660, 0.00020925, 0.00058514, 0.00063170, 0.00022969, -0.00025753, -0.00036651, -0.00010022, 0.00011100, 0.00001771, -0.00015016, -0.00010173, 0.00006603, 0.00007391, -0.00006686, -0.00009796, 0.00003524, 0.00009942, -0.00001941, -0.00012227, -0.00003761, 0.00009340, 0.00006206, -0.00006335, -0.00006970, 0.00004032, 0.00007115, -0.00002530, -0.00008072, -0.00000624, 0.00006721, 0.00002341, -0.00005140, -0.00003271, 0.00003386, 0.00002929, -0.00003130, -0.00003830, 0.00001821, 0.00004055, -0.00000475, -0.00003966, -0.00001425, 0.00001939, 0.00000906, -0.00001385, -0.00000286, 0.00002123, 0.00001395, -0.00001421, -0.00002025, -0.00000080, 0.00001047, 0.00000230, -0.00000310, 0.00000605, 0.00001394, 0.00000695, -0.00000712, -0.00001504, -0.00001404, -0.00000740, 0.00000140, 0.00000706, 0.00000584, 0.00000182, 0.00000124, 0.00000201, -0.00000210, -0.00000895, -0.00001002, -0.00000386, 0.00000211, 0.00000313, 0.00000185, 0.00000110, 0.00000028, 0.00000002, 0.00000291, 0.00000752, 0.00000820, 0.00000315, -0.00000258, -0.00000418, -0.00000279, -0.00000178, -0.00000165, -0.00000142, -0.00000189, -0.00000378, -0.00000462, -0.00000117, 0.00000496, },
		{ -0.00000289, 0.00000293, 0.00000712, 0.00000501, -0.00000036, -0.00000213, 0.00000023, 0.00000060, -0.00000324, -0.00000584, -0.00000375, -0.00000115, -0.00000128, -0.00000028, 0.00000362, 0.00000423, -0.00000089, -0.00000304, 0.00000184, 0.00000219, -0.00000836, -0.00001279, 0.00000344, 0.00002019, 0.00000755, -0.00002362, -0.00003249, -0.00001055, 0.00000628, -0.00000485, -0.00002333, -0.00002411, -0.00001064, -0.00000219, -0.00002821, -0.00012330, -0.00024852, -0.00022510, 0.00010497, 0.00057977, 0.00074382, 0.00035539, -0.00023287, -0.00043698, -0.00015854, 0.00012966, 0.00006308, -0.00015356, -0.00014341, 0.00005415, 0.00010647, -0.00004680, -0.00012363, 0.00000599, 0.00011316, 0.00001368, -0.00012298, -0.00006924, 0.00008192, 0.00008666, -0.00004417, -0.00008517, 0.00001800, 0.00007793, -0.00000260, -0.00007894, -0.00002571, 0.00005917, 0.00003759, -0.00004015, -0.00004129, 0.00002220, 0.00003372, -0.00002019, -0.00003965, 0.00000783, 0.00003929, 0.00000524, -0.00003406, -0.00002015, 0.00001267, 0.00001038, -0.00001047, -0.00000521, 0.00001729, 0.00001629, -0.00000796, -0.00001877, -0.00000531, 0.00000709, 0.00000384, -0.00000094, 0.00000462, 0.00001149, 0.00000834, -0.00000223, -0.00001166, -0.00001530, -0.00001137, -0.00000145, 0.00000680, 0.00000670, 0.00000209, 0.00000110, 0.00000280, -0.00000014, -0.00000746, -0.00001058, -0.00000602, 0.00000022, 0.00000264, 0.00000241, 0.00000184, 0.00000061, -0.00000039, 0.00000180, 0.00000654, 0.00000844, 0.00000464, -0.00000102, -0.00000354, -0.00000291, -0.00000209, -0.00000193, -0.00000154, -0.00000149, -0.00000311, -0.00000484, },
		{ -0.00000236, -0.00000464, -0.00000425, 0.00000083, 0.00000616, 0.00000577, 0.00000094, -0.00000139, 0.00000053, 0.00000088, -0.00000315, -0.00000603, -0.00000381, -0.00000117, -0.00000190, -0.00000125, 0.00000364, 0.00000532, -0.00000071, -0.00000469, 0.00000081, 0.00000419, -0.00000559, -0.00001377, -0.00000092, 0.00001867, 0.00001261, -0.00001706, -0.00003293, -0.00001862, 0.00000045, 0.00000060, -0.00001200, -0.00002375, -0.00002654, -0.00001106, -0.00000041, -0.00007255, -0.00024186, -0.00031178, -0.00002381, 0.00053364, 0.00084098, 0.00050656, -0.00017347, -0.00050179, -0.00023912, 0.00013241, 0.00011797, -0.00014063, -0.00018688, 0.00002741, 0.00013504, -0.00001460, -0.00014084, -0.00003129, 0.00011556, 0.00004952, -0.00011204, -0.00009864, 0.00006066, 0.00010526, -0.00001857, -0.00009272, -0.00000700, 0.00007662, 0.00001939, -0.00007032, -0.00004200, 0.00004641, 0.00004737, -0.00002643, -0.00004539, 0.00000988, 0.00003420, -0.00000960, -0.00003773, -0.00000125, 0.00003532, 0.00001303, -0.00002684, -0.00002336, 0.00000595, 0.00000994, -0.00000738, -0.00000643, 0.00001323, 0.00001692, -0.00000237, -0.00001575, -0.00000827, 0.00000325, 0.00000402, 0.00000116, 0.00000424, 0.00000918, 0.00000837, 0.00000160, -0.00000749, -0.00001477, -0.00001455, -0.00000503, 0.00000551, 0.00000739, 0.00000274, 0.00000095, 0.00000309, 0.00000157, -0.00000548, -0.00001035, -0.00000784, -0.00000200, 0.00000162, 0.00000267, 0.00000258, 0.00000116, -0.00000050, 0.00000088, 0.00000544, 0.00000827, 0.00000571, 0.00000042, -0.00000268, -0.00000270, -0.00000215, -0.00000218, -0.00000184, -0.00000130, },
		{ -0.00000226, -0.00000135, -0.00000165, -0.00000412, -0.00000520, -0.00000120, 0.00000484, 0.00000608, 0.00000201, -0.00000061, 0.00000103, 0.00000134, -0.00000303, -0.00000634, -0.00000393, -0.00000098, -0.00000229, -0.00000240, 0.00000323, 0.00000644, 0.00000012, -0.00000602, -0.00000101, 0.00000512, -0.00000248, -0.00001301, -0.00000448, 0.00001529, 0.00001513, -0.00000989, -0.00002945, -0.00002453, -0.00000892, 0.00000082, 0.00000070, -0.00001527, -0.00003939, -0.00003083, 0.00001347, -0.00001618, -0.00020642, -0.00037543, -0.00016695, 0.00044357, 0.00090944, 0.00067450, -0.00007383, -0.00054949, -0.00033891, 0.00011200, 0.00017622, -0.00010843, -0.00022572, -0.00001402, 0.00015403, 0.00002722, -0.00014550, -0.00007250, 0.00010469, 0.00008348, -0.00008975, -0.00012170, 0.00003168, 0.00011503, 0.00001039, -0.00009115, -0.00003146, 0.00006752, 0.00003797, -0.00005619, -0.00005331, 0.00003076, 0.00005194, -0.00001209, -0.00004498, -0.00000158, 0.00003122, -0.00000068, -0.00003331, -0.00000824, 0.00002958, 0.00001821, -0.00001902, -0.00002398, -0.00000004, 0.00000811, -0.00000495, -0.00000674, 0.00000945, 0.00001613, 0.00000208, -0.00001179, -0.00000953, -0.00000042, 0.00000300, 0.00000274, 0.00000468, 0.00000740, 0.00000746, 0.00000411, -0.00000320, -0.00001267, -0.00001645, -0.00000883, 0.00000321, 0.00000756, 0.00000362, 0.00000098, 0.00000301, 0.00000284, -0.00000329, -0.00000933, -0.00000907, -0.00000430, 0.00000006, 0.00000247, 0.00000323, 0.00000188, -0.00000033, 0.00000018, 0.00000434, 0.00000781, 0.00000640, 0.00000162, -0.00000181, -0.00000225, -0.00000190, -0.00000228, },
		{ -0.00000136, -0.00000209, -0.00000265, -0.00000163, -0.00000107, -0.00000336, -0.00000572, -0.00000308, 0.00000326, 0.00000598, 0.00000282, 0.00000011, 0.00000163, 0.00000200, -0.00000279, -0.00000673, -0.00000419, -0.00000066, -0.00000236, -0.00000353, 0.00000235, 0.00000729, 0.00000145, -0.00000666, -0.00000316, 0.00000477, 0.00000017, -0.00001079, -0.00000633, 0.00001110, 0.00001461, -0.00000388, -0.00002263, -0.00002634, -0.00001978, -0.00000536, 0.00001130, 0.00000071, -0.00004469, -0.00005740, 0.00000992, 0.00003665, -0.00014565, -0.00040691, -0.00031125, 0.00031134, 0.00093653, 0.00084637, 0.00006742, -0.00056794, -0.00045086, 0.00006304, 0.00022947, -0.00005667, -0.00025297, -0.00006744, 0.00015850, 0.00007414, -0.00013508, -0.00011242, 0.00008053, 0.00011079, -0.00005819, -0.00013505, -0.00000168, 0.00011439, 0.00003908, -0.00008061, -0.00005225, 0.00005203, 0.00005102, -0.00003864, -0.00005867, 0.00001438, 0.00005130, 0.00000111, -0.00004069, -0.00001092, 0.00002567, 0.00000579, -0.00002737, -0.00001276, 0.00002309, 0.00002080, -0.00001152, -0.00002245, -0.00000475, 0.00000538, -0.00000340, -0.00000642, 0.00000624, 0.00001430, 0.00000515, -0.00000751, -0.00000919, -0.00000341, 0.00000111, 0.00000353, 0.00000557, 0.00000635, 0.00000609, 0.00000530, 0.00000065, -0.00000944, -0.00001685, -0.00001228, 0.00000008, 0.00000696, 0.00000450, 0.00000130, 0.00000275, 0.00000361, -0.00000116, -0.00000768, -0.00000956, -0.00000643, -0.00000193, 0.00000173, 0.00000363, 0.00000271, 0.00000012, -0.00000027, 0.00000330, 0.00000719, 0.00000678, 0.00000254, -0.00000106, -0.00000170, },
		{ -0.00000056, -0.00000123, -0.00000063, -0.00000155, -0.00000288, -0.00000208, -0.00000070, -0.00000247, -0.00000584, -0.00000470, 0.00000154, 0.00000553, 0.00000335, 0.00000071, 0.00000225, 0.00000282, -0.00000236, -0.00000713, -0.00000464, -0.00000036, -0.00000210, -0.00000443, 0.00000114, 0.00000760, 0.00000296, -0.00000643, -0.00000504, 0.00000331, 0.00000160, -0.00000787, -0.00000597, 0.00000746, 0.00001131, -0.00000073, -0.00001399, -0.00002288, -0.00002938, -0.00001772, 0.00001632, 0.00002173, -0.00003917, -0.00008497, -0.00001153, 0.00007699, -0.00006683, -0.00040069, -0.00044208, 0.00014411, 0.00091262, 0.00100644, 0.00024650, -0.00054617, -0.00056429, -0.00001662, 0.00026819, 0.00001167, -0.00026226, -0.00012762, 0.00014517, 0.00012023, -0.00010919, -0.00014550, 0.00004516, 0.00012739, -0.00002100, -0.00013663, -0.00003532, 0.00010332, 0.00006386, -0.00006260, -0.00006686, 0.00003249, 0.00005732, -0.00002011, -0.00005802, -0.00000067, 0.00004619, 0.00001172, -0.00003362, -0.00001734, 0.00001872, 0.00000953, -0.00002096, -0.00001486, 0.00001678, 0.00002114, -0.00000506, -0.00001936, -0.00000788, 0.00000228, -0.00000276, -0.00000578, 0.00000374, 0.00001185, 0.00000678, -0.00000350, -0.00000756, -0.00000534, -0.00000118, 0.00000343, 0.00000653, 0.00000606, 0.00000469, 0.00000537, 0.00000362, -0.00000561, -0.00001575, -0.00001489, -0.00000352, 0.00000546, 0.00000511, 0.00000187, 0.00000248, 0.00000392, 0.00000068, -0.00000559, -0.00000921, -0.00000811, -0.00000418, 0.00000040, 0.00000363, 0.00000354, 0.00000082, -0.00000047, 0.00000235, 0.00000649, 0.00000698, 0.00000320, },
		{ 0.00000707, 0.00000372, -0.00000031, -0.00000101, 0.00000013, -0.00000067, -0.00000281, -0.00000259, -0.00000057, -0.00000156, -0.00000560, -0.00000602, -0.00000025, 0.00000477, 0.00000361, 0.00000116, 0.00000279, 0.00000371, -0.00000170, -0.00000742, -0.00000527, -0.00000024, -0.00000162, -0.00000493, -0.00000017, 0.00000729, 0.00000422, -0.00000544, -0.00000609, 0.00000125, 0.00000137, -0.00000528, -0.00000347, 0.00000577, 0.00000628, -0.00000160, -0.00000573, -0.00001406, -0.00003480, -0.00003452, 0.00001303, 0.00004381, -0.00002157, -0.00010711, -0.00004818, 0.00009752, 0.00001988, -0.00035585, -0.00054532, -0.00004608, 0.00083277, 0.00113788, 0.00045422, -0.00047613, -0.00066586, -0.00012499, 0.00028307, 0.00009036, -0.00024909, -0.00018751, 0.00011322, 0.00015902, -0.00006982, -0.00016684, 0.00000249, 0.00013061, 0.00001718, -0.00012614, -0.00006501, 0.00008333, 0.00008173, -0.00003970, -0.00007379, 0.00001169, 0.00005675, -0.00000303, -0.00005219, -0.00001270, 0.00003793, 0.00001887, -0.00002514, -0.00002060, 0.00001154, 0.00001067, -0.00001503, -0.00001491, 0.00001134, 0.00001981, -0.00000006, -0.00001540, -0.00000940, -0.00000068, -0.00000294, -0.00000512, 0.00000197, 0.00000921, 0.00000710, -0.00000021, -0.00000511, -0.00000608, -0.00000341, 0.00000251, 0.00000722, 0.00000636, 0.00000361, 0.00000463, 0.00000549, -0.00000175, -0.00001336, -0.00001631, -0.00000713, 0.00000313, 0.00000518, 0.00000256, 0.00000236, 0.00000388, 0.00000208, -0.00000335, -0.00000810, -0.00000912, -0.00000641, -0.00000147, 0.00000310, 0.00000424, 0.00000174, -0.00000038, 0.00000153, 0.00000575, },
		{ 0.00000084, 0.00000498, 0.00000714, 0.00000420, -0.00000026, -0.00000114, 0.00000072, 0.00000046, -0.00000235, -0.00000302, -0.00000067, -0.00000074, -0.00000507, -0.00000699, -0.00000203, 0.00000375, 0.00000364, 0.00000147, 0.00000321, 0.00000458, -0.00000085, -0.00000749, -0.00000597, -0.00000038, -0.00000110, -0.00000498, -0.00000125, 0.00000645, 0.00000484, -0.00000409, -0.00000596, -0.00000060, -0.00000050, -0.00000414, 0.00000041, 0.00000707, 0.00000121, -0.00000679, -0.00000030, -0.00000108, -0.00003361, -0.00005270, 0.00000015, 0.00006220, 0.00000699, -0.00011795, -0.00009444, 0.00009385, 0.00010288, -0.00027643, -0.00060934, -0.00024355, 0.00069789, 0.00122487, 0.00067654, -0.00035443, -0.00074117, -0.00025551, 0.00026668, 0.00017044, -0.00021179, -0.00023916, 0.00006463, 0.00018455, -0.00002123, -0.00017305, -0.00004224, 0.00011975, 0.00005145, -0.00010508, -0.00008717, 0.00005727, 0.00009080, -0.00001513, -0.00007278, -0.00000754, 0.00005020, 0.00001064, -0.00004270, -0.00002069, 0.00002813, 0.00002232, -0.00001662, -0.00002096, 0.00000514, 0.00000973, -0.00001023, -0.00001353, 0.00000718, 0.00001748, 0.00000336, -0.00001122, -0.00000948, -0.00000312, -0.00000371, -0.00000465, 0.00000084, 0.00000672, 0.00000638, 0.00000206, -0.00000234, -0.00000566, -0.00000518, 0.00000102, 0.00000741, 0.00000703, 0.00000304, 0.00000347, 0.00000624, 0.00000167, -0.00001006, -0.00001638, -0.00001030, 0.00000017, 0.00000455, 0.00000319, 0.00000245, 0.00000365, 0.00000299, -0.00000121, -0.00000638, -0.00000930, -0.00000835, -0.00000374, 0.00000193, 0.00000462, 0.00000281, 0.00000000, },
		{ 0.00000391, 0.00000070, 0.00000034, 0.00000417, 0.00000717, 0.00000477, -0.00000027, -0.00000164, 0.00000096, 0.00000166, -0.00000148, -0.00000321, -0.00000092, -0.00000007, -0.00000432, -0.00000758, -0.00000370, 0.00000250, 0.00000342, 0.00000168, 0.00000353, 0.00000534, 0.00000008, -0.00000728, -0.00000659, -0.00000081, -0.00000078, -0.00000467, -0.00000180, 0.00000541, 0.00000456, -0.00000297, -0.00000465, -0.00000139, -0.00000349, -0.00000533, 0.00000438, 0.00001174, -0.00000192, -0.00001556, 0.00000014, 0.00001371, -0.00002447, -0.00006843, -0.00002177, 0.00007226, 0.00004306, -0.00011326, -0.00014278, 0.00006531, 0.00017085, -0.00017104, -0.00062671, -0.00043083, 0.00051509, 0.00125488, 0.00089584, -0.00018335, -0.00077672, -0.00039735, 0.00021488, 0.00024139, -0.00015223, -0.00027502, 0.00000414, 0.00019250, 0.00003071, -0.00016285, -0.00008335, 0.00009620, 0.00007748, -0.00007652, -0.00009940, 0.00002878, 0.00009058, 0.00000784, -0.00006476, -0.00002288, 0.00003940, 0.00001964, -0.00003142, -0.00002434, 0.00001840, 0.00002241, -0.00000920, -0.00001906, 0.00000022, 0.00000743, -0.00000693, -0.00001141, 0.00000441, 0.00001477, 0.00000532, -0.00000733, -0.00000845, -0.00000478, -0.00000481, -0.00000449, 0.00000019, 0.00000462, 0.00000499, 0.00000322, 0.00000027, -0.00000432, -0.00000620, -0.00000072, 0.00000700, 0.00000777, 0.00000304, 0.00000227, 0.00000604, 0.00000427, -0.00000632, -0.00001514, -0.00001262, -0.00000309, 0.00000315, 0.00000351, 0.00000270, 0.00000340, 0.00000344, 0.00000061, -0.00000430, -0.00000864, -0.00000972, -0.00000619, 0.00000011, 0.00000452, },
		{ -0.00000229, 0.00000376, 0.00000486, 0.00000170, 0.00000009, 0.00000331, 0.00000713, 0.00000547, -0.00000018, -0.00000241, 0.00000069, 0.00000270, -0.00000025, -0.00000303, -0.00000120, 0.00000040, -0.00000347, -0.00000779, -0.00000518, 0.00000106, 0.00000293, 0.00000179, 0.00000380, 0.00000594, 0.00000095, -0.00000683, -0.00000694, -0.00000140, -0.00000086, -0.00000426, -0.00000167, 0.00000462, 0.00000341, -0.00000274, -0.00000256, -0.00000041, -0.00000661, -0.00000920, 0.00000686, 0.00001929, -0.00000125, -0.00002615, -0.00000578, 0.00002719, -0.00000752, -0.00007781, -0.00005025, 0.00007043, 0.00008139, -0.00009137, -0.00018495, 0.00001508, 0.00021447, -0.00005171, -0.00059536, -0.00059087, 0.00029716, 0.00122048, 0.00109294, 0.00002876, -0.00076196, -0.00053652, 0.00012796, 0.00029262, -0.00007584, -0.00028927, -0.00006136, 0.00018092, 0.00007934, -0.00013743, -0.00011561, 0.00006325, 0.00009224, -0.00004459, -0.00010089, 0.00000169, 0.00008203, 0.00002641, -0.00005170, -0.00003283, 0.00002652, 0.00002364, -0.00002025, -0.00002409, 0.00001009, 0.00001999, -0.00000363, -0.00001580, -0.00000293, 0.00000455, -0.00000515, -0.00000920, 0.00000287, 0.00001221, 0.00000610, -0.00000408, -0.00000669, -0.00000558, -0.00000595, -0.00000467, -0.00000018, 0.00000303, 0.00000329, 0.00000334, 0.00000232, -0.00000243, -0.00000636, -0.00000237, 0.00000607, 0.00000835, 0.00000350, 0.00000131, 0.00000515, 0.00000590, -0.00000263, -0.00001280, -0.00001384, -0.00000626, 0.00000108, 0.00000335, 0.00000302, 0.00000323, 0.00000355, 0.00000197, -0.00000212, -0.00000722, -0.00001029, -0.00000851, },
		{ -0.00000995, -0.00001039, -0.00000508, 0.00000225, 0.00000545, 0.00000293, 0.00000015, 0.00000244, 0.00000695, 0.00000629, 0.00000018, -0.00000329, -0.00000011, 0.00000335, 0.00000115, -0.00000241, -0.00000136, 0.00000069, -0.00000261, -0.00000765, -0.00000639, -0.00000052, 0.00000213, 0.00000178, 0.00000411, 0.00000642, 0.00000163, -0.00000627, -0.00000691, -0.00000194, -0.00000145, -0.00000414, -0.00000089, 0.00000460, 0.00000173, -0.00000393, -0.00000047, 0.00000265, -0.00000858, -0.00001546, 0.00000631, 0.00002840, 0.00000446, -0.00003605, -0.00001822, 0.00003603, 0.00001546, -0.00007766, -0.00008124, 0.00005497, 0.00011581, -0.00005351, -0.00021339, -0.00005025, 0.00022783, 0.00006777, -0.00051892, -0.00070927, 0.00006121, 0.00112066, 0.00124945, 0.00026806, -0.00069105, -0.00065749, 0.00001103, 0.00031511, 0.00000893, -0.00027885, -0.00012384, 0.00015076, 0.00011838, -0.00010026, -0.00013515, 0.00002556, 0.00009446, -0.00001378, -0.00009253, -0.00002064, 0.00006729, 0.00003872, -0.00003612, -0.00003692, 0.00001379, 0.00002311, -0.00001078, -0.00002093, 0.00000405, 0.00001612, -0.00000020, -0.00001209, -0.00000434, 0.00000180, -0.00000464, -0.00000737, 0.00000223, 0.00001009, 0.00000608, -0.00000161, -0.00000459, -0.00000555, -0.00000690, -0.00000516, -0.00000045, 0.00000196, 0.00000163, 0.00000263, 0.00000357, -0.00000038, -0.00000572, -0.00000365, 0.00000476, 0.00000859, 0.00000427, 0.00000076, 0.00000389, 0.00000654, 0.00000061, -0.00000971, -0.00001384, -0.00000894, -0.00000147, 0.00000257, 0.00000322, 0.00000320, 0.00000347, 0.00000282, -0.00000013, -0.00000524, },
		{ 0.00000147, -0.00000299, -0.00000870, -0.00001152, -0.00000798, -0.00000002, 0.00000548, 0.00000426, 0.00000058, 0.00000163, 0.00000656, 0.00000715, 0.00000089, -0.00000406, -0.00000138, 0.00000340, 0.00000250, -0.00000138, -0.00000124, 0.00000086, -0.00000184, -0.00000721, -0.00000721, -0.00000212, 0.00000096, 0.00000159, 0.00000452, 0.00000689, 0.00000202, -0.00000581, -0.00000647, -0.00000212, -0.00000244, -0.00000467, 0.00000023, 0.00000577, 0.00000015, -0.00000678, 0.00000062, 0.00000757, -0.00000814, -0.00002308, 0.00000162, 0.00003708, 0.00001552, -0.00004251, -0.00003597, 0.00003736, 0.00004130, -0.00006620, -0.00010976, 0.00002643, 0.00014034, -0.00000371, -0.00022249, -0.00012173, 0.00020929, 0.00017379, -0.00040628, -0.00077615, -0.00017341, 0.00096120, 0.00135009, 0.00051656, -0.00056406, -0.00074525, -0.00012616, 0.00030304, 0.00009170, -0.00024423, -0.00017541, 0.00010573, 0.00014291, -0.00005653, -0.00014008, -0.00001167, 0.00008484, 0.00001187, -0.00007665, -0.00003587, 0.00004928, 0.00004408, -0.00002065, -0.00003567, 0.00000309, 0.00001923, -0.00000401, -0.00001616, 0.00000059, 0.00001191, 0.00000123, -0.00000871, -0.00000439, -0.00000035, -0.00000498, -0.00000619, 0.00000211, 0.00000855, 0.00000563, 0.00000007, -0.00000248, -0.00000483, -0.00000748, -0.00000583, -0.00000077, 0.00000134, 0.00000024, 0.00000140, 0.00000394, 0.00000145, -0.00000448, -0.00000436, 0.00000332, 0.00000843, 0.00000514, 0.00000067, 0.00000256, 0.00000632, 0.00000311, -0.00000629, -0.00001271, -0.00001082, -0.00000417, 0.00000117, 0.00000311, 0.00000327, 0.00000335, 0.00000324, },
		{ 0.00000330, 0.00000334, 0.00000256, -0.00000075, -0.00000669, -0.00001170, -0.00001065, -0.00000295, 0.00000473, 0.00000547, 0.00000141, 0.00000100, 0.00000594, 0.00000791, 0.00000197, -0.00000447, -0.00000293, 0.00000277, 0.00000352, -0.00000003, -0.00000073, 0.00000103, -0.00000122, -0.00000656, -0.00000759, -0.00000362, -0.00000058, 0.00000108, 0.00000505, 0.00000754, 0.00000213, -0.00000571, -0.00000573, -0.00000166, -0.00000357, -0.00000617, 0.00000116, 0.00000827, -0.00000050, -0.00001108, -0.00000030, 0.00001353, -0.00000437, -0.00003045, -0.00000756, 0.00004303, 0.00003110, -0.00004301, -0.00005657, 0.00002940, 0.00006587, -0.00004351, -0.00013077, -0.00001233, 0.00015024, 0.00005182, -0.00020960, -0.00018947, 0.00016166, 0.00025484, -0.00027033, -0.00078743, -0.00038742, 0.00075418, 0.00138482, 0.00075418, -0.00038742, -0.00078743, -0.00027033, 0.00025484, 0.00016166, -0.00018947, -0.00020960, 0.00005182, 0.00015024, -0.00001233, -0.00013077, -0.00004351, 0.00006587, 0.00002940, -0.00005657, -0.00004301, 0.00003110, 0.00004303, -0.00000756, -0.00003045, -0.00000437, 0.00001353, -0.00000030, -0.00001108, -0.00000050, 0.00000827, 0.00000116, -0.00000617, -0.00000357, -0.00000166, -0.00000573, -0.00000571, 0.00000213, 0.00000754, 0.00000505, 0.00000108, -0.00000058, -0.00000362, -0.00000759, -0.00000656, -0.00000122, 0.00000103, -0.00000073, -0.00000003, 0.00000352, 0.00000277, -0.00000293, -0.00000447, 0.00000197, 0.00000791, 0.00000594, 0.00000100, 0.00000141, 0.00000547, 0.00000473, -0.00000295, -0.00001065, -0.00001170, -0.00000669, -0.00000075, 0.00000256, 0.00000334, },
		{ 0.00000147, 0.00000324, 0.00000335, 0.00000327, 0.00000311, 0.00000117, -0.00000417, -0.00001082, -0.00001271, -0.00000629, 0.00000311, 0.00000632, 0.00000256, 0.00000067, 0.00000514, 0.00000843, 0.00000332, -0.00000436, -0.00000448, 0.00000145, 0.00000394, 0.00000140, 0.00000024, 0.00000134, -0.00000077, -0.00000583, -0.00000748, -0.00000483, -0.00000248, 0.00000007, 0.00000563, 0.00000855, 0.00000211, -0.00000619, -0.00000498, -0.00000035, -0.00000439, -0.00000871, 0.00000123, 0.00001191, 0.00000059, -0.00001616, -0.00000401, 0.00001923, 0.00000309, -0.00003567, -0.00002065, 0.00004408, 0.00004928, -0.00003587, -0.00007665, 0.00001187, 0.00008484, -0.00001167, -0.00014008, -0.00005653, 0.00014291, 0.00010573, -0.00017541, -0.00024423, 0.00009170, 0.00030304, -0.00012616, -0.00074525, -0.00056406, 0.00051656, 0.00135009, 0.00096120, -0.00017341, -0.00077615, -0.00040628, 0.00017379, 0.00020929, -0.00012173, -0.00022249, -0.00000371, 0.00014034, 0.00002643, -0.00010976, -0.00006620, 0.00004130, 0.00003736, -0.00003597, -0.00004251, 0.00001552, 0.00003708, 0.00000162, -0.00002308, -0.00000814, 0.00000757, 0.00000062, -0.00000678, 0.00000015, 0.00000577, 0.00000023, -0.00000467, -0.00000244, -0.00000212, -0.00000647, -0.00000581, 0.00000202, 0.00000689, 0.00000452, 0.00000159, 0.00000096, -0.00000212, -0.00000721, -0.00000721, -0.00000184, 0.00000086, -0.00000124, -0.00000138, 0.00000250, 0.00000340, -0.00000138, -0.00000406, 0.00000089, 0.00000715, 0.00000656, 0.00000163, 0.00000058, 0.00000426, 0.00000548, -0.00000002, -0.00000798, -0.00001152, -0.00000870, -0.00000299, },
		{ -0.00000995, -0.00000524, -0.00000013, 0.00000282, 0.00000347, 0.00000320, 0.00000322, 0.00000257, -0.00000147, -0.00000894, -0.00001384, -0.00000971, 0.00000061, 0.00000654, 0.00000389, 0.00000076, 0.00000427, 0.00000859, 0.00000476, -0.00000365, -0.00000572, -0.00000038, 0.00000357, 0.00000263, 0.00000163, 0.00000196, -0.00000045, -0.00000516, -0.00000690, -0.00000555, -0.00000459, -0.00000161, 0.00000608, 0.00001009, 0.00000223, -0.00000737, -0.00000464, 0.00000180, -0.00000434, -0.00001209, -0.00000020, 0.00001612, 0.00000405, -0.00002093, -0.00001078, 0.00002311, 0.00001379, -0.00003692, -0.00003612, 0.00003872, 0.00006729, -0.00002064, -0.00009253, -0.00001378, 0.00009446, 0.00002556, -0.00013515, -0.00010026, 0.00011838, 0.00015076, -0.00012384, -0.00027885, 0.00000893, 0.00031511, 0.00001103, -0.00065749, -0.00069105, 0.00026806, 0.00124945, 0.00112066, 0.00006121, -0.00070927, -0.00051892, 0.00006777, 0.00022783, -0.00005025, -0.00021339, -0.00005351, 0.00011581, 0.00005497, -0.00008124, -0.00007766, 0.00001546, 0.00003603, -0.00001822, -0.00003605, 0.00000446, 0.00002840, 0.00000631, -0.00001546, -0.00000858, 0.00000265, -0.00000047, -0.00000393, 0.00000173, 0.00000460, -0.00000089, -0.00000414, -0.00000145, -0.00000194, -0.00000691, -0.00000627, 0.00000163, 0.00000642, 0.00000411, 0.00000178, 0.00000213, -0.00000052, -0.00000639, -0.00000765, -0.00000261, 0.00000069, -0.00000136, -0.00000241, 0.00000115, 0.00000335, -0.00000011, -0.00000329, 0.00000018, 0.00000629, 0.00000695, 0.00000244, 0.00000015, 0.00000293, 0.00000545, 0.00000225, -0.00000508, -0.00001039, },
		{ -0.00000229, -0.00000851, -0.00001029, -0.00000722, -0.00000212, 0.00000197, 0.00000355, 0.00000323, 0.00000302, 0.00000335, 0.00000108, -0.00000626, -0.00001384, -0.00001280, -0.00000263, 0.00000590, 0.00000515, 0.00000131, 0.00000350, 0.00000835, 0.00000607, -0.00000237, -0.00000636, -0.00000243, 0.00000232, 0.00000334, 0.00000329, 0.00000303, -0.00000018, -0.00000467, -0.00000595, -0.00000558, -0.00000669, -0.00000408, 0.00000610, 0.00001221, 0.00000287, -0.00000920, -0.00000515, 0.00000455, -0.00000293, -0.00001580, -0.00000363, 0.00001999, 0.00001009, -0.00002409, -0.00002025, 0.00002364, 0.00002652, -0.00003283, -0.00005170, 0.00002641, 0.00008203, 0.00000169, -0.00010089, -0.00004459, 0.00009224, 0.00006325, -0.00011561, -0.00013743, 0.00007934, 0.00018092, -0.00006136, -0.00028927, -0.00007584, 0.00029262, 0.00012796, -0.00053652, -0.00076196, 0.00002876, 0.00109294, 0.00122048, 0.00029716, -0.00059087, -0.00059536, -0.00005171, 0.00021447, 0.00001508, -0.00018495, -0.00009137, 0.00008139, 0.00007043, -0.00005025, -0.00007781, -0.00000752, 0.00002719, -0.00000578, -0.00002615, -0.00000125, 0.00001929, 0.00000686, -0.00000920, -0.00000661, -0.00000041, -0.00000256, -0.00000274, 0.00000341, 0.00000462, -0.00000167, -0.00000426, -0.00000086, -0.00000140, -0.00000694, -0.00000683, 0.00000095, 0.00000594, 0.00000380, 0.00000179, 0.00000293, 0.00000106, -0.00000518, -0.00000779, -0.00000347, 0.00000040, -0.00000120, -0.00000303, -0.00000025, 0.00000270, 0.00000069, -0.00000241, -0.00000018, 0.00000547, 0.00000713, 0.00000331, 0.00000009, 0.00000170, 0.00000486, 0.00000376, },
		{ 0.00000391, 0.00000452, 0.00000011, -0.00000619, -0.00000972, -0.00000864, -0.00000430, 0.00000061, 0.00000344, 0.00000340, 0.00000270, 0.00000351, 0.00000315, -0.00000309, -0.00001262, -0.00001514, -0.00000632, 0.00000427, 0.00000604, 0.00000227, 0.00000304, 0.00000777, 0.00000700, -0.00000072, -0.00000620, -0.00000432, 0.00000027, 0.00000322, 0.00000499, 0.00000462, 0.00000019, -0.00000449, -0.00000481, -0.00000478, -0.00000845, -0.00000733, 0.00000532, 0.00001477, 0.00000441, -0.00001141, -0.00000693, 0.00000743, 0.00000022, -0.00001906, -0.00000920, 0.00002241, 0.00001840, -0.00002434, -0.00003142, 0.00001964, 0.00003940, -0.00002288, -0.00006476, 0.00000784, 0.00009058, 0.00002878, -0.00009940, -0.00007652, 0.00007748, 0.00009620, -0.00008335, -0.00016285, 0.00003071, 0.00019250, 0.00000414, -0.00027502, -0.00015223, 0.00024139, 0.00021488, -0.00039735, -0.00077672, -0.00018335, 0.00089584, 0.00125488, 0.00051509, -0.00043083, -0.00062671, -0.00017104, 0.00017085, 0.00006531, -0.00014278, -0.00011326, 0.00004306, 0.00007226, -0.00002177, -0.00006843, -0.00002447, 0.00001371, 0.00000014, -0.00001556, -0.00000192, 0.00001174, 0.00000438, -0.00000533, -0.00000349, -0.00000139, -0.00000465, -0.00000297, 0.00000456, 0.00000541, -0.00000180, -0.00000467, -0.00000078, -0.00000081, -0.00000659, -0.00000728, 0.00000008, 0.00000534, 0.00000353, 0.00000168, 0.00000342, 0.00000250, -0.00000370, -0.00000758, -0.00000432, -0.00000007, -0.00000092, -0.00000321, -0.00000148, 0.00000166, 0.00000096, -0.00000164, -0.00000027, 0.00000477, 0.00000717, 0.00000417, 0.00000034, 0.00000070, },
		{ 0.00000084, 0.00000000, 0.00000281, 0.00000462, 0.00000193, -0.00000374, -0.00000835, -0.00000930, -0.00000638, -0.00000121, 0.00000299, 0.00000365, 0.00000245, 0.00000319, 0.00000455, 0.00000017, -0.00001030, -0.00001638, -0.00001006, 0.00000167, 0.00000624, 0.00000347, 0.00000304, 0.00000703, 0.00000741, 0.00000102, -0.00000518, -0.00000566, -0.00000234, 0.00000206, 0.00000638, 0.00000672, 0.00000084, -0.00000465, -0.00000371, -0.00000312, -0.00000948, -0.00001122, 0.00000336, 0.00001748, 0.00000718, -0.00001353, -0.00001023, 0.00000973, 0.00000514, -0.00002096, -0.00001662, 0.00002232, 0.00002813, -0.00002069, -0.00004270, 0.00001064, 0.00005020, -0.00000754, -0.00007278, -0.00001513, 0.00009080, 0.00005727, -0.00008717, -0.00010508, 0.00005145, 0.00011975, -0.00004224, -0.00017305, -0.00002123, 0.00018455, 0.00006463, -0.00023916, -0.00021179, 0.00017044, 0.00026668, -0.00025551, -0.00074117, -0.00035443, 0.00067654, 0.00122487, 0.00069789, -0.00024355, -0.00060934, -0.00027643, 0.00010288, 0.00009385, -0.00009444, -0.00011795, 0.00000699, 0.00006220, 0.00000015, -0.00005270, -0.00003361, -0.00000108, -0.00000030, -0.00000679, 0.00000121, 0.00000707, 0.00000041, -0.00000414, -0.00000050, -0.00000060, -0.00000596, -0.00000409, 0.00000484, 0.00000645, -0.00000125, -0.00000498, -0.00000110, -0.00000038, -0.00000597, -0.00000749, -0.00000085, 0.00000458, 0.00000321, 0.00000147, 0.00000364, 0.00000375, -0.00000203, -0.00000699, -0.00000507, -0.00000074, -0.00000067, -0.00000302, -0.00000235, 0.00000046, 0.00000072, -0.00000114, -0.00000026, 0.00000420, 0.00000714, 0.00000498, },
		{ 0.00000707, 0.00000575, 0.00000153, -0.00000038, 0.00000174, 0.00000424, 0.00000310, -0.00000147, -0.00000641, -0.00000912, -0.00000810, -0.00000335, 0.00000208, 0.00000388, 0.00000236, 0.00000256, 0.00000518, 0.00000313, -0.00000713, -0.00001631, -0.00001336, -0.00000175, 0.00000549, 0.00000463, 0.00000361, 0.00000636, 0.00000722, 0.00000251, -0.00000341, -0.00000608, -0.00000511, -0.00000021, 0.00000710, 0.00000921, 0.00000197, -0.00000512, -0.00000294, -0.00000068, -0.00000940, -0.00001540, -0.00000006, 0.00001981, 0.00001134, -0.00001491, -0.00001503, 0.00001067, 0.00001154, -0.00002060, -0.00002514, 0.00001887, 0.00003793, -0.00001270, -0.00005219, -0.00000303, 0.00005675, 0.00001169, -0.00007379, -0.00003970, 0.00008173, 0.00008333, -0.00006501, -0.00012614, 0.00001718, 0.00013061, 0.00000249, -0.00016684, -0.00006982, 0.00015902, 0.00011322, -0.00018751, -0.00024909, 0.00009036, 0.00028307, -0.00012499, -0.00066586, -0.00047613, 0.00045422, 0.00113788, 0.00083277, -0.00004608, -0.00054532, -0.00035585, 0.00001988, 0.00009752, -0.00004818, -0.00010711, -0.00002157, 0.00004381, 0.00001303, -0.00003452, -0.00003480, -0.00001406, -0.00000573, -0.00000160, 0.00000628, 0.00000577, -0.00000347, -0.00000528, 0.00000137, 0.00000125, -0.00000609, -0.00000544, 0.00000422, 0.00000729, -0.00000017, -0.00000493, -0.00000162, -0.00000024, -0.00000527, -0.00000742, -0.00000170, 0.00000371, 0.00000279, 0.00000116, 0.00000361, 0.00000477, -0.00000025, -0.00000602, -0.00000560, -0.00000156, -0.00000057, -0.00000259, -0.00000281, -0.00000067, 0.00000013, -0.00000101, -0.00000031, 0.00000372, },
		{ -0.00000056, 0.00000320, 0.00000698, 0.00000649, 0.00000235, -0.00000047, 0.00000082, 0.00000354, 0.00000363, 0.00000040, -0.00000418, -0.00000811, -0.00000921, -0.00000559, 0.00000068, 0.00000392, 0.00000248, 0.00000187, 0.00000511, 0.00000546, -0.00000352, -0.00001489, -0.00001575, -0.00000561, 0.00000362, 0.00000537, 0.00000469, 0.00000606, 0.00000653, 0.00000343, -0.00000118, -0.00000534, -0.00000756, -0.00000350, 0.00000678, 0.00001185, 0.00000374, -0.00000578, -0.00000276, 0.00000228, -0.00000788, -0.00001936, -0.00000506, 0.00002114, 0.00001678, -0.00001486, -0.00002096, 0.00000953, 0.00001872, -0.00001734, -0.00003362, 0.00001172, 0.00004619, -0.00000067, -0.00005802, -0.00002011, 0.00005732, 0.00003249, -0.00006686, -0.00006260, 0.00006386, 0.00010332, -0.00003532, -0.00013663, -0.00002100, 0.00012739, 0.00004516, -0.00014550, -0.00010919, 0.00012023, 0.00014517, -0.00012762, -0.00026226, 0.00001167, 0.00026819, -0.00001662, -0.00056429, -0.00054617, 0.00024650, 0.00100644, 0.00091262, 0.00014411, -0.00044208, -0.00040069, -0.00006683, 0.00007699, -0.00001153, -0.00008497, -0.00003917, 0.00002173, 0.00001632, -0.00001772, -0.00002938, -0.00002288, -0.00001399, -0.00000073, 0.00001131, 0.00000746, -0.00000597, -0.00000787, 0.00000160, 0.00000331, -0.00000504, -0.00000643, 0.00000296, 0.00000760, 0.00000114, -0.00000443, -0.00000210, -0.00000036, -0.00000464, -0.00000713, -0.00000236, 0.00000282, 0.00000225, 0.00000071, 0.00000335, 0.00000553, 0.00000154, -0.00000470, -0.00000584, -0.00000247, -0.00000070, -0.00000208, -0.00000288, -0.00000155, -0.00000063, -0.00000123, },
		{ -0.00000136, -0.00000170, -0.00000106, 0.00000254, 0.00000678, 0.00000719, 0.00000330, -0.00000027, 0.00000012, 0.00000271, 0.00000363, 0.00000173, -0.00000193, -0.00000643, -0.00000956, -0.00000768, -0.00000116, 0.00000361, 0.00000275, 0.00000130, 0.00000450, 0.00000696, 0.00000008, -0.00001228, -0.00001685, -0.00000944, 0.00000065, 0.00000530, 0.00000609, 0.00000635, 0.00000557, 0.00000353, 0.00000111, -0.00000341, -0.00000919, -0.00000751, 0.00000515, 0.00001430, 0.00000624, -0.00000642, -0.00000340, 0.00000538, -0.00000475, -0.00002245, -0.00001152, 0.00002080, 0.00002309, -0.00001276, -0.00002737, 0.00000579, 0.00002567, -0.00001092, -0.00004069, 0.00000111, 0.00005130, 0.00001438, -0.00005867, -0.00003864, 0.00005102, 0.00005203, -0.00005225, -0.00008061, 0.00003908, 0.00011439, -0.00000168, -0.00013505, -0.00005819, 0.00011079, 0.00008053, -0.00011242, -0.00013508, 0.00007414, 0.00015850, -0.00006744, -0.00025297, -0.00005667, 0.00022947, 0.00006304, -0.00045086, -0.00056794, 0.00006742, 0.00084637, 0.00093653, 0.00031134, -0.00031125, -0.00040691, -0.00014565, 0.00003665, 0.00000992, -0.00005740, -0.00004469, 0.00000071, 0.00001130, -0.00000536, -0.00001978, -0.00002634, -0.00002263, -0.00000388, 0.00001461, 0.00001110, -0.00000633, -0.00001079, 0.00000017, 0.00000477, -0.00000316, -0.00000666, 0.00000145, 0.00000729, 0.00000235, -0.00000353, -0.00000236, -0.00000066, -0.00000419, -0.00000673, -0.00000279, 0.00000200, 0.00000163, 0.00000011, 0.00000282, 0.00000598, 0.00000326, -0.00000308, -0.00000572, -0.00000336, -0.00000107, -0.00000163, -0.00000265, -0.00000209, },
		{ -0.00000226, -0.00000228, -0.00000190, -0.00000225, -0.00000181, 0.00000162, 0.00000640, 0.00000781, 0.00000434, 0.00000018, -0.00000033, 0.00000188, 0.00000323, 0.00000247, 0.00000006, -0.00000430, -0.00000907, -0.00000933, -0.00000329, 0.00000284, 0.00000301, 0.00000098, 0.00000362, 0.00000756, 0.00000321, -0.00000883, -0.00001645, -0.00001267, -0.00000320, 0.00000411, 0.00000746, 0.00000740, 0.00000468, 0.00000274, 0.00000300, -0.00000042, -0.00000953, -0.00001179, 0.00000208, 0.00001613, 0.00000945, -0.00000674, -0.00000495, 0.00000811, -0.00000004, -0.00002398, -0.00001902, 0.00001821, 0.00002958, -0.00000824, -0.00003331, -0.00000068, 0.00003122, -0.00000158, -0.00004498, -0.00001209, 0.00005194, 0.00003076, -0.00005331, -0.00005619, 0.00003797, 0.00006752, -0.00003146, -0.00009115, 0.00001039, 0.00011503, 0.00003168, -0.00012170, -0.00008975, 0.00008348, 0.00010469, -0.00007250, -0.00014550, 0.00002722, 0.00015403, -0.00001402, -0.00022572, -0.00010843, 0.00017622, 0.00011200, -0.00033891, -0.00054949, -0.00007383, 0.00067450, 0.00090944, 0.00044357, -0.00016695, -0.00037543, -0.00020642, -0.00001618, 0.00001347, -0.00003083, -0.00003939, -0.00001527, 0.00000070, 0.00000082, -0.00000892, -0.00002453, -0.00002945, -0.00000989, 0.00001513, 0.00001529, -0.00000448, -0.00001301, -0.00000248, 0.00000512, -0.00000101, -0.00000602, 0.00000012, 0.00000644, 0.00000323, -0.00000240, -0.00000229, -0.00000098, -0.00000393, -0.00000634, -0.00000303, 0.00000134, 0.00000103, -0.00000061, 0.00000201, 0.00000608, 0.00000484, -0.00000120, -0.00000520, -0.00000412, -0.00000165, -0.00000135, },
		{ -0.00000236, -0.00000130, -0.00000184, -0.00000218, -0.00000215, -0.00000270, -0.00000268, 0.00000042, 0.00000571, 0.00000827, 0.00000544, 0.00000088, -0.00000050, 0.00000116, 0.00000258, 0.00000267, 0.00000162, -0.00000200, -0.00000784, -0.00001035, -0.00000548, 0.00000157, 0.00000309, 0.00000095, 0.00000274, 0.00000739, 0.00000551, -0.00000503, -0.00001455, -0.00001477, -0.00000749, 0.00000160, 0.00000837, 0.00000918, 0.00000424, 0.00000116, 0.00000402, 0.00000325, -0.00000827, -0.00001575, -0.00000237, 0.00001692, 0.00001323, -0.00000643, -0.00000738, 0.00000994, 0.00000595, -0.00002336, -0.00002684, 0.00001303, 0.00003532, -0.00000125, -0.00003773, -0.00000960, 0.00003420, 0.00000988, -0.00004539, -0.00002643, 0.00004737, 0.00004641, -0.00004200, -0.00007032, 0.00001939, 0.00007662, -0.00000700, -0.00009272, -0.00001857, 0.00010526, 0.00006066, -0.00009864, -0.00011204, 0.00004952, 0.00011556, -0.00003129, -0.00014084, -0.00001460, 0.00013504, 0.00002741, -0.00018688, -0.00014063, 0.00011797, 0.00013241, -0.00023912, -0.00050179, -0.00017347, 0.00050656, 0.00084098, 0.00053364, -0.00002381, -0.00031178, -0.00024186, -0.00007255, -0.00000041, -0.00001106, -0.00002654, -0.00002375, -0.00001200, 0.00000060, 0.00000045, -0.00001862, -0.00003293, -0.00001706, 0.00001261, 0.00001867, -0.00000092, -0.00001377, -0.00000559, 0.00000419, 0.00000081, -0.00000469, -0.00000071, 0.00000532, 0.00000364, -0.00000125, -0.00000190, -0.00000117, -0.00000381, -0.00000603, -0.00000315, 0.00000088, 0.00000053, -0.00000139, 0.00000094, 0.00000577, 0.00000616, 0.00000083, -0.00000425, -0.00000464, },
		{ -0.00000289, -0.00000484, -0.00000311, -0.00000149, -0.00000154, -0.00000193, -0.00000209, -0.00000291, -0.00000354, -0.00000102, 0.00000464, 0.00000844, 0.00000654, 0.00000180, -0.00000039, 0.00000061, 0.00000184, 0.00000241, 0.00000264, 0.00000022, -0.00000602, -0.00001058, -0.00000746, -0.00000014, 0.00000280, 0.00000110, 0.00000209, 0.00000670, 0.00000680, -0.00000145, -0.00001137, -0.00001530, -0.00001166, -0.00000223, 0.00000834, 0.00001149, 0.00000462, -0.00000094, 0.00000384, 0.00000709, -0.00000531, -0.00001877, -0.00000796, 0.00001629, 0.00001729, -0.00000521, -0.00001047, 0.00001038, 0.00001267, -0.00002015, -0.00003406, 0.00000524, 0.00003929, 0.00000783, -0.00003965, -0.00002019, 0.00003372, 0.00002220, -0.00004129, -0.00004015, 0.00003759, 0.00005917, -0.00002571, -0.00007894, -0.00000260, 0.00007793, 0.00001800, -0.00008517, -0.00004417, 0.00008666, 0.00008192, -0.00006924, -0.00012298, 0.00001368, 0.00011316, 0.00000599, -0.00012363, -0.00004680, 0.00010647, 0.00005415, -0.00014341, -0.00015356, 0.00006308, 0.00012966, -0.00015854, -0.00043698, -0.00023287, 0.00035539, 0.00074382, 0.00057977, 0.00010497, -0.00022510, -0.00024852, -0.00012330, -0.00002821, -0.00000219, -0.00001064, -0.00002411, -0.00002333, -0.00000485, 0.00000628, -0.00001055, -0.00003249, -0.00002362, 0.00000755, 0.00002019, 0.00000344, -0.00001279, -0.00000836, 0.00000219, 0.00000184, -0.00000304, -0.00000089, 0.00000423, 0.00000362, -0.00000028, -0.00000128, -0.00000115, -0.00000375, -0.00000584, -0.00000324, 0.00000060, 0.00000023, -0.00000213, -0.00000036, 0.00000501, 0.00000712, 0.00000293, },
		{ 0.00000762, 0.00000496, -0.00000117, -0.00000462, -0.00000378, -0.00000189, -0.00000142, -0.00000165, -0.00000178, -0.00000279, -0.00000418, -0.00000258, 0.00000315, 0.00000820, 0.00000752, 0.00000291, 0.00000002, 0.00000028, 0.00000110, 0.00000185, 0.00000313, 0.00000211, -0.00000386, -0.00001002, -0.00000895, -0.00000210, 0.00000201, 0.00000124, 0.00000182, 0.00000584, 0.00000706, 0.00000140, -0.00000740, -0.00001404, -0.00001504, -0.00000712, 0.00000695, 0.00001394, 0.00000605, -0.00000310, 0.00000230, 0.00001047, -0.00000080, -0.00002025, -0.00001421, 0.00001395, 0.00002123, -0.00000286, -0.00001385, 0.00000906, 0.00001939, -0.00001425, -0.00003966, -0.00000475, 0.00004055, 0.00001821, -0.00003830, -0.00003130, 0.00002929, 0.00003386, -0.00003271, -0.00005140, 0.00002341, 0.00006721, -0.00000624, -0.00008072, -0.00002530, 0.00007115, 0.00004032, -0.00006970, -0.00006335, 0.00006206, 0.00009340, -0.00003761, -0.00012227, -0.00001941, 0.00009942, 0.00003524, -0.00009796, -0.00006686, 0.00007391, 0.00006603, -0.00010173, -0.00015016, 0.00001771, 0.00011100, -0.00010022, -0.00036651, -0.00025753, 0.00022969, 0.00063170, 0.00058514, 0.00020925, -0.00012660, -0.00022707, -0.00016056, -0.00006414, -0.00000599, 0.00000349, -0.00001761, -0.00003056, -0.00001333, 0.00000758, -0.00000249, -0.00002852, -0.00002812, 0.00000103, 0.00001934, 0.00000751, -0.00001027, -0.00001014, -0.00000040, 0.00000188, -0.00000149, -0.00000051, 0.00000341, 0.00000333, 0.00000040, -0.00000058, -0.00000090, -0.00000361, -0.00000576, -0.00000338, 0.00000043, 0.00000015, -0.00000270, -0.00000179, 0.00000378, },
		{ -0.00000322, 0.00000213, 0.00000755, 0.00000677, 0.00000083, -0.00000397, -0.00000425, -0.00000239, -0.00000153, -0.00000147, -0.00000135, -0.00000232, -0.00000446, -0.00000407, 0.00000130, 0.00000745, 0.00000822, 0.00000413, 0.00000074, 0.00000022, 0.00000046, 0.00000110, 0.00000314, 0.00000355, -0.00000163, -0.00000878, -0.00000974, -0.00000405, 0.00000067, 0.00000115, 0.00000192, 0.00000514, 0.00000649, 0.00000313, -0.00000323, -0.00001101, -0.00001698, -0.00001259, 0.00000391, 0.00001600, 0.00000859, -0.00000483, -0.00000054, 0.00001280, 0.00000489, -0.00001976, -0.00002050, 0.00000982, 0.00002454, 0.00000067, -0.00001702, 0.00000585, 0.00002522, -0.00000589, -0.00004265, -0.00001615, 0.00003843, 0.00002876, -0.00003337, -0.00004158, 0.00002100, 0.00004328, -0.00002034, -0.00005857, 0.00000636, 0.00006930, 0.00001408, -0.00007531, -0.00004587, 0.00005723, 0.00005720, -0.00004866, -0.00007413, 0.00003500, 0.00009463, -0.00000786, -0.00011138, -0.00004595, 0.00007775, 0.00005402, -0.00006862, -0.00007447, 0.00004266, 0.00006516, -0.00006671, -0.00013520, -0.00001477, 0.00008414, -0.00006356, -0.00029978, -0.00025569, 0.00013364, 0.00051753, 0.00055698, 0.00028294, -0.00002784, -0.00018194, -0.00017890, -0.00010126, -0.00002166, 0.00001175, -0.00000696, -0.00003225, -0.00002223, 0.00000457, 0.00000368, -0.00002216, -0.00002971, -0.00000559, 0.00001625, 0.00001039, -0.00000680, -0.00001056, -0.00000294, 0.00000100, -0.00000044, 0.00000021, 0.00000302, 0.00000298, 0.00000078, 0.00000005, -0.00000047, -0.00000334, -0.00000571, -0.00000363, 0.00000028, 0.00000031, -0.00000300, },
		{ 0.00000063, -0.00000294, -0.00000450, 0.00000014, 0.00000684, 0.00000821, 0.00000300, -0.00000286, -0.00000441, -0.00000292, -0.00000182, -0.00000148, -0.00000094, -0.00000159, -0.00000429, -0.00000529, -0.00000076, 0.00000613, 0.00000849, 0.00000533, 0.00000173, 0.00000045, -0.00000003, 0.00000028, 0.00000280, 0.00000448, 0.00000039, -0.00000706, -0.00000973, -0.00000569, -0.00000111, 0.00000057, 0.00000225, 0.00000491, 0.00000547, 0.00000353, 0.00000046, -0.00000655, -0.00001698, -0.00001794, -0.00000079, 0.00001709, 0.00001208, -0.00000562, -0.00000440, 0.00001356, 0.00001116, -0.00001704, -0.00002613, 0.00000402, 0.00002674, 0.00000529, -0.00001942, 0.00000084, 0.00002932, 0.00000430, -0.00004228, -0.00002782, 0.00003263, 0.00003815, -0.00002503, -0.00004964, 0.00000950, 0.00004910, -0.00000548, -0.00006059, -0.00001156, 0.00006506, 0.00003274, -0.00006342, -0.00006180, 0.00003813, 0.00006676, -0.00002514, -0.00007589, 0.00000914, 0.00008666, 0.00001652, -0.00009309, -0.00006358, 0.00005226, 0.00006179, -0.00004022, -0.00007126, 0.00001685, 0.00005517, -0.00004121, -0.00011410, -0.00003374, 0.00005600, -0.00004512, -0.00024322, -0.00023670, 0.00006709, 0.00041190, 0.00050500, 0.00032436, 0.00006085, -0.00012035, -0.00017606, -0.00013263, -0.00004608, 0.00001147, 0.00000432, -0.00002841, -0.00002911, -0.00000156, 0.00000674, -0.00001499, -0.00002831, -0.00001106, 0.00001156, 0.00001148, -0.00000319, -0.00000966, -0.00000489, -0.00000048, -0.00000010, 0.00000094, 0.00000302, 0.00000276, 0.00000096, 0.00000051, 0.00000004, -0.00000289, -0.00000561, -0.00000397, 0.00000007, },
		{ -0.00000438, -0.00000030, 0.00000102, -0.00000252, -0.00000546, -0.00000203, 0.00000550, 0.00000910, 0.00000518, -0.00000133, -0.00000420, -0.00000334, -0.00000226, -0.00000172, -0.00000068, -0.00000073, -0.00000365, -0.00000604, -0.00000281, 0.00000429, 0.00000818, 0.00000636, 0.00000297, 0.00000101, -0.00000032, -0.00000052, 0.00000223, 0.00000493, 0.00000202, -0.00000515, -0.00000894, -0.00000672, -0.00000313, -0.00000065, 0.00000255, 0.00000529, 0.00000444, 0.00000266, 0.00000303, -0.00000121, -0.00001476, -0.00002241, -0.00000693, 0.00001667, 0.00001613, -0.00000507, -0.00000877, 0.00001245, 0.00001727, -0.00001214, -0.00003036, -0.00000313, 0.00002734, 0.00001069, -0.00002054, -0.00000555, 0.00003097, 0.00001538, -0.00003815, -0.00003842, 0.00002338, 0.00004509, -0.00001400, -0.00005435, -0.00000400, 0.00005042, 0.00001014, -0.00005709, -0.00002819, 0.00005502, 0.00004751, -0.00004671, -0.00007133, 0.00001651, 0.00006828, -0.00000237, -0.00006939, -0.00001233, 0.00007183, 0.00003328, -0.00007100, -0.00007158, 0.00002710, 0.00005980, -0.00001651, -0.00006026, -0.00000105, 0.00004039, -0.00002589, -0.00009194, -0.00004089, 0.00003174, -0.00003979, -0.00019998, -0.00020951, 0.00002643, 0.00032206, 0.00043982, 0.00033590, 0.00013173, -0.00005108, -0.00015306, -0.00015260, -0.00007452, 0.00000194, 0.00001273, -0.00002045, -0.00003228, -0.00000896, 0.00000633, -0.00000860, -0.00002453, -0.00001449, 0.00000627, 0.00001070, -0.00000021, -0.00000779, -0.00000588, -0.00000210, -0.00000046, 0.00000136, 0.00000326, 0.00000279, 0.00000109, 0.00000079, 0.00000051, -0.00000228, -0.00000537, },
		{ -0.00000160, -0.00000495, -0.00000476, -0.00000085, 0.00000135, -0.00000176, -0.00000597, -0.00000420, 0.00000356, 0.00000933, 0.00000719, 0.00000054, -0.00000356, -0.00000356, -0.00000273, -0.00000216, -0.00000066, 0.00000009, -0.00000264, -0.00000619, -0.00000460, 0.00000207, 0.00000718, 0.00000701, 0.00000433, 0.00000192, -0.00000038, -0.00000125, 0.00000156, 0.00000503, 0.00000315, -0.00000335, -0.00000753, -0.00000692, -0.00000506, -0.00000257, 0.00000249, 0.00000627, 0.00000388, 0.00000083, 0.00000402, 0.00000424, -0.00001035, -0.00002522, -0.00001399, 0.00001434, 0.00002017, -0.00000298, -0.00001305, 0.00000939, 0.00002247, -0.00000534, -0.00003258, -0.00001106, 0.00002602, 0.00001643, -0.00001996, -0.00001269, 0.00002970, 0.00002618, -0.00003031, -0.00004662, 0.00001142, 0.00004853, -0.00000145, -0.00005498, -0.00001800, 0.00004696, 0.00002467, -0.00004850, -0.00004155, 0.00004056, 0.00005674, -0.00002746, -0.00007370, -0.00000473, 0.00006230, 0.00001673, -0.00005657, -0.00002724, 0.00005324, 0.00004156, -0.00004876, -0.00007083, 0.00000566, 0.00005058, 0.00000022, -0.00004520, -0.00001039, 0.00002498, -0.00001957, -0.00007268, -0.00003942, 0.00001439, -0.00004196, -0.00017029, -0.00018147, 0.00000580, 0.00025159, 0.00037132, 0.00032306, 0.00018038, 0.00001706, -0.00011381, -0.00015759, -0.00010156, -0.00001555, 0.00001554, -0.00001074, -0.00003115, -0.00001566, 0.00000302, -0.00000421, -0.00001947, -0.00001553, 0.00000141, 0.00000839, 0.00000155, -0.00000554, -0.00000583, -0.00000338, -0.00000132, 0.00000128, 0.00000349, 0.00000308, 0.00000132, 0.00000093, 0.00000086, },
		{ 0.00000105, 0.00000105, -0.00000092, -0.00000434, -0.00000505, -0.00000156, 0.00000151, -0.00000077, -0.00000594, -0.00000616, 0.00000115, 0.00000879, 0.00000882, 0.00000262, -0.00000247, -0.00000349, -0.00000315, -0.00000274, -0.00000093, 0.00000071, -0.00000143, -0.00000573, -0.00000588, -0.00000029, 0.00000549, 0.00000707, 0.00000567, 0.00000317, -0.00000016, -0.00000189, 0.00000087, 0.00000492, 0.00000380, -0.00000193, -0.00000576, -0.00000618, -0.00000652, -0.00000508, 0.00000173, 0.00000763, 0.00000416, -0.00000148, 0.00000322, 0.00000898, -0.00000412, -0.00002573, -0.00002123, 0.00000995, 0.00002353, 0.00000059, -0.00001659, 0.00000460, 0.00002605, 0.00000277, -0.00003234, -0.00001907, 0.00002262, 0.00002190, -0.00001750, -0.00001979, 0.00002542, 0.00003547, -0.00001934, -0.00005128, -0.00000205, 0.00004785, 0.00001117, -0.00005140, -0.00003089, 0.00003911, 0.00003638, -0.00003596, -0.00005014, 0.00002364, 0.00005965, -0.00000824, -0.00006923, -0.00002296, 0.00005041, 0.00003010, -0.00004012, -0.00003467, 0.00003417, 0.00004195, -0.00002952, -0.00006343, -0.00000981, 0.00003739, 0.00000924, -0.00002971, -0.00001217, 0.00001218, -0.00001974, -0.00005865, -0.00003316, 0.00000475, -0.00004660, -0.00015208, -0.00015761, -0.00000155, 0.00020065, 0.00030749, 0.00029327, 0.00020587, 0.00007645, -0.00006424, -0.00014664, -0.00012215, -0.00003804, 0.00001133, -0.00000202, -0.00002624, -0.00002005, -0.00000199, -0.00000242, -0.00001440, -0.00001445, -0.00000217, 0.00000526, 0.00000189, -0.00000358, -0.00000497, -0.00000400, -0.00000232, 0.00000067, 0.00000347, 0.00000351, 0.00000175, },
		{ 0.00000389, 0.00000240, 0.00000127, 0.00000112, -0.00000033, -0.00000358, -0.00000515, -0.00000239, 0.00000141, 0.00000031, -0.00000536, -0.00000768, -0.00000152, 0.00000747, 0.00000988, 0.00000474, -0.00000096, -0.00000305, -0.00000342, -0.00000337, -0.00000146, 0.00000100, -0.00000021, -0.00000472, -0.00000646, -0.00000249, 0.00000321, 0.00000638, 0.00000677, 0.00000471, 0.00000040, -0.00000242, 0.00000021, 0.00000476, 0.00000406, -0.00000108, -0.00000400, -0.00000456, -0.00000716, -0.00000791, 0.00000002, 0.00000900, 0.00000549, -0.00000365, 0.00000069, 0.00001224, 0.00000329, -0.00002359, -0.00002777, 0.00000365, 0.00002555, 0.00000537, -0.00001881, -0.00000148, 0.00002750, 0.00001146, -0.00002946, -0.00002636, 0.00001725, 0.00002647, -0.00001319, -0.00002597, 0.00001841, 0.00004215, -0.00000624, -0.00005168, -0.00001554, 0.00004297, 0.00002234, -0.00004407, -0.00004119, 0.00002788, 0.00004401, -0.00002113, -0.00005322, 0.00000648, 0.00005642, 0.00000859, -0.00005923, -0.00003621, 0.00003494, 0.00003674, -0.00002297, -0.00003496, 0.00001751, 0.00003623, -0.00001539, -0.00005222, -0.00001851, 0.00002349, 0.00001132, -0.00001673, -0.00000852, 0.00000395, -0.00002337, -0.00005051, -0.00002567, 0.00000186, -0.00004995, -0.00014196, -0.00014047, -0.00000205, 0.00016675, 0.00025364, 0.00025445, 0.00021042, 0.00012170, -0.00001120, -0.00012133, -0.00013245, -0.00006158, 0.00000027, 0.00000323, -0.00001907, -0.00002125, -0.00000719, -0.00000310, -0.00001041, -0.00001197, -0.00000404, 0.00000213, 0.00000098, -0.00000239, -0.00000373, -0.00000384, -0.00000307, -0.00000031, 0.00000301, },
		{ -0.00000139, 0.00000208, 0.00000400, 0.00000316, 0.00000169, 0.00000115, 0.00000013, -0.00000273, -0.00000503, -0.00000324, 0.00000101, 0.00000132, -0.00000429, -0.00000860, -0.00000422, 0.00000543, 0.00001020, 0.00000671, 0.00000089, -0.00000220, -0.00000344, -0.00000396, -0.00000217, 0.00000092, 0.00000078, -0.00000336, -0.00000625, -0.00000420, 0.00000057, 0.00000485, 0.00000739, 0.00000644, 0.00000133, -0.00000282, -0.00000042, 0.00000467, 0.00000413, -0.00000086, -0.00000260, -0.00000227, -0.00000670, -0.00001065, -0.00000272, 0.00000992, 0.00000784, -0.00000505, -0.00000317, 0.00001343, 0.00001099, -0.00001874, -0.00003273, -0.00000411, 0.00002569, 0.00001081, -0.00001930, -0.00000821, 0.00002652, 0.00001986, -0.00002402, -0.00003217, 0.00001024, 0.00002954, -0.00000735, -0.00003045, 0.00000934, 0.00004538, 0.00000764, -0.00004759, -0.00002748, 0.00003443, 0.00003072, -0.00003400, -0.00004787, 0.00001473, 0.00004690, -0.00000593, -0.00005084, -0.00000878, 0.00004810, 0.00002118, -0.00004572, -0.00004347, 0.00001851, 0.00003683, -0.00000779, -0.00002959, 0.00000529, 0.00002686, -0.00000726, -0.00004012, -0.00002096, 0.00001158, 0.00000829, -0.00000803, -0.00000212, 0.00000079, -0.00002754, -0.00004749, -0.00001969, 0.00000363, -0.00004993, -0.00013619, -0.00013034, -0.00000089, 0.00014573, 0.00021223, 0.00021380, 0.00019852, 0.00015010, 0.00003873, -0.00008532, -0.00013051, -0.00008199, -0.00001598, 0.00000336, -0.00001168, -0.00001928, -0.00001115, -0.00000551, -0.00000818, -0.00000905, -0.00000422, -0.00000026, -0.00000071, -0.00000226, -0.00000264, -0.00000303, -0.00000327, },
		{ -0.00000191, -0.00000280, -0.00000223, 0.00000079, 0.00000367, 0.00000389, 0.00000232, 0.00000123, 0.00000044, -0.00000187, -0.00000467, -0.00000401, 0.00000032, 0.00000210, -0.00000288, -0.00000882, -0.00000666, 0.00000283, 0.00000968, 0.00000829, 0.00000296, -0.00000092, -0.00000317, -0.00000441, -0.00000295, 0.00000049, 0.00000139, -0.00000190, -0.00000528, -0.00000515, -0.00000210, 0.00000250, 0.00000730, 0.00000817, 0.00000266, -0.00000304, -0.00000107, 0.00000472, 0.00000423, -0.00000121, -0.00000190, 0.00000032, -0.00000500, -0.00001284, -0.00000640, 0.00000993, 0.00001095, -0.00000518, -0.00000777, 0.00001224, 0.00001801, -0.00001151, -0.00003533, -0.00001262, 0.00002363, 0.00001626, -0.00001792, -0.00001487, 0.00002313, 0.00002712, -0.00001644, -0.00003584, 0.00000214, 0.00003065, -0.00000053, -0.00003266, -0.00000082, 0.00004476, 0.00002081, -0.00003940, -0.00003648, 0.00002331, 0.00003540, -0.00002257, -0.00005044, 0.00000128, 0.00004508, 0.00000781, -0.00004385, -0.00002047, 0.00003639, 0.00002849, -0.00003100, -0.00004473, 0.00000357, 0.00003155, 0.00000347, -0.00002072, -0.00000155, 0.00001649, -0.00000480, -0.00002963, -0.00001871, 0.00000338, 0.00000253, -0.00000414, 0.00000446, 0.00000200, -0.00003007, -0.00004790, -0.00001672, 0.00000753, -0.00004602, -0.00013147, -0.00012586, -0.00000146, 0.00013287, 0.00018319, 0.00017689, 0.00017587, 0.00016168, 0.00008021, -0.00004358, -0.00011645, -0.00009569, -0.00003464, -0.00000208, -0.00000617, -0.00001502, -0.00001295, -0.00000852, -0.00000783, -0.00000663, -0.00000318, -0.00000144, -0.00000252, -0.00000311, -0.00000215, },
		{ -0.00000459, -0.00000252, -0.00000092, -0.00000173, -0.00000255, -0.00000062, 0.00000286, 0.00000439, 0.00000311, 0.00000145, 0.00000066, -0.00000107, -0.00000410, -0.00000461, -0.00000061, 0.00000252, -0.00000130, -0.00000832, -0.00000859, -0.00000011, 0.00000827, 0.00000927, 0.00000507, 0.00000075, -0.00000254, -0.00000464, -0.00000368, -0.00000019, 0.00000150, -0.00000064, -0.00000375, -0.00000515, -0.00000445, -0.00000052, 0.00000631, 0.00000970, 0.00000436, -0.00000304, -0.00000178, 0.00000490, 0.00000458, -0.00000191, -0.00000210, 0.00000276, -0.00000212, -0.00001400, -0.00001072, 0.00000866, 0.00001436, -0.00000373, -0.00001231, 0.00000872, 0.00002342, -0.00000257, -0.00003506, -0.00002098, 0.00001934, 0.00002100, -0.00001474, -0.00002072, 0.00001760, 0.00003250, -0.00000734, -0.00003691, -0.00000628, 0.00002955, 0.00000654, -0.00003229, -0.00001099, 0.00004037, 0.00003187, -0.00002804, -0.00004150, 0.00001104, 0.00003602, -0.00001132, -0.00004904, -0.00001090, 0.00003926, 0.00001859, -0.00003366, -0.00002756, 0.00002331, 0.00003041, -0.00001730, -0.00004094, -0.00000803, 0.00002279, 0.00000983, -0.00001074, -0.00000317, 0.00000744, -0.00000669, -0.00002239, -0.00001385, -0.00000051, -0.00000359, -0.00000447, 0.00000923, 0.00000603, -0.00002975, -0.00004967, -0.00001704, 0.00001128, -0.00003897, -0.00012551, -0.00012478, -0.00000532, 0.00012382, 0.00016451, 0.00014719, 0.00014832, 0.00015871, 0.00010981, -0.00000146, -0.00009228, -0.00010036, -0.00005243, -0.00001229, -0.00000414, -0.00000995, -0.00001231, -0.00001098, -0.00000893, -0.00000537, -0.00000164, -0.00000135, -0.00000381, },
		{ -0.00000028, -0.00000414, -0.00000617, -0.00000376, -0.00000048, -0.00000034, -0.00000221, -0.00000185, 0.00000164, 0.00000450, 0.00000393, 0.00000184, 0.00000085, -0.00000036, -0.00000337, -0.00000498, -0.00000166, 0.00000252, 0.00000022, -0.00000719, -0.00000980, -0.00000307, 0.00000606, 0.00000945, 0.00000701, 0.00000275, -0.00000153, -0.00000462, -0.00000425, -0.00000098, 0.00000112, 0.00000017, -0.00000193, -0.00000414, -0.00000607, -0.00000392, 0.00000432, 0.00001075, 0.00000636, -0.00000274, -0.00000258, 0.00000513, 0.00000534, -0.00000269, -0.00000325, 0.00000458, 0.00000168, -0.00001372, -0.00001520, 0.00000589, 0.00001747, -0.00000065, -0.00001602, 0.00000325, 0.00002648, 0.00000717, -0.00003174, -0.00002825, 0.00001311, 0.00002437, -0.00001013, -0.00002516, 0.00001048, 0.00003546, 0.00000244, -0.00003519, -0.00001423, 0.00002626, 0.00001308, -0.00002940, -0.00002004, 0.00003276, 0.00003966, -0.00001483, -0.00004209, -0.00000077, 0.00003284, -0.00000163, -0.00004439, -0.00002059, 0.00003063, 0.00002553, -0.00002199, -0.00002985, 0.00001088, 0.00002766, -0.00000638, -0.00003371, -0.00001524, 0.00001268, 0.00001128, -0.00000182, -0.00000067, 0.00000131, -0.00001107, -0.00001905, -0.00000854, -0.00000053, -0.00000820, -0.00000764, 0.00001108, 0.00001104, -0.00002643, -0.00005094, -0.00001996, 0.00001328, -0.00003027, -0.00011718, -0.00012469, -0.00001245, 0.00011527, 0.00015314, 0.00012599, 0.00012093, 0.00014500, 0.00012628, 0.00003629, -0.00006144, -0.00009525, -0.00006624, -0.00002542, -0.00000636, -0.00000575, -0.00000972, -0.00001198, -0.00001066, -0.00000548, -0.00000039, },
		{ -0.00000673, -0.00000002, 0.00000120, -0.00000340, -0.00000729, -0.00000559, -0.00000085, 0.00000098, -0.00000122, -0.00000262, 0.00000020, 0.00000413, 0.00000461, 0.00000240, 0.00000107, 0.00000026, -0.00000254, -0.00000506, -0.00000272, 0.00000210, 0.00000148, -0.00000560, -0.00001017, -0.00000574, 0.00000322, 0.00000870, 0.00000855, 0.00000497, -0.00000015, -0.00000432, -0.00000459, -0.00000170, 0.00000036, 0.00000035, -0.00000017, -0.00000223, -0.00000666, -0.00000734, 0.00000134, 0.00001107, 0.00000852, -0.00000208, -0.00000348, 0.00000529, 0.00000657, -0.00000319, -0.00000524, 0.00000536, 0.00000597, -0.00001181, -0.00001924, 0.00000165, 0.00001966, 0.00000381, -0.00001820, -0.00000348, 0.00002674, 0.00001664, -0.00002557, -0.00003357, 0.00000552, 0.00002588, -0.00000464, -0.00002778, 0.00000244, 0.00003573, 0.00001200, -0.00003081, -0.00002089, 0.00002108, 0.00001835, -0.00002439, -0.00002706, 0.00002291, 0.00004349, -0.00000135, -0.00003837, -0.00001065, 0.00002669, 0.00000542, -0.00003762, -0.00002708, 0.00002064, 0.00002836, -0.00001058, -0.00002784, 0.00000071, 0.00002158, 0.00000073, -0.00002494, -0.00001795, 0.00000326, 0.00000872, 0.00000448, 0.00000430, -0.00000119, -0.00001595, -0.00001930, -0.00000454, 0.00000211, -0.00001022, -0.00001193, 0.00000986, 0.00001530, -0.00002081, -0.00005043, -0.00002423, 0.00001280, -0.00002158, -0.00010649, -0.00012363, -0.00002175, 0.00010521, 0.00014581, 0.00011279, 0.00009732, 0.00012499, 0.00013040, 0.00006612, -0.00002805, -0.00008129, -0.00007374, -0.00003898, -0.00001255, -0.00000386, -0.00000622, -0.00001117, -0.00001206, },
		{ -0.00000878, -0.00001236, -0.00000847, -0.00000079, 0.00000242, -0.00000177, -0.00000753, -0.00000755, -0.00000209, 0.00000183, 0.00000018, -0.00000275, -0.00000119, 0.00000333, 0.00000500, 0.00000302, 0.00000138, 0.00000080, -0.00000166, -0.00000485, -0.00000365, 0.00000131, 0.00000230, -0.00000379, -0.00000967, -0.00000780, 0.00000002, 0.00000699, 0.00000945, 0.00000724, 0.00000160, -0.00000373, -0.00000467, -0.00000219, -0.00000060, -0.00000019, 0.00000115, 0.00000033, -0.00000600, -0.00001031, -0.00000246, 0.00001044, 0.00001065, -0.00000102, -0.00000446, 0.00000522, 0.00000824, -0.00000310, -0.00000779, 0.00000480, 0.00001019, -0.00000824, -0.00002224, -0.00000377, 0.00002037, 0.00000913, -0.00001838, -0.00001062, 0.00002412, 0.00002479, -0.00001711, -0.00003627, -0.00000266, 0.00002531, 0.00000108, -0.00002841, -0.00000573, 0.00003332, 0.00002045, -0.00002417, -0.00002559, 0.00001458, 0.00002179, -0.00001792, -0.00003140, 0.00001200, 0.00004319, 0.00001086, -0.00003109, -0.00001746, 0.00001880, 0.00000931, -0.00003007, -0.00003025, 0.00001071, 0.00002747, -0.00000087, -0.00002263, -0.00000615, 0.00001385, 0.00000377, -0.00001649, -0.00001683, -0.00000396, 0.00000358, 0.00000737, 0.00000989, -0.00000026, -0.00001966, -0.00002209, -0.00000291, 0.00000587, -0.00000942, -0.00001573, 0.00000621, 0.00001764, -0.00001409, -0.00004761, -0.00002842, 0.00000998, -0.00001428, -0.00009420, -0.00012040, -0.00003162, 0.00009299, 0.00013969, 0.00010584, 0.00007947, 0.00010292, 0.00012448, 0.00008616, 0.00000387, -0.00006073, -0.00007381, -0.00005038, -0.00002152, -0.00000513, -0.00000316, },
		{ -0.00000959, -0.00000178, -0.00000559, -0.00001114, -0.00000992, -0.00000255, 0.00000282, 0.00000024, -0.00000672, -0.00000913, -0.00000400, 0.00000191, 0.00000164, -0.00000222, -0.00000224, 0.00000223, 0.00000501, 0.00000359, 0.00000177, 0.00000129, -0.00000079, -0.00000438, -0.00000433, 0.00000029, 0.00000258, -0.00000203, -0.00000842, -0.00000901, -0.00000320, 0.00000438, 0.00000949, 0.00000935, 0.00000366, -0.00000286, -0.00000450, -0.00000235, -0.00000152, -0.00000137, 0.00000171, 0.00000314, -0.00000407, -0.00001241, -0.00000678, 0.00000873, 0.00001250, 0.00000045, -0.00000543, 0.00000476, 0.00001021, -0.00000216, -0.00001049, 0.00000280, 0.00001371, -0.00000327, -0.00002366, -0.00000988, 0.00001925, 0.00001464, -0.00001638, -0.00001723, 0.00001893, 0.00003075, -0.00000721, -0.00003603, -0.00001054, 0.00002270, 0.00000636, -0.00002714, -0.00001330, 0.00002856, 0.00002704, -0.00001595, -0.00002787, 0.00000748, 0.00002307, -0.00001085, -0.00003283, 0.00000132, 0.00003915, 0.00002051, -0.00002144, -0.00002057, 0.00001059, 0.00001009, -0.00002299, -0.00003050, 0.00000204, 0.00002372, 0.00000619, -0.00001563, -0.00000935, 0.00000616, 0.00000324, -0.00000981, -0.00001312, -0.00000817, -0.00000246, 0.00000693, 0.00001452, 0.00000321, -0.00002114, -0.00002601, -0.00000393, 0.00000925, -0.00000633, -0.00001790, 0.00000124, 0.00001757, -0.00000757, -0.00004267, -0.00003141, 0.00000554, -0.00000920, -0.00008148, -0.00011462, -0.00004054, 0.00007898, 0.00013280, 0.00010286, 0.00006779, 0.00008217, 0.00011174, 0.00009624, 0.00003105, -0.00003662, -0.00006666, -0.00005751, -0.00003138, },
		{ -0.00005913, -0.00004005, -0.00001644, -0.00000293, -0.00000270, -0.00000851, -0.00001035, -0.00000478, 0.00000212, 0.00000204, -0.00000499, -0.00000989, -0.00000617, 0.00000112, 0.00000279, -0.00000119, -0.00000275, 0.00000107, 0.00000461, 0.00000397, 0.00000221, 0.00000176, 0.00000003, -0.00000370, -0.00000467, -0.00000080, 0.00000231, -0.00000060, -0.00000661, -0.00000922, -0.00000606, 0.00000107, 0.00000853, 0.00001108, 0.00000594, -0.00000171, -0.00000412, -0.00000212, -0.00000216, -0.00000301, 0.00000131, 0.00000576, -0.00000099, -0.00001326, -0.00001122, 0.00000591, 0.00001381, 0.00000228, -0.00000629, 0.00000381, 0.00001222, -0.00000024, -0.00001285, -0.00000053, 0.00001599, 0.00000264, -0.00002312, -0.00001600, 0.00001618, 0.00001958, -0.00001237, -0.00002245, 0.00001179, 0.00003390, 0.00000310, -0.00003284, -0.00001725, 0.00001840, 0.00001061, -0.00002428, -0.00001965, 0.00002196, 0.00003124, -0.00000696, -0.00002754, 0.00000061, 0.00002220, -0.00000406, -0.00003152, -0.00000799, 0.00003222, 0.00002672, -0.00001089, -0.00001997, 0.00000343, 0.00000831, -0.00001736, -0.00002867, -0.00000463, 0.00001826, 0.00001020, -0.00000828, -0.00000921, -0.00000013, 0.00000018, -0.00000574, -0.00000830, -0.00000927, -0.00000789, 0.00000386, 0.00001710, 0.00000793, -0.00002005, -0.00002962, -0.00000716, 0.00001113, -0.00000195, -0.00001790, -0.00000379, 0.00001527, -0.00000232, -0.00003635, -0.00003254, 0.00000048, -0.00000653, -0.00006947, -0.00010666, -0.00004738, 0.00006414, 0.00012418, 0.00010159, 0.00006149, 0.00006493, 0.00009551, 0.00009759, 0.00005143, -0.00001222, -0.00005371, },
		{ 0.00000964, -0.00003721, -0.00005509, -0.00004567, -0.00002423, -0.00000677, -0.00000123, -0.00000511, -0.00000933, -0.00000677, 0.00000039, 0.00000309, -0.00000275, -0.00000961, -0.00000811, -0.00000041, 0.00000329, 0.00000004, -0.00000266, 0.00000009, 0.00000392, 0.00000408, 0.00000261, 0.00000221, 0.00000077, -0.00000289, -0.00000465, -0.00000177, 0.00000156, 0.00000027, -0.00000456, -0.00000839, -0.00000819, -0.00000265, 0.00000651, 0.00001219, 0.00000831, -0.00000031, -0.00000358, -0.00000152, -0.00000230, -0.00000482, -0.00000011, 0.00000773, 0.00000291, -0.00001263, -0.00001531, 0.00000210, 0.00001436, 0.00000437, -0.00000691, 0.00000232, 0.00001398, 0.00000263, -0.00001440, -0.00000487, 0.00001660, 0.00000885, -0.00002050, -0.00002140, 0.00001134, 0.00002323, -0.00000679, -0.00002564, 0.00000357, 0.00003399, 0.00001274, -0.00002708, -0.00002209, 0.00001296, 0.00001344, -0.00002031, -0.00002436, 0.00001422, 0.00003279, 0.00000189, -0.00002474, -0.00000524, 0.00001948, 0.00000166, -0.00002801, -0.00001507, 0.00002355, 0.00002918, -0.00000090, -0.00001620, -0.00000165, 0.00000496, -0.00001372, -0.00002575, -0.00000905, 0.00001224, 0.00001133, -0.00000181, -0.00000654, -0.00000415, -0.00000410, -0.00000447, -0.00000378, -0.00000785, -0.00001164, -0.00000068, 0.00001719, 0.00001255, -0.00001670, -0.00003185, -0.00001170, 0.00001098, 0.00000255, -0.00001586, -0.00000786, 0.00001141, 0.00000104, -0.00002959, -0.00003171, -0.00000424, -0.00000591, -0.00005898, -0.00009729, -0.00005164, 0.00004964, 0.00011373, 0.00010029, 0.00005908, 0.00005210, 0.00007872, 0.00009240, 0.00006431, },
		{ 0.00008321, 0.00007025, 0.00002699, -0.00001975, -0.00004631, -0.00004703, -0.00003121, -0.00001274, -0.00000199, -0.00000188, -0.00000689, -0.00000781, -0.00000192, 0.00000303, -0.00000058, -0.00000835, -0.00000938, -0.00000235, 0.00000298, 0.00000112, -0.00000205, -0.00000047, 0.00000309, 0.00000386, 0.00000289, 0.00000263, 0.00000141, -0.00000204, -0.00000425, -0.00000244, 0.00000049, 0.00000046, -0.00000260, -0.00000666, -0.00000931, -0.00000639, 0.00000350, 0.00001244, 0.00001059, 0.00000130, -0.00000296, -0.00000064, -0.00000180, -0.00000646, -0.00000245, 0.00000865, 0.00000719, -0.00001044, -0.00001858, -0.00000246, 0.00001395, 0.00000655, -0.00000718, 0.00000032, 0.00001520, 0.00000628, -0.00001475, -0.00000972, 0.00001532, 0.00001464, -0.00001596, -0.00002541, 0.00000517, 0.00002505, -0.00000033, -0.00002646, -0.00000474, 0.00003117, 0.00002078, -0.00001944, -0.00002458, 0.00000709, 0.00001468, -0.00001579, -0.00002722, 0.00000610, 0.00003172, 0.00000975, -0.00001989, -0.00000944, 0.00001547, 0.00000577, -0.00002308, -0.00001948, 0.00001441, 0.00002811, 0.00000727, -0.00001028, -0.00000404, 0.00000115, -0.00001212, -0.00002273, -0.00001141, 0.00000663, 0.00001018, 0.00000294, -0.00000241, -0.00000559, -0.00000830, -0.00000559, -0.00000061, -0.00000486, -0.00001320, -0.00000546, 0.00001500, 0.00001600, -0.00001190, -0.00003211, -0.00001646, 0.00000885, 0.00000619, -0.00001237, -0.00001032, 0.00000692, 0.00000244, -0.00002328, -0.00002929, -0.00000796, -0.00000668, -0.00005040, -0.00008739, -0.00005335, 0.00003641, 0.00010198, 0.00009790, 0.00005892, 0.00004353, 0.00006345, },
		{ 0.00003839, 0.00005077, 0.00007237, 0.00007070, 0.00003899, -0.00000370, -0.00003450, -0.00004384, -0.00003575, -0.00001966, -0.00000525, 0.00000011, -0.00000353, -0.00000742, -0.00000415, 0.00000189, 0.00000096, -0.00000645, -0.00000968, -0.00000425, 0.00000191, 0.00000171, -0.00000115, -0.00000053, 0.00000235, 0.00000339, 0.00000298, 0.00000299, 0.00000194, -0.00000123, -0.00000355, -0.00000267, -0.00000067, -0.00000006, -0.00000107, -0.00000428, -0.00000923, -0.00000974, -0.00000032, 0.00001168, 0.00001259, 0.00000309, -0.00000229, 0.00000039, -0.00000063, -0.00000759, -0.00000546, 0.00000827, 0.00001131, -0.00000683, -0.00002061, -0.00000740, 0.00001250, 0.00000862, -0.00000699, -0.00000206, 0.00001560, 0.00001038, -0.00001364, -0.00001448, 0.00001219, 0.00001932, -0.00000991, -0.00002750, -0.00000166, 0.00002477, 0.00000621, -0.00002489, -0.00001222, 0.00002593, 0.00002648, -0.00001078, -0.00002457, 0.00000151, 0.00001440, -0.00001128, -0.00002826, -0.00000170, 0.00002832, 0.00001591, -0.00001365, -0.00001157, 0.00001089, 0.00000800, -0.00001763, -0.00002120, 0.00000598, 0.00002422, 0.00001277, -0.00000345, -0.00000371, -0.00000202, -0.00001214, -0.00002035, -0.00001226, 0.00000206, 0.00000756, 0.00000563, 0.00000204, -0.00000470, -0.00001141, -0.00000827, 0.00000066, -0.00000143, -0.00001264, -0.00000942, 0.00001119, 0.00001762, -0.00000666, -0.00003036, -0.00002046, 0.00000526, 0.00000834, -0.00000826, -0.00001107, 0.00000266, 0.00000222, -0.00001803, -0.00002593, -0.00001042, -0.00000813, -0.00004366, -0.00007772, -0.00005297, 0.00002502, 0.00008970, 0.00009409, 0.00005958, },
		{ 0.00008903, 0.00006014, 0.00003557, 0.00004089, 0.00006163, 0.00006753, 0.00004584, 0.00000922, -0.00002175, -0.00003673, -0.00003670, -0.00002600, -0.00001062, 0.00000004, -0.00000012, -0.00000553, -0.00000559, -0.00000001, 0.00000149, -0.00000441, -0.00000899, -0.00000565, 0.00000032, 0.00000160, -0.00000026, -0.00000009, 0.00000190, 0.00000277, 0.00000284, 0.00000325, 0.00000234, -0.00000057, -0.00000266, -0.00000239, -0.00000166, -0.00000122, -0.00000027, -0.00000160, -0.00000792, -0.00001229, -0.00000466, 0.00000984, 0.00001408, 0.00000497, -0.00000163, 0.00000143, 0.00000115, -0.00000791, -0.00000875, 0.00000650, 0.00001474, -0.00000210, -0.00002111, -0.00001229, 0.00001003, 0.00001038, -0.00000632, -0.00000464, 0.00001500, 0.00001450, -0.00001103, -0.00001853, 0.00000748, 0.00002232, -0.00000298, -0.00002739, -0.00000841, 0.00002239, 0.00001203, -0.00002124, -0.00001810, 0.00001901, 0.00002947, -0.00000206, -0.00002221, -0.00000310, 0.00001288, -0.00000723, -0.00002770, -0.00000857, 0.00002311, 0.00001990, -0.00000682, -0.00001154, 0.00000653, 0.00000843, -0.00001250, -0.00002061, -0.00000084, 0.00001854, 0.00001527, 0.00000304, -0.00000112, -0.00000376, -0.00001308, -0.00001900, -0.00001229, -0.00000123, 0.00000433, 0.00000632, 0.00000590, -0.00000211, -0.00001281, -0.00001151, 0.00000000, 0.00000144, -0.00001050, -0.00001189, 0.00000668, 0.00001730, -0.00000196, -0.00002706, -0.00002305, 0.00000101, 0.00000884, -0.00000437, -0.00001039, -0.00000076, 0.00000101, -0.00001407, -0.00002227, -0.00001172, -0.00000970, -0.00003843, -0.00006870, -0.00005111, 0.00001561, 0.00007760, },
		{ 0.00000800, 0.00006616, 0.00008316, 0.00006020, 0.00003404, 0.00003337, 0.00005201, 0.00006258, 0.00004855, 0.00001818, -0.00001008, -0.00002710, -0.00003368, -0.00003023, -0.00001709, -0.00000248, 0.00000231, -0.00000251, -0.00000567, -0.00000203, 0.00000089, -0.00000283, -0.00000755, -0.00000621, -0.00000138, 0.00000074, 0.00000027, 0.00000068, 0.00000186, 0.00000217, 0.00000252, 0.00000339, 0.00000260, -0.00000013, -0.00000173, -0.00000159, -0.00000224, -0.00000280, -0.00000038, 0.00000095, -0.00000550, -0.00001369, -0.00000911, 0.00000693, 0.00001486, 0.00000684, -0.00000098, 0.00000231, 0.00000339, -0.00000722, -0.00001190, 0.00000343, 0.00001700, 0.00000328, -0.00001995, -0.00001667, 0.00000666, 0.00001161, -0.00000516, -0.00000719, 0.00001333, 0.00001819, -0.00000706, -0.00002134, 0.00000171, 0.00002327, 0.00000409, -0.00002508, -0.00001432, 0.00001821, 0.00001646, -0.00001610, -0.00002188, 0.00001132, 0.00002970, 0.00000581, -0.00001795, -0.00000627, 0.00001056, -0.00000398, -0.00002591, -0.00001408, 0.00001676, 0.00002152, -0.00000023, -0.00000955, 0.00000306, 0.00000746, -0.00000834, -0.00001839, -0.00000553, 0.00001221, 0.00001494, 0.00000813, 0.00000291, -0.00000367, -0.00001409, -0.00001869, -0.00001215, -0.00000334, 0.00000116, 0.00000544, 0.00000854, 0.00000137, -0.00001234, -0.00001436, -0.00000216, 0.00000308, -0.00000757, -0.00001267, 0.00000236, 0.00001536, 0.00000154, -0.00002293, -0.00002402, -0.00000310, 0.00000794, -0.00000129, -0.00000886, -0.00000311, -0.00000054, -0.00001124, -0.00001881, -0.00001223, -0.00001110, -0.00003427, -0.00006046, -0.00004830, },
		{ -0.00005288, -0.00004489, 0.00000185, 0.00005555, 0.00007693, 0.00005976, 0.00003314, 0.00002752, 0.00004376, 0.00005724, 0.00004859, 0.00002328, -0.00000100, -0.00001676, -0.00002717, -0.00003123, -0.00002326, -0.00000721, 0.00000287, 0.00000084, -0.00000420, -0.00000345, -0.00000058, -0.00000215, -0.00000581, -0.00000580, -0.00000274, -0.00000069, 0.00000020, 0.00000152, 0.00000225, 0.00000175, 0.00000208, 0.00000341, 0.00000273, 0.00000004, -0.00000092, -0.00000035, -0.00000219, -0.00000450, -0.00000145, 0.00000297, -0.00000225, -0.00001373, -0.00001323, 0.00000313, 0.00001477, 0.00000856, -0.00000037, 0.00000291, 0.00000586, -0.00000547, -0.00001443, -0.00000065, 0.00001777, 0.00000877, -0.00001717, -0.00002011, 0.00000266, 0.00001216, -0.00000361, -0.00000948, 0.00001067, 0.00002101, -0.00000207, -0.00002249, -0.00000449, 0.00002207, 0.00001052, -0.00002084, -0.00001874, 0.00001277, 0.00001907, -0.00001020, -0.00002336, 0.00000375, 0.00002745, 0.00001211, -0.00001243, -0.00000772, 0.00000796, -0.00000167, -0.00002332, -0.00001802, 0.00000996, 0.00002083, 0.00000535, -0.00000607, 0.00000100, 0.00000565, -0.00000553, -0.00001534, -0.00000801, 0.00000629, 0.00001239, 0.00001116, 0.00000735, -0.00000181, -0.00001443, -0.00001906, -0.00001228, -0.00000462, -0.00000149, 0.00000354, 0.00000973, 0.00000490, -0.00001028, -0.00001613, -0.00000515, 0.00000322, -0.00000467, -0.00001198, -0.00000107, 0.00001243, 0.00000355, -0.00001873, -0.00002355, -0.00000647, 0.00000617, 0.00000073, -0.00000709, -0.00000450, -0.00000195, -0.00000917, -0.00001576, -0.00001236, -0.00001233, -0.00003082, },
		{ -0.00001359, -0.00002796, -0.00004576, -0.00004099, -0.00000313, 0.00004567, 0.00007059, 0.00005906, 0.00003260, 0.00002268, 0.00003661, 0.00005228, 0.00004742, 0.00002538, 0.00000474, -0.00000753, -0.00001837, -0.00002850, -0.00002765, -0.00001331, 0.00000102, 0.00000352, -0.00000141, -0.00000365, -0.00000239, -0.00000258, -0.00000434, -0.00000456, -0.00000340, -0.00000238, -0.00000056, 0.00000214, 0.00000298, 0.00000164, 0.00000164, 0.00000336, 0.00000277, -0.00000010, -0.00000042, 0.00000115, -0.00000139, -0.00000599, -0.00000337, 0.00000408, 0.00000142, -0.00001233, -0.00001659, -0.00000131, 0.00001372, 0.00000998, 0.00000020, 0.00000313, 0.00000829, -0.00000271, -0.00001596, -0.00000533, 0.00001687, 0.00001379, -0.00001303, -0.00002229, -0.00000168, 0.00001193, -0.00000178, -0.00001128, 0.00000720, 0.00002263, 0.00000346, -0.00002180, -0.00001040, 0.00001889, 0.00001563, -0.00001521, -0.00002125, 0.00000678, 0.00001968, -0.00000430, -0.00002269, -0.00000290, 0.00002330, 0.00001638, -0.00000641, -0.00000746, 0.00000558, -0.00000026, -0.00002036, -0.00002034, 0.00000339, 0.00001818, 0.00000936, -0.00000176, 0.00000055, 0.00000370, -0.00000409, -0.00001224, -0.00000857, 0.00000156, 0.00000853, 0.00001192, 0.00001122, 0.00000138, -0.00001361, -0.00001957, -0.00001286, -0.00000550, -0.00000350, 0.00000119, 0.00000955, 0.00000784, -0.00000713, -0.00001648, -0.00000820, 0.00000201, -0.00000247, -0.00001032, -0.00000329, 0.00000921, 0.00000415, -0.00001509, -0.00002211, -0.00000880, 0.00000415, 0.00000178, -0.00000557, -0.00000530, -0.00000299, -0.00000743, -0.00001305, -0.00001239, },
		{ -0.00001041, -0.00001241, -0.00001512, -0.00002575, -0.00003899, -0.00003653, -0.00000709, 0.00003634, 0.00006416, 0.00005832, 0.00003247, 0.00001846, 0.00003006, 0.00004784, 0.00004615, 0.00002576, 0.00000726, -0.00000081, -0.00000896, -0.00002237, -0.00002904, -0.00001949, -0.00000320, 0.00000456, 0.00000202, -0.00000227, -0.00000383, -0.00000401, -0.00000364, -0.00000286, -0.00000313, -0.00000390, -0.00000194, 0.00000229, 0.00000385, 0.00000185, 0.00000132, 0.00000331, 0.00000279, -0.00000050, -0.00000037, 0.00000269, 0.00000018, -0.00000692, -0.00000590, 0.00000405, 0.00000504, -0.00000961, -0.00001882, -0.00000601, 0.00001170, 0.00001096, 0.00000070, 0.00000294, 0.00001042, 0.00000086, -0.00001619, -0.00001008, 0.00001436, 0.00001781, -0.00000791, -0.00002300, -0.00000596, 0.00001093, 0.00000015, -0.00001242, 0.00000320, 0.00002281, 0.00000897, -0.00001933, -0.00001537, 0.00001417, 0.00001894, -0.00000887, -0.00002169, 0.00000095, 0.00001843, 0.00000088, -0.00002025, -0.00000806, 0.00001796, 0.00001845, -0.00000064, -0.00000571, 0.00000383, 0.00000043, -0.00001741, -0.00002122, -0.00000242, 0.00001408, 0.00001147, 0.00000264, 0.00000166, 0.00000224, -0.00000377, -0.00000972, -0.00000779, -0.00000154, 0.00000431, 0.00001066, 0.00001377, 0.00000514, -0.00001149, -0.00001967, -0.00001377, -0.00000633, -0.00000494, -0.00000116, 0.00000828, 0.00000980, -0.00000354, -0.00001540, -0.00001065, -0.00000013, -0.00000133, -0.00000834, -0.00000430, 0.00000629, 0.00000374, -0.00001235, -0.00002028, -0.00001014, 0.00000236, 0.00000221, -0.00000451, -0.00000593, -0.00000374, -0.00000570, },
		{ -0.00000447, -0.00000386, -0.00000752, -0.00001226, -0.00001704, -0.00002441, -0.00003263, -0.00003138, -0.00000996, 0.00002743, 0.00005748, 0.00005759, 0.00003299, 0.00001484, 0.00002363, 0.00004357, 0.00004532, 0.00002572, 0.00000739, 0.00000269, -0.00000068, -0.00001392, -0.00002683, -0.00002426, -0.00000911, 0.00000326, 0.00000514, 0.00000061, -0.00000419, -0.00000600, -0.00000404, -0.00000123, -0.00000197, -0.00000487, -0.00000370, 0.00000184, 0.00000462, 0.00000234, 0.00000118, 0.00000334, 0.00000289, -0.00000111, -0.00000087, 0.00000398, 0.00000240, -0.00000700, -0.00000868, 0.00000279, 0.00000811, -0.00000584, -0.00001965, -0.00001056, 0.00000883, 0.00001139, 0.00000112, 0.00000232, 0.00001199, 0.00000494, -0.00001499, -0.00001438, 0.00001046, 0.00002042, -0.00000231, -0.00002220, -0.00000983, 0.00000921, 0.00000201, -0.00001281, -0.00000098, 0.00002153, 0.00001389, -0.00001534, -0.00001888, 0.00000849, 0.00002021, -0.00000257, -0.00002018, -0.00000406, 0.00001569, 0.00000483, -0.00001664, -0.00001141, 0.00001220, 0.00001841, 0.00000427, -0.00000288, 0.00000300, 0.00000067, -0.00001472, -0.00002093, -0.00000708, 0.00000917, 0.00001160, 0.00000642, 0.00000398, 0.00000174, -0.00000409, -0.00000810, -0.00000636, -0.00000297, 0.00000060, 0.00000796, 0.00001459, 0.00000868, -0.00000831, -0.00001890, -0.00001468, -0.00000728, -0.00000602, -0.00000329, 0.00000632, 0.00001065, -0.00000007, -0.00001318, -0.00001210, -0.00000264, -0.00000127, -0.00000658, -0.00000442, 0.00000405, 0.00000283, -0.00001054, -0.00001856, -0.00001078, 0.00000106, 0.00000244, -0.00000388, -0.00000672, },
		{ -0.00000338, -0.00000781, -0.00000552, -0.00000209, -0.00000419, -0.00001161, -0.00001928, -0.00002417, -0.00002694, -0.00002555, -0.00001153, 0.00001898, 0.00005030, 0.00005669, 0.00003431, 0.00001209, 0.00001715, 0.00003893, 0.00004489, 0.00002625, 0.00000639, 0.00000305, 0.00000502, -0.00000472, -0.00002112, -0.00002632, -0.00001556, -0.00000056, 0.00000690, 0.00000448, -0.00000299, -0.00000788, -0.00000554, -0.00000019, -0.00000019, -0.00000501, -0.00000548, 0.00000078, 0.00000506, 0.00000296, 0.00000124, 0.00000352, 0.00000318, -0.00000176, -0.00000194, 0.00000477, 0.00000504, -0.00000605, -0.00001128, 0.00000040, 0.00001022, -0.00000144, -0.00001898, -0.00001454, 0.00000531, 0.00001117, 0.00000141, 0.00000132, 0.00001283, 0.00000916, -0.00001237, -0.00001774, 0.00000558, 0.00002137, 0.00000322, -0.00002001, -0.00001297, 0.00000694, 0.00000361, -0.00001244, -0.00000496, 0.00001891, 0.00001774, -0.00001030, -0.00002062, 0.00000256, 0.00001948, 0.00000301, -0.00001709, -0.00000775, 0.00001198, 0.00000728, -0.00001251, -0.00001293, 0.00000675, 0.00001665, 0.00000785, 0.00000050, 0.00000316, 0.00000079, -0.00001243, -0.00001981, -0.00001040, 0.00000409, 0.00000997, 0.00000901, 0.00000695, 0.00000243, -0.00000447, -0.00000741, -0.00000494, -0.00000306, -0.00000204, 0.00000463, 0.00001369, 0.00001127, -0.00000458, -0.00001710, -0.00001519, -0.00000831, -0.00000695, -0.00000513, 0.00000400, 0.00001048, 0.00000286, -0.00001025, -0.00001240, -0.00000497, -0.00000207, -0.00000541, -0.00000409, 0.00000256, 0.00000188, -0.00000942, -0.00001725, -0.00001114, 0.00000023, 0.00000285, },
		{ -0.00000035, 0.00000359, -0.00000267, -0.00000907, -0.00000714, -0.00000069, -0.00000043, -0.00001013, -0.00002152, -0.00002513, -0.00002231, -0.00001923, -0.00001157, 0.00001126, 0.00004249, 0.00005523, 0.00003641, 0.00001059, 0.00001081, 0.00003345, 0.00004431, 0.00002778, 0.00000559, 0.00000114, 0.00000737, 0.00000344, -0.00001284, -0.00002483, -0.00002108, -0.00000646, 0.00000645, 0.00000846, -0.00000010, -0.00000893, -0.00000784, -0.00000016, 0.00000177, -0.00000427, -0.00000691, -0.00000071, 0.00000504, 0.00000351, 0.00000142, 0.00000388, 0.00000377, -0.00000228, -0.00000351, 0.00000481, 0.00000780, -0.00000402, -0.00001327, -0.00000287, 0.00001106, 0.00000309, -0.00001688, -0.00001757, 0.00000144, 0.00001031, 0.00000154, -0.00000001, 0.00001280, 0.00001315, -0.00000852, -0.00001975, 0.00000023, 0.00002062, 0.00000817, -0.00001668, -0.00001513, 0.00000434, 0.00000482, -0.00001138, -0.00000842, 0.00001522, 0.00002017, -0.00000478, -0.00002049, -0.00000295, 0.00001703, 0.00000732, -0.00001297, -0.00000985, 0.00000793, 0.00000817, -0.00000847, -0.00001283, 0.00000215, 0.00001368, 0.00000988, 0.00000389, 0.00000423, 0.00000106, -0.00001055, -0.00001821, -0.00001239, -0.00000059, 0.00000701, 0.00001005, 0.00000994, 0.00000425, -0.00000437, -0.00000742, -0.00000402, -0.00000233, -0.00000335, 0.00000145, 0.00001146, 0.00001247, -0.00000097, -0.00001439, -0.00001494, -0.00000920, -0.00000785, -0.00000675, 0.00000155, 0.00000949, 0.00000502, -0.00000707, -0.00001166, -0.00000673, -0.00000334, -0.00000494, -0.00000372, 0.00000164, 0.00000120, -0.00000863, -0.00001637, -0.00001159, },
		{ -0.00001576, -0.00001233, -0.00000104, 0.00000459, -0.00000146, -0.00001018, -0.00000936, -0.00000004, 0.00000350, -0.00000761, -0.00002329, -0.00002716, -0.00001911, -0.00001282, -0.00000999, 0.00000468, 0.00003412, 0.00005278, 0.00003895, 0.00001065, 0.00000511, 0.00002699, 0.00004288, 0.00003012, 0.00000604, -0.00000173, 0.00000638, 0.00000904, -0.00000347, -0.00001967, -0.00002424, -0.00001346, 0.00000336, 0.00001150, 0.00000414, -0.00000852, -0.00001038, -0.00000126, 0.00000338, -0.00000280, -0.00000770, -0.00000236, 0.00000453, 0.00000383, 0.00000160, 0.00000437, 0.00000474, -0.00000245, -0.00000541, 0.00000396, 0.00001029, -0.00000103, -0.00001429, -0.00000663, 0.00001050, 0.00000724, -0.00001359, -0.00001940, -0.00000245, 0.00000887, 0.00000147, -0.00000158, 0.00001188, 0.00001652, -0.00000374, -0.00002018, -0.00000502, 0.00001831, 0.00001209, -0.00001260, -0.00001619, 0.00000163, 0.00000556, -0.00000978, -0.00001109, 0.00001084, 0.00002100, 0.00000061, -0.00001868, -0.00000746, 0.00001334, 0.00001004, -0.00000843, -0.00001035, 0.00000411, 0.00000772, -0.00000504, -0.00001152, -0.00000124, 0.00001010, 0.00001038, 0.00000681, 0.00000598, 0.00000171, -0.00000899, -0.00001642, -0.00001320, -0.00000444, 0.00000325, 0.00000946, 0.00001230, 0.00000690, -0.00000342, -0.00000768, -0.00000382, -0.00000137, -0.00000345, -0.00000096, 0.00000852, 0.00001217, 0.00000190, -0.00001117, -0.00001378, -0.00000964, -0.00000865, -0.00000824, -0.00000093, 0.00000791, 0.00000637, -0.00000400, -0.00001016, -0.00000772, -0.00000465, -0.00000504, -0.00000363, 0.00000099, 0.00000087, -0.00000783, },
		{ 0.00000073, -0.00000682, -0.00001510, -0.00001333, -0.00000213, 0.00000557, 0.00000034, -0.00001076, -0.00001198, -0.00000042, 0.00000716, -0.00000406, -0.00002411, -0.00002985, -0.00001760, -0.00000690, -0.00000695, -0.00000030, 0.00002554, 0.00004901, 0.00004136, 0.00001232, 0.00000070, 0.00001981, 0.00003993, 0.00003257, 0.00000821, -0.00000410, 0.00000284, 0.00001109, 0.00000519, -0.00001151, -0.00002398, -0.00002015, -0.00000221, 0.00001260, 0.00000901, -0.00000635, -0.00001242, -0.00000337, 0.00000421, -0.00000094, -0.00000772, -0.00000383, 0.00000365, 0.00000379, 0.00000159, 0.00000486, 0.00000609, -0.00000207, -0.00000739, 0.00000218, 0.00001213, 0.00000265, -0.00001406, -0.00001039, 0.00000859, 0.00001056, -0.00000949, -0.00001990, -0.00000601, 0.00000698, 0.00000121, -0.00000329, 0.00001008, 0.00001898, 0.00000155, -0.00001896, -0.00000963, 0.00001475, 0.00001467, -0.00000820, -0.00001615, -0.00000092, 0.00000580, -0.00000783, -0.00001280, 0.00000621, 0.00002027, 0.00000533, -0.00001555, -0.00001058, 0.00000901, 0.00001108, -0.00000409, -0.00000946, 0.00000099, 0.00000627, -0.00000255, -0.00000948, -0.00000331, 0.00000648, 0.00000955, 0.00000888, 0.00000808, 0.00000282, -0.00000760, -0.00001466, -0.00001309, -0.00000721, -0.00000071, 0.00000741, 0.00001355, 0.00000986, -0.00000147, -0.00000770, -0.00000425, -0.00000072, -0.00000272, -0.00000226, 0.00000557, 0.00001063, 0.00000357, -0.00000802, -0.00001182, -0.00000935, -0.00000916, -0.00000962, -0.00000343, 0.00000588, 0.00000697, -0.00000128, -0.00000821, -0.00000798, -0.00000569, -0.00000544, -0.00000389, 0.00000030, },
		{ -0.00000440, -0.00000065, 0.00000053, -0.00000557, -0.00001412, -0.00001437, -0.00000377, 0.00000615, 0.00000259, -0.00001048, -0.00001458, -0.00000189, 0.00001003, 0.00000021, -0.00002363, -0.00003261, -0.00001773, -0.00000208, -0.00000288, -0.00000337, 0.00001733, 0.00004388, 0.00004298, 0.00001529, -0.00000185, 0.00001258, 0.00003518, 0.00003417, 0.00001189, -0.00000475, -0.00000186, 0.00000943, 0.00001144, -0.00000171, -0.00001988, -0.00002504, -0.00000951, 0.00001111, 0.00001349, -0.00000254, -0.00001330, -0.00000601, 0.00000400, 0.00000083, -0.00000703, -0.00000486, 0.00000263, 0.00000338, 0.00000125, 0.00000517, 0.00000772, -0.00000099, -0.00000913, -0.00000043, 0.00001299, 0.00000664, -0.00001251, -0.00001365, 0.00000556, 0.00001269, -0.00000503, -0.00001909, -0.00000894, 0.00000485, 0.00000077, -0.00000505, 0.00000753, 0.00002027, 0.00000689, -0.00001620, -0.00001315, 0.00001039, 0.00001579, -0.00000393, -0.00001512, -0.00000310, 0.00000556, -0.00000574, -0.00001351, 0.00000177, 0.00001817, 0.00000892, -0.00001160, -0.00001215, 0.00000465, 0.00001061, -0.00000046, -0.00000754, -0.00000112, 0.00000428, -0.00000110, -0.00000720, -0.00000412, 0.00000328, 0.00000775, 0.00000991, 0.00001016, 0.00000439, -0.00000621, -0.00001304, -0.00001236, -0.00000883, -0.00000432, 0.00000428, 0.00001337, 0.00001256, 0.00000136, -0.00000703, -0.00000502, -0.00000068, -0.00000170, -0.00000245, 0.00000323, 0.00000836, 0.00000391, -0.00000550, -0.00000941, -0.00000821, -0.00000909, -0.00001078, -0.00000597, 0.00000347, 0.00000693, 0.00000101, -0.00000608, -0.00000766, -0.00000631, -0.00000583, },
		{ -0.00000653, -0.00000599, -0.00000493, -0.00000186, 0.00000000, -0.00000429, -0.00001265, -0.00001511, -0.00000585, 0.00000601, 0.00000493, -0.00000923, -0.00001667, -0.00000420, 0.00001173, 0.00000466, -0.00002174, -0.00003479, -0.00001914, 0.00000118, 0.00000155, -0.00000450, 0.00001014, 0.00003764, 0.00004322, 0.00001892, -0.00000226, 0.00000619, 0.00002880, 0.00003399, 0.00001630, -0.00000300, -0.00000612, 0.00000470, 0.00001405, 0.00000797, -0.00001229, -0.00002683, -0.00001728, 0.00000686, 0.00001654, 0.00000239, -0.00001254, -0.00000856, 0.00000277, 0.00000209, -0.00000591, -0.00000529, 0.00000175, 0.00000274, 0.00000050, 0.00000508, 0.00000945, 0.00000085, -0.00001029, -0.00000363, 0.00001266, 0.00001049, -0.00000972, -0.00001600, 0.00000180, 0.00001346, -0.00000070, -0.00001716, -0.00001101, 0.00000269, 0.00000023, -0.00000674, 0.00000438, 0.00002030, 0.00001180, -0.00001216, -0.00001524, 0.00000573, 0.00001548, -0.00000019, -0.00001332, -0.00000476, 0.00000495, -0.00000371, -0.00001329, -0.00000211, 0.00001506, 0.00001115, -0.00000739, -0.00001220, 0.00000080, 0.00000895, 0.00000214, -0.00000506, -0.00000214, 0.00000219, -0.00000066, -0.00000512, -0.00000393, 0.00000085, 0.00000540, 0.00000986, 0.00001188, 0.00000630, -0.00000467, -0.00001159, -0.00001131, -0.00000940, -0.00000718, 0.00000058, 0.00001175, 0.00001442, 0.00000471, -0.00000542, -0.00000568, -0.00000130, -0.00000090, -0.00000181, 0.00000188, 0.00000601, 0.00000308, -0.00000405, -0.00000707, -0.00000633, -0.00000820, -0.00001147, -0.00000850, 0.00000067, 0.00000628, 0.00000285, -0.00000394, -0.00000697, },
		{ -0.00000185, -0.00000608, -0.00000648, -0.00000583, -0.00000521, -0.00000320, -0.00000100, -0.00000327, -0.00001076, -0.00001524, -0.00000809, 0.00000499, 0.00000689, -0.00000716, -0.00001781, -0.00000688, 0.00001208, 0.00000866, -0.00001869, -0.00003582, -0.00002122, 0.00000271, 0.00000558, -0.00000398, 0.00000455, 0.00003089, 0.00004175, 0.00002236, -0.00000065, 0.00000150, 0.00002150, 0.00003144, 0.00002027, 0.00000108, -0.00000845, -0.00000175, 0.00001255, 0.00001568, -0.00000230, -0.00002477, -0.00002404, 0.00000032, 0.00001729, 0.00000756, -0.00001005, -0.00001032, 0.00000082, 0.00000251, -0.00000473, -0.00000515, 0.00000125, 0.00000207, -0.00000066, 0.00000441, 0.00001101, 0.00000339, -0.00001059, -0.00000705, 0.00001107, 0.00001371, -0.00000597, -0.00001711, -0.00000222, 0.00001288, 0.00000306, -0.00001441, -0.00001213, 0.00000074, -0.00000033, -0.00000826, 0.00000085, 0.00001903, 0.00001588, -0.00000725, -0.00001575, 0.00000130, 0.00001392, 0.00000273, -0.00001104, -0.00000582, 0.00000408, -0.00000190, -0.00001228, -0.00000514, 0.00001136, 0.00001194, -0.00000346, -0.00001101, -0.00000216, 0.00000655, 0.00000356, -0.00000249, -0.00000215, 0.00000038, -0.00000099, -0.00000352, -0.00000307, -0.00000068, 0.00000293, 0.00000884, 0.00001297, 0.00000833, -0.00000289, -0.00001024, -0.00001018, -0.00000914, -0.00000903, -0.00000313, 0.00000889, 0.00001504, 0.00000806, -0.00000283, -0.00000577, -0.00000238, -0.00000070, -0.00000083, 0.00000160, 0.00000418, 0.00000155, -0.00000384, -0.00000536, -0.00000406, -0.00000637, -0.00001140, -0.00001082, -0.00000251, 0.00000500, 0.00000426, },
		{ 0.00000301, 0.00000517, 0.00000019, -0.00000507, -0.00000633, -0.00000544, -0.00000506, -0.00000436, -0.00000246, -0.00000281, -0.00000870, -0.00001457, -0.00001005, 0.00000319, 0.00000803, -0.00000469, -0.00001778, -0.00000934, 0.00001123, 0.00001164, -0.00001501, -0.00003544, -0.00002318, 0.00000272, 0.00000853, -0.00000245, 0.00000086, 0.00002441, 0.00003863, 0.00002476, 0.00000239, -0.00000085, 0.00001430, 0.00002648, 0.00002257, 0.00000675, -0.00000789, -0.00000820, 0.00000731, 0.00001991, 0.00000840, -0.00001883, -0.00002834, -0.00000751, 0.00001531, 0.00001192, -0.00000614, -0.00001074, -0.00000129, 0.00000203, -0.00000391, -0.00000465, 0.00000128, 0.00000167, -0.00000206, 0.00000304, 0.00001209, 0.00000643, -0.00000981, -0.00001027, 0.00000833, 0.00001591, -0.00000164, -0.00001684, -0.00000600, 0.00001112, 0.00000592, -0.00001121, -0.00001231, -0.00000084, -0.00000080, -0.00000948, -0.00000284, 0.00001657, 0.00001878, -0.00000194, -0.00001469, -0.00000243, 0.00001146, 0.00000463, -0.00000859, -0.00000626, 0.00000310, -0.00000043, -0.00001070, -0.00000717, 0.00000753, 0.00001142, -0.00000024, -0.00000896, -0.00000402, 0.00000390, 0.00000385, -0.00000025, -0.00000138, -0.00000088, -0.00000183, -0.00000257, -0.00000190, -0.00000131, 0.00000072, 0.00000712, 0.00001326, 0.00001023, -0.00000088, -0.00000889, -0.00000911, -0.00000834, -0.00000982, -0.00000635, 0.00000518, 0.00001423, 0.00001085, 0.00000050, -0.00000493, -0.00000350, -0.00000123, -0.00000002, 0.00000215, 0.00000330, -0.00000005, -0.00000473, -0.00000475, -0.00000194, -0.00000373, -0.00001026, -0.00001264, -0.00000599, },
		{ -0.00001354, -0.00000954, 0.00000025, 0.00000547, 0.00000218, -0.00000390, -0.00000619, -0.00000500, -0.00000447, -0.00000507, -0.00000414, -0.00000311, -0.00000687, -0.00001316, -0.00001131, 0.00000094, 0.00000810, -0.00000239, -0.00001667, -0.00001099, 0.00000966, 0.00001327, -0.00001142, -0.00003373, -0.00002432, 0.00000176, 0.00001000, -0.00000073, -0.00000101, 0.00001898, 0.00003435, 0.00002550, 0.00000597, -0.00000071, 0.00000835, 0.00001966, 0.00002224, 0.00001274, -0.00000419, -0.00001291, -0.00000049, 0.00001980, 0.00001794, -0.00000977, -0.00002911, -0.00001525, 0.00001071, 0.00001450, -0.00000151, -0.00000955, -0.00000293, 0.00000085, -0.00000377, -0.00000413, 0.00000183, 0.00000179, -0.00000343, 0.00000098, 0.00001239, 0.00000965, -0.00000788, -0.00001285, 0.00000472, 0.00001681, 0.00000275, -0.00001525, -0.00000908, 0.00000852, 0.00000770, -0.00000797, -0.00001168, -0.00000193, -0.00000108, -0.00001032, -0.00000643, 0.00001310, 0.00002029, 0.00000328, -0.00001225, -0.00000511, 0.00000849, 0.00000547, -0.00000626, -0.00000617, 0.00000213, 0.00000064, -0.00000880, -0.00000819, 0.00000398, 0.00000986, 0.00000199, -0.00000651, -0.00000478, 0.00000143, 0.00000324, 0.00000139, -0.00000013, -0.00000148, -0.00000285, -0.00000228, -0.00000075, -0.00000117, -0.00000095, 0.00000498, 0.00001272, 0.00001178, 0.00000131, -0.00000745, -0.00000817, -0.00000728, -0.00000966, -0.00000871, 0.00000116, 0.00001202, 0.00001258, 0.00000413, -0.00000302, -0.00000419, -0.00000236, 0.00000017, 0.00000308, 0.00000347, -0.00000111, -0.00000626, -0.00000541, -0.00000057, -0.00000065, -0.00000792, },
		{ 0.00000227, -0.00000448, -0.00001314, -0.00001279, -0.00000326, 0.00000495, 0.00000402, -0.00000250, -0.00000607, -0.00000472, -0.00000359, -0.00000511, -0.00000568, -0.00000412, -0.00000566, -0.00001131, -0.00001160, -0.00000124, 0.00000711, -0.00000080, -0.00001486, -0.00001145, 0.00000803, 0.00001355, -0.00000864, -0.00003115, -0.00002415, 0.00000060, 0.00000996, 0.00000036, -0.00000153, 0.00001518, 0.00002969, 0.00002435, 0.00000905, 0.00000155, 0.00000463, 0.00001203, 0.00001888, 0.00001755, 0.00000212, -0.00001454, -0.00000911, 0.00001536, 0.00002454, 0.00000099, -0.00002594, -0.00002140, 0.00000420, 0.00001465, 0.00000289, -0.00000688, -0.00000351, -0.00000058, -0.00000443, -0.00000399, 0.00000269, 0.00000258, -0.00000445, -0.00000160, 0.00001169, 0.00001266, -0.00000491, -0.00001441, 0.00000064, 0.00001631, 0.00000672, -0.00001255, -0.00001116, 0.00000550, 0.00000836, -0.00000502, -0.00001046, -0.00000248, -0.00000108, -0.00001068, -0.00000968, 0.00000890, 0.00002030, 0.00000793, -0.00000875, -0.00000652, 0.00000545, 0.00000535, -0.00000431, -0.00000568, 0.00000130, 0.00000132, -0.00000682, -0.00000829, 0.00000102, 0.00000763, 0.00000314, -0.00000409, -0.00000458, -0.00000054, 0.00000202, 0.00000229, 0.00000126, -0.00000143, -0.00000378, -0.00000255, 0.00000014, -0.00000048, -0.00000195, 0.00000277, 0.00001142, 0.00001275, 0.00000352, -0.00000582, -0.00000733, -0.00000621, -0.00000880, -0.00001003, -0.00000266, 0.00000872, 0.00001293, 0.00000750, -0.00000017, -0.00000402, -0.00000371, -0.00000045, 0.00000383, 0.00000451, -0.00000116, -0.00000780, -0.00000722, -0.00000045, },
		{ -0.00000974, -0.00000184, 0.00000437, -0.00000032, -0.00001121, -0.00001521, -0.00000730, 0.00000343, 0.00000553, -0.00000079, -0.00000585, -0.00000474, -0.00000267, -0.00000447, -0.00000669, -0.00000560, -0.00000535, -0.00000949, -0.00001091, -0.00000280, 0.00000540, -0.00000035, -0.00001299, -0.00001068, 0.00000705, 0.00001289, -0.00000717, -0.00002843, -0.00002258, 0.00000011, 0.00000883, 0.00000017, -0.00000151, 0.00001319, 0.00002555, 0.00002156, 0.00001070, 0.00000511, 0.00000367, 0.00000492, 0.00001274, 0.00001987, 0.00000979, -0.00001242, -0.00001665, 0.00000744, 0.00002698, 0.00001162, -0.00001915, -0.00002469, -0.00000304, 0.00001219, 0.00000606, -0.00000324, -0.00000265, -0.00000167, -0.00000576, -0.00000457, 0.00000351, 0.00000405, -0.00000479, -0.00000440, 0.00000991, 0.00001502, -0.00000117, -0.00001468, -0.00000340, 0.00001448, 0.00000982, -0.00000913, -0.00001204, 0.00000249, 0.00000805, -0.00000264, -0.00000894, -0.00000256, -0.00000077, -0.00001051, -0.00001237, 0.00000430, 0.00001888, 0.00001160, -0.00000461, -0.00000659, 0.00000274, 0.00000447, -0.00000291, -0.00000496, 0.00000066, 0.00000166, -0.00000496, -0.00000766, -0.00000114, 0.00000512, 0.00000326, -0.00000207, -0.00000371, -0.00000182, 0.00000054, 0.00000246, 0.00000251, -0.00000086, -0.00000442, -0.00000320, 0.00000061, 0.00000048, -0.00000223, 0.00000077, 0.00000955, 0.00001303, 0.00000559, -0.00000400, -0.00000650, -0.00000528, -0.00000751, -0.00001030, -0.00000580, 0.00000477, 0.00001180, 0.00001006, 0.00000327, -0.00000274, -0.00000479, -0.00000184, 0.00000391, 0.00000596, -0.00000003, -0.00000868, },
		{ 0.00000211, -0.00000835, -0.00001228, -0.00000466, 0.00000502, 0.00000392, -0.00000775, -0.00001629, -0.00001147, 0.00000080, 0.00000639, 0.00000117, -0.00000533, -0.00000506, -0.00000205, -0.00000331, -0.00000687, -0.00000710, -0.00000600, -0.00000822, -0.00000949, -0.00000332, 0.00000353, -0.00000115, -0.00001173, -0.00000900, 0.00000726, 0.00001195, -0.00000719, -0.00002640, -0.00001993, 0.00000094, 0.00000736, -0.00000154, -0.00000185, 0.00001276, 0.00002276, 0.00001781, 0.00001037, 0.00000885, 0.00000547, -0.00000038, 0.00000473, 0.00001882, 0.00001723, -0.00000675, -0.00002139, -0.00000237, 0.00002480, 0.00002024, -0.00000984, -0.00002434, -0.00000959, 0.00000752, 0.00000721, 0.00000055, -0.00000034, -0.00000183, -0.00000738, -0.00000603, 0.00000388, 0.00000602, -0.00000420, -0.00000702, 0.00000714, 0.00001637, 0.00000295, -0.00001357, -0.00000691, 0.00001160, 0.00001174, -0.00000545, -0.00001176, -0.00000011, 0.00000702, -0.00000096, -0.00000737, -0.00000227, -0.00000017, -0.00000978, -0.00001434, -0.00000036, 0.00001618, 0.00001401, -0.00000030, -0.00000544, 0.00000071, 0.00000311, -0.00000214, -0.00000420, 0.00000023, 0.00000174, -0.00000336, -0.00000653, -0.00000243, 0.00000273, 0.00000256, -0.00000068, -0.00000249, -0.00000239, -0.00000090, 0.00000201, 0.00000341, 0.00000004, -0.00000465, -0.00000402, 0.00000062, 0.00000148, -0.00000188, -0.00000077, 0.00000735, 0.00001257, 0.00000731, -0.00000202, -0.00000558, -0.00000454, -0.00000609, -0.00000967, -0.00000796, 0.00000072, 0.00000936, 0.00001138, 0.00000677, -0.00000035, -0.00000510, -0.00000364, 0.00000302, 0.00000721, },
		{ 0.00000114, 0.00000768, 0.00000481, -0.00000654, -0.00001405, -0.00000847, 0.00000384, 0.00000745, -0.00000308, -0.00001562, -0.00001518, -0.00000283, 0.00000626, 0.00000320, -0.00000432, -0.00000550, -0.00000195, -0.00000201, -0.00000612, -0.00000811, -0.00000736, -0.00000792, -0.00000786, -0.00000265, 0.00000216, -0.00000293, -0.00001166, -0.00000702, 0.00000890, 0.00001155, -0.00000844, -0.00002577, -0.00001689, 0.00000342, 0.00000647, -0.00000457, -0.00000340, 0.00001321, 0.00002178, 0.00001403, 0.00000797, 0.00001160, 0.00000939, -0.00000289, -0.00000381, 0.00001424, 0.00002277, 0.00000149, -0.00002225, -0.00001214, 0.00001844, 0.00002531, 0.00000039, -0.00002028, -0.00001408, 0.00000154, 0.00000593, 0.00000352, 0.00000308, -0.00000059, -0.00000870, -0.00000829, 0.00000340, 0.00000811, -0.00000261, -0.00000902, 0.00000363, 0.00001644, 0.00000694, -0.00001118, -0.00000944, 0.00000808, 0.00001235, -0.00000199, -0.00001049, -0.00000200, 0.00000560, -0.00000001, -0.00000599, -0.00000180, 0.00000066, -0.00000854, -0.00001545, -0.00000474, 0.00001249, 0.00001499, 0.00000370, -0.00000330, -0.00000043, 0.00000159, -0.00000195, -0.00000356, -0.00000001, 0.00000166, -0.00000210, -0.00000515, -0.00000291, 0.00000073, 0.00000134, -0.00000002, -0.00000125, -0.00000235, -0.00000206, 0.00000114, 0.00000384, 0.00000107, -0.00000446, -0.00000484, 0.00000018, 0.00000229, -0.00000106, -0.00000172, 0.00000508, 0.00001143, 0.00000851, 0.00000001, -0.00000449, -0.00000396, -0.00000478, -0.00000843, -0.00000902, -0.00000292, 0.00000598, 0.00001122, 0.00000974, 0.00000289, -0.00000427, -0.00000535, },
		{ -0.00000217, -0.00000638, -0.00000144, 0.00000694, 0.00000738, -0.00000337, -0.00001439, -0.00001253, 0.00000079, 0.00000950, 0.00000214, -0.00001303, -0.00001777, -0.00000712, 0.00000489, 0.00000495, -0.00000273, -0.00000577, -0.00000243, -0.00000099, -0.00000460, -0.00000821, -0.00000898, -0.00000877, -0.00000667, -0.00000098, 0.00000188, -0.00000515, -0.00001307, -0.00000553, 0.00001179, 0.00001243, -0.00001027, -0.00002693, -0.00001440, 0.00000741, 0.00000703, -0.00000826, -0.00000666, 0.00001360, 0.00002267, 0.00001123, 0.00000396, 0.00001245, 0.00001438, -0.00000221, -0.00001134, 0.00000669, 0.00002511, 0.00001074, -0.00001891, -0.00001992, 0.00000915, 0.00002594, 0.00000967, -0.00001319, -0.00001548, -0.00000447, 0.00000233, 0.00000479, 0.00000689, 0.00000218, -0.00000911, -0.00001103, 0.00000182, 0.00000985, -0.00000013, -0.00001004, -0.00000018, 0.00001515, 0.00001031, -0.00000782, -0.00001071, 0.00000442, 0.00001171, 0.00000085, -0.00000855, -0.00000306, 0.00000413, 0.00000034, -0.00000495, -0.00000133, 0.00000159, -0.00000688, -0.00001566, -0.00000851, 0.00000817, 0.00001454, 0.00000698, -0.00000051, -0.00000059, 0.00000023, -0.00000225, -0.00000315, -0.00000013, 0.00000149, -0.00000120, -0.00000374, -0.00000273, -0.00000068, -0.00000010, -0.00000003, -0.00000025, -0.00000187, -0.00000282, 0.00000007, 0.00000381, 0.00000203, -0.00000391, -0.00000548, -0.00000058, 0.00000279, 0.00000005, -0.00000204, 0.00000299, 0.00000975, 0.00000908, 0.00000193, -0.00000319, -0.00000343, -0.00000370, -0.00000688, -0.00000906, -0.00000573, 0.00000218, 0.00000962, 0.00001165, 0.00000649, },
		{ 0.00001215, 0.00000984, 0.00000104, -0.00000623, -0.00000419, 0.00000487, 0.00000910, 0.00000071, -0.00001289, -0.00001594, -0.00000375, 0.00000949, 0.00000705, -0.00000872, -0.00001864, -0.00001149, 0.00000218, 0.00000596, -0.00000067, -0.00000554, -0.00000335, -0.00000061, -0.00000270, -0.00000715, -0.00001023, -0.00001065, -0.00000648, 0.00000121, 0.00000302, -0.00000706, -0.00001587, -0.00000531, 0.00001536, 0.00001505, -0.00001181, -0.00002981, -0.00001335, 0.00001231, 0.00000961, -0.00001165, -0.00001168, 0.00001293, 0.00002495, 0.00001020, -0.00000079, 0.00001093, 0.00001915, 0.00000142, -0.00001652, -0.00000263, 0.00002356, 0.00001918, -0.00001195, -0.00002419, -0.00000125, 0.00002209, 0.00001629, -0.00000437, -0.00001335, -0.00000912, -0.00000292, 0.00000379, 0.00001018, 0.00000626, -0.00000808, -0.00001368, -0.00000089, 0.00001076, 0.00000290, -0.00000984, -0.00000382, 0.00001264, 0.00001263, -0.00000392, -0.00001060, 0.00000112, 0.00001005, 0.00000277, -0.00000633, -0.00000327, 0.00000291, 0.00000028, -0.00000431, -0.00000100, 0.00000248, -0.00000493, -0.00001499, -0.00001144, 0.00000361, 0.00001278, 0.00000925, 0.00000252, 0.00000018, -0.00000070, -0.00000285, -0.00000305, -0.00000020, 0.00000130, -0.00000062, -0.00000249, -0.00000210, -0.00000145, -0.00000145, -0.00000056, 0.00000036, -0.00000118, -0.00000315, -0.00000100, 0.00000336, 0.00000278, -0.00000309, -0.00000587, -0.00000155, 0.00000288, 0.00000122, -0.00000175, 0.00000131, 0.00000773, 0.00000896, 0.00000357, -0.00000171, -0.00000283, -0.00000288, -0.00000532, -0.00000829, -0.00000749, -0.00000149, 0.00000686, },
		{ -0.00000454, 0.00000341, 0.00001112, 0.00001231, 0.00000493, -0.00000464, -0.00000642, 0.00000167, 0.00000940, 0.00000494, -0.00000952, -0.00001784, -0.00000902, 0.00000719, 0.00001070, -0.00000324, -0.00001739, -0.00001518, -0.00000168, 0.00000579, 0.00000154, -0.00000455, -0.00000437, -0.00000109, -0.00000095, -0.00000500, -0.00001053, -0.00001310, -0.00000766, 0.00000319, 0.00000557, -0.00000788, -0.00001956, -0.00000686, 0.00001874, 0.00001943, -0.00001211, -0.00003391, -0.00001440, 0.00001716, 0.00001434, -0.00001371, -0.00001801, 0.00001042, 0.00002775, 0.00001131, -0.00000522, 0.00000715, 0.00002247, 0.00000717, -0.00001845, -0.00001215, 0.00001821, 0.00002514, -0.00000270, -0.00002415, -0.00001076, 0.00001459, 0.00001900, 0.00000446, -0.00000799, -0.00001126, -0.00000875, 0.00000039, 0.00001201, 0.00001107, -0.00000530, -0.00001556, -0.00000447, 0.00001045, 0.00000600, -0.00000839, -0.00000676, 0.00000923, 0.00001361, -0.00000000, -0.00000921, -0.00000138, 0.00000776, 0.00000365, -0.00000422, -0.00000280, 0.00000213, 0.00000004, -0.00000403, -0.00000093, 0.00000316, -0.00000288, -0.00001356, -0.00001335, -0.00000080, 0.00000997, 0.00001030, 0.00000539, 0.00000171, -0.00000103, -0.00000352, -0.00000323, -0.00000030, 0.00000112, -0.00000030, -0.00000148, -0.00000122, -0.00000161, -0.00000249, -0.00000140, 0.00000053, -0.00000048, -0.00000312, -0.00000194, 0.00000261, 0.00000325, -0.00000214, -0.00000594, -0.00000259, 0.00000257, 0.00000228, -0.00000097, 0.00000017, 0.00000561, 0.00000820, 0.00000477, -0.00000015, -0.00000207, -0.00000226, -0.00000396, -0.00000704, -0.00000818, },
		{ -0.00000562, -0.00000794, -0.00000663, -0.00000016, 0.00000875, 0.00001343, 0.00000888, -0.00000162, -0.00000752, -0.00000212, 0.00000796, 0.00000847, -0.00000468, -0.00001761, -0.00001402, 0.00000283, 0.00001228, 0.00000250, -0.00001399, -0.00001740, -0.00000617, 0.00000413, 0.00000341, -0.00000276, -0.00000504, -0.00000237, 0.00000013, -0.00000211, -0.00000947, -0.00001544, -0.00001026, 0.00000423, 0.00000910, -0.00000705, -0.00002334, -0.00001031, 0.00002102, 0.00002516, -0.00001046, -0.00003832, -0.00001779, 0.00002090, 0.00002082, -0.00001359, -0.00002475, 0.00000569, 0.00003001, 0.00001441, -0.00000827, 0.00000173, 0.00002349, 0.00001377, -0.00001690, -0.00002024, 0.00000991, 0.00002742, 0.00000705, -0.00001988, -0.00001756, 0.00000496, 0.00001733, 0.00001153, -0.00000035, -0.00001015, -0.00001383, -0.00000500, 0.00001165, 0.00001574, -0.00000084, -0.00001604, -0.00000841, 0.00000875, 0.00000860, -0.00000589, -0.00000862, 0.00000537, 0.00001318, 0.00000341, -0.00000682, -0.00000280, 0.00000529, 0.00000354, -0.00000257, -0.00000190, 0.00000189, -0.00000013, -0.00000399, -0.00000116, 0.00000353, -0.00000092, -0.00001154, -0.00001420, -0.00000469, 0.00000646, 0.00001011, 0.00000772, 0.00000373, -0.00000069, -0.00000406, -0.00000365, -0.00000051, 0.00000097, -0.00000015, -0.00000078, -0.00000032, -0.00000128, -0.00000309, -0.00000230, 0.00000031, 0.00000008, -0.00000283, -0.00000266, 0.00000167, 0.00000339, -0.00000116, -0.00000570, -0.00000356, 0.00000188, 0.00000307, 0.00000013, -0.00000036, 0.00000364, 0.00000691, 0.00000539, 0.00000136, -0.00000110, -0.00000170, -0.00000289, },
		{ -0.00000107, -0.00000209, -0.00000432, -0.00000707, -0.00000764, -0.00000330, 0.00000543, 0.00001296, 0.00001220, 0.00000249, -0.00000701, -0.00000575, 0.00000487, 0.00001050, 0.00000084, -0.00001502, -0.00001771, -0.00000291, 0.00001131, 0.00000745, -0.00000884, -0.00001752, -0.00001055, 0.00000097, 0.00000434, -0.00000036, -0.00000492, -0.00000413, 0.00000014, 0.00000093, -0.00000692, -0.00001694, -0.00001391, 0.00000374, 0.00001290, -0.00000437, -0.00002631, -0.00001534, 0.00002141, 0.00003140, -0.00000652, -0.00004197, -0.00002322, 0.00002258, 0.00002820, -0.00001080, -0.00003080, -0.00000110, 0.00003071, 0.00001886, -0.00000922, -0.00000431, 0.00002186, 0.00001985, -0.00001232, -0.00002561, 0.00000007, 0.00002559, 0.00001543, -0.00001231, -0.00002043, -0.00000487, 0.00001163, 0.00001538, 0.00000808, -0.00000571, -0.00001688, -0.00001150, 0.00000874, 0.00001928, 0.00000484, -0.00001464, -0.00001202, 0.00000574, 0.00001019, -0.00000276, -0.00000918, 0.00000162, 0.00001147, 0.00000588, -0.00000386, -0.00000303, 0.00000308, 0.00000265, -0.00000160, -0.00000088, 0.00000214, -0.00000005, -0.00000401, -0.00000163, 0.00000350, 0.00000076, -0.00000916, -0.00001402, -0.00000777, 0.00000263, 0.00000877, 0.00000924, 0.00000591, 0.00000029, -0.00000428, -0.00000418, -0.00000085, 0.00000081, -0.00000012, -0.00000036, 0.00000046, -0.00000065, -0.00000323, -0.00000307, -0.00000017, 0.00000042, -0.00000241, -0.00000316, 0.00000065, 0.00000323, -0.00000025, -0.00000519, -0.00000437, 0.00000092, 0.00000348, 0.00000135, -0.00000030, 0.00000201, 0.00000529, 0.00000539, 0.00000263, 0.00000006, },
		{ 0.00000349, 0.00000130, -0.00000026, -0.00000146, -0.00000331, -0.00000592, -0.00000765, -0.00000558, 0.00000175, 0.00001093, 0.00001427, 0.00000708, -0.00000475, -0.00000845, 0.00000059, 0.00001050, 0.00000606, -0.00001033, -0.00001920, -0.00000898, 0.00000778, 0.00001057, -0.00000274, -0.00001524, -0.00001390, -0.00000336, 0.00000383, 0.00000218, -0.00000376, -0.00000587, -0.00000102, 0.00000344, -0.00000315, -0.00001697, -0.00001795, 0.00000144, 0.00001608, -0.00000006, -0.00002765, -0.00002121, 0.00001947, 0.00003711, -0.00000045, -0.00004387, -0.00002994, 0.00002160, 0.00003533, -0.00000543, -0.00003504, -0.00000929, 0.00002911, 0.00002365, -0.00000782, -0.00000981, 0.00001784, 0.00002422, -0.00000571, -0.00002749, -0.00000960, 0.00002000, 0.00002088, -0.00000300, -0.00001892, -0.00001297, 0.00000307, 0.00001515, 0.00001559, 0.00000143, -0.00001697, -0.00001788, 0.00000343, 0.00002080, 0.00001098, -0.00001123, -0.00001457, 0.00000181, 0.00001041, 0.00000045, -0.00000843, -0.00000152, 0.00000883, 0.00000715, -0.00000083, -0.00000216, 0.00000151, 0.00000132, -0.00000137, -0.00000004, 0.00000273, 0.00000038, -0.00000393, -0.00000225, 0.00000306, 0.00000200, -0.00000667, -0.00001294, -0.00000988, -0.00000112, 0.00000650, 0.00000979, 0.00000792, 0.00000178, -0.00000406, -0.00000469, -0.00000132, 0.00000062, -0.00000013, -0.00000018, 0.00000102, 0.00000010, -0.00000295, -0.00000357, -0.00000076, 0.00000054, -0.00000196, -0.00000346, -0.00000037, 0.00000280, 0.00000050, -0.00000446, -0.00000494, -0.00000022, 0.00000347, 0.00000253, 0.00000027, 0.00000087, 0.00000354, 0.00000479, },
		{ 0.00000191, 0.00000371, 0.00000381, 0.00000248, 0.00000079, -0.00000082, -0.00000261, -0.00000482, -0.00000690, -0.00000676, -0.00000171, 0.00000767, 0.00001468, 0.00001139, -0.00000094, -0.00000955, -0.00000409, 0.00000835, 0.00000996, -0.00000427, -0.00001798, -0.00001418, 0.00000219, 0.00001109, 0.00000319, -0.00001069, -0.00001539, -0.00000820, 0.00000162, 0.00000422, -0.00000158, -0.00000700, -0.00000317, 0.00000484, 0.00000129, -0.00001518, -0.00002154, -0.00000250, 0.00001784, 0.00000523, -0.00002686, -0.00002693, 0.00001526, 0.00004127, 0.00000707, -0.00004333, -0.00003686, 0.00001789, 0.00004104, 0.00000192, -0.00003665, -0.00001781, 0.00002500, 0.00002764, -0.00000437, -0.00001376, 0.00001221, 0.00002608, 0.00000156, -0.00002584, -0.00001753, 0.00001174, 0.00002244, 0.00000618, -0.00001349, -0.00001776, -0.00000664, 0.00001076, 0.00002056, 0.00001006, -0.00001369, -0.00002282, -0.00000362, 0.00001973, 0.00001657, -0.00000605, -0.00001542, -0.00000242, 0.00000913, 0.00000315, -0.00000662, -0.00000365, 0.00000571, 0.00000714, 0.00000179, -0.00000047, 0.00000081, -0.00000009, -0.00000183, 0.00000038, 0.00000344, 0.00000115, -0.00000359, -0.00000285, 0.00000228, 0.00000270, -0.00000432, -0.00001119, -0.00001092, -0.00000444, 0.00000361, 0.00000931, 0.00000948, 0.00000356, -0.00000337, -0.00000503, -0.00000187, 0.00000039, -0.00000015, -0.00000016, 0.00000131, 0.00000081, -0.00000237, -0.00000373, -0.00000130, 0.00000051, -0.00000153, -0.00000359, -0.00000136, 0.00000213, 0.00000106, -0.00000359, -0.00000522, -0.00000141, 0.00000305, 0.00000348, 0.00000120, 0.00000032, },
		{ 0.00000230, 0.00000034, 0.00000061, 0.00000232, 0.00000354, 0.00000338, 0.00000202, -0.00000002, -0.00000214, -0.00000400, -0.00000578, -0.00000683, -0.00000439, 0.00000372, 0.00001332, 0.00001469, 0.00000389, -0.00000864, -0.00000827, 0.00000434, 0.00001175, 0.00000212, -0.00001408, -0.00001737, -0.00000447, 0.00000868, 0.00000774, -0.00000452, -0.00001440, -0.00001263, -0.00000224, 0.00000508, 0.00000129, -0.00000706, -0.00000579, 0.00000479, 0.00000563, -0.00001163, -0.00002383, -0.00000753, 0.00001762, 0.00001059, -0.00002389, -0.00003148, 0.00000933, 0.00004313, 0.00001504, -0.00004016, -0.00004279, 0.00001194, 0.00004439, 0.00001020, -0.00003530, -0.00002546, 0.00001870, 0.00002986, 0.00000035, -0.00001552, 0.00000606, 0.00002525, 0.00000811, -0.00002129, -0.00002253, 0.00000237, 0.00001994, 0.00001341, -0.00000536, -0.00001839, -0.00001558, 0.00000295, 0.00002185, 0.00001860, -0.00000729, -0.00002518, -0.00001133, 0.00001590, 0.00002063, 0.00000029, -0.00001419, -0.00000619, 0.00000655, 0.00000484, -0.00000420, -0.00000459, 0.00000264, 0.00000600, 0.00000361, 0.00000166, 0.00000106, -0.00000119, -0.00000275, 0.00000025, 0.00000404, 0.00000218, -0.00000290, -0.00000328, 0.00000127, 0.00000284, -0.00000234, -0.00000904, -0.00001096, -0.00000704, 0.00000045, 0.00000791, 0.00001035, 0.00000538, -0.00000225, -0.00000510, -0.00000244, 0.00000010, -0.00000018, -0.00000023, 0.00000137, 0.00000135, -0.00000165, -0.00000357, -0.00000167, 0.00000041, -0.00000113, -0.00000360, -0.00000229, 0.00000126, 0.00000138, -0.00000264, -0.00000519, -0.00000253, 0.00000227, 0.00000407, },
		{ 0.00000122, 0.00000425, 0.00000337, 0.00000085, -0.00000021, 0.00000086, 0.00000272, 0.00000385, 0.00000331, 0.00000105, -0.00000173, -0.00000357, -0.00000464, -0.00000601, -0.00000595, -0.00000026, 0.00001040, 0.00001637, 0.00000896, -0.00000575, -0.00001106, -0.00000077, 0.00001101, 0.00000764, -0.00000805, -0.00001772, -0.00001091, 0.00000358, 0.00000987, 0.00000225, -0.00001078, -0.00001562, -0.00000724, 0.00000423, 0.00000424, -0.00000580, -0.00000824, 0.00000328, 0.00000910, -0.00000676, -0.00002422, -0.00001277, 0.00001530, 0.00001503, -0.00001917, -0.00003406, 0.00000261, 0.00004236, 0.00002229, -0.00003471, -0.00004673, 0.00000468, 0.00004489, 0.00001817, -0.00003122, -0.00003114, 0.00001102, 0.00002967, 0.00000528, -0.00001499, 0.00000052, 0.00002211, 0.00001281, -0.00001497, -0.00002404, -0.00000643, 0.00001403, 0.00001736, 0.00000375, -0.00001481, -0.00002198, -0.00000684, 0.00001897, 0.00002538, 0.00000133, -0.00002425, -0.00001840, 0.00000970, 0.00002232, 0.00000688, -0.00001086, -0.00000876, 0.00000311, 0.00000521, -0.00000172, -0.00000437, 0.00000007, 0.00000404, 0.00000441, 0.00000378, 0.00000213, -0.00000170, -0.00000387, -0.00000043, 0.00000431, 0.00000330, -0.00000185, -0.00000340, 0.00000021, 0.00000247, -0.00000088, -0.00000676, -0.00001013, -0.00000876, -0.00000261, 0.00000578, 0.00001043, 0.00000699, -0.00000081, -0.00000484, -0.00000293, -0.00000024, -0.00000020, -0.00000033, 0.00000123, 0.00000168, -0.00000092, -0.00000317, -0.00000184, 0.00000035, -0.00000073, -0.00000349, -0.00000314, 0.00000022, 0.00000142, -0.00000169, -0.00000486, -0.00000346, },
		{ -0.00000426, -0.00000413, 0.00000004, 0.00000397, 0.00000423, 0.00000170, -0.00000047, -0.00000045, 0.00000147, 0.00000376, 0.00000447, 0.00000241, -0.00000118, -0.00000347, -0.00000380, -0.00000465, -0.00000628, -0.00000363, 0.00000640, 0.00001613, 0.00001341, -0.00000126, -0.00001179, -0.00000601, 0.00000780, 0.00001120, -0.00000093, -0.00001492, -0.00001576, -0.00000339, 0.00000890, 0.00000831, -0.00000486, -0.00001633, -0.00001254, 0.00000146, 0.00000649, -0.00000335, -0.00000984, 0.00000069, 0.00001110, -0.00000132, -0.00002248, -0.00001725, 0.00001123, 0.00001772, -0.00001353, -0.00003427, -0.00000379, 0.00003921, 0.00002777, -0.00002782, -0.00004810, -0.00000269, 0.00004262, 0.00002465, -0.00002518, -0.00003413, 0.00000311, 0.00002701, 0.00000929, -0.00001258, -0.00000347, 0.00001754, 0.00001506, -0.00000827, -0.00002222, -0.00001317, 0.00000597, 0.00001736, 0.00001197, -0.00000782, -0.00002461, -0.00001681, 0.00001226, 0.00002902, 0.00001077, -0.00001991, -0.00002355, 0.00000198, 0.00002122, 0.00001269, -0.00000581, -0.00000955, -0.00000050, 0.00000421, 0.00000028, -0.00000325, -0.00000164, 0.00000170, 0.00000415, 0.00000549, 0.00000378, -0.00000148, -0.00000484, -0.00000155, 0.00000409, 0.00000431, -0.00000052, -0.00000310, -0.00000074, 0.00000171, -0.00000001, -0.00000462, -0.00000867, -0.00000953, -0.00000526, 0.00000319, 0.00000968, 0.00000816, 0.00000078, -0.00000423, -0.00000327, -0.00000061, -0.00000023, -0.00000042, 0.00000098, 0.00000176, -0.00000031, -0.00000264, -0.00000180, 0.00000040, -0.00000026, -0.00000324, -0.00000389, -0.00000097, 0.00000117, -0.00000084, },
		{ 0.00000058, -0.00000016, -0.00000346, -0.00000446, -0.00000114, 0.00000330, 0.00000475, 0.00000271, -0.00000018, -0.00000139, -0.00000002, 0.00000305, 0.00000528, 0.00000397, -0.00000032, -0.00000353, -0.00000344, -0.00000318, -0.00000551, -0.00000591, 0.00000199, 0.00001400, 0.00001645, 0.00000409, -0.00001021, -0.00001028, 0.00000273, 0.00001204, 0.00000598, -0.00000928, -0.00001788, -0.00001095, 0.00000472, 0.00001238, 0.00000247, -0.00001422, -0.00001707, -0.00000301, 0.00000732, -0.00000020, -0.00001011, -0.00000233, 0.00001136, 0.00000378, -0.00001884, -0.00002012, 0.00000617, 0.00001822, -0.00000802, -0.00003224, -0.00000883, 0.00003440, 0.00003083, -0.00002063, -0.00004685, -0.00000896, 0.00003823, 0.00002877, -0.00001832, -0.00003424, -0.00000381, 0.00002234, 0.00001151, -0.00000914, -0.00000541, 0.00001263, 0.00001479, -0.00000250, -0.00001787, -0.00001687, -0.00000263, 0.00001357, 0.00001769, 0.00000112, -0.00002297, -0.00002511, 0.00000277, 0.00002868, 0.00001942, -0.00001268, -0.00002580, -0.00000606, 0.00001734, 0.00001673, 0.00000023, -0.00000829, -0.00000355, 0.00000210, 0.00000136, -0.00000165, -0.00000236, -0.00000057, 0.00000299, 0.00000648, 0.00000565, -0.00000050, -0.00000540, -0.00000290, 0.00000336, 0.00000501, 0.00000095, -0.00000238, -0.00000140, 0.00000072, 0.00000025, -0.00000285, -0.00000683, -0.00000940, -0.00000727, 0.00000045, 0.00000822, 0.00000875, 0.00000231, -0.00000335, -0.00000340, -0.00000095, -0.00000026, -0.00000047, 0.00000067, 0.00000165, 0.00000011, -0.00000211, -0.00000162, 0.00000058, 0.00000033, -0.00000280, -0.00000450, -0.00000227, },
		{ -0.00000488, -0.00000362, -0.00000033, 0.00000024, -0.00000254, -0.00000444, -0.00000218, 0.00000233, 0.00000483, 0.00000367, 0.00000056, -0.00000181, -0.00000150, 0.00000180, 0.00000556, 0.00000556, 0.00000093, -0.00000353, -0.00000356, -0.00000198, -0.00000397, -0.00000686, -0.00000209, 0.00001036, 0.00001754, 0.00000936, -0.00000649, -0.00001267, -0.00000323, 0.00000993, 0.00001134, -0.00000166, -0.00001659, -0.00001763, -0.00000213, 0.00001345, 0.00000995, -0.00000932, -0.00001977, -0.00000852, 0.00000622, 0.00000290, -0.00000890, -0.00000500, 0.00001004, 0.00000772, -0.00001394, -0.00002083, 0.00000113, 0.00001657, -0.00000368, -0.00002859, -0.00001178, 0.00002900, 0.00003131, -0.00001434, -0.00004352, -0.00001322, 0.00003276, 0.00003020, -0.00001190, -0.00003186, -0.00000876, 0.00001659, 0.00001149, -0.00000574, -0.00000535, 0.00000844, 0.00001254, 0.00000139, -0.00001218, -0.00001723, -0.00001015, 0.00000686, 0.00001983, 0.00001022, -0.00001742, -0.00003023, -0.00000790, 0.00002425, 0.00002574, -0.00000368, -0.00002466, -0.00001314, 0.00001125, 0.00001831, 0.00000632, -0.00000508, -0.00000535, -0.00000064, 0.00000129, -0.00000009, -0.00000217, -0.00000238, 0.00000122, 0.00000659, 0.00000736, 0.00000108, -0.00000535, -0.00000421, 0.00000218, 0.00000525, 0.00000237, -0.00000128, -0.00000166, -0.00000030, -0.00000001, -0.00000159, -0.00000490, -0.00000853, -0.00000849, -0.00000214, 0.00000623, 0.00000869, 0.00000362, -0.00000228, -0.00000331, -0.00000123, -0.00000028, -0.00000047, 0.00000036, 0.00000139, 0.00000034, -0.00000167, -0.00000139, 0.00000085, 0.00000108, -0.00000211, },
		{ 0.00000196, -0.00000112, -0.00000494, -0.00000492, -0.00000156, 0.00000029, -0.00000164, -0.00000407, -0.00000294, 0.00000119, 0.00000445, 0.00000437, 0.00000161, -0.00000165, -0.00000274, 0.00000015, 0.00000520, 0.00000694, 0.00000256, -0.00000323, -0.00000402, -0.00000133, -0.00000213, -0.00000648, -0.00000521, 0.00000586, 0.00001651, 0.00001360, -0.00000125, -0.00001259, -0.00000888, 0.00000519, 0.00001405, 0.00000661, -0.00001182, -0.00002199, -0.00001054, 0.00001100, 0.00001617, -0.00000222, -0.00001980, -0.00001405, 0.00000309, 0.00000509, -0.00000647, -0.00000660, 0.00000769, 0.00000996, -0.00000869, -0.00001933, -0.00000285, 0.00001331, -0.00000126, -0.00002430, -0.00001240, 0.00002416, 0.00002964, -0.00000990, -0.00003907, -0.00001504, 0.00002744, 0.00002919, -0.00000702, -0.00002788, -0.00001117, 0.00001091, 0.00000930, -0.00000340, -0.00000381, 0.00000575, 0.00000921, 0.00000295, -0.00000651, -0.00001462, -0.00001530, -0.00000136, 0.00001806, 0.00001772, -0.00000904, -0.00003127, -0.00001794, 0.00001638, 0.00002858, 0.00000562, -0.00002026, -0.00001806, 0.00000391, 0.00001714, 0.00001143, -0.00000039, -0.00000543, -0.00000335, 0.00000012, 0.00000097, -0.00000134, -0.00000347, -0.00000077, 0.00000583, 0.00000859, 0.00000301, -0.00000460, -0.00000521, 0.00000069, 0.00000497, 0.00000356, 0.00000007, -0.00000147, -0.00000116, -0.00000064, -0.00000091, -0.00000313, -0.00000713, -0.00000890, -0.00000434, 0.00000393, 0.00000800, 0.00000455, -0.00000117, -0.00000302, -0.00000141, -0.00000029, -0.00000042, 0.00000010, 0.00000105, 0.00000038, -0.00000137, -0.00000122, 0.00000115, },
		{ -0.00000118, 0.00000138, 0.00000292, 0.00000017, -0.00000459, -0.00000604, -0.00000303, -0.00000008, -0.00000089, -0.00000342, -0.00000335, 0.00000005, 0.00000366, 0.00000469, 0.00000274, -0.00000093, -0.00000351, -0.00000166, 0.00000418, 0.00000788, 0.00000441, -0.00000246, -0.00000457, -0.00000131, -0.00000046, -0.00000506, -0.00000696, 0.00000130, 0.00001356, 0.00001605, 0.00000459, -0.00000991, -0.00001303, -0.00000132, 0.00001348, 0.00001401, -0.00000418, -0.00002295, -0.00001900, 0.00000521, 0.00001981, 0.00000600, -0.00001681, -0.00001846, -0.00000169, 0.00000563, -0.00000344, -0.00000671, 0.00000509, 0.00001037, -0.00000406, -0.00001602, -0.00000497, 0.00000937, -0.00000106, -0.00002049, -0.00001102, 0.00002088, 0.00002669, -0.00000782, -0.00003469, -0.00001458, 0.00002339, 0.00002652, -0.00000438, -0.00002348, -0.00001104, 0.00000641, 0.00000557, -0.00000282, -0.00000167, 0.00000488, 0.00000585, 0.00000231, -0.00000202, -0.00000997, -0.00001735, -0.00000948, 0.00001284, 0.00002222, 0.00000055, -0.00002811, -0.00002564, 0.00000634, 0.00002742, 0.00001365, -0.00001332, -0.00002003, -0.00000347, 0.00001340, 0.00001469, 0.00000500, -0.00000364, -0.00000534, -0.00000187, 0.00000120, -0.00000025, -0.00000378, -0.00000261, 0.00000438, 0.00000911, 0.00000497, -0.00000322, -0.00000570, -0.00000087, 0.00000417, 0.00000436, 0.00000151, -0.00000084, -0.00000172, -0.00000148, -0.00000079, -0.00000171, -0.00000544, -0.00000856, -0.00000595, 0.00000158, 0.00000679, 0.00000501, -0.00000015, -0.00000258, -0.00000147, -0.00000025, -0.00000031, -0.00000009, 0.00000069, 0.00000029, -0.00000125, },
		{ 0.00000013, -0.00000126, -0.00000132, 0.00000141, 0.00000383, 0.00000174, -0.00000376, -0.00000684, -0.00000464, -0.00000091, -0.00000041, -0.00000261, -0.00000336, -0.00000092, 0.00000257, 0.00000453, 0.00000374, 0.00000021, -0.00000370, -0.00000337, 0.00000259, 0.00000818, 0.00000626, -0.00000116, -0.00000490, -0.00000183, 0.00000067, -0.00000308, -0.00000721, -0.00000256, 0.00000925, 0.00001624, 0.00000996, -0.00000503, -0.00001478, -0.00000836, 0.00000962, 0.00001915, 0.00000513, -0.00002004, -0.00002589, -0.00000310, 0.00001999, 0.00001389, -0.00001100, -0.00002064, -0.00000728, 0.00000409, -0.00000069, -0.00000528, 0.00000311, 0.00000929, -0.00000085, -0.00001173, -0.00000482, 0.00000587, -0.00000282, -0.00001815, -0.00000844, 0.00001972, 0.00002362, -0.00000805, -0.00003153, -0.00001253, 0.00002136, 0.00002330, -0.00000410, -0.00001984, -0.00000892, 0.00000393, 0.00000125, -0.00000428, 0.00000006, 0.00000564, 0.00000339, 0.00000011, 0.00000052, -0.00000450, -0.00001624, -0.00001601, 0.00000526, 0.00002297, 0.00000957, -0.00002143, -0.00002977, -0.00000419, 0.00002245, 0.00001907, -0.00000506, -0.00001876, -0.00000966, 0.00000776, 0.00001554, 0.00001016, -0.00000019, -0.00000603, -0.00000416, 0.00000048, 0.00000067, -0.00000342, -0.00000399, 0.00000252, 0.00000883, 0.00000664, -0.00000139, -0.00000554, -0.00000226, 0.00000299, 0.00000466, 0.00000284, 0.00000015, -0.00000188, -0.00000231, -0.00000113, -0.00000077, -0.00000371, -0.00000761, -0.00000689, -0.00000060, 0.00000521, 0.00000500, 0.00000066, -0.00000209, -0.00000143, -0.00000017, -0.00000015, -0.00000021, 0.00000036, },
		},
		{ { -0.00000080, -0.00000205, -0.00000379, -0.00000246, 0.00000245, 0.00000758, 0.00000908, 0.00000599, 0.00000054, -0.00000363, -0.00000340, 0.00000180, 0.00000848, 0.00001043, 0.00000407, -0.00000675, -0.00001274, -0.00000799, 0.00000340, 0.00001046, 0.00000575, -0.00000742, -0.00001807, -0.00001753, -0.00000691, 0.00000435, 0.00000729, 0.00000200, -0.00000271, 0.00000063, 0.00000802, 0.00000686, -0.00000831, -0.00002489, -0.00002315, -0.00000055, 0.00002021, 0.00001436, -0.00001429, -0.00003394, -0.00002040, 0.00001332, 0.00003104, 0.00001587, -0.00001085, -0.00001712, 0.00000096, 0.00001773, 0.00001276, -0.00000485, -0.00001219, -0.00000361, 0.00000515, 0.00000176, -0.00000621, -0.00000506, 0.00000372, 0.00000643, -0.00000126, -0.00000713, -0.00000083, 0.00000974, 0.00000906, -0.00000321, -0.00001073, -0.00000321, 0.00000906, 0.00000974, -0.00000083, -0.00000713, -0.00000126, 0.00000643, 0.00000372, -0.00000506, -0.00000621, 0.00000176, 0.00000515, -0.00000361, -0.00001219, -0.00000485, 0.00001276, 0.00001773, 0.00000096, -0.00001712, -0.00001085, 0.00001587, 0.00003104, 0.00001332, -0.00002040, -0.00003394, -0.00001429, 0.00001436, 0.00002021, -0.00000055, -0.00002315, -0.00002489, -0.00000831, 0.00000686, 0.00000802, 0.00000063, -0.00000271, 0.00000200, 0.00000729, 0.00000435, -0.00000691, -0.00001753, -0.00001807, -0.00000742, 0.00000575, 0.00001046, 0.00000340, -0.00000799, -0.00001274, -0.00000675, 0.00000407, 0.00001043, 0.00000848, 0.00000180, -0.00000340, -0.00000363, 0.00000054, 0.00000599, 0.00000908, 0.00000758, 0.00000245, -0.00000246, -0.00000379, -0.00000205, },
		{ -0.00000304, -0.00000359, -0.00000162, -0.00000089, -0.00000244, -0.00000381, -0.00000227, 0.00000226, 0.00000730, 0.00000971, 0.00000746, 0.00000123, -0.00000470, -0.00000474, 0.00000264, 0.00001136, 0.00001211, 0.00000191, -0.00001127, -0.00001537, -0.00000634, 0.00000724, 0.00001219, 0.00000321, -0.00001222, -0.00002082, -0.00001552, -0.00000164, 0.00000827, 0.00000671, -0.00000110, -0.00000320, 0.00000423, 0.00001031, 0.00000178, -0.00001796, -0.00002805, -0.00001358, 0.00001295, 0.00002182, 0.00000063, -0.00002795, -0.00003045, -0.00000232, 0.00002581, 0.00002444, -0.00000034, -0.00001665, -0.00000700, 0.00001251, 0.00001621, 0.00000172, -0.00001077, -0.00000769, 0.00000194, 0.00000302, -0.00000356, -0.00000496, 0.00000223, 0.00000657, 0.00000010, -0.00000813, -0.00000473, 0.00000742, 0.00001181, 0.00000159, -0.00000986, -0.00000725, 0.00000540, 0.00001069, 0.00000251, -0.00000611, -0.00000272, 0.00000652, 0.00000661, -0.00000348, -0.00000909, -0.00000209, 0.00000626, 0.00000177, -0.00001012, -0.00001025, 0.00000613, 0.00001923, 0.00000998, -0.00001206, -0.00001793, 0.00000342, 0.00002822, 0.00002540, -0.00000505, -0.00003105, -0.00002595, 0.00000151, 0.00001957, 0.00001033, -0.00001310, -0.00002533, -0.00001701, -0.00000054, 0.00000772, 0.00000460, -0.00000095, -0.00000080, 0.00000385, 0.00000564, -0.00000014, -0.00001051, -0.00001730, -0.00001443, -0.00000367, 0.00000641, 0.00000791, 0.00000073, -0.00000765, -0.00000952, -0.00000376, 0.00000439, 0.00000882, 0.00000730, 0.00000190, -0.00000327, -0.00000445, -0.00000045, 0.00000605, 0.00000991, 0.00000787, 0.00000181, },
		{ 0.00001097, 0.00000733, 0.00000064, -0.00000347, -0.00000323, -0.00000134, -0.00000103, -0.00000268, -0.00000413, -0.00000293, 0.00000184, 0.00000838, 0.00001221, 0.00000909, -0.00000007, -0.00000766, -0.00000558, 0.00000562, 0.00001521, 0.00001226, -0.00000249, -0.00001618, -0.00001622, -0.00000277, 0.00001119, 0.00001222, -0.00000112, -0.00001712, -0.00002163, -0.00001107, 0.00000421, 0.00001019, 0.00000382, -0.00000384, -0.00000113, 0.00000849, 0.00000903, -0.00000718, -0.00002555, -0.00002356, 0.00000053, 0.00002103, 0.00001315, -0.00001568, -0.00003213, -0.00001611, 0.00001473, 0.00002639, 0.00000946, -0.00001171, -0.00001147, 0.00000609, 0.00001620, 0.00000714, -0.00000721, -0.00000936, -0.00000125, 0.00000247, -0.00000234, -0.00000495, 0.00000142, 0.00000764, 0.00000261, -0.00000830, -0.00000921, 0.00000313, 0.00001299, 0.00000693, -0.00000706, -0.00001024, 0.00000092, 0.00001041, 0.00000574, -0.00000473, -0.00000484, 0.00000538, 0.00000993, 0.00000074, -0.00000997, -0.00000739, 0.00000381, 0.00000600, -0.00000503, -0.00001226, -0.00000149, 0.00001613, 0.00001680, -0.00000312, -0.00001902, -0.00000843, 0.00001871, 0.00003015, 0.00001029, -0.00002060, -0.00003031, -0.00001157, 0.00001210, 0.00001528, -0.00000234, -0.00001999, -0.00002109, -0.00000850, 0.00000347, 0.00000600, 0.00000203, -0.00000072, 0.00000098, 0.00000383, 0.00000271, -0.00000389, -0.00001209, -0.00001554, -0.00001098, -0.00000151, 0.00000570, 0.00000567, -0.00000023, -0.00000613, -0.00000704, -0.00000247, 0.00000419, 0.00000839, 0.00000720, 0.00000127, -0.00000490, -0.00000588, -0.00000017, 0.00000767, },
		{ -0.00000624, 0.00000183, 0.00000982, 0.00001124, 0.00000603, -0.00000052, -0.00000364, -0.00000291, -0.00000113, -0.00000102, -0.00000311, -0.00000537, -0.00000418, 0.00000254, 0.00001169, 0.00001558, 0.00000890, -0.00000398, -0.00001101, -0.00000423, 0.00001059, 0.00001839, 0.00000988, -0.00000840, -0.00002008, -0.00001480, 0.00000196, 0.00001419, 0.00001021, -0.00000664, -0.00002098, -0.00001973, -0.00000475, 0.00000899, 0.00000922, -0.00000019, -0.00000436, 0.00000320, 0.00001077, 0.00000292, -0.00001701, -0.00002677, -0.00001190, 0.00001309, 0.00001903, -0.00000234, -0.00002626, -0.00002392, 0.00000236, 0.00002229, 0.00001532, -0.00000541, -0.00001239, 0.00000079, 0.00001427, 0.00001060, -0.00000338, -0.00000929, -0.00000328, 0.00000177, -0.00000202, -0.00000604, -0.00000034, 0.00000874, 0.00000682, -0.00000612, -0.00001308, -0.00000331, 0.00001140, 0.00001186, -0.00000211, -0.00001141, -0.00000413, 0.00000846, 0.00000875, -0.00000217, -0.00000693, 0.00000217, 0.00001172, 0.00000673, -0.00000719, -0.00001159, -0.00000168, 0.00000679, 0.00000076, -0.00001026, -0.00000727, 0.00000980, 0.00001922, 0.00000637, -0.00001432, -0.00001585, 0.00000646, 0.00002712, 0.00002088, -0.00000685, -0.00002684, -0.00002026, 0.00000174, 0.00001377, 0.00000510, -0.00001216, -0.00002009, -0.00001382, -0.00000234, 0.00000417, 0.00000374, 0.00000096, 0.00000017, 0.00000157, 0.00000233, -0.00000028, -0.00000617, -0.00001195, -0.00001327, -0.00000856, -0.00000085, 0.00000463, 0.00000463, 0.00000017, -0.00000472, -0.00000593, -0.00000196, 0.00000478, 0.00000902, 0.00000664, -0.00000105, -0.00000740, },
		{ 0.00000439, -0.00000461, -0.00000912, -0.00000470, 0.00000478, 0.00001132, 0.00001052, 0.00000454, -0.00000138, -0.00000375, -0.00000265, -0.00000072, -0.00000105, -0.00000447, -0.00000760, -0.00000475, 0.00000570, 0.00001679, 0.00001777, 0.00000559, -0.00000949, -0.00001270, -0.00000010, 0.00001616, 0.00001939, 0.00000510, -0.00001447, -0.00002202, -0.00001143, 0.00000686, 0.00001552, 0.00000623, -0.00001237, -0.00002261, -0.00001502, 0.00000196, 0.00001112, 0.00000576, -0.00000317, -0.00000179, 0.00000758, 0.00000854, -0.00000660, -0.00002299, -0.00001969, 0.00000226, 0.00001763, 0.00000726, -0.00001683, -0.00002504, -0.00000705, 0.00001545, 0.00001679, -0.00000066, -0.00001183, -0.00000291, 0.00001220, 0.00001299, -0.00000005, -0.00000890, -0.00000465, 0.00000188, -0.00000102, -0.00000752, -0.00000406, 0.00000791, 0.00001163, -0.00000066, -0.00001422, -0.00001069, 0.00000617, 0.00001464, 0.00000439, -0.00000975, -0.00000878, 0.00000447, 0.00001057, 0.00000177, -0.00000755, -0.00000246, 0.00001038, 0.00001216, -0.00000099, -0.00001222, -0.00000784, 0.00000360, 0.00000449, -0.00000576, -0.00000951, 0.00000295, 0.00001731, 0.00001338, -0.00000636, -0.00001746, -0.00000432, 0.00001887, 0.00002468, 0.00000545, -0.00001826, -0.00002266, -0.00000717, 0.00000829, 0.00000768, -0.00000545, -0.00001609, -0.00001537, -0.00000695, 0.00000061, 0.00000315, 0.00000222, 0.00000088, 0.00000057, 0.00000083, 0.00000027, -0.00000236, -0.00000683, -0.00001086, -0.00001147, -0.00000750, -0.00000094, 0.00000426, 0.00000490, 0.00000085, -0.00000435, -0.00000567, -0.00000089, 0.00000648, 0.00000946, },
		{ 0.00000144, 0.00000833, 0.00000823, 0.00000044, -0.00000787, -0.00000887, -0.00000170, 0.00000741, 0.00001170, 0.00000932, 0.00000326, -0.00000213, -0.00000391, -0.00000216, -0.00000014, -0.00000186, -0.00000719, -0.00000984, -0.00000299, 0.00001151, 0.00002176, 0.00001682, -0.00000054, -0.00001449, -0.00001139, 0.00000595, 0.00002055, 0.00001769, -0.00000105, -0.00001950, -0.00002183, -0.00000682, 0.00001105, 0.00001479, 0.00000084, -0.00001704, -0.00002120, -0.00000839, 0.00000708, 0.00000996, 0.00000163, -0.00000322, 0.00000276, 0.00000889, 0.00000137, -0.00001534, -0.00002119, -0.00000668, 0.00001187, 0.00001125, -0.00000836, -0.00002211, -0.00001204, 0.00000954, 0.00001618, 0.00000189, -0.00001207, -0.00000659, 0.00001024, 0.00001568, 0.00000368, -0.00000869, -0.00000673, 0.00000208, 0.00000165, -0.00000745, -0.00000888, 0.00000364, 0.00001447, 0.00000718, -0.00001081, -0.00001643, -0.00000213, 0.00001343, 0.00001060, -0.00000501, -0.00001141, -0.00000102, 0.00001006, 0.00000614, -0.00000569, -0.00000667, 0.00000594, 0.00001462, 0.00000650, -0.00000847, -0.00001174, -0.00000195, 0.00000462, -0.00000135, -0.00000853, -0.00000213, 0.00001289, 0.00001667, 0.00000182, -0.00001439, -0.00001110, 0.00000914, 0.00002264, 0.00001336, -0.00000850, -0.00001998, -0.00001215, 0.00000235, 0.00000669, -0.00000172, -0.00001201, -0.00001438, -0.00000908, -0.00000246, 0.00000116, 0.00000196, 0.00000157, 0.00000082, -0.00000003, -0.00000067, -0.00000135, -0.00000307, -0.00000634, -0.00000982, -0.00001078, -0.00000736, -0.00000070, 0.00000507, 0.00000564, 0.00000067, -0.00000494, -0.00000502, },
		{ -0.00000066, -0.00000538, -0.00000294, 0.00000445, 0.00000875, 0.00000485, -0.00000390, -0.00000931, -0.00000667, 0.00000147, 0.00000893, 0.00001131, 0.00000819, 0.00000211, -0.00000297, -0.00000387, -0.00000125, 0.00000000, -0.00000421, -0.00001073, -0.00001031, 0.00000201, 0.00001854, 0.00002424, 0.00001217, -0.00000767, -0.00001702, -0.00000723, 0.00001218, 0.00002266, 0.00001382, -0.00000736, -0.00002290, -0.00001980, -0.00000173, 0.00001386, 0.00001200, -0.00000494, -0.00001931, -0.00001679, -0.00000170, 0.00000904, 0.00000652, -0.00000066, -0.00000006, 0.00000632, 0.00000501, -0.00000798, -0.00001826, -0.00001129, 0.00000592, 0.00001135, -0.00000307, -0.00001888, -0.00001440, 0.00000595, 0.00001649, 0.00000428, -0.00001340, -0.00001223, 0.00000660, 0.00001859, 0.00000928, -0.00000739, -0.00001008, 0.00000051, 0.00000512, -0.00000427, -0.00001231, -0.00000360, 0.00001287, 0.00001452, -0.00000288, -0.00001769, -0.00001107, 0.00000766, 0.00001407, 0.00000167, -0.00001073, -0.00000638, 0.00000681, 0.00000927, -0.00000166, -0.00000860, 0.00000004, 0.00001311, 0.00001245, -0.00000167, -0.00001158, -0.00000711, 0.00000166, 0.00000090, -0.00000615, -0.00000473, 0.00000812, 0.00001687, 0.00000829, -0.00000893, -0.00001363, 0.00000077, 0.00001739, 0.00001660, -0.00000052, -0.00001501, -0.00001339, -0.00000170, 0.00000469, -0.00000040, -0.00000955, -0.00001308, -0.00000945, -0.00000385, -0.00000050, 0.00000070, 0.00000124, 0.00000121, 0.00000008, -0.00000145, -0.00000207, -0.00000183, -0.00000253, -0.00000558, -0.00000969, -0.00001115, -0.00000716, 0.00000067, 0.00000657, 0.00000567, },
		{ 0.00000303, 0.00000756, 0.00000429, -0.00000229, -0.00000437, 0.00000045, 0.00000656, 0.00000676, 0.00000027, -0.00000687, -0.00000847, -0.00000368, 0.00000374, 0.00000947, 0.00001081, 0.00000719, 0.00000080, -0.00000375, -0.00000327, -0.00000030, -0.00000126, -0.00000808, -0.00001342, -0.00000748, 0.00000951, 0.00002430, 0.00002280, 0.00000505, -0.00001350, -0.00001642, -0.00000159, 0.00001714, 0.00002247, 0.00000877, -0.00001304, -0.00002461, -0.00001637, 0.00000315, 0.00001477, 0.00000759, -0.00000968, -0.00001837, -0.00001065, 0.00000287, 0.00000770, 0.00000324, 0.00000050, 0.00000424, 0.00000560, -0.00000329, -0.00001439, -0.00001248, 0.00000228, 0.00001091, 0.00000018, -0.00001726, -0.00001730, 0.00000290, 0.00001871, 0.00000943, -0.00001335, -0.00001988, -0.00000112, 0.00001939, 0.00001687, -0.00000281, -0.00001322, -0.00000404, 0.00000693, 0.00000159, -0.00001177, -0.00001108, 0.00000628, 0.00001790, 0.00000711, -0.00001315, -0.00001723, -0.00000116, 0.00001301, 0.00000788, -0.00000677, -0.00000971, 0.00000177, 0.00000984, 0.00000303, -0.00000754, -0.00000505, 0.00000845, 0.00001488, 0.00000554, -0.00000762, -0.00000957, -0.00000234, 0.00000059, -0.00000430, -0.00000570, 0.00000414, 0.00001538, 0.00001265, -0.00000294, -0.00001308, -0.00000516, 0.00001134, 0.00001662, 0.00000481, -0.00001016, -0.00001274, -0.00000368, 0.00000358, 0.00000041, -0.00000841, -0.00001291, -0.00000987, -0.00000411, -0.00000079, -0.00000014, 0.00000020, 0.00000083, 0.00000041, -0.00000148, -0.00000303, -0.00000252, -0.00000095, -0.00000142, -0.00000558, -0.00001077, -0.00001173, -0.00000586, },
		{ -0.00001249, -0.00001134, -0.00000317, 0.00000531, 0.00000700, 0.00000199, -0.00000274, -0.00000151, 0.00000376, 0.00000633, 0.00000278, -0.00000368, -0.00000748, -0.00000620, -0.00000123, 0.00000491, 0.00000972, 0.00001047, 0.00000599, -0.00000069, -0.00000401, -0.00000218, -0.00000034, -0.00000444, -0.00001216, -0.00001328, -0.00000122, 0.00001724, 0.00002661, 0.00001782, -0.00000225, -0.00001666, -0.00001351, 0.00000396, 0.00002031, 0.00002066, 0.00000333, -0.00001780, -0.00002471, -0.00001200, 0.00000707, 0.00001358, 0.00000270, -0.00001191, -0.00001454, -0.00000503, 0.00000399, 0.00000480, 0.00000270, 0.00000447, 0.00000606, -0.00000064, -0.00001187, -0.00001308, 0.00000028, 0.00001203, 0.00000438, -0.00001589, -0.00002229, -0.00000296, 0.00002048, 0.00001830, -0.00000825, -0.00002650, -0.00001349, 0.00001463, 0.00002390, 0.00000610, -0.00001313, -0.00001056, 0.00000471, 0.00000745, -0.00000657, -0.00001528, -0.00000319, 0.00001528, 0.00001523, -0.00000408, -0.00001799, -0.00000986, 0.00000754, 0.00001114, -0.00000112, -0.00000992, -0.00000324, 0.00000772, 0.00000657, -0.00000426, -0.00000773, 0.00000267, 0.00001358, 0.00001075, -0.00000174, -0.00000869, -0.00000520, -0.00000109, -0.00000376, -0.00000636, 0.00000081, 0.00001314, 0.00001545, 0.00000282, -0.00001068, -0.00000895, 0.00000562, 0.00001503, 0.00000819, -0.00000624, -0.00001192, -0.00000492, 0.00000348, 0.00000219, -0.00000697, -0.00001353, -0.00001157, -0.00000484, -0.00000017, 0.00000038, -0.00000036, -0.00000011, 0.00000019, -0.00000118, -0.00000345, -0.00000390, -0.00000154, 0.00000085, -0.00000083, -0.00000697, },
		{ 0.00000257, -0.00000173, -0.00000956, -0.00001361, -0.00000923, 0.00000010, 0.00000617, 0.00000485, 0.00000022, -0.00000106, 0.00000222, 0.00000518, 0.00000354, -0.00000148, -0.00000549, -0.00000613, -0.00000398, 0.00000010, 0.00000559, 0.00001016, 0.00001001, 0.00000440, -0.00000190, -0.00000350, -0.00000147, -0.00000230, -0.00000879, -0.00001432, -0.00000935, 0.00000673, 0.00002258, 0.00002495, 0.00001127, -0.00000782, -0.00001730, -0.00000977, 0.00000856, 0.00002207, 0.00001788, -0.00000222, -0.00002152, -0.00002323, -0.00000728, 0.00000931, 0.00001075, -0.00000104, -0.00001095, -0.00000956, -0.00000212, 0.00000224, 0.00000323, 0.00000584, 0.00000819, 0.00000247, -0.00001011, -0.00001505, -0.00000281, 0.00001368, 0.00001141, -0.00001131, -0.00002740, -0.00001342, 0.00001746, 0.00002808, 0.00000376, -0.00002714, -0.00002745, 0.00000238, 0.00002588, 0.00001743, -0.00000732, -0.00001574, -0.00000176, 0.00000990, 0.00000125, -0.00001398, -0.00001162, 0.00000746, 0.00001814, 0.00000601, -0.00001306, -0.00001513, -0.00000016, 0.00001031, 0.00000385, -0.00000744, -0.00000659, 0.00000402, 0.00000796, -0.00000032, -0.00000776, -0.00000226, 0.00000980, 0.00001289, 0.00000382, -0.00000548, -0.00000587, -0.00000250, -0.00000395, -0.00000731, -0.00000256, 0.00001012, 0.00001700, 0.00000836, -0.00000691, -0.00001103, 0.00000033, 0.00001253, 0.00001057, -0.00000270, -0.00001137, -0.00000676, 0.00000309, 0.00000485, -0.00000389, -0.00001336, -0.00001420, -0.00000724, -0.00000019, 0.00000188, 0.00000047, -0.00000061, -0.00000050, -0.00000116, -0.00000344, -0.00000507, -0.00000333, 0.00000081, },
		{ -0.00000545, -0.00000101, 0.00000373, 0.00000297, -0.00000439, -0.00001223, -0.00001296, -0.00000577, 0.00000246, 0.00000498, 0.00000235, 0.00000046, 0.00000235, 0.00000490, 0.00000382, -0.00000050, -0.00000415, -0.00000500, -0.00000426, -0.00000280, 0.00000071, 0.00000648, 0.00001068, 0.00000903, 0.00000271, -0.00000232, -0.00000279, -0.00000235, -0.00000609, -0.00001230, -0.00001286, -0.00000259, 0.00001359, 0.00002409, 0.00002071, 0.00000538, -0.00001116, -0.00001667, -0.00000625, 0.00001239, 0.00002297, 0.00001426, -0.00000778, -0.00002388, -0.00002034, -0.00000311, 0.00000956, 0.00000758, -0.00000217, -0.00000757, -0.00000593, -0.00000267, 0.00000010, 0.00000529, 0.00001089, 0.00000775, -0.00000649, -0.00001746, -0.00000897, 0.00001246, 0.00001977, -0.00000107, -0.00002805, -0.00002637, 0.00000652, 0.00003311, 0.00002032, -0.00001810, -0.00003703, -0.00001540, 0.00001905, 0.00002642, 0.00000385, -0.00001580, -0.00000983, 0.00000707, 0.00000793, -0.00000778, -0.00001554, -0.00000228, 0.00001494, 0.00001321, -0.00000469, -0.00001530, -0.00000698, 0.00000621, 0.00000634, -0.00000389, -0.00000773, 0.00000021, 0.00000735, 0.00000293, -0.00000598, -0.00000537, 0.00000524, 0.00001235, 0.00000765, -0.00000160, -0.00000475, -0.00000269, -0.00000371, -0.00000807, -0.00000617, 0.00000588, 0.00001693, 0.00001347, -0.00000173, -0.00001130, -0.00000460, 0.00000897, 0.00001213, 0.00000120, -0.00001029, -0.00000932, 0.00000108, 0.00000694, 0.00000069, -0.00001095, -0.00001614, -0.00001108, -0.00000213, 0.00000292, 0.00000246, 0.00000020, -0.00000086, -0.00000143, -0.00000336, -0.00000580, },
		{ -0.00000322, -0.00000614, -0.00000738, -0.00000390, 0.00000263, 0.00000587, 0.00000124, -0.00000805, -0.00001350, -0.00001029, -0.00000235, 0.00000271, 0.00000248, 0.00000127, 0.00000301, 0.00000577, 0.00000501, 0.00000040, -0.00000372, -0.00000440, -0.00000339, -0.00000322, -0.00000257, 0.00000147, 0.00000770, 0.00001076, 0.00000762, 0.00000160, -0.00000211, -0.00000314, -0.00000537, -0.00001011, -0.00001276, -0.00000781, 0.00000443, 0.00001719, 0.00002243, 0.00001620, 0.00000124, -0.00001324, -0.00001597, -0.00000292, 0.00001600, 0.00002302, 0.00000966, -0.00001294, -0.00002450, -0.00001669, -0.00000042, 0.00000840, 0.00000588, -0.00000033, -0.00000400, -0.00000560, -0.00000525, 0.00000069, 0.00001094, 0.00001394, 0.00000100, -0.00001694, -0.00001704, 0.00000552, 0.00002496, 0.00001354, -0.00002017, -0.00003604, -0.00001111, 0.00002815, 0.00003499, -0.00000007, -0.00003639, -0.00003275, 0.00000323, 0.00002774, 0.00001649, -0.00000910, -0.00001526, 0.00000008, 0.00001023, 0.00000017, -0.00001382, -0.00000995, 0.00000756, 0.00001523, 0.00000354, -0.00001113, -0.00001058, 0.00000110, 0.00000610, -0.00000094, -0.00000732, -0.00000277, 0.00000558, 0.00000500, -0.00000348, -0.00000672, 0.00000107, 0.00001020, 0.00000959, 0.00000170, -0.00000298, -0.00000191, -0.00000243, -0.00000767, -0.00000929, 0.00000046, 0.00001454, 0.00001731, 0.00000463, -0.00000921, -0.00000858, 0.00000418, 0.00001220, 0.00000538, -0.00000769, -0.00001148, -0.00000268, 0.00000687, 0.00000522, -0.00000622, -0.00001572, -0.00001483, -0.00000601, 0.00000212, 0.00000434, 0.00000224, -0.00000011, -0.00000139, },
		{ 0.00000174, -0.00000046, -0.00000258, -0.00000592, -0.00000879, -0.00000722, -0.00000036, 0.00000639, 0.00000594, -0.00000239, -0.00001113, -0.00001246, -0.00000662, -0.00000061, 0.00000091, 0.00000048, 0.00000265, 0.00000663, 0.00000722, 0.00000249, -0.00000312, -0.00000469, -0.00000299, -0.00000227, -0.00000334, -0.00000242, 0.00000277, 0.00000873, 0.00001018, 0.00000643, 0.00000134, -0.00000238, -0.00000566, -0.00000955, -0.00001187, -0.00000925, -0.00000130, 0.00000892, 0.00001727, 0.00001977, 0.00001307, -0.00000175, -0.00001541, -0.00001536, 0.00000093, 0.00001949, 0.00002171, 0.00000418, -0.00001695, -0.00002344, -0.00001339, 0.00000051, 0.00000738, 0.00000696, 0.00000312, -0.00000284, -0.00000852, -0.00000671, 0.00000568, 0.00001721, 0.00001123, -0.00001042, -0.00002273, -0.00000653, 0.00002240, 0.00002716, -0.00000369, -0.00003621, -0.00002941, 0.00001220, 0.00004040, 0.00002106, -0.00002359, -0.00004229, -0.00001693, 0.00001881, 0.00002476, 0.00000225, -0.00001479, -0.00000742, 0.00000746, 0.00000615, -0.00000819, -0.00001307, -0.00000051, 0.00001240, 0.00000873, -0.00000518, -0.00001055, -0.00000283, 0.00000433, 0.00000070, -0.00000642, -0.00000494, 0.00000336, 0.00000604, -0.00000086, -0.00000680, -0.00000233, 0.00000733, 0.00001014, 0.00000415, -0.00000142, -0.00000106, -0.00000050, -0.00000563, -0.00001071, -0.00000515, 0.00000968, 0.00001862, 0.00001108, -0.00000459, -0.00001051, -0.00000124, 0.00001009, 0.00000880, -0.00000346, -0.00001181, -0.00000700, 0.00000414, 0.00000784, -0.00000056, -0.00001249, -0.00001667, -0.00001054, -0.00000084, 0.00000474, 0.00000446, },
		{ 0.00000317, 0.00000568, 0.00000385, 0.00000126, -0.00000113, -0.00000473, -0.00000914, -0.00001025, -0.00000472, 0.00000434, 0.00000867, 0.00000348, -0.00000658, -0.00001214, -0.00000950, -0.00000377, -0.00000126, -0.00000148, 0.00000059, 0.00000594, 0.00000918, 0.00000587, -0.00000103, -0.00000482, -0.00000356, -0.00000158, -0.00000244, -0.00000389, -0.00000162, 0.00000423, 0.00000907, 0.00000947, 0.00000613, 0.00000114, -0.00000464, -0.00001012, -0.00001245, -0.00000957, -0.00000311, 0.00000357, 0.00000967, 0.00001570, 0.00001831, 0.00001122, -0.00000512, -0.00001818, -0.00001397, 0.00000576, 0.00002216, 0.00001862, -0.00000137, -0.00001915, -0.00002152, -0.00001146, 0.00000054, 0.00000839, 0.00001063, 0.00000552, -0.00000551, -0.00001233, -0.00000407, 0.00001380, 0.00001968, 0.00000181, -0.00002126, -0.00001892, 0.00001102, 0.00003322, 0.00001629, -0.00002426, -0.00004026, -0.00000988, 0.00003257, 0.00003673, -0.00000246, -0.00003928, -0.00003384, 0.00000186, 0.00002416, 0.00001319, -0.00000840, -0.00001155, 0.00000182, 0.00000806, -0.00000188, -0.00001179, -0.00000625, 0.00000712, 0.00001004, -0.00000011, -0.00000832, -0.00000470, 0.00000259, 0.00000152, -0.00000560, -0.00000675, 0.00000080, 0.00000628, 0.00000175, -0.00000589, -0.00000498, 0.00000408, 0.00000977, 0.00000607, -0.00000015, -0.00000083, 0.00000105, -0.00000247, -0.00000972, -0.00000935, 0.00000329, 0.00001671, 0.00001597, 0.00000179, -0.00000951, -0.00000594, 0.00000597, 0.00001021, 0.00000137, -0.00000970, -0.00001010, -0.00000028, 0.00000755, 0.00000405, -0.00000745, -0.00001579, -0.00001400, -0.00000500, },
		{ -0.00001534, -0.00000894, 0.00000022, 0.00000540, 0.00000531, 0.00000305, 0.00000082, -0.00000250, -0.00000787, -0.00001200, -0.00000952, -0.00000013, 0.00000860, 0.00000831, -0.00000071, -0.00000950, -0.00001069, -0.00000623, -0.00000312, -0.00000344, -0.00000248, 0.00000311, 0.00000920, 0.00000914, 0.00000273, -0.00000337, -0.00000417, -0.00000187, -0.00000146, -0.00000344, -0.00000385, -0.00000029, 0.00000513, 0.00000895, 0.00000955, 0.00000643, -0.00000065, -0.00000948, -0.00001445, -0.00001148, -0.00000350, 0.00000250, 0.00000457, 0.00000789, 0.00001510, 0.00001868, 0.00000905, -0.00000996, -0.00002069, -0.00001075, 0.00001102, 0.00002296, 0.00001409, -0.00000584, -0.00001967, -0.00002005, -0.00001096, 0.00000152, 0.00001241, 0.00001487, 0.00000438, -0.00001101, -0.00001347, 0.00000342, 0.00002109, 0.00001519, -0.00001107, -0.00002518, -0.00000513, 0.00002801, 0.00003178, -0.00000353, -0.00003804, -0.00002936, 0.00001371, 0.00004014, 0.00001877, -0.00002461, -0.00004083, -0.00001672, 0.00001436, 0.00001850, 0.00000052, -0.00001066, -0.00000325, 0.00000647, 0.00000248, -0.00000821, -0.00000868, 0.00000204, 0.00000866, 0.00000279, -0.00000584, -0.00000516, 0.00000164, 0.00000249, -0.00000450, -0.00000836, -0.00000235, 0.00000552, 0.00000422, -0.00000395, -0.00000672, 0.00000051, 0.00000844, 0.00000764, 0.00000130, -0.00000109, 0.00000140, 0.00000051, -0.00000671, -0.00001096, -0.00000288, 0.00001202, 0.00001794, 0.00000827, -0.00000570, -0.00000847, 0.00000096, 0.00000907, 0.00000524, -0.00000577, -0.00001084, -0.00000464, 0.00000482, 0.00000627, -0.00000236, -0.00001268, },
		{ 0.00000134, -0.00000870, -0.00001469, -0.00001167, -0.00000303, 0.00000403, 0.00000580, 0.00000421, 0.00000242, 0.00000015, -0.00000492, -0.00001146, -0.00001329, -0.00000617, 0.00000527, 0.00001074, 0.00000535, -0.00000482, -0.00000999, -0.00000788, -0.00000452, -0.00000462, -0.00000520, -0.00000103, 0.00000655, 0.00001040, 0.00000676, -0.00000000, -0.00000332, -0.00000236, -0.00000131, -0.00000259, -0.00000408, -0.00000296, 0.00000072, 0.00000530, 0.00000940, 0.00001073, 0.00000578, -0.00000544, -0.00001556, -0.00001559, -0.00000573, 0.00000356, 0.00000440, 0.00000189, 0.00000647, 0.00001707, 0.00001948, 0.00000477, -0.00001582, -0.00002136, -0.00000562, 0.00001534, 0.00002148, 0.00000936, -0.00000852, -0.00001961, -0.00001998, -0.00001062, 0.00000529, 0.00001856, 0.00001675, -0.00000096, -0.00001617, -0.00000942, 0.00001326, 0.00002307, 0.00000431, -0.00002122, -0.00001860, 0.00001354, 0.00003630, 0.00001797, -0.00002308, -0.00003812, -0.00000850, 0.00003015, 0.00003167, -0.00000461, -0.00003582, -0.00002973, -0.00000043, 0.00001589, 0.00000721, -0.00000630, -0.00000541, 0.00000375, 0.00000434, -0.00000464, -0.00000877, -0.00000145, 0.00000652, 0.00000399, -0.00000421, -0.00000555, 0.00000099, 0.00000399, -0.00000237, -0.00000915, -0.00000600, 0.00000327, 0.00000599, -0.00000104, -0.00000708, -0.00000304, 0.00000596, 0.00000855, 0.00000321, -0.00000110, 0.00000062, 0.00000218, -0.00000303, -0.00000988, -0.00000717, 0.00000606, 0.00001672, 0.00001312, -0.00000026, -0.00000821, -0.00000335, 0.00000594, 0.00000706, -0.00000144, -0.00000923, -0.00000746, 0.00000107, 0.00000603, },
		{ -0.00000226, 0.00000420, 0.00000323, -0.00000502, -0.00001288, -0.00001316, -0.00000595, 0.00000229, 0.00000579, 0.00000474, 0.00000304, 0.00000204, -0.00000124, -0.00000836, -0.00001442, -0.00001201, -0.00000089, 0.00000960, 0.00000997, 0.00000120, -0.00000710, -0.00000837, -0.00000566, -0.00000511, -0.00000664, -0.00000483, 0.00000211, 0.00000868, 0.00000896, 0.00000398, -0.00000047, -0.00000152, -0.00000135, -0.00000239, -0.00000375, -0.00000362, -0.00000200, 0.00000077, 0.00000560, 0.00001133, 0.00001191, 0.00000215, -0.00001292, -0.00001976, -0.00001131, 0.00000295, 0.00000773, 0.00000143, -0.00000179, 0.00000803, 0.00002090, 0.00001831, -0.00000206, -0.00002088, -0.00001918, 0.00000015, 0.00001741, 0.00001844, 0.00000579, -0.00000983, -0.00002043, -0.00002098, -0.00000832, 0.00001231, 0.00002421, 0.00001411, -0.00000854, -0.00001736, -0.00000056, 0.00002089, 0.00001766, -0.00000836, -0.00002288, -0.00000324, 0.00002880, 0.00003179, -0.00000176, -0.00003309, -0.00002484, 0.00001198, 0.00003209, 0.00001225, -0.00002245, -0.00003307, -0.00001382, 0.00000737, 0.00000858, -0.00000203, -0.00000492, 0.00000203, 0.00000495, -0.00000199, -0.00000808, -0.00000367, 0.00000475, 0.00000475, -0.00000312, -0.00000658, -0.00000049, 0.00000531, 0.00000102, -0.00000815, -0.00000929, -0.00000054, 0.00000617, 0.00000223, -0.00000571, -0.00000576, 0.00000246, 0.00000819, 0.00000526, -0.00000023, -0.00000045, 0.00000224, -0.00000009, -0.00000717, -0.00000890, 0.00000062, 0.00001319, 0.00001540, 0.00000520, -0.00000560, -0.00000585, 0.00000208, 0.00000668, 0.00000195, -0.00000623, -0.00000823, },
		{ -0.00000299, -0.00000732, -0.00000441, 0.00000187, 0.00000369, -0.00000212, -0.00001064, -0.00001386, -0.00000862, 0.00000039, 0.00000586, 0.00000542, 0.00000298, 0.00000232, 0.00000152, -0.00000375, -0.00001211, -0.00001542, -0.00000810, 0.00000464, 0.00001136, 0.00000699, -0.00000214, -0.00000702, -0.00000636, -0.00000549, -0.00000701, -0.00000712, -0.00000227, 0.00000475, 0.00000809, 0.00000633, 0.00000315, 0.00000139, -0.00000001, -0.00000235, -0.00000410, -0.00000368, -0.00000255, -0.00000210, 0.00000032, 0.00000747, 0.00001442, 0.00001078, -0.00000540, -0.00002049, -0.00001871, -0.00000204, 0.00001052, 0.00000632, -0.00000459, -0.00000294, 0.00001311, 0.00002403, 0.00001355, -0.00000997, -0.00002317, -0.00001464, 0.00000469, 0.00001706, 0.00001544, 0.00000391, -0.00001120, -0.00002281, -0.00002128, -0.00000245, 0.00002095, 0.00002634, 0.00000720, -0.00001467, -0.00001287, 0.00000963, 0.00002233, 0.00000696, -0.00001639, -0.00001432, 0.00001429, 0.00003374, 0.00001705, -0.00001781, -0.00002954, -0.00000578, 0.00002217, 0.00002001, -0.00000783, -0.00002781, -0.00002096, -0.00000219, 0.00000508, -0.00000078, -0.00000416, 0.00000150, 0.00000584, 0.00000040, -0.00000733, -0.00000562, 0.00000312, 0.00000583, -0.00000148, -0.00000772, -0.00000338, 0.00000512, 0.00000477, -0.00000488, -0.00001084, -0.00000513, 0.00000414, 0.00000468, -0.00000288, -0.00000681, -0.00000127, 0.00000627, 0.00000662, 0.00000147, -0.00000086, 0.00000135, 0.00000137, -0.00000424, -0.00000854, -0.00000321, 0.00000882, 0.00001529, 0.00000944, -0.00000174, -0.00000628, -0.00000135, 0.00000476, 0.00000377, },
		{ 0.00000212, 0.00000403, -0.00000030, -0.00000541, -0.00000526, -0.00000028, 0.00000326, 0.00000010, -0.00000814, -0.00001399, -0.00001141, -0.00000215, 0.00000587, 0.00000698, 0.00000345, 0.00000139, 0.00000204, 0.00000024, -0.00000722, -0.00001486, -0.00001368, -0.00000272, 0.00000848, 0.00001037, 0.00000367, -0.00000336, -0.00000573, -0.00000586, -0.00000716, -0.00000807, -0.00000514, 0.00000065, 0.00000483, 0.00000553, 0.00000514, 0.00000522, 0.00000371, -0.00000073, -0.00000478, -0.00000467, -0.00000229, -0.00000229, -0.00000347, 0.00000096, 0.00001154, 0.00001659, 0.00000523, -0.00001537, -0.00002413, -0.00001105, 0.00000879, 0.00001285, -0.00000040, -0.00000975, 0.00000056, 0.00001966, 0.00002366, 0.00000567, -0.00001640, -0.00002202, -0.00000966, 0.00000677, 0.00001550, 0.00001386, 0.00000285, -0.00001407, -0.00002588, -0.00001865, 0.00000681, 0.00002817, 0.00002342, -0.00000125, -0.00001609, -0.00000419, 0.00001649, 0.00001690, -0.00000354, -0.00001533, 0.00000077, 0.00002596, 0.00002688, 0.00000000, -0.00002296, -0.00001599, 0.00000863, 0.00001815, 0.00000169, -0.00001911, -0.00002124, -0.00000829, 0.00000012, -0.00000278, -0.00000549, 0.00000047, 0.00000711, 0.00000359, -0.00000587, -0.00000768, 0.00000070, 0.00000675, 0.00000136, -0.00000765, -0.00000701, 0.00000254, 0.00000720, -0.00000008, -0.00000965, -0.00000894, 0.00000026, 0.00000520, 0.00000035, -0.00000589, -0.00000411, 0.00000321, 0.00000656, 0.00000325, -0.00000029, 0.00000051, 0.00000163, -0.00000207, -0.00000720, -0.00000532, 0.00000477, 0.00001364, 0.00001210, 0.00000233, -0.00000505, -0.00000370, },
		{ -0.00000272, -0.00000481, -0.00000061, 0.00000302, 0.00000145, -0.00000307, -0.00000497, -0.00000198, 0.00000222, 0.00000170, -0.00000523, -0.00001318, -0.00001418, -0.00000598, 0.00000477, 0.00000917, 0.00000566, 0.00000091, 0.00000051, 0.00000167, -0.00000214, -0.00001067, -0.00001533, -0.00000964, 0.00000213, 0.00000961, 0.00000805, 0.00000187, -0.00000286, -0.00000532, -0.00000751, -0.00000875, -0.00000659, -0.00000191, 0.00000137, 0.00000212, 0.00000360, 0.00000730, 0.00000881, 0.00000389, -0.00000389, -0.00000663, -0.00000319, -0.00000097, -0.00000393, -0.00000482, 0.00000430, 0.00001656, 0.00001492, -0.00000485, -0.00002387, -0.00002069, 0.00000121, 0.00001619, 0.00000823, -0.00000900, -0.00001039, 0.00000796, 0.00002422, 0.00001875, -0.00000291, -0.00001939, -0.00001863, -0.00000631, 0.00000668, 0.00001457, 0.00001378, 0.00000093, -0.00001871, -0.00002749, -0.00001185, 0.00001722, 0.00003118, 0.00001654, -0.00000745, -0.00001210, 0.00000462, 0.00001712, 0.00000805, -0.00000798, -0.00000554, 0.00001507, 0.00002699, 0.00001307, -0.00001067, -0.00001655, -0.00000146, 0.00001114, 0.00000398, -0.00001267, -0.00001812, -0.00000964, -0.00000251, -0.00000544, -0.00000914, -0.00000301, 0.00000715, 0.00000753, -0.00000259, -0.00000896, -0.00000288, 0.00000627, 0.00000480, -0.00000538, -0.00000970, -0.00000197, 0.00000689, 0.00000448, -0.00000590, -0.00001045, -0.00000416, 0.00000344, 0.00000261, -0.00000361, -0.00000526, 0.00000005, 0.00000505, 0.00000422, 0.00000084, 0.00000033, 0.00000145, -0.00000080, -0.00000579, -0.00000631, 0.00000145, 0.00001127, 0.00001339, 0.00000605, },
		{ 0.00001362, 0.00000925, 0.00000038, -0.00000464, -0.00000300, 0.00000101, 0.00000200, -0.00000079, -0.00000366, -0.00000296, 0.00000069, 0.00000242, -0.00000207, -0.00001088, -0.00001599, -0.00001093, 0.00000133, 0.00001043, 0.00000957, 0.00000272, -0.00000119, 0.00000018, 0.00000051, -0.00000521, -0.00001265, -0.00001318, -0.00000499, 0.00000476, 0.00000870, 0.00000649, 0.00000205, -0.00000256, -0.00000721, -0.00000992, -0.00000794, -0.00000288, -0.00000018, -0.00000144, -0.00000116, 0.00000514, 0.00001235, 0.00001095, 0.00000063, -0.00000764, -0.00000602, -0.00000055, -0.00000161, -0.00000698, -0.00000390, 0.00001039, 0.00001950, 0.00000758, -0.00001661, -0.00002643, -0.00001005, 0.00001321, 0.00001637, -0.00000143, -0.00001482, -0.00000532, 0.00001600, 0.00002413, 0.00001076, -0.00000929, -0.00001888, -0.00001538, -0.00000547, 0.00000610, 0.00001536, 0.00001389, -0.00000315, -0.00002372, -0.00002543, -0.00000172, 0.00002559, 0.00002921, 0.00000891, -0.00000891, -0.00000515, 0.00000945, 0.00001238, 0.00000150, -0.00000378, 0.00000777, 0.00002184, 0.00001857, 0.00000058, -0.00001071, -0.00000473, 0.00000501, 0.00000169, -0.00001114, -0.00001634, -0.00000870, -0.00000147, -0.00000546, -0.00001279, -0.00000917, 0.00000394, 0.00001046, 0.00000258, -0.00000798, -0.00000666, 0.00000360, 0.00000726, -0.00000124, -0.00000993, -0.00000672, 0.00000368, 0.00000690, -0.00000101, -0.00000912, -0.00000742, 0.00000017, 0.00000302, -0.00000116, -0.00000472, -0.00000216, 0.00000279, 0.00000397, 0.00000173, 0.00000070, 0.00000148, 0.00000004, -0.00000461, -0.00000681, -0.00000131, 0.00000856, },
		{ -0.00000374, 0.00000550, 0.00001282, 0.00001173, 0.00000401, -0.00000304, -0.00000453, -0.00000166, 0.00000108, 0.00000083, -0.00000152, -0.00000277, -0.00000092, 0.00000188, 0.00000052, -0.00000713, -0.00001557, -0.00001561, -0.00000464, 0.00000881, 0.00001341, 0.00000736, -0.00000055, -0.00000243, -0.00000035, -0.00000160, -0.00000765, -0.00001213, -0.00000951, -0.00000182, 0.00000497, 0.00000778, 0.00000699, 0.00000276, -0.00000455, -0.00001081, -0.00001041, -0.00000393, 0.00000051, -0.00000238, -0.00000643, -0.00000151, 0.00001103, 0.00001742, 0.00000906, -0.00000496, -0.00000927, -0.00000277, 0.00000098, -0.00000494, -0.00000923, 0.00000077, 0.00001694, 0.00001715, -0.00000439, -0.00002516, -0.00002058, 0.00000415, 0.00001959, 0.00000899, -0.00001143, -0.00001460, 0.00000336, 0.00002081, 0.00001919, 0.00000281, -0.00001189, -0.00001676, -0.00001420, -0.00000625, 0.00000685, 0.00001745, 0.00001235, -0.00000914, -0.00002669, -0.00001901, 0.00000911, 0.00002968, 0.00002404, 0.00000390, -0.00000605, 0.00000070, 0.00000887, 0.00000666, 0.00000176, 0.00000666, 0.00001749, 0.00001889, 0.00000705, -0.00000401, -0.00000273, 0.00000341, -0.00000025, -0.00001286, -0.00001850, -0.00000961, 0.00000112, -0.00000150, -0.00001313, -0.00001570, -0.00000306, 0.00000996, 0.00000799, -0.00000407, -0.00000879, -0.00000067, 0.00000733, 0.00000319, -0.00000736, -0.00000973, -0.00000116, 0.00000630, 0.00000301, -0.00000572, -0.00000843, -0.00000318, 0.00000166, 0.00000032, -0.00000329, -0.00000299, 0.00000073, 0.00000281, 0.00000183, 0.00000104, 0.00000179, 0.00000092, -0.00000344, -0.00000698, },
		{ -0.00000202, -0.00000665, -0.00000573, 0.00000217, 0.00001088, 0.00001306, 0.00000769, -0.00000002, -0.00000450, -0.00000422, -0.00000133, 0.00000096, 0.00000073, -0.00000114, -0.00000172, 0.00000030, 0.00000153, -0.00000302, -0.00001247, -0.00001800, -0.00001164, 0.00000339, 0.00001451, 0.00001318, 0.00000373, -0.00000316, -0.00000328, -0.00000169, -0.00000377, -0.00000804, -0.00000948, -0.00000648, -0.00000121, 0.00000438, 0.00000887, 0.00000887, 0.00000146, -0.00000931, -0.00001353, -0.00000700, 0.00000146, 0.00000022, -0.00000857, -0.00000995, 0.00000352, 0.00001914, 0.00001881, 0.00000290, -0.00000982, -0.00000715, 0.00000112, -0.00000065, -0.00000956, -0.00000803, 0.00000843, 0.00002017, 0.00000837, -0.00001687, -0.00002606, -0.00000773, 0.00001606, 0.00001744, -0.00000246, -0.00001690, -0.00000840, 0.00001152, 0.00002031, 0.00001179, -0.00000228, -0.00001143, -0.00001554, -0.00001539, -0.00000667, 0.00000961, 0.00001912, 0.00000822, -0.00001519, -0.00002572, -0.00000959, 0.00001771, 0.00002953, 0.00001891, 0.00000290, -0.00000192, 0.00000245, 0.00000523, 0.00000515, 0.00000926, 0.00001729, 0.00001896, 0.00000975, -0.00000017, 0.00000019, 0.00000564, 0.00000156, -0.00001360, -0.00002330, -0.00001495, 0.00000120, 0.00000449, -0.00000845, -0.00001902, -0.00001183, 0.00000482, 0.00001092, 0.00000163, -0.00000791, -0.00000475, 0.00000484, 0.00000605, -0.00000310, -0.00000993, -0.00000561, 0.00000322, 0.00000476, -0.00000189, -0.00000723, -0.00000531, -0.00000055, 0.00000043, -0.00000197, -0.00000274, -0.00000048, 0.00000145, 0.00000118, 0.00000087, 0.00000200, 0.00000189, },
		{ 0.00000171, 0.00000266, -0.00000039, -0.00000567, -0.00000697, -0.00000105, 0.00000795, 0.00001282, 0.00001053, 0.00000389, -0.00000245, -0.00000549, -0.00000452, -0.00000098, 0.00000187, 0.00000149, -0.00000080, -0.00000123, 0.00000064, -0.00000022, -0.00000771, -0.00001685, -0.00001707, -0.00000466, 0.00001106, 0.00001708, 0.00001050, 0.00000012, -0.00000504, -0.00000460, -0.00000355, -0.00000442, -0.00000588, -0.00000658, -0.00000598, -0.00000227, 0.00000548, 0.00001229, 0.00000956, -0.00000349, -0.00001484, -0.00001232, -0.00000020, 0.00000447, -0.00000549, -0.00001568, -0.00000810, 0.00001332, 0.00002525, 0.00001448, -0.00000496, -0.00001089, -0.00000217, 0.00000268, -0.00000552, -0.00001245, -0.00000229, 0.00001578, 0.00001697, -0.00000457, -0.00002440, -0.00001799, 0.00000710, 0.00002056, 0.00000797, -0.00001252, -0.00001549, 0.00000052, 0.00001546, 0.00001530, 0.00000536, -0.00000363, -0.00001026, -0.00001678, -0.00001753, -0.00000511, 0.00001342, 0.00001863, 0.00000223, -0.00001895, -0.00002068, 0.00000012, 0.00002260, 0.00002738, 0.00001633, 0.00000445, -0.00000003, 0.00000037, 0.00000324, 0.00001050, 0.00001992, 0.00002216, 0.00001266, 0.00000119, 0.00000058, 0.00000797, 0.00000672, -0.00001003, -0.00002668, -0.00002369, -0.00000420, 0.00000831, -0.00000021, -0.00001664, -0.00001865, -0.00000372, 0.00000937, 0.00000658, -0.00000424, -0.00000677, 0.00000109, 0.00000642, 0.00000101, -0.00000766, -0.00000811, -0.00000076, 0.00000403, 0.00000081, -0.00000486, -0.00000576, -0.00000248, -0.00000047, -0.00000140, -0.00000212, -0.00000079, 0.00000054, 0.00000024, 0.00000010, },
		{ -0.00000047, -0.00000098, 0.00000084, 0.00000291, 0.00000112, -0.00000421, -0.00000731, -0.00000360, 0.00000465, 0.00001105, 0.00001170, 0.00000751, 0.00000128, -0.00000444, -0.00000698, -0.00000461, 0.00000067, 0.00000372, 0.00000193, -0.00000121, -0.00000118, 0.00000019, -0.00000340, -0.00001271, -0.00001876, -0.00001243, 0.00000358, 0.00001635, 0.00001648, 0.00000691, -0.00000280, -0.00000705, -0.00000671, -0.00000447, -0.00000228, -0.00000246, -0.00000590, -0.00000814, -0.00000244, 0.00000981, 0.00001602, 0.00000631, -0.00001129, -0.00001749, -0.00000597, 0.00000660, 0.00000166, -0.00001502, -0.00001892, 0.00000062, 0.00002405, 0.00002546, 0.00000558, -0.00001047, -0.00000724, 0.00000263, 0.00000005, -0.00001128, -0.00001087, 0.00000624, 0.00001865, 0.00000720, -0.00001653, -0.00002315, -0.00000389, 0.00001763, 0.00001606, -0.00000406, -0.00001673, -0.00000858, 0.00000768, 0.00001397, 0.00000886, 0.00000227, -0.00000266, -0.00001079, -0.00002010, -0.00001855, -0.00000145, 0.00001637, 0.00001546, -0.00000357, -0.00001901, -0.00001331, 0.00000780, 0.00002463, 0.00002598, 0.00001641, 0.00000538, -0.00000180, -0.00000207, 0.00000693, 0.00002100, 0.00002774, 0.00001906, 0.00000366, -0.00000135, 0.00000699, 0.00001161, -0.00000228, -0.00002493, -0.00003173, -0.00001464, 0.00000625, 0.00000742, -0.00000913, -0.00002029, -0.00001228, 0.00000354, 0.00000831, 0.00000030, -0.00000611, -0.00000212, 0.00000475, 0.00000356, -0.00000429, -0.00000828, -0.00000402, 0.00000174, 0.00000172, -0.00000261, -0.00000497, -0.00000350, -0.00000164, -0.00000161, -0.00000173, -0.00000058, 0.00000029, },
		{ -0.00000027, 0.00000061, -0.00000066, -0.00000199, -0.00000044, 0.00000257, 0.00000226, -0.00000263, -0.00000698, -0.00000526, 0.00000177, 0.00000849, 0.00001100, 0.00000954, 0.00000536, -0.00000097, -0.00000710, -0.00000836, -0.00000305, 0.00000376, 0.00000519, 0.00000114, -0.00000209, -0.00000128, -0.00000136, -0.00000777, -0.00001650, -0.00001699, -0.00000512, 0.00001067, 0.00001828, 0.00001404, 0.00000370, -0.00000555, -0.00001003, -0.00000861, -0.00000261, 0.00000226, -0.00000048, -0.00000881, -0.00001085, 0.00000097, 0.00001644, 0.00001660, -0.00000173, -0.00001848, -0.00001427, 0.00000342, 0.00000877, -0.00000754, -0.00002370, -0.00001457, 0.00001387, 0.00003055, 0.00001874, -0.00000389, -0.00001071, -0.00000092, 0.00000392, -0.00000601, -0.00001437, -0.00000445, 0.00001351, 0.00001449, -0.00000565, -0.00002206, -0.00001323, 0.00001023, 0.00001962, 0.00000531, -0.00001303, -0.00001398, -0.00000026, 0.00000998, 0.00000884, 0.00000430, 0.00000257, -0.00000196, -0.00001394, -0.00002371, -0.00001724, 0.00000296, 0.00001701, 0.00001075, -0.00000713, -0.00001583, -0.00000599, 0.00001295, 0.00002607, 0.00002665, 0.00001690, 0.00000324, -0.00000536, -0.00000043, 0.00001683, 0.00003160, 0.00002857, 0.00001073, -0.00000204, 0.00000260, 0.00001246, 0.00000609, -0.00001750, -0.00003473, -0.00002632, -0.00000231, 0.00001011, -0.00000004, -0.00001612, -0.00001727, -0.00000405, 0.00000607, 0.00000336, -0.00000377, -0.00000365, 0.00000239, 0.00000428, -0.00000124, -0.00000680, -0.00000573, -0.00000081, 0.00000113, -0.00000128, -0.00000375, -0.00000361, -0.00000252, -0.00000227, -0.00000181, },
		{ -0.00000304, -0.00000237, -0.00000017, 0.00000128, -0.00000024, -0.00000267, -0.00000192, 0.00000174, 0.00000300, -0.00000113, -0.00000639, -0.00000631, -0.00000033, 0.00000618, 0.00000919, 0.00000941, 0.00000807, 0.00000363, -0.00000419, -0.00001011, -0.00000796, 0.00000069, 0.00000687, 0.00000500, -0.00000062, -0.00000274, -0.00000180, -0.00000443, -0.00001224, -0.00001729, -0.00001161, 0.00000250, 0.00001472, 0.00001759, 0.00001152, 0.00000093, -0.00000936, -0.00001371, -0.00000818, 0.00000279, 0.00000683, -0.00000250, -0.00001424, -0.00001067, 0.00000865, 0.00002199, 0.00001138, -0.00001226, -0.00002072, -0.00000517, 0.00001105, 0.00000332, -0.00001979, -0.00002609, -0.00000253, 0.00002631, 0.00002930, 0.00000778, -0.00000942, -0.00000574, 0.00000422, 0.00000035, -0.00001232, -0.00001236, 0.00000416, 0.00001542, 0.00000433, -0.00001591, -0.00001843, 0.00000105, 0.00001837, 0.00001304, -0.00000621, -0.00001541, -0.00000683, 0.00000518, 0.00000729, 0.00000344, 0.00000326, 0.00000429, -0.00000350, -0.00001875, -0.00002573, -0.00001401, 0.00000629, 0.00001533, 0.00000648, -0.00000773, -0.00001119, -0.00000012, 0.00001697, 0.00002884, 0.00002826, 0.00001501, -0.00000168, -0.00000664, 0.00000766, 0.00002956, 0.00003715, 0.00002262, 0.00000262, -0.00000168, 0.00000836, 0.00001068, -0.00000757, -0.00003106, -0.00003442, -0.00001439, 0.00000600, 0.00000612, -0.00000862, -0.00001707, -0.00001005, 0.00000129, 0.00000364, -0.00000154, -0.00000369, 0.00000046, 0.00000381, 0.00000079, -0.00000472, -0.00000597, -0.00000273, -0.00000019, -0.00000099, -0.00000278, -0.00000317, -0.00000292, },
		{ -0.00000255, -0.00000285, -0.00000374, -0.00000338, -0.00000053, 0.00000208, 0.00000086, -0.00000279, -0.00000346, 0.00000042, 0.00000341, 0.00000040, -0.00000571, -0.00000728, -0.00000199, 0.00000469, 0.00000757, 0.00000780, 0.00000834, 0.00000729, 0.00000081, -0.00000847, -0.00001165, -0.00000467, 0.00000532, 0.00000819, 0.00000307, -0.00000223, -0.00000308, -0.00000362, -0.00000867, -0.00001480, -0.00001406, -0.00000443, 0.00000782, 0.00001557, 0.00001623, 0.00000976, -0.00000268, -0.00001484, -0.00001617, -0.00000349, 0.00001031, 0.00000802, -0.00000908, -0.00001881, -0.00000512, 0.00001827, 0.00002240, 0.00000066, -0.00002056, -0.00001555, 0.00000610, 0.00001196, -0.00000880, -0.00002911, -0.00001926, 0.00001311, 0.00003241, 0.00002061, -0.00000236, -0.00000861, 0.00000132, 0.00000490, -0.00000643, -0.00001517, -0.00000572, 0.00001052, 0.00001036, -0.00000739, -0.00001876, -0.00000727, 0.00001334, 0.00001764, 0.00000186, -0.00001332, -0.00001147, 0.00000030, 0.00000561, 0.00000210, 0.00000061, 0.00000500, 0.00000495, -0.00000754, -0.00002334, -0.00002549, -0.00001037, 0.00000758, 0.00001268, 0.00000408, -0.00000626, -0.00000684, 0.00000473, 0.00002184, 0.00003289, 0.00002820, 0.00000967, -0.00000626, -0.00000206, 0.00002082, 0.00003993, 0.00003561, 0.00001393, -0.00000106, 0.00000235, 0.00000944, 0.00000008, -0.00002295, -0.00003614, -0.00002522, -0.00000316, 0.00000654, -0.00000187, -0.00001296, -0.00001236, -0.00000337, 0.00000170, -0.00000072, -0.00000332, -0.00000078, 0.00000291, 0.00000188, -0.00000275, -0.00000527, -0.00000375, -0.00000156, -0.00000144, -0.00000231, },
		{ -0.00000231, -0.00000240, -0.00000196, -0.00000229, -0.00000412, -0.00000478, -0.00000167, 0.00000261, 0.00000260, -0.00000200, -0.00000480, -0.00000149, 0.00000335, 0.00000216, -0.00000461, -0.00000838, -0.00000399, 0.00000366, 0.00000706, 0.00000624, 0.00000654, 0.00000836, 0.00000563, -0.00000382, -0.00001208, -0.00000993, 0.00000069, 0.00000856, 0.00000703, 0.00000075, -0.00000296, -0.00000409, -0.00000726, -0.00001232, -0.00001349, -0.00000769, 0.00000154, 0.00000954, 0.00001495, 0.00001592, 0.00000786, -0.00000864, -0.00002086, -0.00001463, 0.00000553, 0.00001641, 0.00000344, -0.00001782, -0.00001854, 0.00000543, 0.00002530, 0.00001555, -0.00001178, -0.00002209, -0.00000441, 0.00001373, 0.00000412, -0.00002261, -0.00003010, -0.00000475, 0.00002600, 0.00002955, 0.00000857, -0.00000718, -0.00000266, 0.00000619, 0.00000053, -0.00001269, -0.00001276, 0.00000219, 0.00001110, 0.00000052, -0.00001506, -0.00001290, 0.00000616, 0.00001854, 0.00000955, -0.00000829, -0.00001381, -0.00000460, 0.00000379, 0.00000192, -0.00000230, 0.00000111, 0.00000729, 0.00000321, -0.00001274, -0.00002626, -0.00002379, -0.00000781, 0.00000729, 0.00001066, 0.00000358, -0.00000429, -0.00000311, 0.00001027, 0.00002836, 0.00003614, 0.00002443, 0.00000270, -0.00000645, 0.00000900, 0.00003471, 0.00004425, 0.00002903, 0.00000715, -0.00000045, 0.00000434, 0.00000227, -0.00001498, -0.00003240, -0.00003121, -0.00001330, 0.00000168, 0.00000098, -0.00000809, -0.00001128, -0.00000589, -0.00000067, -0.00000123, -0.00000344, -0.00000183, 0.00000187, 0.00000229, -0.00000116, -0.00000409, -0.00000391, -0.00000259, },
		{ -0.00000332, -0.00000303, -0.00000333, -0.00000314, -0.00000166, -0.00000123, -0.00000376, -0.00000623, -0.00000383, 0.00000220, 0.00000464, 0.00000006, -0.00000538, -0.00000396, 0.00000233, 0.00000396, -0.00000256, -0.00000909, -0.00000676, 0.00000199, 0.00000741, 0.00000612, 0.00000438, 0.00000674, 0.00000810, 0.00000168, -0.00000893, -0.00001256, -0.00000495, 0.00000557, 0.00000883, 0.00000459, -0.00000049, -0.00000352, -0.00000712, -0.00001172, -0.00001278, -0.00000789, -0.00000104, 0.00000379, 0.00000865, 0.00001534, 0.00001669, 0.00000372, -0.00001714, -0.00002446, -0.00000726, 0.00001621, 0.00001717, -0.00000645, -0.00002421, -0.00001140, 0.00001729, 0.00002551, 0.00000306, -0.00002038, -0.00001535, 0.00000761, 0.00001304, -0.00000987, -0.00003143, -0.00002106, 0.00001188, 0.00003089, 0.00001952, -0.00000120, -0.00000521, 0.00000450, 0.00000604, -0.00000645, -0.00001502, -0.00000644, 0.00000703, 0.00000546, -0.00000916, -0.00001499, -0.00000134, 0.00001582, 0.00001536, -0.00000118, -0.00001325, -0.00000918, 0.00000098, 0.00000265, -0.00000304, -0.00000385, 0.00000359, 0.00000810, -0.00000058, -0.00001740, -0.00002740, -0.00002207, -0.00000669, 0.00000672, 0.00000995, 0.00000384, -0.00000267, 0.00000121, 0.00001786, 0.00003524, 0.00003605, 0.00001733, -0.00000209, 0.00000008, 0.00002374, 0.00004485, 0.00004245, 0.00002153, 0.00000393, 0.00000019, -0.00000045, -0.00001108, -0.00002701, -0.00003201, -0.00002063, -0.00000543, -0.00000059, -0.00000545, -0.00000902, -0.00000606, -0.00000184, -0.00000191, -0.00000402, -0.00000309, 0.00000051, 0.00000206, -0.00000008, -0.00000272, },
		{ 0.00000018, -0.00000160, -0.00000212, -0.00000257, -0.00000408, -0.00000453, -0.00000216, 0.00000001, -0.00000226, -0.00000692, -0.00000678, 0.00000012, 0.00000603, 0.00000331, -0.00000437, -0.00000636, -0.00000002, 0.00000504, 0.00000048, -0.00000846, -0.00000979, -0.00000117, 0.00000731, 0.00000752, 0.00000367, 0.00000404, 0.00000751, 0.00000551, -0.00000398, -0.00001163, -0.00000886, 0.00000089, 0.00000743, 0.00000675, 0.00000303, -0.00000058, -0.00000577, -0.00001229, -0.00001427, -0.00000821, -0.00000011, 0.00000234, 0.00000202, 0.00000801, 0.00001814, 0.00001649, -0.00000434, -0.00002596, -0.00002232, 0.00000510, 0.00002401, 0.00001063, -0.00001794, -0.00002409, 0.00000102, 0.00002515, 0.00001762, -0.00001009, -0.00002093, -0.00000321, 0.00001430, 0.00000335, -0.00002380, -0.00003054, -0.00000520, 0.00002391, 0.00002645, 0.00000746, -0.00000472, 0.00000131, 0.00000862, 0.00000114, -0.00001221, -0.00001246, 0.00000008, 0.00000624, -0.00000327, -0.00001372, -0.00000748, 0.00001030, 0.00001803, 0.00000662, -0.00000940, -0.00001225, -0.00000319, 0.00000265, -0.00000147, -0.00000649, -0.00000264, 0.00000596, 0.00000677, -0.00000499, -0.00002070, -0.00002778, -0.00002123, -0.00000623, 0.00000688, 0.00000994, 0.00000378, -0.00000073, 0.00000779, 0.00002713, 0.00003989, 0.00003162, 0.00000982, -0.00000126, 0.00001244, 0.00003783, 0.00004931, 0.00003734, 0.00001580, 0.00000170, -0.00000421, -0.00001205, -0.00002402, -0.00003032, -0.00002389, -0.00001136, -0.00000458, -0.00000600, -0.00000803, -0.00000540, -0.00000159, -0.00000168, -0.00000428, -0.00000435, -0.00000120, 0.00000106, },
		{ -0.00000270, -0.00000048, -0.00000067, -0.00000140, -0.00000085, -0.00000105, -0.00000385, -0.00000617, -0.00000392, 0.00000056, 0.00000024, -0.00000590, -0.00000949, -0.00000381, 0.00000549, 0.00000683, -0.00000128, -0.00000751, -0.00000341, 0.00000445, 0.00000372, -0.00000592, -0.00001176, -0.00000558, 0.00000538, 0.00000913, 0.00000502, 0.00000238, 0.00000515, 0.00000641, 0.00000011, -0.00000847, -0.00000965, -0.00000270, 0.00000419, 0.00000583, 0.00000485, 0.00000353, -0.00000144, -0.00001118, -0.00001747, -0.00001179, 0.00000072, 0.00000580, 0.00000041, -0.00000133, 0.00001046, 0.00002227, 0.00001240, -0.00001581, -0.00003115, -0.00001292, 0.00001860, 0.00002476, -0.00000162, -0.00002577, -0.00001619, 0.00001392, 0.00002506, 0.00000455, -0.00001804, -0.00001317, 0.00000826, 0.00001180, -0.00001124, -0.00003101, -0.00001979, 0.00001095, 0.00002694, 0.00001581, -0.00000105, -0.00000162, 0.00000803, 0.00000763, -0.00000558, -0.00001397, -0.00000706, 0.00000309, 0.00000060, -0.00001019, -0.00001100, 0.00000348, 0.00001693, 0.00001313, -0.00000280, -0.00001232, -0.00000768, 0.00000062, 0.00000053, -0.00000578, -0.00000727, -0.00000026, 0.00000682, 0.00000409, -0.00000879, -0.00002300, -0.00002847, -0.00002106, -0.00000538, 0.00000764, 0.00000949, 0.00000349, 0.00000323, 0.00001719, 0.00003592, 0.00004025, 0.00002441, 0.00000569, 0.00000617, 0.00002727, 0.00004813, 0.00004926, 0.00003180, 0.00001081, -0.00000420, -0.00001528, -0.00002523, -0.00002983, -0.00002474, -0.00001443, -0.00000817, -0.00000850, -0.00000930, -0.00000575, -0.00000101, -0.00000050, -0.00000353, -0.00000486, },
		{ -0.00000204, -0.00000425, -0.00000315, -0.00000166, -0.00000228, -0.00000270, -0.00000058, 0.00000105, -0.00000198, -0.00000699, -0.00000678, -0.00000065, 0.00000271, -0.00000278, -0.00001039, -0.00000863, 0.00000217, 0.00000891, 0.00000332, -0.00000629, -0.00000660, 0.00000192, 0.00000578, -0.00000196, -0.00001150, -0.00000988, 0.00000135, 0.00000919, 0.00000734, 0.00000285, 0.00000327, 0.00000532, 0.00000177, -0.00000573, -0.00000838, -0.00000350, 0.00000203, 0.00000304, 0.00000309, 0.00000562, 0.00000486, -0.00000572, -0.00001892, -0.00001879, -0.00000298, 0.00001037, 0.00000601, -0.00000575, -0.00000232, 0.00001676, 0.00002445, 0.00000279, -0.00002737, -0.00002902, 0.00000192, 0.00002791, 0.00001716, -0.00001482, -0.00002595, -0.00000322, 0.00002172, 0.00001683, -0.00000795, -0.00001741, -0.00000122, 0.00001313, 0.00000097, -0.00002389, -0.00002798, -0.00000380, 0.00002098, 0.00002110, 0.00000465, -0.00000294, 0.00000505, 0.00001118, 0.00000258, -0.00001057, -0.00001170, -0.00000243, 0.00000124, -0.00000620, -0.00001154, -0.00000273, 0.00001257, 0.00001654, 0.00000475, -0.00000879, -0.00001050, -0.00000316, 0.00000082, -0.00000332, -0.00000828, -0.00000611, 0.00000158, 0.00000617, 0.00000130, -0.00001178, -0.00002528, -0.00002972, -0.00002065, -0.00000388, 0.00000799, 0.00000818, 0.00000443, 0.00001055, 0.00002809, 0.00004162, 0.00003636, 0.00001777, 0.00000735, 0.00001841, 0.00004115, 0.00005424, 0.00004699, 0.00002585, 0.00000262, -0.00001647, -0.00002927, -0.00003284, -0.00002623, -0.00001550, -0.00000969, -0.00001076, -0.00001203, -0.00000792, -0.00000149, 0.00000066, },
		{ -0.00000353, 0.00000075, -0.00000084, -0.00000311, -0.00000233, -0.00000137, -0.00000338, -0.00000522, -0.00000237, 0.00000233, 0.00000134, -0.00000573, -0.00000954, -0.00000404, 0.00000346, 0.00000156, -0.00000832, -0.00001236, -0.00000346, 0.00000796, 0.00000771, -0.00000256, -0.00000796, -0.00000167, 0.00000572, 0.00000197, -0.00000895, -0.00001242, -0.00000359, 0.00000691, 0.00000872, 0.00000468, 0.00000326, 0.00000449, 0.00000171, -0.00000519, -0.00000783, -0.00000252, 0.00000306, 0.00000179, -0.00000117, 0.00000288, 0.00000924, 0.00000367, -0.00001455, -0.00002535, -0.00001285, 0.00001001, 0.00001559, -0.00000067, -0.00001213, 0.00000201, 0.00002453, 0.00002110, -0.00001099, -0.00003421, -0.00001856, 0.00001731, 0.00002895, 0.00000386, -0.00002337, -0.00001814, 0.00000933, 0.00002118, 0.00000441, -0.00001436, -0.00000925, 0.00000853, 0.00000883, -0.00001295, -0.00002863, -0.00001619, 0.00001060, 0.00002182, 0.00001058, -0.00000202, 0.00000105, 0.00001098, 0.00000960, -0.00000345, -0.00001198, -0.00000784, -0.00000127, -0.00000363, -0.00000991, -0.00000680, 0.00000663, 0.00001610, 0.00001087, -0.00000264, -0.00001004, -0.00000679, -0.00000120, -0.00000163, -0.00000656, -0.00000869, -0.00000455, 0.00000223, 0.00000496, -0.00000101, -0.00001466, -0.00002811, -0.00003087, -0.00001942, -0.00000250, 0.00000701, 0.00000693, 0.00000844, 0.00002100, 0.00003797, 0.00004294, 0.00003048, 0.00001484, 0.00001472, 0.00003254, 0.00005253, 0.00005748, 0.00004295, 0.00001626, -0.00001206, -0.00003273, -0.00003902, -0.00003074, -0.00001694, -0.00000944, -0.00001131, -0.00001457, -0.00001129, },
		{ -0.00001577, -0.00001466, -0.00000653, -0.00000034, -0.00000073, -0.00000274, -0.00000124, 0.00000064, -0.00000236, -0.00000735, -0.00000625, 0.00000110, 0.00000460, -0.00000192, -0.00001032, -0.00000864, 0.00000122, 0.00000505, -0.00000356, -0.00001304, -0.00000945, 0.00000364, 0.00000974, 0.00000233, -0.00000666, -0.00000460, 0.00000384, 0.00000445, -0.00000532, -0.00001261, -0.00000768, 0.00000315, 0.00000801, 0.00000588, 0.00000448, 0.00000552, 0.00000241, -0.00000609, -0.00001016, -0.00000305, 0.00000637, 0.00000506, -0.00000385, -0.00000425, 0.00000740, 0.00001269, -0.00000321, -0.00002566, -0.00002550, 0.00000068, 0.00002186, 0.00001255, -0.00001155, -0.00001391, 0.00001150, 0.00002937, 0.00001085, -0.00002460, -0.00003257, -0.00000256, 0.00002753, 0.00002124, -0.00000970, -0.00002384, -0.00000596, 0.00001615, 0.00001324, -0.00000625, -0.00001246, 0.00000146, 0.00001100, -0.00000231, -0.00002318, -0.00002354, -0.00000120, 0.00001792, 0.00001502, 0.00000100, -0.00000231, 0.00000744, 0.00001322, 0.00000490, -0.00000761, -0.00001056, -0.00000541, -0.00000361, -0.00000781, -0.00000829, 0.00000119, 0.00001264, 0.00001378, 0.00000382, -0.00000632, -0.00000813, -0.00000418, -0.00000210, -0.00000463, -0.00000809, -0.00000807, -0.00000366, 0.00000220, 0.00000401, -0.00000318, -0.00001814, -0.00003110, -0.00003111, -0.00001777, -0.00000244, 0.00000495, 0.00000765, 0.00001634, 0.00003273, 0.00004467, 0.00004071, 0.00002564, 0.00001691, 0.00002598, 0.00004665, 0.00006181, 0.00005820, 0.00003435, -0.00000070, -0.00003210, -0.00004605, -0.00003877, -0.00002092, -0.00000910, -0.00001014, },
		{ -0.00001041, -0.00000825, -0.00001555, -0.00001738, -0.00000964, -0.00000177, -0.00000136, -0.00000391, -0.00000167, 0.00000302, 0.00000136, -0.00000683, -0.00001053, -0.00000337, 0.00000548, 0.00000321, -0.00000777, -0.00001223, -0.00000371, 0.00000559, 0.00000191, -0.00001003, -0.00001325, -0.00000256, 0.00000825, 0.00000611, -0.00000341, -0.00000555, 0.00000162, 0.00000527, -0.00000212, -0.00001130, -0.00000998, -0.00000030, 0.00000592, 0.00000512, 0.00000472, 0.00000770, 0.00000587, -0.00000536, -0.00001461, -0.00000818, 0.00000785, 0.00001254, -0.00000038, -0.00001121, -0.00000183, 0.00001517, 0.00001125, -0.00001601, -0.00003372, -0.00001608, 0.00001789, 0.00002657, 0.00000127, -0.00002115, -0.00000832, 0.00002251, 0.00002731, -0.00000406, -0.00003268, -0.00002211, 0.00001340, 0.00002893, 0.00000826, -0.00001807, -0.00001683, 0.00000495, 0.00001501, 0.00000248, -0.00001050, -0.00000453, 0.00000866, 0.00000521, -0.00001447, -0.00002508, -0.00001154, 0.00001054, 0.00001667, 0.00000529, -0.00000365, 0.00000215, 0.00001242, 0.00001149, -0.00000012, -0.00000904, -0.00000874, -0.00000584, -0.00000696, -0.00000816, -0.00000238, 0.00000812, 0.00001332, 0.00000830, -0.00000109, -0.00000636, -0.00000587, -0.00000403, -0.00000439, -0.00000658, -0.00000834, -0.00000763, -0.00000340, 0.00000232, 0.00000328, -0.00000602, -0.00002221, -0.00003335, -0.00003037, -0.00001693, -0.00000421, 0.00000330, 0.00001194, 0.00002736, 0.00004342, 0.00004771, 0.00003717, 0.00002388, 0.00002344, 0.00003954, 0.00006057, 0.00006905, 0.00005374, 0.00001695, -0.00002512, -0.00005097, -0.00004913, -0.00002846, },
		{ -0.00005989, -0.00003947, -0.00001445, -0.00000660, -0.00001431, -0.00001965, -0.00001287, -0.00000280, -0.00000131, -0.00000580, -0.00000476, 0.00000320, 0.00000628, -0.00000231, -0.00001225, -0.00000969, 0.00000229, 0.00000699, -0.00000236, -0.00001258, -0.00000925, 0.00000250, 0.00000545, -0.00000475, -0.00001338, -0.00000799, 0.00000407, 0.00000710, -0.00000028, -0.00000474, 0.00000049, 0.00000561, 0.00000020, -0.00000998, -0.00001123, -0.00000238, 0.00000444, 0.00000317, 0.00000226, 0.00000819, 0.00001137, 0.00000002, -0.00001721, -0.00001776, 0.00000275, 0.00001959, 0.00001056, -0.00001149, -0.00001438, 0.00000727, 0.00002119, 0.00000168, -0.00003071, -0.00003293, 0.00000190, 0.00003192, 0.00002073, -0.00001392, -0.00002407, 0.00000344, 0.00002975, 0.00001747, -0.00001852, -0.00003173, -0.00000639, 0.00002364, 0.00002186, -0.00000432, -0.00001866, -0.00000650, 0.00001004, 0.00000784, -0.00000555, -0.00000748, 0.00000417, 0.00000867, -0.00000536, -0.00002164, -0.00001842, 0.00000150, 0.00001497, 0.00000951, -0.00000233, -0.00000270, 0.00000786, 0.00001402, 0.00000764, -0.00000353, -0.00000901, -0.00000857, -0.00000804, -0.00000808, -0.00000411, 0.00000438, 0.00001089, 0.00000982, 0.00000325, -0.00000261, -0.00000492, -0.00000531, -0.00000570, -0.00000632, -0.00000724, -0.00000839, -0.00000772, -0.00000303, 0.00000286, 0.00000202, -0.00000983, -0.00002604, -0.00003432, -0.00002961, -0.00001801, -0.00000706, 0.00000396, 0.00002015, 0.00003965, 0.00005171, 0.00004820, 0.00003438, 0.00002565, 0.00003356, 0.00005523, 0.00007428, 0.00007155, 0.00003911, -0.00001086, -0.00005118, },
		{ 0.00001079, -0.00004450, -0.00006890, -0.00005349, -0.00002200, -0.00000578, -0.00001203, -0.00002164, -0.00001713, -0.00000407, 0.00000051, -0.00000614, -0.00000951, -0.00000074, 0.00000912, 0.00000506, -0.00000903, -0.00001455, -0.00000438, 0.00000681, 0.00000340, -0.00000912, -0.00001263, -0.00000275, 0.00000547, 0.00000010, -0.00001047, -0.00001061, -0.00000043, 0.00000557, 0.00000092, -0.00000399, 0.00000041, 0.00000691, 0.00000282, -0.00000909, -0.00001312, -0.00000402, 0.00000492, 0.00000267, -0.00000212, 0.00000422, 0.00001517, 0.00001001, -0.00001316, -0.00002721, -0.00001054, 0.00001932, 0.00002477, -0.00000097, -0.00002219, -0.00000896, 0.00001954, 0.00001967, -0.00001491, -0.00004041, -0.00002055, 0.00002275, 0.00003632, 0.00000574, -0.00002635, -0.00001806, 0.00001665, 0.00002924, 0.00000284, -0.00002702, -0.00002211, 0.00000872, 0.00002521, 0.00001033, -0.00001167, -0.00001290, 0.00000189, 0.00000824, -0.00000064, -0.00000747, -0.00000035, 0.00000853, 0.00000202, -0.00001497, -0.00002085, -0.00000712, 0.00001012, 0.00001211, 0.00000116, -0.00000514, 0.00000167, 0.00001187, 0.00001264, 0.00000384, -0.00000538, -0.00000949, -0.00001017, -0.00000929, -0.00000528, 0.00000209, 0.00000843, 0.00000919, 0.00000525, 0.00000090, -0.00000190, -0.00000431, -0.00000670, -0.00000763, -0.00000711, -0.00000745, -0.00000891, -0.00000782, -0.00000203, 0.00000324, -0.00000040, -0.00001407, -0.00002870, -0.00003445, -0.00003020, -0.00002112, -0.00000934, 0.00000814, 0.00003136, 0.00005163, 0.00005766, 0.00004764, 0.00003324, 0.00003083, 0.00004758, 0.00007339, 0.00008520, 0.00006347, },
		{ 0.00009212, 0.00008689, 0.00003900, -0.00002879, -0.00007339, -0.00006962, -0.00003414, -0.00000675, -0.00000815, -0.00002242, -0.00002306, -0.00000761, 0.00000322, -0.00000288, -0.00001275, -0.00000824, 0.00000674, 0.00001157, -0.00000111, -0.00001454, -0.00001141, 0.00000213, 0.00000632, -0.00000373, -0.00001235, -0.00000735, 0.00000265, 0.00000238, -0.00000695, -0.00001048, -0.00000301, 0.00000368, 0.00000023, -0.00000522, -0.00000058, 0.00000910, 0.00000752, -0.00000707, -0.00001632, -0.00000766, 0.00000627, 0.00000595, -0.00000483, -0.00000407, 0.00001266, 0.00002026, -0.00000059, -0.00002950, -0.00002796, 0.00000715, 0.00003372, 0.00001827, -0.00001772, -0.00002534, 0.00000472, 0.00002813, 0.00000787, -0.00003266, -0.00003895, 0.00000086, 0.00003844, 0.00002847, -0.00001260, -0.00003051, -0.00000534, 0.00002551, 0.00002067, -0.00001102, -0.00002661, -0.00000792, 0.00001800, 0.00001920, -0.00000027, -0.00001218, -0.00000503, 0.00000490, 0.00000217, -0.00000581, -0.00000375, 0.00000595, 0.00000648, -0.00000713, -0.00001893, -0.00001342, 0.00000326, 0.00001192, 0.00000532, -0.00000439, -0.00000350, 0.00000640, 0.00001316, 0.00001001, 0.00000099, -0.00000706, -0.00001135, -0.00001163, -0.00000731, 0.00000045, 0.00000700, 0.00000820, 0.00000523, 0.00000243, 0.00000115, -0.00000117, -0.00000554, -0.00000887, -0.00000865, -0.00000720, -0.00000803, -0.00000964, -0.00000720, -0.00000069, 0.00000260, -0.00000387, -0.00001769, -0.00003001, -0.00003495, -0.00003298, -0.00002530, -0.00000950, 0.00001589, 0.00004436, 0.00006276, 0.00006201, 0.00004676, 0.00003398, 0.00004035, 0.00006670, },
		{ 0.00003788, 0.00005651, 0.00009040, 0.00010503, 0.00007116, -0.00000256, -0.00006952, -0.00008565, -0.00005200, -0.00001178, -0.00000284, -0.00001988, -0.00002938, -0.00001522, 0.00000374, 0.00000359, -0.00001111, -0.00001570, -0.00000134, 0.00001295, 0.00000804, -0.00000863, -0.00001472, -0.00000460, 0.00000487, 0.00000037, -0.00000946, -0.00000924, -0.00000049, 0.00000220, -0.00000512, -0.00000982, -0.00000354, 0.00000381, 0.00000002, -0.00000840, -0.00000491, 0.00000968, 0.00001452, -0.00000114, -0.00001887, -0.00001497, 0.00000487, 0.00001240, -0.00000168, -0.00001243, 0.00000208, 0.00002393, 0.00001675, -0.00001960, -0.00004077, -0.00001521, 0.00002918, 0.00003760, 0.00000080, -0.00003146, -0.00001715, 0.00002097, 0.00002674, -0.00001153, -0.00004354, -0.00002454, 0.00002404, 0.00004256, 0.00001118, -0.00002693, -0.00002488, 0.00000858, 0.00002638, 0.00000765, -0.00001891, -0.00001832, 0.00000514, 0.00001938, 0.00000977, -0.00000604, -0.00000783, 0.00000052, 0.00000258, -0.00000384, -0.00000576, 0.00000209, 0.00000767, -0.00000013, -0.00001373, -0.00001607, -0.00000375, 0.00000871, 0.00000820, -0.00000122, -0.00000574, 0.00000027, 0.00000950, 0.00001253, 0.00000743, -0.00000157, -0.00000976, -0.00001355, -0.00001055, -0.00000197, 0.00000619, 0.00000819, 0.00000482, 0.00000194, 0.00000227, 0.00000222, -0.00000200, -0.00000801, -0.00001047, -0.00000876, -0.00000748, -0.00000905, -0.00000979, -0.00000580, 0.00000009, 0.00000058, -0.00000751, -0.00001996, -0.00003071, -0.00003707, -0.00003785, -0.00002920, -0.00000678, 0.00002674, 0.00005859, 0.00007319, 0.00006495, 0.00004561, },
		{ 0.00008260, 0.00006587, 0.00004519, 0.00004814, 0.00008077, 0.00011339, 0.00010207, 0.00003342, -0.00005312, -0.00009697, -0.00007514, -0.00002419, 0.00000143, -0.00001247, -0.00003263, -0.00002629, -0.00000140, 0.00000971, -0.00000369, -0.00001839, -0.00001182, 0.00000726, 0.00001338, 0.00000046, -0.00001218, -0.00000922, 0.00000062, 0.00000100, -0.00000696, -0.00000894, -0.00000171, 0.00000192, -0.00000504, -0.00001110, -0.00000445, 0.00000619, 0.00000332, -0.00001057, -0.00001301, 0.00000492, 0.00002083, 0.00001015, -0.00001651, -0.00002439, -0.00000295, 0.00001756, 0.00000855, -0.00001440, -0.00001292, 0.00001605, 0.00003039, 0.00000129, -0.00004016, -0.00003901, 0.00000870, 0.00004570, 0.00002692, -0.00002077, -0.00003455, 0.00000016, 0.00003159, 0.00001357, -0.00003106, -0.00004156, -0.00000131, 0.00004034, 0.00003360, -0.00000863, -0.00003206, -0.00001258, 0.00001794, 0.00001961, -0.00000427, -0.00001860, -0.00000645, 0.00001266, 0.00001451, 0.00000163, -0.00000633, -0.00000253, 0.00000159, -0.00000221, -0.00000646, -0.00000196, 0.00000591, 0.00000440, -0.00000709, -0.00001478, -0.00000888, 0.00000357, 0.00000846, 0.00000245, -0.00000479, -0.00000396, 0.00000383, 0.00001077, 0.00001122, 0.00000491, -0.00000497, -0.00001312, -0.00001387, -0.00000591, 0.00000458, 0.00000903, 0.00000562, 0.00000097, 0.00000107, 0.00000368, 0.00000224, -0.00000440, -0.00001038, -0.00001105, -0.00000875, -0.00000835, -0.00000983, -0.00000892, -0.00000431, -0.00000044, -0.00000235, -0.00001027, -0.00002096, -0.00003197, -0.00004138, -0.00004405, -0.00003185, -0.00000096, 0.00004056, 0.00007409, },
		{ 0.00000844, 0.00005784, 0.00009047, 0.00008962, 0.00006483, 0.00004898, 0.00006839, 0.00011008, 0.00012487, 0.00007442, -0.00002204, -0.00009711, -0.00009966, -0.00004627, -0.00000025, -0.00000154, -0.00002909, -0.00003684, -0.00001319, 0.00001045, 0.00000638, -0.00001373, -0.00001918, -0.00000316, 0.00001167, 0.00000760, -0.00000549, -0.00000884, -0.00000247, -0.00000081, -0.00000723, -0.00000931, -0.00000136, 0.00000401, -0.00000406, -0.00001447, -0.00000901, 0.00000769, 0.00001093, -0.00000723, -0.00002144, -0.00000709, 0.00002096, 0.00002424, -0.00000542, -0.00003047, -0.00001751, 0.00001514, 0.00002182, -0.00000548, -0.00002416, -0.00000231, 0.00003154, 0.00002484, -0.00002292, -0.00005234, -0.00002144, 0.00003506, 0.00004823, 0.00000474, -0.00003670, -0.00002437, 0.00001904, 0.00003040, -0.00000703, -0.00004216, -0.00002622, 0.00002234, 0.00004396, 0.00001567, -0.00002351, -0.00002735, 0.00000055, 0.00001969, 0.00000913, -0.00001056, -0.00001172, 0.00000384, 0.00001348, 0.00000731, -0.00000244, -0.00000336, 0.00000044, -0.00000092, -0.00000593, -0.00000517, 0.00000216, 0.00000561, -0.00000120, -0.00001056, -0.00001074, -0.00000156, 0.00000612, 0.00000466, -0.00000205, -0.00000512, -0.00000113, 0.00000616, 0.00001102, 0.00000965, 0.00000153, -0.00000940, -0.00001531, -0.00001066, 0.00000104, 0.00000932, 0.00000787, 0.00000139, -0.00000104, 0.00000237, 0.00000476, 0.00000054, -0.00000729, -0.00001180, -0.00001112, -0.00000929, -0.00000938, -0.00000964, -0.00000736, -0.00000371, -0.00000236, -0.00000521, -0.00001166, -0.00002144, -0.00003472, -0.00004788, -0.00005093, -0.00003270, },
		{ -0.00005648, -0.00005809, -0.00003089, 0.00002265, 0.00007870, 0.00010588, 0.00009291, 0.00006462, 0.00006233, 0.00009870, 0.00013446, 0.00011227, 0.00002114, -0.00008057, -0.00011809, -0.00007638, -0.00001300, 0.00000774, -0.00001788, -0.00004133, -0.00002833, 0.00000280, 0.00001326, -0.00000363, -0.00001931, -0.00001273, 0.00000404, 0.00000866, 0.00000042, -0.00000449, -0.00000113, -0.00000116, -0.00000949, -0.00001316, -0.00000276, 0.00000825, 0.00000125, -0.00001653, -0.00001803, 0.00000346, 0.00001931, 0.00000420, -0.00002375, -0.00002355, 0.00001020, 0.00003428, 0.00001399, -0.00002638, -0.00003361, 0.00000140, 0.00002984, 0.00001276, -0.00002326, -0.00002312, 0.00001676, 0.00003923, 0.00000539, -0.00004667, -0.00004840, 0.00000705, 0.00005275, 0.00003441, -0.00001992, -0.00003964, -0.00000510, 0.00003061, 0.00001697, -0.00002653, -0.00003962, -0.00000305, 0.00003786, 0.00003469, -0.00000380, -0.00002901, -0.00001660, 0.00000919, 0.00001483, 0.00000009, -0.00000984, -0.00000244, 0.00000919, 0.00000955, 0.00000152, -0.00000255, -0.00000032, 0.00000013, -0.00000430, -0.00000669, -0.00000205, 0.00000373, 0.00000213, -0.00000543, -0.00000930, -0.00000471, 0.00000258, 0.00000453, 0.00000044, -0.00000388, -0.00000361, 0.00000127, 0.00000755, 0.00001083, 0.00000718, -0.00000324, -0.00001348, -0.00001431, -0.00000420, 0.00000753, 0.00001027, 0.00000395, -0.00000193, -0.00000059, 0.00000420, 0.00000428, -0.00000221, -0.00000954, -0.00001242, -0.00001152, -0.00001032, -0.00000974, -0.00000836, -0.00000602, -0.00000453, -0.00000501, -0.00000714, -0.00001183, -0.00002225, -0.00003948, },
		{ -0.00001123, -0.00002401, -0.00004671, -0.00006738, -0.00006481, -0.00002458, 0.00004281, 0.00010154, 0.00011766, 0.00009367, 0.00007135, 0.00008876, 0.00013162, 0.00013893, 0.00006857, -0.00004655, -0.00012216, -0.00010783, -0.00003842, 0.00000828, -0.00000295, -0.00003633, -0.00004047, -0.00001153, 0.00001220, 0.00000617, -0.00001264, -0.00001636, -0.00000428, 0.00000386, 0.00000091, -0.00000105, 0.00000394, 0.00000380, -0.00000926, -0.00001988, -0.00000982, 0.00001049, 0.00001149, -0.00001203, -0.00002770, -0.00000935, 0.00002151, 0.00002100, -0.00001425, -0.00003662, -0.00001119, 0.00003211, 0.00003514, -0.00000861, -0.00004205, -0.00002102, 0.00002464, 0.00003200, -0.00000724, -0.00003490, -0.00000873, 0.00003565, 0.00003218, -0.00002237, -0.00005909, -0.00002754, 0.00003573, 0.00005401, 0.00000962, -0.00003669, -0.00002858, 0.00001445, 0.00002954, -0.00000298, -0.00003657, -0.00002440, 0.00001909, 0.00004055, 0.00001758, -0.00001778, -0.00002544, -0.00000550, 0.00001116, 0.00000729, -0.00000392, -0.00000414, 0.00000496, 0.00000922, 0.00000426, -0.00000115, -0.00000081, 0.00000075, -0.00000207, -0.00000616, -0.00000511, 0.00000011, 0.00000223, -0.00000161, -0.00000597, -0.00000502, -0.00000017, 0.00000276, 0.00000126, -0.00000205, -0.00000358, -0.00000178, 0.00000302, 0.00000851, 0.00000982, 0.00000306, -0.00000858, -0.00001508, -0.00000951, 0.00000325, 0.00001096, 0.00000765, -0.00000028, -0.00000292, 0.00000123, 0.00000497, 0.00000235, -0.00000484, -0.00001086, -0.00001301, -0.00001256, -0.00001112, -0.00000893, -0.00000662, -0.00000574, -0.00000664, -0.00000756, -0.00000782, },
		{ -0.00000948, -0.00000943, -0.00000716, -0.00001020, -0.00002744, -0.00005734, -0.00008057, -0.00006912, -0.00001138, 0.00006840, 0.00012294, 0.00012489, 0.00009747, 0.00009157, 0.00012476, 0.00015175, 0.00011046, -0.00000081, -0.00010741, -0.00013115, -0.00007219, -0.00000459, 0.00000820, -0.00002360, -0.00004425, -0.00002657, 0.00000311, 0.00000991, -0.00000433, -0.00001338, -0.00000766, -0.00000172, -0.00000427, -0.00000387, 0.00000730, 0.00001376, -0.00000132, -0.00002411, -0.00002303, 0.00000467, 0.00002208, 0.00000194, -0.00003048, -0.00002792, 0.00001123, 0.00003482, 0.00000729, -0.00003682, -0.00003582, 0.00001354, 0.00004719, 0.00001907, -0.00003454, -0.00004264, 0.00000408, 0.00004082, 0.00001834, -0.00002918, -0.00003271, 0.00001401, 0.00004418, 0.00001040, -0.00004750, -0.00005356, 0.00000303, 0.00005323, 0.00003856, -0.00001575, -0.00003926, -0.00000935, 0.00002558, 0.00001686, -0.00002050, -0.00003325, -0.00000303, 0.00003238, 0.00003148, 0.00000008, -0.00002292, -0.00001689, 0.00000148, 0.00000796, 0.00000149, -0.00000239, 0.00000279, 0.00000816, 0.00000593, 0.00000027, -0.00000130, 0.00000053, -0.00000013, -0.00000404, -0.00000595, -0.00000333, -0.00000013, -0.00000051, -0.00000290, -0.00000316, -0.00000079, 0.00000105, 0.00000048, -0.00000128, -0.00000251, -0.00000247, -0.00000029, 0.00000451, 0.00000893, 0.00000715, -0.00000244, -0.00001247, -0.00001282, -0.00000233, 0.00000886, 0.00001051, 0.00000345, -0.00000279, -0.00000194, 0.00000275, 0.00000420, -0.00000002, -0.00000665, -0.00001182, -0.00001419, -0.00001380, -0.00001095, -0.00000713, -0.00000526, -0.00000682, },
		{ -0.00000501, -0.00000489, -0.00000915, -0.00001257, -0.00001019, -0.00000499, -0.00000921, -0.00003393, -0.00007239, -0.00009464, -0.00006774, 0.00000982, 0.00009620, 0.00013989, 0.00013132, 0.00011281, 0.00012653, 0.00015681, 0.00014116, 0.00004690, -0.00007645, -0.00013916, -0.00010549, -0.00002958, 0.00000894, -0.00000972, -0.00003925, -0.00003603, -0.00000892, 0.00000627, -0.00000057, -0.00000860, -0.00000481, -0.00000171, -0.00000962, -0.00001384, 0.00000200, 0.00002242, 0.00001541, -0.00001821, -0.00003666, -0.00001234, 0.00002445, 0.00002199, -0.00001966, -0.00004326, -0.00001197, 0.00003530, 0.00003316, -0.00001901, -0.00005138, -0.00001728, 0.00004064, 0.00004551, -0.00000960, -0.00005131, -0.00002513, 0.00003128, 0.00004111, -0.00000617, -0.00004217, -0.00001600, 0.00003480, 0.00003646, -0.00001846, -0.00005927, -0.00003145, 0.00003147, 0.00005333, 0.00001366, -0.00003170, -0.00002843, 0.00000873, 0.00002432, -0.00000069, -0.00002817, -0.00001895, 0.00001585, 0.00003359, 0.00001629, -0.00001180, -0.00002034, -0.00000851, 0.00000307, 0.00000325, -0.00000004, 0.00000251, 0.00000767, 0.00000739, 0.00000195, -0.00000161, -0.00000061, 0.00000058, -0.00000154, -0.00000459, -0.00000494, -0.00000314, -0.00000201, -0.00000190, -0.00000102, 0.00000061, 0.00000097, -0.00000051, -0.00000193, -0.00000209, -0.00000184, -0.00000141, 0.00000109, 0.00000579, 0.00000793, 0.00000254, -0.00000770, -0.00001300, -0.00000718, 0.00000451, 0.00001090, 0.00000740, -0.00000003, -0.00000313, -0.00000045, 0.00000296, 0.00000253, -0.00000187, -0.00000784, -0.00001313, -0.00001587, -0.00001450, -0.00000961, },
		{ -0.00001748, -0.00001417, -0.00000740, -0.00000315, -0.00000571, -0.00001255, -0.00001546, -0.00000922, -0.00000118, -0.00000959, -0.00004547, -0.00009169, -0.00010620, -0.00005781, 0.00003703, 0.00012173, 0.00015351, 0.00014590, 0.00014566, 0.00016670, 0.00016414, 0.00008929, -0.00003747, -0.00013151, -0.00013017, -0.00005973, -0.00000180, -0.00000177, -0.00003074, -0.00003823, -0.00001742, -0.00000017, -0.00000328, -0.00000884, -0.00000090, 0.00000593, -0.00000706, -0.00002470, -0.00001458, 0.00002016, 0.00003448, 0.00000209, -0.00004028, -0.00003628, 0.00001117, 0.00003799, 0.00000502, -0.00004427, -0.00003996, 0.00001662, 0.00005011, 0.00001286, -0.00004695, -0.00004790, 0.00001373, 0.00005692, 0.00002468, -0.00003911, -0.00004964, 0.00000444, 0.00004821, 0.00002426, -0.00003067, -0.00003949, 0.00000823, 0.00004355, 0.00001447, -0.00004272, -0.00005279, -0.00000170, 0.00004723, 0.00003785, -0.00000989, -0.00003356, -0.00001117, 0.00001799, 0.00001320, -0.00001454, -0.00002378, -0.00000069, 0.00002574, 0.00002497, 0.00000182, -0.00001589, -0.00001414, -0.00000370, 0.00000109, 0.00000050, 0.00000249, 0.00000763, 0.00000913, 0.00000443, -0.00000104, -0.00000210, -0.00000032, -0.00000008, -0.00000218, -0.00000422, -0.00000487, -0.00000462, -0.00000336, -0.00000050, 0.00000254, 0.00000280, -0.00000012, -0.00000294, -0.00000301, -0.00000159, -0.00000103, -0.00000041, 0.00000256, 0.00000614, 0.00000484, -0.00000294, -0.00001057, -0.00000972, -0.00000038, 0.00000866, 0.00000967, 0.00000387, -0.00000161, -0.00000230, 0.00000022, 0.00000202, 0.00000096, -0.00000303, -0.00000908, -0.00001505, },
		{ -0.00000393, -0.00001083, -0.00001729, -0.00001849, -0.00001283, -0.00000476, -0.00000193, -0.00000791, -0.00001689, -0.00001738, -0.00000564, 0.00000359, -0.00001387, -0.00006347, -0.00011267, -0.00011106, -0.00003933, 0.00006535, 0.00014324, 0.00017159, 0.00017856, 0.00019220, 0.00019118, 0.00012805, 0.00000220, -0.00011455, -0.00014441, -0.00008742, -0.00001873, -0.00000180, -0.00002554, -0.00003785, -0.00002028, -0.00000242, -0.00000750, -0.00001678, -0.00000448, 0.00001539, 0.00000745, -0.00002556, -0.00003632, 0.00000067, 0.00004331, 0.00003221, -0.00002477, -0.00005514, -0.00001766, 0.00003772, 0.00003449, -0.00002479, -0.00005858, -0.00001726, 0.00004609, 0.00004551, -0.00001999, -0.00006279, -0.00002454, 0.00004450, 0.00005318, -0.00000803, -0.00005613, -0.00002881, 0.00003374, 0.00004715, -0.00000252, -0.00004425, -0.00002235, 0.00002894, 0.00003599, -0.00001273, -0.00005276, -0.00003153, 0.00002386, 0.00004618, 0.00001523, -0.00002336, -0.00002386, 0.00000360, 0.00001620, -0.00000067, -0.00001886, -0.00001090, 0.00001375, 0.00002506, 0.00001224, -0.00000720, -0.00001400, -0.00000860, -0.00000286, -0.00000127, 0.00000121, 0.00000694, 0.00001066, 0.00000772, 0.00000112, -0.00000279, -0.00000224, -0.00000047, -0.00000037, -0.00000201, -0.00000436, -0.00000636, -0.00000618, -0.00000239, 0.00000305, 0.00000533, 0.00000225, -0.00000270, -0.00000450, -0.00000267, -0.00000073, -0.00000029, 0.00000081, 0.00000363, 0.00000462, 0.00000010, -0.00000721, -0.00000978, -0.00000407, 0.00000498, 0.00000946, 0.00000694, 0.00000160, -0.00000165, -0.00000170, -0.00000027, 0.00000078, 0.00000001, },
		{ -0.00000135, -0.00000005, -0.00000049, -0.00000509, -0.00001304, -0.00001939, -0.00001872, -0.00001064, -0.00000193, -0.00000171, -0.00001184, -0.00002164, -0.00001690, 0.00000091, 0.00000702, -0.00002484, -0.00008694, -0.00013064, -0.00010699, -0.00001647, 0.00009046, 0.00016595, 0.00020657, 0.00023211, 0.00023443, 0.00017477, 0.00004402, -0.00009494, -0.00015389, -0.00011113, -0.00003497, -0.00000468, -0.00002558, -0.00004220, -0.00002288, 0.00000224, -0.00000431, -0.00002667, -0.00002040, 0.00001510, 0.00002842, -0.00000828, -0.00004928, -0.00003233, 0.00003012, 0.00005804, 0.00001017, -0.00005445, -0.00005132, 0.00001469, 0.00005323, 0.00001073, -0.00005496, -0.00005232, 0.00001799, 0.00006212, 0.00002019, -0.00005180, -0.00005736, 0.00001048, 0.00006119, 0.00002949, -0.00003902, -0.00005278, 0.00000275, 0.00005046, 0.00002899, -0.00002750, -0.00004172, 0.00000086, 0.00003708, 0.00001590, -0.00003382, -0.00004587, -0.00000527, 0.00003643, 0.00003202, -0.00000431, -0.00002415, -0.00001000, 0.00000988, 0.00000710, -0.00000981, -0.00001320, 0.00000352, 0.00001961, 0.00001692, 0.00000124, -0.00000985, -0.00000989, -0.00000596, -0.00000403, -0.00000156, 0.00000465, 0.00001087, 0.00001097, 0.00000486, -0.00000161, -0.00000385, -0.00000240, -0.00000034, 0.00000005, -0.00000208, -0.00000595, -0.00000842, -0.00000591, 0.00000104, 0.00000665, 0.00000569, -0.00000031, -0.00000496, -0.00000461, -0.00000170, 0.00000004, 0.00000064, 0.00000206, 0.00000330, 0.00000106, -0.00000461, -0.00000849, -0.00000590, 0.00000148, 0.00000744, 0.00000803, 0.00000466, 0.00000089, -0.00000137, -0.00000205, },
		{ 0.00000089, -0.00000173, -0.00000315, -0.00000230, -0.00000039, -0.00000097, -0.00000663, -0.00001543, -0.00002118, -0.00001817, -0.00000761, 0.00000080, -0.00000321, -0.00001756, -0.00002534, -0.00001260, 0.00000904, 0.00000551, -0.00004347, -0.00011176, -0.00014146, -0.00009645, 0.00000466, 0.00011407, 0.00020297, 0.00026801, 0.00029448, 0.00024345, 0.00010107, -0.00007044, -0.00016582, -0.00013814, -0.00004994, -0.00000259, -0.00002420, -0.00005360, -0.00003491, 0.00000727, 0.00001138, -0.00002619, -0.00004326, -0.00000431, 0.00004113, 0.00002460, -0.00003932, -0.00006361, -0.00000707, 0.00006200, 0.00005217, -0.00002647, -0.00007178, -0.00002557, 0.00004740, 0.00004700, -0.00002587, -0.00007063, -0.00002457, 0.00005139, 0.00005556, -0.00001680, -0.00006828, -0.00003138, 0.00004269, 0.00005615, -0.00000449, -0.00005526, -0.00003069, 0.00003183, 0.00004861, 0.00000250, -0.00004069, -0.00002607, 0.00001942, 0.00003032, -0.00000716, -0.00004113, -0.00002720, 0.00001521, 0.00003421, 0.00001353, -0.00001406, -0.00001612, 0.00000028, 0.00000732, -0.00000245, -0.00001035, -0.00000200, 0.00001308, 0.00001688, 0.00000673, -0.00000477, -0.00000832, -0.00000679, -0.00000591, -0.00000467, 0.00000084, 0.00000888, 0.00001278, 0.00000916, 0.00000168, -0.00000375, -0.00000457, -0.00000215, 0.00000050, 0.00000043, -0.00000351, -0.00000855, -0.00000921, -0.00000304, 0.00000533, 0.00000831, 0.00000365, -0.00000320, -0.00000588, -0.00000377, -0.00000072, 0.00000089, 0.00000184, 0.00000246, 0.00000094, -0.00000340, -0.00000728, -0.00000647, -0.00000095, 0.00000493, 0.00000734, 0.00000623, 0.00000367, },
		{ 0.00000621, 0.00000529, 0.00000357, 0.00000059, -0.00000284, -0.00000439, -0.00000288, -0.00000053, -0.00000158, -0.00000837, -0.00001793, -0.00002266, -0.00001662, -0.00000370, 0.00000261, -0.00000723, -0.00002409, -0.00002581, -0.00000447, 0.00001524, -0.00000378, -0.00006711, -0.00013251, -0.00014523, -0.00008627, 0.00002174, 0.00014778, 0.00026973, 0.00035241, 0.00033708, 0.00019144, -0.00002465, -0.00017802, -0.00017914, -0.00007424, 0.00000565, -0.00000983, -0.00006347, -0.00006160, -0.00000129, 0.00003334, -0.00000502, -0.00005723, -0.00004086, 0.00002960, 0.00005722, -0.00000245, -0.00007276, -0.00005539, 0.00003363, 0.00007990, 0.00002246, -0.00006315, -0.00006541, 0.00001405, 0.00006474, 0.00001827, -0.00006015, -0.00006253, 0.00001486, 0.00006817, 0.00002765, -0.00005024, -0.00006193, 0.00000460, 0.00005876, 0.00003192, -0.00003445, -0.00005085, 0.00000008, 0.00004722, 0.00003107, -0.00002076, -0.00003881, -0.00000618, 0.00002607, 0.00001380, -0.00002322, -0.00003421, -0.00000642, 0.00002355, 0.00002229, -0.00000060, -0.00001351, -0.00000639, 0.00000311, 0.00000042, -0.00000662, -0.00000351, 0.00000841, 0.00001492, 0.00000921, -0.00000095, -0.00000591, -0.00000579, -0.00000590, -0.00000657, -0.00000326, 0.00000489, 0.00001212, 0.00001247, 0.00000623, -0.00000134, -0.00000542, -0.00000470, -0.00000105, 0.00000156, -0.00000040, -0.00000633, -0.00001053, -0.00000748, 0.00000144, 0.00000840, 0.00000741, 0.00000058, -0.00000509, -0.00000561, -0.00000272, 0.00000025, 0.00000208, 0.00000266, 0.00000109, -0.00000292, -0.00000680, -0.00000687, -0.00000254, 0.00000286, 0.00000590, },
		{ 0.00000118, 0.00000452, 0.00000541, 0.00000548, 0.00000524, 0.00000345, -0.00000037, -0.00000427, -0.00000543, -0.00000325, -0.00000057, -0.00000215, -0.00001033, -0.00002071, -0.00002356, -0.00001362, 0.00000055, 0.00000217, -0.00001367, -0.00002910, -0.00002158, 0.00000486, 0.00001532, -0.00002023, -0.00008999, -0.00014666, -0.00014824, -0.00008278, 0.00004262, 0.00021089, 0.00037300, 0.00043632, 0.00032322, 0.00006781, -0.00017066, -0.00023556, -0.00012546, 0.00000679, 0.00002320, -0.00005432, -0.00009471, -0.00003515, 0.00004335, 0.00003462, -0.00004442, -0.00007757, -0.00001164, 0.00006683, 0.00004852, -0.00004631, -0.00009065, -0.00002127, 0.00007385, 0.00007168, -0.00002217, -0.00008292, -0.00003460, 0.00005134, 0.00005699, -0.00002254, -0.00007677, -0.00003230, 0.00004998, 0.00006089, -0.00001030, -0.00006646, -0.00003594, 0.00003562, 0.00005321, -0.00000031, -0.00004873, -0.00002958, 0.00002658, 0.00004503, 0.00000728, -0.00003253, -0.00002614, 0.00000850, 0.00002049, -0.00000336, -0.00002710, -0.00001926, 0.00000766, 0.00002020, 0.00000890, -0.00000605, -0.00000736, -0.00000089, -0.00000021, -0.00000482, -0.00000362, 0.00000589, 0.00001334, 0.00001035, 0.00000132, -0.00000429, -0.00000451, -0.00000448, -0.00000642, -0.00000607, 0.00000017, 0.00000899, 0.00001347, 0.00001045, 0.00000282, -0.00000405, -0.00000639, -0.00000378, 0.00000061, 0.00000170, -0.00000287, -0.00000925, -0.00001034, -0.00000356, 0.00000556, 0.00000904, 0.00000484, -0.00000203, -0.00000571, -0.00000484, -0.00000166, 0.00000152, 0.00000322, 0.00000209, -0.00000207, -0.00000661, -0.00000765, -0.00000406, },
		{ -0.00000840, -0.00000592, -0.00000070, 0.00000323, 0.00000456, 0.00000492, 0.00000554, 0.00000535, 0.00000280, -0.00000169, -0.00000565, -0.00000641, -0.00000348, -0.00000028, -0.00000271, -0.00001292, -0.00002368, -0.00002311, -0.00000912, 0.00000368, -0.00000140, -0.00002064, -0.00002994, -0.00001379, 0.00001062, 0.00000759, -0.00003869, -0.00010700, -0.00015860, -0.00016013, -0.00008391, 0.00008733, 0.00032069, 0.00049953, 0.00047911, 0.00022492, -0.00010876, -0.00028725, -0.00021414, -0.00002533, 0.00006301, -0.00001131, -0.00011133, -0.00009131, 0.00002046, 0.00007330, 0.00000177, -0.00008895, -0.00006936, 0.00003730, 0.00008704, 0.00001122, -0.00008942, -0.00007978, 0.00002883, 0.00009516, 0.00003676, -0.00006412, -0.00007537, 0.00000945, 0.00007005, 0.00002596, -0.00005869, -0.00006819, 0.00000804, 0.00006674, 0.00003320, -0.00004270, -0.00006020, -0.00000223, 0.00004991, 0.00003072, -0.00002675, -0.00004379, -0.00000217, 0.00003952, 0.00002969, -0.00001239, -0.00003149, -0.00001119, 0.00001299, 0.00000850, -0.00001337, -0.00002044, -0.00000495, 0.00001140, 0.00001095, 0.00000060, -0.00000419, -0.00000175, -0.00000130, -0.00000493, -0.00000451, 0.00000399, 0.00001257, 0.00001175, 0.00000318, -0.00000361, -0.00000420, -0.00000306, -0.00000471, -0.00000669, -0.00000364, 0.00000451, 0.00001186, 0.00001287, 0.00000731, -0.00000068, -0.00000613, -0.00000616, -0.00000189, 0.00000177, 0.00000011, -0.00000616, -0.00001059, -0.00000765, 0.00000090, 0.00000772, 0.00000760, 0.00000209, -0.00000350, -0.00000557, -0.00000396, -0.00000033, 0.00000299, 0.00000338, -0.00000032, -0.00000577, },
		{ 0.00000178, -0.00000382, -0.00000824, -0.00000767, -0.00000308, 0.00000148, 0.00000362, 0.00000429, 0.00000510, 0.00000583, 0.00000501, 0.00000182, -0.00000301, -0.00000714, -0.00000746, -0.00000325, 0.00000051, -0.00000376, -0.00001642, -0.00002603, -0.00002063, -0.00000430, 0.00000395, -0.00000722, -0.00002489, -0.00002598, -0.00000693, 0.00000893, -0.00000438, -0.00005192, -0.00011922, -0.00017972, -0.00018595, -0.00007298, 0.00018175, 0.00047829, 0.00061185, 0.00043724, 0.00004103, -0.00029061, -0.00032657, -0.00011551, 0.00007713, 0.00006183, -0.00008452, -0.00014576, -0.00004335, 0.00008118, 0.00006661, -0.00005607, -0.00011408, -0.00002784, 0.00008596, 0.00007524, -0.00004372, -0.00011150, -0.00003961, 0.00007567, 0.00008638, -0.00001263, -0.00008582, -0.00004243, 0.00004888, 0.00006214, -0.00001572, -0.00007553, -0.00003841, 0.00004223, 0.00005991, -0.00000233, -0.00005759, -0.00003668, 0.00002499, 0.00004447, 0.00000294, -0.00003795, -0.00002503, 0.00001976, 0.00003725, 0.00001036, -0.00002189, -0.00002265, -0.00000140, 0.00000871, -0.00000207, -0.00001371, -0.00000967, 0.00000251, 0.00000713, 0.00000274, -0.00000078, 0.00000013, -0.00000049, -0.00000503, -0.00000636, 0.00000118, 0.00001151, 0.00001365, 0.00000588, -0.00000281, -0.00000496, -0.00000286, -0.00000282, -0.00000543, -0.00000546, 0.00000028, 0.00000845, 0.00001291, 0.00001065, 0.00000352, -0.00000383, -0.00000697, -0.00000459, 0.00000004, 0.00000137, -0.00000286, -0.00000863, -0.00000949, -0.00000365, 0.00000411, 0.00000768, 0.00000537, 0.00000015, -0.00000410, -0.00000518, -0.00000281, 0.00000134, 0.00000384, },
		{ -0.00000115, 0.00000277, 0.00000310, -0.00000130, -0.00000671, -0.00000835, -0.00000538, -0.00000094, 0.00000205, 0.00000351, 0.00000462, 0.00000556, 0.00000568, 0.00000434, 0.00000093, -0.00000447, -0.00000900, -0.00000820, -0.00000209, 0.00000134, -0.00000604, -0.00002029, -0.00002653, -0.00001658, -0.00000135, 0.00000099, -0.00001207, -0.00002402, -0.00002057, -0.00000670, 0.00000079, -0.00001217, -0.00005657, -0.00013732, -0.00022221, -0.00021667, -0.00001985, 0.00034333, 0.00065758, 0.00065673, 0.00028555, -0.00019362, -0.00041640, -0.00026532, 0.00002466, 0.00013327, -0.00000276, -0.00016080, -0.00013063, 0.00003520, 0.00011595, 0.00001804, -0.00011438, -0.00010184, 0.00003451, 0.00011119, 0.00003068, -0.00009495, -0.00010021, 0.00001635, 0.00010006, 0.00004965, -0.00005708, -0.00007855, 0.00000242, 0.00006807, 0.00003179, -0.00005104, -0.00006771, -0.00000059, 0.00005798, 0.00003509, -0.00003106, -0.00005216, -0.00000809, 0.00003640, 0.00002549, -0.00001796, -0.00003280, -0.00000285, 0.00002942, 0.00002506, -0.00000451, -0.00002165, -0.00001330, 0.00000066, 0.00000144, -0.00000604, -0.00000753, -0.00000172, 0.00000208, 0.00000059, -0.00000031, 0.00000206, 0.00000220, -0.00000329, -0.00000767, -0.00000263, 0.00000889, 0.00001499, 0.00000957, -0.00000068, -0.00000572, -0.00000409, -0.00000206, -0.00000353, -0.00000530, -0.00000251, 0.00000462, 0.00001104, 0.00001209, 0.00000720, -0.00000038, -0.00000590, -0.00000622, -0.00000239, 0.00000079, -0.00000074, -0.00000580, -0.00000900, -0.00000654, -0.00000012, 0.00000529, 0.00000640, 0.00000352, -0.00000096, -0.00000434, -0.00000449, },
		{ -0.00000170, -0.00000430, -0.00000317, 0.00000062, 0.00000288, 0.00000061, -0.00000438, -0.00000750, -0.00000657, -0.00000337, -0.00000038, 0.00000192, 0.00000399, 0.00000536, 0.00000555, 0.00000509, 0.00000387, 0.00000005, -0.00000653, -0.00001100, -0.00000790, -0.00000018, 0.00000118, -0.00000952, -0.00002304, -0.00002476, -0.00001315, -0.00000193, -0.00000269, -0.00001200, -0.00001940, -0.00002046, -0.00001546, -0.00000580, -0.00000760, -0.00005949, -0.00017873, -0.00028804, -0.00022436, 0.00010764, 0.00056467, 0.00080530, 0.00058559, 0.00003543, -0.00041487, -0.00043626, -0.00011877, 0.00014989, 0.00011296, -0.00010556, -0.00019938, -0.00006210, 0.00011213, 0.00010312, -0.00005651, -0.00014527, -0.00005115, 0.00009314, 0.00009884, -0.00003162, -0.00012115, -0.00005793, 0.00006684, 0.00009248, -0.00000037, -0.00007963, -0.00004665, 0.00004101, 0.00006116, -0.00000727, -0.00006709, -0.00004107, 0.00003006, 0.00005244, 0.00000489, -0.00004355, -0.00003302, 0.00001327, 0.00003141, 0.00000407, -0.00002529, -0.00001761, 0.00001330, 0.00002711, 0.00001104, -0.00001128, -0.00001673, -0.00000849, -0.00000236, -0.00000291, -0.00000332, -0.00000080, -0.00000007, -0.00000267, -0.00000309, 0.00000146, 0.00000470, 0.00000036, -0.00000683, -0.00000608, 0.00000458, 0.00001445, 0.00001321, 0.00000310, -0.00000517, -0.00000583, -0.00000287, -0.00000232, -0.00000410, -0.00000357, 0.00000156, 0.00000832, 0.00001179, 0.00000954, 0.00000305, -0.00000355, -0.00000627, -0.00000427, -0.00000080, -0.00000023, -0.00000352, -0.00000722, -0.00000731, -0.00000334, 0.00000179, 0.00000501, 0.00000510, 0.00000235, },
		{ 0.00000443, 0.00000431, 0.00000147, -0.00000226, -0.00000358, -0.00000136, 0.00000145, 0.00000105, -0.00000253, -0.00000572, -0.00000608, -0.00000459, -0.00000293, -0.00000081, 0.00000237, 0.00000513, 0.00000566, 0.00000487, 0.00000468, 0.00000371, -0.00000142, -0.00000925, -0.00001224, -0.00000631, 0.00000134, -0.00000062, -0.00001281, -0.00002332, -0.00002238, -0.00001314, -0.00000495, -0.00000221, -0.00000581, -0.00001765, -0.00003176, -0.00002782, 0.00000131, 0.00000924, -0.00007939, -0.00025751, -0.00035874, -0.00016790, 0.00032501, 0.00080436, 0.00085618, 0.00037902, -0.00026055, -0.00055163, -0.00033530, 0.00006204, 0.00020722, 0.00001977, -0.00020114, -0.00017482, 0.00003716, 0.00015271, 0.00004234, -0.00012698, -0.00013181, 0.00002064, 0.00012353, 0.00005141, -0.00008788, -0.00011165, -0.00000075, 0.00009317, 0.00005756, -0.00004391, -0.00007374, -0.00000473, 0.00005945, 0.00003414, -0.00003907, -0.00006086, -0.00000871, 0.00004373, 0.00003244, -0.00001808, -0.00003934, -0.00001146, 0.00002099, 0.00001699, -0.00001014, -0.00002008, -0.00000154, 0.00001928, 0.00001835, 0.00000128, -0.00001168, -0.00001267, -0.00000864, -0.00000541, -0.00000183, 0.00000207, 0.00000175, -0.00000357, -0.00000673, -0.00000219, 0.00000469, 0.00000417, -0.00000359, -0.00000750, -0.00000022, 0.00001160, 0.00001535, 0.00000771, -0.00000268, -0.00000676, -0.00000466, -0.00000244, -0.00000303, -0.00000349, -0.00000031, 0.00000577, 0.00001056, 0.00001055, 0.00000568, -0.00000091, -0.00000513, -0.00000501, -0.00000229, -0.00000079, -0.00000242, -0.00000542, -0.00000662, -0.00000483, -0.00000129, 0.00000220, },
		{ -0.00000307, -0.00000064, 0.00000218, 0.00000426, 0.00000371, 0.00000056, -0.00000223, -0.00000201, 0.00000000, 0.00000030, -0.00000206, -0.00000440, -0.00000458, -0.00000390, -0.00000412, -0.00000387, -0.00000080, 0.00000384, 0.00000609, 0.00000505, 0.00000413, 0.00000484, 0.00000323, -0.00000390, -0.00001175, -0.00001195, -0.00000435, 0.00000119, -0.00000312, -0.00001376, -0.00002175, -0.00002291, -0.00001740, -0.00000547, 0.00000662, 0.00000176, -0.00002806, -0.00005388, -0.00003009, 0.00003043, 0.00002372, -0.00014012, -0.00037061, -0.00039352, -0.00000992, 0.00061383, 0.00099291, 0.00075905, 0.00006320, -0.00052272, -0.00055493, -0.00014346, 0.00021028, 0.00017120, -0.00010965, -0.00024521, -0.00009015, 0.00012833, 0.00013701, -0.00004438, -0.00016381, -0.00007681, 0.00008599, 0.00011328, -0.00001310, -0.00011713, -0.00007143, 0.00004953, 0.00008803, 0.00001116, -0.00006573, -0.00004559, 0.00003008, 0.00005422, 0.00000066, -0.00005323, -0.00003931, 0.00001626, 0.00003987, 0.00000953, -0.00002734, -0.00002557, 0.00000291, 0.00001659, 0.00000271, -0.00001335, -0.00000879, 0.00000857, 0.00001693, 0.00000956, -0.00000280, -0.00001017, -0.00001209, -0.00001049, -0.00000477, 0.00000289, 0.00000531, -0.00000083, -0.00000818, -0.00000699, 0.00000140, 0.00000585, 0.00000065, -0.00000625, -0.00000380, 0.00000725, 0.00001515, 0.00001173, 0.00000133, -0.00000600, -0.00000629, -0.00000366, -0.00000281, -0.00000319, -0.00000131, 0.00000379, 0.00000913, 0.00001080, 0.00000748, 0.00000136, -0.00000360, -0.00000482, -0.00000308, -0.00000153, -0.00000225, -0.00000436, -0.00000556, -0.00000491, },
		{ -0.00000484, -0.00000440, -0.00000358, -0.00000261, -0.00000050, 0.00000260, 0.00000430, 0.00000284, -0.00000003, -0.00000114, -0.00000029, -0.00000041, -0.00000273, -0.00000460, -0.00000366, -0.00000196, -0.00000294, -0.00000551, -0.00000480, 0.00000051, 0.00000558, 0.00000602, 0.00000399, 0.00000406, 0.00000511, 0.00000165, -0.00000676, -0.00001278, -0.00001061, -0.00000386, -0.00000054, -0.00000371, -0.00001152, -0.00002190, -0.00002947, -0.00002227, 0.00000333, 0.00002268, -0.00000007, -0.00005637, -0.00007543, -0.00000625, 0.00007632, 0.00000797, -0.00025650, -0.00048826, -0.00034040, 0.00026097, 0.00091623, 0.00105478, 0.00050286, -0.00029163, -0.00067304, -0.00042029, 0.00007425, 0.00027144, 0.00005552, -0.00022222, -0.00021558, 0.00002426, 0.00017499, 0.00006976, -0.00012330, -0.00015217, -0.00000131, 0.00012052, 0.00006776, -0.00007013, -0.00011096, -0.00001793, 0.00007595, 0.00005795, -0.00002802, -0.00006168, -0.00000973, 0.00004641, 0.00003237, -0.00002547, -0.00004892, -0.00001437, 0.00002700, 0.00002566, -0.00000654, -0.00002440, -0.00001177, 0.00000669, 0.00000696, -0.00000472, -0.00000817, 0.00000127, 0.00001108, 0.00001132, 0.00000436, -0.00000365, -0.00001030, -0.00001376, -0.00001023, -0.00000030, 0.00000715, 0.00000395, -0.00000591, -0.00001011, -0.00000376, 0.00000429, 0.00000375, -0.00000318, -0.00000511, 0.00000292, 0.00001289, 0.00001407, 0.00000574, -0.00000354, -0.00000682, -0.00000518, -0.00000346, -0.00000325, -0.00000204, 0.00000220, 0.00000779, 0.00001081, 0.00000886, 0.00000324, -0.00000222, -0.00000434, -0.00000327, -0.00000186, -0.00000230, -0.00000396, },
		{ -0.00000211, -0.00000373, -0.00000454, -0.00000392, -0.00000337, -0.00000354, -0.00000276, 0.00000021, 0.00000342, 0.00000389, 0.00000185, 0.00000043, 0.00000074, 0.00000012, -0.00000316, -0.00000607, -0.00000467, -0.00000065, 0.00000005, -0.00000422, -0.00000757, -0.00000439, 0.00000270, 0.00000649, 0.00000510, 0.00000343, 0.00000446, 0.00000438, -0.00000085, -0.00000841, -0.00001210, -0.00001031, -0.00000590, -0.00000130, 0.00000019, -0.00000868, -0.00002892, -0.00004111, -0.00001959, 0.00002522, 0.00003656, -0.00002318, -0.00009761, -0.00007622, 0.00005106, 0.00011516, -0.00006801, -0.00041871, -0.00055626, -0.00015862, 0.00061362, 0.00114755, 0.00093906, 0.00012950, -0.00059274, -0.00066631, -0.00019191, 0.00024361, 0.00022637, -0.00009342, -0.00027264, -0.00012206, 0.00012524, 0.00016019, -0.00002242, -0.00016499, -0.00009841, 0.00006600, 0.00011425, 0.00000663, -0.00010041, -0.00007635, 0.00002809, 0.00007365, 0.00001870, -0.00004733, -0.00003906, 0.00001893, 0.00004297, 0.00000638, -0.00003679, -0.00003337, 0.00000375, 0.00002478, 0.00001069, -0.00001225, -0.00001609, -0.00000437, 0.00000290, -0.00000051, -0.00000418, -0.00000055, 0.00000603, 0.00000873, 0.00000690, 0.00000244, -0.00000468, -0.00001258, -0.00001452, -0.00000634, 0.00000509, 0.00000767, -0.00000093, -0.00000971, -0.00000823, 0.00000024, 0.00000423, -0.00000009, -0.00000438, -0.00000016, 0.00000961, 0.00001448, 0.00000947, -0.00000004, -0.00000595, -0.00000620, -0.00000448, -0.00000376, -0.00000288, 0.00000061, 0.00000633, 0.00001067, 0.00001023, 0.00000513, -0.00000093, -0.00000398, -0.00000337, -0.00000186, },
		{ -0.00000373, -0.00000196, -0.00000174, -0.00000335, -0.00000435, -0.00000360, -0.00000287, -0.00000370, -0.00000435, -0.00000226, 0.00000171, 0.00000384, 0.00000286, 0.00000161, 0.00000219, 0.00000206, -0.00000195, -0.00000724, -0.00000751, -0.00000181, 0.00000268, -0.00000024, -0.00000702, -0.00000864, -0.00000256, 0.00000451, 0.00000631, 0.00000441, 0.00000363, 0.00000406, 0.00000222, -0.00000255, -0.00000765, -0.00001165, -0.00001343, -0.00000876, 0.00000304, 0.00000811, -0.00001146, -0.00004568, -0.00005054, -0.00000062, 0.00005551, 0.00003267, -0.00007199, -0.00013324, -0.00003674, 0.00013008, 0.00010957, -0.00021832, -0.00058470, -0.00051315, 0.00015700, 0.00097398, 0.00122489, 0.00064504, -0.00027486, -0.00075868, -0.00050933, 0.00005398, 0.00031139, 0.00009824, -0.00021786, -0.00024317, -0.00000080, 0.00017693, 0.00009318, -0.00010427, -0.00015750, -0.00002566, 0.00010244, 0.00007483, -0.00004651, -0.00009804, -0.00003067, 0.00005238, 0.00005028, -0.00001332, -0.00004495, -0.00001119, 0.00003200, 0.00002707, -0.00001288, -0.00003427, -0.00001650, 0.00001111, 0.00001631, 0.00000153, -0.00001043, -0.00000951, -0.00000411, -0.00000237, -0.00000209, 0.00000078, 0.00000445, 0.00000579, 0.00000565, 0.00000511, 0.00000127, -0.00000749, -0.00001486, -0.00001206, -0.00000039, 0.00000791, 0.00000407, -0.00000608, -0.00000995, -0.00000424, 0.00000220, 0.00000151, -0.00000279, -0.00000169, 0.00000640, 0.00001344, 0.00001197, 0.00000366, -0.00000394, -0.00000635, -0.00000535, -0.00000443, -0.00000382, -0.00000114, 0.00000444, 0.00001007, 0.00001152, 0.00000736, 0.00000070, -0.00000365, },
		{ 0.00000302, -0.00000286, -0.00000420, -0.00000242, -0.00000152, -0.00000288, -0.00000411, -0.00000331, -0.00000218, -0.00000324, -0.00000532, -0.00000462, -0.00000047, 0.00000317, 0.00000324, 0.00000196, 0.00000287, 0.00000429, 0.00000116, -0.00000627, -0.00001036, -0.00000570, 0.00000251, 0.00000413, -0.00000279, -0.00000951, -0.00000805, -0.00000056, 0.00000522, 0.00000610, 0.00000472, 0.00000332, 0.00000168, 0.00000011, -0.00000122, -0.00000561, -0.00001494, -0.00002004, -0.00000757, 0.00001502, 0.00001453, -0.00002666, -0.00006843, -0.00004583, 0.00003721, 0.00007926, -0.00000307, -0.00013599, -0.00013638, 0.00004829, 0.00019708, 0.00002460, -0.00042506, -0.00068774, -0.00031713, 0.00056194, 0.00124485, 0.00110064, 0.00022976, -0.00060826, -0.00075047, -0.00025871, 0.00023955, 0.00026566, -0.00005960, -0.00027404, -0.00014954, 0.00010344, 0.00016613, 0.00000332, -0.00014832, -0.00010997, 0.00003818, 0.00010079, 0.00002198, -0.00007514, -0.00007120, 0.00000770, 0.00005244, 0.00002034, -0.00002857, -0.00002844, 0.00001002, 0.00003002, 0.00000913, -0.00002086, -0.00002461, -0.00000529, 0.00001022, 0.00000867, -0.00000091, -0.00000688, -0.00000800, -0.00000723, -0.00000425, 0.00000121, 0.00000535, 0.00000518, 0.00000360, 0.00000433, 0.00000456, -0.00000130, -0.00001107, -0.00001461, -0.00000672, 0.00000440, 0.00000656, -0.00000131, -0.00000856, -0.00000712, -0.00000095, 0.00000122, -0.00000160, -0.00000224, 0.00000378, 0.00001163, 0.00001326, 0.00000702, -0.00000125, -0.00000564, -0.00000581, -0.00000497, -0.00000461, -0.00000289, 0.00000206, 0.00000863, 0.00001227, 0.00000982, },
		{ 0.00001198, 0.00001196, 0.00000594, -0.00000121, -0.00000431, -0.00000311, -0.00000168, -0.00000260, -0.00000400, -0.00000313, -0.00000127, -0.00000206, -0.00000548, -0.00000683, -0.00000319, 0.00000198, 0.00000353, 0.00000200, 0.00000236, 0.00000529, 0.00000487, -0.00000256, -0.00001085, -0.00001046, -0.00000137, 0.00000581, 0.00000294, -0.00000589, -0.00001045, -0.00000681, 0.00000041, 0.00000567, 0.00000722, 0.00000508, 0.00000050, -0.00000198, 0.00000117, 0.00000360, -0.00000629, -0.00002453, -0.00002595, 0.00000334, 0.00003197, 0.00001004, -0.00005624, -0.00008545, -0.00001661, 0.00008472, 0.00007591, -0.00007280, -0.00018897, -0.00008493, 0.00015964, 0.00020607, -0.00015151, -0.00063410, -0.00065939, 0.00002897, 0.00096787, 0.00133989, 0.00078550, -0.00020811, -0.00078982, -0.00058600, 0.00000207, 0.00031550, 0.00013733, -0.00018813, -0.00024937, -0.00003112, 0.00015746, 0.00010568, -0.00007500, -0.00014591, -0.00004589, 0.00007342, 0.00007013, -0.00002311, -0.00007592, -0.00003596, 0.00002776, 0.00003627, -0.00000292, -0.00002719, -0.00000904, 0.00001919, 0.00001982, -0.00000319, -0.00001973, -0.00001518, -0.00000117, 0.00000662, 0.00000533, -0.00000003, -0.00000602, -0.00001026, -0.00000901, -0.00000166, 0.00000566, 0.00000667, 0.00000329, 0.00000233, 0.00000455, 0.00000310, -0.00000525, -0.00001309, -0.00001107, -0.00000102, 0.00000564, 0.00000224, -0.00000536, -0.00000761, -0.00000358, -0.00000021, -0.00000135, -0.00000258, 0.00000162, 0.00000946, 0.00001355, 0.00000981, 0.00000177, -0.00000424, -0.00000582, -0.00000528, -0.00000506, -0.00000429, -0.00000050, 0.00000628, },
		{ -0.00000277, 0.00000334, 0.00001046, 0.00001317, 0.00000894, 0.00000128, -0.00000365, -0.00000359, -0.00000199, -0.00000258, -0.00000425, -0.00000343, -0.00000049, -0.00000011, -0.00000434, -0.00000841, -0.00000655, -0.00000022, 0.00000370, 0.00000258, 0.00000143, 0.00000443, 0.00000719, 0.00000255, -0.00000778, -0.00001315, -0.00000731, 0.00000308, 0.00000653, 0.00000041, -0.00000776, -0.00001042, -0.00000658, 0.00000077, 0.00000780, 0.00000945, 0.00000284, -0.00000550, -0.00000352, 0.00000794, 0.00000859, -0.00001478, -0.00003868, -0.00002345, 0.00002593, 0.00004462, -0.00001360, -0.00009285, -0.00008085, 0.00003772, 0.00012096, 0.00002990, -0.00015976, -0.00019803, 0.00002314, 0.00025411, 0.00011900, -0.00039210, -0.00076842, -0.00045969, 0.00046785, 0.00126921, 0.00121767, 0.00034958, -0.00056223, -0.00078865, -0.00033111, 0.00019569, 0.00027722, -0.00001674, -0.00024751, -0.00016403, 0.00006862, 0.00015224, 0.00002545, -0.00011781, -0.00010802, 0.00000944, 0.00007567, 0.00002885, -0.00004744, -0.00005738, -0.00000734, 0.00002913, 0.00001612, -0.00001319, -0.00001629, 0.00000467, 0.00001816, 0.00000924, -0.00000791, -0.00001503, -0.00000991, -0.00000114, 0.00000470, 0.00000530, -0.00000007, -0.00000854, -0.00001254, -0.00000703, 0.00000307, 0.00000790, 0.00000500, 0.00000150, 0.00000279, 0.00000453, -0.00000009, -0.00000879, -0.00001200, -0.00000584, 0.00000228, 0.00000328, -0.00000226, -0.00000633, -0.00000478, -0.00000165, -0.00000171, -0.00000308, -0.00000037, 0.00000698, 0.00001295, 0.00001192, 0.00000491, -0.00000222, -0.00000539, -0.00000543, -0.00000520, -0.00000512, },
		{ -0.00000522, -0.00000543, -0.00000442, 0.00000037, 0.00000795, 0.00001313, 0.00001140, 0.00000417, -0.00000222, -0.00000354, -0.00000205, -0.00000242, -0.00000473, -0.00000458, -0.00000059, 0.00000208, -0.00000158, -0.00000837, -0.00000997, -0.00000394, 0.00000286, 0.00000379, 0.00000140, 0.00000258, 0.00000700, 0.00000663, -0.00000222, -0.00001181, -0.00001195, -0.00000290, 0.00000536, 0.00000534, -0.00000125, -0.00000836, -0.00001146, -0.00000771, 0.00000286, 0.00001263, 0.00001002, -0.00000449, -0.00001203, 0.00000173, 0.00001935, 0.00000711, -0.00003326, -0.00004992, -0.00000543, 0.00005453, 0.00004020, -0.00005681, -0.00011986, -0.00004319, 0.00010289, 0.00012092, -0.00005777, -0.00023048, -0.00013987, 0.00016091, 0.00027899, -0.00007184, -0.00063240, -0.00075619, -0.00010061, 0.00090037, 0.00138056, 0.00090037, -0.00010061, -0.00075619, -0.00063240, -0.00007184, 0.00027899, 0.00016091, -0.00013987, -0.00023048, -0.00005777, 0.00012092, 0.00010289, -0.00004319, -0.00011986, -0.00005681, 0.00004020, 0.00005453, -0.00000543, -0.00004992, -0.00003326, 0.00000711, 0.00001935, 0.00000173, -0.00001203, -0.00000449, 0.00001002, 0.00001263, 0.00000286, -0.00000771, -0.00001146, -0.00000836, -0.00000125, 0.00000534, 0.00000536, -0.00000290, -0.00001195, -0.00001181, -0.00000222, 0.00000663, 0.00000700, 0.00000258, 0.00000140, 0.00000379, 0.00000286, -0.00000394, -0.00000997, -0.00000837, -0.00000158, 0.00000208, -0.00000059, -0.00000458, -0.00000473, -0.00000242, -0.00000205, -0.00000354, -0.00000222, 0.00000417, 0.00001140, 0.00001313, 0.00000795, 0.00000037, -0.00000442, -0.00000543, },
		{ -0.00000277, -0.00000512, -0.00000520, -0.00000543, -0.00000539, -0.00000222, 0.00000491, 0.00001192, 0.00001295, 0.00000698, -0.00000037, -0.00000308, -0.00000171, -0.00000165, -0.00000478, -0.00000633, -0.00000226, 0.00000328, 0.00000228, -0.00000584, -0.00001200, -0.00000879, -0.00000009, 0.00000453, 0.00000279, 0.00000150, 0.00000500, 0.00000790, 0.00000307, -0.00000703, -0.00001254, -0.00000854, -0.00000007, 0.00000530, 0.00000470, -0.00000114, -0.00000991, -0.00001503, -0.00000791, 0.00000924, 0.00001816, 0.00000467, -0.00001629, -0.00001319, 0.00001612, 0.00002913, -0.00000734, -0.00005738, -0.00004744, 0.00002885, 0.00007567, 0.00000944, -0.00010802, -0.00011781, 0.00002545, 0.00015224, 0.00006862, -0.00016403, -0.00024751, -0.00001674, 0.00027722, 0.00019569, -0.00033111, -0.00078865, -0.00056223, 0.00034958, 0.00121767, 0.00126921, 0.00046785, -0.00045969, -0.00076842, -0.00039210, 0.00011900, 0.00025411, 0.00002314, -0.00019803, -0.00015976, 0.00002990, 0.00012096, 0.00003772, -0.00008085, -0.00009285, -0.00001360, 0.00004462, 0.00002593, -0.00002345, -0.00003868, -0.00001478, 0.00000859, 0.00000794, -0.00000352, -0.00000550, 0.00000284, 0.00000945, 0.00000780, 0.00000077, -0.00000658, -0.00001042, -0.00000776, 0.00000041, 0.00000653, 0.00000308, -0.00000731, -0.00001315, -0.00000778, 0.00000255, 0.00000719, 0.00000443, 0.00000143, 0.00000258, 0.00000370, -0.00000022, -0.00000655, -0.00000841, -0.00000434, -0.00000011, -0.00000049, -0.00000343, -0.00000425, -0.00000258, -0.00000199, -0.00000359, -0.00000365, 0.00000128, 0.00000894, 0.00001317, 0.00001046, 0.00000334, },
		{ 0.00001198, 0.00000628, -0.00000050, -0.00000429, -0.00000506, -0.00000528, -0.00000582, -0.00000424, 0.00000177, 0.00000981, 0.00001355, 0.00000946, 0.00000162, -0.00000258, -0.00000135, -0.00000021, -0.00000358, -0.00000761, -0.00000536, 0.00000224, 0.00000564, -0.00000102, -0.00001107, -0.00001309, -0.00000525, 0.00000310, 0.00000455, 0.00000233, 0.00000329, 0.00000667, 0.00000566, -0.00000166, -0.00000901, -0.00001026, -0.00000602, -0.00000003, 0.00000533, 0.00000662, -0.00000117, -0.00001518, -0.00001973, -0.00000319, 0.00001982, 0.00001919, -0.00000904, -0.00002719, -0.00000292, 0.00003627, 0.00002776, -0.00003596, -0.00007592, -0.00002311, 0.00007013, 0.00007342, -0.00004589, -0.00014591, -0.00007500, 0.00010568, 0.00015746, -0.00003112, -0.00024937, -0.00018813, 0.00013733, 0.00031550, 0.00000207, -0.00058600, -0.00078982, -0.00020811, 0.00078550, 0.00133989, 0.00096787, 0.00002897, -0.00065939, -0.00063410, -0.00015151, 0.00020607, 0.00015964, -0.00008493, -0.00018897, -0.00007280, 0.00007591, 0.00008472, -0.00001661, -0.00008545, -0.00005624, 0.00001004, 0.00003197, 0.00000334, -0.00002595, -0.00002453, -0.00000629, 0.00000360, 0.00000117, -0.00000198, 0.00000050, 0.00000508, 0.00000722, 0.00000567, 0.00000041, -0.00000681, -0.00001045, -0.00000589, 0.00000294, 0.00000581, -0.00000137, -0.00001046, -0.00001085, -0.00000256, 0.00000487, 0.00000529, 0.00000236, 0.00000200, 0.00000353, 0.00000198, -0.00000319, -0.00000683, -0.00000548, -0.00000206, -0.00000127, -0.00000313, -0.00000400, -0.00000260, -0.00000168, -0.00000311, -0.00000431, -0.00000121, 0.00000594, 0.00001196, },
		{ 0.00000302, 0.00000982, 0.00001227, 0.00000863, 0.00000206, -0.00000289, -0.00000461, -0.00000497, -0.00000581, -0.00000564, -0.00000125, 0.00000702, 0.00001326, 0.00001163, 0.00000378, -0.00000224, -0.00000160, 0.00000122, -0.00000095, -0.00000712, -0.00000856, -0.00000131, 0.00000656, 0.00000440, -0.00000672, -0.00001461, -0.00001107, -0.00000130, 0.00000456, 0.00000433, 0.00000360, 0.00000518, 0.00000535, 0.00000121, -0.00000425, -0.00000723, -0.00000800, -0.00000688, -0.00000091, 0.00000867, 0.00001022, -0.00000529, -0.00002461, -0.00002086, 0.00000913, 0.00003002, 0.00001002, -0.00002844, -0.00002857, 0.00002034, 0.00005244, 0.00000770, -0.00007120, -0.00007514, 0.00002198, 0.00010079, 0.00003818, -0.00010997, -0.00014832, 0.00000332, 0.00016613, 0.00010344, -0.00014954, -0.00027404, -0.00005960, 0.00026566, 0.00023955, -0.00025871, -0.00075047, -0.00060826, 0.00022976, 0.00110064, 0.00124485, 0.00056194, -0.00031713, -0.00068774, -0.00042506, 0.00002460, 0.00019708, 0.00004829, -0.00013638, -0.00013599, -0.00000307, 0.00007926, 0.00003721, -0.00004583, -0.00006843, -0.00002666, 0.00001453, 0.00001502, -0.00000757, -0.00002004, -0.00001494, -0.00000561, -0.00000122, 0.00000011, 0.00000168, 0.00000332, 0.00000472, 0.00000610, 0.00000522, -0.00000056, -0.00000805, -0.00000951, -0.00000279, 0.00000413, 0.00000251, -0.00000570, -0.00001036, -0.00000627, 0.00000116, 0.00000429, 0.00000287, 0.00000196, 0.00000324, 0.00000317, -0.00000047, -0.00000462, -0.00000532, -0.00000324, -0.00000218, -0.00000331, -0.00000411, -0.00000288, -0.00000152, -0.00000242, -0.00000420, -0.00000286, },
		{ -0.00000373, -0.00000365, 0.00000070, 0.00000736, 0.00001152, 0.00001007, 0.00000444, -0.00000114, -0.00000382, -0.00000443, -0.00000535, -0.00000635, -0.00000394, 0.00000366, 0.00001197, 0.00001344, 0.00000640, -0.00000169, -0.00000279, 0.00000151, 0.00000220, -0.00000424, -0.00000995, -0.00000608, 0.00000407, 0.00000791, -0.00000039, -0.00001206, -0.00001486, -0.00000749, 0.00000127, 0.00000511, 0.00000565, 0.00000579, 0.00000445, 0.00000078, -0.00000209, -0.00000237, -0.00000411, -0.00000951, -0.00001043, 0.00000153, 0.00001631, 0.00001111, -0.00001650, -0.00003427, -0.00001288, 0.00002707, 0.00003200, -0.00001119, -0.00004495, -0.00001332, 0.00005028, 0.00005238, -0.00003067, -0.00009804, -0.00004651, 0.00007483, 0.00010244, -0.00002566, -0.00015750, -0.00010427, 0.00009318, 0.00017693, -0.00000080, -0.00024317, -0.00021786, 0.00009824, 0.00031139, 0.00005398, -0.00050933, -0.00075868, -0.00027486, 0.00064504, 0.00122489, 0.00097398, 0.00015700, -0.00051315, -0.00058470, -0.00021832, 0.00010957, 0.00013008, -0.00003674, -0.00013324, -0.00007199, 0.00003267, 0.00005551, -0.00000062, -0.00005054, -0.00004568, -0.00001146, 0.00000811, 0.00000304, -0.00000876, -0.00001343, -0.00001165, -0.00000765, -0.00000255, 0.00000222, 0.00000406, 0.00000363, 0.00000441, 0.00000631, 0.00000451, -0.00000256, -0.00000864, -0.00000702, -0.00000024, 0.00000268, -0.00000181, -0.00000751, -0.00000724, -0.00000195, 0.00000206, 0.00000219, 0.00000161, 0.00000286, 0.00000384, 0.00000171, -0.00000226, -0.00000435, -0.00000370, -0.00000287, -0.00000360, -0.00000435, -0.00000335, -0.00000174, -0.00000196, },
		{ -0.00000211, -0.00000186, -0.00000337, -0.00000398, -0.00000093, 0.00000513, 0.00001023, 0.00001067, 0.00000633, 0.00000061, -0.00000288, -0.00000376, -0.00000448, -0.00000620, -0.00000595, -0.00000004, 0.00000947, 0.00001448, 0.00000961, -0.00000016, -0.00000438, -0.00000009, 0.00000423, 0.00000024, -0.00000823, -0.00000971, -0.00000093, 0.00000767, 0.00000509, -0.00000634, -0.00001452, -0.00001258, -0.00000468, 0.00000244, 0.00000690, 0.00000873, 0.00000603, -0.00000055, -0.00000418, -0.00000051, 0.00000290, -0.00000437, -0.00001609, -0.00001225, 0.00001069, 0.00002478, 0.00000375, -0.00003337, -0.00003679, 0.00000638, 0.00004297, 0.00001893, -0.00003906, -0.00004733, 0.00001870, 0.00007365, 0.00002809, -0.00007635, -0.00010041, 0.00000663, 0.00011425, 0.00006600, -0.00009841, -0.00016499, -0.00002242, 0.00016019, 0.00012524, -0.00012206, -0.00027264, -0.00009342, 0.00022637, 0.00024361, -0.00019191, -0.00066631, -0.00059274, 0.00012950, 0.00093906, 0.00114755, 0.00061362, -0.00015862, -0.00055626, -0.00041871, -0.00006801, 0.00011516, 0.00005106, -0.00007622, -0.00009761, -0.00002318, 0.00003656, 0.00002522, -0.00001959, -0.00004111, -0.00002892, -0.00000868, 0.00000019, -0.00000130, -0.00000590, -0.00001031, -0.00001210, -0.00000841, -0.00000085, 0.00000438, 0.00000446, 0.00000343, 0.00000510, 0.00000649, 0.00000270, -0.00000439, -0.00000757, -0.00000422, 0.00000005, -0.00000065, -0.00000467, -0.00000607, -0.00000316, 0.00000012, 0.00000074, 0.00000043, 0.00000185, 0.00000389, 0.00000342, 0.00000021, -0.00000276, -0.00000354, -0.00000337, -0.00000392, -0.00000454, -0.00000373, },
		{ -0.00000484, -0.00000396, -0.00000230, -0.00000186, -0.00000327, -0.00000434, -0.00000222, 0.00000324, 0.00000886, 0.00001081, 0.00000779, 0.00000220, -0.00000204, -0.00000325, -0.00000346, -0.00000518, -0.00000682, -0.00000354, 0.00000574, 0.00001407, 0.00001289, 0.00000292, -0.00000511, -0.00000318, 0.00000375, 0.00000429, -0.00000376, -0.00001011, -0.00000591, 0.00000395, 0.00000715, -0.00000030, -0.00001023, -0.00001376, -0.00001030, -0.00000365, 0.00000436, 0.00001132, 0.00001108, 0.00000127, -0.00000817, -0.00000472, 0.00000696, 0.00000669, -0.00001177, -0.00002440, -0.00000654, 0.00002566, 0.00002700, -0.00001437, -0.00004892, -0.00002547, 0.00003237, 0.00004641, -0.00000973, -0.00006168, -0.00002802, 0.00005795, 0.00007595, -0.00001793, -0.00011096, -0.00007013, 0.00006776, 0.00012052, -0.00000131, -0.00015217, -0.00012330, 0.00006976, 0.00017499, 0.00002426, -0.00021558, -0.00022222, 0.00005552, 0.00027144, 0.00007425, -0.00042029, -0.00067304, -0.00029163, 0.00050286, 0.00105478, 0.00091623, 0.00026097, -0.00034040, -0.00048826, -0.00025650, 0.00000797, 0.00007632, -0.00000625, -0.00007543, -0.00005637, -0.00000007, 0.00002268, 0.00000333, -0.00002227, -0.00002947, -0.00002190, -0.00001152, -0.00000371, -0.00000054, -0.00000386, -0.00001061, -0.00001278, -0.00000676, 0.00000165, 0.00000511, 0.00000406, 0.00000399, 0.00000602, 0.00000558, 0.00000051, -0.00000480, -0.00000551, -0.00000294, -0.00000196, -0.00000366, -0.00000460, -0.00000273, -0.00000041, -0.00000029, -0.00000114, -0.00000003, 0.00000284, 0.00000430, 0.00000260, -0.00000050, -0.00000261, -0.00000358, -0.00000440, },
		{ -0.00000307, -0.00000491, -0.00000556, -0.00000436, -0.00000225, -0.00000153, -0.00000308, -0.00000482, -0.00000360, 0.00000136, 0.00000748, 0.00001080, 0.00000913, 0.00000379, -0.00000131, -0.00000319, -0.00000281, -0.00000366, -0.00000629, -0.00000600, 0.00000133, 0.00001173, 0.00001515, 0.00000725, -0.00000380, -0.00000625, 0.00000065, 0.00000585, 0.00000140, -0.00000699, -0.00000818, -0.00000083, 0.00000531, 0.00000289, -0.00000477, -0.00001049, -0.00001209, -0.00001017, -0.00000280, 0.00000956, 0.00001693, 0.00000857, -0.00000879, -0.00001335, 0.00000271, 0.00001659, 0.00000291, -0.00002557, -0.00002734, 0.00000953, 0.00003987, 0.00001626, -0.00003931, -0.00005323, 0.00000066, 0.00005422, 0.00003008, -0.00004559, -0.00006573, 0.00001116, 0.00008803, 0.00004953, -0.00007143, -0.00011713, -0.00001310, 0.00011328, 0.00008599, -0.00007681, -0.00016381, -0.00004438, 0.00013701, 0.00012833, -0.00009015, -0.00024521, -0.00010965, 0.00017120, 0.00021028, -0.00014346, -0.00055493, -0.00052272, 0.00006320, 0.00075905, 0.00099291, 0.00061383, -0.00000992, -0.00039352, -0.00037061, -0.00014012, 0.00002372, 0.00003043, -0.00003009, -0.00005388, -0.00002806, 0.00000176, 0.00000662, -0.00000547, -0.00001740, -0.00002291, -0.00002175, -0.00001376, -0.00000312, 0.00000119, -0.00000435, -0.00001195, -0.00001175, -0.00000390, 0.00000323, 0.00000484, 0.00000413, 0.00000505, 0.00000609, 0.00000384, -0.00000080, -0.00000387, -0.00000412, -0.00000390, -0.00000458, -0.00000440, -0.00000206, 0.00000030, 0.00000000, -0.00000201, -0.00000223, 0.00000056, 0.00000371, 0.00000426, 0.00000218, -0.00000064, },
		{ 0.00000443, 0.00000220, -0.00000129, -0.00000483, -0.00000662, -0.00000542, -0.00000242, -0.00000079, -0.00000229, -0.00000501, -0.00000513, -0.00000091, 0.00000568, 0.00001055, 0.00001056, 0.00000577, -0.00000031, -0.00000349, -0.00000303, -0.00000244, -0.00000466, -0.00000676, -0.00000268, 0.00000771, 0.00001535, 0.00001160, -0.00000022, -0.00000750, -0.00000359, 0.00000417, 0.00000469, -0.00000219, -0.00000673, -0.00000357, 0.00000175, 0.00000207, -0.00000183, -0.00000541, -0.00000864, -0.00001267, -0.00001168, 0.00000128, 0.00001835, 0.00001928, -0.00000154, -0.00002008, -0.00001014, 0.00001699, 0.00002099, -0.00001146, -0.00003934, -0.00001808, 0.00003244, 0.00004373, -0.00000871, -0.00006086, -0.00003907, 0.00003414, 0.00005945, -0.00000473, -0.00007374, -0.00004391, 0.00005756, 0.00009317, -0.00000075, -0.00011165, -0.00008788, 0.00005141, 0.00012353, 0.00002064, -0.00013181, -0.00012698, 0.00004234, 0.00015271, 0.00003716, -0.00017482, -0.00020114, 0.00001977, 0.00020722, 0.00006204, -0.00033530, -0.00055163, -0.00026055, 0.00037902, 0.00085618, 0.00080436, 0.00032501, -0.00016790, -0.00035874, -0.00025751, -0.00007939, 0.00000924, 0.00000131, -0.00002782, -0.00003176, -0.00001765, -0.00000581, -0.00000221, -0.00000495, -0.00001314, -0.00002238, -0.00002332, -0.00001281, -0.00000062, 0.00000134, -0.00000631, -0.00001224, -0.00000925, -0.00000142, 0.00000371, 0.00000468, 0.00000487, 0.00000566, 0.00000513, 0.00000237, -0.00000081, -0.00000293, -0.00000459, -0.00000608, -0.00000572, -0.00000253, 0.00000105, 0.00000145, -0.00000136, -0.00000358, -0.00000226, 0.00000147, 0.00000431, },
		{ -0.00000170, 0.00000235, 0.00000510, 0.00000501, 0.00000179, -0.00000334, -0.00000731, -0.00000722, -0.00000352, -0.00000023, -0.00000080, -0.00000427, -0.00000627, -0.00000355, 0.00000305, 0.00000954, 0.00001179, 0.00000832, 0.00000156, -0.00000357, -0.00000410, -0.00000232, -0.00000287, -0.00000583, -0.00000517, 0.00000310, 0.00001321, 0.00001445, 0.00000458, -0.00000608, -0.00000683, 0.00000036, 0.00000470, 0.00000146, -0.00000309, -0.00000267, -0.00000007, -0.00000080, -0.00000332, -0.00000291, -0.00000236, -0.00000849, -0.00001673, -0.00001128, 0.00001104, 0.00002711, 0.00001330, -0.00001761, -0.00002529, 0.00000407, 0.00003141, 0.00001327, -0.00003302, -0.00004355, 0.00000489, 0.00005244, 0.00003006, -0.00004107, -0.00006709, -0.00000727, 0.00006116, 0.00004101, -0.00004665, -0.00007963, -0.00000037, 0.00009248, 0.00006684, -0.00005793, -0.00012115, -0.00003162, 0.00009884, 0.00009314, -0.00005115, -0.00014527, -0.00005651, 0.00010312, 0.00011213, -0.00006210, -0.00019938, -0.00010556, 0.00011296, 0.00014989, -0.00011877, -0.00043626, -0.00041487, 0.00003543, 0.00058559, 0.00080530, 0.00056467, 0.00010764, -0.00022436, -0.00028804, -0.00017873, -0.00005949, -0.00000760, -0.00000580, -0.00001546, -0.00002046, -0.00001940, -0.00001200, -0.00000269, -0.00000193, -0.00001315, -0.00002476, -0.00002304, -0.00000952, 0.00000118, -0.00000018, -0.00000790, -0.00001100, -0.00000653, 0.00000005, 0.00000387, 0.00000509, 0.00000555, 0.00000536, 0.00000399, 0.00000192, -0.00000038, -0.00000337, -0.00000657, -0.00000750, -0.00000438, 0.00000061, 0.00000288, 0.00000062, -0.00000317, -0.00000430, },
		{ -0.00000115, -0.00000449, -0.00000434, -0.00000096, 0.00000352, 0.00000640, 0.00000529, -0.00000012, -0.00000654, -0.00000900, -0.00000580, -0.00000074, 0.00000079, -0.00000239, -0.00000622, -0.00000590, -0.00000038, 0.00000720, 0.00001209, 0.00001104, 0.00000462, -0.00000251, -0.00000530, -0.00000353, -0.00000206, -0.00000409, -0.00000572, -0.00000068, 0.00000957, 0.00001499, 0.00000889, -0.00000263, -0.00000767, -0.00000329, 0.00000220, 0.00000206, -0.00000031, 0.00000059, 0.00000208, -0.00000172, -0.00000753, -0.00000604, 0.00000144, 0.00000066, -0.00001330, -0.00002165, -0.00000451, 0.00002506, 0.00002942, -0.00000285, -0.00003280, -0.00001796, 0.00002549, 0.00003640, -0.00000809, -0.00005216, -0.00003106, 0.00003509, 0.00005798, -0.00000059, -0.00006771, -0.00005104, 0.00003179, 0.00006807, 0.00000242, -0.00007855, -0.00005708, 0.00004965, 0.00010006, 0.00001635, -0.00010021, -0.00009495, 0.00003068, 0.00011119, 0.00003451, -0.00010184, -0.00011438, 0.00001804, 0.00011595, 0.00003520, -0.00013063, -0.00016080, -0.00000276, 0.00013327, 0.00002466, -0.00026532, -0.00041640, -0.00019362, 0.00028555, 0.00065673, 0.00065758, 0.00034333, -0.00001985, -0.00021667, -0.00022221, -0.00013732, -0.00005657, -0.00001217, 0.00000079, -0.00000670, -0.00002057, -0.00002402, -0.00001207, 0.00000099, -0.00000135, -0.00001658, -0.00002653, -0.00002029, -0.00000604, 0.00000134, -0.00000209, -0.00000820, -0.00000900, -0.00000447, 0.00000093, 0.00000434, 0.00000568, 0.00000556, 0.00000462, 0.00000351, 0.00000205, -0.00000094, -0.00000538, -0.00000835, -0.00000671, -0.00000130, 0.00000310, 0.00000277, },
		{ 0.00000178, 0.00000384, 0.00000134, -0.00000281, -0.00000518, -0.00000410, 0.00000015, 0.00000537, 0.00000768, 0.00000411, -0.00000365, -0.00000949, -0.00000863, -0.00000286, 0.00000137, 0.00000004, -0.00000459, -0.00000697, -0.00000383, 0.00000352, 0.00001065, 0.00001291, 0.00000845, 0.00000028, -0.00000546, -0.00000543, -0.00000282, -0.00000286, -0.00000496, -0.00000281, 0.00000588, 0.00001365, 0.00001151, 0.00000118, -0.00000636, -0.00000503, -0.00000049, 0.00000013, -0.00000078, 0.00000274, 0.00000713, 0.00000251, -0.00000967, -0.00001371, -0.00000207, 0.00000871, -0.00000140, -0.00002265, -0.00002189, 0.00001036, 0.00003725, 0.00001976, -0.00002503, -0.00003795, 0.00000294, 0.00004447, 0.00002499, -0.00003668, -0.00005759, -0.00000233, 0.00005991, 0.00004223, -0.00003841, -0.00007553, -0.00001572, 0.00006214, 0.00004888, -0.00004243, -0.00008582, -0.00001263, 0.00008638, 0.00007567, -0.00003961, -0.00011150, -0.00004372, 0.00007524, 0.00008596, -0.00002784, -0.00011408, -0.00005607, 0.00006661, 0.00008118, -0.00004335, -0.00014576, -0.00008452, 0.00006183, 0.00007713, -0.00011551, -0.00032657, -0.00029061, 0.00004103, 0.00043724, 0.00061185, 0.00047829, 0.00018175, -0.00007298, -0.00018595, -0.00017972, -0.00011922, -0.00005192, -0.00000438, 0.00000893, -0.00000693, -0.00002598, -0.00002489, -0.00000722, 0.00000395, -0.00000430, -0.00002063, -0.00002603, -0.00001642, -0.00000376, 0.00000051, -0.00000325, -0.00000746, -0.00000714, -0.00000301, 0.00000182, 0.00000501, 0.00000583, 0.00000510, 0.00000429, 0.00000362, 0.00000148, -0.00000308, -0.00000767, -0.00000824, -0.00000382, },
		{ -0.00000840, -0.00000577, -0.00000032, 0.00000338, 0.00000299, -0.00000033, -0.00000396, -0.00000557, -0.00000350, 0.00000209, 0.00000760, 0.00000772, 0.00000090, -0.00000765, -0.00001059, -0.00000616, 0.00000011, 0.00000177, -0.00000189, -0.00000616, -0.00000613, -0.00000068, 0.00000731, 0.00001287, 0.00001186, 0.00000451, -0.00000364, -0.00000669, -0.00000471, -0.00000306, -0.00000420, -0.00000361, 0.00000318, 0.00001175, 0.00001257, 0.00000399, -0.00000451, -0.00000493, -0.00000130, -0.00000175, -0.00000419, 0.00000060, 0.00001095, 0.00001140, -0.00000495, -0.00002044, -0.00001337, 0.00000850, 0.00001299, -0.00001119, -0.00003149, -0.00001239, 0.00002969, 0.00003952, -0.00000217, -0.00004379, -0.00002675, 0.00003072, 0.00004991, -0.00000223, -0.00006020, -0.00004270, 0.00003320, 0.00006674, 0.00000804, -0.00006819, -0.00005869, 0.00002596, 0.00007005, 0.00000945, -0.00007537, -0.00006412, 0.00003676, 0.00009516, 0.00002883, -0.00007978, -0.00008942, 0.00001122, 0.00008704, 0.00003730, -0.00006936, -0.00008895, 0.00000177, 0.00007330, 0.00002046, -0.00009131, -0.00011133, -0.00001131, 0.00006301, -0.00002533, -0.00021414, -0.00028725, -0.00010876, 0.00022492, 0.00047911, 0.00049953, 0.00032069, 0.00008733, -0.00008391, -0.00016013, -0.00015860, -0.00010700, -0.00003869, 0.00000759, 0.00001062, -0.00001379, -0.00002994, -0.00002064, -0.00000140, 0.00000368, -0.00000912, -0.00002311, -0.00002368, -0.00001292, -0.00000271, -0.00000028, -0.00000348, -0.00000641, -0.00000565, -0.00000169, 0.00000280, 0.00000535, 0.00000554, 0.00000492, 0.00000456, 0.00000323, -0.00000070, -0.00000592, },
		{ 0.00000118, -0.00000406, -0.00000765, -0.00000661, -0.00000207, 0.00000209, 0.00000322, 0.00000152, -0.00000166, -0.00000484, -0.00000571, -0.00000203, 0.00000484, 0.00000904, 0.00000556, -0.00000356, -0.00001034, -0.00000925, -0.00000287, 0.00000170, 0.00000061, -0.00000378, -0.00000639, -0.00000405, 0.00000282, 0.00001045, 0.00001347, 0.00000899, 0.00000017, -0.00000607, -0.00000642, -0.00000448, -0.00000451, -0.00000429, 0.00000132, 0.00001035, 0.00001334, 0.00000589, -0.00000362, -0.00000482, -0.00000021, -0.00000089, -0.00000736, -0.00000605, 0.00000890, 0.00002020, 0.00000766, -0.00001926, -0.00002710, -0.00000336, 0.00002049, 0.00000850, -0.00002614, -0.00003253, 0.00000728, 0.00004503, 0.00002658, -0.00002958, -0.00004873, -0.00000031, 0.00005321, 0.00003562, -0.00003594, -0.00006646, -0.00001030, 0.00006089, 0.00004998, -0.00003230, -0.00007677, -0.00002254, 0.00005699, 0.00005134, -0.00003460, -0.00008292, -0.00002217, 0.00007168, 0.00007385, -0.00002127, -0.00009065, -0.00004631, 0.00004852, 0.00006683, -0.00001164, -0.00007757, -0.00004442, 0.00003462, 0.00004335, -0.00003515, -0.00009471, -0.00005432, 0.00002320, 0.00000679, -0.00012546, -0.00023556, -0.00017066, 0.00006781, 0.00032322, 0.00043632, 0.00037300, 0.00021089, 0.00004262, -0.00008278, -0.00014824, -0.00014666, -0.00008999, -0.00002023, 0.00001532, 0.00000486, -0.00002158, -0.00002910, -0.00001367, 0.00000217, 0.00000055, -0.00001362, -0.00002356, -0.00002071, -0.00001033, -0.00000215, -0.00000057, -0.00000325, -0.00000543, -0.00000427, -0.00000037, 0.00000345, 0.00000524, 0.00000548, 0.00000541, 0.00000452, },
		{ 0.00000621, 0.00000590, 0.00000286, -0.00000254, -0.00000687, -0.00000680, -0.00000292, 0.00000109, 0.00000266, 0.00000208, 0.00000025, -0.00000272, -0.00000561, -0.00000509, 0.00000058, 0.00000741, 0.00000840, 0.00000144, -0.00000748, -0.00001053, -0.00000633, -0.00000040, 0.00000156, -0.00000105, -0.00000470, -0.00000542, -0.00000134, 0.00000623, 0.00001247, 0.00001212, 0.00000489, -0.00000326, -0.00000657, -0.00000590, -0.00000579, -0.00000591, -0.00000095, 0.00000921, 0.00001492, 0.00000841, -0.00000351, -0.00000662, 0.00000042, 0.00000311, -0.00000639, -0.00001351, -0.00000060, 0.00002229, 0.00002355, -0.00000642, -0.00003421, -0.00002322, 0.00001380, 0.00002607, -0.00000618, -0.00003881, -0.00002076, 0.00003107, 0.00004722, 0.00000008, -0.00005085, -0.00003445, 0.00003192, 0.00005876, 0.00000460, -0.00006193, -0.00005024, 0.00002765, 0.00006817, 0.00001486, -0.00006253, -0.00006015, 0.00001827, 0.00006474, 0.00001405, -0.00006541, -0.00006315, 0.00002246, 0.00007990, 0.00003363, -0.00005539, -0.00007276, -0.00000245, 0.00005722, 0.00002960, -0.00004086, -0.00005723, -0.00000502, 0.00003334, -0.00000129, -0.00006160, -0.00006347, -0.00000983, 0.00000565, -0.00007424, -0.00017914, -0.00017802, -0.00002465, 0.00019144, 0.00033708, 0.00035241, 0.00026973, 0.00014778, 0.00002174, -0.00008627, -0.00014523, -0.00013251, -0.00006711, -0.00000378, 0.00001524, -0.00000447, -0.00002581, -0.00002409, -0.00000723, 0.00000261, -0.00000370, -0.00001662, -0.00002266, -0.00001793, -0.00000837, -0.00000158, -0.00000053, -0.00000288, -0.00000439, -0.00000284, 0.00000059, 0.00000357, 0.00000529, },
		{ 0.00000089, 0.00000367, 0.00000623, 0.00000734, 0.00000493, -0.00000095, -0.00000647, -0.00000728, -0.00000340, 0.00000094, 0.00000246, 0.00000184, 0.00000089, -0.00000072, -0.00000377, -0.00000588, -0.00000320, 0.00000365, 0.00000831, 0.00000533, -0.00000304, -0.00000921, -0.00000855, -0.00000351, 0.00000043, 0.00000050, -0.00000215, -0.00000457, -0.00000375, 0.00000168, 0.00000916, 0.00001278, 0.00000888, 0.00000084, -0.00000467, -0.00000591, -0.00000679, -0.00000832, -0.00000477, 0.00000673, 0.00001688, 0.00001308, -0.00000200, -0.00001035, -0.00000245, 0.00000732, 0.00000028, -0.00001612, -0.00001406, 0.00001353, 0.00003421, 0.00001521, -0.00002720, -0.00004113, -0.00000716, 0.00003032, 0.00001942, -0.00002607, -0.00004069, 0.00000250, 0.00004861, 0.00003183, -0.00003069, -0.00005526, -0.00000449, 0.00005615, 0.00004269, -0.00003138, -0.00006828, -0.00001680, 0.00005556, 0.00005139, -0.00002457, -0.00007063, -0.00002587, 0.00004700, 0.00004740, -0.00002557, -0.00007178, -0.00002647, 0.00005217, 0.00006200, -0.00000707, -0.00006361, -0.00003932, 0.00002460, 0.00004113, -0.00000431, -0.00004326, -0.00002619, 0.00001138, 0.00000727, -0.00003491, -0.00005360, -0.00002420, -0.00000259, -0.00004994, -0.00013814, -0.00016582, -0.00007044, 0.00010107, 0.00024345, 0.00029448, 0.00026801, 0.00020297, 0.00011407, 0.00000466, -0.00009645, -0.00014146, -0.00011176, -0.00004347, 0.00000551, 0.00000904, -0.00001260, -0.00002534, -0.00001756, -0.00000321, 0.00000080, -0.00000761, -0.00001817, -0.00002118, -0.00001543, -0.00000663, -0.00000097, -0.00000039, -0.00000230, -0.00000315, -0.00000173, },
		{ -0.00000135, -0.00000205, -0.00000137, 0.00000089, 0.00000466, 0.00000803, 0.00000744, 0.00000148, -0.00000590, -0.00000849, -0.00000461, 0.00000106, 0.00000330, 0.00000206, 0.00000064, 0.00000004, -0.00000170, -0.00000461, -0.00000496, -0.00000031, 0.00000569, 0.00000665, 0.00000104, -0.00000591, -0.00000842, -0.00000595, -0.00000208, 0.00000005, -0.00000034, -0.00000240, -0.00000385, -0.00000161, 0.00000486, 0.00001097, 0.00001087, 0.00000465, -0.00000156, -0.00000403, -0.00000596, -0.00000989, -0.00000985, 0.00000124, 0.00001692, 0.00001961, 0.00000352, -0.00001320, -0.00000981, 0.00000710, 0.00000988, -0.00001000, -0.00002415, -0.00000431, 0.00003202, 0.00003643, -0.00000527, -0.00004587, -0.00003382, 0.00001590, 0.00003708, 0.00000086, -0.00004172, -0.00002750, 0.00002899, 0.00005046, 0.00000275, -0.00005278, -0.00003902, 0.00002949, 0.00006119, 0.00001048, -0.00005736, -0.00005180, 0.00002019, 0.00006212, 0.00001799, -0.00005232, -0.00005496, 0.00001073, 0.00005323, 0.00001469, -0.00005132, -0.00005445, 0.00001017, 0.00005804, 0.00003012, -0.00003233, -0.00004928, -0.00000828, 0.00002842, 0.00001510, -0.00002040, -0.00002667, -0.00000431, 0.00000224, -0.00002288, -0.00004220, -0.00002558, -0.00000468, -0.00003497, -0.00011113, -0.00015389, -0.00009494, 0.00004402, 0.00017477, 0.00023443, 0.00023211, 0.00020657, 0.00016595, 0.00009046, -0.00001647, -0.00010699, -0.00013064, -0.00008694, -0.00002484, 0.00000702, 0.00000091, -0.00001690, -0.00002164, -0.00001184, -0.00000171, -0.00000193, -0.00001064, -0.00001872, -0.00001939, -0.00001304, -0.00000509, -0.00000049, -0.00000005, },
		{ -0.00000393, 0.00000001, 0.00000078, -0.00000027, -0.00000170, -0.00000165, 0.00000160, 0.00000694, 0.00000946, 0.00000498, -0.00000407, -0.00000978, -0.00000721, 0.00000010, 0.00000462, 0.00000363, 0.00000081, -0.00000029, -0.00000073, -0.00000267, -0.00000450, -0.00000270, 0.00000225, 0.00000533, 0.00000305, -0.00000239, -0.00000618, -0.00000636, -0.00000436, -0.00000201, -0.00000037, -0.00000047, -0.00000224, -0.00000279, 0.00000112, 0.00000772, 0.00001066, 0.00000694, 0.00000121, -0.00000127, -0.00000286, -0.00000860, -0.00001400, -0.00000720, 0.00001224, 0.00002506, 0.00001375, -0.00001090, -0.00001886, -0.00000067, 0.00001620, 0.00000360, -0.00002386, -0.00002336, 0.00001523, 0.00004618, 0.00002386, -0.00003153, -0.00005276, -0.00001273, 0.00003599, 0.00002894, -0.00002235, -0.00004425, -0.00000252, 0.00004715, 0.00003374, -0.00002881, -0.00005613, -0.00000803, 0.00005318, 0.00004450, -0.00002454, -0.00006279, -0.00001999, 0.00004551, 0.00004609, -0.00001726, -0.00005858, -0.00002479, 0.00003449, 0.00003772, -0.00001766, -0.00005514, -0.00002477, 0.00003221, 0.00004331, 0.00000067, -0.00003632, -0.00002556, 0.00000745, 0.00001539, -0.00000448, -0.00001678, -0.00000750, -0.00000242, -0.00002028, -0.00003785, -0.00002554, -0.00000180, -0.00001873, -0.00008742, -0.00014441, -0.00011455, 0.00000220, 0.00012805, 0.00019118, 0.00019220, 0.00017856, 0.00017159, 0.00014324, 0.00006535, -0.00003933, -0.00011106, -0.00011267, -0.00006347, -0.00001387, 0.00000359, -0.00000564, -0.00001738, -0.00001689, -0.00000791, -0.00000193, -0.00000476, -0.00001283, -0.00001849, -0.00001729, -0.00001083, },
		{ -0.00001748, -0.00001505, -0.00000908, -0.00000303, 0.00000096, 0.00000202, 0.00000022, -0.00000230, -0.00000161, 0.00000387, 0.00000967, 0.00000866, -0.00000038, -0.00000972, -0.00001057, -0.00000294, 0.00000484, 0.00000614, 0.00000256, -0.00000041, -0.00000103, -0.00000159, -0.00000301, -0.00000294, -0.00000012, 0.00000280, 0.00000254, -0.00000050, -0.00000336, -0.00000462, -0.00000487, -0.00000422, -0.00000218, -0.00000008, -0.00000032, -0.00000210, -0.00000104, 0.00000443, 0.00000913, 0.00000763, 0.00000249, 0.00000050, 0.00000109, -0.00000370, -0.00001414, -0.00001589, 0.00000182, 0.00002497, 0.00002574, -0.00000069, -0.00002378, -0.00001454, 0.00001320, 0.00001799, -0.00001117, -0.00003356, -0.00000989, 0.00003785, 0.00004723, -0.00000170, -0.00005279, -0.00004272, 0.00001447, 0.00004355, 0.00000823, -0.00003949, -0.00003067, 0.00002426, 0.00004821, 0.00000444, -0.00004964, -0.00003911, 0.00002468, 0.00005692, 0.00001373, -0.00004790, -0.00004695, 0.00001286, 0.00005011, 0.00001662, -0.00003996, -0.00004427, 0.00000502, 0.00003799, 0.00001117, -0.00003628, -0.00004028, 0.00000209, 0.00003448, 0.00002016, -0.00001458, -0.00002470, -0.00000706, 0.00000593, -0.00000090, -0.00000884, -0.00000328, -0.00000017, -0.00001742, -0.00003823, -0.00003074, -0.00000177, -0.00000180, -0.00005973, -0.00013017, -0.00013151, -0.00003747, 0.00008929, 0.00016414, 0.00016670, 0.00014566, 0.00014590, 0.00015351, 0.00012173, 0.00003703, -0.00005781, -0.00010620, -0.00009169, -0.00004547, -0.00000959, -0.00000118, -0.00000922, -0.00001546, -0.00001255, -0.00000571, -0.00000315, -0.00000740, -0.00001417, },
		{ -0.00000501, -0.00000961, -0.00001450, -0.00001587, -0.00001313, -0.00000784, -0.00000187, 0.00000253, 0.00000296, -0.00000045, -0.00000313, -0.00000003, 0.00000740, 0.00001090, 0.00000451, -0.00000718, -0.00001300, -0.00000770, 0.00000254, 0.00000793, 0.00000579, 0.00000109, -0.00000141, -0.00000184, -0.00000209, -0.00000193, -0.00000051, 0.00000097, 0.00000061, -0.00000102, -0.00000190, -0.00000201, -0.00000314, -0.00000494, -0.00000459, -0.00000154, 0.00000058, -0.00000061, -0.00000161, 0.00000195, 0.00000739, 0.00000767, 0.00000251, -0.00000004, 0.00000325, 0.00000307, -0.00000851, -0.00002034, -0.00001180, 0.00001629, 0.00003359, 0.00001585, -0.00001895, -0.00002817, -0.00000069, 0.00002432, 0.00000873, -0.00002843, -0.00003170, 0.00001366, 0.00005333, 0.00003147, -0.00003145, -0.00005927, -0.00001846, 0.00003646, 0.00003480, -0.00001600, -0.00004217, -0.00000617, 0.00004111, 0.00003128, -0.00002513, -0.00005131, -0.00000960, 0.00004551, 0.00004064, -0.00001728, -0.00005138, -0.00001901, 0.00003316, 0.00003530, -0.00001197, -0.00004326, -0.00001966, 0.00002199, 0.00002445, -0.00001234, -0.00003666, -0.00001821, 0.00001541, 0.00002242, 0.00000200, -0.00001384, -0.00000962, -0.00000171, -0.00000481, -0.00000860, -0.00000057, 0.00000627, -0.00000892, -0.00003603, -0.00003925, -0.00000972, 0.00000894, -0.00002958, -0.00010549, -0.00013916, -0.00007645, 0.00004690, 0.00014116, 0.00015681, 0.00012653, 0.00011281, 0.00013132, 0.00013989, 0.00009620, 0.00000982, -0.00006774, -0.00009464, -0.00007239, -0.00003393, -0.00000921, -0.00000499, -0.00001019, -0.00001257, -0.00000915, -0.00000489, },
		{ -0.00000948, -0.00000682, -0.00000526, -0.00000713, -0.00001095, -0.00001380, -0.00001419, -0.00001182, -0.00000665, -0.00000002, 0.00000420, 0.00000275, -0.00000194, -0.00000279, 0.00000345, 0.00001051, 0.00000886, -0.00000233, -0.00001282, -0.00001247, -0.00000244, 0.00000715, 0.00000893, 0.00000451, -0.00000029, -0.00000247, -0.00000251, -0.00000128, 0.00000048, 0.00000105, -0.00000079, -0.00000316, -0.00000290, -0.00000051, -0.00000013, -0.00000333, -0.00000595, -0.00000404, -0.00000013, 0.00000053, -0.00000130, 0.00000027, 0.00000593, 0.00000816, 0.00000279, -0.00000239, 0.00000149, 0.00000796, 0.00000148, -0.00001689, -0.00002292, 0.00000008, 0.00003148, 0.00003238, -0.00000303, -0.00003325, -0.00002050, 0.00001686, 0.00002558, -0.00000935, -0.00003926, -0.00001575, 0.00003856, 0.00005323, 0.00000303, -0.00005356, -0.00004750, 0.00001040, 0.00004418, 0.00001401, -0.00003271, -0.00002918, 0.00001834, 0.00004082, 0.00000408, -0.00004264, -0.00003454, 0.00001907, 0.00004719, 0.00001354, -0.00003582, -0.00003682, 0.00000729, 0.00003482, 0.00001123, -0.00002792, -0.00003048, 0.00000194, 0.00002208, 0.00000467, -0.00002303, -0.00002411, -0.00000132, 0.00001376, 0.00000730, -0.00000387, -0.00000427, -0.00000172, -0.00000766, -0.00001338, -0.00000433, 0.00000991, 0.00000311, -0.00002657, -0.00004425, -0.00002360, 0.00000820, -0.00000459, -0.00007219, -0.00013115, -0.00010741, -0.00000081, 0.00011046, 0.00015175, 0.00012476, 0.00009157, 0.00009747, 0.00012489, 0.00012294, 0.00006840, -0.00001138, -0.00006912, -0.00008057, -0.00005734, -0.00002744, -0.00001020, -0.00000716, -0.00000943, },
		{ -0.00001123, -0.00000782, -0.00000756, -0.00000664, -0.00000574, -0.00000662, -0.00000893, -0.00001112, -0.00001256, -0.00001301, -0.00001086, -0.00000484, 0.00000235, 0.00000497, 0.00000123, -0.00000292, -0.00000028, 0.00000765, 0.00001096, 0.00000325, -0.00000951, -0.00001508, -0.00000858, 0.00000306, 0.00000982, 0.00000851, 0.00000302, -0.00000178, -0.00000358, -0.00000205, 0.00000126, 0.00000276, -0.00000017, -0.00000502, -0.00000597, -0.00000161, 0.00000223, 0.00000011, -0.00000511, -0.00000616, -0.00000207, 0.00000075, -0.00000081, -0.00000115, 0.00000426, 0.00000922, 0.00000496, -0.00000414, -0.00000392, 0.00000729, 0.00001116, -0.00000550, -0.00002544, -0.00001778, 0.00001758, 0.00004055, 0.00001909, -0.00002440, -0.00003657, -0.00000298, 0.00002954, 0.00001445, -0.00002858, -0.00003669, 0.00000962, 0.00005401, 0.00003573, -0.00002754, -0.00005909, -0.00002237, 0.00003218, 0.00003565, -0.00000873, -0.00003490, -0.00000724, 0.00003200, 0.00002464, -0.00002102, -0.00004205, -0.00000861, 0.00003514, 0.00003211, -0.00001119, -0.00003662, -0.00001425, 0.00002100, 0.00002151, -0.00000935, -0.00002770, -0.00001203, 0.00001149, 0.00001049, -0.00000982, -0.00001988, -0.00000926, 0.00000380, 0.00000394, -0.00000105, 0.00000091, 0.00000386, -0.00000428, -0.00001636, -0.00001264, 0.00000617, 0.00001220, -0.00001153, -0.00004047, -0.00003633, -0.00000295, 0.00000828, -0.00003842, -0.00010783, -0.00012216, -0.00004655, 0.00006857, 0.00013893, 0.00013162, 0.00008876, 0.00007135, 0.00009367, 0.00011766, 0.00010154, 0.00004281, -0.00002458, -0.00006481, -0.00006738, -0.00004671, -0.00002401, },
		{ -0.00005648, -0.00003948, -0.00002225, -0.00001183, -0.00000714, -0.00000501, -0.00000453, -0.00000602, -0.00000836, -0.00000974, -0.00001032, -0.00001152, -0.00001242, -0.00000954, -0.00000221, 0.00000428, 0.00000420, -0.00000059, -0.00000193, 0.00000395, 0.00001027, 0.00000753, -0.00000420, -0.00001431, -0.00001348, -0.00000324, 0.00000718, 0.00001083, 0.00000755, 0.00000127, -0.00000361, -0.00000388, 0.00000044, 0.00000453, 0.00000258, -0.00000471, -0.00000930, -0.00000543, 0.00000213, 0.00000373, -0.00000205, -0.00000669, -0.00000430, 0.00000013, -0.00000032, -0.00000255, 0.00000152, 0.00000955, 0.00000919, -0.00000244, -0.00000984, 0.00000009, 0.00001483, 0.00000919, -0.00001660, -0.00002901, -0.00000380, 0.00003469, 0.00003786, -0.00000305, -0.00003962, -0.00002653, 0.00001697, 0.00003061, -0.00000510, -0.00003964, -0.00001992, 0.00003441, 0.00005275, 0.00000705, -0.00004840, -0.00004667, 0.00000539, 0.00003923, 0.00001676, -0.00002312, -0.00002326, 0.00001276, 0.00002984, 0.00000140, -0.00003361, -0.00002638, 0.00001399, 0.00003428, 0.00001020, -0.00002355, -0.00002375, 0.00000420, 0.00001931, 0.00000346, -0.00001803, -0.00001653, 0.00000125, 0.00000825, -0.00000276, -0.00001316, -0.00000949, -0.00000116, -0.00000113, -0.00000449, 0.00000042, 0.00000866, 0.00000404, -0.00001273, -0.00001931, -0.00000363, 0.00001326, 0.00000280, -0.00002833, -0.00004133, -0.00001788, 0.00000774, -0.00001300, -0.00007638, -0.00011809, -0.00008057, 0.00002114, 0.00011227, 0.00013446, 0.00009870, 0.00006233, 0.00006462, 0.00009291, 0.00010588, 0.00007870, 0.00002265, -0.00003089, -0.00005809, },
		{ 0.00000844, -0.00003270, -0.00005093, -0.00004788, -0.00003472, -0.00002144, -0.00001166, -0.00000521, -0.00000236, -0.00000371, -0.00000736, -0.00000964, -0.00000938, -0.00000929, -0.00001112, -0.00001180, -0.00000729, 0.00000054, 0.00000476, 0.00000237, -0.00000104, 0.00000139, 0.00000787, 0.00000932, 0.00000104, -0.00001066, -0.00001531, -0.00000940, 0.00000153, 0.00000965, 0.00001102, 0.00000616, -0.00000113, -0.00000512, -0.00000205, 0.00000466, 0.00000612, -0.00000156, -0.00001074, -0.00001056, -0.00000120, 0.00000561, 0.00000216, -0.00000517, -0.00000593, -0.00000092, 0.00000044, -0.00000336, -0.00000244, 0.00000731, 0.00001348, 0.00000384, -0.00001172, -0.00001056, 0.00000913, 0.00001969, 0.00000055, -0.00002735, -0.00002351, 0.00001567, 0.00004396, 0.00002234, -0.00002622, -0.00004216, -0.00000703, 0.00003040, 0.00001904, -0.00002437, -0.00003670, 0.00000474, 0.00004823, 0.00003506, -0.00002144, -0.00005234, -0.00002292, 0.00002484, 0.00003154, -0.00000231, -0.00002416, -0.00000548, 0.00002182, 0.00001514, -0.00001751, -0.00003047, -0.00000542, 0.00002424, 0.00002096, -0.00000709, -0.00002144, -0.00000723, 0.00001093, 0.00000769, -0.00000901, -0.00001447, -0.00000406, 0.00000401, -0.00000136, -0.00000931, -0.00000723, -0.00000081, -0.00000247, -0.00000884, -0.00000549, 0.00000760, 0.00001167, -0.00000316, -0.00001918, -0.00001373, 0.00000638, 0.00001045, -0.00001319, -0.00003684, -0.00002909, -0.00000154, -0.00000025, -0.00004627, -0.00009966, -0.00009711, -0.00002204, 0.00007442, 0.00012487, 0.00011008, 0.00006839, 0.00004898, 0.00006483, 0.00008962, 0.00009047, 0.00005784, },
		{ 0.00008260, 0.00007409, 0.00004056, -0.00000096, -0.00003185, -0.00004405, -0.00004138, -0.00003197, -0.00002096, -0.00001027, -0.00000235, -0.00000044, -0.00000431, -0.00000892, -0.00000983, -0.00000835, -0.00000875, -0.00001105, -0.00001038, -0.00000440, 0.00000224, 0.00000368, 0.00000107, 0.00000097, 0.00000562, 0.00000903, 0.00000458, -0.00000591, -0.00001387, -0.00001312, -0.00000497, 0.00000491, 0.00001122, 0.00001077, 0.00000383, -0.00000396, -0.00000479, 0.00000245, 0.00000846, 0.00000357, -0.00000888, -0.00001478, -0.00000709, 0.00000440, 0.00000591, -0.00000196, -0.00000646, -0.00000221, 0.00000159, -0.00000253, -0.00000633, 0.00000163, 0.00001451, 0.00001266, -0.00000645, -0.00001860, -0.00000427, 0.00001961, 0.00001794, -0.00001258, -0.00003206, -0.00000863, 0.00003360, 0.00004034, -0.00000131, -0.00004156, -0.00003106, 0.00001357, 0.00003159, 0.00000016, -0.00003455, -0.00002077, 0.00002692, 0.00004570, 0.00000870, -0.00003901, -0.00004016, 0.00000129, 0.00003039, 0.00001605, -0.00001292, -0.00001440, 0.00000855, 0.00001756, -0.00000295, -0.00002439, -0.00001651, 0.00001015, 0.00002083, 0.00000492, -0.00001301, -0.00001057, 0.00000332, 0.00000619, -0.00000445, -0.00001110, -0.00000504, 0.00000192, -0.00000171, -0.00000894, -0.00000696, 0.00000100, 0.00000062, -0.00000922, -0.00001218, 0.00000046, 0.00001338, 0.00000726, -0.00001182, -0.00001839, -0.00000369, 0.00000971, -0.00000140, -0.00002629, -0.00003263, -0.00001247, 0.00000143, -0.00002419, -0.00007514, -0.00009697, -0.00005312, 0.00003342, 0.00010207, 0.00011339, 0.00008077, 0.00004814, 0.00004519, 0.00006587, },
		{ 0.00003788, 0.00004561, 0.00006495, 0.00007319, 0.00005859, 0.00002674, -0.00000678, -0.00002920, -0.00003785, -0.00003707, -0.00003071, -0.00001996, -0.00000751, 0.00000058, 0.00000009, -0.00000580, -0.00000979, -0.00000905, -0.00000748, -0.00000876, -0.00001047, -0.00000801, -0.00000200, 0.00000222, 0.00000227, 0.00000194, 0.00000482, 0.00000819, 0.00000619, -0.00000197, -0.00001055, -0.00001355, -0.00000976, -0.00000157, 0.00000743, 0.00001253, 0.00000950, 0.00000027, -0.00000574, -0.00000122, 0.00000820, 0.00000871, -0.00000375, -0.00001607, -0.00001373, -0.00000013, 0.00000767, 0.00000209, -0.00000576, -0.00000384, 0.00000258, 0.00000052, -0.00000783, -0.00000604, 0.00000977, 0.00001938, 0.00000514, -0.00001832, -0.00001891, 0.00000765, 0.00002638, 0.00000858, -0.00002488, -0.00002693, 0.00001118, 0.00004256, 0.00002404, -0.00002454, -0.00004354, -0.00001153, 0.00002674, 0.00002097, -0.00001715, -0.00003146, 0.00000080, 0.00003760, 0.00002918, -0.00001521, -0.00004077, -0.00001960, 0.00001675, 0.00002393, 0.00000208, -0.00001243, -0.00000168, 0.00001240, 0.00000487, -0.00001497, -0.00001887, -0.00000114, 0.00001452, 0.00000968, -0.00000491, -0.00000840, 0.00000002, 0.00000381, -0.00000354, -0.00000982, -0.00000512, 0.00000220, -0.00000049, -0.00000924, -0.00000946, 0.00000037, 0.00000487, -0.00000460, -0.00001472, -0.00000863, 0.00000804, 0.00001295, -0.00000134, -0.00001570, -0.00001111, 0.00000359, 0.00000374, -0.00001522, -0.00002938, -0.00001988, -0.00000284, -0.00001178, -0.00005200, -0.00008565, -0.00006952, -0.00000256, 0.00007116, 0.00010503, 0.00009040, 0.00005651, },
		{ 0.00009212, 0.00006670, 0.00004035, 0.00003398, 0.00004676, 0.00006201, 0.00006276, 0.00004436, 0.00001589, -0.00000950, -0.00002530, -0.00003298, -0.00003495, -0.00003001, -0.00001769, -0.00000387, 0.00000260, -0.00000069, -0.00000720, -0.00000964, -0.00000803, -0.00000720, -0.00000865, -0.00000887, -0.00000554, -0.00000117, 0.00000115, 0.00000243, 0.00000523, 0.00000820, 0.00000700, 0.00000045, -0.00000731, -0.00001163, -0.00001135, -0.00000706, 0.00000099, 0.00001001, 0.00001316, 0.00000640, -0.00000350, -0.00000439, 0.00000532, 0.00001192, 0.00000326, -0.00001342, -0.00001893, -0.00000713, 0.00000648, 0.00000595, -0.00000375, -0.00000581, 0.00000217, 0.00000490, -0.00000503, -0.00001218, -0.00000027, 0.00001920, 0.00001800, -0.00000792, -0.00002661, -0.00001102, 0.00002067, 0.00002551, -0.00000534, -0.00003051, -0.00001260, 0.00002847, 0.00003844, 0.00000086, -0.00003895, -0.00003266, 0.00000787, 0.00002813, 0.00000472, -0.00002534, -0.00001772, 0.00001827, 0.00003372, 0.00000715, -0.00002796, -0.00002950, -0.00000059, 0.00002026, 0.00001266, -0.00000407, -0.00000483, 0.00000595, 0.00000627, -0.00000766, -0.00001632, -0.00000707, 0.00000752, 0.00000910, -0.00000058, -0.00000522, 0.00000023, 0.00000368, -0.00000301, -0.00001048, -0.00000695, 0.00000238, 0.00000265, -0.00000735, -0.00001235, -0.00000373, 0.00000632, 0.00000213, -0.00001141, -0.00001454, -0.00000111, 0.00001157, 0.00000674, -0.00000824, -0.00001275, -0.00000288, 0.00000322, -0.00000761, -0.00002306, -0.00002242, -0.00000815, -0.00000675, -0.00003414, -0.00006962, -0.00007339, -0.00002879, 0.00003900, 0.00008689, },
		{ 0.00001079, 0.00006347, 0.00008520, 0.00007339, 0.00004758, 0.00003083, 0.00003324, 0.00004764, 0.00005766, 0.00005163, 0.00003136, 0.00000814, -0.00000934, -0.00002112, -0.00003020, -0.00003445, -0.00002870, -0.00001407, -0.00000040, 0.00000324, -0.00000203, -0.00000782, -0.00000891, -0.00000745, -0.00000711, -0.00000763, -0.00000670, -0.00000431, -0.00000190, 0.00000090, 0.00000525, 0.00000919, 0.00000843, 0.00000209, -0.00000528, -0.00000929, -0.00001017, -0.00000949, -0.00000538, 0.00000384, 0.00001264, 0.00001187, 0.00000167, -0.00000514, 0.00000116, 0.00001211, 0.00001012, -0.00000712, -0.00002085, -0.00001497, 0.00000202, 0.00000853, -0.00000035, -0.00000747, -0.00000064, 0.00000824, 0.00000189, -0.00001290, -0.00001167, 0.00001033, 0.00002521, 0.00000872, -0.00002211, -0.00002702, 0.00000284, 0.00002924, 0.00001665, -0.00001806, -0.00002635, 0.00000574, 0.00003632, 0.00002275, -0.00002055, -0.00004041, -0.00001491, 0.00001967, 0.00001954, -0.00000896, -0.00002219, -0.00000097, 0.00002477, 0.00001932, -0.00001054, -0.00002721, -0.00001316, 0.00001001, 0.00001517, 0.00000422, -0.00000212, 0.00000267, 0.00000492, -0.00000402, -0.00001312, -0.00000909, 0.00000282, 0.00000691, 0.00000041, -0.00000399, 0.00000092, 0.00000557, -0.00000043, -0.00001061, -0.00001047, 0.00000010, 0.00000547, -0.00000275, -0.00001263, -0.00000912, 0.00000340, 0.00000681, -0.00000438, -0.00001455, -0.00000903, 0.00000506, 0.00000912, -0.00000074, -0.00000951, -0.00000614, 0.00000051, -0.00000407, -0.00001713, -0.00002164, -0.00001203, -0.00000578, -0.00002200, -0.00005349, -0.00006890, -0.00004450, },
		{ -0.00005989, -0.00005118, -0.00001086, 0.00003911, 0.00007155, 0.00007428, 0.00005523, 0.00003356, 0.00002565, 0.00003438, 0.00004820, 0.00005171, 0.00003965, 0.00002015, 0.00000396, -0.00000706, -0.00001801, -0.00002961, -0.00003432, -0.00002604, -0.00000983, 0.00000202, 0.00000286, -0.00000303, -0.00000772, -0.00000839, -0.00000724, -0.00000632, -0.00000570, -0.00000531, -0.00000492, -0.00000261, 0.00000325, 0.00000982, 0.00001089, 0.00000438, -0.00000411, -0.00000808, -0.00000804, -0.00000857, -0.00000901, -0.00000353, 0.00000764, 0.00001402, 0.00000786, -0.00000270, -0.00000233, 0.00000951, 0.00001497, 0.00000150, -0.00001842, -0.00002164, -0.00000536, 0.00000867, 0.00000417, -0.00000748, -0.00000555, 0.00000784, 0.00001004, -0.00000650, -0.00001866, -0.00000432, 0.00002186, 0.00002364, -0.00000639, -0.00003173, -0.00001852, 0.00001747, 0.00002975, 0.00000344, -0.00002407, -0.00001392, 0.00002073, 0.00003192, 0.00000190, -0.00003293, -0.00003071, 0.00000168, 0.00002119, 0.00000727, -0.00001438, -0.00001149, 0.00001056, 0.00001959, 0.00000275, -0.00001776, -0.00001721, 0.00000002, 0.00001137, 0.00000819, 0.00000226, 0.00000317, 0.00000444, -0.00000238, -0.00001123, -0.00000998, 0.00000020, 0.00000561, 0.00000049, -0.00000474, -0.00000028, 0.00000710, 0.00000407, -0.00000799, -0.00001338, -0.00000475, 0.00000545, 0.00000250, -0.00000925, -0.00001258, -0.00000236, 0.00000699, 0.00000229, -0.00000969, -0.00001225, -0.00000231, 0.00000628, 0.00000320, -0.00000476, -0.00000580, -0.00000131, -0.00000280, -0.00001287, -0.00001965, -0.00001431, -0.00000660, -0.00001445, -0.00003947, },
		{ -0.00001041, -0.00002846, -0.00004913, -0.00005097, -0.00002512, 0.00001695, 0.00005374, 0.00006905, 0.00006057, 0.00003954, 0.00002344, 0.00002388, 0.00003717, 0.00004771, 0.00004342, 0.00002736, 0.00001194, 0.00000330, -0.00000421, -0.00001693, -0.00003037, -0.00003335, -0.00002221, -0.00000602, 0.00000328, 0.00000232, -0.00000340, -0.00000763, -0.00000834, -0.00000658, -0.00000439, -0.00000403, -0.00000587, -0.00000636, -0.00000109, 0.00000830, 0.00001332, 0.00000812, -0.00000238, -0.00000816, -0.00000696, -0.00000584, -0.00000874, -0.00000904, -0.00000012, 0.00001149, 0.00001242, 0.00000215, -0.00000365, 0.00000529, 0.00001667, 0.00001054, -0.00001154, -0.00002508, -0.00001447, 0.00000521, 0.00000866, -0.00000453, -0.00001050, 0.00000248, 0.00001501, 0.00000495, -0.00001683, -0.00001807, 0.00000826, 0.00002893, 0.00001340, -0.00002211, -0.00003268, -0.00000406, 0.00002731, 0.00002251, -0.00000832, -0.00002115, 0.00000127, 0.00002657, 0.00001789, -0.00001608, -0.00003372, -0.00001601, 0.00001125, 0.00001517, -0.00000183, -0.00001121, -0.00000038, 0.00001254, 0.00000785, -0.00000818, -0.00001461, -0.00000536, 0.00000587, 0.00000770, 0.00000472, 0.00000512, 0.00000592, -0.00000030, -0.00000998, -0.00001130, -0.00000212, 0.00000527, 0.00000162, -0.00000555, -0.00000341, 0.00000611, 0.00000825, -0.00000256, -0.00001325, -0.00001003, 0.00000191, 0.00000559, -0.00000371, -0.00001223, -0.00000777, 0.00000321, 0.00000548, -0.00000337, -0.00001053, -0.00000683, 0.00000136, 0.00000302, -0.00000167, -0.00000391, -0.00000136, -0.00000177, -0.00000964, -0.00001738, -0.00001555, -0.00000825, },
		{ -0.00001577, -0.00001014, -0.00000910, -0.00002092, -0.00003877, -0.00004605, -0.00003210, -0.00000070, 0.00003435, 0.00005820, 0.00006181, 0.00004665, 0.00002598, 0.00001691, 0.00002564, 0.00004071, 0.00004467, 0.00003273, 0.00001634, 0.00000765, 0.00000495, -0.00000244, -0.00001777, -0.00003111, -0.00003110, -0.00001814, -0.00000318, 0.00000401, 0.00000220, -0.00000366, -0.00000807, -0.00000809, -0.00000463, -0.00000210, -0.00000418, -0.00000813, -0.00000632, 0.00000382, 0.00001378, 0.00001264, 0.00000119, -0.00000829, -0.00000781, -0.00000361, -0.00000541, -0.00001056, -0.00000761, 0.00000490, 0.00001322, 0.00000744, -0.00000231, 0.00000100, 0.00001502, 0.00001792, -0.00000120, -0.00002354, -0.00002318, -0.00000231, 0.00001100, 0.00000146, -0.00001246, -0.00000625, 0.00001324, 0.00001615, -0.00000596, -0.00002384, -0.00000970, 0.00002124, 0.00002753, -0.00000256, -0.00003257, -0.00002460, 0.00001085, 0.00002937, 0.00001150, -0.00001391, -0.00001155, 0.00001255, 0.00002186, 0.00000068, -0.00002550, -0.00002566, -0.00000321, 0.00001269, 0.00000740, -0.00000425, -0.00000385, 0.00000506, 0.00000637, -0.00000305, -0.00001016, -0.00000609, 0.00000241, 0.00000552, 0.00000448, 0.00000588, 0.00000801, 0.00000315, -0.00000768, -0.00001261, -0.00000532, 0.00000445, 0.00000384, -0.00000460, -0.00000666, 0.00000233, 0.00000974, 0.00000364, -0.00000945, -0.00001304, -0.00000356, 0.00000505, 0.00000122, -0.00000864, -0.00001032, -0.00000192, 0.00000460, 0.00000110, -0.00000625, -0.00000735, -0.00000236, 0.00000064, -0.00000124, -0.00000274, -0.00000073, -0.00000034, -0.00000653, -0.00001466, },
		{ -0.00000353, -0.00001129, -0.00001457, -0.00001131, -0.00000944, -0.00001694, -0.00003074, -0.00003902, -0.00003273, -0.00001206, 0.00001626, 0.00004295, 0.00005748, 0.00005253, 0.00003254, 0.00001472, 0.00001484, 0.00003048, 0.00004294, 0.00003797, 0.00002100, 0.00000844, 0.00000693, 0.00000701, -0.00000250, -0.00001942, -0.00003087, -0.00002811, -0.00001466, -0.00000101, 0.00000496, 0.00000223, -0.00000455, -0.00000869, -0.00000656, -0.00000163, -0.00000120, -0.00000679, -0.00001004, -0.00000264, 0.00001087, 0.00001610, 0.00000663, -0.00000680, -0.00000991, -0.00000363, -0.00000127, -0.00000784, -0.00001198, -0.00000345, 0.00000960, 0.00001098, 0.00000105, -0.00000202, 0.00001058, 0.00002182, 0.00001060, -0.00001619, -0.00002863, -0.00001295, 0.00000883, 0.00000853, -0.00000925, -0.00001436, 0.00000441, 0.00002118, 0.00000933, -0.00001814, -0.00002337, 0.00000386, 0.00002895, 0.00001731, -0.00001856, -0.00003421, -0.00001099, 0.00002110, 0.00002453, 0.00000201, -0.00001213, -0.00000067, 0.00001559, 0.00001001, -0.00001285, -0.00002535, -0.00001455, 0.00000367, 0.00000924, 0.00000288, -0.00000117, 0.00000179, 0.00000306, -0.00000252, -0.00000783, -0.00000519, 0.00000171, 0.00000449, 0.00000326, 0.00000468, 0.00000872, 0.00000691, -0.00000359, -0.00001242, -0.00000895, 0.00000197, 0.00000572, -0.00000167, -0.00000796, -0.00000256, 0.00000771, 0.00000796, -0.00000346, -0.00001236, -0.00000832, 0.00000156, 0.00000346, -0.00000404, -0.00000954, -0.00000573, 0.00000134, 0.00000233, -0.00000237, -0.00000522, -0.00000338, -0.00000137, -0.00000233, -0.00000311, -0.00000084, 0.00000075, },
		{ -0.00000204, 0.00000066, -0.00000149, -0.00000792, -0.00001203, -0.00001076, -0.00000969, -0.00001550, -0.00002623, -0.00003284, -0.00002927, -0.00001647, 0.00000262, 0.00002585, 0.00004699, 0.00005424, 0.00004115, 0.00001841, 0.00000735, 0.00001777, 0.00003636, 0.00004162, 0.00002809, 0.00001055, 0.00000443, 0.00000818, 0.00000799, -0.00000388, -0.00002065, -0.00002972, -0.00002528, -0.00001178, 0.00000130, 0.00000617, 0.00000158, -0.00000611, -0.00000828, -0.00000332, 0.00000082, -0.00000316, -0.00001050, -0.00000879, 0.00000475, 0.00001654, 0.00001257, -0.00000273, -0.00001154, -0.00000620, 0.00000124, -0.00000243, -0.00001170, -0.00001057, 0.00000258, 0.00001118, 0.00000505, -0.00000294, 0.00000465, 0.00002110, 0.00002098, -0.00000380, -0.00002798, -0.00002389, 0.00000097, 0.00001313, -0.00000122, -0.00001741, -0.00000795, 0.00001683, 0.00002172, -0.00000322, -0.00002595, -0.00001482, 0.00001716, 0.00002791, 0.00000192, -0.00002902, -0.00002737, 0.00000279, 0.00002445, 0.00001676, -0.00000232, -0.00000575, 0.00000601, 0.00001037, -0.00000298, -0.00001879, -0.00001892, -0.00000572, 0.00000486, 0.00000562, 0.00000309, 0.00000304, 0.00000203, -0.00000350, -0.00000838, -0.00000573, 0.00000177, 0.00000532, 0.00000327, 0.00000285, 0.00000734, 0.00000919, 0.00000135, -0.00000988, -0.00001150, -0.00000196, 0.00000578, 0.00000192, -0.00000660, -0.00000629, 0.00000332, 0.00000891, 0.00000217, -0.00000863, -0.00001039, -0.00000278, 0.00000271, -0.00000065, -0.00000678, -0.00000699, -0.00000198, 0.00000105, -0.00000058, -0.00000270, -0.00000228, -0.00000166, -0.00000315, -0.00000425, },
		{ -0.00000270, -0.00000486, -0.00000353, -0.00000050, -0.00000101, -0.00000575, -0.00000930, -0.00000850, -0.00000817, -0.00001443, -0.00002474, -0.00002983, -0.00002523, -0.00001528, -0.00000420, 0.00001081, 0.00003180, 0.00004926, 0.00004813, 0.00002727, 0.00000617, 0.00000569, 0.00002441, 0.00004025, 0.00003592, 0.00001719, 0.00000323, 0.00000349, 0.00000949, 0.00000764, -0.00000538, -0.00002106, -0.00002847, -0.00002300, -0.00000879, 0.00000409, 0.00000682, -0.00000026, -0.00000727, -0.00000578, 0.00000053, 0.00000062, -0.00000768, -0.00001232, -0.00000280, 0.00001313, 0.00001693, 0.00000348, -0.00001100, -0.00001019, 0.00000060, 0.00000309, -0.00000706, -0.00001397, -0.00000558, 0.00000763, 0.00000803, -0.00000162, -0.00000105, 0.00001581, 0.00002694, 0.00001095, -0.00001979, -0.00003101, -0.00001124, 0.00001180, 0.00000826, -0.00001317, -0.00001804, 0.00000455, 0.00002506, 0.00001392, -0.00001619, -0.00002577, -0.00000162, 0.00002476, 0.00001860, -0.00001292, -0.00003115, -0.00001581, 0.00001240, 0.00002227, 0.00001046, -0.00000133, 0.00000041, 0.00000580, 0.00000072, -0.00001179, -0.00001747, -0.00001118, -0.00000144, 0.00000353, 0.00000485, 0.00000583, 0.00000419, -0.00000270, -0.00000965, -0.00000847, 0.00000011, 0.00000641, 0.00000515, 0.00000238, 0.00000502, 0.00000913, 0.00000538, -0.00000558, -0.00001176, -0.00000592, 0.00000372, 0.00000445, -0.00000341, -0.00000751, -0.00000128, 0.00000683, 0.00000549, -0.00000381, -0.00000949, -0.00000590, 0.00000024, 0.00000056, -0.00000392, -0.00000617, -0.00000385, -0.00000105, -0.00000085, -0.00000140, -0.00000067, -0.00000048, },
		{ 0.00000018, 0.00000106, -0.00000120, -0.00000435, -0.00000428, -0.00000168, -0.00000159, -0.00000540, -0.00000803, -0.00000600, -0.00000458, -0.00001136, -0.00002389, -0.00003032, -0.00002402, -0.00001205, -0.00000421, 0.00000170, 0.00001580, 0.00003734, 0.00004931, 0.00003783, 0.00001244, -0.00000126, 0.00000982, 0.00003162, 0.00003989, 0.00002713, 0.00000779, -0.00000073, 0.00000378, 0.00000994, 0.00000688, -0.00000623, -0.00002123, -0.00002778, -0.00002070, -0.00000499, 0.00000677, 0.00000596, -0.00000264, -0.00000649, -0.00000147, 0.00000265, -0.00000319, -0.00001225, -0.00000940, 0.00000662, 0.00001803, 0.00001030, -0.00000748, -0.00001372, -0.00000327, 0.00000624, 0.00000008, -0.00001246, -0.00001221, 0.00000114, 0.00000862, 0.00000131, -0.00000472, 0.00000746, 0.00002645, 0.00002391, -0.00000520, -0.00003054, -0.00002380, 0.00000335, 0.00001430, -0.00000321, -0.00002093, -0.00001009, 0.00001762, 0.00002515, 0.00000102, -0.00002409, -0.00001794, 0.00001063, 0.00002401, 0.00000510, -0.00002232, -0.00002596, -0.00000434, 0.00001649, 0.00001814, 0.00000801, 0.00000202, 0.00000234, -0.00000011, -0.00000821, -0.00001427, -0.00001229, -0.00000577, -0.00000058, 0.00000303, 0.00000675, 0.00000743, 0.00000089, -0.00000886, -0.00001163, -0.00000398, 0.00000551, 0.00000751, 0.00000404, 0.00000367, 0.00000752, 0.00000731, -0.00000117, -0.00000979, -0.00000846, 0.00000048, 0.00000504, -0.00000002, -0.00000636, -0.00000437, 0.00000331, 0.00000603, 0.00000012, -0.00000678, -0.00000692, -0.00000226, 0.00000001, -0.00000216, -0.00000453, -0.00000408, -0.00000257, -0.00000212, -0.00000160, },
		{ -0.00000332, -0.00000272, -0.00000008, 0.00000206, 0.00000051, -0.00000309, -0.00000402, -0.00000191, -0.00000184, -0.00000606, -0.00000902, -0.00000545, -0.00000059, -0.00000543, -0.00002063, -0.00003201, -0.00002701, -0.00001108, -0.00000045, 0.00000019, 0.00000393, 0.00002153, 0.00004245, 0.00004485, 0.00002374, 0.00000008, -0.00000209, 0.00001733, 0.00003605, 0.00003524, 0.00001786, 0.00000121, -0.00000267, 0.00000384, 0.00000995, 0.00000672, -0.00000669, -0.00002207, -0.00002740, -0.00001740, -0.00000058, 0.00000810, 0.00000359, -0.00000385, -0.00000304, 0.00000265, 0.00000098, -0.00000918, -0.00001325, -0.00000118, 0.00001536, 0.00001582, -0.00000134, -0.00001499, -0.00000916, 0.00000546, 0.00000703, -0.00000644, -0.00001502, -0.00000645, 0.00000604, 0.00000450, -0.00000521, -0.00000120, 0.00001952, 0.00003089, 0.00001188, -0.00002106, -0.00003143, -0.00000987, 0.00001304, 0.00000761, -0.00001535, -0.00002038, 0.00000306, 0.00002551, 0.00001729, -0.00001140, -0.00002421, -0.00000645, 0.00001717, 0.00001621, -0.00000726, -0.00002446, -0.00001714, 0.00000372, 0.00001669, 0.00001534, 0.00000865, 0.00000379, -0.00000104, -0.00000789, -0.00001278, -0.00001172, -0.00000712, -0.00000352, -0.00000049, 0.00000459, 0.00000883, 0.00000557, -0.00000495, -0.00001256, -0.00000893, 0.00000168, 0.00000810, 0.00000674, 0.00000438, 0.00000612, 0.00000741, 0.00000199, -0.00000676, -0.00000909, -0.00000256, 0.00000396, 0.00000233, -0.00000396, -0.00000538, 0.00000006, 0.00000464, 0.00000220, -0.00000383, -0.00000623, -0.00000376, -0.00000123, -0.00000166, -0.00000314, -0.00000333, -0.00000303, },
		{ -0.00000231, -0.00000259, -0.00000391, -0.00000409, -0.00000116, 0.00000229, 0.00000187, -0.00000183, -0.00000344, -0.00000123, -0.00000067, -0.00000589, -0.00001128, -0.00000809, 0.00000098, 0.00000168, -0.00001330, -0.00003121, -0.00003240, -0.00001498, 0.00000227, 0.00000434, -0.00000045, 0.00000715, 0.00002903, 0.00004425, 0.00003471, 0.00000900, -0.00000645, 0.00000270, 0.00002443, 0.00003614, 0.00002836, 0.00001027, -0.00000311, -0.00000429, 0.00000358, 0.00001066, 0.00000729, -0.00000781, -0.00002379, -0.00002626, -0.00001274, 0.00000321, 0.00000729, 0.00000111, -0.00000230, 0.00000192, 0.00000379, -0.00000460, -0.00001381, -0.00000829, 0.00000955, 0.00001854, 0.00000616, -0.00001290, -0.00001506, 0.00000052, 0.00001110, 0.00000219, -0.00001276, -0.00001269, 0.00000053, 0.00000619, -0.00000266, -0.00000718, 0.00000857, 0.00002955, 0.00002600, -0.00000475, -0.00003010, -0.00002261, 0.00000412, 0.00001373, -0.00000441, -0.00002209, -0.00001178, 0.00001555, 0.00002530, 0.00000543, -0.00001854, -0.00001782, 0.00000344, 0.00001641, 0.00000553, -0.00001463, -0.00002086, -0.00000864, 0.00000786, 0.00001592, 0.00001495, 0.00000954, 0.00000154, -0.00000769, -0.00001349, -0.00001232, -0.00000726, -0.00000409, -0.00000296, 0.00000075, 0.00000703, 0.00000856, 0.00000069, -0.00000993, -0.00001208, -0.00000382, 0.00000563, 0.00000836, 0.00000654, 0.00000624, 0.00000706, 0.00000366, -0.00000399, -0.00000838, -0.00000461, 0.00000216, 0.00000335, -0.00000149, -0.00000480, -0.00000200, 0.00000260, 0.00000261, -0.00000167, -0.00000478, -0.00000412, -0.00000229, -0.00000196, -0.00000240, },
		{ -0.00000255, -0.00000231, -0.00000144, -0.00000156, -0.00000375, -0.00000527, -0.00000275, 0.00000188, 0.00000291, -0.00000078, -0.00000332, -0.00000072, 0.00000170, -0.00000337, -0.00001236, -0.00001296, -0.00000187, 0.00000654, -0.00000316, -0.00002522, -0.00003614, -0.00002295, 0.00000008, 0.00000944, 0.00000235, -0.00000106, 0.00001393, 0.00003561, 0.00003993, 0.00002082, -0.00000206, -0.00000626, 0.00000967, 0.00002820, 0.00003289, 0.00002184, 0.00000473, -0.00000684, -0.00000626, 0.00000408, 0.00001268, 0.00000758, -0.00001037, -0.00002549, -0.00002334, -0.00000754, 0.00000495, 0.00000500, 0.00000061, 0.00000210, 0.00000561, 0.00000030, -0.00001147, -0.00001332, 0.00000186, 0.00001764, 0.00001334, -0.00000727, -0.00001876, -0.00000739, 0.00001036, 0.00001052, -0.00000572, -0.00001517, -0.00000643, 0.00000490, 0.00000132, -0.00000861, -0.00000236, 0.00002061, 0.00003241, 0.00001311, -0.00001926, -0.00002911, -0.00000880, 0.00001196, 0.00000610, -0.00001555, -0.00002056, 0.00000066, 0.00002240, 0.00001827, -0.00000512, -0.00001881, -0.00000908, 0.00000802, 0.00001031, -0.00000349, -0.00001617, -0.00001484, -0.00000268, 0.00000976, 0.00001623, 0.00001557, 0.00000782, -0.00000443, -0.00001406, -0.00001480, -0.00000867, -0.00000362, -0.00000308, -0.00000223, 0.00000307, 0.00000819, 0.00000532, -0.00000467, -0.00001165, -0.00000847, 0.00000081, 0.00000729, 0.00000834, 0.00000780, 0.00000757, 0.00000469, -0.00000199, -0.00000728, -0.00000571, 0.00000040, 0.00000341, 0.00000042, -0.00000346, -0.00000279, 0.00000086, 0.00000208, -0.00000053, -0.00000338, -0.00000374, -0.00000285, },
		{ -0.00000304, -0.00000292, -0.00000317, -0.00000278, -0.00000099, -0.00000019, -0.00000273, -0.00000597, -0.00000472, 0.00000079, 0.00000381, 0.00000046, -0.00000369, -0.00000154, 0.00000364, 0.00000129, -0.00001005, -0.00001707, -0.00000862, 0.00000612, 0.00000600, -0.00001439, -0.00003442, -0.00003106, -0.00000757, 0.00001068, 0.00000836, -0.00000168, 0.00000262, 0.00002262, 0.00003715, 0.00002956, 0.00000766, -0.00000664, -0.00000168, 0.00001501, 0.00002826, 0.00002884, 0.00001697, -0.00000012, -0.00001119, -0.00000773, 0.00000648, 0.00001533, 0.00000629, -0.00001401, -0.00002573, -0.00001875, -0.00000350, 0.00000429, 0.00000326, 0.00000344, 0.00000729, 0.00000518, -0.00000683, -0.00001541, -0.00000621, 0.00001304, 0.00001837, 0.00000105, -0.00001843, -0.00001591, 0.00000433, 0.00001542, 0.00000416, -0.00001236, -0.00001232, 0.00000035, 0.00000422, -0.00000574, -0.00000942, 0.00000778, 0.00002930, 0.00002631, -0.00000253, -0.00002609, -0.00001979, 0.00000332, 0.00001105, -0.00000517, -0.00002072, -0.00001226, 0.00001138, 0.00002199, 0.00000865, -0.00001067, -0.00001424, -0.00000250, 0.00000683, 0.00000279, -0.00000818, -0.00001371, -0.00000936, 0.00000093, 0.00001152, 0.00001759, 0.00001472, 0.00000250, -0.00001161, -0.00001729, -0.00001224, -0.00000443, -0.00000180, -0.00000274, -0.00000062, 0.00000500, 0.00000687, 0.00000069, -0.00000796, -0.00001011, -0.00000419, 0.00000363, 0.00000807, 0.00000941, 0.00000919, 0.00000618, -0.00000033, -0.00000631, -0.00000639, -0.00000113, 0.00000300, 0.00000174, -0.00000192, -0.00000267, -0.00000024, 0.00000128, -0.00000017, -0.00000237, },
		{ -0.00000027, -0.00000181, -0.00000227, -0.00000252, -0.00000361, -0.00000375, -0.00000128, 0.00000113, -0.00000081, -0.00000573, -0.00000680, -0.00000124, 0.00000428, 0.00000239, -0.00000365, -0.00000377, 0.00000336, 0.00000607, -0.00000405, -0.00001727, -0.00001612, -0.00000004, 0.00001011, -0.00000231, -0.00002632, -0.00003473, -0.00001750, 0.00000609, 0.00001246, 0.00000260, -0.00000204, 0.00001073, 0.00002857, 0.00003160, 0.00001683, -0.00000043, -0.00000536, 0.00000324, 0.00001690, 0.00002665, 0.00002607, 0.00001295, -0.00000599, -0.00001583, -0.00000713, 0.00001075, 0.00001701, 0.00000296, -0.00001724, -0.00002371, -0.00001394, -0.00000196, 0.00000257, 0.00000430, 0.00000884, 0.00000998, -0.00000026, -0.00001398, -0.00001303, 0.00000531, 0.00001962, 0.00001023, -0.00001323, -0.00002206, -0.00000565, 0.00001449, 0.00001351, -0.00000445, -0.00001437, -0.00000601, 0.00000392, -0.00000092, -0.00001071, -0.00000389, 0.00001874, 0.00003055, 0.00001387, -0.00001457, -0.00002370, -0.00000754, 0.00000877, 0.00000342, -0.00001427, -0.00001848, -0.00000173, 0.00001660, 0.00001644, 0.00000097, -0.00001085, -0.00000881, -0.00000048, 0.00000226, -0.00000261, -0.00000861, -0.00001003, -0.00000555, 0.00000370, 0.00001404, 0.00001828, 0.00001067, -0.00000512, -0.00001699, -0.00001650, -0.00000777, -0.00000136, -0.00000128, -0.00000209, 0.00000114, 0.00000519, 0.00000376, -0.00000305, -0.00000836, -0.00000710, -0.00000097, 0.00000536, 0.00000954, 0.00001100, 0.00000849, 0.00000177, -0.00000526, -0.00000698, -0.00000263, 0.00000226, 0.00000257, -0.00000044, -0.00000199, -0.00000066, 0.00000061, },
		{ -0.00000047, 0.00000029, -0.00000058, -0.00000173, -0.00000161, -0.00000164, -0.00000350, -0.00000497, -0.00000261, 0.00000172, 0.00000174, -0.00000402, -0.00000828, -0.00000429, 0.00000356, 0.00000475, -0.00000212, -0.00000611, 0.00000030, 0.00000831, 0.00000354, -0.00001228, -0.00002029, -0.00000913, 0.00000742, 0.00000625, -0.00001464, -0.00003173, -0.00002493, -0.00000228, 0.00001161, 0.00000699, -0.00000135, 0.00000366, 0.00001906, 0.00002774, 0.00002100, 0.00000693, -0.00000207, -0.00000180, 0.00000538, 0.00001641, 0.00002598, 0.00002463, 0.00000780, -0.00001331, -0.00001901, -0.00000357, 0.00001546, 0.00001637, -0.00000145, -0.00001855, -0.00002010, -0.00001079, -0.00000266, 0.00000227, 0.00000886, 0.00001397, 0.00000768, -0.00000858, -0.00001673, -0.00000406, 0.00001606, 0.00001763, -0.00000389, -0.00002315, -0.00001653, 0.00000720, 0.00001865, 0.00000624, -0.00001087, -0.00001128, 0.00000005, 0.00000263, -0.00000724, -0.00001047, 0.00000558, 0.00002546, 0.00002405, 0.00000062, -0.00001892, -0.00001502, 0.00000166, 0.00000660, -0.00000597, -0.00001749, -0.00001129, 0.00000631, 0.00001602, 0.00000981, -0.00000244, -0.00000814, -0.00000590, -0.00000246, -0.00000228, -0.00000447, -0.00000671, -0.00000705, -0.00000280, 0.00000691, 0.00001648, 0.00001635, 0.00000358, -0.00001243, -0.00001876, -0.00001271, -0.00000340, 0.00000019, -0.00000118, -0.00000121, 0.00000193, 0.00000372, 0.00000067, -0.00000461, -0.00000698, -0.00000444, 0.00000128, 0.00000751, 0.00001170, 0.00001105, 0.00000465, -0.00000360, -0.00000731, -0.00000421, 0.00000112, 0.00000291, 0.00000084, -0.00000098, },
		{ 0.00000171, 0.00000010, 0.00000024, 0.00000054, -0.00000079, -0.00000212, -0.00000140, -0.00000047, -0.00000248, -0.00000576, -0.00000486, 0.00000081, 0.00000403, -0.00000076, -0.00000811, -0.00000766, 0.00000101, 0.00000642, 0.00000109, -0.00000677, -0.00000424, 0.00000658, 0.00000937, -0.00000372, -0.00001865, -0.00001664, -0.00000021, 0.00000831, -0.00000420, -0.00002369, -0.00002668, -0.00001003, 0.00000672, 0.00000797, 0.00000058, 0.00000119, 0.00001266, 0.00002216, 0.00001992, 0.00001050, 0.00000324, 0.00000037, -0.00000003, 0.00000445, 0.00001633, 0.00002738, 0.00002260, 0.00000012, -0.00002068, -0.00001895, 0.00000223, 0.00001863, 0.00001342, -0.00000511, -0.00001753, -0.00001678, -0.00001026, -0.00000363, 0.00000536, 0.00001530, 0.00001546, 0.00000052, -0.00001549, -0.00001252, 0.00000797, 0.00002056, 0.00000710, -0.00001799, -0.00002440, -0.00000457, 0.00001697, 0.00001578, -0.00000229, -0.00001245, -0.00000552, 0.00000268, -0.00000217, -0.00001089, -0.00000496, 0.00001448, 0.00002525, 0.00001332, -0.00000810, -0.00001568, -0.00000549, 0.00000447, -0.00000020, -0.00001232, -0.00001484, -0.00000349, 0.00000956, 0.00001229, 0.00000548, -0.00000227, -0.00000598, -0.00000658, -0.00000588, -0.00000442, -0.00000355, -0.00000460, -0.00000504, 0.00000012, 0.00001050, 0.00001708, 0.00001106, -0.00000466, -0.00001707, -0.00001685, -0.00000771, -0.00000022, 0.00000064, -0.00000123, -0.00000080, 0.00000149, 0.00000187, -0.00000098, -0.00000452, -0.00000549, -0.00000245, 0.00000389, 0.00001053, 0.00001282, 0.00000795, -0.00000105, -0.00000697, -0.00000567, -0.00000039, 0.00000266, },
		{ -0.00000202, 0.00000189, 0.00000200, 0.00000087, 0.00000118, 0.00000145, -0.00000048, -0.00000274, -0.00000197, 0.00000043, -0.00000055, -0.00000531, -0.00000723, -0.00000189, 0.00000476, 0.00000322, -0.00000561, -0.00000993, -0.00000310, 0.00000605, 0.00000484, -0.00000475, -0.00000791, 0.00000163, 0.00001092, 0.00000482, -0.00001183, -0.00001902, -0.00000845, 0.00000449, 0.00000120, -0.00001495, -0.00002330, -0.00001360, 0.00000156, 0.00000564, 0.00000019, -0.00000017, 0.00000975, 0.00001896, 0.00001729, 0.00000926, 0.00000515, 0.00000523, 0.00000245, -0.00000192, 0.00000290, 0.00001891, 0.00002953, 0.00001771, -0.00000959, -0.00002572, -0.00001519, 0.00000822, 0.00001912, 0.00000961, -0.00000667, -0.00001539, -0.00001554, -0.00001143, -0.00000228, 0.00001179, 0.00002031, 0.00001152, -0.00000840, -0.00001690, -0.00000246, 0.00001744, 0.00001606, -0.00000773, -0.00002606, -0.00001687, 0.00000837, 0.00002017, 0.00000843, -0.00000803, -0.00000956, -0.00000065, 0.00000112, -0.00000715, -0.00000982, 0.00000290, 0.00001881, 0.00001914, 0.00000352, -0.00000995, -0.00000857, 0.00000022, 0.00000146, -0.00000700, -0.00001353, -0.00000931, 0.00000146, 0.00000887, 0.00000887, 0.00000438, -0.00000121, -0.00000648, -0.00000948, -0.00000804, -0.00000377, -0.00000169, -0.00000328, -0.00000316, 0.00000373, 0.00001318, 0.00001451, 0.00000339, -0.00001164, -0.00001800, -0.00001247, -0.00000302, 0.00000153, 0.00000030, -0.00000172, -0.00000114, 0.00000073, 0.00000096, -0.00000133, -0.00000422, -0.00000450, -0.00000002, 0.00000769, 0.00001306, 0.00001088, 0.00000217, -0.00000573, -0.00000665, },
		{ -0.00000374, -0.00000698, -0.00000344, 0.00000092, 0.00000179, 0.00000104, 0.00000183, 0.00000281, 0.00000073, -0.00000299, -0.00000329, 0.00000032, 0.00000166, -0.00000318, -0.00000843, -0.00000572, 0.00000301, 0.00000630, -0.00000116, -0.00000973, -0.00000736, 0.00000319, 0.00000733, -0.00000067, -0.00000879, -0.00000407, 0.00000799, 0.00000996, -0.00000306, -0.00001570, -0.00001313, -0.00000150, 0.00000112, -0.00000961, -0.00001850, -0.00001286, -0.00000025, 0.00000341, -0.00000273, -0.00000401, 0.00000705, 0.00001889, 0.00001749, 0.00000666, 0.00000176, 0.00000666, 0.00000887, 0.00000070, -0.00000605, 0.00000390, 0.00002404, 0.00002968, 0.00000911, -0.00001901, -0.00002669, -0.00000914, 0.00001235, 0.00001745, 0.00000685, -0.00000625, -0.00001420, -0.00001676, -0.00001189, 0.00000281, 0.00001919, 0.00002081, 0.00000336, -0.00001460, -0.00001143, 0.00000899, 0.00001959, 0.00000415, -0.00002058, -0.00002516, -0.00000439, 0.00001715, 0.00001694, 0.00000077, -0.00000923, -0.00000494, 0.00000098, -0.00000277, -0.00000927, -0.00000496, 0.00000906, 0.00001742, 0.00001103, -0.00000151, -0.00000643, -0.00000238, 0.00000051, -0.00000393, -0.00001041, -0.00001081, -0.00000455, 0.00000276, 0.00000699, 0.00000778, 0.00000497, -0.00000182, -0.00000951, -0.00001213, -0.00000765, -0.00000160, -0.00000035, -0.00000243, -0.00000055, 0.00000736, 0.00001341, 0.00000881, -0.00000464, -0.00001561, -0.00001557, -0.00000713, 0.00000052, 0.00000188, -0.00000092, -0.00000277, -0.00000152, 0.00000083, 0.00000108, -0.00000166, -0.00000453, -0.00000304, 0.00000401, 0.00001173, 0.00001282, 0.00000550, },
		{ 0.00001362, 0.00000856, -0.00000131, -0.00000681, -0.00000461, 0.00000004, 0.00000148, 0.00000070, 0.00000173, 0.00000397, 0.00000279, -0.00000216, -0.00000472, -0.00000116, 0.00000302, 0.00000017, -0.00000742, -0.00000912, -0.00000101, 0.00000690, 0.00000368, -0.00000672, -0.00000993, -0.00000124, 0.00000726, 0.00000360, -0.00000666, -0.00000798, 0.00000258, 0.00001046, 0.00000394, -0.00000917, -0.00001279, -0.00000546, -0.00000147, -0.00000870, -0.00001634, -0.00001114, 0.00000169, 0.00000501, -0.00000473, -0.00001071, 0.00000058, 0.00001857, 0.00002184, 0.00000777, -0.00000378, 0.00000150, 0.00001238, 0.00000945, -0.00000515, -0.00000891, 0.00000891, 0.00002921, 0.00002559, -0.00000172, -0.00002543, -0.00002372, -0.00000315, 0.00001389, 0.00001536, 0.00000610, -0.00000547, -0.00001538, -0.00001888, -0.00000929, 0.00001076, 0.00002413, 0.00001600, -0.00000532, -0.00001482, -0.00000143, 0.00001637, 0.00001321, -0.00001005, -0.00002643, -0.00001661, 0.00000758, 0.00001950, 0.00001039, -0.00000390, -0.00000698, -0.00000161, -0.00000055, -0.00000602, -0.00000764, 0.00000063, 0.00001095, 0.00001235, 0.00000514, -0.00000116, -0.00000144, -0.00000018, -0.00000288, -0.00000794, -0.00000992, -0.00000721, -0.00000256, 0.00000205, 0.00000649, 0.00000870, 0.00000476, -0.00000499, -0.00001318, -0.00001265, -0.00000521, 0.00000051, 0.00000018, -0.00000119, 0.00000272, 0.00000957, 0.00001043, 0.00000133, -0.00001093, -0.00001599, -0.00001088, -0.00000207, 0.00000242, 0.00000069, -0.00000296, -0.00000366, -0.00000079, 0.00000200, 0.00000101, -0.00000300, -0.00000464, 0.00000038, 0.00000925, },
		{ -0.00000272, 0.00000605, 0.00001339, 0.00001127, 0.00000145, -0.00000631, -0.00000579, -0.00000080, 0.00000145, 0.00000033, 0.00000084, 0.00000422, 0.00000505, 0.00000005, -0.00000526, -0.00000361, 0.00000261, 0.00000344, -0.00000416, -0.00001045, -0.00000590, 0.00000448, 0.00000689, -0.00000197, -0.00000970, -0.00000538, 0.00000480, 0.00000627, -0.00000288, -0.00000896, -0.00000259, 0.00000753, 0.00000715, -0.00000301, -0.00000914, -0.00000544, -0.00000251, -0.00000964, -0.00001812, -0.00001267, 0.00000398, 0.00001114, -0.00000146, -0.00001655, -0.00001067, 0.00001307, 0.00002699, 0.00001507, -0.00000554, -0.00000798, 0.00000805, 0.00001712, 0.00000462, -0.00001210, -0.00000745, 0.00001654, 0.00003118, 0.00001722, -0.00001185, -0.00002749, -0.00001871, 0.00000093, 0.00001378, 0.00001457, 0.00000668, -0.00000631, -0.00001863, -0.00001939, -0.00000291, 0.00001875, 0.00002422, 0.00000796, -0.00001039, -0.00000900, 0.00000823, 0.00001619, 0.00000121, -0.00002069, -0.00002387, -0.00000485, 0.00001492, 0.00001656, 0.00000430, -0.00000482, -0.00000393, -0.00000097, -0.00000319, -0.00000663, -0.00000389, 0.00000389, 0.00000881, 0.00000730, 0.00000360, 0.00000212, 0.00000137, -0.00000191, -0.00000659, -0.00000875, -0.00000751, -0.00000532, -0.00000286, 0.00000187, 0.00000805, 0.00000961, 0.00000213, -0.00000964, -0.00001533, -0.00001067, -0.00000214, 0.00000167, 0.00000051, 0.00000091, 0.00000566, 0.00000917, 0.00000477, -0.00000598, -0.00001418, -0.00001318, -0.00000523, 0.00000170, 0.00000222, -0.00000198, -0.00000497, -0.00000307, 0.00000145, 0.00000302, -0.00000061, -0.00000481, },
		{ 0.00000212, -0.00000370, -0.00000505, 0.00000233, 0.00001210, 0.00001364, 0.00000477, -0.00000532, -0.00000720, -0.00000207, 0.00000163, 0.00000051, -0.00000029, 0.00000325, 0.00000656, 0.00000321, -0.00000411, -0.00000589, 0.00000035, 0.00000520, 0.00000026, -0.00000894, -0.00000965, -0.00000008, 0.00000720, 0.00000254, -0.00000701, -0.00000765, 0.00000136, 0.00000675, 0.00000070, -0.00000768, -0.00000587, 0.00000359, 0.00000711, 0.00000047, -0.00000549, -0.00000278, 0.00000012, -0.00000829, -0.00002124, -0.00001911, 0.00000169, 0.00001815, 0.00000863, -0.00001599, -0.00002296, 0.00000000, 0.00002688, 0.00002596, 0.00000077, -0.00001533, -0.00000354, 0.00001690, 0.00001649, -0.00000419, -0.00001609, -0.00000125, 0.00002342, 0.00002817, 0.00000681, -0.00001865, -0.00002588, -0.00001407, 0.00000285, 0.00001386, 0.00001550, 0.00000677, -0.00000966, -0.00002202, -0.00001640, 0.00000567, 0.00002366, 0.00001966, 0.00000056, -0.00000975, -0.00000040, 0.00001285, 0.00000879, -0.00001105, -0.00002413, -0.00001537, 0.00000523, 0.00001659, 0.00001154, 0.00000096, -0.00000347, -0.00000229, -0.00000229, -0.00000467, -0.00000478, -0.00000073, 0.00000371, 0.00000522, 0.00000514, 0.00000553, 0.00000483, 0.00000065, -0.00000514, -0.00000807, -0.00000716, -0.00000586, -0.00000573, -0.00000336, 0.00000367, 0.00001037, 0.00000848, -0.00000272, -0.00001368, -0.00001486, -0.00000722, 0.00000024, 0.00000204, 0.00000139, 0.00000345, 0.00000698, 0.00000587, -0.00000215, -0.00001141, -0.00001399, -0.00000814, 0.00000010, 0.00000326, -0.00000028, -0.00000526, -0.00000541, -0.00000030, 0.00000403, },
		{ -0.00000299, 0.00000377, 0.00000476, -0.00000135, -0.00000628, -0.00000174, 0.00000944, 0.00001529, 0.00000882, -0.00000321, -0.00000854, -0.00000424, 0.00000137, 0.00000135, -0.00000086, 0.00000147, 0.00000662, 0.00000627, -0.00000127, -0.00000681, -0.00000288, 0.00000468, 0.00000414, -0.00000513, -0.00001084, -0.00000488, 0.00000477, 0.00000512, -0.00000338, -0.00000772, -0.00000148, 0.00000583, 0.00000312, -0.00000562, -0.00000733, 0.00000040, 0.00000584, 0.00000150, -0.00000416, -0.00000078, 0.00000508, -0.00000219, -0.00002096, -0.00002781, -0.00000783, 0.00002001, 0.00002217, -0.00000578, -0.00002954, -0.00001781, 0.00001705, 0.00003374, 0.00001429, -0.00001432, -0.00001639, 0.00000696, 0.00002233, 0.00000963, -0.00001287, -0.00001467, 0.00000720, 0.00002634, 0.00002095, -0.00000245, -0.00002128, -0.00002281, -0.00001120, 0.00000391, 0.00001544, 0.00001706, 0.00000469, -0.00001464, -0.00002317, -0.00000997, 0.00001355, 0.00002403, 0.00001311, -0.00000294, -0.00000459, 0.00000632, 0.00001052, -0.00000204, -0.00001871, -0.00002049, -0.00000540, 0.00001078, 0.00001442, 0.00000747, 0.00000032, -0.00000210, -0.00000255, -0.00000368, -0.00000410, -0.00000235, -0.00000001, 0.00000139, 0.00000315, 0.00000633, 0.00000809, 0.00000475, -0.00000227, -0.00000712, -0.00000701, -0.00000549, -0.00000636, -0.00000702, -0.00000214, 0.00000699, 0.00001136, 0.00000464, -0.00000810, -0.00001542, -0.00001211, -0.00000375, 0.00000152, 0.00000232, 0.00000298, 0.00000542, 0.00000586, 0.00000039, -0.00000862, -0.00001386, -0.00001064, -0.00000212, 0.00000369, 0.00000187, -0.00000441, -0.00000732, },
		{ -0.00000226, -0.00000823, -0.00000623, 0.00000195, 0.00000668, 0.00000208, -0.00000585, -0.00000560, 0.00000520, 0.00001540, 0.00001319, 0.00000062, -0.00000890, -0.00000717, -0.00000009, 0.00000224, -0.00000045, -0.00000023, 0.00000526, 0.00000819, 0.00000246, -0.00000576, -0.00000571, 0.00000223, 0.00000617, -0.00000054, -0.00000929, -0.00000815, 0.00000102, 0.00000531, -0.00000049, -0.00000658, -0.00000312, 0.00000475, 0.00000475, -0.00000367, -0.00000808, -0.00000199, 0.00000495, 0.00000203, -0.00000492, -0.00000203, 0.00000858, 0.00000737, -0.00001382, -0.00003307, -0.00002245, 0.00001225, 0.00003209, 0.00001198, -0.00002484, -0.00003309, -0.00000176, 0.00003179, 0.00002880, -0.00000324, -0.00002288, -0.00000836, 0.00001766, 0.00002089, -0.00000056, -0.00001736, -0.00000854, 0.00001411, 0.00002421, 0.00001231, -0.00000832, -0.00002098, -0.00002043, -0.00000983, 0.00000579, 0.00001844, 0.00001741, 0.00000015, -0.00001918, -0.00002088, -0.00000206, 0.00001831, 0.00002090, 0.00000803, -0.00000179, 0.00000143, 0.00000773, 0.00000295, -0.00001131, -0.00001976, -0.00001292, 0.00000215, 0.00001191, 0.00001133, 0.00000560, 0.00000077, -0.00000200, -0.00000362, -0.00000375, -0.00000239, -0.00000135, -0.00000152, -0.00000047, 0.00000398, 0.00000896, 0.00000868, 0.00000211, -0.00000483, -0.00000664, -0.00000511, -0.00000566, -0.00000837, -0.00000710, 0.00000120, 0.00000997, 0.00000960, -0.00000089, -0.00001201, -0.00001442, -0.00000836, -0.00000124, 0.00000204, 0.00000304, 0.00000474, 0.00000579, 0.00000229, -0.00000595, -0.00001316, -0.00001288, -0.00000502, 0.00000323, 0.00000420, },
		{ 0.00000134, 0.00000603, 0.00000107, -0.00000746, -0.00000923, -0.00000144, 0.00000706, 0.00000594, -0.00000335, -0.00000821, -0.00000026, 0.00001312, 0.00001672, 0.00000606, -0.00000717, -0.00000988, -0.00000303, 0.00000218, 0.00000062, -0.00000110, 0.00000321, 0.00000855, 0.00000596, -0.00000304, -0.00000708, -0.00000104, 0.00000599, 0.00000327, -0.00000600, -0.00000915, -0.00000237, 0.00000399, 0.00000099, -0.00000555, -0.00000421, 0.00000399, 0.00000652, -0.00000145, -0.00000877, -0.00000464, 0.00000434, 0.00000375, -0.00000541, -0.00000630, 0.00000721, 0.00001589, -0.00000043, -0.00002973, -0.00003582, -0.00000461, 0.00003167, 0.00003015, -0.00000850, -0.00003812, -0.00002308, 0.00001797, 0.00003630, 0.00001354, -0.00001860, -0.00002122, 0.00000431, 0.00002307, 0.00001326, -0.00000942, -0.00001617, -0.00000096, 0.00001675, 0.00001856, 0.00000529, -0.00001062, -0.00001998, -0.00001961, -0.00000852, 0.00000936, 0.00002148, 0.00001534, -0.00000562, -0.00002136, -0.00001582, 0.00000477, 0.00001948, 0.00001707, 0.00000647, 0.00000189, 0.00000440, 0.00000356, -0.00000573, -0.00001559, -0.00001556, -0.00000544, 0.00000578, 0.00001073, 0.00000940, 0.00000530, 0.00000072, -0.00000296, -0.00000408, -0.00000259, -0.00000131, -0.00000236, -0.00000332, -0.00000000, 0.00000676, 0.00001040, 0.00000655, -0.00000103, -0.00000520, -0.00000462, -0.00000452, -0.00000788, -0.00000999, -0.00000482, 0.00000535, 0.00001074, 0.00000527, -0.00000617, -0.00001329, -0.00001146, -0.00000492, 0.00000015, 0.00000242, 0.00000421, 0.00000580, 0.00000403, -0.00000303, -0.00001167, -0.00001469, -0.00000870, },
		{ -0.00001534, -0.00001268, -0.00000236, 0.00000627, 0.00000482, -0.00000464, -0.00001084, -0.00000577, 0.00000524, 0.00000907, 0.00000096, -0.00000847, -0.00000570, 0.00000827, 0.00001794, 0.00001202, -0.00000288, -0.00001096, -0.00000671, 0.00000051, 0.00000140, -0.00000109, 0.00000130, 0.00000764, 0.00000844, 0.00000051, -0.00000672, -0.00000395, 0.00000422, 0.00000552, -0.00000235, -0.00000836, -0.00000450, 0.00000249, 0.00000164, -0.00000516, -0.00000584, 0.00000279, 0.00000866, 0.00000204, -0.00000868, -0.00000821, 0.00000248, 0.00000647, -0.00000325, -0.00001066, 0.00000052, 0.00001850, 0.00001436, -0.00001672, -0.00004083, -0.00002461, 0.00001877, 0.00004014, 0.00001371, -0.00002936, -0.00003804, -0.00000353, 0.00003178, 0.00002801, -0.00000513, -0.00002518, -0.00001107, 0.00001519, 0.00002109, 0.00000342, -0.00001347, -0.00001101, 0.00000438, 0.00001487, 0.00001241, 0.00000152, -0.00001096, -0.00002005, -0.00001967, -0.00000584, 0.00001409, 0.00002296, 0.00001102, -0.00001075, -0.00002069, -0.00000996, 0.00000905, 0.00001868, 0.00001510, 0.00000789, 0.00000457, 0.00000250, -0.00000350, -0.00001148, -0.00001445, -0.00000948, -0.00000065, 0.00000643, 0.00000955, 0.00000895, 0.00000513, -0.00000029, -0.00000385, -0.00000344, -0.00000146, -0.00000187, -0.00000417, -0.00000337, 0.00000273, 0.00000914, 0.00000920, 0.00000311, -0.00000248, -0.00000344, -0.00000312, -0.00000623, -0.00001069, -0.00000950, -0.00000071, 0.00000831, 0.00000860, -0.00000013, -0.00000952, -0.00001200, -0.00000787, -0.00000250, 0.00000082, 0.00000305, 0.00000531, 0.00000540, 0.00000022, -0.00000894, },
		{ 0.00000317, -0.00000500, -0.00001400, -0.00001579, -0.00000745, 0.00000405, 0.00000755, -0.00000028, -0.00001010, -0.00000970, 0.00000137, 0.00001021, 0.00000597, -0.00000594, -0.00000951, 0.00000179, 0.00001597, 0.00001671, 0.00000329, -0.00000935, -0.00000972, -0.00000247, 0.00000105, -0.00000083, -0.00000015, 0.00000607, 0.00000977, 0.00000408, -0.00000498, -0.00000589, 0.00000175, 0.00000628, 0.00000080, -0.00000675, -0.00000560, 0.00000152, 0.00000259, -0.00000470, -0.00000832, -0.00000011, 0.00001004, 0.00000712, -0.00000625, -0.00001179, -0.00000188, 0.00000806, 0.00000182, -0.00001155, -0.00000840, 0.00001319, 0.00002416, 0.00000186, -0.00003384, -0.00003928, -0.00000246, 0.00003673, 0.00003257, -0.00000988, -0.00004026, -0.00002426, 0.00001629, 0.00003322, 0.00001102, -0.00001892, -0.00002126, 0.00000181, 0.00001968, 0.00001380, -0.00000407, -0.00001233, -0.00000551, 0.00000552, 0.00001063, 0.00000839, 0.00000054, -0.00001146, -0.00002152, -0.00001915, -0.00000137, 0.00001862, 0.00002216, 0.00000576, -0.00001397, -0.00001818, -0.00000512, 0.00001122, 0.00001831, 0.00001570, 0.00000967, 0.00000357, -0.00000311, -0.00000957, -0.00001245, -0.00001012, -0.00000464, 0.00000114, 0.00000613, 0.00000947, 0.00000907, 0.00000423, -0.00000162, -0.00000389, -0.00000244, -0.00000158, -0.00000356, -0.00000482, -0.00000103, 0.00000587, 0.00000918, 0.00000594, 0.00000059, -0.00000148, -0.00000126, -0.00000377, -0.00000950, -0.00001214, -0.00000658, 0.00000348, 0.00000867, 0.00000434, -0.00000472, -0.00001025, -0.00000914, -0.00000473, -0.00000113, 0.00000126, 0.00000385, 0.00000568, },
		{ 0.00000174, 0.00000446, 0.00000474, -0.00000084, -0.00001054, -0.00001667, -0.00001249, -0.00000056, 0.00000784, 0.00000414, -0.00000700, -0.00001181, -0.00000346, 0.00000880, 0.00001009, -0.00000124, -0.00001051, -0.00000459, 0.00001108, 0.00001862, 0.00000968, -0.00000515, -0.00001071, -0.00000563, -0.00000050, -0.00000106, -0.00000142, 0.00000415, 0.00001014, 0.00000733, -0.00000233, -0.00000680, -0.00000086, 0.00000604, 0.00000336, -0.00000494, -0.00000642, 0.00000070, 0.00000433, -0.00000283, -0.00001055, -0.00000518, 0.00000873, 0.00001240, -0.00000051, -0.00001307, -0.00000819, 0.00000615, 0.00000746, -0.00000742, -0.00001479, 0.00000225, 0.00002476, 0.00001881, -0.00001693, -0.00004229, -0.00002359, 0.00002106, 0.00004040, 0.00001220, -0.00002941, -0.00003621, -0.00000369, 0.00002716, 0.00002240, -0.00000653, -0.00002273, -0.00001042, 0.00001123, 0.00001721, 0.00000568, -0.00000671, -0.00000852, -0.00000284, 0.00000312, 0.00000696, 0.00000738, 0.00000051, -0.00001339, -0.00002344, -0.00001695, 0.00000418, 0.00002171, 0.00001949, 0.00000093, -0.00001536, -0.00001541, -0.00000175, 0.00001307, 0.00001977, 0.00001727, 0.00000892, -0.00000130, -0.00000925, -0.00001187, -0.00000955, -0.00000566, -0.00000238, 0.00000134, 0.00000643, 0.00001018, 0.00000873, 0.00000277, -0.00000242, -0.00000334, -0.00000227, -0.00000299, -0.00000469, -0.00000312, 0.00000249, 0.00000722, 0.00000663, 0.00000265, 0.00000048, 0.00000091, -0.00000061, -0.00000662, -0.00001246, -0.00001113, -0.00000239, 0.00000594, 0.00000639, -0.00000036, -0.00000722, -0.00000879, -0.00000592, -0.00000258, -0.00000046, },
		{ -0.00000322, -0.00000139, -0.00000011, 0.00000224, 0.00000434, 0.00000212, -0.00000601, -0.00001483, -0.00001572, -0.00000622, 0.00000522, 0.00000687, -0.00000268, -0.00001148, -0.00000769, 0.00000538, 0.00001220, 0.00000418, -0.00000858, -0.00000921, 0.00000463, 0.00001731, 0.00001454, 0.00000046, -0.00000929, -0.00000767, -0.00000243, -0.00000191, -0.00000298, 0.00000170, 0.00000959, 0.00001020, 0.00000107, -0.00000672, -0.00000348, 0.00000500, 0.00000558, -0.00000277, -0.00000732, -0.00000094, 0.00000610, 0.00000110, -0.00001058, -0.00001113, 0.00000354, 0.00001523, 0.00000756, -0.00000995, -0.00001382, 0.00000017, 0.00001023, 0.00000008, -0.00001526, -0.00000910, 0.00001649, 0.00002774, 0.00000323, -0.00003275, -0.00003639, -0.00000007, 0.00003499, 0.00002815, -0.00001111, -0.00003604, -0.00002017, 0.00001354, 0.00002496, 0.00000552, -0.00001704, -0.00001694, 0.00000100, 0.00001394, 0.00001094, 0.00000069, -0.00000525, -0.00000560, -0.00000400, -0.00000033, 0.00000588, 0.00000840, -0.00000042, -0.00001669, -0.00002450, -0.00001294, 0.00000966, 0.00002302, 0.00001600, -0.00000292, -0.00001597, -0.00001324, 0.00000124, 0.00001620, 0.00002243, 0.00001719, 0.00000443, -0.00000781, -0.00001276, -0.00001011, -0.00000537, -0.00000314, -0.00000211, 0.00000160, 0.00000762, 0.00001076, 0.00000770, 0.00000147, -0.00000257, -0.00000322, -0.00000339, -0.00000440, -0.00000372, 0.00000040, 0.00000501, 0.00000577, 0.00000301, 0.00000127, 0.00000248, 0.00000271, -0.00000235, -0.00001029, -0.00001350, -0.00000805, 0.00000124, 0.00000587, 0.00000263, -0.00000390, -0.00000738, -0.00000614, },
		{ -0.00000545, -0.00000580, -0.00000336, -0.00000143, -0.00000086, 0.00000020, 0.00000246, 0.00000292, -0.00000213, -0.00001108, -0.00001614, -0.00001095, 0.00000069, 0.00000694, 0.00000108, -0.00000932, -0.00001029, 0.00000120, 0.00001213, 0.00000897, -0.00000460, -0.00001130, -0.00000173, 0.00001347, 0.00001693, 0.00000588, -0.00000617, -0.00000807, -0.00000371, -0.00000269, -0.00000475, -0.00000160, 0.00000765, 0.00001235, 0.00000524, -0.00000537, -0.00000598, 0.00000293, 0.00000735, 0.00000021, -0.00000773, -0.00000389, 0.00000634, 0.00000621, -0.00000698, -0.00001530, -0.00000469, 0.00001321, 0.00001494, -0.00000228, -0.00001554, -0.00000778, 0.00000793, 0.00000707, -0.00000983, -0.00001580, 0.00000385, 0.00002642, 0.00001905, -0.00001540, -0.00003703, -0.00001810, 0.00002032, 0.00003311, 0.00000652, -0.00002637, -0.00002805, -0.00000107, 0.00001977, 0.00001246, -0.00000897, -0.00001746, -0.00000649, 0.00000775, 0.00001089, 0.00000529, 0.00000010, -0.00000267, -0.00000593, -0.00000757, -0.00000217, 0.00000758, 0.00000956, -0.00000311, -0.00002034, -0.00002388, -0.00000778, 0.00001426, 0.00002297, 0.00001239, -0.00000625, -0.00001667, -0.00001116, 0.00000538, 0.00002071, 0.00002409, 0.00001359, -0.00000259, -0.00001286, -0.00001230, -0.00000609, -0.00000235, -0.00000279, -0.00000232, 0.00000271, 0.00000903, 0.00001068, 0.00000648, 0.00000071, -0.00000280, -0.00000426, -0.00000500, -0.00000415, -0.00000050, 0.00000382, 0.00000490, 0.00000235, 0.00000046, 0.00000235, 0.00000498, 0.00000246, -0.00000577, -0.00001296, -0.00001223, -0.00000439, 0.00000297, 0.00000373, -0.00000101, },
		{ 0.00000257, 0.00000081, -0.00000333, -0.00000507, -0.00000344, -0.00000116, -0.00000050, -0.00000061, 0.00000047, 0.00000188, -0.00000019, -0.00000724, -0.00001420, -0.00001336, -0.00000389, 0.00000485, 0.00000309, -0.00000676, -0.00001137, -0.00000270, 0.00001057, 0.00001253, 0.00000033, -0.00001103, -0.00000691, 0.00000836, 0.00001700, 0.00001012, -0.00000256, -0.00000731, -0.00000395, -0.00000250, -0.00000587, -0.00000548, 0.00000382, 0.00001289, 0.00000980, -0.00000226, -0.00000776, -0.00000032, 0.00000796, 0.00000402, -0.00000659, -0.00000744, 0.00000385, 0.00001031, -0.00000016, -0.00001513, -0.00001306, 0.00000601, 0.00001814, 0.00000746, -0.00001162, -0.00001398, 0.00000125, 0.00000990, -0.00000176, -0.00001574, -0.00000732, 0.00001743, 0.00002588, 0.00000238, -0.00002745, -0.00002714, 0.00000376, 0.00002808, 0.00001746, -0.00001342, -0.00002740, -0.00001131, 0.00001141, 0.00001368, -0.00000281, -0.00001505, -0.00001011, 0.00000247, 0.00000819, 0.00000584, 0.00000323, 0.00000224, -0.00000212, -0.00000956, -0.00001095, -0.00000104, 0.00001075, 0.00000931, -0.00000728, -0.00002323, -0.00002152, -0.00000222, 0.00001788, 0.00002207, 0.00000856, -0.00000977, -0.00001730, -0.00000782, 0.00001127, 0.00002495, 0.00002258, 0.00000673, -0.00000935, -0.00001432, -0.00000879, -0.00000230, -0.00000147, -0.00000350, -0.00000190, 0.00000440, 0.00001001, 0.00001016, 0.00000559, 0.00000010, -0.00000398, -0.00000613, -0.00000549, -0.00000148, 0.00000354, 0.00000518, 0.00000222, -0.00000106, 0.00000022, 0.00000485, 0.00000617, 0.00000010, -0.00000923, -0.00001361, -0.00000956, -0.00000173, },
		{ -0.00001249, -0.00000697, -0.00000083, 0.00000085, -0.00000154, -0.00000390, -0.00000345, -0.00000118, 0.00000019, -0.00000011, -0.00000036, 0.00000038, -0.00000017, -0.00000484, -0.00001157, -0.00001353, -0.00000697, 0.00000219, 0.00000348, -0.00000492, -0.00001192, -0.00000624, 0.00000819, 0.00001503, 0.00000562, -0.00000895, -0.00001068, 0.00000282, 0.00001545, 0.00001314, 0.00000081, -0.00000636, -0.00000376, -0.00000109, -0.00000520, -0.00000869, -0.00000174, 0.00001075, 0.00001358, 0.00000267, -0.00000773, -0.00000426, 0.00000657, 0.00000772, -0.00000324, -0.00000992, -0.00000112, 0.00001114, 0.00000754, -0.00000986, -0.00001799, -0.00000408, 0.00001523, 0.00001528, -0.00000319, -0.00001528, -0.00000657, 0.00000745, 0.00000471, -0.00001056, -0.00001313, 0.00000610, 0.00002390, 0.00001463, -0.00001349, -0.00002650, -0.00000825, 0.00001830, 0.00002048, -0.00000296, -0.00002229, -0.00001589, 0.00000438, 0.00001203, 0.00000028, -0.00001308, -0.00001187, -0.00000064, 0.00000606, 0.00000447, 0.00000270, 0.00000480, 0.00000399, -0.00000503, -0.00001454, -0.00001191, 0.00000270, 0.00001358, 0.00000707, -0.00001200, -0.00002471, -0.00001780, 0.00000333, 0.00002066, 0.00002031, 0.00000396, -0.00001351, -0.00001666, -0.00000225, 0.00001782, 0.00002661, 0.00001724, -0.00000122, -0.00001328, -0.00001216, -0.00000444, -0.00000034, -0.00000218, -0.00000401, -0.00000069, 0.00000599, 0.00001047, 0.00000972, 0.00000491, -0.00000123, -0.00000620, -0.00000748, -0.00000368, 0.00000278, 0.00000633, 0.00000376, -0.00000151, -0.00000274, 0.00000199, 0.00000700, 0.00000531, -0.00000317, -0.00001134, },
		{ 0.00000303, -0.00000586, -0.00001173, -0.00001077, -0.00000558, -0.00000142, -0.00000095, -0.00000252, -0.00000303, -0.00000148, 0.00000041, 0.00000083, 0.00000020, -0.00000014, -0.00000079, -0.00000411, -0.00000987, -0.00001291, -0.00000841, 0.00000041, 0.00000358, -0.00000368, -0.00001274, -0.00001016, 0.00000481, 0.00001662, 0.00001134, -0.00000516, -0.00001308, -0.00000294, 0.00001265, 0.00001538, 0.00000414, -0.00000570, -0.00000430, 0.00000059, -0.00000234, -0.00000957, -0.00000762, 0.00000554, 0.00001488, 0.00000845, -0.00000505, -0.00000754, 0.00000303, 0.00000984, 0.00000177, -0.00000971, -0.00000677, 0.00000788, 0.00001301, -0.00000116, -0.00001723, -0.00001315, 0.00000711, 0.00001790, 0.00000628, -0.00001108, -0.00001177, 0.00000159, 0.00000693, -0.00000404, -0.00001322, -0.00000281, 0.00001687, 0.00001939, -0.00000112, -0.00001988, -0.00001335, 0.00000943, 0.00001871, 0.00000290, -0.00001730, -0.00001726, 0.00000018, 0.00001091, 0.00000228, -0.00001248, -0.00001439, -0.00000329, 0.00000560, 0.00000424, 0.00000050, 0.00000324, 0.00000770, 0.00000287, -0.00001065, -0.00001837, -0.00000968, 0.00000759, 0.00001477, 0.00000315, -0.00001637, -0.00002461, -0.00001304, 0.00000877, 0.00002247, 0.00001714, -0.00000159, -0.00001642, -0.00001350, 0.00000505, 0.00002280, 0.00002430, 0.00000951, -0.00000748, -0.00001342, -0.00000808, -0.00000126, -0.00000030, -0.00000327, -0.00000375, 0.00000080, 0.00000719, 0.00001081, 0.00000947, 0.00000374, -0.00000368, -0.00000847, -0.00000687, 0.00000027, 0.00000676, 0.00000656, 0.00000045, -0.00000437, -0.00000229, 0.00000429, 0.00000756, },
		{ -0.00000066, 0.00000567, 0.00000657, 0.00000067, -0.00000716, -0.00001115, -0.00000969, -0.00000558, -0.00000253, -0.00000183, -0.00000207, -0.00000145, 0.00000008, 0.00000121, 0.00000124, 0.00000070, -0.00000050, -0.00000385, -0.00000945, -0.00001308, -0.00000955, -0.00000040, 0.00000469, -0.00000170, -0.00001339, -0.00001501, -0.00000052, 0.00001660, 0.00001739, 0.00000077, -0.00001363, -0.00000893, 0.00000829, 0.00001687, 0.00000812, -0.00000473, -0.00000615, 0.00000090, 0.00000166, -0.00000711, -0.00001158, -0.00000167, 0.00001245, 0.00001311, 0.00000004, -0.00000860, -0.00000166, 0.00000927, 0.00000681, -0.00000638, -0.00001073, 0.00000167, 0.00001407, 0.00000766, -0.00001107, -0.00001769, -0.00000288, 0.00001452, 0.00001287, -0.00000360, -0.00001231, -0.00000427, 0.00000512, 0.00000051, -0.00001008, -0.00000739, 0.00000928, 0.00001859, 0.00000660, -0.00001223, -0.00001340, 0.00000428, 0.00001649, 0.00000595, -0.00001440, -0.00001888, -0.00000307, 0.00001135, 0.00000592, -0.00001129, -0.00001826, -0.00000798, 0.00000501, 0.00000632, -0.00000006, -0.00000066, 0.00000652, 0.00000904, -0.00000170, -0.00001679, -0.00001931, -0.00000494, 0.00001200, 0.00001386, -0.00000173, -0.00001980, -0.00002290, -0.00000736, 0.00001382, 0.00002266, 0.00001218, -0.00000723, -0.00001702, -0.00000767, 0.00001217, 0.00002424, 0.00001854, 0.00000201, -0.00001031, -0.00001073, -0.00000421, 0.00000000, -0.00000125, -0.00000387, -0.00000297, 0.00000211, 0.00000819, 0.00001131, 0.00000893, 0.00000147, -0.00000667, -0.00000931, -0.00000390, 0.00000485, 0.00000875, 0.00000445, -0.00000294, -0.00000538, },
		{ 0.00000144, -0.00000502, -0.00000494, 0.00000067, 0.00000564, 0.00000507, -0.00000070, -0.00000736, -0.00001078, -0.00000982, -0.00000634, -0.00000307, -0.00000135, -0.00000067, -0.00000003, 0.00000082, 0.00000157, 0.00000196, 0.00000116, -0.00000246, -0.00000908, -0.00001438, -0.00001201, -0.00000172, 0.00000669, 0.00000235, -0.00001215, -0.00001998, -0.00000850, 0.00001336, 0.00002264, 0.00000914, -0.00001110, -0.00001439, 0.00000182, 0.00001667, 0.00001289, -0.00000213, -0.00000853, -0.00000135, 0.00000462, -0.00000195, -0.00001174, -0.00000847, 0.00000650, 0.00001462, 0.00000594, -0.00000667, -0.00000569, 0.00000614, 0.00001006, -0.00000102, -0.00001141, -0.00000501, 0.00001060, 0.00001343, -0.00000213, -0.00001643, -0.00001081, 0.00000718, 0.00001447, 0.00000364, -0.00000888, -0.00000745, 0.00000165, 0.00000208, -0.00000673, -0.00000869, 0.00000368, 0.00001568, 0.00001024, -0.00000659, -0.00001207, 0.00000189, 0.00001618, 0.00000954, -0.00001204, -0.00002211, -0.00000836, 0.00001125, 0.00001187, -0.00000668, -0.00002119, -0.00001534, 0.00000137, 0.00000889, 0.00000276, -0.00000322, 0.00000163, 0.00000996, 0.00000708, -0.00000839, -0.00002120, -0.00001704, 0.00000084, 0.00001479, 0.00001105, -0.00000682, -0.00002183, -0.00001950, -0.00000105, 0.00001769, 0.00002055, 0.00000595, -0.00001139, -0.00001449, -0.00000054, 0.00001682, 0.00002176, 0.00001151, -0.00000299, -0.00000984, -0.00000719, -0.00000186, -0.00000014, -0.00000216, -0.00000391, -0.00000213, 0.00000326, 0.00000932, 0.00001170, 0.00000741, -0.00000170, -0.00000887, -0.00000787, 0.00000044, 0.00000823, 0.00000833, },
		{ 0.00000439, 0.00000946, 0.00000648, -0.00000089, -0.00000567, -0.00000435, 0.00000085, 0.00000490, 0.00000426, -0.00000094, -0.00000750, -0.00001147, -0.00001086, -0.00000683, -0.00000236, 0.00000027, 0.00000083, 0.00000057, 0.00000088, 0.00000222, 0.00000315, 0.00000061, -0.00000695, -0.00001537, -0.00001609, -0.00000545, 0.00000768, 0.00000829, -0.00000717, -0.00002266, -0.00001826, 0.00000545, 0.00002468, 0.00001887, -0.00000432, -0.00001746, -0.00000636, 0.00001338, 0.00001731, 0.00000295, -0.00000951, -0.00000576, 0.00000449, 0.00000360, -0.00000784, -0.00001222, -0.00000099, 0.00001216, 0.00001038, -0.00000246, -0.00000755, 0.00000177, 0.00001057, 0.00000447, -0.00000878, -0.00000975, 0.00000439, 0.00001464, 0.00000617, -0.00001069, -0.00001422, -0.00000066, 0.00001163, 0.00000791, -0.00000406, -0.00000752, -0.00000102, 0.00000188, -0.00000465, -0.00000890, -0.00000005, 0.00001299, 0.00001220, -0.00000291, -0.00001183, -0.00000066, 0.00001679, 0.00001545, -0.00000705, -0.00002504, -0.00001683, 0.00000726, 0.00001763, 0.00000226, -0.00001969, -0.00002299, -0.00000660, 0.00000854, 0.00000758, -0.00000179, -0.00000317, 0.00000576, 0.00001112, 0.00000196, -0.00001502, -0.00002261, -0.00001237, 0.00000623, 0.00001552, 0.00000686, -0.00001143, -0.00002202, -0.00001447, 0.00000510, 0.00001939, 0.00001616, -0.00000010, -0.00001270, -0.00000949, 0.00000559, 0.00001777, 0.00001679, 0.00000570, -0.00000475, -0.00000760, -0.00000447, -0.00000105, -0.00000072, -0.00000265, -0.00000375, -0.00000138, 0.00000454, 0.00001052, 0.00001132, 0.00000478, -0.00000470, -0.00000912, -0.00000461, },
		{ -0.00000624, -0.00000740, -0.00000105, 0.00000664, 0.00000902, 0.00000478, -0.00000196, -0.00000593, -0.00000472, 0.00000017, 0.00000463, 0.00000463, -0.00000085, -0.00000856, -0.00001327, -0.00001195, -0.00000617, -0.00000028, 0.00000233, 0.00000157, 0.00000017, 0.00000096, 0.00000374, 0.00000417, -0.00000234, -0.00001382, -0.00002009, -0.00001216, 0.00000510, 0.00001377, 0.00000174, -0.00002026, -0.00002684, -0.00000685, 0.00002088, 0.00002712, 0.00000646, -0.00001585, -0.00001432, 0.00000637, 0.00001922, 0.00000980, -0.00000727, -0.00001026, 0.00000076, 0.00000679, -0.00000168, -0.00001159, -0.00000719, 0.00000673, 0.00001172, 0.00000217, -0.00000693, -0.00000217, 0.00000875, 0.00000846, -0.00000413, -0.00001141, -0.00000211, 0.00001186, 0.00001140, -0.00000331, -0.00001308, -0.00000612, 0.00000682, 0.00000874, -0.00000034, -0.00000604, -0.00000202, 0.00000177, -0.00000328, -0.00000929, -0.00000338, 0.00001060, 0.00001427, 0.00000079, -0.00001239, -0.00000541, 0.00001532, 0.00002229, 0.00000236, -0.00002392, -0.00002626, -0.00000234, 0.00001903, 0.00001309, -0.00001190, -0.00002677, -0.00001701, 0.00000292, 0.00001077, 0.00000320, -0.00000436, -0.00000019, 0.00000922, 0.00000899, -0.00000475, -0.00001973, -0.00002098, -0.00000664, 0.00001021, 0.00001419, 0.00000196, -0.00001480, -0.00002008, -0.00000840, 0.00000988, 0.00001839, 0.00001059, -0.00000423, -0.00001101, -0.00000398, 0.00000890, 0.00001558, 0.00001169, 0.00000254, -0.00000418, -0.00000537, -0.00000311, -0.00000102, -0.00000113, -0.00000291, -0.00000364, -0.00000052, 0.00000603, 0.00001124, 0.00000982, 0.00000183, },
		{ 0.00001097, 0.00000767, -0.00000017, -0.00000588, -0.00000490, 0.00000127, 0.00000720, 0.00000839, 0.00000419, -0.00000247, -0.00000704, -0.00000613, -0.00000023, 0.00000567, 0.00000570, -0.00000151, -0.00001098, -0.00001554, -0.00001209, -0.00000389, 0.00000271, 0.00000383, 0.00000098, -0.00000072, 0.00000203, 0.00000600, 0.00000347, -0.00000850, -0.00002109, -0.00001999, -0.00000234, 0.00001528, 0.00001210, -0.00001157, -0.00003031, -0.00002060, 0.00001029, 0.00003015, 0.00001871, -0.00000843, -0.00001902, -0.00000312, 0.00001680, 0.00001613, -0.00000149, -0.00001226, -0.00000503, 0.00000600, 0.00000381, -0.00000739, -0.00000997, 0.00000074, 0.00000993, 0.00000538, -0.00000484, -0.00000473, 0.00000574, 0.00001041, 0.00000092, -0.00001024, -0.00000706, 0.00000693, 0.00001299, 0.00000313, -0.00000921, -0.00000830, 0.00000261, 0.00000764, 0.00000142, -0.00000495, -0.00000234, 0.00000247, -0.00000125, -0.00000936, -0.00000721, 0.00000714, 0.00001620, 0.00000609, -0.00001147, -0.00001171, 0.00000946, 0.00002639, 0.00001473, -0.00001611, -0.00003213, -0.00001568, 0.00001315, 0.00002103, 0.00000053, -0.00002356, -0.00002555, -0.00000718, 0.00000903, 0.00000849, -0.00000113, -0.00000384, 0.00000382, 0.00001019, 0.00000421, -0.00001107, -0.00002163, -0.00001712, -0.00000112, 0.00001222, 0.00001119, -0.00000277, -0.00001622, -0.00001618, -0.00000249, 0.00001226, 0.00001521, 0.00000562, -0.00000558, -0.00000766, -0.00000007, 0.00000909, 0.00001221, 0.00000838, 0.00000184, -0.00000293, -0.00000413, -0.00000268, -0.00000103, -0.00000134, -0.00000323, -0.00000347, 0.00000064, 0.00000733, },
		{ -0.00000304, 0.00000181, 0.00000787, 0.00000991, 0.00000605, -0.00000045, -0.00000445, -0.00000327, 0.00000190, 0.00000730, 0.00000882, 0.00000439, -0.00000376, -0.00000952, -0.00000765, 0.00000073, 0.00000791, 0.00000641, -0.00000367, -0.00001443, -0.00001730, -0.00001051, -0.00000014, 0.00000564, 0.00000385, -0.00000080, -0.00000095, 0.00000460, 0.00000772, -0.00000054, -0.00001701, -0.00002533, -0.00001310, 0.00001033, 0.00001957, 0.00000151, -0.00002595, -0.00003105, -0.00000505, 0.00002540, 0.00002822, 0.00000342, -0.00001793, -0.00001206, 0.00000998, 0.00001923, 0.00000613, -0.00001025, -0.00001012, 0.00000177, 0.00000626, -0.00000209, -0.00000909, -0.00000348, 0.00000661, 0.00000652, -0.00000272, -0.00000611, 0.00000251, 0.00001069, 0.00000540, -0.00000725, -0.00000986, 0.00000159, 0.00001181, 0.00000742, -0.00000473, -0.00000813, 0.00000010, 0.00000657, 0.00000223, -0.00000496, -0.00000356, 0.00000302, 0.00000194, -0.00000769, -0.00001077, 0.00000172, 0.00001621, 0.00001251, -0.00000700, -0.00001665, -0.00000034, 0.00002444, 0.00002581, -0.00000232, -0.00003045, -0.00002795, 0.00000063, 0.00002182, 0.00001295, -0.00001358, -0.00002805, -0.00001796, 0.00000178, 0.00001031, 0.00000423, -0.00000320, -0.00000110, 0.00000671, 0.00000827, -0.00000164, -0.00001552, -0.00002082, -0.00001222, 0.00000321, 0.00001219, 0.00000724, -0.00000634, -0.00001537, -0.00001127, 0.00000191, 0.00001211, 0.00001136, 0.00000264, -0.00000474, -0.00000470, 0.00000123, 0.00000746, 0.00000971, 0.00000730, 0.00000226, -0.00000227, -0.00000381, -0.00000244, -0.00000089, -0.00000162, -0.00000359, },
		},
		{ { -0.00000402, -0.00000261, -0.00000042, -0.00000147, -0.00000723, -0.00001408, -0.00001546, -0.00000775, 0.00000532, 0.00001433, 0.00001146, -0.00000235, -0.00001718, -0.00002192, -0.00001365, -0.00000012, 0.00000748, 0.00000474, -0.00000210, -0.00000280, 0.00000643, 0.00001861, 0.00002236, 0.00001264, -0.00000451, -0.00001757, -0.00001974, -0.00001373, -0.00000723, -0.00000462, -0.00000357, 0.00000054, 0.00000813, 0.00001425, 0.00001380, 0.00000718, 0.00000012, -0.00000217, 0.00000010, 0.00000234, 0.00000118, -0.00000195, -0.00000300, -0.00000045, 0.00000264, 0.00000211, -0.00000225, -0.00000596, -0.00000461, 0.00000111, 0.00000575, 0.00000468, -0.00000090, -0.00000485, -0.00000224, 0.00000555, 0.00001183, 0.00001057, 0.00000180, -0.00000860, -0.00001374, -0.00001078, -0.00000228, 0.00000620, 0.00000967, 0.00000620, -0.00000228, -0.00001078, -0.00001374, -0.00000860, 0.00000180, 0.00001057, 0.00001183, 0.00000555, -0.00000224, -0.00000485, -0.00000090, 0.00000468, 0.00000575, 0.00000111, -0.00000461, -0.00000596, -0.00000225, 0.00000211, 0.00000264, -0.00000045, -0.00000300, -0.00000195, 0.00000118, 0.00000234, 0.00000010, -0.00000217, 0.00000012, 0.00000718, 0.00001380, 0.00001425, 0.00000813, 0.00000054, -0.00000357, -0.00000462, -0.00000723, -0.00001373, -0.00001974, -0.00001757, -0.00000451, 0.00001264, 0.00002236, 0.00001861, 0.00000643, -0.00000280, -0.00000210, 0.00000474, 0.00000748, -0.00000012, -0.00001365, -0.00002192, -0.00001718, -0.00000235, 0.00001146, 0.00001433, 0.00000532, -0.00000775, -0.00001546, -0.00001408, -0.00000723, -0.00000147, -0.00000042, -0.00000261, },
		{ -0.00000064, -0.00000022, -0.00000263, -0.00000393, -0.00000249, -0.00000069, -0.00000252, -0.00000893, -0.00001556, -0.00001562, -0.00000626, 0.00000747, 0.00001540, 0.00001054, -0.00000426, -0.00001779, -0.00001976, -0.00000977, 0.00000248, 0.00000634, 0.00000028, -0.00000678, -0.00000448, 0.00000845, 0.00002231, 0.00002512, 0.00001356, -0.00000440, -0.00001707, -0.00001903, -0.00001408, -0.00000922, -0.00000711, -0.00000474, 0.00000120, 0.00000912, 0.00001358, 0.00001128, 0.00000497, 0.00000075, 0.00000163, 0.00000457, 0.00000420, -0.00000089, -0.00000628, -0.00000653, -0.00000121, 0.00000429, 0.00000439, -0.00000090, -0.00000596, -0.00000552, -0.00000005, 0.00000489, 0.00000451, -0.00000024, -0.00000380, -0.00000168, 0.00000506, 0.00001076, 0.00001025, 0.00000314, -0.00000625, -0.00001233, -0.00001188, -0.00000546, 0.00000320, 0.00000910, 0.00000841, 0.00000095, -0.00000881, -0.00001400, -0.00001030, 0.00000019, 0.00000995, 0.00001205, 0.00000596, -0.00000213, -0.00000513, -0.00000143, 0.00000419, 0.00000579, 0.00000203, -0.00000314, -0.00000507, -0.00000311, -0.00000040, 0.00000032, -0.00000030, 0.00000026, 0.00000247, 0.00000358, 0.00000097, -0.00000386, -0.00000566, -0.00000071, 0.00000868, 0.00001553, 0.00001466, 0.00000758, 0.00000052, -0.00000229, -0.00000280, -0.00000610, -0.00001350, -0.00001947, -0.00001658, -0.00000362, 0.00001163, 0.00001873, 0.00001403, 0.00000398, -0.00000126, 0.00000229, 0.00000874, 0.00000844, -0.00000215, -0.00001636, -0.00002303, -0.00001638, -0.00000120, 0.00001130, 0.00001262, 0.00000329, -0.00000867, -0.00001478, -0.00001247, -0.00000571, },
		{ -0.00001365, -0.00001075, -0.00000436, -0.00000003, -0.00000013, -0.00000258, -0.00000368, -0.00000224, -0.00000101, -0.00000379, -0.00001080, -0.00001686, -0.00001524, -0.00000423, 0.00000962, 0.00001574, 0.00000858, -0.00000674, -0.00001806, -0.00001663, -0.00000504, 0.00000529, 0.00000496, -0.00000432, -0.00001124, -0.00000600, 0.00000997, 0.00002476, 0.00002663, 0.00001419, -0.00000330, -0.00001496, -0.00001723, -0.00001441, -0.00001191, -0.00001019, -0.00000590, 0.00000227, 0.00001037, 0.00001273, 0.00000830, 0.00000235, 0.00000118, 0.00000538, 0.00000910, 0.00000631, -0.00000243, -0.00001005, -0.00000994, -0.00000257, 0.00000500, 0.00000611, 0.00000077, -0.00000503, -0.00000566, -0.00000124, 0.00000328, 0.00000361, 0.00000036, -0.00000212, -0.00000041, 0.00000467, 0.00000904, 0.00000903, 0.00000401, -0.00000357, -0.00000997, -0.00001199, -0.00000823, -0.00000025, 0.00000746, 0.00000957, 0.00000383, -0.00000622, -0.00001308, -0.00001109, -0.00000144, 0.00000848, 0.00001133, 0.00000611, -0.00000145, -0.00000462, -0.00000170, 0.00000324, 0.00000508, 0.00000257, -0.00000137, -0.00000349, -0.00000342, -0.00000280, -0.00000233, -0.00000071, 0.00000295, 0.00000651, 0.00000602, 0.00000018, -0.00000690, -0.00000849, -0.00000167, 0.00000927, 0.00001623, 0.00001472, 0.00000752, 0.00000124, -0.00000085, -0.00000161, -0.00000582, -0.00001347, -0.00001844, -0.00001436, -0.00000188, 0.00001071, 0.00001468, 0.00000902, 0.00000122, -0.00000017, 0.00000591, 0.00001196, 0.00000926, -0.00000332, -0.00001763, -0.00002304, -0.00001552, -0.00000092, 0.00001004, 0.00001034, 0.00000142, -0.00000902, },
		{ 0.00000755, -0.00000029, -0.00000889, -0.00001211, -0.00000891, -0.00000313, 0.00000036, -0.00000018, -0.00000250, -0.00000326, -0.00000185, -0.00000138, -0.00000528, -0.00001279, -0.00001793, -0.00001429, -0.00000173, 0.00001162, 0.00001527, 0.00000571, -0.00000951, -0.00001778, -0.00001264, 0.00000015, 0.00000793, 0.00000325, -0.00000876, -0.00001504, -0.00000704, 0.00001099, 0.00002574, 0.00002662, 0.00001436, -0.00000131, -0.00001134, -0.00001431, -0.00001455, -0.00001501, -0.00001366, -0.00000711, 0.00000347, 0.00001162, 0.00001171, 0.00000517, -0.00000032, 0.00000148, 0.00000876, 0.00001319, 0.00000840, -0.00000333, -0.00001288, -0.00001287, -0.00000438, 0.00000461, 0.00000698, 0.00000251, -0.00000321, -0.00000491, -0.00000226, 0.00000108, 0.00000198, 0.00000074, -0.00000001, 0.00000147, 0.00000451, 0.00000695, 0.00000707, 0.00000430, -0.00000088, -0.00000695, -0.00001111, -0.00001028, -0.00000374, 0.00000496, 0.00000955, 0.00000604, -0.00000329, -0.00001106, -0.00001083, -0.00000283, 0.00000633, 0.00000968, 0.00000586, -0.00000041, -0.00000341, -0.00000159, 0.00000204, 0.00000376, 0.00000266, 0.00000042, -0.00000150, -0.00000317, -0.00000480, -0.00000493, -0.00000151, 0.00000487, 0.00000978, 0.00000821, -0.00000002, -0.00000877, -0.00001037, -0.00000262, 0.00000891, 0.00001580, 0.00001430, 0.00000788, 0.00000266, 0.00000081, -0.00000093, -0.00000623, -0.00001363, -0.00001688, -0.00001132, 0.00000045, 0.00001000, 0.00001071, 0.00000410, -0.00000164, 0.00000026, 0.00000835, 0.00001411, 0.00000994, -0.00000348, -0.00001733, -0.00002193, -0.00001463, -0.00000153, 0.00000775, },
		{ -0.00001370, -0.00000292, 0.00000456, 0.00000431, -0.00000191, -0.00000838, -0.00001022, -0.00000692, -0.00000198, 0.00000056, -0.00000038, -0.00000242, -0.00000268, -0.00000133, -0.00000180, -0.00000697, -0.00001484, -0.00001867, -0.00001275, 0.00000112, 0.00001330, 0.00001393, 0.00000210, -0.00001227, -0.00001678, -0.00000791, 0.00000541, 0.00001002, 0.00000114, -0.00001278, -0.00001779, -0.00000738, 0.00001153, 0.00002516, 0.00002501, 0.00001396, 0.00000136, -0.00000645, -0.00001042, -0.00001433, -0.00001814, -0.00001719, -0.00000836, 0.00000447, 0.00001253, 0.00001049, 0.00000222, -0.00000264, 0.00000179, 0.00001152, 0.00001642, 0.00001016, -0.00000358, -0.00001447, -0.00001495, -0.00000643, 0.00000312, 0.00000680, 0.00000406, -0.00000071, -0.00000326, -0.00000291, -0.00000144, -0.00000026, 0.00000078, 0.00000224, 0.00000378, 0.00000465, 0.00000476, 0.00000462, 0.00000400, 0.00000151, -0.00000363, -0.00000938, -0.00001140, -0.00000688, 0.00000188, 0.00000838, 0.00000732, -0.00000039, -0.00000816, -0.00000950, -0.00000376, 0.00000380, 0.00000725, 0.00000512, 0.00000076, -0.00000169, -0.00000107, 0.00000083, 0.00000208, 0.00000233, 0.00000198, 0.00000062, -0.00000248, -0.00000620, -0.00000712, -0.00000250, 0.00000593, 0.00001195, 0.00000986, 0.00000030, -0.00000935, -0.00001110, -0.00000338, 0.00000771, 0.00001424, 0.00001332, 0.00000848, 0.00000463, 0.00000267, -0.00000054, -0.00000705, -0.00001388, -0.00001505, -0.00000792, 0.00000302, 0.00000955, 0.00000723, -0.00000023, -0.00000436, -0.00000009, 0.00000930, 0.00001495, 0.00001040, -0.00000258, -0.00001542, -0.00001973, },
		{ -0.00000075, -0.00001206, -0.00001649, -0.00001265, -0.00000487, 0.00000067, 0.00000066, -0.00000355, -0.00000762, -0.00000804, -0.00000476, -0.00000084, 0.00000058, -0.00000076, -0.00000236, -0.00000198, -0.00000071, -0.00000230, -0.00000883, -0.00001684, -0.00001896, -0.00001062, 0.00000415, 0.00001445, 0.00001169, -0.00000203, -0.00001470, -0.00001491, -0.00000263, 0.00001034, 0.00001126, -0.00000138, -0.00001618, -0.00001925, -0.00000688, 0.00001163, 0.00002309, 0.00002188, 0.00001292, 0.00000443, -0.00000070, -0.00000579, -0.00001364, -0.00002086, -0.00002037, -0.00000959, 0.00000498, 0.00001274, 0.00000898, -0.00000032, -0.00000424, 0.00000230, 0.00001354, 0.00001845, 0.00001127, -0.00000329, -0.00001470, -0.00001592, -0.00000843, 0.00000068, 0.00000551, 0.00000516, 0.00000220, -0.00000084, -0.00000303, -0.00000396, -0.00000288, 0.00000042, 0.00000434, 0.00000625, 0.00000511, 0.00000276, 0.00000199, 0.00000316, 0.00000335, -0.00000038, -0.00000703, -0.00001151, -0.00000935, -0.00000141, 0.00000620, 0.00000753, 0.00000215, -0.00000471, -0.00000722, -0.00000408, 0.00000120, 0.00000429, 0.00000388, 0.00000182, 0.00000028, -0.00000021, -0.00000018, 0.00000032, 0.00000166, 0.00000313, 0.00000254, -0.00000150, -0.00000688, -0.00000863, -0.00000345, 0.00000616, 0.00001286, 0.00001073, 0.00000096, -0.00000869, -0.00001061, -0.00000380, 0.00000590, 0.00001172, 0.00001177, 0.00000908, 0.00000686, 0.00000464, -0.00000023, -0.00000792, -0.00001403, -0.00001313, -0.00000464, 0.00000543, 0.00000932, 0.00000458, -0.00000351, -0.00000666, -0.00000119, 0.00000864, 0.00001435, 0.00001055, },
		{ 0.00000643, 0.00001230, 0.00001025, 0.00000177, -0.00000752, -0.00001235, -0.00001134, -0.00000711, -0.00000367, -0.00000334, -0.00000532, -0.00000680, -0.00000567, -0.00000242, 0.00000033, 0.00000048, -0.00000130, -0.00000233, -0.00000119, -0.00000004, -0.00000292, -0.00001080, -0.00001867, -0.00001871, -0.00000795, 0.00000716, 0.00001489, 0.00000855, -0.00000643, -0.00001648, -0.00001206, 0.00000299, 0.00001459, 0.00001142, -0.00000431, -0.00001878, -0.00001930, -0.00000551, 0.00001136, 0.00001971, 0.00001746, 0.00001131, 0.00000760, 0.00000540, -0.00000078, -0.00001240, -0.00002277, -0.00002279, -0.00001069, 0.00000479, 0.00001194, 0.00000705, -0.00000228, -0.00000485, 0.00000317, 0.00001478, 0.00001908, 0.00001151, -0.00000262, -0.00001359, -0.00001561, -0.00001011, -0.00000246, 0.00000316, 0.00000559, 0.00000515, 0.00000212, -0.00000255, -0.00000616, -0.00000557, -0.00000030, 0.00000604, 0.00000858, 0.00000579, 0.00000115, -0.00000047, 0.00000197, 0.00000452, 0.00000245, -0.00000435, -0.00001067, -0.00001090, -0.00000453, 0.00000330, 0.00000668, 0.00000406, -0.00000110, -0.00000423, -0.00000370, -0.00000115, 0.00000110, 0.00000223, 0.00000258, 0.00000223, 0.00000090, -0.00000085, -0.00000126, 0.00000081, 0.00000378, 0.00000401, -0.00000045, -0.00000685, -0.00000927, -0.00000413, 0.00000567, 0.00001251, 0.00001071, 0.00000177, -0.00000700, -0.00000898, -0.00000373, 0.00000377, 0.00000854, 0.00000969, 0.00000943, 0.00000897, 0.00000656, 0.00000019, -0.00000846, -0.00001385, -0.00001127, -0.00000190, 0.00000727, 0.00000919, 0.00000294, -0.00000544, -0.00000828, -0.00000289, },
		{ -0.00000586, -0.00000900, -0.00000494, 0.00000293, 0.00000893, 0.00000938, 0.00000459, -0.00000221, -0.00000750, -0.00000965, -0.00000927, -0.00000816, -0.00000760, -0.00000734, -0.00000609, -0.00000320, 0.00000011, 0.00000159, 0.00000031, -0.00000195, -0.00000232, -0.00000034, 0.00000060, -0.00000369, -0.00001284, -0.00002017, -0.00001781, -0.00000482, 0.00000994, 0.00001443, 0.00000459, -0.00001083, -0.00001735, -0.00000823, 0.00000870, 0.00001787, 0.00001039, -0.00000755, -0.00002047, -0.00001795, -0.00000338, 0.00001074, 0.00001532, 0.00001213, 0.00000926, 0.00001059, 0.00001131, 0.00000417, -0.00001064, -0.00002352, -0.00002403, -0.00001148, 0.00000378, 0.00000992, 0.00000459, -0.00000364, -0.00000434, 0.00000455, 0.00001530, 0.00001827, 0.00001071, -0.00000180, -0.00001134, -0.00001403, -0.00001121, -0.00000590, -0.00000002, 0.00000524, 0.00000774, 0.00000526, -0.00000150, -0.00000776, -0.00000800, -0.00000129, 0.00000715, 0.00001045, 0.00000656, 0.00000009, -0.00000249, 0.00000064, 0.00000497, 0.00000459, -0.00000168, -0.00000903, -0.00001142, -0.00000716, 0.00000004, 0.00000491, 0.00000516, 0.00000229, -0.00000086, -0.00000268, -0.00000300, -0.00000195, 0.00000033, 0.00000290, 0.00000387, 0.00000207, -0.00000110, -0.00000242, -0.00000005, 0.00000391, 0.00000487, 0.00000049, -0.00000621, -0.00000902, -0.00000440, 0.00000467, 0.00001105, 0.00000979, 0.00000250, -0.00000458, -0.00000643, -0.00000310, 0.00000171, 0.00000509, 0.00000721, 0.00000930, 0.00001057, 0.00000820, 0.00000085, -0.00000831, -0.00001307, -0.00000949, 0.00000001, 0.00000821, 0.00000898, 0.00000233, },
		{ 0.00000802, 0.00000847, 0.00000260, -0.00000484, -0.00000870, -0.00000699, -0.00000141, 0.00000453, 0.00000784, 0.00000727, 0.00000336, -0.00000220, -0.00000744, -0.00001098, -0.00001245, -0.00001200, -0.00000971, -0.00000565, -0.00000078, 0.00000275, 0.00000295, 0.00000013, -0.00000263, -0.00000229, 0.00000051, 0.00000108, -0.00000469, -0.00001485, -0.00002118, -0.00001617, -0.00000131, 0.00001227, 0.00001297, -0.00000008, -0.00001491, -0.00001711, -0.00000349, 0.00001423, 0.00001994, 0.00000816, -0.00001093, -0.00002117, -0.00001535, -0.00000068, 0.00000978, 0.00001023, 0.00000634, 0.00000699, 0.00001318, 0.00001649, 0.00000862, -0.00000847, -0.00002287, -0.00002377, -0.00001176, 0.00000197, 0.00000658, 0.00000150, -0.00000444, -0.00000270, 0.00000650, 0.00001525, 0.00001616, 0.00000886, -0.00000107, -0.00000825, -0.00001132, -0.00001152, -0.00000925, -0.00000371, 0.00000408, 0.00000965, 0.00000822, -0.00000001, -0.00000855, -0.00000984, -0.00000237, 0.00000755, 0.00001161, 0.00000723, -0.00000041, -0.00000383, -0.00000060, 0.00000479, 0.00000589, 0.00000069, -0.00000686, -0.00001095, -0.00000904, -0.00000322, 0.00000248, 0.00000539, 0.00000510, 0.00000250, -0.00000116, -0.00000416, -0.00000453, -0.00000160, 0.00000275, 0.00000498, 0.00000313, -0.00000095, -0.00000302, -0.00000074, 0.00000360, 0.00000505, 0.00000116, -0.00000513, -0.00000795, -0.00000419, 0.00000339, 0.00000873, 0.00000807, 0.00000299, -0.00000185, -0.00000331, -0.00000193, 0.00000003, 0.00000183, 0.00000454, 0.00000851, 0.00001130, 0.00000932, 0.00000178, -0.00000722, -0.00001145, -0.00000776, 0.00000094, },
		{ -0.00000884, -0.00000597, 0.00000091, 0.00000663, 0.00000748, 0.00000345, -0.00000261, -0.00000736, -0.00000867, -0.00000604, -0.00000051, 0.00000561, 0.00000937, 0.00000860, 0.00000322, -0.00000464, -0.00001192, -0.00001621, -0.00001637, -0.00001244, -0.00000562, 0.00000145, 0.00000542, 0.00000439, 0.00000002, -0.00000324, -0.00000219, 0.00000129, 0.00000130, -0.00000595, -0.00001676, -0.00002154, -0.00001375, 0.00000242, 0.00001394, 0.00001042, -0.00000527, -0.00001841, -0.00001563, 0.00000201, 0.00001928, 0.00002066, 0.00000484, -0.00001426, -0.00002086, -0.00001173, 0.00000230, 0.00000847, 0.00000481, 0.00000060, 0.00000475, 0.00001519, 0.00002051, 0.00001214, -0.00000605, -0.00002075, -0.00002182, -0.00001135, -0.00000051, 0.00000202, -0.00000224, -0.00000482, -0.00000011, 0.00000898, 0.00001481, 0.00001302, 0.00000608, -0.00000063, -0.00000472, -0.00000775, -0.00001097, -0.00001209, -0.00000751, 0.00000219, 0.00001059, 0.00001059, 0.00000172, -0.00000843, -0.00001081, -0.00000333, 0.00000726, 0.00001190, 0.00000761, -0.00000039, -0.00000437, -0.00000153, 0.00000412, 0.00000631, 0.00000251, -0.00000447, -0.00000964, -0.00001003, -0.00000609, -0.00000027, 0.00000480, 0.00000709, 0.00000549, 0.00000064, -0.00000455, -0.00000639, -0.00000333, 0.00000217, 0.00000545, 0.00000390, -0.00000048, -0.00000301, -0.00000112, 0.00000299, 0.00000460, 0.00000145, -0.00000383, -0.00000628, -0.00000353, 0.00000205, 0.00000593, 0.00000579, 0.00000311, 0.00000079, -0.00000002, -0.00000032, -0.00000096, -0.00000081, 0.00000196, 0.00000702, 0.00001088, 0.00000965, 0.00000294, -0.00000508, },
		{ 0.00000903, 0.00000419, -0.00000193, -0.00000519, -0.00000399, 0.00000011, 0.00000414, 0.00000585, 0.00000448, 0.00000039, -0.00000511, -0.00000963, -0.00001030, -0.00000564, 0.00000276, 0.00001048, 0.00001295, 0.00000838, -0.00000128, -0.00001183, -0.00001911, -0.00002049, -0.00001547, -0.00000608, 0.00000333, 0.00000796, 0.00000587, 0.00000006, -0.00000362, -0.00000194, 0.00000195, 0.00000113, -0.00000754, -0.00001845, -0.00002112, -0.00001051, 0.00000623, 0.00001476, 0.00000680, -0.00001076, -0.00002105, -0.00001287, 0.00000804, 0.00002356, 0.00001995, 0.00000065, -0.00001728, -0.00001953, -0.00000738, 0.00000523, 0.00000678, -0.00000058, -0.00000462, 0.00000282, 0.00001654, 0.00002303, 0.00001439, -0.00000359, -0.00001723, -0.00001816, -0.00001012, -0.00000341, -0.00000354, -0.00000656, -0.00000498, 0.00000312, 0.00001185, 0.00001415, 0.00000922, 0.00000261, -0.00000060, -0.00000120, -0.00000370, -0.00000957, -0.00001406, -0.00001097, -0.00000021, 0.00001042, 0.00001204, 0.00000344, -0.00000743, -0.00001074, -0.00000394, 0.00000639, 0.00001125, 0.00000754, 0.00000001, -0.00000412, -0.00000198, 0.00000319, 0.00000595, 0.00000362, -0.00000217, -0.00000774, -0.00001011, -0.00000828, -0.00000298, 0.00000357, 0.00000810, 0.00000777, 0.00000247, -0.00000423, -0.00000735, -0.00000464, 0.00000129, 0.00000525, 0.00000426, 0.00000019, -0.00000244, -0.00000112, 0.00000224, 0.00000366, 0.00000136, -0.00000251, -0.00000427, -0.00000253, 0.00000084, 0.00000304, 0.00000326, 0.00000284, 0.00000296, 0.00000298, 0.00000155, -0.00000109, -0.00000246, -0.00000025, 0.00000487, 0.00000918, },
		{ -0.00000183, 0.00000222, 0.00000621, 0.00000735, 0.00000534, 0.00000201, -0.00000061, -0.00000167, -0.00000113, 0.00000081, 0.00000353, 0.00000526, 0.00000361, -0.00000221, -0.00000961, -0.00001357, -0.00001029, -0.00000055, 0.00001030, 0.00001590, 0.00001289, 0.00000254, -0.00001056, -0.00002089, -0.00002411, -0.00001867, -0.00000704, 0.00000471, 0.00001018, 0.00000728, 0.00000030, -0.00000364, -0.00000146, 0.00000244, 0.00000046, -0.00000947, -0.00001984, -0.00001980, -0.00000651, 0.00000995, 0.00001459, 0.00000217, -0.00001627, -0.00002263, -0.00000889, 0.00001429, 0.00002681, 0.00001785, -0.00000414, -0.00001977, -0.00001726, -0.00000265, 0.00000780, 0.00000468, -0.00000566, -0.00000893, 0.00000144, 0.00001723, 0.00002390, 0.00001519, -0.00000131, -0.00001256, -0.00001294, -0.00000802, -0.00000643, -0.00000970, -0.00001129, -0.00000510, 0.00000660, 0.00001485, 0.00001342, 0.00000521, -0.00000122, -0.00000105, 0.00000192, 0.00000038, -0.00000746, -0.00001494, -0.00001368, -0.00000285, 0.00000913, 0.00001234, 0.00000485, -0.00000569, -0.00000958, -0.00000404, 0.00000512, 0.00000974, 0.00000692, 0.00000060, -0.00000318, -0.00000188, 0.00000222, 0.00000498, 0.00000401, -0.00000023, -0.00000555, -0.00000940, -0.00000962, -0.00000532, 0.00000194, 0.00000814, 0.00000912, 0.00000404, -0.00000333, -0.00000737, -0.00000536, 0.00000028, 0.00000445, 0.00000415, 0.00000090, -0.00000144, -0.00000074, 0.00000153, 0.00000245, 0.00000094, -0.00000136, -0.00000226, -0.00000138, -0.00000011, 0.00000043, 0.00000082, 0.00000222, 0.00000438, 0.00000532, 0.00000344, -0.00000032, -0.00000288, },
		{ 0.00000511, 0.00000128, -0.00000198, -0.00000257, -0.00000066, 0.00000219, 0.00000465, 0.00000618, 0.00000639, 0.00000465, 0.00000107, -0.00000241, -0.00000292, 0.00000058, 0.00000536, 0.00000644, 0.00000099, -0.00000846, -0.00001529, -0.00001388, -0.00000406, 0.00000869, 0.00001706, 0.00001636, 0.00000658, -0.00000812, -0.00002137, -0.00002695, -0.00002179, -0.00000844, 0.00000546, 0.00001186, 0.00000849, 0.00000079, -0.00000315, -0.00000064, 0.00000273, -0.00000080, -0.00001175, -0.00002085, -0.00001754, -0.00000181, 0.00001339, 0.00001338, -0.00000329, -0.00002153, -0.00002300, -0.00000384, 0.00002043, 0.00002882, 0.00001446, -0.00000920, -0.00002149, -0.00001414, 0.00000212, 0.00000972, 0.00000219, -0.00001016, -0.00001203, 0.00000076, 0.00001730, 0.00002312, 0.00001447, 0.00000060, -0.00000707, -0.00000651, -0.00000509, -0.00000923, -0.00001597, -0.00001616, -0.00000535, 0.00000987, 0.00001767, 0.00001270, 0.00000138, -0.00000500, -0.00000193, 0.00000430, 0.00000406, -0.00000487, -0.00001460, -0.00001533, -0.00000542, 0.00000684, 0.00001137, 0.00000570, -0.00000347, -0.00000744, -0.00000350, 0.00000368, 0.00000755, 0.00000576, 0.00000120, -0.00000176, -0.00000124, 0.00000140, 0.00000365, 0.00000373, 0.00000117, -0.00000340, -0.00000810, -0.00001004, -0.00000701, 0.00000019, 0.00000731, 0.00000943, 0.00000512, -0.00000208, -0.00000655, -0.00000541, -0.00000065, 0.00000325, 0.00000361, 0.00000150, -0.00000022, -0.00000006, 0.00000098, 0.00000121, 0.00000034, -0.00000052, -0.00000055, -0.00000031, -0.00000076, -0.00000158, -0.00000120, 0.00000136, 0.00000488, 0.00000667, },
		{ 0.00000042, 0.00000443, 0.00000687, 0.00000632, 0.00000349, 0.00000015, -0.00000238, -0.00000348, -0.00000251, 0.00000110, 0.00000653, 0.00001075, 0.00001022, 0.00000425, -0.00000330, -0.00000656, -0.00000283, 0.00000446, 0.00000826, 0.00000402, -0.00000618, -0.00001506, -0.00001590, -0.00000742, 0.00000569, 0.00001623, 0.00001845, 0.00001055, -0.00000468, -0.00002049, -0.00002876, -0.00002453, -0.00001010, 0.00000550, 0.00001277, 0.00000931, 0.00000151, -0.00000205, 0.00000060, 0.00000284, -0.00000265, -0.00001436, -0.00002142, -0.00001435, 0.00000343, 0.00001640, 0.00001112, -0.00000935, -0.00002625, -0.00002209, 0.00000201, 0.00002609, 0.00002944, 0.00000997, -0.00001419, -0.00002228, -0.00001033, 0.00000662, 0.00001077, -0.00000066, -0.00001390, -0.00001376, 0.00000087, 0.00001683, 0.00002085, 0.00001236, 0.00000200, -0.00000120, 0.00000066, -0.00000150, -0.00001150, -0.00002180, -0.00002081, -0.00000585, 0.00001252, 0.00001995, 0.00001203, -0.00000188, -0.00000835, -0.00000310, 0.00000575, 0.00000696, -0.00000208, -0.00001310, -0.00001570, -0.00000760, 0.00000382, 0.00000918, 0.00000582, -0.00000108, -0.00000454, -0.00000232, 0.00000234, 0.00000497, 0.00000413, 0.00000163, -0.00000011, -0.00000019, 0.00000089, 0.00000223, 0.00000296, 0.00000194, -0.00000155, -0.00000645, -0.00000960, -0.00000790, -0.00000140, 0.00000584, 0.00000877, 0.00000558, -0.00000072, -0.00000509, -0.00000482, -0.00000133, 0.00000189, 0.00000276, 0.00000190, 0.00000100, 0.00000078, 0.00000067, 0.00000017, -0.00000029, -0.00000003, 0.00000062, 0.00000043, -0.00000112, -0.00000279, -0.00000253, },
		{ -0.00000126, -0.00000313, -0.00000300, -0.00000044, 0.00000315, 0.00000591, 0.00000690, 0.00000603, 0.00000324, -0.00000128, -0.00000593, -0.00000742, -0.00000298, 0.00000625, 0.00001459, 0.00001560, 0.00000777, -0.00000342, -0.00000955, -0.00000642, 0.00000237, 0.00000859, 0.00000642, -0.00000293, -0.00001270, -0.00001596, -0.00001027, 0.00000154, 0.00001341, 0.00001895, 0.00001406, -0.00000056, -0.00001832, -0.00002930, -0.00002653, -0.00001180, 0.00000481, 0.00001273, 0.00000958, 0.00000241, -0.00000027, 0.00000231, 0.00000278, -0.00000507, -0.00001721, -0.00002149, -0.00001033, 0.00000901, 0.00001883, 0.00000792, -0.00001570, -0.00003017, -0.00001995, 0.00000834, 0.00003092, 0.00002861, 0.00000467, -0.00001876, -0.00002204, -0.00000604, 0.00001055, 0.00001085, -0.00000379, -0.00001677, -0.00001412, 0.00000173, 0.00001595, 0.00001739, 0.00000912, 0.00000284, 0.00000459, 0.00000800, 0.00000251, -0.00001299, -0.00002665, -0.00002486, -0.00000662, 0.00001417, 0.00002135, 0.00001133, -0.00000433, -0.00001092, -0.00000439, 0.00000617, 0.00000882, 0.00000060, -0.00001062, -0.00001476, -0.00000912, 0.00000039, 0.00000599, 0.00000511, 0.00000116, -0.00000122, -0.00000057, 0.00000130, 0.00000232, 0.00000220, 0.00000178, 0.00000148, 0.00000111, 0.00000073, 0.00000098, 0.00000191, 0.00000209, -0.00000022, -0.00000475, -0.00000849, -0.00000794, -0.00000262, 0.00000401, 0.00000729, 0.00000535, 0.00000049, -0.00000327, -0.00000369, -0.00000164, 0.00000062, 0.00000178, 0.00000205, 0.00000200, 0.00000160, 0.00000063, -0.00000049, -0.00000076, 0.00000011, 0.00000108, 0.00000068, },
		{ -0.00000002, 0.00000080, 0.00000032, -0.00000130, -0.00000265, -0.00000257, -0.00000105, 0.00000124, 0.00000394, 0.00000679, 0.00000855, 0.00000687, 0.00000057, -0.00000774, -0.00001197, -0.00000717, 0.00000532, 0.00001744, 0.00002026, 0.00001145, -0.00000247, -0.00001139, -0.00000985, -0.00000095, 0.00000706, 0.00000777, 0.00000099, -0.00000827, -0.00001389, -0.00001223, -0.00000337, 0.00000884, 0.00001775, 0.00001672, 0.00000380, -0.00001508, -0.00002845, -0.00002743, -0.00001323, 0.00000347, 0.00001159, 0.00000911, 0.00000340, 0.00000216, 0.00000453, 0.00000262, -0.00000793, -0.00002019, -0.00002107, -0.00000563, 0.00001466, 0.00002058, 0.00000396, -0.00002198, -0.00003304, -0.00001671, 0.00001474, 0.00003460, 0.00002636, -0.00000112, -0.00002261, -0.00002075, -0.00000149, 0.00001368, 0.00000993, -0.00000706, -0.00001869, -0.00001322, 0.00000321, 0.00001475, 0.00001312, 0.00000513, 0.00000313, 0.00000987, 0.00001491, 0.00000659, -0.00001356, -0.00003007, -0.00002792, -0.00000761, 0.00001460, 0.00002157, 0.00001052, -0.00000585, -0.00001245, -0.00000556, 0.00000564, 0.00000949, 0.00000287, -0.00000744, -0.00001261, -0.00000980, -0.00000304, 0.00000211, 0.00000359, 0.00000294, 0.00000213, 0.00000155, 0.00000071, -0.00000006, 0.00000021, 0.00000161, 0.00000279, 0.00000241, 0.00000091, 0.00000008, 0.00000082, 0.00000174, 0.00000052, -0.00000324, -0.00000695, -0.00000724, -0.00000332, 0.00000212, 0.00000527, 0.00000450, 0.00000134, -0.00000142, -0.00000225, -0.00000152, -0.00000029, 0.00000090, 0.00000199, 0.00000263, 0.00000223, 0.00000080, -0.00000064, -0.00000093, },
		{ 0.00000110, -0.00000030, -0.00000071, -0.00000034, -0.00000017, -0.00000067, -0.00000134, -0.00000152, -0.00000135, -0.00000133, -0.00000098, 0.00000126, 0.00000600, 0.00001070, 0.00001054, 0.00000288, -0.00000871, -0.00001563, -0.00001098, 0.00000379, 0.00001892, 0.00002368, 0.00001501, -0.00000034, -0.00001168, -0.00001273, -0.00000533, 0.00000357, 0.00000774, 0.00000517, -0.00000207, -0.00000973, -0.00001300, -0.00000851, 0.00000295, 0.00001492, 0.00001819, 0.00000786, -0.00001110, -0.00002616, -0.00002694, -0.00001406, 0.00000163, 0.00000934, 0.00000777, 0.00000434, 0.00000513, 0.00000722, 0.00000245, -0.00001104, -0.00002311, -0.00002015, -0.00000052, 0.00002010, 0.00002157, -0.00000050, -0.00002780, -0.00003470, -0.00001257, 0.00002079, 0.00003687, 0.00002281, -0.00000704, -0.00002549, -0.00001848, 0.00000304, 0.00001585, 0.00000811, -0.00001028, -0.00001966, -0.00001129, 0.00000509, 0.00001332, 0.00000847, 0.00000086, 0.00000298, 0.00001428, 0.00002083, 0.00001038, -0.00001317, -0.00003173, -0.00002966, -0.00000869, 0.00001371, 0.00002044, 0.00000945, -0.00000643, -0.00001284, -0.00000643, 0.00000434, 0.00000900, 0.00000453, -0.00000393, -0.00000951, -0.00000958, -0.00000612, -0.00000203, 0.00000142, 0.00000402, 0.00000510, 0.00000378, 0.00000065, -0.00000186, -0.00000156, 0.00000118, 0.00000362, 0.00000351, 0.00000133, -0.00000038, -0.00000009, 0.00000107, 0.00000067, -0.00000210, -0.00000527, -0.00000601, -0.00000348, 0.00000044, 0.00000306, 0.00000320, 0.00000172, 0.00000014, -0.00000076, -0.00000101, -0.00000067, 0.00000034, 0.00000180, 0.00000281, 0.00000252, },
		{ 0.00000162, 0.00000255, 0.00000238, 0.00000141, 0.00000047, -0.00000011, -0.00000074, -0.00000165, -0.00000216, -0.00000145, 0.00000001, 0.00000042, -0.00000126, -0.00000320, -0.00000171, 0.00000470, 0.00001220, 0.00001370, 0.00000527, -0.00000876, -0.00001790, -0.00001389, 0.00000185, 0.00001878, 0.00002543, 0.00001815, 0.00000295, -0.00001021, -0.00001473, -0.00001048, -0.00000178, 0.00000614, 0.00000918, 0.00000536, -0.00000374, -0.00001239, -0.00001330, -0.00000365, 0.00001071, 0.00001820, 0.00001110, -0.00000684, -0.00002255, -0.00002482, -0.00001397, -0.00000046, 0.00000604, 0.00000550, 0.00000504, 0.00000845, 0.00001027, 0.00000235, -0.00001415, -0.00002574, -0.00001876, 0.00000473, 0.00002500, 0.00002175, -0.00000516, -0.00003274, -0.00003501, -0.00000783, 0.00002604, 0.00003752, 0.00001817, -0.00001269, -0.00002721, -0.00001538, 0.00000726, 0.00001697, 0.00000556, -0.00001323, -0.00001971, -0.00000862, 0.00000708, 0.00001171, 0.00000384, -0.00000323, 0.00000255, 0.00001756, 0.00002528, 0.00001352, -0.00001191, -0.00003148, -0.00002985, -0.00000968, 0.00001158, 0.00001793, 0.00000801, -0.00000624, -0.00001210, -0.00000682, 0.00000254, 0.00000752, 0.00000543, -0.00000048, -0.00000581, -0.00000850, -0.00000851, -0.00000599, -0.00000118, 0.00000429, 0.00000733, 0.00000582, 0.00000107, -0.00000289, -0.00000286, 0.00000059, 0.00000391, 0.00000420, 0.00000185, -0.00000040, -0.00000066, 0.00000030, 0.00000035, -0.00000142, -0.00000372, -0.00000452, -0.00000318, -0.00000082, 0.00000101, 0.00000170, 0.00000160, 0.00000116, 0.00000050, -0.00000021, -0.00000043, 0.00000029, },
		{ 0.00000068, 0.00000041, 0.00000086, 0.00000161, 0.00000198, 0.00000182, 0.00000161, 0.00000149, 0.00000079, -0.00000112, -0.00000340, -0.00000396, -0.00000168, 0.00000165, 0.00000241, -0.00000090, -0.00000510, -0.00000451, 0.00000311, 0.00001285, 0.00001584, 0.00000732, -0.00000792, -0.00001843, -0.00001545, -0.00000024, 0.00001697, 0.00002520, 0.00002051, 0.00000716, -0.00000695, -0.00001556, -0.00001596, -0.00000862, 0.00000300, 0.00001255, 0.00001334, 0.00000356, -0.00001035, -0.00001716, -0.00001022, 0.00000551, 0.00001663, 0.00001300, -0.00000280, -0.00001786, -0.00002102, -0.00001269, -0.00000250, 0.00000191, 0.00000231, 0.00000534, 0.00001184, 0.00001351, 0.00000241, -0.00001693, -0.00002782, -0.00001692, 0.00000979, 0.00002903, 0.00002111, -0.00000966, -0.00003643, -0.00003395, -0.00000283, 0.00003008, 0.00003645, 0.00001271, -0.00001772, -0.00002766, -0.00001169, 0.00001090, 0.00001701, 0.00000253, -0.00001569, -0.00001890, -0.00000556, 0.00000889, 0.00000995, -0.00000037, -0.00000669, 0.00000205, 0.00001958, 0.00002794, 0.00001570, -0.00000994, -0.00002936, -0.00002841, -0.00001040, 0.00000842, 0.00001415, 0.00000610, -0.00000549, -0.00001042, -0.00000667, 0.00000055, 0.00000532, 0.00000558, 0.00000257, -0.00000196, -0.00000674, -0.00001000, -0.00000934, -0.00000389, 0.00000372, 0.00000853, 0.00000738, 0.00000185, -0.00000305, -0.00000350, 0.00000001, 0.00000366, 0.00000436, 0.00000229, -0.00000006, -0.00000082, -0.00000035, -0.00000023, -0.00000120, -0.00000253, -0.00000309, -0.00000259, -0.00000157, -0.00000058, 0.00000028, 0.00000105, 0.00000147, 0.00000126, },
		{ 0.00000023, 0.00000104, 0.00000136, 0.00000146, 0.00000173, 0.00000205, 0.00000190, 0.00000126, 0.00000094, 0.00000163, 0.00000255, 0.00000179, -0.00000145, -0.00000515, -0.00000577, -0.00000197, 0.00000311, 0.00000425, -0.00000041, -0.00000644, -0.00000667, 0.00000154, 0.00001254, 0.00001656, 0.00000860, -0.00000635, -0.00001707, -0.00001530, -0.00000217, 0.00001362, 0.00002291, 0.00002180, 0.00001196, -0.00000211, -0.00001508, -0.00002127, -0.00001638, -0.00000148, 0.00001492, 0.00002109, 0.00001149, -0.00000703, -0.00001960, -0.00001601, -0.00000012, 0.00001353, 0.00001321, 0.00000052, -0.00001243, -0.00001560, -0.00001003, -0.00000415, -0.00000273, -0.00000165, 0.00000508, 0.00001497, 0.00001671, 0.00000271, -0.00001909, -0.00002906, -0.00001467, 0.00001434, 0.00003191, 0.00001970, -0.00001367, -0.00003857, -0.00003153, 0.00000207, 0.00003254, 0.00003364, 0.00000674, -0.00002182, -0.00002683, -0.00000764, 0.00001372, 0.00001601, -0.00000071, -0.00001745, -0.00001733, -0.00000244, 0.00001023, 0.00000806, -0.00000390, -0.00000918, 0.00000167, 0.00002034, 0.00002868, 0.00001671, -0.00000753, -0.00002560, -0.00002542, -0.00001069, 0.00000458, 0.00000938, 0.00000372, -0.00000448, -0.00000809, -0.00000598, -0.00000132, 0.00000277, 0.00000505, 0.00000494, 0.00000163, -0.00000454, -0.00001049, -0.00001175, -0.00000640, 0.00000242, 0.00000858, 0.00000820, 0.00000280, -0.00000238, -0.00000334, -0.00000039, 0.00000301, 0.00000398, 0.00000249, 0.00000046, -0.00000058, -0.00000073, -0.00000086, -0.00000134, -0.00000184, -0.00000201, -0.00000192, -0.00000179, -0.00000149, -0.00000078, },
		{ -0.00000159, -0.00000160, -0.00000124, -0.00000064, -0.00000004, 0.00000073, 0.00000194, 0.00000330, 0.00000375, 0.00000257, 0.00000063, -0.00000006, 0.00000145, 0.00000341, 0.00000267, -0.00000173, -0.00000664, -0.00000727, -0.00000224, 0.00000415, 0.00000553, 0.00000001, -0.00000707, -0.00000781, 0.00000033, 0.00001132, 0.00001559, 0.00000876, -0.00000429, -0.00001381, -0.00001327, -0.00000363, 0.00000903, 0.00001867, 0.00002181, 0.00001687, 0.00000388, -0.00001332, -0.00002592, -0.00002437, -0.00000689, 0.00001606, 0.00002784, 0.00001925, -0.00000273, -0.00002030, -0.00002033, -0.00000558, 0.00000911, 0.00001153, 0.00000268, -0.00000674, -0.00000885, -0.00000595, -0.00000508, -0.00000744, -0.00000616, 0.00000415, 0.00001748, 0.00001957, 0.00000327, -0.00002030, -0.00002920, -0.00001204, 0.00001809, 0.00003339, 0.00001757, -0.00001689, -0.00003892, -0.00002788, 0.00000650, 0.00003318, 0.00002917, 0.00000059, -0.00002473, -0.00002479, -0.00000354, 0.00001554, 0.00001408, -0.00000388, -0.00001834, -0.00001511, 0.00000045, 0.00001088, 0.00000604, -0.00000654, -0.00001048, 0.00000156, 0.00001994, 0.00002753, 0.00001646, -0.00000494, -0.00002057, -0.00002110, -0.00001041, 0.00000047, 0.00000402, 0.00000093, -0.00000349, -0.00000546, -0.00000484, -0.00000282, 0.00000024, 0.00000402, 0.00000647, 0.00000459, -0.00000219, -0.00001001, -0.00001299, -0.00000840, 0.00000058, 0.00000749, 0.00000812, 0.00000370, -0.00000104, -0.00000239, -0.00000044, 0.00000216, 0.00000314, 0.00000235, 0.00000099, -0.00000006, -0.00000075, -0.00000130, -0.00000166, -0.00000168, -0.00000149, -0.00000144, },
		{ -0.00000198, -0.00000167, -0.00000136, -0.00000114, -0.00000097, -0.00000100, -0.00000135, -0.00000155, -0.00000058, 0.00000197, 0.00000485, 0.00000575, 0.00000364, 0.00000032, -0.00000094, 0.00000111, 0.00000385, 0.00000314, -0.00000204, -0.00000765, -0.00000814, -0.00000234, 0.00000460, 0.00000592, 0.00000014, -0.00000692, -0.00000765, -0.00000024, 0.00000932, 0.00001286, 0.00000757, -0.00000204, -0.00000891, -0.00000935, -0.00000433, 0.00000366, 0.00001280, 0.00002047, 0.00002138, 0.00001045, -0.00001045, -0.00002949, -0.00003181, -0.00001267, 0.00001586, 0.00003289, 0.00002602, 0.00000210, -0.00001913, -0.00002263, -0.00001025, 0.00000379, 0.00000798, 0.00000335, -0.00000128, -0.00000119, -0.00000056, -0.00000499, -0.00001173, -0.00001086, 0.00000250, 0.00001903, 0.00002178, 0.00000407, -0.00002033, -0.00002800, -0.00000903, 0.00002085, 0.00003333, 0.00001484, -0.00001907, -0.00003740, -0.00002319, 0.00001016, 0.00003187, 0.00002325, -0.00000540, -0.00002631, -0.00002170, 0.00000033, 0.00001625, 0.00001139, -0.00000670, -0.00001826, -0.00001240, 0.00000285, 0.00001070, 0.00000391, -0.00000821, -0.00001052, 0.00000184, 0.00001858, 0.00002475, 0.00001501, -0.00000246, -0.00001477, -0.00001584, -0.00000955, -0.00000348, -0.00000148, -0.00000213, -0.00000278, -0.00000294, -0.00000344, -0.00000379, -0.00000193, 0.00000272, 0.00000710, 0.00000664, 0.00000000, -0.00000870, -0.00001299, -0.00000967, -0.00000153, 0.00000541, 0.00000712, 0.00000432, 0.00000072, -0.00000078, -0.00000004, 0.00000133, 0.00000200, 0.00000185, 0.00000131, 0.00000057, -0.00000039, -0.00000137, -0.00000197, },
		{ 0.00000027, -0.00000099, -0.00000207, -0.00000261, -0.00000252, -0.00000181, -0.00000065, 0.00000026, -0.00000004, -0.00000169, -0.00000322, -0.00000238, 0.00000148, 0.00000607, 0.00000774, 0.00000502, 0.00000053, -0.00000141, 0.00000073, 0.00000374, 0.00000295, -0.00000249, -0.00000803, -0.00000810, -0.00000212, 0.00000439, 0.00000519, -0.00000023, -0.00000605, -0.00000606, 0.00000006, 0.00000676, 0.00000848, 0.00000492, 0.00000008, -0.00000274, -0.00000377, -0.00000409, -0.00000197, 0.00000580, 0.00001782, 0.00002503, 0.00001693, -0.00000681, -0.00003166, -0.00003795, -0.00001818, 0.00001443, 0.00003576, 0.00003105, 0.00000688, -0.00001620, -0.00002258, -0.00001355, -0.00000192, 0.00000282, 0.00000237, 0.00000344, 0.00000682, 0.00000588, -0.00000368, -0.00001511, -0.00001534, 0.00000020, 0.00001936, 0.00002304, 0.00000506, -0.00001901, -0.00002531, -0.00000566, 0.00002248, 0.00003172, 0.00001163, -0.00002007, -0.00003407, -0.00001773, 0.00001277, 0.00002861, 0.00001614, -0.00001095, -0.00002653, -0.00001781, 0.00000373, 0.00001583, 0.00000817, -0.00000896, -0.00001719, -0.00000935, 0.00000461, 0.00000965, 0.00000172, -0.00000892, -0.00000936, 0.00000253, 0.00001651, 0.00002069, 0.00001254, -0.00000031, -0.00000875, -0.00001012, -0.00000813, -0.00000689, -0.00000661, -0.00000525, -0.00000253, -0.00000086, -0.00000199, -0.00000414, -0.00000346, 0.00000139, 0.00000689, 0.00000765, 0.00000180, -0.00000680, -0.00001185, -0.00001010, -0.00000363, 0.00000262, 0.00000528, 0.00000450, 0.00000257, 0.00000127, 0.00000084, 0.00000076, 0.00000080, 0.00000102, 0.00000129, 0.00000112, },
		{ 0.00000001, 0.00000084, 0.00000139, 0.00000106, -0.00000015, -0.00000183, -0.00000337, -0.00000394, -0.00000284, -0.00000034, 0.00000181, 0.00000151, -0.00000153, -0.00000470, -0.00000437, 0.00000050, 0.00000671, 0.00000936, 0.00000656, 0.00000139, -0.00000119, 0.00000046, 0.00000300, 0.00000189, -0.00000321, -0.00000774, -0.00000697, -0.00000141, 0.00000355, 0.00000324, -0.00000128, -0.00000462, -0.00000310, 0.00000135, 0.00000392, 0.00000276, 0.00000091, 0.00000181, 0.00000415, 0.00000308, -0.00000283, -0.00000729, -0.00000168, 0.00001410, 0.00002741, 0.00002261, -0.00000285, -0.00003230, -0.00004217, -0.00002277, 0.00001201, 0.00003616, 0.00003372, 0.00001102, -0.00001184, -0.00002011, -0.00001506, -0.00000740, -0.00000348, -0.00000020, 0.00000700, 0.00001456, 0.00001296, -0.00000108, -0.00001712, -0.00001913, -0.00000260, 0.00001829, 0.00002310, 0.00000611, -0.00001631, -0.00002107, -0.00000193, 0.00002301, 0.00002867, 0.00000813, -0.00001986, -0.00002913, -0.00001186, 0.00001417, 0.00002357, 0.00000824, -0.00001581, -0.00002549, -0.00001342, 0.00000643, 0.00001437, 0.00000467, -0.00001048, -0.00001520, -0.00000617, 0.00000563, 0.00000780, -0.00000047, -0.00000879, -0.00000723, 0.00000356, 0.00001401, 0.00001583, 0.00000936, 0.00000133, -0.00000307, -0.00000446, -0.00000631, -0.00000945, -0.00001093, -0.00000816, -0.00000280, 0.00000051, -0.00000072, -0.00000393, -0.00000421, 0.00000026, 0.00000602, 0.00000763, 0.00000303, -0.00000458, -0.00000980, -0.00000968, -0.00000546, -0.00000049, 0.00000282, 0.00000414, 0.00000418, 0.00000344, 0.00000211, 0.00000064, -0.00000020, },
		{ 0.00000363, 0.00000107, -0.00000074, -0.00000099, 0.00000000, 0.00000125, 0.00000181, 0.00000103, -0.00000121, -0.00000406, -0.00000569, -0.00000437, -0.00000038, 0.00000336, 0.00000338, -0.00000085, -0.00000569, -0.00000617, -0.00000082, 0.00000658, 0.00001028, 0.00000801, 0.00000291, -0.00000009, 0.00000051, 0.00000166, -0.00000013, -0.00000434, -0.00000682, -0.00000472, -0.00000010, 0.00000219, 0.00000012, -0.00000307, -0.00000282, 0.00000101, 0.00000359, 0.00000108, -0.00000386, -0.00000420, 0.00000293, 0.00001119, 0.00001065, -0.00000061, -0.00001181, -0.00000895, 0.00000964, 0.00002827, 0.00002686, 0.00000092, -0.00003145, -0.00004404, -0.00002585, 0.00000900, 0.00003409, 0.00003368, 0.00001394, -0.00000659, -0.00001546, -0.00001453, -0.00001205, -0.00001029, -0.00000410, 0.00000909, 0.00002137, 0.00002015, 0.00000273, -0.00001741, -0.00002179, -0.00000566, 0.00001580, 0.00002181, 0.00000710, -0.00001231, -0.00001533, 0.00000215, 0.00002256, 0.00002441, 0.00000452, -0.00001849, -0.00002297, -0.00000596, 0.00001426, 0.00001706, -0.00000002, -0.00001977, -0.00002337, -0.00000889, 0.00000828, 0.00001201, 0.00000116, -0.00001117, -0.00001244, -0.00000303, 0.00000590, 0.00000532, -0.00000257, -0.00000795, -0.00000441, 0.00000482, 0.00001134, 0.00001068, 0.00000583, 0.00000238, 0.00000182, 0.00000065, -0.00000427, -0.00001098, -0.00001408, -0.00001061, -0.00000357, 0.00000101, 0.00000019, -0.00000327, -0.00000418, -0.00000053, 0.00000470, 0.00000673, 0.00000362, -0.00000234, -0.00000717, -0.00000855, -0.00000682, -0.00000353, 0.00000006, 0.00000325, 0.00000527, 0.00000540, },
		{ 0.00000192, 0.00000564, 0.00000681, 0.00000517, 0.00000206, -0.00000065, -0.00000176, -0.00000112, 0.00000065, 0.00000232, 0.00000235, -0.00000023, -0.00000451, -0.00000748, -0.00000620, -0.00000083, 0.00000457, 0.00000524, 0.00000027, -0.00000593, -0.00000740, -0.00000224, 0.00000560, 0.00001018, 0.00000909, 0.00000502, 0.00000202, 0.00000106, -0.00000013, -0.00000311, -0.00000596, -0.00000541, -0.00000141, 0.00000186, 0.00000050, -0.00000397, -0.00000559, -0.00000090, 0.00000593, 0.00000664, -0.00000152, -0.00001086, -0.00001002, 0.00000334, 0.00001778, 0.00001829, 0.00000237, -0.00001510, -0.00001533, 0.00000489, 0.00002752, 0.00002919, 0.00000398, -0.00002929, -0.00004336, -0.00002698, 0.00000585, 0.00002981, 0.00003083, 0.00001520, -0.00000109, -0.00000911, -0.00001199, -0.00001533, -0.00001687, -0.00000890, 0.00000960, 0.00002670, 0.00002688, 0.00000754, -0.00001580, -0.00002294, -0.00000868, 0.00001201, 0.00001914, 0.00000787, -0.00000723, -0.00000832, 0.00000654, 0.00002134, 0.00001934, 0.00000103, -0.00001616, -0.00001608, -0.00000049, 0.00001302, 0.00000949, -0.00000817, -0.00002272, -0.00002046, -0.00000460, 0.00000916, 0.00000899, -0.00000206, -0.00001098, -0.00000911, -0.00000015, 0.00000548, 0.00000245, -0.00000445, -0.00000660, -0.00000128, 0.00000612, 0.00000873, 0.00000574, 0.00000233, 0.00000284, 0.00000554, 0.00000475, -0.00000224, -0.00001144, -0.00001586, -0.00001239, -0.00000471, 0.00000065, 0.00000061, -0.00000233, -0.00000348, -0.00000091, 0.00000318, 0.00000522, 0.00000363, -0.00000032, -0.00000435, -0.00000696, -0.00000760, -0.00000612, -0.00000262, },
		{ -0.00000779, -0.00000794, -0.00000487, 0.00000035, 0.00000519, 0.00000740, 0.00000646, 0.00000349, 0.00000016, -0.00000209, -0.00000231, -0.00000037, 0.00000245, 0.00000357, 0.00000093, -0.00000462, -0.00000899, -0.00000803, -0.00000169, 0.00000515, 0.00000668, 0.00000164, -0.00000526, -0.00000769, -0.00000347, 0.00000383, 0.00000887, 0.00000951, 0.00000751, 0.00000512, 0.00000225, -0.00000218, -0.00000690, -0.00000809, -0.00000370, 0.00000273, 0.00000441, -0.00000131, -0.00000871, -0.00000870, 0.00000092, 0.00001120, 0.00001021, -0.00000368, -0.00001766, -0.00001604, 0.00000305, 0.00002340, 0.00002534, 0.00000581, -0.00001694, -0.00002023, 0.00000033, 0.00002526, 0.00002931, 0.00000590, -0.00002617, -0.00004021, -0.00002592, 0.00000303, 0.00002383, 0.00002539, 0.00001449, 0.00000395, -0.00000178, -0.00000768, -0.00001686, -0.00002247, -0.00001400, 0.00000858, 0.00003012, 0.00003259, 0.00001300, -0.00001227, -0.00002230, -0.00001128, 0.00000723, 0.00001519, 0.00000831, -0.00000140, -0.00000035, 0.00001119, 0.00001970, 0.00001394, -0.00000209, -0.00001313, -0.00000904, 0.00000413, 0.00001054, 0.00000138, -0.00001571, -0.00002459, -0.00001712, -0.00000091, 0.00000904, 0.00000558, -0.00000474, -0.00000995, -0.00000549, 0.00000232, 0.00000448, -0.00000056, -0.00000599, -0.00000495, 0.00000181, 0.00000727, 0.00000635, 0.00000143, -0.00000078, 0.00000278, 0.00000790, 0.00000755, -0.00000045, -0.00001091, -0.00001621, -0.00001334, -0.00000603, -0.00000043, 0.00000051, -0.00000132, -0.00000234, -0.00000093, 0.00000168, 0.00000341, 0.00000322, 0.00000130, -0.00000175, -0.00000519, },
		{ 0.00000242, 0.00000031, -0.00000359, -0.00000748, -0.00000882, -0.00000636, -0.00000123, 0.00000396, 0.00000703, 0.00000727, 0.00000514, 0.00000163, -0.00000182, -0.00000332, -0.00000163, 0.00000212, 0.00000440, 0.00000203, -0.00000437, -0.00000992, -0.00000952, -0.00000282, 0.00000488, 0.00000732, 0.00000298, -0.00000368, -0.00000675, -0.00000416, 0.00000144, 0.00000630, 0.00000904, 0.00001011, 0.00000905, 0.00000415, -0.00000424, -0.00001125, -0.00001068, -0.00000190, 0.00000737, 0.00000741, -0.00000304, -0.00001369, -0.00001215, 0.00000246, 0.00001635, 0.00001392, -0.00000532, -0.00002375, -0.00002172, 0.00000218, 0.00002764, 0.00003120, 0.00000934, -0.00001722, -0.00002324, -0.00000357, 0.00002174, 0.00002718, 0.00000636, -0.00002253, -0.00003493, -0.00002269, 0.00000096, 0.00001681, 0.00001786, 0.00001177, 0.00000788, 0.00000568, -0.00000209, -0.00001647, -0.00002642, -0.00001872, 0.00000632, 0.00003142, 0.00003679, 0.00001867, -0.00000702, -0.00001978, -0.00001314, 0.00000186, 0.00001025, 0.00000833, 0.00000478, 0.00000815, 0.00001601, 0.00001799, 0.00000876, -0.00000463, -0.00000973, -0.00000252, 0.00000746, 0.00000696, -0.00000671, -0.00002219, -0.00002540, -0.00001370, 0.00000183, 0.00000796, 0.00000207, -0.00000667, -0.00000818, -0.00000183, 0.00000423, 0.00000306, -0.00000341, -0.00000710, -0.00000320, 0.00000450, 0.00000809, 0.00000432, -0.00000194, -0.00000323, 0.00000229, 0.00000882, 0.00000892, 0.00000092, -0.00000958, -0.00001526, -0.00001344, -0.00000731, -0.00000199, -0.00000002, -0.00000043, -0.00000104, -0.00000071, 0.00000038, 0.00000166, 0.00000260, },
		{ -0.00000064, 0.00000026, 0.00000202, 0.00000306, 0.00000159, -0.00000244, -0.00000681, -0.00000869, -0.00000688, -0.00000256, 0.00000213, 0.00000567, 0.00000738, 0.00000675, 0.00000360, -0.00000089, -0.00000388, -0.00000289, 0.00000135, 0.00000462, 0.00000278, -0.00000385, -0.00001007, -0.00001030, -0.00000401, 0.00000369, 0.00000684, 0.00000397, -0.00000131, -0.00000446, -0.00000402, -0.00000129, 0.00000255, 0.00000751, 0.00001251, 0.00001355, 0.00000674, -0.00000604, -0.00001580, -0.00001358, -0.00000021, 0.00001213, 0.00001065, -0.00000451, -0.00001848, -0.00001563, 0.00000359, 0.00002089, 0.00001737, -0.00000642, -0.00002868, -0.00002655, 0.00000092, 0.00003027, 0.00003537, 0.00001254, -0.00001607, -0.00002414, -0.00000644, 0.00001735, 0.00002300, 0.00000525, -0.00001880, -0.00002807, -0.00001755, -0.00000008, 0.00000948, 0.00000895, 0.00000719, 0.00001015, 0.00001236, 0.00000411, -0.00001421, -0.00002822, -0.00002240, 0.00000323, 0.00003062, 0.00003916, 0.00002408, -0.00000043, -0.00001546, -0.00001396, -0.00000356, 0.00000472, 0.00000793, 0.00001088, 0.00001668, 0.00002087, 0.00001660, 0.00000434, -0.00000637, -0.00000630, 0.00000287, 0.00000915, 0.00000247, -0.00001426, -0.00002722, -0.00002519, -0.00001057, 0.00000340, 0.00000598, -0.00000125, -0.00000771, -0.00000580, 0.00000160, 0.00000549, 0.00000138, -0.00000585, -0.00000769, -0.00000153, 0.00000654, 0.00000845, 0.00000271, -0.00000420, -0.00000485, 0.00000150, 0.00000842, 0.00000887, 0.00000176, -0.00000772, -0.00001331, -0.00001276, -0.00000835, -0.00000369, -0.00000082, 0.00000022, 0.00000011, -0.00000045, },
		{ 0.00000056, 0.00000084, -0.00000035, -0.00000134, -0.00000056, 0.00000172, 0.00000328, 0.00000198, -0.00000195, -0.00000600, -0.00000764, -0.00000637, -0.00000343, -0.00000003, 0.00000349, 0.00000669, 0.00000800, 0.00000576, 0.00000064, -0.00000376, -0.00000382, 0.00000027, 0.00000405, 0.00000287, -0.00000325, -0.00000932, -0.00001007, -0.00000498, 0.00000161, 0.00000502, 0.00000426, 0.00000162, -0.00000082, -0.00000280, -0.00000401, -0.00000214, 0.00000489, 0.00001436, 0.00001822, 0.00000987, -0.00000735, -0.00002014, -0.00001656, 0.00000119, 0.00001659, 0.00001384, -0.00000561, -0.00002265, -0.00001879, 0.00000429, 0.00002444, 0.00002014, -0.00000706, -0.00003214, -0.00003007, -0.00000045, 0.00003120, 0.00003754, 0.00001506, -0.00001372, -0.00002297, -0.00000800, 0.00001255, 0.00001719, 0.00000264, -0.00001541, -0.00002033, -0.00001100, 0.00000007, 0.00000260, -0.00000047, 0.00000116, 0.00001044, 0.00001744, 0.00001012, -0.00001041, -0.00002765, -0.00002447, -0.00000011, 0.00002798, 0.00003956, 0.00002877, 0.00000698, -0.00000958, -0.00001353, -0.00000849, -0.00000090, 0.00000717, 0.00001650, 0.00002473, 0.00002564, 0.00001589, 0.00000118, -0.00000717, -0.00000322, 0.00000656, 0.00000894, -0.00000267, -0.00002078, -0.00003053, -0.00002409, -0.00000804, 0.00000363, 0.00000326, -0.00000417, -0.00000783, -0.00000302, 0.00000457, 0.00000609, -0.00000035, -0.00000769, -0.00000776, -0.00000010, 0.00000775, 0.00000828, 0.00000155, -0.00000530, -0.00000560, 0.00000053, 0.00000690, 0.00000761, 0.00000204, -0.00000563, -0.00001073, -0.00001148, -0.00000899, -0.00000519, -0.00000165, },
		{ -0.00000914, -0.00000617, -0.00000225, 0.00000063, 0.00000098, -0.00000062, -0.00000180, -0.00000072, 0.00000185, 0.00000321, 0.00000153, -0.00000220, -0.00000522, -0.00000590, -0.00000492, -0.00000369, -0.00000221, 0.00000076, 0.00000523, 0.00000859, 0.00000774, 0.00000262, -0.00000280, -0.00000409, -0.00000089, 0.00000263, 0.00000203, -0.00000280, -0.00000769, -0.00000861, -0.00000542, -0.00000116, 0.00000181, 0.00000360, 0.00000476, 0.00000398, -0.00000040, -0.00000634, -0.00000738, 0.00000125, 0.00001539, 0.00002263, 0.00001331, -0.00000802, -0.00002386, -0.00001933, 0.00000223, 0.00002038, 0.00001667, -0.00000628, -0.00002585, -0.00002131, 0.00000458, 0.00002671, 0.00002189, -0.00000737, -0.00003399, -0.00003199, -0.00000169, 0.00003055, 0.00003761, 0.00001659, -0.00001055, -0.00001996, -0.00000817, 0.00000780, 0.00001034, -0.00000121, -0.00001269, -0.00001246, -0.00000369, 0.00000133, -0.00000320, -0.00000947, -0.00000574, 0.00000864, 0.00002027, 0.00001515, -0.00000559, -0.00002474, -0.00002454, -0.00000313, 0.00002399, 0.00003809, 0.00003239, 0.00001461, -0.00000258, -0.00001177, -0.00001238, -0.00000602, 0.00000623, 0.00002130, 0.00003184, 0.00003016, 0.00001613, -0.00000033, -0.00000694, -0.00000082, 0.00000813, 0.00000670, -0.00000822, -0.00002590, -0.00003198, -0.00002227, -0.00000633, 0.00000253, -0.00000001, -0.00000651, -0.00000706, -0.00000005, 0.00000694, 0.00000609, -0.00000193, -0.00000878, -0.00000735, 0.00000096, 0.00000807, 0.00000761, 0.00000085, -0.00000535, -0.00000557, -0.00000055, 0.00000458, 0.00000547, 0.00000184, -0.00000360, -0.00000797, -0.00000989, },
		{ -0.00000185, -0.00000548, -0.00000828, -0.00000881, -0.00000640, -0.00000238, 0.00000049, 0.00000045, -0.00000137, -0.00000209, -0.00000025, 0.00000247, 0.00000300, 0.00000040, -0.00000312, -0.00000463, -0.00000377, -0.00000276, -0.00000330, -0.00000405, -0.00000211, 0.00000313, 0.00000833, 0.00000912, 0.00000475, -0.00000095, -0.00000342, -0.00000185, 0.00000044, 0.00000009, -0.00000278, -0.00000535, -0.00000583, -0.00000507, -0.00000433, -0.00000267, 0.00000178, 0.00000771, 0.00000961, 0.00000316, -0.00000795, -0.00001273, -0.00000318, 0.00001541, 0.00002633, 0.00001675, -0.00000796, -0.00002660, -0.00002160, 0.00000285, 0.00002317, 0.00001885, -0.00000655, -0.00002783, -0.00002288, 0.00000456, 0.00002756, 0.00002239, -0.00000752, -0.00003423, -0.00003218, -0.00000257, 0.00002854, 0.00003566, 0.00001698, -0.00000698, -0.00001554, -0.00000704, 0.00000354, 0.00000313, -0.00000585, -0.00001082, -0.00000521, 0.00000362, 0.00000349, -0.00000743, -0.00001721, -0.00001279, 0.00000494, 0.00002044, 0.00001847, -0.00000043, -0.00001984, -0.00002247, -0.00000528, 0.00001928, 0.00003509, 0.00003472, 0.00002185, 0.00000500, -0.00000873, -0.00001480, -0.00001011, 0.00000531, 0.00002508, 0.00003767, 0.00003433, 0.00001748, 0.00000001, -0.00000574, 0.00000061, 0.00000732, 0.00000247, -0.00001387, -0.00002938, -0.00003160, -0.00001995, -0.00000557, 0.00000019, -0.00000358, -0.00000816, -0.00000558, 0.00000284, 0.00000859, 0.00000561, -0.00000318, -0.00000907, -0.00000655, 0.00000154, 0.00000752, 0.00000653, 0.00000058, -0.00000453, -0.00000498, -0.00000169, 0.00000179, 0.00000288, 0.00000132, },
		{ -0.00000111, 0.00000027, 0.00000069, -0.00000055, -0.00000362, -0.00000699, -0.00000810, -0.00000578, -0.00000187, 0.00000030, -0.00000066, -0.00000263, -0.00000235, 0.00000070, 0.00000350, 0.00000277, -0.00000115, -0.00000451, -0.00000429, -0.00000159, -0.00000030, -0.00000236, -0.00000528, -0.00000466, 0.00000069, 0.00000708, 0.00000952, 0.00000668, 0.00000169, -0.00000158, -0.00000229, -0.00000235, -0.00000303, -0.00000342, -0.00000254, -0.00000181, -0.00000371, -0.00000754, -0.00000817, -0.00000124, 0.00001010, 0.00001561, 0.00000772, -0.00000857, -0.00001769, -0.00000806, 0.00001433, 0.00002894, 0.00001984, -0.00000721, -0.00002810, -0.00002308, 0.00000313, 0.00002477, 0.00002013, -0.00000650, -0.00002847, -0.00002333, 0.00000433, 0.00002699, 0.00002155, -0.00000765, -0.00003302, -0.00003067, -0.00000291, 0.00002551, 0.00003199, 0.00001618, -0.00000344, -0.00001027, -0.00000483, 0.00000008, -0.00000372, -0.00001072, -0.00000984, 0.00000081, 0.00001016, 0.00000617, -0.00000986, -0.00002299, -0.00001927, -0.00000025, 0.00001785, 0.00001953, 0.00000429, -0.00001356, -0.00001840, -0.00000612, 0.00001459, 0.00003107, 0.00003572, 0.00002817, 0.00001259, -0.00000461, -0.00001544, -0.00001268, 0.00000469, 0.00002780, 0.00004203, 0.00003806, 0.00002001, 0.00000220, -0.00000375, 0.00000083, 0.00000410, -0.00000353, -0.00001934, -0.00003114, -0.00002960, -0.00001738, -0.00000578, -0.00000310, -0.00000718, -0.00000911, -0.00000363, 0.00000539, 0.00000949, 0.00000484, -0.00000391, -0.00000859, -0.00000554, 0.00000158, 0.00000621, 0.00000522, 0.00000073, -0.00000310, -0.00000409, -0.00000290, },
		{ -0.00000135, -0.00000322, -0.00000420, -0.00000383, -0.00000194, 0.00000019, 0.00000023, -0.00000265, -0.00000626, -0.00000716, -0.00000435, -0.00000067, 0.00000019, -0.00000218, -0.00000428, -0.00000265, 0.00000188, 0.00000473, 0.00000256, -0.00000278, -0.00000603, -0.00000416, 0.00000029, 0.00000204, -0.00000108, -0.00000567, -0.00000642, -0.00000173, 0.00000488, 0.00000863, 0.00000800, 0.00000490, 0.00000150, -0.00000192, -0.00000543, -0.00000722, -0.00000493, 0.00000040, 0.00000322, -0.00000128, -0.00001041, -0.00001427, -0.00000536, 0.00001160, 0.00002146, 0.00001296, -0.00000807, -0.00002178, -0.00001299, 0.00001220, 0.00003017, 0.00002224, -0.00000589, -0.00002822, -0.00002355, 0.00000315, 0.00002513, 0.00002038, -0.00000623, -0.00002780, -0.00002260, 0.00000403, 0.00002515, 0.00001941, -0.00000788, -0.00003061, -0.00002769, -0.00000267, 0.00002184, 0.00002704, 0.00001431, -0.00000032, -0.00000479, -0.00000193, -0.00000236, -0.00000958, -0.00001524, -0.00000965, 0.00000514, 0.00001525, 0.00000890, -0.00001048, -0.00002635, -0.00002449, -0.00000635, 0.00001275, 0.00001804, 0.00000785, -0.00000673, -0.00001276, -0.00000544, 0.00001060, 0.00002670, 0.00003556, 0.00003320, 0.00001964, 0.00000031, -0.00001420, -0.00001340, 0.00000466, 0.00002956, 0.00004497, 0.00004136, 0.00002364, 0.00000599, -0.00000129, -0.00000032, -0.00000139, -0.00001089, -0.00002434, -0.00003126, -0.00002634, -0.00001486, -0.00000687, -0.00000700, -0.00001050, -0.00000941, -0.00000153, 0.00000732, 0.00000966, 0.00000401, -0.00000397, -0.00000741, -0.00000448, 0.00000103, 0.00000434, 0.00000387, 0.00000127, },
		{ 0.00000211, 0.00000269, 0.00000216, 0.00000042, -0.00000264, -0.00000562, -0.00000613, -0.00000343, 0.00000000, 0.00000052, -0.00000267, -0.00000624, -0.00000616, -0.00000230, 0.00000114, 0.00000029, -0.00000382, -0.00000608, -0.00000301, 0.00000298, 0.00000584, 0.00000235, -0.00000419, -0.00000727, -0.00000410, 0.00000162, 0.00000375, 0.00000022, -0.00000514, -0.00000698, -0.00000370, 0.00000188, 0.00000627, 0.00000834, 0.00000838, 0.00000576, -0.00000051, -0.00000845, -0.00001226, -0.00000736, 0.00000312, 0.00000891, 0.00000218, -0.00001259, -0.00002048, -0.00001034, 0.00001198, 0.00002663, 0.00001847, -0.00000645, -0.00002464, -0.00001754, 0.00000919, 0.00002985, 0.00002364, -0.00000419, -0.00002695, -0.00002288, 0.00000306, 0.00002430, 0.00001956, -0.00000586, -0.00002596, -0.00002074, 0.00000375, 0.00002227, 0.00001619, -0.00000828, -0.00002733, -0.00002358, -0.00000187, 0.00001790, 0.00002134, 0.00001160, 0.00000210, 0.00000032, 0.00000121, -0.00000375, -0.00001396, -0.00001885, -0.00001001, 0.00000757, 0.00001839, 0.00001119, -0.00000952, -0.00002712, -0.00002790, -0.00001263, 0.00000565, 0.00001400, 0.00000964, -0.00000025, -0.00000622, -0.00000327, 0.00000787, 0.00002266, 0.00003459, 0.00003679, 0.00002573, 0.00000569, -0.00001115, -0.00001213, 0.00000543, 0.00003063, 0.00004670, 0.00004433, 0.00002817, 0.00001093, 0.00000117, -0.00000296, -0.00000876, -0.00001905, -0.00002856, -0.00002995, -0.00002232, -0.00001268, -0.00000870, -0.00001104, -0.00001322, -0.00000914, 0.00000036, 0.00000840, 0.00000916, 0.00000335, -0.00000326, -0.00000566, -0.00000358, -0.00000011, },
		{ -0.00000350, -0.00000298, -0.00000179, -0.00000025, 0.00000185, 0.00000332, 0.00000196, -0.00000258, -0.00000719, -0.00000785, -0.00000401, 0.00000026, 0.00000039, -0.00000363, -0.00000695, -0.00000529, 0.00000006, 0.00000331, 0.00000064, -0.00000524, -0.00000769, -0.00000340, 0.00000371, 0.00000645, 0.00000202, -0.00000508, -0.00000779, -0.00000387, 0.00000223, 0.00000442, 0.00000119, -0.00000375, -0.00000607, -0.00000481, -0.00000163, 0.00000240, 0.00000738, 0.00001171, 0.00001097, 0.00000205, -0.00001102, -0.00001776, -0.00001068, 0.00000528, 0.00001479, 0.00000649, -0.00001382, -0.00002627, -0.00001581, 0.00001116, 0.00003067, 0.00002377, -0.00000385, -0.00002601, -0.00002128, 0.00000559, 0.00002800, 0.00002386, -0.00000237, -0.00002446, -0.00002107, 0.00000300, 0.00002247, 0.00001777, -0.00000548, -0.00002318, -0.00001796, 0.00000356, 0.00001867, 0.00001219, -0.00000881, -0.00002354, -0.00001879, -0.00000065, 0.00001404, 0.00001543, 0.00000838, 0.00000366, 0.00000452, 0.00000413, -0.00000417, -0.00001655, -0.00002109, -0.00001062, 0.00000811, 0.00001930, 0.00001259, -0.00000739, -0.00002545, -0.00002920, -0.00001837, -0.00000266, 0.00000772, 0.00000924, 0.00000495, 0.00000036, 0.00000010, 0.00000674, 0.00001959, 0.00003328, 0.00003903, 0.00003062, 0.00001118, -0.00000655, -0.00000895, 0.00000720, 0.00003140, 0.00004770, 0.00004717, 0.00003334, 0.00001636, 0.00000305, -0.00000716, -0.00001750, -0.00002732, -0.00003172, -0.00002753, -0.00001813, -0.00001116, -0.00001105, -0.00001474, -0.00001507, -0.00000842, 0.00000172, 0.00000843, 0.00000810, 0.00000306, -0.00000172, },
		{ 0.00000657, 0.00000330, 0.00000062, -0.00000109, -0.00000279, -0.00000390, -0.00000252, 0.00000146, 0.00000465, 0.00000306, -0.00000315, -0.00000886, -0.00000888, -0.00000364, 0.00000098, -0.00000006, -0.00000530, -0.00000825, -0.00000468, 0.00000234, 0.00000549, 0.00000122, -0.00000610, -0.00000867, -0.00000366, 0.00000380, 0.00000615, 0.00000137, -0.00000528, -0.00000718, -0.00000318, 0.00000206, 0.00000370, 0.00000142, -0.00000168, -0.00000355, -0.00000469, -0.00000528, -0.00000280, 0.00000495, 0.00001448, 0.00001675, 0.00000572, -0.00001280, -0.00002329, -0.00001470, 0.00000663, 0.00002038, 0.00001132, -0.00001396, -0.00003115, -0.00002132, 0.00000921, 0.00003320, 0.00002836, -0.00000053, -0.00002580, -0.00002389, 0.00000175, 0.00002478, 0.00002281, -0.00000067, -0.00002101, -0.00001825, 0.00000307, 0.00001992, 0.00001522, -0.00000515, -0.00001978, -0.00001455, 0.00000346, 0.00001468, 0.00000781, -0.00000940, -0.00001955, -0.00001378, 0.00000078, 0.00001050, 0.00000983, 0.00000500, 0.00000431, 0.00000746, 0.00000640, -0.00000382, -0.00001726, -0.00002169, -0.00001114, 0.00000699, 0.00001798, 0.00001278, -0.00000463, -0.00002175, -0.00002831, -0.00002292, -0.00001130, -0.00000020, 0.00000653, 0.00000811, 0.00000605, 0.00000410, 0.00000725, 0.00001796, 0.00003217, 0.00004023, 0.00003431, 0.00001651, -0.00000080, -0.00000415, 0.00001001, 0.00003236, 0.00004857, 0.00005018, 0.00003881, 0.00002150, 0.00000367, -0.00001292, -0.00002696, -0.00003494, -0.00003356, -0.00002438, -0.00001438, -0.00001059, -0.00001369, -0.00001763, -0.00001579, -0.00000738, 0.00000224, 0.00000728, },
		{ -0.00000616, 0.00000176, 0.00000493, 0.00000469, 0.00000409, 0.00000363, 0.00000138, -0.00000302, -0.00000627, -0.00000450, 0.00000154, 0.00000598, 0.00000356, -0.00000435, -0.00001051, -0.00000920, -0.00000244, 0.00000204, -0.00000071, -0.00000736, -0.00000984, -0.00000436, 0.00000415, 0.00000721, 0.00000187, -0.00000612, -0.00000860, -0.00000358, 0.00000310, 0.00000457, 0.00000015, -0.00000474, -0.00000516, -0.00000173, 0.00000123, 0.00000143, 0.00000059, 0.00000077, 0.00000051, -0.00000311, -0.00000864, -0.00000902, 0.00000102, 0.00001630, 0.00002265, 0.00001033, -0.00001352, -0.00002833, -0.00001911, 0.00000703, 0.00002519, 0.00001628, -0.00001299, -0.00003471, -0.00002639, 0.00000631, 0.00003405, 0.00003184, 0.00000313, -0.00002411, -0.00002515, -0.00000197, 0.00002051, 0.00002056, 0.00000070, -0.00001696, -0.00001466, 0.00000334, 0.00001695, 0.00001220, -0.00000487, -0.00001608, -0.00001087, 0.00000339, 0.00001062, 0.00000346, -0.00000990, -0.00001565, -0.00000900, 0.00000218, 0.00000742, 0.00000494, 0.00000184, 0.00000414, 0.00000893, 0.00000773, -0.00000298, -0.00001623, -0.00002056, -0.00001128, 0.00000461, 0.00001466, 0.00001159, -0.00000182, -0.00001666, -0.00002543, -0.00002579, -0.00001931, -0.00000896, 0.00000168, 0.00000868, 0.00000997, 0.00000801, 0.00000913, 0.00001798, 0.00003177, 0.00004092, 0.00003707, 0.00002154, 0.00000562, 0.00000183, 0.00001383, 0.00003399, 0.00005002, 0.00005369, 0.00004421, 0.00002548, 0.00000236, -0.00002023, -0.00003646, -0.00004117, -0.00003387, -0.00002091, -0.00001165, -0.00001120, -0.00001638, -0.00001934, -0.00001521, },
		{ -0.00001892, -0.00001961, -0.00001329, -0.00000486, 0.00000020, 0.00000147, 0.00000258, 0.00000540, 0.00000709, 0.00000375, -0.00000362, -0.00000864, -0.00000602, 0.00000200, 0.00000709, 0.00000336, -0.00000606, -0.00001196, -0.00000880, -0.00000064, 0.00000322, -0.00000151, -0.00000941, -0.00001131, -0.00000428, 0.00000511, 0.00000796, 0.00000235, -0.00000514, -0.00000707, -0.00000288, 0.00000159, 0.00000147, -0.00000189, -0.00000352, -0.00000156, 0.00000074, -0.00000006, -0.00000239, -0.00000154, 0.00000325, 0.00000587, 0.00000002, -0.00001130, -0.00001580, -0.00000426, 0.00001686, 0.00002813, 0.00001556, -0.00001305, -0.00003243, -0.00002352, 0.00000644, 0.00002884, 0.00002092, -0.00001106, -0.00003669, -0.00003057, 0.00000279, 0.00003317, 0.00003386, 0.00000674, -0.00002117, -0.00002499, -0.00000523, 0.00001559, 0.00001732, 0.00000157, -0.00001272, -0.00001063, 0.00000381, 0.00001390, 0.00000906, -0.00000460, -0.00001236, -0.00000728, 0.00000326, 0.00000676, -0.00000046, -0.00001015, -0.00001201, -0.00000483, 0.00000331, 0.00000487, 0.00000104, -0.00000083, 0.00000331, 0.00000893, 0.00000793, -0.00000194, -0.00001377, -0.00001784, -0.00001083, 0.00000148, 0.00000983, 0.00000904, 0.00000052, -0.00001094, -0.00002101, -0.00002673, -0.00002586, -0.00001763, -0.00000485, 0.00000646, 0.00001139, 0.00001096, 0.00001178, 0.00001955, 0.00003247, 0.00004174, 0.00003941, 0.00002631, 0.00001228, 0.00000843, 0.00001850, 0.00003673, 0.00005276, 0.00005806, 0.00004917, 0.00002750, -0.00000150, -0.00002899, -0.00004534, -0.00004534, -0.00003250, -0.00001749, -0.00001041, -0.00001314, },
		{ -0.00001443, -0.00001093, -0.00001642, -0.00002116, -0.00001836, -0.00001010, -0.00000356, -0.00000234, -0.00000286, 0.00000034, 0.00000705, 0.00001068, 0.00000586, -0.00000440, -0.00001070, -0.00000694, 0.00000268, 0.00000770, 0.00000238, -0.00000805, -0.00001294, -0.00000768, 0.00000143, 0.00000417, -0.00000244, -0.00001106, -0.00001217, -0.00000427, 0.00000496, 0.00000728, 0.00000236, -0.00000312, -0.00000382, -0.00000126, -0.00000062, -0.00000327, -0.00000496, -0.00000181, 0.00000357, 0.00000440, -0.00000151, -0.00000754, -0.00000504, 0.00000538, 0.00001211, 0.00000459, -0.00001294, -0.00002259, -0.00001057, 0.00001600, 0.00003268, 0.00002096, -0.00001138, -0.00003518, -0.00002749, 0.00000500, 0.00003103, 0.00002481, -0.00000840, -0.00003698, -0.00003348, -0.00000098, 0.00003073, 0.00003426, 0.00000988, -0.00001736, -0.00002352, -0.00000776, 0.00001049, 0.00001341, 0.00000188, -0.00000867, -0.00000655, 0.00000442, 0.00001104, 0.00000616, -0.00000424, -0.00000889, -0.00000412, 0.00000294, 0.00000330, -0.00000365, -0.00001000, -0.00000876, -0.00000151, 0.00000397, 0.00000281, -0.00000177, -0.00000280, 0.00000205, 0.00000763, 0.00000700, -0.00000098, -0.00001033, -0.00001387, -0.00000969, -0.00000187, 0.00000409, 0.00000538, 0.00000198, -0.00000539, -0.00001567, -0.00002575, -0.00003030, -0.00002530, -0.00001234, 0.00000159, 0.00000984, 0.00001210, 0.00001435, 0.00002221, 0.00003443, 0.00004336, 0.00004200, 0.00003103, 0.00001883, 0.00001510, 0.00002382, 0.00004095, 0.00005739, 0.00006356, 0.00005335, 0.00002683, -0.00000837, -0.00003906, -0.00005302, -0.00004698, -0.00002941, },
		{ -0.00005906, -0.00004583, -0.00002470, -0.00001197, -0.00001331, -0.00002092, -0.00002299, -0.00001569, -0.00000590, -0.00000236, -0.00000562, -0.00000771, -0.00000189, 0.00000876, 0.00001397, 0.00000752, -0.00000515, -0.00001206, -0.00000712, 0.00000336, 0.00000749, 0.00000057, -0.00001004, -0.00001312, -0.00000585, 0.00000341, 0.00000449, -0.00000358, -0.00001196, -0.00001195, -0.00000407, 0.00000357, 0.00000481, 0.00000155, -0.00000020, 0.00000127, 0.00000150, -0.00000330, -0.00000955, -0.00000920, 0.00000010, 0.00001000, 0.00000929, -0.00000282, -0.00001365, -0.00000983, 0.00000682, 0.00001871, 0.00001034, -0.00001332, -0.00002884, -0.00001747, 0.00001370, 0.00003585, 0.00002606, -0.00000867, -0.00003632, -0.00003059, 0.00000295, 0.00003166, 0.00002760, -0.00000535, -0.00003564, -0.00003489, -0.00000459, 0.00002702, 0.00003303, 0.00001223, -0.00001313, -0.00002096, -0.00000937, 0.00000566, 0.00000921, 0.00000165, -0.00000516, -0.00000281, 0.00000504, 0.00000859, 0.00000379, -0.00000369, -0.00000583, -0.00000166, 0.00000234, 0.00000034, -0.00000589, -0.00000932, -0.00000591, 0.00000084, 0.00000404, 0.00000115, -0.00000351, -0.00000400, 0.00000059, 0.00000535, 0.00000511, -0.00000031, -0.00000645, -0.00000914, -0.00000792, -0.00000494, -0.00000184, 0.00000098, 0.00000232, -0.00000073, -0.00001013, -0.00002311, -0.00003227, -0.00003115, -0.00001992, -0.00000540, 0.00000522, 0.00001071, 0.00001583, 0.00002520, 0.00003756, 0.00004631, 0.00004560, 0.00003610, 0.00002509, 0.00002141, 0.00002958, 0.00004689, 0.00006434, 0.00007033, 0.00005641, 0.00002296, -0.00001850, -0.00005022, },
		{ 0.00001561, -0.00003182, -0.00006211, -0.00006316, -0.00004192, -0.00001864, -0.00001028, -0.00001737, -0.00002632, -0.00002431, -0.00001192, -0.00000108, -0.00000133, -0.00000926, -0.00001256, -0.00000396, 0.00001021, 0.00001647, 0.00000857, -0.00000560, -0.00001233, -0.00000644, 0.00000377, 0.00000612, -0.00000211, -0.00001171, -0.00001215, -0.00000327, 0.00000497, 0.00000377, -0.00000506, -0.00001186, -0.00001022, -0.00000336, 0.00000098, 0.00000032, -0.00000040, 0.00000335, 0.00000811, 0.00000558, -0.00000611, -0.00001708, -0.00001460, 0.00000189, 0.00001732, 0.00001528, -0.00000365, -0.00002021, -0.00001565, 0.00000733, 0.00002510, 0.00001681, -0.00001237, -0.00003401, -0.00002440, 0.00001012, 0.00003734, 0.00003035, -0.00000519, -0.00003574, -0.00003246, 0.00000059, 0.00003077, 0.00002903, -0.00000227, -0.00003290, -0.00003470, -0.00000768, 0.00002245, 0.00003032, 0.00001354, -0.00000895, -0.00001766, -0.00001004, 0.00000149, 0.00000514, 0.00000101, -0.00000240, 0.00000024, 0.00000552, 0.00000668, 0.00000217, -0.00000286, -0.00000328, -0.00000005, 0.00000137, -0.00000206, -0.00000710, -0.00000806, -0.00000345, 0.00000223, 0.00000350, -0.00000022, -0.00000436, -0.00000448, -0.00000087, 0.00000247, 0.00000252, -0.00000003, -0.00000262, -0.00000421, -0.00000570, -0.00000737, -0.00000728, -0.00000366, 0.00000152, 0.00000252, -0.00000513, -0.00001929, -0.00003175, -0.00003459, -0.00002667, -0.00001370, -0.00000220, 0.00000630, 0.00001517, 0.00002751, 0.00004144, 0.00005093, 0.00005092, 0.00004207, 0.00003116, 0.00002714, 0.00003564, 0.00005464, 0.00007373, 0.00007827, 0.00005801, },
		{ 0.00008537, 0.00008699, 0.00005781, 0.00000478, -0.00004793, -0.00007426, -0.00006520, -0.00003560, -0.00001169, -0.00000942, -0.00002271, -0.00003213, -0.00002507, -0.00000750, 0.00000377, -0.00000060, -0.00001282, -0.00001683, -0.00000569, 0.00001103, 0.00001764, 0.00000879, -0.00000549, -0.00001111, -0.00000478, 0.00000367, 0.00000329, -0.00000569, -0.00001278, -0.00000972, 0.00000010, 0.00000582, 0.00000164, -0.00000708, -0.00001069, -0.00000669, -0.00000183, -0.00000260, -0.00000618, -0.00000374, 0.00000714, 0.00001638, 0.00001099, -0.00000868, -0.00002539, -0.00002101, 0.00000325, 0.00002500, 0.00002207, -0.00000379, -0.00002665, -0.00002207, 0.00000680, 0.00003071, 0.00002348, -0.00001017, -0.00003770, -0.00003080, 0.00000554, 0.00003697, 0.00003341, -0.00000133, -0.00003351, -0.00003287, -0.00000172, 0.00002856, 0.00002901, 0.00000050, -0.00002911, -0.00003300, -0.00000996, 0.00001749, 0.00002643, 0.00001370, -0.00000524, -0.00001403, -0.00000983, -0.00000172, 0.00000159, 0.00000013, -0.00000054, 0.00000236, 0.00000571, 0.00000531, 0.00000143, -0.00000168, -0.00000124, 0.00000067, 0.00000001, -0.00000392, -0.00000732, -0.00000625, -0.00000132, 0.00000280, 0.00000242, -0.00000141, -0.00000455, -0.00000440, -0.00000218, -0.00000058, -0.00000037, -0.00000016, 0.00000068, 0.00000033, -0.00000331, -0.00000891, -0.00001167, -0.00000799, -0.00000026, 0.00000403, -0.00000128, -0.00001496, -0.00002902, -0.00003532, -0.00003173, -0.00002225, -0.00001175, -0.00000131, 0.00001144, 0.00002796, 0.00004533, 0.00005721, 0.00005846, 0.00004950, 0.00003737, 0.00003244, 0.00004206, 0.00006418, },
		{ 0.00003782, 0.00004914, 0.00007540, 0.00009871, 0.00009576, 0.00005538, -0.00000926, -0.00006606, -0.00008600, -0.00006519, -0.00002750, -0.00000450, -0.00000945, -0.00002872, -0.00003767, -0.00002518, -0.00000300, 0.00000800, -0.00000030, -0.00001580, -0.00001985, -0.00000685, 0.00001088, 0.00001697, 0.00000798, -0.00000460, -0.00000806, -0.00000205, 0.00000285, -0.00000127, -0.00001020, -0.00001304, -0.00000557, 0.00000433, 0.00000584, -0.00000215, -0.00000983, -0.00000850, -0.00000126, 0.00000080, -0.00000681, -0.00001447, -0.00000860, 0.00001071, 0.00002558, 0.00001759, -0.00001063, -0.00003389, -0.00002810, 0.00000392, 0.00003242, 0.00002924, -0.00000309, -0.00003237, -0.00002857, 0.00000526, 0.00003505, 0.00002974, -0.00000698, -0.00003963, -0.00003611, 0.00000039, 0.00003479, 0.00003490, 0.00000249, -0.00002987, -0.00003171, -0.00000363, 0.00002537, 0.00002760, 0.00000266, -0.00002469, -0.00003001, -0.00001126, 0.00001260, 0.00002175, 0.00001276, -0.00000236, -0.00001050, -0.00000892, -0.00000383, -0.00000112, -0.00000078, 0.00000044, 0.00000341, 0.00000550, 0.00000445, 0.00000153, -0.00000015, 0.00000031, 0.00000057, -0.00000167, -0.00000531, -0.00000671, -0.00000399, 0.00000056, 0.00000277, 0.00000098, -0.00000249, -0.00000437, -0.00000400, -0.00000326, -0.00000344, -0.00000318, -0.00000063, 0.00000315, 0.00000398, -0.00000108, -0.00000953, -0.00001460, -0.00001151, -0.00000267, 0.00000378, 0.00000097, -0.00001080, -0.00002469, -0.00003340, -0.00003442, -0.00002991, -0.00002248, -0.00001191, 0.00000396, 0.00002543, 0.00004825, 0.00006476, 0.00006842, 0.00005893, 0.00004434, },
		{ 0.00008053, 0.00007067, 0.00005286, 0.00004425, 0.00005750, 0.00008820, 0.00011286, 0.00010347, 0.00005021, -0.00002606, -0.00008512, -0.00009653, -0.00006325, -0.00001853, 0.00000210, -0.00001039, -0.00003466, -0.00004216, -0.00002452, 0.00000094, 0.00001083, -0.00000065, -0.00001773, -0.00002099, -0.00000718, 0.00000949, 0.00001402, 0.00000592, -0.00000280, -0.00000293, 0.00000185, 0.00000121, -0.00000767, -0.00001565, -0.00001237, 0.00000040, 0.00000947, 0.00000498, -0.00000767, -0.00001348, -0.00000551, 0.00000596, 0.00000472, -0.00001118, -0.00002409, -0.00001491, 0.00001357, 0.00003505, 0.00002506, -0.00001165, -0.00004189, -0.00003541, 0.00000375, 0.00003895, 0.00003623, -0.00000156, -0.00003686, -0.00003455, 0.00000288, 0.00003778, 0.00003502, -0.00000314, -0.00003971, -0.00003992, -0.00000487, 0.00003099, 0.00003466, 0.00000586, -0.00002519, -0.00002905, -0.00000486, 0.00002160, 0.00002499, 0.00000405, -0.00002008, -0.00002608, -0.00001154, 0.00000817, 0.00001672, 0.00001086, -0.00000049, -0.00000742, -0.00000756, -0.00000481, -0.00000282, -0.00000152, 0.00000065, 0.00000341, 0.00000480, 0.00000396, 0.00000236, 0.00000168, 0.00000146, -0.00000016, -0.00000357, -0.00000628, -0.00000550, -0.00000149, 0.00000221, 0.00000238, -0.00000060, -0.00000350, -0.00000411, -0.00000353, -0.00000412, -0.00000579, -0.00000556, -0.00000128, 0.00000461, 0.00000638, 0.00000066, -0.00000933, -0.00001590, -0.00001383, -0.00000527, 0.00000202, 0.00000141, -0.00000748, -0.00001954, -0.00002922, -0.00003434, -0.00003559, -0.00003319, -0.00002488, -0.00000755, 0.00001894, 0.00004907, 0.00007273, },
		{ 0.00000784, 0.00004661, 0.00007992, 0.00009400, 0.00008478, 0.00006385, 0.00005308, 0.00006816, 0.00010252, 0.00012671, 0.00010873, 0.00004168, -0.00004497, -0.00010375, -0.00010488, -0.00005953, -0.00000973, 0.00000715, -0.00001231, -0.00003972, -0.00004473, -0.00002293, 0.00000373, 0.00001152, -0.00000190, -0.00001824, -0.00001966, -0.00000639, 0.00000673, 0.00000846, 0.00000243, -0.00000008, 0.00000436, 0.00000696, -0.00000123, -0.00001590, -0.00002201, -0.00001080, 0.00000815, 0.00001553, 0.00000338, -0.00001482, -0.00001814, -0.00000207, 0.00001464, 0.00000996, -0.00001521, -0.00003441, -0.00002246, 0.00001531, 0.00004403, 0.00003294, -0.00001153, -0.00004871, -0.00004238, 0.00000271, 0.00004404, 0.00004247, 0.00000067, -0.00003972, -0.00003945, -0.00000005, 0.00003872, 0.00003883, 0.00000089, -0.00003806, -0.00004191, -0.00000975, 0.00002594, 0.00003269, 0.00000842, -0.00001992, -0.00002511, -0.00000525, 0.00001767, 0.00002151, 0.00000460, -0.00001568, -0.00002161, -0.00001090, 0.00000449, 0.00001180, 0.00000827, 0.00000028, -0.00000510, -0.00000603, -0.00000480, -0.00000344, -0.00000192, 0.00000028, 0.00000250, 0.00000364, 0.00000371, 0.00000368, 0.00000365, 0.00000229, -0.00000130, -0.00000549, -0.00000691, -0.00000395, 0.00000104, 0.00000367, 0.00000189, -0.00000207, -0.00000439, -0.00000396, -0.00000325, -0.00000481, -0.00000748, -0.00000723, -0.00000196, 0.00000502, 0.00000733, 0.00000164, -0.00000855, -0.00001567, -0.00001473, -0.00000758, -0.00000079, 0.00000011, -0.00000554, -0.00001448, -0.00002344, -0.00003142, -0.00003842, -0.00004263, -0.00003928, -0.00002292, },
		{ -0.00004964, -0.00005387, -0.00004143, -0.00000803, 0.00003982, 0.00008477, 0.00010751, 0.00010088, 0.00007821, 0.00006599, 0.00008248, 0.00011845, 0.00013900, 0.00010989, 0.00002902, -0.00006533, -0.00012046, -0.00010995, -0.00005420, -0.00000220, 0.00000960, -0.00001532, -0.00004315, -0.00004451, -0.00002016, 0.00000490, 0.00000941, -0.00000434, -0.00001712, -0.00001543, -0.00000420, 0.00000262, 0.00000016, -0.00000265, 0.00000343, 0.00001371, 0.00001328, -0.00000435, -0.00002577, -0.00002920, -0.00000847, 0.00001745, 0.00002246, 0.00000127, -0.00002329, -0.00002378, 0.00000141, 0.00002425, 0.00001640, -0.00001841, -0.00004466, -0.00003082, 0.00001562, 0.00005174, 0.00004063, -0.00001022, -0.00005375, -0.00004842, 0.00000091, 0.00004726, 0.00004738, 0.00000334, -0.00004074, -0.00004278, -0.00000314, 0.00003789, 0.00004083, 0.00000466, -0.00003493, -0.00004199, -0.00001381, 0.00002011, 0.00002916, 0.00000993, -0.00001456, -0.00002026, -0.00000474, 0.00001400, 0.00001757, 0.00000437, -0.00001181, -0.00001706, -0.00000953, 0.00000175, 0.00000736, 0.00000531, 0.00000004, -0.00000368, -0.00000460, -0.00000403, -0.00000309, -0.00000189, -0.00000043, 0.00000093, 0.00000210, 0.00000354, 0.00000521, 0.00000560, 0.00000287, -0.00000256, -0.00000721, -0.00000722, -0.00000231, 0.00000334, 0.00000490, 0.00000150, -0.00000321, -0.00000513, -0.00000405, -0.00000333, -0.00000541, -0.00000845, -0.00000806, -0.00000250, 0.00000453, 0.00000686, 0.00000170, -0.00000753, -0.00001417, -0.00001420, -0.00000917, -0.00000403, -0.00000261, -0.00000528, -0.00001036, -0.00001696, -0.00002595, -0.00003783, },
		{ -0.00001074, -0.00001857, -0.00003370, -0.00005329, -0.00006730, -0.00006188, -0.00002828, 0.00002793, 0.00008555, 0.00011924, 0.00011815, 0.00009668, 0.00008483, 0.00010211, 0.00013628, 0.00014847, 0.00010518, 0.00001137, -0.00008648, -0.00013372, -0.00011063, -0.00004734, 0.00000299, 0.00000845, -0.00001961, -0.00004433, -0.00004072, -0.00001594, 0.00000413, 0.00000396, -0.00000828, -0.00001433, -0.00000808, -0.00000036, -0.00000266, -0.00001080, -0.00000942, 0.00000745, 0.00002484, 0.00002074, -0.00000789, -0.00003688, -0.00003707, -0.00000563, 0.00002789, 0.00003012, -0.00000103, -0.00003259, -0.00003027, 0.00000449, 0.00003411, 0.00002373, -0.00002036, -0.00005402, -0.00003945, 0.00001436, 0.00005748, 0.00004747, -0.00000782, -0.00005655, -0.00005292, -0.00000141, 0.00004838, 0.00005048, 0.00000611, -0.00003989, -0.00004420, -0.00000598, 0.00003546, 0.00004086, 0.00000776, -0.00003072, -0.00004024, -0.00001674, 0.00001402, 0.00002441, 0.00001027, -0.00000956, -0.00001493, -0.00000340, 0.00001089, 0.00001357, 0.00000353, -0.00000869, -0.00001279, -0.00000770, -0.00000005, 0.00000367, 0.00000231, -0.00000100, -0.00000317, -0.00000348, -0.00000279, -0.00000198, -0.00000144, -0.00000124, -0.00000098, 0.00000036, 0.00000333, 0.00000663, 0.00000727, 0.00000323, -0.00000372, -0.00000853, -0.00000722, -0.00000081, 0.00000519, 0.00000586, 0.00000133, -0.00000383, -0.00000560, -0.00000441, -0.00000385, -0.00000603, -0.00000879, -0.00000809, -0.00000282, 0.00000336, 0.00000519, 0.00000081, -0.00000660, -0.00001188, -0.00001243, -0.00000972, -0.00000703, -0.00000621, -0.00000678, -0.00000790, },
		{ -0.00001002, -0.00000980, -0.00000754, -0.00000570, -0.00001016, -0.00002637, -0.00005303, -0.00007820, -0.00008266, -0.00005193, 0.00001056, 0.00008051, 0.00012697, 0.00013527, 0.00011970, 0.00011144, 0.00012895, 0.00015654, 0.00015403, 0.00009291, -0.00001221, -0.00010793, -0.00014215, -0.00010583, -0.00003893, 0.00000501, 0.00000280, -0.00002544, -0.00004285, -0.00003279, -0.00000996, 0.00000136, -0.00000508, -0.00001399, -0.00001006, 0.00000237, 0.00000532, -0.00000874, -0.00002408, -0.00001784, 0.00001162, 0.00003723, 0.00002913, -0.00001153, -0.00004869, -0.00004538, -0.00000258, 0.00003888, 0.00003823, -0.00000318, -0.00004208, -0.00003731, 0.00000679, 0.00004346, 0.00003149, -0.00002080, -0.00006171, -0.00004767, 0.00001156, 0.00006072, 0.00005283, -0.00000458, -0.00005685, -0.00005541, -0.00000392, 0.00004735, 0.00005144, 0.00000862, -0.00003731, -0.00004354, -0.00000821, 0.00003178, 0.00003895, 0.00000985, -0.00002591, -0.00003690, -0.00001835, 0.00000816, 0.00001888, 0.00000948, -0.00000530, -0.00000960, -0.00000142, 0.00000855, 0.00000991, 0.00000232, -0.00000637, -0.00000913, -0.00000573, -0.00000100, 0.00000088, -0.00000042, -0.00000255, -0.00000346, -0.00000279, -0.00000139, -0.00000041, -0.00000064, -0.00000196, -0.00000288, -0.00000138, 0.00000301, 0.00000769, 0.00000846, 0.00000337, -0.00000456, -0.00000926, -0.00000689, 0.00000042, 0.00000644, 0.00000648, 0.00000143, -0.00000386, -0.00000574, -0.00000495, -0.00000477, -0.00000671, -0.00000867, -0.00000748, -0.00000288, 0.00000179, 0.00000269, -0.00000088, -0.00000605, -0.00000935, -0.00000981, -0.00000910, -0.00000918, },
		{ -0.00000682, -0.00000735, -0.00001001, -0.00001334, -0.00001388, -0.00000939, -0.00000255, -0.00000179, -0.00001664, -0.00004882, -0.00008538, -0.00010187, -0.00007739, -0.00001214, 0.00006809, 0.00012827, 0.00015048, 0.00014729, 0.00014747, 0.00016487, 0.00017998, 0.00015491, 0.00007160, -0.00004261, -0.00012939, -0.00014468, -0.00009464, -0.00002884, 0.00000330, -0.00000802, -0.00003312, -0.00003861, -0.00002042, -0.00000193, -0.00000324, -0.00001771, -0.00002167, -0.00000469, 0.00001559, 0.00001287, -0.00001513, -0.00003911, -0.00002776, 0.00001546, 0.00005018, 0.00003818, -0.00001489, -0.00006049, -0.00005377, 0.00000032, 0.00004973, 0.00004645, -0.00000485, -0.00005105, -0.00004449, 0.00000801, 0.00005155, 0.00003913, -0.00001965, -0.00006706, -0.00005481, 0.00000743, 0.00006110, 0.00005614, -0.00000088, -0.00005463, -0.00005555, -0.00000625, 0.00004434, 0.00005012, 0.00001050, -0.00003334, -0.00004086, -0.00000952, 0.00002729, 0.00003533, 0.00001074, -0.00002097, -0.00003237, -0.00001861, 0.00000299, 0.00001306, 0.00000774, -0.00000205, -0.00000470, 0.00000091, 0.00000706, 0.00000691, 0.00000102, -0.00000483, -0.00000627, -0.00000390, -0.00000130, -0.00000099, -0.00000265, -0.00000429, -0.00000432, -0.00000257, -0.00000013, 0.00000126, 0.00000034, -0.00000242, -0.00000446, -0.00000287, 0.00000256, 0.00000818, 0.00000897, 0.00000327, -0.00000499, -0.00000928, -0.00000623, 0.00000132, 0.00000701, 0.00000671, 0.00000175, -0.00000333, -0.00000550, -0.00000553, -0.00000594, -0.00000747, -0.00000832, -0.00000655, -0.00000278, 0.00000010, -0.00000018, -0.00000310, -0.00000607, -0.00000713, },
		{ -0.00000548, -0.00000673, -0.00000569, -0.00000404, -0.00000474, -0.00000928, -0.00001558, -0.00001837, -0.00001317, -0.00000173, 0.00000547, -0.00000571, -0.00004111, -0.00008798, -0.00011750, -0.00010263, -0.00003944, 0.00004707, 0.00012070, 0.00016171, 0.00017900, 0.00019410, 0.00021164, 0.00020757, 0.00015079, 0.00004020, -0.00008063, -0.00015083, -0.00014067, -0.00007642, -0.00001685, -0.00000231, -0.00002432, -0.00004296, -0.00003183, -0.00000366, 0.00000834, -0.00000924, -0.00003356, -0.00003139, 0.00000125, 0.00003098, 0.00002221, -0.00002126, -0.00005509, -0.00003889, 0.00001852, 0.00006289, 0.00004745, -0.00001759, -0.00007151, -0.00006181, 0.00000277, 0.00005969, 0.00005431, -0.00000577, -0.00005877, -0.00005129, 0.00000800, 0.00005769, 0.00004600, -0.00001698, -0.00006962, -0.00006023, 0.00000234, 0.00005855, 0.00005699, 0.00000286, -0.00005009, -0.00005323, -0.00000803, 0.00003971, 0.00004660, 0.00001148, -0.00002842, -0.00003639, -0.00000974, 0.00002245, 0.00003039, 0.00001040, -0.00001635, -0.00002714, -0.00001763, -0.00000118, 0.00000745, 0.00000531, 0.00000008, -0.00000062, 0.00000328, 0.00000635, 0.00000474, -0.00000010, -0.00000390, -0.00000427, -0.00000244, -0.00000120, -0.00000205, -0.00000426, -0.00000590, -0.00000545, -0.00000274, 0.00000078, 0.00000270, 0.00000130, -0.00000259, -0.00000548, -0.00000391, 0.00000205, 0.00000800, 0.00000869, 0.00000289, -0.00000498, -0.00000857, -0.00000521, 0.00000192, 0.00000693, 0.00000653, 0.00000216, -0.00000237, -0.00000487, -0.00000596, -0.00000709, -0.00000824, -0.00000798, -0.00000565, -0.00000268, -0.00000146, -0.00000291, },
		{ -0.00000282, -0.00000277, -0.00000507, -0.00000761, -0.00000792, -0.00000533, -0.00000197, -0.00000168, -0.00000707, -0.00001628, -0.00002250, -0.00001827, -0.00000331, 0.00001063, 0.00000493, -0.00003089, -0.00008556, -0.00012762, -0.00012529, -0.00007007, 0.00001679, 0.00010210, 0.00016668, 0.00021383, 0.00025191, 0.00027061, 0.00024038, 0.00014189, -0.00000172, -0.00012678, -0.00017248, -0.00013004, -0.00005093, -0.00000266, -0.00001155, -0.00004602, -0.00005519, -0.00002303, 0.00001704, 0.00002095, -0.00001600, -0.00005194, -0.00004306, 0.00000708, 0.00004768, 0.00003306, -0.00002653, -0.00007106, -0.00005076, 0.00002034, 0.00007444, 0.00005646, -0.00001932, -0.00008095, -0.00006902, 0.00000451, 0.00006804, 0.00006129, -0.00000579, -0.00006460, -0.00005716, 0.00000677, 0.00006137, 0.00005151, -0.00001307, -0.00006918, -0.00006341, -0.00000322, 0.00005326, 0.00005518, 0.00000616, -0.00004366, -0.00004853, -0.00000896, 0.00003392, 0.00004116, 0.00001139, -0.00002305, -0.00003055, -0.00000884, 0.00001772, 0.00002462, 0.00000894, -0.00001241, -0.00002170, -0.00001568, -0.00000417, 0.00000249, 0.00000254, 0.00000111, 0.00000240, 0.00000538, 0.00000627, 0.00000347, -0.00000084, -0.00000339, -0.00000307, -0.00000148, -0.00000097, -0.00000251, -0.00000521, -0.00000711, -0.00000653, -0.00000316, 0.00000123, 0.00000363, 0.00000201, -0.00000248, -0.00000581, -0.00000435, 0.00000153, 0.00000717, 0.00000760, 0.00000219, -0.00000461, -0.00000722, -0.00000385, 0.00000230, 0.00000634, 0.00000595, 0.00000253, -0.00000120, -0.00000392, -0.00000603, -0.00000794, -0.00000889, -0.00000785, -0.00000516, },
		{ -0.00000925, -0.00000806, -0.00000542, -0.00000344, -0.00000379, -0.00000636, -0.00000910, -0.00000942, -0.00000612, -0.00000100, 0.00000127, -0.00000370, -0.00001529, -0.00002554, -0.00002377, -0.00000693, 0.00001294, 0.00001370, -0.00001956, -0.00007817, -0.00013063, -0.00014293, -0.00010226, -0.00002278, 0.00007077, 0.00016318, 0.00025027, 0.00032068, 0.00034255, 0.00027945, 0.00012897, -0.00005383, -0.00018117, -0.00019476, -0.00011333, -0.00001845, 0.00001399, -0.00002373, -0.00007255, -0.00006994, -0.00001303, 0.00004085, 0.00003577, -0.00002276, -0.00007186, -0.00005637, 0.00001213, 0.00006462, 0.00004497, -0.00003038, -0.00008598, -0.00006279, 0.00002058, 0.00008396, 0.00006460, -0.00001987, -0.00008809, -0.00007485, 0.00000539, 0.00007415, 0.00006689, -0.00000490, -0.00006804, -0.00006156, 0.00000449, 0.00006231, 0.00005513, -0.00000832, -0.00006577, -0.00006403, -0.00000872, 0.00004564, 0.00005076, 0.00000863, -0.00003592, -0.00004180, -0.00000886, 0.00002756, 0.00003426, 0.00001020, -0.00001771, -0.00002388, -0.00000695, 0.00001348, 0.00001855, 0.00000661, -0.00000935, -0.00001654, -0.00001310, -0.00000594, -0.00000151, -0.00000021, 0.00000120, 0.00000425, 0.00000695, 0.00000658, 0.00000304, -0.00000105, -0.00000305, -0.00000250, -0.00000109, -0.00000084, -0.00000261, -0.00000557, -0.00000777, -0.00000728, -0.00000365, 0.00000119, 0.00000388, 0.00000230, -0.00000218, -0.00000544, -0.00000411, 0.00000110, 0.00000579, 0.00000581, 0.00000117, -0.00000401, -0.00000541, -0.00000221, 0.00000260, 0.00000544, 0.00000506, 0.00000268, -0.00000009, -0.00000278, -0.00000561, -0.00000818, },
		{ -0.00000161, -0.00000460, -0.00000755, -0.00000914, -0.00000862, -0.00000663, -0.00000476, -0.00000460, -0.00000664, -0.00000965, -0.00001090, -0.00000790, -0.00000134, 0.00000358, 0.00000021, -0.00001274, -0.00002684, -0.00002856, -0.00001186, 0.00001200, 0.00001904, -0.00000884, -0.00006642, -0.00012535, -0.00015322, -0.00013391, -0.00007092, 0.00002574, 0.00014924, 0.00028642, 0.00039933, 0.00042745, 0.00032566, 0.00011337, -0.00011499, -0.00024341, -0.00021827, -0.00009168, 0.00002021, 0.00003329, -0.00003779, -0.00010285, -0.00008711, -0.00000281, 0.00006655, 0.00005246, -0.00002868, -0.00009210, -0.00007081, 0.00001576, 0.00008065, 0.00005730, -0.00003239, -0.00009883, -0.00007430, 0.00001903, 0.00009065, 0.00007130, -0.00001915, -0.00009235, -0.00007883, 0.00000540, 0.00007756, 0.00007062, -0.00000321, -0.00006879, -0.00006403, 0.00000145, 0.00006045, 0.00005649, -0.00000324, -0.00005973, -0.00006198, -0.00001363, 0.00003634, 0.00004400, 0.00000998, -0.00002755, -0.00003354, -0.00000767, 0.00002118, 0.00002649, 0.00000803, -0.00001285, -0.00001697, -0.00000431, 0.00001000, 0.00001272, 0.00000375, -0.00000725, -0.00001203, -0.00001025, -0.00000662, -0.00000435, -0.00000262, 0.00000060, 0.00000497, 0.00000781, 0.00000701, 0.00000327, -0.00000072, -0.00000268, -0.00000235, -0.00000119, -0.00000097, -0.00000261, -0.00000551, -0.00000781, -0.00000749, -0.00000399, 0.00000074, 0.00000340, 0.00000206, -0.00000183, -0.00000450, -0.00000326, 0.00000085, 0.00000407, 0.00000351, -0.00000013, -0.00000336, -0.00000339, -0.00000043, 0.00000295, 0.00000450, 0.00000400, 0.00000249, 0.00000069, },
		{ 0.00000293, 0.00000190, 0.00000091, -0.00000060, -0.00000302, -0.00000593, -0.00000839, -0.00000948, -0.00000883, -0.00000694, -0.00000537, -0.00000601, -0.00000914, -0.00001198, -0.00001028, -0.00000293, 0.00000473, 0.00000387, -0.00000907, -0.00002597, -0.00003148, -0.00001707, 0.00000777, 0.00001963, -0.00000057, -0.00005137, -0.00011127, -0.00015422, -0.00016272, -0.00012619, -0.00003304, 0.00012341, 0.00032009, 0.00048587, 0.00052434, 0.00037954, 0.00009684, -0.00018320, -0.00031248, -0.00024360, -0.00006678, 0.00006361, 0.00005521, -0.00005237, -0.00013544, -0.00010636, 0.00000655, 0.00009266, 0.00007045, -0.00003300, -0.00011130, -0.00008568, 0.00001745, 0.00009460, 0.00006929, -0.00003229, -0.00010868, -0.00008455, 0.00001569, 0.00009392, 0.00007596, -0.00001724, -0.00009334, -0.00008053, 0.00000464, 0.00007802, 0.00007211, -0.00000093, -0.00006679, -0.00006426, -0.00000196, 0.00005601, 0.00005542, 0.00000167, -0.00005159, -0.00005740, -0.00001751, 0.00002610, 0.00003540, 0.00001003, -0.00001923, -0.00002441, -0.00000548, 0.00001528, 0.00001849, 0.00000513, -0.00000878, -0.00001039, -0.00000128, 0.00000740, 0.00000757, 0.00000077, -0.00000604, -0.00000842, -0.00000750, -0.00000642, -0.00000601, -0.00000444, -0.00000039, 0.00000472, 0.00000788, 0.00000731, 0.00000389, 0.00000008, -0.00000213, -0.00000237, -0.00000165, -0.00000144, -0.00000271, -0.00000519, -0.00000728, -0.00000706, -0.00000403, 0.00000004, 0.00000227, 0.00000125, -0.00000158, -0.00000319, -0.00000196, 0.00000080, 0.00000226, 0.00000099, -0.00000160, -0.00000284, -0.00000147, 0.00000135, 0.00000345, 0.00000378, },
		{ 0.00000291, 0.00000416, 0.00000350, 0.00000204, 0.00000092, 0.00000043, 0.00000006, -0.00000099, -0.00000333, -0.00000690, -0.00001045, -0.00001187, -0.00000997, -0.00000633, -0.00000476, -0.00000762, -0.00001232, -0.00001269, -0.00000547, 0.00000435, 0.00000637, -0.00000501, -0.00002278, -0.00003145, -0.00002131, 0.00000062, 0.00001453, 0.00000351, -0.00003445, -0.00008854, -0.00014461, -0.00018646, -0.00018652, -0.00010464, 0.00008494, 0.00034902, 0.00057747, 0.00063128, 0.00044112, 0.00008143, -0.00025566, -0.00038670, -0.00027124, -0.00004073, 0.00010980, 0.00007949, -0.00006600, -0.00016847, -0.00012709, 0.00001403, 0.00011751, 0.00008890, -0.00003509, -0.00012808, -0.00010010, 0.00001692, 0.00010541, 0.00008014, -0.00003004, -0.00011480, -0.00009283, 0.00001070, 0.00009340, 0.00007812, -0.00001434, -0.00009092, -0.00007966, 0.00000332, 0.00007551, 0.00007114, 0.00000162, -0.00006222, -0.00006211, -0.00000529, 0.00004942, 0.00005198, 0.00000592, -0.00004208, -0.00005067, -0.00002004, 0.00001574, 0.00002565, 0.00000881, -0.00001160, -0.00001516, -0.00000254, 0.00001027, 0.00001092, 0.00000182, -0.00000571, -0.00000466, 0.00000176, 0.00000567, 0.00000341, -0.00000197, -0.00000553, -0.00000583, -0.00000515, -0.00000564, -0.00000658, -0.00000552, -0.00000148, 0.00000376, 0.00000722, 0.00000729, 0.00000464, 0.00000121, -0.00000132, -0.00000234, -0.00000228, -0.00000221, -0.00000305, -0.00000482, -0.00000630, -0.00000601, -0.00000365, -0.00000072, 0.00000072, -0.00000002, -0.00000153, -0.00000181, -0.00000047, 0.00000093, 0.00000060, -0.00000140, -0.00000309, -0.00000257, 0.00000006, },
		{ -0.00000441, -0.00000264, 0.00000094, 0.00000407, 0.00000504, 0.00000379, 0.00000150, -0.00000034, -0.00000074, 0.00000027, 0.00000131, 0.00000008, -0.00000471, -0.00001133, -0.00001544, -0.00001370, -0.00000769, -0.00000335, -0.00000537, -0.00001163, -0.00001441, -0.00000843, 0.00000225, 0.00000684, -0.00000146, -0.00001741, -0.00002762, -0.00002330, -0.00000869, 0.00000330, 0.00000191, -0.00001736, -0.00005804, -0.00012380, -0.00020318, -0.00024931, -0.00018710, 0.00003396, 0.00037109, 0.00067064, 0.00074531, 0.00050981, 0.00006931, -0.00032894, -0.00046372, -0.00030145, -0.00001589, 0.00015637, 0.00010551, -0.00007723, -0.00019988, -0.00014839, 0.00001874, 0.00013941, 0.00010680, -0.00003455, -0.00014118, -0.00011312, 0.00001409, 0.00011222, 0.00008900, -0.00002583, -0.00011675, -0.00009852, 0.00000441, 0.00008901, 0.00007746, -0.00001078, -0.00008525, -0.00007610, 0.00000173, 0.00007029, 0.00006768, 0.00000410, -0.00005548, -0.00005765, -0.00000815, 0.00004131, 0.00004644, 0.00000913, -0.00003198, -0.00004237, -0.00002107, 0.00000602, 0.00001551, 0.00000647, -0.00000516, -0.00000649, 0.00000082, 0.00000638, 0.00000431, -0.00000149, -0.00000366, -0.00000013, 0.00000442, 0.00000467, 0.00000042, -0.00000414, -0.00000549, -0.00000422, -0.00000338, -0.00000458, -0.00000629, -0.00000586, -0.00000241, 0.00000239, 0.00000596, 0.00000683, 0.00000524, 0.00000244, -0.00000028, -0.00000210, -0.00000289, -0.00000318, -0.00000368, -0.00000454, -0.00000508, -0.00000449, -0.00000286, -0.00000134, -0.00000099, -0.00000157, -0.00000175, -0.00000066, 0.00000090, 0.00000115, -0.00000070, -0.00000334, },
		{ 0.00000130, -0.00000151, -0.00000455, -0.00000533, -0.00000304, 0.00000104, 0.00000463, 0.00000599, 0.00000464, 0.00000143, -0.00000171, -0.00000249, -0.00000000, 0.00000361, 0.00000396, -0.00000194, -0.00001187, -0.00001903, -0.00001782, -0.00000964, -0.00000237, -0.00000286, -0.00000979, -0.00001474, -0.00001110, -0.00000150, 0.00000454, 0.00000058, -0.00001030, -0.00001946, -0.00002188, -0.00001915, -0.00001394, -0.00000637, -0.00000181, -0.00002129, -0.00009212, -0.00021139, -0.00031167, -0.00027752, -0.00002845, 0.00038453, 0.00076145, 0.00086258, 0.00058430, 0.00006249, -0.00039921, -0.00054063, -0.00033409, 0.00000539, 0.00020058, 0.00013234, -0.00008478, -0.00022753, -0.00016911, 0.00002011, 0.00015680, 0.00012302, -0.00003129, -0.00014960, -0.00012377, 0.00000916, 0.00011451, 0.00009515, -0.00002005, -0.00011439, -0.00010116, -0.00000268, 0.00008100, 0.00007384, -0.00000696, -0.00007672, -0.00006996, 0.00000017, 0.00006280, 0.00006189, 0.00000617, -0.00004714, -0.00005116, -0.00001017, 0.00003239, 0.00003926, 0.00001105, -0.00002209, -0.00003318, -0.00002064, -0.00000240, 0.00000575, 0.00000333, -0.00000027, 0.00000097, 0.00000419, 0.00000370, -0.00000092, -0.00000444, -0.00000255, 0.00000298, 0.00000637, 0.00000420, -0.00000140, -0.00000552, -0.00000561, -0.00000341, -0.00000228, -0.00000352, -0.00000543, -0.00000553, -0.00000300, 0.00000093, 0.00000432, 0.00000591, 0.00000549, 0.00000355, 0.00000090, -0.00000158, -0.00000329, -0.00000418, -0.00000451, -0.00000444, -0.00000383, -0.00000272, -0.00000174, -0.00000168, -0.00000252, -0.00000312, -0.00000223, 0.00000001, 0.00000180, },
		{ -0.00000288, 0.00000000, 0.00000195, 0.00000121, -0.00000179, -0.00000483, -0.00000568, -0.00000365, 0.00000034, 0.00000449, 0.00000686, 0.00000595, 0.00000183, -0.00000298, -0.00000452, -0.00000069, 0.00000565, 0.00000783, 0.00000112, -0.00001186, -0.00002207, -0.00002185, -0.00001227, -0.00000242, -0.00000069, -0.00000683, -0.00001305, -0.00001266, -0.00000662, -0.00000103, 0.00000016, -0.00000215, -0.00000689, -0.00001613, -0.00002957, -0.00003649, -0.00002175, 0.00001059, 0.00001964, -0.00005077, -0.00021026, -0.00037061, -0.00037219, -0.00010023, 0.00038809, 0.00084575, 0.00097858, 0.00066257, 0.00006266, -0.00046246, -0.00061403, -0.00036862, 0.00002093, 0.00023963, 0.00015870, -0.00008773, -0.00024944, -0.00018794, 0.00001790, 0.00016839, 0.00013642, -0.00002550, -0.00015264, -0.00013118, 0.00000259, 0.00011208, 0.00009801, -0.00001325, -0.00010794, -0.00010050, -0.00001002, 0.00006988, 0.00006737, -0.00000332, -0.00006597, -0.00006153, -0.00000107, 0.00005368, 0.00005414, 0.00000758, -0.00003789, -0.00004310, -0.00001114, 0.00002340, 0.00003107, 0.00001162, -0.00001313, -0.00002386, -0.00001891, -0.00000904, -0.00000289, -0.00000023, 0.00000293, 0.00000673, 0.00000717, 0.00000216, -0.00000454, -0.00000668, -0.00000218, 0.00000463, 0.00000741, 0.00000399, -0.00000218, -0.00000603, -0.00000564, -0.00000317, -0.00000181, -0.00000267, -0.00000429, -0.00000471, -0.00000317, -0.00000036, 0.00000255, 0.00000461, 0.00000526, 0.00000433, 0.00000206, -0.00000079, -0.00000337, -0.00000502, -0.00000542, -0.00000454, -0.00000275, -0.00000099, -0.00000046, -0.00000164, -0.00000357, -0.00000437, },
		{ -0.00000120, -0.00000389, -0.00000501, -0.00000356, -0.00000075, 0.00000114, 0.00000068, -0.00000164, -0.00000415, -0.00000532, -0.00000431, -0.00000104, 0.00000364, 0.00000744, 0.00000745, 0.00000264, -0.00000393, -0.00000645, -0.00000158, 0.00000718, 0.00001114, 0.00000405, -0.00001116, -0.00002393, -0.00002519, -0.00001553, -0.00000407, 0.00000044, -0.00000296, -0.00000886, -0.00001234, -0.00001261, -0.00001007, -0.00000349, 0.00000616, 0.00000974, -0.00000556, -0.00003874, -0.00006314, -0.00004397, 0.00001857, 0.00006234, -0.00000181, -0.00019973, -0.00042332, -0.00046688, -0.00017842, 0.00038123, 0.00091955, 0.00108834, 0.00074192, 0.00007097, -0.00051487, -0.00068031, -0.00040399, 0.00002901, 0.00027080, 0.00018305, -0.00008559, -0.00026394, -0.00020351, 0.00001226, 0.00017330, 0.00014590, -0.00001769, -0.00015008, -0.00013466, -0.00000501, 0.00010514, 0.00009725, -0.00000610, -0.00009794, -0.00009653, -0.00001701, 0.00005645, 0.00005837, -0.00000025, -0.00005378, -0.00005131, -0.00000173, 0.00004366, 0.00004497, 0.00000813, -0.00002843, -0.00003406, -0.00001098, 0.00001505, 0.00002252, 0.00001091, -0.00000565, -0.00001513, -0.00001622, -0.00001363, -0.00000985, -0.00000375, 0.00000450, 0.00001053, 0.00000940, 0.00000155, -0.00000650, -0.00000801, -0.00000231, 0.00000494, 0.00000743, 0.00000378, -0.00000215, -0.00000569, -0.00000537, -0.00000320, -0.00000183, -0.00000214, -0.00000317, -0.00000363, -0.00000294, -0.00000129, 0.00000089, 0.00000306, 0.00000454, 0.00000463, 0.00000303, 0.00000014, -0.00000310, -0.00000553, -0.00000619, -0.00000476, -0.00000199, 0.00000041, 0.00000077, },
		{ -0.00000162, 0.00000125, 0.00000168, -0.00000046, -0.00000336, -0.00000482, -0.00000408, -0.00000218, -0.00000070, -0.00000044, -0.00000122, -0.00000263, -0.00000421, -0.00000481, -0.00000282, 0.00000217, 0.00000754, 0.00000879, 0.00000365, -0.00000435, -0.00000783, -0.00000238, 0.00000801, 0.00001329, 0.00000636, -0.00000975, -0.00002404, -0.00002722, -0.00001926, -0.00000777, -0.00000014, 0.00000144, -0.00000196, -0.00000946, -0.00001881, -0.00002240, -0.00001087, 0.00001373, 0.00002966, 0.00000985, -0.00004558, -0.00009228, -0.00007210, 0.00002129, 0.00010425, 0.00005202, -0.00018053, -0.00046738, -0.00055713, -0.00025935, 0.00036419, 0.00097929, 0.00118681, 0.00081910, 0.00008787, -0.00055307, -0.00073585, -0.00043871, 0.00002850, 0.00029176, 0.00020372, -0.00007842, -0.00026984, -0.00021450, 0.00000374, 0.00017113, 0.00015061, -0.00000862, -0.00014210, -0.00013377, -0.00001285, 0.00009426, 0.00009281, 0.00000070, -0.00008522, -0.00008953, -0.00002308, 0.00004163, 0.00004739, 0.00000193, -0.00004105, -0.00003996, -0.00000170, 0.00003348, 0.00003501, 0.00000780, -0.00001946, -0.00002469, -0.00000973, 0.00000789, 0.00001432, 0.00000914, -0.00000003, -0.00000762, -0.00001297, -0.00001613, -0.00001475, -0.00000683, 0.00000465, 0.00001229, 0.00001063, 0.00000159, -0.00000696, -0.00000834, -0.00000267, 0.00000418, 0.00000649, 0.00000338, -0.00000162, -0.00000466, -0.00000470, -0.00000324, -0.00000212, -0.00000196, -0.00000228, -0.00000252, -0.00000240, -0.00000178, -0.00000047, 0.00000146, 0.00000340, 0.00000439, 0.00000363, 0.00000106, -0.00000251, -0.00000559, -0.00000662, -0.00000498, },
		{ -0.00000513, -0.00000652, -0.00000502, -0.00000163, 0.00000136, 0.00000208, 0.00000046, -0.00000196, -0.00000367, -0.00000425, -0.00000412, -0.00000348, -0.00000222, -0.00000078, -0.00000056, -0.00000243, -0.00000491, -0.00000461, 0.00000030, 0.00000700, 0.00000953, 0.00000457, -0.00000411, -0.00000817, -0.00000270, 0.00000805, 0.00001376, 0.00000751, -0.00000769, -0.00002195, -0.00002730, -0.00002315, -0.00001374, -0.00000301, 0.00000587, 0.00000759, -0.00000358, -0.00002451, -0.00003750, -0.00002211, 0.00001966, 0.00005171, 0.00002961, -0.00004927, -0.00012204, -0.00010457, 0.00001846, 0.00014287, 0.00010750, -0.00015414, -0.00050094, -0.00063852, -0.00033888, 0.00033796, 0.00102211, 0.00126921, 0.00089056, 0.00011307, -0.00057446, -0.00077731, -0.00047091, 0.00001903, 0.00030072, 0.00021902, -0.00006680, -0.00026655, -0.00021982, -0.00000678, 0.00016207, 0.00014997, 0.00000080, -0.00012935, -0.00012842, -0.00002014, 0.00008035, 0.00008495, 0.00000653, -0.00007078, -0.00007999, -0.00002779, 0.00002646, 0.00003514, 0.00000303, -0.00002865, -0.00002823, -0.00000095, 0.00002385, 0.00002498, 0.00000664, -0.00001152, -0.00001565, -0.00000758, 0.00000233, 0.00000704, 0.00000666, 0.00000360, -0.00000177, -0.00000958, -0.00001671, -0.00001744, -0.00000912, 0.00000371, 0.00001218, 0.00001076, 0.00000198, -0.00000620, -0.00000772, -0.00000302, 0.00000269, 0.00000475, 0.00000266, -0.00000088, -0.00000318, -0.00000360, -0.00000305, -0.00000246, -0.00000204, -0.00000173, -0.00000159, -0.00000172, -0.00000185, -0.00000140, -0.00000001, 0.00000200, 0.00000363, 0.00000374, 0.00000178, -0.00000173, },
		{ 0.00000331, 0.00000211, -0.00000091, -0.00000419, -0.00000577, -0.00000472, -0.00000193, 0.00000070, 0.00000181, 0.00000139, 0.00000016, -0.00000158, -0.00000392, -0.00000628, -0.00000694, -0.00000464, -0.00000061, 0.00000167, -0.00000021, -0.00000444, -0.00000596, -0.00000162, 0.00000573, 0.00000925, 0.00000503, -0.00000318, -0.00000706, -0.00000212, 0.00000732, 0.00001212, 0.00000699, -0.00000519, -0.00001739, -0.00002494, -0.00002682, -0.00002196, -0.00000857, 0.00000977, 0.00001942, 0.00000546, -0.00002899, -0.00005451, -0.00003690, 0.00002325, 0.00007450, 0.00005271, -0.00004933, -0.00015046, -0.00013933, 0.00001037, 0.00017598, 0.00016117, -0.00012269, -0.00052288, -0.00070707, -0.00041273, 0.00030429, 0.00104608, 0.00133135, 0.00095262, 0.00014548, -0.00057741, -0.00080189, -0.00049848, 0.00000103, 0.00029667, 0.00022743, -0.00005181, -0.00025421, -0.00021868, -0.00001819, 0.00014684, 0.00014380, 0.00000960, -0.00011285, -0.00011887, -0.00002615, 0.00006451, 0.00007418, 0.00001085, -0.00005571, -0.00006863, -0.00003084, 0.00001193, 0.00002244, 0.00000298, -0.00001737, -0.00001689, 0.00000042, 0.00001536, 0.00001553, 0.00000484, -0.00000502, -0.00000754, -0.00000482, -0.00000146, 0.00000116, 0.00000383, 0.00000531, 0.00000219, -0.00000644, -0.00001568, -0.00001801, -0.00001041, 0.00000208, 0.00001054, 0.00000980, 0.00000242, -0.00000462, -0.00000634, -0.00000320, 0.00000082, 0.00000249, 0.00000157, -0.00000023, -0.00000153, -0.00000217, -0.00000252, -0.00000261, -0.00000226, -0.00000154, -0.00000099, -0.00000107, -0.00000161, -0.00000188, -0.00000119, 0.00000053, 0.00000246, },
		{ -0.00000198, -0.00000080, 0.00000107, 0.00000236, 0.00000194, -0.00000023, -0.00000288, -0.00000436, -0.00000397, -0.00000238, -0.00000064, 0.00000084, 0.00000214, 0.00000275, 0.00000127, -0.00000303, -0.00000831, -0.00001065, -0.00000753, -0.00000095, 0.00000356, 0.00000212, -0.00000328, -0.00000640, -0.00000318, 0.00000375, 0.00000758, 0.00000462, -0.00000162, -0.00000418, -0.00000026, 0.00000595, 0.00000815, 0.00000439, -0.00000254, -0.00001035, -0.00001979, -0.00002981, -0.00003218, -0.00001693, 0.00001260, 0.00003291, 0.00001748, -0.00003168, -0.00007236, -0.00005455, 0.00002402, 0.00009650, 0.00007774, -0.00004571, -0.00017571, -0.00017404, -0.00000213, 0.00020183, 0.00020965, -0.00008872, -0.00053285, -0.00075949, -0.00047674, 0.00026548, 0.00105034, 0.00137001, 0.00100181, 0.00018331, -0.00056144, -0.00080759, -0.00051928, -0.00002429, 0.00027947, 0.00022775, -0.00003490, -0.00023362, -0.00021075, -0.00002925, 0.00012662, 0.00013233, 0.00001686, -0.00009392, -0.00010571, -0.00003028, 0.00004796, 0.00006128, 0.00001335, -0.00004111, -0.00005629, -0.00003212, -0.00000104, 0.00001012, 0.00000188, -0.00000786, -0.00000667, 0.00000221, 0.00000840, 0.00000727, 0.00000267, -0.00000018, -0.00000081, -0.00000179, -0.00000349, -0.00000306, 0.00000103, 0.00000536, 0.00000424, -0.00000385, -0.00001348, -0.00001675, -0.00001063, 0.00000019, 0.00000782, 0.00000794, 0.00000266, -0.00000266, -0.00000448, -0.00000313, -0.00000108, 0.00000001, 0.00000019, 0.00000015, 0.00000001, -0.00000057, -0.00000159, -0.00000243, -0.00000242, -0.00000162, -0.00000076, -0.00000060, -0.00000122, -0.00000198, },
		{ -0.00000043, -0.00000088, -0.00000183, -0.00000238, -0.00000182, -0.00000035, 0.00000101, 0.00000122, 0.00000015, -0.00000135, -0.00000236, -0.00000274, -0.00000283, -0.00000246, -0.00000073, 0.00000258, 0.00000547, 0.00000456, -0.00000160, -0.00000986, -0.00001411, -0.00001061, -0.00000201, 0.00000457, 0.00000414, -0.00000142, -0.00000553, -0.00000395, 0.00000117, 0.00000426, 0.00000298, 0.00000042, 0.00000061, 0.00000319, 0.00000415, 0.00000182, -0.00000054, -0.00000006, -0.00000102, -0.00001177, -0.00003170, -0.00004386, -0.00002794, 0.00001389, 0.00004718, 0.00003192, -0.00003216, -0.00008983, -0.00007400, 0.00002183, 0.00011622, 0.00010299, -0.00003877, -0.00019620, -0.00020624, -0.00001776, 0.00021926, 0.00024990, -0.00005501, -0.00053128, -0.00079342, -0.00052727, 0.00022421, 0.00103514, 0.00138313, 0.00103514, 0.00022421, -0.00052727, -0.00079342, -0.00053128, -0.00005501, 0.00024990, 0.00021926, -0.00001776, -0.00020624, -0.00019620, -0.00003877, 0.00010299, 0.00011622, 0.00002183, -0.00007400, -0.00008983, -0.00003216, 0.00003192, 0.00004718, 0.00001389, -0.00002794, -0.00004386, -0.00003170, -0.00001177, -0.00000102, -0.00000006, -0.00000054, 0.00000182, 0.00000415, 0.00000319, 0.00000061, 0.00000042, 0.00000298, 0.00000426, 0.00000117, -0.00000395, -0.00000553, -0.00000142, 0.00000414, 0.00000457, -0.00000201, -0.00001061, -0.00001411, -0.00000986, -0.00000160, 0.00000456, 0.00000547, 0.00000258, -0.00000073, -0.00000246, -0.00000283, -0.00000274, -0.00000236, -0.00000135, 0.00000015, 0.00000122, 0.00000101, -0.00000035, -0.00000182, -0.00000238, -0.00000183, -0.00000088, },
		{ -0.00000198, -0.00000198, -0.00000122, -0.00000060, -0.00000076, -0.00000162, -0.00000242, -0.00000243, -0.00000159, -0.00000057, 0.00000001, 0.00000015, 0.00000019, 0.00000001, -0.00000108, -0.00000313, -0.00000448, -0.00000266, 0.00000266, 0.00000794, 0.00000782, 0.00000019, -0.00001063, -0.00001675, -0.00001348, -0.00000385, 0.00000424, 0.00000536, 0.00000103, -0.00000306, -0.00000349, -0.00000179, -0.00000081, -0.00000018, 0.00000267, 0.00000727, 0.00000840, 0.00000221, -0.00000667, -0.00000786, 0.00000188, 0.00001012, -0.00000104, -0.00003212, -0.00005629, -0.00004111, 0.00001335, 0.00006128, 0.00004796, -0.00003028, -0.00010571, -0.00009392, 0.00001686, 0.00013233, 0.00012662, -0.00002925, -0.00021075, -0.00023362, -0.00003490, 0.00022775, 0.00027947, -0.00002429, -0.00051928, -0.00080759, -0.00056144, 0.00018331, 0.00100181, 0.00137001, 0.00105034, 0.00026548, -0.00047674, -0.00075949, -0.00053285, -0.00008872, 0.00020965, 0.00020183, -0.00000213, -0.00017404, -0.00017571, -0.00004571, 0.00007774, 0.00009650, 0.00002402, -0.00005455, -0.00007236, -0.00003168, 0.00001748, 0.00003291, 0.00001260, -0.00001693, -0.00003218, -0.00002981, -0.00001979, -0.00001035, -0.00000254, 0.00000439, 0.00000815, 0.00000595, -0.00000026, -0.00000418, -0.00000162, 0.00000462, 0.00000758, 0.00000375, -0.00000318, -0.00000640, -0.00000328, 0.00000212, 0.00000356, -0.00000095, -0.00000753, -0.00001065, -0.00000831, -0.00000303, 0.00000127, 0.00000275, 0.00000214, 0.00000084, -0.00000064, -0.00000238, -0.00000397, -0.00000436, -0.00000288, -0.00000023, 0.00000194, 0.00000236, 0.00000107, -0.00000080, },
		{ 0.00000331, 0.00000246, 0.00000053, -0.00000119, -0.00000188, -0.00000161, -0.00000107, -0.00000099, -0.00000154, -0.00000226, -0.00000261, -0.00000252, -0.00000217, -0.00000153, -0.00000023, 0.00000157, 0.00000249, 0.00000082, -0.00000320, -0.00000634, -0.00000462, 0.00000242, 0.00000980, 0.00001054, 0.00000208, -0.00001041, -0.00001801, -0.00001568, -0.00000644, 0.00000219, 0.00000531, 0.00000383, 0.00000116, -0.00000146, -0.00000482, -0.00000754, -0.00000502, 0.00000484, 0.00001553, 0.00001536, 0.00000042, -0.00001689, -0.00001737, 0.00000298, 0.00002244, 0.00001193, -0.00003084, -0.00006863, -0.00005571, 0.00001085, 0.00007418, 0.00006451, -0.00002615, -0.00011887, -0.00011285, 0.00000960, 0.00014380, 0.00014684, -0.00001819, -0.00021868, -0.00025421, -0.00005181, 0.00022743, 0.00029667, 0.00000103, -0.00049848, -0.00080189, -0.00057741, 0.00014548, 0.00095262, 0.00133135, 0.00104608, 0.00030429, -0.00041273, -0.00070707, -0.00052288, -0.00012269, 0.00016117, 0.00017598, 0.00001037, -0.00013933, -0.00015046, -0.00004933, 0.00005271, 0.00007450, 0.00002325, -0.00003690, -0.00005451, -0.00002899, 0.00000546, 0.00001942, 0.00000977, -0.00000857, -0.00002196, -0.00002682, -0.00002494, -0.00001739, -0.00000519, 0.00000699, 0.00001212, 0.00000732, -0.00000212, -0.00000706, -0.00000318, 0.00000503, 0.00000925, 0.00000573, -0.00000162, -0.00000596, -0.00000444, -0.00000021, 0.00000167, -0.00000061, -0.00000464, -0.00000694, -0.00000628, -0.00000392, -0.00000158, 0.00000016, 0.00000139, 0.00000181, 0.00000070, -0.00000193, -0.00000472, -0.00000577, -0.00000419, -0.00000091, 0.00000211, },
		{ -0.00000513, -0.00000173, 0.00000178, 0.00000374, 0.00000363, 0.00000200, -0.00000001, -0.00000140, -0.00000185, -0.00000172, -0.00000159, -0.00000173, -0.00000204, -0.00000246, -0.00000305, -0.00000360, -0.00000318, -0.00000088, 0.00000266, 0.00000475, 0.00000269, -0.00000302, -0.00000772, -0.00000620, 0.00000198, 0.00001076, 0.00001218, 0.00000371, -0.00000912, -0.00001744, -0.00001671, -0.00000958, -0.00000177, 0.00000360, 0.00000666, 0.00000704, 0.00000233, -0.00000758, -0.00001565, -0.00001152, 0.00000664, 0.00002498, 0.00002385, -0.00000095, -0.00002823, -0.00002865, 0.00000303, 0.00003514, 0.00002646, -0.00002779, -0.00007999, -0.00007078, 0.00000653, 0.00008495, 0.00008035, -0.00002014, -0.00012842, -0.00012935, 0.00000080, 0.00014997, 0.00016207, -0.00000678, -0.00021982, -0.00026655, -0.00006680, 0.00021902, 0.00030072, 0.00001903, -0.00047091, -0.00077731, -0.00057446, 0.00011307, 0.00089056, 0.00126921, 0.00102211, 0.00033796, -0.00033888, -0.00063852, -0.00050094, -0.00015414, 0.00010750, 0.00014287, 0.00001846, -0.00010457, -0.00012204, -0.00004927, 0.00002961, 0.00005171, 0.00001966, -0.00002211, -0.00003750, -0.00002451, -0.00000358, 0.00000759, 0.00000587, -0.00000301, -0.00001374, -0.00002315, -0.00002730, -0.00002195, -0.00000769, 0.00000751, 0.00001376, 0.00000805, -0.00000270, -0.00000817, -0.00000411, 0.00000457, 0.00000953, 0.00000700, 0.00000030, -0.00000461, -0.00000491, -0.00000243, -0.00000056, -0.00000078, -0.00000222, -0.00000348, -0.00000412, -0.00000425, -0.00000367, -0.00000196, 0.00000046, 0.00000208, 0.00000136, -0.00000163, -0.00000502, -0.00000652, },
		{ -0.00000162, -0.00000498, -0.00000662, -0.00000559, -0.00000251, 0.00000106, 0.00000363, 0.00000439, 0.00000340, 0.00000146, -0.00000047, -0.00000178, -0.00000240, -0.00000252, -0.00000228, -0.00000196, -0.00000212, -0.00000324, -0.00000470, -0.00000466, -0.00000162, 0.00000338, 0.00000649, 0.00000418, -0.00000267, -0.00000834, -0.00000696, 0.00000159, 0.00001063, 0.00001229, 0.00000465, -0.00000683, -0.00001475, -0.00001613, -0.00001297, -0.00000762, -0.00000003, 0.00000914, 0.00001432, 0.00000789, -0.00000973, -0.00002469, -0.00001946, 0.00000780, 0.00003501, 0.00003348, -0.00000170, -0.00003996, -0.00004105, 0.00000193, 0.00004739, 0.00004163, -0.00002308, -0.00008953, -0.00008522, 0.00000070, 0.00009281, 0.00009426, -0.00001285, -0.00013377, -0.00014210, -0.00000862, 0.00015061, 0.00017113, 0.00000374, -0.00021450, -0.00026984, -0.00007842, 0.00020372, 0.00029176, 0.00002850, -0.00043871, -0.00073585, -0.00055307, 0.00008787, 0.00081910, 0.00118681, 0.00097929, 0.00036419, -0.00025935, -0.00055713, -0.00046738, -0.00018053, 0.00005202, 0.00010425, 0.00002129, -0.00007210, -0.00009228, -0.00004558, 0.00000985, 0.00002966, 0.00001373, -0.00001087, -0.00002240, -0.00001881, -0.00000946, -0.00000196, 0.00000144, -0.00000014, -0.00000777, -0.00001926, -0.00002722, -0.00002404, -0.00000975, 0.00000636, 0.00001329, 0.00000801, -0.00000238, -0.00000783, -0.00000435, 0.00000365, 0.00000879, 0.00000754, 0.00000217, -0.00000282, -0.00000481, -0.00000421, -0.00000263, -0.00000122, -0.00000044, -0.00000070, -0.00000218, -0.00000408, -0.00000482, -0.00000336, -0.00000046, 0.00000168, 0.00000125, },
		{ -0.00000120, 0.00000077, 0.00000041, -0.00000199, -0.00000476, -0.00000619, -0.00000553, -0.00000310, 0.00000014, 0.00000303, 0.00000463, 0.00000454, 0.00000306, 0.00000089, -0.00000129, -0.00000294, -0.00000363, -0.00000317, -0.00000214, -0.00000183, -0.00000320, -0.00000537, -0.00000569, -0.00000215, 0.00000378, 0.00000743, 0.00000494, -0.00000231, -0.00000801, -0.00000650, 0.00000155, 0.00000940, 0.00001053, 0.00000450, -0.00000375, -0.00000985, -0.00001363, -0.00001622, -0.00001513, -0.00000565, 0.00001091, 0.00002252, 0.00001505, -0.00001098, -0.00003406, -0.00002843, 0.00000813, 0.00004497, 0.00004366, -0.00000173, -0.00005131, -0.00005378, -0.00000025, 0.00005837, 0.00005645, -0.00001701, -0.00009653, -0.00009794, -0.00000610, 0.00009725, 0.00010514, -0.00000501, -0.00013466, -0.00015008, -0.00001769, 0.00014590, 0.00017330, 0.00001226, -0.00020351, -0.00026394, -0.00008559, 0.00018305, 0.00027080, 0.00002901, -0.00040399, -0.00068031, -0.00051487, 0.00007097, 0.00074192, 0.00108834, 0.00091955, 0.00038123, -0.00017842, -0.00046688, -0.00042332, -0.00019973, -0.00000181, 0.00006234, 0.00001857, -0.00004397, -0.00006314, -0.00003874, -0.00000556, 0.00000974, 0.00000616, -0.00000349, -0.00001007, -0.00001261, -0.00001234, -0.00000886, -0.00000296, 0.00000044, -0.00000407, -0.00001553, -0.00002519, -0.00002393, -0.00001116, 0.00000405, 0.00001114, 0.00000718, -0.00000158, -0.00000645, -0.00000393, 0.00000264, 0.00000745, 0.00000744, 0.00000364, -0.00000104, -0.00000431, -0.00000532, -0.00000415, -0.00000164, 0.00000068, 0.00000114, -0.00000075, -0.00000356, -0.00000501, -0.00000389, },
		{ -0.00000288, -0.00000437, -0.00000357, -0.00000164, -0.00000046, -0.00000099, -0.00000275, -0.00000454, -0.00000542, -0.00000502, -0.00000337, -0.00000079, 0.00000206, 0.00000433, 0.00000526, 0.00000461, 0.00000255, -0.00000036, -0.00000317, -0.00000471, -0.00000429, -0.00000267, -0.00000181, -0.00000317, -0.00000564, -0.00000603, -0.00000218, 0.00000399, 0.00000741, 0.00000463, -0.00000218, -0.00000668, -0.00000454, 0.00000216, 0.00000717, 0.00000673, 0.00000293, -0.00000023, -0.00000289, -0.00000904, -0.00001891, -0.00002386, -0.00001313, 0.00001162, 0.00003107, 0.00002340, -0.00001114, -0.00004310, -0.00003789, 0.00000758, 0.00005414, 0.00005368, -0.00000107, -0.00006153, -0.00006597, -0.00000332, 0.00006737, 0.00006988, -0.00001002, -0.00010050, -0.00010794, -0.00001325, 0.00009801, 0.00011208, 0.00000259, -0.00013118, -0.00015264, -0.00002550, 0.00013642, 0.00016839, 0.00001790, -0.00018794, -0.00024944, -0.00008773, 0.00015870, 0.00023963, 0.00002093, -0.00036862, -0.00061403, -0.00046246, 0.00006266, 0.00066257, 0.00097858, 0.00084575, 0.00038809, -0.00010023, -0.00037219, -0.00037061, -0.00021026, -0.00005077, 0.00001964, 0.00001059, -0.00002175, -0.00003649, -0.00002957, -0.00001613, -0.00000689, -0.00000215, 0.00000016, -0.00000103, -0.00000662, -0.00001266, -0.00001305, -0.00000683, -0.00000069, -0.00000242, -0.00001227, -0.00002185, -0.00002207, -0.00001186, 0.00000112, 0.00000783, 0.00000565, -0.00000069, -0.00000452, -0.00000298, 0.00000183, 0.00000595, 0.00000686, 0.00000449, 0.00000034, -0.00000365, -0.00000568, -0.00000483, -0.00000179, 0.00000121, 0.00000195, 0.00000000, },
		{ 0.00000130, 0.00000180, 0.00000001, -0.00000223, -0.00000312, -0.00000252, -0.00000168, -0.00000174, -0.00000272, -0.00000383, -0.00000444, -0.00000451, -0.00000418, -0.00000329, -0.00000158, 0.00000090, 0.00000355, 0.00000549, 0.00000591, 0.00000432, 0.00000093, -0.00000300, -0.00000553, -0.00000543, -0.00000352, -0.00000228, -0.00000341, -0.00000561, -0.00000552, -0.00000140, 0.00000420, 0.00000637, 0.00000298, -0.00000255, -0.00000444, -0.00000092, 0.00000370, 0.00000419, 0.00000097, -0.00000027, 0.00000333, 0.00000575, -0.00000240, -0.00002064, -0.00003318, -0.00002209, 0.00001105, 0.00003926, 0.00003239, -0.00001017, -0.00005116, -0.00004714, 0.00000617, 0.00006189, 0.00006280, 0.00000017, -0.00006996, -0.00007672, -0.00000696, 0.00007384, 0.00008100, -0.00000268, -0.00010116, -0.00011439, -0.00002005, 0.00009515, 0.00011451, 0.00000916, -0.00012377, -0.00014960, -0.00003129, 0.00012302, 0.00015680, 0.00002011, -0.00016911, -0.00022753, -0.00008478, 0.00013234, 0.00020058, 0.00000539, -0.00033409, -0.00054063, -0.00039921, 0.00006249, 0.00058430, 0.00086258, 0.00076145, 0.00038453, -0.00002845, -0.00027752, -0.00031167, -0.00021139, -0.00009212, -0.00002129, -0.00000181, -0.00000637, -0.00001394, -0.00001915, -0.00002188, -0.00001946, -0.00001030, 0.00000058, 0.00000454, -0.00000150, -0.00001110, -0.00001474, -0.00000979, -0.00000286, -0.00000237, -0.00000964, -0.00001782, -0.00001903, -0.00001187, -0.00000194, 0.00000396, 0.00000361, -0.00000000, -0.00000249, -0.00000171, 0.00000143, 0.00000464, 0.00000599, 0.00000463, 0.00000104, -0.00000304, -0.00000533, -0.00000455, -0.00000151, },
		{ -0.00000441, -0.00000334, -0.00000070, 0.00000115, 0.00000090, -0.00000066, -0.00000175, -0.00000157, -0.00000099, -0.00000134, -0.00000286, -0.00000449, -0.00000508, -0.00000454, -0.00000368, -0.00000318, -0.00000289, -0.00000210, -0.00000028, 0.00000244, 0.00000524, 0.00000683, 0.00000596, 0.00000239, -0.00000241, -0.00000586, -0.00000629, -0.00000458, -0.00000338, -0.00000422, -0.00000549, -0.00000414, 0.00000042, 0.00000467, 0.00000442, -0.00000013, -0.00000366, -0.00000149, 0.00000431, 0.00000638, 0.00000082, -0.00000649, -0.00000516, 0.00000647, 0.00001551, 0.00000602, -0.00002107, -0.00004237, -0.00003198, 0.00000913, 0.00004644, 0.00004131, -0.00000815, -0.00005765, -0.00005548, 0.00000410, 0.00006768, 0.00007029, 0.00000173, -0.00007610, -0.00008525, -0.00001078, 0.00007746, 0.00008901, 0.00000441, -0.00009852, -0.00011675, -0.00002583, 0.00008900, 0.00011222, 0.00001409, -0.00011312, -0.00014118, -0.00003455, 0.00010680, 0.00013941, 0.00001874, -0.00014839, -0.00019988, -0.00007723, 0.00010551, 0.00015637, -0.00001589, -0.00030145, -0.00046372, -0.00032894, 0.00006931, 0.00050981, 0.00074531, 0.00067064, 0.00037109, 0.00003396, -0.00018710, -0.00024931, -0.00020318, -0.00012380, -0.00005804, -0.00001736, 0.00000191, 0.00000330, -0.00000869, -0.00002330, -0.00002762, -0.00001741, -0.00000146, 0.00000684, 0.00000225, -0.00000843, -0.00001441, -0.00001163, -0.00000537, -0.00000335, -0.00000769, -0.00001370, -0.00001544, -0.00001133, -0.00000471, 0.00000008, 0.00000131, 0.00000027, -0.00000074, -0.00000034, 0.00000150, 0.00000379, 0.00000504, 0.00000407, 0.00000094, -0.00000264, },
		{ 0.00000291, 0.00000006, -0.00000257, -0.00000309, -0.00000140, 0.00000060, 0.00000093, -0.00000047, -0.00000181, -0.00000153, -0.00000002, 0.00000072, -0.00000072, -0.00000365, -0.00000601, -0.00000630, -0.00000482, -0.00000305, -0.00000221, -0.00000228, -0.00000234, -0.00000132, 0.00000121, 0.00000464, 0.00000729, 0.00000722, 0.00000376, -0.00000148, -0.00000552, -0.00000658, -0.00000564, -0.00000515, -0.00000583, -0.00000553, -0.00000197, 0.00000341, 0.00000567, 0.00000176, -0.00000466, -0.00000571, 0.00000182, 0.00001092, 0.00001027, -0.00000254, -0.00001516, -0.00001160, 0.00000881, 0.00002565, 0.00001574, -0.00002004, -0.00005067, -0.00004208, 0.00000592, 0.00005198, 0.00004942, -0.00000529, -0.00006211, -0.00006222, 0.00000162, 0.00007114, 0.00007551, 0.00000332, -0.00007966, -0.00009092, -0.00001434, 0.00007812, 0.00009340, 0.00001070, -0.00009283, -0.00011480, -0.00003004, 0.00008014, 0.00010541, 0.00001692, -0.00010010, -0.00012808, -0.00003509, 0.00008890, 0.00011751, 0.00001403, -0.00012709, -0.00016847, -0.00006600, 0.00007949, 0.00010980, -0.00004073, -0.00027124, -0.00038670, -0.00025566, 0.00008143, 0.00044112, 0.00063128, 0.00057747, 0.00034902, 0.00008494, -0.00010464, -0.00018652, -0.00018646, -0.00014461, -0.00008854, -0.00003445, 0.00000351, 0.00001453, 0.00000062, -0.00002131, -0.00003145, -0.00002278, -0.00000501, 0.00000637, 0.00000435, -0.00000547, -0.00001269, -0.00001232, -0.00000762, -0.00000476, -0.00000633, -0.00000997, -0.00001187, -0.00001045, -0.00000690, -0.00000333, -0.00000099, 0.00000006, 0.00000043, 0.00000092, 0.00000204, 0.00000350, 0.00000416, },
		{ 0.00000293, 0.00000378, 0.00000345, 0.00000135, -0.00000147, -0.00000284, -0.00000160, 0.00000099, 0.00000226, 0.00000080, -0.00000196, -0.00000319, -0.00000158, 0.00000125, 0.00000227, 0.00000004, -0.00000403, -0.00000706, -0.00000728, -0.00000519, -0.00000271, -0.00000144, -0.00000165, -0.00000237, -0.00000213, 0.00000008, 0.00000389, 0.00000731, 0.00000788, 0.00000472, -0.00000039, -0.00000444, -0.00000601, -0.00000642, -0.00000750, -0.00000842, -0.00000604, 0.00000077, 0.00000757, 0.00000740, -0.00000128, -0.00001039, -0.00000878, 0.00000513, 0.00001849, 0.00001528, -0.00000548, -0.00002441, -0.00001923, 0.00001003, 0.00003540, 0.00002610, -0.00001751, -0.00005740, -0.00005159, 0.00000167, 0.00005542, 0.00005601, -0.00000196, -0.00006426, -0.00006679, -0.00000093, 0.00007211, 0.00007802, 0.00000464, -0.00008053, -0.00009334, -0.00001724, 0.00007596, 0.00009392, 0.00001569, -0.00008455, -0.00010868, -0.00003229, 0.00006929, 0.00009460, 0.00001745, -0.00008568, -0.00011130, -0.00003300, 0.00007045, 0.00009266, 0.00000655, -0.00010636, -0.00013544, -0.00005237, 0.00005521, 0.00006361, -0.00006678, -0.00024360, -0.00031248, -0.00018320, 0.00009684, 0.00037954, 0.00052434, 0.00048587, 0.00032009, 0.00012341, -0.00003304, -0.00012619, -0.00016272, -0.00015422, -0.00011127, -0.00005137, -0.00000057, 0.00001963, 0.00000777, -0.00001707, -0.00003148, -0.00002597, -0.00000907, 0.00000387, 0.00000473, -0.00000293, -0.00001028, -0.00001198, -0.00000914, -0.00000601, -0.00000537, -0.00000694, -0.00000883, -0.00000948, -0.00000839, -0.00000593, -0.00000302, -0.00000060, 0.00000091, 0.00000190, },
		{ -0.00000161, 0.00000069, 0.00000249, 0.00000400, 0.00000450, 0.00000295, -0.00000043, -0.00000339, -0.00000336, -0.00000013, 0.00000351, 0.00000407, 0.00000085, -0.00000326, -0.00000450, -0.00000183, 0.00000206, 0.00000340, 0.00000074, -0.00000399, -0.00000749, -0.00000781, -0.00000551, -0.00000261, -0.00000097, -0.00000119, -0.00000235, -0.00000268, -0.00000072, 0.00000327, 0.00000701, 0.00000781, 0.00000497, 0.00000060, -0.00000262, -0.00000435, -0.00000662, -0.00001025, -0.00001203, -0.00000725, 0.00000375, 0.00001272, 0.00001000, -0.00000431, -0.00001697, -0.00001285, 0.00000803, 0.00002649, 0.00002118, -0.00000767, -0.00003354, -0.00002755, 0.00000998, 0.00004400, 0.00003634, -0.00001363, -0.00006198, -0.00005973, -0.00000324, 0.00005649, 0.00006045, 0.00000145, -0.00006403, -0.00006879, -0.00000321, 0.00007062, 0.00007756, 0.00000540, -0.00007883, -0.00009235, -0.00001915, 0.00007130, 0.00009065, 0.00001903, -0.00007430, -0.00009883, -0.00003239, 0.00005730, 0.00008065, 0.00001576, -0.00007081, -0.00009210, -0.00002868, 0.00005246, 0.00006655, -0.00000281, -0.00008711, -0.00010285, -0.00003779, 0.00003329, 0.00002021, -0.00009168, -0.00021827, -0.00024341, -0.00011499, 0.00011337, 0.00032566, 0.00042745, 0.00039933, 0.00028642, 0.00014924, 0.00002574, -0.00007092, -0.00013391, -0.00015322, -0.00012535, -0.00006642, -0.00000884, 0.00001904, 0.00001200, -0.00001186, -0.00002856, -0.00002684, -0.00001274, 0.00000021, 0.00000358, -0.00000134, -0.00000790, -0.00001090, -0.00000965, -0.00000664, -0.00000460, -0.00000476, -0.00000663, -0.00000862, -0.00000914, -0.00000755, -0.00000460, },
		{ -0.00000925, -0.00000818, -0.00000561, -0.00000278, -0.00000009, 0.00000268, 0.00000506, 0.00000544, 0.00000260, -0.00000221, -0.00000541, -0.00000401, 0.00000117, 0.00000581, 0.00000579, 0.00000110, -0.00000411, -0.00000544, -0.00000218, 0.00000230, 0.00000388, 0.00000119, -0.00000365, -0.00000728, -0.00000777, -0.00000557, -0.00000261, -0.00000084, -0.00000109, -0.00000250, -0.00000305, -0.00000105, 0.00000304, 0.00000658, 0.00000695, 0.00000425, 0.00000120, -0.00000021, -0.00000151, -0.00000594, -0.00001310, -0.00001654, -0.00000935, 0.00000661, 0.00001855, 0.00001348, -0.00000695, -0.00002388, -0.00001771, 0.00001020, 0.00003426, 0.00002756, -0.00000886, -0.00004180, -0.00003592, 0.00000863, 0.00005076, 0.00004564, -0.00000872, -0.00006403, -0.00006577, -0.00000832, 0.00005513, 0.00006231, 0.00000449, -0.00006156, -0.00006804, -0.00000490, 0.00006689, 0.00007415, 0.00000539, -0.00007485, -0.00008809, -0.00001987, 0.00006460, 0.00008396, 0.00002058, -0.00006279, -0.00008598, -0.00003038, 0.00004497, 0.00006462, 0.00001213, -0.00005637, -0.00007186, -0.00002276, 0.00003577, 0.00004085, -0.00001303, -0.00006994, -0.00007255, -0.00002373, 0.00001399, -0.00001845, -0.00011333, -0.00019476, -0.00018117, -0.00005383, 0.00012897, 0.00027945, 0.00034255, 0.00032068, 0.00025027, 0.00016318, 0.00007077, -0.00002278, -0.00010226, -0.00014293, -0.00013063, -0.00007817, -0.00001956, 0.00001370, 0.00001294, -0.00000693, -0.00002377, -0.00002554, -0.00001529, -0.00000370, 0.00000127, -0.00000100, -0.00000612, -0.00000942, -0.00000910, -0.00000636, -0.00000379, -0.00000344, -0.00000542, -0.00000806, },
		{ -0.00000282, -0.00000516, -0.00000785, -0.00000889, -0.00000794, -0.00000603, -0.00000392, -0.00000120, 0.00000253, 0.00000595, 0.00000634, 0.00000230, -0.00000385, -0.00000722, -0.00000461, 0.00000219, 0.00000760, 0.00000717, 0.00000153, -0.00000435, -0.00000581, -0.00000248, 0.00000201, 0.00000363, 0.00000123, -0.00000316, -0.00000653, -0.00000711, -0.00000521, -0.00000251, -0.00000097, -0.00000148, -0.00000307, -0.00000339, -0.00000084, 0.00000347, 0.00000627, 0.00000538, 0.00000240, 0.00000111, 0.00000254, 0.00000249, -0.00000417, -0.00001568, -0.00002170, -0.00001241, 0.00000894, 0.00002462, 0.00001772, -0.00000884, -0.00003055, -0.00002305, 0.00001139, 0.00004116, 0.00003392, -0.00000896, -0.00004853, -0.00004366, 0.00000616, 0.00005518, 0.00005326, -0.00000322, -0.00006341, -0.00006918, -0.00001307, 0.00005151, 0.00006137, 0.00000677, -0.00005716, -0.00006460, -0.00000579, 0.00006129, 0.00006804, 0.00000451, -0.00006902, -0.00008095, -0.00001932, 0.00005646, 0.00007444, 0.00002034, -0.00005076, -0.00007106, -0.00002653, 0.00003306, 0.00004768, 0.00000708, -0.00004306, -0.00005194, -0.00001600, 0.00002095, 0.00001704, -0.00002303, -0.00005519, -0.00004602, -0.00001155, -0.00000266, -0.00005093, -0.00013004, -0.00017248, -0.00012678, -0.00000172, 0.00014189, 0.00024038, 0.00027061, 0.00025191, 0.00021383, 0.00016668, 0.00010210, 0.00001679, -0.00007007, -0.00012529, -0.00012762, -0.00008556, -0.00003089, 0.00000493, 0.00001063, -0.00000331, -0.00001827, -0.00002250, -0.00001628, -0.00000707, -0.00000168, -0.00000197, -0.00000533, -0.00000792, -0.00000761, -0.00000507, -0.00000277, },
		{ -0.00000548, -0.00000291, -0.00000146, -0.00000268, -0.00000565, -0.00000798, -0.00000824, -0.00000709, -0.00000596, -0.00000487, -0.00000237, 0.00000216, 0.00000653, 0.00000693, 0.00000192, -0.00000521, -0.00000857, -0.00000498, 0.00000289, 0.00000869, 0.00000800, 0.00000205, -0.00000391, -0.00000548, -0.00000259, 0.00000130, 0.00000270, 0.00000078, -0.00000274, -0.00000545, -0.00000590, -0.00000426, -0.00000205, -0.00000120, -0.00000244, -0.00000427, -0.00000390, -0.00000010, 0.00000474, 0.00000635, 0.00000328, -0.00000062, 0.00000008, 0.00000531, 0.00000745, -0.00000118, -0.00001763, -0.00002714, -0.00001635, 0.00001040, 0.00003039, 0.00002245, -0.00000974, -0.00003639, -0.00002842, 0.00001148, 0.00004660, 0.00003971, -0.00000803, -0.00005323, -0.00005009, 0.00000286, 0.00005699, 0.00005855, 0.00000234, -0.00006023, -0.00006962, -0.00001698, 0.00004600, 0.00005769, 0.00000800, -0.00005129, -0.00005877, -0.00000577, 0.00005431, 0.00005969, 0.00000277, -0.00006181, -0.00007151, -0.00001759, 0.00004745, 0.00006289, 0.00001852, -0.00003889, -0.00005509, -0.00002126, 0.00002221, 0.00003098, 0.00000125, -0.00003139, -0.00003356, -0.00000924, 0.00000834, -0.00000366, -0.00003183, -0.00004296, -0.00002432, -0.00000231, -0.00001685, -0.00007642, -0.00014067, -0.00015083, -0.00008063, 0.00004020, 0.00015079, 0.00020757, 0.00021164, 0.00019410, 0.00017900, 0.00016171, 0.00012070, 0.00004707, -0.00003944, -0.00010263, -0.00011750, -0.00008798, -0.00004111, -0.00000571, 0.00000547, -0.00000173, -0.00001317, -0.00001837, -0.00001558, -0.00000928, -0.00000474, -0.00000404, -0.00000569, -0.00000673, },
		{ -0.00000682, -0.00000713, -0.00000607, -0.00000310, -0.00000018, 0.00000010, -0.00000278, -0.00000655, -0.00000832, -0.00000747, -0.00000594, -0.00000553, -0.00000550, -0.00000333, 0.00000175, 0.00000671, 0.00000701, 0.00000132, -0.00000623, -0.00000928, -0.00000499, 0.00000327, 0.00000897, 0.00000818, 0.00000256, -0.00000287, -0.00000446, -0.00000242, 0.00000034, 0.00000126, -0.00000013, -0.00000257, -0.00000432, -0.00000429, -0.00000265, -0.00000099, -0.00000130, -0.00000390, -0.00000627, -0.00000483, 0.00000102, 0.00000691, 0.00000706, 0.00000091, -0.00000470, -0.00000205, 0.00000774, 0.00001306, 0.00000299, -0.00001861, -0.00003237, -0.00002097, 0.00001074, 0.00003533, 0.00002729, -0.00000952, -0.00004086, -0.00003334, 0.00001050, 0.00005012, 0.00004434, -0.00000625, -0.00005555, -0.00005463, -0.00000088, 0.00005614, 0.00006110, 0.00000743, -0.00005481, -0.00006706, -0.00001965, 0.00003913, 0.00005155, 0.00000801, -0.00004449, -0.00005105, -0.00000485, 0.00004645, 0.00004973, 0.00000032, -0.00005377, -0.00006049, -0.00001489, 0.00003818, 0.00005018, 0.00001546, -0.00002776, -0.00003911, -0.00001513, 0.00001287, 0.00001559, -0.00000469, -0.00002167, -0.00001771, -0.00000324, -0.00000193, -0.00002042, -0.00003861, -0.00003312, -0.00000802, 0.00000330, -0.00002884, -0.00009464, -0.00014468, -0.00012939, -0.00004261, 0.00007160, 0.00015491, 0.00017998, 0.00016487, 0.00014747, 0.00014729, 0.00015048, 0.00012827, 0.00006809, -0.00001214, -0.00007739, -0.00010187, -0.00008538, -0.00004882, -0.00001664, -0.00000179, -0.00000255, -0.00000939, -0.00001388, -0.00001334, -0.00001001, -0.00000735, },
		{ -0.00001002, -0.00000918, -0.00000910, -0.00000981, -0.00000935, -0.00000605, -0.00000088, 0.00000269, 0.00000179, -0.00000288, -0.00000748, -0.00000867, -0.00000671, -0.00000477, -0.00000495, -0.00000574, -0.00000386, 0.00000143, 0.00000648, 0.00000644, 0.00000042, -0.00000689, -0.00000926, -0.00000456, 0.00000337, 0.00000846, 0.00000769, 0.00000301, -0.00000138, -0.00000288, -0.00000196, -0.00000064, -0.00000041, -0.00000139, -0.00000279, -0.00000346, -0.00000255, -0.00000042, 0.00000088, -0.00000100, -0.00000573, -0.00000913, -0.00000637, 0.00000232, 0.00000991, 0.00000855, -0.00000142, -0.00000960, -0.00000530, 0.00000948, 0.00001888, 0.00000816, -0.00001835, -0.00003690, -0.00002591, 0.00000985, 0.00003895, 0.00003178, -0.00000821, -0.00004354, -0.00003731, 0.00000862, 0.00005144, 0.00004735, -0.00000392, -0.00005541, -0.00005685, -0.00000458, 0.00005283, 0.00006072, 0.00001156, -0.00004767, -0.00006171, -0.00002080, 0.00003149, 0.00004346, 0.00000679, -0.00003731, -0.00004208, -0.00000318, 0.00003823, 0.00003888, -0.00000258, -0.00004538, -0.00004869, -0.00001153, 0.00002913, 0.00003723, 0.00001162, -0.00001784, -0.00002408, -0.00000874, 0.00000532, 0.00000237, -0.00001006, -0.00001399, -0.00000508, 0.00000136, -0.00000996, -0.00003279, -0.00004285, -0.00002544, 0.00000280, 0.00000501, -0.00003893, -0.00010583, -0.00014215, -0.00010793, -0.00001221, 0.00009291, 0.00015403, 0.00015654, 0.00012895, 0.00011144, 0.00011970, 0.00013527, 0.00012697, 0.00008051, 0.00001056, -0.00005193, -0.00008266, -0.00007820, -0.00005303, -0.00002637, -0.00001016, -0.00000570, -0.00000754, -0.00000980, },
		{ -0.00001074, -0.00000790, -0.00000678, -0.00000621, -0.00000703, -0.00000972, -0.00001243, -0.00001188, -0.00000660, 0.00000081, 0.00000519, 0.00000336, -0.00000282, -0.00000809, -0.00000879, -0.00000603, -0.00000385, -0.00000441, -0.00000560, -0.00000383, 0.00000133, 0.00000586, 0.00000519, -0.00000081, -0.00000722, -0.00000853, -0.00000372, 0.00000323, 0.00000727, 0.00000663, 0.00000333, 0.00000036, -0.00000098, -0.00000124, -0.00000144, -0.00000198, -0.00000279, -0.00000348, -0.00000317, -0.00000100, 0.00000231, 0.00000367, -0.00000005, -0.00000770, -0.00001279, -0.00000869, 0.00000353, 0.00001357, 0.00001089, -0.00000340, -0.00001493, -0.00000956, 0.00001027, 0.00002441, 0.00001402, -0.00001674, -0.00004024, -0.00003072, 0.00000776, 0.00004086, 0.00003546, -0.00000598, -0.00004420, -0.00003989, 0.00000611, 0.00005048, 0.00004838, -0.00000141, -0.00005292, -0.00005655, -0.00000782, 0.00004747, 0.00005748, 0.00001436, -0.00003945, -0.00005402, -0.00002036, 0.00002373, 0.00003411, 0.00000449, -0.00003027, -0.00003259, -0.00000103, 0.00003012, 0.00002789, -0.00000563, -0.00003707, -0.00003688, -0.00000789, 0.00002074, 0.00002484, 0.00000745, -0.00000942, -0.00001080, -0.00000266, -0.00000036, -0.00000808, -0.00001433, -0.00000828, 0.00000396, 0.00000413, -0.00001594, -0.00004072, -0.00004433, -0.00001961, 0.00000845, 0.00000299, -0.00004734, -0.00011063, -0.00013372, -0.00008648, 0.00001137, 0.00010518, 0.00014847, 0.00013628, 0.00010211, 0.00008483, 0.00009668, 0.00011815, 0.00011924, 0.00008555, 0.00002793, -0.00002828, -0.00006188, -0.00006730, -0.00005329, -0.00003370, -0.00001857, },
		{ -0.00004964, -0.00003783, -0.00002595, -0.00001696, -0.00001036, -0.00000528, -0.00000261, -0.00000403, -0.00000917, -0.00001420, -0.00001417, -0.00000753, 0.00000170, 0.00000686, 0.00000453, -0.00000250, -0.00000806, -0.00000845, -0.00000541, -0.00000333, -0.00000405, -0.00000513, -0.00000321, 0.00000150, 0.00000490, 0.00000334, -0.00000231, -0.00000722, -0.00000721, -0.00000256, 0.00000287, 0.00000560, 0.00000521, 0.00000354, 0.00000210, 0.00000093, -0.00000043, -0.00000189, -0.00000309, -0.00000403, -0.00000460, -0.00000368, 0.00000004, 0.00000531, 0.00000736, 0.00000175, -0.00000953, -0.00001706, -0.00001181, 0.00000437, 0.00001757, 0.00001400, -0.00000474, -0.00002026, -0.00001456, 0.00000993, 0.00002916, 0.00002011, -0.00001381, -0.00004199, -0.00003493, 0.00000466, 0.00004083, 0.00003789, -0.00000314, -0.00004278, -0.00004074, 0.00000334, 0.00004738, 0.00004726, 0.00000091, -0.00004842, -0.00005375, -0.00001022, 0.00004063, 0.00005174, 0.00001562, -0.00003082, -0.00004466, -0.00001841, 0.00001640, 0.00002425, 0.00000141, -0.00002378, -0.00002329, 0.00000127, 0.00002246, 0.00001745, -0.00000847, -0.00002920, -0.00002577, -0.00000435, 0.00001328, 0.00001371, 0.00000343, -0.00000265, 0.00000016, 0.00000262, -0.00000420, -0.00001543, -0.00001712, -0.00000434, 0.00000941, 0.00000490, -0.00002016, -0.00004451, -0.00004315, -0.00001532, 0.00000960, -0.00000220, -0.00005420, -0.00010995, -0.00012046, -0.00006533, 0.00002902, 0.00010989, 0.00013900, 0.00011845, 0.00008248, 0.00006599, 0.00007821, 0.00010088, 0.00010751, 0.00008477, 0.00003982, -0.00000803, -0.00004143, -0.00005387, },
		{ 0.00000784, -0.00002292, -0.00003928, -0.00004263, -0.00003842, -0.00003142, -0.00002344, -0.00001448, -0.00000554, 0.00000011, -0.00000079, -0.00000758, -0.00001473, -0.00001567, -0.00000855, 0.00000164, 0.00000733, 0.00000502, -0.00000196, -0.00000723, -0.00000748, -0.00000481, -0.00000325, -0.00000396, -0.00000439, -0.00000207, 0.00000189, 0.00000367, 0.00000104, -0.00000395, -0.00000691, -0.00000549, -0.00000130, 0.00000229, 0.00000365, 0.00000368, 0.00000371, 0.00000364, 0.00000250, 0.00000028, -0.00000192, -0.00000344, -0.00000480, -0.00000603, -0.00000510, 0.00000028, 0.00000827, 0.00001180, 0.00000449, -0.00001090, -0.00002161, -0.00001568, 0.00000460, 0.00002151, 0.00001767, -0.00000525, -0.00002511, -0.00001992, 0.00000842, 0.00003269, 0.00002594, -0.00000975, -0.00004191, -0.00003806, 0.00000089, 0.00003883, 0.00003872, -0.00000005, -0.00003945, -0.00003972, 0.00000067, 0.00004247, 0.00004404, 0.00000271, -0.00004238, -0.00004871, -0.00001153, 0.00003294, 0.00004403, 0.00001531, -0.00002246, -0.00003441, -0.00001521, 0.00000996, 0.00001464, -0.00000207, -0.00001814, -0.00001482, 0.00000338, 0.00001553, 0.00000815, -0.00001080, -0.00002201, -0.00001590, -0.00000123, 0.00000696, 0.00000436, -0.00000008, 0.00000243, 0.00000846, 0.00000673, -0.00000639, -0.00001966, -0.00001824, -0.00000190, 0.00001152, 0.00000373, -0.00002293, -0.00004473, -0.00003972, -0.00001231, 0.00000715, -0.00000973, -0.00005953, -0.00010488, -0.00010375, -0.00004497, 0.00004168, 0.00010873, 0.00012671, 0.00010252, 0.00006816, 0.00005308, 0.00006385, 0.00008478, 0.00009400, 0.00007992, 0.00004661, },
		{ 0.00008053, 0.00007273, 0.00004907, 0.00001894, -0.00000755, -0.00002488, -0.00003319, -0.00003559, -0.00003434, -0.00002922, -0.00001954, -0.00000748, 0.00000141, 0.00000202, -0.00000527, -0.00001383, -0.00001590, -0.00000933, 0.00000066, 0.00000638, 0.00000461, -0.00000128, -0.00000556, -0.00000579, -0.00000412, -0.00000353, -0.00000411, -0.00000350, -0.00000060, 0.00000238, 0.00000221, -0.00000149, -0.00000550, -0.00000628, -0.00000357, -0.00000016, 0.00000146, 0.00000168, 0.00000236, 0.00000396, 0.00000480, 0.00000341, 0.00000065, -0.00000152, -0.00000282, -0.00000481, -0.00000756, -0.00000742, -0.00000049, 0.00001086, 0.00001672, 0.00000817, -0.00001154, -0.00002608, -0.00002008, 0.00000405, 0.00002499, 0.00002160, -0.00000486, -0.00002905, -0.00002519, 0.00000586, 0.00003466, 0.00003099, -0.00000487, -0.00003992, -0.00003971, -0.00000314, 0.00003502, 0.00003778, 0.00000288, -0.00003455, -0.00003686, -0.00000156, 0.00003623, 0.00003895, 0.00000375, -0.00003541, -0.00004189, -0.00001165, 0.00002506, 0.00003505, 0.00001357, -0.00001491, -0.00002409, -0.00001118, 0.00000472, 0.00000596, -0.00000551, -0.00001348, -0.00000767, 0.00000498, 0.00000947, 0.00000040, -0.00001237, -0.00001565, -0.00000767, 0.00000121, 0.00000185, -0.00000293, -0.00000280, 0.00000592, 0.00001402, 0.00000949, -0.00000718, -0.00002099, -0.00001773, -0.00000065, 0.00001083, 0.00000094, -0.00002452, -0.00004216, -0.00003466, -0.00001039, 0.00000210, -0.00001853, -0.00006325, -0.00009653, -0.00008512, -0.00002606, 0.00005021, 0.00010347, 0.00011286, 0.00008820, 0.00005750, 0.00004425, 0.00005286, 0.00007067, },
		{ 0.00003782, 0.00004434, 0.00005893, 0.00006842, 0.00006476, 0.00004825, 0.00002543, 0.00000396, -0.00001191, -0.00002248, -0.00002991, -0.00003442, -0.00003340, -0.00002469, -0.00001080, 0.00000097, 0.00000378, -0.00000267, -0.00001151, -0.00001460, -0.00000953, -0.00000108, 0.00000398, 0.00000315, -0.00000063, -0.00000318, -0.00000344, -0.00000326, -0.00000400, -0.00000437, -0.00000249, 0.00000098, 0.00000277, 0.00000056, -0.00000399, -0.00000671, -0.00000531, -0.00000167, 0.00000057, 0.00000031, -0.00000015, 0.00000153, 0.00000445, 0.00000550, 0.00000341, 0.00000044, -0.00000078, -0.00000112, -0.00000383, -0.00000892, -0.00001050, -0.00000236, 0.00001276, 0.00002175, 0.00001260, -0.00001126, -0.00003001, -0.00002469, 0.00000266, 0.00002760, 0.00002537, -0.00000363, -0.00003171, -0.00002987, 0.00000249, 0.00003490, 0.00003479, 0.00000039, -0.00003611, -0.00003963, -0.00000698, 0.00002974, 0.00003505, 0.00000526, -0.00002857, -0.00003237, -0.00000309, 0.00002924, 0.00003242, 0.00000392, -0.00002810, -0.00003389, -0.00001063, 0.00001759, 0.00002558, 0.00001071, -0.00000860, -0.00001447, -0.00000681, 0.00000080, -0.00000126, -0.00000850, -0.00000983, -0.00000215, 0.00000584, 0.00000433, -0.00000557, -0.00001304, -0.00001020, -0.00000127, 0.00000285, -0.00000205, -0.00000806, -0.00000460, 0.00000798, 0.00001697, 0.00001088, -0.00000685, -0.00001985, -0.00001580, -0.00000030, 0.00000800, -0.00000300, -0.00002518, -0.00003767, -0.00002872, -0.00000945, -0.00000450, -0.00002750, -0.00006519, -0.00008600, -0.00006606, -0.00000926, 0.00005538, 0.00009576, 0.00009871, 0.00007540, 0.00004914, },
		{ 0.00008537, 0.00006418, 0.00004206, 0.00003244, 0.00003737, 0.00004950, 0.00005846, 0.00005721, 0.00004533, 0.00002796, 0.00001144, -0.00000131, -0.00001175, -0.00002225, -0.00003173, -0.00003532, -0.00002902, -0.00001496, -0.00000128, 0.00000403, -0.00000026, -0.00000799, -0.00001167, -0.00000891, -0.00000331, 0.00000033, 0.00000068, -0.00000016, -0.00000037, -0.00000058, -0.00000218, -0.00000440, -0.00000455, -0.00000141, 0.00000242, 0.00000280, -0.00000132, -0.00000625, -0.00000732, -0.00000392, 0.00000001, 0.00000067, -0.00000124, -0.00000168, 0.00000143, 0.00000531, 0.00000571, 0.00000236, -0.00000054, 0.00000013, 0.00000159, -0.00000172, -0.00000983, -0.00001403, -0.00000524, 0.00001370, 0.00002643, 0.00001749, -0.00000996, -0.00003300, -0.00002911, 0.00000050, 0.00002901, 0.00002856, -0.00000172, -0.00003287, -0.00003351, -0.00000133, 0.00003341, 0.00003697, 0.00000554, -0.00003080, -0.00003770, -0.00001017, 0.00002348, 0.00003071, 0.00000680, -0.00002207, -0.00002665, -0.00000379, 0.00002207, 0.00002500, 0.00000325, -0.00002101, -0.00002539, -0.00000868, 0.00001099, 0.00001638, 0.00000714, -0.00000374, -0.00000618, -0.00000260, -0.00000183, -0.00000669, -0.00001069, -0.00000708, 0.00000164, 0.00000582, 0.00000010, -0.00000972, -0.00001278, -0.00000569, 0.00000329, 0.00000367, -0.00000478, -0.00001111, -0.00000549, 0.00000879, 0.00001764, 0.00001103, -0.00000569, -0.00001683, -0.00001282, -0.00000060, 0.00000377, -0.00000750, -0.00002507, -0.00003213, -0.00002271, -0.00000942, -0.00001169, -0.00003560, -0.00006520, -0.00007426, -0.00004793, 0.00000478, 0.00005781, 0.00008699, },
		{ 0.00001561, 0.00005801, 0.00007827, 0.00007373, 0.00005464, 0.00003564, 0.00002714, 0.00003116, 0.00004207, 0.00005092, 0.00005093, 0.00004144, 0.00002751, 0.00001517, 0.00000630, -0.00000220, -0.00001370, -0.00002667, -0.00003459, -0.00003175, -0.00001929, -0.00000513, 0.00000252, 0.00000152, -0.00000366, -0.00000728, -0.00000737, -0.00000570, -0.00000421, -0.00000262, -0.00000003, 0.00000252, 0.00000247, -0.00000087, -0.00000448, -0.00000436, -0.00000022, 0.00000350, 0.00000223, -0.00000345, -0.00000806, -0.00000710, -0.00000206, 0.00000137, -0.00000005, -0.00000328, -0.00000286, 0.00000217, 0.00000668, 0.00000552, 0.00000024, -0.00000240, 0.00000101, 0.00000514, 0.00000149, -0.00001004, -0.00001766, -0.00000895, 0.00001354, 0.00003032, 0.00002245, -0.00000768, -0.00003470, -0.00003290, -0.00000227, 0.00002903, 0.00003077, 0.00000059, -0.00003246, -0.00003574, -0.00000519, 0.00003035, 0.00003734, 0.00001012, -0.00002440, -0.00003401, -0.00001237, 0.00001681, 0.00002510, 0.00000733, -0.00001565, -0.00002021, -0.00000365, 0.00001528, 0.00001732, 0.00000189, -0.00001460, -0.00001708, -0.00000611, 0.00000558, 0.00000811, 0.00000335, -0.00000040, 0.00000032, 0.00000098, -0.00000336, -0.00001022, -0.00001186, -0.00000506, 0.00000377, 0.00000497, -0.00000327, -0.00001215, -0.00001171, -0.00000211, 0.00000612, 0.00000377, -0.00000644, -0.00001233, -0.00000560, 0.00000857, 0.00001647, 0.00001021, -0.00000396, -0.00001256, -0.00000926, -0.00000133, -0.00000108, -0.00001192, -0.00002431, -0.00002632, -0.00001737, -0.00001028, -0.00001864, -0.00004192, -0.00006316, -0.00006211, -0.00003182, },
		{ -0.00005906, -0.00005022, -0.00001850, 0.00002296, 0.00005641, 0.00007033, 0.00006434, 0.00004689, 0.00002958, 0.00002141, 0.00002509, 0.00003610, 0.00004560, 0.00004631, 0.00003756, 0.00002520, 0.00001583, 0.00001071, 0.00000522, -0.00000540, -0.00001992, -0.00003115, -0.00003227, -0.00002311, -0.00001013, -0.00000073, 0.00000232, 0.00000098, -0.00000184, -0.00000494, -0.00000792, -0.00000914, -0.00000645, -0.00000031, 0.00000511, 0.00000535, 0.00000059, -0.00000400, -0.00000351, 0.00000115, 0.00000404, 0.00000084, -0.00000591, -0.00000932, -0.00000589, 0.00000034, 0.00000234, -0.00000166, -0.00000583, -0.00000369, 0.00000379, 0.00000859, 0.00000504, -0.00000281, -0.00000516, 0.00000165, 0.00000921, 0.00000566, -0.00000937, -0.00002096, -0.00001313, 0.00001223, 0.00003303, 0.00002702, -0.00000459, -0.00003489, -0.00003564, -0.00000535, 0.00002760, 0.00003166, 0.00000295, -0.00003059, -0.00003632, -0.00000867, 0.00002606, 0.00003585, 0.00001370, -0.00001747, -0.00002884, -0.00001332, 0.00001034, 0.00001871, 0.00000682, -0.00000983, -0.00001365, -0.00000282, 0.00000929, 0.00001000, 0.00000010, -0.00000920, -0.00000955, -0.00000330, 0.00000150, 0.00000127, -0.00000020, 0.00000155, 0.00000481, 0.00000357, -0.00000407, -0.00001195, -0.00001196, -0.00000358, 0.00000449, 0.00000341, -0.00000585, -0.00001312, -0.00001004, 0.00000057, 0.00000749, 0.00000336, -0.00000712, -0.00001206, -0.00000515, 0.00000752, 0.00001397, 0.00000876, -0.00000189, -0.00000771, -0.00000562, -0.00000236, -0.00000590, -0.00001569, -0.00002299, -0.00002092, -0.00001331, -0.00001197, -0.00002470, -0.00004583, },
		{ -0.00001443, -0.00002941, -0.00004698, -0.00005302, -0.00003906, -0.00000837, 0.00002683, 0.00005335, 0.00006356, 0.00005739, 0.00004095, 0.00002382, 0.00001510, 0.00001883, 0.00003103, 0.00004200, 0.00004336, 0.00003443, 0.00002221, 0.00001435, 0.00001210, 0.00000984, 0.00000159, -0.00001234, -0.00002530, -0.00003030, -0.00002575, -0.00001567, -0.00000539, 0.00000198, 0.00000538, 0.00000409, -0.00000187, -0.00000969, -0.00001387, -0.00001033, -0.00000098, 0.00000700, 0.00000763, 0.00000205, -0.00000280, -0.00000177, 0.00000281, 0.00000397, -0.00000151, -0.00000876, -0.00001000, -0.00000365, 0.00000330, 0.00000294, -0.00000412, -0.00000889, -0.00000424, 0.00000616, 0.00001104, 0.00000442, -0.00000655, -0.00000867, 0.00000188, 0.00001341, 0.00001049, -0.00000776, -0.00002352, -0.00001736, 0.00000988, 0.00003426, 0.00003073, -0.00000098, -0.00003348, -0.00003698, -0.00000840, 0.00002481, 0.00003103, 0.00000500, -0.00002749, -0.00003518, -0.00001138, 0.00002096, 0.00003268, 0.00001600, -0.00001057, -0.00002259, -0.00001294, 0.00000459, 0.00001211, 0.00000538, -0.00000504, -0.00000754, -0.00000151, 0.00000440, 0.00000357, -0.00000181, -0.00000496, -0.00000327, -0.00000062, -0.00000126, -0.00000382, -0.00000312, 0.00000236, 0.00000728, 0.00000496, -0.00000427, -0.00001217, -0.00001106, -0.00000244, 0.00000417, 0.00000143, -0.00000768, -0.00001294, -0.00000805, 0.00000238, 0.00000770, 0.00000268, -0.00000694, -0.00001070, -0.00000440, 0.00000586, 0.00001068, 0.00000705, 0.00000034, -0.00000286, -0.00000234, -0.00000356, -0.00001010, -0.00001836, -0.00002116, -0.00001642, -0.00001093, },
		{ -0.00001892, -0.00001314, -0.00001041, -0.00001749, -0.00003250, -0.00004534, -0.00004534, -0.00002899, -0.00000150, 0.00002750, 0.00004917, 0.00005806, 0.00005276, 0.00003673, 0.00001850, 0.00000843, 0.00001228, 0.00002631, 0.00003941, 0.00004174, 0.00003247, 0.00001955, 0.00001178, 0.00001096, 0.00001139, 0.00000646, -0.00000485, -0.00001763, -0.00002586, -0.00002673, -0.00002101, -0.00001094, 0.00000052, 0.00000904, 0.00000983, 0.00000148, -0.00001083, -0.00001784, -0.00001377, -0.00000194, 0.00000793, 0.00000893, 0.00000331, -0.00000083, 0.00000104, 0.00000487, 0.00000331, -0.00000483, -0.00001201, -0.00001015, -0.00000046, 0.00000676, 0.00000326, -0.00000728, -0.00001236, -0.00000460, 0.00000906, 0.00001390, 0.00000381, -0.00001063, -0.00001272, 0.00000157, 0.00001732, 0.00001559, -0.00000523, -0.00002499, -0.00002117, 0.00000674, 0.00003386, 0.00003317, 0.00000279, -0.00003057, -0.00003669, -0.00001106, 0.00002092, 0.00002884, 0.00000644, -0.00002352, -0.00003243, -0.00001305, 0.00001556, 0.00002813, 0.00001686, -0.00000426, -0.00001580, -0.00001130, 0.00000002, 0.00000587, 0.00000325, -0.00000154, -0.00000239, -0.00000006, 0.00000074, -0.00000156, -0.00000352, -0.00000189, 0.00000147, 0.00000159, -0.00000288, -0.00000707, -0.00000514, 0.00000235, 0.00000796, 0.00000511, -0.00000428, -0.00001131, -0.00000941, -0.00000151, 0.00000322, -0.00000064, -0.00000880, -0.00001196, -0.00000606, 0.00000336, 0.00000709, 0.00000200, -0.00000602, -0.00000864, -0.00000362, 0.00000375, 0.00000709, 0.00000540, 0.00000258, 0.00000147, 0.00000020, -0.00000486, -0.00001329, -0.00001961, },
		{ -0.00000616, -0.00001521, -0.00001934, -0.00001638, -0.00001120, -0.00001165, -0.00002091, -0.00003387, -0.00004117, -0.00003646, -0.00002023, 0.00000236, 0.00002548, 0.00004421, 0.00005369, 0.00005002, 0.00003399, 0.00001383, 0.00000183, 0.00000562, 0.00002154, 0.00003707, 0.00004092, 0.00003177, 0.00001798, 0.00000913, 0.00000801, 0.00000997, 0.00000868, 0.00000168, -0.00000896, -0.00001931, -0.00002579, -0.00002543, -0.00001666, -0.00000182, 0.00001159, 0.00001466, 0.00000461, -0.00001128, -0.00002056, -0.00001623, -0.00000298, 0.00000773, 0.00000893, 0.00000414, 0.00000184, 0.00000494, 0.00000742, 0.00000218, -0.00000900, -0.00001565, -0.00000990, 0.00000346, 0.00001062, 0.00000339, -0.00001087, -0.00001608, -0.00000487, 0.00001220, 0.00001695, 0.00000334, -0.00001466, -0.00001696, 0.00000070, 0.00002056, 0.00002051, -0.00000197, -0.00002515, -0.00002411, 0.00000313, 0.00003184, 0.00003405, 0.00000631, -0.00002639, -0.00003471, -0.00001299, 0.00001628, 0.00002519, 0.00000703, -0.00001911, -0.00002833, -0.00001352, 0.00001033, 0.00002265, 0.00001630, 0.00000102, -0.00000902, -0.00000864, -0.00000311, 0.00000051, 0.00000077, 0.00000059, 0.00000143, 0.00000123, -0.00000173, -0.00000516, -0.00000474, 0.00000015, 0.00000457, 0.00000310, -0.00000358, -0.00000860, -0.00000612, 0.00000187, 0.00000721, 0.00000415, -0.00000436, -0.00000984, -0.00000736, -0.00000071, 0.00000204, -0.00000244, -0.00000920, -0.00001051, -0.00000435, 0.00000356, 0.00000598, 0.00000154, -0.00000450, -0.00000627, -0.00000302, 0.00000138, 0.00000363, 0.00000409, 0.00000469, 0.00000493, 0.00000176, },
		{ 0.00000657, 0.00000728, 0.00000224, -0.00000738, -0.00001579, -0.00001763, -0.00001369, -0.00001059, -0.00001438, -0.00002438, -0.00003356, -0.00003494, -0.00002696, -0.00001292, 0.00000367, 0.00002150, 0.00003881, 0.00005018, 0.00004857, 0.00003236, 0.00001001, -0.00000415, -0.00000080, 0.00001651, 0.00003431, 0.00004023, 0.00003217, 0.00001796, 0.00000725, 0.00000410, 0.00000605, 0.00000811, 0.00000653, -0.00000020, -0.00001130, -0.00002292, -0.00002831, -0.00002175, -0.00000463, 0.00001278, 0.00001798, 0.00000699, -0.00001114, -0.00002169, -0.00001726, -0.00000382, 0.00000640, 0.00000746, 0.00000431, 0.00000500, 0.00000983, 0.00001050, 0.00000078, -0.00001378, -0.00001955, -0.00000940, 0.00000781, 0.00001468, 0.00000346, -0.00001455, -0.00001978, -0.00000515, 0.00001522, 0.00001992, 0.00000307, -0.00001825, -0.00002101, -0.00000067, 0.00002281, 0.00002478, 0.00000175, -0.00002389, -0.00002580, -0.00000053, 0.00002836, 0.00003320, 0.00000921, -0.00002132, -0.00003115, -0.00001396, 0.00001132, 0.00002038, 0.00000663, -0.00001470, -0.00002329, -0.00001280, 0.00000572, 0.00001675, 0.00001448, 0.00000495, -0.00000280, -0.00000528, -0.00000469, -0.00000355, -0.00000168, 0.00000142, 0.00000370, 0.00000206, -0.00000318, -0.00000718, -0.00000528, 0.00000137, 0.00000615, 0.00000380, -0.00000366, -0.00000867, -0.00000610, 0.00000122, 0.00000549, 0.00000234, -0.00000468, -0.00000825, -0.00000530, -0.00000006, 0.00000098, -0.00000364, -0.00000888, -0.00000886, -0.00000315, 0.00000306, 0.00000465, 0.00000146, -0.00000252, -0.00000390, -0.00000279, -0.00000109, 0.00000062, 0.00000330, },
		{ -0.00000350, -0.00000172, 0.00000306, 0.00000810, 0.00000843, 0.00000172, -0.00000842, -0.00001507, -0.00001474, -0.00001105, -0.00001116, -0.00001813, -0.00002753, -0.00003172, -0.00002732, -0.00001750, -0.00000716, 0.00000305, 0.00001636, 0.00003334, 0.00004717, 0.00004770, 0.00003140, 0.00000720, -0.00000895, -0.00000655, 0.00001118, 0.00003062, 0.00003903, 0.00003328, 0.00001959, 0.00000674, 0.00000010, 0.00000036, 0.00000495, 0.00000924, 0.00000772, -0.00000266, -0.00001837, -0.00002920, -0.00002545, -0.00000739, 0.00001259, 0.00001930, 0.00000811, -0.00001062, -0.00002109, -0.00001655, -0.00000417, 0.00000413, 0.00000452, 0.00000366, 0.00000838, 0.00001543, 0.00001404, -0.00000065, -0.00001879, -0.00002354, -0.00000881, 0.00001219, 0.00001867, 0.00000356, -0.00001796, -0.00002318, -0.00000548, 0.00001777, 0.00002247, 0.00000300, -0.00002107, -0.00002446, -0.00000237, 0.00002386, 0.00002800, 0.00000559, -0.00002128, -0.00002601, -0.00000385, 0.00002377, 0.00003067, 0.00001116, -0.00001581, -0.00002627, -0.00001382, 0.00000649, 0.00001479, 0.00000528, -0.00001068, -0.00001776, -0.00001102, 0.00000205, 0.00001097, 0.00001171, 0.00000738, 0.00000240, -0.00000163, -0.00000481, -0.00000607, -0.00000375, 0.00000119, 0.00000442, 0.00000223, -0.00000387, -0.00000779, -0.00000508, 0.00000202, 0.00000645, 0.00000371, -0.00000340, -0.00000769, -0.00000524, 0.00000064, 0.00000331, 0.00000006, -0.00000529, -0.00000695, -0.00000363, 0.00000039, 0.00000026, -0.00000401, -0.00000785, -0.00000719, -0.00000258, 0.00000196, 0.00000332, 0.00000185, -0.00000025, -0.00000179, -0.00000298, },
		{ 0.00000211, -0.00000011, -0.00000358, -0.00000566, -0.00000326, 0.00000335, 0.00000916, 0.00000840, 0.00000036, -0.00000914, -0.00001322, -0.00001104, -0.00000870, -0.00001268, -0.00002232, -0.00002995, -0.00002856, -0.00001905, -0.00000876, -0.00000296, 0.00000117, 0.00001093, 0.00002817, 0.00004433, 0.00004670, 0.00003063, 0.00000543, -0.00001213, -0.00001115, 0.00000569, 0.00002573, 0.00003679, 0.00003459, 0.00002266, 0.00000787, -0.00000327, -0.00000622, -0.00000025, 0.00000964, 0.00001400, 0.00000565, -0.00001263, -0.00002790, -0.00002712, -0.00000952, 0.00001119, 0.00001839, 0.00000757, -0.00001001, -0.00001885, -0.00001396, -0.00000375, 0.00000121, 0.00000032, 0.00000210, 0.00001160, 0.00002134, 0.00001790, -0.00000187, -0.00002358, -0.00002733, -0.00000828, 0.00001619, 0.00002227, 0.00000375, -0.00002074, -0.00002596, -0.00000586, 0.00001956, 0.00002430, 0.00000306, -0.00002288, -0.00002695, -0.00000419, 0.00002364, 0.00002985, 0.00000919, -0.00001754, -0.00002464, -0.00000645, 0.00001847, 0.00002663, 0.00001198, -0.00001034, -0.00002048, -0.00001259, 0.00000218, 0.00000891, 0.00000312, -0.00000736, -0.00001226, -0.00000845, -0.00000051, 0.00000576, 0.00000838, 0.00000834, 0.00000627, 0.00000188, -0.00000370, -0.00000698, -0.00000514, 0.00000022, 0.00000375, 0.00000162, -0.00000410, -0.00000727, -0.00000419, 0.00000235, 0.00000584, 0.00000298, -0.00000301, -0.00000608, -0.00000382, 0.00000029, 0.00000114, -0.00000230, -0.00000616, -0.00000624, -0.00000267, 0.00000052, 0.00000000, -0.00000343, -0.00000613, -0.00000562, -0.00000264, 0.00000042, 0.00000216, 0.00000269, },
		{ -0.00000135, 0.00000127, 0.00000387, 0.00000434, 0.00000103, -0.00000448, -0.00000741, -0.00000397, 0.00000401, 0.00000966, 0.00000732, -0.00000153, -0.00000941, -0.00001050, -0.00000700, -0.00000687, -0.00001486, -0.00002634, -0.00003126, -0.00002434, -0.00001089, -0.00000139, -0.00000032, -0.00000129, 0.00000599, 0.00002364, 0.00004136, 0.00004497, 0.00002956, 0.00000466, -0.00001340, -0.00001420, 0.00000031, 0.00001964, 0.00003320, 0.00003556, 0.00002670, 0.00001060, -0.00000544, -0.00001276, -0.00000673, 0.00000785, 0.00001804, 0.00001275, -0.00000635, -0.00002449, -0.00002635, -0.00001048, 0.00000890, 0.00001525, 0.00000514, -0.00000965, -0.00001524, -0.00000958, -0.00000236, -0.00000193, -0.00000479, -0.00000032, 0.00001431, 0.00002704, 0.00002184, -0.00000267, -0.00002769, -0.00003061, -0.00000788, 0.00001941, 0.00002515, 0.00000403, -0.00002260, -0.00002780, -0.00000623, 0.00002038, 0.00002513, 0.00000315, -0.00002355, -0.00002822, -0.00000589, 0.00002224, 0.00003017, 0.00001220, -0.00001299, -0.00002178, -0.00000807, 0.00001296, 0.00002146, 0.00001160, -0.00000536, -0.00001427, -0.00001041, -0.00000128, 0.00000322, 0.00000040, -0.00000493, -0.00000722, -0.00000543, -0.00000192, 0.00000150, 0.00000490, 0.00000800, 0.00000863, 0.00000488, -0.00000173, -0.00000642, -0.00000567, -0.00000108, 0.00000204, 0.00000029, -0.00000416, -0.00000603, -0.00000278, 0.00000256, 0.00000473, 0.00000188, -0.00000265, -0.00000428, -0.00000218, 0.00000019, -0.00000067, -0.00000435, -0.00000716, -0.00000626, -0.00000265, 0.00000023, 0.00000019, -0.00000194, -0.00000383, -0.00000420, -0.00000322, },
		{ -0.00000111, -0.00000290, -0.00000409, -0.00000310, 0.00000073, 0.00000522, 0.00000621, 0.00000158, -0.00000554, -0.00000859, -0.00000391, 0.00000484, 0.00000949, 0.00000539, -0.00000363, -0.00000911, -0.00000718, -0.00000310, -0.00000578, -0.00001738, -0.00002960, -0.00003114, -0.00001934, -0.00000353, 0.00000410, 0.00000083, -0.00000375, 0.00000220, 0.00002001, 0.00003806, 0.00004203, 0.00002780, 0.00000469, -0.00001268, -0.00001544, -0.00000461, 0.00001259, 0.00002817, 0.00003572, 0.00003107, 0.00001459, -0.00000612, -0.00001840, -0.00001356, 0.00000429, 0.00001953, 0.00001785, -0.00000025, -0.00001927, -0.00002299, -0.00000986, 0.00000617, 0.00001016, 0.00000081, -0.00000984, -0.00001072, -0.00000372, 0.00000008, -0.00000483, -0.00001027, -0.00000344, 0.00001618, 0.00003199, 0.00002551, -0.00000291, -0.00003067, -0.00003302, -0.00000765, 0.00002155, 0.00002699, 0.00000433, -0.00002333, -0.00002847, -0.00000650, 0.00002013, 0.00002477, 0.00000313, -0.00002308, -0.00002810, -0.00000721, 0.00001984, 0.00002894, 0.00001433, -0.00000806, -0.00001769, -0.00000857, 0.00000772, 0.00001561, 0.00001010, -0.00000124, -0.00000817, -0.00000754, -0.00000371, -0.00000181, -0.00000254, -0.00000342, -0.00000303, -0.00000235, -0.00000229, -0.00000158, 0.00000169, 0.00000668, 0.00000952, 0.00000708, 0.00000069, -0.00000466, -0.00000528, -0.00000236, -0.00000030, -0.00000159, -0.00000429, -0.00000451, -0.00000115, 0.00000277, 0.00000350, 0.00000070, -0.00000235, -0.00000263, -0.00000066, 0.00000030, -0.00000187, -0.00000578, -0.00000810, -0.00000699, -0.00000362, -0.00000055, 0.00000069, 0.00000027, },
		{ -0.00000185, 0.00000132, 0.00000288, 0.00000179, -0.00000169, -0.00000498, -0.00000453, 0.00000058, 0.00000653, 0.00000752, 0.00000154, -0.00000655, -0.00000907, -0.00000318, 0.00000561, 0.00000859, 0.00000284, -0.00000558, -0.00000816, -0.00000358, 0.00000019, -0.00000557, -0.00001995, -0.00003160, -0.00002938, -0.00001387, 0.00000247, 0.00000732, 0.00000061, -0.00000574, 0.00000001, 0.00001748, 0.00003433, 0.00003767, 0.00002508, 0.00000531, -0.00001011, -0.00001480, -0.00000873, 0.00000500, 0.00002185, 0.00003472, 0.00003509, 0.00001928, -0.00000528, -0.00002247, -0.00001984, -0.00000043, 0.00001847, 0.00002044, 0.00000494, -0.00001279, -0.00001721, -0.00000743, 0.00000349, 0.00000362, -0.00000521, -0.00001082, -0.00000585, 0.00000313, 0.00000354, -0.00000704, -0.00001554, -0.00000698, 0.00001698, 0.00003566, 0.00002854, -0.00000257, -0.00003218, -0.00003423, -0.00000752, 0.00002239, 0.00002756, 0.00000456, -0.00002288, -0.00002783, -0.00000655, 0.00001885, 0.00002317, 0.00000285, -0.00002160, -0.00002660, -0.00000796, 0.00001675, 0.00002633, 0.00001541, -0.00000318, -0.00001273, -0.00000795, 0.00000316, 0.00000961, 0.00000771, 0.00000178, -0.00000267, -0.00000433, -0.00000507, -0.00000583, -0.00000535, -0.00000278, 0.00000009, 0.00000044, -0.00000185, -0.00000342, -0.00000095, 0.00000475, 0.00000912, 0.00000833, 0.00000313, -0.00000211, -0.00000405, -0.00000330, -0.00000276, -0.00000377, -0.00000463, -0.00000312, 0.00000040, 0.00000300, 0.00000247, -0.00000025, -0.00000209, -0.00000137, 0.00000045, 0.00000049, -0.00000238, -0.00000640, -0.00000881, -0.00000828, -0.00000548, },
		{ -0.00000914, -0.00000989, -0.00000797, -0.00000360, 0.00000184, 0.00000547, 0.00000458, -0.00000055, -0.00000557, -0.00000535, 0.00000085, 0.00000761, 0.00000807, 0.00000096, -0.00000735, -0.00000878, -0.00000193, 0.00000609, 0.00000694, -0.00000005, -0.00000706, -0.00000651, -0.00000001, 0.00000253, -0.00000633, -0.00002227, -0.00003198, -0.00002590, -0.00000822, 0.00000670, 0.00000813, -0.00000082, -0.00000694, -0.00000033, 0.00001613, 0.00003016, 0.00003184, 0.00002130, 0.00000623, -0.00000602, -0.00001238, -0.00001177, -0.00000258, 0.00001461, 0.00003239, 0.00003809, 0.00002399, -0.00000313, -0.00002454, -0.00002474, -0.00000559, 0.00001515, 0.00002027, 0.00000864, -0.00000574, -0.00000947, -0.00000320, 0.00000133, -0.00000369, -0.00001246, -0.00001269, -0.00000121, 0.00001034, 0.00000780, -0.00000817, -0.00001996, -0.00001055, 0.00001659, 0.00003761, 0.00003055, -0.00000169, -0.00003199, -0.00003399, -0.00000737, 0.00002189, 0.00002671, 0.00000458, -0.00002131, -0.00002585, -0.00000628, 0.00001667, 0.00002038, 0.00000223, -0.00001933, -0.00002386, -0.00000802, 0.00001331, 0.00002263, 0.00001539, 0.00000125, -0.00000738, -0.00000634, -0.00000040, 0.00000398, 0.00000476, 0.00000360, 0.00000181, -0.00000116, -0.00000542, -0.00000861, -0.00000769, -0.00000280, 0.00000203, 0.00000263, -0.00000089, -0.00000409, -0.00000280, 0.00000262, 0.00000774, 0.00000859, 0.00000523, 0.00000076, -0.00000221, -0.00000369, -0.00000492, -0.00000590, -0.00000522, -0.00000220, 0.00000153, 0.00000321, 0.00000185, -0.00000072, -0.00000180, -0.00000062, 0.00000098, 0.00000063, -0.00000225, -0.00000617, },
		{ 0.00000056, -0.00000165, -0.00000519, -0.00000899, -0.00001148, -0.00001073, -0.00000563, 0.00000204, 0.00000761, 0.00000690, 0.00000053, -0.00000560, -0.00000530, 0.00000155, 0.00000828, 0.00000775, -0.00000010, -0.00000776, -0.00000769, -0.00000035, 0.00000609, 0.00000457, -0.00000302, -0.00000783, -0.00000417, 0.00000326, 0.00000363, -0.00000804, -0.00002409, -0.00003053, -0.00002078, -0.00000267, 0.00000894, 0.00000656, -0.00000322, -0.00000717, 0.00000118, 0.00001589, 0.00002564, 0.00002473, 0.00001650, 0.00000717, -0.00000090, -0.00000849, -0.00001353, -0.00000958, 0.00000698, 0.00002877, 0.00003956, 0.00002798, -0.00000011, -0.00002447, -0.00002765, -0.00001041, 0.00001012, 0.00001744, 0.00001044, 0.00000116, -0.00000047, 0.00000260, 0.00000007, -0.00001100, -0.00002033, -0.00001541, 0.00000264, 0.00001719, 0.00001255, -0.00000800, -0.00002297, -0.00001372, 0.00001506, 0.00003754, 0.00003120, -0.00000045, -0.00003007, -0.00003214, -0.00000706, 0.00002014, 0.00002444, 0.00000429, -0.00001879, -0.00002265, -0.00000561, 0.00001384, 0.00001659, 0.00000119, -0.00001656, -0.00002014, -0.00000735, 0.00000987, 0.00001822, 0.00001436, 0.00000489, -0.00000214, -0.00000401, -0.00000280, -0.00000082, 0.00000162, 0.00000426, 0.00000502, 0.00000161, -0.00000498, -0.00001007, -0.00000932, -0.00000325, 0.00000287, 0.00000405, 0.00000027, -0.00000382, -0.00000376, 0.00000064, 0.00000576, 0.00000800, 0.00000669, 0.00000349, -0.00000003, -0.00000343, -0.00000637, -0.00000764, -0.00000600, -0.00000195, 0.00000198, 0.00000328, 0.00000172, -0.00000056, -0.00000134, -0.00000035, 0.00000084, },
		{ -0.00000064, -0.00000045, 0.00000011, 0.00000022, -0.00000082, -0.00000369, -0.00000835, -0.00001276, -0.00001331, -0.00000772, 0.00000176, 0.00000887, 0.00000842, 0.00000150, -0.00000485, -0.00000420, 0.00000271, 0.00000845, 0.00000654, -0.00000153, -0.00000769, -0.00000585, 0.00000138, 0.00000549, 0.00000160, -0.00000580, -0.00000771, -0.00000125, 0.00000598, 0.00000340, -0.00001057, -0.00002519, -0.00002722, -0.00001426, 0.00000247, 0.00000915, 0.00000287, -0.00000630, -0.00000637, 0.00000434, 0.00001660, 0.00002087, 0.00001668, 0.00001088, 0.00000793, 0.00000472, -0.00000356, -0.00001396, -0.00001546, -0.00000043, 0.00002408, 0.00003916, 0.00003062, 0.00000323, -0.00002240, -0.00002822, -0.00001421, 0.00000411, 0.00001236, 0.00001015, 0.00000719, 0.00000895, 0.00000948, -0.00000008, -0.00001755, -0.00002807, -0.00001880, 0.00000525, 0.00002300, 0.00001735, -0.00000644, -0.00002414, -0.00001607, 0.00001254, 0.00003537, 0.00003027, 0.00000092, -0.00002655, -0.00002868, -0.00000642, 0.00001737, 0.00002089, 0.00000359, -0.00001563, -0.00001848, -0.00000451, 0.00001065, 0.00001213, -0.00000021, -0.00001358, -0.00001580, -0.00000604, 0.00000674, 0.00001355, 0.00001251, 0.00000751, 0.00000255, -0.00000129, -0.00000402, -0.00000446, -0.00000131, 0.00000397, 0.00000684, 0.00000369, -0.00000401, -0.00001030, -0.00001007, -0.00000385, 0.00000278, 0.00000462, 0.00000135, -0.00000289, -0.00000388, -0.00000089, 0.00000360, 0.00000675, 0.00000738, 0.00000567, 0.00000213, -0.00000256, -0.00000688, -0.00000869, -0.00000681, -0.00000244, 0.00000159, 0.00000306, 0.00000202, 0.00000026, },
		{ 0.00000242, 0.00000260, 0.00000166, 0.00000038, -0.00000071, -0.00000104, -0.00000043, -0.00000002, -0.00000199, -0.00000731, -0.00001344, -0.00001526, -0.00000958, 0.00000092, 0.00000892, 0.00000882, 0.00000229, -0.00000323, -0.00000194, 0.00000432, 0.00000809, 0.00000450, -0.00000320, -0.00000710, -0.00000341, 0.00000306, 0.00000423, -0.00000183, -0.00000818, -0.00000667, 0.00000207, 0.00000796, 0.00000183, -0.00001370, -0.00002540, -0.00002219, -0.00000671, 0.00000696, 0.00000746, -0.00000252, -0.00000973, -0.00000463, 0.00000876, 0.00001799, 0.00001601, 0.00000815, 0.00000478, 0.00000833, 0.00001025, 0.00000186, -0.00001314, -0.00001978, -0.00000702, 0.00001867, 0.00003679, 0.00003142, 0.00000632, -0.00001872, -0.00002642, -0.00001647, -0.00000209, 0.00000568, 0.00000788, 0.00001177, 0.00001786, 0.00001681, 0.00000096, -0.00002269, -0.00003493, -0.00002253, 0.00000636, 0.00002718, 0.00002174, -0.00000357, -0.00002324, -0.00001722, 0.00000934, 0.00003120, 0.00002764, 0.00000218, -0.00002172, -0.00002375, -0.00000532, 0.00001392, 0.00001635, 0.00000246, -0.00001215, -0.00001369, -0.00000304, 0.00000741, 0.00000737, -0.00000190, -0.00001068, -0.00001125, -0.00000424, 0.00000415, 0.00000905, 0.00001011, 0.00000904, 0.00000630, 0.00000144, -0.00000416, -0.00000675, -0.00000368, 0.00000298, 0.00000732, 0.00000488, -0.00000282, -0.00000952, -0.00000992, -0.00000437, 0.00000203, 0.00000440, 0.00000212, -0.00000163, -0.00000332, -0.00000182, 0.00000163, 0.00000514, 0.00000727, 0.00000703, 0.00000396, -0.00000123, -0.00000636, -0.00000882, -0.00000748, -0.00000359, 0.00000031, },
		{ -0.00000779, -0.00000519, -0.00000175, 0.00000130, 0.00000322, 0.00000341, 0.00000168, -0.00000093, -0.00000234, -0.00000132, 0.00000051, -0.00000043, -0.00000603, -0.00001334, -0.00001621, -0.00001091, -0.00000045, 0.00000755, 0.00000790, 0.00000278, -0.00000078, 0.00000143, 0.00000635, 0.00000727, 0.00000181, -0.00000495, -0.00000599, -0.00000056, 0.00000448, 0.00000232, -0.00000549, -0.00000995, -0.00000474, 0.00000558, 0.00000904, -0.00000091, -0.00001712, -0.00002459, -0.00001571, 0.00000138, 0.00001054, 0.00000413, -0.00000904, -0.00001313, -0.00000209, 0.00001394, 0.00001970, 0.00001119, -0.00000035, -0.00000140, 0.00000831, 0.00001519, 0.00000723, -0.00001128, -0.00002230, -0.00001227, 0.00001300, 0.00003259, 0.00003012, 0.00000858, -0.00001400, -0.00002247, -0.00001686, -0.00000768, -0.00000178, 0.00000395, 0.00001449, 0.00002539, 0.00002383, 0.00000303, -0.00002592, -0.00004021, -0.00002617, 0.00000590, 0.00002931, 0.00002526, 0.00000033, -0.00002023, -0.00001694, 0.00000581, 0.00002534, 0.00002340, 0.00000305, -0.00001604, -0.00001766, -0.00000368, 0.00001021, 0.00001120, 0.00000092, -0.00000870, -0.00000871, -0.00000131, 0.00000441, 0.00000273, -0.00000370, -0.00000809, -0.00000690, -0.00000218, 0.00000225, 0.00000512, 0.00000751, 0.00000951, 0.00000887, 0.00000383, -0.00000347, -0.00000769, -0.00000526, 0.00000164, 0.00000668, 0.00000515, -0.00000169, -0.00000803, -0.00000899, -0.00000462, 0.00000093, 0.00000357, 0.00000245, -0.00000037, -0.00000231, -0.00000209, 0.00000016, 0.00000349, 0.00000646, 0.00000740, 0.00000519, 0.00000035, -0.00000487, -0.00000794, },
		{ 0.00000192, -0.00000262, -0.00000612, -0.00000760, -0.00000696, -0.00000435, -0.00000032, 0.00000363, 0.00000522, 0.00000318, -0.00000091, -0.00000348, -0.00000233, 0.00000061, 0.00000065, -0.00000471, -0.00001239, -0.00001586, -0.00001144, -0.00000224, 0.00000475, 0.00000554, 0.00000284, 0.00000233, 0.00000574, 0.00000873, 0.00000612, -0.00000128, -0.00000660, -0.00000445, 0.00000245, 0.00000548, -0.00000015, -0.00000911, -0.00001098, -0.00000206, 0.00000899, 0.00000916, -0.00000460, -0.00002046, -0.00002272, -0.00000817, 0.00000949, 0.00001302, -0.00000049, -0.00001608, -0.00001616, 0.00000103, 0.00001934, 0.00002134, 0.00000654, -0.00000832, -0.00000723, 0.00000787, 0.00001914, 0.00001201, -0.00000868, -0.00002294, -0.00001580, 0.00000754, 0.00002688, 0.00002670, 0.00000960, -0.00000890, -0.00001687, -0.00001533, -0.00001199, -0.00000911, -0.00000109, 0.00001520, 0.00003083, 0.00002981, 0.00000585, -0.00002698, -0.00004336, -0.00002929, 0.00000398, 0.00002919, 0.00002752, 0.00000489, -0.00001533, -0.00001510, 0.00000237, 0.00001829, 0.00001778, 0.00000334, -0.00001002, -0.00001086, -0.00000152, 0.00000664, 0.00000593, -0.00000090, -0.00000559, -0.00000397, 0.00000050, 0.00000186, -0.00000141, -0.00000541, -0.00000596, -0.00000311, -0.00000013, 0.00000106, 0.00000202, 0.00000502, 0.00000909, 0.00001018, 0.00000560, -0.00000224, -0.00000740, -0.00000593, 0.00000027, 0.00000524, 0.00000457, -0.00000083, -0.00000620, -0.00000748, -0.00000451, -0.00000023, 0.00000235, 0.00000232, 0.00000065, -0.00000112, -0.00000176, -0.00000065, 0.00000206, 0.00000517, 0.00000681, 0.00000564, },
		{ 0.00000363, 0.00000540, 0.00000527, 0.00000325, 0.00000006, -0.00000353, -0.00000682, -0.00000855, -0.00000717, -0.00000234, 0.00000362, 0.00000673, 0.00000470, -0.00000053, -0.00000418, -0.00000327, 0.00000019, 0.00000101, -0.00000357, -0.00001061, -0.00001408, -0.00001098, -0.00000427, 0.00000065, 0.00000182, 0.00000238, 0.00000583, 0.00001068, 0.00001134, 0.00000482, -0.00000441, -0.00000795, -0.00000257, 0.00000532, 0.00000590, -0.00000303, -0.00001244, -0.00001117, 0.00000116, 0.00001201, 0.00000828, -0.00000889, -0.00002337, -0.00001977, -0.00000002, 0.00001706, 0.00001426, -0.00000596, -0.00002297, -0.00001849, 0.00000452, 0.00002441, 0.00002256, 0.00000215, -0.00001533, -0.00001231, 0.00000710, 0.00002181, 0.00001580, -0.00000566, -0.00002179, -0.00001741, 0.00000273, 0.00002015, 0.00002137, 0.00000909, -0.00000410, -0.00001029, -0.00001205, -0.00001453, -0.00001546, -0.00000659, 0.00001394, 0.00003368, 0.00003409, 0.00000900, -0.00002585, -0.00004404, -0.00003145, 0.00000092, 0.00002686, 0.00002827, 0.00000964, -0.00000895, -0.00001181, -0.00000061, 0.00001065, 0.00001119, 0.00000293, -0.00000420, -0.00000386, 0.00000108, 0.00000359, 0.00000101, -0.00000282, -0.00000307, 0.00000012, 0.00000219, -0.00000010, -0.00000472, -0.00000682, -0.00000434, -0.00000013, 0.00000166, 0.00000051, -0.00000009, 0.00000291, 0.00000801, 0.00001028, 0.00000658, -0.00000082, -0.00000617, -0.00000569, -0.00000085, 0.00000338, 0.00000336, -0.00000038, -0.00000437, -0.00000569, -0.00000406, -0.00000121, 0.00000103, 0.00000181, 0.00000125, 0.00000000, -0.00000099, -0.00000074, 0.00000107, },
		{ 0.00000001, -0.00000020, 0.00000064, 0.00000211, 0.00000344, 0.00000418, 0.00000414, 0.00000282, -0.00000049, -0.00000546, -0.00000968, -0.00000980, -0.00000458, 0.00000303, 0.00000763, 0.00000602, 0.00000026, -0.00000421, -0.00000393, -0.00000072, 0.00000051, -0.00000280, -0.00000816, -0.00001093, -0.00000945, -0.00000631, -0.00000446, -0.00000307, 0.00000133, 0.00000936, 0.00001583, 0.00001401, 0.00000356, -0.00000723, -0.00000879, -0.00000047, 0.00000780, 0.00000563, -0.00000617, -0.00001520, -0.00001048, 0.00000467, 0.00001437, 0.00000643, -0.00001342, -0.00002549, -0.00001581, 0.00000824, 0.00002357, 0.00001417, -0.00001186, -0.00002913, -0.00001986, 0.00000813, 0.00002867, 0.00002301, -0.00000193, -0.00002107, -0.00001631, 0.00000611, 0.00002310, 0.00001829, -0.00000260, -0.00001913, -0.00001712, -0.00000108, 0.00001296, 0.00001456, 0.00000700, -0.00000020, -0.00000348, -0.00000740, -0.00001506, -0.00002011, -0.00001184, 0.00001102, 0.00003372, 0.00003616, 0.00001201, -0.00002277, -0.00004217, -0.00003230, -0.00000285, 0.00002261, 0.00002741, 0.00001410, -0.00000168, -0.00000729, -0.00000283, 0.00000308, 0.00000415, 0.00000181, 0.00000091, 0.00000276, 0.00000392, 0.00000135, -0.00000310, -0.00000462, -0.00000128, 0.00000324, 0.00000355, -0.00000141, -0.00000697, -0.00000774, -0.00000321, 0.00000189, 0.00000300, 0.00000046, -0.00000119, 0.00000139, 0.00000656, 0.00000936, 0.00000671, 0.00000050, -0.00000437, -0.00000470, -0.00000153, 0.00000151, 0.00000181, -0.00000034, -0.00000284, -0.00000394, -0.00000337, -0.00000183, -0.00000015, 0.00000106, 0.00000139, 0.00000084, },
		{ 0.00000027, 0.00000112, 0.00000129, 0.00000102, 0.00000080, 0.00000076, 0.00000084, 0.00000127, 0.00000257, 0.00000450, 0.00000528, 0.00000262, -0.00000363, -0.00001010, -0.00001185, -0.00000680, 0.00000180, 0.00000765, 0.00000689, 0.00000139, -0.00000346, -0.00000414, -0.00000199, -0.00000086, -0.00000253, -0.00000525, -0.00000661, -0.00000689, -0.00000813, -0.00001012, -0.00000875, -0.00000031, 0.00001254, 0.00002069, 0.00001651, 0.00000253, -0.00000936, -0.00000892, 0.00000172, 0.00000965, 0.00000461, -0.00000935, -0.00001719, -0.00000896, 0.00000817, 0.00001583, 0.00000373, -0.00001781, -0.00002653, -0.00001095, 0.00001614, 0.00002861, 0.00001277, -0.00001773, -0.00003407, -0.00002007, 0.00001163, 0.00003172, 0.00002248, -0.00000566, -0.00002531, -0.00001901, 0.00000506, 0.00002304, 0.00001936, 0.00000020, -0.00001534, -0.00001511, -0.00000368, 0.00000588, 0.00000682, 0.00000344, 0.00000237, 0.00000282, -0.00000192, -0.00001355, -0.00002258, -0.00001620, 0.00000688, 0.00003105, 0.00003576, 0.00001443, -0.00001818, -0.00003795, -0.00003166, -0.00000681, 0.00001693, 0.00002503, 0.00001782, 0.00000580, -0.00000197, -0.00000409, -0.00000377, -0.00000274, 0.00000008, 0.00000492, 0.00000848, 0.00000676, 0.00000006, -0.00000606, -0.00000605, -0.00000023, 0.00000519, 0.00000439, -0.00000212, -0.00000810, -0.00000803, -0.00000249, 0.00000295, 0.00000374, 0.00000073, -0.00000141, 0.00000053, 0.00000502, 0.00000774, 0.00000607, 0.00000148, -0.00000238, -0.00000322, -0.00000169, -0.00000004, 0.00000026, -0.00000065, -0.00000181, -0.00000252, -0.00000261, -0.00000207, -0.00000099, },
		{ -0.00000198, -0.00000197, -0.00000137, -0.00000039, 0.00000057, 0.00000131, 0.00000185, 0.00000200, 0.00000133, -0.00000004, -0.00000078, 0.00000072, 0.00000432, 0.00000712, 0.00000541, -0.00000153, -0.00000967, -0.00001299, -0.00000870, 0.00000000, 0.00000664, 0.00000710, 0.00000272, -0.00000193, -0.00000379, -0.00000344, -0.00000294, -0.00000278, -0.00000213, -0.00000148, -0.00000348, -0.00000955, -0.00001584, -0.00001477, -0.00000246, 0.00001501, 0.00002475, 0.00001858, 0.00000184, -0.00001052, -0.00000821, 0.00000391, 0.00001070, 0.00000285, -0.00001240, -0.00001826, -0.00000670, 0.00001139, 0.00001625, 0.00000033, -0.00002170, -0.00002631, -0.00000540, 0.00002325, 0.00003187, 0.00001016, -0.00002319, -0.00003740, -0.00001907, 0.00001484, 0.00003333, 0.00002085, -0.00000903, -0.00002800, -0.00002033, 0.00000407, 0.00002178, 0.00001903, 0.00000250, -0.00001086, -0.00001173, -0.00000499, -0.00000056, -0.00000119, -0.00000128, 0.00000335, 0.00000798, 0.00000379, -0.00001025, -0.00002263, -0.00001913, 0.00000210, 0.00002602, 0.00003289, 0.00001586, -0.00001267, -0.00003181, -0.00002949, -0.00001045, 0.00001045, 0.00002138, 0.00002047, 0.00001280, 0.00000366, -0.00000433, -0.00000935, -0.00000891, -0.00000204, 0.00000757, 0.00001286, 0.00000932, -0.00000024, -0.00000765, -0.00000692, 0.00000014, 0.00000592, 0.00000460, -0.00000234, -0.00000814, -0.00000765, -0.00000204, 0.00000314, 0.00000385, 0.00000111, -0.00000094, 0.00000032, 0.00000364, 0.00000575, 0.00000485, 0.00000197, -0.00000058, -0.00000155, -0.00000135, -0.00000100, -0.00000097, -0.00000114, -0.00000136, -0.00000167, },
		{ -0.00000159, -0.00000144, -0.00000149, -0.00000168, -0.00000166, -0.00000130, -0.00000075, -0.00000006, 0.00000099, 0.00000235, 0.00000314, 0.00000216, -0.00000044, -0.00000239, -0.00000104, 0.00000370, 0.00000812, 0.00000749, 0.00000058, -0.00000840, -0.00001299, -0.00001001, -0.00000219, 0.00000459, 0.00000647, 0.00000402, 0.00000024, -0.00000282, -0.00000484, -0.00000546, -0.00000349, 0.00000093, 0.00000402, 0.00000047, -0.00001041, -0.00002110, -0.00002057, -0.00000494, 0.00001646, 0.00002753, 0.00001994, 0.00000156, -0.00001048, -0.00000654, 0.00000604, 0.00001088, 0.00000045, -0.00001511, -0.00001834, -0.00000388, 0.00001408, 0.00001554, -0.00000354, -0.00002479, -0.00002473, 0.00000059, 0.00002917, 0.00003318, 0.00000650, -0.00002788, -0.00003892, -0.00001689, 0.00001757, 0.00003339, 0.00001809, -0.00001204, -0.00002920, -0.00002030, 0.00000327, 0.00001957, 0.00001748, 0.00000415, -0.00000616, -0.00000744, -0.00000508, -0.00000595, -0.00000885, -0.00000674, 0.00000268, 0.00001153, 0.00000911, -0.00000558, -0.00002033, -0.00002030, -0.00000273, 0.00001925, 0.00002784, 0.00001606, -0.00000689, -0.00002437, -0.00002592, -0.00001332, 0.00000388, 0.00001687, 0.00002181, 0.00001867, 0.00000903, -0.00000363, -0.00001327, -0.00001381, -0.00000429, 0.00000876, 0.00001559, 0.00001132, 0.00000033, -0.00000781, -0.00000707, 0.00000001, 0.00000553, 0.00000415, -0.00000224, -0.00000727, -0.00000664, -0.00000173, 0.00000267, 0.00000341, 0.00000145, -0.00000006, 0.00000063, 0.00000257, 0.00000375, 0.00000330, 0.00000194, 0.00000073, -0.00000004, -0.00000064, -0.00000124, -0.00000160, },
		{ 0.00000023, -0.00000078, -0.00000149, -0.00000179, -0.00000192, -0.00000201, -0.00000184, -0.00000134, -0.00000086, -0.00000073, -0.00000058, 0.00000046, 0.00000249, 0.00000398, 0.00000301, -0.00000039, -0.00000334, -0.00000238, 0.00000280, 0.00000820, 0.00000858, 0.00000242, -0.00000640, -0.00001175, -0.00001049, -0.00000454, 0.00000163, 0.00000494, 0.00000505, 0.00000277, -0.00000132, -0.00000598, -0.00000809, -0.00000448, 0.00000372, 0.00000938, 0.00000458, -0.00001069, -0.00002542, -0.00002560, -0.00000753, 0.00001671, 0.00002868, 0.00002034, 0.00000167, -0.00000918, -0.00000390, 0.00000806, 0.00001023, -0.00000244, -0.00001733, -0.00001745, -0.00000071, 0.00001601, 0.00001372, -0.00000764, -0.00002683, -0.00002182, 0.00000674, 0.00003364, 0.00003254, 0.00000207, -0.00003153, -0.00003857, -0.00001367, 0.00001970, 0.00003191, 0.00001434, -0.00001467, -0.00002906, -0.00001909, 0.00000271, 0.00001671, 0.00001497, 0.00000508, -0.00000165, -0.00000273, -0.00000415, -0.00001003, -0.00001560, -0.00001243, 0.00000052, 0.00001321, 0.00001353, -0.00000012, -0.00001601, -0.00001960, -0.00000703, 0.00001149, 0.00002109, 0.00001492, -0.00000148, -0.00001638, -0.00002127, -0.00001508, -0.00000211, 0.00001196, 0.00002180, 0.00002291, 0.00001362, -0.00000217, -0.00001530, -0.00001707, -0.00000635, 0.00000860, 0.00001656, 0.00001254, 0.00000154, -0.00000667, -0.00000644, -0.00000041, 0.00000425, 0.00000311, -0.00000197, -0.00000577, -0.00000515, -0.00000145, 0.00000179, 0.00000255, 0.00000163, 0.00000094, 0.00000126, 0.00000190, 0.00000205, 0.00000173, 0.00000146, 0.00000136, 0.00000104, },
		{ 0.00000068, 0.00000126, 0.00000147, 0.00000105, 0.00000028, -0.00000058, -0.00000157, -0.00000259, -0.00000309, -0.00000253, -0.00000120, -0.00000023, -0.00000035, -0.00000082, -0.00000006, 0.00000229, 0.00000436, 0.00000366, 0.00000001, -0.00000350, -0.00000305, 0.00000185, 0.00000738, 0.00000853, 0.00000372, -0.00000389, -0.00000934, -0.00001000, -0.00000674, -0.00000196, 0.00000257, 0.00000558, 0.00000532, 0.00000055, -0.00000667, -0.00001042, -0.00000549, 0.00000610, 0.00001415, 0.00000842, -0.00001040, -0.00002841, -0.00002936, -0.00000994, 0.00001570, 0.00002794, 0.00001958, 0.00000205, -0.00000669, -0.00000037, 0.00000995, 0.00000889, -0.00000556, -0.00001890, -0.00001569, 0.00000253, 0.00001701, 0.00001090, -0.00001169, -0.00002766, -0.00001772, 0.00001271, 0.00003645, 0.00003008, -0.00000283, -0.00003395, -0.00003643, -0.00000966, 0.00002111, 0.00002903, 0.00000979, -0.00001692, -0.00002782, -0.00001693, 0.00000241, 0.00001351, 0.00001184, 0.00000534, 0.00000231, 0.00000191, -0.00000250, -0.00001269, -0.00002102, -0.00001786, -0.00000280, 0.00001300, 0.00001663, 0.00000551, -0.00001022, -0.00001716, -0.00001035, 0.00000356, 0.00001334, 0.00001255, 0.00000300, -0.00000862, -0.00001596, -0.00001556, -0.00000695, 0.00000716, 0.00002051, 0.00002520, 0.00001697, -0.00000024, -0.00001545, -0.00001843, -0.00000792, 0.00000732, 0.00001584, 0.00001285, 0.00000311, -0.00000451, -0.00000510, -0.00000090, 0.00000241, 0.00000165, -0.00000168, -0.00000396, -0.00000340, -0.00000112, 0.00000079, 0.00000149, 0.00000161, 0.00000182, 0.00000198, 0.00000161, 0.00000086, 0.00000041, },
		{ 0.00000162, 0.00000029, -0.00000043, -0.00000021, 0.00000050, 0.00000116, 0.00000160, 0.00000170, 0.00000101, -0.00000082, -0.00000318, -0.00000452, -0.00000372, -0.00000142, 0.00000035, 0.00000030, -0.00000066, -0.00000040, 0.00000185, 0.00000420, 0.00000391, 0.00000059, -0.00000286, -0.00000289, 0.00000107, 0.00000582, 0.00000733, 0.00000429, -0.00000118, -0.00000599, -0.00000851, -0.00000850, -0.00000581, -0.00000048, 0.00000543, 0.00000752, 0.00000254, -0.00000682, -0.00001210, -0.00000624, 0.00000801, 0.00001793, 0.00001158, -0.00000968, -0.00002985, -0.00003148, -0.00001191, 0.00001352, 0.00002528, 0.00001756, 0.00000255, -0.00000323, 0.00000384, 0.00001171, 0.00000708, -0.00000862, -0.00001971, -0.00001323, 0.00000556, 0.00001697, 0.00000726, -0.00001538, -0.00002721, -0.00001269, 0.00001817, 0.00003752, 0.00002604, -0.00000783, -0.00003501, -0.00003274, -0.00000516, 0.00002175, 0.00002500, 0.00000473, -0.00001876, -0.00002574, -0.00001415, 0.00000235, 0.00001027, 0.00000845, 0.00000504, 0.00000550, 0.00000604, -0.00000046, -0.00001397, -0.00002482, -0.00002255, -0.00000684, 0.00001110, 0.00001820, 0.00001071, -0.00000365, -0.00001330, -0.00001239, -0.00000374, 0.00000536, 0.00000918, 0.00000614, -0.00000178, -0.00001048, -0.00001473, -0.00001021, 0.00000295, 0.00001815, 0.00002543, 0.00001878, 0.00000185, -0.00001389, -0.00001790, -0.00000876, 0.00000527, 0.00001370, 0.00001220, 0.00000470, -0.00000171, -0.00000320, -0.00000126, 0.00000042, 0.00000001, -0.00000145, -0.00000216, -0.00000165, -0.00000074, -0.00000011, 0.00000047, 0.00000141, 0.00000238, 0.00000255, },
		{ 0.00000110, 0.00000252, 0.00000281, 0.00000180, 0.00000034, -0.00000067, -0.00000101, -0.00000076, 0.00000014, 0.00000172, 0.00000320, 0.00000306, 0.00000044, -0.00000348, -0.00000601, -0.00000527, -0.00000210, 0.00000067, 0.00000107, -0.00000009, -0.00000038, 0.00000133, 0.00000351, 0.00000362, 0.00000118, -0.00000156, -0.00000186, 0.00000065, 0.00000378, 0.00000510, 0.00000402, 0.00000142, -0.00000203, -0.00000612, -0.00000958, -0.00000951, -0.00000393, 0.00000453, 0.00000900, 0.00000434, -0.00000643, -0.00001284, -0.00000643, 0.00000945, 0.00002044, 0.00001371, -0.00000869, -0.00002966, -0.00003173, -0.00001317, 0.00001038, 0.00002083, 0.00001428, 0.00000298, 0.00000086, 0.00000847, 0.00001332, 0.00000509, -0.00001129, -0.00001966, -0.00001028, 0.00000811, 0.00001585, 0.00000304, -0.00001848, -0.00002549, -0.00000704, 0.00002281, 0.00003687, 0.00002079, -0.00001257, -0.00003470, -0.00002780, -0.00000050, 0.00002157, 0.00002010, -0.00000052, -0.00002015, -0.00002311, -0.00001104, 0.00000245, 0.00000722, 0.00000513, 0.00000434, 0.00000777, 0.00000934, 0.00000163, -0.00001406, -0.00002694, -0.00002616, -0.00001110, 0.00000786, 0.00001819, 0.00001492, 0.00000295, -0.00000851, -0.00001300, -0.00000973, -0.00000207, 0.00000517, 0.00000774, 0.00000357, -0.00000533, -0.00001273, -0.00001168, -0.00000034, 0.00001501, 0.00002368, 0.00001892, 0.00000379, -0.00001098, -0.00001563, -0.00000871, 0.00000288, 0.00001054, 0.00001070, 0.00000600, 0.00000126, -0.00000098, -0.00000133, -0.00000135, -0.00000152, -0.00000134, -0.00000067, -0.00000017, -0.00000034, -0.00000071, -0.00000030, },
		{ -0.00000002, -0.00000093, -0.00000064, 0.00000080, 0.00000223, 0.00000263, 0.00000199, 0.00000090, -0.00000029, -0.00000152, -0.00000225, -0.00000142, 0.00000134, 0.00000450, 0.00000527, 0.00000212, -0.00000332, -0.00000724, -0.00000695, -0.00000324, 0.00000052, 0.00000174, 0.00000082, 0.00000008, 0.00000091, 0.00000241, 0.00000279, 0.00000161, 0.00000021, -0.00000006, 0.00000071, 0.00000155, 0.00000213, 0.00000294, 0.00000359, 0.00000211, -0.00000304, -0.00000980, -0.00001261, -0.00000744, 0.00000287, 0.00000949, 0.00000564, -0.00000556, -0.00001245, -0.00000585, 0.00001052, 0.00002157, 0.00001460, -0.00000761, -0.00002792, -0.00003007, -0.00001356, 0.00000659, 0.00001491, 0.00000987, 0.00000313, 0.00000513, 0.00001312, 0.00001475, 0.00000321, -0.00001322, -0.00001869, -0.00000706, 0.00000993, 0.00001368, -0.00000149, -0.00002075, -0.00002261, -0.00000112, 0.00002636, 0.00003460, 0.00001474, -0.00001671, -0.00003304, -0.00002198, 0.00000396, 0.00002058, 0.00001466, -0.00000563, -0.00002107, -0.00002019, -0.00000793, 0.00000262, 0.00000453, 0.00000216, 0.00000340, 0.00000911, 0.00001159, 0.00000347, -0.00001323, -0.00002743, -0.00002845, -0.00001508, 0.00000380, 0.00001672, 0.00001775, 0.00000884, -0.00000337, -0.00001223, -0.00001389, -0.00000827, 0.00000099, 0.00000777, 0.00000706, -0.00000095, -0.00000985, -0.00001139, -0.00000247, 0.00001145, 0.00002026, 0.00001744, 0.00000532, -0.00000717, -0.00001197, -0.00000774, 0.00000057, 0.00000687, 0.00000855, 0.00000679, 0.00000394, 0.00000124, -0.00000105, -0.00000257, -0.00000265, -0.00000130, 0.00000032, 0.00000080, },
		{ -0.00000126, 0.00000068, 0.00000108, 0.00000011, -0.00000076, -0.00000049, 0.00000063, 0.00000160, 0.00000200, 0.00000205, 0.00000178, 0.00000062, -0.00000164, -0.00000369, -0.00000327, 0.00000049, 0.00000535, 0.00000729, 0.00000401, -0.00000262, -0.00000794, -0.00000849, -0.00000475, -0.00000022, 0.00000209, 0.00000191, 0.00000098, 0.00000073, 0.00000111, 0.00000148, 0.00000178, 0.00000220, 0.00000232, 0.00000130, -0.00000057, -0.00000122, 0.00000116, 0.00000511, 0.00000599, 0.00000039, -0.00000912, -0.00001476, -0.00001062, 0.00000060, 0.00000882, 0.00000617, -0.00000439, -0.00001092, -0.00000433, 0.00001133, 0.00002135, 0.00001417, -0.00000662, -0.00002486, -0.00002665, -0.00001299, 0.00000251, 0.00000800, 0.00000459, 0.00000284, 0.00000912, 0.00001739, 0.00001595, 0.00000173, -0.00001412, -0.00001677, -0.00000379, 0.00001085, 0.00001055, -0.00000604, -0.00002204, -0.00001876, 0.00000467, 0.00002861, 0.00003092, 0.00000834, -0.00001995, -0.00003017, -0.00001570, 0.00000792, 0.00001883, 0.00000901, -0.00001033, -0.00002149, -0.00001721, -0.00000507, 0.00000278, 0.00000231, -0.00000027, 0.00000241, 0.00000958, 0.00001273, 0.00000481, -0.00001180, -0.00002653, -0.00002930, -0.00001832, -0.00000056, 0.00001406, 0.00001895, 0.00001341, 0.00000154, -0.00001027, -0.00001596, -0.00001270, -0.00000293, 0.00000642, 0.00000859, 0.00000237, -0.00000642, -0.00000955, -0.00000342, 0.00000777, 0.00001560, 0.00001459, 0.00000625, -0.00000298, -0.00000742, -0.00000593, -0.00000128, 0.00000324, 0.00000603, 0.00000690, 0.00000591, 0.00000315, -0.00000044, -0.00000300, -0.00000313, },
		{ 0.00000042, -0.00000253, -0.00000279, -0.00000112, 0.00000043, 0.00000062, -0.00000003, -0.00000029, 0.00000017, 0.00000067, 0.00000078, 0.00000100, 0.00000190, 0.00000276, 0.00000189, -0.00000133, -0.00000482, -0.00000509, -0.00000072, 0.00000558, 0.00000877, 0.00000584, -0.00000140, -0.00000790, -0.00000960, -0.00000645, -0.00000155, 0.00000194, 0.00000296, 0.00000223, 0.00000089, -0.00000019, -0.00000011, 0.00000163, 0.00000413, 0.00000497, 0.00000234, -0.00000232, -0.00000454, -0.00000108, 0.00000582, 0.00000918, 0.00000382, -0.00000760, -0.00001570, -0.00001310, -0.00000208, 0.00000696, 0.00000575, -0.00000310, -0.00000835, -0.00000188, 0.00001203, 0.00001995, 0.00001252, -0.00000585, -0.00002081, -0.00002180, -0.00001150, -0.00000150, 0.00000066, -0.00000120, 0.00000200, 0.00001236, 0.00002085, 0.00001683, 0.00000087, -0.00001376, -0.00001390, -0.00000066, 0.00001077, 0.00000662, -0.00001033, -0.00002228, -0.00001419, 0.00000997, 0.00002944, 0.00002609, 0.00000201, -0.00002209, -0.00002625, -0.00000935, 0.00001112, 0.00001640, 0.00000343, -0.00001435, -0.00002142, -0.00001436, -0.00000265, 0.00000284, 0.00000060, -0.00000205, 0.00000151, 0.00000931, 0.00001277, 0.00000550, -0.00001010, -0.00002453, -0.00002876, -0.00002049, -0.00000468, 0.00001055, 0.00001845, 0.00001623, 0.00000569, -0.00000742, -0.00001590, -0.00001506, -0.00000618, 0.00000402, 0.00000826, 0.00000446, -0.00000283, -0.00000656, -0.00000330, 0.00000425, 0.00001022, 0.00001075, 0.00000653, 0.00000110, -0.00000251, -0.00000348, -0.00000238, 0.00000015, 0.00000349, 0.00000632, 0.00000687, 0.00000443, },
		{ 0.00000511, 0.00000667, 0.00000488, 0.00000136, -0.00000120, -0.00000158, -0.00000076, -0.00000031, -0.00000055, -0.00000052, 0.00000034, 0.00000121, 0.00000098, -0.00000006, -0.00000022, 0.00000150, 0.00000361, 0.00000325, -0.00000065, -0.00000541, -0.00000655, -0.00000208, 0.00000512, 0.00000943, 0.00000731, 0.00000019, -0.00000701, -0.00001004, -0.00000810, -0.00000340, 0.00000117, 0.00000373, 0.00000365, 0.00000140, -0.00000124, -0.00000176, 0.00000120, 0.00000576, 0.00000755, 0.00000368, -0.00000350, -0.00000744, -0.00000347, 0.00000570, 0.00001137, 0.00000684, -0.00000542, -0.00001533, -0.00001460, -0.00000487, 0.00000406, 0.00000430, -0.00000193, -0.00000500, 0.00000138, 0.00001270, 0.00001767, 0.00000987, -0.00000535, -0.00001616, -0.00001597, -0.00000923, -0.00000509, -0.00000651, -0.00000707, 0.00000060, 0.00001447, 0.00002312, 0.00001730, 0.00000076, -0.00001203, -0.00001016, 0.00000219, 0.00000972, 0.00000212, -0.00001414, -0.00002149, -0.00000920, 0.00001446, 0.00002882, 0.00002043, -0.00000384, -0.00002300, -0.00002153, -0.00000329, 0.00001338, 0.00001339, -0.00000181, -0.00001754, -0.00002085, -0.00001175, -0.00000080, 0.00000273, -0.00000064, -0.00000315, 0.00000079, 0.00000849, 0.00001186, 0.00000546, -0.00000844, -0.00002179, -0.00002695, -0.00002137, -0.00000812, 0.00000658, 0.00001636, 0.00001706, 0.00000869, -0.00000406, -0.00001388, -0.00001529, -0.00000846, 0.00000099, 0.00000644, 0.00000536, 0.00000058, -0.00000292, -0.00000241, 0.00000107, 0.00000465, 0.00000639, 0.00000618, 0.00000465, 0.00000219, -0.00000066, -0.00000257, -0.00000198, 0.00000128, },
		{ -0.00000183, -0.00000288, -0.00000032, 0.00000344, 0.00000532, 0.00000438, 0.00000222, 0.00000082, 0.00000043, -0.00000011, -0.00000138, -0.00000226, -0.00000136, 0.00000094, 0.00000245, 0.00000153, -0.00000074, -0.00000144, 0.00000090, 0.00000415, 0.00000445, 0.00000028, -0.00000536, -0.00000737, -0.00000333, 0.00000404, 0.00000912, 0.00000814, 0.00000194, -0.00000532, -0.00000962, -0.00000940, -0.00000555, -0.00000023, 0.00000401, 0.00000498, 0.00000222, -0.00000188, -0.00000318, 0.00000060, 0.00000692, 0.00000974, 0.00000512, -0.00000404, -0.00000958, -0.00000569, 0.00000485, 0.00001234, 0.00000913, -0.00000285, -0.00001368, -0.00001494, -0.00000746, 0.00000038, 0.00000192, -0.00000105, -0.00000122, 0.00000521, 0.00001342, 0.00001485, 0.00000660, -0.00000510, -0.00001129, -0.00000970, -0.00000643, -0.00000802, -0.00001294, -0.00001256, -0.00000131, 0.00001519, 0.00002390, 0.00001723, 0.00000144, -0.00000893, -0.00000566, 0.00000468, 0.00000780, -0.00000265, -0.00001726, -0.00001977, -0.00000414, 0.00001785, 0.00002681, 0.00001429, -0.00000889, -0.00002263, -0.00001627, 0.00000217, 0.00001459, 0.00000995, -0.00000651, -0.00001980, -0.00001984, -0.00000947, 0.00000046, 0.00000244, -0.00000146, -0.00000364, 0.00000030, 0.00000728, 0.00001018, 0.00000471, -0.00000704, -0.00001867, -0.00002411, -0.00002089, -0.00001056, 0.00000254, 0.00001289, 0.00001590, 0.00001030, -0.00000055, -0.00001029, -0.00001357, -0.00000961, -0.00000221, 0.00000361, 0.00000526, 0.00000353, 0.00000081, -0.00000113, -0.00000167, -0.00000061, 0.00000201, 0.00000534, 0.00000735, 0.00000621, 0.00000222, },
		{ 0.00000903, 0.00000918, 0.00000487, -0.00000025, -0.00000246, -0.00000109, 0.00000155, 0.00000298, 0.00000296, 0.00000284, 0.00000326, 0.00000304, 0.00000084, -0.00000253, -0.00000427, -0.00000251, 0.00000136, 0.00000366, 0.00000224, -0.00000112, -0.00000244, 0.00000019, 0.00000426, 0.00000525, 0.00000129, -0.00000464, -0.00000735, -0.00000423, 0.00000247, 0.00000777, 0.00000810, 0.00000357, -0.00000298, -0.00000828, -0.00001011, -0.00000774, -0.00000217, 0.00000362, 0.00000595, 0.00000319, -0.00000198, -0.00000412, 0.00000001, 0.00000754, 0.00001125, 0.00000639, -0.00000394, -0.00001074, -0.00000743, 0.00000344, 0.00001204, 0.00001042, -0.00000021, -0.00001097, -0.00001406, -0.00000957, -0.00000370, -0.00000120, -0.00000060, 0.00000261, 0.00000922, 0.00001415, 0.00001185, 0.00000312, -0.00000498, -0.00000656, -0.00000354, -0.00000341, -0.00001012, -0.00001816, -0.00001723, -0.00000359, 0.00001439, 0.00002303, 0.00001654, 0.00000282, -0.00000462, -0.00000058, 0.00000678, 0.00000523, -0.00000738, -0.00001953, -0.00001728, 0.00000065, 0.00001995, 0.00002356, 0.00000804, -0.00001287, -0.00002105, -0.00001076, 0.00000680, 0.00001476, 0.00000623, -0.00001051, -0.00002112, -0.00001845, -0.00000754, 0.00000113, 0.00000195, -0.00000194, -0.00000362, 0.00000006, 0.00000587, 0.00000796, 0.00000333, -0.00000608, -0.00001547, -0.00002049, -0.00001911, -0.00001183, -0.00000128, 0.00000838, 0.00001295, 0.00001048, 0.00000276, -0.00000564, -0.00001030, -0.00000963, -0.00000511, 0.00000039, 0.00000448, 0.00000585, 0.00000414, 0.00000011, -0.00000399, -0.00000519, -0.00000193, 0.00000419, },
		{ -0.00000884, -0.00000508, 0.00000294, 0.00000965, 0.00001088, 0.00000702, 0.00000196, -0.00000081, -0.00000096, -0.00000032, -0.00000002, 0.00000079, 0.00000311, 0.00000579, 0.00000593, 0.00000205, -0.00000353, -0.00000628, -0.00000383, 0.00000145, 0.00000460, 0.00000299, -0.00000112, -0.00000301, -0.00000048, 0.00000390, 0.00000545, 0.00000217, -0.00000333, -0.00000639, -0.00000455, 0.00000064, 0.00000549, 0.00000709, 0.00000480, -0.00000027, -0.00000609, -0.00001003, -0.00000964, -0.00000447, 0.00000251, 0.00000631, 0.00000412, -0.00000153, -0.00000437, -0.00000039, 0.00000761, 0.00001190, 0.00000726, -0.00000333, -0.00001081, -0.00000843, 0.00000172, 0.00001059, 0.00001059, 0.00000219, -0.00000751, -0.00001209, -0.00001097, -0.00000775, -0.00000472, -0.00000063, 0.00000608, 0.00001302, 0.00001481, 0.00000898, -0.00000011, -0.00000482, -0.00000224, 0.00000202, -0.00000051, -0.00001135, -0.00002182, -0.00002075, -0.00000605, 0.00001214, 0.00002051, 0.00001519, 0.00000475, 0.00000060, 0.00000481, 0.00000847, 0.00000230, -0.00001173, -0.00002086, -0.00001426, 0.00000484, 0.00002066, 0.00001928, 0.00000201, -0.00001563, -0.00001841, -0.00000527, 0.00001042, 0.00001394, 0.00000242, -0.00001375, -0.00002154, -0.00001676, -0.00000595, 0.00000130, 0.00000129, -0.00000219, -0.00000324, 0.00000002, 0.00000439, 0.00000542, 0.00000145, -0.00000562, -0.00001244, -0.00001637, -0.00001621, -0.00001192, -0.00000464, 0.00000322, 0.00000860, 0.00000937, 0.00000561, -0.00000051, -0.00000604, -0.00000867, -0.00000736, -0.00000261, 0.00000345, 0.00000748, 0.00000663, 0.00000091, -0.00000597, },
		{ 0.00000802, 0.00000094, -0.00000776, -0.00001145, -0.00000722, 0.00000178, 0.00000932, 0.00001130, 0.00000851, 0.00000454, 0.00000183, 0.00000003, -0.00000193, -0.00000331, -0.00000185, 0.00000299, 0.00000807, 0.00000873, 0.00000339, -0.00000419, -0.00000795, -0.00000513, 0.00000116, 0.00000505, 0.00000360, -0.00000074, -0.00000302, -0.00000095, 0.00000313, 0.00000498, 0.00000275, -0.00000160, -0.00000453, -0.00000416, -0.00000116, 0.00000250, 0.00000510, 0.00000539, 0.00000248, -0.00000322, -0.00000904, -0.00001095, -0.00000686, 0.00000069, 0.00000589, 0.00000479, -0.00000060, -0.00000383, -0.00000041, 0.00000723, 0.00001161, 0.00000755, -0.00000237, -0.00000984, -0.00000855, -0.00000001, 0.00000822, 0.00000965, 0.00000408, -0.00000371, -0.00000925, -0.00001152, -0.00001132, -0.00000825, -0.00000107, 0.00000886, 0.00001616, 0.00001525, 0.00000650, -0.00000270, -0.00000444, 0.00000150, 0.00000658, 0.00000197, -0.00001176, -0.00002377, -0.00002287, -0.00000847, 0.00000862, 0.00001649, 0.00001318, 0.00000699, 0.00000634, 0.00001023, 0.00000978, -0.00000068, -0.00001535, -0.00002117, -0.00001093, 0.00000816, 0.00001994, 0.00001423, -0.00000349, -0.00001711, -0.00001491, -0.00000008, 0.00001297, 0.00001227, -0.00000131, -0.00001617, -0.00002118, -0.00001485, -0.00000469, 0.00000108, 0.00000051, -0.00000229, -0.00000263, 0.00000013, 0.00000295, 0.00000275, -0.00000078, -0.00000565, -0.00000971, -0.00001200, -0.00001245, -0.00001098, -0.00000744, -0.00000220, 0.00000336, 0.00000727, 0.00000784, 0.00000453, -0.00000141, -0.00000699, -0.00000870, -0.00000484, 0.00000260, 0.00000847, },
		{ -0.00000586, 0.00000233, 0.00000898, 0.00000821, 0.00000001, -0.00000949, -0.00001307, -0.00000831, 0.00000085, 0.00000820, 0.00001057, 0.00000930, 0.00000721, 0.00000509, 0.00000171, -0.00000310, -0.00000643, -0.00000458, 0.00000250, 0.00000979, 0.00001105, 0.00000467, -0.00000440, -0.00000902, -0.00000621, 0.00000049, 0.00000487, 0.00000391, -0.00000005, -0.00000242, -0.00000110, 0.00000207, 0.00000387, 0.00000290, 0.00000033, -0.00000195, -0.00000300, -0.00000268, -0.00000086, 0.00000229, 0.00000516, 0.00000491, 0.00000004, -0.00000716, -0.00001142, -0.00000903, -0.00000168, 0.00000459, 0.00000497, 0.00000064, -0.00000249, 0.00000009, 0.00000656, 0.00001045, 0.00000715, -0.00000129, -0.00000800, -0.00000776, -0.00000150, 0.00000526, 0.00000774, 0.00000524, -0.00000002, -0.00000590, -0.00001121, -0.00001403, -0.00001134, -0.00000180, 0.00001071, 0.00001827, 0.00001530, 0.00000455, -0.00000434, -0.00000364, 0.00000459, 0.00000992, 0.00000378, -0.00001148, -0.00002403, -0.00002352, -0.00001064, 0.00000417, 0.00001131, 0.00001059, 0.00000926, 0.00001213, 0.00001532, 0.00001074, -0.00000338, -0.00001795, -0.00002047, -0.00000755, 0.00001039, 0.00001787, 0.00000870, -0.00000823, -0.00001735, -0.00001083, 0.00000459, 0.00001443, 0.00000994, -0.00000482, -0.00001781, -0.00002017, -0.00001284, -0.00000369, 0.00000060, -0.00000034, -0.00000232, -0.00000195, 0.00000031, 0.00000159, 0.00000011, -0.00000320, -0.00000609, -0.00000734, -0.00000760, -0.00000816, -0.00000927, -0.00000965, -0.00000750, -0.00000221, 0.00000459, 0.00000938, 0.00000893, 0.00000293, -0.00000494, -0.00000900, },
		{ 0.00000643, -0.00000289, -0.00000828, -0.00000544, 0.00000294, 0.00000919, 0.00000727, -0.00000190, -0.00001127, -0.00001385, -0.00000846, 0.00000019, 0.00000656, 0.00000897, 0.00000943, 0.00000969, 0.00000854, 0.00000377, -0.00000373, -0.00000898, -0.00000700, 0.00000177, 0.00001071, 0.00001251, 0.00000567, -0.00000413, -0.00000927, -0.00000685, -0.00000045, 0.00000401, 0.00000378, 0.00000081, -0.00000126, -0.00000085, 0.00000090, 0.00000223, 0.00000258, 0.00000223, 0.00000110, -0.00000115, -0.00000370, -0.00000423, -0.00000110, 0.00000406, 0.00000668, 0.00000330, -0.00000453, -0.00001090, -0.00001067, -0.00000435, 0.00000245, 0.00000452, 0.00000197, -0.00000047, 0.00000115, 0.00000579, 0.00000858, 0.00000604, -0.00000030, -0.00000557, -0.00000616, -0.00000255, 0.00000212, 0.00000515, 0.00000559, 0.00000316, -0.00000246, -0.00001011, -0.00001561, -0.00001359, -0.00000262, 0.00001151, 0.00001908, 0.00001478, 0.00000317, -0.00000485, -0.00000228, 0.00000705, 0.00001194, 0.00000479, -0.00001069, -0.00002279, -0.00002277, -0.00001240, -0.00000078, 0.00000540, 0.00000760, 0.00001131, 0.00001746, 0.00001971, 0.00001136, -0.00000551, -0.00001930, -0.00001878, -0.00000431, 0.00001142, 0.00001459, 0.00000299, -0.00001206, -0.00001648, -0.00000643, 0.00000855, 0.00001489, 0.00000716, -0.00000795, -0.00001871, -0.00001867, -0.00001080, -0.00000292, -0.00000004, -0.00000119, -0.00000233, -0.00000130, 0.00000048, 0.00000033, -0.00000242, -0.00000567, -0.00000680, -0.00000532, -0.00000334, -0.00000367, -0.00000711, -0.00001134, -0.00001235, -0.00000752, 0.00000177, 0.00001025, 0.00001230, },
		{ -0.00000075, 0.00001055, 0.00001435, 0.00000864, -0.00000119, -0.00000666, -0.00000351, 0.00000458, 0.00000932, 0.00000543, -0.00000464, -0.00001313, -0.00001403, -0.00000792, -0.00000023, 0.00000464, 0.00000686, 0.00000908, 0.00001177, 0.00001172, 0.00000590, -0.00000380, -0.00001061, -0.00000869, 0.00000096, 0.00001073, 0.00001286, 0.00000616, -0.00000345, -0.00000863, -0.00000688, -0.00000150, 0.00000254, 0.00000313, 0.00000166, 0.00000032, -0.00000018, -0.00000021, 0.00000028, 0.00000182, 0.00000388, 0.00000429, 0.00000120, -0.00000408, -0.00000722, -0.00000471, 0.00000215, 0.00000753, 0.00000620, -0.00000141, -0.00000935, -0.00001151, -0.00000703, -0.00000038, 0.00000335, 0.00000316, 0.00000199, 0.00000276, 0.00000511, 0.00000625, 0.00000434, 0.00000042, -0.00000288, -0.00000396, -0.00000303, -0.00000084, 0.00000220, 0.00000516, 0.00000551, 0.00000068, -0.00000843, -0.00001592, -0.00001470, -0.00000329, 0.00001127, 0.00001845, 0.00001354, 0.00000230, -0.00000424, -0.00000032, 0.00000898, 0.00001274, 0.00000498, -0.00000959, -0.00002037, -0.00002086, -0.00001364, -0.00000579, -0.00000070, 0.00000443, 0.00001292, 0.00002188, 0.00002309, 0.00001163, -0.00000688, -0.00001925, -0.00001618, -0.00000138, 0.00001126, 0.00001034, -0.00000263, -0.00001491, -0.00001470, -0.00000203, 0.00001169, 0.00001445, 0.00000415, -0.00001062, -0.00001896, -0.00001684, -0.00000883, -0.00000230, -0.00000071, -0.00000198, -0.00000236, -0.00000076, 0.00000058, -0.00000084, -0.00000476, -0.00000804, -0.00000762, -0.00000355, 0.00000066, 0.00000067, -0.00000487, -0.00001265, -0.00001649, -0.00001206, },
		{ -0.00001370, -0.00001973, -0.00001542, -0.00000258, 0.00001040, 0.00001495, 0.00000930, -0.00000009, -0.00000436, -0.00000023, 0.00000723, 0.00000955, 0.00000302, -0.00000792, -0.00001505, -0.00001388, -0.00000705, -0.00000054, 0.00000267, 0.00000463, 0.00000848, 0.00001332, 0.00001424, 0.00000771, -0.00000338, -0.00001110, -0.00000935, 0.00000030, 0.00000986, 0.00001195, 0.00000593, -0.00000250, -0.00000712, -0.00000620, -0.00000248, 0.00000062, 0.00000198, 0.00000233, 0.00000208, 0.00000083, -0.00000107, -0.00000169, 0.00000076, 0.00000512, 0.00000725, 0.00000380, -0.00000376, -0.00000950, -0.00000816, -0.00000039, 0.00000732, 0.00000838, 0.00000188, -0.00000688, -0.00001140, -0.00000938, -0.00000363, 0.00000151, 0.00000400, 0.00000462, 0.00000476, 0.00000465, 0.00000378, 0.00000224, 0.00000078, -0.00000026, -0.00000144, -0.00000291, -0.00000326, -0.00000071, 0.00000406, 0.00000680, 0.00000312, -0.00000643, -0.00001495, -0.00001447, -0.00000358, 0.00001016, 0.00001642, 0.00001152, 0.00000179, -0.00000264, 0.00000222, 0.00001049, 0.00001253, 0.00000447, -0.00000836, -0.00001719, -0.00001814, -0.00001433, -0.00001042, -0.00000645, 0.00000136, 0.00001396, 0.00002501, 0.00002516, 0.00001153, -0.00000738, -0.00001779, -0.00001278, 0.00000114, 0.00001002, 0.00000541, -0.00000791, -0.00001678, -0.00001227, 0.00000210, 0.00001393, 0.00001330, 0.00000112, -0.00001275, -0.00001867, -0.00001484, -0.00000697, -0.00000180, -0.00000133, -0.00000268, -0.00000242, -0.00000038, 0.00000056, -0.00000198, -0.00000692, -0.00001022, -0.00000838, -0.00000191, 0.00000431, 0.00000456, -0.00000292, },
		{ 0.00000755, 0.00000775, -0.00000153, -0.00001463, -0.00002193, -0.00001733, -0.00000348, 0.00000994, 0.00001411, 0.00000835, 0.00000026, -0.00000164, 0.00000410, 0.00001071, 0.00001000, 0.00000045, -0.00001132, -0.00001688, -0.00001363, -0.00000623, -0.00000093, 0.00000081, 0.00000266, 0.00000788, 0.00001430, 0.00001580, 0.00000891, -0.00000262, -0.00001037, -0.00000877, -0.00000002, 0.00000821, 0.00000978, 0.00000487, -0.00000151, -0.00000493, -0.00000480, -0.00000317, -0.00000150, 0.00000042, 0.00000266, 0.00000376, 0.00000204, -0.00000159, -0.00000341, -0.00000041, 0.00000586, 0.00000968, 0.00000633, -0.00000283, -0.00001083, -0.00001106, -0.00000329, 0.00000604, 0.00000955, 0.00000496, -0.00000374, -0.00001028, -0.00001111, -0.00000695, -0.00000088, 0.00000430, 0.00000707, 0.00000695, 0.00000451, 0.00000147, -0.00000001, 0.00000074, 0.00000198, 0.00000108, -0.00000226, -0.00000491, -0.00000321, 0.00000251, 0.00000698, 0.00000461, -0.00000438, -0.00001287, -0.00001288, -0.00000333, 0.00000840, 0.00001319, 0.00000876, 0.00000148, -0.00000032, 0.00000517, 0.00001171, 0.00001162, 0.00000347, -0.00000711, -0.00001366, -0.00001501, -0.00001455, -0.00001431, -0.00001134, -0.00000131, 0.00001436, 0.00002662, 0.00002574, 0.00001099, -0.00000704, -0.00001504, -0.00000876, 0.00000325, 0.00000793, 0.00000015, -0.00001264, -0.00001778, -0.00000951, 0.00000571, 0.00001527, 0.00001162, -0.00000173, -0.00001429, -0.00001793, -0.00001279, -0.00000528, -0.00000138, -0.00000185, -0.00000326, -0.00000250, -0.00000018, 0.00000036, -0.00000313, -0.00000891, -0.00001211, -0.00000889, -0.00000029, },
		{ -0.00001365, -0.00000902, 0.00000142, 0.00001034, 0.00001004, -0.00000092, -0.00001552, -0.00002304, -0.00001763, -0.00000332, 0.00000926, 0.00001196, 0.00000591, -0.00000017, 0.00000122, 0.00000902, 0.00001468, 0.00001071, -0.00000188, -0.00001436, -0.00001844, -0.00001347, -0.00000582, -0.00000161, -0.00000085, 0.00000124, 0.00000752, 0.00001472, 0.00001623, 0.00000927, -0.00000167, -0.00000849, -0.00000690, 0.00000018, 0.00000602, 0.00000651, 0.00000295, -0.00000071, -0.00000233, -0.00000280, -0.00000342, -0.00000349, -0.00000137, 0.00000257, 0.00000508, 0.00000324, -0.00000170, -0.00000462, -0.00000145, 0.00000611, 0.00001133, 0.00000848, -0.00000144, -0.00001109, -0.00001308, -0.00000622, 0.00000383, 0.00000957, 0.00000746, -0.00000025, -0.00000823, -0.00001199, -0.00000997, -0.00000357, 0.00000401, 0.00000903, 0.00000904, 0.00000467, -0.00000041, -0.00000212, 0.00000036, 0.00000361, 0.00000328, -0.00000124, -0.00000566, -0.00000503, 0.00000077, 0.00000611, 0.00000500, -0.00000257, -0.00000994, -0.00001005, -0.00000243, 0.00000631, 0.00000910, 0.00000538, 0.00000118, 0.00000235, 0.00000830, 0.00001273, 0.00001037, 0.00000227, -0.00000590, -0.00001019, -0.00001191, -0.00001441, -0.00001723, -0.00001496, -0.00000330, 0.00001419, 0.00002663, 0.00002476, 0.00000997, -0.00000600, -0.00001124, -0.00000432, 0.00000496, 0.00000529, -0.00000504, -0.00001663, -0.00001806, -0.00000674, 0.00000858, 0.00001574, 0.00000962, -0.00000423, -0.00001524, -0.00001686, -0.00001080, -0.00000379, -0.00000101, -0.00000224, -0.00000368, -0.00000258, -0.00000013, -0.00000003, -0.00000436, -0.00001075, },
		{ -0.00000064, -0.00000571, -0.00001247, -0.00001478, -0.00000867, 0.00000329, 0.00001262, 0.00001130, -0.00000120, -0.00001638, -0.00002303, -0.00001636, -0.00000215, 0.00000844, 0.00000874, 0.00000229, -0.00000126, 0.00000398, 0.00001403, 0.00001873, 0.00001163, -0.00000362, -0.00001658, -0.00001947, -0.00001350, -0.00000610, -0.00000280, -0.00000229, 0.00000052, 0.00000758, 0.00001466, 0.00001553, 0.00000868, -0.00000071, -0.00000566, -0.00000386, 0.00000097, 0.00000358, 0.00000247, 0.00000026, -0.00000030, 0.00000032, -0.00000040, -0.00000311, -0.00000507, -0.00000314, 0.00000203, 0.00000579, 0.00000419, -0.00000143, -0.00000513, -0.00000213, 0.00000596, 0.00001205, 0.00000995, 0.00000019, -0.00001030, -0.00001400, -0.00000881, 0.00000095, 0.00000841, 0.00000910, 0.00000320, -0.00000546, -0.00001188, -0.00001233, -0.00000625, 0.00000314, 0.00001025, 0.00001076, 0.00000506, -0.00000168, -0.00000380, -0.00000024, 0.00000451, 0.00000489, -0.00000005, -0.00000552, -0.00000596, -0.00000090, 0.00000439, 0.00000429, -0.00000121, -0.00000653, -0.00000628, -0.00000089, 0.00000420, 0.00000457, 0.00000163, 0.00000075, 0.00000497, 0.00001128, 0.00001358, 0.00000912, 0.00000120, -0.00000474, -0.00000711, -0.00000922, -0.00001408, -0.00001903, -0.00001707, -0.00000440, 0.00001356, 0.00002512, 0.00002231, 0.00000845, -0.00000448, -0.00000678, 0.00000028, 0.00000634, 0.00000248, -0.00000977, -0.00001976, -0.00001779, -0.00000426, 0.00001054, 0.00001540, 0.00000747, -0.00000626, -0.00001562, -0.00001556, -0.00000893, -0.00000252, -0.00000069, -0.00000249, -0.00000393, -0.00000263, -0.00000022, },
		},
		{ { 0.00000265, 0.00000240, 0.00000189, 0.00000161, 0.00000169, 0.00000167, 0.00000083, -0.00000101, -0.00000308, -0.00000387, -0.00000217, 0.00000200, 0.00000711, 0.00001094, 0.00001178, 0.00000933, 0.00000464, -0.00000068, -0.00000539, -0.00000891, -0.00001105, -0.00001154, -0.00000999, -0.00000643, -0.00000174, 0.00000241, 0.00000444, 0.00000393, 0.00000196, 0.00000037, 0.00000034, 0.00000146, 0.00000202, 0.00000056, -0.00000265, -0.00000543, -0.00000511, -0.00000075, 0.00000562, 0.00000990, 0.00000846, 0.00000073, -0.00001007, -0.00001869, -0.00002091, -0.00001593, -0.00000648, 0.00000321, 0.00000996, 0.00001293, 0.00001297, 0.00001101, 0.00000705, 0.00000073, -0.00000717, -0.00001382, -0.00001544, -0.00000987, 0.00000123, 0.00001238, 0.00001723, 0.00001257, 0.00000077, -0.00001128, -0.00001633, -0.00001128, 0.00000077, 0.00001257, 0.00001723, 0.00001238, 0.00000123, -0.00000987, -0.00001544, -0.00001382, -0.00000717, 0.00000073, 0.00000705, 0.00001101, 0.00001297, 0.00001293, 0.00000996, 0.00000321, -0.00000648, -0.00001593, -0.00002091, -0.00001869, -0.00001007, 0.00000073, 0.00000846, 0.00000990, 0.00000562, -0.00000075, -0.00000511, -0.00000543, -0.00000265, 0.00000056, 0.00000202, 0.00000146, 0.00000034, 0.00000037, 0.00000196, 0.00000393, 0.00000444, 0.00000241, -0.00000174, -0.00000643, -0.00000999, -0.00001154, -0.00001105, -0.00000891, -0.00000539, -0.00000068, 0.00000464, 0.00000933, 0.00001178, 0.00001094, 0.00000711, 0.00000200, -0.00000217, -0.00000387, -0.00000308, -0.00000101, 0.00000083, 0.00000167, 0.00000169, 0.00000161, 0.00000189, 0.00000240, },
		{ 0.00000174, 0.00000180, 0.00000225, 0.00000263, 0.00000253, 0.00000203, 0.00000158, 0.00000148, 0.00000146, 0.00000084, -0.00000079, -0.00000294, -0.00000431, -0.00000352, -0.00000017, 0.00000473, 0.00000921, 0.00001142, 0.00001059, 0.00000717, 0.00000231, -0.00000288, -0.00000764, -0.00001128, -0.00001293, -0.00001169, -0.00000735, -0.00000107, 0.00000471, 0.00000745, 0.00000607, 0.00000181, -0.00000249, -0.00000422, -0.00000280, 0.00000002, 0.00000157, 0.00000040, -0.00000243, -0.00000409, -0.00000218, 0.00000315, 0.00000887, 0.00001094, 0.00000701, -0.00000181, -0.00001156, -0.00001786, -0.00001844, -0.00001406, -0.00000733, -0.00000066, 0.00000500, 0.00000981, 0.00001355, 0.00001472, 0.00001130, 0.00000282, -0.00000808, -0.00001621, -0.00001673, -0.00000849, 0.00000450, 0.00001504, 0.00001697, 0.00000920, -0.00000350, -0.00001342, -0.00001474, -0.00000706, 0.00000463, 0.00001335, 0.00001448, 0.00000822, -0.00000127, -0.00000901, -0.00001211, -0.00001081, -0.00000706, -0.00000251, 0.00000245, 0.00000798, 0.00001329, 0.00001607, 0.00001366, 0.00000524, -0.00000675, -0.00001728, -0.00002146, -0.00001746, -0.00000774, 0.00000241, 0.00000804, 0.00000729, 0.00000205, -0.00000375, -0.00000678, -0.00000606, -0.00000295, 0.00000037, 0.00000257, 0.00000365, 0.00000435, 0.00000502, 0.00000521, 0.00000402, 0.00000105, -0.00000311, -0.00000719, -0.00001001, -0.00001115, -0.00001082, -0.00000938, -0.00000681, -0.00000287, 0.00000226, 0.00000753, 0.00001115, 0.00001161, 0.00000869, 0.00000378, -0.00000080, -0.00000316, -0.00000286, -0.00000092, 0.00000108, 0.00000206, 0.00000205, },
		{ 0.00000158, 0.00000287, 0.00000283, 0.00000220, 0.00000186, 0.00000211, 0.00000255, 0.00000264, 0.00000222, 0.00000165, 0.00000134, 0.00000125, 0.00000082, -0.00000047, -0.00000247, -0.00000416, -0.00000422, -0.00000193, 0.00000220, 0.00000664, 0.00000970, 0.00001039, 0.00000871, 0.00000522, 0.00000054, -0.00000482, -0.00001007, -0.00001379, -0.00001422, -0.00001030, -0.00000278, 0.00000548, 0.00001078, 0.00001065, 0.00000544, -0.00000173, -0.00000689, -0.00000772, -0.00000485, -0.00000115, 0.00000058, -0.00000029, -0.00000190, -0.00000151, 0.00000209, 0.00000738, 0.00001096, 0.00000995, 0.00000403, -0.00000429, -0.00001140, -0.00001491, -0.00001479, -0.00001261, -0.00000967, -0.00000587, -0.00000018, 0.00000747, 0.00001490, 0.00001817, 0.00001422, 0.00000356, -0.00000904, -0.00001688, -0.00001541, -0.00000539, 0.00000726, 0.00001506, 0.00001361, 0.00000433, -0.00000668, -0.00001265, -0.00001046, -0.00000218, 0.00000674, 0.00001124, 0.00000964, 0.00000399, -0.00000212, -0.00000615, -0.00000785, -0.00000826, -0.00000797, -0.00000616, -0.00000147, 0.00000612, 0.00001410, 0.00001828, 0.00001538, 0.00000548, -0.00000739, -0.00001732, -0.00001991, -0.00001476, -0.00000546, 0.00000272, 0.00000599, 0.00000387, -0.00000118, -0.00000576, -0.00000760, -0.00000640, -0.00000328, 0.00000040, 0.00000380, 0.00000659, 0.00000838, 0.00000853, 0.00000647, 0.00000231, -0.00000293, -0.00000770, -0.00001082, -0.00001199, -0.00001168, -0.00001040, -0.00000818, -0.00000462, 0.00000042, 0.00000606, 0.00001050, 0.00001192, 0.00000968, 0.00000492, -0.00000000, -0.00000287, -0.00000284, -0.00000078, },
		{ -0.00000333, -0.00000371, -0.00000136, 0.00000177, 0.00000383, 0.00000414, 0.00000329, 0.00000240, 0.00000217, 0.00000245, 0.00000266, 0.00000240, 0.00000182, 0.00000132, 0.00000106, 0.00000068, -0.00000033, -0.00000203, -0.00000372, -0.00000425, -0.00000282, 0.00000036, 0.00000418, 0.00000726, 0.00000873, 0.00000847, 0.00000672, 0.00000356, -0.00000113, -0.00000703, -0.00001270, -0.00001572, -0.00001385, -0.00000670, 0.00000334, 0.00001189, 0.00001495, 0.00001120, 0.00000289, -0.00000550, -0.00001007, -0.00000967, -0.00000620, -0.00000278, -0.00000129, -0.00000122, -0.00000045, 0.00000258, 0.00000726, 0.00001102, 0.00001117, 0.00000700, 0.00000037, -0.00000571, -0.00000921, -0.00001039, -0.00001103, -0.00001232, -0.00001333, -0.00001141, -0.00000455, 0.00000621, 0.00001636, 0.00002036, 0.00001529, 0.00000335, -0.00000904, -0.00001503, -0.00001166, -0.00000169, 0.00000823, 0.00001203, 0.00000798, -0.00000052, -0.00000763, -0.00000908, -0.00000475, 0.00000183, 0.00000639, 0.00000683, 0.00000415, 0.00000090, -0.00000112, -0.00000221, -0.00000385, -0.00000677, -0.00000953, -0.00000926, -0.00000396, 0.00000541, 0.00001470, 0.00001883, 0.00001496, 0.00000450, -0.00000759, -0.00001568, -0.00001667, -0.00001144, -0.00000389, 0.00000166, 0.00000292, 0.00000033, -0.00000380, -0.00000703, -0.00000798, -0.00000652, -0.00000320, 0.00000121, 0.00000585, 0.00000962, 0.00001130, 0.00000998, 0.00000572, -0.00000027, -0.00000616, -0.00001046, -0.00001266, -0.00001314, -0.00001245, -0.00001061, -0.00000713, -0.00000177, 0.00000468, 0.00001025, 0.00001273, 0.00001104, 0.00000608, 0.00000039, },
		{ 0.00001341, 0.00000836, 0.00000146, -0.00000389, -0.00000544, -0.00000324, 0.00000076, 0.00000413, 0.00000547, 0.00000494, 0.00000371, 0.00000285, 0.00000265, 0.00000271, 0.00000253, 0.00000201, 0.00000143, 0.00000101, 0.00000056, -0.00000036, -0.00000186, -0.00000342, -0.00000408, -0.00000307, -0.00000052, 0.00000264, 0.00000522, 0.00000659, 0.00000685, 0.00000632, 0.00000487, 0.00000184, -0.00000329, -0.00000972, -0.00001500, -0.00001601, -0.00001094, -0.00000096, 0.00000989, 0.00001661, 0.00001615, 0.00000917, -0.00000050, -0.00000822, -0.00001136, -0.00001033, -0.00000761, -0.00000536, -0.00000385, -0.00000169, 0.00000242, 0.00000795, 0.00001242, 0.00001311, 0.00000925, 0.00000281, -0.00000282, -0.00000533, -0.00000525, -0.00000537, -0.00000820, -0.00001345, -0.00001757, -0.00001617, -0.00000744, 0.00000574, 0.00001706, 0.00002056, 0.00001460, 0.00000305, -0.00000719, -0.00001059, -0.00000642, 0.00000125, 0.00000668, 0.00000645, 0.00000154, -0.00000393, -0.00000599, -0.00000366, 0.00000077, 0.00000384, 0.00000372, 0.00000139, -0.00000057, -0.00000041, 0.00000119, 0.00000167, -0.00000103, -0.00000636, -0.00001100, -0.00001101, -0.00000480, 0.00000534, 0.00001435, 0.00001739, 0.00001281, 0.00000308, -0.00000685, -0.00001256, -0.00001249, -0.00000828, -0.00000327, -0.00000034, -0.00000050, -0.00000292, -0.00000590, -0.00000792, -0.00000809, -0.00000611, -0.00000217, 0.00000304, 0.00000823, 0.00001174, 0.00001218, 0.00000920, 0.00000375, -0.00000241, -0.00000765, -0.00001126, -0.00001340, -0.00001436, -0.00001389, -0.00001114, -0.00000556, 0.00000210, 0.00000958, 0.00001396, },
		{ -0.00000290, 0.00000683, 0.00001422, 0.00001618, 0.00001210, 0.00000432, -0.00000317, -0.00000696, -0.00000602, -0.00000184, 0.00000283, 0.00000578, 0.00000640, 0.00000550, 0.00000433, 0.00000358, 0.00000320, 0.00000279, 0.00000219, 0.00000156, 0.00000106, 0.00000055, -0.00000036, -0.00000183, -0.00000341, -0.00000415, -0.00000330, -0.00000093, 0.00000199, 0.00000422, 0.00000515, 0.00000509, 0.00000476, 0.00000432, 0.00000294, -0.00000056, -0.00000632, -0.00001255, -0.00001594, -0.00001362, -0.00000529, 0.00000602, 0.00001528, 0.00001831, 0.00001416, 0.00000548, -0.00000334, -0.00000898, -0.00001083, -0.00001049, -0.00000977, -0.00000889, -0.00000640, -0.00000095, 0.00000679, 0.00001380, 0.00001641, 0.00001303, 0.00000561, -0.00000148, -0.00000447, -0.00000299, -0.00000035, -0.00000103, -0.00000696, -0.00001562, -0.00002135, -0.00001924, -0.00000877, 0.00000539, 0.00001618, 0.00001863, 0.00001285, 0.00000362, -0.00000313, -0.00000427, -0.00000108, 0.00000236, 0.00000266, -0.00000043, -0.00000409, -0.00000505, -0.00000230, 0.00000208, 0.00000467, 0.00000349, -0.00000037, -0.00000364, -0.00000352, -0.00000003, 0.00000386, 0.00000446, 0.00000029, -0.00000656, -0.00001161, -0.00001104, -0.00000434, 0.00000523, 0.00001264, 0.00001422, 0.00000967, 0.00000189, -0.00000511, -0.00000857, -0.00000820, -0.00000573, -0.00000343, -0.00000273, -0.00000379, -0.00000582, -0.00000774, -0.00000860, -0.00000769, -0.00000471, 0.00000006, 0.00000553, 0.00001002, 0.00001197, 0.00001071, 0.00000684, 0.00000180, -0.00000304, -0.00000716, -0.00001077, -0.00001403, -0.00001621, -0.00001573, -0.00001122, },
		{ -0.00001532, -0.00001857, -0.00001716, -0.00001003, 0.00000099, 0.00001167, 0.00001748, 0.00001616, 0.00000895, -0.00000014, -0.00000669, -0.00000825, -0.00000524, -0.00000015, 0.00000425, 0.00000653, 0.00000681, 0.00000610, 0.00000525, 0.00000450, 0.00000367, 0.00000269, 0.00000177, 0.00000112, 0.00000059, -0.00000026, -0.00000171, -0.00000344, -0.00000446, -0.00000386, -0.00000152, 0.00000156, 0.00000395, 0.00000476, 0.00000426, 0.00000345, 0.00000302, 0.00000245, 0.00000037, -0.00000412, -0.00001008, -0.00001462, -0.00001442, -0.00000809, 0.00000244, 0.00001267, 0.00001799, 0.00001644, 0.00000969, 0.00000158, -0.00000449, -0.00000763, -0.00000926, -0.00001107, -0.00001296, -0.00001274, -0.00000797, 0.00000143, 0.00001212, 0.00001893, 0.00001826, 0.00001078, 0.00000124, -0.00000459, -0.00000400, 0.00000082, 0.00000429, 0.00000167, -0.00000734, -0.00001802, -0.00002369, -0.00002026, -0.00000900, 0.00000435, 0.00001340, 0.00001509, 0.00001105, 0.00000572, 0.00000275, 0.00000263, 0.00000298, 0.00000122, -0.00000292, -0.00000696, -0.00000767, -0.00000389, 0.00000222, 0.00000658, 0.00000614, 0.00000123, -0.00000457, -0.00000705, -0.00000441, 0.00000134, 0.00000588, 0.00000562, 0.00000034, -0.00000669, -0.00001091, -0.00000952, -0.00000321, 0.00000450, 0.00000959, 0.00000997, 0.00000635, 0.00000127, -0.00000271, -0.00000445, -0.00000441, -0.00000388, -0.00000398, -0.00000506, -0.00000680, -0.00000850, -0.00000942, -0.00000883, -0.00000632, -0.00000200, 0.00000321, 0.00000775, 0.00001017, 0.00000985, 0.00000742, 0.00000417, 0.00000111, -0.00000176, -0.00000531, -0.00001011, },
		{ 0.00000231, 0.00000092, -0.00000331, -0.00001030, -0.00001741, -0.00002067, -0.00001714, -0.00000712, 0.00000556, 0.00001542, 0.00001833, 0.00001373, 0.00000474, -0.00000386, -0.00000833, -0.00000771, -0.00000366, 0.00000115, 0.00000470, 0.00000647, 0.00000698, 0.00000688, 0.00000640, 0.00000543, 0.00000401, 0.00000251, 0.00000137, 0.00000062, -0.00000018, -0.00000153, -0.00000334, -0.00000472, -0.00000458, -0.00000249, 0.00000076, 0.00000362, 0.00000485, 0.00000437, 0.00000317, 0.00000230, 0.00000172, 0.00000023, -0.00000341, -0.00000890, -0.00001380, -0.00001474, -0.00000969, 0.00000021, 0.00001080, 0.00001728, 0.00001711, 0.00001147, 0.00000419, -0.00000109, -0.00000345, -0.00000481, -0.00000775, -0.00001267, -0.00001680, -0.00001590, -0.00000779, 0.00000529, 0.00001731, 0.00002212, 0.00001755, 0.00000704, -0.00000256, -0.00000561, -0.00000150, 0.00000508, 0.00000757, 0.00000232, -0.00000880, -0.00001966, -0.00002399, -0.00001949, -0.00000894, 0.00000207, 0.00000902, 0.00001096, 0.00001014, 0.00000935, 0.00000938, 0.00000860, 0.00000491, -0.00000176, -0.00000861, -0.00001166, -0.00000869, -0.00000125, 0.00000610, 0.00000876, 0.00000521, -0.00000190, -0.00000767, -0.00000829, -0.00000365, 0.00000281, 0.00000659, 0.00000522, -0.00000025, -0.00000618, -0.00000890, -0.00000706, -0.00000212, 0.00000295, 0.00000568, 0.00000547, 0.00000341, 0.00000116, -0.00000024, -0.00000083, -0.00000136, -0.00000255, -0.00000461, -0.00000720, -0.00000962, -0.00001104, -0.00001069, -0.00000812, -0.00000358, 0.00000182, 0.00000637, 0.00000857, 0.00000803, 0.00000580, 0.00000360, 0.00000259, },
		{ 0.00000380, 0.00000104, 0.00000084, 0.00000304, 0.00000511, 0.00000378, -0.00000245, -0.00001179, -0.00001966, -0.00002133, -0.00001490, -0.00000273, 0.00000971, 0.00001691, 0.00001631, 0.00000944, 0.00000059, -0.00000589, -0.00000784, -0.00000582, -0.00000199, 0.00000163, 0.00000421, 0.00000593, 0.00000714, 0.00000781, 0.00000757, 0.00000625, 0.00000428, 0.00000237, 0.00000098, -0.00000011, -0.00000144, -0.00000321, -0.00000481, -0.00000515, -0.00000352, -0.00000039, 0.00000281, 0.00000461, 0.00000459, 0.00000354, 0.00000256, 0.00000191, 0.00000053, -0.00000297, -0.00000865, -0.00001429, -0.00001626, -0.00001189, -0.00000177, 0.00000999, 0.00001791, 0.00001855, 0.00001270, 0.00000473, -0.00000064, -0.00000165, -0.00000060, -0.00000169, -0.00000726, -0.00001530, -0.00002033, -0.00001730, -0.00000559, 0.00000980, 0.00002106, 0.00002255, 0.00001454, 0.00000297, -0.00000463, -0.00000421, 0.00000230, 0.00000864, 0.00000880, 0.00000120, -0.00001043, -0.00001980, -0.00002232, -0.00001768, -0.00000933, -0.00000145, 0.00000389, 0.00000735, 0.00001057, 0.00001386, 0.00001532, 0.00001243, 0.00000458, -0.00000550, -0.00001291, -0.00001366, -0.00000751, 0.00000162, 0.00000812, 0.00000829, 0.00000262, -0.00000473, -0.00000890, -0.00000753, -0.00000206, 0.00000361, 0.00000589, 0.00000379, -0.00000081, -0.00000481, -0.00000604, -0.00000442, -0.00000153, 0.00000077, 0.00000161, 0.00000138, 0.00000104, 0.00000124, 0.00000181, 0.00000198, 0.00000097, -0.00000155, -0.00000526, -0.00000926, -0.00001232, -0.00001316, -0.00001097, -0.00000594, 0.00000046, 0.00000589, 0.00000828, 0.00000710, },
		{ 0.00000508, 0.00000887, 0.00000782, 0.00000322, -0.00000145, -0.00000274, 0.00000038, 0.00000555, 0.00000837, 0.00000534, -0.00000352, -0.00001418, -0.00002087, -0.00001953, -0.00001039, 0.00000215, 0.00001211, 0.00001537, 0.00001172, 0.00000439, -0.00000236, -0.00000575, -0.00000555, -0.00000333, -0.00000083, 0.00000123, 0.00000312, 0.00000523, 0.00000737, 0.00000872, 0.00000857, 0.00000695, 0.00000460, 0.00000236, 0.00000048, -0.00000129, -0.00000324, -0.00000498, -0.00000561, -0.00000436, -0.00000146, 0.00000178, 0.00000388, 0.00000424, 0.00000354, 0.00000296, 0.00000286, 0.00000207, -0.00000122, -0.00000759, -0.00001496, -0.00001912, -0.00001619, -0.00000571, 0.00000835, 0.00001935, 0.00002208, 0.00001604, 0.00000585, -0.00000197, -0.00000348, -0.00000003, 0.00000300, 0.00000048, -0.00000822, -0.00001824, -0.00002240, -0.00001630, -0.00000182, 0.00001374, 0.00002227, 0.00002009, 0.00001019, -0.00000013, -0.00000431, -0.00000080, 0.00000630, 0.00001060, 0.00000795, -0.00000091, -0.00001132, -0.00001821, -0.00001926, -0.00001573, -0.00001054, -0.00000564, -0.00000084, 0.00000510, 0.00001218, 0.00001809, 0.00001925, 0.00001359, 0.00000264, -0.00000866, -0.00001475, -0.00001294, -0.00000512, 0.00000349, 0.00000773, 0.00000566, -0.00000052, -0.00000636, -0.00000822, -0.00000548, -0.00000049, 0.00000340, 0.00000416, 0.00000208, -0.00000090, -0.00000283, -0.00000304, -0.00000222, -0.00000153, -0.00000156, -0.00000199, -0.00000194, -0.00000083, 0.00000119, 0.00000319, 0.00000399, 0.00000269, -0.00000092, -0.00000615, -0.00001143, -0.00001471, -0.00001425, -0.00000957, -0.00000213, },
		{ -0.00001619, -0.00001348, -0.00000610, 0.00000280, 0.00000896, 0.00000942, 0.00000450, -0.00000223, -0.00000599, -0.00000399, 0.00000260, 0.00000914, 0.00001048, 0.00000448, -0.00000638, -0.00001632, -0.00001978, -0.00001487, -0.00000437, 0.00000613, 0.00001170, 0.00001080, 0.00000557, -0.00000014, -0.00000338, -0.00000362, -0.00000226, -0.00000098, -0.00000038, 0.00000024, 0.00000184, 0.00000457, 0.00000752, 0.00000935, 0.00000931, 0.00000763, 0.00000511, 0.00000239, -0.00000035, -0.00000308, -0.00000537, -0.00000633, -0.00000524, -0.00000231, 0.00000108, 0.00000327, 0.00000354, 0.00000272, 0.00000236, 0.00000321, 0.00000400, 0.00000212, -0.00000418, -0.00001355, -0.00002133, -0.00002195, -0.00001280, 0.00000327, 0.00001888, 0.00002631, 0.00002234, 0.00001040, -0.00000168, -0.00000691, -0.00000379, 0.00000295, 0.00000597, 0.00000089, -0.00001029, -0.00002040, -0.00002206, -0.00001295, 0.00000246, 0.00001583, 0.00002043, 0.00001539, 0.00000577, -0.00000132, -0.00000165, 0.00000363, 0.00000941, 0.00001060, 0.00000562, -0.00000301, -0.00001095, -0.00001523, -0.00001575, -0.00001434, -0.00001241, -0.00000955, -0.00000414, 0.00000445, 0.00001418, 0.00002079, 0.00002039, 0.00001235, 0.00000021, -0.00001010, -0.00001384, -0.00001024, -0.00000270, 0.00000362, 0.00000517, 0.00000190, -0.00000318, -0.00000648, -0.00000618, -0.00000306, 0.00000048, 0.00000237, 0.00000211, 0.00000070, -0.00000046, -0.00000074, -0.00000054, -0.00000076, -0.00000191, -0.00000358, -0.00000476, -0.00000447, -0.00000239, 0.00000087, 0.00000396, 0.00000532, 0.00000375, -0.00000096, -0.00000759, -0.00001363, },
		{ 0.00000383, -0.00000211, -0.00000970, -0.00001537, -0.00001585, -0.00001021, -0.00000081, 0.00000761, 0.00001057, 0.00000672, -0.00000113, -0.00000755, -0.00000802, -0.00000194, 0.00000678, 0.00001214, 0.00001002, 0.00000099, -0.00001000, -0.00001673, -0.00001562, -0.00000786, 0.00000170, 0.00000784, 0.00000815, 0.00000407, -0.00000074, -0.00000312, -0.00000240, -0.00000023, 0.00000114, 0.00000077, -0.00000049, -0.00000088, 0.00000070, 0.00000392, 0.00000735, 0.00000955, 0.00000985, 0.00000844, 0.00000578, 0.00000225, -0.00000176, -0.00000542, -0.00000741, -0.00000672, -0.00000349, 0.00000061, 0.00000333, 0.00000345, 0.00000179, 0.00000064, 0.00000176, 0.00000443, 0.00000542, 0.00000125, -0.00000862, -0.00002012, -0.00002629, -0.00002158, -0.00000609, 0.00001358, 0.00002771, 0.00002938, 0.00001879, 0.00000318, -0.00000795, -0.00000907, -0.00000195, 0.00000598, 0.00000718, -0.00000053, -0.00001252, -0.00002063, -0.00001899, -0.00000805, 0.00000593, 0.00001522, 0.00001583, 0.00000965, 0.00000251, -0.00000025, 0.00000261, 0.00000791, 0.00001094, 0.00000893, 0.00000277, -0.00000428, -0.00000929, -0.00001164, -0.00001269, -0.00001375, -0.00001431, -0.00001219, -0.00000550, 0.00000501, 0.00001553, 0.00002109, 0.00001873, 0.00000960, -0.00000157, -0.00000932, -0.00001070, -0.00000673, -0.00000122, 0.00000197, 0.00000131, -0.00000186, -0.00000478, -0.00000542, -0.00000367, -0.00000105, 0.00000075, 0.00000104, 0.00000040, -0.00000007, 0.00000022, 0.00000092, 0.00000109, -0.00000002, -0.00000235, -0.00000501, -0.00000668, -0.00000633, -0.00000370, 0.00000044, 0.00000436, 0.00000600, },
		{ -0.00000455, 0.00000023, 0.00000454, 0.00000577, 0.00000242, -0.00000464, -0.00001210, -0.00001573, -0.00001288, -0.00000444, 0.00000522, 0.00001065, 0.00000875, 0.00000095, -0.00000748, -0.00001076, -0.00000640, 0.00000297, 0.00001122, 0.00001280, 0.00000637, -0.00000427, -0.00001273, -0.00001423, -0.00000856, 0.00000009, 0.00000617, 0.00000651, 0.00000203, -0.00000338, -0.00000585, -0.00000412, 0.00000002, 0.00000340, 0.00000395, 0.00000191, -0.00000070, -0.00000169, -0.00000021, 0.00000307, 0.00000667, 0.00000933, 0.00001037, 0.00000946, 0.00000646, 0.00000165, -0.00000385, -0.00000798, -0.00000879, -0.00000574, -0.00000053, 0.00000372, 0.00000456, 0.00000211, -0.00000090, -0.00000119, 0.00000215, 0.00000629, 0.00000627, -0.00000114, -0.00001418, -0.00002597, -0.00002837, -0.00001752, 0.00000291, 0.00002316, 0.00003300, 0.00002809, 0.00001258, -0.00000366, -0.00001153, -0.00000836, 0.00000090, 0.00000772, 0.00000616, -0.00000311, -0.00001370, -0.00001825, -0.00001364, -0.00000287, 0.00000742, 0.00001179, 0.00000945, 0.00000421, 0.00000112, 0.00000267, 0.00000731, 0.00001107, 0.00001080, 0.00000639, 0.00000039, -0.00000435, -0.00000685, -0.00000836, -0.00001061, -0.00001369, -0.00001542, -0.00001289, -0.00000501, 0.00000596, 0.00001534, 0.00001878, 0.00001498, 0.00000649, -0.00000199, -0.00000660, -0.00000642, -0.00000352, -0.00000101, -0.00000083, -0.00000270, -0.00000478, -0.00000530, -0.00000388, -0.00000155, 0.00000020, 0.00000060, 0.00000000, -0.00000056, -0.00000029, 0.00000077, 0.00000181, 0.00000181, 0.00000022, -0.00000268, -0.00000587, -0.00000791, -0.00000759, },
		{ -0.00000296, -0.00000643, -0.00000863, -0.00000811, -0.00000458, 0.00000051, 0.00000431, 0.00000410, -0.00000087, -0.00000833, -0.00001392, -0.00001372, -0.00000697, 0.00000295, 0.00001026, 0.00001044, 0.00000332, -0.00000648, -0.00001240, -0.00001025, -0.00000114, 0.00000913, 0.00001389, 0.00001004, 0.00000009, -0.00000955, -0.00001294, -0.00000850, 0.00000021, 0.00000710, 0.00000778, 0.00000229, -0.00000530, -0.00000993, -0.00000892, -0.00000343, 0.00000285, 0.00000638, 0.00000591, 0.00000277, -0.00000052, -0.00000198, -0.00000103, 0.00000176, 0.00000542, 0.00000889, 0.00001104, 0.00001061, 0.00000683, 0.00000033, -0.00000644, -0.00001024, -0.00000895, -0.00000334, 0.00000306, 0.00000618, 0.00000429, -0.00000049, -0.00000373, -0.00000211, 0.00000347, 0.00000799, 0.00000571, -0.00000511, -0.00001984, -0.00002956, -0.00002665, -0.00001024, 0.00001243, 0.00002996, 0.00003372, 0.00002309, 0.00000555, -0.00000837, -0.00001184, -0.00000564, 0.00000332, 0.00000731, 0.00000326, -0.00000566, -0.00001284, -0.00001332, -0.00000714, 0.00000120, 0.00000638, 0.00000618, 0.00000268, 0.00000018, 0.00000164, 0.00000649, 0.00001131, 0.00001261, 0.00000949, 0.00000397, -0.00000087, -0.00000336, -0.00000437, -0.00000604, -0.00000953, -0.00001356, -0.00001507, -0.00001159, -0.00000334, 0.00000640, 0.00001324, 0.00001440, 0.00001028, 0.00000396, -0.00000100, -0.00000283, -0.00000221, -0.00000129, -0.00000177, -0.00000376, -0.00000588, -0.00000656, -0.00000523, -0.00000266, -0.00000031, 0.00000076, 0.00000048, -0.00000036, -0.00000076, -0.00000025, 0.00000091, 0.00000191, 0.00000183, 0.00000014, },
		{ 0.00000067, 0.00000149, 0.00000139, -0.00000031, -0.00000348, -0.00000693, -0.00000875, -0.00000751, -0.00000344, 0.00000119, 0.00000317, 0.00000047, -0.00000599, -0.00001234, -0.00001396, -0.00000869, 0.00000127, 0.00001022, 0.00001247, 0.00000627, -0.00000468, -0.00001333, -0.00001384, -0.00000555, 0.00000635, 0.00001425, 0.00001313, 0.00000393, -0.00000712, -0.00001281, -0.00000971, -0.00000035, 0.00000873, 0.00001144, 0.00000612, -0.00000374, -0.00001209, -0.00001433, -0.00000989, -0.00000195, 0.00000505, 0.00000818, 0.00000721, 0.00000388, 0.00000036, -0.00000186, -0.00000208, -0.00000011, 0.00000379, 0.00000852, 0.00001193, 0.00001161, 0.00000651, -0.00000173, -0.00000913, -0.00001155, -0.00000760, 0.00000013, 0.00000648, 0.00000730, 0.00000252, -0.00000364, -0.00000585, -0.00000187, 0.00000515, 0.00000860, 0.00000330, -0.00001002, -0.00002416, -0.00002964, -0.00002105, -0.00000115, 0.00002034, 0.00003252, 0.00003000, 0.00001596, -0.00000036, -0.00000992, -0.00000936, -0.00000230, 0.00000413, 0.00000471, -0.00000049, -0.00000693, -0.00000955, -0.00000668, -0.00000094, 0.00000318, 0.00000299, -0.00000041, -0.00000317, -0.00000195, 0.00000349, 0.00001008, 0.00001381, 0.00001265, 0.00000780, 0.00000245, -0.00000082, -0.00000184, -0.00000253, -0.00000492, -0.00000906, -0.00001274, -0.00001308, -0.00000881, -0.00000140, 0.00000575, 0.00000953, 0.00000901, 0.00000575, 0.00000240, 0.00000078, 0.00000081, 0.00000101, -0.00000009, -0.00000276, -0.00000584, -0.00000771, -0.00000740, -0.00000518, -0.00000227, -0.00000001, 0.00000094, 0.00000071, -0.00000000, -0.00000044, -0.00000018, },
		{ 0.00000132, 0.00000079, 0.00000010, -0.00000012, 0.00000030, 0.00000085, 0.00000054, -0.00000137, -0.00000454, -0.00000739, -0.00000795, -0.00000542, -0.00000115, 0.00000174, 0.00000044, -0.00000528, -0.00001219, -0.00001529, -0.00001119, -0.00000086, 0.00001021, 0.00001522, 0.00001060, -0.00000126, -0.00001303, -0.00001713, -0.00001068, 0.00000235, 0.00001367, 0.00001611, 0.00000839, -0.00000410, -0.00001298, -0.00001251, -0.00000328, 0.00000842, 0.00001492, 0.00001213, 0.00000186, -0.00000985, -0.00001678, -0.00001611, -0.00000941, -0.00000080, 0.00000589, 0.00000886, 0.00000838, 0.00000564, 0.00000189, -0.00000168, -0.00000360, -0.00000244, 0.00000215, 0.00000845, 0.00001288, 0.00001206, 0.00000529, -0.00000430, -0.00001128, -0.00001144, -0.00000484, 0.00000394, 0.00000887, 0.00000678, -0.00000027, -0.00000647, -0.00000671, -0.00000072, 0.00000630, 0.00000740, -0.00000075, -0.00001466, -0.00002570, -0.00002568, -0.00001255, 0.00000776, 0.00002487, 0.00003041, 0.00002304, 0.00000872, -0.00000370, -0.00000831, -0.00000532, 0.00000021, 0.00000283, 0.00000067, -0.00000372, -0.00000611, -0.00000424, 0.00000032, 0.00000365, 0.00000276, -0.00000188, -0.00000658, -0.00000721, -0.00000235, 0.00000568, 0.00001246, 0.00001458, 0.00001175, 0.00000657, 0.00000224, 0.00000024, -0.00000038, -0.00000167, -0.00000474, -0.00000861, -0.00001093, -0.00000982, -0.00000545, 0.00000005, 0.00000402, 0.00000507, 0.00000381, 0.00000211, 0.00000159, 0.00000239, 0.00000334, 0.00000292, 0.00000052, -0.00000314, -0.00000652, -0.00000823, -0.00000778, -0.00000563, -0.00000282, -0.00000038, 0.00000105, },
		{ -0.00000666, -0.00000399, -0.00000106, 0.00000116, 0.00000205, 0.00000166, 0.00000074, 0.00000015, 0.00000015, 0.00000013, -0.00000092, -0.00000336, -0.00000619, -0.00000746, -0.00000576, -0.00000174, 0.00000174, 0.00000127, -0.00000436, -0.00001258, -0.00001798, -0.00001574, -0.00000528, 0.00000845, 0.00001753, 0.00001612, 0.00000459, -0.00001020, -0.00001897, -0.00001607, -0.00000330, 0.00001117, 0.00001814, 0.00001344, 0.00000059, -0.00001170, -0.00001534, -0.00000819, 0.00000490, 0.00001551, 0.00001721, 0.00000929, -0.00000345, -0.00001424, -0.00001846, -0.00001566, -0.00000862, -0.00000088, 0.00000524, 0.00000892, 0.00000995, 0.00000814, 0.00000371, -0.00000185, -0.00000563, -0.00000487, 0.00000090, 0.00000870, 0.00001353, 0.00001156, 0.00000316, -0.00000687, -0.00001229, -0.00000973, -0.00000120, 0.00000722, 0.00000967, 0.00000482, -0.00000326, -0.00000818, -0.00000625, 0.00000060, 0.00000605, 0.00000423, -0.00000553, -0.00001759, -0.00002357, -0.00001814, -0.00000289, 0.00001447, 0.00002505, 0.00002437, 0.00001477, 0.00000316, -0.00000394, -0.00000448, -0.00000131, 0.00000093, -0.00000025, -0.00000352, -0.00000529, -0.00000311, 0.00000202, 0.00000625, 0.00000587, 0.00000035, -0.00000699, -0.00001121, -0.00000919, -0.00000170, 0.00000725, 0.00001316, 0.00001392, 0.00001064, 0.00000628, 0.00000321, 0.00000177, 0.00000060, -0.00000165, -0.00000494, -0.00000774, -0.00000829, -0.00000610, -0.00000242, 0.00000069, 0.00000177, 0.00000096, -0.00000032, -0.00000052, 0.00000084, 0.00000293, 0.00000427, 0.00000373, 0.00000122, -0.00000237, -0.00000576, -0.00000786, -0.00000814, },
		{ -0.00000045, -0.00000402, -0.00000711, -0.00000864, -0.00000798, -0.00000534, -0.00000184, 0.00000108, 0.00000246, 0.00000236, 0.00000161, 0.00000097, 0.00000042, -0.00000083, -0.00000331, -0.00000637, -0.00000806, -0.00000655, -0.00000191, 0.00000304, 0.00000417, -0.00000107, -0.00001104, -0.00002006, -0.00002157, -0.00001272, 0.00000296, 0.00001696, 0.00002089, 0.00001212, -0.00000411, -0.00001769, -0.00002004, -0.00000977, 0.00000635, 0.00001793, 0.00001775, 0.00000646, -0.00000810, -0.00001628, -0.00001302, -0.00000077, 0.00001254, 0.00001891, 0.00001517, 0.00000417, -0.00000788, -0.00001562, -0.00001719, -0.00001404, -0.00000863, -0.00000258, 0.00000354, 0.00000905, 0.00001222, 0.00001106, 0.00000524, -0.00000264, -0.00000792, -0.00000690, 0.00000028, 0.00000909, 0.00001344, 0.00000993, 0.00000048, -0.00000878, -0.00001174, -0.00000667, 0.00000252, 0.00000918, 0.00000875, 0.00000205, -0.00000557, -0.00000843, -0.00000493, 0.00000121, 0.00000392, -0.00000039, -0.00000971, -0.00001761, -0.00001773, -0.00000844, 0.00000584, 0.00001756, 0.00002106, 0.00001607, 0.00000721, 0.00000036, -0.00000155, 0.00000005, 0.00000135, -0.00000030, -0.00000409, -0.00000654, -0.00000463, 0.00000142, 0.00000787, 0.00001002, 0.00000568, -0.00000307, -0.00001113, -0.00001375, -0.00000949, -0.00000088, 0.00000755, 0.00001223, 0.00001243, 0.00000990, 0.00000701, 0.00000495, 0.00000330, 0.00000107, -0.00000202, -0.00000500, -0.00000638, -0.00000543, -0.00000282, -0.00000026, 0.00000074, -0.00000013, -0.00000189, -0.00000303, -0.00000257, -0.00000058, 0.00000193, 0.00000376, 0.00000403, 0.00000250, },
		{ -0.00000047, 0.00000238, 0.00000435, 0.00000442, 0.00000216, -0.00000190, -0.00000624, -0.00000902, -0.00000906, -0.00000653, -0.00000283, 0.00000041, 0.00000229, 0.00000296, 0.00000300, 0.00000252, 0.00000091, -0.00000243, -0.00000682, -0.00001002, -0.00000937, -0.00000407, 0.00000334, 0.00000780, 0.00000488, -0.00000565, -0.00001855, -0.00002574, -0.00002139, -0.00000634, 0.00001158, 0.00002202, 0.00001875, 0.00000397, -0.00001280, -0.00002079, -0.00001504, 0.00000041, 0.00001526, 0.00001995, 0.00001189, -0.00000306, -0.00001482, -0.00001595, -0.00000626, 0.00000766, 0.00001725, 0.00001754, 0.00000961, -0.00000127, -0.00000976, -0.00001365, -0.00001394, -0.00001258, -0.00001012, -0.00000559, 0.00000163, 0.00000981, 0.00001502, 0.00001376, 0.00000597, -0.00000403, -0.00001000, -0.00000809, 0.00000029, 0.00000919, 0.00001226, 0.00000730, -0.00000216, -0.00000944, -0.00000961, -0.00000293, 0.00000542, 0.00000941, 0.00000650, -0.00000065, -0.00000658, -0.00000736, -0.00000358, 0.00000038, 0.00000005, -0.00000531, -0.00001191, -0.00001420, -0.00000911, 0.00000145, 0.00001183, 0.00001658, 0.00001414, 0.00000757, 0.00000192, 0.00000039, 0.00000220, 0.00000377, 0.00000202, -0.00000285, -0.00000736, -0.00000746, -0.00000194, 0.00000633, 0.00001206, 0.00001119, 0.00000370, -0.00000637, -0.00001353, -0.00001425, -0.00000883, -0.00000062, 0.00000645, 0.00001016, 0.00001074, 0.00000974, 0.00000842, 0.00000691, 0.00000464, 0.00000134, -0.00000221, -0.00000464, -0.00000491, -0.00000315, -0.00000066, 0.00000097, 0.00000083, -0.00000087, -0.00000307, -0.00000458, -0.00000462, -0.00000310, },
		{ -0.00000281, -0.00000556, -0.00000713, -0.00000657, -0.00000363, 0.00000081, 0.00000488, 0.00000654, 0.00000474, 0.00000013, -0.00000517, -0.00000880, -0.00000947, -0.00000753, -0.00000432, -0.00000113, 0.00000156, 0.00000377, 0.00000514, 0.00000461, 0.00000109, -0.00000506, -0.00001120, -0.00001341, -0.00000918, 0.00000011, 0.00000902, 0.00001098, 0.00000290, -0.00001195, -0.00002501, -0.00002760, -0.00001679, 0.00000197, 0.00001785, 0.00002144, 0.00001105, -0.00000602, -0.00001824, -0.00001751, -0.00000459, 0.00001161, 0.00002022, 0.00001578, 0.00000175, -0.00001222, -0.00001706, -0.00001021, 0.00000319, 0.00001445, 0.00001734, 0.00001170, 0.00000241, -0.00000500, -0.00000826, -0.00000903, -0.00001016, -0.00001233, -0.00001308, -0.00000906, 0.00000034, 0.00001135, 0.00001782, 0.00001554, 0.00000566, -0.00000566, -0.00001133, -0.00000824, 0.00000065, 0.00000857, 0.00000991, 0.00000413, -0.00000408, -0.00000858, -0.00000632, 0.00000061, 0.00000686, 0.00000799, 0.00000368, -0.00000248, -0.00000613, -0.00000562, -0.00000305, -0.00000212, -0.00000475, -0.00000916, -0.00001121, -0.00000776, 0.00000058, 0.00000951, 0.00001408, 0.00001221, 0.00000617, 0.00000073, -0.00000046, 0.00000242, 0.00000578, 0.00000561, 0.00000085, -0.00000561, -0.00000895, -0.00000607, 0.00000195, 0.00001036, 0.00001386, 0.00001006, 0.00000094, -0.00000863, -0.00001399, -0.00001328, -0.00000796, -0.00000122, 0.00000428, 0.00000759, 0.00000923, 0.00001000, 0.00001002, 0.00000877, 0.00000590, 0.00000190, -0.00000187, -0.00000402, -0.00000392, -0.00000208, 0.00000024, 0.00000173, 0.00000163, -0.00000008, },
		{ 0.00000022, 0.00000262, 0.00000351, 0.00000201, -0.00000177, -0.00000652, -0.00001005, -0.00001032, -0.00000668, -0.00000051, 0.00000535, 0.00000812, 0.00000666, 0.00000198, -0.00000350, -0.00000753, -0.00000913, -0.00000861, -0.00000662, -0.00000350, 0.00000068, 0.00000515, 0.00000793, 0.00000660, 0.00000029, -0.00000877, -0.00001560, -0.00001533, -0.00000686, 0.00000540, 0.00001354, 0.00001114, -0.00000190, -0.00001844, -0.00002816, -0.00002430, -0.00000851, 0.00000970, 0.00001914, 0.00001437, -0.00000049, -0.00001456, -0.00001772, -0.00000758, 0.00000888, 0.00002031, 0.00001892, 0.00000583, -0.00000996, -0.00001798, -0.00001347, -0.00000023, 0.00001263, 0.00001730, 0.00001241, 0.00000299, -0.00000413, -0.00000562, -0.00000366, -0.00000325, -0.00000730, -0.00001371, -0.00001679, -0.00001193, 0.00000020, 0.00001336, 0.00001984, 0.00001590, 0.00000444, -0.00000700, -0.00001153, -0.00000749, 0.00000089, 0.00000696, 0.00000665, 0.00000111, -0.00000477, -0.00000633, -0.00000265, 0.00000313, 0.00000658, 0.00000547, 0.00000116, -0.00000299, -0.00000460, -0.00000408, -0.00000389, -0.00000588, -0.00000922, -0.00001073, -0.00000748, 0.00000043, 0.00000934, 0.00001430, 0.00001264, 0.00000596, -0.00000089, -0.00000332, -0.00000028, 0.00000510, 0.00000792, 0.00000533, -0.00000131, -0.00000738, -0.00000830, -0.00000286, 0.00000587, 0.00001259, 0.00001324, 0.00000742, -0.00000164, -0.00000944, -0.00001287, -0.00001157, -0.00000738, -0.00000255, 0.00000161, 0.00000499, 0.00000792, 0.00001031, 0.00001142, 0.00001048, 0.00000742, 0.00000310, -0.00000101, -0.00000358, -0.00000398, -0.00000241, },
		{ -0.00000001, -0.00000383, -0.00000525, -0.00000378, -0.00000005, 0.00000409, 0.00000624, 0.00000464, -0.00000071, -0.00000764, -0.00001278, -0.00001339, -0.00000898, -0.00000166, 0.00000510, 0.00000848, 0.00000770, 0.00000394, -0.00000081, -0.00000511, -0.00000835, -0.00001015, -0.00000974, -0.00000620, 0.00000024, 0.00000722, 0.00001073, 0.00000756, -0.00000194, -0.00001300, -0.00001872, -0.00001467, -0.00000245, 0.00001044, 0.00001514, 0.00000749, -0.00000853, -0.00002292, -0.00002614, -0.00001583, 0.00000132, 0.00001391, 0.00001377, 0.00000176, -0.00001271, -0.00001831, -0.00001002, 0.00000720, 0.00002170, 0.00002337, 0.00001082, -0.00000765, -0.00001983, -0.00001810, -0.00000443, 0.00001145, 0.00001921, 0.00001502, 0.00000357, -0.00000618, -0.00000800, -0.00000279, 0.00000277, 0.00000197, -0.00000628, -0.00001635, -0.00002016, -0.00001335, 0.00000119, 0.00001511, 0.00002040, 0.00001475, 0.00000279, -0.00000754, -0.00001055, -0.00000624, 0.00000055, 0.00000437, 0.00000309, -0.00000110, -0.00000406, -0.00000324, 0.00000051, 0.00000409, 0.00000486, 0.00000270, -0.00000036, -0.00000224, -0.00000274, -0.00000352, -0.00000611, -0.00000994, -0.00001209, -0.00000943, -0.00000146, 0.00000860, 0.00001546, 0.00001531, 0.00000854, -0.00000030, -0.00000558, -0.00000443, 0.00000142, 0.00000705, 0.00000800, 0.00000344, -0.00000334, -0.00000738, -0.00000560, 0.00000116, 0.00000877, 0.00001269, 0.00001077, 0.00000426, -0.00000340, -0.00000891, -0.00001083, -0.00000975, -0.00000717, -0.00000419, -0.00000103, 0.00000259, 0.00000660, 0.00001028, 0.00001247, 0.00001225, 0.00000950, 0.00000495, },
		{ 0.00001334, 0.00001433, 0.00001213, 0.00000707, 0.00000059, -0.00000501, -0.00000740, -0.00000543, 0.00000003, 0.00000606, 0.00000910, 0.00000688, -0.00000006, -0.00000853, -0.00001439, -0.00001497, -0.00001037, -0.00000316, 0.00000345, 0.00000733, 0.00000813, 0.00000650, 0.00000305, -0.00000191, -0.00000768, -0.00001236, -0.00001321, -0.00000843, 0.00000071, 0.00000960, 0.00001254, 0.00000663, -0.00000549, -0.00001655, -0.00001912, -0.00001090, 0.00000310, 0.00001339, 0.00001249, 0.00000038, -0.00001501, -0.00002310, -0.00001818, -0.00000376, 0.00000969, 0.00001221, 0.00000183, -0.00001363, -0.00002187, -0.00001509, 0.00000396, 0.00002337, 0.00003009, 0.00001918, -0.00000257, -0.00002115, -0.00002481, -0.00001200, 0.00000795, 0.00002164, 0.00002090, 0.00000795, -0.00000692, -0.00001332, -0.00000816, 0.00000265, 0.00000919, 0.00000521, -0.00000719, -0.00001927, -0.00002209, -0.00001298, 0.00000278, 0.00001583, 0.00001915, 0.00001240, 0.00000129, -0.00000700, -0.00000867, -0.00000507, -0.00000057, 0.00000122, -0.00000005, -0.00000203, -0.00000221, -0.00000015, 0.00000243, 0.00000346, 0.00000238, 0.00000048, -0.00000062, -0.00000074, -0.00000140, -0.00000428, -0.00000911, -0.00001307, -0.00001242, -0.00000550, 0.00000542, 0.00001496, 0.00001800, 0.00001304, 0.00000341, -0.00000493, -0.00000732, -0.00000331, 0.00000330, 0.00000741, 0.00000621, 0.00000095, -0.00000425, -0.00000545, -0.00000162, 0.00000487, 0.00001003, 0.00001094, 0.00000734, 0.00000136, -0.00000418, -0.00000752, -0.00000852, -0.00000809, -0.00000709, -0.00000566, -0.00000333, 0.00000030, 0.00000500, 0.00000981, },
		{ -0.00000530, -0.00000206, 0.00000303, 0.00000905, 0.00001429, 0.00001674, 0.00001491, 0.00000881, 0.00000041, -0.00000685, -0.00000962, -0.00000655, 0.00000066, 0.00000793, 0.00001104, 0.00000802, 0.00000028, -0.00000837, -0.00001401, -0.00001476, -0.00001127, -0.00000568, 0.00000006, 0.00000498, 0.00000857, 0.00000994, 0.00000772, 0.00000129, -0.00000766, -0.00001498, -0.00001612, -0.00000936, 0.00000209, 0.00001136, 0.00001218, 0.00000342, -0.00000951, -0.00001791, -0.00001580, -0.00000446, 0.00000806, 0.00001252, 0.00000531, -0.00000857, -0.00001884, -0.00001751, -0.00000514, 0.00000902, 0.00001357, 0.00000368, -0.00001461, -0.00002791, -0.00002484, -0.00000467, 0.00002128, 0.00003669, 0.00003134, 0.00000825, -0.00001800, -0.00003096, -0.00002321, -0.00000099, 0.00002048, 0.00002755, 0.00001711, -0.00000209, -0.00001633, -0.00001668, -0.00000482, 0.00000900, 0.00001385, 0.00000588, -0.00000934, -0.00002126, -0.00002191, -0.00001110, 0.00000417, 0.00001493, 0.00001622, 0.00000949, 0.00000043, -0.00000550, -0.00000649, -0.00000443, -0.00000236, -0.00000183, -0.00000211, -0.00000163, 0.00000013, 0.00000211, 0.00000276, 0.00000165, -0.00000002, -0.00000062, 0.00000016, 0.00000072, -0.00000118, -0.00000616, -0.00001189, -0.00001422, -0.00001003, 0.00000005, 0.00001146, 0.00001828, 0.00001698, 0.00000882, -0.00000107, -0.00000703, -0.00000646, -0.00000119, 0.00000426, 0.00000601, 0.00000337, -0.00000107, -0.00000359, -0.00000211, 0.00000255, 0.00000744, 0.00000964, 0.00000807, 0.00000383, -0.00000086, -0.00000425, -0.00000587, -0.00000634, -0.00000655, -0.00000681, -0.00000667, },
		{ -0.00000438, -0.00000491, -0.00000608, -0.00000721, -0.00000710, -0.00000454, 0.00000091, 0.00000837, 0.00001546, 0.00001913, 0.00001714, 0.00000962, -0.00000048, -0.00000859, -0.00001093, -0.00000668, 0.00000142, 0.00000869, 0.00001120, 0.00000796, 0.00000100, -0.00000626, -0.00001123, -0.00001312, -0.00001244, -0.00000969, -0.00000477, 0.00000221, 0.00000962, 0.00001413, 0.00001235, 0.00000362, -0.00000842, -0.00001724, -0.00001739, -0.00000851, 0.00000378, 0.00001129, 0.00000889, -0.00000163, -0.00001254, -0.00001569, -0.00000866, 0.00000318, 0.00001043, 0.00000690, -0.00000531, -0.00001682, -0.00001786, -0.00000613, 0.00001050, 0.00001907, 0.00001102, -0.00001059, -0.00003189, -0.00003721, -0.00002017, 0.00001091, 0.00003768, 0.00004351, 0.00002471, -0.00000685, -0.00003144, -0.00003437, -0.00001534, 0.00001204, 0.00002981, 0.00002763, 0.00000895, -0.00001241, -0.00002231, -0.00001574, 0.00000068, 0.00001433, 0.00001563, 0.00000427, -0.00001155, -0.00002137, -0.00001961, -0.00000848, 0.00000455, 0.00001230, 0.00001219, 0.00000669, 0.00000039, -0.00000349, -0.00000462, -0.00000449, -0.00000434, -0.00000404, -0.00000275, -0.00000030, 0.00000214, 0.00000295, 0.00000163, -0.00000058, -0.00000162, -0.00000054, 0.00000137, 0.00000141, -0.00000223, -0.00000844, -0.00001338, -0.00001286, -0.00000552, 0.00000574, 0.00001528, 0.00001817, 0.00001337, 0.00000425, -0.00000373, -0.00000658, -0.00000405, 0.00000072, 0.00000385, 0.00000343, 0.00000055, -0.00000190, -0.00000155, 0.00000173, 0.00000596, 0.00000856, 0.00000805, 0.00000489, 0.00000080, -0.00000245, -0.00000405, -0.00000437, },
		{ -0.00000160, -0.00000367, -0.00000393, -0.00000313, -0.00000248, -0.00000300, -0.00000497, -0.00000753, -0.00000886, -0.00000695, -0.00000090, 0.00000808, 0.00001664, 0.00002079, 0.00001807, 0.00000933, -0.00000145, -0.00000921, -0.00001063, -0.00000597, 0.00000142, 0.00000735, 0.00000927, 0.00000726, 0.00000305, -0.00000169, -0.00000639, -0.00001100, -0.00001462, -0.00001506, -0.00001013, 0.00000001, 0.00001152, 0.00001831, 0.00001579, 0.00000448, -0.00000952, -0.00001802, -0.00001620, -0.00000610, 0.00000466, 0.00000830, 0.00000263, -0.00000727, -0.00001295, -0.00000926, 0.00000124, 0.00000990, 0.00000867, -0.00000307, -0.00001704, -0.00002166, -0.00001107, 0.00000928, 0.00002522, 0.00002349, 0.00000190, -0.00002758, -0.00004574, -0.00003888, -0.00000870, 0.00002780, 0.00004882, 0.00004185, 0.00001183, -0.00002215, -0.00003931, -0.00003051, -0.00000338, 0.00002356, 0.00003352, 0.00002206, -0.00000123, -0.00002027, -0.00002360, -0.00001130, 0.00000642, 0.00001710, 0.00001440, 0.00000143, -0.00001263, -0.00001926, -0.00001582, -0.00000607, 0.00000352, 0.00000834, 0.00000786, 0.00000447, 0.00000093, -0.00000162, -0.00000347, -0.00000504, -0.00000583, -0.00000490, -0.00000209, 0.00000127, 0.00000310, 0.00000225, -0.00000039, -0.00000245, -0.00000208, 0.00000035, 0.00000228, 0.00000094, -0.00000412, -0.00001017, -0.00001285, -0.00000921, -0.00000009, 0.00001012, 0.00001616, 0.00001523, 0.00000864, 0.00000071, -0.00000418, -0.00000434, -0.00000133, 0.00000165, 0.00000228, 0.00000061, -0.00000133, -0.00000136, 0.00000121, 0.00000510, 0.00000806, 0.00000840, 0.00000596, 0.00000203, },
		{ 0.00000879, 0.00000747, 0.00000394, -0.00000027, -0.00000353, -0.00000479, -0.00000397, -0.00000199, -0.00000047, -0.00000091, -0.00000380, -0.00000787, -0.00001045, -0.00000879, -0.00000193, 0.00000816, 0.00001722, 0.00002088, 0.00001730, 0.00000834, -0.00000152, -0.00000783, -0.00000867, -0.00000524, -0.00000042, 0.00000344, 0.00000574, 0.00000698, 0.00000720, 0.00000517, -0.00000055, -0.00000949, -0.00001807, -0.00002098, -0.00001473, -0.00000085, 0.00001392, 0.00002135, 0.00001707, 0.00000385, -0.00001001, -0.00001626, -0.00001235, -0.00000304, 0.00000345, 0.00000194, -0.00000563, -0.00001173, -0.00000957, 0.00000084, 0.00001183, 0.00001360, 0.00000233, -0.00001555, -0.00002687, -0.00002112, 0.00000081, 0.00002556, 0.00003546, 0.00002091, -0.00001178, -0.00004311, -0.00005259, -0.00003246, 0.00000652, 0.00004139, 0.00005162, 0.00003211, -0.00000383, -0.00003352, -0.00003961, -0.00002062, 0.00000925, 0.00003044, 0.00003083, 0.00001265, -0.00001053, -0.00002375, -0.00002038, -0.00000513, 0.00001053, 0.00001667, 0.00001095, -0.00000130, -0.00001185, -0.00001527, -0.00001155, -0.00000463, 0.00000114, 0.00000380, 0.00000392, 0.00000294, 0.00000156, -0.00000044, -0.00000313, -0.00000558, -0.00000624, -0.00000428, -0.00000064, 0.00000233, 0.00000266, 0.00000035, -0.00000251, -0.00000339, -0.00000150, 0.00000141, 0.00000231, -0.00000057, -0.00000607, -0.00001049, -0.00001018, -0.00000419, 0.00000482, 0.00001219, 0.00001428, 0.00001070, 0.00000428, -0.00000105, -0.00000287, -0.00000152, 0.00000068, 0.00000147, 0.00000031, -0.00000145, -0.00000185, 0.00000013, 0.00000382, 0.00000732, },
		{ -0.00000112, 0.00000189, 0.00000557, 0.00000819, 0.00000851, 0.00000628, 0.00000227, -0.00000210, -0.00000520, -0.00000586, -0.00000394, -0.00000070, 0.00000162, 0.00000099, -0.00000297, -0.00000829, -0.00001145, -0.00000947, -0.00000197, 0.00000818, 0.00001632, 0.00001880, 0.00001503, 0.00000760, 0.00000035, -0.00000415, -0.00000571, -0.00000563, -0.00000481, -0.00000271, 0.00000178, 0.00000823, 0.00001356, 0.00001326, 0.00000482, -0.00000941, -0.00002233, -0.00002610, -0.00001737, -0.00000022, 0.00001588, 0.00002204, 0.00001575, 0.00000242, -0.00000875, -0.00001142, -0.00000648, -0.00000068, -0.00000070, -0.00000724, -0.00001412, -0.00001333, -0.00000224, 0.00001290, 0.00002064, 0.00001293, -0.00000756, -0.00002764, -0.00003206, -0.00001476, 0.00001535, 0.00003893, 0.00003877, 0.00001208, -0.00002640, -0.00005308, -0.00005073, -0.00001991, 0.00002107, 0.00004778, 0.00004524, 0.00001700, -0.00001794, -0.00003796, -0.00003249, -0.00000770, 0.00001897, 0.00003095, 0.00002278, 0.00000220, -0.00001648, -0.00002232, -0.00001406, 0.00000071, 0.00001184, 0.00001345, 0.00000665, -0.00000280, -0.00000918, -0.00001034, -0.00000782, -0.00000446, -0.00000203, -0.00000050, 0.00000078, 0.00000180, 0.00000168, -0.00000023, -0.00000326, -0.00000552, -0.00000527, -0.00000250, 0.00000084, 0.00000232, 0.00000093, -0.00000205, -0.00000402, -0.00000326, -0.00000039, 0.00000196, 0.00000132, -0.00000257, -0.00000721, -0.00000902, -0.00000597, 0.00000091, 0.00000805, 0.00001177, 0.00001060, 0.00000607, 0.00000131, -0.00000111, -0.00000077, 0.00000079, 0.00000155, 0.00000064, -0.00000115, -0.00000216, },
		{ 0.00000147, -0.00000006, -0.00000144, -0.00000146, 0.00000029, 0.00000324, 0.00000620, 0.00000796, 0.00000767, 0.00000512, 0.00000085, -0.00000371, -0.00000668, -0.00000661, -0.00000350, 0.00000081, 0.00000343, 0.00000218, -0.00000272, -0.00000846, -0.00001121, -0.00000857, -0.00000127, 0.00000729, 0.00001319, 0.00001451, 0.00001209, 0.00000826, 0.00000473, 0.00000132, -0.00000302, -0.00000821, -0.00001173, -0.00000993, -0.00000113, 0.00001168, 0.00002141, 0.00002096, 0.00000834, -0.00001089, -0.00002628, -0.00002893, -0.00001736, 0.00000134, 0.00001619, 0.00001962, 0.00001219, 0.00000132, -0.00000486, -0.00000380, 0.00000006, -0.00000033, -0.00000782, -0.00001771, -0.00002079, -0.00001111, 0.00000796, 0.00002440, 0.00002554, 0.00000790, -0.00001862, -0.00003619, -0.00003105, -0.00000392, 0.00002905, 0.00004616, 0.00003461, -0.00000027, -0.00003773, -0.00005475, -0.00004092, -0.00000511, 0.00003053, 0.00004494, 0.00003136, 0.00000069, -0.00002669, -0.00003446, -0.00002022, 0.00000453, 0.00002320, 0.00002523, 0.00001184, -0.00000641, -0.00001778, -0.00001685, -0.00000676, 0.00000453, 0.00001023, 0.00000865, 0.00000291, -0.00000256, -0.00000527, -0.00000560, -0.00000530, -0.00000533, -0.00000521, -0.00000396, -0.00000159, 0.00000058, 0.00000093, -0.00000085, -0.00000333, -0.00000441, -0.00000305, -0.00000026, 0.00000168, 0.00000112, -0.00000154, -0.00000416, -0.00000455, -0.00000235, 0.00000058, 0.00000164, -0.00000039, -0.00000425, -0.00000695, -0.00000600, -0.00000128, 0.00000483, 0.00000906, 0.00000944, 0.00000648, 0.00000254, 0.00000002, -0.00000019, 0.00000100, 0.00000191, },
		{ -0.00000001, 0.00000072, 0.00000167, 0.00000182, 0.00000106, 0.00000006, -0.00000037, 0.00000022, 0.00000175, 0.00000383, 0.00000586, 0.00000708, 0.00000660, 0.00000389, -0.00000060, -0.00000523, -0.00000772, -0.00000660, -0.00000240, 0.00000230, 0.00000437, 0.00000223, -0.00000292, -0.00000778, -0.00000917, -0.00000617, -0.00000060, 0.00000457, 0.00000753, 0.00000873, 0.00000978, 0.00001124, 0.00001138, 0.00000729, -0.00000205, -0.00001343, -0.00002023, -0.00001647, -0.00000173, 0.00001713, 0.00002924, 0.00002649, 0.00000925, -0.00001324, -0.00002842, -0.00002841, -0.00001484, 0.00000274, 0.00001384, 0.00001412, 0.00000746, 0.00000173, 0.00000179, 0.00000543, 0.00000564, -0.00000265, -0.00001688, -0.00002728, -0.00002395, -0.00000529, 0.00001887, 0.00003275, 0.00002555, 0.00000006, -0.00002775, -0.00003880, -0.00002405, 0.00000852, 0.00003835, 0.00004565, 0.00002467, -0.00001218, -0.00004232, -0.00004769, -0.00002626, 0.00000732, 0.00003182, 0.00003339, 0.00001371, -0.00001246, -0.00002794, -0.00002435, -0.00000640, 0.00001276, 0.00002109, 0.00001517, 0.00000107, -0.00001117, -0.00001463, -0.00000929, -0.00000053, 0.00000561, 0.00000654, 0.00000381, 0.00000069, -0.00000082, -0.00000116, -0.00000198, -0.00000413, -0.00000666, -0.00000772, -0.00000638, -0.00000352, -0.00000115, -0.00000069, -0.00000181, -0.00000280, -0.00000221, -0.00000016, 0.00000167, 0.00000149, -0.00000093, -0.00000397, -0.00000539, -0.00000413, -0.00000125, 0.00000089, 0.00000049, -0.00000217, -0.00000489, -0.00000521, -0.00000224, 0.00000265, 0.00000678, 0.00000807, 0.00000634, 0.00000318, 0.00000070, },
		{ 0.00000596, 0.00000370, 0.00000148, 0.00000037, 0.00000038, 0.00000082, 0.00000107, 0.00000103, 0.00000102, 0.00000126, 0.00000171, 0.00000220, 0.00000281, 0.00000377, 0.00000505, 0.00000593, 0.00000524, 0.00000227, -0.00000234, -0.00000648, -0.00000781, -0.00000542, -0.00000076, 0.00000319, 0.00000388, 0.00000108, -0.00000302, -0.00000554, -0.00000515, -0.00000290, -0.00000103, -0.00000059, -0.00000020, 0.00000279, 0.00000939, 0.00001677, 0.00001920, 0.00001204, -0.00000381, -0.00002082, -0.00002858, -0.00002060, 0.00000048, 0.00002348, 0.00003514, 0.00002854, 0.00000767, -0.00001517, -0.00002744, -0.00002425, -0.00001072, 0.00000277, 0.00000848, 0.00000644, 0.00000300, 0.00000435, 0.00001043, 0.00001450, 0.00000892, -0.00000736, -0.00002593, -0.00003380, -0.00002291, 0.00000266, 0.00002785, 0.00003615, 0.00002121, -0.00000793, -0.00003208, -0.00003464, -0.00001306, 0.00001905, 0.00004081, 0.00003798, 0.00001234, -0.00001979, -0.00003865, -0.00003402, -0.00001122, 0.00001346, 0.00002424, 0.00001609, -0.00000314, -0.00001926, -0.00002186, -0.00001074, 0.00000520, 0.00001513, 0.00001368, 0.00000372, -0.00000691, -0.00001146, -0.00000852, -0.00000188, 0.00000326, 0.00000425, 0.00000222, 0.00000023, 0.00000024, 0.00000163, 0.00000212, 0.00000008, -0.00000397, -0.00000778, -0.00000924, -0.00000798, -0.00000546, -0.00000350, -0.00000278, -0.00000249, -0.00000142, 0.00000068, 0.00000259, 0.00000268, 0.00000037, -0.00000313, -0.00000562, -0.00000554, -0.00000320, -0.00000057, 0.00000032, -0.00000104, -0.00000327, -0.00000419, -0.00000251, 0.00000117, 0.00000488, 0.00000667, },
		{ 0.00000024, 0.00000325, 0.00000509, 0.00000512, 0.00000386, 0.00000239, 0.00000139, 0.00000084, 0.00000034, -0.00000021, -0.00000040, 0.00000026, 0.00000172, 0.00000329, 0.00000410, 0.00000396, 0.00000348, 0.00000353, 0.00000422, 0.00000458, 0.00000327, -0.00000011, -0.00000430, -0.00000697, -0.00000642, -0.00000300, 0.00000092, 0.00000275, 0.00000165, -0.00000083, -0.00000219, -0.00000129, 0.00000041, 0.00000010, -0.00000352, -0.00000814, -0.00000879, -0.00000171, 0.00001170, 0.00002413, 0.00002641, 0.00001401, -0.00000849, -0.00002897, -0.00003477, -0.00002115, 0.00000500, 0.00002899, 0.00003744, 0.00002668, 0.00000459, -0.00001523, -0.00002267, -0.00001713, -0.00000637, 0.00000061, 0.00000064, -0.00000189, 0.00000009, 0.00000907, 0.00001953, 0.00002164, 0.00000939, -0.00001311, -0.00003274, -0.00003584, -0.00001829, 0.00001040, 0.00003260, 0.00003422, 0.00001450, -0.00001312, -0.00002996, -0.00002477, -0.00000126, 0.00002468, 0.00003595, 0.00002577, 0.00000161, -0.00002062, -0.00002772, -0.00001777, -0.00000030, 0.00001139, 0.00000971, -0.00000252, -0.00001515, -0.00001865, -0.00001078, 0.00000244, 0.00001182, 0.00001170, 0.00000351, -0.00000613, -0.00001066, -0.00000810, -0.00000162, 0.00000356, 0.00000434, 0.00000158, -0.00000129, -0.00000145, 0.00000109, 0.00000380, 0.00000393, 0.00000071, -0.00000422, -0.00000827, -0.00000985, -0.00000921, -0.00000766, -0.00000619, -0.00000468, -0.00000243, 0.00000065, 0.00000349, 0.00000442, 0.00000259, -0.00000118, -0.00000472, -0.00000606, -0.00000482, -0.00000237, -0.00000070, -0.00000088, -0.00000228, -0.00000322, -0.00000236, },
		{ -0.00000219, -0.00000269, -0.00000203, -0.00000012, 0.00000209, 0.00000348, 0.00000365, 0.00000314, 0.00000272, 0.00000262, 0.00000233, 0.00000123, -0.00000057, -0.00000205, -0.00000197, 0.00000004, 0.00000306, 0.00000546, 0.00000614, 0.00000531, 0.00000417, 0.00000366, 0.00000357, 0.00000273, 0.00000024, -0.00000332, -0.00000602, -0.00000601, -0.00000324, 0.00000026, 0.00000181, 0.00000045, -0.00000207, -0.00000263, 0.00000034, 0.00000489, 0.00000646, 0.00000155, -0.00000853, -0.00001718, -0.00001651, -0.00000343, 0.00001662, 0.00003178, 0.00003109, 0.00001234, -0.00001524, -0.00003589, -0.00003713, -0.00001795, 0.00001045, 0.00003182, 0.00003523, 0.00002153, 0.00000150, -0.00001239, -0.00001445, -0.00000849, -0.00000310, -0.00000378, -0.00000833, -0.00000918, -0.00000065, 0.00001486, 0.00002723, 0.00002569, 0.00000764, -0.00001792, -0.00003543, -0.00003320, -0.00001187, 0.00001552, 0.00003201, 0.00002817, 0.00000821, -0.00001323, -0.00002149, -0.00001180, 0.00000807, 0.00002395, 0.00002547, 0.00001284, -0.00000428, -0.00001447, -0.00001280, -0.00000367, 0.00000346, 0.00000176, -0.00000789, -0.00001789, -0.00002006, -0.00001188, 0.00000163, 0.00001177, 0.00001246, 0.00000430, -0.00000636, -0.00001226, -0.00001026, -0.00000295, 0.00000400, 0.00000619, 0.00000340, -0.00000099, -0.00000306, -0.00000135, 0.00000238, 0.00000494, 0.00000420, 0.00000042, -0.00000437, -0.00000812, -0.00001003, -0.00001048, -0.00001002, -0.00000859, -0.00000575, -0.00000155, 0.00000280, 0.00000539, 0.00000490, 0.00000167, -0.00000240, -0.00000508, -0.00000531, -0.00000377, -0.00000212, -0.00000162, },
		{ -0.00000378, -0.00000296, -0.00000266, -0.00000295, -0.00000300, -0.00000204, -0.00000018, 0.00000161, 0.00000238, 0.00000204, 0.00000150, 0.00000177, 0.00000293, 0.00000391, 0.00000336, 0.00000093, -0.00000218, -0.00000392, -0.00000292, 0.00000047, 0.00000442, 0.00000698, 0.00000745, 0.00000656, 0.00000546, 0.00000447, 0.00000286, -0.00000017, -0.00000413, -0.00000701, -0.00000672, -0.00000306, 0.00000153, 0.00000350, 0.00000108, -0.00000380, -0.00000649, -0.00000329, 0.00000491, 0.00001222, 0.00001156, 0.00000047, -0.00001572, -0.00002615, -0.00002163, -0.00000180, 0.00002308, 0.00003774, 0.00003182, 0.00000720, -0.00002235, -0.00003963, -0.00003490, -0.00001193, 0.00001496, 0.00003063, 0.00002871, 0.00001454, -0.00000016, -0.00000643, -0.00000403, -0.00000002, -0.00000168, -0.00000957, -0.00001667, -0.00001417, 0.00000051, 0.00002018, 0.00003197, 0.00002647, 0.00000514, -0.00001980, -0.00003317, -0.00002703, -0.00000596, 0.00001634, 0.00002653, 0.00002044, 0.00000503, -0.00000748, -0.00000858, 0.00000093, 0.00001263, 0.00001750, 0.00001267, 0.00000289, -0.00000397, -0.00000356, 0.00000159, 0.00000441, -0.00000058, -0.00001242, -0.00002390, -0.00002676, -0.00001794, -0.00000203, 0.00001163, 0.00001536, 0.00000811, -0.00000426, -0.00001336, -0.00001400, -0.00000707, 0.00000188, 0.00000704, 0.00000623, 0.00000169, -0.00000235, -0.00000291, -0.00000017, 0.00000330, 0.00000482, 0.00000336, -0.00000021, -0.00000427, -0.00000774, -0.00001030, -0.00001188, -0.00001201, -0.00000997, -0.00000565, -0.00000022, 0.00000421, 0.00000577, 0.00000410, 0.00000059, -0.00000260, -0.00000404, },
		{ 0.00000253, 0.00000009, -0.00000144, -0.00000188, -0.00000207, -0.00000275, -0.00000375, -0.00000409, -0.00000296, -0.00000067, 0.00000146, 0.00000208, 0.00000108, -0.00000025, -0.00000027, 0.00000159, 0.00000414, 0.00000528, 0.00000365, -0.00000012, -0.00000376, -0.00000493, -0.00000290, 0.00000110, 0.00000497, 0.00000733, 0.00000821, 0.00000833, 0.00000783, 0.00000585, 0.00000150, -0.00000458, -0.00000962, -0.00001034, -0.00000558, 0.00000199, 0.00000711, 0.00000565, -0.00000178, -0.00000949, -0.00001045, -0.00000198, 0.00001122, 0.00001934, 0.00001435, -0.00000342, -0.00002389, -0.00003317, -0.00002299, 0.00000273, 0.00002927, 0.00004019, 0.00002811, -0.00000015, -0.00002779, -0.00003889, -0.00002847, -0.00000485, 0.00001682, 0.00002508, 0.00001910, 0.00000748, 0.00000046, 0.00000183, 0.00000677, 0.00000684, -0.00000204, -0.00001532, -0.00002276, -0.00001639, 0.00000255, 0.00002343, 0.00003307, 0.00002494, 0.00000379, -0.00001748, -0.00002648, -0.00001940, -0.00000263, 0.00001256, 0.00001808, 0.00001389, 0.00000655, 0.00000314, 0.00000573, 0.00001052, 0.00001194, 0.00000780, 0.00000132, -0.00000176, 0.00000150, 0.00000817, 0.00001112, 0.00000466, -0.00001052, -0.00002679, -0.00003439, -0.00002809, -0.00001103, 0.00000716, 0.00001659, 0.00001325, 0.00000112, -0.00001117, -0.00001619, -0.00001215, -0.00000306, 0.00000479, 0.00000732, 0.00000474, 0.00000042, -0.00000205, -0.00000138, 0.00000120, 0.00000342, 0.00000374, 0.00000207, -0.00000082, -0.00000419, -0.00000770, -0.00001097, -0.00001310, -0.00001292, -0.00000982, -0.00000456, 0.00000083, 0.00000416, 0.00000447, },
		{ -0.00000309, 0.00000079, 0.00000232, 0.00000191, 0.00000098, 0.00000070, 0.00000097, 0.00000078, -0.00000071, -0.00000310, -0.00000486, -0.00000455, -0.00000217, 0.00000070, 0.00000203, 0.00000103, -0.00000114, -0.00000226, -0.00000085, 0.00000249, 0.00000544, 0.00000573, 0.00000289, -0.00000133, -0.00000433, -0.00000450, -0.00000219, 0.00000106, 0.00000398, 0.00000647, 0.00000901, 0.00001123, 0.00001132, 0.00000718, -0.00000123, -0.00001058, -0.00001552, -0.00001227, -0.00000208, 0.00000865, 0.00001233, 0.00000578, -0.00000667, -0.00001552, -0.00001277, 0.00000155, 0.00001827, 0.00002474, 0.00001396, -0.00000949, -0.00003133, -0.00003667, -0.00002038, 0.00000877, 0.00003324, 0.00003808, 0.00002067, -0.00000775, -0.00002982, -0.00003349, -0.00001932, 0.00000131, 0.00001507, 0.00001596, 0.00000829, 0.00000185, 0.00000330, 0.00001085, 0.00001610, 0.00001137, -0.00000335, -0.00001945, -0.00002566, -0.00001632, 0.00000408, 0.00002364, 0.00003101, 0.00002278, 0.00000515, -0.00001090, -0.00001712, -0.00001263, -0.00000292, 0.00000541, 0.00000942, 0.00001081, 0.00001279, 0.00001618, 0.00001827, 0.00001550, 0.00000742, -0.00000173, -0.00000565, -0.00000103, 0.00000905, 0.00001637, 0.00001312, -0.00000213, -0.00002258, -0.00003711, -0.00003736, -0.00002321, -0.00000282, 0.00001259, 0.00001574, 0.00000726, -0.00000569, -0.00001462, -0.00001507, -0.00000846, -0.00000005, 0.00000520, 0.00000562, 0.00000297, 0.00000029, -0.00000046, 0.00000062, 0.00000216, 0.00000288, 0.00000237, 0.00000085, -0.00000149, -0.00000467, -0.00000846, -0.00001190, -0.00001350, -0.00001216, -0.00000812, },
		{ -0.00001255, -0.00001240, -0.00000952, -0.00000540, -0.00000211, -0.00000083, -0.00000104, -0.00000110, 0.00000019, 0.00000250, 0.00000407, 0.00000327, 0.00000010, -0.00000359, -0.00000537, -0.00000420, -0.00000121, 0.00000114, 0.00000106, -0.00000111, -0.00000317, -0.00000295, -0.00000006, 0.00000367, 0.00000567, 0.00000463, 0.00000144, -0.00000169, -0.00000306, -0.00000264, -0.00000155, -0.00000050, 0.00000122, 0.00000495, 0.00001065, 0.00001557, 0.00001540, 0.00000756, -0.00000577, -0.00001770, -0.00002065, -0.00001186, 0.00000371, 0.00001584, 0.00001593, 0.00000345, -0.00001284, -0.00002057, -0.00001266, 0.00000682, 0.00002460, 0.00002713, 0.00001040, -0.00001637, -0.00003625, -0.00003587, -0.00001469, 0.00001439, 0.00003347, 0.00003150, 0.00001119, -0.00001358, -0.00002761, -0.00002443, -0.00000954, 0.00000497, 0.00000983, 0.00000495, -0.00000174, -0.00000159, 0.00000740, 0.00001878, 0.00002266, 0.00001370, -0.00000436, -0.00002083, -0.00002537, -0.00001504, 0.00000399, 0.00002087, 0.00002721, 0.00002178, 0.00000988, -0.00000129, -0.00000741, -0.00000841, -0.00000642, -0.00000273, 0.00000327, 0.00001218, 0.00002220, 0.00002865, 0.00002689, 0.00001624, 0.00000184, -0.00000799, -0.00000728, 0.00000295, 0.00001465, 0.00001782, 0.00000740, -0.00001288, -0.00003248, -0.00004066, -0.00003328, -0.00001516, 0.00000329, 0.00001274, 0.00001026, 0.00000009, -0.00001007, -0.00001427, -0.00001144, -0.00000487, 0.00000101, 0.00000360, 0.00000323, 0.00000186, 0.00000120, 0.00000159, 0.00000230, 0.00000259, 0.00000223, 0.00000134, -0.00000018, -0.00000268, -0.00000623, -0.00001004, },
		{ -0.00000136, -0.00000491, -0.00000901, -0.00001196, -0.00001217, -0.00000945, -0.00000545, -0.00000261, -0.00000237, -0.00000400, -0.00000513, -0.00000364, 0.00000051, 0.00000505, 0.00000705, 0.00000514, 0.00000061, -0.00000357, -0.00000495, -0.00000341, -0.00000100, -0.00000007, -0.00000132, -0.00000331, -0.00000380, -0.00000181, 0.00000153, 0.00000393, 0.00000393, 0.00000208, 0.00000026, -0.00000017, 0.00000031, -0.00000002, -0.00000203, -0.00000412, -0.00000291, 0.00000372, 0.00001377, 0.00002107, 0.00001910, 0.00000613, -0.00001207, -0.00002488, -0.00002377, -0.00000865, 0.00001105, 0.00002217, 0.00001696, -0.00000109, -0.00001906, -0.00002346, -0.00001002, 0.00001270, 0.00002877, 0.00002595, 0.00000455, -0.00002231, -0.00003740, -0.00003105, -0.00000760, 0.00001773, 0.00002940, 0.00002170, 0.00000182, -0.00001618, -0.00002151, -0.00001361, -0.00000118, 0.00000543, 0.00000223, -0.00000588, -0.00000958, -0.00000302, 0.00001128, 0.00002406, 0.00002600, 0.00001460, -0.00000399, -0.00001917, -0.00002276, -0.00001378, 0.00000193, 0.00001622, 0.00002358, 0.00002317, 0.00001742, 0.00000923, 0.00000050, -0.00000721, -0.00001147, -0.00000915, 0.00000133, 0.00001743, 0.00003228, 0.00003808, 0.00003107, 0.00001469, -0.00000189, -0.00000945, -0.00000488, 0.00000659, 0.00001481, 0.00001158, -0.00000368, -0.00002343, -0.00003690, -0.00003703, -0.00002468, -0.00000754, 0.00000514, 0.00000813, 0.00000264, -0.00000570, -0.00001112, -0.00001118, -0.00000730, -0.00000269, 0.00000023, 0.00000118, 0.00000127, 0.00000164, 0.00000245, 0.00000317, 0.00000331, 0.00000290, 0.00000214, 0.00000090, },
		{ 0.00000389, 0.00000372, 0.00000296, 0.00000081, -0.00000326, -0.00000845, -0.00001258, -0.00001337, -0.00001017, -0.00000489, -0.00000090, -0.00000072, -0.00000411, -0.00000803, -0.00000886, -0.00000502, 0.00000173, 0.00000756, 0.00000924, 0.00000639, 0.00000142, -0.00000248, -0.00000370, -0.00000302, -0.00000244, -0.00000310, -0.00000428, -0.00000426, -0.00000218, 0.00000088, 0.00000273, 0.00000209, -0.00000005, -0.00000114, 0.00000053, 0.00000385, 0.00000535, 0.00000224, -0.00000448, -0.00000967, -0.00000744, 0.00000375, 0.00001848, 0.00002687, 0.00002125, 0.00000245, -0.00001939, -0.00003075, -0.00002392, -0.00000288, 0.00001866, 0.00002633, 0.00001510, -0.00000694, -0.00002398, -0.00002357, -0.00000551, 0.00001775, 0.00002976, 0.00002155, -0.00000203, -0.00002573, -0.00003448, -0.00002351, -0.00000114, 0.00001764, 0.00002164, 0.00001069, -0.00000548, -0.00001509, -0.00001292, -0.00000326, 0.00000436, 0.00000308, -0.00000595, -0.00001472, -0.00001466, -0.00000336, 0.00001346, 0.00002588, 0.00002648, 0.00001506, -0.00000175, -0.00001510, -0.00001911, -0.00001331, -0.00000146, 0.00001152, 0.00002183, 0.00002717, 0.00002618, 0.00001839, 0.00000540, -0.00000820, -0.00001570, -0.00001174, 0.00000390, 0.00002477, 0.00004067, 0.00004342, 0.00003201, 0.00001340, -0.00000196, -0.00000680, -0.00000159, 0.00000627, 0.00000792, -0.00000069, -0.00001612, -0.00002996, -0.00003463, -0.00002820, -0.00001519, -0.00000312, 0.00000246, 0.00000086, -0.00000436, -0.00000863, -0.00000942, -0.00000724, -0.00000426, -0.00000223, -0.00000137, -0.00000079, 0.00000029, 0.00000180, 0.00000310, 0.00000376, },
		{ 0.00000002, 0.00000148, 0.00000252, 0.00000348, 0.00000465, 0.00000549, 0.00000450, 0.00000040, -0.00000634, -0.00001300, -0.00001600, -0.00001338, -0.00000651, 0.00000036, 0.00000287, -0.00000041, -0.00000692, -0.00001179, -0.00001125, -0.00000514, 0.00000313, 0.00000912, 0.00001034, 0.00000750, 0.00000324, -0.00000016, -0.00000227, -0.00000404, -0.00000604, -0.00000739, -0.00000648, -0.00000281, 0.00000178, 0.00000409, 0.00000218, -0.00000249, -0.00000571, -0.00000370, 0.00000327, 0.00001025, 0.00001093, 0.00000294, -0.00000922, -0.00001636, -0.00001114, 0.00000567, 0.00002428, 0.00003171, 0.00002095, -0.00000325, -0.00002650, -0.00003399, -0.00002077, 0.00000451, 0.00002508, 0.00002747, 0.00001078, -0.00001281, -0.00002658, -0.00002102, -0.00000035, 0.00002070, 0.00002737, 0.00001511, -0.00000763, -0.00002577, -0.00002824, -0.00001519, 0.00000305, 0.00001398, 0.00001181, 0.00000066, -0.00000953, -0.00001102, -0.00000387, 0.00000476, 0.00000667, -0.00000085, -0.00001288, -0.00002051, -0.00001732, -0.00000383, 0.00001297, 0.00002431, 0.00002497, 0.00001574, 0.00000198, -0.00000993, -0.00001561, -0.00001359, -0.00000471, 0.00000868, 0.00002286, 0.00003292, 0.00003415, 0.00002459, 0.00000735, -0.00000959, -0.00001697, -0.00000963, 0.00000988, 0.00003212, 0.00004601, 0.00004528, 0.00003193, 0.00001431, 0.00000144, -0.00000275, -0.00000110, -0.00000019, -0.00000501, -0.00001522, -0.00002568, -0.00003041, -0.00002690, -0.00001765, -0.00000813, -0.00000296, -0.00000314, -0.00000623, -0.00000873, -0.00000870, -0.00000662, -0.00000429, -0.00000304, -0.00000277, -0.00000251, -0.00000153, },
		{ -0.00000322, -0.00000306, -0.00000225, -0.00000111, -0.00000036, -0.00000006, 0.00000073, 0.00000295, 0.00000617, 0.00000810, 0.00000598, -0.00000104, -0.00001052, -0.00001760, -0.00001809, -0.00001154, -0.00000189, 0.00000491, 0.00000481, -0.00000163, -0.00000975, -0.00001404, -0.00001179, -0.00000450, 0.00000372, 0.00000915, 0.00001062, 0.00000922, 0.00000653, 0.00000300, -0.00000172, -0.00000729, -0.00001161, -0.00001175, -0.00000655, 0.00000156, 0.00000731, 0.00000632, -0.00000119, -0.00000966, -0.00001179, -0.00000427, 0.00000883, 0.00001813, 0.00001562, 0.00000124, -0.00001583, -0.00002289, -0.00001292, 0.00000947, 0.00003007, 0.00003437, 0.00001786, -0.00001008, -0.00003193, -0.00003377, -0.00001478, 0.00001212, 0.00002907, 0.00002549, 0.00000504, -0.00001742, -0.00002643, -0.00001663, 0.00000411, 0.00002084, 0.00002227, 0.00000825, -0.00001093, -0.00002251, -0.00002025, -0.00000801, 0.00000411, 0.00000771, 0.00000197, -0.00000670, -0.00001025, -0.00000550, 0.00000364, 0.00000943, 0.00000633, -0.00000478, -0.00001723, -0.00002312, -0.00001848, -0.00000543, 0.00000966, 0.00002013, 0.00002237, 0.00001667, 0.00000612, -0.00000510, -0.00001287, -0.00001378, -0.00000607, 0.00000903, 0.00002652, 0.00003890, 0.00003964, 0.00002735, 0.00000756, -0.00000941, -0.00001409, -0.00000337, 0.00001750, 0.00003801, 0.00004853, 0.00004558, 0.00003284, 0.00001768, 0.00000597, -0.00000107, -0.00000618, -0.00001256, -0.00002062, -0.00002749, -0.00002946, -0.00002516, -0.00001689, -0.00000910, -0.00000536, -0.00000613, -0.00000898, -0.00001075, -0.00000992, -0.00000725, -0.00000465, -0.00000338, },
		{ -0.00000885, -0.00000616, -0.00000481, -0.00000438, -0.00000362, -0.00000203, -0.00000048, -0.00000032, -0.00000170, -0.00000292, -0.00000159, 0.00000300, 0.00000856, 0.00001077, 0.00000641, -0.00000383, -0.00001501, -0.00002085, -0.00001793, -0.00000816, 0.00000235, 0.00000736, 0.00000447, -0.00000358, -0.00001120, -0.00001388, -0.00001071, -0.00000413, 0.00000260, 0.00000761, 0.00001081, 0.00001233, 0.00001125, 0.00000607, -0.00000312, -0.00001304, -0.00001822, -0.00001465, -0.00000348, 0.00000851, 0.00001298, 0.00000612, -0.00000775, -0.00001837, -0.00001667, -0.00000191, 0.00001675, 0.00002603, 0.00001811, -0.00000300, -0.00002323, -0.00002788, -0.00001221, 0.00001446, 0.00003452, 0.00003400, 0.00001239, -0.00001667, -0.00003449, -0.00003000, -0.00000706, 0.00001850, 0.00002998, 0.00002100, -0.00000082, -0.00001990, -0.00002388, -0.00001167, 0.00000677, 0.00001827, 0.00001579, 0.00000255, -0.00001140, -0.00001699, -0.00001242, -0.00000334, 0.00000228, 0.00000052, -0.00000596, -0.00001062, -0.00000853, -0.00000035, 0.00000824, 0.00001076, 0.00000458, -0.00000733, -0.00001855, -0.00002321, -0.00001919, -0.00000857, 0.00000414, 0.00001441, 0.00001914, 0.00001715, 0.00000924, -0.00000166, -0.00001070, -0.00001256, -0.00000414, 0.00001291, 0.00003182, 0.00004359, 0.00004195, 0.00002745, 0.00000778, -0.00000635, -0.00000724, 0.00000553, 0.00002523, 0.00004233, 0.00004986, 0.00004643, 0.00003544, 0.00002159, 0.00000788, -0.00000509, -0.00001743, -0.00002804, -0.00003428, -0.00003383, -0.00002690, -0.00001684, -0.00000846, -0.00000509, -0.00000667, -0.00001027, -0.00001244, -0.00001164, },
		{ -0.00001010, -0.00001263, -0.00001202, -0.00000947, -0.00000739, -0.00000697, -0.00000722, -0.00000614, -0.00000300, 0.00000064, 0.00000200, -0.00000025, -0.00000437, -0.00000642, -0.00000336, 0.00000408, 0.00001116, 0.00001225, 0.00000498, -0.00000753, -0.00001838, -0.00002141, -0.00001531, -0.00000439, 0.00000455, 0.00000676, 0.00000230, -0.00000483, -0.00001003, -0.00001114, -0.00000893, -0.00000518, -0.00000065, 0.00000511, 0.00001191, 0.00001721, 0.00001673, 0.00000781, -0.00000721, -0.00002078, -0.00002438, -0.00001469, 0.00000290, 0.00001683, 0.00001704, 0.00000255, -0.00001690, -0.00002698, -0.00001892, 0.00000351, 0.00002579, 0.00003232, 0.00001762, -0.00000914, -0.00002996, -0.00003025, -0.00000918, 0.00001944, 0.00003646, 0.00003048, 0.00000559, -0.00002162, -0.00003354, -0.00002336, 0.00000091, 0.00002256, 0.00002792, 0.00001514, -0.00000560, -0.00002006, -0.00001986, -0.00000738, 0.00000720, 0.00001385, 0.00000949, -0.00000097, -0.00000942, -0.00001082, -0.00000639, -0.00000164, -0.00000134, -0.00000577, -0.00001078, -0.00001140, -0.00000588, 0.00000290, 0.00000949, 0.00000960, 0.00000276, -0.00000778, -0.00001723, -0.00002181, -0.00002012, -0.00001293, -0.00000251, 0.00000801, 0.00001516, 0.00001609, 0.00001012, 0.00000003, -0.00000834, -0.00000875, 0.00000161, 0.00001969, 0.00003748, 0.00004622, 0.00004160, 0.00002651, 0.00000947, -0.00000022, 0.00000235, 0.00001548, 0.00003261, 0.00004637, 0.00005211, 0.00004892, 0.00003834, 0.00002251, 0.00000348, -0.00001616, -0.00003273, -0.00004208, -0.00004172, -0.00003276, -0.00001993, -0.00000926, -0.00000470, -0.00000614, },
		{ -0.00001224, -0.00000597, -0.00000680, -0.00001066, -0.00001282, -0.00001132, -0.00000795, -0.00000609, -0.00000736, -0.00001011, -0.00001076, -0.00000715, -0.00000072, 0.00000419, 0.00000385, -0.00000162, -0.00000783, -0.00000926, -0.00000364, 0.00000587, 0.00001264, 0.00001128, 0.00000164, -0.00001085, -0.00001893, -0.00001849, -0.00001099, -0.00000196, 0.00000329, 0.00000309, -0.00000040, -0.00000380, -0.00000558, -0.00000653, -0.00000781, -0.00000854, -0.00000573, 0.00000274, 0.00001478, 0.00002362, 0.00002178, 0.00000708, -0.00001404, -0.00002927, -0.00002842, -0.00001106, 0.00001187, 0.00002475, 0.00001803, -0.00000444, -0.00002724, -0.00003369, -0.00001772, 0.00001126, 0.00003427, 0.00003568, 0.00001419, -0.00001596, -0.00003467, -0.00002962, -0.00000467, 0.00002311, 0.00003526, 0.00002445, -0.00000110, -0.00002386, -0.00002925, -0.00001513, 0.00000772, 0.00002382, 0.00002363, 0.00000919, -0.00000863, -0.00001835, -0.00001554, -0.00000460, 0.00000570, 0.00000886, 0.00000465, -0.00000212, -0.00000611, -0.00000561, -0.00000308, -0.00000236, -0.00000519, -0.00000976, -0.00001229, -0.00001014, -0.00000378, 0.00000350, 0.00000789, 0.00000722, 0.00000181, -0.00000622, -0.00001424, -0.00001990, -0.00002137, -0.00001767, -0.00000925, 0.00000132, 0.00000985, 0.00001248, 0.00000821, 0.00000032, -0.00000492, -0.00000187, 0.00001068, 0.00002817, 0.00004258, 0.00004708, 0.00004009, 0.00002618, 0.00001325, 0.00000809, 0.00001307, 0.00002566, 0.00004049, 0.00005213, 0.00005670, 0.00005219, 0.00003827, 0.00001646, -0.00000932, -0.00003292, -0.00004779, -0.00005000, -0.00004062, -0.00002550, },
		{ -0.00005756, -0.00004868, -0.00003149, -0.00001556, -0.00000779, -0.00000880, -0.00001356, -0.00001573, -0.00001244, -0.00000614, -0.00000206, -0.00000369, -0.00000963, -0.00001459, -0.00001363, -0.00000631, 0.00000270, 0.00000710, 0.00000385, -0.00000428, -0.00001071, -0.00001003, -0.00000224, 0.00000719, 0.00001140, 0.00000720, -0.00000270, -0.00001195, -0.00001535, -0.00001233, -0.00000664, -0.00000270, -0.00000202, -0.00000264, -0.00000181, 0.00000068, 0.00000185, -0.00000148, -0.00000873, -0.00001441, -0.00001166, 0.00000173, 0.00001974, 0.00003068, 0.00002502, 0.00000323, -0.00002284, -0.00003681, -0.00002899, -0.00000388, 0.00002187, 0.00003041, 0.00001518, -0.00001394, -0.00003689, -0.00003709, -0.00001313, 0.00001991, 0.00004055, 0.00003551, 0.00000862, -0.00002200, -0.00003645, -0.00002636, 0.00000006, 0.00002444, 0.00003105, 0.00001712, -0.00000633, -0.00002298, -0.00002250, -0.00000682, 0.00001241, 0.00002252, 0.00001827, 0.00000419, -0.00000981, -0.00001564, -0.00001195, -0.00000353, 0.00000313, 0.00000459, 0.00000188, -0.00000148, -0.00000290, -0.00000253, -0.00000243, -0.00000432, -0.00000785, -0.00001088, -0.00001122, -0.00000825, -0.00000325, 0.00000163, 0.00000459, 0.00000478, 0.00000204, -0.00000339, -0.00001068, -0.00001799, -0.00002255, -0.00002184, -0.00001541, -0.00000578, 0.00000254, 0.00000580, 0.00000368, -0.00000012, -0.00000004, 0.00000759, 0.00002179, 0.00003719, 0.00004710, 0.00004750, 0.00003927, 0.00002755, 0.00001869, 0.00001718, 0.00002392, 0.00003655, 0.00005067, 0.00006115, 0.00006314, 0.00005325, 0.00003113, 0.00000076, -0.00002983, -0.00005135, },
		{ 0.00001361, -0.00002453, -0.00005426, -0.00006598, -0.00005805, -0.00003787, -0.00001767, -0.00000741, -0.00000922, -0.00001709, -0.00002194, -0.00001833, -0.00000815, 0.00000124, 0.00000302, -0.00000379, -0.00001354, -0.00001837, -0.00001417, -0.00000368, 0.00000570, 0.00000763, 0.00000158, -0.00000703, -0.00001124, -0.00000792, -0.00000000, 0.00000623, 0.00000618, 0.00000043, -0.00000620, -0.00000899, -0.00000722, -0.00000435, -0.00000441, -0.00000787, -0.00001084, -0.00000851, -0.00000016, 0.00000894, 0.00001098, 0.00000223, -0.00001255, -0.00002220, -0.00001701, 0.00000305, 0.00002645, 0.00003701, 0.00002529, -0.00000362, -0.00003215, -0.00004168, -0.00002551, 0.00000579, 0.00003094, 0.00003242, 0.00000870, -0.00002432, -0.00004402, -0.00003652, -0.00000609, 0.00002772, 0.00004347, 0.00003204, 0.00000221, -0.00002601, -0.00003514, -0.00002145, 0.00000378, 0.00002303, 0.00002466, 0.00000994, -0.00000920, -0.00001934, -0.00001463, 0.00000029, 0.00001466, 0.00001946, 0.00001298, 0.00000067, -0.00000961, -0.00001292, -0.00000963, -0.00000377, 0.00000059, 0.00000187, 0.00000104, -0.00000017, -0.00000098, -0.00000189, -0.00000364, -0.00000617, -0.00000849, -0.00000949, -0.00000882, -0.00000693, -0.00000449, -0.00000186, 0.00000083, 0.00000294, 0.00000307, -0.00000029, -0.00000736, -0.00001613, -0.00002301, -0.00002479, -0.00002085, -0.00001364, -0.00000708, -0.00000369, -0.00000275, -0.00000078, 0.00000591, 0.00001827, 0.00003346, 0.00004622, 0.00005197, 0.00004928, 0.00004058, 0.00003073, 0.00002480, 0.00002608, 0.00003517, 0.00004974, 0.00006471, 0.00007319, 0.00006861, 0.00004777, },
		{ 0.00008261, 0.00008549, 0.00006840, 0.00003168, -0.00001503, -0.00005596, -0.00007667, -0.00007176, -0.00004800, -0.00002056, -0.00000418, -0.00000473, -0.00001668, -0.00002784, -0.00002824, -0.00001675, -0.00000122, 0.00000784, 0.00000502, -0.00000634, -0.00001706, -0.00001920, -0.00001171, -0.00000061, 0.00000612, 0.00000467, -0.00000216, -0.00000790, -0.00000796, -0.00000322, 0.00000129, 0.00000118, -0.00000330, -0.00000745, -0.00000663, -0.00000084, 0.00000454, 0.00000317, -0.00000626, -0.00001766, -0.00002146, -0.00001227, 0.00000566, 0.00002021, 0.00001987, 0.00000305, -0.00001933, -0.00003058, -0.00002031, 0.00000702, 0.00003386, 0.00004115, 0.00002202, -0.00001250, -0.00004019, -0.00004269, -0.00001835, 0.00001620, 0.00003725, 0.00003023, -0.00000025, -0.00003362, -0.00004738, -0.00003227, 0.00000185, 0.00003316, 0.00004265, 0.00002623, -0.00000362, -0.00002731, -0.00003132, -0.00001617, 0.00000563, 0.00001922, 0.00001738, 0.00000415, -0.00000948, -0.00001390, -0.00000703, 0.00000539, 0.00001477, 0.00001565, 0.00000859, -0.00000139, -0.00000876, -0.00001087, -0.00000856, -0.00000450, -0.00000104, 0.00000087, 0.00000134, 0.00000069, -0.00000092, -0.00000320, -0.00000547, -0.00000690, -0.00000711, -0.00000664, -0.00000639, -0.00000675, -0.00000699, -0.00000577, -0.00000245, 0.00000177, 0.00000418, 0.00000222, -0.00000461, -0.00001405, -0.00002229, -0.00002637, -0.00002581, -0.00002245, -0.00001864, -0.00001509, -0.00001030, -0.00000178, 0.00001166, 0.00002845, 0.00004461, 0.00005561, 0.00005867, 0.00005397, 0.00004455, 0.00003519, 0.00003084, 0.00003508, 0.00004839, 0.00006693, },
		{ 0.00003733, 0.00004603, 0.00006569, 0.00008856, 0.00010174, 0.00009288, 0.00005741, 0.00000306, -0.00005192, -0.00008714, -0.00009073, -0.00006601, -0.00002970, -0.00000220, 0.00000408, -0.00000888, -0.00002755, -0.00003673, -0.00002963, -0.00001147, 0.00000535, 0.00001061, 0.00000309, -0.00000958, -0.00001744, -0.00001563, -0.00000716, 0.00000055, 0.00000216, -0.00000159, -0.00000541, -0.00000481, -0.00000044, 0.00000253, -0.00000061, -0.00000902, -0.00001594, -0.00001400, -0.00000205, 0.00001239, 0.00001779, 0.00000764, -0.00001329, -0.00003089, -0.00003137, -0.00001196, 0.00001567, 0.00003270, 0.00002641, 0.00000010, -0.00002819, -0.00003786, -0.00002054, 0.00001319, 0.00004049, 0.00004198, 0.00001550, -0.00002186, -0.00004533, -0.00003946, -0.00000878, 0.00002537, 0.00003958, 0.00002431, -0.00000993, -0.00004016, -0.00004656, -0.00002548, 0.00000902, 0.00003531, 0.00003855, 0.00001944, -0.00000779, -0.00002596, -0.00002607, -0.00001167, 0.00000537, 0.00001391, 0.00001056, 0.00000048, -0.00000761, -0.00000788, -0.00000080, 0.00000829, 0.00001345, 0.00001200, 0.00000542, -0.00000240, -0.00000795, -0.00000973, -0.00000823, -0.00000494, -0.00000141, 0.00000112, 0.00000179, 0.00000035, -0.00000255, -0.00000545, -0.00000683, -0.00000616, -0.00000445, -0.00000356, -0.00000479, -0.00000760, -0.00000980, -0.00000906, -0.00000483, 0.00000090, 0.00000468, 0.00000369, -0.00000242, -0.00001150, -0.00002041, -0.00002692, -0.00003054, -0.00003187, -0.00003112, -0.00002730, -0.00001862, -0.00000401, 0.00001551, 0.00003667, 0.00005496, 0.00006630, 0.00006852, 0.00006223, 0.00005092, 0.00004044, },
		{ 0.00008200, 0.00007373, 0.00005925, 0.00004715, 0.00004652, 0.00006178, 0.00008836, 0.00011245, 0.00011656, 0.00008898, 0.00003208, -0.00003632, -0.00009059, -0.00011041, -0.00009189, -0.00004974, -0.00000862, 0.00001106, 0.00000415, -0.00001813, -0.00003720, -0.00003973, -0.00002542, -0.00000520, 0.00000782, 0.00000765, -0.00000160, -0.00001053, -0.00001237, -0.00000779, -0.00000284, -0.00000274, -0.00000686, -0.00000950, -0.00000563, 0.00000353, 0.00001034, 0.00000684, -0.00000764, -0.00002408, -0.00002938, -0.00001654, 0.00000844, 0.00002908, 0.00002959, 0.00000696, -0.00002525, -0.00004520, -0.00003792, -0.00000657, 0.00002867, 0.00004405, 0.00002888, -0.00000651, -0.00003755, -0.00004244, -0.00001741, 0.00002041, 0.00004488, 0.00003904, 0.00000679, -0.00002994, -0.00004654, -0.00003260, 0.00000137, 0.00003161, 0.00003767, 0.00001597, -0.00001847, -0.00004289, -0.00004208, -0.00001769, 0.00001408, 0.00003404, 0.00003224, 0.00001299, -0.00000981, -0.00002265, -0.00002060, -0.00000865, 0.00000342, 0.00000831, 0.00000524, -0.00000099, -0.00000449, -0.00000247, 0.00000350, 0.00000934, 0.00001154, 0.00000907, 0.00000338, -0.00000292, -0.00000754, -0.00000925, -0.00000798, -0.00000460, -0.00000072, 0.00000176, 0.00000154, -0.00000130, -0.00000510, -0.00000752, -0.00000709, -0.00000433, -0.00000156, -0.00000128, -0.00000428, -0.00000889, -0.00001199, -0.00001118, -0.00000640, -0.00000012, 0.00000419, 0.00000413, -0.00000062, -0.00000859, -0.00001783, -0.00002691, -0.00003495, -0.00004095, -0.00004320, -0.00003947, -0.00002804, -0.00000885, 0.00001595, 0.00004225, 0.00006495, 0.00007921, },
		{ 0.00001223, 0.00004535, 0.00007515, 0.00009444, 0.00009837, 0.00008763, 0.00006996, 0.00005782, 0.00006214, 0.00008488, 0.00011526, 0.00013317, 0.00011954, 0.00006828, -0.00000758, -0.00008052, -0.00012226, -0.00011908, -0.00007925, -0.00002782, 0.00000808, 0.00001441, -0.00000371, -0.00002804, -0.00004057, -0.00003476, -0.00001757, -0.00000204, 0.00000318, -0.00000081, -0.00000611, -0.00000602, -0.00000101, 0.00000233, -0.00000183, -0.00001212, -0.00001990, -0.00001626, -0.00000057, 0.00001699, 0.00002195, 0.00000701, -0.00002057, -0.00004217, -0.00004051, -0.00001305, 0.00002419, 0.00004639, 0.00003699, 0.00000024, -0.00004041, -0.00005764, -0.00003909, 0.00000363, 0.00004247, 0.00005193, 0.00002652, -0.00001565, -0.00004555, -0.00004329, -0.00001148, 0.00002715, 0.00004597, 0.00003275, -0.00000250, -0.00003523, -0.00004367, -0.00002349, 0.00001020, 0.00003395, 0.00003217, 0.00000697, -0.00002436, -0.00004169, -0.00003517, -0.00001049, 0.00001631, 0.00002995, 0.00002506, 0.00000781, -0.00000985, -0.00001840, -0.00001589, -0.00000721, 0.00000065, 0.00000353, 0.00000191, -0.00000076, -0.00000117, 0.00000155, 0.00000585, 0.00000918, 0.00000968, 0.00000699, 0.00000208, -0.00000336, -0.00000750, -0.00000895, -0.00000729, -0.00000348, 0.00000037, 0.00000197, 0.00000027, -0.00000377, -0.00000757, -0.00000864, -0.00000628, -0.00000220, 0.00000055, -0.00000026, -0.00000455, -0.00000986, -0.00001299, -0.00001203, -0.00000744, -0.00000156, 0.00000280, 0.00000374, 0.00000071, -0.00000587, -0.00001527, -0.00002658, -0.00003839, -0.00004843, -0.00005377, -0.00005145, -0.00003947, -0.00001756, },
		{ -0.00006238, -0.00006373, -0.00005382, -0.00003069, 0.00000460, 0.00004653, 0.00008549, 0.00011082, 0.00011608, 0.00010367, 0.00008538, 0.00007687, 0.00008821, 0.00011581, 0.00014172, 0.00014247, 0.00010312, 0.00002841, -0.00005697, -0.00012023, -0.00013817, -0.00010983, -0.00005636, -0.00000791, 0.00001377, 0.00000623, -0.00001578, -0.00003278, -0.00003398, -0.00002260, -0.00001025, -0.00000608, -0.00000953, -0.00001218, -0.00000657, 0.00000597, 0.00001539, 0.00001134, -0.00000689, -0.00002760, -0.00003431, -0.00001878, 0.00001081, 0.00003375, 0.00003121, 0.00000102, -0.00003858, -0.00006018, -0.00004630, -0.00000290, 0.00004299, 0.00006090, 0.00003788, -0.00001188, -0.00005597, -0.00006537, -0.00003405, 0.00001706, 0.00005446, 0.00005468, 0.00001968, -0.00002552, -0.00005054, -0.00004023, -0.00000401, 0.00003197, 0.00004343, 0.00002424, -0.00001070, -0.00003690, -0.00003748, -0.00001392, 0.00001625, 0.00003236, 0.00002449, -0.00000096, -0.00002684, -0.00003726, -0.00002738, -0.00000505, 0.00001575, 0.00002412, 0.00001817, 0.00000424, -0.00000857, -0.00001425, -0.00001245, -0.00000692, -0.00000197, 0.00000030, 0.00000046, 0.00000039, 0.00000149, 0.00000391, 0.00000666, 0.00000843, 0.00000819, 0.00000560, 0.00000113, -0.00000386, -0.00000753, -0.00000834, -0.00000602, -0.00000202, 0.00000112, 0.00000128, -0.00000181, -0.00000625, -0.00000916, -0.00000859, -0.00000492, -0.00000063, 0.00000136, -0.00000040, -0.00000506, -0.00001003, -0.00001271, -0.00001189, -0.00000816, -0.00000326, 0.00000083, 0.00000262, 0.00000115, -0.00000411, -0.00001325, -0.00002569, -0.00003985, -0.00005319, },
		{ -0.00001183, -0.00002353, -0.00003840, -0.00005478, -0.00006939, -0.00007704, -0.00007140, -0.00004763, -0.00000598, 0.00004596, 0.00009457, 0.00012600, 0.00013375, 0.00012334, 0.00011005, 0.00010967, 0.00012723, 0.00015145, 0.00015953, 0.00013100, 0.00006219, -0.00002817, -0.00010706, -0.00014493, -0.00013220, -0.00008412, -0.00003064, 0.00000175, 0.00000456, -0.00001114, -0.00002593, -0.00002710, -0.00001668, -0.00000699, -0.00000847, -0.00001981, -0.00002862, -0.00002222, 0.00000039, 0.00002541, 0.00003359, 0.00001546, -0.00001940, -0.00004718, -0.00004643, -0.00001467, 0.00002776, 0.00005058, 0.00003501, -0.00001163, -0.00005915, -0.00007450, -0.00004475, 0.00001281, 0.00006149, 0.00006951, 0.00003163, -0.00002724, -0.00006870, -0.00006647, -0.00002343, 0.00003123, 0.00006227, 0.00005174, 0.00000973, -0.00003415, -0.00005156, -0.00003396, 0.00000344, 0.00003386, 0.00003773, 0.00001505, -0.00001648, -0.00003499, -0.00002936, -0.00000561, 0.00001882, 0.00002767, 0.00001630, -0.00000657, -0.00002599, -0.00003085, -0.00002011, -0.00000186, 0.00001305, 0.00001775, 0.00001234, 0.00000209, -0.00000682, -0.00001090, -0.00001022, -0.00000704, -0.00000365, -0.00000121, 0.00000035, 0.00000160, 0.00000304, 0.00000477, 0.00000650, 0.00000755, 0.00000709, 0.00000458, 0.00000032, -0.00000426, -0.00000722, -0.00000718, -0.00000440, -0.00000083, 0.00000102, -0.00000029, -0.00000408, -0.00000800, -0.00000952, -0.00000770, -0.00000373, -0.00000014, 0.00000088, -0.00000125, -0.00000535, -0.00000933, -0.00001144, -0.00001102, -0.00000847, -0.00000484, -0.00000142, 0.00000057, 0.00000001, -0.00000397, },
		{ -0.00000396, -0.00000292, -0.00000332, -0.00000559, -0.00001046, -0.00001933, -0.00003373, -0.00005370, -0.00007569, -0.00009169, -0.00009132, -0.00006685, -0.00001867, 0.00004253, 0.00009957, 0.00013737, 0.00015152, 0.00015066, 0.00015026, 0.00016077, 0.00017822, 0.00018441, 0.00015814, 0.00009092, -0.00000355, -0.00009360, -0.00014616, -0.00014570, -0.00010312, -0.00004836, -0.00001152, -0.00000511, -0.00001914, -0.00003180, -0.00002745, -0.00000830, 0.00000862, 0.00000642, -0.00001640, -0.00004294, -0.00004992, -0.00002626, 0.00001637, 0.00004981, 0.00004914, 0.00001177, -0.00003861, -0.00006723, -0.00005270, -0.00000320, 0.00004771, 0.00006372, 0.00003136, -0.00002961, -0.00007867, -0.00008196, -0.00003555, 0.00003141, 0.00007604, 0.00007030, 0.00001933, -0.00004270, -0.00007581, -0.00006046, -0.00000920, 0.00004338, 0.00006440, 0.00004375, -0.00000132, -0.00003988, -0.00004854, -0.00002580, 0.00000944, 0.00003256, 0.00002999, 0.00000674, -0.00001920, -0.00003029, -0.00002096, 0.00000026, 0.00001809, 0.00002124, 0.00000911, -0.00000935, -0.00002259, -0.00002386, -0.00001431, -0.00000081, 0.00000918, 0.00001177, 0.00000777, 0.00000081, -0.00000534, -0.00000859, -0.00000878, -0.00000686, -0.00000405, -0.00000130, 0.00000084, 0.00000231, 0.00000342, 0.00000459, 0.00000589, 0.00000675, 0.00000622, 0.00000369, -0.00000035, -0.00000432, -0.00000637, -0.00000562, -0.00000294, -0.00000041, -0.00000002, -0.00000228, -0.00000590, -0.00000865, -0.00000889, -0.00000658, -0.00000321, -0.00000075, -0.00000046, -0.00000234, -0.00000532, -0.00000806, -0.00000956, -0.00000945, -0.00000802, -0.00000591, },
		{ -0.00000723, -0.00000698, -0.00000645, -0.00000640, -0.00000718, -0.00000836, -0.00000891, -0.00000836, -0.00000831, -0.00001280, -0.00002655, -0.00005124, -0.00008196, -0.00010686, -0.00011153, -0.00008655, -0.00003375, 0.00003348, 0.00009689, 0.00014350, 0.00017226, 0.00019212, 0.00021232, 0.00023167, 0.00023541, 0.00020372, 0.00012718, 0.00001864, -0.00008837, -0.00015580, -0.00016269, -0.00011801, -0.00005544, -0.00001253, -0.00000775, -0.00003073, -0.00005246, -0.00004796, -0.00001537, 0.00002265, 0.00003623, 0.00001180, -0.00003577, -0.00007216, -0.00006791, -0.00002082, 0.00004038, 0.00007475, 0.00005771, -0.00000074, -0.00006167, -0.00008345, -0.00005075, 0.00001431, 0.00006684, 0.00006990, 0.00002007, -0.00004992, -0.00009328, -0.00008080, -0.00002026, 0.00004923, 0.00008366, 0.00006307, 0.00000354, -0.00005496, -0.00007567, -0.00004842, 0.00000590, 0.00005121, 0.00006057, 0.00003238, -0.00001138, -0.00004180, -0.00004226, -0.00001734, 0.00001303, 0.00002854, 0.00002163, 0.00000041, -0.00001898, -0.00002411, -0.00001371, 0.00000325, 0.00001492, 0.00001457, 0.00000389, -0.00000952, -0.00001780, -0.00001745, -0.00001032, -0.00000130, 0.00000506, 0.00000670, 0.00000425, -0.00000021, -0.00000450, -0.00000712, -0.00000751, -0.00000594, -0.00000327, -0.00000059, 0.00000129, 0.00000228, 0.00000292, 0.00000387, 0.00000516, 0.00000602, 0.00000540, 0.00000288, -0.00000076, -0.00000389, -0.00000506, -0.00000405, -0.00000205, -0.00000086, -0.00000162, -0.00000410, -0.00000686, -0.00000834, -0.00000781, -0.00000575, -0.00000342, -0.00000204, -0.00000214, -0.00000346, -0.00000523, -0.00000666, },
		{ -0.00000382, -0.00000474, -0.00000546, -0.00000535, -0.00000435, -0.00000333, -0.00000377, -0.00000674, -0.00001163, -0.00001580, -0.00001601, -0.00001120, -0.00000486, -0.00000469, -0.00001859, -0.00004880, -0.00008792, -0.00012037, -0.00012936, -0.00010584, -0.00005334, 0.00001471, 0.00008357, 0.00014565, 0.00020199, 0.00025538, 0.00030032, 0.00031889, 0.00028822, 0.00019692, 0.00005951, -0.00008307, -0.00018072, -0.00020095, -0.00014842, -0.00006349, 0.00000123, 0.00001240, -0.00002471, -0.00007222, -0.00008720, -0.00005229, 0.00001183, 0.00006095, 0.00005908, 0.00000506, -0.00006500, -0.00010183, -0.00007730, -0.00000480, 0.00006885, 0.00009498, 0.00005639, -0.00002065, -0.00008421, -0.00009206, -0.00004023, 0.00003472, 0.00008110, 0.00006735, 0.00000297, -0.00006864, -0.00009999, -0.00007115, -0.00000194, 0.00006265, 0.00008283, 0.00004940, -0.00001237, -0.00006141, -0.00006826, -0.00003268, 0.00001903, 0.00005343, 0.00005173, 0.00001980, -0.00001879, -0.00003991, -0.00003407, -0.00000997, 0.00001394, 0.00002278, 0.00001395, -0.00000345, -0.00001661, -0.00001785, -0.00000850, 0.00000365, 0.00001056, 0.00000886, 0.00000093, -0.00000783, -0.00001277, -0.00001231, -0.00000796, -0.00000258, 0.00000134, 0.00000261, 0.00000138, -0.00000138, -0.00000427, -0.00000603, -0.00000596, -0.00000422, -0.00000176, 0.00000028, 0.00000130, 0.00000160, 0.00000198, 0.00000303, 0.00000447, 0.00000531, 0.00000458, 0.00000222, -0.00000079, -0.00000304, -0.00000364, -0.00000287, -0.00000189, -0.00000188, -0.00000322, -0.00000532, -0.00000707, -0.00000761, -0.00000685, -0.00000537, -0.00000402, -0.00000345, },
		{ -0.00000631, -0.00000521, -0.00000446, -0.00000459, -0.00000553, -0.00000647, -0.00000621, -0.00000402, -0.00000066, 0.00000146, -0.00000054, -0.00000770, -0.00001752, -0.00002444, -0.00002315, -0.00001303, -0.00000052, 0.00000307, -0.00001202, -0.00004699, -0.00009178, -0.00012903, -0.00014273, -0.00012580, -0.00008150, -0.00001798, 0.00005909, 0.00014882, 0.00024918, 0.00034704, 0.00041366, 0.00041318, 0.00032280, 0.00015324, -0.00004599, -0.00020446, -0.00026533, -0.00021694, -0.00010106, 0.00000984, 0.00005435, 0.00001752, -0.00006262, -0.00012206, -0.00011406, -0.00004020, 0.00005135, 0.00009915, 0.00007076, -0.00001473, -0.00009937, -0.00012549, -0.00007470, 0.00001993, 0.00009647, 0.00010593, 0.00004466, -0.00004438, -0.00010159, -0.00009092, -0.00002299, 0.00005385, 0.00008735, 0.00005633, -0.00001659, -0.00008197, -0.00009745, -0.00005506, 0.00001561, 0.00006898, 0.00007389, 0.00003224, -0.00002514, -0.00006083, -0.00005507, -0.00001618, 0.00002800, 0.00005001, 0.00003972, 0.00000814, -0.00002269, -0.00003500, -0.00002548, -0.00000459, 0.00001258, 0.00001647, 0.00000783, -0.00000505, -0.00001319, -0.00001259, -0.00000553, 0.00000233, 0.00000623, 0.00000479, -0.00000009, -0.00000527, -0.00000836, -0.00000862, -0.00000674, -0.00000402, -0.00000172, -0.00000072, -0.00000124, -0.00000278, -0.00000432, -0.00000484, -0.00000391, -0.00000199, -0.00000013, 0.00000082, 0.00000080, 0.00000056, 0.00000098, 0.00000231, 0.00000389, 0.00000463, 0.00000384, 0.00000178, -0.00000055, -0.00000212, -0.00000255, -0.00000230, -0.00000222, -0.00000291, -0.00000434, -0.00000593, -0.00000697, -0.00000707, },
		{ -0.00000485, -0.00000631, -0.00000719, -0.00000708, -0.00000605, -0.00000479, -0.00000436, -0.00000542, -0.00000750, -0.00000880, -0.00000724, -0.00000226, 0.00000390, 0.00000671, 0.00000216, -0.00000992, -0.00002428, -0.00003257, -0.00002861, -0.00001357, 0.00000300, 0.00000777, -0.00000847, -0.00004484, -0.00009030, -0.00012975, -0.00015141, -0.00014995, -0.00012332, -0.00006701, 0.00002686, 0.00016162, 0.00032273, 0.00046990, 0.00054547, 0.00050025, 0.00032492, 0.00006684, -0.00018281, -0.00033005, -0.00032649, -0.00019664, -0.00002546, 0.00008916, 0.00009167, -0.00000145, -0.00011612, -0.00016998, -0.00012529, -0.00001143, 0.00009664, 0.00012914, 0.00006744, -0.00004485, -0.00013219, -0.00013765, -0.00005966, 0.00004892, 0.00011759, 0.00010501, 0.00002462, -0.00006713, -0.00011013, -0.00008019, -0.00000263, 0.00006766, 0.00008423, 0.00003913, -0.00003457, -0.00008726, -0.00008635, -0.00003594, 0.00002887, 0.00006719, 0.00005889, 0.00001509, -0.00003241, -0.00005356, -0.00003869, -0.00000174, 0.00003172, 0.00004208, 0.00002677, -0.00000100, -0.00002310, -0.00002834, -0.00001778, -0.00000145, 0.00000973, 0.00001064, 0.00000358, -0.00000507, -0.00000979, -0.00000889, -0.00000439, 0.00000032, 0.00000278, 0.00000242, 0.00000014, -0.00000270, -0.00000499, -0.00000616, -0.00000612, -0.00000525, -0.00000419, -0.00000361, -0.00000374, -0.00000420, -0.00000423, -0.00000329, -0.00000152, 0.00000025, 0.00000111, 0.00000080, -0.00000008, -0.00000050, 0.00000022, 0.00000186, 0.00000348, 0.00000408, 0.00000326, 0.00000149, -0.00000036, -0.00000158, -0.00000205, -0.00000217, -0.00000251, -0.00000344, },
		{ -0.00000206, -0.00000201, -0.00000224, -0.00000325, -0.00000506, -0.00000703, -0.00000811, -0.00000756, -0.00000563, -0.00000374, -0.00000372, -0.00000627, -0.00000991, -0.00001139, -0.00000790, 0.00000027, 0.00000875, 0.00001110, 0.00000313, -0.00001332, -0.00003026, -0.00003791, -0.00003121, -0.00001391, 0.00000283, 0.00000684, -0.00000806, -0.00003948, -0.00007965, -0.00012094, -0.00015804, -0.00018390, -0.00018314, -0.00013112, -0.00000518, 0.00019481, 0.00042920, 0.00062065, 0.00068243, 0.00056322, 0.00028363, -0.00006053, -0.00033706, -0.00044153, -0.00035183, -0.00014017, 0.00006703, 0.00015975, 0.00010436, -0.00004280, -0.00017547, -0.00020565, -0.00011693, 0.00002970, 0.00013901, 0.00014439, 0.00004908, -0.00007951, -0.00015657, -0.00013535, -0.00003489, 0.00007634, 0.00012782, 0.00009237, 0.00000042, -0.00008420, -0.00010818, -0.00006230, 0.00001651, 0.00007331, 0.00007255, 0.00001941, -0.00004736, -0.00008371, -0.00006920, -0.00001761, 0.00003555, 0.00005810, 0.00004101, 0.00000106, -0.00003329, -0.00004138, -0.00002208, 0.00000863, 0.00003035, 0.00003152, 0.00001491, -0.00000683, -0.00002074, -0.00002125, -0.00001173, -0.00000026, 0.00000629, 0.00000588, 0.00000094, -0.00000438, -0.00000714, -0.00000674, -0.00000432, -0.00000149, 0.00000058, 0.00000139, 0.00000088, -0.00000065, -0.00000267, -0.00000452, -0.00000573, -0.00000622, -0.00000629, -0.00000623, -0.00000603, -0.00000530, -0.00000368, -0.00000136, 0.00000086, 0.00000203, 0.00000167, 0.00000025, -0.00000111, -0.00000134, -0.00000016, 0.00000179, 0.00000339, 0.00000378, 0.00000282, 0.00000106, -0.00000067, -0.00000174, },
		{ 0.00000229, 0.00000009, -0.00000180, -0.00000259, -0.00000218, -0.00000131, -0.00000116, -0.00000262, -0.00000552, -0.00000849, -0.00000965, -0.00000805, -0.00000466, -0.00000215, -0.00000304, -0.00000749, -0.00001247, -0.00001337, -0.00000743, 0.00000336, 0.00001253, 0.00001289, 0.00000169, -0.00001672, -0.00003295, -0.00003845, -0.00003116, -0.00001668, -0.00000428, -0.00000096, -0.00000862, -0.00002669, -0.00005688, -0.00010394, -0.00016824, -0.00023381, -0.00026237, -0.00020402, -0.00002528, 0.00025864, 0.00056989, 0.00078869, 0.00080648, 0.00058625, 0.00019464, -0.00021869, -0.00048964, -0.00052174, -0.00033554, -0.00005533, 0.00016096, 0.00020860, 0.00008936, -0.00009911, -0.00022902, -0.00022142, -0.00008991, 0.00007536, 0.00016999, 0.00014175, 0.00001946, -0.00011144, -0.00016733, -0.00011900, -0.00000553, 0.00009655, 0.00012519, 0.00007081, -0.00002289, -0.00009227, -0.00009651, -0.00004118, 0.00003054, 0.00006999, 0.00005502, 0.00000118, -0.00005276, -0.00007249, -0.00004956, -0.00000333, 0.00003507, 0.00004405, 0.00002369, -0.00000779, -0.00002843, -0.00002694, -0.00000787, 0.00001409, 0.00002508, 0.00002044, 0.00000557, -0.00000940, -0.00001672, -0.00001482, -0.00000756, -0.00000043, 0.00000300, 0.00000232, -0.00000065, -0.00000371, -0.00000550, -0.00000568, -0.00000454, -0.00000258, -0.00000046, 0.00000111, 0.00000155, 0.00000068, -0.00000117, -0.00000339, -0.00000544, -0.00000708, -0.00000818, -0.00000856, -0.00000782, -0.00000574, -0.00000259, 0.00000068, 0.00000282, 0.00000302, 0.00000148, -0.00000068, -0.00000206, -0.00000180, -0.00000004, 0.00000222, 0.00000375, 0.00000376, },
		{ 0.00000070, 0.00000329, 0.00000458, 0.00000382, 0.00000130, -0.00000175, -0.00000376, -0.00000371, -0.00000186, 0.00000020, 0.00000047, -0.00000213, -0.00000670, -0.00001061, -0.00001124, -0.00000799, -0.00000316, -0.00000063, -0.00000297, -0.00000906, -0.00001426, -0.00001353, -0.00000546, 0.00000593, 0.00001322, 0.00001066, -0.00000167, -0.00001770, -0.00002965, -0.00003337, -0.00003050, -0.00002552, -0.00002072, -0.00001451, -0.00000587, -0.00000215, -0.00002155, -0.00008367, -0.00018957, -0.00030415, -0.00035761, -0.00027442, -0.00002021, 0.00035965, 0.00073876, 0.00095688, 0.00089807, 0.00055851, 0.00006238, -0.00039019, -0.00061999, -0.00055854, -0.00027999, 0.00004391, 0.00024031, 0.00022746, 0.00005015, -0.00015914, -0.00026619, -0.00021427, -0.00004985, 0.00011632, 0.00018354, 0.00012225, -0.00001486, -0.00013387, -0.00016235, -0.00009221, 0.00002230, 0.00010560, 0.00011063, 0.00004496, -0.00004075, -0.00009025, -0.00007799, -0.00002119, 0.00003707, 0.00005897, 0.00003541, -0.00001228, -0.00005043, -0.00005637, -0.00003113, 0.00000500, 0.00002855, 0.00002821, 0.00000973, -0.00001093, -0.00001970, -0.00001296, 0.00000228, 0.00001494, 0.00001770, 0.00001059, -0.00000063, -0.00000936, -0.00001213, -0.00000965, -0.00000504, -0.00000131, 0.00000019, -0.00000029, -0.00000180, -0.00000346, -0.00000469, -0.00000510, -0.00000446, -0.00000284, -0.00000074, 0.00000107, 0.00000188, 0.00000141, -0.00000020, -0.00000257, -0.00000531, -0.00000795, -0.00000989, -0.00001038, -0.00000885, -0.00000542, -0.00000112, 0.00000248, 0.00000403, 0.00000315, 0.00000068, -0.00000179, -0.00000280, -0.00000178, },
		{ -0.00000053, -0.00000289, -0.00000319, -0.00000117, 0.00000215, 0.00000499, 0.00000568, 0.00000356, -0.00000051, -0.00000448, -0.00000612, -0.00000450, -0.00000077, 0.00000228, 0.00000194, -0.00000242, -0.00000858, -0.00001272, -0.00001208, -0.00000720, -0.00000180, -0.00000020, -0.00000393, -0.00001021, -0.00001377, -0.00001087, -0.00000251, 0.00000597, 0.00000876, 0.00000403, -0.00000497, -0.00001317, -0.00001854, -0.00002388, -0.00003296, -0.00004391, -0.00004671, -0.00002991, 0.00000568, 0.00003674, 0.00002297, -0.00006828, -0.00022948, -0.00039536, -0.00045971, -0.00032756, 0.00002155, 0.00049771, 0.00092197, 0.00110409, 0.00094039, 0.00047731, -0.00009987, -0.00055338, -0.00071044, -0.00054773, -0.00019541, 0.00014051, 0.00029222, 0.00021479, -0.00000430, -0.00021074, -0.00028011, -0.00018651, -0.00000537, 0.00014437, 0.00017762, 0.00009070, -0.00004641, -0.00014228, -0.00014303, -0.00006066, 0.00004318, 0.00010225, 0.00008754, 0.00001994, -0.00005028, -0.00007941, -0.00005674, -0.00000590, 0.00003573, 0.00004316, 0.00001751, -0.00001920, -0.00004187, -0.00003882, -0.00001674, 0.00000719, 0.00001840, 0.00001367, 0.00000073, -0.00000919, -0.00000952, -0.00000163, 0.00000782, 0.00001240, 0.00001000, 0.00000311, -0.00000383, -0.00000764, -0.00000782, -0.00000589, -0.00000374, -0.00000245, -0.00000210, -0.00000236, -0.00000294, -0.00000366, -0.00000429, -0.00000447, -0.00000387, -0.00000249, -0.00000069, 0.00000093, 0.00000184, 0.00000172, 0.00000040, -0.00000208, -0.00000543, -0.00000887, -0.00001126, -0.00001144, -0.00000896, -0.00000448, 0.00000038, 0.00000372, 0.00000438, 0.00000251, },
		{ 0.00000157, 0.00000422, 0.00000392, 0.00000129, -0.00000195, -0.00000383, -0.00000308, 0.00000015, 0.00000432, 0.00000709, 0.00000658, 0.00000254, -0.00000321, -0.00000762, -0.00000810, -0.00000441, 0.00000095, 0.00000410, 0.00000237, -0.00000371, -0.00001050, -0.00001374, -0.00001162, -0.00000618, -0.00000181, -0.00000177, -0.00000559, -0.00000943, -0.00000945, -0.00000522, -0.00000030, 0.00000099, -0.00000212, -0.00000573, -0.00000501, -0.00000033, 0.00000030, -0.00001338, -0.00004297, -0.00007374, -0.00007941, -0.00004104, 0.00003092, 0.00008914, 0.00006927, -0.00006741, -0.00029272, -0.00050212, -0.00055447, -0.00034820, 0.00010628, 0.00066445, 0.00109950, 0.00120925, 0.00092338, 0.00034961, -0.00027214, -0.00068677, -0.00074999, -0.00049391, -0.00009755, 0.00021829, 0.00030979, 0.00017574, -0.00006193, -0.00024390, -0.00026921, -0.00014489, 0.00003418, 0.00015446, 0.00015447, 0.00005427, -0.00006875, -0.00013550, -0.00011375, -0.00003057, 0.00005371, 0.00008808, 0.00006082, 0.00000003, -0.00005087, -0.00006288, -0.00003688, 0.00000273, 0.00002808, 0.00002619, 0.00000414, -0.00001966, -0.00002981, -0.00002310, -0.00000779, 0.00000454, 0.00000746, 0.00000269, -0.00000313, -0.00000438, -0.00000021, 0.00000583, 0.00000929, 0.00000806, 0.00000342, -0.00000163, -0.00000466, -0.00000516, -0.00000426, -0.00000338, -0.00000323, -0.00000364, -0.00000409, -0.00000427, -0.00000419, -0.00000405, -0.00000387, -0.00000352, -0.00000286, -0.00000186, -0.00000063, 0.00000063, 0.00000159, 0.00000177, 0.00000067, -0.00000199, -0.00000583, -0.00000969, -0.00001203, -0.00001160, -0.00000827, -0.00000321, },
		{ -0.00001202, -0.00001089, -0.00000706, -0.00000200, 0.00000222, 0.00000398, 0.00000283, -0.00000026, -0.00000334, -0.00000437, -0.00000229, 0.00000223, 0.00000695, 0.00000902, 0.00000667, 0.00000054, -0.00000637, -0.00001026, -0.00000893, -0.00000341, 0.00000247, 0.00000455, 0.00000120, -0.00000548, -0.00001121, -0.00001268, -0.00000997, -0.00000618, -0.00000458, -0.00000562, -0.00000665, -0.00000477, -0.00000035, 0.00000228, -0.00000148, -0.00001114, -0.00001911, -0.00001560, 0.00000189, 0.00002213, 0.00002428, -0.00000692, -0.00006427, -0.00011398, -0.00011294, -0.00004103, 0.00007218, 0.00014978, 0.00010693, -0.00008969, -0.00037887, -0.00061300, -0.00062498, -0.00032428, 0.00023240, 0.00084359, 0.00124851, 0.00125580, 0.00084644, 0.00019113, -0.00043155, -0.00077348, -0.00073660, -0.00040938, -0.00000401, 0.00026550, 0.00029332, 0.00012059, -0.00011075, -0.00025310, -0.00023724, -0.00009860, 0.00006127, 0.00014577, 0.00011984, 0.00002043, -0.00007816, -0.00011582, -0.00008054, -0.00000702, 0.00005326, 0.00006685, 0.00003541, -0.00001236, -0.00004410, -0.00004461, -0.00002153, 0.00000473, 0.00001696, 0.00001134, -0.00000347, -0.00001528, -0.00001735, -0.00001142, -0.00000408, -0.00000071, -0.00000174, -0.00000371, -0.00000291, 0.00000137, 0.00000659, 0.00000931, 0.00000788, 0.00000343, -0.00000127, -0.00000388, -0.00000391, -0.00000257, -0.00000155, -0.00000183, -0.00000324, -0.00000491, -0.00000600, -0.00000614, -0.00000544, -0.00000430, -0.00000315, -0.00000226, -0.00000169, -0.00000126, -0.00000067, 0.00000024, 0.00000122, 0.00000159, 0.00000053, -0.00000232, -0.00000637, -0.00001017, },
		{ -0.00000006, -0.00000297, -0.00000681, -0.00001006, -0.00001120, -0.00000957, -0.00000569, -0.00000114, 0.00000224, 0.00000309, 0.00000130, -0.00000192, -0.00000441, -0.00000420, -0.00000063, 0.00000491, 0.00000945, 0.00000999, 0.00000548, -0.00000219, -0.00000903, -0.00001133, -0.00000819, -0.00000214, 0.00000249, 0.00000263, -0.00000141, -0.00000646, -0.00000926, -0.00000912, -0.00000820, -0.00000895, -0.00001104, -0.00001091, -0.00000501, 0.00000545, 0.00001315, 0.00000911, -0.00000895, -0.00003126, -0.00003979, -0.00002136, 0.00001855, 0.00005302, 0.00004841, -0.00000987, -0.00009822, -0.00016012, -0.00013940, -0.00002441, 0.00012755, 0.00020926, 0.00012472, -0.00013982, -0.00048107, -0.00071189, -0.00065515, -0.00025046, 0.00038942, 0.00101337, 0.00134792, 0.00123554, 0.00071908, 0.00002348, -0.00055705, -0.00080466, -0.00067732, -0.00031102, 0.00006974, 0.00027716, 0.00024974, 0.00006178, -0.00014176, -0.00023832, -0.00019193, -0.00005678, 0.00007190, 0.00012164, 0.00008120, -0.00000491, -0.00007430, -0.00008805, -0.00004939, 0.00000717, 0.00004380, 0.00004332, 0.00001508, -0.00001703, -0.00003301, -0.00002820, -0.00001208, 0.00000172, 0.00000551, 0.00000077, -0.00000576, -0.00000855, -0.00000706, -0.00000456, -0.00000422, -0.00000617, -0.00000765, -0.00000580, -0.00000029, 0.00000626, 0.00001021, 0.00000952, 0.00000501, -0.00000043, -0.00000390, -0.00000423, -0.00000235, -0.00000025, 0.00000037, -0.00000101, -0.00000367, -0.00000635, -0.00000793, -0.00000788, -0.00000640, -0.00000418, -0.00000214, -0.00000091, -0.00000062, -0.00000080, -0.00000077, -0.00000014, 0.00000079, 0.00000112, },
		{ -0.00000082, -0.00000046, 0.00000021, 0.00000027, -0.00000102, -0.00000368, -0.00000687, -0.00000927, -0.00000975, -0.00000795, -0.00000450, -0.00000079, 0.00000161, 0.00000166, -0.00000050, -0.00000338, -0.00000477, -0.00000298, 0.00000191, 0.00000765, 0.00001089, 0.00000921, 0.00000295, -0.00000477, -0.00000986, -0.00001001, -0.00000621, -0.00000195, -0.00000040, -0.00000198, -0.00000435, -0.00000491, -0.00000368, -0.00000373, -0.00000819, -0.00001605, -0.00002104, -0.00001563, 0.00000167, 0.00002185, 0.00002904, 0.00001186, -0.00002482, -0.00005808, -0.00005993, -0.00001876, 0.00004572, 0.00008830, 0.00006631, -0.00002619, -0.00014272, -0.00020458, -0.00015089, 0.00001082, 0.00019030, 0.00025563, 0.00011364, -0.00021639, -0.00058625, -0.00078097, -0.00063363, -0.00013032, 0.00055920, 0.00115063, 0.00138284, 0.00115063, 0.00055920, -0.00013032, -0.00063363, -0.00078097, -0.00058625, -0.00021639, 0.00011364, 0.00025563, 0.00019030, 0.00001082, -0.00015089, -0.00020458, -0.00014272, -0.00002619, 0.00006631, 0.00008830, 0.00004572, -0.00001876, -0.00005993, -0.00005808, -0.00002482, 0.00001186, 0.00002904, 0.00002185, 0.00000167, -0.00001563, -0.00002104, -0.00001605, -0.00000819, -0.00000373, -0.00000368, -0.00000491, -0.00000435, -0.00000198, -0.00000040, -0.00000195, -0.00000621, -0.00001001, -0.00000986, -0.00000477, 0.00000295, 0.00000921, 0.00001089, 0.00000765, 0.00000191, -0.00000298, -0.00000477, -0.00000338, -0.00000050, 0.00000166, 0.00000161, -0.00000079, -0.00000450, -0.00000795, -0.00000975, -0.00000927, -0.00000687, -0.00000368, -0.00000102, 0.00000027, 0.00000021, -0.00000046, },
		{ -0.00000006, 0.00000112, 0.00000079, -0.00000014, -0.00000077, -0.00000080, -0.00000062, -0.00000091, -0.00000214, -0.00000418, -0.00000640, -0.00000788, -0.00000793, -0.00000635, -0.00000367, -0.00000101, 0.00000037, -0.00000025, -0.00000235, -0.00000423, -0.00000390, -0.00000043, 0.00000501, 0.00000952, 0.00001021, 0.00000626, -0.00000029, -0.00000580, -0.00000765, -0.00000617, -0.00000422, -0.00000456, -0.00000706, -0.00000855, -0.00000576, 0.00000077, 0.00000551, 0.00000172, -0.00001208, -0.00002820, -0.00003301, -0.00001703, 0.00001508, 0.00004332, 0.00004380, 0.00000717, -0.00004939, -0.00008805, -0.00007430, -0.00000491, 0.00008120, 0.00012164, 0.00007190, -0.00005678, -0.00019193, -0.00023832, -0.00014176, 0.00006178, 0.00024974, 0.00027716, 0.00006974, -0.00031102, -0.00067732, -0.00080466, -0.00055705, 0.00002348, 0.00071908, 0.00123554, 0.00134792, 0.00101337, 0.00038942, -0.00025046, -0.00065515, -0.00071189, -0.00048107, -0.00013982, 0.00012472, 0.00020926, 0.00012755, -0.00002441, -0.00013940, -0.00016012, -0.00009822, -0.00000987, 0.00004841, 0.00005302, 0.00001855, -0.00002136, -0.00003979, -0.00003126, -0.00000895, 0.00000911, 0.00001315, 0.00000545, -0.00000501, -0.00001091, -0.00001104, -0.00000895, -0.00000820, -0.00000912, -0.00000926, -0.00000646, -0.00000141, 0.00000263, 0.00000249, -0.00000214, -0.00000819, -0.00001133, -0.00000903, -0.00000219, 0.00000548, 0.00000999, 0.00000945, 0.00000491, -0.00000063, -0.00000420, -0.00000441, -0.00000192, 0.00000130, 0.00000309, 0.00000224, -0.00000114, -0.00000569, -0.00000957, -0.00001120, -0.00001006, -0.00000681, -0.00000297, },
		{ -0.00001202, -0.00001017, -0.00000637, -0.00000232, 0.00000053, 0.00000159, 0.00000122, 0.00000024, -0.00000067, -0.00000126, -0.00000169, -0.00000226, -0.00000315, -0.00000430, -0.00000544, -0.00000614, -0.00000600, -0.00000491, -0.00000324, -0.00000183, -0.00000155, -0.00000257, -0.00000391, -0.00000388, -0.00000127, 0.00000343, 0.00000788, 0.00000931, 0.00000659, 0.00000137, -0.00000291, -0.00000371, -0.00000174, -0.00000071, -0.00000408, -0.00001142, -0.00001735, -0.00001528, -0.00000347, 0.00001134, 0.00001696, 0.00000473, -0.00002153, -0.00004461, -0.00004410, -0.00001236, 0.00003541, 0.00006685, 0.00005326, -0.00000702, -0.00008054, -0.00011582, -0.00007816, 0.00002043, 0.00011984, 0.00014577, 0.00006127, -0.00009860, -0.00023724, -0.00025310, -0.00011075, 0.00012059, 0.00029332, 0.00026550, -0.00000401, -0.00040938, -0.00073660, -0.00077348, -0.00043155, 0.00019113, 0.00084644, 0.00125580, 0.00124851, 0.00084359, 0.00023240, -0.00032428, -0.00062498, -0.00061300, -0.00037887, -0.00008969, 0.00010693, 0.00014978, 0.00007218, -0.00004103, -0.00011294, -0.00011398, -0.00006427, -0.00000692, 0.00002428, 0.00002213, 0.00000189, -0.00001560, -0.00001911, -0.00001114, -0.00000148, 0.00000228, -0.00000035, -0.00000477, -0.00000665, -0.00000562, -0.00000458, -0.00000618, -0.00000997, -0.00001268, -0.00001121, -0.00000548, 0.00000120, 0.00000455, 0.00000247, -0.00000341, -0.00000893, -0.00001026, -0.00000637, 0.00000054, 0.00000667, 0.00000902, 0.00000695, 0.00000223, -0.00000229, -0.00000437, -0.00000334, -0.00000026, 0.00000283, 0.00000398, 0.00000222, -0.00000200, -0.00000706, -0.00001089, },
		{ 0.00000157, -0.00000321, -0.00000827, -0.00001160, -0.00001203, -0.00000969, -0.00000583, -0.00000199, 0.00000067, 0.00000177, 0.00000159, 0.00000063, -0.00000063, -0.00000186, -0.00000286, -0.00000352, -0.00000387, -0.00000405, -0.00000419, -0.00000427, -0.00000409, -0.00000364, -0.00000323, -0.00000338, -0.00000426, -0.00000516, -0.00000466, -0.00000163, 0.00000342, 0.00000806, 0.00000929, 0.00000583, -0.00000021, -0.00000438, -0.00000313, 0.00000269, 0.00000746, 0.00000454, -0.00000779, -0.00002310, -0.00002981, -0.00001966, 0.00000414, 0.00002619, 0.00002808, 0.00000273, -0.00003688, -0.00006288, -0.00005087, 0.00000003, 0.00006082, 0.00008808, 0.00005371, -0.00003057, -0.00011375, -0.00013550, -0.00006875, 0.00005427, 0.00015447, 0.00015446, 0.00003418, -0.00014489, -0.00026921, -0.00024390, -0.00006193, 0.00017574, 0.00030979, 0.00021829, -0.00009755, -0.00049391, -0.00074999, -0.00068677, -0.00027214, 0.00034961, 0.00092338, 0.00120925, 0.00109950, 0.00066445, 0.00010628, -0.00034820, -0.00055447, -0.00050212, -0.00029272, -0.00006741, 0.00006927, 0.00008914, 0.00003092, -0.00004104, -0.00007941, -0.00007374, -0.00004297, -0.00001338, 0.00000030, -0.00000033, -0.00000501, -0.00000573, -0.00000212, 0.00000099, -0.00000030, -0.00000522, -0.00000945, -0.00000943, -0.00000559, -0.00000177, -0.00000181, -0.00000618, -0.00001162, -0.00001374, -0.00001050, -0.00000371, 0.00000237, 0.00000410, 0.00000095, -0.00000441, -0.00000810, -0.00000762, -0.00000321, 0.00000254, 0.00000658, 0.00000709, 0.00000432, 0.00000015, -0.00000308, -0.00000383, -0.00000195, 0.00000129, 0.00000392, 0.00000422, },
		{ -0.00000053, 0.00000251, 0.00000438, 0.00000372, 0.00000038, -0.00000448, -0.00000896, -0.00001144, -0.00001126, -0.00000887, -0.00000543, -0.00000208, 0.00000040, 0.00000172, 0.00000184, 0.00000093, -0.00000069, -0.00000249, -0.00000387, -0.00000447, -0.00000429, -0.00000366, -0.00000294, -0.00000236, -0.00000210, -0.00000245, -0.00000374, -0.00000589, -0.00000782, -0.00000764, -0.00000383, 0.00000311, 0.00001000, 0.00001240, 0.00000782, -0.00000163, -0.00000952, -0.00000919, 0.00000073, 0.00001367, 0.00001840, 0.00000719, -0.00001674, -0.00003882, -0.00004187, -0.00001920, 0.00001751, 0.00004316, 0.00003573, -0.00000590, -0.00005674, -0.00007941, -0.00005028, 0.00001994, 0.00008754, 0.00010225, 0.00004318, -0.00006066, -0.00014303, -0.00014228, -0.00004641, 0.00009070, 0.00017762, 0.00014437, -0.00000537, -0.00018651, -0.00028011, -0.00021074, -0.00000430, 0.00021479, 0.00029222, 0.00014051, -0.00019541, -0.00054773, -0.00071044, -0.00055338, -0.00009987, 0.00047731, 0.00094039, 0.00110409, 0.00092197, 0.00049771, 0.00002155, -0.00032756, -0.00045971, -0.00039536, -0.00022948, -0.00006828, 0.00002297, 0.00003674, 0.00000568, -0.00002991, -0.00004671, -0.00004391, -0.00003296, -0.00002388, -0.00001854, -0.00001317, -0.00000497, 0.00000403, 0.00000876, 0.00000597, -0.00000251, -0.00001087, -0.00001377, -0.00001021, -0.00000393, -0.00000020, -0.00000180, -0.00000720, -0.00001208, -0.00001272, -0.00000858, -0.00000242, 0.00000194, 0.00000228, -0.00000077, -0.00000450, -0.00000612, -0.00000448, -0.00000051, 0.00000356, 0.00000568, 0.00000499, 0.00000215, -0.00000117, -0.00000319, -0.00000289, },
		{ 0.00000070, -0.00000178, -0.00000280, -0.00000179, 0.00000068, 0.00000315, 0.00000403, 0.00000248, -0.00000112, -0.00000542, -0.00000885, -0.00001038, -0.00000989, -0.00000795, -0.00000531, -0.00000257, -0.00000020, 0.00000141, 0.00000188, 0.00000107, -0.00000074, -0.00000284, -0.00000446, -0.00000510, -0.00000469, -0.00000346, -0.00000180, -0.00000029, 0.00000019, -0.00000131, -0.00000504, -0.00000965, -0.00001213, -0.00000936, -0.00000063, 0.00001059, 0.00001770, 0.00001494, 0.00000228, -0.00001296, -0.00001970, -0.00001093, 0.00000973, 0.00002821, 0.00002855, 0.00000500, -0.00003113, -0.00005637, -0.00005043, -0.00001228, 0.00003541, 0.00005897, 0.00003707, -0.00002119, -0.00007799, -0.00009025, -0.00004075, 0.00004496, 0.00011063, 0.00010560, 0.00002230, -0.00009221, -0.00016235, -0.00013387, -0.00001486, 0.00012225, 0.00018354, 0.00011632, -0.00004985, -0.00021427, -0.00026619, -0.00015914, 0.00005015, 0.00022746, 0.00024031, 0.00004391, -0.00027999, -0.00055854, -0.00061999, -0.00039019, 0.00006238, 0.00055851, 0.00089807, 0.00095688, 0.00073876, 0.00035965, -0.00002021, -0.00027442, -0.00035761, -0.00030415, -0.00018957, -0.00008367, -0.00002155, -0.00000215, -0.00000587, -0.00001451, -0.00002072, -0.00002552, -0.00003050, -0.00003337, -0.00002965, -0.00001770, -0.00000167, 0.00001066, 0.00001322, 0.00000593, -0.00000546, -0.00001353, -0.00001426, -0.00000906, -0.00000297, -0.00000063, -0.00000316, -0.00000799, -0.00001124, -0.00001061, -0.00000670, -0.00000213, 0.00000047, 0.00000020, -0.00000186, -0.00000371, -0.00000376, -0.00000175, 0.00000130, 0.00000382, 0.00000458, 0.00000329, },
		{ 0.00000229, 0.00000376, 0.00000375, 0.00000222, -0.00000004, -0.00000180, -0.00000206, -0.00000068, 0.00000148, 0.00000302, 0.00000282, 0.00000068, -0.00000259, -0.00000574, -0.00000782, -0.00000856, -0.00000818, -0.00000708, -0.00000544, -0.00000339, -0.00000117, 0.00000068, 0.00000155, 0.00000111, -0.00000046, -0.00000258, -0.00000454, -0.00000568, -0.00000550, -0.00000371, -0.00000065, 0.00000232, 0.00000300, -0.00000043, -0.00000756, -0.00001482, -0.00001672, -0.00000940, 0.00000557, 0.00002044, 0.00002508, 0.00001409, -0.00000787, -0.00002694, -0.00002843, -0.00000779, 0.00002369, 0.00004405, 0.00003507, -0.00000333, -0.00004956, -0.00007249, -0.00005276, 0.00000118, 0.00005502, 0.00006999, 0.00003054, -0.00004118, -0.00009651, -0.00009227, -0.00002289, 0.00007081, 0.00012519, 0.00009655, -0.00000553, -0.00011900, -0.00016733, -0.00011144, 0.00001946, 0.00014175, 0.00016999, 0.00007536, -0.00008991, -0.00022142, -0.00022902, -0.00009911, 0.00008936, 0.00020860, 0.00016096, -0.00005533, -0.00033554, -0.00052174, -0.00048964, -0.00021869, 0.00019464, 0.00058625, 0.00080648, 0.00078869, 0.00056989, 0.00025864, -0.00002528, -0.00020402, -0.00026237, -0.00023381, -0.00016824, -0.00010394, -0.00005688, -0.00002669, -0.00000862, -0.00000096, -0.00000428, -0.00001668, -0.00003116, -0.00003845, -0.00003295, -0.00001672, 0.00000169, 0.00001289, 0.00001253, 0.00000336, -0.00000743, -0.00001337, -0.00001247, -0.00000749, -0.00000304, -0.00000215, -0.00000466, -0.00000805, -0.00000965, -0.00000849, -0.00000552, -0.00000262, -0.00000116, -0.00000131, -0.00000218, -0.00000259, -0.00000180, 0.00000009, },
		{ -0.00000206, -0.00000174, -0.00000067, 0.00000106, 0.00000282, 0.00000378, 0.00000339, 0.00000179, -0.00000016, -0.00000134, -0.00000111, 0.00000025, 0.00000167, 0.00000203, 0.00000086, -0.00000136, -0.00000368, -0.00000530, -0.00000603, -0.00000623, -0.00000629, -0.00000622, -0.00000573, -0.00000452, -0.00000267, -0.00000065, 0.00000088, 0.00000139, 0.00000058, -0.00000149, -0.00000432, -0.00000674, -0.00000714, -0.00000438, 0.00000094, 0.00000588, 0.00000629, -0.00000026, -0.00001173, -0.00002125, -0.00002074, -0.00000683, 0.00001491, 0.00003152, 0.00003035, 0.00000863, -0.00002208, -0.00004138, -0.00003329, 0.00000106, 0.00004101, 0.00005810, 0.00003555, -0.00001761, -0.00006920, -0.00008371, -0.00004736, 0.00001941, 0.00007255, 0.00007331, 0.00001651, -0.00006230, -0.00010818, -0.00008420, 0.00000042, 0.00009237, 0.00012782, 0.00007634, -0.00003489, -0.00013535, -0.00015657, -0.00007951, 0.00004908, 0.00014439, 0.00013901, 0.00002970, -0.00011693, -0.00020565, -0.00017547, -0.00004280, 0.00010436, 0.00015975, 0.00006703, -0.00014017, -0.00035183, -0.00044153, -0.00033706, -0.00006053, 0.00028363, 0.00056322, 0.00068243, 0.00062065, 0.00042920, 0.00019481, -0.00000518, -0.00013112, -0.00018314, -0.00018390, -0.00015804, -0.00012094, -0.00007965, -0.00003948, -0.00000806, 0.00000684, 0.00000283, -0.00001391, -0.00003121, -0.00003791, -0.00003026, -0.00001332, 0.00000313, 0.00001110, 0.00000875, 0.00000027, -0.00000790, -0.00001139, -0.00000991, -0.00000627, -0.00000372, -0.00000374, -0.00000563, -0.00000756, -0.00000811, -0.00000703, -0.00000506, -0.00000325, -0.00000224, -0.00000201, },
		{ -0.00000485, -0.00000344, -0.00000251, -0.00000217, -0.00000205, -0.00000158, -0.00000036, 0.00000149, 0.00000326, 0.00000408, 0.00000348, 0.00000186, 0.00000022, -0.00000050, -0.00000008, 0.00000080, 0.00000111, 0.00000025, -0.00000152, -0.00000329, -0.00000423, -0.00000420, -0.00000374, -0.00000361, -0.00000419, -0.00000525, -0.00000612, -0.00000616, -0.00000499, -0.00000270, 0.00000014, 0.00000242, 0.00000278, 0.00000032, -0.00000439, -0.00000889, -0.00000979, -0.00000507, 0.00000358, 0.00001064, 0.00000973, -0.00000145, -0.00001778, -0.00002834, -0.00002310, -0.00000100, 0.00002677, 0.00004208, 0.00003172, -0.00000174, -0.00003869, -0.00005356, -0.00003241, 0.00001509, 0.00005889, 0.00006719, 0.00002887, -0.00003594, -0.00008635, -0.00008726, -0.00003457, 0.00003913, 0.00008423, 0.00006766, -0.00000263, -0.00008019, -0.00011013, -0.00006713, 0.00002462, 0.00010501, 0.00011759, 0.00004892, -0.00005966, -0.00013765, -0.00013219, -0.00004485, 0.00006744, 0.00012914, 0.00009664, -0.00001143, -0.00012529, -0.00016998, -0.00011612, -0.00000145, 0.00009167, 0.00008916, -0.00002546, -0.00019664, -0.00032649, -0.00033005, -0.00018281, 0.00006684, 0.00032492, 0.00050025, 0.00054547, 0.00046990, 0.00032273, 0.00016162, 0.00002686, -0.00006701, -0.00012332, -0.00014995, -0.00015141, -0.00012975, -0.00009030, -0.00004484, -0.00000847, 0.00000777, 0.00000300, -0.00001357, -0.00002861, -0.00003257, -0.00002428, -0.00000992, 0.00000216, 0.00000671, 0.00000390, -0.00000226, -0.00000724, -0.00000880, -0.00000750, -0.00000542, -0.00000436, -0.00000479, -0.00000605, -0.00000708, -0.00000719, -0.00000631, },
		{ -0.00000631, -0.00000707, -0.00000697, -0.00000593, -0.00000434, -0.00000291, -0.00000222, -0.00000230, -0.00000255, -0.00000212, -0.00000055, 0.00000178, 0.00000384, 0.00000463, 0.00000389, 0.00000231, 0.00000098, 0.00000056, 0.00000080, 0.00000082, -0.00000013, -0.00000199, -0.00000391, -0.00000484, -0.00000432, -0.00000278, -0.00000124, -0.00000072, -0.00000172, -0.00000402, -0.00000674, -0.00000862, -0.00000836, -0.00000527, -0.00000009, 0.00000479, 0.00000623, 0.00000233, -0.00000553, -0.00001259, -0.00001319, -0.00000505, 0.00000783, 0.00001647, 0.00001258, -0.00000459, -0.00002548, -0.00003500, -0.00002269, 0.00000814, 0.00003972, 0.00005001, 0.00002800, -0.00001618, -0.00005507, -0.00006083, -0.00002514, 0.00003224, 0.00007389, 0.00006898, 0.00001561, -0.00005506, -0.00009745, -0.00008197, -0.00001659, 0.00005633, 0.00008735, 0.00005385, -0.00002299, -0.00009092, -0.00010159, -0.00004438, 0.00004466, 0.00010593, 0.00009647, 0.00001993, -0.00007470, -0.00012549, -0.00009937, -0.00001473, 0.00007076, 0.00009915, 0.00005135, -0.00004020, -0.00011406, -0.00012206, -0.00006262, 0.00001752, 0.00005435, 0.00000984, -0.00010106, -0.00021694, -0.00026533, -0.00020446, -0.00004599, 0.00015324, 0.00032280, 0.00041318, 0.00041366, 0.00034704, 0.00024918, 0.00014882, 0.00005909, -0.00001798, -0.00008150, -0.00012580, -0.00014273, -0.00012903, -0.00009178, -0.00004699, -0.00001202, 0.00000307, -0.00000052, -0.00001303, -0.00002315, -0.00002444, -0.00001752, -0.00000770, -0.00000054, 0.00000146, -0.00000066, -0.00000402, -0.00000621, -0.00000647, -0.00000553, -0.00000459, -0.00000446, -0.00000521, },
		{ -0.00000382, -0.00000345, -0.00000402, -0.00000537, -0.00000685, -0.00000761, -0.00000707, -0.00000532, -0.00000322, -0.00000188, -0.00000189, -0.00000287, -0.00000364, -0.00000304, -0.00000079, 0.00000222, 0.00000458, 0.00000531, 0.00000447, 0.00000303, 0.00000198, 0.00000160, 0.00000130, 0.00000028, -0.00000176, -0.00000422, -0.00000596, -0.00000603, -0.00000427, -0.00000138, 0.00000138, 0.00000261, 0.00000134, -0.00000258, -0.00000796, -0.00001231, -0.00001277, -0.00000783, 0.00000093, 0.00000886, 0.00001056, 0.00000365, -0.00000850, -0.00001785, -0.00001661, -0.00000345, 0.00001395, 0.00002278, 0.00001394, -0.00000997, -0.00003407, -0.00003991, -0.00001879, 0.00001980, 0.00005173, 0.00005343, 0.00001903, -0.00003268, -0.00006826, -0.00006141, -0.00001237, 0.00004940, 0.00008283, 0.00006265, -0.00000194, -0.00007115, -0.00009999, -0.00006864, 0.00000297, 0.00006735, 0.00008110, 0.00003472, -0.00004023, -0.00009206, -0.00008421, -0.00002065, 0.00005639, 0.00009498, 0.00006885, -0.00000480, -0.00007730, -0.00010183, -0.00006500, 0.00000506, 0.00005908, 0.00006095, 0.00001183, -0.00005229, -0.00008720, -0.00007222, -0.00002471, 0.00001240, 0.00000123, -0.00006349, -0.00014842, -0.00020095, -0.00018072, -0.00008307, 0.00005951, 0.00019692, 0.00028822, 0.00031889, 0.00030032, 0.00025538, 0.00020199, 0.00014565, 0.00008357, 0.00001471, -0.00005334, -0.00010584, -0.00012936, -0.00012037, -0.00008792, -0.00004880, -0.00001859, -0.00000469, -0.00000486, -0.00001120, -0.00001601, -0.00001580, -0.00001163, -0.00000674, -0.00000377, -0.00000333, -0.00000435, -0.00000535, -0.00000546, -0.00000474, },
		{ -0.00000723, -0.00000666, -0.00000523, -0.00000346, -0.00000214, -0.00000204, -0.00000342, -0.00000575, -0.00000781, -0.00000834, -0.00000686, -0.00000410, -0.00000162, -0.00000086, -0.00000205, -0.00000405, -0.00000506, -0.00000389, -0.00000076, 0.00000288, 0.00000540, 0.00000602, 0.00000516, 0.00000387, 0.00000292, 0.00000228, 0.00000129, -0.00000059, -0.00000327, -0.00000594, -0.00000751, -0.00000712, -0.00000450, -0.00000021, 0.00000425, 0.00000670, 0.00000506, -0.00000130, -0.00001032, -0.00001745, -0.00001780, -0.00000952, 0.00000389, 0.00001457, 0.00001492, 0.00000325, -0.00001371, -0.00002411, -0.00001898, 0.00000041, 0.00002163, 0.00002854, 0.00001303, -0.00001734, -0.00004226, -0.00004180, -0.00001138, 0.00003238, 0.00006057, 0.00005121, 0.00000590, -0.00004842, -0.00007567, -0.00005496, 0.00000354, 0.00006307, 0.00008366, 0.00004923, -0.00002026, -0.00008080, -0.00009328, -0.00004992, 0.00002007, 0.00006990, 0.00006684, 0.00001431, -0.00005075, -0.00008345, -0.00006167, -0.00000074, 0.00005771, 0.00007475, 0.00004038, -0.00002082, -0.00006791, -0.00007216, -0.00003577, 0.00001180, 0.00003623, 0.00002265, -0.00001537, -0.00004796, -0.00005246, -0.00003073, -0.00000775, -0.00001253, -0.00005544, -0.00011801, -0.00016269, -0.00015580, -0.00008837, 0.00001864, 0.00012718, 0.00020372, 0.00023541, 0.00023167, 0.00021232, 0.00019212, 0.00017226, 0.00014350, 0.00009689, 0.00003348, -0.00003375, -0.00008655, -0.00011153, -0.00010686, -0.00008196, -0.00005124, -0.00002655, -0.00001280, -0.00000831, -0.00000836, -0.00000891, -0.00000836, -0.00000718, -0.00000640, -0.00000645, -0.00000698, },
		{ -0.00000396, -0.00000591, -0.00000802, -0.00000945, -0.00000956, -0.00000806, -0.00000532, -0.00000234, -0.00000046, -0.00000075, -0.00000321, -0.00000658, -0.00000889, -0.00000865, -0.00000590, -0.00000228, -0.00000002, -0.00000041, -0.00000294, -0.00000562, -0.00000637, -0.00000432, -0.00000035, 0.00000369, 0.00000622, 0.00000675, 0.00000589, 0.00000459, 0.00000342, 0.00000231, 0.00000084, -0.00000130, -0.00000405, -0.00000686, -0.00000878, -0.00000859, -0.00000534, 0.00000081, 0.00000777, 0.00001177, 0.00000918, -0.00000081, -0.00001431, -0.00002386, -0.00002259, -0.00000935, 0.00000911, 0.00002124, 0.00001809, 0.00000026, -0.00002096, -0.00003029, -0.00001920, 0.00000674, 0.00002999, 0.00003256, 0.00000944, -0.00002580, -0.00004854, -0.00003988, -0.00000132, 0.00004375, 0.00006440, 0.00004338, -0.00000920, -0.00006046, -0.00007581, -0.00004270, 0.00001933, 0.00007030, 0.00007604, 0.00003141, -0.00003555, -0.00008196, -0.00007867, -0.00002961, 0.00003136, 0.00006372, 0.00004771, -0.00000320, -0.00005270, -0.00006723, -0.00003861, 0.00001177, 0.00004914, 0.00004981, 0.00001637, -0.00002626, -0.00004992, -0.00004294, -0.00001640, 0.00000642, 0.00000862, -0.00000830, -0.00002745, -0.00003180, -0.00001914, -0.00000511, -0.00001152, -0.00004836, -0.00010312, -0.00014570, -0.00014616, -0.00009360, -0.00000355, 0.00009092, 0.00015814, 0.00018441, 0.00017822, 0.00016077, 0.00015026, 0.00015066, 0.00015152, 0.00013737, 0.00009957, 0.00004253, -0.00001867, -0.00006685, -0.00009132, -0.00009169, -0.00007569, -0.00005370, -0.00003373, -0.00001933, -0.00001046, -0.00000559, -0.00000332, -0.00000292, },
		{ -0.00001183, -0.00000397, 0.00000001, 0.00000057, -0.00000142, -0.00000484, -0.00000847, -0.00001102, -0.00001144, -0.00000933, -0.00000535, -0.00000125, 0.00000088, -0.00000014, -0.00000373, -0.00000770, -0.00000952, -0.00000800, -0.00000408, -0.00000029, 0.00000102, -0.00000083, -0.00000440, -0.00000718, -0.00000722, -0.00000426, 0.00000032, 0.00000458, 0.00000709, 0.00000755, 0.00000650, 0.00000477, 0.00000304, 0.00000160, 0.00000035, -0.00000121, -0.00000365, -0.00000704, -0.00001022, -0.00001090, -0.00000682, 0.00000209, 0.00001234, 0.00001775, 0.00001305, -0.00000186, -0.00002011, -0.00003085, -0.00002599, -0.00000657, 0.00001630, 0.00002767, 0.00001882, -0.00000561, -0.00002936, -0.00003499, -0.00001648, 0.00001505, 0.00003773, 0.00003386, 0.00000344, -0.00003396, -0.00005156, -0.00003415, 0.00000973, 0.00005174, 0.00006227, 0.00003123, -0.00002343, -0.00006647, -0.00006870, -0.00002724, 0.00003163, 0.00006951, 0.00006149, 0.00001281, -0.00004475, -0.00007450, -0.00005915, -0.00001163, 0.00003501, 0.00005058, 0.00002776, -0.00001467, -0.00004643, -0.00004718, -0.00001940, 0.00001546, 0.00003359, 0.00002541, 0.00000039, -0.00002222, -0.00002862, -0.00001981, -0.00000847, -0.00000699, -0.00001668, -0.00002710, -0.00002593, -0.00001114, 0.00000456, 0.00000175, -0.00003064, -0.00008412, -0.00013220, -0.00014493, -0.00010706, -0.00002817, 0.00006219, 0.00013100, 0.00015953, 0.00015145, 0.00012723, 0.00010967, 0.00011005, 0.00012334, 0.00013375, 0.00012600, 0.00009457, 0.00004596, -0.00000598, -0.00004763, -0.00007140, -0.00007704, -0.00006939, -0.00005478, -0.00003840, -0.00002353, },
		{ -0.00006238, -0.00005319, -0.00003985, -0.00002569, -0.00001325, -0.00000411, 0.00000115, 0.00000262, 0.00000083, -0.00000326, -0.00000816, -0.00001189, -0.00001271, -0.00001003, -0.00000506, -0.00000040, 0.00000136, -0.00000063, -0.00000492, -0.00000859, -0.00000916, -0.00000625, -0.00000181, 0.00000128, 0.00000112, -0.00000202, -0.00000602, -0.00000834, -0.00000753, -0.00000386, 0.00000113, 0.00000560, 0.00000819, 0.00000843, 0.00000666, 0.00000391, 0.00000149, 0.00000039, 0.00000046, 0.00000030, -0.00000197, -0.00000692, -0.00001245, -0.00001425, -0.00000857, 0.00000424, 0.00001817, 0.00002412, 0.00001575, -0.00000505, -0.00002738, -0.00003726, -0.00002684, -0.00000096, 0.00002449, 0.00003236, 0.00001625, -0.00001392, -0.00003748, -0.00003690, -0.00001070, 0.00002424, 0.00004343, 0.00003197, -0.00000401, -0.00004023, -0.00005054, -0.00002552, 0.00001968, 0.00005468, 0.00005446, 0.00001706, -0.00003405, -0.00006537, -0.00005597, -0.00001188, 0.00003788, 0.00006090, 0.00004299, -0.00000290, -0.00004630, -0.00006018, -0.00003858, 0.00000102, 0.00003121, 0.00003375, 0.00001081, -0.00001878, -0.00003431, -0.00002760, -0.00000689, 0.00001134, 0.00001539, 0.00000597, -0.00000657, -0.00001218, -0.00000953, -0.00000608, -0.00001025, -0.00002260, -0.00003398, -0.00003278, -0.00001578, 0.00000623, 0.00001377, -0.00000791, -0.00005636, -0.00010983, -0.00013817, -0.00012023, -0.00005697, 0.00002841, 0.00010312, 0.00014247, 0.00014172, 0.00011581, 0.00008821, 0.00007687, 0.00008538, 0.00010367, 0.00011608, 0.00011082, 0.00008549, 0.00004653, 0.00000460, -0.00003069, -0.00005382, -0.00006373, },
		{ 0.00001223, -0.00001756, -0.00003947, -0.00005145, -0.00005377, -0.00004843, -0.00003839, -0.00002658, -0.00001527, -0.00000587, 0.00000071, 0.00000374, 0.00000280, -0.00000156, -0.00000744, -0.00001203, -0.00001299, -0.00000986, -0.00000455, -0.00000026, 0.00000055, -0.00000220, -0.00000628, -0.00000864, -0.00000757, -0.00000377, 0.00000027, 0.00000197, 0.00000037, -0.00000348, -0.00000729, -0.00000895, -0.00000750, -0.00000336, 0.00000208, 0.00000699, 0.00000968, 0.00000918, 0.00000585, 0.00000155, -0.00000117, -0.00000076, 0.00000191, 0.00000353, 0.00000065, -0.00000721, -0.00001589, -0.00001840, -0.00000985, 0.00000781, 0.00002506, 0.00002995, 0.00001631, -0.00001049, -0.00003517, -0.00004169, -0.00002436, 0.00000697, 0.00003217, 0.00003395, 0.00001020, -0.00002349, -0.00004367, -0.00003523, -0.00000250, 0.00003275, 0.00004597, 0.00002715, -0.00001148, -0.00004329, -0.00004555, -0.00001565, 0.00002652, 0.00005193, 0.00004247, 0.00000363, -0.00003909, -0.00005764, -0.00004041, 0.00000024, 0.00003699, 0.00004639, 0.00002419, -0.00001305, -0.00004051, -0.00004217, -0.00002057, 0.00000701, 0.00002195, 0.00001699, -0.00000057, -0.00001626, -0.00001990, -0.00001212, -0.00000183, 0.00000233, -0.00000101, -0.00000602, -0.00000611, -0.00000081, 0.00000318, -0.00000204, -0.00001757, -0.00003476, -0.00004057, -0.00002804, -0.00000371, 0.00001441, 0.00000808, -0.00002782, -0.00007925, -0.00011908, -0.00012226, -0.00008052, -0.00000758, 0.00006828, 0.00011954, 0.00013317, 0.00011526, 0.00008488, 0.00006214, 0.00005782, 0.00006996, 0.00008763, 0.00009837, 0.00009444, 0.00007515, 0.00004535, },
		{ 0.00008200, 0.00007921, 0.00006495, 0.00004225, 0.00001595, -0.00000885, -0.00002804, -0.00003947, -0.00004320, -0.00004095, -0.00003495, -0.00002691, -0.00001783, -0.00000859, -0.00000062, 0.00000413, 0.00000419, -0.00000012, -0.00000640, -0.00001118, -0.00001199, -0.00000889, -0.00000428, -0.00000128, -0.00000156, -0.00000433, -0.00000709, -0.00000752, -0.00000510, -0.00000130, 0.00000154, 0.00000176, -0.00000072, -0.00000460, -0.00000798, -0.00000925, -0.00000754, -0.00000292, 0.00000338, 0.00000907, 0.00001154, 0.00000934, 0.00000350, -0.00000247, -0.00000449, -0.00000099, 0.00000524, 0.00000831, 0.00000342, -0.00000865, -0.00002060, -0.00002265, -0.00000981, 0.00001299, 0.00003224, 0.00003404, 0.00001408, -0.00001769, -0.00004208, -0.00004289, -0.00001847, 0.00001597, 0.00003767, 0.00003161, 0.00000137, -0.00003260, -0.00004654, -0.00002994, 0.00000679, 0.00003904, 0.00004488, 0.00002041, -0.00001741, -0.00004244, -0.00003755, -0.00000651, 0.00002888, 0.00004405, 0.00002867, -0.00000657, -0.00003792, -0.00004520, -0.00002525, 0.00000696, 0.00002959, 0.00002908, 0.00000844, -0.00001654, -0.00002938, -0.00002408, -0.00000764, 0.00000684, 0.00001034, 0.00000353, -0.00000563, -0.00000950, -0.00000686, -0.00000274, -0.00000284, -0.00000779, -0.00001237, -0.00001053, -0.00000160, 0.00000765, 0.00000782, -0.00000520, -0.00002542, -0.00003973, -0.00003720, -0.00001813, 0.00000415, 0.00001106, -0.00000862, -0.00004974, -0.00009189, -0.00011041, -0.00009059, -0.00003632, 0.00003208, 0.00008898, 0.00011656, 0.00011245, 0.00008836, 0.00006178, 0.00004652, 0.00004715, 0.00005925, 0.00007373, },
		{ 0.00003733, 0.00004044, 0.00005092, 0.00006223, 0.00006852, 0.00006630, 0.00005496, 0.00003667, 0.00001551, -0.00000401, -0.00001862, -0.00002730, -0.00003112, -0.00003187, -0.00003054, -0.00002692, -0.00002041, -0.00001150, -0.00000242, 0.00000369, 0.00000468, 0.00000090, -0.00000483, -0.00000906, -0.00000980, -0.00000760, -0.00000479, -0.00000356, -0.00000445, -0.00000616, -0.00000683, -0.00000545, -0.00000255, 0.00000035, 0.00000179, 0.00000112, -0.00000141, -0.00000494, -0.00000823, -0.00000973, -0.00000795, -0.00000240, 0.00000542, 0.00001200, 0.00001345, 0.00000829, -0.00000080, -0.00000788, -0.00000761, 0.00000048, 0.00001056, 0.00001391, 0.00000537, -0.00001167, -0.00002607, -0.00002596, -0.00000779, 0.00001944, 0.00003855, 0.00003531, 0.00000902, -0.00002548, -0.00004656, -0.00004016, -0.00000993, 0.00002431, 0.00003958, 0.00002537, -0.00000878, -0.00003946, -0.00004533, -0.00002186, 0.00001550, 0.00004198, 0.00004049, 0.00001319, -0.00002054, -0.00003786, -0.00002819, 0.00000010, 0.00002641, 0.00003270, 0.00001567, -0.00001196, -0.00003137, -0.00003089, -0.00001329, 0.00000764, 0.00001779, 0.00001239, -0.00000205, -0.00001400, -0.00001594, -0.00000902, -0.00000061, 0.00000253, -0.00000044, -0.00000481, -0.00000541, -0.00000159, 0.00000216, 0.00000055, -0.00000716, -0.00001563, -0.00001744, -0.00000958, 0.00000309, 0.00001061, 0.00000535, -0.00001147, -0.00002963, -0.00003673, -0.00002755, -0.00000888, 0.00000408, -0.00000220, -0.00002970, -0.00006601, -0.00009073, -0.00008714, -0.00005192, 0.00000306, 0.00005741, 0.00009288, 0.00010174, 0.00008856, 0.00006569, 0.00004603, },
		{ 0.00008261, 0.00006693, 0.00004839, 0.00003508, 0.00003084, 0.00003519, 0.00004455, 0.00005397, 0.00005867, 0.00005561, 0.00004461, 0.00002845, 0.00001166, -0.00000178, -0.00001030, -0.00001509, -0.00001864, -0.00002245, -0.00002581, -0.00002637, -0.00002229, -0.00001405, -0.00000461, 0.00000222, 0.00000418, 0.00000177, -0.00000245, -0.00000577, -0.00000699, -0.00000675, -0.00000639, -0.00000664, -0.00000711, -0.00000690, -0.00000547, -0.00000320, -0.00000092, 0.00000069, 0.00000134, 0.00000087, -0.00000104, -0.00000450, -0.00000856, -0.00001087, -0.00000876, -0.00000139, 0.00000859, 0.00001565, 0.00001477, 0.00000539, -0.00000703, -0.00001390, -0.00000948, 0.00000415, 0.00001738, 0.00001922, 0.00000563, -0.00001617, -0.00003132, -0.00002731, -0.00000362, 0.00002623, 0.00004265, 0.00003316, 0.00000185, -0.00003227, -0.00004738, -0.00003362, -0.00000025, 0.00003023, 0.00003725, 0.00001620, -0.00001835, -0.00004269, -0.00004019, -0.00001250, 0.00002202, 0.00004115, 0.00003386, 0.00000702, -0.00002031, -0.00003058, -0.00001933, 0.00000305, 0.00001987, 0.00002021, 0.00000566, -0.00001227, -0.00002146, -0.00001766, -0.00000626, 0.00000317, 0.00000454, -0.00000084, -0.00000663, -0.00000745, -0.00000330, 0.00000118, 0.00000129, -0.00000322, -0.00000796, -0.00000790, -0.00000216, 0.00000467, 0.00000612, -0.00000061, -0.00001171, -0.00001920, -0.00001706, -0.00000634, 0.00000502, 0.00000784, -0.00000122, -0.00001675, -0.00002824, -0.00002784, -0.00001668, -0.00000473, -0.00000418, -0.00002056, -0.00004800, -0.00007176, -0.00007667, -0.00005596, -0.00001503, 0.00003168, 0.00006840, 0.00008549, },
		{ 0.00001361, 0.00004777, 0.00006861, 0.00007319, 0.00006471, 0.00004974, 0.00003517, 0.00002608, 0.00002480, 0.00003073, 0.00004058, 0.00004928, 0.00005197, 0.00004622, 0.00003346, 0.00001827, 0.00000591, -0.00000078, -0.00000275, -0.00000369, -0.00000708, -0.00001364, -0.00002085, -0.00002479, -0.00002301, -0.00001613, -0.00000736, -0.00000029, 0.00000307, 0.00000294, 0.00000083, -0.00000186, -0.00000449, -0.00000693, -0.00000882, -0.00000949, -0.00000849, -0.00000617, -0.00000364, -0.00000189, -0.00000098, -0.00000017, 0.00000104, 0.00000187, 0.00000059, -0.00000377, -0.00000963, -0.00001292, -0.00000961, 0.00000067, 0.00001298, 0.00001946, 0.00001466, 0.00000029, -0.00001463, -0.00001934, -0.00000920, 0.00000994, 0.00002466, 0.00002303, 0.00000378, -0.00002145, -0.00003514, -0.00002601, 0.00000221, 0.00003204, 0.00004347, 0.00002772, -0.00000609, -0.00003652, -0.00004402, -0.00002432, 0.00000870, 0.00003242, 0.00003094, 0.00000579, -0.00002551, -0.00004168, -0.00003215, -0.00000362, 0.00002529, 0.00003701, 0.00002645, 0.00000305, -0.00001701, -0.00002220, -0.00001255, 0.00000223, 0.00001098, 0.00000894, -0.00000016, -0.00000851, -0.00001084, -0.00000787, -0.00000441, -0.00000435, -0.00000722, -0.00000899, -0.00000620, 0.00000043, 0.00000618, 0.00000623, -0.00000000, -0.00000792, -0.00001124, -0.00000703, 0.00000158, 0.00000763, 0.00000570, -0.00000368, -0.00001417, -0.00001837, -0.00001354, -0.00000379, 0.00000302, 0.00000124, -0.00000815, -0.00001833, -0.00002194, -0.00001709, -0.00000922, -0.00000741, -0.00001767, -0.00003787, -0.00005805, -0.00006598, -0.00005426, -0.00002453, },
		{ -0.00005756, -0.00005135, -0.00002983, 0.00000076, 0.00003113, 0.00005325, 0.00006314, 0.00006115, 0.00005067, 0.00003655, 0.00002392, 0.00001718, 0.00001869, 0.00002755, 0.00003927, 0.00004750, 0.00004710, 0.00003719, 0.00002179, 0.00000759, -0.00000004, -0.00000012, 0.00000368, 0.00000580, 0.00000254, -0.00000578, -0.00001541, -0.00002184, -0.00002255, -0.00001799, -0.00001068, -0.00000339, 0.00000204, 0.00000478, 0.00000459, 0.00000163, -0.00000325, -0.00000825, -0.00001122, -0.00001088, -0.00000785, -0.00000432, -0.00000243, -0.00000253, -0.00000290, -0.00000148, 0.00000188, 0.00000459, 0.00000313, -0.00000353, -0.00001195, -0.00001564, -0.00000981, 0.00000419, 0.00001827, 0.00002252, 0.00001241, -0.00000682, -0.00002250, -0.00002298, -0.00000633, 0.00001712, 0.00003105, 0.00002444, 0.00000006, -0.00002636, -0.00003645, -0.00002200, 0.00000862, 0.00003551, 0.00004055, 0.00001991, -0.00001313, -0.00003709, -0.00003689, -0.00001394, 0.00001518, 0.00003041, 0.00002187, -0.00000388, -0.00002899, -0.00003681, -0.00002284, 0.00000323, 0.00002502, 0.00003068, 0.00001974, 0.00000173, -0.00001166, -0.00001441, -0.00000873, -0.00000148, 0.00000185, 0.00000068, -0.00000181, -0.00000264, -0.00000202, -0.00000270, -0.00000664, -0.00001233, -0.00001535, -0.00001195, -0.00000270, 0.00000720, 0.00001140, 0.00000719, -0.00000224, -0.00001003, -0.00001071, -0.00000428, 0.00000385, 0.00000710, 0.00000270, -0.00000631, -0.00001363, -0.00001459, -0.00000963, -0.00000369, -0.00000206, -0.00000614, -0.00001244, -0.00001573, -0.00001356, -0.00000880, -0.00000779, -0.00001556, -0.00003149, -0.00004868, },
		{ -0.00001224, -0.00002550, -0.00004062, -0.00005000, -0.00004779, -0.00003292, -0.00000932, 0.00001646, 0.00003827, 0.00005219, 0.00005670, 0.00005213, 0.00004049, 0.00002566, 0.00001307, 0.00000809, 0.00001325, 0.00002618, 0.00004009, 0.00004708, 0.00004258, 0.00002817, 0.00001068, -0.00000187, -0.00000492, 0.00000032, 0.00000821, 0.00001248, 0.00000985, 0.00000132, -0.00000925, -0.00001767, -0.00002137, -0.00001990, -0.00001424, -0.00000622, 0.00000181, 0.00000722, 0.00000789, 0.00000350, -0.00000378, -0.00001014, -0.00001229, -0.00000976, -0.00000519, -0.00000236, -0.00000308, -0.00000561, -0.00000611, -0.00000212, 0.00000465, 0.00000886, 0.00000570, -0.00000460, -0.00001554, -0.00001835, -0.00000863, 0.00000919, 0.00002363, 0.00002382, 0.00000772, -0.00001513, -0.00002925, -0.00002386, -0.00000110, 0.00002445, 0.00003526, 0.00002311, -0.00000467, -0.00002962, -0.00003467, -0.00001596, 0.00001419, 0.00003568, 0.00003427, 0.00001126, -0.00001772, -0.00003369, -0.00002724, -0.00000444, 0.00001803, 0.00002475, 0.00001187, -0.00001106, -0.00002842, -0.00002927, -0.00001404, 0.00000708, 0.00002178, 0.00002362, 0.00001478, 0.00000274, -0.00000573, -0.00000854, -0.00000781, -0.00000653, -0.00000558, -0.00000380, -0.00000040, 0.00000309, 0.00000329, -0.00000196, -0.00001099, -0.00001849, -0.00001893, -0.00001085, 0.00000164, 0.00001128, 0.00001264, 0.00000587, -0.00000364, -0.00000926, -0.00000783, -0.00000162, 0.00000385, 0.00000419, -0.00000072, -0.00000715, -0.00001076, -0.00001011, -0.00000736, -0.00000609, -0.00000795, -0.00001132, -0.00001282, -0.00001066, -0.00000680, -0.00000597, },
		{ -0.00001010, -0.00000614, -0.00000470, -0.00000926, -0.00001993, -0.00003276, -0.00004172, -0.00004208, -0.00003273, -0.00001616, 0.00000348, 0.00002251, 0.00003834, 0.00004892, 0.00005211, 0.00004637, 0.00003261, 0.00001548, 0.00000235, -0.00000022, 0.00000947, 0.00002651, 0.00004160, 0.00004622, 0.00003748, 0.00001969, 0.00000161, -0.00000875, -0.00000834, 0.00000003, 0.00001012, 0.00001609, 0.00001516, 0.00000801, -0.00000251, -0.00001293, -0.00002012, -0.00002181, -0.00001723, -0.00000778, 0.00000276, 0.00000960, 0.00000949, 0.00000290, -0.00000588, -0.00001140, -0.00001078, -0.00000577, -0.00000134, -0.00000164, -0.00000639, -0.00001082, -0.00000942, -0.00000097, 0.00000949, 0.00001385, 0.00000720, -0.00000738, -0.00001986, -0.00002006, -0.00000560, 0.00001514, 0.00002792, 0.00002256, 0.00000091, -0.00002336, -0.00003354, -0.00002162, 0.00000559, 0.00003048, 0.00003646, 0.00001944, -0.00000918, -0.00003025, -0.00002996, -0.00000914, 0.00001762, 0.00003232, 0.00002579, 0.00000351, -0.00001892, -0.00002698, -0.00001690, 0.00000255, 0.00001704, 0.00001683, 0.00000290, -0.00001469, -0.00002438, -0.00002078, -0.00000721, 0.00000781, 0.00001673, 0.00001721, 0.00001191, 0.00000511, -0.00000065, -0.00000518, -0.00000893, -0.00001114, -0.00001003, -0.00000483, 0.00000230, 0.00000676, 0.00000455, -0.00000439, -0.00001531, -0.00002141, -0.00001838, -0.00000753, 0.00000498, 0.00001225, 0.00001116, 0.00000408, -0.00000336, -0.00000642, -0.00000437, -0.00000025, 0.00000200, 0.00000064, -0.00000300, -0.00000614, -0.00000722, -0.00000697, -0.00000739, -0.00000947, -0.00001202, -0.00001263, },
		{ -0.00000885, -0.00001164, -0.00001244, -0.00001027, -0.00000667, -0.00000509, -0.00000846, -0.00001684, -0.00002690, -0.00003383, -0.00003428, -0.00002804, -0.00001743, -0.00000509, 0.00000788, 0.00002159, 0.00003544, 0.00004643, 0.00004986, 0.00004233, 0.00002523, 0.00000553, -0.00000724, -0.00000635, 0.00000778, 0.00002745, 0.00004195, 0.00004359, 0.00003182, 0.00001291, -0.00000414, -0.00001256, -0.00001070, -0.00000166, 0.00000924, 0.00001715, 0.00001914, 0.00001441, 0.00000414, -0.00000857, -0.00001919, -0.00002321, -0.00001855, -0.00000733, 0.00000458, 0.00001076, 0.00000824, -0.00000035, -0.00000853, -0.00001062, -0.00000596, 0.00000052, 0.00000228, -0.00000334, -0.00001242, -0.00001699, -0.00001140, 0.00000255, 0.00001579, 0.00001827, 0.00000677, -0.00001167, -0.00002388, -0.00001990, -0.00000082, 0.00002100, 0.00002998, 0.00001850, -0.00000706, -0.00003000, -0.00003449, -0.00001667, 0.00001239, 0.00003400, 0.00003452, 0.00001446, -0.00001221, -0.00002788, -0.00002323, -0.00000300, 0.00001811, 0.00002603, 0.00001675, -0.00000191, -0.00001667, -0.00001837, -0.00000775, 0.00000612, 0.00001298, 0.00000851, -0.00000348, -0.00001465, -0.00001822, -0.00001304, -0.00000312, 0.00000607, 0.00001125, 0.00001233, 0.00001081, 0.00000761, 0.00000260, -0.00000413, -0.00001071, -0.00001388, -0.00001120, -0.00000358, 0.00000447, 0.00000736, 0.00000235, -0.00000816, -0.00001793, -0.00002085, -0.00001501, -0.00000383, 0.00000641, 0.00001077, 0.00000856, 0.00000300, -0.00000159, -0.00000292, -0.00000170, -0.00000032, -0.00000048, -0.00000203, -0.00000362, -0.00000438, -0.00000481, -0.00000616, },
		{ -0.00000322, -0.00000338, -0.00000465, -0.00000725, -0.00000992, -0.00001075, -0.00000898, -0.00000613, -0.00000536, -0.00000910, -0.00001689, -0.00002516, -0.00002946, -0.00002749, -0.00002062, -0.00001256, -0.00000618, -0.00000107, 0.00000597, 0.00001768, 0.00003284, 0.00004558, 0.00004853, 0.00003801, 0.00001750, -0.00000337, -0.00001409, -0.00000941, 0.00000756, 0.00002735, 0.00003964, 0.00003890, 0.00002652, 0.00000903, -0.00000607, -0.00001378, -0.00001287, -0.00000510, 0.00000612, 0.00001667, 0.00002237, 0.00002013, 0.00000966, -0.00000543, -0.00001848, -0.00002312, -0.00001723, -0.00000478, 0.00000633, 0.00000943, 0.00000364, -0.00000550, -0.00001025, -0.00000670, 0.00000197, 0.00000771, 0.00000411, -0.00000801, -0.00002025, -0.00002251, -0.00001093, 0.00000825, 0.00002227, 0.00002084, 0.00000411, -0.00001663, -0.00002643, -0.00001742, 0.00000504, 0.00002549, 0.00002907, 0.00001212, -0.00001478, -0.00003377, -0.00003193, -0.00001008, 0.00001786, 0.00003437, 0.00003007, 0.00000947, -0.00001292, -0.00002289, -0.00001583, 0.00000124, 0.00001562, 0.00001813, 0.00000883, -0.00000427, -0.00001179, -0.00000966, -0.00000119, 0.00000632, 0.00000731, 0.00000156, -0.00000655, -0.00001175, -0.00001161, -0.00000729, -0.00000172, 0.00000300, 0.00000653, 0.00000922, 0.00001062, 0.00000915, 0.00000372, -0.00000450, -0.00001179, -0.00001404, -0.00000975, -0.00000163, 0.00000481, 0.00000491, -0.00000189, -0.00001154, -0.00001809, -0.00001760, -0.00001052, -0.00000104, 0.00000598, 0.00000810, 0.00000617, 0.00000295, 0.00000073, -0.00000006, -0.00000036, -0.00000111, -0.00000225, -0.00000306, },
		{ 0.00000002, -0.00000153, -0.00000251, -0.00000277, -0.00000304, -0.00000429, -0.00000662, -0.00000870, -0.00000873, -0.00000623, -0.00000314, -0.00000296, -0.00000813, -0.00001765, -0.00002690, -0.00003041, -0.00002568, -0.00001522, -0.00000501, -0.00000019, -0.00000110, -0.00000275, 0.00000144, 0.00001431, 0.00003193, 0.00004528, 0.00004601, 0.00003212, 0.00000988, -0.00000963, -0.00001697, -0.00000959, 0.00000735, 0.00002459, 0.00003415, 0.00003292, 0.00002286, 0.00000868, -0.00000471, -0.00001359, -0.00001561, -0.00000993, 0.00000198, 0.00001574, 0.00002497, 0.00002431, 0.00001297, -0.00000383, -0.00001732, -0.00002051, -0.00001288, -0.00000085, 0.00000667, 0.00000476, -0.00000387, -0.00001102, -0.00000953, 0.00000066, 0.00001181, 0.00001398, 0.00000305, -0.00001519, -0.00002824, -0.00002577, -0.00000763, 0.00001511, 0.00002737, 0.00002070, -0.00000035, -0.00002102, -0.00002658, -0.00001281, 0.00001078, 0.00002747, 0.00002508, 0.00000451, -0.00002077, -0.00003399, -0.00002650, -0.00000325, 0.00002095, 0.00003171, 0.00002428, 0.00000567, -0.00001114, -0.00001636, -0.00000922, 0.00000294, 0.00001093, 0.00001025, 0.00000327, -0.00000370, -0.00000571, -0.00000249, 0.00000218, 0.00000409, 0.00000178, -0.00000281, -0.00000648, -0.00000739, -0.00000604, -0.00000404, -0.00000227, -0.00000016, 0.00000324, 0.00000750, 0.00001034, 0.00000912, 0.00000313, -0.00000514, -0.00001125, -0.00001179, -0.00000692, -0.00000041, 0.00000287, 0.00000036, -0.00000651, -0.00001338, -0.00001600, -0.00001300, -0.00000634, 0.00000040, 0.00000450, 0.00000549, 0.00000465, 0.00000348, 0.00000252, 0.00000148, },
		{ 0.00000389, 0.00000376, 0.00000310, 0.00000180, 0.00000029, -0.00000079, -0.00000137, -0.00000223, -0.00000426, -0.00000724, -0.00000942, -0.00000863, -0.00000436, 0.00000086, 0.00000246, -0.00000312, -0.00001519, -0.00002820, -0.00003463, -0.00002996, -0.00001612, -0.00000069, 0.00000792, 0.00000627, -0.00000159, -0.00000680, -0.00000196, 0.00001340, 0.00003201, 0.00004342, 0.00004067, 0.00002477, 0.00000390, -0.00001174, -0.00001570, -0.00000820, 0.00000540, 0.00001839, 0.00002618, 0.00002717, 0.00002183, 0.00001152, -0.00000146, -0.00001331, -0.00001911, -0.00001510, -0.00000175, 0.00001506, 0.00002648, 0.00002588, 0.00001346, -0.00000336, -0.00001466, -0.00001472, -0.00000595, 0.00000308, 0.00000436, -0.00000326, -0.00001292, -0.00001509, -0.00000548, 0.00001069, 0.00002164, 0.00001764, -0.00000114, -0.00002351, -0.00003448, -0.00002573, -0.00000203, 0.00002155, 0.00002976, 0.00001775, -0.00000551, -0.00002357, -0.00002398, -0.00000694, 0.00001510, 0.00002633, 0.00001866, -0.00000288, -0.00002392, -0.00003075, -0.00001939, 0.00000245, 0.00002125, 0.00002687, 0.00001848, 0.00000375, -0.00000744, -0.00000967, -0.00000448, 0.00000224, 0.00000535, 0.00000385, 0.00000053, -0.00000114, -0.00000005, 0.00000209, 0.00000273, 0.00000088, -0.00000218, -0.00000426, -0.00000428, -0.00000310, -0.00000244, -0.00000302, -0.00000370, -0.00000248, 0.00000142, 0.00000639, 0.00000924, 0.00000756, 0.00000173, -0.00000502, -0.00000886, -0.00000803, -0.00000411, -0.00000072, -0.00000090, -0.00000489, -0.00001017, -0.00001337, -0.00001258, -0.00000845, -0.00000326, 0.00000081, 0.00000296, 0.00000372, },
		{ -0.00000136, 0.00000090, 0.00000214, 0.00000290, 0.00000331, 0.00000317, 0.00000245, 0.00000164, 0.00000127, 0.00000118, 0.00000023, -0.00000269, -0.00000730, -0.00001118, -0.00001112, -0.00000570, 0.00000264, 0.00000813, 0.00000514, -0.00000754, -0.00002468, -0.00003703, -0.00003690, -0.00002343, -0.00000368, 0.00001158, 0.00001481, 0.00000659, -0.00000488, -0.00000945, -0.00000189, 0.00001469, 0.00003107, 0.00003808, 0.00003228, 0.00001743, 0.00000133, -0.00000915, -0.00001147, -0.00000721, 0.00000050, 0.00000923, 0.00001742, 0.00002317, 0.00002358, 0.00001622, 0.00000193, -0.00001378, -0.00002276, -0.00001917, -0.00000399, 0.00001460, 0.00002600, 0.00002406, 0.00001128, -0.00000302, -0.00000958, -0.00000588, 0.00000223, 0.00000543, -0.00000118, -0.00001361, -0.00002151, -0.00001618, 0.00000182, 0.00002170, 0.00002940, 0.00001773, -0.00000760, -0.00003105, -0.00003740, -0.00002231, 0.00000455, 0.00002595, 0.00002877, 0.00001270, -0.00001002, -0.00002346, -0.00001906, -0.00000109, 0.00001696, 0.00002217, 0.00001105, -0.00000865, -0.00002377, -0.00002488, -0.00001207, 0.00000613, 0.00001910, 0.00002107, 0.00001377, 0.00000372, -0.00000291, -0.00000412, -0.00000203, -0.00000002, 0.00000031, -0.00000017, 0.00000026, 0.00000208, 0.00000393, 0.00000393, 0.00000153, -0.00000181, -0.00000380, -0.00000331, -0.00000132, -0.00000007, -0.00000100, -0.00000341, -0.00000495, -0.00000357, 0.00000061, 0.00000514, 0.00000705, 0.00000505, 0.00000051, -0.00000364, -0.00000513, -0.00000400, -0.00000237, -0.00000261, -0.00000545, -0.00000945, -0.00001217, -0.00001196, -0.00000901, -0.00000491, },
		{ -0.00001255, -0.00001004, -0.00000623, -0.00000268, -0.00000018, 0.00000134, 0.00000223, 0.00000259, 0.00000230, 0.00000159, 0.00000120, 0.00000186, 0.00000323, 0.00000360, 0.00000101, -0.00000487, -0.00001144, -0.00001427, -0.00001007, 0.00000009, 0.00001026, 0.00001274, 0.00000329, -0.00001516, -0.00003328, -0.00004066, -0.00003248, -0.00001288, 0.00000740, 0.00001782, 0.00001465, 0.00000295, -0.00000728, -0.00000799, 0.00000184, 0.00001624, 0.00002689, 0.00002865, 0.00002220, 0.00001218, 0.00000327, -0.00000273, -0.00000642, -0.00000841, -0.00000741, -0.00000129, 0.00000988, 0.00002178, 0.00002721, 0.00002087, 0.00000399, -0.00001504, -0.00002537, -0.00002083, -0.00000436, 0.00001370, 0.00002266, 0.00001878, 0.00000740, -0.00000159, -0.00000174, 0.00000495, 0.00000983, 0.00000497, -0.00000954, -0.00002443, -0.00002761, -0.00001358, 0.00001119, 0.00003150, 0.00003347, 0.00001439, -0.00001469, -0.00003587, -0.00003625, -0.00001637, 0.00001040, 0.00002713, 0.00002460, 0.00000682, -0.00001266, -0.00002057, -0.00001284, 0.00000345, 0.00001593, 0.00001584, 0.00000371, -0.00001186, -0.00002065, -0.00001770, -0.00000577, 0.00000756, 0.00001540, 0.00001557, 0.00001065, 0.00000495, 0.00000122, -0.00000050, -0.00000155, -0.00000264, -0.00000306, -0.00000169, 0.00000144, 0.00000463, 0.00000567, 0.00000367, -0.00000006, -0.00000295, -0.00000317, -0.00000111, 0.00000106, 0.00000114, -0.00000121, -0.00000420, -0.00000537, -0.00000359, 0.00000010, 0.00000327, 0.00000407, 0.00000250, 0.00000019, -0.00000110, -0.00000104, -0.00000083, -0.00000211, -0.00000540, -0.00000952, -0.00001240, },
		{ -0.00000309, -0.00000812, -0.00001216, -0.00001350, -0.00001190, -0.00000846, -0.00000467, -0.00000149, 0.00000085, 0.00000237, 0.00000288, 0.00000216, 0.00000062, -0.00000046, 0.00000029, 0.00000297, 0.00000562, 0.00000520, -0.00000005, -0.00000846, -0.00001507, -0.00001462, -0.00000569, 0.00000726, 0.00001574, 0.00001259, -0.00000282, -0.00002321, -0.00003736, -0.00003711, -0.00002258, -0.00000213, 0.00001312, 0.00001637, 0.00000905, -0.00000103, -0.00000565, -0.00000173, 0.00000742, 0.00001550, 0.00001827, 0.00001618, 0.00001279, 0.00001081, 0.00000942, 0.00000541, -0.00000292, -0.00001263, -0.00001712, -0.00001090, 0.00000515, 0.00002278, 0.00003101, 0.00002364, 0.00000408, -0.00001632, -0.00002566, -0.00001945, -0.00000335, 0.00001137, 0.00001610, 0.00001085, 0.00000330, 0.00000185, 0.00000829, 0.00001596, 0.00001507, 0.00000131, -0.00001932, -0.00003349, -0.00002982, -0.00000775, 0.00002067, 0.00003808, 0.00003324, 0.00000877, -0.00002038, -0.00003667, -0.00003133, -0.00000949, 0.00001396, 0.00002474, 0.00001827, 0.00000155, -0.00001277, -0.00001552, -0.00000667, 0.00000578, 0.00001233, 0.00000865, -0.00000208, -0.00001227, -0.00001552, -0.00001058, -0.00000123, 0.00000718, 0.00001132, 0.00001123, 0.00000901, 0.00000647, 0.00000398, 0.00000106, -0.00000219, -0.00000450, -0.00000433, -0.00000133, 0.00000289, 0.00000573, 0.00000544, 0.00000249, -0.00000085, -0.00000226, -0.00000114, 0.00000103, 0.00000203, 0.00000070, -0.00000217, -0.00000455, -0.00000486, -0.00000310, -0.00000071, 0.00000078, 0.00000097, 0.00000070, 0.00000098, 0.00000191, 0.00000232, 0.00000079, },
		{ 0.00000253, 0.00000447, 0.00000416, 0.00000083, -0.00000456, -0.00000982, -0.00001292, -0.00001310, -0.00001097, -0.00000770, -0.00000419, -0.00000082, 0.00000207, 0.00000374, 0.00000342, 0.00000120, -0.00000138, -0.00000205, 0.00000042, 0.00000474, 0.00000732, 0.00000479, -0.00000306, -0.00001215, -0.00001619, -0.00001117, 0.00000112, 0.00001325, 0.00001659, 0.00000716, -0.00001103, -0.00002809, -0.00003439, -0.00002679, -0.00001052, 0.00000466, 0.00001112, 0.00000817, 0.00000150, -0.00000176, 0.00000132, 0.00000780, 0.00001194, 0.00001052, 0.00000573, 0.00000314, 0.00000655, 0.00001389, 0.00001808, 0.00001256, -0.00000263, -0.00001940, -0.00002648, -0.00001748, 0.00000379, 0.00002494, 0.00003307, 0.00002343, 0.00000255, -0.00001639, -0.00002276, -0.00001532, -0.00000204, 0.00000684, 0.00000677, 0.00000183, 0.00000046, 0.00000748, 0.00001910, 0.00002508, 0.00001682, -0.00000485, -0.00002847, -0.00003889, -0.00002779, -0.00000015, 0.00002811, 0.00004019, 0.00002927, 0.00000273, -0.00002299, -0.00003317, -0.00002389, -0.00000342, 0.00001435, 0.00001934, 0.00001122, -0.00000198, -0.00001045, -0.00000949, -0.00000178, 0.00000565, 0.00000711, 0.00000199, -0.00000558, -0.00001034, -0.00000962, -0.00000458, 0.00000150, 0.00000585, 0.00000783, 0.00000833, 0.00000821, 0.00000733, 0.00000497, 0.00000110, -0.00000290, -0.00000493, -0.00000376, -0.00000012, 0.00000365, 0.00000528, 0.00000414, 0.00000159, -0.00000027, -0.00000025, 0.00000108, 0.00000208, 0.00000146, -0.00000067, -0.00000296, -0.00000409, -0.00000375, -0.00000275, -0.00000207, -0.00000188, -0.00000144, 0.00000009, },
		{ -0.00000378, -0.00000404, -0.00000260, 0.00000059, 0.00000410, 0.00000577, 0.00000421, -0.00000022, -0.00000565, -0.00000997, -0.00001201, -0.00001188, -0.00001030, -0.00000774, -0.00000427, -0.00000021, 0.00000336, 0.00000482, 0.00000330, -0.00000017, -0.00000291, -0.00000235, 0.00000169, 0.00000623, 0.00000704, 0.00000188, -0.00000707, -0.00001400, -0.00001336, -0.00000426, 0.00000811, 0.00001536, 0.00001163, -0.00000203, -0.00001794, -0.00002676, -0.00002390, -0.00001242, -0.00000058, 0.00000441, 0.00000159, -0.00000356, -0.00000397, 0.00000289, 0.00001267, 0.00001750, 0.00001263, 0.00000093, -0.00000858, -0.00000748, 0.00000503, 0.00002044, 0.00002653, 0.00001634, -0.00000596, -0.00002703, -0.00003317, -0.00001980, 0.00000514, 0.00002647, 0.00003197, 0.00002018, 0.00000051, -0.00001417, -0.00001667, -0.00000957, -0.00000168, -0.00000002, -0.00000403, -0.00000643, -0.00000016, 0.00001454, 0.00002871, 0.00003063, 0.00001496, -0.00001193, -0.00003490, -0.00003963, -0.00002235, 0.00000720, 0.00003182, 0.00003774, 0.00002308, -0.00000180, -0.00002163, -0.00002615, -0.00001572, 0.00000047, 0.00001156, 0.00001222, 0.00000491, -0.00000329, -0.00000649, -0.00000380, 0.00000108, 0.00000350, 0.00000153, -0.00000306, -0.00000672, -0.00000701, -0.00000413, -0.00000017, 0.00000286, 0.00000447, 0.00000546, 0.00000656, 0.00000745, 0.00000698, 0.00000442, 0.00000047, -0.00000292, -0.00000392, -0.00000218, 0.00000093, 0.00000336, 0.00000391, 0.00000293, 0.00000177, 0.00000150, 0.00000204, 0.00000238, 0.00000161, -0.00000018, -0.00000204, -0.00000300, -0.00000295, -0.00000266, -0.00000296, },
		{ -0.00000219, -0.00000162, -0.00000212, -0.00000377, -0.00000531, -0.00000508, -0.00000240, 0.00000167, 0.00000490, 0.00000539, 0.00000280, -0.00000155, -0.00000575, -0.00000859, -0.00001002, -0.00001048, -0.00001003, -0.00000812, -0.00000437, 0.00000042, 0.00000420, 0.00000494, 0.00000238, -0.00000135, -0.00000306, -0.00000099, 0.00000340, 0.00000619, 0.00000400, -0.00000295, -0.00001026, -0.00001226, -0.00000636, 0.00000430, 0.00001246, 0.00001177, 0.00000163, -0.00001188, -0.00002006, -0.00001789, -0.00000789, 0.00000176, 0.00000346, -0.00000367, -0.00001280, -0.00001447, -0.00000428, 0.00001284, 0.00002547, 0.00002395, 0.00000807, -0.00001180, -0.00002149, -0.00001323, 0.00000821, 0.00002817, 0.00003201, 0.00001552, -0.00001187, -0.00003320, -0.00003543, -0.00001792, 0.00000764, 0.00002569, 0.00002723, 0.00001486, -0.00000065, -0.00000918, -0.00000833, -0.00000378, -0.00000310, -0.00000849, -0.00001445, -0.00001239, 0.00000150, 0.00002153, 0.00003523, 0.00003182, 0.00001045, -0.00001795, -0.00003713, -0.00003589, -0.00001524, 0.00001234, 0.00003109, 0.00003178, 0.00001662, -0.00000343, -0.00001651, -0.00001718, -0.00000853, 0.00000155, 0.00000646, 0.00000489, 0.00000034, -0.00000263, -0.00000207, 0.00000045, 0.00000181, 0.00000026, -0.00000324, -0.00000601, -0.00000602, -0.00000332, 0.00000024, 0.00000273, 0.00000357, 0.00000366, 0.00000417, 0.00000531, 0.00000614, 0.00000546, 0.00000306, 0.00000004, -0.00000197, -0.00000205, -0.00000057, 0.00000123, 0.00000233, 0.00000262, 0.00000272, 0.00000314, 0.00000365, 0.00000348, 0.00000209, -0.00000012, -0.00000203, -0.00000269, },
		{ 0.00000024, -0.00000236, -0.00000322, -0.00000228, -0.00000088, -0.00000070, -0.00000237, -0.00000482, -0.00000606, -0.00000472, -0.00000118, 0.00000259, 0.00000442, 0.00000349, 0.00000065, -0.00000243, -0.00000468, -0.00000619, -0.00000766, -0.00000921, -0.00000985, -0.00000827, -0.00000422, 0.00000071, 0.00000393, 0.00000380, 0.00000109, -0.00000145, -0.00000129, 0.00000158, 0.00000434, 0.00000356, -0.00000162, -0.00000810, -0.00001066, -0.00000613, 0.00000351, 0.00001170, 0.00001182, 0.00000244, -0.00001078, -0.00001865, -0.00001515, -0.00000252, 0.00000971, 0.00001139, -0.00000030, -0.00001777, -0.00002772, -0.00002062, 0.00000161, 0.00002577, 0.00003595, 0.00002468, -0.00000126, -0.00002477, -0.00002996, -0.00001312, 0.00001450, 0.00003422, 0.00003260, 0.00001040, -0.00001829, -0.00003584, -0.00003274, -0.00001311, 0.00000939, 0.00002164, 0.00001953, 0.00000907, 0.00000009, -0.00000189, 0.00000064, 0.00000061, -0.00000637, -0.00001713, -0.00002267, -0.00001523, 0.00000459, 0.00002668, 0.00003744, 0.00002899, 0.00000500, -0.00002115, -0.00003477, -0.00002897, -0.00000849, 0.00001401, 0.00002641, 0.00002413, 0.00001170, -0.00000171, -0.00000879, -0.00000814, -0.00000352, 0.00000010, 0.00000041, -0.00000129, -0.00000219, -0.00000083, 0.00000165, 0.00000275, 0.00000092, -0.00000300, -0.00000642, -0.00000697, -0.00000430, -0.00000011, 0.00000327, 0.00000458, 0.00000422, 0.00000353, 0.00000348, 0.00000396, 0.00000410, 0.00000329, 0.00000172, 0.00000026, -0.00000040, -0.00000021, 0.00000034, 0.00000084, 0.00000139, 0.00000239, 0.00000386, 0.00000512, 0.00000509, 0.00000325, },
		{ 0.00000596, 0.00000667, 0.00000488, 0.00000117, -0.00000251, -0.00000419, -0.00000327, -0.00000104, 0.00000032, -0.00000057, -0.00000320, -0.00000554, -0.00000562, -0.00000313, 0.00000037, 0.00000268, 0.00000259, 0.00000068, -0.00000142, -0.00000249, -0.00000278, -0.00000350, -0.00000546, -0.00000798, -0.00000924, -0.00000778, -0.00000397, 0.00000008, 0.00000212, 0.00000163, 0.00000024, 0.00000023, 0.00000222, 0.00000425, 0.00000326, -0.00000188, -0.00000852, -0.00001146, -0.00000691, 0.00000372, 0.00001368, 0.00001513, 0.00000520, -0.00001074, -0.00002186, -0.00001926, -0.00000314, 0.00001609, 0.00002424, 0.00001346, -0.00001122, -0.00003402, -0.00003865, -0.00001979, 0.00001234, 0.00003798, 0.00004081, 0.00001905, -0.00001306, -0.00003464, -0.00003208, -0.00000793, 0.00002121, 0.00003615, 0.00002785, 0.00000266, -0.00002291, -0.00003380, -0.00002593, -0.00000736, 0.00000892, 0.00001450, 0.00001043, 0.00000435, 0.00000300, 0.00000644, 0.00000848, 0.00000277, -0.00001072, -0.00002425, -0.00002744, -0.00001517, 0.00000767, 0.00002854, 0.00003514, 0.00002348, 0.00000048, -0.00002060, -0.00002858, -0.00002082, -0.00000381, 0.00001204, 0.00001920, 0.00001677, 0.00000939, 0.00000279, -0.00000020, -0.00000059, -0.00000103, -0.00000290, -0.00000515, -0.00000554, -0.00000302, 0.00000108, 0.00000388, 0.00000319, -0.00000076, -0.00000542, -0.00000781, -0.00000648, -0.00000234, 0.00000227, 0.00000524, 0.00000593, 0.00000505, 0.00000377, 0.00000281, 0.00000220, 0.00000171, 0.00000126, 0.00000102, 0.00000103, 0.00000107, 0.00000082, 0.00000038, 0.00000037, 0.00000148, 0.00000370, },
		{ -0.00000001, 0.00000070, 0.00000318, 0.00000634, 0.00000807, 0.00000678, 0.00000265, -0.00000224, -0.00000521, -0.00000489, -0.00000217, 0.00000049, 0.00000089, -0.00000125, -0.00000413, -0.00000539, -0.00000397, -0.00000093, 0.00000149, 0.00000167, -0.00000016, -0.00000221, -0.00000280, -0.00000181, -0.00000069, -0.00000115, -0.00000352, -0.00000638, -0.00000772, -0.00000666, -0.00000413, -0.00000198, -0.00000116, -0.00000082, 0.00000069, 0.00000381, 0.00000654, 0.00000561, -0.00000053, -0.00000929, -0.00001463, -0.00001117, 0.00000107, 0.00001517, 0.00002109, 0.00001276, -0.00000640, -0.00002435, -0.00002794, -0.00001246, 0.00001371, 0.00003339, 0.00003182, 0.00000732, -0.00002626, -0.00004769, -0.00004232, -0.00001218, 0.00002467, 0.00004565, 0.00003835, 0.00000852, -0.00002405, -0.00003880, -0.00002775, 0.00000006, 0.00002555, 0.00003275, 0.00001887, -0.00000529, -0.00002395, -0.00002728, -0.00001688, -0.00000265, 0.00000564, 0.00000543, 0.00000179, 0.00000173, 0.00000746, 0.00001412, 0.00001384, 0.00000274, -0.00001484, -0.00002841, -0.00002842, -0.00001324, 0.00000925, 0.00002649, 0.00002924, 0.00001713, -0.00000173, -0.00001647, -0.00002023, -0.00001343, -0.00000205, 0.00000729, 0.00001138, 0.00001124, 0.00000978, 0.00000873, 0.00000753, 0.00000457, -0.00000060, -0.00000617, -0.00000917, -0.00000778, -0.00000292, 0.00000223, 0.00000437, 0.00000230, -0.00000240, -0.00000660, -0.00000772, -0.00000523, -0.00000060, 0.00000389, 0.00000660, 0.00000708, 0.00000586, 0.00000383, 0.00000175, 0.00000022, -0.00000037, 0.00000006, 0.00000106, 0.00000182, 0.00000167, 0.00000072, },
		{ 0.00000147, 0.00000191, 0.00000100, -0.00000019, 0.00000002, 0.00000254, 0.00000648, 0.00000944, 0.00000906, 0.00000483, -0.00000128, -0.00000600, -0.00000695, -0.00000425, -0.00000039, 0.00000164, 0.00000058, -0.00000235, -0.00000455, -0.00000416, -0.00000154, 0.00000112, 0.00000168, -0.00000026, -0.00000305, -0.00000441, -0.00000333, -0.00000085, 0.00000093, 0.00000058, -0.00000159, -0.00000396, -0.00000521, -0.00000533, -0.00000530, -0.00000560, -0.00000527, -0.00000256, 0.00000291, 0.00000865, 0.00001023, 0.00000453, -0.00000676, -0.00001685, -0.00001778, -0.00000641, 0.00001184, 0.00002523, 0.00002320, 0.00000453, -0.00002022, -0.00003446, -0.00002669, 0.00000069, 0.00003136, 0.00004494, 0.00003053, -0.00000511, -0.00004092, -0.00005475, -0.00003773, -0.00000027, 0.00003461, 0.00004616, 0.00002905, -0.00000392, -0.00003105, -0.00003619, -0.00001862, 0.00000790, 0.00002554, 0.00002440, 0.00000796, -0.00001111, -0.00002079, -0.00001771, -0.00000782, -0.00000033, 0.00000006, -0.00000380, -0.00000486, 0.00000132, 0.00001219, 0.00001962, 0.00001619, 0.00000134, -0.00001736, -0.00002893, -0.00002628, -0.00001089, 0.00000834, 0.00002096, 0.00002141, 0.00001168, -0.00000113, -0.00000993, -0.00001173, -0.00000821, -0.00000302, 0.00000132, 0.00000473, 0.00000826, 0.00001209, 0.00001451, 0.00001319, 0.00000729, -0.00000127, -0.00000857, -0.00001121, -0.00000846, -0.00000272, 0.00000218, 0.00000343, 0.00000081, -0.00000350, -0.00000661, -0.00000668, -0.00000371, 0.00000085, 0.00000512, 0.00000767, 0.00000796, 0.00000620, 0.00000324, 0.00000029, -0.00000146, -0.00000144, -0.00000006, },
		{ -0.00000112, -0.00000216, -0.00000115, 0.00000064, 0.00000155, 0.00000079, -0.00000077, -0.00000111, 0.00000131, 0.00000607, 0.00001060, 0.00001177, 0.00000805, 0.00000091, -0.00000597, -0.00000902, -0.00000721, -0.00000257, 0.00000132, 0.00000196, -0.00000039, -0.00000326, -0.00000402, -0.00000205, 0.00000093, 0.00000232, 0.00000084, -0.00000250, -0.00000527, -0.00000552, -0.00000326, -0.00000023, 0.00000168, 0.00000180, 0.00000078, -0.00000050, -0.00000203, -0.00000446, -0.00000782, -0.00001034, -0.00000918, -0.00000280, 0.00000665, 0.00001345, 0.00001184, 0.00000071, -0.00001406, -0.00002232, -0.00001648, 0.00000220, 0.00002278, 0.00003095, 0.00001897, -0.00000770, -0.00003249, -0.00003796, -0.00001794, 0.00001700, 0.00004524, 0.00004778, 0.00002107, -0.00001991, -0.00005073, -0.00005308, -0.00002640, 0.00001208, 0.00003877, 0.00003893, 0.00001535, -0.00001476, -0.00003206, -0.00002764, -0.00000756, 0.00001293, 0.00002064, 0.00001290, -0.00000224, -0.00001333, -0.00001412, -0.00000724, -0.00000070, -0.00000068, -0.00000648, -0.00001142, -0.00000875, 0.00000242, 0.00001575, 0.00002204, 0.00001588, -0.00000022, -0.00001737, -0.00002610, -0.00002233, -0.00000941, 0.00000482, 0.00001326, 0.00001356, 0.00000823, 0.00000178, -0.00000271, -0.00000481, -0.00000563, -0.00000571, -0.00000415, 0.00000035, 0.00000760, 0.00001503, 0.00001880, 0.00001632, 0.00000818, -0.00000197, -0.00000947, -0.00001145, -0.00000829, -0.00000297, 0.00000099, 0.00000162, -0.00000070, -0.00000394, -0.00000586, -0.00000520, -0.00000210, 0.00000227, 0.00000628, 0.00000851, 0.00000819, 0.00000557, 0.00000189, },
		{ 0.00000879, 0.00000732, 0.00000382, 0.00000013, -0.00000185, -0.00000145, 0.00000031, 0.00000147, 0.00000068, -0.00000152, -0.00000287, -0.00000105, 0.00000428, 0.00001070, 0.00001428, 0.00001219, 0.00000482, -0.00000419, -0.00001018, -0.00001049, -0.00000607, -0.00000057, 0.00000231, 0.00000141, -0.00000150, -0.00000339, -0.00000251, 0.00000035, 0.00000266, 0.00000233, -0.00000064, -0.00000428, -0.00000624, -0.00000558, -0.00000313, -0.00000044, 0.00000156, 0.00000294, 0.00000392, 0.00000380, 0.00000114, -0.00000463, -0.00001155, -0.00001527, -0.00001185, -0.00000130, 0.00001095, 0.00001667, 0.00001053, -0.00000513, -0.00002038, -0.00002375, -0.00001053, 0.00001265, 0.00003083, 0.00003044, 0.00000925, -0.00002062, -0.00003961, -0.00003352, -0.00000383, 0.00003211, 0.00005162, 0.00004139, 0.00000652, -0.00003246, -0.00005259, -0.00004311, -0.00001178, 0.00002091, 0.00003546, 0.00002556, 0.00000081, -0.00002112, -0.00002687, -0.00001555, 0.00000233, 0.00001360, 0.00001183, 0.00000084, -0.00000957, -0.00001173, -0.00000563, 0.00000194, 0.00000345, -0.00000304, -0.00001235, -0.00001626, -0.00001001, 0.00000385, 0.00001707, 0.00002135, 0.00001392, -0.00000085, -0.00001473, -0.00002098, -0.00001807, -0.00000949, -0.00000055, 0.00000517, 0.00000720, 0.00000698, 0.00000574, 0.00000344, -0.00000042, -0.00000524, -0.00000867, -0.00000783, -0.00000152, 0.00000834, 0.00001730, 0.00002088, 0.00001722, 0.00000816, -0.00000193, -0.00000879, -0.00001045, -0.00000787, -0.00000380, -0.00000091, -0.00000047, -0.00000199, -0.00000397, -0.00000479, -0.00000353, -0.00000027, 0.00000394, 0.00000747, },
		{ -0.00000160, 0.00000203, 0.00000596, 0.00000840, 0.00000806, 0.00000510, 0.00000121, -0.00000136, -0.00000133, 0.00000061, 0.00000228, 0.00000165, -0.00000133, -0.00000434, -0.00000418, 0.00000071, 0.00000864, 0.00001523, 0.00001616, 0.00001012, -0.00000009, -0.00000921, -0.00001285, -0.00001017, -0.00000412, 0.00000094, 0.00000228, 0.00000035, -0.00000208, -0.00000245, -0.00000039, 0.00000225, 0.00000310, 0.00000127, -0.00000209, -0.00000490, -0.00000583, -0.00000504, -0.00000347, -0.00000162, 0.00000093, 0.00000447, 0.00000786, 0.00000834, 0.00000352, -0.00000607, -0.00001582, -0.00001926, -0.00001263, 0.00000143, 0.00001440, 0.00001710, 0.00000642, -0.00001130, -0.00002360, -0.00002027, -0.00000123, 0.00002206, 0.00003352, 0.00002356, -0.00000338, -0.00003051, -0.00003931, -0.00002215, 0.00001183, 0.00004185, 0.00004882, 0.00002780, -0.00000870, -0.00003888, -0.00004574, -0.00002758, 0.00000190, 0.00002349, 0.00002522, 0.00000928, -0.00001107, -0.00002166, -0.00001704, -0.00000307, 0.00000867, 0.00000990, 0.00000124, -0.00000926, -0.00001295, -0.00000727, 0.00000263, 0.00000830, 0.00000466, -0.00000610, -0.00001620, -0.00001802, -0.00000952, 0.00000448, 0.00001579, 0.00001831, 0.00001152, 0.00000001, -0.00001013, -0.00001506, -0.00001462, -0.00001100, -0.00000639, -0.00000169, 0.00000305, 0.00000726, 0.00000927, 0.00000735, 0.00000142, -0.00000597, -0.00001063, -0.00000921, -0.00000145, 0.00000933, 0.00001807, 0.00002079, 0.00001664, 0.00000808, -0.00000090, -0.00000695, -0.00000886, -0.00000753, -0.00000497, -0.00000300, -0.00000248, -0.00000313, -0.00000393, -0.00000367, },
		{ -0.00000438, -0.00000437, -0.00000405, -0.00000245, 0.00000080, 0.00000489, 0.00000805, 0.00000856, 0.00000596, 0.00000173, -0.00000155, -0.00000190, 0.00000055, 0.00000343, 0.00000385, 0.00000072, -0.00000405, -0.00000658, -0.00000373, 0.00000425, 0.00001337, 0.00001817, 0.00001528, 0.00000574, -0.00000552, -0.00001286, -0.00001338, -0.00000844, -0.00000223, 0.00000141, 0.00000137, -0.00000054, -0.00000162, -0.00000058, 0.00000163, 0.00000295, 0.00000214, -0.00000030, -0.00000275, -0.00000404, -0.00000434, -0.00000449, -0.00000462, -0.00000349, 0.00000039, 0.00000669, 0.00001219, 0.00001230, 0.00000455, -0.00000848, -0.00001961, -0.00002137, -0.00001155, 0.00000427, 0.00001563, 0.00001433, 0.00000068, -0.00001574, -0.00002231, -0.00001241, 0.00000895, 0.00002763, 0.00002981, 0.00001204, -0.00001534, -0.00003437, -0.00003144, -0.00000685, 0.00002471, 0.00004351, 0.00003768, 0.00001091, -0.00002017, -0.00003721, -0.00003189, -0.00001059, 0.00001102, 0.00001907, 0.00001050, -0.00000613, -0.00001786, -0.00001682, -0.00000531, 0.00000690, 0.00001043, 0.00000318, -0.00000866, -0.00001569, -0.00001254, -0.00000163, 0.00000889, 0.00001129, 0.00000378, -0.00000851, -0.00001739, -0.00001724, -0.00000842, 0.00000362, 0.00001235, 0.00001413, 0.00000962, 0.00000221, -0.00000477, -0.00000969, -0.00001244, -0.00001312, -0.00001123, -0.00000626, 0.00000100, 0.00000796, 0.00001120, 0.00000869, 0.00000142, -0.00000668, -0.00001093, -0.00000859, -0.00000048, 0.00000962, 0.00001714, 0.00001913, 0.00001546, 0.00000837, 0.00000091, -0.00000454, -0.00000710, -0.00000721, -0.00000608, -0.00000491, },
		{ -0.00000530, -0.00000667, -0.00000681, -0.00000655, -0.00000634, -0.00000587, -0.00000425, -0.00000086, 0.00000383, 0.00000807, 0.00000964, 0.00000744, 0.00000255, -0.00000211, -0.00000359, -0.00000107, 0.00000337, 0.00000601, 0.00000426, -0.00000119, -0.00000646, -0.00000703, -0.00000107, 0.00000882, 0.00001698, 0.00001828, 0.00001146, 0.00000005, -0.00001003, -0.00001422, -0.00001189, -0.00000616, -0.00000118, 0.00000072, 0.00000016, -0.00000062, -0.00000002, 0.00000165, 0.00000276, 0.00000211, 0.00000013, -0.00000163, -0.00000211, -0.00000183, -0.00000236, -0.00000443, -0.00000649, -0.00000550, 0.00000043, 0.00000949, 0.00001622, 0.00001493, 0.00000417, -0.00001110, -0.00002191, -0.00002126, -0.00000934, 0.00000588, 0.00001385, 0.00000900, -0.00000482, -0.00001668, -0.00001633, -0.00000209, 0.00001711, 0.00002755, 0.00002048, -0.00000099, -0.00002321, -0.00003096, -0.00001800, 0.00000825, 0.00003134, 0.00003669, 0.00002128, -0.00000467, -0.00002484, -0.00002791, -0.00001461, 0.00000368, 0.00001357, 0.00000902, -0.00000514, -0.00001751, -0.00001884, -0.00000857, 0.00000531, 0.00001252, 0.00000806, -0.00000446, -0.00001580, -0.00001791, -0.00000951, 0.00000342, 0.00001218, 0.00001136, 0.00000209, -0.00000936, -0.00001612, -0.00001498, -0.00000766, 0.00000129, 0.00000772, 0.00000994, 0.00000857, 0.00000498, 0.00000006, -0.00000568, -0.00001127, -0.00001476, -0.00001401, -0.00000837, 0.00000028, 0.00000802, 0.00001104, 0.00000793, 0.00000066, -0.00000655, -0.00000962, -0.00000685, 0.00000041, 0.00000881, 0.00001491, 0.00001674, 0.00001429, 0.00000905, 0.00000303, -0.00000206, },
		{ 0.00001334, 0.00000981, 0.00000500, 0.00000030, -0.00000333, -0.00000566, -0.00000709, -0.00000809, -0.00000852, -0.00000752, -0.00000418, 0.00000136, 0.00000734, 0.00001094, 0.00001003, 0.00000487, -0.00000162, -0.00000545, -0.00000425, 0.00000095, 0.00000621, 0.00000741, 0.00000330, -0.00000331, -0.00000732, -0.00000493, 0.00000341, 0.00001304, 0.00001800, 0.00001496, 0.00000542, -0.00000550, -0.00001242, -0.00001307, -0.00000911, -0.00000428, -0.00000140, -0.00000074, -0.00000062, 0.00000048, 0.00000238, 0.00000346, 0.00000243, -0.00000015, -0.00000221, -0.00000203, -0.00000005, 0.00000122, -0.00000057, -0.00000507, -0.00000867, -0.00000700, 0.00000129, 0.00001240, 0.00001915, 0.00001583, 0.00000278, -0.00001298, -0.00002209, -0.00001927, -0.00000719, 0.00000521, 0.00000919, 0.00000265, -0.00000816, -0.00001332, -0.00000692, 0.00000795, 0.00002090, 0.00002164, 0.00000795, -0.00001200, -0.00002481, -0.00002115, -0.00000257, 0.00001918, 0.00003009, 0.00002337, 0.00000396, -0.00001509, -0.00002187, -0.00001363, 0.00000183, 0.00001221, 0.00000969, -0.00000376, -0.00001818, -0.00002310, -0.00001501, 0.00000038, 0.00001249, 0.00001339, 0.00000310, -0.00001090, -0.00001912, -0.00001655, -0.00000549, 0.00000663, 0.00001254, 0.00000960, 0.00000071, -0.00000843, -0.00001321, -0.00001236, -0.00000768, -0.00000191, 0.00000305, 0.00000650, 0.00000813, 0.00000733, 0.00000345, -0.00000316, -0.00001037, -0.00001497, -0.00001439, -0.00000853, -0.00000006, 0.00000688, 0.00000910, 0.00000606, 0.00000003, -0.00000543, -0.00000740, -0.00000501, 0.00000059, 0.00000707, 0.00001213, 0.00001433, },
		{ -0.00000001, 0.00000495, 0.00000950, 0.00001225, 0.00001247, 0.00001028, 0.00000660, 0.00000259, -0.00000103, -0.00000419, -0.00000717, -0.00000975, -0.00001083, -0.00000891, -0.00000340, 0.00000426, 0.00001077, 0.00001269, 0.00000877, 0.00000116, -0.00000560, -0.00000738, -0.00000334, 0.00000344, 0.00000800, 0.00000705, 0.00000142, -0.00000443, -0.00000558, -0.00000030, 0.00000854, 0.00001531, 0.00001546, 0.00000860, -0.00000146, -0.00000943, -0.00001209, -0.00000994, -0.00000611, -0.00000352, -0.00000274, -0.00000224, -0.00000036, 0.00000270, 0.00000486, 0.00000409, 0.00000051, -0.00000324, -0.00000406, -0.00000110, 0.00000309, 0.00000437, 0.00000055, -0.00000624, -0.00001055, -0.00000754, 0.00000279, 0.00001475, 0.00002040, 0.00001511, 0.00000119, -0.00001335, -0.00002016, -0.00001635, -0.00000628, 0.00000197, 0.00000277, -0.00000279, -0.00000800, -0.00000618, 0.00000357, 0.00001502, 0.00001921, 0.00001145, -0.00000443, -0.00001810, -0.00001983, -0.00000765, 0.00001082, 0.00002337, 0.00002170, 0.00000720, -0.00001002, -0.00001831, -0.00001271, 0.00000176, 0.00001377, 0.00001391, 0.00000132, -0.00001583, -0.00002614, -0.00002292, -0.00000853, 0.00000749, 0.00001514, 0.00001044, -0.00000245, -0.00001467, -0.00001872, -0.00001300, -0.00000194, 0.00000756, 0.00001073, 0.00000722, 0.00000024, -0.00000620, -0.00000974, -0.00001015, -0.00000835, -0.00000511, -0.00000081, 0.00000394, 0.00000770, 0.00000848, 0.00000510, -0.00000166, -0.00000898, -0.00001339, -0.00001278, -0.00000764, -0.00000071, 0.00000464, 0.00000624, 0.00000409, -0.00000005, -0.00000378, -0.00000525, -0.00000383, },
		{ 0.00000022, -0.00000241, -0.00000398, -0.00000358, -0.00000101, 0.00000310, 0.00000742, 0.00001048, 0.00001142, 0.00001031, 0.00000792, 0.00000499, 0.00000161, -0.00000255, -0.00000738, -0.00001157, -0.00001287, -0.00000944, -0.00000164, 0.00000742, 0.00001324, 0.00001259, 0.00000587, -0.00000286, -0.00000830, -0.00000738, -0.00000131, 0.00000533, 0.00000792, 0.00000510, -0.00000028, -0.00000332, -0.00000089, 0.00000596, 0.00001264, 0.00001430, 0.00000934, 0.00000043, -0.00000748, -0.00001073, -0.00000922, -0.00000588, -0.00000389, -0.00000408, -0.00000460, -0.00000299, 0.00000116, 0.00000547, 0.00000658, 0.00000313, -0.00000265, -0.00000633, -0.00000477, 0.00000111, 0.00000665, 0.00000696, 0.00000089, -0.00000749, -0.00001153, -0.00000700, 0.00000444, 0.00001590, 0.00001984, 0.00001336, 0.00000020, -0.00001193, -0.00001679, -0.00001371, -0.00000730, -0.00000325, -0.00000366, -0.00000562, -0.00000413, 0.00000299, 0.00001241, 0.00001730, 0.00001263, -0.00000023, -0.00001347, -0.00001798, -0.00000996, 0.00000583, 0.00001892, 0.00002031, 0.00000888, -0.00000758, -0.00001772, -0.00001456, -0.00000049, 0.00001437, 0.00001914, 0.00000970, -0.00000851, -0.00002430, -0.00002816, -0.00001844, -0.00000190, 0.00001114, 0.00001354, 0.00000540, -0.00000686, -0.00001533, -0.00001560, -0.00000877, 0.00000029, 0.00000660, 0.00000793, 0.00000515, 0.00000068, -0.00000350, -0.00000662, -0.00000861, -0.00000913, -0.00000753, -0.00000350, 0.00000198, 0.00000666, 0.00000812, 0.00000535, -0.00000051, -0.00000668, -0.00001032, -0.00001005, -0.00000652, -0.00000177, 0.00000201, 0.00000351, 0.00000262, },
		{ -0.00000281, -0.00000008, 0.00000163, 0.00000173, 0.00000024, -0.00000208, -0.00000392, -0.00000402, -0.00000187, 0.00000190, 0.00000590, 0.00000877, 0.00001002, 0.00001000, 0.00000923, 0.00000759, 0.00000428, -0.00000122, -0.00000796, -0.00001328, -0.00001399, -0.00000863, 0.00000094, 0.00001006, 0.00001386, 0.00001036, 0.00000195, -0.00000607, -0.00000895, -0.00000561, 0.00000085, 0.00000561, 0.00000578, 0.00000242, -0.00000046, 0.00000073, 0.00000617, 0.00001221, 0.00001408, 0.00000951, 0.00000058, -0.00000776, -0.00001121, -0.00000916, -0.00000475, -0.00000212, -0.00000305, -0.00000562, -0.00000613, -0.00000248, 0.00000368, 0.00000799, 0.00000686, 0.00000061, -0.00000632, -0.00000858, -0.00000408, 0.00000413, 0.00000991, 0.00000857, 0.00000065, -0.00000824, -0.00001133, -0.00000566, 0.00000566, 0.00001554, 0.00001782, 0.00001135, 0.00000034, -0.00000906, -0.00001308, -0.00001233, -0.00001016, -0.00000903, -0.00000826, -0.00000500, 0.00000241, 0.00001170, 0.00001734, 0.00001445, 0.00000319, -0.00001021, -0.00001706, -0.00001222, 0.00000175, 0.00001578, 0.00002022, 0.00001161, -0.00000459, -0.00001751, -0.00001824, -0.00000602, 0.00001105, 0.00002144, 0.00001785, 0.00000197, -0.00001679, -0.00002760, -0.00002501, -0.00001195, 0.00000290, 0.00001098, 0.00000902, 0.00000011, -0.00000918, -0.00001341, -0.00001120, -0.00000506, 0.00000109, 0.00000461, 0.00000514, 0.00000377, 0.00000156, -0.00000113, -0.00000432, -0.00000753, -0.00000947, -0.00000880, -0.00000517, 0.00000013, 0.00000474, 0.00000654, 0.00000488, 0.00000081, -0.00000363, -0.00000657, -0.00000713, -0.00000556, },
		{ -0.00000047, -0.00000310, -0.00000462, -0.00000458, -0.00000307, -0.00000087, 0.00000083, 0.00000097, -0.00000066, -0.00000315, -0.00000491, -0.00000464, -0.00000221, 0.00000134, 0.00000464, 0.00000691, 0.00000842, 0.00000974, 0.00001074, 0.00001016, 0.00000645, -0.00000062, -0.00000883, -0.00001425, -0.00001353, -0.00000637, 0.00000370, 0.00001119, 0.00001206, 0.00000633, -0.00000194, -0.00000746, -0.00000736, -0.00000285, 0.00000202, 0.00000377, 0.00000220, 0.00000039, 0.00000192, 0.00000757, 0.00001414, 0.00001658, 0.00001183, 0.00000145, -0.00000911, -0.00001420, -0.00001191, -0.00000531, 0.00000005, 0.00000038, -0.00000358, -0.00000736, -0.00000658, -0.00000065, 0.00000650, 0.00000941, 0.00000542, -0.00000293, -0.00000961, -0.00000944, -0.00000216, 0.00000730, 0.00001226, 0.00000919, 0.00000029, -0.00000809, -0.00001000, -0.00000403, 0.00000597, 0.00001376, 0.00001502, 0.00000981, 0.00000163, -0.00000559, -0.00001012, -0.00001258, -0.00001394, -0.00001365, -0.00000976, -0.00000127, 0.00000961, 0.00001754, 0.00001725, 0.00000766, -0.00000626, -0.00001595, -0.00001482, -0.00000306, 0.00001189, 0.00001995, 0.00001526, 0.00000041, -0.00001504, -0.00002079, -0.00001280, 0.00000397, 0.00001875, 0.00002202, 0.00001158, -0.00000634, -0.00002139, -0.00002574, -0.00001855, -0.00000565, 0.00000488, 0.00000780, 0.00000334, -0.00000407, -0.00000937, -0.00001002, -0.00000682, -0.00000243, 0.00000091, 0.00000252, 0.00000300, 0.00000296, 0.00000229, 0.00000041, -0.00000283, -0.00000653, -0.00000906, -0.00000902, -0.00000624, -0.00000190, 0.00000216, 0.00000442, 0.00000435, 0.00000238, },
		{ -0.00000045, 0.00000250, 0.00000403, 0.00000376, 0.00000193, -0.00000058, -0.00000257, -0.00000303, -0.00000189, -0.00000013, 0.00000074, -0.00000026, -0.00000282, -0.00000543, -0.00000638, -0.00000500, -0.00000202, 0.00000107, 0.00000330, 0.00000495, 0.00000701, 0.00000990, 0.00001243, 0.00001223, 0.00000755, -0.00000088, -0.00000949, -0.00001375, -0.00001113, -0.00000307, 0.00000568, 0.00001002, 0.00000787, 0.00000142, -0.00000463, -0.00000654, -0.00000409, -0.00000030, 0.00000135, 0.00000005, -0.00000155, 0.00000036, 0.00000721, 0.00001607, 0.00002106, 0.00001756, 0.00000584, -0.00000844, -0.00001773, -0.00001761, -0.00000971, -0.00000039, 0.00000392, 0.00000121, -0.00000493, -0.00000843, -0.00000557, 0.00000205, 0.00000875, 0.00000918, 0.00000252, -0.00000667, -0.00001174, -0.00000878, 0.00000048, 0.00000993, 0.00001344, 0.00000909, 0.00000028, -0.00000690, -0.00000792, -0.00000264, 0.00000524, 0.00001106, 0.00001222, 0.00000905, 0.00000354, -0.00000258, -0.00000863, -0.00001404, -0.00001719, -0.00001562, -0.00000788, 0.00000417, 0.00001517, 0.00001891, 0.00001254, -0.00000077, -0.00001302, -0.00001628, -0.00000810, 0.00000646, 0.00001775, 0.00001793, 0.00000635, -0.00000977, -0.00002004, -0.00001769, -0.00000411, 0.00001212, 0.00002089, 0.00001696, 0.00000296, -0.00001272, -0.00002157, -0.00002006, -0.00001104, -0.00000107, 0.00000417, 0.00000304, -0.00000191, -0.00000655, -0.00000806, -0.00000637, -0.00000331, -0.00000083, 0.00000042, 0.00000097, 0.00000161, 0.00000236, 0.00000246, 0.00000108, -0.00000184, -0.00000534, -0.00000798, -0.00000864, -0.00000711, -0.00000402, },
		{ -0.00000666, -0.00000814, -0.00000786, -0.00000576, -0.00000237, 0.00000122, 0.00000373, 0.00000427, 0.00000293, 0.00000084, -0.00000052, -0.00000032, 0.00000096, 0.00000177, 0.00000069, -0.00000242, -0.00000610, -0.00000829, -0.00000774, -0.00000494, -0.00000165, 0.00000060, 0.00000177, 0.00000321, 0.00000628, 0.00001064, 0.00001392, 0.00001316, 0.00000725, -0.00000170, -0.00000919, -0.00001121, -0.00000699, 0.00000035, 0.00000587, 0.00000625, 0.00000202, -0.00000311, -0.00000529, -0.00000352, -0.00000025, 0.00000093, -0.00000131, -0.00000448, -0.00000394, 0.00000316, 0.00001477, 0.00002437, 0.00002505, 0.00001447, -0.00000289, -0.00001814, -0.00002357, -0.00001759, -0.00000553, 0.00000423, 0.00000605, 0.00000060, -0.00000625, -0.00000818, -0.00000326, 0.00000482, 0.00000967, 0.00000722, -0.00000120, -0.00000973, -0.00001229, -0.00000687, 0.00000316, 0.00001156, 0.00001353, 0.00000870, 0.00000090, -0.00000487, -0.00000563, -0.00000185, 0.00000371, 0.00000814, 0.00000995, 0.00000892, 0.00000524, -0.00000088, -0.00000862, -0.00001566, -0.00001846, -0.00001424, -0.00000345, 0.00000929, 0.00001721, 0.00001551, 0.00000490, -0.00000819, -0.00001534, -0.00001170, 0.00000059, 0.00001344, 0.00001814, 0.00001117, -0.00000330, -0.00001607, -0.00001897, -0.00001020, 0.00000459, 0.00001612, 0.00001753, 0.00000845, -0.00000528, -0.00001574, -0.00001798, -0.00001258, -0.00000436, 0.00000127, 0.00000174, -0.00000174, -0.00000576, -0.00000746, -0.00000619, -0.00000336, -0.00000092, 0.00000013, 0.00000015, 0.00000015, 0.00000074, 0.00000166, 0.00000205, 0.00000116, -0.00000106, -0.00000399, },
		{ 0.00000132, 0.00000105, -0.00000038, -0.00000282, -0.00000563, -0.00000778, -0.00000823, -0.00000652, -0.00000314, 0.00000052, 0.00000292, 0.00000334, 0.00000239, 0.00000159, 0.00000211, 0.00000381, 0.00000507, 0.00000402, 0.00000005, -0.00000545, -0.00000982, -0.00001093, -0.00000861, -0.00000474, -0.00000167, -0.00000038, 0.00000024, 0.00000224, 0.00000657, 0.00001175, 0.00001458, 0.00001246, 0.00000568, -0.00000235, -0.00000721, -0.00000658, -0.00000188, 0.00000276, 0.00000365, 0.00000032, -0.00000424, -0.00000611, -0.00000372, 0.00000067, 0.00000283, 0.00000021, -0.00000532, -0.00000831, -0.00000370, 0.00000872, 0.00002304, 0.00003041, 0.00002487, 0.00000776, -0.00001255, -0.00002568, -0.00002570, -0.00001466, -0.00000075, 0.00000740, 0.00000630, -0.00000072, -0.00000671, -0.00000647, -0.00000027, 0.00000678, 0.00000887, 0.00000394, -0.00000484, -0.00001144, -0.00001128, -0.00000430, 0.00000529, 0.00001206, 0.00001288, 0.00000845, 0.00000215, -0.00000244, -0.00000360, -0.00000168, 0.00000189, 0.00000564, 0.00000838, 0.00000886, 0.00000589, -0.00000080, -0.00000941, -0.00001611, -0.00001678, -0.00000985, 0.00000186, 0.00001213, 0.00001492, 0.00000842, -0.00000328, -0.00001251, -0.00001298, -0.00000410, 0.00000839, 0.00001611, 0.00001367, 0.00000235, -0.00001068, -0.00001713, -0.00001303, -0.00000126, 0.00001060, 0.00001522, 0.00001021, -0.00000086, -0.00001119, -0.00001529, -0.00001219, -0.00000528, 0.00000044, 0.00000174, -0.00000115, -0.00000542, -0.00000795, -0.00000739, -0.00000454, -0.00000137, 0.00000054, 0.00000085, 0.00000030, -0.00000012, 0.00000010, 0.00000079, },
		{ 0.00000067, -0.00000018, -0.00000044, -0.00000000, 0.00000071, 0.00000094, -0.00000001, -0.00000227, -0.00000518, -0.00000740, -0.00000771, -0.00000584, -0.00000276, -0.00000009, 0.00000101, 0.00000081, 0.00000078, 0.00000240, 0.00000575, 0.00000901, 0.00000953, 0.00000575, -0.00000140, -0.00000881, -0.00001308, -0.00001274, -0.00000906, -0.00000492, -0.00000253, -0.00000184, -0.00000082, 0.00000245, 0.00000780, 0.00001265, 0.00001381, 0.00001008, 0.00000349, -0.00000195, -0.00000317, -0.00000041, 0.00000299, 0.00000318, -0.00000094, -0.00000668, -0.00000955, -0.00000693, -0.00000049, 0.00000471, 0.00000413, -0.00000230, -0.00000936, -0.00000992, -0.00000036, 0.00001596, 0.00003000, 0.00003252, 0.00002034, -0.00000115, -0.00002105, -0.00002964, -0.00002416, -0.00001002, 0.00000330, 0.00000860, 0.00000515, -0.00000187, -0.00000585, -0.00000364, 0.00000252, 0.00000730, 0.00000648, 0.00000013, -0.00000760, -0.00001155, -0.00000913, -0.00000173, 0.00000651, 0.00001161, 0.00001193, 0.00000852, 0.00000379, -0.00000011, -0.00000208, -0.00000186, 0.00000036, 0.00000388, 0.00000721, 0.00000818, 0.00000505, -0.00000195, -0.00000989, -0.00001433, -0.00001209, -0.00000374, 0.00000612, 0.00001144, 0.00000873, -0.00000035, -0.00000971, -0.00001281, -0.00000712, 0.00000393, 0.00001313, 0.00001425, 0.00000635, -0.00000555, -0.00001384, -0.00001333, -0.00000468, 0.00000627, 0.00001247, 0.00001022, 0.00000127, -0.00000869, -0.00001396, -0.00001234, -0.00000599, 0.00000047, 0.00000317, 0.00000119, -0.00000344, -0.00000751, -0.00000875, -0.00000693, -0.00000348, -0.00000031, 0.00000139, 0.00000149, },
		{ -0.00000296, 0.00000014, 0.00000183, 0.00000191, 0.00000091, -0.00000025, -0.00000076, -0.00000036, 0.00000048, 0.00000076, -0.00000031, -0.00000266, -0.00000523, -0.00000656, -0.00000588, -0.00000376, -0.00000177, -0.00000129, -0.00000221, -0.00000283, -0.00000100, 0.00000396, 0.00001028, 0.00001440, 0.00001324, 0.00000640, -0.00000334, -0.00001159, -0.00001507, -0.00001356, -0.00000953, -0.00000604, -0.00000437, -0.00000336, -0.00000087, 0.00000397, 0.00000949, 0.00001261, 0.00001131, 0.00000649, 0.00000164, 0.00000018, 0.00000268, 0.00000618, 0.00000638, 0.00000120, -0.00000714, -0.00001332, -0.00001284, -0.00000566, 0.00000326, 0.00000731, 0.00000332, -0.00000564, -0.00001184, -0.00000837, 0.00000555, 0.00002309, 0.00003372, 0.00002996, 0.00001243, -0.00001024, -0.00002665, -0.00002956, -0.00001984, -0.00000511, 0.00000571, 0.00000799, 0.00000347, -0.00000211, -0.00000373, -0.00000049, 0.00000429, 0.00000618, 0.00000306, -0.00000334, -0.00000895, -0.00001024, -0.00000644, 0.00000033, 0.00000683, 0.00001061, 0.00001104, 0.00000889, 0.00000542, 0.00000176, -0.00000103, -0.00000198, -0.00000052, 0.00000277, 0.00000591, 0.00000638, 0.00000285, -0.00000343, -0.00000892, -0.00000993, -0.00000530, 0.00000229, 0.00000778, 0.00000710, 0.00000021, -0.00000850, -0.00001294, -0.00000955, 0.00000009, 0.00001004, 0.00001389, 0.00000913, -0.00000114, -0.00001025, -0.00001240, -0.00000648, 0.00000332, 0.00001044, 0.00001026, 0.00000295, -0.00000697, -0.00001372, -0.00001392, -0.00000833, -0.00000087, 0.00000410, 0.00000431, 0.00000051, -0.00000458, -0.00000811, -0.00000863, -0.00000643, },
		{ -0.00000455, -0.00000759, -0.00000791, -0.00000587, -0.00000268, 0.00000022, 0.00000181, 0.00000181, 0.00000077, -0.00000029, -0.00000056, 0.00000000, 0.00000060, 0.00000020, -0.00000155, -0.00000388, -0.00000530, -0.00000478, -0.00000270, -0.00000083, -0.00000101, -0.00000352, -0.00000642, -0.00000660, -0.00000199, 0.00000649, 0.00001498, 0.00001878, 0.00001534, 0.00000596, -0.00000501, -0.00001289, -0.00001542, -0.00001369, -0.00001061, -0.00000836, -0.00000685, -0.00000435, 0.00000039, 0.00000639, 0.00001080, 0.00001107, 0.00000731, 0.00000267, 0.00000112, 0.00000421, 0.00000945, 0.00001179, 0.00000742, -0.00000287, -0.00001364, -0.00001825, -0.00001370, -0.00000311, 0.00000616, 0.00000772, 0.00000090, -0.00000836, -0.00001153, -0.00000366, 0.00001258, 0.00002809, 0.00003300, 0.00002316, 0.00000291, -0.00001752, -0.00002837, -0.00002597, -0.00001418, -0.00000114, 0.00000627, 0.00000629, 0.00000215, -0.00000119, -0.00000090, 0.00000211, 0.00000456, 0.00000372, -0.00000053, -0.00000574, -0.00000879, -0.00000798, -0.00000385, 0.00000165, 0.00000646, 0.00000946, 0.00001037, 0.00000933, 0.00000667, 0.00000307, -0.00000021, -0.00000169, -0.00000070, 0.00000191, 0.00000395, 0.00000340, 0.00000002, -0.00000412, -0.00000585, -0.00000338, 0.00000203, 0.00000651, 0.00000617, 0.00000009, -0.00000856, -0.00001423, -0.00001273, -0.00000427, 0.00000637, 0.00001280, 0.00001122, 0.00000297, -0.00000640, -0.00001076, -0.00000748, 0.00000095, 0.00000875, 0.00001065, 0.00000522, -0.00000444, -0.00001288, -0.00001573, -0.00001210, -0.00000464, 0.00000242, 0.00000577, 0.00000454, 0.00000023, },
		{ 0.00000383, 0.00000600, 0.00000436, 0.00000044, -0.00000370, -0.00000633, -0.00000668, -0.00000501, -0.00000235, -0.00000002, 0.00000109, 0.00000092, 0.00000022, -0.00000007, 0.00000040, 0.00000104, 0.00000075, -0.00000105, -0.00000367, -0.00000542, -0.00000478, -0.00000186, 0.00000131, 0.00000197, -0.00000122, -0.00000673, -0.00001070, -0.00000932, -0.00000157, 0.00000960, 0.00001873, 0.00002109, 0.00001553, 0.00000501, -0.00000550, -0.00001219, -0.00001431, -0.00001375, -0.00001269, -0.00001164, -0.00000929, -0.00000428, 0.00000277, 0.00000893, 0.00001094, 0.00000791, 0.00000261, -0.00000025, 0.00000251, 0.00000965, 0.00001583, 0.00001522, 0.00000593, -0.00000805, -0.00001899, -0.00002063, -0.00001252, -0.00000053, 0.00000718, 0.00000598, -0.00000195, -0.00000907, -0.00000795, 0.00000318, 0.00001879, 0.00002938, 0.00002771, 0.00001358, -0.00000609, -0.00002158, -0.00002629, -0.00002012, -0.00000862, 0.00000125, 0.00000542, 0.00000443, 0.00000176, 0.00000064, 0.00000179, 0.00000345, 0.00000333, 0.00000061, -0.00000349, -0.00000672, -0.00000741, -0.00000542, -0.00000176, 0.00000225, 0.00000578, 0.00000844, 0.00000985, 0.00000955, 0.00000735, 0.00000392, 0.00000070, -0.00000088, -0.00000049, 0.00000077, 0.00000114, -0.00000023, -0.00000240, -0.00000312, -0.00000074, 0.00000407, 0.00000815, 0.00000784, 0.00000170, -0.00000786, -0.00001562, -0.00001673, -0.00001000, 0.00000099, 0.00001002, 0.00001214, 0.00000678, -0.00000194, -0.00000802, -0.00000755, -0.00000113, 0.00000672, 0.00001057, 0.00000761, -0.00000081, -0.00001021, -0.00001585, -0.00001537, -0.00000970, -0.00000211, },
		{ -0.00001619, -0.00001363, -0.00000759, -0.00000096, 0.00000375, 0.00000532, 0.00000396, 0.00000087, -0.00000239, -0.00000447, -0.00000476, -0.00000358, -0.00000191, -0.00000076, -0.00000054, -0.00000074, -0.00000046, 0.00000070, 0.00000211, 0.00000237, 0.00000048, -0.00000306, -0.00000618, -0.00000648, -0.00000318, 0.00000190, 0.00000517, 0.00000362, -0.00000270, -0.00001024, -0.00001384, -0.00001010, 0.00000021, 0.00001235, 0.00002039, 0.00002079, 0.00001418, 0.00000445, -0.00000414, -0.00000955, -0.00001241, -0.00001434, -0.00001575, -0.00001523, -0.00001095, -0.00000301, 0.00000562, 0.00001060, 0.00000941, 0.00000363, -0.00000165, -0.00000132, 0.00000577, 0.00001539, 0.00002043, 0.00001583, 0.00000246, -0.00001295, -0.00002206, -0.00002040, -0.00001029, 0.00000089, 0.00000597, 0.00000295, -0.00000379, -0.00000691, -0.00000168, 0.00001040, 0.00002234, 0.00002631, 0.00001888, 0.00000327, -0.00001280, -0.00002195, -0.00002133, -0.00001355, -0.00000418, 0.00000212, 0.00000400, 0.00000321, 0.00000236, 0.00000272, 0.00000354, 0.00000327, 0.00000108, -0.00000231, -0.00000524, -0.00000633, -0.00000537, -0.00000308, -0.00000035, 0.00000239, 0.00000511, 0.00000763, 0.00000931, 0.00000935, 0.00000752, 0.00000457, 0.00000184, 0.00000024, -0.00000038, -0.00000098, -0.00000226, -0.00000362, -0.00000338, -0.00000014, 0.00000557, 0.00001080, 0.00001170, 0.00000613, -0.00000437, -0.00001487, -0.00001978, -0.00001632, -0.00000638, 0.00000448, 0.00001048, 0.00000914, 0.00000260, -0.00000399, -0.00000599, -0.00000223, 0.00000450, 0.00000942, 0.00000896, 0.00000280, -0.00000610, -0.00001348, },
		{ 0.00000508, -0.00000213, -0.00000957, -0.00001425, -0.00001471, -0.00001143, -0.00000615, -0.00000092, 0.00000269, 0.00000399, 0.00000319, 0.00000119, -0.00000083, -0.00000194, -0.00000199, -0.00000156, -0.00000153, -0.00000222, -0.00000304, -0.00000283, -0.00000090, 0.00000208, 0.00000416, 0.00000340, -0.00000049, -0.00000548, -0.00000822, -0.00000636, -0.00000052, 0.00000566, 0.00000773, 0.00000349, -0.00000512, -0.00001294, -0.00001475, -0.00000866, 0.00000264, 0.00001359, 0.00001925, 0.00001809, 0.00001218, 0.00000510, -0.00000084, -0.00000564, -0.00001054, -0.00001573, -0.00001926, -0.00001821, -0.00001132, -0.00000091, 0.00000795, 0.00001060, 0.00000630, -0.00000080, -0.00000431, -0.00000013, 0.00001019, 0.00002009, 0.00002227, 0.00001374, -0.00000182, -0.00001630, -0.00002240, -0.00001824, -0.00000822, 0.00000048, 0.00000300, -0.00000003, -0.00000348, -0.00000197, 0.00000585, 0.00001604, 0.00002208, 0.00001935, 0.00000835, -0.00000571, -0.00001619, -0.00001912, -0.00001496, -0.00000759, -0.00000122, 0.00000207, 0.00000286, 0.00000296, 0.00000354, 0.00000424, 0.00000388, 0.00000178, -0.00000146, -0.00000436, -0.00000561, -0.00000498, -0.00000324, -0.00000129, 0.00000048, 0.00000236, 0.00000460, 0.00000695, 0.00000857, 0.00000872, 0.00000737, 0.00000523, 0.00000312, 0.00000123, -0.00000083, -0.00000333, -0.00000555, -0.00000575, -0.00000236, 0.00000439, 0.00001172, 0.00001537, 0.00001211, 0.00000215, -0.00001039, -0.00001953, -0.00002087, -0.00001418, -0.00000352, 0.00000534, 0.00000837, 0.00000555, 0.00000038, -0.00000274, -0.00000145, 0.00000322, 0.00000782, 0.00000887, },
		{ 0.00000380, 0.00000710, 0.00000828, 0.00000589, 0.00000046, -0.00000594, -0.00001097, -0.00001316, -0.00001232, -0.00000926, -0.00000526, -0.00000155, 0.00000097, 0.00000198, 0.00000181, 0.00000124, 0.00000104, 0.00000138, 0.00000161, 0.00000077, -0.00000153, -0.00000442, -0.00000604, -0.00000481, -0.00000081, 0.00000379, 0.00000589, 0.00000361, -0.00000206, -0.00000753, -0.00000890, -0.00000473, 0.00000262, 0.00000829, 0.00000812, 0.00000162, -0.00000751, -0.00001366, -0.00001291, -0.00000550, 0.00000458, 0.00001243, 0.00001532, 0.00001386, 0.00001057, 0.00000735, 0.00000389, -0.00000145, -0.00000933, -0.00001768, -0.00002232, -0.00001980, -0.00001043, 0.00000120, 0.00000880, 0.00000864, 0.00000230, -0.00000421, -0.00000463, 0.00000297, 0.00001454, 0.00002255, 0.00002106, 0.00000980, -0.00000559, -0.00001730, -0.00002033, -0.00001530, -0.00000726, -0.00000169, -0.00000060, -0.00000165, -0.00000064, 0.00000473, 0.00001270, 0.00001855, 0.00001791, 0.00000999, -0.00000177, -0.00001189, -0.00001626, -0.00001429, -0.00000865, -0.00000297, 0.00000053, 0.00000191, 0.00000256, 0.00000354, 0.00000459, 0.00000461, 0.00000281, -0.00000039, -0.00000352, -0.00000515, -0.00000481, -0.00000321, -0.00000144, -0.00000011, 0.00000098, 0.00000237, 0.00000428, 0.00000625, 0.00000757, 0.00000781, 0.00000714, 0.00000593, 0.00000421, 0.00000163, -0.00000199, -0.00000582, -0.00000784, -0.00000589, 0.00000059, 0.00000944, 0.00001631, 0.00001691, 0.00000971, -0.00000273, -0.00001490, -0.00002133, -0.00001966, -0.00001179, -0.00000245, 0.00000378, 0.00000511, 0.00000304, 0.00000084, 0.00000104, },
		{ 0.00000231, 0.00000259, 0.00000360, 0.00000580, 0.00000803, 0.00000857, 0.00000637, 0.00000182, -0.00000358, -0.00000812, -0.00001069, -0.00001104, -0.00000962, -0.00000720, -0.00000461, -0.00000255, -0.00000136, -0.00000083, -0.00000024, 0.00000116, 0.00000341, 0.00000547, 0.00000568, 0.00000295, -0.00000212, -0.00000706, -0.00000890, -0.00000618, -0.00000025, 0.00000522, 0.00000659, 0.00000281, -0.00000365, -0.00000829, -0.00000767, -0.00000190, 0.00000521, 0.00000876, 0.00000610, -0.00000125, -0.00000869, -0.00001166, -0.00000861, -0.00000176, 0.00000491, 0.00000860, 0.00000938, 0.00000935, 0.00001014, 0.00001096, 0.00000902, 0.00000207, -0.00000894, -0.00001949, -0.00002399, -0.00001966, -0.00000880, 0.00000232, 0.00000757, 0.00000508, -0.00000150, -0.00000561, -0.00000256, 0.00000704, 0.00001755, 0.00002212, 0.00001731, 0.00000529, -0.00000779, -0.00001590, -0.00001680, -0.00001267, -0.00000775, -0.00000481, -0.00000345, -0.00000109, 0.00000419, 0.00001147, 0.00001711, 0.00001728, 0.00001080, 0.00000021, -0.00000969, -0.00001474, -0.00001380, -0.00000890, -0.00000341, 0.00000023, 0.00000172, 0.00000230, 0.00000317, 0.00000437, 0.00000485, 0.00000362, 0.00000076, -0.00000249, -0.00000458, -0.00000472, -0.00000334, -0.00000153, -0.00000018, 0.00000062, 0.00000137, 0.00000251, 0.00000401, 0.00000543, 0.00000640, 0.00000688, 0.00000698, 0.00000647, 0.00000470, 0.00000115, -0.00000366, -0.00000771, -0.00000833, -0.00000386, 0.00000474, 0.00001373, 0.00001833, 0.00001542, 0.00000556, -0.00000712, -0.00001714, -0.00002067, -0.00001741, -0.00001030, -0.00000331, 0.00000092, },
		{ -0.00001532, -0.00001011, -0.00000531, -0.00000176, 0.00000111, 0.00000417, 0.00000742, 0.00000985, 0.00001017, 0.00000775, 0.00000321, -0.00000200, -0.00000632, -0.00000883, -0.00000942, -0.00000850, -0.00000680, -0.00000506, -0.00000398, -0.00000388, -0.00000441, -0.00000445, -0.00000271, 0.00000127, 0.00000635, 0.00000997, 0.00000959, 0.00000450, -0.00000321, -0.00000952, -0.00001091, -0.00000669, 0.00000034, 0.00000562, 0.00000588, 0.00000134, -0.00000441, -0.00000705, -0.00000457, 0.00000123, 0.00000614, 0.00000658, 0.00000222, -0.00000389, -0.00000767, -0.00000696, -0.00000292, 0.00000122, 0.00000298, 0.00000263, 0.00000275, 0.00000572, 0.00001105, 0.00001509, 0.00001340, 0.00000435, -0.00000900, -0.00002026, -0.00002369, -0.00001802, -0.00000734, 0.00000167, 0.00000429, 0.00000082, -0.00000400, -0.00000459, 0.00000124, 0.00001078, 0.00001826, 0.00001893, 0.00001212, 0.00000143, -0.00000797, -0.00001274, -0.00001296, -0.00001107, -0.00000926, -0.00000763, -0.00000449, 0.00000158, 0.00000969, 0.00001644, 0.00001799, 0.00001267, 0.00000244, -0.00000809, -0.00001442, -0.00001462, -0.00001008, -0.00000412, 0.00000037, 0.00000245, 0.00000302, 0.00000345, 0.00000426, 0.00000476, 0.00000395, 0.00000156, -0.00000152, -0.00000386, -0.00000446, -0.00000344, -0.00000171, -0.00000026, 0.00000059, 0.00000112, 0.00000177, 0.00000269, 0.00000367, 0.00000450, 0.00000525, 0.00000610, 0.00000681, 0.00000653, 0.00000425, -0.00000015, -0.00000524, -0.00000825, -0.00000669, -0.00000014, 0.00000895, 0.00001616, 0.00001748, 0.00001167, 0.00000099, -0.00001003, -0.00001716, -0.00001857, },
		{ -0.00000290, -0.00001122, -0.00001573, -0.00001621, -0.00001403, -0.00001077, -0.00000716, -0.00000304, 0.00000180, 0.00000684, 0.00001071, 0.00001197, 0.00001002, 0.00000553, 0.00000006, -0.00000471, -0.00000769, -0.00000860, -0.00000774, -0.00000582, -0.00000379, -0.00000273, -0.00000343, -0.00000573, -0.00000820, -0.00000857, -0.00000511, 0.00000189, 0.00000967, 0.00001422, 0.00001264, 0.00000523, -0.00000434, -0.00001104, -0.00001161, -0.00000656, 0.00000029, 0.00000446, 0.00000386, -0.00000003, -0.00000352, -0.00000364, -0.00000037, 0.00000349, 0.00000467, 0.00000208, -0.00000230, -0.00000505, -0.00000409, -0.00000043, 0.00000266, 0.00000236, -0.00000108, -0.00000427, -0.00000313, 0.00000362, 0.00001285, 0.00001863, 0.00001618, 0.00000539, -0.00000877, -0.00001924, -0.00002135, -0.00001562, -0.00000696, -0.00000103, -0.00000035, -0.00000299, -0.00000447, -0.00000148, 0.00000561, 0.00001303, 0.00001641, 0.00001380, 0.00000679, -0.00000095, -0.00000640, -0.00000889, -0.00000977, -0.00001049, -0.00001083, -0.00000898, -0.00000334, 0.00000548, 0.00001416, 0.00001831, 0.00001528, 0.00000602, -0.00000529, -0.00001362, -0.00001594, -0.00001255, -0.00000632, -0.00000056, 0.00000294, 0.00000432, 0.00000476, 0.00000509, 0.00000515, 0.00000422, 0.00000199, -0.00000093, -0.00000330, -0.00000415, -0.00000341, -0.00000183, -0.00000036, 0.00000055, 0.00000106, 0.00000156, 0.00000219, 0.00000279, 0.00000320, 0.00000358, 0.00000433, 0.00000550, 0.00000640, 0.00000578, 0.00000283, -0.00000184, -0.00000602, -0.00000696, -0.00000317, 0.00000432, 0.00001210, 0.00001618, 0.00001422, 0.00000683, },
		{ 0.00001341, 0.00001396, 0.00000958, 0.00000210, -0.00000556, -0.00001114, -0.00001389, -0.00001436, -0.00001340, -0.00001126, -0.00000765, -0.00000241, 0.00000375, 0.00000920, 0.00001218, 0.00001174, 0.00000823, 0.00000304, -0.00000217, -0.00000611, -0.00000809, -0.00000792, -0.00000590, -0.00000292, -0.00000050, -0.00000034, -0.00000327, -0.00000828, -0.00001249, -0.00001256, -0.00000685, 0.00000308, 0.00001281, 0.00001739, 0.00001435, 0.00000534, -0.00000480, -0.00001101, -0.00001100, -0.00000636, -0.00000103, 0.00000167, 0.00000119, -0.00000041, -0.00000057, 0.00000139, 0.00000372, 0.00000384, 0.00000077, -0.00000366, -0.00000599, -0.00000393, 0.00000154, 0.00000645, 0.00000668, 0.00000125, -0.00000642, -0.00001059, -0.00000719, 0.00000305, 0.00001460, 0.00002056, 0.00001706, 0.00000574, -0.00000744, -0.00001617, -0.00001757, -0.00001345, -0.00000820, -0.00000537, -0.00000525, -0.00000533, -0.00000282, 0.00000281, 0.00000925, 0.00001311, 0.00001242, 0.00000795, 0.00000242, -0.00000169, -0.00000385, -0.00000536, -0.00000761, -0.00001033, -0.00001136, -0.00000822, -0.00000050, 0.00000917, 0.00001615, 0.00001661, 0.00000989, -0.00000096, -0.00001094, -0.00001601, -0.00001500, -0.00000972, -0.00000329, 0.00000184, 0.00000487, 0.00000632, 0.00000685, 0.00000659, 0.00000522, 0.00000264, -0.00000052, -0.00000307, -0.00000408, -0.00000342, -0.00000186, -0.00000036, 0.00000056, 0.00000101, 0.00000143, 0.00000201, 0.00000253, 0.00000271, 0.00000265, 0.00000285, 0.00000371, 0.00000494, 0.00000547, 0.00000413, 0.00000076, -0.00000324, -0.00000544, -0.00000389, 0.00000146, 0.00000836, },
		{ -0.00000333, 0.00000039, 0.00000608, 0.00001104, 0.00001273, 0.00001025, 0.00000468, -0.00000177, -0.00000713, -0.00001061, -0.00001245, -0.00001314, -0.00001266, -0.00001046, -0.00000616, -0.00000027, 0.00000572, 0.00000998, 0.00001130, 0.00000962, 0.00000585, 0.00000121, -0.00000320, -0.00000652, -0.00000798, -0.00000703, -0.00000380, 0.00000033, 0.00000292, 0.00000166, -0.00000389, -0.00001144, -0.00001667, -0.00001568, -0.00000759, 0.00000450, 0.00001496, 0.00001883, 0.00001470, 0.00000541, -0.00000396, -0.00000926, -0.00000953, -0.00000677, -0.00000385, -0.00000221, -0.00000112, 0.00000090, 0.00000415, 0.00000683, 0.00000639, 0.00000183, -0.00000475, -0.00000908, -0.00000763, -0.00000052, 0.00000798, 0.00001203, 0.00000823, -0.00000169, -0.00001166, -0.00001503, -0.00000904, 0.00000335, 0.00001529, 0.00002036, 0.00001636, 0.00000621, -0.00000455, -0.00001141, -0.00001333, -0.00001232, -0.00001103, -0.00001039, -0.00000921, -0.00000571, 0.00000037, 0.00000700, 0.00001117, 0.00001102, 0.00000726, 0.00000258, -0.00000045, -0.00000122, -0.00000129, -0.00000278, -0.00000620, -0.00000967, -0.00001007, -0.00000550, 0.00000289, 0.00001120, 0.00001495, 0.00001189, 0.00000334, -0.00000670, -0.00001385, -0.00001572, -0.00001270, -0.00000703, -0.00000113, 0.00000356, 0.00000672, 0.00000847, 0.00000873, 0.00000726, 0.00000418, 0.00000036, -0.00000282, -0.00000425, -0.00000372, -0.00000203, -0.00000033, 0.00000068, 0.00000106, 0.00000132, 0.00000182, 0.00000240, 0.00000266, 0.00000245, 0.00000217, 0.00000240, 0.00000329, 0.00000414, 0.00000383, 0.00000177, -0.00000136, -0.00000371, },
		{ 0.00000158, -0.00000078, -0.00000284, -0.00000287, -0.00000000, 0.00000492, 0.00000968, 0.00001192, 0.00001050, 0.00000606, 0.00000042, -0.00000462, -0.00000818, -0.00001040, -0.00001168, -0.00001199, -0.00001082, -0.00000770, -0.00000293, 0.00000231, 0.00000647, 0.00000853, 0.00000838, 0.00000659, 0.00000380, 0.00000040, -0.00000328, -0.00000640, -0.00000760, -0.00000576, -0.00000118, 0.00000387, 0.00000599, 0.00000272, -0.00000546, -0.00001476, -0.00001991, -0.00001732, -0.00000739, 0.00000548, 0.00001538, 0.00001828, 0.00001410, 0.00000612, -0.00000147, -0.00000616, -0.00000797, -0.00000826, -0.00000785, -0.00000615, -0.00000212, 0.00000399, 0.00000964, 0.00001124, 0.00000674, -0.00000218, -0.00001046, -0.00001265, -0.00000668, 0.00000433, 0.00001361, 0.00001506, 0.00000726, -0.00000539, -0.00001541, -0.00001688, -0.00000904, 0.00000356, 0.00001422, 0.00001817, 0.00001490, 0.00000747, -0.00000018, -0.00000587, -0.00000967, -0.00001261, -0.00001479, -0.00001491, -0.00001140, -0.00000429, 0.00000403, 0.00000995, 0.00001096, 0.00000738, 0.00000209, -0.00000151, -0.00000190, -0.00000029, 0.00000058, -0.00000115, -0.00000485, -0.00000772, -0.00000689, -0.00000173, 0.00000544, 0.00001065, 0.00001078, 0.00000548, -0.00000278, -0.00001030, -0.00001422, -0.00001379, -0.00001007, -0.00000482, 0.00000054, 0.00000522, 0.00000871, 0.00001039, 0.00000970, 0.00000664, 0.00000220, -0.00000193, -0.00000422, -0.00000416, -0.00000247, -0.00000047, 0.00000082, 0.00000125, 0.00000134, 0.00000165, 0.00000222, 0.00000264, 0.00000255, 0.00000211, 0.00000186, 0.00000220, 0.00000283, 0.00000287, },
		{ 0.00000174, 0.00000205, 0.00000206, 0.00000108, -0.00000092, -0.00000286, -0.00000316, -0.00000080, 0.00000378, 0.00000869, 0.00001161, 0.00001115, 0.00000753, 0.00000226, -0.00000287, -0.00000681, -0.00000938, -0.00001082, -0.00001115, -0.00001001, -0.00000719, -0.00000311, 0.00000105, 0.00000402, 0.00000521, 0.00000502, 0.00000435, 0.00000365, 0.00000257, 0.00000037, -0.00000295, -0.00000606, -0.00000678, -0.00000375, 0.00000205, 0.00000729, 0.00000804, 0.00000241, -0.00000774, -0.00001746, -0.00002146, -0.00001728, -0.00000675, 0.00000524, 0.00001366, 0.00001607, 0.00001329, 0.00000798, 0.00000245, -0.00000251, -0.00000706, -0.00001081, -0.00001211, -0.00000901, -0.00000127, 0.00000822, 0.00001448, 0.00001335, 0.00000463, -0.00000706, -0.00001474, -0.00001342, -0.00000350, 0.00000920, 0.00001697, 0.00001504, 0.00000450, -0.00000849, -0.00001673, -0.00001621, -0.00000808, 0.00000282, 0.00001130, 0.00001472, 0.00001355, 0.00000981, 0.00000500, -0.00000066, -0.00000733, -0.00001406, -0.00001844, -0.00001786, -0.00001156, -0.00000181, 0.00000701, 0.00001094, 0.00000887, 0.00000315, -0.00000218, -0.00000409, -0.00000243, 0.00000040, 0.00000157, 0.00000002, -0.00000280, -0.00000422, -0.00000249, 0.00000181, 0.00000607, 0.00000745, 0.00000471, -0.00000107, -0.00000735, -0.00001169, -0.00001293, -0.00001128, -0.00000764, -0.00000288, 0.00000231, 0.00000717, 0.00001059, 0.00001142, 0.00000921, 0.00000473, -0.00000017, -0.00000352, -0.00000431, -0.00000294, -0.00000079, 0.00000084, 0.00000146, 0.00000148, 0.00000158, 0.00000203, 0.00000253, 0.00000263, 0.00000225, 0.00000180, },
		},
		{ { 0.00000444, 0.00000423, 0.00000365, 0.00000285, 0.00000201, 0.00000133, 0.00000089, 0.00000067, 0.00000053, 0.00000024, -0.00000038, -0.00000144, -0.00000283, -0.00000431, -0.00000553, -0.00000617, -0.00000603, -0.00000510, -0.00000351, -0.00000147, 0.00000080, 0.00000315, 0.00000544, 0.00000749, 0.00000898, 0.00000954, 0.00000883, 0.00000676, 0.00000362, 0.00000016, -0.00000262, -0.00000381, -0.00000295, -0.00000033, 0.00000308, 0.00000596, 0.00000719, 0.00000635, 0.00000400, 0.00000143, 0.00000009, 0.00000089, 0.00000365, 0.00000698, 0.00000881, 0.00000727, 0.00000161, -0.00000724, -0.00001681, -0.00002390, -0.00002579, -0.00002133, -0.00001150, 0.00000086, 0.00001212, 0.00001914, 0.00002041, 0.00001651, 0.00000966, 0.00000271, -0.00000203, -0.00000370, -0.00000297, -0.00000145, -0.00000073, -0.00000145, -0.00000297, -0.00000370, -0.00000203, 0.00000271, 0.00000966, 0.00001651, 0.00002041, 0.00001914, 0.00001212, 0.00000086, -0.00001150, -0.00002133, -0.00002579, -0.00002390, -0.00001681, -0.00000724, 0.00000161, 0.00000727, 0.00000881, 0.00000698, 0.00000365, 0.00000089, 0.00000009, 0.00000143, 0.00000400, 0.00000635, 0.00000719, 0.00000596, 0.00000308, -0.00000033, -0.00000295, -0.00000381, -0.00000262, 0.00000016, 0.00000362, 0.00000676, 0.00000883, 0.00000954, 0.00000898, 0.00000749, 0.00000544, 0.00000315, 0.00000080, -0.00000147, -0.00000351, -0.00000510, -0.00000603, -0.00000617, -0.00000553, -0.00000431, -0.00000283, -0.00000144, -0.00000038, 0.00000024, 0.00000053, 0.00000067, 0.00000089, 0.00000133, 0.00000201, 0.00000285, 0.00000365, 0.00000423, },
		{ 0.00000182, 0.00000237, 0.00000302, 0.00000367, 0.00000418, 0.00000437, 0.00000414, 0.00000350, 0.00000261, 0.00000170, 0.00000100, 0.00000066, 0.00000065, 0.00000078, 0.00000076, 0.00000035, -0.00000058, -0.00000195, -0.00000349, -0.00000489, -0.00000583, -0.00000614, -0.00000575, -0.00000468, -0.00000296, -0.00000062, 0.00000226, 0.00000545, 0.00000849, 0.00001070, 0.00001137, 0.00001001, 0.00000665, 0.00000198, -0.00000275, -0.00000616, -0.00000717, -0.00000551, -0.00000184, 0.00000248, 0.00000593, 0.00000747, 0.00000700, 0.00000540, 0.00000409, 0.00000428, 0.00000631, 0.00000929, 0.00001140, 0.00001062, 0.00000573, -0.00000297, -0.00001351, -0.00002287, -0.00002798, -0.00002699, -0.00002000, -0.00000907, 0.00000256, 0.00001163, 0.00001613, 0.00001583, 0.00001222, 0.00000764, 0.00000415, 0.00000268, 0.00000276, 0.00000295, 0.00000182, -0.00000117, -0.00000523, -0.00000847, -0.00000879, -0.00000497, 0.00000255, 0.00001162, 0.00001909, 0.00002200, 0.00001881, 0.00001011, -0.00000158, -0.00001264, -0.00001979, -0.00002117, -0.00001703, -0.00000941, -0.00000130, 0.00000460, 0.00000680, 0.00000545, 0.00000201, -0.00000149, -0.00000337, -0.00000291, -0.00000053, 0.00000257, 0.00000505, 0.00000598, 0.00000522, 0.00000334, 0.00000132, 0.00000006, 0.00000006, 0.00000124, 0.00000312, 0.00000507, 0.00000655, 0.00000731, 0.00000738, 0.00000692, 0.00000612, 0.00000507, 0.00000374, 0.00000209, 0.00000016, -0.00000191, -0.00000383, -0.00000530, -0.00000606, -0.00000604, -0.00000531, -0.00000411, -0.00000272, -0.00000141, -0.00000035, 0.00000041, 0.00000094, 0.00000137, },
		{ -0.00000242, -0.00000081, 0.00000049, 0.00000137, 0.00000191, 0.00000226, 0.00000260, 0.00000303, 0.00000352, 0.00000394, 0.00000411, 0.00000389, 0.00000329, 0.00000246, 0.00000164, 0.00000105, 0.00000081, 0.00000086, 0.00000099, 0.00000093, 0.00000050, -0.00000034, -0.00000150, -0.00000276, -0.00000393, -0.00000488, -0.00000553, -0.00000578, -0.00000546, -0.00000432, -0.00000213, 0.00000114, 0.00000515, 0.00000910, 0.00001193, 0.00001261, 0.00001059, 0.00000611, 0.00000026, -0.00000531, -0.00000899, -0.00000973, -0.00000751, -0.00000330, 0.00000137, 0.00000503, 0.00000694, 0.00000732, 0.00000717, 0.00000767, 0.00000946, 0.00001214, 0.00001431, 0.00001411, 0.00001013, 0.00000217, -0.00000838, -0.00001885, -0.00002626, -0.00002839, -0.00002468, -0.00001642, -0.00000629, 0.00000272, 0.00000847, 0.00001035, 0.00000936, 0.00000739, 0.00000622, 0.00000664, 0.00000810, 0.00000906, 0.00000791, 0.00000389, -0.00000226, -0.00000847, -0.00001217, -0.00001137, -0.00000569, 0.00000333, 0.00001272, 0.00001914, 0.00002021, 0.00001545, 0.00000644, -0.00000378, -0.00001188, -0.00001552, -0.00001408, -0.00000881, -0.00000214, 0.00000333, 0.00000581, 0.00000492, 0.00000162, -0.00000236, -0.00000529, -0.00000612, -0.00000481, -0.00000212, 0.00000081, 0.00000299, 0.00000398, 0.00000393, 0.00000342, 0.00000308, 0.00000329, 0.00000405, 0.00000507, 0.00000592, 0.00000629, 0.00000613, 0.00000561, 0.00000502, 0.00000461, 0.00000442, 0.00000433, 0.00000405, 0.00000332, 0.00000197, 0.00000009, -0.00000209, -0.00000417, -0.00000576, -0.00000658, -0.00000650, -0.00000560, -0.00000413, },
		{ -0.00000548, -0.00000718, -0.00000770, -0.00000700, -0.00000534, -0.00000317, -0.00000101, 0.00000071, 0.00000182, 0.00000236, 0.00000255, 0.00000266, 0.00000286, 0.00000317, 0.00000348, 0.00000363, 0.00000350, 0.00000309, 0.00000253, 0.00000199, 0.00000160, 0.00000138, 0.00000122, 0.00000098, 0.00000056, -0.00000004, -0.00000072, -0.00000139, -0.00000199, -0.00000260, -0.00000336, -0.00000431, -0.00000530, -0.00000587, -0.00000545, -0.00000352, 0.00000001, 0.00000464, 0.00000924, 0.00001240, 0.00001293, 0.00001034, 0.00000511, -0.00000135, -0.00000719, -0.00001080, -0.00001131, -0.00000896, -0.00000482, -0.00000032, 0.00000340, 0.00000596, 0.00000776, 0.00000960, 0.00001205, 0.00001498, 0.00001734, 0.00001758, 0.00001434, 0.00000721, -0.00000280, -0.00001348, -0.00002205, -0.00002624, -0.00002516, -0.00001959, -0.00001172, -0.00000416, 0.00000103, 0.00000318, 0.00000309, 0.00000250, 0.00000310, 0.00000564, 0.00000947, 0.00001285, 0.00001375, 0.00001090, 0.00000457, -0.00000338, -0.00001017, -0.00001317, -0.00001107, -0.00000445, 0.00000439, 0.00001229, 0.00001645, 0.00001554, 0.00001010, 0.00000232, -0.00000492, -0.00000919, -0.00000939, -0.00000602, -0.00000089, 0.00000376, 0.00000614, 0.00000556, 0.00000252, -0.00000161, -0.00000526, -0.00000730, -0.00000737, -0.00000585, -0.00000357, -0.00000136, 0.00000032, 0.00000146, 0.00000237, 0.00000343, 0.00000483, 0.00000644, 0.00000785, 0.00000861, 0.00000842, 0.00000729, 0.00000556, 0.00000380, 0.00000253, 0.00000209, 0.00000247, 0.00000336, 0.00000425, 0.00000457, 0.00000396, 0.00000231, -0.00000015, -0.00000295, },
		{ 0.00000587, 0.00000596, 0.00000447, 0.00000163, -0.00000191, -0.00000531, -0.00000773, -0.00000867, -0.00000801, -0.00000610, -0.00000355, -0.00000102, 0.00000095, 0.00000215, 0.00000263, 0.00000266, 0.00000255, 0.00000250, 0.00000259, 0.00000276, 0.00000291, 0.00000298, 0.00000298, 0.00000293, 0.00000284, 0.00000264, 0.00000224, 0.00000156, 0.00000063, -0.00000036, -0.00000116, -0.00000153, -0.00000144, -0.00000113, -0.00000101, -0.00000150, -0.00000275, -0.00000445, -0.00000588, -0.00000612, -0.00000449, -0.00000090, 0.00000396, 0.00000872, 0.00001181, 0.00001206, 0.00000913, 0.00000372, -0.00000262, -0.00000815, -0.00001144, -0.00001193, -0.00000994, -0.00000643, -0.00000241, 0.00000145, 0.00000507, 0.00000873, 0.00001270, 0.00001671, 0.00001986, 0.00002079, 0.00001826, 0.00001188, 0.00000245, -0.00000804, -0.00001702, -0.00002229, -0.00002280, -0.00001910, -0.00001308, -0.00000717, -0.00000332, -0.00000221, -0.00000306, -0.00000407, -0.00000337, 0.00000003, 0.00000564, 0.00001166, 0.00001569, 0.00001585, 0.00001162, 0.00000423, -0.00000382, -0.00000967, -0.00001131, -0.00000832, -0.00000205, 0.00000498, 0.00001010, 0.00001157, 0.00000918, 0.00000423, -0.00000108, -0.00000461, -0.00000512, -0.00000271, 0.00000138, 0.00000531, 0.00000748, 0.00000711, 0.00000442, 0.00000044, -0.00000349, -0.00000630, -0.00000752, -0.00000732, -0.00000626, -0.00000495, -0.00000375, -0.00000261, -0.00000123, 0.00000072, 0.00000335, 0.00000639, 0.00000924, 0.00001120, 0.00001169, 0.00001054, 0.00000803, 0.00000486, 0.00000191, -0.00000003, -0.00000052, 0.00000044, 0.00000236, 0.00000445, },
		{ -0.00000252, -0.00000297, -0.00000141, 0.00000150, 0.00000470, 0.00000703, 0.00000763, 0.00000618, 0.00000297, -0.00000115, -0.00000512, -0.00000796, -0.00000906, -0.00000838, -0.00000633, -0.00000362, -0.00000099, 0.00000104, 0.00000224, 0.00000268, 0.00000262, 0.00000233, 0.00000202, 0.00000183, 0.00000181, 0.00000201, 0.00000243, 0.00000301, 0.00000363, 0.00000404, 0.00000395, 0.00000318, 0.00000176, -0.00000001, -0.00000161, -0.00000252, -0.00000245, -0.00000152, -0.00000028, 0.00000054, 0.00000031, -0.00000114, -0.00000334, -0.00000534, -0.00000606, -0.00000476, -0.00000144, 0.00000307, 0.00000736, 0.00000996, 0.00000987, 0.00000699, 0.00000211, -0.00000341, -0.00000813, -0.00001103, -0.00001176, -0.00001057, -0.00000799, -0.00000447, -0.00000021, 0.00000479, 0.00001039, 0.00001608, 0.00002076, 0.00002301, 0.00002160, 0.00001610, 0.00000725, -0.00000302, -0.00001222, -0.00001808, -0.00001945, -0.00001673, -0.00001166, -0.00000666, -0.00000378, -0.00000385, -0.00000616, -0.00000885, -0.00000973, -0.00000736, -0.00000174, 0.00000562, 0.00001228, 0.00001590, 0.00001517, 0.00001044, 0.00000353, -0.00000303, -0.00000697, -0.00000721, -0.00000417, 0.00000047, 0.00000458, 0.00000648, 0.00000561, 0.00000268, -0.00000075, -0.00000298, -0.00000295, -0.00000058, 0.00000316, 0.00000680, 0.00000894, 0.00000883, 0.00000657, 0.00000299, -0.00000079, -0.00000380, -0.00000558, -0.00000625, -0.00000626, -0.00000613, -0.00000608, -0.00000599, -0.00000538, -0.00000375, -0.00000085, 0.00000315, 0.00000758, 0.00001148, 0.00001388, 0.00001414, 0.00001215, 0.00000844, 0.00000399, -0.00000001, },
		{ 0.00001576, 0.00001309, 0.00000840, 0.00000294, -0.00000182, -0.00000467, -0.00000495, -0.00000281, 0.00000090, 0.00000486, 0.00000774, 0.00000857, 0.00000708, 0.00000369, -0.00000062, -0.00000469, -0.00000754, -0.00000863, -0.00000798, -0.00000606, -0.00000355, -0.00000111, 0.00000081, 0.00000201, 0.00000251, 0.00000245, 0.00000204, 0.00000149, 0.00000100, 0.00000080, 0.00000109, 0.00000192, 0.00000316, 0.00000445, 0.00000527, 0.00000517, 0.00000394, 0.00000177, -0.00000074, -0.00000279, -0.00000369, -0.00000319, -0.00000158, 0.00000037, 0.00000169, 0.00000171, 0.00000031, -0.00000195, -0.00000408, -0.00000503, -0.00000417, -0.00000161, 0.00000186, 0.00000504, 0.00000680, 0.00000649, 0.00000415, 0.00000044, -0.00000368, -0.00000732, -0.00000987, -0.00001107, -0.00001085, -0.00000916, -0.00000589, -0.00000095, 0.00000544, 0.00001255, 0.00001902, 0.00002317, 0.00002352, 0.00001943, 0.00001150, 0.00000156, -0.00000783, -0.00001426, -0.00001632, -0.00001415, -0.00000936, -0.00000440, -0.00000155, -0.00000196, -0.00000527, -0.00000971, -0.00001294, -0.00001302, -0.00000929, -0.00000263, 0.00000489, 0.00001086, 0.00001349, 0.00001230, 0.00000823, 0.00000316, -0.00000092, -0.00000270, -0.00000205, 0.00000009, 0.00000225, 0.00000312, 0.00000220, -0.00000007, -0.00000253, -0.00000384, -0.00000316, -0.00000046, 0.00000344, 0.00000720, 0.00000959, 0.00000987, 0.00000812, 0.00000504, 0.00000167, -0.00000109, -0.00000281, -0.00000361, -0.00000398, -0.00000449, -0.00000543, -0.00000667, -0.00000762, -0.00000753, -0.00000577, -0.00000217, 0.00000285, 0.00000832, 0.00001300, 0.00001572, },
		{ -0.00000303, 0.00000253, 0.00000859, 0.00001368, 0.00001650, 0.00001630, 0.00001310, 0.00000778, 0.00000179, -0.00000327, -0.00000610, -0.00000610, -0.00000356, 0.00000051, 0.00000468, 0.00000760, 0.00000841, 0.00000696, 0.00000378, -0.00000014, -0.00000374, -0.00000623, -0.00000722, -0.00000681, -0.00000539, -0.00000347, -0.00000151, 0.00000018, 0.00000140, 0.00000206, 0.00000216, 0.00000175, 0.00000103, 0.00000031, -0.00000001, 0.00000037, 0.00000155, 0.00000329, 0.00000504, 0.00000610, 0.00000592, 0.00000433, 0.00000169, -0.00000121, -0.00000344, -0.00000429, -0.00000356, -0.00000166, 0.00000055, 0.00000214, 0.00000250, 0.00000154, -0.00000024, -0.00000202, -0.00000303, -0.00000284, -0.00000158, 0.00000021, 0.00000181, 0.00000260, 0.00000229, 0.00000094, -0.00000115, -0.00000360, -0.00000611, -0.00000838, -0.00001006, -0.00001063, -0.00000948, -0.00000609, -0.00000038, 0.00000700, 0.00001464, 0.00002062, 0.00002310, 0.00002103, 0.00001463, 0.00000550, -0.00000385, -0.00001083, -0.00001369, -0.00001216, -0.00000756, -0.00000222, 0.00000139, 0.00000168, -0.00000148, -0.00000674, -0.00001186, -0.00001465, -0.00001386, -0.00000961, -0.00000330, 0.00000301, 0.00000750, 0.00000919, 0.00000829, 0.00000592, 0.00000352, 0.00000214, 0.00000209, 0.00000282, 0.00000337, 0.00000285, 0.00000098, -0.00000176, -0.00000434, -0.00000559, -0.00000478, -0.00000193, 0.00000216, 0.00000622, 0.00000903, 0.00000982, 0.00000859, 0.00000597, 0.00000297, 0.00000054, -0.00000082, -0.00000119, -0.00000110, -0.00000128, -0.00000226, -0.00000413, -0.00000641, -0.00000824, -0.00000865, -0.00000696, },
		{ -0.00000225, -0.00000515, -0.00000757, -0.00000842, -0.00000698, -0.00000314, 0.00000243, 0.00000846, 0.00001343, 0.00001601, 0.00001548, 0.00001201, 0.00000657, 0.00000068, -0.00000405, -0.00000647, -0.00000617, -0.00000359, 0.00000020, 0.00000387, 0.00000631, 0.00000692, 0.00000575, 0.00000333, 0.00000046, -0.00000212, -0.00000395, -0.00000487, -0.00000496, -0.00000443, -0.00000347, -0.00000225, -0.00000091, 0.00000034, 0.00000128, 0.00000170, 0.00000150, 0.00000082, 0.00000004, -0.00000035, 0.00000004, 0.00000131, 0.00000318, 0.00000503, 0.00000612, 0.00000591, 0.00000431, 0.00000171, -0.00000109, -0.00000323, -0.00000410, -0.00000355, -0.00000193, 0.00000009, 0.00000179, 0.00000268, 0.00000263, 0.00000186, 0.00000077, -0.00000029, -0.00000109, -0.00000161, -0.00000191, -0.00000209, -0.00000218, -0.00000223, -0.00000231, -0.00000264, -0.00000347, -0.00000494, -0.00000692, -0.00000880, -0.00000966, -0.00000850, -0.00000471, 0.00000151, 0.00000901, 0.00001591, 0.00002014, 0.00002021, 0.00001585, 0.00000817, -0.00000057, -0.00000777, -0.00001141, -0.00001076, -0.00000663, -0.00000108, 0.00000346, 0.00000508, 0.00000313, -0.00000164, -0.00000739, -0.00001204, -0.00001401, -0.00001281, -0.00000912, -0.00000435, -0.00000001, 0.00000288, 0.00000415, 0.00000432, 0.00000424, 0.00000456, 0.00000539, 0.00000625, 0.00000637, 0.00000512, 0.00000240, -0.00000119, -0.00000457, -0.00000656, -0.00000637, -0.00000394, 0.00000001, 0.00000426, 0.00000752, 0.00000891, 0.00000823, 0.00000598, 0.00000313, 0.00000072, -0.00000055, -0.00000055, 0.00000030, 0.00000118, 0.00000128, 0.00000012, },
		{ -0.00000081, 0.00000097, 0.00000258, 0.00000317, 0.00000228, 0.00000000, -0.00000296, -0.00000553, -0.00000663, -0.00000553, -0.00000217, 0.00000277, 0.00000805, 0.00001222, 0.00001409, 0.00001316, 0.00000973, 0.00000482, -0.00000017, -0.00000391, -0.00000559, -0.00000510, -0.00000301, -0.00000024, 0.00000224, 0.00000377, 0.00000415, 0.00000359, 0.00000249, 0.00000127, 0.00000015, -0.00000083, -0.00000172, -0.00000254, -0.00000321, -0.00000352, -0.00000328, -0.00000245, -0.00000120, 0.00000010, 0.00000103, 0.00000134, 0.00000103, 0.00000043, -0.00000000, 0.00000016, 0.00000109, 0.00000257, 0.00000410, 0.00000510, 0.00000511, 0.00000401, 0.00000207, -0.00000017, -0.00000210, -0.00000322, -0.00000333, -0.00000248, -0.00000097, 0.00000083, 0.00000251, 0.00000376, 0.00000431, 0.00000397, 0.00000269, 0.00000060, -0.00000195, -0.00000441, -0.00000621, -0.00000693, -0.00000652, -0.00000534, -0.00000412, -0.00000358, -0.00000416, -0.00000564, -0.00000721, -0.00000769, -0.00000603, -0.00000184, 0.00000425, 0.00001074, 0.00001564, 0.00001726, 0.00001485, 0.00000898, 0.00000143, -0.00000544, -0.00000952, -0.00000972, -0.00000638, -0.00000106, 0.00000406, 0.00000701, 0.00000675, 0.00000346, -0.00000160, -0.00000670, -0.00001032, -0.00001165, -0.00001082, -0.00000859, -0.00000597, -0.00000368, -0.00000194, -0.00000045, 0.00000124, 0.00000340, 0.00000586, 0.00000799, 0.00000896, 0.00000813, 0.00000543, 0.00000143, -0.00000274, -0.00000581, -0.00000681, -0.00000545, -0.00000220, 0.00000187, 0.00000547, 0.00000754, 0.00000759, 0.00000583, 0.00000304, 0.00000029, -0.00000151, -0.00000184, },
		{ 0.00000365, 0.00000064, -0.00000204, -0.00000354, -0.00000346, -0.00000193, 0.00000041, 0.00000265, 0.00000393, 0.00000377, 0.00000222, -0.00000011, -0.00000224, -0.00000322, -0.00000243, 0.00000012, 0.00000378, 0.00000749, 0.00001009, 0.00001080, 0.00000941, 0.00000641, 0.00000273, -0.00000059, -0.00000277, -0.00000353, -0.00000309, -0.00000203, -0.00000095, -0.00000020, 0.00000016, 0.00000037, 0.00000073, 0.00000138, 0.00000220, 0.00000283, 0.00000283, 0.00000196, 0.00000029, -0.00000175, -0.00000354, -0.00000452, -0.00000438, -0.00000323, -0.00000150, 0.00000021, 0.00000139, 0.00000181, 0.00000158, 0.00000104, 0.00000064, 0.00000068, 0.00000124, 0.00000217, 0.00000312, 0.00000371, 0.00000367, 0.00000290, 0.00000149, -0.00000027, -0.00000195, -0.00000310, -0.00000331, -0.00000237, -0.00000033, 0.00000239, 0.00000509, 0.00000690, 0.00000709, 0.00000530, 0.00000176, -0.00000273, -0.00000704, -0.00001004, -0.00001108, -0.00001017, -0.00000801, -0.00000567, -0.00000413, -0.00000384, -0.00000449, -0.00000514, -0.00000465, -0.00000228, 0.00000194, 0.00000703, 0.00001142, 0.00001349, 0.00001230, 0.00000799, 0.00000181, -0.00000426, -0.00000830, -0.00000908, -0.00000655, -0.00000180, 0.00000339, 0.00000722, 0.00000850, 0.00000702, 0.00000350, -0.00000081, -0.00000468, -0.00000733, -0.00000859, -0.00000879, -0.00000845, -0.00000791, -0.00000717, -0.00000592, -0.00000378, -0.00000066, 0.00000312, 0.00000676, 0.00000930, 0.00000992, 0.00000833, 0.00000491, 0.00000063, -0.00000324, -0.00000557, -0.00000574, -0.00000382, -0.00000052, 0.00000305, 0.00000576, 0.00000682, 0.00000598, },
		{ 0.00000034, 0.00000326, 0.00000534, 0.00000593, 0.00000483, 0.00000237, -0.00000072, -0.00000351, -0.00000519, -0.00000530, -0.00000387, -0.00000144, 0.00000120, 0.00000324, 0.00000416, 0.00000393, 0.00000296, 0.00000197, 0.00000162, 0.00000224, 0.00000373, 0.00000553, 0.00000691, 0.00000729, 0.00000643, 0.00000461, 0.00000241, 0.00000054, -0.00000057, -0.00000086, -0.00000067, -0.00000055, -0.00000093, -0.00000192, -0.00000316, -0.00000405, -0.00000394, -0.00000251, 0.00000005, 0.00000304, 0.00000552, 0.00000661, 0.00000585, 0.00000339, -0.00000005, -0.00000345, -0.00000584, -0.00000659, -0.00000562, -0.00000338, -0.00000064, 0.00000175, 0.00000318, 0.00000339, 0.00000256, 0.00000117, -0.00000013, -0.00000081, -0.00000056, 0.00000056, 0.00000213, 0.00000354, 0.00000415, 0.00000356, 0.00000177, -0.00000072, -0.00000304, -0.00000425, -0.00000365, -0.00000115, 0.00000267, 0.00000662, 0.00000929, 0.00000956, 0.00000699, 0.00000208, -0.00000394, -0.00000947, -0.00001314, -0.00001424, -0.00001298, -0.00001027, -0.00000732, -0.00000507, -0.00000384, -0.00000325, -0.00000249, -0.00000079, 0.00000210, 0.00000571, 0.00000895, 0.00001054, 0.00000963, 0.00000615, 0.00000101, -0.00000425, -0.00000797, -0.00000901, -0.00000712, -0.00000303, 0.00000187, 0.00000605, 0.00000840, 0.00000852, 0.00000675, 0.00000390, 0.00000082, -0.00000188, -0.00000402, -0.00000574, -0.00000724, -0.00000859, -0.00000954, -0.00000961, -0.00000836, -0.00000561, -0.00000162, 0.00000284, 0.00000675, 0.00000915, 0.00000946, 0.00000766, 0.00000436, 0.00000055, -0.00000266, -0.00000441, -0.00000430, -0.00000249, },
		{ 0.00000136, -0.00000138, -0.00000308, -0.00000333, -0.00000216, 0.00000004, 0.00000254, 0.00000452, 0.00000528, 0.00000448, 0.00000225, -0.00000085, -0.00000395, -0.00000621, -0.00000704, -0.00000630, -0.00000430, -0.00000165, 0.00000099, 0.00000315, 0.00000469, 0.00000575, 0.00000660, 0.00000739, 0.00000808, 0.00000838, 0.00000792, 0.00000649, 0.00000423, 0.00000164, -0.00000056, -0.00000171, -0.00000153, -0.00000026, 0.00000136, 0.00000238, 0.00000203, 0.00000009, -0.00000299, -0.00000618, -0.00000826, -0.00000826, -0.00000585, -0.00000153, 0.00000352, 0.00000782, 0.00001008, 0.00000961, 0.00000655, 0.00000181, -0.00000322, -0.00000712, -0.00000887, -0.00000811, -0.00000525, -0.00000130, 0.00000249, 0.00000497, 0.00000549, 0.00000408, 0.00000140, -0.00000145, -0.00000337, -0.00000358, -0.00000198, 0.00000082, 0.00000374, 0.00000558, 0.00000550, 0.00000343, 0.00000011, -0.00000316, -0.00000495, -0.00000431, -0.00000117, 0.00000353, 0.00000819, 0.00001105, 0.00001089, 0.00000742, 0.00000146, -0.00000542, -0.00001143, -0.00001523, -0.00001627, -0.00001490, -0.00001210, -0.00000891, -0.00000608, -0.00000378, -0.00000172, 0.00000056, 0.00000328, 0.00000616, 0.00000848, 0.00000934, 0.00000811, 0.00000478, 0.00000008, -0.00000472, -0.00000824, -0.00000945, -0.00000807, -0.00000463, -0.00000019, 0.00000398, 0.00000693, 0.00000821, 0.00000794, 0.00000661, 0.00000478, 0.00000282, 0.00000082, -0.00000130, -0.00000367, -0.00000619, -0.00000854, -0.00001015, -0.00001046, -0.00000913, -0.00000623, -0.00000224, 0.00000202, 0.00000563, 0.00000786, 0.00000829, 0.00000699, 0.00000443, },
		{ 0.00000368, 0.00000609, 0.00000724, 0.00000694, 0.00000532, 0.00000280, 0.00000004, -0.00000222, -0.00000332, -0.00000296, -0.00000126, 0.00000119, 0.00000356, 0.00000497, 0.00000484, 0.00000308, 0.00000010, -0.00000333, -0.00000634, -0.00000824, -0.00000867, -0.00000767, -0.00000553, -0.00000263, 0.00000072, 0.00000426, 0.00000777, 0.00001094, 0.00001329, 0.00001429, 0.00001346, 0.00001071, 0.00000644, 0.00000156, -0.00000271, -0.00000523, -0.00000538, -0.00000331, 0.00000005, 0.00000328, 0.00000496, 0.00000418, 0.00000093, -0.00000389, -0.00000872, -0.00001188, -0.00001210, -0.00000905, -0.00000337, 0.00000341, 0.00000939, 0.00001289, 0.00001290, 0.00000948, 0.00000364, -0.00000290, -0.00000824, -0.00001091, -0.00001027, -0.00000671, -0.00000152, 0.00000354, 0.00000681, 0.00000732, 0.00000505, 0.00000100, -0.00000323, -0.00000600, -0.00000625, -0.00000389, 0.00000012, 0.00000421, 0.00000675, 0.00000676, 0.00000426, 0.00000034, -0.00000331, -0.00000504, -0.00000394, -0.00000020, 0.00000491, 0.00000951, 0.00001183, 0.00001082, 0.00000653, 0.00000005, -0.00000693, -0.00001274, -0.00001628, -0.00001723, -0.00001599, -0.00001336, -0.00001012, -0.00000673, -0.00000333, 0.00000012, 0.00000358, 0.00000672, 0.00000893, 0.00000952, 0.00000807, 0.00000467, 0.00000003, -0.00000471, -0.00000832, -0.00000990, -0.00000914, -0.00000641, -0.00000257, 0.00000138, 0.00000459, 0.00000664, 0.00000751, 0.00000746, 0.00000681, 0.00000578, 0.00000440, 0.00000260, 0.00000029, -0.00000241, -0.00000522, -0.00000768, -0.00000927, -0.00000963, -0.00000859, -0.00000629, -0.00000310, 0.00000043, },
		{ -0.00000759, -0.00000744, -0.00000640, -0.00000444, -0.00000168, 0.00000152, 0.00000459, 0.00000684, 0.00000765, 0.00000674, 0.00000430, 0.00000101, -0.00000215, -0.00000421, -0.00000453, -0.00000308, -0.00000039, 0.00000258, 0.00000479, 0.00000549, 0.00000438, 0.00000174, -0.00000181, -0.00000547, -0.00000851, -0.00001042, -0.00001089, -0.00000977, -0.00000703, -0.00000279, 0.00000263, 0.00000861, 0.00001420, 0.00001827, 0.00001979, 0.00001817, 0.00001358, 0.00000698, -0.00000003, -0.00000568, -0.00000855, -0.00000809, -0.00000478, -0.00000001, 0.00000433, 0.00000654, 0.00000560, 0.00000158, -0.00000437, -0.00001038, -0.00001448, -0.00001515, -0.00001190, -0.00000540, 0.00000265, 0.00001006, 0.00001474, 0.00001536, 0.00001179, 0.00000513, -0.00000259, -0.00000904, -0.00001233, -0.00001160, -0.00000734, -0.00000116, 0.00000470, 0.00000821, 0.00000826, 0.00000501, -0.00000016, -0.00000519, -0.00000814, -0.00000793, -0.00000474, 0.00000011, 0.00000469, 0.00000724, 0.00000692, 0.00000406, 0.00000007, -0.00000321, -0.00000424, -0.00000243, 0.00000165, 0.00000650, 0.00001025, 0.00001142, 0.00000938, 0.00000454, -0.00000186, -0.00000829, -0.00001341, -0.00001645, -0.00001728, -0.00001624, -0.00001384, -0.00001054, -0.00000664, -0.00000237, 0.00000197, 0.00000588, 0.00000868, 0.00000973, 0.00000866, 0.00000559, 0.00000118, -0.00000353, -0.00000741, -0.00000959, -0.00000972, -0.00000798, -0.00000498, -0.00000151, 0.00000174, 0.00000433, 0.00000609, 0.00000707, 0.00000737, 0.00000708, 0.00000623, 0.00000484, 0.00000297, 0.00000076, -0.00000157, -0.00000377, -0.00000562, -0.00000694, },
		{ 0.00000056, -0.00000056, -0.00000148, -0.00000246, -0.00000369, -0.00000511, -0.00000636, -0.00000684, -0.00000603, -0.00000368, -0.00000010, 0.00000389, 0.00000717, 0.00000872, 0.00000801, 0.00000521, 0.00000119, -0.00000278, -0.00000550, -0.00000619, -0.00000478, -0.00000185, 0.00000158, 0.00000443, 0.00000588, 0.00000557, 0.00000356, 0.00000028, -0.00000369, -0.00000771, -0.00001111, -0.00001324, -0.00001348, -0.00001135, -0.00000671, 0.00000006, 0.00000799, 0.00001559, 0.00002121, 0.00002346, 0.00002167, 0.00001618, 0.00000832, 0.00000005, -0.00000655, -0.00000993, -0.00000954, -0.00000594, -0.00000068, 0.00000421, 0.00000688, 0.00000618, 0.00000210, -0.00000427, -0.00001099, -0.00001589, -0.00001722, -0.00001423, -0.00000746, 0.00000137, 0.00000981, 0.00001545, 0.00001664, 0.00001309, 0.00000598, -0.00000238, -0.00000933, -0.00001272, -0.00001168, -0.00000688, -0.00000028, 0.00000561, 0.00000865, 0.00000785, 0.00000374, -0.00000193, -0.00000690, -0.00000927, -0.00000825, -0.00000441, 0.00000062, 0.00000484, 0.00000675, 0.00000589, 0.00000300, -0.00000039, -0.00000258, -0.00000248, 0.00000002, 0.00000399, 0.00000788, 0.00001017, 0.00000987, 0.00000686, 0.00000185, -0.00000399, -0.00000942, -0.00001355, -0.00001594, -0.00001654, -0.00001556, -0.00001326, -0.00000989, -0.00000571, -0.00000114, 0.00000325, 0.00000673, 0.00000861, 0.00000850, 0.00000640, 0.00000283, -0.00000135, -0.00000520, -0.00000793, -0.00000910, -0.00000867, -0.00000694, -0.00000437, -0.00000141, 0.00000151, 0.00000409, 0.00000605, 0.00000722, 0.00000747, 0.00000683, 0.00000548, 0.00000375, 0.00000202, },
		{ 0.00000719, 0.00000549, 0.00000329, 0.00000139, 0.00000037, 0.00000037, 0.00000101, 0.00000153, 0.00000118, -0.00000041, -0.00000299, -0.00000570, -0.00000742, -0.00000721, -0.00000477, -0.00000060, 0.00000409, 0.00000783, 0.00000943, 0.00000837, 0.00000505, 0.00000055, -0.00000372, -0.00000653, -0.00000721, -0.00000580, -0.00000290, 0.00000055, 0.00000362, 0.00000561, 0.00000612, 0.00000503, 0.00000243, -0.00000139, -0.00000594, -0.00001049, -0.00001409, -0.00001572, -0.00001454, -0.00001018, -0.00000303, 0.00000577, 0.00001448, 0.00002120, 0.00002438, 0.00002329, 0.00001823, 0.00001050, 0.00000206, -0.00000504, -0.00000918, -0.00000964, -0.00000679, -0.00000195, 0.00000299, 0.00000610, 0.00000606, 0.00000252, -0.00000366, -0.00001065, -0.00001620, -0.00001828, -0.00001586, -0.00000926, -0.00000019, 0.00000874, 0.00001490, 0.00001644, 0.00001304, 0.00000598, -0.00000226, -0.00000888, -0.00001178, -0.00001028, -0.00000534, 0.00000089, 0.00000587, 0.00000774, 0.00000595, 0.00000140, -0.00000394, -0.00000793, -0.00000910, -0.00000719, -0.00000316, 0.00000126, 0.00000439, 0.00000525, 0.00000395, 0.00000149, -0.00000066, -0.00000130, 0.00000005, 0.00000295, 0.00000627, 0.00000869, 0.00000918, 0.00000740, 0.00000366, -0.00000117, -0.00000611, -0.00001033, -0.00001330, -0.00001485, -0.00001497, -0.00001379, -0.00001145, -0.00000815, -0.00000423, -0.00000021, 0.00000330, 0.00000569, 0.00000656, 0.00000582, 0.00000374, 0.00000086, -0.00000222, -0.00000494, -0.00000691, -0.00000790, -0.00000783, -0.00000670, -0.00000457, -0.00000169, 0.00000157, 0.00000461, 0.00000683, 0.00000774, },
		{ -0.00000151, 0.00000258, 0.00000613, 0.00000812, 0.00000799, 0.00000595, 0.00000287, 0.00000002, -0.00000148, -0.00000112, 0.00000077, 0.00000313, 0.00000465, 0.00000433, 0.00000197, -0.00000173, -0.00000539, -0.00000754, -0.00000721, -0.00000433, 0.00000020, 0.00000488, 0.00000817, 0.00000903, 0.00000731, 0.00000366, -0.00000069, -0.00000448, -0.00000674, -0.00000713, -0.00000582, -0.00000336, -0.00000042, 0.00000238, 0.00000455, 0.00000573, 0.00000561, 0.00000397, 0.00000081, -0.00000359, -0.00000853, -0.00001295, -0.00001567, -0.00001565, -0.00001241, -0.00000620, 0.00000199, 0.00001058, 0.00001780, 0.00002213, 0.00002271, 0.00001949, 0.00001334, 0.00000576, -0.00000142, -0.00000655, -0.00000855, -0.00000726, -0.00000351, 0.00000109, 0.00000463, 0.00000546, 0.00000284, -0.00000276, -0.00000969, -0.00001563, -0.00001836, -0.00001660, -0.00001052, -0.00000178, 0.00000698, 0.00001304, 0.00001463, 0.00001152, 0.00000511, -0.00000211, -0.00000755, -0.00000944, -0.00000753, -0.00000303, 0.00000191, 0.00000516, 0.00000544, 0.00000279, -0.00000153, -0.00000567, -0.00000798, -0.00000768, -0.00000512, -0.00000152, 0.00000163, 0.00000323, 0.00000303, 0.00000161, 0.00000010, -0.00000043, 0.00000058, 0.00000293, 0.00000578, 0.00000800, 0.00000868, 0.00000740, 0.00000434, 0.00000018, -0.00000423, -0.00000813, -0.00001102, -0.00001268, -0.00001311, -0.00001245, -0.00001085, -0.00000852, -0.00000574, -0.00000287, -0.00000028, 0.00000169, 0.00000286, 0.00000319, 0.00000279, 0.00000182, 0.00000041, -0.00000136, -0.00000339, -0.00000544, -0.00000712, -0.00000789, -0.00000727, -0.00000505, },
		{ -0.00000468, -0.00000746, -0.00000888, -0.00000808, -0.00000495, -0.00000021, 0.00000471, 0.00000820, 0.00000912, 0.00000725, 0.00000346, -0.00000067, -0.00000344, -0.00000378, -0.00000165, 0.00000193, 0.00000532, 0.00000697, 0.00000605, 0.00000286, -0.00000137, -0.00000497, -0.00000653, -0.00000546, -0.00000219, 0.00000203, 0.00000569, 0.00000756, 0.00000713, 0.00000473, 0.00000126, -0.00000219, -0.00000474, -0.00000595, -0.00000586, -0.00000477, -0.00000306, -0.00000107, 0.00000097, 0.00000285, 0.00000427, 0.00000488, 0.00000428, 0.00000223, -0.00000115, -0.00000538, -0.00000961, -0.00001283, -0.00001412, -0.00001289, -0.00000905, -0.00000306, 0.00000413, 0.00001124, 0.00001692, 0.00002000, 0.00001982, 0.00001636, 0.00001044, 0.00000350, -0.00000268, -0.00000653, -0.00000717, -0.00000483, -0.00000078, 0.00000302, 0.00000466, 0.00000297, -0.00000190, -0.00000848, -0.00001445, -0.00001754, -0.00001636, -0.00001104, -0.00000322, 0.00000460, 0.00000993, 0.00001129, 0.00000870, 0.00000361, -0.00000174, -0.00000529, -0.00000593, -0.00000390, -0.00000053, 0.00000236, 0.00000337, 0.00000204, -0.00000101, -0.00000441, -0.00000671, -0.00000703, -0.00000541, -0.00000269, -0.00000007, 0.00000146, 0.00000154, 0.00000058, -0.00000049, -0.00000072, 0.00000042, 0.00000279, 0.00000566, 0.00000799, 0.00000887, 0.00000784, 0.00000501, 0.00000101, -0.00000333, -0.00000718, -0.00000996, -0.00001139, -0.00001151, -0.00001055, -0.00000886, -0.00000687, -0.00000496, -0.00000343, -0.00000242, -0.00000186, -0.00000150, -0.00000105, -0.00000030, 0.00000066, 0.00000147, 0.00000160, 0.00000057, -0.00000168, },
		{ 0.00000033, 0.00000284, 0.00000375, 0.00000240, -0.00000102, -0.00000535, -0.00000890, -0.00001011, -0.00000818, -0.00000352, 0.00000237, 0.00000744, 0.00000988, 0.00000891, 0.00000507, -0.00000001, -0.00000424, -0.00000596, -0.00000460, -0.00000088, 0.00000355, 0.00000682, 0.00000764, 0.00000581, 0.00000224, -0.00000151, -0.00000388, -0.00000400, -0.00000197, 0.00000124, 0.00000428, 0.00000597, 0.00000577, 0.00000388, 0.00000107, -0.00000169, -0.00000366, -0.00000453, -0.00000443, -0.00000372, -0.00000280, -0.00000189, -0.00000102, -0.00000012, 0.00000086, 0.00000185, 0.00000263, 0.00000290, 0.00000239, 0.00000097, -0.00000131, -0.00000420, -0.00000728, -0.00000995, -0.00001153, -0.00001133, -0.00000886, -0.00000409, 0.00000244, 0.00000954, 0.00001561, 0.00001905, 0.00001885, 0.00001499, 0.00000858, 0.00000157, -0.00000390, -0.00000631, -0.00000532, -0.00000196, 0.00000181, 0.00000387, 0.00000284, -0.00000134, -0.00000730, -0.00001287, -0.00001589, -0.00001515, -0.00001085, -0.00000451, 0.00000168, 0.00000576, 0.00000676, 0.00000502, 0.00000188, -0.00000099, -0.00000233, -0.00000182, -0.00000013, 0.00000151, 0.00000195, 0.00000070, -0.00000186, -0.00000469, -0.00000661, -0.00000684, -0.00000537, -0.00000292, -0.00000056, 0.00000077, 0.00000073, -0.00000032, -0.00000150, -0.00000184, -0.00000074, 0.00000175, 0.00000494, 0.00000779, 0.00000928, 0.00000879, 0.00000628, 0.00000224, -0.00000243, -0.00000676, -0.00000993, -0.00001144, -0.00001121, -0.00000952, -0.00000697, -0.00000435, -0.00000241, -0.00000164, -0.00000214, -0.00000349, -0.00000493, -0.00000556, -0.00000478, -0.00000260, },
		{ -0.00000615, -0.00000877, -0.00000919, -0.00000695, -0.00000272, 0.00000198, 0.00000524, 0.00000565, 0.00000285, -0.00000216, -0.00000741, -0.00001069, -0.00001050, -0.00000668, -0.00000052, 0.00000572, 0.00000973, 0.00001010, 0.00000689, 0.00000155, -0.00000370, -0.00000678, -0.00000663, -0.00000356, 0.00000094, 0.00000495, 0.00000691, 0.00000631, 0.00000375, 0.00000064, -0.00000151, -0.00000176, -0.00000009, 0.00000262, 0.00000510, 0.00000622, 0.00000546, 0.00000310, -0.00000003, -0.00000289, -0.00000466, -0.00000504, -0.00000423, -0.00000277, -0.00000126, -0.00000019, 0.00000025, 0.00000010, -0.00000042, -0.00000099, -0.00000130, -0.00000107, -0.00000020, 0.00000115, 0.00000255, 0.00000332, 0.00000272, 0.00000030, -0.00000381, -0.00000872, -0.00001290, -0.00001468, -0.00001282, -0.00000714, 0.00000124, 0.00001016, 0.00001708, 0.00002000, 0.00001823, 0.00001264, 0.00000539, -0.00000099, -0.00000452, -0.00000456, -0.00000197, 0.00000133, 0.00000325, 0.00000245, -0.00000112, -0.00000623, -0.00001096, -0.00001355, -0.00001317, -0.00001015, -0.00000579, -0.00000172, 0.00000088, 0.00000165, 0.00000113, 0.00000035, 0.00000018, 0.00000092, 0.00000216, 0.00000300, 0.00000261, 0.00000067, -0.00000236, -0.00000549, -0.00000754, -0.00000779, -0.00000620, -0.00000351, -0.00000086, 0.00000075, 0.00000084, -0.00000033, -0.00000188, -0.00000277, -0.00000221, -0.00000003, 0.00000327, 0.00000669, 0.00000909, 0.00000961, 0.00000788, 0.00000416, -0.00000079, -0.00000589, -0.00001001, -0.00001229, -0.00001227, -0.00001009, -0.00000644, -0.00000249, 0.00000053, 0.00000161, 0.00000041, -0.00000258, },
		{ 0.00000239, 0.00000481, 0.00000406, 0.00000036, -0.00000492, -0.00000973, -0.00001210, -0.00001092, -0.00000649, -0.00000042, 0.00000492, 0.00000737, 0.00000590, 0.00000108, -0.00000511, -0.00001006, -0.00001164, -0.00000906, -0.00000324, 0.00000357, 0.00000878, 0.00001045, 0.00000817, 0.00000309, -0.00000258, -0.00000655, -0.00000741, -0.00000514, -0.00000102, 0.00000304, 0.00000539, 0.00000533, 0.00000333, 0.00000071, -0.00000100, -0.00000083, 0.00000126, 0.00000437, 0.00000714, 0.00000832, 0.00000732, 0.00000435, 0.00000033, -0.00000356, -0.00000628, -0.00000725, -0.00000648, -0.00000438, -0.00000166, 0.00000094, 0.00000275, 0.00000331, 0.00000242, 0.00000030, -0.00000243, -0.00000481, -0.00000577, -0.00000456, -0.00000113, 0.00000363, 0.00000807, 0.00001023, 0.00000867, 0.00000312, -0.00000518, -0.00001374, -0.00001964, -0.00002059, -0.00001586, -0.00000663, 0.00000437, 0.00001386, 0.00001918, 0.00001923, 0.00001481, 0.00000814, 0.00000190, -0.00000187, -0.00000250, -0.00000077, 0.00000161, 0.00000285, 0.00000193, -0.00000108, -0.00000510, -0.00000869, -0.00001071, -0.00001077, -0.00000931, -0.00000725, -0.00000542, -0.00000419, -0.00000335, -0.00000234, -0.00000069, 0.00000157, 0.00000385, 0.00000523, 0.00000489, 0.00000258, -0.00000117, -0.00000519, -0.00000817, -0.00000915, -0.00000794, -0.00000518, -0.00000203, 0.00000029, 0.00000107, 0.00000030, -0.00000133, -0.00000277, -0.00000304, -0.00000168, 0.00000115, 0.00000469, 0.00000785, 0.00000955, 0.00000909, 0.00000628, 0.00000157, -0.00000406, -0.00000933, -0.00001293, -0.00001391, -0.00001198, -0.00000768, -0.00000230, },
		{ -0.00001511, -0.00001426, -0.00001012, -0.00000390, 0.00000243, 0.00000667, 0.00000729, 0.00000398, -0.00000206, -0.00000854, -0.00001291, -0.00001335, -0.00000961, -0.00000310, 0.00000359, 0.00000776, 0.00000774, 0.00000357, -0.00000295, -0.00000901, -0.00001200, -0.00001061, -0.00000535, 0.00000168, 0.00000767, 0.00001034, 0.00000887, 0.00000414, -0.00000169, -0.00000618, -0.00000768, -0.00000592, -0.00000206, 0.00000199, 0.00000444, 0.00000447, 0.00000246, -0.00000024, -0.00000201, -0.00000174, 0.00000071, 0.00000447, 0.00000812, 0.00001023, 0.00001000, 0.00000741, 0.00000319, -0.00000155, -0.00000569, -0.00000841, -0.00000927, -0.00000819, -0.00000545, -0.00000161, 0.00000246, 0.00000569, 0.00000701, 0.00000573, 0.00000194, -0.00000333, -0.00000823, -0.00001069, -0.00000924, -0.00000370, 0.00000445, 0.00001244, 0.00001716, 0.00001631, 0.00000944, -0.00000173, -0.00001381, -0.00002291, -0.00002605, -0.00002228, -0.00001294, -0.00000114, 0.00000951, 0.00001613, 0.00001762, 0.00001472, 0.00000954, 0.00000449, 0.00000130, 0.00000046, 0.00000128, 0.00000243, 0.00000271, 0.00000154, -0.00000084, -0.00000364, -0.00000605, -0.00000761, -0.00000837, -0.00000870, -0.00000895, -0.00000913, -0.00000885, -0.00000758, -0.00000496, -0.00000124, 0.00000277, 0.00000582, 0.00000684, 0.00000531, 0.00000164, -0.00000302, -0.00000717, -0.00000952, -0.00000950, -0.00000743, -0.00000430, -0.00000137, 0.00000038, 0.00000056, -0.00000048, -0.00000193, -0.00000282, -0.00000243, -0.00000054, 0.00000248, 0.00000581, 0.00000841, 0.00000930, 0.00000788, 0.00000408, -0.00000147, -0.00000754, -0.00001257, },
		{ 0.00000602, 0.00000123, -0.00000499, -0.00001107, -0.00001520, -0.00001592, -0.00001275, -0.00000650, 0.00000085, 0.00000677, 0.00000908, 0.00000690, 0.00000101, -0.00000630, -0.00001213, -0.00001408, -0.00001132, -0.00000493, 0.00000247, 0.00000782, 0.00000891, 0.00000531, -0.00000138, -0.00000828, -0.00001239, -0.00001195, -0.00000710, 0.00000015, 0.00000685, 0.00001039, 0.00000956, 0.00000504, -0.00000104, -0.00000607, -0.00000813, -0.00000672, -0.00000290, 0.00000136, 0.00000411, 0.00000434, 0.00000231, -0.00000067, -0.00000291, -0.00000312, -0.00000099, 0.00000277, 0.00000683, 0.00000982, 0.00001085, 0.00000969, 0.00000674, 0.00000270, -0.00000170, -0.00000580, -0.00000900, -0.00001069, -0.00001031, -0.00000756, -0.00000272, 0.00000315, 0.00000829, 0.00001081, 0.00000938, 0.00000400, -0.00000373, -0.00001102, -0.00001481, -0.00001305, -0.00000563, 0.00000528, 0.00001594, 0.00002236, 0.00002180, 0.00001392, 0.00000094, -0.00001312, -0.00002393, -0.00002835, -0.00002555, -0.00001702, -0.00000591, 0.00000434, 0.00001117, 0.00001368, 0.00001259, 0.00000958, 0.00000647, 0.00000439, 0.00000357, 0.00000351, 0.00000343, 0.00000281, 0.00000157, -0.00000003, -0.00000163, -0.00000308, -0.00000455, -0.00000631, -0.00000848, -0.00001074, -0.00001239, -0.00001255, -0.00001062, -0.00000668, -0.00000158, 0.00000326, 0.00000638, 0.00000682, 0.00000451, 0.00000033, -0.00000431, -0.00000792, -0.00000954, -0.00000897, -0.00000678, -0.00000401, -0.00000167, -0.00000043, -0.00000039, -0.00000114, -0.00000197, -0.00000217, -0.00000125, 0.00000084, 0.00000372, 0.00000660, 0.00000850, 0.00000849, },
		{ 0.00000225, 0.00000487, 0.00000719, 0.00000824, 0.00000711, 0.00000343, -0.00000235, -0.00000888, -0.00001423, -0.00001652, -0.00001469, -0.00000907, -0.00000141, 0.00000561, 0.00000940, 0.00000849, 0.00000321, -0.00000432, -0.00001098, -0.00001396, -0.00001193, -0.00000564, 0.00000231, 0.00000856, 0.00001044, 0.00000713, 0.00000001, -0.00000789, -0.00001322, -0.00001372, -0.00000920, -0.00000157, 0.00000604, 0.00001063, 0.00001058, 0.00000632, -0.00000006, -0.00000577, -0.00000857, -0.00000772, -0.00000409, 0.00000033, 0.00000349, 0.00000417, 0.00000248, -0.00000035, -0.00000269, -0.00000328, -0.00000177, 0.00000124, 0.00000461, 0.00000726, 0.00000858, 0.00000857, 0.00000756, 0.00000583, 0.00000341, 0.00000017, -0.00000387, -0.00000819, -0.00001162, -0.00001271, -0.00001039, -0.00000465, 0.00000307, 0.00001020, 0.00001395, 0.00001246, 0.00000581, -0.00000379, -0.00001268, -0.00001714, -0.00001488, -0.00000612, 0.00000634, 0.00001814, 0.00002497, 0.00002419, 0.00001582, 0.00000246, -0.00001172, -0.00002257, -0.00002736, -0.00002557, -0.00001875, -0.00000965, -0.00000106, 0.00000513, 0.00000835, 0.00000912, 0.00000853, 0.00000755, 0.00000668, 0.00000595, 0.00000517, 0.00000420, 0.00000313, 0.00000219, 0.00000152, 0.00000095, 0.00000003, -0.00000178, -0.00000472, -0.00000850, -0.00001223, -0.00001471, -0.00001490, -0.00001242, -0.00000777, -0.00000223, 0.00000260, 0.00000537, 0.00000543, 0.00000306, -0.00000073, -0.00000461, -0.00000744, -0.00000859, -0.00000809, -0.00000647, -0.00000447, -0.00000276, -0.00000172, -0.00000139, -0.00000150, -0.00000157, -0.00000114, 0.00000012, },
		{ -0.00000225, -0.00000165, -0.00000105, -0.00000011, 0.00000146, 0.00000365, 0.00000602, 0.00000768, 0.00000760, 0.00000501, -0.00000007, -0.00000660, -0.00001271, -0.00001627, -0.00001575, -0.00001100, -0.00000346, 0.00000417, 0.00000900, 0.00000908, 0.00000434, -0.00000327, -0.00001051, -0.00001419, -0.00001251, -0.00000597, 0.00000283, 0.00001021, 0.00001300, 0.00000992, 0.00000218, -0.00000707, -0.00001401, -0.00001584, -0.00001193, -0.00000403, 0.00000456, 0.00001041, 0.00001147, 0.00000781, 0.00000143, -0.00000476, -0.00000828, -0.00000812, -0.00000502, -0.00000089, 0.00000217, 0.00000296, 0.00000159, -0.00000072, -0.00000239, -0.00000236, -0.00000057, 0.00000207, 0.00000434, 0.00000536, 0.00000509, 0.00000426, 0.00000382, 0.00000426, 0.00000519, 0.00000543, 0.00000372, -0.00000049, -0.00000643, -0.00001208, -0.00001493, -0.00001319, -0.00000675, 0.00000245, 0.00001106, 0.00001561, 0.00001405, 0.00000672, -0.00000362, -0.00001280, -0.00001695, -0.00001404, -0.00000478, 0.00000763, 0.00001878, 0.00002468, 0.00002329, 0.00001512, 0.00000286, -0.00000971, -0.00001922, -0.00002370, -0.00002301, -0.00001851, -0.00001221, -0.00000594, -0.00000082, 0.00000285, 0.00000527, 0.00000676, 0.00000752, 0.00000756, 0.00000689, 0.00000570, 0.00000443, 0.00000355, 0.00000335, 0.00000364, 0.00000380, 0.00000298, 0.00000058, -0.00000343, -0.00000833, -0.00001287, -0.00001567, -0.00001585, -0.00001329, -0.00000879, -0.00000368, 0.00000060, 0.00000303, 0.00000325, 0.00000162, -0.00000106, -0.00000385, -0.00000604, -0.00000721, -0.00000735, -0.00000667, -0.00000551, -0.00000422, -0.00000309, },
		{ -0.00000694, -0.00000720, -0.00000669, -0.00000560, -0.00000429, -0.00000303, -0.00000192, -0.00000079, 0.00000069, 0.00000273, 0.00000516, 0.00000728, 0.00000801, 0.00000634, 0.00000194, -0.00000448, -0.00001112, -0.00001569, -0.00001629, -0.00001233, -0.00000498, 0.00000310, 0.00000873, 0.00000954, 0.00000507, -0.00000290, -0.00001093, -0.00001540, -0.00001405, -0.00000702, 0.00000303, 0.00001204, 0.00001626, 0.00001382, 0.00000563, -0.00000502, -0.00001387, -0.00001748, -0.00001468, -0.00000695, 0.00000229, 0.00000925, 0.00001143, 0.00000858, 0.00000258, -0.00000356, -0.00000723, -0.00000729, -0.00000450, -0.00000089, 0.00000141, 0.00000130, -0.00000080, -0.00000327, -0.00000426, -0.00000272, 0.00000093, 0.00000503, 0.00000759, 0.00000731, 0.00000439, 0.00000042, -0.00000236, -0.00000230, 0.00000072, 0.00000512, 0.00000831, 0.00000798, 0.00000333, -0.00000432, -0.00001200, -0.00001632, -0.00001508, -0.00000832, 0.00000151, 0.00001056, 0.00001518, 0.00001352, 0.00000635, -0.00000326, -0.00001122, -0.00001409, -0.00001054, -0.00000177, 0.00000900, 0.00001789, 0.00002185, 0.00001973, 0.00001245, 0.00000246, -0.00000736, -0.00001472, -0.00001856, -0.00001900, -0.00001696, -0.00001353, -0.00000955, -0.00000550, -0.00000161, 0.00000191, 0.00000470, 0.00000639, 0.00000680, 0.00000612, 0.00000491, 0.00000395, 0.00000383, 0.00000462, 0.00000578, 0.00000637, 0.00000545, 0.00000255, -0.00000206, -0.00000740, -0.00001215, -0.00001509, -0.00001558, -0.00001372, -0.00001022, -0.00000616, -0.00000256, -0.00000010, 0.00000095, 0.00000069, -0.00000056, -0.00000235, -0.00000427, -0.00000590, },
		{ 0.00000036, 0.00000043, -0.00000070, -0.00000267, -0.00000488, -0.00000668, -0.00000759, -0.00000748, -0.00000659, -0.00000532, -0.00000400, -0.00000265, -0.00000099, 0.00000126, 0.00000408, 0.00000687, 0.00000852, 0.00000780, 0.00000406, -0.00000225, -0.00000944, -0.00001503, -0.00001673, -0.00001350, -0.00000623, 0.00000242, 0.00000896, 0.00001053, 0.00000620, -0.00000239, -0.00001163, -0.00001737, -0.00001671, -0.00000938, 0.00000203, 0.00001307, 0.00001926, 0.00001805, 0.00000993, -0.00000182, -0.00001248, -0.00001795, -0.00001648, -0.00000926, 0.00000020, 0.00000777, 0.00001055, 0.00000805, 0.00000221, -0.00000372, -0.00000687, -0.00000612, -0.00000249, 0.00000149, 0.00000324, 0.00000153, -0.00000281, -0.00000734, -0.00000926, -0.00000691, -0.00000073, 0.00000681, 0.00001237, 0.00001333, 0.00000912, 0.00000158, -0.00000586, -0.00000976, -0.00000830, -0.00000222, 0.00000554, 0.00001114, 0.00001167, 0.00000651, -0.00000235, -0.00001120, -0.00001624, -0.00001532, -0.00000885, 0.00000041, 0.00000861, 0.00001251, 0.00001083, 0.00000476, -0.00000270, -0.00000812, -0.00000908, -0.00000510, 0.00000224, 0.00001015, 0.00001580, 0.00001741, 0.00001473, 0.00000889, 0.00000173, -0.00000498, -0.00001013, -0.00001335, -0.00001480, -0.00001482, -0.00001364, -0.00001136, -0.00000810, -0.00000422, -0.00000036, 0.00000272, 0.00000444, 0.00000469, 0.00000394, 0.00000302, 0.00000278, 0.00000363, 0.00000535, 0.00000716, 0.00000803, 0.00000715, 0.00000427, -0.00000017, -0.00000523, -0.00000984, -0.00001309, -0.00001452, -0.00001413, -0.00001228, -0.00000948, -0.00000632, -0.00000332, -0.00000096, },
		{ -0.00001467, -0.00001299, -0.00000990, -0.00000607, -0.00000238, 0.00000026, 0.00000123, 0.00000047, -0.00000157, -0.00000403, -0.00000610, -0.00000728, -0.00000753, -0.00000714, -0.00000643, -0.00000547, -0.00000398, -0.00000158, 0.00000182, 0.00000563, 0.00000862, 0.00000926, 0.00000648, 0.00000042, -0.00000732, -0.00001408, -0.00001708, -0.00001475, -0.00000759, 0.00000178, 0.00000950, 0.00001211, 0.00000812, -0.00000115, -0.00001190, -0.00001940, -0.00001999, -0.00001279, -0.00000022, 0.00001290, 0.00002137, 0.00002179, 0.00001411, 0.00000158, -0.00001063, -0.00001772, -0.00001729, -0.00001028, -0.00000043, 0.00000762, 0.00001042, 0.00000729, 0.00000044, -0.00000626, -0.00000925, -0.00000710, -0.00000114, 0.00000521, 0.00000825, 0.00000591, -0.00000111, -0.00000956, -0.00001518, -0.00001473, -0.00000769, 0.00000341, 0.00001401, 0.00001959, 0.00001770, 0.00000910, -0.00000253, -0.00001228, -0.00001614, -0.00001278, -0.00000402, 0.00000603, 0.00001289, 0.00001367, 0.00000825, -0.00000080, -0.00000953, -0.00001435, -0.00001363, -0.00000818, -0.00000073, 0.00000544, 0.00000802, 0.00000654, 0.00000240, -0.00000192, -0.00000411, -0.00000302, 0.00000097, 0.00000624, 0.00001080, 0.00001312, 0.00001263, 0.00000979, 0.00000561, 0.00000113, -0.00000294, -0.00000638, -0.00000920, -0.00001136, -0.00001265, -0.00001271, -0.00001126, -0.00000844, -0.00000484, -0.00000134, 0.00000117, 0.00000223, 0.00000200, 0.00000116, 0.00000060, 0.00000102, 0.00000256, 0.00000483, 0.00000701, 0.00000823, 0.00000790, 0.00000589, 0.00000249, -0.00000174, -0.00000613, -0.00001008, -0.00001307, -0.00001469, },
		{ 0.00000262, -0.00000168, -0.00000665, -0.00001146, -0.00001509, -0.00001665, -0.00001572, -0.00001255, -0.00000807, -0.00000358, -0.00000029, 0.00000113, 0.00000070, -0.00000094, -0.00000297, -0.00000471, -0.00000595, -0.00000681, -0.00000748, -0.00000785, -0.00000746, -0.00000565, -0.00000211, 0.00000269, 0.00000732, 0.00000986, 0.00000867, 0.00000339, -0.00000463, -0.00001254, -0.00001706, -0.00001592, -0.00000908, 0.00000100, 0.00001013, 0.00001414, 0.00001080, 0.00000098, -0.00001145, -0.00002113, -0.00002349, -0.00001685, -0.00000333, 0.00001194, 0.00002289, 0.00002514, 0.00001794, 0.00000446, -0.00000955, -0.00001833, -0.00001862, -0.00001105, 0.00000030, 0.00000984, 0.00001314, 0.00000892, -0.00000037, -0.00000989, -0.00001472, -0.00001232, -0.00000376, 0.00000675, 0.00001391, 0.00001385, 0.00000609, -0.00000616, -0.00001737, -0.00002224, -0.00001809, -0.00000621, 0.00000872, 0.00002062, 0.00002469, 0.00001942, 0.00000719, -0.00000688, -0.00001718, -0.00001997, -0.00001484, -0.00000462, 0.00000603, 0.00001278, 0.00001334, 0.00000824, 0.00000026, -0.00000701, -0.00001082, -0.00001031, -0.00000656, -0.00000185, 0.00000161, 0.00000270, 0.00000170, -0.00000005, -0.00000095, -0.00000003, 0.00000267, 0.00000620, 0.00000925, 0.00001080, 0.00001050, 0.00000871, 0.00000616, 0.00000348, 0.00000098, -0.00000144, -0.00000397, -0.00000663, -0.00000910, -0.00001077, -0.00001105, -0.00000971, -0.00000709, -0.00000395, -0.00000124, 0.00000032, 0.00000054, -0.00000020, -0.00000117, -0.00000164, -0.00000113, 0.00000040, 0.00000258, 0.00000489, 0.00000674, 0.00000768, 0.00000739, 0.00000571, },
		{ 0.00000093, 0.00000373, 0.00000650, 0.00000848, 0.00000880, 0.00000686, 0.00000258, -0.00000334, -0.00000960, -0.00001463, -0.00001718, -0.00001675, -0.00001377, -0.00000943, -0.00000513, -0.00000195, -0.00000032, -0.00000002, -0.00000050, -0.00000129, -0.00000230, -0.00000372, -0.00000570, -0.00000795, -0.00000964, -0.00000959, -0.00000698, -0.00000196, 0.00000405, 0.00000870, 0.00000968, 0.00000586, -0.00000187, -0.00001064, -0.00001668, -0.00001693, -0.00001059, 0.00000018, 0.00001093, 0.00001673, 0.00001440, 0.00000416, -0.00001017, -0.00002257, -0.00002735, -0.00002176, -0.00000737, 0.00001041, 0.00002447, 0.00002912, 0.00002253, 0.00000761, -0.00000920, -0.00002080, -0.00002251, -0.00001415, -0.00000002, 0.00001307, 0.00001898, 0.00001502, 0.00000317, -0.00001090, -0.00002048, -0.00002095, -0.00001195, 0.00000243, 0.00001555, 0.00002125, 0.00001668, 0.00000362, -0.00001230, -0.00002409, -0.00002653, -0.00001837, -0.00000288, 0.00001368, 0.00002473, 0.00002620, 0.00001801, 0.00000392, -0.00001029, -0.00001931, -0.00002044, -0.00001432, -0.00000434, 0.00000511, 0.00001057, 0.00001074, 0.00000668, 0.00000091, -0.00000396, -0.00000634, -0.00000615, -0.00000450, -0.00000282, -0.00000204, -0.00000218, -0.00000246, -0.00000191, 0.00000004, 0.00000318, 0.00000665, 0.00000938, 0.00001057, 0.00001011, 0.00000847, 0.00000638, 0.00000447, 0.00000289, 0.00000139, -0.00000042, -0.00000276, -0.00000542, -0.00000781, -0.00000921, -0.00000914, -0.00000761, -0.00000512, -0.00000250, -0.00000053, 0.00000034, 0.00000010, -0.00000090, -0.00000212, -0.00000306, -0.00000335, -0.00000279, -0.00000134, },
		{ -0.00000116, -0.00000323, -0.00000523, -0.00000646, -0.00000623, -0.00000411, -0.00000029, 0.00000439, 0.00000852, 0.00001063, 0.00000970, 0.00000567, -0.00000050, -0.00000713, -0.00001249, -0.00001536, -0.00001550, -0.00001352, -0.00001052, -0.00000752, -0.00000502, -0.00000303, -0.00000131, 0.00000020, 0.00000114, 0.00000083, -0.00000123, -0.00000483, -0.00000882, -0.00001143, -0.00001102, -0.00000703, -0.00000058, 0.00000577, 0.00000899, 0.00000706, 0.00000014, -0.00000915, -0.00001661, -0.00001833, -0.00001257, -0.00000095, 0.00001190, 0.00002020, 0.00001957, 0.00000923, -0.00000729, -0.00002329, -0.00003173, -0.00002828, -0.00001348, 0.00000724, 0.00002565, 0.00003423, 0.00002942, 0.00001326, -0.00000740, -0.00002385, -0.00002924, -0.00002166, -0.00000483, 0.00001347, 0.00002499, 0.00002460, 0.00001258, -0.00000557, -0.00002170, -0.00002861, -0.00002331, -0.00000828, 0.00000975, 0.00002279, 0.00002517, 0.00001595, -0.00000075, -0.00001769, -0.00002771, -0.00002676, -0.00001549, 0.00000127, 0.00001672, 0.00002502, 0.00002358, 0.00001382, 0.00000023, -0.00001181, -0.00001822, -0.00001769, -0.00001175, -0.00000367, 0.00000318, 0.00000669, 0.00000660, 0.00000418, 0.00000121, -0.00000096, -0.00000194, -0.00000220, -0.00000254, -0.00000342, -0.00000464, -0.00000544, -0.00000494, -0.00000275, 0.00000082, 0.00000483, 0.00000814, 0.00000992, 0.00001000, 0.00000882, 0.00000717, 0.00000567, 0.00000457, 0.00000362, 0.00000237, 0.00000048, -0.00000203, -0.00000470, -0.00000685, -0.00000789, -0.00000756, -0.00000604, -0.00000383, -0.00000155, 0.00000025, 0.00000125, 0.00000132, 0.00000046, },
		{ -0.00000189, 0.00000061, 0.00000279, 0.00000403, 0.00000372, 0.00000160, -0.00000203, -0.00000623, -0.00000961, -0.00001080, -0.00000905, -0.00000460, 0.00000129, 0.00000673, 0.00000996, 0.00000997, 0.00000689, 0.00000184, -0.00000359, -0.00000803, -0.00001081, -0.00001201, -0.00001218, -0.00001179, -0.00001089, -0.00000917, -0.00000630, -0.00000245, 0.00000148, 0.00000401, 0.00000381, 0.00000051, -0.00000486, -0.00001004, -0.00001251, -0.00001068, -0.00000494, 0.00000225, 0.00000733, 0.00000728, 0.00000129, -0.00000848, -0.00001760, -0.00002118, -0.00001631, -0.00000374, 0.00001195, 0.00002407, 0.00002666, 0.00001739, -0.00000102, -0.00002149, -0.00003541, -0.00003631, -0.00002284, 0.00000028, 0.00002405, 0.00003882, 0.00003848, 0.00002308, -0.00000102, -0.00002382, -0.00003594, -0.00003262, -0.00001570, 0.00000722, 0.00002616, 0.00003306, 0.00002527, 0.00000648, -0.00001498, -0.00002989, -0.00003216, -0.00002131, -0.00000248, 0.00001600, 0.00002633, 0.00002454, 0.00001197, -0.00000558, -0.00002058, -0.00002702, -0.00002280, -0.00001033, 0.00000490, 0.00001679, 0.00002122, 0.00001746, 0.00000798, -0.00000290, -0.00001112, -0.00001438, -0.00001275, -0.00000813, -0.00000305, 0.00000059, 0.00000211, 0.00000205, 0.00000146, 0.00000116, 0.00000129, 0.00000134, 0.00000066, -0.00000105, -0.00000343, -0.00000554, -0.00000636, -0.00000531, -0.00000251, 0.00000120, 0.00000471, 0.00000710, 0.00000804, 0.00000780, 0.00000702, 0.00000629, 0.00000588, 0.00000561, 0.00000508, 0.00000386, 0.00000186, -0.00000070, -0.00000331, -0.00000542, -0.00000663, -0.00000678, -0.00000591, -0.00000419, },
		{ -0.00000288, -0.00000527, -0.00000700, -0.00000757, -0.00000654, -0.00000381, 0.00000018, 0.00000439, 0.00000740, 0.00000796, 0.00000547, 0.00000042, -0.00000572, -0.00001094, -0.00001349, -0.00001255, -0.00000854, -0.00000294, 0.00000240, 0.00000594, 0.00000707, 0.00000610, 0.00000390, 0.00000133, -0.00000125, -0.00000394, -0.00000704, -0.00001045, -0.00001344, -0.00001478, -0.00001337, -0.00000897, -0.00000267, 0.00000331, 0.00000652, 0.00000546, 0.00000046, -0.00000619, -0.00001123, -0.00001194, -0.00000765, -0.00000029, 0.00000625, 0.00000799, 0.00000292, -0.00000762, -0.00001914, -0.00002575, -0.00002301, -0.00001022, 0.00000858, 0.00002598, 0.00003421, 0.00002862, 0.00001021, -0.00001439, -0.00003527, -0.00004342, -0.00003466, -0.00001171, 0.00001666, 0.00003912, 0.00004651, 0.00003575, 0.00001120, -0.00001711, -0.00003770, -0.00004248, -0.00003004, -0.00000614, 0.00001892, 0.00003476, 0.00003525, 0.00002093, -0.00000152, -0.00002240, -0.00003324, -0.00003026, -0.00001569, 0.00000358, 0.00001921, 0.00002507, 0.00001958, 0.00000602, -0.00000929, -0.00001992, -0.00002204, -0.00001569, -0.00000438, 0.00000685, 0.00001370, 0.00001424, 0.00000936, 0.00000193, -0.00000481, -0.00000863, -0.00000905, -0.00000712, -0.00000455, -0.00000270, -0.00000201, -0.00000201, -0.00000183, -0.00000090, 0.00000069, 0.00000223, 0.00000282, 0.00000197, -0.00000015, -0.00000269, -0.00000458, -0.00000500, -0.00000380, -0.00000146, 0.00000115, 0.00000321, 0.00000435, 0.00000466, 0.00000462, 0.00000470, 0.00000514, 0.00000583, 0.00000641, 0.00000648, 0.00000580, 0.00000433, 0.00000222, -0.00000028, },
		{ 0.00000721, 0.00000810, 0.00000822, 0.00000721, 0.00000477, 0.00000097, -0.00000365, -0.00000795, -0.00001058, -0.00001039, -0.00000703, -0.00000126, 0.00000516, 0.00001008, 0.00001171, 0.00000940, 0.00000387, -0.00000307, -0.00000925, -0.00001297, -0.00001363, -0.00001170, -0.00000835, -0.00000474, -0.00000153, 0.00000123, 0.00000376, 0.00000603, 0.00000746, 0.00000705, 0.00000394, -0.00000186, -0.00000905, -0.00001529, -0.00001813, -0.00001612, -0.00000974, -0.00000145, 0.00000530, 0.00000762, 0.00000460, -0.00000206, -0.00000876, -0.00001173, -0.00000899, -0.00000159, 0.00000659, 0.00001063, 0.00000706, -0.00000408, -0.00001869, -0.00003007, -0.00003188, -0.00002128, -0.00000078, 0.00002222, 0.00003829, 0.00004002, 0.00002547, -0.00000045, -0.00002771, -0.00004509, -0.00004498, -0.00002675, 0.00000270, 0.00003170, 0.00004856, 0.00004647, 0.00002638, -0.00000350, -0.00003110, -0.00004561, -0.00004187, -0.00002229, 0.00000442, 0.00002712, 0.00003702, 0.00003108, 0.00001295, -0.00000904, -0.00002569, -0.00003081, -0.00002355, -0.00000821, 0.00000800, 0.00001835, 0.00001931, 0.00001172, -0.00000011, -0.00001067, -0.00001574, -0.00001404, -0.00000734, 0.00000074, 0.00000660, 0.00000823, 0.00000582, 0.00000126, -0.00000303, -0.00000534, -0.00000530, -0.00000381, -0.00000230, -0.00000183, -0.00000260, -0.00000392, -0.00000475, -0.00000432, -0.00000259, -0.00000024, 0.00000172, 0.00000248, 0.00000186, 0.00000032, -0.00000131, -0.00000224, -0.00000216, -0.00000125, -0.00000006, 0.00000087, 0.00000126, 0.00000120, 0.00000104, 0.00000116, 0.00000178, 0.00000290, 0.00000435, 0.00000587, },
		{ -0.00000201, -0.00000184, -0.00000098, 0.00000075, 0.00000346, 0.00000696, 0.00001061, 0.00001332, 0.00001388, 0.00001139, 0.00000587, -0.00000156, -0.00000887, -0.00001375, -0.00001452, -0.00001083, -0.00000385, 0.00000411, 0.00001049, 0.00001340, 0.00001231, 0.00000799, 0.00000206, -0.00000383, -0.00000860, -0.00001192, -0.00001396, -0.00001481, -0.00001420, -0.00001151, -0.00000636, 0.00000082, 0.00000845, 0.00001405, 0.00001526, 0.00001095, 0.00000206, -0.00000846, -0.00001672, -0.00001953, -0.00001591, -0.00000768, 0.00000129, 0.00000679, 0.00000638, 0.00000063, -0.00000704, -0.00001203, -0.00001092, -0.00000346, 0.00000693, 0.00001460, 0.00001435, 0.00000426, -0.00001280, -0.00002994, -0.00003903, -0.00003446, -0.00001606, 0.00001019, 0.00003423, 0.00004607, 0.00004013, 0.00001801, -0.00001184, -0.00003741, -0.00004812, -0.00003930, -0.00001431, 0.00001678, 0.00004139, 0.00004962, 0.00003837, 0.00001249, -0.00001741, -0.00003951, -0.00004563, -0.00003443, -0.00001158, 0.00001303, 0.00002958, 0.00003237, 0.00002180, 0.00000362, -0.00001396, -0.00002395, -0.00002337, -0.00001399, -0.00000103, 0.00000953, 0.00001365, 0.00001068, 0.00000317, -0.00000473, -0.00000935, -0.00000917, -0.00000513, 0.00000015, 0.00000386, 0.00000439, 0.00000190, -0.00000196, -0.00000511, -0.00000612, -0.00000484, -0.00000229, -0.00000005, 0.00000069, -0.00000038, -0.00000259, -0.00000475, -0.00000580, -0.00000531, -0.00000360, -0.00000150, 0.00000017, 0.00000097, 0.00000097, 0.00000061, 0.00000036, 0.00000047, 0.00000086, 0.00000122, 0.00000124, 0.00000079, -0.00000001, -0.00000091, -0.00000165, },
		{ 0.00000219, 0.00000140, 0.00000033, -0.00000104, -0.00000272, -0.00000458, -0.00000611, -0.00000649, -0.00000481, -0.00000060, 0.00000579, 0.00001294, 0.00001871, 0.00002096, 0.00001838, 0.00001121, 0.00000126, -0.00000858, -0.00001542, -0.00001740, -0.00001433, -0.00000763, 0.00000038, 0.00000738, 0.00001187, 0.00001345, 0.00001248, 0.00000958, 0.00000513, -0.00000077, -0.00000791, -0.00001549, -0.00002181, -0.00002463, -0.00002210, -0.00001383, -0.00000155, 0.00001110, 0.00001983, 0.00002139, 0.00001514, 0.00000353, -0.00000877, -0.00001686, -0.00001775, -0.00001178, -0.00000247, 0.00000517, 0.00000713, 0.00000245, -0.00000619, -0.00001369, -0.00001515, -0.00000852, 0.00000397, 0.00001647, 0.00002212, 0.00001646, -0.00000000, -0.00002144, -0.00003883, -0.00004383, -0.00003271, -0.00000843, 0.00002012, 0.00004169, 0.00004737, 0.00003457, 0.00000831, -0.00002071, -0.00004057, -0.00004315, -0.00002756, -0.00000048, 0.00002674, 0.00004291, 0.00004167, 0.00002397, -0.00000257, -0.00002713, -0.00004033, -0.00003794, -0.00002232, -0.00000088, 0.00001737, 0.00002568, 0.00002217, 0.00000999, -0.00000454, -0.00001504, -0.00001782, -0.00001305, -0.00000422, 0.00000399, 0.00000793, 0.00000658, 0.00000159, -0.00000384, -0.00000677, -0.00000591, -0.00000207, 0.00000242, 0.00000511, 0.00000457, 0.00000109, -0.00000362, -0.00000741, -0.00000870, -0.00000722, -0.00000393, -0.00000051, 0.00000155, 0.00000157, -0.00000014, -0.00000256, -0.00000460, -0.00000551, -0.00000522, -0.00000411, -0.00000274, -0.00000153, -0.00000059, 0.00000018, 0.00000093, 0.00000169, 0.00000234, 0.00000270, 0.00000265, },
		{ 0.00000128, 0.00000202, 0.00000236, 0.00000246, 0.00000256, 0.00000279, 0.00000296, 0.00000256, 0.00000094, -0.00000222, -0.00000650, -0.00001063, -0.00001276, -0.00001123, -0.00000536, 0.00000399, 0.00001446, 0.00002291, 0.00002659, 0.00002416, 0.00001623, 0.00000517, -0.00000584, -0.00001397, -0.00001764, -0.00001679, -0.00001254, -0.00000643, 0.00000027, 0.00000679, 0.00001261, 0.00001699, 0.00001873, 0.00001637, 0.00000903, -0.00000278, -0.00001658, -0.00002840, -0.00003402, -0.00003081, -0.00001908, -0.00000238, 0.00001362, 0.00002332, 0.00002357, 0.00001504, 0.00000197, -0.00000966, -0.00001493, -0.00001215, -0.00000368, 0.00000532, 0.00000949, 0.00000604, -0.00000366, -0.00001463, -0.00002059, -0.00001728, -0.00000481, 0.00001186, 0.00002496, 0.00002744, 0.00001652, -0.00000461, -0.00002772, -0.00004291, -0.00004299, -0.00002685, -0.00000025, 0.00002644, 0.00004258, 0.00004181, 0.00002484, -0.00000087, -0.00002438, -0.00003593, -0.00003112, -0.00001265, 0.00001108, 0.00002974, 0.00003552, 0.00002641, 0.00000668, -0.00001521, -0.00003046, -0.00003368, -0.00002485, -0.00000886, 0.00000702, 0.00001646, 0.00001670, 0.00000925, -0.00000128, -0.00000964, -0.00001240, -0.00000926, -0.00000275, 0.00000331, 0.00000588, 0.00000405, -0.00000072, -0.00000555, -0.00000776, -0.00000616, -0.00000160, 0.00000364, 0.00000698, 0.00000683, 0.00000328, -0.00000210, -0.00000706, -0.00000975, -0.00000946, -0.00000679, -0.00000320, -0.00000022, 0.00000115, 0.00000077, -0.00000077, -0.00000263, -0.00000409, -0.00000480, -0.00000479, -0.00000426, -0.00000338, -0.00000229, -0.00000105, 0.00000020, },
		{ -0.00000408, -0.00000285, -0.00000125, 0.00000031, 0.00000132, 0.00000150, 0.00000094, 0.00000022, 0.00000010, 0.00000114, 0.00000320, 0.00000535, 0.00000611, 0.00000418, -0.00000082, -0.00000784, -0.00001458, -0.00001825, -0.00001671, -0.00000948, 0.00000190, 0.00001432, 0.00002420, 0.00002881, 0.00002720, 0.00002032, 0.00001046, 0.00000022, -0.00000843, -0.00001451, -0.00001780, -0.00001835, -0.00001595, -0.00001021, -0.00000115, 0.00001015, 0.00002111, 0.00002810, 0.00002775, 0.00001852, 0.00000192, -0.00001747, -0.00003337, -0.00004015, -0.00003525, -0.00002038, -0.00000111, 0.00001540, 0.00002334, 0.00002073, 0.00001020, -0.00000236, -0.00001057, -0.00001064, -0.00000314, 0.00000733, 0.00001447, 0.00001353, 0.00000384, -0.00001055, -0.00002271, -0.00002608, -0.00001786, -0.00000068, 0.00001826, 0.00003031, 0.00002934, 0.00001474, -0.00000799, -0.00002950, -0.00004058, -0.00003629, -0.00001815, 0.00000643, 0.00002741, 0.00003652, 0.00003075, 0.00001355, -0.00000695, -0.00002163, -0.00002448, -0.00001506, 0.00000154, 0.00001726, 0.00002482, 0.00002092, 0.00000750, -0.00000943, -0.00002266, -0.00002709, -0.00002178, -0.00001002, 0.00000249, 0.00001037, 0.00001099, 0.00000530, -0.00000298, -0.00000940, -0.00001095, -0.00000734, -0.00000089, 0.00000488, 0.00000708, 0.00000480, -0.00000057, -0.00000618, -0.00000921, -0.00000820, -0.00000371, 0.00000212, 0.00000667, 0.00000800, 0.00000573, 0.00000099, -0.00000422, -0.00000800, -0.00000928, -0.00000817, -0.00000562, -0.00000290, -0.00000102, -0.00000038, -0.00000083, -0.00000191, -0.00000314, -0.00000414, -0.00000471, -0.00000471, },
		{ -0.00000333, -0.00000452, -0.00000554, -0.00000594, -0.00000530, -0.00000348, -0.00000084, 0.00000180, 0.00000342, 0.00000326, 0.00000134, -0.00000142, -0.00000351, -0.00000351, -0.00000088, 0.00000356, 0.00000783, 0.00000956, 0.00000710, 0.00000043, -0.00000859, -0.00001681, -0.00002103, -0.00001929, -0.00001168, -0.00000029, 0.00001167, 0.00002113, 0.00002620, 0.00002657, 0.00002315, 0.00001722, 0.00000974, 0.00000116, -0.00000820, -0.00001741, -0.00002455, -0.00002696, -0.00002239, -0.00001036, 0.00000685, 0.00002428, 0.00003584, 0.00003664, 0.00002524, 0.00000470, -0.00001820, -0.00003550, -0.00004119, -0.00003378, -0.00001691, 0.00000219, 0.00001589, 0.00001945, 0.00001303, 0.00000141, -0.00000847, -0.00001096, -0.00000443, 0.00000784, 0.00001929, 0.00002335, 0.00001677, 0.00000145, -0.00001623, -0.00002825, -0.00002874, -0.00001686, 0.00000262, 0.00002135, 0.00003109, 0.00002749, 0.00001203, -0.00000860, -0.00002553, -0.00003168, -0.00002482, -0.00000845, 0.00001005, 0.00002282, 0.00002506, 0.00001702, 0.00000356, -0.00000843, -0.00001331, -0.00000934, 0.00000075, 0.00001127, 0.00001647, 0.00001330, 0.00000280, -0.00001048, -0.00002077, -0.00002380, -0.00001870, -0.00000824, 0.00000262, 0.00000915, 0.00000898, 0.00000300, -0.00000529, -0.00001161, -0.00001299, -0.00000902, -0.00000187, 0.00000493, 0.00000828, 0.00000691, 0.00000183, -0.00000430, -0.00000856, -0.00000913, -0.00000597, -0.00000074, 0.00000418, 0.00000678, 0.00000624, 0.00000311, -0.00000114, -0.00000486, -0.00000698, -0.00000723, -0.00000609, -0.00000439, -0.00000286, -0.00000197, -0.00000182, -0.00000234, },
		{ -0.00000349, -0.00000260, -0.00000209, -0.00000222, -0.00000320, -0.00000496, -0.00000703, -0.00000853, -0.00000851, -0.00000637, -0.00000233, 0.00000243, 0.00000618, 0.00000733, 0.00000523, 0.00000068, -0.00000432, -0.00000736, -0.00000678, -0.00000249, 0.00000379, 0.00000920, 0.00001103, 0.00000791, 0.00000055, -0.00000853, -0.00001606, -0.00001937, -0.00001745, -0.00001117, -0.00000262, 0.00000593, 0.00001305, 0.00001838, 0.00002224, 0.00002474, 0.00002512, 0.00002187, 0.00001357, 0.00000018, -0.00001595, -0.00003032, -0.00003757, -0.00003385, -0.00001880, 0.00000356, 0.00002586, 0.00004006, 0.00004067, 0.00002721, 0.00000461, -0.00001864, -0.00003398, -0.00003633, -0.00002614, -0.00000903, 0.00000692, 0.00001486, 0.00001228, 0.00000199, -0.00000943, -0.00001493, -0.00001057, 0.00000253, 0.00001857, 0.00002986, 0.00003047, 0.00001917, 0.00000018, -0.00001869, -0.00002951, -0.00002785, -0.00001472, 0.00000395, 0.00002007, 0.00002703, 0.00002260, 0.00000963, -0.00000547, -0.00001591, -0.00001759, -0.00001077, 0.00000040, 0.00001020, 0.00001419, 0.00001129, 0.00000405, -0.00000294, -0.00000553, -0.00000224, 0.00000502, 0.00001194, 0.00001408, 0.00000918, -0.00000162, -0.00001419, -0.00002335, -0.00002532, -0.00001939, -0.00000831, 0.00000309, 0.00001010, 0.00001021, 0.00000410, -0.00000488, -0.00001239, -0.00001514, -0.00001223, -0.00000538, 0.00000209, 0.00000694, 0.00000742, 0.00000391, -0.00000148, -0.00000608, -0.00000790, -0.00000644, -0.00000270, 0.00000146, 0.00000431, 0.00000489, 0.00000334, 0.00000058, -0.00000227, -0.00000432, -0.00000524, -0.00000516, -0.00000445, },
		{ -0.00000215, -0.00000340, -0.00000418, -0.00000439, -0.00000395, -0.00000291, -0.00000165, -0.00000085, -0.00000129, -0.00000340, -0.00000684, -0.00001044, -0.00001251, -0.00001161, -0.00000732, -0.00000069, 0.00000605, 0.00001035, 0.00001046, 0.00000633, -0.00000025, -0.00000639, -0.00000937, -0.00000794, -0.00000283, 0.00000349, 0.00000806, 0.00000873, 0.00000521, -0.00000097, -0.00000725, -0.00001136, -0.00001233, -0.00001071, -0.00000796, -0.00000533, -0.00000290, 0.00000046, 0.00000627, 0.00001506, 0.00002529, 0.00003337, 0.00003485, 0.00002663, 0.00000894, -0.00001392, -0.00003464, -0.00004560, -0.00004207, -0.00002456, 0.00000098, 0.00002526, 0.00003936, 0.00003838, 0.00002352, 0.00000145, -0.00001858, -0.00002879, -0.00002618, -0.00001361, 0.00000179, 0.00001215, 0.00001259, 0.00000347, -0.00000987, -0.00001982, -0.00002025, -0.00000951, 0.00000865, 0.00002675, 0.00003699, 0.00003484, 0.00002104, 0.00000124, -0.00001661, -0.00002577, -0.00002354, -0.00001213, 0.00000273, 0.00001458, 0.00001922, 0.00001623, 0.00000871, 0.00000132, -0.00000225, -0.00000096, 0.00000333, 0.00000721, 0.00000781, 0.00000443, -0.00000104, -0.00000523, -0.00000521, -0.00000025, 0.00000753, 0.00001409, 0.00001540, 0.00000953, -0.00000223, -0.00001575, -0.00002587, -0.00002864, -0.00002312, -0.00001178, 0.00000068, 0.00000935, 0.00001127, 0.00000651, -0.00000207, -0.00001033, -0.00001475, -0.00001386, -0.00000866, -0.00000189, 0.00000344, 0.00000538, 0.00000375, -0.00000002, -0.00000380, -0.00000583, -0.00000544, -0.00000315, -0.00000019, 0.00000214, 0.00000311, 0.00000264, 0.00000123, -0.00000053, },
		{ 0.00000133, 0.00000178, 0.00000176, 0.00000131, 0.00000036, -0.00000115, -0.00000303, -0.00000470, -0.00000539, -0.00000459, -0.00000244, 0.00000008, 0.00000146, 0.00000042, -0.00000337, -0.00000887, -0.00001390, -0.00001606, -0.00001382, -0.00000744, 0.00000105, 0.00000855, 0.00001230, 0.00001110, 0.00000585, -0.00000087, -0.00000606, -0.00000769, -0.00000563, -0.00000155, 0.00000206, 0.00000325, 0.00000165, -0.00000145, -0.00000396, -0.00000430, -0.00000248, -0.00000019, 0.00000018, -0.00000286, -0.00000863, -0.00001407, -0.00001495, -0.00000811, 0.00000654, 0.00002509, 0.00004071, 0.00004631, 0.00003778, 0.00001626, -0.00001172, -0.00003651, -0.00004915, -0.00004513, -0.00002637, -0.00000056, 0.00002215, 0.00003332, 0.00002976, 0.00001464, -0.00000418, -0.00001792, -0.00002085, -0.00001277, 0.00000121, 0.00001331, 0.00001692, 0.00000984, -0.00000471, -0.00001963, -0.00002733, -0.00002329, -0.00000817, 0.00001246, 0.00003065, 0.00003954, 0.00003627, 0.00002293, 0.00000532, -0.00000978, -0.00001753, -0.00001671, -0.00000969, -0.00000070, 0.00000639, 0.00000983, 0.00001030, 0.00000991, 0.00001061, 0.00001281, 0.00001515, 0.00001537, 0.00001187, 0.00000492, -0.00000312, -0.00000881, -0.00000938, -0.00000428, 0.00000424, 0.00001201, 0.00001480, 0.00001030, -0.00000062, -0.00001419, -0.00002534, -0.00002985, -0.00002622, -0.00001620, -0.00000398, 0.00000570, 0.00000960, 0.00000711, 0.00000031, -0.00000725, -0.00001223, -0.00001291, -0.00000968, -0.00000454, 0.00000005, 0.00000232, 0.00000188, -0.00000038, -0.00000294, -0.00000449, -0.00000449, -0.00000320, -0.00000137, 0.00000027, },
		{ -0.00000384, -0.00000354, -0.00000304, -0.00000246, -0.00000160, -0.00000024, 0.00000156, 0.00000325, 0.00000393, 0.00000287, 0.00000003, -0.00000370, -0.00000675, -0.00000763, -0.00000574, -0.00000187, 0.00000203, 0.00000368, 0.00000166, -0.00000378, -0.00001065, -0.00001599, -0.00001727, -0.00001359, -0.00000619, 0.00000212, 0.00000822, 0.00001016, 0.00000800, 0.00000362, -0.00000035, -0.00000201, -0.00000120, 0.00000056, 0.00000105, -0.00000116, -0.00000563, -0.00001004, -0.00001135, -0.00000761, 0.00000060, 0.00000988, 0.00001541, 0.00001328, 0.00000279, -0.00001256, -0.00002607, -0.00003063, -0.00002208, -0.00000158, 0.00002427, 0.00004573, 0.00005393, 0.00004472, 0.00002071, -0.00000960, -0.00003523, -0.00004721, -0.00004227, -0.00002402, -0.00000110, 0.00001670, 0.00002276, 0.00001631, 0.00000234, -0.00001109, -0.00001682, -0.00001195, 0.00000090, 0.00001506, 0.00002309, 0.00002045, 0.00000763, -0.00001006, -0.00002488, -0.00003024, -0.00002349, -0.00000698, 0.00001316, 0.00002977, 0.00003763, 0.00003533, 0.00002529, 0.00001218, 0.00000068, -0.00000639, -0.00000867, -0.00000761, -0.00000505, -0.00000205, 0.00000155, 0.00000651, 0.00001316, 0.00002054, 0.00002642, 0.00002820, 0.00002431, 0.00001528, 0.00000390, -0.00000588, -0.00001061, -0.00000902, -0.00000268, 0.00000463, 0.00000863, 0.00000656, -0.00000159, -0.00001297, -0.00002318, -0.00002822, -0.00002626, -0.00001833, -0.00000778, 0.00000131, 0.00000587, 0.00000505, 0.00000028, -0.00000564, -0.00000992, -0.00001099, -0.00000898, -0.00000534, -0.00000192, -0.00000004, -0.00000001, -0.00000122, -0.00000268, -0.00000363, },
		{ -0.00000159, -0.00000171, -0.00000181, -0.00000179, -0.00000199, -0.00000283, -0.00000434, -0.00000592, -0.00000647, -0.00000506, -0.00000155, 0.00000305, 0.00000689, 0.00000813, 0.00000587, 0.00000076, -0.00000511, -0.00000920, -0.00000969, -0.00000648, -0.00000132, 0.00000301, 0.00000406, 0.00000095, -0.00000512, -0.00001142, -0.00001505, -0.00001437, -0.00000979, -0.00000355, 0.00000157, 0.00000370, 0.00000286, 0.00000085, 0.00000012, 0.00000213, 0.00000627, 0.00000994, 0.00000987, 0.00000409, -0.00000649, -0.00001795, -0.00002481, -0.00002269, -0.00001074, 0.00000725, 0.00002403, 0.00003184, 0.00002602, 0.00000755, -0.00001679, -0.00003694, -0.00004375, -0.00003307, -0.00000785, 0.00002279, 0.00004720, 0.00005603, 0.00004610, 0.00002164, -0.00000762, -0.00003071, -0.00004001, -0.00003419, -0.00001829, -0.00000097, 0.00000966, 0.00000966, 0.00000078, -0.00001084, -0.00001793, -0.00001573, -0.00000429, 0.00001161, 0.00002481, 0.00002913, 0.00002226, 0.00000676, -0.00001117, -0.00002443, -0.00002803, -0.00002093, -0.00000612, 0.00001101, 0.00002506, 0.00003263, 0.00003314, 0.00002828, 0.00002072, 0.00001270, 0.00000544, -0.00000073, -0.00000559, -0.00000845, -0.00000806, -0.00000331, 0.00000586, 0.00001777, 0.00002922, 0.00003653, 0.00003707, 0.00003047, 0.00001894, 0.00000640, -0.00000317, -0.00000747, -0.00000674, -0.00000352, -0.00000129, -0.00000272, -0.00000836, -0.00001641, -0.00002360, -0.00002683, -0.00002460, -0.00001770, -0.00000880, -0.00000119, 0.00000270, 0.00000222, -0.00000140, -0.00000590, -0.00000909, -0.00000985, -0.00000835, -0.00000576, -0.00000337, -0.00000198, },
		{ -0.00000802, -0.00000661, -0.00000560, -0.00000510, -0.00000455, -0.00000330, -0.00000120, 0.00000104, 0.00000215, 0.00000098, -0.00000260, -0.00000742, -0.00001128, -0.00001201, -0.00000857, -0.00000181, 0.00000580, 0.00001116, 0.00001198, 0.00000794, 0.00000086, -0.00000613, -0.00001013, -0.00000981, -0.00000604, -0.00000134, 0.00000148, 0.00000082, -0.00000287, -0.00000744, -0.00001025, -0.00000974, -0.00000632, -0.00000220, -0.00000008, -0.00000147, -0.00000556, -0.00000942, -0.00000948, -0.00000367, 0.00000698, 0.00001822, 0.00002424, 0.00002053, 0.00000644, -0.00001373, -0.00003201, -0.00004009, -0.00003308, -0.00001226, 0.00001494, 0.00003763, 0.00004604, 0.00003582, 0.00001029, -0.00002071, -0.00004473, -0.00005193, -0.00003919, -0.00001143, 0.00002051, 0.00004450, 0.00005210, 0.00004181, 0.00001924, -0.00000576, -0.00002361, -0.00002901, -0.00002278, -0.00001075, -0.00000054, 0.00000229, -0.00000330, -0.00001358, -0.00002220, -0.00002349, -0.00001527, 0.00000009, 0.00001681, 0.00002832, 0.00003025, 0.00002213, 0.00000747, -0.00000798, -0.00001871, -0.00002147, -0.00001618, -0.00000536, 0.00000733, 0.00001867, 0.00002673, 0.00003099, 0.00003172, 0.00002929, 0.00002390, 0.00001590, 0.00000629, -0.00000295, -0.00000909, -0.00000963, -0.00000342, 0.00000848, 0.00002292, 0.00003560, 0.00004272, 0.00004236, 0.00003521, 0.00002405, 0.00001245, 0.00000321, -0.00000273, -0.00000627, -0.00000937, -0.00001366, -0.00001943, -0.00002539, -0.00002933, -0.00002931, -0.00002479, -0.00001697, -0.00000838, -0.00000172, 0.00000126, 0.00000046, -0.00000274, -0.00000632, -0.00000862, -0.00000903, },
		{ -0.00000365, -0.00000605, -0.00000703, -0.00000667, -0.00000601, -0.00000619, -0.00000759, -0.00000942, -0.00001013, -0.00000833, -0.00000383, 0.00000197, 0.00000655, 0.00000744, 0.00000359, -0.00000388, -0.00001197, -0.00001708, -0.00001666, -0.00001049, -0.00000090, 0.00000831, 0.00001358, 0.00001318, 0.00000797, 0.00000079, -0.00000507, -0.00000751, -0.00000656, -0.00000416, -0.00000279, -0.00000394, -0.00000711, -0.00001007, -0.00001022, -0.00000631, 0.00000045, 0.00000664, 0.00000830, 0.00000322, -0.00000738, -0.00001889, -0.00002507, -0.00002116, -0.00000660, 0.00001396, 0.00003209, 0.00003912, 0.00003021, 0.00000711, -0.00002197, -0.00004537, -0.00005285, -0.00004023, -0.00001149, 0.00002247, 0.00004819, 0.00005530, 0.00004083, 0.00001054, -0.00002351, -0.00004801, -0.00005388, -0.00003978, -0.00001228, 0.00001731, 0.00003787, 0.00004297, 0.00003305, 0.00001443, -0.00000407, -0.00001528, -0.00001665, -0.00001071, -0.00000326, -0.00000006, -0.00000392, -0.00001344, -0.00002388, -0.00002964, -0.00002701, -0.00001594, 0.00000002, 0.00001535, 0.00002497, 0.00002629, 0.00001993, 0.00000911, -0.00000195, -0.00000972, -0.00001243, -0.00001014, -0.00000409, 0.00000417, 0.00001332, 0.00002232, 0.00003008, 0.00003527, 0.00003639, 0.00003246, 0.00002374, 0.00001216, 0.00000104, -0.00000591, -0.00000604, 0.00000120, 0.00001386, 0.00002821, 0.00004008, 0.00004641, 0.00004612, 0.00004016, 0.00003069, 0.00001993, 0.00000924, -0.00000104, -0.00001117, -0.00002116, -0.00003021, -0.00003674, -0.00003901, -0.00003613, -0.00002868, -0.00001876, -0.00000925, -0.00000264, -0.00000007, -0.00000100, },
		{ -0.00001191, -0.00000536, -0.00000342, -0.00000469, -0.00000669, -0.00000723, -0.00000556, -0.00000273, -0.00000086, -0.00000182, -0.00000600, -0.00001176, -0.00001618, -0.00001642, -0.00001139, -0.00000249, 0.00000675, 0.00001228, 0.00001140, 0.00000418, -0.00000644, -0.00001595, -0.00002031, -0.00001778, -0.00000968, 0.00000039, 0.00000836, 0.00001157, 0.00000992, 0.00000558, 0.00000152, -0.00000028, 0.00000007, 0.00000057, -0.00000125, -0.00000652, -0.00001384, -0.00001964, -0.00001989, -0.00001256, 0.00000076, 0.00001483, 0.00002291, 0.00001989, 0.00000530, -0.00001580, -0.00003441, -0.00004141, -0.00003184, -0.00000774, 0.00002207, 0.00004533, 0.00005154, 0.00003674, 0.00000567, -0.00002979, -0.00005538, -0.00006047, -0.00004255, -0.00000846, 0.00002843, 0.00005374, 0.00005799, 0.00004032, 0.00000853, -0.00002450, -0.00004629, -0.00004972, -0.00003547, -0.00001097, 0.00001328, 0.00002841, 0.00003059, 0.00002189, 0.00000848, -0.00000275, -0.00000748, -0.00000558, -0.00000057, 0.00000262, 0.00000040, -0.00000768, -0.00001882, -0.00002837, -0.00003213, -0.00002824, -0.00001784, -0.00000441, 0.00000785, 0.00001574, 0.00001803, 0.00001552, 0.00001022, 0.00000433, -0.00000050, -0.00000336, -0.00000383, -0.00000163, 0.00000349, 0.00001142, 0.00002130, 0.00003126, 0.00003875, 0.00004134, 0.00003781, 0.00002888, 0.00001725, 0.00000673, 0.00000091, 0.00000188, 0.00000942, 0.00002130, 0.00003411, 0.00004457, 0.00005044, 0.00005091, 0.00004625, 0.00003729, 0.00002486, 0.00000978, -0.00000689, -0.00002347, -0.00003762, -0.00004679, -0.00004916, -0.00004451, -0.00003454, -0.00002249, },
		{ -0.00005855, -0.00005233, -0.00003963, -0.00002511, -0.00001347, -0.00000765, -0.00000771, -0.00001109, -0.00001396, -0.00001327, -0.00000826, -0.00000093, 0.00000504, 0.00000618, 0.00000121, -0.00000816, -0.00001776, -0.00002298, -0.00002094, -0.00001200, 0.00000026, 0.00001067, 0.00001485, 0.00001125, 0.00000181, -0.00000904, -0.00001657, -0.00001795, -0.00001344, -0.00000597, 0.00000064, 0.00000377, 0.00000334, 0.00000153, 0.00000129, 0.00000419, 0.00000916, 0.00001273, 0.00001091, 0.00000173, -0.00001299, -0.00002768, -0.00003527, -0.00003060, -0.00001346, 0.00001048, 0.00003147, 0.00003979, 0.00003031, 0.00000538, -0.00002552, -0.00004940, -0.00005530, -0.00003933, -0.00000677, 0.00002971, 0.00005520, 0.00005886, 0.00003866, 0.00000237, -0.00003555, -0.00005997, -0.00006137, -0.00003963, -0.00000386, 0.00003163, 0.00005333, 0.00005395, 0.00003489, 0.00000506, -0.00002339, -0.00004019, -0.00004087, -0.00002785, -0.00000845, 0.00000879, 0.00001789, 0.00001751, 0.00001068, 0.00000263, -0.00000212, -0.00000183, 0.00000198, 0.00000575, 0.00000595, 0.00000091, -0.00000836, -0.00001882, -0.00002686, -0.00002990, -0.00002736, -0.00002053, -0.00001180, -0.00000350, 0.00000289, 0.00000690, 0.00000877, 0.00000895, 0.00000782, 0.00000571, 0.00000328, 0.00000165, 0.00000227, 0.00000637, 0.00001420, 0.00002458, 0.00003503, 0.00004255, 0.00004476, 0.00004096, 0.00003248, 0.00002227, 0.00001385, 0.00001008, 0.00001220, 0.00001967, 0.00003052, 0.00004211, 0.00005190, 0.00005785, 0.00005848, 0.00005285, 0.00004065, 0.00002250, 0.00000025, -0.00002285, -0.00004270, -0.00005540, },
		{ 0.00000888, -0.00002152, -0.00004830, -0.00006562, -0.00006997, -0.00006159, -0.00004457, -0.00002553, -0.00001108, -0.00000531, -0.00000827, -0.00001612, -0.00002310, -0.00002428, -0.00001787, -0.00000608, 0.00000594, 0.00001267, 0.00001085, 0.00000119, -0.00001196, -0.00002262, -0.00002599, -0.00002070, -0.00000937, 0.00000279, 0.00001050, 0.00001099, 0.00000511, -0.00000333, -0.00000975, -0.00001125, -0.00000804, -0.00000310, -0.00000026, -0.00000177, -0.00000672, -0.00001138, -0.00001120, -0.00000369, 0.00000956, 0.00002293, 0.00002908, 0.00002271, 0.00000369, -0.00002186, -0.00004362, -0.00005140, -0.00004002, -0.00001226, 0.00002156, 0.00004754, 0.00005415, 0.00003746, 0.00000332, -0.00003472, -0.00006091, -0.00006402, -0.00004223, -0.00000406, 0.00003512, 0.00005952, 0.00005955, 0.00003569, -0.00000192, -0.00003790, -0.00005815, -0.00005551, -0.00003232, 0.00000116, 0.00003139, 0.00004731, 0.00004446, 0.00002612, 0.00000113, -0.00002039, -0.00003125, -0.00002965, -0.00001902, -0.00000565, 0.00000447, 0.00000823, 0.00000614, 0.00000140, -0.00000226, -0.00000249, 0.00000061, 0.00000493, 0.00000760, 0.00000657, 0.00000151, -0.00000621, -0.00001432, -0.00002070, -0.00002420, -0.00002476, -0.00002302, -0.00001978, -0.00001559, -0.00001071, -0.00000544, -0.00000033, 0.00000376, 0.00000605, 0.00000636, 0.00000549, 0.00000506, 0.00000695, 0.00001239, 0.00002126, 0.00003188, 0.00004159, 0.00004770, 0.00004857, 0.00004424, 0.00003646, 0.00002800, 0.00002173, 0.00001975, 0.00002292, 0.00003083, 0.00004198, 0.00005416, 0.00006461, 0.00007034, 0.00006857, 0.00005741, 0.00003675, },
		{ 0.00008342, 0.00008688, 0.00007821, 0.00005586, 0.00002199, -0.00001745, -0.00005396, -0.00007898, -0.00008687, -0.00007705, -0.00005445, -0.00002789, -0.00000676, 0.00000245, -0.00000140, -0.00001396, -0.00002754, -0.00003462, -0.00003120, -0.00001841, -0.00000184, 0.00001131, 0.00001559, 0.00000982, -0.00000257, -0.00001535, -0.00002262, -0.00002162, -0.00001384, -0.00000398, 0.00000274, 0.00000337, -0.00000128, -0.00000728, -0.00000992, -0.00000659, 0.00000144, 0.00000953, 0.00001206, 0.00000577, -0.00000792, -0.00002290, -0.00003102, -0.00002617, -0.00000787, 0.00001759, 0.00003930, 0.00004647, 0.00003362, 0.00000380, -0.00003196, -0.00005891, -0.00006493, -0.00004610, -0.00000890, 0.00003205, 0.00006002, 0.00006324, 0.00004002, -0.00000034, -0.00004143, -0.00006660, -0.00006593, -0.00004025, -0.00000056, 0.00003685, 0.00005727, 0.00005358, 0.00002874, -0.00000599, -0.00003617, -0.00005032, -0.00004431, -0.00002237, 0.00000542, 0.00002785, 0.00003718, 0.00003181, 0.00001610, -0.00000233, -0.00001621, -0.00002152, -0.00001856, -0.00001092, -0.00000327, 0.00000099, 0.00000097, -0.00000185, -0.00000483, -0.00000576, -0.00000393, -0.00000024, 0.00000350, 0.00000554, 0.00000501, 0.00000208, -0.00000238, -0.00000739, -0.00001232, -0.00001688, -0.00002093, -0.00002412, -0.00002581, -0.00002530, -0.00002231, -0.00001728, -0.00001135, -0.00000595, -0.00000208, 0.00000024, 0.00000205, 0.00000501, 0.00001064, 0.00001942, 0.00003050, 0.00004181, 0.00005088, 0.00005567, 0.00005534, 0.00005050, 0.00004298, 0.00003524, 0.00002981, 0.00002871, 0.00003311, 0.00004299, 0.00005689, 0.00007181, },
		{ 0.00003760, 0.00004401, 0.00005786, 0.00007658, 0.00009495, 0.00010599, 0.00010295, 0.00008186, 0.00004362, -0.00000533, -0.00005428, -0.00009138, -0.00010768, -0.00010035, -0.00007398, -0.00003897, -0.00000777, 0.00000980, 0.00001005, -0.00000359, -0.00002246, -0.00003679, -0.00003989, -0.00003084, -0.00001439, 0.00000174, 0.00001070, 0.00000972, 0.00000113, -0.00000929, -0.00001552, -0.00001454, -0.00000783, -0.00000032, 0.00000249, -0.00000210, -0.00001215, -0.00002182, -0.00002434, -0.00001592, 0.00000163, 0.00002106, 0.00003274, 0.00002941, 0.00001026, -0.00001786, -0.00004280, -0.00005246, -0.00004054, -0.00001004, 0.00002724, 0.00005545, 0.00006157, 0.00004137, 0.00000183, -0.00004142, -0.00007053, -0.00007311, -0.00004757, -0.00000404, 0.00003987, 0.00006648, 0.00006541, 0.00003778, -0.00000451, -0.00004406, -0.00006533, -0.00006095, -0.00003429, 0.00000254, 0.00003427, 0.00004891, 0.00004242, 0.00001958, -0.00000872, -0.00003062, -0.00003819, -0.00003025, -0.00001197, 0.00000806, 0.00002185, 0.00002517, 0.00001868, 0.00000680, -0.00000480, -0.00001183, -0.00001294, -0.00000961, -0.00000489, -0.00000162, -0.00000118, -0.00000315, -0.00000595, -0.00000786, -0.00000791, -0.00000619, -0.00000350, -0.00000084, 0.00000113, 0.00000217, 0.00000230, 0.00000146, -0.00000066, -0.00000439, -0.00000981, -0.00001637, -0.00002289, -0.00002794, -0.00003040, -0.00002994, -0.00002715, -0.00002323, -0.00001930, -0.00001581, -0.00001226, -0.00000743, -0.00000002, 0.00001063, 0.00002405, 0.00003868, 0.00005222, 0.00006242, 0.00006763, 0.00006733, 0.00006214, 0.00005381, 0.00004494, 0.00003857, },
		{ 0.00008525, 0.00008011, 0.00007008, 0.00005866, 0.00005043, 0.00004981, 0.00005940, 0.00007835, 0.00010166, 0.00012084, 0.00012633, 0.00011082, 0.00007257, 0.00001705, -0.00004389, -0.00009547, -0.00012487, -0.00012581, -0.00010092, -0.00006082, -0.00002011, 0.00000807, 0.00001684, 0.00000755, -0.00001157, -0.00002963, -0.00003803, -0.00003401, -0.00002118, -0.00000706, 0.00000118, 0.00000038, -0.00000711, -0.00001502, -0.00001691, -0.00000996, 0.00000308, 0.00001501, 0.00001794, 0.00000777, -0.00001293, -0.00003533, -0.00004813, -0.00004303, -0.00001934, 0.00001474, 0.00004511, 0.00005781, 0.00004559, 0.00001179, -0.00003037, -0.00006312, -0.00007180, -0.00005152, -0.00000972, 0.00003664, 0.00006816, 0.00007134, 0.00004443, -0.00000170, -0.00004816, -0.00007601, -0.00007425, -0.00004424, 0.00000118, 0.00004335, 0.00006572, 0.00006057, 0.00003166, -0.00000795, -0.00004192, -0.00005753, -0.00005051, -0.00002607, 0.00000420, 0.00002774, 0.00003625, 0.00002858, 0.00001031, -0.00000945, -0.00002238, -0.00002427, -0.00001615, -0.00000308, 0.00000875, 0.00001473, 0.00001362, 0.00000734, -0.00000041, -0.00000617, -0.00000819, -0.00000682, -0.00000387, -0.00000139, -0.00000072, -0.00000198, -0.00000437, -0.00000677, -0.00000839, -0.00000896, -0.00000862, -0.00000760, -0.00000599, -0.00000376, -0.00000101, 0.00000176, 0.00000365, 0.00000360, 0.00000089, -0.00000447, -0.00001162, -0.00001913, -0.00002560, -0.00003015, -0.00003271, -0.00003382, -0.00003415, -0.00003395, -0.00003271, -0.00002933, -0.00002251, -0.00001139, 0.00000398, 0.00002251, 0.00004219, 0.00006052, 0.00007501, 0.00008362, },
		{ 0.00001182, 0.00003794, 0.00006394, 0.00008629, 0.00010157, 0.00010740, 0.00010349, 0.00009234, 0.00007911, 0.00007034, 0.00007163, 0.00008513, 0.00010772, 0.00013103, 0.00014371, 0.00013538, 0.00010098, 0.00004368, -0.00002512, -0.00008899, -0.00013195, -0.00014420, -0.00012566, -0.00008595, -0.00004056, -0.00000484, 0.00001175, 0.00000866, -0.00000665, -0.00002291, -0.00003068, -0.00002664, -0.00001455, -0.00000269, 0.00000099, -0.00000648, -0.00002122, -0.00003428, -0.00003630, -0.00002284, 0.00000257, 0.00002923, 0.00004396, 0.00003754, 0.00000980, -0.00002926, -0.00006312, -0.00007588, -0.00005968, -0.00001889, 0.00003099, 0.00006954, 0.00008014, 0.00005741, 0.00000992, -0.00004316, -0.00007998, -0.00008534, -0.00005715, -0.00000727, 0.00004359, 0.00007466, 0.00007382, 0.00004242, -0.00000573, -0.00005050, -0.00007410, -0.00006824, -0.00003703, 0.00000548, 0.00004174, 0.00005811, 0.00005011, 0.00002336, -0.00000966, -0.00003543, -0.00004501, -0.00003710, -0.00001757, 0.00000394, 0.00001861, 0.00002191, 0.00001489, 0.00000274, -0.00000814, -0.00001315, -0.00001116, -0.00000443, 0.00000306, 0.00000771, 0.00000789, 0.00000430, -0.00000083, -0.00000499, -0.00000669, -0.00000586, -0.00000358, -0.00000131, -0.00000015, -0.00000042, -0.00000180, -0.00000371, -0.00000571, -0.00000756, -0.00000919, -0.00001041, -0.00001081, -0.00000992, -0.00000749, -0.00000383, 0.00000015, 0.00000319, 0.00000414, 0.00000244, -0.00000168, -0.00000740, -0.00001368, -0.00001976, -0.00002541, -0.00003085, -0.00003636, -0.00004185, -0.00004661, -0.00004928, -0.00004824, -0.00004202, -0.00002974, -0.00001144, },
		{ -0.00006194, -0.00006598, -0.00006403, -0.00005435, -0.00003597, -0.00000918, 0.00002398, 0.00005957, 0.00009234, 0.00011687, 0.00012944, 0.00012961, 0.00012098, 0.00011040, 0.00010547, 0.00011133, 0.00012776, 0.00014819, 0.00016134, 0.00015534, 0.00012273, 0.00006448, -0.00000906, -0.00008091, -0.00013319, -0.00015370, -0.00014048, -0.00010255, -0.00005620, -0.00001841, -0.00000014, -0.00000246, -0.00001708, -0.00003115, -0.00003386, -0.00002185, -0.00000082, 0.00001754, 0.00002172, 0.00000675, -0.00002245, -0.00005266, -0.00006825, -0.00005882, -0.00002488, 0.00002128, 0.00006029, 0.00007414, 0.00005456, 0.00000745, -0.00004880, -0.00009078, -0.00009997, -0.00007103, -0.00001448, 0.00004736, 0.00008966, 0.00009544, 0.00006272, 0.00000524, -0.00005346, -0.00008996, -0.00009060, -0.00005677, -0.00000376, 0.00004622, 0.00007345, 0.00006865, 0.00003612, -0.00000900, -0.00004769, -0.00006515, -0.00005646, -0.00002765, 0.00000785, 0.00003546, 0.00004550, 0.00003652, 0.00001481, -0.00000924, -0.00002602, -0.00003049, -0.00002353, -0.00001048, 0.00000193, 0.00000876, 0.00000866, 0.00000375, -0.00000207, -0.00000536, -0.00000467, -0.00000092, 0.00000348, 0.00000607, 0.00000559, 0.00000239, -0.00000193, -0.00000554, -0.00000720, -0.00000674, -0.00000489, -0.00000273, -0.00000111, -0.00000042, -0.00000057, -0.00000134, -0.00000258, -0.00000428, -0.00000644, -0.00000882, -0.00001088, -0.00001191, -0.00001136, -0.00000917, -0.00000587, -0.00000242, 0.00000015, 0.00000116, 0.00000046, -0.00000163, -0.00000463, -0.00000829, -0.00001269, -0.00001826, -0.00002541, -0.00003423, -0.00004415, -0.00005398, },
		{ -0.00001126, -0.00001794, -0.00002754, -0.00003980, -0.00005385, -0.00006802, -0.00007978, -0.00008582, -0.00008263, -0.00006740, -0.00003928, -0.00000033, 0.00004435, 0.00008767, 0.00012268, 0.00014507, 0.00015491, 0.00015680, 0.00015783, 0.00016412, 0.00017730, 0.00019257, 0.00019974, 0.00018705, 0.00014662, 0.00007914, -0.00000454, -0.00008592, -0.00014505, -0.00016796, -0.00015232, -0.00010865, -0.00005631, -0.00001579, -0.00000039, -0.00001099, -0.00003630, -0.00005851, -0.00006191, -0.00004064, -0.00000180, 0.00003747, 0.00005792, 0.00004753, 0.00000789, -0.00004554, -0.00008955, -0.00010336, -0.00007822, -0.00002201, 0.00004334, 0.00009083, 0.00009994, 0.00006564, 0.00000103, -0.00006790, -0.00011301, -0.00011579, -0.00007505, -0.00000729, 0.00006036, 0.00010141, 0.00010089, 0.00006093, -0.00000076, -0.00005885, -0.00009105, -0.00008688, -0.00005109, -0.00000054, 0.00004358, 0.00006470, 0.00005712, 0.00002720, -0.00001052, -0.00004014, -0.00005101, -0.00004141, -0.00001803, 0.00000793, 0.00002600, 0.00003062, 0.00002259, 0.00000762, -0.00000699, -0.00001572, -0.00001680, -0.00001217, -0.00000574, -0.00000110, 0.00000012, -0.00000138, -0.00000346, -0.00000397, -0.00000201, 0.00000172, 0.00000544, 0.00000731, 0.00000641, 0.00000310, -0.00000130, -0.00000519, -0.00000744, -0.00000776, -0.00000664, -0.00000489, -0.00000324, -0.00000205, -0.00000136, -0.00000111, -0.00000132, -0.00000212, -0.00000362, -0.00000571, -0.00000797, -0.00000979, -0.00001061, -0.00001021, -0.00000880, -0.00000696, -0.00000531, -0.00000426, -0.00000384, -0.00000385, -0.00000401, -0.00000433, -0.00000517, -0.00000721, },
		{ -0.00000877, -0.00000825, -0.00000724, -0.00000618, -0.00000570, -0.00000668, -0.00001016, -0.00001734, -0.00002920, -0.00004599, -0.00006645, -0.00008739, -0.00010379, -0.00010992, -0.00010100, -0.00007516, -0.00003454, 0.00001502, 0.00006582, 0.00011108, 0.00014749, 0.00017616, 0.00020131, 0.00022705, 0.00025370, 0.00027554, 0.00028154, 0.00025938, 0.00020137, 0.00010988, -0.00000065, -0.00010658, -0.00018237, -0.00020989, -0.00018576, -0.00012354, -0.00004927, 0.00000827, 0.00002878, 0.00000847, -0.00003857, -0.00008667, -0.00010983, -0.00009336, -0.00004069, 0.00002763, 0.00008271, 0.00009987, 0.00006988, 0.00000358, -0.00007214, -0.00012548, -0.00013324, -0.00009090, -0.00001482, 0.00006448, 0.00011496, 0.00011644, 0.00006888, -0.00000784, -0.00008247, -0.00012535, -0.00012049, -0.00007150, 0.00000066, 0.00006697, 0.00010242, 0.00009576, 0.00005300, -0.00000631, -0.00005796, -0.00008316, -0.00007542, -0.00004185, 0.00000124, 0.00003597, 0.00005024, 0.00004177, 0.00001767, -0.00000988, -0.00002936, -0.00003451, -0.00002597, -0.00000981, 0.00000604, 0.00001547, 0.00001640, 0.00001075, 0.00000269, -0.00000377, -0.00000659, -0.00000616, -0.00000443, -0.00000345, -0.00000414, -0.00000590, -0.00000720, -0.00000657, -0.00000356, 0.00000101, 0.00000544, 0.00000802, 0.00000782, 0.00000504, 0.00000083, -0.00000335, -0.00000634, -0.00000766, -0.00000753, -0.00000652, -0.00000524, -0.00000405, -0.00000309, -0.00000239, -0.00000203, -0.00000214, -0.00000284, -0.00000407, -0.00000554, -0.00000684, -0.00000760, -0.00000773, -0.00000742, -0.00000707, -0.00000703, -0.00000744, -0.00000811, -0.00000867, },
		{ -0.00000499, -0.00000603, -0.00000761, -0.00000945, -0.00001125, -0.00001273, -0.00001358, -0.00001345, -0.00001199, -0.00000920, -0.00000586, -0.00000383, -0.00000586, -0.00001484, -0.00003245, -0.00005786, -0.00008711, -0.00011362, -0.00012998, -0.00013029, -0.00011222, -0.00007765, -0.00003154, 0.00002054, 0.00007498, 0.00013143, 0.00019170, 0.00025667, 0.00032247, 0.00037826, 0.00040728, 0.00039193, 0.00032125, 0.00019790, 0.00004128, -0.00011568, -0.00023567, -0.00028952, -0.00026710, -0.00018224, -0.00006879, 0.00003109, 0.00008236, 0.00007024, 0.00000563, -0.00007967, -0.00014609, -0.00016267, -0.00011990, -0.00003364, 0.00006158, 0.00012700, 0.00013590, 0.00008480, -0.00000515, -0.00009703, -0.00015356, -0.00015236, -0.00009489, -0.00000529, 0.00008009, 0.00012779, 0.00012070, 0.00006428, -0.00001672, -0.00008918, -0.00012536, -0.00011358, -0.00006179, 0.00000711, 0.00006556, 0.00009262, 0.00008157, 0.00004109, -0.00000990, -0.00005092, -0.00006814, -0.00005894, -0.00003129, 0.00000103, 0.00002491, 0.00003299, 0.00002563, 0.00000928, -0.00000739, -0.00001755, -0.00001867, -0.00001266, -0.00000395, 0.00000306, 0.00000600, 0.00000512, 0.00000241, 0.00000010, -0.00000063, -0.00000011, 0.00000035, -0.00000050, -0.00000301, -0.00000629, -0.00000876, -0.00000898, -0.00000645, -0.00000190, 0.00000306, 0.00000668, 0.00000781, 0.00000632, 0.00000299, -0.00000094, -0.00000431, -0.00000647, -0.00000731, -0.00000714, -0.00000640, -0.00000544, -0.00000449, -0.00000371, -0.00000324, -0.00000320, -0.00000362, -0.00000437, -0.00000516, -0.00000568, -0.00000572, -0.00000535, -0.00000485, -0.00000462, },
		{ -0.00000574, -0.00000512, -0.00000434, -0.00000369, -0.00000337, -0.00000344, -0.00000396, -0.00000511, -0.00000719, -0.00001044, -0.00001470, -0.00001904, -0.00002189, -0.00002150, -0.00001695, -0.00000904, -0.00000072, 0.00000346, -0.00000111, -0.00001719, -0.00004416, -0.00007764, -0.00011072, -0.00013613, -0.00014859, -0.00014595, -0.00012867, -0.00009774, -0.00005247, 0.00001039, 0.00009496, 0.00020218, 0.00032500, 0.00044550, 0.00053632, 0.00056728, 0.00051585, 0.00037764, 0.00017253, -0.00005726, -0.00025776, -0.00037977, -0.00039562, -0.00030959, -0.00015764, 0.00000454, 0.00012087, 0.00015487, 0.00010234, -0.00000813, -0.00012714, -0.00020430, -0.00020801, -0.00013734, -0.00002138, 0.00009331, 0.00016198, 0.00015961, 0.00009023, -0.00001574, -0.00011478, -0.00016823, -0.00015746, -0.00009025, 0.00000382, 0.00008654, 0.00012680, 0.00011241, 0.00005354, -0.00002313, -0.00008640, -0.00011321, -0.00009700, -0.00004850, 0.00001035, 0.00005634, 0.00007419, 0.00006150, 0.00002784, -0.00001080, -0.00003931, -0.00004921, -0.00004082, -0.00002143, -0.00000095, 0.00001266, 0.00001620, 0.00001145, 0.00000313, -0.00000390, -0.00000691, -0.00000595, -0.00000304, -0.00000059, 0.00000006, -0.00000086, -0.00000204, -0.00000214, -0.00000071, 0.00000153, 0.00000316, 0.00000289, 0.00000040, -0.00000346, -0.00000707, -0.00000882, -0.00000787, -0.00000452, -0.00000004, 0.00000396, 0.00000620, 0.00000616, 0.00000415, 0.00000104, -0.00000218, -0.00000479, -0.00000642, -0.00000706, -0.00000692, -0.00000627, -0.00000538, -0.00000455, -0.00000402, -0.00000396, -0.00000437, -0.00000504, -0.00000568, -0.00000596, },
		{ -0.00000467, -0.00000505, -0.00000541, -0.00000559, -0.00000557, -0.00000550, -0.00000552, -0.00000557, -0.00000536, -0.00000450, -0.00000290, -0.00000107, -0.00000018, -0.00000167, -0.00000651, -0.00001438, -0.00002333, -0.00003019, -0.00003181, -0.00002657, -0.00001557, -0.00000273, 0.00000639, 0.00000657, -0.00000500, -0.00002771, -0.00005801, -0.00009109, -0.00012286, -0.00015083, -0.00017327, -0.00018686, -0.00018429, -0.00015374, -0.00008171, 0.00004076, 0.00021061, 0.00040728, 0.00059284, 0.00071952, 0.00074357, 0.00064121, 0.00042071, 0.00012507, -0.00017733, -0.00041215, -0.00052257, -0.00048835, -0.00033320, -0.00011707, 0.00008436, 0.00020590, 0.00021489, 0.00012117, -0.00002890, -0.00017161, -0.00025073, -0.00023834, -0.00014374, -0.00000756, 0.00011596, 0.00018093, 0.00016717, 0.00008631, -0.00002499, -0.00012126, -0.00016664, -0.00014819, -0.00007885, 0.00000991, 0.00008225, 0.00011242, 0.00009384, 0.00003949, -0.00002538, -0.00007464, -0.00009166, -0.00007448, -0.00003457, 0.00000970, 0.00004130, 0.00005097, 0.00003953, 0.00001562, -0.00000921, -0.00002569, -0.00003006, -0.00002425, -0.00001363, -0.00000387, 0.00000163, 0.00000260, 0.00000105, -0.00000049, -0.00000049, 0.00000090, 0.00000236, 0.00000248, 0.00000077, -0.00000206, -0.00000450, -0.00000517, -0.00000363, -0.00000062, 0.00000229, 0.00000360, 0.00000258, -0.00000036, -0.00000392, -0.00000652, -0.00000707, -0.00000539, -0.00000219, 0.00000130, 0.00000391, 0.00000491, 0.00000420, 0.00000221, -0.00000041, -0.00000302, -0.00000513, -0.00000649, -0.00000703, -0.00000686, -0.00000622, -0.00000542, -0.00000478, -0.00000453, },
		{ -0.00000683, -0.00000651, -0.00000617, -0.00000593, -0.00000574, -0.00000539, -0.00000475, -0.00000395, -0.00000341, -0.00000365, -0.00000500, -0.00000716, -0.00000917, -0.00000970, -0.00000777, -0.00000341, 0.00000190, 0.00000560, 0.00000508, -0.00000106, -0.00001202, -0.00002473, -0.00003489, -0.00003875, -0.00003478, -0.00002445, -0.00001172, -0.00000133, 0.00000306, -0.00000009, -0.00001048, -0.00002763, -0.00005247, -0.00008759, -0.00013527, -0.00019344, -0.00025158, -0.00028912, -0.00027864, -0.00019437, -0.00002386, 0.00022149, 0.00050207, 0.00075561, 0.00091277, 0.00091876, 0.00075378, 0.00044436, 0.00006008, -0.00030487, -0.00055986, -0.00064646, -0.00055723, -0.00033754, -0.00006952, 0.00015647, 0.00027236, 0.00025391, 0.00012588, -0.00005078, -0.00020379, -0.00027690, -0.00024919, -0.00013964, 0.00000364, 0.00012440, 0.00018036, 0.00015806, 0.00007504, -0.00003009, -0.00011479, -0.00014940, -0.00012706, -0.00006364, 0.00001139, 0.00006810, 0.00008790, 0.00006912, 0.00002515, -0.00002306, -0.00005650, -0.00006517, -0.00005045, -0.00002245, 0.00000569, 0.00002364, 0.00002737, 0.00001937, 0.00000600, -0.00000610, -0.00001289, -0.00001393, -0.00001144, -0.00000839, -0.00000666, -0.00000636, -0.00000627, -0.00000497, -0.00000196, 0.00000197, 0.00000519, 0.00000621, 0.00000453, 0.00000093, -0.00000291, -0.00000530, -0.00000533, -0.00000329, -0.00000042, 0.00000179, 0.00000229, 0.00000095, -0.00000149, -0.00000385, -0.00000507, -0.00000468, -0.00000290, -0.00000043, 0.00000185, 0.00000325, 0.00000347, 0.00000255, 0.00000079, -0.00000137, -0.00000351, -0.00000526, -0.00000640, -0.00000688, },
		{ -0.00000201, -0.00000357, -0.00000479, -0.00000560, -0.00000617, -0.00000672, -0.00000742, -0.00000812, -0.00000843, -0.00000780, -0.00000598, -0.00000327, -0.00000063, 0.00000059, -0.00000065, -0.00000447, -0.00000966, -0.00001400, -0.00001517, -0.00001188, -0.00000477, 0.00000355, 0.00000947, 0.00000984, 0.00000354, -0.00000798, -0.00002121, -0.00003215, -0.00003807, -0.00003848, -0.00003486, -0.00002926, -0.00002276, -0.00001487, -0.00000468, 0.00000661, 0.00001326, 0.00000456, -0.00003201, -0.00010439, -0.00020820, -0.00032167, -0.00040651, -0.00041693, -0.00031505, -0.00008812, 0.00023934, 0.00060460, 0.00091971, 0.00109602, 0.00107279, 0.00083991, 0.00044586, -0.00001348, -0.00042247, -0.00068108, -0.00073585, -0.00059554, -0.00032592, -0.00002569, 0.00020768, 0.00030972, 0.00026743, 0.00011840, -0.00006760, -0.00021684, -0.00027842, -0.00023995, -0.00012749, 0.00000887, 0.00011662, 0.00016083, 0.00013496, 0.00005949, -0.00002958, -0.00009670, -0.00012024, -0.00009863, -0.00004774, 0.00000816, 0.00004728, 0.00005834, 0.00004306, 0.00001297, -0.00001716, -0.00003584, -0.00003879, -0.00002897, -0.00001364, -0.00000023, 0.00000685, 0.00000725, 0.00000365, -0.00000046, -0.00000280, -0.00000314, -0.00000279, -0.00000334, -0.00000545, -0.00000837, -0.00001047, -0.00001020, -0.00000708, -0.00000203, 0.00000307, 0.00000631, 0.00000662, 0.00000428, 0.00000062, -0.00000267, -0.00000432, -0.00000402, -0.00000239, -0.00000051, 0.00000059, 0.00000048, -0.00000065, -0.00000211, -0.00000317, -0.00000334, -0.00000256, -0.00000113, 0.00000046, 0.00000171, 0.00000230, 0.00000209, 0.00000114, -0.00000033, },
		{ 0.00000121, 0.00000115, 0.00000054, -0.00000048, -0.00000160, -0.00000251, -0.00000304, -0.00000329, -0.00000368, -0.00000468, -0.00000656, -0.00000904, -0.00001123, -0.00001195, -0.00001030, -0.00000624, -0.00000094, 0.00000352, 0.00000497, 0.00000223, -0.00000415, -0.00001188, -0.00001775, -0.00001908, -0.00001498, -0.00000690, 0.00000195, 0.00000803, 0.00000908, 0.00000501, -0.00000240, -0.00001079, -0.00001881, -0.00002676, -0.00003590, -0.00004656, -0.00005613, -0.00005870, -0.00004727, -0.00001822, 0.00002376, 0.00006309, 0.00007576, 0.00003745, -0.00006469, -0.00022101, -0.00039464, -0.00052691, -0.00055357, -0.00042745, -0.00014004, 0.00026702, 0.00070657, 0.00106746, 0.00124746, 0.00118660, 0.00088930, 0.00042653, -0.00008386, -0.00051271, -0.00075907, -0.00078011, -0.00060135, -0.00030412, 0.00000462, 0.00022875, 0.00031286, 0.00025545, 0.00010249, -0.00007477, -0.00020808, -0.00025578, -0.00021364, -0.00011069, 0.00000663, 0.00009429, 0.00012664, 0.00010283, 0.00004284, -0.00002378, -0.00007089, -0.00008507, -0.00006821, -0.00003364, 0.00000160, 0.00002427, 0.00002933, 0.00001998, 0.00000426, -0.00000960, -0.00001675, -0.00001683, -0.00001279, -0.00000844, -0.00000621, -0.00000623, -0.00000693, -0.00000648, -0.00000407, -0.00000051, 0.00000238, 0.00000287, 0.00000039, -0.00000410, -0.00000858, -0.00001096, -0.00001015, -0.00000647, -0.00000144, 0.00000297, 0.00000530, 0.00000510, 0.00000301, 0.00000028, -0.00000188, -0.00000282, -0.00000259, -0.00000173, -0.00000093, -0.00000062, -0.00000090, -0.00000148, -0.00000201, -0.00000215, -0.00000180, -0.00000105, -0.00000011, 0.00000073, },
		{ -0.00000116, -0.00000060, 0.00000008, 0.00000066, 0.00000091, 0.00000059, -0.00000033, -0.00000161, -0.00000278, -0.00000332, -0.00000293, -0.00000183, -0.00000077, -0.00000079, -0.00000267, -0.00000641, -0.00001099, -0.00001457, -0.00001532, -0.00001228, -0.00000608, 0.00000112, 0.00000641, 0.00000744, 0.00000351, -0.00000393, -0.00001194, -0.00001740, -0.00001846, -0.00001536, -0.00001007, -0.00000505, -0.00000175, 0.00000012, 0.00000204, 0.00000501, 0.00000782, 0.00000648, -0.00000421, -0.00002726, -0.00005985, -0.00009161, -0.00010676, -0.00009047, -0.00003731, 0.00004182, 0.00011841, 0.00015319, 0.00011034, -0.00002541, -0.00023570, -0.00046610, -0.00063769, -0.00067093, -0.00051457, -0.00016917, 0.00030446, 0.00079699, 0.00118050, 0.00134747, 0.00124574, 0.00089693, 0.00039144, -0.00013896, -0.00056173, -0.00078340, -0.00077535, -0.00057733, -0.00027865, 0.00001493, 0.00021645, 0.00028279, 0.00022199, 0.00008236, -0.00007060, -0.00017959, -0.00021401, -0.00017580, -0.00009247, -0.00000222, 0.00006225, 0.00008462, 0.00006756, 0.00002751, -0.00001466, -0.00004276, -0.00005031, -0.00004060, -0.00002267, -0.00000600, 0.00000366, 0.00000557, 0.00000273, -0.00000086, -0.00000263, -0.00000243, -0.00000191, -0.00000295, -0.00000621, -0.00001064, -0.00001406, -0.00001450, -0.00001134, -0.00000569, 0.00000014, 0.00000378, 0.00000394, 0.00000093, -0.00000357, -0.00000740, -0.00000894, -0.00000772, -0.00000450, -0.00000072, 0.00000220, 0.00000345, 0.00000304, 0.00000161, -0.00000002, -0.00000122, -0.00000175, -0.00000177, -0.00000155, -0.00000139, -0.00000138, -0.00000148, -0.00000157, -0.00000148, },
		{ -0.00000158, -0.00000159, -0.00000157, -0.00000144, -0.00000109, -0.00000045, 0.00000044, 0.00000132, 0.00000177, 0.00000138, -0.00000004, -0.00000225, -0.00000450, -0.00000581, -0.00000542, -0.00000327, -0.00000024, 0.00000211, 0.00000219, -0.00000078, -0.00000623, -0.00001227, -0.00001640, -0.00001666, -0.00001265, -0.00000581, 0.00000113, 0.00000541, 0.00000555, 0.00000201, -0.00000319, -0.00000768, -0.00001015, -0.00001108, -0.00001221, -0.00001503, -0.00001909, -0.00002143, -0.00001774, -0.00000512, 0.00001483, 0.00003458, 0.00004237, 0.00002717, -0.00001481, -0.00007517, -0.00013305, -0.00016127, -0.00013752, -0.00005666, 0.00006148, 0.00017375, 0.00022611, 0.00017404, 0.00000371, -0.00025488, -0.00052961, -0.00072482, -0.00075192, -0.00056214, -0.00016908, 0.00034809, 0.00086393, 0.00124316, 0.00138244, 0.00124316, 0.00086393, 0.00034809, -0.00016908, -0.00056214, -0.00075192, -0.00072482, -0.00052961, -0.00025488, 0.00000371, 0.00017404, 0.00022611, 0.00017375, 0.00006148, -0.00005666, -0.00013752, -0.00016127, -0.00013305, -0.00007517, -0.00001481, 0.00002717, 0.00004237, 0.00003458, 0.00001483, -0.00000512, -0.00001774, -0.00002143, -0.00001909, -0.00001503, -0.00001221, -0.00001108, -0.00001015, -0.00000768, -0.00000319, 0.00000201, 0.00000555, 0.00000541, 0.00000113, -0.00000581, -0.00001265, -0.00001666, -0.00001640, -0.00001227, -0.00000623, -0.00000078, 0.00000219, 0.00000211, -0.00000024, -0.00000327, -0.00000542, -0.00000581, -0.00000450, -0.00000225, -0.00000004, 0.00000138, 0.00000177, 0.00000132, 0.00000044, -0.00000045, -0.00000109, -0.00000144, -0.00000157, -0.00000159, },
		{ -0.00000116, -0.00000148, -0.00000157, -0.00000148, -0.00000138, -0.00000139, -0.00000155, -0.00000177, -0.00000175, -0.00000122, -0.00000002, 0.00000161, 0.00000304, 0.00000345, 0.00000220, -0.00000072, -0.00000450, -0.00000772, -0.00000894, -0.00000740, -0.00000357, 0.00000093, 0.00000394, 0.00000378, 0.00000014, -0.00000569, -0.00001134, -0.00001450, -0.00001406, -0.00001064, -0.00000621, -0.00000295, -0.00000191, -0.00000243, -0.00000263, -0.00000086, 0.00000273, 0.00000557, 0.00000366, -0.00000600, -0.00002267, -0.00004060, -0.00005031, -0.00004276, -0.00001466, 0.00002751, 0.00006756, 0.00008462, 0.00006225, -0.00000222, -0.00009247, -0.00017580, -0.00021401, -0.00017959, -0.00007060, 0.00008236, 0.00022199, 0.00028279, 0.00021645, 0.00001493, -0.00027865, -0.00057733, -0.00077535, -0.00078340, -0.00056173, -0.00013896, 0.00039144, 0.00089693, 0.00124574, 0.00134747, 0.00118050, 0.00079699, 0.00030446, -0.00016917, -0.00051457, -0.00067093, -0.00063769, -0.00046610, -0.00023570, -0.00002541, 0.00011034, 0.00015319, 0.00011841, 0.00004182, -0.00003731, -0.00009047, -0.00010676, -0.00009161, -0.00005985, -0.00002726, -0.00000421, 0.00000648, 0.00000782, 0.00000501, 0.00000204, 0.00000012, -0.00000175, -0.00000505, -0.00001007, -0.00001536, -0.00001846, -0.00001740, -0.00001194, -0.00000393, 0.00000351, 0.00000744, 0.00000641, 0.00000112, -0.00000608, -0.00001228, -0.00001532, -0.00001457, -0.00001099, -0.00000641, -0.00000267, -0.00000079, -0.00000077, -0.00000183, -0.00000293, -0.00000332, -0.00000278, -0.00000161, -0.00000033, 0.00000059, 0.00000091, 0.00000066, 0.00000008, -0.00000060, },
		{ 0.00000121, 0.00000073, -0.00000011, -0.00000105, -0.00000180, -0.00000215, -0.00000201, -0.00000148, -0.00000090, -0.00000062, -0.00000093, -0.00000173, -0.00000259, -0.00000282, -0.00000188, 0.00000028, 0.00000301, 0.00000510, 0.00000530, 0.00000297, -0.00000144, -0.00000647, -0.00001015, -0.00001096, -0.00000858, -0.00000410, 0.00000039, 0.00000287, 0.00000238, -0.00000051, -0.00000407, -0.00000648, -0.00000693, -0.00000623, -0.00000621, -0.00000844, -0.00001279, -0.00001683, -0.00001675, -0.00000960, 0.00000426, 0.00001998, 0.00002933, 0.00002427, 0.00000160, -0.00003364, -0.00006821, -0.00008507, -0.00007089, -0.00002378, 0.00004284, 0.00010283, 0.00012664, 0.00009429, 0.00000663, -0.00011069, -0.00021364, -0.00025578, -0.00020808, -0.00007477, 0.00010249, 0.00025545, 0.00031286, 0.00022875, 0.00000462, -0.00030412, -0.00060135, -0.00078011, -0.00075907, -0.00051271, -0.00008386, 0.00042653, 0.00088930, 0.00118660, 0.00124746, 0.00106746, 0.00070657, 0.00026702, -0.00014004, -0.00042745, -0.00055357, -0.00052691, -0.00039464, -0.00022101, -0.00006469, 0.00003745, 0.00007576, 0.00006309, 0.00002376, -0.00001822, -0.00004727, -0.00005870, -0.00005613, -0.00004656, -0.00003590, -0.00002676, -0.00001881, -0.00001079, -0.00000240, 0.00000501, 0.00000908, 0.00000803, 0.00000195, -0.00000690, -0.00001498, -0.00001908, -0.00001775, -0.00001188, -0.00000415, 0.00000223, 0.00000497, 0.00000352, -0.00000094, -0.00000624, -0.00001030, -0.00001195, -0.00001123, -0.00000904, -0.00000656, -0.00000468, -0.00000368, -0.00000329, -0.00000304, -0.00000251, -0.00000160, -0.00000048, 0.00000054, 0.00000115, },
		{ -0.00000201, -0.00000033, 0.00000114, 0.00000209, 0.00000230, 0.00000171, 0.00000046, -0.00000113, -0.00000256, -0.00000334, -0.00000317, -0.00000211, -0.00000065, 0.00000048, 0.00000059, -0.00000051, -0.00000239, -0.00000402, -0.00000432, -0.00000267, 0.00000062, 0.00000428, 0.00000662, 0.00000631, 0.00000307, -0.00000203, -0.00000708, -0.00001020, -0.00001047, -0.00000837, -0.00000545, -0.00000334, -0.00000279, -0.00000314, -0.00000280, -0.00000046, 0.00000365, 0.00000725, 0.00000685, -0.00000023, -0.00001364, -0.00002897, -0.00003879, -0.00003584, -0.00001716, 0.00001297, 0.00004306, 0.00005834, 0.00004728, 0.00000816, -0.00004774, -0.00009863, -0.00012024, -0.00009670, -0.00002958, 0.00005949, 0.00013496, 0.00016083, 0.00011662, 0.00000887, -0.00012749, -0.00023995, -0.00027842, -0.00021684, -0.00006760, 0.00011840, 0.00026743, 0.00030972, 0.00020768, -0.00002569, -0.00032592, -0.00059554, -0.00073585, -0.00068108, -0.00042247, -0.00001348, 0.00044586, 0.00083991, 0.00107279, 0.00109602, 0.00091971, 0.00060460, 0.00023934, -0.00008812, -0.00031505, -0.00041693, -0.00040651, -0.00032167, -0.00020820, -0.00010439, -0.00003201, 0.00000456, 0.00001326, 0.00000661, -0.00000468, -0.00001487, -0.00002276, -0.00002926, -0.00003486, -0.00003848, -0.00003807, -0.00003215, -0.00002121, -0.00000798, 0.00000354, 0.00000984, 0.00000947, 0.00000355, -0.00000477, -0.00001188, -0.00001517, -0.00001400, -0.00000966, -0.00000447, -0.00000065, 0.00000059, -0.00000063, -0.00000327, -0.00000598, -0.00000780, -0.00000843, -0.00000812, -0.00000742, -0.00000672, -0.00000617, -0.00000560, -0.00000479, -0.00000357, },
		{ -0.00000683, -0.00000688, -0.00000640, -0.00000526, -0.00000351, -0.00000137, 0.00000079, 0.00000255, 0.00000347, 0.00000325, 0.00000185, -0.00000043, -0.00000290, -0.00000468, -0.00000507, -0.00000385, -0.00000149, 0.00000095, 0.00000229, 0.00000179, -0.00000042, -0.00000329, -0.00000533, -0.00000530, -0.00000291, 0.00000093, 0.00000453, 0.00000621, 0.00000519, 0.00000197, -0.00000196, -0.00000497, -0.00000627, -0.00000636, -0.00000666, -0.00000839, -0.00001144, -0.00001393, -0.00001289, -0.00000610, 0.00000600, 0.00001937, 0.00002737, 0.00002364, 0.00000569, -0.00002245, -0.00005045, -0.00006517, -0.00005650, -0.00002306, 0.00002515, 0.00006912, 0.00008790, 0.00006810, 0.00001139, -0.00006364, -0.00012706, -0.00014940, -0.00011479, -0.00003009, 0.00007504, 0.00015806, 0.00018036, 0.00012440, 0.00000364, -0.00013964, -0.00024919, -0.00027690, -0.00020379, -0.00005078, 0.00012588, 0.00025391, 0.00027236, 0.00015647, -0.00006952, -0.00033754, -0.00055723, -0.00064646, -0.00055986, -0.00030487, 0.00006008, 0.00044436, 0.00075378, 0.00091876, 0.00091277, 0.00075561, 0.00050207, 0.00022149, -0.00002386, -0.00019437, -0.00027864, -0.00028912, -0.00025158, -0.00019344, -0.00013527, -0.00008759, -0.00005247, -0.00002763, -0.00001048, -0.00000009, 0.00000306, -0.00000133, -0.00001172, -0.00002445, -0.00003478, -0.00003875, -0.00003489, -0.00002473, -0.00001202, -0.00000106, 0.00000508, 0.00000560, 0.00000190, -0.00000341, -0.00000777, -0.00000970, -0.00000917, -0.00000716, -0.00000500, -0.00000365, -0.00000341, -0.00000395, -0.00000475, -0.00000539, -0.00000574, -0.00000593, -0.00000617, -0.00000651, },
		{ -0.00000467, -0.00000453, -0.00000478, -0.00000542, -0.00000622, -0.00000686, -0.00000703, -0.00000649, -0.00000513, -0.00000302, -0.00000041, 0.00000221, 0.00000420, 0.00000491, 0.00000391, 0.00000130, -0.00000219, -0.00000539, -0.00000707, -0.00000652, -0.00000392, -0.00000036, 0.00000258, 0.00000360, 0.00000229, -0.00000062, -0.00000363, -0.00000517, -0.00000450, -0.00000206, 0.00000077, 0.00000248, 0.00000236, 0.00000090, -0.00000049, -0.00000049, 0.00000105, 0.00000260, 0.00000163, -0.00000387, -0.00001363, -0.00002425, -0.00003006, -0.00002569, -0.00000921, 0.00001562, 0.00003953, 0.00005097, 0.00004130, 0.00000970, -0.00003457, -0.00007448, -0.00009166, -0.00007464, -0.00002538, 0.00003949, 0.00009384, 0.00011242, 0.00008225, 0.00000991, -0.00007885, -0.00014819, -0.00016664, -0.00012126, -0.00002499, 0.00008631, 0.00016717, 0.00018093, 0.00011596, -0.00000756, -0.00014374, -0.00023834, -0.00025073, -0.00017161, -0.00002890, 0.00012117, 0.00021489, 0.00020590, 0.00008436, -0.00011707, -0.00033320, -0.00048835, -0.00052257, -0.00041215, -0.00017733, 0.00012507, 0.00042071, 0.00064121, 0.00074357, 0.00071952, 0.00059284, 0.00040728, 0.00021061, 0.00004076, -0.00008171, -0.00015374, -0.00018429, -0.00018686, -0.00017327, -0.00015083, -0.00012286, -0.00009109, -0.00005801, -0.00002771, -0.00000500, 0.00000657, 0.00000639, -0.00000273, -0.00001557, -0.00002657, -0.00003181, -0.00003019, -0.00002333, -0.00001438, -0.00000651, -0.00000167, -0.00000018, -0.00000107, -0.00000290, -0.00000450, -0.00000536, -0.00000557, -0.00000552, -0.00000550, -0.00000557, -0.00000559, -0.00000541, -0.00000505, },
		{ -0.00000574, -0.00000596, -0.00000568, -0.00000504, -0.00000437, -0.00000396, -0.00000402, -0.00000455, -0.00000538, -0.00000627, -0.00000692, -0.00000706, -0.00000642, -0.00000479, -0.00000218, 0.00000104, 0.00000415, 0.00000616, 0.00000620, 0.00000396, -0.00000004, -0.00000452, -0.00000787, -0.00000882, -0.00000707, -0.00000346, 0.00000040, 0.00000289, 0.00000316, 0.00000153, -0.00000071, -0.00000214, -0.00000204, -0.00000086, 0.00000006, -0.00000059, -0.00000304, -0.00000595, -0.00000691, -0.00000390, 0.00000313, 0.00001145, 0.00001620, 0.00001266, -0.00000095, -0.00002143, -0.00004082, -0.00004921, -0.00003931, -0.00001080, 0.00002784, 0.00006150, 0.00007419, 0.00005634, 0.00001035, -0.00004850, -0.00009700, -0.00011321, -0.00008640, -0.00002313, 0.00005354, 0.00011241, 0.00012680, 0.00008654, 0.00000382, -0.00009025, -0.00015746, -0.00016823, -0.00011478, -0.00001574, 0.00009023, 0.00015961, 0.00016198, 0.00009331, -0.00002138, -0.00013734, -0.00020801, -0.00020430, -0.00012714, -0.00000813, 0.00010234, 0.00015487, 0.00012087, 0.00000454, -0.00015764, -0.00030959, -0.00039562, -0.00037977, -0.00025776, -0.00005726, 0.00017253, 0.00037764, 0.00051585, 0.00056728, 0.00053632, 0.00044550, 0.00032500, 0.00020218, 0.00009496, 0.00001039, -0.00005247, -0.00009774, -0.00012867, -0.00014595, -0.00014859, -0.00013613, -0.00011072, -0.00007764, -0.00004416, -0.00001719, -0.00000111, 0.00000346, -0.00000072, -0.00000904, -0.00001695, -0.00002150, -0.00002189, -0.00001904, -0.00001470, -0.00001044, -0.00000719, -0.00000511, -0.00000396, -0.00000344, -0.00000337, -0.00000369, -0.00000434, -0.00000512, },
		{ -0.00000499, -0.00000462, -0.00000485, -0.00000535, -0.00000572, -0.00000568, -0.00000516, -0.00000437, -0.00000362, -0.00000320, -0.00000324, -0.00000371, -0.00000449, -0.00000544, -0.00000640, -0.00000714, -0.00000731, -0.00000647, -0.00000431, -0.00000094, 0.00000299, 0.00000632, 0.00000781, 0.00000668, 0.00000306, -0.00000190, -0.00000645, -0.00000898, -0.00000876, -0.00000629, -0.00000301, -0.00000050, 0.00000035, -0.00000011, -0.00000063, 0.00000010, 0.00000241, 0.00000512, 0.00000600, 0.00000306, -0.00000395, -0.00001266, -0.00001867, -0.00001755, -0.00000739, 0.00000928, 0.00002563, 0.00003299, 0.00002491, 0.00000103, -0.00003129, -0.00005894, -0.00006814, -0.00005092, -0.00000990, 0.00004109, 0.00008157, 0.00009262, 0.00006556, 0.00000711, -0.00006179, -0.00011358, -0.00012536, -0.00008918, -0.00001672, 0.00006428, 0.00012070, 0.00012779, 0.00008009, -0.00000529, -0.00009489, -0.00015236, -0.00015356, -0.00009703, -0.00000515, 0.00008480, 0.00013590, 0.00012700, 0.00006158, -0.00003364, -0.00011990, -0.00016267, -0.00014609, -0.00007967, 0.00000563, 0.00007024, 0.00008236, 0.00003109, -0.00006879, -0.00018224, -0.00026710, -0.00028952, -0.00023567, -0.00011568, 0.00004128, 0.00019790, 0.00032125, 0.00039193, 0.00040728, 0.00037826, 0.00032247, 0.00025667, 0.00019170, 0.00013143, 0.00007498, 0.00002054, -0.00003154, -0.00007765, -0.00011222, -0.00013029, -0.00012998, -0.00011362, -0.00008711, -0.00005786, -0.00003245, -0.00001484, -0.00000586, -0.00000383, -0.00000586, -0.00000920, -0.00001199, -0.00001345, -0.00001358, -0.00001273, -0.00001125, -0.00000945, -0.00000761, -0.00000603, },
		{ -0.00000877, -0.00000867, -0.00000811, -0.00000744, -0.00000703, -0.00000707, -0.00000742, -0.00000773, -0.00000760, -0.00000684, -0.00000554, -0.00000407, -0.00000284, -0.00000214, -0.00000203, -0.00000239, -0.00000309, -0.00000405, -0.00000524, -0.00000652, -0.00000753, -0.00000766, -0.00000634, -0.00000335, 0.00000083, 0.00000504, 0.00000782, 0.00000802, 0.00000544, 0.00000101, -0.00000356, -0.00000657, -0.00000720, -0.00000590, -0.00000414, -0.00000345, -0.00000443, -0.00000616, -0.00000659, -0.00000377, 0.00000269, 0.00001075, 0.00001640, 0.00001547, 0.00000604, -0.00000981, -0.00002597, -0.00003451, -0.00002936, -0.00000988, 0.00001767, 0.00004177, 0.00005024, 0.00003597, 0.00000124, -0.00004185, -0.00007542, -0.00008316, -0.00005796, -0.00000631, 0.00005300, 0.00009576, 0.00010242, 0.00006697, 0.00000066, -0.00007150, -0.00012049, -0.00012535, -0.00008247, -0.00000784, 0.00006888, 0.00011644, 0.00011496, 0.00006448, -0.00001482, -0.00009090, -0.00013324, -0.00012548, -0.00007214, 0.00000358, 0.00006988, 0.00009987, 0.00008271, 0.00002763, -0.00004069, -0.00009336, -0.00010983, -0.00008667, -0.00003857, 0.00000847, 0.00002878, 0.00000827, -0.00004927, -0.00012354, -0.00018576, -0.00020989, -0.00018237, -0.00010658, -0.00000065, 0.00010988, 0.00020137, 0.00025938, 0.00028154, 0.00027554, 0.00025370, 0.00022705, 0.00020131, 0.00017616, 0.00014749, 0.00011108, 0.00006582, 0.00001502, -0.00003454, -0.00007516, -0.00010100, -0.00010992, -0.00010379, -0.00008739, -0.00006645, -0.00004599, -0.00002920, -0.00001734, -0.00001016, -0.00000668, -0.00000570, -0.00000618, -0.00000724, -0.00000825, },
		{ -0.00001126, -0.00000721, -0.00000517, -0.00000433, -0.00000401, -0.00000385, -0.00000384, -0.00000426, -0.00000531, -0.00000696, -0.00000880, -0.00001021, -0.00001061, -0.00000979, -0.00000797, -0.00000571, -0.00000362, -0.00000212, -0.00000132, -0.00000111, -0.00000136, -0.00000205, -0.00000324, -0.00000489, -0.00000664, -0.00000776, -0.00000744, -0.00000519, -0.00000130, 0.00000310, 0.00000641, 0.00000731, 0.00000544, 0.00000172, -0.00000201, -0.00000397, -0.00000346, -0.00000138, 0.00000012, -0.00000110, -0.00000574, -0.00001217, -0.00001680, -0.00001572, -0.00000699, 0.00000762, 0.00002259, 0.00003062, 0.00002600, 0.00000793, -0.00001803, -0.00004141, -0.00005101, -0.00004014, -0.00001052, 0.00002720, 0.00005712, 0.00006470, 0.00004358, -0.00000054, -0.00005109, -0.00008688, -0.00009105, -0.00005885, -0.00000076, 0.00006093, 0.00010089, 0.00010141, 0.00006036, -0.00000729, -0.00007505, -0.00011579, -0.00011301, -0.00006790, 0.00000103, 0.00006564, 0.00009994, 0.00009083, 0.00004334, -0.00002201, -0.00007822, -0.00010336, -0.00008955, -0.00004554, 0.00000789, 0.00004753, 0.00005792, 0.00003747, -0.00000180, -0.00004064, -0.00006191, -0.00005851, -0.00003630, -0.00001099, -0.00000039, -0.00001579, -0.00005631, -0.00010865, -0.00015232, -0.00016796, -0.00014505, -0.00008592, -0.00000454, 0.00007914, 0.00014662, 0.00018705, 0.00019974, 0.00019257, 0.00017730, 0.00016412, 0.00015783, 0.00015680, 0.00015491, 0.00014507, 0.00012268, 0.00008767, 0.00004435, -0.00000033, -0.00003928, -0.00006740, -0.00008263, -0.00008582, -0.00007978, -0.00006802, -0.00005385, -0.00003980, -0.00002754, -0.00001794, },
		{ -0.00006194, -0.00005398, -0.00004415, -0.00003423, -0.00002541, -0.00001826, -0.00001269, -0.00000829, -0.00000463, -0.00000163, 0.00000046, 0.00000116, 0.00000015, -0.00000242, -0.00000587, -0.00000917, -0.00001136, -0.00001191, -0.00001088, -0.00000882, -0.00000644, -0.00000428, -0.00000258, -0.00000134, -0.00000057, -0.00000042, -0.00000111, -0.00000273, -0.00000489, -0.00000674, -0.00000720, -0.00000554, -0.00000193, 0.00000239, 0.00000559, 0.00000607, 0.00000348, -0.00000092, -0.00000467, -0.00000536, -0.00000207, 0.00000375, 0.00000866, 0.00000876, 0.00000193, -0.00001048, -0.00002353, -0.00003049, -0.00002602, -0.00000924, 0.00001481, 0.00003652, 0.00004550, 0.00003546, 0.00000785, -0.00002765, -0.00005646, -0.00006515, -0.00004769, -0.00000900, 0.00003612, 0.00006865, 0.00007345, 0.00004622, -0.00000376, -0.00005677, -0.00009060, -0.00008996, -0.00005346, 0.00000524, 0.00006272, 0.00009544, 0.00008966, 0.00004736, -0.00001448, -0.00007103, -0.00009997, -0.00009078, -0.00004880, 0.00000745, 0.00005456, 0.00007414, 0.00006029, 0.00002128, -0.00002488, -0.00005882, -0.00006825, -0.00005266, -0.00002245, 0.00000675, 0.00002172, 0.00001754, -0.00000082, -0.00002185, -0.00003386, -0.00003115, -0.00001708, -0.00000246, -0.00000014, -0.00001841, -0.00005620, -0.00010255, -0.00014048, -0.00015370, -0.00013319, -0.00008091, -0.00000906, 0.00006448, 0.00012273, 0.00015534, 0.00016134, 0.00014819, 0.00012776, 0.00011133, 0.00010547, 0.00011040, 0.00012098, 0.00012961, 0.00012944, 0.00011687, 0.00009234, 0.00005957, 0.00002398, -0.00000918, -0.00003597, -0.00005435, -0.00006403, -0.00006598, },
		{ 0.00001182, -0.00001144, -0.00002974, -0.00004202, -0.00004824, -0.00004928, -0.00004661, -0.00004185, -0.00003636, -0.00003085, -0.00002541, -0.00001976, -0.00001368, -0.00000740, -0.00000168, 0.00000244, 0.00000414, 0.00000319, 0.00000015, -0.00000383, -0.00000749, -0.00000992, -0.00001081, -0.00001041, -0.00000919, -0.00000756, -0.00000571, -0.00000371, -0.00000180, -0.00000042, -0.00000015, -0.00000131, -0.00000358, -0.00000586, -0.00000669, -0.00000499, -0.00000083, 0.00000430, 0.00000789, 0.00000771, 0.00000306, -0.00000443, -0.00001116, -0.00001315, -0.00000814, 0.00000274, 0.00001489, 0.00002191, 0.00001861, 0.00000394, -0.00001757, -0.00003710, -0.00004501, -0.00003543, -0.00000966, 0.00002336, 0.00005011, 0.00005811, 0.00004174, 0.00000548, -0.00003703, -0.00006824, -0.00007410, -0.00005050, -0.00000573, 0.00004242, 0.00007382, 0.00007466, 0.00004359, -0.00000727, -0.00005715, -0.00008534, -0.00007998, -0.00004316, 0.00000992, 0.00005741, 0.00008014, 0.00006954, 0.00003099, -0.00001889, -0.00005968, -0.00007588, -0.00006312, -0.00002926, 0.00000980, 0.00003754, 0.00004396, 0.00002923, 0.00000257, -0.00002284, -0.00003630, -0.00003428, -0.00002122, -0.00000648, 0.00000099, -0.00000269, -0.00001455, -0.00002664, -0.00003068, -0.00002291, -0.00000665, 0.00000866, 0.00001175, -0.00000484, -0.00004056, -0.00008595, -0.00012566, -0.00014420, -0.00013195, -0.00008899, -0.00002512, 0.00004368, 0.00010098, 0.00013538, 0.00014371, 0.00013103, 0.00010772, 0.00008513, 0.00007163, 0.00007034, 0.00007911, 0.00009234, 0.00010349, 0.00010740, 0.00010157, 0.00008629, 0.00006394, 0.00003794, },
		{ 0.00008525, 0.00008362, 0.00007501, 0.00006052, 0.00004219, 0.00002251, 0.00000398, -0.00001139, -0.00002251, -0.00002933, -0.00003271, -0.00003395, -0.00003415, -0.00003382, -0.00003271, -0.00003015, -0.00002560, -0.00001913, -0.00001162, -0.00000447, 0.00000089, 0.00000360, 0.00000365, 0.00000176, -0.00000101, -0.00000376, -0.00000599, -0.00000760, -0.00000862, -0.00000896, -0.00000839, -0.00000677, -0.00000437, -0.00000198, -0.00000072, -0.00000139, -0.00000387, -0.00000682, -0.00000819, -0.00000617, -0.00000041, 0.00000734, 0.00001362, 0.00001473, 0.00000875, -0.00000308, -0.00001615, -0.00002427, -0.00002238, -0.00000945, 0.00001031, 0.00002858, 0.00003625, 0.00002774, 0.00000420, -0.00002607, -0.00005051, -0.00005753, -0.00004192, -0.00000795, 0.00003166, 0.00006057, 0.00006572, 0.00004335, 0.00000118, -0.00004424, -0.00007425, -0.00007601, -0.00004816, -0.00000170, 0.00004443, 0.00007134, 0.00006816, 0.00003664, -0.00000972, -0.00005152, -0.00007180, -0.00006312, -0.00003037, 0.00001179, 0.00004559, 0.00005781, 0.00004511, 0.00001474, -0.00001934, -0.00004303, -0.00004813, -0.00003533, -0.00001293, 0.00000777, 0.00001794, 0.00001501, 0.00000308, -0.00000996, -0.00001691, -0.00001502, -0.00000711, 0.00000038, 0.00000118, -0.00000706, -0.00002118, -0.00003401, -0.00003803, -0.00002963, -0.00001157, 0.00000755, 0.00001684, 0.00000807, -0.00002011, -0.00006082, -0.00010092, -0.00012581, -0.00012487, -0.00009547, -0.00004389, 0.00001705, 0.00007257, 0.00011082, 0.00012633, 0.00012084, 0.00010166, 0.00007835, 0.00005940, 0.00004981, 0.00005043, 0.00005866, 0.00007008, 0.00008011, },
		{ 0.00003760, 0.00003857, 0.00004494, 0.00005381, 0.00006214, 0.00006733, 0.00006763, 0.00006242, 0.00005222, 0.00003868, 0.00002405, 0.00001063, -0.00000002, -0.00000743, -0.00001226, -0.00001581, -0.00001930, -0.00002323, -0.00002715, -0.00002994, -0.00003040, -0.00002794, -0.00002289, -0.00001637, -0.00000981, -0.00000439, -0.00000066, 0.00000146, 0.00000230, 0.00000217, 0.00000113, -0.00000084, -0.00000350, -0.00000619, -0.00000791, -0.00000786, -0.00000595, -0.00000315, -0.00000118, -0.00000162, -0.00000489, -0.00000961, -0.00001294, -0.00001183, -0.00000480, 0.00000680, 0.00001868, 0.00002517, 0.00002185, 0.00000806, -0.00001197, -0.00003025, -0.00003819, -0.00003062, -0.00000872, 0.00001958, 0.00004242, 0.00004891, 0.00003427, 0.00000254, -0.00003429, -0.00006095, -0.00006533, -0.00004406, -0.00000451, 0.00003778, 0.00006541, 0.00006648, 0.00003987, -0.00000404, -0.00004757, -0.00007311, -0.00007053, -0.00004142, 0.00000183, 0.00004137, 0.00006157, 0.00005545, 0.00002724, -0.00001004, -0.00004054, -0.00005246, -0.00004280, -0.00001786, 0.00001026, 0.00002941, 0.00003274, 0.00002106, 0.00000163, -0.00001592, -0.00002434, -0.00002182, -0.00001215, -0.00000210, 0.00000249, -0.00000032, -0.00000783, -0.00001454, -0.00001552, -0.00000929, 0.00000113, 0.00000972, 0.00001070, 0.00000174, -0.00001439, -0.00003084, -0.00003989, -0.00003679, -0.00002246, -0.00000359, 0.00001005, 0.00000980, -0.00000777, -0.00003897, -0.00007398, -0.00010035, -0.00010768, -0.00009138, -0.00005428, -0.00000533, 0.00004362, 0.00008186, 0.00010295, 0.00010599, 0.00009495, 0.00007658, 0.00005786, 0.00004401, },
		{ 0.00008342, 0.00007181, 0.00005689, 0.00004299, 0.00003311, 0.00002871, 0.00002981, 0.00003524, 0.00004298, 0.00005050, 0.00005534, 0.00005567, 0.00005088, 0.00004181, 0.00003050, 0.00001942, 0.00001064, 0.00000501, 0.00000205, 0.00000024, -0.00000208, -0.00000595, -0.00001135, -0.00001728, -0.00002231, -0.00002530, -0.00002581, -0.00002412, -0.00002093, -0.00001688, -0.00001232, -0.00000739, -0.00000238, 0.00000208, 0.00000501, 0.00000554, 0.00000350, -0.00000024, -0.00000393, -0.00000576, -0.00000483, -0.00000185, 0.00000097, 0.00000099, -0.00000327, -0.00001092, -0.00001856, -0.00002152, -0.00001621, -0.00000233, 0.00001610, 0.00003181, 0.00003718, 0.00002785, 0.00000542, -0.00002237, -0.00004431, -0.00005032, -0.00003617, -0.00000599, 0.00002874, 0.00005358, 0.00005727, 0.00003685, -0.00000056, -0.00004025, -0.00006593, -0.00006660, -0.00004143, -0.00000034, 0.00004002, 0.00006324, 0.00006002, 0.00003205, -0.00000890, -0.00004610, -0.00006493, -0.00005891, -0.00003196, 0.00000380, 0.00003362, 0.00004647, 0.00003930, 0.00001759, -0.00000787, -0.00002617, -0.00003102, -0.00002290, -0.00000792, 0.00000577, 0.00001206, 0.00000953, 0.00000144, -0.00000659, -0.00000992, -0.00000728, -0.00000128, 0.00000337, 0.00000274, -0.00000398, -0.00001384, -0.00002162, -0.00002262, -0.00001535, -0.00000257, 0.00000982, 0.00001559, 0.00001131, -0.00000184, -0.00001841, -0.00003120, -0.00003462, -0.00002754, -0.00001396, -0.00000140, 0.00000245, -0.00000676, -0.00002789, -0.00005445, -0.00007705, -0.00008687, -0.00007898, -0.00005396, -0.00001745, 0.00002199, 0.00005586, 0.00007821, 0.00008688, },
		{ 0.00000888, 0.00003675, 0.00005741, 0.00006857, 0.00007034, 0.00006461, 0.00005416, 0.00004198, 0.00003083, 0.00002292, 0.00001975, 0.00002173, 0.00002800, 0.00003646, 0.00004424, 0.00004857, 0.00004770, 0.00004159, 0.00003188, 0.00002126, 0.00001239, 0.00000695, 0.00000506, 0.00000549, 0.00000636, 0.00000605, 0.00000376, -0.00000033, -0.00000544, -0.00001071, -0.00001559, -0.00001978, -0.00002302, -0.00002476, -0.00002420, -0.00002070, -0.00001432, -0.00000621, 0.00000151, 0.00000657, 0.00000760, 0.00000493, 0.00000061, -0.00000249, -0.00000226, 0.00000140, 0.00000614, 0.00000823, 0.00000447, -0.00000565, -0.00001902, -0.00002965, -0.00003125, -0.00002039, 0.00000113, 0.00002612, 0.00004446, 0.00004731, 0.00003139, 0.00000116, -0.00003232, -0.00005551, -0.00005815, -0.00003790, -0.00000192, 0.00003569, 0.00005955, 0.00005952, 0.00003512, -0.00000406, -0.00004223, -0.00006402, -0.00006091, -0.00003472, 0.00000332, 0.00003746, 0.00005415, 0.00004754, 0.00002156, -0.00001226, -0.00004002, -0.00005140, -0.00004362, -0.00002186, 0.00000369, 0.00002271, 0.00002908, 0.00002293, 0.00000956, -0.00000369, -0.00001120, -0.00001138, -0.00000672, -0.00000177, -0.00000026, -0.00000310, -0.00000804, -0.00001125, -0.00000975, -0.00000333, 0.00000511, 0.00001099, 0.00001050, 0.00000279, -0.00000937, -0.00002070, -0.00002599, -0.00002262, -0.00001196, 0.00000119, 0.00001085, 0.00001267, 0.00000594, -0.00000608, -0.00001787, -0.00002428, -0.00002310, -0.00001612, -0.00000827, -0.00000531, -0.00001108, -0.00002553, -0.00004457, -0.00006159, -0.00006997, -0.00006562, -0.00004830, -0.00002152, },
		{ -0.00005855, -0.00005540, -0.00004270, -0.00002285, 0.00000025, 0.00002250, 0.00004065, 0.00005285, 0.00005848, 0.00005785, 0.00005190, 0.00004211, 0.00003052, 0.00001967, 0.00001220, 0.00001008, 0.00001385, 0.00002227, 0.00003248, 0.00004096, 0.00004476, 0.00004255, 0.00003503, 0.00002458, 0.00001420, 0.00000637, 0.00000227, 0.00000165, 0.00000328, 0.00000571, 0.00000782, 0.00000895, 0.00000877, 0.00000690, 0.00000289, -0.00000350, -0.00001180, -0.00002053, -0.00002736, -0.00002990, -0.00002686, -0.00001882, -0.00000836, 0.00000091, 0.00000595, 0.00000575, 0.00000198, -0.00000183, -0.00000212, 0.00000263, 0.00001068, 0.00001751, 0.00001789, 0.00000879, -0.00000845, -0.00002785, -0.00004087, -0.00004019, -0.00002339, 0.00000506, 0.00003489, 0.00005395, 0.00005333, 0.00003163, -0.00000386, -0.00003963, -0.00006137, -0.00005997, -0.00003555, 0.00000237, 0.00003866, 0.00005886, 0.00005520, 0.00002971, -0.00000677, -0.00003933, -0.00005530, -0.00004940, -0.00002552, 0.00000538, 0.00003031, 0.00003979, 0.00003147, 0.00001048, -0.00001346, -0.00003060, -0.00003527, -0.00002768, -0.00001299, 0.00000173, 0.00001091, 0.00001273, 0.00000916, 0.00000419, 0.00000129, 0.00000153, 0.00000334, 0.00000377, 0.00000064, -0.00000597, -0.00001344, -0.00001795, -0.00001657, -0.00000904, 0.00000181, 0.00001125, 0.00001485, 0.00001067, 0.00000026, -0.00001200, -0.00002094, -0.00002298, -0.00001776, -0.00000816, 0.00000121, 0.00000618, 0.00000504, -0.00000093, -0.00000826, -0.00001327, -0.00001396, -0.00001109, -0.00000771, -0.00000765, -0.00001347, -0.00002511, -0.00003963, -0.00005233, },
		{ -0.00001191, -0.00002249, -0.00003454, -0.00004451, -0.00004916, -0.00004679, -0.00003762, -0.00002347, -0.00000689, 0.00000978, 0.00002486, 0.00003729, 0.00004625, 0.00005091, 0.00005044, 0.00004457, 0.00003411, 0.00002130, 0.00000942, 0.00000188, 0.00000091, 0.00000673, 0.00001725, 0.00002888, 0.00003781, 0.00004134, 0.00003875, 0.00003126, 0.00002130, 0.00001142, 0.00000349, -0.00000163, -0.00000383, -0.00000336, -0.00000050, 0.00000433, 0.00001022, 0.00001552, 0.00001803, 0.00001574, 0.00000785, -0.00000441, -0.00001784, -0.00002824, -0.00003213, -0.00002837, -0.00001882, -0.00000768, 0.00000040, 0.00000262, -0.00000057, -0.00000558, -0.00000748, -0.00000275, 0.00000848, 0.00002189, 0.00003059, 0.00002841, 0.00001328, -0.00001097, -0.00003547, -0.00004972, -0.00004629, -0.00002450, 0.00000853, 0.00004032, 0.00005799, 0.00005374, 0.00002843, -0.00000846, -0.00004255, -0.00006047, -0.00005538, -0.00002979, 0.00000567, 0.00003674, 0.00005154, 0.00004533, 0.00002207, -0.00000774, -0.00003184, -0.00004141, -0.00003441, -0.00001580, 0.00000530, 0.00001989, 0.00002291, 0.00001483, 0.00000076, -0.00001256, -0.00001989, -0.00001964, -0.00001384, -0.00000652, -0.00000125, 0.00000057, 0.00000007, -0.00000028, 0.00000152, 0.00000558, 0.00000992, 0.00001157, 0.00000836, 0.00000039, -0.00000968, -0.00001778, -0.00002031, -0.00001595, -0.00000644, 0.00000418, 0.00001140, 0.00001228, 0.00000675, -0.00000249, -0.00001139, -0.00001642, -0.00001618, -0.00001176, -0.00000600, -0.00000182, -0.00000086, -0.00000273, -0.00000556, -0.00000723, -0.00000669, -0.00000469, -0.00000342, -0.00000536, },
		{ -0.00000365, -0.00000100, -0.00000007, -0.00000264, -0.00000925, -0.00001876, -0.00002868, -0.00003613, -0.00003901, -0.00003674, -0.00003021, -0.00002116, -0.00001117, -0.00000104, 0.00000924, 0.00001993, 0.00003069, 0.00004016, 0.00004612, 0.00004641, 0.00004008, 0.00002821, 0.00001386, 0.00000120, -0.00000604, -0.00000591, 0.00000104, 0.00001216, 0.00002374, 0.00003246, 0.00003639, 0.00003527, 0.00003008, 0.00002232, 0.00001332, 0.00000417, -0.00000409, -0.00001014, -0.00001243, -0.00000972, -0.00000195, 0.00000911, 0.00001993, 0.00002629, 0.00002497, 0.00001535, 0.00000002, -0.00001594, -0.00002701, -0.00002964, -0.00002388, -0.00001344, -0.00000392, -0.00000006, -0.00000326, -0.00001071, -0.00001665, -0.00001528, -0.00000407, 0.00001443, 0.00003305, 0.00004297, 0.00003787, 0.00001731, -0.00001228, -0.00003978, -0.00005388, -0.00004801, -0.00002351, 0.00001054, 0.00004083, 0.00005530, 0.00004819, 0.00002247, -0.00001149, -0.00004023, -0.00005285, -0.00004537, -0.00002197, 0.00000711, 0.00003021, 0.00003912, 0.00003209, 0.00001396, -0.00000660, -0.00002116, -0.00002507, -0.00001889, -0.00000738, 0.00000322, 0.00000830, 0.00000664, 0.00000045, -0.00000631, -0.00001022, -0.00001007, -0.00000711, -0.00000394, -0.00000279, -0.00000416, -0.00000656, -0.00000751, -0.00000507, 0.00000079, 0.00000797, 0.00001318, 0.00001358, 0.00000831, -0.00000090, -0.00001049, -0.00001666, -0.00001708, -0.00001197, -0.00000388, 0.00000359, 0.00000744, 0.00000655, 0.00000197, -0.00000383, -0.00000833, -0.00001013, -0.00000942, -0.00000759, -0.00000619, -0.00000601, -0.00000667, -0.00000703, -0.00000605, },
		{ -0.00000802, -0.00000903, -0.00000862, -0.00000632, -0.00000274, 0.00000046, 0.00000126, -0.00000172, -0.00000838, -0.00001697, -0.00002479, -0.00002931, -0.00002933, -0.00002539, -0.00001943, -0.00001366, -0.00000937, -0.00000627, -0.00000273, 0.00000321, 0.00001245, 0.00002405, 0.00003521, 0.00004236, 0.00004272, 0.00003560, 0.00002292, 0.00000848, -0.00000342, -0.00000963, -0.00000909, -0.00000295, 0.00000629, 0.00001590, 0.00002390, 0.00002929, 0.00003172, 0.00003099, 0.00002673, 0.00001867, 0.00000733, -0.00000536, -0.00001618, -0.00002147, -0.00001871, -0.00000798, 0.00000747, 0.00002213, 0.00003025, 0.00002832, 0.00001681, 0.00000009, -0.00001527, -0.00002349, -0.00002220, -0.00001358, -0.00000330, 0.00000229, -0.00000054, -0.00001075, -0.00002278, -0.00002901, -0.00002361, -0.00000576, 0.00001924, 0.00004181, 0.00005210, 0.00004450, 0.00002051, -0.00001143, -0.00003919, -0.00005193, -0.00004473, -0.00002071, 0.00001029, 0.00003582, 0.00004604, 0.00003763, 0.00001494, -0.00001226, -0.00003308, -0.00004009, -0.00003201, -0.00001373, 0.00000644, 0.00002053, 0.00002424, 0.00001822, 0.00000698, -0.00000367, -0.00000948, -0.00000942, -0.00000556, -0.00000147, -0.00000008, -0.00000220, -0.00000632, -0.00000974, -0.00001025, -0.00000744, -0.00000287, 0.00000082, 0.00000148, -0.00000134, -0.00000604, -0.00000981, -0.00001013, -0.00000613, 0.00000086, 0.00000794, 0.00001198, 0.00001116, 0.00000580, -0.00000181, -0.00000857, -0.00001201, -0.00001128, -0.00000742, -0.00000260, 0.00000098, 0.00000215, 0.00000104, -0.00000120, -0.00000330, -0.00000455, -0.00000510, -0.00000560, -0.00000661, },
		{ -0.00000159, -0.00000198, -0.00000337, -0.00000576, -0.00000835, -0.00000985, -0.00000909, -0.00000590, -0.00000140, 0.00000222, 0.00000270, -0.00000119, -0.00000880, -0.00001770, -0.00002460, -0.00002683, -0.00002360, -0.00001641, -0.00000836, -0.00000272, -0.00000129, -0.00000352, -0.00000674, -0.00000747, -0.00000317, 0.00000640, 0.00001894, 0.00003047, 0.00003707, 0.00003653, 0.00002922, 0.00001777, 0.00000586, -0.00000331, -0.00000806, -0.00000845, -0.00000559, -0.00000073, 0.00000544, 0.00001270, 0.00002072, 0.00002828, 0.00003314, 0.00003263, 0.00002506, 0.00001101, -0.00000612, -0.00002093, -0.00002803, -0.00002443, -0.00001117, 0.00000676, 0.00002226, 0.00002913, 0.00002481, 0.00001161, -0.00000429, -0.00001573, -0.00001793, -0.00001084, 0.00000078, 0.00000966, 0.00000966, -0.00000097, -0.00001829, -0.00003419, -0.00004001, -0.00003071, -0.00000762, 0.00002164, 0.00004610, 0.00005603, 0.00004720, 0.00002279, -0.00000785, -0.00003307, -0.00004375, -0.00003694, -0.00001679, 0.00000755, 0.00002602, 0.00003184, 0.00002403, 0.00000725, -0.00001074, -0.00002269, -0.00002481, -0.00001795, -0.00000649, 0.00000409, 0.00000987, 0.00000994, 0.00000627, 0.00000213, 0.00000012, 0.00000085, 0.00000286, 0.00000370, 0.00000157, -0.00000355, -0.00000979, -0.00001437, -0.00001505, -0.00001142, -0.00000512, 0.00000095, 0.00000406, 0.00000301, -0.00000132, -0.00000648, -0.00000969, -0.00000920, -0.00000511, 0.00000076, 0.00000587, 0.00000813, 0.00000689, 0.00000305, -0.00000155, -0.00000506, -0.00000647, -0.00000592, -0.00000434, -0.00000283, -0.00000199, -0.00000179, -0.00000181, -0.00000171, },
		{ -0.00000384, -0.00000363, -0.00000268, -0.00000122, -0.00000001, -0.00000004, -0.00000192, -0.00000534, -0.00000898, -0.00001099, -0.00000992, -0.00000564, 0.00000028, 0.00000505, 0.00000587, 0.00000131, -0.00000778, -0.00001833, -0.00002626, -0.00002822, -0.00002318, -0.00001297, -0.00000159, 0.00000656, 0.00000863, 0.00000463, -0.00000268, -0.00000902, -0.00001061, -0.00000588, 0.00000390, 0.00001528, 0.00002431, 0.00002820, 0.00002642, 0.00002054, 0.00001316, 0.00000651, 0.00000155, -0.00000205, -0.00000505, -0.00000761, -0.00000867, -0.00000639, 0.00000068, 0.00001218, 0.00002529, 0.00003533, 0.00003763, 0.00002977, 0.00001316, -0.00000698, -0.00002349, -0.00003024, -0.00002488, -0.00001006, 0.00000763, 0.00002045, 0.00002309, 0.00001506, 0.00000090, -0.00001195, -0.00001682, -0.00001109, 0.00000234, 0.00001631, 0.00002276, 0.00001670, -0.00000110, -0.00002402, -0.00004227, -0.00004721, -0.00003523, -0.00000960, 0.00002071, 0.00004472, 0.00005393, 0.00004573, 0.00002427, -0.00000158, -0.00002208, -0.00003063, -0.00002607, -0.00001256, 0.00000279, 0.00001328, 0.00001541, 0.00000988, 0.00000060, -0.00000761, -0.00001135, -0.00001004, -0.00000563, -0.00000116, 0.00000105, 0.00000056, -0.00000120, -0.00000201, -0.00000035, 0.00000362, 0.00000800, 0.00001016, 0.00000822, 0.00000212, -0.00000619, -0.00001359, -0.00001727, -0.00001599, -0.00001065, -0.00000378, 0.00000166, 0.00000368, 0.00000203, -0.00000187, -0.00000574, -0.00000763, -0.00000675, -0.00000370, 0.00000003, 0.00000287, 0.00000393, 0.00000325, 0.00000156, -0.00000024, -0.00000160, -0.00000246, -0.00000304, -0.00000354, },
		{ 0.00000133, 0.00000027, -0.00000137, -0.00000320, -0.00000449, -0.00000449, -0.00000294, -0.00000038, 0.00000188, 0.00000232, 0.00000005, -0.00000454, -0.00000968, -0.00001291, -0.00001223, -0.00000725, 0.00000031, 0.00000711, 0.00000960, 0.00000570, -0.00000398, -0.00001620, -0.00002622, -0.00002985, -0.00002534, -0.00001419, -0.00000062, 0.00001030, 0.00001480, 0.00001201, 0.00000424, -0.00000428, -0.00000938, -0.00000881, -0.00000312, 0.00000492, 0.00001187, 0.00001537, 0.00001515, 0.00001281, 0.00001061, 0.00000991, 0.00001030, 0.00000983, 0.00000639, -0.00000070, -0.00000969, -0.00001671, -0.00001753, -0.00000978, 0.00000532, 0.00002293, 0.00003627, 0.00003954, 0.00003065, 0.00001246, -0.00000817, -0.00002329, -0.00002733, -0.00001963, -0.00000471, 0.00000984, 0.00001692, 0.00001331, 0.00000121, -0.00001277, -0.00002085, -0.00001792, -0.00000418, 0.00001464, 0.00002976, 0.00003332, 0.00002215, -0.00000056, -0.00002637, -0.00004513, -0.00004915, -0.00003651, -0.00001172, 0.00001626, 0.00003778, 0.00004631, 0.00004071, 0.00002509, 0.00000654, -0.00000811, -0.00001495, -0.00001407, -0.00000863, -0.00000286, 0.00000018, -0.00000019, -0.00000248, -0.00000430, -0.00000396, -0.00000145, 0.00000165, 0.00000325, 0.00000206, -0.00000155, -0.00000563, -0.00000769, -0.00000606, -0.00000087, 0.00000585, 0.00001110, 0.00001230, 0.00000855, 0.00000105, -0.00000744, -0.00001382, -0.00001606, -0.00001390, -0.00000887, -0.00000337, 0.00000042, 0.00000146, 0.00000008, -0.00000244, -0.00000459, -0.00000539, -0.00000470, -0.00000303, -0.00000115, 0.00000036, 0.00000131, 0.00000176, 0.00000178, },
		{ -0.00000215, -0.00000053, 0.00000123, 0.00000264, 0.00000311, 0.00000214, -0.00000019, -0.00000315, -0.00000544, -0.00000583, -0.00000380, -0.00000002, 0.00000375, 0.00000538, 0.00000344, -0.00000189, -0.00000866, -0.00001386, -0.00001475, -0.00001033, -0.00000207, 0.00000651, 0.00001127, 0.00000935, 0.00000068, -0.00001178, -0.00002312, -0.00002864, -0.00002587, -0.00001575, -0.00000223, 0.00000953, 0.00001540, 0.00001409, 0.00000753, -0.00000025, -0.00000521, -0.00000523, -0.00000104, 0.00000443, 0.00000781, 0.00000721, 0.00000333, -0.00000096, -0.00000225, 0.00000132, 0.00000871, 0.00001623, 0.00001922, 0.00001458, 0.00000273, -0.00001213, -0.00002354, -0.00002577, -0.00001661, 0.00000124, 0.00002104, 0.00003484, 0.00003699, 0.00002675, 0.00000865, -0.00000951, -0.00002025, -0.00001982, -0.00000987, 0.00000347, 0.00001259, 0.00001215, 0.00000179, -0.00001361, -0.00002618, -0.00002879, -0.00001858, 0.00000145, 0.00002352, 0.00003838, 0.00003936, 0.00002526, 0.00000098, -0.00002456, -0.00004207, -0.00004560, -0.00003464, -0.00001392, 0.00000894, 0.00002663, 0.00003485, 0.00003337, 0.00002529, 0.00001506, 0.00000627, 0.00000046, -0.00000290, -0.00000533, -0.00000796, -0.00001071, -0.00001233, -0.00001136, -0.00000725, -0.00000097, 0.00000521, 0.00000873, 0.00000806, 0.00000349, -0.00000283, -0.00000794, -0.00000937, -0.00000639, -0.00000025, 0.00000633, 0.00001046, 0.00001035, 0.00000605, -0.00000069, -0.00000732, -0.00001161, -0.00001251, -0.00001044, -0.00000684, -0.00000340, -0.00000129, -0.00000085, -0.00000165, -0.00000291, -0.00000395, -0.00000439, -0.00000418, -0.00000340, },
		{ -0.00000349, -0.00000445, -0.00000516, -0.00000524, -0.00000432, -0.00000227, 0.00000058, 0.00000334, 0.00000489, 0.00000431, 0.00000146, -0.00000270, -0.00000644, -0.00000790, -0.00000608, -0.00000148, 0.00000391, 0.00000742, 0.00000694, 0.00000209, -0.00000538, -0.00001223, -0.00001514, -0.00001239, -0.00000488, 0.00000410, 0.00001021, 0.00001010, 0.00000309, -0.00000831, -0.00001939, -0.00002532, -0.00002335, -0.00001419, -0.00000162, 0.00000918, 0.00001408, 0.00001194, 0.00000502, -0.00000224, -0.00000553, -0.00000294, 0.00000405, 0.00001129, 0.00001419, 0.00001020, 0.00000040, -0.00001077, -0.00001759, -0.00001591, -0.00000547, 0.00000963, 0.00002260, 0.00002703, 0.00002007, 0.00000395, -0.00001472, -0.00002785, -0.00002951, -0.00001869, 0.00000018, 0.00001917, 0.00003047, 0.00002986, 0.00001857, 0.00000253, -0.00001057, -0.00001493, -0.00000943, 0.00000199, 0.00001228, 0.00001486, 0.00000692, -0.00000903, -0.00002614, -0.00003633, -0.00003398, -0.00001864, 0.00000461, 0.00002721, 0.00004067, 0.00004006, 0.00002586, 0.00000356, -0.00001880, -0.00003385, -0.00003757, -0.00003032, -0.00001595, 0.00000018, 0.00001357, 0.00002187, 0.00002512, 0.00002474, 0.00002224, 0.00001838, 0.00001305, 0.00000593, -0.00000262, -0.00001117, -0.00001745, -0.00001937, -0.00001606, -0.00000853, 0.00000055, 0.00000791, 0.00001103, 0.00000920, 0.00000379, -0.00000249, -0.00000678, -0.00000736, -0.00000432, 0.00000068, 0.00000523, 0.00000733, 0.00000618, 0.00000243, -0.00000233, -0.00000637, -0.00000851, -0.00000853, -0.00000703, -0.00000496, -0.00000320, -0.00000222, -0.00000209, -0.00000260, },
		{ -0.00000333, -0.00000234, -0.00000182, -0.00000197, -0.00000286, -0.00000439, -0.00000609, -0.00000723, -0.00000698, -0.00000486, -0.00000114, 0.00000311, 0.00000624, 0.00000678, 0.00000418, -0.00000074, -0.00000597, -0.00000913, -0.00000856, -0.00000430, 0.00000183, 0.00000691, 0.00000828, 0.00000493, -0.00000187, -0.00000902, -0.00001299, -0.00001161, -0.00000529, 0.00000300, 0.00000898, 0.00000915, 0.00000262, -0.00000824, -0.00001870, -0.00002380, -0.00002077, -0.00001048, 0.00000280, 0.00001330, 0.00001647, 0.00001127, 0.00000075, -0.00000934, -0.00001331, -0.00000843, 0.00000356, 0.00001702, 0.00002506, 0.00002282, 0.00001005, -0.00000845, -0.00002482, -0.00003168, -0.00002553, -0.00000860, 0.00001203, 0.00002749, 0.00003109, 0.00002135, 0.00000262, -0.00001686, -0.00002874, -0.00002825, -0.00001623, 0.00000145, 0.00001677, 0.00002335, 0.00001929, 0.00000784, -0.00000443, -0.00001096, -0.00000847, 0.00000141, 0.00001303, 0.00001945, 0.00001589, 0.00000219, -0.00001691, -0.00003378, -0.00004119, -0.00003550, -0.00001820, 0.00000470, 0.00002524, 0.00003664, 0.00003584, 0.00002428, 0.00000685, -0.00001036, -0.00002239, -0.00002696, -0.00002455, -0.00001741, -0.00000820, 0.00000116, 0.00000974, 0.00001722, 0.00002315, 0.00002657, 0.00002620, 0.00002113, 0.00001167, -0.00000029, -0.00001168, -0.00001929, -0.00002103, -0.00001681, -0.00000859, 0.00000043, 0.00000710, 0.00000956, 0.00000783, 0.00000356, -0.00000088, -0.00000351, -0.00000351, -0.00000142, 0.00000134, 0.00000326, 0.00000342, 0.00000180, -0.00000084, -0.00000348, -0.00000530, -0.00000594, -0.00000554, -0.00000452, },
		{ -0.00000408, -0.00000471, -0.00000471, -0.00000414, -0.00000314, -0.00000191, -0.00000083, -0.00000038, -0.00000102, -0.00000290, -0.00000562, -0.00000817, -0.00000928, -0.00000800, -0.00000422, 0.00000099, 0.00000573, 0.00000800, 0.00000667, 0.00000212, -0.00000371, -0.00000820, -0.00000921, -0.00000618, -0.00000057, 0.00000480, 0.00000708, 0.00000488, -0.00000089, -0.00000734, -0.00001095, -0.00000940, -0.00000298, 0.00000530, 0.00001099, 0.00001037, 0.00000249, -0.00001002, -0.00002178, -0.00002709, -0.00002266, -0.00000943, 0.00000750, 0.00002092, 0.00002482, 0.00001726, 0.00000154, -0.00001506, -0.00002448, -0.00002163, -0.00000695, 0.00001355, 0.00003075, 0.00003652, 0.00002741, 0.00000643, -0.00001815, -0.00003629, -0.00004058, -0.00002950, -0.00000799, 0.00001474, 0.00002934, 0.00003031, 0.00001826, -0.00000068, -0.00001786, -0.00002608, -0.00002271, -0.00001055, 0.00000384, 0.00001353, 0.00001447, 0.00000733, -0.00000314, -0.00001064, -0.00001057, -0.00000236, 0.00001020, 0.00002073, 0.00002334, 0.00001540, -0.00000111, -0.00002038, -0.00003525, -0.00004015, -0.00003337, -0.00001747, 0.00000192, 0.00001852, 0.00002775, 0.00002810, 0.00002111, 0.00001015, -0.00000115, -0.00001021, -0.00001595, -0.00001835, -0.00001780, -0.00001451, -0.00000843, 0.00000022, 0.00001046, 0.00002032, 0.00002720, 0.00002881, 0.00002420, 0.00001432, 0.00000190, -0.00000948, -0.00001671, -0.00001825, -0.00001458, -0.00000784, -0.00000082, 0.00000418, 0.00000611, 0.00000535, 0.00000320, 0.00000114, 0.00000010, 0.00000022, 0.00000094, 0.00000150, 0.00000132, 0.00000031, -0.00000125, -0.00000285, },
		{ 0.00000128, 0.00000020, -0.00000105, -0.00000229, -0.00000338, -0.00000426, -0.00000479, -0.00000480, -0.00000409, -0.00000263, -0.00000077, 0.00000077, 0.00000115, -0.00000022, -0.00000320, -0.00000679, -0.00000946, -0.00000975, -0.00000706, -0.00000210, 0.00000328, 0.00000683, 0.00000698, 0.00000364, -0.00000160, -0.00000616, -0.00000776, -0.00000555, -0.00000072, 0.00000405, 0.00000588, 0.00000331, -0.00000275, -0.00000926, -0.00001240, -0.00000964, -0.00000128, 0.00000925, 0.00001670, 0.00001646, 0.00000702, -0.00000886, -0.00002485, -0.00003368, -0.00003046, -0.00001521, 0.00000668, 0.00002641, 0.00003552, 0.00002974, 0.00001108, -0.00001265, -0.00003112, -0.00003593, -0.00002438, -0.00000087, 0.00002484, 0.00004181, 0.00004258, 0.00002644, -0.00000025, -0.00002685, -0.00004299, -0.00004291, -0.00002772, -0.00000461, 0.00001652, 0.00002744, 0.00002496, 0.00001186, -0.00000481, -0.00001728, -0.00002059, -0.00001463, -0.00000366, 0.00000604, 0.00000949, 0.00000532, -0.00000368, -0.00001215, -0.00001493, -0.00000966, 0.00000197, 0.00001504, 0.00002357, 0.00002332, 0.00001362, -0.00000238, -0.00001908, -0.00003081, -0.00003402, -0.00002840, -0.00001658, -0.00000278, 0.00000903, 0.00001637, 0.00001873, 0.00001699, 0.00001261, 0.00000679, 0.00000027, -0.00000643, -0.00001254, -0.00001679, -0.00001764, -0.00001397, -0.00000584, 0.00000517, 0.00001623, 0.00002416, 0.00002659, 0.00002291, 0.00001446, 0.00000399, -0.00000536, -0.00001123, -0.00001276, -0.00001063, -0.00000650, -0.00000222, 0.00000094, 0.00000256, 0.00000296, 0.00000279, 0.00000256, 0.00000246, 0.00000236, 0.00000202, },
		{ 0.00000219, 0.00000265, 0.00000270, 0.00000234, 0.00000169, 0.00000093, 0.00000018, -0.00000059, -0.00000153, -0.00000274, -0.00000411, -0.00000522, -0.00000551, -0.00000460, -0.00000256, -0.00000014, 0.00000157, 0.00000155, -0.00000051, -0.00000393, -0.00000722, -0.00000870, -0.00000741, -0.00000362, 0.00000109, 0.00000457, 0.00000511, 0.00000242, -0.00000207, -0.00000591, -0.00000677, -0.00000384, 0.00000159, 0.00000658, 0.00000793, 0.00000399, -0.00000422, -0.00001305, -0.00001782, -0.00001504, -0.00000454, 0.00000999, 0.00002217, 0.00002568, 0.00001737, -0.00000088, -0.00002232, -0.00003794, -0.00004033, -0.00002713, -0.00000257, 0.00002397, 0.00004167, 0.00004291, 0.00002674, -0.00000048, -0.00002756, -0.00004315, -0.00004057, -0.00002071, 0.00000831, 0.00003457, 0.00004737, 0.00004169, 0.00002012, -0.00000843, -0.00003271, -0.00004383, -0.00003883, -0.00002144, -0.00000000, 0.00001646, 0.00002212, 0.00001647, 0.00000397, -0.00000852, -0.00001515, -0.00001369, -0.00000619, 0.00000245, 0.00000713, 0.00000517, -0.00000247, -0.00001178, -0.00001775, -0.00001686, -0.00000877, 0.00000353, 0.00001514, 0.00002139, 0.00001983, 0.00001110, -0.00000155, -0.00001383, -0.00002210, -0.00002463, -0.00002181, -0.00001549, -0.00000791, -0.00000077, 0.00000513, 0.00000958, 0.00001248, 0.00001345, 0.00001187, 0.00000738, 0.00000038, -0.00000763, -0.00001433, -0.00001740, -0.00001542, -0.00000858, 0.00000126, 0.00001121, 0.00001838, 0.00002096, 0.00001871, 0.00001294, 0.00000579, -0.00000060, -0.00000481, -0.00000649, -0.00000611, -0.00000458, -0.00000272, -0.00000104, 0.00000033, 0.00000140, },
		{ -0.00000201, -0.00000165, -0.00000091, -0.00000001, 0.00000079, 0.00000124, 0.00000122, 0.00000086, 0.00000047, 0.00000036, 0.00000061, 0.00000097, 0.00000097, 0.00000017, -0.00000150, -0.00000360, -0.00000531, -0.00000580, -0.00000475, -0.00000259, -0.00000038, 0.00000069, -0.00000005, -0.00000229, -0.00000484, -0.00000612, -0.00000511, -0.00000196, 0.00000190, 0.00000439, 0.00000386, 0.00000015, -0.00000513, -0.00000917, -0.00000935, -0.00000473, 0.00000317, 0.00001068, 0.00001365, 0.00000953, -0.00000103, -0.00001399, -0.00002337, -0.00002395, -0.00001396, 0.00000362, 0.00002180, 0.00003237, 0.00002958, 0.00001303, -0.00001158, -0.00003443, -0.00004563, -0.00003951, -0.00001741, 0.00001249, 0.00003837, 0.00004962, 0.00004139, 0.00001678, -0.00001431, -0.00003930, -0.00004812, -0.00003741, -0.00001184, 0.00001801, 0.00004013, 0.00004607, 0.00003423, 0.00001019, -0.00001606, -0.00003446, -0.00003903, -0.00002994, -0.00001280, 0.00000426, 0.00001435, 0.00001460, 0.00000693, -0.00000346, -0.00001092, -0.00001203, -0.00000704, 0.00000063, 0.00000638, 0.00000679, 0.00000129, -0.00000768, -0.00001591, -0.00001953, -0.00001672, -0.00000846, 0.00000206, 0.00001095, 0.00001526, 0.00001405, 0.00000845, 0.00000082, -0.00000636, -0.00001151, -0.00001420, -0.00001481, -0.00001396, -0.00001192, -0.00000860, -0.00000383, 0.00000206, 0.00000799, 0.00001231, 0.00001340, 0.00001049, 0.00000411, -0.00000385, -0.00001083, -0.00001452, -0.00001375, -0.00000887, -0.00000156, 0.00000587, 0.00001139, 0.00001388, 0.00001332, 0.00001061, 0.00000696, 0.00000346, 0.00000075, -0.00000098, -0.00000184, },
		{ 0.00000721, 0.00000587, 0.00000435, 0.00000290, 0.00000178, 0.00000116, 0.00000104, 0.00000120, 0.00000126, 0.00000087, -0.00000006, -0.00000125, -0.00000216, -0.00000224, -0.00000131, 0.00000032, 0.00000186, 0.00000248, 0.00000172, -0.00000024, -0.00000259, -0.00000432, -0.00000475, -0.00000392, -0.00000260, -0.00000183, -0.00000230, -0.00000381, -0.00000530, -0.00000534, -0.00000303, 0.00000126, 0.00000582, 0.00000823, 0.00000660, 0.00000074, -0.00000734, -0.00001404, -0.00001574, -0.00001067, -0.00000011, 0.00001172, 0.00001931, 0.00001835, 0.00000800, -0.00000821, -0.00002355, -0.00003081, -0.00002569, -0.00000904, 0.00001295, 0.00003108, 0.00003702, 0.00002712, 0.00000442, -0.00002229, -0.00004187, -0.00004561, -0.00003110, -0.00000350, 0.00002638, 0.00004647, 0.00004856, 0.00003170, 0.00000270, -0.00002675, -0.00004498, -0.00004509, -0.00002771, -0.00000045, 0.00002547, 0.00004002, 0.00003829, 0.00002222, -0.00000078, -0.00002128, -0.00003188, -0.00003007, -0.00001869, -0.00000408, 0.00000706, 0.00001063, 0.00000659, -0.00000159, -0.00000899, -0.00001173, -0.00000876, -0.00000206, 0.00000460, 0.00000762, 0.00000530, -0.00000145, -0.00000974, -0.00001612, -0.00001813, -0.00001529, -0.00000905, -0.00000186, 0.00000394, 0.00000705, 0.00000746, 0.00000603, 0.00000376, 0.00000123, -0.00000153, -0.00000474, -0.00000835, -0.00001170, -0.00001363, -0.00001297, -0.00000925, -0.00000307, 0.00000387, 0.00000940, 0.00001171, 0.00001008, 0.00000516, -0.00000126, -0.00000703, -0.00001039, -0.00001058, -0.00000795, -0.00000365, 0.00000097, 0.00000477, 0.00000721, 0.00000822, 0.00000810, },
		{ -0.00000288, -0.00000028, 0.00000222, 0.00000433, 0.00000580, 0.00000648, 0.00000641, 0.00000583, 0.00000514, 0.00000470, 0.00000462, 0.00000466, 0.00000435, 0.00000321, 0.00000115, -0.00000146, -0.00000380, -0.00000500, -0.00000458, -0.00000269, -0.00000015, 0.00000197, 0.00000282, 0.00000223, 0.00000069, -0.00000090, -0.00000183, -0.00000201, -0.00000201, -0.00000270, -0.00000455, -0.00000712, -0.00000905, -0.00000863, -0.00000481, 0.00000193, 0.00000936, 0.00001424, 0.00001370, 0.00000685, -0.00000438, -0.00001569, -0.00002204, -0.00001992, -0.00000929, 0.00000602, 0.00001958, 0.00002507, 0.00001921, 0.00000358, -0.00001569, -0.00003026, -0.00003324, -0.00002240, -0.00000152, 0.00002093, 0.00003525, 0.00003476, 0.00001892, -0.00000614, -0.00003004, -0.00004248, -0.00003770, -0.00001711, 0.00001120, 0.00003575, 0.00004651, 0.00003912, 0.00001666, -0.00001171, -0.00003466, -0.00004342, -0.00003527, -0.00001439, 0.00001021, 0.00002862, 0.00003421, 0.00002598, 0.00000858, -0.00001022, -0.00002301, -0.00002575, -0.00001914, -0.00000762, 0.00000292, 0.00000799, 0.00000625, -0.00000029, -0.00000765, -0.00001194, -0.00001123, -0.00000619, 0.00000046, 0.00000546, 0.00000652, 0.00000331, -0.00000267, -0.00000897, -0.00001337, -0.00001478, -0.00001344, -0.00001045, -0.00000704, -0.00000394, -0.00000125, 0.00000133, 0.00000390, 0.00000610, 0.00000707, 0.00000594, 0.00000240, -0.00000294, -0.00000854, -0.00001255, -0.00001349, -0.00001094, -0.00000572, 0.00000042, 0.00000547, 0.00000796, 0.00000740, 0.00000439, 0.00000018, -0.00000381, -0.00000654, -0.00000757, -0.00000700, -0.00000527, },
		{ -0.00000189, -0.00000419, -0.00000591, -0.00000678, -0.00000663, -0.00000542, -0.00000331, -0.00000070, 0.00000186, 0.00000386, 0.00000508, 0.00000561, 0.00000588, 0.00000629, 0.00000702, 0.00000780, 0.00000804, 0.00000710, 0.00000471, 0.00000120, -0.00000251, -0.00000531, -0.00000636, -0.00000554, -0.00000343, -0.00000105, 0.00000066, 0.00000134, 0.00000129, 0.00000116, 0.00000146, 0.00000205, 0.00000211, 0.00000059, -0.00000305, -0.00000813, -0.00001275, -0.00001438, -0.00001112, -0.00000290, 0.00000798, 0.00001746, 0.00002122, 0.00001679, 0.00000490, -0.00001033, -0.00002280, -0.00002702, -0.00002058, -0.00000558, 0.00001197, 0.00002454, 0.00002633, 0.00001600, -0.00000248, -0.00002131, -0.00003216, -0.00002989, -0.00001498, 0.00000648, 0.00002527, 0.00003306, 0.00002616, 0.00000722, -0.00001570, -0.00003262, -0.00003594, -0.00002382, -0.00000102, 0.00002308, 0.00003848, 0.00003882, 0.00002405, 0.00000028, -0.00002284, -0.00003631, -0.00003541, -0.00002149, -0.00000102, 0.00001739, 0.00002666, 0.00002407, 0.00001195, -0.00000374, -0.00001631, -0.00002118, -0.00001760, -0.00000848, 0.00000129, 0.00000728, 0.00000733, 0.00000225, -0.00000494, -0.00001068, -0.00001251, -0.00001004, -0.00000486, 0.00000051, 0.00000381, 0.00000401, 0.00000148, -0.00000245, -0.00000630, -0.00000917, -0.00001089, -0.00001179, -0.00001218, -0.00001201, -0.00001081, -0.00000803, -0.00000359, 0.00000184, 0.00000689, 0.00000997, 0.00000996, 0.00000673, 0.00000129, -0.00000460, -0.00000905, -0.00001080, -0.00000961, -0.00000623, -0.00000203, 0.00000160, 0.00000372, 0.00000403, 0.00000279, 0.00000061, },
		{ -0.00000116, 0.00000046, 0.00000132, 0.00000125, 0.00000025, -0.00000155, -0.00000383, -0.00000604, -0.00000756, -0.00000789, -0.00000685, -0.00000470, -0.00000203, 0.00000048, 0.00000237, 0.00000362, 0.00000457, 0.00000567, 0.00000717, 0.00000882, 0.00001000, 0.00000992, 0.00000814, 0.00000483, 0.00000082, -0.00000275, -0.00000494, -0.00000544, -0.00000464, -0.00000342, -0.00000254, -0.00000220, -0.00000194, -0.00000096, 0.00000121, 0.00000418, 0.00000660, 0.00000669, 0.00000318, -0.00000367, -0.00001175, -0.00001769, -0.00001822, -0.00001181, 0.00000023, 0.00001382, 0.00002358, 0.00002502, 0.00001672, 0.00000127, -0.00001549, -0.00002676, -0.00002771, -0.00001769, -0.00000075, 0.00001595, 0.00002517, 0.00002279, 0.00000975, -0.00000828, -0.00002331, -0.00002861, -0.00002170, -0.00000557, 0.00001258, 0.00002460, 0.00002499, 0.00001347, -0.00000483, -0.00002166, -0.00002924, -0.00002385, -0.00000740, 0.00001326, 0.00002942, 0.00003423, 0.00002565, 0.00000724, -0.00001348, -0.00002828, -0.00003173, -0.00002329, -0.00000729, 0.00000923, 0.00001957, 0.00002020, 0.00001190, -0.00000095, -0.00001257, -0.00001833, -0.00001661, -0.00000915, 0.00000014, 0.00000706, 0.00000899, 0.00000577, -0.00000058, -0.00000703, -0.00001102, -0.00001143, -0.00000882, -0.00000483, -0.00000123, 0.00000083, 0.00000114, 0.00000020, -0.00000131, -0.00000303, -0.00000502, -0.00000752, -0.00001052, -0.00001352, -0.00001550, -0.00001536, -0.00001249, -0.00000713, -0.00000050, 0.00000567, 0.00000970, 0.00001063, 0.00000852, 0.00000439, -0.00000029, -0.00000411, -0.00000623, -0.00000646, -0.00000523, -0.00000323, },
		{ 0.00000093, -0.00000134, -0.00000279, -0.00000335, -0.00000306, -0.00000212, -0.00000090, 0.00000010, 0.00000034, -0.00000053, -0.00000250, -0.00000512, -0.00000761, -0.00000914, -0.00000921, -0.00000781, -0.00000542, -0.00000276, -0.00000042, 0.00000139, 0.00000289, 0.00000447, 0.00000638, 0.00000847, 0.00001011, 0.00001057, 0.00000938, 0.00000665, 0.00000318, 0.00000004, -0.00000191, -0.00000246, -0.00000218, -0.00000204, -0.00000282, -0.00000450, -0.00000615, -0.00000634, -0.00000396, 0.00000091, 0.00000668, 0.00001074, 0.00001057, 0.00000511, -0.00000434, -0.00001432, -0.00002044, -0.00001931, -0.00001029, 0.00000392, 0.00001801, 0.00002620, 0.00002473, 0.00001368, -0.00000288, -0.00001837, -0.00002653, -0.00002409, -0.00001230, 0.00000362, 0.00001668, 0.00002125, 0.00001555, 0.00000243, -0.00001195, -0.00002095, -0.00002048, -0.00001090, 0.00000317, 0.00001502, 0.00001898, 0.00001307, -0.00000002, -0.00001415, -0.00002251, -0.00002080, -0.00000920, 0.00000761, 0.00002253, 0.00002912, 0.00002447, 0.00001041, -0.00000737, -0.00002176, -0.00002735, -0.00002257, -0.00001017, 0.00000416, 0.00001440, 0.00001673, 0.00001093, 0.00000018, -0.00001059, -0.00001693, -0.00001668, -0.00001064, -0.00000187, 0.00000586, 0.00000968, 0.00000870, 0.00000405, -0.00000196, -0.00000698, -0.00000959, -0.00000964, -0.00000795, -0.00000570, -0.00000372, -0.00000230, -0.00000129, -0.00000050, -0.00000002, -0.00000032, -0.00000195, -0.00000513, -0.00000943, -0.00001377, -0.00001675, -0.00001718, -0.00001463, -0.00000960, -0.00000334, 0.00000258, 0.00000686, 0.00000880, 0.00000848, 0.00000650, 0.00000373, },
		{ 0.00000262, 0.00000571, 0.00000739, 0.00000768, 0.00000674, 0.00000489, 0.00000258, 0.00000040, -0.00000113, -0.00000164, -0.00000117, -0.00000020, 0.00000054, 0.00000032, -0.00000124, -0.00000395, -0.00000709, -0.00000971, -0.00001105, -0.00001077, -0.00000910, -0.00000663, -0.00000397, -0.00000144, 0.00000098, 0.00000348, 0.00000616, 0.00000871, 0.00001050, 0.00001080, 0.00000925, 0.00000620, 0.00000267, -0.00000003, -0.00000095, -0.00000005, 0.00000170, 0.00000270, 0.00000161, -0.00000185, -0.00000656, -0.00001031, -0.00001082, -0.00000701, 0.00000026, 0.00000824, 0.00001334, 0.00001278, 0.00000603, -0.00000462, -0.00001484, -0.00001997, -0.00001718, -0.00000688, 0.00000719, 0.00001942, 0.00002469, 0.00002062, 0.00000872, -0.00000621, -0.00001809, -0.00002224, -0.00001737, -0.00000616, 0.00000609, 0.00001385, 0.00001391, 0.00000675, -0.00000376, -0.00001232, -0.00001472, -0.00000989, -0.00000037, 0.00000892, 0.00001314, 0.00000984, 0.00000030, -0.00001105, -0.00001862, -0.00001833, -0.00000955, 0.00000446, 0.00001794, 0.00002514, 0.00002289, 0.00001194, -0.00000333, -0.00001685, -0.00002349, -0.00002113, -0.00001145, 0.00000098, 0.00001080, 0.00001414, 0.00001013, 0.00000100, -0.00000908, -0.00001592, -0.00001706, -0.00001254, -0.00000463, 0.00000339, 0.00000867, 0.00000986, 0.00000732, 0.00000269, -0.00000211, -0.00000565, -0.00000746, -0.00000785, -0.00000748, -0.00000681, -0.00000595, -0.00000471, -0.00000297, -0.00000094, 0.00000070, 0.00000113, -0.00000029, -0.00000358, -0.00000807, -0.00001255, -0.00001572, -0.00001665, -0.00001509, -0.00001146, -0.00000665, -0.00000168, },
		{ -0.00001467, -0.00001469, -0.00001307, -0.00001008, -0.00000613, -0.00000174, 0.00000249, 0.00000589, 0.00000790, 0.00000823, 0.00000701, 0.00000483, 0.00000256, 0.00000102, 0.00000060, 0.00000116, 0.00000200, 0.00000223, 0.00000117, -0.00000134, -0.00000484, -0.00000844, -0.00001126, -0.00001271, -0.00001265, -0.00001136, -0.00000920, -0.00000638, -0.00000294, 0.00000113, 0.00000561, 0.00000979, 0.00001263, 0.00001312, 0.00001080, 0.00000624, 0.00000097, -0.00000302, -0.00000411, -0.00000192, 0.00000240, 0.00000654, 0.00000802, 0.00000544, -0.00000073, -0.00000818, -0.00001363, -0.00001435, -0.00000953, -0.00000080, 0.00000825, 0.00001367, 0.00001289, 0.00000603, -0.00000402, -0.00001278, -0.00001614, -0.00001228, -0.00000253, 0.00000910, 0.00001770, 0.00001959, 0.00001401, 0.00000341, -0.00000769, -0.00001473, -0.00001518, -0.00000956, -0.00000111, 0.00000591, 0.00000825, 0.00000521, -0.00000114, -0.00000710, -0.00000925, -0.00000626, 0.00000044, 0.00000729, 0.00001042, 0.00000762, -0.00000043, -0.00001028, -0.00001729, -0.00001772, -0.00001063, 0.00000158, 0.00001411, 0.00002179, 0.00002137, 0.00001290, -0.00000022, -0.00001279, -0.00001999, -0.00001940, -0.00001190, -0.00000115, 0.00000812, 0.00001211, 0.00000950, 0.00000178, -0.00000759, -0.00001475, -0.00001708, -0.00001408, -0.00000732, 0.00000042, 0.00000648, 0.00000926, 0.00000862, 0.00000563, 0.00000182, -0.00000158, -0.00000398, -0.00000547, -0.00000643, -0.00000714, -0.00000753, -0.00000728, -0.00000610, -0.00000403, -0.00000157, 0.00000047, 0.00000123, 0.00000026, -0.00000238, -0.00000607, -0.00000990, -0.00001299, },
		{ 0.00000036, -0.00000096, -0.00000332, -0.00000632, -0.00000948, -0.00001228, -0.00001413, -0.00001452, -0.00001309, -0.00000984, -0.00000523, -0.00000017, 0.00000427, 0.00000715, 0.00000803, 0.00000716, 0.00000535, 0.00000363, 0.00000278, 0.00000302, 0.00000394, 0.00000469, 0.00000444, 0.00000272, -0.00000036, -0.00000422, -0.00000810, -0.00001136, -0.00001364, -0.00001482, -0.00001480, -0.00001335, -0.00001013, -0.00000498, 0.00000173, 0.00000889, 0.00001473, 0.00001741, 0.00001580, 0.00001015, 0.00000224, -0.00000510, -0.00000908, -0.00000812, -0.00000270, 0.00000476, 0.00001083, 0.00001251, 0.00000861, 0.00000041, -0.00000885, -0.00001532, -0.00001624, -0.00001120, -0.00000235, 0.00000651, 0.00001167, 0.00001114, 0.00000554, -0.00000222, -0.00000830, -0.00000976, -0.00000586, 0.00000158, 0.00000912, 0.00001333, 0.00001237, 0.00000681, -0.00000073, -0.00000691, -0.00000926, -0.00000734, -0.00000281, 0.00000153, 0.00000324, 0.00000149, -0.00000249, -0.00000612, -0.00000687, -0.00000372, 0.00000221, 0.00000805, 0.00001055, 0.00000777, 0.00000020, -0.00000926, -0.00001648, -0.00001795, -0.00001248, -0.00000182, 0.00000993, 0.00001805, 0.00001926, 0.00001307, 0.00000203, -0.00000938, -0.00001671, -0.00001737, -0.00001163, -0.00000239, 0.00000620, 0.00001053, 0.00000896, 0.00000242, -0.00000623, -0.00001350, -0.00001673, -0.00001503, -0.00000944, -0.00000225, 0.00000406, 0.00000780, 0.00000852, 0.00000687, 0.00000408, 0.00000126, -0.00000099, -0.00000265, -0.00000400, -0.00000532, -0.00000659, -0.00000748, -0.00000759, -0.00000668, -0.00000488, -0.00000267, -0.00000070, 0.00000043, },
		{ -0.00000694, -0.00000590, -0.00000427, -0.00000235, -0.00000056, 0.00000069, 0.00000095, -0.00000010, -0.00000256, -0.00000616, -0.00001022, -0.00001372, -0.00001558, -0.00001509, -0.00001215, -0.00000740, -0.00000206, 0.00000255, 0.00000545, 0.00000637, 0.00000578, 0.00000462, 0.00000383, 0.00000395, 0.00000491, 0.00000612, 0.00000680, 0.00000639, 0.00000470, 0.00000191, -0.00000161, -0.00000550, -0.00000955, -0.00001353, -0.00001696, -0.00001900, -0.00001856, -0.00001472, -0.00000736, 0.00000246, 0.00001245, 0.00001973, 0.00002185, 0.00001789, 0.00000900, -0.00000177, -0.00001054, -0.00001409, -0.00001122, -0.00000326, 0.00000635, 0.00001352, 0.00001518, 0.00001056, 0.00000151, -0.00000832, -0.00001508, -0.00001632, -0.00001200, -0.00000432, 0.00000333, 0.00000798, 0.00000831, 0.00000512, 0.00000072, -0.00000230, -0.00000236, 0.00000042, 0.00000439, 0.00000731, 0.00000759, 0.00000503, 0.00000093, -0.00000272, -0.00000426, -0.00000327, -0.00000080, 0.00000130, 0.00000141, -0.00000089, -0.00000450, -0.00000729, -0.00000723, -0.00000356, 0.00000258, 0.00000858, 0.00001143, 0.00000925, 0.00000229, -0.00000695, -0.00001468, -0.00001748, -0.00001387, -0.00000502, 0.00000563, 0.00001382, 0.00001626, 0.00001204, 0.00000303, -0.00000702, -0.00001405, -0.00001540, -0.00001093, -0.00000290, 0.00000507, 0.00000954, 0.00000873, 0.00000310, -0.00000498, -0.00001233, -0.00001629, -0.00001569, -0.00001112, -0.00000448, 0.00000194, 0.00000634, 0.00000801, 0.00000728, 0.00000516, 0.00000273, 0.00000069, -0.00000079, -0.00000192, -0.00000303, -0.00000429, -0.00000560, -0.00000669, -0.00000720, },
		{ -0.00000225, -0.00000309, -0.00000422, -0.00000551, -0.00000667, -0.00000735, -0.00000721, -0.00000604, -0.00000385, -0.00000106, 0.00000162, 0.00000325, 0.00000303, 0.00000060, -0.00000368, -0.00000879, -0.00001329, -0.00001585, -0.00001567, -0.00001287, -0.00000833, -0.00000343, 0.00000058, 0.00000298, 0.00000380, 0.00000364, 0.00000335, 0.00000355, 0.00000443, 0.00000570, 0.00000689, 0.00000756, 0.00000752, 0.00000676, 0.00000527, 0.00000285, -0.00000082, -0.00000594, -0.00001221, -0.00001851, -0.00002301, -0.00002370, -0.00001922, -0.00000971, 0.00000286, 0.00001512, 0.00002329, 0.00002468, 0.00001878, 0.00000763, -0.00000478, -0.00001404, -0.00001695, -0.00001280, -0.00000362, 0.00000672, 0.00001405, 0.00001561, 0.00001106, 0.00000245, -0.00000675, -0.00001319, -0.00001493, -0.00001208, -0.00000643, -0.00000049, 0.00000372, 0.00000543, 0.00000519, 0.00000426, 0.00000382, 0.00000426, 0.00000509, 0.00000536, 0.00000434, 0.00000207, -0.00000057, -0.00000236, -0.00000239, -0.00000072, 0.00000159, 0.00000296, 0.00000217, -0.00000089, -0.00000502, -0.00000812, -0.00000828, -0.00000476, 0.00000143, 0.00000781, 0.00001147, 0.00001041, 0.00000456, -0.00000403, -0.00001193, -0.00001584, -0.00001401, -0.00000707, 0.00000218, 0.00000992, 0.00001300, 0.00001021, 0.00000283, -0.00000597, -0.00001251, -0.00001419, -0.00001051, -0.00000327, 0.00000434, 0.00000908, 0.00000900, 0.00000417, -0.00000346, -0.00001100, -0.00001575, -0.00001627, -0.00001271, -0.00000660, -0.00000007, 0.00000501, 0.00000760, 0.00000768, 0.00000602, 0.00000365, 0.00000146, -0.00000011, -0.00000105, -0.00000165, },
		{ 0.00000225, 0.00000012, -0.00000114, -0.00000157, -0.00000150, -0.00000139, -0.00000172, -0.00000276, -0.00000447, -0.00000647, -0.00000809, -0.00000859, -0.00000744, -0.00000461, -0.00000073, 0.00000306, 0.00000543, 0.00000537, 0.00000260, -0.00000223, -0.00000777, -0.00001242, -0.00001490, -0.00001471, -0.00001223, -0.00000850, -0.00000472, -0.00000178, 0.00000003, 0.00000095, 0.00000152, 0.00000219, 0.00000313, 0.00000420, 0.00000517, 0.00000595, 0.00000668, 0.00000755, 0.00000853, 0.00000912, 0.00000835, 0.00000513, -0.00000106, -0.00000965, -0.00001875, -0.00002557, -0.00002736, -0.00002257, -0.00001172, 0.00000246, 0.00001582, 0.00002419, 0.00002497, 0.00001814, 0.00000634, -0.00000612, -0.00001488, -0.00001714, -0.00001268, -0.00000379, 0.00000581, 0.00001246, 0.00001395, 0.00001020, 0.00000307, -0.00000465, -0.00001039, -0.00001271, -0.00001162, -0.00000819, -0.00000387, 0.00000017, 0.00000341, 0.00000583, 0.00000756, 0.00000857, 0.00000858, 0.00000726, 0.00000461, 0.00000124, -0.00000177, -0.00000328, -0.00000269, -0.00000035, 0.00000248, 0.00000417, 0.00000349, 0.00000033, -0.00000409, -0.00000772, -0.00000857, -0.00000577, -0.00000006, 0.00000632, 0.00001058, 0.00001063, 0.00000604, -0.00000157, -0.00000920, -0.00001372, -0.00001322, -0.00000789, 0.00000001, 0.00000713, 0.00001044, 0.00000856, 0.00000231, -0.00000564, -0.00001193, -0.00001396, -0.00001098, -0.00000432, 0.00000321, 0.00000849, 0.00000940, 0.00000561, -0.00000141, -0.00000907, -0.00001469, -0.00001652, -0.00001423, -0.00000888, -0.00000235, 0.00000343, 0.00000711, 0.00000824, 0.00000719, 0.00000487, },
		{ 0.00000602, 0.00000849, 0.00000850, 0.00000660, 0.00000372, 0.00000084, -0.00000125, -0.00000217, -0.00000197, -0.00000114, -0.00000039, -0.00000043, -0.00000167, -0.00000401, -0.00000678, -0.00000897, -0.00000954, -0.00000792, -0.00000431, 0.00000033, 0.00000451, 0.00000682, 0.00000638, 0.00000326, -0.00000158, -0.00000668, -0.00001062, -0.00001255, -0.00001239, -0.00001074, -0.00000848, -0.00000631, -0.00000455, -0.00000308, -0.00000163, -0.00000003, 0.00000157, 0.00000281, 0.00000343, 0.00000351, 0.00000357, 0.00000439, 0.00000647, 0.00000958, 0.00001259, 0.00001368, 0.00001117, 0.00000434, -0.00000591, -0.00001702, -0.00002555, -0.00002835, -0.00002393, -0.00001312, 0.00000094, 0.00001392, 0.00002180, 0.00002236, 0.00001594, 0.00000528, -0.00000563, -0.00001305, -0.00001481, -0.00001102, -0.00000373, 0.00000400, 0.00000938, 0.00001081, 0.00000829, 0.00000315, -0.00000272, -0.00000756, -0.00001031, -0.00001069, -0.00000900, -0.00000580, -0.00000170, 0.00000270, 0.00000674, 0.00000969, 0.00001085, 0.00000982, 0.00000683, 0.00000277, -0.00000099, -0.00000312, -0.00000291, -0.00000067, 0.00000231, 0.00000434, 0.00000411, 0.00000136, -0.00000290, -0.00000672, -0.00000813, -0.00000607, -0.00000104, 0.00000504, 0.00000956, 0.00001039, 0.00000685, 0.00000015, -0.00000710, -0.00001195, -0.00001239, -0.00000828, -0.00000138, 0.00000531, 0.00000891, 0.00000782, 0.00000247, -0.00000493, -0.00001132, -0.00001408, -0.00001213, -0.00000630, 0.00000101, 0.00000690, 0.00000908, 0.00000677, 0.00000085, -0.00000650, -0.00001275, -0.00001592, -0.00001520, -0.00001107, -0.00000499, 0.00000123, },
		{ -0.00001511, -0.00001257, -0.00000754, -0.00000147, 0.00000408, 0.00000788, 0.00000930, 0.00000841, 0.00000581, 0.00000248, -0.00000054, -0.00000243, -0.00000282, -0.00000193, -0.00000048, 0.00000056, 0.00000038, -0.00000137, -0.00000430, -0.00000743, -0.00000950, -0.00000952, -0.00000717, -0.00000302, 0.00000164, 0.00000531, 0.00000684, 0.00000582, 0.00000277, -0.00000124, -0.00000496, -0.00000758, -0.00000885, -0.00000913, -0.00000895, -0.00000870, -0.00000837, -0.00000761, -0.00000605, -0.00000364, -0.00000084, 0.00000154, 0.00000271, 0.00000243, 0.00000128, 0.00000046, 0.00000130, 0.00000449, 0.00000954, 0.00001472, 0.00001762, 0.00001613, 0.00000951, -0.00000114, -0.00001294, -0.00002228, -0.00002605, -0.00002291, -0.00001381, -0.00000173, 0.00000944, 0.00001631, 0.00001716, 0.00001244, 0.00000445, -0.00000370, -0.00000924, -0.00001069, -0.00000823, -0.00000333, 0.00000194, 0.00000573, 0.00000701, 0.00000569, 0.00000246, -0.00000161, -0.00000545, -0.00000819, -0.00000927, -0.00000841, -0.00000569, -0.00000155, 0.00000319, 0.00000741, 0.00001000, 0.00001023, 0.00000812, 0.00000447, 0.00000071, -0.00000174, -0.00000201, -0.00000024, 0.00000246, 0.00000447, 0.00000444, 0.00000199, -0.00000206, -0.00000592, -0.00000768, -0.00000618, -0.00000169, 0.00000414, 0.00000887, 0.00001034, 0.00000767, 0.00000168, -0.00000535, -0.00001061, -0.00001200, -0.00000901, -0.00000295, 0.00000357, 0.00000774, 0.00000776, 0.00000359, -0.00000310, -0.00000961, -0.00001335, -0.00001291, -0.00000854, -0.00000206, 0.00000398, 0.00000729, 0.00000667, 0.00000243, -0.00000390, -0.00001012, -0.00001426, },
		{ 0.00000239, -0.00000230, -0.00000768, -0.00001198, -0.00001391, -0.00001293, -0.00000933, -0.00000406, 0.00000157, 0.00000628, 0.00000909, 0.00000955, 0.00000785, 0.00000469, 0.00000115, -0.00000168, -0.00000304, -0.00000277, -0.00000133, 0.00000030, 0.00000107, 0.00000029, -0.00000203, -0.00000518, -0.00000794, -0.00000915, -0.00000817, -0.00000519, -0.00000117, 0.00000258, 0.00000489, 0.00000523, 0.00000385, 0.00000157, -0.00000069, -0.00000234, -0.00000335, -0.00000419, -0.00000542, -0.00000725, -0.00000931, -0.00001077, -0.00001071, -0.00000869, -0.00000510, -0.00000108, 0.00000193, 0.00000285, 0.00000161, -0.00000077, -0.00000250, -0.00000187, 0.00000190, 0.00000814, 0.00001481, 0.00001923, 0.00001918, 0.00001386, 0.00000437, -0.00000663, -0.00001586, -0.00002059, -0.00001964, -0.00001374, -0.00000518, 0.00000312, 0.00000867, 0.00001023, 0.00000807, 0.00000363, -0.00000113, -0.00000456, -0.00000577, -0.00000481, -0.00000243, 0.00000030, 0.00000242, 0.00000331, 0.00000275, 0.00000094, -0.00000166, -0.00000438, -0.00000648, -0.00000725, -0.00000628, -0.00000356, 0.00000033, 0.00000435, 0.00000732, 0.00000832, 0.00000714, 0.00000437, 0.00000126, -0.00000083, -0.00000100, 0.00000071, 0.00000333, 0.00000533, 0.00000539, 0.00000304, -0.00000102, -0.00000514, -0.00000741, -0.00000655, -0.00000258, 0.00000309, 0.00000817, 0.00001045, 0.00000878, 0.00000357, -0.00000324, -0.00000906, -0.00001164, -0.00001006, -0.00000511, 0.00000108, 0.00000590, 0.00000737, 0.00000492, -0.00000042, -0.00000649, -0.00001092, -0.00001210, -0.00000973, -0.00000492, 0.00000036, 0.00000406, 0.00000481, },
		{ -0.00000615, -0.00000258, 0.00000041, 0.00000161, 0.00000053, -0.00000249, -0.00000644, -0.00001009, -0.00001227, -0.00001229, -0.00001001, -0.00000589, -0.00000079, 0.00000416, 0.00000788, 0.00000961, 0.00000909, 0.00000669, 0.00000327, -0.00000003, -0.00000221, -0.00000277, -0.00000188, -0.00000033, 0.00000084, 0.00000075, -0.00000086, -0.00000351, -0.00000620, -0.00000779, -0.00000754, -0.00000549, -0.00000236, 0.00000067, 0.00000261, 0.00000300, 0.00000216, 0.00000092, 0.00000018, 0.00000035, 0.00000113, 0.00000165, 0.00000088, -0.00000172, -0.00000579, -0.00001015, -0.00001317, -0.00001355, -0.00001096, -0.00000623, -0.00000112, 0.00000245, 0.00000325, 0.00000133, -0.00000197, -0.00000456, -0.00000452, -0.00000099, 0.00000539, 0.00001264, 0.00001823, 0.00002000, 0.00001708, 0.00001016, 0.00000124, -0.00000714, -0.00001282, -0.00001468, -0.00001290, -0.00000872, -0.00000381, 0.00000030, 0.00000272, 0.00000332, 0.00000255, 0.00000115, -0.00000020, -0.00000107, -0.00000130, -0.00000099, -0.00000042, 0.00000010, 0.00000025, -0.00000019, -0.00000126, -0.00000277, -0.00000423, -0.00000504, -0.00000466, -0.00000289, -0.00000003, 0.00000310, 0.00000546, 0.00000622, 0.00000510, 0.00000262, -0.00000009, -0.00000176, -0.00000151, 0.00000064, 0.00000375, 0.00000631, 0.00000691, 0.00000495, 0.00000094, -0.00000356, -0.00000663, -0.00000678, -0.00000370, 0.00000155, 0.00000689, 0.00001010, 0.00000973, 0.00000572, -0.00000052, -0.00000668, -0.00001050, -0.00001069, -0.00000741, -0.00000216, 0.00000285, 0.00000565, 0.00000524, 0.00000198, -0.00000272, -0.00000695, -0.00000919, -0.00000877, },
		{ 0.00000033, -0.00000260, -0.00000478, -0.00000556, -0.00000493, -0.00000349, -0.00000214, -0.00000164, -0.00000241, -0.00000435, -0.00000697, -0.00000952, -0.00001121, -0.00001144, -0.00000993, -0.00000676, -0.00000243, 0.00000224, 0.00000628, 0.00000879, 0.00000928, 0.00000779, 0.00000494, 0.00000175, -0.00000074, -0.00000184, -0.00000150, -0.00000032, 0.00000073, 0.00000077, -0.00000056, -0.00000292, -0.00000537, -0.00000684, -0.00000661, -0.00000469, -0.00000186, 0.00000070, 0.00000195, 0.00000151, -0.00000013, -0.00000182, -0.00000233, -0.00000099, 0.00000188, 0.00000502, 0.00000676, 0.00000576, 0.00000168, -0.00000451, -0.00001085, -0.00001515, -0.00001589, -0.00001287, -0.00000730, -0.00000134, 0.00000284, 0.00000387, 0.00000181, -0.00000196, -0.00000532, -0.00000631, -0.00000390, 0.00000157, 0.00000858, 0.00001499, 0.00001885, 0.00001905, 0.00001561, 0.00000954, 0.00000244, -0.00000409, -0.00000886, -0.00001133, -0.00001153, -0.00000995, -0.00000728, -0.00000420, -0.00000131, 0.00000097, 0.00000239, 0.00000290, 0.00000263, 0.00000185, 0.00000086, -0.00000012, -0.00000102, -0.00000189, -0.00000280, -0.00000372, -0.00000443, -0.00000453, -0.00000366, -0.00000169, 0.00000107, 0.00000388, 0.00000577, 0.00000597, 0.00000428, 0.00000124, -0.00000197, -0.00000400, -0.00000388, -0.00000151, 0.00000224, 0.00000581, 0.00000764, 0.00000682, 0.00000355, -0.00000088, -0.00000460, -0.00000596, -0.00000424, -0.00000001, 0.00000507, 0.00000891, 0.00000988, 0.00000744, 0.00000237, -0.00000352, -0.00000818, -0.00001011, -0.00000890, -0.00000535, -0.00000102, 0.00000240, 0.00000375, 0.00000284, },
		{ -0.00000468, -0.00000168, 0.00000057, 0.00000160, 0.00000147, 0.00000066, -0.00000030, -0.00000105, -0.00000150, -0.00000186, -0.00000242, -0.00000343, -0.00000496, -0.00000687, -0.00000886, -0.00001055, -0.00001151, -0.00001139, -0.00000996, -0.00000718, -0.00000333, 0.00000101, 0.00000501, 0.00000784, 0.00000887, 0.00000799, 0.00000566, 0.00000279, 0.00000042, -0.00000072, -0.00000049, 0.00000058, 0.00000154, 0.00000146, -0.00000007, -0.00000269, -0.00000541, -0.00000703, -0.00000671, -0.00000441, -0.00000101, 0.00000204, 0.00000337, 0.00000236, -0.00000053, -0.00000390, -0.00000593, -0.00000529, -0.00000174, 0.00000361, 0.00000870, 0.00001129, 0.00000993, 0.00000460, -0.00000322, -0.00001104, -0.00001636, -0.00001754, -0.00001445, -0.00000848, -0.00000190, 0.00000297, 0.00000466, 0.00000302, -0.00000078, -0.00000483, -0.00000717, -0.00000653, -0.00000268, 0.00000350, 0.00001044, 0.00001636, 0.00001982, 0.00002000, 0.00001692, 0.00001124, 0.00000413, -0.00000306, -0.00000905, -0.00001289, -0.00001412, -0.00001283, -0.00000961, -0.00000538, -0.00000115, 0.00000223, 0.00000428, 0.00000488, 0.00000427, 0.00000285, 0.00000097, -0.00000107, -0.00000306, -0.00000477, -0.00000586, -0.00000595, -0.00000474, -0.00000219, 0.00000126, 0.00000473, 0.00000713, 0.00000756, 0.00000569, 0.00000203, -0.00000219, -0.00000546, -0.00000653, -0.00000497, -0.00000137, 0.00000286, 0.00000605, 0.00000697, 0.00000532, 0.00000193, -0.00000165, -0.00000378, -0.00000344, -0.00000067, 0.00000346, 0.00000725, 0.00000912, 0.00000820, 0.00000471, -0.00000021, -0.00000495, -0.00000808, -0.00000888, -0.00000746, },
		{ -0.00000151, -0.00000505, -0.00000727, -0.00000789, -0.00000712, -0.00000544, -0.00000339, -0.00000136, 0.00000041, 0.00000182, 0.00000279, 0.00000319, 0.00000286, 0.00000169, -0.00000028, -0.00000287, -0.00000574, -0.00000852, -0.00001085, -0.00001245, -0.00001311, -0.00001268, -0.00001102, -0.00000813, -0.00000423, 0.00000018, 0.00000434, 0.00000740, 0.00000868, 0.00000800, 0.00000578, 0.00000293, 0.00000058, -0.00000043, 0.00000010, 0.00000161, 0.00000303, 0.00000323, 0.00000163, -0.00000152, -0.00000512, -0.00000768, -0.00000798, -0.00000567, -0.00000153, 0.00000279, 0.00000544, 0.00000516, 0.00000191, -0.00000303, -0.00000753, -0.00000944, -0.00000755, -0.00000211, 0.00000511, 0.00001152, 0.00001463, 0.00001304, 0.00000698, -0.00000178, -0.00001052, -0.00001660, -0.00001836, -0.00001563, -0.00000969, -0.00000276, 0.00000284, 0.00000546, 0.00000463, 0.00000109, -0.00000351, -0.00000726, -0.00000855, -0.00000655, -0.00000142, 0.00000576, 0.00001334, 0.00001949, 0.00002271, 0.00002213, 0.00001780, 0.00001058, 0.00000199, -0.00000620, -0.00001241, -0.00001565, -0.00001567, -0.00001295, -0.00000853, -0.00000359, 0.00000081, 0.00000397, 0.00000561, 0.00000573, 0.00000455, 0.00000238, -0.00000042, -0.00000336, -0.00000582, -0.00000713, -0.00000674, -0.00000448, -0.00000069, 0.00000366, 0.00000731, 0.00000903, 0.00000817, 0.00000488, 0.00000020, -0.00000433, -0.00000721, -0.00000754, -0.00000539, -0.00000173, 0.00000197, 0.00000433, 0.00000465, 0.00000313, 0.00000077, -0.00000112, -0.00000148, 0.00000002, 0.00000287, 0.00000595, 0.00000799, 0.00000812, 0.00000613, 0.00000258, },
		{ 0.00000719, 0.00000774, 0.00000683, 0.00000461, 0.00000157, -0.00000169, -0.00000457, -0.00000670, -0.00000783, -0.00000790, -0.00000691, -0.00000494, -0.00000222, 0.00000086, 0.00000374, 0.00000582, 0.00000656, 0.00000569, 0.00000330, -0.00000021, -0.00000423, -0.00000815, -0.00001145, -0.00001379, -0.00001497, -0.00001485, -0.00001330, -0.00001033, -0.00000611, -0.00000117, 0.00000366, 0.00000740, 0.00000918, 0.00000869, 0.00000627, 0.00000295, 0.00000005, -0.00000130, -0.00000066, 0.00000149, 0.00000395, 0.00000525, 0.00000439, 0.00000126, -0.00000316, -0.00000719, -0.00000910, -0.00000793, -0.00000394, 0.00000140, 0.00000595, 0.00000774, 0.00000587, 0.00000089, -0.00000534, -0.00001028, -0.00001178, -0.00000888, -0.00000226, 0.00000598, 0.00001304, 0.00001644, 0.00001490, 0.00000874, -0.00000019, -0.00000926, -0.00001586, -0.00001828, -0.00001620, -0.00001065, -0.00000366, 0.00000252, 0.00000606, 0.00000610, 0.00000299, -0.00000195, -0.00000679, -0.00000964, -0.00000918, -0.00000504, 0.00000206, 0.00001050, 0.00001823, 0.00002329, 0.00002438, 0.00002120, 0.00001448, 0.00000577, -0.00000303, -0.00001018, -0.00001454, -0.00001572, -0.00001409, -0.00001049, -0.00000594, -0.00000139, 0.00000243, 0.00000503, 0.00000612, 0.00000561, 0.00000362, 0.00000055, -0.00000290, -0.00000580, -0.00000721, -0.00000653, -0.00000372, 0.00000055, 0.00000505, 0.00000837, 0.00000943, 0.00000783, 0.00000409, -0.00000060, -0.00000477, -0.00000721, -0.00000742, -0.00000570, -0.00000299, -0.00000041, 0.00000118, 0.00000153, 0.00000101, 0.00000037, 0.00000037, 0.00000139, 0.00000329, 0.00000549, },
		{ 0.00000056, 0.00000202, 0.00000375, 0.00000548, 0.00000683, 0.00000747, 0.00000722, 0.00000605, 0.00000409, 0.00000151, -0.00000141, -0.00000437, -0.00000694, -0.00000867, -0.00000910, -0.00000793, -0.00000520, -0.00000135, 0.00000283, 0.00000640, 0.00000850, 0.00000861, 0.00000673, 0.00000325, -0.00000114, -0.00000571, -0.00000989, -0.00001326, -0.00001556, -0.00001654, -0.00001594, -0.00001355, -0.00000942, -0.00000399, 0.00000185, 0.00000686, 0.00000987, 0.00001017, 0.00000788, 0.00000399, 0.00000002, -0.00000248, -0.00000258, -0.00000039, 0.00000300, 0.00000589, 0.00000675, 0.00000484, 0.00000062, -0.00000441, -0.00000825, -0.00000927, -0.00000690, -0.00000193, 0.00000374, 0.00000785, 0.00000865, 0.00000561, -0.00000028, -0.00000688, -0.00001168, -0.00001272, -0.00000933, -0.00000238, 0.00000598, 0.00001309, 0.00001664, 0.00001545, 0.00000981, 0.00000137, -0.00000746, -0.00001423, -0.00001722, -0.00001589, -0.00001099, -0.00000427, 0.00000210, 0.00000618, 0.00000688, 0.00000421, -0.00000068, -0.00000594, -0.00000954, -0.00000993, -0.00000655, 0.00000005, 0.00000832, 0.00001618, 0.00002167, 0.00002346, 0.00002121, 0.00001559, 0.00000799, 0.00000006, -0.00000671, -0.00001135, -0.00001348, -0.00001324, -0.00001111, -0.00000771, -0.00000369, 0.00000028, 0.00000356, 0.00000557, 0.00000588, 0.00000443, 0.00000158, -0.00000185, -0.00000478, -0.00000619, -0.00000550, -0.00000278, 0.00000119, 0.00000521, 0.00000801, 0.00000872, 0.00000717, 0.00000389, -0.00000010, -0.00000368, -0.00000603, -0.00000684, -0.00000636, -0.00000511, -0.00000369, -0.00000246, -0.00000148, -0.00000056, },
		{ -0.00000759, -0.00000694, -0.00000562, -0.00000377, -0.00000157, 0.00000076, 0.00000297, 0.00000484, 0.00000623, 0.00000708, 0.00000737, 0.00000707, 0.00000609, 0.00000433, 0.00000174, -0.00000151, -0.00000498, -0.00000798, -0.00000972, -0.00000959, -0.00000741, -0.00000353, 0.00000118, 0.00000559, 0.00000866, 0.00000973, 0.00000868, 0.00000588, 0.00000197, -0.00000237, -0.00000664, -0.00001054, -0.00001384, -0.00001624, -0.00001728, -0.00001645, -0.00001341, -0.00000829, -0.00000186, 0.00000454, 0.00000938, 0.00001142, 0.00001025, 0.00000650, 0.00000165, -0.00000243, -0.00000424, -0.00000321, 0.00000007, 0.00000406, 0.00000692, 0.00000724, 0.00000469, 0.00000011, -0.00000474, -0.00000793, -0.00000814, -0.00000519, -0.00000016, 0.00000501, 0.00000826, 0.00000821, 0.00000470, -0.00000116, -0.00000734, -0.00001160, -0.00001233, -0.00000904, -0.00000259, 0.00000513, 0.00001179, 0.00001536, 0.00001474, 0.00001006, 0.00000265, -0.00000540, -0.00001190, -0.00001515, -0.00001448, -0.00001038, -0.00000437, 0.00000158, 0.00000560, 0.00000654, 0.00000433, -0.00000001, -0.00000478, -0.00000809, -0.00000855, -0.00000568, -0.00000003, 0.00000698, 0.00001358, 0.00001817, 0.00001979, 0.00001827, 0.00001420, 0.00000861, 0.00000263, -0.00000279, -0.00000703, -0.00000977, -0.00001089, -0.00001042, -0.00000851, -0.00000547, -0.00000181, 0.00000174, 0.00000438, 0.00000549, 0.00000479, 0.00000258, -0.00000039, -0.00000308, -0.00000453, -0.00000421, -0.00000215, 0.00000101, 0.00000430, 0.00000674, 0.00000765, 0.00000684, 0.00000459, 0.00000152, -0.00000168, -0.00000444, -0.00000640, -0.00000744, },
		{ 0.00000368, 0.00000043, -0.00000310, -0.00000629, -0.00000859, -0.00000963, -0.00000927, -0.00000768, -0.00000522, -0.00000241, 0.00000029, 0.00000260, 0.00000440, 0.00000578, 0.00000681, 0.00000746, 0.00000751, 0.00000664, 0.00000459, 0.00000138, -0.00000257, -0.00000641, -0.00000914, -0.00000990, -0.00000832, -0.00000471, 0.00000003, 0.00000467, 0.00000807, 0.00000952, 0.00000893, 0.00000672, 0.00000358, 0.00000012, -0.00000333, -0.00000673, -0.00001012, -0.00001336, -0.00001599, -0.00001723, -0.00001628, -0.00001274, -0.00000693, 0.00000005, 0.00000653, 0.00001082, 0.00001183, 0.00000951, 0.00000491, -0.00000020, -0.00000394, -0.00000504, -0.00000331, 0.00000034, 0.00000426, 0.00000676, 0.00000675, 0.00000421, 0.00000012, -0.00000389, -0.00000625, -0.00000600, -0.00000323, 0.00000100, 0.00000505, 0.00000732, 0.00000681, 0.00000354, -0.00000152, -0.00000671, -0.00001027, -0.00001091, -0.00000824, -0.00000290, 0.00000364, 0.00000948, 0.00001290, 0.00001289, 0.00000939, 0.00000341, -0.00000337, -0.00000905, -0.00001210, -0.00001188, -0.00000872, -0.00000389, 0.00000093, 0.00000418, 0.00000496, 0.00000328, 0.00000005, -0.00000331, -0.00000538, -0.00000523, -0.00000271, 0.00000156, 0.00000644, 0.00001071, 0.00001346, 0.00001429, 0.00001329, 0.00001094, 0.00000777, 0.00000426, 0.00000072, -0.00000263, -0.00000553, -0.00000767, -0.00000867, -0.00000824, -0.00000634, -0.00000333, 0.00000010, 0.00000308, 0.00000484, 0.00000497, 0.00000356, 0.00000119, -0.00000126, -0.00000296, -0.00000332, -0.00000222, 0.00000004, 0.00000280, 0.00000532, 0.00000694, 0.00000724, 0.00000609, },
		{ 0.00000136, 0.00000443, 0.00000699, 0.00000829, 0.00000786, 0.00000563, 0.00000202, -0.00000224, -0.00000623, -0.00000913, -0.00001046, -0.00001015, -0.00000854, -0.00000619, -0.00000367, -0.00000130, 0.00000082, 0.00000282, 0.00000478, 0.00000661, 0.00000794, 0.00000821, 0.00000693, 0.00000398, -0.00000019, -0.00000463, -0.00000807, -0.00000945, -0.00000824, -0.00000472, 0.00000008, 0.00000478, 0.00000811, 0.00000934, 0.00000848, 0.00000616, 0.00000328, 0.00000056, -0.00000172, -0.00000378, -0.00000608, -0.00000891, -0.00001210, -0.00001490, -0.00001627, -0.00001523, -0.00001143, -0.00000542, 0.00000146, 0.00000742, 0.00001089, 0.00001105, 0.00000819, 0.00000353, -0.00000117, -0.00000431, -0.00000495, -0.00000316, 0.00000011, 0.00000343, 0.00000550, 0.00000558, 0.00000374, 0.00000082, -0.00000198, -0.00000358, -0.00000337, -0.00000145, 0.00000140, 0.00000408, 0.00000549, 0.00000497, 0.00000249, -0.00000130, -0.00000525, -0.00000811, -0.00000887, -0.00000712, -0.00000322, 0.00000181, 0.00000655, 0.00000961, 0.00001008, 0.00000782, 0.00000352, -0.00000153, -0.00000585, -0.00000826, -0.00000826, -0.00000618, -0.00000299, 0.00000009, 0.00000203, 0.00000238, 0.00000136, -0.00000026, -0.00000153, -0.00000171, -0.00000056, 0.00000164, 0.00000423, 0.00000649, 0.00000792, 0.00000838, 0.00000808, 0.00000739, 0.00000660, 0.00000575, 0.00000469, 0.00000315, 0.00000099, -0.00000165, -0.00000430, -0.00000630, -0.00000704, -0.00000621, -0.00000395, -0.00000085, 0.00000225, 0.00000448, 0.00000528, 0.00000452, 0.00000254, 0.00000004, -0.00000216, -0.00000333, -0.00000308, -0.00000138, },
		{ 0.00000034, -0.00000249, -0.00000430, -0.00000441, -0.00000266, 0.00000055, 0.00000436, 0.00000766, 0.00000946, 0.00000915, 0.00000675, 0.00000284, -0.00000162, -0.00000561, -0.00000836, -0.00000961, -0.00000954, -0.00000859, -0.00000724, -0.00000574, -0.00000402, -0.00000188, 0.00000082, 0.00000390, 0.00000675, 0.00000852, 0.00000840, 0.00000605, 0.00000187, -0.00000303, -0.00000712, -0.00000901, -0.00000797, -0.00000425, 0.00000101, 0.00000615, 0.00000963, 0.00001054, 0.00000895, 0.00000571, 0.00000210, -0.00000079, -0.00000249, -0.00000325, -0.00000384, -0.00000507, -0.00000732, -0.00001027, -0.00001298, -0.00001424, -0.00001314, -0.00000947, -0.00000394, 0.00000208, 0.00000699, 0.00000956, 0.00000929, 0.00000662, 0.00000267, -0.00000115, -0.00000365, -0.00000425, -0.00000304, -0.00000072, 0.00000177, 0.00000356, 0.00000415, 0.00000354, 0.00000213, 0.00000056, -0.00000056, -0.00000081, -0.00000013, 0.00000117, 0.00000256, 0.00000339, 0.00000318, 0.00000175, -0.00000064, -0.00000338, -0.00000562, -0.00000659, -0.00000584, -0.00000345, -0.00000005, 0.00000339, 0.00000585, 0.00000661, 0.00000552, 0.00000304, 0.00000005, -0.00000251, -0.00000394, -0.00000405, -0.00000316, -0.00000192, -0.00000093, -0.00000055, -0.00000067, -0.00000086, -0.00000057, 0.00000054, 0.00000241, 0.00000461, 0.00000643, 0.00000729, 0.00000691, 0.00000553, 0.00000373, 0.00000224, 0.00000162, 0.00000197, 0.00000296, 0.00000393, 0.00000416, 0.00000324, 0.00000120, -0.00000144, -0.00000387, -0.00000530, -0.00000519, -0.00000351, -0.00000072, 0.00000237, 0.00000483, 0.00000593, 0.00000534, 0.00000326, },
		{ 0.00000365, 0.00000598, 0.00000682, 0.00000576, 0.00000305, -0.00000052, -0.00000382, -0.00000574, -0.00000557, -0.00000324, 0.00000063, 0.00000491, 0.00000833, 0.00000992, 0.00000930, 0.00000676, 0.00000312, -0.00000066, -0.00000378, -0.00000592, -0.00000717, -0.00000791, -0.00000845, -0.00000879, -0.00000859, -0.00000733, -0.00000468, -0.00000081, 0.00000350, 0.00000702, 0.00000850, 0.00000722, 0.00000339, -0.00000180, -0.00000655, -0.00000908, -0.00000830, -0.00000426, 0.00000181, 0.00000799, 0.00001230, 0.00001349, 0.00001142, 0.00000703, 0.00000194, -0.00000228, -0.00000465, -0.00000514, -0.00000449, -0.00000384, -0.00000413, -0.00000567, -0.00000801, -0.00001017, -0.00001108, -0.00001004, -0.00000704, -0.00000273, 0.00000176, 0.00000530, 0.00000709, 0.00000690, 0.00000509, 0.00000239, -0.00000033, -0.00000237, -0.00000331, -0.00000310, -0.00000195, -0.00000027, 0.00000149, 0.00000290, 0.00000367, 0.00000371, 0.00000312, 0.00000217, 0.00000124, 0.00000068, 0.00000064, 0.00000104, 0.00000158, 0.00000181, 0.00000139, 0.00000021, -0.00000150, -0.00000323, -0.00000438, -0.00000452, -0.00000354, -0.00000175, 0.00000029, 0.00000196, 0.00000283, 0.00000283, 0.00000220, 0.00000138, 0.00000073, 0.00000037, 0.00000016, -0.00000020, -0.00000095, -0.00000203, -0.00000309, -0.00000353, -0.00000277, -0.00000059, 0.00000273, 0.00000641, 0.00000941, 0.00001080, 0.00001009, 0.00000749, 0.00000378, 0.00000012, -0.00000243, -0.00000322, -0.00000224, -0.00000011, 0.00000222, 0.00000377, 0.00000393, 0.00000265, 0.00000041, -0.00000193, -0.00000346, -0.00000354, -0.00000204, 0.00000064, },
		{ -0.00000081, -0.00000184, -0.00000151, 0.00000029, 0.00000304, 0.00000583, 0.00000759, 0.00000754, 0.00000547, 0.00000187, -0.00000220, -0.00000545, -0.00000681, -0.00000581, -0.00000274, 0.00000143, 0.00000543, 0.00000813, 0.00000896, 0.00000799, 0.00000586, 0.00000340, 0.00000124, -0.00000045, -0.00000194, -0.00000368, -0.00000597, -0.00000859, -0.00001082, -0.00001165, -0.00001032, -0.00000670, -0.00000160, 0.00000346, 0.00000675, 0.00000701, 0.00000406, -0.00000106, -0.00000638, -0.00000972, -0.00000952, -0.00000544, 0.00000143, 0.00000898, 0.00001485, 0.00001726, 0.00001564, 0.00001074, 0.00000425, -0.00000184, -0.00000603, -0.00000769, -0.00000721, -0.00000564, -0.00000416, -0.00000358, -0.00000412, -0.00000534, -0.00000652, -0.00000693, -0.00000621, -0.00000441, -0.00000195, 0.00000060, 0.00000269, 0.00000397, 0.00000431, 0.00000376, 0.00000251, 0.00000083, -0.00000097, -0.00000248, -0.00000333, -0.00000322, -0.00000210, -0.00000017, 0.00000207, 0.00000401, 0.00000511, 0.00000510, 0.00000410, 0.00000257, 0.00000109, 0.00000016, -0.00000000, 0.00000043, 0.00000103, 0.00000134, 0.00000103, 0.00000010, -0.00000120, -0.00000245, -0.00000328, -0.00000352, -0.00000321, -0.00000254, -0.00000172, -0.00000083, 0.00000015, 0.00000127, 0.00000249, 0.00000359, 0.00000415, 0.00000377, 0.00000224, -0.00000024, -0.00000301, -0.00000510, -0.00000559, -0.00000391, -0.00000017, 0.00000482, 0.00000973, 0.00001316, 0.00001409, 0.00001222, 0.00000805, 0.00000277, -0.00000217, -0.00000553, -0.00000663, -0.00000553, -0.00000296, 0.00000000, 0.00000228, 0.00000317, 0.00000258, 0.00000097, },
		{ -0.00000225, 0.00000012, 0.00000128, 0.00000118, 0.00000030, -0.00000055, -0.00000055, 0.00000072, 0.00000313, 0.00000598, 0.00000823, 0.00000891, 0.00000752, 0.00000426, 0.00000001, -0.00000394, -0.00000637, -0.00000656, -0.00000457, -0.00000119, 0.00000240, 0.00000512, 0.00000637, 0.00000625, 0.00000539, 0.00000456, 0.00000424, 0.00000432, 0.00000415, 0.00000288, -0.00000001, -0.00000435, -0.00000912, -0.00001281, -0.00001401, -0.00001204, -0.00000739, -0.00000164, 0.00000313, 0.00000508, 0.00000346, -0.00000108, -0.00000663, -0.00001076, -0.00001141, -0.00000777, -0.00000057, 0.00000817, 0.00001585, 0.00002021, 0.00002014, 0.00001591, 0.00000901, 0.00000151, -0.00000471, -0.00000850, -0.00000966, -0.00000880, -0.00000692, -0.00000494, -0.00000347, -0.00000264, -0.00000231, -0.00000223, -0.00000218, -0.00000209, -0.00000191, -0.00000161, -0.00000109, -0.00000029, 0.00000077, 0.00000186, 0.00000263, 0.00000268, 0.00000179, 0.00000009, -0.00000193, -0.00000355, -0.00000410, -0.00000323, -0.00000109, 0.00000171, 0.00000431, 0.00000591, 0.00000612, 0.00000503, 0.00000318, 0.00000131, 0.00000004, -0.00000035, 0.00000004, 0.00000082, 0.00000150, 0.00000170, 0.00000128, 0.00000034, -0.00000091, -0.00000225, -0.00000347, -0.00000443, -0.00000496, -0.00000487, -0.00000395, -0.00000212, 0.00000046, 0.00000333, 0.00000575, 0.00000692, 0.00000631, 0.00000387, 0.00000020, -0.00000359, -0.00000617, -0.00000647, -0.00000405, 0.00000068, 0.00000657, 0.00001201, 0.00001548, 0.00001601, 0.00001343, 0.00000846, 0.00000243, -0.00000314, -0.00000698, -0.00000842, -0.00000757, -0.00000515, },
		{ -0.00000303, -0.00000696, -0.00000865, -0.00000824, -0.00000641, -0.00000413, -0.00000226, -0.00000128, -0.00000110, -0.00000119, -0.00000082, 0.00000054, 0.00000297, 0.00000597, 0.00000859, 0.00000982, 0.00000903, 0.00000622, 0.00000216, -0.00000193, -0.00000478, -0.00000559, -0.00000434, -0.00000176, 0.00000098, 0.00000285, 0.00000337, 0.00000282, 0.00000209, 0.00000214, 0.00000352, 0.00000592, 0.00000829, 0.00000919, 0.00000750, 0.00000301, -0.00000330, -0.00000961, -0.00001386, -0.00001465, -0.00001186, -0.00000674, -0.00000148, 0.00000168, 0.00000139, -0.00000222, -0.00000756, -0.00001216, -0.00001369, -0.00001083, -0.00000385, 0.00000550, 0.00001463, 0.00002103, 0.00002310, 0.00002062, 0.00001464, 0.00000700, -0.00000038, -0.00000609, -0.00000948, -0.00001063, -0.00001006, -0.00000838, -0.00000611, -0.00000360, -0.00000115, 0.00000094, 0.00000229, 0.00000260, 0.00000181, 0.00000021, -0.00000158, -0.00000284, -0.00000303, -0.00000202, -0.00000024, 0.00000154, 0.00000250, 0.00000214, 0.00000055, -0.00000166, -0.00000356, -0.00000429, -0.00000344, -0.00000121, 0.00000169, 0.00000433, 0.00000592, 0.00000610, 0.00000504, 0.00000329, 0.00000155, 0.00000037, -0.00000001, 0.00000031, 0.00000103, 0.00000175, 0.00000216, 0.00000206, 0.00000140, 0.00000018, -0.00000151, -0.00000347, -0.00000539, -0.00000681, -0.00000722, -0.00000623, -0.00000374, -0.00000014, 0.00000378, 0.00000696, 0.00000841, 0.00000760, 0.00000468, 0.00000051, -0.00000356, -0.00000610, -0.00000610, -0.00000327, 0.00000179, 0.00000778, 0.00001310, 0.00001630, 0.00001650, 0.00001368, 0.00000859, 0.00000253, },
		{ 0.00001576, 0.00001572, 0.00001300, 0.00000832, 0.00000285, -0.00000217, -0.00000577, -0.00000753, -0.00000762, -0.00000667, -0.00000543, -0.00000449, -0.00000398, -0.00000361, -0.00000281, -0.00000109, 0.00000167, 0.00000504, 0.00000812, 0.00000987, 0.00000959, 0.00000720, 0.00000344, -0.00000046, -0.00000316, -0.00000384, -0.00000253, -0.00000007, 0.00000220, 0.00000312, 0.00000225, 0.00000009, -0.00000205, -0.00000270, -0.00000092, 0.00000316, 0.00000823, 0.00001230, 0.00001349, 0.00001086, 0.00000489, -0.00000263, -0.00000929, -0.00001302, -0.00001294, -0.00000971, -0.00000527, -0.00000196, -0.00000155, -0.00000440, -0.00000936, -0.00001415, -0.00001632, -0.00001426, -0.00000783, 0.00000156, 0.00001150, 0.00001943, 0.00002352, 0.00002317, 0.00001902, 0.00001255, 0.00000544, -0.00000095, -0.00000589, -0.00000916, -0.00001085, -0.00001107, -0.00000987, -0.00000732, -0.00000368, 0.00000044, 0.00000415, 0.00000649, 0.00000680, 0.00000504, 0.00000186, -0.00000161, -0.00000417, -0.00000503, -0.00000408, -0.00000195, 0.00000031, 0.00000171, 0.00000169, 0.00000037, -0.00000158, -0.00000319, -0.00000369, -0.00000279, -0.00000074, 0.00000177, 0.00000394, 0.00000517, 0.00000527, 0.00000445, 0.00000316, 0.00000192, 0.00000109, 0.00000080, 0.00000100, 0.00000149, 0.00000204, 0.00000245, 0.00000251, 0.00000201, 0.00000081, -0.00000111, -0.00000355, -0.00000606, -0.00000798, -0.00000863, -0.00000754, -0.00000469, -0.00000062, 0.00000369, 0.00000708, 0.00000857, 0.00000774, 0.00000486, 0.00000090, -0.00000281, -0.00000495, -0.00000467, -0.00000182, 0.00000294, 0.00000840, 0.00001309, },
		{ -0.00000252, -0.00000001, 0.00000399, 0.00000844, 0.00001215, 0.00001414, 0.00001388, 0.00001148, 0.00000758, 0.00000315, -0.00000085, -0.00000375, -0.00000538, -0.00000599, -0.00000608, -0.00000613, -0.00000626, -0.00000625, -0.00000558, -0.00000380, -0.00000079, 0.00000299, 0.00000657, 0.00000883, 0.00000894, 0.00000680, 0.00000316, -0.00000058, -0.00000295, -0.00000298, -0.00000075, 0.00000268, 0.00000561, 0.00000648, 0.00000458, 0.00000047, -0.00000417, -0.00000721, -0.00000697, -0.00000303, 0.00000353, 0.00001044, 0.00001517, 0.00001590, 0.00001228, 0.00000562, -0.00000174, -0.00000736, -0.00000973, -0.00000885, -0.00000616, -0.00000385, -0.00000378, -0.00000666, -0.00001166, -0.00001673, -0.00001945, -0.00001808, -0.00001222, -0.00000302, 0.00000725, 0.00001610, 0.00002160, 0.00002301, 0.00002076, 0.00001608, 0.00001039, 0.00000479, -0.00000021, -0.00000447, -0.00000799, -0.00001057, -0.00001176, -0.00001103, -0.00000813, -0.00000341, 0.00000211, 0.00000699, 0.00000987, 0.00000996, 0.00000736, 0.00000307, -0.00000144, -0.00000476, -0.00000606, -0.00000534, -0.00000334, -0.00000114, 0.00000031, 0.00000054, -0.00000028, -0.00000152, -0.00000245, -0.00000252, -0.00000161, -0.00000001, 0.00000176, 0.00000318, 0.00000395, 0.00000404, 0.00000363, 0.00000301, 0.00000243, 0.00000201, 0.00000181, 0.00000183, 0.00000202, 0.00000233, 0.00000262, 0.00000268, 0.00000224, 0.00000104, -0.00000099, -0.00000362, -0.00000633, -0.00000838, -0.00000906, -0.00000796, -0.00000512, -0.00000115, 0.00000297, 0.00000618, 0.00000763, 0.00000703, 0.00000470, 0.00000150, -0.00000141, -0.00000297, },
		{ 0.00000587, 0.00000445, 0.00000236, 0.00000044, -0.00000052, -0.00000003, 0.00000191, 0.00000486, 0.00000803, 0.00001054, 0.00001169, 0.00001120, 0.00000924, 0.00000639, 0.00000335, 0.00000072, -0.00000123, -0.00000261, -0.00000375, -0.00000495, -0.00000626, -0.00000732, -0.00000752, -0.00000630, -0.00000349, 0.00000044, 0.00000442, 0.00000711, 0.00000748, 0.00000531, 0.00000138, -0.00000271, -0.00000512, -0.00000461, -0.00000108, 0.00000423, 0.00000918, 0.00001157, 0.00001010, 0.00000498, -0.00000205, -0.00000832, -0.00001131, -0.00000967, -0.00000382, 0.00000423, 0.00001162, 0.00001585, 0.00001569, 0.00001166, 0.00000564, 0.00000003, -0.00000337, -0.00000407, -0.00000306, -0.00000221, -0.00000332, -0.00000717, -0.00001308, -0.00001910, -0.00002280, -0.00002229, -0.00001702, -0.00000804, 0.00000245, 0.00001188, 0.00001826, 0.00002079, 0.00001986, 0.00001671, 0.00001270, 0.00000873, 0.00000507, 0.00000145, -0.00000241, -0.00000643, -0.00000994, -0.00001193, -0.00001144, -0.00000815, -0.00000262, 0.00000372, 0.00000913, 0.00001206, 0.00001181, 0.00000872, 0.00000396, -0.00000090, -0.00000449, -0.00000612, -0.00000588, -0.00000445, -0.00000275, -0.00000150, -0.00000101, -0.00000113, -0.00000144, -0.00000153, -0.00000116, -0.00000036, 0.00000063, 0.00000156, 0.00000224, 0.00000264, 0.00000284, 0.00000293, 0.00000298, 0.00000298, 0.00000291, 0.00000276, 0.00000259, 0.00000250, 0.00000255, 0.00000266, 0.00000263, 0.00000215, 0.00000095, -0.00000102, -0.00000355, -0.00000610, -0.00000801, -0.00000867, -0.00000773, -0.00000531, -0.00000191, 0.00000163, 0.00000447, 0.00000596, },
		{ -0.00000548, -0.00000295, -0.00000015, 0.00000231, 0.00000396, 0.00000457, 0.00000425, 0.00000336, 0.00000247, 0.00000209, 0.00000253, 0.00000380, 0.00000556, 0.00000729, 0.00000842, 0.00000861, 0.00000785, 0.00000644, 0.00000483, 0.00000343, 0.00000237, 0.00000146, 0.00000032, -0.00000136, -0.00000357, -0.00000585, -0.00000737, -0.00000730, -0.00000526, -0.00000161, 0.00000252, 0.00000556, 0.00000614, 0.00000376, -0.00000089, -0.00000602, -0.00000939, -0.00000919, -0.00000492, 0.00000232, 0.00001010, 0.00001554, 0.00001645, 0.00001229, 0.00000439, -0.00000445, -0.00001107, -0.00001317, -0.00001017, -0.00000338, 0.00000457, 0.00001090, 0.00001375, 0.00001285, 0.00000947, 0.00000564, 0.00000310, 0.00000250, 0.00000309, 0.00000318, 0.00000103, -0.00000416, -0.00001172, -0.00001959, -0.00002516, -0.00002624, -0.00002205, -0.00001348, -0.00000280, 0.00000721, 0.00001434, 0.00001758, 0.00001734, 0.00001498, 0.00001205, 0.00000960, 0.00000776, 0.00000596, 0.00000340, -0.00000032, -0.00000482, -0.00000896, -0.00001131, -0.00001080, -0.00000719, -0.00000135, 0.00000511, 0.00001034, 0.00001293, 0.00001240, 0.00000924, 0.00000464, 0.00000001, -0.00000352, -0.00000545, -0.00000587, -0.00000530, -0.00000431, -0.00000336, -0.00000260, -0.00000199, -0.00000139, -0.00000072, -0.00000004, 0.00000056, 0.00000098, 0.00000122, 0.00000138, 0.00000160, 0.00000199, 0.00000253, 0.00000309, 0.00000350, 0.00000363, 0.00000348, 0.00000317, 0.00000286, 0.00000266, 0.00000255, 0.00000236, 0.00000182, 0.00000071, -0.00000101, -0.00000317, -0.00000534, -0.00000700, -0.00000770, -0.00000718, },
		{ -0.00000242, -0.00000413, -0.00000560, -0.00000650, -0.00000658, -0.00000576, -0.00000417, -0.00000209, 0.00000009, 0.00000197, 0.00000332, 0.00000405, 0.00000433, 0.00000442, 0.00000461, 0.00000502, 0.00000561, 0.00000613, 0.00000629, 0.00000592, 0.00000507, 0.00000405, 0.00000329, 0.00000308, 0.00000342, 0.00000393, 0.00000398, 0.00000299, 0.00000081, -0.00000212, -0.00000481, -0.00000612, -0.00000529, -0.00000236, 0.00000162, 0.00000492, 0.00000581, 0.00000333, -0.00000214, -0.00000881, -0.00001408, -0.00001552, -0.00001188, -0.00000378, 0.00000644, 0.00001545, 0.00002021, 0.00001914, 0.00001272, 0.00000333, -0.00000569, -0.00001137, -0.00001217, -0.00000847, -0.00000226, 0.00000389, 0.00000791, 0.00000906, 0.00000810, 0.00000664, 0.00000622, 0.00000739, 0.00000936, 0.00001035, 0.00000847, 0.00000272, -0.00000629, -0.00001642, -0.00002468, -0.00002839, -0.00002626, -0.00001885, -0.00000838, 0.00000217, 0.00001013, 0.00001411, 0.00001431, 0.00001214, 0.00000946, 0.00000767, 0.00000717, 0.00000732, 0.00000694, 0.00000503, 0.00000137, -0.00000330, -0.00000751, -0.00000973, -0.00000899, -0.00000531, 0.00000026, 0.00000611, 0.00001059, 0.00001261, 0.00001193, 0.00000910, 0.00000515, 0.00000114, -0.00000213, -0.00000432, -0.00000546, -0.00000578, -0.00000553, -0.00000488, -0.00000393, -0.00000276, -0.00000150, -0.00000034, 0.00000050, 0.00000093, 0.00000099, 0.00000086, 0.00000081, 0.00000105, 0.00000164, 0.00000246, 0.00000329, 0.00000389, 0.00000411, 0.00000394, 0.00000352, 0.00000303, 0.00000260, 0.00000226, 0.00000191, 0.00000137, 0.00000049, -0.00000081, },
		{ 0.00000182, 0.00000137, 0.00000094, 0.00000041, -0.00000035, -0.00000141, -0.00000272, -0.00000411, -0.00000531, -0.00000604, -0.00000606, -0.00000530, -0.00000383, -0.00000191, 0.00000016, 0.00000209, 0.00000374, 0.00000507, 0.00000612, 0.00000692, 0.00000738, 0.00000731, 0.00000655, 0.00000507, 0.00000312, 0.00000124, 0.00000006, 0.00000006, 0.00000132, 0.00000334, 0.00000522, 0.00000598, 0.00000505, 0.00000257, -0.00000053, -0.00000291, -0.00000337, -0.00000149, 0.00000201, 0.00000545, 0.00000680, 0.00000460, -0.00000130, -0.00000941, -0.00001703, -0.00002117, -0.00001979, -0.00001264, -0.00000158, 0.00001011, 0.00001881, 0.00002200, 0.00001909, 0.00001162, 0.00000255, -0.00000497, -0.00000879, -0.00000847, -0.00000523, -0.00000117, 0.00000182, 0.00000295, 0.00000276, 0.00000268, 0.00000415, 0.00000764, 0.00001222, 0.00001583, 0.00001613, 0.00001163, 0.00000256, -0.00000907, -0.00002000, -0.00002699, -0.00002798, -0.00002287, -0.00001351, -0.00000297, 0.00000573, 0.00001062, 0.00001140, 0.00000929, 0.00000631, 0.00000428, 0.00000409, 0.00000540, 0.00000700, 0.00000747, 0.00000593, 0.00000248, -0.00000184, -0.00000551, -0.00000717, -0.00000616, -0.00000275, 0.00000198, 0.00000665, 0.00001001, 0.00001137, 0.00001070, 0.00000849, 0.00000545, 0.00000226, -0.00000062, -0.00000296, -0.00000468, -0.00000575, -0.00000614, -0.00000583, -0.00000489, -0.00000349, -0.00000195, -0.00000058, 0.00000035, 0.00000076, 0.00000078, 0.00000065, 0.00000066, 0.00000100, 0.00000170, 0.00000261, 0.00000350, 0.00000414, 0.00000437, 0.00000418, 0.00000367, 0.00000302, 0.00000237, },
		},
		{ { -0.00000693, -0.00000683, -0.00000658, -0.00000627, -0.00000601, -0.00000588, -0.00000586, -0.00000583, -0.00000559, -0.00000490, -0.00000356, -0.00000150, 0.00000117, 0.00000418, 0.00000712, 0.00000951, 0.00001095, 0.00001119, 0.00001021, 0.00000825, 0.00000574, 0.00000319, 0.00000110, -0.00000019, -0.00000058, -0.00000021, 0.00000055, 0.00000122, 0.00000134, 0.00000062, -0.00000102, -0.00000339, -0.00000610, -0.00000863, -0.00001046, -0.00001123, -0.00001078, -0.00000920, -0.00000679, -0.00000395, -0.00000110, 0.00000143, 0.00000349, 0.00000509, 0.00000633, 0.00000740, 0.00000839, 0.00000928, 0.00000990, 0.00000997, 0.00000918, 0.00000733, 0.00000437, 0.00000056, -0.00000362, -0.00000750, -0.00001038, -0.00001169, -0.00001113, -0.00000877, -0.00000509, -0.00000082, 0.00000312, 0.00000588, 0.00000688, 0.00000588, 0.00000312, -0.00000082, -0.00000509, -0.00000877, -0.00001113, -0.00001169, -0.00001038, -0.00000750, -0.00000362, 0.00000056, 0.00000437, 0.00000733, 0.00000918, 0.00000997, 0.00000990, 0.00000928, 0.00000839, 0.00000740, 0.00000633, 0.00000509, 0.00000349, 0.00000143, -0.00000110, -0.00000395, -0.00000679, -0.00000920, -0.00001078, -0.00001123, -0.00001046, -0.00000863, -0.00000610, -0.00000339, -0.00000102, 0.00000062, 0.00000134, 0.00000122, 0.00000055, -0.00000021, -0.00000058, -0.00000019, 0.00000110, 0.00000319, 0.00000574, 0.00000825, 0.00001021, 0.00001119, 0.00001095, 0.00000951, 0.00000712, 0.00000418, 0.00000117, -0.00000150, -0.00000356, -0.00000490, -0.00000559, -0.00000583, -0.00000586, -0.00000588, -0.00000601, -0.00000627, -0.00000658, -0.00000683, },
		{ -0.00000421, -0.00000473, -0.00000528, -0.00000580, -0.00000620, -0.00000639, -0.00000633, -0.00000607, -0.00000572, -0.00000546, -0.00000543, -0.00000570, -0.00000623, -0.00000683, -0.00000722, -0.00000712, -0.00000629, -0.00000464, -0.00000226, 0.00000057, 0.00000344, 0.00000590, 0.00000759, 0.00000829, 0.00000803, 0.00000706, 0.00000575, 0.00000456, 0.00000384, 0.00000379, 0.00000435, 0.00000525, 0.00000604, 0.00000626, 0.00000552, 0.00000365, 0.00000074, -0.00000285, -0.00000658, -0.00000984, -0.00001209, -0.00001301, -0.00001250, -0.00001073, -0.00000808, -0.00000502, -0.00000199, 0.00000070, 0.00000290, 0.00000465, 0.00000608, 0.00000733, 0.00000848, 0.00000946, 0.00001006, 0.00001000, 0.00000898, 0.00000684, 0.00000363, -0.00000033, -0.00000448, -0.00000814, -0.00001062, -0.00001143, -0.00001037, -0.00000763, -0.00000377, 0.00000041, 0.00000400, 0.00000625, 0.00000665, 0.00000511, 0.00000198, -0.00000207, -0.00000619, -0.00000950, -0.00001134, -0.00001138, -0.00000966, -0.00000654, -0.00000265, 0.00000134, 0.00000482, 0.00000741, 0.00000894, 0.00000950, 0.00000934, 0.00000874, 0.00000795, 0.00000707, 0.00000609, 0.00000488, 0.00000329, 0.00000127, -0.00000115, -0.00000374, -0.00000616, -0.00000803, -0.00000899, -0.00000887, -0.00000772, -0.00000580, -0.00000356, -0.00000152, -0.00000013, 0.00000031, -0.00000022, -0.00000148, -0.00000303, -0.00000433, -0.00000486, -0.00000430, -0.00000257, 0.00000013, 0.00000337, 0.00000658, 0.00000921, 0.00001081, 0.00001117, 0.00001031, 0.00000849, 0.00000609, 0.00000354, 0.00000120, -0.00000070, -0.00000208, -0.00000302, -0.00000368, },
		{ 0.00000393, 0.00000225, 0.00000086, -0.00000022, -0.00000109, -0.00000189, -0.00000270, -0.00000357, -0.00000441, -0.00000512, -0.00000554, -0.00000560, -0.00000528, -0.00000470, -0.00000406, -0.00000359, -0.00000349, -0.00000384, -0.00000460, -0.00000557, -0.00000645, -0.00000693, -0.00000678, -0.00000589, -0.00000436, -0.00000243, -0.00000045, 0.00000120, 0.00000228, 0.00000269, 0.00000255, 0.00000215, 0.00000188, 0.00000211, 0.00000308, 0.00000478, 0.00000698, 0.00000922, 0.00001095, 0.00001163, 0.00001088, 0.00000861, 0.00000503, 0.00000060, -0.00000400, -0.00000808, -0.00001106, -0.00001257, -0.00001254, -0.00001115, -0.00000881, -0.00000597, -0.00000308, -0.00000044, 0.00000180, 0.00000368, 0.00000528, 0.00000670, 0.00000796, 0.00000896, 0.00000945, 0.00000918, 0.00000789, 0.00000551, 0.00000219, -0.00000169, -0.00000553, -0.00000867, -0.00001049, -0.00001062, -0.00000898, -0.00000589, -0.00000199, 0.00000191, 0.00000497, 0.00000651, 0.00000620, 0.00000411, 0.00000067, -0.00000339, -0.00000721, -0.00001003, -0.00001130, -0.00001082, -0.00000875, -0.00000553, -0.00000176, 0.00000191, 0.00000500, 0.00000721, 0.00000846, 0.00000890, 0.00000876, 0.00000828, 0.00000763, 0.00000685, 0.00000585, 0.00000451, 0.00000273, 0.00000051, -0.00000199, -0.00000449, -0.00000658, -0.00000789, -0.00000814, -0.00000727, -0.00000546, -0.00000310, -0.00000072, 0.00000111, 0.00000198, 0.00000168, 0.00000030, -0.00000180, -0.00000407, -0.00000592, -0.00000682, -0.00000647, -0.00000484, -0.00000218, 0.00000103, 0.00000424, 0.00000690, 0.00000861, 0.00000922, 0.00000876, 0.00000750, 0.00000577, },
		{ 0.00000516, 0.00000646, 0.00000684, 0.00000640, 0.00000540, 0.00000417, 0.00000296, 0.00000195, 0.00000115, 0.00000049, -0.00000017, -0.00000097, -0.00000196, -0.00000308, -0.00000420, -0.00000510, -0.00000559, -0.00000557, -0.00000504, -0.00000414, -0.00000313, -0.00000225, -0.00000175, -0.00000174, -0.00000218, -0.00000290, -0.00000364, -0.00000414, -0.00000422, -0.00000384, -0.00000310, -0.00000227, -0.00000162, -0.00000143, -0.00000180, -0.00000265, -0.00000372, -0.00000458, -0.00000478, -0.00000394, -0.00000191, 0.00000120, 0.00000500, 0.00000887, 0.00001211, 0.00001408, 0.00001435, 0.00001280, 0.00000963, 0.00000537, 0.00000070, -0.00000365, -0.00000707, -0.00000916, -0.00000981, -0.00000917, -0.00000757, -0.00000541, -0.00000307, -0.00000082, 0.00000121, 0.00000301, 0.00000461, 0.00000604, 0.00000723, 0.00000804, 0.00000822, 0.00000753, 0.00000583, 0.00000315, -0.00000027, -0.00000396, -0.00000731, -0.00000969, -0.00001061, -0.00000983, -0.00000745, -0.00000392, 0.00000007, 0.00000368, 0.00000616, 0.00000699, 0.00000599, 0.00000337, -0.00000032, -0.00000432, -0.00000783, -0.00001017, -0.00001096, -0.00001012, -0.00000789, -0.00000476, -0.00000131, 0.00000193, 0.00000457, 0.00000643, 0.00000753, 0.00000802, 0.00000811, 0.00000796, 0.00000762, 0.00000703, 0.00000606, 0.00000456, 0.00000248, -0.00000009, -0.00000288, -0.00000550, -0.00000749, -0.00000848, -0.00000823, -0.00000676, -0.00000438, -0.00000158, 0.00000105, 0.00000291, 0.00000362, 0.00000304, 0.00000134, -0.00000105, -0.00000354, -0.00000553, -0.00000653, -0.00000629, -0.00000485, -0.00000250, 0.00000030, 0.00000300, },
		{ -0.00000521, -0.00000469, -0.00000317, -0.00000099, 0.00000137, 0.00000344, 0.00000486, 0.00000546, 0.00000528, 0.00000452, 0.00000349, 0.00000248, 0.00000170, 0.00000121, 0.00000093, 0.00000069, 0.00000028, -0.00000044, -0.00000152, -0.00000285, -0.00000422, -0.00000539, -0.00000611, -0.00000625, -0.00000581, -0.00000491, -0.00000379, -0.00000272, -0.00000189, -0.00000141, -0.00000124, -0.00000124, -0.00000120, -0.00000096, -0.00000044, 0.00000029, 0.00000101, 0.00000141, 0.00000119, 0.00000014, -0.00000175, -0.00000426, -0.00000694, -0.00000923, -0.00001054, -0.00001042, -0.00000867, -0.00000541, -0.00000108, 0.00000367, 0.00000805, 0.00001135, 0.00001307, 0.00001300, 0.00001128, 0.00000832, 0.00000473, 0.00000115, -0.00000185, -0.00000390, -0.00000486, -0.00000480, -0.00000395, -0.00000260, -0.00000100, 0.00000064, 0.00000224, 0.00000374, 0.00000514, 0.00000635, 0.00000724, 0.00000760, 0.00000718, 0.00000581, 0.00000345, 0.00000024, -0.00000343, -0.00000702, -0.00000991, -0.00001154, -0.00001153, -0.00000980, -0.00000662, -0.00000258, 0.00000158, 0.00000504, 0.00000713, 0.00000746, 0.00000602, 0.00000313, -0.00000058, -0.00000435, -0.00000746, -0.00000937, -0.00000981, -0.00000884, -0.00000677, -0.00000407, -0.00000124, 0.00000131, 0.00000334, 0.00000480, 0.00000577, 0.00000644, 0.00000694, 0.00000732, 0.00000752, 0.00000735, 0.00000659, 0.00000509, 0.00000282, -0.00000003, -0.00000311, -0.00000592, -0.00000798, -0.00000887, -0.00000842, -0.00000671, -0.00000412, -0.00000119, 0.00000144, 0.00000321, 0.00000380, 0.00000313, 0.00000144, -0.00000080, -0.00000299, -0.00000459, },
		{ 0.00000264, 0.00000172, 0.00000004, -0.00000191, -0.00000356, -0.00000444, -0.00000428, -0.00000309, -0.00000112, 0.00000121, 0.00000340, 0.00000503, 0.00000581, 0.00000572, 0.00000491, 0.00000369, 0.00000243, 0.00000142, 0.00000084, 0.00000068, 0.00000077, 0.00000087, 0.00000073, 0.00000017, -0.00000084, -0.00000221, -0.00000370, -0.00000504, -0.00000601, -0.00000644, -0.00000635, -0.00000584, -0.00000511, -0.00000436, -0.00000373, -0.00000326, -0.00000284, -0.00000231, -0.00000149, -0.00000025, 0.00000136, 0.00000316, 0.00000480, 0.00000584, 0.00000591, 0.00000474, 0.00000232, -0.00000108, -0.00000498, -0.00000870, -0.00001158, -0.00001306, -0.00001282, -0.00001090, -0.00000762, -0.00000354, 0.00000063, 0.00000424, 0.00000677, 0.00000799, 0.00000790, 0.00000679, 0.00000506, 0.00000318, 0.00000156, 0.00000047, 0.00000000, 0.00000014, 0.00000074, 0.00000165, 0.00000275, 0.00000393, 0.00000515, 0.00000637, 0.00000751, 0.00000841, 0.00000887, 0.00000861, 0.00000741, 0.00000516, 0.00000192, -0.00000202, -0.00000616, -0.00000989, -0.00001258, -0.00001372, -0.00001305, -0.00001065, -0.00000693, -0.00000254, 0.00000170, 0.00000504, 0.00000691, 0.00000704, 0.00000554, 0.00000284, -0.00000044, -0.00000359, -0.00000602, -0.00000734, -0.00000744, -0.00000648, -0.00000480, -0.00000283, -0.00000095, 0.00000059, 0.00000173, 0.00000256, 0.00000327, 0.00000403, 0.00000493, 0.00000589, 0.00000672, 0.00000711, 0.00000676, 0.00000551, 0.00000333, 0.00000048, -0.00000263, -0.00000545, -0.00000746, -0.00000828, -0.00000777, -0.00000609, -0.00000366, -0.00000104, 0.00000116, 0.00000247, },
		{ -0.00000600, -0.00000435, -0.00000220, -0.00000013, 0.00000132, 0.00000179, 0.00000117, -0.00000034, -0.00000231, -0.00000417, -0.00000539, -0.00000558, -0.00000459, -0.00000258, 0.00000007, 0.00000283, 0.00000516, 0.00000666, 0.00000711, 0.00000659, 0.00000534, 0.00000377, 0.00000229, 0.00000120, 0.00000065, 0.00000061, 0.00000088, 0.00000116, 0.00000120, 0.00000081, -0.00000004, -0.00000125, -0.00000259, -0.00000383, -0.00000475, -0.00000527, -0.00000541, -0.00000531, -0.00000514, -0.00000507, -0.00000516, -0.00000534, -0.00000543, -0.00000518, -0.00000433, -0.00000277, -0.00000051, 0.00000218, 0.00000491, 0.00000715, 0.00000842, 0.00000837, 0.00000688, 0.00000410, 0.00000044, -0.00000351, -0.00000711, -0.00000982, -0.00001126, -0.00001132, -0.00001016, -0.00000814, -0.00000572, -0.00000336, -0.00000143, -0.00000011, 0.00000058, 0.00000080, 0.00000077, 0.00000074, 0.00000088, 0.00000130, 0.00000201, 0.00000294, 0.00000400, 0.00000515, 0.00000633, 0.00000755, 0.00000881, 0.00001006, 0.00001117, 0.00001193, 0.00001205, 0.00001127, 0.00000935, 0.00000627, 0.00000219, -0.00000248, -0.00000716, -0.00001118, -0.00001391, -0.00001491, -0.00001404, -0.00001148, -0.00000773, -0.00000351, 0.00000042, 0.00000338, 0.00000493, 0.00000496, 0.00000366, 0.00000151, -0.00000089, -0.00000297, -0.00000430, -0.00000469, -0.00000421, -0.00000312, -0.00000181, -0.00000063, 0.00000018, 0.00000056, 0.00000063, 0.00000063, 0.00000085, 0.00000145, 0.00000249, 0.00000380, 0.00000507, 0.00000591, 0.00000596, 0.00000503, 0.00000315, 0.00000060, -0.00000217, -0.00000461, -0.00000623, -0.00000671, },
		{ 0.00000186, -0.00000036, -0.00000263, -0.00000441, -0.00000530, -0.00000508, -0.00000383, -0.00000190, 0.00000017, 0.00000180, 0.00000252, 0.00000208, 0.00000052, -0.00000181, -0.00000434, -0.00000644, -0.00000757, -0.00000739, -0.00000587, -0.00000326, -0.00000008, 0.00000308, 0.00000565, 0.00000722, 0.00000764, 0.00000702, 0.00000569, 0.00000407, 0.00000259, 0.00000156, 0.00000109, 0.00000112, 0.00000142, 0.00000171, 0.00000173, 0.00000135, 0.00000056, -0.00000049, -0.00000160, -0.00000253, -0.00000312, -0.00000333, -0.00000327, -0.00000312, -0.00000312, -0.00000343, -0.00000411, -0.00000507, -0.00000606, -0.00000675, -0.00000682, -0.00000604, -0.00000435, -0.00000191, 0.00000095, 0.00000376, 0.00000603, 0.00000739, 0.00000759, 0.00000664, 0.00000474, 0.00000224, -0.00000043, -0.00000288, -0.00000483, -0.00000615, -0.00000688, -0.00000716, -0.00000717, -0.00000709, -0.00000702, -0.00000696, -0.00000684, -0.00000654, -0.00000592, -0.00000492, -0.00000354, -0.00000186, 0.00000002, 0.00000195, 0.00000387, 0.00000573, 0.00000755, 0.00000936, 0.00001117, 0.00001291, 0.00001440, 0.00001537, 0.00001551, 0.00001452, 0.00001222, 0.00000866, 0.00000409, -0.00000099, -0.00000595, -0.00001009, -0.00001284, -0.00001385, -0.00001308, -0.00001081, -0.00000760, -0.00000413, -0.00000110, 0.00000097, 0.00000181, 0.00000148, 0.00000028, -0.00000128, -0.00000268, -0.00000348, -0.00000347, -0.00000268, -0.00000135, 0.00000013, 0.00000135, 0.00000200, 0.00000199, 0.00000142, 0.00000057, -0.00000019, -0.00000051, -0.00000020, 0.00000072, 0.00000203, 0.00000335, 0.00000425, 0.00000436, 0.00000354, },
		{ 0.00000045, 0.00000165, 0.00000276, 0.00000335, 0.00000311, 0.00000196, 0.00000009, -0.00000210, -0.00000406, -0.00000528, -0.00000542, -0.00000438, -0.00000240, 0.00000006, 0.00000236, 0.00000390, 0.00000422, 0.00000317, 0.00000092, -0.00000207, -0.00000512, -0.00000755, -0.00000881, -0.00000859, -0.00000694, -0.00000419, -0.00000089, 0.00000230, 0.00000483, 0.00000634, 0.00000671, 0.00000612, 0.00000492, 0.00000354, 0.00000239, 0.00000172, 0.00000160, 0.00000189, 0.00000234, 0.00000266, 0.00000262, 0.00000210, 0.00000117, 0.00000002, -0.00000109, -0.00000189, -0.00000221, -0.00000204, -0.00000150, -0.00000084, -0.00000036, -0.00000030, -0.00000079, -0.00000183, -0.00000321, -0.00000465, -0.00000582, -0.00000641, -0.00000627, -0.00000537, -0.00000385, -0.00000197, -0.00000002, 0.00000171, 0.00000302, 0.00000381, 0.00000410, 0.00000398, 0.00000355, 0.00000290, 0.00000208, 0.00000108, -0.00000016, -0.00000166, -0.00000344, -0.00000539, -0.00000737, -0.00000914, -0.00001048, -0.00001118, -0.00001112, -0.00001029, -0.00000880, -0.00000680, -0.00000449, -0.00000203, 0.00000047, 0.00000298, 0.00000553, 0.00000813, 0.00001075, 0.00001327, 0.00001543, 0.00001692, 0.00001738, 0.00001654, 0.00001428, 0.00001071, 0.00000618, 0.00000124, -0.00000344, -0.00000723, -0.00000963, -0.00001042, -0.00000969, -0.00000781, -0.00000537, -0.00000301, -0.00000128, -0.00000053, -0.00000081, -0.00000189, -0.00000335, -0.00000465, -0.00000534, -0.00000512, -0.00000398, -0.00000214, -0.00000001, 0.00000191, 0.00000321, 0.00000365, 0.00000323, 0.00000219, 0.00000093, -0.00000013, -0.00000063, -0.00000041, },
		{ 0.00000255, 0.00000140, 0.00000022, -0.00000056, -0.00000063, 0.00000005, 0.00000129, 0.00000268, 0.00000374, 0.00000402, 0.00000329, 0.00000158, -0.00000080, -0.00000332, -0.00000538, -0.00000644, -0.00000619, -0.00000465, -0.00000214, 0.00000075, 0.00000332, 0.00000494, 0.00000517, 0.00000392, 0.00000143, -0.00000175, -0.00000491, -0.00000737, -0.00000858, -0.00000831, -0.00000666, -0.00000404, -0.00000101, 0.00000179, 0.00000384, 0.00000486, 0.00000483, 0.00000398, 0.00000273, 0.00000153, 0.00000075, 0.00000060, 0.00000108, 0.00000197, 0.00000294, 0.00000366, 0.00000384, 0.00000337, 0.00000233, 0.00000095, -0.00000045, -0.00000154, -0.00000208, -0.00000197, -0.00000126, -0.00000018, 0.00000098, 0.00000193, 0.00000242, 0.00000233, 0.00000166, 0.00000053, -0.00000086, -0.00000230, -0.00000360, -0.00000461, -0.00000526, -0.00000552, -0.00000539, -0.00000488, -0.00000402, -0.00000281, -0.00000130, 0.00000041, 0.00000219, 0.00000381, 0.00000502, 0.00000559, 0.00000532, 0.00000413, 0.00000208, -0.00000062, -0.00000366, -0.00000666, -0.00000925, -0.00001113, -0.00001215, -0.00001227, -0.00001158, -0.00001026, -0.00000847, -0.00000635, -0.00000398, -0.00000137, 0.00000150, 0.00000460, 0.00000784, 0.00001100, 0.00001377, 0.00001575, 0.00001659, 0.00001604, 0.00001404, 0.00001078, 0.00000670, 0.00000239, -0.00000149, -0.00000439, -0.00000595, -0.00000608, -0.00000503, -0.00000326, -0.00000138, 0.00000003, 0.00000054, 0.00000000, -0.00000145, -0.00000341, -0.00000535, -0.00000673, -0.00000714, -0.00000644, -0.00000476, -0.00000247, -0.00000009, 0.00000186, 0.00000301, 0.00000320, },
		{ -0.00000393, -0.00000188, 0.00000001, 0.00000126, 0.00000160, 0.00000104, -0.00000016, -0.00000152, -0.00000254, -0.00000278, -0.00000205, -0.00000044, 0.00000168, 0.00000378, 0.00000526, 0.00000567, 0.00000480, 0.00000276, -0.00000002, -0.00000291, -0.00000524, -0.00000645, -0.00000625, -0.00000469, -0.00000216, 0.00000070, 0.00000317, 0.00000463, 0.00000470, 0.00000333, 0.00000085, -0.00000217, -0.00000502, -0.00000703, -0.00000777, -0.00000707, -0.00000514, -0.00000245, 0.00000038, 0.00000273, 0.00000413, 0.00000438, 0.00000355, 0.00000198, 0.00000016, -0.00000138, -0.00000223, -0.00000219, -0.00000128, 0.00000024, 0.00000198, 0.00000350, 0.00000444, 0.00000460, 0.00000396, 0.00000270, 0.00000112, -0.00000042, -0.00000159, -0.00000215, -0.00000200, -0.00000120, 0.00000010, 0.00000166, 0.00000321, 0.00000453, 0.00000540, 0.00000569, 0.00000534, 0.00000435, 0.00000276, 0.00000070, -0.00000165, -0.00000405, -0.00000623, -0.00000789, -0.00000876, -0.00000867, -0.00000756, -0.00000552, -0.00000283, 0.00000014, 0.00000293, 0.00000509, 0.00000627, 0.00000630, 0.00000516, 0.00000307, 0.00000034, -0.00000262, -0.00000542, -0.00000774, -0.00000942, -0.00001039, -0.00001069, -0.00001044, -0.00000970, -0.00000853, -0.00000692, -0.00000482, -0.00000221, 0.00000087, 0.00000424, 0.00000762, 0.00001060, 0.00001276, 0.00001372, 0.00001327, 0.00001144, 0.00000850, 0.00000496, 0.00000143, -0.00000145, -0.00000323, -0.00000368, -0.00000287, -0.00000119, 0.00000082, 0.00000254, 0.00000344, 0.00000323, 0.00000191, -0.00000022, -0.00000265, -0.00000481, -0.00000617, -0.00000645, -0.00000562, },
		{ 0.00000105, -0.00000111, -0.00000277, -0.00000354, -0.00000328, -0.00000212, -0.00000049, 0.00000108, 0.00000205, 0.00000205, 0.00000097, -0.00000096, -0.00000328, -0.00000536, -0.00000661, -0.00000662, -0.00000527, -0.00000278, 0.00000033, 0.00000335, 0.00000561, 0.00000660, 0.00000611, 0.00000429, 0.00000163, -0.00000121, -0.00000352, -0.00000474, -0.00000462, -0.00000324, -0.00000102, 0.00000139, 0.00000330, 0.00000415, 0.00000365, 0.00000186, -0.00000084, -0.00000382, -0.00000636, -0.00000785, -0.00000790, -0.00000648, -0.00000389, -0.00000068, 0.00000246, 0.00000485, 0.00000602, 0.00000579, 0.00000426, 0.00000185, -0.00000088, -0.00000333, -0.00000498, -0.00000553, -0.00000493, -0.00000335, -0.00000119, 0.00000110, 0.00000306, 0.00000436, 0.00000481, 0.00000441, 0.00000331, 0.00000179, 0.00000018, -0.00000121, -0.00000208, -0.00000228, -0.00000171, -0.00000042, 0.00000145, 0.00000368, 0.00000595, 0.00000794, 0.00000932, 0.00000980, 0.00000922, 0.00000751, 0.00000482, 0.00000144, -0.00000222, -0.00000566, -0.00000838, -0.00001000, -0.00001030, -0.00000929, -0.00000718, -0.00000438, -0.00000140, 0.00000128, 0.00000322, 0.00000419, 0.00000412, 0.00000315, 0.00000153, -0.00000042, -0.00000240, -0.00000419, -0.00000566, -0.00000679, -0.00000759, -0.00000809, -0.00000828, -0.00000808, -0.00000737, -0.00000603, -0.00000399, -0.00000131, 0.00000179, 0.00000496, 0.00000775, 0.00000969, 0.00001042, 0.00000978, 0.00000788, 0.00000507, 0.00000191, -0.00000099, -0.00000304, -0.00000387, -0.00000339, -0.00000180, 0.00000041, 0.00000266, 0.00000436, 0.00000506, 0.00000460, 0.00000313, },
		{ 0.00000026, 0.00000230, 0.00000364, 0.00000398, 0.00000331, 0.00000193, 0.00000031, -0.00000099, -0.00000152, -0.00000103, 0.00000042, 0.00000245, 0.00000448, 0.00000589, 0.00000613, 0.00000495, 0.00000240, -0.00000107, -0.00000480, -0.00000799, -0.00000995, -0.00001025, -0.00000882, -0.00000598, -0.00000235, 0.00000128, 0.00000415, 0.00000571, 0.00000578, 0.00000451, 0.00000239, 0.00000009, -0.00000170, -0.00000248, -0.00000205, -0.00000057, 0.00000152, 0.00000356, 0.00000489, 0.00000503, 0.00000377, 0.00000125, -0.00000205, -0.00000544, -0.00000817, -0.00000962, -0.00000942, -0.00000758, -0.00000442, -0.00000057, 0.00000324, 0.00000628, 0.00000798, 0.00000809, 0.00000665, 0.00000400, 0.00000069, -0.00000267, -0.00000547, -0.00000727, -0.00000784, -0.00000718, -0.00000552, -0.00000320, -0.00000066, 0.00000168, 0.00000346, 0.00000446, 0.00000459, 0.00000390, 0.00000260, 0.00000097, -0.00000065, -0.00000191, -0.00000249, -0.00000219, -0.00000092, 0.00000121, 0.00000396, 0.00000694, 0.00000966, 0.00001165, 0.00001253, 0.00001206, 0.00001023, 0.00000725, 0.00000354, -0.00000036, -0.00000390, -0.00000659, -0.00000812, -0.00000840, -0.00000757, -0.00000595, -0.00000396, -0.00000204, -0.00000051, 0.00000041, 0.00000067, 0.00000039, -0.00000026, -0.00000107, -0.00000190, -0.00000265, -0.00000334, -0.00000400, -0.00000469, -0.00000540, -0.00000603, -0.00000640, -0.00000631, -0.00000556, -0.00000403, -0.00000180, 0.00000091, 0.00000369, 0.00000608, 0.00000760, 0.00000794, 0.00000699, 0.00000493, 0.00000216, -0.00000073, -0.00000316, -0.00000461, -0.00000484, -0.00000386, -0.00000199, },
		{ -0.00000212, -0.00000395, -0.00000480, -0.00000453, -0.00000332, -0.00000155, 0.00000023, 0.00000152, 0.00000195, 0.00000143, 0.00000015, -0.00000143, -0.00000275, -0.00000323, -0.00000253, -0.00000058, 0.00000232, 0.00000560, 0.00000848, 0.00001023, 0.00001031, 0.00000850, 0.00000503, 0.00000049, -0.00000429, -0.00000840, -0.00001107, -0.00001186, -0.00001071, -0.00000801, -0.00000442, -0.00000077, 0.00000216, 0.00000386, 0.00000414, 0.00000319, 0.00000152, -0.00000022, -0.00000139, -0.00000154, -0.00000054, 0.00000140, 0.00000379, 0.00000596, 0.00000727, 0.00000723, 0.00000567, 0.00000276, -0.00000101, -0.00000495, -0.00000827, -0.00001033, -0.00001072, -0.00000935, -0.00000647, -0.00000262, 0.00000151, 0.00000520, 0.00000783, 0.00000901, 0.00000861, 0.00000678, 0.00000388, 0.00000042, -0.00000304, -0.00000600, -0.00000805, -0.00000896, -0.00000870, -0.00000737, -0.00000526, -0.00000271, -0.00000014, 0.00000206, 0.00000359, 0.00000423, 0.00000397, 0.00000293, 0.00000138, -0.00000027, -0.00000162, -0.00000226, -0.00000193, -0.00000056, 0.00000175, 0.00000466, 0.00000770, 0.00001036, 0.00001215, 0.00001276, 0.00001206, 0.00001016, 0.00000741, 0.00000425, 0.00000119, -0.00000135, -0.00000309, -0.00000395, -0.00000403, -0.00000357, -0.00000289, -0.00000227, -0.00000190, -0.00000186, -0.00000208, -0.00000244, -0.00000278, -0.00000297, -0.00000299, -0.00000289, -0.00000281, -0.00000288, -0.00000320, -0.00000377, -0.00000445, -0.00000502, -0.00000519, -0.00000473, -0.00000352, -0.00000162, 0.00000071, 0.00000308, 0.00000502, 0.00000611, 0.00000608, 0.00000492, 0.00000286, 0.00000032, },
		{ 0.00000367, 0.00000422, 0.00000381, 0.00000251, 0.00000061, -0.00000143, -0.00000310, -0.00000399, -0.00000390, -0.00000288, -0.00000127, 0.00000045, 0.00000173, 0.00000211, 0.00000141, -0.00000028, -0.00000255, -0.00000476, -0.00000624, -0.00000641, -0.00000499, -0.00000205, 0.00000192, 0.00000617, 0.00000980, 0.00001200, 0.00001220, 0.00001028, 0.00000652, 0.00000162, -0.00000349, -0.00000786, -0.00001070, -0.00001159, -0.00001053, -0.00000793, -0.00000448, -0.00000103, 0.00000169, 0.00000317, 0.00000327, 0.00000221, 0.00000050, -0.00000121, -0.00000229, -0.00000232, -0.00000116, 0.00000099, 0.00000366, 0.00000621, 0.00000802, 0.00000858, 0.00000767, 0.00000537, 0.00000204, -0.00000174, -0.00000528, -0.00000798, -0.00000938, -0.00000927, -0.00000772, -0.00000501, -0.00000164, 0.00000184, 0.00000488, 0.00000704, 0.00000802, 0.00000772, 0.00000622, 0.00000377, 0.00000072, -0.00000249, -0.00000543, -0.00000771, -0.00000903, -0.00000926, -0.00000840, -0.00000663, -0.00000427, -0.00000173, 0.00000058, 0.00000228, 0.00000315, 0.00000312, 0.00000233, 0.00000110, -0.00000018, -0.00000109, -0.00000127, -0.00000056, 0.00000104, 0.00000329, 0.00000582, 0.00000816, 0.00000991, 0.00001077, 0.00001062, 0.00000956, 0.00000786, 0.00000586, 0.00000396, 0.00000244, 0.00000144, 0.00000096, 0.00000083, 0.00000082, 0.00000069, 0.00000029, -0.00000044, -0.00000140, -0.00000242, -0.00000329, -0.00000382, -0.00000394, -0.00000368, -0.00000319, -0.00000271, -0.00000244, -0.00000251, -0.00000292, -0.00000354, -0.00000411, -0.00000434, -0.00000402, -0.00000303, -0.00000145, 0.00000046, 0.00000230, },
		{ -0.00000297, -0.00000267, -0.00000185, -0.00000064, 0.00000067, 0.00000175, 0.00000224, 0.00000194, 0.00000084, -0.00000084, -0.00000272, -0.00000430, -0.00000514, -0.00000493, -0.00000366, -0.00000158, 0.00000081, 0.00000290, 0.00000410, 0.00000402, 0.00000257, 0.00000001, -0.00000307, -0.00000591, -0.00000775, -0.00000801, -0.00000646, -0.00000327, 0.00000097, 0.00000542, 0.00000913, 0.00001126, 0.00001133, 0.00000923, 0.00000536, 0.00000046, -0.00000453, -0.00000864, -0.00001113, -0.00001164, -0.00001020, -0.00000726, -0.00000354, 0.00000013, 0.00000301, 0.00000461, 0.00000477, 0.00000367, 0.00000179, -0.00000027, -0.00000191, -0.00000268, -0.00000237, -0.00000103, 0.00000102, 0.00000331, 0.00000533, 0.00000663, 0.00000694, 0.00000618, 0.00000447, 0.00000213, -0.00000046, -0.00000286, -0.00000472, -0.00000577, -0.00000587, -0.00000506, -0.00000348, -0.00000137, 0.00000095, 0.00000313, 0.00000486, 0.00000586, 0.00000596, 0.00000508, 0.00000332, 0.00000089, -0.00000189, -0.00000460, -0.00000686, -0.00000833, -0.00000881, -0.00000826, -0.00000685, -0.00000486, -0.00000271, -0.00000077, 0.00000064, 0.00000134, 0.00000138, 0.00000093, 0.00000031, -0.00000013, -0.00000011, 0.00000054, 0.00000181, 0.00000350, 0.00000531, 0.00000689, 0.00000795, 0.00000830, 0.00000795, 0.00000704, 0.00000585, 0.00000470, 0.00000386, 0.00000348, 0.00000355, 0.00000392, 0.00000434, 0.00000453, 0.00000427, 0.00000347, 0.00000220, 0.00000064, -0.00000092, -0.00000220, -0.00000300, -0.00000324, -0.00000300, -0.00000247, -0.00000191, -0.00000155, -0.00000152, -0.00000183, -0.00000233, -0.00000280, },
		{ -0.00000064, -0.00000031, -0.00000024, -0.00000039, -0.00000062, -0.00000073, -0.00000054, 0.00000003, 0.00000091, 0.00000190, 0.00000267, 0.00000289, 0.00000229, 0.00000082, -0.00000138, -0.00000394, -0.00000631, -0.00000796, -0.00000845, -0.00000757, -0.00000542, -0.00000243, 0.00000078, 0.00000349, 0.00000508, 0.00000518, 0.00000375, 0.00000116, -0.00000195, -0.00000477, -0.00000655, -0.00000675, -0.00000520, -0.00000215, 0.00000177, 0.00000572, 0.00000877, 0.00001018, 0.00000954, 0.00000686, 0.00000258, -0.00000251, -0.00000749, -0.00001144, -0.00001367, -0.00001384, -0.00001202, -0.00000864, -0.00000438, -0.00000002, 0.00000371, 0.00000629, 0.00000746, 0.00000728, 0.00000601, 0.00000411, 0.00000205, 0.00000025, -0.00000098, -0.00000152, -0.00000137, -0.00000069, 0.00000033, 0.00000144, 0.00000244, 0.00000316, 0.00000351, 0.00000343, 0.00000295, 0.00000213, 0.00000107, -0.00000006, -0.00000110, -0.00000186, -0.00000218, -0.00000198, -0.00000124, -0.00000006, 0.00000134, 0.00000268, 0.00000366, 0.00000402, 0.00000358, 0.00000234, 0.00000045, -0.00000181, -0.00000406, -0.00000594, -0.00000715, -0.00000753, -0.00000710, -0.00000605, -0.00000465, -0.00000322, -0.00000204, -0.00000125, -0.00000086, -0.00000073, -0.00000063, -0.00000033, 0.00000036, 0.00000149, 0.00000295, 0.00000452, 0.00000592, 0.00000686, 0.00000714, 0.00000674, 0.00000577, 0.00000450, 0.00000328, 0.00000242, 0.00000215, 0.00000251, 0.00000339, 0.00000452, 0.00000556, 0.00000618, 0.00000616, 0.00000544, 0.00000413, 0.00000248, 0.00000082, -0.00000055, -0.00000143, -0.00000174, -0.00000158, -0.00000113, },
		{ 0.00000182, 0.00000063, -0.00000014, -0.00000046, -0.00000042, -0.00000018, 0.00000007, 0.00000023, 0.00000028, 0.00000032, 0.00000051, 0.00000103, 0.00000196, 0.00000327, 0.00000474, 0.00000604, 0.00000675, 0.00000652, 0.00000513, 0.00000261, -0.00000073, -0.00000440, -0.00000773, -0.00001010, -0.00001106, -0.00001041, -0.00000833, -0.00000527, -0.00000191, 0.00000102, 0.00000291, 0.00000344, 0.00000259, 0.00000075, -0.00000149, -0.00000338, -0.00000429, -0.00000381, -0.00000188, 0.00000118, 0.00000475, 0.00000800, 0.00001016, 0.00001061, 0.00000906, 0.00000564, 0.00000081, -0.00000467, -0.00000993, -0.00001415, -0.00001671, -0.00001728, -0.00001588, -0.00001282, -0.00000861, -0.00000388, 0.00000074, 0.00000476, 0.00000781, 0.00000970, 0.00001042, 0.00001006, 0.00000880, 0.00000689, 0.00000457, 0.00000214, -0.00000015, -0.00000203, -0.00000330, -0.00000380, -0.00000350, -0.00000246, -0.00000089, 0.00000094, 0.00000266, 0.00000398, 0.00000465, 0.00000459, 0.00000384, 0.00000263, 0.00000127, 0.00000010, -0.00000063, -0.00000076, -0.00000032, 0.00000052, 0.00000147, 0.00000221, 0.00000246, 0.00000208, 0.00000105, -0.00000044, -0.00000214, -0.00000374, -0.00000496, -0.00000565, -0.00000578, -0.00000546, -0.00000488, -0.00000426, -0.00000375, -0.00000342, -0.00000321, -0.00000295, -0.00000245, -0.00000155, -0.00000018, 0.00000159, 0.00000355, 0.00000537, 0.00000674, 0.00000738, 0.00000717, 0.00000618, 0.00000464, 0.00000292, 0.00000143, 0.00000050, 0.00000034, 0.00000097, 0.00000219, 0.00000367, 0.00000505, 0.00000596, 0.00000621, 0.00000573, 0.00000466, 0.00000326, },
		{ 0.00000374, 0.00000467, 0.00000502, 0.00000477, 0.00000406, 0.00000309, 0.00000211, 0.00000131, 0.00000077, 0.00000048, 0.00000032, 0.00000017, -0.00000007, -0.00000039, -0.00000067, -0.00000070, -0.00000025, 0.00000085, 0.00000264, 0.00000496, 0.00000747, 0.00000968, 0.00001109, 0.00001126, 0.00000996, 0.00000726, 0.00000348, -0.00000080, -0.00000489, -0.00000813, -0.00001007, -0.00001050, -0.00000954, -0.00000762, -0.00000530, -0.00000320, -0.00000178, -0.00000129, -0.00000166, -0.00000255, -0.00000347, -0.00000385, -0.00000328, -0.00000156, 0.00000119, 0.00000458, 0.00000799, 0.00001074, 0.00001219, 0.00001195, 0.00000989, 0.00000620, 0.00000134, -0.00000405, -0.00000927, -0.00001371, -0.00001689, -0.00001852, -0.00001853, -0.00001704, -0.00001426, -0.00001051, -0.00000611, -0.00000142, 0.00000321, 0.00000742, 0.00001086, 0.00001323, 0.00001427, 0.00001384, 0.00001200, 0.00000895, 0.00000512, 0.00000106, -0.00000264, -0.00000542, -0.00000689, -0.00000688, -0.00000552, -0.00000315, -0.00000028, 0.00000250, 0.00000468, 0.00000593, 0.00000613, 0.00000542, 0.00000409, 0.00000255, 0.00000120, 0.00000032, 0.00000002, 0.00000024, 0.00000076, 0.00000131, 0.00000161, 0.00000151, 0.00000094, 0.00000000, -0.00000111, -0.00000219, -0.00000304, -0.00000359, -0.00000384, -0.00000390, -0.00000393, -0.00000405, -0.00000432, -0.00000468, -0.00000497, -0.00000495, -0.00000442, -0.00000324, -0.00000142, 0.00000086, 0.00000330, 0.00000549, 0.00000705, 0.00000771, 0.00000737, 0.00000615, 0.00000434, 0.00000237, 0.00000067, -0.00000041, -0.00000067, -0.00000013, 0.00000101, 0.00000242, },
		{ 0.00000088, -0.00000019, -0.00000056, -0.00000026, 0.00000051, 0.00000149, 0.00000239, 0.00000298, 0.00000318, 0.00000303, 0.00000265, 0.00000221, 0.00000187, 0.00000168, 0.00000159, 0.00000149, 0.00000122, 0.00000064, -0.00000024, -0.00000132, -0.00000231, -0.00000290, -0.00000274, -0.00000160, 0.00000055, 0.00000351, 0.00000689, 0.00001013, 0.00001263, 0.00001393, 0.00001374, 0.00001206, 0.00000917, 0.00000555, 0.00000179, -0.00000156, -0.00000411, -0.00000569, -0.00000638, -0.00000643, -0.00000623, -0.00000609, -0.00000624, -0.00000668, -0.00000724, -0.00000759, -0.00000735, -0.00000624, -0.00000413, -0.00000111, 0.00000248, 0.00000615, 0.00000933, 0.00001150, 0.00001230, 0.00001161, 0.00000950, 0.00000628, 0.00000235, -0.00000184, -0.00000590, -0.00000955, -0.00001258, -0.00001490, -0.00001644, -0.00001713, -0.00001686, -0.00001553, -0.00001305, -0.00000941, -0.00000479, 0.00000049, 0.00000589, 0.00001078, 0.00001450, 0.00001652, 0.00001652, 0.00001454, 0.00001091, 0.00000626, 0.00000139, -0.00000288, -0.00000589, -0.00000725, -0.00000690, -0.00000513, -0.00000244, 0.00000050, 0.00000309, 0.00000488, 0.00000565, 0.00000545, 0.00000452, 0.00000324, 0.00000199, 0.00000107, 0.00000062, 0.00000061, 0.00000088, 0.00000121, 0.00000140, 0.00000131, 0.00000093, 0.00000034, -0.00000030, -0.00000085, -0.00000123, -0.00000142, -0.00000153, -0.00000171, -0.00000211, -0.00000282, -0.00000379, -0.00000487, -0.00000579, -0.00000626, -0.00000602, -0.00000492, -0.00000301, -0.00000053, 0.00000216, 0.00000460, 0.00000639, 0.00000724, 0.00000707, 0.00000600, 0.00000434, 0.00000250, },
		{ 0.00000446, 0.00000544, 0.00000555, 0.00000492, 0.00000382, 0.00000258, 0.00000151, 0.00000083, 0.00000061, 0.00000076, 0.00000113, 0.00000148, 0.00000166, 0.00000159, 0.00000131, 0.00000094, 0.00000068, 0.00000067, 0.00000096, 0.00000149, 0.00000207, 0.00000245, 0.00000239, 0.00000172, 0.00000045, -0.00000124, -0.00000299, -0.00000439, -0.00000502, -0.00000457, -0.00000297, -0.00000035, 0.00000294, 0.00000640, 0.00000951, 0.00001184, 0.00001311, 0.00001326, 0.00001245, 0.00001094, 0.00000905, 0.00000707, 0.00000514, 0.00000329, 0.00000144, -0.00000056, -0.00000278, -0.00000520, -0.00000765, -0.00000980, -0.00001126, -0.00001165, -0.00001074, -0.00000850, -0.00000515, -0.00000115, 0.00000290, 0.00000637, 0.00000877, 0.00000978, 0.00000939, 0.00000783, 0.00000550, 0.00000288, 0.00000038, -0.00000175, -0.00000346, -0.00000488, -0.00000625, -0.00000779, -0.00000962, -0.00001164, -0.00001355, -0.00001488, -0.00001514, -0.00001392, -0.00001104, -0.00000665, -0.00000123, 0.00000448, 0.00000963, 0.00001342, 0.00001526, 0.00001495, 0.00001267, 0.00000896, 0.00000461, 0.00000047, -0.00000272, -0.00000449, -0.00000470, -0.00000355, -0.00000151, 0.00000083, 0.00000290, 0.00000429, 0.00000480, 0.00000448, 0.00000357, 0.00000241, 0.00000132, 0.00000056, 0.00000020, 0.00000021, 0.00000042, 0.00000067, 0.00000079, 0.00000073, 0.00000052, 0.00000027, 0.00000012, 0.00000017, 0.00000041, 0.00000077, 0.00000105, 0.00000103, 0.00000056, -0.00000045, -0.00000190, -0.00000357, -0.00000512, -0.00000619, -0.00000648, -0.00000584, -0.00000431, -0.00000211, 0.00000036, 0.00000268, },
		{ -0.00000477, -0.00000478, -0.00000401, -0.00000266, -0.00000104, 0.00000050, 0.00000169, 0.00000235, 0.00000249, 0.00000224, 0.00000183, 0.00000149, 0.00000139, 0.00000159, 0.00000199, 0.00000244, 0.00000271, 0.00000263, 0.00000214, 0.00000130, 0.00000030, -0.00000060, -0.00000114, -0.00000114, -0.00000059, 0.00000039, 0.00000152, 0.00000247, 0.00000291, 0.00000263, 0.00000161, -0.00000001, -0.00000190, -0.00000367, -0.00000494, -0.00000543, -0.00000501, -0.00000374, -0.00000182, 0.00000045, 0.00000280, 0.00000501, 0.00000698, 0.00000870, 0.00001021, 0.00001156, 0.00001271, 0.00001350, 0.00001371, 0.00001304, 0.00001125, 0.00000826, 0.00000421, -0.00000050, -0.00000526, -0.00000932, -0.00001200, -0.00001279, -0.00001153, -0.00000844, -0.00000410, 0.00000062, 0.00000482, 0.00000771, 0.00000877, 0.00000792, 0.00000548, 0.00000211, -0.00000141, -0.00000429, -0.00000600, -0.00000637, -0.00000560, -0.00000419, -0.00000281, -0.00000208, -0.00000243, -0.00000392, -0.00000625, -0.00000885, -0.00001096, -0.00001190, -0.00001122, -0.00000881, -0.00000498, -0.00000037, 0.00000420, 0.00000790, 0.00001011, 0.00001052, 0.00000924, 0.00000670, 0.00000361, 0.00000068, -0.00000143, -0.00000234, -0.00000200, -0.00000063, 0.00000131, 0.00000329, 0.00000482, 0.00000560, 0.00000552, 0.00000472, 0.00000346, 0.00000209, 0.00000089, 0.00000005, -0.00000041, -0.00000056, -0.00000057, -0.00000059, -0.00000071, -0.00000091, -0.00000110, -0.00000112, -0.00000083, -0.00000017, 0.00000080, 0.00000190, 0.00000286, 0.00000341, 0.00000333, 0.00000254, 0.00000112, -0.00000067, -0.00000249, -0.00000396, },
		{ 0.00000427, 0.00000425, 0.00000358, 0.00000247, 0.00000121, 0.00000009, -0.00000068, -0.00000099, -0.00000089, -0.00000055, -0.00000021, -0.00000004, -0.00000018, -0.00000059, -0.00000114, -0.00000161, -0.00000176, -0.00000143, -0.00000059, 0.00000065, 0.00000204, 0.00000326, 0.00000401, 0.00000411, 0.00000350, 0.00000233, 0.00000087, -0.00000055, -0.00000159, -0.00000204, -0.00000184, -0.00000111, -0.00000008, 0.00000093, 0.00000165, 0.00000188, 0.00000159, 0.00000084, -0.00000016, -0.00000122, -0.00000216, -0.00000286, -0.00000334, -0.00000364, -0.00000384, -0.00000395, -0.00000388, -0.00000345, -0.00000240, -0.00000050, 0.00000236, 0.00000609, 0.00001031, 0.00001444, 0.00001774, 0.00001947, 0.00001911, 0.00001646, 0.00001176, 0.00000571, -0.00000068, -0.00000625, -0.00000998, -0.00001116, -0.00000962, -0.00000575, -0.00000046, 0.00000506, 0.00000954, 0.00001197, 0.00001177, 0.00000898, 0.00000416, -0.00000167, -0.00000733, -0.00001172, -0.00001408, -0.00001415, -0.00001219, -0.00000889, -0.00000519, -0.00000200, -0.00000003, 0.00000042, -0.00000050, -0.00000229, -0.00000423, -0.00000559, -0.00000584, -0.00000477, -0.00000258, 0.00000026, 0.00000306, 0.00000515, 0.00000607, 0.00000563, 0.00000400, 0.00000165, -0.00000080, -0.00000272, -0.00000363, -0.00000332, -0.00000188, 0.00000033, 0.00000281, 0.00000502, 0.00000655, 0.00000718, 0.00000691, 0.00000595, 0.00000460, 0.00000321, 0.00000200, 0.00000110, 0.00000049, 0.00000003, -0.00000044, -0.00000103, -0.00000176, -0.00000252, -0.00000315, -0.00000341, -0.00000315, -0.00000230, -0.00000094, 0.00000070, 0.00000232, 0.00000359, },
		{ -0.00000475, -0.00000448, -0.00000358, -0.00000223, -0.00000070, 0.00000073, 0.00000187, 0.00000262, 0.00000302, 0.00000319, 0.00000331, 0.00000351, 0.00000384, 0.00000422, 0.00000449, 0.00000443, 0.00000387, 0.00000273, 0.00000106, -0.00000090, -0.00000283, -0.00000436, -0.00000516, -0.00000506, -0.00000407, -0.00000239, -0.00000036, 0.00000161, 0.00000314, 0.00000399, 0.00000406, 0.00000345, 0.00000239, 0.00000117, 0.00000005, -0.00000079, -0.00000128, -0.00000147, -0.00000147, -0.00000139, -0.00000130, -0.00000121, -0.00000104, -0.00000069, -0.00000013, 0.00000061, 0.00000136, 0.00000183, 0.00000172, 0.00000077, -0.00000111, -0.00000374, -0.00000668, -0.00000927, -0.00001075, -0.00001047, -0.00000804, -0.00000350, 0.00000262, 0.00000939, 0.00001563, 0.00002013, 0.00002199, 0.00002077, 0.00001666, 0.00001047, 0.00000350, -0.00000280, -0.00000708, -0.00000844, -0.00000660, -0.00000203, 0.00000420, 0.00001063, 0.00001574, 0.00001829, 0.00001759, 0.00001366, 0.00000718, -0.00000066, -0.00000845, -0.00001484, -0.00001888, -0.00002015, -0.00001882, -0.00001557, -0.00001135, -0.00000714, -0.00000373, -0.00000153, -0.00000054, -0.00000041, -0.00000059, -0.00000052, 0.00000016, 0.00000157, 0.00000346, 0.00000540, 0.00000682, 0.00000723, 0.00000639, 0.00000434, 0.00000144, -0.00000173, -0.00000452, -0.00000634, -0.00000684, -0.00000595, -0.00000393, -0.00000122, 0.00000160, 0.00000402, 0.00000567, 0.00000641, 0.00000634, 0.00000570, 0.00000482, 0.00000399, 0.00000337, 0.00000298, 0.00000271, 0.00000236, 0.00000176, 0.00000079, -0.00000050, -0.00000196, -0.00000333, -0.00000434, },
		{ 0.00000414, 0.00000364, 0.00000259, 0.00000114, -0.00000046, -0.00000194, -0.00000308, -0.00000377, -0.00000401, -0.00000389, -0.00000355, -0.00000311, -0.00000263, -0.00000205, -0.00000127, -0.00000017, 0.00000134, 0.00000322, 0.00000532, 0.00000734, 0.00000891, 0.00000970, 0.00000944, 0.00000807, 0.00000573, 0.00000276, -0.00000038, -0.00000321, -0.00000530, -0.00000639, -0.00000645, -0.00000561, -0.00000415, -0.00000242, -0.00000073, 0.00000070, 0.00000175, 0.00000244, 0.00000281, 0.00000295, 0.00000286, 0.00000250, 0.00000179, 0.00000068, -0.00000082, -0.00000256, -0.00000427, -0.00000553, -0.00000597, -0.00000529, -0.00000342, -0.00000059, 0.00000266, 0.00000560, 0.00000741, 0.00000745, 0.00000538, 0.00000135, -0.00000399, -0.00000962, -0.00001429, -0.00001686, -0.00001652, -0.00001307, -0.00000695, 0.00000080, 0.00000872, 0.00001528, 0.00001920, 0.00001978, 0.00001705, 0.00001176, 0.00000526, -0.00000084, -0.00000502, -0.00000624, -0.00000411, 0.00000092, 0.00000776, 0.00001488, 0.00002064, 0.00002371, 0.00002332, 0.00001943, 0.00001268, 0.00000426, -0.00000441, -0.00001196, -0.00001740, -0.00002024, -0.00002055, -0.00001887, -0.00001599, -0.00001273, -0.00000973, -0.00000732, -0.00000549, -0.00000397, -0.00000236, -0.00000036, 0.00000217, 0.00000508, 0.00000797, 0.00001029, 0.00001151, 0.00001127, 0.00000948, 0.00000639, 0.00000255, -0.00000136, -0.00000463, -0.00000671, -0.00000733, -0.00000656, -0.00000473, -0.00000237, -0.00000004, 0.00000181, 0.00000290, 0.00000323, 0.00000299, 0.00000249, 0.00000206, 0.00000195, 0.00000224, 0.00000285, 0.00000355, 0.00000406, },
		{ -0.00000242, -0.00000256, -0.00000198, -0.00000082, 0.00000065, 0.00000212, 0.00000329, 0.00000399, 0.00000415, 0.00000387, 0.00000326, 0.00000248, 0.00000161, 0.00000070, -0.00000029, -0.00000141, -0.00000268, -0.00000402, -0.00000526, -0.00000614, -0.00000636, -0.00000568, -0.00000398, -0.00000136, 0.00000192, 0.00000542, 0.00000860, 0.00001099, 0.00001220, 0.00001209, 0.00001073, 0.00000839, 0.00000548, 0.00000244, -0.00000036, -0.00000266, -0.00000435, -0.00000544, -0.00000601, -0.00000613, -0.00000582, -0.00000507, -0.00000383, -0.00000209, 0.00000007, 0.00000239, 0.00000450, 0.00000593, 0.00000624, 0.00000513, 0.00000260, -0.00000103, -0.00000512, -0.00000879, -0.00001116, -0.00001152, -0.00000956, -0.00000549, -0.00000004, 0.00000566, 0.00001032, 0.00001277, 0.00001228, 0.00000872, 0.00000267, -0.00000469, -0.00001184, -0.00001723, -0.00001968, -0.00001859, -0.00001418, -0.00000737, 0.00000039, 0.00000746, 0.00001239, 0.00001429, 0.00001299, 0.00000911, 0.00000387, -0.00000118, -0.00000456, -0.00000522, -0.00000277, 0.00000243, 0.00000933, 0.00001649, 0.00002238, 0.00002575, 0.00002588, 0.00002271, 0.00001684, 0.00000933, 0.00000143, -0.00000568, -0.00001116, -0.00001463, -0.00001615, -0.00001618, -0.00001529, -0.00001405, -0.00001283, -0.00001171, -0.00001052, -0.00000894, -0.00000668, -0.00000358, 0.00000025, 0.00000443, 0.00000836, 0.00001140, 0.00001299, 0.00001284, 0.00001100, 0.00000786, 0.00000407, 0.00000037, -0.00000254, -0.00000422, -0.00000453, -0.00000365, -0.00000204, -0.00000025, 0.00000117, 0.00000187, 0.00000173, 0.00000086, -0.00000039, -0.00000161, },
		{ 0.00000527, 0.00000529, 0.00000419, 0.00000220, -0.00000024, -0.00000264, -0.00000456, -0.00000572, -0.00000602, -0.00000555, -0.00000452, -0.00000317, -0.00000171, -0.00000027, 0.00000108, 0.00000235, 0.00000353, 0.00000457, 0.00000534, 0.00000567, 0.00000536, 0.00000428, 0.00000241, -0.00000007, -0.00000284, -0.00000546, -0.00000745, -0.00000840, -0.00000808, -0.00000649, -0.00000383, -0.00000052, 0.00000295, 0.00000613, 0.00000864, 0.00001030, 0.00001106, 0.00001104, 0.00001038, 0.00000923, 0.00000766, 0.00000567, 0.00000326, 0.00000043, -0.00000267, -0.00000575, -0.00000838, -0.00001005, -0.00001031, -0.00000890, -0.00000586, -0.00000163, 0.00000302, 0.00000710, 0.00000965, 0.00000993, 0.00000764, 0.00000307, -0.00000294, -0.00000914, -0.00001416, -0.00001676, -0.00001622, -0.00001248, -0.00000620, 0.00000133, 0.00000851, 0.00001376, 0.00001589, 0.00001441, 0.00000961, 0.00000252, -0.00000533, -0.00001224, -0.00001681, -0.00001817, -0.00001623, -0.00001165, -0.00000568, 0.00000020, 0.00000463, 0.00000670, 0.00000618, 0.00000353, -0.00000020, -0.00000371, -0.00000573, -0.00000538, -0.00000239, 0.00000284, 0.00000937, 0.00001592, 0.00002120, 0.00002419, 0.00002439, 0.00002185, 0.00001717, 0.00001127, 0.00000518, -0.00000024, -0.00000442, -0.00000722, -0.00000883, -0.00000966, -0.00001017, -0.00001067, -0.00001124, -0.00001169, -0.00001168, -0.00001080, -0.00000878, -0.00000558, -0.00000150, 0.00000292, 0.00000696, 0.00000991, 0.00001128, 0.00001089, 0.00000892, 0.00000590, 0.00000256, -0.00000032, -0.00000216, -0.00000263, -0.00000178, 0.00000003, 0.00000221, 0.00000414, },
		{ -0.00000490, -0.00000434, -0.00000233, 0.00000062, 0.00000382, 0.00000656, 0.00000828, 0.00000870, 0.00000782, 0.00000588, 0.00000331, 0.00000056, -0.00000199, -0.00000408, -0.00000563, -0.00000664, -0.00000720, -0.00000737, -0.00000715, -0.00000651, -0.00000535, -0.00000364, -0.00000141, 0.00000116, 0.00000378, 0.00000604, 0.00000757, 0.00000805, 0.00000738, 0.00000563, 0.00000310, 0.00000023, -0.00000250, -0.00000468, -0.00000603, -0.00000648, -0.00000610, -0.00000509, -0.00000368, -0.00000202, -0.00000019, 0.00000182, 0.00000407, 0.00000656, 0.00000918, 0.00001163, 0.00001346, 0.00001414, 0.00001322, 0.00001048, 0.00000603, 0.00000041, -0.00000551, -0.00001064, -0.00001393, -0.00001460, -0.00001240, -0.00000769, -0.00000142, 0.00000505, 0.00001021, 0.00001280, 0.00001207, 0.00000802, 0.00000140, -0.00000640, -0.00001368, -0.00001882, -0.00002063, -0.00001866, -0.00001328, -0.00000565, 0.00000260, 0.00000972, 0.00001426, 0.00001537, 0.00001301, 0.00000790, 0.00000133, -0.00000517, -0.00001022, -0.00001289, -0.00001293, -0.00001073, -0.00000724, -0.00000362, -0.00000094, 0.00000009, -0.00000067, -0.00000279, -0.00000540, -0.00000744, -0.00000798, -0.00000643, -0.00000276, 0.00000253, 0.00000853, 0.00001414, 0.00001833, 0.00002042, 0.00002018, 0.00001786, 0.00001410, 0.00000974, 0.00000557, 0.00000216, -0.00000026, -0.00000179, -0.00000278, -0.00000365, -0.00000473, -0.00000614, -0.00000770, -0.00000901, -0.00000959, -0.00000899, -0.00000702, -0.00000378, 0.00000026, 0.00000441, 0.00000786, 0.00000994, 0.00001023, 0.00000873, 0.00000583, 0.00000223, -0.00000125, -0.00000380, },
		{ 0.00001260, 0.00000991, 0.00000583, 0.00000120, -0.00000304, -0.00000608, -0.00000741, -0.00000691, -0.00000486, -0.00000182, 0.00000153, 0.00000451, 0.00000666, 0.00000773, 0.00000776, 0.00000696, 0.00000564, 0.00000408, 0.00000246, 0.00000087, -0.00000074, -0.00000242, -0.00000421, -0.00000604, -0.00000771, -0.00000896, -0.00000949, -0.00000910, -0.00000772, -0.00000547, -0.00000267, 0.00000029, 0.00000297, 0.00000503, 0.00000627, 0.00000668, 0.00000640, 0.00000567, 0.00000469, 0.00000361, 0.00000243, 0.00000107, -0.00000058, -0.00000257, -0.00000477, -0.00000688, -0.00000840, -0.00000878, -0.00000752, -0.00000441, 0.00000039, 0.00000627, 0.00001226, 0.00001717, 0.00001990, 0.00001965, 0.00001620, 0.00001001, 0.00000216, -0.00000584, -0.00001238, -0.00001608, -0.00001619, -0.00001272, -0.00000650, 0.00000100, 0.00000799, 0.00001277, 0.00001416, 0.00001174, 0.00000597, -0.00000192, -0.00001021, -0.00001710, -0.00002111, -0.00002147, -0.00001819, -0.00001211, -0.00000463, 0.00000262, 0.00000819, 0.00001110, 0.00001108, 0.00000853, 0.00000440, -0.00000013, -0.00000395, -0.00000631, -0.00000695, -0.00000619, -0.00000468, -0.00000327, -0.00000268, -0.00000326, -0.00000496, -0.00000727, -0.00000940, -0.00001052, -0.00000999, -0.00000757, -0.00000345, 0.00000171, 0.00000703, 0.00001155, 0.00001453, 0.00001559, 0.00001479, 0.00001256, 0.00000959, 0.00000658, 0.00000411, 0.00000245, 0.00000152, 0.00000102, 0.00000049, -0.00000045, -0.00000199, -0.00000399, -0.00000605, -0.00000758, -0.00000800, -0.00000690, -0.00000419, -0.00000021, 0.00000440, 0.00000874, 0.00001194, 0.00001333, },
		{ -0.00000390, 0.00000117, 0.00000679, 0.00001192, 0.00001559, 0.00001714, 0.00001635, 0.00001347, 0.00000916, 0.00000433, -0.00000009, -0.00000339, -0.00000514, -0.00000533, -0.00000426, -0.00000248, -0.00000054, 0.00000108, 0.00000212, 0.00000257, 0.00000260, 0.00000246, 0.00000241, 0.00000256, 0.00000288, 0.00000318, 0.00000320, 0.00000269, 0.00000153, -0.00000025, -0.00000243, -0.00000467, -0.00000662, -0.00000799, -0.00000860, -0.00000848, -0.00000777, -0.00000668, -0.00000539, -0.00000400, -0.00000247, -0.00000071, 0.00000143, 0.00000397, 0.00000676, 0.00000943, 0.00001141, 0.00001208, 0.00001093, 0.00000775, 0.00000276, -0.00000335, -0.00000950, -0.00001439, -0.00001686, -0.00001608, -0.00001189, -0.00000484, 0.00000383, 0.00001247, 0.00001932, 0.00002292, 0.00002248, 0.00001807, 0.00001060, 0.00000169, -0.00000677, -0.00001296, -0.00001563, -0.00001435, -0.00000961, -0.00000268, 0.00000466, 0.00001058, 0.00001364, 0.00001306, 0.00000898, 0.00000231, -0.00000546, -0.00001266, -0.00001780, -0.00001995, -0.00001889, -0.00001514, -0.00000975, -0.00000403, 0.00000083, 0.00000395, 0.00000504, 0.00000433, 0.00000247, 0.00000028, -0.00000149, -0.00000239, -0.00000237, -0.00000173, -0.00000102, -0.00000084, -0.00000157, -0.00000331, -0.00000574, -0.00000828, -0.00001020, -0.00001084, -0.00000983, -0.00000717, -0.00000327, 0.00000116, 0.00000528, 0.00000836, 0.00000995, 0.00000997, 0.00000873, 0.00000678, 0.00000479, 0.00000330, 0.00000256, 0.00000253, 0.00000285, 0.00000299, 0.00000246, 0.00000097, -0.00000145, -0.00000440, -0.00000720, -0.00000906, -0.00000931, -0.00000757, },
		{ -0.00000095, -0.00000508, -0.00000891, -0.00001148, -0.00001203, -0.00001022, -0.00000623, -0.00000072, 0.00000533, 0.00001082, 0.00001482, 0.00001674, 0.00001648, 0.00001438, 0.00001117, 0.00000768, 0.00000468, 0.00000268, 0.00000183, 0.00000193, 0.00000253, 0.00000308, 0.00000313, 0.00000242, 0.00000101, -0.00000083, -0.00000265, -0.00000401, -0.00000456, -0.00000420, -0.00000305, -0.00000141, 0.00000032, 0.00000178, 0.00000271, 0.00000302, 0.00000277, 0.00000213, 0.00000123, 0.00000015, -0.00000113, -0.00000270, -0.00000467, -0.00000702, -0.00000954, -0.00001180, -0.00001321, -0.00001313, -0.00001106, -0.00000685, -0.00000080, 0.00000628, 0.00001318, 0.00001852, 0.00002107, 0.00002001, 0.00001526, 0.00000749, -0.00000191, -0.00001110, -0.00001820, -0.00002166, -0.00002070, -0.00001545, -0.00000696, 0.00000300, 0.00001234, 0.00001914, 0.00002206, 0.00002062, 0.00001537, 0.00000765, -0.00000068, -0.00000771, -0.00001193, -0.00001256, -0.00000972, -0.00000436, 0.00000203, 0.00000776, 0.00001140, 0.00001207, 0.00000965, 0.00000477, -0.00000144, -0.00000759, -0.00001244, -0.00001515, -0.00001546, -0.00001369, -0.00001060, -0.00000713, -0.00000415, -0.00000223, -0.00000152, -0.00000182, -0.00000260, -0.00000332, -0.00000354, -0.00000307, -0.00000206, -0.00000091, -0.00000009, -0.00000003, -0.00000092, -0.00000263, -0.00000475, -0.00000668, -0.00000782, -0.00000772, -0.00000628, -0.00000373, -0.00000060, 0.00000241, 0.00000464, 0.00000566, 0.00000537, 0.00000405, 0.00000226, 0.00000065, -0.00000019, 0.00000002, 0.00000121, 0.00000295, 0.00000457, 0.00000534, 0.00000473, 0.00000255, },
		{ -0.00000020, 0.00000315, 0.00000598, 0.00000741, 0.00000686, 0.00000428, 0.00000011, -0.00000476, -0.00000923, -0.00001224, -0.00001307, -0.00001149, -0.00000780, -0.00000278, 0.00000253, 0.00000710, 0.00001014, 0.00001133, 0.00001081, 0.00000913, 0.00000711, 0.00000554, 0.00000500, 0.00000571, 0.00000746, 0.00000970, 0.00001172, 0.00001281, 0.00001251, 0.00001066, 0.00000748, 0.00000346, -0.00000076, -0.00000454, -0.00000743, -0.00000919, -0.00000983, -0.00000951, -0.00000848, -0.00000694, -0.00000500, -0.00000270, -0.00000007, 0.00000282, 0.00000571, 0.00000815, 0.00000956, 0.00000933, 0.00000703, 0.00000259, -0.00000360, -0.00001062, -0.00001720, -0.00002189, -0.00002346, -0.00002115, -0.00001499, -0.00000579, 0.00000487, 0.00001502, 0.00002264, 0.00002616, 0.00002480, 0.00001882, 0.00000943, -0.00000142, -0.00001148, -0.00001867, -0.00002159, -0.00001980, -0.00001389, -0.00000533, 0.00000387, 0.00001168, 0.00001650, 0.00001750, 0.00001478, 0.00000931, 0.00000262, -0.00000357, -0.00000781, -0.00000925, -0.00000777, -0.00000402, 0.00000085, 0.00000547, 0.00000864, 0.00000961, 0.00000821, 0.00000488, 0.00000050, -0.00000392, -0.00000745, -0.00000951, -0.00000996, -0.00000909, -0.00000751, -0.00000587, -0.00000476, -0.00000446, -0.00000494, -0.00000590, -0.00000690, -0.00000751, -0.00000747, -0.00000678, -0.00000567, -0.00000452, -0.00000372, -0.00000352, -0.00000393, -0.00000470, -0.00000541, -0.00000558, -0.00000485, -0.00000310, -0.00000056, 0.00000229, 0.00000479, 0.00000629, 0.00000638, 0.00000500, 0.00000246, -0.00000056, -0.00000323, -0.00000480, -0.00000477, -0.00000310, },
		{ 0.00000197, -0.00000249, -0.00000595, -0.00000759, -0.00000703, -0.00000444, -0.00000052, 0.00000367, 0.00000697, 0.00000846, 0.00000766, 0.00000466, 0.00000015, -0.00000481, -0.00000905, -0.00001156, -0.00001178, -0.00000973, -0.00000597, -0.00000148, 0.00000263, 0.00000542, 0.00000633, 0.00000537, 0.00000304, 0.00000024, -0.00000203, -0.00000289, -0.00000185, 0.00000110, 0.00000548, 0.00001045, 0.00001507, 0.00001846, 0.00002005, 0.00001964, 0.00001742, 0.00001382, 0.00000943, 0.00000475, 0.00000016, -0.00000413, -0.00000804, -0.00001150, -0.00001436, -0.00001632, -0.00001694, -0.00001578, -0.00001255, -0.00000727, -0.00000043, 0.00000701, 0.00001374, 0.00001832, 0.00001956, 0.00001679, 0.00001013, 0.00000054, -0.00001029, -0.00002026, -0.00002731, -0.00002985, -0.00002720, -0.00001971, -0.00000879, 0.00000344, 0.00001455, 0.00002238, 0.00002548, 0.00002346, 0.00001702, 0.00000776, -0.00000214, -0.00001053, -0.00001572, -0.00001682, -0.00001399, -0.00000821, -0.00000113, 0.00000550, 0.00001014, 0.00001191, 0.00001069, 0.00000712, 0.00000238, -0.00000218, -0.00000539, -0.00000653, -0.00000550, -0.00000277, 0.00000076, 0.00000405, 0.00000622, 0.00000675, 0.00000560, 0.00000318, 0.00000018, -0.00000262, -0.00000461, -0.00000545, -0.00000521, -0.00000422, -0.00000302, -0.00000214, -0.00000196, -0.00000259, -0.00000390, -0.00000556, -0.00000719, -0.00000847, -0.00000927, -0.00000963, -0.00000975, -0.00000987, -0.00001016, -0.00001060, -0.00001097, -0.00001090, -0.00000998, -0.00000790, -0.00000464, -0.00000049, 0.00000395, 0.00000786, 0.00001044, 0.00001110, 0.00000964, 0.00000635, },
		{ 0.00000445, 0.00000957, 0.00001263, 0.00001310, 0.00001099, 0.00000688, 0.00000184, -0.00000289, -0.00000612, -0.00000708, -0.00000559, -0.00000210, 0.00000239, 0.00000662, 0.00000939, 0.00000986, 0.00000779, 0.00000362, -0.00000168, -0.00000681, -0.00001050, -0.00001187, -0.00001060, -0.00000706, -0.00000220, 0.00000272, 0.00000643, 0.00000799, 0.00000702, 0.00000380, -0.00000084, -0.00000572, -0.00000960, -0.00001151, -0.00001097, -0.00000803, -0.00000324, 0.00000251, 0.00000828, 0.00001328, 0.00001701, 0.00001931, 0.00002033, 0.00002029, 0.00001942, 0.00001775, 0.00001515, 0.00001135, 0.00000615, -0.00000042, -0.00000792, -0.00001550, -0.00002196, -0.00002600, -0.00002651, -0.00002294, -0.00001550, -0.00000524, 0.00000606, 0.00001626, 0.00002325, 0.00002550, 0.00002237, 0.00001434, 0.00000294, -0.00000955, -0.00002061, -0.00002801, -0.00003029, -0.00002714, -0.00001937, -0.00000874, 0.00000243, 0.00001187, 0.00001779, 0.00001932, 0.00001661, 0.00001076, 0.00000346, -0.00000342, -0.00000829, -0.00001025, -0.00000916, -0.00000570, -0.00000105, 0.00000342, 0.00000652, 0.00000755, 0.00000644, 0.00000370, 0.00000023, -0.00000293, -0.00000493, -0.00000532, -0.00000411, -0.00000178, 0.00000092, 0.00000318, 0.00000440, 0.00000430, 0.00000298, 0.00000089, -0.00000135, -0.00000313, -0.00000404, -0.00000393, -0.00000297, -0.00000153, -0.00000010, 0.00000090, 0.00000120, 0.00000076, -0.00000030, -0.00000175, -0.00000337, -0.00000503, -0.00000672, -0.00000854, -0.00001060, -0.00001291, -0.00001533, -0.00001748, -0.00001885, -0.00001886, -0.00001709, -0.00001340, -0.00000808, -0.00000180, },
		{ -0.00002337, -0.00002418, -0.00002249, -0.00001830, -0.00001215, -0.00000503, 0.00000179, 0.00000708, 0.00000989, 0.00000983, 0.00000718, 0.00000286, -0.00000183, -0.00000550, -0.00000703, -0.00000587, -0.00000221, 0.00000310, 0.00000870, 0.00001310, 0.00001506, 0.00001394, 0.00000983, 0.00000359, -0.00000336, -0.00000942, -0.00001321, -0.00001387, -0.00001132, -0.00000626, 0.00000002, 0.00000597, 0.00001018, 0.00001168, 0.00001020, 0.00000617, 0.00000061, -0.00000521, -0.00001007, -0.00001308, -0.00001389, -0.00001268, -0.00001001, -0.00000663, -0.00000319, -0.00000004, 0.00000282, 0.00000568, 0.00000894, 0.00001285, 0.00001733, 0.00002181, 0.00002534, 0.00002678, 0.00002511, 0.00001977, 0.00001095, -0.00000033, -0.00001235, -0.00002298, -0.00003014, -0.00003232, -0.00002893, -0.00002052, -0.00000871, 0.00000413, 0.00001539, 0.00002279, 0.00002488, 0.00002138, 0.00001318, 0.00000217, -0.00000924, -0.00001866, -0.00002427, -0.00002519, -0.00002164, -0.00001478, -0.00000641, 0.00000152, 0.00000737, 0.00001016, 0.00000978, 0.00000689, 0.00000272, -0.00000137, -0.00000418, -0.00000502, -0.00000385, -0.00000120, 0.00000202, 0.00000478, 0.00000627, 0.00000608, 0.00000431, 0.00000151, -0.00000154, -0.00000401, -0.00000530, -0.00000516, -0.00000376, -0.00000161, 0.00000061, 0.00000227, 0.00000292, 0.00000245, 0.00000106, -0.00000081, -0.00000262, -0.00000391, -0.00000438, -0.00000398, -0.00000290, -0.00000144, 0.00000003, 0.00000126, 0.00000209, 0.00000251, 0.00000256, 0.00000230, 0.00000169, 0.00000059, -0.00000120, -0.00000387, -0.00000746, -0.00001179, -0.00001639, -0.00002052, },
		{ 0.00000373, 0.00000106, -0.00000313, -0.00000850, -0.00001430, -0.00001954, -0.00002319, -0.00002440, -0.00002276, -0.00001848, -0.00001234, -0.00000560, 0.00000031, 0.00000412, 0.00000506, 0.00000306, -0.00000118, -0.00000635, -0.00001085, -0.00001318, -0.00001235, -0.00000814, -0.00000121, 0.00000704, 0.00001482, 0.00002034, 0.00002232, 0.00002024, 0.00001455, 0.00000652, -0.00000206, -0.00000930, -0.00001368, -0.00001443, -0.00001162, -0.00000621, 0.00000031, 0.00000630, 0.00001037, 0.00001171, 0.00001028, 0.00000672, 0.00000215, -0.00000223, -0.00000548, -0.00000712, -0.00000733, -0.00000672, -0.00000616, -0.00000640, -0.00000778, -0.00001006, -0.00001240, -0.00001362, -0.00001252, -0.00000832, -0.00000094, 0.00000882, 0.00001937, 0.00002863, 0.00003452, 0.00003546, 0.00003080, 0.00002108, 0.00000793, -0.00000625, -0.00001875, -0.00002721, -0.00003012, -0.00002718, -0.00001930, -0.00000846, 0.00000287, 0.00001221, 0.00001769, 0.00001839, 0.00001455, 0.00000737, -0.00000126, -0.00000934, -0.00001518, -0.00001777, -0.00001700, -0.00001356, -0.00000870, -0.00000382, -0.00000015, 0.00000162, 0.00000143, -0.00000022, -0.00000241, -0.00000417, -0.00000474, -0.00000377, -0.00000145, 0.00000165, 0.00000468, 0.00000684, 0.00000756, 0.00000667, 0.00000442, 0.00000141, -0.00000160, -0.00000392, -0.00000508, -0.00000497, -0.00000381, -0.00000209, -0.00000039, 0.00000081, 0.00000120, 0.00000075, -0.00000033, -0.00000170, -0.00000295, -0.00000380, -0.00000410, -0.00000387, -0.00000325, -0.00000240, -0.00000147, -0.00000051, 0.00000051, 0.00000164, 0.00000286, 0.00000403, 0.00000485, 0.00000490, },
		{ -0.00000163, -0.00000028, 0.00000159, 0.00000371, 0.00000553, 0.00000641, 0.00000576, 0.00000323, -0.00000113, -0.00000678, -0.00001275, -0.00001786, -0.00002097, -0.00002128, -0.00001859, -0.00001342, -0.00000693, -0.00000066, 0.00000379, 0.00000521, 0.00000309, -0.00000218, -0.00000938, -0.00001669, -0.00002214, -0.00002410, -0.00002166, -0.00001495, -0.00000510, 0.00000599, 0.00001609, 0.00002317, 0.00002591, 0.00002395, 0.00001802, 0.00000971, 0.00000108, -0.00000590, -0.00000978, -0.00001003, -0.00000704, -0.00000205, 0.00000330, 0.00000740, 0.00000915, 0.00000823, 0.00000514, 0.00000105, -0.00000258, -0.00000451, -0.00000407, -0.00000140, 0.00000252, 0.00000618, 0.00000794, 0.00000657, 0.00000162, -0.00000634, -0.00001577, -0.00002449, -0.00003020, -0.00003110, -0.00002634, -0.00001630, -0.00000259, 0.00001226, 0.00002542, 0.00003434, 0.00003734, 0.00003400, 0.00002525, 0.00001311, 0.00000018, -0.00001089, -0.00001808, -0.00002036, -0.00001787, -0.00001182, -0.00000409, 0.00000325, 0.00000846, 0.00001050, 0.00000923, 0.00000534, 0.00000009, -0.00000509, -0.00000898, -0.00001084, -0.00001062, -0.00000882, -0.00000634, -0.00000414, -0.00000297, -0.00000315, -0.00000449, -0.00000643, -0.00000815, -0.00000890, -0.00000820, -0.00000599, -0.00000263, 0.00000120, 0.00000467, 0.00000708, 0.00000796, 0.00000728, 0.00000532, 0.00000267, -0.00000002, -0.00000217, -0.00000343, -0.00000375, -0.00000331, -0.00000248, -0.00000166, -0.00000114, -0.00000106, -0.00000137, -0.00000191, -0.00000246, -0.00000285, -0.00000301, -0.00000299, -0.00000288, -0.00000280, -0.00000277, -0.00000270, -0.00000239, },
		{ -0.00000205, -0.00000187, -0.00000208, -0.00000259, -0.00000312, -0.00000328, -0.00000271, -0.00000124, 0.00000100, 0.00000356, 0.00000571, 0.00000666, 0.00000574, 0.00000266, -0.00000234, -0.00000843, -0.00001438, -0.00001880, -0.00002050, -0.00001882, -0.00001389, -0.00000665, 0.00000127, 0.00000795, 0.00001160, 0.00001103, 0.00000604, -0.00000254, -0.00001294, -0.00002283, -0.00002991, -0.00003240, -0.00002953, -0.00002176, -0.00001062, 0.00000158, 0.00001234, 0.00001956, 0.00002207, 0.00001984, 0.00001403, 0.00000657, -0.00000030, -0.00000467, -0.00000543, -0.00000260, 0.00000277, 0.00000885, 0.00001361, 0.00001538, 0.00001333, 0.00000779, 0.00000011, -0.00000762, -0.00001316, -0.00001477, -0.00001169, -0.00000442, 0.00000536, 0.00001512, 0.00002216, 0.00002428, 0.00002033, 0.00001060, -0.00000325, -0.00001849, -0.00003196, -0.00004083, -0.00004317, -0.00003848, -0.00002772, -0.00001308, 0.00000253, 0.00001617, 0.00002551, 0.00002928, 0.00002753, 0.00002144, 0.00001297, 0.00000436, -0.00000248, -0.00000635, -0.00000694, -0.00000485, -0.00000128, 0.00000236, 0.00000485, 0.00000547, 0.00000414, 0.00000137, -0.00000197, -0.00000490, -0.00000669, -0.00000703, -0.00000609, -0.00000445, -0.00000289, -0.00000213, -0.00000260, -0.00000434, -0.00000693, -0.00000969, -0.00001184, -0.00001270, -0.00001193, -0.00000958, -0.00000608, -0.00000212, 0.00000155, 0.00000431, 0.00000581, 0.00000600, 0.00000513, 0.00000364, 0.00000199, 0.00000054, -0.00000050, -0.00000115, -0.00000153, -0.00000184, -0.00000222, -0.00000272, -0.00000326, -0.00000369, -0.00000385, -0.00000366, -0.00000317, -0.00000256, },
		{ -0.00000414, -0.00000493, -0.00000506, -0.00000455, -0.00000363, -0.00000265, -0.00000199, -0.00000190, -0.00000239, -0.00000324, -0.00000398, -0.00000410, -0.00000318, -0.00000111, 0.00000183, 0.00000494, 0.00000728, 0.00000789, 0.00000609, 0.00000174, -0.00000463, -0.00001184, -0.00001829, -0.00002234, -0.00002274, -0.00001899, -0.00001154, -0.00000184, 0.00000803, 0.00001573, 0.00001929, 0.00001759, 0.00001070, -0.00000009, -0.00001251, -0.00002385, -0.00003155, -0.00003391, -0.00003042, -0.00002198, -0.00001063, 0.00000092, 0.00001001, 0.00001465, 0.00001409, 0.00000896, 0.00000111, -0.00000685, -0.00001232, -0.00001333, -0.00000915, -0.00000052, 0.00001057, 0.00002130, 0.00002881, 0.00003090, 0.00002667, 0.00001682, 0.00000347, -0.00001030, -0.00002118, -0.00002646, -0.00002474, -0.00001625, -0.00000288, 0.00001231, 0.00002575, 0.00003421, 0.00003558, 0.00002930, 0.00001651, -0.00000028, -0.00001773, -0.00003252, -0.00004199, -0.00004475, -0.00004083, -0.00003163, -0.00001943, -0.00000685, 0.00000382, 0.00001103, 0.00001427, 0.00001400, 0.00001141, 0.00000798, 0.00000505, 0.00000348, 0.00000350, 0.00000472, 0.00000636, 0.00000753, 0.00000756, 0.00000617, 0.00000359, 0.00000040, -0.00000260, -0.00000470, -0.00000545, -0.00000484, -0.00000326, -0.00000141, -0.00000004, 0.00000023, -0.00000088, -0.00000327, -0.00000647, -0.00000976, -0.00001241, -0.00001384, -0.00001378, -0.00001230, -0.00000976, -0.00000670, -0.00000369, -0.00000115, 0.00000067, 0.00000177, 0.00000231, 0.00000250, 0.00000253, 0.00000249, 0.00000233, 0.00000194, 0.00000120, 0.00000007, -0.00000135, -0.00000286, },
		{ 0.00000283, 0.00000334, 0.00000299, 0.00000184, 0.00000014, -0.00000169, -0.00000320, -0.00000406, -0.00000412, -0.00000352, -0.00000263, -0.00000192, -0.00000180, -0.00000247, -0.00000378, -0.00000530, -0.00000634, -0.00000626, -0.00000463, -0.00000146, 0.00000270, 0.00000685, 0.00000974, 0.00001023, 0.00000760, 0.00000190, -0.00000603, -0.00001460, -0.00002183, -0.00002583, -0.00002528, -0.00001984, -0.00001035, 0.00000129, 0.00001253, 0.00002071, 0.00002377, 0.00002076, 0.00001214, -0.00000025, -0.00001362, -0.00002484, -0.00003126, -0.00003136, -0.00002515, -0.00001423, -0.00000141, 0.00000998, 0.00001694, 0.00001754, 0.00001149, 0.00000026, -0.00001329, -0.00002554, -0.00003305, -0.00003339, -0.00002585, -0.00001159, 0.00000651, 0.00002457, 0.00003860, 0.00004544, 0.00004353, 0.00003331, 0.00001710, -0.00000143, -0.00001814, -0.00002934, -0.00003267, -0.00002758, -0.00001546, 0.00000080, 0.00001744, 0.00003077, 0.00003794, 0.00003758, 0.00002995, 0.00001684, 0.00000102, -0.00001443, -0.00002681, -0.00003437, -0.00003654, -0.00003392, -0.00002797, -0.00002054, -0.00001335, -0.00000761, -0.00000384, -0.00000183, -0.00000092, -0.00000024, 0.00000092, 0.00000287, 0.00000549, 0.00000826, 0.00001047, 0.00001147, 0.00001091, 0.00000884, 0.00000570, 0.00000223, -0.00000077, -0.00000267, -0.00000317, -0.00000236, -0.00000072, 0.00000108, 0.00000234, 0.00000253, 0.00000145, -0.00000077, -0.00000372, -0.00000681, -0.00000949, -0.00001134, -0.00001219, -0.00001210, -0.00001131, -0.00001014, -0.00000884, -0.00000756, -0.00000630, -0.00000497, -0.00000349, -0.00000181, -0.00000003, 0.00000162, },
		{ -0.00001211, -0.00001169, -0.00001037, -0.00000822, -0.00000555, -0.00000283, -0.00000056, 0.00000086, 0.00000126, 0.00000075, -0.00000031, -0.00000140, -0.00000206, -0.00000199, -0.00000121, -0.00000004, 0.00000097, 0.00000124, 0.00000038, -0.00000166, -0.00000448, -0.00000731, -0.00000921, -0.00000931, -0.00000719, -0.00000299, 0.00000244, 0.00000773, 0.00001132, 0.00001185, 0.00000860, 0.00000177, -0.00000742, -0.00001700, -0.00002459, -0.00002808, -0.00002616, -0.00001876, -0.00000715, 0.00000621, 0.00001826, 0.00002602, 0.00002738, 0.00002169, 0.00000996, -0.00000522, -0.00002030, -0.00003161, -0.00003620, -0.00003271, -0.00002169, -0.00000558, 0.00001179, 0.00002614, 0.00003371, 0.00003225, 0.00002160, 0.00000378, -0.00001740, -0.00003721, -0.00005108, -0.00005568, -0.00004966, -0.00003406, -0.00001197, 0.00001207, 0.00003318, 0.00004721, 0.00005164, 0.00004610, 0.00003239, 0.00001398, -0.00000484, -0.00001994, -0.00002829, -0.00002855, -0.00002125, -0.00000858, 0.00000627, 0.00001987, 0.00002936, 0.00003301, 0.00003050, 0.00002285, 0.00001200, 0.00000030, -0.00001013, -0.00001779, -0.00002207, -0.00002324, -0.00002216, -0.00001998, -0.00001771, -0.00001597, -0.00001491, -0.00001419, -0.00001324, -0.00001151, -0.00000867, -0.00000478, -0.00000030, 0.00000405, 0.00000752, 0.00000949, 0.00000972, 0.00000836, 0.00000591, 0.00000312, 0.00000070, -0.00000078, -0.00000109, -0.00000035, 0.00000104, 0.00000254, 0.00000362, 0.00000394, 0.00000337, 0.00000204, 0.00000023, -0.00000171, -0.00000352, -0.00000505, -0.00000632, -0.00000744, -0.00000856, -0.00000972, -0.00001085, -0.00001174, },
		{ 0.00000240, 0.00000110, -0.00000109, -0.00000402, -0.00000734, -0.00001051, -0.00001298, -0.00001430, -0.00001428, -0.00001304, -0.00001097, -0.00000866, -0.00000667, -0.00000540, -0.00000493, -0.00000502, -0.00000520, -0.00000493, -0.00000379, -0.00000169, 0.00000109, 0.00000388, 0.00000586, 0.00000629, 0.00000477, 0.00000143, -0.00000301, -0.00000740, -0.00001045, -0.00001108, -0.00000880, -0.00000392, 0.00000240, 0.00000843, 0.00001225, 0.00001233, 0.00000795, -0.00000041, -0.00001115, -0.00002176, -0.00002944, -0.00003184, -0.00002764, -0.00001712, -0.00000212, 0.00001419, 0.00002803, 0.00003593, 0.00003557, 0.00002646, 0.00001020, -0.00000983, -0.00002912, -0.00004311, -0.00004825, -0.00004289, -0.00002779, -0.00000598, 0.00001781, 0.00003825, 0.00005058, 0.00005173, 0.00004106, 0.00002051, -0.00000578, -0.00003241, -0.00005392, -0.00006599, -0.00006634, -0.00005522, -0.00003524, -0.00001072, 0.00001331, 0.00003226, 0.00004292, 0.00004410, 0.00003672, 0.00002344, 0.00000794, -0.00000602, -0.00001544, -0.00001869, -0.00001574, -0.00000798, 0.00000224, 0.00001227, 0.00001983, 0.00002349, 0.00002290, 0.00001870, 0.00001224, 0.00000510, -0.00000129, -0.00000605, -0.00000894, -0.00001028, -0.00001079, -0.00001121, -0.00001209, -0.00001358, -0.00001542, -0.00001704, -0.00001781, -0.00001719, -0.00001500, -0.00001144, -0.00000707, -0.00000265, 0.00000106, 0.00000352, 0.00000450, 0.00000414, 0.00000285, 0.00000122, -0.00000021, -0.00000106, -0.00000118, -0.00000069, 0.00000017, 0.00000107, 0.00000178, 0.00000220, 0.00000238, 0.00000245, 0.00000255, 0.00000275, 0.00000296, 0.00000294, },
		{ -0.00000353, -0.00000230, -0.00000023, 0.00000248, 0.00000540, 0.00000791, 0.00000941, 0.00000942, 0.00000779, 0.00000472, 0.00000072, -0.00000349, -0.00000720, -0.00000987, -0.00001131, -0.00001166, -0.00001139, -0.00001106, -0.00001113, -0.00001181, -0.00001292, -0.00001394, -0.00001419, -0.00001302, -0.00001010, -0.00000560, -0.00000020, 0.00000501, 0.00000882, 0.00001022, 0.00000876, 0.00000475, -0.00000078, -0.00000626, -0.00001003, -0.00001076, -0.00000794, -0.00000207, 0.00000533, 0.00001207, 0.00001584, 0.00001485, 0.00000842, -0.00000272, -0.00001639, -0.00002946, -0.00003849, -0.00004066, -0.00003449, -0.00002042, -0.00000084, 0.00002032, 0.00003841, 0.00004914, 0.00004961, 0.00003911, 0.00001938, -0.00000565, -0.00003068, -0.00005019, -0.00005974, -0.00005694, -0.00004210, -0.00001819, 0.00000980, 0.00003593, 0.00005466, 0.00006201, 0.00005647, 0.00003926, 0.00001402, -0.00001406, -0.00003940, -0.00005726, -0.00006467, -0.00006101, -0.00004798, -0.00002900, -0.00000842, 0.00000959, 0.00002190, 0.00002702, 0.00002524, 0.00001836, 0.00000911, 0.00000034, -0.00000562, -0.00000753, -0.00000533, -0.00000007, 0.00000654, 0.00001263, 0.00001670, 0.00001794, 0.00001636, 0.00001268, 0.00000801, 0.00000353, 0.00000010, -0.00000187, -0.00000256, -0.00000253, -0.00000250, -0.00000314, -0.00000476, -0.00000732, -0.00001036, -0.00001322, -0.00001522, -0.00001587, -0.00001499, -0.00001279, -0.00000975, -0.00000649, -0.00000361, -0.00000152, -0.00000037, -0.00000009, -0.00000038, -0.00000094, -0.00000150, -0.00000193, -0.00000224, -0.00000253, -0.00000291, -0.00000337, -0.00000380, -0.00000395, },
		{ -0.00000025, -0.00000173, -0.00000400, -0.00000675, -0.00000946, -0.00001149, -0.00001220, -0.00001120, -0.00000842, -0.00000421, 0.00000073, 0.00000552, 0.00000931, 0.00001145, 0.00001172, 0.00001032, 0.00000778, 0.00000482, 0.00000212, 0.00000006, -0.00000134, -0.00000241, -0.00000371, -0.00000581, -0.00000898, -0.00001308, -0.00001749, -0.00002123, -0.00002319, -0.00002251, -0.00001888, -0.00001269, -0.00000504, 0.00000245, 0.00000812, 0.00001062, 0.00000943, 0.00000496, -0.00000137, -0.00000756, -0.00001148, -0.00001154, -0.00000713, 0.00000099, 0.00001088, 0.00001976, 0.00002470, 0.00002348, 0.00001522, 0.00000081, -0.00001714, -0.00003478, -0.00004785, -0.00005275, -0.00004746, -0.00003212, -0.00000923, 0.00001680, 0.00004062, 0.00005705, 0.00006230, 0.00005491, 0.00003614, 0.00000974, -0.00001881, -0.00004353, -0.00005923, -0.00006268, -0.00005331, -0.00003329, -0.00000705, 0.00001981, 0.00004169, 0.00005425, 0.00005533, 0.00004524, 0.00002660, 0.00000366, -0.00001881, -0.00003654, -0.00004665, -0.00004819, -0.00004211, -0.00003087, -0.00001771, -0.00000575, 0.00000265, 0.00000643, 0.00000587, 0.00000235, -0.00000216, -0.00000567, -0.00000677, -0.00000489, -0.00000043, 0.00000549, 0.00001139, 0.00001587, 0.00001797, 0.00001741, 0.00001460, 0.00001044, 0.00000605, 0.00000243, 0.00000017, -0.00000059, -0.00000022, 0.00000061, 0.00000118, 0.00000090, -0.00000046, -0.00000278, -0.00000564, -0.00000843, -0.00001061, -0.00001177, -0.00001181, -0.00001087, -0.00000927, -0.00000742, -0.00000562, -0.00000407, -0.00000280, -0.00000177, -0.00000088, -0.00000013, 0.00000037, 0.00000040, },
		{ -0.00000556, -0.00000330, -0.00000052, 0.00000242, 0.00000497, 0.00000649, 0.00000645, 0.00000456, 0.00000090, -0.00000399, -0.00000926, -0.00001390, -0.00001698, -0.00001787, -0.00001642, -0.00001298, -0.00000832, -0.00000338, 0.00000095, 0.00000407, 0.00000582, 0.00000649, 0.00000663, 0.00000687, 0.00000761, 0.00000887, 0.00001023, 0.00001088, 0.00000995, 0.00000674, 0.00000108, -0.00000650, -0.00001485, -0.00002232, -0.00002727, -0.00002842, -0.00002533, -0.00001854, -0.00000956, -0.00000052, 0.00000634, 0.00000931, 0.00000764, 0.00000192, -0.00000611, -0.00001386, -0.00001864, -0.00001836, -0.00001219, -0.00000093, 0.00001308, 0.00002641, 0.00003537, 0.00003696, 0.00002971, 0.00001421, -0.00000690, -0.00002937, -0.00004825, -0.00005908, -0.00005890, -0.00004707, -0.00002551, 0.00000166, 0.00002898, 0.00005083, 0.00006271, 0.00006223, 0.00004967, 0.00002787, 0.00000157, -0.00002364, -0.00004262, -0.00005175, -0.00004972, -0.00003770, -0.00001891, 0.00000212, 0.00002074, 0.00003314, 0.00003722, 0.00003286, 0.00002183, 0.00000718, -0.00000757, -0.00001933, -0.00002608, -0.00002729, -0.00002382, -0.00001756, -0.00001083, -0.00000571, -0.00000349, -0.00000445, -0.00000783, -0.00001218, -0.00001579, -0.00001723, -0.00001571, -0.00001126, -0.00000470, 0.00000264, 0.00000931, 0.00001407, 0.00001626, 0.00001584, 0.00001340, 0.00000988, 0.00000631, 0.00000354, 0.00000201, 0.00000172, 0.00000229, 0.00000312, 0.00000363, 0.00000337, 0.00000223, 0.00000032, -0.00000200, -0.00000433, -0.00000631, -0.00000777, -0.00000865, -0.00000905, -0.00000908, -0.00000882, -0.00000823, -0.00000719, },
		{ -0.00000407, -0.00000642, -0.00000891, -0.00001120, -0.00001280, -0.00001317, -0.00001189, -0.00000880, -0.00000418, 0.00000133, 0.00000674, 0.00001100, 0.00001317, 0.00001275, 0.00000974, 0.00000470, -0.00000137, -0.00000731, -0.00001205, -0.00001491, -0.00001573, -0.00001486, -0.00001299, -0.00001095, -0.00000936, -0.00000846, -0.00000800, -0.00000733, -0.00000568, -0.00000241, 0.00000263, 0.00000893, 0.00001540, 0.00002049, 0.00002269, 0.00002089, 0.00001482, 0.00000523, -0.00000617, -0.00001710, -0.00002516, -0.00002855, -0.00002649, -0.00001956, -0.00000957, 0.00000079, 0.00000867, 0.00001178, 0.00000906, 0.00000103, -0.00001026, -0.00002166, -0.00002975, -0.00003169, -0.00002603, -0.00001322, 0.00000438, 0.00002298, 0.00003822, 0.00004621, 0.00004452, 0.00003284, 0.00001319, -0.00001049, -0.00003321, -0.00005001, -0.00005716, -0.00005298, -0.00003828, -0.00001614, 0.00000878, 0.00003133, 0.00004700, 0.00005294, 0.00004852, 0.00003538, 0.00001694, -0.00000246, -0.00001864, -0.00002852, -0.00003071, -0.00002576, -0.00001583, -0.00000404, 0.00000638, 0.00001290, 0.00001424, 0.00001062, 0.00000352, -0.00000478, -0.00001192, -0.00001605, -0.00001632, -0.00001303, -0.00000750, -0.00000158, 0.00000284, 0.00000437, 0.00000247, -0.00000242, -0.00000907, -0.00001579, -0.00002091, -0.00002320, -0.00002215, -0.00001808, -0.00001193, -0.00000505, 0.00000122, 0.00000584, 0.00000831, 0.00000872, 0.00000760, 0.00000575, 0.00000398, 0.00000285, 0.00000263, 0.00000319, 0.00000421, 0.00000523, 0.00000587, 0.00000593, 0.00000538, 0.00000433, 0.00000298, 0.00000146, -0.00000018, -0.00000200, },
		{ 0.00000637, 0.00000744, 0.00000842, 0.00000907, 0.00000898, 0.00000775, 0.00000506, 0.00000090, -0.00000438, -0.00001003, -0.00001506, -0.00001843, -0.00001930, -0.00001731, -0.00001268, -0.00000618, 0.00000097, 0.00000744, 0.00001210, 0.00001423, 0.00001377, 0.00001123, 0.00000751, 0.00000366, 0.00000050, -0.00000156, -0.00000269, -0.00000347, -0.00000473, -0.00000713, -0.00001091, -0.00001567, -0.00002038, -0.00002364, -0.00002400, -0.00002050, -0.00001298, -0.00000234, 0.00000960, 0.00002042, 0.00002770, 0.00002962, 0.00002546, 0.00001595, 0.00000310, -0.00001016, -0.00002071, -0.00002601, -0.00002479, -0.00001740, -0.00000582, 0.00000681, 0.00001695, 0.00002159, 0.00001903, 0.00000937, -0.00000537, -0.00002175, -0.00003568, -0.00004345, -0.00004267, -0.00003289, -0.00001582, 0.00000505, 0.00002525, 0.00004035, 0.00004702, 0.00004384, 0.00003160, 0.00001312, -0.00000744, -0.00002553, -0.00003734, -0.00004056, -0.00003494, -0.00002220, -0.00000555, 0.00001115, 0.00002435, 0.00003158, 0.00003201, 0.00002647, 0.00001710, 0.00000671, -0.00000201, -0.00000718, -0.00000815, -0.00000557, -0.00000109, 0.00000319, 0.00000535, 0.00000422, -0.00000031, -0.00000727, -0.00001493, -0.00002125, -0.00002451, -0.00002374, -0.00001899, -0.00001133, -0.00000252, 0.00000539, 0.00001066, 0.00001219, 0.00000983, 0.00000433, -0.00000291, -0.00001019, -0.00001598, -0.00001923, -0.00001960, -0.00001745, -0.00001363, -0.00000925, -0.00000532, -0.00000253, -0.00000113, -0.00000094, -0.00000147, -0.00000216, -0.00000251, -0.00000226, -0.00000140, -0.00000009, 0.00000140, 0.00000285, 0.00000414, 0.00000529, },
		{ -0.00000732, -0.00000683, -0.00000631, -0.00000565, -0.00000456, -0.00000271, 0.00000015, 0.00000402, 0.00000852, 0.00001295, 0.00001638, 0.00001785, 0.00001669, 0.00001271, 0.00000632, -0.00000151, -0.00000942, -0.00001599, -0.00002004, -0.00002098, -0.00001887, -0.00001444, -0.00000886, -0.00000337, 0.00000100, 0.00000376, 0.00000502, 0.00000542, 0.00000586, 0.00000711, 0.00000952, 0.00001278, 0.00001596, 0.00001768, 0.00001657, 0.00001173, 0.00000311, -0.00000829, -0.00002054, -0.00003112, -0.00003752, -0.00003793, -0.00003175, -0.00001983, -0.00000446, 0.00001116, 0.00002363, 0.00003017, 0.00002933, 0.00002142, 0.00000847, -0.00000623, -0.00001888, -0.00002615, -0.00002602, -0.00001831, -0.00000481, 0.00001116, 0.00002554, 0.00003451, 0.00003546, 0.00002766, 0.00001248, -0.00000692, -0.00002631, -0.00004143, -0.00004893, -0.00004721, -0.00003675, -0.00001994, -0.00000048, 0.00001749, 0.00003044, 0.00003615, 0.00003419, 0.00002592, 0.00001398, 0.00000161, -0.00000823, -0.00001352, -0.00001362, -0.00000927, -0.00000229, 0.00000505, 0.00001068, 0.00001328, 0.00001259, 0.00000938, 0.00000514, 0.00000155, -0.00000004, 0.00000095, 0.00000416, 0.00000839, 0.00001193, 0.00001315, 0.00001090, 0.00000496, -0.00000391, -0.00001411, -0.00002359, -0.00003035, -0.00003298, -0.00003096, -0.00002480, -0.00001590, -0.00000616, 0.00000245, 0.00000837, 0.00001079, 0.00000972, 0.00000595, 0.00000077, -0.00000446, -0.00000856, -0.00001089, -0.00001134, -0.00001031, -0.00000852, -0.00000671, -0.00000546, -0.00000504, -0.00000539, -0.00000621, -0.00000709, -0.00000772, -0.00000793, -0.00000774, },
		{ -0.00000440, -0.00000522, -0.00000572, -0.00000593, -0.00000609, -0.00000660, -0.00000777, -0.00000970, -0.00001214, -0.00001450, -0.00001592, -0.00001559, -0.00001291, -0.00000780, -0.00000076, 0.00000712, 0.00001446, 0.00001982, 0.00002213, 0.00002094, 0.00001653, 0.00000987, 0.00000237, -0.00000451, -0.00000961, -0.00001241, -0.00001305, -0.00001230, -0.00001120, -0.00001071, -0.00001132, -0.00001284, -0.00001441, -0.00001469, -0.00001237, -0.00000660, 0.00000260, 0.00001414, 0.00002595, 0.00003542, 0.00004002, 0.00003797, 0.00002882, 0.00001369, -0.00000482, -0.00002314, -0.00003753, -0.00004494, -0.00004379, -0.00003439, -0.00001893, -0.00000103, 0.00001515, 0.00002584, 0.00002865, 0.00002311, 0.00001082, -0.00000492, -0.00001993, -0.00003017, -0.00003274, -0.00002660, -0.00001283, 0.00000559, 0.00002452, 0.00003966, 0.00004752, 0.00004628, 0.00003613, 0.00001922, -0.00000091, -0.00002017, -0.00003490, -0.00004263, -0.00004254, -0.00003552, -0.00002378, -0.00001023, 0.00000232, 0.00001176, 0.00001712, 0.00001862, 0.00001741, 0.00001509, 0.00001314, 0.00001250, 0.00001334, 0.00001506, 0.00001658, 0.00001680, 0.00001494, 0.00001093, 0.00000543, -0.00000034, -0.00000491, -0.00000706, -0.00000617, -0.00000245, 0.00000304, 0.00000867, 0.00001265, 0.00001351, 0.00001051, 0.00000383, -0.00000540, -0.00001543, -0.00002428, -0.00003023, -0.00003223, -0.00003011, -0.00002459, -0.00001707, -0.00000920, -0.00000252, 0.00000188, 0.00000363, 0.00000302, 0.00000083, -0.00000192, -0.00000428, -0.00000566, -0.00000587, -0.00000517, -0.00000404, -0.00000303, -0.00000256, -0.00000275, -0.00000346, },
		{ -0.00000357, -0.00000274, -0.00000266, -0.00000329, -0.00000426, -0.00000504, -0.00000508, -0.00000412, -0.00000228, -0.00000008, 0.00000163, 0.00000197, 0.00000031, -0.00000350, -0.00000897, -0.00001507, -0.00002041, -0.00002363, -0.00002372, -0.00002036, -0.00001401, -0.00000584, 0.00000258, 0.00000968, 0.00001426, 0.00001585, 0.00001476, 0.00001194, 0.00000869, 0.00000619, 0.00000509, 0.00000531, 0.00000602, 0.00000590, 0.00000359, -0.00000180, -0.00001022, -0.00002055, -0.00003069, -0.00003801, -0.00004000, -0.00003500, -0.00002276, -0.00000472, 0.00001616, 0.00003594, 0.00005056, 0.00005681, 0.00005308, 0.00003988, 0.00001979, -0.00000309, -0.00002401, -0.00003868, -0.00004424, -0.00003996, -0.00002736, -0.00000988, 0.00000802, 0.00002186, 0.00002829, 0.00002583, 0.00001524, -0.00000073, -0.00001800, -0.00003218, -0.00003961, -0.00003825, -0.00002809, -0.00001111, 0.00000921, 0.00002873, 0.00004362, 0.00005120, 0.00005042, 0.00004198, 0.00002803, 0.00001154, -0.00000440, -0.00001729, -0.00002562, -0.00002907, -0.00002828, -0.00002454, -0.00001922, -0.00001345, -0.00000782, -0.00000238, 0.00000317, 0.00000922, 0.00001588, 0.00002287, 0.00002939, 0.00003437, 0.00003674, 0.00003576, 0.00003131, 0.00002400, 0.00001508, 0.00000618, -0.00000112, -0.00000568, -0.00000706, -0.00000566, -0.00000261, 0.00000060, 0.00000245, 0.00000186, -0.00000151, -0.00000725, -0.00001424, -0.00002099, -0.00002604, -0.00002834, -0.00002751, -0.00002390, -0.00001845, -0.00001245, -0.00000715, -0.00000349, -0.00000184, -0.00000200, -0.00000333, -0.00000500, -0.00000624, -0.00000660, -0.00000602, -0.00000484, },
		{ -0.00000642, -0.00000722, -0.00000685, -0.00000541, -0.00000347, -0.00000184, -0.00000130, -0.00000227, -0.00000468, -0.00000790, -0.00001093, -0.00001267, -0.00001228, -0.00000947, -0.00000467, 0.00000108, 0.00000632, 0.00000960, 0.00000988, 0.00000682, 0.00000094, -0.00000652, -0.00001387, -0.00001948, -0.00002219, -0.00002159, -0.00001818, -0.00001315, -0.00000799, -0.00000408, -0.00000220, -0.00000228, -0.00000342, -0.00000417, -0.00000300, 0.00000108, 0.00000814, 0.00001707, 0.00002578, 0.00003161, 0.00003206, 0.00002553, 0.00001196, -0.00000699, -0.00002807, -0.00004707, -0.00005977, -0.00006289, -0.00005501, -0.00003696, -0.00001184, 0.00001565, 0.00004016, 0.00005688, 0.00006263, 0.00005662, 0.00004054, 0.00001825, -0.00000519, -0.00002464, -0.00003610, -0.00003757, -0.00002946, -0.00001440, 0.00000344, 0.00001942, 0.00002953, 0.00003130, 0.00002438, 0.00001048, -0.00000706, -0.00002413, -0.00003683, -0.00004236, -0.00003954, -0.00002903, -0.00001302, 0.00000537, 0.00002280, 0.00003646, 0.00004458, 0.00004661, 0.00004316, 0.00003562, 0.00002576, 0.00001521, 0.00000523, -0.00000346, -0.00001058, -0.00001609, -0.00001997, -0.00002206, -0.00002203, -0.00001944, -0.00001399, -0.00000569, 0.00000490, 0.00001674, 0.00002837, 0.00003818, 0.00004479, 0.00004733, 0.00004568, 0.00004042, 0.00003275, 0.00002413, 0.00001593, 0.00000912, 0.00000405, 0.00000046, -0.00000237, -0.00000533, -0.00000914, -0.00001408, -0.00001988, -0.00002576, -0.00003066, -0.00003354, -0.00003370, -0.00003101, -0.00002595, -0.00001955, -0.00001307, -0.00000769, -0.00000422, -0.00000290, -0.00000338, -0.00000487, },
		{ -0.00001164, -0.00000731, -0.00000584, -0.00000670, -0.00000875, -0.00001062, -0.00001109, -0.00000951, -0.00000596, -0.00000127, 0.00000324, 0.00000615, 0.00000636, 0.00000350, -0.00000198, -0.00000885, -0.00001538, -0.00001986, -0.00002105, -0.00001854, -0.00001286, -0.00000540, 0.00000201, 0.00000759, 0.00001007, 0.00000910, 0.00000526, -0.00000008, -0.00000522, -0.00000866, -0.00000955, -0.00000802, -0.00000511, -0.00000249, -0.00000189, -0.00000449, -0.00001045, -0.00001866, -0.00002693, -0.00003247, -0.00003268, -0.00002590, -0.00001212, 0.00000687, 0.00002763, 0.00004573, 0.00005681, 0.00005760, 0.00004683, 0.00002569, -0.00000226, -0.00003177, -0.00005700, -0.00007280, -0.00007587, -0.00006553, -0.00004393, -0.00001551, 0.00001390, 0.00003844, 0.00005348, 0.00005664, 0.00004819, 0.00003091, 0.00000936, -0.00001125, -0.00002627, -0.00003264, -0.00002951, -0.00001831, -0.00000226, 0.00001446, 0.00002777, 0.00003459, 0.00003351, 0.00002498, 0.00001109, -0.00000503, -0.00001996, -0.00003081, -0.00003571, -0.00003414, -0.00002682, -0.00001542, -0.00000204, 0.00001124, 0.00002277, 0.00003152, 0.00003707, 0.00003948, 0.00003906, 0.00003621, 0.00003127, 0.00002456, 0.00001642, 0.00000740, -0.00000173, -0.00000992, -0.00001603, -0.00001898, -0.00001802, -0.00001298, -0.00000431, 0.00000690, 0.00001915, 0.00003079, 0.00004037, 0.00004690, 0.00004998, 0.00004978, 0.00004690, 0.00004213, 0.00003617, 0.00002943, 0.00002203, 0.00001384, 0.00000466, -0.00000552, -0.00001637, -0.00002712, -0.00003667, -0.00004380, -0.00004748, -0.00004716, -0.00004300, -0.00003586, -0.00002716, -0.00001859, },
		{ -0.00006068, -0.00005653, -0.00004775, -0.00003633, -0.00002467, -0.00001506, -0.00000916, -0.00000761, -0.00000990, -0.00001452, -0.00001940, -0.00002246, -0.00002219, -0.00001805, -0.00001067, -0.00000163, 0.00000694, 0.00001292, 0.00001482, 0.00001218, 0.00000570, -0.00000296, -0.00001160, -0.00001813, -0.00002111, -0.00002012, -0.00001590, -0.00001001, -0.00000443, -0.00000087, -0.00000025, -0.00000243, -0.00000617, -0.00000958, -0.00001068, -0.00000811, -0.00000166, 0.00000750, 0.00001697, 0.00002372, 0.00002490, 0.00001874, 0.00000516, -0.00001390, -0.00003475, -0.00005273, -0.00006321, -0.00006282, -0.00005036, -0.00002726, 0.00000252, 0.00003328, 0.00005874, 0.00007346, 0.00007411, 0.00006021, 0.00003437, 0.00000167, -0.00003134, -0.00005817, -0.00007371, -0.00007538, -0.00006355, -0.00004141, -0.00001413, 0.00001235, 0.00003265, 0.00004313, 0.00004256, 0.00003224, 0.00001552, -0.00000310, -0.00001907, -0.00002883, -0.00003055, -0.00002436, -0.00001218, 0.00000287, 0.00001731, 0.00002808, 0.00003321, 0.00003209, 0.00002545, 0.00001507, 0.00000322, -0.00000785, -0.00001635, -0.00002122, -0.00002219, -0.00001958, -0.00001413, -0.00000673, 0.00000177, 0.00001064, 0.00001927, 0.00002711, 0.00003358, 0.00003810, 0.00004012, 0.00003924, 0.00003539, 0.00002890, 0.00002058, 0.00001162, 0.00000341, -0.00000270, -0.00000570, -0.00000506, -0.00000086, 0.00000627, 0.00001530, 0.00002508, 0.00003450, 0.00004270, 0.00004915, 0.00005353, 0.00005569, 0.00005547, 0.00005266, 0.00004693, 0.00003803, 0.00002590, 0.00001093, -0.00000596, -0.00002326, -0.00003912, -0.00005159, -0.00005907, },
		{ 0.00001154, -0.00001274, -0.00003663, -0.00005708, -0.00007134, -0.00007752, -0.00007508, -0.00006500, -0.00004963, -0.00003225, -0.00001641, -0.00000509, -0.00000013, -0.00000178, -0.00000872, -0.00001840, -0.00002775, -0.00003392, -0.00003500, -0.00003051, -0.00002142, -0.00000991, 0.00000129, 0.00000963, 0.00001343, 0.00001234, 0.00000731, 0.00000033, -0.00000621, -0.00001026, -0.00001075, -0.00000789, -0.00000313, 0.00000128, 0.00000303, 0.00000053, -0.00000649, -0.00001672, -0.00002748, -0.00003537, -0.00003719, -0.00003095, -0.00001656, 0.00000386, 0.00002630, 0.00004566, 0.00005694, 0.00005652, 0.00004317, 0.00001852, -0.00001307, -0.00004536, -0.00007161, -0.00008600, -0.00008507, -0.00006850, -0.00003931, -0.00000322, 0.00003259, 0.00006099, 0.00007650, 0.00007639, 0.00006126, 0.00003477, 0.00000279, -0.00002801, -0.00005159, -0.00006379, -0.00006318, -0.00005111, -0.00003127, -0.00000866, 0.00001157, 0.00002533, 0.00003032, 0.00002644, 0.00001556, 0.00000093, -0.00001372, -0.00002503, -0.00003075, -0.00003007, -0.00002363, -0.00001323, -0.00000124, 0.00000998, 0.00001856, 0.00002343, 0.00002438, 0.00002194, 0.00001711, 0.00001106, 0.00000486, -0.00000064, -0.00000487, -0.00000749, -0.00000831, -0.00000723, -0.00000422, 0.00000064, 0.00000712, 0.00001471, 0.00002269, 0.00003014, 0.00003610, 0.00003978, 0.00004069, 0.00003878, 0.00003450, 0.00002871, 0.00002250, 0.00001699, 0.00001312, 0.00001150, 0.00001240, 0.00001571, 0.00002110, 0.00002809, 0.00003617, 0.00004480, 0.00005335, 0.00006108, 0.00006704, 0.00007009, 0.00006903, 0.00006283, 0.00005090, 0.00003344, },
		{ 0.00008677, 0.00009276, 0.00009255, 0.00008437, 0.00006744, 0.00004234, 0.00001114, -0.00002276, -0.00005511, -0.00008152, -0.00009836, -0.00010345, -0.00009663, -0.00007982, -0.00005669, -0.00003190, -0.00001017, 0.00000475, 0.00001085, 0.00000816, -0.00000130, -0.00001415, -0.00002653, -0.00003510, -0.00003782, -0.00003441, -0.00002630, -0.00001612, -0.00000689, -0.00000110, 0.00000001, -0.00000318, -0.00000880, -0.00001408, -0.00001620, -0.00001324, -0.00000490, 0.00000730, 0.00002021, 0.00002990, 0.00003270, 0.00002631, 0.00001065, -0.00001191, -0.00003682, -0.00005836, -0.00007098, -0.00007067, -0.00005610, -0.00002914, 0.00000533, 0.00004042, 0.00006869, 0.00008383, 0.00008213, 0.00006337, 0.00003099, -0.00000863, -0.00004754, -0.00007794, -0.00009386, -0.00009244, -0.00007447, -0.00004412, -0.00000795, 0.00002659, 0.00005277, 0.00006602, 0.00006472, 0.00005039, 0.00002714, 0.00000056, -0.00002358, -0.00004060, -0.00004784, -0.00004501, -0.00003405, -0.00001842, -0.00000220, 0.00001087, 0.00001824, 0.00001890, 0.00001345, 0.00000374, -0.00000770, -0.00001832, -0.00002608, -0.00002984, -0.00002941, -0.00002542, -0.00001904, -0.00001160, -0.00000430, 0.00000198, 0.00000676, 0.00000989, 0.00001149, 0.00001177, 0.00001099, 0.00000942, 0.00000735, 0.00000515, 0.00000327, 0.00000221, 0.00000248, 0.00000443, 0.00000819, 0.00001359, 0.00002010, 0.00002698, 0.00003338, 0.00003851, 0.00004178, 0.00004297, 0.00004217, 0.00003979, 0.00003643, 0.00003276, 0.00002944, 0.00002703, 0.00002604, 0.00002687, 0.00002988, 0.00003532, 0.00004326, 0.00005342, 0.00006503, 0.00007678, },
		{ 0.00003737, 0.00004126, 0.00004948, 0.00006199, 0.00007778, 0.00009479, 0.00011004, 0.00012005, 0.00012144, 0.00011169, 0.00008982, 0.00005685, 0.00001584, -0.00002837, -0.00007003, -0.00010343, -0.00012408, -0.00012959, -0.00012019, -0.00009873, -0.00007004, -0.00003996, -0.00001406, 0.00000352, 0.00001086, 0.00000859, -0.00000052, -0.00001237, -0.00002268, -0.00002818, -0.00002743, -0.00002111, -0.00001180, -0.00000312, 0.00000138, -0.00000060, -0.00000940, -0.00002304, -0.00003770, -0.00004863, -0.00005148, -0.00004360, -0.00002494, 0.00000166, 0.00003088, 0.00005613, 0.00007104, 0.00007110, 0.00005489, 0.00002464, -0.00001404, -0.00005333, -0.00008484, -0.00010153, -0.00009933, -0.00007811, -0.00004185, 0.00000226, 0.00004530, 0.00007860, 0.00009558, 0.00009316, 0.00007233, 0.00003784, -0.00000294, -0.00004167, -0.00007086, -0.00008541, -0.00008356, -0.00006702, -0.00004037, -0.00000988, 0.00001797, 0.00003790, 0.00004684, 0.00004436, 0.00003248, 0.00001498, -0.00000365, -0.00001928, -0.00002899, -0.00003156, -0.00002753, -0.00001883, -0.00000814, 0.00000181, 0.00000886, 0.00001178, 0.00001044, 0.00000558, -0.00000144, -0.00000912, -0.00001612, -0.00002152, -0.00002490, -0.00002626, -0.00002590, -0.00002426, -0.00002174, -0.00001869, -0.00001534, -0.00001188, -0.00000849, -0.00000537, -0.00000273, -0.00000070, 0.00000069, 0.00000157, 0.00000226, 0.00000320, 0.00000487, 0.00000767, 0.00001181, 0.00001727, 0.00002373, 0.00003072, 0.00003764, 0.00004391, 0.00004904, 0.00005269, 0.00005468, 0.00005499, 0.00005370, 0.00005102, 0.00004733, 0.00004320, 0.00003945, 0.00003712, },
		{ 0.00008159, 0.00008126, 0.00007785, 0.00007192, 0.00006468, 0.00005792, 0.00005373, 0.00005416, 0.00006059, 0.00007335, 0.00009126, 0.00011160, 0.00013036, 0.00014293, 0.00014493, 0.00013322, 0.00010676, 0.00006712, 0.00001841, -0.00003327, -0.00008089, -0.00011787, -0.00013934, -0.00014324, -0.00013070, -0.00010580, -0.00007468, -0.00004414, -0.00002015, -0.00000650, -0.00000407, -0.00001073, -0.00002209, -0.00003271, -0.00003768, -0.00003395, -0.00002130, -0.00000240, 0.00001774, 0.00003316, 0.00003855, 0.00003076, 0.00000992, -0.00002038, -0.00005368, -0.00008210, -0.00009821, -0.00009686, -0.00007659, -0.00004024, 0.00000552, 0.00005154, 0.00008816, 0.00010739, 0.00010473, 0.00008027, 0.00003872, -0.00001155, -0.00006034, -0.00009781, -0.00011659, -0.00011331, -0.00008923, -0.00004984, -0.00000354, 0.00004021, 0.00007299, 0.00008906, 0.00008649, 0.00006723, 0.00003648, 0.00000132, -0.00003092, -0.00005424, -0.00006510, -0.00006293, -0.00004992, -0.00003021, -0.00000883, 0.00000957, 0.00002164, 0.00002589, 0.00002277, 0.00001425, 0.00000319, -0.00000747, -0.00001540, -0.00001928, -0.00001893, -0.00001519, -0.00000949, -0.00000344, 0.00000157, 0.00000464, 0.00000547, 0.00000425, 0.00000150, -0.00000214, -0.00000610, -0.00000998, -0.00001357, -0.00001680, -0.00001967, -0.00002214, -0.00002415, -0.00002562, -0.00002644, -0.00002658, -0.00002610, -0.00002514, -0.00002387, -0.00002247, -0.00002098, -0.00001935, -0.00001737, -0.00001475, -0.00001116, -0.00000635, -0.00000016, 0.00000738, 0.00001614, 0.00002582, 0.00003606, 0.00004645, 0.00005649, 0.00006565, 0.00007330, 0.00007881, },
		{ 0.00000876, 0.00002463, 0.00004217, 0.00006053, 0.00007849, 0.00009456, 0.00010720, 0.00011515, 0.00011777, 0.00011535, 0.00010924, 0.00010172, 0.00009561, 0.00009371, 0.00009798, 0.00010896, 0.00012528, 0.00014369, 0.00015945, 0.00016723, 0.00016229, 0.00014162, 0.00010495, 0.00005510, -0.00000222, -0.00005931, -0.00010798, -0.00014118, -0.00015455, -0.00014744, -0.00012313, -0.00008821, -0.00005115, -0.00002042, -0.00000255, -0.00000065, -0.00001366, -0.00003661, -0.00006188, -0.00008104, -0.00008703, -0.00007600, -0.00004841, -0.00000917, 0.00003346, 0.00006967, 0.00009046, 0.00008988, 0.00006655, 0.00002425, -0.00002875, -0.00008145, -0.00012247, -0.00014262, -0.00013699, -0.00010605, -0.00005568, 0.00000417, 0.00006154, 0.00010510, 0.00012651, 0.00012216, 0.00009378, 0.00004790, -0.00000564, -0.00005594, -0.00009337, -0.00011143, -0.00010801, -0.00008544, -0.00004974, -0.00000905, 0.00002824, 0.00005525, 0.00006792, 0.00006555, 0.00005057, 0.00002767, 0.00000254, -0.00001949, -0.00003452, -0.00004075, -0.00003851, -0.00002990, -0.00001802, -0.00000613, 0.00000315, 0.00000828, 0.00000900, 0.00000610, 0.00000107, -0.00000442, -0.00000894, -0.00001162, -0.00001221, -0.00001105, -0.00000879, -0.00000618, -0.00000385, -0.00000217, -0.00000125, -0.00000098, -0.00000117, -0.00000164, -0.00000231, -0.00000322, -0.00000448, -0.00000619, -0.00000844, -0.00001120, -0.00001438, -0.00001780, -0.00002131, -0.00002475, -0.00002805, -0.00003118, -0.00003410, -0.00003678, -0.00003909, -0.00004087, -0.00004186, -0.00004177, -0.00004034, -0.00003729, -0.00003241, -0.00002550, -0.00001639, -0.00000498, },
		{ -0.00006021, -0.00006410, -0.00006655, -0.00006692, -0.00006436, -0.00005797, -0.00004695, -0.00003082, -0.00000968, 0.00001564, 0.00004353, 0.00007181, 0.00009806, 0.00012008, 0.00013647, 0.00014698, 0.00015270, 0.00015587, 0.00015935, 0.00016583, 0.00017697, 0.00019259, 0.00021038, 0.00022593, 0.00023355, 0.00022741, 0.00020303, 0.00015864, 0.00009613, 0.00002128, -0.00005692, -0.00012772, -0.00018065, -0.00020772, -0.00020521, -0.00017475, -0.00012327, -0.00006178, -0.00000323, 0.00004015, 0.00005952, 0.00005132, 0.00001820, -0.00003155, -0.00008575, -0.00013100, -0.00015571, -0.00015271, -0.00012100, -0.00006603, 0.00000137, 0.00006736, 0.00011811, 0.00014281, 0.00013608, 0.00009917, 0.00003964, -0.00003025, -0.00009621, -0.00014497, -0.00016707, -0.00015872, -0.00012244, -0.00006631, -0.00000206, 0.00005753, 0.00010133, 0.00012201, 0.00011739, 0.00009043, 0.00004833, 0.00000064, -0.00004288, -0.00007427, -0.00008889, -0.00008594, -0.00006828, -0.00004135, -0.00001172, 0.00001442, 0.00003253, 0.00004042, 0.00003840, 0.00002877, 0.00001503, 0.00000087, -0.00001066, -0.00001772, -0.00001988, -0.00001793, -0.00001342, -0.00000816, -0.00000370, -0.00000099, -0.00000028, -0.00000125, -0.00000316, -0.00000524, -0.00000686, -0.00000770, -0.00000778, -0.00000734, -0.00000675, -0.00000632, -0.00000620, -0.00000639, -0.00000672, -0.00000695, -0.00000690, -0.00000646, -0.00000567, -0.00000467, -0.00000371, -0.00000301, -0.00000281, -0.00000325, -0.00000441, -0.00000632, -0.00000899, -0.00001240, -0.00001651, -0.00002125, -0.00002654, -0.00003223, -0.00003816, -0.00004413, -0.00004994, -0.00005538, },
		{ -0.00001209, -0.00001631, -0.00002130, -0.00002718, -0.00003417, -0.00004250, -0.00005231, -0.00006348, -0.00007546, -0.00008720, -0.00009717, -0.00010351, -0.00010435, -0.00009819, -0.00008427, -0.00006285, -0.00003522, -0.00000354, 0.00002964, 0.00006201, 0.00009205, 0.00011951, 0.00014549, 0.00017209, 0.00020168, 0.00023592, 0.00027472, 0.00031552, 0.00035310, 0.00038006, 0.00038810, 0.00036970, 0.00032013, 0.00023911, 0.00013180, 0.00000873, -0.00011542, -0.00022413, -0.00030194, -0.00033749, -0.00032600, -0.00027052, -0.00018182, -0.00007658, 0.00002557, 0.00010585, 0.00014993, 0.00015077, 0.00011005, 0.00003773, -0.00005001, -0.00013413, -0.00019671, -0.00022462, -0.00021213, -0.00016202, -0.00008473, 0.00000389, 0.00008611, 0.00014595, 0.00017251, 0.00016198, 0.00011817, 0.00005136, -0.00002408, -0.00009288, -0.00014204, -0.00016332, -0.00015459, -0.00011984, -0.00006787, -0.00001011, 0.00004197, 0.00007916, 0.00009619, 0.00009237, 0.00007121, 0.00003921, 0.00000411, -0.00002681, -0.00004826, -0.00005767, -0.00005538, -0.00004405, -0.00002775, -0.00001077, 0.00000329, 0.00001224, 0.00001549, 0.00001385, 0.00000906, 0.00000313, -0.00000215, -0.00000567, -0.00000709, -0.00000675, -0.00000541, -0.00000392, -0.00000295, -0.00000281, -0.00000344, -0.00000452, -0.00000558, -0.00000625, -0.00000633, -0.00000587, -0.00000511, -0.00000440, -0.00000405, -0.00000428, -0.00000511, -0.00000641, -0.00000792, -0.00000932, -0.00001035, -0.00001082, -0.00001067, -0.00000992, -0.00000871, -0.00000722, -0.00000566, -0.00000426, -0.00000323, -0.00000275, -0.00000299, -0.00000403, -0.00000591, -0.00000861, },
		{ -0.00000830, -0.00000763, -0.00000732, -0.00000733, -0.00000753, -0.00000771, -0.00000770, -0.00000748, -0.00000731, -0.00000775, -0.00000969, -0.00001415, -0.00002208, -0.00003404, -0.00004995, -0.00006891, -0.00008920, -0.00010850, -0.00012430, -0.00013437, -0.00013723, -0.00013242, -0.00012058, -0.00010307, -0.00008147, -0.00005685, -0.00002919, 0.00000299, 0.00004241, 0.00009236, 0.00015558, 0.00023283, 0.00032167, 0.00041568, 0.00050454, 0.00057508, 0.00061328, 0.00060688, 0.00054822, 0.00043657, 0.00027941, 0.00009228, -0.00010299, -0.00028149, -0.00041933, -0.00049790, -0.00050739, -0.00044890, -0.00033449, -0.00018530, -0.00002782, 0.00011090, 0.00020815, 0.00024959, 0.00023176, 0.00016229, 0.00005811, -0.00005821, -0.00016305, -0.00023636, -0.00026543, -0.00024695, -0.00018714, -0.00009997, -0.00000381, 0.00008243, 0.00014320, 0.00016911, 0.00015836, 0.00011648, 0.00005457, -0.00001341, -0.00007375, -0.00011569, -0.00013331, -0.00012615, -0.00009872, -0.00005891, -0.00001599, 0.00002149, 0.00004735, 0.00005872, 0.00005609, 0.00004271, 0.00002341, 0.00000328, -0.00001344, -0.00002416, -0.00002815, -0.00002635, -0.00002076, -0.00001373, -0.00000736, -0.00000296, -0.00000098, -0.00000104, -0.00000229, -0.00000377, -0.00000470, -0.00000471, -0.00000388, -0.00000261, -0.00000144, -0.00000083, -0.00000102, -0.00000198, -0.00000341, -0.00000490, -0.00000602, -0.00000650, -0.00000625, -0.00000540, -0.00000425, -0.00000315, -0.00000243, -0.00000230, -0.00000285, -0.00000401, -0.00000560, -0.00000737, -0.00000908, -0.00001050, -0.00001146, -0.00001188, -0.00001176, -0.00001117, -0.00001026, -0.00000923, },
		{ -0.00000837, -0.00000888, -0.00000901, -0.00000890, -0.00000877, -0.00000894, -0.00000967, -0.00001111, -0.00001319, -0.00001556, -0.00001763, -0.00001870, -0.00001812, -0.00001552, -0.00001105, -0.00000545, -0.00000003, 0.00000354, 0.00000355, -0.00000136, -0.00001180, -0.00002746, -0.00004708, -0.00006876, -0.00009038, -0.00011015, -0.00012706, -0.00014105, -0.00015284, -0.00016340, -0.00017311, -0.00018091, -0.00018365, -0.00017593, -0.00015073, -0.00010066, -0.00001985, 0.00009393, 0.00023747, 0.00040105, 0.00056859, 0.00071923, 0.00083015, 0.00088045, 0.00085512, 0.00074853, 0.00056663, 0.00032711, 0.00005752, -0.00020867, -0.00043723, -0.00059886, -0.00067420, -0.00065716, -0.00055604, -0.00039195, -0.00019509, 0.00000075, 0.00016407, 0.00027109, 0.00030949, 0.00027990, 0.00019499, 0.00007621, -0.00005076, -0.00016120, -0.00023584, -0.00026410, -0.00024525, -0.00018761, -0.00010605, -0.00001839, 0.00005839, 0.00011142, 0.00013398, 0.00012610, 0.00009373, 0.00004674, -0.00000358, -0.00004698, -0.00007623, -0.00008808, -0.00008333, -0.00006600, -0.00004189, -0.00001706, 0.00000351, 0.00001680, 0.00002204, 0.00002035, 0.00001414, 0.00000621, -0.00000099, -0.00000592, -0.00000806, -0.00000785, -0.00000630, -0.00000456, -0.00000354, -0.00000365, -0.00000481, -0.00000652, -0.00000811, -0.00000899, -0.00000884, -0.00000767, -0.00000581, -0.00000378, -0.00000209, -0.00000114, -0.00000108, -0.00000183, -0.00000309, -0.00000447, -0.00000560, -0.00000622, -0.00000624, -0.00000570, -0.00000481, -0.00000381, -0.00000296, -0.00000247, -0.00000246, -0.00000293, -0.00000384, -0.00000503, -0.00000630, -0.00000748, },
		{ -0.00000410, -0.00000454, -0.00000528, -0.00000616, -0.00000693, -0.00000734, -0.00000717, -0.00000632, -0.00000489, -0.00000322, -0.00000185, -0.00000141, -0.00000247, -0.00000538, -0.00001006, -0.00001599, -0.00002217, -0.00002738, -0.00003038, -0.00003029, -0.00002683, -0.00002051, -0.00001254, -0.00000464, 0.00000141, 0.00000416, 0.00000293, -0.00000214, -0.00001014, -0.00001983, -0.00003027, -0.00004138, -0.00005432, -0.00007146, -0.00009585, -0.00013011, -0.00017513, -0.00022869, -0.00028462, -0.00033264, -0.00035935, -0.00035019, -0.00029236, -0.00017794, -0.00000678, 0.00021164, 0.00045799, 0.00070475, 0.00091972, 0.00107092, 0.00113205, 0.00108744, 0.00093566, 0.00069075, 0.00038089, 0.00004443, -0.00027593, -0.00053991, -0.00071616, -0.00078713, -0.00075169, -0.00062474, -0.00043408, -0.00021519, -0.00000481, 0.00016524, 0.00027314, 0.00030985, 0.00027952, 0.00019735, 0.00008570, -0.00003076, -0.00012988, -0.00019577, -0.00022099, -0.00020680, -0.00016174, -0.00009890, -0.00003270, 0.00002413, 0.00006283, 0.00007970, 0.00007600, 0.00005683, 0.00002934, 0.00000091, -0.00002251, -0.00003735, -0.00004273, -0.00004003, -0.00003213, -0.00002230, -0.00001341, -0.00000722, -0.00000428, -0.00000404, -0.00000528, -0.00000665, -0.00000710, -0.00000617, -0.00000403, -0.00000133, 0.00000110, 0.00000250, 0.00000247, 0.00000100, -0.00000148, -0.00000434, -0.00000688, -0.00000856, -0.00000912, -0.00000858, -0.00000727, -0.00000561, -0.00000406, -0.00000298, -0.00000253, -0.00000270, -0.00000332, -0.00000415, -0.00000495, -0.00000551, -0.00000572, -0.00000559, -0.00000520, -0.00000470, -0.00000425, -0.00000401, },
		{ -0.00000473, -0.00000453, -0.00000418, -0.00000383, -0.00000366, -0.00000386, -0.00000459, -0.00000588, -0.00000762, -0.00000952, -0.00001113, -0.00001194, -0.00001152, -0.00000964, -0.00000639, -0.00000227, 0.00000190, 0.00000509, 0.00000632, 0.00000491, 0.00000067, -0.00000591, -0.00001380, -0.00002163, -0.00002799, -0.00003184, -0.00003281, -0.00003128, -0.00002828, -0.00002514, -0.00002297, -0.00002221, -0.00002234, -0.00002188, -0.00001887, -0.00001161, 0.00000034, 0.00001536, 0.00002928, 0.00003561, 0.00002657, -0.00000518, -0.00006433, -0.00015081, -0.00025814, -0.00037296, -0.00047600, -0.00054462, -0.00055648, -0.00049404, -0.00034877, -0.00012428, 0.00016237, 0.00048180, 0.00079581, 0.00106285, 0.00124461, 0.00131245, 0.00125265, 0.00106924, 0.00078393, 0.00043303, 0.00006176, -0.00028281, -0.00055925, -0.00073798, -0.00080544, -0.00076528, -0.00063666, -0.00044991, -0.00024068, -0.00004359, 0.00011334, 0.00021263, 0.00024891, 0.00022823, 0.00016535, 0.00007981, -0.00000837, -0.00008239, -0.00013129, -0.00015096, -0.00014367, -0.00011643, -0.00007863, -0.00003971, -0.00000716, 0.00001462, 0.00002453, 0.00002429, 0.00001736, 0.00000772, -0.00000124, -0.00000738, -0.00001005, -0.00000989, -0.00000826, -0.00000671, -0.00000635, -0.00000763, -0.00001024, -0.00001335, -0.00001591, -0.00001702, -0.00001621, -0.00001354, -0.00000958, -0.00000517, -0.00000123, 0.00000153, 0.00000271, 0.00000234, 0.00000077, -0.00000143, -0.00000368, -0.00000548, -0.00000655, -0.00000684, -0.00000648, -0.00000574, -0.00000491, -0.00000421, -0.00000379, -0.00000370, -0.00000387, -0.00000419, -0.00000452, -0.00000472, },
		{ -0.00000423, -0.00000432, -0.00000455, -0.00000483, -0.00000503, -0.00000499, -0.00000462, -0.00000387, -0.00000283, -0.00000173, -0.00000090, -0.00000072, -0.00000150, -0.00000340, -0.00000633, -0.00000989, -0.00001342, -0.00001611, -0.00001724, -0.00001631, -0.00001328, -0.00000859, -0.00000316, 0.00000186, 0.00000533, 0.00000647, 0.00000507, 0.00000157, -0.00000306, -0.00000763, -0.00001113, -0.00001317, -0.00001421, -0.00001549, -0.00001874, -0.00002549, -0.00003637, -0.00005045, -0.00006499, -0.00007562, -0.00007724, -0.00006536, -0.00003767, 0.00000449, 0.00005529, 0.00010452, 0.00013886, 0.00014423, 0.00010889, 0.00002674, -0.00010003, -0.00025948, -0.00043037, -0.00058452, -0.00069109, -0.00072186, -0.00065681, -0.00048885, -0.00022649, 0.00010593, 0.00047095, 0.00082319, 0.00111628, 0.00131024, 0.00137808, 0.00131024, 0.00111628, 0.00082319, 0.00047095, 0.00010593, -0.00022649, -0.00048885, -0.00065681, -0.00072186, -0.00069109, -0.00058452, -0.00043037, -0.00025948, -0.00010003, 0.00002674, 0.00010889, 0.00014423, 0.00013886, 0.00010452, 0.00005529, 0.00000449, -0.00003767, -0.00006536, -0.00007724, -0.00007562, -0.00006499, -0.00005045, -0.00003637, -0.00002549, -0.00001874, -0.00001549, -0.00001421, -0.00001317, -0.00001113, -0.00000763, -0.00000306, 0.00000157, 0.00000507, 0.00000647, 0.00000533, 0.00000186, -0.00000316, -0.00000859, -0.00001328, -0.00001631, -0.00001724, -0.00001611, -0.00001342, -0.00000989, -0.00000633, -0.00000340, -0.00000150, -0.00000072, -0.00000090, -0.00000173, -0.00000283, -0.00000387, -0.00000462, -0.00000499, -0.00000503, -0.00000483, -0.00000455, -0.00000432, },
		{ -0.00000473, -0.00000472, -0.00000452, -0.00000419, -0.00000387, -0.00000370, -0.00000379, -0.00000421, -0.00000491, -0.00000574, -0.00000648, -0.00000684, -0.00000655, -0.00000548, -0.00000368, -0.00000143, 0.00000077, 0.00000234, 0.00000271, 0.00000153, -0.00000123, -0.00000517, -0.00000958, -0.00001354, -0.00001621, -0.00001702, -0.00001591, -0.00001335, -0.00001024, -0.00000763, -0.00000635, -0.00000671, -0.00000826, -0.00000989, -0.00001005, -0.00000738, -0.00000124, 0.00000772, 0.00001736, 0.00002429, 0.00002453, 0.00001462, -0.00000716, -0.00003971, -0.00007863, -0.00011643, -0.00014367, -0.00015096, -0.00013129, -0.00008239, -0.00000837, 0.00007981, 0.00016535, 0.00022823, 0.00024891, 0.00021263, 0.00011334, -0.00004359, -0.00024068, -0.00044991, -0.00063666, -0.00076528, -0.00080544, -0.00073798, -0.00055925, -0.00028281, 0.00006176, 0.00043303, 0.00078393, 0.00106924, 0.00125265, 0.00131245, 0.00124461, 0.00106285, 0.00079581, 0.00048180, 0.00016237, -0.00012428, -0.00034877, -0.00049404, -0.00055648, -0.00054462, -0.00047600, -0.00037296, -0.00025814, -0.00015081, -0.00006433, -0.00000518, 0.00002657, 0.00003561, 0.00002928, 0.00001536, 0.00000034, -0.00001161, -0.00001887, -0.00002188, -0.00002234, -0.00002221, -0.00002297, -0.00002514, -0.00002828, -0.00003128, -0.00003281, -0.00003184, -0.00002799, -0.00002163, -0.00001380, -0.00000591, 0.00000067, 0.00000491, 0.00000632, 0.00000509, 0.00000190, -0.00000227, -0.00000639, -0.00000964, -0.00001152, -0.00001194, -0.00001113, -0.00000952, -0.00000762, -0.00000588, -0.00000459, -0.00000386, -0.00000366, -0.00000383, -0.00000418, -0.00000453, },
		{ -0.00000410, -0.00000401, -0.00000425, -0.00000470, -0.00000520, -0.00000559, -0.00000572, -0.00000551, -0.00000495, -0.00000415, -0.00000332, -0.00000270, -0.00000253, -0.00000298, -0.00000406, -0.00000561, -0.00000727, -0.00000858, -0.00000912, -0.00000856, -0.00000688, -0.00000434, -0.00000148, 0.00000100, 0.00000247, 0.00000250, 0.00000110, -0.00000133, -0.00000403, -0.00000617, -0.00000710, -0.00000665, -0.00000528, -0.00000404, -0.00000428, -0.00000722, -0.00001341, -0.00002230, -0.00003213, -0.00004003, -0.00004273, -0.00003735, -0.00002251, 0.00000091, 0.00002934, 0.00005683, 0.00007600, 0.00007970, 0.00006283, 0.00002413, -0.00003270, -0.00009890, -0.00016174, -0.00020680, -0.00022099, -0.00019577, -0.00012988, -0.00003076, 0.00008570, 0.00019735, 0.00027952, 0.00030985, 0.00027314, 0.00016524, -0.00000481, -0.00021519, -0.00043408, -0.00062474, -0.00075169, -0.00078713, -0.00071616, -0.00053991, -0.00027593, 0.00004443, 0.00038089, 0.00069075, 0.00093566, 0.00108744, 0.00113205, 0.00107092, 0.00091972, 0.00070475, 0.00045799, 0.00021164, -0.00000678, -0.00017794, -0.00029236, -0.00035019, -0.00035935, -0.00033264, -0.00028462, -0.00022869, -0.00017513, -0.00013011, -0.00009585, -0.00007146, -0.00005432, -0.00004138, -0.00003027, -0.00001983, -0.00001014, -0.00000214, 0.00000293, 0.00000416, 0.00000141, -0.00000464, -0.00001254, -0.00002051, -0.00002683, -0.00003029, -0.00003038, -0.00002738, -0.00002217, -0.00001599, -0.00001006, -0.00000538, -0.00000247, -0.00000141, -0.00000185, -0.00000322, -0.00000489, -0.00000632, -0.00000717, -0.00000734, -0.00000693, -0.00000616, -0.00000528, -0.00000454, },
		{ -0.00000837, -0.00000748, -0.00000630, -0.00000503, -0.00000384, -0.00000293, -0.00000246, -0.00000247, -0.00000296, -0.00000381, -0.00000481, -0.00000570, -0.00000624, -0.00000622, -0.00000560, -0.00000447, -0.00000309, -0.00000183, -0.00000108, -0.00000114, -0.00000209, -0.00000378, -0.00000581, -0.00000767, -0.00000884, -0.00000899, -0.00000811, -0.00000652, -0.00000481, -0.00000365, -0.00000354, -0.00000456, -0.00000630, -0.00000785, -0.00000806, -0.00000592, -0.00000099, 0.00000621, 0.00001414, 0.00002035, 0.00002204, 0.00001680, 0.00000351, -0.00001706, -0.00004189, -0.00006600, -0.00008333, -0.00008808, -0.00007623, -0.00004698, -0.00000358, 0.00004674, 0.00009373, 0.00012610, 0.00013398, 0.00011142, 0.00005839, -0.00001839, -0.00010605, -0.00018761, -0.00024525, -0.00026410, -0.00023584, -0.00016120, -0.00005076, 0.00007621, 0.00019499, 0.00027990, 0.00030949, 0.00027109, 0.00016407, 0.00000075, -0.00019509, -0.00039195, -0.00055604, -0.00065716, -0.00067420, -0.00059886, -0.00043723, -0.00020867, 0.00005752, 0.00032711, 0.00056663, 0.00074853, 0.00085512, 0.00088045, 0.00083015, 0.00071923, 0.00056859, 0.00040105, 0.00023747, 0.00009393, -0.00001985, -0.00010066, -0.00015073, -0.00017593, -0.00018365, -0.00018091, -0.00017311, -0.00016340, -0.00015284, -0.00014105, -0.00012706, -0.00011015, -0.00009038, -0.00006876, -0.00004708, -0.00002746, -0.00001180, -0.00000136, 0.00000355, 0.00000354, -0.00000003, -0.00000545, -0.00001105, -0.00001552, -0.00001812, -0.00001870, -0.00001763, -0.00001556, -0.00001319, -0.00001111, -0.00000967, -0.00000894, -0.00000877, -0.00000890, -0.00000901, -0.00000888, },
		{ -0.00000830, -0.00000923, -0.00001026, -0.00001117, -0.00001176, -0.00001188, -0.00001146, -0.00001050, -0.00000908, -0.00000737, -0.00000560, -0.00000401, -0.00000285, -0.00000230, -0.00000243, -0.00000315, -0.00000425, -0.00000540, -0.00000625, -0.00000650, -0.00000602, -0.00000490, -0.00000341, -0.00000198, -0.00000102, -0.00000083, -0.00000144, -0.00000261, -0.00000388, -0.00000471, -0.00000470, -0.00000377, -0.00000229, -0.00000104, -0.00000098, -0.00000296, -0.00000736, -0.00001373, -0.00002076, -0.00002635, -0.00002815, -0.00002416, -0.00001344, 0.00000328, 0.00002341, 0.00004271, 0.00005609, 0.00005872, 0.00004735, 0.00002149, -0.00001599, -0.00005891, -0.00009872, -0.00012615, -0.00013331, -0.00011569, -0.00007375, -0.00001341, 0.00005457, 0.00011648, 0.00015836, 0.00016911, 0.00014320, 0.00008243, -0.00000381, -0.00009997, -0.00018714, -0.00024695, -0.00026543, -0.00023636, -0.00016305, -0.00005821, 0.00005811, 0.00016229, 0.00023176, 0.00024959, 0.00020815, 0.00011090, -0.00002782, -0.00018530, -0.00033449, -0.00044890, -0.00050739, -0.00049790, -0.00041933, -0.00028149, -0.00010299, 0.00009228, 0.00027941, 0.00043657, 0.00054822, 0.00060688, 0.00061328, 0.00057508, 0.00050454, 0.00041568, 0.00032167, 0.00023283, 0.00015558, 0.00009236, 0.00004241, 0.00000299, -0.00002919, -0.00005685, -0.00008147, -0.00010307, -0.00012058, -0.00013242, -0.00013723, -0.00013437, -0.00012430, -0.00010850, -0.00008920, -0.00006891, -0.00004995, -0.00003404, -0.00002208, -0.00001415, -0.00000969, -0.00000775, -0.00000731, -0.00000748, -0.00000770, -0.00000771, -0.00000753, -0.00000733, -0.00000732, -0.00000763, },
		{ -0.00001209, -0.00000861, -0.00000591, -0.00000403, -0.00000299, -0.00000275, -0.00000323, -0.00000426, -0.00000566, -0.00000722, -0.00000871, -0.00000992, -0.00001067, -0.00001082, -0.00001035, -0.00000932, -0.00000792, -0.00000641, -0.00000511, -0.00000428, -0.00000405, -0.00000440, -0.00000511, -0.00000587, -0.00000633, -0.00000625, -0.00000558, -0.00000452, -0.00000344, -0.00000281, -0.00000295, -0.00000392, -0.00000541, -0.00000675, -0.00000709, -0.00000567, -0.00000215, 0.00000313, 0.00000906, 0.00001385, 0.00001549, 0.00001224, 0.00000329, -0.00001077, -0.00002775, -0.00004405, -0.00005538, -0.00005767, -0.00004826, -0.00002681, 0.00000411, 0.00003921, 0.00007121, 0.00009237, 0.00009619, 0.00007916, 0.00004197, -0.00001011, -0.00006787, -0.00011984, -0.00015459, -0.00016332, -0.00014204, -0.00009288, -0.00002408, 0.00005136, 0.00011817, 0.00016198, 0.00017251, 0.00014595, 0.00008611, 0.00000389, -0.00008473, -0.00016202, -0.00021213, -0.00022462, -0.00019671, -0.00013413, -0.00005001, 0.00003773, 0.00011005, 0.00015077, 0.00014993, 0.00010585, 0.00002557, -0.00007658, -0.00018182, -0.00027052, -0.00032600, -0.00033749, -0.00030194, -0.00022413, -0.00011542, 0.00000873, 0.00013180, 0.00023911, 0.00032013, 0.00036970, 0.00038810, 0.00038006, 0.00035310, 0.00031552, 0.00027472, 0.00023592, 0.00020168, 0.00017209, 0.00014549, 0.00011951, 0.00009205, 0.00006201, 0.00002964, -0.00000354, -0.00003522, -0.00006285, -0.00008427, -0.00009819, -0.00010435, -0.00010351, -0.00009717, -0.00008720, -0.00007546, -0.00006348, -0.00005231, -0.00004250, -0.00003417, -0.00002718, -0.00002130, -0.00001631, },
		{ -0.00006021, -0.00005538, -0.00004994, -0.00004413, -0.00003816, -0.00003223, -0.00002654, -0.00002125, -0.00001651, -0.00001240, -0.00000899, -0.00000632, -0.00000441, -0.00000325, -0.00000281, -0.00000301, -0.00000371, -0.00000467, -0.00000567, -0.00000646, -0.00000690, -0.00000695, -0.00000672, -0.00000639, -0.00000620, -0.00000632, -0.00000675, -0.00000734, -0.00000778, -0.00000770, -0.00000686, -0.00000524, -0.00000316, -0.00000125, -0.00000028, -0.00000099, -0.00000370, -0.00000816, -0.00001342, -0.00001793, -0.00001988, -0.00001772, -0.00001066, 0.00000087, 0.00001503, 0.00002877, 0.00003840, 0.00004042, 0.00003253, 0.00001442, -0.00001172, -0.00004135, -0.00006828, -0.00008594, -0.00008889, -0.00007427, -0.00004288, 0.00000064, 0.00004833, 0.00009043, 0.00011739, 0.00012201, 0.00010133, 0.00005753, -0.00000206, -0.00006631, -0.00012244, -0.00015872, -0.00016707, -0.00014497, -0.00009621, -0.00003025, 0.00003964, 0.00009917, 0.00013608, 0.00014281, 0.00011811, 0.00006736, 0.00000137, -0.00006603, -0.00012100, -0.00015271, -0.00015571, -0.00013100, -0.00008575, -0.00003155, 0.00001820, 0.00005132, 0.00005952, 0.00004015, -0.00000323, -0.00006178, -0.00012327, -0.00017475, -0.00020521, -0.00020772, -0.00018065, -0.00012772, -0.00005692, 0.00002128, 0.00009613, 0.00015864, 0.00020303, 0.00022741, 0.00023355, 0.00022593, 0.00021038, 0.00019259, 0.00017697, 0.00016583, 0.00015935, 0.00015587, 0.00015270, 0.00014698, 0.00013647, 0.00012008, 0.00009806, 0.00007181, 0.00004353, 0.00001564, -0.00000968, -0.00003082, -0.00004695, -0.00005797, -0.00006436, -0.00006692, -0.00006655, -0.00006410, },
		{ 0.00000876, -0.00000498, -0.00001639, -0.00002550, -0.00003241, -0.00003729, -0.00004034, -0.00004177, -0.00004186, -0.00004087, -0.00003909, -0.00003678, -0.00003410, -0.00003118, -0.00002805, -0.00002475, -0.00002131, -0.00001780, -0.00001438, -0.00001120, -0.00000844, -0.00000619, -0.00000448, -0.00000322, -0.00000231, -0.00000164, -0.00000117, -0.00000098, -0.00000125, -0.00000217, -0.00000385, -0.00000618, -0.00000879, -0.00001105, -0.00001221, -0.00001162, -0.00000894, -0.00000442, 0.00000107, 0.00000610, 0.00000900, 0.00000828, 0.00000315, -0.00000613, -0.00001802, -0.00002990, -0.00003851, -0.00004075, -0.00003452, -0.00001949, 0.00000254, 0.00002767, 0.00005057, 0.00006555, 0.00006792, 0.00005525, 0.00002824, -0.00000905, -0.00004974, -0.00008544, -0.00010801, -0.00011143, -0.00009337, -0.00005594, -0.00000564, 0.00004790, 0.00009378, 0.00012216, 0.00012651, 0.00010510, 0.00006154, 0.00000417, -0.00005568, -0.00010605, -0.00013699, -0.00014262, -0.00012247, -0.00008145, -0.00002875, 0.00002425, 0.00006655, 0.00008988, 0.00009046, 0.00006967, 0.00003346, -0.00000917, -0.00004841, -0.00007600, -0.00008703, -0.00008104, -0.00006188, -0.00003661, -0.00001366, -0.00000065, -0.00000255, -0.00002042, -0.00005115, -0.00008821, -0.00012313, -0.00014744, -0.00015455, -0.00014118, -0.00010798, -0.00005931, -0.00000222, 0.00005510, 0.00010495, 0.00014162, 0.00016229, 0.00016723, 0.00015945, 0.00014369, 0.00012528, 0.00010896, 0.00009798, 0.00009371, 0.00009561, 0.00010172, 0.00010924, 0.00011535, 0.00011777, 0.00011515, 0.00010720, 0.00009456, 0.00007849, 0.00006053, 0.00004217, 0.00002463, },
		{ 0.00008159, 0.00007881, 0.00007330, 0.00006565, 0.00005649, 0.00004645, 0.00003606, 0.00002582, 0.00001614, 0.00000738, -0.00000016, -0.00000635, -0.00001116, -0.00001475, -0.00001737, -0.00001935, -0.00002098, -0.00002247, -0.00002387, -0.00002514, -0.00002610, -0.00002658, -0.00002644, -0.00002562, -0.00002415, -0.00002214, -0.00001967, -0.00001680, -0.00001357, -0.00000998, -0.00000610, -0.00000214, 0.00000150, 0.00000425, 0.00000547, 0.00000464, 0.00000157, -0.00000344, -0.00000949, -0.00001519, -0.00001893, -0.00001928, -0.00001540, -0.00000747, 0.00000319, 0.00001425, 0.00002277, 0.00002589, 0.00002164, 0.00000957, -0.00000883, -0.00003021, -0.00004992, -0.00006293, -0.00006510, -0.00005424, -0.00003092, 0.00000132, 0.00003648, 0.00006723, 0.00008649, 0.00008906, 0.00007299, 0.00004021, -0.00000354, -0.00004984, -0.00008923, -0.00011331, -0.00011659, -0.00009781, -0.00006034, -0.00001155, 0.00003872, 0.00008027, 0.00010473, 0.00010739, 0.00008816, 0.00005154, 0.00000552, -0.00004024, -0.00007659, -0.00009686, -0.00009821, -0.00008210, -0.00005368, -0.00002038, 0.00000992, 0.00003076, 0.00003855, 0.00003316, 0.00001774, -0.00000240, -0.00002130, -0.00003395, -0.00003768, -0.00003271, -0.00002209, -0.00001073, -0.00000407, -0.00000650, -0.00002015, -0.00004414, -0.00007468, -0.00010580, -0.00013070, -0.00014324, -0.00013934, -0.00011787, -0.00008089, -0.00003327, 0.00001841, 0.00006712, 0.00010676, 0.00013322, 0.00014493, 0.00014293, 0.00013036, 0.00011160, 0.00009126, 0.00007335, 0.00006059, 0.00005416, 0.00005373, 0.00005792, 0.00006468, 0.00007192, 0.00007785, 0.00008126, },
		{ 0.00003737, 0.00003712, 0.00003945, 0.00004320, 0.00004733, 0.00005102, 0.00005370, 0.00005499, 0.00005468, 0.00005269, 0.00004904, 0.00004391, 0.00003764, 0.00003072, 0.00002373, 0.00001727, 0.00001181, 0.00000767, 0.00000487, 0.00000320, 0.00000226, 0.00000157, 0.00000069, -0.00000070, -0.00000273, -0.00000537, -0.00000849, -0.00001188, -0.00001534, -0.00001869, -0.00002174, -0.00002426, -0.00002590, -0.00002626, -0.00002490, -0.00002152, -0.00001612, -0.00000912, -0.00000144, 0.00000558, 0.00001044, 0.00001178, 0.00000886, 0.00000181, -0.00000814, -0.00001883, -0.00002753, -0.00003156, -0.00002899, -0.00001928, -0.00000365, 0.00001498, 0.00003248, 0.00004436, 0.00004684, 0.00003790, 0.00001797, -0.00000988, -0.00004037, -0.00006702, -0.00008356, -0.00008541, -0.00007086, -0.00004167, -0.00000294, 0.00003784, 0.00007233, 0.00009316, 0.00009558, 0.00007860, 0.00004530, 0.00000226, -0.00004185, -0.00007811, -0.00009933, -0.00010153, -0.00008484, -0.00005333, -0.00001404, 0.00002464, 0.00005489, 0.00007110, 0.00007104, 0.00005613, 0.00003088, 0.00000166, -0.00002494, -0.00004360, -0.00005148, -0.00004863, -0.00003770, -0.00002304, -0.00000940, -0.00000060, 0.00000138, -0.00000312, -0.00001180, -0.00002111, -0.00002743, -0.00002818, -0.00002268, -0.00001237, -0.00000052, 0.00000859, 0.00001086, 0.00000352, -0.00001406, -0.00003996, -0.00007004, -0.00009873, -0.00012019, -0.00012959, -0.00012408, -0.00010343, -0.00007003, -0.00002837, 0.00001584, 0.00005685, 0.00008982, 0.00011169, 0.00012144, 0.00012005, 0.00011004, 0.00009479, 0.00007778, 0.00006199, 0.00004948, 0.00004126, },
		{ 0.00008677, 0.00007678, 0.00006503, 0.00005342, 0.00004326, 0.00003532, 0.00002988, 0.00002687, 0.00002604, 0.00002703, 0.00002944, 0.00003276, 0.00003643, 0.00003979, 0.00004217, 0.00004297, 0.00004178, 0.00003851, 0.00003338, 0.00002698, 0.00002010, 0.00001359, 0.00000819, 0.00000443, 0.00000248, 0.00000221, 0.00000327, 0.00000515, 0.00000735, 0.00000942, 0.00001099, 0.00001177, 0.00001149, 0.00000989, 0.00000676, 0.00000198, -0.00000430, -0.00001160, -0.00001904, -0.00002542, -0.00002941, -0.00002984, -0.00002608, -0.00001832, -0.00000770, 0.00000374, 0.00001345, 0.00001890, 0.00001824, 0.00001087, -0.00000220, -0.00001842, -0.00003405, -0.00004501, -0.00004784, -0.00004060, -0.00002358, 0.00000056, 0.00002714, 0.00005039, 0.00006472, 0.00006602, 0.00005277, 0.00002659, -0.00000795, -0.00004412, -0.00007447, -0.00009244, -0.00009386, -0.00007794, -0.00004754, -0.00000863, 0.00003099, 0.00006337, 0.00008213, 0.00008383, 0.00006869, 0.00004042, 0.00000533, -0.00002914, -0.00005610, -0.00007067, -0.00007098, -0.00005836, -0.00003682, -0.00001191, 0.00001065, 0.00002631, 0.00003270, 0.00002990, 0.00002021, 0.00000730, -0.00000490, -0.00001324, -0.00001620, -0.00001408, -0.00000880, -0.00000318, 0.00000001, -0.00000110, -0.00000689, -0.00001612, -0.00002630, -0.00003441, -0.00003782, -0.00003510, -0.00002653, -0.00001415, -0.00000130, 0.00000816, 0.00001085, 0.00000475, -0.00001017, -0.00003190, -0.00005669, -0.00007982, -0.00009663, -0.00010345, -0.00009836, -0.00008152, -0.00005511, -0.00002276, 0.00001114, 0.00004234, 0.00006744, 0.00008437, 0.00009255, 0.00009276, },
		{ 0.00001154, 0.00003344, 0.00005090, 0.00006283, 0.00006903, 0.00007009, 0.00006704, 0.00006108, 0.00005335, 0.00004480, 0.00003617, 0.00002809, 0.00002110, 0.00001571, 0.00001240, 0.00001150, 0.00001312, 0.00001699, 0.00002250, 0.00002871, 0.00003450, 0.00003878, 0.00004069, 0.00003978, 0.00003610, 0.00003014, 0.00002269, 0.00001471, 0.00000712, 0.00000064, -0.00000422, -0.00000723, -0.00000831, -0.00000749, -0.00000487, -0.00000064, 0.00000486, 0.00001106, 0.00001711, 0.00002194, 0.00002438, 0.00002343, 0.00001856, 0.00000998, -0.00000124, -0.00001323, -0.00002363, -0.00003007, -0.00003075, -0.00002503, -0.00001372, 0.00000093, 0.00001556, 0.00002644, 0.00003032, 0.00002533, 0.00001157, -0.00000866, -0.00003127, -0.00005111, -0.00006318, -0.00006379, -0.00005159, -0.00002801, 0.00000279, 0.00003477, 0.00006126, 0.00007639, 0.00007650, 0.00006099, 0.00003259, -0.00000322, -0.00003931, -0.00006850, -0.00008507, -0.00008600, -0.00007161, -0.00004536, -0.00001307, 0.00001852, 0.00004317, 0.00005652, 0.00005694, 0.00004566, 0.00002630, 0.00000386, -0.00001656, -0.00003095, -0.00003719, -0.00003537, -0.00002748, -0.00001672, -0.00000649, 0.00000053, 0.00000303, 0.00000128, -0.00000313, -0.00000789, -0.00001075, -0.00001026, -0.00000621, 0.00000033, 0.00000731, 0.00001234, 0.00001343, 0.00000963, 0.00000129, -0.00000991, -0.00002142, -0.00003051, -0.00003500, -0.00003392, -0.00002775, -0.00001840, -0.00000872, -0.00000178, -0.00000013, -0.00000509, -0.00001641, -0.00003225, -0.00004963, -0.00006500, -0.00007508, -0.00007752, -0.00007134, -0.00005708, -0.00003663, -0.00001274, },
		{ -0.00006068, -0.00005907, -0.00005159, -0.00003912, -0.00002326, -0.00000596, 0.00001093, 0.00002590, 0.00003803, 0.00004693, 0.00005266, 0.00005547, 0.00005569, 0.00005353, 0.00004915, 0.00004270, 0.00003450, 0.00002508, 0.00001530, 0.00000627, -0.00000086, -0.00000506, -0.00000570, -0.00000270, 0.00000341, 0.00001162, 0.00002058, 0.00002890, 0.00003539, 0.00003924, 0.00004012, 0.00003810, 0.00003358, 0.00002711, 0.00001927, 0.00001064, 0.00000177, -0.00000673, -0.00001413, -0.00001958, -0.00002219, -0.00002122, -0.00001635, -0.00000785, 0.00000322, 0.00001507, 0.00002545, 0.00003209, 0.00003321, 0.00002808, 0.00001731, 0.00000287, -0.00001218, -0.00002436, -0.00003055, -0.00002883, -0.00001907, -0.00000310, 0.00001552, 0.00003224, 0.00004256, 0.00004313, 0.00003265, 0.00001235, -0.00001413, -0.00004141, -0.00006355, -0.00007538, -0.00007371, -0.00005817, -0.00003134, 0.00000167, 0.00003437, 0.00006021, 0.00007411, 0.00007346, 0.00005874, 0.00003328, 0.00000252, -0.00002726, -0.00005036, -0.00006282, -0.00006321, -0.00005273, -0.00003475, -0.00001390, 0.00000516, 0.00001874, 0.00002490, 0.00002372, 0.00001697, 0.00000750, -0.00000166, -0.00000811, -0.00001068, -0.00000958, -0.00000617, -0.00000243, -0.00000025, -0.00000087, -0.00000443, -0.00001001, -0.00001590, -0.00002012, -0.00002111, -0.00001813, -0.00001160, -0.00000296, 0.00000570, 0.00001218, 0.00001482, 0.00001292, 0.00000694, -0.00000163, -0.00001067, -0.00001805, -0.00002219, -0.00002246, -0.00001940, -0.00001452, -0.00000990, -0.00000761, -0.00000916, -0.00001506, -0.00002467, -0.00003633, -0.00004775, -0.00005653, },
		{ -0.00001164, -0.00001859, -0.00002716, -0.00003586, -0.00004300, -0.00004716, -0.00004748, -0.00004380, -0.00003667, -0.00002712, -0.00001637, -0.00000552, 0.00000466, 0.00001384, 0.00002203, 0.00002943, 0.00003617, 0.00004213, 0.00004690, 0.00004978, 0.00004998, 0.00004690, 0.00004037, 0.00003079, 0.00001915, 0.00000690, -0.00000431, -0.00001298, -0.00001802, -0.00001898, -0.00001603, -0.00000992, -0.00000173, 0.00000740, 0.00001642, 0.00002456, 0.00003127, 0.00003621, 0.00003906, 0.00003948, 0.00003707, 0.00003152, 0.00002277, 0.00001124, -0.00000204, -0.00001542, -0.00002682, -0.00003414, -0.00003571, -0.00003081, -0.00001996, -0.00000503, 0.00001109, 0.00002498, 0.00003351, 0.00003459, 0.00002777, 0.00001446, -0.00000226, -0.00001831, -0.00002951, -0.00003264, -0.00002627, -0.00001125, 0.00000936, 0.00003091, 0.00004819, 0.00005664, 0.00005348, 0.00003844, 0.00001390, -0.00001551, -0.00004393, -0.00006553, -0.00007587, -0.00007280, -0.00005700, -0.00003177, -0.00000226, 0.00002569, 0.00004683, 0.00005760, 0.00005681, 0.00004573, 0.00002763, 0.00000687, -0.00001212, -0.00002590, -0.00003268, -0.00003247, -0.00002693, -0.00001866, -0.00001045, -0.00000449, -0.00000189, -0.00000249, -0.00000511, -0.00000802, -0.00000955, -0.00000866, -0.00000522, -0.00000008, 0.00000526, 0.00000910, 0.00001007, 0.00000759, 0.00000201, -0.00000540, -0.00001286, -0.00001854, -0.00002105, -0.00001986, -0.00001538, -0.00000885, -0.00000198, 0.00000350, 0.00000636, 0.00000615, 0.00000324, -0.00000127, -0.00000596, -0.00000951, -0.00001109, -0.00001062, -0.00000875, -0.00000670, -0.00000584, -0.00000731, },
		{ -0.00000642, -0.00000487, -0.00000338, -0.00000290, -0.00000422, -0.00000769, -0.00001307, -0.00001955, -0.00002595, -0.00003101, -0.00003370, -0.00003354, -0.00003066, -0.00002576, -0.00001988, -0.00001408, -0.00000914, -0.00000533, -0.00000237, 0.00000046, 0.00000405, 0.00000912, 0.00001593, 0.00002413, 0.00003275, 0.00004042, 0.00004568, 0.00004733, 0.00004479, 0.00003818, 0.00002837, 0.00001674, 0.00000490, -0.00000569, -0.00001399, -0.00001944, -0.00002203, -0.00002206, -0.00001997, -0.00001609, -0.00001058, -0.00000346, 0.00000523, 0.00001521, 0.00002576, 0.00003562, 0.00004316, 0.00004661, 0.00004458, 0.00003646, 0.00002280, 0.00000537, -0.00001302, -0.00002903, -0.00003954, -0.00004236, -0.00003683, -0.00002413, -0.00000706, 0.00001048, 0.00002438, 0.00003130, 0.00002953, 0.00001942, 0.00000344, -0.00001440, -0.00002946, -0.00003757, -0.00003610, -0.00002464, -0.00000519, 0.00001825, 0.00004054, 0.00005662, 0.00006263, 0.00005688, 0.00004016, 0.00001565, -0.00001184, -0.00003696, -0.00005501, -0.00006289, -0.00005977, -0.00004707, -0.00002807, -0.00000699, 0.00001196, 0.00002553, 0.00003206, 0.00003161, 0.00002578, 0.00001707, 0.00000814, 0.00000108, -0.00000300, -0.00000417, -0.00000342, -0.00000228, -0.00000220, -0.00000408, -0.00000799, -0.00001315, -0.00001818, -0.00002159, -0.00002219, -0.00001948, -0.00001387, -0.00000652, 0.00000094, 0.00000682, 0.00000988, 0.00000960, 0.00000632, 0.00000108, -0.00000467, -0.00000947, -0.00001228, -0.00001267, -0.00001093, -0.00000790, -0.00000468, -0.00000227, -0.00000130, -0.00000184, -0.00000347, -0.00000541, -0.00000685, -0.00000722, },
		{ -0.00000357, -0.00000484, -0.00000602, -0.00000660, -0.00000624, -0.00000500, -0.00000333, -0.00000200, -0.00000184, -0.00000349, -0.00000715, -0.00001245, -0.00001845, -0.00002390, -0.00002751, -0.00002834, -0.00002604, -0.00002099, -0.00001424, -0.00000725, -0.00000151, 0.00000186, 0.00000245, 0.00000060, -0.00000261, -0.00000566, -0.00000706, -0.00000568, -0.00000112, 0.00000618, 0.00001508, 0.00002400, 0.00003131, 0.00003576, 0.00003674, 0.00003437, 0.00002939, 0.00002287, 0.00001588, 0.00000922, 0.00000317, -0.00000238, -0.00000782, -0.00001345, -0.00001922, -0.00002454, -0.00002828, -0.00002907, -0.00002562, -0.00001729, -0.00000440, 0.00001154, 0.00002803, 0.00004198, 0.00005042, 0.00005120, 0.00004362, 0.00002873, 0.00000921, -0.00001111, -0.00002809, -0.00003825, -0.00003961, -0.00003218, -0.00001800, -0.00000073, 0.00001524, 0.00002583, 0.00002829, 0.00002186, 0.00000802, -0.00000988, -0.00002736, -0.00003996, -0.00004424, -0.00003868, -0.00002401, -0.00000309, 0.00001979, 0.00003988, 0.00005308, 0.00005681, 0.00005056, 0.00003594, 0.00001616, -0.00000472, -0.00002276, -0.00003500, -0.00004000, -0.00003801, -0.00003069, -0.00002055, -0.00001022, -0.00000180, 0.00000359, 0.00000590, 0.00000602, 0.00000531, 0.00000509, 0.00000619, 0.00000869, 0.00001194, 0.00001476, 0.00001585, 0.00001426, 0.00000968, 0.00000258, -0.00000584, -0.00001401, -0.00002036, -0.00002372, -0.00002363, -0.00002041, -0.00001507, -0.00000897, -0.00000350, 0.00000031, 0.00000197, 0.00000163, -0.00000008, -0.00000228, -0.00000412, -0.00000508, -0.00000504, -0.00000426, -0.00000329, -0.00000266, -0.00000274, },
		{ -0.00000440, -0.00000346, -0.00000275, -0.00000256, -0.00000303, -0.00000404, -0.00000517, -0.00000587, -0.00000566, -0.00000428, -0.00000192, 0.00000083, 0.00000302, 0.00000363, 0.00000188, -0.00000252, -0.00000920, -0.00001707, -0.00002459, -0.00003011, -0.00003223, -0.00003023, -0.00002428, -0.00001543, -0.00000540, 0.00000383, 0.00001051, 0.00001351, 0.00001265, 0.00000867, 0.00000304, -0.00000245, -0.00000617, -0.00000706, -0.00000491, -0.00000034, 0.00000543, 0.00001093, 0.00001494, 0.00001680, 0.00001658, 0.00001506, 0.00001334, 0.00001250, 0.00001314, 0.00001509, 0.00001741, 0.00001862, 0.00001712, 0.00001176, 0.00000232, -0.00001023, -0.00002378, -0.00003552, -0.00004254, -0.00004263, -0.00003490, -0.00002017, -0.00000091, 0.00001922, 0.00003613, 0.00004628, 0.00004752, 0.00003966, 0.00002452, 0.00000559, -0.00001283, -0.00002660, -0.00003274, -0.00003017, -0.00001993, -0.00000492, 0.00001082, 0.00002311, 0.00002865, 0.00002584, 0.00001515, -0.00000103, -0.00001893, -0.00003439, -0.00004379, -0.00004494, -0.00003753, -0.00002314, -0.00000482, 0.00001369, 0.00002882, 0.00003797, 0.00004002, 0.00003542, 0.00002595, 0.00001414, 0.00000260, -0.00000660, -0.00001237, -0.00001469, -0.00001441, -0.00001284, -0.00001132, -0.00001071, -0.00001120, -0.00001230, -0.00001305, -0.00001241, -0.00000961, -0.00000451, 0.00000237, 0.00000987, 0.00001653, 0.00002094, 0.00002213, 0.00001982, 0.00001446, 0.00000712, -0.00000076, -0.00000780, -0.00001291, -0.00001559, -0.00001592, -0.00001450, -0.00001214, -0.00000970, -0.00000777, -0.00000660, -0.00000609, -0.00000593, -0.00000572, -0.00000522, },
		{ -0.00000732, -0.00000774, -0.00000793, -0.00000772, -0.00000709, -0.00000621, -0.00000539, -0.00000504, -0.00000546, -0.00000671, -0.00000852, -0.00001031, -0.00001134, -0.00001089, -0.00000856, -0.00000446, 0.00000077, 0.00000595, 0.00000972, 0.00001079, 0.00000837, 0.00000245, -0.00000616, -0.00001590, -0.00002480, -0.00003096, -0.00003298, -0.00003035, -0.00002359, -0.00001411, -0.00000391, 0.00000496, 0.00001090, 0.00001315, 0.00001193, 0.00000839, 0.00000416, 0.00000095, -0.00000004, 0.00000155, 0.00000514, 0.00000938, 0.00001259, 0.00001328, 0.00001068, 0.00000505, -0.00000229, -0.00000927, -0.00001362, -0.00001352, -0.00000823, 0.00000161, 0.00001398, 0.00002592, 0.00003419, 0.00003615, 0.00003044, 0.00001749, -0.00000048, -0.00001994, -0.00003675, -0.00004721, -0.00004893, -0.00004143, -0.00002631, -0.00000692, 0.00001248, 0.00002766, 0.00003546, 0.00003451, 0.00002554, 0.00001116, -0.00000481, -0.00001831, -0.00002602, -0.00002615, -0.00001888, -0.00000623, 0.00000847, 0.00002142, 0.00002933, 0.00003017, 0.00002363, 0.00001116, -0.00000446, -0.00001983, -0.00003175, -0.00003793, -0.00003752, -0.00003112, -0.00002054, -0.00000829, 0.00000311, 0.00001173, 0.00001657, 0.00001768, 0.00001596, 0.00001278, 0.00000952, 0.00000711, 0.00000586, 0.00000542, 0.00000502, 0.00000376, 0.00000100, -0.00000337, -0.00000886, -0.00001444, -0.00001887, -0.00002098, -0.00002004, -0.00001599, -0.00000942, -0.00000151, 0.00000632, 0.00001271, 0.00001669, 0.00001785, 0.00001638, 0.00001295, 0.00000852, 0.00000402, 0.00000015, -0.00000271, -0.00000456, -0.00000565, -0.00000631, -0.00000683, },
		{ 0.00000637, 0.00000529, 0.00000414, 0.00000285, 0.00000140, -0.00000009, -0.00000140, -0.00000226, -0.00000251, -0.00000216, -0.00000147, -0.00000094, -0.00000113, -0.00000253, -0.00000532, -0.00000925, -0.00001363, -0.00001745, -0.00001960, -0.00001923, -0.00001598, -0.00001019, -0.00000291, 0.00000433, 0.00000983, 0.00001219, 0.00001066, 0.00000539, -0.00000252, -0.00001133, -0.00001899, -0.00002374, -0.00002451, -0.00002125, -0.00001493, -0.00000727, -0.00000031, 0.00000422, 0.00000535, 0.00000319, -0.00000109, -0.00000557, -0.00000815, -0.00000718, -0.00000201, 0.00000671, 0.00001710, 0.00002647, 0.00003201, 0.00003158, 0.00002435, 0.00001115, -0.00000555, -0.00002220, -0.00003494, -0.00004056, -0.00003734, -0.00002553, -0.00000744, 0.00001312, 0.00003160, 0.00004384, 0.00004702, 0.00004035, 0.00002525, 0.00000505, -0.00001582, -0.00003289, -0.00004267, -0.00004345, -0.00003568, -0.00002175, -0.00000537, 0.00000937, 0.00001903, 0.00002159, 0.00001695, 0.00000681, -0.00000582, -0.00001740, -0.00002479, -0.00002601, -0.00002071, -0.00001016, 0.00000310, 0.00001595, 0.00002546, 0.00002962, 0.00002770, 0.00002042, 0.00000960, -0.00000234, -0.00001298, -0.00002050, -0.00002400, -0.00002364, -0.00002038, -0.00001567, -0.00001091, -0.00000713, -0.00000473, -0.00000347, -0.00000269, -0.00000156, 0.00000050, 0.00000366, 0.00000751, 0.00001123, 0.00001377, 0.00001423, 0.00001210, 0.00000744, 0.00000097, -0.00000618, -0.00001268, -0.00001731, -0.00001930, -0.00001843, -0.00001506, -0.00001003, -0.00000438, 0.00000090, 0.00000506, 0.00000775, 0.00000898, 0.00000907, 0.00000842, 0.00000744, },
		{ -0.00000407, -0.00000200, -0.00000018, 0.00000146, 0.00000298, 0.00000433, 0.00000538, 0.00000593, 0.00000587, 0.00000523, 0.00000421, 0.00000319, 0.00000263, 0.00000285, 0.00000398, 0.00000575, 0.00000760, 0.00000872, 0.00000831, 0.00000584, 0.00000122, -0.00000505, -0.00001193, -0.00001808, -0.00002215, -0.00002320, -0.00002091, -0.00001579, -0.00000907, -0.00000242, 0.00000247, 0.00000437, 0.00000284, -0.00000158, -0.00000750, -0.00001303, -0.00001632, -0.00001605, -0.00001192, -0.00000478, 0.00000352, 0.00001062, 0.00001424, 0.00001290, 0.00000638, -0.00000404, -0.00001583, -0.00002576, -0.00003071, -0.00002852, -0.00001864, -0.00000246, 0.00001694, 0.00003538, 0.00004852, 0.00005294, 0.00004700, 0.00003133, 0.00000878, -0.00001614, -0.00003828, -0.00005298, -0.00005716, -0.00005001, -0.00003321, -0.00001049, 0.00001319, 0.00003284, 0.00004452, 0.00004621, 0.00003822, 0.00002298, 0.00000438, -0.00001322, -0.00002603, -0.00003169, -0.00002975, -0.00002166, -0.00001026, 0.00000103, 0.00000906, 0.00001178, 0.00000867, 0.00000079, -0.00000957, -0.00001956, -0.00002649, -0.00002855, -0.00002516, -0.00001710, -0.00000617, 0.00000523, 0.00001482, 0.00002089, 0.00002269, 0.00002049, 0.00001540, 0.00000893, 0.00000263, -0.00000241, -0.00000568, -0.00000733, -0.00000800, -0.00000846, -0.00000936, -0.00001095, -0.00001299, -0.00001486, -0.00001573, -0.00001491, -0.00001205, -0.00000731, -0.00000137, 0.00000470, 0.00000974, 0.00001275, 0.00001317, 0.00001100, 0.00000674, 0.00000133, -0.00000418, -0.00000880, -0.00001189, -0.00001317, -0.00001280, -0.00001120, -0.00000891, -0.00000642, },
		{ -0.00000556, -0.00000719, -0.00000823, -0.00000882, -0.00000908, -0.00000905, -0.00000865, -0.00000777, -0.00000631, -0.00000433, -0.00000200, 0.00000032, 0.00000223, 0.00000337, 0.00000363, 0.00000312, 0.00000229, 0.00000172, 0.00000201, 0.00000354, 0.00000631, 0.00000988, 0.00001340, 0.00001584, 0.00001626, 0.00001407, 0.00000931, 0.00000264, -0.00000470, -0.00001126, -0.00001571, -0.00001723, -0.00001579, -0.00001218, -0.00000783, -0.00000445, -0.00000349, -0.00000571, -0.00001083, -0.00001756, -0.00002382, -0.00002729, -0.00002608, -0.00001933, -0.00000757, 0.00000718, 0.00002183, 0.00003286, 0.00003722, 0.00003314, 0.00002074, 0.00000212, -0.00001891, -0.00003770, -0.00004972, -0.00005175, -0.00004262, -0.00002364, 0.00000157, 0.00002787, 0.00004967, 0.00006223, 0.00006271, 0.00005083, 0.00002898, 0.00000166, -0.00002551, -0.00004707, -0.00005890, -0.00005908, -0.00004825, -0.00002937, -0.00000690, 0.00001421, 0.00002971, 0.00003696, 0.00003537, 0.00002641, 0.00001308, -0.00000093, -0.00001219, -0.00001836, -0.00001864, -0.00001386, -0.00000611, 0.00000192, 0.00000764, 0.00000931, 0.00000634, -0.00000052, -0.00000956, -0.00001854, -0.00002533, -0.00002842, -0.00002727, -0.00002232, -0.00001485, -0.00000650, 0.00000108, 0.00000674, 0.00000995, 0.00001088, 0.00001023, 0.00000887, 0.00000761, 0.00000687, 0.00000663, 0.00000649, 0.00000582, 0.00000407, 0.00000095, -0.00000338, -0.00000832, -0.00001298, -0.00001642, -0.00001787, -0.00001698, -0.00001390, -0.00000926, -0.00000399, 0.00000090, 0.00000456, 0.00000645, 0.00000649, 0.00000497, 0.00000242, -0.00000052, -0.00000330, },
		{ -0.00000025, 0.00000040, 0.00000037, -0.00000013, -0.00000088, -0.00000177, -0.00000280, -0.00000407, -0.00000562, -0.00000742, -0.00000927, -0.00001087, -0.00001181, -0.00001177, -0.00001061, -0.00000843, -0.00000564, -0.00000278, -0.00000046, 0.00000090, 0.00000118, 0.00000061, -0.00000022, -0.00000059, 0.00000017, 0.00000243, 0.00000605, 0.00001044, 0.00001460, 0.00001741, 0.00001797, 0.00001587, 0.00001139, 0.00000549, -0.00000043, -0.00000489, -0.00000677, -0.00000567, -0.00000216, 0.00000235, 0.00000587, 0.00000643, 0.00000265, -0.00000575, -0.00001771, -0.00003087, -0.00004211, -0.00004819, -0.00004665, -0.00003654, -0.00001881, 0.00000366, 0.00002660, 0.00004524, 0.00005533, 0.00005425, 0.00004169, 0.00001981, -0.00000705, -0.00003329, -0.00005331, -0.00006268, -0.00005923, -0.00004353, -0.00001881, 0.00000974, 0.00003614, 0.00005491, 0.00006230, 0.00005705, 0.00004062, 0.00001680, -0.00000923, -0.00003212, -0.00004746, -0.00005275, -0.00004785, -0.00003478, -0.00001714, 0.00000081, 0.00001522, 0.00002348, 0.00002470, 0.00001976, 0.00001088, 0.00000099, -0.00000713, -0.00001154, -0.00001148, -0.00000756, -0.00000137, 0.00000496, 0.00000943, 0.00001062, 0.00000812, 0.00000245, -0.00000504, -0.00001269, -0.00001888, -0.00002251, -0.00002319, -0.00002123, -0.00001749, -0.00001308, -0.00000898, -0.00000581, -0.00000371, -0.00000241, -0.00000134, 0.00000006, 0.00000212, 0.00000482, 0.00000778, 0.00001032, 0.00001172, 0.00001145, 0.00000931, 0.00000552, 0.00000073, -0.00000421, -0.00000842, -0.00001120, -0.00001220, -0.00001149, -0.00000946, -0.00000675, -0.00000400, -0.00000173, },
		{ -0.00000353, -0.00000395, -0.00000380, -0.00000337, -0.00000291, -0.00000253, -0.00000224, -0.00000193, -0.00000150, -0.00000094, -0.00000038, -0.00000009, -0.00000037, -0.00000152, -0.00000361, -0.00000649, -0.00000975, -0.00001279, -0.00001499, -0.00001587, -0.00001522, -0.00001322, -0.00001036, -0.00000732, -0.00000476, -0.00000314, -0.00000250, -0.00000253, -0.00000256, -0.00000187, 0.00000010, 0.00000353, 0.00000801, 0.00001268, 0.00001636, 0.00001794, 0.00001670, 0.00001263, 0.00000654, -0.00000007, -0.00000533, -0.00000753, -0.00000562, 0.00000034, 0.00000911, 0.00001836, 0.00002524, 0.00002702, 0.00002190, 0.00000959, -0.00000842, -0.00002900, -0.00004798, -0.00006101, -0.00006467, -0.00005726, -0.00003940, -0.00001406, 0.00001402, 0.00003926, 0.00005647, 0.00006201, 0.00005466, 0.00003593, 0.00000980, -0.00001819, -0.00004210, -0.00005694, -0.00005974, -0.00005019, -0.00003068, -0.00000565, 0.00001938, 0.00003911, 0.00004961, 0.00004914, 0.00003841, 0.00002032, -0.00000084, -0.00002042, -0.00003449, -0.00004066, -0.00003849, -0.00002946, -0.00001639, -0.00000272, 0.00000842, 0.00001485, 0.00001584, 0.00001207, 0.00000533, -0.00000207, -0.00000794, -0.00001076, -0.00001003, -0.00000626, -0.00000078, 0.00000475, 0.00000876, 0.00001022, 0.00000882, 0.00000501, -0.00000020, -0.00000560, -0.00001010, -0.00001302, -0.00001419, -0.00001394, -0.00001292, -0.00001181, -0.00001113, -0.00001106, -0.00001139, -0.00001166, -0.00001131, -0.00000987, -0.00000720, -0.00000349, 0.00000072, 0.00000472, 0.00000779, 0.00000942, 0.00000941, 0.00000791, 0.00000540, 0.00000248, -0.00000023, -0.00000230, },
		{ 0.00000240, 0.00000294, 0.00000296, 0.00000275, 0.00000255, 0.00000245, 0.00000238, 0.00000220, 0.00000178, 0.00000107, 0.00000017, -0.00000069, -0.00000118, -0.00000106, -0.00000021, 0.00000122, 0.00000285, 0.00000414, 0.00000450, 0.00000352, 0.00000106, -0.00000265, -0.00000707, -0.00001144, -0.00001500, -0.00001719, -0.00001781, -0.00001704, -0.00001542, -0.00001358, -0.00001209, -0.00001121, -0.00001079, -0.00001028, -0.00000894, -0.00000605, -0.00000129, 0.00000510, 0.00001224, 0.00001870, 0.00002290, 0.00002349, 0.00001983, 0.00001227, 0.00000224, -0.00000798, -0.00001574, -0.00001869, -0.00001544, -0.00000602, 0.00000794, 0.00002344, 0.00003672, 0.00004410, 0.00004292, 0.00003226, 0.00001331, -0.00001072, -0.00003524, -0.00005522, -0.00006634, -0.00006599, -0.00005392, -0.00003241, -0.00000578, 0.00002051, 0.00004106, 0.00005173, 0.00005058, 0.00003825, 0.00001781, -0.00000598, -0.00002779, -0.00004289, -0.00004825, -0.00004311, -0.00002912, -0.00000983, 0.00001020, 0.00002646, 0.00003557, 0.00003593, 0.00002803, 0.00001419, -0.00000212, -0.00001712, -0.00002764, -0.00003184, -0.00002944, -0.00002176, -0.00001115, -0.00000041, 0.00000795, 0.00001233, 0.00001225, 0.00000843, 0.00000240, -0.00000392, -0.00000880, -0.00001108, -0.00001045, -0.00000740, -0.00000301, 0.00000143, 0.00000477, 0.00000629, 0.00000586, 0.00000388, 0.00000109, -0.00000169, -0.00000379, -0.00000493, -0.00000520, -0.00000502, -0.00000493, -0.00000540, -0.00000667, -0.00000866, -0.00001097, -0.00001304, -0.00001428, -0.00001430, -0.00001298, -0.00001051, -0.00000734, -0.00000402, -0.00000109, 0.00000110, },
		{ -0.00001211, -0.00001174, -0.00001085, -0.00000972, -0.00000856, -0.00000744, -0.00000632, -0.00000505, -0.00000352, -0.00000171, 0.00000023, 0.00000204, 0.00000337, 0.00000394, 0.00000362, 0.00000254, 0.00000104, -0.00000035, -0.00000109, -0.00000078, 0.00000070, 0.00000312, 0.00000591, 0.00000836, 0.00000972, 0.00000949, 0.00000752, 0.00000405, -0.00000030, -0.00000478, -0.00000867, -0.00001151, -0.00001324, -0.00001419, -0.00001491, -0.00001597, -0.00001771, -0.00001998, -0.00002216, -0.00002324, -0.00002207, -0.00001779, -0.00001013, 0.00000030, 0.00001200, 0.00002285, 0.00003050, 0.00003301, 0.00002936, 0.00001987, 0.00000627, -0.00000858, -0.00002125, -0.00002855, -0.00002829, -0.00001994, -0.00000484, 0.00001398, 0.00003239, 0.00004610, 0.00005164, 0.00004721, 0.00003318, 0.00001207, -0.00001197, -0.00003406, -0.00004966, -0.00005568, -0.00005108, -0.00003721, -0.00001740, 0.00000378, 0.00002160, 0.00003225, 0.00003371, 0.00002614, 0.00001179, -0.00000558, -0.00002169, -0.00003271, -0.00003620, -0.00003161, -0.00002030, -0.00000522, 0.00000996, 0.00002169, 0.00002738, 0.00002602, 0.00001826, 0.00000621, -0.00000715, -0.00001876, -0.00002616, -0.00002808, -0.00002459, -0.00001700, -0.00000742, 0.00000177, 0.00000860, 0.00001185, 0.00001132, 0.00000773, 0.00000244, -0.00000299, -0.00000719, -0.00000931, -0.00000921, -0.00000731, -0.00000448, -0.00000166, 0.00000038, 0.00000124, 0.00000097, -0.00000004, -0.00000121, -0.00000199, -0.00000206, -0.00000140, -0.00000031, 0.00000075, 0.00000126, 0.00000086, -0.00000056, -0.00000283, -0.00000555, -0.00000822, -0.00001037, -0.00001169, },
		{ 0.00000283, 0.00000162, -0.00000003, -0.00000181, -0.00000349, -0.00000497, -0.00000630, -0.00000756, -0.00000884, -0.00001014, -0.00001131, -0.00001210, -0.00001219, -0.00001134, -0.00000949, -0.00000681, -0.00000372, -0.00000077, 0.00000145, 0.00000253, 0.00000234, 0.00000108, -0.00000072, -0.00000236, -0.00000317, -0.00000267, -0.00000077, 0.00000223, 0.00000570, 0.00000884, 0.00001091, 0.00001147, 0.00001047, 0.00000826, 0.00000549, 0.00000287, 0.00000092, -0.00000024, -0.00000092, -0.00000183, -0.00000384, -0.00000761, -0.00001335, -0.00002054, -0.00002797, -0.00003392, -0.00003654, -0.00003437, -0.00002681, -0.00001443, 0.00000102, 0.00001684, 0.00002995, 0.00003758, 0.00003794, 0.00003077, 0.00001744, 0.00000080, -0.00001546, -0.00002758, -0.00003267, -0.00002934, -0.00001814, -0.00000143, 0.00001710, 0.00003331, 0.00004353, 0.00004544, 0.00003860, 0.00002457, 0.00000651, -0.00001159, -0.00002585, -0.00003339, -0.00003305, -0.00002554, -0.00001329, 0.00000026, 0.00001149, 0.00001754, 0.00001694, 0.00000998, -0.00000141, -0.00001423, -0.00002515, -0.00003136, -0.00003126, -0.00002484, -0.00001362, -0.00000025, 0.00001214, 0.00002076, 0.00002377, 0.00002071, 0.00001253, 0.00000129, -0.00001035, -0.00001984, -0.00002528, -0.00002583, -0.00002183, -0.00001460, -0.00000603, 0.00000190, 0.00000760, 0.00001023, 0.00000974, 0.00000685, 0.00000270, -0.00000146, -0.00000463, -0.00000626, -0.00000634, -0.00000530, -0.00000378, -0.00000247, -0.00000180, -0.00000192, -0.00000263, -0.00000352, -0.00000412, -0.00000406, -0.00000320, -0.00000169, 0.00000014, 0.00000184, 0.00000299, 0.00000334, },
		{ -0.00000414, -0.00000286, -0.00000135, 0.00000007, 0.00000120, 0.00000194, 0.00000233, 0.00000249, 0.00000253, 0.00000250, 0.00000231, 0.00000177, 0.00000067, -0.00000115, -0.00000369, -0.00000670, -0.00000976, -0.00001230, -0.00001378, -0.00001384, -0.00001241, -0.00000976, -0.00000647, -0.00000327, -0.00000088, 0.00000023, -0.00000004, -0.00000141, -0.00000326, -0.00000484, -0.00000545, -0.00000470, -0.00000260, 0.00000040, 0.00000359, 0.00000617, 0.00000756, 0.00000753, 0.00000636, 0.00000472, 0.00000350, 0.00000348, 0.00000505, 0.00000798, 0.00001141, 0.00001400, 0.00001427, 0.00001103, 0.00000382, -0.00000685, -0.00001943, -0.00003163, -0.00004083, -0.00004475, -0.00004199, -0.00003252, -0.00001773, -0.00000028, 0.00001651, 0.00002930, 0.00003558, 0.00003421, 0.00002575, 0.00001231, -0.00000288, -0.00001625, -0.00002474, -0.00002646, -0.00002118, -0.00001030, 0.00000347, 0.00001682, 0.00002667, 0.00003090, 0.00002881, 0.00002130, 0.00001057, -0.00000052, -0.00000915, -0.00001333, -0.00001232, -0.00000685, 0.00000111, 0.00000896, 0.00001409, 0.00001465, 0.00001001, 0.00000092, -0.00001063, -0.00002198, -0.00003042, -0.00003391, -0.00003155, -0.00002385, -0.00001251, -0.00000009, 0.00001070, 0.00001759, 0.00001929, 0.00001573, 0.00000803, -0.00000184, -0.00001154, -0.00001899, -0.00002274, -0.00002234, -0.00001829, -0.00001184, -0.00000463, 0.00000174, 0.00000609, 0.00000789, 0.00000728, 0.00000494, 0.00000183, -0.00000111, -0.00000318, -0.00000410, -0.00000398, -0.00000324, -0.00000239, -0.00000190, -0.00000199, -0.00000265, -0.00000363, -0.00000455, -0.00000506, -0.00000493, },
		{ -0.00000205, -0.00000256, -0.00000317, -0.00000366, -0.00000385, -0.00000369, -0.00000326, -0.00000272, -0.00000222, -0.00000184, -0.00000153, -0.00000115, -0.00000050, 0.00000054, 0.00000199, 0.00000364, 0.00000513, 0.00000600, 0.00000581, 0.00000431, 0.00000155, -0.00000212, -0.00000608, -0.00000958, -0.00001193, -0.00001270, -0.00001184, -0.00000969, -0.00000693, -0.00000434, -0.00000260, -0.00000213, -0.00000289, -0.00000445, -0.00000609, -0.00000703, -0.00000669, -0.00000490, -0.00000197, 0.00000137, 0.00000414, 0.00000547, 0.00000485, 0.00000236, -0.00000128, -0.00000485, -0.00000694, -0.00000635, -0.00000248, 0.00000436, 0.00001297, 0.00002144, 0.00002753, 0.00002928, 0.00002551, 0.00001617, 0.00000253, -0.00001308, -0.00002772, -0.00003848, -0.00004317, -0.00004083, -0.00003196, -0.00001849, -0.00000325, 0.00001060, 0.00002033, 0.00002428, 0.00002216, 0.00001512, 0.00000536, -0.00000442, -0.00001169, -0.00001477, -0.00001316, -0.00000762, 0.00000011, 0.00000779, 0.00001333, 0.00001538, 0.00001361, 0.00000885, 0.00000277, -0.00000260, -0.00000543, -0.00000467, -0.00000030, 0.00000657, 0.00001403, 0.00001984, 0.00002207, 0.00001956, 0.00001234, 0.00000158, -0.00001062, -0.00002176, -0.00002953, -0.00003240, -0.00002991, -0.00002283, -0.00001294, -0.00000254, 0.00000604, 0.00001103, 0.00001160, 0.00000795, 0.00000127, -0.00000665, -0.00001389, -0.00001882, -0.00002050, -0.00001880, -0.00001438, -0.00000843, -0.00000234, 0.00000266, 0.00000574, 0.00000666, 0.00000571, 0.00000356, 0.00000100, -0.00000124, -0.00000271, -0.00000328, -0.00000312, -0.00000259, -0.00000208, -0.00000187, },
		{ -0.00000163, -0.00000239, -0.00000270, -0.00000277, -0.00000280, -0.00000288, -0.00000299, -0.00000301, -0.00000285, -0.00000246, -0.00000191, -0.00000137, -0.00000106, -0.00000114, -0.00000166, -0.00000248, -0.00000331, -0.00000375, -0.00000343, -0.00000217, -0.00000002, 0.00000267, 0.00000532, 0.00000728, 0.00000796, 0.00000708, 0.00000467, 0.00000120, -0.00000263, -0.00000599, -0.00000820, -0.00000890, -0.00000815, -0.00000643, -0.00000449, -0.00000315, -0.00000297, -0.00000414, -0.00000634, -0.00000882, -0.00001062, -0.00001084, -0.00000898, -0.00000509, 0.00000009, 0.00000534, 0.00000923, 0.00001050, 0.00000846, 0.00000325, -0.00000409, -0.00001182, -0.00001787, -0.00002036, -0.00001808, -0.00001089, 0.00000018, 0.00001311, 0.00002525, 0.00003400, 0.00003734, 0.00003434, 0.00002542, 0.00001226, -0.00000259, -0.00001630, -0.00002634, -0.00003110, -0.00003020, -0.00002449, -0.00001577, -0.00000634, 0.00000162, 0.00000657, 0.00000794, 0.00000618, 0.00000252, -0.00000140, -0.00000407, -0.00000451, -0.00000258, 0.00000105, 0.00000514, 0.00000823, 0.00000915, 0.00000740, 0.00000330, -0.00000205, -0.00000704, -0.00001003, -0.00000978, -0.00000590, 0.00000108, 0.00000971, 0.00001802, 0.00002395, 0.00002591, 0.00002317, 0.00001609, 0.00000599, -0.00000510, -0.00001495, -0.00002166, -0.00002410, -0.00002214, -0.00001669, -0.00000938, -0.00000218, 0.00000309, 0.00000521, 0.00000379, -0.00000066, -0.00000693, -0.00001342, -0.00001859, -0.00002128, -0.00002097, -0.00001786, -0.00001275, -0.00000678, -0.00000113, 0.00000323, 0.00000576, 0.00000641, 0.00000553, 0.00000371, 0.00000159, -0.00000028, },
		{ 0.00000373, 0.00000490, 0.00000485, 0.00000403, 0.00000286, 0.00000164, 0.00000051, -0.00000051, -0.00000147, -0.00000240, -0.00000325, -0.00000387, -0.00000410, -0.00000380, -0.00000295, -0.00000170, -0.00000033, 0.00000075, 0.00000120, 0.00000081, -0.00000039, -0.00000209, -0.00000381, -0.00000497, -0.00000508, -0.00000392, -0.00000160, 0.00000141, 0.00000442, 0.00000667, 0.00000756, 0.00000684, 0.00000468, 0.00000165, -0.00000145, -0.00000377, -0.00000474, -0.00000417, -0.00000241, -0.00000022, 0.00000143, 0.00000162, -0.00000015, -0.00000382, -0.00000870, -0.00001356, -0.00001700, -0.00001777, -0.00001518, -0.00000934, -0.00000126, 0.00000737, 0.00001455, 0.00001839, 0.00001769, 0.00001221, 0.00000287, -0.00000846, -0.00001930, -0.00002718, -0.00003012, -0.00002721, -0.00001875, -0.00000625, 0.00000793, 0.00002108, 0.00003080, 0.00003546, 0.00003452, 0.00002863, 0.00001937, 0.00000882, -0.00000094, -0.00000832, -0.00001252, -0.00001362, -0.00001240, -0.00001006, -0.00000778, -0.00000640, -0.00000616, -0.00000672, -0.00000733, -0.00000712, -0.00000548, -0.00000223, 0.00000215, 0.00000672, 0.00001028, 0.00001171, 0.00001037, 0.00000630, 0.00000031, -0.00000621, -0.00001162, -0.00001443, -0.00001368, -0.00000930, -0.00000206, 0.00000652, 0.00001455, 0.00002024, 0.00002232, 0.00002034, 0.00001482, 0.00000704, -0.00000121, -0.00000814, -0.00001235, -0.00001318, -0.00001085, -0.00000635, -0.00000118, 0.00000306, 0.00000506, 0.00000412, 0.00000031, -0.00000560, -0.00001234, -0.00001848, -0.00002276, -0.00002440, -0.00002319, -0.00001954, -0.00001430, -0.00000850, -0.00000313, 0.00000106, },
		{ -0.00002337, -0.00002052, -0.00001639, -0.00001179, -0.00000746, -0.00000387, -0.00000120, 0.00000059, 0.00000169, 0.00000230, 0.00000256, 0.00000251, 0.00000209, 0.00000126, 0.00000003, -0.00000144, -0.00000290, -0.00000398, -0.00000438, -0.00000391, -0.00000262, -0.00000081, 0.00000106, 0.00000245, 0.00000292, 0.00000227, 0.00000061, -0.00000161, -0.00000376, -0.00000516, -0.00000530, -0.00000401, -0.00000154, 0.00000151, 0.00000431, 0.00000608, 0.00000627, 0.00000478, 0.00000202, -0.00000120, -0.00000385, -0.00000502, -0.00000418, -0.00000137, 0.00000272, 0.00000689, 0.00000978, 0.00001016, 0.00000737, 0.00000152, -0.00000641, -0.00001478, -0.00002164, -0.00002519, -0.00002427, -0.00001866, -0.00000924, 0.00000217, 0.00001318, 0.00002138, 0.00002488, 0.00002279, 0.00001539, 0.00000413, -0.00000871, -0.00002052, -0.00002893, -0.00003232, -0.00003014, -0.00002298, -0.00001235, -0.00000033, 0.00001095, 0.00001977, 0.00002511, 0.00002678, 0.00002534, 0.00002181, 0.00001733, 0.00001285, 0.00000894, 0.00000568, 0.00000282, -0.00000004, -0.00000319, -0.00000663, -0.00001001, -0.00001268, -0.00001389, -0.00001308, -0.00001007, -0.00000521, 0.00000061, 0.00000617, 0.00001020, 0.00001168, 0.00001018, 0.00000597, 0.00000002, -0.00000626, -0.00001132, -0.00001387, -0.00001321, -0.00000942, -0.00000336, 0.00000359, 0.00000983, 0.00001394, 0.00001506, 0.00001310, 0.00000870, 0.00000310, -0.00000221, -0.00000587, -0.00000703, -0.00000550, -0.00000183, 0.00000286, 0.00000718, 0.00000983, 0.00000989, 0.00000708, 0.00000179, -0.00000503, -0.00001215, -0.00001830, -0.00002249, -0.00002418, },
		{ 0.00000445, -0.00000180, -0.00000808, -0.00001340, -0.00001709, -0.00001886, -0.00001885, -0.00001748, -0.00001533, -0.00001291, -0.00001060, -0.00000854, -0.00000672, -0.00000503, -0.00000337, -0.00000175, -0.00000030, 0.00000076, 0.00000120, 0.00000090, -0.00000010, -0.00000153, -0.00000297, -0.00000393, -0.00000404, -0.00000313, -0.00000135, 0.00000089, 0.00000298, 0.00000430, 0.00000440, 0.00000318, 0.00000092, -0.00000178, -0.00000411, -0.00000532, -0.00000493, -0.00000293, 0.00000023, 0.00000370, 0.00000644, 0.00000755, 0.00000652, 0.00000342, -0.00000105, -0.00000570, -0.00000916, -0.00001025, -0.00000829, -0.00000342, 0.00000346, 0.00001076, 0.00001661, 0.00001932, 0.00001779, 0.00001187, 0.00000243, -0.00000874, -0.00001937, -0.00002714, -0.00003029, -0.00002801, -0.00002061, -0.00000955, 0.00000294, 0.00001434, 0.00002237, 0.00002550, 0.00002325, 0.00001626, 0.00000606, -0.00000524, -0.00001550, -0.00002294, -0.00002651, -0.00002600, -0.00002196, -0.00001550, -0.00000792, -0.00000042, 0.00000615, 0.00001135, 0.00001515, 0.00001775, 0.00001942, 0.00002029, 0.00002033, 0.00001931, 0.00001701, 0.00001328, 0.00000828, 0.00000251, -0.00000324, -0.00000803, -0.00001097, -0.00001151, -0.00000960, -0.00000572, -0.00000084, 0.00000380, 0.00000702, 0.00000799, 0.00000643, 0.00000272, -0.00000220, -0.00000706, -0.00001060, -0.00001187, -0.00001050, -0.00000681, -0.00000168, 0.00000362, 0.00000779, 0.00000986, 0.00000939, 0.00000662, 0.00000239, -0.00000210, -0.00000559, -0.00000708, -0.00000612, -0.00000289, 0.00000184, 0.00000688, 0.00001099, 0.00001310, 0.00001263, 0.00000957, },
		{ 0.00000197, 0.00000635, 0.00000964, 0.00001110, 0.00001044, 0.00000786, 0.00000395, -0.00000049, -0.00000464, -0.00000790, -0.00000998, -0.00001090, -0.00001097, -0.00001060, -0.00001016, -0.00000987, -0.00000975, -0.00000963, -0.00000927, -0.00000847, -0.00000719, -0.00000556, -0.00000390, -0.00000259, -0.00000196, -0.00000214, -0.00000302, -0.00000422, -0.00000521, -0.00000545, -0.00000461, -0.00000262, 0.00000018, 0.00000318, 0.00000560, 0.00000675, 0.00000622, 0.00000405, 0.00000076, -0.00000277, -0.00000550, -0.00000653, -0.00000539, -0.00000218, 0.00000238, 0.00000712, 0.00001069, 0.00001191, 0.00001014, 0.00000550, -0.00000113, -0.00000821, -0.00001399, -0.00001682, -0.00001572, -0.00001053, -0.00000214, 0.00000776, 0.00001702, 0.00002346, 0.00002548, 0.00002238, 0.00001455, 0.00000344, -0.00000879, -0.00001971, -0.00002720, -0.00002985, -0.00002731, -0.00002026, -0.00001029, 0.00000054, 0.00001013, 0.00001679, 0.00001956, 0.00001832, 0.00001374, 0.00000701, -0.00000043, -0.00000727, -0.00001255, -0.00001578, -0.00001694, -0.00001632, -0.00001436, -0.00001150, -0.00000804, -0.00000413, 0.00000016, 0.00000475, 0.00000943, 0.00001382, 0.00001742, 0.00001964, 0.00002005, 0.00001846, 0.00001507, 0.00001045, 0.00000548, 0.00000110, -0.00000185, -0.00000289, -0.00000203, 0.00000024, 0.00000304, 0.00000537, 0.00000633, 0.00000542, 0.00000263, -0.00000148, -0.00000597, -0.00000973, -0.00001178, -0.00001156, -0.00000905, -0.00000481, 0.00000015, 0.00000466, 0.00000766, 0.00000846, 0.00000697, 0.00000367, -0.00000052, -0.00000444, -0.00000703, -0.00000759, -0.00000595, -0.00000249, },
		{ -0.00000020, -0.00000310, -0.00000477, -0.00000480, -0.00000323, -0.00000056, 0.00000246, 0.00000500, 0.00000638, 0.00000629, 0.00000479, 0.00000229, -0.00000056, -0.00000310, -0.00000485, -0.00000558, -0.00000541, -0.00000470, -0.00000393, -0.00000352, -0.00000372, -0.00000452, -0.00000567, -0.00000678, -0.00000747, -0.00000751, -0.00000690, -0.00000590, -0.00000494, -0.00000446, -0.00000476, -0.00000587, -0.00000751, -0.00000909, -0.00000996, -0.00000951, -0.00000745, -0.00000392, 0.00000050, 0.00000488, 0.00000821, 0.00000961, 0.00000864, 0.00000547, 0.00000085, -0.00000402, -0.00000777, -0.00000925, -0.00000781, -0.00000357, 0.00000262, 0.00000931, 0.00001478, 0.00001750, 0.00001650, 0.00001168, 0.00000387, -0.00000533, -0.00001389, -0.00001980, -0.00002159, -0.00001867, -0.00001148, -0.00000142, 0.00000943, 0.00001882, 0.00002480, 0.00002616, 0.00002264, 0.00001502, 0.00000487, -0.00000579, -0.00001499, -0.00002115, -0.00002346, -0.00002189, -0.00001720, -0.00001062, -0.00000360, 0.00000259, 0.00000703, 0.00000933, 0.00000956, 0.00000815, 0.00000571, 0.00000282, -0.00000007, -0.00000270, -0.00000500, -0.00000694, -0.00000848, -0.00000951, -0.00000983, -0.00000919, -0.00000743, -0.00000454, -0.00000076, 0.00000346, 0.00000748, 0.00001066, 0.00001251, 0.00001281, 0.00001172, 0.00000970, 0.00000746, 0.00000571, 0.00000500, 0.00000554, 0.00000711, 0.00000913, 0.00001081, 0.00001133, 0.00001014, 0.00000710, 0.00000253, -0.00000278, -0.00000780, -0.00001149, -0.00001307, -0.00001224, -0.00000923, -0.00000476, 0.00000011, 0.00000428, 0.00000686, 0.00000741, 0.00000598, 0.00000315, },
		{ -0.00000095, 0.00000255, 0.00000473, 0.00000534, 0.00000457, 0.00000295, 0.00000121, 0.00000002, -0.00000019, 0.00000065, 0.00000226, 0.00000405, 0.00000537, 0.00000566, 0.00000464, 0.00000241, -0.00000060, -0.00000373, -0.00000628, -0.00000772, -0.00000782, -0.00000668, -0.00000475, -0.00000263, -0.00000092, -0.00000003, -0.00000009, -0.00000091, -0.00000206, -0.00000307, -0.00000354, -0.00000332, -0.00000260, -0.00000182, -0.00000152, -0.00000223, -0.00000415, -0.00000713, -0.00001060, -0.00001369, -0.00001546, -0.00001515, -0.00001244, -0.00000759, -0.00000144, 0.00000477, 0.00000965, 0.00001207, 0.00001140, 0.00000776, 0.00000203, -0.00000436, -0.00000972, -0.00001256, -0.00001193, -0.00000771, -0.00000068, 0.00000765, 0.00001537, 0.00002062, 0.00002206, 0.00001914, 0.00001234, 0.00000300, -0.00000696, -0.00001545, -0.00002070, -0.00002166, -0.00001820, -0.00001110, -0.00000191, 0.00000749, 0.00001526, 0.00002001, 0.00002107, 0.00001852, 0.00001318, 0.00000628, -0.00000080, -0.00000685, -0.00001106, -0.00001313, -0.00001321, -0.00001180, -0.00000954, -0.00000702, -0.00000467, -0.00000270, -0.00000113, 0.00000015, 0.00000123, 0.00000213, 0.00000277, 0.00000302, 0.00000271, 0.00000178, 0.00000032, -0.00000141, -0.00000305, -0.00000420, -0.00000456, -0.00000401, -0.00000265, -0.00000083, 0.00000101, 0.00000242, 0.00000313, 0.00000308, 0.00000253, 0.00000193, 0.00000183, 0.00000268, 0.00000468, 0.00000768, 0.00001117, 0.00001438, 0.00001648, 0.00001674, 0.00001482, 0.00001082, 0.00000533, -0.00000072, -0.00000623, -0.00001022, -0.00001203, -0.00001148, -0.00000891, -0.00000508, },
		{ -0.00000390, -0.00000757, -0.00000931, -0.00000906, -0.00000720, -0.00000440, -0.00000145, 0.00000097, 0.00000246, 0.00000299, 0.00000285, 0.00000253, 0.00000256, 0.00000330, 0.00000479, 0.00000678, 0.00000873, 0.00000997, 0.00000995, 0.00000836, 0.00000528, 0.00000116, -0.00000327, -0.00000717, -0.00000983, -0.00001084, -0.00001020, -0.00000828, -0.00000574, -0.00000331, -0.00000157, -0.00000084, -0.00000102, -0.00000173, -0.00000237, -0.00000239, -0.00000149, 0.00000028, 0.00000247, 0.00000433, 0.00000504, 0.00000395, 0.00000083, -0.00000403, -0.00000975, -0.00001514, -0.00001889, -0.00001995, -0.00001780, -0.00001266, -0.00000546, 0.00000231, 0.00000898, 0.00001306, 0.00001364, 0.00001058, 0.00000466, -0.00000268, -0.00000961, -0.00001435, -0.00001563, -0.00001296, -0.00000677, 0.00000169, 0.00001060, 0.00001807, 0.00002248, 0.00002292, 0.00001932, 0.00001247, 0.00000383, -0.00000484, -0.00001189, -0.00001608, -0.00001686, -0.00001439, -0.00000950, -0.00000335, 0.00000276, 0.00000775, 0.00001093, 0.00001208, 0.00001141, 0.00000943, 0.00000676, 0.00000397, 0.00000143, -0.00000071, -0.00000247, -0.00000400, -0.00000539, -0.00000668, -0.00000777, -0.00000848, -0.00000860, -0.00000799, -0.00000662, -0.00000467, -0.00000243, -0.00000025, 0.00000153, 0.00000269, 0.00000320, 0.00000318, 0.00000288, 0.00000256, 0.00000241, 0.00000246, 0.00000260, 0.00000257, 0.00000212, 0.00000108, -0.00000054, -0.00000248, -0.00000426, -0.00000533, -0.00000514, -0.00000339, -0.00000009, 0.00000433, 0.00000916, 0.00001347, 0.00001635, 0.00001714, 0.00001559, 0.00001192, 0.00000679, 0.00000117, },
		{ 0.00001260, 0.00001333, 0.00001194, 0.00000874, 0.00000440, -0.00000021, -0.00000419, -0.00000690, -0.00000800, -0.00000758, -0.00000605, -0.00000399, -0.00000199, -0.00000045, 0.00000049, 0.00000102, 0.00000152, 0.00000245, 0.00000411, 0.00000658, 0.00000959, 0.00001256, 0.00001479, 0.00001559, 0.00001453, 0.00001155, 0.00000703, 0.00000171, -0.00000345, -0.00000757, -0.00000999, -0.00001052, -0.00000940, -0.00000727, -0.00000496, -0.00000326, -0.00000268, -0.00000327, -0.00000468, -0.00000619, -0.00000695, -0.00000631, -0.00000395, -0.00000013, 0.00000440, 0.00000853, 0.00001108, 0.00001110, 0.00000819, 0.00000262, -0.00000463, -0.00001211, -0.00001819, -0.00002147, -0.00002111, -0.00001710, -0.00001021, -0.00000192, 0.00000597, 0.00001174, 0.00001416, 0.00001277, 0.00000799, 0.00000100, -0.00000650, -0.00001272, -0.00001619, -0.00001608, -0.00001238, -0.00000584, 0.00000216, 0.00001001, 0.00001620, 0.00001965, 0.00001990, 0.00001717, 0.00001226, 0.00000627, 0.00000039, -0.00000441, -0.00000752, -0.00000878, -0.00000840, -0.00000688, -0.00000477, -0.00000257, -0.00000058, 0.00000107, 0.00000243, 0.00000361, 0.00000469, 0.00000567, 0.00000640, 0.00000668, 0.00000627, 0.00000503, 0.00000297, 0.00000029, -0.00000267, -0.00000547, -0.00000772, -0.00000910, -0.00000949, -0.00000896, -0.00000771, -0.00000604, -0.00000421, -0.00000242, -0.00000074, 0.00000087, 0.00000246, 0.00000408, 0.00000564, 0.00000696, 0.00000776, 0.00000773, 0.00000666, 0.00000451, 0.00000153, -0.00000182, -0.00000486, -0.00000691, -0.00000741, -0.00000608, -0.00000304, 0.00000120, 0.00000583, 0.00000991, },
		{ -0.00000490, -0.00000380, -0.00000125, 0.00000223, 0.00000583, 0.00000873, 0.00001023, 0.00000994, 0.00000786, 0.00000441, 0.00000026, -0.00000378, -0.00000702, -0.00000899, -0.00000959, -0.00000901, -0.00000770, -0.00000614, -0.00000473, -0.00000365, -0.00000278, -0.00000179, -0.00000026, 0.00000216, 0.00000557, 0.00000974, 0.00001410, 0.00001786, 0.00002018, 0.00002042, 0.00001833, 0.00001414, 0.00000853, 0.00000253, -0.00000276, -0.00000643, -0.00000798, -0.00000744, -0.00000540, -0.00000279, -0.00000067, 0.00000009, -0.00000094, -0.00000362, -0.00000724, -0.00001073, -0.00001293, -0.00001289, -0.00001022, -0.00000517, 0.00000133, 0.00000790, 0.00001301, 0.00001537, 0.00001426, 0.00000972, 0.00000260, -0.00000565, -0.00001328, -0.00001866, -0.00002063, -0.00001882, -0.00001368, -0.00000640, 0.00000140, 0.00000802, 0.00001207, 0.00001280, 0.00001021, 0.00000505, -0.00000142, -0.00000769, -0.00001240, -0.00001460, -0.00001393, -0.00001064, -0.00000551, 0.00000041, 0.00000603, 0.00001048, 0.00001322, 0.00001414, 0.00001346, 0.00001163, 0.00000918, 0.00000656, 0.00000407, 0.00000182, -0.00000019, -0.00000202, -0.00000368, -0.00000509, -0.00000610, -0.00000648, -0.00000603, -0.00000468, -0.00000250, 0.00000023, 0.00000310, 0.00000563, 0.00000738, 0.00000805, 0.00000757, 0.00000604, 0.00000378, 0.00000116, -0.00000141, -0.00000364, -0.00000535, -0.00000651, -0.00000715, -0.00000737, -0.00000720, -0.00000664, -0.00000563, -0.00000408, -0.00000199, 0.00000056, 0.00000331, 0.00000588, 0.00000782, 0.00000870, 0.00000828, 0.00000656, 0.00000382, 0.00000062, -0.00000233, -0.00000434, },
		{ 0.00000527, 0.00000414, 0.00000221, 0.00000003, -0.00000178, -0.00000263, -0.00000216, -0.00000032, 0.00000256, 0.00000590, 0.00000892, 0.00001089, 0.00001128, 0.00000991, 0.00000696, 0.00000292, -0.00000150, -0.00000558, -0.00000878, -0.00001080, -0.00001168, -0.00001169, -0.00001124, -0.00001067, -0.00001017, -0.00000966, -0.00000883, -0.00000722, -0.00000442, -0.00000024, 0.00000518, 0.00001127, 0.00001717, 0.00002185, 0.00002439, 0.00002419, 0.00002120, 0.00001592, 0.00000937, 0.00000284, -0.00000239, -0.00000538, -0.00000573, -0.00000371, -0.00000020, 0.00000353, 0.00000618, 0.00000670, 0.00000463, 0.00000020, -0.00000568, -0.00001165, -0.00001623, -0.00001817, -0.00001681, -0.00001224, -0.00000533, 0.00000252, 0.00000961, 0.00001441, 0.00001589, 0.00001376, 0.00000851, 0.00000133, -0.00000620, -0.00001248, -0.00001622, -0.00001676, -0.00001416, -0.00000914, -0.00000294, 0.00000307, 0.00000764, 0.00000993, 0.00000965, 0.00000710, 0.00000302, -0.00000163, -0.00000586, -0.00000890, -0.00001031, -0.00001005, -0.00000838, -0.00000575, -0.00000267, 0.00000043, 0.00000326, 0.00000567, 0.00000766, 0.00000923, 0.00001038, 0.00001104, 0.00001106, 0.00001030, 0.00000864, 0.00000613, 0.00000295, -0.00000052, -0.00000383, -0.00000649, -0.00000808, -0.00000840, -0.00000745, -0.00000546, -0.00000284, -0.00000007, 0.00000241, 0.00000428, 0.00000536, 0.00000567, 0.00000534, 0.00000457, 0.00000353, 0.00000235, 0.00000108, -0.00000027, -0.00000171, -0.00000317, -0.00000452, -0.00000555, -0.00000602, -0.00000572, -0.00000456, -0.00000264, -0.00000024, 0.00000220, 0.00000419, 0.00000529, },
		{ -0.00000242, -0.00000161, -0.00000039, 0.00000086, 0.00000173, 0.00000187, 0.00000117, -0.00000025, -0.00000204, -0.00000365, -0.00000453, -0.00000422, -0.00000254, 0.00000037, 0.00000407, 0.00000786, 0.00001100, 0.00001284, 0.00001299, 0.00001140, 0.00000836, 0.00000443, 0.00000025, -0.00000358, -0.00000668, -0.00000894, -0.00001052, -0.00001171, -0.00001283, -0.00001405, -0.00001529, -0.00001618, -0.00001615, -0.00001463, -0.00001116, -0.00000568, 0.00000143, 0.00000933, 0.00001684, 0.00002271, 0.00002588, 0.00002575, 0.00002238, 0.00001649, 0.00000933, 0.00000243, -0.00000277, -0.00000522, -0.00000456, -0.00000118, 0.00000387, 0.00000911, 0.00001299, 0.00001429, 0.00001239, 0.00000746, 0.00000039, -0.00000737, -0.00001418, -0.00001859, -0.00001968, -0.00001723, -0.00001184, -0.00000469, 0.00000267, 0.00000872, 0.00001228, 0.00001277, 0.00001032, 0.00000566, -0.00000004, -0.00000549, -0.00000956, -0.00001152, -0.00001116, -0.00000879, -0.00000512, -0.00000103, 0.00000260, 0.00000513, 0.00000624, 0.00000593, 0.00000450, 0.00000239, 0.00000007, -0.00000209, -0.00000383, -0.00000507, -0.00000582, -0.00000613, -0.00000601, -0.00000544, -0.00000435, -0.00000266, -0.00000036, 0.00000244, 0.00000548, 0.00000839, 0.00001073, 0.00001209, 0.00001220, 0.00001099, 0.00000860, 0.00000542, 0.00000192, -0.00000136, -0.00000398, -0.00000568, -0.00000636, -0.00000614, -0.00000526, -0.00000402, -0.00000268, -0.00000141, -0.00000029, 0.00000070, 0.00000161, 0.00000248, 0.00000326, 0.00000387, 0.00000415, 0.00000399, 0.00000329, 0.00000212, 0.00000065, -0.00000082, -0.00000198, -0.00000256, },
		{ 0.00000414, 0.00000406, 0.00000355, 0.00000285, 0.00000224, 0.00000195, 0.00000206, 0.00000249, 0.00000299, 0.00000323, 0.00000290, 0.00000181, -0.00000004, -0.00000237, -0.00000473, -0.00000656, -0.00000733, -0.00000671, -0.00000463, -0.00000136, 0.00000255, 0.00000639, 0.00000948, 0.00001127, 0.00001151, 0.00001029, 0.00000797, 0.00000508, 0.00000217, -0.00000036, -0.00000236, -0.00000397, -0.00000549, -0.00000732, -0.00000973, -0.00001273, -0.00001599, -0.00001887, -0.00002055, -0.00002024, -0.00001740, -0.00001196, -0.00000441, 0.00000426, 0.00001268, 0.00001943, 0.00002332, 0.00002371, 0.00002064, 0.00001488, 0.00000776, 0.00000092, -0.00000411, -0.00000624, -0.00000502, -0.00000084, 0.00000526, 0.00001176, 0.00001705, 0.00001978, 0.00001920, 0.00001528, 0.00000872, 0.00000080, -0.00000695, -0.00001307, -0.00001652, -0.00001686, -0.00001429, -0.00000962, -0.00000399, 0.00000135, 0.00000538, 0.00000745, 0.00000741, 0.00000560, 0.00000266, -0.00000059, -0.00000342, -0.00000529, -0.00000597, -0.00000553, -0.00000427, -0.00000256, -0.00000082, 0.00000068, 0.00000179, 0.00000250, 0.00000286, 0.00000295, 0.00000281, 0.00000244, 0.00000175, 0.00000070, -0.00000073, -0.00000242, -0.00000415, -0.00000561, -0.00000645, -0.00000639, -0.00000530, -0.00000321, -0.00000038, 0.00000276, 0.00000573, 0.00000807, 0.00000944, 0.00000970, 0.00000891, 0.00000734, 0.00000532, 0.00000322, 0.00000134, -0.00000017, -0.00000127, -0.00000205, -0.00000263, -0.00000311, -0.00000355, -0.00000389, -0.00000401, -0.00000377, -0.00000308, -0.00000194, -0.00000046, 0.00000114, 0.00000259, 0.00000364, },
		{ -0.00000475, -0.00000434, -0.00000333, -0.00000196, -0.00000050, 0.00000079, 0.00000176, 0.00000236, 0.00000271, 0.00000298, 0.00000337, 0.00000399, 0.00000482, 0.00000570, 0.00000634, 0.00000641, 0.00000567, 0.00000402, 0.00000160, -0.00000122, -0.00000393, -0.00000595, -0.00000684, -0.00000634, -0.00000452, -0.00000173, 0.00000144, 0.00000434, 0.00000639, 0.00000723, 0.00000682, 0.00000540, 0.00000346, 0.00000157, 0.00000016, -0.00000052, -0.00000059, -0.00000041, -0.00000054, -0.00000153, -0.00000373, -0.00000714, -0.00001135, -0.00001557, -0.00001882, -0.00002015, -0.00001888, -0.00001484, -0.00000845, -0.00000066, 0.00000718, 0.00001366, 0.00001759, 0.00001829, 0.00001574, 0.00001063, 0.00000420, -0.00000203, -0.00000660, -0.00000844, -0.00000708, -0.00000280, 0.00000350, 0.00001047, 0.00001666, 0.00002077, 0.00002199, 0.00002013, 0.00001563, 0.00000939, 0.00000262, -0.00000350, -0.00000804, -0.00001047, -0.00001075, -0.00000927, -0.00000668, -0.00000374, -0.00000111, 0.00000077, 0.00000172, 0.00000183, 0.00000136, 0.00000061, -0.00000013, -0.00000069, -0.00000104, -0.00000121, -0.00000130, -0.00000139, -0.00000147, -0.00000147, -0.00000128, -0.00000079, 0.00000005, 0.00000117, 0.00000239, 0.00000345, 0.00000406, 0.00000399, 0.00000314, 0.00000161, -0.00000036, -0.00000239, -0.00000407, -0.00000506, -0.00000516, -0.00000436, -0.00000283, -0.00000090, 0.00000106, 0.00000273, 0.00000387, 0.00000443, 0.00000449, 0.00000422, 0.00000384, 0.00000351, 0.00000331, 0.00000319, 0.00000302, 0.00000262, 0.00000187, 0.00000073, -0.00000070, -0.00000223, -0.00000358, -0.00000448, },
		{ 0.00000427, 0.00000359, 0.00000232, 0.00000070, -0.00000094, -0.00000230, -0.00000315, -0.00000341, -0.00000315, -0.00000252, -0.00000176, -0.00000103, -0.00000044, 0.00000003, 0.00000049, 0.00000110, 0.00000200, 0.00000321, 0.00000460, 0.00000595, 0.00000691, 0.00000718, 0.00000655, 0.00000502, 0.00000281, 0.00000033, -0.00000188, -0.00000332, -0.00000363, -0.00000272, -0.00000080, 0.00000165, 0.00000400, 0.00000563, 0.00000607, 0.00000515, 0.00000306, 0.00000026, -0.00000258, -0.00000477, -0.00000584, -0.00000559, -0.00000423, -0.00000229, -0.00000050, 0.00000042, -0.00000003, -0.00000200, -0.00000519, -0.00000889, -0.00001219, -0.00001415, -0.00001408, -0.00001172, -0.00000733, -0.00000167, 0.00000416, 0.00000898, 0.00001177, 0.00001197, 0.00000954, 0.00000506, -0.00000046, -0.00000575, -0.00000962, -0.00001116, -0.00000998, -0.00000625, -0.00000068, 0.00000571, 0.00001176, 0.00001646, 0.00001911, 0.00001947, 0.00001774, 0.00001444, 0.00001031, 0.00000609, 0.00000236, -0.00000050, -0.00000240, -0.00000345, -0.00000388, -0.00000395, -0.00000384, -0.00000364, -0.00000334, -0.00000286, -0.00000216, -0.00000122, -0.00000016, 0.00000084, 0.00000159, 0.00000188, 0.00000165, 0.00000093, -0.00000008, -0.00000111, -0.00000184, -0.00000204, -0.00000159, -0.00000055, 0.00000087, 0.00000233, 0.00000350, 0.00000411, 0.00000401, 0.00000326, 0.00000204, 0.00000065, -0.00000059, -0.00000143, -0.00000176, -0.00000161, -0.00000114, -0.00000059, -0.00000018, -0.00000004, -0.00000021, -0.00000055, -0.00000089, -0.00000099, -0.00000068, 0.00000009, 0.00000121, 0.00000247, 0.00000358, 0.00000425, },
		{ -0.00000477, -0.00000396, -0.00000249, -0.00000067, 0.00000112, 0.00000254, 0.00000333, 0.00000341, 0.00000286, 0.00000190, 0.00000080, -0.00000017, -0.00000083, -0.00000112, -0.00000110, -0.00000091, -0.00000071, -0.00000059, -0.00000057, -0.00000056, -0.00000041, 0.00000005, 0.00000089, 0.00000209, 0.00000346, 0.00000472, 0.00000552, 0.00000560, 0.00000482, 0.00000329, 0.00000131, -0.00000063, -0.00000200, -0.00000234, -0.00000143, 0.00000068, 0.00000361, 0.00000670, 0.00000924, 0.00001052, 0.00001011, 0.00000790, 0.00000420, -0.00000037, -0.00000498, -0.00000881, -0.00001122, -0.00001190, -0.00001096, -0.00000885, -0.00000625, -0.00000392, -0.00000243, -0.00000208, -0.00000281, -0.00000419, -0.00000560, -0.00000637, -0.00000600, -0.00000429, -0.00000141, 0.00000211, 0.00000548, 0.00000792, 0.00000877, 0.00000771, 0.00000482, 0.00000062, -0.00000410, -0.00000844, -0.00001153, -0.00001279, -0.00001200, -0.00000932, -0.00000526, -0.00000050, 0.00000421, 0.00000826, 0.00001125, 0.00001304, 0.00001371, 0.00001350, 0.00001271, 0.00001156, 0.00001021, 0.00000870, 0.00000698, 0.00000501, 0.00000280, 0.00000045, -0.00000182, -0.00000374, -0.00000501, -0.00000543, -0.00000494, -0.00000367, -0.00000190, -0.00000001, 0.00000161, 0.00000263, 0.00000291, 0.00000247, 0.00000152, 0.00000039, -0.00000059, -0.00000114, -0.00000114, -0.00000060, 0.00000030, 0.00000130, 0.00000214, 0.00000263, 0.00000271, 0.00000244, 0.00000199, 0.00000159, 0.00000139, 0.00000149, 0.00000183, 0.00000224, 0.00000249, 0.00000235, 0.00000169, 0.00000050, -0.00000104, -0.00000266, -0.00000401, -0.00000478, },
		{ 0.00000446, 0.00000268, 0.00000036, -0.00000211, -0.00000431, -0.00000584, -0.00000648, -0.00000619, -0.00000512, -0.00000357, -0.00000190, -0.00000045, 0.00000056, 0.00000103, 0.00000105, 0.00000077, 0.00000041, 0.00000017, 0.00000012, 0.00000027, 0.00000052, 0.00000073, 0.00000079, 0.00000067, 0.00000042, 0.00000021, 0.00000020, 0.00000056, 0.00000132, 0.00000241, 0.00000357, 0.00000448, 0.00000480, 0.00000429, 0.00000290, 0.00000083, -0.00000151, -0.00000355, -0.00000470, -0.00000449, -0.00000272, 0.00000047, 0.00000461, 0.00000896, 0.00001267, 0.00001495, 0.00001526, 0.00001342, 0.00000963, 0.00000448, -0.00000123, -0.00000665, -0.00001104, -0.00001392, -0.00001514, -0.00001488, -0.00001355, -0.00001164, -0.00000962, -0.00000779, -0.00000625, -0.00000488, -0.00000346, -0.00000175, 0.00000038, 0.00000288, 0.00000550, 0.00000783, 0.00000939, 0.00000978, 0.00000877, 0.00000637, 0.00000290, -0.00000115, -0.00000515, -0.00000850, -0.00001074, -0.00001165, -0.00001126, -0.00000980, -0.00000765, -0.00000520, -0.00000278, -0.00000056, 0.00000144, 0.00000329, 0.00000514, 0.00000707, 0.00000905, 0.00001094, 0.00001245, 0.00001326, 0.00001311, 0.00001184, 0.00000951, 0.00000640, 0.00000294, -0.00000035, -0.00000297, -0.00000457, -0.00000502, -0.00000439, -0.00000299, -0.00000124, 0.00000045, 0.00000172, 0.00000239, 0.00000245, 0.00000207, 0.00000149, 0.00000096, 0.00000067, 0.00000068, 0.00000094, 0.00000131, 0.00000159, 0.00000166, 0.00000148, 0.00000113, 0.00000076, 0.00000061, 0.00000083, 0.00000151, 0.00000258, 0.00000382, 0.00000492, 0.00000555, 0.00000544, },
		{ 0.00000088, 0.00000250, 0.00000434, 0.00000600, 0.00000707, 0.00000724, 0.00000639, 0.00000460, 0.00000216, -0.00000053, -0.00000301, -0.00000492, -0.00000602, -0.00000626, -0.00000579, -0.00000487, -0.00000379, -0.00000282, -0.00000211, -0.00000171, -0.00000153, -0.00000142, -0.00000123, -0.00000085, -0.00000030, 0.00000034, 0.00000093, 0.00000131, 0.00000140, 0.00000121, 0.00000088, 0.00000061, 0.00000062, 0.00000107, 0.00000199, 0.00000324, 0.00000452, 0.00000545, 0.00000565, 0.00000488, 0.00000309, 0.00000050, -0.00000244, -0.00000513, -0.00000690, -0.00000725, -0.00000589, -0.00000288, 0.00000139, 0.00000626, 0.00001091, 0.00001454, 0.00001652, 0.00001652, 0.00001450, 0.00001078, 0.00000589, 0.00000049, -0.00000479, -0.00000941, -0.00001305, -0.00001553, -0.00001686, -0.00001713, -0.00001644, -0.00001490, -0.00001258, -0.00000955, -0.00000590, -0.00000184, 0.00000235, 0.00000628, 0.00000950, 0.00001161, 0.00001230, 0.00001150, 0.00000933, 0.00000615, 0.00000248, -0.00000111, -0.00000413, -0.00000624, -0.00000735, -0.00000759, -0.00000724, -0.00000668, -0.00000624, -0.00000609, -0.00000623, -0.00000643, -0.00000638, -0.00000569, -0.00000411, -0.00000156, 0.00000179, 0.00000555, 0.00000917, 0.00001206, 0.00001374, 0.00001393, 0.00001263, 0.00001013, 0.00000689, 0.00000351, 0.00000055, -0.00000160, -0.00000274, -0.00000290, -0.00000231, -0.00000132, -0.00000024, 0.00000064, 0.00000122, 0.00000149, 0.00000159, 0.00000168, 0.00000187, 0.00000221, 0.00000265, 0.00000303, 0.00000318, 0.00000298, 0.00000239, 0.00000149, 0.00000051, -0.00000026, -0.00000056, -0.00000019, },
		{ 0.00000374, 0.00000242, 0.00000101, -0.00000013, -0.00000067, -0.00000041, 0.00000067, 0.00000237, 0.00000434, 0.00000615, 0.00000737, 0.00000771, 0.00000705, 0.00000549, 0.00000330, 0.00000086, -0.00000142, -0.00000324, -0.00000442, -0.00000495, -0.00000497, -0.00000468, -0.00000432, -0.00000405, -0.00000393, -0.00000390, -0.00000384, -0.00000359, -0.00000304, -0.00000219, -0.00000111, 0.00000000, 0.00000094, 0.00000151, 0.00000161, 0.00000131, 0.00000076, 0.00000024, 0.00000002, 0.00000032, 0.00000120, 0.00000255, 0.00000409, 0.00000542, 0.00000613, 0.00000593, 0.00000468, 0.00000250, -0.00000028, -0.00000315, -0.00000552, -0.00000688, -0.00000689, -0.00000542, -0.00000264, 0.00000106, 0.00000512, 0.00000895, 0.00001200, 0.00001384, 0.00001427, 0.00001323, 0.00001086, 0.00000742, 0.00000321, -0.00000142, -0.00000611, -0.00001051, -0.00001426, -0.00001704, -0.00001853, -0.00001852, -0.00001689, -0.00001371, -0.00000927, -0.00000405, 0.00000134, 0.00000620, 0.00000989, 0.00001195, 0.00001219, 0.00001074, 0.00000799, 0.00000458, 0.00000119, -0.00000156, -0.00000328, -0.00000385, -0.00000347, -0.00000255, -0.00000166, -0.00000129, -0.00000178, -0.00000320, -0.00000530, -0.00000762, -0.00000954, -0.00001050, -0.00001007, -0.00000813, -0.00000489, -0.00000080, 0.00000348, 0.00000726, 0.00000996, 0.00001126, 0.00001109, 0.00000968, 0.00000747, 0.00000496, 0.00000264, 0.00000085, -0.00000025, -0.00000070, -0.00000067, -0.00000039, -0.00000007, 0.00000017, 0.00000032, 0.00000048, 0.00000077, 0.00000131, 0.00000211, 0.00000309, 0.00000406, 0.00000477, 0.00000502, 0.00000467, },
		{ 0.00000182, 0.00000326, 0.00000466, 0.00000573, 0.00000621, 0.00000596, 0.00000505, 0.00000367, 0.00000219, 0.00000097, 0.00000034, 0.00000050, 0.00000143, 0.00000292, 0.00000464, 0.00000618, 0.00000717, 0.00000738, 0.00000674, 0.00000537, 0.00000355, 0.00000159, -0.00000018, -0.00000155, -0.00000245, -0.00000295, -0.00000321, -0.00000342, -0.00000375, -0.00000426, -0.00000488, -0.00000546, -0.00000578, -0.00000565, -0.00000496, -0.00000374, -0.00000214, -0.00000044, 0.00000105, 0.00000208, 0.00000246, 0.00000221, 0.00000147, 0.00000052, -0.00000032, -0.00000076, -0.00000063, 0.00000010, 0.00000127, 0.00000263, 0.00000384, 0.00000459, 0.00000465, 0.00000398, 0.00000266, 0.00000094, -0.00000089, -0.00000246, -0.00000350, -0.00000380, -0.00000330, -0.00000203, -0.00000015, 0.00000214, 0.00000457, 0.00000689, 0.00000880, 0.00001006, 0.00001042, 0.00000970, 0.00000781, 0.00000476, 0.00000074, -0.00000388, -0.00000861, -0.00001282, -0.00001588, -0.00001728, -0.00001671, -0.00001415, -0.00000993, -0.00000467, 0.00000081, 0.00000564, 0.00000906, 0.00001061, 0.00001016, 0.00000800, 0.00000475, 0.00000118, -0.00000188, -0.00000381, -0.00000429, -0.00000338, -0.00000149, 0.00000075, 0.00000259, 0.00000344, 0.00000291, 0.00000102, -0.00000191, -0.00000527, -0.00000833, -0.00001041, -0.00001106, -0.00001010, -0.00000773, -0.00000440, -0.00000073, 0.00000261, 0.00000513, 0.00000652, 0.00000675, 0.00000604, 0.00000474, 0.00000327, 0.00000196, 0.00000103, 0.00000051, 0.00000032, 0.00000028, 0.00000023, 0.00000007, -0.00000018, -0.00000042, -0.00000046, -0.00000014, 0.00000063, },
		{ -0.00000064, -0.00000113, -0.00000158, -0.00000174, -0.00000143, -0.00000055, 0.00000082, 0.00000248, 0.00000413, 0.00000544, 0.00000616, 0.00000618, 0.00000556, 0.00000452, 0.00000339, 0.00000251, 0.00000215, 0.00000242, 0.00000328, 0.00000450, 0.00000577, 0.00000674, 0.00000714, 0.00000686, 0.00000592, 0.00000452, 0.00000295, 0.00000149, 0.00000036, -0.00000033, -0.00000063, -0.00000073, -0.00000086, -0.00000125, -0.00000204, -0.00000322, -0.00000465, -0.00000605, -0.00000710, -0.00000753, -0.00000715, -0.00000594, -0.00000406, -0.00000181, 0.00000045, 0.00000234, 0.00000358, 0.00000402, 0.00000366, 0.00000268, 0.00000134, -0.00000006, -0.00000124, -0.00000198, -0.00000218, -0.00000186, -0.00000110, -0.00000006, 0.00000107, 0.00000213, 0.00000295, 0.00000343, 0.00000351, 0.00000316, 0.00000244, 0.00000144, 0.00000033, -0.00000069, -0.00000137, -0.00000152, -0.00000098, 0.00000025, 0.00000205, 0.00000411, 0.00000601, 0.00000728, 0.00000746, 0.00000629, 0.00000371, -0.00000002, -0.00000438, -0.00000864, -0.00001202, -0.00001384, -0.00001367, -0.00001144, -0.00000749, -0.00000251, 0.00000258, 0.00000686, 0.00000954, 0.00001018, 0.00000877, 0.00000572, 0.00000177, -0.00000215, -0.00000520, -0.00000675, -0.00000655, -0.00000477, -0.00000195, 0.00000116, 0.00000375, 0.00000518, 0.00000508, 0.00000349, 0.00000078, -0.00000243, -0.00000542, -0.00000757, -0.00000845, -0.00000796, -0.00000631, -0.00000394, -0.00000138, 0.00000082, 0.00000229, 0.00000289, 0.00000267, 0.00000190, 0.00000091, 0.00000003, -0.00000054, -0.00000073, -0.00000062, -0.00000039, -0.00000024, -0.00000031, },
		{ -0.00000297, -0.00000280, -0.00000233, -0.00000183, -0.00000152, -0.00000155, -0.00000191, -0.00000247, -0.00000300, -0.00000324, -0.00000300, -0.00000220, -0.00000092, 0.00000064, 0.00000220, 0.00000347, 0.00000427, 0.00000453, 0.00000434, 0.00000392, 0.00000355, 0.00000348, 0.00000386, 0.00000470, 0.00000585, 0.00000704, 0.00000795, 0.00000830, 0.00000795, 0.00000689, 0.00000531, 0.00000350, 0.00000181, 0.00000054, -0.00000011, -0.00000013, 0.00000031, 0.00000093, 0.00000138, 0.00000134, 0.00000064, -0.00000077, -0.00000271, -0.00000486, -0.00000685, -0.00000826, -0.00000881, -0.00000833, -0.00000686, -0.00000460, -0.00000189, 0.00000089, 0.00000332, 0.00000508, 0.00000596, 0.00000586, 0.00000486, 0.00000313, 0.00000095, -0.00000137, -0.00000348, -0.00000506, -0.00000587, -0.00000577, -0.00000472, -0.00000286, -0.00000046, 0.00000213, 0.00000447, 0.00000618, 0.00000694, 0.00000663, 0.00000533, 0.00000331, 0.00000102, -0.00000103, -0.00000237, -0.00000268, -0.00000191, -0.00000027, 0.00000179, 0.00000367, 0.00000477, 0.00000461, 0.00000301, 0.00000013, -0.00000354, -0.00000726, -0.00001020, -0.00001164, -0.00001113, -0.00000864, -0.00000453, 0.00000046, 0.00000536, 0.00000923, 0.00001133, 0.00001126, 0.00000913, 0.00000542, 0.00000097, -0.00000327, -0.00000646, -0.00000801, -0.00000775, -0.00000591, -0.00000307, 0.00000001, 0.00000257, 0.00000402, 0.00000410, 0.00000290, 0.00000081, -0.00000158, -0.00000366, -0.00000493, -0.00000514, -0.00000430, -0.00000272, -0.00000084, 0.00000084, 0.00000194, 0.00000224, 0.00000175, 0.00000067, -0.00000064, -0.00000185, -0.00000267, },
		{ 0.00000367, 0.00000230, 0.00000046, -0.00000145, -0.00000303, -0.00000402, -0.00000434, -0.00000411, -0.00000354, -0.00000292, -0.00000251, -0.00000244, -0.00000271, -0.00000319, -0.00000368, -0.00000394, -0.00000382, -0.00000329, -0.00000242, -0.00000140, -0.00000044, 0.00000029, 0.00000069, 0.00000082, 0.00000083, 0.00000096, 0.00000144, 0.00000244, 0.00000396, 0.00000586, 0.00000786, 0.00000956, 0.00001062, 0.00001077, 0.00000991, 0.00000816, 0.00000582, 0.00000329, 0.00000104, -0.00000056, -0.00000127, -0.00000109, -0.00000018, 0.00000110, 0.00000233, 0.00000312, 0.00000315, 0.00000228, 0.00000058, -0.00000173, -0.00000427, -0.00000663, -0.00000840, -0.00000926, -0.00000903, -0.00000771, -0.00000543, -0.00000249, 0.00000072, 0.00000377, 0.00000622, 0.00000772, 0.00000802, 0.00000704, 0.00000488, 0.00000184, -0.00000164, -0.00000501, -0.00000772, -0.00000927, -0.00000938, -0.00000798, -0.00000528, -0.00000174, 0.00000204, 0.00000537, 0.00000767, 0.00000858, 0.00000802, 0.00000621, 0.00000366, 0.00000099, -0.00000116, -0.00000232, -0.00000229, -0.00000121, 0.00000050, 0.00000221, 0.00000327, 0.00000317, 0.00000169, -0.00000103, -0.00000448, -0.00000793, -0.00001053, -0.00001159, -0.00001070, -0.00000786, -0.00000349, 0.00000162, 0.00000652, 0.00001028, 0.00001220, 0.00001200, 0.00000980, 0.00000617, 0.00000192, -0.00000205, -0.00000499, -0.00000641, -0.00000624, -0.00000476, -0.00000255, -0.00000028, 0.00000141, 0.00000211, 0.00000173, 0.00000045, -0.00000127, -0.00000288, -0.00000390, -0.00000399, -0.00000310, -0.00000143, 0.00000061, 0.00000251, 0.00000381, 0.00000422, },
		{ -0.00000212, 0.00000032, 0.00000286, 0.00000492, 0.00000608, 0.00000611, 0.00000502, 0.00000308, 0.00000071, -0.00000162, -0.00000352, -0.00000473, -0.00000519, -0.00000502, -0.00000445, -0.00000377, -0.00000320, -0.00000288, -0.00000281, -0.00000289, -0.00000299, -0.00000297, -0.00000278, -0.00000244, -0.00000208, -0.00000186, -0.00000190, -0.00000227, -0.00000289, -0.00000357, -0.00000403, -0.00000395, -0.00000309, -0.00000135, 0.00000119, 0.00000425, 0.00000741, 0.00001016, 0.00001206, 0.00001276, 0.00001215, 0.00001036, 0.00000770, 0.00000466, 0.00000175, -0.00000056, -0.00000193, -0.00000226, -0.00000162, -0.00000027, 0.00000138, 0.00000293, 0.00000397, 0.00000423, 0.00000359, 0.00000206, -0.00000014, -0.00000271, -0.00000526, -0.00000737, -0.00000870, -0.00000896, -0.00000805, -0.00000600, -0.00000304, 0.00000042, 0.00000388, 0.00000678, 0.00000861, 0.00000901, 0.00000783, 0.00000520, 0.00000151, -0.00000262, -0.00000647, -0.00000935, -0.00001072, -0.00001033, -0.00000827, -0.00000495, -0.00000101, 0.00000276, 0.00000567, 0.00000723, 0.00000727, 0.00000596, 0.00000379, 0.00000140, -0.00000054, -0.00000154, -0.00000139, -0.00000022, 0.00000152, 0.00000319, 0.00000414, 0.00000386, 0.00000216, -0.00000077, -0.00000442, -0.00000801, -0.00001071, -0.00001186, -0.00001107, -0.00000840, -0.00000429, 0.00000049, 0.00000503, 0.00000850, 0.00001031, 0.00001023, 0.00000848, 0.00000560, 0.00000232, -0.00000058, -0.00000253, -0.00000323, -0.00000275, -0.00000143, 0.00000015, 0.00000143, 0.00000195, 0.00000152, 0.00000023, -0.00000155, -0.00000332, -0.00000453, -0.00000480, -0.00000395, },
		{ 0.00000026, -0.00000199, -0.00000386, -0.00000484, -0.00000461, -0.00000316, -0.00000073, 0.00000216, 0.00000493, 0.00000699, 0.00000794, 0.00000760, 0.00000608, 0.00000369, 0.00000091, -0.00000180, -0.00000403, -0.00000556, -0.00000631, -0.00000640, -0.00000603, -0.00000540, -0.00000469, -0.00000400, -0.00000334, -0.00000265, -0.00000190, -0.00000107, -0.00000026, 0.00000039, 0.00000067, 0.00000041, -0.00000051, -0.00000204, -0.00000396, -0.00000595, -0.00000757, -0.00000840, -0.00000812, -0.00000659, -0.00000390, -0.00000036, 0.00000354, 0.00000725, 0.00001023, 0.00001206, 0.00001253, 0.00001165, 0.00000966, 0.00000694, 0.00000396, 0.00000121, -0.00000092, -0.00000219, -0.00000249, -0.00000191, -0.00000065, 0.00000097, 0.00000260, 0.00000390, 0.00000459, 0.00000446, 0.00000346, 0.00000168, -0.00000066, -0.00000320, -0.00000552, -0.00000718, -0.00000784, -0.00000727, -0.00000547, -0.00000267, 0.00000069, 0.00000400, 0.00000665, 0.00000809, 0.00000798, 0.00000628, 0.00000324, -0.00000057, -0.00000442, -0.00000758, -0.00000942, -0.00000962, -0.00000817, -0.00000544, -0.00000205, 0.00000125, 0.00000377, 0.00000503, 0.00000489, 0.00000356, 0.00000152, -0.00000057, -0.00000205, -0.00000248, -0.00000170, 0.00000009, 0.00000239, 0.00000451, 0.00000578, 0.00000571, 0.00000415, 0.00000128, -0.00000235, -0.00000598, -0.00000882, -0.00001025, -0.00000995, -0.00000799, -0.00000480, -0.00000107, 0.00000240, 0.00000495, 0.00000613, 0.00000589, 0.00000448, 0.00000245, 0.00000042, -0.00000103, -0.00000152, -0.00000099, 0.00000031, 0.00000193, 0.00000331, 0.00000398, 0.00000364, 0.00000230, },
		{ 0.00000105, 0.00000313, 0.00000460, 0.00000506, 0.00000436, 0.00000266, 0.00000041, -0.00000180, -0.00000339, -0.00000387, -0.00000304, -0.00000099, 0.00000191, 0.00000507, 0.00000788, 0.00000978, 0.00001042, 0.00000969, 0.00000775, 0.00000496, 0.00000179, -0.00000131, -0.00000399, -0.00000603, -0.00000737, -0.00000808, -0.00000828, -0.00000809, -0.00000759, -0.00000679, -0.00000566, -0.00000419, -0.00000240, -0.00000042, 0.00000153, 0.00000315, 0.00000412, 0.00000419, 0.00000322, 0.00000128, -0.00000140, -0.00000438, -0.00000718, -0.00000929, -0.00001030, -0.00001000, -0.00000838, -0.00000566, -0.00000222, 0.00000144, 0.00000482, 0.00000751, 0.00000922, 0.00000980, 0.00000932, 0.00000794, 0.00000595, 0.00000368, 0.00000145, -0.00000042, -0.00000171, -0.00000228, -0.00000208, -0.00000121, 0.00000018, 0.00000179, 0.00000331, 0.00000441, 0.00000481, 0.00000436, 0.00000306, 0.00000110, -0.00000119, -0.00000335, -0.00000493, -0.00000553, -0.00000498, -0.00000333, -0.00000088, 0.00000185, 0.00000426, 0.00000579, 0.00000602, 0.00000485, 0.00000246, -0.00000068, -0.00000389, -0.00000648, -0.00000790, -0.00000785, -0.00000636, -0.00000382, -0.00000084, 0.00000186, 0.00000365, 0.00000415, 0.00000330, 0.00000139, -0.00000102, -0.00000324, -0.00000462, -0.00000474, -0.00000352, -0.00000121, 0.00000163, 0.00000429, 0.00000611, 0.00000660, 0.00000561, 0.00000335, 0.00000033, -0.00000278, -0.00000527, -0.00000662, -0.00000661, -0.00000536, -0.00000328, -0.00000096, 0.00000097, 0.00000205, 0.00000205, 0.00000108, -0.00000049, -0.00000212, -0.00000328, -0.00000354, -0.00000277, -0.00000111, },
		{ -0.00000393, -0.00000562, -0.00000645, -0.00000617, -0.00000481, -0.00000265, -0.00000022, 0.00000191, 0.00000323, 0.00000344, 0.00000254, 0.00000082, -0.00000119, -0.00000287, -0.00000368, -0.00000323, -0.00000145, 0.00000143, 0.00000496, 0.00000850, 0.00001144, 0.00001327, 0.00001372, 0.00001276, 0.00001060, 0.00000762, 0.00000424, 0.00000087, -0.00000221, -0.00000482, -0.00000692, -0.00000853, -0.00000970, -0.00001044, -0.00001069, -0.00001039, -0.00000942, -0.00000774, -0.00000542, -0.00000262, 0.00000034, 0.00000307, 0.00000516, 0.00000630, 0.00000627, 0.00000509, 0.00000293, 0.00000014, -0.00000283, -0.00000552, -0.00000756, -0.00000867, -0.00000876, -0.00000789, -0.00000623, -0.00000405, -0.00000165, 0.00000070, 0.00000276, 0.00000435, 0.00000534, 0.00000569, 0.00000540, 0.00000453, 0.00000321, 0.00000166, 0.00000010, -0.00000120, -0.00000200, -0.00000215, -0.00000159, -0.00000042, 0.00000112, 0.00000270, 0.00000396, 0.00000460, 0.00000444, 0.00000350, 0.00000198, 0.00000024, -0.00000128, -0.00000219, -0.00000223, -0.00000138, 0.00000016, 0.00000198, 0.00000355, 0.00000438, 0.00000413, 0.00000273, 0.00000038, -0.00000245, -0.00000514, -0.00000707, -0.00000777, -0.00000703, -0.00000502, -0.00000217, 0.00000085, 0.00000333, 0.00000470, 0.00000463, 0.00000317, 0.00000070, -0.00000216, -0.00000469, -0.00000625, -0.00000645, -0.00000524, -0.00000291, -0.00000002, 0.00000276, 0.00000480, 0.00000567, 0.00000526, 0.00000378, 0.00000168, -0.00000044, -0.00000205, -0.00000278, -0.00000254, -0.00000152, -0.00000016, 0.00000104, 0.00000160, 0.00000126, 0.00000001, -0.00000188, },
		{ 0.00000255, 0.00000320, 0.00000301, 0.00000186, -0.00000009, -0.00000247, -0.00000476, -0.00000644, -0.00000714, -0.00000673, -0.00000535, -0.00000341, -0.00000145, 0.00000000, 0.00000054, 0.00000003, -0.00000138, -0.00000326, -0.00000503, -0.00000608, -0.00000595, -0.00000439, -0.00000149, 0.00000239, 0.00000670, 0.00001078, 0.00001404, 0.00001604, 0.00001659, 0.00001575, 0.00001377, 0.00001100, 0.00000784, 0.00000460, 0.00000150, -0.00000137, -0.00000398, -0.00000635, -0.00000847, -0.00001026, -0.00001158, -0.00001227, -0.00001215, -0.00001113, -0.00000925, -0.00000666, -0.00000366, -0.00000062, 0.00000208, 0.00000413, 0.00000532, 0.00000559, 0.00000502, 0.00000381, 0.00000219, 0.00000041, -0.00000130, -0.00000281, -0.00000402, -0.00000488, -0.00000539, -0.00000552, -0.00000526, -0.00000461, -0.00000360, -0.00000230, -0.00000086, 0.00000053, 0.00000166, 0.00000233, 0.00000242, 0.00000193, 0.00000098, -0.00000018, -0.00000126, -0.00000197, -0.00000208, -0.00000154, -0.00000045, 0.00000095, 0.00000233, 0.00000337, 0.00000384, 0.00000366, 0.00000294, 0.00000197, 0.00000108, 0.00000060, 0.00000075, 0.00000153, 0.00000273, 0.00000398, 0.00000483, 0.00000486, 0.00000384, 0.00000179, -0.00000101, -0.00000404, -0.00000666, -0.00000831, -0.00000858, -0.00000737, -0.00000491, -0.00000175, 0.00000143, 0.00000392, 0.00000517, 0.00000494, 0.00000332, 0.00000075, -0.00000214, -0.00000465, -0.00000619, -0.00000644, -0.00000538, -0.00000332, -0.00000080, 0.00000158, 0.00000329, 0.00000402, 0.00000374, 0.00000268, 0.00000129, 0.00000005, -0.00000063, -0.00000056, 0.00000022, 0.00000140, },
		{ 0.00000045, -0.00000041, -0.00000063, -0.00000013, 0.00000093, 0.00000219, 0.00000323, 0.00000365, 0.00000321, 0.00000191, -0.00000001, -0.00000214, -0.00000398, -0.00000512, -0.00000534, -0.00000465, -0.00000335, -0.00000189, -0.00000081, -0.00000053, -0.00000128, -0.00000301, -0.00000537, -0.00000781, -0.00000969, -0.00001042, -0.00000963, -0.00000723, -0.00000344, 0.00000124, 0.00000618, 0.00001071, 0.00001428, 0.00001654, 0.00001738, 0.00001692, 0.00001543, 0.00001327, 0.00001075, 0.00000813, 0.00000553, 0.00000298, 0.00000047, -0.00000203, -0.00000449, -0.00000680, -0.00000880, -0.00001029, -0.00001112, -0.00001118, -0.00001048, -0.00000914, -0.00000737, -0.00000539, -0.00000344, -0.00000166, -0.00000016, 0.00000108, 0.00000208, 0.00000290, 0.00000355, 0.00000398, 0.00000410, 0.00000381, 0.00000302, 0.00000171, -0.00000002, -0.00000197, -0.00000385, -0.00000537, -0.00000627, -0.00000641, -0.00000582, -0.00000465, -0.00000321, -0.00000183, -0.00000079, -0.00000030, -0.00000036, -0.00000084, -0.00000150, -0.00000204, -0.00000221, -0.00000189, -0.00000109, 0.00000002, 0.00000117, 0.00000210, 0.00000262, 0.00000266, 0.00000234, 0.00000189, 0.00000160, 0.00000172, 0.00000239, 0.00000354, 0.00000492, 0.00000612, 0.00000671, 0.00000634, 0.00000483, 0.00000230, -0.00000089, -0.00000419, -0.00000694, -0.00000859, -0.00000881, -0.00000755, -0.00000512, -0.00000207, 0.00000092, 0.00000317, 0.00000422, 0.00000390, 0.00000236, 0.00000006, -0.00000240, -0.00000438, -0.00000542, -0.00000528, -0.00000406, -0.00000210, 0.00000009, 0.00000196, 0.00000311, 0.00000335, 0.00000276, 0.00000165, },
		{ 0.00000186, 0.00000354, 0.00000436, 0.00000425, 0.00000335, 0.00000203, 0.00000072, -0.00000020, -0.00000051, -0.00000019, 0.00000057, 0.00000142, 0.00000199, 0.00000200, 0.00000135, 0.00000013, -0.00000135, -0.00000268, -0.00000347, -0.00000348, -0.00000268, -0.00000128, 0.00000028, 0.00000148, 0.00000181, 0.00000097, -0.00000110, -0.00000413, -0.00000760, -0.00001081, -0.00001308, -0.00001385, -0.00001284, -0.00001009, -0.00000595, -0.00000099, 0.00000409, 0.00000866, 0.00001222, 0.00001452, 0.00001551, 0.00001537, 0.00001440, 0.00001291, 0.00001117, 0.00000936, 0.00000755, 0.00000573, 0.00000387, 0.00000195, 0.00000002, -0.00000186, -0.00000354, -0.00000492, -0.00000592, -0.00000654, -0.00000684, -0.00000696, -0.00000702, -0.00000709, -0.00000717, -0.00000716, -0.00000688, -0.00000615, -0.00000483, -0.00000288, -0.00000043, 0.00000224, 0.00000474, 0.00000664, 0.00000759, 0.00000739, 0.00000603, 0.00000376, 0.00000095, -0.00000191, -0.00000435, -0.00000604, -0.00000682, -0.00000675, -0.00000606, -0.00000507, -0.00000411, -0.00000343, -0.00000312, -0.00000312, -0.00000327, -0.00000333, -0.00000312, -0.00000253, -0.00000160, -0.00000049, 0.00000056, 0.00000135, 0.00000173, 0.00000171, 0.00000142, 0.00000112, 0.00000109, 0.00000156, 0.00000259, 0.00000407, 0.00000569, 0.00000702, 0.00000764, 0.00000722, 0.00000565, 0.00000308, -0.00000008, -0.00000326, -0.00000587, -0.00000739, -0.00000757, -0.00000644, -0.00000434, -0.00000181, 0.00000052, 0.00000208, 0.00000252, 0.00000180, 0.00000017, -0.00000190, -0.00000383, -0.00000508, -0.00000530, -0.00000441, -0.00000263, -0.00000036, },
		{ -0.00000600, -0.00000671, -0.00000623, -0.00000461, -0.00000217, 0.00000060, 0.00000315, 0.00000503, 0.00000596, 0.00000591, 0.00000507, 0.00000380, 0.00000249, 0.00000145, 0.00000085, 0.00000063, 0.00000063, 0.00000056, 0.00000018, -0.00000063, -0.00000181, -0.00000312, -0.00000421, -0.00000469, -0.00000430, -0.00000297, -0.00000089, 0.00000151, 0.00000366, 0.00000496, 0.00000493, 0.00000338, 0.00000042, -0.00000351, -0.00000773, -0.00001148, -0.00001404, -0.00001491, -0.00001391, -0.00001118, -0.00000716, -0.00000248, 0.00000219, 0.00000627, 0.00000935, 0.00001127, 0.00001205, 0.00001193, 0.00001117, 0.00001006, 0.00000881, 0.00000755, 0.00000633, 0.00000515, 0.00000400, 0.00000294, 0.00000201, 0.00000130, 0.00000088, 0.00000074, 0.00000077, 0.00000080, 0.00000058, -0.00000011, -0.00000143, -0.00000336, -0.00000572, -0.00000814, -0.00001016, -0.00001132, -0.00001126, -0.00000982, -0.00000711, -0.00000351, 0.00000044, 0.00000410, 0.00000688, 0.00000837, 0.00000842, 0.00000715, 0.00000491, 0.00000218, -0.00000051, -0.00000277, -0.00000433, -0.00000518, -0.00000543, -0.00000534, -0.00000516, -0.00000507, -0.00000514, -0.00000531, -0.00000541, -0.00000527, -0.00000475, -0.00000383, -0.00000259, -0.00000125, -0.00000004, 0.00000081, 0.00000120, 0.00000116, 0.00000088, 0.00000061, 0.00000065, 0.00000120, 0.00000229, 0.00000377, 0.00000534, 0.00000659, 0.00000711, 0.00000666, 0.00000516, 0.00000283, 0.00000007, -0.00000258, -0.00000459, -0.00000558, -0.00000539, -0.00000417, -0.00000231, -0.00000034, 0.00000117, 0.00000179, 0.00000132, -0.00000013, -0.00000220, -0.00000435, },
		{ 0.00000264, 0.00000247, 0.00000116, -0.00000104, -0.00000366, -0.00000609, -0.00000777, -0.00000828, -0.00000746, -0.00000545, -0.00000263, 0.00000048, 0.00000333, 0.00000551, 0.00000676, 0.00000711, 0.00000672, 0.00000589, 0.00000493, 0.00000403, 0.00000327, 0.00000256, 0.00000173, 0.00000059, -0.00000095, -0.00000283, -0.00000480, -0.00000648, -0.00000744, -0.00000734, -0.00000602, -0.00000359, -0.00000044, 0.00000284, 0.00000554, 0.00000704, 0.00000691, 0.00000504, 0.00000170, -0.00000254, -0.00000693, -0.00001065, -0.00001305, -0.00001372, -0.00001258, -0.00000989, -0.00000616, -0.00000202, 0.00000192, 0.00000516, 0.00000741, 0.00000861, 0.00000887, 0.00000841, 0.00000751, 0.00000637, 0.00000515, 0.00000393, 0.00000275, 0.00000165, 0.00000074, 0.00000014, 0.00000000, 0.00000047, 0.00000156, 0.00000318, 0.00000506, 0.00000679, 0.00000790, 0.00000799, 0.00000677, 0.00000424, 0.00000063, -0.00000354, -0.00000762, -0.00001090, -0.00001282, -0.00001306, -0.00001158, -0.00000870, -0.00000498, -0.00000108, 0.00000232, 0.00000474, 0.00000591, 0.00000584, 0.00000480, 0.00000316, 0.00000136, -0.00000025, -0.00000149, -0.00000231, -0.00000284, -0.00000326, -0.00000373, -0.00000436, -0.00000511, -0.00000584, -0.00000635, -0.00000644, -0.00000601, -0.00000504, -0.00000370, -0.00000221, -0.00000084, 0.00000017, 0.00000073, 0.00000087, 0.00000077, 0.00000068, 0.00000084, 0.00000142, 0.00000243, 0.00000369, 0.00000491, 0.00000572, 0.00000581, 0.00000503, 0.00000340, 0.00000121, -0.00000112, -0.00000309, -0.00000428, -0.00000444, -0.00000356, -0.00000191, 0.00000004, 0.00000172, },
		{ -0.00000521, -0.00000459, -0.00000299, -0.00000080, 0.00000144, 0.00000313, 0.00000380, 0.00000321, 0.00000144, -0.00000119, -0.00000412, -0.00000671, -0.00000842, -0.00000887, -0.00000798, -0.00000592, -0.00000311, -0.00000003, 0.00000282, 0.00000509, 0.00000659, 0.00000735, 0.00000752, 0.00000732, 0.00000694, 0.00000644, 0.00000577, 0.00000480, 0.00000334, 0.00000131, -0.00000124, -0.00000407, -0.00000677, -0.00000884, -0.00000981, -0.00000937, -0.00000746, -0.00000435, -0.00000058, 0.00000313, 0.00000602, 0.00000746, 0.00000713, 0.00000504, 0.00000158, -0.00000258, -0.00000662, -0.00000980, -0.00001153, -0.00001154, -0.00000991, -0.00000702, -0.00000343, 0.00000024, 0.00000345, 0.00000581, 0.00000718, 0.00000760, 0.00000724, 0.00000635, 0.00000514, 0.00000374, 0.00000224, 0.00000064, -0.00000100, -0.00000260, -0.00000395, -0.00000480, -0.00000486, -0.00000390, -0.00000185, 0.00000115, 0.00000473, 0.00000832, 0.00001128, 0.00001300, 0.00001307, 0.00001135, 0.00000805, 0.00000367, -0.00000108, -0.00000541, -0.00000867, -0.00001042, -0.00001054, -0.00000923, -0.00000694, -0.00000426, -0.00000175, 0.00000014, 0.00000119, 0.00000141, 0.00000101, 0.00000029, -0.00000044, -0.00000096, -0.00000120, -0.00000124, -0.00000124, -0.00000141, -0.00000189, -0.00000272, -0.00000379, -0.00000491, -0.00000581, -0.00000625, -0.00000611, -0.00000539, -0.00000422, -0.00000285, -0.00000152, -0.00000044, 0.00000028, 0.00000069, 0.00000093, 0.00000121, 0.00000170, 0.00000248, 0.00000349, 0.00000452, 0.00000528, 0.00000546, 0.00000486, 0.00000344, 0.00000137, -0.00000099, -0.00000317, -0.00000469, },
		{ 0.00000516, 0.00000300, 0.00000030, -0.00000250, -0.00000485, -0.00000629, -0.00000653, -0.00000553, -0.00000354, -0.00000105, 0.00000134, 0.00000304, 0.00000362, 0.00000291, 0.00000105, -0.00000158, -0.00000438, -0.00000676, -0.00000823, -0.00000848, -0.00000749, -0.00000550, -0.00000288, -0.00000009, 0.00000248, 0.00000456, 0.00000606, 0.00000703, 0.00000762, 0.00000796, 0.00000811, 0.00000802, 0.00000753, 0.00000643, 0.00000457, 0.00000193, -0.00000131, -0.00000476, -0.00000789, -0.00001012, -0.00001096, -0.00001017, -0.00000783, -0.00000432, -0.00000032, 0.00000337, 0.00000599, 0.00000699, 0.00000616, 0.00000368, 0.00000007, -0.00000392, -0.00000745, -0.00000983, -0.00001061, -0.00000969, -0.00000731, -0.00000396, -0.00000027, 0.00000315, 0.00000583, 0.00000753, 0.00000822, 0.00000804, 0.00000723, 0.00000604, 0.00000461, 0.00000301, 0.00000121, -0.00000082, -0.00000307, -0.00000541, -0.00000757, -0.00000917, -0.00000981, -0.00000916, -0.00000707, -0.00000365, 0.00000070, 0.00000537, 0.00000963, 0.00001280, 0.00001435, 0.00001408, 0.00001211, 0.00000887, 0.00000500, 0.00000120, -0.00000191, -0.00000394, -0.00000478, -0.00000458, -0.00000372, -0.00000265, -0.00000180, -0.00000143, -0.00000162, -0.00000227, -0.00000310, -0.00000384, -0.00000422, -0.00000414, -0.00000364, -0.00000290, -0.00000218, -0.00000174, -0.00000175, -0.00000225, -0.00000313, -0.00000414, -0.00000504, -0.00000557, -0.00000559, -0.00000510, -0.00000420, -0.00000308, -0.00000196, -0.00000097, -0.00000017, 0.00000049, 0.00000115, 0.00000195, 0.00000296, 0.00000417, 0.00000540, 0.00000640, 0.00000684, 0.00000646, },
		{ 0.00000393, 0.00000577, 0.00000750, 0.00000876, 0.00000922, 0.00000861, 0.00000690, 0.00000424, 0.00000103, -0.00000218, -0.00000484, -0.00000647, -0.00000682, -0.00000592, -0.00000407, -0.00000180, 0.00000030, 0.00000168, 0.00000198, 0.00000111, -0.00000072, -0.00000310, -0.00000546, -0.00000727, -0.00000814, -0.00000789, -0.00000658, -0.00000449, -0.00000199, 0.00000051, 0.00000273, 0.00000451, 0.00000585, 0.00000685, 0.00000763, 0.00000828, 0.00000876, 0.00000890, 0.00000846, 0.00000721, 0.00000500, 0.00000191, -0.00000176, -0.00000553, -0.00000875, -0.00001082, -0.00001130, -0.00001003, -0.00000721, -0.00000339, 0.00000067, 0.00000411, 0.00000620, 0.00000651, 0.00000497, 0.00000191, -0.00000199, -0.00000589, -0.00000898, -0.00001062, -0.00001049, -0.00000867, -0.00000553, -0.00000169, 0.00000219, 0.00000551, 0.00000789, 0.00000918, 0.00000945, 0.00000896, 0.00000796, 0.00000670, 0.00000528, 0.00000368, 0.00000180, -0.00000044, -0.00000308, -0.00000597, -0.00000881, -0.00001115, -0.00001254, -0.00001257, -0.00001106, -0.00000808, -0.00000400, 0.00000060, 0.00000503, 0.00000861, 0.00001088, 0.00001163, 0.00001095, 0.00000922, 0.00000698, 0.00000478, 0.00000308, 0.00000211, 0.00000188, 0.00000215, 0.00000255, 0.00000269, 0.00000228, 0.00000120, -0.00000045, -0.00000243, -0.00000436, -0.00000589, -0.00000678, -0.00000693, -0.00000645, -0.00000557, -0.00000460, -0.00000384, -0.00000349, -0.00000359, -0.00000406, -0.00000470, -0.00000528, -0.00000560, -0.00000554, -0.00000512, -0.00000441, -0.00000357, -0.00000270, -0.00000189, -0.00000109, -0.00000022, 0.00000086, 0.00000225, },
		{ -0.00000421, -0.00000368, -0.00000302, -0.00000208, -0.00000070, 0.00000120, 0.00000354, 0.00000609, 0.00000849, 0.00001031, 0.00001117, 0.00001081, 0.00000921, 0.00000658, 0.00000337, 0.00000013, -0.00000257, -0.00000430, -0.00000486, -0.00000433, -0.00000303, -0.00000148, -0.00000022, 0.00000031, -0.00000013, -0.00000152, -0.00000356, -0.00000580, -0.00000772, -0.00000887, -0.00000899, -0.00000803, -0.00000616, -0.00000374, -0.00000115, 0.00000127, 0.00000329, 0.00000488, 0.00000609, 0.00000707, 0.00000795, 0.00000874, 0.00000934, 0.00000950, 0.00000894, 0.00000741, 0.00000482, 0.00000134, -0.00000265, -0.00000654, -0.00000966, -0.00001138, -0.00001134, -0.00000950, -0.00000619, -0.00000207, 0.00000198, 0.00000511, 0.00000665, 0.00000625, 0.00000400, 0.00000041, -0.00000377, -0.00000763, -0.00001037, -0.00001143, -0.00001062, -0.00000814, -0.00000448, -0.00000033, 0.00000363, 0.00000684, 0.00000898, 0.00001000, 0.00001006, 0.00000946, 0.00000848, 0.00000733, 0.00000608, 0.00000465, 0.00000290, 0.00000070, -0.00000199, -0.00000502, -0.00000808, -0.00001073, -0.00001250, -0.00001301, -0.00001209, -0.00000984, -0.00000658, -0.00000285, 0.00000074, 0.00000365, 0.00000552, 0.00000626, 0.00000604, 0.00000525, 0.00000435, 0.00000379, 0.00000384, 0.00000456, 0.00000575, 0.00000706, 0.00000803, 0.00000829, 0.00000759, 0.00000590, 0.00000344, 0.00000057, -0.00000226, -0.00000464, -0.00000629, -0.00000712, -0.00000722, -0.00000683, -0.00000623, -0.00000570, -0.00000543, -0.00000546, -0.00000572, -0.00000607, -0.00000633, -0.00000639, -0.00000620, -0.00000580, -0.00000528, -0.00000473, },
		},
		{ { 0.00000484, 0.00000488, 0.00000500, 0.00000516, 0.00000531, 0.00000539, 0.00000534, 0.00000510, 0.00000463, 0.00000389, 0.00000290, 0.00000167, 0.00000026, -0.00000125, -0.00000277, -0.00000421, -0.00000546, -0.00000643, -0.00000706, -0.00000729, -0.00000712, -0.00000655, -0.00000563, -0.00000442, -0.00000303, -0.00000153, -0.00000003, 0.00000137, 0.00000261, 0.00000361, 0.00000436, 0.00000483, 0.00000504, 0.00000503, 0.00000484, 0.00000453, 0.00000415, 0.00000376, 0.00000339, 0.00000307, 0.00000281, 0.00000262, 0.00000246, 0.00000233, 0.00000219, 0.00000201, 0.00000176, 0.00000145, 0.00000106, 0.00000060, 0.00000009, -0.00000044, -0.00000097, -0.00000145, -0.00000187, -0.00000219, -0.00000242, -0.00000254, -0.00000257, -0.00000253, -0.00000244, -0.00000233, -0.00000223, -0.00000216, -0.00000213, -0.00000216, -0.00000223, -0.00000233, -0.00000244, -0.00000253, -0.00000257, -0.00000254, -0.00000242, -0.00000219, -0.00000187, -0.00000145, -0.00000097, -0.00000044, 0.00000009, 0.00000060, 0.00000106, 0.00000145, 0.00000176, 0.00000201, 0.00000219, 0.00000233, 0.00000246, 0.00000262, 0.00000281, 0.00000307, 0.00000339, 0.00000376, 0.00000415, 0.00000453, 0.00000484, 0.00000503, 0.00000504, 0.00000483, 0.00000436, 0.00000361, 0.00000261, 0.00000137, -0.00000003, -0.00000153, -0.00000303, -0.00000442, -0.00000563, -0.00000655, -0.00000712, -0.00000729, -0.00000706, -0.00000643, -0.00000546, -0.00000421, -0.00000277, -0.00000125, 0.00000026, 0.00000167, 0.00000290, 0.00000389, 0.00000463, 0.00000510, 0.00000534, 0.00000539, 0.00000531, 0.00000516, 0.00000500, 0.00000488, },
		{ 0.00000359, 0.00000399, 0.00000423, 0.00000433, 0.00000436, 0.00000438, 0.00000443, 0.00000457, 0.00000482, 0.00000518, 0.00000564, 0.00000615, 0.00000666, 0.00000708, 0.00000734, 0.00000737, 0.00000710, 0.00000648, 0.00000552, 0.00000422, 0.00000264, 0.00000084, -0.00000108, -0.00000299, -0.00000478, -0.00000635, -0.00000758, -0.00000842, -0.00000882, -0.00000875, -0.00000826, -0.00000739, -0.00000622, -0.00000486, -0.00000339, -0.00000193, -0.00000058, 0.00000059, 0.00000154, 0.00000222, 0.00000264, 0.00000283, 0.00000281, 0.00000266, 0.00000242, 0.00000217, 0.00000194, 0.00000178, 0.00000171, 0.00000174, 0.00000187, 0.00000206, 0.00000229, 0.00000253, 0.00000273, 0.00000287, 0.00000294, 0.00000292, 0.00000281, 0.00000262, 0.00000238, 0.00000211, 0.00000184, 0.00000158, 0.00000136, 0.00000119, 0.00000105, 0.00000094, 0.00000085, 0.00000075, 0.00000062, 0.00000045, 0.00000021, -0.00000010, -0.00000048, -0.00000090, -0.00000136, -0.00000182, -0.00000225, -0.00000262, -0.00000291, -0.00000309, -0.00000317, -0.00000314, -0.00000301, -0.00000281, -0.00000255, -0.00000228, -0.00000202, -0.00000178, -0.00000158, -0.00000141, -0.00000128, -0.00000116, -0.00000103, -0.00000085, -0.00000061, -0.00000028, 0.00000015, 0.00000067, 0.00000127, 0.00000191, 0.00000254, 0.00000312, 0.00000357, 0.00000385, 0.00000392, 0.00000373, 0.00000328, 0.00000258, 0.00000166, 0.00000059, -0.00000057, -0.00000174, -0.00000281, -0.00000372, -0.00000438, -0.00000475, -0.00000479, -0.00000452, -0.00000394, -0.00000311, -0.00000210, -0.00000098, 0.00000016, 0.00000124, 0.00000220, 0.00000299, },
		{ -0.00000349, -0.00000286, -0.00000208, -0.00000122, -0.00000036, 0.00000044, 0.00000113, 0.00000165, 0.00000200, 0.00000218, 0.00000222, 0.00000216, 0.00000208, 0.00000202, 0.00000207, 0.00000226, 0.00000264, 0.00000321, 0.00000396, 0.00000483, 0.00000577, 0.00000669, 0.00000748, 0.00000806, 0.00000834, 0.00000826, 0.00000776, 0.00000686, 0.00000557, 0.00000396, 0.00000212, 0.00000016, -0.00000178, -0.00000358, -0.00000513, -0.00000631, -0.00000706, -0.00000734, -0.00000716, -0.00000654, -0.00000556, -0.00000430, -0.00000288, -0.00000142, -0.00000002, 0.00000122, 0.00000221, 0.00000291, 0.00000328, 0.00000334, 0.00000312, 0.00000266, 0.00000203, 0.00000130, 0.00000056, -0.00000015, -0.00000076, -0.00000124, -0.00000157, -0.00000174, -0.00000177, -0.00000167, -0.00000148, -0.00000123, -0.00000094, -0.00000065, -0.00000035, -0.00000007, 0.00000019, 0.00000045, 0.00000070, 0.00000095, 0.00000121, 0.00000148, 0.00000175, 0.00000201, 0.00000224, 0.00000243, 0.00000256, 0.00000261, 0.00000256, 0.00000242, 0.00000218, 0.00000186, 0.00000148, 0.00000106, 0.00000064, 0.00000024, -0.00000011, -0.00000039, -0.00000060, -0.00000073, -0.00000078, -0.00000079, -0.00000078, -0.00000076, -0.00000077, -0.00000083, -0.00000095, -0.00000113, -0.00000135, -0.00000161, -0.00000186, -0.00000208, -0.00000223, -0.00000227, -0.00000219, -0.00000198, -0.00000163, -0.00000118, -0.00000065, -0.00000010, 0.00000043, 0.00000088, 0.00000118, 0.00000131, 0.00000123, 0.00000094, 0.00000045, -0.00000020, -0.00000096, -0.00000177, -0.00000255, -0.00000324, -0.00000375, -0.00000406, -0.00000412, -0.00000392, },
		{ -0.00000117, -0.00000172, -0.00000226, -0.00000272, -0.00000306, -0.00000323, -0.00000320, -0.00000296, -0.00000252, -0.00000193, -0.00000122, -0.00000048, 0.00000023, 0.00000085, 0.00000132, 0.00000159, 0.00000166, 0.00000152, 0.00000122, 0.00000079, 0.00000032, -0.00000011, -0.00000043, -0.00000058, -0.00000048, -0.00000013, 0.00000047, 0.00000130, 0.00000229, 0.00000336, 0.00000439, 0.00000529, 0.00000594, 0.00000626, 0.00000617, 0.00000564, 0.00000468, 0.00000333, 0.00000165, -0.00000023, -0.00000220, -0.00000411, -0.00000582, -0.00000721, -0.00000817, -0.00000863, -0.00000855, -0.00000794, -0.00000684, -0.00000533, -0.00000352, -0.00000152, 0.00000053, 0.00000250, 0.00000427, 0.00000576, 0.00000689, 0.00000762, 0.00000794, 0.00000788, 0.00000746, 0.00000676, 0.00000583, 0.00000477, 0.00000362, 0.00000247, 0.00000137, 0.00000035, -0.00000055, -0.00000132, -0.00000196, -0.00000246, -0.00000284, -0.00000311, -0.00000326, -0.00000332, -0.00000328, -0.00000315, -0.00000294, -0.00000265, -0.00000229, -0.00000187, -0.00000142, -0.00000096, -0.00000050, -0.00000008, 0.00000029, 0.00000057, 0.00000077, 0.00000087, 0.00000087, 0.00000079, 0.00000065, 0.00000047, 0.00000029, 0.00000014, 0.00000003, -0.00000002, 0.00000002, 0.00000012, 0.00000028, 0.00000049, 0.00000069, 0.00000088, 0.00000100, 0.00000104, 0.00000098, 0.00000080, 0.00000052, 0.00000015, -0.00000027, -0.00000072, -0.00000114, -0.00000150, -0.00000176, -0.00000190, -0.00000190, -0.00000176, -0.00000151, -0.00000117, -0.00000080, -0.00000044, -0.00000015, 0.00000004, 0.00000010, -0.00000001, -0.00000027, -0.00000067, },
		{ -0.00000186, -0.00000163, -0.00000140, -0.00000124, -0.00000115, -0.00000118, -0.00000132, -0.00000157, -0.00000189, -0.00000224, -0.00000258, -0.00000285, -0.00000300, -0.00000298, -0.00000277, -0.00000236, -0.00000175, -0.00000100, -0.00000014, 0.00000075, 0.00000161, 0.00000235, 0.00000292, 0.00000326, 0.00000335, 0.00000318, 0.00000279, 0.00000222, 0.00000155, 0.00000085, 0.00000023, -0.00000024, -0.00000048, -0.00000044, -0.00000011, 0.00000050, 0.00000135, 0.00000236, 0.00000343, 0.00000444, 0.00000527, 0.00000581, 0.00000596, 0.00000565, 0.00000485, 0.00000357, 0.00000185, -0.00000022, -0.00000253, -0.00000494, -0.00000729, -0.00000944, -0.00001124, -0.00001257, -0.00001336, -0.00001356, -0.00001314, -0.00001215, -0.00001064, -0.00000872, -0.00000649, -0.00000409, -0.00000165, 0.00000073, 0.00000292, 0.00000484, 0.00000643, 0.00000765, 0.00000849, 0.00000896, 0.00000908, 0.00000890, 0.00000846, 0.00000781, 0.00000701, 0.00000611, 0.00000515, 0.00000416, 0.00000319, 0.00000225, 0.00000137, 0.00000057, -0.00000014, -0.00000074, -0.00000122, -0.00000157, -0.00000179, -0.00000189, -0.00000188, -0.00000176, -0.00000157, -0.00000133, -0.00000107, -0.00000083, -0.00000062, -0.00000049, -0.00000043, -0.00000047, -0.00000059, -0.00000079, -0.00000104, -0.00000130, -0.00000155, -0.00000175, -0.00000187, -0.00000188, -0.00000178, -0.00000157, -0.00000126, -0.00000088, -0.00000045, -0.00000003, 0.00000035, 0.00000064, 0.00000082, 0.00000086, 0.00000076, 0.00000052, 0.00000017, -0.00000026, -0.00000073, -0.00000119, -0.00000161, -0.00000193, -0.00000214, -0.00000223, -0.00000220, -0.00000207, },
		{ -0.00000024, -0.00000072, -0.00000120, -0.00000164, -0.00000200, -0.00000227, -0.00000243, -0.00000249, -0.00000247, -0.00000241, -0.00000233, -0.00000228, -0.00000229, -0.00000237, -0.00000255, -0.00000282, -0.00000315, -0.00000352, -0.00000387, -0.00000415, -0.00000431, -0.00000431, -0.00000412, -0.00000371, -0.00000309, -0.00000230, -0.00000137, -0.00000037, 0.00000062, 0.00000152, 0.00000227, 0.00000281, 0.00000309, 0.00000309, 0.00000284, 0.00000237, 0.00000174, 0.00000103, 0.00000034, -0.00000023, -0.00000060, -0.00000070, -0.00000048, 0.00000007, 0.00000092, 0.00000203, 0.00000332, 0.00000467, 0.00000596, 0.00000706, 0.00000787, 0.00000827, 0.00000819, 0.00000759, 0.00000647, 0.00000486, 0.00000284, 0.00000051, -0.00000201, -0.00000456, -0.00000703, -0.00000926, -0.00001114, -0.00001258, -0.00001353, -0.00001394, -0.00001382, -0.00001321, -0.00001217, -0.00001077, -0.00000911, -0.00000728, -0.00000538, -0.00000349, -0.00000170, -0.00000005, 0.00000140, 0.00000262, 0.00000360, 0.00000434, 0.00000484, 0.00000512, 0.00000520, 0.00000509, 0.00000484, 0.00000447, 0.00000400, 0.00000347, 0.00000290, 0.00000233, 0.00000179, 0.00000130, 0.00000089, 0.00000058, 0.00000039, 0.00000031, 0.00000035, 0.00000048, 0.00000069, 0.00000094, 0.00000121, 0.00000145, 0.00000162, 0.00000170, 0.00000166, 0.00000150, 0.00000121, 0.00000082, 0.00000034, -0.00000018, -0.00000070, -0.00000118, -0.00000157, -0.00000183, -0.00000196, -0.00000192, -0.00000174, -0.00000143, -0.00000103, -0.00000057, -0.00000011, 0.00000031, 0.00000063, 0.00000083, 0.00000089, 0.00000079, 0.00000056, 0.00000020, },
		{ -0.00000005, 0.00000040, 0.00000081, 0.00000113, 0.00000132, 0.00000138, 0.00000130, 0.00000108, 0.00000076, 0.00000037, -0.00000006, -0.00000047, -0.00000084, -0.00000113, -0.00000134, -0.00000145, -0.00000148, -0.00000146, -0.00000140, -0.00000136, -0.00000135, -0.00000142, -0.00000157, -0.00000182, -0.00000214, -0.00000252, -0.00000291, -0.00000327, -0.00000355, -0.00000371, -0.00000370, -0.00000351, -0.00000312, -0.00000256, -0.00000185, -0.00000104, -0.00000021, 0.00000057, 0.00000124, 0.00000173, 0.00000199, 0.00000199, 0.00000172, 0.00000122, 0.00000052, -0.00000030, -0.00000116, -0.00000197, -0.00000264, -0.00000308, -0.00000323, -0.00000306, -0.00000254, -0.00000171, -0.00000061, 0.00000069, 0.00000208, 0.00000346, 0.00000471, 0.00000575, 0.00000647, 0.00000681, 0.00000674, 0.00000625, 0.00000535, 0.00000411, 0.00000260, 0.00000091, -0.00000086, -0.00000261, -0.00000425, -0.00000568, -0.00000685, -0.00000771, -0.00000824, -0.00000843, -0.00000832, -0.00000793, -0.00000731, -0.00000651, -0.00000559, -0.00000460, -0.00000360, -0.00000262, -0.00000171, -0.00000089, -0.00000019, 0.00000039, 0.00000083, 0.00000114, 0.00000130, 0.00000132, 0.00000121, 0.00000100, 0.00000068, 0.00000029, -0.00000014, -0.00000059, -0.00000102, -0.00000140, -0.00000168, -0.00000185, -0.00000188, -0.00000177, -0.00000150, -0.00000110, -0.00000059, -0.00000001, 0.00000061, 0.00000120, 0.00000173, 0.00000215, 0.00000242, 0.00000252, 0.00000244, 0.00000219, 0.00000179, 0.00000128, 0.00000071, 0.00000012, -0.00000042, -0.00000087, -0.00000119, -0.00000135, -0.00000135, -0.00000119, -0.00000089, -0.00000050, },
		{ -0.00000015, -0.00000066, -0.00000108, -0.00000138, -0.00000154, -0.00000154, -0.00000140, -0.00000115, -0.00000081, -0.00000043, -0.00000007, 0.00000024, 0.00000046, 0.00000056, 0.00000052, 0.00000036, 0.00000009, -0.00000027, -0.00000066, -0.00000107, -0.00000143, -0.00000173, -0.00000194, -0.00000203, -0.00000202, -0.00000192, -0.00000175, -0.00000153, -0.00000131, -0.00000112, -0.00000097, -0.00000088, -0.00000086, -0.00000090, -0.00000097, -0.00000105, -0.00000110, -0.00000109, -0.00000097, -0.00000074, -0.00000038, 0.00000010, 0.00000067, 0.00000132, 0.00000197, 0.00000258, 0.00000309, 0.00000344, 0.00000359, 0.00000352, 0.00000321, 0.00000267, 0.00000195, 0.00000108, 0.00000015, -0.00000079, -0.00000164, -0.00000235, -0.00000285, -0.00000311, -0.00000311, -0.00000286, -0.00000238, -0.00000172, -0.00000096, -0.00000016, 0.00000058, 0.00000121, 0.00000165, 0.00000185, 0.00000180, 0.00000148, 0.00000091, 0.00000014, -0.00000079, -0.00000180, -0.00000283, -0.00000381, -0.00000468, -0.00000537, -0.00000585, -0.00000611, -0.00000613, -0.00000592, -0.00000550, -0.00000492, -0.00000420, -0.00000338, -0.00000251, -0.00000162, -0.00000075, 0.00000007, 0.00000082, 0.00000149, 0.00000204, 0.00000248, 0.00000279, 0.00000295, 0.00000297, 0.00000284, 0.00000256, 0.00000214, 0.00000159, 0.00000092, 0.00000017, -0.00000062, -0.00000142, -0.00000218, -0.00000284, -0.00000337, -0.00000373, -0.00000389, -0.00000383, -0.00000357, -0.00000311, -0.00000250, -0.00000178, -0.00000100, -0.00000022, 0.00000049, 0.00000108, 0.00000152, 0.00000176, 0.00000180, 0.00000166, 0.00000134, 0.00000090, 0.00000039, },
		{ 0.00000042, 0.00000106, 0.00000154, 0.00000186, 0.00000198, 0.00000193, 0.00000172, 0.00000140, 0.00000101, 0.00000060, 0.00000023, -0.00000007, -0.00000027, -0.00000034, -0.00000029, -0.00000013, 0.00000010, 0.00000038, 0.00000065, 0.00000088, 0.00000101, 0.00000101, 0.00000088, 0.00000059, 0.00000016, -0.00000038, -0.00000101, -0.00000169, -0.00000237, -0.00000301, -0.00000359, -0.00000405, -0.00000441, -0.00000463, -0.00000474, -0.00000473, -0.00000463, -0.00000445, -0.00000422, -0.00000394, -0.00000363, -0.00000329, -0.00000290, -0.00000246, -0.00000197, -0.00000140, -0.00000075, -0.00000002, 0.00000077, 0.00000161, 0.00000247, 0.00000331, 0.00000407, 0.00000472, 0.00000520, 0.00000548, 0.00000554, 0.00000537, 0.00000498, 0.00000439, 0.00000366, 0.00000284, 0.00000201, 0.00000123, 0.00000057, 0.00000008, -0.00000019, -0.00000022, -0.00000002, 0.00000039, 0.00000096, 0.00000162, 0.00000230, 0.00000292, 0.00000340, 0.00000366, 0.00000366, 0.00000337, 0.00000277, 0.00000188, 0.00000074, -0.00000059, -0.00000205, -0.00000355, -0.00000501, -0.00000635, -0.00000751, -0.00000843, -0.00000908, -0.00000943, -0.00000949, -0.00000927, -0.00000880, -0.00000811, -0.00000726, -0.00000627, -0.00000519, -0.00000407, -0.00000292, -0.00000178, -0.00000068, 0.00000037, 0.00000135, 0.00000223, 0.00000300, 0.00000363, 0.00000411, 0.00000440, 0.00000450, 0.00000439, 0.00000408, 0.00000357, 0.00000288, 0.00000205, 0.00000111, 0.00000013, -0.00000083, -0.00000173, -0.00000249, -0.00000306, -0.00000342, -0.00000354, -0.00000342, -0.00000307, -0.00000253, -0.00000185, -0.00000109, -0.00000031, },
		{ -0.00000305, -0.00000349, -0.00000370, -0.00000368, -0.00000344, -0.00000301, -0.00000244, -0.00000178, -0.00000109, -0.00000043, 0.00000014, 0.00000060, 0.00000092, 0.00000109, 0.00000112, 0.00000105, 0.00000090, 0.00000073, 0.00000057, 0.00000048, 0.00000049, 0.00000062, 0.00000087, 0.00000124, 0.00000170, 0.00000223, 0.00000277, 0.00000327, 0.00000371, 0.00000402, 0.00000417, 0.00000415, 0.00000395, 0.00000356, 0.00000301, 0.00000231, 0.00000149, 0.00000059, -0.00000036, -0.00000132, -0.00000228, -0.00000321, -0.00000408, -0.00000488, -0.00000560, -0.00000621, -0.00000671, -0.00000707, -0.00000727, -0.00000730, -0.00000713, -0.00000674, -0.00000613, -0.00000529, -0.00000424, -0.00000301, -0.00000163, -0.00000017, 0.00000131, 0.00000272, 0.00000400, 0.00000505, 0.00000582, 0.00000626, 0.00000634, 0.00000607, 0.00000546, 0.00000457, 0.00000348, 0.00000228, 0.00000108, -0.00000002, -0.00000092, -0.00000153, -0.00000179, -0.00000169, -0.00000120, -0.00000038, 0.00000073, 0.00000202, 0.00000340, 0.00000475, 0.00000596, 0.00000693, 0.00000758, 0.00000784, 0.00000769, 0.00000713, 0.00000618, 0.00000491, 0.00000338, 0.00000168, -0.00000008, -0.00000183, -0.00000348, -0.00000495, -0.00000621, -0.00000720, -0.00000793, -0.00000839, -0.00000860, -0.00000857, -0.00000835, -0.00000795, -0.00000741, -0.00000675, -0.00000600, -0.00000517, -0.00000427, -0.00000332, -0.00000233, -0.00000133, -0.00000033, 0.00000062, 0.00000149, 0.00000226, 0.00000286, 0.00000328, 0.00000348, 0.00000345, 0.00000318, 0.00000269, 0.00000201, 0.00000118, 0.00000026, -0.00000069, -0.00000160, -0.00000241, },
		{ 0.00000309, 0.00000299, 0.00000267, 0.00000214, 0.00000145, 0.00000065, -0.00000020, -0.00000103, -0.00000179, -0.00000241, -0.00000287, -0.00000314, -0.00000321, -0.00000311, -0.00000286, -0.00000250, -0.00000209, -0.00000168, -0.00000132, -0.00000105, -0.00000090, -0.00000088, -0.00000098, -0.00000119, -0.00000146, -0.00000177, -0.00000206, -0.00000229, -0.00000242, -0.00000241, -0.00000225, -0.00000193, -0.00000145, -0.00000084, -0.00000012, 0.00000067, 0.00000150, 0.00000231, 0.00000309, 0.00000380, 0.00000440, 0.00000489, 0.00000524, 0.00000545, 0.00000551, 0.00000541, 0.00000515, 0.00000472, 0.00000412, 0.00000335, 0.00000242, 0.00000134, 0.00000014, -0.00000117, -0.00000252, -0.00000385, -0.00000511, -0.00000621, -0.00000709, -0.00000766, -0.00000787, -0.00000768, -0.00000707, -0.00000605, -0.00000466, -0.00000297, -0.00000108, 0.00000091, 0.00000287, 0.00000466, 0.00000617, 0.00000728, 0.00000792, 0.00000806, 0.00000767, 0.00000681, 0.00000553, 0.00000395, 0.00000220, 0.00000041, -0.00000127, -0.00000271, -0.00000380, -0.00000446, -0.00000463, -0.00000432, -0.00000357, -0.00000243, -0.00000101, 0.00000057, 0.00000219, 0.00000372, 0.00000506, 0.00000610, 0.00000680, 0.00000711, 0.00000703, 0.00000658, 0.00000583, 0.00000482, 0.00000366, 0.00000240, 0.00000114, -0.00000007, -0.00000117, -0.00000214, -0.00000295, -0.00000359, -0.00000408, -0.00000444, -0.00000468, -0.00000481, -0.00000486, -0.00000482, -0.00000470, -0.00000449, -0.00000418, -0.00000377, -0.00000324, -0.00000261, -0.00000188, -0.00000108, -0.00000024, 0.00000060, 0.00000139, 0.00000207, 0.00000261, 0.00000296, },
		{ -0.00000244, -0.00000199, -0.00000142, -0.00000076, -0.00000005, 0.00000067, 0.00000132, 0.00000187, 0.00000228, 0.00000250, 0.00000253, 0.00000236, 0.00000203, 0.00000156, 0.00000101, 0.00000043, -0.00000012, -0.00000059, -0.00000093, -0.00000111, -0.00000113, -0.00000099, -0.00000071, -0.00000033, 0.00000010, 0.00000054, 0.00000094, 0.00000124, 0.00000141, 0.00000143, 0.00000130, 0.00000101, 0.00000059, 0.00000006, -0.00000054, -0.00000117, -0.00000179, -0.00000238, -0.00000291, -0.00000335, -0.00000370, -0.00000394, -0.00000409, -0.00000412, -0.00000406, -0.00000389, -0.00000361, -0.00000322, -0.00000272, -0.00000211, -0.00000138, -0.00000056, 0.00000035, 0.00000130, 0.00000225, 0.00000315, 0.00000393, 0.00000453, 0.00000489, 0.00000495, 0.00000469, 0.00000408, 0.00000315, 0.00000193, 0.00000050, -0.00000106, -0.00000263, -0.00000408, -0.00000530, -0.00000617, -0.00000660, -0.00000654, -0.00000594, -0.00000484, -0.00000329, -0.00000138, 0.00000075, 0.00000296, 0.00000509, 0.00000698, 0.00000851, 0.00000955, 0.00001005, 0.00000997, 0.00000934, 0.00000821, 0.00000670, 0.00000493, 0.00000304, 0.00000118, -0.00000049, -0.00000187, -0.00000287, -0.00000342, -0.00000353, -0.00000321, -0.00000253, -0.00000158, -0.00000045, 0.00000073, 0.00000186, 0.00000285, 0.00000362, 0.00000412, 0.00000433, 0.00000427, 0.00000396, 0.00000346, 0.00000282, 0.00000211, 0.00000139, 0.00000071, 0.00000010, -0.00000042, -0.00000083, -0.00000115, -0.00000140, -0.00000161, -0.00000179, -0.00000197, -0.00000217, -0.00000237, -0.00000258, -0.00000276, -0.00000290, -0.00000296, -0.00000292, -0.00000275, },
		{ 0.00000055, 0.00000038, 0.00000011, -0.00000024, -0.00000062, -0.00000100, -0.00000133, -0.00000158, -0.00000171, -0.00000170, -0.00000155, -0.00000127, -0.00000090, -0.00000047, -0.00000003, 0.00000036, 0.00000065, 0.00000082, 0.00000082, 0.00000065, 0.00000033, -0.00000012, -0.00000066, -0.00000124, -0.00000180, -0.00000228, -0.00000264, -0.00000284, -0.00000285, -0.00000267, -0.00000231, -0.00000179, -0.00000114, -0.00000042, 0.00000035, 0.00000111, 0.00000183, 0.00000248, 0.00000305, 0.00000352, 0.00000389, 0.00000416, 0.00000435, 0.00000444, 0.00000443, 0.00000434, 0.00000413, 0.00000380, 0.00000334, 0.00000273, 0.00000196, 0.00000104, -0.00000000, -0.00000115, -0.00000234, -0.00000352, -0.00000461, -0.00000555, -0.00000625, -0.00000667, -0.00000676, -0.00000649, -0.00000589, -0.00000499, -0.00000387, -0.00000260, -0.00000131, -0.00000009, 0.00000092, 0.00000165, 0.00000201, 0.00000197, 0.00000152, 0.00000071, -0.00000039, -0.00000168, -0.00000302, -0.00000428, -0.00000531, -0.00000600, -0.00000626, -0.00000603, -0.00000529, -0.00000407, -0.00000246, -0.00000055, 0.00000150, 0.00000356, 0.00000547, 0.00000710, 0.00000832, 0.00000907, 0.00000931, 0.00000905, 0.00000835, 0.00000728, 0.00000596, 0.00000453, 0.00000311, 0.00000182, 0.00000077, 0.00000003, -0.00000037, -0.00000042, -0.00000016, 0.00000034, 0.00000102, 0.00000178, 0.00000252, 0.00000316, 0.00000364, 0.00000391, 0.00000396, 0.00000380, 0.00000346, 0.00000298, 0.00000243, 0.00000186, 0.00000134, 0.00000091, 0.00000060, 0.00000041, 0.00000034, 0.00000037, 0.00000045, 0.00000055, 0.00000062, 0.00000063, },
		{ 0.00000008, 0.00000009, 0.00000028, 0.00000061, 0.00000104, 0.00000151, 0.00000197, 0.00000235, 0.00000262, 0.00000275, 0.00000274, 0.00000259, 0.00000234, 0.00000203, 0.00000171, 0.00000142, 0.00000123, 0.00000115, 0.00000120, 0.00000138, 0.00000167, 0.00000204, 0.00000243, 0.00000280, 0.00000308, 0.00000323, 0.00000320, 0.00000299, 0.00000259, 0.00000200, 0.00000127, 0.00000042, -0.00000048, -0.00000138, -0.00000225, -0.00000302, -0.00000369, -0.00000422, -0.00000461, -0.00000487, -0.00000501, -0.00000504, -0.00000498, -0.00000484, -0.00000461, -0.00000429, -0.00000387, -0.00000332, -0.00000263, -0.00000177, -0.00000075, 0.00000043, 0.00000175, 0.00000317, 0.00000462, 0.00000602, 0.00000729, 0.00000832, 0.00000905, 0.00000939, 0.00000929, 0.00000874, 0.00000775, 0.00000637, 0.00000468, 0.00000279, 0.00000082, -0.00000109, -0.00000282, -0.00000426, -0.00000531, -0.00000593, -0.00000611, -0.00000588, -0.00000531, -0.00000449, -0.00000355, -0.00000261, -0.00000179, -0.00000119, -0.00000090, -0.00000095, -0.00000133, -0.00000201, -0.00000291, -0.00000393, -0.00000493, -0.00000581, -0.00000643, -0.00000672, -0.00000660, -0.00000606, -0.00000512, -0.00000383, -0.00000228, -0.00000060, 0.00000109, 0.00000267, 0.00000400, 0.00000501, 0.00000563, 0.00000585, 0.00000568, 0.00000517, 0.00000440, 0.00000349, 0.00000253, 0.00000164, 0.00000091, 0.00000040, 0.00000017, 0.00000021, 0.00000050, 0.00000100, 0.00000162, 0.00000228, 0.00000291, 0.00000342, 0.00000375, 0.00000387, 0.00000376, 0.00000345, 0.00000296, 0.00000237, 0.00000174, 0.00000114, 0.00000063, 0.00000027, },
		{ 0.00000316, 0.00000265, 0.00000194, 0.00000109, 0.00000019, -0.00000067, -0.00000142, -0.00000199, -0.00000234, -0.00000244, -0.00000231, -0.00000197, -0.00000147, -0.00000089, -0.00000029, 0.00000026, 0.00000070, 0.00000100, 0.00000113, 0.00000111, 0.00000095, 0.00000071, 0.00000044, 0.00000020, 0.00000005, 0.00000004, 0.00000020, 0.00000053, 0.00000104, 0.00000169, 0.00000244, 0.00000324, 0.00000402, 0.00000475, 0.00000535, 0.00000581, 0.00000611, 0.00000623, 0.00000621, 0.00000605, 0.00000579, 0.00000546, 0.00000508, 0.00000467, 0.00000423, 0.00000375, 0.00000322, 0.00000259, 0.00000185, 0.00000097, -0.00000006, -0.00000125, -0.00000256, -0.00000394, -0.00000533, -0.00000664, -0.00000776, -0.00000861, -0.00000909, -0.00000912, -0.00000866, -0.00000770, -0.00000627, -0.00000443, -0.00000230, -0.00000000, 0.00000231, 0.00000449, 0.00000638, 0.00000785, 0.00000883, 0.00000925, 0.00000910, 0.00000844, 0.00000733, 0.00000590, 0.00000428, 0.00000262, 0.00000104, -0.00000032, -0.00000139, -0.00000211, -0.00000247, -0.00000250, -0.00000226, -0.00000185, -0.00000137, -0.00000094, -0.00000063, -0.00000053, -0.00000069, -0.00000110, -0.00000174, -0.00000255, -0.00000345, -0.00000433, -0.00000509, -0.00000564, -0.00000589, -0.00000582, -0.00000540, -0.00000465, -0.00000364, -0.00000245, -0.00000118, 0.00000006, 0.00000116, 0.00000204, 0.00000263, 0.00000291, 0.00000287, 0.00000255, 0.00000202, 0.00000136, 0.00000067, 0.00000004, -0.00000044, -0.00000072, -0.00000074, -0.00000051, -0.00000006, 0.00000056, 0.00000128, 0.00000201, 0.00000265, 0.00000314, 0.00000341, 0.00000342, },
		{ -0.00000010, 0.00000080, 0.00000178, 0.00000277, 0.00000365, 0.00000433, 0.00000474, 0.00000483, 0.00000458, 0.00000402, 0.00000319, 0.00000217, 0.00000105, -0.00000006, -0.00000108, -0.00000192, -0.00000252, -0.00000284, -0.00000289, -0.00000269, -0.00000230, -0.00000179, -0.00000126, -0.00000077, -0.00000041, -0.00000023, -0.00000026, -0.00000050, -0.00000095, -0.00000154, -0.00000223, -0.00000293, -0.00000358, -0.00000412, -0.00000448, -0.00000466, -0.00000463, -0.00000441, -0.00000402, -0.00000352, -0.00000295, -0.00000234, -0.00000175, -0.00000119, -0.00000066, -0.00000014, 0.00000038, 0.00000095, 0.00000161, 0.00000240, 0.00000334, 0.00000443, 0.00000565, 0.00000696, 0.00000827, 0.00000949, 0.00001052, 0.00001126, 0.00001160, 0.00001147, 0.00001082, 0.00000966, 0.00000803, 0.00000599, 0.00000367, 0.00000123, -0.00000119, -0.00000341, -0.00000528, -0.00000667, -0.00000750, -0.00000773, -0.00000736, -0.00000645, -0.00000511, -0.00000347, -0.00000169, 0.00000008, 0.00000168, 0.00000298, 0.00000390, 0.00000438, 0.00000442, 0.00000407, 0.00000339, 0.00000249, 0.00000149, 0.00000050, -0.00000037, -0.00000104, -0.00000147, -0.00000162, -0.00000154, -0.00000126, -0.00000086, -0.00000043, -0.00000006, 0.00000018, 0.00000024, 0.00000007, -0.00000031, -0.00000089, -0.00000159, -0.00000235, -0.00000307, -0.00000368, -0.00000408, -0.00000424, -0.00000412, -0.00000373, -0.00000311, -0.00000230, -0.00000141, -0.00000051, 0.00000030, 0.00000094, 0.00000135, 0.00000150, 0.00000140, 0.00000107, 0.00000057, -0.00000002, -0.00000061, -0.00000109, -0.00000141, -0.00000148, -0.00000128, -0.00000081, },
		{ -0.00000124, -0.00000191, -0.00000254, -0.00000304, -0.00000331, -0.00000328, -0.00000293, -0.00000225, -0.00000127, -0.00000005, 0.00000129, 0.00000266, 0.00000392, 0.00000498, 0.00000572, 0.00000610, 0.00000608, 0.00000567, 0.00000491, 0.00000390, 0.00000273, 0.00000152, 0.00000039, -0.00000055, -0.00000123, -0.00000159, -0.00000162, -0.00000134, -0.00000082, -0.00000012, 0.00000065, 0.00000140, 0.00000202, 0.00000244, 0.00000260, 0.00000249, 0.00000211, 0.00000151, 0.00000074, -0.00000013, -0.00000102, -0.00000186, -0.00000259, -0.00000319, -0.00000364, -0.00000398, -0.00000422, -0.00000445, -0.00000471, -0.00000506, -0.00000555, -0.00000619, -0.00000697, -0.00000785, -0.00000875, -0.00000959, -0.00001025, -0.00001061, -0.00001058, -0.00001007, -0.00000903, -0.00000748, -0.00000544, -0.00000300, -0.00000030, 0.00000250, 0.00000524, 0.00000774, 0.00000985, 0.00001142, 0.00001239, 0.00001270, 0.00001239, 0.00001151, 0.00001019, 0.00000855, 0.00000678, 0.00000503, 0.00000345, 0.00000217, 0.00000127, 0.00000079, 0.00000071, 0.00000100, 0.00000155, 0.00000226, 0.00000301, 0.00000367, 0.00000414, 0.00000435, 0.00000426, 0.00000387, 0.00000320, 0.00000232, 0.00000131, 0.00000028, -0.00000068, -0.00000149, -0.00000209, -0.00000244, -0.00000254, -0.00000242, -0.00000214, -0.00000176, -0.00000137, -0.00000104, -0.00000083, -0.00000079, -0.00000094, -0.00000127, -0.00000174, -0.00000228, -0.00000284, -0.00000334, -0.00000370, -0.00000387, -0.00000383, -0.00000357, -0.00000310, -0.00000249, -0.00000180, -0.00000112, -0.00000052, -0.00000009, 0.00000013, 0.00000011, -0.00000015, -0.00000062, },
		{ -0.00000189, -0.00000137, -0.00000095, -0.00000070, -0.00000067, -0.00000091, -0.00000141, -0.00000213, -0.00000299, -0.00000392, -0.00000480, -0.00000551, -0.00000595, -0.00000604, -0.00000572, -0.00000499, -0.00000387, -0.00000244, -0.00000080, 0.00000092, 0.00000257, 0.00000401, 0.00000513, 0.00000583, 0.00000607, 0.00000583, 0.00000516, 0.00000414, 0.00000288, 0.00000152, 0.00000020, -0.00000094, -0.00000180, -0.00000230, -0.00000241, -0.00000214, -0.00000153, -0.00000068, 0.00000031, 0.00000132, 0.00000223, 0.00000295, 0.00000340, 0.00000354, 0.00000340, 0.00000301, 0.00000244, 0.00000180, 0.00000117, 0.00000068, 0.00000038, 0.00000033, 0.00000054, 0.00000098, 0.00000157, 0.00000222, 0.00000279, 0.00000317, 0.00000322, 0.00000285, 0.00000200, 0.00000065, -0.00000116, -0.00000335, -0.00000579, -0.00000831, -0.00001074, -0.00001289, -0.00001460, -0.00001575, -0.00001623, -0.00001602, -0.00001513, -0.00001365, -0.00001169, -0.00000940, -0.00000696, -0.00000454, -0.00000229, -0.00000033, 0.00000124, 0.00000238, 0.00000311, 0.00000347, 0.00000355, 0.00000345, 0.00000329, 0.00000317, 0.00000318, 0.00000339, 0.00000381, 0.00000444, 0.00000523, 0.00000610, 0.00000697, 0.00000774, 0.00000833, 0.00000865, 0.00000866, 0.00000835, 0.00000774, 0.00000687, 0.00000582, 0.00000468, 0.00000353, 0.00000248, 0.00000158, 0.00000088, 0.00000041, 0.00000016, 0.00000009, 0.00000015, 0.00000027, 0.00000037, 0.00000040, 0.00000031, 0.00000007, -0.00000032, -0.00000083, -0.00000142, -0.00000203, -0.00000258, -0.00000302, -0.00000329, -0.00000335, -0.00000321, -0.00000289, -0.00000242, },
		{ 0.00000055, -0.00000008, -0.00000062, -0.00000100, -0.00000118, -0.00000115, -0.00000092, -0.00000055, -0.00000011, 0.00000033, 0.00000067, 0.00000081, 0.00000070, 0.00000029, -0.00000042, -0.00000141, -0.00000259, -0.00000387, -0.00000513, -0.00000622, -0.00000702, -0.00000742, -0.00000734, -0.00000676, -0.00000569, -0.00000420, -0.00000240, -0.00000043, 0.00000154, 0.00000334, 0.00000481, 0.00000583, 0.00000632, 0.00000624, 0.00000562, 0.00000454, 0.00000313, 0.00000153, -0.00000007, -0.00000151, -0.00000264, -0.00000336, -0.00000360, -0.00000335, -0.00000268, -0.00000166, -0.00000045, 0.00000080, 0.00000192, 0.00000278, 0.00000324, 0.00000325, 0.00000278, 0.00000187, 0.00000063, -0.00000082, -0.00000230, -0.00000365, -0.00000468, -0.00000527, -0.00000531, -0.00000475, -0.00000363, -0.00000202, -0.00000004, 0.00000212, 0.00000428, 0.00000624, 0.00000780, 0.00000883, 0.00000921, 0.00000891, 0.00000793, 0.00000635, 0.00000430, 0.00000192, -0.00000059, -0.00000307, -0.00000535, -0.00000729, -0.00000879, -0.00000982, -0.00001037, -0.00001049, -0.00001026, -0.00000978, -0.00000915, -0.00000848, -0.00000786, -0.00000736, -0.00000698, -0.00000674, -0.00000658, -0.00000646, -0.00000631, -0.00000603, -0.00000558, -0.00000491, -0.00000400, -0.00000286, -0.00000153, -0.00000008, 0.00000140, 0.00000282, 0.00000408, 0.00000510, 0.00000583, 0.00000623, 0.00000631, 0.00000610, 0.00000565, 0.00000504, 0.00000437, 0.00000370, 0.00000312, 0.00000269, 0.00000242, 0.00000233, 0.00000239, 0.00000255, 0.00000275, 0.00000292, 0.00000301, 0.00000297, 0.00000275, 0.00000237, 0.00000185, 0.00000122, },
		{ 0.00000094, 0.00000147, 0.00000190, 0.00000218, 0.00000226, 0.00000213, 0.00000181, 0.00000133, 0.00000076, 0.00000018, -0.00000033, -0.00000071, -0.00000089, -0.00000085, -0.00000060, -0.00000017, 0.00000037, 0.00000092, 0.00000138, 0.00000166, 0.00000165, 0.00000131, 0.00000061, -0.00000043, -0.00000174, -0.00000321, -0.00000471, -0.00000608, -0.00000716, -0.00000781, -0.00000792, -0.00000744, -0.00000635, -0.00000472, -0.00000265, -0.00000031, 0.00000211, 0.00000440, 0.00000634, 0.00000777, 0.00000855, 0.00000862, 0.00000796, 0.00000665, 0.00000482, 0.00000266, 0.00000039, -0.00000175, -0.00000354, -0.00000480, -0.00000540, -0.00000526, -0.00000441, -0.00000293, -0.00000097, 0.00000126, 0.00000352, 0.00000557, 0.00000718, 0.00000817, 0.00000840, 0.00000783, 0.00000647, 0.00000442, 0.00000184, -0.00000105, -0.00000401, -0.00000679, -0.00000917, -0.00001095, -0.00001201, -0.00001227, -0.00001175, -0.00001050, -0.00000867, -0.00000642, -0.00000395, -0.00000146, 0.00000086, 0.00000285, 0.00000439, 0.00000544, 0.00000597, 0.00000603, 0.00000568, 0.00000504, 0.00000420, 0.00000329, 0.00000239, 0.00000158, 0.00000090, 0.00000036, -0.00000006, -0.00000040, -0.00000071, -0.00000106, -0.00000148, -0.00000201, -0.00000266, -0.00000340, -0.00000419, -0.00000498, -0.00000569, -0.00000624, -0.00000658, -0.00000664, -0.00000640, -0.00000587, -0.00000508, -0.00000410, -0.00000299, -0.00000185, -0.00000078, 0.00000013, 0.00000083, 0.00000127, 0.00000145, 0.00000137, 0.00000109, 0.00000068, 0.00000021, -0.00000022, -0.00000055, -0.00000072, -0.00000071, -0.00000049, -0.00000011, 0.00000038, },
		{ 0.00000008, -0.00000071, -0.00000135, -0.00000178, -0.00000196, -0.00000187, -0.00000152, -0.00000099, -0.00000032, 0.00000038, 0.00000102, 0.00000152, 0.00000182, 0.00000188, 0.00000169, 0.00000130, 0.00000074, 0.00000011, -0.00000051, -0.00000103, -0.00000138, -0.00000151, -0.00000140, -0.00000109, -0.00000064, -0.00000014, 0.00000031, 0.00000057, 0.00000055, 0.00000015, -0.00000067, -0.00000191, -0.00000350, -0.00000534, -0.00000726, -0.00000909, -0.00001063, -0.00001168, -0.00001207, -0.00001169, -0.00001050, -0.00000851, -0.00000583, -0.00000263, 0.00000086, 0.00000436, 0.00000760, 0.00001031, 0.00001223, 0.00001322, 0.00001316, 0.00001207, 0.00001004, 0.00000725, 0.00000395, 0.00000043, -0.00000298, -0.00000596, -0.00000825, -0.00000964, -0.00000999, -0.00000928, -0.00000757, -0.00000501, -0.00000184, 0.00000167, 0.00000520, 0.00000846, 0.00001115, 0.00001306, 0.00001404, 0.00001402, 0.00001301, 0.00001113, 0.00000854, 0.00000545, 0.00000211, -0.00000124, -0.00000435, -0.00000704, -0.00000918, -0.00001067, -0.00001150, -0.00001171, -0.00001138, -0.00001063, -0.00000958, -0.00000838, -0.00000714, -0.00000596, -0.00000491, -0.00000401, -0.00000327, -0.00000267, -0.00000216, -0.00000170, -0.00000125, -0.00000078, -0.00000027, 0.00000027, 0.00000081, 0.00000131, 0.00000173, 0.00000200, 0.00000210, 0.00000200, 0.00000170, 0.00000122, 0.00000060, -0.00000009, -0.00000078, -0.00000138, -0.00000181, -0.00000203, -0.00000200, -0.00000171, -0.00000118, -0.00000047, 0.00000034, 0.00000118, 0.00000194, 0.00000253, 0.00000290, 0.00000300, 0.00000281, 0.00000237, 0.00000171, 0.00000092, },
		{ 0.00000356, 0.00000451, 0.00000517, 0.00000547, 0.00000539, 0.00000494, 0.00000418, 0.00000320, 0.00000208, 0.00000097, -0.00000003, -0.00000081, -0.00000129, -0.00000144, -0.00000126, -0.00000077, -0.00000005, 0.00000081, 0.00000170, 0.00000251, 0.00000316, 0.00000358, 0.00000374, 0.00000365, 0.00000334, 0.00000291, 0.00000243, 0.00000200, 0.00000172, 0.00000165, 0.00000182, 0.00000221, 0.00000278, 0.00000343, 0.00000401, 0.00000439, 0.00000440, 0.00000393, 0.00000287, 0.00000119, -0.00000109, -0.00000386, -0.00000697, -0.00001019, -0.00001326, -0.00001591, -0.00001788, -0.00001893, -0.00001890, -0.00001771, -0.00001537, -0.00001199, -0.00000777, -0.00000301, 0.00000196, 0.00000676, 0.00001104, 0.00001446, 0.00001676, 0.00001776, 0.00001740, 0.00001573, 0.00001290, 0.00000915, 0.00000482, 0.00000026, -0.00000415, -0.00000805, -0.00001113, -0.00001318, -0.00001406, -0.00001374, -0.00001229, -0.00000988, -0.00000672, -0.00000312, 0.00000064, 0.00000425, 0.00000744, 0.00001002, 0.00001182, 0.00001278, 0.00001291, 0.00001226, 0.00001098, 0.00000920, 0.00000713, 0.00000492, 0.00000274, 0.00000072, -0.00000105, -0.00000252, -0.00000367, -0.00000453, -0.00000514, -0.00000556, -0.00000584, -0.00000604, -0.00000618, -0.00000630, -0.00000638, -0.00000642, -0.00000639, -0.00000626, -0.00000602, -0.00000565, -0.00000516, -0.00000458, -0.00000393, -0.00000328, -0.00000269, -0.00000220, -0.00000187, -0.00000172, -0.00000177, -0.00000198, -0.00000232, -0.00000271, -0.00000310, -0.00000338, -0.00000348, -0.00000334, -0.00000293, -0.00000223, -0.00000128, -0.00000013, 0.00000113, 0.00000240, },
		{ -0.00000654, -0.00000664, -0.00000636, -0.00000568, -0.00000460, -0.00000320, -0.00000157, 0.00000017, 0.00000187, 0.00000340, 0.00000462, 0.00000543, 0.00000578, 0.00000564, 0.00000505, 0.00000407, 0.00000282, 0.00000144, 0.00000006, -0.00000118, -0.00000215, -0.00000279, -0.00000305, -0.00000294, -0.00000250, -0.00000182, -0.00000100, -0.00000017, 0.00000057, 0.00000113, 0.00000145, 0.00000153, 0.00000139, 0.00000111, 0.00000080, 0.00000057, 0.00000056, 0.00000086, 0.00000154, 0.00000264, 0.00000410, 0.00000584, 0.00000771, 0.00000950, 0.00001100, 0.00001200, 0.00001228, 0.00001170, 0.00001018, 0.00000771, 0.00000439, 0.00000038, -0.00000406, -0.00000861, -0.00001293, -0.00001666, -0.00001949, -0.00002116, -0.00002148, -0.00002039, -0.00001793, -0.00001425, -0.00000961, -0.00000433, 0.00000118, 0.00000653, 0.00001132, 0.00001521, 0.00001792, 0.00001929, 0.00001927, 0.00001790, 0.00001535, 0.00001185, 0.00000774, 0.00000334, -0.00000099, -0.00000493, -0.00000822, -0.00001064, -0.00001210, -0.00001257, -0.00001209, -0.00001081, -0.00000891, -0.00000659, -0.00000409, -0.00000162, 0.00000063, 0.00000253, 0.00000399, 0.00000497, 0.00000547, 0.00000554, 0.00000527, 0.00000475, 0.00000407, 0.00000331, 0.00000256, 0.00000185, 0.00000120, 0.00000063, 0.00000012, -0.00000035, -0.00000080, -0.00000124, -0.00000168, -0.00000212, -0.00000254, -0.00000290, -0.00000318, -0.00000334, -0.00000338, -0.00000327, -0.00000304, -0.00000271, -0.00000233, -0.00000196, -0.00000166, -0.00000151, -0.00000154, -0.00000179, -0.00000226, -0.00000292, -0.00000373, -0.00000459, -0.00000542, -0.00000610, },
		{ 0.00000145, 0.00000124, 0.00000067, -0.00000024, -0.00000144, -0.00000281, -0.00000426, -0.00000562, -0.00000676, -0.00000755, -0.00000788, -0.00000769, -0.00000695, -0.00000571, -0.00000403, -0.00000206, 0.00000006, 0.00000215, 0.00000404, 0.00000559, 0.00000666, 0.00000720, 0.00000719, 0.00000666, 0.00000571, 0.00000446, 0.00000306, 0.00000167, 0.00000043, -0.00000053, -0.00000115, -0.00000140, -0.00000130, -0.00000094, -0.00000043, 0.00000010, 0.00000050, 0.00000065, 0.00000047, -0.00000011, -0.00000106, -0.00000232, -0.00000376, -0.00000521, -0.00000649, -0.00000740, -0.00000777, -0.00000747, -0.00000642, -0.00000464, -0.00000221, 0.00000072, 0.00000391, 0.00000710, 0.00000999, 0.00001230, 0.00001377, 0.00001424, 0.00001358, 0.00001180, 0.00000900, 0.00000536, 0.00000115, -0.00000328, -0.00000758, -0.00001139, -0.00001440, -0.00001633, -0.00001703, -0.00001643, -0.00001456, -0.00001158, -0.00000772, -0.00000328, 0.00000139, 0.00000593, 0.00001003, 0.00001339, 0.00001580, 0.00001714, 0.00001737, 0.00001655, 0.00001483, 0.00001240, 0.00000950, 0.00000640, 0.00000334, 0.00000054, -0.00000182, -0.00000362, -0.00000482, -0.00000541, -0.00000545, -0.00000505, -0.00000432, -0.00000341, -0.00000244, -0.00000152, -0.00000074, -0.00000014, 0.00000024, 0.00000042, 0.00000042, 0.00000029, 0.00000006, -0.00000021, -0.00000050, -0.00000078, -0.00000105, -0.00000131, -0.00000157, -0.00000184, -0.00000213, -0.00000245, -0.00000277, -0.00000309, -0.00000335, -0.00000353, -0.00000359, -0.00000348, -0.00000319, -0.00000271, -0.00000208, -0.00000132, -0.00000052, 0.00000025, 0.00000089, 0.00000131, },
		{ -0.00000586, -0.00000531, -0.00000444, -0.00000330, -0.00000198, -0.00000059, 0.00000074, 0.00000187, 0.00000268, 0.00000306, 0.00000295, 0.00000232, 0.00000121, -0.00000031, -0.00000211, -0.00000404, -0.00000591, -0.00000755, -0.00000879, -0.00000949, -0.00000959, -0.00000905, -0.00000791, -0.00000627, -0.00000426, -0.00000207, 0.00000011, 0.00000209, 0.00000371, 0.00000485, 0.00000544, 0.00000548, 0.00000503, 0.00000420, 0.00000316, 0.00000207, 0.00000113, 0.00000049, 0.00000027, 0.00000055, 0.00000132, 0.00000251, 0.00000401, 0.00000564, 0.00000718, 0.00000843, 0.00000918, 0.00000930, 0.00000866, 0.00000727, 0.00000516, 0.00000248, -0.00000058, -0.00000378, -0.00000683, -0.00000949, -0.00001151, -0.00001270, -0.00001297, -0.00001230, -0.00001074, -0.00000845, -0.00000566, -0.00000263, 0.00000035, 0.00000298, 0.00000503, 0.00000630, 0.00000668, 0.00000615, 0.00000477, 0.00000269, 0.00000013, -0.00000265, -0.00000536, -0.00000774, -0.00000954, -0.00001059, -0.00001076, -0.00001003, -0.00000845, -0.00000616, -0.00000333, -0.00000021, 0.00000298, 0.00000598, 0.00000858, 0.00001063, 0.00001201, 0.00001267, 0.00001264, 0.00001200, 0.00001087, 0.00000940, 0.00000774, 0.00000607, 0.00000452, 0.00000319, 0.00000216, 0.00000145, 0.00000106, 0.00000095, 0.00000105, 0.00000129, 0.00000159, 0.00000190, 0.00000215, 0.00000230, 0.00000236, 0.00000230, 0.00000216, 0.00000194, 0.00000168, 0.00000138, 0.00000105, 0.00000070, 0.00000031, -0.00000014, -0.00000066, -0.00000126, -0.00000195, -0.00000271, -0.00000352, -0.00000431, -0.00000503, -0.00000562, -0.00000599, -0.00000609, },
		{ 0.00000218, 0.00000110, -0.00000017, -0.00000158, -0.00000302, -0.00000439, -0.00000557, -0.00000644, -0.00000690, -0.00000688, -0.00000637, -0.00000537, -0.00000395, -0.00000223, -0.00000036, 0.00000150, 0.00000316, 0.00000446, 0.00000526, 0.00000548, 0.00000506, 0.00000405, 0.00000253, 0.00000063, -0.00000145, -0.00000353, -0.00000538, -0.00000682, -0.00000772, -0.00000798, -0.00000757, -0.00000656, -0.00000506, -0.00000323, -0.00000129, 0.00000055, 0.00000206, 0.00000306, 0.00000344, 0.00000312, 0.00000214, 0.00000058, -0.00000138, -0.00000353, -0.00000561, -0.00000737, -0.00000857, -0.00000902, -0.00000861, -0.00000729, -0.00000512, -0.00000224, 0.00000116, 0.00000479, 0.00000837, 0.00001159, 0.00001420, 0.00001596, 0.00001674, 0.00001648, 0.00001520, 0.00001302, 0.00001014, 0.00000679, 0.00000326, -0.00000018, -0.00000329, -0.00000585, -0.00000772, -0.00000883, -0.00000921, -0.00000892, -0.00000812, -0.00000698, -0.00000570, -0.00000449, -0.00000352, -0.00000290, -0.00000272, -0.00000298, -0.00000364, -0.00000459, -0.00000570, -0.00000680, -0.00000773, -0.00000835, -0.00000854, -0.00000823, -0.00000741, -0.00000610, -0.00000440, -0.00000243, -0.00000032, 0.00000177, 0.00000370, 0.00000533, 0.00000658, 0.00000740, 0.00000777, 0.00000772, 0.00000732, 0.00000666, 0.00000583, 0.00000495, 0.00000410, 0.00000335, 0.00000276, 0.00000235, 0.00000213, 0.00000207, 0.00000214, 0.00000230, 0.00000250, 0.00000272, 0.00000292, 0.00000310, 0.00000326, 0.00000340, 0.00000353, 0.00000367, 0.00000382, 0.00000398, 0.00000411, 0.00000418, 0.00000416, 0.00000399, 0.00000362, 0.00000302, },
		{ 0.00000153, 0.00000222, 0.00000303, 0.00000387, 0.00000467, 0.00000533, 0.00000576, 0.00000586, 0.00000556, 0.00000484, 0.00000369, 0.00000217, 0.00000037, -0.00000158, -0.00000353, -0.00000531, -0.00000676, -0.00000774, -0.00000814, -0.00000793, -0.00000710, -0.00000573, -0.00000393, -0.00000189, 0.00000019, 0.00000212, 0.00000370, 0.00000475, 0.00000519, 0.00000495, 0.00000408, 0.00000267, 0.00000088, -0.00000106, -0.00000293, -0.00000449, -0.00000552, -0.00000586, -0.00000541, -0.00000417, -0.00000219, 0.00000035, 0.00000325, 0.00000622, 0.00000896, 0.00001119, 0.00001264, 0.00001312, 0.00001253, 0.00001083, 0.00000811, 0.00000455, 0.00000042, -0.00000398, -0.00000831, -0.00001220, -0.00001536, -0.00001754, -0.00001856, -0.00001834, -0.00001692, -0.00001441, -0.00001103, -0.00000705, -0.00000276, 0.00000151, 0.00000547, 0.00000886, 0.00001148, 0.00001323, 0.00001406, 0.00001402, 0.00001320, 0.00001177, 0.00000990, 0.00000780, 0.00000565, 0.00000361, 0.00000179, 0.00000028, -0.00000092, -0.00000182, -0.00000248, -0.00000297, -0.00000337, -0.00000377, -0.00000422, -0.00000476, -0.00000539, -0.00000607, -0.00000676, -0.00000738, -0.00000786, -0.00000810, -0.00000807, -0.00000771, -0.00000701, -0.00000600, -0.00000474, -0.00000330, -0.00000176, -0.00000025, 0.00000116, 0.00000237, 0.00000333, 0.00000399, 0.00000436, 0.00000444, 0.00000429, 0.00000396, 0.00000351, 0.00000301, 0.00000251, 0.00000207, 0.00000169, 0.00000140, 0.00000118, 0.00000102, 0.00000089, 0.00000078, 0.00000066, 0.00000054, 0.00000042, 0.00000034, 0.00000032, 0.00000040, 0.00000062, 0.00000099, },
		{ 0.00000122, 0.00000053, -0.00000025, -0.00000106, -0.00000182, -0.00000243, -0.00000282, -0.00000291, -0.00000265, -0.00000202, -0.00000106, 0.00000019, 0.00000161, 0.00000309, 0.00000446, 0.00000559, 0.00000632, 0.00000657, 0.00000625, 0.00000537, 0.00000396, 0.00000213, 0.00000003, -0.00000217, -0.00000425, -0.00000602, -0.00000730, -0.00000798, -0.00000798, -0.00000730, -0.00000603, -0.00000429, -0.00000229, -0.00000026, 0.00000155, 0.00000292, 0.00000366, 0.00000363, 0.00000278, 0.00000115, -0.00000114, -0.00000388, -0.00000681, -0.00000962, -0.00001199, -0.00001365, -0.00001433, -0.00001388, -0.00001223, -0.00000940, -0.00000555, -0.00000092, 0.00000419, 0.00000940, 0.00001433, 0.00001860, 0.00002188, 0.00002391, 0.00002455, 0.00002373, 0.00002152, 0.00001810, 0.00001370, 0.00000866, 0.00000334, -0.00000191, -0.00000673, -0.00001083, -0.00001399, -0.00001608, -0.00001704, -0.00001692, -0.00001582, -0.00001394, -0.00001148, -0.00000868, -0.00000578, -0.00000299, -0.00000047, 0.00000165, 0.00000329, 0.00000442, 0.00000508, 0.00000531, 0.00000518, 0.00000479, 0.00000421, 0.00000353, 0.00000278, 0.00000202, 0.00000125, 0.00000049, -0.00000027, -0.00000104, -0.00000182, -0.00000260, -0.00000337, -0.00000411, -0.00000476, -0.00000528, -0.00000564, -0.00000578, -0.00000568, -0.00000533, -0.00000474, -0.00000393, -0.00000296, -0.00000188, -0.00000078, 0.00000029, 0.00000126, 0.00000208, 0.00000271, 0.00000314, 0.00000339, 0.00000348, 0.00000344, 0.00000333, 0.00000318, 0.00000304, 0.00000293, 0.00000285, 0.00000280, 0.00000274, 0.00000265, 0.00000249, 0.00000221, 0.00000179, },
		{ 0.00000330, 0.00000393, 0.00000462, 0.00000532, 0.00000594, 0.00000640, 0.00000660, 0.00000648, 0.00000601, 0.00000518, 0.00000404, 0.00000264, 0.00000111, -0.00000043, -0.00000184, -0.00000296, -0.00000369, -0.00000394, -0.00000367, -0.00000289, -0.00000168, -0.00000015, 0.00000154, 0.00000320, 0.00000465, 0.00000570, 0.00000623, 0.00000612, 0.00000537, 0.00000400, 0.00000214, -0.00000005, -0.00000234, -0.00000451, -0.00000631, -0.00000753, -0.00000801, -0.00000766, -0.00000646, -0.00000451, -0.00000197, 0.00000094, 0.00000391, 0.00000664, 0.00000881, 0.00001015, 0.00001045, 0.00000959, 0.00000754, 0.00000440, 0.00000035, -0.00000432, -0.00000927, -0.00001410, -0.00001842, -0.00002187, -0.00002413, -0.00002500, -0.00002436, -0.00002221, -0.00001869, -0.00001400, -0.00000847, -0.00000247, 0.00000362, 0.00000939, 0.00001448, 0.00001860, 0.00002154, 0.00002317, 0.00002349, 0.00002255, 0.00002053, 0.00001764, 0.00001416, 0.00001035, 0.00000651, 0.00000287, -0.00000037, -0.00000304, -0.00000507, -0.00000643, -0.00000715, -0.00000729, -0.00000696, -0.00000628, -0.00000537, -0.00000436, -0.00000334, -0.00000241, -0.00000162, -0.00000100, -0.00000058, -0.00000036, -0.00000032, -0.00000044, -0.00000071, -0.00000111, -0.00000161, -0.00000220, -0.00000286, -0.00000356, -0.00000428, -0.00000498, -0.00000564, -0.00000620, -0.00000663, -0.00000689, -0.00000695, -0.00000678, -0.00000639, -0.00000579, -0.00000501, -0.00000408, -0.00000308, -0.00000205, -0.00000107, -0.00000019, 0.00000055, 0.00000112, 0.00000152, 0.00000178, 0.00000191, 0.00000199, 0.00000207, 0.00000220, 0.00000243, 0.00000280, },
		{ -0.00000282, -0.00000303, -0.00000332, -0.00000362, -0.00000382, -0.00000382, -0.00000355, -0.00000295, -0.00000200, -0.00000072, 0.00000083, 0.00000256, 0.00000435, 0.00000604, 0.00000750, 0.00000860, 0.00000922, 0.00000930, 0.00000884, 0.00000786, 0.00000647, 0.00000480, 0.00000302, 0.00000132, -0.00000012, -0.00000114, -0.00000161, -0.00000149, -0.00000077, 0.00000047, 0.00000210, 0.00000394, 0.00000576, 0.00000732, 0.00000841, 0.00000885, 0.00000850, 0.00000732, 0.00000534, 0.00000268, -0.00000047, -0.00000383, -0.00000712, -0.00001002, -0.00001223, -0.00001351, -0.00001369, -0.00001268, -0.00001049, -0.00000727, -0.00000322, 0.00000134, 0.00000605, 0.00001053, 0.00001440, 0.00001732, 0.00001904, 0.00001936, 0.00001824, 0.00001572, 0.00001197, 0.00000725, 0.00000188, -0.00000374, -0.00000923, -0.00001420, -0.00001832, -0.00002134, -0.00002308, -0.00002348, -0.00002256, -0.00002045, -0.00001735, -0.00001353, -0.00000927, -0.00000489, -0.00000067, 0.00000313, 0.00000633, 0.00000878, 0.00001043, 0.00001126, 0.00001134, 0.00001078, 0.00000971, 0.00000830, 0.00000670, 0.00000506, 0.00000351, 0.00000215, 0.00000105, 0.00000024, -0.00000026, -0.00000049, -0.00000048, -0.00000028, 0.00000005, 0.00000045, 0.00000085, 0.00000121, 0.00000148, 0.00000163, 0.00000162, 0.00000144, 0.00000107, 0.00000052, -0.00000020, -0.00000109, -0.00000210, -0.00000321, -0.00000435, -0.00000548, -0.00000654, -0.00000747, -0.00000821, -0.00000872, -0.00000896, -0.00000892, -0.00000861, -0.00000806, -0.00000733, -0.00000646, -0.00000556, -0.00000468, -0.00000392, -0.00000332, -0.00000294, -0.00000278, },
		{ -0.00000588, -0.00000519, -0.00000447, -0.00000383, -0.00000336, -0.00000315, -0.00000327, -0.00000372, -0.00000451, -0.00000556, -0.00000680, -0.00000810, -0.00000931, -0.00001029, -0.00001091, -0.00001106, -0.00001066, -0.00000969, -0.00000818, -0.00000621, -0.00000391, -0.00000146, 0.00000097, 0.00000317, 0.00000496, 0.00000621, 0.00000681, 0.00000675, 0.00000606, 0.00000484, 0.00000326, 0.00000152, -0.00000016, -0.00000153, -0.00000239, -0.00000258, -0.00000201, -0.00000066, 0.00000141, 0.00000406, 0.00000706, 0.00001014, 0.00001302, 0.00001538, 0.00001697, 0.00001756, 0.00001702, 0.00001531, 0.00001248, 0.00000869, 0.00000418, -0.00000072, -0.00000566, -0.00001027, -0.00001418, -0.00001710, -0.00001880, -0.00001913, -0.00001808, -0.00001571, -0.00001221, -0.00000784, -0.00000294, 0.00000213, 0.00000700, 0.00001133, 0.00001482, 0.00001725, 0.00001851, 0.00001854, 0.00001742, 0.00001528, 0.00001234, 0.00000885, 0.00000511, 0.00000139, -0.00000205, -0.00000499, -0.00000728, -0.00000882, -0.00000957, -0.00000959, -0.00000895, -0.00000780, -0.00000628, -0.00000458, -0.00000285, -0.00000124, 0.00000013, 0.00000117, 0.00000183, 0.00000211, 0.00000203, 0.00000164, 0.00000102, 0.00000024, -0.00000061, -0.00000144, -0.00000217, -0.00000275, -0.00000313, -0.00000329, -0.00000321, -0.00000291, -0.00000240, -0.00000173, -0.00000094, -0.00000010, 0.00000075, 0.00000154, 0.00000221, 0.00000270, 0.00000297, 0.00000299, 0.00000273, 0.00000221, 0.00000143, 0.00000043, -0.00000072, -0.00000196, -0.00000320, -0.00000437, -0.00000538, -0.00000617, -0.00000668, -0.00000689, -0.00000680, -0.00000644, },
		{ -0.00000148, -0.00000246, -0.00000335, -0.00000407, -0.00000455, -0.00000472, -0.00000459, -0.00000415, -0.00000347, -0.00000263, -0.00000174, -0.00000093, -0.00000032, -0.00000003, -0.00000015, -0.00000074, -0.00000182, -0.00000335, -0.00000524, -0.00000737, -0.00000957, -0.00001166, -0.00001344, -0.00001475, -0.00001544, -0.00001542, -0.00001466, -0.00001318, -0.00001109, -0.00000854, -0.00000575, -0.00000295, -0.00000038, 0.00000173, 0.00000318, 0.00000385, 0.00000366, 0.00000263, 0.00000088, -0.00000144, -0.00000407, -0.00000675, -0.00000916, -0.00001103, -0.00001211, -0.00001220, -0.00001121, -0.00000912, -0.00000602, -0.00000207, 0.00000247, 0.00000728, 0.00001202, 0.00001634, 0.00001990, 0.00002245, 0.00002377, 0.00002377, 0.00002243, 0.00001986, 0.00001625, 0.00001184, 0.00000696, 0.00000194, -0.00000286, -0.00000715, -0.00001065, -0.00001318, -0.00001463, -0.00001497, -0.00001426, -0.00001264, -0.00001030, -0.00000746, -0.00000438, -0.00000130, 0.00000157, 0.00000403, 0.00000597, 0.00000732, 0.00000804, 0.00000820, 0.00000787, 0.00000716, 0.00000623, 0.00000520, 0.00000420, 0.00000336, 0.00000276, 0.00000243, 0.00000241, 0.00000267, 0.00000316, 0.00000381, 0.00000454, 0.00000527, 0.00000589, 0.00000634, 0.00000655, 0.00000648, 0.00000612, 0.00000547, 0.00000455, 0.00000340, 0.00000210, 0.00000069, -0.00000073, -0.00000209, -0.00000332, -0.00000435, -0.00000513, -0.00000562, -0.00000580, -0.00000567, -0.00000525, -0.00000458, -0.00000371, -0.00000271, -0.00000167, -0.00000066, 0.00000024, 0.00000094, 0.00000139, 0.00000156, 0.00000143, 0.00000101, 0.00000034, -0.00000052, },
		{ -0.00000047, 0.00000048, 0.00000129, 0.00000191, 0.00000226, 0.00000234, 0.00000214, 0.00000169, 0.00000107, 0.00000035, -0.00000037, -0.00000100, -0.00000143, -0.00000160, -0.00000147, -0.00000102, -0.00000029, 0.00000065, 0.00000170, 0.00000272, 0.00000358, 0.00000412, 0.00000421, 0.00000375, 0.00000268, 0.00000099, -0.00000127, -0.00000399, -0.00000701, -0.00001014, -0.00001313, -0.00001576, -0.00001781, -0.00001910, -0.00001948, -0.00001890, -0.00001737, -0.00001497, -0.00001187, -0.00000830, -0.00000453, -0.00000086, 0.00000240, 0.00000499, 0.00000667, 0.00000729, 0.00000679, 0.00000517, 0.00000255, -0.00000087, -0.00000483, -0.00000902, -0.00001310, -0.00001675, -0.00001965, -0.00002157, -0.00002233, -0.00002185, -0.00002013, -0.00001727, -0.00001345, -0.00000892, -0.00000396, 0.00000110, 0.00000597, 0.00001035, 0.00001401, 0.00001677, 0.00001852, 0.00001923, 0.00001895, 0.00001778, 0.00001589, 0.00001347, 0.00001073, 0.00000788, 0.00000512, 0.00000261, 0.00000048, -0.00000120, -0.00000239, -0.00000310, -0.00000339, -0.00000333, -0.00000302, -0.00000255, -0.00000201, -0.00000150, -0.00000106, -0.00000074, -0.00000055, -0.00000047, -0.00000047, -0.00000050, -0.00000051, -0.00000044, -0.00000024, 0.00000013, 0.00000071, 0.00000148, 0.00000245, 0.00000358, 0.00000482, 0.00000611, 0.00000738, 0.00000855, 0.00000954, 0.00001030, 0.00001076, 0.00001089, 0.00001066, 0.00001008, 0.00000915, 0.00000793, 0.00000647, 0.00000485, 0.00000314, 0.00000146, -0.00000012, -0.00000151, -0.00000263, -0.00000343, -0.00000388, -0.00000397, -0.00000372, -0.00000317, -0.00000238, -0.00000146, },
		{ -0.00000049, -0.00000194, -0.00000309, -0.00000390, -0.00000433, -0.00000437, -0.00000406, -0.00000345, -0.00000262, -0.00000164, -0.00000063, 0.00000032, 0.00000115, 0.00000178, 0.00000219, 0.00000238, 0.00000238, 0.00000224, 0.00000205, 0.00000190, 0.00000188, 0.00000208, 0.00000256, 0.00000336, 0.00000449, 0.00000588, 0.00000747, 0.00000912, 0.00001069, 0.00001201, 0.00001290, 0.00001321, 0.00001282, 0.00001163, 0.00000963, 0.00000684, 0.00000336, -0.00000064, -0.00000496, -0.00000934, -0.00001351, -0.00001719, -0.00002014, -0.00002215, -0.00002306, -0.00002279, -0.00002134, -0.00001881, -0.00001534, -0.00001117, -0.00000659, -0.00000190, 0.00000258, 0.00000654, 0.00000970, 0.00001185, 0.00001286, 0.00001265, 0.00001126, 0.00000878, 0.00000539, 0.00000131, -0.00000318, -0.00000778, -0.00001222, -0.00001623, -0.00001957, -0.00002208, -0.00002364, -0.00002420, -0.00002379, -0.00002247, -0.00002038, -0.00001768, -0.00001454, -0.00001118, -0.00000777, -0.00000448, -0.00000146, 0.00000119, 0.00000339, 0.00000513, 0.00000639, 0.00000722, 0.00000767, 0.00000780, 0.00000769, 0.00000741, 0.00000703, 0.00000660, 0.00000617, 0.00000576, 0.00000537, 0.00000500, 0.00000464, 0.00000427, 0.00000388, 0.00000344, 0.00000295, 0.00000242, 0.00000185, 0.00000126, 0.00000070, 0.00000019, -0.00000022, -0.00000048, -0.00000055, -0.00000041, -0.00000003, 0.00000059, 0.00000145, 0.00000251, 0.00000373, 0.00000504, 0.00000637, 0.00000765, 0.00000878, 0.00000970, 0.00001032, 0.00001060, 0.00001050, 0.00001000, 0.00000912, 0.00000791, 0.00000641, 0.00000473, 0.00000295, 0.00000117, },
		{ 0.00001278, 0.00001292, 0.00001263, 0.00001192, 0.00001080, 0.00000932, 0.00000755, 0.00000559, 0.00000355, 0.00000154, -0.00000034, -0.00000199, -0.00000334, -0.00000433, -0.00000496, -0.00000523, -0.00000519, -0.00000488, -0.00000440, -0.00000384, -0.00000326, -0.00000276, -0.00000237, -0.00000215, -0.00000208, -0.00000214, -0.00000227, -0.00000240, -0.00000242, -0.00000226, -0.00000181, -0.00000100, 0.00000022, 0.00000186, 0.00000389, 0.00000625, 0.00000882, 0.00001147, 0.00001402, 0.00001630, 0.00001812, 0.00001932, 0.00001976, 0.00001934, 0.00001801, 0.00001579, 0.00001274, 0.00000899, 0.00000471, 0.00000013, -0.00000451, -0.00000897, -0.00001298, -0.00001633, -0.00001882, -0.00002030, -0.00002071, -0.00002001, -0.00001827, -0.00001560, -0.00001215, -0.00000815, -0.00000383, 0.00000054, 0.00000473, 0.00000848, 0.00001160, 0.00001391, 0.00001531, 0.00001574, 0.00001520, 0.00001375, 0.00001148, 0.00000855, 0.00000513, 0.00000141, -0.00000242, -0.00000616, -0.00000965, -0.00001273, -0.00001531, -0.00001729, -0.00001866, -0.00001940, -0.00001954, -0.00001916, -0.00001833, -0.00001715, -0.00001571, -0.00001411, -0.00001245, -0.00001080, -0.00000921, -0.00000774, -0.00000640, -0.00000520, -0.00000414, -0.00000319, -0.00000235, -0.00000159, -0.00000089, -0.00000024, 0.00000038, 0.00000095, 0.00000147, 0.00000192, 0.00000230, 0.00000257, 0.00000275, 0.00000281, 0.00000277, 0.00000264, 0.00000245, 0.00000225, 0.00000208, 0.00000198, 0.00000202, 0.00000223, 0.00000264, 0.00000328, 0.00000414, 0.00000520, 0.00000642, 0.00000773, 0.00000906, 0.00001032, 0.00001141, 0.00001226, },
		{ -0.00000090, 0.00000022, 0.00000157, 0.00000311, 0.00000480, 0.00000656, 0.00000830, 0.00000992, 0.00001133, 0.00001244, 0.00001318, 0.00001350, 0.00001337, 0.00001281, 0.00001185, 0.00001055, 0.00000901, 0.00000731, 0.00000557, 0.00000389, 0.00000236, 0.00000108, 0.00000008, -0.00000061, -0.00000098, -0.00000106, -0.00000090, -0.00000058, -0.00000018, 0.00000023, 0.00000056, 0.00000074, 0.00000073, 0.00000050, 0.00000006, -0.00000058, -0.00000137, -0.00000224, -0.00000310, -0.00000386, -0.00000444, -0.00000475, -0.00000472, -0.00000431, -0.00000349, -0.00000227, -0.00000068, 0.00000123, 0.00000335, 0.00000559, 0.00000782, 0.00000991, 0.00001174, 0.00001317, 0.00001412, 0.00001448, 0.00001423, 0.00001332, 0.00001179, 0.00000969, 0.00000711, 0.00000418, 0.00000103, -0.00000216, -0.00000523, -0.00000799, -0.00001030, -0.00001201, -0.00001302, -0.00001326, -0.00001271, -0.00001139, -0.00000937, -0.00000677, -0.00000371, -0.00000039, 0.00000301, 0.00000631, 0.00000931, 0.00001185, 0.00001380, 0.00001506, 0.00001559, 0.00001537, 0.00001444, 0.00001289, 0.00001082, 0.00000837, 0.00000569, 0.00000292, 0.00000022, -0.00000230, -0.00000454, -0.00000643, -0.00000794, -0.00000906, -0.00000982, -0.00001027, -0.00001046, -0.00001047, -0.00001036, -0.00001020, -0.00001004, -0.00000990, -0.00000981, -0.00000975, -0.00000972, -0.00000969, -0.00000963, -0.00000950, -0.00000929, -0.00000897, -0.00000855, -0.00000804, -0.00000746, -0.00000684, -0.00000621, -0.00000562, -0.00000508, -0.00000463, -0.00000426, -0.00000398, -0.00000374, -0.00000353, -0.00000328, -0.00000294, -0.00000246, -0.00000179, },
		{ -0.00000674, -0.00000680, -0.00000693, -0.00000710, -0.00000728, -0.00000742, -0.00000746, -0.00000734, -0.00000701, -0.00000644, -0.00000561, -0.00000451, -0.00000317, -0.00000166, -0.00000003, 0.00000163, 0.00000322, 0.00000464, 0.00000580, 0.00000664, 0.00000709, 0.00000714, 0.00000678, 0.00000604, 0.00000498, 0.00000368, 0.00000223, 0.00000073, -0.00000072, -0.00000201, -0.00000306, -0.00000379, -0.00000416, -0.00000413, -0.00000369, -0.00000287, -0.00000169, -0.00000022, 0.00000148, 0.00000332, 0.00000522, 0.00000707, 0.00000879, 0.00001028, 0.00001146, 0.00001225, 0.00001259, 0.00001245, 0.00001181, 0.00001068, 0.00000910, 0.00000712, 0.00000484, 0.00000237, -0.00000016, -0.00000260, -0.00000483, -0.00000669, -0.00000808, -0.00000890, -0.00000909, -0.00000865, -0.00000760, -0.00000601, -0.00000400, -0.00000171, 0.00000068, 0.00000299, 0.00000506, 0.00000672, 0.00000784, 0.00000834, 0.00000818, 0.00000739, 0.00000604, 0.00000423, 0.00000213, -0.00000008, -0.00000221, -0.00000408, -0.00000554, -0.00000645, -0.00000673, -0.00000637, -0.00000539, -0.00000386, -0.00000190, 0.00000032, 0.00000264, 0.00000489, 0.00000691, 0.00000856, 0.00000974, 0.00001041, 0.00001054, 0.00001018, 0.00000939, 0.00000828, 0.00000697, 0.00000558, 0.00000422, 0.00000300, 0.00000199, 0.00000122, 0.00000072, 0.00000045, 0.00000036, 0.00000040, 0.00000049, 0.00000055, 0.00000052, 0.00000035, 0.00000000, -0.00000052, -0.00000119, -0.00000199, -0.00000285, -0.00000373, -0.00000456, -0.00000530, -0.00000590, -0.00000635, -0.00000665, -0.00000680, -0.00000684, -0.00000682, -0.00000677, -0.00000673, },
		{ 0.00000045, -0.00000009, -0.00000051, -0.00000080, -0.00000099, -0.00000109, -0.00000115, -0.00000121, -0.00000131, -0.00000148, -0.00000174, -0.00000208, -0.00000247, -0.00000288, -0.00000324, -0.00000352, -0.00000363, -0.00000353, -0.00000319, -0.00000259, -0.00000173, -0.00000064, 0.00000063, 0.00000199, 0.00000338, 0.00000469, 0.00000582, 0.00000670, 0.00000725, 0.00000741, 0.00000715, 0.00000645, 0.00000531, 0.00000377, 0.00000188, -0.00000032, -0.00000273, -0.00000529, -0.00000789, -0.00001045, -0.00001285, -0.00001499, -0.00001677, -0.00001808, -0.00001882, -0.00001889, -0.00001823, -0.00001678, -0.00001452, -0.00001147, -0.00000769, -0.00000329, 0.00000158, 0.00000674, 0.00001193, 0.00001692, 0.00002144, 0.00002522, 0.00002804, 0.00002971, 0.00003009, 0.00002914, 0.00002687, 0.00002340, 0.00001890, 0.00001365, 0.00000794, 0.00000211, -0.00000350, -0.00000855, -0.00001278, -0.00001596, -0.00001794, -0.00001869, -0.00001824, -0.00001672, -0.00001431, -0.00001127, -0.00000788, -0.00000442, -0.00000116, 0.00000167, 0.00000391, 0.00000544, 0.00000623, 0.00000634, 0.00000586, 0.00000495, 0.00000378, 0.00000254, 0.00000141, 0.00000054, 0.00000003, -0.00000006, 0.00000027, 0.00000096, 0.00000192, 0.00000302, 0.00000413, 0.00000510, 0.00000582, 0.00000620, 0.00000617, 0.00000574, 0.00000495, 0.00000386, 0.00000258, 0.00000124, -0.00000004, -0.00000114, -0.00000195, -0.00000243, -0.00000254, -0.00000229, -0.00000172, -0.00000091, 0.00000004, 0.00000103, 0.00000196, 0.00000273, 0.00000330, 0.00000360, 0.00000363, 0.00000342, 0.00000299, 0.00000241, 0.00000175, 0.00000107, },
		{ -0.00000109, -0.00000044, -0.00000002, 0.00000016, 0.00000010, -0.00000014, -0.00000051, -0.00000094, -0.00000136, -0.00000171, -0.00000197, -0.00000211, -0.00000213, -0.00000206, -0.00000194, -0.00000181, -0.00000171, -0.00000169, -0.00000176, -0.00000192, -0.00000218, -0.00000248, -0.00000279, -0.00000304, -0.00000317, -0.00000313, -0.00000285, -0.00000232, -0.00000152, -0.00000045, 0.00000087, 0.00000237, 0.00000402, 0.00000575, 0.00000750, 0.00000919, 0.00001077, 0.00001219, 0.00001339, 0.00001434, 0.00001499, 0.00001530, 0.00001522, 0.00001470, 0.00001369, 0.00001215, 0.00001002, 0.00000729, 0.00000395, 0.00000004, -0.00000436, -0.00000912, -0.00001407, -0.00001899, -0.00002363, -0.00002769, -0.00003091, -0.00003301, -0.00003376, -0.00003300, -0.00003065, -0.00002672, -0.00002132, -0.00001469, -0.00000713, 0.00000096, 0.00000915, 0.00001696, 0.00002393, 0.00002967, 0.00003384, 0.00003621, 0.00003667, 0.00003524, 0.00003206, 0.00002740, 0.00002161, 0.00001510, 0.00000832, 0.00000171, -0.00000432, -0.00000944, -0.00001342, -0.00001610, -0.00001747, -0.00001759, -0.00001663, -0.00001481, -0.00001241, -0.00000970, -0.00000696, -0.00000441, -0.00000222, -0.00000050, 0.00000073, 0.00000150, 0.00000189, 0.00000204, 0.00000209, 0.00000218, 0.00000243, 0.00000293, 0.00000370, 0.00000473, 0.00000595, 0.00000727, 0.00000855, 0.00000965, 0.00001044, 0.00001080, 0.00001065, 0.00000998, 0.00000880, 0.00000717, 0.00000522, 0.00000307, 0.00000089, -0.00000117, -0.00000296, -0.00000438, -0.00000536, -0.00000585, -0.00000588, -0.00000551, -0.00000482, -0.00000391, -0.00000292, -0.00000195, },
		{ -0.00000122, -0.00000236, -0.00000303, -0.00000325, -0.00000308, -0.00000263, -0.00000201, -0.00000133, -0.00000072, -0.00000025, 0.00000001, 0.00000004, -0.00000017, -0.00000057, -0.00000110, -0.00000171, -0.00000231, -0.00000286, -0.00000330, -0.00000361, -0.00000378, -0.00000383, -0.00000381, -0.00000374, -0.00000370, -0.00000370, -0.00000380, -0.00000399, -0.00000429, -0.00000466, -0.00000508, -0.00000549, -0.00000587, -0.00000615, -0.00000632, -0.00000636, -0.00000627, -0.00000606, -0.00000576, -0.00000539, -0.00000497, -0.00000453, -0.00000404, -0.00000349, -0.00000283, -0.00000200, -0.00000092, 0.00000047, 0.00000224, 0.00000440, 0.00000695, 0.00000983, 0.00001291, 0.00001602, 0.00001896, 0.00002146, 0.00002327, 0.00002414, 0.00002386, 0.00002228, 0.00001935, 0.00001508, 0.00000964, 0.00000327, -0.00000370, -0.00001083, -0.00001768, -0.00002378, -0.00002870, -0.00003205, -0.00003359, -0.00003315, -0.00003074, -0.00002650, -0.00002070, -0.00001373, -0.00000607, 0.00000176, 0.00000924, 0.00001587, 0.00002125, 0.00002507, 0.00002716, 0.00002746, 0.00002608, 0.00002324, 0.00001925, 0.00001448, 0.00000935, 0.00000424, -0.00000047, -0.00000450, -0.00000766, -0.00000985, -0.00001106, -0.00001138, -0.00001097, -0.00001002, -0.00000876, -0.00000740, -0.00000611, -0.00000502, -0.00000421, -0.00000367, -0.00000336, -0.00000318, -0.00000301, -0.00000270, -0.00000215, -0.00000126, 0.00000002, 0.00000167, 0.00000364, 0.00000582, 0.00000805, 0.00001017, 0.00001201, 0.00001340, 0.00001423, 0.00001442, 0.00001394, 0.00001284, 0.00001121, 0.00000917, 0.00000689, 0.00000456, 0.00000233, 0.00000036, },
		{ 0.00001323, 0.00001289, 0.00001191, 0.00001042, 0.00000857, 0.00000654, 0.00000451, 0.00000266, 0.00000113, 0.00000002, -0.00000062, -0.00000079, -0.00000056, -0.00000001, 0.00000075, 0.00000157, 0.00000234, 0.00000296, 0.00000333, 0.00000343, 0.00000325, 0.00000281, 0.00000217, 0.00000139, 0.00000057, -0.00000024, -0.00000098, -0.00000160, -0.00000210, -0.00000248, -0.00000275, -0.00000296, -0.00000313, -0.00000328, -0.00000342, -0.00000355, -0.00000364, -0.00000365, -0.00000353, -0.00000324, -0.00000274, -0.00000203, -0.00000114, -0.00000013, 0.00000092, 0.00000185, 0.00000254, 0.00000282, 0.00000257, 0.00000167, 0.00000007, -0.00000220, -0.00000507, -0.00000836, -0.00001185, -0.00001526, -0.00001827, -0.00002057, -0.00002189, -0.00002199, -0.00002075, -0.00001813, -0.00001423, -0.00000925, -0.00000351, 0.00000261, 0.00000865, 0.00001417, 0.00001873, 0.00002196, 0.00002362, 0.00002356, 0.00002180, 0.00001849, 0.00001391, 0.00000846, 0.00000260, -0.00000317, -0.00000838, -0.00001258, -0.00001545, -0.00001678, -0.00001650, -0.00001467, -0.00001151, -0.00000732, -0.00000251, 0.00000249, 0.00000726, 0.00001139, 0.00001457, 0.00001658, 0.00001733, 0.00001684, 0.00001523, 0.00001272, 0.00000959, 0.00000616, 0.00000275, -0.00000039, -0.00000301, -0.00000499, -0.00000627, -0.00000688, -0.00000690, -0.00000650, -0.00000585, -0.00000514, -0.00000452, -0.00000413, -0.00000404, -0.00000424, -0.00000468, -0.00000526, -0.00000583, -0.00000625, -0.00000635, -0.00000601, -0.00000516, -0.00000376, -0.00000187, 0.00000042, 0.00000296, 0.00000557, 0.00000805, 0.00001020, 0.00001185, 0.00001288, },
		{ -0.00000437, -0.00000240, -0.00000004, 0.00000253, 0.00000508, 0.00000739, 0.00000926, 0.00001051, 0.00001106, 0.00001086, 0.00000996, 0.00000845, 0.00000650, 0.00000430, 0.00000206, -0.00000001, -0.00000174, -0.00000299, -0.00000367, -0.00000379, -0.00000336, -0.00000249, -0.00000130, 0.00000007, 0.00000147, 0.00000277, 0.00000387, 0.00000469, 0.00000521, 0.00000543, 0.00000538, 0.00000511, 0.00000469, 0.00000415, 0.00000354, 0.00000287, 0.00000212, 0.00000127, 0.00000028, -0.00000089, -0.00000225, -0.00000380, -0.00000550, -0.00000726, -0.00000894, -0.00001038, -0.00001139, -0.00001177, -0.00001137, -0.00001005, -0.00000776, -0.00000451, -0.00000044, 0.00000425, 0.00000926, 0.00001422, 0.00001873, 0.00002239, 0.00002483, 0.00002574, 0.00002493, 0.00002234, 0.00001804, 0.00001225, 0.00000531, -0.00000234, -0.00001016, -0.00001763, -0.00002422, -0.00002949, -0.00003309, -0.00003481, -0.00003462, -0.00003261, -0.00002904, -0.00002427, -0.00001877, -0.00001301, -0.00000748, -0.00000261, 0.00000125, 0.00000390, 0.00000525, 0.00000536, 0.00000441, 0.00000267, 0.00000051, -0.00000172, -0.00000365, -0.00000497, -0.00000547, -0.00000502, -0.00000364, -0.00000142, 0.00000142, 0.00000461, 0.00000784, 0.00001080, 0.00001320, 0.00001482, 0.00001553, 0.00001529, 0.00001414, 0.00001224, 0.00000979, 0.00000705, 0.00000429, 0.00000173, -0.00000040, -0.00000197, -0.00000293, -0.00000328, -0.00000312, -0.00000259, -0.00000188, -0.00000116, -0.00000063, -0.00000040, -0.00000056, -0.00000113, -0.00000205, -0.00000321, -0.00000444, -0.00000557, -0.00000641, -0.00000680, -0.00000661, -0.00000580, },
		{ 0.00000037, -0.00000104, -0.00000262, -0.00000413, -0.00000536, -0.00000609, -0.00000618, -0.00000552, -0.00000409, -0.00000197, 0.00000069, 0.00000369, 0.00000676, 0.00000964, 0.00001205, 0.00001378, 0.00001467, 0.00001463, 0.00001367, 0.00001188, 0.00000940, 0.00000647, 0.00000332, 0.00000020, -0.00000265, -0.00000505, -0.00000686, -0.00000801, -0.00000851, -0.00000842, -0.00000783, -0.00000688, -0.00000569, -0.00000440, -0.00000310, -0.00000184, -0.00000065, 0.00000050, 0.00000164, 0.00000283, 0.00000411, 0.00000550, 0.00000698, 0.00000847, 0.00000985, 0.00001095, 0.00001159, 0.00001156, 0.00001068, 0.00000883, 0.00000595, 0.00000207, -0.00000266, -0.00000798, -0.00001357, -0.00001899, -0.00002380, -0.00002754, -0.00002979, -0.00003024, -0.00002869, -0.00002508, -0.00001954, -0.00001234, -0.00000391, 0.00000521, 0.00001439, 0.00002300, 0.00003042, 0.00003613, 0.00003973, 0.00004098, 0.00003983, 0.00003640, 0.00003100, 0.00002406, 0.00001611, 0.00000776, -0.00000043, -0.00000793, -0.00001430, -0.00001925, -0.00002262, -0.00002441, -0.00002476, -0.00002393, -0.00002222, -0.00002002, -0.00001767, -0.00001548, -0.00001369, -0.00001244, -0.00001176, -0.00001159, -0.00001178, -0.00001212, -0.00001238, -0.00001233, -0.00001180, -0.00001067, -0.00000889, -0.00000651, -0.00000367, -0.00000057, 0.00000256, 0.00000545, 0.00000787, 0.00000962, 0.00001056, 0.00001065, 0.00000993, 0.00000851, 0.00000659, 0.00000438, 0.00000215, 0.00000013, -0.00000148, -0.00000255, -0.00000301, -0.00000288, -0.00000226, -0.00000129, -0.00000017, 0.00000090, 0.00000172, 0.00000213, 0.00000204, 0.00000144, },
		{ -0.00000453, -0.00000321, -0.00000178, -0.00000049, 0.00000042, 0.00000078, 0.00000045, -0.00000060, -0.00000230, -0.00000450, -0.00000697, -0.00000944, -0.00001160, -0.00001317, -0.00001389, -0.00001361, -0.00001225, -0.00000983, -0.00000650, -0.00000248, 0.00000192, 0.00000639, 0.00001056, 0.00001412, 0.00001683, 0.00001850, 0.00001907, 0.00001855, 0.00001707, 0.00001481, 0.00001203, 0.00000897, 0.00000590, 0.00000301, 0.00000048, -0.00000164, -0.00000331, -0.00000461, -0.00000563, -0.00000650, -0.00000734, -0.00000826, -0.00000929, -0.00001042, -0.00001155, -0.00001254, -0.00001318, -0.00001325, -0.00001254, -0.00001088, -0.00000818, -0.00000445, 0.00000017, 0.00000546, 0.00001104, 0.00001649, 0.00002132, 0.00002505, 0.00002723, 0.00002754, 0.00002575, 0.00002184, 0.00001594, 0.00000838, -0.00000037, -0.00000969, -0.00001892, -0.00002736, -0.00003435, -0.00003934, -0.00004193, -0.00004191, -0.00003928, -0.00003423, -0.00002715, -0.00001858, -0.00000916, 0.00000042, 0.00000950, 0.00001746, 0.00002383, 0.00002828, 0.00003067, 0.00003101, 0.00002950, 0.00002645, 0.00002227, 0.00001743, 0.00001237, 0.00000750, 0.00000315, -0.00000045, -0.00000321, -0.00000512, -0.00000628, -0.00000685, -0.00000705, -0.00000708, -0.00000712, -0.00000731, -0.00000772, -0.00000833, -0.00000908, -0.00000982, -0.00001041, -0.00001069, -0.00001050, -0.00000976, -0.00000844, -0.00000658, -0.00000429, -0.00000174, 0.00000085, 0.00000324, 0.00000521, 0.00000657, 0.00000720, 0.00000705, 0.00000616, 0.00000463, 0.00000266, 0.00000046, -0.00000170, -0.00000359, -0.00000501, -0.00000583, -0.00000599, -0.00000552, },
		{ 0.00000781, 0.00000550, 0.00000332, 0.00000152, 0.00000034, -0.00000007, 0.00000033, 0.00000150, 0.00000329, 0.00000545, 0.00000768, 0.00000964, 0.00001101, 0.00001152, 0.00001097, 0.00000926, 0.00000644, 0.00000265, -0.00000184, -0.00000667, -0.00001145, -0.00001578, -0.00001926, -0.00002160, -0.00002261, -0.00002220, -0.00002044, -0.00001750, -0.00001368, -0.00000932, -0.00000480, -0.00000052, 0.00000322, 0.00000617, 0.00000820, 0.00000928, 0.00000951, 0.00000908, 0.00000823, 0.00000723, 0.00000634, 0.00000578, 0.00000567, 0.00000604, 0.00000680, 0.00000777, 0.00000870, 0.00000926, 0.00000917, 0.00000815, 0.00000602, 0.00000271, -0.00000168, -0.00000695, -0.00001273, -0.00001857, -0.00002393, -0.00002827, -0.00003109, -0.00003198, -0.00003071, -0.00002718, -0.00002154, -0.00001410, -0.00000536, 0.00000402, 0.00001334, 0.00002186, 0.00002887, 0.00003381, 0.00003627, 0.00003605, 0.00003315, 0.00002782, 0.00002051, 0.00001181, 0.00000243, -0.00000691, -0.00001549, -0.00002269, -0.00002804, -0.00003124, -0.00003219, -0.00003099, -0.00002791, -0.00002337, -0.00001786, -0.00001194, -0.00000612, -0.00000088, 0.00000341, 0.00000653, 0.00000836, 0.00000896, 0.00000846, 0.00000710, 0.00000517, 0.00000296, 0.00000075, -0.00000125, -0.00000287, -0.00000405, -0.00000479, -0.00000515, -0.00000526, -0.00000523, -0.00000519, -0.00000523, -0.00000540, -0.00000569, -0.00000603, -0.00000633, -0.00000645, -0.00000626, -0.00000564, -0.00000451, -0.00000286, -0.00000073, 0.00000176, 0.00000444, 0.00000709, 0.00000948, 0.00001140, 0.00001266, 0.00001313, 0.00001279, 0.00001169, 0.00000996, },
		{ 0.00000091, 0.00000395, 0.00000641, 0.00000808, 0.00000884, 0.00000864, 0.00000755, 0.00000575, 0.00000351, 0.00000115, -0.00000096, -0.00000246, -0.00000304, -0.00000249, -0.00000070, 0.00000231, 0.00000635, 0.00001113, 0.00001627, 0.00002129, 0.00002575, 0.00002919, 0.00003127, 0.00003176, 0.00003056, 0.00002777, 0.00002359, 0.00001839, 0.00001261, 0.00000675, 0.00000128, -0.00000335, -0.00000683, -0.00000898, -0.00000976, -0.00000930, -0.00000788, -0.00000584, -0.00000362, -0.00000164, -0.00000029, 0.00000015, -0.00000045, -0.00000211, -0.00000463, -0.00000773, -0.00001098, -0.00001391, -0.00001605, -0.00001697, -0.00001638, -0.00001412, -0.00001024, -0.00000496, 0.00000131, 0.00000801, 0.00001451, 0.00002013, 0.00002424, 0.00002633, 0.00002603, 0.00002321, 0.00001796, 0.00001059, 0.00000165, -0.00000818, -0.00001811, -0.00002730, -0.00003501, -0.00004057, -0.00004353, -0.00004365, -0.00004093, -0.00003564, -0.00002824, -0.00001937, -0.00000978, -0.00000025, 0.00000847, 0.00001574, 0.00002107, 0.00002418, 0.00002499, 0.00002365, 0.00002046, 0.00001590, 0.00001051, 0.00000487, -0.00000046, -0.00000501, -0.00000844, -0.00001054, -0.00001126, -0.00001072, -0.00000914, -0.00000682, -0.00000413, -0.00000144, 0.00000094, 0.00000273, 0.00000377, 0.00000400, 0.00000343, 0.00000218, 0.00000042, -0.00000166, -0.00000385, -0.00000598, -0.00000791, -0.00000955, -0.00001090, -0.00001197, -0.00001283, -0.00001355, -0.00001422, -0.00001489, -0.00001556, -0.00001622, -0.00001679, -0.00001715, -0.00001718, -0.00001675, -0.00001575, -0.00001413, -0.00001187, -0.00000907, -0.00000585, -0.00000244, },
		{ -0.00002095, -0.00002162, -0.00002152, -0.00002055, -0.00001873, -0.00001613, -0.00001293, -0.00000941, -0.00000588, -0.00000270, -0.00000022, 0.00000127, 0.00000155, 0.00000054, -0.00000173, -0.00000507, -0.00000916, -0.00001359, -0.00001787, -0.00002149, -0.00002398, -0.00002497, -0.00002421, -0.00002161, -0.00001726, -0.00001143, -0.00000453, 0.00000290, 0.00001029, 0.00001704, 0.00002263, 0.00002667, 0.00002892, 0.00002933, 0.00002804, 0.00002540, 0.00002186, 0.00001799, 0.00001438, 0.00001157, 0.00000999, 0.00000993, 0.00001146, 0.00001444, 0.00001854, 0.00002325, 0.00002794, 0.00003194, 0.00003459, 0.00003534, 0.00003382, 0.00002985, 0.00002354, 0.00001521, 0.00000544, -0.00000504, -0.00001537, -0.00002467, -0.00003212, -0.00003704, -0.00003896, -0.00003768, -0.00003329, -0.00002614, -0.00001687, -0.00000630, 0.00000465, 0.00001501, 0.00002385, 0.00003042, 0.00003416, 0.00003477, 0.00003224, 0.00002684, 0.00001909, 0.00000968, -0.00000055, -0.00001074, -0.00002006, -0.00002781, -0.00003346, -0.00003671, -0.00003748, -0.00003592, -0.00003240, -0.00002741, -0.00002155, -0.00001545, -0.00000969, -0.00000475, -0.00000099, 0.00000140, 0.00000239, 0.00000214, 0.00000089, -0.00000101, -0.00000316, -0.00000519, -0.00000676, -0.00000764, -0.00000768, -0.00000687, -0.00000528, -0.00000309, -0.00000053, 0.00000212, 0.00000462, 0.00000673, 0.00000828, 0.00000917, 0.00000939, 0.00000898, 0.00000803, 0.00000667, 0.00000504, 0.00000327, 0.00000146, -0.00000033, -0.00000208, -0.00000382, -0.00000558, -0.00000742, -0.00000937, -0.00001145, -0.00001362, -0.00001580, -0.00001786, -0.00001963, },
		{ 0.00000181, 0.00000041, -0.00000143, -0.00000374, -0.00000644, -0.00000942, -0.00001247, -0.00001536, -0.00001779, -0.00001950, -0.00002024, -0.00001984, -0.00001822, -0.00001543, -0.00001164, -0.00000717, -0.00000239, 0.00000221, 0.00000616, 0.00000899, 0.00001034, 0.00000995, 0.00000776, 0.00000385, -0.00000149, -0.00000782, -0.00001458, -0.00002114, -0.00002686, -0.00003117, -0.00003364, -0.00003401, -0.00003221, -0.00002843, -0.00002306, -0.00001664, -0.00000987, -0.00000346, 0.00000189, 0.00000560, 0.00000726, 0.00000671, 0.00000403, -0.00000043, -0.00000610, -0.00001221, -0.00001792, -0.00002236, -0.00002475, -0.00002447, -0.00002116, -0.00001474, -0.00000546, 0.00000614, 0.00001924, 0.00003285, 0.00004585, 0.00005714, 0.00006572, 0.00007080, 0.00007187, 0.00006876, 0.00006164, 0.00005106, 0.00003782, 0.00002299, 0.00000772, -0.00000678, -0.00001941, -0.00002927, -0.00003569, -0.00003835, -0.00003724, -0.00003271, -0.00002537, -0.00001605, -0.00000573, 0.00000460, 0.00001400, 0.00002165, 0.00002699, 0.00002966, 0.00002960, 0.00002702, 0.00002231, 0.00001607, 0.00000897, 0.00000170, -0.00000508, -0.00001084, -0.00001520, -0.00001796, -0.00001912, -0.00001883, -0.00001739, -0.00001519, -0.00001266, -0.00001020, -0.00000817, -0.00000682, -0.00000627, -0.00000655, -0.00000755, -0.00000906, -0.00001083, -0.00001259, -0.00001406, -0.00001501, -0.00001531, -0.00001489, -0.00001376, -0.00001202, -0.00000984, -0.00000740, -0.00000490, -0.00000253, -0.00000043, 0.00000130, 0.00000261, 0.00000353, 0.00000409, 0.00000437, 0.00000445, 0.00000439, 0.00000422, 0.00000393, 0.00000348, 0.00000281, },
		{ -0.00000210, -0.00000134, -0.00000047, 0.00000054, 0.00000165, 0.00000280, 0.00000386, 0.00000464, 0.00000495, 0.00000462, 0.00000349, 0.00000148, -0.00000140, -0.00000504, -0.00000922, -0.00001361, -0.00001784, -0.00002149, -0.00002416, -0.00002550, -0.00002527, -0.00002339, -0.00001992, -0.00001511, -0.00000935, -0.00000318, 0.00000280, 0.00000798, 0.00001179, 0.00001378, 0.00001368, 0.00001143, 0.00000720, 0.00000138, -0.00000543, -0.00001252, -0.00001908, -0.00002434, -0.00002764, -0.00002849, -0.00002665, -0.00002217, -0.00001540, -0.00000695, 0.00000233, 0.00001144, 0.00001935, 0.00002507, 0.00002779, 0.00002693, 0.00002226, 0.00001390, 0.00000235, -0.00001156, -0.00002672, -0.00004185, -0.00005561, -0.00006672, -0.00007409, -0.00007691, -0.00007475, -0.00006759, -0.00005585, -0.00004031, -0.00002210, -0.00000254, 0.00001691, 0.00003485, 0.00005000, 0.00006136, 0.00006827, 0.00007044, 0.00006800, 0.00006145, 0.00005162, 0.00003956, 0.00002643, 0.00001343, 0.00000165, -0.00000801, -0.00001491, -0.00001873, -0.00001944, -0.00001732, -0.00001290, -0.00000687, -0.00000003, 0.00000682, 0.00001295, 0.00001778, 0.00002091, 0.00002216, 0.00002157, 0.00001935, 0.00001587, 0.00001162, 0.00000709, 0.00000275, -0.00000099, -0.00000384, -0.00000565, -0.00000642, -0.00000626, -0.00000540, -0.00000412, -0.00000272, -0.00000150, -0.00000068, -0.00000043, -0.00000080, -0.00000176, -0.00000320, -0.00000495, -0.00000679, -0.00000851, -0.00000994, -0.00001093, -0.00001141, -0.00001137, -0.00001087, -0.00001000, -0.00000889, -0.00000767, -0.00000646, -0.00000536, -0.00000439, -0.00000356, -0.00000282, },
		{ -0.00000565, -0.00000529, -0.00000502, -0.00000490, -0.00000496, -0.00000517, -0.00000543, -0.00000563, -0.00000559, -0.00000516, -0.00000421, -0.00000266, -0.00000053, 0.00000210, 0.00000502, 0.00000797, 0.00001063, 0.00001265, 0.00001372, 0.00001358, 0.00001210, 0.00000926, 0.00000521, 0.00000024, -0.00000521, -0.00001065, -0.00001551, -0.00001927, -0.00002147, -0.00002182, -0.00002020, -0.00001672, -0.00001171, -0.00000570, 0.00000063, 0.00000652, 0.00001119, 0.00001394, 0.00001423, 0.00001177, 0.00000654, -0.00000115, -0.00001071, -0.00002130, -0.00003189, -0.00004135, -0.00004859, -0.00005265, -0.00005281, -0.00004870, -0.00004032, -0.00002807, -0.00001276, 0.00000447, 0.00002225, 0.00003908, 0.00005347, 0.00006410, 0.00006991, 0.00007022, 0.00006481, 0.00005393, 0.00003831, 0.00001906, -0.00000238, -0.00002439, -0.00004533, -0.00006363, -0.00007798, -0.00008740, -0.00009134, -0.00008972, -0.00008291, -0.00007167, -0.00005711, -0.00004055, -0.00002338, -0.00000697, 0.00000749, 0.00001904, 0.00002709, 0.00003138, 0.00003203, 0.00002948, 0.00002442, 0.00001773, 0.00001033, 0.00000316, -0.00000299, -0.00000752, -0.00001004, -0.00001043, -0.00000882, -0.00000552, -0.00000105, 0.00000403, 0.00000913, 0.00001368, 0.00001724, 0.00001952, 0.00002038, 0.00001985, 0.00001812, 0.00001548, 0.00001229, 0.00000894, 0.00000576, 0.00000307, 0.00000103, -0.00000026, -0.00000082, -0.00000078, -0.00000032, 0.00000035, 0.00000101, 0.00000147, 0.00000158, 0.00000130, 0.00000061, -0.00000041, -0.00000163, -0.00000292, -0.00000412, -0.00000510, -0.00000579, -0.00000614, -0.00000619, -0.00000599, },
		{ -0.00000013, -0.00000120, -0.00000204, -0.00000258, -0.00000277, -0.00000264, -0.00000228, -0.00000183, -0.00000148, -0.00000141, -0.00000176, -0.00000265, -0.00000409, -0.00000601, -0.00000826, -0.00001059, -0.00001271, -0.00001432, -0.00001511, -0.00001488, -0.00001348, -0.00001093, -0.00000738, -0.00000312, 0.00000147, 0.00000591, 0.00000970, 0.00001239, 0.00001361, 0.00001316, 0.00001102, 0.00000736, 0.00000257, -0.00000279, -0.00000803, -0.00001245, -0.00001533, -0.00001610, -0.00001438, -0.00001004, -0.00000326, 0.00000552, 0.00001554, 0.00002585, 0.00003537, 0.00004297, 0.00004762, 0.00004852, 0.00004512, 0.00003727, 0.00002523, 0.00000969, -0.00000832, -0.00002746, -0.00004619, -0.00006296, -0.00007627, -0.00008490, -0.00008795, -0.00008499, -0.00007607, -0.00006177, -0.00004310, -0.00002146, 0.00000147, 0.00002394, 0.00004419, 0.00006071, 0.00007226, 0.00007806, 0.00007782, 0.00007173, 0.00006047, 0.00004512, 0.00002706, 0.00000780, -0.00001110, -0.00002822, -0.00004236, -0.00005268, -0.00005872, -0.00006044, -0.00005817, -0.00005259, -0.00004463, -0.00003532, -0.00002576, -0.00001693, -0.00000965, -0.00000450, -0.00000177, -0.00000145, -0.00000327, -0.00000675, -0.00001126, -0.00001609, -0.00002058, -0.00002411, -0.00002625, -0.00002674, -0.00002552, -0.00002272, -0.00001865, -0.00001371, -0.00000839, -0.00000315, 0.00000158, 0.00000547, 0.00000830, 0.00001000, 0.00001064, 0.00001036, 0.00000941, 0.00000807, 0.00000661, 0.00000527, 0.00000421, 0.00000354, 0.00000325, 0.00000329, 0.00000354, 0.00000383, 0.00000403, 0.00000401, 0.00000368, 0.00000304, 0.00000212, 0.00000102, },
		{ -0.00000245, -0.00000132, -0.00000051, -0.00000011, -0.00000016, -0.00000061, -0.00000137, -0.00000226, -0.00000308, -0.00000362, -0.00000371, -0.00000322, -0.00000212, -0.00000045, 0.00000161, 0.00000383, 0.00000591, 0.00000752, 0.00000837, 0.00000822, 0.00000694, 0.00000454, 0.00000115, -0.00000294, -0.00000734, -0.00001159, -0.00001521, -0.00001776, -0.00001891, -0.00001845, -0.00001640, -0.00001296, -0.00000852, -0.00000362, 0.00000106, 0.00000485, 0.00000711, 0.00000732, 0.00000521, 0.00000071, -0.00000592, -0.00001414, -0.00002318, -0.00003206, -0.00003972, -0.00004511, -0.00004731, -0.00004563, -0.00003970, -0.00002955, -0.00001561, 0.00000128, 0.00001995, 0.00003898, 0.00005682, 0.00007194, 0.00008296, 0.00008880, 0.00008876, 0.00008264, 0.00007073, 0.00005382, 0.00003314, 0.00001023, -0.00001317, -0.00003526, -0.00005439, -0.00006915, -0.00007853, -0.00008196, -0.00007943, -0.00007138, -0.00005873, -0.00004275, -0.00002495, -0.00000690, 0.00000988, 0.00002405, 0.00003460, 0.00004090, 0.00004276, 0.00004038, 0.00003436, 0.00002560, 0.00001517, 0.00000424, -0.00000609, -0.00001486, -0.00002137, -0.00002520, -0.00002625, -0.00002473, -0.00002112, -0.00001608, -0.00001040, -0.00000486, -0.00000019, 0.00000305, 0.00000446, 0.00000391, 0.00000147, -0.00000256, -0.00000774, -0.00001352, -0.00001931, -0.00002456, -0.00002882, -0.00003176, -0.00003322, -0.00003320, -0.00003185, -0.00002944, -0.00002631, -0.00002284, -0.00001939, -0.00001624, -0.00001361, -0.00001160, -0.00001021, -0.00000936, -0.00000887, -0.00000858, -0.00000829, -0.00000786, -0.00000718, -0.00000623, -0.00000505, -0.00000374, },
		{ -0.00001782, -0.00001847, -0.00001877, -0.00001859, -0.00001788, -0.00001669, -0.00001513, -0.00001339, -0.00001171, -0.00001032, -0.00000943, -0.00000918, -0.00000964, -0.00001073, -0.00001231, -0.00001411, -0.00001581, -0.00001708, -0.00001760, -0.00001713, -0.00001552, -0.00001279, -0.00000908, -0.00000470, -0.00000004, 0.00000439, 0.00000811, 0.00001067, 0.00001173, 0.00001110, 0.00000879, 0.00000504, 0.00000024, -0.00000500, -0.00001003, -0.00001414, -0.00001669, -0.00001717, -0.00001530, -0.00001106, -0.00000472, 0.00000317, 0.00001181, 0.00002023, 0.00002738, 0.00003225, 0.00003395, 0.00003187, 0.00002571, 0.00001558, 0.00000199, -0.00001419, -0.00003173, -0.00004923, -0.00006515, -0.00007804, -0.00008661, -0.00008990, -0.00008736, -0.00007892, -0.00006503, -0.00004659, -0.00002491, -0.00000156, 0.00002171, 0.00004317, 0.00006126, 0.00007472, 0.00008268, 0.00008476, 0.00008109, 0.00007226, 0.00005928, 0.00004346, 0.00002628, 0.00000926, -0.00000621, -0.00001898, -0.00002820, -0.00003347, -0.00003476, -0.00003246, -0.00002727, -0.00002015, -0.00001215, -0.00000436, 0.00000227, 0.00000696, 0.00000925, 0.00000894, 0.00000619, 0.00000142, -0.00000474, -0.00001150, -0.00001805, -0.00002361, -0.00002757, -0.00002949, -0.00002917, -0.00002668, -0.00002232, -0.00001656, -0.00001002, -0.00000338, 0.00000272, 0.00000771, 0.00001118, 0.00001289, 0.00001279, 0.00001104, 0.00000793, 0.00000387, -0.00000066, -0.00000521, -0.00000936, -0.00001280, -0.00001535, -0.00001693, -0.00001764, -0.00001762, -0.00001713, -0.00001643, -0.00001576, -0.00001533, -0.00001525, -0.00001555, -0.00001617, -0.00001699, },
		{ -0.00000227, -0.00000132, -0.00000088, -0.00000110, -0.00000203, -0.00000364, -0.00000574, -0.00000811, -0.00001045, -0.00001245, -0.00001385, -0.00001447, -0.00001421, -0.00001313, -0.00001140, -0.00000930, -0.00000716, -0.00000538, -0.00000429, -0.00000417, -0.00000518, -0.00000732, -0.00001042, -0.00001418, -0.00001816, -0.00002186, -0.00002475, -0.00002637, -0.00002638, -0.00002461, -0.00002110, -0.00001610, -0.00001010, -0.00000372, 0.00000231, 0.00000724, 0.00001038, 0.00001121, 0.00000945, 0.00000508, -0.00000159, -0.00000997, -0.00001922, -0.00002831, -0.00003614, -0.00004165, -0.00004395, -0.00004238, -0.00003667, -0.00002693, -0.00001367, 0.00000220, 0.00001943, 0.00003661, 0.00005220, 0.00006475, 0.00007301, 0.00007605, 0.00007336, 0.00006494, 0.00005124, 0.00003322, 0.00001220, -0.00001025, -0.00003242, -0.00005263, -0.00006937, -0.00008146, -0.00008813, -0.00008906, -0.00008444, -0.00007490, -0.00006147, -0.00004544, -0.00002823, -0.00001129, 0.00000408, 0.00001683, 0.00002623, 0.00003192, 0.00003396, 0.00003272, 0.00002890, 0.00002335, 0.00001703, 0.00001087, 0.00000566, 0.00000198, 0.00000015, 0.00000020, 0.00000191, 0.00000481, 0.00000828, 0.00001163, 0.00001419, 0.00001539, 0.00001483, 0.00001231, 0.00000791, 0.00000189, -0.00000526, -0.00001294, -0.00002048, -0.00002720, -0.00003255, -0.00003607, -0.00003755, -0.00003695, -0.00003444, -0.00003037, -0.00002524, -0.00001958, -0.00001396, -0.00000887, -0.00000472, -0.00000175, -0.00000005, 0.00000045, -0.00000004, -0.00000122, -0.00000275, -0.00000427, -0.00000548, -0.00000617, -0.00000625, -0.00000573, -0.00000474, -0.00000350, },
		{ -0.00002009, -0.00002098, -0.00002101, -0.00002002, -0.00001800, -0.00001508, -0.00001151, -0.00000766, -0.00000395, -0.00000077, 0.00000152, 0.00000269, 0.00000262, 0.00000138, -0.00000083, -0.00000364, -0.00000665, -0.00000938, -0.00001141, -0.00001239, -0.00001213, -0.00001060, -0.00000795, -0.00000453, -0.00000079, 0.00000270, 0.00000540, 0.00000681, 0.00000655, 0.00000447, 0.00000063, -0.00000468, -0.00001093, -0.00001743, -0.00002338, -0.00002797, -0.00003047, -0.00003035, -0.00002731, -0.00002137, -0.00001291, -0.00000258, 0.00000867, 0.00001971, 0.00002936, 0.00003645, 0.00004001, 0.00003936, 0.00003420, 0.00002462, 0.00001120, -0.00000511, -0.00002301, -0.00004098, -0.00005743, -0.00007083, -0.00007986, -0.00008354, -0.00008135, -0.00007325, -0.00005972, -0.00004171, -0.00002054, 0.00000217, 0.00002470, 0.00004534, 0.00006260, 0.00007526, 0.00008255, 0.00008413, 0.00008020, 0.00007136, 0.00005860, 0.00004319, 0.00002652, 0.00000999, -0.00000514, -0.00001784, -0.00002740, -0.00003347, -0.00003607, -0.00003555, -0.00003253, -0.00002780, -0.00002222, -0.00001663, -0.00001174, -0.00000808, -0.00000592, -0.00000531, -0.00000606, -0.00000777, -0.00000997, -0.00001211, -0.00001368, -0.00001427, -0.00001366, -0.00001179, -0.00000879, -0.00000499, -0.00000082, 0.00000318, 0.00000650, 0.00000866, 0.00000933, 0.00000832, 0.00000565, 0.00000149, -0.00000378, -0.00000969, -0.00001570, -0.00002126, -0.00002589, -0.00002924, -0.00003108, -0.00003139, -0.00003030, -0.00002809, -0.00002516, -0.00002195, -0.00001892, -0.00001643, -0.00001476, -0.00001407, -0.00001432, -0.00001536, -0.00001691, -0.00001861, },
		{ -0.00001109, -0.00000932, -0.00000911, -0.00001054, -0.00001346, -0.00001754, -0.00002232, -0.00002722, -0.00003166, -0.00003509, -0.00003709, -0.00003736, -0.00003586, -0.00003272, -0.00002827, -0.00002300, -0.00001751, -0.00001241, -0.00000826, -0.00000548, -0.00000434, -0.00000487, -0.00000687, -0.00000996, -0.00001356, -0.00001704, -0.00001972, -0.00002104, -0.00002056, -0.00001809, -0.00001371, -0.00000775, -0.00000080, 0.00000635, 0.00001282, 0.00001772, 0.00002024, 0.00001982, 0.00001615, 0.00000929, -0.00000033, -0.00001193, -0.00002447, -0.00003667, -0.00004721, -0.00005482, -0.00005845, -0.00005734, -0.00005117, -0.00004009, -0.00002471, -0.00000611, 0.00001431, 0.00003488, 0.00005386, 0.00006959, 0.00008064, 0.00008593, 0.00008490, 0.00007747, 0.00006414, 0.00004590, 0.00002415, 0.00000057, -0.00002299, -0.00004475, -0.00006310, -0.00007675, -0.00008484, -0.00008699, -0.00008333, -0.00007444, -0.00006132, -0.00004523, -0.00002757, -0.00000977, 0.00000686, 0.00002124, 0.00003262, 0.00004057, 0.00004505, 0.00004634, 0.00004496, 0.00004165, 0.00003721, 0.00003243, 0.00002798, 0.00002439, 0.00002194, 0.00002071, 0.00002057, 0.00002122, 0.00002224, 0.00002318, 0.00002361, 0.00002319, 0.00002171, 0.00001913, 0.00001555, 0.00001121, 0.00000647, 0.00000170, -0.00000269, -0.00000639, -0.00000918, -0.00001097, -0.00001182, -0.00001192, -0.00001156, -0.00001110, -0.00001089, -0.00001126, -0.00001242, -0.00001448, -0.00001738, -0.00002093, -0.00002483, -0.00002866, -0.00003202, -0.00003449, -0.00003575, -0.00003561, -0.00003404, -0.00003116, -0.00002728, -0.00002282, -0.00001829, -0.00001422, },
		{ -0.00005960, -0.00005707, -0.00005201, -0.00004478, -0.00003593, -0.00002618, -0.00001634, -0.00000728, 0.00000023, 0.00000554, 0.00000822, 0.00000810, 0.00000530, 0.00000021, -0.00000657, -0.00001424, -0.00002195, -0.00002886, -0.00003423, -0.00003755, -0.00003853, -0.00003719, -0.00003385, -0.00002906, -0.00002357, -0.00001822, -0.00001383, -0.00001113, -0.00001062, -0.00001250, -0.00001669, -0.00002275, -0.00002995, -0.00003735, -0.00004392, -0.00004859, -0.00005046, -0.00004886, -0.00004348, -0.00003442, -0.00002218, -0.00000769, 0.00000783, 0.00002293, 0.00003609, 0.00004588, 0.00005111, 0.00005094, 0.00004503, 0.00003355, 0.00001722, -0.00000276, -0.00002479, -0.00004702, -0.00006752, -0.00008443, -0.00009618, -0.00010157, -0.00009996, -0.00009129, -0.00007608, -0.00005540, -0.00003079, -0.00000409, 0.00002271, 0.00004766, 0.00006898, 0.00008525, 0.00009551, 0.00009931, 0.00009675, 0.00008843, 0.00007535, 0.00005885, 0.00004043, 0.00002162, 0.00000382, -0.00001176, -0.00002425, -0.00003314, -0.00003828, -0.00003988, -0.00003843, -0.00003461, -0.00002921, -0.00002304, -0.00001680, -0.00001105, -0.00000618, -0.00000232, 0.00000056, 0.00000268, 0.00000434, 0.00000594, 0.00000781, 0.00001024, 0.00001342, 0.00001737, 0.00002200, 0.00002708, 0.00003230, 0.00003730, 0.00004172, 0.00004525, 0.00004766, 0.00004884, 0.00004876, 0.00004754, 0.00004535, 0.00004242, 0.00003897, 0.00003522, 0.00003130, 0.00002726, 0.00002307, 0.00001863, 0.00001379, 0.00000835, 0.00000220, -0.00000475, -0.00001246, -0.00002075, -0.00002933, -0.00003777, -0.00004554, -0.00005210, -0.00005691, -0.00005951, },
		{ 0.00000890, -0.00000731, -0.00002444, -0.00004168, -0.00005811, -0.00007275, -0.00008467, -0.00009305, -0.00009730, -0.00009711, -0.00009249, -0.00008379, -0.00007172, -0.00005723, -0.00004149, -0.00002576, -0.00001126, 0.00000091, 0.00000992, 0.00001527, 0.00001684, 0.00001491, 0.00001010, 0.00000334, -0.00000426, -0.00001152, -0.00001735, -0.00002086, -0.00002145, -0.00001892, -0.00001351, -0.00000583, 0.00000315, 0.00001223, 0.00002007, 0.00002543, 0.00002720, 0.00002465, 0.00001746, 0.00000581, -0.00000959, -0.00002759, -0.00004667, -0.00006508, -0.00008101, -0.00009276, -0.00009895, -0.00009864, -0.00009143, -0.00007758, -0.00005794, -0.00003393, -0.00000741, 0.00001948, 0.00004451, 0.00006556, 0.00008081, 0.00008894, 0.00008919, 0.00008148, 0.00006642, 0.00004518, 0.00001950, -0.00000859, -0.00003686, -0.00006312, -0.00008541, -0.00010213, -0.00011221, -0.00011512, -0.00011095, -0.00010034, -0.00008438, -0.00006453, -0.00004244, -0.00001979, 0.00000182, 0.00002107, 0.00003693, 0.00004879, 0.00005642, 0.00005997, 0.00005993, 0.00005701, 0.00005204, 0.00004589, 0.00003938, 0.00003316, 0.00002771, 0.00002331, 0.00002001, 0.00001771, 0.00001615, 0.00001503, 0.00001404, 0.00001291, 0.00001146, 0.00000963, 0.00000748, 0.00000519, 0.00000300, 0.00000122, 0.00000013, 0.00000001, 0.00000102, 0.00000325, 0.00000668, 0.00001117, 0.00001652, 0.00002247, 0.00002872, 0.00003499, 0.00004105, 0.00004669, 0.00005179, 0.00005626, 0.00006006, 0.00006314, 0.00006546, 0.00006693, 0.00006740, 0.00006667, 0.00006449, 0.00006058, 0.00005467, 0.00004657, 0.00003618, 0.00002354, },
		{ 0.00009720, 0.00010342, 0.00010768, 0.00010932, 0.00010767, 0.00010219, 0.00009253, 0.00007854, 0.00006041, 0.00003861, 0.00001398, -0.00001239, -0.00003916, -0.00006486, -0.00008800, -0.00010722, -0.00012136, -0.00012965, -0.00013171, -0.00012767, -0.00011813, -0.00010416, -0.00008716, -0.00006877, -0.00005071, -0.00003461, -0.00002185, -0.00001345, -0.00000988, -0.00001113, -0.00001658, -0.00002516, -0.00003539, -0.00004556, -0.00005391, -0.00005882, -0.00005900, -0.00005361, -0.00004242, -0.00002583, -0.00000485, 0.00001894, 0.00004356, 0.00006681, 0.00008646, 0.00010048, 0.00010725, 0.00010575, 0.00009562, 0.00007728, 0.00005189, 0.00002125, -0.00001234, -0.00004627, -0.00007785, -0.00010456, -0.00012427, -0.00013541, -0.00013713, -0.00012935, -0.00011278, -0.00008883, -0.00005948, -0.00002712, 0.00000570, 0.00003646, 0.00006291, 0.00008320, 0.00009611, 0.00010102, 0.00009801, 0.00008777, 0.00007156, 0.00005097, 0.00002787, 0.00000415, -0.00001839, -0.00003823, -0.00005421, -0.00006560, -0.00007211, -0.00007390, -0.00007147, -0.00006560, -0.00005724, -0.00004737, -0.00003694, -0.00002675, -0.00001741, -0.00000931, -0.00000261, 0.00000275, 0.00000695, 0.00001029, 0.00001309, 0.00001566, 0.00001822, 0.00002090, 0.00002374, 0.00002662, 0.00002941, 0.00003187, 0.00003378, 0.00003495, 0.00003526, 0.00003465, 0.00003318, 0.00003099, 0.00002830, 0.00002539, 0.00002255, 0.00002006, 0.00001817, 0.00001708, 0.00001692, 0.00001775, 0.00001958, 0.00002239, 0.00002611, 0.00003069, 0.00003608, 0.00004221, 0.00004906, 0.00005654, 0.00006455, 0.00007294, 0.00008144, 0.00008968, },
		{ 0.00003096, 0.00003396, 0.00003852, 0.00004484, 0.00005303, 0.00006307, 0.00007474, 0.00008764, 0.00010114, 0.00011440, 0.00012642, 0.00013609, 0.00014228, 0.00014392, 0.00014018, 0.00013047, 0.00011463, 0.00009293, 0.00006613, 0.00003544, 0.00000249, -0.00003083, -0.00006244, -0.00009033, -0.00011264, -0.00012791, -0.00013520, -0.00013422, -0.00012532, -0.00010959, -0.00008870, -0.00006480, -0.00004035, -0.00001789, 0.00000022, 0.00001197, 0.00001595, 0.00001149, -0.00000127, -0.00002134, -0.00004700, -0.00007588, -0.00010521, -0.00013203, -0.00015351, -0.00016718, -0.00017117, -0.00016442, -0.00014679, -0.00011905, -0.00008291, -0.00004080, 0.00000428, 0.00004901, 0.00009008, 0.00012443, 0.00014955, 0.00016364, 0.00016581, 0.00015610, 0.00013551, 0.00010584, 0.00006958, 0.00002964, -0.00001088, -0.00004896, -0.00008190, -0.00010755, -0.00012444, -0.00013185, -0.00012988, -0.00011934, -0.00010166, -0.00007873, -0.00005268, -0.00002572, 0.00000009, 0.00002298, 0.00004160, 0.00005508, 0.00006310, 0.00006580, 0.00006375, 0.00005784, 0.00004914, 0.00003882, 0.00002797, 0.00001755, 0.00000829, 0.00000070, -0.00000501, -0.00000884, -0.00001098, -0.00001176, -0.00001156, -0.00001077, -0.00000971, -0.00000860, -0.00000757, -0.00000663, -0.00000569, -0.00000460, -0.00000319, -0.00000130, 0.00000119, 0.00000432, 0.00000808, 0.00001233, 0.00001690, 0.00002156, 0.00002607, 0.00003019, 0.00003370, 0.00003648, 0.00003844, 0.00003956, 0.00003990, 0.00003956, 0.00003866, 0.00003736, 0.00003580, 0.00003412, 0.00003246, 0.00003094, 0.00002970, 0.00002888, 0.00002868, 0.00002929, },
		{ 0.00008009, 0.00008171, 0.00008249, 0.00008239, 0.00008145, 0.00007978, 0.00007758, 0.00007519, 0.00007303, 0.00007161, 0.00007149, 0.00007319, 0.00007716, 0.00008369, 0.00009283, 0.00010432, 0.00011761, 0.00013178, 0.00014563, 0.00015769, 0.00016638, 0.00017011, 0.00016744, 0.00015724, 0.00013884, 0.00011215, 0.00007775, 0.00003691, -0.00000844, -0.00005583, -0.00010236, -0.00014496, -0.00018061, -0.00020660, -0.00022079, -0.00022181, -0.00020922, -0.00018364, -0.00014668, -0.00010092, -0.00004972, 0.00000306, 0.00005330, 0.00009700, 0.00013059, 0.00015122, 0.00015706, 0.00014740, 0.00012276, 0.00008487, 0.00003651, -0.00001867, -0.00007649, -0.00013255, -0.00018261, -0.00022291, -0.00025049, -0.00026338, -0.00026080, -0.00024316, -0.00021201, -0.00016988, -0.00012007, -0.00006631, -0.00001250, 0.00003764, 0.00008084, 0.00011456, 0.00013709, 0.00014770, 0.00014661, 0.00013489, 0.00011436, 0.00008735, 0.00005648, 0.00002439, -0.00000640, -0.00003378, -0.00005612, -0.00007242, -0.00008224, -0.00008576, -0.00008363, -0.00007689, -0.00006680, -0.00005471, -0.00004191, -0.00002955, -0.00001851, -0.00000937, -0.00000241, 0.00000235, 0.00000515, 0.00000638, 0.00000650, 0.00000601, 0.00000534, 0.00000483, 0.00000468, 0.00000500, 0.00000575, 0.00000681, 0.00000802, 0.00000920, 0.00001021, 0.00001094, 0.00001138, 0.00001157, 0.00001162, 0.00001172, 0.00001204, 0.00001277, 0.00001407, 0.00001606, 0.00001879, 0.00002225, 0.00002636, 0.00003102, 0.00003609, 0.00004140, 0.00004680, 0.00005215, 0.00005735, 0.00006227, 0.00006685, 0.00007100, 0.00007465, 0.00007771, },
		{ 0.00001057, 0.00001781, 0.00002590, 0.00003484, 0.00004460, 0.00005510, 0.00006619, 0.00007766, 0.00008924, 0.00010062, 0.00011147, 0.00012149, 0.00013046, 0.00013825, 0.00014492, 0.00015070, 0.00015599, 0.00016139, 0.00016761, 0.00017540, 0.00018548, 0.00019842, 0.00021452, 0.00023369, 0.00025543, 0.00027871, 0.00030200, 0.00032334, 0.00034042, 0.00035075, 0.00035186, 0.00034153, 0.00031801, 0.00028024, 0.00022807, 0.00016235, 0.00008497, -0.00000111, -0.00009206, -0.00018335, -0.00027003, -0.00034710, -0.00040985, -0.00045424, -0.00047724, -0.00047708, -0.00045346, -0.00040759, -0.00034217, -0.00026124, -0.00016987, -0.00007386, 0.00002071, 0.00010790, 0.00018238, 0.00023973, 0.00027685, 0.00029209, 0.00028538, 0.00025817, 0.00021329, 0.00015470, 0.00008715, 0.00001577, -0.00005432, -0.00011835, -0.00017234, -0.00021326, -0.00023925, -0.00024969, -0.00024513, -0.00022721, -0.00019840, -0.00016174, -0.00012060, -0.00007832, -0.00003798, -0.00000218, 0.00002715, 0.00004879, 0.00006230, 0.00006794, 0.00006656, 0.00005944, 0.00004815, 0.00003435, 0.00001963, 0.00000537, -0.00000734, -0.00001778, -0.00002559, -0.00003075, -0.00003351, -0.00003435, -0.00003382, -0.00003252, -0.00003100, -0.00002968, -0.00002884, -0.00002862, -0.00002901, -0.00002987, -0.00003100, -0.00003217, -0.00003314, -0.00003374, -0.00003386, -0.00003346, -0.00003260, -0.00003137, -0.00002992, -0.00002840, -0.00002695, -0.00002569, -0.00002465, -0.00002385, -0.00002322, -0.00002267, -0.00002208, -0.00002130, -0.00002019, -0.00001863, -0.00001653, -0.00001380, -0.00001041, -0.00000630, -0.00000145, 0.00000416, },
		{ -0.00005912, -0.00006220, -0.00006534, -0.00006848, -0.00007157, -0.00007451, -0.00007715, -0.00007932, -0.00008076, -0.00008123, -0.00008043, -0.00007812, -0.00007407, -0.00006814, -0.00006032, -0.00005070, -0.00003948, -0.00002699, -0.00001362, 0.00000023, 0.00001420, 0.00002808, 0.00004182, 0.00005566, 0.00007009, 0.00008590, 0.00010413, 0.00012597, 0.00015268, 0.00018541, 0.00022503, 0.00027199, 0.00032611, 0.00038649, 0.00045143, 0.00051839, 0.00058408, 0.00064462, 0.00069568, 0.00073288, 0.00075200, 0.00074941, 0.00072238, 0.00066936, 0.00059025, 0.00048652, 0.00036124, 0.00021897, 0.00006559, -0.00009206, -0.00024657, -0.00039046, -0.00051663, -0.00061887, -0.00069229, -0.00073367, -0.00074164, -0.00071682, -0.00066175, -0.00058071, -0.00047938, -0.00036448, -0.00024321, -0.00012285, -0.00001017, 0.00008893, 0.00016981, 0.00022936, 0.00026607, 0.00028008, 0.00027299, 0.00024770, 0.00020803, 0.00015840, 0.00010346, 0.00004767, -0.00000496, -0.00005118, -0.00008861, -0.00011587, -0.00013255, -0.00013911, -0.00013675, -0.00012719, -0.00011246, -0.00009467, -0.00007584, -0.00005770, -0.00004158, -0.00002838, -0.00001853, -0.00001206, -0.00000863, -0.00000766, -0.00000845, -0.00001024, -0.00001236, -0.00001426, -0.00001557, -0.00001611, -0.00001588, -0.00001502, -0.00001378, -0.00001246, -0.00001135, -0.00001070, -0.00001068, -0.00001138, -0.00001278, -0.00001478, -0.00001725, -0.00001999, -0.00002283, -0.00002560, -0.00002819, -0.00003054, -0.00003264, -0.00003451, -0.00003624, -0.00003791, -0.00003960, -0.00004141, -0.00004338, -0.00004555, -0.00004793, -0.00005050, -0.00005324, -0.00005613, },
		{ -0.00001759, -0.00001835, -0.00001922, -0.00002022, -0.00002136, -0.00002271, -0.00002432, -0.00002629, -0.00002877, -0.00003191, -0.00003588, -0.00004081, -0.00004682, -0.00005396, -0.00006219, -0.00007137, -0.00008126, -0.00009154, -0.00010181, -0.00011166, -0.00012072, -0.00012867, -0.00013534, -0.00014071, -0.00014494, -0.00014837, -0.00015145, -0.00015473, -0.00015870, -0.00016374, -0.00016997, -0.00017713, -0.00018455, -0.00019102, -0.00019484, -0.00019385, -0.00018556, -0.00016730, -0.00013647, -0.00009078, -0.00002855, 0.00005107, 0.00014785, 0.00026029, 0.00038559, 0.00051964, 0.00065714, 0.00079187, 0.00091698, 0.00102545, 0.00111050, 0.00116613, 0.00118751, 0.00117142, 0.00111655, 0.00102366, 0.00089564, 0.00073738, 0.00055553, 0.00035813, 0.00015407, -0.00004738, -0.00023722, -0.00040723, -0.00055050, -0.00066182, -0.00073795, -0.00077776, -0.00078219, -0.00075413, -0.00069809, -0.00061983, -0.00052588, -0.00042307, -0.00031808, -0.00021694, -0.00012477, -0.00004547, 0.00001841, 0.00006569, 0.00009645, 0.00011193, 0.00011420, 0.00010595, 0.00009017, 0.00006984, 0.00004772, 0.00002617, 0.00000697, -0.00000872, -0.00002030, -0.00002777, -0.00003153, -0.00003231, -0.00003102, -0.00002858, -0.00002584, -0.00002348, -0.00002198, -0.00002156, -0.00002222, -0.00002378, -0.00002592, -0.00002828, -0.00003046, -0.00003214, -0.00003310, -0.00003319, -0.00003242, -0.00003088, -0.00002875, -0.00002625, -0.00002363, -0.00002109, -0.00001883, -0.00001696, -0.00001555, -0.00001460, -0.00001406, -0.00001387, -0.00001393, -0.00001417, -0.00001451, -0.00001491, -0.00001535, -0.00001582, -0.00001634, -0.00001692, },
		{ -0.00002172, -0.00002174, -0.00002180, -0.00002191, -0.00002206, -0.00002225, -0.00002244, -0.00002259, -0.00002262, -0.00002244, -0.00002193, -0.00002100, -0.00001957, -0.00001762, -0.00001518, -0.00001238, -0.00000939, -0.00000649, -0.00000397, -0.00000217, -0.00000140, -0.00000189, -0.00000379, -0.00000712, -0.00001174, -0.00001739, -0.00002368, -0.00003019, -0.00003650, -0.00004225, -0.00004730, -0.00005172, -0.00005588, -0.00006048, -0.00006653, -0.00007524, -0.00008800, -0.00010615, -0.00013084, -0.00016284, -0.00020234, -0.00024878, -0.00030076, -0.00035594, -0.00041111, -0.00046227, -0.00050485, -0.00053398, -0.00054483, -0.00053301, -0.00049492, -0.00042819, -0.00033191, -0.00020694, -0.00005596, 0.00011652, 0.00030433, 0.00049988, 0.00069465, 0.00087960, 0.00104579, 0.00118489, 0.00128980, 0.00135503, 0.00137717, 0.00135503, 0.00128980, 0.00118489, 0.00104579, 0.00087960, 0.00069465, 0.00049988, 0.00030433, 0.00011652, -0.00005596, -0.00020694, -0.00033191, -0.00042819, -0.00049492, -0.00053301, -0.00054483, -0.00053398, -0.00050485, -0.00046227, -0.00041111, -0.00035594, -0.00030076, -0.00024878, -0.00020234, -0.00016284, -0.00013084, -0.00010615, -0.00008800, -0.00007524, -0.00006653, -0.00006048, -0.00005588, -0.00005172, -0.00004730, -0.00004225, -0.00003650, -0.00003019, -0.00002368, -0.00001739, -0.00001174, -0.00000712, -0.00000379, -0.00000189, -0.00000140, -0.00000217, -0.00000397, -0.00000649, -0.00000939, -0.00001238, -0.00001518, -0.00001762, -0.00001957, -0.00002100, -0.00002193, -0.00002244, -0.00002262, -0.00002259, -0.00002244, -0.00002225, -0.00002206, -0.00002191, -0.00002180, -0.00002174, },
		{ -0.00001759, -0.00001692, -0.00001634, -0.00001582, -0.00001535, -0.00001491, -0.00001451, -0.00001417, -0.00001393, -0.00001387, -0.00001406, -0.00001460, -0.00001555, -0.00001696, -0.00001883, -0.00002109, -0.00002363, -0.00002625, -0.00002875, -0.00003088, -0.00003242, -0.00003319, -0.00003310, -0.00003214, -0.00003046, -0.00002828, -0.00002592, -0.00002378, -0.00002222, -0.00002156, -0.00002198, -0.00002348, -0.00002584, -0.00002858, -0.00003102, -0.00003231, -0.00003153, -0.00002777, -0.00002030, -0.00000872, 0.00000697, 0.00002617, 0.00004772, 0.00006984, 0.00009017, 0.00010595, 0.00011420, 0.00011193, 0.00009645, 0.00006569, 0.00001841, -0.00004547, -0.00012477, -0.00021694, -0.00031808, -0.00042307, -0.00052588, -0.00061983, -0.00069809, -0.00075413, -0.00078219, -0.00077776, -0.00073795, -0.00066182, -0.00055050, -0.00040723, -0.00023722, -0.00004738, 0.00015407, 0.00035813, 0.00055553, 0.00073738, 0.00089564, 0.00102366, 0.00111655, 0.00117142, 0.00118751, 0.00116613, 0.00111050, 0.00102545, 0.00091698, 0.00079187, 0.00065714, 0.00051964, 0.00038559, 0.00026029, 0.00014785, 0.00005107, -0.00002855, -0.00009078, -0.00013647, -0.00016730, -0.00018556, -0.00019385, -0.00019484, -0.00019102, -0.00018455, -0.00017713, -0.00016997, -0.00016374, -0.00015870, -0.00015473, -0.00015145, -0.00014837, -0.00014494, -0.00014071, -0.00013534, -0.00012867, -0.00012072, -0.00011166, -0.00010181, -0.00009154, -0.00008126, -0.00007137, -0.00006219, -0.00005396, -0.00004682, -0.00004081, -0.00003588, -0.00003191, -0.00002877, -0.00002629, -0.00002432, -0.00002271, -0.00002136, -0.00002022, -0.00001922, -0.00001835, },
		{ -0.00005912, -0.00005613, -0.00005324, -0.00005050, -0.00004793, -0.00004555, -0.00004338, -0.00004141, -0.00003960, -0.00003791, -0.00003624, -0.00003451, -0.00003264, -0.00003054, -0.00002819, -0.00002560, -0.00002283, -0.00001999, -0.00001725, -0.00001478, -0.00001278, -0.00001138, -0.00001068, -0.00001070, -0.00001135, -0.00001246, -0.00001378, -0.00001502, -0.00001588, -0.00001611, -0.00001557, -0.00001426, -0.00001236, -0.00001024, -0.00000845, -0.00000766, -0.00000863, -0.00001206, -0.00001853, -0.00002838, -0.00004158, -0.00005770, -0.00007584, -0.00009467, -0.00011246, -0.00012719, -0.00013675, -0.00013911, -0.00013255, -0.00011587, -0.00008861, -0.00005118, -0.00000496, 0.00004767, 0.00010346, 0.00015840, 0.00020803, 0.00024770, 0.00027299, 0.00028008, 0.00026607, 0.00022936, 0.00016981, 0.00008893, -0.00001017, -0.00012285, -0.00024321, -0.00036448, -0.00047938, -0.00058071, -0.00066175, -0.00071682, -0.00074164, -0.00073367, -0.00069229, -0.00061887, -0.00051663, -0.00039046, -0.00024657, -0.00009206, 0.00006559, 0.00021897, 0.00036124, 0.00048652, 0.00059025, 0.00066936, 0.00072238, 0.00074941, 0.00075200, 0.00073288, 0.00069568, 0.00064462, 0.00058408, 0.00051839, 0.00045143, 0.00038649, 0.00032611, 0.00027199, 0.00022503, 0.00018541, 0.00015268, 0.00012597, 0.00010413, 0.00008590, 0.00007009, 0.00005566, 0.00004182, 0.00002808, 0.00001420, 0.00000023, -0.00001362, -0.00002699, -0.00003948, -0.00005070, -0.00006032, -0.00006814, -0.00007407, -0.00007812, -0.00008043, -0.00008123, -0.00008076, -0.00007932, -0.00007715, -0.00007451, -0.00007157, -0.00006848, -0.00006534, -0.00006220, },
		{ 0.00001057, 0.00000416, -0.00000145, -0.00000630, -0.00001041, -0.00001380, -0.00001653, -0.00001863, -0.00002019, -0.00002130, -0.00002208, -0.00002267, -0.00002322, -0.00002385, -0.00002465, -0.00002569, -0.00002695, -0.00002840, -0.00002992, -0.00003137, -0.00003260, -0.00003346, -0.00003386, -0.00003374, -0.00003314, -0.00003217, -0.00003100, -0.00002987, -0.00002901, -0.00002862, -0.00002884, -0.00002968, -0.00003100, -0.00003252, -0.00003382, -0.00003435, -0.00003351, -0.00003075, -0.00002559, -0.00001778, -0.00000734, 0.00000537, 0.00001963, 0.00003435, 0.00004815, 0.00005944, 0.00006656, 0.00006794, 0.00006230, 0.00004879, 0.00002715, -0.00000218, -0.00003798, -0.00007832, -0.00012060, -0.00016174, -0.00019840, -0.00022721, -0.00024513, -0.00024969, -0.00023925, -0.00021326, -0.00017234, -0.00011835, -0.00005432, 0.00001577, 0.00008715, 0.00015470, 0.00021329, 0.00025817, 0.00028538, 0.00029209, 0.00027685, 0.00023973, 0.00018238, 0.00010790, 0.00002071, -0.00007386, -0.00016987, -0.00026124, -0.00034217, -0.00040759, -0.00045346, -0.00047708, -0.00047724, -0.00045424, -0.00040985, -0.00034710, -0.00027003, -0.00018335, -0.00009206, -0.00000111, 0.00008497, 0.00016235, 0.00022807, 0.00028024, 0.00031801, 0.00034153, 0.00035186, 0.00035075, 0.00034042, 0.00032334, 0.00030200, 0.00027871, 0.00025543, 0.00023369, 0.00021452, 0.00019842, 0.00018548, 0.00017540, 0.00016761, 0.00016139, 0.00015599, 0.00015070, 0.00014492, 0.00013825, 0.00013046, 0.00012149, 0.00011147, 0.00010062, 0.00008924, 0.00007766, 0.00006619, 0.00005510, 0.00004460, 0.00003484, 0.00002590, 0.00001781, },
		{ 0.00008009, 0.00007771, 0.00007465, 0.00007100, 0.00006685, 0.00006227, 0.00005735, 0.00005215, 0.00004680, 0.00004140, 0.00003609, 0.00003102, 0.00002636, 0.00002225, 0.00001879, 0.00001606, 0.00001407, 0.00001277, 0.00001204, 0.00001172, 0.00001162, 0.00001157, 0.00001138, 0.00001094, 0.00001021, 0.00000920, 0.00000802, 0.00000681, 0.00000575, 0.00000500, 0.00000468, 0.00000483, 0.00000534, 0.00000601, 0.00000650, 0.00000638, 0.00000515, 0.00000235, -0.00000241, -0.00000937, -0.00001851, -0.00002955, -0.00004191, -0.00005471, -0.00006680, -0.00007689, -0.00008363, -0.00008576, -0.00008224, -0.00007242, -0.00005612, -0.00003378, -0.00000640, 0.00002439, 0.00005648, 0.00008735, 0.00011436, 0.00013489, 0.00014661, 0.00014770, 0.00013709, 0.00011456, 0.00008084, 0.00003764, -0.00001250, -0.00006631, -0.00012007, -0.00016988, -0.00021201, -0.00024316, -0.00026080, -0.00026338, -0.00025049, -0.00022291, -0.00018261, -0.00013255, -0.00007649, -0.00001867, 0.00003651, 0.00008487, 0.00012276, 0.00014740, 0.00015706, 0.00015122, 0.00013059, 0.00009700, 0.00005330, 0.00000306, -0.00004972, -0.00010092, -0.00014668, -0.00018364, -0.00020922, -0.00022181, -0.00022079, -0.00020660, -0.00018061, -0.00014496, -0.00010236, -0.00005583, -0.00000844, 0.00003691, 0.00007775, 0.00011215, 0.00013884, 0.00015724, 0.00016744, 0.00017011, 0.00016638, 0.00015769, 0.00014563, 0.00013178, 0.00011761, 0.00010432, 0.00009283, 0.00008369, 0.00007716, 0.00007319, 0.00007149, 0.00007161, 0.00007303, 0.00007519, 0.00007758, 0.00007978, 0.00008145, 0.00008239, 0.00008249, 0.00008171, },
		{ 0.00003096, 0.00002929, 0.00002868, 0.00002888, 0.00002970, 0.00003094, 0.00003246, 0.00003412, 0.00003580, 0.00003736, 0.00003866, 0.00003956, 0.00003990, 0.00003956, 0.00003844, 0.00003648, 0.00003370, 0.00003019, 0.00002607, 0.00002156, 0.00001690, 0.00001233, 0.00000808, 0.00000432, 0.00000119, -0.00000130, -0.00000319, -0.00000460, -0.00000569, -0.00000663, -0.00000757, -0.00000860, -0.00000971, -0.00001077, -0.00001156, -0.00001176, -0.00001098, -0.00000884, -0.00000501, 0.00000070, 0.00000829, 0.00001755, 0.00002797, 0.00003882, 0.00004914, 0.00005784, 0.00006375, 0.00006580, 0.00006310, 0.00005508, 0.00004160, 0.00002298, 0.00000009, -0.00002572, -0.00005268, -0.00007873, -0.00010166, -0.00011934, -0.00012988, -0.00013185, -0.00012444, -0.00010755, -0.00008190, -0.00004896, -0.00001088, 0.00002964, 0.00006958, 0.00010584, 0.00013551, 0.00015610, 0.00016581, 0.00016364, 0.00014955, 0.00012443, 0.00009008, 0.00004901, 0.00000428, -0.00004080, -0.00008291, -0.00011905, -0.00014679, -0.00016442, -0.00017117, -0.00016718, -0.00015351, -0.00013203, -0.00010521, -0.00007588, -0.00004700, -0.00002134, -0.00000127, 0.00001149, 0.00001595, 0.00001197, 0.00000022, -0.00001789, -0.00004035, -0.00006480, -0.00008870, -0.00010959, -0.00012532, -0.00013422, -0.00013520, -0.00012791, -0.00011264, -0.00009033, -0.00006244, -0.00003083, 0.00000249, 0.00003544, 0.00006613, 0.00009293, 0.00011463, 0.00013047, 0.00014018, 0.00014392, 0.00014228, 0.00013609, 0.00012642, 0.00011440, 0.00010114, 0.00008764, 0.00007474, 0.00006307, 0.00005303, 0.00004484, 0.00003852, 0.00003396, },
		{ 0.00009720, 0.00008968, 0.00008144, 0.00007294, 0.00006455, 0.00005654, 0.00004906, 0.00004221, 0.00003608, 0.00003069, 0.00002611, 0.00002239, 0.00001958, 0.00001775, 0.00001692, 0.00001708, 0.00001817, 0.00002006, 0.00002255, 0.00002539, 0.00002830, 0.00003099, 0.00003318, 0.00003465, 0.00003526, 0.00003495, 0.00003378, 0.00003187, 0.00002941, 0.00002662, 0.00002374, 0.00002090, 0.00001822, 0.00001566, 0.00001309, 0.00001029, 0.00000695, 0.00000275, -0.00000261, -0.00000931, -0.00001741, -0.00002675, -0.00003694, -0.00004737, -0.00005724, -0.00006560, -0.00007147, -0.00007390, -0.00007211, -0.00006560, -0.00005421, -0.00003823, -0.00001839, 0.00000415, 0.00002787, 0.00005097, 0.00007156, 0.00008777, 0.00009801, 0.00010102, 0.00009611, 0.00008320, 0.00006291, 0.00003646, 0.00000570, -0.00002712, -0.00005948, -0.00008883, -0.00011278, -0.00012935, -0.00013713, -0.00013541, -0.00012427, -0.00010456, -0.00007785, -0.00004627, -0.00001234, 0.00002125, 0.00005189, 0.00007728, 0.00009562, 0.00010575, 0.00010725, 0.00010048, 0.00008646, 0.00006681, 0.00004356, 0.00001894, -0.00000485, -0.00002583, -0.00004242, -0.00005361, -0.00005900, -0.00005882, -0.00005391, -0.00004556, -0.00003539, -0.00002516, -0.00001658, -0.00001113, -0.00000988, -0.00001345, -0.00002185, -0.00003461, -0.00005071, -0.00006877, -0.00008716, -0.00010416, -0.00011813, -0.00012767, -0.00013171, -0.00012965, -0.00012136, -0.00010722, -0.00008800, -0.00006486, -0.00003916, -0.00001239, 0.00001398, 0.00003861, 0.00006041, 0.00007854, 0.00009253, 0.00010219, 0.00010767, 0.00010932, 0.00010768, 0.00010342, },
		{ 0.00000890, 0.00002354, 0.00003618, 0.00004657, 0.00005467, 0.00006058, 0.00006449, 0.00006667, 0.00006740, 0.00006693, 0.00006546, 0.00006314, 0.00006006, 0.00005626, 0.00005179, 0.00004669, 0.00004105, 0.00003499, 0.00002872, 0.00002247, 0.00001652, 0.00001117, 0.00000668, 0.00000325, 0.00000102, 0.00000001, 0.00000013, 0.00000122, 0.00000300, 0.00000519, 0.00000748, 0.00000963, 0.00001146, 0.00001291, 0.00001404, 0.00001503, 0.00001615, 0.00001771, 0.00002001, 0.00002331, 0.00002771, 0.00003316, 0.00003938, 0.00004589, 0.00005204, 0.00005701, 0.00005993, 0.00005997, 0.00005642, 0.00004879, 0.00003693, 0.00002107, 0.00000182, -0.00001979, -0.00004244, -0.00006453, -0.00008438, -0.00010034, -0.00011095, -0.00011512, -0.00011221, -0.00010213, -0.00008541, -0.00006312, -0.00003686, -0.00000859, 0.00001950, 0.00004518, 0.00006642, 0.00008148, 0.00008919, 0.00008894, 0.00008081, 0.00006556, 0.00004451, 0.00001948, -0.00000741, -0.00003393, -0.00005794, -0.00007758, -0.00009143, -0.00009864, -0.00009895, -0.00009276, -0.00008101, -0.00006508, -0.00004667, -0.00002759, -0.00000959, 0.00000581, 0.00001746, 0.00002465, 0.00002720, 0.00002543, 0.00002007, 0.00001223, 0.00000315, -0.00000583, -0.00001351, -0.00001892, -0.00002145, -0.00002086, -0.00001735, -0.00001152, -0.00000426, 0.00000334, 0.00001010, 0.00001491, 0.00001684, 0.00001527, 0.00000992, 0.00000091, -0.00001126, -0.00002576, -0.00004149, -0.00005723, -0.00007172, -0.00008379, -0.00009249, -0.00009711, -0.00009730, -0.00009305, -0.00008467, -0.00007275, -0.00005811, -0.00004168, -0.00002444, -0.00000731, },
		{ -0.00005960, -0.00005951, -0.00005691, -0.00005210, -0.00004554, -0.00003777, -0.00002933, -0.00002075, -0.00001246, -0.00000475, 0.00000220, 0.00000835, 0.00001379, 0.00001863, 0.00002307, 0.00002726, 0.00003130, 0.00003522, 0.00003897, 0.00004242, 0.00004535, 0.00004754, 0.00004876, 0.00004884, 0.00004766, 0.00004525, 0.00004172, 0.00003730, 0.00003230, 0.00002708, 0.00002200, 0.00001737, 0.00001342, 0.00001024, 0.00000781, 0.00000594, 0.00000434, 0.00000268, 0.00000056, -0.00000232, -0.00000618, -0.00001105, -0.00001680, -0.00002304, -0.00002921, -0.00003461, -0.00003843, -0.00003988, -0.00003828, -0.00003314, -0.00002425, -0.00001176, 0.00000382, 0.00002162, 0.00004043, 0.00005885, 0.00007535, 0.00008843, 0.00009675, 0.00009931, 0.00009551, 0.00008525, 0.00006898, 0.00004766, 0.00002271, -0.00000409, -0.00003079, -0.00005540, -0.00007608, -0.00009129, -0.00009996, -0.00010157, -0.00009618, -0.00008443, -0.00006752, -0.00004702, -0.00002479, -0.00000276, 0.00001722, 0.00003355, 0.00004503, 0.00005094, 0.00005111, 0.00004588, 0.00003609, 0.00002293, 0.00000783, -0.00000769, -0.00002218, -0.00003442, -0.00004348, -0.00004886, -0.00005046, -0.00004859, -0.00004392, -0.00003735, -0.00002995, -0.00002275, -0.00001669, -0.00001250, -0.00001062, -0.00001113, -0.00001383, -0.00001822, -0.00002357, -0.00002906, -0.00003385, -0.00003719, -0.00003853, -0.00003755, -0.00003423, -0.00002886, -0.00002195, -0.00001424, -0.00000657, 0.00000021, 0.00000530, 0.00000810, 0.00000822, 0.00000554, 0.00000023, -0.00000728, -0.00001634, -0.00002618, -0.00003593, -0.00004478, -0.00005201, -0.00005707, },
		{ -0.00001109, -0.00001422, -0.00001829, -0.00002282, -0.00002728, -0.00003116, -0.00003404, -0.00003561, -0.00003575, -0.00003449, -0.00003202, -0.00002866, -0.00002483, -0.00002093, -0.00001738, -0.00001448, -0.00001242, -0.00001126, -0.00001089, -0.00001110, -0.00001156, -0.00001192, -0.00001182, -0.00001097, -0.00000918, -0.00000639, -0.00000269, 0.00000170, 0.00000647, 0.00001121, 0.00001555, 0.00001913, 0.00002171, 0.00002319, 0.00002361, 0.00002318, 0.00002224, 0.00002122, 0.00002057, 0.00002071, 0.00002194, 0.00002439, 0.00002798, 0.00003243, 0.00003721, 0.00004165, 0.00004496, 0.00004634, 0.00004505, 0.00004057, 0.00003262, 0.00002124, 0.00000686, -0.00000977, -0.00002757, -0.00004523, -0.00006132, -0.00007444, -0.00008333, -0.00008699, -0.00008484, -0.00007675, -0.00006310, -0.00004475, -0.00002299, 0.00000057, 0.00002415, 0.00004590, 0.00006414, 0.00007747, 0.00008490, 0.00008593, 0.00008064, 0.00006959, 0.00005386, 0.00003488, 0.00001431, -0.00000611, -0.00002471, -0.00004009, -0.00005117, -0.00005734, -0.00005845, -0.00005482, -0.00004721, -0.00003667, -0.00002447, -0.00001193, -0.00000033, 0.00000929, 0.00001615, 0.00001982, 0.00002024, 0.00001772, 0.00001282, 0.00000635, -0.00000080, -0.00000775, -0.00001371, -0.00001809, -0.00002056, -0.00002104, -0.00001972, -0.00001704, -0.00001356, -0.00000996, -0.00000687, -0.00000487, -0.00000434, -0.00000548, -0.00000826, -0.00001241, -0.00001751, -0.00002300, -0.00002827, -0.00003272, -0.00003586, -0.00003736, -0.00003709, -0.00003509, -0.00003166, -0.00002722, -0.00002232, -0.00001754, -0.00001346, -0.00001054, -0.00000911, -0.00000932, },
		{ -0.00002009, -0.00001861, -0.00001691, -0.00001536, -0.00001432, -0.00001407, -0.00001476, -0.00001643, -0.00001892, -0.00002195, -0.00002516, -0.00002809, -0.00003030, -0.00003139, -0.00003108, -0.00002924, -0.00002589, -0.00002126, -0.00001570, -0.00000969, -0.00000378, 0.00000149, 0.00000565, 0.00000832, 0.00000933, 0.00000866, 0.00000650, 0.00000318, -0.00000082, -0.00000499, -0.00000879, -0.00001179, -0.00001366, -0.00001427, -0.00001368, -0.00001211, -0.00000997, -0.00000777, -0.00000606, -0.00000531, -0.00000592, -0.00000808, -0.00001174, -0.00001663, -0.00002222, -0.00002780, -0.00003253, -0.00003555, -0.00003607, -0.00003347, -0.00002740, -0.00001784, -0.00000514, 0.00000999, 0.00002652, 0.00004319, 0.00005860, 0.00007136, 0.00008020, 0.00008413, 0.00008255, 0.00007526, 0.00006260, 0.00004534, 0.00002470, 0.00000217, -0.00002054, -0.00004171, -0.00005972, -0.00007325, -0.00008135, -0.00008354, -0.00007986, -0.00007083, -0.00005743, -0.00004098, -0.00002301, -0.00000511, 0.00001120, 0.00002462, 0.00003420, 0.00003936, 0.00004001, 0.00003645, 0.00002936, 0.00001971, 0.00000867, -0.00000258, -0.00001291, -0.00002137, -0.00002731, -0.00003035, -0.00003047, -0.00002797, -0.00002338, -0.00001743, -0.00001093, -0.00000468, 0.00000063, 0.00000447, 0.00000655, 0.00000681, 0.00000540, 0.00000270, -0.00000079, -0.00000453, -0.00000795, -0.00001060, -0.00001213, -0.00001239, -0.00001141, -0.00000938, -0.00000665, -0.00000364, -0.00000083, 0.00000138, 0.00000262, 0.00000269, 0.00000152, -0.00000077, -0.00000395, -0.00000766, -0.00001151, -0.00001508, -0.00001800, -0.00002002, -0.00002101, -0.00002098, },
		{ -0.00000227, -0.00000350, -0.00000474, -0.00000573, -0.00000625, -0.00000617, -0.00000548, -0.00000427, -0.00000275, -0.00000122, -0.00000004, 0.00000045, -0.00000005, -0.00000175, -0.00000472, -0.00000887, -0.00001396, -0.00001958, -0.00002524, -0.00003037, -0.00003444, -0.00003695, -0.00003755, -0.00003607, -0.00003255, -0.00002720, -0.00002048, -0.00001294, -0.00000526, 0.00000189, 0.00000791, 0.00001231, 0.00001483, 0.00001539, 0.00001419, 0.00001163, 0.00000828, 0.00000481, 0.00000191, 0.00000020, 0.00000015, 0.00000198, 0.00000566, 0.00001087, 0.00001703, 0.00002335, 0.00002890, 0.00003272, 0.00003396, 0.00003192, 0.00002623, 0.00001683, 0.00000408, -0.00001129, -0.00002823, -0.00004544, -0.00006147, -0.00007490, -0.00008444, -0.00008906, -0.00008813, -0.00008146, -0.00006937, -0.00005263, -0.00003242, -0.00001025, 0.00001220, 0.00003322, 0.00005124, 0.00006494, 0.00007336, 0.00007605, 0.00007301, 0.00006475, 0.00005220, 0.00003661, 0.00001943, 0.00000220, -0.00001367, -0.00002693, -0.00003667, -0.00004238, -0.00004395, -0.00004165, -0.00003614, -0.00002831, -0.00001922, -0.00000997, -0.00000159, 0.00000508, 0.00000945, 0.00001121, 0.00001038, 0.00000724, 0.00000231, -0.00000372, -0.00001010, -0.00001610, -0.00002110, -0.00002461, -0.00002638, -0.00002637, -0.00002475, -0.00002186, -0.00001816, -0.00001418, -0.00001042, -0.00000732, -0.00000518, -0.00000417, -0.00000429, -0.00000538, -0.00000716, -0.00000930, -0.00001140, -0.00001313, -0.00001421, -0.00001447, -0.00001385, -0.00001245, -0.00001045, -0.00000811, -0.00000574, -0.00000364, -0.00000203, -0.00000110, -0.00000088, -0.00000132, },
		{ -0.00001782, -0.00001699, -0.00001617, -0.00001555, -0.00001525, -0.00001533, -0.00001576, -0.00001643, -0.00001713, -0.00001762, -0.00001764, -0.00001693, -0.00001535, -0.00001280, -0.00000936, -0.00000521, -0.00000066, 0.00000387, 0.00000793, 0.00001104, 0.00001279, 0.00001289, 0.00001118, 0.00000771, 0.00000272, -0.00000338, -0.00001002, -0.00001656, -0.00002232, -0.00002668, -0.00002917, -0.00002949, -0.00002757, -0.00002361, -0.00001805, -0.00001150, -0.00000474, 0.00000142, 0.00000619, 0.00000894, 0.00000925, 0.00000696, 0.00000227, -0.00000436, -0.00001215, -0.00002015, -0.00002727, -0.00003246, -0.00003476, -0.00003347, -0.00002820, -0.00001898, -0.00000621, 0.00000926, 0.00002628, 0.00004346, 0.00005928, 0.00007226, 0.00008109, 0.00008476, 0.00008268, 0.00007472, 0.00006126, 0.00004317, 0.00002171, -0.00000156, -0.00002491, -0.00004659, -0.00006503, -0.00007892, -0.00008736, -0.00008990, -0.00008661, -0.00007804, -0.00006515, -0.00004923, -0.00003173, -0.00001419, 0.00000199, 0.00001558, 0.00002571, 0.00003187, 0.00003395, 0.00003225, 0.00002738, 0.00002023, 0.00001181, 0.00000317, -0.00000472, -0.00001106, -0.00001530, -0.00001717, -0.00001669, -0.00001414, -0.00001003, -0.00000500, 0.00000024, 0.00000504, 0.00000879, 0.00001110, 0.00001173, 0.00001067, 0.00000811, 0.00000439, -0.00000004, -0.00000470, -0.00000908, -0.00001279, -0.00001552, -0.00001713, -0.00001760, -0.00001708, -0.00001581, -0.00001411, -0.00001231, -0.00001073, -0.00000964, -0.00000918, -0.00000943, -0.00001032, -0.00001171, -0.00001339, -0.00001513, -0.00001669, -0.00001788, -0.00001859, -0.00001877, -0.00001847, },
		{ -0.00000245, -0.00000374, -0.00000505, -0.00000623, -0.00000718, -0.00000786, -0.00000829, -0.00000858, -0.00000887, -0.00000936, -0.00001021, -0.00001160, -0.00001361, -0.00001624, -0.00001939, -0.00002284, -0.00002631, -0.00002944, -0.00003185, -0.00003320, -0.00003322, -0.00003176, -0.00002882, -0.00002456, -0.00001931, -0.00001352, -0.00000774, -0.00000256, 0.00000147, 0.00000391, 0.00000446, 0.00000305, -0.00000019, -0.00000486, -0.00001040, -0.00001608, -0.00002112, -0.00002473, -0.00002625, -0.00002520, -0.00002137, -0.00001486, -0.00000609, 0.00000424, 0.00001517, 0.00002560, 0.00003436, 0.00004038, 0.00004276, 0.00004090, 0.00003460, 0.00002405, 0.00000988, -0.00000690, -0.00002495, -0.00004275, -0.00005873, -0.00007138, -0.00007943, -0.00008196, -0.00007853, -0.00006915, -0.00005439, -0.00003526, -0.00001317, 0.00001023, 0.00003314, 0.00005382, 0.00007073, 0.00008264, 0.00008876, 0.00008880, 0.00008296, 0.00007194, 0.00005682, 0.00003898, 0.00001995, 0.00000128, -0.00001561, -0.00002955, -0.00003970, -0.00004563, -0.00004731, -0.00004511, -0.00003972, -0.00003206, -0.00002318, -0.00001414, -0.00000592, 0.00000071, 0.00000521, 0.00000732, 0.00000711, 0.00000485, 0.00000106, -0.00000362, -0.00000852, -0.00001296, -0.00001640, -0.00001845, -0.00001891, -0.00001776, -0.00001521, -0.00001159, -0.00000734, -0.00000294, 0.00000115, 0.00000454, 0.00000694, 0.00000822, 0.00000837, 0.00000752, 0.00000591, 0.00000383, 0.00000161, -0.00000045, -0.00000212, -0.00000322, -0.00000371, -0.00000362, -0.00000308, -0.00000226, -0.00000137, -0.00000061, -0.00000016, -0.00000011, -0.00000051, -0.00000132, },
		{ -0.00000013, 0.00000102, 0.00000212, 0.00000304, 0.00000368, 0.00000401, 0.00000403, 0.00000383, 0.00000354, 0.00000329, 0.00000325, 0.00000354, 0.00000421, 0.00000527, 0.00000661, 0.00000807, 0.00000941, 0.00001036, 0.00001064, 0.00001000, 0.00000830, 0.00000547, 0.00000158, -0.00000315, -0.00000839, -0.00001371, -0.00001865, -0.00002272, -0.00002552, -0.00002674, -0.00002625, -0.00002411, -0.00002058, -0.00001609, -0.00001126, -0.00000675, -0.00000327, -0.00000145, -0.00000177, -0.00000450, -0.00000965, -0.00001693, -0.00002576, -0.00003532, -0.00004463, -0.00005259, -0.00005817, -0.00006044, -0.00005872, -0.00005268, -0.00004236, -0.00002822, -0.00001110, 0.00000780, 0.00002706, 0.00004512, 0.00006047, 0.00007173, 0.00007782, 0.00007806, 0.00007226, 0.00006071, 0.00004419, 0.00002394, 0.00000147, -0.00002146, -0.00004310, -0.00006177, -0.00007607, -0.00008499, -0.00008795, -0.00008490, -0.00007627, -0.00006296, -0.00004619, -0.00002746, -0.00000832, 0.00000969, 0.00002523, 0.00003727, 0.00004512, 0.00004852, 0.00004762, 0.00004297, 0.00003537, 0.00002585, 0.00001554, 0.00000552, -0.00000326, -0.00001004, -0.00001438, -0.00001610, -0.00001533, -0.00001245, -0.00000803, -0.00000279, 0.00000257, 0.00000736, 0.00001102, 0.00001316, 0.00001361, 0.00001239, 0.00000970, 0.00000591, 0.00000147, -0.00000312, -0.00000738, -0.00001093, -0.00001348, -0.00001488, -0.00001511, -0.00001432, -0.00001271, -0.00001059, -0.00000826, -0.00000601, -0.00000409, -0.00000265, -0.00000176, -0.00000141, -0.00000148, -0.00000183, -0.00000228, -0.00000264, -0.00000277, -0.00000258, -0.00000204, -0.00000120, },
		{ -0.00000565, -0.00000599, -0.00000619, -0.00000614, -0.00000579, -0.00000510, -0.00000412, -0.00000292, -0.00000163, -0.00000041, 0.00000061, 0.00000130, 0.00000158, 0.00000147, 0.00000101, 0.00000035, -0.00000032, -0.00000078, -0.00000082, -0.00000026, 0.00000103, 0.00000307, 0.00000576, 0.00000894, 0.00001229, 0.00001548, 0.00001812, 0.00001985, 0.00002038, 0.00001952, 0.00001724, 0.00001368, 0.00000913, 0.00000403, -0.00000105, -0.00000552, -0.00000882, -0.00001043, -0.00001004, -0.00000752, -0.00000299, 0.00000316, 0.00001033, 0.00001773, 0.00002442, 0.00002948, 0.00003203, 0.00003138, 0.00002709, 0.00001904, 0.00000749, -0.00000697, -0.00002338, -0.00004055, -0.00005711, -0.00007167, -0.00008291, -0.00008972, -0.00009134, -0.00008740, -0.00007798, -0.00006363, -0.00004533, -0.00002439, -0.00000238, 0.00001906, 0.00003831, 0.00005393, 0.00006481, 0.00007022, 0.00006991, 0.00006410, 0.00005347, 0.00003908, 0.00002225, 0.00000447, -0.00001276, -0.00002807, -0.00004032, -0.00004870, -0.00005281, -0.00005265, -0.00004859, -0.00004135, -0.00003189, -0.00002130, -0.00001071, -0.00000115, 0.00000654, 0.00001177, 0.00001423, 0.00001394, 0.00001119, 0.00000652, 0.00000063, -0.00000570, -0.00001171, -0.00001672, -0.00002020, -0.00002182, -0.00002147, -0.00001927, -0.00001551, -0.00001065, -0.00000521, 0.00000024, 0.00000521, 0.00000926, 0.00001210, 0.00001358, 0.00001372, 0.00001265, 0.00001063, 0.00000797, 0.00000502, 0.00000210, -0.00000053, -0.00000266, -0.00000421, -0.00000516, -0.00000559, -0.00000563, -0.00000543, -0.00000517, -0.00000496, -0.00000490, -0.00000502, -0.00000529, },
		{ -0.00000210, -0.00000282, -0.00000356, -0.00000439, -0.00000536, -0.00000646, -0.00000767, -0.00000889, -0.00001000, -0.00001087, -0.00001137, -0.00001141, -0.00001093, -0.00000994, -0.00000851, -0.00000679, -0.00000495, -0.00000320, -0.00000176, -0.00000080, -0.00000043, -0.00000068, -0.00000150, -0.00000272, -0.00000412, -0.00000540, -0.00000626, -0.00000642, -0.00000565, -0.00000384, -0.00000099, 0.00000275, 0.00000709, 0.00001162, 0.00001587, 0.00001935, 0.00002157, 0.00002216, 0.00002091, 0.00001778, 0.00001295, 0.00000682, -0.00000003, -0.00000687, -0.00001290, -0.00001732, -0.00001944, -0.00001873, -0.00001491, -0.00000801, 0.00000165, 0.00001343, 0.00002643, 0.00003956, 0.00005162, 0.00006145, 0.00006800, 0.00007044, 0.00006827, 0.00006136, 0.00005000, 0.00003485, 0.00001691, -0.00000254, -0.00002210, -0.00004031, -0.00005585, -0.00006759, -0.00007475, -0.00007691, -0.00007409, -0.00006672, -0.00005561, -0.00004185, -0.00002672, -0.00001156, 0.00000235, 0.00001390, 0.00002226, 0.00002693, 0.00002779, 0.00002507, 0.00001935, 0.00001144, 0.00000233, -0.00000695, -0.00001540, -0.00002217, -0.00002665, -0.00002849, -0.00002764, -0.00002434, -0.00001908, -0.00001252, -0.00000543, 0.00000138, 0.00000720, 0.00001143, 0.00001368, 0.00001378, 0.00001179, 0.00000798, 0.00000280, -0.00000318, -0.00000935, -0.00001511, -0.00001992, -0.00002339, -0.00002527, -0.00002550, -0.00002416, -0.00002149, -0.00001784, -0.00001361, -0.00000922, -0.00000504, -0.00000140, 0.00000148, 0.00000349, 0.00000462, 0.00000495, 0.00000464, 0.00000386, 0.00000280, 0.00000165, 0.00000054, -0.00000047, -0.00000134, },
		{ 0.00000181, 0.00000281, 0.00000348, 0.00000393, 0.00000422, 0.00000439, 0.00000445, 0.00000437, 0.00000409, 0.00000353, 0.00000261, 0.00000130, -0.00000043, -0.00000253, -0.00000490, -0.00000740, -0.00000984, -0.00001202, -0.00001376, -0.00001489, -0.00001531, -0.00001501, -0.00001406, -0.00001259, -0.00001083, -0.00000906, -0.00000755, -0.00000655, -0.00000627, -0.00000682, -0.00000817, -0.00001020, -0.00001266, -0.00001519, -0.00001739, -0.00001883, -0.00001912, -0.00001796, -0.00001520, -0.00001084, -0.00000508, 0.00000170, 0.00000897, 0.00001607, 0.00002231, 0.00002702, 0.00002960, 0.00002966, 0.00002699, 0.00002165, 0.00001400, 0.00000460, -0.00000573, -0.00001605, -0.00002537, -0.00003271, -0.00003724, -0.00003835, -0.00003569, -0.00002927, -0.00001941, -0.00000678, 0.00000772, 0.00002299, 0.00003782, 0.00005106, 0.00006164, 0.00006876, 0.00007187, 0.00007080, 0.00006572, 0.00005714, 0.00004585, 0.00003285, 0.00001924, 0.00000614, -0.00000546, -0.00001474, -0.00002116, -0.00002447, -0.00002475, -0.00002236, -0.00001792, -0.00001221, -0.00000610, -0.00000043, 0.00000403, 0.00000671, 0.00000726, 0.00000560, 0.00000189, -0.00000346, -0.00000987, -0.00001664, -0.00002306, -0.00002843, -0.00003221, -0.00003401, -0.00003364, -0.00003117, -0.00002686, -0.00002114, -0.00001458, -0.00000782, -0.00000149, 0.00000385, 0.00000776, 0.00000995, 0.00001034, 0.00000899, 0.00000616, 0.00000221, -0.00000239, -0.00000717, -0.00001164, -0.00001543, -0.00001822, -0.00001984, -0.00002024, -0.00001950, -0.00001779, -0.00001536, -0.00001247, -0.00000942, -0.00000644, -0.00000374, -0.00000143, 0.00000041, },
		{ -0.00002095, -0.00001963, -0.00001786, -0.00001580, -0.00001362, -0.00001145, -0.00000937, -0.00000742, -0.00000558, -0.00000382, -0.00000208, -0.00000033, 0.00000146, 0.00000327, 0.00000504, 0.00000667, 0.00000803, 0.00000898, 0.00000939, 0.00000917, 0.00000828, 0.00000673, 0.00000462, 0.00000212, -0.00000053, -0.00000309, -0.00000528, -0.00000687, -0.00000768, -0.00000764, -0.00000676, -0.00000519, -0.00000316, -0.00000101, 0.00000089, 0.00000214, 0.00000239, 0.00000140, -0.00000099, -0.00000475, -0.00000969, -0.00001545, -0.00002155, -0.00002741, -0.00003240, -0.00003592, -0.00003748, -0.00003671, -0.00003346, -0.00002781, -0.00002006, -0.00001074, -0.00000055, 0.00000968, 0.00001909, 0.00002684, 0.00003224, 0.00003477, 0.00003416, 0.00003042, 0.00002385, 0.00001501, 0.00000465, -0.00000630, -0.00001687, -0.00002614, -0.00003329, -0.00003768, -0.00003896, -0.00003704, -0.00003212, -0.00002467, -0.00001537, -0.00000504, 0.00000544, 0.00001521, 0.00002354, 0.00002985, 0.00003382, 0.00003534, 0.00003459, 0.00003194, 0.00002794, 0.00002325, 0.00001854, 0.00001444, 0.00001146, 0.00000993, 0.00000999, 0.00001157, 0.00001438, 0.00001799, 0.00002186, 0.00002540, 0.00002804, 0.00002933, 0.00002892, 0.00002667, 0.00002263, 0.00001704, 0.00001029, 0.00000290, -0.00000453, -0.00001143, -0.00001726, -0.00002161, -0.00002421, -0.00002497, -0.00002398, -0.00002149, -0.00001787, -0.00001359, -0.00000916, -0.00000507, -0.00000173, 0.00000054, 0.00000155, 0.00000127, -0.00000022, -0.00000270, -0.00000588, -0.00000941, -0.00001293, -0.00001613, -0.00001873, -0.00002055, -0.00002152, -0.00002162, },
		{ 0.00000091, -0.00000244, -0.00000585, -0.00000907, -0.00001187, -0.00001413, -0.00001575, -0.00001675, -0.00001718, -0.00001715, -0.00001679, -0.00001622, -0.00001556, -0.00001489, -0.00001422, -0.00001355, -0.00001283, -0.00001197, -0.00001090, -0.00000955, -0.00000791, -0.00000598, -0.00000385, -0.00000166, 0.00000042, 0.00000218, 0.00000343, 0.00000400, 0.00000377, 0.00000273, 0.00000094, -0.00000144, -0.00000413, -0.00000682, -0.00000914, -0.00001072, -0.00001126, -0.00001054, -0.00000844, -0.00000501, -0.00000046, 0.00000487, 0.00001051, 0.00001590, 0.00002046, 0.00002365, 0.00002499, 0.00002418, 0.00002107, 0.00001574, 0.00000847, -0.00000025, -0.00000978, -0.00001937, -0.00002824, -0.00003564, -0.00004093, -0.00004365, -0.00004353, -0.00004057, -0.00003501, -0.00002730, -0.00001811, -0.00000818, 0.00000165, 0.00001059, 0.00001796, 0.00002321, 0.00002603, 0.00002633, 0.00002424, 0.00002013, 0.00001451, 0.00000801, 0.00000131, -0.00000496, -0.00001024, -0.00001412, -0.00001638, -0.00001697, -0.00001605, -0.00001391, -0.00001098, -0.00000773, -0.00000463, -0.00000211, -0.00000045, 0.00000015, -0.00000029, -0.00000164, -0.00000362, -0.00000584, -0.00000788, -0.00000930, -0.00000976, -0.00000898, -0.00000683, -0.00000335, 0.00000128, 0.00000675, 0.00001261, 0.00001839, 0.00002359, 0.00002777, 0.00003056, 0.00003176, 0.00003127, 0.00002919, 0.00002575, 0.00002129, 0.00001627, 0.00001113, 0.00000635, 0.00000231, -0.00000070, -0.00000249, -0.00000304, -0.00000246, -0.00000096, 0.00000115, 0.00000351, 0.00000575, 0.00000755, 0.00000864, 0.00000884, 0.00000808, 0.00000641, 0.00000395, },
		{ 0.00000781, 0.00000996, 0.00001169, 0.00001279, 0.00001313, 0.00001266, 0.00001140, 0.00000948, 0.00000709, 0.00000444, 0.00000176, -0.00000073, -0.00000286, -0.00000451, -0.00000564, -0.00000626, -0.00000645, -0.00000633, -0.00000603, -0.00000569, -0.00000540, -0.00000523, -0.00000519, -0.00000523, -0.00000526, -0.00000515, -0.00000479, -0.00000405, -0.00000287, -0.00000125, 0.00000075, 0.00000296, 0.00000517, 0.00000710, 0.00000846, 0.00000896, 0.00000836, 0.00000653, 0.00000341, -0.00000088, -0.00000612, -0.00001194, -0.00001786, -0.00002337, -0.00002791, -0.00003099, -0.00003219, -0.00003124, -0.00002804, -0.00002269, -0.00001549, -0.00000691, 0.00000243, 0.00001181, 0.00002051, 0.00002782, 0.00003315, 0.00003605, 0.00003627, 0.00003381, 0.00002887, 0.00002186, 0.00001334, 0.00000402, -0.00000536, -0.00001410, -0.00002154, -0.00002718, -0.00003071, -0.00003198, -0.00003109, -0.00002827, -0.00002393, -0.00001857, -0.00001273, -0.00000695, -0.00000168, 0.00000271, 0.00000602, 0.00000815, 0.00000917, 0.00000926, 0.00000870, 0.00000777, 0.00000680, 0.00000604, 0.00000567, 0.00000578, 0.00000634, 0.00000723, 0.00000823, 0.00000908, 0.00000951, 0.00000928, 0.00000820, 0.00000617, 0.00000322, -0.00000052, -0.00000480, -0.00000932, -0.00001368, -0.00001750, -0.00002044, -0.00002220, -0.00002261, -0.00002160, -0.00001926, -0.00001578, -0.00001145, -0.00000667, -0.00000184, 0.00000265, 0.00000644, 0.00000926, 0.00001097, 0.00001152, 0.00001101, 0.00000964, 0.00000768, 0.00000545, 0.00000329, 0.00000150, 0.00000033, -0.00000007, 0.00000034, 0.00000152, 0.00000332, 0.00000550, },
		{ -0.00000453, -0.00000552, -0.00000599, -0.00000583, -0.00000501, -0.00000359, -0.00000170, 0.00000046, 0.00000266, 0.00000463, 0.00000616, 0.00000705, 0.00000720, 0.00000657, 0.00000521, 0.00000324, 0.00000085, -0.00000174, -0.00000429, -0.00000658, -0.00000844, -0.00000976, -0.00001050, -0.00001069, -0.00001041, -0.00000982, -0.00000908, -0.00000833, -0.00000772, -0.00000731, -0.00000712, -0.00000708, -0.00000705, -0.00000685, -0.00000628, -0.00000512, -0.00000321, -0.00000045, 0.00000315, 0.00000750, 0.00001237, 0.00001743, 0.00002227, 0.00002645, 0.00002950, 0.00003101, 0.00003067, 0.00002828, 0.00002383, 0.00001746, 0.00000950, 0.00000042, -0.00000916, -0.00001858, -0.00002715, -0.00003423, -0.00003928, -0.00004191, -0.00004193, -0.00003934, -0.00003435, -0.00002736, -0.00001892, -0.00000969, -0.00000037, 0.00000838, 0.00001594, 0.00002184, 0.00002575, 0.00002754, 0.00002723, 0.00002505, 0.00002132, 0.00001649, 0.00001104, 0.00000546, 0.00000017, -0.00000445, -0.00000818, -0.00001088, -0.00001254, -0.00001325, -0.00001318, -0.00001254, -0.00001155, -0.00001042, -0.00000929, -0.00000826, -0.00000734, -0.00000650, -0.00000563, -0.00000461, -0.00000331, -0.00000164, 0.00000048, 0.00000301, 0.00000590, 0.00000897, 0.00001203, 0.00001481, 0.00001707, 0.00001855, 0.00001907, 0.00001850, 0.00001683, 0.00001412, 0.00001056, 0.00000639, 0.00000192, -0.00000248, -0.00000650, -0.00000983, -0.00001225, -0.00001361, -0.00001389, -0.00001317, -0.00001160, -0.00000944, -0.00000697, -0.00000450, -0.00000230, -0.00000060, 0.00000045, 0.00000078, 0.00000042, -0.00000049, -0.00000178, -0.00000321, },
		{ 0.00000037, 0.00000144, 0.00000204, 0.00000213, 0.00000172, 0.00000090, -0.00000017, -0.00000129, -0.00000226, -0.00000288, -0.00000301, -0.00000255, -0.00000148, 0.00000013, 0.00000215, 0.00000438, 0.00000659, 0.00000851, 0.00000993, 0.00001065, 0.00001056, 0.00000962, 0.00000787, 0.00000545, 0.00000256, -0.00000057, -0.00000367, -0.00000651, -0.00000889, -0.00001067, -0.00001180, -0.00001233, -0.00001238, -0.00001212, -0.00001178, -0.00001159, -0.00001176, -0.00001244, -0.00001369, -0.00001548, -0.00001767, -0.00002002, -0.00002222, -0.00002393, -0.00002476, -0.00002441, -0.00002262, -0.00001925, -0.00001430, -0.00000793, -0.00000043, 0.00000776, 0.00001611, 0.00002406, 0.00003100, 0.00003640, 0.00003983, 0.00004098, 0.00003973, 0.00003613, 0.00003042, 0.00002300, 0.00001439, 0.00000521, -0.00000391, -0.00001234, -0.00001954, -0.00002508, -0.00002869, -0.00003024, -0.00002979, -0.00002754, -0.00002380, -0.00001899, -0.00001357, -0.00000798, -0.00000266, 0.00000207, 0.00000595, 0.00000883, 0.00001068, 0.00001156, 0.00001159, 0.00001095, 0.00000985, 0.00000847, 0.00000698, 0.00000550, 0.00000411, 0.00000283, 0.00000164, 0.00000050, -0.00000065, -0.00000184, -0.00000310, -0.00000440, -0.00000569, -0.00000688, -0.00000783, -0.00000842, -0.00000851, -0.00000801, -0.00000686, -0.00000505, -0.00000265, 0.00000020, 0.00000332, 0.00000647, 0.00000940, 0.00001188, 0.00001367, 0.00001463, 0.00001467, 0.00001378, 0.00001205, 0.00000964, 0.00000676, 0.00000369, 0.00000069, -0.00000197, -0.00000409, -0.00000552, -0.00000618, -0.00000609, -0.00000536, -0.00000413, -0.00000262, -0.00000104, },
		{ -0.00000437, -0.00000580, -0.00000661, -0.00000680, -0.00000641, -0.00000557, -0.00000444, -0.00000321, -0.00000205, -0.00000113, -0.00000056, -0.00000040, -0.00000063, -0.00000116, -0.00000188, -0.00000259, -0.00000312, -0.00000328, -0.00000293, -0.00000197, -0.00000040, 0.00000173, 0.00000429, 0.00000705, 0.00000979, 0.00001224, 0.00001414, 0.00001529, 0.00001553, 0.00001482, 0.00001320, 0.00001080, 0.00000784, 0.00000461, 0.00000142, -0.00000142, -0.00000364, -0.00000502, -0.00000547, -0.00000497, -0.00000365, -0.00000172, 0.00000051, 0.00000267, 0.00000441, 0.00000536, 0.00000525, 0.00000390, 0.00000125, -0.00000261, -0.00000748, -0.00001301, -0.00001877, -0.00002427, -0.00002904, -0.00003261, -0.00003462, -0.00003481, -0.00003309, -0.00002949, -0.00002422, -0.00001763, -0.00001016, -0.00000234, 0.00000531, 0.00001225, 0.00001804, 0.00002234, 0.00002493, 0.00002574, 0.00002483, 0.00002239, 0.00001873, 0.00001422, 0.00000926, 0.00000425, -0.00000044, -0.00000451, -0.00000776, -0.00001005, -0.00001137, -0.00001177, -0.00001139, -0.00001038, -0.00000894, -0.00000726, -0.00000550, -0.00000380, -0.00000225, -0.00000089, 0.00000028, 0.00000127, 0.00000212, 0.00000287, 0.00000354, 0.00000415, 0.00000469, 0.00000511, 0.00000538, 0.00000543, 0.00000521, 0.00000469, 0.00000387, 0.00000277, 0.00000147, 0.00000007, -0.00000130, -0.00000249, -0.00000336, -0.00000379, -0.00000367, -0.00000299, -0.00000174, -0.00000001, 0.00000206, 0.00000430, 0.00000650, 0.00000845, 0.00000996, 0.00001086, 0.00001106, 0.00001051, 0.00000926, 0.00000739, 0.00000508, 0.00000253, -0.00000004, -0.00000240, },
		{ 0.00001323, 0.00001288, 0.00001185, 0.00001020, 0.00000805, 0.00000557, 0.00000296, 0.00000042, -0.00000187, -0.00000376, -0.00000516, -0.00000601, -0.00000635, -0.00000625, -0.00000583, -0.00000526, -0.00000468, -0.00000424, -0.00000404, -0.00000413, -0.00000452, -0.00000514, -0.00000585, -0.00000650, -0.00000690, -0.00000688, -0.00000627, -0.00000499, -0.00000301, -0.00000039, 0.00000275, 0.00000616, 0.00000959, 0.00001272, 0.00001523, 0.00001684, 0.00001733, 0.00001658, 0.00001457, 0.00001139, 0.00000726, 0.00000249, -0.00000251, -0.00000732, -0.00001151, -0.00001467, -0.00001650, -0.00001678, -0.00001545, -0.00001258, -0.00000838, -0.00000317, 0.00000260, 0.00000846, 0.00001391, 0.00001849, 0.00002180, 0.00002356, 0.00002362, 0.00002196, 0.00001873, 0.00001417, 0.00000865, 0.00000261, -0.00000351, -0.00000925, -0.00001423, -0.00001813, -0.00002075, -0.00002199, -0.00002189, -0.00002057, -0.00001827, -0.00001526, -0.00001185, -0.00000836, -0.00000507, -0.00000220, 0.00000007, 0.00000167, 0.00000257, 0.00000282, 0.00000254, 0.00000185, 0.00000092, -0.00000013, -0.00000114, -0.00000203, -0.00000274, -0.00000324, -0.00000353, -0.00000365, -0.00000364, -0.00000355, -0.00000342, -0.00000328, -0.00000313, -0.00000296, -0.00000275, -0.00000248, -0.00000210, -0.00000160, -0.00000098, -0.00000024, 0.00000057, 0.00000139, 0.00000217, 0.00000281, 0.00000325, 0.00000343, 0.00000333, 0.00000296, 0.00000234, 0.00000157, 0.00000075, -0.00000001, -0.00000056, -0.00000079, -0.00000062, 0.00000002, 0.00000113, 0.00000266, 0.00000451, 0.00000654, 0.00000857, 0.00001042, 0.00001191, 0.00001289, },
		{ -0.00000122, 0.00000036, 0.00000233, 0.00000456, 0.00000689, 0.00000917, 0.00001121, 0.00001284, 0.00001394, 0.00001442, 0.00001423, 0.00001340, 0.00001201, 0.00001017, 0.00000805, 0.00000582, 0.00000364, 0.00000167, 0.00000002, -0.00000126, -0.00000215, -0.00000270, -0.00000301, -0.00000318, -0.00000336, -0.00000367, -0.00000421, -0.00000502, -0.00000611, -0.00000740, -0.00000876, -0.00001002, -0.00001097, -0.00001138, -0.00001106, -0.00000985, -0.00000766, -0.00000450, -0.00000047, 0.00000424, 0.00000935, 0.00001448, 0.00001925, 0.00002324, 0.00002608, 0.00002746, 0.00002716, 0.00002507, 0.00002125, 0.00001587, 0.00000924, 0.00000176, -0.00000607, -0.00001373, -0.00002070, -0.00002650, -0.00003074, -0.00003315, -0.00003359, -0.00003205, -0.00002870, -0.00002378, -0.00001768, -0.00001083, -0.00000370, 0.00000327, 0.00000964, 0.00001508, 0.00001935, 0.00002228, 0.00002386, 0.00002414, 0.00002327, 0.00002146, 0.00001896, 0.00001602, 0.00001291, 0.00000983, 0.00000695, 0.00000440, 0.00000224, 0.00000047, -0.00000092, -0.00000200, -0.00000283, -0.00000349, -0.00000404, -0.00000453, -0.00000497, -0.00000539, -0.00000576, -0.00000606, -0.00000627, -0.00000636, -0.00000632, -0.00000615, -0.00000587, -0.00000549, -0.00000508, -0.00000466, -0.00000429, -0.00000399, -0.00000380, -0.00000370, -0.00000370, -0.00000374, -0.00000381, -0.00000383, -0.00000378, -0.00000361, -0.00000330, -0.00000286, -0.00000231, -0.00000171, -0.00000110, -0.00000057, -0.00000017, 0.00000004, 0.00000001, -0.00000025, -0.00000072, -0.00000133, -0.00000201, -0.00000263, -0.00000308, -0.00000325, -0.00000303, -0.00000236, },
		{ -0.00000109, -0.00000195, -0.00000292, -0.00000391, -0.00000482, -0.00000551, -0.00000588, -0.00000585, -0.00000536, -0.00000438, -0.00000296, -0.00000117, 0.00000089, 0.00000307, 0.00000522, 0.00000717, 0.00000880, 0.00000998, 0.00001065, 0.00001080, 0.00001044, 0.00000965, 0.00000855, 0.00000727, 0.00000595, 0.00000473, 0.00000370, 0.00000293, 0.00000243, 0.00000218, 0.00000209, 0.00000204, 0.00000189, 0.00000150, 0.00000073, -0.00000050, -0.00000222, -0.00000441, -0.00000696, -0.00000970, -0.00001241, -0.00001481, -0.00001663, -0.00001759, -0.00001747, -0.00001610, -0.00001342, -0.00000944, -0.00000432, 0.00000171, 0.00000832, 0.00001510, 0.00002161, 0.00002740, 0.00003206, 0.00003524, 0.00003667, 0.00003621, 0.00003384, 0.00002967, 0.00002393, 0.00001696, 0.00000915, 0.00000096, -0.00000713, -0.00001469, -0.00002132, -0.00002672, -0.00003065, -0.00003300, -0.00003376, -0.00003301, -0.00003091, -0.00002769, -0.00002363, -0.00001899, -0.00001407, -0.00000912, -0.00000436, 0.00000004, 0.00000395, 0.00000729, 0.00001002, 0.00001215, 0.00001369, 0.00001470, 0.00001522, 0.00001530, 0.00001499, 0.00001434, 0.00001339, 0.00001219, 0.00001077, 0.00000919, 0.00000750, 0.00000575, 0.00000402, 0.00000237, 0.00000087, -0.00000045, -0.00000152, -0.00000232, -0.00000285, -0.00000313, -0.00000317, -0.00000304, -0.00000279, -0.00000248, -0.00000218, -0.00000192, -0.00000176, -0.00000169, -0.00000171, -0.00000181, -0.00000194, -0.00000206, -0.00000213, -0.00000211, -0.00000197, -0.00000171, -0.00000136, -0.00000094, -0.00000051, -0.00000014, 0.00000010, 0.00000016, -0.00000002, -0.00000044, },
		{ 0.00000045, 0.00000107, 0.00000175, 0.00000241, 0.00000299, 0.00000342, 0.00000363, 0.00000360, 0.00000330, 0.00000273, 0.00000196, 0.00000103, 0.00000004, -0.00000091, -0.00000172, -0.00000229, -0.00000254, -0.00000243, -0.00000195, -0.00000114, -0.00000004, 0.00000124, 0.00000258, 0.00000386, 0.00000495, 0.00000574, 0.00000617, 0.00000620, 0.00000582, 0.00000510, 0.00000413, 0.00000302, 0.00000192, 0.00000096, 0.00000027, -0.00000006, 0.00000003, 0.00000054, 0.00000141, 0.00000254, 0.00000378, 0.00000495, 0.00000586, 0.00000634, 0.00000623, 0.00000544, 0.00000391, 0.00000167, -0.00000116, -0.00000442, -0.00000788, -0.00001127, -0.00001431, -0.00001672, -0.00001824, -0.00001869, -0.00001794, -0.00001596, -0.00001278, -0.00000855, -0.00000350, 0.00000211, 0.00000794, 0.00001365, 0.00001890, 0.00002340, 0.00002687, 0.00002914, 0.00003009, 0.00002971, 0.00002804, 0.00002522, 0.00002144, 0.00001692, 0.00001193, 0.00000674, 0.00000158, -0.00000329, -0.00000769, -0.00001147, -0.00001452, -0.00001678, -0.00001823, -0.00001889, -0.00001882, -0.00001808, -0.00001677, -0.00001499, -0.00001285, -0.00001045, -0.00000789, -0.00000529, -0.00000273, -0.00000032, 0.00000188, 0.00000377, 0.00000531, 0.00000645, 0.00000715, 0.00000741, 0.00000725, 0.00000670, 0.00000582, 0.00000469, 0.00000338, 0.00000199, 0.00000063, -0.00000064, -0.00000173, -0.00000259, -0.00000319, -0.00000353, -0.00000363, -0.00000352, -0.00000324, -0.00000288, -0.00000247, -0.00000208, -0.00000174, -0.00000148, -0.00000131, -0.00000121, -0.00000115, -0.00000109, -0.00000099, -0.00000080, -0.00000051, -0.00000009, },
		{ -0.00000674, -0.00000673, -0.00000677, -0.00000682, -0.00000684, -0.00000680, -0.00000665, -0.00000635, -0.00000590, -0.00000530, -0.00000456, -0.00000373, -0.00000285, -0.00000199, -0.00000119, -0.00000052, 0.00000000, 0.00000035, 0.00000052, 0.00000055, 0.00000049, 0.00000040, 0.00000036, 0.00000045, 0.00000072, 0.00000122, 0.00000199, 0.00000300, 0.00000422, 0.00000558, 0.00000697, 0.00000828, 0.00000939, 0.00001018, 0.00001054, 0.00001041, 0.00000974, 0.00000856, 0.00000691, 0.00000489, 0.00000264, 0.00000032, -0.00000190, -0.00000386, -0.00000539, -0.00000637, -0.00000673, -0.00000645, -0.00000554, -0.00000408, -0.00000221, -0.00000008, 0.00000213, 0.00000423, 0.00000604, 0.00000739, 0.00000818, 0.00000834, 0.00000784, 0.00000672, 0.00000506, 0.00000299, 0.00000068, -0.00000171, -0.00000400, -0.00000601, -0.00000760, -0.00000865, -0.00000909, -0.00000890, -0.00000808, -0.00000669, -0.00000483, -0.00000260, -0.00000016, 0.00000237, 0.00000484, 0.00000712, 0.00000910, 0.00001068, 0.00001181, 0.00001245, 0.00001259, 0.00001225, 0.00001146, 0.00001028, 0.00000879, 0.00000707, 0.00000522, 0.00000332, 0.00000148, -0.00000022, -0.00000169, -0.00000287, -0.00000369, -0.00000413, -0.00000416, -0.00000379, -0.00000306, -0.00000201, -0.00000072, 0.00000073, 0.00000223, 0.00000368, 0.00000498, 0.00000604, 0.00000678, 0.00000714, 0.00000709, 0.00000664, 0.00000580, 0.00000464, 0.00000322, 0.00000163, -0.00000003, -0.00000166, -0.00000317, -0.00000451, -0.00000561, -0.00000644, -0.00000701, -0.00000734, -0.00000746, -0.00000742, -0.00000728, -0.00000710, -0.00000693, -0.00000680, },
		{ -0.00000090, -0.00000179, -0.00000246, -0.00000294, -0.00000328, -0.00000353, -0.00000374, -0.00000398, -0.00000426, -0.00000463, -0.00000508, -0.00000562, -0.00000621, -0.00000684, -0.00000746, -0.00000804, -0.00000855, -0.00000897, -0.00000929, -0.00000950, -0.00000963, -0.00000969, -0.00000972, -0.00000975, -0.00000981, -0.00000990, -0.00001004, -0.00001020, -0.00001036, -0.00001047, -0.00001046, -0.00001027, -0.00000982, -0.00000906, -0.00000794, -0.00000643, -0.00000454, -0.00000230, 0.00000022, 0.00000292, 0.00000569, 0.00000837, 0.00001082, 0.00001289, 0.00001444, 0.00001537, 0.00001559, 0.00001506, 0.00001380, 0.00001185, 0.00000931, 0.00000631, 0.00000301, -0.00000039, -0.00000371, -0.00000677, -0.00000937, -0.00001139, -0.00001271, -0.00001326, -0.00001302, -0.00001201, -0.00001030, -0.00000799, -0.00000523, -0.00000216, 0.00000103, 0.00000418, 0.00000711, 0.00000969, 0.00001179, 0.00001332, 0.00001423, 0.00001448, 0.00001412, 0.00001317, 0.00001174, 0.00000991, 0.00000782, 0.00000559, 0.00000335, 0.00000123, -0.00000068, -0.00000227, -0.00000349, -0.00000431, -0.00000472, -0.00000475, -0.00000444, -0.00000386, -0.00000310, -0.00000224, -0.00000137, -0.00000058, 0.00000006, 0.00000050, 0.00000073, 0.00000074, 0.00000056, 0.00000023, -0.00000018, -0.00000058, -0.00000090, -0.00000106, -0.00000098, -0.00000061, 0.00000008, 0.00000108, 0.00000236, 0.00000389, 0.00000557, 0.00000731, 0.00000901, 0.00001055, 0.00001185, 0.00001281, 0.00001337, 0.00001350, 0.00001318, 0.00001244, 0.00001133, 0.00000992, 0.00000830, 0.00000656, 0.00000480, 0.00000311, 0.00000157, 0.00000022, },
		{ 0.00001278, 0.00001226, 0.00001141, 0.00001032, 0.00000906, 0.00000773, 0.00000642, 0.00000520, 0.00000414, 0.00000328, 0.00000264, 0.00000223, 0.00000202, 0.00000198, 0.00000208, 0.00000225, 0.00000245, 0.00000264, 0.00000277, 0.00000281, 0.00000275, 0.00000257, 0.00000230, 0.00000192, 0.00000147, 0.00000095, 0.00000038, -0.00000024, -0.00000089, -0.00000159, -0.00000235, -0.00000319, -0.00000414, -0.00000520, -0.00000640, -0.00000774, -0.00000921, -0.00001080, -0.00001245, -0.00001411, -0.00001571, -0.00001715, -0.00001833, -0.00001916, -0.00001954, -0.00001940, -0.00001866, -0.00001729, -0.00001531, -0.00001273, -0.00000965, -0.00000616, -0.00000242, 0.00000141, 0.00000513, 0.00000855, 0.00001148, 0.00001375, 0.00001520, 0.00001574, 0.00001531, 0.00001391, 0.00001160, 0.00000848, 0.00000473, 0.00000054, -0.00000383, -0.00000815, -0.00001215, -0.00001560, -0.00001827, -0.00002001, -0.00002071, -0.00002030, -0.00001882, -0.00001633, -0.00001298, -0.00000897, -0.00000451, 0.00000013, 0.00000471, 0.00000899, 0.00001274, 0.00001579, 0.00001801, 0.00001934, 0.00001976, 0.00001932, 0.00001812, 0.00001630, 0.00001402, 0.00001147, 0.00000882, 0.00000625, 0.00000389, 0.00000186, 0.00000022, -0.00000100, -0.00000181, -0.00000226, -0.00000242, -0.00000240, -0.00000227, -0.00000214, -0.00000208, -0.00000215, -0.00000237, -0.00000276, -0.00000326, -0.00000384, -0.00000440, -0.00000488, -0.00000519, -0.00000523, -0.00000496, -0.00000433, -0.00000334, -0.00000199, -0.00000034, 0.00000154, 0.00000355, 0.00000559, 0.00000755, 0.00000932, 0.00001080, 0.00001192, 0.00001263, 0.00001292, },
		{ -0.00000049, 0.00000117, 0.00000295, 0.00000473, 0.00000641, 0.00000791, 0.00000912, 0.00001000, 0.00001050, 0.00001060, 0.00001032, 0.00000970, 0.00000878, 0.00000765, 0.00000637, 0.00000504, 0.00000373, 0.00000251, 0.00000145, 0.00000059, -0.00000003, -0.00000041, -0.00000055, -0.00000048, -0.00000022, 0.00000019, 0.00000070, 0.00000126, 0.00000185, 0.00000242, 0.00000295, 0.00000344, 0.00000388, 0.00000427, 0.00000464, 0.00000500, 0.00000537, 0.00000576, 0.00000617, 0.00000660, 0.00000703, 0.00000741, 0.00000769, 0.00000780, 0.00000767, 0.00000722, 0.00000639, 0.00000513, 0.00000339, 0.00000119, -0.00000146, -0.00000448, -0.00000777, -0.00001118, -0.00001454, -0.00001768, -0.00002038, -0.00002247, -0.00002379, -0.00002420, -0.00002364, -0.00002208, -0.00001957, -0.00001623, -0.00001222, -0.00000778, -0.00000318, 0.00000131, 0.00000539, 0.00000878, 0.00001126, 0.00001265, 0.00001286, 0.00001185, 0.00000970, 0.00000654, 0.00000258, -0.00000190, -0.00000659, -0.00001117, -0.00001534, -0.00001881, -0.00002134, -0.00002279, -0.00002306, -0.00002215, -0.00002014, -0.00001719, -0.00001351, -0.00000934, -0.00000496, -0.00000064, 0.00000336, 0.00000684, 0.00000963, 0.00001163, 0.00001282, 0.00001321, 0.00001290, 0.00001201, 0.00001069, 0.00000912, 0.00000747, 0.00000588, 0.00000449, 0.00000336, 0.00000256, 0.00000208, 0.00000188, 0.00000190, 0.00000205, 0.00000224, 0.00000238, 0.00000238, 0.00000219, 0.00000178, 0.00000115, 0.00000032, -0.00000063, -0.00000164, -0.00000262, -0.00000345, -0.00000406, -0.00000437, -0.00000433, -0.00000390, -0.00000309, -0.00000194, },
		{ -0.00000047, -0.00000146, -0.00000238, -0.00000317, -0.00000372, -0.00000397, -0.00000388, -0.00000343, -0.00000263, -0.00000151, -0.00000012, 0.00000146, 0.00000314, 0.00000485, 0.00000647, 0.00000793, 0.00000915, 0.00001008, 0.00001066, 0.00001089, 0.00001076, 0.00001030, 0.00000954, 0.00000855, 0.00000738, 0.00000611, 0.00000482, 0.00000358, 0.00000245, 0.00000148, 0.00000071, 0.00000013, -0.00000024, -0.00000044, -0.00000051, -0.00000050, -0.00000047, -0.00000047, -0.00000055, -0.00000074, -0.00000106, -0.00000150, -0.00000201, -0.00000255, -0.00000302, -0.00000333, -0.00000339, -0.00000310, -0.00000239, -0.00000120, 0.00000048, 0.00000261, 0.00000512, 0.00000788, 0.00001073, 0.00001347, 0.00001589, 0.00001778, 0.00001895, 0.00001923, 0.00001852, 0.00001677, 0.00001401, 0.00001035, 0.00000597, 0.00000110, -0.00000396, -0.00000892, -0.00001345, -0.00001727, -0.00002013, -0.00002185, -0.00002233, -0.00002157, -0.00001965, -0.00001675, -0.00001310, -0.00000902, -0.00000483, -0.00000087, 0.00000255, 0.00000517, 0.00000679, 0.00000729, 0.00000667, 0.00000499, 0.00000240, -0.00000086, -0.00000453, -0.00000830, -0.00001187, -0.00001497, -0.00001737, -0.00001890, -0.00001948, -0.00001910, -0.00001781, -0.00001576, -0.00001313, -0.00001014, -0.00000701, -0.00000399, -0.00000127, 0.00000099, 0.00000268, 0.00000375, 0.00000421, 0.00000412, 0.00000358, 0.00000272, 0.00000170, 0.00000065, -0.00000029, -0.00000102, -0.00000147, -0.00000160, -0.00000143, -0.00000100, -0.00000037, 0.00000035, 0.00000107, 0.00000169, 0.00000214, 0.00000234, 0.00000226, 0.00000191, 0.00000129, 0.00000048, },
		{ -0.00000148, -0.00000052, 0.00000034, 0.00000101, 0.00000143, 0.00000156, 0.00000139, 0.00000094, 0.00000024, -0.00000066, -0.00000167, -0.00000271, -0.00000371, -0.00000458, -0.00000525, -0.00000567, -0.00000580, -0.00000562, -0.00000513, -0.00000435, -0.00000332, -0.00000209, -0.00000073, 0.00000069, 0.00000210, 0.00000340, 0.00000455, 0.00000547, 0.00000612, 0.00000648, 0.00000655, 0.00000634, 0.00000589, 0.00000527, 0.00000454, 0.00000381, 0.00000316, 0.00000267, 0.00000241, 0.00000243, 0.00000276, 0.00000336, 0.00000420, 0.00000520, 0.00000623, 0.00000716, 0.00000787, 0.00000820, 0.00000804, 0.00000732, 0.00000597, 0.00000403, 0.00000157, -0.00000130, -0.00000438, -0.00000746, -0.00001030, -0.00001264, -0.00001426, -0.00001497, -0.00001463, -0.00001318, -0.00001065, -0.00000715, -0.00000286, 0.00000194, 0.00000696, 0.00001184, 0.00001625, 0.00001986, 0.00002243, 0.00002377, 0.00002377, 0.00002245, 0.00001990, 0.00001634, 0.00001202, 0.00000728, 0.00000247, -0.00000207, -0.00000602, -0.00000912, -0.00001121, -0.00001220, -0.00001211, -0.00001103, -0.00000916, -0.00000675, -0.00000407, -0.00000144, 0.00000088, 0.00000263, 0.00000366, 0.00000385, 0.00000318, 0.00000173, -0.00000038, -0.00000295, -0.00000575, -0.00000854, -0.00001109, -0.00001318, -0.00001466, -0.00001542, -0.00001544, -0.00001475, -0.00001344, -0.00001166, -0.00000957, -0.00000737, -0.00000524, -0.00000335, -0.00000182, -0.00000074, -0.00000015, -0.00000003, -0.00000032, -0.00000093, -0.00000174, -0.00000263, -0.00000347, -0.00000415, -0.00000459, -0.00000472, -0.00000455, -0.00000407, -0.00000335, -0.00000246, },
		{ -0.00000588, -0.00000644, -0.00000680, -0.00000689, -0.00000668, -0.00000617, -0.00000538, -0.00000437, -0.00000320, -0.00000196, -0.00000072, 0.00000043, 0.00000143, 0.00000221, 0.00000273, 0.00000299, 0.00000297, 0.00000270, 0.00000221, 0.00000154, 0.00000075, -0.00000010, -0.00000094, -0.00000173, -0.00000240, -0.00000291, -0.00000321, -0.00000329, -0.00000313, -0.00000275, -0.00000217, -0.00000144, -0.00000061, 0.00000024, 0.00000102, 0.00000164, 0.00000203, 0.00000211, 0.00000183, 0.00000117, 0.00000013, -0.00000124, -0.00000285, -0.00000458, -0.00000628, -0.00000780, -0.00000895, -0.00000959, -0.00000957, -0.00000882, -0.00000728, -0.00000499, -0.00000205, 0.00000139, 0.00000511, 0.00000885, 0.00001234, 0.00001528, 0.00001742, 0.00001854, 0.00001851, 0.00001725, 0.00001482, 0.00001133, 0.00000700, 0.00000213, -0.00000294, -0.00000784, -0.00001221, -0.00001571, -0.00001808, -0.00001913, -0.00001880, -0.00001710, -0.00001418, -0.00001027, -0.00000566, -0.00000072, 0.00000418, 0.00000869, 0.00001248, 0.00001531, 0.00001702, 0.00001756, 0.00001697, 0.00001538, 0.00001302, 0.00001014, 0.00000706, 0.00000406, 0.00000141, -0.00000066, -0.00000201, -0.00000258, -0.00000239, -0.00000153, -0.00000016, 0.00000152, 0.00000326, 0.00000484, 0.00000606, 0.00000675, 0.00000681, 0.00000621, 0.00000496, 0.00000317, 0.00000097, -0.00000146, -0.00000391, -0.00000621, -0.00000818, -0.00000969, -0.00001066, -0.00001106, -0.00001091, -0.00001029, -0.00000931, -0.00000810, -0.00000680, -0.00000556, -0.00000451, -0.00000372, -0.00000327, -0.00000315, -0.00000336, -0.00000383, -0.00000447, -0.00000519, },
		{ -0.00000282, -0.00000278, -0.00000294, -0.00000332, -0.00000392, -0.00000468, -0.00000556, -0.00000646, -0.00000733, -0.00000806, -0.00000861, -0.00000892, -0.00000896, -0.00000872, -0.00000821, -0.00000747, -0.00000654, -0.00000548, -0.00000435, -0.00000321, -0.00000210, -0.00000109, -0.00000020, 0.00000052, 0.00000107, 0.00000144, 0.00000162, 0.00000163, 0.00000148, 0.00000121, 0.00000085, 0.00000045, 0.00000005, -0.00000028, -0.00000048, -0.00000049, -0.00000026, 0.00000024, 0.00000105, 0.00000215, 0.00000351, 0.00000506, 0.00000670, 0.00000830, 0.00000971, 0.00001078, 0.00001134, 0.00001126, 0.00001043, 0.00000878, 0.00000633, 0.00000313, -0.00000067, -0.00000489, -0.00000927, -0.00001353, -0.00001735, -0.00002045, -0.00002256, -0.00002348, -0.00002308, -0.00002134, -0.00001832, -0.00001420, -0.00000923, -0.00000374, 0.00000188, 0.00000725, 0.00001197, 0.00001572, 0.00001824, 0.00001936, 0.00001904, 0.00001732, 0.00001440, 0.00001053, 0.00000605, 0.00000134, -0.00000322, -0.00000727, -0.00001049, -0.00001268, -0.00001369, -0.00001351, -0.00001223, -0.00001002, -0.00000712, -0.00000383, -0.00000047, 0.00000268, 0.00000534, 0.00000732, 0.00000850, 0.00000885, 0.00000841, 0.00000732, 0.00000576, 0.00000394, 0.00000210, 0.00000047, -0.00000077, -0.00000149, -0.00000161, -0.00000114, -0.00000012, 0.00000132, 0.00000302, 0.00000480, 0.00000647, 0.00000786, 0.00000884, 0.00000930, 0.00000922, 0.00000860, 0.00000750, 0.00000604, 0.00000435, 0.00000256, 0.00000083, -0.00000072, -0.00000200, -0.00000295, -0.00000355, -0.00000382, -0.00000382, -0.00000362, -0.00000332, -0.00000303, },
		{ 0.00000330, 0.00000280, 0.00000243, 0.00000220, 0.00000207, 0.00000199, 0.00000191, 0.00000178, 0.00000152, 0.00000112, 0.00000055, -0.00000019, -0.00000107, -0.00000205, -0.00000308, -0.00000408, -0.00000501, -0.00000579, -0.00000639, -0.00000678, -0.00000695, -0.00000689, -0.00000663, -0.00000620, -0.00000564, -0.00000498, -0.00000428, -0.00000356, -0.00000286, -0.00000220, -0.00000161, -0.00000111, -0.00000071, -0.00000044, -0.00000032, -0.00000036, -0.00000058, -0.00000100, -0.00000162, -0.00000241, -0.00000334, -0.00000436, -0.00000537, -0.00000628, -0.00000696, -0.00000729, -0.00000715, -0.00000643, -0.00000507, -0.00000304, -0.00000037, 0.00000287, 0.00000651, 0.00001035, 0.00001416, 0.00001764, 0.00002053, 0.00002255, 0.00002349, 0.00002317, 0.00002154, 0.00001860, 0.00001448, 0.00000939, 0.00000362, -0.00000247, -0.00000847, -0.00001400, -0.00001869, -0.00002221, -0.00002436, -0.00002500, -0.00002413, -0.00002187, -0.00001842, -0.00001410, -0.00000927, -0.00000432, 0.00000035, 0.00000440, 0.00000754, 0.00000959, 0.00001045, 0.00001015, 0.00000881, 0.00000664, 0.00000391, 0.00000094, -0.00000197, -0.00000451, -0.00000646, -0.00000766, -0.00000801, -0.00000753, -0.00000631, -0.00000451, -0.00000234, -0.00000005, 0.00000214, 0.00000400, 0.00000537, 0.00000612, 0.00000623, 0.00000570, 0.00000465, 0.00000320, 0.00000154, -0.00000015, -0.00000168, -0.00000289, -0.00000367, -0.00000394, -0.00000369, -0.00000296, -0.00000184, -0.00000043, 0.00000111, 0.00000264, 0.00000404, 0.00000518, 0.00000601, 0.00000648, 0.00000660, 0.00000640, 0.00000594, 0.00000532, 0.00000462, 0.00000393, },
		{ 0.00000122, 0.00000179, 0.00000221, 0.00000249, 0.00000265, 0.00000274, 0.00000280, 0.00000285, 0.00000293, 0.00000304, 0.00000318, 0.00000333, 0.00000344, 0.00000348, 0.00000339, 0.00000314, 0.00000271, 0.00000208, 0.00000126, 0.00000029, -0.00000078, -0.00000188, -0.00000296, -0.00000393, -0.00000474, -0.00000533, -0.00000568, -0.00000578, -0.00000564, -0.00000528, -0.00000476, -0.00000411, -0.00000337, -0.00000260, -0.00000182, -0.00000104, -0.00000027, 0.00000049, 0.00000125, 0.00000202, 0.00000278, 0.00000353, 0.00000421, 0.00000479, 0.00000518, 0.00000531, 0.00000508, 0.00000442, 0.00000329, 0.00000165, -0.00000047, -0.00000299, -0.00000578, -0.00000868, -0.00001148, -0.00001394, -0.00001582, -0.00001692, -0.00001704, -0.00001608, -0.00001399, -0.00001083, -0.00000673, -0.00000191, 0.00000334, 0.00000866, 0.00001370, 0.00001810, 0.00002152, 0.00002373, 0.00002455, 0.00002391, 0.00002188, 0.00001860, 0.00001433, 0.00000940, 0.00000419, -0.00000092, -0.00000555, -0.00000940, -0.00001223, -0.00001388, -0.00001433, -0.00001365, -0.00001199, -0.00000962, -0.00000681, -0.00000388, -0.00000114, 0.00000115, 0.00000278, 0.00000363, 0.00000366, 0.00000292, 0.00000155, -0.00000026, -0.00000229, -0.00000429, -0.00000603, -0.00000730, -0.00000798, -0.00000798, -0.00000730, -0.00000602, -0.00000425, -0.00000217, 0.00000003, 0.00000213, 0.00000396, 0.00000537, 0.00000625, 0.00000657, 0.00000632, 0.00000559, 0.00000446, 0.00000309, 0.00000161, 0.00000019, -0.00000106, -0.00000202, -0.00000265, -0.00000291, -0.00000282, -0.00000243, -0.00000182, -0.00000106, -0.00000025, 0.00000053, },
		{ 0.00000153, 0.00000099, 0.00000062, 0.00000040, 0.00000032, 0.00000034, 0.00000042, 0.00000054, 0.00000066, 0.00000078, 0.00000089, 0.00000102, 0.00000118, 0.00000140, 0.00000169, 0.00000207, 0.00000251, 0.00000301, 0.00000351, 0.00000396, 0.00000429, 0.00000444, 0.00000436, 0.00000399, 0.00000333, 0.00000237, 0.00000116, -0.00000025, -0.00000176, -0.00000330, -0.00000474, -0.00000600, -0.00000701, -0.00000771, -0.00000807, -0.00000810, -0.00000786, -0.00000738, -0.00000676, -0.00000607, -0.00000539, -0.00000476, -0.00000422, -0.00000377, -0.00000337, -0.00000297, -0.00000248, -0.00000182, -0.00000092, 0.00000028, 0.00000179, 0.00000361, 0.00000565, 0.00000780, 0.00000990, 0.00001177, 0.00001320, 0.00001402, 0.00001406, 0.00001323, 0.00001148, 0.00000886, 0.00000547, 0.00000151, -0.00000276, -0.00000705, -0.00001103, -0.00001441, -0.00001692, -0.00001834, -0.00001856, -0.00001754, -0.00001536, -0.00001220, -0.00000831, -0.00000398, 0.00000042, 0.00000455, 0.00000811, 0.00001083, 0.00001253, 0.00001312, 0.00001264, 0.00001119, 0.00000896, 0.00000622, 0.00000325, 0.00000035, -0.00000219, -0.00000417, -0.00000541, -0.00000586, -0.00000552, -0.00000449, -0.00000293, -0.00000106, 0.00000088, 0.00000267, 0.00000408, 0.00000495, 0.00000519, 0.00000475, 0.00000370, 0.00000212, 0.00000019, -0.00000189, -0.00000393, -0.00000573, -0.00000710, -0.00000793, -0.00000814, -0.00000774, -0.00000676, -0.00000531, -0.00000353, -0.00000158, 0.00000037, 0.00000217, 0.00000369, 0.00000484, 0.00000556, 0.00000586, 0.00000576, 0.00000533, 0.00000467, 0.00000387, 0.00000303, 0.00000222, },
		{ 0.00000218, 0.00000302, 0.00000362, 0.00000399, 0.00000416, 0.00000418, 0.00000411, 0.00000398, 0.00000382, 0.00000367, 0.00000353, 0.00000340, 0.00000326, 0.00000310, 0.00000292, 0.00000272, 0.00000250, 0.00000230, 0.00000214, 0.00000207, 0.00000213, 0.00000235, 0.00000276, 0.00000335, 0.00000410, 0.00000495, 0.00000583, 0.00000666, 0.00000732, 0.00000772, 0.00000777, 0.00000740, 0.00000658, 0.00000533, 0.00000370, 0.00000177, -0.00000032, -0.00000243, -0.00000440, -0.00000610, -0.00000741, -0.00000823, -0.00000854, -0.00000835, -0.00000773, -0.00000680, -0.00000570, -0.00000459, -0.00000364, -0.00000298, -0.00000272, -0.00000290, -0.00000352, -0.00000449, -0.00000570, -0.00000698, -0.00000812, -0.00000892, -0.00000921, -0.00000883, -0.00000772, -0.00000585, -0.00000329, -0.00000018, 0.00000326, 0.00000679, 0.00001014, 0.00001302, 0.00001520, 0.00001648, 0.00001674, 0.00001596, 0.00001420, 0.00001159, 0.00000837, 0.00000479, 0.00000116, -0.00000224, -0.00000512, -0.00000729, -0.00000861, -0.00000902, -0.00000857, -0.00000737, -0.00000561, -0.00000353, -0.00000138, 0.00000058, 0.00000214, 0.00000312, 0.00000344, 0.00000306, 0.00000206, 0.00000055, -0.00000129, -0.00000323, -0.00000506, -0.00000656, -0.00000757, -0.00000798, -0.00000772, -0.00000682, -0.00000538, -0.00000353, -0.00000145, 0.00000063, 0.00000253, 0.00000405, 0.00000506, 0.00000548, 0.00000526, 0.00000446, 0.00000316, 0.00000150, -0.00000036, -0.00000223, -0.00000395, -0.00000537, -0.00000637, -0.00000688, -0.00000690, -0.00000644, -0.00000557, -0.00000439, -0.00000302, -0.00000158, -0.00000017, 0.00000110, },
		{ -0.00000586, -0.00000609, -0.00000599, -0.00000562, -0.00000503, -0.00000431, -0.00000352, -0.00000271, -0.00000195, -0.00000126, -0.00000066, -0.00000014, 0.00000031, 0.00000070, 0.00000105, 0.00000138, 0.00000168, 0.00000194, 0.00000216, 0.00000230, 0.00000236, 0.00000230, 0.00000215, 0.00000190, 0.00000159, 0.00000129, 0.00000105, 0.00000095, 0.00000106, 0.00000145, 0.00000216, 0.00000319, 0.00000452, 0.00000607, 0.00000774, 0.00000940, 0.00001087, 0.00001200, 0.00001264, 0.00001267, 0.00001201, 0.00001063, 0.00000858, 0.00000598, 0.00000298, -0.00000021, -0.00000333, -0.00000616, -0.00000845, -0.00001003, -0.00001076, -0.00001059, -0.00000954, -0.00000774, -0.00000536, -0.00000265, 0.00000013, 0.00000269, 0.00000477, 0.00000615, 0.00000668, 0.00000630, 0.00000503, 0.00000298, 0.00000035, -0.00000263, -0.00000566, -0.00000845, -0.00001074, -0.00001230, -0.00001297, -0.00001270, -0.00001151, -0.00000949, -0.00000683, -0.00000378, -0.00000058, 0.00000248, 0.00000516, 0.00000727, 0.00000866, 0.00000930, 0.00000918, 0.00000843, 0.00000718, 0.00000564, 0.00000401, 0.00000251, 0.00000132, 0.00000055, 0.00000027, 0.00000049, 0.00000113, 0.00000207, 0.00000316, 0.00000420, 0.00000503, 0.00000548, 0.00000544, 0.00000485, 0.00000371, 0.00000209, 0.00000011, -0.00000207, -0.00000426, -0.00000627, -0.00000791, -0.00000905, -0.00000959, -0.00000949, -0.00000879, -0.00000755, -0.00000591, -0.00000404, -0.00000211, -0.00000031, 0.00000121, 0.00000232, 0.00000295, 0.00000306, 0.00000268, 0.00000187, 0.00000074, -0.00000059, -0.00000198, -0.00000330, -0.00000444, -0.00000531, },
		{ 0.00000145, 0.00000131, 0.00000089, 0.00000025, -0.00000052, -0.00000132, -0.00000208, -0.00000271, -0.00000319, -0.00000348, -0.00000359, -0.00000353, -0.00000335, -0.00000309, -0.00000277, -0.00000245, -0.00000213, -0.00000184, -0.00000157, -0.00000131, -0.00000105, -0.00000078, -0.00000050, -0.00000021, 0.00000006, 0.00000029, 0.00000042, 0.00000042, 0.00000024, -0.00000014, -0.00000074, -0.00000152, -0.00000244, -0.00000341, -0.00000432, -0.00000505, -0.00000545, -0.00000541, -0.00000482, -0.00000362, -0.00000182, 0.00000054, 0.00000334, 0.00000640, 0.00000950, 0.00001240, 0.00001483, 0.00001655, 0.00001737, 0.00001714, 0.00001580, 0.00001339, 0.00001003, 0.00000593, 0.00000139, -0.00000328, -0.00000772, -0.00001158, -0.00001456, -0.00001643, -0.00001703, -0.00001633, -0.00001440, -0.00001139, -0.00000758, -0.00000328, 0.00000115, 0.00000536, 0.00000900, 0.00001180, 0.00001358, 0.00001424, 0.00001377, 0.00001230, 0.00000999, 0.00000710, 0.00000391, 0.00000072, -0.00000221, -0.00000464, -0.00000642, -0.00000747, -0.00000777, -0.00000740, -0.00000649, -0.00000521, -0.00000376, -0.00000232, -0.00000106, -0.00000011, 0.00000047, 0.00000065, 0.00000050, 0.00000010, -0.00000043, -0.00000094, -0.00000130, -0.00000140, -0.00000115, -0.00000053, 0.00000043, 0.00000167, 0.00000306, 0.00000446, 0.00000571, 0.00000666, 0.00000719, 0.00000720, 0.00000666, 0.00000559, 0.00000404, 0.00000215, 0.00000006, -0.00000206, -0.00000403, -0.00000571, -0.00000695, -0.00000769, -0.00000788, -0.00000755, -0.00000676, -0.00000562, -0.00000426, -0.00000281, -0.00000144, -0.00000024, 0.00000067, 0.00000124, },
		{ -0.00000654, -0.00000610, -0.00000542, -0.00000459, -0.00000373, -0.00000292, -0.00000226, -0.00000179, -0.00000154, -0.00000151, -0.00000166, -0.00000196, -0.00000233, -0.00000271, -0.00000304, -0.00000327, -0.00000338, -0.00000334, -0.00000318, -0.00000290, -0.00000254, -0.00000212, -0.00000168, -0.00000124, -0.00000080, -0.00000035, 0.00000012, 0.00000063, 0.00000120, 0.00000185, 0.00000256, 0.00000331, 0.00000407, 0.00000475, 0.00000527, 0.00000554, 0.00000547, 0.00000497, 0.00000399, 0.00000253, 0.00000063, -0.00000162, -0.00000409, -0.00000659, -0.00000891, -0.00001081, -0.00001209, -0.00001257, -0.00001210, -0.00001064, -0.00000822, -0.00000493, -0.00000099, 0.00000334, 0.00000774, 0.00001185, 0.00001535, 0.00001790, 0.00001927, 0.00001929, 0.00001792, 0.00001521, 0.00001132, 0.00000653, 0.00000118, -0.00000433, -0.00000961, -0.00001425, -0.00001793, -0.00002039, -0.00002148, -0.00002116, -0.00001949, -0.00001666, -0.00001293, -0.00000861, -0.00000406, 0.00000038, 0.00000439, 0.00000771, 0.00001018, 0.00001170, 0.00001228, 0.00001200, 0.00001100, 0.00000950, 0.00000771, 0.00000584, 0.00000410, 0.00000264, 0.00000154, 0.00000086, 0.00000056, 0.00000057, 0.00000080, 0.00000111, 0.00000139, 0.00000153, 0.00000145, 0.00000113, 0.00000057, -0.00000017, -0.00000100, -0.00000182, -0.00000250, -0.00000294, -0.00000305, -0.00000279, -0.00000215, -0.00000118, 0.00000006, 0.00000144, 0.00000282, 0.00000407, 0.00000505, 0.00000564, 0.00000578, 0.00000543, 0.00000462, 0.00000340, 0.00000187, 0.00000017, -0.00000157, -0.00000320, -0.00000460, -0.00000568, -0.00000636, -0.00000664, },
		{ 0.00000356, 0.00000240, 0.00000113, -0.00000013, -0.00000128, -0.00000223, -0.00000293, -0.00000334, -0.00000348, -0.00000338, -0.00000310, -0.00000271, -0.00000232, -0.00000198, -0.00000177, -0.00000172, -0.00000187, -0.00000220, -0.00000269, -0.00000328, -0.00000393, -0.00000458, -0.00000516, -0.00000565, -0.00000602, -0.00000626, -0.00000639, -0.00000642, -0.00000638, -0.00000630, -0.00000618, -0.00000604, -0.00000584, -0.00000556, -0.00000514, -0.00000453, -0.00000367, -0.00000252, -0.00000105, 0.00000072, 0.00000274, 0.00000492, 0.00000713, 0.00000920, 0.00001098, 0.00001226, 0.00001291, 0.00001278, 0.00001182, 0.00001002, 0.00000744, 0.00000425, 0.00000064, -0.00000312, -0.00000672, -0.00000988, -0.00001229, -0.00001374, -0.00001406, -0.00001318, -0.00001113, -0.00000805, -0.00000415, 0.00000026, 0.00000482, 0.00000915, 0.00001290, 0.00001573, 0.00001740, 0.00001776, 0.00001676, 0.00001446, 0.00001104, 0.00000676, 0.00000196, -0.00000301, -0.00000777, -0.00001199, -0.00001537, -0.00001771, -0.00001890, -0.00001893, -0.00001788, -0.00001591, -0.00001326, -0.00001019, -0.00000697, -0.00000386, -0.00000109, 0.00000119, 0.00000287, 0.00000393, 0.00000440, 0.00000439, 0.00000401, 0.00000343, 0.00000278, 0.00000221, 0.00000182, 0.00000165, 0.00000172, 0.00000200, 0.00000243, 0.00000291, 0.00000334, 0.00000365, 0.00000374, 0.00000358, 0.00000316, 0.00000251, 0.00000170, 0.00000081, -0.00000005, -0.00000077, -0.00000126, -0.00000144, -0.00000129, -0.00000081, -0.00000003, 0.00000097, 0.00000208, 0.00000320, 0.00000418, 0.00000494, 0.00000539, 0.00000547, 0.00000517, 0.00000451, },
		{ 0.00000008, 0.00000092, 0.00000171, 0.00000237, 0.00000281, 0.00000300, 0.00000290, 0.00000253, 0.00000194, 0.00000118, 0.00000034, -0.00000047, -0.00000118, -0.00000171, -0.00000200, -0.00000203, -0.00000181, -0.00000138, -0.00000078, -0.00000009, 0.00000060, 0.00000122, 0.00000170, 0.00000200, 0.00000210, 0.00000200, 0.00000173, 0.00000131, 0.00000081, 0.00000027, -0.00000027, -0.00000078, -0.00000125, -0.00000170, -0.00000216, -0.00000267, -0.00000327, -0.00000401, -0.00000491, -0.00000596, -0.00000714, -0.00000838, -0.00000958, -0.00001063, -0.00001138, -0.00001171, -0.00001150, -0.00001067, -0.00000918, -0.00000704, -0.00000435, -0.00000124, 0.00000211, 0.00000545, 0.00000854, 0.00001113, 0.00001301, 0.00001402, 0.00001404, 0.00001306, 0.00001115, 0.00000846, 0.00000520, 0.00000167, -0.00000184, -0.00000501, -0.00000757, -0.00000928, -0.00000999, -0.00000964, -0.00000825, -0.00000596, -0.00000298, 0.00000043, 0.00000395, 0.00000725, 0.00001004, 0.00001207, 0.00001316, 0.00001322, 0.00001223, 0.00001031, 0.00000760, 0.00000436, 0.00000086, -0.00000263, -0.00000583, -0.00000851, -0.00001050, -0.00001169, -0.00001207, -0.00001168, -0.00001063, -0.00000909, -0.00000726, -0.00000534, -0.00000350, -0.00000191, -0.00000067, 0.00000015, 0.00000055, 0.00000057, 0.00000031, -0.00000014, -0.00000064, -0.00000109, -0.00000140, -0.00000151, -0.00000138, -0.00000103, -0.00000051, 0.00000011, 0.00000074, 0.00000130, 0.00000169, 0.00000188, 0.00000182, 0.00000152, 0.00000102, 0.00000038, -0.00000032, -0.00000099, -0.00000152, -0.00000187, -0.00000196, -0.00000178, -0.00000135, -0.00000071, },
		{ 0.00000094, 0.00000038, -0.00000011, -0.00000049, -0.00000071, -0.00000072, -0.00000055, -0.00000022, 0.00000021, 0.00000068, 0.00000109, 0.00000137, 0.00000145, 0.00000127, 0.00000083, 0.00000013, -0.00000078, -0.00000185, -0.00000299, -0.00000410, -0.00000508, -0.00000587, -0.00000640, -0.00000664, -0.00000658, -0.00000624, -0.00000569, -0.00000498, -0.00000419, -0.00000340, -0.00000266, -0.00000201, -0.00000148, -0.00000106, -0.00000071, -0.00000040, -0.00000006, 0.00000036, 0.00000090, 0.00000158, 0.00000239, 0.00000329, 0.00000420, 0.00000504, 0.00000568, 0.00000603, 0.00000597, 0.00000544, 0.00000439, 0.00000285, 0.00000086, -0.00000146, -0.00000395, -0.00000642, -0.00000867, -0.00001050, -0.00001175, -0.00001227, -0.00001201, -0.00001095, -0.00000917, -0.00000679, -0.00000401, -0.00000105, 0.00000184, 0.00000442, 0.00000647, 0.00000783, 0.00000840, 0.00000817, 0.00000718, 0.00000557, 0.00000352, 0.00000126, -0.00000097, -0.00000293, -0.00000441, -0.00000526, -0.00000540, -0.00000480, -0.00000354, -0.00000175, 0.00000039, 0.00000266, 0.00000482, 0.00000665, 0.00000796, 0.00000862, 0.00000855, 0.00000777, 0.00000634, 0.00000440, 0.00000211, -0.00000031, -0.00000265, -0.00000472, -0.00000635, -0.00000744, -0.00000792, -0.00000781, -0.00000716, -0.00000608, -0.00000471, -0.00000321, -0.00000174, -0.00000043, 0.00000061, 0.00000131, 0.00000165, 0.00000166, 0.00000138, 0.00000092, 0.00000037, -0.00000017, -0.00000060, -0.00000085, -0.00000089, -0.00000071, -0.00000033, 0.00000018, 0.00000076, 0.00000133, 0.00000181, 0.00000213, 0.00000226, 0.00000218, 0.00000190, 0.00000147, },
		{ 0.00000055, 0.00000122, 0.00000185, 0.00000237, 0.00000275, 0.00000297, 0.00000301, 0.00000292, 0.00000275, 0.00000255, 0.00000239, 0.00000233, 0.00000242, 0.00000269, 0.00000312, 0.00000370, 0.00000437, 0.00000504, 0.00000565, 0.00000610, 0.00000631, 0.00000623, 0.00000583, 0.00000510, 0.00000408, 0.00000282, 0.00000140, -0.00000008, -0.00000153, -0.00000286, -0.00000400, -0.00000491, -0.00000558, -0.00000603, -0.00000631, -0.00000646, -0.00000658, -0.00000674, -0.00000698, -0.00000736, -0.00000786, -0.00000848, -0.00000915, -0.00000978, -0.00001026, -0.00001049, -0.00001037, -0.00000982, -0.00000879, -0.00000729, -0.00000535, -0.00000307, -0.00000059, 0.00000192, 0.00000430, 0.00000635, 0.00000793, 0.00000891, 0.00000921, 0.00000883, 0.00000780, 0.00000624, 0.00000428, 0.00000212, -0.00000004, -0.00000202, -0.00000363, -0.00000475, -0.00000531, -0.00000527, -0.00000468, -0.00000365, -0.00000230, -0.00000082, 0.00000063, 0.00000187, 0.00000278, 0.00000325, 0.00000324, 0.00000278, 0.00000192, 0.00000080, -0.00000045, -0.00000166, -0.00000268, -0.00000335, -0.00000360, -0.00000336, -0.00000264, -0.00000151, -0.00000007, 0.00000153, 0.00000313, 0.00000454, 0.00000562, 0.00000624, 0.00000632, 0.00000583, 0.00000481, 0.00000334, 0.00000154, -0.00000043, -0.00000240, -0.00000420, -0.00000569, -0.00000676, -0.00000734, -0.00000742, -0.00000702, -0.00000622, -0.00000513, -0.00000387, -0.00000259, -0.00000141, -0.00000042, 0.00000029, 0.00000070, 0.00000081, 0.00000067, 0.00000033, -0.00000011, -0.00000055, -0.00000092, -0.00000115, -0.00000118, -0.00000100, -0.00000062, -0.00000008, },
		{ -0.00000189, -0.00000242, -0.00000289, -0.00000321, -0.00000335, -0.00000329, -0.00000302, -0.00000258, -0.00000203, -0.00000142, -0.00000083, -0.00000032, 0.00000007, 0.00000031, 0.00000040, 0.00000037, 0.00000027, 0.00000015, 0.00000009, 0.00000016, 0.00000041, 0.00000088, 0.00000158, 0.00000248, 0.00000353, 0.00000468, 0.00000582, 0.00000687, 0.00000774, 0.00000835, 0.00000866, 0.00000865, 0.00000833, 0.00000774, 0.00000697, 0.00000610, 0.00000523, 0.00000444, 0.00000381, 0.00000339, 0.00000318, 0.00000317, 0.00000329, 0.00000345, 0.00000355, 0.00000347, 0.00000311, 0.00000238, 0.00000124, -0.00000033, -0.00000229, -0.00000454, -0.00000696, -0.00000940, -0.00001169, -0.00001365, -0.00001513, -0.00001602, -0.00001623, -0.00001575, -0.00001460, -0.00001289, -0.00001074, -0.00000831, -0.00000579, -0.00000335, -0.00000116, 0.00000065, 0.00000200, 0.00000285, 0.00000322, 0.00000317, 0.00000279, 0.00000222, 0.00000157, 0.00000098, 0.00000054, 0.00000033, 0.00000038, 0.00000068, 0.00000117, 0.00000180, 0.00000244, 0.00000301, 0.00000340, 0.00000354, 0.00000340, 0.00000295, 0.00000223, 0.00000132, 0.00000031, -0.00000068, -0.00000153, -0.00000214, -0.00000241, -0.00000230, -0.00000180, -0.00000094, 0.00000020, 0.00000152, 0.00000288, 0.00000414, 0.00000516, 0.00000583, 0.00000607, 0.00000583, 0.00000513, 0.00000401, 0.00000257, 0.00000092, -0.00000080, -0.00000244, -0.00000387, -0.00000499, -0.00000572, -0.00000604, -0.00000595, -0.00000551, -0.00000480, -0.00000392, -0.00000299, -0.00000213, -0.00000141, -0.00000091, -0.00000067, -0.00000070, -0.00000095, -0.00000137, },
		{ -0.00000124, -0.00000062, -0.00000015, 0.00000011, 0.00000013, -0.00000009, -0.00000052, -0.00000112, -0.00000180, -0.00000249, -0.00000310, -0.00000357, -0.00000383, -0.00000387, -0.00000370, -0.00000334, -0.00000284, -0.00000228, -0.00000174, -0.00000127, -0.00000094, -0.00000079, -0.00000083, -0.00000104, -0.00000137, -0.00000176, -0.00000214, -0.00000242, -0.00000254, -0.00000244, -0.00000209, -0.00000149, -0.00000068, 0.00000028, 0.00000131, 0.00000232, 0.00000320, 0.00000387, 0.00000426, 0.00000435, 0.00000414, 0.00000367, 0.00000301, 0.00000226, 0.00000155, 0.00000100, 0.00000071, 0.00000079, 0.00000127, 0.00000217, 0.00000345, 0.00000503, 0.00000678, 0.00000855, 0.00001019, 0.00001151, 0.00001239, 0.00001270, 0.00001239, 0.00001142, 0.00000985, 0.00000774, 0.00000524, 0.00000250, -0.00000030, -0.00000300, -0.00000544, -0.00000748, -0.00000903, -0.00001007, -0.00001058, -0.00001061, -0.00001025, -0.00000959, -0.00000875, -0.00000785, -0.00000697, -0.00000619, -0.00000555, -0.00000506, -0.00000471, -0.00000445, -0.00000422, -0.00000398, -0.00000364, -0.00000319, -0.00000259, -0.00000186, -0.00000102, -0.00000013, 0.00000074, 0.00000151, 0.00000211, 0.00000249, 0.00000260, 0.00000244, 0.00000202, 0.00000140, 0.00000065, -0.00000012, -0.00000082, -0.00000134, -0.00000162, -0.00000159, -0.00000123, -0.00000055, 0.00000039, 0.00000152, 0.00000273, 0.00000390, 0.00000491, 0.00000567, 0.00000608, 0.00000610, 0.00000572, 0.00000498, 0.00000392, 0.00000266, 0.00000129, -0.00000005, -0.00000127, -0.00000225, -0.00000293, -0.00000328, -0.00000331, -0.00000304, -0.00000254, -0.00000191, },
		{ -0.00000010, -0.00000081, -0.00000128, -0.00000148, -0.00000141, -0.00000109, -0.00000061, -0.00000002, 0.00000057, 0.00000107, 0.00000140, 0.00000150, 0.00000135, 0.00000094, 0.00000030, -0.00000051, -0.00000141, -0.00000230, -0.00000311, -0.00000373, -0.00000412, -0.00000424, -0.00000408, -0.00000368, -0.00000307, -0.00000235, -0.00000159, -0.00000089, -0.00000031, 0.00000007, 0.00000024, 0.00000018, -0.00000006, -0.00000043, -0.00000086, -0.00000126, -0.00000154, -0.00000162, -0.00000147, -0.00000104, -0.00000037, 0.00000050, 0.00000149, 0.00000249, 0.00000339, 0.00000407, 0.00000442, 0.00000438, 0.00000390, 0.00000298, 0.00000168, 0.00000008, -0.00000169, -0.00000347, -0.00000511, -0.00000645, -0.00000736, -0.00000773, -0.00000750, -0.00000667, -0.00000528, -0.00000341, -0.00000119, 0.00000123, 0.00000367, 0.00000599, 0.00000803, 0.00000966, 0.00001082, 0.00001147, 0.00001160, 0.00001126, 0.00001052, 0.00000949, 0.00000827, 0.00000696, 0.00000565, 0.00000443, 0.00000334, 0.00000240, 0.00000161, 0.00000095, 0.00000038, -0.00000014, -0.00000066, -0.00000119, -0.00000175, -0.00000234, -0.00000295, -0.00000352, -0.00000402, -0.00000441, -0.00000463, -0.00000466, -0.00000448, -0.00000412, -0.00000358, -0.00000293, -0.00000223, -0.00000154, -0.00000095, -0.00000050, -0.00000026, -0.00000023, -0.00000041, -0.00000077, -0.00000126, -0.00000179, -0.00000230, -0.00000269, -0.00000289, -0.00000284, -0.00000252, -0.00000192, -0.00000108, -0.00000006, 0.00000105, 0.00000217, 0.00000319, 0.00000402, 0.00000458, 0.00000483, 0.00000474, 0.00000433, 0.00000365, 0.00000277, 0.00000178, 0.00000080, },
		{ 0.00000316, 0.00000342, 0.00000341, 0.00000314, 0.00000265, 0.00000201, 0.00000128, 0.00000056, -0.00000006, -0.00000051, -0.00000074, -0.00000072, -0.00000044, 0.00000004, 0.00000067, 0.00000136, 0.00000202, 0.00000255, 0.00000287, 0.00000291, 0.00000263, 0.00000204, 0.00000116, 0.00000006, -0.00000118, -0.00000245, -0.00000364, -0.00000465, -0.00000540, -0.00000582, -0.00000589, -0.00000564, -0.00000509, -0.00000433, -0.00000345, -0.00000255, -0.00000174, -0.00000110, -0.00000069, -0.00000053, -0.00000063, -0.00000094, -0.00000137, -0.00000185, -0.00000226, -0.00000250, -0.00000247, -0.00000211, -0.00000139, -0.00000032, 0.00000104, 0.00000262, 0.00000428, 0.00000590, 0.00000733, 0.00000844, 0.00000910, 0.00000925, 0.00000883, 0.00000785, 0.00000638, 0.00000449, 0.00000231, -0.00000000, -0.00000230, -0.00000443, -0.00000627, -0.00000770, -0.00000866, -0.00000912, -0.00000909, -0.00000861, -0.00000776, -0.00000664, -0.00000533, -0.00000394, -0.00000256, -0.00000125, -0.00000006, 0.00000097, 0.00000185, 0.00000259, 0.00000322, 0.00000375, 0.00000423, 0.00000467, 0.00000508, 0.00000546, 0.00000579, 0.00000605, 0.00000621, 0.00000623, 0.00000611, 0.00000581, 0.00000535, 0.00000475, 0.00000402, 0.00000324, 0.00000244, 0.00000169, 0.00000104, 0.00000053, 0.00000020, 0.00000004, 0.00000005, 0.00000020, 0.00000044, 0.00000071, 0.00000095, 0.00000111, 0.00000113, 0.00000100, 0.00000070, 0.00000026, -0.00000029, -0.00000089, -0.00000147, -0.00000197, -0.00000231, -0.00000244, -0.00000234, -0.00000199, -0.00000142, -0.00000067, 0.00000019, 0.00000109, 0.00000194, 0.00000265, },
		{ 0.00000008, 0.00000027, 0.00000063, 0.00000114, 0.00000174, 0.00000237, 0.00000296, 0.00000345, 0.00000376, 0.00000387, 0.00000375, 0.00000342, 0.00000291, 0.00000228, 0.00000162, 0.00000100, 0.00000050, 0.00000021, 0.00000017, 0.00000040, 0.00000091, 0.00000164, 0.00000253, 0.00000349, 0.00000440, 0.00000517, 0.00000568, 0.00000585, 0.00000563, 0.00000501, 0.00000400, 0.00000267, 0.00000109, -0.00000060, -0.00000228, -0.00000383, -0.00000512, -0.00000606, -0.00000660, -0.00000672, -0.00000643, -0.00000581, -0.00000493, -0.00000393, -0.00000291, -0.00000201, -0.00000133, -0.00000095, -0.00000090, -0.00000119, -0.00000179, -0.00000261, -0.00000355, -0.00000449, -0.00000531, -0.00000588, -0.00000611, -0.00000593, -0.00000531, -0.00000426, -0.00000282, -0.00000109, 0.00000082, 0.00000279, 0.00000468, 0.00000637, 0.00000775, 0.00000874, 0.00000929, 0.00000939, 0.00000905, 0.00000832, 0.00000729, 0.00000602, 0.00000462, 0.00000317, 0.00000175, 0.00000043, -0.00000075, -0.00000177, -0.00000263, -0.00000332, -0.00000387, -0.00000429, -0.00000461, -0.00000484, -0.00000498, -0.00000504, -0.00000501, -0.00000487, -0.00000461, -0.00000422, -0.00000369, -0.00000302, -0.00000225, -0.00000138, -0.00000048, 0.00000042, 0.00000127, 0.00000200, 0.00000259, 0.00000299, 0.00000320, 0.00000323, 0.00000308, 0.00000280, 0.00000243, 0.00000204, 0.00000167, 0.00000138, 0.00000120, 0.00000115, 0.00000123, 0.00000142, 0.00000171, 0.00000203, 0.00000234, 0.00000259, 0.00000274, 0.00000275, 0.00000262, 0.00000235, 0.00000197, 0.00000151, 0.00000104, 0.00000061, 0.00000028, 0.00000009, },
		{ 0.00000055, 0.00000063, 0.00000062, 0.00000055, 0.00000045, 0.00000037, 0.00000034, 0.00000041, 0.00000060, 0.00000091, 0.00000134, 0.00000186, 0.00000243, 0.00000298, 0.00000346, 0.00000380, 0.00000396, 0.00000391, 0.00000364, 0.00000316, 0.00000252, 0.00000178, 0.00000102, 0.00000034, -0.00000016, -0.00000042, -0.00000037, 0.00000003, 0.00000077, 0.00000182, 0.00000311, 0.00000453, 0.00000596, 0.00000728, 0.00000835, 0.00000905, 0.00000931, 0.00000907, 0.00000832, 0.00000710, 0.00000547, 0.00000356, 0.00000150, -0.00000055, -0.00000246, -0.00000407, -0.00000529, -0.00000603, -0.00000626, -0.00000600, -0.00000531, -0.00000428, -0.00000302, -0.00000168, -0.00000039, 0.00000071, 0.00000152, 0.00000197, 0.00000201, 0.00000165, 0.00000092, -0.00000009, -0.00000131, -0.00000260, -0.00000387, -0.00000499, -0.00000589, -0.00000649, -0.00000676, -0.00000667, -0.00000625, -0.00000555, -0.00000461, -0.00000352, -0.00000234, -0.00000115, -0.00000000, 0.00000104, 0.00000196, 0.00000273, 0.00000334, 0.00000380, 0.00000413, 0.00000434, 0.00000443, 0.00000444, 0.00000435, 0.00000416, 0.00000389, 0.00000352, 0.00000305, 0.00000248, 0.00000183, 0.00000111, 0.00000035, -0.00000042, -0.00000114, -0.00000179, -0.00000231, -0.00000267, -0.00000285, -0.00000284, -0.00000264, -0.00000228, -0.00000180, -0.00000124, -0.00000066, -0.00000012, 0.00000033, 0.00000065, 0.00000082, 0.00000082, 0.00000065, 0.00000036, -0.00000003, -0.00000047, -0.00000090, -0.00000127, -0.00000155, -0.00000170, -0.00000171, -0.00000158, -0.00000133, -0.00000100, -0.00000062, -0.00000024, 0.00000011, 0.00000038, },
		{ -0.00000244, -0.00000275, -0.00000292, -0.00000296, -0.00000290, -0.00000276, -0.00000258, -0.00000237, -0.00000217, -0.00000197, -0.00000179, -0.00000161, -0.00000140, -0.00000115, -0.00000083, -0.00000042, 0.00000010, 0.00000071, 0.00000139, 0.00000211, 0.00000282, 0.00000346, 0.00000396, 0.00000427, 0.00000433, 0.00000412, 0.00000362, 0.00000285, 0.00000186, 0.00000073, -0.00000045, -0.00000158, -0.00000253, -0.00000321, -0.00000353, -0.00000342, -0.00000287, -0.00000187, -0.00000049, 0.00000118, 0.00000304, 0.00000493, 0.00000670, 0.00000821, 0.00000934, 0.00000997, 0.00001005, 0.00000955, 0.00000851, 0.00000698, 0.00000509, 0.00000296, 0.00000075, -0.00000138, -0.00000329, -0.00000484, -0.00000594, -0.00000654, -0.00000660, -0.00000617, -0.00000530, -0.00000408, -0.00000263, -0.00000106, 0.00000050, 0.00000193, 0.00000315, 0.00000408, 0.00000469, 0.00000495, 0.00000489, 0.00000453, 0.00000393, 0.00000315, 0.00000225, 0.00000130, 0.00000035, -0.00000056, -0.00000138, -0.00000211, -0.00000272, -0.00000322, -0.00000361, -0.00000389, -0.00000406, -0.00000412, -0.00000409, -0.00000394, -0.00000370, -0.00000335, -0.00000291, -0.00000238, -0.00000179, -0.00000117, -0.00000054, 0.00000006, 0.00000059, 0.00000101, 0.00000130, 0.00000143, 0.00000141, 0.00000124, 0.00000094, 0.00000054, 0.00000010, -0.00000033, -0.00000071, -0.00000099, -0.00000113, -0.00000111, -0.00000093, -0.00000059, -0.00000012, 0.00000043, 0.00000101, 0.00000156, 0.00000203, 0.00000236, 0.00000253, 0.00000250, 0.00000228, 0.00000187, 0.00000132, 0.00000067, -0.00000005, -0.00000076, -0.00000142, -0.00000199, },
		{ 0.00000309, 0.00000296, 0.00000261, 0.00000207, 0.00000139, 0.00000060, -0.00000024, -0.00000108, -0.00000188, -0.00000261, -0.00000324, -0.00000377, -0.00000418, -0.00000449, -0.00000470, -0.00000482, -0.00000486, -0.00000481, -0.00000468, -0.00000444, -0.00000408, -0.00000359, -0.00000295, -0.00000214, -0.00000117, -0.00000007, 0.00000114, 0.00000240, 0.00000366, 0.00000482, 0.00000583, 0.00000658, 0.00000703, 0.00000711, 0.00000680, 0.00000610, 0.00000506, 0.00000372, 0.00000219, 0.00000057, -0.00000101, -0.00000243, -0.00000357, -0.00000432, -0.00000463, -0.00000446, -0.00000380, -0.00000271, -0.00000127, 0.00000041, 0.00000220, 0.00000395, 0.00000553, 0.00000681, 0.00000767, 0.00000806, 0.00000792, 0.00000728, 0.00000617, 0.00000466, 0.00000287, 0.00000091, -0.00000108, -0.00000297, -0.00000466, -0.00000605, -0.00000707, -0.00000768, -0.00000787, -0.00000766, -0.00000709, -0.00000621, -0.00000511, -0.00000385, -0.00000252, -0.00000117, 0.00000014, 0.00000134, 0.00000242, 0.00000335, 0.00000412, 0.00000472, 0.00000515, 0.00000541, 0.00000551, 0.00000545, 0.00000524, 0.00000489, 0.00000440, 0.00000380, 0.00000309, 0.00000231, 0.00000150, 0.00000067, -0.00000012, -0.00000084, -0.00000145, -0.00000193, -0.00000225, -0.00000241, -0.00000242, -0.00000229, -0.00000206, -0.00000177, -0.00000146, -0.00000119, -0.00000098, -0.00000088, -0.00000090, -0.00000105, -0.00000132, -0.00000168, -0.00000209, -0.00000250, -0.00000286, -0.00000311, -0.00000321, -0.00000314, -0.00000287, -0.00000241, -0.00000179, -0.00000103, -0.00000020, 0.00000065, 0.00000145, 0.00000214, 0.00000267, 0.00000299, },
		{ -0.00000305, -0.00000241, -0.00000160, -0.00000069, 0.00000026, 0.00000118, 0.00000201, 0.00000269, 0.00000318, 0.00000345, 0.00000348, 0.00000328, 0.00000286, 0.00000226, 0.00000149, 0.00000062, -0.00000033, -0.00000133, -0.00000233, -0.00000332, -0.00000427, -0.00000517, -0.00000600, -0.00000675, -0.00000741, -0.00000795, -0.00000835, -0.00000857, -0.00000860, -0.00000839, -0.00000793, -0.00000720, -0.00000621, -0.00000495, -0.00000348, -0.00000183, -0.00000008, 0.00000168, 0.00000338, 0.00000491, 0.00000618, 0.00000713, 0.00000769, 0.00000784, 0.00000758, 0.00000693, 0.00000596, 0.00000475, 0.00000340, 0.00000202, 0.00000073, -0.00000038, -0.00000120, -0.00000169, -0.00000179, -0.00000153, -0.00000092, -0.00000002, 0.00000108, 0.00000228, 0.00000348, 0.00000457, 0.00000546, 0.00000607, 0.00000634, 0.00000626, 0.00000582, 0.00000505, 0.00000400, 0.00000272, 0.00000131, -0.00000017, -0.00000163, -0.00000301, -0.00000424, -0.00000529, -0.00000613, -0.00000674, -0.00000713, -0.00000730, -0.00000727, -0.00000707, -0.00000671, -0.00000621, -0.00000560, -0.00000488, -0.00000408, -0.00000321, -0.00000228, -0.00000132, -0.00000036, 0.00000059, 0.00000149, 0.00000231, 0.00000301, 0.00000356, 0.00000395, 0.00000415, 0.00000417, 0.00000402, 0.00000371, 0.00000327, 0.00000277, 0.00000223, 0.00000170, 0.00000124, 0.00000087, 0.00000062, 0.00000049, 0.00000048, 0.00000057, 0.00000073, 0.00000090, 0.00000105, 0.00000112, 0.00000109, 0.00000092, 0.00000060, 0.00000014, -0.00000043, -0.00000109, -0.00000178, -0.00000244, -0.00000301, -0.00000344, -0.00000368, -0.00000370, -0.00000349, },
		{ 0.00000042, -0.00000031, -0.00000109, -0.00000185, -0.00000253, -0.00000307, -0.00000342, -0.00000354, -0.00000342, -0.00000306, -0.00000249, -0.00000173, -0.00000083, 0.00000013, 0.00000111, 0.00000205, 0.00000288, 0.00000357, 0.00000408, 0.00000439, 0.00000450, 0.00000440, 0.00000411, 0.00000363, 0.00000300, 0.00000223, 0.00000135, 0.00000037, -0.00000068, -0.00000178, -0.00000292, -0.00000407, -0.00000519, -0.00000627, -0.00000726, -0.00000811, -0.00000880, -0.00000927, -0.00000949, -0.00000943, -0.00000908, -0.00000843, -0.00000751, -0.00000635, -0.00000501, -0.00000355, -0.00000205, -0.00000059, 0.00000074, 0.00000188, 0.00000277, 0.00000337, 0.00000366, 0.00000366, 0.00000340, 0.00000292, 0.00000230, 0.00000162, 0.00000096, 0.00000039, -0.00000002, -0.00000022, -0.00000019, 0.00000008, 0.00000057, 0.00000123, 0.00000201, 0.00000284, 0.00000366, 0.00000439, 0.00000498, 0.00000537, 0.00000554, 0.00000548, 0.00000520, 0.00000472, 0.00000407, 0.00000331, 0.00000247, 0.00000161, 0.00000077, -0.00000002, -0.00000075, -0.00000140, -0.00000197, -0.00000246, -0.00000290, -0.00000329, -0.00000363, -0.00000394, -0.00000422, -0.00000445, -0.00000463, -0.00000473, -0.00000474, -0.00000463, -0.00000441, -0.00000405, -0.00000359, -0.00000301, -0.00000237, -0.00000169, -0.00000101, -0.00000038, 0.00000016, 0.00000059, 0.00000088, 0.00000101, 0.00000101, 0.00000088, 0.00000065, 0.00000038, 0.00000010, -0.00000013, -0.00000029, -0.00000034, -0.00000027, -0.00000007, 0.00000023, 0.00000060, 0.00000101, 0.00000140, 0.00000172, 0.00000193, 0.00000198, 0.00000186, 0.00000154, 0.00000106, },
		{ -0.00000015, 0.00000039, 0.00000090, 0.00000134, 0.00000166, 0.00000180, 0.00000176, 0.00000152, 0.00000108, 0.00000049, -0.00000022, -0.00000100, -0.00000178, -0.00000250, -0.00000311, -0.00000357, -0.00000383, -0.00000389, -0.00000373, -0.00000337, -0.00000284, -0.00000218, -0.00000142, -0.00000062, 0.00000017, 0.00000092, 0.00000159, 0.00000214, 0.00000256, 0.00000284, 0.00000297, 0.00000295, 0.00000279, 0.00000248, 0.00000204, 0.00000149, 0.00000082, 0.00000007, -0.00000075, -0.00000162, -0.00000251, -0.00000338, -0.00000420, -0.00000492, -0.00000550, -0.00000592, -0.00000613, -0.00000611, -0.00000585, -0.00000537, -0.00000468, -0.00000381, -0.00000283, -0.00000180, -0.00000079, 0.00000014, 0.00000091, 0.00000148, 0.00000180, 0.00000185, 0.00000165, 0.00000121, 0.00000058, -0.00000016, -0.00000096, -0.00000172, -0.00000238, -0.00000286, -0.00000311, -0.00000311, -0.00000285, -0.00000235, -0.00000164, -0.00000079, 0.00000015, 0.00000108, 0.00000195, 0.00000267, 0.00000321, 0.00000352, 0.00000359, 0.00000344, 0.00000309, 0.00000258, 0.00000197, 0.00000132, 0.00000067, 0.00000010, -0.00000038, -0.00000074, -0.00000097, -0.00000109, -0.00000110, -0.00000105, -0.00000097, -0.00000090, -0.00000086, -0.00000088, -0.00000097, -0.00000112, -0.00000131, -0.00000153, -0.00000175, -0.00000192, -0.00000202, -0.00000203, -0.00000194, -0.00000173, -0.00000143, -0.00000107, -0.00000066, -0.00000027, 0.00000009, 0.00000036, 0.00000052, 0.00000056, 0.00000046, 0.00000024, -0.00000007, -0.00000043, -0.00000081, -0.00000115, -0.00000140, -0.00000154, -0.00000154, -0.00000138, -0.00000108, -0.00000066, },
		{ -0.00000005, -0.00000050, -0.00000089, -0.00000119, -0.00000135, -0.00000135, -0.00000119, -0.00000087, -0.00000042, 0.00000012, 0.00000071, 0.00000128, 0.00000179, 0.00000219, 0.00000244, 0.00000252, 0.00000242, 0.00000215, 0.00000173, 0.00000120, 0.00000061, -0.00000001, -0.00000059, -0.00000110, -0.00000150, -0.00000177, -0.00000188, -0.00000185, -0.00000168, -0.00000140, -0.00000102, -0.00000059, -0.00000014, 0.00000029, 0.00000068, 0.00000100, 0.00000121, 0.00000132, 0.00000130, 0.00000114, 0.00000083, 0.00000039, -0.00000019, -0.00000089, -0.00000171, -0.00000262, -0.00000360, -0.00000460, -0.00000559, -0.00000651, -0.00000731, -0.00000793, -0.00000832, -0.00000843, -0.00000824, -0.00000771, -0.00000685, -0.00000568, -0.00000425, -0.00000261, -0.00000086, 0.00000091, 0.00000260, 0.00000411, 0.00000535, 0.00000625, 0.00000674, 0.00000681, 0.00000647, 0.00000575, 0.00000471, 0.00000346, 0.00000208, 0.00000069, -0.00000061, -0.00000171, -0.00000254, -0.00000306, -0.00000323, -0.00000308, -0.00000264, -0.00000197, -0.00000116, -0.00000030, 0.00000052, 0.00000122, 0.00000172, 0.00000199, 0.00000199, 0.00000173, 0.00000124, 0.00000057, -0.00000021, -0.00000104, -0.00000185, -0.00000256, -0.00000312, -0.00000351, -0.00000370, -0.00000371, -0.00000355, -0.00000327, -0.00000291, -0.00000252, -0.00000214, -0.00000182, -0.00000157, -0.00000142, -0.00000135, -0.00000136, -0.00000140, -0.00000146, -0.00000148, -0.00000145, -0.00000134, -0.00000113, -0.00000084, -0.00000047, -0.00000006, 0.00000037, 0.00000076, 0.00000108, 0.00000130, 0.00000138, 0.00000132, 0.00000113, 0.00000081, 0.00000040, },
		{ -0.00000024, 0.00000020, 0.00000056, 0.00000079, 0.00000089, 0.00000083, 0.00000063, 0.00000031, -0.00000011, -0.00000057, -0.00000103, -0.00000143, -0.00000174, -0.00000192, -0.00000196, -0.00000183, -0.00000157, -0.00000118, -0.00000070, -0.00000018, 0.00000034, 0.00000082, 0.00000121, 0.00000150, 0.00000166, 0.00000170, 0.00000162, 0.00000145, 0.00000121, 0.00000094, 0.00000069, 0.00000048, 0.00000035, 0.00000031, 0.00000039, 0.00000058, 0.00000089, 0.00000130, 0.00000179, 0.00000233, 0.00000290, 0.00000347, 0.00000400, 0.00000447, 0.00000484, 0.00000509, 0.00000520, 0.00000512, 0.00000484, 0.00000434, 0.00000360, 0.00000262, 0.00000140, -0.00000005, -0.00000170, -0.00000349, -0.00000538, -0.00000728, -0.00000911, -0.00001077, -0.00001217, -0.00001321, -0.00001382, -0.00001394, -0.00001353, -0.00001258, -0.00001114, -0.00000926, -0.00000703, -0.00000456, -0.00000201, 0.00000051, 0.00000284, 0.00000486, 0.00000647, 0.00000759, 0.00000819, 0.00000827, 0.00000787, 0.00000706, 0.00000596, 0.00000467, 0.00000332, 0.00000203, 0.00000092, 0.00000007, -0.00000048, -0.00000070, -0.00000060, -0.00000023, 0.00000034, 0.00000103, 0.00000174, 0.00000237, 0.00000284, 0.00000309, 0.00000309, 0.00000281, 0.00000227, 0.00000152, 0.00000062, -0.00000037, -0.00000137, -0.00000230, -0.00000309, -0.00000371, -0.00000412, -0.00000431, -0.00000431, -0.00000415, -0.00000387, -0.00000352, -0.00000315, -0.00000282, -0.00000255, -0.00000237, -0.00000229, -0.00000228, -0.00000233, -0.00000241, -0.00000247, -0.00000249, -0.00000243, -0.00000227, -0.00000200, -0.00000164, -0.00000120, -0.00000072, },
		{ -0.00000186, -0.00000207, -0.00000220, -0.00000223, -0.00000214, -0.00000193, -0.00000161, -0.00000119, -0.00000073, -0.00000026, 0.00000017, 0.00000052, 0.00000076, 0.00000086, 0.00000082, 0.00000064, 0.00000035, -0.00000003, -0.00000045, -0.00000088, -0.00000126, -0.00000157, -0.00000178, -0.00000188, -0.00000187, -0.00000175, -0.00000155, -0.00000130, -0.00000104, -0.00000079, -0.00000059, -0.00000047, -0.00000043, -0.00000049, -0.00000062, -0.00000083, -0.00000107, -0.00000133, -0.00000157, -0.00000176, -0.00000188, -0.00000189, -0.00000179, -0.00000157, -0.00000122, -0.00000074, -0.00000014, 0.00000057, 0.00000137, 0.00000225, 0.00000319, 0.00000416, 0.00000515, 0.00000611, 0.00000701, 0.00000781, 0.00000846, 0.00000890, 0.00000908, 0.00000896, 0.00000849, 0.00000765, 0.00000643, 0.00000484, 0.00000292, 0.00000073, -0.00000165, -0.00000409, -0.00000649, -0.00000872, -0.00001064, -0.00001215, -0.00001314, -0.00001356, -0.00001336, -0.00001257, -0.00001124, -0.00000944, -0.00000729, -0.00000494, -0.00000253, -0.00000022, 0.00000185, 0.00000357, 0.00000485, 0.00000565, 0.00000596, 0.00000581, 0.00000527, 0.00000444, 0.00000343, 0.00000236, 0.00000135, 0.00000050, -0.00000011, -0.00000044, -0.00000048, -0.00000024, 0.00000023, 0.00000085, 0.00000155, 0.00000222, 0.00000279, 0.00000318, 0.00000335, 0.00000326, 0.00000292, 0.00000235, 0.00000161, 0.00000075, -0.00000014, -0.00000100, -0.00000175, -0.00000236, -0.00000277, -0.00000298, -0.00000300, -0.00000285, -0.00000258, -0.00000224, -0.00000189, -0.00000157, -0.00000132, -0.00000118, -0.00000115, -0.00000124, -0.00000140, -0.00000163, },
		{ -0.00000117, -0.00000067, -0.00000027, -0.00000001, 0.00000010, 0.00000004, -0.00000015, -0.00000044, -0.00000080, -0.00000117, -0.00000151, -0.00000176, -0.00000190, -0.00000190, -0.00000176, -0.00000150, -0.00000114, -0.00000072, -0.00000027, 0.00000015, 0.00000052, 0.00000080, 0.00000098, 0.00000104, 0.00000100, 0.00000088, 0.00000069, 0.00000049, 0.00000028, 0.00000012, 0.00000002, -0.00000002, 0.00000003, 0.00000014, 0.00000029, 0.00000047, 0.00000065, 0.00000079, 0.00000087, 0.00000087, 0.00000077, 0.00000057, 0.00000029, -0.00000008, -0.00000050, -0.00000096, -0.00000142, -0.00000187, -0.00000229, -0.00000265, -0.00000294, -0.00000315, -0.00000328, -0.00000332, -0.00000326, -0.00000311, -0.00000284, -0.00000246, -0.00000196, -0.00000132, -0.00000055, 0.00000035, 0.00000137, 0.00000247, 0.00000362, 0.00000477, 0.00000583, 0.00000676, 0.00000746, 0.00000788, 0.00000794, 0.00000762, 0.00000689, 0.00000576, 0.00000427, 0.00000250, 0.00000053, -0.00000152, -0.00000352, -0.00000533, -0.00000684, -0.00000794, -0.00000855, -0.00000863, -0.00000817, -0.00000721, -0.00000582, -0.00000411, -0.00000220, -0.00000023, 0.00000165, 0.00000333, 0.00000468, 0.00000564, 0.00000617, 0.00000626, 0.00000594, 0.00000529, 0.00000439, 0.00000336, 0.00000229, 0.00000130, 0.00000047, -0.00000013, -0.00000048, -0.00000058, -0.00000043, -0.00000011, 0.00000032, 0.00000079, 0.00000122, 0.00000152, 0.00000166, 0.00000159, 0.00000132, 0.00000085, 0.00000023, -0.00000048, -0.00000122, -0.00000193, -0.00000252, -0.00000296, -0.00000320, -0.00000323, -0.00000306, -0.00000272, -0.00000226, -0.00000172, },
		{ -0.00000349, -0.00000392, -0.00000412, -0.00000406, -0.00000375, -0.00000324, -0.00000255, -0.00000177, -0.00000096, -0.00000020, 0.00000045, 0.00000094, 0.00000123, 0.00000131, 0.00000118, 0.00000088, 0.00000043, -0.00000010, -0.00000065, -0.00000118, -0.00000163, -0.00000198, -0.00000219, -0.00000227, -0.00000223, -0.00000208, -0.00000186, -0.00000161, -0.00000135, -0.00000113, -0.00000095, -0.00000083, -0.00000077, -0.00000076, -0.00000078, -0.00000079, -0.00000078, -0.00000073, -0.00000060, -0.00000039, -0.00000011, 0.00000024, 0.00000064, 0.00000106, 0.00000148, 0.00000186, 0.00000218, 0.00000242, 0.00000256, 0.00000261, 0.00000256, 0.00000243, 0.00000224, 0.00000201, 0.00000175, 0.00000148, 0.00000121, 0.00000095, 0.00000070, 0.00000045, 0.00000019, -0.00000007, -0.00000035, -0.00000065, -0.00000094, -0.00000123, -0.00000148, -0.00000167, -0.00000177, -0.00000174, -0.00000157, -0.00000124, -0.00000076, -0.00000015, 0.00000056, 0.00000130, 0.00000203, 0.00000266, 0.00000312, 0.00000334, 0.00000328, 0.00000291, 0.00000221, 0.00000122, -0.00000002, -0.00000142, -0.00000288, -0.00000430, -0.00000556, -0.00000654, -0.00000716, -0.00000734, -0.00000706, -0.00000631, -0.00000513, -0.00000358, -0.00000178, 0.00000016, 0.00000212, 0.00000396, 0.00000557, 0.00000686, 0.00000776, 0.00000826, 0.00000834, 0.00000806, 0.00000748, 0.00000669, 0.00000577, 0.00000483, 0.00000396, 0.00000321, 0.00000264, 0.00000226, 0.00000207, 0.00000202, 0.00000208, 0.00000216, 0.00000222, 0.00000218, 0.00000200, 0.00000165, 0.00000113, 0.00000044, -0.00000036, -0.00000122, -0.00000208, -0.00000286, },
		{ 0.00000359, 0.00000299, 0.00000220, 0.00000124, 0.00000016, -0.00000098, -0.00000210, -0.00000311, -0.00000394, -0.00000452, -0.00000479, -0.00000475, -0.00000438, -0.00000372, -0.00000281, -0.00000174, -0.00000057, 0.00000059, 0.00000166, 0.00000258, 0.00000328, 0.00000373, 0.00000392, 0.00000385, 0.00000357, 0.00000312, 0.00000254, 0.00000191, 0.00000127, 0.00000067, 0.00000015, -0.00000028, -0.00000061, -0.00000085, -0.00000103, -0.00000116, -0.00000128, -0.00000141, -0.00000158, -0.00000178, -0.00000202, -0.00000228, -0.00000255, -0.00000281, -0.00000301, -0.00000314, -0.00000317, -0.00000309, -0.00000291, -0.00000262, -0.00000225, -0.00000182, -0.00000136, -0.00000090, -0.00000048, -0.00000010, 0.00000021, 0.00000045, 0.00000062, 0.00000075, 0.00000085, 0.00000094, 0.00000105, 0.00000119, 0.00000136, 0.00000158, 0.00000184, 0.00000211, 0.00000238, 0.00000262, 0.00000281, 0.00000292, 0.00000294, 0.00000287, 0.00000273, 0.00000253, 0.00000229, 0.00000206, 0.00000187, 0.00000174, 0.00000171, 0.00000178, 0.00000194, 0.00000217, 0.00000242, 0.00000266, 0.00000281, 0.00000283, 0.00000264, 0.00000222, 0.00000154, 0.00000059, -0.00000058, -0.00000193, -0.00000339, -0.00000486, -0.00000622, -0.00000739, -0.00000826, -0.00000875, -0.00000882, -0.00000842, -0.00000758, -0.00000635, -0.00000478, -0.00000299, -0.00000108, 0.00000084, 0.00000264, 0.00000422, 0.00000552, 0.00000648, 0.00000710, 0.00000737, 0.00000734, 0.00000708, 0.00000666, 0.00000615, 0.00000564, 0.00000518, 0.00000482, 0.00000457, 0.00000443, 0.00000438, 0.00000436, 0.00000433, 0.00000423, 0.00000399, },
		},
		{ { -0.00000216, -0.00000215, -0.00000213, -0.00000209, -0.00000203, -0.00000196, -0.00000188, -0.00000179, -0.00000170, -0.00000160, -0.00000150, -0.00000141, -0.00000132, -0.00000124, -0.00000118, -0.00000113, -0.00000110, -0.00000109, -0.00000110, -0.00000114, -0.00000119, -0.00000128, -0.00000138, -0.00000150, -0.00000165, -0.00000181, -0.00000199, -0.00000217, -0.00000237, -0.00000257, -0.00000277, -0.00000296, -0.00000315, -0.00000332, -0.00000348, -0.00000362, -0.00000373, -0.00000381, -0.00000386, -0.00000388, -0.00000386, -0.00000381, -0.00000372, -0.00000360, -0.00000344, -0.00000325, -0.00000303, -0.00000278, -0.00000250, -0.00000221, -0.00000190, -0.00000158, -0.00000125, -0.00000092, -0.00000060, -0.00000029, 0.00000000, 0.00000027, 0.00000052, 0.00000074, 0.00000092, 0.00000106, 0.00000117, 0.00000123, 0.00000126, 0.00000123, 0.00000117, 0.00000106, 0.00000092, 0.00000074, 0.00000052, 0.00000027, 0.00000000, -0.00000029, -0.00000060, -0.00000092, -0.00000125, -0.00000158, -0.00000190, -0.00000221, -0.00000250, -0.00000278, -0.00000303, -0.00000325, -0.00000344, -0.00000360, -0.00000372, -0.00000381, -0.00000386, -0.00000388, -0.00000386, -0.00000381, -0.00000373, -0.00000362, -0.00000348, -0.00000332, -0.00000315, -0.00000296, -0.00000277, -0.00000257, -0.00000237, -0.00000217, -0.00000199, -0.00000181, -0.00000165, -0.00000150, -0.00000138, -0.00000128, -0.00000119, -0.00000114, -0.00000110, -0.00000109, -0.00000110, -0.00000113, -0.00000118, -0.00000124, -0.00000132, -0.00000141, -0.00000150, -0.00000160, -0.00000170, -0.00000179, -0.00000188, -0.00000196, -0.00000203, -0.00000209, -0.00000213, -0.00000215, },
		{ -0.00000132, -0.00000138, -0.00000145, -0.00000153, -0.00000162, -0.00000171, -0.00000180, -0.00000190, -0.00000199, -0.00000207, -0.00000214, -0.00000220, -0.00000225, -0.00000227, -0.00000228, -0.00000226, -0.00000222, -0.00000216, -0.00000208, -0.00000197, -0.00000184, -0.00000170, -0.00000153, -0.00000136, -0.00000117, -0.00000097, -0.00000077, -0.00000057, -0.00000038, -0.00000019, -0.00000001, 0.00000015, 0.00000028, 0.00000040, 0.00000049, 0.00000055, 0.00000058, 0.00000057, 0.00000054, 0.00000047, 0.00000036, 0.00000023, 0.00000007, -0.00000012, -0.00000034, -0.00000058, -0.00000083, -0.00000109, -0.00000137, -0.00000164, -0.00000191, -0.00000218, -0.00000243, -0.00000267, -0.00000288, -0.00000307, -0.00000322, -0.00000335, -0.00000343, -0.00000348, -0.00000350, -0.00000347, -0.00000340, -0.00000329, -0.00000315, -0.00000297, -0.00000275, -0.00000251, -0.00000225, -0.00000196, -0.00000166, -0.00000135, -0.00000104, -0.00000072, -0.00000041, -0.00000011, 0.00000017, 0.00000044, 0.00000067, 0.00000089, 0.00000107, 0.00000121, 0.00000132, 0.00000140, 0.00000143, 0.00000143, 0.00000139, 0.00000132, 0.00000121, 0.00000108, 0.00000092, 0.00000073, 0.00000053, 0.00000030, 0.00000007, -0.00000016, -0.00000040, -0.00000064, -0.00000087, -0.00000109, -0.00000130, -0.00000149, -0.00000165, -0.00000180, -0.00000193, -0.00000202, -0.00000210, -0.00000215, -0.00000217, -0.00000218, -0.00000216, -0.00000212, -0.00000207, -0.00000200, -0.00000192, -0.00000183, -0.00000175, -0.00000166, -0.00000157, -0.00000149, -0.00000141, -0.00000135, -0.00000130, -0.00000126, -0.00000124, -0.00000124, -0.00000125, -0.00000128, },
		{ 0.00000007, 0.00000012, 0.00000019, 0.00000025, 0.00000031, 0.00000037, 0.00000042, 0.00000047, 0.00000049, 0.00000051, 0.00000050, 0.00000048, 0.00000044, 0.00000038, 0.00000029, 0.00000019, 0.00000007, -0.00000008, -0.00000024, -0.00000041, -0.00000060, -0.00000079, -0.00000099, -0.00000119, -0.00000139, -0.00000158, -0.00000176, -0.00000193, -0.00000208, -0.00000221, -0.00000232, -0.00000240, -0.00000245, -0.00000248, -0.00000247, -0.00000242, -0.00000235, -0.00000225, -0.00000212, -0.00000195, -0.00000177, -0.00000156, -0.00000133, -0.00000109, -0.00000084, -0.00000058, -0.00000032, -0.00000007, 0.00000018, 0.00000042, 0.00000063, 0.00000083, 0.00000101, 0.00000115, 0.00000127, 0.00000135, 0.00000140, 0.00000141, 0.00000139, 0.00000134, 0.00000125, 0.00000113, 0.00000098, 0.00000081, 0.00000061, 0.00000040, 0.00000017, -0.00000007, -0.00000032, -0.00000057, -0.00000081, -0.00000104, -0.00000126, -0.00000146, -0.00000164, -0.00000179, -0.00000192, -0.00000202, -0.00000208, -0.00000211, -0.00000211, -0.00000207, -0.00000200, -0.00000190, -0.00000177, -0.00000162, -0.00000144, -0.00000124, -0.00000102, -0.00000079, -0.00000055, -0.00000031, -0.00000007, 0.00000017, 0.00000039, 0.00000061, 0.00000081, 0.00000099, 0.00000115, 0.00000128, 0.00000139, 0.00000148, 0.00000153, 0.00000156, 0.00000157, 0.00000155, 0.00000150, 0.00000144, 0.00000136, 0.00000126, 0.00000115, 0.00000102, 0.00000090, 0.00000077, 0.00000064, 0.00000052, 0.00000040, 0.00000029, 0.00000019, 0.00000011, 0.00000004, -0.00000001, -0.00000004, -0.00000006, -0.00000006, -0.00000005, -0.00000002, 0.00000002, },
		{ 0.00000048, 0.00000039, 0.00000030, 0.00000022, 0.00000015, 0.00000008, 0.00000004, 0.00000000, -0.00000001, -0.00000001, 0.00000001, 0.00000005, 0.00000010, 0.00000017, 0.00000026, 0.00000036, 0.00000047, 0.00000059, 0.00000072, 0.00000084, 0.00000097, 0.00000109, 0.00000120, 0.00000130, 0.00000139, 0.00000145, 0.00000150, 0.00000152, 0.00000152, 0.00000149, 0.00000143, 0.00000135, 0.00000124, 0.00000110, 0.00000093, 0.00000074, 0.00000053, 0.00000030, 0.00000006, -0.00000020, -0.00000046, -0.00000073, -0.00000099, -0.00000125, -0.00000151, -0.00000174, -0.00000196, -0.00000216, -0.00000233, -0.00000248, -0.00000259, -0.00000268, -0.00000273, -0.00000274, -0.00000272, -0.00000267, -0.00000258, -0.00000246, -0.00000231, -0.00000214, -0.00000195, -0.00000173, -0.00000150, -0.00000126, -0.00000102, -0.00000077, -0.00000052, -0.00000029, -0.00000006, 0.00000015, 0.00000034, 0.00000051, 0.00000066, 0.00000077, 0.00000086, 0.00000092, 0.00000095, 0.00000095, 0.00000092, 0.00000086, 0.00000078, 0.00000067, 0.00000054, 0.00000039, 0.00000024, 0.00000007, -0.00000011, -0.00000028, -0.00000046, -0.00000063, -0.00000079, -0.00000093, -0.00000106, -0.00000117, -0.00000126, -0.00000133, -0.00000137, -0.00000139, -0.00000138, -0.00000134, -0.00000129, -0.00000121, -0.00000111, -0.00000099, -0.00000085, -0.00000070, -0.00000055, -0.00000038, -0.00000021, -0.00000005, 0.00000011, 0.00000027, 0.00000042, 0.00000055, 0.00000067, 0.00000077, 0.00000085, 0.00000091, 0.00000096, 0.00000098, 0.00000098, 0.00000097, 0.00000093, 0.00000088, 0.00000082, 0.00000075, 0.00000066, 0.00000057, },
		{ 0.00000077, 0.00000086, 0.00000094, 0.00000099, 0.00000103, 0.00000105, 0.00000105, 0.00000102, 0.00000098, 0.00000092, 0.00000084, 0.00000075, 0.00000064, 0.00000053, 0.00000040, 0.00000027, 0.00000014, 0.00000001, -0.00000011, -0.00000022, -0.00000032, -0.00000041, -0.00000048, -0.00000053, -0.00000056, -0.00000057, -0.00000055, -0.00000051, -0.00000045, -0.00000036, -0.00000025, -0.00000012, 0.00000002, 0.00000018, 0.00000035, 0.00000052, 0.00000070, 0.00000088, 0.00000106, 0.00000122, 0.00000138, 0.00000152, 0.00000163, 0.00000173, 0.00000180, 0.00000184, 0.00000185, 0.00000183, 0.00000177, 0.00000169, 0.00000157, 0.00000142, 0.00000124, 0.00000104, 0.00000081, 0.00000056, 0.00000030, 0.00000002, -0.00000026, -0.00000055, -0.00000084, -0.00000112, -0.00000139, -0.00000165, -0.00000188, -0.00000210, -0.00000229, -0.00000244, -0.00000257, -0.00000266, -0.00000272, -0.00000275, -0.00000273, -0.00000269, -0.00000261, -0.00000249, -0.00000235, -0.00000219, -0.00000200, -0.00000179, -0.00000157, -0.00000133, -0.00000109, -0.00000085, -0.00000061, -0.00000038, -0.00000016, 0.00000005, 0.00000024, 0.00000041, 0.00000056, 0.00000068, 0.00000077, 0.00000084, 0.00000089, 0.00000090, 0.00000089, 0.00000086, 0.00000081, 0.00000073, 0.00000064, 0.00000054, 0.00000042, 0.00000030, 0.00000018, 0.00000006, -0.00000006, -0.00000018, -0.00000028, -0.00000037, -0.00000044, -0.00000050, -0.00000054, -0.00000056, -0.00000056, -0.00000054, -0.00000050, -0.00000044, -0.00000037, -0.00000028, -0.00000017, -0.00000006, 0.00000006, 0.00000019, 0.00000031, 0.00000044, 0.00000056, 0.00000067, },
		{ 0.00000029, 0.00000032, 0.00000037, 0.00000043, 0.00000052, 0.00000062, 0.00000074, 0.00000087, 0.00000101, 0.00000116, 0.00000131, 0.00000147, 0.00000163, 0.00000178, 0.00000193, 0.00000206, 0.00000218, 0.00000228, 0.00000237, 0.00000243, 0.00000246, 0.00000248, 0.00000246, 0.00000242, 0.00000236, 0.00000227, 0.00000216, 0.00000203, 0.00000188, 0.00000172, 0.00000154, 0.00000136, 0.00000118, 0.00000099, 0.00000082, 0.00000065, 0.00000049, 0.00000035, 0.00000023, 0.00000014, 0.00000007, 0.00000003, 0.00000003, 0.00000005, 0.00000010, 0.00000019, 0.00000030, 0.00000045, 0.00000062, 0.00000081, 0.00000103, 0.00000126, 0.00000150, 0.00000175, 0.00000200, 0.00000224, 0.00000248, 0.00000271, 0.00000292, 0.00000311, 0.00000326, 0.00000339, 0.00000349, 0.00000355, 0.00000357, 0.00000355, 0.00000349, 0.00000339, 0.00000325, 0.00000307, 0.00000285, 0.00000260, 0.00000233, 0.00000203, 0.00000170, 0.00000136, 0.00000101, 0.00000066, 0.00000030, -0.00000005, -0.00000039, -0.00000072, -0.00000102, -0.00000130, -0.00000156, -0.00000178, -0.00000196, -0.00000212, -0.00000223, -0.00000230, -0.00000234, -0.00000234, -0.00000231, -0.00000224, -0.00000213, -0.00000200, -0.00000185, -0.00000167, -0.00000148, -0.00000128, -0.00000106, -0.00000085, -0.00000063, -0.00000042, -0.00000022, -0.00000003, 0.00000015, 0.00000031, 0.00000044, 0.00000056, 0.00000066, 0.00000073, 0.00000078, 0.00000081, 0.00000082, 0.00000081, 0.00000078, 0.00000074, 0.00000069, 0.00000064, 0.00000057, 0.00000051, 0.00000045, 0.00000040, 0.00000035, 0.00000031, 0.00000029, 0.00000028, },
		{ 0.00000031, 0.00000032, 0.00000031, 0.00000029, 0.00000025, 0.00000021, 0.00000016, 0.00000010, 0.00000004, -0.00000001, -0.00000006, -0.00000010, -0.00000012, -0.00000013, -0.00000012, -0.00000009, -0.00000004, 0.00000004, 0.00000014, 0.00000027, 0.00000042, 0.00000060, 0.00000079, 0.00000101, 0.00000124, 0.00000148, 0.00000173, 0.00000199, 0.00000224, 0.00000250, 0.00000274, 0.00000296, 0.00000317, 0.00000336, 0.00000351, 0.00000364, 0.00000373, 0.00000379, 0.00000381, 0.00000378, 0.00000372, 0.00000362, 0.00000349, 0.00000332, 0.00000311, 0.00000287, 0.00000261, 0.00000233, 0.00000204, 0.00000173, 0.00000142, 0.00000110, 0.00000080, 0.00000051, 0.00000024, -0.00000001, -0.00000023, -0.00000041, -0.00000056, -0.00000066, -0.00000073, -0.00000075, -0.00000072, -0.00000065, -0.00000053, -0.00000037, -0.00000018, 0.00000006, 0.00000033, 0.00000062, 0.00000094, 0.00000127, 0.00000162, 0.00000196, 0.00000231, 0.00000265, 0.00000297, 0.00000327, 0.00000355, 0.00000379, 0.00000400, 0.00000417, 0.00000430, 0.00000438, 0.00000442, 0.00000441, 0.00000435, 0.00000425, 0.00000411, 0.00000392, 0.00000370, 0.00000344, 0.00000315, 0.00000284, 0.00000251, 0.00000217, 0.00000182, 0.00000146, 0.00000112, 0.00000078, 0.00000045, 0.00000014, -0.00000014, -0.00000039, -0.00000062, -0.00000082, -0.00000098, -0.00000110, -0.00000120, -0.00000125, -0.00000128, -0.00000127, -0.00000124, -0.00000118, -0.00000110, -0.00000100, -0.00000088, -0.00000076, -0.00000062, -0.00000049, -0.00000036, -0.00000023, -0.00000011, -0.00000000, 0.00000009, 0.00000017, 0.00000024, 0.00000028, },
		{ -0.00000022, -0.00000015, -0.00000006, 0.00000004, 0.00000015, 0.00000027, 0.00000039, 0.00000050, 0.00000061, 0.00000070, 0.00000078, 0.00000085, 0.00000089, 0.00000090, 0.00000089, 0.00000086, 0.00000080, 0.00000071, 0.00000060, 0.00000047, 0.00000032, 0.00000015, -0.00000004, -0.00000023, -0.00000042, -0.00000061, -0.00000080, -0.00000098, -0.00000113, -0.00000127, -0.00000138, -0.00000146, -0.00000151, -0.00000151, -0.00000148, -0.00000141, -0.00000130, -0.00000114, -0.00000094, -0.00000070, -0.00000043, -0.00000013, 0.00000021, 0.00000057, 0.00000094, 0.00000133, 0.00000172, 0.00000212, 0.00000250, 0.00000287, 0.00000321, 0.00000353, 0.00000381, 0.00000405, 0.00000425, 0.00000440, 0.00000449, 0.00000453, 0.00000451, 0.00000444, 0.00000430, 0.00000412, 0.00000388, 0.00000359, 0.00000326, 0.00000288, 0.00000248, 0.00000205, 0.00000159, 0.00000113, 0.00000066, 0.00000019, -0.00000027, -0.00000071, -0.00000113, -0.00000151, -0.00000186, -0.00000217, -0.00000243, -0.00000264, -0.00000279, -0.00000290, -0.00000294, -0.00000294, -0.00000288, -0.00000277, -0.00000262, -0.00000242, -0.00000218, -0.00000191, -0.00000161, -0.00000129, -0.00000096, -0.00000062, -0.00000028, 0.00000006, 0.00000038, 0.00000069, 0.00000098, 0.00000123, 0.00000146, 0.00000165, 0.00000181, 0.00000193, 0.00000201, 0.00000204, 0.00000205, 0.00000201, 0.00000194, 0.00000185, 0.00000172, 0.00000157, 0.00000141, 0.00000123, 0.00000105, 0.00000086, 0.00000067, 0.00000049, 0.00000032, 0.00000016, 0.00000002, -0.00000009, -0.00000019, -0.00000026, -0.00000030, -0.00000032, -0.00000031, -0.00000028, },
		{ -0.00000044, -0.00000063, -0.00000081, -0.00000100, -0.00000118, -0.00000135, -0.00000150, -0.00000164, -0.00000175, -0.00000183, -0.00000188, -0.00000190, -0.00000189, -0.00000184, -0.00000176, -0.00000165, -0.00000151, -0.00000135, -0.00000116, -0.00000095, -0.00000073, -0.00000050, -0.00000027, -0.00000004, 0.00000018, 0.00000038, 0.00000057, 0.00000073, 0.00000085, 0.00000094, 0.00000100, 0.00000101, 0.00000097, 0.00000089, 0.00000077, 0.00000060, 0.00000039, 0.00000013, -0.00000016, -0.00000048, -0.00000082, -0.00000119, -0.00000157, -0.00000195, -0.00000234, -0.00000271, -0.00000307, -0.00000340, -0.00000370, -0.00000396, -0.00000417, -0.00000434, -0.00000445, -0.00000449, -0.00000448, -0.00000440, -0.00000426, -0.00000405, -0.00000378, -0.00000345, -0.00000306, -0.00000263, -0.00000215, -0.00000163, -0.00000107, -0.00000050, 0.00000009, 0.00000068, 0.00000127, 0.00000185, 0.00000241, 0.00000294, 0.00000343, 0.00000388, 0.00000428, 0.00000463, 0.00000492, 0.00000514, 0.00000530, 0.00000539, 0.00000542, 0.00000539, 0.00000529, 0.00000514, 0.00000493, 0.00000467, 0.00000437, 0.00000403, 0.00000367, 0.00000328, 0.00000288, 0.00000247, 0.00000206, 0.00000166, 0.00000127, 0.00000091, 0.00000057, 0.00000026, -0.00000002, -0.00000025, -0.00000045, -0.00000061, -0.00000073, -0.00000080, -0.00000084, -0.00000084, -0.00000081, -0.00000075, -0.00000067, -0.00000056, -0.00000044, -0.00000032, -0.00000019, -0.00000006, 0.00000007, 0.00000018, 0.00000027, 0.00000035, 0.00000040, 0.00000043, 0.00000044, 0.00000041, 0.00000036, 0.00000028, 0.00000017, 0.00000005, -0.00000010, -0.00000026, },
		{ 0.00000038, 0.00000047, 0.00000055, 0.00000061, 0.00000067, 0.00000070, 0.00000070, 0.00000069, 0.00000064, 0.00000056, 0.00000045, 0.00000032, 0.00000015, -0.00000004, -0.00000026, -0.00000049, -0.00000074, -0.00000100, -0.00000126, -0.00000153, -0.00000178, -0.00000202, -0.00000223, -0.00000242, -0.00000258, -0.00000270, -0.00000277, -0.00000280, -0.00000278, -0.00000270, -0.00000257, -0.00000238, -0.00000214, -0.00000184, -0.00000150, -0.00000111, -0.00000069, -0.00000022, 0.00000027, 0.00000077, 0.00000130, 0.00000182, 0.00000234, 0.00000284, 0.00000333, 0.00000378, 0.00000419, 0.00000455, 0.00000486, 0.00000512, 0.00000530, 0.00000542, 0.00000547, 0.00000544, 0.00000535, 0.00000518, 0.00000494, 0.00000463, 0.00000427, 0.00000385, 0.00000338, 0.00000288, 0.00000234, 0.00000178, 0.00000120, 0.00000062, 0.00000005, -0.00000051, -0.00000104, -0.00000154, -0.00000201, -0.00000242, -0.00000278, -0.00000308, -0.00000332, -0.00000349, -0.00000359, -0.00000363, -0.00000359, -0.00000350, -0.00000334, -0.00000312, -0.00000285, -0.00000254, -0.00000218, -0.00000179, -0.00000138, -0.00000095, -0.00000052, -0.00000008, 0.00000034, 0.00000076, 0.00000115, 0.00000151, 0.00000183, 0.00000212, 0.00000236, 0.00000256, 0.00000271, 0.00000282, 0.00000287, 0.00000288, 0.00000284, 0.00000277, 0.00000265, 0.00000251, 0.00000233, 0.00000213, 0.00000192, 0.00000170, 0.00000147, 0.00000124, 0.00000102, 0.00000081, 0.00000062, 0.00000045, 0.00000030, 0.00000018, 0.00000008, 0.00000001, -0.00000003, -0.00000005, -0.00000003, -0.00000000, 0.00000005, 0.00000012, 0.00000020, 0.00000029, },
		{ 0.00000112, 0.00000103, 0.00000096, 0.00000091, 0.00000088, 0.00000088, 0.00000090, 0.00000095, 0.00000102, 0.00000113, 0.00000126, 0.00000141, 0.00000158, 0.00000177, 0.00000197, 0.00000217, 0.00000237, 0.00000257, 0.00000276, 0.00000293, 0.00000307, 0.00000318, 0.00000326, 0.00000329, 0.00000328, 0.00000322, 0.00000311, 0.00000295, 0.00000273, 0.00000246, 0.00000214, 0.00000177, 0.00000135, 0.00000089, 0.00000040, -0.00000012, -0.00000066, -0.00000121, -0.00000176, -0.00000232, -0.00000285, -0.00000337, -0.00000385, -0.00000429, -0.00000468, -0.00000502, -0.00000529, -0.00000550, -0.00000564, -0.00000570, -0.00000568, -0.00000559, -0.00000542, -0.00000517, -0.00000485, -0.00000447, -0.00000402, -0.00000352, -0.00000297, -0.00000239, -0.00000177, -0.00000114, -0.00000049, 0.00000015, 0.00000078, 0.00000139, 0.00000198, 0.00000252, 0.00000302, 0.00000346, 0.00000385, 0.00000417, 0.00000442, 0.00000460, 0.00000471, 0.00000475, 0.00000471, 0.00000461, 0.00000445, 0.00000423, 0.00000395, 0.00000363, 0.00000327, 0.00000288, 0.00000246, 0.00000204, 0.00000160, 0.00000117, 0.00000076, 0.00000036, -0.00000002, -0.00000036, -0.00000067, -0.00000093, -0.00000114, -0.00000131, -0.00000143, -0.00000150, -0.00000152, -0.00000149, -0.00000142, -0.00000131, -0.00000116, -0.00000097, -0.00000077, -0.00000054, -0.00000029, -0.00000004, 0.00000021, 0.00000046, 0.00000070, 0.00000093, 0.00000113, 0.00000132, 0.00000147, 0.00000160, 0.00000170, 0.00000177, 0.00000181, 0.00000182, 0.00000180, 0.00000176, 0.00000170, 0.00000162, 0.00000153, 0.00000142, 0.00000132, 0.00000122, },
		{ 0.00000052, 0.00000058, 0.00000061, 0.00000062, 0.00000059, 0.00000054, 0.00000045, 0.00000033, 0.00000019, 0.00000002, -0.00000018, -0.00000039, -0.00000062, -0.00000085, -0.00000109, -0.00000132, -0.00000155, -0.00000176, -0.00000195, -0.00000211, -0.00000223, -0.00000232, -0.00000237, -0.00000237, -0.00000233, -0.00000224, -0.00000209, -0.00000190, -0.00000166, -0.00000138, -0.00000106, -0.00000070, -0.00000031, 0.00000011, 0.00000054, 0.00000098, 0.00000142, 0.00000185, 0.00000227, 0.00000267, 0.00000303, 0.00000335, 0.00000363, 0.00000385, 0.00000402, 0.00000412, 0.00000415, 0.00000412, 0.00000401, 0.00000384, 0.00000359, 0.00000328, 0.00000291, 0.00000248, 0.00000200, 0.00000147, 0.00000091, 0.00000031, -0.00000030, -0.00000092, -0.00000154, -0.00000215, -0.00000275, -0.00000331, -0.00000384, -0.00000433, -0.00000476, -0.00000513, -0.00000544, -0.00000569, -0.00000586, -0.00000596, -0.00000600, -0.00000595, -0.00000584, -0.00000567, -0.00000543, -0.00000514, -0.00000479, -0.00000441, -0.00000399, -0.00000354, -0.00000307, -0.00000259, -0.00000212, -0.00000164, -0.00000118, -0.00000075, -0.00000034, 0.00000004, 0.00000038, 0.00000067, 0.00000092, 0.00000112, 0.00000127, 0.00000137, 0.00000143, 0.00000144, 0.00000141, 0.00000134, 0.00000123, 0.00000110, 0.00000094, 0.00000077, 0.00000058, 0.00000039, 0.00000019, 0.00000001, -0.00000017, -0.00000033, -0.00000047, -0.00000059, -0.00000068, -0.00000074, -0.00000078, -0.00000078, -0.00000076, -0.00000071, -0.00000064, -0.00000055, -0.00000044, -0.00000032, -0.00000018, -0.00000005, 0.00000009, 0.00000022, 0.00000033, 0.00000044, },
		{ -0.00000219, -0.00000216, -0.00000210, -0.00000201, -0.00000188, -0.00000173, -0.00000155, -0.00000134, -0.00000111, -0.00000087, -0.00000062, -0.00000036, -0.00000010, 0.00000016, 0.00000040, 0.00000063, 0.00000083, 0.00000100, 0.00000114, 0.00000124, 0.00000129, 0.00000131, 0.00000127, 0.00000119, 0.00000105, 0.00000087, 0.00000064, 0.00000037, 0.00000005, -0.00000029, -0.00000067, -0.00000108, -0.00000150, -0.00000193, -0.00000236, -0.00000279, -0.00000321, -0.00000360, -0.00000397, -0.00000430, -0.00000459, -0.00000483, -0.00000502, -0.00000516, -0.00000523, -0.00000524, -0.00000518, -0.00000507, -0.00000489, -0.00000465, -0.00000436, -0.00000401, -0.00000362, -0.00000319, -0.00000272, -0.00000223, -0.00000172, -0.00000120, -0.00000069, -0.00000018, 0.00000032, 0.00000079, 0.00000123, 0.00000163, 0.00000199, 0.00000229, 0.00000255, 0.00000274, 0.00000288, 0.00000296, 0.00000298, 0.00000293, 0.00000283, 0.00000268, 0.00000248, 0.00000223, 0.00000194, 0.00000162, 0.00000127, 0.00000091, 0.00000053, 0.00000014, -0.00000024, -0.00000061, -0.00000097, -0.00000131, -0.00000162, -0.00000190, -0.00000214, -0.00000235, -0.00000252, -0.00000264, -0.00000273, -0.00000277, -0.00000278, -0.00000275, -0.00000268, -0.00000259, -0.00000247, -0.00000233, -0.00000218, -0.00000202, -0.00000185, -0.00000168, -0.00000152, -0.00000137, -0.00000123, -0.00000111, -0.00000101, -0.00000093, -0.00000088, -0.00000085, -0.00000085, -0.00000088, -0.00000093, -0.00000100, -0.00000110, -0.00000120, -0.00000132, -0.00000145, -0.00000158, -0.00000171, -0.00000183, -0.00000194, -0.00000204, -0.00000211, -0.00000217, -0.00000220, },
		{ -0.00000056, -0.00000074, -0.00000094, -0.00000117, -0.00000141, -0.00000167, -0.00000194, -0.00000222, -0.00000249, -0.00000277, -0.00000303, -0.00000328, -0.00000351, -0.00000371, -0.00000388, -0.00000401, -0.00000410, -0.00000415, -0.00000416, -0.00000411, -0.00000402, -0.00000387, -0.00000368, -0.00000343, -0.00000315, -0.00000282, -0.00000245, -0.00000205, -0.00000162, -0.00000117, -0.00000070, -0.00000023, 0.00000025, 0.00000072, 0.00000117, 0.00000160, 0.00000201, 0.00000238, 0.00000271, 0.00000300, 0.00000323, 0.00000341, 0.00000354, 0.00000361, 0.00000362, 0.00000357, 0.00000346, 0.00000330, 0.00000308, 0.00000282, 0.00000251, 0.00000217, 0.00000179, 0.00000139, 0.00000097, 0.00000054, 0.00000010, -0.00000033, -0.00000075, -0.00000115, -0.00000153, -0.00000188, -0.00000220, -0.00000247, -0.00000271, -0.00000290, -0.00000304, -0.00000314, -0.00000318, -0.00000318, -0.00000314, -0.00000305, -0.00000292, -0.00000276, -0.00000256, -0.00000234, -0.00000210, -0.00000184, -0.00000158, -0.00000131, -0.00000104, -0.00000077, -0.00000052, -0.00000028, -0.00000006, 0.00000013, 0.00000030, 0.00000044, 0.00000055, 0.00000063, 0.00000069, 0.00000071, 0.00000071, 0.00000068, 0.00000063, 0.00000056, 0.00000048, 0.00000038, 0.00000027, 0.00000016, 0.00000004, -0.00000007, -0.00000018, -0.00000027, -0.00000036, -0.00000043, -0.00000049, -0.00000053, -0.00000055, -0.00000056, -0.00000055, -0.00000052, -0.00000049, -0.00000044, -0.00000038, -0.00000032, -0.00000025, -0.00000019, -0.00000013, -0.00000009, -0.00000005, -0.00000003, -0.00000003, -0.00000006, -0.00000010, -0.00000018, -0.00000028, -0.00000040, },
		{ 0.00000018, 0.00000034, 0.00000052, 0.00000069, 0.00000086, 0.00000103, 0.00000119, 0.00000133, 0.00000146, 0.00000156, 0.00000164, 0.00000168, 0.00000170, 0.00000168, 0.00000162, 0.00000152, 0.00000139, 0.00000121, 0.00000100, 0.00000075, 0.00000046, 0.00000014, -0.00000021, -0.00000059, -0.00000099, -0.00000140, -0.00000182, -0.00000225, -0.00000268, -0.00000310, -0.00000351, -0.00000389, -0.00000425, -0.00000458, -0.00000488, -0.00000513, -0.00000533, -0.00000549, -0.00000560, -0.00000566, -0.00000566, -0.00000561, -0.00000550, -0.00000534, -0.00000513, -0.00000488, -0.00000458, -0.00000424, -0.00000387, -0.00000347, -0.00000305, -0.00000261, -0.00000216, -0.00000170, -0.00000125, -0.00000080, -0.00000037, 0.00000004, 0.00000043, 0.00000079, 0.00000112, 0.00000141, 0.00000167, 0.00000188, 0.00000205, 0.00000219, 0.00000228, 0.00000233, 0.00000234, 0.00000232, 0.00000226, 0.00000217, 0.00000206, 0.00000192, 0.00000177, 0.00000160, 0.00000142, 0.00000124, 0.00000106, 0.00000087, 0.00000069, 0.00000053, 0.00000037, 0.00000023, 0.00000010, -0.00000001, -0.00000011, -0.00000018, -0.00000024, -0.00000028, -0.00000031, -0.00000032, -0.00000032, -0.00000030, -0.00000028, -0.00000026, -0.00000023, -0.00000019, -0.00000016, -0.00000014, -0.00000011, -0.00000010, -0.00000009, -0.00000009, -0.00000010, -0.00000012, -0.00000015, -0.00000019, -0.00000024, -0.00000029, -0.00000034, -0.00000040, -0.00000046, -0.00000052, -0.00000057, -0.00000061, -0.00000065, -0.00000067, -0.00000068, -0.00000067, -0.00000065, -0.00000061, -0.00000055, -0.00000047, -0.00000037, -0.00000026, -0.00000012, 0.00000002, },
		{ 0.00000028, 0.00000024, 0.00000019, 0.00000015, 0.00000012, 0.00000010, 0.00000010, 0.00000010, 0.00000013, 0.00000017, 0.00000024, 0.00000032, 0.00000042, 0.00000054, 0.00000068, 0.00000083, 0.00000100, 0.00000118, 0.00000137, 0.00000157, 0.00000178, 0.00000198, 0.00000217, 0.00000236, 0.00000254, 0.00000270, 0.00000284, 0.00000296, 0.00000305, 0.00000311, 0.00000314, 0.00000313, 0.00000308, 0.00000300, 0.00000288, 0.00000272, 0.00000252, 0.00000229, 0.00000201, 0.00000171, 0.00000137, 0.00000101, 0.00000062, 0.00000021, -0.00000021, -0.00000064, -0.00000108, -0.00000152, -0.00000196, -0.00000239, -0.00000280, -0.00000320, -0.00000358, -0.00000393, -0.00000426, -0.00000455, -0.00000481, -0.00000504, -0.00000523, -0.00000539, -0.00000550, -0.00000559, -0.00000563, -0.00000565, -0.00000563, -0.00000558, -0.00000551, -0.00000541, -0.00000529, -0.00000516, -0.00000501, -0.00000485, -0.00000468, -0.00000451, -0.00000434, -0.00000417, -0.00000400, -0.00000384, -0.00000369, -0.00000354, -0.00000341, -0.00000329, -0.00000318, -0.00000308, -0.00000299, -0.00000291, -0.00000284, -0.00000278, -0.00000272, -0.00000266, -0.00000261, -0.00000256, -0.00000251, -0.00000245, -0.00000239, -0.00000232, -0.00000225, -0.00000216, -0.00000207, -0.00000197, -0.00000186, -0.00000174, -0.00000161, -0.00000147, -0.00000133, -0.00000118, -0.00000102, -0.00000087, -0.00000071, -0.00000056, -0.00000041, -0.00000027, -0.00000014, -0.00000001, 0.00000010, 0.00000020, 0.00000029, 0.00000036, 0.00000042, 0.00000046, 0.00000049, 0.00000050, 0.00000050, 0.00000048, 0.00000046, 0.00000042, 0.00000038, 0.00000033, },
		{ -0.00000017, -0.00000015, -0.00000014, -0.00000014, -0.00000015, -0.00000017, -0.00000021, -0.00000026, -0.00000033, -0.00000041, -0.00000049, -0.00000059, -0.00000070, -0.00000081, -0.00000092, -0.00000104, -0.00000116, -0.00000127, -0.00000138, -0.00000147, -0.00000156, -0.00000163, -0.00000169, -0.00000173, -0.00000175, -0.00000176, -0.00000174, -0.00000170, -0.00000164, -0.00000157, -0.00000147, -0.00000136, -0.00000123, -0.00000109, -0.00000094, -0.00000079, -0.00000063, -0.00000047, -0.00000031, -0.00000016, -0.00000002, 0.00000011, 0.00000022, 0.00000032, 0.00000039, 0.00000044, 0.00000047, 0.00000047, 0.00000044, 0.00000039, 0.00000031, 0.00000021, 0.00000008, -0.00000008, -0.00000025, -0.00000044, -0.00000064, -0.00000086, -0.00000108, -0.00000131, -0.00000154, -0.00000176, -0.00000198, -0.00000218, -0.00000237, -0.00000254, -0.00000269, -0.00000281, -0.00000291, -0.00000298, -0.00000301, -0.00000302, -0.00000300, -0.00000294, -0.00000286, -0.00000275, -0.00000261, -0.00000245, -0.00000226, -0.00000206, -0.00000184, -0.00000161, -0.00000138, -0.00000113, -0.00000089, -0.00000065, -0.00000042, -0.00000020, 0.00000001, 0.00000020, 0.00000037, 0.00000052, 0.00000065, 0.00000076, 0.00000084, 0.00000090, 0.00000093, 0.00000094, 0.00000093, 0.00000090, 0.00000085, 0.00000078, 0.00000070, 0.00000061, 0.00000051, 0.00000040, 0.00000029, 0.00000018, 0.00000008, -0.00000003, -0.00000012, -0.00000021, -0.00000028, -0.00000035, -0.00000040, -0.00000044, -0.00000046, -0.00000048, -0.00000048, -0.00000047, -0.00000045, -0.00000042, -0.00000039, -0.00000035, -0.00000031, -0.00000027, -0.00000023, -0.00000019, },
		{ -0.00000026, -0.00000023, -0.00000018, -0.00000012, -0.00000006, 0.00000002, 0.00000011, 0.00000021, 0.00000032, 0.00000043, 0.00000055, 0.00000067, 0.00000079, 0.00000091, 0.00000103, 0.00000113, 0.00000123, 0.00000132, 0.00000139, 0.00000146, 0.00000150, 0.00000154, 0.00000156, 0.00000157, 0.00000157, 0.00000155, 0.00000153, 0.00000150, 0.00000147, 0.00000144, 0.00000141, 0.00000139, 0.00000137, 0.00000137, 0.00000139, 0.00000143, 0.00000149, 0.00000157, 0.00000168, 0.00000182, 0.00000199, 0.00000219, 0.00000241, 0.00000267, 0.00000295, 0.00000326, 0.00000358, 0.00000393, 0.00000428, 0.00000465, 0.00000502, 0.00000538, 0.00000574, 0.00000608, 0.00000640, 0.00000669, 0.00000695, 0.00000717, 0.00000735, 0.00000748, 0.00000756, 0.00000757, 0.00000754, 0.00000743, 0.00000727, 0.00000705, 0.00000676, 0.00000641, 0.00000601, 0.00000555, 0.00000505, 0.00000450, 0.00000391, 0.00000329, 0.00000264, 0.00000198, 0.00000130, 0.00000062, -0.00000005, -0.00000071, -0.00000136, -0.00000197, -0.00000255, -0.00000310, -0.00000360, -0.00000405, -0.00000445, -0.00000479, -0.00000507, -0.00000530, -0.00000547, -0.00000557, -0.00000562, -0.00000562, -0.00000556, -0.00000546, -0.00000531, -0.00000512, -0.00000489, -0.00000464, -0.00000437, -0.00000407, -0.00000377, -0.00000346, -0.00000314, -0.00000283, -0.00000253, -0.00000224, -0.00000197, -0.00000171, -0.00000148, -0.00000127, -0.00000108, -0.00000092, -0.00000078, -0.00000066, -0.00000057, -0.00000049, -0.00000044, -0.00000040, -0.00000037, -0.00000035, -0.00000034, -0.00000033, -0.00000032, -0.00000032, -0.00000031, -0.00000029, },
		{ -0.00000033, -0.00000013, 0.00000006, 0.00000022, 0.00000037, 0.00000051, 0.00000063, 0.00000075, 0.00000085, 0.00000095, 0.00000104, 0.00000113, 0.00000122, 0.00000132, 0.00000142, 0.00000153, 0.00000164, 0.00000177, 0.00000190, 0.00000204, 0.00000219, 0.00000235, 0.00000251, 0.00000267, 0.00000283, 0.00000299, 0.00000314, 0.00000328, 0.00000340, 0.00000350, 0.00000357, 0.00000362, 0.00000363, 0.00000361, 0.00000355, 0.00000345, 0.00000331, 0.00000312, 0.00000289, 0.00000262, 0.00000231, 0.00000196, 0.00000158, 0.00000118, 0.00000074, 0.00000030, -0.00000016, -0.00000062, -0.00000108, -0.00000153, -0.00000195, -0.00000235, -0.00000271, -0.00000302, -0.00000329, -0.00000349, -0.00000364, -0.00000372, -0.00000372, -0.00000366, -0.00000351, -0.00000330, -0.00000300, -0.00000264, -0.00000221, -0.00000171, -0.00000115, -0.00000055, 0.00000011, 0.00000079, 0.00000150, 0.00000223, 0.00000296, 0.00000369, 0.00000440, 0.00000508, 0.00000573, 0.00000634, 0.00000689, 0.00000738, 0.00000780, 0.00000815, 0.00000841, 0.00000860, 0.00000869, 0.00000871, 0.00000863, 0.00000847, 0.00000823, 0.00000791, 0.00000752, 0.00000707, 0.00000656, 0.00000600, 0.00000539, 0.00000475, 0.00000409, 0.00000342, 0.00000274, 0.00000207, 0.00000141, 0.00000078, 0.00000017, -0.00000040, -0.00000092, -0.00000140, -0.00000183, -0.00000219, -0.00000250, -0.00000276, -0.00000295, -0.00000308, -0.00000316, -0.00000319, -0.00000316, -0.00000309, -0.00000298, -0.00000283, -0.00000265, -0.00000245, -0.00000223, -0.00000199, -0.00000174, -0.00000149, -0.00000125, -0.00000100, -0.00000076, -0.00000054, },
		{ -0.00000110, -0.00000126, -0.00000140, -0.00000150, -0.00000157, -0.00000161, -0.00000163, -0.00000164, -0.00000163, -0.00000161, -0.00000159, -0.00000157, -0.00000156, -0.00000155, -0.00000155, -0.00000156, -0.00000159, -0.00000163, -0.00000168, -0.00000174, -0.00000182, -0.00000190, -0.00000198, -0.00000207, -0.00000215, -0.00000222, -0.00000228, -0.00000231, -0.00000232, -0.00000231, -0.00000225, -0.00000216, -0.00000203, -0.00000185, -0.00000163, -0.00000136, -0.00000104, -0.00000068, -0.00000027, 0.00000017, 0.00000066, 0.00000117, 0.00000171, 0.00000227, 0.00000283, 0.00000339, 0.00000395, 0.00000448, 0.00000499, 0.00000547, 0.00000589, 0.00000626, 0.00000657, 0.00000681, 0.00000697, 0.00000706, 0.00000705, 0.00000696, 0.00000679, 0.00000652, 0.00000617, 0.00000574, 0.00000523, 0.00000465, 0.00000400, 0.00000330, 0.00000256, 0.00000178, 0.00000098, 0.00000017, -0.00000064, -0.00000143, -0.00000220, -0.00000293, -0.00000361, -0.00000423, -0.00000478, -0.00000526, -0.00000564, -0.00000594, -0.00000614, -0.00000625, -0.00000626, -0.00000616, -0.00000598, -0.00000570, -0.00000533, -0.00000489, -0.00000437, -0.00000379, -0.00000315, -0.00000248, -0.00000177, -0.00000105, -0.00000031, 0.00000042, 0.00000113, 0.00000182, 0.00000247, 0.00000308, 0.00000363, 0.00000412, 0.00000455, 0.00000490, 0.00000517, 0.00000537, 0.00000550, 0.00000554, 0.00000551, 0.00000541, 0.00000525, 0.00000502, 0.00000474, 0.00000441, 0.00000405, 0.00000365, 0.00000323, 0.00000279, 0.00000235, 0.00000190, 0.00000147, 0.00000105, 0.00000065, 0.00000027, -0.00000007, -0.00000038, -0.00000066, -0.00000090, },
		{ 0.00000119, 0.00000108, 0.00000095, 0.00000079, 0.00000062, 0.00000044, 0.00000025, 0.00000007, -0.00000011, -0.00000027, -0.00000040, -0.00000052, -0.00000060, -0.00000065, -0.00000067, -0.00000066, -0.00000061, -0.00000053, -0.00000043, -0.00000030, -0.00000015, 0.00000002, 0.00000019, 0.00000037, 0.00000054, 0.00000070, 0.00000084, 0.00000095, 0.00000104, 0.00000108, 0.00000107, 0.00000102, 0.00000092, 0.00000076, 0.00000055, 0.00000029, -0.00000003, -0.00000040, -0.00000082, -0.00000127, -0.00000176, -0.00000228, -0.00000281, -0.00000334, -0.00000388, -0.00000440, -0.00000489, -0.00000535, -0.00000576, -0.00000612, -0.00000641, -0.00000662, -0.00000676, -0.00000680, -0.00000676, -0.00000662, -0.00000638, -0.00000604, -0.00000561, -0.00000509, -0.00000448, -0.00000380, -0.00000304, -0.00000222, -0.00000135, -0.00000045, 0.00000049, 0.00000143, 0.00000237, 0.00000330, 0.00000419, 0.00000505, 0.00000584, 0.00000657, 0.00000723, 0.00000779, 0.00000826, 0.00000863, 0.00000889, 0.00000904, 0.00000908, 0.00000901, 0.00000884, 0.00000856, 0.00000819, 0.00000772, 0.00000718, 0.00000657, 0.00000590, 0.00000518, 0.00000443, 0.00000366, 0.00000287, 0.00000210, 0.00000134, 0.00000061, -0.00000009, -0.00000073, -0.00000131, -0.00000183, -0.00000228, -0.00000265, -0.00000295, -0.00000316, -0.00000330, -0.00000336, -0.00000334, -0.00000326, -0.00000311, -0.00000290, -0.00000265, -0.00000236, -0.00000203, -0.00000169, -0.00000133, -0.00000097, -0.00000061, -0.00000027, 0.00000006, 0.00000035, 0.00000061, 0.00000084, 0.00000102, 0.00000116, 0.00000125, 0.00000130, 0.00000130, 0.00000127, },
		{ -0.00000079, -0.00000073, -0.00000066, -0.00000058, -0.00000050, -0.00000042, -0.00000037, -0.00000033, -0.00000031, -0.00000033, -0.00000037, -0.00000045, -0.00000057, -0.00000072, -0.00000090, -0.00000112, -0.00000136, -0.00000163, -0.00000191, -0.00000220, -0.00000250, -0.00000278, -0.00000306, -0.00000331, -0.00000352, -0.00000370, -0.00000383, -0.00000390, -0.00000391, -0.00000385, -0.00000372, -0.00000352, -0.00000324, -0.00000289, -0.00000246, -0.00000196, -0.00000141, -0.00000079, -0.00000012, 0.00000059, 0.00000132, 0.00000208, 0.00000284, 0.00000359, 0.00000433, 0.00000503, 0.00000569, 0.00000629, 0.00000682, 0.00000727, 0.00000764, 0.00000790, 0.00000807, 0.00000812, 0.00000807, 0.00000790, 0.00000762, 0.00000724, 0.00000675, 0.00000617, 0.00000550, 0.00000475, 0.00000394, 0.00000308, 0.00000218, 0.00000125, 0.00000031, -0.00000062, -0.00000153, -0.00000240, -0.00000322, -0.00000398, -0.00000466, -0.00000526, -0.00000576, -0.00000616, -0.00000645, -0.00000663, -0.00000669, -0.00000664, -0.00000648, -0.00000621, -0.00000583, -0.00000537, -0.00000482, -0.00000420, -0.00000351, -0.00000278, -0.00000201, -0.00000121, -0.00000041, 0.00000038, 0.00000116, 0.00000190, 0.00000261, 0.00000326, 0.00000384, 0.00000436, 0.00000479, 0.00000515, 0.00000542, 0.00000560, 0.00000570, 0.00000571, 0.00000564, 0.00000550, 0.00000529, 0.00000501, 0.00000468, 0.00000431, 0.00000390, 0.00000346, 0.00000301, 0.00000255, 0.00000210, 0.00000166, 0.00000124, 0.00000085, 0.00000049, 0.00000017, -0.00000010, -0.00000033, -0.00000052, -0.00000066, -0.00000076, -0.00000082, -0.00000084, -0.00000083, },
		{ 0.00000073, 0.00000056, 0.00000039, 0.00000022, 0.00000007, -0.00000006, -0.00000017, -0.00000024, -0.00000029, -0.00000030, -0.00000027, -0.00000020, -0.00000010, 0.00000004, 0.00000020, 0.00000040, 0.00000061, 0.00000083, 0.00000106, 0.00000129, 0.00000151, 0.00000170, 0.00000187, 0.00000200, 0.00000208, 0.00000211, 0.00000208, 0.00000198, 0.00000182, 0.00000158, 0.00000127, 0.00000089, 0.00000044, -0.00000008, -0.00000066, -0.00000129, -0.00000197, -0.00000269, -0.00000343, -0.00000418, -0.00000494, -0.00000568, -0.00000640, -0.00000707, -0.00000769, -0.00000825, -0.00000872, -0.00000911, -0.00000940, -0.00000958, -0.00000965, -0.00000960, -0.00000943, -0.00000914, -0.00000872, -0.00000819, -0.00000754, -0.00000679, -0.00000595, -0.00000502, -0.00000402, -0.00000296, -0.00000186, -0.00000074, 0.00000040, 0.00000153, 0.00000264, 0.00000371, 0.00000473, 0.00000567, 0.00000654, 0.00000731, 0.00000797, 0.00000852, 0.00000895, 0.00000925, 0.00000943, 0.00000948, 0.00000940, 0.00000920, 0.00000889, 0.00000846, 0.00000794, 0.00000733, 0.00000665, 0.00000590, 0.00000511, 0.00000428, 0.00000344, 0.00000259, 0.00000175, 0.00000093, 0.00000015, -0.00000058, -0.00000125, -0.00000185, -0.00000238, -0.00000282, -0.00000318, -0.00000345, -0.00000364, -0.00000373, -0.00000375, -0.00000369, -0.00000355, -0.00000335, -0.00000309, -0.00000279, -0.00000245, -0.00000208, -0.00000169, -0.00000130, -0.00000090, -0.00000052, -0.00000016, 0.00000017, 0.00000046, 0.00000072, 0.00000093, 0.00000110, 0.00000122, 0.00000129, 0.00000132, 0.00000130, 0.00000124, 0.00000115, 0.00000103, 0.00000089, },
		{ 0.00000083, 0.00000105, 0.00000125, 0.00000142, 0.00000156, 0.00000167, 0.00000174, 0.00000176, 0.00000174, 0.00000168, 0.00000156, 0.00000140, 0.00000121, 0.00000097, 0.00000070, 0.00000041, 0.00000010, -0.00000023, -0.00000055, -0.00000086, -0.00000116, -0.00000144, -0.00000168, -0.00000188, -0.00000203, -0.00000213, -0.00000216, -0.00000214, -0.00000205, -0.00000189, -0.00000166, -0.00000138, -0.00000103, -0.00000064, -0.00000019, 0.00000029, 0.00000080, 0.00000133, 0.00000186, 0.00000239, 0.00000289, 0.00000337, 0.00000380, 0.00000418, 0.00000449, 0.00000472, 0.00000487, 0.00000492, 0.00000487, 0.00000471, 0.00000444, 0.00000407, 0.00000358, 0.00000299, 0.00000231, 0.00000152, 0.00000066, -0.00000028, -0.00000128, -0.00000233, -0.00000342, -0.00000452, -0.00000562, -0.00000671, -0.00000778, -0.00000879, -0.00000975, -0.00001063, -0.00001142, -0.00001211, -0.00001269, -0.00001316, -0.00001350, -0.00001371, -0.00001379, -0.00001374, -0.00001355, -0.00001325, -0.00001282, -0.00001228, -0.00001164, -0.00001091, -0.00001011, -0.00000923, -0.00000831, -0.00000735, -0.00000637, -0.00000539, -0.00000441, -0.00000346, -0.00000255, -0.00000168, -0.00000087, -0.00000013, 0.00000053, 0.00000111, 0.00000160, 0.00000201, 0.00000232, 0.00000255, 0.00000269, 0.00000274, 0.00000272, 0.00000264, 0.00000249, 0.00000229, 0.00000205, 0.00000178, 0.00000149, 0.00000119, 0.00000088, 0.00000059, 0.00000031, 0.00000006, -0.00000016, -0.00000034, -0.00000048, -0.00000057, -0.00000062, -0.00000063, -0.00000058, -0.00000050, -0.00000038, -0.00000022, -0.00000004, 0.00000017, 0.00000039, 0.00000061, },
		{ -0.00000037, -0.00000061, -0.00000083, -0.00000102, -0.00000117, -0.00000129, -0.00000137, -0.00000140, -0.00000139, -0.00000134, -0.00000125, -0.00000113, -0.00000097, -0.00000079, -0.00000059, -0.00000037, -0.00000016, 0.00000005, 0.00000024, 0.00000042, 0.00000056, 0.00000066, 0.00000071, 0.00000072, 0.00000066, 0.00000054, 0.00000036, 0.00000010, -0.00000022, -0.00000060, -0.00000105, -0.00000156, -0.00000212, -0.00000272, -0.00000337, -0.00000403, -0.00000472, -0.00000541, -0.00000609, -0.00000675, -0.00000738, -0.00000796, -0.00000849, -0.00000896, -0.00000934, -0.00000965, -0.00000986, -0.00000997, -0.00000998, -0.00000988, -0.00000968, -0.00000937, -0.00000897, -0.00000846, -0.00000787, -0.00000719, -0.00000644, -0.00000563, -0.00000477, -0.00000387, -0.00000296, -0.00000203, -0.00000111, -0.00000021, 0.00000066, 0.00000147, 0.00000223, 0.00000292, 0.00000353, 0.00000405, 0.00000448, 0.00000480, 0.00000502, 0.00000514, 0.00000515, 0.00000506, 0.00000487, 0.00000460, 0.00000424, 0.00000381, 0.00000331, 0.00000276, 0.00000217, 0.00000155, 0.00000091, 0.00000028, -0.00000035, -0.00000096, -0.00000153, -0.00000206, -0.00000254, -0.00000295, -0.00000330, -0.00000358, -0.00000378, -0.00000391, -0.00000395, -0.00000393, -0.00000382, -0.00000366, -0.00000343, -0.00000314, -0.00000281, -0.00000244, -0.00000204, -0.00000162, -0.00000119, -0.00000076, -0.00000035, 0.00000005, 0.00000042, 0.00000075, 0.00000104, 0.00000129, 0.00000148, 0.00000163, 0.00000171, 0.00000174, 0.00000172, 0.00000165, 0.00000153, 0.00000137, 0.00000117, 0.00000094, 0.00000069, 0.00000043, 0.00000016, -0.00000011, },
		{ -0.00000100, -0.00000078, -0.00000059, -0.00000045, -0.00000034, -0.00000027, -0.00000024, -0.00000026, -0.00000031, -0.00000039, -0.00000051, -0.00000064, -0.00000080, -0.00000097, -0.00000114, -0.00000131, -0.00000147, -0.00000160, -0.00000171, -0.00000178, -0.00000181, -0.00000179, -0.00000171, -0.00000157, -0.00000137, -0.00000111, -0.00000077, -0.00000038, 0.00000008, 0.00000060, 0.00000117, 0.00000179, 0.00000246, 0.00000315, 0.00000386, 0.00000459, 0.00000532, 0.00000603, 0.00000672, 0.00000738, 0.00000800, 0.00000855, 0.00000904, 0.00000946, 0.00000979, 0.00001003, 0.00001017, 0.00001022, 0.00001016, 0.00001000, 0.00000973, 0.00000937, 0.00000891, 0.00000837, 0.00000774, 0.00000704, 0.00000628, 0.00000546, 0.00000460, 0.00000372, 0.00000282, 0.00000192, 0.00000102, 0.00000015, -0.00000069, -0.00000148, -0.00000222, -0.00000289, -0.00000349, -0.00000402, -0.00000447, -0.00000483, -0.00000510, -0.00000529, -0.00000539, -0.00000541, -0.00000536, -0.00000524, -0.00000505, -0.00000481, -0.00000452, -0.00000419, -0.00000384, -0.00000347, -0.00000309, -0.00000272, -0.00000236, -0.00000202, -0.00000171, -0.00000143, -0.00000120, -0.00000101, -0.00000088, -0.00000079, -0.00000076, -0.00000078, -0.00000085, -0.00000097, -0.00000113, -0.00000133, -0.00000156, -0.00000181, -0.00000209, -0.00000237, -0.00000266, -0.00000295, -0.00000322, -0.00000347, -0.00000370, -0.00000390, -0.00000406, -0.00000417, -0.00000425, -0.00000428, -0.00000426, -0.00000419, -0.00000409, -0.00000393, -0.00000375, -0.00000352, -0.00000327, -0.00000299, -0.00000270, -0.00000240, -0.00000210, -0.00000180, -0.00000151, -0.00000124, },
		{ -0.00000052, -0.00000045, -0.00000034, -0.00000020, -0.00000003, 0.00000017, 0.00000040, 0.00000066, 0.00000094, 0.00000124, 0.00000155, 0.00000187, 0.00000218, 0.00000249, 0.00000279, 0.00000307, 0.00000332, 0.00000353, 0.00000371, 0.00000383, 0.00000391, 0.00000393, 0.00000389, 0.00000379, 0.00000362, 0.00000340, 0.00000311, 0.00000276, 0.00000235, 0.00000190, 0.00000140, 0.00000086, 0.00000029, -0.00000030, -0.00000091, -0.00000151, -0.00000212, -0.00000270, -0.00000326, -0.00000379, -0.00000427, -0.00000470, -0.00000507, -0.00000537, -0.00000560, -0.00000576, -0.00000583, -0.00000582, -0.00000573, -0.00000556, -0.00000531, -0.00000498, -0.00000459, -0.00000413, -0.00000361, -0.00000304, -0.00000244, -0.00000180, -0.00000114, -0.00000048, 0.00000019, 0.00000084, 0.00000148, 0.00000208, 0.00000265, 0.00000317, 0.00000363, 0.00000403, 0.00000437, 0.00000464, 0.00000484, 0.00000496, 0.00000502, 0.00000500, 0.00000491, 0.00000476, 0.00000455, 0.00000429, 0.00000398, 0.00000363, 0.00000324, 0.00000283, 0.00000241, 0.00000198, 0.00000155, 0.00000112, 0.00000072, 0.00000033, -0.00000003, -0.00000036, -0.00000065, -0.00000089, -0.00000110, -0.00000127, -0.00000139, -0.00000146, -0.00000150, -0.00000150, -0.00000146, -0.00000139, -0.00000130, -0.00000118, -0.00000105, -0.00000091, -0.00000076, -0.00000061, -0.00000047, -0.00000033, -0.00000021, -0.00000010, -0.00000002, 0.00000005, 0.00000009, 0.00000011, 0.00000011, 0.00000009, 0.00000005, -0.00000001, -0.00000008, -0.00000015, -0.00000024, -0.00000032, -0.00000040, -0.00000047, -0.00000052, -0.00000056, -0.00000057, -0.00000056, },
		{ 0.00000004, 0.00000000, -0.00000006, -0.00000014, -0.00000026, -0.00000039, -0.00000055, -0.00000072, -0.00000091, -0.00000111, -0.00000131, -0.00000152, -0.00000173, -0.00000192, -0.00000211, -0.00000227, -0.00000241, -0.00000253, -0.00000261, -0.00000266, -0.00000266, -0.00000263, -0.00000255, -0.00000243, -0.00000226, -0.00000204, -0.00000179, -0.00000149, -0.00000116, -0.00000079, -0.00000039, 0.00000003, 0.00000046, 0.00000091, 0.00000137, 0.00000182, 0.00000226, 0.00000269, 0.00000309, 0.00000347, 0.00000380, 0.00000410, 0.00000435, 0.00000455, 0.00000469, 0.00000478, 0.00000480, 0.00000477, 0.00000467, 0.00000452, 0.00000431, 0.00000404, 0.00000373, 0.00000337, 0.00000296, 0.00000253, 0.00000206, 0.00000157, 0.00000107, 0.00000056, 0.00000005, -0.00000046, -0.00000095, -0.00000142, -0.00000187, -0.00000228, -0.00000266, -0.00000300, -0.00000330, -0.00000355, -0.00000376, -0.00000392, -0.00000403, -0.00000409, -0.00000411, -0.00000409, -0.00000403, -0.00000394, -0.00000381, -0.00000366, -0.00000348, -0.00000329, -0.00000309, -0.00000288, -0.00000267, -0.00000246, -0.00000226, -0.00000206, -0.00000189, -0.00000173, -0.00000159, -0.00000146, -0.00000136, -0.00000129, -0.00000123, -0.00000119, -0.00000117, -0.00000117, -0.00000119, -0.00000121, -0.00000124, -0.00000128, -0.00000133, -0.00000137, -0.00000141, -0.00000144, -0.00000146, -0.00000147, -0.00000147, -0.00000146, -0.00000143, -0.00000138, -0.00000132, -0.00000124, -0.00000115, -0.00000105, -0.00000094, -0.00000082, -0.00000070, -0.00000058, -0.00000046, -0.00000034, -0.00000023, -0.00000014, -0.00000006, -0.00000001, 0.00000003, 0.00000005, },
		{ -0.00000386, -0.00000378, -0.00000367, -0.00000352, -0.00000335, -0.00000315, -0.00000293, -0.00000269, -0.00000244, -0.00000217, -0.00000189, -0.00000160, -0.00000132, -0.00000103, -0.00000076, -0.00000050, -0.00000025, -0.00000003, 0.00000017, 0.00000034, 0.00000048, 0.00000059, 0.00000065, 0.00000069, 0.00000068, 0.00000063, 0.00000054, 0.00000042, 0.00000026, 0.00000006, -0.00000017, -0.00000043, -0.00000071, -0.00000102, -0.00000134, -0.00000167, -0.00000201, -0.00000235, -0.00000269, -0.00000302, -0.00000333, -0.00000363, -0.00000390, -0.00000414, -0.00000434, -0.00000451, -0.00000464, -0.00000473, -0.00000477, -0.00000477, -0.00000472, -0.00000462, -0.00000448, -0.00000429, -0.00000405, -0.00000378, -0.00000347, -0.00000312, -0.00000274, -0.00000233, -0.00000190, -0.00000146, -0.00000099, -0.00000052, -0.00000005, 0.00000042, 0.00000089, 0.00000135, 0.00000180, 0.00000223, 0.00000263, 0.00000302, 0.00000337, 0.00000370, 0.00000400, 0.00000426, 0.00000449, 0.00000469, 0.00000486, 0.00000499, 0.00000509, 0.00000515, 0.00000519, 0.00000519, 0.00000517, 0.00000512, 0.00000504, 0.00000494, 0.00000482, 0.00000468, 0.00000452, 0.00000434, 0.00000415, 0.00000394, 0.00000373, 0.00000350, 0.00000325, 0.00000300, 0.00000274, 0.00000247, 0.00000220, 0.00000191, 0.00000163, 0.00000133, 0.00000103, 0.00000073, 0.00000042, 0.00000011, -0.00000020, -0.00000051, -0.00000082, -0.00000113, -0.00000143, -0.00000172, -0.00000200, -0.00000228, -0.00000254, -0.00000278, -0.00000301, -0.00000321, -0.00000340, -0.00000356, -0.00000369, -0.00000380, -0.00000387, -0.00000392, -0.00000393, -0.00000391, },
		{ 0.00000176, 0.00000130, 0.00000083, 0.00000036, -0.00000012, -0.00000060, -0.00000108, -0.00000155, -0.00000201, -0.00000245, -0.00000288, -0.00000329, -0.00000368, -0.00000404, -0.00000436, -0.00000466, -0.00000491, -0.00000513, -0.00000530, -0.00000543, -0.00000551, -0.00000554, -0.00000552, -0.00000544, -0.00000531, -0.00000513, -0.00000490, -0.00000461, -0.00000427, -0.00000389, -0.00000346, -0.00000298, -0.00000247, -0.00000191, -0.00000133, -0.00000073, -0.00000010, 0.00000054, 0.00000119, 0.00000184, 0.00000248, 0.00000312, 0.00000373, 0.00000432, 0.00000487, 0.00000539, 0.00000587, 0.00000629, 0.00000666, 0.00000697, 0.00000722, 0.00000741, 0.00000752, 0.00000757, 0.00000755, 0.00000746, 0.00000730, 0.00000707, 0.00000678, 0.00000643, 0.00000602, 0.00000556, 0.00000505, 0.00000451, 0.00000393, 0.00000332, 0.00000270, 0.00000206, 0.00000142, 0.00000078, 0.00000015, -0.00000047, -0.00000105, -0.00000161, -0.00000213, -0.00000261, -0.00000304, -0.00000341, -0.00000373, -0.00000400, -0.00000420, -0.00000434, -0.00000442, -0.00000443, -0.00000438, -0.00000427, -0.00000411, -0.00000389, -0.00000362, -0.00000330, -0.00000294, -0.00000254, -0.00000210, -0.00000164, -0.00000116, -0.00000066, -0.00000015, 0.00000036, 0.00000087, 0.00000138, 0.00000187, 0.00000235, 0.00000281, 0.00000324, 0.00000364, 0.00000401, 0.00000435, 0.00000465, 0.00000491, 0.00000512, 0.00000530, 0.00000543, 0.00000552, 0.00000556, 0.00000556, 0.00000552, 0.00000543, 0.00000531, 0.00000514, 0.00000494, 0.00000470, 0.00000442, 0.00000412, 0.00000378, 0.00000342, 0.00000304, 0.00000263, 0.00000220, },
		{ 0.00000267, 0.00000299, 0.00000328, 0.00000354, 0.00000377, 0.00000397, 0.00000415, 0.00000429, 0.00000439, 0.00000447, 0.00000452, 0.00000453, 0.00000452, 0.00000448, 0.00000440, 0.00000430, 0.00000417, 0.00000401, 0.00000382, 0.00000361, 0.00000336, 0.00000309, 0.00000279, 0.00000246, 0.00000210, 0.00000171, 0.00000129, 0.00000085, 0.00000038, -0.00000011, -0.00000063, -0.00000116, -0.00000172, -0.00000229, -0.00000286, -0.00000345, -0.00000404, -0.00000462, -0.00000520, -0.00000576, -0.00000631, -0.00000682, -0.00000730, -0.00000775, -0.00000814, -0.00000849, -0.00000877, -0.00000900, -0.00000915, -0.00000923, -0.00000923, -0.00000916, -0.00000900, -0.00000876, -0.00000843, -0.00000802, -0.00000753, -0.00000697, -0.00000633, -0.00000562, -0.00000485, -0.00000402, -0.00000315, -0.00000224, -0.00000130, -0.00000034, 0.00000063, 0.00000160, 0.00000255, 0.00000348, 0.00000438, 0.00000523, 0.00000603, 0.00000677, 0.00000743, 0.00000802, 0.00000853, 0.00000894, 0.00000927, 0.00000949, 0.00000963, 0.00000966, 0.00000960, 0.00000944, 0.00000919, 0.00000886, 0.00000845, 0.00000797, 0.00000742, 0.00000681, 0.00000616, 0.00000547, 0.00000475, 0.00000402, 0.00000327, 0.00000253, 0.00000180, 0.00000109, 0.00000041, -0.00000023, -0.00000083, -0.00000137, -0.00000187, -0.00000230, -0.00000266, -0.00000296, -0.00000319, -0.00000335, -0.00000345, -0.00000348, -0.00000344, -0.00000334, -0.00000319, -0.00000298, -0.00000273, -0.00000243, -0.00000210, -0.00000174, -0.00000135, -0.00000094, -0.00000053, -0.00000010, 0.00000033, 0.00000075, 0.00000117, 0.00000157, 0.00000196, 0.00000232, },
		{ 0.00000149, 0.00000164, 0.00000181, 0.00000201, 0.00000223, 0.00000247, 0.00000271, 0.00000297, 0.00000322, 0.00000347, 0.00000371, 0.00000394, 0.00000415, 0.00000434, 0.00000451, 0.00000465, 0.00000477, 0.00000487, 0.00000493, 0.00000497, 0.00000499, 0.00000499, 0.00000497, 0.00000493, 0.00000487, 0.00000481, 0.00000475, 0.00000468, 0.00000461, 0.00000454, 0.00000448, 0.00000444, 0.00000440, 0.00000437, 0.00000436, 0.00000436, 0.00000437, 0.00000439, 0.00000441, 0.00000444, 0.00000447, 0.00000450, 0.00000451, 0.00000450, 0.00000448, 0.00000443, 0.00000434, 0.00000421, 0.00000404, 0.00000382, 0.00000354, 0.00000321, 0.00000281, 0.00000235, 0.00000183, 0.00000124, 0.00000059, -0.00000012, -0.00000088, -0.00000170, -0.00000256, -0.00000346, -0.00000439, -0.00000534, -0.00000630, -0.00000726, -0.00000822, -0.00000914, -0.00001004, -0.00001089, -0.00001169, -0.00001243, -0.00001309, -0.00001366, -0.00001415, -0.00001454, -0.00001482, -0.00001500, -0.00001507, -0.00001503, -0.00001488, -0.00001462, -0.00001426, -0.00001380, -0.00001325, -0.00001261, -0.00001189, -0.00001111, -0.00001026, -0.00000938, -0.00000846, -0.00000752, -0.00000656, -0.00000561, -0.00000467, -0.00000376, -0.00000287, -0.00000203, -0.00000124, -0.00000051, 0.00000016, 0.00000075, 0.00000128, 0.00000173, 0.00000211, 0.00000242, 0.00000265, 0.00000281, 0.00000291, 0.00000295, 0.00000294, 0.00000288, 0.00000279, 0.00000266, 0.00000250, 0.00000233, 0.00000216, 0.00000198, 0.00000181, 0.00000165, 0.00000151, 0.00000140, 0.00000131, 0.00000126, 0.00000123, 0.00000125, 0.00000130, 0.00000138, },
		{ -0.00000110, -0.00000130, -0.00000150, -0.00000171, -0.00000190, -0.00000209, -0.00000225, -0.00000239, -0.00000250, -0.00000257, -0.00000260, -0.00000258, -0.00000252, -0.00000241, -0.00000225, -0.00000205, -0.00000180, -0.00000152, -0.00000120, -0.00000085, -0.00000048, -0.00000009, 0.00000030, 0.00000069, 0.00000107, 0.00000144, 0.00000177, 0.00000207, 0.00000232, 0.00000252, 0.00000266, 0.00000275, 0.00000276, 0.00000271, 0.00000259, 0.00000240, 0.00000214, 0.00000181, 0.00000143, 0.00000099, 0.00000051, -0.00000001, -0.00000057, -0.00000114, -0.00000172, -0.00000229, -0.00000285, -0.00000339, -0.00000389, -0.00000435, -0.00000474, -0.00000507, -0.00000533, -0.00000551, -0.00000560, -0.00000560, -0.00000552, -0.00000534, -0.00000507, -0.00000471, -0.00000427, -0.00000375, -0.00000317, -0.00000252, -0.00000182, -0.00000109, -0.00000032, 0.00000046, 0.00000125, 0.00000202, 0.00000278, 0.00000350, 0.00000418, 0.00000480, 0.00000535, 0.00000584, 0.00000624, 0.00000656, 0.00000679, 0.00000693, 0.00000697, 0.00000692, 0.00000679, 0.00000657, 0.00000626, 0.00000589, 0.00000545, 0.00000496, 0.00000442, 0.00000384, 0.00000324, 0.00000263, 0.00000201, 0.00000140, 0.00000081, 0.00000024, -0.00000029, -0.00000078, -0.00000122, -0.00000161, -0.00000194, -0.00000221, -0.00000241, -0.00000256, -0.00000264, -0.00000267, -0.00000265, -0.00000258, -0.00000246, -0.00000232, -0.00000214, -0.00000194, -0.00000173, -0.00000152, -0.00000130, -0.00000110, -0.00000091, -0.00000075, -0.00000061, -0.00000051, -0.00000043, -0.00000040, -0.00000040, -0.00000044, -0.00000051, -0.00000062, -0.00000076, -0.00000092, },
		{ -0.00000241, -0.00000227, -0.00000213, -0.00000201, -0.00000191, -0.00000183, -0.00000178, -0.00000176, -0.00000177, -0.00000182, -0.00000191, -0.00000203, -0.00000218, -0.00000237, -0.00000258, -0.00000282, -0.00000306, -0.00000332, -0.00000357, -0.00000381, -0.00000403, -0.00000421, -0.00000436, -0.00000445, -0.00000447, -0.00000442, -0.00000430, -0.00000408, -0.00000376, -0.00000335, -0.00000283, -0.00000220, -0.00000147, -0.00000063, 0.00000030, 0.00000133, 0.00000244, 0.00000363, 0.00000488, 0.00000618, 0.00000752, 0.00000888, 0.00001025, 0.00001160, 0.00001292, 0.00001419, 0.00001540, 0.00001652, 0.00001755, 0.00001846, 0.00001924, 0.00001988, 0.00002038, 0.00002071, 0.00002089, 0.00002089, 0.00002073, 0.00002040, 0.00001991, 0.00001926, 0.00001846, 0.00001752, 0.00001646, 0.00001528, 0.00001400, 0.00001265, 0.00001123, 0.00000976, 0.00000827, 0.00000676, 0.00000527, 0.00000381, 0.00000239, 0.00000103, -0.00000026, -0.00000146, -0.00000256, -0.00000356, -0.00000444, -0.00000521, -0.00000585, -0.00000637, -0.00000677, -0.00000705, -0.00000722, -0.00000729, -0.00000725, -0.00000713, -0.00000693, -0.00000667, -0.00000634, -0.00000598, -0.00000559, -0.00000517, -0.00000475, -0.00000433, -0.00000393, -0.00000354, -0.00000319, -0.00000287, -0.00000259, -0.00000236, -0.00000217, -0.00000203, -0.00000194, -0.00000189, -0.00000188, -0.00000191, -0.00000197, -0.00000206, -0.00000217, -0.00000230, -0.00000243, -0.00000256, -0.00000270, -0.00000282, -0.00000292, -0.00000301, -0.00000308, -0.00000311, -0.00000313, -0.00000311, -0.00000307, -0.00000300, -0.00000291, -0.00000280, -0.00000268, -0.00000254, },
		{ -0.00000171, -0.00000152, -0.00000131, -0.00000109, -0.00000084, -0.00000058, -0.00000029, 0.00000002, 0.00000036, 0.00000072, 0.00000110, 0.00000150, 0.00000191, 0.00000234, 0.00000276, 0.00000319, 0.00000359, 0.00000398, 0.00000434, 0.00000466, 0.00000492, 0.00000513, 0.00000527, 0.00000533, 0.00000531, 0.00000519, 0.00000497, 0.00000466, 0.00000423, 0.00000370, 0.00000307, 0.00000233, 0.00000150, 0.00000057, -0.00000044, -0.00000151, -0.00000265, -0.00000383, -0.00000504, -0.00000626, -0.00000747, -0.00000866, -0.00000980, -0.00001088, -0.00001189, -0.00001279, -0.00001357, -0.00001423, -0.00001474, -0.00001509, -0.00001527, -0.00001528, -0.00001510, -0.00001474, -0.00001419, -0.00001346, -0.00001255, -0.00001147, -0.00001024, -0.00000886, -0.00000734, -0.00000572, -0.00000400, -0.00000221, -0.00000036, 0.00000151, 0.00000339, 0.00000525, 0.00000707, 0.00000883, 0.00001051, 0.00001208, 0.00001354, 0.00001485, 0.00001601, 0.00001701, 0.00001784, 0.00001848, 0.00001895, 0.00001922, 0.00001932, 0.00001923, 0.00001897, 0.00001855, 0.00001797, 0.00001725, 0.00001640, 0.00001545, 0.00001439, 0.00001326, 0.00001207, 0.00001084, 0.00000958, 0.00000832, 0.00000706, 0.00000583, 0.00000464, 0.00000350, 0.00000242, 0.00000141, 0.00000048, -0.00000036, -0.00000112, -0.00000179, -0.00000236, -0.00000285, -0.00000325, -0.00000357, -0.00000380, -0.00000397, -0.00000407, -0.00000412, -0.00000411, -0.00000406, -0.00000398, -0.00000387, -0.00000373, -0.00000358, -0.00000342, -0.00000326, -0.00000309, -0.00000291, -0.00000274, -0.00000258, -0.00000241, -0.00000224, -0.00000207, -0.00000189, },
		{ -0.00000102, -0.00000151, -0.00000199, -0.00000246, -0.00000292, -0.00000337, -0.00000382, -0.00000426, -0.00000470, -0.00000514, -0.00000557, -0.00000600, -0.00000642, -0.00000683, -0.00000722, -0.00000758, -0.00000792, -0.00000822, -0.00000848, -0.00000869, -0.00000883, -0.00000892, -0.00000893, -0.00000886, -0.00000871, -0.00000847, -0.00000814, -0.00000772, -0.00000721, -0.00000660, -0.00000591, -0.00000514, -0.00000428, -0.00000337, -0.00000239, -0.00000137, -0.00000031, 0.00000076, 0.00000183, 0.00000289, 0.00000393, 0.00000491, 0.00000583, 0.00000667, 0.00000741, 0.00000805, 0.00000856, 0.00000893, 0.00000916, 0.00000923, 0.00000915, 0.00000890, 0.00000849, 0.00000793, 0.00000720, 0.00000633, 0.00000532, 0.00000418, 0.00000292, 0.00000158, 0.00000015, -0.00000133, -0.00000285, -0.00000439, -0.00000592, -0.00000742, -0.00000886, -0.00001024, -0.00001152, -0.00001269, -0.00001372, -0.00001461, -0.00001534, -0.00001590, -0.00001628, -0.00001648, -0.00001648, -0.00001630, -0.00001593, -0.00001538, -0.00001465, -0.00001377, -0.00001273, -0.00001156, -0.00001027, -0.00000889, -0.00000742, -0.00000589, -0.00000432, -0.00000273, -0.00000115, 0.00000041, 0.00000193, 0.00000340, 0.00000478, 0.00000608, 0.00000727, 0.00000835, 0.00000930, 0.00001012, 0.00001080, 0.00001135, 0.00001176, 0.00001203, 0.00001217, 0.00001219, 0.00001208, 0.00001186, 0.00001154, 0.00001113, 0.00001064, 0.00001008, 0.00000947, 0.00000881, 0.00000811, 0.00000740, 0.00000666, 0.00000593, 0.00000520, 0.00000447, 0.00000377, 0.00000308, 0.00000242, 0.00000178, 0.00000117, 0.00000059, 0.00000003, -0.00000050, },
		{ 0.00000526, 0.00000520, 0.00000512, 0.00000501, 0.00000489, 0.00000475, 0.00000461, 0.00000447, 0.00000433, 0.00000419, 0.00000405, 0.00000392, 0.00000380, 0.00000367, 0.00000355, 0.00000342, 0.00000328, 0.00000312, 0.00000294, 0.00000274, 0.00000250, 0.00000223, 0.00000191, 0.00000153, 0.00000110, 0.00000062, 0.00000007, -0.00000055, -0.00000122, -0.00000196, -0.00000276, -0.00000361, -0.00000451, -0.00000545, -0.00000643, -0.00000742, -0.00000843, -0.00000943, -0.00001041, -0.00001136, -0.00001227, -0.00001311, -0.00001389, -0.00001457, -0.00001515, -0.00001562, -0.00001596, -0.00001617, -0.00001624, -0.00001616, -0.00001594, -0.00001556, -0.00001503, -0.00001436, -0.00001354, -0.00001260, -0.00001153, -0.00001035, -0.00000908, -0.00000773, -0.00000632, -0.00000487, -0.00000339, -0.00000191, -0.00000044, 0.00000099, 0.00000236, 0.00000366, 0.00000487, 0.00000598, 0.00000696, 0.00000780, 0.00000851, 0.00000906, 0.00000946, 0.00000969, 0.00000977, 0.00000969, 0.00000946, 0.00000908, 0.00000857, 0.00000793, 0.00000719, 0.00000635, 0.00000543, 0.00000445, 0.00000343, 0.00000238, 0.00000133, 0.00000028, -0.00000073, -0.00000170, -0.00000261, -0.00000344, -0.00000419, -0.00000484, -0.00000538, -0.00000581, -0.00000613, -0.00000633, -0.00000641, -0.00000638, -0.00000623, -0.00000599, -0.00000564, -0.00000521, -0.00000470, -0.00000412, -0.00000349, -0.00000282, -0.00000212, -0.00000140, -0.00000068, 0.00000004, 0.00000073, 0.00000140, 0.00000202, 0.00000261, 0.00000314, 0.00000361, 0.00000403, 0.00000438, 0.00000467, 0.00000490, 0.00000508, 0.00000519, 0.00000526, 0.00000528, },
		{ -0.00000141, -0.00000129, -0.00000116, -0.00000104, -0.00000093, -0.00000084, -0.00000076, -0.00000071, -0.00000068, -0.00000067, -0.00000070, -0.00000074, -0.00000081, -0.00000090, -0.00000100, -0.00000111, -0.00000123, -0.00000134, -0.00000144, -0.00000153, -0.00000158, -0.00000161, -0.00000159, -0.00000153, -0.00000141, -0.00000124, -0.00000101, -0.00000071, -0.00000034, 0.00000009, 0.00000058, 0.00000113, 0.00000174, 0.00000239, 0.00000308, 0.00000381, 0.00000455, 0.00000531, 0.00000605, 0.00000678, 0.00000748, 0.00000814, 0.00000873, 0.00000926, 0.00000970, 0.00001005, 0.00001029, 0.00001041, 0.00001042, 0.00001029, 0.00001004, 0.00000965, 0.00000914, 0.00000849, 0.00000773, 0.00000684, 0.00000585, 0.00000477, 0.00000360, 0.00000237, 0.00000108, -0.00000024, -0.00000158, -0.00000292, -0.00000425, -0.00000553, -0.00000677, -0.00000793, -0.00000901, -0.00000998, -0.00001085, -0.00001159, -0.00001220, -0.00001267, -0.00001299, -0.00001317, -0.00001320, -0.00001309, -0.00001284, -0.00001246, -0.00001195, -0.00001133, -0.00001061, -0.00000980, -0.00000892, -0.00000798, -0.00000700, -0.00000599, -0.00000497, -0.00000396, -0.00000298, -0.00000202, -0.00000112, -0.00000028, 0.00000049, 0.00000118, 0.00000179, 0.00000230, 0.00000271, 0.00000303, 0.00000326, 0.00000339, 0.00000343, 0.00000338, 0.00000326, 0.00000307, 0.00000282, 0.00000252, 0.00000218, 0.00000181, 0.00000143, 0.00000103, 0.00000063, 0.00000025, -0.00000012, -0.00000046, -0.00000077, -0.00000104, -0.00000127, -0.00000145, -0.00000160, -0.00000170, -0.00000175, -0.00000177, -0.00000175, -0.00000170, -0.00000162, -0.00000152, },
		{ -0.00000100, -0.00000093, -0.00000087, -0.00000080, -0.00000073, -0.00000066, -0.00000057, -0.00000046, -0.00000035, -0.00000021, -0.00000006, 0.00000011, 0.00000029, 0.00000048, 0.00000068, 0.00000088, 0.00000107, 0.00000125, 0.00000142, 0.00000156, 0.00000167, 0.00000174, 0.00000176, 0.00000174, 0.00000165, 0.00000151, 0.00000130, 0.00000103, 0.00000069, 0.00000028, -0.00000019, -0.00000072, -0.00000131, -0.00000195, -0.00000264, -0.00000335, -0.00000409, -0.00000485, -0.00000560, -0.00000634, -0.00000706, -0.00000773, -0.00000836, -0.00000892, -0.00000940, -0.00000980, -0.00001010, -0.00001030, -0.00001038, -0.00001035, -0.00001019, -0.00000991, -0.00000951, -0.00000899, -0.00000835, -0.00000760, -0.00000675, -0.00000581, -0.00000479, -0.00000371, -0.00000258, -0.00000141, -0.00000022, 0.00000097, 0.00000215, 0.00000329, 0.00000438, 0.00000541, 0.00000636, 0.00000721, 0.00000796, 0.00000859, 0.00000909, 0.00000946, 0.00000969, 0.00000979, 0.00000974, 0.00000956, 0.00000924, 0.00000880, 0.00000823, 0.00000755, 0.00000678, 0.00000592, 0.00000498, 0.00000399, 0.00000296, 0.00000191, 0.00000084, -0.00000023, -0.00000127, -0.00000228, -0.00000324, -0.00000414, -0.00000497, -0.00000572, -0.00000638, -0.00000695, -0.00000742, -0.00000780, -0.00000807, -0.00000825, -0.00000833, -0.00000833, -0.00000823, -0.00000806, -0.00000783, -0.00000753, -0.00000718, -0.00000678, -0.00000636, -0.00000592, -0.00000546, -0.00000500, -0.00000455, -0.00000411, -0.00000368, -0.00000329, -0.00000292, -0.00000258, -0.00000228, -0.00000202, -0.00000179, -0.00000159, -0.00000142, -0.00000129, -0.00000117, -0.00000108, },
		{ -0.00000140, -0.00000104, -0.00000068, -0.00000031, 0.00000004, 0.00000038, 0.00000070, 0.00000099, 0.00000125, 0.00000148, 0.00000167, 0.00000182, 0.00000194, 0.00000201, 0.00000205, 0.00000206, 0.00000204, 0.00000200, 0.00000194, 0.00000187, 0.00000179, 0.00000172, 0.00000165, 0.00000160, 0.00000157, 0.00000156, 0.00000159, 0.00000166, 0.00000176, 0.00000191, 0.00000209, 0.00000232, 0.00000258, 0.00000289, 0.00000322, 0.00000357, 0.00000395, 0.00000433, 0.00000471, 0.00000508, 0.00000543, 0.00000575, 0.00000603, 0.00000625, 0.00000641, 0.00000650, 0.00000650, 0.00000641, 0.00000623, 0.00000594, 0.00000555, 0.00000504, 0.00000443, 0.00000371, 0.00000288, 0.00000196, 0.00000095, -0.00000015, -0.00000132, -0.00000255, -0.00000383, -0.00000514, -0.00000646, -0.00000779, -0.00000911, -0.00001039, -0.00001162, -0.00001280, -0.00001389, -0.00001490, -0.00001580, -0.00001659, -0.00001726, -0.00001780, -0.00001820, -0.00001846, -0.00001858, -0.00001856, -0.00001841, -0.00001812, -0.00001771, -0.00001718, -0.00001655, -0.00001582, -0.00001502, -0.00001414, -0.00001321, -0.00001224, -0.00001125, -0.00001026, -0.00000927, -0.00000829, -0.00000736, -0.00000646, -0.00000563, -0.00000485, -0.00000415, -0.00000353, -0.00000299, -0.00000254, -0.00000217, -0.00000189, -0.00000168, -0.00000156, -0.00000150, -0.00000151, -0.00000158, -0.00000169, -0.00000185, -0.00000203, -0.00000222, -0.00000243, -0.00000264, -0.00000284, -0.00000301, -0.00000316, -0.00000328, -0.00000336, -0.00000340, -0.00000339, -0.00000334, -0.00000323, -0.00000308, -0.00000289, -0.00000265, -0.00000238, -0.00000208, -0.00000175, },
		{ -0.00000228, -0.00000276, -0.00000324, -0.00000371, -0.00000414, -0.00000455, -0.00000492, -0.00000524, -0.00000551, -0.00000573, -0.00000588, -0.00000598, -0.00000601, -0.00000599, -0.00000591, -0.00000577, -0.00000559, -0.00000537, -0.00000511, -0.00000483, -0.00000454, -0.00000424, -0.00000395, -0.00000367, -0.00000342, -0.00000320, -0.00000302, -0.00000290, -0.00000283, -0.00000283, -0.00000289, -0.00000302, -0.00000323, -0.00000350, -0.00000384, -0.00000424, -0.00000470, -0.00000520, -0.00000575, -0.00000632, -0.00000691, -0.00000751, -0.00000810, -0.00000867, -0.00000920, -0.00000969, -0.00001012, -0.00001047, -0.00001074, -0.00001092, -0.00001100, -0.00001097, -0.00001082, -0.00001056, -0.00001018, -0.00000969, -0.00000908, -0.00000836, -0.00000754, -0.00000662, -0.00000563, -0.00000456, -0.00000343, -0.00000227, -0.00000107, 0.00000014, 0.00000134, 0.00000253, 0.00000368, 0.00000478, 0.00000581, 0.00000677, 0.00000764, 0.00000840, 0.00000906, 0.00000961, 0.00001004, 0.00001034, 0.00001053, 0.00001059, 0.00001054, 0.00001037, 0.00001010, 0.00000974, 0.00000929, 0.00000877, 0.00000818, 0.00000755, 0.00000688, 0.00000619, 0.00000549, 0.00000480, 0.00000412, 0.00000346, 0.00000285, 0.00000228, 0.00000177, 0.00000132, 0.00000093, 0.00000061, 0.00000036, 0.00000017, 0.00000005, -0.00000001, -0.00000002, 0.00000003, 0.00000011, 0.00000023, 0.00000037, 0.00000053, 0.00000069, 0.00000085, 0.00000099, 0.00000111, 0.00000120, 0.00000126, 0.00000127, 0.00000123, 0.00000114, 0.00000100, 0.00000081, 0.00000056, 0.00000026, -0.00000008, -0.00000046, -0.00000088, -0.00000133, -0.00000180, },
		{ -0.00000134, -0.00000100, -0.00000068, -0.00000038, -0.00000010, 0.00000015, 0.00000035, 0.00000051, 0.00000062, 0.00000067, 0.00000066, 0.00000060, 0.00000049, 0.00000032, 0.00000011, -0.00000014, -0.00000042, -0.00000073, -0.00000104, -0.00000136, -0.00000166, -0.00000194, -0.00000218, -0.00000237, -0.00000249, -0.00000254, -0.00000251, -0.00000238, -0.00000215, -0.00000181, -0.00000136, -0.00000079, -0.00000011, 0.00000068, 0.00000158, 0.00000257, 0.00000366, 0.00000483, 0.00000607, 0.00000735, 0.00000867, 0.00001001, 0.00001134, 0.00001266, 0.00001393, 0.00001514, 0.00001628, 0.00001731, 0.00001824, 0.00001903, 0.00001968, 0.00002017, 0.00002050, 0.00002066, 0.00002064, 0.00002044, 0.00002007, 0.00001952, 0.00001880, 0.00001793, 0.00001690, 0.00001575, 0.00001447, 0.00001309, 0.00001163, 0.00001011, 0.00000854, 0.00000695, 0.00000536, 0.00000378, 0.00000225, 0.00000077, -0.00000063, -0.00000195, -0.00000316, -0.00000425, -0.00000523, -0.00000606, -0.00000677, -0.00000733, -0.00000775, -0.00000804, -0.00000820, -0.00000823, -0.00000815, -0.00000796, -0.00000769, -0.00000733, -0.00000690, -0.00000643, -0.00000592, -0.00000538, -0.00000484, -0.00000431, -0.00000379, -0.00000330, -0.00000285, -0.00000245, -0.00000210, -0.00000182, -0.00000159, -0.00000143, -0.00000134, -0.00000130, -0.00000132, -0.00000140, -0.00000152, -0.00000167, -0.00000186, -0.00000207, -0.00000229, -0.00000252, -0.00000273, -0.00000293, -0.00000311, -0.00000325, -0.00000336, -0.00000343, -0.00000345, -0.00000342, -0.00000334, -0.00000322, -0.00000305, -0.00000283, -0.00000258, -0.00000230, -0.00000199, -0.00000167, },
		{ 0.00000144, 0.00000137, 0.00000132, 0.00000129, 0.00000129, 0.00000133, 0.00000142, 0.00000156, 0.00000175, 0.00000200, 0.00000231, 0.00000267, 0.00000309, 0.00000355, 0.00000406, 0.00000460, 0.00000516, 0.00000573, 0.00000630, 0.00000685, 0.00000737, 0.00000785, 0.00000827, 0.00000861, 0.00000887, 0.00000902, 0.00000906, 0.00000898, 0.00000877, 0.00000842, 0.00000793, 0.00000730, 0.00000654, 0.00000563, 0.00000460, 0.00000346, 0.00000220, 0.00000086, -0.00000056, -0.00000203, -0.00000354, -0.00000506, -0.00000657, -0.00000805, -0.00000947, -0.00001081, -0.00001205, -0.00001317, -0.00001415, -0.00001496, -0.00001560, -0.00001605, -0.00001630, -0.00001635, -0.00001619, -0.00001582, -0.00001524, -0.00001446, -0.00001349, -0.00001235, -0.00001104, -0.00000958, -0.00000800, -0.00000632, -0.00000455, -0.00000274, -0.00000089, 0.00000096, 0.00000278, 0.00000456, 0.00000627, 0.00000788, 0.00000937, 0.00001073, 0.00001195, 0.00001300, 0.00001387, 0.00001457, 0.00001508, 0.00001540, 0.00001554, 0.00001549, 0.00001528, 0.00001490, 0.00001438, 0.00001372, 0.00001294, 0.00001207, 0.00001111, 0.00001009, 0.00000903, 0.00000795, 0.00000687, 0.00000580, 0.00000477, 0.00000378, 0.00000286, 0.00000201, 0.00000125, 0.00000058, 0.00000000, -0.00000047, -0.00000085, -0.00000112, -0.00000131, -0.00000140, -0.00000140, -0.00000134, -0.00000121, -0.00000103, -0.00000080, -0.00000054, -0.00000027, 0.00000002, 0.00000030, 0.00000058, 0.00000083, 0.00000106, 0.00000126, 0.00000142, 0.00000154, 0.00000163, 0.00000168, 0.00000169, 0.00000168, 0.00000164, 0.00000158, 0.00000151, },
		{ 0.00000287, 0.00000303, 0.00000316, 0.00000325, 0.00000329, 0.00000327, 0.00000318, 0.00000302, 0.00000277, 0.00000245, 0.00000205, 0.00000157, 0.00000102, 0.00000039, -0.00000028, -0.00000101, -0.00000177, -0.00000254, -0.00000332, -0.00000409, -0.00000483, -0.00000552, -0.00000614, -0.00000667, -0.00000711, -0.00000743, -0.00000763, -0.00000768, -0.00000758, -0.00000733, -0.00000692, -0.00000636, -0.00000563, -0.00000475, -0.00000374, -0.00000259, -0.00000132, 0.00000004, 0.00000148, 0.00000298, 0.00000451, 0.00000605, 0.00000758, 0.00000905, 0.00001046, 0.00001177, 0.00001296, 0.00001401, 0.00001490, 0.00001560, 0.00001610, 0.00001640, 0.00001647, 0.00001631, 0.00001593, 0.00001532, 0.00001448, 0.00001344, 0.00001219, 0.00001077, 0.00000917, 0.00000744, 0.00000558, 0.00000363, 0.00000162, -0.00000043, -0.00000248, -0.00000451, -0.00000649, -0.00000840, -0.00001020, -0.00001187, -0.00001339, -0.00001474, -0.00001590, -0.00001687, -0.00001762, -0.00001817, -0.00001849, -0.00001861, -0.00001851, -0.00001821, -0.00001772, -0.00001706, -0.00001623, -0.00001527, -0.00001418, -0.00001299, -0.00001173, -0.00001041, -0.00000907, -0.00000771, -0.00000637, -0.00000506, -0.00000381, -0.00000262, -0.00000152, -0.00000051, 0.00000040, 0.00000119, 0.00000187, 0.00000243, 0.00000288, 0.00000321, 0.00000344, 0.00000358, 0.00000362, 0.00000359, 0.00000349, 0.00000335, 0.00000316, 0.00000295, 0.00000273, 0.00000250, 0.00000229, 0.00000210, 0.00000194, 0.00000182, 0.00000173, 0.00000170, 0.00000170, 0.00000175, 0.00000184, 0.00000197, 0.00000213, 0.00000231, 0.00000250, 0.00000269, },
		{ -0.00000845, -0.00000845, -0.00000839, -0.00000824, -0.00000801, -0.00000768, -0.00000725, -0.00000672, -0.00000609, -0.00000535, -0.00000452, -0.00000360, -0.00000260, -0.00000154, -0.00000042, 0.00000073, 0.00000190, 0.00000307, 0.00000421, 0.00000531, 0.00000634, 0.00000728, 0.00000811, 0.00000880, 0.00000935, 0.00000974, 0.00000995, 0.00000996, 0.00000978, 0.00000940, 0.00000881, 0.00000802, 0.00000703, 0.00000587, 0.00000453, 0.00000303, 0.00000141, -0.00000033, -0.00000215, -0.00000403, -0.00000593, -0.00000783, -0.00000969, -0.00001149, -0.00001319, -0.00001475, -0.00001617, -0.00001740, -0.00001842, -0.00001922, -0.00001977, -0.00002007, -0.00002010, -0.00001986, -0.00001935, -0.00001858, -0.00001755, -0.00001627, -0.00001476, -0.00001305, -0.00001114, -0.00000909, -0.00000690, -0.00000461, -0.00000226, 0.00000012, 0.00000250, 0.00000485, 0.00000713, 0.00000931, 0.00001137, 0.00001327, 0.00001499, 0.00001651, 0.00001782, 0.00001890, 0.00001974, 0.00002034, 0.00002069, 0.00002080, 0.00002067, 0.00002032, 0.00001976, 0.00001899, 0.00001805, 0.00001695, 0.00001572, 0.00001437, 0.00001294, 0.00001145, 0.00000992, 0.00000838, 0.00000684, 0.00000534, 0.00000389, 0.00000250, 0.00000120, -0.00000002, -0.00000112, -0.00000212, -0.00000301, -0.00000378, -0.00000445, -0.00000500, -0.00000545, -0.00000581, -0.00000609, -0.00000629, -0.00000643, -0.00000653, -0.00000658, -0.00000662, -0.00000664, -0.00000665, -0.00000667, -0.00000671, -0.00000677, -0.00000685, -0.00000695, -0.00000708, -0.00000723, -0.00000740, -0.00000758, -0.00000777, -0.00000796, -0.00000813, -0.00000828, -0.00000839, },
		{ 0.00000036, -0.00000038, -0.00000118, -0.00000204, -0.00000296, -0.00000394, -0.00000498, -0.00000607, -0.00000722, -0.00000841, -0.00000963, -0.00001087, -0.00001212, -0.00001335, -0.00001456, -0.00001571, -0.00001680, -0.00001779, -0.00001867, -0.00001941, -0.00002000, -0.00002041, -0.00002063, -0.00002064, -0.00002043, -0.00001999, -0.00001930, -0.00001837, -0.00001720, -0.00001579, -0.00001415, -0.00001229, -0.00001022, -0.00000797, -0.00000557, -0.00000303, -0.00000038, 0.00000233, 0.00000508, 0.00000783, 0.00001054, 0.00001318, 0.00001570, 0.00001808, 0.00002028, 0.00002227, 0.00002401, 0.00002549, 0.00002667, 0.00002755, 0.00002810, 0.00002832, 0.00002821, 0.00002776, 0.00002698, 0.00002588, 0.00002448, 0.00002280, 0.00002087, 0.00001870, 0.00001634, 0.00001382, 0.00001118, 0.00000844, 0.00000566, 0.00000287, 0.00000012, -0.00000257, -0.00000514, -0.00000758, -0.00000985, -0.00001191, -0.00001374, -0.00001533, -0.00001666, -0.00001771, -0.00001848, -0.00001896, -0.00001916, -0.00001908, -0.00001873, -0.00001813, -0.00001730, -0.00001625, -0.00001502, -0.00001361, -0.00001208, -0.00001043, -0.00000870, -0.00000693, -0.00000513, -0.00000334, -0.00000158, 0.00000012, 0.00000175, 0.00000327, 0.00000469, 0.00000597, 0.00000712, 0.00000813, 0.00000898, 0.00000968, 0.00001024, 0.00001065, 0.00001092, 0.00001106, 0.00001108, 0.00001099, 0.00001080, 0.00001053, 0.00001019, 0.00000979, 0.00000935, 0.00000887, 0.00000836, 0.00000784, 0.00000730, 0.00000676, 0.00000622, 0.00000568, 0.00000514, 0.00000460, 0.00000406, 0.00000350, 0.00000293, 0.00000234, 0.00000172, 0.00000106, },
		{ 0.00000073, 0.00000121, 0.00000170, 0.00000219, 0.00000269, 0.00000321, 0.00000373, 0.00000426, 0.00000480, 0.00000533, 0.00000584, 0.00000634, 0.00000680, 0.00000721, 0.00000756, 0.00000784, 0.00000803, 0.00000811, 0.00000806, 0.00000789, 0.00000756, 0.00000708, 0.00000642, 0.00000559, 0.00000458, 0.00000339, 0.00000201, 0.00000046, -0.00000126, -0.00000314, -0.00000515, -0.00000729, -0.00000953, -0.00001184, -0.00001421, -0.00001658, -0.00001895, -0.00002127, -0.00002351, -0.00002563, -0.00002761, -0.00002940, -0.00003098, -0.00003232, -0.00003339, -0.00003418, -0.00003465, -0.00003479, -0.00003460, -0.00003406, -0.00003318, -0.00003195, -0.00003039, -0.00002851, -0.00002633, -0.00002387, -0.00002116, -0.00001823, -0.00001512, -0.00001185, -0.00000848, -0.00000505, -0.00000159, 0.00000185, 0.00000523, 0.00000850, 0.00001163, 0.00001457, 0.00001730, 0.00001978, 0.00002198, 0.00002389, 0.00002547, 0.00002672, 0.00002763, 0.00002820, 0.00002842, 0.00002830, 0.00002786, 0.00002711, 0.00002606, 0.00002475, 0.00002320, 0.00002144, 0.00001949, 0.00001740, 0.00001520, 0.00001293, 0.00001061, 0.00000828, 0.00000597, 0.00000372, 0.00000155, -0.00000051, -0.00000244, -0.00000422, -0.00000583, -0.00000727, -0.00000852, -0.00000958, -0.00001044, -0.00001112, -0.00001161, -0.00001192, -0.00001206, -0.00001205, -0.00001189, -0.00001161, -0.00001122, -0.00001074, -0.00001018, -0.00000956, -0.00000889, -0.00000820, -0.00000749, -0.00000678, -0.00000607, -0.00000537, -0.00000470, -0.00000405, -0.00000343, -0.00000284, -0.00000227, -0.00000173, -0.00000121, -0.00000071, -0.00000022, 0.00000026, },
		{ 0.00000071, 0.00000090, 0.00000108, 0.00000124, 0.00000137, 0.00000147, 0.00000154, 0.00000157, 0.00000157, 0.00000153, 0.00000146, 0.00000136, 0.00000125, 0.00000113, 0.00000100, 0.00000089, 0.00000079, 0.00000073, 0.00000072, 0.00000076, 0.00000086, 0.00000103, 0.00000129, 0.00000163, 0.00000206, 0.00000258, 0.00000319, 0.00000389, 0.00000467, 0.00000551, 0.00000642, 0.00000737, 0.00000834, 0.00000933, 0.00001030, 0.00001124, 0.00001212, 0.00001292, 0.00001362, 0.00001419, 0.00001462, 0.00001487, 0.00001493, 0.00001478, 0.00001441, 0.00001381, 0.00001297, 0.00001187, 0.00001053, 0.00000895, 0.00000713, 0.00000508, 0.00000283, 0.00000038, -0.00000223, -0.00000499, -0.00000786, -0.00001081, -0.00001381, -0.00001683, -0.00001982, -0.00002275, -0.00002558, -0.00002829, -0.00003083, -0.00003317, -0.00003529, -0.00003715, -0.00003874, -0.00004004, -0.00004102, -0.00004169, -0.00004203, -0.00004204, -0.00004173, -0.00004110, -0.00004016, -0.00003894, -0.00003745, -0.00003572, -0.00003376, -0.00003162, -0.00002932, -0.00002690, -0.00002438, -0.00002181, -0.00001922, -0.00001663, -0.00001409, -0.00001161, -0.00000924, -0.00000699, -0.00000489, -0.00000295, -0.00000119, 0.00000038, 0.00000175, 0.00000292, 0.00000389, 0.00000465, 0.00000523, 0.00000562, 0.00000584, 0.00000590, 0.00000582, 0.00000562, 0.00000532, 0.00000493, 0.00000448, 0.00000398, 0.00000346, 0.00000293, 0.00000241, 0.00000191, 0.00000145, 0.00000104, 0.00000068, 0.00000038, 0.00000014, -0.00000003, -0.00000013, -0.00000018, -0.00000016, -0.00000009, 0.00000002, 0.00000016, 0.00000033, 0.00000052, },
		{ -0.00000110, -0.00000159, -0.00000203, -0.00000241, -0.00000274, -0.00000299, -0.00000318, -0.00000329, -0.00000333, -0.00000331, -0.00000322, -0.00000307, -0.00000289, -0.00000266, -0.00000241, -0.00000215, -0.00000189, -0.00000165, -0.00000144, -0.00000127, -0.00000117, -0.00000113, -0.00000116, -0.00000129, -0.00000152, -0.00000184, -0.00000227, -0.00000281, -0.00000344, -0.00000418, -0.00000500, -0.00000590, -0.00000687, -0.00000790, -0.00000897, -0.00001005, -0.00001114, -0.00001222, -0.00001326, -0.00001424, -0.00001514, -0.00001596, -0.00001665, -0.00001723, -0.00001765, -0.00001793, -0.00001804, -0.00001797, -0.00001774, -0.00001732, -0.00001673, -0.00001597, -0.00001505, -0.00001398, -0.00001277, -0.00001143, -0.00001000, -0.00000848, -0.00000691, -0.00000529, -0.00000366, -0.00000204, -0.00000045, 0.00000108, 0.00000254, 0.00000390, 0.00000514, 0.00000624, 0.00000720, 0.00000800, 0.00000863, 0.00000909, 0.00000937, 0.00000948, 0.00000942, 0.00000919, 0.00000881, 0.00000829, 0.00000764, 0.00000688, 0.00000603, 0.00000510, 0.00000413, 0.00000312, 0.00000211, 0.00000110, 0.00000013, -0.00000079, -0.00000165, -0.00000242, -0.00000310, -0.00000367, -0.00000413, -0.00000446, -0.00000467, -0.00000475, -0.00000471, -0.00000456, -0.00000429, -0.00000393, -0.00000347, -0.00000294, -0.00000235, -0.00000171, -0.00000105, -0.00000037, 0.00000030, 0.00000096, 0.00000158, 0.00000215, 0.00000266, 0.00000309, 0.00000345, 0.00000372, 0.00000389, 0.00000397, 0.00000396, 0.00000385, 0.00000365, 0.00000337, 0.00000302, 0.00000260, 0.00000213, 0.00000162, 0.00000107, 0.00000052, -0.00000004, -0.00000058, },
		{ -0.00000258, -0.00000247, -0.00000244, -0.00000249, -0.00000261, -0.00000280, -0.00000307, -0.00000339, -0.00000377, -0.00000418, -0.00000463, -0.00000509, -0.00000555, -0.00000600, -0.00000642, -0.00000680, -0.00000711, -0.00000734, -0.00000748, -0.00000751, -0.00000741, -0.00000718, -0.00000681, -0.00000628, -0.00000560, -0.00000475, -0.00000375, -0.00000258, -0.00000125, 0.00000022, 0.00000184, 0.00000358, 0.00000543, 0.00000737, 0.00000939, 0.00001147, 0.00001357, 0.00001569, 0.00001779, 0.00001985, 0.00002184, 0.00002375, 0.00002555, 0.00002721, 0.00002872, 0.00003006, 0.00003120, 0.00003215, 0.00003288, 0.00003339, 0.00003367, 0.00003372, 0.00003354, 0.00003313, 0.00003250, 0.00003166, 0.00003061, 0.00002938, 0.00002797, 0.00002641, 0.00002471, 0.00002290, 0.00002099, 0.00001900, 0.00001697, 0.00001490, 0.00001283, 0.00001076, 0.00000873, 0.00000674, 0.00000481, 0.00000296, 0.00000120, -0.00000046, -0.00000201, -0.00000345, -0.00000478, -0.00000598, -0.00000707, -0.00000804, -0.00000890, -0.00000965, -0.00001029, -0.00001084, -0.00001131, -0.00001169, -0.00001201, -0.00001226, -0.00001246, -0.00001262, -0.00001273, -0.00001282, -0.00001288, -0.00001292, -0.00001295, -0.00001296, -0.00001296, -0.00001295, -0.00001292, -0.00001289, -0.00001284, -0.00001277, -0.00001268, -0.00001256, -0.00001242, -0.00001225, -0.00001205, -0.00001181, -0.00001153, -0.00001121, -0.00001085, -0.00001046, -0.00001002, -0.00000955, -0.00000905, -0.00000852, -0.00000797, -0.00000740, -0.00000683, -0.00000626, -0.00000569, -0.00000515, -0.00000463, -0.00000415, -0.00000372, -0.00000334, -0.00000302, -0.00000276, },
		{ -0.00000253, -0.00000142, -0.00000028, 0.00000089, 0.00000207, 0.00000325, 0.00000442, 0.00000557, 0.00000669, 0.00000775, 0.00000876, 0.00000970, 0.00001056, 0.00001131, 0.00001196, 0.00001249, 0.00001289, 0.00001315, 0.00001325, 0.00001321, 0.00001300, 0.00001262, 0.00001208, 0.00001136, 0.00001048, 0.00000943, 0.00000823, 0.00000687, 0.00000537, 0.00000373, 0.00000199, 0.00000014, -0.00000179, -0.00000379, -0.00000583, -0.00000789, -0.00000996, -0.00001199, -0.00001398, -0.00001590, -0.00001773, -0.00001943, -0.00002100, -0.00002240, -0.00002363, -0.00002465, -0.00002546, -0.00002605, -0.00002640, -0.00002650, -0.00002636, -0.00002595, -0.00002530, -0.00002440, -0.00002325, -0.00002188, -0.00002028, -0.00001847, -0.00001648, -0.00001432, -0.00001201, -0.00000958, -0.00000705, -0.00000445, -0.00000180, 0.00000087, 0.00000354, 0.00000616, 0.00000873, 0.00001121, 0.00001359, 0.00001583, 0.00001792, 0.00001984, 0.00002157, 0.00002311, 0.00002443, 0.00002553, 0.00002641, 0.00002706, 0.00002748, 0.00002766, 0.00002763, 0.00002738, 0.00002692, 0.00002626, 0.00002542, 0.00002440, 0.00002323, 0.00002192, 0.00002049, 0.00001895, 0.00001733, 0.00001564, 0.00001390, 0.00001213, 0.00001034, 0.00000856, 0.00000679, 0.00000506, 0.00000337, 0.00000175, 0.00000019, -0.00000128, -0.00000266, -0.00000395, -0.00000513, -0.00000619, -0.00000714, -0.00000798, -0.00000869, -0.00000928, -0.00000975, -0.00001009, -0.00001031, -0.00001041, -0.00001040, -0.00001027, -0.00001002, -0.00000967, -0.00000922, -0.00000866, -0.00000802, -0.00000728, -0.00000647, -0.00000558, -0.00000462, -0.00000360, },
		{ -0.00000356, -0.00000491, -0.00000617, -0.00000734, -0.00000842, -0.00000939, -0.00001026, -0.00001102, -0.00001167, -0.00001220, -0.00001263, -0.00001294, -0.00001315, -0.00001324, -0.00001323, -0.00001311, -0.00001289, -0.00001257, -0.00001216, -0.00001165, -0.00001104, -0.00001035, -0.00000957, -0.00000872, -0.00000778, -0.00000677, -0.00000570, -0.00000456, -0.00000337, -0.00000214, -0.00000086, 0.00000044, 0.00000177, 0.00000311, 0.00000445, 0.00000577, 0.00000706, 0.00000832, 0.00000952, 0.00001066, 0.00001171, 0.00001266, 0.00001351, 0.00001422, 0.00001480, 0.00001523, 0.00001549, 0.00001558, 0.00001549, 0.00001521, 0.00001474, 0.00001408, 0.00001321, 0.00001215, 0.00001091, 0.00000948, 0.00000788, 0.00000612, 0.00000421, 0.00000217, 0.00000003, -0.00000220, -0.00000450, -0.00000684, -0.00000920, -0.00001154, -0.00001385, -0.00001609, -0.00001824, -0.00002027, -0.00002215, -0.00002387, -0.00002540, -0.00002673, -0.00002782, -0.00002867, -0.00002927, -0.00002960, -0.00002966, -0.00002946, -0.00002898, -0.00002823, -0.00002722, -0.00002597, -0.00002448, -0.00002278, -0.00002087, -0.00001879, -0.00001656, -0.00001420, -0.00001175, -0.00000922, -0.00000665, -0.00000407, -0.00000150, 0.00000102, 0.00000348, 0.00000583, 0.00000807, 0.00001016, 0.00001208, 0.00001383, 0.00001538, 0.00001673, 0.00001785, 0.00001876, 0.00001943, 0.00001988, 0.00002010, 0.00002010, 0.00001989, 0.00001948, 0.00001887, 0.00001808, 0.00001713, 0.00001603, 0.00001480, 0.00001346, 0.00001203, 0.00001052, 0.00000896, 0.00000735, 0.00000573, 0.00000410, 0.00000249, 0.00000090, -0.00000064, -0.00000213, },
		{ 0.00001124, 0.00001153, 0.00001167, 0.00001167, 0.00001153, 0.00001127, 0.00001090, 0.00001042, 0.00000986, 0.00000922, 0.00000853, 0.00000779, 0.00000702, 0.00000623, 0.00000544, 0.00000465, 0.00000387, 0.00000311, 0.00000237, 0.00000166, 0.00000098, 0.00000033, -0.00000030, -0.00000090, -0.00000149, -0.00000206, -0.00000264, -0.00000321, -0.00000380, -0.00000440, -0.00000503, -0.00000569, -0.00000638, -0.00000711, -0.00000789, -0.00000870, -0.00000956, -0.00001045, -0.00001137, -0.00001232, -0.00001328, -0.00001423, -0.00001518, -0.00001609, -0.00001696, -0.00001777, -0.00001850, -0.00001913, -0.00001965, -0.00002003, -0.00002027, -0.00002034, -0.00002024, -0.00001995, -0.00001946, -0.00001878, -0.00001790, -0.00001681, -0.00001553, -0.00001406, -0.00001241, -0.00001060, -0.00000865, -0.00000657, -0.00000439, -0.00000214, 0.00000016, 0.00000248, 0.00000478, 0.00000704, 0.00000923, 0.00001131, 0.00001327, 0.00001506, 0.00001666, 0.00001806, 0.00001922, 0.00002014, 0.00002081, 0.00002120, 0.00002132, 0.00002117, 0.00002076, 0.00002008, 0.00001915, 0.00001799, 0.00001662, 0.00001505, 0.00001333, 0.00001146, 0.00000949, 0.00000745, 0.00000536, 0.00000327, 0.00000120, -0.00000081, -0.00000273, -0.00000454, -0.00000621, -0.00000770, -0.00000902, -0.00001012, -0.00001102, -0.00001168, -0.00001211, -0.00001231, -0.00001227, -0.00001201, -0.00001154, -0.00001086, -0.00001000, -0.00000897, -0.00000780, -0.00000651, -0.00000512, -0.00000366, -0.00000216, -0.00000063, 0.00000088, 0.00000237, 0.00000380, 0.00000515, 0.00000641, 0.00000756, 0.00000858, 0.00000947, 0.00001021, 0.00001080, },
		{ -0.00000728, -0.00000711, -0.00000681, -0.00000639, -0.00000587, -0.00000528, -0.00000462, -0.00000393, -0.00000322, -0.00000252, -0.00000185, -0.00000122, -0.00000065, -0.00000016, 0.00000024, 0.00000055, 0.00000076, 0.00000088, 0.00000090, 0.00000083, 0.00000069, 0.00000048, 0.00000023, -0.00000006, -0.00000035, -0.00000063, -0.00000088, -0.00000107, -0.00000117, -0.00000117, -0.00000105, -0.00000079, -0.00000038, 0.00000020, 0.00000095, 0.00000189, 0.00000299, 0.00000427, 0.00000572, 0.00000731, 0.00000903, 0.00001086, 0.00001277, 0.00001473, 0.00001672, 0.00001869, 0.00002061, 0.00002245, 0.00002416, 0.00002571, 0.00002707, 0.00002821, 0.00002908, 0.00002968, 0.00002997, 0.00002994, 0.00002957, 0.00002887, 0.00002782, 0.00002643, 0.00002472, 0.00002269, 0.00002038, 0.00001781, 0.00001500, 0.00001200, 0.00000885, 0.00000559, 0.00000227, -0.00000108, -0.00000439, -0.00000763, -0.00001074, -0.00001369, -0.00001642, -0.00001891, -0.00002111, -0.00002301, -0.00002456, -0.00002577, -0.00002660, -0.00002707, -0.00002716, -0.00002688, -0.00002624, -0.00002528, -0.00002399, -0.00002243, -0.00002061, -0.00001857, -0.00001636, -0.00001402, -0.00001158, -0.00000910, -0.00000661, -0.00000417, -0.00000180, 0.00000045, 0.00000254, 0.00000445, 0.00000615, 0.00000762, 0.00000883, 0.00000978, 0.00001047, 0.00001088, 0.00001104, 0.00001093, 0.00001059, 0.00001002, 0.00000925, 0.00000831, 0.00000722, 0.00000601, 0.00000472, 0.00000337, 0.00000201, 0.00000065, -0.00000066, -0.00000192, -0.00000308, -0.00000412, -0.00000504, -0.00000581, -0.00000643, -0.00000689, -0.00000718, -0.00000731, },
		{ -0.00000363, -0.00000386, -0.00000420, -0.00000465, -0.00000516, -0.00000571, -0.00000628, -0.00000682, -0.00000732, -0.00000775, -0.00000807, -0.00000828, -0.00000836, -0.00000828, -0.00000806, -0.00000767, -0.00000714, -0.00000645, -0.00000564, -0.00000471, -0.00000368, -0.00000259, -0.00000147, -0.00000034, 0.00000075, 0.00000179, 0.00000271, 0.00000351, 0.00000413, 0.00000456, 0.00000476, 0.00000471, 0.00000440, 0.00000381, 0.00000294, 0.00000178, 0.00000034, -0.00000136, -0.00000330, -0.00000546, -0.00000780, -0.00001030, -0.00001289, -0.00001555, -0.00001821, -0.00002083, -0.00002335, -0.00002572, -0.00002788, -0.00002979, -0.00003140, -0.00003266, -0.00003353, -0.00003399, -0.00003400, -0.00003355, -0.00003262, -0.00003122, -0.00002935, -0.00002702, -0.00002426, -0.00002110, -0.00001757, -0.00001372, -0.00000961, -0.00000528, -0.00000081, 0.00000375, 0.00000832, 0.00001284, 0.00001724, 0.00002146, 0.00002543, 0.00002909, 0.00003238, 0.00003526, 0.00003769, 0.00003962, 0.00004104, 0.00004192, 0.00004226, 0.00004205, 0.00004130, 0.00004004, 0.00003828, 0.00003606, 0.00003342, 0.00003041, 0.00002708, 0.00002349, 0.00001969, 0.00001575, 0.00001174, 0.00000771, 0.00000372, -0.00000016, -0.00000388, -0.00000739, -0.00001064, -0.00001359, -0.00001622, -0.00001849, -0.00002039, -0.00002190, -0.00002302, -0.00002376, -0.00002413, -0.00002415, -0.00002384, -0.00002323, -0.00002235, -0.00002125, -0.00001996, -0.00001853, -0.00001699, -0.00001540, -0.00001379, -0.00001220, -0.00001068, -0.00000924, -0.00000794, -0.00000677, -0.00000578, -0.00000496, -0.00000433, -0.00000389, -0.00000363, -0.00000355, },
		{ -0.00000122, 0.00000011, 0.00000153, 0.00000299, 0.00000447, 0.00000591, 0.00000727, 0.00000850, 0.00000958, 0.00001046, 0.00001112, 0.00001153, 0.00001167, 0.00001155, 0.00001114, 0.00001046, 0.00000952, 0.00000834, 0.00000694, 0.00000536, 0.00000364, 0.00000181, -0.00000007, -0.00000196, -0.00000381, -0.00000557, -0.00000718, -0.00000860, -0.00000979, -0.00001071, -0.00001132, -0.00001161, -0.00001153, -0.00001110, -0.00001030, -0.00000914, -0.00000763, -0.00000580, -0.00000368, -0.00000130, 0.00000127, 0.00000400, 0.00000682, 0.00000967, 0.00001248, 0.00001518, 0.00001771, 0.00002000, 0.00002198, 0.00002359, 0.00002478, 0.00002550, 0.00002571, 0.00002537, 0.00002446, 0.00002297, 0.00002089, 0.00001824, 0.00001504, 0.00001130, 0.00000708, 0.00000242, -0.00000262, -0.00000798, -0.00001357, -0.00001932, -0.00002515, -0.00003098, -0.00003671, -0.00004227, -0.00004757, -0.00005252, -0.00005706, -0.00006112, -0.00006464, -0.00006757, -0.00006986, -0.00007148, -0.00007242, -0.00007267, -0.00007223, -0.00007111, -0.00006935, -0.00006697, -0.00006402, -0.00006055, -0.00005664, -0.00005234, -0.00004773, -0.00004289, -0.00003789, -0.00003282, -0.00002775, -0.00002277, -0.00001793, -0.00001332, -0.00000899, -0.00000499, -0.00000138, 0.00000182, 0.00000457, 0.00000685, 0.00000867, 0.00001001, 0.00001090, 0.00001135, 0.00001139, 0.00001106, 0.00001040, 0.00000945, 0.00000827, 0.00000691, 0.00000543, 0.00000388, 0.00000232, 0.00000080, -0.00000064, -0.00000193, -0.00000306, -0.00000397, -0.00000466, -0.00000509, -0.00000527, -0.00000518, -0.00000484, -0.00000424, -0.00000342, -0.00000240, },
		{ -0.00000557, -0.00000779, -0.00001002, -0.00001221, -0.00001430, -0.00001623, -0.00001794, -0.00001941, -0.00002058, -0.00002141, -0.00002190, -0.00002202, -0.00002176, -0.00002114, -0.00002015, -0.00001884, -0.00001722, -0.00001534, -0.00001325, -0.00001101, -0.00000868, -0.00000631, -0.00000399, -0.00000177, 0.00000027, 0.00000206, 0.00000355, 0.00000468, 0.00000538, 0.00000564, 0.00000539, 0.00000464, 0.00000335, 0.00000154, -0.00000079, -0.00000362, -0.00000691, -0.00001062, -0.00001469, -0.00001906, -0.00002365, -0.00002839, -0.00003318, -0.00003794, -0.00004258, -0.00004700, -0.00005112, -0.00005484, -0.00005809, -0.00006078, -0.00006286, -0.00006426, -0.00006495, -0.00006487, -0.00006403, -0.00006240, -0.00006000, -0.00005684, -0.00005296, -0.00004841, -0.00004324, -0.00003753, -0.00003135, -0.00002479, -0.00001796, -0.00001094, -0.00000384, 0.00000322, 0.00001016, 0.00001686, 0.00002323, 0.00002918, 0.00003462, 0.00003948, 0.00004370, 0.00004723, 0.00005002, 0.00005205, 0.00005332, 0.00005381, 0.00005356, 0.00005257, 0.00005091, 0.00004860, 0.00004573, 0.00004236, 0.00003856, 0.00003442, 0.00003003, 0.00002548, 0.00002085, 0.00001624, 0.00001172, 0.00000738, 0.00000329, -0.00000048, -0.00000388, -0.00000687, -0.00000939, -0.00001144, -0.00001300, -0.00001407, -0.00001465, -0.00001476, -0.00001444, -0.00001373, -0.00001267, -0.00001132, -0.00000973, -0.00000797, -0.00000610, -0.00000419, -0.00000230, -0.00000050, 0.00000116, 0.00000262, 0.00000384, 0.00000478, 0.00000540, 0.00000570, 0.00000565, 0.00000525, 0.00000451, 0.00000345, 0.00000209, 0.00000046, -0.00000139, -0.00000341, },
		{ -0.00001277, -0.00001070, -0.00000874, -0.00000693, -0.00000533, -0.00000400, -0.00000298, -0.00000231, -0.00000201, -0.00000211, -0.00000260, -0.00000348, -0.00000474, -0.00000634, -0.00000825, -0.00001042, -0.00001278, -0.00001527, -0.00001781, -0.00002033, -0.00002274, -0.00002496, -0.00002690, -0.00002849, -0.00002963, -0.00003027, -0.00003034, -0.00002979, -0.00002857, -0.00002665, -0.00002401, -0.00002066, -0.00001660, -0.00001187, -0.00000650, -0.00000054, 0.00000592, 0.00001282, 0.00002005, 0.00002751, 0.00003510, 0.00004270, 0.00005019, 0.00005744, 0.00006434, 0.00007077, 0.00007662, 0.00008178, 0.00008615, 0.00008966, 0.00009224, 0.00009382, 0.00009437, 0.00009386, 0.00009230, 0.00008968, 0.00008604, 0.00008143, 0.00007590, 0.00006954, 0.00006242, 0.00005467, 0.00004638, 0.00003768, 0.00002870, 0.00001956, 0.00001040, 0.00000135, -0.00000747, -0.00001593, -0.00002391, -0.00003133, -0.00003807, -0.00004407, -0.00004926, -0.00005360, -0.00005704, -0.00005957, -0.00006119, -0.00006192, -0.00006180, -0.00006085, -0.00005915, -0.00005677, -0.00005378, -0.00005028, -0.00004637, -0.00004214, -0.00003770, -0.00003315, -0.00002859, -0.00002412, -0.00001984, -0.00001581, -0.00001212, -0.00000884, -0.00000601, -0.00000367, -0.00000185, -0.00000056, 0.00000019, 0.00000043, 0.00000016, -0.00000056, -0.00000170, -0.00000320, -0.00000501, -0.00000705, -0.00000925, -0.00001154, -0.00001386, -0.00001612, -0.00001826, -0.00002023, -0.00002196, -0.00002340, -0.00002453, -0.00002530, -0.00002570, -0.00002573, -0.00002539, -0.00002468, -0.00002364, -0.00002230, -0.00002070, -0.00001889, -0.00001692, -0.00001486, },
		{ -0.00001557, -0.00001631, -0.00001683, -0.00001707, -0.00001698, -0.00001652, -0.00001566, -0.00001437, -0.00001266, -0.00001053, -0.00000799, -0.00000508, -0.00000183, 0.00000169, 0.00000543, 0.00000930, 0.00001322, 0.00001710, 0.00002086, 0.00002438, 0.00002758, 0.00003036, 0.00003263, 0.00003430, 0.00003530, 0.00003556, 0.00003503, 0.00003367, 0.00003146, 0.00002839, 0.00002447, 0.00001971, 0.00001418, 0.00000792, 0.00000101, -0.00000645, -0.00001437, -0.00002263, -0.00003109, -0.00003964, -0.00004813, -0.00005641, -0.00006435, -0.00007182, -0.00007867, -0.00008478, -0.00009004, -0.00009435, -0.00009761, -0.00009976, -0.00010074, -0.00010052, -0.00009909, -0.00009646, -0.00009264, -0.00008770, -0.00008168, -0.00007469, -0.00006682, -0.00005818, -0.00004892, -0.00003915, -0.00002905, -0.00001875, -0.00000841, 0.00000180, 0.00001175, 0.00002127, 0.00003024, 0.00003852, 0.00004599, 0.00005257, 0.00005816, 0.00006271, 0.00006615, 0.00006848, 0.00006967, 0.00006974, 0.00006873, 0.00006668, 0.00006365, 0.00005974, 0.00005502, 0.00004962, 0.00004364, 0.00003721, 0.00003045, 0.00002349, 0.00001646, 0.00000948, 0.00000266, -0.00000387, -0.00001003, -0.00001572, -0.00002087, -0.00002541, -0.00002931, -0.00003252, -0.00003504, -0.00003686, -0.00003800, -0.00003848, -0.00003835, -0.00003767, -0.00003648, -0.00003487, -0.00003292, -0.00003069, -0.00002828, -0.00002576, -0.00002322, -0.00002074, -0.00001837, -0.00001619, -0.00001425, -0.00001259, -0.00001124, -0.00001022, -0.00000954, -0.00000920, -0.00000918, -0.00000945, -0.00000998, -0.00001073, -0.00001162, -0.00001262, -0.00001366, -0.00001466, },
		{ -0.00001457, -0.00001391, -0.00001360, -0.00001371, -0.00001427, -0.00001531, -0.00001686, -0.00001892, -0.00002147, -0.00002449, -0.00002794, -0.00003176, -0.00003589, -0.00004025, -0.00004474, -0.00004928, -0.00005375, -0.00005804, -0.00006206, -0.00006568, -0.00006880, -0.00007131, -0.00007314, -0.00007419, -0.00007439, -0.00007370, -0.00007207, -0.00006948, -0.00006593, -0.00006143, -0.00005603, -0.00004977, -0.00004274, -0.00003501, -0.00002670, -0.00001794, -0.00000884, 0.00000044, 0.00000976, 0.00001895, 0.00002787, 0.00003635, 0.00004425, 0.00005142, 0.00005773, 0.00006305, 0.00006729, 0.00007034, 0.00007215, 0.00007266, 0.00007184, 0.00006970, 0.00006624, 0.00006152, 0.00005559, 0.00004854, 0.00004046, 0.00003148, 0.00002174, 0.00001139, 0.00000057, -0.00001053, -0.00002175, -0.00003291, -0.00004385, -0.00005440, -0.00006441, -0.00007371, -0.00008219, -0.00008973, -0.00009621, -0.00010157, -0.00010574, -0.00010869, -0.00011039, -0.00011086, -0.00011011, -0.00010819, -0.00010518, -0.00010115, -0.00009620, -0.00009046, -0.00008403, -0.00007706, -0.00006968, -0.00006204, -0.00005427, -0.00004652, -0.00003892, -0.00003160, -0.00002467, -0.00001823, -0.00001238, -0.00000718, -0.00000270, 0.00000102, 0.00000397, 0.00000613, 0.00000752, 0.00000816, 0.00000811, 0.00000740, 0.00000612, 0.00000434, 0.00000215, -0.00000037, -0.00000312, -0.00000600, -0.00000893, -0.00001181, -0.00001456, -0.00001711, -0.00001939, -0.00002135, -0.00002295, -0.00002416, -0.00002496, -0.00002536, -0.00002537, -0.00002502, -0.00002433, -0.00002338, -0.00002220, -0.00002087, -0.00001946, -0.00001805, -0.00001671, -0.00001553, },
		{ -0.00006803, -0.00006849, -0.00006831, -0.00006743, -0.00006585, -0.00006356, -0.00006058, -0.00005694, -0.00005269, -0.00004789, -0.00004262, -0.00003698, -0.00003108, -0.00002504, -0.00001897, -0.00001302, -0.00000731, -0.00000199, 0.00000282, 0.00000699, 0.00001039, 0.00001294, 0.00001452, 0.00001507, 0.00001453, 0.00001285, 0.00001003, 0.00000606, 0.00000097, -0.00000518, -0.00001232, -0.00002038, -0.00002922, -0.00003872, -0.00004873, -0.00005910, -0.00006965, -0.00008021, -0.00009059, -0.00010061, -0.00011010, -0.00011887, -0.00012677, -0.00013364, -0.00013933, -0.00014374, -0.00014677, -0.00014833, -0.00014837, -0.00014687, -0.00014382, -0.00013925, -0.00013321, -0.00012575, -0.00011700, -0.00010705, -0.00009606, -0.00008416, -0.00007154, -0.00005837, -0.00004484, -0.00003114, -0.00001746, -0.00000400, 0.00000907, 0.00002156, 0.00003331, 0.00004417, 0.00005402, 0.00006273, 0.00007022, 0.00007642, 0.00008128, 0.00008478, 0.00008692, 0.00008773, 0.00008724, 0.00008554, 0.00008269, 0.00007880, 0.00007399, 0.00006838, 0.00006211, 0.00005532, 0.00004815, 0.00004075, 0.00003324, 0.00002578, 0.00001847, 0.00001144, 0.00000479, -0.00000140, -0.00000705, -0.00001212, -0.00001656, -0.00002036, -0.00002351, -0.00002603, -0.00002796, -0.00002933, -0.00003020, -0.00003064, -0.00003073, -0.00003055, -0.00003017, -0.00002970, -0.00002921, -0.00002879, -0.00002851, -0.00002844, -0.00002863, -0.00002914, -0.00002999, -0.00003121, -0.00003279, -0.00003474, -0.00003702, -0.00003961, -0.00004245, -0.00004548, -0.00004863, -0.00005182, -0.00005496, -0.00005798, -0.00006077, -0.00006325, -0.00006534, -0.00006696, },
		{ -0.00000521, -0.00001211, -0.00001968, -0.00002790, -0.00003671, -0.00004606, -0.00005585, -0.00006599, -0.00007637, -0.00008685, -0.00009730, -0.00010757, -0.00011750, -0.00012693, -0.00013571, -0.00014366, -0.00015064, -0.00015650, -0.00016111, -0.00016435, -0.00016612, -0.00016633, -0.00016493, -0.00016187, -0.00015716, -0.00015080, -0.00014284, -0.00013335, -0.00012243, -0.00011019, -0.00009679, -0.00008238, -0.00006717, -0.00005134, -0.00003511, -0.00001872, -0.00000239, 0.00001365, 0.00002915, 0.00004391, 0.00005769, 0.00007030, 0.00008156, 0.00009129, 0.00009935, 0.00010564, 0.00011005, 0.00011254, 0.00011306, 0.00011164, 0.00010828, 0.00010307, 0.00009609, 0.00008747, 0.00007734, 0.00006587, 0.00005325, 0.00003968, 0.00002539, 0.00001058, -0.00000450, -0.00001964, -0.00003459, -0.00004914, -0.00006309, -0.00007623, -0.00008838, -0.00009940, -0.00010913, -0.00011747, -0.00012433, -0.00012965, -0.00013340, -0.00013556, -0.00013615, -0.00013522, -0.00013283, -0.00012907, -0.00012405, -0.00011788, -0.00011072, -0.00010270, -0.00009398, -0.00008471, -0.00007506, -0.00006519, -0.00005524, -0.00004537, -0.00003570, -0.00002637, -0.00001747, -0.00000910, -0.00000133, 0.00000577, 0.00001217, 0.00001784, 0.00002277, 0.00002698, 0.00003048, 0.00003332, 0.00003554, 0.00003719, 0.00003833, 0.00003904, 0.00003938, 0.00003941, 0.00003920, 0.00003880, 0.00003828, 0.00003767, 0.00003701, 0.00003631, 0.00003560, 0.00003487, 0.00003411, 0.00003329, 0.00003239, 0.00003136, 0.00003014, 0.00002868, 0.00002692, 0.00002479, 0.00002223, 0.00001917, 0.00001557, 0.00001136, 0.00000652, 0.00000100, },
		{ 0.00001940, 0.00002674, 0.00003415, 0.00004157, 0.00004889, 0.00005603, 0.00006286, 0.00006925, 0.00007509, 0.00008022, 0.00008452, 0.00008784, 0.00009004, 0.00009100, 0.00009058, 0.00008868, 0.00008521, 0.00008008, 0.00007323, 0.00006464, 0.00005429, 0.00004219, 0.00002839, 0.00001296, -0.00000401, -0.00002238, -0.00004201, -0.00006271, -0.00008428, -0.00010650, -0.00012913, -0.00015191, -0.00017457, -0.00019684, -0.00021844, -0.00023910, -0.00025854, -0.00027650, -0.00029273, -0.00030702, -0.00031915, -0.00032894, -0.00033624, -0.00034094, -0.00034296, -0.00034223, -0.00033876, -0.00033256, -0.00032371, -0.00031230, -0.00029847, -0.00028239, -0.00026424, -0.00024427, -0.00022271, -0.00019985, -0.00017596, -0.00015134, -0.00012629, -0.00010113, -0.00007614, -0.00005164, -0.00002790, -0.00000519, 0.00001624, 0.00003617, 0.00005441, 0.00007077, 0.00008513, 0.00009737, 0.00010744, 0.00011528, 0.00012089, 0.00012430, 0.00012556, 0.00012477, 0.00012203, 0.00011748, 0.00011129, 0.00010361, 0.00009465, 0.00008460, 0.00007366, 0.00006205, 0.00004996, 0.00003759, 0.00002515, 0.00001281, 0.00000074, -0.00001090, -0.00002198, -0.00003239, -0.00004202, -0.00005080, -0.00005867, -0.00006559, -0.00007155, -0.00007655, -0.00008059, -0.00008372, -0.00008597, -0.00008738, -0.00008804, -0.00008799, -0.00008731, -0.00008606, -0.00008432, -0.00008214, -0.00007959, -0.00007673, -0.00007359, -0.00007021, -0.00006662, -0.00006285, -0.00005890, -0.00005478, -0.00005048, -0.00004599, -0.00004131, -0.00003640, -0.00003126, -0.00002586, -0.00002020, -0.00001426, -0.00000804, -0.00000154, 0.00000522, 0.00001221, },
		{ 0.00005267, 0.00005547, 0.00005827, 0.00006111, 0.00006406, 0.00006718, 0.00007054, 0.00007422, 0.00007829, 0.00008283, 0.00008790, 0.00009358, 0.00009992, 0.00010697, 0.00011477, 0.00012333, 0.00013265, 0.00014273, 0.00015353, 0.00016498, 0.00017701, 0.00018951, 0.00020236, 0.00021542, 0.00022850, 0.00024143, 0.00025400, 0.00026599, 0.00027717, 0.00028731, 0.00029616, 0.00030347, 0.00030901, 0.00031255, 0.00031386, 0.00031274, 0.00030901, 0.00030251, 0.00029311, 0.00028070, 0.00026523, 0.00024665, 0.00022499, 0.00020029, 0.00017263, 0.00014214, 0.00010900, 0.00007340, 0.00003558, -0.00000418, -0.00004558, -0.00008828, -0.00013195, -0.00017620, -0.00022065, -0.00026493, -0.00030863, -0.00035137, -0.00039276, -0.00043245, -0.00047009, -0.00050534, -0.00053792, -0.00056757, -0.00059405, -0.00061717, -0.00063678, -0.00065278, -0.00066509, -0.00067369, -0.00067860, -0.00067988, -0.00067762, -0.00067197, -0.00066309, -0.00065118, -0.00063647, -0.00061922, -0.00059968, -0.00057814, -0.00055491, -0.00053026, -0.00050451, -0.00047794, -0.00045084, -0.00042348, -0.00039613, -0.00036902, -0.00034237, -0.00031638, -0.00029122, -0.00026703, -0.00024392, -0.00022200, -0.00020132, -0.00018194, -0.00016386, -0.00014708, -0.00013157, -0.00011730, -0.00010421, -0.00009223, -0.00008128, -0.00007127, -0.00006212, -0.00005374, -0.00004603, -0.00003891, -0.00003230, -0.00002612, -0.00002031, -0.00001480, -0.00000955, -0.00000452, 0.00000033, 0.00000501, 0.00000955, 0.00001394, 0.00001820, 0.00002231, 0.00002628, 0.00003010, 0.00003376, 0.00003726, 0.00004061, 0.00004381, 0.00004687, 0.00004981, },
		{ 0.00015504, 0.00015521, 0.00015573, 0.00015657, 0.00015774, 0.00015920, 0.00016094, 0.00016293, 0.00016514, 0.00016755, 0.00017014, 0.00017288, 0.00017577, 0.00017879, 0.00018196, 0.00018530, 0.00018882, 0.00019257, 0.00019661, 0.00020101, 0.00020585, 0.00021123, 0.00021727, 0.00022409, 0.00023182, 0.00024061, 0.00025061, 0.00026195, 0.00027481, 0.00028932, 0.00030562, 0.00032385, 0.00034413, 0.00036655, 0.00039119, 0.00041810, 0.00044733, 0.00047886, 0.00051266, 0.00054867, 0.00058679, 0.00062688, 0.00066878, 0.00071229, 0.00075716, 0.00080312, 0.00084989, 0.00089713, 0.00094451, 0.00099165, 0.00103819, 0.00108373, 0.00112788, 0.00117025, 0.00121046, 0.00124812, 0.00128290, 0.00131444, 0.00134245, 0.00136664, 0.00138678, 0.00140266, 0.00141412, 0.00142104, 0.00142336, 0.00142104, 0.00141412, 0.00140266, 0.00138678, 0.00136664, 0.00134245, 0.00131444, 0.00128290, 0.00124812, 0.00121046, 0.00117025, 0.00112788, 0.00108373, 0.00103819, 0.00099165, 0.00094451, 0.00089713, 0.00084989, 0.00080312, 0.00075716, 0.00071229, 0.00066878, 0.00062688, 0.00058679, 0.00054867, 0.00051266, 0.00047886, 0.00044733, 0.00041810, 0.00039119, 0.00036655, 0.00034413, 0.00032385, 0.00030562, 0.00028932, 0.00027481, 0.00026195, 0.00025061, 0.00024061, 0.00023182, 0.00022409, 0.00021727, 0.00021123, 0.00020585, 0.00020101, 0.00019661, 0.00019257, 0.00018882, 0.00018530, 0.00018196, 0.00017879, 0.00017577, 0.00017288, 0.00017014, 0.00016755, 0.00016514, 0.00016293, 0.00016094, 0.00015920, 0.00015774, 0.00015657, 0.00015573, 0.00015521, },
		{ 0.00005267, 0.00004981, 0.00004687, 0.00004381, 0.00004061, 0.00003726, 0.00003376, 0.00003010, 0.00002628, 0.00002231, 0.00001820, 0.00001394, 0.00000955, 0.00000501, 0.00000033, -0.00000452, -0.00000955, -0.00001480, -0.00002031, -0.00002612, -0.00003230, -0.00003891, -0.00004603, -0.00005374, -0.00006212, -0.00007127, -0.00008128, -0.00009223, -0.00010421, -0.00011730, -0.00013157, -0.00014708, -0.00016386, -0.00018194, -0.00020132, -0.00022200, -0.00024392, -0.00026703, -0.00029122, -0.00031638, -0.00034237, -0.00036902, -0.00039613, -0.00042348, -0.00045084, -0.00047794, -0.00050451, -0.00053026, -0.00055491, -0.00057814, -0.00059968, -0.00061922, -0.00063647, -0.00065118, -0.00066309, -0.00067197, -0.00067762, -0.00067988, -0.00067860, -0.00067369, -0.00066509, -0.00065278, -0.00063678, -0.00061717, -0.00059405, -0.00056757, -0.00053792, -0.00050534, -0.00047009, -0.00043245, -0.00039276, -0.00035137, -0.00030863, -0.00026493, -0.00022065, -0.00017620, -0.00013195, -0.00008828, -0.00004558, -0.00000418, 0.00003558, 0.00007340, 0.00010900, 0.00014214, 0.00017263, 0.00020029, 0.00022499, 0.00024665, 0.00026523, 0.00028070, 0.00029311, 0.00030251, 0.00030901, 0.00031274, 0.00031386, 0.00031255, 0.00030901, 0.00030347, 0.00029616, 0.00028731, 0.00027717, 0.00026599, 0.00025400, 0.00024143, 0.00022850, 0.00021542, 0.00020236, 0.00018951, 0.00017701, 0.00016498, 0.00015353, 0.00014273, 0.00013265, 0.00012333, 0.00011477, 0.00010697, 0.00009992, 0.00009358, 0.00008790, 0.00008283, 0.00007829, 0.00007422, 0.00007054, 0.00006718, 0.00006406, 0.00006111, 0.00005827, 0.00005547, },
		{ 0.00001940, 0.00001221, 0.00000522, -0.00000154, -0.00000804, -0.00001426, -0.00002020, -0.00002586, -0.00003126, -0.00003640, -0.00004131, -0.00004599, -0.00005048, -0.00005478, -0.00005890, -0.00006285, -0.00006662, -0.00007021, -0.00007359, -0.00007673, -0.00007959, -0.00008214, -0.00008432, -0.00008606, -0.00008731, -0.00008799, -0.00008804, -0.00008738, -0.00008597, -0.00008372, -0.00008059, -0.00007655, -0.00007155, -0.00006559, -0.00005867, -0.00005080, -0.00004202, -0.00003239, -0.00002198, -0.00001090, 0.00000074, 0.00001281, 0.00002515, 0.00003759, 0.00004996, 0.00006205, 0.00007366, 0.00008460, 0.00009465, 0.00010361, 0.00011129, 0.00011748, 0.00012203, 0.00012477, 0.00012556, 0.00012430, 0.00012089, 0.00011528, 0.00010744, 0.00009737, 0.00008513, 0.00007077, 0.00005441, 0.00003617, 0.00001624, -0.00000519, -0.00002790, -0.00005164, -0.00007614, -0.00010113, -0.00012629, -0.00015134, -0.00017596, -0.00019985, -0.00022271, -0.00024427, -0.00026424, -0.00028239, -0.00029847, -0.00031230, -0.00032371, -0.00033256, -0.00033876, -0.00034223, -0.00034296, -0.00034094, -0.00033624, -0.00032894, -0.00031915, -0.00030702, -0.00029273, -0.00027650, -0.00025854, -0.00023910, -0.00021844, -0.00019684, -0.00017457, -0.00015191, -0.00012913, -0.00010650, -0.00008428, -0.00006271, -0.00004201, -0.00002238, -0.00000401, 0.00001296, 0.00002839, 0.00004219, 0.00005429, 0.00006464, 0.00007323, 0.00008008, 0.00008521, 0.00008868, 0.00009058, 0.00009100, 0.00009004, 0.00008784, 0.00008452, 0.00008022, 0.00007509, 0.00006925, 0.00006286, 0.00005603, 0.00004889, 0.00004157, 0.00003415, 0.00002674, },
		{ -0.00000521, 0.00000100, 0.00000652, 0.00001136, 0.00001557, 0.00001917, 0.00002223, 0.00002479, 0.00002692, 0.00002868, 0.00003014, 0.00003136, 0.00003239, 0.00003329, 0.00003411, 0.00003487, 0.00003560, 0.00003631, 0.00003701, 0.00003767, 0.00003828, 0.00003880, 0.00003920, 0.00003941, 0.00003938, 0.00003904, 0.00003833, 0.00003719, 0.00003554, 0.00003332, 0.00003048, 0.00002698, 0.00002277, 0.00001784, 0.00001217, 0.00000577, -0.00000133, -0.00000910, -0.00001747, -0.00002637, -0.00003570, -0.00004537, -0.00005524, -0.00006519, -0.00007506, -0.00008471, -0.00009398, -0.00010270, -0.00011072, -0.00011788, -0.00012405, -0.00012907, -0.00013283, -0.00013522, -0.00013615, -0.00013556, -0.00013340, -0.00012965, -0.00012433, -0.00011747, -0.00010913, -0.00009940, -0.00008838, -0.00007623, -0.00006309, -0.00004914, -0.00003459, -0.00001964, -0.00000450, 0.00001058, 0.00002539, 0.00003968, 0.00005325, 0.00006587, 0.00007734, 0.00008747, 0.00009609, 0.00010307, 0.00010828, 0.00011164, 0.00011306, 0.00011254, 0.00011005, 0.00010564, 0.00009935, 0.00009129, 0.00008156, 0.00007030, 0.00005769, 0.00004391, 0.00002915, 0.00001365, -0.00000239, -0.00001872, -0.00003511, -0.00005134, -0.00006717, -0.00008238, -0.00009679, -0.00011019, -0.00012243, -0.00013335, -0.00014284, -0.00015080, -0.00015716, -0.00016187, -0.00016493, -0.00016633, -0.00016612, -0.00016435, -0.00016111, -0.00015650, -0.00015064, -0.00014366, -0.00013571, -0.00012693, -0.00011750, -0.00010757, -0.00009730, -0.00008685, -0.00007637, -0.00006599, -0.00005585, -0.00004606, -0.00003671, -0.00002790, -0.00001968, -0.00001211, },
		{ -0.00006803, -0.00006696, -0.00006534, -0.00006325, -0.00006077, -0.00005798, -0.00005496, -0.00005182, -0.00004863, -0.00004548, -0.00004245, -0.00003961, -0.00003702, -0.00003474, -0.00003279, -0.00003121, -0.00002999, -0.00002914, -0.00002863, -0.00002844, -0.00002851, -0.00002879, -0.00002921, -0.00002970, -0.00003017, -0.00003055, -0.00003073, -0.00003064, -0.00003020, -0.00002933, -0.00002796, -0.00002603, -0.00002351, -0.00002036, -0.00001656, -0.00001212, -0.00000705, -0.00000140, 0.00000479, 0.00001144, 0.00001847, 0.00002578, 0.00003324, 0.00004075, 0.00004815, 0.00005532, 0.00006211, 0.00006838, 0.00007399, 0.00007880, 0.00008269, 0.00008554, 0.00008724, 0.00008773, 0.00008692, 0.00008478, 0.00008128, 0.00007642, 0.00007022, 0.00006273, 0.00005402, 0.00004417, 0.00003331, 0.00002156, 0.00000907, -0.00000400, -0.00001746, -0.00003114, -0.00004484, -0.00005837, -0.00007154, -0.00008416, -0.00009606, -0.00010705, -0.00011700, -0.00012575, -0.00013321, -0.00013925, -0.00014382, -0.00014687, -0.00014837, -0.00014833, -0.00014677, -0.00014374, -0.00013933, -0.00013364, -0.00012677, -0.00011887, -0.00011010, -0.00010061, -0.00009059, -0.00008021, -0.00006965, -0.00005910, -0.00004873, -0.00003872, -0.00002922, -0.00002038, -0.00001232, -0.00000518, 0.00000097, 0.00000606, 0.00001003, 0.00001285, 0.00001453, 0.00001507, 0.00001452, 0.00001294, 0.00001039, 0.00000699, 0.00000282, -0.00000199, -0.00000731, -0.00001302, -0.00001897, -0.00002504, -0.00003108, -0.00003698, -0.00004262, -0.00004789, -0.00005269, -0.00005694, -0.00006058, -0.00006356, -0.00006585, -0.00006743, -0.00006831, -0.00006849, },
		{ -0.00001457, -0.00001553, -0.00001671, -0.00001805, -0.00001946, -0.00002087, -0.00002220, -0.00002338, -0.00002433, -0.00002502, -0.00002537, -0.00002536, -0.00002496, -0.00002416, -0.00002295, -0.00002135, -0.00001939, -0.00001711, -0.00001456, -0.00001181, -0.00000893, -0.00000600, -0.00000312, -0.00000037, 0.00000215, 0.00000434, 0.00000612, 0.00000740, 0.00000811, 0.00000816, 0.00000752, 0.00000613, 0.00000397, 0.00000102, -0.00000270, -0.00000718, -0.00001238, -0.00001823, -0.00002467, -0.00003160, -0.00003892, -0.00004652, -0.00005427, -0.00006204, -0.00006968, -0.00007706, -0.00008403, -0.00009046, -0.00009620, -0.00010115, -0.00010518, -0.00010819, -0.00011011, -0.00011086, -0.00011039, -0.00010869, -0.00010574, -0.00010157, -0.00009621, -0.00008973, -0.00008219, -0.00007371, -0.00006441, -0.00005440, -0.00004385, -0.00003291, -0.00002175, -0.00001053, 0.00000057, 0.00001139, 0.00002174, 0.00003148, 0.00004046, 0.00004854, 0.00005559, 0.00006152, 0.00006624, 0.00006970, 0.00007184, 0.00007266, 0.00007215, 0.00007034, 0.00006729, 0.00006305, 0.00005773, 0.00005142, 0.00004425, 0.00003635, 0.00002787, 0.00001895, 0.00000976, 0.00000044, -0.00000884, -0.00001794, -0.00002670, -0.00003501, -0.00004274, -0.00004977, -0.00005603, -0.00006143, -0.00006593, -0.00006948, -0.00007207, -0.00007370, -0.00007439, -0.00007419, -0.00007314, -0.00007131, -0.00006880, -0.00006568, -0.00006206, -0.00005804, -0.00005375, -0.00004928, -0.00004474, -0.00004025, -0.00003589, -0.00003176, -0.00002794, -0.00002449, -0.00002147, -0.00001892, -0.00001686, -0.00001531, -0.00001427, -0.00001371, -0.00001360, -0.00001391, },
		{ -0.00001557, -0.00001466, -0.00001366, -0.00001262, -0.00001162, -0.00001073, -0.00000998, -0.00000945, -0.00000918, -0.00000920, -0.00000954, -0.00001022, -0.00001124, -0.00001259, -0.00001425, -0.00001619, -0.00001837, -0.00002074, -0.00002322, -0.00002576, -0.00002828, -0.00003069, -0.00003292, -0.00003487, -0.00003648, -0.00003767, -0.00003835, -0.00003848, -0.00003800, -0.00003686, -0.00003504, -0.00003252, -0.00002931, -0.00002541, -0.00002087, -0.00001572, -0.00001003, -0.00000387, 0.00000266, 0.00000948, 0.00001646, 0.00002349, 0.00003045, 0.00003721, 0.00004364, 0.00004962, 0.00005502, 0.00005974, 0.00006365, 0.00006668, 0.00006873, 0.00006974, 0.00006967, 0.00006848, 0.00006615, 0.00006271, 0.00005816, 0.00005257, 0.00004599, 0.00003852, 0.00003024, 0.00002127, 0.00001175, 0.00000180, -0.00000841, -0.00001875, -0.00002905, -0.00003915, -0.00004892, -0.00005818, -0.00006682, -0.00007469, -0.00008168, -0.00008770, -0.00009264, -0.00009646, -0.00009909, -0.00010052, -0.00010074, -0.00009976, -0.00009761, -0.00009435, -0.00009004, -0.00008478, -0.00007867, -0.00007182, -0.00006435, -0.00005641, -0.00004813, -0.00003964, -0.00003109, -0.00002263, -0.00001437, -0.00000645, 0.00000101, 0.00000792, 0.00001418, 0.00001971, 0.00002447, 0.00002839, 0.00003146, 0.00003367, 0.00003503, 0.00003556, 0.00003530, 0.00003430, 0.00003263, 0.00003036, 0.00002758, 0.00002438, 0.00002086, 0.00001710, 0.00001322, 0.00000930, 0.00000543, 0.00000169, -0.00000183, -0.00000508, -0.00000799, -0.00001053, -0.00001266, -0.00001437, -0.00001566, -0.00001652, -0.00001698, -0.00001707, -0.00001683, -0.00001631, },
		{ -0.00001277, -0.00001486, -0.00001692, -0.00001889, -0.00002070, -0.00002230, -0.00002364, -0.00002468, -0.00002539, -0.00002573, -0.00002570, -0.00002530, -0.00002453, -0.00002340, -0.00002196, -0.00002023, -0.00001826, -0.00001612, -0.00001386, -0.00001154, -0.00000925, -0.00000705, -0.00000501, -0.00000320, -0.00000170, -0.00000056, 0.00000016, 0.00000043, 0.00000019, -0.00000056, -0.00000185, -0.00000367, -0.00000601, -0.00000884, -0.00001212, -0.00001581, -0.00001984, -0.00002412, -0.00002859, -0.00003315, -0.00003770, -0.00004214, -0.00004637, -0.00005028, -0.00005378, -0.00005677, -0.00005915, -0.00006085, -0.00006180, -0.00006192, -0.00006119, -0.00005957, -0.00005704, -0.00005360, -0.00004926, -0.00004407, -0.00003807, -0.00003133, -0.00002391, -0.00001593, -0.00000747, 0.00000135, 0.00001040, 0.00001956, 0.00002870, 0.00003768, 0.00004638, 0.00005467, 0.00006242, 0.00006954, 0.00007590, 0.00008143, 0.00008604, 0.00008968, 0.00009230, 0.00009386, 0.00009437, 0.00009382, 0.00009224, 0.00008966, 0.00008615, 0.00008178, 0.00007662, 0.00007077, 0.00006434, 0.00005744, 0.00005019, 0.00004270, 0.00003510, 0.00002751, 0.00002005, 0.00001282, 0.00000592, -0.00000054, -0.00000650, -0.00001187, -0.00001660, -0.00002066, -0.00002401, -0.00002665, -0.00002857, -0.00002979, -0.00003034, -0.00003027, -0.00002963, -0.00002849, -0.00002690, -0.00002496, -0.00002274, -0.00002033, -0.00001781, -0.00001527, -0.00001278, -0.00001042, -0.00000825, -0.00000634, -0.00000474, -0.00000348, -0.00000260, -0.00000211, -0.00000201, -0.00000231, -0.00000298, -0.00000400, -0.00000533, -0.00000693, -0.00000874, -0.00001070, },
		{ -0.00000557, -0.00000341, -0.00000139, 0.00000046, 0.00000209, 0.00000345, 0.00000451, 0.00000525, 0.00000565, 0.00000570, 0.00000540, 0.00000478, 0.00000384, 0.00000262, 0.00000116, -0.00000050, -0.00000230, -0.00000419, -0.00000610, -0.00000797, -0.00000973, -0.00001132, -0.00001267, -0.00001373, -0.00001444, -0.00001476, -0.00001465, -0.00001407, -0.00001300, -0.00001144, -0.00000939, -0.00000687, -0.00000388, -0.00000048, 0.00000329, 0.00000738, 0.00001172, 0.00001624, 0.00002085, 0.00002548, 0.00003003, 0.00003442, 0.00003856, 0.00004236, 0.00004573, 0.00004860, 0.00005091, 0.00005257, 0.00005356, 0.00005381, 0.00005332, 0.00005205, 0.00005002, 0.00004723, 0.00004370, 0.00003948, 0.00003462, 0.00002918, 0.00002323, 0.00001686, 0.00001016, 0.00000322, -0.00000384, -0.00001094, -0.00001796, -0.00002479, -0.00003135, -0.00003753, -0.00004324, -0.00004841, -0.00005296, -0.00005684, -0.00006000, -0.00006240, -0.00006403, -0.00006487, -0.00006495, -0.00006426, -0.00006286, -0.00006078, -0.00005809, -0.00005484, -0.00005112, -0.00004700, -0.00004258, -0.00003794, -0.00003318, -0.00002839, -0.00002365, -0.00001906, -0.00001469, -0.00001062, -0.00000691, -0.00000362, -0.00000079, 0.00000154, 0.00000335, 0.00000464, 0.00000539, 0.00000564, 0.00000538, 0.00000468, 0.00000355, 0.00000206, 0.00000027, -0.00000177, -0.00000399, -0.00000631, -0.00000868, -0.00001101, -0.00001325, -0.00001534, -0.00001722, -0.00001884, -0.00002015, -0.00002114, -0.00002176, -0.00002202, -0.00002190, -0.00002141, -0.00002058, -0.00001941, -0.00001794, -0.00001623, -0.00001430, -0.00001221, -0.00001002, -0.00000779, },
		{ -0.00000122, -0.00000240, -0.00000342, -0.00000424, -0.00000484, -0.00000518, -0.00000527, -0.00000509, -0.00000466, -0.00000397, -0.00000306, -0.00000193, -0.00000064, 0.00000080, 0.00000232, 0.00000388, 0.00000543, 0.00000691, 0.00000827, 0.00000945, 0.00001040, 0.00001106, 0.00001139, 0.00001135, 0.00001090, 0.00001001, 0.00000867, 0.00000685, 0.00000457, 0.00000182, -0.00000138, -0.00000499, -0.00000899, -0.00001332, -0.00001793, -0.00002277, -0.00002775, -0.00003282, -0.00003789, -0.00004289, -0.00004773, -0.00005234, -0.00005664, -0.00006055, -0.00006402, -0.00006697, -0.00006935, -0.00007111, -0.00007223, -0.00007267, -0.00007242, -0.00007148, -0.00006986, -0.00006757, -0.00006464, -0.00006112, -0.00005706, -0.00005252, -0.00004757, -0.00004227, -0.00003671, -0.00003098, -0.00002515, -0.00001932, -0.00001357, -0.00000798, -0.00000262, 0.00000242, 0.00000708, 0.00001130, 0.00001504, 0.00001824, 0.00002089, 0.00002297, 0.00002446, 0.00002537, 0.00002571, 0.00002550, 0.00002478, 0.00002359, 0.00002198, 0.00002000, 0.00001771, 0.00001518, 0.00001248, 0.00000967, 0.00000682, 0.00000400, 0.00000127, -0.00000130, -0.00000368, -0.00000580, -0.00000763, -0.00000914, -0.00001030, -0.00001110, -0.00001153, -0.00001161, -0.00001132, -0.00001071, -0.00000979, -0.00000860, -0.00000718, -0.00000557, -0.00000381, -0.00000196, -0.00000007, 0.00000181, 0.00000364, 0.00000536, 0.00000694, 0.00000834, 0.00000952, 0.00001046, 0.00001114, 0.00001155, 0.00001167, 0.00001153, 0.00001112, 0.00001046, 0.00000958, 0.00000850, 0.00000727, 0.00000591, 0.00000447, 0.00000299, 0.00000153, 0.00000011, },
		{ -0.00000363, -0.00000355, -0.00000363, -0.00000389, -0.00000433, -0.00000496, -0.00000578, -0.00000677, -0.00000794, -0.00000924, -0.00001068, -0.00001220, -0.00001379, -0.00001540, -0.00001699, -0.00001853, -0.00001996, -0.00002125, -0.00002235, -0.00002323, -0.00002384, -0.00002415, -0.00002413, -0.00002376, -0.00002302, -0.00002190, -0.00002039, -0.00001849, -0.00001622, -0.00001359, -0.00001064, -0.00000739, -0.00000388, -0.00000016, 0.00000372, 0.00000771, 0.00001174, 0.00001575, 0.00001969, 0.00002349, 0.00002708, 0.00003041, 0.00003342, 0.00003606, 0.00003828, 0.00004004, 0.00004130, 0.00004205, 0.00004226, 0.00004192, 0.00004104, 0.00003962, 0.00003769, 0.00003526, 0.00003238, 0.00002909, 0.00002543, 0.00002146, 0.00001724, 0.00001284, 0.00000832, 0.00000375, -0.00000081, -0.00000528, -0.00000961, -0.00001372, -0.00001757, -0.00002110, -0.00002426, -0.00002702, -0.00002935, -0.00003122, -0.00003262, -0.00003355, -0.00003400, -0.00003399, -0.00003353, -0.00003266, -0.00003140, -0.00002979, -0.00002788, -0.00002572, -0.00002335, -0.00002083, -0.00001821, -0.00001555, -0.00001289, -0.00001030, -0.00000780, -0.00000546, -0.00000330, -0.00000136, 0.00000034, 0.00000178, 0.00000294, 0.00000381, 0.00000440, 0.00000471, 0.00000476, 0.00000456, 0.00000413, 0.00000351, 0.00000271, 0.00000179, 0.00000075, -0.00000034, -0.00000147, -0.00000259, -0.00000368, -0.00000471, -0.00000564, -0.00000645, -0.00000714, -0.00000767, -0.00000806, -0.00000828, -0.00000836, -0.00000828, -0.00000807, -0.00000775, -0.00000732, -0.00000682, -0.00000628, -0.00000571, -0.00000516, -0.00000465, -0.00000420, -0.00000386, },
		{ -0.00000728, -0.00000731, -0.00000718, -0.00000689, -0.00000643, -0.00000581, -0.00000504, -0.00000412, -0.00000308, -0.00000192, -0.00000066, 0.00000065, 0.00000201, 0.00000337, 0.00000472, 0.00000601, 0.00000722, 0.00000831, 0.00000925, 0.00001002, 0.00001059, 0.00001093, 0.00001104, 0.00001088, 0.00001047, 0.00000978, 0.00000883, 0.00000762, 0.00000615, 0.00000445, 0.00000254, 0.00000045, -0.00000180, -0.00000417, -0.00000661, -0.00000910, -0.00001158, -0.00001402, -0.00001636, -0.00001857, -0.00002061, -0.00002243, -0.00002399, -0.00002528, -0.00002624, -0.00002688, -0.00002716, -0.00002707, -0.00002660, -0.00002577, -0.00002456, -0.00002301, -0.00002111, -0.00001891, -0.00001642, -0.00001369, -0.00001074, -0.00000763, -0.00000439, -0.00000108, 0.00000227, 0.00000559, 0.00000885, 0.00001200, 0.00001500, 0.00001781, 0.00002038, 0.00002269, 0.00002472, 0.00002643, 0.00002782, 0.00002887, 0.00002957, 0.00002994, 0.00002997, 0.00002968, 0.00002908, 0.00002821, 0.00002707, 0.00002571, 0.00002416, 0.00002245, 0.00002061, 0.00001869, 0.00001672, 0.00001473, 0.00001277, 0.00001086, 0.00000903, 0.00000731, 0.00000572, 0.00000427, 0.00000299, 0.00000189, 0.00000095, 0.00000020, -0.00000038, -0.00000079, -0.00000105, -0.00000117, -0.00000117, -0.00000107, -0.00000088, -0.00000063, -0.00000035, -0.00000006, 0.00000023, 0.00000048, 0.00000069, 0.00000083, 0.00000090, 0.00000088, 0.00000076, 0.00000055, 0.00000024, -0.00000016, -0.00000065, -0.00000122, -0.00000185, -0.00000252, -0.00000322, -0.00000393, -0.00000462, -0.00000528, -0.00000587, -0.00000639, -0.00000681, -0.00000711, },
		{ 0.00001124, 0.00001080, 0.00001021, 0.00000947, 0.00000858, 0.00000756, 0.00000641, 0.00000515, 0.00000380, 0.00000237, 0.00000088, -0.00000063, -0.00000216, -0.00000366, -0.00000512, -0.00000651, -0.00000780, -0.00000897, -0.00001000, -0.00001086, -0.00001154, -0.00001201, -0.00001227, -0.00001231, -0.00001211, -0.00001168, -0.00001102, -0.00001012, -0.00000902, -0.00000770, -0.00000621, -0.00000454, -0.00000273, -0.00000081, 0.00000120, 0.00000327, 0.00000536, 0.00000745, 0.00000949, 0.00001146, 0.00001333, 0.00001505, 0.00001662, 0.00001799, 0.00001915, 0.00002008, 0.00002076, 0.00002117, 0.00002132, 0.00002120, 0.00002081, 0.00002014, 0.00001922, 0.00001806, 0.00001666, 0.00001506, 0.00001327, 0.00001131, 0.00000923, 0.00000704, 0.00000478, 0.00000248, 0.00000016, -0.00000214, -0.00000439, -0.00000657, -0.00000865, -0.00001060, -0.00001241, -0.00001406, -0.00001553, -0.00001681, -0.00001790, -0.00001878, -0.00001946, -0.00001995, -0.00002024, -0.00002034, -0.00002027, -0.00002003, -0.00001965, -0.00001913, -0.00001850, -0.00001777, -0.00001696, -0.00001609, -0.00001518, -0.00001423, -0.00001328, -0.00001232, -0.00001137, -0.00001045, -0.00000956, -0.00000870, -0.00000789, -0.00000711, -0.00000638, -0.00000569, -0.00000503, -0.00000440, -0.00000380, -0.00000321, -0.00000264, -0.00000206, -0.00000149, -0.00000090, -0.00000030, 0.00000033, 0.00000098, 0.00000166, 0.00000237, 0.00000311, 0.00000387, 0.00000465, 0.00000544, 0.00000623, 0.00000702, 0.00000779, 0.00000853, 0.00000922, 0.00000986, 0.00001042, 0.00001090, 0.00001127, 0.00001153, 0.00001167, 0.00001167, 0.00001153, },
		{ -0.00000356, -0.00000213, -0.00000064, 0.00000090, 0.00000249, 0.00000410, 0.00000573, 0.00000735, 0.00000896, 0.00001052, 0.00001203, 0.00001346, 0.00001480, 0.00001603, 0.00001713, 0.00001808, 0.00001887, 0.00001948, 0.00001989, 0.00002010, 0.00002010, 0.00001988, 0.00001943, 0.00001876, 0.00001785, 0.00001673, 0.00001538, 0.00001383, 0.00001208, 0.00001016, 0.00000807, 0.00000583, 0.00000348, 0.00000102, -0.00000150, -0.00000407, -0.00000665, -0.00000922, -0.00001175, -0.00001420, -0.00001656, -0.00001879, -0.00002087, -0.00002278, -0.00002448, -0.00002597, -0.00002722, -0.00002823, -0.00002898, -0.00002946, -0.00002966, -0.00002960, -0.00002927, -0.00002867, -0.00002782, -0.00002673, -0.00002540, -0.00002387, -0.00002215, -0.00002027, -0.00001824, -0.00001609, -0.00001385, -0.00001154, -0.00000920, -0.00000684, -0.00000450, -0.00000220, 0.00000003, 0.00000217, 0.00000421, 0.00000612, 0.00000788, 0.00000948, 0.00001091, 0.00001215, 0.00001321, 0.00001408, 0.00001474, 0.00001521, 0.00001549, 0.00001558, 0.00001549, 0.00001523, 0.00001480, 0.00001422, 0.00001351, 0.00001266, 0.00001171, 0.00001066, 0.00000952, 0.00000832, 0.00000706, 0.00000577, 0.00000445, 0.00000311, 0.00000177, 0.00000044, -0.00000086, -0.00000214, -0.00000337, -0.00000456, -0.00000570, -0.00000677, -0.00000778, -0.00000872, -0.00000957, -0.00001035, -0.00001104, -0.00001165, -0.00001216, -0.00001257, -0.00001289, -0.00001311, -0.00001323, -0.00001324, -0.00001315, -0.00001294, -0.00001263, -0.00001220, -0.00001167, -0.00001102, -0.00001026, -0.00000939, -0.00000842, -0.00000734, -0.00000617, -0.00000491, },
		{ -0.00000253, -0.00000360, -0.00000462, -0.00000558, -0.00000647, -0.00000728, -0.00000802, -0.00000866, -0.00000922, -0.00000967, -0.00001002, -0.00001027, -0.00001040, -0.00001041, -0.00001031, -0.00001009, -0.00000975, -0.00000928, -0.00000869, -0.00000798, -0.00000714, -0.00000619, -0.00000513, -0.00000395, -0.00000266, -0.00000128, 0.00000019, 0.00000175, 0.00000337, 0.00000506, 0.00000679, 0.00000856, 0.00001034, 0.00001213, 0.00001390, 0.00001564, 0.00001733, 0.00001895, 0.00002049, 0.00002192, 0.00002323, 0.00002440, 0.00002542, 0.00002626, 0.00002692, 0.00002738, 0.00002763, 0.00002766, 0.00002748, 0.00002706, 0.00002641, 0.00002553, 0.00002443, 0.00002311, 0.00002157, 0.00001984, 0.00001792, 0.00001583, 0.00001359, 0.00001121, 0.00000873, 0.00000616, 0.00000354, 0.00000087, -0.00000180, -0.00000445, -0.00000705, -0.00000958, -0.00001201, -0.00001432, -0.00001648, -0.00001847, -0.00002028, -0.00002188, -0.00002325, -0.00002440, -0.00002530, -0.00002595, -0.00002636, -0.00002650, -0.00002640, -0.00002605, -0.00002546, -0.00002465, -0.00002363, -0.00002240, -0.00002100, -0.00001943, -0.00001773, -0.00001590, -0.00001398, -0.00001199, -0.00000996, -0.00000789, -0.00000583, -0.00000379, -0.00000179, 0.00000014, 0.00000199, 0.00000373, 0.00000537, 0.00000687, 0.00000823, 0.00000943, 0.00001048, 0.00001136, 0.00001208, 0.00001262, 0.00001300, 0.00001321, 0.00001325, 0.00001315, 0.00001289, 0.00001249, 0.00001196, 0.00001131, 0.00001056, 0.00000970, 0.00000876, 0.00000775, 0.00000669, 0.00000557, 0.00000442, 0.00000325, 0.00000207, 0.00000089, -0.00000028, -0.00000142, },
		{ -0.00000258, -0.00000276, -0.00000302, -0.00000334, -0.00000372, -0.00000415, -0.00000463, -0.00000515, -0.00000569, -0.00000626, -0.00000683, -0.00000740, -0.00000797, -0.00000852, -0.00000905, -0.00000955, -0.00001002, -0.00001046, -0.00001085, -0.00001121, -0.00001153, -0.00001181, -0.00001205, -0.00001225, -0.00001242, -0.00001256, -0.00001268, -0.00001277, -0.00001284, -0.00001289, -0.00001292, -0.00001295, -0.00001296, -0.00001296, -0.00001295, -0.00001292, -0.00001288, -0.00001282, -0.00001273, -0.00001262, -0.00001246, -0.00001226, -0.00001201, -0.00001169, -0.00001131, -0.00001084, -0.00001029, -0.00000965, -0.00000890, -0.00000804, -0.00000707, -0.00000598, -0.00000478, -0.00000345, -0.00000201, -0.00000046, 0.00000120, 0.00000296, 0.00000481, 0.00000674, 0.00000873, 0.00001076, 0.00001283, 0.00001490, 0.00001697, 0.00001900, 0.00002099, 0.00002290, 0.00002471, 0.00002641, 0.00002797, 0.00002938, 0.00003061, 0.00003166, 0.00003250, 0.00003313, 0.00003354, 0.00003372, 0.00003367, 0.00003339, 0.00003288, 0.00003215, 0.00003120, 0.00003006, 0.00002872, 0.00002721, 0.00002555, 0.00002375, 0.00002184, 0.00001985, 0.00001779, 0.00001569, 0.00001357, 0.00001147, 0.00000939, 0.00000737, 0.00000543, 0.00000358, 0.00000184, 0.00000022, -0.00000125, -0.00000258, -0.00000375, -0.00000475, -0.00000560, -0.00000628, -0.00000681, -0.00000718, -0.00000741, -0.00000751, -0.00000748, -0.00000734, -0.00000711, -0.00000680, -0.00000642, -0.00000600, -0.00000555, -0.00000509, -0.00000463, -0.00000418, -0.00000377, -0.00000339, -0.00000307, -0.00000280, -0.00000261, -0.00000249, -0.00000244, -0.00000247, },
		{ -0.00000110, -0.00000058, -0.00000004, 0.00000052, 0.00000107, 0.00000162, 0.00000213, 0.00000260, 0.00000302, 0.00000337, 0.00000365, 0.00000385, 0.00000396, 0.00000397, 0.00000389, 0.00000372, 0.00000345, 0.00000309, 0.00000266, 0.00000215, 0.00000158, 0.00000096, 0.00000030, -0.00000037, -0.00000105, -0.00000171, -0.00000235, -0.00000294, -0.00000347, -0.00000393, -0.00000429, -0.00000456, -0.00000471, -0.00000475, -0.00000467, -0.00000446, -0.00000413, -0.00000367, -0.00000310, -0.00000242, -0.00000165, -0.00000079, 0.00000013, 0.00000110, 0.00000211, 0.00000312, 0.00000413, 0.00000510, 0.00000603, 0.00000688, 0.00000764, 0.00000829, 0.00000881, 0.00000919, 0.00000942, 0.00000948, 0.00000937, 0.00000909, 0.00000863, 0.00000800, 0.00000720, 0.00000624, 0.00000514, 0.00000390, 0.00000254, 0.00000108, -0.00000045, -0.00000204, -0.00000366, -0.00000529, -0.00000691, -0.00000848, -0.00001000, -0.00001143, -0.00001277, -0.00001398, -0.00001505, -0.00001597, -0.00001673, -0.00001732, -0.00001774, -0.00001797, -0.00001804, -0.00001793, -0.00001765, -0.00001723, -0.00001665, -0.00001596, -0.00001514, -0.00001424, -0.00001326, -0.00001222, -0.00001114, -0.00001005, -0.00000897, -0.00000790, -0.00000687, -0.00000590, -0.00000500, -0.00000418, -0.00000344, -0.00000281, -0.00000227, -0.00000184, -0.00000152, -0.00000129, -0.00000116, -0.00000113, -0.00000117, -0.00000127, -0.00000144, -0.00000165, -0.00000189, -0.00000215, -0.00000241, -0.00000266, -0.00000289, -0.00000307, -0.00000322, -0.00000331, -0.00000333, -0.00000329, -0.00000318, -0.00000299, -0.00000274, -0.00000241, -0.00000203, -0.00000159, },
		{ 0.00000071, 0.00000052, 0.00000033, 0.00000016, 0.00000002, -0.00000009, -0.00000016, -0.00000018, -0.00000013, -0.00000003, 0.00000014, 0.00000038, 0.00000068, 0.00000104, 0.00000145, 0.00000191, 0.00000241, 0.00000293, 0.00000346, 0.00000398, 0.00000448, 0.00000493, 0.00000532, 0.00000562, 0.00000582, 0.00000590, 0.00000584, 0.00000562, 0.00000523, 0.00000465, 0.00000389, 0.00000292, 0.00000175, 0.00000038, -0.00000119, -0.00000295, -0.00000489, -0.00000699, -0.00000924, -0.00001161, -0.00001409, -0.00001663, -0.00001922, -0.00002181, -0.00002438, -0.00002690, -0.00002932, -0.00003162, -0.00003376, -0.00003572, -0.00003745, -0.00003894, -0.00004016, -0.00004110, -0.00004173, -0.00004204, -0.00004203, -0.00004169, -0.00004102, -0.00004004, -0.00003874, -0.00003715, -0.00003529, -0.00003317, -0.00003083, -0.00002829, -0.00002558, -0.00002275, -0.00001982, -0.00001683, -0.00001381, -0.00001081, -0.00000786, -0.00000499, -0.00000223, 0.00000038, 0.00000283, 0.00000508, 0.00000713, 0.00000895, 0.00001053, 0.00001187, 0.00001297, 0.00001381, 0.00001441, 0.00001478, 0.00001493, 0.00001487, 0.00001462, 0.00001419, 0.00001362, 0.00001292, 0.00001212, 0.00001124, 0.00001030, 0.00000933, 0.00000834, 0.00000737, 0.00000642, 0.00000551, 0.00000467, 0.00000389, 0.00000319, 0.00000258, 0.00000206, 0.00000163, 0.00000129, 0.00000103, 0.00000086, 0.00000076, 0.00000072, 0.00000073, 0.00000079, 0.00000089, 0.00000100, 0.00000113, 0.00000125, 0.00000136, 0.00000146, 0.00000153, 0.00000157, 0.00000157, 0.00000154, 0.00000147, 0.00000137, 0.00000124, 0.00000108, 0.00000090, },
		{ 0.00000073, 0.00000026, -0.00000022, -0.00000071, -0.00000121, -0.00000173, -0.00000227, -0.00000284, -0.00000343, -0.00000405, -0.00000470, -0.00000537, -0.00000607, -0.00000678, -0.00000749, -0.00000820, -0.00000889, -0.00000956, -0.00001018, -0.00001074, -0.00001122, -0.00001161, -0.00001189, -0.00001205, -0.00001206, -0.00001192, -0.00001161, -0.00001112, -0.00001044, -0.00000958, -0.00000852, -0.00000727, -0.00000583, -0.00000422, -0.00000244, -0.00000051, 0.00000155, 0.00000372, 0.00000597, 0.00000828, 0.00001061, 0.00001293, 0.00001520, 0.00001740, 0.00001949, 0.00002144, 0.00002320, 0.00002475, 0.00002606, 0.00002711, 0.00002786, 0.00002830, 0.00002842, 0.00002820, 0.00002763, 0.00002672, 0.00002547, 0.00002389, 0.00002198, 0.00001978, 0.00001730, 0.00001457, 0.00001163, 0.00000850, 0.00000523, 0.00000185, -0.00000159, -0.00000505, -0.00000848, -0.00001185, -0.00001512, -0.00001823, -0.00002116, -0.00002387, -0.00002633, -0.00002851, -0.00003039, -0.00003195, -0.00003318, -0.00003406, -0.00003460, -0.00003479, -0.00003465, -0.00003418, -0.00003339, -0.00003232, -0.00003098, -0.00002940, -0.00002761, -0.00002563, -0.00002351, -0.00002127, -0.00001895, -0.00001658, -0.00001421, -0.00001184, -0.00000953, -0.00000729, -0.00000515, -0.00000314, -0.00000126, 0.00000046, 0.00000201, 0.00000339, 0.00000458, 0.00000559, 0.00000642, 0.00000708, 0.00000756, 0.00000789, 0.00000806, 0.00000811, 0.00000803, 0.00000784, 0.00000756, 0.00000721, 0.00000680, 0.00000634, 0.00000584, 0.00000533, 0.00000480, 0.00000426, 0.00000373, 0.00000321, 0.00000269, 0.00000219, 0.00000170, 0.00000121, },
		{ 0.00000036, 0.00000106, 0.00000172, 0.00000234, 0.00000293, 0.00000350, 0.00000406, 0.00000460, 0.00000514, 0.00000568, 0.00000622, 0.00000676, 0.00000730, 0.00000784, 0.00000836, 0.00000887, 0.00000935, 0.00000979, 0.00001019, 0.00001053, 0.00001080, 0.00001099, 0.00001108, 0.00001106, 0.00001092, 0.00001065, 0.00001024, 0.00000968, 0.00000898, 0.00000813, 0.00000712, 0.00000597, 0.00000469, 0.00000327, 0.00000175, 0.00000012, -0.00000158, -0.00000334, -0.00000513, -0.00000693, -0.00000870, -0.00001043, -0.00001208, -0.00001361, -0.00001502, -0.00001625, -0.00001730, -0.00001813, -0.00001873, -0.00001908, -0.00001916, -0.00001896, -0.00001848, -0.00001771, -0.00001666, -0.00001533, -0.00001374, -0.00001191, -0.00000985, -0.00000758, -0.00000514, -0.00000257, 0.00000012, 0.00000287, 0.00000566, 0.00000844, 0.00001118, 0.00001382, 0.00001634, 0.00001870, 0.00002087, 0.00002280, 0.00002448, 0.00002588, 0.00002698, 0.00002776, 0.00002821, 0.00002832, 0.00002810, 0.00002755, 0.00002667, 0.00002549, 0.00002401, 0.00002227, 0.00002028, 0.00001808, 0.00001570, 0.00001318, 0.00001054, 0.00000783, 0.00000508, 0.00000233, -0.00000038, -0.00000303, -0.00000557, -0.00000797, -0.00001022, -0.00001229, -0.00001415, -0.00001579, -0.00001720, -0.00001837, -0.00001930, -0.00001999, -0.00002043, -0.00002064, -0.00002063, -0.00002041, -0.00002000, -0.00001941, -0.00001867, -0.00001779, -0.00001680, -0.00001571, -0.00001456, -0.00001335, -0.00001212, -0.00001087, -0.00000963, -0.00000841, -0.00000722, -0.00000607, -0.00000498, -0.00000394, -0.00000296, -0.00000204, -0.00000118, -0.00000038, },
		{ -0.00000845, -0.00000839, -0.00000828, -0.00000813, -0.00000796, -0.00000777, -0.00000758, -0.00000740, -0.00000723, -0.00000708, -0.00000695, -0.00000685, -0.00000677, -0.00000671, -0.00000667, -0.00000665, -0.00000664, -0.00000662, -0.00000658, -0.00000653, -0.00000643, -0.00000629, -0.00000609, -0.00000581, -0.00000545, -0.00000500, -0.00000445, -0.00000378, -0.00000301, -0.00000212, -0.00000112, -0.00000002, 0.00000120, 0.00000250, 0.00000389, 0.00000534, 0.00000684, 0.00000838, 0.00000992, 0.00001145, 0.00001294, 0.00001437, 0.00001572, 0.00001695, 0.00001805, 0.00001899, 0.00001976, 0.00002032, 0.00002067, 0.00002080, 0.00002069, 0.00002034, 0.00001974, 0.00001890, 0.00001782, 0.00001651, 0.00001499, 0.00001327, 0.00001137, 0.00000931, 0.00000713, 0.00000485, 0.00000250, 0.00000012, -0.00000226, -0.00000461, -0.00000690, -0.00000909, -0.00001114, -0.00001305, -0.00001476, -0.00001627, -0.00001755, -0.00001858, -0.00001935, -0.00001986, -0.00002010, -0.00002007, -0.00001977, -0.00001922, -0.00001842, -0.00001740, -0.00001617, -0.00001475, -0.00001319, -0.00001149, -0.00000969, -0.00000783, -0.00000593, -0.00000403, -0.00000215, -0.00000033, 0.00000141, 0.00000303, 0.00000453, 0.00000587, 0.00000703, 0.00000802, 0.00000881, 0.00000940, 0.00000978, 0.00000996, 0.00000995, 0.00000974, 0.00000935, 0.00000880, 0.00000811, 0.00000728, 0.00000634, 0.00000531, 0.00000421, 0.00000307, 0.00000190, 0.00000073, -0.00000042, -0.00000154, -0.00000260, -0.00000360, -0.00000452, -0.00000535, -0.00000609, -0.00000672, -0.00000725, -0.00000768, -0.00000801, -0.00000824, -0.00000839, -0.00000845, },
		{ 0.00000287, 0.00000269, 0.00000250, 0.00000231, 0.00000213, 0.00000197, 0.00000184, 0.00000175, 0.00000170, 0.00000170, 0.00000173, 0.00000182, 0.00000194, 0.00000210, 0.00000229, 0.00000250, 0.00000273, 0.00000295, 0.00000316, 0.00000335, 0.00000349, 0.00000359, 0.00000362, 0.00000358, 0.00000344, 0.00000321, 0.00000288, 0.00000243, 0.00000187, 0.00000119, 0.00000040, -0.00000051, -0.00000152, -0.00000262, -0.00000381, -0.00000506, -0.00000637, -0.00000771, -0.00000907, -0.00001041, -0.00001173, -0.00001299, -0.00001418, -0.00001527, -0.00001623, -0.00001706, -0.00001772, -0.00001821, -0.00001851, -0.00001861, -0.00001849, -0.00001817, -0.00001762, -0.00001687, -0.00001590, -0.00001474, -0.00001339, -0.00001187, -0.00001020, -0.00000840, -0.00000649, -0.00000451, -0.00000248, -0.00000043, 0.00000162, 0.00000363, 0.00000558, 0.00000744, 0.00000917, 0.00001077, 0.00001219, 0.00001344, 0.00001448, 0.00001532, 0.00001593, 0.00001631, 0.00001647, 0.00001640, 0.00001610, 0.00001560, 0.00001490, 0.00001401, 0.00001296, 0.00001177, 0.00001046, 0.00000905, 0.00000758, 0.00000605, 0.00000451, 0.00000298, 0.00000148, 0.00000004, -0.00000132, -0.00000259, -0.00000374, -0.00000475, -0.00000563, -0.00000636, -0.00000692, -0.00000733, -0.00000758, -0.00000768, -0.00000763, -0.00000743, -0.00000711, -0.00000667, -0.00000614, -0.00000552, -0.00000483, -0.00000409, -0.00000332, -0.00000254, -0.00000177, -0.00000101, -0.00000028, 0.00000039, 0.00000102, 0.00000157, 0.00000205, 0.00000245, 0.00000277, 0.00000302, 0.00000318, 0.00000327, 0.00000329, 0.00000325, 0.00000316, 0.00000303, },
		{ 0.00000144, 0.00000151, 0.00000158, 0.00000164, 0.00000168, 0.00000169, 0.00000168, 0.00000163, 0.00000154, 0.00000142, 0.00000126, 0.00000106, 0.00000083, 0.00000058, 0.00000030, 0.00000002, -0.00000027, -0.00000054, -0.00000080, -0.00000103, -0.00000121, -0.00000134, -0.00000140, -0.00000140, -0.00000131, -0.00000112, -0.00000085, -0.00000047, 0.00000000, 0.00000058, 0.00000125, 0.00000201, 0.00000286, 0.00000378, 0.00000477, 0.00000580, 0.00000687, 0.00000795, 0.00000903, 0.00001009, 0.00001111, 0.00001207, 0.00001294, 0.00001372, 0.00001438, 0.00001490, 0.00001528, 0.00001549, 0.00001554, 0.00001540, 0.00001508, 0.00001457, 0.00001387, 0.00001300, 0.00001195, 0.00001073, 0.00000937, 0.00000788, 0.00000627, 0.00000456, 0.00000278, 0.00000096, -0.00000089, -0.00000274, -0.00000455, -0.00000632, -0.00000800, -0.00000958, -0.00001104, -0.00001235, -0.00001349, -0.00001446, -0.00001524, -0.00001582, -0.00001619, -0.00001635, -0.00001630, -0.00001605, -0.00001560, -0.00001496, -0.00001415, -0.00001317, -0.00001205, -0.00001081, -0.00000947, -0.00000805, -0.00000657, -0.00000506, -0.00000354, -0.00000203, -0.00000056, 0.00000086, 0.00000220, 0.00000346, 0.00000460, 0.00000563, 0.00000654, 0.00000730, 0.00000793, 0.00000842, 0.00000877, 0.00000898, 0.00000906, 0.00000902, 0.00000887, 0.00000861, 0.00000827, 0.00000785, 0.00000737, 0.00000685, 0.00000630, 0.00000573, 0.00000516, 0.00000460, 0.00000406, 0.00000355, 0.00000309, 0.00000267, 0.00000231, 0.00000200, 0.00000175, 0.00000156, 0.00000142, 0.00000133, 0.00000129, 0.00000129, 0.00000132, 0.00000137, },
		{ -0.00000134, -0.00000167, -0.00000199, -0.00000230, -0.00000258, -0.00000283, -0.00000305, -0.00000322, -0.00000334, -0.00000342, -0.00000345, -0.00000343, -0.00000336, -0.00000325, -0.00000311, -0.00000293, -0.00000273, -0.00000252, -0.00000229, -0.00000207, -0.00000186, -0.00000167, -0.00000152, -0.00000140, -0.00000132, -0.00000130, -0.00000134, -0.00000143, -0.00000159, -0.00000182, -0.00000210, -0.00000245, -0.00000285, -0.00000330, -0.00000379, -0.00000431, -0.00000484, -0.00000538, -0.00000592, -0.00000643, -0.00000690, -0.00000733, -0.00000769, -0.00000796, -0.00000815, -0.00000823, -0.00000820, -0.00000804, -0.00000775, -0.00000733, -0.00000677, -0.00000606, -0.00000523, -0.00000425, -0.00000316, -0.00000195, -0.00000063, 0.00000077, 0.00000225, 0.00000378, 0.00000536, 0.00000695, 0.00000854, 0.00001011, 0.00001163, 0.00001309, 0.00001447, 0.00001575, 0.00001690, 0.00001793, 0.00001880, 0.00001952, 0.00002007, 0.00002044, 0.00002064, 0.00002066, 0.00002050, 0.00002017, 0.00001968, 0.00001903, 0.00001824, 0.00001731, 0.00001628, 0.00001514, 0.00001393, 0.00001266, 0.00001134, 0.00001001, 0.00000867, 0.00000735, 0.00000607, 0.00000483, 0.00000366, 0.00000257, 0.00000158, 0.00000068, -0.00000011, -0.00000079, -0.00000136, -0.00000181, -0.00000215, -0.00000238, -0.00000251, -0.00000254, -0.00000249, -0.00000237, -0.00000218, -0.00000194, -0.00000166, -0.00000136, -0.00000104, -0.00000073, -0.00000042, -0.00000014, 0.00000011, 0.00000032, 0.00000049, 0.00000060, 0.00000066, 0.00000067, 0.00000062, 0.00000051, 0.00000035, 0.00000015, -0.00000010, -0.00000038, -0.00000068, -0.00000100, },
		{ -0.00000228, -0.00000180, -0.00000133, -0.00000088, -0.00000046, -0.00000008, 0.00000026, 0.00000056, 0.00000081, 0.00000100, 0.00000114, 0.00000123, 0.00000127, 0.00000126, 0.00000120, 0.00000111, 0.00000099, 0.00000085, 0.00000069, 0.00000053, 0.00000037, 0.00000023, 0.00000011, 0.00000003, -0.00000002, -0.00000001, 0.00000005, 0.00000017, 0.00000036, 0.00000061, 0.00000093, 0.00000132, 0.00000177, 0.00000228, 0.00000285, 0.00000346, 0.00000412, 0.00000480, 0.00000549, 0.00000619, 0.00000688, 0.00000755, 0.00000818, 0.00000877, 0.00000929, 0.00000974, 0.00001010, 0.00001037, 0.00001054, 0.00001059, 0.00001053, 0.00001034, 0.00001004, 0.00000961, 0.00000906, 0.00000840, 0.00000764, 0.00000677, 0.00000581, 0.00000478, 0.00000368, 0.00000253, 0.00000134, 0.00000014, -0.00000107, -0.00000227, -0.00000343, -0.00000456, -0.00000563, -0.00000662, -0.00000754, -0.00000836, -0.00000908, -0.00000969, -0.00001018, -0.00001056, -0.00001082, -0.00001097, -0.00001100, -0.00001092, -0.00001074, -0.00001047, -0.00001012, -0.00000969, -0.00000920, -0.00000867, -0.00000810, -0.00000751, -0.00000691, -0.00000632, -0.00000575, -0.00000520, -0.00000470, -0.00000424, -0.00000384, -0.00000350, -0.00000323, -0.00000302, -0.00000289, -0.00000283, -0.00000283, -0.00000290, -0.00000302, -0.00000320, -0.00000342, -0.00000367, -0.00000395, -0.00000424, -0.00000454, -0.00000483, -0.00000511, -0.00000537, -0.00000559, -0.00000577, -0.00000591, -0.00000599, -0.00000601, -0.00000598, -0.00000588, -0.00000573, -0.00000551, -0.00000524, -0.00000492, -0.00000455, -0.00000414, -0.00000371, -0.00000324, -0.00000276, },
		{ -0.00000140, -0.00000175, -0.00000208, -0.00000238, -0.00000265, -0.00000289, -0.00000308, -0.00000323, -0.00000334, -0.00000339, -0.00000340, -0.00000336, -0.00000328, -0.00000316, -0.00000301, -0.00000284, -0.00000264, -0.00000243, -0.00000222, -0.00000203, -0.00000185, -0.00000169, -0.00000158, -0.00000151, -0.00000150, -0.00000156, -0.00000168, -0.00000189, -0.00000217, -0.00000254, -0.00000299, -0.00000353, -0.00000415, -0.00000485, -0.00000563, -0.00000646, -0.00000736, -0.00000829, -0.00000927, -0.00001026, -0.00001125, -0.00001224, -0.00001321, -0.00001414, -0.00001502, -0.00001582, -0.00001655, -0.00001718, -0.00001771, -0.00001812, -0.00001841, -0.00001856, -0.00001858, -0.00001846, -0.00001820, -0.00001780, -0.00001726, -0.00001659, -0.00001580, -0.00001490, -0.00001389, -0.00001280, -0.00001162, -0.00001039, -0.00000911, -0.00000779, -0.00000646, -0.00000514, -0.00000383, -0.00000255, -0.00000132, -0.00000015, 0.00000095, 0.00000196, 0.00000288, 0.00000371, 0.00000443, 0.00000504, 0.00000555, 0.00000594, 0.00000623, 0.00000641, 0.00000650, 0.00000650, 0.00000641, 0.00000625, 0.00000603, 0.00000575, 0.00000543, 0.00000508, 0.00000471, 0.00000433, 0.00000395, 0.00000357, 0.00000322, 0.00000289, 0.00000258, 0.00000232, 0.00000209, 0.00000191, 0.00000176, 0.00000166, 0.00000159, 0.00000156, 0.00000157, 0.00000160, 0.00000165, 0.00000172, 0.00000179, 0.00000187, 0.00000194, 0.00000200, 0.00000204, 0.00000206, 0.00000205, 0.00000201, 0.00000194, 0.00000182, 0.00000167, 0.00000148, 0.00000125, 0.00000099, 0.00000070, 0.00000038, 0.00000004, -0.00000031, -0.00000068, -0.00000104, },
		{ -0.00000100, -0.00000108, -0.00000117, -0.00000129, -0.00000142, -0.00000159, -0.00000179, -0.00000202, -0.00000228, -0.00000258, -0.00000292, -0.00000329, -0.00000368, -0.00000411, -0.00000455, -0.00000500, -0.00000546, -0.00000592, -0.00000636, -0.00000678, -0.00000718, -0.00000753, -0.00000783, -0.00000806, -0.00000823, -0.00000833, -0.00000833, -0.00000825, -0.00000807, -0.00000780, -0.00000742, -0.00000695, -0.00000638, -0.00000572, -0.00000497, -0.00000414, -0.00000324, -0.00000228, -0.00000127, -0.00000023, 0.00000084, 0.00000191, 0.00000296, 0.00000399, 0.00000498, 0.00000592, 0.00000678, 0.00000755, 0.00000823, 0.00000880, 0.00000924, 0.00000956, 0.00000974, 0.00000979, 0.00000969, 0.00000946, 0.00000909, 0.00000859, 0.00000796, 0.00000721, 0.00000636, 0.00000541, 0.00000438, 0.00000329, 0.00000215, 0.00000097, -0.00000022, -0.00000141, -0.00000258, -0.00000371, -0.00000479, -0.00000581, -0.00000675, -0.00000760, -0.00000835, -0.00000899, -0.00000951, -0.00000991, -0.00001019, -0.00001035, -0.00001038, -0.00001030, -0.00001010, -0.00000980, -0.00000940, -0.00000892, -0.00000836, -0.00000773, -0.00000706, -0.00000634, -0.00000560, -0.00000485, -0.00000409, -0.00000335, -0.00000264, -0.00000195, -0.00000131, -0.00000072, -0.00000019, 0.00000028, 0.00000069, 0.00000103, 0.00000130, 0.00000151, 0.00000165, 0.00000174, 0.00000176, 0.00000174, 0.00000167, 0.00000156, 0.00000142, 0.00000125, 0.00000107, 0.00000088, 0.00000068, 0.00000048, 0.00000029, 0.00000011, -0.00000006, -0.00000021, -0.00000035, -0.00000046, -0.00000057, -0.00000066, -0.00000073, -0.00000080, -0.00000087, -0.00000093, },
		{ -0.00000141, -0.00000152, -0.00000162, -0.00000170, -0.00000175, -0.00000177, -0.00000175, -0.00000170, -0.00000160, -0.00000145, -0.00000127, -0.00000104, -0.00000077, -0.00000046, -0.00000012, 0.00000025, 0.00000063, 0.00000103, 0.00000143, 0.00000181, 0.00000218, 0.00000252, 0.00000282, 0.00000307, 0.00000326, 0.00000338, 0.00000343, 0.00000339, 0.00000326, 0.00000303, 0.00000271, 0.00000230, 0.00000179, 0.00000118, 0.00000049, -0.00000028, -0.00000112, -0.00000202, -0.00000298, -0.00000396, -0.00000497, -0.00000599, -0.00000700, -0.00000798, -0.00000892, -0.00000980, -0.00001061, -0.00001133, -0.00001195, -0.00001246, -0.00001284, -0.00001309, -0.00001320, -0.00001317, -0.00001299, -0.00001267, -0.00001220, -0.00001159, -0.00001085, -0.00000998, -0.00000901, -0.00000793, -0.00000677, -0.00000553, -0.00000425, -0.00000292, -0.00000158, -0.00000024, 0.00000108, 0.00000237, 0.00000360, 0.00000477, 0.00000585, 0.00000684, 0.00000773, 0.00000849, 0.00000914, 0.00000965, 0.00001004, 0.00001029, 0.00001042, 0.00001041, 0.00001029, 0.00001005, 0.00000970, 0.00000926, 0.00000873, 0.00000814, 0.00000748, 0.00000678, 0.00000605, 0.00000531, 0.00000455, 0.00000381, 0.00000308, 0.00000239, 0.00000174, 0.00000113, 0.00000058, 0.00000009, -0.00000034, -0.00000071, -0.00000101, -0.00000124, -0.00000141, -0.00000153, -0.00000159, -0.00000161, -0.00000158, -0.00000153, -0.00000144, -0.00000134, -0.00000123, -0.00000111, -0.00000100, -0.00000090, -0.00000081, -0.00000074, -0.00000070, -0.00000067, -0.00000068, -0.00000071, -0.00000076, -0.00000084, -0.00000093, -0.00000104, -0.00000116, -0.00000129, },
		{ 0.00000526, 0.00000528, 0.00000526, 0.00000519, 0.00000508, 0.00000490, 0.00000467, 0.00000438, 0.00000403, 0.00000361, 0.00000314, 0.00000261, 0.00000202, 0.00000140, 0.00000073, 0.00000004, -0.00000068, -0.00000140, -0.00000212, -0.00000282, -0.00000349, -0.00000412, -0.00000470, -0.00000521, -0.00000564, -0.00000599, -0.00000623, -0.00000638, -0.00000641, -0.00000633, -0.00000613, -0.00000581, -0.00000538, -0.00000484, -0.00000419, -0.00000344, -0.00000261, -0.00000170, -0.00000073, 0.00000028, 0.00000133, 0.00000238, 0.00000343, 0.00000445, 0.00000543, 0.00000635, 0.00000719, 0.00000793, 0.00000857, 0.00000908, 0.00000946, 0.00000969, 0.00000977, 0.00000969, 0.00000946, 0.00000906, 0.00000851, 0.00000780, 0.00000696, 0.00000598, 0.00000487, 0.00000366, 0.00000236, 0.00000099, -0.00000044, -0.00000191, -0.00000339, -0.00000487, -0.00000632, -0.00000773, -0.00000908, -0.00001035, -0.00001153, -0.00001260, -0.00001354, -0.00001436, -0.00001503, -0.00001556, -0.00001594, -0.00001616, -0.00001624, -0.00001617, -0.00001596, -0.00001562, -0.00001515, -0.00001457, -0.00001389, -0.00001311, -0.00001227, -0.00001136, -0.00001041, -0.00000943, -0.00000843, -0.00000742, -0.00000643, -0.00000545, -0.00000451, -0.00000361, -0.00000276, -0.00000196, -0.00000122, -0.00000055, 0.00000007, 0.00000062, 0.00000110, 0.00000153, 0.00000191, 0.00000223, 0.00000250, 0.00000274, 0.00000294, 0.00000312, 0.00000328, 0.00000342, 0.00000355, 0.00000367, 0.00000380, 0.00000392, 0.00000405, 0.00000419, 0.00000433, 0.00000447, 0.00000461, 0.00000475, 0.00000489, 0.00000501, 0.00000512, 0.00000520, },
		{ -0.00000102, -0.00000050, 0.00000003, 0.00000059, 0.00000117, 0.00000178, 0.00000242, 0.00000308, 0.00000377, 0.00000447, 0.00000520, 0.00000593, 0.00000666, 0.00000740, 0.00000811, 0.00000881, 0.00000947, 0.00001008, 0.00001064, 0.00001113, 0.00001154, 0.00001186, 0.00001208, 0.00001219, 0.00001217, 0.00001203, 0.00001176, 0.00001135, 0.00001080, 0.00001012, 0.00000930, 0.00000835, 0.00000727, 0.00000608, 0.00000478, 0.00000340, 0.00000193, 0.00000041, -0.00000115, -0.00000273, -0.00000432, -0.00000589, -0.00000742, -0.00000889, -0.00001027, -0.00001156, -0.00001273, -0.00001377, -0.00001465, -0.00001538, -0.00001593, -0.00001630, -0.00001648, -0.00001648, -0.00001628, -0.00001590, -0.00001534, -0.00001461, -0.00001372, -0.00001269, -0.00001152, -0.00001024, -0.00000886, -0.00000742, -0.00000592, -0.00000439, -0.00000285, -0.00000133, 0.00000015, 0.00000158, 0.00000292, 0.00000418, 0.00000532, 0.00000633, 0.00000720, 0.00000793, 0.00000849, 0.00000890, 0.00000915, 0.00000923, 0.00000916, 0.00000893, 0.00000856, 0.00000805, 0.00000741, 0.00000667, 0.00000583, 0.00000491, 0.00000393, 0.00000289, 0.00000183, 0.00000076, -0.00000031, -0.00000137, -0.00000239, -0.00000337, -0.00000428, -0.00000514, -0.00000591, -0.00000660, -0.00000721, -0.00000772, -0.00000814, -0.00000847, -0.00000871, -0.00000886, -0.00000893, -0.00000892, -0.00000883, -0.00000869, -0.00000848, -0.00000822, -0.00000792, -0.00000758, -0.00000722, -0.00000683, -0.00000642, -0.00000600, -0.00000557, -0.00000514, -0.00000470, -0.00000426, -0.00000382, -0.00000337, -0.00000292, -0.00000246, -0.00000199, -0.00000151, },
		{ -0.00000171, -0.00000189, -0.00000207, -0.00000224, -0.00000241, -0.00000258, -0.00000274, -0.00000291, -0.00000309, -0.00000326, -0.00000342, -0.00000358, -0.00000373, -0.00000387, -0.00000398, -0.00000406, -0.00000411, -0.00000412, -0.00000407, -0.00000397, -0.00000380, -0.00000357, -0.00000325, -0.00000285, -0.00000236, -0.00000179, -0.00000112, -0.00000036, 0.00000048, 0.00000141, 0.00000242, 0.00000350, 0.00000464, 0.00000583, 0.00000706, 0.00000832, 0.00000958, 0.00001084, 0.00001207, 0.00001326, 0.00001439, 0.00001545, 0.00001640, 0.00001725, 0.00001797, 0.00001855, 0.00001897, 0.00001923, 0.00001932, 0.00001922, 0.00001895, 0.00001848, 0.00001784, 0.00001701, 0.00001601, 0.00001485, 0.00001354, 0.00001208, 0.00001051, 0.00000883, 0.00000707, 0.00000525, 0.00000339, 0.00000151, -0.00000036, -0.00000221, -0.00000400, -0.00000572, -0.00000734, -0.00000886, -0.00001024, -0.00001147, -0.00001255, -0.00001346, -0.00001419, -0.00001474, -0.00001510, -0.00001528, -0.00001527, -0.00001509, -0.00001474, -0.00001423, -0.00001357, -0.00001279, -0.00001189, -0.00001088, -0.00000980, -0.00000866, -0.00000747, -0.00000626, -0.00000504, -0.00000383, -0.00000265, -0.00000151, -0.00000044, 0.00000057, 0.00000150, 0.00000233, 0.00000307, 0.00000370, 0.00000423, 0.00000466, 0.00000497, 0.00000519, 0.00000531, 0.00000533, 0.00000527, 0.00000513, 0.00000492, 0.00000466, 0.00000434, 0.00000398, 0.00000359, 0.00000319, 0.00000276, 0.00000234, 0.00000191, 0.00000150, 0.00000110, 0.00000072, 0.00000036, 0.00000002, -0.00000029, -0.00000058, -0.00000084, -0.00000109, -0.00000131, -0.00000152, },
		{ -0.00000241, -0.00000254, -0.00000268, -0.00000280, -0.00000291, -0.00000300, -0.00000307, -0.00000311, -0.00000313, -0.00000311, -0.00000308, -0.00000301, -0.00000292, -0.00000282, -0.00000270, -0.00000256, -0.00000243, -0.00000230, -0.00000217, -0.00000206, -0.00000197, -0.00000191, -0.00000188, -0.00000189, -0.00000194, -0.00000203, -0.00000217, -0.00000236, -0.00000259, -0.00000287, -0.00000319, -0.00000354, -0.00000393, -0.00000433, -0.00000475, -0.00000517, -0.00000559, -0.00000598, -0.00000634, -0.00000667, -0.00000693, -0.00000713, -0.00000725, -0.00000729, -0.00000722, -0.00000705, -0.00000677, -0.00000637, -0.00000585, -0.00000521, -0.00000444, -0.00000356, -0.00000256, -0.00000146, -0.00000026, 0.00000103, 0.00000239, 0.00000381, 0.00000527, 0.00000676, 0.00000827, 0.00000976, 0.00001123, 0.00001265, 0.00001400, 0.00001528, 0.00001646, 0.00001752, 0.00001846, 0.00001926, 0.00001991, 0.00002040, 0.00002073, 0.00002089, 0.00002089, 0.00002071, 0.00002038, 0.00001988, 0.00001924, 0.00001846, 0.00001755, 0.00001652, 0.00001540, 0.00001419, 0.00001292, 0.00001160, 0.00001025, 0.00000888, 0.00000752, 0.00000618, 0.00000488, 0.00000363, 0.00000244, 0.00000133, 0.00000030, -0.00000063, -0.00000147, -0.00000220, -0.00000283, -0.00000335, -0.00000376, -0.00000408, -0.00000430, -0.00000442, -0.00000447, -0.00000445, -0.00000436, -0.00000421, -0.00000403, -0.00000381, -0.00000357, -0.00000332, -0.00000306, -0.00000282, -0.00000258, -0.00000237, -0.00000218, -0.00000203, -0.00000191, -0.00000182, -0.00000177, -0.00000176, -0.00000178, -0.00000183, -0.00000191, -0.00000201, -0.00000213, -0.00000227, },
		{ -0.00000110, -0.00000092, -0.00000076, -0.00000062, -0.00000051, -0.00000044, -0.00000040, -0.00000040, -0.00000043, -0.00000051, -0.00000061, -0.00000075, -0.00000091, -0.00000110, -0.00000130, -0.00000152, -0.00000173, -0.00000194, -0.00000214, -0.00000232, -0.00000246, -0.00000258, -0.00000265, -0.00000267, -0.00000264, -0.00000256, -0.00000241, -0.00000221, -0.00000194, -0.00000161, -0.00000122, -0.00000078, -0.00000029, 0.00000024, 0.00000081, 0.00000140, 0.00000201, 0.00000263, 0.00000324, 0.00000384, 0.00000442, 0.00000496, 0.00000545, 0.00000589, 0.00000626, 0.00000657, 0.00000679, 0.00000692, 0.00000697, 0.00000693, 0.00000679, 0.00000656, 0.00000624, 0.00000584, 0.00000535, 0.00000480, 0.00000418, 0.00000350, 0.00000278, 0.00000202, 0.00000125, 0.00000046, -0.00000032, -0.00000109, -0.00000182, -0.00000252, -0.00000317, -0.00000375, -0.00000427, -0.00000471, -0.00000507, -0.00000534, -0.00000552, -0.00000560, -0.00000560, -0.00000551, -0.00000533, -0.00000507, -0.00000474, -0.00000435, -0.00000389, -0.00000339, -0.00000285, -0.00000229, -0.00000172, -0.00000114, -0.00000057, -0.00000001, 0.00000051, 0.00000099, 0.00000143, 0.00000181, 0.00000214, 0.00000240, 0.00000259, 0.00000271, 0.00000276, 0.00000275, 0.00000266, 0.00000252, 0.00000232, 0.00000207, 0.00000177, 0.00000144, 0.00000107, 0.00000069, 0.00000030, -0.00000009, -0.00000048, -0.00000085, -0.00000120, -0.00000152, -0.00000180, -0.00000205, -0.00000225, -0.00000241, -0.00000252, -0.00000258, -0.00000260, -0.00000257, -0.00000250, -0.00000239, -0.00000225, -0.00000209, -0.00000190, -0.00000171, -0.00000150, -0.00000130, },
		{ 0.00000149, 0.00000138, 0.00000130, 0.00000125, 0.00000123, 0.00000126, 0.00000131, 0.00000140, 0.00000151, 0.00000165, 0.00000181, 0.00000198, 0.00000216, 0.00000233, 0.00000250, 0.00000266, 0.00000279, 0.00000288, 0.00000294, 0.00000295, 0.00000291, 0.00000281, 0.00000265, 0.00000242, 0.00000211, 0.00000173, 0.00000128, 0.00000075, 0.00000016, -0.00000051, -0.00000124, -0.00000203, -0.00000287, -0.00000376, -0.00000467, -0.00000561, -0.00000656, -0.00000752, -0.00000846, -0.00000938, -0.00001026, -0.00001111, -0.00001189, -0.00001261, -0.00001325, -0.00001380, -0.00001426, -0.00001462, -0.00001488, -0.00001503, -0.00001507, -0.00001500, -0.00001482, -0.00001454, -0.00001415, -0.00001366, -0.00001309, -0.00001243, -0.00001169, -0.00001089, -0.00001004, -0.00000914, -0.00000822, -0.00000726, -0.00000630, -0.00000534, -0.00000439, -0.00000346, -0.00000256, -0.00000170, -0.00000088, -0.00000012, 0.00000059, 0.00000124, 0.00000183, 0.00000235, 0.00000281, 0.00000321, 0.00000354, 0.00000382, 0.00000404, 0.00000421, 0.00000434, 0.00000443, 0.00000448, 0.00000450, 0.00000451, 0.00000450, 0.00000447, 0.00000444, 0.00000441, 0.00000439, 0.00000437, 0.00000436, 0.00000436, 0.00000437, 0.00000440, 0.00000444, 0.00000448, 0.00000454, 0.00000461, 0.00000468, 0.00000475, 0.00000481, 0.00000487, 0.00000493, 0.00000497, 0.00000499, 0.00000499, 0.00000497, 0.00000493, 0.00000487, 0.00000477, 0.00000465, 0.00000451, 0.00000434, 0.00000415, 0.00000394, 0.00000371, 0.00000347, 0.00000322, 0.00000297, 0.00000271, 0.00000247, 0.00000223, 0.00000201, 0.00000181, 0.00000164, },
		{ 0.00000267, 0.00000232, 0.00000196, 0.00000157, 0.00000117, 0.00000075, 0.00000033, -0.00000010, -0.00000053, -0.00000094, -0.00000135, -0.00000174, -0.00000210, -0.00000243, -0.00000273, -0.00000298, -0.00000319, -0.00000334, -0.00000344, -0.00000348, -0.00000345, -0.00000335, -0.00000319, -0.00000296, -0.00000266, -0.00000230, -0.00000187, -0.00000137, -0.00000083, -0.00000023, 0.00000041, 0.00000109, 0.00000180, 0.00000253, 0.00000327, 0.00000402, 0.00000475, 0.00000547, 0.00000616, 0.00000681, 0.00000742, 0.00000797, 0.00000845, 0.00000886, 0.00000919, 0.00000944, 0.00000960, 0.00000966, 0.00000963, 0.00000949, 0.00000927, 0.00000894, 0.00000853, 0.00000802, 0.00000743, 0.00000677, 0.00000603, 0.00000523, 0.00000438, 0.00000348, 0.00000255, 0.00000160, 0.00000063, -0.00000034, -0.00000130, -0.00000224, -0.00000315, -0.00000402, -0.00000485, -0.00000562, -0.00000633, -0.00000697, -0.00000753, -0.00000802, -0.00000843, -0.00000876, -0.00000900, -0.00000916, -0.00000923, -0.00000923, -0.00000915, -0.00000900, -0.00000877, -0.00000849, -0.00000814, -0.00000775, -0.00000730, -0.00000682, -0.00000631, -0.00000576, -0.00000520, -0.00000462, -0.00000404, -0.00000345, -0.00000286, -0.00000229, -0.00000172, -0.00000116, -0.00000063, -0.00000011, 0.00000038, 0.00000085, 0.00000129, 0.00000171, 0.00000210, 0.00000246, 0.00000279, 0.00000309, 0.00000336, 0.00000361, 0.00000382, 0.00000401, 0.00000417, 0.00000430, 0.00000440, 0.00000448, 0.00000452, 0.00000453, 0.00000452, 0.00000447, 0.00000439, 0.00000429, 0.00000415, 0.00000397, 0.00000377, 0.00000354, 0.00000328, 0.00000299, },
		{ 0.00000176, 0.00000220, 0.00000263, 0.00000304, 0.00000342, 0.00000378, 0.00000412, 0.00000442, 0.00000470, 0.00000494, 0.00000514, 0.00000531, 0.00000543, 0.00000552, 0.00000556, 0.00000556, 0.00000552, 0.00000543, 0.00000530, 0.00000512, 0.00000491, 0.00000465, 0.00000435, 0.00000401, 0.00000364, 0.00000324, 0.00000281, 0.00000235, 0.00000187, 0.00000138, 0.00000087, 0.00000036, -0.00000015, -0.00000066, -0.00000116, -0.00000164, -0.00000210, -0.00000254, -0.00000294, -0.00000330, -0.00000362, -0.00000389, -0.00000411, -0.00000427, -0.00000438, -0.00000443, -0.00000442, -0.00000434, -0.00000420, -0.00000400, -0.00000373, -0.00000341, -0.00000304, -0.00000261, -0.00000213, -0.00000161, -0.00000105, -0.00000047, 0.00000015, 0.00000078, 0.00000142, 0.00000206, 0.00000270, 0.00000332, 0.00000393, 0.00000451, 0.00000505, 0.00000556, 0.00000602, 0.00000643, 0.00000678, 0.00000707, 0.00000730, 0.00000746, 0.00000755, 0.00000757, 0.00000752, 0.00000741, 0.00000722, 0.00000697, 0.00000666, 0.00000629, 0.00000587, 0.00000539, 0.00000487, 0.00000432, 0.00000373, 0.00000312, 0.00000248, 0.00000184, 0.00000119, 0.00000054, -0.00000010, -0.00000073, -0.00000133, -0.00000191, -0.00000247, -0.00000298, -0.00000346, -0.00000389, -0.00000427, -0.00000461, -0.00000490, -0.00000513, -0.00000531, -0.00000544, -0.00000552, -0.00000554, -0.00000551, -0.00000543, -0.00000530, -0.00000513, -0.00000491, -0.00000466, -0.00000436, -0.00000404, -0.00000368, -0.00000329, -0.00000288, -0.00000245, -0.00000201, -0.00000155, -0.00000108, -0.00000060, -0.00000012, 0.00000036, 0.00000083, 0.00000130, },
		{ -0.00000386, -0.00000391, -0.00000393, -0.00000392, -0.00000387, -0.00000380, -0.00000369, -0.00000356, -0.00000340, -0.00000321, -0.00000301, -0.00000278, -0.00000254, -0.00000228, -0.00000200, -0.00000172, -0.00000143, -0.00000113, -0.00000082, -0.00000051, -0.00000020, 0.00000011, 0.00000042, 0.00000073, 0.00000103, 0.00000133, 0.00000163, 0.00000191, 0.00000220, 0.00000247, 0.00000274, 0.00000300, 0.00000325, 0.00000350, 0.00000373, 0.00000394, 0.00000415, 0.00000434, 0.00000452, 0.00000468, 0.00000482, 0.00000494, 0.00000504, 0.00000512, 0.00000517, 0.00000519, 0.00000519, 0.00000515, 0.00000509, 0.00000499, 0.00000486, 0.00000469, 0.00000449, 0.00000426, 0.00000400, 0.00000370, 0.00000337, 0.00000302, 0.00000263, 0.00000223, 0.00000180, 0.00000135, 0.00000089, 0.00000042, -0.00000005, -0.00000052, -0.00000099, -0.00000146, -0.00000190, -0.00000233, -0.00000274, -0.00000312, -0.00000347, -0.00000378, -0.00000405, -0.00000429, -0.00000448, -0.00000462, -0.00000472, -0.00000477, -0.00000477, -0.00000473, -0.00000464, -0.00000451, -0.00000434, -0.00000414, -0.00000390, -0.00000363, -0.00000333, -0.00000302, -0.00000269, -0.00000235, -0.00000201, -0.00000167, -0.00000134, -0.00000102, -0.00000071, -0.00000043, -0.00000017, 0.00000006, 0.00000026, 0.00000042, 0.00000054, 0.00000063, 0.00000068, 0.00000069, 0.00000065, 0.00000059, 0.00000048, 0.00000034, 0.00000017, -0.00000003, -0.00000025, -0.00000050, -0.00000076, -0.00000103, -0.00000132, -0.00000160, -0.00000189, -0.00000217, -0.00000244, -0.00000269, -0.00000293, -0.00000315, -0.00000335, -0.00000352, -0.00000367, -0.00000378, },
		{ 0.00000004, 0.00000005, 0.00000003, -0.00000001, -0.00000006, -0.00000014, -0.00000023, -0.00000034, -0.00000046, -0.00000058, -0.00000070, -0.00000082, -0.00000094, -0.00000105, -0.00000115, -0.00000124, -0.00000132, -0.00000138, -0.00000143, -0.00000146, -0.00000147, -0.00000147, -0.00000146, -0.00000144, -0.00000141, -0.00000137, -0.00000133, -0.00000128, -0.00000124, -0.00000121, -0.00000119, -0.00000117, -0.00000117, -0.00000119, -0.00000123, -0.00000129, -0.00000136, -0.00000146, -0.00000159, -0.00000173, -0.00000189, -0.00000206, -0.00000226, -0.00000246, -0.00000267, -0.00000288, -0.00000309, -0.00000329, -0.00000348, -0.00000366, -0.00000381, -0.00000394, -0.00000403, -0.00000409, -0.00000411, -0.00000409, -0.00000403, -0.00000392, -0.00000376, -0.00000355, -0.00000330, -0.00000300, -0.00000266, -0.00000228, -0.00000187, -0.00000142, -0.00000095, -0.00000046, 0.00000005, 0.00000056, 0.00000107, 0.00000157, 0.00000206, 0.00000253, 0.00000296, 0.00000337, 0.00000373, 0.00000404, 0.00000431, 0.00000452, 0.00000467, 0.00000477, 0.00000480, 0.00000478, 0.00000469, 0.00000455, 0.00000435, 0.00000410, 0.00000380, 0.00000347, 0.00000309, 0.00000269, 0.00000226, 0.00000182, 0.00000137, 0.00000091, 0.00000046, 0.00000003, -0.00000039, -0.00000079, -0.00000116, -0.00000149, -0.00000179, -0.00000204, -0.00000226, -0.00000243, -0.00000255, -0.00000263, -0.00000266, -0.00000266, -0.00000261, -0.00000253, -0.00000241, -0.00000227, -0.00000211, -0.00000192, -0.00000173, -0.00000152, -0.00000131, -0.00000111, -0.00000091, -0.00000072, -0.00000055, -0.00000039, -0.00000026, -0.00000014, -0.00000006, 0.00000000, },
		{ -0.00000052, -0.00000056, -0.00000057, -0.00000056, -0.00000052, -0.00000047, -0.00000040, -0.00000032, -0.00000024, -0.00000015, -0.00000008, -0.00000001, 0.00000005, 0.00000009, 0.00000011, 0.00000011, 0.00000009, 0.00000005, -0.00000002, -0.00000010, -0.00000021, -0.00000033, -0.00000047, -0.00000061, -0.00000076, -0.00000091, -0.00000105, -0.00000118, -0.00000130, -0.00000139, -0.00000146, -0.00000150, -0.00000150, -0.00000146, -0.00000139, -0.00000127, -0.00000110, -0.00000089, -0.00000065, -0.00000036, -0.00000003, 0.00000033, 0.00000072, 0.00000112, 0.00000155, 0.00000198, 0.00000241, 0.00000283, 0.00000324, 0.00000363, 0.00000398, 0.00000429, 0.00000455, 0.00000476, 0.00000491, 0.00000500, 0.00000502, 0.00000496, 0.00000484, 0.00000464, 0.00000437, 0.00000403, 0.00000363, 0.00000317, 0.00000265, 0.00000208, 0.00000148, 0.00000084, 0.00000019, -0.00000048, -0.00000114, -0.00000180, -0.00000244, -0.00000304, -0.00000361, -0.00000413, -0.00000459, -0.00000498, -0.00000531, -0.00000556, -0.00000573, -0.00000582, -0.00000583, -0.00000576, -0.00000560, -0.00000537, -0.00000507, -0.00000470, -0.00000427, -0.00000379, -0.00000326, -0.00000270, -0.00000212, -0.00000151, -0.00000091, -0.00000030, 0.00000029, 0.00000086, 0.00000140, 0.00000190, 0.00000235, 0.00000276, 0.00000311, 0.00000340, 0.00000362, 0.00000379, 0.00000389, 0.00000393, 0.00000391, 0.00000383, 0.00000371, 0.00000353, 0.00000332, 0.00000307, 0.00000279, 0.00000249, 0.00000218, 0.00000187, 0.00000155, 0.00000124, 0.00000094, 0.00000066, 0.00000040, 0.00000017, -0.00000003, -0.00000020, -0.00000034, -0.00000045, },
		{ -0.00000100, -0.00000124, -0.00000151, -0.00000180, -0.00000210, -0.00000240, -0.00000270, -0.00000299, -0.00000327, -0.00000352, -0.00000375, -0.00000393, -0.00000409, -0.00000419, -0.00000426, -0.00000428, -0.00000425, -0.00000417, -0.00000406, -0.00000390, -0.00000370, -0.00000347, -0.00000322, -0.00000295, -0.00000266, -0.00000237, -0.00000209, -0.00000181, -0.00000156, -0.00000133, -0.00000113, -0.00000097, -0.00000085, -0.00000078, -0.00000076, -0.00000079, -0.00000088, -0.00000101, -0.00000120, -0.00000143, -0.00000171, -0.00000202, -0.00000236, -0.00000272, -0.00000309, -0.00000347, -0.00000384, -0.00000419, -0.00000452, -0.00000481, -0.00000505, -0.00000524, -0.00000536, -0.00000541, -0.00000539, -0.00000529, -0.00000510, -0.00000483, -0.00000447, -0.00000402, -0.00000349, -0.00000289, -0.00000222, -0.00000148, -0.00000069, 0.00000015, 0.00000102, 0.00000192, 0.00000282, 0.00000372, 0.00000460, 0.00000546, 0.00000628, 0.00000704, 0.00000774, 0.00000837, 0.00000891, 0.00000937, 0.00000973, 0.00001000, 0.00001016, 0.00001022, 0.00001017, 0.00001003, 0.00000979, 0.00000946, 0.00000904, 0.00000855, 0.00000800, 0.00000738, 0.00000672, 0.00000603, 0.00000532, 0.00000459, 0.00000386, 0.00000315, 0.00000246, 0.00000179, 0.00000117, 0.00000060, 0.00000008, -0.00000038, -0.00000077, -0.00000111, -0.00000137, -0.00000157, -0.00000171, -0.00000179, -0.00000181, -0.00000178, -0.00000171, -0.00000160, -0.00000147, -0.00000131, -0.00000114, -0.00000097, -0.00000080, -0.00000064, -0.00000051, -0.00000039, -0.00000031, -0.00000026, -0.00000024, -0.00000027, -0.00000034, -0.00000045, -0.00000059, -0.00000078, },
		{ -0.00000037, -0.00000011, 0.00000016, 0.00000043, 0.00000069, 0.00000094, 0.00000117, 0.00000137, 0.00000153, 0.00000165, 0.00000172, 0.00000174, 0.00000171, 0.00000163, 0.00000148, 0.00000129, 0.00000104, 0.00000075, 0.00000042, 0.00000005, -0.00000035, -0.00000076, -0.00000119, -0.00000162, -0.00000204, -0.00000244, -0.00000281, -0.00000314, -0.00000343, -0.00000366, -0.00000382, -0.00000393, -0.00000395, -0.00000391, -0.00000378, -0.00000358, -0.00000330, -0.00000295, -0.00000254, -0.00000206, -0.00000153, -0.00000096, -0.00000035, 0.00000028, 0.00000091, 0.00000155, 0.00000217, 0.00000276, 0.00000331, 0.00000381, 0.00000424, 0.00000460, 0.00000487, 0.00000506, 0.00000515, 0.00000514, 0.00000502, 0.00000480, 0.00000448, 0.00000405, 0.00000353, 0.00000292, 0.00000223, 0.00000147, 0.00000066, -0.00000021, -0.00000111, -0.00000203, -0.00000296, -0.00000387, -0.00000477, -0.00000563, -0.00000644, -0.00000719, -0.00000787, -0.00000846, -0.00000897, -0.00000937, -0.00000968, -0.00000988, -0.00000998, -0.00000997, -0.00000986, -0.00000965, -0.00000934, -0.00000896, -0.00000849, -0.00000796, -0.00000738, -0.00000675, -0.00000609, -0.00000541, -0.00000472, -0.00000403, -0.00000337, -0.00000272, -0.00000212, -0.00000156, -0.00000105, -0.00000060, -0.00000022, 0.00000010, 0.00000036, 0.00000054, 0.00000066, 0.00000072, 0.00000071, 0.00000066, 0.00000056, 0.00000042, 0.00000024, 0.00000005, -0.00000016, -0.00000037, -0.00000059, -0.00000079, -0.00000097, -0.00000113, -0.00000125, -0.00000134, -0.00000139, -0.00000140, -0.00000137, -0.00000129, -0.00000117, -0.00000102, -0.00000083, -0.00000061, },
		{ 0.00000083, 0.00000061, 0.00000039, 0.00000017, -0.00000004, -0.00000022, -0.00000038, -0.00000050, -0.00000058, -0.00000063, -0.00000062, -0.00000057, -0.00000048, -0.00000034, -0.00000016, 0.00000006, 0.00000031, 0.00000059, 0.00000088, 0.00000119, 0.00000149, 0.00000178, 0.00000205, 0.00000229, 0.00000249, 0.00000264, 0.00000272, 0.00000274, 0.00000269, 0.00000255, 0.00000232, 0.00000201, 0.00000160, 0.00000111, 0.00000053, -0.00000013, -0.00000087, -0.00000168, -0.00000255, -0.00000346, -0.00000441, -0.00000539, -0.00000637, -0.00000735, -0.00000831, -0.00000923, -0.00001011, -0.00001091, -0.00001164, -0.00001228, -0.00001282, -0.00001325, -0.00001355, -0.00001374, -0.00001379, -0.00001371, -0.00001350, -0.00001316, -0.00001269, -0.00001211, -0.00001142, -0.00001063, -0.00000975, -0.00000879, -0.00000778, -0.00000671, -0.00000562, -0.00000452, -0.00000342, -0.00000233, -0.00000128, -0.00000028, 0.00000066, 0.00000152, 0.00000231, 0.00000299, 0.00000358, 0.00000407, 0.00000444, 0.00000471, 0.00000487, 0.00000492, 0.00000487, 0.00000472, 0.00000449, 0.00000418, 0.00000380, 0.00000337, 0.00000289, 0.00000239, 0.00000186, 0.00000133, 0.00000080, 0.00000029, -0.00000019, -0.00000064, -0.00000103, -0.00000138, -0.00000166, -0.00000189, -0.00000205, -0.00000214, -0.00000216, -0.00000213, -0.00000203, -0.00000188, -0.00000168, -0.00000144, -0.00000116, -0.00000086, -0.00000055, -0.00000023, 0.00000010, 0.00000041, 0.00000070, 0.00000097, 0.00000121, 0.00000140, 0.00000156, 0.00000168, 0.00000174, 0.00000176, 0.00000174, 0.00000167, 0.00000156, 0.00000142, 0.00000125, 0.00000105, },
		{ 0.00000073, 0.00000089, 0.00000103, 0.00000115, 0.00000124, 0.00000130, 0.00000132, 0.00000129, 0.00000122, 0.00000110, 0.00000093, 0.00000072, 0.00000046, 0.00000017, -0.00000016, -0.00000052, -0.00000090, -0.00000130, -0.00000169, -0.00000208, -0.00000245, -0.00000279, -0.00000309, -0.00000335, -0.00000355, -0.00000369, -0.00000375, -0.00000373, -0.00000364, -0.00000345, -0.00000318, -0.00000282, -0.00000238, -0.00000185, -0.00000125, -0.00000058, 0.00000015, 0.00000093, 0.00000175, 0.00000259, 0.00000344, 0.00000428, 0.00000511, 0.00000590, 0.00000665, 0.00000733, 0.00000794, 0.00000846, 0.00000889, 0.00000920, 0.00000940, 0.00000948, 0.00000943, 0.00000925, 0.00000895, 0.00000852, 0.00000797, 0.00000731, 0.00000654, 0.00000567, 0.00000473, 0.00000371, 0.00000264, 0.00000153, 0.00000040, -0.00000074, -0.00000186, -0.00000296, -0.00000402, -0.00000502, -0.00000595, -0.00000679, -0.00000754, -0.00000819, -0.00000872, -0.00000914, -0.00000943, -0.00000960, -0.00000965, -0.00000958, -0.00000940, -0.00000911, -0.00000872, -0.00000825, -0.00000769, -0.00000707, -0.00000640, -0.00000568, -0.00000494, -0.00000418, -0.00000343, -0.00000269, -0.00000197, -0.00000129, -0.00000066, -0.00000008, 0.00000044, 0.00000089, 0.00000127, 0.00000158, 0.00000182, 0.00000198, 0.00000208, 0.00000211, 0.00000208, 0.00000200, 0.00000187, 0.00000170, 0.00000151, 0.00000129, 0.00000106, 0.00000083, 0.00000061, 0.00000040, 0.00000020, 0.00000004, -0.00000010, -0.00000020, -0.00000027, -0.00000030, -0.00000029, -0.00000024, -0.00000017, -0.00000006, 0.00000007, 0.00000022, 0.00000039, 0.00000056, },
		{ -0.00000079, -0.00000083, -0.00000084, -0.00000082, -0.00000076, -0.00000066, -0.00000052, -0.00000033, -0.00000010, 0.00000017, 0.00000049, 0.00000085, 0.00000124, 0.00000166, 0.00000210, 0.00000255, 0.00000301, 0.00000346, 0.00000390, 0.00000431, 0.00000468, 0.00000501, 0.00000529, 0.00000550, 0.00000564, 0.00000571, 0.00000570, 0.00000560, 0.00000542, 0.00000515, 0.00000479, 0.00000436, 0.00000384, 0.00000326, 0.00000261, 0.00000190, 0.00000116, 0.00000038, -0.00000041, -0.00000121, -0.00000201, -0.00000278, -0.00000351, -0.00000420, -0.00000482, -0.00000537, -0.00000583, -0.00000621, -0.00000648, -0.00000664, -0.00000669, -0.00000663, -0.00000645, -0.00000616, -0.00000576, -0.00000526, -0.00000466, -0.00000398, -0.00000322, -0.00000240, -0.00000153, -0.00000062, 0.00000031, 0.00000125, 0.00000218, 0.00000308, 0.00000394, 0.00000475, 0.00000550, 0.00000617, 0.00000675, 0.00000724, 0.00000762, 0.00000790, 0.00000807, 0.00000812, 0.00000807, 0.00000790, 0.00000764, 0.00000727, 0.00000682, 0.00000629, 0.00000569, 0.00000503, 0.00000433, 0.00000359, 0.00000284, 0.00000208, 0.00000132, 0.00000059, -0.00000012, -0.00000079, -0.00000141, -0.00000196, -0.00000246, -0.00000289, -0.00000324, -0.00000352, -0.00000372, -0.00000385, -0.00000391, -0.00000390, -0.00000383, -0.00000370, -0.00000352, -0.00000331, -0.00000306, -0.00000278, -0.00000250, -0.00000220, -0.00000191, -0.00000163, -0.00000136, -0.00000112, -0.00000090, -0.00000072, -0.00000057, -0.00000045, -0.00000037, -0.00000033, -0.00000031, -0.00000033, -0.00000037, -0.00000042, -0.00000050, -0.00000058, -0.00000066, -0.00000073, },
		{ 0.00000119, 0.00000127, 0.00000130, 0.00000130, 0.00000125, 0.00000116, 0.00000102, 0.00000084, 0.00000061, 0.00000035, 0.00000006, -0.00000027, -0.00000061, -0.00000097, -0.00000133, -0.00000169, -0.00000203, -0.00000236, -0.00000265, -0.00000290, -0.00000311, -0.00000326, -0.00000334, -0.00000336, -0.00000330, -0.00000316, -0.00000295, -0.00000265, -0.00000228, -0.00000183, -0.00000131, -0.00000073, -0.00000009, 0.00000061, 0.00000134, 0.00000210, 0.00000287, 0.00000366, 0.00000443, 0.00000518, 0.00000590, 0.00000657, 0.00000718, 0.00000772, 0.00000819, 0.00000856, 0.00000884, 0.00000901, 0.00000908, 0.00000904, 0.00000889, 0.00000863, 0.00000826, 0.00000779, 0.00000723, 0.00000657, 0.00000584, 0.00000505, 0.00000419, 0.00000330, 0.00000237, 0.00000143, 0.00000049, -0.00000045, -0.00000135, -0.00000222, -0.00000304, -0.00000380, -0.00000448, -0.00000509, -0.00000561, -0.00000604, -0.00000638, -0.00000662, -0.00000676, -0.00000680, -0.00000676, -0.00000662, -0.00000641, -0.00000612, -0.00000576, -0.00000535, -0.00000489, -0.00000440, -0.00000388, -0.00000334, -0.00000281, -0.00000228, -0.00000176, -0.00000127, -0.00000082, -0.00000040, -0.00000003, 0.00000029, 0.00000055, 0.00000076, 0.00000092, 0.00000102, 0.00000107, 0.00000108, 0.00000104, 0.00000095, 0.00000084, 0.00000070, 0.00000054, 0.00000037, 0.00000019, 0.00000002, -0.00000015, -0.00000030, -0.00000043, -0.00000053, -0.00000061, -0.00000066, -0.00000067, -0.00000065, -0.00000060, -0.00000052, -0.00000040, -0.00000027, -0.00000011, 0.00000007, 0.00000025, 0.00000044, 0.00000062, 0.00000079, 0.00000095, 0.00000108, },
		{ -0.00000110, -0.00000090, -0.00000066, -0.00000038, -0.00000007, 0.00000027, 0.00000065, 0.00000105, 0.00000147, 0.00000190, 0.00000235, 0.00000279, 0.00000323, 0.00000365, 0.00000405, 0.00000441, 0.00000474, 0.00000502, 0.00000525, 0.00000541, 0.00000551, 0.00000554, 0.00000550, 0.00000537, 0.00000517, 0.00000490, 0.00000455, 0.00000412, 0.00000363, 0.00000308, 0.00000247, 0.00000182, 0.00000113, 0.00000042, -0.00000031, -0.00000105, -0.00000177, -0.00000248, -0.00000315, -0.00000379, -0.00000437, -0.00000489, -0.00000533, -0.00000570, -0.00000598, -0.00000616, -0.00000626, -0.00000625, -0.00000614, -0.00000594, -0.00000564, -0.00000526, -0.00000478, -0.00000423, -0.00000361, -0.00000293, -0.00000220, -0.00000143, -0.00000064, 0.00000017, 0.00000098, 0.00000178, 0.00000256, 0.00000330, 0.00000400, 0.00000465, 0.00000523, 0.00000574, 0.00000617, 0.00000652, 0.00000679, 0.00000696, 0.00000705, 0.00000706, 0.00000697, 0.00000681, 0.00000657, 0.00000626, 0.00000589, 0.00000547, 0.00000499, 0.00000448, 0.00000395, 0.00000339, 0.00000283, 0.00000227, 0.00000171, 0.00000117, 0.00000066, 0.00000017, -0.00000027, -0.00000068, -0.00000104, -0.00000136, -0.00000163, -0.00000185, -0.00000203, -0.00000216, -0.00000225, -0.00000231, -0.00000232, -0.00000231, -0.00000228, -0.00000222, -0.00000215, -0.00000207, -0.00000198, -0.00000190, -0.00000182, -0.00000174, -0.00000168, -0.00000163, -0.00000159, -0.00000156, -0.00000155, -0.00000155, -0.00000156, -0.00000157, -0.00000159, -0.00000161, -0.00000163, -0.00000164, -0.00000163, -0.00000161, -0.00000157, -0.00000150, -0.00000140, -0.00000126, },
		{ -0.00000033, -0.00000054, -0.00000076, -0.00000100, -0.00000125, -0.00000149, -0.00000174, -0.00000199, -0.00000223, -0.00000245, -0.00000265, -0.00000283, -0.00000298, -0.00000309, -0.00000316, -0.00000319, -0.00000316, -0.00000308, -0.00000295, -0.00000276, -0.00000250, -0.00000219, -0.00000183, -0.00000140, -0.00000092, -0.00000040, 0.00000017, 0.00000078, 0.00000141, 0.00000207, 0.00000274, 0.00000342, 0.00000409, 0.00000475, 0.00000539, 0.00000600, 0.00000656, 0.00000707, 0.00000752, 0.00000791, 0.00000823, 0.00000847, 0.00000863, 0.00000871, 0.00000869, 0.00000860, 0.00000841, 0.00000815, 0.00000780, 0.00000738, 0.00000689, 0.00000634, 0.00000573, 0.00000508, 0.00000440, 0.00000369, 0.00000296, 0.00000223, 0.00000150, 0.00000079, 0.00000011, -0.00000055, -0.00000115, -0.00000171, -0.00000221, -0.00000264, -0.00000300, -0.00000330, -0.00000351, -0.00000366, -0.00000372, -0.00000372, -0.00000364, -0.00000349, -0.00000329, -0.00000302, -0.00000271, -0.00000235, -0.00000195, -0.00000153, -0.00000108, -0.00000062, -0.00000016, 0.00000030, 0.00000074, 0.00000118, 0.00000158, 0.00000196, 0.00000231, 0.00000262, 0.00000289, 0.00000312, 0.00000331, 0.00000345, 0.00000355, 0.00000361, 0.00000363, 0.00000362, 0.00000357, 0.00000350, 0.00000340, 0.00000328, 0.00000314, 0.00000299, 0.00000283, 0.00000267, 0.00000251, 0.00000235, 0.00000219, 0.00000204, 0.00000190, 0.00000177, 0.00000164, 0.00000153, 0.00000142, 0.00000132, 0.00000122, 0.00000113, 0.00000104, 0.00000095, 0.00000085, 0.00000075, 0.00000063, 0.00000051, 0.00000037, 0.00000022, 0.00000006, -0.00000013, },
		{ -0.00000026, -0.00000029, -0.00000031, -0.00000032, -0.00000032, -0.00000033, -0.00000034, -0.00000035, -0.00000037, -0.00000040, -0.00000044, -0.00000049, -0.00000057, -0.00000066, -0.00000078, -0.00000092, -0.00000108, -0.00000127, -0.00000148, -0.00000171, -0.00000197, -0.00000224, -0.00000253, -0.00000283, -0.00000314, -0.00000346, -0.00000377, -0.00000407, -0.00000437, -0.00000464, -0.00000489, -0.00000512, -0.00000531, -0.00000546, -0.00000556, -0.00000562, -0.00000562, -0.00000557, -0.00000547, -0.00000530, -0.00000507, -0.00000479, -0.00000445, -0.00000405, -0.00000360, -0.00000310, -0.00000255, -0.00000197, -0.00000136, -0.00000071, -0.00000005, 0.00000062, 0.00000130, 0.00000198, 0.00000264, 0.00000329, 0.00000391, 0.00000450, 0.00000505, 0.00000555, 0.00000601, 0.00000641, 0.00000676, 0.00000705, 0.00000727, 0.00000743, 0.00000754, 0.00000757, 0.00000756, 0.00000748, 0.00000735, 0.00000717, 0.00000695, 0.00000669, 0.00000640, 0.00000608, 0.00000574, 0.00000538, 0.00000502, 0.00000465, 0.00000428, 0.00000393, 0.00000358, 0.00000326, 0.00000295, 0.00000267, 0.00000241, 0.00000219, 0.00000199, 0.00000182, 0.00000168, 0.00000157, 0.00000149, 0.00000143, 0.00000139, 0.00000137, 0.00000137, 0.00000139, 0.00000141, 0.00000144, 0.00000147, 0.00000150, 0.00000153, 0.00000155, 0.00000157, 0.00000157, 0.00000156, 0.00000154, 0.00000150, 0.00000146, 0.00000139, 0.00000132, 0.00000123, 0.00000113, 0.00000103, 0.00000091, 0.00000079, 0.00000067, 0.00000055, 0.00000043, 0.00000032, 0.00000021, 0.00000011, 0.00000002, -0.00000006, -0.00000012, -0.00000018, -0.00000023, },
		{ -0.00000017, -0.00000019, -0.00000023, -0.00000027, -0.00000031, -0.00000035, -0.00000039, -0.00000042, -0.00000045, -0.00000047, -0.00000048, -0.00000048, -0.00000046, -0.00000044, -0.00000040, -0.00000035, -0.00000028, -0.00000021, -0.00000012, -0.00000003, 0.00000008, 0.00000018, 0.00000029, 0.00000040, 0.00000051, 0.00000061, 0.00000070, 0.00000078, 0.00000085, 0.00000090, 0.00000093, 0.00000094, 0.00000093, 0.00000090, 0.00000084, 0.00000076, 0.00000065, 0.00000052, 0.00000037, 0.00000020, 0.00000001, -0.00000020, -0.00000042, -0.00000065, -0.00000089, -0.00000113, -0.00000138, -0.00000161, -0.00000184, -0.00000206, -0.00000226, -0.00000245, -0.00000261, -0.00000275, -0.00000286, -0.00000294, -0.00000300, -0.00000302, -0.00000301, -0.00000298, -0.00000291, -0.00000281, -0.00000269, -0.00000254, -0.00000237, -0.00000218, -0.00000198, -0.00000176, -0.00000154, -0.00000131, -0.00000108, -0.00000086, -0.00000064, -0.00000044, -0.00000025, -0.00000008, 0.00000008, 0.00000021, 0.00000031, 0.00000039, 0.00000044, 0.00000047, 0.00000047, 0.00000044, 0.00000039, 0.00000032, 0.00000022, 0.00000011, -0.00000002, -0.00000016, -0.00000031, -0.00000047, -0.00000063, -0.00000079, -0.00000094, -0.00000109, -0.00000123, -0.00000136, -0.00000147, -0.00000157, -0.00000164, -0.00000170, -0.00000174, -0.00000176, -0.00000175, -0.00000173, -0.00000169, -0.00000163, -0.00000156, -0.00000147, -0.00000138, -0.00000127, -0.00000116, -0.00000104, -0.00000092, -0.00000081, -0.00000070, -0.00000059, -0.00000049, -0.00000041, -0.00000033, -0.00000026, -0.00000021, -0.00000017, -0.00000015, -0.00000014, -0.00000014, -0.00000015, },
		{ 0.00000028, 0.00000033, 0.00000038, 0.00000042, 0.00000046, 0.00000048, 0.00000050, 0.00000050, 0.00000049, 0.00000046, 0.00000042, 0.00000036, 0.00000029, 0.00000020, 0.00000010, -0.00000001, -0.00000014, -0.00000027, -0.00000041, -0.00000056, -0.00000071, -0.00000087, -0.00000102, -0.00000118, -0.00000133, -0.00000147, -0.00000161, -0.00000174, -0.00000186, -0.00000197, -0.00000207, -0.00000216, -0.00000225, -0.00000232, -0.00000239, -0.00000245, -0.00000251, -0.00000256, -0.00000261, -0.00000266, -0.00000272, -0.00000278, -0.00000284, -0.00000291, -0.00000299, -0.00000308, -0.00000318, -0.00000329, -0.00000341, -0.00000354, -0.00000369, -0.00000384, -0.00000400, -0.00000417, -0.00000434, -0.00000451, -0.00000468, -0.00000485, -0.00000501, -0.00000516, -0.00000529, -0.00000541, -0.00000551, -0.00000558, -0.00000563, -0.00000565, -0.00000563, -0.00000559, -0.00000550, -0.00000539, -0.00000523, -0.00000504, -0.00000481, -0.00000455, -0.00000426, -0.00000393, -0.00000358, -0.00000320, -0.00000280, -0.00000239, -0.00000196, -0.00000152, -0.00000108, -0.00000064, -0.00000021, 0.00000021, 0.00000062, 0.00000101, 0.00000137, 0.00000171, 0.00000201, 0.00000229, 0.00000252, 0.00000272, 0.00000288, 0.00000300, 0.00000308, 0.00000313, 0.00000314, 0.00000311, 0.00000305, 0.00000296, 0.00000284, 0.00000270, 0.00000254, 0.00000236, 0.00000217, 0.00000198, 0.00000178, 0.00000157, 0.00000137, 0.00000118, 0.00000100, 0.00000083, 0.00000068, 0.00000054, 0.00000042, 0.00000032, 0.00000024, 0.00000017, 0.00000013, 0.00000010, 0.00000010, 0.00000010, 0.00000012, 0.00000015, 0.00000019, 0.00000024, },
		{ 0.00000018, 0.00000002, -0.00000012, -0.00000026, -0.00000037, -0.00000047, -0.00000055, -0.00000061, -0.00000065, -0.00000067, -0.00000068, -0.00000067, -0.00000065, -0.00000061, -0.00000057, -0.00000052, -0.00000046, -0.00000040, -0.00000034, -0.00000029, -0.00000024, -0.00000019, -0.00000015, -0.00000012, -0.00000010, -0.00000009, -0.00000009, -0.00000010, -0.00000011, -0.00000014, -0.00000016, -0.00000019, -0.00000023, -0.00000026, -0.00000028, -0.00000030, -0.00000032, -0.00000032, -0.00000031, -0.00000028, -0.00000024, -0.00000018, -0.00000011, -0.00000001, 0.00000010, 0.00000023, 0.00000037, 0.00000053, 0.00000069, 0.00000087, 0.00000106, 0.00000124, 0.00000142, 0.00000160, 0.00000177, 0.00000192, 0.00000206, 0.00000217, 0.00000226, 0.00000232, 0.00000234, 0.00000233, 0.00000228, 0.00000219, 0.00000205, 0.00000188, 0.00000167, 0.00000141, 0.00000112, 0.00000079, 0.00000043, 0.00000004, -0.00000037, -0.00000080, -0.00000125, -0.00000170, -0.00000216, -0.00000261, -0.00000305, -0.00000347, -0.00000387, -0.00000424, -0.00000458, -0.00000488, -0.00000513, -0.00000534, -0.00000550, -0.00000561, -0.00000566, -0.00000566, -0.00000560, -0.00000549, -0.00000533, -0.00000513, -0.00000488, -0.00000458, -0.00000425, -0.00000389, -0.00000351, -0.00000310, -0.00000268, -0.00000225, -0.00000182, -0.00000140, -0.00000099, -0.00000059, -0.00000021, 0.00000014, 0.00000046, 0.00000075, 0.00000100, 0.00000121, 0.00000139, 0.00000152, 0.00000162, 0.00000168, 0.00000170, 0.00000168, 0.00000164, 0.00000156, 0.00000146, 0.00000133, 0.00000119, 0.00000103, 0.00000086, 0.00000069, 0.00000052, 0.00000034, },
		{ -0.00000056, -0.00000040, -0.00000028, -0.00000018, -0.00000010, -0.00000006, -0.00000003, -0.00000003, -0.00000005, -0.00000009, -0.00000013, -0.00000019, -0.00000025, -0.00000032, -0.00000038, -0.00000044, -0.00000049, -0.00000052, -0.00000055, -0.00000056, -0.00000055, -0.00000053, -0.00000049, -0.00000043, -0.00000036, -0.00000027, -0.00000018, -0.00000007, 0.00000004, 0.00000016, 0.00000027, 0.00000038, 0.00000048, 0.00000056, 0.00000063, 0.00000068, 0.00000071, 0.00000071, 0.00000069, 0.00000063, 0.00000055, 0.00000044, 0.00000030, 0.00000013, -0.00000006, -0.00000028, -0.00000052, -0.00000077, -0.00000104, -0.00000131, -0.00000158, -0.00000184, -0.00000210, -0.00000234, -0.00000256, -0.00000276, -0.00000292, -0.00000305, -0.00000314, -0.00000318, -0.00000318, -0.00000314, -0.00000304, -0.00000290, -0.00000271, -0.00000247, -0.00000220, -0.00000188, -0.00000153, -0.00000115, -0.00000075, -0.00000033, 0.00000010, 0.00000054, 0.00000097, 0.00000139, 0.00000179, 0.00000217, 0.00000251, 0.00000282, 0.00000308, 0.00000330, 0.00000346, 0.00000357, 0.00000362, 0.00000361, 0.00000354, 0.00000341, 0.00000323, 0.00000300, 0.00000271, 0.00000238, 0.00000201, 0.00000160, 0.00000117, 0.00000072, 0.00000025, -0.00000023, -0.00000070, -0.00000117, -0.00000162, -0.00000205, -0.00000245, -0.00000282, -0.00000315, -0.00000343, -0.00000368, -0.00000387, -0.00000402, -0.00000411, -0.00000416, -0.00000415, -0.00000410, -0.00000401, -0.00000388, -0.00000371, -0.00000351, -0.00000328, -0.00000303, -0.00000277, -0.00000249, -0.00000222, -0.00000194, -0.00000167, -0.00000141, -0.00000117, -0.00000094, -0.00000074, },
		{ -0.00000219, -0.00000220, -0.00000217, -0.00000211, -0.00000204, -0.00000194, -0.00000183, -0.00000171, -0.00000158, -0.00000145, -0.00000132, -0.00000120, -0.00000110, -0.00000100, -0.00000093, -0.00000088, -0.00000085, -0.00000085, -0.00000088, -0.00000093, -0.00000101, -0.00000111, -0.00000123, -0.00000137, -0.00000152, -0.00000168, -0.00000185, -0.00000202, -0.00000218, -0.00000233, -0.00000247, -0.00000259, -0.00000268, -0.00000275, -0.00000278, -0.00000277, -0.00000273, -0.00000264, -0.00000252, -0.00000235, -0.00000214, -0.00000190, -0.00000162, -0.00000131, -0.00000097, -0.00000061, -0.00000024, 0.00000014, 0.00000053, 0.00000091, 0.00000127, 0.00000162, 0.00000194, 0.00000223, 0.00000248, 0.00000268, 0.00000283, 0.00000293, 0.00000298, 0.00000296, 0.00000288, 0.00000274, 0.00000255, 0.00000229, 0.00000199, 0.00000163, 0.00000123, 0.00000079, 0.00000032, -0.00000018, -0.00000069, -0.00000120, -0.00000172, -0.00000223, -0.00000272, -0.00000319, -0.00000362, -0.00000401, -0.00000436, -0.00000465, -0.00000489, -0.00000507, -0.00000518, -0.00000524, -0.00000523, -0.00000516, -0.00000502, -0.00000483, -0.00000459, -0.00000430, -0.00000397, -0.00000360, -0.00000321, -0.00000279, -0.00000236, -0.00000193, -0.00000150, -0.00000108, -0.00000067, -0.00000029, 0.00000005, 0.00000037, 0.00000064, 0.00000087, 0.00000105, 0.00000119, 0.00000127, 0.00000131, 0.00000129, 0.00000124, 0.00000114, 0.00000100, 0.00000083, 0.00000063, 0.00000040, 0.00000016, -0.00000010, -0.00000036, -0.00000062, -0.00000087, -0.00000111, -0.00000134, -0.00000155, -0.00000173, -0.00000188, -0.00000201, -0.00000210, -0.00000216, },
		{ 0.00000052, 0.00000044, 0.00000033, 0.00000022, 0.00000009, -0.00000005, -0.00000018, -0.00000032, -0.00000044, -0.00000055, -0.00000064, -0.00000071, -0.00000076, -0.00000078, -0.00000078, -0.00000074, -0.00000068, -0.00000059, -0.00000047, -0.00000033, -0.00000017, 0.00000001, 0.00000019, 0.00000039, 0.00000058, 0.00000077, 0.00000094, 0.00000110, 0.00000123, 0.00000134, 0.00000141, 0.00000144, 0.00000143, 0.00000137, 0.00000127, 0.00000112, 0.00000092, 0.00000067, 0.00000038, 0.00000004, -0.00000034, -0.00000075, -0.00000118, -0.00000164, -0.00000212, -0.00000259, -0.00000307, -0.00000354, -0.00000399, -0.00000441, -0.00000479, -0.00000514, -0.00000543, -0.00000567, -0.00000584, -0.00000595, -0.00000600, -0.00000596, -0.00000586, -0.00000569, -0.00000544, -0.00000513, -0.00000476, -0.00000433, -0.00000384, -0.00000331, -0.00000275, -0.00000215, -0.00000154, -0.00000092, -0.00000030, 0.00000031, 0.00000091, 0.00000147, 0.00000200, 0.00000248, 0.00000291, 0.00000328, 0.00000359, 0.00000384, 0.00000401, 0.00000412, 0.00000415, 0.00000412, 0.00000402, 0.00000385, 0.00000363, 0.00000335, 0.00000303, 0.00000267, 0.00000227, 0.00000185, 0.00000142, 0.00000098, 0.00000054, 0.00000011, -0.00000031, -0.00000070, -0.00000106, -0.00000138, -0.00000166, -0.00000190, -0.00000209, -0.00000224, -0.00000233, -0.00000237, -0.00000237, -0.00000232, -0.00000223, -0.00000211, -0.00000195, -0.00000176, -0.00000155, -0.00000132, -0.00000109, -0.00000085, -0.00000062, -0.00000039, -0.00000018, 0.00000002, 0.00000019, 0.00000033, 0.00000045, 0.00000054, 0.00000059, 0.00000062, 0.00000061, 0.00000058, },
		{ 0.00000112, 0.00000122, 0.00000132, 0.00000142, 0.00000153, 0.00000162, 0.00000170, 0.00000176, 0.00000180, 0.00000182, 0.00000181, 0.00000177, 0.00000170, 0.00000160, 0.00000147, 0.00000132, 0.00000113, 0.00000093, 0.00000070, 0.00000046, 0.00000021, -0.00000004, -0.00000029, -0.00000054, -0.00000077, -0.00000097, -0.00000116, -0.00000131, -0.00000142, -0.00000149, -0.00000152, -0.00000150, -0.00000143, -0.00000131, -0.00000114, -0.00000093, -0.00000067, -0.00000036, -0.00000002, 0.00000036, 0.00000076, 0.00000117, 0.00000160, 0.00000204, 0.00000246, 0.00000288, 0.00000327, 0.00000363, 0.00000395, 0.00000423, 0.00000445, 0.00000461, 0.00000471, 0.00000475, 0.00000471, 0.00000460, 0.00000442, 0.00000417, 0.00000385, 0.00000346, 0.00000302, 0.00000252, 0.00000198, 0.00000139, 0.00000078, 0.00000015, -0.00000049, -0.00000114, -0.00000177, -0.00000239, -0.00000297, -0.00000352, -0.00000402, -0.00000447, -0.00000485, -0.00000517, -0.00000542, -0.00000559, -0.00000568, -0.00000570, -0.00000564, -0.00000550, -0.00000529, -0.00000502, -0.00000468, -0.00000429, -0.00000385, -0.00000337, -0.00000285, -0.00000232, -0.00000176, -0.00000121, -0.00000066, -0.00000012, 0.00000040, 0.00000089, 0.00000135, 0.00000177, 0.00000214, 0.00000246, 0.00000273, 0.00000295, 0.00000311, 0.00000322, 0.00000328, 0.00000329, 0.00000326, 0.00000318, 0.00000307, 0.00000293, 0.00000276, 0.00000257, 0.00000237, 0.00000217, 0.00000197, 0.00000177, 0.00000158, 0.00000141, 0.00000126, 0.00000113, 0.00000102, 0.00000095, 0.00000090, 0.00000088, 0.00000088, 0.00000091, 0.00000096, 0.00000103, },
		{ 0.00000038, 0.00000029, 0.00000020, 0.00000012, 0.00000005, -0.00000000, -0.00000003, -0.00000005, -0.00000003, 0.00000001, 0.00000008, 0.00000018, 0.00000030, 0.00000045, 0.00000062, 0.00000081, 0.00000102, 0.00000124, 0.00000147, 0.00000170, 0.00000192, 0.00000213, 0.00000233, 0.00000251, 0.00000265, 0.00000277, 0.00000284, 0.00000288, 0.00000287, 0.00000282, 0.00000271, 0.00000256, 0.00000236, 0.00000212, 0.00000183, 0.00000151, 0.00000115, 0.00000076, 0.00000034, -0.00000008, -0.00000052, -0.00000095, -0.00000138, -0.00000179, -0.00000218, -0.00000254, -0.00000285, -0.00000312, -0.00000334, -0.00000350, -0.00000359, -0.00000363, -0.00000359, -0.00000349, -0.00000332, -0.00000308, -0.00000278, -0.00000242, -0.00000201, -0.00000154, -0.00000104, -0.00000051, 0.00000005, 0.00000062, 0.00000120, 0.00000178, 0.00000234, 0.00000288, 0.00000338, 0.00000385, 0.00000427, 0.00000463, 0.00000494, 0.00000518, 0.00000535, 0.00000544, 0.00000547, 0.00000542, 0.00000530, 0.00000512, 0.00000486, 0.00000455, 0.00000419, 0.00000378, 0.00000333, 0.00000284, 0.00000234, 0.00000182, 0.00000130, 0.00000077, 0.00000027, -0.00000022, -0.00000069, -0.00000111, -0.00000150, -0.00000184, -0.00000214, -0.00000238, -0.00000257, -0.00000270, -0.00000278, -0.00000280, -0.00000277, -0.00000270, -0.00000258, -0.00000242, -0.00000223, -0.00000202, -0.00000178, -0.00000153, -0.00000126, -0.00000100, -0.00000074, -0.00000049, -0.00000026, -0.00000004, 0.00000015, 0.00000032, 0.00000045, 0.00000056, 0.00000064, 0.00000069, 0.00000070, 0.00000070, 0.00000067, 0.00000061, 0.00000055, 0.00000047, },
		{ -0.00000044, -0.00000026, -0.00000010, 0.00000005, 0.00000017, 0.00000028, 0.00000036, 0.00000041, 0.00000044, 0.00000043, 0.00000040, 0.00000035, 0.00000027, 0.00000018, 0.00000007, -0.00000006, -0.00000019, -0.00000032, -0.00000044, -0.00000056, -0.00000067, -0.00000075, -0.00000081, -0.00000084, -0.00000084, -0.00000080, -0.00000073, -0.00000061, -0.00000045, -0.00000025, -0.00000002, 0.00000026, 0.00000057, 0.00000091, 0.00000127, 0.00000166, 0.00000206, 0.00000247, 0.00000288, 0.00000328, 0.00000367, 0.00000403, 0.00000437, 0.00000467, 0.00000493, 0.00000514, 0.00000529, 0.00000539, 0.00000542, 0.00000539, 0.00000530, 0.00000514, 0.00000492, 0.00000463, 0.00000428, 0.00000388, 0.00000343, 0.00000294, 0.00000241, 0.00000185, 0.00000127, 0.00000068, 0.00000009, -0.00000050, -0.00000107, -0.00000163, -0.00000215, -0.00000263, -0.00000306, -0.00000345, -0.00000378, -0.00000405, -0.00000426, -0.00000440, -0.00000448, -0.00000449, -0.00000445, -0.00000434, -0.00000417, -0.00000396, -0.00000370, -0.00000340, -0.00000307, -0.00000271, -0.00000234, -0.00000195, -0.00000157, -0.00000119, -0.00000082, -0.00000048, -0.00000016, 0.00000013, 0.00000039, 0.00000060, 0.00000077, 0.00000089, 0.00000097, 0.00000101, 0.00000100, 0.00000094, 0.00000085, 0.00000073, 0.00000057, 0.00000038, 0.00000018, -0.00000004, -0.00000027, -0.00000050, -0.00000073, -0.00000095, -0.00000116, -0.00000135, -0.00000151, -0.00000165, -0.00000176, -0.00000184, -0.00000189, -0.00000190, -0.00000188, -0.00000183, -0.00000175, -0.00000164, -0.00000150, -0.00000135, -0.00000118, -0.00000100, -0.00000081, -0.00000063, },
		{ -0.00000022, -0.00000028, -0.00000031, -0.00000032, -0.00000030, -0.00000026, -0.00000019, -0.00000009, 0.00000002, 0.00000016, 0.00000032, 0.00000049, 0.00000067, 0.00000086, 0.00000105, 0.00000123, 0.00000141, 0.00000157, 0.00000172, 0.00000185, 0.00000194, 0.00000201, 0.00000205, 0.00000204, 0.00000201, 0.00000193, 0.00000181, 0.00000165, 0.00000146, 0.00000123, 0.00000098, 0.00000069, 0.00000038, 0.00000006, -0.00000028, -0.00000062, -0.00000096, -0.00000129, -0.00000161, -0.00000191, -0.00000218, -0.00000242, -0.00000262, -0.00000277, -0.00000288, -0.00000294, -0.00000294, -0.00000290, -0.00000279, -0.00000264, -0.00000243, -0.00000217, -0.00000186, -0.00000151, -0.00000113, -0.00000071, -0.00000027, 0.00000019, 0.00000066, 0.00000113, 0.00000159, 0.00000205, 0.00000248, 0.00000288, 0.00000326, 0.00000359, 0.00000388, 0.00000412, 0.00000430, 0.00000444, 0.00000451, 0.00000453, 0.00000449, 0.00000440, 0.00000425, 0.00000405, 0.00000381, 0.00000353, 0.00000321, 0.00000287, 0.00000250, 0.00000212, 0.00000172, 0.00000133, 0.00000094, 0.00000057, 0.00000021, -0.00000013, -0.00000043, -0.00000070, -0.00000094, -0.00000114, -0.00000130, -0.00000141, -0.00000148, -0.00000151, -0.00000151, -0.00000146, -0.00000138, -0.00000127, -0.00000113, -0.00000098, -0.00000080, -0.00000061, -0.00000042, -0.00000023, -0.00000004, 0.00000015, 0.00000032, 0.00000047, 0.00000060, 0.00000071, 0.00000080, 0.00000086, 0.00000089, 0.00000090, 0.00000089, 0.00000085, 0.00000078, 0.00000070, 0.00000061, 0.00000050, 0.00000039, 0.00000027, 0.00000015, 0.00000004, -0.00000006, -0.00000015, },
		{ 0.00000031, 0.00000028, 0.00000024, 0.00000017, 0.00000009, -0.00000000, -0.00000011, -0.00000023, -0.00000036, -0.00000049, -0.00000062, -0.00000076, -0.00000088, -0.00000100, -0.00000110, -0.00000118, -0.00000124, -0.00000127, -0.00000128, -0.00000125, -0.00000120, -0.00000110, -0.00000098, -0.00000082, -0.00000062, -0.00000039, -0.00000014, 0.00000014, 0.00000045, 0.00000078, 0.00000112, 0.00000146, 0.00000182, 0.00000217, 0.00000251, 0.00000284, 0.00000315, 0.00000344, 0.00000370, 0.00000392, 0.00000411, 0.00000425, 0.00000435, 0.00000441, 0.00000442, 0.00000438, 0.00000430, 0.00000417, 0.00000400, 0.00000379, 0.00000355, 0.00000327, 0.00000297, 0.00000265, 0.00000231, 0.00000196, 0.00000162, 0.00000127, 0.00000094, 0.00000062, 0.00000033, 0.00000006, -0.00000018, -0.00000037, -0.00000053, -0.00000065, -0.00000072, -0.00000075, -0.00000073, -0.00000066, -0.00000056, -0.00000041, -0.00000023, -0.00000001, 0.00000024, 0.00000051, 0.00000080, 0.00000110, 0.00000142, 0.00000173, 0.00000204, 0.00000233, 0.00000261, 0.00000287, 0.00000311, 0.00000332, 0.00000349, 0.00000362, 0.00000372, 0.00000378, 0.00000381, 0.00000379, 0.00000373, 0.00000364, 0.00000351, 0.00000336, 0.00000317, 0.00000296, 0.00000274, 0.00000250, 0.00000224, 0.00000199, 0.00000173, 0.00000148, 0.00000124, 0.00000101, 0.00000079, 0.00000060, 0.00000042, 0.00000027, 0.00000014, 0.00000004, -0.00000004, -0.00000009, -0.00000012, -0.00000013, -0.00000012, -0.00000010, -0.00000006, -0.00000001, 0.00000004, 0.00000010, 0.00000016, 0.00000021, 0.00000025, 0.00000029, 0.00000031, 0.00000032, },
		{ 0.00000029, 0.00000028, 0.00000029, 0.00000031, 0.00000035, 0.00000040, 0.00000045, 0.00000051, 0.00000057, 0.00000064, 0.00000069, 0.00000074, 0.00000078, 0.00000081, 0.00000082, 0.00000081, 0.00000078, 0.00000073, 0.00000066, 0.00000056, 0.00000044, 0.00000031, 0.00000015, -0.00000003, -0.00000022, -0.00000042, -0.00000063, -0.00000085, -0.00000106, -0.00000128, -0.00000148, -0.00000167, -0.00000185, -0.00000200, -0.00000213, -0.00000224, -0.00000231, -0.00000234, -0.00000234, -0.00000230, -0.00000223, -0.00000212, -0.00000196, -0.00000178, -0.00000156, -0.00000130, -0.00000102, -0.00000072, -0.00000039, -0.00000005, 0.00000030, 0.00000066, 0.00000101, 0.00000136, 0.00000170, 0.00000203, 0.00000233, 0.00000260, 0.00000285, 0.00000307, 0.00000325, 0.00000339, 0.00000349, 0.00000355, 0.00000357, 0.00000355, 0.00000349, 0.00000339, 0.00000326, 0.00000311, 0.00000292, 0.00000271, 0.00000248, 0.00000224, 0.00000200, 0.00000175, 0.00000150, 0.00000126, 0.00000103, 0.00000081, 0.00000062, 0.00000045, 0.00000030, 0.00000019, 0.00000010, 0.00000005, 0.00000003, 0.00000003, 0.00000007, 0.00000014, 0.00000023, 0.00000035, 0.00000049, 0.00000065, 0.00000082, 0.00000099, 0.00000118, 0.00000136, 0.00000154, 0.00000172, 0.00000188, 0.00000203, 0.00000216, 0.00000227, 0.00000236, 0.00000242, 0.00000246, 0.00000248, 0.00000246, 0.00000243, 0.00000237, 0.00000228, 0.00000218, 0.00000206, 0.00000193, 0.00000178, 0.00000163, 0.00000147, 0.00000131, 0.00000116, 0.00000101, 0.00000087, 0.00000074, 0.00000062, 0.00000052, 0.00000043, 0.00000037, 0.00000032, },
		{ 0.00000077, 0.00000067, 0.00000056, 0.00000044, 0.00000031, 0.00000019, 0.00000006, -0.00000006, -0.00000017, -0.00000028, -0.00000037, -0.00000044, -0.00000050, -0.00000054, -0.00000056, -0.00000056, -0.00000054, -0.00000050, -0.00000044, -0.00000037, -0.00000028, -0.00000018, -0.00000006, 0.00000006, 0.00000018, 0.00000030, 0.00000042, 0.00000054, 0.00000064, 0.00000073, 0.00000081, 0.00000086, 0.00000089, 0.00000090, 0.00000089, 0.00000084, 0.00000077, 0.00000068, 0.00000056, 0.00000041, 0.00000024, 0.00000005, -0.00000016, -0.00000038, -0.00000061, -0.00000085, -0.00000109, -0.00000133, -0.00000157, -0.00000179, -0.00000200, -0.00000219, -0.00000235, -0.00000249, -0.00000261, -0.00000269, -0.00000273, -0.00000275, -0.00000272, -0.00000266, -0.00000257, -0.00000244, -0.00000229, -0.00000210, -0.00000188, -0.00000165, -0.00000139, -0.00000112, -0.00000084, -0.00000055, -0.00000026, 0.00000002, 0.00000030, 0.00000056, 0.00000081, 0.00000104, 0.00000124, 0.00000142, 0.00000157, 0.00000169, 0.00000177, 0.00000183, 0.00000185, 0.00000184, 0.00000180, 0.00000173, 0.00000163, 0.00000152, 0.00000138, 0.00000122, 0.00000106, 0.00000088, 0.00000070, 0.00000052, 0.00000035, 0.00000018, 0.00000002, -0.00000012, -0.00000025, -0.00000036, -0.00000045, -0.00000051, -0.00000055, -0.00000057, -0.00000056, -0.00000053, -0.00000048, -0.00000041, -0.00000032, -0.00000022, -0.00000011, 0.00000001, 0.00000014, 0.00000027, 0.00000040, 0.00000053, 0.00000064, 0.00000075, 0.00000084, 0.00000092, 0.00000098, 0.00000102, 0.00000105, 0.00000105, 0.00000103, 0.00000099, 0.00000094, 0.00000086, },
		{ 0.00000048, 0.00000057, 0.00000066, 0.00000075, 0.00000082, 0.00000088, 0.00000093, 0.00000097, 0.00000098, 0.00000098, 0.00000096, 0.00000091, 0.00000085, 0.00000077, 0.00000067, 0.00000055, 0.00000042, 0.00000027, 0.00000011, -0.00000005, -0.00000021, -0.00000038, -0.00000055, -0.00000070, -0.00000085, -0.00000099, -0.00000111, -0.00000121, -0.00000129, -0.00000134, -0.00000138, -0.00000139, -0.00000137, -0.00000133, -0.00000126, -0.00000117, -0.00000106, -0.00000093, -0.00000079, -0.00000063, -0.00000046, -0.00000028, -0.00000011, 0.00000007, 0.00000024, 0.00000039, 0.00000054, 0.00000067, 0.00000078, 0.00000086, 0.00000092, 0.00000095, 0.00000095, 0.00000092, 0.00000086, 0.00000077, 0.00000066, 0.00000051, 0.00000034, 0.00000015, -0.00000006, -0.00000029, -0.00000052, -0.00000077, -0.00000102, -0.00000126, -0.00000150, -0.00000173, -0.00000195, -0.00000214, -0.00000231, -0.00000246, -0.00000258, -0.00000267, -0.00000272, -0.00000274, -0.00000273, -0.00000268, -0.00000259, -0.00000248, -0.00000233, -0.00000216, -0.00000196, -0.00000174, -0.00000151, -0.00000125, -0.00000099, -0.00000073, -0.00000046, -0.00000020, 0.00000006, 0.00000030, 0.00000053, 0.00000074, 0.00000093, 0.00000110, 0.00000124, 0.00000135, 0.00000143, 0.00000149, 0.00000152, 0.00000152, 0.00000150, 0.00000145, 0.00000139, 0.00000130, 0.00000120, 0.00000109, 0.00000097, 0.00000084, 0.00000072, 0.00000059, 0.00000047, 0.00000036, 0.00000026, 0.00000017, 0.00000010, 0.00000005, 0.00000001, -0.00000001, -0.00000001, 0.00000000, 0.00000004, 0.00000008, 0.00000015, 0.00000022, 0.00000030, 0.00000039, },
		{ 0.00000007, 0.00000002, -0.00000002, -0.00000005, -0.00000006, -0.00000006, -0.00000004, -0.00000001, 0.00000004, 0.00000011, 0.00000019, 0.00000029, 0.00000040, 0.00000052, 0.00000064, 0.00000077, 0.00000090, 0.00000102, 0.00000115, 0.00000126, 0.00000136, 0.00000144, 0.00000150, 0.00000155, 0.00000157, 0.00000156, 0.00000153, 0.00000148, 0.00000139, 0.00000128, 0.00000115, 0.00000099, 0.00000081, 0.00000061, 0.00000039, 0.00000017, -0.00000007, -0.00000031, -0.00000055, -0.00000079, -0.00000102, -0.00000124, -0.00000144, -0.00000162, -0.00000177, -0.00000190, -0.00000200, -0.00000207, -0.00000211, -0.00000211, -0.00000208, -0.00000202, -0.00000192, -0.00000179, -0.00000164, -0.00000146, -0.00000126, -0.00000104, -0.00000081, -0.00000057, -0.00000032, -0.00000007, 0.00000017, 0.00000040, 0.00000061, 0.00000081, 0.00000098, 0.00000113, 0.00000125, 0.00000134, 0.00000139, 0.00000141, 0.00000140, 0.00000135, 0.00000127, 0.00000115, 0.00000101, 0.00000083, 0.00000063, 0.00000042, 0.00000018, -0.00000007, -0.00000032, -0.00000058, -0.00000084, -0.00000109, -0.00000133, -0.00000156, -0.00000177, -0.00000195, -0.00000212, -0.00000225, -0.00000235, -0.00000242, -0.00000247, -0.00000248, -0.00000245, -0.00000240, -0.00000232, -0.00000221, -0.00000208, -0.00000193, -0.00000176, -0.00000158, -0.00000139, -0.00000119, -0.00000099, -0.00000079, -0.00000060, -0.00000041, -0.00000024, -0.00000008, 0.00000007, 0.00000019, 0.00000029, 0.00000038, 0.00000044, 0.00000048, 0.00000050, 0.00000051, 0.00000049, 0.00000047, 0.00000042, 0.00000037, 0.00000031, 0.00000025, 0.00000019, 0.00000012, },
		{ -0.00000132, -0.00000128, -0.00000125, -0.00000124, -0.00000124, -0.00000126, -0.00000130, -0.00000135, -0.00000141, -0.00000149, -0.00000157, -0.00000166, -0.00000175, -0.00000183, -0.00000192, -0.00000200, -0.00000207, -0.00000212, -0.00000216, -0.00000218, -0.00000217, -0.00000215, -0.00000210, -0.00000202, -0.00000193, -0.00000180, -0.00000165, -0.00000149, -0.00000130, -0.00000109, -0.00000087, -0.00000064, -0.00000040, -0.00000016, 0.00000007, 0.00000030, 0.00000053, 0.00000073, 0.00000092, 0.00000108, 0.00000121, 0.00000132, 0.00000139, 0.00000143, 0.00000143, 0.00000140, 0.00000132, 0.00000121, 0.00000107, 0.00000089, 0.00000067, 0.00000044, 0.00000017, -0.00000011, -0.00000041, -0.00000072, -0.00000104, -0.00000135, -0.00000166, -0.00000196, -0.00000225, -0.00000251, -0.00000275, -0.00000297, -0.00000315, -0.00000329, -0.00000340, -0.00000347, -0.00000350, -0.00000348, -0.00000343, -0.00000335, -0.00000322, -0.00000307, -0.00000288, -0.00000267, -0.00000243, -0.00000218, -0.00000191, -0.00000164, -0.00000137, -0.00000109, -0.00000083, -0.00000058, -0.00000034, -0.00000012, 0.00000007, 0.00000023, 0.00000036, 0.00000047, 0.00000054, 0.00000057, 0.00000058, 0.00000055, 0.00000049, 0.00000040, 0.00000028, 0.00000015, -0.00000001, -0.00000019, -0.00000038, -0.00000057, -0.00000077, -0.00000097, -0.00000117, -0.00000136, -0.00000153, -0.00000170, -0.00000184, -0.00000197, -0.00000208, -0.00000216, -0.00000222, -0.00000226, -0.00000228, -0.00000227, -0.00000225, -0.00000220, -0.00000214, -0.00000207, -0.00000199, -0.00000190, -0.00000180, -0.00000171, -0.00000162, -0.00000153, -0.00000145, -0.00000138, },
		},
		{ { -0.00000049, -0.00000049, -0.00000047, -0.00000044, -0.00000040, -0.00000035, -0.00000029, -0.00000022, -0.00000015, -0.00000006, 0.00000003, 0.00000013, 0.00000024, 0.00000035, 0.00000046, 0.00000058, 0.00000070, 0.00000082, 0.00000094, 0.00000105, 0.00000116, 0.00000127, 0.00000138, 0.00000148, 0.00000157, 0.00000166, 0.00000173, 0.00000180, 0.00000186, 0.00000191, 0.00000194, 0.00000197, 0.00000199, 0.00000200, 0.00000199, 0.00000198, 0.00000195, 0.00000191, 0.00000187, 0.00000181, 0.00000175, 0.00000168, 0.00000160, 0.00000152, 0.00000143, 0.00000133, 0.00000123, 0.00000113, 0.00000103, 0.00000093, 0.00000083, 0.00000073, 0.00000063, 0.00000054, 0.00000045, 0.00000037, 0.00000029, 0.00000022, 0.00000016, 0.00000010, 0.00000006, 0.00000002, -0.00000000, -0.00000002, -0.00000002, -0.00000002, -0.00000000, 0.00000002, 0.00000006, 0.00000010, 0.00000016, 0.00000022, 0.00000029, 0.00000037, 0.00000045, 0.00000054, 0.00000063, 0.00000073, 0.00000083, 0.00000093, 0.00000103, 0.00000113, 0.00000123, 0.00000133, 0.00000143, 0.00000152, 0.00000160, 0.00000168, 0.00000175, 0.00000181, 0.00000187, 0.00000191, 0.00000195, 0.00000198, 0.00000199, 0.00000200, 0.00000199, 0.00000197, 0.00000194, 0.00000191, 0.00000186, 0.00000180, 0.00000173, 0.00000166, 0.00000157, 0.00000148, 0.00000138, 0.00000127, 0.00000116, 0.00000105, 0.00000094, 0.00000082, 0.00000070, 0.00000058, 0.00000046, 0.00000035, 0.00000024, 0.00000013, 0.00000003, -0.00000006, -0.00000015, -0.00000022, -0.00000029, -0.00000035, -0.00000040, -0.00000044, -0.00000047, -0.00000049, },
		{ -0.00000069, -0.00000038, -0.00000004, 0.00000032, 0.00000068, 0.00000103, 0.00000138, 0.00000170, 0.00000199, 0.00000225, 0.00000247, 0.00000264, 0.00000276, 0.00000284, 0.00000286, 0.00000283, 0.00000276, 0.00000265, 0.00000251, 0.00000233, 0.00000214, 0.00000192, 0.00000170, 0.00000148, 0.00000126, 0.00000106, 0.00000086, 0.00000069, 0.00000054, 0.00000042, 0.00000031, 0.00000024, 0.00000018, 0.00000014, 0.00000011, 0.00000009, 0.00000008, 0.00000007, 0.00000006, 0.00000005, 0.00000002, -0.00000002, -0.00000008, -0.00000015, -0.00000023, -0.00000033, -0.00000044, -0.00000055, -0.00000067, -0.00000079, -0.00000090, -0.00000100, -0.00000109, -0.00000115, -0.00000119, -0.00000120, -0.00000118, -0.00000112, -0.00000103, -0.00000091, -0.00000075, -0.00000056, -0.00000035, -0.00000012, 0.00000012, 0.00000037, 0.00000062, 0.00000087, 0.00000110, 0.00000132, 0.00000151, 0.00000166, 0.00000178, 0.00000186, 0.00000191, 0.00000190, 0.00000186, 0.00000178, 0.00000166, 0.00000152, 0.00000134, 0.00000115, 0.00000094, 0.00000072, 0.00000051, 0.00000030, 0.00000010, -0.00000007, -0.00000022, -0.00000034, -0.00000043, -0.00000049, -0.00000052, -0.00000051, -0.00000047, -0.00000041, -0.00000033, -0.00000023, -0.00000012, -0.00000001, 0.00000010, 0.00000020, 0.00000028, 0.00000035, 0.00000038, 0.00000039, 0.00000037, 0.00000031, 0.00000022, 0.00000010, -0.00000005, -0.00000023, -0.00000042, -0.00000062, -0.00000083, -0.00000104, -0.00000123, -0.00000141, -0.00000156, -0.00000167, -0.00000175, -0.00000179, -0.00000177, -0.00000171, -0.00000160, -0.00000144, -0.00000123, -0.00000098, },
		{ 0.00000364, 0.00000364, 0.00000354, 0.00000337, 0.00000313, 0.00000282, 0.00000247, 0.00000207, 0.00000166, 0.00000123, 0.00000081, 0.00000041, 0.00000004, -0.00000030, -0.00000059, -0.00000083, -0.00000102, -0.00000115, -0.00000123, -0.00000127, -0.00000126, -0.00000122, -0.00000115, -0.00000106, -0.00000096, -0.00000087, -0.00000078, -0.00000070, -0.00000065, -0.00000063, -0.00000063, -0.00000066, -0.00000072, -0.00000081, -0.00000091, -0.00000103, -0.00000115, -0.00000128, -0.00000139, -0.00000148, -0.00000155, -0.00000159, -0.00000159, -0.00000155, -0.00000146, -0.00000133, -0.00000116, -0.00000095, -0.00000070, -0.00000043, -0.00000013, 0.00000017, 0.00000047, 0.00000076, 0.00000102, 0.00000126, 0.00000145, 0.00000159, 0.00000168, 0.00000171, 0.00000167, 0.00000157, 0.00000140, 0.00000118, 0.00000090, 0.00000058, 0.00000023, -0.00000015, -0.00000055, -0.00000094, -0.00000133, -0.00000169, -0.00000202, -0.00000230, -0.00000253, -0.00000270, -0.00000281, -0.00000285, -0.00000282, -0.00000273, -0.00000258, -0.00000237, -0.00000211, -0.00000182, -0.00000150, -0.00000116, -0.00000083, -0.00000051, -0.00000021, 0.00000006, 0.00000027, 0.00000043, 0.00000053, 0.00000056, 0.00000053, 0.00000042, 0.00000025, 0.00000001, -0.00000027, -0.00000060, -0.00000096, -0.00000135, -0.00000173, -0.00000211, -0.00000246, -0.00000278, -0.00000304, -0.00000324, -0.00000337, -0.00000342, -0.00000338, -0.00000326, -0.00000305, -0.00000275, -0.00000238, -0.00000194, -0.00000144, -0.00000090, -0.00000033, 0.00000025, 0.00000083, 0.00000139, 0.00000191, 0.00000238, 0.00000279, 0.00000313, 0.00000339, 0.00000356, },
		{ -0.00000022, -0.00000063, -0.00000097, -0.00000123, -0.00000142, -0.00000152, -0.00000155, -0.00000152, -0.00000142, -0.00000129, -0.00000112, -0.00000093, -0.00000074, -0.00000056, -0.00000040, -0.00000026, -0.00000016, -0.00000011, -0.00000009, -0.00000012, -0.00000019, -0.00000030, -0.00000044, -0.00000060, -0.00000076, -0.00000093, -0.00000109, -0.00000123, -0.00000134, -0.00000141, -0.00000144, -0.00000143, -0.00000136, -0.00000125, -0.00000110, -0.00000091, -0.00000069, -0.00000045, -0.00000020, 0.00000006, 0.00000030, 0.00000052, 0.00000071, 0.00000085, 0.00000095, 0.00000099, 0.00000097, 0.00000089, 0.00000075, 0.00000055, 0.00000030, 0.00000001, -0.00000032, -0.00000067, -0.00000102, -0.00000138, -0.00000173, -0.00000204, -0.00000232, -0.00000255, -0.00000272, -0.00000283, -0.00000286, -0.00000282, -0.00000271, -0.00000253, -0.00000228, -0.00000198, -0.00000162, -0.00000123, -0.00000081, -0.00000037, 0.00000006, 0.00000047, 0.00000085, 0.00000119, 0.00000147, 0.00000168, 0.00000182, 0.00000187, 0.00000183, 0.00000171, 0.00000150, 0.00000120, 0.00000084, 0.00000040, -0.00000008, -0.00000061, -0.00000115, -0.00000170, -0.00000223, -0.00000272, -0.00000317, -0.00000354, -0.00000383, -0.00000403, -0.00000412, -0.00000409, -0.00000395, -0.00000369, -0.00000332, -0.00000285, -0.00000228, -0.00000163, -0.00000092, -0.00000017, 0.00000060, 0.00000138, 0.00000213, 0.00000284, 0.00000349, 0.00000406, 0.00000453, 0.00000489, 0.00000514, 0.00000526, 0.00000526, 0.00000514, 0.00000491, 0.00000458, 0.00000416, 0.00000367, 0.00000313, 0.00000255, 0.00000195, 0.00000136, 0.00000079, 0.00000026, },
		{ -0.00000070, -0.00000046, -0.00000025, -0.00000008, 0.00000004, 0.00000010, 0.00000012, 0.00000007, -0.00000002, -0.00000015, -0.00000032, -0.00000051, -0.00000071, -0.00000090, -0.00000108, -0.00000124, -0.00000136, -0.00000143, -0.00000146, -0.00000143, -0.00000135, -0.00000122, -0.00000104, -0.00000082, -0.00000057, -0.00000030, -0.00000002, 0.00000026, 0.00000053, 0.00000077, 0.00000097, 0.00000114, 0.00000125, 0.00000131, 0.00000131, 0.00000126, 0.00000116, 0.00000100, 0.00000081, 0.00000058, 0.00000032, 0.00000005, -0.00000022, -0.00000048, -0.00000073, -0.00000095, -0.00000114, -0.00000129, -0.00000138, -0.00000143, -0.00000143, -0.00000137, -0.00000127, -0.00000112, -0.00000094, -0.00000072, -0.00000048, -0.00000023, 0.00000002, 0.00000027, 0.00000049, 0.00000069, 0.00000084, 0.00000095, 0.00000100, 0.00000099, 0.00000091, 0.00000076, 0.00000055, 0.00000027, -0.00000007, -0.00000047, -0.00000091, -0.00000138, -0.00000188, -0.00000238, -0.00000288, -0.00000335, -0.00000377, -0.00000415, -0.00000444, -0.00000465, -0.00000476, -0.00000476, -0.00000464, -0.00000441, -0.00000405, -0.00000357, -0.00000298, -0.00000230, -0.00000153, -0.00000069, 0.00000020, 0.00000111, 0.00000203, 0.00000292, 0.00000376, 0.00000454, 0.00000523, 0.00000580, 0.00000626, 0.00000657, 0.00000674, 0.00000677, 0.00000665, 0.00000639, 0.00000601, 0.00000551, 0.00000491, 0.00000425, 0.00000353, 0.00000278, 0.00000203, 0.00000130, 0.00000061, -0.00000002, -0.00000058, -0.00000104, -0.00000142, -0.00000169, -0.00000186, -0.00000193, -0.00000191, -0.00000182, -0.00000166, -0.00000145, -0.00000121, -0.00000096, },
		{ -0.00000112, -0.00000132, -0.00000150, -0.00000165, -0.00000176, -0.00000183, -0.00000184, -0.00000180, -0.00000171, -0.00000156, -0.00000137, -0.00000115, -0.00000089, -0.00000063, -0.00000036, -0.00000010, 0.00000013, 0.00000034, 0.00000050, 0.00000062, 0.00000068, 0.00000068, 0.00000063, 0.00000052, 0.00000036, 0.00000015, -0.00000009, -0.00000036, -0.00000066, -0.00000096, -0.00000126, -0.00000156, -0.00000184, -0.00000209, -0.00000231, -0.00000249, -0.00000264, -0.00000274, -0.00000280, -0.00000281, -0.00000279, -0.00000273, -0.00000264, -0.00000252, -0.00000238, -0.00000222, -0.00000205, -0.00000188, -0.00000172, -0.00000156, -0.00000142, -0.00000130, -0.00000121, -0.00000115, -0.00000112, -0.00000113, -0.00000118, -0.00000127, -0.00000140, -0.00000156, -0.00000176, -0.00000198, -0.00000221, -0.00000246, -0.00000270, -0.00000293, -0.00000314, -0.00000331, -0.00000343, -0.00000348, -0.00000347, -0.00000337, -0.00000319, -0.00000291, -0.00000254, -0.00000207, -0.00000151, -0.00000087, -0.00000016, 0.00000060, 0.00000140, 0.00000221, 0.00000303, 0.00000381, 0.00000454, 0.00000520, 0.00000576, 0.00000621, 0.00000654, 0.00000672, 0.00000676, 0.00000665, 0.00000640, 0.00000600, 0.00000548, 0.00000485, 0.00000413, 0.00000334, 0.00000251, 0.00000166, 0.00000082, 0.00000001, -0.00000074, -0.00000141, -0.00000199, -0.00000246, -0.00000282, -0.00000306, -0.00000318, -0.00000320, -0.00000311, -0.00000294, -0.00000270, -0.00000241, -0.00000208, -0.00000174, -0.00000141, -0.00000110, -0.00000083, -0.00000061, -0.00000045, -0.00000035, -0.00000032, -0.00000035, -0.00000043, -0.00000056, -0.00000073, -0.00000092, },
		{ -0.00000118, -0.00000098, -0.00000077, -0.00000058, -0.00000041, -0.00000028, -0.00000019, -0.00000015, -0.00000015, -0.00000021, -0.00000031, -0.00000046, -0.00000065, -0.00000086, -0.00000110, -0.00000135, -0.00000160, -0.00000185, -0.00000208, -0.00000228, -0.00000246, -0.00000260, -0.00000271, -0.00000277, -0.00000279, -0.00000277, -0.00000271, -0.00000262, -0.00000249, -0.00000233, -0.00000215, -0.00000194, -0.00000173, -0.00000150, -0.00000127, -0.00000103, -0.00000081, -0.00000059, -0.00000039, -0.00000022, -0.00000007, 0.00000005, 0.00000013, 0.00000016, 0.00000015, 0.00000010, -0.00000001, -0.00000017, -0.00000038, -0.00000064, -0.00000094, -0.00000127, -0.00000163, -0.00000200, -0.00000238, -0.00000274, -0.00000308, -0.00000337, -0.00000360, -0.00000376, -0.00000383, -0.00000380, -0.00000367, -0.00000342, -0.00000307, -0.00000261, -0.00000205, -0.00000140, -0.00000068, 0.00000008, 0.00000088, 0.00000168, 0.00000246, 0.00000320, 0.00000386, 0.00000443, 0.00000488, 0.00000519, 0.00000537, 0.00000539, 0.00000527, 0.00000499, 0.00000458, 0.00000405, 0.00000341, 0.00000269, 0.00000192, 0.00000112, 0.00000032, -0.00000046, -0.00000118, -0.00000182, -0.00000237, -0.00000281, -0.00000313, -0.00000332, -0.00000340, -0.00000336, -0.00000322, -0.00000299, -0.00000269, -0.00000233, -0.00000195, -0.00000155, -0.00000117, -0.00000082, -0.00000051, -0.00000027, -0.00000009, 0.00000001, 0.00000004, 0.00000000, -0.00000010, -0.00000026, -0.00000046, -0.00000069, -0.00000092, -0.00000116, -0.00000138, -0.00000157, -0.00000172, -0.00000182, -0.00000187, -0.00000187, -0.00000181, -0.00000170, -0.00000156, -0.00000138, },
		{ -0.00000079, -0.00000095, -0.00000110, -0.00000125, -0.00000138, -0.00000149, -0.00000157, -0.00000162, -0.00000163, -0.00000161, -0.00000155, -0.00000146, -0.00000134, -0.00000119, -0.00000103, -0.00000085, -0.00000066, -0.00000048, -0.00000030, -0.00000014, 0.00000001, 0.00000014, 0.00000025, 0.00000034, 0.00000039, 0.00000042, 0.00000042, 0.00000038, 0.00000031, 0.00000021, 0.00000006, -0.00000012, -0.00000034, -0.00000060, -0.00000091, -0.00000125, -0.00000163, -0.00000204, -0.00000248, -0.00000293, -0.00000339, -0.00000384, -0.00000427, -0.00000466, -0.00000498, -0.00000524, -0.00000540, -0.00000545, -0.00000538, -0.00000518, -0.00000483, -0.00000433, -0.00000369, -0.00000291, -0.00000200, -0.00000098, 0.00000012, 0.00000129, 0.00000249, 0.00000369, 0.00000485, 0.00000595, 0.00000694, 0.00000781, 0.00000852, 0.00000904, 0.00000938, 0.00000950, 0.00000942, 0.00000915, 0.00000868, 0.00000803, 0.00000725, 0.00000634, 0.00000535, 0.00000431, 0.00000325, 0.00000222, 0.00000123, 0.00000033, -0.00000046, -0.00000113, -0.00000166, -0.00000204, -0.00000227, -0.00000236, -0.00000233, -0.00000217, -0.00000193, -0.00000161, -0.00000125, -0.00000087, -0.00000050, -0.00000015, 0.00000016, 0.00000040, 0.00000057, 0.00000066, 0.00000068, 0.00000061, 0.00000048, 0.00000029, 0.00000005, -0.00000021, -0.00000049, -0.00000077, -0.00000103, -0.00000126, -0.00000145, -0.00000159, -0.00000167, -0.00000170, -0.00000168, -0.00000161, -0.00000149, -0.00000135, -0.00000119, -0.00000101, -0.00000084, -0.00000069, -0.00000056, -0.00000046, -0.00000039, -0.00000037, -0.00000039, -0.00000044, -0.00000053, -0.00000065, },
		{ 0.00000053, 0.00000077, 0.00000100, 0.00000120, 0.00000138, 0.00000153, 0.00000163, 0.00000168, 0.00000169, 0.00000165, 0.00000156, 0.00000143, 0.00000126, 0.00000106, 0.00000082, 0.00000056, 0.00000029, 0.00000000, -0.00000030, -0.00000060, -0.00000091, -0.00000122, -0.00000152, -0.00000182, -0.00000212, -0.00000240, -0.00000266, -0.00000291, -0.00000312, -0.00000330, -0.00000343, -0.00000349, -0.00000349, -0.00000339, -0.00000320, -0.00000291, -0.00000249, -0.00000196, -0.00000130, -0.00000052, 0.00000037, 0.00000136, 0.00000244, 0.00000358, 0.00000475, 0.00000593, 0.00000708, 0.00000817, 0.00000916, 0.00001002, 0.00001071, 0.00001122, 0.00001151, 0.00001157, 0.00001139, 0.00001098, 0.00001033, 0.00000947, 0.00000841, 0.00000718, 0.00000583, 0.00000439, 0.00000291, 0.00000142, -0.00000003, -0.00000139, -0.00000263, -0.00000372, -0.00000462, -0.00000533, -0.00000583, -0.00000611, -0.00000619, -0.00000607, -0.00000578, -0.00000535, -0.00000479, -0.00000415, -0.00000346, -0.00000276, -0.00000207, -0.00000142, -0.00000084, -0.00000035, 0.00000004, 0.00000032, 0.00000048, 0.00000054, 0.00000050, 0.00000038, 0.00000019, -0.00000005, -0.00000032, -0.00000060, -0.00000086, -0.00000111, -0.00000131, -0.00000147, -0.00000157, -0.00000161, -0.00000160, -0.00000154, -0.00000144, -0.00000131, -0.00000116, -0.00000100, -0.00000084, -0.00000069, -0.00000057, -0.00000048, -0.00000042, -0.00000039, -0.00000039, -0.00000042, -0.00000048, -0.00000054, -0.00000061, -0.00000067, -0.00000072, -0.00000075, -0.00000074, -0.00000070, -0.00000062, -0.00000050, -0.00000035, -0.00000016, 0.00000005, 0.00000029, },
		{ 0.00000069, 0.00000021, -0.00000029, -0.00000080, -0.00000129, -0.00000177, -0.00000221, -0.00000262, -0.00000299, -0.00000331, -0.00000358, -0.00000380, -0.00000397, -0.00000409, -0.00000417, -0.00000421, -0.00000420, -0.00000415, -0.00000405, -0.00000390, -0.00000370, -0.00000344, -0.00000312, -0.00000273, -0.00000228, -0.00000175, -0.00000116, -0.00000050, 0.00000022, 0.00000099, 0.00000178, 0.00000259, 0.00000339, 0.00000416, 0.00000488, 0.00000550, 0.00000602, 0.00000640, 0.00000662, 0.00000668, 0.00000654, 0.00000622, 0.00000570, 0.00000500, 0.00000414, 0.00000313, 0.00000200, 0.00000080, -0.00000045, -0.00000170, -0.00000291, -0.00000404, -0.00000504, -0.00000589, -0.00000656, -0.00000702, -0.00000725, -0.00000726, -0.00000705, -0.00000662, -0.00000601, -0.00000522, -0.00000431, -0.00000330, -0.00000224, -0.00000116, -0.00000011, 0.00000087, 0.00000176, 0.00000251, 0.00000312, 0.00000356, 0.00000383, 0.00000393, 0.00000386, 0.00000365, 0.00000331, 0.00000286, 0.00000233, 0.00000176, 0.00000116, 0.00000057, 0.00000001, -0.00000050, -0.00000095, -0.00000131, -0.00000159, -0.00000178, -0.00000190, -0.00000194, -0.00000191, -0.00000184, -0.00000173, -0.00000160, -0.00000147, -0.00000134, -0.00000123, -0.00000114, -0.00000108, -0.00000104, -0.00000102, -0.00000102, -0.00000103, -0.00000104, -0.00000104, -0.00000101, -0.00000095, -0.00000086, -0.00000071, -0.00000053, -0.00000029, -0.00000002, 0.00000029, 0.00000063, 0.00000097, 0.00000132, 0.00000164, 0.00000194, 0.00000219, 0.00000238, 0.00000250, 0.00000254, 0.00000250, 0.00000238, 0.00000217, 0.00000189, 0.00000155, 0.00000114, },
		{ -0.00000610, -0.00000599, -0.00000580, -0.00000556, -0.00000526, -0.00000492, -0.00000456, -0.00000416, -0.00000375, -0.00000331, -0.00000286, -0.00000239, -0.00000190, -0.00000139, -0.00000085, -0.00000029, 0.00000028, 0.00000087, 0.00000147, 0.00000207, 0.00000265, 0.00000321, 0.00000372, 0.00000417, 0.00000454, 0.00000482, 0.00000499, 0.00000505, 0.00000497, 0.00000477, 0.00000445, 0.00000400, 0.00000345, 0.00000282, 0.00000212, 0.00000139, 0.00000064, -0.00000007, -0.00000074, -0.00000132, -0.00000180, -0.00000214, -0.00000234, -0.00000239, -0.00000227, -0.00000201, -0.00000160, -0.00000107, -0.00000045, 0.00000024, 0.00000096, 0.00000167, 0.00000235, 0.00000295, 0.00000345, 0.00000382, 0.00000404, 0.00000409, 0.00000397, 0.00000368, 0.00000323, 0.00000264, 0.00000192, 0.00000111, 0.00000023, -0.00000067, -0.00000158, -0.00000244, -0.00000323, -0.00000393, -0.00000452, -0.00000498, -0.00000530, -0.00000547, -0.00000552, -0.00000543, -0.00000524, -0.00000496, -0.00000460, -0.00000420, -0.00000378, -0.00000336, -0.00000296, -0.00000259, -0.00000226, -0.00000199, -0.00000177, -0.00000160, -0.00000147, -0.00000137, -0.00000129, -0.00000121, -0.00000112, -0.00000101, -0.00000086, -0.00000066, -0.00000041, -0.00000012, 0.00000023, 0.00000062, 0.00000104, 0.00000147, 0.00000189, 0.00000229, 0.00000264, 0.00000293, 0.00000313, 0.00000324, 0.00000323, 0.00000311, 0.00000286, 0.00000250, 0.00000202, 0.00000145, 0.00000079, 0.00000008, -0.00000068, -0.00000146, -0.00000224, -0.00000298, -0.00000368, -0.00000431, -0.00000486, -0.00000531, -0.00000567, -0.00000592, -0.00000608, -0.00000613, },
		{ 0.00000076, 0.00000119, 0.00000160, 0.00000198, 0.00000235, 0.00000271, 0.00000306, 0.00000340, 0.00000373, 0.00000404, 0.00000434, 0.00000460, 0.00000482, 0.00000499, 0.00000510, 0.00000513, 0.00000508, 0.00000495, 0.00000473, 0.00000442, 0.00000403, 0.00000357, 0.00000306, 0.00000251, 0.00000193, 0.00000136, 0.00000080, 0.00000029, -0.00000016, -0.00000054, -0.00000084, -0.00000104, -0.00000115, -0.00000117, -0.00000111, -0.00000098, -0.00000079, -0.00000058, -0.00000036, -0.00000016, 0.00000002, 0.00000013, 0.00000017, 0.00000013, -0.00000000, -0.00000023, -0.00000054, -0.00000093, -0.00000138, -0.00000187, -0.00000238, -0.00000287, -0.00000334, -0.00000374, -0.00000405, -0.00000426, -0.00000435, -0.00000431, -0.00000413, -0.00000383, -0.00000340, -0.00000286, -0.00000224, -0.00000157, -0.00000085, -0.00000014, 0.00000055, 0.00000118, 0.00000174, 0.00000220, 0.00000255, 0.00000279, 0.00000290, 0.00000291, 0.00000281, 0.00000263, 0.00000238, 0.00000211, 0.00000182, 0.00000154, 0.00000131, 0.00000113, 0.00000104, 0.00000104, 0.00000113, 0.00000131, 0.00000159, 0.00000193, 0.00000233, 0.00000276, 0.00000318, 0.00000358, 0.00000393, 0.00000418, 0.00000433, 0.00000435, 0.00000423, 0.00000395, 0.00000352, 0.00000295, 0.00000225, 0.00000143, 0.00000052, -0.00000044, -0.00000144, -0.00000242, -0.00000337, -0.00000425, -0.00000503, -0.00000569, -0.00000621, -0.00000658, -0.00000680, -0.00000685, -0.00000676, -0.00000653, -0.00000617, -0.00000571, -0.00000517, -0.00000457, -0.00000393, -0.00000327, -0.00000261, -0.00000196, -0.00000134, -0.00000076, -0.00000022, 0.00000029, },
		{ 0.00000349, 0.00000358, 0.00000366, 0.00000371, 0.00000373, 0.00000371, 0.00000363, 0.00000349, 0.00000329, 0.00000304, 0.00000272, 0.00000237, 0.00000197, 0.00000155, 0.00000113, 0.00000072, 0.00000033, -0.00000002, -0.00000031, -0.00000055, -0.00000071, -0.00000081, -0.00000083, -0.00000079, -0.00000070, -0.00000056, -0.00000038, -0.00000019, 0.00000001, 0.00000020, 0.00000037, 0.00000052, 0.00000063, 0.00000072, 0.00000077, 0.00000079, 0.00000080, 0.00000081, 0.00000081, 0.00000084, 0.00000090, 0.00000100, 0.00000115, 0.00000135, 0.00000161, 0.00000191, 0.00000226, 0.00000263, 0.00000301, 0.00000339, 0.00000373, 0.00000401, 0.00000422, 0.00000433, 0.00000432, 0.00000419, 0.00000392, 0.00000351, 0.00000298, 0.00000234, 0.00000160, 0.00000079, -0.00000005, -0.00000090, -0.00000171, -0.00000245, -0.00000309, -0.00000358, -0.00000392, -0.00000408, -0.00000404, -0.00000382, -0.00000341, -0.00000283, -0.00000210, -0.00000127, -0.00000036, 0.00000057, 0.00000150, 0.00000236, 0.00000312, 0.00000374, 0.00000419, 0.00000443, 0.00000447, 0.00000428, 0.00000387, 0.00000326, 0.00000247, 0.00000153, 0.00000048, -0.00000065, -0.00000180, -0.00000293, -0.00000400, -0.00000496, -0.00000578, -0.00000644, -0.00000690, -0.00000716, -0.00000721, -0.00000706, -0.00000672, -0.00000622, -0.00000557, -0.00000481, -0.00000397, -0.00000309, -0.00000220, -0.00000133, -0.00000051, 0.00000025, 0.00000092, 0.00000150, 0.00000197, 0.00000235, 0.00000264, 0.00000284, 0.00000297, 0.00000306, 0.00000310, 0.00000313, 0.00000314, 0.00000317, 0.00000320, 0.00000325, 0.00000332, 0.00000340, },
		{ 0.00000226, 0.00000199, 0.00000169, 0.00000136, 0.00000101, 0.00000067, 0.00000035, 0.00000006, -0.00000018, -0.00000036, -0.00000049, -0.00000055, -0.00000055, -0.00000050, -0.00000039, -0.00000025, -0.00000008, 0.00000011, 0.00000030, 0.00000049, 0.00000067, 0.00000083, 0.00000096, 0.00000107, 0.00000115, 0.00000120, 0.00000124, 0.00000126, 0.00000126, 0.00000125, 0.00000123, 0.00000121, 0.00000117, 0.00000112, 0.00000105, 0.00000096, 0.00000082, 0.00000063, 0.00000039, 0.00000008, -0.00000030, -0.00000076, -0.00000129, -0.00000190, -0.00000257, -0.00000328, -0.00000401, -0.00000475, -0.00000545, -0.00000609, -0.00000664, -0.00000706, -0.00000731, -0.00000738, -0.00000724, -0.00000688, -0.00000629, -0.00000547, -0.00000444, -0.00000321, -0.00000182, -0.00000032, 0.00000126, 0.00000286, 0.00000443, 0.00000590, 0.00000722, 0.00000834, 0.00000922, 0.00000981, 0.00001010, 0.00001005, 0.00000969, 0.00000900, 0.00000802, 0.00000677, 0.00000531, 0.00000369, 0.00000195, 0.00000018, -0.00000157, -0.00000325, -0.00000478, -0.00000613, -0.00000725, -0.00000810, -0.00000868, -0.00000897, -0.00000897, -0.00000870, -0.00000819, -0.00000747, -0.00000659, -0.00000558, -0.00000449, -0.00000338, -0.00000229, -0.00000126, -0.00000031, 0.00000050, 0.00000118, 0.00000171, 0.00000208, 0.00000231, 0.00000240, 0.00000238, 0.00000227, 0.00000209, 0.00000188, 0.00000165, 0.00000144, 0.00000126, 0.00000113, 0.00000107, 0.00000107, 0.00000113, 0.00000126, 0.00000143, 0.00000164, 0.00000187, 0.00000210, 0.00000231, 0.00000249, 0.00000261, 0.00000268, 0.00000268, 0.00000261, 0.00000246, },
		{ -0.00000076, -0.00000082, -0.00000080, -0.00000072, -0.00000059, -0.00000041, -0.00000020, 0.00000003, 0.00000026, 0.00000048, 0.00000068, 0.00000085, 0.00000099, 0.00000109, 0.00000116, 0.00000119, 0.00000118, 0.00000113, 0.00000106, 0.00000096, 0.00000083, 0.00000067, 0.00000049, 0.00000028, 0.00000005, -0.00000021, -0.00000050, -0.00000082, -0.00000116, -0.00000152, -0.00000190, -0.00000228, -0.00000265, -0.00000299, -0.00000330, -0.00000355, -0.00000371, -0.00000377, -0.00000370, -0.00000349, -0.00000313, -0.00000260, -0.00000191, -0.00000105, -0.00000005, 0.00000109, 0.00000233, 0.00000364, 0.00000497, 0.00000628, 0.00000752, 0.00000864, 0.00000959, 0.00001031, 0.00001077, 0.00001093, 0.00001076, 0.00001026, 0.00000941, 0.00000824, 0.00000676, 0.00000501, 0.00000303, 0.00000089, -0.00000135, -0.00000363, -0.00000587, -0.00000799, -0.00000994, -0.00001165, -0.00001306, -0.00001413, -0.00001485, -0.00001518, -0.00001513, -0.00001471, -0.00001395, -0.00001288, -0.00001156, -0.00001004, -0.00000838, -0.00000665, -0.00000491, -0.00000322, -0.00000163, -0.00000020, 0.00000104, 0.00000207, 0.00000286, 0.00000342, 0.00000375, 0.00000387, 0.00000381, 0.00000360, 0.00000327, 0.00000286, 0.00000243, 0.00000199, 0.00000160, 0.00000128, 0.00000104, 0.00000091, 0.00000089, 0.00000098, 0.00000117, 0.00000145, 0.00000179, 0.00000217, 0.00000256, 0.00000295, 0.00000330, 0.00000358, 0.00000379, 0.00000390, 0.00000392, 0.00000382, 0.00000363, 0.00000335, 0.00000299, 0.00000256, 0.00000210, 0.00000161, 0.00000113, 0.00000067, 0.00000025, -0.00000011, -0.00000040, -0.00000062, },
		{ 0.00000153, 0.00000172, 0.00000185, 0.00000191, 0.00000192, 0.00000187, 0.00000177, 0.00000163, 0.00000144, 0.00000123, 0.00000099, 0.00000073, 0.00000045, 0.00000016, -0.00000013, -0.00000043, -0.00000073, -0.00000103, -0.00000132, -0.00000158, -0.00000183, -0.00000204, -0.00000220, -0.00000231, -0.00000235, -0.00000232, -0.00000219, -0.00000198, -0.00000166, -0.00000125, -0.00000074, -0.00000014, 0.00000053, 0.00000125, 0.00000200, 0.00000275, 0.00000347, 0.00000412, 0.00000467, 0.00000508, 0.00000533, 0.00000537, 0.00000520, 0.00000480, 0.00000417, 0.00000330, 0.00000222, 0.00000094, -0.00000049, -0.00000204, -0.00000365, -0.00000527, -0.00000684, -0.00000830, -0.00000958, -0.00001065, -0.00001145, -0.00001194, -0.00001210, -0.00001191, -0.00001137, -0.00001049, -0.00000930, -0.00000783, -0.00000614, -0.00000428, -0.00000231, -0.00000030, 0.00000168, 0.00000357, 0.00000531, 0.00000683, 0.00000811, 0.00000910, 0.00000979, 0.00001017, 0.00001024, 0.00001003, 0.00000956, 0.00000887, 0.00000801, 0.00000702, 0.00000598, 0.00000492, 0.00000389, 0.00000295, 0.00000213, 0.00000147, 0.00000098, 0.00000067, 0.00000054, 0.00000058, 0.00000078, 0.00000110, 0.00000152, 0.00000200, 0.00000250, 0.00000299, 0.00000343, 0.00000380, 0.00000406, 0.00000420, 0.00000421, 0.00000409, 0.00000384, 0.00000347, 0.00000300, 0.00000246, 0.00000187, 0.00000126, 0.00000065, 0.00000008, -0.00000044, -0.00000088, -0.00000123, -0.00000147, -0.00000160, -0.00000163, -0.00000155, -0.00000138, -0.00000114, -0.00000083, -0.00000048, -0.00000010, 0.00000028, 0.00000064, 0.00000098, 0.00000128, },
		{ 0.00000080, 0.00000051, 0.00000024, -0.00000001, -0.00000024, -0.00000044, -0.00000062, -0.00000077, -0.00000088, -0.00000097, -0.00000102, -0.00000103, -0.00000099, -0.00000091, -0.00000077, -0.00000058, -0.00000033, -0.00000001, 0.00000036, 0.00000078, 0.00000124, 0.00000174, 0.00000226, 0.00000277, 0.00000326, 0.00000371, 0.00000410, 0.00000440, 0.00000460, 0.00000467, 0.00000461, 0.00000441, 0.00000407, 0.00000359, 0.00000298, 0.00000227, 0.00000147, 0.00000062, -0.00000025, -0.00000111, -0.00000191, -0.00000262, -0.00000320, -0.00000362, -0.00000387, -0.00000392, -0.00000377, -0.00000342, -0.00000288, -0.00000217, -0.00000132, -0.00000036, 0.00000066, 0.00000169, 0.00000269, 0.00000361, 0.00000440, 0.00000502, 0.00000545, 0.00000565, 0.00000561, 0.00000532, 0.00000481, 0.00000407, 0.00000315, 0.00000207, 0.00000088, -0.00000037, -0.00000162, -0.00000283, -0.00000394, -0.00000491, -0.00000570, -0.00000627, -0.00000662, -0.00000672, -0.00000657, -0.00000620, -0.00000561, -0.00000485, -0.00000394, -0.00000293, -0.00000188, -0.00000081, 0.00000021, 0.00000114, 0.00000196, 0.00000262, 0.00000311, 0.00000341, 0.00000352, 0.00000345, 0.00000321, 0.00000282, 0.00000232, 0.00000172, 0.00000108, 0.00000042, -0.00000022, -0.00000080, -0.00000131, -0.00000171, -0.00000199, -0.00000213, -0.00000215, -0.00000204, -0.00000180, -0.00000147, -0.00000105, -0.00000057, -0.00000006, 0.00000046, 0.00000097, 0.00000145, 0.00000188, 0.00000223, 0.00000251, 0.00000271, 0.00000282, 0.00000284, 0.00000279, 0.00000267, 0.00000249, 0.00000226, 0.00000199, 0.00000170, 0.00000140, 0.00000110, },
		{ -0.00000074, -0.00000074, -0.00000072, -0.00000070, -0.00000065, -0.00000059, -0.00000051, -0.00000040, -0.00000026, -0.00000010, 0.00000007, 0.00000026, 0.00000044, 0.00000062, 0.00000077, 0.00000088, 0.00000094, 0.00000093, 0.00000084, 0.00000067, 0.00000041, 0.00000005, -0.00000040, -0.00000093, -0.00000153, -0.00000219, -0.00000290, -0.00000361, -0.00000433, -0.00000502, -0.00000566, -0.00000623, -0.00000671, -0.00000710, -0.00000737, -0.00000754, -0.00000760, -0.00000756, -0.00000743, -0.00000723, -0.00000698, -0.00000670, -0.00000642, -0.00000615, -0.00000593, -0.00000575, -0.00000564, -0.00000560, -0.00000563, -0.00000573, -0.00000589, -0.00000608, -0.00000628, -0.00000648, -0.00000663, -0.00000672, -0.00000671, -0.00000658, -0.00000630, -0.00000588, -0.00000528, -0.00000452, -0.00000361, -0.00000255, -0.00000138, -0.00000012, 0.00000119, 0.00000251, 0.00000379, 0.00000501, 0.00000610, 0.00000704, 0.00000779, 0.00000833, 0.00000864, 0.00000871, 0.00000854, 0.00000814, 0.00000754, 0.00000676, 0.00000583, 0.00000479, 0.00000369, 0.00000257, 0.00000147, 0.00000044, -0.00000049, -0.00000129, -0.00000192, -0.00000239, -0.00000267, -0.00000277, -0.00000270, -0.00000247, -0.00000211, -0.00000164, -0.00000108, -0.00000048, 0.00000013, 0.00000074, 0.00000130, 0.00000180, 0.00000221, 0.00000253, 0.00000274, 0.00000285, 0.00000286, 0.00000277, 0.00000260, 0.00000237, 0.00000208, 0.00000176, 0.00000142, 0.00000109, 0.00000076, 0.00000047, 0.00000020, -0.00000003, -0.00000022, -0.00000038, -0.00000050, -0.00000058, -0.00000064, -0.00000068, -0.00000071, -0.00000073, -0.00000074, -0.00000074, },
		{ -0.00000367, -0.00000366, -0.00000361, -0.00000354, -0.00000345, -0.00000336, -0.00000329, -0.00000324, -0.00000323, -0.00000327, -0.00000337, -0.00000353, -0.00000375, -0.00000403, -0.00000437, -0.00000474, -0.00000515, -0.00000557, -0.00000599, -0.00000639, -0.00000675, -0.00000705, -0.00000730, -0.00000746, -0.00000753, -0.00000751, -0.00000740, -0.00000719, -0.00000690, -0.00000652, -0.00000607, -0.00000556, -0.00000500, -0.00000440, -0.00000378, -0.00000313, -0.00000248, -0.00000182, -0.00000115, -0.00000049, 0.00000017, 0.00000083, 0.00000149, 0.00000215, 0.00000282, 0.00000349, 0.00000415, 0.00000481, 0.00000546, 0.00000608, 0.00000666, 0.00000718, 0.00000763, 0.00000798, 0.00000822, 0.00000833, 0.00000829, 0.00000809, 0.00000772, 0.00000718, 0.00000646, 0.00000558, 0.00000456, 0.00000340, 0.00000215, 0.00000082, -0.00000054, -0.00000189, -0.00000320, -0.00000443, -0.00000553, -0.00000649, -0.00000726, -0.00000782, -0.00000817, -0.00000828, -0.00000818, -0.00000785, -0.00000733, -0.00000663, -0.00000579, -0.00000484, -0.00000382, -0.00000276, -0.00000171, -0.00000070, 0.00000023, 0.00000106, 0.00000176, 0.00000233, 0.00000274, 0.00000300, 0.00000311, 0.00000308, 0.00000294, 0.00000269, 0.00000237, 0.00000199, 0.00000158, 0.00000117, 0.00000077, 0.00000040, 0.00000008, -0.00000019, -0.00000040, -0.00000056, -0.00000066, -0.00000072, -0.00000074, -0.00000075, -0.00000075, -0.00000076, -0.00000078, -0.00000084, -0.00000093, -0.00000106, -0.00000123, -0.00000144, -0.00000168, -0.00000194, -0.00000222, -0.00000251, -0.00000278, -0.00000303, -0.00000324, -0.00000342, -0.00000355, -0.00000363, },
		{ -0.00000169, -0.00000159, -0.00000157, -0.00000164, -0.00000180, -0.00000203, -0.00000231, -0.00000264, -0.00000300, -0.00000335, -0.00000369, -0.00000398, -0.00000423, -0.00000440, -0.00000448, -0.00000448, -0.00000437, -0.00000417, -0.00000386, -0.00000347, -0.00000299, -0.00000243, -0.00000183, -0.00000117, -0.00000050, 0.00000019, 0.00000087, 0.00000153, 0.00000214, 0.00000271, 0.00000321, 0.00000364, 0.00000398, 0.00000423, 0.00000439, 0.00000444, 0.00000439, 0.00000423, 0.00000397, 0.00000361, 0.00000315, 0.00000261, 0.00000197, 0.00000127, 0.00000050, -0.00000031, -0.00000115, -0.00000200, -0.00000284, -0.00000365, -0.00000441, -0.00000509, -0.00000567, -0.00000612, -0.00000644, -0.00000660, -0.00000659, -0.00000640, -0.00000602, -0.00000547, -0.00000475, -0.00000386, -0.00000284, -0.00000171, -0.00000049, 0.00000079, 0.00000208, 0.00000335, 0.00000456, 0.00000569, 0.00000669, 0.00000754, 0.00000821, 0.00000870, 0.00000898, 0.00000906, 0.00000894, 0.00000863, 0.00000816, 0.00000754, 0.00000680, 0.00000598, 0.00000511, 0.00000422, 0.00000335, 0.00000252, 0.00000176, 0.00000110, 0.00000054, 0.00000010, -0.00000022, -0.00000043, -0.00000053, -0.00000055, -0.00000049, -0.00000038, -0.00000025, -0.00000012, -0.00000002, 0.00000004, 0.00000004, -0.00000003, -0.00000019, -0.00000044, -0.00000077, -0.00000117, -0.00000164, -0.00000215, -0.00000268, -0.00000322, -0.00000374, -0.00000422, -0.00000463, -0.00000497, -0.00000521, -0.00000535, -0.00000539, -0.00000532, -0.00000514, -0.00000489, -0.00000455, -0.00000417, -0.00000375, -0.00000331, -0.00000289, -0.00000250, -0.00000216, -0.00000189, },
		{ -0.00000364, -0.00000408, -0.00000443, -0.00000466, -0.00000477, -0.00000474, -0.00000458, -0.00000428, -0.00000386, -0.00000332, -0.00000268, -0.00000196, -0.00000118, -0.00000037, 0.00000045, 0.00000126, 0.00000202, 0.00000273, 0.00000335, 0.00000388, 0.00000430, 0.00000460, 0.00000478, 0.00000484, 0.00000478, 0.00000461, 0.00000434, 0.00000398, 0.00000357, 0.00000310, 0.00000261, 0.00000211, 0.00000163, 0.00000118, 0.00000080, 0.00000049, 0.00000027, 0.00000016, 0.00000016, 0.00000029, 0.00000053, 0.00000089, 0.00000137, 0.00000194, 0.00000260, 0.00000333, 0.00000409, 0.00000488, 0.00000566, 0.00000641, 0.00000709, 0.00000768, 0.00000815, 0.00000849, 0.00000866, 0.00000867, 0.00000849, 0.00000812, 0.00000757, 0.00000685, 0.00000597, 0.00000495, 0.00000381, 0.00000259, 0.00000132, 0.00000004, -0.00000122, -0.00000242, -0.00000352, -0.00000449, -0.00000530, -0.00000594, -0.00000637, -0.00000661, -0.00000663, -0.00000646, -0.00000610, -0.00000557, -0.00000491, -0.00000414, -0.00000330, -0.00000242, -0.00000155, -0.00000073, 0.00000001, 0.00000064, 0.00000114, 0.00000147, 0.00000163, 0.00000161, 0.00000141, 0.00000105, 0.00000053, -0.00000011, -0.00000085, -0.00000165, -0.00000248, -0.00000331, -0.00000410, -0.00000481, -0.00000542, -0.00000590, -0.00000623, -0.00000641, -0.00000642, -0.00000627, -0.00000598, -0.00000555, -0.00000500, -0.00000437, -0.00000368, -0.00000297, -0.00000227, -0.00000160, -0.00000100, -0.00000049, -0.00000009, 0.00000018, 0.00000032, 0.00000032, 0.00000018, -0.00000008, -0.00000046, -0.00000092, -0.00000144, -0.00000200, -0.00000258, -0.00000313, },
		{ -0.00000086, 0.00000016, 0.00000119, 0.00000220, 0.00000315, 0.00000403, 0.00000480, 0.00000544, 0.00000594, 0.00000629, 0.00000648, 0.00000652, 0.00000640, 0.00000615, 0.00000578, 0.00000530, 0.00000475, 0.00000415, 0.00000352, 0.00000289, 0.00000228, 0.00000172, 0.00000123, 0.00000082, 0.00000050, 0.00000028, 0.00000016, 0.00000013, 0.00000019, 0.00000032, 0.00000051, 0.00000074, 0.00000097, 0.00000120, 0.00000139, 0.00000153, 0.00000158, 0.00000154, 0.00000138, 0.00000111, 0.00000072, 0.00000020, -0.00000044, -0.00000118, -0.00000200, -0.00000288, -0.00000379, -0.00000470, -0.00000559, -0.00000641, -0.00000713, -0.00000771, -0.00000814, -0.00000838, -0.00000842, -0.00000823, -0.00000782, -0.00000718, -0.00000632, -0.00000527, -0.00000403, -0.00000264, -0.00000114, 0.00000043, 0.00000203, 0.00000361, 0.00000512, 0.00000652, 0.00000776, 0.00000880, 0.00000962, 0.00001017, 0.00001045, 0.00001044, 0.00001013, 0.00000955, 0.00000870, 0.00000760, 0.00000629, 0.00000482, 0.00000321, 0.00000153, -0.00000019, -0.00000188, -0.00000350, -0.00000501, -0.00000636, -0.00000752, -0.00000846, -0.00000916, -0.00000961, -0.00000980, -0.00000975, -0.00000947, -0.00000897, -0.00000830, -0.00000748, -0.00000656, -0.00000556, -0.00000455, -0.00000355, -0.00000261, -0.00000176, -0.00000104, -0.00000046, -0.00000004, 0.00000020, 0.00000027, 0.00000017, -0.00000009, -0.00000049, -0.00000100, -0.00000159, -0.00000224, -0.00000290, -0.00000355, -0.00000414, -0.00000465, -0.00000505, -0.00000531, -0.00000543, -0.00000538, -0.00000516, -0.00000478, -0.00000424, -0.00000356, -0.00000275, -0.00000184, },
		{ 0.00000767, 0.00000722, 0.00000664, 0.00000596, 0.00000519, 0.00000438, 0.00000354, 0.00000272, 0.00000193, 0.00000121, 0.00000057, 0.00000003, -0.00000040, -0.00000071, -0.00000091, -0.00000100, -0.00000099, -0.00000090, -0.00000075, -0.00000056, -0.00000035, -0.00000014, 0.00000004, 0.00000017, 0.00000025, 0.00000026, 0.00000019, 0.00000005, -0.00000017, -0.00000044, -0.00000076, -0.00000109, -0.00000143, -0.00000173, -0.00000199, -0.00000216, -0.00000222, -0.00000216, -0.00000195, -0.00000158, -0.00000104, -0.00000034, 0.00000051, 0.00000150, 0.00000261, 0.00000381, 0.00000507, 0.00000634, 0.00000759, 0.00000876, 0.00000982, 0.00001073, 0.00001144, 0.00001192, 0.00001215, 0.00001209, 0.00001174, 0.00001109, 0.00001014, 0.00000891, 0.00000742, 0.00000569, 0.00000377, 0.00000170, -0.00000048, -0.00000271, -0.00000493, -0.00000711, -0.00000918, -0.00001109, -0.00001280, -0.00001427, -0.00001547, -0.00001637, -0.00001695, -0.00001722, -0.00001716, -0.00001679, -0.00001612, -0.00001519, -0.00001403, -0.00001266, -0.00001114, -0.00000952, -0.00000782, -0.00000612, -0.00000445, -0.00000285, -0.00000137, -0.00000004, 0.00000111, 0.00000206, 0.00000279, 0.00000329, 0.00000357, 0.00000363, 0.00000349, 0.00000316, 0.00000268, 0.00000207, 0.00000137, 0.00000062, -0.00000015, -0.00000089, -0.00000157, -0.00000216, -0.00000263, -0.00000296, -0.00000312, -0.00000311, -0.00000292, -0.00000256, -0.00000202, -0.00000134, -0.00000053, 0.00000038, 0.00000136, 0.00000238, 0.00000341, 0.00000440, 0.00000533, 0.00000616, 0.00000686, 0.00000742, 0.00000782, 0.00000805, 0.00000809, 0.00000797, },
		{ -0.00000089, -0.00000119, -0.00000135, -0.00000141, -0.00000135, -0.00000121, -0.00000100, -0.00000075, -0.00000047, -0.00000019, 0.00000007, 0.00000029, 0.00000045, 0.00000055, 0.00000058, 0.00000054, 0.00000045, 0.00000030, 0.00000012, -0.00000008, -0.00000027, -0.00000043, -0.00000054, -0.00000059, -0.00000055, -0.00000041, -0.00000017, 0.00000018, 0.00000063, 0.00000117, 0.00000177, 0.00000243, 0.00000311, 0.00000378, 0.00000440, 0.00000495, 0.00000539, 0.00000568, 0.00000582, 0.00000576, 0.00000551, 0.00000505, 0.00000438, 0.00000352, 0.00000248, 0.00000128, -0.00000004, -0.00000144, -0.00000289, -0.00000434, -0.00000576, -0.00000708, -0.00000828, -0.00000931, -0.00001014, -0.00001075, -0.00001111, -0.00001120, -0.00001104, -0.00001061, -0.00000993, -0.00000902, -0.00000790, -0.00000660, -0.00000515, -0.00000359, -0.00000196, -0.00000030, 0.00000136, 0.00000297, 0.00000450, 0.00000592, 0.00000719, 0.00000830, 0.00000922, 0.00000994, 0.00001045, 0.00001075, 0.00001083, 0.00001070, 0.00001038, 0.00000987, 0.00000919, 0.00000837, 0.00000743, 0.00000639, 0.00000530, 0.00000416, 0.00000303, 0.00000192, 0.00000087, -0.00000010, -0.00000096, -0.00000168, -0.00000224, -0.00000264, -0.00000286, -0.00000290, -0.00000275, -0.00000242, -0.00000193, -0.00000128, -0.00000051, 0.00000036, 0.00000131, 0.00000229, 0.00000328, 0.00000424, 0.00000514, 0.00000594, 0.00000663, 0.00000717, 0.00000755, 0.00000776, 0.00000779, 0.00000765, 0.00000734, 0.00000687, 0.00000627, 0.00000557, 0.00000478, 0.00000394, 0.00000308, 0.00000224, 0.00000144, 0.00000070, 0.00000006, -0.00000048, },
		{ 0.00000058, 0.00000077, 0.00000090, 0.00000097, 0.00000098, 0.00000095, 0.00000089, 0.00000081, 0.00000074, 0.00000069, 0.00000068, 0.00000073, 0.00000085, 0.00000105, 0.00000134, 0.00000171, 0.00000216, 0.00000267, 0.00000323, 0.00000382, 0.00000441, 0.00000497, 0.00000549, 0.00000593, 0.00000627, 0.00000650, 0.00000660, 0.00000656, 0.00000638, 0.00000607, 0.00000564, 0.00000511, 0.00000450, 0.00000384, 0.00000317, 0.00000250, 0.00000188, 0.00000133, 0.00000088, 0.00000055, 0.00000035, 0.00000030, 0.00000038, 0.00000061, 0.00000096, 0.00000142, 0.00000195, 0.00000254, 0.00000315, 0.00000374, 0.00000429, 0.00000477, 0.00000514, 0.00000539, 0.00000549, 0.00000545, 0.00000524, 0.00000488, 0.00000437, 0.00000372, 0.00000296, 0.00000209, 0.00000116, 0.00000018, -0.00000083, -0.00000183, -0.00000281, -0.00000374, -0.00000461, -0.00000540, -0.00000611, -0.00000673, -0.00000725, -0.00000767, -0.00000800, -0.00000823, -0.00000838, -0.00000843, -0.00000841, -0.00000830, -0.00000811, -0.00000784, -0.00000749, -0.00000705, -0.00000653, -0.00000593, -0.00000524, -0.00000447, -0.00000362, -0.00000270, -0.00000172, -0.00000069, 0.00000036, 0.00000143, 0.00000248, 0.00000349, 0.00000444, 0.00000531, 0.00000606, 0.00000667, 0.00000714, 0.00000745, 0.00000758, 0.00000754, 0.00000733, 0.00000695, 0.00000644, 0.00000580, 0.00000507, 0.00000427, 0.00000343, 0.00000259, 0.00000177, 0.00000100, 0.00000032, -0.00000027, -0.00000074, -0.00000109, -0.00000132, -0.00000142, -0.00000140, -0.00000129, -0.00000109, -0.00000084, -0.00000054, -0.00000023, 0.00000007, 0.00000035, },
		{ 0.00000090, 0.00000104, 0.00000124, 0.00000150, 0.00000181, 0.00000216, 0.00000253, 0.00000292, 0.00000329, 0.00000363, 0.00000391, 0.00000412, 0.00000423, 0.00000424, 0.00000413, 0.00000391, 0.00000357, 0.00000312, 0.00000258, 0.00000196, 0.00000130, 0.00000060, -0.00000010, -0.00000078, -0.00000141, -0.00000199, -0.00000249, -0.00000291, -0.00000325, -0.00000352, -0.00000371, -0.00000386, -0.00000398, -0.00000409, -0.00000423, -0.00000441, -0.00000466, -0.00000501, -0.00000545, -0.00000601, -0.00000668, -0.00000746, -0.00000832, -0.00000925, -0.00001022, -0.00001119, -0.00001212, -0.00001297, -0.00001370, -0.00001428, -0.00001466, -0.00001482, -0.00001474, -0.00001441, -0.00001383, -0.00001300, -0.00001194, -0.00001069, -0.00000926, -0.00000772, -0.00000609, -0.00000444, -0.00000279, -0.00000121, 0.00000027, 0.00000161, 0.00000279, 0.00000378, 0.00000459, 0.00000521, 0.00000564, 0.00000591, 0.00000603, 0.00000605, 0.00000598, 0.00000586, 0.00000573, 0.00000560, 0.00000552, 0.00000549, 0.00000553, 0.00000564, 0.00000583, 0.00000607, 0.00000636, 0.00000667, 0.00000699, 0.00000727, 0.00000750, 0.00000765, 0.00000769, 0.00000761, 0.00000740, 0.00000706, 0.00000658, 0.00000597, 0.00000525, 0.00000444, 0.00000356, 0.00000266, 0.00000175, 0.00000087, 0.00000005, -0.00000068, -0.00000131, -0.00000181, -0.00000218, -0.00000241, -0.00000250, -0.00000247, -0.00000233, -0.00000210, -0.00000180, -0.00000145, -0.00000108, -0.00000071, -0.00000037, -0.00000006, 0.00000020, 0.00000041, 0.00000056, 0.00000065, 0.00000071, 0.00000073, 0.00000073, 0.00000074, 0.00000075, 0.00000080, },
		{ 0.00000057, 0.00000047, 0.00000028, 0.00000000, -0.00000035, -0.00000078, -0.00000126, -0.00000178, -0.00000232, -0.00000286, -0.00000337, -0.00000384, -0.00000425, -0.00000458, -0.00000483, -0.00000498, -0.00000506, -0.00000506, -0.00000500, -0.00000490, -0.00000478, -0.00000466, -0.00000457, -0.00000452, -0.00000454, -0.00000464, -0.00000481, -0.00000506, -0.00000538, -0.00000575, -0.00000615, -0.00000653, -0.00000687, -0.00000713, -0.00000726, -0.00000723, -0.00000700, -0.00000656, -0.00000587, -0.00000494, -0.00000377, -0.00000238, -0.00000079, 0.00000095, 0.00000279, 0.00000468, 0.00000654, 0.00000831, 0.00000992, 0.00001130, 0.00001241, 0.00001318, 0.00001359, 0.00001361, 0.00001324, 0.00001249, 0.00001138, 0.00000995, 0.00000826, 0.00000637, 0.00000436, 0.00000230, 0.00000027, -0.00000164, -0.00000338, -0.00000487, -0.00000607, -0.00000694, -0.00000747, -0.00000764, -0.00000746, -0.00000697, -0.00000621, -0.00000521, -0.00000405, -0.00000278, -0.00000148, -0.00000021, 0.00000098, 0.00000202, 0.00000288, 0.00000351, 0.00000391, 0.00000406, 0.00000398, 0.00000366, 0.00000316, 0.00000249, 0.00000171, 0.00000086, -0.00000002, -0.00000087, -0.00000166, -0.00000236, -0.00000293, -0.00000335, -0.00000362, -0.00000373, -0.00000370, -0.00000353, -0.00000325, -0.00000289, -0.00000247, -0.00000202, -0.00000157, -0.00000114, -0.00000077, -0.00000045, -0.00000020, -0.00000003, 0.00000007, 0.00000011, 0.00000009, 0.00000003, -0.00000004, -0.00000012, -0.00000019, -0.00000023, -0.00000025, -0.00000022, -0.00000015, -0.00000005, 0.00000008, 0.00000022, 0.00000036, 0.00000048, 0.00000057, 0.00000060, },
		{ -0.00000180, -0.00000184, -0.00000180, -0.00000167, -0.00000148, -0.00000122, -0.00000092, -0.00000060, -0.00000028, 0.00000002, 0.00000027, 0.00000044, 0.00000052, 0.00000050, 0.00000037, 0.00000014, -0.00000019, -0.00000059, -0.00000105, -0.00000153, -0.00000199, -0.00000240, -0.00000272, -0.00000291, -0.00000295, -0.00000282, -0.00000251, -0.00000201, -0.00000134, -0.00000052, 0.00000042, 0.00000143, 0.00000245, 0.00000343, 0.00000430, 0.00000501, 0.00000550, 0.00000572, 0.00000562, 0.00000519, 0.00000442, 0.00000330, 0.00000188, 0.00000018, -0.00000173, -0.00000378, -0.00000588, -0.00000796, -0.00000990, -0.00001163, -0.00001305, -0.00001408, -0.00001467, -0.00001476, -0.00001433, -0.00001337, -0.00001190, -0.00000996, -0.00000761, -0.00000491, -0.00000197, 0.00000112, 0.00000425, 0.00000730, 0.00001018, 0.00001278, 0.00001501, 0.00001681, 0.00001813, 0.00001892, 0.00001918, 0.00001892, 0.00001817, 0.00001697, 0.00001540, 0.00001353, 0.00001144, 0.00000922, 0.00000697, 0.00000475, 0.00000267, 0.00000077, -0.00000089, -0.00000227, -0.00000335, -0.00000412, -0.00000459, -0.00000477, -0.00000471, -0.00000444, -0.00000400, -0.00000346, -0.00000284, -0.00000221, -0.00000161, -0.00000106, -0.00000059, -0.00000022, 0.00000005, 0.00000021, 0.00000027, 0.00000025, 0.00000016, 0.00000004, -0.00000011, -0.00000026, -0.00000038, -0.00000048, -0.00000053, -0.00000053, -0.00000050, -0.00000042, -0.00000031, -0.00000019, -0.00000007, 0.00000003, 0.00000010, 0.00000012, 0.00000010, 0.00000001, -0.00000012, -0.00000031, -0.00000054, -0.00000079, -0.00000105, -0.00000130, -0.00000152, -0.00000169, },
		{ 0.00000184, 0.00000179, 0.00000162, 0.00000131, 0.00000088, 0.00000034, -0.00000027, -0.00000094, -0.00000161, -0.00000226, -0.00000284, -0.00000332, -0.00000366, -0.00000383, -0.00000382, -0.00000363, -0.00000325, -0.00000271, -0.00000203, -0.00000124, -0.00000040, 0.00000044, 0.00000124, 0.00000193, 0.00000248, 0.00000284, 0.00000298, 0.00000288, 0.00000254, 0.00000197, 0.00000120, 0.00000028, -0.00000075, -0.00000181, -0.00000283, -0.00000374, -0.00000444, -0.00000488, -0.00000499, -0.00000471, -0.00000402, -0.00000290, -0.00000136, 0.00000058, 0.00000285, 0.00000540, 0.00000814, 0.00001095, 0.00001374, 0.00001639, 0.00001877, 0.00002080, 0.00002235, 0.00002337, 0.00002377, 0.00002354, 0.00002264, 0.00002111, 0.00001896, 0.00001628, 0.00001315, 0.00000966, 0.00000593, 0.00000209, -0.00000174, -0.00000542, -0.00000885, -0.00001192, -0.00001455, -0.00001666, -0.00001821, -0.00001919, -0.00001959, -0.00001943, -0.00001878, -0.00001767, -0.00001620, -0.00001445, -0.00001250, -0.00001045, -0.00000839, -0.00000639, -0.00000453, -0.00000286, -0.00000142, -0.00000023, 0.00000068, 0.00000133, 0.00000174, 0.00000192, 0.00000192, 0.00000178, 0.00000154, 0.00000123, 0.00000091, 0.00000060, 0.00000033, 0.00000011, -0.00000003, -0.00000011, -0.00000013, -0.00000010, -0.00000002, 0.00000007, 0.00000016, 0.00000023, 0.00000028, 0.00000028, 0.00000023, 0.00000014, 0.00000001, -0.00000015, -0.00000033, -0.00000050, -0.00000066, -0.00000077, -0.00000083, -0.00000082, -0.00000073, -0.00000057, -0.00000034, -0.00000004, 0.00000029, 0.00000065, 0.00000101, 0.00000133, 0.00000159, 0.00000177, },
		{ -0.00000321, -0.00000364, -0.00000389, -0.00000395, -0.00000382, -0.00000349, -0.00000299, -0.00000235, -0.00000161, -0.00000081, -0.00000000, 0.00000076, 0.00000143, 0.00000196, 0.00000231, 0.00000247, 0.00000242, 0.00000216, 0.00000171, 0.00000110, 0.00000036, -0.00000043, -0.00000124, -0.00000199, -0.00000262, -0.00000308, -0.00000331, -0.00000328, -0.00000296, -0.00000234, -0.00000145, -0.00000030, 0.00000104, 0.00000253, 0.00000407, 0.00000560, 0.00000700, 0.00000820, 0.00000911, 0.00000965, 0.00000975, 0.00000938, 0.00000851, 0.00000714, 0.00000530, 0.00000304, 0.00000041, -0.00000247, -0.00000551, -0.00000860, -0.00001160, -0.00001440, -0.00001687, -0.00001893, -0.00002046, -0.00002141, -0.00002174, -0.00002141, -0.00002045, -0.00001888, -0.00001676, -0.00001418, -0.00001123, -0.00000802, -0.00000469, -0.00000134, 0.00000190, 0.00000492, 0.00000761, 0.00000991, 0.00001174, 0.00001307, 0.00001388, 0.00001418, 0.00001400, 0.00001337, 0.00001236, 0.00001104, 0.00000950, 0.00000781, 0.00000606, 0.00000433, 0.00000267, 0.00000115, -0.00000018, -0.00000131, -0.00000221, -0.00000290, -0.00000336, -0.00000364, -0.00000376, -0.00000376, -0.00000367, -0.00000352, -0.00000336, -0.00000320, -0.00000306, -0.00000296, -0.00000290, -0.00000288, -0.00000288, -0.00000289, -0.00000288, -0.00000285, -0.00000275, -0.00000259, -0.00000235, -0.00000201, -0.00000159, -0.00000110, -0.00000054, 0.00000005, 0.00000065, 0.00000123, 0.00000175, 0.00000217, 0.00000248, 0.00000264, 0.00000265, 0.00000249, 0.00000216, 0.00000168, 0.00000107, 0.00000036, -0.00000040, -0.00000119, -0.00000195, -0.00000264, },
		{ -0.00000002, 0.00000062, 0.00000114, 0.00000150, 0.00000168, 0.00000167, 0.00000146, 0.00000107, 0.00000054, -0.00000010, -0.00000080, -0.00000151, -0.00000215, -0.00000269, -0.00000307, -0.00000325, -0.00000320, -0.00000290, -0.00000236, -0.00000160, -0.00000063, 0.00000047, 0.00000167, 0.00000290, 0.00000407, 0.00000512, 0.00000599, 0.00000661, 0.00000693, 0.00000693, 0.00000658, 0.00000589, 0.00000489, 0.00000362, 0.00000214, 0.00000053, -0.00000114, -0.00000277, -0.00000426, -0.00000554, -0.00000652, -0.00000713, -0.00000733, -0.00000709, -0.00000641, -0.00000530, -0.00000381, -0.00000200, 0.00000006, 0.00000225, 0.00000449, 0.00000665, 0.00000864, 0.00001034, 0.00001166, 0.00001254, 0.00001291, 0.00001274, 0.00001203, 0.00001078, 0.00000906, 0.00000691, 0.00000441, 0.00000167, -0.00000121, -0.00000412, -0.00000695, -0.00000959, -0.00001197, -0.00001399, -0.00001559, -0.00001675, -0.00001743, -0.00001765, -0.00001743, -0.00001681, -0.00001583, -0.00001458, -0.00001313, -0.00001154, -0.00000990, -0.00000827, -0.00000672, -0.00000528, -0.00000400, -0.00000290, -0.00000197, -0.00000121, -0.00000061, -0.00000012, 0.00000027, 0.00000062, 0.00000096, 0.00000132, 0.00000174, 0.00000222, 0.00000279, 0.00000344, 0.00000415, 0.00000490, 0.00000567, 0.00000642, 0.00000709, 0.00000766, 0.00000808, 0.00000832, 0.00000835, 0.00000815, 0.00000772, 0.00000707, 0.00000622, 0.00000519, 0.00000404, 0.00000281, 0.00000156, 0.00000033, -0.00000080, -0.00000180, -0.00000261, -0.00000322, -0.00000359, -0.00000372, -0.00000362, -0.00000331, -0.00000282, -0.00000220, -0.00000148, -0.00000074, },
		{ -0.00000125, -0.00000180, -0.00000227, -0.00000259, -0.00000274, -0.00000268, -0.00000241, -0.00000191, -0.00000121, -0.00000034, 0.00000065, 0.00000172, 0.00000280, 0.00000382, 0.00000473, 0.00000545, 0.00000596, 0.00000620, 0.00000615, 0.00000581, 0.00000518, 0.00000430, 0.00000319, 0.00000192, 0.00000055, -0.00000085, -0.00000222, -0.00000348, -0.00000457, -0.00000544, -0.00000605, -0.00000639, -0.00000644, -0.00000623, -0.00000578, -0.00000514, -0.00000437, -0.00000354, -0.00000272, -0.00000199, -0.00000141, -0.00000105, -0.00000095, -0.00000115, -0.00000166, -0.00000248, -0.00000357, -0.00000489, -0.00000639, -0.00000797, -0.00000957, -0.00001107, -0.00001240, -0.00001346, -0.00001418, -0.00001449, -0.00001435, -0.00001373, -0.00001263, -0.00001107, -0.00000908, -0.00000674, -0.00000412, -0.00000131, 0.00000159, 0.00000447, 0.00000724, 0.00000979, 0.00001205, 0.00001394, 0.00001543, 0.00001648, 0.00001710, 0.00001728, 0.00001708, 0.00001654, 0.00001573, 0.00001472, 0.00001360, 0.00001245, 0.00001135, 0.00001035, 0.00000951, 0.00000888, 0.00000847, 0.00000829, 0.00000831, 0.00000851, 0.00000885, 0.00000925, 0.00000968, 0.00001005, 0.00001030, 0.00001038, 0.00001024, 0.00000984, 0.00000917, 0.00000823, 0.00000703, 0.00000559, 0.00000398, 0.00000224, 0.00000045, -0.00000132, -0.00000301, -0.00000454, -0.00000585, -0.00000690, -0.00000764, -0.00000805, -0.00000814, -0.00000792, -0.00000741, -0.00000667, -0.00000574, -0.00000469, -0.00000360, -0.00000251, -0.00000151, -0.00000063, 0.00000006, 0.00000055, 0.00000082, 0.00000086, 0.00000070, 0.00000036, -0.00000011, -0.00000067, },
		{ 0.00000152, 0.00000257, 0.00000357, 0.00000446, 0.00000519, 0.00000572, 0.00000599, 0.00000599, 0.00000572, 0.00000518, 0.00000440, 0.00000342, 0.00000229, 0.00000108, -0.00000017, -0.00000137, -0.00000248, -0.00000343, -0.00000418, -0.00000470, -0.00000498, -0.00000500, -0.00000479, -0.00000436, -0.00000376, -0.00000303, -0.00000222, -0.00000139, -0.00000058, 0.00000016, 0.00000078, 0.00000127, 0.00000161, 0.00000180, 0.00000186, 0.00000182, 0.00000171, 0.00000158, 0.00000149, 0.00000148, 0.00000160, 0.00000191, 0.00000242, 0.00000317, 0.00000414, 0.00000533, 0.00000671, 0.00000822, 0.00000980, 0.00001138, 0.00001288, 0.00001422, 0.00001530, 0.00001606, 0.00001642, 0.00001634, 0.00001579, 0.00001476, 0.00001327, 0.00001134, 0.00000904, 0.00000645, 0.00000366, 0.00000078, -0.00000207, -0.00000479, -0.00000727, -0.00000940, -0.00001111, -0.00001233, -0.00001302, -0.00001317, -0.00001280, -0.00001194, -0.00001065, -0.00000902, -0.00000715, -0.00000514, -0.00000310, -0.00000116, 0.00000058, 0.00000203, 0.00000311, 0.00000377, 0.00000397, 0.00000371, 0.00000301, 0.00000191, 0.00000047, -0.00000122, -0.00000307, -0.00000498, -0.00000684, -0.00000857, -0.00001005, -0.00001123, -0.00001205, -0.00001246, -0.00001246, -0.00001204, -0.00001125, -0.00001013, -0.00000875, -0.00000718, -0.00000551, -0.00000382, -0.00000221, -0.00000074, 0.00000051, 0.00000150, 0.00000220, 0.00000259, 0.00000269, 0.00000250, 0.00000208, 0.00000148, 0.00000077, 0.00000000, -0.00000074, -0.00000140, -0.00000192, -0.00000224, -0.00000234, -0.00000220, -0.00000182, -0.00000122, -0.00000043, 0.00000050, },
		{ 0.00000286, 0.00000162, 0.00000027, -0.00000110, -0.00000244, -0.00000369, -0.00000478, -0.00000567, -0.00000634, -0.00000675, -0.00000691, -0.00000683, -0.00000654, -0.00000607, -0.00000546, -0.00000476, -0.00000403, -0.00000331, -0.00000264, -0.00000206, -0.00000160, -0.00000128, -0.00000109, -0.00000104, -0.00000111, -0.00000128, -0.00000151, -0.00000178, -0.00000206, -0.00000232, -0.00000253, -0.00000270, -0.00000281, -0.00000288, -0.00000292, -0.00000296, -0.00000303, -0.00000318, -0.00000345, -0.00000388, -0.00000451, -0.00000536, -0.00000644, -0.00000776, -0.00000930, -0.00001102, -0.00001287, -0.00001479, -0.00001669, -0.00001848, -0.00002007, -0.00002135, -0.00002224, -0.00002264, -0.00002249, -0.00002175, -0.00002038, -0.00001839, -0.00001581, -0.00001269, -0.00000911, -0.00000519, -0.00000105, 0.00000317, 0.00000731, 0.00001124, 0.00001480, 0.00001785, 0.00002029, 0.00002203, 0.00002299, 0.00002316, 0.00002254, 0.00002116, 0.00001909, 0.00001644, 0.00001331, 0.00000986, 0.00000624, 0.00000259, -0.00000093, -0.00000418, -0.00000703, -0.00000938, -0.00001116, -0.00001233, -0.00001286, -0.00001277, -0.00001212, -0.00001097, -0.00000941, -0.00000755, -0.00000551, -0.00000342, -0.00000138, 0.00000050, 0.00000212, 0.00000341, 0.00000432, 0.00000484, 0.00000494, 0.00000467, 0.00000407, 0.00000320, 0.00000213, 0.00000096, -0.00000023, -0.00000135, -0.00000232, -0.00000309, -0.00000358, -0.00000378, -0.00000368, -0.00000327, -0.00000259, -0.00000168, -0.00000060, 0.00000057, 0.00000177, 0.00000292, 0.00000394, 0.00000478, 0.00000537, 0.00000568, 0.00000569, 0.00000539, 0.00000479, 0.00000394, },
		{ -0.00001110, -0.00001067, -0.00001002, -0.00000921, -0.00000829, -0.00000731, -0.00000635, -0.00000544, -0.00000462, -0.00000394, -0.00000340, -0.00000303, -0.00000281, -0.00000274, -0.00000279, -0.00000295, -0.00000318, -0.00000345, -0.00000374, -0.00000403, -0.00000428, -0.00000450, -0.00000468, -0.00000482, -0.00000493, -0.00000502, -0.00000512, -0.00000523, -0.00000537, -0.00000555, -0.00000577, -0.00000602, -0.00000629, -0.00000654, -0.00000674, -0.00000685, -0.00000681, -0.00000657, -0.00000608, -0.00000528, -0.00000413, -0.00000262, -0.00000072, 0.00000157, 0.00000420, 0.00000713, 0.00001030, 0.00001361, 0.00001695, 0.00002021, 0.00002325, 0.00002593, 0.00002814, 0.00002975, 0.00003065, 0.00003077, 0.00003006, 0.00002848, 0.00002606, 0.00002283, 0.00001887, 0.00001430, 0.00000926, 0.00000389, -0.00000162, -0.00000708, -0.00001232, -0.00001715, -0.00002141, -0.00002496, -0.00002770, -0.00002955, -0.00003046, -0.00003044, -0.00002952, -0.00002778, -0.00002532, -0.00002227, -0.00001879, -0.00001503, -0.00001117, -0.00000738, -0.00000381, -0.00000060, 0.00000214, 0.00000433, 0.00000591, 0.00000686, 0.00000720, 0.00000696, 0.00000622, 0.00000507, 0.00000362, 0.00000198, 0.00000028, -0.00000136, -0.00000283, -0.00000406, -0.00000496, -0.00000549, -0.00000562, -0.00000537, -0.00000475, -0.00000382, -0.00000265, -0.00000132, 0.00000007, 0.00000144, 0.00000270, 0.00000376, 0.00000454, 0.00000501, 0.00000511, 0.00000485, 0.00000422, 0.00000325, 0.00000200, 0.00000051, -0.00000114, -0.00000286, -0.00000458, -0.00000623, -0.00000772, -0.00000900, -0.00001002, -0.00001074, -0.00001116, -0.00001128, },
		{ 0.00000288, 0.00000311, 0.00000314, 0.00000298, 0.00000267, 0.00000225, 0.00000177, 0.00000125, 0.00000074, 0.00000027, -0.00000015, -0.00000051, -0.00000079, -0.00000099, -0.00000112, -0.00000120, -0.00000123, -0.00000122, -0.00000119, -0.00000114, -0.00000107, -0.00000097, -0.00000084, -0.00000065, -0.00000039, -0.00000003, 0.00000044, 0.00000106, 0.00000183, 0.00000276, 0.00000384, 0.00000507, 0.00000641, 0.00000782, 0.00000925, 0.00001063, 0.00001188, 0.00001293, 0.00001368, 0.00001407, 0.00001403, 0.00001349, 0.00001241, 0.00001077, 0.00000859, 0.00000588, 0.00000271, -0.00000084, -0.00000467, -0.00000865, -0.00001264, -0.00001649, -0.00002004, -0.00002314, -0.00002566, -0.00002746, -0.00002847, -0.00002860, -0.00002782, -0.00002615, -0.00002361, -0.00002028, -0.00001627, -0.00001173, -0.00000681, -0.00000169, 0.00000343, 0.00000838, 0.00001297, 0.00001703, 0.00002044, 0.00002306, 0.00002484, 0.00002572, 0.00002571, 0.00002485, 0.00002321, 0.00002089, 0.00001803, 0.00001477, 0.00001128, 0.00000773, 0.00000427, 0.00000106, -0.00000178, -0.00000414, -0.00000594, -0.00000714, -0.00000773, -0.00000772, -0.00000718, -0.00000618, -0.00000482, -0.00000320, -0.00000146, 0.00000029, 0.00000193, 0.00000334, 0.00000445, 0.00000517, 0.00000547, 0.00000533, 0.00000476, 0.00000378, 0.00000246, 0.00000086, -0.00000093, -0.00000281, -0.00000469, -0.00000647, -0.00000806, -0.00000939, -0.00001040, -0.00001105, -0.00001132, -0.00001120, -0.00001072, -0.00000991, -0.00000883, -0.00000753, -0.00000609, -0.00000458, -0.00000308, -0.00000164, -0.00000033, 0.00000080, 0.00000173, 0.00000242, },
		{ -0.00000281, -0.00000312, -0.00000327, -0.00000328, -0.00000315, -0.00000292, -0.00000261, -0.00000224, -0.00000183, -0.00000140, -0.00000096, -0.00000051, -0.00000005, 0.00000042, 0.00000092, 0.00000145, 0.00000202, 0.00000265, 0.00000333, 0.00000408, 0.00000486, 0.00000568, 0.00000650, 0.00000729, 0.00000802, 0.00000864, 0.00000910, 0.00000938, 0.00000941, 0.00000919, 0.00000869, 0.00000790, 0.00000684, 0.00000553, 0.00000402, 0.00000236, 0.00000063, -0.00000110, -0.00000273, -0.00000416, -0.00000532, -0.00000611, -0.00000647, -0.00000634, -0.00000569, -0.00000453, -0.00000286, -0.00000073, 0.00000179, 0.00000459, 0.00000757, 0.00001060, 0.00001352, 0.00001621, 0.00001851, 0.00002031, 0.00002148, 0.00002194, 0.00002164, 0.00002055, 0.00001867, 0.00001604, 0.00001275, 0.00000890, 0.00000463, 0.00000009, -0.00000455, -0.00000911, -0.00001343, -0.00001733, -0.00002067, -0.00002334, -0.00002523, -0.00002629, -0.00002649, -0.00002586, -0.00002445, -0.00002233, -0.00001963, -0.00001647, -0.00001302, -0.00000943, -0.00000587, -0.00000248, 0.00000058, 0.00000320, 0.00000528, 0.00000676, 0.00000761, 0.00000782, 0.00000743, 0.00000649, 0.00000508, 0.00000332, 0.00000132, -0.00000080, -0.00000292, -0.00000491, -0.00000667, -0.00000811, -0.00000916, -0.00000978, -0.00000995, -0.00000967, -0.00000898, -0.00000792, -0.00000656, -0.00000498, -0.00000327, -0.00000151, 0.00000019, 0.00000177, 0.00000316, 0.00000429, 0.00000512, 0.00000564, 0.00000584, 0.00000574, 0.00000536, 0.00000474, 0.00000394, 0.00000300, 0.00000199, 0.00000097, -0.00000001, -0.00000091, -0.00000170, -0.00000233, },
		{ 0.00000323, 0.00000377, 0.00000421, 0.00000456, 0.00000483, 0.00000504, 0.00000521, 0.00000537, 0.00000552, 0.00000568, 0.00000585, 0.00000602, 0.00000619, 0.00000633, 0.00000643, 0.00000646, 0.00000640, 0.00000622, 0.00000590, 0.00000543, 0.00000480, 0.00000402, 0.00000310, 0.00000207, 0.00000095, -0.00000021, -0.00000137, -0.00000247, -0.00000347, -0.00000433, -0.00000500, -0.00000546, -0.00000569, -0.00000568, -0.00000546, -0.00000504, -0.00000447, -0.00000380, -0.00000310, -0.00000245, -0.00000190, -0.00000155, -0.00000145, -0.00000166, -0.00000221, -0.00000313, -0.00000442, -0.00000606, -0.00000800, -0.00001017, -0.00001248, -0.00001484, -0.00001712, -0.00001921, -0.00002099, -0.00002234, -0.00002317, -0.00002338, -0.00002292, -0.00002176, -0.00001989, -0.00001735, -0.00001418, -0.00001048, -0.00000637, -0.00000199, 0.00000252, 0.00000699, 0.00001126, 0.00001517, 0.00001858, 0.00002136, 0.00002340, 0.00002465, 0.00002506, 0.00002463, 0.00002340, 0.00002142, 0.00001881, 0.00001567, 0.00001214, 0.00000839, 0.00000456, 0.00000082, -0.00000270, -0.00000585, -0.00000854, -0.00001067, -0.00001219, -0.00001309, -0.00001336, -0.00001303, -0.00001218, -0.00001088, -0.00000924, -0.00000735, -0.00000535, -0.00000333, -0.00000142, 0.00000030, 0.00000175, 0.00000287, 0.00000361, 0.00000398, 0.00000397, 0.00000361, 0.00000294, 0.00000204, 0.00000096, -0.00000021, -0.00000140, -0.00000254, -0.00000357, -0.00000441, -0.00000505, -0.00000544, -0.00000557, -0.00000545, -0.00000509, -0.00000452, -0.00000378, -0.00000291, -0.00000196, -0.00000098, 0.00000000, 0.00000094, 0.00000180, 0.00000257, },
		{ 0.00000224, 0.00000184, 0.00000155, 0.00000135, 0.00000123, 0.00000116, 0.00000110, 0.00000103, 0.00000092, 0.00000075, 0.00000051, 0.00000018, -0.00000023, -0.00000071, -0.00000125, -0.00000182, -0.00000240, -0.00000296, -0.00000346, -0.00000389, -0.00000422, -0.00000442, -0.00000450, -0.00000445, -0.00000428, -0.00000400, -0.00000364, -0.00000322, -0.00000278, -0.00000235, -0.00000195, -0.00000162, -0.00000138, -0.00000123, -0.00000118, -0.00000121, -0.00000131, -0.00000145, -0.00000158, -0.00000166, -0.00000164, -0.00000147, -0.00000112, -0.00000052, 0.00000032, 0.00000145, 0.00000284, 0.00000448, 0.00000633, 0.00000834, 0.00001043, 0.00001252, 0.00001451, 0.00001630, 0.00001779, 0.00001888, 0.00001949, 0.00001953, 0.00001896, 0.00001774, 0.00001587, 0.00001337, 0.00001030, 0.00000672, 0.00000274, -0.00000153, -0.00000593, -0.00001033, -0.00001458, -0.00001852, -0.00002202, -0.00002496, -0.00002724, -0.00002877, -0.00002952, -0.00002947, -0.00002863, -0.00002706, -0.00002483, -0.00002204, -0.00001881, -0.00001528, -0.00001158, -0.00000787, -0.00000429, -0.00000095, 0.00000203, 0.00000457, 0.00000659, 0.00000807, 0.00000899, 0.00000937, 0.00000926, 0.00000872, 0.00000783, 0.00000668, 0.00000537, 0.00000401, 0.00000269, 0.00000150, 0.00000051, -0.00000021, -0.00000064, -0.00000074, -0.00000052, 0.00000000, 0.00000078, 0.00000177, 0.00000291, 0.00000414, 0.00000537, 0.00000655, 0.00000762, 0.00000851, 0.00000920, 0.00000965, 0.00000985, 0.00000980, 0.00000952, 0.00000903, 0.00000838, 0.00000760, 0.00000675, 0.00000586, 0.00000498, 0.00000415, 0.00000341, 0.00000277, },
		{ -0.00000023, 0.00000019, 0.00000045, 0.00000055, 0.00000051, 0.00000035, 0.00000008, -0.00000025, -0.00000061, -0.00000098, -0.00000132, -0.00000160, -0.00000181, -0.00000194, -0.00000199, -0.00000194, -0.00000181, -0.00000162, -0.00000137, -0.00000109, -0.00000081, -0.00000052, -0.00000026, -0.00000004, 0.00000015, 0.00000029, 0.00000038, 0.00000044, 0.00000047, 0.00000048, 0.00000049, 0.00000049, 0.00000051, 0.00000053, 0.00000057, 0.00000061, 0.00000063, 0.00000061, 0.00000054, 0.00000037, 0.00000009, -0.00000034, -0.00000094, -0.00000173, -0.00000273, -0.00000392, -0.00000531, -0.00000687, -0.00000856, -0.00001032, -0.00001209, -0.00001381, -0.00001538, -0.00001673, -0.00001777, -0.00001842, -0.00001861, -0.00001828, -0.00001739, -0.00001592, -0.00001388, -0.00001128, -0.00000817, -0.00000462, -0.00000072, 0.00000341, 0.00000766, 0.00001190, 0.00001599, 0.00001981, 0.00002323, 0.00002614, 0.00002845, 0.00003009, 0.00003103, 0.00003125, 0.00003075, 0.00002959, 0.00002784, 0.00002557, 0.00002290, 0.00001995, 0.00001684, 0.00001372, 0.00001070, 0.00000789, 0.00000541, 0.00000333, 0.00000171, 0.00000058, -0.00000004, -0.00000018, 0.00000012, 0.00000081, 0.00000180, 0.00000300, 0.00000433, 0.00000567, 0.00000694, 0.00000804, 0.00000891, 0.00000947, 0.00000970, 0.00000956, 0.00000907, 0.00000823, 0.00000709, 0.00000570, 0.00000413, 0.00000244, 0.00000071, -0.00000098, -0.00000256, -0.00000397, -0.00000515, -0.00000607, -0.00000670, -0.00000703, -0.00000708, -0.00000686, -0.00000641, -0.00000576, -0.00000498, -0.00000411, -0.00000321, -0.00000233, -0.00000152, -0.00000081, },
		{ 0.00000035, -0.00000028, -0.00000077, -0.00000113, -0.00000133, -0.00000138, -0.00000129, -0.00000107, -0.00000074, -0.00000033, 0.00000013, 0.00000062, 0.00000111, 0.00000158, 0.00000200, 0.00000236, 0.00000265, 0.00000286, 0.00000299, 0.00000304, 0.00000302, 0.00000294, 0.00000281, 0.00000264, 0.00000244, 0.00000222, 0.00000200, 0.00000178, 0.00000157, 0.00000139, 0.00000123, 0.00000111, 0.00000105, 0.00000104, 0.00000110, 0.00000125, 0.00000150, 0.00000187, 0.00000237, 0.00000301, 0.00000382, 0.00000480, 0.00000594, 0.00000726, 0.00000874, 0.00001035, 0.00001208, 0.00001388, 0.00001569, 0.00001748, 0.00001916, 0.00002068, 0.00002196, 0.00002293, 0.00002352, 0.00002367, 0.00002334, 0.00002248, 0.00002109, 0.00001915, 0.00001669, 0.00001374, 0.00001037, 0.00000665, 0.00000269, -0.00000143, -0.00000556, -0.00000960, -0.00001341, -0.00001689, -0.00001991, -0.00002238, -0.00002423, -0.00002541, -0.00002588, -0.00002564, -0.00002471, -0.00002316, -0.00002104, -0.00001846, -0.00001554, -0.00001238, -0.00000913, -0.00000592, -0.00000288, -0.00000013, 0.00000224, 0.00000413, 0.00000548, 0.00000626, 0.00000646, 0.00000610, 0.00000520, 0.00000385, 0.00000211, 0.00000008, -0.00000212, -0.00000439, -0.00000662, -0.00000869, -0.00001053, -0.00001205, -0.00001318, -0.00001388, -0.00001413, -0.00001394, -0.00001332, -0.00001231, -0.00001096, -0.00000936, -0.00000756, -0.00000567, -0.00000375, -0.00000188, -0.00000015, 0.00000140, 0.00000270, 0.00000373, 0.00000446, 0.00000490, 0.00000505, 0.00000493, 0.00000459, 0.00000405, 0.00000338, 0.00000263, 0.00000184, 0.00000106, },
		{ 0.00000055, 0.00000113, 0.00000164, 0.00000204, 0.00000231, 0.00000244, 0.00000241, 0.00000222, 0.00000189, 0.00000143, 0.00000086, 0.00000020, -0.00000052, -0.00000127, -0.00000202, -0.00000274, -0.00000340, -0.00000398, -0.00000446, -0.00000482, -0.00000506, -0.00000516, -0.00000514, -0.00000499, -0.00000472, -0.00000436, -0.00000392, -0.00000342, -0.00000289, -0.00000235, -0.00000185, -0.00000142, -0.00000107, -0.00000086, -0.00000081, -0.00000094, -0.00000129, -0.00000187, -0.00000271, -0.00000380, -0.00000516, -0.00000677, -0.00000862, -0.00001069, -0.00001293, -0.00001529, -0.00001773, -0.00002018, -0.00002256, -0.00002480, -0.00002682, -0.00002854, -0.00002988, -0.00003076, -0.00003113, -0.00003092, -0.00003011, -0.00002868, -0.00002661, -0.00002393, -0.00002067, -0.00001691, -0.00001271, -0.00000818, -0.00000342, 0.00000143, 0.00000624, 0.00001089, 0.00001523, 0.00001913, 0.00002249, 0.00002519, 0.00002716, 0.00002834, 0.00002870, 0.00002824, 0.00002696, 0.00002494, 0.00002224, 0.00001896, 0.00001521, 0.00001114, 0.00000688, 0.00000258, -0.00000162, -0.00000557, -0.00000916, -0.00001228, -0.00001483, -0.00001676, -0.00001801, -0.00001859, -0.00001850, -0.00001778, -0.00001648, -0.00001470, -0.00001252, -0.00001005, -0.00000740, -0.00000469, -0.00000202, 0.00000051, 0.00000281, 0.00000480, 0.00000643, 0.00000766, 0.00000847, 0.00000887, 0.00000887, 0.00000850, 0.00000783, 0.00000691, 0.00000580, 0.00000458, 0.00000331, 0.00000207, 0.00000090, -0.00000013, -0.00000098, -0.00000163, -0.00000207, -0.00000228, -0.00000229, -0.00000209, -0.00000174, -0.00000126, -0.00000069, -0.00000007, },
		{ -0.00000655, -0.00000759, -0.00000858, -0.00000948, -0.00001024, -0.00001084, -0.00001126, -0.00001146, -0.00001145, -0.00001123, -0.00001080, -0.00001019, -0.00000941, -0.00000851, -0.00000751, -0.00000647, -0.00000543, -0.00000443, -0.00000352, -0.00000272, -0.00000209, -0.00000165, -0.00000141, -0.00000139, -0.00000160, -0.00000201, -0.00000262, -0.00000339, -0.00000429, -0.00000527, -0.00000628, -0.00000726, -0.00000816, -0.00000891, -0.00000945, -0.00000973, -0.00000969, -0.00000930, -0.00000852, -0.00000732, -0.00000571, -0.00000368, -0.00000127, 0.00000149, 0.00000456, 0.00000786, 0.00001131, 0.00001483, 0.00001833, 0.00002168, 0.00002480, 0.00002757, 0.00002990, 0.00003169, 0.00003287, 0.00003336, 0.00003312, 0.00003212, 0.00003036, 0.00002783, 0.00002460, 0.00002070, 0.00001623, 0.00001129, 0.00000598, 0.00000044, -0.00000519, -0.00001077, -0.00001615, -0.00002121, -0.00002580, -0.00002981, -0.00003314, -0.00003571, -0.00003746, -0.00003837, -0.00003842, -0.00003763, -0.00003605, -0.00003375, -0.00003081, -0.00002733, -0.00002344, -0.00001926, -0.00001493, -0.00001058, -0.00000633, -0.00000230, 0.00000140, 0.00000468, 0.00000747, 0.00000973, 0.00001141, 0.00001252, 0.00001307, 0.00001309, 0.00001262, 0.00001173, 0.00001050, 0.00000901, 0.00000734, 0.00000557, 0.00000380, 0.00000209, 0.00000051, -0.00000089, -0.00000206, -0.00000298, -0.00000364, -0.00000404, -0.00000419, -0.00000413, -0.00000388, -0.00000350, -0.00000303, -0.00000252, -0.00000202, -0.00000159, -0.00000126, -0.00000107, -0.00000104, -0.00000118, -0.00000152, -0.00000203, -0.00000272, -0.00000354, -0.00000448, -0.00000550, },
		{ -0.00000311, -0.00000164, -0.00000013, 0.00000138, 0.00000281, 0.00000413, 0.00000528, 0.00000622, 0.00000691, 0.00000733, 0.00000746, 0.00000730, 0.00000687, 0.00000618, 0.00000527, 0.00000420, 0.00000300, 0.00000175, 0.00000050, -0.00000067, -0.00000171, -0.00000256, -0.00000317, -0.00000349, -0.00000350, -0.00000318, -0.00000254, -0.00000159, -0.00000036, 0.00000111, 0.00000276, 0.00000451, 0.00000630, 0.00000803, 0.00000964, 0.00001103, 0.00001212, 0.00001285, 0.00001316, 0.00001299, 0.00001233, 0.00001115, 0.00000947, 0.00000730, 0.00000469, 0.00000170, -0.00000159, -0.00000511, -0.00000875, -0.00001240, -0.00001596, -0.00001932, -0.00002237, -0.00002502, -0.00002717, -0.00002876, -0.00002973, -0.00003003, -0.00002965, -0.00002859, -0.00002686, -0.00002450, -0.00002157, -0.00001814, -0.00001431, -0.00001016, -0.00000580, -0.00000134, 0.00000311, 0.00000743, 0.00001153, 0.00001531, 0.00001868, 0.00002158, 0.00002394, 0.00002573, 0.00002693, 0.00002752, 0.00002753, 0.00002698, 0.00002591, 0.00002437, 0.00002244, 0.00002019, 0.00001769, 0.00001503, 0.00001230, 0.00000957, 0.00000691, 0.00000440, 0.00000210, 0.00000006, -0.00000170, -0.00000314, -0.00000426, -0.00000505, -0.00000553, -0.00000573, -0.00000567, -0.00000540, -0.00000497, -0.00000443, -0.00000383, -0.00000322, -0.00000265, -0.00000216, -0.00000180, -0.00000158, -0.00000153, -0.00000166, -0.00000196, -0.00000242, -0.00000303, -0.00000375, -0.00000454, -0.00000537, -0.00000619, -0.00000696, -0.00000762, -0.00000814, -0.00000848, -0.00000860, -0.00000850, -0.00000815, -0.00000756, -0.00000674, -0.00000570, -0.00000447, },
		{ 0.00000564, 0.00000446, 0.00000309, 0.00000158, -0.00000001, -0.00000159, -0.00000310, -0.00000448, -0.00000564, -0.00000654, -0.00000712, -0.00000736, -0.00000723, -0.00000673, -0.00000587, -0.00000470, -0.00000324, -0.00000156, 0.00000027, 0.00000218, 0.00000409, 0.00000590, 0.00000756, 0.00000899, 0.00001012, 0.00001092, 0.00001136, 0.00001142, 0.00001111, 0.00001045, 0.00000948, 0.00000825, 0.00000684, 0.00000531, 0.00000375, 0.00000224, 0.00000086, -0.00000032, -0.00000124, -0.00000183, -0.00000208, -0.00000195, -0.00000144, -0.00000058, 0.00000061, 0.00000208, 0.00000376, 0.00000559, 0.00000748, 0.00000935, 0.00001111, 0.00001269, 0.00001402, 0.00001502, 0.00001564, 0.00001586, 0.00001565, 0.00001500, 0.00001393, 0.00001246, 0.00001063, 0.00000850, 0.00000611, 0.00000356, 0.00000090, -0.00000179, -0.00000444, -0.00000698, -0.00000934, -0.00001147, -0.00001333, -0.00001488, -0.00001610, -0.00001697, -0.00001749, -0.00001767, -0.00001752, -0.00001706, -0.00001634, -0.00001538, -0.00001423, -0.00001291, -0.00001149, -0.00000999, -0.00000847, -0.00000696, -0.00000550, -0.00000413, -0.00000287, -0.00000175, -0.00000079, -0.00000002, 0.00000057, 0.00000095, 0.00000114, 0.00000112, 0.00000092, 0.00000055, 0.00000003, -0.00000062, -0.00000136, -0.00000216, -0.00000298, -0.00000379, -0.00000454, -0.00000520, -0.00000573, -0.00000610, -0.00000627, -0.00000624, -0.00000598, -0.00000550, -0.00000479, -0.00000389, -0.00000280, -0.00000158, -0.00000026, 0.00000111, 0.00000248, 0.00000379, 0.00000498, 0.00000600, 0.00000681, 0.00000736, 0.00000762, 0.00000758, 0.00000723, 0.00000657, },
		{ -0.00000607, -0.00000513, -0.00000383, -0.00000224, -0.00000039, 0.00000163, 0.00000376, 0.00000590, 0.00000797, 0.00000990, 0.00001159, 0.00001298, 0.00001402, 0.00001468, 0.00001492, 0.00001475, 0.00001418, 0.00001326, 0.00001202, 0.00001054, 0.00000887, 0.00000710, 0.00000529, 0.00000353, 0.00000188, 0.00000038, -0.00000092, -0.00000200, -0.00000286, -0.00000350, -0.00000396, -0.00000427, -0.00000450, -0.00000470, -0.00000494, -0.00000529, -0.00000579, -0.00000652, -0.00000749, -0.00000874, -0.00001027, -0.00001206, -0.00001408, -0.00001628, -0.00001858, -0.00002090, -0.00002315, -0.00002522, -0.00002701, -0.00002843, -0.00002939, -0.00002981, -0.00002964, -0.00002884, -0.00002739, -0.00002530, -0.00002261, -0.00001937, -0.00001566, -0.00001157, -0.00000722, -0.00000272, 0.00000181, 0.00000625, 0.00001047, 0.00001437, 0.00001785, 0.00002083, 0.00002325, 0.00002508, 0.00002628, 0.00002686, 0.00002684, 0.00002626, 0.00002516, 0.00002361, 0.00002168, 0.00001944, 0.00001698, 0.00001436, 0.00001165, 0.00000894, 0.00000626, 0.00000369, 0.00000124, -0.00000103, -0.00000311, -0.00000498, -0.00000663, -0.00000804, -0.00000922, -0.00001017, -0.00001088, -0.00001136, -0.00001161, -0.00001164, -0.00001144, -0.00001103, -0.00001041, -0.00000960, -0.00000862, -0.00000748, -0.00000621, -0.00000485, -0.00000344, -0.00000201, -0.00000061, 0.00000070, 0.00000189, 0.00000291, 0.00000372, 0.00000428, 0.00000457, 0.00000457, 0.00000429, 0.00000374, 0.00000293, 0.00000191, 0.00000073, -0.00000056, -0.00000190, -0.00000320, -0.00000439, -0.00000541, -0.00000620, -0.00000669, -0.00000685, -0.00000665, },
		{ 0.00001644, 0.00001654, 0.00001622, 0.00001547, 0.00001433, 0.00001284, 0.00001106, 0.00000905, 0.00000691, 0.00000470, 0.00000252, 0.00000044, -0.00000148, -0.00000317, -0.00000461, -0.00000577, -0.00000665, -0.00000726, -0.00000765, -0.00000785, -0.00000793, -0.00000793, -0.00000792, -0.00000796, -0.00000809, -0.00000835, -0.00000875, -0.00000930, -0.00000996, -0.00001072, -0.00001149, -0.00001222, -0.00001281, -0.00001318, -0.00001323, -0.00001288, -0.00001204, -0.00001066, -0.00000869, -0.00000612, -0.00000297, 0.00000072, 0.00000488, 0.00000941, 0.00001419, 0.00001906, 0.00002386, 0.00002843, 0.00003260, 0.00003621, 0.00003909, 0.00004114, 0.00004224, 0.00004232, 0.00004136, 0.00003934, 0.00003632, 0.00003237, 0.00002759, 0.00002213, 0.00001613, 0.00000978, 0.00000326, -0.00000324, -0.00000955, -0.00001550, -0.00002094, -0.00002575, -0.00002983, -0.00003311, -0.00003556, -0.00003716, -0.00003795, -0.00003796, -0.00003726, -0.00003595, -0.00003411, -0.00003184, -0.00002924, -0.00002642, -0.00002347, -0.00002046, -0.00001746, -0.00001453, -0.00001171, -0.00000902, -0.00000648, -0.00000409, -0.00000186, 0.00000022, 0.00000216, 0.00000396, 0.00000562, 0.00000715, 0.00000854, 0.00000977, 0.00001084, 0.00001171, 0.00001237, 0.00001280, 0.00001299, 0.00001291, 0.00001258, 0.00001200, 0.00001118, 0.00001016, 0.00000897, 0.00000766, 0.00000631, 0.00000496, 0.00000369, 0.00000257, 0.00000165, 0.00000100, 0.00000066, 0.00000065, 0.00000099, 0.00000168, 0.00000269, 0.00000399, 0.00000551, 0.00000719, 0.00000895, 0.00001070, 0.00001235, 0.00001382, 0.00001504, 0.00001593, },
		{ -0.00000581, -0.00000722, -0.00000830, -0.00000904, -0.00000944, -0.00000952, -0.00000930, -0.00000886, -0.00000825, -0.00000754, -0.00000681, -0.00000612, -0.00000553, -0.00000510, -0.00000484, -0.00000476, -0.00000486, -0.00000510, -0.00000544, -0.00000580, -0.00000612, -0.00000630, -0.00000628, -0.00000597, -0.00000532, -0.00000428, -0.00000284, -0.00000098, 0.00000124, 0.00000377, 0.00000652, 0.00000938, 0.00001221, 0.00001488, 0.00001724, 0.00001912, 0.00002041, 0.00002098, 0.00002073, 0.00001961, 0.00001759, 0.00001468, 0.00001094, 0.00000647, 0.00000141, -0.00000409, -0.00000983, -0.00001560, -0.00002118, -0.00002635, -0.00003091, -0.00003467, -0.00003746, -0.00003917, -0.00003972, -0.00003906, -0.00003721, -0.00003422, -0.00003019, -0.00002525, -0.00001958, -0.00001337, -0.00000681, -0.00000013, 0.00000647, 0.00001279, 0.00001866, 0.00002393, 0.00002847, 0.00003222, 0.00003512, 0.00003716, 0.00003837, 0.00003881, 0.00003856, 0.00003772, 0.00003640, 0.00003471, 0.00003278, 0.00003070, 0.00002859, 0.00002651, 0.00002453, 0.00002268, 0.00002099, 0.00001945, 0.00001805, 0.00001675, 0.00001553, 0.00001433, 0.00001313, 0.00001189, 0.00001058, 0.00000919, 0.00000773, 0.00000621, 0.00000466, 0.00000312, 0.00000165, 0.00000030, -0.00000087, -0.00000180, -0.00000244, -0.00000276, -0.00000273, -0.00000234, -0.00000161, -0.00000057, 0.00000074, 0.00000224, 0.00000386, 0.00000551, 0.00000711, 0.00000855, 0.00000976, 0.00001067, 0.00001121, 0.00001134, 0.00001105, 0.00001033, 0.00000922, 0.00000776, 0.00000600, 0.00000404, 0.00000196, -0.00000016, -0.00000221, -0.00000412, },
		{ -0.00000126, -0.00000016, 0.00000070, 0.00000129, 0.00000159, 0.00000164, 0.00000147, 0.00000115, 0.00000073, 0.00000032, -0.00000001, -0.00000016, -0.00000007, 0.00000033, 0.00000108, 0.00000218, 0.00000364, 0.00000540, 0.00000741, 0.00000957, 0.00001178, 0.00001392, 0.00001588, 0.00001751, 0.00001873, 0.00001943, 0.00001955, 0.00001905, 0.00001792, 0.00001620, 0.00001396, 0.00001130, 0.00000835, 0.00000527, 0.00000223, -0.00000059, -0.00000302, -0.00000489, -0.00000606, -0.00000643, -0.00000591, -0.00000449, -0.00000218, 0.00000096, 0.00000482, 0.00000925, 0.00001406, 0.00001905, 0.00002399, 0.00002865, 0.00003279, 0.00003622, 0.00003875, 0.00004024, 0.00004058, 0.00003972, 0.00003766, 0.00003446, 0.00003020, 0.00002505, 0.00001917, 0.00001277, 0.00000609, -0.00000065, -0.00000720, -0.00001336, -0.00001894, -0.00002377, -0.00002774, -0.00003077, -0.00003282, -0.00003391, -0.00003410, -0.00003347, -0.00003214, -0.00003026, -0.00002798, -0.00002546, -0.00002286, -0.00002031, -0.00001794, -0.00001584, -0.00001408, -0.00001268, -0.00001166, -0.00001098, -0.00001059, -0.00001041, -0.00001037, -0.00001037, -0.00001032, -0.00001013, -0.00000974, -0.00000909, -0.00000815, -0.00000693, -0.00000543, -0.00000371, -0.00000184, 0.00000012, 0.00000205, 0.00000386, 0.00000547, 0.00000677, 0.00000770, 0.00000819, 0.00000822, 0.00000779, 0.00000690, 0.00000560, 0.00000397, 0.00000208, 0.00000003, -0.00000206, -0.00000409, -0.00000596, -0.00000757, -0.00000885, -0.00000974, -0.00001021, -0.00001025, -0.00000988, -0.00000914, -0.00000810, -0.00000683, -0.00000542, -0.00000396, -0.00000255, },
		{ 0.00000164, 0.00000080, 0.00000026, 0.00000006, 0.00000025, 0.00000080, 0.00000168, 0.00000284, 0.00000419, 0.00000562, 0.00000703, 0.00000829, 0.00000929, 0.00000993, 0.00001013, 0.00000984, 0.00000902, 0.00000768, 0.00000586, 0.00000363, 0.00000108, -0.00000167, -0.00000449, -0.00000724, -0.00000979, -0.00001201, -0.00001380, -0.00001508, -0.00001581, -0.00001597, -0.00001561, -0.00001479, -0.00001361, -0.00001221, -0.00001075, -0.00000939, -0.00000830, -0.00000766, -0.00000759, -0.00000823, -0.00000966, -0.00001191, -0.00001497, -0.00001878, -0.00002323, -0.00002817, -0.00003342, -0.00003873, -0.00004389, -0.00004862, -0.00005270, -0.00005589, -0.00005800, -0.00005887, -0.00005840, -0.00005654, -0.00005329, -0.00004873, -0.00004298, -0.00003623, -0.00002869, -0.00002062, -0.00001229, -0.00000401, 0.00000396, 0.00001134, 0.00001791, 0.00002348, 0.00002789, 0.00003107, 0.00003297, 0.00003364, 0.00003314, 0.00003161, 0.00002922, 0.00002616, 0.00002264, 0.00001890, 0.00001514, 0.00001157, 0.00000836, 0.00000564, 0.00000351, 0.00000203, 0.00000121, 0.00000100, 0.00000133, 0.00000210, 0.00000318, 0.00000442, 0.00000568, 0.00000681, 0.00000769, 0.00000820, 0.00000828, 0.00000788, 0.00000699, 0.00000564, 0.00000388, 0.00000181, -0.00000045, -0.00000279, -0.00000508, -0.00000717, -0.00000896, -0.00001033, -0.00001121, -0.00001155, -0.00001133, -0.00001057, -0.00000932, -0.00000765, -0.00000565, -0.00000345, -0.00000117, 0.00000107, 0.00000314, 0.00000494, 0.00000639, 0.00000742, 0.00000800, 0.00000813, 0.00000784, 0.00000719, 0.00000625, 0.00000511, 0.00000389, 0.00000270, },
		{ -0.00000509, -0.00000382, -0.00000282, -0.00000219, -0.00000199, -0.00000226, -0.00000302, -0.00000422, -0.00000583, -0.00000774, -0.00000986, -0.00001206, -0.00001422, -0.00001623, -0.00001797, -0.00001936, -0.00002032, -0.00002084, -0.00002090, -0.00002053, -0.00001980, -0.00001880, -0.00001762, -0.00001638, -0.00001521, -0.00001421, -0.00001348, -0.00001311, -0.00001312, -0.00001353, -0.00001431, -0.00001539, -0.00001666, -0.00001800, -0.00001923, -0.00002019, -0.00002070, -0.00002059, -0.00001971, -0.00001794, -0.00001519, -0.00001145, -0.00000672, -0.00000109, 0.00000531, 0.00001230, 0.00001964, 0.00002707, 0.00003429, 0.00004100, 0.00004691, 0.00005175, 0.00005526, 0.00005726, 0.00005760, 0.00005621, 0.00005310, 0.00004834, 0.00004207, 0.00003450, 0.00002590, 0.00001659, 0.00000690, -0.00000282, -0.00001220, -0.00002092, -0.00002867, -0.00003520, -0.00004031, -0.00004389, -0.00004587, -0.00004627, -0.00004518, -0.00004275, -0.00003918, -0.00003472, -0.00002965, -0.00002426, -0.00001882, -0.00001362, -0.00000890, -0.00000484, -0.00000160, 0.00000072, 0.00000211, 0.00000257, 0.00000219, 0.00000110, -0.00000054, -0.00000255, -0.00000473, -0.00000686, -0.00000876, -0.00001027, -0.00001125, -0.00001160, -0.00001128, -0.00001028, -0.00000865, -0.00000647, -0.00000387, -0.00000100, 0.00000199, 0.00000491, 0.00000759, 0.00000988, 0.00001165, 0.00001281, 0.00001328, 0.00001304, 0.00001213, 0.00001059, 0.00000852, 0.00000604, 0.00000330, 0.00000046, -0.00000234, -0.00000494, -0.00000722, -0.00000906, -0.00001039, -0.00001117, -0.00001141, -0.00001112, -0.00001039, -0.00000929, -0.00000796, -0.00000651, },
		{ -0.00000138, -0.00000346, -0.00000549, -0.00000737, -0.00000896, -0.00001021, -0.00001106, -0.00001149, -0.00001152, -0.00001118, -0.00001055, -0.00000972, -0.00000877, -0.00000781, -0.00000693, -0.00000620, -0.00000569, -0.00000543, -0.00000541, -0.00000562, -0.00000599, -0.00000644, -0.00000688, -0.00000720, -0.00000728, -0.00000702, -0.00000632, -0.00000513, -0.00000341, -0.00000118, 0.00000151, 0.00000458, 0.00000788, 0.00001125, 0.00001448, 0.00001737, 0.00001968, 0.00002122, 0.00002177, 0.00002119, 0.00001937, 0.00001624, 0.00001181, 0.00000617, -0.00000056, -0.00000817, -0.00001638, -0.00002490, -0.00003337, -0.00004144, -0.00004873, -0.00005491, -0.00005966, -0.00006271, -0.00006388, -0.00006304, -0.00006016, -0.00005529, -0.00004857, -0.00004021, -0.00003053, -0.00001986, -0.00000861, 0.00000280, 0.00001394, 0.00002440, 0.00003380, 0.00004180, 0.00004815, 0.00005266, 0.00005523, 0.00005586, 0.00005461, 0.00005165, 0.00004722, 0.00004159, 0.00003511, 0.00002813, 0.00002102, 0.00001413, 0.00000778, 0.00000225, -0.00000224, -0.00000555, -0.00000760, -0.00000839, -0.00000798, -0.00000651, -0.00000417, -0.00000119, 0.00000217, 0.00000564, 0.00000895, 0.00001186, 0.00001416, 0.00001567, 0.00001629, 0.00001595, 0.00001467, 0.00001250, 0.00000956, 0.00000602, 0.00000205, -0.00000211, -0.00000625, -0.00001014, -0.00001360, -0.00001644, -0.00001854, -0.00001981, -0.00002021, -0.00001976, -0.00001853, -0.00001660, -0.00001414, -0.00001129, -0.00000825, -0.00000520, -0.00000232, 0.00000023, 0.00000233, 0.00000386, 0.00000477, 0.00000504, 0.00000468, 0.00000376, 0.00000237, 0.00000061, },
		{ -0.00001894, -0.00001800, -0.00001687, -0.00001567, -0.00001450, -0.00001342, -0.00001249, -0.00001174, -0.00001117, -0.00001074, -0.00001041, -0.00001010, -0.00000973, -0.00000924, -0.00000853, -0.00000757, -0.00000630, -0.00000472, -0.00000286, -0.00000075, 0.00000149, 0.00000378, 0.00000597, 0.00000792, 0.00000949, 0.00001053, 0.00001092, 0.00001058, 0.00000946, 0.00000755, 0.00000490, 0.00000161, -0.00000217, -0.00000624, -0.00001038, -0.00001432, -0.00001779, -0.00002052, -0.00002225, -0.00002277, -0.00002192, -0.00001957, -0.00001570, -0.00001034, -0.00000363, 0.00000424, 0.00001299, 0.00002229, 0.00003176, 0.00004096, 0.00004949, 0.00005691, 0.00006284, 0.00006695, 0.00006897, 0.00006870, 0.00006608, 0.00006110, 0.00005389, 0.00004467, 0.00003374, 0.00002151, 0.00000842, -0.00000504, -0.00001835, -0.00003101, -0.00004254, -0.00005251, -0.00006057, -0.00006645, -0.00006999, -0.00007111, -0.00006988, -0.00006643, -0.00006102, -0.00005395, -0.00004564, -0.00003651, -0.00002700, -0.00001759, -0.00000869, -0.00000070, 0.00000607, 0.00001137, 0.00001504, 0.00001701, 0.00001730, 0.00001602, 0.00001335, 0.00000956, 0.00000494, -0.00000018, -0.00000545, -0.00001054, -0.00001515, -0.00001900, -0.00002188, -0.00002364, -0.00002422, -0.00002360, -0.00002186, -0.00001911, -0.00001556, -0.00001141, -0.00000693, -0.00000237, 0.00000201, 0.00000597, 0.00000930, 0.00001184, 0.00001347, 0.00001414, 0.00001383, 0.00001261, 0.00001057, 0.00000784, 0.00000460, 0.00000102, -0.00000268, -0.00000633, -0.00000976, -0.00001281, -0.00001538, -0.00001738, -0.00001878, -0.00001959, -0.00001983, -0.00001958, },
		{ -0.00000234, -0.00000160, -0.00000096, -0.00000034, 0.00000033, 0.00000110, 0.00000201, 0.00000304, 0.00000418, 0.00000538, 0.00000655, 0.00000760, 0.00000843, 0.00000894, 0.00000903, 0.00000863, 0.00000769, 0.00000620, 0.00000417, 0.00000166, -0.00000121, -0.00000431, -0.00000749, -0.00001055, -0.00001330, -0.00001556, -0.00001716, -0.00001795, -0.00001785, -0.00001680, -0.00001482, -0.00001198, -0.00000843, -0.00000436, -0.00000001, 0.00000432, 0.00000832, 0.00001167, 0.00001408, 0.00001524, 0.00001494, 0.00001300, 0.00000933, 0.00000393, -0.00000312, -0.00001163, -0.00002133, -0.00003187, -0.00004283, -0.00005373, -0.00006409, -0.00007341, -0.00008121, -0.00008705, -0.00009058, -0.00009153, -0.00008971, -0.00008508, -0.00007771, -0.00006778, -0.00005560, -0.00004157, -0.00002620, -0.00001003, 0.00000633, 0.00002228, 0.00003722, 0.00005059, 0.00006193, 0.00007082, 0.00007700, 0.00008029, 0.00008065, 0.00007817, 0.00007307, 0.00006564, 0.00005631, 0.00004554, 0.00003387, 0.00002183, 0.00000996, -0.00000124, -0.00001131, -0.00001989, -0.00002669, -0.00003153, -0.00003433, -0.00003512, -0.00003405, -0.00003131, -0.00002721, -0.00002210, -0.00001635, -0.00001036, -0.00000452, 0.00000083, 0.00000537, 0.00000885, 0.00001111, 0.00001204, 0.00001163, 0.00000994, 0.00000711, 0.00000331, -0.00000122, -0.00000621, -0.00001139, -0.00001648, -0.00002123, -0.00002542, -0.00002886, -0.00003142, -0.00003304, -0.00003368, -0.00003339, -0.00003225, -0.00003038, -0.00002793, -0.00002506, -0.00002195, -0.00001877, -0.00001566, -0.00001275, -0.00001013, -0.00000787, -0.00000598, -0.00000446, -0.00000327, },
		{ 0.00000397, 0.00000370, 0.00000334, 0.00000280, 0.00000201, 0.00000091, -0.00000053, -0.00000231, -0.00000437, -0.00000664, -0.00000904, -0.00001141, -0.00001363, -0.00001556, -0.00001704, -0.00001798, -0.00001826, -0.00001785, -0.00001673, -0.00001494, -0.00001258, -0.00000978, -0.00000671, -0.00000358, -0.00000062, 0.00000195, 0.00000391, 0.00000506, 0.00000525, 0.00000439, 0.00000243, -0.00000057, -0.00000453, -0.00000924, -0.00001447, -0.00001991, -0.00002522, -0.00003002, -0.00003395, -0.00003665, -0.00003782, -0.00003719, -0.00003460, -0.00002995, -0.00002328, -0.00001470, -0.00000447, 0.00000710, 0.00001957, 0.00003244, 0.00004517, 0.00005719, 0.00006793, 0.00007685, 0.00008347, 0.00008740, 0.00008834, 0.00008613, 0.00008072, 0.00007223, 0.00006088, 0.00004705, 0.00003120, 0.00001392, -0.00000416, -0.00002237, -0.00004002, -0.00005645, -0.00007106, -0.00008332, -0.00009282, -0.00009924, -0.00010244, -0.00010239, -0.00009920, -0.00009312, -0.00008453, -0.00007387, -0.00006170, -0.00004858, -0.00003511, -0.00002190, -0.00000947, 0.00000167, 0.00001114, 0.00001863, 0.00002395, 0.00002704, 0.00002792, 0.00002675, 0.00002377, 0.00001929, 0.00001369, 0.00000737, 0.00000075, -0.00000576, -0.00001179, -0.00001702, -0.00002119, -0.00002412, -0.00002571, -0.00002593, -0.00002486, -0.00002260, -0.00001936, -0.00001535, -0.00001083, -0.00000606, -0.00000131, 0.00000319, 0.00000724, 0.00001067, 0.00001338, 0.00001530, 0.00001644, 0.00001683, 0.00001655, 0.00001574, 0.00001451, 0.00001302, 0.00001141, 0.00000981, 0.00000832, 0.00000704, 0.00000600, 0.00000520, 0.00000464, 0.00000426, },
		{ -0.00001395, -0.00001519, -0.00001633, -0.00001724, -0.00001779, -0.00001790, -0.00001750, -0.00001658, -0.00001516, -0.00001332, -0.00001115, -0.00000880, -0.00000644, -0.00000425, -0.00000242, -0.00000110, -0.00000045, -0.00000057, -0.00000151, -0.00000329, -0.00000583, -0.00000904, -0.00001274, -0.00001671, -0.00002071, -0.00002448, -0.00002773, -0.00003021, -0.00003170, -0.00003202, -0.00003106, -0.00002877, -0.00002521, -0.00002049, -0.00001483, -0.00000850, -0.00000184, 0.00000476, 0.00001089, 0.00001614, 0.00002013, 0.00002250, 0.00002300, 0.00002142, 0.00001769, 0.00001184, 0.00000400, -0.00000556, -0.00001648, -0.00002829, -0.00004046, -0.00005241, -0.00006354, -0.00007326, -0.00008100, -0.00008629, -0.00008872, -0.00008802, -0.00008403, -0.00007676, -0.00006635, -0.00005309, -0.00003738, -0.00001977, -0.00000086, 0.00001864, 0.00003802, 0.00005657, 0.00007361, 0.00008851, 0.00010075, 0.00010992, 0.00011574, 0.00011808, 0.00011697, 0.00011255, 0.00010513, 0.00009513, 0.00008304, 0.00006945, 0.00005498, 0.00004026, 0.00002589, 0.00001244, 0.00000039, -0.00000987, -0.00001805, -0.00002397, -0.00002760, -0.00002899, -0.00002831, -0.00002583, -0.00002188, -0.00001685, -0.00001114, -0.00000516, 0.00000068, 0.00000603, 0.00001059, 0.00001414, 0.00001652, 0.00001767, 0.00001759, 0.00001638, 0.00001416, 0.00001114, 0.00000756, 0.00000364, -0.00000035, -0.00000419, -0.00000768, -0.00001066, -0.00001301, -0.00001469, -0.00001568, -0.00001602, -0.00001581, -0.00001517, -0.00001425, -0.00001318, -0.00001213, -0.00001124, -0.00001061, -0.00001031, -0.00001040, -0.00001087, -0.00001167, -0.00001273, },
		{ -0.00001371, -0.00001163, -0.00000953, -0.00000759, -0.00000598, -0.00000485, -0.00000433, -0.00000448, -0.00000533, -0.00000685, -0.00000895, -0.00001150, -0.00001431, -0.00001719, -0.00001990, -0.00002222, -0.00002394, -0.00002487, -0.00002487, -0.00002385, -0.00002180, -0.00001874, -0.00001480, -0.00001014, -0.00000497, 0.00000044, 0.00000580, 0.00001082, 0.00001523, 0.00001878, 0.00002128, 0.00002258, 0.00002264, 0.00002148, 0.00001921, 0.00001603, 0.00001220, 0.00000805, 0.00000395, 0.00000029, -0.00000254, -0.00000418, -0.00000431, -0.00000269, 0.00000082, 0.00000628, 0.00001361, 0.00002264, 0.00003306, 0.00004447, 0.00005638, 0.00006826, 0.00007950, 0.00008951, 0.00009771, 0.00010357, 0.00010664, 0.00010658, 0.00010318, 0.00009635, 0.00008617, 0.00007285, 0.00005675, 0.00003837, 0.00001829, -0.00000281, -0.00002419, -0.00004513, -0.00006488, -0.00008277, -0.00009821, -0.00011070, -0.00011988, -0.00012553, -0.00012759, -0.00012613, -0.00012138, -0.00011372, -0.00010360, -0.00009157, -0.00007827, -0.00006431, -0.00005034, -0.00003695, -0.00002467, -0.00001395, -0.00000514, 0.00000154, 0.00000599, 0.00000820, 0.00000832, 0.00000657, 0.00000326, -0.00000125, -0.00000655, -0.00001221, -0.00001785, -0.00002308, -0.00002760, -0.00003115, -0.00003357, -0.00003477, -0.00003475, -0.00003359, -0.00003144, -0.00002850, -0.00002501, -0.00002123, -0.00001743, -0.00001386, -0.00001073, -0.00000822, -0.00000645, -0.00000547, -0.00000529, -0.00000585, -0.00000703, -0.00000869, -0.00001065, -0.00001271, -0.00001469, -0.00001639, -0.00001767, -0.00001842, -0.00001858, -0.00001812, -0.00001710, -0.00001558, },
		{ -0.00000807, -0.00001135, -0.00001504, -0.00001888, -0.00002264, -0.00002606, -0.00002894, -0.00003108, -0.00003237, -0.00003272, -0.00003212, -0.00003062, -0.00002835, -0.00002545, -0.00002213, -0.00001861, -0.00001515, -0.00001195, -0.00000924, -0.00000718, -0.00000590, -0.00000545, -0.00000584, -0.00000701, -0.00000883, -0.00001114, -0.00001372, -0.00001635, -0.00001879, -0.00002080, -0.00002220, -0.00002285, -0.00002264, -0.00002156, -0.00001968, -0.00001711, -0.00001408, -0.00001085, -0.00000773, -0.00000507, -0.00000322, -0.00000252, -0.00000327, -0.00000572, -0.00001002, -0.00001624, -0.00002433, -0.00003415, -0.00004542, -0.00005778, -0.00007077, -0.00008384, -0.00009642, -0.00010791, -0.00011771, -0.00012529, -0.00013015, -0.00013191, -0.00013032, -0.00012524, -0.00011671, -0.00010488, -0.00009009, -0.00007279, -0.00005355, -0.00003304, -0.00001199, 0.00000885, 0.00002873, 0.00004693, 0.00006281, 0.00007584, 0.00008561, 0.00009185, 0.00009444, 0.00009344, 0.00008905, 0.00008160, 0.00007156, 0.00005948, 0.00004599, 0.00003175, 0.00001742, 0.00000365, -0.00000901, -0.00002005, -0.00002911, -0.00003591, -0.00004033, -0.00004238, -0.00004216, -0.00003992, -0.00003599, -0.00003076, -0.00002466, -0.00001818, -0.00001174, -0.00000577, -0.00000063, 0.00000339, 0.00000611, 0.00000741, 0.00000729, 0.00000584, 0.00000323, -0.00000032, -0.00000451, -0.00000904, -0.00001357, -0.00001781, -0.00002148, -0.00002435, -0.00002626, -0.00002712, -0.00002692, -0.00002571, -0.00002363, -0.00002085, -0.00001761, -0.00001417, -0.00001078, -0.00000770, -0.00000518, -0.00000339, -0.00000247, -0.00000251, -0.00000350, -0.00000539, },
		{ -0.00007370, -0.00007245, -0.00006991, -0.00006626, -0.00006175, -0.00005664, -0.00005119, -0.00004567, -0.00004032, -0.00003531, -0.00003081, -0.00002691, -0.00002363, -0.00002095, -0.00001882, -0.00001711, -0.00001569, -0.00001441, -0.00001311, -0.00001167, -0.00000998, -0.00000796, -0.00000559, -0.00000290, 0.00000005, 0.00000314, 0.00000623, 0.00000913, 0.00001167, 0.00001367, 0.00001498, 0.00001549, 0.00001513, 0.00001389, 0.00001184, 0.00000912, 0.00000590, 0.00000246, -0.00000091, -0.00000387, -0.00000609, -0.00000723, -0.00000700, -0.00000516, -0.00000155, 0.00000391, 0.00001118, 0.00002012, 0.00003048, 0.00004190, 0.00005393, 0.00006604, 0.00007765, 0.00008816, 0.00009698, 0.00010355, 0.00010738, 0.00010808, 0.00010537, 0.00009911, 0.00008932, 0.00007619, 0.00006002, 0.00004130, 0.00002062, -0.00000133, -0.00002378, -0.00004593, -0.00006698, -0.00008616, -0.00010280, -0.00011630, -0.00012623, -0.00013227, -0.00013431, -0.00013238, -0.00012671, -0.00011765, -0.00010572, -0.00009153, -0.00007577, -0.00005919, -0.00004255, -0.00002656, -0.00001188, 0.00000092, 0.00001138, 0.00001920, 0.00002420, 0.00002636, 0.00002581, 0.00002282, 0.00001774, 0.00001105, 0.00000326, -0.00000508, -0.00001342, -0.00002124, -0.00002810, -0.00003362, -0.00003754, -0.00003970, -0.00004007, -0.00003873, -0.00003585, -0.00003174, -0.00002672, -0.00002122, -0.00001564, -0.00001042, -0.00000594, -0.00000255, -0.00000052, -0.00000004, -0.00000120, -0.00000398, -0.00000830, -0.00001395, -0.00002068, -0.00002816, -0.00003603, -0.00004391, -0.00005144, -0.00005826, -0.00006408, -0.00006866, -0.00007184, -0.00007352, },
		{ 0.00001040, 0.00001981, 0.00002796, 0.00003483, 0.00004048, 0.00004499, 0.00004851, 0.00005118, 0.00005315, 0.00005458, 0.00005560, 0.00005631, 0.00005677, 0.00005703, 0.00005707, 0.00005687, 0.00005637, 0.00005550, 0.00005419, 0.00005239, 0.00005005, 0.00004716, 0.00004374, 0.00003985, 0.00003559, 0.00003108, 0.00002649, 0.00002200, 0.00001777, 0.00001398, 0.00001079, 0.00000830, 0.00000658, 0.00000563, 0.00000540, 0.00000576, 0.00000652, 0.00000744, 0.00000823, 0.00000858, 0.00000817, 0.00000668, 0.00000385, -0.00000055, -0.00000667, -0.00001454, -0.00002410, -0.00003520, -0.00004754, -0.00006074, -0.00007431, -0.00008769, -0.00010026, -0.00011139, -0.00012046, -0.00012686, -0.00013010, -0.00012976, -0.00012556, -0.00011737, -0.00010524, -0.00008938, -0.00007015, -0.00004812, -0.00002394, 0.00000158, 0.00002757, 0.00005312, 0.00007729, 0.00009923, 0.00011811, 0.00013326, 0.00014415, 0.00015040, 0.00015186, 0.00014855, 0.00014069, 0.00012871, 0.00011317, 0.00009480, 0.00007441, 0.00005290, 0.00003117, 0.00001010, -0.00000947, -0.00002684, -0.00004141, -0.00005275, -0.00006060, -0.00006488, -0.00006572, -0.00006338, -0.00005829, -0.00005101, -0.00004218, -0.00003251, -0.00002272, -0.00001350, -0.00000550, 0.00000073, 0.00000474, 0.00000624, 0.00000508, 0.00000126, -0.00000508, -0.00001365, -0.00002404, -0.00003574, -0.00004821, -0.00006082, -0.00007299, -0.00008413, -0.00009373, -0.00010135, -0.00010664, -0.00010940, -0.00010951, -0.00010699, -0.00010197, -0.00009468, -0.00008544, -0.00007462, -0.00006265, -0.00004996, -0.00003699, -0.00002414, -0.00001176, -0.00000017, },
		{ 0.00008053, 0.00007519, 0.00006970, 0.00006420, 0.00005878, 0.00005349, 0.00004836, 0.00004337, 0.00003853, 0.00003380, 0.00002917, 0.00002465, 0.00002024, 0.00001599, 0.00001195, 0.00000819, 0.00000480, 0.00000185, -0.00000057, -0.00000240, -0.00000360, -0.00000414, -0.00000407, -0.00000342, -0.00000229, -0.00000082, 0.00000086, 0.00000258, 0.00000417, 0.00000549, 0.00000640, 0.00000681, 0.00000667, 0.00000601, 0.00000491, 0.00000349, 0.00000198, 0.00000063, -0.00000027, -0.00000040, 0.00000057, 0.00000293, 0.00000696, 0.00001285, 0.00002070, 0.00003052, 0.00004221, 0.00005551, 0.00007009, 0.00008545, 0.00010102, 0.00011615, 0.00013010, 0.00014214, 0.00015154, 0.00015761, 0.00015976, 0.00015750, 0.00015052, 0.00013866, 0.00012196, 0.00010067, 0.00007523, 0.00004628, 0.00001463, -0.00001878, -0.00005290, -0.00008660, -0.00011877, -0.00014830, -0.00017419, -0.00019554, -0.00021166, -0.00022204, -0.00022638, -0.00022466, -0.00021709, -0.00020411, -0.00018638, -0.00016477, -0.00014027, -0.00011400, -0.00008713, -0.00006082, -0.00003621, -0.00001431, 0.00000399, 0.00001800, 0.00002722, 0.00003140, 0.00003050, 0.00002473, 0.00001451, 0.00000047, -0.00001662, -0.00003587, -0.00005629, -0.00007689, -0.00009666, -0.00011469, -0.00013013, -0.00014231, -0.00015070, -0.00015493, -0.00015486, -0.00015050, -0.00014207, -0.00012992, -0.00011457, -0.00009662, -0.00007674, -0.00005567, -0.00003413, -0.00001280, 0.00000766, 0.00002670, 0.00004388, 0.00005884, 0.00007137, 0.00008134, 0.00008877, 0.00009375, 0.00009646, 0.00009715, 0.00009609, 0.00009360, 0.00008999, 0.00008554, },
		{ 0.00003414, 0.00003242, 0.00003104, 0.00002995, 0.00002910, 0.00002848, 0.00002807, 0.00002787, 0.00002789, 0.00002813, 0.00002857, 0.00002919, 0.00002995, 0.00003078, 0.00003161, 0.00003236, 0.00003292, 0.00003321, 0.00003314, 0.00003264, 0.00003167, 0.00003021, 0.00002829, 0.00002597, 0.00002333, 0.00002049, 0.00001760, 0.00001479, 0.00001221, 0.00000999, 0.00000824, 0.00000699, 0.00000627, 0.00000599, 0.00000602, 0.00000618, 0.00000618, 0.00000572, 0.00000444, 0.00000197, -0.00000207, -0.00000803, -0.00001618, -0.00002673, -0.00003979, -0.00005534, -0.00007321, -0.00009310, -0.00011454, -0.00013695, -0.00015958, -0.00018163, -0.00020219, -0.00022033, -0.00023512, -0.00024571, -0.00025130, -0.00025127, -0.00024517, -0.00023274, -0.00021398, -0.00018914, -0.00015871, -0.00012344, -0.00008429, -0.00004245, 0.00000076, 0.00004390, 0.00008549, 0.00012406, 0.00015818, 0.00018659, 0.00020820, 0.00022213, 0.00022780, 0.00022492, 0.00021352, 0.00019395, 0.00016687, 0.00013319, 0.00009411, 0.00005100, 0.00000539, -0.00004114, -0.00008694, -0.00013047, -0.00017025, -0.00020498, -0.00023358, -0.00025521, -0.00026931, -0.00027562, -0.00027414, -0.00026518, -0.00024931, -0.00022730, -0.00020013, -0.00016889, -0.00013478, -0.00009902, -0.00006281, -0.00002730, 0.00000649, 0.00003766, 0.00006549, 0.00008944, 0.00010915, 0.00012448, 0.00013543, 0.00014220, 0.00014510, 0.00014457, 0.00014109, 0.00013525, 0.00012759, 0.00011869, 0.00010907, 0.00009919, 0.00008946, 0.00008019, 0.00007162, 0.00006390, 0.00005712, 0.00005129, 0.00004637, 0.00004230, 0.00003898, 0.00003630, },
		{ 0.00007498, 0.00007087, 0.00006673, 0.00006256, 0.00005838, 0.00005415, 0.00004987, 0.00004552, 0.00004107, 0.00003653, 0.00003189, 0.00002718, 0.00002245, 0.00001775, 0.00001316, 0.00000875, 0.00000463, 0.00000085, -0.00000249, -0.00000536, -0.00000772, -0.00000960, -0.00001104, -0.00001212, -0.00001295, -0.00001367, -0.00001444, -0.00001542, -0.00001676, -0.00001857, -0.00002095, -0.00002392, -0.00002743, -0.00003137, -0.00003553, -0.00003965, -0.00004335, -0.00004623, -0.00004782, -0.00004764, -0.00004521, -0.00004010, -0.00003194, -0.00002047, -0.00000554, 0.00001281, 0.00003440, 0.00005882, 0.00008551, 0.00011368, 0.00014239, 0.00017056, 0.00019698, 0.00022039, 0.00023950, 0.00025306, 0.00025991, 0.00025905, 0.00024969, 0.00023126, 0.00020353, 0.00016654, 0.00012070, 0.00006675, 0.00000575, -0.00006093, -0.00013166, -0.00020460, -0.00027775, -0.00034904, -0.00041639, -0.00047780, -0.00053139, -0.00057551, -0.00060878, -0.00063016, -0.00063896, -0.00063488, -0.00061804, -0.00058893, -0.00054842, -0.00049772, -0.00043832, -0.00037193, -0.00030045, -0.00022584, -0.00015009, -0.00007514, -0.00000279, 0.00006532, 0.00012777, 0.00018342, 0.00023142, 0.00027123, 0.00030259, 0.00032556, 0.00034044, 0.00034778, 0.00034832, 0.00034294, 0.00033262, 0.00031840, 0.00030132, 0.00028236, 0.00026246, 0.00024244, 0.00022298, 0.00020464, 0.00018782, 0.00017278, 0.00015965, 0.00014845, 0.00013908, 0.00013137, 0.00012510, 0.00012001, 0.00011584, 0.00011231, 0.00010920, 0.00010630, 0.00010344, 0.00010049, 0.00009739, 0.00009409, 0.00009058, 0.00008689, 0.00008303, 0.00007905, },
		{ -0.00000489, -0.00000719, -0.00000924, -0.00001106, -0.00001263, -0.00001395, -0.00001502, -0.00001581, -0.00001634, -0.00001661, -0.00001665, -0.00001651, -0.00001625, -0.00001594, -0.00001569, -0.00001559, -0.00001574, -0.00001622, -0.00001711, -0.00001845, -0.00002025, -0.00002247, -0.00002506, -0.00002789, -0.00003082, -0.00003369, -0.00003630, -0.00003848, -0.00004006, -0.00004093, -0.00004100, -0.00004029, -0.00003891, -0.00003704, -0.00003501, -0.00003324, -0.00003224, -0.00003263, -0.00003508, -0.00004029, -0.00004900, -0.00006187, -0.00007951, -0.00010240, -0.00013085, -0.00016500, -0.00020473, -0.00024968, -0.00029921, -0.00035242, -0.00040814, -0.00046496, -0.00052127, -0.00057529, -0.00062514, -0.00066892, -0.00070472, -0.00073078, -0.00074547, -0.00074744, -0.00073563, -0.00070934, -0.00066828, -0.00061258, -0.00054282, -0.00045999, -0.00036551, -0.00026116, -0.00014904, -0.00003150, 0.00008893, 0.00020963, 0.00032798, 0.00044143, 0.00054759, 0.00064433, 0.00072981, 0.00080254, 0.00086145, 0.00090587, 0.00093558, 0.00095077, 0.00095201, 0.00094025, 0.00091674, 0.00088297, 0.00084060, 0.00079143, 0.00073728, 0.00067996, 0.00062117, 0.00056250, 0.00050534, 0.00045085, 0.00039999, 0.00035343, 0.00031161, 0.00027475, 0.00024284, 0.00021567, 0.00019292, 0.00017411, 0.00015871, 0.00014615, 0.00013583, 0.00012720, 0.00011975, 0.00011303, 0.00010668, 0.00010042, 0.00009407, 0.00008751, 0.00008071, 0.00007368, 0.00006651, 0.00005927, 0.00005210, 0.00004510, 0.00003838, 0.00003203, 0.00002612, 0.00002069, 0.00001575, 0.00001129, 0.00000730, 0.00000372, 0.00000053, -0.00000233, },
		{ -0.00010526, -0.00010538, -0.00010575, -0.00010638, -0.00010729, -0.00010851, -0.00011006, -0.00011195, -0.00011419, -0.00011676, -0.00011963, -0.00012274, -0.00012602, -0.00012937, -0.00013269, -0.00013588, -0.00013883, -0.00014146, -0.00014371, -0.00014558, -0.00014707, -0.00014828, -0.00014934, -0.00015042, -0.00015175, -0.00015360, -0.00015623, -0.00015992, -0.00016491, -0.00017139, -0.00017948, -0.00018917, -0.00020035, -0.00021274, -0.00022590, -0.00023924, -0.00025200, -0.00026325, -0.00027196, -0.00027697, -0.00027707, -0.00027104, -0.00025769, -0.00023594, -0.00020484, -0.00016366, -0.00011192, -0.00004945, 0.00002360, 0.00010669, 0.00019894, 0.00029907, 0.00040549, 0.00051627, 0.00062924, 0.00074199, 0.00085203, 0.00095678, 0.00105372, 0.00114042, 0.00121469, 0.00127460, 0.00131857, 0.00134543, 0.00135446, 0.00134543, 0.00131857, 0.00127460, 0.00121469, 0.00114042, 0.00105372, 0.00095678, 0.00085203, 0.00074199, 0.00062924, 0.00051627, 0.00040549, 0.00029907, 0.00019894, 0.00010669, 0.00002360, -0.00004945, -0.00011192, -0.00016366, -0.00020484, -0.00023594, -0.00025769, -0.00027104, -0.00027707, -0.00027697, -0.00027196, -0.00026325, -0.00025200, -0.00023924, -0.00022590, -0.00021274, -0.00020035, -0.00018917, -0.00017948, -0.00017139, -0.00016491, -0.00015992, -0.00015623, -0.00015360, -0.00015175, -0.00015042, -0.00014934, -0.00014828, -0.00014707, -0.00014558, -0.00014371, -0.00014146, -0.00013883, -0.00013588, -0.00013269, -0.00012937, -0.00012602, -0.00012274, -0.00011963, -0.00011676, -0.00011419, -0.00011195, -0.00011006, -0.00010851, -0.00010729, -0.00010638, -0.00010575, -0.00010538, },
		{ -0.00000489, -0.00000233, 0.00000053, 0.00000372, 0.00000730, 0.00001129, 0.00001575, 0.00002069, 0.00002612, 0.00003203, 0.00003838, 0.00004510, 0.00005210, 0.00005927, 0.00006651, 0.00007368, 0.00008071, 0.00008751, 0.00009407, 0.00010042, 0.00010668, 0.00011303, 0.00011975, 0.00012720, 0.00013583, 0.00014615, 0.00015871, 0.00017411, 0.00019292, 0.00021567, 0.00024284, 0.00027475, 0.00031161, 0.00035343, 0.00039999, 0.00045085, 0.00050534, 0.00056250, 0.00062117, 0.00067996, 0.00073728, 0.00079143, 0.00084060, 0.00088297, 0.00091674, 0.00094025, 0.00095201, 0.00095077, 0.00093558, 0.00090587, 0.00086145, 0.00080254, 0.00072981, 0.00064433, 0.00054759, 0.00044143, 0.00032798, 0.00020963, 0.00008893, -0.00003150, -0.00014904, -0.00026116, -0.00036551, -0.00045999, -0.00054282, -0.00061258, -0.00066828, -0.00070934, -0.00073563, -0.00074744, -0.00074547, -0.00073078, -0.00070472, -0.00066892, -0.00062514, -0.00057529, -0.00052127, -0.00046496, -0.00040814, -0.00035242, -0.00029921, -0.00024968, -0.00020473, -0.00016500, -0.00013085, -0.00010240, -0.00007951, -0.00006187, -0.00004900, -0.00004029, -0.00003508, -0.00003263, -0.00003224, -0.00003324, -0.00003501, -0.00003704, -0.00003891, -0.00004029, -0.00004100, -0.00004093, -0.00004006, -0.00003848, -0.00003630, -0.00003369, -0.00003082, -0.00002789, -0.00002506, -0.00002247, -0.00002025, -0.00001845, -0.00001711, -0.00001622, -0.00001574, -0.00001559, -0.00001569, -0.00001594, -0.00001625, -0.00001651, -0.00001665, -0.00001661, -0.00001634, -0.00001581, -0.00001502, -0.00001395, -0.00001263, -0.00001106, -0.00000924, -0.00000719, },
		{ 0.00007498, 0.00007905, 0.00008303, 0.00008689, 0.00009058, 0.00009409, 0.00009739, 0.00010049, 0.00010344, 0.00010630, 0.00010920, 0.00011231, 0.00011584, 0.00012001, 0.00012510, 0.00013137, 0.00013908, 0.00014845, 0.00015965, 0.00017278, 0.00018782, 0.00020464, 0.00022298, 0.00024244, 0.00026246, 0.00028236, 0.00030132, 0.00031840, 0.00033262, 0.00034294, 0.00034832, 0.00034778, 0.00034044, 0.00032556, 0.00030259, 0.00027123, 0.00023142, 0.00018342, 0.00012777, 0.00006532, -0.00000279, -0.00007514, -0.00015009, -0.00022584, -0.00030045, -0.00037193, -0.00043832, -0.00049772, -0.00054842, -0.00058893, -0.00061804, -0.00063488, -0.00063896, -0.00063016, -0.00060878, -0.00057551, -0.00053139, -0.00047780, -0.00041639, -0.00034904, -0.00027775, -0.00020460, -0.00013166, -0.00006093, 0.00000575, 0.00006675, 0.00012070, 0.00016654, 0.00020353, 0.00023126, 0.00024969, 0.00025905, 0.00025991, 0.00025306, 0.00023950, 0.00022039, 0.00019698, 0.00017056, 0.00014239, 0.00011368, 0.00008551, 0.00005882, 0.00003440, 0.00001281, -0.00000554, -0.00002047, -0.00003194, -0.00004010, -0.00004521, -0.00004764, -0.00004782, -0.00004623, -0.00004335, -0.00003965, -0.00003553, -0.00003137, -0.00002743, -0.00002392, -0.00002095, -0.00001857, -0.00001676, -0.00001542, -0.00001444, -0.00001367, -0.00001295, -0.00001212, -0.00001104, -0.00000960, -0.00000772, -0.00000536, -0.00000249, 0.00000085, 0.00000463, 0.00000875, 0.00001316, 0.00001775, 0.00002245, 0.00002718, 0.00003189, 0.00003653, 0.00004107, 0.00004552, 0.00004987, 0.00005415, 0.00005838, 0.00006256, 0.00006673, 0.00007087, },
		{ 0.00003414, 0.00003630, 0.00003898, 0.00004230, 0.00004637, 0.00005129, 0.00005712, 0.00006390, 0.00007162, 0.00008019, 0.00008946, 0.00009919, 0.00010907, 0.00011869, 0.00012759, 0.00013525, 0.00014109, 0.00014457, 0.00014510, 0.00014220, 0.00013543, 0.00012448, 0.00010915, 0.00008944, 0.00006549, 0.00003766, 0.00000649, -0.00002730, -0.00006281, -0.00009902, -0.00013478, -0.00016889, -0.00020013, -0.00022730, -0.00024931, -0.00026518, -0.00027414, -0.00027562, -0.00026931, -0.00025521, -0.00023358, -0.00020498, -0.00017025, -0.00013047, -0.00008694, -0.00004114, 0.00000539, 0.00005100, 0.00009411, 0.00013319, 0.00016687, 0.00019395, 0.00021352, 0.00022492, 0.00022780, 0.00022213, 0.00020820, 0.00018659, 0.00015818, 0.00012406, 0.00008549, 0.00004390, 0.00000076, -0.00004245, -0.00008429, -0.00012344, -0.00015871, -0.00018914, -0.00021398, -0.00023274, -0.00024517, -0.00025127, -0.00025130, -0.00024571, -0.00023512, -0.00022033, -0.00020219, -0.00018163, -0.00015958, -0.00013695, -0.00011454, -0.00009310, -0.00007321, -0.00005534, -0.00003979, -0.00002673, -0.00001618, -0.00000803, -0.00000207, 0.00000197, 0.00000444, 0.00000572, 0.00000618, 0.00000618, 0.00000602, 0.00000599, 0.00000627, 0.00000699, 0.00000824, 0.00000999, 0.00001221, 0.00001479, 0.00001760, 0.00002049, 0.00002333, 0.00002597, 0.00002829, 0.00003021, 0.00003167, 0.00003264, 0.00003314, 0.00003321, 0.00003292, 0.00003236, 0.00003161, 0.00003078, 0.00002995, 0.00002919, 0.00002857, 0.00002813, 0.00002789, 0.00002787, 0.00002807, 0.00002848, 0.00002910, 0.00002995, 0.00003104, 0.00003242, },
		{ 0.00008053, 0.00008554, 0.00008999, 0.00009360, 0.00009609, 0.00009715, 0.00009646, 0.00009375, 0.00008877, 0.00008134, 0.00007137, 0.00005884, 0.00004388, 0.00002670, 0.00000766, -0.00001280, -0.00003413, -0.00005567, -0.00007674, -0.00009662, -0.00011457, -0.00012992, -0.00014207, -0.00015050, -0.00015486, -0.00015493, -0.00015070, -0.00014231, -0.00013013, -0.00011469, -0.00009666, -0.00007689, -0.00005629, -0.00003587, -0.00001662, 0.00000047, 0.00001451, 0.00002473, 0.00003050, 0.00003140, 0.00002722, 0.00001800, 0.00000399, -0.00001431, -0.00003621, -0.00006082, -0.00008713, -0.00011400, -0.00014027, -0.00016477, -0.00018638, -0.00020411, -0.00021709, -0.00022466, -0.00022638, -0.00022204, -0.00021166, -0.00019554, -0.00017419, -0.00014830, -0.00011877, -0.00008660, -0.00005290, -0.00001878, 0.00001463, 0.00004628, 0.00007523, 0.00010067, 0.00012196, 0.00013866, 0.00015052, 0.00015750, 0.00015976, 0.00015761, 0.00015154, 0.00014214, 0.00013010, 0.00011615, 0.00010102, 0.00008545, 0.00007009, 0.00005551, 0.00004221, 0.00003052, 0.00002070, 0.00001285, 0.00000696, 0.00000293, 0.00000057, -0.00000040, -0.00000027, 0.00000063, 0.00000198, 0.00000349, 0.00000491, 0.00000601, 0.00000667, 0.00000681, 0.00000640, 0.00000549, 0.00000417, 0.00000258, 0.00000086, -0.00000082, -0.00000229, -0.00000342, -0.00000407, -0.00000414, -0.00000360, -0.00000240, -0.00000057, 0.00000185, 0.00000480, 0.00000819, 0.00001195, 0.00001599, 0.00002024, 0.00002465, 0.00002917, 0.00003380, 0.00003853, 0.00004337, 0.00004836, 0.00005349, 0.00005878, 0.00006420, 0.00006970, 0.00007519, },
		{ 0.00001040, -0.00000017, -0.00001176, -0.00002414, -0.00003699, -0.00004996, -0.00006265, -0.00007462, -0.00008544, -0.00009468, -0.00010197, -0.00010699, -0.00010951, -0.00010940, -0.00010664, -0.00010135, -0.00009373, -0.00008413, -0.00007299, -0.00006082, -0.00004821, -0.00003574, -0.00002404, -0.00001365, -0.00000508, 0.00000126, 0.00000508, 0.00000624, 0.00000474, 0.00000073, -0.00000550, -0.00001350, -0.00002272, -0.00003251, -0.00004218, -0.00005101, -0.00005829, -0.00006338, -0.00006572, -0.00006488, -0.00006060, -0.00005275, -0.00004141, -0.00002684, -0.00000947, 0.00001010, 0.00003117, 0.00005290, 0.00007441, 0.00009480, 0.00011317, 0.00012871, 0.00014069, 0.00014855, 0.00015186, 0.00015040, 0.00014415, 0.00013326, 0.00011811, 0.00009923, 0.00007729, 0.00005312, 0.00002757, 0.00000158, -0.00002394, -0.00004812, -0.00007015, -0.00008938, -0.00010524, -0.00011737, -0.00012556, -0.00012976, -0.00013010, -0.00012686, -0.00012046, -0.00011139, -0.00010026, -0.00008769, -0.00007431, -0.00006074, -0.00004754, -0.00003520, -0.00002410, -0.00001454, -0.00000667, -0.00000055, 0.00000385, 0.00000668, 0.00000817, 0.00000858, 0.00000823, 0.00000744, 0.00000652, 0.00000576, 0.00000540, 0.00000563, 0.00000658, 0.00000830, 0.00001079, 0.00001398, 0.00001777, 0.00002200, 0.00002649, 0.00003108, 0.00003559, 0.00003985, 0.00004374, 0.00004716, 0.00005005, 0.00005239, 0.00005419, 0.00005550, 0.00005637, 0.00005687, 0.00005707, 0.00005703, 0.00005677, 0.00005631, 0.00005560, 0.00005458, 0.00005315, 0.00005118, 0.00004851, 0.00004499, 0.00004048, 0.00003483, 0.00002796, 0.00001981, },
		{ -0.00007370, -0.00007352, -0.00007184, -0.00006866, -0.00006408, -0.00005826, -0.00005144, -0.00004391, -0.00003603, -0.00002816, -0.00002068, -0.00001395, -0.00000830, -0.00000398, -0.00000120, -0.00000004, -0.00000052, -0.00000255, -0.00000594, -0.00001042, -0.00001564, -0.00002122, -0.00002672, -0.00003174, -0.00003585, -0.00003873, -0.00004007, -0.00003970, -0.00003754, -0.00003362, -0.00002810, -0.00002124, -0.00001342, -0.00000508, 0.00000326, 0.00001105, 0.00001774, 0.00002282, 0.00002581, 0.00002636, 0.00002420, 0.00001920, 0.00001138, 0.00000092, -0.00001188, -0.00002656, -0.00004255, -0.00005919, -0.00007577, -0.00009153, -0.00010572, -0.00011765, -0.00012671, -0.00013238, -0.00013431, -0.00013227, -0.00012623, -0.00011630, -0.00010280, -0.00008616, -0.00006698, -0.00004593, -0.00002378, -0.00000133, 0.00002062, 0.00004130, 0.00006002, 0.00007619, 0.00008932, 0.00009911, 0.00010537, 0.00010808, 0.00010738, 0.00010355, 0.00009698, 0.00008816, 0.00007765, 0.00006604, 0.00005393, 0.00004190, 0.00003048, 0.00002012, 0.00001118, 0.00000391, -0.00000155, -0.00000516, -0.00000700, -0.00000723, -0.00000609, -0.00000387, -0.00000091, 0.00000246, 0.00000590, 0.00000912, 0.00001184, 0.00001389, 0.00001513, 0.00001549, 0.00001498, 0.00001367, 0.00001167, 0.00000913, 0.00000623, 0.00000314, 0.00000005, -0.00000290, -0.00000559, -0.00000796, -0.00000998, -0.00001167, -0.00001311, -0.00001441, -0.00001569, -0.00001711, -0.00001882, -0.00002095, -0.00002363, -0.00002691, -0.00003081, -0.00003531, -0.00004032, -0.00004567, -0.00005119, -0.00005664, -0.00006175, -0.00006626, -0.00006991, -0.00007245, },
		{ -0.00000807, -0.00000539, -0.00000350, -0.00000251, -0.00000247, -0.00000339, -0.00000518, -0.00000770, -0.00001078, -0.00001417, -0.00001761, -0.00002085, -0.00002363, -0.00002571, -0.00002692, -0.00002712, -0.00002626, -0.00002435, -0.00002148, -0.00001781, -0.00001357, -0.00000904, -0.00000451, -0.00000032, 0.00000323, 0.00000584, 0.00000729, 0.00000741, 0.00000611, 0.00000339, -0.00000063, -0.00000577, -0.00001174, -0.00001818, -0.00002466, -0.00003076, -0.00003599, -0.00003992, -0.00004216, -0.00004238, -0.00004033, -0.00003591, -0.00002911, -0.00002005, -0.00000901, 0.00000365, 0.00001742, 0.00003175, 0.00004599, 0.00005948, 0.00007156, 0.00008160, 0.00008905, 0.00009344, 0.00009444, 0.00009185, 0.00008561, 0.00007584, 0.00006281, 0.00004693, 0.00002873, 0.00000885, -0.00001199, -0.00003304, -0.00005355, -0.00007279, -0.00009009, -0.00010488, -0.00011671, -0.00012524, -0.00013032, -0.00013191, -0.00013015, -0.00012529, -0.00011771, -0.00010791, -0.00009642, -0.00008384, -0.00007077, -0.00005778, -0.00004542, -0.00003415, -0.00002433, -0.00001624, -0.00001002, -0.00000572, -0.00000327, -0.00000252, -0.00000322, -0.00000507, -0.00000773, -0.00001085, -0.00001408, -0.00001711, -0.00001968, -0.00002156, -0.00002264, -0.00002285, -0.00002220, -0.00002080, -0.00001879, -0.00001635, -0.00001372, -0.00001114, -0.00000883, -0.00000701, -0.00000584, -0.00000545, -0.00000590, -0.00000718, -0.00000924, -0.00001195, -0.00001515, -0.00001861, -0.00002213, -0.00002545, -0.00002835, -0.00003062, -0.00003212, -0.00003272, -0.00003237, -0.00003108, -0.00002894, -0.00002606, -0.00002264, -0.00001888, -0.00001504, -0.00001135, },
		{ -0.00001371, -0.00001558, -0.00001710, -0.00001812, -0.00001858, -0.00001842, -0.00001767, -0.00001639, -0.00001469, -0.00001271, -0.00001065, -0.00000869, -0.00000703, -0.00000585, -0.00000529, -0.00000547, -0.00000645, -0.00000822, -0.00001073, -0.00001386, -0.00001743, -0.00002123, -0.00002501, -0.00002850, -0.00003144, -0.00003359, -0.00003475, -0.00003477, -0.00003357, -0.00003115, -0.00002760, -0.00002308, -0.00001785, -0.00001221, -0.00000655, -0.00000125, 0.00000326, 0.00000657, 0.00000832, 0.00000820, 0.00000599, 0.00000154, -0.00000514, -0.00001395, -0.00002467, -0.00003695, -0.00005034, -0.00006431, -0.00007827, -0.00009157, -0.00010360, -0.00011372, -0.00012138, -0.00012613, -0.00012759, -0.00012553, -0.00011988, -0.00011070, -0.00009821, -0.00008277, -0.00006488, -0.00004513, -0.00002419, -0.00000281, 0.00001829, 0.00003837, 0.00005675, 0.00007285, 0.00008617, 0.00009635, 0.00010318, 0.00010658, 0.00010664, 0.00010357, 0.00009771, 0.00008951, 0.00007950, 0.00006826, 0.00005638, 0.00004447, 0.00003306, 0.00002264, 0.00001361, 0.00000628, 0.00000082, -0.00000269, -0.00000431, -0.00000418, -0.00000254, 0.00000029, 0.00000395, 0.00000805, 0.00001220, 0.00001603, 0.00001921, 0.00002148, 0.00002264, 0.00002258, 0.00002128, 0.00001878, 0.00001523, 0.00001082, 0.00000580, 0.00000044, -0.00000497, -0.00001014, -0.00001480, -0.00001874, -0.00002180, -0.00002385, -0.00002487, -0.00002487, -0.00002394, -0.00002222, -0.00001990, -0.00001719, -0.00001431, -0.00001150, -0.00000895, -0.00000685, -0.00000533, -0.00000448, -0.00000433, -0.00000485, -0.00000598, -0.00000759, -0.00000953, -0.00001163, },
		{ -0.00001395, -0.00001273, -0.00001167, -0.00001087, -0.00001040, -0.00001031, -0.00001061, -0.00001124, -0.00001213, -0.00001318, -0.00001425, -0.00001517, -0.00001581, -0.00001602, -0.00001568, -0.00001469, -0.00001301, -0.00001066, -0.00000768, -0.00000419, -0.00000035, 0.00000364, 0.00000756, 0.00001114, 0.00001416, 0.00001638, 0.00001759, 0.00001767, 0.00001652, 0.00001414, 0.00001059, 0.00000603, 0.00000068, -0.00000516, -0.00001114, -0.00001685, -0.00002188, -0.00002583, -0.00002831, -0.00002899, -0.00002760, -0.00002397, -0.00001805, -0.00000987, 0.00000039, 0.00001244, 0.00002589, 0.00004026, 0.00005498, 0.00006945, 0.00008304, 0.00009513, 0.00010513, 0.00011255, 0.00011697, 0.00011808, 0.00011574, 0.00010992, 0.00010075, 0.00008851, 0.00007361, 0.00005657, 0.00003802, 0.00001864, -0.00000086, -0.00001977, -0.00003738, -0.00005309, -0.00006635, -0.00007676, -0.00008403, -0.00008802, -0.00008872, -0.00008629, -0.00008100, -0.00007326, -0.00006354, -0.00005241, -0.00004046, -0.00002829, -0.00001648, -0.00000556, 0.00000400, 0.00001184, 0.00001769, 0.00002142, 0.00002300, 0.00002250, 0.00002013, 0.00001614, 0.00001089, 0.00000476, -0.00000184, -0.00000850, -0.00001483, -0.00002049, -0.00002521, -0.00002877, -0.00003106, -0.00003202, -0.00003170, -0.00003021, -0.00002773, -0.00002448, -0.00002071, -0.00001671, -0.00001274, -0.00000904, -0.00000583, -0.00000329, -0.00000151, -0.00000057, -0.00000045, -0.00000110, -0.00000242, -0.00000425, -0.00000644, -0.00000880, -0.00001115, -0.00001332, -0.00001516, -0.00001658, -0.00001750, -0.00001790, -0.00001779, -0.00001724, -0.00001633, -0.00001519, },
		{ 0.00000397, 0.00000426, 0.00000464, 0.00000520, 0.00000600, 0.00000704, 0.00000832, 0.00000981, 0.00001141, 0.00001302, 0.00001451, 0.00001574, 0.00001655, 0.00001683, 0.00001644, 0.00001530, 0.00001338, 0.00001067, 0.00000724, 0.00000319, -0.00000131, -0.00000606, -0.00001083, -0.00001535, -0.00001936, -0.00002260, -0.00002486, -0.00002593, -0.00002571, -0.00002412, -0.00002119, -0.00001702, -0.00001179, -0.00000576, 0.00000075, 0.00000737, 0.00001369, 0.00001929, 0.00002377, 0.00002675, 0.00002792, 0.00002704, 0.00002395, 0.00001863, 0.00001114, 0.00000167, -0.00000947, -0.00002190, -0.00003511, -0.00004858, -0.00006170, -0.00007387, -0.00008453, -0.00009312, -0.00009920, -0.00010239, -0.00010244, -0.00009924, -0.00009282, -0.00008332, -0.00007106, -0.00005645, -0.00004002, -0.00002237, -0.00000416, 0.00001392, 0.00003120, 0.00004705, 0.00006088, 0.00007223, 0.00008072, 0.00008613, 0.00008834, 0.00008740, 0.00008347, 0.00007685, 0.00006793, 0.00005719, 0.00004517, 0.00003244, 0.00001957, 0.00000710, -0.00000447, -0.00001470, -0.00002328, -0.00002995, -0.00003460, -0.00003719, -0.00003782, -0.00003665, -0.00003395, -0.00003002, -0.00002522, -0.00001991, -0.00001447, -0.00000924, -0.00000453, -0.00000057, 0.00000243, 0.00000439, 0.00000525, 0.00000506, 0.00000391, 0.00000195, -0.00000062, -0.00000358, -0.00000671, -0.00000978, -0.00001258, -0.00001494, -0.00001673, -0.00001785, -0.00001826, -0.00001798, -0.00001704, -0.00001556, -0.00001363, -0.00001141, -0.00000904, -0.00000664, -0.00000437, -0.00000231, -0.00000053, 0.00000091, 0.00000201, 0.00000280, 0.00000334, 0.00000370, },
		{ -0.00000234, -0.00000327, -0.00000446, -0.00000598, -0.00000787, -0.00001013, -0.00001275, -0.00001566, -0.00001877, -0.00002195, -0.00002506, -0.00002793, -0.00003038, -0.00003225, -0.00003339, -0.00003368, -0.00003304, -0.00003142, -0.00002886, -0.00002542, -0.00002123, -0.00001648, -0.00001139, -0.00000621, -0.00000122, 0.00000331, 0.00000711, 0.00000994, 0.00001163, 0.00001204, 0.00001111, 0.00000885, 0.00000537, 0.00000083, -0.00000452, -0.00001036, -0.00001635, -0.00002210, -0.00002721, -0.00003131, -0.00003405, -0.00003512, -0.00003433, -0.00003153, -0.00002669, -0.00001989, -0.00001131, -0.00000124, 0.00000996, 0.00002183, 0.00003387, 0.00004554, 0.00005631, 0.00006564, 0.00007307, 0.00007817, 0.00008065, 0.00008029, 0.00007700, 0.00007082, 0.00006193, 0.00005059, 0.00003722, 0.00002228, 0.00000633, -0.00001003, -0.00002620, -0.00004157, -0.00005560, -0.00006778, -0.00007771, -0.00008508, -0.00008971, -0.00009153, -0.00009058, -0.00008705, -0.00008121, -0.00007341, -0.00006409, -0.00005373, -0.00004283, -0.00003187, -0.00002133, -0.00001163, -0.00000312, 0.00000393, 0.00000933, 0.00001300, 0.00001494, 0.00001524, 0.00001408, 0.00001167, 0.00000832, 0.00000432, -0.00000001, -0.00000436, -0.00000843, -0.00001198, -0.00001482, -0.00001680, -0.00001785, -0.00001795, -0.00001716, -0.00001556, -0.00001330, -0.00001055, -0.00000749, -0.00000431, -0.00000121, 0.00000166, 0.00000417, 0.00000620, 0.00000769, 0.00000863, 0.00000903, 0.00000894, 0.00000843, 0.00000760, 0.00000655, 0.00000538, 0.00000418, 0.00000304, 0.00000201, 0.00000110, 0.00000033, -0.00000034, -0.00000096, -0.00000160, },
		{ -0.00001894, -0.00001958, -0.00001983, -0.00001959, -0.00001878, -0.00001738, -0.00001538, -0.00001281, -0.00000976, -0.00000633, -0.00000268, 0.00000102, 0.00000460, 0.00000784, 0.00001057, 0.00001261, 0.00001383, 0.00001414, 0.00001347, 0.00001184, 0.00000930, 0.00000597, 0.00000201, -0.00000237, -0.00000693, -0.00001141, -0.00001556, -0.00001911, -0.00002186, -0.00002360, -0.00002422, -0.00002364, -0.00002188, -0.00001900, -0.00001515, -0.00001054, -0.00000545, -0.00000018, 0.00000494, 0.00000956, 0.00001335, 0.00001602, 0.00001730, 0.00001701, 0.00001504, 0.00001137, 0.00000607, -0.00000070, -0.00000869, -0.00001759, -0.00002700, -0.00003651, -0.00004564, -0.00005395, -0.00006102, -0.00006643, -0.00006988, -0.00007111, -0.00006999, -0.00006645, -0.00006057, -0.00005251, -0.00004254, -0.00003101, -0.00001835, -0.00000504, 0.00000842, 0.00002151, 0.00003374, 0.00004467, 0.00005389, 0.00006110, 0.00006608, 0.00006870, 0.00006897, 0.00006695, 0.00006284, 0.00005691, 0.00004949, 0.00004096, 0.00003176, 0.00002229, 0.00001299, 0.00000424, -0.00000363, -0.00001034, -0.00001570, -0.00001957, -0.00002192, -0.00002277, -0.00002225, -0.00002052, -0.00001779, -0.00001432, -0.00001038, -0.00000624, -0.00000217, 0.00000161, 0.00000490, 0.00000755, 0.00000946, 0.00001058, 0.00001092, 0.00001053, 0.00000949, 0.00000792, 0.00000597, 0.00000378, 0.00000149, -0.00000075, -0.00000286, -0.00000472, -0.00000630, -0.00000757, -0.00000853, -0.00000924, -0.00000973, -0.00001010, -0.00001041, -0.00001074, -0.00001117, -0.00001174, -0.00001249, -0.00001342, -0.00001450, -0.00001567, -0.00001687, -0.00001800, },
		{ -0.00000138, 0.00000061, 0.00000237, 0.00000376, 0.00000468, 0.00000504, 0.00000477, 0.00000386, 0.00000233, 0.00000023, -0.00000232, -0.00000520, -0.00000825, -0.00001129, -0.00001414, -0.00001660, -0.00001853, -0.00001976, -0.00002021, -0.00001981, -0.00001854, -0.00001644, -0.00001360, -0.00001014, -0.00000625, -0.00000211, 0.00000205, 0.00000602, 0.00000956, 0.00001250, 0.00001467, 0.00001595, 0.00001629, 0.00001567, 0.00001416, 0.00001186, 0.00000895, 0.00000564, 0.00000217, -0.00000119, -0.00000417, -0.00000651, -0.00000798, -0.00000839, -0.00000760, -0.00000555, -0.00000224, 0.00000225, 0.00000778, 0.00001413, 0.00002102, 0.00002813, 0.00003511, 0.00004159, 0.00004722, 0.00005165, 0.00005461, 0.00005586, 0.00005523, 0.00005266, 0.00004815, 0.00004180, 0.00003380, 0.00002440, 0.00001394, 0.00000280, -0.00000861, -0.00001986, -0.00003053, -0.00004021, -0.00004857, -0.00005529, -0.00006016, -0.00006304, -0.00006388, -0.00006271, -0.00005966, -0.00005491, -0.00004873, -0.00004144, -0.00003337, -0.00002490, -0.00001638, -0.00000817, -0.00000056, 0.00000617, 0.00001181, 0.00001624, 0.00001937, 0.00002119, 0.00002177, 0.00002122, 0.00001968, 0.00001737, 0.00001448, 0.00001125, 0.00000788, 0.00000458, 0.00000151, -0.00000118, -0.00000341, -0.00000513, -0.00000632, -0.00000702, -0.00000728, -0.00000720, -0.00000688, -0.00000644, -0.00000599, -0.00000562, -0.00000541, -0.00000543, -0.00000569, -0.00000620, -0.00000693, -0.00000781, -0.00000877, -0.00000972, -0.00001055, -0.00001118, -0.00001152, -0.00001149, -0.00001106, -0.00001021, -0.00000896, -0.00000737, -0.00000549, -0.00000346, },
		{ -0.00000509, -0.00000651, -0.00000796, -0.00000929, -0.00001039, -0.00001112, -0.00001141, -0.00001117, -0.00001039, -0.00000906, -0.00000722, -0.00000494, -0.00000234, 0.00000046, 0.00000330, 0.00000604, 0.00000852, 0.00001059, 0.00001213, 0.00001304, 0.00001328, 0.00001281, 0.00001165, 0.00000988, 0.00000759, 0.00000491, 0.00000199, -0.00000100, -0.00000387, -0.00000647, -0.00000865, -0.00001028, -0.00001128, -0.00001160, -0.00001125, -0.00001027, -0.00000876, -0.00000686, -0.00000473, -0.00000255, -0.00000054, 0.00000110, 0.00000219, 0.00000257, 0.00000211, 0.00000072, -0.00000160, -0.00000484, -0.00000890, -0.00001362, -0.00001882, -0.00002426, -0.00002965, -0.00003472, -0.00003918, -0.00004275, -0.00004518, -0.00004627, -0.00004587, -0.00004389, -0.00004031, -0.00003520, -0.00002867, -0.00002092, -0.00001220, -0.00000282, 0.00000690, 0.00001659, 0.00002590, 0.00003450, 0.00004207, 0.00004834, 0.00005310, 0.00005621, 0.00005760, 0.00005726, 0.00005526, 0.00005175, 0.00004691, 0.00004100, 0.00003429, 0.00002707, 0.00001964, 0.00001230, 0.00000531, -0.00000109, -0.00000672, -0.00001145, -0.00001519, -0.00001794, -0.00001971, -0.00002059, -0.00002070, -0.00002019, -0.00001923, -0.00001800, -0.00001666, -0.00001539, -0.00001431, -0.00001353, -0.00001312, -0.00001311, -0.00001348, -0.00001421, -0.00001521, -0.00001638, -0.00001762, -0.00001880, -0.00001980, -0.00002053, -0.00002090, -0.00002084, -0.00002032, -0.00001936, -0.00001797, -0.00001623, -0.00001422, -0.00001206, -0.00000986, -0.00000774, -0.00000583, -0.00000422, -0.00000302, -0.00000226, -0.00000199, -0.00000219, -0.00000282, -0.00000382, },
		{ 0.00000164, 0.00000270, 0.00000389, 0.00000511, 0.00000625, 0.00000719, 0.00000784, 0.00000813, 0.00000800, 0.00000742, 0.00000639, 0.00000494, 0.00000314, 0.00000107, -0.00000117, -0.00000345, -0.00000565, -0.00000765, -0.00000932, -0.00001057, -0.00001133, -0.00001155, -0.00001121, -0.00001033, -0.00000896, -0.00000717, -0.00000508, -0.00000279, -0.00000045, 0.00000181, 0.00000388, 0.00000564, 0.00000699, 0.00000788, 0.00000828, 0.00000820, 0.00000769, 0.00000681, 0.00000568, 0.00000442, 0.00000318, 0.00000210, 0.00000133, 0.00000100, 0.00000121, 0.00000203, 0.00000351, 0.00000564, 0.00000836, 0.00001157, 0.00001514, 0.00001890, 0.00002264, 0.00002616, 0.00002922, 0.00003161, 0.00003314, 0.00003364, 0.00003297, 0.00003107, 0.00002789, 0.00002348, 0.00001791, 0.00001134, 0.00000396, -0.00000401, -0.00001229, -0.00002062, -0.00002869, -0.00003623, -0.00004298, -0.00004873, -0.00005329, -0.00005654, -0.00005840, -0.00005887, -0.00005800, -0.00005589, -0.00005270, -0.00004862, -0.00004389, -0.00003873, -0.00003342, -0.00002817, -0.00002323, -0.00001878, -0.00001497, -0.00001191, -0.00000966, -0.00000823, -0.00000759, -0.00000766, -0.00000830, -0.00000939, -0.00001075, -0.00001221, -0.00001361, -0.00001479, -0.00001561, -0.00001597, -0.00001581, -0.00001508, -0.00001380, -0.00001201, -0.00000979, -0.00000724, -0.00000449, -0.00000167, 0.00000108, 0.00000363, 0.00000586, 0.00000768, 0.00000902, 0.00000984, 0.00001013, 0.00000993, 0.00000929, 0.00000829, 0.00000703, 0.00000562, 0.00000419, 0.00000284, 0.00000168, 0.00000080, 0.00000025, 0.00000006, 0.00000026, 0.00000080, },
		{ -0.00000126, -0.00000255, -0.00000396, -0.00000542, -0.00000683, -0.00000810, -0.00000914, -0.00000988, -0.00001025, -0.00001021, -0.00000974, -0.00000885, -0.00000757, -0.00000596, -0.00000409, -0.00000206, 0.00000003, 0.00000208, 0.00000397, 0.00000560, 0.00000690, 0.00000779, 0.00000822, 0.00000819, 0.00000770, 0.00000677, 0.00000547, 0.00000386, 0.00000205, 0.00000012, -0.00000184, -0.00000371, -0.00000543, -0.00000693, -0.00000815, -0.00000909, -0.00000974, -0.00001013, -0.00001032, -0.00001037, -0.00001037, -0.00001041, -0.00001059, -0.00001098, -0.00001166, -0.00001268, -0.00001408, -0.00001584, -0.00001794, -0.00002031, -0.00002286, -0.00002546, -0.00002798, -0.00003026, -0.00003214, -0.00003347, -0.00003410, -0.00003391, -0.00003282, -0.00003077, -0.00002774, -0.00002377, -0.00001894, -0.00001336, -0.00000720, -0.00000065, 0.00000609, 0.00001277, 0.00001917, 0.00002505, 0.00003020, 0.00003446, 0.00003766, 0.00003972, 0.00004058, 0.00004024, 0.00003875, 0.00003622, 0.00003279, 0.00002865, 0.00002399, 0.00001905, 0.00001406, 0.00000925, 0.00000482, 0.00000096, -0.00000218, -0.00000449, -0.00000591, -0.00000643, -0.00000606, -0.00000489, -0.00000302, -0.00000059, 0.00000223, 0.00000527, 0.00000835, 0.00001130, 0.00001396, 0.00001620, 0.00001792, 0.00001905, 0.00001955, 0.00001943, 0.00001873, 0.00001751, 0.00001588, 0.00001392, 0.00001178, 0.00000957, 0.00000741, 0.00000540, 0.00000364, 0.00000218, 0.00000108, 0.00000033, -0.00000007, -0.00000016, -0.00000001, 0.00000032, 0.00000073, 0.00000115, 0.00000147, 0.00000164, 0.00000159, 0.00000129, 0.00000070, -0.00000016, },
		{ -0.00000581, -0.00000412, -0.00000221, -0.00000016, 0.00000196, 0.00000404, 0.00000600, 0.00000776, 0.00000922, 0.00001033, 0.00001105, 0.00001134, 0.00001121, 0.00001067, 0.00000976, 0.00000855, 0.00000711, 0.00000551, 0.00000386, 0.00000224, 0.00000074, -0.00000057, -0.00000161, -0.00000234, -0.00000273, -0.00000276, -0.00000244, -0.00000180, -0.00000087, 0.00000030, 0.00000165, 0.00000312, 0.00000466, 0.00000621, 0.00000773, 0.00000919, 0.00001058, 0.00001189, 0.00001313, 0.00001433, 0.00001553, 0.00001675, 0.00001805, 0.00001945, 0.00002099, 0.00002268, 0.00002453, 0.00002651, 0.00002859, 0.00003070, 0.00003278, 0.00003471, 0.00003640, 0.00003772, 0.00003856, 0.00003881, 0.00003837, 0.00003716, 0.00003512, 0.00003222, 0.00002847, 0.00002393, 0.00001866, 0.00001279, 0.00000647, -0.00000013, -0.00000681, -0.00001337, -0.00001958, -0.00002525, -0.00003019, -0.00003422, -0.00003721, -0.00003906, -0.00003972, -0.00003917, -0.00003746, -0.00003467, -0.00003091, -0.00002635, -0.00002118, -0.00001560, -0.00000983, -0.00000409, 0.00000141, 0.00000647, 0.00001094, 0.00001468, 0.00001759, 0.00001961, 0.00002073, 0.00002098, 0.00002041, 0.00001912, 0.00001724, 0.00001488, 0.00001221, 0.00000938, 0.00000652, 0.00000377, 0.00000124, -0.00000098, -0.00000284, -0.00000428, -0.00000532, -0.00000597, -0.00000628, -0.00000630, -0.00000612, -0.00000580, -0.00000544, -0.00000510, -0.00000486, -0.00000476, -0.00000484, -0.00000510, -0.00000553, -0.00000612, -0.00000681, -0.00000754, -0.00000825, -0.00000886, -0.00000930, -0.00000952, -0.00000944, -0.00000904, -0.00000830, -0.00000722, },
		{ 0.00001644, 0.00001593, 0.00001504, 0.00001382, 0.00001235, 0.00001070, 0.00000895, 0.00000719, 0.00000551, 0.00000399, 0.00000269, 0.00000168, 0.00000099, 0.00000065, 0.00000066, 0.00000100, 0.00000165, 0.00000257, 0.00000369, 0.00000496, 0.00000631, 0.00000766, 0.00000897, 0.00001016, 0.00001118, 0.00001200, 0.00001258, 0.00001291, 0.00001299, 0.00001280, 0.00001237, 0.00001171, 0.00001084, 0.00000977, 0.00000854, 0.00000715, 0.00000562, 0.00000396, 0.00000216, 0.00000022, -0.00000186, -0.00000409, -0.00000648, -0.00000902, -0.00001171, -0.00001453, -0.00001746, -0.00002046, -0.00002347, -0.00002642, -0.00002924, -0.00003184, -0.00003411, -0.00003595, -0.00003726, -0.00003796, -0.00003795, -0.00003716, -0.00003556, -0.00003311, -0.00002983, -0.00002575, -0.00002094, -0.00001550, -0.00000955, -0.00000324, 0.00000326, 0.00000978, 0.00001613, 0.00002213, 0.00002759, 0.00003237, 0.00003632, 0.00003934, 0.00004136, 0.00004232, 0.00004224, 0.00004114, 0.00003909, 0.00003621, 0.00003260, 0.00002843, 0.00002386, 0.00001906, 0.00001419, 0.00000941, 0.00000488, 0.00000072, -0.00000297, -0.00000612, -0.00000869, -0.00001066, -0.00001204, -0.00001288, -0.00001323, -0.00001318, -0.00001281, -0.00001222, -0.00001149, -0.00001072, -0.00000996, -0.00000930, -0.00000875, -0.00000835, -0.00000809, -0.00000796, -0.00000792, -0.00000793, -0.00000793, -0.00000785, -0.00000765, -0.00000726, -0.00000665, -0.00000577, -0.00000461, -0.00000317, -0.00000148, 0.00000044, 0.00000252, 0.00000470, 0.00000691, 0.00000905, 0.00001106, 0.00001284, 0.00001433, 0.00001547, 0.00001622, 0.00001654, },
		{ -0.00000607, -0.00000665, -0.00000685, -0.00000669, -0.00000620, -0.00000541, -0.00000439, -0.00000320, -0.00000190, -0.00000056, 0.00000073, 0.00000191, 0.00000293, 0.00000374, 0.00000429, 0.00000457, 0.00000457, 0.00000428, 0.00000372, 0.00000291, 0.00000189, 0.00000070, -0.00000061, -0.00000201, -0.00000344, -0.00000485, -0.00000621, -0.00000748, -0.00000862, -0.00000960, -0.00001041, -0.00001103, -0.00001144, -0.00001164, -0.00001161, -0.00001136, -0.00001088, -0.00001017, -0.00000922, -0.00000804, -0.00000663, -0.00000498, -0.00000311, -0.00000103, 0.00000124, 0.00000369, 0.00000626, 0.00000894, 0.00001165, 0.00001436, 0.00001698, 0.00001944, 0.00002168, 0.00002361, 0.00002516, 0.00002626, 0.00002684, 0.00002686, 0.00002628, 0.00002508, 0.00002325, 0.00002083, 0.00001785, 0.00001437, 0.00001047, 0.00000625, 0.00000181, -0.00000272, -0.00000722, -0.00001157, -0.00001566, -0.00001937, -0.00002261, -0.00002530, -0.00002739, -0.00002884, -0.00002964, -0.00002981, -0.00002939, -0.00002843, -0.00002701, -0.00002522, -0.00002315, -0.00002090, -0.00001858, -0.00001628, -0.00001408, -0.00001206, -0.00001027, -0.00000874, -0.00000749, -0.00000652, -0.00000579, -0.00000529, -0.00000494, -0.00000470, -0.00000450, -0.00000427, -0.00000396, -0.00000350, -0.00000286, -0.00000200, -0.00000092, 0.00000038, 0.00000188, 0.00000353, 0.00000529, 0.00000710, 0.00000887, 0.00001054, 0.00001202, 0.00001326, 0.00001418, 0.00001475, 0.00001492, 0.00001468, 0.00001402, 0.00001298, 0.00001159, 0.00000990, 0.00000797, 0.00000590, 0.00000376, 0.00000163, -0.00000039, -0.00000224, -0.00000383, -0.00000513, },
		{ 0.00000564, 0.00000657, 0.00000723, 0.00000758, 0.00000762, 0.00000736, 0.00000681, 0.00000600, 0.00000498, 0.00000379, 0.00000248, 0.00000111, -0.00000026, -0.00000158, -0.00000280, -0.00000389, -0.00000479, -0.00000550, -0.00000598, -0.00000624, -0.00000627, -0.00000610, -0.00000573, -0.00000520, -0.00000454, -0.00000379, -0.00000298, -0.00000216, -0.00000136, -0.00000062, 0.00000003, 0.00000055, 0.00000092, 0.00000112, 0.00000114, 0.00000095, 0.00000057, -0.00000002, -0.00000079, -0.00000175, -0.00000287, -0.00000413, -0.00000550, -0.00000696, -0.00000847, -0.00000999, -0.00001149, -0.00001291, -0.00001423, -0.00001538, -0.00001634, -0.00001706, -0.00001752, -0.00001767, -0.00001749, -0.00001697, -0.00001610, -0.00001488, -0.00001333, -0.00001147, -0.00000934, -0.00000698, -0.00000444, -0.00000179, 0.00000090, 0.00000356, 0.00000611, 0.00000850, 0.00001063, 0.00001246, 0.00001393, 0.00001500, 0.00001565, 0.00001586, 0.00001564, 0.00001502, 0.00001402, 0.00001269, 0.00001111, 0.00000935, 0.00000748, 0.00000559, 0.00000376, 0.00000208, 0.00000061, -0.00000058, -0.00000144, -0.00000195, -0.00000208, -0.00000183, -0.00000124, -0.00000032, 0.00000086, 0.00000224, 0.00000375, 0.00000531, 0.00000684, 0.00000825, 0.00000948, 0.00001045, 0.00001111, 0.00001142, 0.00001136, 0.00001092, 0.00001012, 0.00000899, 0.00000756, 0.00000590, 0.00000409, 0.00000218, 0.00000027, -0.00000156, -0.00000324, -0.00000470, -0.00000587, -0.00000673, -0.00000723, -0.00000736, -0.00000712, -0.00000654, -0.00000564, -0.00000448, -0.00000310, -0.00000159, -0.00000001, 0.00000158, 0.00000309, 0.00000446, },
		{ -0.00000311, -0.00000447, -0.00000570, -0.00000674, -0.00000756, -0.00000815, -0.00000850, -0.00000860, -0.00000848, -0.00000814, -0.00000762, -0.00000696, -0.00000619, -0.00000537, -0.00000454, -0.00000375, -0.00000303, -0.00000242, -0.00000196, -0.00000166, -0.00000153, -0.00000158, -0.00000180, -0.00000216, -0.00000265, -0.00000322, -0.00000383, -0.00000443, -0.00000497, -0.00000540, -0.00000567, -0.00000573, -0.00000553, -0.00000505, -0.00000426, -0.00000314, -0.00000170, 0.00000006, 0.00000210, 0.00000440, 0.00000691, 0.00000957, 0.00001230, 0.00001503, 0.00001769, 0.00002019, 0.00002244, 0.00002437, 0.00002591, 0.00002698, 0.00002753, 0.00002752, 0.00002693, 0.00002573, 0.00002394, 0.00002158, 0.00001868, 0.00001531, 0.00001153, 0.00000743, 0.00000311, -0.00000134, -0.00000580, -0.00001016, -0.00001431, -0.00001814, -0.00002157, -0.00002450, -0.00002686, -0.00002859, -0.00002965, -0.00003003, -0.00002973, -0.00002876, -0.00002717, -0.00002502, -0.00002237, -0.00001932, -0.00001596, -0.00001240, -0.00000875, -0.00000511, -0.00000159, 0.00000170, 0.00000469, 0.00000730, 0.00000947, 0.00001115, 0.00001233, 0.00001299, 0.00001316, 0.00001285, 0.00001212, 0.00001103, 0.00000964, 0.00000803, 0.00000630, 0.00000451, 0.00000276, 0.00000111, -0.00000036, -0.00000159, -0.00000254, -0.00000318, -0.00000350, -0.00000349, -0.00000317, -0.00000256, -0.00000171, -0.00000067, 0.00000050, 0.00000175, 0.00000300, 0.00000420, 0.00000527, 0.00000618, 0.00000687, 0.00000730, 0.00000746, 0.00000733, 0.00000691, 0.00000622, 0.00000528, 0.00000413, 0.00000281, 0.00000138, -0.00000013, -0.00000164, },
		{ -0.00000655, -0.00000550, -0.00000448, -0.00000354, -0.00000272, -0.00000203, -0.00000152, -0.00000118, -0.00000104, -0.00000107, -0.00000126, -0.00000159, -0.00000202, -0.00000252, -0.00000303, -0.00000350, -0.00000388, -0.00000413, -0.00000419, -0.00000404, -0.00000364, -0.00000298, -0.00000206, -0.00000089, 0.00000051, 0.00000209, 0.00000380, 0.00000557, 0.00000734, 0.00000901, 0.00001050, 0.00001173, 0.00001262, 0.00001309, 0.00001307, 0.00001252, 0.00001141, 0.00000973, 0.00000747, 0.00000468, 0.00000140, -0.00000230, -0.00000633, -0.00001058, -0.00001493, -0.00001926, -0.00002344, -0.00002733, -0.00003081, -0.00003375, -0.00003605, -0.00003763, -0.00003842, -0.00003837, -0.00003746, -0.00003571, -0.00003314, -0.00002981, -0.00002580, -0.00002121, -0.00001615, -0.00001077, -0.00000519, 0.00000044, 0.00000598, 0.00001129, 0.00001623, 0.00002070, 0.00002460, 0.00002783, 0.00003036, 0.00003212, 0.00003312, 0.00003336, 0.00003287, 0.00003169, 0.00002990, 0.00002757, 0.00002480, 0.00002168, 0.00001833, 0.00001483, 0.00001131, 0.00000786, 0.00000456, 0.00000149, -0.00000127, -0.00000368, -0.00000571, -0.00000732, -0.00000852, -0.00000930, -0.00000969, -0.00000973, -0.00000945, -0.00000891, -0.00000816, -0.00000726, -0.00000628, -0.00000527, -0.00000429, -0.00000339, -0.00000262, -0.00000201, -0.00000160, -0.00000139, -0.00000141, -0.00000165, -0.00000209, -0.00000272, -0.00000352, -0.00000443, -0.00000543, -0.00000647, -0.00000751, -0.00000851, -0.00000941, -0.00001019, -0.00001080, -0.00001123, -0.00001145, -0.00001146, -0.00001126, -0.00001084, -0.00001024, -0.00000948, -0.00000858, -0.00000759, },
		{ 0.00000055, -0.00000007, -0.00000069, -0.00000126, -0.00000174, -0.00000209, -0.00000229, -0.00000228, -0.00000207, -0.00000163, -0.00000098, -0.00000013, 0.00000090, 0.00000207, 0.00000331, 0.00000458, 0.00000580, 0.00000691, 0.00000783, 0.00000850, 0.00000887, 0.00000887, 0.00000847, 0.00000766, 0.00000643, 0.00000480, 0.00000281, 0.00000051, -0.00000202, -0.00000469, -0.00000740, -0.00001005, -0.00001252, -0.00001470, -0.00001648, -0.00001778, -0.00001850, -0.00001859, -0.00001801, -0.00001676, -0.00001483, -0.00001228, -0.00000916, -0.00000557, -0.00000162, 0.00000258, 0.00000688, 0.00001114, 0.00001521, 0.00001896, 0.00002224, 0.00002494, 0.00002696, 0.00002824, 0.00002870, 0.00002834, 0.00002716, 0.00002519, 0.00002249, 0.00001913, 0.00001523, 0.00001089, 0.00000624, 0.00000143, -0.00000342, -0.00000818, -0.00001271, -0.00001691, -0.00002067, -0.00002393, -0.00002661, -0.00002868, -0.00003011, -0.00003092, -0.00003113, -0.00003076, -0.00002988, -0.00002854, -0.00002682, -0.00002480, -0.00002256, -0.00002018, -0.00001773, -0.00001529, -0.00001293, -0.00001069, -0.00000862, -0.00000677, -0.00000516, -0.00000380, -0.00000271, -0.00000187, -0.00000129, -0.00000094, -0.00000081, -0.00000086, -0.00000107, -0.00000142, -0.00000185, -0.00000235, -0.00000289, -0.00000342, -0.00000392, -0.00000436, -0.00000472, -0.00000499, -0.00000514, -0.00000516, -0.00000506, -0.00000482, -0.00000446, -0.00000398, -0.00000340, -0.00000274, -0.00000202, -0.00000127, -0.00000052, 0.00000020, 0.00000086, 0.00000143, 0.00000189, 0.00000222, 0.00000241, 0.00000244, 0.00000231, 0.00000204, 0.00000164, 0.00000113, },
		{ 0.00000035, 0.00000106, 0.00000184, 0.00000263, 0.00000338, 0.00000405, 0.00000459, 0.00000493, 0.00000505, 0.00000490, 0.00000446, 0.00000373, 0.00000270, 0.00000140, -0.00000015, -0.00000188, -0.00000375, -0.00000567, -0.00000756, -0.00000936, -0.00001096, -0.00001231, -0.00001332, -0.00001394, -0.00001413, -0.00001388, -0.00001318, -0.00001205, -0.00001053, -0.00000869, -0.00000662, -0.00000439, -0.00000212, 0.00000008, 0.00000211, 0.00000385, 0.00000520, 0.00000610, 0.00000646, 0.00000626, 0.00000548, 0.00000413, 0.00000224, -0.00000013, -0.00000288, -0.00000592, -0.00000913, -0.00001238, -0.00001554, -0.00001846, -0.00002104, -0.00002316, -0.00002471, -0.00002564, -0.00002588, -0.00002541, -0.00002423, -0.00002238, -0.00001991, -0.00001689, -0.00001341, -0.00000960, -0.00000556, -0.00000143, 0.00000269, 0.00000665, 0.00001037, 0.00001374, 0.00001669, 0.00001915, 0.00002109, 0.00002248, 0.00002334, 0.00002367, 0.00002352, 0.00002293, 0.00002196, 0.00002068, 0.00001916, 0.00001748, 0.00001569, 0.00001388, 0.00001208, 0.00001035, 0.00000874, 0.00000726, 0.00000594, 0.00000480, 0.00000382, 0.00000301, 0.00000237, 0.00000187, 0.00000150, 0.00000125, 0.00000110, 0.00000104, 0.00000105, 0.00000111, 0.00000123, 0.00000139, 0.00000157, 0.00000178, 0.00000200, 0.00000222, 0.00000244, 0.00000264, 0.00000281, 0.00000294, 0.00000302, 0.00000304, 0.00000299, 0.00000286, 0.00000265, 0.00000236, 0.00000200, 0.00000158, 0.00000111, 0.00000062, 0.00000013, -0.00000033, -0.00000074, -0.00000107, -0.00000129, -0.00000138, -0.00000133, -0.00000113, -0.00000077, -0.00000028, },
		{ -0.00000023, -0.00000081, -0.00000152, -0.00000233, -0.00000321, -0.00000411, -0.00000498, -0.00000576, -0.00000641, -0.00000686, -0.00000708, -0.00000703, -0.00000670, -0.00000607, -0.00000515, -0.00000397, -0.00000256, -0.00000098, 0.00000071, 0.00000244, 0.00000413, 0.00000570, 0.00000709, 0.00000823, 0.00000907, 0.00000956, 0.00000970, 0.00000947, 0.00000891, 0.00000804, 0.00000694, 0.00000567, 0.00000433, 0.00000300, 0.00000180, 0.00000081, 0.00000012, -0.00000018, -0.00000004, 0.00000058, 0.00000171, 0.00000333, 0.00000541, 0.00000789, 0.00001070, 0.00001372, 0.00001684, 0.00001995, 0.00002290, 0.00002557, 0.00002784, 0.00002959, 0.00003075, 0.00003125, 0.00003103, 0.00003009, 0.00002845, 0.00002614, 0.00002323, 0.00001981, 0.00001599, 0.00001190, 0.00000766, 0.00000341, -0.00000072, -0.00000462, -0.00000817, -0.00001128, -0.00001388, -0.00001592, -0.00001739, -0.00001828, -0.00001861, -0.00001842, -0.00001777, -0.00001673, -0.00001538, -0.00001381, -0.00001209, -0.00001032, -0.00000856, -0.00000687, -0.00000531, -0.00000392, -0.00000273, -0.00000173, -0.00000094, -0.00000034, 0.00000009, 0.00000037, 0.00000054, 0.00000061, 0.00000063, 0.00000061, 0.00000057, 0.00000053, 0.00000051, 0.00000049, 0.00000049, 0.00000048, 0.00000047, 0.00000044, 0.00000038, 0.00000029, 0.00000015, -0.00000004, -0.00000026, -0.00000052, -0.00000081, -0.00000109, -0.00000137, -0.00000162, -0.00000181, -0.00000194, -0.00000199, -0.00000194, -0.00000181, -0.00000160, -0.00000132, -0.00000098, -0.00000061, -0.00000025, 0.00000008, 0.00000035, 0.00000051, 0.00000055, 0.00000045, 0.00000019, },
		{ 0.00000224, 0.00000277, 0.00000341, 0.00000415, 0.00000498, 0.00000586, 0.00000675, 0.00000760, 0.00000838, 0.00000903, 0.00000952, 0.00000980, 0.00000985, 0.00000965, 0.00000920, 0.00000851, 0.00000762, 0.00000655, 0.00000537, 0.00000414, 0.00000291, 0.00000177, 0.00000078, 0.00000000, -0.00000052, -0.00000074, -0.00000064, -0.00000021, 0.00000051, 0.00000150, 0.00000269, 0.00000401, 0.00000537, 0.00000668, 0.00000783, 0.00000872, 0.00000926, 0.00000937, 0.00000899, 0.00000807, 0.00000659, 0.00000457, 0.00000203, -0.00000095, -0.00000429, -0.00000787, -0.00001158, -0.00001528, -0.00001881, -0.00002204, -0.00002483, -0.00002706, -0.00002863, -0.00002947, -0.00002952, -0.00002877, -0.00002724, -0.00002496, -0.00002202, -0.00001852, -0.00001458, -0.00001033, -0.00000593, -0.00000153, 0.00000274, 0.00000672, 0.00001030, 0.00001337, 0.00001587, 0.00001774, 0.00001896, 0.00001953, 0.00001949, 0.00001888, 0.00001779, 0.00001630, 0.00001451, 0.00001252, 0.00001043, 0.00000834, 0.00000633, 0.00000448, 0.00000284, 0.00000145, 0.00000032, -0.00000052, -0.00000112, -0.00000147, -0.00000164, -0.00000166, -0.00000158, -0.00000145, -0.00000131, -0.00000121, -0.00000118, -0.00000123, -0.00000138, -0.00000162, -0.00000195, -0.00000235, -0.00000278, -0.00000322, -0.00000364, -0.00000400, -0.00000428, -0.00000445, -0.00000450, -0.00000442, -0.00000422, -0.00000389, -0.00000346, -0.00000296, -0.00000240, -0.00000182, -0.00000125, -0.00000071, -0.00000023, 0.00000018, 0.00000051, 0.00000075, 0.00000092, 0.00000103, 0.00000110, 0.00000116, 0.00000123, 0.00000135, 0.00000155, 0.00000184, },
		{ 0.00000323, 0.00000257, 0.00000180, 0.00000094, 0.00000000, -0.00000098, -0.00000196, -0.00000291, -0.00000378, -0.00000452, -0.00000509, -0.00000545, -0.00000557, -0.00000544, -0.00000505, -0.00000441, -0.00000357, -0.00000254, -0.00000140, -0.00000021, 0.00000096, 0.00000204, 0.00000294, 0.00000361, 0.00000397, 0.00000398, 0.00000361, 0.00000287, 0.00000175, 0.00000030, -0.00000142, -0.00000333, -0.00000535, -0.00000735, -0.00000924, -0.00001088, -0.00001218, -0.00001303, -0.00001336, -0.00001309, -0.00001219, -0.00001067, -0.00000854, -0.00000585, -0.00000270, 0.00000082, 0.00000456, 0.00000839, 0.00001214, 0.00001567, 0.00001881, 0.00002142, 0.00002340, 0.00002463, 0.00002506, 0.00002465, 0.00002340, 0.00002136, 0.00001858, 0.00001517, 0.00001126, 0.00000699, 0.00000252, -0.00000199, -0.00000637, -0.00001048, -0.00001418, -0.00001735, -0.00001989, -0.00002176, -0.00002292, -0.00002338, -0.00002317, -0.00002234, -0.00002099, -0.00001921, -0.00001712, -0.00001484, -0.00001248, -0.00001017, -0.00000800, -0.00000606, -0.00000442, -0.00000313, -0.00000221, -0.00000166, -0.00000145, -0.00000155, -0.00000190, -0.00000245, -0.00000310, -0.00000380, -0.00000447, -0.00000504, -0.00000546, -0.00000568, -0.00000569, -0.00000546, -0.00000500, -0.00000433, -0.00000347, -0.00000247, -0.00000137, -0.00000021, 0.00000095, 0.00000207, 0.00000310, 0.00000402, 0.00000480, 0.00000543, 0.00000590, 0.00000622, 0.00000640, 0.00000646, 0.00000643, 0.00000633, 0.00000619, 0.00000602, 0.00000585, 0.00000568, 0.00000552, 0.00000537, 0.00000521, 0.00000504, 0.00000483, 0.00000456, 0.00000421, 0.00000377, },
		{ -0.00000281, -0.00000233, -0.00000170, -0.00000091, -0.00000001, 0.00000097, 0.00000199, 0.00000300, 0.00000394, 0.00000474, 0.00000536, 0.00000574, 0.00000584, 0.00000564, 0.00000512, 0.00000429, 0.00000316, 0.00000177, 0.00000019, -0.00000151, -0.00000327, -0.00000498, -0.00000656, -0.00000792, -0.00000898, -0.00000967, -0.00000995, -0.00000978, -0.00000916, -0.00000811, -0.00000667, -0.00000491, -0.00000292, -0.00000080, 0.00000132, 0.00000332, 0.00000508, 0.00000649, 0.00000743, 0.00000782, 0.00000761, 0.00000676, 0.00000528, 0.00000320, 0.00000058, -0.00000248, -0.00000587, -0.00000943, -0.00001302, -0.00001647, -0.00001963, -0.00002233, -0.00002445, -0.00002586, -0.00002649, -0.00002629, -0.00002523, -0.00002334, -0.00002067, -0.00001733, -0.00001343, -0.00000911, -0.00000455, 0.00000009, 0.00000463, 0.00000890, 0.00001275, 0.00001604, 0.00001867, 0.00002055, 0.00002164, 0.00002194, 0.00002148, 0.00002031, 0.00001851, 0.00001621, 0.00001352, 0.00001060, 0.00000757, 0.00000459, 0.00000179, -0.00000073, -0.00000286, -0.00000453, -0.00000569, -0.00000634, -0.00000647, -0.00000611, -0.00000532, -0.00000416, -0.00000273, -0.00000110, 0.00000063, 0.00000236, 0.00000402, 0.00000553, 0.00000684, 0.00000790, 0.00000869, 0.00000919, 0.00000941, 0.00000938, 0.00000910, 0.00000864, 0.00000802, 0.00000729, 0.00000650, 0.00000568, 0.00000486, 0.00000408, 0.00000333, 0.00000265, 0.00000202, 0.00000145, 0.00000092, 0.00000042, -0.00000005, -0.00000051, -0.00000096, -0.00000140, -0.00000183, -0.00000224, -0.00000261, -0.00000292, -0.00000315, -0.00000328, -0.00000327, -0.00000312, },
		{ 0.00000288, 0.00000242, 0.00000173, 0.00000080, -0.00000033, -0.00000164, -0.00000308, -0.00000458, -0.00000609, -0.00000753, -0.00000883, -0.00000991, -0.00001072, -0.00001120, -0.00001132, -0.00001105, -0.00001040, -0.00000939, -0.00000806, -0.00000647, -0.00000469, -0.00000281, -0.00000093, 0.00000086, 0.00000246, 0.00000378, 0.00000476, 0.00000533, 0.00000547, 0.00000517, 0.00000445, 0.00000334, 0.00000193, 0.00000029, -0.00000146, -0.00000320, -0.00000482, -0.00000618, -0.00000718, -0.00000772, -0.00000773, -0.00000714, -0.00000594, -0.00000414, -0.00000178, 0.00000106, 0.00000427, 0.00000773, 0.00001128, 0.00001477, 0.00001803, 0.00002089, 0.00002321, 0.00002485, 0.00002571, 0.00002572, 0.00002484, 0.00002306, 0.00002044, 0.00001703, 0.00001297, 0.00000838, 0.00000343, -0.00000169, -0.00000681, -0.00001173, -0.00001627, -0.00002028, -0.00002361, -0.00002615, -0.00002782, -0.00002860, -0.00002847, -0.00002746, -0.00002566, -0.00002314, -0.00002004, -0.00001649, -0.00001264, -0.00000865, -0.00000467, -0.00000084, 0.00000271, 0.00000588, 0.00000859, 0.00001077, 0.00001241, 0.00001349, 0.00001403, 0.00001407, 0.00001368, 0.00001293, 0.00001188, 0.00001063, 0.00000925, 0.00000782, 0.00000641, 0.00000507, 0.00000384, 0.00000276, 0.00000183, 0.00000106, 0.00000044, -0.00000003, -0.00000039, -0.00000065, -0.00000084, -0.00000097, -0.00000107, -0.00000114, -0.00000119, -0.00000122, -0.00000123, -0.00000120, -0.00000112, -0.00000099, -0.00000079, -0.00000051, -0.00000015, 0.00000027, 0.00000074, 0.00000125, 0.00000177, 0.00000225, 0.00000267, 0.00000298, 0.00000314, 0.00000311, },
		{ -0.00001110, -0.00001128, -0.00001116, -0.00001074, -0.00001002, -0.00000900, -0.00000772, -0.00000623, -0.00000458, -0.00000286, -0.00000114, 0.00000051, 0.00000200, 0.00000325, 0.00000422, 0.00000485, 0.00000511, 0.00000501, 0.00000454, 0.00000376, 0.00000270, 0.00000144, 0.00000007, -0.00000132, -0.00000265, -0.00000382, -0.00000475, -0.00000537, -0.00000562, -0.00000549, -0.00000496, -0.00000406, -0.00000283, -0.00000136, 0.00000028, 0.00000198, 0.00000362, 0.00000507, 0.00000622, 0.00000696, 0.00000720, 0.00000686, 0.00000591, 0.00000433, 0.00000214, -0.00000060, -0.00000381, -0.00000738, -0.00001117, -0.00001503, -0.00001879, -0.00002227, -0.00002532, -0.00002778, -0.00002952, -0.00003044, -0.00003046, -0.00002955, -0.00002770, -0.00002496, -0.00002141, -0.00001715, -0.00001232, -0.00000708, -0.00000162, 0.00000389, 0.00000926, 0.00001430, 0.00001887, 0.00002283, 0.00002606, 0.00002848, 0.00003006, 0.00003077, 0.00003065, 0.00002975, 0.00002814, 0.00002593, 0.00002325, 0.00002021, 0.00001695, 0.00001361, 0.00001030, 0.00000713, 0.00000420, 0.00000157, -0.00000072, -0.00000262, -0.00000413, -0.00000528, -0.00000608, -0.00000657, -0.00000681, -0.00000685, -0.00000674, -0.00000654, -0.00000629, -0.00000602, -0.00000577, -0.00000555, -0.00000537, -0.00000523, -0.00000512, -0.00000502, -0.00000493, -0.00000482, -0.00000468, -0.00000450, -0.00000428, -0.00000403, -0.00000374, -0.00000345, -0.00000318, -0.00000295, -0.00000279, -0.00000274, -0.00000281, -0.00000303, -0.00000340, -0.00000394, -0.00000462, -0.00000544, -0.00000635, -0.00000731, -0.00000829, -0.00000921, -0.00001002, -0.00001067, },
		{ 0.00000286, 0.00000394, 0.00000479, 0.00000539, 0.00000569, 0.00000568, 0.00000537, 0.00000478, 0.00000394, 0.00000292, 0.00000177, 0.00000057, -0.00000060, -0.00000168, -0.00000259, -0.00000327, -0.00000368, -0.00000378, -0.00000358, -0.00000309, -0.00000232, -0.00000135, -0.00000023, 0.00000096, 0.00000213, 0.00000320, 0.00000407, 0.00000467, 0.00000494, 0.00000484, 0.00000432, 0.00000341, 0.00000212, 0.00000050, -0.00000138, -0.00000342, -0.00000551, -0.00000755, -0.00000941, -0.00001097, -0.00001212, -0.00001277, -0.00001286, -0.00001233, -0.00001116, -0.00000938, -0.00000703, -0.00000418, -0.00000093, 0.00000259, 0.00000624, 0.00000986, 0.00001331, 0.00001644, 0.00001909, 0.00002116, 0.00002254, 0.00002316, 0.00002299, 0.00002203, 0.00002029, 0.00001785, 0.00001480, 0.00001124, 0.00000731, 0.00000317, -0.00000105, -0.00000519, -0.00000911, -0.00001269, -0.00001581, -0.00001839, -0.00002038, -0.00002175, -0.00002249, -0.00002264, -0.00002224, -0.00002135, -0.00002007, -0.00001848, -0.00001669, -0.00001479, -0.00001287, -0.00001102, -0.00000930, -0.00000776, -0.00000644, -0.00000536, -0.00000451, -0.00000388, -0.00000345, -0.00000318, -0.00000303, -0.00000296, -0.00000292, -0.00000288, -0.00000281, -0.00000270, -0.00000253, -0.00000232, -0.00000206, -0.00000178, -0.00000151, -0.00000128, -0.00000111, -0.00000104, -0.00000109, -0.00000128, -0.00000160, -0.00000206, -0.00000264, -0.00000331, -0.00000403, -0.00000476, -0.00000546, -0.00000607, -0.00000654, -0.00000683, -0.00000691, -0.00000675, -0.00000634, -0.00000567, -0.00000478, -0.00000369, -0.00000244, -0.00000110, 0.00000027, 0.00000162, },
		{ 0.00000152, 0.00000050, -0.00000043, -0.00000122, -0.00000182, -0.00000220, -0.00000234, -0.00000224, -0.00000192, -0.00000140, -0.00000074, 0.00000000, 0.00000077, 0.00000148, 0.00000208, 0.00000250, 0.00000269, 0.00000259, 0.00000220, 0.00000150, 0.00000051, -0.00000074, -0.00000221, -0.00000382, -0.00000551, -0.00000718, -0.00000875, -0.00001013, -0.00001125, -0.00001204, -0.00001246, -0.00001246, -0.00001205, -0.00001123, -0.00001005, -0.00000857, -0.00000684, -0.00000498, -0.00000307, -0.00000122, 0.00000047, 0.00000191, 0.00000301, 0.00000371, 0.00000397, 0.00000377, 0.00000311, 0.00000203, 0.00000058, -0.00000116, -0.00000310, -0.00000514, -0.00000715, -0.00000902, -0.00001065, -0.00001194, -0.00001280, -0.00001317, -0.00001302, -0.00001233, -0.00001111, -0.00000940, -0.00000727, -0.00000479, -0.00000207, 0.00000078, 0.00000366, 0.00000645, 0.00000904, 0.00001134, 0.00001327, 0.00001476, 0.00001579, 0.00001634, 0.00001642, 0.00001606, 0.00001530, 0.00001422, 0.00001288, 0.00001138, 0.00000980, 0.00000822, 0.00000671, 0.00000533, 0.00000414, 0.00000317, 0.00000242, 0.00000191, 0.00000160, 0.00000148, 0.00000149, 0.00000158, 0.00000171, 0.00000182, 0.00000186, 0.00000180, 0.00000161, 0.00000127, 0.00000078, 0.00000016, -0.00000058, -0.00000139, -0.00000222, -0.00000303, -0.00000376, -0.00000436, -0.00000479, -0.00000500, -0.00000498, -0.00000470, -0.00000418, -0.00000343, -0.00000248, -0.00000137, -0.00000017, 0.00000108, 0.00000229, 0.00000342, 0.00000440, 0.00000518, 0.00000572, 0.00000599, 0.00000599, 0.00000572, 0.00000519, 0.00000446, 0.00000357, 0.00000257, },
		{ -0.00000125, -0.00000067, -0.00000011, 0.00000036, 0.00000070, 0.00000086, 0.00000082, 0.00000055, 0.00000006, -0.00000063, -0.00000151, -0.00000251, -0.00000360, -0.00000469, -0.00000574, -0.00000667, -0.00000741, -0.00000792, -0.00000814, -0.00000805, -0.00000764, -0.00000690, -0.00000585, -0.00000454, -0.00000301, -0.00000132, 0.00000045, 0.00000224, 0.00000398, 0.00000559, 0.00000703, 0.00000823, 0.00000917, 0.00000984, 0.00001024, 0.00001038, 0.00001030, 0.00001005, 0.00000968, 0.00000925, 0.00000885, 0.00000851, 0.00000831, 0.00000829, 0.00000847, 0.00000888, 0.00000951, 0.00001035, 0.00001135, 0.00001245, 0.00001360, 0.00001472, 0.00001573, 0.00001654, 0.00001708, 0.00001728, 0.00001710, 0.00001648, 0.00001543, 0.00001394, 0.00001205, 0.00000979, 0.00000724, 0.00000447, 0.00000159, -0.00000131, -0.00000412, -0.00000674, -0.00000908, -0.00001107, -0.00001263, -0.00001373, -0.00001435, -0.00001449, -0.00001418, -0.00001346, -0.00001240, -0.00001107, -0.00000957, -0.00000797, -0.00000639, -0.00000489, -0.00000357, -0.00000248, -0.00000166, -0.00000115, -0.00000095, -0.00000105, -0.00000141, -0.00000199, -0.00000272, -0.00000354, -0.00000437, -0.00000514, -0.00000578, -0.00000623, -0.00000644, -0.00000639, -0.00000605, -0.00000544, -0.00000457, -0.00000348, -0.00000222, -0.00000085, 0.00000055, 0.00000192, 0.00000319, 0.00000430, 0.00000518, 0.00000581, 0.00000615, 0.00000620, 0.00000596, 0.00000545, 0.00000473, 0.00000382, 0.00000280, 0.00000172, 0.00000065, -0.00000034, -0.00000121, -0.00000191, -0.00000241, -0.00000268, -0.00000274, -0.00000259, -0.00000227, -0.00000180, },
		{ -0.00000002, -0.00000074, -0.00000148, -0.00000220, -0.00000282, -0.00000331, -0.00000362, -0.00000372, -0.00000359, -0.00000322, -0.00000261, -0.00000180, -0.00000080, 0.00000033, 0.00000156, 0.00000281, 0.00000404, 0.00000519, 0.00000622, 0.00000707, 0.00000772, 0.00000815, 0.00000835, 0.00000832, 0.00000808, 0.00000766, 0.00000709, 0.00000642, 0.00000567, 0.00000490, 0.00000415, 0.00000344, 0.00000279, 0.00000222, 0.00000174, 0.00000132, 0.00000096, 0.00000062, 0.00000027, -0.00000012, -0.00000061, -0.00000121, -0.00000197, -0.00000290, -0.00000400, -0.00000528, -0.00000672, -0.00000827, -0.00000990, -0.00001154, -0.00001313, -0.00001458, -0.00001583, -0.00001681, -0.00001743, -0.00001765, -0.00001743, -0.00001675, -0.00001559, -0.00001399, -0.00001197, -0.00000959, -0.00000695, -0.00000412, -0.00000121, 0.00000167, 0.00000441, 0.00000691, 0.00000906, 0.00001078, 0.00001203, 0.00001274, 0.00001291, 0.00001254, 0.00001166, 0.00001034, 0.00000864, 0.00000665, 0.00000449, 0.00000225, 0.00000006, -0.00000200, -0.00000381, -0.00000530, -0.00000641, -0.00000709, -0.00000733, -0.00000713, -0.00000652, -0.00000554, -0.00000426, -0.00000277, -0.00000114, 0.00000053, 0.00000214, 0.00000362, 0.00000489, 0.00000589, 0.00000658, 0.00000693, 0.00000693, 0.00000661, 0.00000599, 0.00000512, 0.00000407, 0.00000290, 0.00000167, 0.00000047, -0.00000063, -0.00000160, -0.00000236, -0.00000290, -0.00000320, -0.00000325, -0.00000307, -0.00000269, -0.00000215, -0.00000151, -0.00000080, -0.00000010, 0.00000054, 0.00000107, 0.00000146, 0.00000167, 0.00000168, 0.00000150, 0.00000114, 0.00000062, },
		{ -0.00000321, -0.00000264, -0.00000195, -0.00000119, -0.00000040, 0.00000036, 0.00000107, 0.00000168, 0.00000216, 0.00000249, 0.00000265, 0.00000264, 0.00000248, 0.00000217, 0.00000175, 0.00000123, 0.00000065, 0.00000005, -0.00000054, -0.00000110, -0.00000159, -0.00000201, -0.00000235, -0.00000259, -0.00000275, -0.00000285, -0.00000288, -0.00000289, -0.00000288, -0.00000288, -0.00000290, -0.00000296, -0.00000306, -0.00000320, -0.00000336, -0.00000352, -0.00000367, -0.00000376, -0.00000376, -0.00000364, -0.00000336, -0.00000290, -0.00000221, -0.00000131, -0.00000018, 0.00000115, 0.00000267, 0.00000433, 0.00000606, 0.00000781, 0.00000950, 0.00001104, 0.00001236, 0.00001337, 0.00001400, 0.00001418, 0.00001388, 0.00001307, 0.00001174, 0.00000991, 0.00000761, 0.00000492, 0.00000190, -0.00000134, -0.00000469, -0.00000802, -0.00001123, -0.00001418, -0.00001676, -0.00001888, -0.00002045, -0.00002141, -0.00002174, -0.00002141, -0.00002046, -0.00001893, -0.00001687, -0.00001440, -0.00001160, -0.00000860, -0.00000551, -0.00000247, 0.00000041, 0.00000304, 0.00000530, 0.00000714, 0.00000851, 0.00000938, 0.00000975, 0.00000965, 0.00000911, 0.00000820, 0.00000700, 0.00000560, 0.00000407, 0.00000253, 0.00000104, -0.00000030, -0.00000145, -0.00000234, -0.00000296, -0.00000328, -0.00000331, -0.00000308, -0.00000262, -0.00000199, -0.00000124, -0.00000043, 0.00000036, 0.00000110, 0.00000171, 0.00000216, 0.00000242, 0.00000247, 0.00000231, 0.00000196, 0.00000143, 0.00000076, -0.00000000, -0.00000081, -0.00000161, -0.00000235, -0.00000299, -0.00000349, -0.00000382, -0.00000395, -0.00000389, -0.00000364, },
		{ 0.00000184, 0.00000177, 0.00000159, 0.00000133, 0.00000101, 0.00000065, 0.00000029, -0.00000004, -0.00000034, -0.00000057, -0.00000073, -0.00000082, -0.00000083, -0.00000077, -0.00000066, -0.00000050, -0.00000033, -0.00000015, 0.00000001, 0.00000014, 0.00000023, 0.00000028, 0.00000028, 0.00000023, 0.00000016, 0.00000007, -0.00000002, -0.00000010, -0.00000013, -0.00000011, -0.00000003, 0.00000011, 0.00000033, 0.00000060, 0.00000091, 0.00000123, 0.00000154, 0.00000178, 0.00000192, 0.00000192, 0.00000174, 0.00000133, 0.00000068, -0.00000023, -0.00000142, -0.00000286, -0.00000453, -0.00000639, -0.00000839, -0.00001045, -0.00001250, -0.00001445, -0.00001620, -0.00001767, -0.00001878, -0.00001943, -0.00001959, -0.00001919, -0.00001821, -0.00001666, -0.00001455, -0.00001192, -0.00000885, -0.00000542, -0.00000174, 0.00000209, 0.00000593, 0.00000966, 0.00001315, 0.00001628, 0.00001896, 0.00002111, 0.00002264, 0.00002354, 0.00002377, 0.00002337, 0.00002235, 0.00002080, 0.00001877, 0.00001639, 0.00001374, 0.00001095, 0.00000814, 0.00000540, 0.00000285, 0.00000058, -0.00000136, -0.00000290, -0.00000402, -0.00000471, -0.00000499, -0.00000488, -0.00000444, -0.00000374, -0.00000283, -0.00000181, -0.00000075, 0.00000028, 0.00000120, 0.00000197, 0.00000254, 0.00000288, 0.00000298, 0.00000284, 0.00000248, 0.00000193, 0.00000124, 0.00000044, -0.00000040, -0.00000124, -0.00000203, -0.00000271, -0.00000325, -0.00000363, -0.00000382, -0.00000383, -0.00000366, -0.00000332, -0.00000284, -0.00000226, -0.00000161, -0.00000094, -0.00000027, 0.00000034, 0.00000088, 0.00000131, 0.00000162, 0.00000179, },
		{ -0.00000180, -0.00000169, -0.00000152, -0.00000130, -0.00000105, -0.00000079, -0.00000054, -0.00000031, -0.00000012, 0.00000001, 0.00000010, 0.00000012, 0.00000010, 0.00000003, -0.00000007, -0.00000019, -0.00000031, -0.00000042, -0.00000050, -0.00000053, -0.00000053, -0.00000048, -0.00000038, -0.00000026, -0.00000011, 0.00000004, 0.00000016, 0.00000025, 0.00000027, 0.00000021, 0.00000005, -0.00000022, -0.00000059, -0.00000106, -0.00000161, -0.00000221, -0.00000284, -0.00000346, -0.00000400, -0.00000444, -0.00000471, -0.00000477, -0.00000459, -0.00000412, -0.00000335, -0.00000227, -0.00000089, 0.00000077, 0.00000267, 0.00000475, 0.00000697, 0.00000922, 0.00001144, 0.00001353, 0.00001540, 0.00001697, 0.00001817, 0.00001892, 0.00001918, 0.00001892, 0.00001813, 0.00001681, 0.00001501, 0.00001278, 0.00001018, 0.00000730, 0.00000425, 0.00000112, -0.00000197, -0.00000491, -0.00000761, -0.00000996, -0.00001190, -0.00001337, -0.00001433, -0.00001476, -0.00001467, -0.00001408, -0.00001305, -0.00001163, -0.00000990, -0.00000796, -0.00000588, -0.00000378, -0.00000173, 0.00000018, 0.00000188, 0.00000330, 0.00000442, 0.00000519, 0.00000562, 0.00000572, 0.00000550, 0.00000501, 0.00000430, 0.00000343, 0.00000245, 0.00000143, 0.00000042, -0.00000052, -0.00000134, -0.00000201, -0.00000251, -0.00000282, -0.00000295, -0.00000291, -0.00000272, -0.00000240, -0.00000199, -0.00000153, -0.00000105, -0.00000059, -0.00000019, 0.00000014, 0.00000037, 0.00000050, 0.00000052, 0.00000044, 0.00000027, 0.00000002, -0.00000028, -0.00000060, -0.00000092, -0.00000122, -0.00000148, -0.00000167, -0.00000180, -0.00000184, },
		{ 0.00000057, 0.00000060, 0.00000057, 0.00000048, 0.00000036, 0.00000022, 0.00000008, -0.00000005, -0.00000015, -0.00000022, -0.00000025, -0.00000023, -0.00000019, -0.00000012, -0.00000004, 0.00000003, 0.00000009, 0.00000011, 0.00000007, -0.00000003, -0.00000020, -0.00000045, -0.00000077, -0.00000114, -0.00000157, -0.00000202, -0.00000247, -0.00000289, -0.00000325, -0.00000353, -0.00000370, -0.00000373, -0.00000362, -0.00000335, -0.00000293, -0.00000236, -0.00000166, -0.00000087, -0.00000002, 0.00000086, 0.00000171, 0.00000249, 0.00000316, 0.00000366, 0.00000398, 0.00000406, 0.00000391, 0.00000351, 0.00000288, 0.00000202, 0.00000098, -0.00000021, -0.00000148, -0.00000278, -0.00000405, -0.00000521, -0.00000621, -0.00000697, -0.00000746, -0.00000764, -0.00000747, -0.00000694, -0.00000607, -0.00000487, -0.00000338, -0.00000164, 0.00000027, 0.00000230, 0.00000436, 0.00000637, 0.00000826, 0.00000995, 0.00001138, 0.00001249, 0.00001324, 0.00001361, 0.00001359, 0.00001318, 0.00001241, 0.00001130, 0.00000992, 0.00000831, 0.00000654, 0.00000468, 0.00000279, 0.00000095, -0.00000079, -0.00000238, -0.00000377, -0.00000494, -0.00000587, -0.00000656, -0.00000700, -0.00000723, -0.00000726, -0.00000713, -0.00000687, -0.00000653, -0.00000615, -0.00000575, -0.00000538, -0.00000506, -0.00000481, -0.00000464, -0.00000454, -0.00000452, -0.00000457, -0.00000466, -0.00000478, -0.00000490, -0.00000500, -0.00000506, -0.00000506, -0.00000498, -0.00000483, -0.00000458, -0.00000425, -0.00000384, -0.00000337, -0.00000286, -0.00000232, -0.00000178, -0.00000126, -0.00000078, -0.00000035, 0.00000000, 0.00000028, 0.00000047, },
		{ 0.00000090, 0.00000080, 0.00000075, 0.00000074, 0.00000073, 0.00000073, 0.00000071, 0.00000065, 0.00000056, 0.00000041, 0.00000020, -0.00000006, -0.00000037, -0.00000071, -0.00000108, -0.00000145, -0.00000180, -0.00000210, -0.00000233, -0.00000247, -0.00000250, -0.00000241, -0.00000218, -0.00000181, -0.00000131, -0.00000068, 0.00000005, 0.00000087, 0.00000175, 0.00000266, 0.00000356, 0.00000444, 0.00000525, 0.00000597, 0.00000658, 0.00000706, 0.00000740, 0.00000761, 0.00000769, 0.00000765, 0.00000750, 0.00000727, 0.00000699, 0.00000667, 0.00000636, 0.00000607, 0.00000583, 0.00000564, 0.00000553, 0.00000549, 0.00000552, 0.00000560, 0.00000573, 0.00000586, 0.00000598, 0.00000605, 0.00000603, 0.00000591, 0.00000564, 0.00000521, 0.00000459, 0.00000378, 0.00000279, 0.00000161, 0.00000027, -0.00000121, -0.00000279, -0.00000444, -0.00000609, -0.00000772, -0.00000926, -0.00001069, -0.00001194, -0.00001300, -0.00001383, -0.00001441, -0.00001474, -0.00001482, -0.00001466, -0.00001428, -0.00001370, -0.00001297, -0.00001212, -0.00001119, -0.00001022, -0.00000925, -0.00000832, -0.00000746, -0.00000668, -0.00000601, -0.00000545, -0.00000501, -0.00000466, -0.00000441, -0.00000423, -0.00000409, -0.00000398, -0.00000386, -0.00000371, -0.00000352, -0.00000325, -0.00000291, -0.00000249, -0.00000199, -0.00000141, -0.00000078, -0.00000010, 0.00000060, 0.00000130, 0.00000196, 0.00000258, 0.00000312, 0.00000357, 0.00000391, 0.00000413, 0.00000424, 0.00000423, 0.00000412, 0.00000391, 0.00000363, 0.00000329, 0.00000292, 0.00000253, 0.00000216, 0.00000181, 0.00000150, 0.00000124, 0.00000104, },
		{ 0.00000058, 0.00000035, 0.00000007, -0.00000023, -0.00000054, -0.00000084, -0.00000109, -0.00000129, -0.00000140, -0.00000142, -0.00000132, -0.00000109, -0.00000074, -0.00000027, 0.00000032, 0.00000100, 0.00000177, 0.00000259, 0.00000343, 0.00000427, 0.00000507, 0.00000580, 0.00000644, 0.00000695, 0.00000733, 0.00000754, 0.00000758, 0.00000745, 0.00000714, 0.00000667, 0.00000606, 0.00000531, 0.00000444, 0.00000349, 0.00000248, 0.00000143, 0.00000036, -0.00000069, -0.00000172, -0.00000270, -0.00000362, -0.00000447, -0.00000524, -0.00000593, -0.00000653, -0.00000705, -0.00000749, -0.00000784, -0.00000811, -0.00000830, -0.00000841, -0.00000843, -0.00000838, -0.00000823, -0.00000800, -0.00000767, -0.00000725, -0.00000673, -0.00000611, -0.00000540, -0.00000461, -0.00000374, -0.00000281, -0.00000183, -0.00000083, 0.00000018, 0.00000116, 0.00000209, 0.00000296, 0.00000372, 0.00000437, 0.00000488, 0.00000524, 0.00000545, 0.00000549, 0.00000539, 0.00000514, 0.00000477, 0.00000429, 0.00000374, 0.00000315, 0.00000254, 0.00000195, 0.00000142, 0.00000096, 0.00000061, 0.00000038, 0.00000030, 0.00000035, 0.00000055, 0.00000088, 0.00000133, 0.00000188, 0.00000250, 0.00000317, 0.00000384, 0.00000450, 0.00000511, 0.00000564, 0.00000607, 0.00000638, 0.00000656, 0.00000660, 0.00000650, 0.00000627, 0.00000593, 0.00000549, 0.00000497, 0.00000441, 0.00000382, 0.00000323, 0.00000267, 0.00000216, 0.00000171, 0.00000134, 0.00000105, 0.00000085, 0.00000073, 0.00000068, 0.00000069, 0.00000074, 0.00000081, 0.00000089, 0.00000095, 0.00000098, 0.00000097, 0.00000090, 0.00000077, },
		{ -0.00000089, -0.00000048, 0.00000006, 0.00000070, 0.00000144, 0.00000224, 0.00000308, 0.00000394, 0.00000478, 0.00000557, 0.00000627, 0.00000687, 0.00000734, 0.00000765, 0.00000779, 0.00000776, 0.00000755, 0.00000717, 0.00000663, 0.00000594, 0.00000514, 0.00000424, 0.00000328, 0.00000229, 0.00000131, 0.00000036, -0.00000051, -0.00000128, -0.00000193, -0.00000242, -0.00000275, -0.00000290, -0.00000286, -0.00000264, -0.00000224, -0.00000168, -0.00000096, -0.00000010, 0.00000087, 0.00000192, 0.00000303, 0.00000416, 0.00000530, 0.00000639, 0.00000743, 0.00000837, 0.00000919, 0.00000987, 0.00001038, 0.00001070, 0.00001083, 0.00001075, 0.00001045, 0.00000994, 0.00000922, 0.00000830, 0.00000719, 0.00000592, 0.00000450, 0.00000297, 0.00000136, -0.00000030, -0.00000196, -0.00000359, -0.00000515, -0.00000660, -0.00000790, -0.00000902, -0.00000993, -0.00001061, -0.00001104, -0.00001120, -0.00001111, -0.00001075, -0.00001014, -0.00000931, -0.00000828, -0.00000708, -0.00000576, -0.00000434, -0.00000289, -0.00000144, -0.00000004, 0.00000128, 0.00000248, 0.00000352, 0.00000438, 0.00000505, 0.00000551, 0.00000576, 0.00000582, 0.00000568, 0.00000539, 0.00000495, 0.00000440, 0.00000378, 0.00000311, 0.00000243, 0.00000177, 0.00000117, 0.00000063, 0.00000018, -0.00000017, -0.00000041, -0.00000055, -0.00000059, -0.00000054, -0.00000043, -0.00000027, -0.00000008, 0.00000012, 0.00000030, 0.00000045, 0.00000054, 0.00000058, 0.00000055, 0.00000045, 0.00000029, 0.00000007, -0.00000019, -0.00000047, -0.00000075, -0.00000100, -0.00000121, -0.00000135, -0.00000141, -0.00000135, -0.00000119, },
		{ 0.00000767, 0.00000797, 0.00000809, 0.00000805, 0.00000782, 0.00000742, 0.00000686, 0.00000616, 0.00000533, 0.00000440, 0.00000341, 0.00000238, 0.00000136, 0.00000038, -0.00000053, -0.00000134, -0.00000202, -0.00000256, -0.00000292, -0.00000311, -0.00000312, -0.00000296, -0.00000263, -0.00000216, -0.00000157, -0.00000089, -0.00000015, 0.00000062, 0.00000137, 0.00000207, 0.00000268, 0.00000316, 0.00000349, 0.00000363, 0.00000357, 0.00000329, 0.00000279, 0.00000206, 0.00000111, -0.00000004, -0.00000137, -0.00000285, -0.00000445, -0.00000612, -0.00000782, -0.00000952, -0.00001114, -0.00001266, -0.00001403, -0.00001519, -0.00001612, -0.00001679, -0.00001716, -0.00001722, -0.00001695, -0.00001637, -0.00001547, -0.00001427, -0.00001280, -0.00001109, -0.00000918, -0.00000711, -0.00000493, -0.00000271, -0.00000048, 0.00000170, 0.00000377, 0.00000569, 0.00000742, 0.00000891, 0.00001014, 0.00001109, 0.00001174, 0.00001209, 0.00001215, 0.00001192, 0.00001144, 0.00001073, 0.00000982, 0.00000876, 0.00000759, 0.00000634, 0.00000507, 0.00000381, 0.00000261, 0.00000150, 0.00000051, -0.00000034, -0.00000104, -0.00000158, -0.00000195, -0.00000216, -0.00000222, -0.00000216, -0.00000199, -0.00000173, -0.00000143, -0.00000109, -0.00000076, -0.00000044, -0.00000017, 0.00000005, 0.00000019, 0.00000026, 0.00000025, 0.00000017, 0.00000004, -0.00000014, -0.00000035, -0.00000056, -0.00000075, -0.00000090, -0.00000099, -0.00000100, -0.00000091, -0.00000071, -0.00000040, 0.00000003, 0.00000057, 0.00000121, 0.00000193, 0.00000272, 0.00000354, 0.00000438, 0.00000519, 0.00000596, 0.00000664, 0.00000722, },
		{ -0.00000086, -0.00000184, -0.00000275, -0.00000356, -0.00000424, -0.00000478, -0.00000516, -0.00000538, -0.00000543, -0.00000531, -0.00000505, -0.00000465, -0.00000414, -0.00000355, -0.00000290, -0.00000224, -0.00000159, -0.00000100, -0.00000049, -0.00000009, 0.00000017, 0.00000027, 0.00000020, -0.00000004, -0.00000046, -0.00000104, -0.00000176, -0.00000261, -0.00000355, -0.00000455, -0.00000556, -0.00000656, -0.00000748, -0.00000830, -0.00000897, -0.00000947, -0.00000975, -0.00000980, -0.00000961, -0.00000916, -0.00000846, -0.00000752, -0.00000636, -0.00000501, -0.00000350, -0.00000188, -0.00000019, 0.00000153, 0.00000321, 0.00000482, 0.00000629, 0.00000760, 0.00000870, 0.00000955, 0.00001013, 0.00001044, 0.00001045, 0.00001017, 0.00000962, 0.00000880, 0.00000776, 0.00000652, 0.00000512, 0.00000361, 0.00000203, 0.00000043, -0.00000114, -0.00000264, -0.00000403, -0.00000527, -0.00000632, -0.00000718, -0.00000782, -0.00000823, -0.00000842, -0.00000838, -0.00000814, -0.00000771, -0.00000713, -0.00000641, -0.00000559, -0.00000470, -0.00000379, -0.00000288, -0.00000200, -0.00000118, -0.00000044, 0.00000020, 0.00000072, 0.00000111, 0.00000138, 0.00000154, 0.00000158, 0.00000153, 0.00000139, 0.00000120, 0.00000097, 0.00000074, 0.00000051, 0.00000032, 0.00000019, 0.00000013, 0.00000016, 0.00000028, 0.00000050, 0.00000082, 0.00000123, 0.00000172, 0.00000228, 0.00000289, 0.00000352, 0.00000415, 0.00000475, 0.00000530, 0.00000578, 0.00000615, 0.00000640, 0.00000652, 0.00000648, 0.00000629, 0.00000594, 0.00000544, 0.00000480, 0.00000403, 0.00000315, 0.00000220, 0.00000119, 0.00000016, },
		{ -0.00000364, -0.00000313, -0.00000258, -0.00000200, -0.00000144, -0.00000092, -0.00000046, -0.00000008, 0.00000018, 0.00000032, 0.00000032, 0.00000018, -0.00000009, -0.00000049, -0.00000100, -0.00000160, -0.00000227, -0.00000297, -0.00000368, -0.00000437, -0.00000500, -0.00000555, -0.00000598, -0.00000627, -0.00000642, -0.00000641, -0.00000623, -0.00000590, -0.00000542, -0.00000481, -0.00000410, -0.00000331, -0.00000248, -0.00000165, -0.00000085, -0.00000011, 0.00000053, 0.00000105, 0.00000141, 0.00000161, 0.00000163, 0.00000147, 0.00000114, 0.00000064, 0.00000001, -0.00000073, -0.00000155, -0.00000242, -0.00000330, -0.00000414, -0.00000491, -0.00000557, -0.00000610, -0.00000646, -0.00000663, -0.00000661, -0.00000637, -0.00000594, -0.00000530, -0.00000449, -0.00000352, -0.00000242, -0.00000122, 0.00000004, 0.00000132, 0.00000259, 0.00000381, 0.00000495, 0.00000597, 0.00000685, 0.00000757, 0.00000812, 0.00000849, 0.00000867, 0.00000866, 0.00000849, 0.00000815, 0.00000768, 0.00000709, 0.00000641, 0.00000566, 0.00000488, 0.00000409, 0.00000333, 0.00000260, 0.00000194, 0.00000137, 0.00000089, 0.00000053, 0.00000029, 0.00000016, 0.00000016, 0.00000027, 0.00000049, 0.00000080, 0.00000118, 0.00000163, 0.00000211, 0.00000261, 0.00000310, 0.00000357, 0.00000398, 0.00000434, 0.00000461, 0.00000478, 0.00000484, 0.00000478, 0.00000460, 0.00000430, 0.00000388, 0.00000335, 0.00000273, 0.00000202, 0.00000126, 0.00000045, -0.00000037, -0.00000118, -0.00000196, -0.00000268, -0.00000332, -0.00000386, -0.00000428, -0.00000458, -0.00000474, -0.00000477, -0.00000466, -0.00000443, -0.00000408, },
		{ -0.00000169, -0.00000189, -0.00000216, -0.00000250, -0.00000289, -0.00000331, -0.00000375, -0.00000417, -0.00000455, -0.00000489, -0.00000514, -0.00000532, -0.00000539, -0.00000535, -0.00000521, -0.00000497, -0.00000463, -0.00000422, -0.00000374, -0.00000322, -0.00000268, -0.00000215, -0.00000164, -0.00000117, -0.00000077, -0.00000044, -0.00000019, -0.00000003, 0.00000004, 0.00000004, -0.00000002, -0.00000012, -0.00000025, -0.00000038, -0.00000049, -0.00000055, -0.00000053, -0.00000043, -0.00000022, 0.00000010, 0.00000054, 0.00000110, 0.00000176, 0.00000252, 0.00000335, 0.00000422, 0.00000511, 0.00000598, 0.00000680, 0.00000754, 0.00000816, 0.00000863, 0.00000894, 0.00000906, 0.00000898, 0.00000870, 0.00000821, 0.00000754, 0.00000669, 0.00000569, 0.00000456, 0.00000335, 0.00000208, 0.00000079, -0.00000049, -0.00000171, -0.00000284, -0.00000386, -0.00000475, -0.00000547, -0.00000602, -0.00000640, -0.00000659, -0.00000660, -0.00000644, -0.00000612, -0.00000567, -0.00000509, -0.00000441, -0.00000365, -0.00000284, -0.00000200, -0.00000115, -0.00000031, 0.00000050, 0.00000127, 0.00000197, 0.00000261, 0.00000315, 0.00000361, 0.00000397, 0.00000423, 0.00000439, 0.00000444, 0.00000439, 0.00000423, 0.00000398, 0.00000364, 0.00000321, 0.00000271, 0.00000214, 0.00000153, 0.00000087, 0.00000019, -0.00000050, -0.00000117, -0.00000183, -0.00000243, -0.00000299, -0.00000347, -0.00000386, -0.00000417, -0.00000437, -0.00000448, -0.00000448, -0.00000440, -0.00000423, -0.00000398, -0.00000369, -0.00000335, -0.00000300, -0.00000264, -0.00000231, -0.00000203, -0.00000180, -0.00000164, -0.00000157, -0.00000159, },
		{ -0.00000367, -0.00000363, -0.00000355, -0.00000342, -0.00000324, -0.00000303, -0.00000278, -0.00000251, -0.00000222, -0.00000194, -0.00000168, -0.00000144, -0.00000123, -0.00000106, -0.00000093, -0.00000084, -0.00000078, -0.00000076, -0.00000075, -0.00000075, -0.00000074, -0.00000072, -0.00000066, -0.00000056, -0.00000040, -0.00000019, 0.00000008, 0.00000040, 0.00000077, 0.00000117, 0.00000158, 0.00000199, 0.00000237, 0.00000269, 0.00000294, 0.00000308, 0.00000311, 0.00000300, 0.00000274, 0.00000233, 0.00000176, 0.00000106, 0.00000023, -0.00000070, -0.00000171, -0.00000276, -0.00000382, -0.00000484, -0.00000579, -0.00000663, -0.00000733, -0.00000785, -0.00000818, -0.00000828, -0.00000817, -0.00000782, -0.00000726, -0.00000649, -0.00000553, -0.00000443, -0.00000320, -0.00000189, -0.00000054, 0.00000082, 0.00000215, 0.00000340, 0.00000456, 0.00000558, 0.00000646, 0.00000718, 0.00000772, 0.00000809, 0.00000829, 0.00000833, 0.00000822, 0.00000798, 0.00000763, 0.00000718, 0.00000666, 0.00000608, 0.00000546, 0.00000481, 0.00000415, 0.00000349, 0.00000282, 0.00000215, 0.00000149, 0.00000083, 0.00000017, -0.00000049, -0.00000115, -0.00000182, -0.00000248, -0.00000313, -0.00000378, -0.00000440, -0.00000500, -0.00000556, -0.00000607, -0.00000652, -0.00000690, -0.00000719, -0.00000740, -0.00000751, -0.00000753, -0.00000746, -0.00000730, -0.00000705, -0.00000675, -0.00000639, -0.00000599, -0.00000557, -0.00000515, -0.00000474, -0.00000437, -0.00000403, -0.00000375, -0.00000353, -0.00000337, -0.00000327, -0.00000323, -0.00000324, -0.00000329, -0.00000336, -0.00000345, -0.00000354, -0.00000361, -0.00000366, },
		{ -0.00000074, -0.00000074, -0.00000074, -0.00000073, -0.00000071, -0.00000068, -0.00000064, -0.00000058, -0.00000050, -0.00000038, -0.00000022, -0.00000003, 0.00000020, 0.00000047, 0.00000076, 0.00000109, 0.00000142, 0.00000176, 0.00000208, 0.00000237, 0.00000260, 0.00000277, 0.00000286, 0.00000285, 0.00000274, 0.00000253, 0.00000221, 0.00000180, 0.00000130, 0.00000074, 0.00000013, -0.00000048, -0.00000108, -0.00000164, -0.00000211, -0.00000247, -0.00000270, -0.00000277, -0.00000267, -0.00000239, -0.00000192, -0.00000129, -0.00000049, 0.00000044, 0.00000147, 0.00000257, 0.00000369, 0.00000479, 0.00000583, 0.00000676, 0.00000754, 0.00000814, 0.00000854, 0.00000871, 0.00000864, 0.00000833, 0.00000779, 0.00000704, 0.00000610, 0.00000501, 0.00000379, 0.00000251, 0.00000119, -0.00000012, -0.00000138, -0.00000255, -0.00000361, -0.00000452, -0.00000528, -0.00000588, -0.00000630, -0.00000658, -0.00000671, -0.00000672, -0.00000663, -0.00000648, -0.00000628, -0.00000608, -0.00000589, -0.00000573, -0.00000563, -0.00000560, -0.00000564, -0.00000575, -0.00000593, -0.00000615, -0.00000642, -0.00000670, -0.00000698, -0.00000723, -0.00000743, -0.00000756, -0.00000760, -0.00000754, -0.00000737, -0.00000710, -0.00000671, -0.00000623, -0.00000566, -0.00000502, -0.00000433, -0.00000361, -0.00000290, -0.00000219, -0.00000153, -0.00000093, -0.00000040, 0.00000005, 0.00000041, 0.00000067, 0.00000084, 0.00000093, 0.00000094, 0.00000088, 0.00000077, 0.00000062, 0.00000044, 0.00000026, 0.00000007, -0.00000010, -0.00000026, -0.00000040, -0.00000051, -0.00000059, -0.00000065, -0.00000070, -0.00000072, -0.00000074, },
		{ 0.00000080, 0.00000110, 0.00000140, 0.00000170, 0.00000199, 0.00000226, 0.00000249, 0.00000267, 0.00000279, 0.00000284, 0.00000282, 0.00000271, 0.00000251, 0.00000223, 0.00000188, 0.00000145, 0.00000097, 0.00000046, -0.00000006, -0.00000057, -0.00000105, -0.00000147, -0.00000180, -0.00000204, -0.00000215, -0.00000213, -0.00000199, -0.00000171, -0.00000131, -0.00000080, -0.00000022, 0.00000042, 0.00000108, 0.00000172, 0.00000232, 0.00000282, 0.00000321, 0.00000345, 0.00000352, 0.00000341, 0.00000311, 0.00000262, 0.00000196, 0.00000114, 0.00000021, -0.00000081, -0.00000188, -0.00000293, -0.00000394, -0.00000485, -0.00000561, -0.00000620, -0.00000657, -0.00000672, -0.00000662, -0.00000627, -0.00000570, -0.00000491, -0.00000394, -0.00000283, -0.00000162, -0.00000037, 0.00000088, 0.00000207, 0.00000315, 0.00000407, 0.00000481, 0.00000532, 0.00000561, 0.00000565, 0.00000545, 0.00000502, 0.00000440, 0.00000361, 0.00000269, 0.00000169, 0.00000066, -0.00000036, -0.00000132, -0.00000217, -0.00000288, -0.00000342, -0.00000377, -0.00000392, -0.00000387, -0.00000362, -0.00000320, -0.00000262, -0.00000191, -0.00000111, -0.00000025, 0.00000062, 0.00000147, 0.00000227, 0.00000298, 0.00000359, 0.00000407, 0.00000441, 0.00000461, 0.00000467, 0.00000460, 0.00000440, 0.00000410, 0.00000371, 0.00000326, 0.00000277, 0.00000226, 0.00000174, 0.00000124, 0.00000078, 0.00000036, -0.00000001, -0.00000033, -0.00000058, -0.00000077, -0.00000091, -0.00000099, -0.00000103, -0.00000102, -0.00000097, -0.00000088, -0.00000077, -0.00000062, -0.00000044, -0.00000024, -0.00000001, 0.00000024, 0.00000051, },
		{ 0.00000153, 0.00000128, 0.00000098, 0.00000064, 0.00000028, -0.00000010, -0.00000048, -0.00000083, -0.00000114, -0.00000138, -0.00000155, -0.00000163, -0.00000160, -0.00000147, -0.00000123, -0.00000088, -0.00000044, 0.00000008, 0.00000065, 0.00000126, 0.00000187, 0.00000246, 0.00000300, 0.00000347, 0.00000384, 0.00000409, 0.00000421, 0.00000420, 0.00000406, 0.00000380, 0.00000343, 0.00000299, 0.00000250, 0.00000200, 0.00000152, 0.00000110, 0.00000078, 0.00000058, 0.00000054, 0.00000067, 0.00000098, 0.00000147, 0.00000213, 0.00000295, 0.00000389, 0.00000492, 0.00000598, 0.00000702, 0.00000801, 0.00000887, 0.00000956, 0.00001003, 0.00001024, 0.00001017, 0.00000979, 0.00000910, 0.00000811, 0.00000683, 0.00000531, 0.00000357, 0.00000168, -0.00000030, -0.00000231, -0.00000428, -0.00000614, -0.00000783, -0.00000930, -0.00001049, -0.00001137, -0.00001191, -0.00001210, -0.00001194, -0.00001145, -0.00001065, -0.00000958, -0.00000830, -0.00000684, -0.00000527, -0.00000365, -0.00000204, -0.00000049, 0.00000094, 0.00000222, 0.00000330, 0.00000417, 0.00000480, 0.00000520, 0.00000537, 0.00000533, 0.00000508, 0.00000467, 0.00000412, 0.00000347, 0.00000275, 0.00000200, 0.00000125, 0.00000053, -0.00000014, -0.00000074, -0.00000125, -0.00000166, -0.00000198, -0.00000219, -0.00000232, -0.00000235, -0.00000231, -0.00000220, -0.00000204, -0.00000183, -0.00000158, -0.00000132, -0.00000103, -0.00000073, -0.00000043, -0.00000013, 0.00000016, 0.00000045, 0.00000073, 0.00000099, 0.00000123, 0.00000144, 0.00000163, 0.00000177, 0.00000187, 0.00000192, 0.00000191, 0.00000185, 0.00000172, },
		{ -0.00000076, -0.00000062, -0.00000040, -0.00000011, 0.00000025, 0.00000067, 0.00000113, 0.00000161, 0.00000210, 0.00000256, 0.00000299, 0.00000335, 0.00000363, 0.00000382, 0.00000392, 0.00000390, 0.00000379, 0.00000358, 0.00000330, 0.00000295, 0.00000256, 0.00000217, 0.00000179, 0.00000145, 0.00000117, 0.00000098, 0.00000089, 0.00000091, 0.00000104, 0.00000128, 0.00000160, 0.00000199, 0.00000243, 0.00000286, 0.00000327, 0.00000360, 0.00000381, 0.00000387, 0.00000375, 0.00000342, 0.00000286, 0.00000207, 0.00000104, -0.00000020, -0.00000163, -0.00000322, -0.00000491, -0.00000665, -0.00000838, -0.00001004, -0.00001156, -0.00001288, -0.00001395, -0.00001471, -0.00001513, -0.00001518, -0.00001485, -0.00001413, -0.00001306, -0.00001165, -0.00000994, -0.00000799, -0.00000587, -0.00000363, -0.00000135, 0.00000089, 0.00000303, 0.00000501, 0.00000676, 0.00000824, 0.00000941, 0.00001026, 0.00001076, 0.00001093, 0.00001077, 0.00001031, 0.00000959, 0.00000864, 0.00000752, 0.00000628, 0.00000497, 0.00000364, 0.00000233, 0.00000109, -0.00000005, -0.00000105, -0.00000191, -0.00000260, -0.00000313, -0.00000349, -0.00000370, -0.00000377, -0.00000371, -0.00000355, -0.00000330, -0.00000299, -0.00000265, -0.00000228, -0.00000190, -0.00000152, -0.00000116, -0.00000082, -0.00000050, -0.00000021, 0.00000005, 0.00000028, 0.00000049, 0.00000067, 0.00000083, 0.00000096, 0.00000106, 0.00000113, 0.00000118, 0.00000119, 0.00000116, 0.00000109, 0.00000099, 0.00000085, 0.00000068, 0.00000048, 0.00000026, 0.00000003, -0.00000020, -0.00000041, -0.00000059, -0.00000072, -0.00000080, -0.00000082, },
		{ 0.00000226, 0.00000246, 0.00000261, 0.00000268, 0.00000268, 0.00000261, 0.00000249, 0.00000231, 0.00000210, 0.00000187, 0.00000164, 0.00000143, 0.00000126, 0.00000113, 0.00000107, 0.00000107, 0.00000113, 0.00000126, 0.00000144, 0.00000165, 0.00000188, 0.00000209, 0.00000227, 0.00000238, 0.00000240, 0.00000231, 0.00000208, 0.00000171, 0.00000118, 0.00000050, -0.00000031, -0.00000126, -0.00000229, -0.00000338, -0.00000449, -0.00000558, -0.00000659, -0.00000747, -0.00000819, -0.00000870, -0.00000897, -0.00000897, -0.00000868, -0.00000810, -0.00000725, -0.00000613, -0.00000478, -0.00000325, -0.00000157, 0.00000018, 0.00000195, 0.00000369, 0.00000531, 0.00000677, 0.00000802, 0.00000900, 0.00000969, 0.00001005, 0.00001010, 0.00000981, 0.00000922, 0.00000834, 0.00000722, 0.00000590, 0.00000443, 0.00000286, 0.00000126, -0.00000032, -0.00000182, -0.00000321, -0.00000444, -0.00000547, -0.00000629, -0.00000688, -0.00000724, -0.00000738, -0.00000731, -0.00000706, -0.00000664, -0.00000609, -0.00000545, -0.00000475, -0.00000401, -0.00000328, -0.00000257, -0.00000190, -0.00000129, -0.00000076, -0.00000030, 0.00000008, 0.00000039, 0.00000063, 0.00000082, 0.00000096, 0.00000105, 0.00000112, 0.00000117, 0.00000121, 0.00000123, 0.00000125, 0.00000126, 0.00000126, 0.00000124, 0.00000120, 0.00000115, 0.00000107, 0.00000096, 0.00000083, 0.00000067, 0.00000049, 0.00000030, 0.00000011, -0.00000008, -0.00000025, -0.00000039, -0.00000050, -0.00000055, -0.00000055, -0.00000049, -0.00000036, -0.00000018, 0.00000006, 0.00000035, 0.00000067, 0.00000101, 0.00000136, 0.00000169, 0.00000199, },
		{ 0.00000349, 0.00000340, 0.00000332, 0.00000325, 0.00000320, 0.00000317, 0.00000314, 0.00000313, 0.00000310, 0.00000306, 0.00000297, 0.00000284, 0.00000264, 0.00000235, 0.00000197, 0.00000150, 0.00000092, 0.00000025, -0.00000051, -0.00000133, -0.00000220, -0.00000309, -0.00000397, -0.00000481, -0.00000557, -0.00000622, -0.00000672, -0.00000706, -0.00000721, -0.00000716, -0.00000690, -0.00000644, -0.00000578, -0.00000496, -0.00000400, -0.00000293, -0.00000180, -0.00000065, 0.00000048, 0.00000153, 0.00000247, 0.00000326, 0.00000387, 0.00000428, 0.00000447, 0.00000443, 0.00000419, 0.00000374, 0.00000312, 0.00000236, 0.00000150, 0.00000057, -0.00000036, -0.00000127, -0.00000210, -0.00000283, -0.00000341, -0.00000382, -0.00000404, -0.00000408, -0.00000392, -0.00000358, -0.00000309, -0.00000245, -0.00000171, -0.00000090, -0.00000005, 0.00000079, 0.00000160, 0.00000234, 0.00000298, 0.00000351, 0.00000392, 0.00000419, 0.00000432, 0.00000433, 0.00000422, 0.00000401, 0.00000373, 0.00000339, 0.00000301, 0.00000263, 0.00000226, 0.00000191, 0.00000161, 0.00000135, 0.00000115, 0.00000100, 0.00000090, 0.00000084, 0.00000081, 0.00000081, 0.00000080, 0.00000079, 0.00000077, 0.00000072, 0.00000063, 0.00000052, 0.00000037, 0.00000020, 0.00000001, -0.00000019, -0.00000038, -0.00000056, -0.00000070, -0.00000079, -0.00000083, -0.00000081, -0.00000071, -0.00000055, -0.00000031, -0.00000002, 0.00000033, 0.00000072, 0.00000113, 0.00000155, 0.00000197, 0.00000237, 0.00000272, 0.00000304, 0.00000329, 0.00000349, 0.00000363, 0.00000371, 0.00000373, 0.00000371, 0.00000366, 0.00000358, },
		{ 0.00000076, 0.00000029, -0.00000022, -0.00000076, -0.00000134, -0.00000196, -0.00000261, -0.00000327, -0.00000393, -0.00000457, -0.00000517, -0.00000571, -0.00000617, -0.00000653, -0.00000676, -0.00000685, -0.00000680, -0.00000658, -0.00000621, -0.00000569, -0.00000503, -0.00000425, -0.00000337, -0.00000242, -0.00000144, -0.00000044, 0.00000052, 0.00000143, 0.00000225, 0.00000295, 0.00000352, 0.00000395, 0.00000423, 0.00000435, 0.00000433, 0.00000418, 0.00000393, 0.00000358, 0.00000318, 0.00000276, 0.00000233, 0.00000193, 0.00000159, 0.00000131, 0.00000113, 0.00000104, 0.00000104, 0.00000113, 0.00000131, 0.00000154, 0.00000182, 0.00000211, 0.00000238, 0.00000263, 0.00000281, 0.00000291, 0.00000290, 0.00000279, 0.00000255, 0.00000220, 0.00000174, 0.00000118, 0.00000055, -0.00000014, -0.00000085, -0.00000157, -0.00000224, -0.00000286, -0.00000340, -0.00000383, -0.00000413, -0.00000431, -0.00000435, -0.00000426, -0.00000405, -0.00000374, -0.00000334, -0.00000287, -0.00000238, -0.00000187, -0.00000138, -0.00000093, -0.00000054, -0.00000023, -0.00000000, 0.00000013, 0.00000017, 0.00000013, 0.00000002, -0.00000016, -0.00000036, -0.00000058, -0.00000079, -0.00000098, -0.00000111, -0.00000117, -0.00000115, -0.00000104, -0.00000084, -0.00000054, -0.00000016, 0.00000029, 0.00000080, 0.00000136, 0.00000193, 0.00000251, 0.00000306, 0.00000357, 0.00000403, 0.00000442, 0.00000473, 0.00000495, 0.00000508, 0.00000513, 0.00000510, 0.00000499, 0.00000482, 0.00000460, 0.00000434, 0.00000404, 0.00000373, 0.00000340, 0.00000306, 0.00000271, 0.00000235, 0.00000198, 0.00000160, 0.00000119, },
		{ -0.00000610, -0.00000613, -0.00000608, -0.00000592, -0.00000567, -0.00000531, -0.00000486, -0.00000431, -0.00000368, -0.00000298, -0.00000224, -0.00000146, -0.00000068, 0.00000008, 0.00000079, 0.00000145, 0.00000202, 0.00000250, 0.00000286, 0.00000311, 0.00000323, 0.00000324, 0.00000313, 0.00000293, 0.00000264, 0.00000229, 0.00000189, 0.00000147, 0.00000104, 0.00000062, 0.00000023, -0.00000012, -0.00000041, -0.00000066, -0.00000086, -0.00000101, -0.00000112, -0.00000121, -0.00000129, -0.00000137, -0.00000147, -0.00000160, -0.00000177, -0.00000199, -0.00000226, -0.00000259, -0.00000296, -0.00000336, -0.00000378, -0.00000420, -0.00000460, -0.00000496, -0.00000524, -0.00000543, -0.00000552, -0.00000547, -0.00000530, -0.00000498, -0.00000452, -0.00000393, -0.00000323, -0.00000244, -0.00000158, -0.00000067, 0.00000023, 0.00000111, 0.00000192, 0.00000264, 0.00000323, 0.00000368, 0.00000397, 0.00000409, 0.00000404, 0.00000382, 0.00000345, 0.00000295, 0.00000235, 0.00000167, 0.00000096, 0.00000024, -0.00000045, -0.00000107, -0.00000160, -0.00000201, -0.00000227, -0.00000239, -0.00000234, -0.00000214, -0.00000180, -0.00000132, -0.00000074, -0.00000007, 0.00000064, 0.00000139, 0.00000212, 0.00000282, 0.00000345, 0.00000400, 0.00000445, 0.00000477, 0.00000497, 0.00000505, 0.00000499, 0.00000482, 0.00000454, 0.00000417, 0.00000372, 0.00000321, 0.00000265, 0.00000207, 0.00000147, 0.00000087, 0.00000028, -0.00000029, -0.00000085, -0.00000139, -0.00000190, -0.00000239, -0.00000286, -0.00000331, -0.00000375, -0.00000416, -0.00000456, -0.00000492, -0.00000526, -0.00000556, -0.00000580, -0.00000599, },
		{ 0.00000069, 0.00000114, 0.00000155, 0.00000189, 0.00000217, 0.00000238, 0.00000250, 0.00000254, 0.00000250, 0.00000238, 0.00000219, 0.00000194, 0.00000164, 0.00000132, 0.00000097, 0.00000063, 0.00000029, -0.00000002, -0.00000029, -0.00000053, -0.00000071, -0.00000086, -0.00000095, -0.00000101, -0.00000104, -0.00000104, -0.00000103, -0.00000102, -0.00000102, -0.00000104, -0.00000108, -0.00000114, -0.00000123, -0.00000134, -0.00000147, -0.00000160, -0.00000173, -0.00000184, -0.00000191, -0.00000194, -0.00000190, -0.00000178, -0.00000159, -0.00000131, -0.00000095, -0.00000050, 0.00000001, 0.00000057, 0.00000116, 0.00000176, 0.00000233, 0.00000286, 0.00000331, 0.00000365, 0.00000386, 0.00000393, 0.00000383, 0.00000356, 0.00000312, 0.00000251, 0.00000176, 0.00000087, -0.00000011, -0.00000116, -0.00000224, -0.00000330, -0.00000431, -0.00000522, -0.00000601, -0.00000662, -0.00000705, -0.00000726, -0.00000725, -0.00000702, -0.00000656, -0.00000589, -0.00000504, -0.00000404, -0.00000291, -0.00000170, -0.00000045, 0.00000080, 0.00000200, 0.00000313, 0.00000414, 0.00000500, 0.00000570, 0.00000622, 0.00000654, 0.00000668, 0.00000662, 0.00000640, 0.00000602, 0.00000550, 0.00000488, 0.00000416, 0.00000339, 0.00000259, 0.00000178, 0.00000099, 0.00000022, -0.00000050, -0.00000116, -0.00000175, -0.00000228, -0.00000273, -0.00000312, -0.00000344, -0.00000370, -0.00000390, -0.00000405, -0.00000415, -0.00000420, -0.00000421, -0.00000417, -0.00000409, -0.00000397, -0.00000380, -0.00000358, -0.00000331, -0.00000299, -0.00000262, -0.00000221, -0.00000177, -0.00000129, -0.00000080, -0.00000029, 0.00000021, },
		{ 0.00000053, 0.00000029, 0.00000005, -0.00000016, -0.00000035, -0.00000050, -0.00000062, -0.00000070, -0.00000074, -0.00000075, -0.00000072, -0.00000067, -0.00000061, -0.00000054, -0.00000048, -0.00000042, -0.00000039, -0.00000039, -0.00000042, -0.00000048, -0.00000057, -0.00000069, -0.00000084, -0.00000100, -0.00000116, -0.00000131, -0.00000144, -0.00000154, -0.00000160, -0.00000161, -0.00000157, -0.00000147, -0.00000131, -0.00000111, -0.00000086, -0.00000060, -0.00000032, -0.00000005, 0.00000019, 0.00000038, 0.00000050, 0.00000054, 0.00000048, 0.00000032, 0.00000004, -0.00000035, -0.00000084, -0.00000142, -0.00000207, -0.00000276, -0.00000346, -0.00000415, -0.00000479, -0.00000535, -0.00000578, -0.00000607, -0.00000619, -0.00000611, -0.00000583, -0.00000533, -0.00000462, -0.00000372, -0.00000263, -0.00000139, -0.00000003, 0.00000142, 0.00000291, 0.00000439, 0.00000583, 0.00000718, 0.00000841, 0.00000947, 0.00001033, 0.00001098, 0.00001139, 0.00001157, 0.00001151, 0.00001122, 0.00001071, 0.00001002, 0.00000916, 0.00000817, 0.00000708, 0.00000593, 0.00000475, 0.00000358, 0.00000244, 0.00000136, 0.00000037, -0.00000052, -0.00000130, -0.00000196, -0.00000249, -0.00000291, -0.00000320, -0.00000339, -0.00000349, -0.00000349, -0.00000343, -0.00000330, -0.00000312, -0.00000291, -0.00000266, -0.00000240, -0.00000212, -0.00000182, -0.00000152, -0.00000122, -0.00000091, -0.00000060, -0.00000030, 0.00000000, 0.00000029, 0.00000056, 0.00000082, 0.00000106, 0.00000126, 0.00000143, 0.00000156, 0.00000165, 0.00000169, 0.00000168, 0.00000163, 0.00000153, 0.00000138, 0.00000120, 0.00000100, 0.00000077, },
		{ -0.00000079, -0.00000065, -0.00000053, -0.00000044, -0.00000039, -0.00000037, -0.00000039, -0.00000046, -0.00000056, -0.00000069, -0.00000084, -0.00000101, -0.00000119, -0.00000135, -0.00000149, -0.00000161, -0.00000168, -0.00000170, -0.00000167, -0.00000159, -0.00000145, -0.00000126, -0.00000103, -0.00000077, -0.00000049, -0.00000021, 0.00000005, 0.00000029, 0.00000048, 0.00000061, 0.00000068, 0.00000066, 0.00000057, 0.00000040, 0.00000016, -0.00000015, -0.00000050, -0.00000087, -0.00000125, -0.00000161, -0.00000193, -0.00000217, -0.00000233, -0.00000236, -0.00000227, -0.00000204, -0.00000166, -0.00000113, -0.00000046, 0.00000033, 0.00000123, 0.00000222, 0.00000325, 0.00000431, 0.00000535, 0.00000634, 0.00000725, 0.00000803, 0.00000868, 0.00000915, 0.00000942, 0.00000950, 0.00000938, 0.00000904, 0.00000852, 0.00000781, 0.00000694, 0.00000595, 0.00000485, 0.00000369, 0.00000249, 0.00000129, 0.00000012, -0.00000098, -0.00000200, -0.00000291, -0.00000369, -0.00000433, -0.00000483, -0.00000518, -0.00000538, -0.00000545, -0.00000540, -0.00000524, -0.00000498, -0.00000466, -0.00000427, -0.00000384, -0.00000339, -0.00000293, -0.00000248, -0.00000204, -0.00000163, -0.00000125, -0.00000091, -0.00000060, -0.00000034, -0.00000012, 0.00000006, 0.00000021, 0.00000031, 0.00000038, 0.00000042, 0.00000042, 0.00000039, 0.00000034, 0.00000025, 0.00000014, 0.00000001, -0.00000014, -0.00000030, -0.00000048, -0.00000066, -0.00000085, -0.00000103, -0.00000119, -0.00000134, -0.00000146, -0.00000155, -0.00000161, -0.00000163, -0.00000162, -0.00000157, -0.00000149, -0.00000138, -0.00000125, -0.00000110, -0.00000095, },
		{ -0.00000118, -0.00000138, -0.00000156, -0.00000170, -0.00000181, -0.00000187, -0.00000187, -0.00000182, -0.00000172, -0.00000157, -0.00000138, -0.00000116, -0.00000092, -0.00000069, -0.00000046, -0.00000026, -0.00000010, 0.00000000, 0.00000004, 0.00000001, -0.00000009, -0.00000027, -0.00000051, -0.00000082, -0.00000117, -0.00000155, -0.00000195, -0.00000233, -0.00000269, -0.00000299, -0.00000322, -0.00000336, -0.00000340, -0.00000332, -0.00000313, -0.00000281, -0.00000237, -0.00000182, -0.00000118, -0.00000046, 0.00000032, 0.00000112, 0.00000192, 0.00000269, 0.00000341, 0.00000405, 0.00000458, 0.00000499, 0.00000527, 0.00000539, 0.00000537, 0.00000519, 0.00000488, 0.00000443, 0.00000386, 0.00000320, 0.00000246, 0.00000168, 0.00000088, 0.00000008, -0.00000068, -0.00000140, -0.00000205, -0.00000261, -0.00000307, -0.00000342, -0.00000367, -0.00000380, -0.00000383, -0.00000376, -0.00000360, -0.00000337, -0.00000308, -0.00000274, -0.00000238, -0.00000200, -0.00000163, -0.00000127, -0.00000094, -0.00000064, -0.00000038, -0.00000017, -0.00000001, 0.00000010, 0.00000015, 0.00000016, 0.00000013, 0.00000005, -0.00000007, -0.00000022, -0.00000039, -0.00000059, -0.00000081, -0.00000103, -0.00000127, -0.00000150, -0.00000173, -0.00000194, -0.00000215, -0.00000233, -0.00000249, -0.00000262, -0.00000271, -0.00000277, -0.00000279, -0.00000277, -0.00000271, -0.00000260, -0.00000246, -0.00000228, -0.00000208, -0.00000185, -0.00000160, -0.00000135, -0.00000110, -0.00000086, -0.00000065, -0.00000046, -0.00000031, -0.00000021, -0.00000015, -0.00000015, -0.00000019, -0.00000028, -0.00000041, -0.00000058, -0.00000077, -0.00000098, },
		{ -0.00000112, -0.00000092, -0.00000073, -0.00000056, -0.00000043, -0.00000035, -0.00000032, -0.00000035, -0.00000045, -0.00000061, -0.00000083, -0.00000110, -0.00000141, -0.00000174, -0.00000208, -0.00000241, -0.00000270, -0.00000294, -0.00000311, -0.00000320, -0.00000318, -0.00000306, -0.00000282, -0.00000246, -0.00000199, -0.00000141, -0.00000074, 0.00000001, 0.00000082, 0.00000166, 0.00000251, 0.00000334, 0.00000413, 0.00000485, 0.00000548, 0.00000600, 0.00000640, 0.00000665, 0.00000676, 0.00000672, 0.00000654, 0.00000621, 0.00000576, 0.00000520, 0.00000454, 0.00000381, 0.00000303, 0.00000221, 0.00000140, 0.00000060, -0.00000016, -0.00000087, -0.00000151, -0.00000207, -0.00000254, -0.00000291, -0.00000319, -0.00000337, -0.00000347, -0.00000348, -0.00000343, -0.00000331, -0.00000314, -0.00000293, -0.00000270, -0.00000246, -0.00000221, -0.00000198, -0.00000176, -0.00000156, -0.00000140, -0.00000127, -0.00000118, -0.00000113, -0.00000112, -0.00000115, -0.00000121, -0.00000130, -0.00000142, -0.00000156, -0.00000172, -0.00000188, -0.00000205, -0.00000222, -0.00000238, -0.00000252, -0.00000264, -0.00000273, -0.00000279, -0.00000281, -0.00000280, -0.00000274, -0.00000264, -0.00000249, -0.00000231, -0.00000209, -0.00000184, -0.00000156, -0.00000126, -0.00000096, -0.00000066, -0.00000036, -0.00000009, 0.00000015, 0.00000036, 0.00000052, 0.00000063, 0.00000068, 0.00000068, 0.00000062, 0.00000050, 0.00000034, 0.00000013, -0.00000010, -0.00000036, -0.00000063, -0.00000089, -0.00000115, -0.00000137, -0.00000156, -0.00000171, -0.00000180, -0.00000184, -0.00000183, -0.00000176, -0.00000165, -0.00000150, -0.00000132, },
		{ -0.00000070, -0.00000096, -0.00000121, -0.00000145, -0.00000166, -0.00000182, -0.00000191, -0.00000193, -0.00000186, -0.00000169, -0.00000142, -0.00000104, -0.00000058, -0.00000002, 0.00000061, 0.00000130, 0.00000203, 0.00000278, 0.00000353, 0.00000425, 0.00000491, 0.00000551, 0.00000601, 0.00000639, 0.00000665, 0.00000677, 0.00000674, 0.00000657, 0.00000626, 0.00000580, 0.00000523, 0.00000454, 0.00000376, 0.00000292, 0.00000203, 0.00000111, 0.00000020, -0.00000069, -0.00000153, -0.00000230, -0.00000298, -0.00000357, -0.00000405, -0.00000441, -0.00000464, -0.00000476, -0.00000476, -0.00000465, -0.00000444, -0.00000415, -0.00000377, -0.00000335, -0.00000288, -0.00000238, -0.00000188, -0.00000138, -0.00000091, -0.00000047, -0.00000007, 0.00000027, 0.00000055, 0.00000076, 0.00000091, 0.00000099, 0.00000100, 0.00000095, 0.00000084, 0.00000069, 0.00000049, 0.00000027, 0.00000002, -0.00000023, -0.00000048, -0.00000072, -0.00000094, -0.00000112, -0.00000127, -0.00000137, -0.00000143, -0.00000143, -0.00000138, -0.00000129, -0.00000114, -0.00000095, -0.00000073, -0.00000048, -0.00000022, 0.00000005, 0.00000032, 0.00000058, 0.00000081, 0.00000100, 0.00000116, 0.00000126, 0.00000131, 0.00000131, 0.00000125, 0.00000114, 0.00000097, 0.00000077, 0.00000053, 0.00000026, -0.00000002, -0.00000030, -0.00000057, -0.00000082, -0.00000104, -0.00000122, -0.00000135, -0.00000143, -0.00000146, -0.00000143, -0.00000136, -0.00000124, -0.00000108, -0.00000090, -0.00000071, -0.00000051, -0.00000032, -0.00000015, -0.00000002, 0.00000007, 0.00000012, 0.00000010, 0.00000004, -0.00000008, -0.00000025, -0.00000046, },
		{ -0.00000022, 0.00000026, 0.00000079, 0.00000136, 0.00000195, 0.00000255, 0.00000313, 0.00000367, 0.00000416, 0.00000458, 0.00000491, 0.00000514, 0.00000526, 0.00000526, 0.00000514, 0.00000489, 0.00000453, 0.00000406, 0.00000349, 0.00000284, 0.00000213, 0.00000138, 0.00000060, -0.00000017, -0.00000092, -0.00000163, -0.00000228, -0.00000285, -0.00000332, -0.00000369, -0.00000395, -0.00000409, -0.00000412, -0.00000403, -0.00000383, -0.00000354, -0.00000317, -0.00000272, -0.00000223, -0.00000170, -0.00000115, -0.00000061, -0.00000008, 0.00000040, 0.00000084, 0.00000120, 0.00000150, 0.00000171, 0.00000183, 0.00000187, 0.00000182, 0.00000168, 0.00000147, 0.00000119, 0.00000085, 0.00000047, 0.00000006, -0.00000037, -0.00000081, -0.00000123, -0.00000162, -0.00000198, -0.00000228, -0.00000253, -0.00000271, -0.00000282, -0.00000286, -0.00000283, -0.00000272, -0.00000255, -0.00000232, -0.00000204, -0.00000173, -0.00000138, -0.00000102, -0.00000067, -0.00000032, 0.00000001, 0.00000030, 0.00000055, 0.00000075, 0.00000089, 0.00000097, 0.00000099, 0.00000095, 0.00000085, 0.00000071, 0.00000052, 0.00000030, 0.00000006, -0.00000020, -0.00000045, -0.00000069, -0.00000091, -0.00000110, -0.00000125, -0.00000136, -0.00000143, -0.00000144, -0.00000141, -0.00000134, -0.00000123, -0.00000109, -0.00000093, -0.00000076, -0.00000060, -0.00000044, -0.00000030, -0.00000019, -0.00000012, -0.00000009, -0.00000011, -0.00000016, -0.00000026, -0.00000040, -0.00000056, -0.00000074, -0.00000093, -0.00000112, -0.00000129, -0.00000142, -0.00000152, -0.00000155, -0.00000152, -0.00000142, -0.00000123, -0.00000097, -0.00000063, },
		{ 0.00000364, 0.00000356, 0.00000339, 0.00000313, 0.00000279, 0.00000238, 0.00000191, 0.00000139, 0.00000083, 0.00000025, -0.00000033, -0.00000090, -0.00000144, -0.00000194, -0.00000238, -0.00000275, -0.00000305, -0.00000326, -0.00000338, -0.00000342, -0.00000337, -0.00000324, -0.00000304, -0.00000278, -0.00000246, -0.00000211, -0.00000173, -0.00000135, -0.00000096, -0.00000060, -0.00000027, 0.00000001, 0.00000025, 0.00000042, 0.00000053, 0.00000056, 0.00000053, 0.00000043, 0.00000027, 0.00000006, -0.00000021, -0.00000051, -0.00000083, -0.00000116, -0.00000150, -0.00000182, -0.00000211, -0.00000237, -0.00000258, -0.00000273, -0.00000282, -0.00000285, -0.00000281, -0.00000270, -0.00000253, -0.00000230, -0.00000202, -0.00000169, -0.00000133, -0.00000094, -0.00000055, -0.00000015, 0.00000023, 0.00000058, 0.00000090, 0.00000118, 0.00000140, 0.00000157, 0.00000167, 0.00000171, 0.00000168, 0.00000159, 0.00000145, 0.00000126, 0.00000102, 0.00000076, 0.00000047, 0.00000017, -0.00000013, -0.00000043, -0.00000070, -0.00000095, -0.00000116, -0.00000133, -0.00000146, -0.00000155, -0.00000159, -0.00000159, -0.00000155, -0.00000148, -0.00000139, -0.00000128, -0.00000115, -0.00000103, -0.00000091, -0.00000081, -0.00000072, -0.00000066, -0.00000063, -0.00000063, -0.00000065, -0.00000070, -0.00000078, -0.00000087, -0.00000096, -0.00000106, -0.00000115, -0.00000122, -0.00000126, -0.00000127, -0.00000123, -0.00000115, -0.00000102, -0.00000083, -0.00000059, -0.00000030, 0.00000004, 0.00000041, 0.00000081, 0.00000123, 0.00000166, 0.00000207, 0.00000247, 0.00000282, 0.00000313, 0.00000337, 0.00000354, 0.00000364, },
		{ -0.00000069, -0.00000098, -0.00000123, -0.00000144, -0.00000160, -0.00000171, -0.00000177, -0.00000179, -0.00000175, -0.00000167, -0.00000156, -0.00000141, -0.00000123, -0.00000104, -0.00000083, -0.00000062, -0.00000042, -0.00000023, -0.00000005, 0.00000010, 0.00000022, 0.00000031, 0.00000037, 0.00000039, 0.00000038, 0.00000035, 0.00000028, 0.00000020, 0.00000010, -0.00000001, -0.00000012, -0.00000023, -0.00000033, -0.00000041, -0.00000047, -0.00000051, -0.00000052, -0.00000049, -0.00000043, -0.00000034, -0.00000022, -0.00000007, 0.00000010, 0.00000030, 0.00000051, 0.00000072, 0.00000094, 0.00000115, 0.00000134, 0.00000152, 0.00000166, 0.00000178, 0.00000186, 0.00000190, 0.00000191, 0.00000186, 0.00000178, 0.00000166, 0.00000151, 0.00000132, 0.00000110, 0.00000087, 0.00000062, 0.00000037, 0.00000012, -0.00000012, -0.00000035, -0.00000056, -0.00000075, -0.00000091, -0.00000103, -0.00000112, -0.00000118, -0.00000120, -0.00000119, -0.00000115, -0.00000109, -0.00000100, -0.00000090, -0.00000079, -0.00000067, -0.00000055, -0.00000044, -0.00000033, -0.00000023, -0.00000015, -0.00000008, -0.00000002, 0.00000002, 0.00000005, 0.00000006, 0.00000007, 0.00000008, 0.00000009, 0.00000011, 0.00000014, 0.00000018, 0.00000024, 0.00000031, 0.00000042, 0.00000054, 0.00000069, 0.00000086, 0.00000106, 0.00000126, 0.00000148, 0.00000170, 0.00000192, 0.00000214, 0.00000233, 0.00000251, 0.00000265, 0.00000276, 0.00000283, 0.00000286, 0.00000284, 0.00000276, 0.00000264, 0.00000247, 0.00000225, 0.00000199, 0.00000170, 0.00000138, 0.00000103, 0.00000068, 0.00000032, -0.00000004, -0.00000038, },
		},
		{ { 0.00000243, 0.00000231, 0.00000198, 0.00000153, 0.00000109, 0.00000080, 0.00000079, 0.00000110, 0.00000175, 0.00000266, 0.00000368, 0.00000465, 0.00000536, 0.00000567, 0.00000547, 0.00000474, 0.00000354, 0.00000200, 0.00000032, -0.00000133, -0.00000274, -0.00000381, -0.00000448, -0.00000478, -0.00000481, -0.00000471, -0.00000463, -0.00000470, -0.00000500, -0.00000550, -0.00000611, -0.00000667, -0.00000700, -0.00000688, -0.00000617, -0.00000480, -0.00000281, -0.00000036, 0.00000233, 0.00000493, 0.00000713, 0.00000865, 0.00000929, 0.00000899, 0.00000779, 0.00000589, 0.00000359, 0.00000121, -0.00000090, -0.00000247, -0.00000333, -0.00000344, -0.00000289, -0.00000187, -0.00000067, 0.00000045, 0.00000121, 0.00000144, 0.00000109, 0.00000022, -0.00000102, -0.00000238, -0.00000360, -0.00000444, -0.00000474, -0.00000444, -0.00000360, -0.00000238, -0.00000102, 0.00000022, 0.00000109, 0.00000144, 0.00000121, 0.00000045, -0.00000067, -0.00000187, -0.00000289, -0.00000344, -0.00000333, -0.00000247, -0.00000090, 0.00000121, 0.00000359, 0.00000589, 0.00000779, 0.00000899, 0.00000929, 0.00000865, 0.00000713, 0.00000493, 0.00000233, -0.00000036, -0.00000281, -0.00000480, -0.00000617, -0.00000688, -0.00000700, -0.00000667, -0.00000611, -0.00000550, -0.00000500, -0.00000470, -0.00000463, -0.00000471, -0.00000481, -0.00000478, -0.00000448, -0.00000381, -0.00000274, -0.00000133, 0.00000032, 0.00000200, 0.00000354, 0.00000474, 0.00000547, 0.00000567, 0.00000536, 0.00000465, 0.00000368, 0.00000266, 0.00000175, 0.00000110, 0.00000079, 0.00000080, 0.00000109, 0.00000153, 0.00000198, 0.00000231, },
		{ 0.00000166, 0.00000149, 0.00000154, 0.00000176, 0.00000204, 0.00000224, 0.00000224, 0.00000193, 0.00000128, 0.00000029, -0.00000095, -0.00000230, -0.00000360, -0.00000472, -0.00000554, -0.00000600, -0.00000612, -0.00000597, -0.00000568, -0.00000538, -0.00000519, -0.00000518, -0.00000537, -0.00000567, -0.00000597, -0.00000607, -0.00000580, -0.00000503, -0.00000367, -0.00000175, 0.00000060, 0.00000317, 0.00000567, 0.00000779, 0.00000926, 0.00000989, 0.00000959, 0.00000841, 0.00000651, 0.00000418, 0.00000174, -0.00000047, -0.00000216, -0.00000315, -0.00000337, -0.00000290, -0.00000192, -0.00000069, 0.00000048, 0.00000133, 0.00000166, 0.00000139, 0.00000056, -0.00000069, -0.00000213, -0.00000348, -0.00000449, -0.00000498, -0.00000485, -0.00000414, -0.00000300, -0.00000164, -0.00000034, 0.00000065, 0.00000113, 0.00000102, 0.00000034, -0.00000075, -0.00000201, -0.00000316, -0.00000390, -0.00000400, -0.00000336, -0.00000196, 0.00000005, 0.00000242, 0.00000481, 0.00000689, 0.00000835, 0.00000897, 0.00000866, 0.00000745, 0.00000552, 0.00000313, 0.00000061, -0.00000174, -0.00000366, -0.00000498, -0.00000565, -0.00000573, -0.00000537, -0.00000477, -0.00000414, -0.00000366, -0.00000342, -0.00000346, -0.00000368, -0.00000397, -0.00000413, -0.00000400, -0.00000345, -0.00000243, -0.00000098, 0.00000079, 0.00000266, 0.00000442, 0.00000582, 0.00000670, 0.00000694, 0.00000655, 0.00000562, 0.00000432, 0.00000288, 0.00000153, 0.00000048, -0.00000013, -0.00000024, 0.00000011, 0.00000081, 0.00000169, 0.00000257, 0.00000328, 0.00000371, 0.00000380, 0.00000358, 0.00000312, 0.00000257, 0.00000204, },
		{ -0.00000270, -0.00000276, -0.00000301, -0.00000341, -0.00000392, -0.00000444, -0.00000486, -0.00000510, -0.00000511, -0.00000488, -0.00000448, -0.00000400, -0.00000355, -0.00000324, -0.00000313, -0.00000324, -0.00000351, -0.00000382, -0.00000402, -0.00000394, -0.00000343, -0.00000240, -0.00000085, 0.00000112, 0.00000333, 0.00000554, 0.00000745, 0.00000882, 0.00000945, 0.00000922, 0.00000818, 0.00000646, 0.00000430, 0.00000202, -0.00000008, -0.00000172, -0.00000271, -0.00000298, -0.00000258, -0.00000169, -0.00000055, 0.00000054, 0.00000132, 0.00000158, 0.00000124, 0.00000033, -0.00000102, -0.00000257, -0.00000405, -0.00000520, -0.00000580, -0.00000575, -0.00000508, -0.00000392, -0.00000249, -0.00000108, 0.00000005, 0.00000068, 0.00000070, 0.00000011, -0.00000095, -0.00000224, -0.00000345, -0.00000428, -0.00000445, -0.00000382, -0.00000236, -0.00000018, 0.00000246, 0.00000524, 0.00000777, 0.00000972, 0.00001082, 0.00001092, 0.00001001, 0.00000825, 0.00000587, 0.00000320, 0.00000056, -0.00000175, -0.00000354, -0.00000470, -0.00000526, -0.00000535, -0.00000514, -0.00000483, -0.00000462, -0.00000459, -0.00000479, -0.00000513, -0.00000547, -0.00000563, -0.00000542, -0.00000469, -0.00000339, -0.00000154, 0.00000071, 0.00000313, 0.00000544, 0.00000739, 0.00000872, 0.00000929, 0.00000905, 0.00000809, 0.00000659, 0.00000481, 0.00000302, 0.00000150, 0.00000044, -0.00000003, 0.00000008, 0.00000067, 0.00000157, 0.00000256, 0.00000341, 0.00000393, 0.00000401, 0.00000362, 0.00000281, 0.00000171, 0.00000049, -0.00000070, -0.00000169, -0.00000240, -0.00000281, -0.00000294, -0.00000289, -0.00000277, },
		{ -0.00000191, -0.00000186, -0.00000175, -0.00000158, -0.00000140, -0.00000128, -0.00000128, -0.00000147, -0.00000183, -0.00000233, -0.00000286, -0.00000328, -0.00000344, -0.00000318, -0.00000242, -0.00000114, 0.00000059, 0.00000263, 0.00000473, 0.00000665, 0.00000813, 0.00000896, 0.00000901, 0.00000828, 0.00000686, 0.00000496, 0.00000285, 0.00000082, -0.00000086, -0.00000200, -0.00000251, -0.00000241, -0.00000184, -0.00000102, -0.00000021, 0.00000034, 0.00000045, 0.00000002, -0.00000093, -0.00000227, -0.00000377, -0.00000517, -0.00000622, -0.00000672, -0.00000655, -0.00000575, -0.00000443, -0.00000283, -0.00000122, 0.00000011, 0.00000093, 0.00000112, 0.00000069, -0.00000024, -0.00000142, -0.00000254, -0.00000326, -0.00000333, -0.00000255, -0.00000091, 0.00000150, 0.00000440, 0.00000743, 0.00001020, 0.00001231, 0.00001345, 0.00001345, 0.00001228, 0.00001007, 0.00000710, 0.00000371, 0.00000028, -0.00000282, -0.00000534, -0.00000711, -0.00000813, -0.00000848, -0.00000833, -0.00000791, -0.00000740, -0.00000696, -0.00000665, -0.00000644, -0.00000620, -0.00000577, -0.00000498, -0.00000368, -0.00000182, 0.00000057, 0.00000333, 0.00000621, 0.00000890, 0.00001110, 0.00001253, 0.00001301, 0.00001247, 0.00001100, 0.00000879, 0.00000614, 0.00000340, 0.00000092, -0.00000104, -0.00000229, -0.00000277, -0.00000256, -0.00000184, -0.00000086, 0.00000010, 0.00000081, 0.00000107, 0.00000081, 0.00000003, -0.00000115, -0.00000255, -0.00000395, -0.00000516, -0.00000602, -0.00000644, -0.00000640, -0.00000597, -0.00000526, -0.00000443, -0.00000360, -0.00000291, -0.00000241, -0.00000210, -0.00000197, -0.00000193, },
		{ -0.00000178, -0.00000223, -0.00000272, -0.00000317, -0.00000349, -0.00000354, -0.00000322, -0.00000243, -0.00000115, 0.00000055, 0.00000256, 0.00000465, 0.00000660, 0.00000816, 0.00000911, 0.00000931, 0.00000873, 0.00000743, 0.00000558, 0.00000343, 0.00000125, -0.00000068, -0.00000216, -0.00000306, -0.00000336, -0.00000317, -0.00000265, -0.00000203, -0.00000154, -0.00000137, -0.00000161, -0.00000228, -0.00000327, -0.00000440, -0.00000542, -0.00000613, -0.00000632, -0.00000591, -0.00000489, -0.00000341, -0.00000168, 0.00000005, 0.00000148, 0.00000241, 0.00000271, 0.00000237, 0.00000153, 0.00000041, -0.00000066, -0.00000138, -0.00000146, -0.00000074, 0.00000081, 0.00000308, 0.00000579, 0.00000858, 0.00001103, 0.00001274, 0.00001341, 0.00001285, 0.00001107, 0.00000823, 0.00000463, 0.00000069, -0.00000317, -0.00000654, -0.00000911, -0.00001071, -0.00001130, -0.00001101, -0.00001004, -0.00000866, -0.00000716, -0.00000575, -0.00000455, -0.00000358, -0.00000277, -0.00000197, -0.00000098, 0.00000034, 0.00000208, 0.00000422, 0.00000665, 0.00000913, 0.00001136, 0.00001302, 0.00001382, 0.00001358, 0.00001224, 0.00000987, 0.00000670, 0.00000308, -0.00000059, -0.00000389, -0.00000650, -0.00000816, -0.00000878, -0.00000843, -0.00000730, -0.00000567, -0.00000390, -0.00000231, -0.00000116, -0.00000063, -0.00000073, -0.00000141, -0.00000247, -0.00000368, -0.00000480, -0.00000561, -0.00000596, -0.00000579, -0.00000514, -0.00000411, -0.00000288, -0.00000162, -0.00000050, 0.00000035, 0.00000088, 0.00000108, 0.00000101, 0.00000074, 0.00000037, -0.00000003, -0.00000040, -0.00000074, -0.00000106, -0.00000139, },
		{ -0.00000132, -0.00000015, 0.00000135, 0.00000309, 0.00000491, 0.00000660, 0.00000794, 0.00000875, 0.00000887, 0.00000824, 0.00000693, 0.00000506, 0.00000285, 0.00000057, -0.00000150, -0.00000316, -0.00000425, -0.00000471, -0.00000461, -0.00000409, -0.00000336, -0.00000262, -0.00000209, -0.00000189, -0.00000205, -0.00000253, -0.00000318, -0.00000382, -0.00000423, -0.00000426, -0.00000379, -0.00000283, -0.00000148, 0.00000008, 0.00000161, 0.00000286, 0.00000362, 0.00000378, 0.00000334, 0.00000239, 0.00000117, -0.00000004, -0.00000093, -0.00000124, -0.00000081, 0.00000039, 0.00000227, 0.00000456, 0.00000692, 0.00000895, 0.00001028, 0.00001063, 0.00000985, 0.00000796, 0.00000515, 0.00000174, -0.00000186, -0.00000519, -0.00000787, -0.00000961, -0.00001026, -0.00000985, -0.00000855, -0.00000665, -0.00000449, -0.00000240, -0.00000066, 0.00000057, 0.00000127, 0.00000154, 0.00000156, 0.00000158, 0.00000184, 0.00000248, 0.00000356, 0.00000502, 0.00000665, 0.00000817, 0.00000926, 0.00000961, 0.00000900, 0.00000735, 0.00000472, 0.00000134, -0.00000243, -0.00000617, -0.00000941, -0.00001179, -0.00001302, -0.00001299, -0.00001177, -0.00000959, -0.00000679, -0.00000379, -0.00000100, 0.00000121, 0.00000262, 0.00000312, 0.00000276, 0.00000174, 0.00000032, -0.00000119, -0.00000251, -0.00000341, -0.00000376, -0.00000354, -0.00000284, -0.00000182, -0.00000069, 0.00000037, 0.00000117, 0.00000162, 0.00000169, 0.00000141, 0.00000087, 0.00000019, -0.00000051, -0.00000115, -0.00000166, -0.00000203, -0.00000229, -0.00000247, -0.00000261, -0.00000272, -0.00000280, -0.00000277, -0.00000258, -0.00000211, },
		{ 0.00000721, 0.00000735, 0.00000687, 0.00000578, 0.00000417, 0.00000223, 0.00000018, -0.00000172, -0.00000325, -0.00000426, -0.00000467, -0.00000450, -0.00000388, -0.00000299, -0.00000204, -0.00000123, -0.00000073, -0.00000059, -0.00000081, -0.00000130, -0.00000188, -0.00000237, -0.00000260, -0.00000246, -0.00000191, -0.00000102, 0.00000009, 0.00000121, 0.00000212, 0.00000263, 0.00000264, 0.00000211, 0.00000114, -0.00000007, -0.00000128, -0.00000220, -0.00000258, -0.00000226, -0.00000119, 0.00000053, 0.00000270, 0.00000499, 0.00000705, 0.00000854, 0.00000918, 0.00000884, 0.00000753, 0.00000541, 0.00000278, 0.00000002, -0.00000248, -0.00000438, -0.00000545, -0.00000559, -0.00000488, -0.00000353, -0.00000184, -0.00000017, 0.00000114, 0.00000186, 0.00000188, 0.00000122, 0.00000006, -0.00000131, -0.00000257, -0.00000342, -0.00000362, -0.00000308, -0.00000185, -0.00000015, 0.00000169, 0.00000331, 0.00000433, 0.00000445, 0.00000353, 0.00000159, -0.00000116, -0.00000437, -0.00000759, -0.00001033, -0.00001218, -0.00001281, -0.00001209, -0.00001010, -0.00000707, -0.00000341, 0.00000040, 0.00000388, 0.00000659, 0.00000824, 0.00000868, 0.00000797, 0.00000630, 0.00000402, 0.00000149, -0.00000089, -0.00000281, -0.00000406, -0.00000456, -0.00000436, -0.00000362, -0.00000255, -0.00000141, -0.00000042, 0.00000027, 0.00000057, 0.00000050, 0.00000012, -0.00000043, -0.00000103, -0.00000154, -0.00000189, -0.00000205, -0.00000204, -0.00000190, -0.00000170, -0.00000150, -0.00000133, -0.00000117, -0.00000097, -0.00000067, -0.00000017, 0.00000057, 0.00000157, 0.00000279, 0.00000413, 0.00000544, 0.00000652, },
		{ -0.00000037, -0.00000186, -0.00000295, -0.00000354, -0.00000360, -0.00000321, -0.00000253, -0.00000173, -0.00000103, -0.00000058, -0.00000047, -0.00000073, -0.00000129, -0.00000201, -0.00000272, -0.00000324, -0.00000344, -0.00000325, -0.00000270, -0.00000187, -0.00000093, -0.00000006, 0.00000054, 0.00000076, 0.00000056, -0.00000002, -0.00000082, -0.00000163, -0.00000220, -0.00000231, -0.00000178, -0.00000056, 0.00000128, 0.00000356, 0.00000601, 0.00000828, 0.00001006, 0.00001107, 0.00001117, 0.00001033, 0.00000867, 0.00000642, 0.00000392, 0.00000149, -0.00000056, -0.00000203, -0.00000283, -0.00000302, -0.00000278, -0.00000236, -0.00000207, -0.00000215, -0.00000276, -0.00000395, -0.00000561, -0.00000748, -0.00000925, -0.00001056, -0.00001110, -0.00001067, -0.00000920, -0.00000682, -0.00000382, -0.00000061, 0.00000233, 0.00000455, 0.00000570, 0.00000558, 0.00000420, 0.00000178, -0.00000129, -0.00000450, -0.00000730, -0.00000919, -0.00000981, -0.00000900, -0.00000681, -0.00000351, 0.00000045, 0.00000454, 0.00000818, 0.00001089, 0.00001230, 0.00001227, 0.00001084, 0.00000826, 0.00000491, 0.00000128, -0.00000216, -0.00000500, -0.00000695, -0.00000787, -0.00000781, -0.00000692, -0.00000548, -0.00000379, -0.00000216, -0.00000082, 0.00000009, 0.00000052, 0.00000054, 0.00000025, -0.00000018, -0.00000060, -0.00000090, -0.00000102, -0.00000095, -0.00000076, -0.00000051, -0.00000028, -0.00000013, -0.00000008, -0.00000010, -0.00000011, -0.00000001, 0.00000028, 0.00000083, 0.00000166, 0.00000271, 0.00000387, 0.00000498, 0.00000585, 0.00000634, 0.00000630, 0.00000571, 0.00000460, 0.00000308, 0.00000135, },
		{ -0.00000386, -0.00000328, -0.00000281, -0.00000258, -0.00000265, -0.00000298, -0.00000349, -0.00000402, -0.00000442, -0.00000453, -0.00000427, -0.00000363, -0.00000266, -0.00000150, -0.00000032, 0.00000070, 0.00000142, 0.00000176, 0.00000174, 0.00000146, 0.00000108, 0.00000082, 0.00000089, 0.00000144, 0.00000256, 0.00000420, 0.00000623, 0.00000839, 0.00001040, 0.00001195, 0.00001276, 0.00001268, 0.00001164, 0.00000974, 0.00000715, 0.00000417, 0.00000112, -0.00000173, -0.00000413, -0.00000597, -0.00000725, -0.00000806, -0.00000860, -0.00000907, -0.00000965, -0.00001046, -0.00001149, -0.00001263, -0.00001366, -0.00001429, -0.00001423, -0.00001326, -0.00001126, -0.00000827, -0.00000450, -0.00000032, 0.00000380, 0.00000734, 0.00000984, 0.00001096, 0.00001053, 0.00000864, 0.00000557, 0.00000180, -0.00000207, -0.00000542, -0.00000770, -0.00000852, -0.00000772, -0.00000537, -0.00000181, 0.00000245, 0.00000679, 0.00001056, 0.00001320, 0.00001432, 0.00001375, 0.00001158, 0.00000811, 0.00000381, -0.00000075, -0.00000499, -0.00000840, -0.00001065, -0.00001157, -0.00001121, -0.00000976, -0.00000757, -0.00000505, -0.00000256, -0.00000045, 0.00000108, 0.00000195, 0.00000220, 0.00000196, 0.00000143, 0.00000080, 0.00000026, -0.00000010, -0.00000023, -0.00000017, 0.00000000, 0.00000019, 0.00000030, 0.00000031, 0.00000021, 0.00000007, -0.00000001, 0.00000007, 0.00000043, 0.00000111, 0.00000208, 0.00000328, 0.00000454, 0.00000568, 0.00000650, 0.00000683, 0.00000658, 0.00000572, 0.00000434, 0.00000257, 0.00000065, -0.00000121, -0.00000279, -0.00000392, -0.00000455, -0.00000467, -0.00000439, },
		{ -0.00000254, -0.00000252, -0.00000234, -0.00000191, -0.00000116, -0.00000012, 0.00000111, 0.00000239, 0.00000355, 0.00000444, 0.00000497, 0.00000509, 0.00000486, 0.00000443, 0.00000397, 0.00000367, 0.00000370, 0.00000417, 0.00000508, 0.00000634, 0.00000775, 0.00000906, 0.00000999, 0.00001030, 0.00000980, 0.00000843, 0.00000623, 0.00000336, 0.00000009, -0.00000329, -0.00000648, -0.00000922, -0.00001136, -0.00001284, -0.00001370, -0.00001407, -0.00001411, -0.00001399, -0.00001382, -0.00001364, -0.00001339, -0.00001294, -0.00001209, -0.00001066, -0.00000852, -0.00000562, -0.00000208, 0.00000188, 0.00000588, 0.00000949, 0.00001225, 0.00001377, 0.00001381, 0.00001230, 0.00000942, 0.00000553, 0.00000117, -0.00000302, -0.00000642, -0.00000850, -0.00000891, -0.00000756, -0.00000462, -0.00000051, 0.00000418, 0.00000874, 0.00001248, 0.00001483, 0.00001541, 0.00001409, 0.00001102, 0.00000662, 0.00000144, -0.00000383, -0.00000853, -0.00001212, -0.00001422, -0.00001468, -0.00001359, -0.00001125, -0.00000809, -0.00000461, -0.00000130, 0.00000145, 0.00000337, 0.00000437, 0.00000451, 0.00000397, 0.00000302, 0.00000193, 0.00000095, 0.00000024, -0.00000011, -0.00000014, 0.00000007, 0.00000039, 0.00000067, 0.00000084, 0.00000085, 0.00000074, 0.00000059, 0.00000053, 0.00000068, 0.00000112, 0.00000188, 0.00000289, 0.00000403, 0.00000511, 0.00000592, 0.00000626, 0.00000600, 0.00000507, 0.00000353, 0.00000151, -0.00000077, -0.00000305, -0.00000506, -0.00000661, -0.00000754, -0.00000782, -0.00000752, -0.00000677, -0.00000576, -0.00000471, -0.00000378, -0.00000310, -0.00000271, -0.00000256, },
		{ 0.00000431, 0.00000504, 0.00000558, 0.00000580, 0.00000566, 0.00000515, 0.00000437, 0.00000348, 0.00000266, 0.00000207, 0.00000186, 0.00000208, 0.00000267, 0.00000351, 0.00000439, 0.00000506, 0.00000529, 0.00000488, 0.00000374, 0.00000188, -0.00000058, -0.00000342, -0.00000638, -0.00000916, -0.00001150, -0.00001323, -0.00001426, -0.00001459, -0.00001432, -0.00001361, -0.00001263, -0.00001151, -0.00001036, -0.00000919, -0.00000793, -0.00000649, -0.00000475, -0.00000261, -0.00000007, 0.00000280, 0.00000581, 0.00000867, 0.00001103, 0.00001255, 0.00001296, 0.00001209, 0.00000997, 0.00000681, 0.00000298, -0.00000101, -0.00000458, -0.00000719, -0.00000840, -0.00000799, -0.00000594, -0.00000252, 0.00000181, 0.00000640, 0.00001056, 0.00001365, 0.00001513, 0.00001473, 0.00001242, 0.00000846, 0.00000335, -0.00000226, -0.00000764, -0.00001212, -0.00001515, -0.00001642, -0.00001585, -0.00001365, -0.00001020, -0.00000606, -0.00000182, 0.00000198, 0.00000489, 0.00000667, 0.00000726, 0.00000679, 0.00000554, 0.00000387, 0.00000213, 0.00000066, -0.00000035, -0.00000079, -0.00000073, -0.00000028, 0.00000033, 0.00000092, 0.00000131, 0.00000141, 0.00000122, 0.00000083, 0.00000038, 0.00000003, -0.00000008, 0.00000014, 0.00000069, 0.00000147, 0.00000232, 0.00000303, 0.00000340, 0.00000324, 0.00000246, 0.00000108, -0.00000077, -0.00000290, -0.00000503, -0.00000687, -0.00000819, -0.00000880, -0.00000864, -0.00000775, -0.00000630, -0.00000449, -0.00000259, -0.00000083, 0.00000060, 0.00000160, 0.00000216, 0.00000235, 0.00000232, 0.00000222, 0.00000223, 0.00000244, 0.00000289, 0.00000355, },
		{ 0.00000055, -0.00000033, -0.00000104, -0.00000144, -0.00000143, -0.00000100, -0.00000023, 0.00000071, 0.00000161, 0.00000222, 0.00000236, 0.00000187, 0.00000073, -0.00000100, -0.00000316, -0.00000552, -0.00000781, -0.00000977, -0.00001120, -0.00001199, -0.00001211, -0.00001164, -0.00001070, -0.00000946, -0.00000808, -0.00000670, -0.00000537, -0.00000409, -0.00000280, -0.00000142, 0.00000013, 0.00000189, 0.00000384, 0.00000584, 0.00000771, 0.00000919, 0.00001004, 0.00001005, 0.00000910, 0.00000723, 0.00000461, 0.00000154, -0.00000154, -0.00000418, -0.00000595, -0.00000652, -0.00000573, -0.00000363, -0.00000049, 0.00000324, 0.00000700, 0.00001017, 0.00001222, 0.00001275, 0.00001157, 0.00000876, 0.00000465, -0.00000026, -0.00000530, -0.00000979, -0.00001315, -0.00001493, -0.00001492, -0.00001319, -0.00001003, -0.00000593, -0.00000149, 0.00000267, 0.00000603, 0.00000819, 0.00000901, 0.00000851, 0.00000695, 0.00000471, 0.00000224, -0.00000003, -0.00000175, -0.00000273, -0.00000291, -0.00000243, -0.00000151, -0.00000044, 0.00000049, 0.00000103, 0.00000106, 0.00000059, -0.00000029, -0.00000135, -0.00000236, -0.00000308, -0.00000335, -0.00000309, -0.00000236, -0.00000130, -0.00000014, 0.00000087, 0.00000149, 0.00000157, 0.00000106, 0.00000001, -0.00000139, -0.00000290, -0.00000422, -0.00000509, -0.00000531, -0.00000480, -0.00000358, -0.00000181, 0.00000027, 0.00000238, 0.00000423, 0.00000560, 0.00000636, 0.00000647, 0.00000600, 0.00000513, 0.00000408, 0.00000305, 0.00000224, 0.00000176, 0.00000163, 0.00000180, 0.00000212, 0.00000245, 0.00000262, 0.00000252, 0.00000211, 0.00000141, },
		{ -0.00000013, 0.00000094, 0.00000186, 0.00000242, 0.00000245, 0.00000185, 0.00000062, -0.00000111, -0.00000317, -0.00000529, -0.00000724, -0.00000877, -0.00000974, -0.00001007, -0.00000979, -0.00000900, -0.00000785, -0.00000651, -0.00000513, -0.00000381, -0.00000260, -0.00000147, -0.00000037, 0.00000077, 0.00000202, 0.00000337, 0.00000476, 0.00000609, 0.00000716, 0.00000779, 0.00000781, 0.00000711, 0.00000570, 0.00000368, 0.00000130, -0.00000113, -0.00000324, -0.00000467, -0.00000516, -0.00000458, -0.00000296, -0.00000051, 0.00000241, 0.00000534, 0.00000779, 0.00000933, 0.00000964, 0.00000861, 0.00000631, 0.00000304, -0.00000075, -0.00000451, -0.00000768, -0.00000981, -0.00001057, -0.00000988, -0.00000787, -0.00000487, -0.00000136, 0.00000212, 0.00000502, 0.00000694, 0.00000762, 0.00000703, 0.00000533, 0.00000287, 0.00000009, -0.00000253, -0.00000460, -0.00000581, -0.00000606, -0.00000542, -0.00000411, -0.00000247, -0.00000088, 0.00000030, 0.00000083, 0.00000061, -0.00000032, -0.00000174, -0.00000336, -0.00000481, -0.00000577, -0.00000600, -0.00000539, -0.00000398, -0.00000196, 0.00000036, 0.00000264, 0.00000453, 0.00000578, 0.00000624, 0.00000591, 0.00000492, 0.00000353, 0.00000206, 0.00000080, 0.00000003, -0.00000010, 0.00000041, 0.00000147, 0.00000285, 0.00000427, 0.00000545, 0.00000612, 0.00000615, 0.00000550, 0.00000424, 0.00000258, 0.00000077, -0.00000092, -0.00000226, -0.00000307, -0.00000331, -0.00000302, -0.00000235, -0.00000150, -0.00000068, -0.00000008, 0.00000017, 0.00000005, -0.00000039, -0.00000100, -0.00000161, -0.00000203, -0.00000212, -0.00000180, -0.00000110, },
		{ 0.00000041, -0.00000143, -0.00000349, -0.00000550, -0.00000722, -0.00000844, -0.00000903, -0.00000898, -0.00000835, -0.00000727, -0.00000592, -0.00000446, -0.00000305, -0.00000178, -0.00000067, 0.00000029, 0.00000116, 0.00000201, 0.00000287, 0.00000375, 0.00000458, 0.00000525, 0.00000563, 0.00000559, 0.00000503, 0.00000392, 0.00000235, 0.00000046, -0.00000149, -0.00000321, -0.00000441, -0.00000484, -0.00000438, -0.00000303, -0.00000092, 0.00000166, 0.00000434, 0.00000671, 0.00000840, 0.00000913, 0.00000875, 0.00000732, 0.00000502, 0.00000220, -0.00000071, -0.00000328, -0.00000515, -0.00000606, -0.00000596, -0.00000494, -0.00000327, -0.00000131, 0.00000051, 0.00000181, 0.00000233, 0.00000193, 0.00000068, -0.00000122, -0.00000342, -0.00000552, -0.00000714, -0.00000798, -0.00000788, -0.00000687, -0.00000513, -0.00000299, -0.00000084, 0.00000093, 0.00000199, 0.00000215, 0.00000142, -0.00000004, -0.00000190, -0.00000376, -0.00000518, -0.00000578, -0.00000533, -0.00000374, -0.00000116, 0.00000213, 0.00000570, 0.00000905, 0.00001175, 0.00001342, 0.00001387, 0.00001309, 0.00001125, 0.00000867, 0.00000576, 0.00000296, 0.00000062, -0.00000097, -0.00000173, -0.00000171, -0.00000110, -0.00000019, 0.00000070, 0.00000126, 0.00000131, 0.00000075, -0.00000035, -0.00000182, -0.00000340, -0.00000481, -0.00000578, -0.00000614, -0.00000582, -0.00000488, -0.00000347, -0.00000186, -0.00000031, 0.00000093, 0.00000167, 0.00000185, 0.00000150, 0.00000076, -0.00000016, -0.00000102, -0.00000159, -0.00000174, -0.00000140, -0.00000065, 0.00000038, 0.00000144, 0.00000231, 0.00000274, 0.00000259, 0.00000180, },
		{ -0.00000841, -0.00000805, -0.00000713, -0.00000583, -0.00000435, -0.00000289, -0.00000159, -0.00000053, 0.00000027, 0.00000086, 0.00000129, 0.00000166, 0.00000202, 0.00000237, 0.00000269, 0.00000289, 0.00000288, 0.00000257, 0.00000193, 0.00000098, -0.00000020, -0.00000143, -0.00000250, -0.00000318, -0.00000327, -0.00000265, -0.00000130, 0.00000068, 0.00000307, 0.00000558, 0.00000784, 0.00000952, 0.00001035, 0.00001018, 0.00000898, 0.00000690, 0.00000419, 0.00000120, -0.00000170, -0.00000418, -0.00000598, -0.00000698, -0.00000721, -0.00000681, -0.00000603, -0.00000515, -0.00000444, -0.00000409, -0.00000419, -0.00000470, -0.00000546, -0.00000623, -0.00000674, -0.00000675, -0.00000611, -0.00000479, -0.00000291, -0.00000069, 0.00000153, 0.00000341, 0.00000463, 0.00000499, 0.00000443, 0.00000308, 0.00000121, -0.00000077, -0.00000241, -0.00000329, -0.00000309, -0.00000165, 0.00000098, 0.00000452, 0.00000853, 0.00001248, 0.00001580, 0.00001796, 0.00001863, 0.00001765, 0.00001511, 0.00001130, 0.00000670, 0.00000189, -0.00000257, -0.00000618, -0.00000861, -0.00000971, -0.00000955, -0.00000839, -0.00000660, -0.00000461, -0.00000284, -0.00000158, -0.00000103, -0.00000117, -0.00000187, -0.00000288, -0.00000387, -0.00000457, -0.00000473, -0.00000427, -0.00000322, -0.00000172, -0.00000003, 0.00000155, 0.00000274, 0.00000333, 0.00000324, 0.00000250, 0.00000127, -0.00000019, -0.00000160, -0.00000267, -0.00000318, -0.00000304, -0.00000227, -0.00000101, 0.00000048, 0.00000191, 0.00000299, 0.00000348, 0.00000324, 0.00000227, 0.00000066, -0.00000136, -0.00000352, -0.00000553, -0.00000712, -0.00000811, },
		{ -0.00000131, -0.00000045, 0.00000007, 0.00000033, 0.00000044, 0.00000052, 0.00000064, 0.00000083, 0.00000108, 0.00000133, 0.00000149, 0.00000150, 0.00000131, 0.00000094, 0.00000048, 0.00000004, -0.00000022, -0.00000015, 0.00000037, 0.00000138, 0.00000282, 0.00000455, 0.00000634, 0.00000792, 0.00000900, 0.00000936, 0.00000884, 0.00000741, 0.00000516, 0.00000231, -0.00000085, -0.00000398, -0.00000675, -0.00000890, -0.00001027, -0.00001081, -0.00001059, -0.00000978, -0.00000860, -0.00000728, -0.00000603, -0.00000496, -0.00000410, -0.00000342, -0.00000280, -0.00000208, -0.00000116, 0.00000005, 0.00000153, 0.00000316, 0.00000478, 0.00000614, 0.00000702, 0.00000725, 0.00000674, 0.00000557, 0.00000392, 0.00000210, 0.00000049, -0.00000051, -0.00000059, 0.00000042, 0.00000255, 0.00000560, 0.00000918, 0.00001279, 0.00001585, 0.00001782, 0.00001828, 0.00001699, 0.00001397, 0.00000947, 0.00000396, -0.00000192, -0.00000750, -0.00001213, -0.00001530, -0.00001673, -0.00001636, -0.00001439, -0.00001125, -0.00000747, -0.00000366, -0.00000035, 0.00000205, 0.00000331, 0.00000342, 0.00000256, 0.00000107, -0.00000066, -0.00000221, -0.00000326, -0.00000361, -0.00000323, -0.00000225, -0.00000093, 0.00000044, 0.00000153, 0.00000209, 0.00000200, 0.00000126, 0.00000003, -0.00000144, -0.00000283, -0.00000382, -0.00000417, -0.00000375, -0.00000258, -0.00000082, 0.00000125, 0.00000328, 0.00000490, 0.00000583, 0.00000586, 0.00000494, 0.00000316, 0.00000076, -0.00000195, -0.00000461, -0.00000688, -0.00000851, -0.00000933, -0.00000932, -0.00000858, -0.00000729, -0.00000569, -0.00000403, -0.00000252, },
		{ 0.00000068, 0.00000085, 0.00000119, 0.00000162, 0.00000204, 0.00000235, 0.00000249, 0.00000245, 0.00000226, 0.00000203, 0.00000187, 0.00000190, 0.00000218, 0.00000275, 0.00000354, 0.00000441, 0.00000520, 0.00000570, 0.00000570, 0.00000510, 0.00000382, 0.00000193, -0.00000043, -0.00000301, -0.00000553, -0.00000772, -0.00000933, -0.00001019, -0.00001024, -0.00000952, -0.00000817, -0.00000639, -0.00000440, -0.00000241, -0.00000061, 0.00000092, 0.00000216, 0.00000313, 0.00000393, 0.00000464, 0.00000533, 0.00000603, 0.00000668, 0.00000722, 0.00000752, 0.00000747, 0.00000700, 0.00000611, 0.00000488, 0.00000349, 0.00000218, 0.00000124, 0.00000091, 0.00000137, 0.00000267, 0.00000472, 0.00000725, 0.00000988, 0.00001215, 0.00001357, 0.00001376, 0.00001245, 0.00000960, 0.00000539, 0.00000022, -0.00000535, -0.00001066, -0.00001505, -0.00001795, -0.00001902, -0.00001813, -0.00001543, -0.00001132, -0.00000639, -0.00000134, 0.00000316, 0.00000653, 0.00000838, 0.00000859, 0.00000725, 0.00000472, 0.00000150, -0.00000185, -0.00000478, -0.00000685, -0.00000781, -0.00000761, -0.00000642, -0.00000455, -0.00000240, -0.00000040, 0.00000110, 0.00000188, 0.00000188, 0.00000122, 0.00000015, -0.00000099, -0.00000186, -0.00000217, -0.00000175, -0.00000060, 0.00000114, 0.00000319, 0.00000516, 0.00000668, 0.00000741, 0.00000712, 0.00000576, 0.00000343, 0.00000038, -0.00000300, -0.00000631, -0.00000912, -0.00001112, -0.00001210, -0.00001203, -0.00001100, -0.00000924, -0.00000704, -0.00000475, -0.00000264, -0.00000094, 0.00000024, 0.00000090, 0.00000111, 0.00000104, 0.00000084, 0.00000069, },
		{ 0.00000284, 0.00000290, 0.00000273, 0.00000240, 0.00000205, 0.00000179, 0.00000170, 0.00000183, 0.00000214, 0.00000254, 0.00000289, 0.00000305, 0.00000287, 0.00000228, 0.00000126, -0.00000011, -0.00000169, -0.00000327, -0.00000464, -0.00000560, -0.00000600, -0.00000576, -0.00000489, -0.00000350, -0.00000176, 0.00000015, 0.00000201, 0.00000365, 0.00000497, 0.00000590, 0.00000647, 0.00000675, 0.00000681, 0.00000675, 0.00000664, 0.00000649, 0.00000631, 0.00000604, 0.00000563, 0.00000504, 0.00000426, 0.00000336, 0.00000243, 0.00000162, 0.00000110, 0.00000101, 0.00000143, 0.00000235, 0.00000367, 0.00000517, 0.00000655, 0.00000748, 0.00000763, 0.00000678, 0.00000483, 0.00000185, -0.00000191, -0.00000605, -0.00001005, -0.00001336, -0.00001551, -0.00001612, -0.00001506, -0.00001240, -0.00000846, -0.00000376, 0.00000105, 0.00000531, 0.00000840, 0.00000991, 0.00000963, 0.00000762, 0.00000422, -0.00000005, -0.00000453, -0.00000856, -0.00001153, -0.00001304, -0.00001292, -0.00001124, -0.00000831, -0.00000462, -0.00000076, 0.00000272, 0.00000534, 0.00000683, 0.00000711, 0.00000630, 0.00000474, 0.00000283, 0.00000104, -0.00000025, -0.00000079, -0.00000048, 0.00000056, 0.00000206, 0.00000364, 0.00000488, 0.00000541, 0.00000498, 0.00000350, 0.00000106, -0.00000207, -0.00000550, -0.00000876, -0.00001143, -0.00001315, -0.00001370, -0.00001304, -0.00001132, -0.00000880, -0.00000587, -0.00000293, -0.00000033, 0.00000164, 0.00000284, 0.00000326, 0.00000303, 0.00000235, 0.00000148, 0.00000068, 0.00000012, -0.00000008, 0.00000010, 0.00000059, 0.00000126, 0.00000194, 0.00000250, },
		{ 0.00000212, 0.00000193, 0.00000200, 0.00000224, 0.00000254, 0.00000276, 0.00000279, 0.00000253, 0.00000197, 0.00000118, 0.00000026, -0.00000064, -0.00000133, -0.00000169, -0.00000162, -0.00000110, -0.00000016, 0.00000109, 0.00000248, 0.00000385, 0.00000505, 0.00000596, 0.00000652, 0.00000672, 0.00000661, 0.00000627, 0.00000580, 0.00000526, 0.00000473, 0.00000421, 0.00000371, 0.00000318, 0.00000259, 0.00000193, 0.00000122, 0.00000049, -0.00000016, -0.00000064, -0.00000086, -0.00000075, -0.00000032, 0.00000038, 0.00000120, 0.00000194, 0.00000240, 0.00000238, 0.00000172, 0.00000040, -0.00000152, -0.00000383, -0.00000624, -0.00000838, -0.00000990, -0.00001050, -0.00000999, -0.00000836, -0.00000579, -0.00000262, 0.00000070, 0.00000364, 0.00000571, 0.00000653, 0.00000588, 0.00000380, 0.00000053, -0.00000347, -0.00000761, -0.00001124, -0.00001376, -0.00001475, -0.00001398, -0.00001148, -0.00000756, -0.00000272, 0.00000237, 0.00000703, 0.00001066, 0.00001280, 0.00001323, 0.00001200, 0.00000942, 0.00000595, 0.00000221, -0.00000123, -0.00000384, -0.00000530, -0.00000549, -0.00000455, -0.00000279, -0.00000067, 0.00000129, 0.00000262, 0.00000299, 0.00000225, 0.00000044, -0.00000219, -0.00000521, -0.00000816, -0.00001057, -0.00001203, -0.00001232, -0.00001139, -0.00000940, -0.00000665, -0.00000355, -0.00000055, 0.00000194, 0.00000362, 0.00000436, 0.00000416, 0.00000320, 0.00000177, 0.00000020, -0.00000118, -0.00000210, -0.00000239, -0.00000204, -0.00000111, 0.00000019, 0.00000163, 0.00000297, 0.00000400, 0.00000460, 0.00000473, 0.00000444, 0.00000388, 0.00000320, 0.00000257, },
		{ 0.00000233, 0.00000254, 0.00000250, 0.00000223, 0.00000181, 0.00000138, 0.00000106, 0.00000097, 0.00000118, 0.00000170, 0.00000247, 0.00000339, 0.00000434, 0.00000517, 0.00000575, 0.00000603, 0.00000596, 0.00000556, 0.00000490, 0.00000407, 0.00000315, 0.00000221, 0.00000132, 0.00000050, -0.00000027, -0.00000098, -0.00000166, -0.00000230, -0.00000288, -0.00000336, -0.00000366, -0.00000373, -0.00000351, -0.00000301, -0.00000225, -0.00000132, -0.00000037, 0.00000044, 0.00000095, 0.00000104, 0.00000065, -0.00000021, -0.00000143, -0.00000283, -0.00000418, -0.00000523, -0.00000580, -0.00000576, -0.00000510, -0.00000394, -0.00000249, -0.00000106, 0.00000004, 0.00000052, 0.00000019, -0.00000100, -0.00000293, -0.00000533, -0.00000779, -0.00000984, -0.00001105, -0.00001105, -0.00000964, -0.00000685, -0.00000290, 0.00000174, 0.00000651, 0.00001075, 0.00001387, 0.00001540, 0.00001510, 0.00001296, 0.00000928, 0.00000454, -0.00000058, -0.00000539, -0.00000922, -0.00001159, -0.00001224, -0.00001116, -0.00000864, -0.00000517, -0.00000137, 0.00000211, 0.00000468, 0.00000593, 0.00000568, 0.00000400, 0.00000121, -0.00000223, -0.00000574, -0.00000875, -0.00001079, -0.00001157, -0.00001102, -0.00000927, -0.00000665, -0.00000361, -0.00000064, 0.00000182, 0.00000343, 0.00000404, 0.00000368, 0.00000254, 0.00000095, -0.00000068, -0.00000199, -0.00000266, -0.00000253, -0.00000156, 0.00000011, 0.00000222, 0.00000446, 0.00000648, 0.00000799, 0.00000880, 0.00000884, 0.00000815, 0.00000691, 0.00000535, 0.00000375, 0.00000234, 0.00000130, 0.00000073, 0.00000062, 0.00000088, 0.00000136, 0.00000190, },
		{ 0.00000413, 0.00000394, 0.00000390, 0.00000402, 0.00000425, 0.00000452, 0.00000471, 0.00000471, 0.00000445, 0.00000387, 0.00000299, 0.00000186, 0.00000057, -0.00000079, -0.00000210, -0.00000329, -0.00000429, -0.00000508, -0.00000564, -0.00000599, -0.00000615, -0.00000612, -0.00000591, -0.00000549, -0.00000486, -0.00000400, -0.00000292, -0.00000169, -0.00000037, 0.00000091, 0.00000201, 0.00000279, 0.00000313, 0.00000296, 0.00000226, 0.00000110, -0.00000039, -0.00000202, -0.00000360, -0.00000492, -0.00000584, -0.00000627, -0.00000624, -0.00000582, -0.00000518, -0.00000449, -0.00000395, -0.00000368, -0.00000373, -0.00000403, -0.00000446, -0.00000478, -0.00000476, -0.00000417, -0.00000286, -0.00000081, 0.00000185, 0.00000486, 0.00000783, 0.00001031, 0.00001186, 0.00001214, 0.00001094, 0.00000827, 0.00000437, -0.00000034, -0.00000527, -0.00000977, -0.00001321, -0.00001510, -0.00001515, -0.00001332, -0.00000987, -0.00000526, -0.00000018, 0.00000465, 0.00000852, 0.00001087, 0.00001136, 0.00000993, 0.00000682, 0.00000249, -0.00000241, -0.00000717, -0.00001111, -0.00001369, -0.00001462, -0.00001383, -0.00001153, -0.00000812, -0.00000417, -0.00000026, 0.00000307, 0.00000543, 0.00000662, 0.00000665, 0.00000574, 0.00000426, 0.00000265, 0.00000133, 0.00000065, 0.00000082, 0.00000185, 0.00000360, 0.00000577, 0.00000798, 0.00000985, 0.00001102, 0.00001128, 0.00001056, 0.00000892, 0.00000660, 0.00000393, 0.00000126, -0.00000105, -0.00000274, -0.00000363, -0.00000369, -0.00000302, -0.00000180, -0.00000028, 0.00000130, 0.00000269, 0.00000375, 0.00000441, 0.00000467, 0.00000462, 0.00000440, },
		{ 0.00000085, 0.00000003, -0.00000095, -0.00000211, -0.00000339, -0.00000474, -0.00000604, -0.00000719, -0.00000809, -0.00000866, -0.00000886, -0.00000871, -0.00000821, -0.00000744, -0.00000644, -0.00000528, -0.00000400, -0.00000264, -0.00000125, 0.00000015, 0.00000149, 0.00000271, 0.00000372, 0.00000444, 0.00000477, 0.00000462, 0.00000396, 0.00000281, 0.00000122, -0.00000066, -0.00000267, -0.00000460, -0.00000624, -0.00000743, -0.00000805, -0.00000806, -0.00000750, -0.00000648, -0.00000516, -0.00000372, -0.00000234, -0.00000113, -0.00000016, 0.00000056, 0.00000110, 0.00000157, 0.00000206, 0.00000266, 0.00000338, 0.00000417, 0.00000489, 0.00000538, 0.00000540, 0.00000478, 0.00000340, 0.00000124, -0.00000155, -0.00000471, -0.00000787, -0.00001059, -0.00001241, -0.00001299, -0.00001210, -0.00000973, -0.00000607, -0.00000153, 0.00000332, 0.00000783, 0.00001135, 0.00001335, 0.00001348, 0.00001165, 0.00000805, 0.00000311, -0.00000250, -0.00000804, -0.00001276, -0.00001599, -0.00001732, -0.00001655, -0.00001382, -0.00000950, -0.00000420, 0.00000139, 0.00000655, 0.00001068, 0.00001335, 0.00001438, 0.00001385, 0.00001206, 0.00000949, 0.00000669, 0.00000418, 0.00000239, 0.00000159, 0.00000181, 0.00000290, 0.00000455, 0.00000634, 0.00000783, 0.00000863, 0.00000850, 0.00000735, 0.00000528, 0.00000254, -0.00000048, -0.00000336, -0.00000569, -0.00000714, -0.00000752, -0.00000681, -0.00000513, -0.00000272, 0.00000008, 0.00000290, 0.00000542, 0.00000738, 0.00000863, 0.00000913, 0.00000896, 0.00000827, 0.00000724, 0.00000607, 0.00000492, 0.00000388, 0.00000300, 0.00000225, 0.00000156, },
		{ -0.00000949, -0.00000964, -0.00000945, -0.00000888, -0.00000795, -0.00000671, -0.00000522, -0.00000360, -0.00000194, -0.00000033, 0.00000116, 0.00000247, 0.00000357, 0.00000443, 0.00000502, 0.00000532, 0.00000531, 0.00000494, 0.00000422, 0.00000313, 0.00000172, 0.00000005, -0.00000175, -0.00000355, -0.00000517, -0.00000644, -0.00000723, -0.00000744, -0.00000704, -0.00000607, -0.00000466, -0.00000297, -0.00000119, 0.00000047, 0.00000184, 0.00000279, 0.00000327, 0.00000330, 0.00000294, 0.00000231, 0.00000152, 0.00000070, -0.00000009, -0.00000081, -0.00000148, -0.00000215, -0.00000287, -0.00000369, -0.00000459, -0.00000552, -0.00000635, -0.00000692, -0.00000702, -0.00000649, -0.00000522, -0.00000319, -0.00000052, 0.00000256, 0.00000569, 0.00000846, 0.00001044, 0.00001125, 0.00001066, 0.00000862, 0.00000527, 0.00000096, -0.00000377, -0.00000830, -0.00001199, -0.00001425, -0.00001470, -0.00001316, -0.00000973, -0.00000476, 0.00000115, 0.00000730, 0.00001292, 0.00001734, 0.00002004, 0.00002074, 0.00001945, 0.00001644, 0.00001220, 0.00000738, 0.00000267, -0.00000131, -0.00000409, -0.00000541, -0.00000527, -0.00000389, -0.00000168, 0.00000084, 0.00000313, 0.00000474, 0.00000534, 0.00000482, 0.00000327, 0.00000096, -0.00000170, -0.00000424, -0.00000621, -0.00000725, -0.00000715, -0.00000588, -0.00000359, -0.00000058, 0.00000275, 0.00000599, 0.00000873, 0.00001067, 0.00001163, 0.00001156, 0.00001057, 0.00000885, 0.00000666, 0.00000429, 0.00000197, -0.00000011, -0.00000184, -0.00000320, -0.00000424, -0.00000505, -0.00000574, -0.00000639, -0.00000707, -0.00000777, -0.00000846, -0.00000907, },
		{ -0.00000044, 0.00000114, 0.00000256, 0.00000377, 0.00000469, 0.00000531, 0.00000563, 0.00000566, 0.00000543, 0.00000496, 0.00000427, 0.00000337, 0.00000229, 0.00000104, -0.00000031, -0.00000169, -0.00000303, -0.00000420, -0.00000509, -0.00000562, -0.00000571, -0.00000534, -0.00000455, -0.00000344, -0.00000215, -0.00000085, 0.00000029, 0.00000112, 0.00000151, 0.00000143, 0.00000089, -0.00000001, -0.00000115, -0.00000236, -0.00000348, -0.00000435, -0.00000490, -0.00000508, -0.00000491, -0.00000444, -0.00000376, -0.00000295, -0.00000210, -0.00000124, -0.00000039, 0.00000047, 0.00000138, 0.00000235, 0.00000338, 0.00000444, 0.00000541, 0.00000617, 0.00000656, 0.00000643, 0.00000568, 0.00000427, 0.00000230, -0.00000006, -0.00000250, -0.00000468, -0.00000621, -0.00000674, -0.00000605, -0.00000404, -0.00000080, 0.00000336, 0.00000801, 0.00001258, 0.00001646, 0.00001910, 0.00002008, 0.00001918, 0.00001643, 0.00001208, 0.00000665, 0.00000076, -0.00000486, -0.00000954, -0.00001273, -0.00001410, -0.00001355, -0.00001128, -0.00000770, -0.00000338, 0.00000101, 0.00000484, 0.00000758, 0.00000891, 0.00000873, 0.00000717, 0.00000458, 0.00000146, -0.00000166, -0.00000423, -0.00000584, -0.00000626, -0.00000542, -0.00000349, -0.00000078, 0.00000229, 0.00000526, 0.00000768, 0.00000924, 0.00000973, 0.00000914, 0.00000757, 0.00000526, 0.00000253, -0.00000030, -0.00000292, -0.00000511, -0.00000675, -0.00000780, -0.00000833, -0.00000846, -0.00000835, -0.00000815, -0.00000796, -0.00000784, -0.00000778, -0.00000775, -0.00000765, -0.00000738, -0.00000686, -0.00000606, -0.00000495, -0.00000359, -0.00000205, },
		{ 0.00000588, 0.00000548, 0.00000487, 0.00000410, 0.00000320, 0.00000219, 0.00000109, -0.00000008, -0.00000129, -0.00000249, -0.00000360, -0.00000456, -0.00000528, -0.00000569, -0.00000577, -0.00000550, -0.00000493, -0.00000415, -0.00000328, -0.00000246, -0.00000179, -0.00000140, -0.00000131, -0.00000153, -0.00000200, -0.00000260, -0.00000320, -0.00000364, -0.00000381, -0.00000362, -0.00000303, -0.00000208, -0.00000085, 0.00000053, 0.00000190, 0.00000313, 0.00000411, 0.00000477, 0.00000508, 0.00000507, 0.00000482, 0.00000441, 0.00000393, 0.00000346, 0.00000305, 0.00000271, 0.00000244, 0.00000223, 0.00000206, 0.00000194, 0.00000189, 0.00000198, 0.00000228, 0.00000288, 0.00000383, 0.00000514, 0.00000675, 0.00000851, 0.00001020, 0.00001156, 0.00001229, 0.00001213, 0.00001089, 0.00000849, 0.00000503, 0.00000075, -0.00000396, -0.00000860, -0.00001264, -0.00001555, -0.00001693, -0.00001653, -0.00001432, -0.00001051, -0.00000552, 0.00000009, 0.00000563, 0.00001044, 0.00001394, 0.00001573, 0.00001562, 0.00001369, 0.00001024, 0.00000577, 0.00000090, -0.00000373, -0.00000753, -0.00001008, -0.00001111, -0.00001062, -0.00000880, -0.00000601, -0.00000275, 0.00000049, 0.00000323, 0.00000511, 0.00000592, 0.00000564, 0.00000438, 0.00000239, 0.00000002, -0.00000240, -0.00000455, -0.00000620, -0.00000723, -0.00000763, -0.00000748, -0.00000696, -0.00000624, -0.00000552, -0.00000493, -0.00000455, -0.00000439, -0.00000439, -0.00000443, -0.00000441, -0.00000419, -0.00000370, -0.00000290, -0.00000180, -0.00000049, 0.00000094, 0.00000236, 0.00000365, 0.00000471, 0.00000548, 0.00000592, 0.00000605, },
		{ -0.00000095, -0.00000226, -0.00000356, -0.00000479, -0.00000588, -0.00000676, -0.00000734, -0.00000758, -0.00000744, -0.00000693, -0.00000611, -0.00000505, -0.00000388, -0.00000273, -0.00000171, -0.00000091, -0.00000037, -0.00000008, 0.00000001, -0.00000000, -0.00000001, 0.00000010, 0.00000040, 0.00000095, 0.00000172, 0.00000266, 0.00000366, 0.00000460, 0.00000533, 0.00000577, 0.00000584, 0.00000555, 0.00000496, 0.00000417, 0.00000333, 0.00000258, 0.00000207, 0.00000191, 0.00000214, 0.00000275, 0.00000369, 0.00000482, 0.00000603, 0.00000715, 0.00000807, 0.00000866, 0.00000888, 0.00000867, 0.00000806, 0.00000705, 0.00000569, 0.00000402, 0.00000207, -0.00000012, -0.00000247, -0.00000493, -0.00000737, -0.00000965, -0.00001158, -0.00001293, -0.00001351, -0.00001312, -0.00001163, -0.00000903, -0.00000541, -0.00000099, 0.00000385, 0.00000868, 0.00001300, 0.00001631, 0.00001820, 0.00001839, 0.00001677, 0.00001344, 0.00000871, 0.00000304, -0.00000297, -0.00000870, -0.00001354, -0.00001702, -0.00001881, -0.00001881, -0.00001713, -0.00001409, -0.00001014, -0.00000581, -0.00000164, 0.00000190, 0.00000445, 0.00000585, 0.00000608, 0.00000529, 0.00000377, 0.00000184, -0.00000014, -0.00000187, -0.00000312, -0.00000380, -0.00000391, -0.00000357, -0.00000295, -0.00000226, -0.00000169, -0.00000137, -0.00000137, -0.00000166, -0.00000217, -0.00000274, -0.00000322, -0.00000344, -0.00000331, -0.00000276, -0.00000181, -0.00000053, 0.00000095, 0.00000249, 0.00000392, 0.00000511, 0.00000597, 0.00000645, 0.00000653, 0.00000626, 0.00000569, 0.00000487, 0.00000389, 0.00000279, 0.00000160, 0.00000034, },
		{ -0.00000775, -0.00000727, -0.00000641, -0.00000519, -0.00000372, -0.00000211, -0.00000050, 0.00000097, 0.00000217, 0.00000304, 0.00000353, 0.00000368, 0.00000355, 0.00000326, 0.00000293, 0.00000268, 0.00000259, 0.00000271, 0.00000303, 0.00000350, 0.00000401, 0.00000448, 0.00000479, 0.00000489, 0.00000474, 0.00000438, 0.00000388, 0.00000337, 0.00000298, 0.00000282, 0.00000300, 0.00000352, 0.00000436, 0.00000541, 0.00000650, 0.00000744, 0.00000802, 0.00000808, 0.00000748, 0.00000618, 0.00000421, 0.00000169, -0.00000118, -0.00000419, -0.00000706, -0.00000957, -0.00001150, -0.00001270, -0.00001306, -0.00001257, -0.00001125, -0.00000919, -0.00000653, -0.00000340, 0.00000002, 0.00000353, 0.00000695, 0.00001007, 0.00001268, 0.00001458, 0.00001559, 0.00001555, 0.00001438, 0.00001205, 0.00000863, 0.00000430, -0.00000067, -0.00000592, -0.00001104, -0.00001559, -0.00001918, -0.00002148, -0.00002229, -0.00002154, -0.00001934, -0.00001591, -0.00001164, -0.00000695, -0.00000233, 0.00000181, 0.00000513, 0.00000739, 0.00000854, 0.00000863, 0.00000786, 0.00000650, 0.00000487, 0.00000327, 0.00000193, 0.00000101, 0.00000055, 0.00000050, 0.00000073, 0.00000106, 0.00000129, 0.00000129, 0.00000094, 0.00000023, -0.00000077, -0.00000193, -0.00000306, -0.00000397, -0.00000450, -0.00000452, -0.00000399, -0.00000293, -0.00000145, 0.00000029, 0.00000210, 0.00000378, 0.00000516, 0.00000612, 0.00000659, 0.00000657, 0.00000609, 0.00000524, 0.00000412, 0.00000282, 0.00000143, 0.00000002, -0.00000137, -0.00000270, -0.00000395, -0.00000511, -0.00000613, -0.00000696, -0.00000755, -0.00000783, },
		{ 0.00000318, 0.00000414, 0.00000471, 0.00000485, 0.00000460, 0.00000402, 0.00000324, 0.00000242, 0.00000169, 0.00000119, 0.00000100, 0.00000113, 0.00000157, 0.00000221, 0.00000295, 0.00000365, 0.00000420, 0.00000452, 0.00000458, 0.00000440, 0.00000407, 0.00000367, 0.00000333, 0.00000314, 0.00000314, 0.00000334, 0.00000365, 0.00000395, 0.00000410, 0.00000391, 0.00000325, 0.00000204, 0.00000027, -0.00000197, -0.00000449, -0.00000705, -0.00000936, -0.00001111, -0.00001205, -0.00001200, -0.00001088, -0.00000872, -0.00000569, -0.00000203, 0.00000192, 0.00000581, 0.00000928, 0.00001204, 0.00001385, 0.00001457, 0.00001417, 0.00001271, 0.00001032, 0.00000721, 0.00000359, -0.00000029, -0.00000420, -0.00000791, -0.00001126, -0.00001408, -0.00001625, -0.00001769, -0.00001835, -0.00001820, -0.00001725, -0.00001557, -0.00001325, -0.00001044, -0.00000732, -0.00000407, -0.00000091, 0.00000197, 0.00000441, 0.00000630, 0.00000758, 0.00000828, 0.00000848, 0.00000829, 0.00000787, 0.00000735, 0.00000687, 0.00000649, 0.00000623, 0.00000606, 0.00000588, 0.00000559, 0.00000507, 0.00000421, 0.00000298, 0.00000139, -0.00000046, -0.00000244, -0.00000435, -0.00000598, -0.00000714, -0.00000766, -0.00000748, -0.00000659, -0.00000507, -0.00000309, -0.00000086, 0.00000139, 0.00000342, 0.00000505, 0.00000615, 0.00000664, 0.00000653, 0.00000589, 0.00000484, 0.00000352, 0.00000205, 0.00000057, -0.00000084, -0.00000211, -0.00000321, -0.00000414, -0.00000488, -0.00000545, -0.00000582, -0.00000597, -0.00000586, -0.00000547, -0.00000477, -0.00000377, -0.00000251, -0.00000107, 0.00000045, 0.00000191, },
		{ 0.00000003, -0.00000056, -0.00000086, -0.00000080, -0.00000035, 0.00000044, 0.00000148, 0.00000262, 0.00000370, 0.00000457, 0.00000509, 0.00000522, 0.00000495, 0.00000433, 0.00000350, 0.00000259, 0.00000174, 0.00000106, 0.00000062, 0.00000042, 0.00000038, 0.00000040, 0.00000033, 0.00000005, -0.00000055, -0.00000149, -0.00000273, -0.00000412, -0.00000548, -0.00000656, -0.00000712, -0.00000695, -0.00000596, -0.00000412, -0.00000156, 0.00000149, 0.00000471, 0.00000771, 0.00001011, 0.00001159, 0.00001191, 0.00001097, 0.00000882, 0.00000566, 0.00000178, -0.00000241, -0.00000651, -0.00001012, -0.00001295, -0.00001477, -0.00001552, -0.00001525, -0.00001412, -0.00001238, -0.00001031, -0.00000819, -0.00000625, -0.00000464, -0.00000345, -0.00000267, -0.00000221, -0.00000193, -0.00000167, -0.00000127, -0.00000061, 0.00000041, 0.00000181, 0.00000354, 0.00000551, 0.00000761, 0.00000967, 0.00001154, 0.00001305, 0.00001409, 0.00001453, 0.00001431, 0.00001340, 0.00001182, 0.00000961, 0.00000688, 0.00000376, 0.00000042, -0.00000293, -0.00000608, -0.00000881, -0.00001093, -0.00001228, -0.00001274, -0.00001229, -0.00001097, -0.00000890, -0.00000626, -0.00000329, -0.00000026, 0.00000257, 0.00000498, 0.00000679, 0.00000790, 0.00000828, 0.00000799, 0.00000713, 0.00000586, 0.00000435, 0.00000277, 0.00000125, -0.00000011, -0.00000124, -0.00000213, -0.00000280, -0.00000328, -0.00000359, -0.00000378, -0.00000383, -0.00000375, -0.00000351, -0.00000309, -0.00000249, -0.00000171, -0.00000081, 0.00000016, 0.00000109, 0.00000188, 0.00000243, 0.00000268, 0.00000259, 0.00000219, 0.00000155, 0.00000078, },
		{ 0.00000428, 0.00000529, 0.00000598, 0.00000626, 0.00000605, 0.00000537, 0.00000430, 0.00000298, 0.00000161, 0.00000035, -0.00000061, -0.00000119, -0.00000134, -0.00000111, -0.00000061, -0.00000001, 0.00000052, 0.00000081, 0.00000075, 0.00000032, -0.00000043, -0.00000136, -0.00000230, -0.00000301, -0.00000332, -0.00000308, -0.00000227, -0.00000094, 0.00000072, 0.00000246, 0.00000395, 0.00000491, 0.00000507, 0.00000430, 0.00000259, 0.00000007, -0.00000298, -0.00000619, -0.00000915, -0.00001145, -0.00001277, -0.00001292, -0.00001186, -0.00000975, -0.00000688, -0.00000366, -0.00000055, 0.00000201, 0.00000366, 0.00000419, 0.00000356, 0.00000191, -0.00000045, -0.00000310, -0.00000554, -0.00000731, -0.00000800, -0.00000737, -0.00000532, -0.00000197, 0.00000240, 0.00000739, 0.00001248, 0.00001715, 0.00002090, 0.00002334, 0.00002421, 0.00002341, 0.00002100, 0.00001719, 0.00001230, 0.00000675, 0.00000097, -0.00000461, -0.00000959, -0.00001366, -0.00001660, -0.00001827, -0.00001865, -0.00001780, -0.00001587, -0.00001305, -0.00000961, -0.00000581, -0.00000193, 0.00000176, 0.00000503, 0.00000770, 0.00000963, 0.00001077, 0.00001111, 0.00001073, 0.00000975, 0.00000831, 0.00000661, 0.00000481, 0.00000306, 0.00000146, 0.00000011, -0.00000099, -0.00000185, -0.00000252, -0.00000304, -0.00000348, -0.00000387, -0.00000423, -0.00000455, -0.00000479, -0.00000490, -0.00000486, -0.00000462, -0.00000419, -0.00000360, -0.00000290, -0.00000217, -0.00000150, -0.00000096, -0.00000060, -0.00000044, -0.00000044, -0.00000054, -0.00000062, -0.00000060, -0.00000037, 0.00000013, 0.00000092, 0.00000195, 0.00000311, },
		{ 0.00000174, 0.00000025, -0.00000100, -0.00000184, -0.00000217, -0.00000198, -0.00000131, -0.00000031, 0.00000082, 0.00000183, 0.00000252, 0.00000271, 0.00000233, 0.00000139, -0.00000000, -0.00000164, -0.00000328, -0.00000466, -0.00000557, -0.00000589, -0.00000557, -0.00000471, -0.00000349, -0.00000216, -0.00000100, -0.00000024, -0.00000005, -0.00000049, -0.00000146, -0.00000278, -0.00000416, -0.00000528, -0.00000583, -0.00000559, -0.00000447, -0.00000255, -0.00000005, 0.00000267, 0.00000518, 0.00000702, 0.00000781, 0.00000732, 0.00000550, 0.00000251, -0.00000127, -0.00000531, -0.00000899, -0.00001168, -0.00001283, -0.00001209, -0.00000933, -0.00000468, 0.00000146, 0.00000847, 0.00001559, 0.00002204, 0.00002705, 0.00003003, 0.00003061, 0.00002868, 0.00002441, 0.00001821, 0.00001069, 0.00000258, -0.00000537, -0.00001246, -0.00001812, -0.00002197, -0.00002381, -0.00002368, -0.00002178, -0.00001844, -0.00001410, -0.00000923, -0.00000425, 0.00000043, 0.00000453, 0.00000783, 0.00001024, 0.00001173, 0.00001235, 0.00001223, 0.00001148, 0.00001026, 0.00000873, 0.00000702, 0.00000526, 0.00000354, 0.00000194, 0.00000051, -0.00000076, -0.00000185, -0.00000281, -0.00000367, -0.00000448, -0.00000528, -0.00000608, -0.00000690, -0.00000769, -0.00000841, -0.00000899, -0.00000934, -0.00000940, -0.00000912, -0.00000847, -0.00000748, -0.00000619, -0.00000472, -0.00000316, -0.00000163, -0.00000024, 0.00000094, 0.00000189, 0.00000260, 0.00000313, 0.00000356, 0.00000395, 0.00000438, 0.00000488, 0.00000545, 0.00000603, 0.00000653, 0.00000685, 0.00000689, 0.00000655, 0.00000580, 0.00000468, 0.00000328, },
		{ 0.00000027, 0.00000155, 0.00000260, 0.00000319, 0.00000315, 0.00000240, 0.00000094, -0.00000107, -0.00000341, -0.00000575, -0.00000780, -0.00000924, -0.00000988, -0.00000961, -0.00000847, -0.00000662, -0.00000433, -0.00000190, 0.00000034, 0.00000214, 0.00000333, 0.00000387, 0.00000385, 0.00000345, 0.00000293, 0.00000255, 0.00000251, 0.00000293, 0.00000378, 0.00000492, 0.00000607, 0.00000690, 0.00000709, 0.00000637, 0.00000461, 0.00000188, -0.00000159, -0.00000535, -0.00000886, -0.00001153, -0.00001278, -0.00001222, -0.00000965, -0.00000514, 0.00000095, 0.00000801, 0.00001527, 0.00002183, 0.00002689, 0.00002974, 0.00002994, 0.00002738, 0.00002225, 0.00001507, 0.00000660, -0.00000227, -0.00001059, -0.00001751, -0.00002238, -0.00002481, -0.00002469, -0.00002223, -0.00001787, -0.00001223, -0.00000604, 0.00000002, 0.00000533, 0.00000944, 0.00001211, 0.00001327, 0.00001307, 0.00001178, 0.00000973, 0.00000732, 0.00000488, 0.00000268, 0.00000089, -0.00000040, -0.00000124, -0.00000170, -0.00000194, -0.00000211, -0.00000234, -0.00000275, -0.00000341, -0.00000433, -0.00000550, -0.00000685, -0.00000829, -0.00000971, -0.00001101, -0.00001207, -0.00001278, -0.00001305, -0.00001282, -0.00001204, -0.00001073, -0.00000892, -0.00000669, -0.00000415, -0.00000147, 0.00000119, 0.00000367, 0.00000581, 0.00000751, 0.00000870, 0.00000936, 0.00000955, 0.00000934, 0.00000886, 0.00000822, 0.00000752, 0.00000683, 0.00000620, 0.00000561, 0.00000502, 0.00000436, 0.00000359, 0.00000266, 0.00000158, 0.00000041, -0.00000074, -0.00000175, -0.00000248, -0.00000280, -0.00000264, -0.00000201, -0.00000098, },
		{ -0.00000301, -0.00000563, -0.00000815, -0.00001020, -0.00001148, -0.00001173, -0.00001086, -0.00000889, -0.00000600, -0.00000250, 0.00000122, 0.00000476, 0.00000775, 0.00000993, 0.00001115, 0.00001143, 0.00001091, 0.00000984, 0.00000851, 0.00000720, 0.00000614, 0.00000541, 0.00000501, 0.00000478, 0.00000452, 0.00000397, 0.00000292, 0.00000125, -0.00000100, -0.00000366, -0.00000638, -0.00000873, -0.00001024, -0.00001048, -0.00000917, -0.00000622, -0.00000180, 0.00000370, 0.00000968, 0.00001539, 0.00002007, 0.00002304, 0.00002382, 0.00002220, 0.00001826, 0.00001242, 0.00000537, -0.00000205, -0.00000893, -0.00001443, -0.00001790, -0.00001898, -0.00001761, -0.00001408, -0.00000895, -0.00000298, 0.00000298, 0.00000813, 0.00001183, 0.00001364, 0.00001345, 0.00001139, 0.00000786, 0.00000344, -0.00000122, -0.00000549, -0.00000885, -0.00001093, -0.00001161, -0.00001097, -0.00000928, -0.00000694, -0.00000440, -0.00000213, -0.00000050, 0.00000024, -0.00000001, -0.00000123, -0.00000322, -0.00000571, -0.00000839, -0.00001093, -0.00001303, -0.00001445, -0.00001506, -0.00001477, -0.00001360, -0.00001164, -0.00000902, -0.00000590, -0.00000248, 0.00000105, 0.00000447, 0.00000760, 0.00001028, 0.00001236, 0.00001374, 0.00001436, 0.00001424, 0.00001342, 0.00001201, 0.00001018, 0.00000811, 0.00000599, 0.00000399, 0.00000226, 0.00000087, -0.00000014, -0.00000080, -0.00000117, -0.00000136, -0.00000144, -0.00000150, -0.00000158, -0.00000165, -0.00000169, -0.00000161, -0.00000134, -0.00000085, -0.00000011, 0.00000079, 0.00000176, 0.00000261, 0.00000315, 0.00000320, 0.00000260, 0.00000132, -0.00000060, },
		{ -0.00000608, -0.00000225, 0.00000215, 0.00000668, 0.00001089, 0.00001435, 0.00001672, 0.00001781, 0.00001762, 0.00001627, 0.00001406, 0.00001134, 0.00000851, 0.00000589, 0.00000371, 0.00000209, 0.00000098, 0.00000023, -0.00000039, -0.00000110, -0.00000209, -0.00000342, -0.00000504, -0.00000676, -0.00000825, -0.00000917, -0.00000917, -0.00000801, -0.00000560, -0.00000204, 0.00000236, 0.00000712, 0.00001164, 0.00001532, 0.00001764, 0.00001822, 0.00001696, 0.00001399, 0.00000972, 0.00000475, -0.00000021, -0.00000445, -0.00000737, -0.00000859, -0.00000800, -0.00000580, -0.00000245, 0.00000138, 0.00000495, 0.00000753, 0.00000854, 0.00000766, 0.00000486, 0.00000042, -0.00000511, -0.00001100, -0.00001641, -0.00002062, -0.00002300, -0.00002325, -0.00002132, -0.00001749, -0.00001230, -0.00000648, -0.00000079, 0.00000400, 0.00000731, 0.00000876, 0.00000826, 0.00000598, 0.00000232, -0.00000215, -0.00000679, -0.00001097, -0.00001415, -0.00001596, -0.00001621, -0.00001490, -0.00001222, -0.00000848, -0.00000408, 0.00000057, 0.00000507, 0.00000909, 0.00001237, 0.00001477, 0.00001621, 0.00001669, 0.00001628, 0.00001508, 0.00001321, 0.00001085, 0.00000815, 0.00000531, 0.00000250, -0.00000008, -0.00000228, -0.00000396, -0.00000502, -0.00000544, -0.00000523, -0.00000450, -0.00000339, -0.00000206, -0.00000069, 0.00000054, 0.00000153, 0.00000222, 0.00000259, 0.00000270, 0.00000263, 0.00000248, 0.00000231, 0.00000218, 0.00000205, 0.00000185, 0.00000149, 0.00000084, -0.00000020, -0.00000167, -0.00000353, -0.00000564, -0.00000779, -0.00000967, -0.00001099, -0.00001143, -0.00001081, -0.00000900, },
		{ 0.00002121, 0.00001997, 0.00001741, 0.00001386, 0.00000975, 0.00000554, 0.00000167, -0.00000152, -0.00000384, -0.00000525, -0.00000587, -0.00000594, -0.00000575, -0.00000556, -0.00000558, -0.00000588, -0.00000641, -0.00000698, -0.00000731, -0.00000710, -0.00000609, -0.00000413, -0.00000123, 0.00000243, 0.00000650, 0.00001053, 0.00001401, 0.00001648, 0.00001760, 0.00001723, 0.00001541, 0.00001243, 0.00000873, 0.00000483, 0.00000126, -0.00000154, -0.00000332, -0.00000402, -0.00000385, -0.00000318, -0.00000253, -0.00000242, -0.00000329, -0.00000541, -0.00000879, -0.00001318, -0.00001808, -0.00002280, -0.00002659, -0.00002876, -0.00002878, -0.00002639, -0.00002167, -0.00001502, -0.00000716, 0.00000102, 0.00000854, 0.00001450, 0.00001820, 0.00001924, 0.00001756, 0.00001349, 0.00000766, 0.00000092, -0.00000579, -0.00001156, -0.00001564, -0.00001753, -0.00001707, -0.00001438, -0.00000988, -0.00000418, 0.00000198, 0.00000790, 0.00001292, 0.00001659, 0.00001863, 0.00001901, 0.00001785, 0.00001546, 0.00001220, 0.00000848, 0.00000468, 0.00000112, -0.00000196, -0.00000442, -0.00000618, -0.00000723, -0.00000760, -0.00000734, -0.00000651, -0.00000521, -0.00000351, -0.00000156, 0.00000052, 0.00000256, 0.00000439, 0.00000585, 0.00000681, 0.00000718, 0.00000693, 0.00000611, 0.00000485, 0.00000330, 0.00000165, 0.00000007, -0.00000128, -0.00000234, -0.00000310, -0.00000361, -0.00000398, -0.00000434, -0.00000479, -0.00000538, -0.00000609, -0.00000680, -0.00000732, -0.00000743, -0.00000689, -0.00000550, -0.00000318, 0.00000003, 0.00000395, 0.00000827, 0.00001257, 0.00001640, 0.00001932, 0.00002099, },
		{ -0.00000749, -0.00000954, -0.00001036, -0.00001008, -0.00000893, -0.00000729, -0.00000552, -0.00000395, -0.00000280, -0.00000215, -0.00000191, -0.00000188, -0.00000177, -0.00000130, -0.00000022, 0.00000156, 0.00000401, 0.00000692, 0.00000994, 0.00001267, 0.00001469, 0.00001564, 0.00001531, 0.00001366, 0.00001085, 0.00000718, 0.00000307, -0.00000104, -0.00000477, -0.00000783, -0.00001012, -0.00001170, -0.00001278, -0.00001367, -0.00001468, -0.00001607, -0.00001792, -0.00002016, -0.00002251, -0.00002453, -0.00002568, -0.00002544, -0.00002342, -0.00001942, -0.00001353, -0.00000615, 0.00000206, 0.00001023, 0.00001742, 0.00002275, 0.00002553, 0.00002540, 0.00002234, 0.00001677, 0.00000941, 0.00000127, -0.00000655, -0.00001302, -0.00001728, -0.00001880, -0.00001743, -0.00001345, -0.00000746, -0.00000035, 0.00000688, 0.00001325, 0.00001795, 0.00002042, 0.00002044, 0.00001813, 0.00001391, 0.00000842, 0.00000243, -0.00000332, -0.00000816, -0.00001162, -0.00001344, -0.00001360, -0.00001229, -0.00000982, -0.00000660, -0.00000306, 0.00000043, 0.00000359, 0.00000620, 0.00000817, 0.00000948, 0.00001015, 0.00001022, 0.00000975, 0.00000879, 0.00000739, 0.00000559, 0.00000348, 0.00000116, -0.00000120, -0.00000345, -0.00000538, -0.00000682, -0.00000764, -0.00000779, -0.00000728, -0.00000624, -0.00000484, -0.00000332, -0.00000188, -0.00000072, 0.00000007, 0.00000050, 0.00000067, 0.00000078, 0.00000106, 0.00000176, 0.00000305, 0.00000498, 0.00000751, 0.00001042, 0.00001338, 0.00001600, 0.00001787, 0.00001864, 0.00001810, 0.00001618, 0.00001303, 0.00000894, 0.00000436, -0.00000021, -0.00000430, },
		{ 0.00000064, 0.00000238, 0.00000331, 0.00000343, 0.00000289, 0.00000198, 0.00000105, 0.00000043, 0.00000037, 0.00000099, 0.00000223, 0.00000388, 0.00000563, 0.00000708, 0.00000789, 0.00000777, 0.00000659, 0.00000437, 0.00000132, -0.00000227, -0.00000601, -0.00000952, -0.00001251, -0.00001479, -0.00001631, -0.00001717, -0.00001754, -0.00001764, -0.00001766, -0.00001768, -0.00001769, -0.00001752, -0.00001690, -0.00001552, -0.00001309, -0.00000943, -0.00000455, 0.00000132, 0.00000772, 0.00001400, 0.00001944, 0.00002329, 0.00002497, 0.00002412, 0.00002073, 0.00001511, 0.00000793, 0.00000010, -0.00000733, -0.00001336, -0.00001715, -0.00001818, -0.00001632, -0.00001189, -0.00000557, 0.00000168, 0.00000875, 0.00001457, 0.00001823, 0.00001919, 0.00001729, 0.00001280, 0.00000638, -0.00000105, -0.00000844, -0.00001474, -0.00001911, -0.00002099, -0.00002018, -0.00001686, -0.00001154, -0.00000496, 0.00000203, 0.00000856, 0.00001393, 0.00001762, 0.00001939, 0.00001925, 0.00001745, 0.00001438, 0.00001052, 0.00000636, 0.00000230, -0.00000135, -0.00000439, -0.00000674, -0.00000840, -0.00000942, -0.00000985, -0.00000974, -0.00000910, -0.00000795, -0.00000630, -0.00000421, -0.00000176, 0.00000088, 0.00000348, 0.00000582, 0.00000762, 0.00000871, 0.00000896, 0.00000840, 0.00000715, 0.00000549, 0.00000374, 0.00000227, 0.00000140, 0.00000135, 0.00000219, 0.00000383, 0.00000600, 0.00000832, 0.00001036, 0.00001168, 0.00001197, 0.00001103, 0.00000888, 0.00000575, 0.00000200, -0.00000188, -0.00000539, -0.00000806, -0.00000959, -0.00000983, -0.00000885, -0.00000691, -0.00000439, -0.00000173, },
		{ -0.00000385, -0.00000517, -0.00000568, -0.00000530, -0.00000411, -0.00000236, -0.00000040, 0.00000136, 0.00000256, 0.00000293, 0.00000237, 0.00000091, -0.00000126, -0.00000383, -0.00000647, -0.00000886, -0.00001076, -0.00001203, -0.00001266, -0.00001273, -0.00001241, -0.00001184, -0.00001115, -0.00001037, -0.00000946, -0.00000826, -0.00000662, -0.00000435, -0.00000138, 0.00000225, 0.00000633, 0.00001051, 0.00001430, 0.00001716, 0.00001859, 0.00001823, 0.00001596, 0.00001191, 0.00000649, 0.00000037, -0.00000563, -0.00001067, -0.00001400, -0.00001510, -0.00001380, -0.00001029, -0.00000513, 0.00000086, 0.00000665, 0.00001126, 0.00001384, 0.00001385, 0.00001118, 0.00000611, -0.00000062, -0.00000802, -0.00001494, -0.00002025, -0.00002304, -0.00002277, -0.00001932, -0.00001305, -0.00000473, 0.00000458, 0.00001366, 0.00002136, 0.00002670, 0.00002907, 0.00002826, 0.00002448, 0.00001832, 0.00001065, 0.00000246, -0.00000529, -0.00001176, -0.00001636, -0.00001882, -0.00001914, -0.00001759, -0.00001461, -0.00001073, -0.00000646, -0.00000224, 0.00000162, 0.00000492, 0.00000760, 0.00000968, 0.00001120, 0.00001219, 0.00001265, 0.00001252, 0.00001174, 0.00001025, 0.00000804, 0.00000519, 0.00000191, -0.00000152, -0.00000470, -0.00000721, -0.00000869, -0.00000887, -0.00000763, -0.00000509, -0.00000153, 0.00000257, 0.00000664, 0.00001010, 0.00001244, 0.00001331, 0.00001259, 0.00001040, 0.00000706, 0.00000310, -0.00000090, -0.00000436, -0.00000682, -0.00000798, -0.00000780, -0.00000645, -0.00000428, -0.00000178, 0.00000055, 0.00000228, 0.00000310, 0.00000290, 0.00000179, 0.00000003, -0.00000199, },
		{ 0.00000225, 0.00000463, 0.00000628, 0.00000694, 0.00000650, 0.00000505, 0.00000281, 0.00000011, -0.00000271, -0.00000530, -0.00000743, -0.00000895, -0.00000983, -0.00001016, -0.00001003, -0.00000958, -0.00000889, -0.00000799, -0.00000684, -0.00000537, -0.00000350, -0.00000118, 0.00000156, 0.00000455, 0.00000754, 0.00001016, 0.00001204, 0.00001281, 0.00001224, 0.00001025, 0.00000699, 0.00000283, -0.00000171, -0.00000597, -0.00000932, -0.00001124, -0.00001142, -0.00000983, -0.00000676, -0.00000277, 0.00000141, 0.00000494, 0.00000711, 0.00000737, 0.00000553, 0.00000177, -0.00000338, -0.00000905, -0.00001423, -0.00001790, -0.00001922, -0.00001767, -0.00001316, -0.00000607, 0.00000276, 0.00001220, 0.00002093, 0.00002770, 0.00003148, 0.00003163, 0.00002803, 0.00002106, 0.00001160, 0.00000085, -0.00000981, -0.00001905, -0.00002574, -0.00002915, -0.00002899, -0.00002548, -0.00001922, -0.00001116, -0.00000238, 0.00000603, 0.00001318, 0.00001841, 0.00002140, 0.00002216, 0.00002097, 0.00001829, 0.00001467, 0.00001061, 0.00000654, 0.00000272, -0.00000073, -0.00000380, -0.00000655, -0.00000903, -0.00001122, -0.00001301, -0.00001422, -0.00001458, -0.00001383, -0.00001180, -0.00000843, -0.00000386, 0.00000157, 0.00000732, 0.00001275, 0.00001719, 0.00002005, 0.00002089, 0.00001957, 0.00001619, 0.00001118, 0.00000516, -0.00000110, -0.00000682, -0.00001130, -0.00001410, -0.00001499, -0.00001409, -0.00001175, -0.00000853, -0.00000507, -0.00000200, 0.00000019, 0.00000122, 0.00000104, -0.00000015, -0.00000198, -0.00000395, -0.00000557, -0.00000644, -0.00000631, -0.00000513, -0.00000308, -0.00000047, },
		{ 0.00000091, -0.00000262, -0.00000606, -0.00000906, -0.00001134, -0.00001277, -0.00001335, -0.00001316, -0.00001235, -0.00001108, -0.00000950, -0.00000768, -0.00000566, -0.00000346, -0.00000109, 0.00000142, 0.00000395, 0.00000634, 0.00000833, 0.00000965, 0.00001007, 0.00000941, 0.00000764, 0.00000487, 0.00000141, -0.00000232, -0.00000581, -0.00000855, -0.00001014, -0.00001035, -0.00000918, -0.00000687, -0.00000389, -0.00000083, 0.00000169, 0.00000312, 0.00000312, 0.00000163, -0.00000111, -0.00000456, -0.00000798, -0.00001057, -0.00001158, -0.00001049, -0.00000712, -0.00000169, 0.00000518, 0.00001253, 0.00001922, 0.00002413, 0.00002629, 0.00002510, 0.00002045, 0.00001274, 0.00000287, -0.00000788, -0.00001806, -0.00002622, -0.00003118, -0.00003217, -0.00002899, -0.00002201, -0.00001212, -0.00000064, 0.00001094, 0.00002114, 0.00002871, 0.00003280, 0.00003307, 0.00002971, 0.00002338, 0.00001504, 0.00000587, -0.00000300, -0.00001061, -0.00001631, -0.00001980, -0.00002115, -0.00002068, -0.00001891, -0.00001640, -0.00001364, -0.00001095, -0.00000846, -0.00000612, -0.00000372, -0.00000104, 0.00000211, 0.00000577, 0.00000980, 0.00001384, 0.00001738, 0.00001986, 0.00002072, 0.00001958, 0.00001629, 0.00001101, 0.00000420, -0.00000341, -0.00001092, -0.00001743, -0.00002213, -0.00002444, -0.00002414, -0.00002135, -0.00001654, -0.00001045, -0.00000396, 0.00000205, 0.00000681, 0.00000984, 0.00001093, 0.00001022, 0.00000810, 0.00000519, 0.00000216, -0.00000036, -0.00000189, -0.00000221, -0.00000131, 0.00000055, 0.00000293, 0.00000532, 0.00000719, 0.00000815, 0.00000795, 0.00000655, 0.00000410, },
		{ -0.00001395, -0.00001244, -0.00001038, -0.00000802, -0.00000558, -0.00000320, -0.00000096, 0.00000108, 0.00000293, 0.00000454, 0.00000586, 0.00000678, 0.00000718, 0.00000693, 0.00000593, 0.00000419, 0.00000181, -0.00000100, -0.00000391, -0.00000652, -0.00000847, -0.00000944, -0.00000926, -0.00000794, -0.00000570, -0.00000292, -0.00000008, 0.00000229, 0.00000375, 0.00000404, 0.00000309, 0.00000113, -0.00000141, -0.00000394, -0.00000583, -0.00000650, -0.00000562, -0.00000311, 0.00000074, 0.00000537, 0.00000996, 0.00001361, 0.00001548, 0.00001497, 0.00001184, 0.00000628, -0.00000104, -0.00000912, -0.00001670, -0.00002251, -0.00002547, -0.00002485, -0.00002045, -0.00001265, -0.00000237, 0.00000904, 0.00001999, 0.00002892, 0.00003449, 0.00003583, 0.00003265, 0.00002532, 0.00001477, 0.00000238, -0.00001023, -0.00002148, -0.00003001, -0.00003492, -0.00003586, -0.00003303, -0.00002718, -0.00001938, -0.00001087, -0.00000286, 0.00000371, 0.00000822, 0.00001052, 0.00001084, 0.00000972, 0.00000788, 0.00000603, 0.00000474, 0.00000431, 0.00000472, 0.00000567, 0.00000664, 0.00000702, 0.00000627, 0.00000406, 0.00000033, -0.00000462, -0.00001020, -0.00001559, -0.00001989, -0.00002228, -0.00002220, -0.00001939, -0.00001405, -0.00000674, 0.00000166, 0.00001009, 0.00001744, 0.00002282, 0.00002557, 0.00002547, 0.00002268, 0.00001776, 0.00001153, 0.00000495, -0.00000105, -0.00000571, -0.00000853, -0.00000939, -0.00000846, -0.00000622, -0.00000331, -0.00000043, 0.00000180, 0.00000292, 0.00000271, 0.00000119, -0.00000137, -0.00000455, -0.00000788, -0.00001086, -0.00001312, -0.00001442, -0.00001467, },
		{ 0.00000130, 0.00000231, 0.00000297, 0.00000336, 0.00000351, 0.00000342, 0.00000306, 0.00000238, 0.00000137, 0.00000007, -0.00000143, -0.00000294, -0.00000425, -0.00000510, -0.00000531, -0.00000477, -0.00000350, -0.00000166, 0.00000045, 0.00000246, 0.00000395, 0.00000459, 0.00000418, 0.00000269, 0.00000034, -0.00000251, -0.00000533, -0.00000758, -0.00000877, -0.00000860, -0.00000700, -0.00000419, -0.00000063, 0.00000301, 0.00000601, 0.00000773, 0.00000772, 0.00000584, 0.00000233, -0.00000225, -0.00000703, -0.00001105, -0.00001337, -0.00001330, -0.00001051, -0.00000514, 0.00000217, 0.00001037, 0.00001814, 0.00002413, 0.00002711, 0.00002629, 0.00002140, 0.00001279, 0.00000141, -0.00001129, -0.00002364, -0.00003391, -0.00004066, -0.00004292, -0.00004038, -0.00003340, -0.00002301, -0.00001071, 0.00000175, 0.00001265, 0.00002055, 0.00002454, 0.00002431, 0.00002024, 0.00001326, 0.00000470, -0.00000396, -0.00001131, -0.00001629, -0.00001830, -0.00001730, -0.00001379, -0.00000866, -0.00000301, 0.00000204, 0.00000558, 0.00000708, 0.00000648, 0.00000417, 0.00000090, -0.00000236, -0.00000462, -0.00000508, -0.00000330, 0.00000074, 0.00000659, 0.00001340, 0.00002010, 0.00002553, 0.00002871, 0.00002894, 0.00002598, 0.00002008, 0.00001192, 0.00000253, -0.00000689, -0.00001515, -0.00002128, -0.00002465, -0.00002506, -0.00002274, -0.00001831, -0.00001264, -0.00000672, -0.00000148, 0.00000237, 0.00000438, 0.00000446, 0.00000281, -0.00000008, -0.00000361, -0.00000712, -0.00001006, -0.00001201, -0.00001279, -0.00001240, -0.00001102, -0.00000896, -0.00000655, -0.00000413, -0.00000193, -0.00000010, },
		{ 0.00000289, 0.00000274, 0.00000243, 0.00000193, 0.00000127, 0.00000052, -0.00000018, -0.00000071, -0.00000097, -0.00000092, -0.00000062, -0.00000021, 0.00000009, 0.00000006, -0.00000053, -0.00000179, -0.00000369, -0.00000606, -0.00000858, -0.00001083, -0.00001236, -0.00001279, -0.00001188, -0.00000960, -0.00000616, -0.00000198, 0.00000233, 0.00000613, 0.00000882, 0.00000997, 0.00000943, 0.00000734, 0.00000414, 0.00000050, -0.00000281, -0.00000505, -0.00000570, -0.00000454, -0.00000174, 0.00000218, 0.00000635, 0.00000976, 0.00001144, 0.00001062, 0.00000691, 0.00000045, -0.00000813, -0.00001772, -0.00002692, -0.00003425, -0.00003840, -0.00003845, -0.00003408, -0.00002566, -0.00001420, -0.00000123, 0.00001139, 0.00002179, 0.00002841, 0.00003020, 0.00002684, 0.00001881, 0.00000732, -0.00000589, -0.00001877, -0.00002932, -0.00003589, -0.00003744, -0.00003370, -0.00002519, -0.00001317, 0.00000065, 0.00001429, 0.00002589, 0.00003400, 0.00003774, 0.00003698, 0.00003226, 0.00002475, 0.00001594, 0.00000745, 0.00000066, -0.00000343, -0.00000442, -0.00000253, 0.00000147, 0.00000642, 0.00001099, 0.00001391, 0.00001428, 0.00001162, 0.00000606, -0.00000176, -0.00001077, -0.00001964, -0.00002707, -0.00003195, -0.00003355, -0.00003168, -0.00002666, -0.00001925, -0.00001058, -0.00000189, 0.00000565, 0.00001110, 0.00001388, 0.00001385, 0.00001132, 0.00000689, 0.00000142, -0.00000419, -0.00000913, -0.00001276, -0.00001475, -0.00001505, -0.00001386, -0.00001158, -0.00000868, -0.00000565, -0.00000288, -0.00000063, 0.00000100, 0.00000203, 0.00000259, 0.00000284, 0.00000292, 0.00000294, 0.00000294, },
		{ -0.00000228, -0.00000380, -0.00000515, -0.00000625, -0.00000708, -0.00000768, -0.00000817, -0.00000866, -0.00000925, -0.00000995, -0.00001070, -0.00001134, -0.00001161, -0.00001125, -0.00001002, -0.00000781, -0.00000463, -0.00000071, 0.00000356, 0.00000767, 0.00001104, 0.00001316, 0.00001367, 0.00001242, 0.00000957, 0.00000550, 0.00000082, -0.00000376, -0.00000756, -0.00001004, -0.00001095, -0.00001032, -0.00000854, -0.00000624, -0.00000419, -0.00000313, -0.00000365, -0.00000598, -0.00000998, -0.00001510, -0.00002045, -0.00002494, -0.00002752, -0.00002732, -0.00002391, -0.00001739, -0.00000842, 0.00000182, 0.00001181, 0.00001992, 0.00002470, 0.00002512, 0.00002083, 0.00001220, 0.00000039, -0.00001287, -0.00002546, -0.00003523, -0.00004036, -0.00003961, -0.00003260, -0.00001990, -0.00000293, 0.00001617, 0.00003488, 0.00005069, 0.00006143, 0.00006566, 0.00006283, 0.00005339, 0.00003871, 0.00002085, 0.00000225, -0.00001467, -0.00002784, -0.00003587, -0.00003822, -0.00003524, -0.00002809, -0.00001846, -0.00000832, 0.00000042, 0.00000627, 0.00000835, 0.00000649, 0.00000126, -0.00000618, -0.00001431, -0.00002155, -0.00002649, -0.00002818, -0.00002625, -0.00002096, -0.00001313, -0.00000399, 0.00000505, 0.00001262, 0.00001764, 0.00001946, 0.00001796, 0.00001354, 0.00000702, -0.00000053, -0.00000795, -0.00001419, -0.00001848, -0.00002039, -0.00001991, -0.00001735, -0.00001330, -0.00000849, -0.00000362, 0.00000070, 0.00000406, 0.00000627, 0.00000739, 0.00000760, 0.00000720, 0.00000650, 0.00000574, 0.00000506, 0.00000452, 0.00000404, 0.00000352, 0.00000284, 0.00000190, 0.00000069, -0.00000074, },
		{ -0.00001080, -0.00000966, -0.00000842, -0.00000718, -0.00000597, -0.00000473, -0.00000336, -0.00000176, 0.00000016, 0.00000240, 0.00000481, 0.00000717, 0.00000912, 0.00001029, 0.00001033, 0.00000901, 0.00000626, 0.00000227, -0.00000260, -0.00000779, -0.00001265, -0.00001657, -0.00001904, -0.00001981, -0.00001890, -0.00001663, -0.00001353, -0.00001031, -0.00000766, -0.00000615, -0.00000606, -0.00000739, -0.00000975, -0.00001250, -0.00001483, -0.00001592, -0.00001516, -0.00001224, -0.00000731, -0.00000094, 0.00000589, 0.00001199, 0.00001618, 0.00001747, 0.00001532, 0.00000978, 0.00000155, -0.00000807, -0.00001741, -0.00002459, -0.00002793, -0.00002622, -0.00001895, -0.00000651, 0.00000982, 0.00002804, 0.00004564, 0.00005999, 0.00006874, 0.00007022, 0.00006368, 0.00004949, 0.00002911, 0.00000489, -0.00002022, -0.00004313, -0.00006102, -0.00007179, -0.00007431, -0.00006862, -0.00005589, -0.00003821, -0.00001827, 0.00000107, 0.00001717, 0.00002807, 0.00003265, 0.00003086, 0.00002358, 0.00001250, -0.00000023, -0.00001238, -0.00002195, -0.00002753, -0.00002846, -0.00002491, -0.00001777, -0.00000849, 0.00000123, 0.00000970, 0.00001556, 0.00001798, 0.00001674, 0.00001226, 0.00000548, -0.00000234, -0.00000985, -0.00001583, -0.00001939, -0.00002008, -0.00001797, -0.00001354, -0.00000759, -0.00000112, 0.00000495, 0.00000981, 0.00001296, 0.00001422, 0.00001373, 0.00001186, 0.00000913, 0.00000608, 0.00000317, 0.00000074, -0.00000109, -0.00000235, -0.00000321, -0.00000390, -0.00000465, -0.00000559, -0.00000678, -0.00000817, -0.00000959, -0.00001087, -0.00001181, -0.00001229, -0.00001224, -0.00001171, },
		{ 0.00000202, 0.00000191, 0.00000173, 0.00000153, 0.00000129, 0.00000090, 0.00000019, -0.00000104, -0.00000291, -0.00000548, -0.00000865, -0.00001216, -0.00001562, -0.00001856, -0.00002053, -0.00002117, -0.00002031, -0.00001798, -0.00001450, -0.00001037, -0.00000623, -0.00000270, -0.00000034, 0.00000056, -0.00000004, -0.00000186, -0.00000433, -0.00000676, -0.00000841, -0.00000868, -0.00000724, -0.00000412, 0.00000029, 0.00000526, 0.00000986, 0.00001318, 0.00001443, 0.00001323, 0.00000965, 0.00000428, -0.00000185, -0.00000740, -0.00001100, -0.00001147, -0.00000810, -0.00000084, 0.00000963, 0.00002189, 0.00003402, 0.00004383, 0.00004922, 0.00004855, 0.00004097, 0.00002660, 0.00000663, -0.00001684, -0.00004098, -0.00006270, -0.00007904, -0.00008765, -0.00008716, -0.00007741, -0.00005948, -0.00003562, -0.00000887, 0.00001732, 0.00003961, 0.00005526, 0.00006248, 0.00006071, 0.00005066, 0.00003414, 0.00001382, -0.00000723, -0.00002603, -0.00004007, -0.00004770, -0.00004832, -0.00004239, -0.00003133, -0.00001721, -0.00000239, 0.00001085, 0.00002068, 0.00002597, 0.00002640, 0.00002248, 0.00001536, 0.00000661, -0.00000208, -0.00000919, -0.00001361, -0.00001480, -0.00001283, -0.00000833, -0.00000228, 0.00000411, 0.00000969, 0.00001353, 0.00001508, 0.00001422, 0.00001123, 0.00000673, 0.00000149, -0.00000366, -0.00000804, -0.00001119, -0.00001293, -0.00001334, -0.00001273, -0.00001153, -0.00001018, -0.00000905, -0.00000836, -0.00000817, -0.00000837, -0.00000873, -0.00000899, -0.00000892, -0.00000835, -0.00000727, -0.00000575, -0.00000397, -0.00000216, -0.00000055, 0.00000071, 0.00000153, 0.00000194, },
		{ -0.00001126, -0.00001141, -0.00001166, -0.00001209, -0.00001267, -0.00001326, -0.00001364, -0.00001354, -0.00001272, -0.00001105, -0.00000855, -0.00000540, -0.00000197, 0.00000127, 0.00000385, 0.00000536, 0.00000554, 0.00000440, 0.00000220, -0.00000057, -0.00000330, -0.00000530, -0.00000602, -0.00000514, -0.00000261, 0.00000123, 0.00000577, 0.00001019, 0.00001365, 0.00001540, 0.00001500, 0.00001241, 0.00000802, 0.00000260, -0.00000280, -0.00000706, -0.00000924, -0.00000874, -0.00000550, -0.00000005, 0.00000652, 0.00001273, 0.00001694, 0.00001767, 0.00001388, 0.00000523, -0.00000779, -0.00002384, -0.00004087, -0.00005642, -0.00006793, -0.00007325, -0.00007093, -0.00006054, -0.00004282, -0.00001963, 0.00000627, 0.00003160, 0.00005301, 0.00006759, 0.00007331, 0.00006934, 0.00005619, 0.00003565, 0.00001055, -0.00001566, -0.00003943, -0.00005760, -0.00006786, -0.00006906, -0.00006138, -0.00004620, -0.00002593, -0.00000355, 0.00001780, 0.00003533, 0.00004698, 0.00005165, 0.00004937, 0.00004114, 0.00002877, 0.00001450, 0.00000063, -0.00001083, -0.00001849, -0.00002171, -0.00002064, -0.00001614, -0.00000955, -0.00000243, 0.00000376, 0.00000785, 0.00000919, 0.00000768, 0.00000375, -0.00000175, -0.00000778, -0.00001327, -0.00001735, -0.00001944, -0.00001939, -0.00001742, -0.00001405, -0.00000999, -0.00000596, -0.00000258, -0.00000027, 0.00000081, 0.00000079, -0.00000004, -0.00000126, -0.00000246, -0.00000328, -0.00000356, -0.00000328, -0.00000259, -0.00000177, -0.00000109, -0.00000083, -0.00000114, -0.00000205, -0.00000346, -0.00000518, -0.00000694, -0.00000852, -0.00000976, -0.00001058, -0.00001103, },
		{ -0.00000060, -0.00000040, -0.00000008, 0.00000051, 0.00000140, 0.00000249, 0.00000356, 0.00000432, 0.00000450, 0.00000392, 0.00000253, 0.00000051, -0.00000183, -0.00000403, -0.00000557, -0.00000603, -0.00000515, -0.00000293, 0.00000035, 0.00000416, 0.00000776, 0.00001039, 0.00001137, 0.00001026, 0.00000697, 0.00000183, -0.00000449, -0.00001104, -0.00001678, -0.00002077, -0.00002234, -0.00002127, -0.00001784, -0.00001283, -0.00000736, -0.00000276, -0.00000023, -0.00000067, -0.00000443, -0.00001117, -0.00001993, -0.00002917, -0.00003706, -0.00004178, -0.00004184, -0.00003643, -0.00002556, -0.00001021, 0.00000782, 0.00002607, 0.00004186, 0.00005265, 0.00005649, 0.00005236, 0.00004041, 0.00002197, -0.00000060, -0.00002420, -0.00004549, -0.00006132, -0.00006929, -0.00006806, -0.00005760, -0.00003922, -0.00001535, 0.00001081, 0.00003573, 0.00005613, 0.00006941, 0.00007402, 0.00006968, 0.00005736, 0.00003912, 0.00001769, -0.00000388, -0.00002274, -0.00003666, -0.00004428, -0.00004529, -0.00004038, -0.00003108, -0.00001942, -0.00000758, 0.00000249, 0.00000932, 0.00001216, 0.00001105, 0.00000668, 0.00000027, -0.00000674, -0.00001291, -0.00001711, -0.00001867, -0.00001743, -0.00001378, -0.00000852, -0.00000265, 0.00000279, 0.00000691, 0.00000914, 0.00000929, 0.00000755, 0.00000441, 0.00000053, -0.00000337, -0.00000671, -0.00000908, -0.00001036, -0.00001065, -0.00001027, -0.00000961, -0.00000908, -0.00000899, -0.00000946, -0.00001046, -0.00001177, -0.00001309, -0.00001406, -0.00001442, -0.00001399, -0.00001277, -0.00001090, -0.00000865, -0.00000634, -0.00000426, -0.00000262, -0.00000151, -0.00000090, },
		{ -0.00000429, -0.00000436, -0.00000447, -0.00000480, -0.00000541, -0.00000625, -0.00000709, -0.00000766, -0.00000766, -0.00000690, -0.00000534, -0.00000315, -0.00000069, 0.00000151, 0.00000285, 0.00000281, 0.00000105, -0.00000249, -0.00000752, -0.00001345, -0.00001941, -0.00002445, -0.00002769, -0.00002846, -0.00002648, -0.00002195, -0.00001551, -0.00000818, -0.00000118, 0.00000428, 0.00000724, 0.00000716, 0.00000405, -0.00000147, -0.00000830, -0.00001502, -0.00002016, -0.00002246, -0.00002113, -0.00001607, -0.00000787, 0.00000220, 0.00001241, 0.00002086, 0.00002579, 0.00002598, 0.00002094, 0.00001111, -0.00000219, -0.00001689, -0.00003052, -0.00004058, -0.00004499, -0.00004241, -0.00003259, -0.00001640, 0.00000423, 0.00002653, 0.00004738, 0.00006370, 0.00007298, 0.00007364, 0.00006535, 0.00004902, 0.00002675, 0.00000145, -0.00002355, -0.00004499, -0.00006023, -0.00006754, -0.00006639, -0.00005746, -0.00004252, -0.00002408, -0.00000498, 0.00001202, 0.00002469, 0.00003163, 0.00003245, 0.00002772, 0.00001882, 0.00000768, -0.00000359, -0.00001307, -0.00001934, -0.00002165, -0.00002001, -0.00001512, -0.00000818, -0.00000069, 0.00000592, 0.00001046, 0.00001219, 0.00001095, 0.00000709, 0.00000137, -0.00000516, -0.00001146, -0.00001659, -0.00001995, -0.00002130, -0.00002077, -0.00001881, -0.00001604, -0.00001312, -0.00001062, -0.00000888, -0.00000799, -0.00000779, -0.00000795, -0.00000803, -0.00000764, -0.00000650, -0.00000454, -0.00000188, 0.00000117, 0.00000418, 0.00000670, 0.00000836, 0.00000893, 0.00000837, 0.00000686, 0.00000468, 0.00000225, -0.00000007, -0.00000196, -0.00000327, -0.00000400, },
		{ -0.00000561, -0.00000548, -0.00000534, -0.00000500, -0.00000439, -0.00000359, -0.00000286, -0.00000254, -0.00000298, -0.00000445, -0.00000704, -0.00001059, -0.00001466, -0.00001863, -0.00002175, -0.00002330, -0.00002272, -0.00001979, -0.00001466, -0.00000790, -0.00000041, 0.00000669, 0.00001224, 0.00001527, 0.00001519, 0.00001188, 0.00000577, -0.00000220, -0.00001076, -0.00001850, -0.00002411, -0.00002664, -0.00002565, -0.00002135, -0.00001450, -0.00000635, 0.00000164, 0.00000805, 0.00001176, 0.00001221, 0.00000950, 0.00000440, -0.00000177, -0.00000739, -0.00001086, -0.00001089, -0.00000680, 0.00000127, 0.00001238, 0.00002488, 0.00003661, 0.00004532, 0.00004905, 0.00004647, 0.00003717, 0.00002182, 0.00000209, -0.00001955, -0.00004019, -0.00005689, -0.00006721, -0.00006954, -0.00006341, -0.00004955, -0.00002983, -0.00000697, 0.00001590, 0.00003567, 0.00004974, 0.00005641, 0.00005510, 0.00004642, 0.00003202, 0.00001431, -0.00000394, -0.00002005, -0.00003185, -0.00003798, -0.00003806, -0.00003268, -0.00002326, -0.00001178, -0.00000038, 0.00000899, 0.00001486, 0.00001647, 0.00001382, 0.00000765, -0.00000077, -0.00000987, -0.00001812, -0.00002423, -0.00002738, -0.00002731, -0.00002434, -0.00001921, -0.00001292, -0.00000657, -0.00000108, 0.00000291, 0.00000515, 0.00000575, 0.00000515, 0.00000395, 0.00000278, 0.00000215, 0.00000234, 0.00000334, 0.00000490, 0.00000655, 0.00000778, 0.00000810, 0.00000717, 0.00000493, 0.00000153, -0.00000262, -0.00000700, -0.00001101, -0.00001414, -0.00001605, -0.00001661, -0.00001595, -0.00001436, -0.00001225, -0.00001007, -0.00000817, -0.00000679, -0.00000598, },
		{ -0.00000695, -0.00000739, -0.00000798, -0.00000897, -0.00001043, -0.00001225, -0.00001412, -0.00001557, -0.00001610, -0.00001526, -0.00001281, -0.00000876, -0.00000351, 0.00000228, 0.00000773, 0.00001187, 0.00001386, 0.00001313, 0.00000953, 0.00000338, -0.00000456, -0.00001315, -0.00002107, -0.00002708, -0.00003016, -0.00002974, -0.00002582, -0.00001893, -0.00001010, -0.00000064, 0.00000807, 0.00001480, 0.00001876, 0.00001965, 0.00001775, 0.00001384, 0.00000905, 0.00000458, 0.00000150, 0.00000048, 0.00000167, 0.00000463, 0.00000843, 0.00001181, 0.00001345, 0.00001227, 0.00000766, -0.00000029, -0.00001079, -0.00002236, -0.00003311, -0.00004101, -0.00004427, -0.00004167, -0.00003284, -0.00001841, 0.00000006, 0.00002023, 0.00003934, 0.00005461, 0.00006371, 0.00006511, 0.00005837, 0.00004421, 0.00002448, 0.00000182, -0.00002069, -0.00004001, -0.00005360, -0.00005981, -0.00005812, -0.00004916, -0.00003465, -0.00001703, 0.00000089, 0.00001640, 0.00002731, 0.00003227, 0.00003094, 0.00002398, 0.00001293, -0.00000016, -0.00001299, -0.00002351, -0.00003015, -0.00003208, -0.00002930, -0.00002257, -0.00001321, -0.00000289, 0.00000674, 0.00001430, 0.00001888, 0.00002015, 0.00001839, 0.00001432, 0.00000897, 0.00000342, -0.00000138, -0.00000480, -0.00000661, -0.00000698, -0.00000639, -0.00000551, -0.00000501, -0.00000543, -0.00000701, -0.00000970, -0.00001314, -0.00001675, -0.00001984, -0.00002179, -0.00002217, -0.00002080, -0.00001783, -0.00001368, -0.00000896, -0.00000435, -0.00000050, 0.00000216, 0.00000339, 0.00000325, 0.00000202, 0.00000011, -0.00000201, -0.00000393, -0.00000540, -0.00000637, },
		{ -0.00000813, -0.00000799, -0.00000748, -0.00000623, -0.00000403, -0.00000092, 0.00000278, 0.00000652, 0.00000958, 0.00001124, 0.00001094, 0.00000838, 0.00000365, -0.00000275, -0.00001000, -0.00001705, -0.00002280, -0.00002631, -0.00002695, -0.00002457, -0.00001948, -0.00001248, -0.00000464, 0.00000278, 0.00000866, 0.00001212, 0.00001274, 0.00001060, 0.00000625, 0.00000061, -0.00000521, -0.00001014, -0.00001335, -0.00001438, -0.00001325, -0.00001041, -0.00000665, -0.00000289, 0.00000003, 0.00000152, 0.00000144, 0.00000009, -0.00000177, -0.00000315, -0.00000296, -0.00000034, 0.00000514, 0.00001333, 0.00002339, 0.00003394, 0.00004320, 0.00004929, 0.00005058, 0.00004603, 0.00003541, 0.00001942, -0.00000030, -0.00002140, -0.00004112, -0.00005674, -0.00006596, -0.00006732, -0.00006047, -0.00004620, -0.00002643, -0.00000387, 0.00001836, 0.00003716, 0.00005000, 0.00005525, 0.00005242, 0.00004226, 0.00002657, 0.00000792, -0.00001076, -0.00002667, -0.00003753, -0.00004199, -0.00003973, -0.00003150, -0.00001895, -0.00000434, 0.00000990, 0.00002153, 0.00002888, 0.00003102, 0.00002796, 0.00002048, 0.00001003, -0.00000162, -0.00001268, -0.00002162, -0.00002745, -0.00002979, -0.00002889, -0.00002552, -0.00002078, -0.00001581, -0.00001163, -0.00000890, -0.00000786, -0.00000829, -0.00000962, -0.00001111, -0.00001197, -0.00001160, -0.00000969, -0.00000632, -0.00000191, 0.00000286, 0.00000717, 0.00001026, 0.00001155, 0.00001078, 0.00000806, 0.00000380, -0.00000132, -0.00000652, -0.00001106, -0.00001437, -0.00001618, -0.00001647, -0.00001554, -0.00001383, -0.00001189, -0.00001017, -0.00000897, -0.00000835, },
		{ 0.00000426, 0.00000585, 0.00000706, 0.00000721, 0.00000576, 0.00000243, -0.00000271, -0.00000922, -0.00001636, -0.00002321, -0.00002886, -0.00003254, -0.00003379, -0.00003255, -0.00002916, -0.00002432, -0.00001893, -0.00001391, -0.00001005, -0.00000785, -0.00000742, -0.00000845, -0.00001032, -0.00001219, -0.00001319, -0.00001260, -0.00001000, -0.00000534, 0.00000101, 0.00000836, 0.00001585, 0.00002256, 0.00002775, 0.00003098, 0.00003217, 0.00003161, 0.00002987, 0.00002766, 0.00002564, 0.00002424, 0.00002357, 0.00002333, 0.00002290, 0.00002141, 0.00001803, 0.00001211, 0.00000345, -0.00000760, -0.00002008, -0.00003249, -0.00004301, -0.00004981, -0.00005132, -0.00004658, -0.00003550, -0.00001892, 0.00000141, 0.00002300, 0.00004302, 0.00005867, 0.00006764, 0.00006849, 0.00006090, 0.00004577, 0.00002510, 0.00000174, -0.00002107, -0.00004016, -0.00005296, -0.00005781, -0.00005433, -0.00004334, -0.00002682, -0.00000753, 0.00001141, 0.00002704, 0.00003699, 0.00003985, 0.00003536, 0.00002441, 0.00000886, -0.00000881, -0.00002589, -0.00003990, -0.00004896, -0.00005206, -0.00004915, -0.00004109, -0.00002946, -0.00001623, -0.00000341, 0.00000729, 0.00001469, 0.00001831, 0.00001837, 0.00001565, 0.00001133, 0.00000665, 0.00000273, 0.00000030, -0.00000038, 0.00000047, 0.00000222, 0.00000401, 0.00000495, 0.00000436, 0.00000188, -0.00000242, -0.00000804, -0.00001418, -0.00001987, -0.00002416, -0.00002634, -0.00002603, -0.00002328, -0.00001855, -0.00001258, -0.00000631, -0.00000063, 0.00000376, 0.00000644, 0.00000738, 0.00000689, 0.00000552, 0.00000394, 0.00000275, 0.00000237, 0.00000294, },
		{ -0.00001552, -0.00002234, -0.00003008, -0.00003775, -0.00004431, -0.00004887, -0.00005081, -0.00004995, -0.00004646, -0.00004087, -0.00003393, -0.00002644, -0.00001912, -0.00001246, -0.00000664, -0.00000159, 0.00000302, 0.00000761, 0.00001259, 0.00001821, 0.00002445, 0.00003104, 0.00003746, 0.00004300, 0.00004695, 0.00004872, 0.00004795, 0.00004463, 0.00003907, 0.00003188, 0.00002387, 0.00001586, 0.00000858, 0.00000253, -0.00000212, -0.00000548, -0.00000791, -0.00000986, -0.00001173, -0.00001375, -0.00001584, -0.00001758, -0.00001832, -0.00001722, -0.00001355, -0.00000681, 0.00000302, 0.00001539, 0.00002913, 0.00004258, 0.00005380, 0.00006084, 0.00006212, 0.00005671, 0.00004455, 0.00002662, 0.00000479, -0.00001833, -0.00003976, -0.00005660, -0.00006651, -0.00006802, -0.00006089, -0.00004611, -0.00002586, -0.00000313, 0.00001866, 0.00003623, 0.00004691, 0.00004909, 0.00004246, 0.00002806, 0.00000813, -0.00001427, -0.00003570, -0.00005291, -0.00006332, -0.00006543, -0.00005899, -0.00004505, -0.00002572, -0.00000382, 0.00001754, 0.00003551, 0.00004789, 0.00005342, 0.00005193, 0.00004430, 0.00003220, 0.00001779, 0.00000332, -0.00000924, -0.00001852, -0.00002388, -0.00002546, -0.00002404, -0.00002083, -0.00001719, -0.00001430, -0.00001294, -0.00001340, -0.00001537, -0.00001812, -0.00002063, -0.00002187, -0.00002098, -0.00001756, -0.00001167, -0.00000389, 0.00000479, 0.00001315, 0.00001999, 0.00002432, 0.00002559, 0.00002371, 0.00001915, 0.00001277, 0.00000567, -0.00000100, -0.00000632, -0.00000967, -0.00001094, -0.00001046, -0.00000894, -0.00000731, -0.00000655, -0.00000743, -0.00001041, },
		{ -0.00006281, -0.00006044, -0.00005421, -0.00004473, -0.00003295, -0.00001998, -0.00000684, 0.00000559, 0.00001673, 0.00002627, 0.00003417, 0.00004055, 0.00004558, 0.00004941, 0.00005205, 0.00005341, 0.00005328, 0.00005144, 0.00004771, 0.00004209, 0.00003479, 0.00002628, 0.00001722, 0.00000842, 0.00000070, -0.00000523, -0.00000890, -0.00001013, -0.00000905, -0.00000608, -0.00000182, 0.00000306, 0.00000797, 0.00001250, 0.00001647, 0.00001993, 0.00002310, 0.00002624, 0.00002954, 0.00003299, 0.00003627, 0.00003879, 0.00003973, 0.00003819, 0.00003341, 0.00002495, 0.00001288, -0.00000208, -0.00001862, -0.00003489, -0.00004879, -0.00005825, -0.00006161, -0.00005790, -0.00004713, -0.00003035, -0.00000957, 0.00001244, 0.00003258, 0.00004783, 0.00005579, 0.00005503, 0.00004541, 0.00002812, 0.00000559, -0.00001888, -0.00004161, -0.00005905, -0.00006839, -0.00006797, -0.00005758, -0.00003851, -0.00001333, 0.00001441, 0.00004082, 0.00006216, 0.00007547, 0.00007899, 0.00007234, 0.00005665, 0.00003424, 0.00000830, -0.00001766, -0.00004032, -0.00005706, -0.00006629, -0.00006764, -0.00006190, -0.00005078, -0.00003661, -0.00002183, -0.00000863, 0.00000143, 0.00000759, 0.00000995, 0.00000936, 0.00000713, 0.00000475, 0.00000353, 0.00000433, 0.00000740, 0.00001231, 0.00001811, 0.00002349, 0.00002709, 0.00002777, 0.00002487, 0.00001836, 0.00000885, -0.00000248, -0.00001412, -0.00002451, -0.00003226, -0.00003649, -0.00003690, -0.00003390, -0.00002847, -0.00002200, -0.00001605, -0.00001204, -0.00001103, -0.00001352, -0.00001937, -0.00002784, -0.00003771, -0.00004750, -0.00005573, -0.00006112, },
		{ 0.00001270, 0.00002991, 0.00004411, 0.00005481, 0.00006194, 0.00006572, 0.00006660, 0.00006511, 0.00006173, 0.00005689, 0.00005094, 0.00004416, 0.00003685, 0.00002934, 0.00002204, 0.00001543, 0.00001004, 0.00000633, 0.00000468, 0.00000525, 0.00000796, 0.00001245, 0.00001814, 0.00002429, 0.00003011, 0.00003489, 0.00003810, 0.00003944, 0.00003890, 0.00003672, 0.00003329, 0.00002903, 0.00002436, 0.00001950, 0.00001454, 0.00000936, 0.00000377, -0.00000241, -0.00000923, -0.00001648, -0.00002366, -0.00002997, -0.00003441, -0.00003594, -0.00003370, -0.00002723, -0.00001664, -0.00000273, 0.00001305, 0.00002870, 0.00004198, 0.00005070, 0.00005312, 0.00004829, 0.00003629, 0.00001829, -0.00000346, -0.00002600, -0.00004599, -0.00006024, -0.00006624, -0.00006258, -0.00004925, -0.00002773, -0.00000079, 0.00002782, 0.00005393, 0.00007357, 0.00008359, 0.00008213, 0.00006897, 0.00004556, 0.00001485, -0.00001914, -0.00005192, -0.00007917, -0.00009740, -0.00010442, -0.00009963, -0.00008410, -0.00006032, -0.00003177, -0.00000240, 0.00002402, 0.00004444, 0.00005691, 0.00006085, 0.00005700, 0.00004714, 0.00003380, 0.00001967, 0.00000718, -0.00000192, -0.00000681, -0.00000765, -0.00000548, -0.00000190, 0.00000125, 0.00000234, 0.00000026, -0.00000534, -0.00001399, -0.00002447, -0.00003508, -0.00004398, -0.00004954, -0.00005064, -0.00004697, -0.00003904, -0.00002818, -0.00001626, -0.00000542, 0.00000231, 0.00000539, 0.00000296, -0.00000493, -0.00001736, -0.00003270, -0.00004881, -0.00006344, -0.00007449, -0.00008038, -0.00008020, -0.00007382, -0.00006186, -0.00004557, -0.00002655, -0.00000658, },
		{ 0.00008599, 0.00007809, 0.00006895, 0.00005950, 0.00005044, 0.00004227, 0.00003530, 0.00002973, 0.00002568, 0.00002321, 0.00002236, 0.00002307, 0.00002518, 0.00002842, 0.00003235, 0.00003642, 0.00004004, 0.00004258, 0.00004356, 0.00004266, 0.00003979, 0.00003514, 0.00002912, 0.00002231, 0.00001535, 0.00000885, 0.00000328, -0.00000107, -0.00000415, -0.00000607, -0.00000705, -0.00000737, -0.00000720, -0.00000659, -0.00000543, -0.00000350, -0.00000051, 0.00000372, 0.00000921, 0.00001565, 0.00002239, 0.00002848, 0.00003275, 0.00003403, 0.00003138, 0.00002429, 0.00001291, -0.00000188, -0.00001848, -0.00003474, -0.00004822, -0.00005663, -0.00005814, -0.00005181, -0.00003782, -0.00001759, 0.00000638, 0.00003074, 0.00005180, 0.00006605, 0.00007075, 0.00006441, 0.00004709, 0.00002049, -0.00001217, -0.00004663, -0.00007812, -0.00010210, -0.00011489, -0.00011432, -0.00010001, -0.00007352, -0.00003811, 0.00000169, 0.00004075, 0.00007409, 0.00009759, 0.00010856, 0.00010612, 0.00009125, 0.00006662, 0.00003606, 0.00000402, -0.00002521, -0.00004808, -0.00006230, -0.00006710, -0.00006325, -0.00005280, -0.00003865, -0.00002398, -0.00001170, -0.00000394, -0.00000171, -0.00000483, -0.00001202, -0.00002121, -0.00003000, -0.00003611, -0.00003792, -0.00003468, -0.00002678, -0.00001559, -0.00000328, 0.00000757, 0.00001448, 0.00001548, 0.00000949, -0.00000346, -0.00002220, -0.00004459, -0.00006782, -0.00008881, -0.00010469, -0.00011321, -0.00011301, -0.00010383, -0.00008646, -0.00006263, -0.00003467, -0.00000523, 0.00002311, 0.00004813, 0.00006824, 0.00008252, 0.00009081, 0.00009352, 0.00009153, },
		{ 0.00003798, 0.00003666, 0.00003710, 0.00003881, 0.00004134, 0.00004427, 0.00004723, 0.00004984, 0.00005177, 0.00005268, 0.00005231, 0.00005049, 0.00004715, 0.00004243, 0.00003662, 0.00003015, 0.00002358, 0.00001749, 0.00001237, 0.00000860, 0.00000634, 0.00000555, 0.00000596, 0.00000717, 0.00000872, 0.00001016, 0.00001119, 0.00001161, 0.00001143, 0.00001077, 0.00000981, 0.00000871, 0.00000752, 0.00000613, 0.00000429, 0.00000161, -0.00000226, -0.00000760, -0.00001439, -0.00002228, -0.00003052, -0.00003803, -0.00004351, -0.00004566, -0.00004345, -0.00003636, -0.00002459, -0.00000915, 0.00000817, 0.00002498, 0.00003862, 0.00004655, 0.00004682, 0.00003845, 0.00002171, -0.00000175, -0.00002908, -0.00005651, -0.00007987, -0.00009522, -0.00009946, -0.00009092, -0.00006967, -0.00003767, 0.00000145, 0.00004279, 0.00008087, 0.00011041, 0.00012713, 0.00012835, 0.00011350, 0.00008419, 0.00004405, -0.00000181, -0.00004750, -0.00008720, -0.00011603, -0.00013070, -0.00012998, -0.00011483, -0.00008819, -0.00005446, -0.00001881, 0.00001362, 0.00003854, 0.00005305, 0.00005609, 0.00004844, 0.00003251, 0.00001183, -0.00000959, -0.00002798, -0.00004046, -0.00004545, -0.00004296, -0.00003446, -0.00002261, -0.00001073, -0.00000218, 0.00000025, -0.00000515, -0.00001871, -0.00003928, -0.00006441, -0.00009069, -0.00011435, -0.00013175, -0.00014000, -0.00013733, -0.00012333, -0.00009903, -0.00006664, -0.00002927, 0.00000953, 0.00004630, 0.00007803, 0.00010254, 0.00011865, 0.00012620, 0.00012597, 0.00011942, 0.00010842, 0.00009495, 0.00008086, 0.00006764, 0.00005634, 0.00004757, 0.00004150, },
		{ 0.00008187, 0.00007972, 0.00007602, 0.00007094, 0.00006464, 0.00005738, 0.00004942, 0.00004108, 0.00003275, 0.00002482, 0.00001767, 0.00001161, 0.00000684, 0.00000340, 0.00000116, -0.00000015, -0.00000094, -0.00000163, -0.00000262, -0.00000422, -0.00000655, -0.00000956, -0.00001305, -0.00001671, -0.00002019, -0.00002320, -0.00002555, -0.00002721, -0.00002827, -0.00002891, -0.00002935, -0.00002971, -0.00002997, -0.00002995, -0.00002925, -0.00002741, -0.00002396, -0.00001862, -0.00001138, -0.00000267, 0.00000665, 0.00001535, 0.00002195, 0.00002501, 0.00002338, 0.00001650, 0.00000459, -0.00001119, -0.00002886, -0.00004574, -0.00005890, -0.00006553, -0.00006350, -0.00005179, -0.00003076, -0.00000233, 0.00003021, 0.00006252, 0.00008983, 0.00010759, 0.00011226, 0.00010184, 0.00007639, 0.00003811, -0.00000881, -0.00005872, -0.00010520, -0.00014202, -0.00016402, -0.00016790, -0.00015274, -0.00012022, -0.00007442, -0.00002127, 0.00003227, 0.00007920, 0.00011349, 0.00013093, 0.00012969, 0.00011061, 0.00007694, 0.00003388, -0.00001227, -0.00005511, -0.00008901, -0.00011004, -0.00011640, -0.00010870, -0.00008971, -0.00006387, -0.00003652, -0.00001301, 0.00000215, 0.00000598, -0.00000250, -0.00002218, -0.00005006, -0.00008177, -0.00011227, -0.00013658, -0.00015058, -0.00015155, -0.00013856, -0.00011251, -0.00007591, -0.00003254, 0.00001322, 0.00005695, 0.00009479, 0.00012392, 0.00014278, 0.00015117, 0.00015013, 0.00014160, 0.00012811, 0.00011231, 0.00009663, 0.00008302, 0.00007273, 0.00006629, 0.00006361, 0.00006408, 0.00006681, 0.00007077, 0.00007500, 0.00007869, 0.00008127, 0.00008238, },
		{ 0.00000955, -0.00000106, -0.00001024, -0.00001790, -0.00002399, -0.00002855, -0.00003166, -0.00003351, -0.00003436, -0.00003454, -0.00003439, -0.00003422, -0.00003425, -0.00003460, -0.00003521, -0.00003592, -0.00003646, -0.00003656, -0.00003593, -0.00003444, -0.00003205, -0.00002891, -0.00002525, -0.00002141, -0.00001770, -0.00001436, -0.00001151, -0.00000910, -0.00000699, -0.00000492, -0.00000269, -0.00000021, 0.00000244, 0.00000496, 0.00000684, 0.00000747, 0.00000629, 0.00000291, -0.00000269, -0.00001004, -0.00001822, -0.00002583, -0.00003129, -0.00003298, -0.00002968, -0.00002080, -0.00000665, 0.00001142, 0.00003112, 0.00004943, 0.00006299, 0.00006866, 0.00006405, 0.00004803, 0.00002111, -0.00001444, -0.00005474, -0.00009469, -0.00012865, -0.00015124, -0.00015815, -0.00014694, -0.00011752, -0.00007235, -0.00001629, 0.00004396, 0.00010071, 0.00014634, 0.00017436, 0.00018043, 0.00016299, 0.00012363, 0.00006692, -0.00000015, -0.00006913, -0.00013126, -0.00017871, -0.00020568, -0.00020921, -0.00018964, -0.00015049, -0.00009795, -0.00004000, 0.00001477, 0.00005838, 0.00008457, 0.00008970, 0.00007316, 0.00003744, -0.00001234, -0.00006914, -0.00012509, -0.00017256, -0.00020516, -0.00021858, -0.00021104, -0.00018340, -0.00013887, -0.00008244, -0.00002014, 0.00004188, 0.00009809, 0.00014425, 0.00017779, 0.00019794, 0.00020560, 0.00020295, 0.00019300, 0.00017903, 0.00016410, 0.00015062, 0.00014017, 0.00013339, 0.00013004, 0.00012928, 0.00012986, 0.00013042, 0.00012976, 0.00012698, 0.00012160, 0.00011352, 0.00010303, 0.00009063, 0.00007695, 0.00006265, 0.00004833, 0.00003448, 0.00002146, },
		{ -0.00006042, -0.00005601, -0.00005140, -0.00004682, -0.00004245, -0.00003847, -0.00003495, -0.00003190, -0.00002921, -0.00002674, -0.00002428, -0.00002167, -0.00001879, -0.00001562, -0.00001227, -0.00000893, -0.00000587, -0.00000335, -0.00000158, -0.00000068, -0.00000060, -0.00000118, -0.00000215, -0.00000319, -0.00000402, -0.00000448, -0.00000454, -0.00000433, -0.00000413, -0.00000423, -0.00000492, -0.00000631, -0.00000828, -0.00001046, -0.00001226, -0.00001298, -0.00001196, -0.00000878, -0.00000341, 0.00000362, 0.00001126, 0.00001795, 0.00002190, 0.00002136, 0.00001499, 0.00000223, -0.00001647, -0.00003946, -0.00006397, -0.00008643, -0.00010287, -0.00010956, -0.00010368, -0.00008386, -0.00005071, -0.00000689, 0.00004293, 0.00009262, 0.00013528, 0.00016422, 0.00017398, 0.00016123, 0.00012547, 0.00006929, -0.00000169, -0.00007945, -0.00015452, -0.00021722, -0.00025903, -0.00027378, -0.00025869, -0.00021483, -0.00014718, -0.00006404, 0.00002396, 0.00010522, 0.00016876, 0.00020565, 0.00021026, 0.00018100, 0.00012062, 0.00003582, -0.00006359, -0.00016593, -0.00025917, -0.00033239, -0.00037715, -0.00038835, -0.00036479, -0.00030910, -0.00022722, -0.00012749, -0.00001951, 0.00008708, 0.00018370, 0.00026374, 0.00032310, 0.00036031, 0.00037637, 0.00037424, 0.00035817, 0.00033293, 0.00030311, 0.00027254, 0.00024396, 0.00021883, 0.00019748, 0.00017931, 0.00016314, 0.00014759, 0.00013142, 0.00011378, 0.00009431, 0.00007319, 0.00005103, 0.00002874, 0.00000731, -0.00001233, -0.00002945, -0.00004359, -0.00005458, -0.00006249, -0.00006759, -0.00007025, -0.00007089, -0.00006989, -0.00006762, -0.00006437, },
		{ -0.00001182, -0.00001032, -0.00000926, -0.00000850, -0.00000789, -0.00000730, -0.00000666, -0.00000596, -0.00000528, -0.00000472, -0.00000444, -0.00000454, -0.00000510, -0.00000606, -0.00000731, -0.00000863, -0.00000975, -0.00001045, -0.00001054, -0.00000999, -0.00000889, -0.00000747, -0.00000604, -0.00000494, -0.00000442, -0.00000463, -0.00000551, -0.00000684, -0.00000826, -0.00000938, -0.00000987, -0.00000958, -0.00000866, -0.00000751, -0.00000679, -0.00000730, -0.00000974, -0.00001454, -0.00002165, -0.00003040, -0.00003949, -0.00004709, -0.00005107, -0.00004940, -0.00004053, -0.00002385, 0.00000001, 0.00002901, 0.00005975, 0.00008782, 0.00010832, 0.00011658, 0.00010899, 0.00008367, 0.00004109, -0.00001567, -0.00008106, -0.00014753, -0.00020642, -0.00024911, -0.00026825, -0.00025893, -0.00021968, -0.00015294, -0.00006509, 0.00003410, 0.00013250, 0.00021714, 0.00027582, 0.00029883, 0.00028028, 0.00021914, 0.00011955, -0.00000947, -0.00015493, -0.00030138, -0.00043268, -0.00053392, -0.00059318, -0.00060291, -0.00056072, -0.00046966, -0.00033763, -0.00017645, -0.00000034, 0.00017576, 0.00033779, 0.00047409, 0.00057639, 0.00064041, 0.00066586, 0.00065604, 0.00061698, 0.00055643, 0.00048268, 0.00040363, 0.00032586, 0.00025423, 0.00019161, 0.00013900, 0.00009593, 0.00006089, 0.00003188, 0.00000695, -0.00001550, -0.00003642, -0.00005613, -0.00007437, -0.00009047, -0.00010364, -0.00011316, -0.00011856, -0.00011974, -0.00011699, -0.00011091, -0.00010234, -0.00009219, -0.00008134, -0.00007053, -0.00006029, -0.00005098, -0.00004274, -0.00003559, -0.00002949, -0.00002435, -0.00002008, -0.00001662, -0.00001389, },
		{ -0.00001289, -0.00001271, -0.00001236, -0.00001193, -0.00001151, -0.00001118, -0.00001099, -0.00001091, -0.00001088, -0.00001078, -0.00001048, -0.00000990, -0.00000901, -0.00000787, -0.00000665, -0.00000557, -0.00000489, -0.00000481, -0.00000546, -0.00000682, -0.00000872, -0.00001085, -0.00001283, -0.00001431, -0.00001498, -0.00001475, -0.00001370, -0.00001213, -0.00001047, -0.00000917, -0.00000860, -0.00000889, -0.00000985, -0.00001096, -0.00001144, -0.00001038, -0.00000697, -0.00000077, 0.00000808, 0.00001871, 0.00002944, 0.00003799, 0.00004171, 0.00003806, 0.00002509, 0.00000199, -0.00003053, -0.00006997, -0.00011215, -0.00015152, -0.00018179, -0.00019687, -0.00019178, -0.00016367, -0.00011256, -0.00004183, 0.00004182, 0.00012884, 0.00020779, 0.00026671, 0.00029471, 0.00028350, 0.00022883, 0.00013141, -0.00000268, -0.00016234, -0.00033239, -0.00049512, -0.00063227, -0.00072715, -0.00076661, -0.00074270, -0.00065370, -0.00050446, -0.00030589, -0.00007384, 0.00017266, 0.00041347, 0.00062948, 0.00080455, 0.00092710, 0.00099096, 0.00099570, 0.00094620, 0.00085169, 0.00072433, 0.00057768, 0.00042510, 0.00027837, 0.00014674, 0.00003629, -0.00005009, -0.00011247, -0.00015321, -0.00017617, -0.00018590, -0.00018683, -0.00018273, -0.00017637, -0.00016936, -0.00016231, -0.00015508, -0.00014709, -0.00013766, -0.00012630, -0.00011290, -0.00009774, -0.00008150, -0.00006508, -0.00004947, -0.00003555, -0.00002398, -0.00001510, -0.00000894, -0.00000523, -0.00000353, -0.00000329, -0.00000398, -0.00000516, -0.00000648, -0.00000776, -0.00000892, -0.00000992, -0.00001080, -0.00001155, -0.00001217, -0.00001262, -0.00001287, },
		{ -0.00000789, -0.00000747, -0.00000721, -0.00000705, -0.00000692, -0.00000676, -0.00000654, -0.00000629, -0.00000609, -0.00000604, -0.00000624, -0.00000676, -0.00000757, -0.00000858, -0.00000960, -0.00001036, -0.00001063, -0.00001020, -0.00000900, -0.00000712, -0.00000480, -0.00000244, -0.00000047, 0.00000069, 0.00000074, -0.00000040, -0.00000257, -0.00000540, -0.00000834, -0.00001088, -0.00001259, -0.00001337, -0.00001348, -0.00001359, -0.00001470, -0.00001793, -0.00002425, -0.00003420, -0.00004756, -0.00006321, -0.00007903, -0.00009207, -0.00009895, -0.00009632, -0.00008159, -0.00005355, -0.00001296, 0.00003711, 0.00009124, 0.00014205, 0.00018087, 0.00019886, 0.00018824, 0.00014361, 0.00006313, -0.00005070, -0.00019068, -0.00034524, -0.00049926, -0.00063564, -0.00073706, -0.00078810, -0.00077717, -0.00069817, -0.00055155, -0.00034464, -0.00009121, 0.00018983, 0.00047625, 0.00074479, 0.00097364, 0.00114466, 0.00124521, 0.00126935, 0.00121824, 0.00109969, 0.00092709, 0.00071766, 0.00049046, 0.00026433, 0.00005595, -0.00012160, -0.00025975, -0.00035481, -0.00040765, -0.00042290, -0.00040791, -0.00037138, -0.00032221, -0.00026841, -0.00021632, -0.00017026, -0.00013246, -0.00010332, -0.00008186, -0.00006631, -0.00005462, -0.00004498, -0.00003607, -0.00002723, -0.00001845, -0.00001014, -0.00000300, 0.00000228, 0.00000520, 0.00000559, 0.00000363, -0.00000018, -0.00000512, -0.00001040, -0.00001530, -0.00001926, -0.00002197, -0.00002335, -0.00002353, -0.00002280, -0.00002149, -0.00001991, -0.00001830, -0.00001682, -0.00001550, -0.00001432, -0.00001324, -0.00001220, -0.00001117, -0.00001018, -0.00000927, -0.00000849, },
		{ -0.00000652, -0.00000647, -0.00000635, -0.00000621, -0.00000610, -0.00000609, -0.00000617, -0.00000633, -0.00000650, -0.00000658, -0.00000649, -0.00000623, -0.00000584, -0.00000548, -0.00000538, -0.00000583, -0.00000707, -0.00000926, -0.00001242, -0.00001637, -0.00002074, -0.00002503, -0.00002867, -0.00003116, -0.00003218, -0.00003166, -0.00002982, -0.00002717, -0.00002438, -0.00002213, -0.00002091, -0.00002088, -0.00002171, -0.00002256, -0.00002224, -0.00001939, -0.00001286, -0.00000207, 0.00001255, 0.00002935, 0.00004532, 0.00005628, 0.00005726, 0.00004323, 0.00000990, -0.00004533, -0.00012252, -0.00021854, -0.00032686, -0.00043765, -0.00053859, -0.00061593, -0.00065608, -0.00064732, -0.00058147, -0.00045534, -0.00027167, -0.00003945, 0.00022656, 0.00050694, 0.00077947, 0.00102130, 0.00121140, 0.00133282, 0.00137456, 0.00133282, 0.00121140, 0.00102130, 0.00077947, 0.00050694, 0.00022656, -0.00003945, -0.00027167, -0.00045534, -0.00058147, -0.00064732, -0.00065608, -0.00061593, -0.00053859, -0.00043765, -0.00032686, -0.00021854, -0.00012252, -0.00004533, 0.00000990, 0.00004323, 0.00005726, 0.00005628, 0.00004532, 0.00002935, 0.00001255, -0.00000207, -0.00001286, -0.00001939, -0.00002224, -0.00002256, -0.00002171, -0.00002088, -0.00002091, -0.00002213, -0.00002438, -0.00002717, -0.00002982, -0.00003166, -0.00003218, -0.00003116, -0.00002867, -0.00002503, -0.00002074, -0.00001637, -0.00001242, -0.00000926, -0.00000707, -0.00000583, -0.00000538, -0.00000548, -0.00000584, -0.00000623, -0.00000649, -0.00000658, -0.00000650, -0.00000633, -0.00000617, -0.00000609, -0.00000610, -0.00000621, -0.00000635, -0.00000647, },
		{ -0.00000789, -0.00000849, -0.00000927, -0.00001018, -0.00001117, -0.00001220, -0.00001324, -0.00001432, -0.00001550, -0.00001682, -0.00001830, -0.00001991, -0.00002149, -0.00002280, -0.00002353, -0.00002335, -0.00002197, -0.00001926, -0.00001530, -0.00001040, -0.00000512, -0.00000018, 0.00000363, 0.00000559, 0.00000520, 0.00000228, -0.00000300, -0.00001014, -0.00001845, -0.00002723, -0.00003607, -0.00004498, -0.00005462, -0.00006631, -0.00008186, -0.00010332, -0.00013246, -0.00017026, -0.00021632, -0.00026841, -0.00032221, -0.00037138, -0.00040791, -0.00042290, -0.00040765, -0.00035481, -0.00025975, -0.00012160, 0.00005595, 0.00026433, 0.00049046, 0.00071766, 0.00092709, 0.00109969, 0.00121824, 0.00126935, 0.00124521, 0.00114466, 0.00097364, 0.00074479, 0.00047625, 0.00018983, -0.00009121, -0.00034464, -0.00055155, -0.00069817, -0.00077717, -0.00078810, -0.00073706, -0.00063564, -0.00049926, -0.00034524, -0.00019068, -0.00005070, 0.00006313, 0.00014361, 0.00018824, 0.00019886, 0.00018087, 0.00014205, 0.00009124, 0.00003711, -0.00001296, -0.00005355, -0.00008159, -0.00009632, -0.00009895, -0.00009207, -0.00007903, -0.00006321, -0.00004756, -0.00003420, -0.00002425, -0.00001793, -0.00001470, -0.00001359, -0.00001348, -0.00001337, -0.00001259, -0.00001088, -0.00000834, -0.00000540, -0.00000257, -0.00000040, 0.00000074, 0.00000069, -0.00000047, -0.00000244, -0.00000480, -0.00000712, -0.00000900, -0.00001020, -0.00001063, -0.00001036, -0.00000960, -0.00000858, -0.00000757, -0.00000676, -0.00000624, -0.00000604, -0.00000609, -0.00000629, -0.00000654, -0.00000676, -0.00000692, -0.00000705, -0.00000721, -0.00000747, },
		{ -0.00001289, -0.00001287, -0.00001262, -0.00001217, -0.00001155, -0.00001080, -0.00000992, -0.00000892, -0.00000776, -0.00000648, -0.00000516, -0.00000398, -0.00000329, -0.00000353, -0.00000523, -0.00000894, -0.00001510, -0.00002398, -0.00003555, -0.00004947, -0.00006508, -0.00008150, -0.00009774, -0.00011290, -0.00012630, -0.00013766, -0.00014709, -0.00015508, -0.00016231, -0.00016936, -0.00017637, -0.00018273, -0.00018683, -0.00018590, -0.00017617, -0.00015321, -0.00011247, -0.00005009, 0.00003629, 0.00014674, 0.00027837, 0.00042510, 0.00057768, 0.00072433, 0.00085169, 0.00094620, 0.00099570, 0.00099096, 0.00092710, 0.00080455, 0.00062948, 0.00041347, 0.00017266, -0.00007384, -0.00030589, -0.00050446, -0.00065370, -0.00074270, -0.00076661, -0.00072715, -0.00063227, -0.00049512, -0.00033239, -0.00016234, -0.00000268, 0.00013141, 0.00022883, 0.00028350, 0.00029471, 0.00026671, 0.00020779, 0.00012884, 0.00004182, -0.00004183, -0.00011256, -0.00016367, -0.00019178, -0.00019687, -0.00018179, -0.00015152, -0.00011215, -0.00006997, -0.00003053, 0.00000199, 0.00002509, 0.00003806, 0.00004171, 0.00003799, 0.00002944, 0.00001871, 0.00000808, -0.00000077, -0.00000697, -0.00001038, -0.00001144, -0.00001096, -0.00000985, -0.00000889, -0.00000860, -0.00000917, -0.00001047, -0.00001213, -0.00001370, -0.00001475, -0.00001498, -0.00001431, -0.00001283, -0.00001085, -0.00000872, -0.00000682, -0.00000546, -0.00000481, -0.00000489, -0.00000557, -0.00000665, -0.00000787, -0.00000901, -0.00000990, -0.00001048, -0.00001078, -0.00001088, -0.00001091, -0.00001099, -0.00001118, -0.00001151, -0.00001193, -0.00001236, -0.00001271, },
		{ -0.00001182, -0.00001389, -0.00001662, -0.00002008, -0.00002435, -0.00002949, -0.00003559, -0.00004274, -0.00005098, -0.00006029, -0.00007053, -0.00008134, -0.00009219, -0.00010234, -0.00011091, -0.00011699, -0.00011974, -0.00011856, -0.00011316, -0.00010364, -0.00009047, -0.00007437, -0.00005613, -0.00003642, -0.00001550, 0.00000695, 0.00003188, 0.00006089, 0.00009593, 0.00013900, 0.00019161, 0.00025423, 0.00032586, 0.00040363, 0.00048268, 0.00055643, 0.00061698, 0.00065604, 0.00066586, 0.00064041, 0.00057639, 0.00047409, 0.00033779, 0.00017576, -0.00000034, -0.00017645, -0.00033763, -0.00046966, -0.00056072, -0.00060291, -0.00059318, -0.00053392, -0.00043268, -0.00030138, -0.00015493, -0.00000947, 0.00011955, 0.00021914, 0.00028028, 0.00029883, 0.00027582, 0.00021714, 0.00013250, 0.00003410, -0.00006509, -0.00015294, -0.00021968, -0.00025893, -0.00026825, -0.00024911, -0.00020642, -0.00014753, -0.00008106, -0.00001567, 0.00004109, 0.00008367, 0.00010899, 0.00011658, 0.00010832, 0.00008782, 0.00005975, 0.00002901, 0.00000001, -0.00002385, -0.00004053, -0.00004940, -0.00005107, -0.00004709, -0.00003949, -0.00003040, -0.00002165, -0.00001454, -0.00000974, -0.00000730, -0.00000679, -0.00000751, -0.00000866, -0.00000958, -0.00000987, -0.00000938, -0.00000826, -0.00000684, -0.00000551, -0.00000463, -0.00000442, -0.00000494, -0.00000604, -0.00000747, -0.00000889, -0.00000999, -0.00001054, -0.00001045, -0.00000975, -0.00000863, -0.00000731, -0.00000606, -0.00000510, -0.00000454, -0.00000444, -0.00000472, -0.00000528, -0.00000596, -0.00000666, -0.00000730, -0.00000789, -0.00000850, -0.00000926, -0.00001032, },
		{ -0.00006042, -0.00006437, -0.00006762, -0.00006989, -0.00007089, -0.00007025, -0.00006759, -0.00006249, -0.00005458, -0.00004359, -0.00002945, -0.00001233, 0.00000731, 0.00002874, 0.00005103, 0.00007319, 0.00009431, 0.00011378, 0.00013142, 0.00014759, 0.00016314, 0.00017931, 0.00019748, 0.00021883, 0.00024396, 0.00027254, 0.00030311, 0.00033293, 0.00035817, 0.00037424, 0.00037637, 0.00036031, 0.00032310, 0.00026374, 0.00018370, 0.00008708, -0.00001951, -0.00012749, -0.00022722, -0.00030910, -0.00036479, -0.00038835, -0.00037715, -0.00033239, -0.00025917, -0.00016593, -0.00006359, 0.00003582, 0.00012062, 0.00018100, 0.00021026, 0.00020565, 0.00016876, 0.00010522, 0.00002396, -0.00006404, -0.00014718, -0.00021483, -0.00025869, -0.00027378, -0.00025903, -0.00021722, -0.00015452, -0.00007945, -0.00000169, 0.00006929, 0.00012547, 0.00016123, 0.00017398, 0.00016422, 0.00013528, 0.00009262, 0.00004293, -0.00000689, -0.00005071, -0.00008386, -0.00010368, -0.00010956, -0.00010287, -0.00008643, -0.00006397, -0.00003946, -0.00001647, 0.00000223, 0.00001499, 0.00002136, 0.00002190, 0.00001795, 0.00001126, 0.00000362, -0.00000341, -0.00000878, -0.00001196, -0.00001298, -0.00001226, -0.00001046, -0.00000828, -0.00000631, -0.00000492, -0.00000423, -0.00000413, -0.00000433, -0.00000454, -0.00000448, -0.00000402, -0.00000319, -0.00000215, -0.00000118, -0.00000060, -0.00000068, -0.00000158, -0.00000335, -0.00000587, -0.00000893, -0.00001227, -0.00001562, -0.00001879, -0.00002167, -0.00002428, -0.00002674, -0.00002921, -0.00003190, -0.00003495, -0.00003847, -0.00004245, -0.00004682, -0.00005140, -0.00005601, },
		{ 0.00000955, 0.00002146, 0.00003448, 0.00004833, 0.00006265, 0.00007695, 0.00009063, 0.00010303, 0.00011352, 0.00012160, 0.00012698, 0.00012976, 0.00013042, 0.00012986, 0.00012928, 0.00013004, 0.00013339, 0.00014017, 0.00015062, 0.00016410, 0.00017903, 0.00019300, 0.00020295, 0.00020560, 0.00019794, 0.00017779, 0.00014425, 0.00009809, 0.00004188, -0.00002014, -0.00008244, -0.00013887, -0.00018340, -0.00021104, -0.00021858, -0.00020516, -0.00017256, -0.00012509, -0.00006914, -0.00001234, 0.00003744, 0.00007316, 0.00008970, 0.00008457, 0.00005838, 0.00001477, -0.00004000, -0.00009795, -0.00015049, -0.00018964, -0.00020921, -0.00020568, -0.00017871, -0.00013126, -0.00006913, -0.00000015, 0.00006692, 0.00012363, 0.00016299, 0.00018043, 0.00017436, 0.00014634, 0.00010071, 0.00004396, -0.00001629, -0.00007235, -0.00011752, -0.00014694, -0.00015815, -0.00015124, -0.00012865, -0.00009469, -0.00005474, -0.00001444, 0.00002111, 0.00004803, 0.00006405, 0.00006866, 0.00006299, 0.00004943, 0.00003112, 0.00001142, -0.00000665, -0.00002080, -0.00002968, -0.00003298, -0.00003129, -0.00002583, -0.00001822, -0.00001004, -0.00000269, 0.00000291, 0.00000629, 0.00000747, 0.00000684, 0.00000496, 0.00000244, -0.00000021, -0.00000269, -0.00000492, -0.00000699, -0.00000910, -0.00001151, -0.00001436, -0.00001770, -0.00002141, -0.00002525, -0.00002891, -0.00003205, -0.00003444, -0.00003593, -0.00003656, -0.00003646, -0.00003592, -0.00003521, -0.00003460, -0.00003425, -0.00003422, -0.00003439, -0.00003454, -0.00003436, -0.00003351, -0.00003166, -0.00002855, -0.00002399, -0.00001790, -0.00001024, -0.00000106, },
		{ 0.00008187, 0.00008238, 0.00008127, 0.00007869, 0.00007500, 0.00007077, 0.00006681, 0.00006408, 0.00006361, 0.00006629, 0.00007273, 0.00008302, 0.00009663, 0.00011231, 0.00012811, 0.00014160, 0.00015013, 0.00015117, 0.00014278, 0.00012392, 0.00009479, 0.00005695, 0.00001322, -0.00003254, -0.00007591, -0.00011251, -0.00013856, -0.00015155, -0.00015058, -0.00013658, -0.00011227, -0.00008177, -0.00005006, -0.00002218, -0.00000250, 0.00000598, 0.00000215, -0.00001301, -0.00003652, -0.00006387, -0.00008971, -0.00010870, -0.00011640, -0.00011004, -0.00008901, -0.00005511, -0.00001227, 0.00003388, 0.00007694, 0.00011061, 0.00012969, 0.00013093, 0.00011349, 0.00007920, 0.00003227, -0.00002127, -0.00007442, -0.00012022, -0.00015274, -0.00016790, -0.00016402, -0.00014202, -0.00010520, -0.00005872, -0.00000881, 0.00003811, 0.00007639, 0.00010184, 0.00011226, 0.00010759, 0.00008983, 0.00006252, 0.00003021, -0.00000233, -0.00003076, -0.00005179, -0.00006350, -0.00006553, -0.00005890, -0.00004574, -0.00002886, -0.00001119, 0.00000459, 0.00001650, 0.00002338, 0.00002501, 0.00002195, 0.00001535, 0.00000665, -0.00000267, -0.00001138, -0.00001862, -0.00002396, -0.00002741, -0.00002925, -0.00002995, -0.00002997, -0.00002971, -0.00002935, -0.00002891, -0.00002827, -0.00002721, -0.00002555, -0.00002320, -0.00002019, -0.00001671, -0.00001305, -0.00000956, -0.00000655, -0.00000422, -0.00000262, -0.00000163, -0.00000094, -0.00000015, 0.00000116, 0.00000340, 0.00000684, 0.00001161, 0.00001767, 0.00002482, 0.00003275, 0.00004108, 0.00004942, 0.00005738, 0.00006464, 0.00007094, 0.00007602, 0.00007972, },
		{ 0.00003798, 0.00004150, 0.00004757, 0.00005634, 0.00006764, 0.00008086, 0.00009495, 0.00010842, 0.00011942, 0.00012597, 0.00012620, 0.00011865, 0.00010254, 0.00007803, 0.00004630, 0.00000953, -0.00002927, -0.00006664, -0.00009903, -0.00012333, -0.00013733, -0.00014000, -0.00013175, -0.00011435, -0.00009069, -0.00006441, -0.00003928, -0.00001871, -0.00000515, 0.00000025, -0.00000218, -0.00001073, -0.00002261, -0.00003446, -0.00004296, -0.00004545, -0.00004046, -0.00002798, -0.00000959, 0.00001183, 0.00003251, 0.00004844, 0.00005609, 0.00005305, 0.00003854, 0.00001362, -0.00001881, -0.00005446, -0.00008819, -0.00011483, -0.00012998, -0.00013070, -0.00011603, -0.00008720, -0.00004750, -0.00000181, 0.00004405, 0.00008419, 0.00011350, 0.00012835, 0.00012713, 0.00011041, 0.00008087, 0.00004279, 0.00000145, -0.00003767, -0.00006967, -0.00009092, -0.00009946, -0.00009522, -0.00007987, -0.00005651, -0.00002908, -0.00000175, 0.00002171, 0.00003845, 0.00004682, 0.00004655, 0.00003862, 0.00002498, 0.00000817, -0.00000915, -0.00002459, -0.00003636, -0.00004345, -0.00004566, -0.00004351, -0.00003803, -0.00003052, -0.00002228, -0.00001439, -0.00000760, -0.00000226, 0.00000161, 0.00000429, 0.00000613, 0.00000752, 0.00000871, 0.00000981, 0.00001077, 0.00001143, 0.00001161, 0.00001119, 0.00001016, 0.00000872, 0.00000717, 0.00000596, 0.00000555, 0.00000634, 0.00000860, 0.00001237, 0.00001749, 0.00002358, 0.00003015, 0.00003662, 0.00004243, 0.00004715, 0.00005049, 0.00005231, 0.00005268, 0.00005177, 0.00004984, 0.00004723, 0.00004427, 0.00004134, 0.00003881, 0.00003710, 0.00003666, },
		{ 0.00008599, 0.00009153, 0.00009352, 0.00009081, 0.00008252, 0.00006824, 0.00004813, 0.00002311, -0.00000523, -0.00003467, -0.00006263, -0.00008646, -0.00010383, -0.00011301, -0.00011321, -0.00010469, -0.00008881, -0.00006782, -0.00004459, -0.00002220, -0.00000346, 0.00000949, 0.00001548, 0.00001448, 0.00000757, -0.00000328, -0.00001559, -0.00002678, -0.00003468, -0.00003792, -0.00003611, -0.00003000, -0.00002121, -0.00001202, -0.00000483, -0.00000171, -0.00000394, -0.00001170, -0.00002398, -0.00003865, -0.00005280, -0.00006325, -0.00006710, -0.00006230, -0.00004808, -0.00002521, 0.00000402, 0.00003606, 0.00006662, 0.00009125, 0.00010612, 0.00010856, 0.00009759, 0.00007409, 0.00004075, 0.00000169, -0.00003811, -0.00007352, -0.00010001, -0.00011432, -0.00011489, -0.00010210, -0.00007812, -0.00004663, -0.00001217, 0.00002049, 0.00004709, 0.00006441, 0.00007075, 0.00006605, 0.00005180, 0.00003074, 0.00000638, -0.00001759, -0.00003782, -0.00005181, -0.00005814, -0.00005663, -0.00004822, -0.00003474, -0.00001848, -0.00000188, 0.00001291, 0.00002429, 0.00003138, 0.00003403, 0.00003275, 0.00002848, 0.00002239, 0.00001565, 0.00000921, 0.00000372, -0.00000051, -0.00000350, -0.00000543, -0.00000659, -0.00000720, -0.00000737, -0.00000705, -0.00000607, -0.00000415, -0.00000107, 0.00000328, 0.00000885, 0.00001535, 0.00002231, 0.00002912, 0.00003514, 0.00003979, 0.00004266, 0.00004356, 0.00004258, 0.00004004, 0.00003642, 0.00003235, 0.00002842, 0.00002518, 0.00002307, 0.00002236, 0.00002321, 0.00002568, 0.00002973, 0.00003530, 0.00004227, 0.00005044, 0.00005950, 0.00006895, 0.00007809, },
		{ 0.00001270, -0.00000658, -0.00002655, -0.00004557, -0.00006186, -0.00007382, -0.00008020, -0.00008038, -0.00007449, -0.00006344, -0.00004881, -0.00003270, -0.00001736, -0.00000493, 0.00000296, 0.00000539, 0.00000231, -0.00000542, -0.00001626, -0.00002818, -0.00003904, -0.00004697, -0.00005064, -0.00004954, -0.00004398, -0.00003508, -0.00002447, -0.00001399, -0.00000534, 0.00000026, 0.00000234, 0.00000125, -0.00000190, -0.00000548, -0.00000765, -0.00000681, -0.00000192, 0.00000718, 0.00001967, 0.00003380, 0.00004714, 0.00005700, 0.00006085, 0.00005691, 0.00004444, 0.00002402, -0.00000240, -0.00003177, -0.00006032, -0.00008410, -0.00009963, -0.00010442, -0.00009740, -0.00007917, -0.00005192, -0.00001914, 0.00001485, 0.00004556, 0.00006897, 0.00008213, 0.00008359, 0.00007357, 0.00005393, 0.00002782, -0.00000079, -0.00002773, -0.00004925, -0.00006258, -0.00006624, -0.00006024, -0.00004599, -0.00002600, -0.00000346, 0.00001829, 0.00003629, 0.00004829, 0.00005312, 0.00005070, 0.00004198, 0.00002870, 0.00001305, -0.00000273, -0.00001664, -0.00002723, -0.00003370, -0.00003594, -0.00003441, -0.00002997, -0.00002366, -0.00001648, -0.00000923, -0.00000241, 0.00000377, 0.00000936, 0.00001454, 0.00001950, 0.00002436, 0.00002903, 0.00003329, 0.00003672, 0.00003890, 0.00003944, 0.00003810, 0.00003489, 0.00003011, 0.00002429, 0.00001814, 0.00001245, 0.00000796, 0.00000525, 0.00000468, 0.00000633, 0.00001004, 0.00001543, 0.00002204, 0.00002934, 0.00003685, 0.00004416, 0.00005094, 0.00005689, 0.00006173, 0.00006511, 0.00006660, 0.00006572, 0.00006194, 0.00005481, 0.00004411, 0.00002991, },
		{ -0.00006281, -0.00006112, -0.00005573, -0.00004750, -0.00003771, -0.00002784, -0.00001937, -0.00001352, -0.00001103, -0.00001204, -0.00001605, -0.00002200, -0.00002847, -0.00003390, -0.00003690, -0.00003649, -0.00003226, -0.00002451, -0.00001412, -0.00000248, 0.00000885, 0.00001836, 0.00002487, 0.00002777, 0.00002709, 0.00002349, 0.00001811, 0.00001231, 0.00000740, 0.00000433, 0.00000353, 0.00000475, 0.00000713, 0.00000936, 0.00000995, 0.00000759, 0.00000143, -0.00000863, -0.00002183, -0.00003661, -0.00005078, -0.00006190, -0.00006764, -0.00006629, -0.00005706, -0.00004032, -0.00001766, 0.00000830, 0.00003424, 0.00005665, 0.00007234, 0.00007899, 0.00007547, 0.00006216, 0.00004082, 0.00001441, -0.00001333, -0.00003851, -0.00005758, -0.00006797, -0.00006839, -0.00005905, -0.00004161, -0.00001888, 0.00000559, 0.00002812, 0.00004541, 0.00005503, 0.00005579, 0.00004783, 0.00003258, 0.00001244, -0.00000957, -0.00003035, -0.00004713, -0.00005790, -0.00006161, -0.00005825, -0.00004879, -0.00003489, -0.00001862, -0.00000208, 0.00001288, 0.00002495, 0.00003341, 0.00003819, 0.00003973, 0.00003879, 0.00003627, 0.00003299, 0.00002954, 0.00002624, 0.00002310, 0.00001993, 0.00001647, 0.00001250, 0.00000797, 0.00000306, -0.00000182, -0.00000608, -0.00000905, -0.00001013, -0.00000890, -0.00000523, 0.00000070, 0.00000842, 0.00001722, 0.00002628, 0.00003479, 0.00004209, 0.00004771, 0.00005144, 0.00005328, 0.00005341, 0.00005205, 0.00004941, 0.00004558, 0.00004055, 0.00003417, 0.00002627, 0.00001673, 0.00000559, -0.00000684, -0.00001998, -0.00003295, -0.00004473, -0.00005421, -0.00006044, },
		{ -0.00001552, -0.00001041, -0.00000743, -0.00000655, -0.00000731, -0.00000894, -0.00001046, -0.00001094, -0.00000967, -0.00000632, -0.00000100, 0.00000567, 0.00001277, 0.00001915, 0.00002371, 0.00002559, 0.00002432, 0.00001999, 0.00001315, 0.00000479, -0.00000389, -0.00001167, -0.00001756, -0.00002098, -0.00002187, -0.00002063, -0.00001812, -0.00001537, -0.00001340, -0.00001294, -0.00001430, -0.00001719, -0.00002083, -0.00002404, -0.00002546, -0.00002388, -0.00001852, -0.00000924, 0.00000332, 0.00001779, 0.00003220, 0.00004430, 0.00005193, 0.00005342, 0.00004789, 0.00003551, 0.00001754, -0.00000382, -0.00002572, -0.00004505, -0.00005899, -0.00006543, -0.00006332, -0.00005291, -0.00003570, -0.00001427, 0.00000813, 0.00002806, 0.00004246, 0.00004909, 0.00004691, 0.00003623, 0.00001866, -0.00000313, -0.00002586, -0.00004611, -0.00006089, -0.00006802, -0.00006651, -0.00005660, -0.00003976, -0.00001833, 0.00000479, 0.00002662, 0.00004455, 0.00005671, 0.00006212, 0.00006084, 0.00005380, 0.00004258, 0.00002913, 0.00001539, 0.00000302, -0.00000681, -0.00001355, -0.00001722, -0.00001832, -0.00001758, -0.00001584, -0.00001375, -0.00001173, -0.00000986, -0.00000791, -0.00000548, -0.00000212, 0.00000253, 0.00000858, 0.00001586, 0.00002387, 0.00003188, 0.00003907, 0.00004463, 0.00004795, 0.00004872, 0.00004695, 0.00004300, 0.00003746, 0.00003104, 0.00002445, 0.00001821, 0.00001259, 0.00000761, 0.00000302, -0.00000159, -0.00000664, -0.00001246, -0.00001912, -0.00002644, -0.00003393, -0.00004087, -0.00004646, -0.00004995, -0.00005081, -0.00004887, -0.00004431, -0.00003775, -0.00003008, -0.00002234, },
		{ 0.00000426, 0.00000294, 0.00000237, 0.00000275, 0.00000394, 0.00000552, 0.00000689, 0.00000738, 0.00000644, 0.00000376, -0.00000063, -0.00000631, -0.00001258, -0.00001855, -0.00002328, -0.00002603, -0.00002634, -0.00002416, -0.00001987, -0.00001418, -0.00000804, -0.00000242, 0.00000188, 0.00000436, 0.00000495, 0.00000401, 0.00000222, 0.00000047, -0.00000038, 0.00000030, 0.00000273, 0.00000665, 0.00001133, 0.00001565, 0.00001837, 0.00001831, 0.00001469, 0.00000729, -0.00000341, -0.00001623, -0.00002946, -0.00004109, -0.00004915, -0.00005206, -0.00004896, -0.00003990, -0.00002589, -0.00000881, 0.00000886, 0.00002441, 0.00003536, 0.00003985, 0.00003699, 0.00002704, 0.00001141, -0.00000753, -0.00002682, -0.00004334, -0.00005433, -0.00005781, -0.00005296, -0.00004016, -0.00002107, 0.00000174, 0.00002510, 0.00004577, 0.00006090, 0.00006849, 0.00006764, 0.00005867, 0.00004302, 0.00002300, 0.00000141, -0.00001892, -0.00003550, -0.00004658, -0.00005132, -0.00004981, -0.00004301, -0.00003249, -0.00002008, -0.00000760, 0.00000345, 0.00001211, 0.00001803, 0.00002141, 0.00002290, 0.00002333, 0.00002357, 0.00002424, 0.00002564, 0.00002766, 0.00002987, 0.00003161, 0.00003217, 0.00003098, 0.00002775, 0.00002256, 0.00001585, 0.00000836, 0.00000101, -0.00000534, -0.00001000, -0.00001260, -0.00001319, -0.00001219, -0.00001032, -0.00000845, -0.00000742, -0.00000785, -0.00001005, -0.00001391, -0.00001893, -0.00002432, -0.00002916, -0.00003255, -0.00003379, -0.00003254, -0.00002886, -0.00002321, -0.00001636, -0.00000922, -0.00000271, 0.00000243, 0.00000576, 0.00000721, 0.00000706, 0.00000585, },
		{ -0.00000813, -0.00000835, -0.00000897, -0.00001017, -0.00001189, -0.00001383, -0.00001554, -0.00001647, -0.00001618, -0.00001437, -0.00001106, -0.00000652, -0.00000132, 0.00000380, 0.00000806, 0.00001078, 0.00001155, 0.00001026, 0.00000717, 0.00000286, -0.00000191, -0.00000632, -0.00000969, -0.00001160, -0.00001197, -0.00001111, -0.00000962, -0.00000829, -0.00000786, -0.00000890, -0.00001163, -0.00001581, -0.00002078, -0.00002552, -0.00002889, -0.00002979, -0.00002745, -0.00002162, -0.00001268, -0.00000162, 0.00001003, 0.00002048, 0.00002796, 0.00003102, 0.00002888, 0.00002153, 0.00000990, -0.00000434, -0.00001895, -0.00003150, -0.00003973, -0.00004199, -0.00003753, -0.00002667, -0.00001076, 0.00000792, 0.00002657, 0.00004226, 0.00005242, 0.00005525, 0.00005000, 0.00003716, 0.00001836, -0.00000387, -0.00002643, -0.00004620, -0.00006047, -0.00006732, -0.00006596, -0.00005674, -0.00004112, -0.00002140, -0.00000030, 0.00001942, 0.00003541, 0.00004603, 0.00005058, 0.00004929, 0.00004320, 0.00003394, 0.00002339, 0.00001333, 0.00000514, -0.00000034, -0.00000296, -0.00000315, -0.00000177, 0.00000009, 0.00000144, 0.00000152, 0.00000003, -0.00000289, -0.00000665, -0.00001041, -0.00001325, -0.00001438, -0.00001335, -0.00001014, -0.00000521, 0.00000061, 0.00000625, 0.00001060, 0.00001274, 0.00001212, 0.00000866, 0.00000278, -0.00000464, -0.00001248, -0.00001948, -0.00002457, -0.00002695, -0.00002631, -0.00002280, -0.00001705, -0.00001000, -0.00000275, 0.00000365, 0.00000838, 0.00001094, 0.00001124, 0.00000958, 0.00000652, 0.00000278, -0.00000092, -0.00000403, -0.00000623, -0.00000748, -0.00000799, },
		{ -0.00000695, -0.00000637, -0.00000540, -0.00000393, -0.00000201, 0.00000011, 0.00000202, 0.00000325, 0.00000339, 0.00000216, -0.00000050, -0.00000435, -0.00000896, -0.00001368, -0.00001783, -0.00002080, -0.00002217, -0.00002179, -0.00001984, -0.00001675, -0.00001314, -0.00000970, -0.00000701, -0.00000543, -0.00000501, -0.00000551, -0.00000639, -0.00000698, -0.00000661, -0.00000480, -0.00000138, 0.00000342, 0.00000897, 0.00001432, 0.00001839, 0.00002015, 0.00001888, 0.00001430, 0.00000674, -0.00000289, -0.00001321, -0.00002257, -0.00002930, -0.00003208, -0.00003015, -0.00002351, -0.00001299, -0.00000016, 0.00001293, 0.00002398, 0.00003094, 0.00003227, 0.00002731, 0.00001640, 0.00000089, -0.00001703, -0.00003465, -0.00004916, -0.00005812, -0.00005981, -0.00005360, -0.00004001, -0.00002069, 0.00000182, 0.00002448, 0.00004421, 0.00005837, 0.00006511, 0.00006371, 0.00005461, 0.00003934, 0.00002023, 0.00000006, -0.00001841, -0.00003284, -0.00004167, -0.00004427, -0.00004101, -0.00003311, -0.00002236, -0.00001079, -0.00000029, 0.00000766, 0.00001227, 0.00001345, 0.00001181, 0.00000843, 0.00000463, 0.00000167, 0.00000048, 0.00000150, 0.00000458, 0.00000905, 0.00001384, 0.00001775, 0.00001965, 0.00001876, 0.00001480, 0.00000807, -0.00000064, -0.00001010, -0.00001893, -0.00002582, -0.00002974, -0.00003016, -0.00002708, -0.00002107, -0.00001315, -0.00000456, 0.00000338, 0.00000953, 0.00001313, 0.00001386, 0.00001187, 0.00000773, 0.00000228, -0.00000351, -0.00000876, -0.00001281, -0.00001526, -0.00001610, -0.00001557, -0.00001412, -0.00001225, -0.00001043, -0.00000897, -0.00000798, -0.00000739, },
		{ -0.00000561, -0.00000598, -0.00000679, -0.00000817, -0.00001007, -0.00001225, -0.00001436, -0.00001595, -0.00001661, -0.00001605, -0.00001414, -0.00001101, -0.00000700, -0.00000262, 0.00000153, 0.00000493, 0.00000717, 0.00000810, 0.00000778, 0.00000655, 0.00000490, 0.00000334, 0.00000234, 0.00000215, 0.00000278, 0.00000395, 0.00000515, 0.00000575, 0.00000515, 0.00000291, -0.00000108, -0.00000657, -0.00001292, -0.00001921, -0.00002434, -0.00002731, -0.00002738, -0.00002423, -0.00001812, -0.00000987, -0.00000077, 0.00000765, 0.00001382, 0.00001647, 0.00001486, 0.00000899, -0.00000038, -0.00001178, -0.00002326, -0.00003268, -0.00003806, -0.00003798, -0.00003185, -0.00002005, -0.00000394, 0.00001431, 0.00003202, 0.00004642, 0.00005510, 0.00005641, 0.00004974, 0.00003567, 0.00001590, -0.00000697, -0.00002983, -0.00004955, -0.00006341, -0.00006954, -0.00006721, -0.00005689, -0.00004019, -0.00001955, 0.00000209, 0.00002182, 0.00003717, 0.00004647, 0.00004905, 0.00004532, 0.00003661, 0.00002488, 0.00001238, 0.00000127, -0.00000680, -0.00001089, -0.00001086, -0.00000739, -0.00000177, 0.00000440, 0.00000950, 0.00001221, 0.00001176, 0.00000805, 0.00000164, -0.00000635, -0.00001450, -0.00002135, -0.00002565, -0.00002664, -0.00002411, -0.00001850, -0.00001076, -0.00000220, 0.00000577, 0.00001188, 0.00001519, 0.00001527, 0.00001224, 0.00000669, -0.00000041, -0.00000790, -0.00001466, -0.00001979, -0.00002272, -0.00002330, -0.00002175, -0.00001863, -0.00001466, -0.00001059, -0.00000704, -0.00000445, -0.00000298, -0.00000254, -0.00000286, -0.00000359, -0.00000439, -0.00000500, -0.00000534, -0.00000548, },
		{ -0.00000429, -0.00000400, -0.00000327, -0.00000196, -0.00000007, 0.00000225, 0.00000468, 0.00000686, 0.00000837, 0.00000893, 0.00000836, 0.00000670, 0.00000418, 0.00000117, -0.00000188, -0.00000454, -0.00000650, -0.00000764, -0.00000803, -0.00000795, -0.00000779, -0.00000799, -0.00000888, -0.00001062, -0.00001312, -0.00001604, -0.00001881, -0.00002077, -0.00002130, -0.00001995, -0.00001659, -0.00001146, -0.00000516, 0.00000137, 0.00000709, 0.00001095, 0.00001219, 0.00001046, 0.00000592, -0.00000069, -0.00000818, -0.00001512, -0.00002001, -0.00002165, -0.00001934, -0.00001307, -0.00000359, 0.00000768, 0.00001882, 0.00002772, 0.00003245, 0.00003163, 0.00002469, 0.00001202, -0.00000498, -0.00002408, -0.00004252, -0.00005746, -0.00006639, -0.00006754, -0.00006023, -0.00004499, -0.00002355, 0.00000145, 0.00002675, 0.00004902, 0.00006535, 0.00007364, 0.00007298, 0.00006370, 0.00004738, 0.00002653, 0.00000423, -0.00001640, -0.00003259, -0.00004241, -0.00004499, -0.00004058, -0.00003052, -0.00001689, -0.00000219, 0.00001111, 0.00002094, 0.00002598, 0.00002579, 0.00002086, 0.00001241, 0.00000220, -0.00000787, -0.00001607, -0.00002113, -0.00002246, -0.00002016, -0.00001502, -0.00000830, -0.00000147, 0.00000405, 0.00000716, 0.00000724, 0.00000428, -0.00000118, -0.00000818, -0.00001551, -0.00002195, -0.00002648, -0.00002846, -0.00002769, -0.00002445, -0.00001941, -0.00001345, -0.00000752, -0.00000249, 0.00000105, 0.00000281, 0.00000285, 0.00000151, -0.00000069, -0.00000315, -0.00000534, -0.00000690, -0.00000766, -0.00000766, -0.00000709, -0.00000625, -0.00000541, -0.00000480, -0.00000447, -0.00000436, },
		{ -0.00000060, -0.00000090, -0.00000151, -0.00000262, -0.00000426, -0.00000634, -0.00000865, -0.00001090, -0.00001277, -0.00001399, -0.00001442, -0.00001406, -0.00001309, -0.00001177, -0.00001046, -0.00000946, -0.00000899, -0.00000908, -0.00000961, -0.00001027, -0.00001065, -0.00001036, -0.00000908, -0.00000671, -0.00000337, 0.00000053, 0.00000441, 0.00000755, 0.00000929, 0.00000914, 0.00000691, 0.00000279, -0.00000265, -0.00000852, -0.00001378, -0.00001743, -0.00001867, -0.00001711, -0.00001291, -0.00000674, 0.00000027, 0.00000668, 0.00001105, 0.00001216, 0.00000932, 0.00000249, -0.00000758, -0.00001942, -0.00003108, -0.00004038, -0.00004529, -0.00004428, -0.00003666, -0.00002274, -0.00000388, 0.00001769, 0.00003912, 0.00005736, 0.00006968, 0.00007402, 0.00006941, 0.00005613, 0.00003573, 0.00001081, -0.00001535, -0.00003922, -0.00005760, -0.00006806, -0.00006929, -0.00006132, -0.00004549, -0.00002420, -0.00000060, 0.00002197, 0.00004041, 0.00005236, 0.00005649, 0.00005265, 0.00004186, 0.00002607, 0.00000782, -0.00001021, -0.00002556, -0.00003643, -0.00004184, -0.00004178, -0.00003706, -0.00002917, -0.00001993, -0.00001117, -0.00000443, -0.00000067, -0.00000023, -0.00000276, -0.00000736, -0.00001283, -0.00001784, -0.00002127, -0.00002234, -0.00002077, -0.00001678, -0.00001104, -0.00000449, 0.00000183, 0.00000697, 0.00001026, 0.00001137, 0.00001039, 0.00000776, 0.00000416, 0.00000035, -0.00000293, -0.00000515, -0.00000603, -0.00000557, -0.00000403, -0.00000183, 0.00000051, 0.00000253, 0.00000392, 0.00000450, 0.00000432, 0.00000356, 0.00000249, 0.00000140, 0.00000051, -0.00000008, -0.00000040, },
		{ -0.00001126, -0.00001103, -0.00001058, -0.00000976, -0.00000852, -0.00000694, -0.00000518, -0.00000346, -0.00000205, -0.00000114, -0.00000083, -0.00000109, -0.00000177, -0.00000259, -0.00000328, -0.00000356, -0.00000328, -0.00000246, -0.00000126, -0.00000004, 0.00000079, 0.00000081, -0.00000027, -0.00000258, -0.00000596, -0.00000999, -0.00001405, -0.00001742, -0.00001939, -0.00001944, -0.00001735, -0.00001327, -0.00000778, -0.00000175, 0.00000375, 0.00000768, 0.00000919, 0.00000785, 0.00000376, -0.00000243, -0.00000955, -0.00001614, -0.00002064, -0.00002171, -0.00001849, -0.00001083, 0.00000063, 0.00001450, 0.00002877, 0.00004114, 0.00004937, 0.00005165, 0.00004698, 0.00003533, 0.00001780, -0.00000355, -0.00002593, -0.00004620, -0.00006138, -0.00006906, -0.00006786, -0.00005760, -0.00003943, -0.00001566, 0.00001055, 0.00003565, 0.00005619, 0.00006934, 0.00007331, 0.00006759, 0.00005301, 0.00003160, 0.00000627, -0.00001963, -0.00004282, -0.00006054, -0.00007093, -0.00007325, -0.00006793, -0.00005642, -0.00004087, -0.00002384, -0.00000779, 0.00000523, 0.00001388, 0.00001767, 0.00001694, 0.00001273, 0.00000652, -0.00000005, -0.00000550, -0.00000874, -0.00000924, -0.00000706, -0.00000280, 0.00000260, 0.00000802, 0.00001241, 0.00001500, 0.00001540, 0.00001365, 0.00001019, 0.00000577, 0.00000123, -0.00000261, -0.00000514, -0.00000602, -0.00000530, -0.00000330, -0.00000057, 0.00000220, 0.00000440, 0.00000554, 0.00000536, 0.00000385, 0.00000127, -0.00000197, -0.00000540, -0.00000855, -0.00001105, -0.00001272, -0.00001354, -0.00001364, -0.00001326, -0.00001267, -0.00001209, -0.00001166, -0.00001141, },
		{ 0.00000202, 0.00000194, 0.00000153, 0.00000071, -0.00000055, -0.00000216, -0.00000397, -0.00000575, -0.00000727, -0.00000835, -0.00000892, -0.00000899, -0.00000873, -0.00000837, -0.00000817, -0.00000836, -0.00000905, -0.00001018, -0.00001153, -0.00001273, -0.00001334, -0.00001293, -0.00001119, -0.00000804, -0.00000366, 0.00000149, 0.00000673, 0.00001123, 0.00001422, 0.00001508, 0.00001353, 0.00000969, 0.00000411, -0.00000228, -0.00000833, -0.00001283, -0.00001480, -0.00001361, -0.00000919, -0.00000208, 0.00000661, 0.00001536, 0.00002248, 0.00002640, 0.00002597, 0.00002068, 0.00001085, -0.00000239, -0.00001721, -0.00003133, -0.00004239, -0.00004832, -0.00004770, -0.00004007, -0.00002603, -0.00000723, 0.00001382, 0.00003414, 0.00005066, 0.00006071, 0.00006248, 0.00005526, 0.00003961, 0.00001732, -0.00000887, -0.00003562, -0.00005948, -0.00007741, -0.00008716, -0.00008765, -0.00007904, -0.00006270, -0.00004098, -0.00001684, 0.00000663, 0.00002660, 0.00004097, 0.00004855, 0.00004922, 0.00004383, 0.00003402, 0.00002189, 0.00000963, -0.00000084, -0.00000810, -0.00001147, -0.00001100, -0.00000740, -0.00000185, 0.00000428, 0.00000965, 0.00001323, 0.00001443, 0.00001318, 0.00000986, 0.00000526, 0.00000029, -0.00000412, -0.00000724, -0.00000868, -0.00000841, -0.00000676, -0.00000433, -0.00000186, -0.00000004, 0.00000056, -0.00000034, -0.00000270, -0.00000623, -0.00001037, -0.00001450, -0.00001798, -0.00002031, -0.00002117, -0.00002053, -0.00001856, -0.00001562, -0.00001216, -0.00000865, -0.00000548, -0.00000291, -0.00000104, 0.00000019, 0.00000090, 0.00000129, 0.00000153, 0.00000173, 0.00000191, },
		{ -0.00001080, -0.00001171, -0.00001224, -0.00001229, -0.00001181, -0.00001087, -0.00000959, -0.00000817, -0.00000678, -0.00000559, -0.00000465, -0.00000390, -0.00000321, -0.00000235, -0.00000109, 0.00000074, 0.00000317, 0.00000608, 0.00000913, 0.00001186, 0.00001373, 0.00001422, 0.00001296, 0.00000981, 0.00000495, -0.00000112, -0.00000759, -0.00001354, -0.00001797, -0.00002008, -0.00001939, -0.00001583, -0.00000985, -0.00000234, 0.00000548, 0.00001226, 0.00001674, 0.00001798, 0.00001556, 0.00000970, 0.00000123, -0.00000849, -0.00001777, -0.00002491, -0.00002846, -0.00002753, -0.00002195, -0.00001238, -0.00000023, 0.00001250, 0.00002358, 0.00003086, 0.00003265, 0.00002807, 0.00001717, 0.00000107, -0.00001827, -0.00003821, -0.00005589, -0.00006862, -0.00007431, -0.00007179, -0.00006102, -0.00004313, -0.00002022, 0.00000489, 0.00002911, 0.00004949, 0.00006368, 0.00007022, 0.00006874, 0.00005999, 0.00004564, 0.00002804, 0.00000982, -0.00000651, -0.00001895, -0.00002622, -0.00002793, -0.00002459, -0.00001741, -0.00000807, 0.00000155, 0.00000978, 0.00001532, 0.00001747, 0.00001618, 0.00001199, 0.00000589, -0.00000094, -0.00000731, -0.00001224, -0.00001516, -0.00001592, -0.00001483, -0.00001250, -0.00000975, -0.00000739, -0.00000606, -0.00000615, -0.00000766, -0.00001031, -0.00001353, -0.00001663, -0.00001890, -0.00001981, -0.00001904, -0.00001657, -0.00001265, -0.00000779, -0.00000260, 0.00000227, 0.00000626, 0.00000901, 0.00001033, 0.00001029, 0.00000912, 0.00000717, 0.00000481, 0.00000240, 0.00000016, -0.00000176, -0.00000336, -0.00000473, -0.00000597, -0.00000718, -0.00000842, -0.00000966, },
		{ -0.00000228, -0.00000074, 0.00000069, 0.00000190, 0.00000284, 0.00000352, 0.00000404, 0.00000452, 0.00000506, 0.00000574, 0.00000650, 0.00000720, 0.00000760, 0.00000739, 0.00000627, 0.00000406, 0.00000070, -0.00000362, -0.00000849, -0.00001330, -0.00001735, -0.00001991, -0.00002039, -0.00001848, -0.00001419, -0.00000795, -0.00000053, 0.00000702, 0.00001354, 0.00001796, 0.00001946, 0.00001764, 0.00001262, 0.00000505, -0.00000399, -0.00001313, -0.00002096, -0.00002625, -0.00002818, -0.00002649, -0.00002155, -0.00001431, -0.00000618, 0.00000126, 0.00000649, 0.00000835, 0.00000627, 0.00000042, -0.00000832, -0.00001846, -0.00002809, -0.00003524, -0.00003822, -0.00003587, -0.00002784, -0.00001467, 0.00000225, 0.00002085, 0.00003871, 0.00005339, 0.00006283, 0.00006566, 0.00006143, 0.00005069, 0.00003488, 0.00001617, -0.00000293, -0.00001990, -0.00003260, -0.00003961, -0.00004036, -0.00003523, -0.00002546, -0.00001287, 0.00000039, 0.00001220, 0.00002083, 0.00002512, 0.00002470, 0.00001992, 0.00001181, 0.00000182, -0.00000842, -0.00001739, -0.00002391, -0.00002732, -0.00002752, -0.00002494, -0.00002045, -0.00001510, -0.00000998, -0.00000598, -0.00000365, -0.00000313, -0.00000419, -0.00000624, -0.00000854, -0.00001032, -0.00001095, -0.00001004, -0.00000756, -0.00000376, 0.00000082, 0.00000550, 0.00000957, 0.00001242, 0.00001367, 0.00001316, 0.00001104, 0.00000767, 0.00000356, -0.00000071, -0.00000463, -0.00000781, -0.00001002, -0.00001125, -0.00001161, -0.00001134, -0.00001070, -0.00000995, -0.00000925, -0.00000866, -0.00000817, -0.00000768, -0.00000708, -0.00000625, -0.00000515, -0.00000380, },
		{ 0.00000289, 0.00000294, 0.00000294, 0.00000292, 0.00000284, 0.00000259, 0.00000203, 0.00000100, -0.00000063, -0.00000288, -0.00000565, -0.00000868, -0.00001158, -0.00001386, -0.00001505, -0.00001475, -0.00001276, -0.00000913, -0.00000419, 0.00000142, 0.00000689, 0.00001132, 0.00001385, 0.00001388, 0.00001110, 0.00000565, -0.00000189, -0.00001058, -0.00001925, -0.00002666, -0.00003168, -0.00003355, -0.00003195, -0.00002707, -0.00001964, -0.00001077, -0.00000176, 0.00000606, 0.00001162, 0.00001428, 0.00001391, 0.00001099, 0.00000642, 0.00000147, -0.00000253, -0.00000442, -0.00000343, 0.00000066, 0.00000745, 0.00001594, 0.00002475, 0.00003226, 0.00003698, 0.00003774, 0.00003400, 0.00002589, 0.00001429, 0.00000065, -0.00001317, -0.00002519, -0.00003370, -0.00003744, -0.00003589, -0.00002932, -0.00001877, -0.00000589, 0.00000732, 0.00001881, 0.00002684, 0.00003020, 0.00002841, 0.00002179, 0.00001139, -0.00000123, -0.00001420, -0.00002566, -0.00003408, -0.00003845, -0.00003840, -0.00003425, -0.00002692, -0.00001772, -0.00000813, 0.00000045, 0.00000691, 0.00001062, 0.00001144, 0.00000976, 0.00000635, 0.00000218, -0.00000174, -0.00000454, -0.00000570, -0.00000505, -0.00000281, 0.00000050, 0.00000414, 0.00000734, 0.00000943, 0.00000997, 0.00000882, 0.00000613, 0.00000233, -0.00000198, -0.00000616, -0.00000960, -0.00001188, -0.00001279, -0.00001236, -0.00001083, -0.00000858, -0.00000606, -0.00000369, -0.00000179, -0.00000053, 0.00000006, 0.00000009, -0.00000021, -0.00000062, -0.00000092, -0.00000097, -0.00000071, -0.00000018, 0.00000052, 0.00000127, 0.00000193, 0.00000243, 0.00000274, },
		{ 0.00000130, -0.00000010, -0.00000193, -0.00000413, -0.00000655, -0.00000896, -0.00001102, -0.00001240, -0.00001279, -0.00001201, -0.00001006, -0.00000712, -0.00000361, -0.00000008, 0.00000281, 0.00000446, 0.00000438, 0.00000237, -0.00000148, -0.00000672, -0.00001264, -0.00001831, -0.00002274, -0.00002506, -0.00002465, -0.00002128, -0.00001515, -0.00000689, 0.00000253, 0.00001192, 0.00002008, 0.00002598, 0.00002894, 0.00002871, 0.00002553, 0.00002010, 0.00001340, 0.00000659, 0.00000074, -0.00000330, -0.00000508, -0.00000462, -0.00000236, 0.00000090, 0.00000417, 0.00000648, 0.00000708, 0.00000558, 0.00000204, -0.00000301, -0.00000866, -0.00001379, -0.00001730, -0.00001830, -0.00001629, -0.00001131, -0.00000396, 0.00000470, 0.00001326, 0.00002024, 0.00002431, 0.00002454, 0.00002055, 0.00001265, 0.00000175, -0.00001071, -0.00002301, -0.00003340, -0.00004038, -0.00004292, -0.00004066, -0.00003391, -0.00002364, -0.00001129, 0.00000141, 0.00001279, 0.00002140, 0.00002629, 0.00002711, 0.00002413, 0.00001814, 0.00001037, 0.00000217, -0.00000514, -0.00001051, -0.00001330, -0.00001337, -0.00001105, -0.00000703, -0.00000225, 0.00000233, 0.00000584, 0.00000772, 0.00000773, 0.00000601, 0.00000301, -0.00000063, -0.00000419, -0.00000700, -0.00000860, -0.00000877, -0.00000758, -0.00000533, -0.00000251, 0.00000034, 0.00000269, 0.00000418, 0.00000459, 0.00000395, 0.00000246, 0.00000045, -0.00000166, -0.00000350, -0.00000477, -0.00000531, -0.00000510, -0.00000425, -0.00000294, -0.00000143, 0.00000007, 0.00000137, 0.00000238, 0.00000306, 0.00000342, 0.00000351, 0.00000336, 0.00000297, 0.00000231, },
		{ -0.00001395, -0.00001467, -0.00001442, -0.00001312, -0.00001086, -0.00000788, -0.00000455, -0.00000137, 0.00000119, 0.00000271, 0.00000292, 0.00000180, -0.00000043, -0.00000331, -0.00000622, -0.00000846, -0.00000939, -0.00000853, -0.00000571, -0.00000105, 0.00000495, 0.00001153, 0.00001776, 0.00002268, 0.00002547, 0.00002557, 0.00002282, 0.00001744, 0.00001009, 0.00000166, -0.00000674, -0.00001405, -0.00001939, -0.00002220, -0.00002228, -0.00001989, -0.00001559, -0.00001020, -0.00000462, 0.00000033, 0.00000406, 0.00000627, 0.00000702, 0.00000664, 0.00000567, 0.00000472, 0.00000431, 0.00000474, 0.00000603, 0.00000788, 0.00000972, 0.00001084, 0.00001052, 0.00000822, 0.00000371, -0.00000286, -0.00001087, -0.00001938, -0.00002718, -0.00003303, -0.00003586, -0.00003492, -0.00003001, -0.00002148, -0.00001023, 0.00000238, 0.00001477, 0.00002532, 0.00003265, 0.00003583, 0.00003449, 0.00002892, 0.00001999, 0.00000904, -0.00000237, -0.00001265, -0.00002045, -0.00002485, -0.00002547, -0.00002251, -0.00001670, -0.00000912, -0.00000104, 0.00000628, 0.00001184, 0.00001497, 0.00001548, 0.00001361, 0.00000996, 0.00000537, 0.00000074, -0.00000311, -0.00000562, -0.00000650, -0.00000583, -0.00000394, -0.00000141, 0.00000113, 0.00000309, 0.00000404, 0.00000375, 0.00000229, -0.00000008, -0.00000292, -0.00000570, -0.00000794, -0.00000926, -0.00000944, -0.00000847, -0.00000652, -0.00000391, -0.00000100, 0.00000181, 0.00000419, 0.00000593, 0.00000693, 0.00000718, 0.00000678, 0.00000586, 0.00000454, 0.00000293, 0.00000108, -0.00000096, -0.00000320, -0.00000558, -0.00000802, -0.00001038, -0.00001244, },
		{ 0.00000091, 0.00000410, 0.00000655, 0.00000795, 0.00000815, 0.00000719, 0.00000532, 0.00000293, 0.00000055, -0.00000131, -0.00000221, -0.00000189, -0.00000036, 0.00000216, 0.00000519, 0.00000810, 0.00001022, 0.00001093, 0.00000984, 0.00000681, 0.00000205, -0.00000396, -0.00001045, -0.00001654, -0.00002135, -0.00002414, -0.00002444, -0.00002213, -0.00001743, -0.00001092, -0.00000341, 0.00000420, 0.00001101, 0.00001629, 0.00001958, 0.00002072, 0.00001986, 0.00001738, 0.00001384, 0.00000980, 0.00000577, 0.00000211, -0.00000104, -0.00000372, -0.00000612, -0.00000846, -0.00001095, -0.00001364, -0.00001640, -0.00001891, -0.00002068, -0.00002115, -0.00001980, -0.00001631, -0.00001061, -0.00000300, 0.00000587, 0.00001504, 0.00002338, 0.00002971, 0.00003307, 0.00003280, 0.00002871, 0.00002114, 0.00001094, -0.00000064, -0.00001212, -0.00002201, -0.00002899, -0.00003217, -0.00003118, -0.00002622, -0.00001806, -0.00000788, 0.00000287, 0.00001274, 0.00002045, 0.00002510, 0.00002629, 0.00002413, 0.00001922, 0.00001253, 0.00000518, -0.00000169, -0.00000712, -0.00001049, -0.00001158, -0.00001057, -0.00000798, -0.00000456, -0.00000111, 0.00000163, 0.00000312, 0.00000312, 0.00000169, -0.00000083, -0.00000389, -0.00000687, -0.00000918, -0.00001035, -0.00001014, -0.00000855, -0.00000581, -0.00000232, 0.00000141, 0.00000487, 0.00000764, 0.00000941, 0.00001007, 0.00000965, 0.00000833, 0.00000634, 0.00000395, 0.00000142, -0.00000109, -0.00000346, -0.00000566, -0.00000768, -0.00000950, -0.00001108, -0.00001235, -0.00001316, -0.00001335, -0.00001277, -0.00001134, -0.00000906, -0.00000606, -0.00000262, },
		{ 0.00000225, -0.00000047, -0.00000308, -0.00000513, -0.00000631, -0.00000644, -0.00000557, -0.00000395, -0.00000198, -0.00000015, 0.00000104, 0.00000122, 0.00000019, -0.00000200, -0.00000507, -0.00000853, -0.00001175, -0.00001409, -0.00001499, -0.00001410, -0.00001130, -0.00000682, -0.00000110, 0.00000516, 0.00001118, 0.00001619, 0.00001957, 0.00002089, 0.00002005, 0.00001719, 0.00001275, 0.00000732, 0.00000157, -0.00000386, -0.00000843, -0.00001180, -0.00001383, -0.00001458, -0.00001422, -0.00001301, -0.00001122, -0.00000903, -0.00000655, -0.00000380, -0.00000073, 0.00000272, 0.00000654, 0.00001061, 0.00001467, 0.00001829, 0.00002097, 0.00002216, 0.00002140, 0.00001841, 0.00001318, 0.00000603, -0.00000238, -0.00001116, -0.00001922, -0.00002548, -0.00002899, -0.00002915, -0.00002574, -0.00001905, -0.00000981, 0.00000085, 0.00001160, 0.00002106, 0.00002803, 0.00003163, 0.00003148, 0.00002770, 0.00002093, 0.00001220, 0.00000276, -0.00000607, -0.00001316, -0.00001767, -0.00001922, -0.00001790, -0.00001423, -0.00000905, -0.00000338, 0.00000177, 0.00000553, 0.00000737, 0.00000711, 0.00000494, 0.00000141, -0.00000277, -0.00000676, -0.00000983, -0.00001142, -0.00001124, -0.00000932, -0.00000597, -0.00000171, 0.00000283, 0.00000699, 0.00001025, 0.00001224, 0.00001281, 0.00001204, 0.00001016, 0.00000754, 0.00000455, 0.00000156, -0.00000118, -0.00000350, -0.00000537, -0.00000684, -0.00000799, -0.00000889, -0.00000958, -0.00001003, -0.00001016, -0.00000983, -0.00000895, -0.00000743, -0.00000530, -0.00000271, 0.00000011, 0.00000281, 0.00000505, 0.00000650, 0.00000694, 0.00000628, 0.00000463, },
		{ -0.00000385, -0.00000199, 0.00000003, 0.00000179, 0.00000290, 0.00000310, 0.00000228, 0.00000055, -0.00000178, -0.00000428, -0.00000645, -0.00000780, -0.00000798, -0.00000682, -0.00000436, -0.00000090, 0.00000310, 0.00000706, 0.00001040, 0.00001259, 0.00001331, 0.00001244, 0.00001010, 0.00000664, 0.00000257, -0.00000153, -0.00000509, -0.00000763, -0.00000887, -0.00000869, -0.00000721, -0.00000470, -0.00000152, 0.00000191, 0.00000519, 0.00000804, 0.00001025, 0.00001174, 0.00001252, 0.00001265, 0.00001219, 0.00001120, 0.00000968, 0.00000760, 0.00000492, 0.00000162, -0.00000224, -0.00000646, -0.00001073, -0.00001461, -0.00001759, -0.00001914, -0.00001882, -0.00001636, -0.00001176, -0.00000529, 0.00000246, 0.00001065, 0.00001832, 0.00002448, 0.00002826, 0.00002907, 0.00002670, 0.00002136, 0.00001366, 0.00000458, -0.00000473, -0.00001305, -0.00001932, -0.00002277, -0.00002304, -0.00002025, -0.00001494, -0.00000802, -0.00000062, 0.00000611, 0.00001118, 0.00001385, 0.00001384, 0.00001126, 0.00000665, 0.00000086, -0.00000513, -0.00001029, -0.00001380, -0.00001510, -0.00001400, -0.00001067, -0.00000563, 0.00000037, 0.00000649, 0.00001191, 0.00001596, 0.00001823, 0.00001859, 0.00001716, 0.00001430, 0.00001051, 0.00000633, 0.00000225, -0.00000138, -0.00000435, -0.00000662, -0.00000826, -0.00000946, -0.00001037, -0.00001115, -0.00001184, -0.00001241, -0.00001273, -0.00001266, -0.00001203, -0.00001076, -0.00000886, -0.00000647, -0.00000383, -0.00000126, 0.00000091, 0.00000237, 0.00000293, 0.00000256, 0.00000136, -0.00000040, -0.00000236, -0.00000411, -0.00000530, -0.00000568, -0.00000517, },
		{ 0.00000064, -0.00000173, -0.00000439, -0.00000691, -0.00000885, -0.00000983, -0.00000959, -0.00000806, -0.00000539, -0.00000188, 0.00000200, 0.00000575, 0.00000888, 0.00001103, 0.00001197, 0.00001168, 0.00001036, 0.00000832, 0.00000600, 0.00000383, 0.00000219, 0.00000135, 0.00000140, 0.00000227, 0.00000374, 0.00000549, 0.00000715, 0.00000840, 0.00000896, 0.00000871, 0.00000762, 0.00000582, 0.00000348, 0.00000088, -0.00000176, -0.00000421, -0.00000630, -0.00000795, -0.00000910, -0.00000974, -0.00000985, -0.00000942, -0.00000840, -0.00000674, -0.00000439, -0.00000135, 0.00000230, 0.00000636, 0.00001052, 0.00001438, 0.00001745, 0.00001925, 0.00001939, 0.00001762, 0.00001393, 0.00000856, 0.00000203, -0.00000496, -0.00001154, -0.00001686, -0.00002018, -0.00002099, -0.00001911, -0.00001474, -0.00000844, -0.00000105, 0.00000638, 0.00001280, 0.00001729, 0.00001919, 0.00001823, 0.00001457, 0.00000875, 0.00000168, -0.00000557, -0.00001189, -0.00001632, -0.00001818, -0.00001715, -0.00001336, -0.00000733, 0.00000010, 0.00000793, 0.00001511, 0.00002073, 0.00002412, 0.00002497, 0.00002329, 0.00001944, 0.00001400, 0.00000772, 0.00000132, -0.00000455, -0.00000943, -0.00001309, -0.00001552, -0.00001690, -0.00001752, -0.00001769, -0.00001768, -0.00001766, -0.00001764, -0.00001754, -0.00001717, -0.00001631, -0.00001479, -0.00001251, -0.00000952, -0.00000601, -0.00000227, 0.00000132, 0.00000437, 0.00000659, 0.00000777, 0.00000789, 0.00000708, 0.00000563, 0.00000388, 0.00000223, 0.00000099, 0.00000037, 0.00000043, 0.00000105, 0.00000198, 0.00000289, 0.00000343, 0.00000331, 0.00000238, },
		{ -0.00000749, -0.00000430, -0.00000021, 0.00000436, 0.00000894, 0.00001303, 0.00001618, 0.00001810, 0.00001864, 0.00001787, 0.00001600, 0.00001338, 0.00001042, 0.00000751, 0.00000498, 0.00000305, 0.00000176, 0.00000106, 0.00000078, 0.00000067, 0.00000050, 0.00000007, -0.00000072, -0.00000188, -0.00000332, -0.00000484, -0.00000624, -0.00000728, -0.00000779, -0.00000764, -0.00000682, -0.00000538, -0.00000345, -0.00000120, 0.00000116, 0.00000348, 0.00000559, 0.00000739, 0.00000879, 0.00000975, 0.00001022, 0.00001015, 0.00000948, 0.00000817, 0.00000620, 0.00000359, 0.00000043, -0.00000306, -0.00000660, -0.00000982, -0.00001229, -0.00001360, -0.00001344, -0.00001162, -0.00000816, -0.00000332, 0.00000243, 0.00000842, 0.00001391, 0.00001813, 0.00002044, 0.00002042, 0.00001795, 0.00001325, 0.00000688, -0.00000035, -0.00000746, -0.00001345, -0.00001743, -0.00001880, -0.00001728, -0.00001302, -0.00000655, 0.00000127, 0.00000941, 0.00001677, 0.00002234, 0.00002540, 0.00002553, 0.00002275, 0.00001742, 0.00001023, 0.00000206, -0.00000615, -0.00001353, -0.00001942, -0.00002342, -0.00002544, -0.00002568, -0.00002453, -0.00002251, -0.00002016, -0.00001792, -0.00001607, -0.00001468, -0.00001367, -0.00001278, -0.00001170, -0.00001012, -0.00000783, -0.00000477, -0.00000104, 0.00000307, 0.00000718, 0.00001085, 0.00001366, 0.00001531, 0.00001564, 0.00001469, 0.00001267, 0.00000994, 0.00000692, 0.00000401, 0.00000156, -0.00000022, -0.00000130, -0.00000177, -0.00000188, -0.00000191, -0.00000215, -0.00000280, -0.00000395, -0.00000552, -0.00000729, -0.00000893, -0.00001008, -0.00001036, -0.00000954, },
		{ 0.00002121, 0.00002099, 0.00001932, 0.00001640, 0.00001257, 0.00000827, 0.00000395, 0.00000003, -0.00000318, -0.00000550, -0.00000689, -0.00000743, -0.00000732, -0.00000680, -0.00000609, -0.00000538, -0.00000479, -0.00000434, -0.00000398, -0.00000361, -0.00000310, -0.00000234, -0.00000128, 0.00000007, 0.00000165, 0.00000330, 0.00000485, 0.00000611, 0.00000693, 0.00000718, 0.00000681, 0.00000585, 0.00000439, 0.00000256, 0.00000052, -0.00000156, -0.00000351, -0.00000521, -0.00000651, -0.00000734, -0.00000760, -0.00000723, -0.00000618, -0.00000442, -0.00000196, 0.00000112, 0.00000468, 0.00000848, 0.00001220, 0.00001546, 0.00001785, 0.00001901, 0.00001863, 0.00001659, 0.00001292, 0.00000790, 0.00000198, -0.00000418, -0.00000988, -0.00001438, -0.00001707, -0.00001753, -0.00001564, -0.00001156, -0.00000579, 0.00000092, 0.00000766, 0.00001349, 0.00001756, 0.00001924, 0.00001820, 0.00001450, 0.00000854, 0.00000102, -0.00000716, -0.00001502, -0.00002167, -0.00002639, -0.00002878, -0.00002876, -0.00002659, -0.00002280, -0.00001808, -0.00001318, -0.00000879, -0.00000541, -0.00000329, -0.00000242, -0.00000253, -0.00000318, -0.00000385, -0.00000402, -0.00000332, -0.00000154, 0.00000126, 0.00000483, 0.00000873, 0.00001243, 0.00001541, 0.00001723, 0.00001760, 0.00001648, 0.00001401, 0.00001053, 0.00000650, 0.00000243, -0.00000123, -0.00000413, -0.00000609, -0.00000710, -0.00000731, -0.00000698, -0.00000641, -0.00000588, -0.00000558, -0.00000556, -0.00000575, -0.00000594, -0.00000587, -0.00000525, -0.00000384, -0.00000152, 0.00000167, 0.00000554, 0.00000975, 0.00001386, 0.00001741, 0.00001997, },
		{ -0.00000608, -0.00000900, -0.00001081, -0.00001143, -0.00001099, -0.00000967, -0.00000779, -0.00000564, -0.00000353, -0.00000167, -0.00000020, 0.00000084, 0.00000149, 0.00000185, 0.00000205, 0.00000218, 0.00000231, 0.00000248, 0.00000263, 0.00000270, 0.00000259, 0.00000222, 0.00000153, 0.00000054, -0.00000069, -0.00000206, -0.00000339, -0.00000450, -0.00000523, -0.00000544, -0.00000502, -0.00000396, -0.00000228, -0.00000008, 0.00000250, 0.00000531, 0.00000815, 0.00001085, 0.00001321, 0.00001508, 0.00001628, 0.00001669, 0.00001621, 0.00001477, 0.00001237, 0.00000909, 0.00000507, 0.00000057, -0.00000408, -0.00000848, -0.00001222, -0.00001490, -0.00001621, -0.00001596, -0.00001415, -0.00001097, -0.00000679, -0.00000215, 0.00000232, 0.00000598, 0.00000826, 0.00000876, 0.00000731, 0.00000400, -0.00000079, -0.00000648, -0.00001230, -0.00001749, -0.00002132, -0.00002325, -0.00002300, -0.00002062, -0.00001641, -0.00001100, -0.00000511, 0.00000042, 0.00000486, 0.00000766, 0.00000854, 0.00000753, 0.00000495, 0.00000138, -0.00000245, -0.00000580, -0.00000800, -0.00000859, -0.00000737, -0.00000445, -0.00000021, 0.00000475, 0.00000972, 0.00001399, 0.00001696, 0.00001822, 0.00001764, 0.00001532, 0.00001164, 0.00000712, 0.00000236, -0.00000204, -0.00000560, -0.00000801, -0.00000917, -0.00000917, -0.00000825, -0.00000676, -0.00000504, -0.00000342, -0.00000209, -0.00000110, -0.00000039, 0.00000023, 0.00000098, 0.00000209, 0.00000371, 0.00000589, 0.00000851, 0.00001134, 0.00001406, 0.00001627, 0.00001762, 0.00001781, 0.00001672, 0.00001435, 0.00001089, 0.00000668, 0.00000215, -0.00000225, },
		{ -0.00000301, -0.00000060, 0.00000132, 0.00000260, 0.00000320, 0.00000315, 0.00000261, 0.00000176, 0.00000079, -0.00000011, -0.00000085, -0.00000134, -0.00000161, -0.00000169, -0.00000165, -0.00000158, -0.00000150, -0.00000144, -0.00000136, -0.00000117, -0.00000080, -0.00000014, 0.00000087, 0.00000226, 0.00000399, 0.00000599, 0.00000811, 0.00001018, 0.00001201, 0.00001342, 0.00001424, 0.00001436, 0.00001374, 0.00001236, 0.00001028, 0.00000760, 0.00000447, 0.00000105, -0.00000248, -0.00000590, -0.00000902, -0.00001164, -0.00001360, -0.00001477, -0.00001506, -0.00001445, -0.00001303, -0.00001093, -0.00000839, -0.00000571, -0.00000322, -0.00000123, -0.00000001, 0.00000024, -0.00000050, -0.00000213, -0.00000440, -0.00000694, -0.00000928, -0.00001097, -0.00001161, -0.00001093, -0.00000885, -0.00000549, -0.00000122, 0.00000344, 0.00000786, 0.00001139, 0.00001345, 0.00001364, 0.00001183, 0.00000813, 0.00000298, -0.00000298, -0.00000895, -0.00001408, -0.00001761, -0.00001898, -0.00001790, -0.00001443, -0.00000893, -0.00000205, 0.00000537, 0.00001242, 0.00001826, 0.00002220, 0.00002382, 0.00002304, 0.00002007, 0.00001539, 0.00000968, 0.00000370, -0.00000180, -0.00000622, -0.00000917, -0.00001048, -0.00001024, -0.00000873, -0.00000638, -0.00000366, -0.00000100, 0.00000125, 0.00000292, 0.00000397, 0.00000452, 0.00000478, 0.00000501, 0.00000541, 0.00000614, 0.00000720, 0.00000851, 0.00000984, 0.00001091, 0.00001143, 0.00001115, 0.00000993, 0.00000775, 0.00000476, 0.00000122, -0.00000250, -0.00000600, -0.00000889, -0.00001086, -0.00001173, -0.00001148, -0.00001020, -0.00000815, -0.00000563, },
		{ 0.00000027, -0.00000098, -0.00000201, -0.00000264, -0.00000280, -0.00000248, -0.00000175, -0.00000074, 0.00000041, 0.00000158, 0.00000266, 0.00000359, 0.00000436, 0.00000502, 0.00000561, 0.00000620, 0.00000683, 0.00000752, 0.00000822, 0.00000886, 0.00000934, 0.00000955, 0.00000936, 0.00000870, 0.00000751, 0.00000581, 0.00000367, 0.00000119, -0.00000147, -0.00000415, -0.00000669, -0.00000892, -0.00001073, -0.00001204, -0.00001282, -0.00001305, -0.00001278, -0.00001207, -0.00001101, -0.00000971, -0.00000829, -0.00000685, -0.00000550, -0.00000433, -0.00000341, -0.00000275, -0.00000234, -0.00000211, -0.00000194, -0.00000170, -0.00000124, -0.00000040, 0.00000089, 0.00000268, 0.00000488, 0.00000732, 0.00000973, 0.00001178, 0.00001307, 0.00001327, 0.00001211, 0.00000944, 0.00000533, 0.00000002, -0.00000604, -0.00001223, -0.00001787, -0.00002223, -0.00002469, -0.00002481, -0.00002238, -0.00001751, -0.00001059, -0.00000227, 0.00000660, 0.00001507, 0.00002225, 0.00002738, 0.00002994, 0.00002974, 0.00002689, 0.00002183, 0.00001527, 0.00000801, 0.00000095, -0.00000514, -0.00000965, -0.00001222, -0.00001278, -0.00001153, -0.00000886, -0.00000535, -0.00000159, 0.00000188, 0.00000461, 0.00000637, 0.00000709, 0.00000690, 0.00000607, 0.00000492, 0.00000378, 0.00000293, 0.00000251, 0.00000255, 0.00000293, 0.00000345, 0.00000385, 0.00000387, 0.00000333, 0.00000214, 0.00000034, -0.00000190, -0.00000433, -0.00000662, -0.00000847, -0.00000961, -0.00000988, -0.00000924, -0.00000780, -0.00000575, -0.00000341, -0.00000107, 0.00000094, 0.00000240, 0.00000315, 0.00000319, 0.00000260, 0.00000155, },
		{ 0.00000174, 0.00000328, 0.00000468, 0.00000580, 0.00000655, 0.00000689, 0.00000685, 0.00000653, 0.00000603, 0.00000545, 0.00000488, 0.00000438, 0.00000395, 0.00000356, 0.00000313, 0.00000260, 0.00000189, 0.00000094, -0.00000024, -0.00000163, -0.00000316, -0.00000472, -0.00000619, -0.00000748, -0.00000847, -0.00000912, -0.00000940, -0.00000934, -0.00000899, -0.00000841, -0.00000769, -0.00000690, -0.00000608, -0.00000528, -0.00000448, -0.00000367, -0.00000281, -0.00000185, -0.00000076, 0.00000051, 0.00000194, 0.00000354, 0.00000526, 0.00000702, 0.00000873, 0.00001026, 0.00001148, 0.00001223, 0.00001235, 0.00001173, 0.00001024, 0.00000783, 0.00000453, 0.00000043, -0.00000425, -0.00000923, -0.00001410, -0.00001844, -0.00002178, -0.00002368, -0.00002381, -0.00002197, -0.00001812, -0.00001246, -0.00000537, 0.00000258, 0.00001069, 0.00001821, 0.00002441, 0.00002868, 0.00003061, 0.00003003, 0.00002705, 0.00002204, 0.00001559, 0.00000847, 0.00000146, -0.00000468, -0.00000933, -0.00001209, -0.00001283, -0.00001168, -0.00000899, -0.00000531, -0.00000127, 0.00000251, 0.00000550, 0.00000732, 0.00000781, 0.00000702, 0.00000518, 0.00000267, -0.00000005, -0.00000255, -0.00000447, -0.00000559, -0.00000583, -0.00000528, -0.00000416, -0.00000278, -0.00000146, -0.00000049, -0.00000005, -0.00000024, -0.00000100, -0.00000216, -0.00000349, -0.00000471, -0.00000557, -0.00000589, -0.00000557, -0.00000466, -0.00000328, -0.00000164, -0.00000000, 0.00000139, 0.00000233, 0.00000271, 0.00000252, 0.00000183, 0.00000082, -0.00000031, -0.00000131, -0.00000198, -0.00000217, -0.00000184, -0.00000100, 0.00000025, },
		{ 0.00000428, 0.00000311, 0.00000195, 0.00000092, 0.00000013, -0.00000037, -0.00000060, -0.00000062, -0.00000054, -0.00000044, -0.00000044, -0.00000060, -0.00000096, -0.00000150, -0.00000217, -0.00000290, -0.00000360, -0.00000419, -0.00000462, -0.00000486, -0.00000490, -0.00000479, -0.00000455, -0.00000423, -0.00000387, -0.00000348, -0.00000304, -0.00000252, -0.00000185, -0.00000099, 0.00000011, 0.00000146, 0.00000306, 0.00000481, 0.00000661, 0.00000831, 0.00000975, 0.00001073, 0.00001111, 0.00001077, 0.00000963, 0.00000770, 0.00000503, 0.00000176, -0.00000193, -0.00000581, -0.00000961, -0.00001305, -0.00001587, -0.00001780, -0.00001865, -0.00001827, -0.00001660, -0.00001366, -0.00000959, -0.00000461, 0.00000097, 0.00000675, 0.00001230, 0.00001719, 0.00002100, 0.00002341, 0.00002421, 0.00002334, 0.00002090, 0.00001715, 0.00001248, 0.00000739, 0.00000240, -0.00000197, -0.00000532, -0.00000737, -0.00000800, -0.00000731, -0.00000554, -0.00000310, -0.00000045, 0.00000191, 0.00000356, 0.00000419, 0.00000366, 0.00000201, -0.00000055, -0.00000366, -0.00000688, -0.00000975, -0.00001186, -0.00001292, -0.00001277, -0.00001145, -0.00000915, -0.00000619, -0.00000298, 0.00000007, 0.00000259, 0.00000430, 0.00000507, 0.00000491, 0.00000395, 0.00000246, 0.00000072, -0.00000094, -0.00000227, -0.00000308, -0.00000332, -0.00000301, -0.00000230, -0.00000136, -0.00000043, 0.00000032, 0.00000075, 0.00000081, 0.00000052, -0.00000001, -0.00000061, -0.00000111, -0.00000134, -0.00000119, -0.00000061, 0.00000035, 0.00000161, 0.00000298, 0.00000430, 0.00000537, 0.00000605, 0.00000626, 0.00000598, 0.00000529, },
		{ 0.00000003, 0.00000078, 0.00000155, 0.00000219, 0.00000259, 0.00000268, 0.00000243, 0.00000188, 0.00000109, 0.00000016, -0.00000081, -0.00000171, -0.00000249, -0.00000309, -0.00000351, -0.00000375, -0.00000383, -0.00000378, -0.00000359, -0.00000328, -0.00000280, -0.00000213, -0.00000124, -0.00000011, 0.00000125, 0.00000277, 0.00000435, 0.00000586, 0.00000713, 0.00000799, 0.00000828, 0.00000790, 0.00000679, 0.00000498, 0.00000257, -0.00000026, -0.00000329, -0.00000626, -0.00000890, -0.00001097, -0.00001229, -0.00001274, -0.00001228, -0.00001093, -0.00000881, -0.00000608, -0.00000293, 0.00000042, 0.00000376, 0.00000688, 0.00000961, 0.00001182, 0.00001340, 0.00001431, 0.00001453, 0.00001409, 0.00001305, 0.00001154, 0.00000967, 0.00000761, 0.00000551, 0.00000354, 0.00000181, 0.00000041, -0.00000061, -0.00000127, -0.00000167, -0.00000193, -0.00000221, -0.00000267, -0.00000345, -0.00000464, -0.00000625, -0.00000819, -0.00001031, -0.00001238, -0.00001412, -0.00001525, -0.00001552, -0.00001477, -0.00001295, -0.00001012, -0.00000651, -0.00000241, 0.00000178, 0.00000566, 0.00000882, 0.00001097, 0.00001191, 0.00001159, 0.00001011, 0.00000771, 0.00000471, 0.00000149, -0.00000156, -0.00000412, -0.00000596, -0.00000695, -0.00000712, -0.00000656, -0.00000548, -0.00000412, -0.00000273, -0.00000149, -0.00000055, 0.00000005, 0.00000033, 0.00000040, 0.00000038, 0.00000042, 0.00000062, 0.00000106, 0.00000174, 0.00000259, 0.00000350, 0.00000433, 0.00000495, 0.00000522, 0.00000509, 0.00000457, 0.00000370, 0.00000262, 0.00000148, 0.00000044, -0.00000035, -0.00000080, -0.00000086, -0.00000056, },
		{ 0.00000318, 0.00000191, 0.00000045, -0.00000107, -0.00000251, -0.00000377, -0.00000477, -0.00000547, -0.00000586, -0.00000597, -0.00000582, -0.00000545, -0.00000488, -0.00000414, -0.00000321, -0.00000211, -0.00000084, 0.00000057, 0.00000205, 0.00000352, 0.00000484, 0.00000589, 0.00000653, 0.00000664, 0.00000615, 0.00000505, 0.00000342, 0.00000139, -0.00000086, -0.00000309, -0.00000507, -0.00000659, -0.00000748, -0.00000766, -0.00000714, -0.00000598, -0.00000435, -0.00000244, -0.00000046, 0.00000139, 0.00000298, 0.00000421, 0.00000507, 0.00000559, 0.00000588, 0.00000606, 0.00000623, 0.00000649, 0.00000687, 0.00000735, 0.00000787, 0.00000829, 0.00000848, 0.00000828, 0.00000758, 0.00000630, 0.00000441, 0.00000197, -0.00000091, -0.00000407, -0.00000732, -0.00001044, -0.00001325, -0.00001557, -0.00001725, -0.00001820, -0.00001835, -0.00001769, -0.00001625, -0.00001408, -0.00001126, -0.00000791, -0.00000420, -0.00000029, 0.00000359, 0.00000721, 0.00001032, 0.00001271, 0.00001417, 0.00001457, 0.00001385, 0.00001204, 0.00000928, 0.00000581, 0.00000192, -0.00000203, -0.00000569, -0.00000872, -0.00001088, -0.00001200, -0.00001205, -0.00001111, -0.00000936, -0.00000705, -0.00000449, -0.00000197, 0.00000027, 0.00000204, 0.00000325, 0.00000391, 0.00000410, 0.00000395, 0.00000365, 0.00000334, 0.00000314, 0.00000314, 0.00000333, 0.00000367, 0.00000407, 0.00000440, 0.00000458, 0.00000452, 0.00000420, 0.00000365, 0.00000295, 0.00000221, 0.00000157, 0.00000113, 0.00000100, 0.00000119, 0.00000169, 0.00000242, 0.00000324, 0.00000402, 0.00000460, 0.00000485, 0.00000471, 0.00000414, },
		{ -0.00000775, -0.00000783, -0.00000755, -0.00000696, -0.00000613, -0.00000511, -0.00000395, -0.00000270, -0.00000137, 0.00000002, 0.00000143, 0.00000282, 0.00000412, 0.00000524, 0.00000609, 0.00000657, 0.00000659, 0.00000612, 0.00000516, 0.00000378, 0.00000210, 0.00000029, -0.00000145, -0.00000293, -0.00000399, -0.00000452, -0.00000450, -0.00000397, -0.00000306, -0.00000193, -0.00000077, 0.00000023, 0.00000094, 0.00000129, 0.00000129, 0.00000106, 0.00000073, 0.00000050, 0.00000055, 0.00000101, 0.00000193, 0.00000327, 0.00000487, 0.00000650, 0.00000786, 0.00000863, 0.00000854, 0.00000739, 0.00000513, 0.00000181, -0.00000233, -0.00000695, -0.00001164, -0.00001591, -0.00001934, -0.00002154, -0.00002229, -0.00002148, -0.00001918, -0.00001559, -0.00001104, -0.00000592, -0.00000067, 0.00000430, 0.00000863, 0.00001205, 0.00001438, 0.00001555, 0.00001559, 0.00001458, 0.00001268, 0.00001007, 0.00000695, 0.00000353, 0.00000002, -0.00000340, -0.00000653, -0.00000919, -0.00001125, -0.00001257, -0.00001306, -0.00001270, -0.00001150, -0.00000957, -0.00000706, -0.00000419, -0.00000118, 0.00000169, 0.00000421, 0.00000618, 0.00000748, 0.00000808, 0.00000802, 0.00000744, 0.00000650, 0.00000541, 0.00000436, 0.00000352, 0.00000300, 0.00000282, 0.00000298, 0.00000337, 0.00000388, 0.00000438, 0.00000474, 0.00000489, 0.00000479, 0.00000448, 0.00000401, 0.00000350, 0.00000303, 0.00000271, 0.00000259, 0.00000268, 0.00000293, 0.00000326, 0.00000355, 0.00000368, 0.00000353, 0.00000304, 0.00000217, 0.00000097, -0.00000050, -0.00000211, -0.00000372, -0.00000519, -0.00000641, -0.00000727, },
		{ -0.00000095, 0.00000034, 0.00000160, 0.00000279, 0.00000389, 0.00000487, 0.00000569, 0.00000626, 0.00000653, 0.00000645, 0.00000597, 0.00000511, 0.00000392, 0.00000249, 0.00000095, -0.00000053, -0.00000181, -0.00000276, -0.00000331, -0.00000344, -0.00000322, -0.00000274, -0.00000217, -0.00000166, -0.00000137, -0.00000137, -0.00000169, -0.00000226, -0.00000295, -0.00000357, -0.00000391, -0.00000380, -0.00000312, -0.00000187, -0.00000014, 0.00000184, 0.00000377, 0.00000529, 0.00000608, 0.00000585, 0.00000445, 0.00000190, -0.00000164, -0.00000581, -0.00001014, -0.00001409, -0.00001713, -0.00001881, -0.00001881, -0.00001702, -0.00001354, -0.00000870, -0.00000297, 0.00000304, 0.00000871, 0.00001344, 0.00001677, 0.00001839, 0.00001820, 0.00001631, 0.00001300, 0.00000868, 0.00000385, -0.00000099, -0.00000541, -0.00000903, -0.00001163, -0.00001312, -0.00001351, -0.00001293, -0.00001158, -0.00000965, -0.00000737, -0.00000493, -0.00000247, -0.00000012, 0.00000207, 0.00000402, 0.00000569, 0.00000705, 0.00000806, 0.00000867, 0.00000888, 0.00000866, 0.00000807, 0.00000715, 0.00000603, 0.00000482, 0.00000369, 0.00000275, 0.00000214, 0.00000191, 0.00000207, 0.00000258, 0.00000333, 0.00000417, 0.00000496, 0.00000555, 0.00000584, 0.00000577, 0.00000533, 0.00000460, 0.00000366, 0.00000266, 0.00000172, 0.00000095, 0.00000040, 0.00000010, -0.00000001, -0.00000000, 0.00000001, -0.00000008, -0.00000037, -0.00000091, -0.00000171, -0.00000273, -0.00000388, -0.00000505, -0.00000611, -0.00000693, -0.00000744, -0.00000758, -0.00000734, -0.00000676, -0.00000588, -0.00000479, -0.00000356, -0.00000226, },
		{ 0.00000588, 0.00000605, 0.00000592, 0.00000548, 0.00000471, 0.00000365, 0.00000236, 0.00000094, -0.00000049, -0.00000180, -0.00000290, -0.00000370, -0.00000419, -0.00000441, -0.00000443, -0.00000439, -0.00000439, -0.00000455, -0.00000493, -0.00000552, -0.00000624, -0.00000696, -0.00000748, -0.00000763, -0.00000723, -0.00000620, -0.00000455, -0.00000240, 0.00000002, 0.00000239, 0.00000438, 0.00000564, 0.00000592, 0.00000511, 0.00000323, 0.00000049, -0.00000275, -0.00000601, -0.00000880, -0.00001062, -0.00001111, -0.00001008, -0.00000753, -0.00000373, 0.00000090, 0.00000577, 0.00001024, 0.00001369, 0.00001562, 0.00001573, 0.00001394, 0.00001044, 0.00000563, 0.00000009, -0.00000552, -0.00001051, -0.00001432, -0.00001653, -0.00001693, -0.00001555, -0.00001264, -0.00000860, -0.00000396, 0.00000075, 0.00000503, 0.00000849, 0.00001089, 0.00001213, 0.00001229, 0.00001156, 0.00001020, 0.00000851, 0.00000675, 0.00000514, 0.00000383, 0.00000288, 0.00000228, 0.00000198, 0.00000189, 0.00000194, 0.00000206, 0.00000223, 0.00000244, 0.00000271, 0.00000305, 0.00000346, 0.00000393, 0.00000441, 0.00000482, 0.00000507, 0.00000508, 0.00000477, 0.00000411, 0.00000313, 0.00000190, 0.00000053, -0.00000085, -0.00000208, -0.00000303, -0.00000362, -0.00000381, -0.00000364, -0.00000320, -0.00000260, -0.00000200, -0.00000153, -0.00000131, -0.00000140, -0.00000179, -0.00000246, -0.00000328, -0.00000415, -0.00000493, -0.00000550, -0.00000577, -0.00000569, -0.00000528, -0.00000456, -0.00000360, -0.00000249, -0.00000129, -0.00000008, 0.00000109, 0.00000219, 0.00000320, 0.00000410, 0.00000487, 0.00000548, },
		{ -0.00000044, -0.00000205, -0.00000359, -0.00000495, -0.00000606, -0.00000686, -0.00000738, -0.00000765, -0.00000775, -0.00000778, -0.00000784, -0.00000796, -0.00000815, -0.00000835, -0.00000846, -0.00000833, -0.00000780, -0.00000675, -0.00000511, -0.00000292, -0.00000030, 0.00000253, 0.00000526, 0.00000757, 0.00000914, 0.00000973, 0.00000924, 0.00000768, 0.00000526, 0.00000229, -0.00000078, -0.00000349, -0.00000542, -0.00000626, -0.00000584, -0.00000423, -0.00000166, 0.00000146, 0.00000458, 0.00000717, 0.00000873, 0.00000891, 0.00000758, 0.00000484, 0.00000101, -0.00000338, -0.00000770, -0.00001128, -0.00001355, -0.00001410, -0.00001273, -0.00000954, -0.00000486, 0.00000076, 0.00000665, 0.00001208, 0.00001643, 0.00001918, 0.00002008, 0.00001910, 0.00001646, 0.00001258, 0.00000801, 0.00000336, -0.00000080, -0.00000404, -0.00000605, -0.00000674, -0.00000621, -0.00000468, -0.00000250, -0.00000006, 0.00000230, 0.00000427, 0.00000568, 0.00000643, 0.00000656, 0.00000617, 0.00000541, 0.00000444, 0.00000338, 0.00000235, 0.00000138, 0.00000047, -0.00000039, -0.00000124, -0.00000210, -0.00000295, -0.00000376, -0.00000444, -0.00000491, -0.00000508, -0.00000490, -0.00000435, -0.00000348, -0.00000236, -0.00000115, -0.00000001, 0.00000089, 0.00000143, 0.00000151, 0.00000112, 0.00000029, -0.00000085, -0.00000215, -0.00000344, -0.00000455, -0.00000534, -0.00000571, -0.00000562, -0.00000509, -0.00000420, -0.00000303, -0.00000169, -0.00000031, 0.00000104, 0.00000229, 0.00000337, 0.00000427, 0.00000496, 0.00000543, 0.00000566, 0.00000563, 0.00000531, 0.00000469, 0.00000377, 0.00000256, 0.00000114, },
		{ -0.00000949, -0.00000907, -0.00000846, -0.00000777, -0.00000707, -0.00000639, -0.00000574, -0.00000505, -0.00000424, -0.00000320, -0.00000184, -0.00000011, 0.00000197, 0.00000429, 0.00000666, 0.00000885, 0.00001057, 0.00001156, 0.00001163, 0.00001067, 0.00000873, 0.00000599, 0.00000275, -0.00000058, -0.00000359, -0.00000588, -0.00000715, -0.00000725, -0.00000621, -0.00000424, -0.00000170, 0.00000096, 0.00000327, 0.00000482, 0.00000534, 0.00000474, 0.00000313, 0.00000084, -0.00000168, -0.00000389, -0.00000527, -0.00000541, -0.00000409, -0.00000131, 0.00000267, 0.00000738, 0.00001220, 0.00001644, 0.00001945, 0.00002074, 0.00002004, 0.00001734, 0.00001292, 0.00000730, 0.00000115, -0.00000476, -0.00000973, -0.00001316, -0.00001470, -0.00001425, -0.00001199, -0.00000830, -0.00000377, 0.00000096, 0.00000527, 0.00000862, 0.00001066, 0.00001125, 0.00001044, 0.00000846, 0.00000569, 0.00000256, -0.00000052, -0.00000319, -0.00000522, -0.00000649, -0.00000702, -0.00000692, -0.00000635, -0.00000552, -0.00000459, -0.00000369, -0.00000287, -0.00000215, -0.00000148, -0.00000081, -0.00000009, 0.00000070, 0.00000152, 0.00000231, 0.00000294, 0.00000330, 0.00000327, 0.00000279, 0.00000184, 0.00000047, -0.00000119, -0.00000297, -0.00000466, -0.00000607, -0.00000704, -0.00000744, -0.00000723, -0.00000644, -0.00000517, -0.00000355, -0.00000175, 0.00000005, 0.00000172, 0.00000313, 0.00000422, 0.00000494, 0.00000531, 0.00000532, 0.00000502, 0.00000443, 0.00000357, 0.00000247, 0.00000116, -0.00000033, -0.00000194, -0.00000360, -0.00000522, -0.00000671, -0.00000795, -0.00000888, -0.00000945, -0.00000964, },
		{ 0.00000085, 0.00000156, 0.00000225, 0.00000300, 0.00000388, 0.00000492, 0.00000607, 0.00000724, 0.00000827, 0.00000896, 0.00000913, 0.00000863, 0.00000738, 0.00000542, 0.00000290, 0.00000008, -0.00000272, -0.00000513, -0.00000681, -0.00000752, -0.00000714, -0.00000569, -0.00000336, -0.00000048, 0.00000254, 0.00000528, 0.00000735, 0.00000850, 0.00000863, 0.00000783, 0.00000634, 0.00000455, 0.00000290, 0.00000181, 0.00000159, 0.00000239, 0.00000418, 0.00000669, 0.00000949, 0.00001206, 0.00001385, 0.00001438, 0.00001335, 0.00001068, 0.00000655, 0.00000139, -0.00000420, -0.00000950, -0.00001382, -0.00001655, -0.00001732, -0.00001599, -0.00001276, -0.00000804, -0.00000250, 0.00000311, 0.00000805, 0.00001165, 0.00001348, 0.00001335, 0.00001135, 0.00000783, 0.00000332, -0.00000153, -0.00000607, -0.00000973, -0.00001210, -0.00001299, -0.00001241, -0.00001059, -0.00000787, -0.00000471, -0.00000155, 0.00000124, 0.00000340, 0.00000478, 0.00000540, 0.00000538, 0.00000489, 0.00000417, 0.00000338, 0.00000266, 0.00000206, 0.00000157, 0.00000110, 0.00000056, -0.00000016, -0.00000113, -0.00000234, -0.00000372, -0.00000516, -0.00000648, -0.00000750, -0.00000806, -0.00000805, -0.00000743, -0.00000624, -0.00000460, -0.00000267, -0.00000066, 0.00000122, 0.00000281, 0.00000396, 0.00000462, 0.00000477, 0.00000444, 0.00000372, 0.00000271, 0.00000149, 0.00000015, -0.00000125, -0.00000264, -0.00000400, -0.00000528, -0.00000644, -0.00000744, -0.00000821, -0.00000871, -0.00000886, -0.00000866, -0.00000809, -0.00000719, -0.00000604, -0.00000474, -0.00000339, -0.00000211, -0.00000095, 0.00000003, },
		{ 0.00000413, 0.00000440, 0.00000462, 0.00000467, 0.00000441, 0.00000375, 0.00000269, 0.00000130, -0.00000028, -0.00000180, -0.00000302, -0.00000369, -0.00000363, -0.00000274, -0.00000105, 0.00000126, 0.00000393, 0.00000660, 0.00000892, 0.00001056, 0.00001128, 0.00001102, 0.00000985, 0.00000798, 0.00000577, 0.00000360, 0.00000185, 0.00000082, 0.00000065, 0.00000133, 0.00000265, 0.00000426, 0.00000574, 0.00000665, 0.00000662, 0.00000543, 0.00000307, -0.00000026, -0.00000417, -0.00000812, -0.00001153, -0.00001383, -0.00001462, -0.00001369, -0.00001111, -0.00000717, -0.00000241, 0.00000249, 0.00000682, 0.00000993, 0.00001136, 0.00001087, 0.00000852, 0.00000465, -0.00000018, -0.00000526, -0.00000987, -0.00001332, -0.00001515, -0.00001510, -0.00001321, -0.00000977, -0.00000527, -0.00000034, 0.00000437, 0.00000827, 0.00001094, 0.00001214, 0.00001186, 0.00001031, 0.00000783, 0.00000486, 0.00000185, -0.00000081, -0.00000286, -0.00000417, -0.00000476, -0.00000478, -0.00000446, -0.00000403, -0.00000373, -0.00000368, -0.00000395, -0.00000449, -0.00000518, -0.00000582, -0.00000624, -0.00000627, -0.00000584, -0.00000492, -0.00000360, -0.00000202, -0.00000039, 0.00000110, 0.00000226, 0.00000296, 0.00000313, 0.00000279, 0.00000201, 0.00000091, -0.00000037, -0.00000169, -0.00000292, -0.00000400, -0.00000486, -0.00000549, -0.00000591, -0.00000612, -0.00000615, -0.00000599, -0.00000564, -0.00000508, -0.00000429, -0.00000329, -0.00000210, -0.00000079, 0.00000057, 0.00000186, 0.00000299, 0.00000387, 0.00000445, 0.00000471, 0.00000471, 0.00000452, 0.00000425, 0.00000402, 0.00000390, 0.00000394, },
		{ 0.00000233, 0.00000190, 0.00000136, 0.00000088, 0.00000062, 0.00000073, 0.00000130, 0.00000234, 0.00000375, 0.00000535, 0.00000691, 0.00000815, 0.00000884, 0.00000880, 0.00000799, 0.00000648, 0.00000446, 0.00000222, 0.00000011, -0.00000156, -0.00000253, -0.00000266, -0.00000199, -0.00000068, 0.00000095, 0.00000254, 0.00000368, 0.00000404, 0.00000343, 0.00000182, -0.00000064, -0.00000361, -0.00000665, -0.00000927, -0.00001102, -0.00001157, -0.00001079, -0.00000875, -0.00000574, -0.00000223, 0.00000121, 0.00000400, 0.00000568, 0.00000593, 0.00000468, 0.00000211, -0.00000137, -0.00000517, -0.00000864, -0.00001116, -0.00001224, -0.00001159, -0.00000922, -0.00000539, -0.00000058, 0.00000454, 0.00000928, 0.00001296, 0.00001510, 0.00001540, 0.00001387, 0.00001075, 0.00000651, 0.00000174, -0.00000290, -0.00000685, -0.00000964, -0.00001105, -0.00001105, -0.00000984, -0.00000779, -0.00000533, -0.00000293, -0.00000100, 0.00000019, 0.00000052, 0.00000004, -0.00000106, -0.00000249, -0.00000394, -0.00000510, -0.00000576, -0.00000580, -0.00000523, -0.00000418, -0.00000283, -0.00000143, -0.00000021, 0.00000065, 0.00000104, 0.00000095, 0.00000044, -0.00000037, -0.00000132, -0.00000225, -0.00000301, -0.00000351, -0.00000373, -0.00000366, -0.00000336, -0.00000288, -0.00000230, -0.00000166, -0.00000098, -0.00000027, 0.00000050, 0.00000132, 0.00000221, 0.00000315, 0.00000407, 0.00000490, 0.00000556, 0.00000596, 0.00000603, 0.00000575, 0.00000517, 0.00000434, 0.00000339, 0.00000247, 0.00000170, 0.00000118, 0.00000097, 0.00000106, 0.00000138, 0.00000181, 0.00000223, 0.00000250, 0.00000254, },
		{ 0.00000212, 0.00000257, 0.00000320, 0.00000388, 0.00000444, 0.00000473, 0.00000460, 0.00000400, 0.00000297, 0.00000163, 0.00000019, -0.00000111, -0.00000204, -0.00000239, -0.00000210, -0.00000118, 0.00000020, 0.00000177, 0.00000320, 0.00000416, 0.00000436, 0.00000362, 0.00000194, -0.00000055, -0.00000355, -0.00000665, -0.00000940, -0.00001139, -0.00001232, -0.00001203, -0.00001057, -0.00000816, -0.00000521, -0.00000219, 0.00000044, 0.00000225, 0.00000299, 0.00000262, 0.00000129, -0.00000067, -0.00000279, -0.00000455, -0.00000549, -0.00000530, -0.00000384, -0.00000123, 0.00000221, 0.00000595, 0.00000942, 0.00001200, 0.00001323, 0.00001280, 0.00001066, 0.00000703, 0.00000237, -0.00000272, -0.00000756, -0.00001148, -0.00001398, -0.00001475, -0.00001376, -0.00001124, -0.00000761, -0.00000347, 0.00000053, 0.00000380, 0.00000588, 0.00000653, 0.00000571, 0.00000364, 0.00000070, -0.00000262, -0.00000579, -0.00000836, -0.00000999, -0.00001050, -0.00000990, -0.00000838, -0.00000624, -0.00000383, -0.00000152, 0.00000040, 0.00000172, 0.00000238, 0.00000240, 0.00000194, 0.00000120, 0.00000038, -0.00000032, -0.00000075, -0.00000086, -0.00000064, -0.00000016, 0.00000049, 0.00000122, 0.00000193, 0.00000259, 0.00000318, 0.00000371, 0.00000421, 0.00000473, 0.00000526, 0.00000580, 0.00000627, 0.00000661, 0.00000672, 0.00000652, 0.00000596, 0.00000505, 0.00000385, 0.00000248, 0.00000109, -0.00000016, -0.00000110, -0.00000162, -0.00000169, -0.00000133, -0.00000064, 0.00000026, 0.00000118, 0.00000197, 0.00000253, 0.00000279, 0.00000276, 0.00000254, 0.00000224, 0.00000200, 0.00000193, },
		{ 0.00000284, 0.00000250, 0.00000194, 0.00000126, 0.00000059, 0.00000010, -0.00000008, 0.00000012, 0.00000068, 0.00000148, 0.00000235, 0.00000303, 0.00000326, 0.00000284, 0.00000164, -0.00000033, -0.00000293, -0.00000587, -0.00000880, -0.00001132, -0.00001304, -0.00001370, -0.00001315, -0.00001143, -0.00000876, -0.00000550, -0.00000207, 0.00000106, 0.00000350, 0.00000498, 0.00000541, 0.00000488, 0.00000364, 0.00000206, 0.00000056, -0.00000048, -0.00000079, -0.00000025, 0.00000104, 0.00000283, 0.00000474, 0.00000630, 0.00000711, 0.00000683, 0.00000534, 0.00000272, -0.00000076, -0.00000462, -0.00000831, -0.00001124, -0.00001292, -0.00001304, -0.00001153, -0.00000856, -0.00000453, -0.00000005, 0.00000422, 0.00000762, 0.00000963, 0.00000991, 0.00000840, 0.00000531, 0.00000105, -0.00000376, -0.00000846, -0.00001240, -0.00001506, -0.00001612, -0.00001551, -0.00001336, -0.00001005, -0.00000605, -0.00000191, 0.00000185, 0.00000483, 0.00000678, 0.00000763, 0.00000748, 0.00000655, 0.00000517, 0.00000367, 0.00000235, 0.00000143, 0.00000101, 0.00000110, 0.00000162, 0.00000243, 0.00000336, 0.00000426, 0.00000504, 0.00000563, 0.00000604, 0.00000631, 0.00000649, 0.00000664, 0.00000675, 0.00000681, 0.00000675, 0.00000647, 0.00000590, 0.00000497, 0.00000365, 0.00000201, 0.00000015, -0.00000176, -0.00000350, -0.00000489, -0.00000576, -0.00000600, -0.00000560, -0.00000464, -0.00000327, -0.00000169, -0.00000011, 0.00000126, 0.00000228, 0.00000287, 0.00000305, 0.00000289, 0.00000254, 0.00000214, 0.00000183, 0.00000170, 0.00000179, 0.00000205, 0.00000240, 0.00000273, 0.00000290, },
		{ 0.00000068, 0.00000069, 0.00000084, 0.00000104, 0.00000111, 0.00000090, 0.00000024, -0.00000094, -0.00000264, -0.00000475, -0.00000704, -0.00000924, -0.00001100, -0.00001203, -0.00001210, -0.00001112, -0.00000912, -0.00000631, -0.00000300, 0.00000038, 0.00000343, 0.00000576, 0.00000712, 0.00000741, 0.00000668, 0.00000516, 0.00000319, 0.00000114, -0.00000060, -0.00000175, -0.00000217, -0.00000186, -0.00000099, 0.00000015, 0.00000122, 0.00000188, 0.00000188, 0.00000110, -0.00000040, -0.00000240, -0.00000455, -0.00000642, -0.00000761, -0.00000781, -0.00000685, -0.00000478, -0.00000185, 0.00000150, 0.00000472, 0.00000725, 0.00000859, 0.00000838, 0.00000653, 0.00000316, -0.00000134, -0.00000639, -0.00001132, -0.00001543, -0.00001813, -0.00001902, -0.00001795, -0.00001505, -0.00001066, -0.00000535, 0.00000022, 0.00000539, 0.00000960, 0.00001245, 0.00001376, 0.00001357, 0.00001215, 0.00000988, 0.00000725, 0.00000472, 0.00000267, 0.00000137, 0.00000091, 0.00000124, 0.00000218, 0.00000349, 0.00000488, 0.00000611, 0.00000700, 0.00000747, 0.00000752, 0.00000722, 0.00000668, 0.00000603, 0.00000533, 0.00000464, 0.00000393, 0.00000313, 0.00000216, 0.00000092, -0.00000061, -0.00000241, -0.00000440, -0.00000639, -0.00000817, -0.00000952, -0.00001024, -0.00001019, -0.00000933, -0.00000772, -0.00000553, -0.00000301, -0.00000043, 0.00000193, 0.00000382, 0.00000510, 0.00000570, 0.00000570, 0.00000520, 0.00000441, 0.00000354, 0.00000275, 0.00000218, 0.00000190, 0.00000187, 0.00000203, 0.00000226, 0.00000245, 0.00000249, 0.00000235, 0.00000204, 0.00000162, 0.00000119, 0.00000085, },
		{ -0.00000131, -0.00000252, -0.00000403, -0.00000569, -0.00000729, -0.00000858, -0.00000932, -0.00000933, -0.00000851, -0.00000688, -0.00000461, -0.00000195, 0.00000076, 0.00000316, 0.00000494, 0.00000586, 0.00000583, 0.00000490, 0.00000328, 0.00000125, -0.00000082, -0.00000258, -0.00000375, -0.00000417, -0.00000382, -0.00000283, -0.00000144, 0.00000003, 0.00000126, 0.00000200, 0.00000209, 0.00000153, 0.00000044, -0.00000093, -0.00000225, -0.00000323, -0.00000361, -0.00000326, -0.00000221, -0.00000066, 0.00000107, 0.00000256, 0.00000342, 0.00000331, 0.00000205, -0.00000035, -0.00000366, -0.00000747, -0.00001125, -0.00001439, -0.00001636, -0.00001673, -0.00001530, -0.00001213, -0.00000750, -0.00000192, 0.00000396, 0.00000947, 0.00001397, 0.00001699, 0.00001828, 0.00001782, 0.00001585, 0.00001279, 0.00000918, 0.00000560, 0.00000255, 0.00000042, -0.00000059, -0.00000051, 0.00000049, 0.00000210, 0.00000392, 0.00000557, 0.00000674, 0.00000725, 0.00000702, 0.00000614, 0.00000478, 0.00000316, 0.00000153, 0.00000005, -0.00000116, -0.00000208, -0.00000280, -0.00000342, -0.00000410, -0.00000496, -0.00000603, -0.00000728, -0.00000860, -0.00000978, -0.00001059, -0.00001081, -0.00001027, -0.00000890, -0.00000675, -0.00000398, -0.00000085, 0.00000231, 0.00000516, 0.00000741, 0.00000884, 0.00000936, 0.00000900, 0.00000792, 0.00000634, 0.00000455, 0.00000282, 0.00000138, 0.00000037, -0.00000015, -0.00000022, 0.00000004, 0.00000048, 0.00000094, 0.00000131, 0.00000150, 0.00000149, 0.00000133, 0.00000108, 0.00000083, 0.00000064, 0.00000052, 0.00000044, 0.00000033, 0.00000007, -0.00000045, },
		{ -0.00000841, -0.00000811, -0.00000712, -0.00000553, -0.00000352, -0.00000136, 0.00000066, 0.00000227, 0.00000324, 0.00000348, 0.00000299, 0.00000191, 0.00000048, -0.00000101, -0.00000227, -0.00000304, -0.00000318, -0.00000267, -0.00000160, -0.00000019, 0.00000127, 0.00000250, 0.00000324, 0.00000333, 0.00000274, 0.00000155, -0.00000003, -0.00000172, -0.00000322, -0.00000427, -0.00000473, -0.00000457, -0.00000387, -0.00000288, -0.00000187, -0.00000117, -0.00000103, -0.00000158, -0.00000284, -0.00000461, -0.00000660, -0.00000839, -0.00000955, -0.00000971, -0.00000861, -0.00000618, -0.00000257, 0.00000189, 0.00000670, 0.00001130, 0.00001511, 0.00001765, 0.00001863, 0.00001796, 0.00001580, 0.00001248, 0.00000853, 0.00000452, 0.00000098, -0.00000165, -0.00000309, -0.00000329, -0.00000241, -0.00000077, 0.00000121, 0.00000308, 0.00000443, 0.00000499, 0.00000463, 0.00000341, 0.00000153, -0.00000069, -0.00000291, -0.00000479, -0.00000611, -0.00000675, -0.00000674, -0.00000623, -0.00000546, -0.00000470, -0.00000419, -0.00000409, -0.00000444, -0.00000515, -0.00000603, -0.00000681, -0.00000721, -0.00000698, -0.00000598, -0.00000418, -0.00000170, 0.00000120, 0.00000419, 0.00000690, 0.00000898, 0.00001018, 0.00001035, 0.00000952, 0.00000784, 0.00000558, 0.00000307, 0.00000068, -0.00000130, -0.00000265, -0.00000327, -0.00000318, -0.00000250, -0.00000143, -0.00000020, 0.00000098, 0.00000193, 0.00000257, 0.00000288, 0.00000289, 0.00000269, 0.00000237, 0.00000202, 0.00000166, 0.00000129, 0.00000086, 0.00000027, -0.00000053, -0.00000159, -0.00000289, -0.00000435, -0.00000583, -0.00000713, -0.00000805, },
		{ 0.00000041, 0.00000180, 0.00000259, 0.00000274, 0.00000231, 0.00000144, 0.00000038, -0.00000065, -0.00000140, -0.00000174, -0.00000159, -0.00000102, -0.00000016, 0.00000076, 0.00000150, 0.00000185, 0.00000167, 0.00000093, -0.00000031, -0.00000186, -0.00000347, -0.00000488, -0.00000582, -0.00000614, -0.00000578, -0.00000481, -0.00000340, -0.00000182, -0.00000035, 0.00000075, 0.00000131, 0.00000126, 0.00000070, -0.00000019, -0.00000110, -0.00000171, -0.00000173, -0.00000097, 0.00000062, 0.00000296, 0.00000576, 0.00000867, 0.00001125, 0.00001309, 0.00001387, 0.00001342, 0.00001175, 0.00000905, 0.00000570, 0.00000213, -0.00000116, -0.00000374, -0.00000533, -0.00000578, -0.00000518, -0.00000376, -0.00000190, -0.00000004, 0.00000142, 0.00000215, 0.00000199, 0.00000093, -0.00000084, -0.00000299, -0.00000513, -0.00000687, -0.00000788, -0.00000798, -0.00000714, -0.00000552, -0.00000342, -0.00000122, 0.00000068, 0.00000193, 0.00000233, 0.00000181, 0.00000051, -0.00000131, -0.00000327, -0.00000494, -0.00000596, -0.00000606, -0.00000515, -0.00000328, -0.00000071, 0.00000220, 0.00000502, 0.00000732, 0.00000875, 0.00000913, 0.00000840, 0.00000671, 0.00000434, 0.00000166, -0.00000092, -0.00000303, -0.00000438, -0.00000484, -0.00000441, -0.00000321, -0.00000149, 0.00000046, 0.00000235, 0.00000392, 0.00000503, 0.00000559, 0.00000563, 0.00000525, 0.00000458, 0.00000375, 0.00000287, 0.00000201, 0.00000116, 0.00000029, -0.00000067, -0.00000178, -0.00000305, -0.00000446, -0.00000592, -0.00000727, -0.00000835, -0.00000898, -0.00000903, -0.00000844, -0.00000722, -0.00000550, -0.00000349, -0.00000143, },
		{ -0.00000013, -0.00000110, -0.00000180, -0.00000212, -0.00000203, -0.00000161, -0.00000100, -0.00000039, 0.00000005, 0.00000017, -0.00000008, -0.00000068, -0.00000150, -0.00000235, -0.00000302, -0.00000331, -0.00000307, -0.00000226, -0.00000092, 0.00000077, 0.00000258, 0.00000424, 0.00000550, 0.00000615, 0.00000612, 0.00000545, 0.00000427, 0.00000285, 0.00000147, 0.00000041, -0.00000010, 0.00000003, 0.00000080, 0.00000206, 0.00000353, 0.00000492, 0.00000591, 0.00000624, 0.00000578, 0.00000453, 0.00000264, 0.00000036, -0.00000196, -0.00000398, -0.00000539, -0.00000600, -0.00000577, -0.00000481, -0.00000336, -0.00000174, -0.00000032, 0.00000061, 0.00000083, 0.00000030, -0.00000088, -0.00000247, -0.00000411, -0.00000542, -0.00000606, -0.00000581, -0.00000460, -0.00000253, 0.00000009, 0.00000287, 0.00000533, 0.00000703, 0.00000762, 0.00000694, 0.00000502, 0.00000212, -0.00000136, -0.00000487, -0.00000787, -0.00000988, -0.00001057, -0.00000981, -0.00000768, -0.00000451, -0.00000075, 0.00000304, 0.00000631, 0.00000861, 0.00000964, 0.00000933, 0.00000779, 0.00000534, 0.00000241, -0.00000051, -0.00000296, -0.00000458, -0.00000516, -0.00000467, -0.00000324, -0.00000113, 0.00000130, 0.00000368, 0.00000570, 0.00000711, 0.00000781, 0.00000779, 0.00000716, 0.00000609, 0.00000476, 0.00000337, 0.00000202, 0.00000077, -0.00000037, -0.00000147, -0.00000260, -0.00000381, -0.00000513, -0.00000651, -0.00000785, -0.00000900, -0.00000979, -0.00001007, -0.00000974, -0.00000877, -0.00000724, -0.00000529, -0.00000317, -0.00000111, 0.00000062, 0.00000185, 0.00000245, 0.00000242, 0.00000186, 0.00000094, },
		{ 0.00000055, 0.00000141, 0.00000211, 0.00000252, 0.00000262, 0.00000245, 0.00000212, 0.00000180, 0.00000163, 0.00000176, 0.00000224, 0.00000305, 0.00000408, 0.00000513, 0.00000600, 0.00000647, 0.00000636, 0.00000560, 0.00000423, 0.00000238, 0.00000027, -0.00000181, -0.00000358, -0.00000480, -0.00000531, -0.00000509, -0.00000422, -0.00000290, -0.00000139, 0.00000001, 0.00000106, 0.00000157, 0.00000149, 0.00000087, -0.00000014, -0.00000130, -0.00000236, -0.00000309, -0.00000335, -0.00000308, -0.00000236, -0.00000135, -0.00000029, 0.00000059, 0.00000106, 0.00000103, 0.00000049, -0.00000044, -0.00000151, -0.00000243, -0.00000291, -0.00000273, -0.00000175, -0.00000003, 0.00000224, 0.00000471, 0.00000695, 0.00000851, 0.00000901, 0.00000819, 0.00000603, 0.00000267, -0.00000149, -0.00000593, -0.00001003, -0.00001319, -0.00001492, -0.00001493, -0.00001315, -0.00000979, -0.00000530, -0.00000026, 0.00000465, 0.00000876, 0.00001157, 0.00001275, 0.00001222, 0.00001017, 0.00000700, 0.00000324, -0.00000049, -0.00000363, -0.00000573, -0.00000652, -0.00000595, -0.00000418, -0.00000154, 0.00000154, 0.00000461, 0.00000723, 0.00000910, 0.00001005, 0.00001004, 0.00000919, 0.00000771, 0.00000584, 0.00000384, 0.00000189, 0.00000013, -0.00000142, -0.00000280, -0.00000409, -0.00000537, -0.00000670, -0.00000808, -0.00000946, -0.00001070, -0.00001164, -0.00001211, -0.00001199, -0.00001120, -0.00000977, -0.00000781, -0.00000552, -0.00000316, -0.00000100, 0.00000073, 0.00000187, 0.00000236, 0.00000222, 0.00000161, 0.00000071, -0.00000023, -0.00000100, -0.00000143, -0.00000144, -0.00000104, -0.00000033, },
		{ 0.00000431, 0.00000355, 0.00000289, 0.00000244, 0.00000223, 0.00000222, 0.00000232, 0.00000235, 0.00000216, 0.00000160, 0.00000060, -0.00000083, -0.00000259, -0.00000449, -0.00000630, -0.00000775, -0.00000864, -0.00000880, -0.00000819, -0.00000687, -0.00000503, -0.00000290, -0.00000077, 0.00000108, 0.00000246, 0.00000324, 0.00000340, 0.00000303, 0.00000232, 0.00000147, 0.00000069, 0.00000014, -0.00000008, 0.00000003, 0.00000038, 0.00000083, 0.00000122, 0.00000141, 0.00000131, 0.00000092, 0.00000033, -0.00000028, -0.00000073, -0.00000079, -0.00000035, 0.00000066, 0.00000213, 0.00000387, 0.00000554, 0.00000679, 0.00000726, 0.00000667, 0.00000489, 0.00000198, -0.00000182, -0.00000606, -0.00001020, -0.00001365, -0.00001585, -0.00001642, -0.00001515, -0.00001212, -0.00000764, -0.00000226, 0.00000335, 0.00000846, 0.00001242, 0.00001473, 0.00001513, 0.00001365, 0.00001056, 0.00000640, 0.00000181, -0.00000252, -0.00000594, -0.00000799, -0.00000840, -0.00000719, -0.00000458, -0.00000101, 0.00000298, 0.00000681, 0.00000997, 0.00001209, 0.00001296, 0.00001255, 0.00001103, 0.00000867, 0.00000581, 0.00000280, -0.00000007, -0.00000261, -0.00000475, -0.00000649, -0.00000793, -0.00000919, -0.00001036, -0.00001151, -0.00001263, -0.00001361, -0.00001432, -0.00001459, -0.00001426, -0.00001323, -0.00001150, -0.00000916, -0.00000638, -0.00000342, -0.00000058, 0.00000188, 0.00000374, 0.00000488, 0.00000529, 0.00000506, 0.00000439, 0.00000351, 0.00000267, 0.00000208, 0.00000186, 0.00000207, 0.00000266, 0.00000348, 0.00000437, 0.00000515, 0.00000566, 0.00000580, 0.00000558, 0.00000504, },
		{ -0.00000254, -0.00000256, -0.00000271, -0.00000310, -0.00000378, -0.00000471, -0.00000576, -0.00000677, -0.00000752, -0.00000782, -0.00000754, -0.00000661, -0.00000506, -0.00000305, -0.00000077, 0.00000151, 0.00000353, 0.00000507, 0.00000600, 0.00000626, 0.00000592, 0.00000511, 0.00000403, 0.00000289, 0.00000188, 0.00000112, 0.00000068, 0.00000053, 0.00000059, 0.00000074, 0.00000085, 0.00000084, 0.00000067, 0.00000039, 0.00000007, -0.00000014, -0.00000011, 0.00000024, 0.00000095, 0.00000193, 0.00000302, 0.00000397, 0.00000451, 0.00000437, 0.00000337, 0.00000145, -0.00000130, -0.00000461, -0.00000809, -0.00001125, -0.00001359, -0.00001468, -0.00001422, -0.00001212, -0.00000853, -0.00000383, 0.00000144, 0.00000662, 0.00001102, 0.00001409, 0.00001541, 0.00001483, 0.00001248, 0.00000874, 0.00000418, -0.00000051, -0.00000462, -0.00000756, -0.00000891, -0.00000850, -0.00000642, -0.00000302, 0.00000117, 0.00000553, 0.00000942, 0.00001230, 0.00001381, 0.00001377, 0.00001225, 0.00000949, 0.00000588, 0.00000188, -0.00000208, -0.00000562, -0.00000852, -0.00001066, -0.00001209, -0.00001294, -0.00001339, -0.00001364, -0.00001382, -0.00001399, -0.00001411, -0.00001407, -0.00001370, -0.00001284, -0.00001136, -0.00000922, -0.00000648, -0.00000329, 0.00000009, 0.00000336, 0.00000623, 0.00000843, 0.00000980, 0.00001030, 0.00000999, 0.00000906, 0.00000775, 0.00000634, 0.00000508, 0.00000417, 0.00000370, 0.00000367, 0.00000397, 0.00000443, 0.00000486, 0.00000509, 0.00000497, 0.00000444, 0.00000355, 0.00000239, 0.00000111, -0.00000012, -0.00000116, -0.00000191, -0.00000234, -0.00000252, },
		{ -0.00000386, -0.00000439, -0.00000467, -0.00000455, -0.00000392, -0.00000279, -0.00000121, 0.00000065, 0.00000257, 0.00000434, 0.00000572, 0.00000658, 0.00000683, 0.00000650, 0.00000568, 0.00000454, 0.00000328, 0.00000208, 0.00000111, 0.00000043, 0.00000007, -0.00000001, 0.00000007, 0.00000021, 0.00000031, 0.00000030, 0.00000019, 0.00000000, -0.00000017, -0.00000023, -0.00000010, 0.00000026, 0.00000080, 0.00000143, 0.00000196, 0.00000220, 0.00000195, 0.00000108, -0.00000045, -0.00000256, -0.00000505, -0.00000757, -0.00000976, -0.00001121, -0.00001157, -0.00001065, -0.00000840, -0.00000499, -0.00000075, 0.00000381, 0.00000811, 0.00001158, 0.00001375, 0.00001432, 0.00001320, 0.00001056, 0.00000679, 0.00000245, -0.00000181, -0.00000537, -0.00000772, -0.00000852, -0.00000770, -0.00000542, -0.00000207, 0.00000180, 0.00000557, 0.00000864, 0.00001053, 0.00001096, 0.00000984, 0.00000734, 0.00000380, -0.00000032, -0.00000450, -0.00000827, -0.00001126, -0.00001326, -0.00001423, -0.00001429, -0.00001366, -0.00001263, -0.00001149, -0.00001046, -0.00000965, -0.00000907, -0.00000860, -0.00000806, -0.00000725, -0.00000597, -0.00000413, -0.00000173, 0.00000112, 0.00000417, 0.00000715, 0.00000974, 0.00001164, 0.00001268, 0.00001276, 0.00001195, 0.00001040, 0.00000839, 0.00000623, 0.00000420, 0.00000256, 0.00000144, 0.00000089, 0.00000082, 0.00000108, 0.00000146, 0.00000174, 0.00000176, 0.00000142, 0.00000070, -0.00000032, -0.00000150, -0.00000266, -0.00000363, -0.00000427, -0.00000453, -0.00000442, -0.00000402, -0.00000349, -0.00000298, -0.00000265, -0.00000258, -0.00000281, -0.00000328, },
		{ -0.00000037, 0.00000135, 0.00000308, 0.00000460, 0.00000571, 0.00000630, 0.00000634, 0.00000585, 0.00000498, 0.00000387, 0.00000271, 0.00000166, 0.00000083, 0.00000028, -0.00000001, -0.00000011, -0.00000010, -0.00000008, -0.00000013, -0.00000028, -0.00000051, -0.00000076, -0.00000095, -0.00000102, -0.00000090, -0.00000060, -0.00000018, 0.00000025, 0.00000054, 0.00000052, 0.00000009, -0.00000082, -0.00000216, -0.00000379, -0.00000548, -0.00000692, -0.00000781, -0.00000787, -0.00000695, -0.00000500, -0.00000216, 0.00000128, 0.00000491, 0.00000826, 0.00001084, 0.00001227, 0.00001230, 0.00001089, 0.00000818, 0.00000454, 0.00000045, -0.00000351, -0.00000681, -0.00000900, -0.00000981, -0.00000919, -0.00000730, -0.00000450, -0.00000129, 0.00000178, 0.00000420, 0.00000558, 0.00000570, 0.00000455, 0.00000233, -0.00000061, -0.00000382, -0.00000682, -0.00000920, -0.00001067, -0.00001110, -0.00001056, -0.00000925, -0.00000748, -0.00000561, -0.00000395, -0.00000276, -0.00000215, -0.00000207, -0.00000236, -0.00000278, -0.00000302, -0.00000283, -0.00000203, -0.00000056, 0.00000149, 0.00000392, 0.00000642, 0.00000867, 0.00001033, 0.00001117, 0.00001107, 0.00001006, 0.00000828, 0.00000601, 0.00000356, 0.00000128, -0.00000056, -0.00000178, -0.00000231, -0.00000220, -0.00000163, -0.00000082, -0.00000002, 0.00000056, 0.00000076, 0.00000054, -0.00000006, -0.00000093, -0.00000187, -0.00000270, -0.00000325, -0.00000344, -0.00000324, -0.00000272, -0.00000201, -0.00000129, -0.00000073, -0.00000047, -0.00000058, -0.00000103, -0.00000173, -0.00000253, -0.00000321, -0.00000360, -0.00000354, -0.00000295, -0.00000186, },
		{ 0.00000721, 0.00000652, 0.00000544, 0.00000413, 0.00000279, 0.00000157, 0.00000057, -0.00000017, -0.00000067, -0.00000097, -0.00000117, -0.00000133, -0.00000150, -0.00000170, -0.00000190, -0.00000204, -0.00000205, -0.00000189, -0.00000154, -0.00000103, -0.00000043, 0.00000012, 0.00000050, 0.00000057, 0.00000027, -0.00000042, -0.00000141, -0.00000255, -0.00000362, -0.00000436, -0.00000456, -0.00000406, -0.00000281, -0.00000089, 0.00000149, 0.00000402, 0.00000630, 0.00000797, 0.00000868, 0.00000824, 0.00000659, 0.00000388, 0.00000040, -0.00000341, -0.00000707, -0.00001010, -0.00001209, -0.00001281, -0.00001218, -0.00001033, -0.00000759, -0.00000437, -0.00000116, 0.00000159, 0.00000353, 0.00000445, 0.00000433, 0.00000331, 0.00000169, -0.00000015, -0.00000185, -0.00000308, -0.00000362, -0.00000342, -0.00000257, -0.00000131, 0.00000006, 0.00000122, 0.00000188, 0.00000186, 0.00000114, -0.00000017, -0.00000184, -0.00000353, -0.00000488, -0.00000559, -0.00000545, -0.00000438, -0.00000248, 0.00000002, 0.00000278, 0.00000541, 0.00000753, 0.00000884, 0.00000918, 0.00000854, 0.00000705, 0.00000499, 0.00000270, 0.00000053, -0.00000119, -0.00000226, -0.00000258, -0.00000220, -0.00000128, -0.00000007, 0.00000114, 0.00000211, 0.00000264, 0.00000263, 0.00000212, 0.00000121, 0.00000009, -0.00000102, -0.00000191, -0.00000246, -0.00000260, -0.00000237, -0.00000188, -0.00000130, -0.00000081, -0.00000059, -0.00000073, -0.00000123, -0.00000204, -0.00000299, -0.00000388, -0.00000450, -0.00000467, -0.00000426, -0.00000325, -0.00000172, 0.00000018, 0.00000223, 0.00000417, 0.00000578, 0.00000687, 0.00000735, },
		{ -0.00000132, -0.00000211, -0.00000258, -0.00000277, -0.00000280, -0.00000272, -0.00000261, -0.00000247, -0.00000229, -0.00000203, -0.00000166, -0.00000115, -0.00000051, 0.00000019, 0.00000087, 0.00000141, 0.00000169, 0.00000162, 0.00000117, 0.00000037, -0.00000069, -0.00000182, -0.00000284, -0.00000354, -0.00000376, -0.00000341, -0.00000251, -0.00000119, 0.00000032, 0.00000174, 0.00000276, 0.00000312, 0.00000262, 0.00000121, -0.00000100, -0.00000379, -0.00000679, -0.00000959, -0.00001177, -0.00001299, -0.00001302, -0.00001179, -0.00000941, -0.00000617, -0.00000243, 0.00000134, 0.00000472, 0.00000735, 0.00000900, 0.00000961, 0.00000926, 0.00000817, 0.00000665, 0.00000502, 0.00000356, 0.00000248, 0.00000184, 0.00000158, 0.00000156, 0.00000154, 0.00000127, 0.00000057, -0.00000066, -0.00000240, -0.00000449, -0.00000665, -0.00000855, -0.00000985, -0.00001026, -0.00000961, -0.00000787, -0.00000519, -0.00000186, 0.00000174, 0.00000515, 0.00000796, 0.00000985, 0.00001063, 0.00001028, 0.00000895, 0.00000692, 0.00000456, 0.00000227, 0.00000039, -0.00000081, -0.00000124, -0.00000093, -0.00000004, 0.00000117, 0.00000239, 0.00000334, 0.00000378, 0.00000362, 0.00000286, 0.00000161, 0.00000008, -0.00000148, -0.00000283, -0.00000379, -0.00000426, -0.00000423, -0.00000382, -0.00000318, -0.00000253, -0.00000205, -0.00000189, -0.00000209, -0.00000262, -0.00000336, -0.00000409, -0.00000461, -0.00000471, -0.00000425, -0.00000316, -0.00000150, 0.00000057, 0.00000285, 0.00000506, 0.00000693, 0.00000824, 0.00000887, 0.00000875, 0.00000794, 0.00000660, 0.00000491, 0.00000309, 0.00000135, -0.00000015, },
		{ -0.00000178, -0.00000139, -0.00000106, -0.00000074, -0.00000040, -0.00000003, 0.00000037, 0.00000074, 0.00000101, 0.00000108, 0.00000088, 0.00000035, -0.00000050, -0.00000162, -0.00000288, -0.00000411, -0.00000514, -0.00000579, -0.00000596, -0.00000561, -0.00000480, -0.00000368, -0.00000247, -0.00000141, -0.00000073, -0.00000063, -0.00000116, -0.00000231, -0.00000390, -0.00000567, -0.00000730, -0.00000843, -0.00000878, -0.00000816, -0.00000650, -0.00000389, -0.00000059, 0.00000308, 0.00000670, 0.00000987, 0.00001224, 0.00001358, 0.00001382, 0.00001302, 0.00001136, 0.00000913, 0.00000665, 0.00000422, 0.00000208, 0.00000034, -0.00000098, -0.00000197, -0.00000277, -0.00000358, -0.00000455, -0.00000575, -0.00000716, -0.00000866, -0.00001004, -0.00001101, -0.00001130, -0.00001071, -0.00000911, -0.00000654, -0.00000317, 0.00000069, 0.00000463, 0.00000823, 0.00001107, 0.00001285, 0.00001341, 0.00001274, 0.00001103, 0.00000858, 0.00000579, 0.00000308, 0.00000081, -0.00000074, -0.00000146, -0.00000138, -0.00000066, 0.00000041, 0.00000153, 0.00000237, 0.00000271, 0.00000241, 0.00000148, 0.00000005, -0.00000168, -0.00000341, -0.00000489, -0.00000591, -0.00000632, -0.00000613, -0.00000542, -0.00000440, -0.00000327, -0.00000228, -0.00000161, -0.00000137, -0.00000154, -0.00000203, -0.00000265, -0.00000317, -0.00000336, -0.00000306, -0.00000216, -0.00000068, 0.00000125, 0.00000343, 0.00000558, 0.00000743, 0.00000873, 0.00000931, 0.00000911, 0.00000816, 0.00000660, 0.00000465, 0.00000256, 0.00000055, -0.00000115, -0.00000243, -0.00000322, -0.00000354, -0.00000349, -0.00000317, -0.00000272, -0.00000223, },
		{ -0.00000191, -0.00000193, -0.00000197, -0.00000210, -0.00000241, -0.00000291, -0.00000360, -0.00000443, -0.00000526, -0.00000597, -0.00000640, -0.00000644, -0.00000602, -0.00000516, -0.00000395, -0.00000255, -0.00000115, 0.00000003, 0.00000081, 0.00000107, 0.00000081, 0.00000010, -0.00000086, -0.00000184, -0.00000256, -0.00000277, -0.00000229, -0.00000104, 0.00000092, 0.00000340, 0.00000614, 0.00000879, 0.00001100, 0.00001247, 0.00001301, 0.00001253, 0.00001110, 0.00000890, 0.00000621, 0.00000333, 0.00000057, -0.00000182, -0.00000368, -0.00000498, -0.00000577, -0.00000620, -0.00000644, -0.00000665, -0.00000696, -0.00000740, -0.00000791, -0.00000833, -0.00000848, -0.00000813, -0.00000711, -0.00000534, -0.00000282, 0.00000028, 0.00000371, 0.00000710, 0.00001007, 0.00001228, 0.00001345, 0.00001345, 0.00001231, 0.00001020, 0.00000743, 0.00000440, 0.00000150, -0.00000091, -0.00000255, -0.00000333, -0.00000326, -0.00000254, -0.00000142, -0.00000024, 0.00000069, 0.00000112, 0.00000093, 0.00000011, -0.00000122, -0.00000283, -0.00000443, -0.00000575, -0.00000655, -0.00000672, -0.00000622, -0.00000517, -0.00000377, -0.00000227, -0.00000093, 0.00000002, 0.00000045, 0.00000034, -0.00000021, -0.00000102, -0.00000184, -0.00000241, -0.00000251, -0.00000200, -0.00000086, 0.00000082, 0.00000285, 0.00000496, 0.00000686, 0.00000828, 0.00000901, 0.00000896, 0.00000813, 0.00000665, 0.00000473, 0.00000263, 0.00000059, -0.00000114, -0.00000242, -0.00000318, -0.00000344, -0.00000328, -0.00000286, -0.00000233, -0.00000183, -0.00000147, -0.00000128, -0.00000128, -0.00000140, -0.00000158, -0.00000175, -0.00000186, },
		{ -0.00000270, -0.00000277, -0.00000289, -0.00000294, -0.00000281, -0.00000240, -0.00000169, -0.00000070, 0.00000049, 0.00000171, 0.00000281, 0.00000362, 0.00000401, 0.00000393, 0.00000341, 0.00000256, 0.00000157, 0.00000067, 0.00000008, -0.00000003, 0.00000044, 0.00000150, 0.00000302, 0.00000481, 0.00000659, 0.00000809, 0.00000905, 0.00000929, 0.00000872, 0.00000739, 0.00000544, 0.00000313, 0.00000071, -0.00000154, -0.00000339, -0.00000469, -0.00000542, -0.00000563, -0.00000547, -0.00000513, -0.00000479, -0.00000459, -0.00000462, -0.00000483, -0.00000514, -0.00000535, -0.00000526, -0.00000470, -0.00000354, -0.00000175, 0.00000056, 0.00000320, 0.00000587, 0.00000825, 0.00001001, 0.00001092, 0.00001082, 0.00000972, 0.00000777, 0.00000524, 0.00000246, -0.00000018, -0.00000236, -0.00000382, -0.00000445, -0.00000428, -0.00000345, -0.00000224, -0.00000095, 0.00000011, 0.00000070, 0.00000068, 0.00000005, -0.00000108, -0.00000249, -0.00000392, -0.00000508, -0.00000575, -0.00000580, -0.00000520, -0.00000405, -0.00000257, -0.00000102, 0.00000033, 0.00000124, 0.00000158, 0.00000132, 0.00000054, -0.00000055, -0.00000169, -0.00000258, -0.00000298, -0.00000271, -0.00000172, -0.00000008, 0.00000202, 0.00000430, 0.00000646, 0.00000818, 0.00000922, 0.00000945, 0.00000882, 0.00000745, 0.00000554, 0.00000333, 0.00000112, -0.00000085, -0.00000240, -0.00000343, -0.00000394, -0.00000402, -0.00000382, -0.00000351, -0.00000324, -0.00000313, -0.00000324, -0.00000355, -0.00000400, -0.00000448, -0.00000488, -0.00000511, -0.00000510, -0.00000486, -0.00000444, -0.00000392, -0.00000341, -0.00000301, -0.00000276, },
		{ 0.00000166, 0.00000204, 0.00000257, 0.00000312, 0.00000358, 0.00000380, 0.00000371, 0.00000328, 0.00000257, 0.00000169, 0.00000081, 0.00000011, -0.00000024, -0.00000013, 0.00000048, 0.00000153, 0.00000288, 0.00000432, 0.00000562, 0.00000655, 0.00000694, 0.00000670, 0.00000582, 0.00000442, 0.00000266, 0.00000079, -0.00000098, -0.00000243, -0.00000345, -0.00000400, -0.00000413, -0.00000397, -0.00000368, -0.00000346, -0.00000342, -0.00000366, -0.00000414, -0.00000477, -0.00000537, -0.00000573, -0.00000565, -0.00000498, -0.00000366, -0.00000174, 0.00000061, 0.00000313, 0.00000552, 0.00000745, 0.00000866, 0.00000897, 0.00000835, 0.00000689, 0.00000481, 0.00000242, 0.00000005, -0.00000196, -0.00000336, -0.00000400, -0.00000390, -0.00000316, -0.00000201, -0.00000075, 0.00000034, 0.00000102, 0.00000113, 0.00000065, -0.00000034, -0.00000164, -0.00000300, -0.00000414, -0.00000485, -0.00000498, -0.00000449, -0.00000348, -0.00000213, -0.00000069, 0.00000056, 0.00000139, 0.00000166, 0.00000133, 0.00000048, -0.00000069, -0.00000192, -0.00000290, -0.00000337, -0.00000315, -0.00000216, -0.00000047, 0.00000174, 0.00000418, 0.00000651, 0.00000841, 0.00000959, 0.00000989, 0.00000926, 0.00000779, 0.00000567, 0.00000317, 0.00000060, -0.00000175, -0.00000367, -0.00000503, -0.00000580, -0.00000607, -0.00000597, -0.00000567, -0.00000537, -0.00000518, -0.00000519, -0.00000538, -0.00000568, -0.00000597, -0.00000612, -0.00000600, -0.00000554, -0.00000472, -0.00000360, -0.00000230, -0.00000095, 0.00000029, 0.00000128, 0.00000193, 0.00000224, 0.00000224, 0.00000204, 0.00000176, 0.00000154, 0.00000149, },
		},
		{ { -0.00000075, -0.00000055, -0.00000006, 0.00000048, 0.00000074, 0.00000048, -0.00000040, -0.00000181, -0.00000349, -0.00000514, -0.00000648, -0.00000733, -0.00000760, -0.00000726, -0.00000627, -0.00000454, -0.00000197, 0.00000144, 0.00000553, 0.00000985, 0.00001367, 0.00001613, 0.00001645, 0.00001421, 0.00000955, 0.00000327, -0.00000335, -0.00000879, -0.00001180, -0.00001167, -0.00000856, -0.00000339, 0.00000236, 0.00000709, 0.00000959, 0.00000934, 0.00000665, 0.00000252, -0.00000176, -0.00000501, -0.00000654, -0.00000634, -0.00000496, -0.00000329, -0.00000211, -0.00000188, -0.00000249, -0.00000340, -0.00000390, -0.00000341, -0.00000179, 0.00000060, 0.00000299, 0.00000451, 0.00000456, 0.00000304, 0.00000043, -0.00000232, -0.00000416, -0.00000433, -0.00000264, 0.00000045, 0.00000394, 0.00000666, 0.00000768, 0.00000666, 0.00000394, 0.00000045, -0.00000264, -0.00000433, -0.00000416, -0.00000232, 0.00000043, 0.00000304, 0.00000456, 0.00000451, 0.00000299, 0.00000060, -0.00000179, -0.00000341, -0.00000390, -0.00000340, -0.00000249, -0.00000188, -0.00000211, -0.00000329, -0.00000496, -0.00000634, -0.00000654, -0.00000501, -0.00000176, 0.00000252, 0.00000665, 0.00000934, 0.00000959, 0.00000709, 0.00000236, -0.00000339, -0.00000856, -0.00001167, -0.00001180, -0.00000879, -0.00000335, 0.00000327, 0.00000955, 0.00001421, 0.00001645, 0.00001613, 0.00001367, 0.00000985, 0.00000553, 0.00000144, -0.00000197, -0.00000454, -0.00000627, -0.00000726, -0.00000760, -0.00000733, -0.00000648, -0.00000514, -0.00000349, -0.00000181, -0.00000040, 0.00000048, 0.00000074, 0.00000048, -0.00000006, -0.00000055, },
		{ -0.00000105, -0.00000115, -0.00000174, -0.00000274, -0.00000391, -0.00000497, -0.00000566, -0.00000577, -0.00000521, -0.00000393, -0.00000193, 0.00000076, 0.00000404, 0.00000767, 0.00001124, 0.00001416, 0.00001576, 0.00001548, 0.00001305, 0.00000865, 0.00000297, -0.00000289, -0.00000768, -0.00001037, -0.00001040, -0.00000794, -0.00000383, 0.00000063, 0.00000408, 0.00000551, 0.00000457, 0.00000164, -0.00000225, -0.00000584, -0.00000801, -0.00000816, -0.00000639, -0.00000341, -0.00000025, 0.00000210, 0.00000301, 0.00000247, 0.00000097, -0.00000068, -0.00000171, -0.00000169, -0.00000069, 0.00000082, 0.00000212, 0.00000261, 0.00000206, 0.00000072, -0.00000077, -0.00000166, -0.00000138, 0.00000018, 0.00000258, 0.00000498, 0.00000642, 0.00000623, 0.00000427, 0.00000105, -0.00000242, -0.00000499, -0.00000577, -0.00000444, -0.00000144, 0.00000225, 0.00000536, 0.00000680, 0.00000605, 0.00000328, -0.00000068, -0.00000470, -0.00000769, -0.00000901, -0.00000864, -0.00000711, -0.00000522, -0.00000374, -0.00000303, -0.00000294, -0.00000289, -0.00000213, -0.00000015, 0.00000307, 0.00000695, 0.00001040, 0.00001223, 0.00001155, 0.00000813, 0.00000258, -0.00000379, -0.00000935, -0.00001255, -0.00001250, -0.00000913, -0.00000329, 0.00000358, 0.00000983, 0.00001409, 0.00001560, 0.00001434, 0.00001095, 0.00000645, 0.00000188, -0.00000196, -0.00000467, -0.00000626, -0.00000697, -0.00000713, -0.00000696, -0.00000652, -0.00000576, -0.00000461, -0.00000308, -0.00000134, 0.00000030, 0.00000150, 0.00000200, 0.00000173, 0.00000086, -0.00000029, -0.00000134, -0.00000199, -0.00000211, -0.00000181, -0.00000135, },
		{ -0.00000188, -0.00000267, -0.00000306, -0.00000286, -0.00000194, -0.00000029, 0.00000202, 0.00000484, 0.00000791, 0.00001086, 0.00001322, 0.00001445, 0.00001412, 0.00001200, 0.00000822, 0.00000331, -0.00000187, -0.00000632, -0.00000917, -0.00000993, -0.00000867, -0.00000602, -0.00000297, -0.00000059, 0.00000033, -0.00000047, -0.00000261, -0.00000521, -0.00000722, -0.00000772, -0.00000634, -0.00000331, 0.00000056, 0.00000417, 0.00000649, 0.00000693, 0.00000551, 0.00000285, -0.00000009, -0.00000233, -0.00000323, -0.00000265, -0.00000100, 0.00000099, 0.00000257, 0.00000325, 0.00000297, 0.00000209, 0.00000121, 0.00000085, 0.00000125, 0.00000223, 0.00000323, 0.00000363, 0.00000293, 0.00000111, -0.00000139, -0.00000374, -0.00000502, -0.00000460, -0.00000242, 0.00000095, 0.00000442, 0.00000675, 0.00000698, 0.00000477, 0.00000051, -0.00000472, -0.00000955, -0.00001274, -0.00001356, -0.00001204, -0.00000885, -0.00000505, -0.00000168, 0.00000062, 0.00000177, 0.00000228, 0.00000289, 0.00000424, 0.00000655, 0.00000943, 0.00001197, 0.00001309, 0.00001193, 0.00000825, 0.00000259, -0.00000382, -0.00000938, -0.00001262, -0.00001266, -0.00000949, -0.00000398, 0.00000240, 0.00000799, 0.00001149, 0.00001224, 0.00001040, 0.00000677, 0.00000251, -0.00000128, -0.00000386, -0.00000504, -0.00000508, -0.00000455, -0.00000400, -0.00000376, -0.00000388, -0.00000409, -0.00000405, -0.00000349, -0.00000237, -0.00000092, 0.00000044, 0.00000128, 0.00000132, 0.00000055, -0.00000078, -0.00000222, -0.00000331, -0.00000374, -0.00000342, -0.00000255, -0.00000147, -0.00000058, -0.00000017, -0.00000035, -0.00000100, },
		{ -0.00000012, 0.00000180, 0.00000431, 0.00000713, 0.00000988, 0.00001211, 0.00001335, 0.00001321, 0.00001146, 0.00000817, 0.00000371, -0.00000123, -0.00000581, -0.00000924, -0.00001101, -0.00001100, -0.00000957, -0.00000742, -0.00000538, -0.00000409, -0.00000383, -0.00000439, -0.00000518, -0.00000543, -0.00000453, -0.00000225, 0.00000109, 0.00000472, 0.00000766, 0.00000906, 0.00000847, 0.00000606, 0.00000255, -0.00000100, -0.00000354, -0.00000438, -0.00000339, -0.00000105, 0.00000180, 0.00000421, 0.00000549, 0.00000538, 0.00000415, 0.00000238, 0.00000075, -0.00000023, -0.00000046, -0.00000018, 0.00000014, 0.00000006, -0.00000058, -0.00000160, -0.00000252, -0.00000273, -0.00000187, 0.00000004, 0.00000248, 0.00000453, 0.00000525, 0.00000395, 0.00000055, -0.00000433, -0.00000955, -0.00001373, -0.00001570, -0.00001493, -0.00001162, -0.00000669, -0.00000141, 0.00000300, 0.00000577, 0.00000682, 0.00000672, 0.00000637, 0.00000662, 0.00000790, 0.00000999, 0.00001209, 0.00001311, 0.00001210, 0.00000867, 0.00000319, -0.00000323, -0.00000907, -0.00001287, -0.00001370, -0.00001146, -0.00000689, -0.00000135, 0.00000361, 0.00000678, 0.00000757, 0.00000621, 0.00000352, 0.00000063, -0.00000146, -0.00000224, -0.00000176, -0.00000058, 0.00000051, 0.00000079, -0.00000003, -0.00000182, -0.00000398, -0.00000579, -0.00000661, -0.00000617, -0.00000464, -0.00000257, -0.00000065, 0.00000051, 0.00000062, -0.00000022, -0.00000157, -0.00000284, -0.00000350, -0.00000326, -0.00000218, -0.00000058, 0.00000103, 0.00000219, 0.00000259, 0.00000220, 0.00000121, -0.00000001, -0.00000105, -0.00000153, -0.00000125, },
		{ 0.00001140, 0.00001267, 0.00001258, 0.00001091, 0.00000771, 0.00000329, -0.00000176, -0.00000668, -0.00001070, -0.00001321, -0.00001396, -0.00001309, -0.00001107, -0.00000859, -0.00000627, -0.00000452, -0.00000338, -0.00000257, -0.00000161, -0.00000011, 0.00000209, 0.00000476, 0.00000733, 0.00000907, 0.00000937, 0.00000798, 0.00000515, 0.00000161, -0.00000164, -0.00000369, -0.00000395, -0.00000240, 0.00000039, 0.00000350, 0.00000590, 0.00000686, 0.00000613, 0.00000399, 0.00000119, -0.00000143, -0.00000316, -0.00000367, -0.00000309, -0.00000190, -0.00000068, 0.00000011, 0.00000034, 0.00000017, -0.00000004, 0.00000007, 0.00000057, 0.00000123, 0.00000152, 0.00000084, -0.00000119, -0.00000451, -0.00000854, -0.00001228, -0.00001461, -0.00001468, -0.00001220, -0.00000759, -0.00000188, 0.00000363, 0.00000778, 0.00000989, 0.00001005, 0.00000895, 0.00000767, 0.00000718, 0.00000801, 0.00000996, 0.00001217, 0.00001347, 0.00001274, 0.00000942, 0.00000376, -0.00000321, -0.00000997, -0.00001494, -0.00001701, -0.00001588, -0.00001210, -0.00000690, -0.00000171, 0.00000222, 0.00000428, 0.00000455, 0.00000372, 0.00000272, 0.00000233, 0.00000289, 0.00000411, 0.00000529, 0.00000556, 0.00000427, 0.00000130, -0.00000287, -0.00000727, -0.00001078, -0.00001245, -0.00001187, -0.00000927, -0.00000546, -0.00000151, 0.00000155, 0.00000307, 0.00000295, 0.00000164, -0.00000012, -0.00000150, -0.00000193, -0.00000124, 0.00000033, 0.00000223, 0.00000382, 0.00000460, 0.00000435, 0.00000320, 0.00000154, -0.00000011, -0.00000125, -0.00000152, -0.00000075, 0.00000100, 0.00000348, 0.00000634, 0.00000914, },
		{ 0.00000172, -0.00000345, -0.00000851, -0.00001269, -0.00001537, -0.00001620, -0.00001520, -0.00001273, -0.00000944, -0.00000599, -0.00000294, -0.00000057, 0.00000115, 0.00000250, 0.00000378, 0.00000520, 0.00000671, 0.00000803, 0.00000870, 0.00000834, 0.00000679, 0.00000429, 0.00000141, -0.00000107, -0.00000243, -0.00000227, -0.00000065, 0.00000188, 0.00000442, 0.00000609, 0.00000623, 0.00000472, 0.00000195, -0.00000128, -0.00000402, -0.00000551, -0.00000543, -0.00000397, -0.00000171, 0.00000054, 0.00000210, 0.00000259, 0.00000204, 0.00000082, -0.00000060, -0.00000182, -0.00000272, -0.00000347, -0.00000437, -0.00000573, -0.00000756, -0.00000956, -0.00001108, -0.00001141, -0.00001001, -0.00000677, -0.00000214, 0.00000298, 0.00000752, 0.00001053, 0.00001160, 0.00001093, 0.00000932, 0.00000781, 0.00000734, 0.00000832, 0.00001045, 0.00001279, 0.00001403, 0.00001301, 0.00000913, 0.00000262, -0.00000542, -0.00001333, -0.00001939, -0.00002230, -0.00002156, -0.00001761, -0.00001161, -0.00000506, 0.00000067, 0.00000474, 0.00000702, 0.00000793, 0.00000821, 0.00000849, 0.00000900, 0.00000945, 0.00000918, 0.00000747, 0.00000389, -0.00000140, -0.00000759, -0.00001341, -0.00001745, -0.00001866, -0.00001666, -0.00001190, -0.00000556, 0.00000082, 0.00000581, 0.00000842, 0.00000845, 0.00000643, 0.00000338, 0.00000049, -0.00000129, -0.00000150, -0.00000027, 0.00000179, 0.00000385, 0.00000518, 0.00000533, 0.00000430, 0.00000245, 0.00000039, -0.00000123, -0.00000189, -0.00000136, 0.00000032, 0.00000286, 0.00000581, 0.00000863, 0.00001082, 0.00001190, 0.00001157, 0.00000968, 0.00000628, },
		{ -0.00001683, -0.00001532, -0.00001226, -0.00000826, -0.00000405, -0.00000028, 0.00000263, 0.00000456, 0.00000568, 0.00000630, 0.00000675, 0.00000717, 0.00000754, 0.00000761, 0.00000714, 0.00000598, 0.00000420, 0.00000217, 0.00000039, -0.00000061, -0.00000056, 0.00000051, 0.00000215, 0.00000370, 0.00000447, 0.00000398, 0.00000220, -0.00000047, -0.00000325, -0.00000532, -0.00000603, -0.00000517, -0.00000304, -0.00000034, 0.00000203, 0.00000330, 0.00000310, 0.00000149, -0.00000103, -0.00000374, -0.00000598, -0.00000735, -0.00000781, -0.00000763, -0.00000722, -0.00000688, -0.00000668, -0.00000638, -0.00000560, -0.00000393, -0.00000126, 0.00000219, 0.00000583, 0.00000894, 0.00001087, 0.00001136, 0.00001062, 0.00000929, 0.00000824, 0.00000815, 0.00000926, 0.00001116, 0.00001290, 0.00001326, 0.00001118, 0.00000620, -0.00000130, -0.00001012, -0.00001850, -0.00002464, -0.00002718, -0.00002560, -0.00002035, -0.00001270, -0.00000429, 0.00000331, 0.00000902, 0.00001249, 0.00001398, 0.00001412, 0.00001355, 0.00001259, 0.00001113, 0.00000874, 0.00000494, -0.00000044, -0.00000703, -0.00001384, -0.00001946, -0.00002249, -0.00002195, -0.00001768, -0.00001044, -0.00000178, 0.00000640, 0.00001235, 0.00001496, 0.00001405, 0.00001037, 0.00000531, 0.00000047, -0.00000282, -0.00000386, -0.00000272, -0.00000013, 0.00000284, 0.00000514, 0.00000606, 0.00000541, 0.00000355, 0.00000117, -0.00000091, -0.00000202, -0.00000181, -0.00000027, 0.00000224, 0.00000518, 0.00000792, 0.00000988, 0.00001063, 0.00000990, 0.00000766, 0.00000405, -0.00000055, -0.00000558, -0.00001036, -0.00001418, -0.00001645, },
		{ 0.00000243, 0.00000529, 0.00000697, 0.00000759, 0.00000752, 0.00000716, 0.00000684, 0.00000667, 0.00000658, 0.00000634, 0.00000572, 0.00000466, 0.00000326, 0.00000183, 0.00000073, 0.00000025, 0.00000043, 0.00000107, 0.00000175, 0.00000197, 0.00000141, 0.00000003, -0.00000188, -0.00000374, -0.00000493, -0.00000500, -0.00000385, -0.00000182, 0.00000040, 0.00000201, 0.00000236, 0.00000118, -0.00000131, -0.00000447, -0.00000746, -0.00000949, -0.00001009, -0.00000926, -0.00000737, -0.00000506, -0.00000293, -0.00000136, -0.00000038, 0.00000030, 0.00000109, 0.00000235, 0.00000417, 0.00000636, 0.00000845, 0.00000998, 0.00001066, 0.00001050, 0.00000985, 0.00000925, 0.00000914, 0.00000966, 0.00001045, 0.00001077, 0.00000968, 0.00000639, 0.00000068, -0.00000697, -0.00001534, -0.00002276, -0.00002756, -0.00002854, -0.00002533, -0.00001850, -0.00000944, 0.00000010, 0.00000843, 0.00001439, 0.00001755, 0.00001815, 0.00001689, 0.00001455, 0.00001165, 0.00000835, 0.00000444, -0.00000034, -0.00000604, -0.00001223, -0.00001797, -0.00002199, -0.00002305, -0.00002041, -0.00001416, -0.00000530, 0.00000439, 0.00001287, 0.00001832, 0.00001970, 0.00001700, 0.00001126, 0.00000423, -0.00000217, -0.00000635, -0.00000752, -0.00000579, -0.00000210, 0.00000217, 0.00000566, 0.00000740, 0.00000708, 0.00000506, 0.00000218, -0.00000055, -0.00000227, -0.00000247, -0.00000115, 0.00000130, 0.00000421, 0.00000684, 0.00000857, 0.00000895, 0.00000783, 0.00000527, 0.00000157, -0.00000285, -0.00000742, -0.00001152, -0.00001457, -0.00001608, -0.00001579, -0.00001373, -0.00001024, -0.00000591, -0.00000146, },
		{ 0.00000687, 0.00000616, 0.00000580, 0.00000576, 0.00000581, 0.00000566, 0.00000508, 0.00000402, 0.00000263, 0.00000122, 0.00000011, -0.00000050, -0.00000061, -0.00000042, -0.00000026, -0.00000043, -0.00000106, -0.00000204, -0.00000306, -0.00000372, -0.00000371, -0.00000297, -0.00000176, -0.00000057, 0.00000001, -0.00000047, -0.00000212, -0.00000464, -0.00000738, -0.00000954, -0.00001041, -0.00000963, -0.00000733, -0.00000406, -0.00000060, 0.00000227, 0.00000405, 0.00000465, 0.00000441, 0.00000392, 0.00000376, 0.00000433, 0.00000566, 0.00000743, 0.00000916, 0.00001037, 0.00001078, 0.00001045, 0.00000963, 0.00000869, 0.00000784, 0.00000700, 0.00000578, 0.00000361, 0.00000000, -0.00000515, -0.00001143, -0.00001784, -0.00002305, -0.00002572, -0.00002496, -0.00002055, -0.00001316, -0.00000413, 0.00000488, 0.00001230, 0.00001705, 0.00001878, 0.00001785, 0.00001505, 0.00001131, 0.00000731, 0.00000338, -0.00000056, -0.00000471, -0.00000917, -0.00001364, -0.00001741, -0.00001946, -0.00001882, -0.00001496, -0.00000810, 0.00000069, 0.00000970, 0.00001696, 0.00002082, 0.00002043, 0.00001601, 0.00000881, 0.00000075, -0.00000610, -0.00001011, -0.00001055, -0.00000769, -0.00000271, 0.00000276, 0.00000711, 0.00000925, 0.00000883, 0.00000632, 0.00000272, -0.00000075, -0.00000304, -0.00000357, -0.00000233, 0.00000017, 0.00000314, 0.00000572, 0.00000721, 0.00000723, 0.00000572, 0.00000291, -0.00000078, -0.00000482, -0.00000867, -0.00001181, -0.00001381, -0.00001437, -0.00001336, -0.00001088, -0.00000728, -0.00000307, 0.00000109, 0.00000460, 0.00000703, 0.00000823, 0.00000833, 0.00000773, },
		{ 0.00000558, 0.00000527, 0.00000434, 0.00000287, 0.00000114, -0.00000046, -0.00000161, -0.00000214, -0.00000211, -0.00000176, -0.00000141, -0.00000133, -0.00000162, -0.00000220, -0.00000282, -0.00000324, -0.00000332, -0.00000314, -0.00000293, -0.00000304, -0.00000371, -0.00000500, -0.00000665, -0.00000815, -0.00000889, -0.00000837, -0.00000641, -0.00000323, 0.00000054, 0.00000411, 0.00000672, 0.00000791, 0.00000767, 0.00000645, 0.00000494, 0.00000391, 0.00000384, 0.00000486, 0.00000662, 0.00000854, 0.00000994, 0.00001033, 0.00000958, 0.00000788, 0.00000568, 0.00000343, 0.00000138, -0.00000048, -0.00000245, -0.00000489, -0.00000803, -0.00001174, -0.00001547, -0.00001835, -0.00001946, -0.00001810, -0.00001410, -0.00000796, -0.00000072, 0.00000627, 0.00001174, 0.00001483, 0.00001531, 0.00001355, 0.00001037, 0.00000665, 0.00000311, 0.00000007, -0.00000256, -0.00000505, -0.00000763, -0.00001023, -0.00001238, -0.00001333, -0.00001228, -0.00000879, -0.00000302, 0.00000415, 0.00001124, 0.00001658, 0.00001877, 0.00001716, 0.00001207, 0.00000475, -0.00000292, -0.00000900, -0.00001200, -0.00001135, -0.00000749, -0.00000171, 0.00000423, 0.00000870, 0.00001060, 0.00000968, 0.00000652, 0.00000230, -0.00000163, -0.00000415, -0.00000468, -0.00000330, -0.00000062, 0.00000240, 0.00000483, 0.00000597, 0.00000551, 0.00000357, 0.00000056, -0.00000295, -0.00000637, -0.00000921, -0.00001111, -0.00001187, -0.00001141, -0.00000978, -0.00000715, -0.00000384, -0.00000025, 0.00000313, 0.00000585, 0.00000758, 0.00000823, 0.00000795, 0.00000708, 0.00000606, 0.00000529, 0.00000498, 0.00000510, 0.00000542, },
		{ -0.00000116, -0.00000262, -0.00000334, -0.00000331, -0.00000273, -0.00000198, -0.00000144, -0.00000137, -0.00000184, -0.00000272, -0.00000372, -0.00000460, -0.00000521, -0.00000554, -0.00000573, -0.00000594, -0.00000624, -0.00000652, -0.00000653, -0.00000591, -0.00000437, -0.00000187, 0.00000135, 0.00000476, 0.00000768, 0.00000950, 0.00000989, 0.00000892, 0.00000707, 0.00000506, 0.00000362, 0.00000324, 0.00000399, 0.00000552, 0.00000714, 0.00000813, 0.00000794, 0.00000643, 0.00000389, 0.00000090, -0.00000185, -0.00000386, -0.00000496, -0.00000538, -0.00000560, -0.00000614, -0.00000730, -0.00000902, -0.00001082, -0.00001200, -0.00001188, -0.00001005, -0.00000659, -0.00000207, 0.00000262, 0.00000653, 0.00000892, 0.00000954, 0.00000859, 0.00000667, 0.00000446, 0.00000253, 0.00000110, 0.00000005, -0.00000097, -0.00000229, -0.00000399, -0.00000578, -0.00000703, -0.00000702, -0.00000524, -0.00000162, 0.00000326, 0.00000828, 0.00001210, 0.00001354, 0.00001200, 0.00000769, 0.00000166, -0.00000453, -0.00000919, -0.00001110, -0.00000979, -0.00000573, -0.00000016, 0.00000531, 0.00000916, 0.00001045, 0.00000904, 0.00000558, 0.00000128, -0.00000254, -0.00000482, -0.00000508, -0.00000351, -0.00000083, 0.00000196, 0.00000393, 0.00000447, 0.00000344, 0.00000115, -0.00000181, -0.00000478, -0.00000718, -0.00000866, -0.00000909, -0.00000855, -0.00000721, -0.00000530, -0.00000301, -0.00000054, 0.00000188, 0.00000402, 0.00000565, 0.00000662, 0.00000686, 0.00000649, 0.00000576, 0.00000501, 0.00000452, 0.00000447, 0.00000480, 0.00000529, 0.00000556, 0.00000531, 0.00000435, 0.00000274, 0.00000077, },
		{ -0.00000241, -0.00000163, -0.00000145, -0.00000199, -0.00000316, -0.00000465, -0.00000605, -0.00000703, -0.00000742, -0.00000722, -0.00000656, -0.00000563, -0.00000452, -0.00000323, -0.00000167, 0.00000030, 0.00000268, 0.00000531, 0.00000782, 0.00000975, 0.00001067, 0.00001036, 0.00000893, 0.00000678, 0.00000454, 0.00000283, 0.00000208, 0.00000232, 0.00000323, 0.00000419, 0.00000454, 0.00000385, 0.00000203, -0.00000054, -0.00000319, -0.00000517, -0.00000593, -0.00000535, -0.00000377, -0.00000192, -0.00000064, -0.00000057, -0.00000194, -0.00000446, -0.00000739, -0.00000979, -0.00001086, -0.00001017, -0.00000777, -0.00000423, -0.00000035, 0.00000308, 0.00000551, 0.00000679, 0.00000712, 0.00000691, 0.00000654, 0.00000620, 0.00000579, 0.00000505, 0.00000367, 0.00000154, -0.00000109, -0.00000366, -0.00000542, -0.00000571, -0.00000424, -0.00000125, 0.00000249, 0.00000587, 0.00000781, 0.00000761, 0.00000519, 0.00000120, -0.00000320, -0.00000667, -0.00000813, -0.00000712, -0.00000393, 0.00000050, 0.00000484, 0.00000784, 0.00000872, 0.00000734, 0.00000429, 0.00000059, -0.00000261, -0.00000443, -0.00000452, -0.00000312, -0.00000093, 0.00000115, 0.00000235, 0.00000223, 0.00000085, -0.00000135, -0.00000370, -0.00000556, -0.00000649, -0.00000638, -0.00000540, -0.00000389, -0.00000223, -0.00000072, 0.00000053, 0.00000151, 0.00000232, 0.00000301, 0.00000361, 0.00000409, 0.00000439, 0.00000449, 0.00000446, 0.00000442, 0.00000449, 0.00000476, 0.00000518, 0.00000555, 0.00000562, 0.00000513, 0.00000395, 0.00000215, 0.00000001, -0.00000204, -0.00000358, -0.00000431, -0.00000418, -0.00000341, },
		{ -0.00000557, -0.00000717, -0.00000814, -0.00000821, -0.00000733, -0.00000567, -0.00000355, -0.00000128, 0.00000091, 0.00000292, 0.00000479, 0.00000654, 0.00000815, 0.00000949, 0.00001030, 0.00001035, 0.00000947, 0.00000773, 0.00000544, 0.00000306, 0.00000108, -0.00000012, -0.00000046, -0.00000018, 0.00000028, 0.00000041, -0.00000013, -0.00000136, -0.00000295, -0.00000428, -0.00000475, -0.00000395, -0.00000190, 0.00000088, 0.00000351, 0.00000500, 0.00000458, 0.00000199, -0.00000239, -0.00000761, -0.00001243, -0.00001563, -0.00001638, -0.00001446, -0.00001030, -0.00000481, 0.00000094, 0.00000598, 0.00000972, 0.00001199, 0.00001297, 0.00001300, 0.00001239, 0.00001125, 0.00000950, 0.00000700, 0.00000371, -0.00000017, -0.00000411, -0.00000735, -0.00000916, -0.00000904, -0.00000696, -0.00000347, 0.00000050, 0.00000383, 0.00000560, 0.00000537, 0.00000335, 0.00000028, -0.00000275, -0.00000471, -0.00000493, -0.00000335, -0.00000051, 0.00000267, 0.00000517, 0.00000625, 0.00000564, 0.00000365, 0.00000097, -0.00000153, -0.00000317, -0.00000361, -0.00000299, -0.00000182, -0.00000072, -0.00000024, -0.00000059, -0.00000164, -0.00000295, -0.00000397, -0.00000425, -0.00000361, -0.00000220, -0.00000044, 0.00000117, 0.00000220, 0.00000245, 0.00000201, 0.00000119, 0.00000039, -0.00000002, 0.00000015, 0.00000086, 0.00000196, 0.00000317, 0.00000426, 0.00000510, 0.00000565, 0.00000594, 0.00000598, 0.00000576, 0.00000519, 0.00000417, 0.00000267, 0.00000078, -0.00000126, -0.00000310, -0.00000438, -0.00000488, -0.00000456, -0.00000365, -0.00000256, -0.00000178, -0.00000167, -0.00000239, -0.00000381, },
		{ -0.00000342, -0.00000032, 0.00000272, 0.00000531, 0.00000724, 0.00000849, 0.00000918, 0.00000948, 0.00000946, 0.00000911, 0.00000832, 0.00000700, 0.00000515, 0.00000291, 0.00000060, -0.00000141, -0.00000279, -0.00000339, -0.00000330, -0.00000281, -0.00000228, -0.00000200, -0.00000203, -0.00000220, -0.00000210, -0.00000134, 0.00000031, 0.00000270, 0.00000527, 0.00000717, 0.00000749, 0.00000559, 0.00000135, -0.00000468, -0.00001135, -0.00001718, -0.00002079, -0.00002121, -0.00001818, -0.00001225, -0.00000454, 0.00000352, 0.00001060, 0.00001575, 0.00001861, 0.00001932, 0.00001836, 0.00001629, 0.00001352, 0.00001025, 0.00000648, 0.00000222, -0.00000241, -0.00000698, -0.00001083, -0.00001320, -0.00001350, -0.00001151, -0.00000758, -0.00000259, 0.00000230, 0.00000593, 0.00000751, 0.00000682, 0.00000433, 0.00000095, -0.00000217, -0.00000409, -0.00000429, -0.00000286, -0.00000041, 0.00000220, 0.00000410, 0.00000475, 0.00000403, 0.00000227, 0.00000006, -0.00000197, -0.00000336, -0.00000397, -0.00000393, -0.00000354, -0.00000313, -0.00000284, -0.00000265, -0.00000233, -0.00000164, -0.00000044, 0.00000117, 0.00000288, 0.00000421, 0.00000474, 0.00000419, 0.00000264, 0.00000045, -0.00000176, -0.00000337, -0.00000391, -0.00000318, -0.00000137, 0.00000109, 0.00000364, 0.00000575, 0.00000707, 0.00000749, 0.00000712, 0.00000617, 0.00000489, 0.00000345, 0.00000194, 0.00000040, -0.00000112, -0.00000254, -0.00000368, -0.00000436, -0.00000447, -0.00000402, -0.00000321, -0.00000238, -0.00000193, -0.00000214, -0.00000312, -0.00000469, -0.00000644, -0.00000784, -0.00000840, -0.00000781, -0.00000606, },
		{ 0.00001051, 0.00000980, 0.00000862, 0.00000728, 0.00000594, 0.00000463, 0.00000325, 0.00000171, -0.00000000, -0.00000176, -0.00000332, -0.00000441, -0.00000482, -0.00000452, -0.00000366, -0.00000250, -0.00000133, -0.00000033, 0.00000053, 0.00000139, 0.00000246, 0.00000387, 0.00000545, 0.00000678, 0.00000721, 0.00000607, 0.00000292, -0.00000217, -0.00000856, -0.00001507, -0.00002026, -0.00002280, -0.00002181, -0.00001719, -0.00000962, -0.00000044, 0.00000871, 0.00001633, 0.00002135, 0.00002341, 0.00002276, 0.00002011, 0.00001629, 0.00001201, 0.00000764, 0.00000328, -0.00000114, -0.00000563, -0.00000996, -0.00001361, -0.00001586, -0.00001604, -0.00001380, -0.00000936, -0.00000349, 0.00000258, 0.00000752, 0.00001023, 0.00001020, 0.00000767, 0.00000352, -0.00000096, -0.00000450, -0.00000618, -0.00000570, -0.00000345, -0.00000028, 0.00000273, 0.00000465, 0.00000496, 0.00000366, 0.00000120, -0.00000169, -0.00000427, -0.00000599, -0.00000659, -0.00000613, -0.00000483, -0.00000303, -0.00000098, 0.00000109, 0.00000307, 0.00000480, 0.00000609, 0.00000668, 0.00000633, 0.00000491, 0.00000251, -0.00000051, -0.00000356, -0.00000594, -0.00000704, -0.00000650, -0.00000439, -0.00000112, 0.00000258, 0.00000590, 0.00000817, 0.00000902, 0.00000846, 0.00000680, 0.00000456, 0.00000227, 0.00000033, -0.00000107, -0.00000194, -0.00000243, -0.00000268, -0.00000282, -0.00000288, -0.00000284, -0.00000269, -0.00000252, -0.00000246, -0.00000269, -0.00000336, -0.00000447, -0.00000582, -0.00000705, -0.00000770, -0.00000737, -0.00000586, -0.00000323, 0.00000015, 0.00000373, 0.00000691, 0.00000921, 0.00001040, },
		{ 0.00000036, -0.00000039, -0.00000099, -0.00000164, -0.00000241, -0.00000322, -0.00000382, -0.00000399, -0.00000358, -0.00000260, -0.00000122, 0.00000029, 0.00000168, 0.00000280, 0.00000365, 0.00000432, 0.00000488, 0.00000530, 0.00000535, 0.00000462, 0.00000271, -0.00000063, -0.00000529, -0.00001070, -0.00001583, -0.00001949, -0.00002055, -0.00001833, -0.00001285, -0.00000486, 0.00000429, 0.00001299, 0.00001977, 0.00002364, 0.00002433, 0.00002223, 0.00001823, 0.00001333, 0.00000836, 0.00000380, -0.00000030, -0.00000413, -0.00000787, -0.00001149, -0.00001459, -0.00001649, -0.00001649, -0.00001411, -0.00000941, -0.00000309, 0.00000362, 0.00000924, 0.00001249, 0.00001264, 0.00000979, 0.00000481, -0.00000087, -0.00000566, -0.00000835, -0.00000837, -0.00000598, -0.00000212, 0.00000188, 0.00000476, 0.00000567, 0.00000441, 0.00000144, -0.00000232, -0.00000576, -0.00000796, -0.00000838, -0.00000697, -0.00000411, -0.00000047, 0.00000321, 0.00000631, 0.00000837, 0.00000918, 0.00000870, 0.00000705, 0.00000444, 0.00000119, -0.00000228, -0.00000546, -0.00000780, -0.00000881, -0.00000820, -0.00000597, -0.00000248, 0.00000160, 0.00000545, 0.00000826, 0.00000949, 0.00000897, 0.00000697, 0.00000408, 0.00000107, -0.00000138, -0.00000283, -0.00000319, -0.00000268, -0.00000173, -0.00000081, -0.00000028, -0.00000030, -0.00000084, -0.00000173, -0.00000276, -0.00000377, -0.00000469, -0.00000549, -0.00000613, -0.00000654, -0.00000653, -0.00000587, -0.00000438, -0.00000201, 0.00000108, 0.00000448, 0.00000765, 0.00001004, 0.00001125, 0.00001111, 0.00000980, 0.00000770, 0.00000534, 0.00000318, 0.00000150, },
		{ -0.00000106, -0.00000160, -0.00000203, -0.00000207, -0.00000155, -0.00000054, 0.00000076, 0.00000206, 0.00000311, 0.00000377, 0.00000404, 0.00000400, 0.00000371, 0.00000311, 0.00000205, 0.00000028, -0.00000236, -0.00000582, -0.00000969, -0.00001326, -0.00001559, -0.00001581, -0.00001338, -0.00000831, -0.00000124, 0.00000666, 0.00001397, 0.00001940, 0.00002209, 0.00002183, 0.00001906, 0.00001467, 0.00000968, 0.00000493, 0.00000085, -0.00000255, -0.00000555, -0.00000846, -0.00001135, -0.00001394, -0.00001559, -0.00001552, -0.00001319, -0.00000854, -0.00000219, 0.00000465, 0.00001046, 0.00001380, 0.00001385, 0.00001060, 0.00000494, -0.00000157, -0.00000718, -0.00001044, -0.00001064, -0.00000799, -0.00000353, 0.00000122, 0.00000472, 0.00000594, 0.00000460, 0.00000125, -0.00000296, -0.00000664, -0.00000859, -0.00000817, -0.00000543, -0.00000104, 0.00000390, 0.00000824, 0.00001103, 0.00001176, 0.00001042, 0.00000740, 0.00000336, -0.00000095, -0.00000484, -0.00000777, -0.00000939, -0.00000953, -0.00000822, -0.00000567, -0.00000226, 0.00000145, 0.00000483, 0.00000726, 0.00000828, 0.00000771, 0.00000575, 0.00000289, -0.00000013, -0.00000258, -0.00000388, -0.00000383, -0.00000258, -0.00000067, 0.00000124, 0.00000248, 0.00000265, 0.00000166, -0.00000026, -0.00000261, -0.00000487, -0.00000656, -0.00000744, -0.00000746, -0.00000672, -0.00000540, -0.00000368, -0.00000164, 0.00000065, 0.00000313, 0.00000563, 0.00000789, 0.00000958, 0.00001035, 0.00000998, 0.00000849, 0.00000614, 0.00000339, 0.00000080, -0.00000116, -0.00000222, -0.00000239, -0.00000195, -0.00000129, -0.00000081, -0.00000073, },
		{ -0.00000164, -0.00000050, 0.00000090, 0.00000218, 0.00000307, 0.00000346, 0.00000339, 0.00000297, 0.00000230, 0.00000136, 0.00000005, -0.00000176, -0.00000411, -0.00000680, -0.00000937, -0.00001117, -0.00001147, -0.00000979, -0.00000600, -0.00000053, 0.00000574, 0.00001166, 0.00001612, 0.00001833, 0.00001804, 0.00001557, 0.00001168, 0.00000728, 0.00000313, -0.00000034, -0.00000313, -0.00000551, -0.00000783, -0.00001021, -0.00001241, -0.00001384, -0.00001377, -0.00001161, -0.00000724, -0.00000122, 0.00000531, 0.00001083, 0.00001391, 0.00001366, 0.00001005, 0.00000398, -0.00000294, -0.00000888, -0.00001229, -0.00001239, -0.00000940, -0.00000443, 0.00000085, 0.00000478, 0.00000623, 0.00000493, 0.00000152, -0.00000268, -0.00000609, -0.00000743, -0.00000610, -0.00000235, 0.00000283, 0.00000801, 0.00001175, 0.00001306, 0.00001162, 0.00000780, 0.00000256, -0.00000290, -0.00000745, -0.00001028, -0.00001106, -0.00000994, -0.00000737, -0.00000401, -0.00000053, 0.00000250, 0.00000467, 0.00000575, 0.00000567, 0.00000454, 0.00000263, 0.00000037, -0.00000176, -0.00000324, -0.00000369, -0.00000297, -0.00000122, 0.00000110, 0.00000333, 0.00000478, 0.00000494, 0.00000360, 0.00000098, -0.00000237, -0.00000569, -0.00000821, -0.00000938, -0.00000900, -0.00000722, -0.00000447, -0.00000131, 0.00000173, 0.00000428, 0.00000617, 0.00000740, 0.00000804, 0.00000815, 0.00000775, 0.00000681, 0.00000532, 0.00000338, 0.00000119, -0.00000088, -0.00000245, -0.00000320, -0.00000301, -0.00000200, -0.00000055, 0.00000086, 0.00000177, 0.00000190, 0.00000126, 0.00000010, -0.00000114, -0.00000201, -0.00000220, },
		{ 0.00000440, 0.00000447, 0.00000399, 0.00000317, 0.00000215, 0.00000096, -0.00000049, -0.00000228, -0.00000440, -0.00000659, -0.00000840, -0.00000925, -0.00000860, -0.00000621, -0.00000223, 0.00000276, 0.00000785, 0.00001203, 0.00001452, 0.00001493, 0.00001336, 0.00001036, 0.00000669, 0.00000306, -0.00000008, -0.00000262, -0.00000474, -0.00000672, -0.00000869, -0.00001051, -0.00001169, -0.00001159, -0.00000968, -0.00000585, -0.00000054, 0.00000521, 0.00001000, 0.00001251, 0.00001189, 0.00000813, 0.00000208, -0.00000468, -0.00001035, -0.00001342, -0.00001311, -0.00000970, -0.00000432, 0.00000131, 0.00000550, 0.00000711, 0.00000590, 0.00000260, -0.00000139, -0.00000446, -0.00000534, -0.00000355, 0.00000047, 0.00000550, 0.00000993, 0.00001228, 0.00001171, 0.00000821, 0.00000264, -0.00000359, -0.00000893, -0.00001215, -0.00001266, -0.00001059, -0.00000675, -0.00000226, 0.00000173, 0.00000438, 0.00000530, 0.00000458, 0.00000269, 0.00000028, -0.00000199, -0.00000361, -0.00000428, -0.00000390, -0.00000257, -0.00000053, 0.00000185, 0.00000415, 0.00000587, 0.00000659, 0.00000599, 0.00000402, 0.00000090, -0.00000284, -0.00000645, -0.00000915, -0.00001029, -0.00000955, -0.00000704, -0.00000327, 0.00000100, 0.00000492, 0.00000779, 0.00000922, 0.00000921, 0.00000804, 0.00000618, 0.00000411, 0.00000221, 0.00000065, -0.00000053, -0.00000138, -0.00000195, -0.00000221, -0.00000208, -0.00000148, -0.00000043, 0.00000094, 0.00000230, 0.00000328, 0.00000354, 0.00000288, 0.00000141, -0.00000054, -0.00000245, -0.00000378, -0.00000415, -0.00000345, -0.00000189, 0.00000014, 0.00000211, 0.00000361, },
		{ 0.00000089, -0.00000056, -0.00000210, -0.00000384, -0.00000575, -0.00000754, -0.00000876, -0.00000888, -0.00000751, -0.00000456, -0.00000036, 0.00000438, 0.00000875, 0.00001191, 0.00001328, 0.00001274, 0.00001060, 0.00000747, 0.00000406, 0.00000091, -0.00000170, -0.00000380, -0.00000556, -0.00000716, -0.00000854, -0.00000942, -0.00000933, -0.00000785, -0.00000484, -0.00000065, 0.00000391, 0.00000768, 0.00000952, 0.00000872, 0.00000524, -0.00000015, -0.00000602, -0.00001077, -0.00001300, -0.00001208, -0.00000827, -0.00000273, 0.00000290, 0.00000699, 0.00000847, 0.00000717, 0.00000385, -0.00000007, -0.00000303, -0.00000383, -0.00000211, 0.00000155, 0.00000585, 0.00000916, 0.00001013, 0.00000814, 0.00000351, -0.00000255, -0.00000840, -0.00001240, -0.00001350, -0.00001150, -0.00000712, -0.00000174, 0.00000306, 0.00000598, 0.00000633, 0.00000425, 0.00000056, -0.00000353, -0.00000679, -0.00000830, -0.00000770, -0.00000522, -0.00000150, 0.00000255, 0.00000608, 0.00000842, 0.00000920, 0.00000835, 0.00000609, 0.00000282, -0.00000097, -0.00000467, -0.00000769, -0.00000951, -0.00000974, -0.00000822, -0.00000513, -0.00000096, 0.00000353, 0.00000749, 0.00001012, 0.00001094, 0.00000987, 0.00000726, 0.00000377, 0.00000024, -0.00000260, -0.00000427, -0.00000464, -0.00000391, -0.00000249, -0.00000083, 0.00000069, 0.00000189, 0.00000272, 0.00000324, 0.00000349, 0.00000345, 0.00000304, 0.00000213, 0.00000070, -0.00000113, -0.00000303, -0.00000459, -0.00000534, -0.00000499, -0.00000350, -0.00000113, 0.00000160, 0.00000412, 0.00000590, 0.00000667, 0.00000642, 0.00000540, 0.00000395, 0.00000239, },
		{ -0.00000727, -0.00000852, -0.00000905, -0.00000836, -0.00000615, -0.00000252, 0.00000204, 0.00000670, 0.00001056, 0.00001286, 0.00001324, 0.00001179, 0.00000899, 0.00000556, 0.00000216, -0.00000074, -0.00000299, -0.00000466, -0.00000594, -0.00000694, -0.00000758, -0.00000761, -0.00000672, -0.00000476, -0.00000188, 0.00000135, 0.00000412, 0.00000553, 0.00000499, 0.00000242, -0.00000158, -0.00000591, -0.00000922, -0.00001040, -0.00000892, -0.00000508, 0.00000008, 0.00000508, 0.00000848, 0.00000935, 0.00000762, 0.00000405, 0.00000003, -0.00000298, -0.00000389, -0.00000245, 0.00000072, 0.00000431, 0.00000679, 0.00000698, 0.00000445, -0.00000026, -0.00000580, -0.00001048, -0.00001281, -0.00001205, -0.00000840, -0.00000304, 0.00000233, 0.00000600, 0.00000682, 0.00000455, -0.00000006, -0.00000551, -0.00001005, -0.00001222, -0.00001127, -0.00000738, -0.00000152, 0.00000479, 0.00001001, 0.00001296, 0.00001308, 0.00001055, 0.00000612, 0.00000088, -0.00000406, -0.00000780, -0.00000979, -0.00000985, -0.00000815, -0.00000506, -0.00000112, 0.00000305, 0.00000682, 0.00000959, 0.00001089, 0.00001044, 0.00000828, 0.00000477, 0.00000059, -0.00000343, -0.00000645, -0.00000787, -0.00000744, -0.00000538, -0.00000227, 0.00000110, 0.00000398, 0.00000579, 0.00000632, 0.00000568, 0.00000424, 0.00000243, 0.00000064, -0.00000092, -0.00000217, -0.00000317, -0.00000393, -0.00000441, -0.00000445, -0.00000385, -0.00000248, -0.00000038, 0.00000220, 0.00000480, 0.00000690, 0.00000805, 0.00000801, 0.00000686, 0.00000494, 0.00000271, 0.00000062, -0.00000107, -0.00000235, -0.00000341, -0.00000451, -0.00000582, },
		{ -0.00000366, 0.00000071, 0.00000562, 0.00001009, 0.00001323, 0.00001441, 0.00001351, 0.00001088, 0.00000721, 0.00000332, -0.00000012, -0.00000276, -0.00000454, -0.00000566, -0.00000636, -0.00000679, -0.00000692, -0.00000657, -0.00000555, -0.00000385, -0.00000171, 0.00000038, 0.00000179, 0.00000202, 0.00000091, -0.00000125, -0.00000370, -0.00000550, -0.00000581, -0.00000424, -0.00000103, 0.00000299, 0.00000663, 0.00000872, 0.00000857, 0.00000620, 0.00000238, -0.00000163, -0.00000451, -0.00000533, -0.00000392, -0.00000094, 0.00000236, 0.00000455, 0.00000458, 0.00000218, -0.00000203, -0.00000672, -0.00001030, -0.00001149, -0.00000980, -0.00000573, -0.00000064, 0.00000371, 0.00000573, 0.00000461, 0.00000059, -0.00000510, -0.00001055, -0.00001385, -0.00001370, -0.00000981, -0.00000303, 0.00000490, 0.00001189, 0.00001611, 0.00001651, 0.00001312, 0.00000694, -0.00000033, -0.00000687, -0.00001117, -0.00001242, -0.00001062, -0.00000648, -0.00000113, 0.00000421, 0.00000850, 0.00001109, 0.00001174, 0.00001053, 0.00000784, 0.00000416, 0.00000005, -0.00000387, -0.00000702, -0.00000888, -0.00000911, -0.00000762, -0.00000462, -0.00000070, 0.00000334, 0.00000666, 0.00000854, 0.00000860, 0.00000691, 0.00000396, 0.00000049, -0.00000271, -0.00000499, -0.00000605, -0.00000590, -0.00000484, -0.00000329, -0.00000161, -0.00000001, 0.00000146, 0.00000288, 0.00000430, 0.00000567, 0.00000680, 0.00000741, 0.00000722, 0.00000612, 0.00000421, 0.00000184, -0.00000055, -0.00000249, -0.00000376, -0.00000435, -0.00000456, -0.00000478, -0.00000533, -0.00000631, -0.00000749, -0.00000835, -0.00000827, -0.00000676, },
		{ 0.00001597, 0.00001538, 0.00001263, 0.00000843, 0.00000373, -0.00000056, -0.00000383, -0.00000588, -0.00000687, -0.00000718, -0.00000717, -0.00000705, -0.00000683, -0.00000632, -0.00000535, -0.00000389, -0.00000212, -0.00000041, 0.00000081, 0.00000125, 0.00000093, 0.00000017, -0.00000048, -0.00000045, 0.00000055, 0.00000242, 0.00000464, 0.00000637, 0.00000685, 0.00000560, 0.00000277, -0.00000096, -0.00000450, -0.00000675, -0.00000699, -0.00000519, -0.00000202, 0.00000131, 0.00000350, 0.00000363, 0.00000148, -0.00000229, -0.00000643, -0.00000948, -0.00001030, -0.00000858, -0.00000489, -0.00000059, 0.00000266, 0.00000353, 0.00000149, -0.00000291, -0.00000820, -0.00001242, -0.00001381, -0.00001143, -0.00000550, 0.00000256, 0.00001059, 0.00001630, 0.00001804, 0.00001526, 0.00000874, 0.00000030, -0.00000768, -0.00001306, -0.00001445, -0.00001164, -0.00000556, 0.00000204, 0.00000916, 0.00001409, 0.00001582, 0.00001423, 0.00001000, 0.00000431, -0.00000154, -0.00000644, -0.00000965, -0.00001091, -0.00001028, -0.00000812, -0.00000487, -0.00000107, 0.00000276, 0.00000604, 0.00000826, 0.00000900, 0.00000804, 0.00000548, 0.00000178, -0.00000234, -0.00000598, -0.00000835, -0.00000892, -0.00000762, -0.00000480, -0.00000118, 0.00000245, 0.00000536, 0.00000716, 0.00000778, 0.00000746, 0.00000660, 0.00000556, 0.00000452, 0.00000351, 0.00000241, 0.00000108, -0.00000050, -0.00000218, -0.00000368, -0.00000469, -0.00000504, -0.00000478, -0.00000422, -0.00000377, -0.00000384, -0.00000459, -0.00000585, -0.00000707, -0.00000755, -0.00000664, -0.00000400, 0.00000021, 0.00000534, 0.00001035, 0.00001418, },
		{ -0.00000272, -0.00000663, -0.00000880, -0.00000937, -0.00000884, -0.00000784, -0.00000688, -0.00000618, -0.00000566, -0.00000504, -0.00000404, -0.00000255, -0.00000069, 0.00000124, 0.00000286, 0.00000391, 0.00000432, 0.00000428, 0.00000410, 0.00000404, 0.00000417, 0.00000430, 0.00000406, 0.00000303, 0.00000101, -0.00000181, -0.00000488, -0.00000737, -0.00000848, -0.00000776, -0.00000532, -0.00000186, 0.00000148, 0.00000355, 0.00000355, 0.00000137, -0.00000235, -0.00000638, -0.00000936, -0.00001028, -0.00000883, -0.00000560, -0.00000186, 0.00000090, 0.00000151, -0.00000038, -0.00000416, -0.00000837, -0.00001122, -0.00001121, -0.00000774, -0.00000138, 0.00000624, 0.00001289, 0.00001648, 0.00001577, 0.00001079, 0.00000293, -0.00000549, -0.00001191, -0.00001436, -0.00001201, -0.00000550, 0.00000331, 0.00001191, 0.00001788, 0.00001961, 0.00001670, 0.00001003, 0.00000143, -0.00000693, -0.00001311, -0.00001594, -0.00001519, -0.00001150, -0.00000608, -0.00000030, 0.00000464, 0.00000799, 0.00000943, 0.00000906, 0.00000720, 0.00000427, 0.00000073, -0.00000295, -0.00000625, -0.00000865, -0.00000966, -0.00000893, -0.00000642, -0.00000244, 0.00000233, 0.00000697, 0.00001056, 0.00001240, 0.00001221, 0.00001021, 0.00000700, 0.00000339, 0.00000013, -0.00000226, -0.00000365, -0.00000420, -0.00000427, -0.00000420, -0.00000419, -0.00000424, -0.00000419, -0.00000390, -0.00000332, -0.00000262, -0.00000212, -0.00000215, -0.00000288, -0.00000419, -0.00000558, -0.00000634, -0.00000577, -0.00000341, 0.00000067, 0.00000587, 0.00001114, 0.00001527, 0.00001725, 0.00001655, 0.00001330, 0.00000826, 0.00000253, },
		{ -0.00000737, -0.00000504, -0.00000347, -0.00000270, -0.00000241, -0.00000210, -0.00000130, 0.00000016, 0.00000216, 0.00000431, 0.00000613, 0.00000723, 0.00000745, 0.00000685, 0.00000565, 0.00000410, 0.00000235, 0.00000042, -0.00000170, -0.00000401, -0.00000632, -0.00000825, -0.00000930, -0.00000904, -0.00000728, -0.00000429, -0.00000079, 0.00000223, 0.00000380, 0.00000332, 0.00000078, -0.00000311, -0.00000718, -0.00001014, -0.00001105, -0.00000965, -0.00000650, -0.00000278, 0.00000009, 0.00000105, -0.00000027, -0.00000330, -0.00000670, -0.00000888, -0.00000856, -0.00000529, 0.00000030, 0.00000662, 0.00001168, 0.00001370, 0.00001185, 0.00000653, -0.00000062, -0.00000725, -0.00001105, -0.00001051, -0.00000552, 0.00000261, 0.00001145, 0.00001826, 0.00002080, 0.00001806, 0.00001058, 0.00000030, -0.00001008, -0.00001784, -0.00002111, -0.00001928, -0.00001315, -0.00000461, 0.00000401, 0.00001061, 0.00001382, 0.00001332, 0.00000972, 0.00000428, -0.00000156, -0.00000655, -0.00000989, -0.00001127, -0.00001076, -0.00000867, -0.00000538, -0.00000127, 0.00000326, 0.00000773, 0.00001158, 0.00001419, 0.00001499, 0.00001371, 0.00001042, 0.00000566, 0.00000031, -0.00000463, -0.00000826, -0.00001008, -0.00001006, -0.00000862, -0.00000646, -0.00000431, -0.00000268, -0.00000176, -0.00000142, -0.00000137, -0.00000131, -0.00000114, -0.00000096, -0.00000101, -0.00000151, -0.00000245, -0.00000353, -0.00000418, -0.00000373, -0.00000171, 0.00000193, 0.00000669, 0.00001158, 0.00001538, 0.00001700, 0.00001582, 0.00001190, 0.00000603, -0.00000053, -0.00000642, -0.00001052, -0.00001230, -0.00001188, -0.00000993, },
		{ 0.00000088, 0.00000035, 0.00000074, 0.00000216, 0.00000430, 0.00000652, 0.00000814, 0.00000864, 0.00000783, 0.00000583, 0.00000300, -0.00000020, -0.00000336, -0.00000616, -0.00000841, -0.00000993, -0.00001056, -0.00001013, -0.00000857, -0.00000599, -0.00000279, 0.00000039, 0.00000272, 0.00000349, 0.00000233, -0.00000059, -0.00000454, -0.00000841, -0.00001101, -0.00001154, -0.00000980, -0.00000639, -0.00000245, 0.00000066, 0.00000189, 0.00000087, -0.00000189, -0.00000519, -0.00000753, -0.00000768, -0.00000519, -0.00000058, 0.00000476, 0.00000908, 0.00001087, 0.00000945, 0.00000530, -0.00000008, -0.00000460, -0.00000638, -0.00000441, 0.00000099, 0.00000821, 0.00001480, 0.00001829, 0.00001703, 0.00001077, 0.00000090, -0.00001002, -0.00001895, -0.00002334, -0.00002192, -0.00001511, -0.00000484, 0.00000598, 0.00001440, 0.00001827, 0.00001678, 0.00001067, 0.00000186, -0.00000717, -0.00001411, -0.00001744, -0.00001673, -0.00001255, -0.00000620, 0.00000082, 0.00000719, 0.00001207, 0.00001509, 0.00001626, 0.00001577, 0.00001384, 0.00001064, 0.00000635, 0.00000127, -0.00000406, -0.00000895, -0.00001261, -0.00001442, -0.00001410, -0.00001186, -0.00000834, -0.00000447, -0.00000115, 0.00000100, 0.00000180, 0.00000150, 0.00000066, -0.00000014, -0.00000051, -0.00000038, -0.00000000, 0.00000027, 0.00000016, -0.00000032, -0.00000082, -0.00000077, 0.00000037, 0.00000283, 0.00000634, 0.00001014, 0.00001310, 0.00001414, 0.00001258, 0.00000838, 0.00000229, -0.00000440, -0.00001019, -0.00001382, -0.00001465, -0.00001280, -0.00000911, -0.00000479, -0.00000103, 0.00000136, 0.00000216, 0.00000175, },
		{ 0.00000499, 0.00000726, 0.00000836, 0.00000777, 0.00000547, 0.00000187, -0.00000234, -0.00000640, -0.00000969, -0.00001179, -0.00001256, -0.00001201, -0.00001032, -0.00000773, -0.00000463, -0.00000150, 0.00000109, 0.00000257, 0.00000251, 0.00000080, -0.00000226, -0.00000591, -0.00000918, -0.00001107, -0.00001094, -0.00000876, -0.00000514, -0.00000121, 0.00000174, 0.00000275, 0.00000149, -0.00000155, -0.00000516, -0.00000790, -0.00000852, -0.00000651, -0.00000224, 0.00000304, 0.00000771, 0.00001031, 0.00001014, 0.00000747, 0.00000354, 0.00000008, -0.00000135, 0.00000004, 0.00000386, 0.00000864, 0.00001231, 0.00001292, 0.00000940, 0.00000208, -0.00000731, -0.00001611, -0.00002160, -0.00002185, -0.00001645, -0.00000675, 0.00000456, 0.00001420, 0.00001930, 0.00001836, 0.00001164, 0.00000109, -0.00001022, -0.00001907, -0.00002304, -0.00002112, -0.00001397, -0.00000355, 0.00000749, 0.00001664, 0.00002214, 0.00002340, 0.00002087, 0.00001573, 0.00000942, 0.00000314, -0.00000236, -0.00000680, -0.00001022, -0.00001273, -0.00001434, -0.00001488, -0.00001412, -0.00001197, -0.00000862, -0.00000461, -0.00000071, 0.00000231, 0.00000391, 0.00000403, 0.00000303, 0.00000161, 0.00000051, 0.00000024, 0.00000089, 0.00000211, 0.00000330, 0.00000390, 0.00000364, 0.00000266, 0.00000154, 0.00000099, 0.00000154, 0.00000329, 0.00000577, 0.00000803, 0.00000898, 0.00000781, 0.00000428, -0.00000106, -0.00000701, -0.00001205, -0.00001483, -0.00001459, -0.00001143, -0.00000626, -0.00000052, 0.00000430, 0.00000712, 0.00000755, 0.00000602, 0.00000351, 0.00000120, 0.00000007, 0.00000055, 0.00000243, },
		{ 0.00000134, -0.00000357, -0.00000838, -0.00001216, -0.00001427, -0.00001451, -0.00001308, -0.00001040, -0.00000706, -0.00000364, -0.00000070, 0.00000129, 0.00000201, 0.00000132, -0.00000066, -0.00000352, -0.00000657, -0.00000899, -0.00001005, -0.00000933, -0.00000695, -0.00000354, -0.00000013, 0.00000214, 0.00000248, 0.00000068, -0.00000268, -0.00000644, -0.00000915, -0.00000962, -0.00000732, -0.00000262, 0.00000328, 0.00000875, 0.00001229, 0.00001303, 0.00001106, 0.00000736, 0.00000349, 0.00000091, 0.00000047, 0.00000202, 0.00000448, 0.00000622, 0.00000572, 0.00000225, -0.00000375, -0.00001070, -0.00001634, -0.00001853, -0.00001606, -0.00000916, 0.00000038, 0.00000975, 0.00001595, 0.00001681, 0.00001174, 0.00000204, -0.00000947, -0.00001927, -0.00002421, -0.00002252, -0.00001433, -0.00000164, 0.00001228, 0.00002390, 0.00003042, 0.00003055, 0.00002471, 0.00001480, 0.00000344, -0.00000678, -0.00001406, -0.00001771, -0.00001813, -0.00001642, -0.00001383, -0.00001132, -0.00000930, -0.00000765, -0.00000600, -0.00000400, -0.00000158, 0.00000096, 0.00000312, 0.00000442, 0.00000467, 0.00000408, 0.00000318, 0.00000261, 0.00000285, 0.00000395, 0.00000551, 0.00000684, 0.00000719, 0.00000618, 0.00000391, 0.00000102, -0.00000154, -0.00000287, -0.00000254, -0.00000073, 0.00000171, 0.00000361, 0.00000387, 0.00000194, -0.00000194, -0.00000676, -0.00001101, -0.00001322, -0.00001244, -0.00000862, -0.00000265, 0.00000392, 0.00000935, 0.00001228, 0.00001216, 0.00000938, 0.00000511, 0.00000089, -0.00000190, -0.00000249, -0.00000089, 0.00000210, 0.00000525, 0.00000732, 0.00000742, 0.00000530, },
		{ -0.00001613, -0.00001413, -0.00001073, -0.00000677, -0.00000307, -0.00000025, 0.00000132, 0.00000152, 0.00000045, -0.00000157, -0.00000404, -0.00000639, -0.00000800, -0.00000841, -0.00000742, -0.00000526, -0.00000251, -0.00000003, 0.00000130, 0.00000091, -0.00000122, -0.00000448, -0.00000775, -0.00000973, -0.00000936, -0.00000626, -0.00000086, 0.00000559, 0.00001144, 0.00001515, 0.00001579, 0.00001337, 0.00000881, 0.00000364, -0.00000057, -0.00000277, -0.00000279, -0.00000140, 0.00000009, 0.00000031, -0.00000150, -0.00000513, -0.00000944, -0.00001271, -0.00001333, -0.00001048, -0.00000460, 0.00000268, 0.00000895, 0.00001189, 0.00001011, 0.00000377, -0.00000535, -0.00001428, -0.00001978, -0.00001948, -0.00001264, -0.00000056, 0.00001373, 0.00002638, 0.00003381, 0.00003387, 0.00002641, 0.00001332, -0.00000208, -0.00001605, -0.00002555, -0.00002903, -0.00002672, -0.00002034, -0.00001242, -0.00000539, -0.00000086, 0.00000073, 0.00000004, -0.00000161, -0.00000284, -0.00000275, -0.00000117, 0.00000141, 0.00000416, 0.00000631, 0.00000746, 0.00000773, 0.00000755, 0.00000743, 0.00000763, 0.00000799, 0.00000800, 0.00000704, 0.00000471, 0.00000111, -0.00000308, -0.00000679, -0.00000893, -0.00000885, -0.00000659, -0.00000297, 0.00000065, 0.00000286, 0.00000275, 0.00000023, -0.00000383, -0.00000792, -0.00001033, -0.00000983, -0.00000611, 0.00000008, 0.00000711, 0.00001302, 0.00001614, 0.00001564, 0.00001180, 0.00000592, -0.00000016, -0.00000465, -0.00000637, -0.00000510, -0.00000162, 0.00000262, 0.00000597, 0.00000716, 0.00000563, 0.00000167, -0.00000375, -0.00000929, -0.00001369, -0.00001607, },
		{ 0.00000010, 0.00000150, 0.00000141, 0.00000015, -0.00000180, -0.00000392, -0.00000571, -0.00000676, -0.00000683, -0.00000593, -0.00000431, -0.00000244, -0.00000096, -0.00000044, -0.00000121, -0.00000316, -0.00000570, -0.00000787, -0.00000863, -0.00000719, -0.00000336, 0.00000230, 0.00000857, 0.00001385, 0.00001672, 0.00001633, 0.00001278, 0.00000704, 0.00000071, -0.00000453, -0.00000747, -0.00000780, -0.00000610, -0.00000365, -0.00000188, -0.00000175, -0.00000344, -0.00000621, -0.00000871, -0.00000958, -0.00000801, -0.00000419, 0.00000069, 0.00000485, 0.00000656, 0.00000490, 0.00000023, -0.00000579, -0.00001068, -0.00001200, -0.00000826, 0.00000037, 0.00001195, 0.00002323, 0.00003062, 0.00003136, 0.00002441, 0.00001098, -0.00000585, -0.00002193, -0.00003333, -0.00003746, -0.00003383, -0.00002410, -0.00001148, 0.00000034, 0.00000832, 0.00001096, 0.00000859, 0.00000307, -0.00000296, -0.00000704, -0.00000769, -0.00000474, 0.00000072, 0.00000692, 0.00001206, 0.00001497, 0.00001535, 0.00001371, 0.00001102, 0.00000821, 0.00000577, 0.00000367, 0.00000147, -0.00000129, -0.00000473, -0.00000846, -0.00001158, -0.00001306, -0.00001219, -0.00000891, -0.00000401, 0.00000107, 0.00000474, 0.00000578, 0.00000389, -0.00000017, -0.00000479, -0.00000803, -0.00000831, -0.00000502, 0.00000124, 0.00000878, 0.00001537, 0.00001896, 0.00001839, 0.00001375, 0.00000639, -0.00000156, -0.00000783, -0.00001080, -0.00000993, -0.00000592, -0.00000041, 0.00000456, 0.00000725, 0.00000671, 0.00000302, -0.00000279, -0.00000913, -0.00001440, -0.00001740, -0.00001764, -0.00001538, -0.00001145, -0.00000691, -0.00000282, },
		{ -0.00000349, -0.00000500, -0.00000580, -0.00000582, -0.00000518, -0.00000414, -0.00000305, -0.00000230, -0.00000220, -0.00000286, -0.00000413, -0.00000554, -0.00000638, -0.00000594, -0.00000375, 0.00000020, 0.00000531, 0.00001046, 0.00001430, 0.00001565, 0.00001392, 0.00000933, 0.00000293, -0.00000367, -0.00000884, -0.00001139, -0.00001101, -0.00000833, -0.00000469, -0.00000161, -0.00000029, -0.00000113, -0.00000358, -0.00000644, -0.00000831, -0.00000817, -0.00000582, -0.00000199, 0.00000195, 0.00000455, 0.00000491, 0.00000311, 0.00000025, -0.00000189, -0.00000163, 0.00000194, 0.00000834, 0.00001576, 0.00002147, 0.00002275, 0.00001792, 0.00000704, -0.00000785, -0.00002322, -0.00003508, -0.00004015, -0.00003699, -0.00002647, -0.00001154, 0.00000362, 0.00001495, 0.00001971, 0.00001731, 0.00000939, -0.00000079, -0.00000949, -0.00001366, -0.00001195, -0.00000494, 0.00000509, 0.00001506, 0.00002215, 0.00002464, 0.00002233, 0.00001642, 0.00000892, 0.00000186, -0.00000340, -0.00000648, -0.00000788, -0.00000857, -0.00000937, -0.00001057, -0.00001175, -0.00001206, -0.00001066, -0.00000723, -0.00000225, 0.00000303, 0.00000702, 0.00000838, 0.00000659, 0.00000228, -0.00000294, -0.00000696, -0.00000794, -0.00000499, 0.00000143, 0.00000958, 0.00001698, 0.00002118, 0.00002066, 0.00001530, 0.00000648, -0.00000335, -0.00001147, -0.00001573, -0.00001521, -0.00001050, -0.00000344, 0.00000352, 0.00000808, 0.00000884, 0.00000559, -0.00000064, -0.00000801, -0.00001452, -0.00001856, -0.00001935, -0.00001709, -0.00001268, -0.00000747, -0.00000275, 0.00000056, 0.00000206, 0.00000188, 0.00000047, -0.00000152, },
		{ -0.00000440, -0.00000373, -0.00000328, -0.00000316, -0.00000336, -0.00000375, -0.00000401, -0.00000375, -0.00000259, -0.00000032, 0.00000295, 0.00000669, 0.00001006, 0.00001205, 0.00001184, 0.00000909, 0.00000417, -0.00000193, -0.00000776, -0.00001188, -0.00001332, -0.00001190, -0.00000832, -0.00000396, -0.00000035, 0.00000128, 0.00000051, -0.00000216, -0.00000548, -0.00000795, -0.00000836, -0.00000625, -0.00000212, 0.00000280, 0.00000699, 0.00000930, 0.00000935, 0.00000770, 0.00000557, 0.00000433, 0.00000478, 0.00000680, 0.00000917, 0.00001005, 0.00000767, 0.00000112, -0.00000903, -0.00002069, -0.00003078, -0.00003612, -0.00003459, -0.00002593, -0.00001199, 0.00000367, 0.00001683, 0.00002395, 0.00002331, 0.00001558, 0.00000364, -0.00000830, -0.00001612, -0.00001707, -0.00001064, 0.00000127, 0.00001503, 0.00002650, 0.00003232, 0.00003094, 0.00002302, 0.00001105, -0.00000158, -0.00001171, -0.00001737, -0.00001828, -0.00001563, -0.00001146, -0.00000777, -0.00000577, -0.00000549, -0.00000604, -0.00000607, -0.00000454, -0.00000124, 0.00000303, 0.00000680, 0.00000855, 0.00000738, 0.00000353, -0.00000161, -0.00000595, -0.00000744, -0.00000491, 0.00000140, 0.00000983, 0.00001773, 0.00002237, 0.00002187, 0.00001590, 0.00000581, -0.00000567, -0.00001538, -0.00002067, -0.00002025, -0.00001459, -0.00000572, 0.00000345, 0.00001006, 0.00001216, 0.00000925, 0.00000236, -0.00000640, -0.00001456, -0.00002002, -0.00002163, -0.00001943, -0.00001445, -0.00000830, -0.00000261, 0.00000141, 0.00000323, 0.00000300, 0.00000134, -0.00000094, -0.00000311, -0.00000467, -0.00000546, -0.00000552, -0.00000508, },
		{ -0.00000222, -0.00000176, -0.00000101, 0.00000016, 0.00000177, 0.00000370, 0.00000563, 0.00000702, 0.00000730, 0.00000601, 0.00000303, -0.00000129, -0.00000612, -0.00001033, -0.00001284, -0.00001295, -0.00001064, -0.00000664, -0.00000221, 0.00000125, 0.00000269, 0.00000178, -0.00000095, -0.00000424, -0.00000656, -0.00000671, -0.00000418, 0.00000053, 0.00000615, 0.00001103, 0.00001375, 0.00001365, 0.00001098, 0.00000681, 0.00000256, -0.00000064, -0.00000236, -0.00000308, -0.00000395, -0.00000620, -0.00001049, -0.00001647, -0.00002264, -0.00002682, -0.00002688, -0.00002164, -0.00001149, 0.00000146, 0.00001392, 0.00002243, 0.00002446, 0.00001942, 0.00000898, -0.00000337, -0.00001337, -0.00001738, -0.00001360, -0.00000280, 0.00001194, 0.00002609, 0.00003516, 0.00003614, 0.00002845, 0.00001412, -0.00000291, -0.00001807, -0.00002761, -0.00002969, -0.00002489, -0.00001571, -0.00000569, 0.00000196, 0.00000535, 0.00000442, 0.00000076, -0.00000325, -0.00000540, -0.00000455, -0.00000106, 0.00000352, 0.00000710, 0.00000802, 0.00000578, 0.00000132, -0.00000336, -0.00000592, -0.00000472, 0.00000051, 0.00000841, 0.00001635, 0.00002138, 0.00002125, 0.00001527, 0.00000468, -0.00000774, -0.00001845, -0.00002439, -0.00002388, -0.00001725, -0.00000668, 0.00000452, 0.00001294, 0.00001610, 0.00001324, 0.00000537, -0.00000508, -0.00001512, -0.00002211, -0.00002452, -0.00002221, -0.00001637, -0.00000893, -0.00000195, 0.00000299, 0.00000517, 0.00000474, 0.00000252, -0.00000044, -0.00000316, -0.00000504, -0.00000588, -0.00000585, -0.00000528, -0.00000451, -0.00000379, -0.00000323, -0.00000284, -0.00000254, },
		{ 0.00000177, 0.00000191, 0.00000169, 0.00000091, -0.00000061, -0.00000287, -0.00000565, -0.00000842, -0.00001048, -0.00001118, -0.00001010, -0.00000733, -0.00000348, 0.00000043, 0.00000330, 0.00000431, 0.00000328, 0.00000074, -0.00000215, -0.00000404, -0.00000385, -0.00000121, 0.00000335, 0.00000852, 0.00001264, 0.00001426, 0.00001268, 0.00000815, 0.00000183, -0.00000467, -0.00000984, -0.00001283, -0.00001367, -0.00001322, -0.00001263, -0.00001285, -0.00001408, -0.00001562, -0.00001612, -0.00001417, -0.00000899, -0.00000095, 0.00000831, 0.00001631, 0.00002056, 0.00001950, 0.00001322, 0.00000365, -0.00000596, -0.00001202, -0.00001191, -0.00000493, 0.00000721, 0.00002085, 0.00003145, 0.00003508, 0.00002979, 0.00001632, -0.00000198, -0.00002020, -0.00003336, -0.00003803, -0.00003334, -0.00002128, -0.00000593, 0.00000790, 0.00001628, 0.00001735, 0.00001182, 0.00000254, -0.00000662, -0.00001222, -0.00001243, -0.00000758, 0.00000012, 0.00000750, 0.00001173, 0.00001137, 0.00000693, 0.00000058, -0.00000472, -0.00000643, -0.00000344, 0.00000341, 0.00001161, 0.00001786, 0.00001932, 0.00001469, 0.00000479, -0.00000770, -0.00001902, -0.00002564, -0.00002544, -0.00001845, -0.00000683, 0.00000579, 0.00001554, 0.00001948, 0.00001652, 0.00000765, -0.00000441, -0.00001618, -0.00002449, -0.00002743, -0.00002473, -0.00001774, -0.00000880, -0.00000044, 0.00000536, 0.00000767, 0.00000670, 0.00000349, -0.00000051, -0.00000403, -0.00000624, -0.00000694, -0.00000641, -0.00000519, -0.00000380, -0.00000262, -0.00000176, -0.00000118, -0.00000074, -0.00000032, 0.00000011, 0.00000057, 0.00000103, 0.00000145, },
		{ -0.00000685, -0.00000733, -0.00000758, -0.00000739, -0.00000649, -0.00000471, -0.00000212, 0.00000086, 0.00000360, 0.00000537, 0.00000565, 0.00000437, 0.00000200, -0.00000053, -0.00000216, -0.00000203, 0.00000009, 0.00000367, 0.00000752, 0.00001016, 0.00001032, 0.00000742, 0.00000184, -0.00000518, -0.00001194, -0.00001680, -0.00001878, -0.00001783, -0.00001483, -0.00001112, -0.00000803, -0.00000629, -0.00000576, -0.00000555, -0.00000447, -0.00000163, 0.00000306, 0.00000868, 0.00001358, 0.00001599, 0.00001480, 0.00001013, 0.00000349, -0.00000264, -0.00000564, -0.00000378, 0.00000294, 0.00001262, 0.00002190, 0.00002700, 0.00002511, 0.00001546, -0.00000016, -0.00001779, -0.00003245, -0.00003973, -0.00003729, -0.00002567, -0.00000826, 0.00000975, 0.00002305, 0.00002788, 0.00002328, 0.00001131, -0.00000363, -0.00001638, -0.00002273, -0.00002083, -0.00001173, 0.00000110, 0.00001303, 0.00002000, 0.00001988, 0.00001317, 0.00000276, -0.00000723, -0.00001297, -0.00001233, -0.00000567, 0.00000435, 0.00001375, 0.00001869, 0.00001690, 0.00000852, -0.00000392, -0.00001637, -0.00002465, -0.00002595, -0.00001973, -0.00000796, 0.00000558, 0.00001655, 0.00002146, 0.00001881, 0.00000946, -0.00000371, -0.00001680, -0.00002614, -0.00002943, -0.00002629, -0.00001819, -0.00000789, 0.00000162, 0.00000796, 0.00001004, 0.00000816, 0.00000370, -0.00000149, -0.00000573, -0.00000799, -0.00000808, -0.00000648, -0.00000406, -0.00000168, 0.00000009, 0.00000104, 0.00000129, 0.00000109, 0.00000069, 0.00000021, -0.00000035, -0.00000104, -0.00000189, -0.00000286, -0.00000387, -0.00000480, -0.00000560, -0.00000627, },
		{ 0.00000232, 0.00000369, 0.00000486, 0.00000563, 0.00000574, 0.00000497, 0.00000331, 0.00000108, -0.00000109, -0.00000248, -0.00000253, -0.00000108, 0.00000140, 0.00000397, 0.00000544, 0.00000482, 0.00000170, -0.00000351, -0.00000964, -0.00001509, -0.00001838, -0.00001863, -0.00001588, -0.00001108, -0.00000574, -0.00000135, 0.00000112, 0.00000160, 0.00000086, 0.00000017, 0.00000071, 0.00000303, 0.00000677, 0.00001076, 0.00001351, 0.00001385, 0.00001150, 0.00000728, 0.00000289, 0.00000026, 0.00000069, 0.00000423, 0.00000947, 0.00001389, 0.00001480, 0.00001037, 0.00000058, -0.00001253, -0.00002532, -0.00003362, -0.00003420, -0.00002601, -0.00001075, 0.00000746, 0.00002327, 0.00003185, 0.00003046, 0.00001950, 0.00000246, -0.00001517, -0.00002765, -0.00003091, -0.00002394, -0.00000910, 0.00000863, 0.00002336, 0.00003024, 0.00002715, 0.00001534, -0.00000101, -0.00001626, -0.00002526, -0.00002512, -0.00001614, -0.00000169, 0.00001302, 0.00002279, 0.00002419, 0.00001673, 0.00000301, -0.00001227, -0.00002395, -0.00002812, -0.00002350, -0.00001178, 0.00000303, 0.00001598, 0.00002282, 0.00002142, 0.00001232, -0.00000151, -0.00001578, -0.00002630, -0.00003027, -0.00002709, -0.00001833, -0.00000707, 0.00000325, 0.00000989, 0.00001158, 0.00000875, 0.00000312, -0.00000301, -0.00000759, -0.00000941, -0.00000836, -0.00000524, -0.00000138, 0.00000197, 0.00000394, 0.00000428, 0.00000324, 0.00000136, -0.00000080, -0.00000287, -0.00000472, -0.00000637, -0.00000784, -0.00000910, -0.00001000, -0.00001031, -0.00000990, -0.00000873, -0.00000698, -0.00000492, -0.00000283, -0.00000090, 0.00000080, },
		{ 0.00000081, -0.00000084, -0.00000239, -0.00000362, -0.00000422, -0.00000394, -0.00000273, -0.00000090, 0.00000088, 0.00000178, 0.00000109, -0.00000146, -0.00000551, -0.00001011, -0.00001394, -0.00001579, -0.00001493, -0.00001148, -0.00000634, -0.00000095, 0.00000326, 0.00000535, 0.00000523, 0.00000368, 0.00000202, 0.00000156, 0.00000306, 0.00000639, 0.00001054, 0.00001399, 0.00001536, 0.00001396, 0.00001012, 0.00000510, 0.00000061, -0.00000194, -0.00000201, -0.00000026, 0.00000167, 0.00000187, -0.00000104, -0.00000711, -0.00001484, -0.00002152, -0.00002418, -0.00002072, -0.00001094, 0.00000310, 0.00001752, 0.00002775, 0.00003013, 0.00002324, 0.00000863, -0.00000944, -0.00002528, -0.00003360, -0.00003133, -0.00001877, 0.00000034, 0.00001994, 0.00003359, 0.00003661, 0.00002770, 0.00000946, -0.00001241, -0.00003092, -0.00004011, -0.00003703, -0.00002270, -0.00000178, 0.00001898, 0.00003294, 0.00003579, 0.00002689, 0.00000943, -0.00001074, -0.00002705, -0.00003443, -0.00003086, -0.00001797, -0.00000033, 0.00001611, 0.00002608, 0.00002661, 0.00001788, 0.00000301, -0.00001316, -0.00002570, -0.00003117, -0.00002861, -0.00001959, -0.00000751, 0.00000370, 0.00001086, 0.00001250, 0.00000911, 0.00000272, -0.00000395, -0.00000848, -0.00000955, -0.00000718, -0.00000259, 0.00000243, 0.00000614, 0.00000746, 0.00000617, 0.00000282, -0.00000158, -0.00000599, -0.00000967, -0.00001226, -0.00001380, -0.00001445, -0.00001437, -0.00001356, -0.00001194, -0.00000942, -0.00000607, -0.00000221, 0.00000168, 0.00000502, 0.00000734, 0.00000842, 0.00000832, 0.00000734, 0.00000586, 0.00000418, 0.00000249, },
		{ -0.00000329, -0.00000202, -0.00000074, 0.00000026, 0.00000056, -0.00000023, -0.00000227, -0.00000528, -0.00000858, -0.00001120, -0.00001219, -0.00001100, -0.00000770, -0.00000304, 0.00000180, 0.00000556, 0.00000742, 0.00000727, 0.00000578, 0.00000413, 0.00000355, 0.00000477, 0.00000770, 0.00001138, 0.00001434, 0.00001513, 0.00001295, 0.00000798, 0.00000141, -0.00000500, -0.00000952, -0.00001117, -0.00001005, -0.00000732, -0.00000469, -0.00000365, -0.00000482, -0.00000758, -0.00001022, -0.00001070, -0.00000749, -0.00000046, 0.00000885, 0.00001750, 0.00002216, 0.00002035, 0.00001158, -0.00000221, -0.00001696, -0.00002777, -0.00003051, -0.00002343, -0.00000797, 0.00001139, 0.00002842, 0.00003717, 0.00003398, 0.00001891, -0.00000406, -0.00002812, -0.00004572, -0.00005105, -0.00004195, -0.00002078, 0.00000628, 0.00003108, 0.00004614, 0.00004700, 0.00003361, 0.00001037, -0.00001536, -0.00003564, -0.00004446, -0.00003957, -0.00002311, -0.00000072, 0.00002036, 0.00003366, 0.00003546, 0.00002585, 0.00000840, -0.00001112, -0.00002679, -0.00003437, -0.00003249, -0.00002282, -0.00000923, 0.00000368, 0.00001214, 0.00001436, 0.00001082, 0.00000389, -0.00000328, -0.00000790, -0.00000849, -0.00000525, 0.00000023, 0.00000565, 0.00000887, 0.00000865, 0.00000493, -0.00000126, -0.00000830, -0.00001455, -0.00001884, -0.00002071, -0.00002035, -0.00001831, -0.00001522, -0.00001151, -0.00000740, -0.00000297, 0.00000166, 0.00000617, 0.00001005, 0.00001269, 0.00001358, 0.00001254, 0.00000980, 0.00000598, 0.00000189, -0.00000173, -0.00000436, -0.00000584, -0.00000630, -0.00000605, -0.00000536, -0.00000442, },
		{ -0.00000144, -0.00000331, -0.00000548, -0.00000757, -0.00000900, -0.00000918, -0.00000770, -0.00000459, -0.00000038, 0.00000401, 0.00000756, 0.00000951, 0.00000968, 0.00000850, 0.00000692, 0.00000593, 0.00000621, 0.00000773, 0.00000976, 0.00001104, 0.00001036, 0.00000702, 0.00000128, -0.00000569, -0.00001209, -0.00001614, -0.00001672, -0.00001382, -0.00000853, -0.00000271, 0.00000177, 0.00000370, 0.00000305, 0.00000088, -0.00000107, -0.00000121, 0.00000116, 0.00000538, 0.00000965, 0.00001164, 0.00000956, 0.00000301, -0.00000651, -0.00001594, -0.00002163, -0.00002071, -0.00001236, 0.00000158, 0.00001688, 0.00002814, 0.00003067, 0.00002216, 0.00000387, -0.00001948, -0.00004093, -0.00005345, -0.00005229, -0.00003667, -0.00001033, 0.00001944, 0.00004396, 0.00005585, 0.00005143, 0.00003193, 0.00000321, -0.00002608, -0.00004714, -0.00005384, -0.00004448, -0.00002235, 0.00000549, 0.00003048, 0.00004526, 0.00004589, 0.00003292, 0.00001093, -0.00001309, -0.00003207, -0.00004104, -0.00003853, -0.00002663, -0.00001004, 0.00000571, 0.00001612, 0.00001903, 0.00001503, 0.00000692, -0.00000156, -0.00000712, -0.00000807, -0.00000471, 0.00000097, 0.00000618, 0.00000842, 0.00000636, 0.00000024, -0.00000838, -0.00001716, -0.00002388, -0.00002706, -0.00002630, -0.00002223, -0.00001608, -0.00000921, -0.00000265, 0.00000301, 0.00000763, 0.00001128, 0.00001394, 0.00001543, 0.00001542, 0.00001361, 0.00000998, 0.00000493, -0.00000071, -0.00000591, -0.00000970, -0.00001146, -0.00001111, -0.00000908, -0.00000614, -0.00000313, -0.00000068, 0.00000089, 0.00000157, 0.00000154, 0.00000098, -0.00000001, },
		{ -0.00000544, -0.00000331, -0.00000026, 0.00000341, 0.00000710, 0.00001006, 0.00001165, 0.00001160, 0.00001015, 0.00000795, 0.00000586, 0.00000451, 0.00000408, 0.00000418, 0.00000393, 0.00000246, -0.00000078, -0.00000554, -0.00001085, -0.00001515, -0.00001694, -0.00001529, -0.00001031, -0.00000322, 0.00000405, 0.00000942, 0.00001148, 0.00000996, 0.00000583, 0.00000093, -0.00000277, -0.00000393, -0.00000242, 0.00000065, 0.00000343, 0.00000410, 0.00000177, -0.00000302, -0.00000838, -0.00001174, -0.00001098, -0.00000543, 0.00000357, 0.00001284, 0.00001836, 0.00001676, 0.00000677, -0.00000995, -0.00002896, -0.00004421, -0.00005004, -0.00004311, -0.00002392, 0.00000307, 0.00003046, 0.00005014, 0.00005587, 0.00004538, 0.00002127, -0.00000953, -0.00003791, -0.00005527, -0.00005630, -0.00004063, -0.00001297, 0.00001836, 0.00004404, 0.00005662, 0.00005278, 0.00003421, 0.00000688, -0.00002093, -0.00004134, -0.00004919, -0.00004345, -0.00002720, -0.00000634, 0.00001255, 0.00002429, 0.00002661, 0.00002053, 0.00000960, -0.00000159, -0.00000912, -0.00001104, -0.00000782, -0.00000190, 0.00000336, 0.00000508, 0.00000187, -0.00000573, -0.00001554, -0.00002457, -0.00003009, -0.00003051, -0.00002575, -0.00001713, -0.00000681, 0.00000301, 0.00001067, 0.00001543, 0.00001742, 0.00001725, 0.00001565, 0.00001314, 0.00000989, 0.00000587, 0.00000112, -0.00000404, -0.00000894, -0.00001268, -0.00001442, -0.00001368, -0.00001061, -0.00000597, -0.00000091, 0.00000335, 0.00000591, 0.00000642, 0.00000510, 0.00000261, -0.00000029, -0.00000292, -0.00000493, -0.00000623, -0.00000692, -0.00000707, -0.00000663, },
		{ 0.00001098, 0.00001123, 0.00001022, 0.00000802, 0.00000507, 0.00000201, -0.00000051, -0.00000212, -0.00000291, -0.00000336, -0.00000413, -0.00000570, -0.00000803, -0.00001054, -0.00001215, -0.00001180, -0.00000884, -0.00000347, 0.00000320, 0.00000940, 0.00001330, 0.00001363, 0.00001025, 0.00000422, -0.00000248, -0.00000765, -0.00000966, -0.00000806, -0.00000374, 0.00000139, 0.00000516, 0.00000604, 0.00000374, -0.00000066, -0.00000510, -0.00000752, -0.00000673, -0.00000306, 0.00000156, 0.00000435, 0.00000277, -0.00000424, -0.00001555, -0.00002785, -0.00003664, -0.00003773, -0.00002893, -0.00001119, 0.00001133, 0.00003227, 0.00004500, 0.00004489, 0.00003102, 0.00000678, -0.00002088, -0.00004352, -0.00005384, -0.00004806, -0.00002730, 0.00000265, 0.00003301, 0.00005477, 0.00006143, 0.00005106, 0.00002685, -0.00000394, -0.00003233, -0.00005037, -0.00005358, -0.00004214, -0.00002052, 0.00000411, 0.00002435, 0.00003484, 0.00003380, 0.00002320, 0.00000773, -0.00000702, -0.00001651, -0.00001872, -0.00001453, -0.00000711, -0.00000049, 0.00000196, -0.00000120, -0.00000908, -0.00001888, -0.00002701, -0.00003037, -0.00002742, -0.00001863, -0.00000621, 0.00000670, 0.00001711, 0.00002302, 0.00002390, 0.00002056, 0.00001466, 0.00000801, 0.00000196, -0.00000283, -0.00000636, -0.00000896, -0.00001091, -0.00001217, -0.00001237, -0.00001106, -0.00000796, -0.00000334, 0.00000199, 0.00000685, 0.00000999, 0.00001060, 0.00000852, 0.00000436, -0.00000074, -0.00000546, -0.00000872, -0.00000999, -0.00000931, -0.00000722, -0.00000443, -0.00000152, 0.00000115, 0.00000356, 0.00000579, 0.00000790, 0.00000974, },
		{ -0.00000320, -0.00000539, -0.00000674, -0.00000710, -0.00000670, -0.00000602, -0.00000556, -0.00000556, -0.00000585, -0.00000587, -0.00000489, -0.00000245, 0.00000141, 0.00000591, 0.00000977, 0.00001161, 0.00001050, 0.00000639, 0.00000031, -0.00000591, -0.00001026, -0.00001121, -0.00000839, -0.00000274, 0.00000372, 0.00000861, 0.00001007, 0.00000746, 0.00000160, -0.00000550, -0.00001141, -0.00001425, -0.00001349, -0.00001011, -0.00000627, -0.00000440, -0.00000606, -0.00001122, -0.00001799, -0.00002326, -0.00002387, -0.00001795, -0.00000594, 0.00000924, 0.00002293, 0.00003025, 0.00002791, 0.00001559, -0.00000359, -0.00002378, -0.00003821, -0.00004142, -0.00003119, -0.00000960, 0.00001746, 0.00004198, 0.00005631, 0.00005569, 0.00003983, 0.00001314, -0.00001663, -0.00004082, -0.00005252, -0.00004875, -0.00003130, -0.00000601, 0.00001917, 0.00003681, 0.00004223, 0.00003486, 0.00001810, -0.00000206, -0.00001919, -0.00002858, -0.00002863, -0.00002108, -0.00001012, -0.00000065, 0.00000351, 0.00000098, -0.00000686, -0.00001653, -0.00002380, -0.00002532, -0.00001979, -0.00000833, 0.00000590, 0.00001891, 0.00002717, 0.00002874, 0.00002373, 0.00001408, 0.00000274, -0.00000732, -0.00001405, -0.00001671, -0.00001586, -0.00001279, -0.00000893, -0.00000527, -0.00000216, 0.00000059, 0.00000333, 0.00000621, 0.00000890, 0.00001068, 0.00001070, 0.00000842, 0.00000393, -0.00000194, -0.00000780, -0.00001217, -0.00001389, -0.00001255, -0.00000858, -0.00000309, 0.00000257, 0.00000718, 0.00001005, 0.00001113, 0.00001083, 0.00000977, 0.00000848, 0.00000719, 0.00000582, 0.00000414, 0.00000198, -0.00000058, },
		{ -0.00000401, -0.00000266, -0.00000166, -0.00000116, -0.00000096, -0.00000062, 0.00000034, 0.00000209, 0.00000435, 0.00000637, 0.00000720, 0.00000613, 0.00000306, -0.00000134, -0.00000571, -0.00000849, -0.00000848, -0.00000542, -0.00000019, 0.00000536, 0.00000906, 0.00000916, 0.00000507, -0.00000234, -0.00001098, -0.00001821, -0.00002180, -0.00002078, -0.00001582, -0.00000899, -0.00000303, -0.00000023, -0.00000157, -0.00000625, -0.00001197, -0.00001585, -0.00001559, -0.00001050, -0.00000193, 0.00000710, 0.00001300, 0.00001306, 0.00000660, -0.00000448, -0.00001615, -0.00002362, -0.00002304, -0.00001311, 0.00000422, 0.00002408, 0.00004020, 0.00004696, 0.00004136, 0.00002429, 0.00000039, -0.00002328, -0.00003951, -0.00004325, -0.00003336, -0.00001300, 0.00001137, 0.00003204, 0.00004262, 0.00004000, 0.00002532, 0.00000347, -0.00001866, -0.00003452, -0.00003996, -0.00003451, -0.00002125, -0.00000558, 0.00000685, 0.00001206, 0.00000907, 0.00000005, -0.00001067, -0.00001826, -0.00001919, -0.00001254, -0.00000018, 0.00001387, 0.00002495, 0.00002934, 0.00002557, 0.00001476, 0.00000016, -0.00001397, -0.00002381, -0.00002714, -0.00002379, -0.00001551, -0.00000513, 0.00000448, 0.00001126, 0.00001447, 0.00001455, 0.00001267, 0.00001004, 0.00000742, 0.00000493, 0.00000219, -0.00000121, -0.00000536, -0.00000971, -0.00001320, -0.00001464, -0.00001315, -0.00000862, -0.00000184, 0.00000569, 0.00001218, 0.00001619, 0.00001700, 0.00001480, 0.00001056, 0.00000564, 0.00000128, -0.00000179, -0.00000343, -0.00000406, -0.00000428, -0.00000460, -0.00000521, -0.00000592, -0.00000636, -0.00000619, -0.00000533, },
		{ 0.00000021, 0.00000021, 0.00000069, 0.00000173, 0.00000305, 0.00000401, 0.00000399, 0.00000265, 0.00000018, -0.00000263, -0.00000471, -0.00000512, -0.00000353, -0.00000047, 0.00000273, 0.00000442, 0.00000322, -0.00000131, -0.00000838, -0.00001607, -0.00002196, -0.00002397, -0.00002116, -0.00001420, -0.00000526, 0.00000272, 0.00000704, 0.00000620, 0.00000053, -0.00000792, -0.00001605, -0.00002090, -0.00002070, -0.00001555, -0.00000739, 0.00000075, 0.00000595, 0.00000657, 0.00000296, -0.00000267, -0.00000706, -0.00000723, -0.00000180, 0.00000833, 0.00002005, 0.00002915, 0.00003181, 0.00002617, 0.00001321, -0.00000334, -0.00001817, -0.00002615, -0.00002422, -0.00001261, 0.00000509, 0.00002297, 0.00003477, 0.00003602, 0.00002561, 0.00000627, -0.00001630, -0.00003528, -0.00004497, -0.00004272, -0.00002973, -0.00001055, 0.00000849, 0.00002150, 0.00002497, 0.00001889, 0.00000652, -0.00000682, -0.00001582, -0.00001703, -0.00001007, 0.00000232, 0.00001534, 0.00002391, 0.00002448, 0.00001634, 0.00000177, -0.00001471, -0.00002792, -0.00003374, -0.00003043, -0.00001908, -0.00000310, 0.00001299, 0.00002502, 0.00003041, 0.00002876, 0.00002161, 0.00001171, 0.00000191, -0.00000571, -0.00001033, -0.00001228, -0.00001255, -0.00001211, -0.00001145, -0.00001035, -0.00000818, -0.00000433, 0.00000127, 0.00000792, 0.00001423, 0.00001856, 0.00001960, 0.00001689, 0.00001102, 0.00000346, -0.00000388, -0.00000932, -0.00001186, -0.00001151, -0.00000907, -0.00000585, -0.00000307, -0.00000148, -0.00000118, -0.00000172, -0.00000242, -0.00000269, -0.00000230, -0.00000142, -0.00000047, 0.00000016, 0.00000033, },
		{ 0.00000312, 0.00000379, 0.00000353, 0.00000213, -0.00000015, -0.00000262, -0.00000445, -0.00000508, -0.00000450, -0.00000339, -0.00000281, -0.00000382, -0.00000688, -0.00001157, -0.00001653, -0.00001991, -0.00002006, -0.00001625, -0.00000910, -0.00000057, 0.00000662, 0.00000987, 0.00000768, 0.00000030, -0.00001026, -0.00002080, -0.00002796, -0.00002943, -0.00002467, -0.00001517, -0.00000389, 0.00000580, 0.00001135, 0.00001186, 0.00000839, 0.00000341, -0.00000018, -0.00000027, 0.00000358, 0.00001002, 0.00001641, 0.00001994, 0.00001884, 0.00001322, 0.00000509, -0.00000228, -0.00000572, -0.00000345, 0.00000401, 0.00001384, 0.00002184, 0.00002391, 0.00001773, 0.00000379, -0.00001447, -0.00003169, -0.00004230, -0.00004247, -0.00003151, -0.00001230, 0.00000962, 0.00002780, 0.00003703, 0.00003505, 0.00002327, 0.00000625, -0.00000995, -0.00001990, -0.00002064, -0.00001268, 0.00000027, 0.00001276, 0.00001956, 0.00001759, 0.00000706, -0.00000859, -0.00002390, -0.00003329, -0.00003299, -0.00002235, -0.00000408, 0.00001671, 0.00003413, 0.00004341, 0.00004234, 0.00003175, 0.00001508, -0.00000292, -0.00001773, -0.00002635, -0.00002798, -0.00002385, -0.00001645, -0.00000842, -0.00000171, 0.00000298, 0.00000600, 0.00000828, 0.00001061, 0.00001313, 0.00001520, 0.00001572, 0.00001368, 0.00000873, 0.00000151, -0.00000643, -0.00001314, -0.00001689, -0.00001682, -0.00001321, -0.00000739, -0.00000123, 0.00000348, 0.00000562, 0.00000506, 0.00000257, -0.00000060, -0.00000317, -0.00000437, -0.00000409, -0.00000279, -0.00000122, -0.00000002, 0.00000055, 0.00000067, 0.00000075, 0.00000118, 0.00000207, },
		{ -0.00000329, -0.00000621, -0.00000841, -0.00000941, -0.00000925, -0.00000854, -0.00000812, -0.00000869, -0.00001036, -0.00001254, -0.00001407, -0.00001374, -0.00001084, -0.00000566, 0.00000043, 0.00000534, 0.00000694, 0.00000392, -0.00000359, -0.00001384, -0.00002397, -0.00003083, -0.00003208, -0.00002702, -0.00001691, -0.00000458, 0.00000651, 0.00001343, 0.00001476, 0.00001108, 0.00000464, -0.00000148, -0.00000450, -0.00000297, 0.00000275, 0.00001076, 0.00001836, 0.00002318, 0.00002402, 0.00002126, 0.00001657, 0.00001206, 0.00000934, 0.00000868, 0.00000896, 0.00000808, 0.00000404, -0.00000398, -0.00001495, -0.00002601, -0.00003332, -0.00003342, -0.00002469, -0.00000822, 0.00001223, 0.00003116, 0.00004311, 0.00004449, 0.00003481, 0.00001705, -0.00000334, -0.00002013, -0.00002848, -0.00002648, -0.00001585, -0.00000122, 0.00001152, 0.00001734, 0.00001394, 0.00000257, -0.00001238, -0.00002492, -0.00002963, -0.00002361, -0.00000763, 0.00001404, 0.00003498, 0.00004872, 0.00005083, 0.00004034, 0.00002006, -0.00000444, -0.00002662, -0.00004094, -0.00004449, -0.00003760, -0.00002342, -0.00000661, 0.00000822, 0.00001792, 0.00002147, 0.00001991, 0.00001548, 0.00001058, 0.00000681, 0.00000454, 0.00000308, 0.00000123, -0.00000190, -0.00000635, -0.00001119, -0.00001489, -0.00001594, -0.00001357, -0.00000814, -0.00000110, 0.00000548, 0.00000967, 0.00001031, 0.00000745, 0.00000227, -0.00000338, -0.00000766, -0.00000940, -0.00000837, -0.00000533, -0.00000155, 0.00000166, 0.00000350, 0.00000385, 0.00000322, 0.00000237, 0.00000192, 0.00000205, 0.00000244, 0.00000246, 0.00000156, -0.00000047, },
		{ -0.00001114, -0.00000940, -0.00000786, -0.00000724, -0.00000763, -0.00000854, -0.00000903, -0.00000825, -0.00000594, -0.00000263, 0.00000035, 0.00000142, -0.00000068, -0.00000617, -0.00001404, -0.00002218, -0.00002806, -0.00002960, -0.00002595, -0.00001791, -0.00000770, 0.00000172, 0.00000773, 0.00000893, 0.00000571, 0.00000011, -0.00000492, -0.00000659, -0.00000327, 0.00000491, 0.00001607, 0.00002724, 0.00003541, 0.00003847, 0.00003592, 0.00002879, 0.00001917, 0.00000934, 0.00000094, -0.00000543, -0.00001016, -0.00001404, -0.00001757, -0.00002039, -0.00002123, -0.00001839, -0.00001059, 0.00000209, 0.00001778, 0.00003299, 0.00004350, 0.00004577, 0.00003825, 0.00002220, 0.00000157, -0.00001809, -0.00003136, -0.00003472, -0.00002778, -0.00001349, 0.00000285, 0.00001540, 0.00001985, 0.00001493, 0.00000297, -0.00001096, -0.00002080, -0.00002177, -0.00001215, 0.00000590, 0.00002693, 0.00004396, 0.00005080, 0.00004421, 0.00002508, -0.00000175, -0.00002896, -0.00004907, -0.00005670, -0.00005023, -0.00003211, -0.00000794, 0.00001543, 0.00003201, 0.00003840, 0.00003454, 0.00002327, 0.00000909, -0.00000353, -0.00001155, -0.00001410, -0.00001236, -0.00000867, -0.00000545, -0.00000415, -0.00000483, -0.00000638, -0.00000715, -0.00000588, -0.00000231, 0.00000263, 0.00000718, 0.00000953, 0.00000853, 0.00000426, -0.00000198, -0.00000814, -0.00001215, -0.00001271, -0.00000974, -0.00000433, 0.00000165, 0.00000635, 0.00000850, 0.00000789, 0.00000527, 0.00000196, -0.00000074, -0.00000208, -0.00000209, -0.00000147, -0.00000121, -0.00000205, -0.00000419, -0.00000714, -0.00001000, -0.00001186, -0.00001219, },
		{ -0.00000521, -0.00000661, -0.00000780, -0.00000791, -0.00000666, -0.00000460, -0.00000290, -0.00000290, -0.00000545, -0.00001049, -0.00001691, -0.00002290, -0.00002656, -0.00002666, -0.00002311, -0.00001704, -0.00001037, -0.00000508, -0.00000241, -0.00000231, -0.00000345, -0.00000366, -0.00000088, 0.00000603, 0.00001668, 0.00002910, 0.00004032, 0.00004728, 0.00004786, 0.00004157, 0.00002969, 0.00001485, 0.00000017, -0.00001163, -0.00001896, -0.00002149, -0.00001997, -0.00001558, -0.00000936, -0.00000188, 0.00000670, 0.00001623, 0.00002610, 0.00003486, 0.00004041, 0.00004057, 0.00003392, 0.00002070, 0.00000310, -0.00001498, -0.00002899, -0.00003513, -0.00003169, -0.00001980, -0.00000322, 0.00001272, 0.00002292, 0.00002429, 0.00001682, 0.00000378, -0.00000947, -0.00001729, -0.00001591, -0.00000488, 0.00001253, 0.00003027, 0.00004157, 0.00004129, 0.00002783, 0.00000394, -0.00002398, -0.00004784, -0.00006038, -0.00005761, -0.00004003, -0.00001259, 0.00001693, 0.00004035, 0.00005157, 0.00004837, 0.00003284, 0.00001047, -0.00001179, -0.00002778, -0.00003397, -0.00003022, -0.00001943, -0.00000622, 0.00000489, 0.00001090, 0.00001114, 0.00000711, 0.00000155, -0.00000281, -0.00000432, -0.00000296, -0.00000009, 0.00000230, 0.00000254, 0.00000010, -0.00000423, -0.00000867, -0.00001121, -0.00001051, -0.00000643, -0.00000020, 0.00000613, 0.00001043, 0.00001130, 0.00000858, 0.00000339, -0.00000238, -0.00000682, -0.00000870, -0.00000792, -0.00000541, -0.00000275, -0.00000140, -0.00000217, -0.00000487, -0.00000851, -0.00001169, -0.00001325, -0.00001273, -0.00001055, -0.00000775, -0.00000550, -0.00000463, },
		{ -0.00000785, -0.00000552, -0.00000359, -0.00000358, -0.00000644, -0.00001207, -0.00001933, -0.00002639, -0.00003139, -0.00003307, -0.00003120, -0.00002649, -0.00002026, -0.00001375, -0.00000765, -0.00000182, 0.00000452, 0.00001227, 0.00002182, 0.00003257, 0.00004282, 0.00005023, 0.00005252, 0.00004831, 0.00003774, 0.00002262, 0.00000597, -0.00000873, -0.00001858, -0.00002195, -0.00001883, -0.00001062, 0.00000043, 0.00001199, 0.00002224, 0.00003011, 0.00003521, 0.00003745, 0.00003673, 0.00003278, 0.00002533, 0.00001457, 0.00000150, -0.00001185, -0.00002272, -0.00002833, -0.00002684, -0.00001819, -0.00000441, 0.00001068, 0.00002263, 0.00002774, 0.00002442, 0.00001389, -0.00000006, -0.00001230, -0.00001811, -0.00001497, -0.00000362, 0.00001194, 0.00002575, 0.00003187, 0.00002660, 0.00001000, -0.00001381, -0.00003776, -0.00005408, -0.00005686, -0.00004415, -0.00001881, 0.00001221, 0.00004001, 0.00005648, 0.00005687, 0.00004132, 0.00001469, -0.00001495, -0.00003898, -0.00005086, -0.00004807, -0.00003265, -0.00001024, 0.00001191, 0.00002738, 0.00003248, 0.00002716, 0.00001461, -0.00000018, -0.00001227, -0.00001843, -0.00001796, -0.00001257, -0.00000538, 0.00000047, 0.00000302, 0.00000208, -0.00000094, -0.00000382, -0.00000466, -0.00000267, 0.00000147, 0.00000597, 0.00000870, 0.00000819, 0.00000422, -0.00000202, -0.00000839, -0.00001268, -0.00001347, -0.00001069, -0.00000563, -0.00000043, 0.00000275, 0.00000260, -0.00000088, -0.00000642, -0.00001201, -0.00001569, -0.00001632, -0.00001396, -0.00000978, -0.00000551, -0.00000276, -0.00000232, -0.00000394, -0.00000652, -0.00000862, -0.00000914, },
		{ -0.00000948, -0.00001796, -0.00002827, -0.00003790, -0.00004439, -0.00004603, -0.00004236, -0.00003404, -0.00002253, -0.00000949, 0.00000375, 0.00001637, 0.00002798, 0.00003830, 0.00004682, 0.00005260, 0.00005447, 0.00005143, 0.00004321, 0.00003065, 0.00001580, 0.00000153, -0.00000914, -0.00001392, -0.00001190, -0.00000379, 0.00000828, 0.00002139, 0.00003267, 0.00004004, 0.00004257, 0.00004047, 0.00003476, 0.00002672, 0.00001754, 0.00000810, -0.00000092, -0.00000880, -0.00001462, -0.00001729, -0.00001588, -0.00001013, -0.00000083, 0.00000998, 0.00001942, 0.00002454, 0.00002335, 0.00001568, 0.00000348, -0.00000957, -0.00001921, -0.00002203, -0.00001686, -0.00000536, 0.00000829, 0.00001874, 0.00002132, 0.00001381, -0.00000240, -0.00002256, -0.00003990, -0.00004786, -0.00004243, -0.00002381, 0.00000338, 0.00003131, 0.00005132, 0.00005678, 0.00004533, 0.00001983, -0.00001240, -0.00004183, -0.00005968, -0.00006071, -0.00004484, -0.00001712, 0.00001394, 0.00003914, 0.00005144, 0.00004808, 0.00003116, 0.00000672, -0.00001743, -0.00003430, -0.00003984, -0.00003392, -0.00001994, -0.00000335, 0.00001041, 0.00001767, 0.00001754, 0.00001178, 0.00000375, -0.00000306, -0.00000638, -0.00000587, -0.00000294, 0.00000006, 0.00000104, -0.00000094, -0.00000521, -0.00000993, -0.00001284, -0.00001239, -0.00000839, -0.00000222, 0.00000374, 0.00000699, 0.00000596, 0.00000065, -0.00000733, -0.00001535, -0.00002073, -0.00002173, -0.00001821, -0.00001161, -0.00000438, 0.00000098, 0.00000283, 0.00000099, -0.00000332, -0.00000804, -0.00001113, -0.00001145, -0.00000923, -0.00000595, -0.00000375, -0.00000460, },
		{ -0.00005948, -0.00005717, -0.00004713, -0.00003096, -0.00001132, 0.00000892, 0.00002731, 0.00004218, 0.00005266, 0.00005851, 0.00005978, 0.00005671, 0.00004970, 0.00003953, 0.00002752, 0.00001556, 0.00000582, 0.00000031, 0.00000030, 0.00000584, 0.00001561, 0.00002725, 0.00003788, 0.00004495, 0.00004688, 0.00004344, 0.00003567, 0.00002545, 0.00001490, 0.00000573, -0.00000101, -0.00000504, -0.00000657, -0.00000602, -0.00000374, -0.00000004, 0.00000468, 0.00000971, 0.00001392, 0.00001587, 0.00001430, 0.00000864, -0.00000046, -0.00001107, -0.00002035, -0.00002538, -0.00002426, -0.00001695, -0.00000556, 0.00000610, 0.00001374, 0.00001411, 0.00000633, -0.00000748, -0.00002268, -0.00003349, -0.00003506, -0.00002528, -0.00000591, 0.00001765, 0.00003791, 0.00004770, 0.00004269, 0.00002308, -0.00000608, -0.00003626, -0.00005804, -0.00006417, -0.00005200, -0.00002450, 0.00001046, 0.00004258, 0.00006230, 0.00006381, 0.00004687, 0.00001681, -0.00001721, -0.00004518, -0.00005933, -0.00005645, -0.00003861, -0.00001223, 0.00001425, 0.00003316, 0.00003990, 0.00003410, 0.00001927, 0.00000124, -0.00001405, -0.00002252, -0.00002305, -0.00001740, -0.00000913, -0.00000202, 0.00000143, 0.00000078, -0.00000249, -0.00000585, -0.00000706, -0.00000513, -0.00000084, 0.00000367, 0.00000587, 0.00000396, -0.00000223, -0.00001106, -0.00001965, -0.00002491, -0.00002483, -0.00001920, -0.00000983, 0.00000016, 0.00000742, 0.00000964, 0.00000632, -0.00000101, -0.00000950, -0.00001602, -0.00001840, -0.00001628, -0.00001117, -0.00000597, -0.00000381, -0.00000693, -0.00001576, -0.00002874, -0.00004275, -0.00005407, },
		{ 0.00001149, 0.00003650, 0.00005564, 0.00006737, 0.00007163, 0.00006946, 0.00006250, 0.00005255, 0.00004137, 0.00003061, 0.00002178, 0.00001615, 0.00001461, 0.00001737, 0.00002379, 0.00003235, 0.00004093, 0.00004725, 0.00004952, 0.00004693, 0.00003990, 0.00002995, 0.00001920, 0.00000980, 0.00000325, 0.00000013, 0.00000003, 0.00000187, 0.00000439, 0.00000651, 0.00000761, 0.00000751, 0.00000625, 0.00000387, 0.00000034, -0.00000444, -0.00001031, -0.00001667, -0.00002231, -0.00002569, -0.00002540, -0.00002078, -0.00001244, -0.00000236, 0.00000655, 0.00001134, 0.00001011, 0.00000293, -0.00000792, -0.00001850, -0.00002445, -0.00002259, -0.00001230, 0.00000390, 0.00002090, 0.00003252, 0.00003367, 0.00002234, 0.00000068, -0.00002528, -0.00004731, -0.00005761, -0.00005150, -0.00002929, 0.00000346, 0.00003737, 0.00006209, 0.00006960, 0.00005684, 0.00002695, -0.00001161, -0.00004759, -0.00007043, -0.00007356, -0.00005641, -0.00002450, 0.00001236, 0.00004326, 0.00005960, 0.00005757, 0.00003912, 0.00001097, -0.00001784, -0.00003894, -0.00004717, -0.00004191, -0.00002680, -0.00000801, 0.00000805, 0.00001690, 0.00001726, 0.00001100, 0.00000196, -0.00000575, -0.00000945, -0.00000872, -0.00000531, -0.00000210, -0.00000170, -0.00000520, -0.00001168, -0.00001853, -0.00002264, -0.00002167, -0.00001521, -0.00000503, 0.00000544, 0.00001245, 0.00001326, 0.00000731, -0.00000353, -0.00001560, -0.00002480, -0.00002807, -0.00002459, -0.00001615, -0.00000656, -0.00000033, -0.00000108, -0.00001022, -0.00002639, -0.00004578, -0.00006322, -0.00007377, -0.00007404, -0.00006312, -0.00004268, -0.00001631, },
		{ 0.00008451, 0.00007341, 0.00005996, 0.00004701, 0.00003662, 0.00002996, 0.00002742, 0.00002877, 0.00003325, 0.00003964, 0.00004641, 0.00005184, 0.00005443, 0.00005316, 0.00004789, 0.00003938, 0.00002917, 0.00001916, 0.00001107, 0.00000593, 0.00000381, 0.00000391, 0.00000484, 0.00000518, 0.00000393, 0.00000076, -0.00000393, -0.00000933, -0.00001454, -0.00001888, -0.00002204, -0.00002397, -0.00002469, -0.00002411, -0.00002196, -0.00001796, -0.00001213, -0.00000507, 0.00000192, 0.00000714, 0.00000897, 0.00000658, 0.00000048, -0.00000741, -0.00001423, -0.00001708, -0.00001417, -0.00000573, 0.00000575, 0.00001616, 0.00002106, 0.00001740, 0.00000489, -0.00001343, -0.00003185, -0.00004365, -0.00004345, -0.00002929, -0.00000377, 0.00002630, 0.00005177, 0.00006399, 0.00005778, 0.00003339, -0.00000308, -0.00004137, -0.00007001, -0.00008001, -0.00006775, -0.00003634, 0.00000522, 0.00004472, 0.00007055, 0.00007527, 0.00005789, 0.00002408, -0.00001574, -0.00004977, -0.00006852, -0.00006765, -0.00004898, -0.00001958, 0.00001084, 0.00003320, 0.00004184, 0.00003599, 0.00001953, -0.00000083, -0.00001808, -0.00002736, -0.00002733, -0.00002013, -0.00001020, -0.00000218, 0.00000085, -0.00000153, -0.00000718, -0.00001258, -0.00001447, -0.00001133, -0.00000412, 0.00000413, 0.00000955, 0.00000909, 0.00000186, -0.00001030, -0.00002344, -0.00003287, -0.00003500, -0.00002881, -0.00001640, -0.00000251, 0.00000709, 0.00000751, -0.00000349, -0.00002453, -0.00005091, -0.00007584, -0.00009242, -0.00009557, -0.00008343, -0.00005772, -0.00002314, 0.00001399, 0.00004743, 0.00007243, 0.00008658, 0.00008994, },
		{ 0.00003825, 0.00003928, 0.00004418, 0.00005107, 0.00005807, 0.00006352, 0.00006614, 0.00006512, 0.00006026, 0.00005196, 0.00004124, 0.00002956, 0.00001849, 0.00000933, 0.00000277, -0.00000128, -0.00000363, -0.00000547, -0.00000792, -0.00001166, -0.00001661, -0.00002207, -0.00002694, -0.00003013, -0.00003099, -0.00002945, -0.00002599, -0.00002141, -0.00001647, -0.00001168, -0.00000726, -0.00000321, 0.00000038, 0.00000319, 0.00000468, 0.00000425, 0.00000168, -0.00000262, -0.00000746, -0.00001115, -0.00001207, -0.00000940, -0.00000359, 0.00000344, 0.00000890, 0.00001003, 0.00000530, -0.00000469, -0.00001706, -0.00002731, -0.00003078, -0.00002446, -0.00000838, 0.00001379, 0.00003555, 0.00004943, 0.00004949, 0.00003363, 0.00000478, -0.00002948, -0.00005899, -0.00007409, -0.00006889, -0.00004345, -0.00000422, 0.00003767, 0.00006966, 0.00008172, 0.00006960, 0.00003635, -0.00000850, -0.00005180, -0.00008083, -0.00008732, -0.00006991, -0.00003449, 0.00000778, 0.00004413, 0.00006417, 0.00006304, 0.00004263, 0.00001061, -0.00002240, -0.00004642, -0.00005519, -0.00004792, -0.00002900, -0.00000606, 0.00001294, 0.00002242, 0.00002086, 0.00001083, -0.00000245, -0.00001336, -0.00001810, -0.00001601, -0.00000956, -0.00000304, -0.00000060, -0.00000439, -0.00001358, -0.00002462, -0.00003272, -0.00003378, -0.00002632, -0.00001234, 0.00000310, 0.00001341, 0.00001283, -0.00000146, -0.00002803, -0.00006132, -0.00009305, -0.00011452, -0.00011910, -0.00010412, -0.00007153, -0.00002729, 0.00002032, 0.00006290, 0.00009397, 0.00011026, 0.00011199, 0.00010232, 0.00008602, 0.00006814, 0.00005281, 0.00004257, },
		{ 0.00008679, 0.00008431, 0.00007692, 0.00006544, 0.00005109, 0.00003536, 0.00001978, 0.00000572, -0.00000581, -0.00001436, -0.00002009, -0.00002367, -0.00002607, -0.00002816, -0.00003044, -0.00003285, -0.00003479, -0.00003542, -0.00003402, -0.00003030, -0.00002460, -0.00001782, -0.00001111, -0.00000555, -0.00000177, 0.00000016, 0.00000063, 0.00000025, -0.00000046, -0.00000127, -0.00000224, -0.00000356, -0.00000523, -0.00000695, -0.00000809, -0.00000794, -0.00000611, -0.00000289, 0.00000066, 0.00000302, 0.00000271, -0.00000088, -0.00000706, -0.00001372, -0.00001790, -0.00001687, -0.00000932, 0.00000376, 0.00001890, 0.00003103, 0.00003501, 0.00002762, 0.00000907, -0.00001646, -0.00004168, -0.00005824, -0.00005952, -0.00004308, -0.00001206, 0.00002534, 0.00005795, 0.00007514, 0.00007026, 0.00004315, 0.00000064, -0.00004526, -0.00008087, -0.00009514, -0.00008321, -0.00004817, -0.00000025, 0.00004627, 0.00007752, 0.00008433, 0.00006510, 0.00002622, -0.00002007, -0.00005965, -0.00008099, -0.00007868, -0.00005484, -0.00001811, 0.00001943, 0.00004635, 0.00005542, 0.00004564, 0.00002204, -0.00000651, -0.00003056, -0.00004336, -0.00004288, -0.00003197, -0.00001683, -0.00000433, 0.00000069, -0.00000286, -0.00001214, -0.00002186, -0.00002660, -0.00002327, -0.00001263, 0.00000077, 0.00001009, 0.00000875, -0.00000697, -0.00003617, -0.00007314, -0.00010869, -0.00013280, -0.00013756, -0.00011948, -0.00008059, -0.00002785, 0.00002880, 0.00007909, 0.00011503, 0.00013262, 0.00013240, 0.00011869, 0.00009798, 0.00007697, 0.00006092, 0.00005267, 0.00005250, 0.00005858, 0.00006793, 0.00007737, 0.00008424, },
		{ 0.00001005, -0.00000903, -0.00002434, -0.00003541, -0.00004229, -0.00004552, -0.00004595, -0.00004460, -0.00004237, -0.00003987, -0.00003728, -0.00003436, -0.00003071, -0.00002595, -0.00002009, -0.00001356, -0.00000722, -0.00000208, 0.00000104, 0.00000182, 0.00000053, -0.00000200, -0.00000475, -0.00000680, -0.00000767, -0.00000738, -0.00000639, -0.00000528, -0.00000445, -0.00000400, -0.00000368, -0.00000318, -0.00000234, -0.00000137, -0.00000083, -0.00000134, -0.00000319, -0.00000602, -0.00000870, -0.00000966, -0.00000753, -0.00000189, 0.00000624, 0.00001438, 0.00001923, 0.00001785, 0.00000893, -0.00000626, -0.00002376, -0.00003788, -0.00004294, -0.00003536, -0.00001539, 0.00001238, 0.00003994, 0.00005812, 0.00005961, 0.00004169, 0.00000767, -0.00003350, -0.00006967, -0.00008915, -0.00008460, -0.00005573, -0.00000990, 0.00003978, 0.00007831, 0.00009350, 0.00007997, 0.00004107, -0.00001192, -0.00006315, -0.00009713, -0.00010359, -0.00008071, -0.00003568, 0.00001761, 0.00006302, 0.00008721, 0.00008372, 0.00005480, 0.00001029, -0.00003578, -0.00006986, -0.00008311, -0.00007391, -0.00004784, -0.00001529, 0.00001239, 0.00002681, 0.00002512, 0.00001048, -0.00000959, -0.00002642, -0.00003361, -0.00002951, -0.00001774, -0.00000579, -0.00000206, -0.00001255, -0.00003833, -0.00007467, -0.00011221, -0.00013975, -0.00014766, -0.00013087, -0.00009048, -0.00003349, 0.00002927, 0.00008612, 0.00012765, 0.00014897, 0.00015048, 0.00013714, 0.00011650, 0.00009638, 0.00008273, 0.00007834, 0.00008268, 0.00009263, 0.00010377, 0.00011184, 0.00011373, 0.00010807, 0.00009519, 0.00007666, 0.00005475, 0.00003185, },
		{ -0.00006230, -0.00005588, -0.00004851, -0.00004112, -0.00003434, -0.00002840, -0.00002322, -0.00001851, -0.00001399, -0.00000957, -0.00000544, -0.00000211, -0.00000012, 0.00000010, -0.00000145, -0.00000430, -0.00000755, -0.00001017, -0.00001132, -0.00001066, -0.00000843, -0.00000538, -0.00000245, -0.00000044, 0.00000026, -0.00000020, -0.00000133, -0.00000253, -0.00000338, -0.00000380, -0.00000403, -0.00000435, -0.00000487, -0.00000532, -0.00000510, -0.00000360, -0.00000065, 0.00000322, 0.00000668, 0.00000797, 0.00000562, -0.00000074, -0.00000989, -0.00001904, -0.00002455, -0.00002321, -0.00001364, 0.00000271, 0.00002146, 0.00003637, 0.00004126, 0.00003229, 0.00000979, -0.00002118, -0.00005179, -0.00007199, -0.00007380, -0.00005425, -0.00001705, 0.00002796, 0.00006733, 0.00008819, 0.00008244, 0.00004978, -0.00000154, -0.00005685, -0.00009933, -0.00011528, -0.00009860, -0.00005305, 0.00000851, 0.00006795, 0.00010737, 0.00011464, 0.00008733, 0.00003341, -0.00003116, -0.00008746, -0.00011952, -0.00011913, -0.00008824, -0.00003796, 0.00001527, 0.00005509, 0.00007041, 0.00005864, 0.00002599, -0.00001507, -0.00005042, -0.00006918, -0.00006727, -0.00004857, -0.00002322, -0.00000382, -0.00000088, -0.00001903, -0.00005524, -0.00009968, -0.00013875, -0.00015935, -0.00015287, -0.00011772, -0.00005968, 0.00001002, 0.00007807, 0.00013277, 0.00016703, 0.00017964, 0.00017477, 0.00016000, 0.00014358, 0.00013189, 0.00012790, 0.00013080, 0.00013690, 0.00014122, 0.00013925, 0.00012830, 0.00010803, 0.00008034, 0.00004856, 0.00001649, -0.00001250, -0.00003604, -0.00005295, -0.00006313, -0.00006727, -0.00006655, },
		{ -0.00001493, -0.00001164, -0.00000937, -0.00000768, -0.00000629, -0.00000516, -0.00000443, -0.00000436, -0.00000516, -0.00000677, -0.00000884, -0.00001072, -0.00001168, -0.00001118, -0.00000909, -0.00000580, -0.00000212, 0.00000096, 0.00000262, 0.00000252, 0.00000087, -0.00000165, -0.00000416, -0.00000593, -0.00000660, -0.00000629, -0.00000543, -0.00000451, -0.00000381, -0.00000330, -0.00000269, -0.00000169, -0.00000030, 0.00000104, 0.00000160, 0.00000063, -0.00000209, -0.00000596, -0.00000954, -0.00001092, -0.00000854, -0.00000202, 0.00000726, 0.00001626, 0.00002109, 0.00001835, 0.00000671, -0.00001208, -0.00003313, -0.00004954, -0.00005460, -0.00004422, -0.00001895, 0.00001544, 0.00004910, 0.00007085, 0.00007189, 0.00004906, 0.00000665, -0.00004413, -0.00008811, -0.00011071, -0.00010274, -0.00006393, -0.00000369, 0.00006107, 0.00011089, 0.00012967, 0.00010997, 0.00005572, -0.00001847, -0.00009149, -0.00014209, -0.00015541, -0.00012766, -0.00006743, 0.00000696, 0.00007306, 0.00011131, 0.00011100, 0.00007348, 0.00001147, -0.00005521, -0.00010608, -0.00012654, -0.00011239, -0.00007096, -0.00001846, 0.00002538, 0.00004374, 0.00002811, -0.00001903, -0.00008502, -0.00015086, -0.00019688, -0.00020841, -0.00017972, -0.00011514, -0.00002730, 0.00006685, 0.00015078, 0.00021255, 0.00024711, 0.00025635, 0.00024724, 0.00022874, 0.00020877, 0.00019198, 0.00017909, 0.00016757, 0.00015335, 0.00013277, 0.00010409, 0.00006818, 0.00002823, -0.00001126, -0.00004574, -0.00007178, -0.00008770, -0.00009366, -0.00009126, -0.00008292, -0.00007120, -0.00005834, -0.00004599, -0.00003518, -0.00002640, -0.00001972, },
		{ -0.00000988, -0.00000929, -0.00000875, -0.00000850, -0.00000861, -0.00000895, -0.00000918, -0.00000888, -0.00000769, -0.00000557, -0.00000281, -0.00000005, 0.00000196, 0.00000259, 0.00000160, -0.00000074, -0.00000373, -0.00000648, -0.00000821, -0.00000855, -0.00000762, -0.00000594, -0.00000419, -0.00000293, -0.00000234, -0.00000224, -0.00000220, -0.00000184, -0.00000108, -0.00000017, 0.00000040, 0.00000022, -0.00000078, -0.00000213, -0.00000302, -0.00000262, -0.00000065, 0.00000228, 0.00000468, 0.00000463, 0.00000058, -0.00000767, -0.00001846, -0.00002839, -0.00003319, -0.00002930, -0.00001550, 0.00000601, 0.00002958, 0.00004743, 0.00005211, 0.00003922, 0.00000962, -0.00002993, -0.00006809, -0.00009213, -0.00009203, -0.00006426, -0.00001384, 0.00004623, 0.00009825, 0.00012510, 0.00011579, 0.00006954, -0.00000300, -0.00008231, -0.00014536, -0.00017269, -0.00015462, -0.00009480, -0.00000970, 0.00007593, 0.00013646, 0.00015317, 0.00012026, 0.00004685, -0.00004550, -0.00012932, -0.00017958, -0.00018149, -0.00013519, -0.00005580, 0.00003100, 0.00009673, 0.00011872, 0.00008696, 0.00000720, -0.00010063, -0.00020798, -0.00028563, -0.00031150, -0.00027626, -0.00018510, -0.00005555, 0.00008772, 0.00021953, 0.00032014, 0.00037903, 0.00039574, 0.00037794, 0.00033771, 0.00028746, 0.00023663, 0.00019010, 0.00014842, 0.00010940, 0.00007027, 0.00002954, -0.00001203, -0.00005155, -0.00008502, -0.00010869, -0.00012030, -0.00011974, -0.00010901, -0.00009153, -0.00007122, -0.00005158, -0.00003503, -0.00002276, -0.00001481, -0.00001049, -0.00000877, -0.00000863, -0.00000920, -0.00000987, -0.00001028, -0.00001027, },
		{ -0.00000546, -0.00000546, -0.00000541, -0.00000498, -0.00000404, -0.00000268, -0.00000127, -0.00000029, -0.00000022, -0.00000125, -0.00000321, -0.00000558, -0.00000765, -0.00000879, -0.00000866, -0.00000736, -0.00000538, -0.00000340, -0.00000199, -0.00000140, -0.00000148, -0.00000178, -0.00000178, -0.00000119, -0.00000007, 0.00000114, 0.00000187, 0.00000167, 0.00000049, -0.00000131, -0.00000309, -0.00000430, -0.00000476, -0.00000485, -0.00000533, -0.00000694, -0.00000983, -0.00001324, -0.00001551, -0.00001461, -0.00000907, 0.00000108, 0.00001370, 0.00002479, 0.00002952, 0.00002401, 0.00000711, -0.00001835, -0.00004568, -0.00006586, -0.00007032, -0.00005408, -0.00001828, 0.00002906, 0.00007462, 0.00010336, 0.00010332, 0.00007002, 0.00000903, -0.00006460, -0.00013006, -0.00016677, -0.00016098, -0.00011085, -0.00002806, 0.00006468, 0.00013986, 0.00017342, 0.00015245, 0.00007992, -0.00002535, -0.00013352, -0.00021241, -0.00023722, -0.00019847, -0.00010563, 0.00001491, 0.00012696, 0.00019487, 0.00019399, 0.00011820, -0.00001771, -0.00018157, -0.00033215, -0.00042995, -0.00044733, -0.00037533, -0.00022532, -0.00002539, 0.00018720, 0.00037529, 0.00051016, 0.00057682, 0.00057533, 0.00051825, 0.00042544, 0.00031816, 0.00021398, 0.00012391, 0.00005197, -0.00000318, -0.00004551, -0.00007900, -0.00010586, -0.00012605, -0.00013803, -0.00014005, -0.00013147, -0.00011351, -0.00008926, -0.00006290, -0.00003866, -0.00001978, -0.00000785, -0.00000272, -0.00000291, -0.00000628, -0.00001068, -0.00001449, -0.00001679, -0.00001737, -0.00001649, -0.00001461, -0.00001227, -0.00000994, -0.00000795, -0.00000652, -0.00000573, },
		{ -0.00000547, -0.00000455, -0.00000363, -0.00000312, -0.00000327, -0.00000412, -0.00000540, -0.00000662, -0.00000729, -0.00000709, -0.00000601, -0.00000438, -0.00000269, -0.00000144, -0.00000090, -0.00000099, -0.00000135, -0.00000151, -0.00000115, -0.00000026, 0.00000080, 0.00000147, 0.00000120, -0.00000020, -0.00000252, -0.00000515, -0.00000734, -0.00000849, -0.00000847, -0.00000757, -0.00000641, -0.00000549, -0.00000496, -0.00000440, -0.00000307, -0.00000035, 0.00000369, 0.00000795, 0.00001040, 0.00000876, 0.00000147, -0.00001115, -0.00002640, -0.00003943, -0.00004456, -0.00003731, -0.00001651, 0.00001440, 0.00004737, 0.00007165, 0.00007694, 0.00005718, 0.00001348, -0.00004490, -0.00010225, -0.00014056, -0.00014514, -0.00010995, -0.00004080, 0.00004497, 0.00012259, 0.00016691, 0.00016037, 0.00009934, -0.00000326, -0.00012038, -0.00021795, -0.00026478, -0.00024242, -0.00015194, -0.00001522, 0.00012991, 0.00023963, 0.00027653, 0.00022079, 0.00007700, -0.00012564, -0.00033994, -0.00051234, -0.00059653, -0.00056525, -0.00041708, -0.00017661, 0.00011190, 0.00039545, 0.00062470, 0.00076457, 0.00080055, 0.00073943, 0.00060499, 0.00043030, 0.00024910, 0.00008854, -0.00003506, -0.00011689, -0.00016132, -0.00017791, -0.00017722, -0.00016764, -0.00015402, -0.00013792, -0.00011907, -0.00009700, -0.00007233, -0.00004712, -0.00002440, -0.00000720, 0.00000251, 0.00000446, 0.00000010, -0.00000789, -0.00001649, -0.00002315, -0.00002641, -0.00002599, -0.00002264, -0.00001763, -0.00001228, -0.00000763, -0.00000428, -0.00000237, -0.00000175, -0.00000212, -0.00000309, -0.00000429, -0.00000536, -0.00000599, -0.00000602, },
		{ -0.00000281, -0.00000340, -0.00000429, -0.00000507, -0.00000536, -0.00000500, -0.00000408, -0.00000292, -0.00000191, -0.00000133, -0.00000125, -0.00000147, -0.00000166, -0.00000154, -0.00000103, -0.00000036, 0.00000006, -0.00000021, -0.00000141, -0.00000338, -0.00000562, -0.00000741, -0.00000814, -0.00000752, -0.00000578, -0.00000352, -0.00000151, -0.00000031, -0.00000008, -0.00000051, -0.00000107, -0.00000137, -0.00000149, -0.00000203, -0.00000388, -0.00000765, -0.00001309, -0.00001867, -0.00002186, -0.00001986, -0.00001091, 0.00000452, 0.00002305, 0.00003879, 0.00004484, 0.00003575, 0.00000994, -0.00002859, -0.00007035, -0.00010245, -0.00011236, -0.00009241, -0.00004336, 0.00002420, 0.00009163, 0.00013706, 0.00014208, 0.00009835, 0.00001183, -0.00009703, -0.00019776, -0.00025821, -0.00025443, -0.00017926, -0.00004659, 0.00011019, 0.00024603, 0.00031614, 0.00028917, 0.00015745, -0.00005846, -0.00031298, -0.00054563, -0.00069583, -0.00071828, -0.00059482, -0.00033966, 0.00000359, 0.00037208, 0.00069836, 0.00092562, 0.00101961, 0.00097451, 0.00081167, 0.00057213, 0.00030551, 0.00005830, -0.00013538, -0.00025864, -0.00031181, -0.00030833, -0.00026842, -0.00021266, -0.00015711, -0.00011104, -0.00007714, -0.00005359, -0.00003676, -0.00002354, -0.00001254, -0.00000415, 0.00000032, -0.00000015, -0.00000544, -0.00001401, -0.00002330, -0.00003059, -0.00003384, -0.00003232, -0.00002666, -0.00001857, -0.00001011, -0.00000311, 0.00000132, 0.00000293, 0.00000217, -0.00000011, -0.00000298, -0.00000571, -0.00000783, -0.00000911, -0.00000951, -0.00000909, -0.00000801, -0.00000650, -0.00000491, -0.00000359, -0.00000285, },
		{ -0.00000515, -0.00000485, -0.00000406, -0.00000311, -0.00000230, -0.00000185, -0.00000175, -0.00000182, -0.00000183, -0.00000163, -0.00000126, -0.00000095, -0.00000101, -0.00000166, -0.00000288, -0.00000436, -0.00000559, -0.00000608, -0.00000558, -0.00000422, -0.00000250, -0.00000111, -0.00000062, -0.00000125, -0.00000276, -0.00000455, -0.00000595, -0.00000654, -0.00000636, -0.00000586, -0.00000565, -0.00000605, -0.00000676, -0.00000688, -0.00000521, -0.00000101, 0.00000536, 0.00001197, 0.00001564, 0.00001290, 0.00000151, -0.00001806, -0.00004183, -0.00006273, -0.00007243, -0.00006417, -0.00003581, 0.00000816, 0.00005647, 0.00009351, 0.00010381, 0.00007743, 0.00001460, -0.00007227, -0.00016028, -0.00022169, -0.00023202, -0.00017866, -0.00006692, 0.00007851, 0.00021803, 0.00030679, 0.00030728, 0.00020134, -0.00000194, -0.00026526, -0.00052994, -0.00072921, -0.00080474, -0.00072204, -0.00048053, -0.00011540, 0.00030952, 0.00071589, 0.00102873, 0.00119307, 0.00118545, 0.00101755, 0.00073132, 0.00038749, 0.00005078, -0.00022434, -0.00040385, -0.00047827, -0.00046040, -0.00037836, -0.00026612, -0.00015463, -0.00006553, -0.00000859, 0.00001711, 0.00001965, 0.00000968, -0.00000344, -0.00001392, -0.00001993, -0.00002257, -0.00002404, -0.00002607, -0.00002901, -0.00003185, -0.00003291, -0.00003078, -0.00002511, -0.00001677, -0.00000759, 0.00000030, 0.00000525, 0.00000651, 0.00000443, 0.00000017, -0.00000480, -0.00000915, -0.00001201, -0.00001315, -0.00001276, -0.00001131, -0.00000929, -0.00000709, -0.00000498, -0.00000316, -0.00000181, -0.00000107, -0.00000104, -0.00000167, -0.00000274, -0.00000392, -0.00000481, },
		{ -0.00000216, -0.00000170, -0.00000158, -0.00000163, -0.00000164, -0.00000150, -0.00000127, -0.00000113, -0.00000131, -0.00000193, -0.00000289, -0.00000392, -0.00000467, -0.00000488, -0.00000451, -0.00000381, -0.00000318, -0.00000305, -0.00000356, -0.00000456, -0.00000556, -0.00000601, -0.00000553, -0.00000412, -0.00000226, -0.00000068, -0.00000008, -0.00000072, -0.00000231, -0.00000408, -0.00000521, -0.00000530, -0.00000476, -0.00000483, -0.00000709, -0.00001257, -0.00002091, -0.00002990, -0.00003570, -0.00003413, -0.00002241, -0.00000101, 0.00002535, 0.00004813, 0.00005712, 0.00004388, 0.00000550, -0.00005269, -0.00011680, -0.00016701, -0.00018304, -0.00015099, -0.00006958, 0.00004645, 0.00016754, 0.00025547, 0.00027348, 0.00019765, 0.00002645, -0.00021492, -0.00047714, -0.00069662, -0.00081043, -0.00077260, -0.00056755, -0.00021664, 0.00022435, 0.00067635, 0.00105448, 0.00128718, 0.00133239, 0.00118654, 0.00088424, 0.00048902, 0.00007785, -0.00027659, -0.00052113, -0.00063144, -0.00061365, -0.00049862, -0.00033124, -0.00015791, -0.00001580, 0.00007384, 0.00010714, 0.00009445, 0.00005425, 0.00000629, -0.00003388, -0.00005789, -0.00006475, -0.00005875, -0.00004646, -0.00003382, -0.00002439, -0.00001890, -0.00001604, -0.00001375, -0.00001050, -0.00000600, -0.00000113, 0.00000261, 0.00000389, 0.00000219, -0.00000199, -0.00000733, -0.00001224, -0.00001539, -0.00001618, -0.00001476, -0.00001185, -0.00000842, -0.00000529, -0.00000293, -0.00000146, -0.00000072, -0.00000051, -0.00000068, -0.00000117, -0.00000198, -0.00000305, -0.00000421, -0.00000521, -0.00000577, -0.00000572, -0.00000508, -0.00000407, -0.00000300, },
		{ -0.00000147, -0.00000140, -0.00000126, -0.00000124, -0.00000154, -0.00000225, -0.00000327, -0.00000434, -0.00000517, -0.00000554, -0.00000543, -0.00000498, -0.00000445, -0.00000404, -0.00000377, -0.00000350, -0.00000299, -0.00000209, -0.00000090, 0.00000017, 0.00000051, -0.00000044, -0.00000286, -0.00000642, -0.00001026, -0.00001331, -0.00001468, -0.00001409, -0.00001206, -0.00000971, -0.00000823, -0.00000828, -0.00000947, -0.00001041, -0.00000923, -0.00000468, 0.00000285, 0.00001069, 0.00001426, 0.00000848, -0.00000999, -0.00004050, -0.00007703, -0.00010865, -0.00012208, -0.00010597, -0.00005591, 0.00002181, 0.00010900, 0.00017835, 0.00019988, 0.00014988, 0.00001992, -0.00017695, -0.00040520, -0.00061230, -0.00073991, -0.00073848, -0.00058141, -0.00027488, 0.00013989, 0.00059319, 0.00100107, 0.00128397, 0.00138512, 0.00128397, 0.00100107, 0.00059319, 0.00013989, -0.00027488, -0.00058141, -0.00073848, -0.00073991, -0.00061230, -0.00040520, -0.00017695, 0.00001992, 0.00014988, 0.00019988, 0.00017835, 0.00010900, 0.00002181, -0.00005591, -0.00010597, -0.00012208, -0.00010865, -0.00007703, -0.00004050, -0.00000999, 0.00000848, 0.00001426, 0.00001069, 0.00000285, -0.00000468, -0.00000923, -0.00001041, -0.00000947, -0.00000828, -0.00000823, -0.00000971, -0.00001206, -0.00001409, -0.00001468, -0.00001331, -0.00001026, -0.00000642, -0.00000286, -0.00000044, 0.00000051, 0.00000017, -0.00000090, -0.00000209, -0.00000299, -0.00000350, -0.00000377, -0.00000404, -0.00000445, -0.00000498, -0.00000543, -0.00000554, -0.00000517, -0.00000434, -0.00000327, -0.00000225, -0.00000154, -0.00000124, -0.00000126, -0.00000140, },
		{ -0.00000216, -0.00000300, -0.00000407, -0.00000508, -0.00000572, -0.00000577, -0.00000521, -0.00000421, -0.00000305, -0.00000198, -0.00000117, -0.00000068, -0.00000051, -0.00000072, -0.00000146, -0.00000293, -0.00000529, -0.00000842, -0.00001185, -0.00001476, -0.00001618, -0.00001539, -0.00001224, -0.00000733, -0.00000199, 0.00000219, 0.00000389, 0.00000261, -0.00000113, -0.00000600, -0.00001050, -0.00001375, -0.00001604, -0.00001890, -0.00002439, -0.00003382, -0.00004646, -0.00005875, -0.00006475, -0.00005789, -0.00003388, 0.00000629, 0.00005425, 0.00009445, 0.00010714, 0.00007384, -0.00001580, -0.00015791, -0.00033124, -0.00049862, -0.00061365, -0.00063144, -0.00052113, -0.00027659, 0.00007785, 0.00048902, 0.00088424, 0.00118654, 0.00133239, 0.00128718, 0.00105448, 0.00067635, 0.00022435, -0.00021664, -0.00056755, -0.00077260, -0.00081043, -0.00069662, -0.00047714, -0.00021492, 0.00002645, 0.00019765, 0.00027348, 0.00025547, 0.00016754, 0.00004645, -0.00006958, -0.00015099, -0.00018304, -0.00016701, -0.00011680, -0.00005269, 0.00000550, 0.00004388, 0.00005712, 0.00004813, 0.00002535, -0.00000101, -0.00002241, -0.00003413, -0.00003570, -0.00002990, -0.00002091, -0.00001257, -0.00000709, -0.00000483, -0.00000476, -0.00000530, -0.00000521, -0.00000408, -0.00000231, -0.00000072, -0.00000008, -0.00000068, -0.00000226, -0.00000412, -0.00000553, -0.00000601, -0.00000556, -0.00000456, -0.00000356, -0.00000305, -0.00000318, -0.00000381, -0.00000451, -0.00000488, -0.00000467, -0.00000392, -0.00000289, -0.00000193, -0.00000131, -0.00000113, -0.00000127, -0.00000150, -0.00000164, -0.00000163, -0.00000158, -0.00000170, },
		{ -0.00000515, -0.00000481, -0.00000392, -0.00000274, -0.00000167, -0.00000104, -0.00000107, -0.00000181, -0.00000316, -0.00000498, -0.00000709, -0.00000929, -0.00001131, -0.00001276, -0.00001315, -0.00001201, -0.00000915, -0.00000480, 0.00000017, 0.00000443, 0.00000651, 0.00000525, 0.00000030, -0.00000759, -0.00001677, -0.00002511, -0.00003078, -0.00003291, -0.00003185, -0.00002901, -0.00002607, -0.00002404, -0.00002257, -0.00001993, -0.00001392, -0.00000344, 0.00000968, 0.00001965, 0.00001711, -0.00000859, -0.00006553, -0.00015463, -0.00026612, -0.00037836, -0.00046040, -0.00047827, -0.00040385, -0.00022434, 0.00005078, 0.00038749, 0.00073132, 0.00101755, 0.00118545, 0.00119307, 0.00102873, 0.00071589, 0.00030952, -0.00011540, -0.00048053, -0.00072204, -0.00080474, -0.00072921, -0.00052994, -0.00026526, -0.00000194, 0.00020134, 0.00030728, 0.00030679, 0.00021803, 0.00007851, -0.00006692, -0.00017866, -0.00023202, -0.00022169, -0.00016028, -0.00007227, 0.00001460, 0.00007743, 0.00010381, 0.00009351, 0.00005647, 0.00000816, -0.00003581, -0.00006417, -0.00007243, -0.00006273, -0.00004183, -0.00001806, 0.00000151, 0.00001290, 0.00001564, 0.00001197, 0.00000536, -0.00000101, -0.00000521, -0.00000688, -0.00000676, -0.00000605, -0.00000565, -0.00000586, -0.00000636, -0.00000654, -0.00000595, -0.00000455, -0.00000276, -0.00000125, -0.00000062, -0.00000111, -0.00000250, -0.00000422, -0.00000558, -0.00000608, -0.00000559, -0.00000436, -0.00000288, -0.00000166, -0.00000101, -0.00000095, -0.00000126, -0.00000163, -0.00000183, -0.00000182, -0.00000175, -0.00000185, -0.00000230, -0.00000311, -0.00000406, -0.00000485, },
		{ -0.00000281, -0.00000285, -0.00000359, -0.00000491, -0.00000650, -0.00000801, -0.00000909, -0.00000951, -0.00000911, -0.00000783, -0.00000571, -0.00000298, -0.00000011, 0.00000217, 0.00000293, 0.00000132, -0.00000311, -0.00001011, -0.00001857, -0.00002666, -0.00003232, -0.00003384, -0.00003059, -0.00002330, -0.00001401, -0.00000544, -0.00000015, 0.00000032, -0.00000415, -0.00001254, -0.00002354, -0.00003676, -0.00005359, -0.00007714, -0.00011104, -0.00015711, -0.00021266, -0.00026842, -0.00030833, -0.00031181, -0.00025864, -0.00013538, 0.00005830, 0.00030551, 0.00057213, 0.00081167, 0.00097451, 0.00101961, 0.00092562, 0.00069836, 0.00037208, 0.00000359, -0.00033966, -0.00059482, -0.00071828, -0.00069583, -0.00054563, -0.00031298, -0.00005846, 0.00015745, 0.00028917, 0.00031614, 0.00024603, 0.00011019, -0.00004659, -0.00017926, -0.00025443, -0.00025821, -0.00019776, -0.00009703, 0.00001183, 0.00009835, 0.00014208, 0.00013706, 0.00009163, 0.00002420, -0.00004336, -0.00009241, -0.00011236, -0.00010245, -0.00007035, -0.00002859, 0.00000994, 0.00003575, 0.00004484, 0.00003879, 0.00002305, 0.00000452, -0.00001091, -0.00001986, -0.00002186, -0.00001867, -0.00001309, -0.00000765, -0.00000388, -0.00000203, -0.00000149, -0.00000137, -0.00000107, -0.00000051, -0.00000008, -0.00000031, -0.00000151, -0.00000352, -0.00000578, -0.00000752, -0.00000814, -0.00000741, -0.00000562, -0.00000338, -0.00000141, -0.00000021, 0.00000006, -0.00000036, -0.00000103, -0.00000154, -0.00000166, -0.00000147, -0.00000125, -0.00000133, -0.00000191, -0.00000292, -0.00000408, -0.00000500, -0.00000536, -0.00000507, -0.00000429, -0.00000340, },
		{ -0.00000547, -0.00000602, -0.00000599, -0.00000536, -0.00000429, -0.00000309, -0.00000212, -0.00000175, -0.00000237, -0.00000428, -0.00000763, -0.00001228, -0.00001763, -0.00002264, -0.00002599, -0.00002641, -0.00002315, -0.00001649, -0.00000789, 0.00000010, 0.00000446, 0.00000251, -0.00000720, -0.00002440, -0.00004712, -0.00007233, -0.00009700, -0.00011907, -0.00013792, -0.00015402, -0.00016764, -0.00017722, -0.00017791, -0.00016132, -0.00011689, -0.00003506, 0.00008854, 0.00024910, 0.00043030, 0.00060499, 0.00073943, 0.00080055, 0.00076457, 0.00062470, 0.00039545, 0.00011190, -0.00017661, -0.00041708, -0.00056525, -0.00059653, -0.00051234, -0.00033994, -0.00012564, 0.00007700, 0.00022079, 0.00027653, 0.00023963, 0.00012991, -0.00001522, -0.00015194, -0.00024242, -0.00026478, -0.00021795, -0.00012038, -0.00000326, 0.00009934, 0.00016037, 0.00016691, 0.00012259, 0.00004497, -0.00004080, -0.00010995, -0.00014514, -0.00014056, -0.00010225, -0.00004490, 0.00001348, 0.00005718, 0.00007694, 0.00007165, 0.00004737, 0.00001440, -0.00001651, -0.00003731, -0.00004456, -0.00003943, -0.00002640, -0.00001115, 0.00000147, 0.00000876, 0.00001040, 0.00000795, 0.00000369, -0.00000035, -0.00000307, -0.00000440, -0.00000496, -0.00000549, -0.00000641, -0.00000757, -0.00000847, -0.00000849, -0.00000734, -0.00000515, -0.00000252, -0.00000020, 0.00000120, 0.00000147, 0.00000080, -0.00000026, -0.00000115, -0.00000151, -0.00000135, -0.00000099, -0.00000090, -0.00000144, -0.00000269, -0.00000438, -0.00000601, -0.00000709, -0.00000729, -0.00000662, -0.00000540, -0.00000412, -0.00000327, -0.00000312, -0.00000363, -0.00000455, },
		{ -0.00000546, -0.00000573, -0.00000652, -0.00000795, -0.00000994, -0.00001227, -0.00001461, -0.00001649, -0.00001737, -0.00001679, -0.00001449, -0.00001068, -0.00000628, -0.00000291, -0.00000272, -0.00000785, -0.00001978, -0.00003866, -0.00006290, -0.00008926, -0.00011351, -0.00013147, -0.00014005, -0.00013803, -0.00012605, -0.00010586, -0.00007900, -0.00004551, -0.00000318, 0.00005197, 0.00012391, 0.00021398, 0.00031816, 0.00042544, 0.00051825, 0.00057533, 0.00057682, 0.00051016, 0.00037529, 0.00018720, -0.00002539, -0.00022532, -0.00037533, -0.00044733, -0.00042995, -0.00033215, -0.00018157, -0.00001771, 0.00011820, 0.00019399, 0.00019487, 0.00012696, 0.00001491, -0.00010563, -0.00019847, -0.00023722, -0.00021241, -0.00013352, -0.00002535, 0.00007992, 0.00015245, 0.00017342, 0.00013986, 0.00006468, -0.00002806, -0.00011085, -0.00016098, -0.00016677, -0.00013006, -0.00006460, 0.00000903, 0.00007002, 0.00010332, 0.00010336, 0.00007462, 0.00002906, -0.00001828, -0.00005408, -0.00007032, -0.00006586, -0.00004568, -0.00001835, 0.00000711, 0.00002401, 0.00002952, 0.00002479, 0.00001370, 0.00000108, -0.00000907, -0.00001461, -0.00001551, -0.00001324, -0.00000983, -0.00000694, -0.00000533, -0.00000485, -0.00000476, -0.00000430, -0.00000309, -0.00000131, 0.00000049, 0.00000167, 0.00000187, 0.00000114, -0.00000007, -0.00000119, -0.00000178, -0.00000178, -0.00000148, -0.00000140, -0.00000199, -0.00000340, -0.00000538, -0.00000736, -0.00000866, -0.00000879, -0.00000765, -0.00000558, -0.00000321, -0.00000125, -0.00000022, -0.00000029, -0.00000127, -0.00000268, -0.00000404, -0.00000498, -0.00000541, -0.00000546, },
		{ -0.00000988, -0.00001027, -0.00001028, -0.00000987, -0.00000920, -0.00000863, -0.00000877, -0.00001049, -0.00001481, -0.00002276, -0.00003503, -0.00005158, -0.00007122, -0.00009153, -0.00010901, -0.00011974, -0.00012030, -0.00010869, -0.00008502, -0.00005155, -0.00001203, 0.00002954, 0.00007027, 0.00010940, 0.00014842, 0.00019010, 0.00023663, 0.00028746, 0.00033771, 0.00037794, 0.00039574, 0.00037903, 0.00032014, 0.00021953, 0.00008772, -0.00005555, -0.00018510, -0.00027626, -0.00031150, -0.00028563, -0.00020798, -0.00010063, 0.00000720, 0.00008696, 0.00011872, 0.00009673, 0.00003100, -0.00005580, -0.00013519, -0.00018149, -0.00017958, -0.00012932, -0.00004550, 0.00004685, 0.00012026, 0.00015317, 0.00013646, 0.00007593, -0.00000970, -0.00009480, -0.00015462, -0.00017269, -0.00014536, -0.00008231, -0.00000300, 0.00006954, 0.00011579, 0.00012510, 0.00009825, 0.00004623, -0.00001384, -0.00006426, -0.00009203, -0.00009213, -0.00006809, -0.00002993, 0.00000962, 0.00003922, 0.00005211, 0.00004743, 0.00002958, 0.00000601, -0.00001550, -0.00002930, -0.00003319, -0.00002839, -0.00001846, -0.00000767, 0.00000058, 0.00000463, 0.00000468, 0.00000228, -0.00000065, -0.00000262, -0.00000302, -0.00000213, -0.00000078, 0.00000022, 0.00000040, -0.00000017, -0.00000108, -0.00000184, -0.00000220, -0.00000224, -0.00000234, -0.00000293, -0.00000419, -0.00000594, -0.00000762, -0.00000855, -0.00000821, -0.00000648, -0.00000373, -0.00000074, 0.00000160, 0.00000259, 0.00000196, -0.00000005, -0.00000281, -0.00000557, -0.00000769, -0.00000888, -0.00000918, -0.00000895, -0.00000861, -0.00000850, -0.00000875, -0.00000929, },
		{ -0.00001493, -0.00001972, -0.00002640, -0.00003518, -0.00004599, -0.00005834, -0.00007120, -0.00008292, -0.00009126, -0.00009366, -0.00008770, -0.00007178, -0.00004574, -0.00001126, 0.00002823, 0.00006818, 0.00010409, 0.00013277, 0.00015335, 0.00016757, 0.00017909, 0.00019198, 0.00020877, 0.00022874, 0.00024724, 0.00025635, 0.00024711, 0.00021255, 0.00015078, 0.00006685, -0.00002730, -0.00011514, -0.00017972, -0.00020841, -0.00019688, -0.00015086, -0.00008502, -0.00001903, 0.00002811, 0.00004374, 0.00002538, -0.00001846, -0.00007096, -0.00011239, -0.00012654, -0.00010608, -0.00005521, 0.00001147, 0.00007348, 0.00011100, 0.00011131, 0.00007306, 0.00000696, -0.00006743, -0.00012766, -0.00015541, -0.00014209, -0.00009149, -0.00001847, 0.00005572, 0.00010997, 0.00012967, 0.00011089, 0.00006107, -0.00000369, -0.00006393, -0.00010274, -0.00011071, -0.00008811, -0.00004413, 0.00000665, 0.00004906, 0.00007189, 0.00007085, 0.00004910, 0.00001544, -0.00001895, -0.00004422, -0.00005460, -0.00004954, -0.00003313, -0.00001208, 0.00000671, 0.00001835, 0.00002109, 0.00001626, 0.00000726, -0.00000202, -0.00000854, -0.00001092, -0.00000954, -0.00000596, -0.00000209, 0.00000063, 0.00000160, 0.00000104, -0.00000030, -0.00000169, -0.00000269, -0.00000330, -0.00000381, -0.00000451, -0.00000543, -0.00000629, -0.00000660, -0.00000593, -0.00000416, -0.00000165, 0.00000087, 0.00000252, 0.00000262, 0.00000096, -0.00000212, -0.00000580, -0.00000909, -0.00001118, -0.00001168, -0.00001072, -0.00000884, -0.00000677, -0.00000516, -0.00000436, -0.00000443, -0.00000516, -0.00000629, -0.00000768, -0.00000937, -0.00001164, },
		{ -0.00006230, -0.00006655, -0.00006727, -0.00006313, -0.00005295, -0.00003604, -0.00001250, 0.00001649, 0.00004856, 0.00008034, 0.00010803, 0.00012830, 0.00013925, 0.00014122, 0.00013690, 0.00013080, 0.00012790, 0.00013189, 0.00014358, 0.00016000, 0.00017477, 0.00017964, 0.00016703, 0.00013277, 0.00007807, 0.00001002, -0.00005968, -0.00011772, -0.00015287, -0.00015935, -0.00013875, -0.00009968, -0.00005524, -0.00001903, -0.00000088, -0.00000382, -0.00002322, -0.00004857, -0.00006727, -0.00006918, -0.00005042, -0.00001507, 0.00002599, 0.00005864, 0.00007041, 0.00005509, 0.00001527, -0.00003796, -0.00008824, -0.00011913, -0.00011952, -0.00008746, -0.00003116, 0.00003341, 0.00008733, 0.00011464, 0.00010737, 0.00006795, 0.00000851, -0.00005305, -0.00009860, -0.00011528, -0.00009933, -0.00005685, -0.00000154, 0.00004978, 0.00008244, 0.00008819, 0.00006733, 0.00002796, -0.00001705, -0.00005425, -0.00007380, -0.00007199, -0.00005179, -0.00002118, 0.00000979, 0.00003229, 0.00004126, 0.00003637, 0.00002146, 0.00000271, -0.00001364, -0.00002321, -0.00002455, -0.00001904, -0.00000989, -0.00000074, 0.00000562, 0.00000797, 0.00000668, 0.00000322, -0.00000065, -0.00000360, -0.00000510, -0.00000532, -0.00000487, -0.00000435, -0.00000403, -0.00000380, -0.00000338, -0.00000253, -0.00000133, -0.00000020, 0.00000026, -0.00000044, -0.00000245, -0.00000538, -0.00000843, -0.00001066, -0.00001132, -0.00001017, -0.00000755, -0.00000430, -0.00000145, 0.00000010, -0.00000012, -0.00000211, -0.00000544, -0.00000957, -0.00001399, -0.00001851, -0.00002322, -0.00002840, -0.00003434, -0.00004112, -0.00004851, -0.00005588, },
		{ 0.00001005, 0.00003185, 0.00005475, 0.00007666, 0.00009519, 0.00010807, 0.00011373, 0.00011184, 0.00010377, 0.00009263, 0.00008268, 0.00007834, 0.00008273, 0.00009638, 0.00011650, 0.00013714, 0.00015048, 0.00014897, 0.00012765, 0.00008612, 0.00002927, -0.00003349, -0.00009048, -0.00013087, -0.00014766, -0.00013975, -0.00011221, -0.00007467, -0.00003833, -0.00001255, -0.00000206, -0.00000579, -0.00001774, -0.00002951, -0.00003361, -0.00002642, -0.00000959, 0.00001048, 0.00002512, 0.00002681, 0.00001239, -0.00001529, -0.00004784, -0.00007391, -0.00008311, -0.00006986, -0.00003578, 0.00001029, 0.00005480, 0.00008372, 0.00008721, 0.00006302, 0.00001761, -0.00003568, -0.00008071, -0.00010359, -0.00009713, -0.00006315, -0.00001192, 0.00004107, 0.00007997, 0.00009350, 0.00007831, 0.00003978, -0.00000990, -0.00005573, -0.00008460, -0.00008915, -0.00006967, -0.00003350, 0.00000767, 0.00004169, 0.00005961, 0.00005812, 0.00003994, 0.00001238, -0.00001539, -0.00003536, -0.00004294, -0.00003788, -0.00002376, -0.00000626, 0.00000893, 0.00001785, 0.00001923, 0.00001438, 0.00000624, -0.00000189, -0.00000753, -0.00000966, -0.00000870, -0.00000602, -0.00000319, -0.00000134, -0.00000083, -0.00000137, -0.00000234, -0.00000318, -0.00000368, -0.00000400, -0.00000445, -0.00000528, -0.00000639, -0.00000738, -0.00000767, -0.00000680, -0.00000475, -0.00000200, 0.00000053, 0.00000182, 0.00000104, -0.00000208, -0.00000722, -0.00001356, -0.00002009, -0.00002595, -0.00003071, -0.00003436, -0.00003728, -0.00003987, -0.00004237, -0.00004460, -0.00004595, -0.00004552, -0.00004229, -0.00003541, -0.00002434, -0.00000903, },
		{ 0.00008679, 0.00008424, 0.00007737, 0.00006793, 0.00005858, 0.00005250, 0.00005267, 0.00006092, 0.00007697, 0.00009798, 0.00011869, 0.00013240, 0.00013262, 0.00011503, 0.00007909, 0.00002880, -0.00002785, -0.00008059, -0.00011948, -0.00013756, -0.00013280, -0.00010869, -0.00007314, -0.00003617, -0.00000697, 0.00000875, 0.00001009, 0.00000077, -0.00001263, -0.00002327, -0.00002660, -0.00002186, -0.00001214, -0.00000286, 0.00000069, -0.00000433, -0.00001683, -0.00003197, -0.00004288, -0.00004336, -0.00003056, -0.00000651, 0.00002204, 0.00004564, 0.00005542, 0.00004635, 0.00001943, -0.00001811, -0.00005484, -0.00007868, -0.00008099, -0.00005965, -0.00002007, 0.00002622, 0.00006510, 0.00008433, 0.00007752, 0.00004627, -0.00000025, -0.00004817, -0.00008321, -0.00009514, -0.00008087, -0.00004526, 0.00000064, 0.00004315, 0.00007026, 0.00007514, 0.00005795, 0.00002534, -0.00001206, -0.00004308, -0.00005952, -0.00005824, -0.00004168, -0.00001646, 0.00000907, 0.00002762, 0.00003501, 0.00003103, 0.00001890, 0.00000376, -0.00000932, -0.00001687, -0.00001790, -0.00001372, -0.00000706, -0.00000088, 0.00000271, 0.00000302, 0.00000066, -0.00000289, -0.00000611, -0.00000794, -0.00000809, -0.00000695, -0.00000523, -0.00000356, -0.00000224, -0.00000127, -0.00000046, 0.00000025, 0.00000063, 0.00000016, -0.00000177, -0.00000555, -0.00001111, -0.00001782, -0.00002460, -0.00003030, -0.00003402, -0.00003542, -0.00003479, -0.00003285, -0.00003044, -0.00002816, -0.00002607, -0.00002367, -0.00002009, -0.00001436, -0.00000581, 0.00000572, 0.00001978, 0.00003536, 0.00005109, 0.00006544, 0.00007692, 0.00008431, },
		{ 0.00003825, 0.00004257, 0.00005281, 0.00006814, 0.00008602, 0.00010232, 0.00011199, 0.00011026, 0.00009397, 0.00006290, 0.00002032, -0.00002729, -0.00007153, -0.00010412, -0.00011910, -0.00011452, -0.00009305, -0.00006132, -0.00002803, -0.00000146, 0.00001283, 0.00001341, 0.00000310, -0.00001234, -0.00002632, -0.00003378, -0.00003272, -0.00002462, -0.00001358, -0.00000439, -0.00000060, -0.00000304, -0.00000956, -0.00001601, -0.00001810, -0.00001336, -0.00000245, 0.00001083, 0.00002086, 0.00002242, 0.00001294, -0.00000606, -0.00002900, -0.00004792, -0.00005519, -0.00004642, -0.00002240, 0.00001061, 0.00004263, 0.00006304, 0.00006417, 0.00004413, 0.00000778, -0.00003449, -0.00006991, -0.00008732, -0.00008083, -0.00005180, -0.00000850, 0.00003635, 0.00006960, 0.00008172, 0.00006966, 0.00003767, -0.00000422, -0.00004345, -0.00006889, -0.00007409, -0.00005899, -0.00002948, 0.00000478, 0.00003363, 0.00004949, 0.00004943, 0.00003555, 0.00001379, -0.00000838, -0.00002446, -0.00003078, -0.00002731, -0.00001706, -0.00000469, 0.00000530, 0.00001003, 0.00000890, 0.00000344, -0.00000359, -0.00000940, -0.00001207, -0.00001115, -0.00000746, -0.00000262, 0.00000168, 0.00000425, 0.00000468, 0.00000319, 0.00000038, -0.00000321, -0.00000726, -0.00001168, -0.00001647, -0.00002141, -0.00002599, -0.00002945, -0.00003099, -0.00003013, -0.00002694, -0.00002207, -0.00001661, -0.00001166, -0.00000792, -0.00000547, -0.00000363, -0.00000128, 0.00000277, 0.00000933, 0.00001849, 0.00002956, 0.00004124, 0.00005196, 0.00006026, 0.00006512, 0.00006614, 0.00006352, 0.00005807, 0.00005107, 0.00004418, 0.00003928, },
		{ 0.00008451, 0.00008994, 0.00008658, 0.00007243, 0.00004743, 0.00001399, -0.00002314, -0.00005772, -0.00008343, -0.00009557, -0.00009242, -0.00007584, -0.00005091, -0.00002453, -0.00000349, 0.00000751, 0.00000709, -0.00000251, -0.00001640, -0.00002881, -0.00003500, -0.00003287, -0.00002344, -0.00001030, 0.00000186, 0.00000909, 0.00000955, 0.00000413, -0.00000412, -0.00001133, -0.00001447, -0.00001258, -0.00000718, -0.00000153, 0.00000085, -0.00000218, -0.00001020, -0.00002013, -0.00002733, -0.00002736, -0.00001808, -0.00000083, 0.00001953, 0.00003599, 0.00004184, 0.00003320, 0.00001084, -0.00001958, -0.00004898, -0.00006765, -0.00006852, -0.00004977, -0.00001574, 0.00002408, 0.00005789, 0.00007527, 0.00007055, 0.00004472, 0.00000522, -0.00003634, -0.00006775, -0.00008001, -0.00007001, -0.00004137, -0.00000308, 0.00003339, 0.00005778, 0.00006399, 0.00005177, 0.00002630, -0.00000377, -0.00002929, -0.00004345, -0.00004365, -0.00003185, -0.00001343, 0.00000489, 0.00001740, 0.00002106, 0.00001616, 0.00000575, -0.00000573, -0.00001417, -0.00001708, -0.00001423, -0.00000741, 0.00000048, 0.00000658, 0.00000897, 0.00000714, 0.00000192, -0.00000507, -0.00001213, -0.00001796, -0.00002196, -0.00002411, -0.00002469, -0.00002397, -0.00002204, -0.00001888, -0.00001454, -0.00000933, -0.00000393, 0.00000076, 0.00000393, 0.00000518, 0.00000484, 0.00000391, 0.00000381, 0.00000593, 0.00001107, 0.00001916, 0.00002917, 0.00003938, 0.00004789, 0.00005316, 0.00005443, 0.00005184, 0.00004641, 0.00003964, 0.00003325, 0.00002877, 0.00002742, 0.00002996, 0.00003662, 0.00004701, 0.00005996, 0.00007341, },
		{ 0.00001149, -0.00001631, -0.00004268, -0.00006312, -0.00007404, -0.00007377, -0.00006322, -0.00004578, -0.00002639, -0.00001022, -0.00000108, -0.00000033, -0.00000656, -0.00001615, -0.00002459, -0.00002807, -0.00002480, -0.00001560, -0.00000353, 0.00000731, 0.00001326, 0.00001245, 0.00000544, -0.00000503, -0.00001521, -0.00002167, -0.00002264, -0.00001853, -0.00001168, -0.00000520, -0.00000170, -0.00000210, -0.00000531, -0.00000872, -0.00000945, -0.00000575, 0.00000196, 0.00001100, 0.00001726, 0.00001690, 0.00000805, -0.00000801, -0.00002680, -0.00004191, -0.00004717, -0.00003894, -0.00001784, 0.00001097, 0.00003912, 0.00005757, 0.00005960, 0.00004326, 0.00001236, -0.00002450, -0.00005641, -0.00007356, -0.00007043, -0.00004759, -0.00001161, 0.00002695, 0.00005684, 0.00006960, 0.00006209, 0.00003737, 0.00000346, -0.00002929, -0.00005150, -0.00005761, -0.00004731, -0.00002528, 0.00000068, 0.00002234, 0.00003367, 0.00003252, 0.00002090, 0.00000390, -0.00001230, -0.00002259, -0.00002445, -0.00001850, -0.00000792, 0.00000293, 0.00001011, 0.00001134, 0.00000655, -0.00000236, -0.00001244, -0.00002078, -0.00002540, -0.00002569, -0.00002231, -0.00001667, -0.00001031, -0.00000444, 0.00000034, 0.00000387, 0.00000625, 0.00000751, 0.00000761, 0.00000651, 0.00000439, 0.00000187, 0.00000003, 0.00000013, 0.00000325, 0.00000980, 0.00001920, 0.00002995, 0.00003990, 0.00004693, 0.00004952, 0.00004725, 0.00004093, 0.00003235, 0.00002379, 0.00001737, 0.00001461, 0.00001615, 0.00002178, 0.00003061, 0.00004137, 0.00005255, 0.00006250, 0.00006946, 0.00007163, 0.00006737, 0.00005564, 0.00003650, },
		{ -0.00005948, -0.00005407, -0.00004275, -0.00002874, -0.00001576, -0.00000693, -0.00000381, -0.00000597, -0.00001117, -0.00001628, -0.00001840, -0.00001602, -0.00000950, -0.00000101, 0.00000632, 0.00000964, 0.00000742, 0.00000016, -0.00000983, -0.00001920, -0.00002483, -0.00002491, -0.00001965, -0.00001106, -0.00000223, 0.00000396, 0.00000587, 0.00000367, -0.00000084, -0.00000513, -0.00000706, -0.00000585, -0.00000249, 0.00000078, 0.00000143, -0.00000202, -0.00000913, -0.00001740, -0.00002305, -0.00002252, -0.00001405, 0.00000124, 0.00001927, 0.00003410, 0.00003990, 0.00003316, 0.00001425, -0.00001223, -0.00003861, -0.00005645, -0.00005933, -0.00004518, -0.00001721, 0.00001681, 0.00004687, 0.00006381, 0.00006230, 0.00004258, 0.00001046, -0.00002450, -0.00005200, -0.00006417, -0.00005804, -0.00003626, -0.00000608, 0.00002308, 0.00004269, 0.00004770, 0.00003791, 0.00001765, -0.00000591, -0.00002528, -0.00003506, -0.00003349, -0.00002268, -0.00000748, 0.00000633, 0.00001411, 0.00001374, 0.00000610, -0.00000556, -0.00001695, -0.00002426, -0.00002538, -0.00002035, -0.00001107, -0.00000046, 0.00000864, 0.00001430, 0.00001587, 0.00001392, 0.00000971, 0.00000468, -0.00000004, -0.00000374, -0.00000602, -0.00000657, -0.00000504, -0.00000101, 0.00000573, 0.00001490, 0.00002545, 0.00003567, 0.00004344, 0.00004688, 0.00004495, 0.00003788, 0.00002725, 0.00001561, 0.00000584, 0.00000030, 0.00000031, 0.00000582, 0.00001556, 0.00002752, 0.00003953, 0.00004970, 0.00005671, 0.00005978, 0.00005851, 0.00005266, 0.00004218, 0.00002731, 0.00000892, -0.00001132, -0.00003096, -0.00004713, -0.00005717, },
		{ -0.00000948, -0.00000460, -0.00000375, -0.00000595, -0.00000923, -0.00001145, -0.00001113, -0.00000804, -0.00000332, 0.00000099, 0.00000283, 0.00000098, -0.00000438, -0.00001161, -0.00001821, -0.00002173, -0.00002073, -0.00001535, -0.00000733, 0.00000065, 0.00000596, 0.00000699, 0.00000374, -0.00000222, -0.00000839, -0.00001239, -0.00001284, -0.00000993, -0.00000521, -0.00000094, 0.00000104, 0.00000006, -0.00000294, -0.00000587, -0.00000638, -0.00000306, 0.00000375, 0.00001178, 0.00001754, 0.00001767, 0.00001041, -0.00000335, -0.00001994, -0.00003392, -0.00003984, -0.00003430, -0.00001743, 0.00000672, 0.00003116, 0.00004808, 0.00005144, 0.00003914, 0.00001394, -0.00001712, -0.00004484, -0.00006071, -0.00005968, -0.00004183, -0.00001240, 0.00001983, 0.00004533, 0.00005678, 0.00005132, 0.00003131, 0.00000338, -0.00002381, -0.00004243, -0.00004786, -0.00003990, -0.00002256, -0.00000240, 0.00001381, 0.00002132, 0.00001874, 0.00000829, -0.00000536, -0.00001686, -0.00002203, -0.00001921, -0.00000957, 0.00000348, 0.00001568, 0.00002335, 0.00002454, 0.00001942, 0.00000998, -0.00000083, -0.00001013, -0.00001588, -0.00001729, -0.00001462, -0.00000880, -0.00000092, 0.00000810, 0.00001754, 0.00002672, 0.00003476, 0.00004047, 0.00004257, 0.00004004, 0.00003267, 0.00002139, 0.00000828, -0.00000379, -0.00001190, -0.00001392, -0.00000914, 0.00000153, 0.00001580, 0.00003065, 0.00004321, 0.00005143, 0.00005447, 0.00005260, 0.00004682, 0.00003830, 0.00002798, 0.00001637, 0.00000375, -0.00000949, -0.00002253, -0.00003404, -0.00004236, -0.00004603, -0.00004439, -0.00003790, -0.00002827, -0.00001796, },
		{ -0.00000785, -0.00000914, -0.00000862, -0.00000652, -0.00000394, -0.00000232, -0.00000276, -0.00000551, -0.00000978, -0.00001396, -0.00001632, -0.00001569, -0.00001201, -0.00000642, -0.00000088, 0.00000260, 0.00000275, -0.00000043, -0.00000563, -0.00001069, -0.00001347, -0.00001268, -0.00000839, -0.00000202, 0.00000422, 0.00000819, 0.00000870, 0.00000597, 0.00000147, -0.00000267, -0.00000466, -0.00000382, -0.00000094, 0.00000208, 0.00000302, 0.00000047, -0.00000538, -0.00001257, -0.00001796, -0.00001843, -0.00001227, -0.00000018, 0.00001461, 0.00002716, 0.00003248, 0.00002738, 0.00001191, -0.00001024, -0.00003265, -0.00004807, -0.00005086, -0.00003898, -0.00001495, 0.00001469, 0.00004132, 0.00005687, 0.00005648, 0.00004001, 0.00001221, -0.00001881, -0.00004415, -0.00005686, -0.00005408, -0.00003776, -0.00001381, 0.00001000, 0.00002660, 0.00003187, 0.00002575, 0.00001194, -0.00000362, -0.00001497, -0.00001811, -0.00001230, -0.00000006, 0.00001389, 0.00002442, 0.00002774, 0.00002263, 0.00001068, -0.00000441, -0.00001819, -0.00002684, -0.00002833, -0.00002272, -0.00001185, 0.00000150, 0.00001457, 0.00002533, 0.00003278, 0.00003673, 0.00003745, 0.00003521, 0.00003011, 0.00002224, 0.00001199, 0.00000043, -0.00001062, -0.00001883, -0.00002195, -0.00001858, -0.00000873, 0.00000597, 0.00002262, 0.00003774, 0.00004831, 0.00005252, 0.00005023, 0.00004282, 0.00003257, 0.00002182, 0.00001227, 0.00000452, -0.00000182, -0.00000765, -0.00001375, -0.00002026, -0.00002649, -0.00003120, -0.00003307, -0.00003139, -0.00002639, -0.00001933, -0.00001207, -0.00000644, -0.00000358, -0.00000359, -0.00000552, },
		{ -0.00000521, -0.00000463, -0.00000550, -0.00000775, -0.00001055, -0.00001273, -0.00001325, -0.00001169, -0.00000851, -0.00000487, -0.00000217, -0.00000140, -0.00000275, -0.00000541, -0.00000792, -0.00000870, -0.00000682, -0.00000238, 0.00000339, 0.00000858, 0.00001130, 0.00001043, 0.00000613, -0.00000020, -0.00000643, -0.00001051, -0.00001121, -0.00000867, -0.00000423, 0.00000010, 0.00000254, 0.00000230, -0.00000009, -0.00000296, -0.00000432, -0.00000281, 0.00000155, 0.00000711, 0.00001114, 0.00001090, 0.00000489, -0.00000622, -0.00001943, -0.00003022, -0.00003397, -0.00002778, -0.00001179, 0.00001047, 0.00003284, 0.00004837, 0.00005157, 0.00004035, 0.00001693, -0.00001259, -0.00004003, -0.00005761, -0.00006038, -0.00004784, -0.00002398, 0.00000394, 0.00002783, 0.00004129, 0.00004157, 0.00003027, 0.00001253, -0.00000488, -0.00001591, -0.00001729, -0.00000947, 0.00000378, 0.00001682, 0.00002429, 0.00002292, 0.00001272, -0.00000322, -0.00001980, -0.00003169, -0.00003513, -0.00002899, -0.00001498, 0.00000310, 0.00002070, 0.00003392, 0.00004057, 0.00004041, 0.00003486, 0.00002610, 0.00001623, 0.00000670, -0.00000188, -0.00000936, -0.00001558, -0.00001997, -0.00002149, -0.00001896, -0.00001163, 0.00000017, 0.00001485, 0.00002969, 0.00004157, 0.00004786, 0.00004728, 0.00004032, 0.00002910, 0.00001668, 0.00000603, -0.00000088, -0.00000366, -0.00000345, -0.00000231, -0.00000241, -0.00000508, -0.00001037, -0.00001704, -0.00002311, -0.00002666, -0.00002656, -0.00002290, -0.00001691, -0.00001049, -0.00000545, -0.00000290, -0.00000290, -0.00000460, -0.00000666, -0.00000791, -0.00000780, -0.00000661, },
		{ -0.00001114, -0.00001219, -0.00001186, -0.00001000, -0.00000714, -0.00000419, -0.00000205, -0.00000121, -0.00000147, -0.00000209, -0.00000208, -0.00000074, 0.00000196, 0.00000527, 0.00000789, 0.00000850, 0.00000635, 0.00000165, -0.00000433, -0.00000974, -0.00001271, -0.00001215, -0.00000814, -0.00000198, 0.00000426, 0.00000853, 0.00000953, 0.00000718, 0.00000263, -0.00000231, -0.00000588, -0.00000715, -0.00000638, -0.00000483, -0.00000415, -0.00000545, -0.00000867, -0.00001236, -0.00001410, -0.00001155, -0.00000353, 0.00000909, 0.00002327, 0.00003454, 0.00003840, 0.00003201, 0.00001543, -0.00000794, -0.00003211, -0.00005023, -0.00005670, -0.00004907, -0.00002896, -0.00000175, 0.00002508, 0.00004421, 0.00005080, 0.00004396, 0.00002693, 0.00000590, -0.00001215, -0.00002177, -0.00002080, -0.00001096, 0.00000297, 0.00001493, 0.00001985, 0.00001540, 0.00000285, -0.00001349, -0.00002778, -0.00003472, -0.00003136, -0.00001809, 0.00000157, 0.00002220, 0.00003825, 0.00004577, 0.00004350, 0.00003299, 0.00001778, 0.00000209, -0.00001059, -0.00001839, -0.00002123, -0.00002039, -0.00001757, -0.00001404, -0.00001016, -0.00000543, 0.00000094, 0.00000934, 0.00001917, 0.00002879, 0.00003592, 0.00003847, 0.00003541, 0.00002724, 0.00001607, 0.00000491, -0.00000327, -0.00000659, -0.00000492, 0.00000011, 0.00000571, 0.00000893, 0.00000773, 0.00000172, -0.00000770, -0.00001791, -0.00002595, -0.00002960, -0.00002806, -0.00002218, -0.00001404, -0.00000617, -0.00000068, 0.00000142, 0.00000035, -0.00000263, -0.00000594, -0.00000825, -0.00000903, -0.00000854, -0.00000763, -0.00000724, -0.00000786, -0.00000940, },
		{ -0.00000329, -0.00000047, 0.00000156, 0.00000246, 0.00000244, 0.00000205, 0.00000192, 0.00000237, 0.00000322, 0.00000385, 0.00000350, 0.00000166, -0.00000155, -0.00000533, -0.00000837, -0.00000940, -0.00000766, -0.00000338, 0.00000227, 0.00000745, 0.00001031, 0.00000967, 0.00000548, -0.00000110, -0.00000814, -0.00001357, -0.00001594, -0.00001489, -0.00001119, -0.00000635, -0.00000190, 0.00000123, 0.00000308, 0.00000454, 0.00000681, 0.00001058, 0.00001548, 0.00001991, 0.00002147, 0.00001792, 0.00000822, -0.00000661, -0.00002342, -0.00003760, -0.00004449, -0.00004094, -0.00002662, -0.00000444, 0.00002006, 0.00004034, 0.00005083, 0.00004872, 0.00003498, 0.00001404, -0.00000763, -0.00002361, -0.00002963, -0.00002492, -0.00001238, 0.00000257, 0.00001394, 0.00001734, 0.00001152, -0.00000122, -0.00001585, -0.00002648, -0.00002848, -0.00002013, -0.00000334, 0.00001705, 0.00003481, 0.00004449, 0.00004311, 0.00003116, 0.00001223, -0.00000822, -0.00002469, -0.00003342, -0.00003332, -0.00002601, -0.00001495, -0.00000398, 0.00000404, 0.00000808, 0.00000896, 0.00000868, 0.00000934, 0.00001206, 0.00001657, 0.00002126, 0.00002402, 0.00002318, 0.00001836, 0.00001076, 0.00000275, -0.00000297, -0.00000450, -0.00000148, 0.00000464, 0.00001108, 0.00001476, 0.00001343, 0.00000651, -0.00000458, -0.00001691, -0.00002702, -0.00003208, -0.00003083, -0.00002397, -0.00001384, -0.00000359, 0.00000392, 0.00000694, 0.00000534, 0.00000043, -0.00000566, -0.00001084, -0.00001374, -0.00001407, -0.00001254, -0.00001036, -0.00000869, -0.00000812, -0.00000854, -0.00000925, -0.00000941, -0.00000841, -0.00000621, },
		{ 0.00000312, 0.00000207, 0.00000118, 0.00000075, 0.00000067, 0.00000055, -0.00000002, -0.00000122, -0.00000279, -0.00000409, -0.00000437, -0.00000317, -0.00000060, 0.00000257, 0.00000506, 0.00000562, 0.00000348, -0.00000123, -0.00000739, -0.00001321, -0.00001682, -0.00001689, -0.00001314, -0.00000643, 0.00000151, 0.00000873, 0.00001368, 0.00001572, 0.00001520, 0.00001313, 0.00001061, 0.00000828, 0.00000600, 0.00000298, -0.00000171, -0.00000842, -0.00001645, -0.00002385, -0.00002798, -0.00002635, -0.00001773, -0.00000292, 0.00001508, 0.00003175, 0.00004234, 0.00004341, 0.00003413, 0.00001671, -0.00000408, -0.00002235, -0.00003299, -0.00003329, -0.00002390, -0.00000859, 0.00000706, 0.00001759, 0.00001956, 0.00001276, 0.00000027, -0.00001268, -0.00002064, -0.00001990, -0.00000995, 0.00000625, 0.00002327, 0.00003505, 0.00003703, 0.00002780, 0.00000962, -0.00001230, -0.00003151, -0.00004247, -0.00004230, -0.00003169, -0.00001447, 0.00000379, 0.00001773, 0.00002391, 0.00002184, 0.00001384, 0.00000401, -0.00000345, -0.00000572, -0.00000228, 0.00000509, 0.00001322, 0.00001884, 0.00001994, 0.00001641, 0.00001002, 0.00000358, -0.00000027, -0.00000018, 0.00000341, 0.00000839, 0.00001186, 0.00001135, 0.00000580, -0.00000389, -0.00001517, -0.00002467, -0.00002943, -0.00002796, -0.00002080, -0.00001026, 0.00000030, 0.00000768, 0.00000987, 0.00000662, -0.00000057, -0.00000910, -0.00001625, -0.00002006, -0.00001991, -0.00001653, -0.00001157, -0.00000688, -0.00000382, -0.00000281, -0.00000339, -0.00000450, -0.00000508, -0.00000445, -0.00000262, -0.00000015, 0.00000213, 0.00000353, 0.00000379, },
		{ 0.00000021, 0.00000033, 0.00000016, -0.00000047, -0.00000142, -0.00000230, -0.00000269, -0.00000242, -0.00000172, -0.00000118, -0.00000148, -0.00000307, -0.00000585, -0.00000907, -0.00001151, -0.00001186, -0.00000932, -0.00000388, 0.00000346, 0.00001102, 0.00001689, 0.00001960, 0.00001856, 0.00001423, 0.00000792, 0.00000127, -0.00000433, -0.00000818, -0.00001035, -0.00001145, -0.00001211, -0.00001255, -0.00001228, -0.00001033, -0.00000571, 0.00000191, 0.00001171, 0.00002161, 0.00002876, 0.00003041, 0.00002502, 0.00001299, -0.00000310, -0.00001908, -0.00003043, -0.00003374, -0.00002792, -0.00001471, 0.00000177, 0.00001634, 0.00002448, 0.00002391, 0.00001534, 0.00000232, -0.00001007, -0.00001703, -0.00001582, -0.00000682, 0.00000652, 0.00001889, 0.00002497, 0.00002150, 0.00000849, -0.00001055, -0.00002973, -0.00004272, -0.00004497, -0.00003528, -0.00001630, 0.00000627, 0.00002561, 0.00003602, 0.00003477, 0.00002297, 0.00000509, -0.00001261, -0.00002422, -0.00002615, -0.00001817, -0.00000334, 0.00001321, 0.00002617, 0.00003181, 0.00002915, 0.00002005, 0.00000833, -0.00000180, -0.00000723, -0.00000706, -0.00000267, 0.00000296, 0.00000657, 0.00000595, 0.00000075, -0.00000739, -0.00001555, -0.00002070, -0.00002090, -0.00001605, -0.00000792, 0.00000053, 0.00000620, 0.00000704, 0.00000272, -0.00000526, -0.00001420, -0.00002116, -0.00002397, -0.00002196, -0.00001607, -0.00000838, -0.00000131, 0.00000322, 0.00000442, 0.00000273, -0.00000047, -0.00000353, -0.00000512, -0.00000471, -0.00000263, 0.00000018, 0.00000265, 0.00000399, 0.00000401, 0.00000305, 0.00000173, 0.00000069, 0.00000021, },
		{ -0.00000401, -0.00000533, -0.00000619, -0.00000636, -0.00000592, -0.00000521, -0.00000460, -0.00000428, -0.00000406, -0.00000343, -0.00000179, 0.00000128, 0.00000564, 0.00001056, 0.00001480, 0.00001700, 0.00001619, 0.00001218, 0.00000569, -0.00000184, -0.00000862, -0.00001315, -0.00001464, -0.00001320, -0.00000971, -0.00000536, -0.00000121, 0.00000219, 0.00000493, 0.00000742, 0.00001004, 0.00001267, 0.00001455, 0.00001447, 0.00001126, 0.00000448, -0.00000513, -0.00001551, -0.00002379, -0.00002714, -0.00002381, -0.00001397, 0.00000016, 0.00001476, 0.00002557, 0.00002934, 0.00002495, 0.00001387, -0.00000018, -0.00001254, -0.00001919, -0.00001826, -0.00001067, 0.00000005, 0.00000907, 0.00001206, 0.00000685, -0.00000558, -0.00002125, -0.00003451, -0.00003996, -0.00003452, -0.00001866, 0.00000347, 0.00002532, 0.00004000, 0.00004262, 0.00003204, 0.00001137, -0.00001300, -0.00003336, -0.00004325, -0.00003951, -0.00002328, 0.00000039, 0.00002429, 0.00004136, 0.00004696, 0.00004020, 0.00002408, 0.00000422, -0.00001311, -0.00002304, -0.00002362, -0.00001615, -0.00000448, 0.00000660, 0.00001306, 0.00001300, 0.00000710, -0.00000193, -0.00001050, -0.00001559, -0.00001585, -0.00001197, -0.00000625, -0.00000157, -0.00000023, -0.00000303, -0.00000899, -0.00001582, -0.00002078, -0.00002180, -0.00001821, -0.00001098, -0.00000234, 0.00000507, 0.00000916, 0.00000906, 0.00000536, -0.00000019, -0.00000542, -0.00000848, -0.00000849, -0.00000571, -0.00000134, 0.00000306, 0.00000613, 0.00000720, 0.00000637, 0.00000435, 0.00000209, 0.00000034, -0.00000062, -0.00000096, -0.00000116, -0.00000166, -0.00000266, },
		{ -0.00000320, -0.00000058, 0.00000198, 0.00000414, 0.00000582, 0.00000719, 0.00000848, 0.00000977, 0.00001083, 0.00001113, 0.00001005, 0.00000718, 0.00000257, -0.00000309, -0.00000858, -0.00001255, -0.00001389, -0.00001217, -0.00000780, -0.00000194, 0.00000393, 0.00000842, 0.00001070, 0.00001068, 0.00000890, 0.00000621, 0.00000333, 0.00000059, -0.00000216, -0.00000527, -0.00000893, -0.00001279, -0.00001586, -0.00001671, -0.00001405, -0.00000732, 0.00000274, 0.00001408, 0.00002373, 0.00002874, 0.00002717, 0.00001891, 0.00000590, -0.00000833, -0.00001979, -0.00002532, -0.00002380, -0.00001653, -0.00000686, 0.00000098, 0.00000351, -0.00000065, -0.00001012, -0.00002108, -0.00002863, -0.00002858, -0.00001919, -0.00000206, 0.00001810, 0.00003486, 0.00004223, 0.00003681, 0.00001917, -0.00000601, -0.00003130, -0.00004875, -0.00005252, -0.00004082, -0.00001663, 0.00001314, 0.00003983, 0.00005569, 0.00005631, 0.00004198, 0.00001746, -0.00000960, -0.00003119, -0.00004142, -0.00003821, -0.00002378, -0.00000359, 0.00001559, 0.00002791, 0.00003025, 0.00002293, 0.00000924, -0.00000594, -0.00001795, -0.00002387, -0.00002326, -0.00001799, -0.00001122, -0.00000606, -0.00000440, -0.00000627, -0.00001011, -0.00001349, -0.00001425, -0.00001141, -0.00000550, 0.00000160, 0.00000746, 0.00001007, 0.00000861, 0.00000372, -0.00000274, -0.00000839, -0.00001121, -0.00001026, -0.00000591, 0.00000031, 0.00000639, 0.00001050, 0.00001161, 0.00000977, 0.00000591, 0.00000141, -0.00000245, -0.00000489, -0.00000587, -0.00000585, -0.00000556, -0.00000556, -0.00000602, -0.00000670, -0.00000710, -0.00000674, -0.00000539, },
		{ 0.00001098, 0.00000974, 0.00000790, 0.00000579, 0.00000356, 0.00000115, -0.00000152, -0.00000443, -0.00000722, -0.00000931, -0.00000999, -0.00000872, -0.00000546, -0.00000074, 0.00000436, 0.00000852, 0.00001060, 0.00000999, 0.00000685, 0.00000199, -0.00000334, -0.00000796, -0.00001106, -0.00001237, -0.00001217, -0.00001091, -0.00000896, -0.00000636, -0.00000283, 0.00000196, 0.00000801, 0.00001466, 0.00002056, 0.00002390, 0.00002302, 0.00001711, 0.00000670, -0.00000621, -0.00001863, -0.00002742, -0.00003037, -0.00002701, -0.00001888, -0.00000908, -0.00000120, 0.00000196, -0.00000049, -0.00000711, -0.00001453, -0.00001872, -0.00001651, -0.00000702, 0.00000773, 0.00002320, 0.00003380, 0.00003484, 0.00002435, 0.00000411, -0.00002052, -0.00004214, -0.00005358, -0.00005037, -0.00003233, -0.00000394, 0.00002685, 0.00005106, 0.00006143, 0.00005477, 0.00003301, 0.00000265, -0.00002730, -0.00004806, -0.00005384, -0.00004352, -0.00002088, 0.00000678, 0.00003102, 0.00004489, 0.00004500, 0.00003227, 0.00001133, -0.00001119, -0.00002893, -0.00003773, -0.00003664, -0.00002785, -0.00001555, -0.00000424, 0.00000277, 0.00000435, 0.00000156, -0.00000306, -0.00000673, -0.00000752, -0.00000510, -0.00000066, 0.00000374, 0.00000604, 0.00000516, 0.00000139, -0.00000374, -0.00000806, -0.00000966, -0.00000765, -0.00000248, 0.00000422, 0.00001025, 0.00001363, 0.00001330, 0.00000940, 0.00000320, -0.00000347, -0.00000884, -0.00001180, -0.00001215, -0.00001054, -0.00000803, -0.00000570, -0.00000413, -0.00000336, -0.00000291, -0.00000212, -0.00000051, 0.00000201, 0.00000507, 0.00000802, 0.00001022, 0.00001123, },
		{ -0.00000544, -0.00000663, -0.00000707, -0.00000692, -0.00000623, -0.00000493, -0.00000292, -0.00000029, 0.00000261, 0.00000510, 0.00000642, 0.00000591, 0.00000335, -0.00000091, -0.00000597, -0.00001061, -0.00001368, -0.00001442, -0.00001268, -0.00000894, -0.00000404, 0.00000112, 0.00000587, 0.00000989, 0.00001314, 0.00001565, 0.00001725, 0.00001742, 0.00001543, 0.00001067, 0.00000301, -0.00000681, -0.00001713, -0.00002575, -0.00003051, -0.00003009, -0.00002457, -0.00001554, -0.00000573, 0.00000187, 0.00000508, 0.00000336, -0.00000190, -0.00000782, -0.00001104, -0.00000912, -0.00000159, 0.00000960, 0.00002053, 0.00002661, 0.00002429, 0.00001255, -0.00000634, -0.00002720, -0.00004345, -0.00004919, -0.00004134, -0.00002093, 0.00000688, 0.00003421, 0.00005278, 0.00005662, 0.00004404, 0.00001836, -0.00001297, -0.00004063, -0.00005630, -0.00005527, -0.00003791, -0.00000953, 0.00002127, 0.00004538, 0.00005587, 0.00005014, 0.00003046, 0.00000307, -0.00002392, -0.00004311, -0.00005004, -0.00004421, -0.00002896, -0.00000995, 0.00000677, 0.00001676, 0.00001836, 0.00001284, 0.00000357, -0.00000543, -0.00001098, -0.00001174, -0.00000838, -0.00000302, 0.00000177, 0.00000410, 0.00000343, 0.00000065, -0.00000242, -0.00000393, -0.00000277, 0.00000093, 0.00000583, 0.00000996, 0.00001148, 0.00000942, 0.00000405, -0.00000322, -0.00001031, -0.00001529, -0.00001694, -0.00001515, -0.00001085, -0.00000554, -0.00000078, 0.00000246, 0.00000393, 0.00000418, 0.00000408, 0.00000451, 0.00000586, 0.00000795, 0.00001015, 0.00001160, 0.00001165, 0.00001006, 0.00000710, 0.00000341, -0.00000026, -0.00000331, },
		{ -0.00000144, -0.00000001, 0.00000098, 0.00000154, 0.00000157, 0.00000089, -0.00000068, -0.00000313, -0.00000614, -0.00000908, -0.00001111, -0.00001146, -0.00000970, -0.00000591, -0.00000071, 0.00000493, 0.00000998, 0.00001361, 0.00001542, 0.00001543, 0.00001394, 0.00001128, 0.00000763, 0.00000301, -0.00000265, -0.00000921, -0.00001608, -0.00002223, -0.00002630, -0.00002706, -0.00002388, -0.00001716, -0.00000838, 0.00000024, 0.00000636, 0.00000842, 0.00000618, 0.00000097, -0.00000471, -0.00000807, -0.00000712, -0.00000156, 0.00000692, 0.00001503, 0.00001903, 0.00001612, 0.00000571, -0.00001004, -0.00002663, -0.00003853, -0.00004104, -0.00003207, -0.00001309, 0.00001093, 0.00003292, 0.00004589, 0.00004526, 0.00003048, 0.00000549, -0.00002235, -0.00004448, -0.00005384, -0.00004714, -0.00002608, 0.00000321, 0.00003193, 0.00005143, 0.00005585, 0.00004396, 0.00001944, -0.00001033, -0.00003667, -0.00005229, -0.00005345, -0.00004093, -0.00001948, 0.00000387, 0.00002216, 0.00003067, 0.00002814, 0.00001688, 0.00000158, -0.00001236, -0.00002071, -0.00002163, -0.00001594, -0.00000651, 0.00000301, 0.00000956, 0.00001164, 0.00000965, 0.00000538, 0.00000116, -0.00000121, -0.00000107, 0.00000088, 0.00000305, 0.00000370, 0.00000177, -0.00000271, -0.00000853, -0.00001382, -0.00001672, -0.00001614, -0.00001209, -0.00000569, 0.00000128, 0.00000702, 0.00001036, 0.00001104, 0.00000976, 0.00000773, 0.00000621, 0.00000593, 0.00000692, 0.00000850, 0.00000968, 0.00000951, 0.00000756, 0.00000401, -0.00000038, -0.00000459, -0.00000770, -0.00000918, -0.00000900, -0.00000757, -0.00000548, -0.00000331, },
		{ -0.00000329, -0.00000442, -0.00000536, -0.00000605, -0.00000630, -0.00000584, -0.00000436, -0.00000173, 0.00000189, 0.00000598, 0.00000980, 0.00001254, 0.00001358, 0.00001269, 0.00001005, 0.00000617, 0.00000166, -0.00000297, -0.00000740, -0.00001151, -0.00001522, -0.00001831, -0.00002035, -0.00002071, -0.00001884, -0.00001455, -0.00000830, -0.00000126, 0.00000493, 0.00000865, 0.00000887, 0.00000565, 0.00000023, -0.00000525, -0.00000849, -0.00000790, -0.00000328, 0.00000389, 0.00001082, 0.00001436, 0.00001214, 0.00000368, -0.00000923, -0.00002282, -0.00003249, -0.00003437, -0.00002679, -0.00001112, 0.00000840, 0.00002585, 0.00003546, 0.00003366, 0.00002036, -0.00000072, -0.00002311, -0.00003957, -0.00004446, -0.00003564, -0.00001536, 0.00001037, 0.00003361, 0.00004700, 0.00004614, 0.00003108, 0.00000628, -0.00002078, -0.00004195, -0.00005105, -0.00004572, -0.00002812, -0.00000406, 0.00001891, 0.00003398, 0.00003717, 0.00002842, 0.00001139, -0.00000797, -0.00002343, -0.00003051, -0.00002777, -0.00001696, -0.00000221, 0.00001158, 0.00002035, 0.00002216, 0.00001750, 0.00000885, -0.00000046, -0.00000749, -0.00001070, -0.00001022, -0.00000758, -0.00000482, -0.00000365, -0.00000469, -0.00000732, -0.00001005, -0.00001117, -0.00000952, -0.00000500, 0.00000141, 0.00000798, 0.00001295, 0.00001513, 0.00001434, 0.00001138, 0.00000770, 0.00000477, 0.00000355, 0.00000413, 0.00000578, 0.00000727, 0.00000742, 0.00000556, 0.00000180, -0.00000304, -0.00000770, -0.00001100, -0.00001219, -0.00001120, -0.00000858, -0.00000528, -0.00000227, -0.00000023, 0.00000056, 0.00000026, -0.00000074, -0.00000202, },
		{ 0.00000081, 0.00000249, 0.00000418, 0.00000586, 0.00000734, 0.00000832, 0.00000842, 0.00000734, 0.00000502, 0.00000168, -0.00000221, -0.00000607, -0.00000942, -0.00001194, -0.00001356, -0.00001437, -0.00001445, -0.00001380, -0.00001226, -0.00000967, -0.00000599, -0.00000158, 0.00000282, 0.00000617, 0.00000746, 0.00000614, 0.00000243, -0.00000259, -0.00000718, -0.00000955, -0.00000848, -0.00000395, 0.00000272, 0.00000911, 0.00001250, 0.00001086, 0.00000370, -0.00000751, -0.00001959, -0.00002861, -0.00003117, -0.00002570, -0.00001316, 0.00000301, 0.00001788, 0.00002661, 0.00002608, 0.00001611, -0.00000033, -0.00001797, -0.00003086, -0.00003443, -0.00002705, -0.00001074, 0.00000943, 0.00002689, 0.00003579, 0.00003294, 0.00001898, -0.00000178, -0.00002270, -0.00003703, -0.00004011, -0.00003092, -0.00001241, 0.00000946, 0.00002770, 0.00003661, 0.00003359, 0.00001994, 0.00000034, -0.00001877, -0.00003133, -0.00003360, -0.00002528, -0.00000944, 0.00000863, 0.00002324, 0.00003013, 0.00002775, 0.00001752, 0.00000310, -0.00001094, -0.00002072, -0.00002418, -0.00002152, -0.00001484, -0.00000711, -0.00000104, 0.00000187, 0.00000167, -0.00000026, -0.00000201, -0.00000194, 0.00000061, 0.00000510, 0.00001012, 0.00001396, 0.00001536, 0.00001399, 0.00001054, 0.00000639, 0.00000306, 0.00000156, 0.00000202, 0.00000368, 0.00000523, 0.00000535, 0.00000326, -0.00000095, -0.00000634, -0.00001148, -0.00001493, -0.00001579, -0.00001394, -0.00001011, -0.00000551, -0.00000146, 0.00000109, 0.00000178, 0.00000088, -0.00000090, -0.00000273, -0.00000394, -0.00000422, -0.00000362, -0.00000239, -0.00000084, },
		{ 0.00000232, 0.00000080, -0.00000090, -0.00000283, -0.00000492, -0.00000698, -0.00000873, -0.00000990, -0.00001031, -0.00001000, -0.00000910, -0.00000784, -0.00000637, -0.00000472, -0.00000287, -0.00000080, 0.00000136, 0.00000324, 0.00000428, 0.00000394, 0.00000197, -0.00000138, -0.00000524, -0.00000836, -0.00000941, -0.00000759, -0.00000301, 0.00000312, 0.00000875, 0.00001158, 0.00000989, 0.00000325, -0.00000707, -0.00001833, -0.00002709, -0.00003027, -0.00002630, -0.00001578, -0.00000151, 0.00001232, 0.00002142, 0.00002282, 0.00001598, 0.00000303, -0.00001178, -0.00002350, -0.00002812, -0.00002395, -0.00001227, 0.00000301, 0.00001673, 0.00002419, 0.00002279, 0.00001302, -0.00000169, -0.00001614, -0.00002512, -0.00002526, -0.00001626, -0.00000101, 0.00001534, 0.00002715, 0.00003024, 0.00002336, 0.00000863, -0.00000910, -0.00002394, -0.00003091, -0.00002765, -0.00001517, 0.00000246, 0.00001950, 0.00003046, 0.00003185, 0.00002327, 0.00000746, -0.00001075, -0.00002601, -0.00003420, -0.00003362, -0.00002532, -0.00001253, 0.00000058, 0.00001037, 0.00001480, 0.00001389, 0.00000947, 0.00000423, 0.00000069, 0.00000026, 0.00000289, 0.00000728, 0.00001150, 0.00001385, 0.00001351, 0.00001076, 0.00000677, 0.00000303, 0.00000071, 0.00000017, 0.00000086, 0.00000160, 0.00000112, -0.00000135, -0.00000574, -0.00001108, -0.00001588, -0.00001863, -0.00001838, -0.00001509, -0.00000964, -0.00000351, 0.00000170, 0.00000482, 0.00000544, 0.00000397, 0.00000140, -0.00000108, -0.00000253, -0.00000248, -0.00000109, 0.00000108, 0.00000331, 0.00000497, 0.00000574, 0.00000563, 0.00000486, 0.00000369, },
		{ -0.00000685, -0.00000627, -0.00000560, -0.00000480, -0.00000387, -0.00000286, -0.00000189, -0.00000104, -0.00000035, 0.00000021, 0.00000069, 0.00000109, 0.00000129, 0.00000104, 0.00000009, -0.00000168, -0.00000406, -0.00000648, -0.00000808, -0.00000799, -0.00000573, -0.00000149, 0.00000370, 0.00000816, 0.00001004, 0.00000796, 0.00000162, -0.00000789, -0.00001819, -0.00002629, -0.00002943, -0.00002614, -0.00001680, -0.00000371, 0.00000946, 0.00001881, 0.00002146, 0.00001655, 0.00000558, -0.00000796, -0.00001973, -0.00002595, -0.00002465, -0.00001637, -0.00000392, 0.00000852, 0.00001690, 0.00001869, 0.00001375, 0.00000435, -0.00000567, -0.00001233, -0.00001297, -0.00000723, 0.00000276, 0.00001317, 0.00001988, 0.00002000, 0.00001303, 0.00000110, -0.00001173, -0.00002083, -0.00002273, -0.00001638, -0.00000363, 0.00001131, 0.00002328, 0.00002788, 0.00002305, 0.00000975, -0.00000826, -0.00002567, -0.00003729, -0.00003973, -0.00003245, -0.00001779, -0.00000016, 0.00001546, 0.00002511, 0.00002700, 0.00002190, 0.00001262, 0.00000294, -0.00000378, -0.00000564, -0.00000264, 0.00000349, 0.00001013, 0.00001480, 0.00001599, 0.00001358, 0.00000868, 0.00000306, -0.00000163, -0.00000447, -0.00000555, -0.00000576, -0.00000629, -0.00000803, -0.00001112, -0.00001483, -0.00001783, -0.00001878, -0.00001680, -0.00001194, -0.00000518, 0.00000184, 0.00000742, 0.00001032, 0.00001016, 0.00000752, 0.00000367, 0.00000009, -0.00000203, -0.00000216, -0.00000053, 0.00000200, 0.00000437, 0.00000565, 0.00000537, 0.00000360, 0.00000086, -0.00000212, -0.00000471, -0.00000649, -0.00000739, -0.00000758, -0.00000733, },
		{ 0.00000177, 0.00000145, 0.00000103, 0.00000057, 0.00000011, -0.00000032, -0.00000074, -0.00000118, -0.00000176, -0.00000262, -0.00000380, -0.00000519, -0.00000641, -0.00000694, -0.00000624, -0.00000403, -0.00000051, 0.00000349, 0.00000670, 0.00000767, 0.00000536, -0.00000044, -0.00000880, -0.00001774, -0.00002473, -0.00002743, -0.00002449, -0.00001618, -0.00000441, 0.00000765, 0.00001652, 0.00001948, 0.00001554, 0.00000579, -0.00000683, -0.00001845, -0.00002544, -0.00002564, -0.00001902, -0.00000770, 0.00000479, 0.00001469, 0.00001932, 0.00001786, 0.00001161, 0.00000341, -0.00000344, -0.00000643, -0.00000472, 0.00000058, 0.00000693, 0.00001137, 0.00001173, 0.00000750, 0.00000012, -0.00000758, -0.00001243, -0.00001222, -0.00000662, 0.00000254, 0.00001182, 0.00001735, 0.00001628, 0.00000790, -0.00000593, -0.00002128, -0.00003334, -0.00003803, -0.00003336, -0.00002020, -0.00000198, 0.00001632, 0.00002979, 0.00003508, 0.00003145, 0.00002085, 0.00000721, -0.00000493, -0.00001191, -0.00001202, -0.00000596, 0.00000365, 0.00001322, 0.00001950, 0.00002056, 0.00001631, 0.00000831, -0.00000095, -0.00000899, -0.00001417, -0.00001612, -0.00001562, -0.00001408, -0.00001285, -0.00001263, -0.00001322, -0.00001367, -0.00001283, -0.00000984, -0.00000467, 0.00000183, 0.00000815, 0.00001268, 0.00001426, 0.00001264, 0.00000852, 0.00000335, -0.00000121, -0.00000385, -0.00000404, -0.00000215, 0.00000074, 0.00000328, 0.00000431, 0.00000330, 0.00000043, -0.00000348, -0.00000733, -0.00001010, -0.00001118, -0.00001048, -0.00000842, -0.00000565, -0.00000287, -0.00000061, 0.00000091, 0.00000169, 0.00000191, },
		{ -0.00000222, -0.00000254, -0.00000284, -0.00000323, -0.00000379, -0.00000451, -0.00000528, -0.00000585, -0.00000588, -0.00000504, -0.00000316, -0.00000044, 0.00000252, 0.00000474, 0.00000517, 0.00000299, -0.00000195, -0.00000893, -0.00001637, -0.00002221, -0.00002452, -0.00002211, -0.00001512, -0.00000508, 0.00000537, 0.00001324, 0.00001610, 0.00001294, 0.00000452, -0.00000668, -0.00001725, -0.00002388, -0.00002439, -0.00001845, -0.00000774, 0.00000468, 0.00001527, 0.00002125, 0.00002138, 0.00001635, 0.00000841, 0.00000051, -0.00000472, -0.00000592, -0.00000336, 0.00000132, 0.00000578, 0.00000802, 0.00000710, 0.00000352, -0.00000106, -0.00000455, -0.00000540, -0.00000325, 0.00000076, 0.00000442, 0.00000535, 0.00000196, -0.00000569, -0.00001571, -0.00002489, -0.00002969, -0.00002761, -0.00001807, -0.00000291, 0.00001412, 0.00002845, 0.00003614, 0.00003516, 0.00002609, 0.00001194, -0.00000280, -0.00001360, -0.00001738, -0.00001337, -0.00000337, 0.00000898, 0.00001942, 0.00002446, 0.00002243, 0.00001392, 0.00000146, -0.00001149, -0.00002164, -0.00002688, -0.00002682, -0.00002264, -0.00001647, -0.00001049, -0.00000620, -0.00000395, -0.00000308, -0.00000236, -0.00000064, 0.00000256, 0.00000681, 0.00001098, 0.00001365, 0.00001375, 0.00001103, 0.00000615, 0.00000053, -0.00000418, -0.00000671, -0.00000656, -0.00000424, -0.00000095, 0.00000178, 0.00000269, 0.00000125, -0.00000221, -0.00000664, -0.00001064, -0.00001295, -0.00001284, -0.00001033, -0.00000612, -0.00000129, 0.00000303, 0.00000601, 0.00000730, 0.00000702, 0.00000563, 0.00000370, 0.00000177, 0.00000016, -0.00000101, -0.00000176, },
		{ -0.00000440, -0.00000508, -0.00000552, -0.00000546, -0.00000467, -0.00000311, -0.00000094, 0.00000134, 0.00000300, 0.00000323, 0.00000141, -0.00000261, -0.00000830, -0.00001445, -0.00001943, -0.00002163, -0.00002002, -0.00001456, -0.00000640, 0.00000236, 0.00000925, 0.00001216, 0.00001006, 0.00000345, -0.00000572, -0.00001459, -0.00002025, -0.00002067, -0.00001538, -0.00000567, 0.00000581, 0.00001590, 0.00002187, 0.00002237, 0.00001773, 0.00000983, 0.00000140, -0.00000491, -0.00000744, -0.00000595, -0.00000161, 0.00000353, 0.00000738, 0.00000855, 0.00000680, 0.00000303, -0.00000124, -0.00000454, -0.00000607, -0.00000604, -0.00000549, -0.00000577, -0.00000777, -0.00001146, -0.00001563, -0.00001828, -0.00001737, -0.00001171, -0.00000158, 0.00001105, 0.00002302, 0.00003094, 0.00003232, 0.00002650, 0.00001503, 0.00000127, -0.00001064, -0.00001707, -0.00001612, -0.00000830, 0.00000364, 0.00001558, 0.00002331, 0.00002395, 0.00001683, 0.00000367, -0.00001199, -0.00002593, -0.00003459, -0.00003612, -0.00003078, -0.00002069, -0.00000903, 0.00000112, 0.00000767, 0.00001005, 0.00000917, 0.00000680, 0.00000478, 0.00000433, 0.00000557, 0.00000770, 0.00000935, 0.00000930, 0.00000699, 0.00000280, -0.00000212, -0.00000625, -0.00000836, -0.00000795, -0.00000548, -0.00000216, 0.00000051, 0.00000128, -0.00000035, -0.00000396, -0.00000832, -0.00001190, -0.00001332, -0.00001188, -0.00000776, -0.00000193, 0.00000417, 0.00000909, 0.00001184, 0.00001205, 0.00001006, 0.00000669, 0.00000295, -0.00000032, -0.00000259, -0.00000375, -0.00000401, -0.00000375, -0.00000336, -0.00000316, -0.00000328, -0.00000373, },
		{ -0.00000349, -0.00000152, 0.00000047, 0.00000188, 0.00000206, 0.00000056, -0.00000275, -0.00000747, -0.00001268, -0.00001709, -0.00001935, -0.00001856, -0.00001452, -0.00000801, -0.00000064, 0.00000559, 0.00000884, 0.00000808, 0.00000352, -0.00000344, -0.00001050, -0.00001521, -0.00001573, -0.00001147, -0.00000335, 0.00000648, 0.00001530, 0.00002066, 0.00002118, 0.00001698, 0.00000958, 0.00000143, -0.00000499, -0.00000794, -0.00000696, -0.00000294, 0.00000228, 0.00000659, 0.00000838, 0.00000702, 0.00000303, -0.00000225, -0.00000723, -0.00001066, -0.00001206, -0.00001175, -0.00001057, -0.00000937, -0.00000857, -0.00000788, -0.00000648, -0.00000340, 0.00000186, 0.00000892, 0.00001642, 0.00002233, 0.00002464, 0.00002215, 0.00001506, 0.00000509, -0.00000494, -0.00001195, -0.00001366, -0.00000949, -0.00000079, 0.00000939, 0.00001731, 0.00001971, 0.00001495, 0.00000362, -0.00001154, -0.00002647, -0.00003699, -0.00004015, -0.00003508, -0.00002322, -0.00000785, 0.00000704, 0.00001792, 0.00002275, 0.00002147, 0.00001576, 0.00000834, 0.00000194, -0.00000163, -0.00000189, 0.00000025, 0.00000311, 0.00000491, 0.00000455, 0.00000195, -0.00000199, -0.00000582, -0.00000817, -0.00000831, -0.00000644, -0.00000358, -0.00000113, -0.00000029, -0.00000161, -0.00000469, -0.00000833, -0.00001101, -0.00001139, -0.00000884, -0.00000367, 0.00000293, 0.00000933, 0.00001392, 0.00001565, 0.00001430, 0.00001046, 0.00000531, 0.00000020, -0.00000375, -0.00000594, -0.00000638, -0.00000554, -0.00000413, -0.00000286, -0.00000220, -0.00000230, -0.00000305, -0.00000414, -0.00000518, -0.00000582, -0.00000580, -0.00000500, },
		{ 0.00000010, -0.00000282, -0.00000691, -0.00001145, -0.00001538, -0.00001764, -0.00001740, -0.00001440, -0.00000913, -0.00000279, 0.00000302, 0.00000671, 0.00000725, 0.00000456, -0.00000041, -0.00000592, -0.00000993, -0.00001080, -0.00000783, -0.00000156, 0.00000639, 0.00001375, 0.00001839, 0.00001896, 0.00001537, 0.00000878, 0.00000124, -0.00000502, -0.00000831, -0.00000803, -0.00000479, -0.00000017, 0.00000389, 0.00000578, 0.00000474, 0.00000107, -0.00000401, -0.00000891, -0.00001219, -0.00001306, -0.00001158, -0.00000846, -0.00000473, -0.00000129, 0.00000147, 0.00000367, 0.00000577, 0.00000821, 0.00001102, 0.00001371, 0.00001535, 0.00001497, 0.00001206, 0.00000692, 0.00000072, -0.00000474, -0.00000769, -0.00000704, -0.00000296, 0.00000307, 0.00000859, 0.00001096, 0.00000832, 0.00000034, -0.00001148, -0.00002410, -0.00003383, -0.00003746, -0.00003333, -0.00002193, -0.00000585, 0.00001098, 0.00002441, 0.00003136, 0.00003062, 0.00002323, 0.00001195, 0.00000037, -0.00000826, -0.00001200, -0.00001068, -0.00000579, 0.00000023, 0.00000490, 0.00000656, 0.00000485, 0.00000069, -0.00000419, -0.00000801, -0.00000958, -0.00000871, -0.00000621, -0.00000344, -0.00000175, -0.00000188, -0.00000365, -0.00000610, -0.00000780, -0.00000747, -0.00000453, 0.00000071, 0.00000704, 0.00001278, 0.00001633, 0.00001672, 0.00001385, 0.00000857, 0.00000230, -0.00000336, -0.00000719, -0.00000863, -0.00000787, -0.00000570, -0.00000316, -0.00000121, -0.00000044, -0.00000096, -0.00000244, -0.00000431, -0.00000593, -0.00000683, -0.00000676, -0.00000571, -0.00000392, -0.00000180, 0.00000015, 0.00000141, 0.00000150, },
		{ -0.00001613, -0.00001607, -0.00001369, -0.00000929, -0.00000375, 0.00000167, 0.00000563, 0.00000716, 0.00000597, 0.00000262, -0.00000162, -0.00000510, -0.00000637, -0.00000465, -0.00000016, 0.00000592, 0.00001180, 0.00001564, 0.00001614, 0.00001302, 0.00000711, 0.00000008, -0.00000611, -0.00000983, -0.00001033, -0.00000792, -0.00000383, 0.00000023, 0.00000275, 0.00000286, 0.00000065, -0.00000297, -0.00000659, -0.00000885, -0.00000893, -0.00000679, -0.00000308, 0.00000111, 0.00000471, 0.00000704, 0.00000800, 0.00000799, 0.00000763, 0.00000743, 0.00000755, 0.00000773, 0.00000746, 0.00000631, 0.00000416, 0.00000141, -0.00000117, -0.00000275, -0.00000284, -0.00000161, 0.00000004, 0.00000073, -0.00000086, -0.00000539, -0.00001242, -0.00002034, -0.00002672, -0.00002903, -0.00002555, -0.00001605, -0.00000208, 0.00001332, 0.00002641, 0.00003387, 0.00003381, 0.00002638, 0.00001373, -0.00000056, -0.00001264, -0.00001948, -0.00001978, -0.00001428, -0.00000535, 0.00000377, 0.00001011, 0.00001189, 0.00000895, 0.00000268, -0.00000460, -0.00001048, -0.00001333, -0.00001271, -0.00000944, -0.00000513, -0.00000150, 0.00000031, 0.00000009, -0.00000140, -0.00000279, -0.00000277, -0.00000057, 0.00000364, 0.00000881, 0.00001337, 0.00001579, 0.00001515, 0.00001144, 0.00000559, -0.00000086, -0.00000626, -0.00000936, -0.00000973, -0.00000775, -0.00000448, -0.00000122, 0.00000091, 0.00000130, -0.00000003, -0.00000251, -0.00000526, -0.00000742, -0.00000841, -0.00000800, -0.00000639, -0.00000404, -0.00000157, 0.00000045, 0.00000152, 0.00000132, -0.00000025, -0.00000307, -0.00000677, -0.00001073, -0.00001413, },
		{ 0.00000134, 0.00000530, 0.00000742, 0.00000732, 0.00000525, 0.00000210, -0.00000089, -0.00000249, -0.00000190, 0.00000089, 0.00000511, 0.00000938, 0.00001216, 0.00001228, 0.00000935, 0.00000392, -0.00000265, -0.00000862, -0.00001244, -0.00001322, -0.00001101, -0.00000676, -0.00000194, 0.00000194, 0.00000387, 0.00000361, 0.00000171, -0.00000073, -0.00000254, -0.00000287, -0.00000154, 0.00000102, 0.00000391, 0.00000618, 0.00000719, 0.00000684, 0.00000551, 0.00000395, 0.00000285, 0.00000261, 0.00000318, 0.00000408, 0.00000467, 0.00000442, 0.00000312, 0.00000096, -0.00000158, -0.00000400, -0.00000600, -0.00000765, -0.00000930, -0.00001132, -0.00001383, -0.00001642, -0.00001813, -0.00001771, -0.00001406, -0.00000678, 0.00000344, 0.00001480, 0.00002471, 0.00003055, 0.00003042, 0.00002390, 0.00001228, -0.00000164, -0.00001433, -0.00002252, -0.00002421, -0.00001927, -0.00000947, 0.00000204, 0.00001174, 0.00001681, 0.00001595, 0.00000975, 0.00000038, -0.00000916, -0.00001606, -0.00001853, -0.00001634, -0.00001070, -0.00000375, 0.00000225, 0.00000572, 0.00000622, 0.00000448, 0.00000202, 0.00000047, 0.00000091, 0.00000349, 0.00000736, 0.00001106, 0.00001303, 0.00001229, 0.00000875, 0.00000328, -0.00000262, -0.00000732, -0.00000962, -0.00000915, -0.00000644, -0.00000268, 0.00000068, 0.00000248, 0.00000214, -0.00000013, -0.00000354, -0.00000695, -0.00000933, -0.00001005, -0.00000899, -0.00000657, -0.00000352, -0.00000066, 0.00000132, 0.00000201, 0.00000129, -0.00000070, -0.00000364, -0.00000706, -0.00001040, -0.00001308, -0.00001451, -0.00001427, -0.00001216, -0.00000838, -0.00000357, },
		{ 0.00000499, 0.00000243, 0.00000055, 0.00000007, 0.00000120, 0.00000351, 0.00000602, 0.00000755, 0.00000712, 0.00000430, -0.00000052, -0.00000626, -0.00001143, -0.00001459, -0.00001483, -0.00001205, -0.00000701, -0.00000106, 0.00000428, 0.00000781, 0.00000898, 0.00000803, 0.00000577, 0.00000329, 0.00000154, 0.00000099, 0.00000154, 0.00000266, 0.00000364, 0.00000390, 0.00000330, 0.00000211, 0.00000089, 0.00000024, 0.00000051, 0.00000161, 0.00000303, 0.00000403, 0.00000391, 0.00000231, -0.00000071, -0.00000461, -0.00000862, -0.00001197, -0.00001412, -0.00001488, -0.00001434, -0.00001273, -0.00001022, -0.00000680, -0.00000236, 0.00000314, 0.00000942, 0.00001573, 0.00002087, 0.00002340, 0.00002214, 0.00001664, 0.00000749, -0.00000355, -0.00001397, -0.00002112, -0.00002304, -0.00001907, -0.00001022, 0.00000109, 0.00001164, 0.00001836, 0.00001930, 0.00001420, 0.00000456, -0.00000675, -0.00001645, -0.00002185, -0.00002160, -0.00001611, -0.00000731, 0.00000208, 0.00000940, 0.00001292, 0.00001231, 0.00000864, 0.00000386, 0.00000004, -0.00000135, 0.00000008, 0.00000354, 0.00000747, 0.00001014, 0.00001031, 0.00000771, 0.00000304, -0.00000224, -0.00000651, -0.00000852, -0.00000790, -0.00000516, -0.00000155, 0.00000149, 0.00000275, 0.00000174, -0.00000121, -0.00000514, -0.00000876, -0.00001094, -0.00001107, -0.00000918, -0.00000591, -0.00000226, 0.00000080, 0.00000251, 0.00000257, 0.00000109, -0.00000150, -0.00000463, -0.00000773, -0.00001032, -0.00001201, -0.00001256, -0.00001179, -0.00000969, -0.00000640, -0.00000234, 0.00000187, 0.00000547, 0.00000777, 0.00000836, 0.00000726, },
		{ 0.00000088, 0.00000175, 0.00000216, 0.00000136, -0.00000103, -0.00000479, -0.00000911, -0.00001280, -0.00001465, -0.00001382, -0.00001019, -0.00000440, 0.00000229, 0.00000838, 0.00001258, 0.00001414, 0.00001310, 0.00001014, 0.00000634, 0.00000283, 0.00000037, -0.00000077, -0.00000082, -0.00000032, 0.00000016, 0.00000027, -0.00000000, -0.00000038, -0.00000051, -0.00000014, 0.00000066, 0.00000150, 0.00000180, 0.00000100, -0.00000115, -0.00000447, -0.00000834, -0.00001186, -0.00001410, -0.00001442, -0.00001261, -0.00000895, -0.00000406, 0.00000127, 0.00000635, 0.00001064, 0.00001384, 0.00001577, 0.00001626, 0.00001509, 0.00001207, 0.00000719, 0.00000082, -0.00000620, -0.00001255, -0.00001673, -0.00001744, -0.00001411, -0.00000717, 0.00000186, 0.00001067, 0.00001678, 0.00001827, 0.00001440, 0.00000598, -0.00000484, -0.00001511, -0.00002192, -0.00002334, -0.00001895, -0.00001002, 0.00000090, 0.00001077, 0.00001703, 0.00001829, 0.00001480, 0.00000821, 0.00000099, -0.00000441, -0.00000638, -0.00000460, -0.00000008, 0.00000530, 0.00000945, 0.00001087, 0.00000908, 0.00000476, -0.00000058, -0.00000519, -0.00000768, -0.00000753, -0.00000519, -0.00000189, 0.00000087, 0.00000189, 0.00000066, -0.00000245, -0.00000639, -0.00000980, -0.00001154, -0.00001101, -0.00000841, -0.00000454, -0.00000059, 0.00000233, 0.00000349, 0.00000272, 0.00000039, -0.00000279, -0.00000599, -0.00000857, -0.00001013, -0.00001056, -0.00000993, -0.00000841, -0.00000616, -0.00000336, -0.00000020, 0.00000300, 0.00000583, 0.00000783, 0.00000864, 0.00000814, 0.00000652, 0.00000430, 0.00000216, 0.00000074, 0.00000035, },
		{ -0.00000737, -0.00000993, -0.00001188, -0.00001230, -0.00001052, -0.00000642, -0.00000053, 0.00000603, 0.00001190, 0.00001582, 0.00001700, 0.00001538, 0.00001158, 0.00000669, 0.00000193, -0.00000171, -0.00000373, -0.00000418, -0.00000353, -0.00000245, -0.00000151, -0.00000101, -0.00000096, -0.00000114, -0.00000131, -0.00000137, -0.00000142, -0.00000176, -0.00000268, -0.00000431, -0.00000646, -0.00000862, -0.00001006, -0.00001008, -0.00000826, -0.00000463, 0.00000031, 0.00000566, 0.00001042, 0.00001371, 0.00001499, 0.00001419, 0.00001158, 0.00000773, 0.00000326, -0.00000127, -0.00000538, -0.00000867, -0.00001076, -0.00001127, -0.00000989, -0.00000655, -0.00000156, 0.00000428, 0.00000972, 0.00001332, 0.00001382, 0.00001061, 0.00000401, -0.00000461, -0.00001315, -0.00001928, -0.00002111, -0.00001784, -0.00001008, 0.00000030, 0.00001058, 0.00001806, 0.00002080, 0.00001826, 0.00001145, 0.00000261, -0.00000552, -0.00001051, -0.00001105, -0.00000725, -0.00000062, 0.00000653, 0.00001185, 0.00001370, 0.00001168, 0.00000662, 0.00000030, -0.00000529, -0.00000856, -0.00000888, -0.00000670, -0.00000330, -0.00000027, 0.00000105, 0.00000009, -0.00000278, -0.00000650, -0.00000965, -0.00001105, -0.00001014, -0.00000718, -0.00000311, 0.00000078, 0.00000332, 0.00000380, 0.00000223, -0.00000079, -0.00000429, -0.00000728, -0.00000904, -0.00000930, -0.00000825, -0.00000632, -0.00000401, -0.00000170, 0.00000042, 0.00000235, 0.00000410, 0.00000565, 0.00000685, 0.00000745, 0.00000723, 0.00000613, 0.00000431, 0.00000216, 0.00000016, -0.00000130, -0.00000210, -0.00000241, -0.00000270, -0.00000347, -0.00000504, },
		{ -0.00000272, 0.00000253, 0.00000826, 0.00001330, 0.00001655, 0.00001725, 0.00001527, 0.00001114, 0.00000587, 0.00000067, -0.00000341, -0.00000577, -0.00000634, -0.00000558, -0.00000419, -0.00000288, -0.00000215, -0.00000212, -0.00000262, -0.00000332, -0.00000390, -0.00000419, -0.00000424, -0.00000419, -0.00000420, -0.00000427, -0.00000420, -0.00000365, -0.00000226, 0.00000013, 0.00000339, 0.00000700, 0.00001021, 0.00001221, 0.00001240, 0.00001056, 0.00000697, 0.00000233, -0.00000244, -0.00000642, -0.00000893, -0.00000966, -0.00000865, -0.00000625, -0.00000295, 0.00000073, 0.00000427, 0.00000720, 0.00000906, 0.00000943, 0.00000799, 0.00000464, -0.00000030, -0.00000608, -0.00001150, -0.00001519, -0.00001594, -0.00001311, -0.00000693, 0.00000143, 0.00001003, 0.00001670, 0.00001961, 0.00001788, 0.00001191, 0.00000331, -0.00000550, -0.00001201, -0.00001436, -0.00001191, -0.00000549, 0.00000293, 0.00001079, 0.00001577, 0.00001648, 0.00001289, 0.00000624, -0.00000138, -0.00000774, -0.00001121, -0.00001122, -0.00000837, -0.00000416, -0.00000038, 0.00000151, 0.00000090, -0.00000186, -0.00000560, -0.00000883, -0.00001028, -0.00000936, -0.00000638, -0.00000235, 0.00000137, 0.00000355, 0.00000355, 0.00000148, -0.00000186, -0.00000532, -0.00000776, -0.00000848, -0.00000737, -0.00000488, -0.00000181, 0.00000101, 0.00000303, 0.00000406, 0.00000430, 0.00000417, 0.00000404, 0.00000410, 0.00000428, 0.00000432, 0.00000391, 0.00000286, 0.00000124, -0.00000069, -0.00000255, -0.00000404, -0.00000504, -0.00000566, -0.00000618, -0.00000688, -0.00000784, -0.00000884, -0.00000937, -0.00000880, -0.00000663, },
		{ 0.00001597, 0.00001418, 0.00001035, 0.00000534, 0.00000021, -0.00000400, -0.00000664, -0.00000755, -0.00000707, -0.00000585, -0.00000459, -0.00000384, -0.00000377, -0.00000422, -0.00000478, -0.00000504, -0.00000469, -0.00000368, -0.00000218, -0.00000050, 0.00000108, 0.00000241, 0.00000351, 0.00000452, 0.00000556, 0.00000660, 0.00000746, 0.00000778, 0.00000716, 0.00000536, 0.00000245, -0.00000118, -0.00000480, -0.00000762, -0.00000892, -0.00000835, -0.00000598, -0.00000234, 0.00000178, 0.00000548, 0.00000804, 0.00000900, 0.00000826, 0.00000604, 0.00000276, -0.00000107, -0.00000487, -0.00000812, -0.00001028, -0.00001091, -0.00000965, -0.00000644, -0.00000154, 0.00000431, 0.00001000, 0.00001423, 0.00001582, 0.00001409, 0.00000916, 0.00000204, -0.00000556, -0.00001164, -0.00001445, -0.00001306, -0.00000768, 0.00000030, 0.00000874, 0.00001526, 0.00001804, 0.00001630, 0.00001059, 0.00000256, -0.00000550, -0.00001143, -0.00001381, -0.00001242, -0.00000820, -0.00000291, 0.00000149, 0.00000353, 0.00000266, -0.00000059, -0.00000489, -0.00000858, -0.00001030, -0.00000948, -0.00000643, -0.00000229, 0.00000148, 0.00000363, 0.00000350, 0.00000131, -0.00000202, -0.00000519, -0.00000699, -0.00000675, -0.00000450, -0.00000096, 0.00000277, 0.00000560, 0.00000685, 0.00000637, 0.00000464, 0.00000242, 0.00000055, -0.00000045, -0.00000048, 0.00000017, 0.00000093, 0.00000125, 0.00000081, -0.00000041, -0.00000212, -0.00000389, -0.00000535, -0.00000632, -0.00000683, -0.00000705, -0.00000717, -0.00000718, -0.00000687, -0.00000588, -0.00000383, -0.00000056, 0.00000373, 0.00000843, 0.00001263, 0.00001538, },
		{ -0.00000366, -0.00000676, -0.00000827, -0.00000835, -0.00000749, -0.00000631, -0.00000533, -0.00000478, -0.00000456, -0.00000435, -0.00000376, -0.00000249, -0.00000055, 0.00000184, 0.00000421, 0.00000612, 0.00000722, 0.00000741, 0.00000680, 0.00000567, 0.00000430, 0.00000288, 0.00000146, -0.00000001, -0.00000161, -0.00000329, -0.00000484, -0.00000590, -0.00000605, -0.00000499, -0.00000271, 0.00000049, 0.00000396, 0.00000691, 0.00000860, 0.00000854, 0.00000666, 0.00000334, -0.00000070, -0.00000462, -0.00000762, -0.00000911, -0.00000888, -0.00000702, -0.00000387, 0.00000005, 0.00000416, 0.00000784, 0.00001053, 0.00001174, 0.00001109, 0.00000850, 0.00000421, -0.00000113, -0.00000648, -0.00001062, -0.00001242, -0.00001117, -0.00000687, -0.00000033, 0.00000694, 0.00001312, 0.00001651, 0.00001611, 0.00001189, 0.00000490, -0.00000303, -0.00000981, -0.00001370, -0.00001385, -0.00001055, -0.00000510, 0.00000059, 0.00000461, 0.00000573, 0.00000371, -0.00000064, -0.00000573, -0.00000980, -0.00001149, -0.00001030, -0.00000672, -0.00000203, 0.00000218, 0.00000458, 0.00000455, 0.00000236, -0.00000094, -0.00000392, -0.00000533, -0.00000451, -0.00000163, 0.00000238, 0.00000620, 0.00000857, 0.00000872, 0.00000663, 0.00000299, -0.00000103, -0.00000424, -0.00000581, -0.00000550, -0.00000370, -0.00000125, 0.00000091, 0.00000202, 0.00000179, 0.00000038, -0.00000171, -0.00000385, -0.00000555, -0.00000657, -0.00000692, -0.00000679, -0.00000636, -0.00000566, -0.00000454, -0.00000276, -0.00000012, 0.00000332, 0.00000721, 0.00001088, 0.00001351, 0.00001441, 0.00001323, 0.00001009, 0.00000562, 0.00000071, },
		{ -0.00000727, -0.00000582, -0.00000451, -0.00000341, -0.00000235, -0.00000107, 0.00000062, 0.00000271, 0.00000494, 0.00000686, 0.00000801, 0.00000805, 0.00000690, 0.00000480, 0.00000220, -0.00000038, -0.00000248, -0.00000385, -0.00000445, -0.00000441, -0.00000393, -0.00000317, -0.00000217, -0.00000092, 0.00000064, 0.00000243, 0.00000424, 0.00000568, 0.00000632, 0.00000579, 0.00000398, 0.00000110, -0.00000227, -0.00000538, -0.00000744, -0.00000787, -0.00000645, -0.00000343, 0.00000059, 0.00000477, 0.00000828, 0.00001044, 0.00001089, 0.00000959, 0.00000682, 0.00000305, -0.00000112, -0.00000506, -0.00000815, -0.00000985, -0.00000979, -0.00000780, -0.00000406, 0.00000088, 0.00000612, 0.00001055, 0.00001308, 0.00001296, 0.00001001, 0.00000479, -0.00000152, -0.00000738, -0.00001127, -0.00001222, -0.00001005, -0.00000551, -0.00000006, 0.00000455, 0.00000682, 0.00000600, 0.00000233, -0.00000304, -0.00000840, -0.00001205, -0.00001281, -0.00001048, -0.00000580, -0.00000026, 0.00000445, 0.00000698, 0.00000679, 0.00000431, 0.00000072, -0.00000245, -0.00000389, -0.00000298, 0.00000003, 0.00000405, 0.00000762, 0.00000935, 0.00000848, 0.00000508, 0.00000008, -0.00000508, -0.00000892, -0.00001040, -0.00000922, -0.00000591, -0.00000158, 0.00000242, 0.00000499, 0.00000553, 0.00000412, 0.00000135, -0.00000188, -0.00000476, -0.00000672, -0.00000761, -0.00000758, -0.00000694, -0.00000594, -0.00000466, -0.00000299, -0.00000074, 0.00000216, 0.00000556, 0.00000899, 0.00001179, 0.00001324, 0.00001286, 0.00001056, 0.00000670, 0.00000204, -0.00000252, -0.00000615, -0.00000836, -0.00000905, -0.00000852, },
		{ 0.00000089, 0.00000239, 0.00000395, 0.00000540, 0.00000642, 0.00000667, 0.00000590, 0.00000412, 0.00000160, -0.00000113, -0.00000350, -0.00000499, -0.00000534, -0.00000459, -0.00000303, -0.00000113, 0.00000070, 0.00000213, 0.00000304, 0.00000345, 0.00000349, 0.00000324, 0.00000272, 0.00000189, 0.00000069, -0.00000083, -0.00000249, -0.00000391, -0.00000464, -0.00000427, -0.00000260, 0.00000024, 0.00000377, 0.00000726, 0.00000987, 0.00001094, 0.00001012, 0.00000749, 0.00000353, -0.00000096, -0.00000513, -0.00000822, -0.00000974, -0.00000951, -0.00000769, -0.00000467, -0.00000097, 0.00000282, 0.00000609, 0.00000835, 0.00000920, 0.00000842, 0.00000608, 0.00000255, -0.00000150, -0.00000522, -0.00000770, -0.00000830, -0.00000679, -0.00000353, 0.00000056, 0.00000425, 0.00000633, 0.00000598, 0.00000306, -0.00000174, -0.00000712, -0.00001150, -0.00001350, -0.00001240, -0.00000840, -0.00000255, 0.00000351, 0.00000814, 0.00001013, 0.00000916, 0.00000585, 0.00000155, -0.00000211, -0.00000383, -0.00000303, -0.00000007, 0.00000385, 0.00000717, 0.00000847, 0.00000699, 0.00000290, -0.00000273, -0.00000827, -0.00001208, -0.00001300, -0.00001077, -0.00000602, -0.00000015, 0.00000524, 0.00000872, 0.00000952, 0.00000768, 0.00000391, -0.00000065, -0.00000484, -0.00000785, -0.00000933, -0.00000942, -0.00000854, -0.00000716, -0.00000556, -0.00000380, -0.00000170, 0.00000091, 0.00000406, 0.00000747, 0.00001060, 0.00001274, 0.00001328, 0.00001191, 0.00000875, 0.00000438, -0.00000036, -0.00000456, -0.00000751, -0.00000888, -0.00000876, -0.00000754, -0.00000575, -0.00000384, -0.00000210, -0.00000056, },
		{ 0.00000440, 0.00000361, 0.00000211, 0.00000014, -0.00000189, -0.00000345, -0.00000415, -0.00000378, -0.00000245, -0.00000054, 0.00000141, 0.00000288, 0.00000354, 0.00000328, 0.00000230, 0.00000094, -0.00000043, -0.00000148, -0.00000208, -0.00000221, -0.00000195, -0.00000138, -0.00000053, 0.00000065, 0.00000221, 0.00000411, 0.00000618, 0.00000804, 0.00000921, 0.00000922, 0.00000779, 0.00000492, 0.00000100, -0.00000327, -0.00000704, -0.00000955, -0.00001029, -0.00000915, -0.00000645, -0.00000284, 0.00000090, 0.00000402, 0.00000599, 0.00000659, 0.00000587, 0.00000415, 0.00000185, -0.00000053, -0.00000257, -0.00000390, -0.00000428, -0.00000361, -0.00000199, 0.00000028, 0.00000269, 0.00000458, 0.00000530, 0.00000438, 0.00000173, -0.00000226, -0.00000675, -0.00001059, -0.00001266, -0.00001215, -0.00000893, -0.00000359, 0.00000264, 0.00000821, 0.00001171, 0.00001228, 0.00000993, 0.00000550, 0.00000047, -0.00000355, -0.00000534, -0.00000446, -0.00000139, 0.00000260, 0.00000590, 0.00000711, 0.00000550, 0.00000131, -0.00000432, -0.00000970, -0.00001311, -0.00001342, -0.00001035, -0.00000468, 0.00000208, 0.00000813, 0.00001189, 0.00001251, 0.00001000, 0.00000521, -0.00000054, -0.00000585, -0.00000968, -0.00001159, -0.00001169, -0.00001051, -0.00000869, -0.00000672, -0.00000474, -0.00000262, -0.00000008, 0.00000306, 0.00000669, 0.00001036, 0.00001336, 0.00001493, 0.00001452, 0.00001203, 0.00000785, 0.00000276, -0.00000223, -0.00000621, -0.00000860, -0.00000925, -0.00000840, -0.00000659, -0.00000440, -0.00000228, -0.00000049, 0.00000096, 0.00000215, 0.00000317, 0.00000399, 0.00000447, },
		{ -0.00000164, -0.00000220, -0.00000201, -0.00000114, 0.00000010, 0.00000126, 0.00000190, 0.00000177, 0.00000086, -0.00000055, -0.00000200, -0.00000301, -0.00000320, -0.00000245, -0.00000088, 0.00000119, 0.00000338, 0.00000532, 0.00000681, 0.00000775, 0.00000815, 0.00000804, 0.00000740, 0.00000617, 0.00000428, 0.00000173, -0.00000131, -0.00000447, -0.00000722, -0.00000900, -0.00000938, -0.00000821, -0.00000569, -0.00000237, 0.00000098, 0.00000360, 0.00000494, 0.00000478, 0.00000333, 0.00000110, -0.00000122, -0.00000297, -0.00000369, -0.00000324, -0.00000176, 0.00000037, 0.00000263, 0.00000454, 0.00000567, 0.00000575, 0.00000467, 0.00000250, -0.00000053, -0.00000401, -0.00000737, -0.00000994, -0.00001106, -0.00001028, -0.00000745, -0.00000290, 0.00000256, 0.00000780, 0.00001162, 0.00001306, 0.00001175, 0.00000801, 0.00000283, -0.00000235, -0.00000610, -0.00000743, -0.00000609, -0.00000268, 0.00000152, 0.00000493, 0.00000623, 0.00000478, 0.00000085, -0.00000443, -0.00000940, -0.00001239, -0.00001229, -0.00000888, -0.00000294, 0.00000398, 0.00001005, 0.00001366, 0.00001391, 0.00001083, 0.00000531, -0.00000122, -0.00000724, -0.00001161, -0.00001377, -0.00001384, -0.00001241, -0.00001021, -0.00000783, -0.00000551, -0.00000313, -0.00000034, 0.00000313, 0.00000728, 0.00001168, 0.00001557, 0.00001804, 0.00001833, 0.00001612, 0.00001166, 0.00000574, -0.00000053, -0.00000600, -0.00000979, -0.00001147, -0.00001117, -0.00000937, -0.00000680, -0.00000411, -0.00000176, 0.00000005, 0.00000136, 0.00000230, 0.00000297, 0.00000339, 0.00000346, 0.00000307, 0.00000218, 0.00000090, -0.00000050, },
		{ -0.00000106, -0.00000073, -0.00000081, -0.00000129, -0.00000195, -0.00000239, -0.00000222, -0.00000116, 0.00000080, 0.00000339, 0.00000614, 0.00000849, 0.00000998, 0.00001035, 0.00000958, 0.00000789, 0.00000563, 0.00000313, 0.00000065, -0.00000164, -0.00000368, -0.00000540, -0.00000672, -0.00000746, -0.00000744, -0.00000656, -0.00000487, -0.00000261, -0.00000026, 0.00000166, 0.00000265, 0.00000248, 0.00000124, -0.00000067, -0.00000258, -0.00000383, -0.00000388, -0.00000258, -0.00000013, 0.00000289, 0.00000575, 0.00000771, 0.00000828, 0.00000726, 0.00000483, 0.00000145, -0.00000226, -0.00000567, -0.00000822, -0.00000953, -0.00000939, -0.00000777, -0.00000484, -0.00000095, 0.00000336, 0.00000740, 0.00001042, 0.00001176, 0.00001103, 0.00000824, 0.00000390, -0.00000104, -0.00000543, -0.00000817, -0.00000859, -0.00000664, -0.00000296, 0.00000125, 0.00000460, 0.00000594, 0.00000472, 0.00000122, -0.00000353, -0.00000799, -0.00001064, -0.00001044, -0.00000718, -0.00000157, 0.00000494, 0.00001060, 0.00001385, 0.00001380, 0.00001046, 0.00000465, -0.00000219, -0.00000854, -0.00001319, -0.00001552, -0.00001559, -0.00001394, -0.00001135, -0.00000846, -0.00000555, -0.00000255, 0.00000085, 0.00000493, 0.00000968, 0.00001467, 0.00001906, 0.00002183, 0.00002209, 0.00001940, 0.00001397, 0.00000666, -0.00000124, -0.00000831, -0.00001338, -0.00001581, -0.00001559, -0.00001326, -0.00000969, -0.00000582, -0.00000236, 0.00000028, 0.00000205, 0.00000311, 0.00000371, 0.00000400, 0.00000404, 0.00000377, 0.00000311, 0.00000206, 0.00000076, -0.00000054, -0.00000155, -0.00000207, -0.00000203, -0.00000160, },
		{ 0.00000036, 0.00000150, 0.00000318, 0.00000534, 0.00000770, 0.00000980, 0.00001111, 0.00001125, 0.00001004, 0.00000765, 0.00000448, 0.00000108, -0.00000201, -0.00000438, -0.00000587, -0.00000653, -0.00000654, -0.00000613, -0.00000549, -0.00000469, -0.00000377, -0.00000276, -0.00000173, -0.00000084, -0.00000030, -0.00000028, -0.00000081, -0.00000173, -0.00000268, -0.00000319, -0.00000283, -0.00000138, 0.00000107, 0.00000408, 0.00000697, 0.00000897, 0.00000949, 0.00000826, 0.00000545, 0.00000160, -0.00000248, -0.00000597, -0.00000820, -0.00000881, -0.00000780, -0.00000546, -0.00000228, 0.00000119, 0.00000444, 0.00000705, 0.00000870, 0.00000918, 0.00000837, 0.00000631, 0.00000321, -0.00000047, -0.00000411, -0.00000697, -0.00000838, -0.00000796, -0.00000576, -0.00000232, 0.00000144, 0.00000441, 0.00000567, 0.00000476, 0.00000188, -0.00000212, -0.00000598, -0.00000837, -0.00000835, -0.00000566, -0.00000087, 0.00000481, 0.00000979, 0.00001264, 0.00001249, 0.00000924, 0.00000362, -0.00000309, -0.00000941, -0.00001411, -0.00001649, -0.00001649, -0.00001459, -0.00001149, -0.00000787, -0.00000413, -0.00000030, 0.00000380, 0.00000836, 0.00001333, 0.00001823, 0.00002223, 0.00002433, 0.00002364, 0.00001977, 0.00001299, 0.00000429, -0.00000486, -0.00001285, -0.00001833, -0.00002055, -0.00001949, -0.00001583, -0.00001070, -0.00000529, -0.00000063, 0.00000271, 0.00000462, 0.00000535, 0.00000530, 0.00000488, 0.00000432, 0.00000365, 0.00000280, 0.00000168, 0.00000029, -0.00000122, -0.00000260, -0.00000358, -0.00000399, -0.00000382, -0.00000322, -0.00000241, -0.00000164, -0.00000099, -0.00000039, },
		{ 0.00001051, 0.00001040, 0.00000921, 0.00000691, 0.00000373, 0.00000015, -0.00000323, -0.00000586, -0.00000737, -0.00000770, -0.00000705, -0.00000582, -0.00000447, -0.00000336, -0.00000269, -0.00000246, -0.00000252, -0.00000269, -0.00000284, -0.00000288, -0.00000282, -0.00000268, -0.00000243, -0.00000194, -0.00000107, 0.00000033, 0.00000227, 0.00000456, 0.00000680, 0.00000846, 0.00000902, 0.00000817, 0.00000590, 0.00000258, -0.00000112, -0.00000439, -0.00000650, -0.00000704, -0.00000594, -0.00000356, -0.00000051, 0.00000251, 0.00000491, 0.00000633, 0.00000668, 0.00000609, 0.00000480, 0.00000307, 0.00000109, -0.00000098, -0.00000303, -0.00000483, -0.00000613, -0.00000659, -0.00000599, -0.00000427, -0.00000169, 0.00000120, 0.00000366, 0.00000496, 0.00000465, 0.00000273, -0.00000028, -0.00000345, -0.00000570, -0.00000618, -0.00000450, -0.00000096, 0.00000352, 0.00000767, 0.00001020, 0.00001023, 0.00000752, 0.00000258, -0.00000349, -0.00000936, -0.00001380, -0.00001604, -0.00001586, -0.00001361, -0.00000996, -0.00000563, -0.00000114, 0.00000328, 0.00000764, 0.00001201, 0.00001629, 0.00002011, 0.00002276, 0.00002341, 0.00002135, 0.00001633, 0.00000871, -0.00000044, -0.00000962, -0.00001719, -0.00002181, -0.00002280, -0.00002026, -0.00001507, -0.00000856, -0.00000217, 0.00000292, 0.00000607, 0.00000721, 0.00000678, 0.00000545, 0.00000387, 0.00000246, 0.00000139, 0.00000053, -0.00000033, -0.00000133, -0.00000250, -0.00000366, -0.00000452, -0.00000482, -0.00000441, -0.00000332, -0.00000176, -0.00000000, 0.00000171, 0.00000325, 0.00000463, 0.00000594, 0.00000728, 0.00000862, 0.00000980, },
		{ -0.00000342, -0.00000606, -0.00000781, -0.00000840, -0.00000784, -0.00000644, -0.00000469, -0.00000312, -0.00000214, -0.00000193, -0.00000238, -0.00000321, -0.00000402, -0.00000447, -0.00000436, -0.00000368, -0.00000254, -0.00000112, 0.00000040, 0.00000194, 0.00000345, 0.00000489, 0.00000617, 0.00000712, 0.00000749, 0.00000707, 0.00000575, 0.00000364, 0.00000109, -0.00000137, -0.00000318, -0.00000391, -0.00000337, -0.00000176, 0.00000045, 0.00000264, 0.00000419, 0.00000474, 0.00000421, 0.00000288, 0.00000117, -0.00000044, -0.00000164, -0.00000233, -0.00000265, -0.00000284, -0.00000313, -0.00000354, -0.00000393, -0.00000397, -0.00000336, -0.00000197, 0.00000006, 0.00000227, 0.00000403, 0.00000475, 0.00000410, 0.00000220, -0.00000041, -0.00000286, -0.00000429, -0.00000409, -0.00000217, 0.00000095, 0.00000433, 0.00000682, 0.00000751, 0.00000593, 0.00000230, -0.00000259, -0.00000758, -0.00001151, -0.00001350, -0.00001320, -0.00001083, -0.00000698, -0.00000241, 0.00000222, 0.00000648, 0.00001025, 0.00001352, 0.00001629, 0.00001836, 0.00001932, 0.00001861, 0.00001575, 0.00001060, 0.00000352, -0.00000454, -0.00001225, -0.00001818, -0.00002121, -0.00002079, -0.00001718, -0.00001135, -0.00000468, 0.00000135, 0.00000559, 0.00000749, 0.00000717, 0.00000527, 0.00000270, 0.00000031, -0.00000134, -0.00000210, -0.00000220, -0.00000203, -0.00000200, -0.00000228, -0.00000281, -0.00000330, -0.00000339, -0.00000279, -0.00000141, 0.00000060, 0.00000291, 0.00000515, 0.00000700, 0.00000832, 0.00000911, 0.00000946, 0.00000948, 0.00000918, 0.00000849, 0.00000724, 0.00000531, 0.00000272, -0.00000032, },
		{ -0.00000557, -0.00000381, -0.00000239, -0.00000167, -0.00000178, -0.00000256, -0.00000365, -0.00000456, -0.00000488, -0.00000438, -0.00000310, -0.00000126, 0.00000078, 0.00000267, 0.00000417, 0.00000519, 0.00000576, 0.00000598, 0.00000594, 0.00000565, 0.00000510, 0.00000426, 0.00000317, 0.00000196, 0.00000086, 0.00000015, -0.00000002, 0.00000039, 0.00000119, 0.00000201, 0.00000245, 0.00000220, 0.00000117, -0.00000044, -0.00000220, -0.00000361, -0.00000425, -0.00000397, -0.00000295, -0.00000164, -0.00000059, -0.00000024, -0.00000072, -0.00000182, -0.00000299, -0.00000361, -0.00000317, -0.00000153, 0.00000097, 0.00000365, 0.00000564, 0.00000625, 0.00000517, 0.00000267, -0.00000051, -0.00000335, -0.00000493, -0.00000471, -0.00000275, 0.00000028, 0.00000335, 0.00000537, 0.00000560, 0.00000383, 0.00000050, -0.00000347, -0.00000696, -0.00000904, -0.00000916, -0.00000735, -0.00000411, -0.00000017, 0.00000371, 0.00000700, 0.00000950, 0.00001125, 0.00001239, 0.00001300, 0.00001297, 0.00001199, 0.00000972, 0.00000598, 0.00000094, -0.00000481, -0.00001030, -0.00001446, -0.00001638, -0.00001563, -0.00001243, -0.00000761, -0.00000239, 0.00000199, 0.00000458, 0.00000500, 0.00000351, 0.00000088, -0.00000190, -0.00000395, -0.00000475, -0.00000428, -0.00000295, -0.00000136, -0.00000013, 0.00000041, 0.00000028, -0.00000018, -0.00000046, -0.00000012, 0.00000108, 0.00000306, 0.00000544, 0.00000773, 0.00000947, 0.00001035, 0.00001030, 0.00000949, 0.00000815, 0.00000654, 0.00000479, 0.00000292, 0.00000091, -0.00000128, -0.00000355, -0.00000567, -0.00000733, -0.00000821, -0.00000814, -0.00000717, },
		{ -0.00000241, -0.00000341, -0.00000418, -0.00000431, -0.00000358, -0.00000204, 0.00000001, 0.00000215, 0.00000395, 0.00000513, 0.00000562, 0.00000555, 0.00000518, 0.00000476, 0.00000449, 0.00000442, 0.00000446, 0.00000449, 0.00000439, 0.00000409, 0.00000361, 0.00000301, 0.00000232, 0.00000151, 0.00000053, -0.00000072, -0.00000223, -0.00000389, -0.00000540, -0.00000638, -0.00000649, -0.00000556, -0.00000370, -0.00000135, 0.00000085, 0.00000223, 0.00000235, 0.00000115, -0.00000093, -0.00000312, -0.00000452, -0.00000443, -0.00000261, 0.00000059, 0.00000429, 0.00000734, 0.00000872, 0.00000784, 0.00000484, 0.00000050, -0.00000393, -0.00000712, -0.00000813, -0.00000667, -0.00000320, 0.00000120, 0.00000519, 0.00000761, 0.00000781, 0.00000587, 0.00000249, -0.00000125, -0.00000424, -0.00000571, -0.00000542, -0.00000366, -0.00000109, 0.00000154, 0.00000367, 0.00000505, 0.00000579, 0.00000620, 0.00000654, 0.00000691, 0.00000712, 0.00000679, 0.00000551, 0.00000308, -0.00000035, -0.00000423, -0.00000777, -0.00001017, -0.00001086, -0.00000979, -0.00000739, -0.00000446, -0.00000194, -0.00000057, -0.00000064, -0.00000192, -0.00000377, -0.00000535, -0.00000593, -0.00000517, -0.00000319, -0.00000054, 0.00000203, 0.00000385, 0.00000454, 0.00000419, 0.00000323, 0.00000232, 0.00000208, 0.00000283, 0.00000454, 0.00000678, 0.00000893, 0.00001036, 0.00001067, 0.00000975, 0.00000782, 0.00000531, 0.00000268, 0.00000030, -0.00000167, -0.00000323, -0.00000452, -0.00000563, -0.00000656, -0.00000722, -0.00000742, -0.00000703, -0.00000605, -0.00000465, -0.00000316, -0.00000199, -0.00000145, -0.00000163, },
		{ -0.00000116, 0.00000077, 0.00000274, 0.00000435, 0.00000531, 0.00000556, 0.00000529, 0.00000480, 0.00000447, 0.00000452, 0.00000501, 0.00000576, 0.00000649, 0.00000686, 0.00000662, 0.00000565, 0.00000402, 0.00000188, -0.00000054, -0.00000301, -0.00000530, -0.00000721, -0.00000855, -0.00000909, -0.00000866, -0.00000718, -0.00000478, -0.00000181, 0.00000115, 0.00000344, 0.00000447, 0.00000393, 0.00000196, -0.00000083, -0.00000351, -0.00000508, -0.00000482, -0.00000254, 0.00000128, 0.00000558, 0.00000904, 0.00001045, 0.00000916, 0.00000531, -0.00000016, -0.00000573, -0.00000979, -0.00001110, -0.00000919, -0.00000453, 0.00000166, 0.00000769, 0.00001200, 0.00001354, 0.00001210, 0.00000828, 0.00000326, -0.00000162, -0.00000524, -0.00000702, -0.00000703, -0.00000578, -0.00000399, -0.00000229, -0.00000097, 0.00000005, 0.00000110, 0.00000253, 0.00000446, 0.00000667, 0.00000859, 0.00000954, 0.00000892, 0.00000653, 0.00000262, -0.00000207, -0.00000659, -0.00001005, -0.00001188, -0.00001200, -0.00001082, -0.00000902, -0.00000730, -0.00000614, -0.00000560, -0.00000538, -0.00000496, -0.00000386, -0.00000185, 0.00000090, 0.00000389, 0.00000643, 0.00000794, 0.00000813, 0.00000714, 0.00000552, 0.00000399, 0.00000324, 0.00000362, 0.00000506, 0.00000707, 0.00000892, 0.00000989, 0.00000950, 0.00000768, 0.00000476, 0.00000135, -0.00000187, -0.00000437, -0.00000591, -0.00000653, -0.00000652, -0.00000624, -0.00000594, -0.00000573, -0.00000554, -0.00000521, -0.00000460, -0.00000372, -0.00000272, -0.00000184, -0.00000137, -0.00000144, -0.00000198, -0.00000273, -0.00000331, -0.00000334, -0.00000262, },
		{ 0.00000558, 0.00000542, 0.00000510, 0.00000498, 0.00000529, 0.00000606, 0.00000708, 0.00000795, 0.00000823, 0.00000758, 0.00000585, 0.00000313, -0.00000025, -0.00000384, -0.00000715, -0.00000978, -0.00001141, -0.00001187, -0.00001111, -0.00000921, -0.00000637, -0.00000295, 0.00000056, 0.00000357, 0.00000551, 0.00000597, 0.00000483, 0.00000240, -0.00000062, -0.00000330, -0.00000468, -0.00000415, -0.00000163, 0.00000230, 0.00000652, 0.00000968, 0.00001060, 0.00000870, 0.00000423, -0.00000171, -0.00000749, -0.00001135, -0.00001200, -0.00000900, -0.00000292, 0.00000475, 0.00001207, 0.00001716, 0.00001877, 0.00001658, 0.00001124, 0.00000415, -0.00000302, -0.00000879, -0.00001228, -0.00001333, -0.00001238, -0.00001023, -0.00000763, -0.00000505, -0.00000256, 0.00000007, 0.00000311, 0.00000665, 0.00001037, 0.00001355, 0.00001531, 0.00001483, 0.00001174, 0.00000627, -0.00000072, -0.00000796, -0.00001410, -0.00001810, -0.00001946, -0.00001835, -0.00001547, -0.00001174, -0.00000803, -0.00000489, -0.00000245, -0.00000048, 0.00000138, 0.00000343, 0.00000568, 0.00000788, 0.00000958, 0.00001033, 0.00000994, 0.00000854, 0.00000662, 0.00000486, 0.00000384, 0.00000391, 0.00000494, 0.00000645, 0.00000767, 0.00000791, 0.00000672, 0.00000411, 0.00000054, -0.00000323, -0.00000641, -0.00000837, -0.00000889, -0.00000815, -0.00000665, -0.00000500, -0.00000371, -0.00000304, -0.00000293, -0.00000314, -0.00000332, -0.00000324, -0.00000282, -0.00000220, -0.00000162, -0.00000133, -0.00000141, -0.00000176, -0.00000211, -0.00000214, -0.00000161, -0.00000046, 0.00000114, 0.00000287, 0.00000434, 0.00000527, },
		{ 0.00000687, 0.00000773, 0.00000833, 0.00000823, 0.00000703, 0.00000460, 0.00000109, -0.00000307, -0.00000728, -0.00001088, -0.00001336, -0.00001437, -0.00001381, -0.00001181, -0.00000867, -0.00000482, -0.00000078, 0.00000291, 0.00000572, 0.00000723, 0.00000721, 0.00000572, 0.00000314, 0.00000017, -0.00000233, -0.00000357, -0.00000304, -0.00000075, 0.00000272, 0.00000632, 0.00000883, 0.00000925, 0.00000711, 0.00000276, -0.00000271, -0.00000769, -0.00001055, -0.00001011, -0.00000610, 0.00000075, 0.00000881, 0.00001601, 0.00002043, 0.00002082, 0.00001696, 0.00000970, 0.00000069, -0.00000810, -0.00001496, -0.00001882, -0.00001946, -0.00001741, -0.00001364, -0.00000917, -0.00000471, -0.00000056, 0.00000338, 0.00000731, 0.00001131, 0.00001505, 0.00001785, 0.00001878, 0.00001705, 0.00001230, 0.00000488, -0.00000413, -0.00001316, -0.00002055, -0.00002496, -0.00002572, -0.00002305, -0.00001784, -0.00001143, -0.00000515, 0.00000000, 0.00000361, 0.00000578, 0.00000700, 0.00000784, 0.00000869, 0.00000963, 0.00001045, 0.00001078, 0.00001037, 0.00000916, 0.00000743, 0.00000566, 0.00000433, 0.00000376, 0.00000392, 0.00000441, 0.00000465, 0.00000405, 0.00000227, -0.00000060, -0.00000406, -0.00000733, -0.00000963, -0.00001041, -0.00000954, -0.00000738, -0.00000464, -0.00000212, -0.00000047, 0.00000001, -0.00000057, -0.00000176, -0.00000297, -0.00000371, -0.00000372, -0.00000306, -0.00000204, -0.00000106, -0.00000043, -0.00000026, -0.00000042, -0.00000061, -0.00000050, 0.00000011, 0.00000122, 0.00000263, 0.00000402, 0.00000508, 0.00000566, 0.00000581, 0.00000576, 0.00000580, 0.00000616, },
		{ 0.00000243, -0.00000146, -0.00000591, -0.00001024, -0.00001373, -0.00001579, -0.00001608, -0.00001457, -0.00001152, -0.00000742, -0.00000285, 0.00000157, 0.00000527, 0.00000783, 0.00000895, 0.00000857, 0.00000684, 0.00000421, 0.00000130, -0.00000115, -0.00000247, -0.00000227, -0.00000055, 0.00000218, 0.00000506, 0.00000708, 0.00000740, 0.00000566, 0.00000217, -0.00000210, -0.00000579, -0.00000752, -0.00000635, -0.00000217, 0.00000423, 0.00001126, 0.00001700, 0.00001970, 0.00001832, 0.00001287, 0.00000439, -0.00000530, -0.00001416, -0.00002041, -0.00002305, -0.00002199, -0.00001797, -0.00001223, -0.00000604, -0.00000034, 0.00000444, 0.00000835, 0.00001165, 0.00001455, 0.00001689, 0.00001815, 0.00001755, 0.00001439, 0.00000843, 0.00000010, -0.00000944, -0.00001850, -0.00002533, -0.00002854, -0.00002756, -0.00002276, -0.00001534, -0.00000697, 0.00000068, 0.00000639, 0.00000968, 0.00001077, 0.00001045, 0.00000966, 0.00000914, 0.00000925, 0.00000985, 0.00001050, 0.00001066, 0.00000998, 0.00000845, 0.00000636, 0.00000417, 0.00000235, 0.00000109, 0.00000030, -0.00000038, -0.00000136, -0.00000293, -0.00000506, -0.00000737, -0.00000926, -0.00001009, -0.00000949, -0.00000746, -0.00000447, -0.00000131, 0.00000118, 0.00000236, 0.00000201, 0.00000040, -0.00000182, -0.00000385, -0.00000500, -0.00000493, -0.00000374, -0.00000188, 0.00000003, 0.00000141, 0.00000197, 0.00000175, 0.00000107, 0.00000043, 0.00000025, 0.00000073, 0.00000183, 0.00000326, 0.00000466, 0.00000572, 0.00000634, 0.00000658, 0.00000667, 0.00000684, 0.00000716, 0.00000752, 0.00000759, 0.00000697, 0.00000529, },
		{ -0.00001683, -0.00001645, -0.00001418, -0.00001036, -0.00000558, -0.00000055, 0.00000405, 0.00000766, 0.00000990, 0.00001063, 0.00000988, 0.00000792, 0.00000518, 0.00000224, -0.00000027, -0.00000181, -0.00000202, -0.00000091, 0.00000117, 0.00000355, 0.00000541, 0.00000606, 0.00000514, 0.00000284, -0.00000013, -0.00000272, -0.00000386, -0.00000282, 0.00000047, 0.00000531, 0.00001037, 0.00001405, 0.00001496, 0.00001235, 0.00000640, -0.00000178, -0.00001044, -0.00001768, -0.00002195, -0.00002249, -0.00001946, -0.00001384, -0.00000703, -0.00000044, 0.00000494, 0.00000874, 0.00001113, 0.00001259, 0.00001355, 0.00001412, 0.00001398, 0.00001249, 0.00000902, 0.00000331, -0.00000429, -0.00001270, -0.00002035, -0.00002560, -0.00002718, -0.00002464, -0.00001850, -0.00001012, -0.00000130, 0.00000620, 0.00001118, 0.00001326, 0.00001290, 0.00001116, 0.00000926, 0.00000815, 0.00000824, 0.00000929, 0.00001062, 0.00001136, 0.00001087, 0.00000894, 0.00000583, 0.00000219, -0.00000126, -0.00000393, -0.00000560, -0.00000638, -0.00000668, -0.00000688, -0.00000722, -0.00000763, -0.00000781, -0.00000735, -0.00000598, -0.00000374, -0.00000103, 0.00000149, 0.00000310, 0.00000330, 0.00000203, -0.00000034, -0.00000304, -0.00000517, -0.00000603, -0.00000532, -0.00000325, -0.00000047, 0.00000220, 0.00000398, 0.00000447, 0.00000370, 0.00000215, 0.00000051, -0.00000056, -0.00000061, 0.00000039, 0.00000217, 0.00000420, 0.00000598, 0.00000714, 0.00000761, 0.00000754, 0.00000717, 0.00000675, 0.00000630, 0.00000568, 0.00000456, 0.00000263, -0.00000028, -0.00000405, -0.00000826, -0.00001226, -0.00001532, },
		{ 0.00000172, 0.00000628, 0.00000968, 0.00001157, 0.00001190, 0.00001082, 0.00000863, 0.00000581, 0.00000286, 0.00000032, -0.00000136, -0.00000189, -0.00000123, 0.00000039, 0.00000245, 0.00000430, 0.00000533, 0.00000518, 0.00000385, 0.00000179, -0.00000027, -0.00000150, -0.00000129, 0.00000049, 0.00000338, 0.00000643, 0.00000845, 0.00000842, 0.00000581, 0.00000082, -0.00000556, -0.00001190, -0.00001666, -0.00001866, -0.00001745, -0.00001341, -0.00000759, -0.00000140, 0.00000389, 0.00000747, 0.00000918, 0.00000945, 0.00000900, 0.00000849, 0.00000821, 0.00000793, 0.00000702, 0.00000474, 0.00000067, -0.00000506, -0.00001161, -0.00001761, -0.00002156, -0.00002230, -0.00001939, -0.00001333, -0.00000542, 0.00000262, 0.00000913, 0.00001301, 0.00001403, 0.00001279, 0.00001045, 0.00000832, 0.00000734, 0.00000781, 0.00000932, 0.00001093, 0.00001160, 0.00001053, 0.00000752, 0.00000298, -0.00000214, -0.00000677, -0.00001001, -0.00001141, -0.00001108, -0.00000956, -0.00000756, -0.00000573, -0.00000437, -0.00000347, -0.00000272, -0.00000182, -0.00000060, 0.00000082, 0.00000204, 0.00000259, 0.00000210, 0.00000054, -0.00000171, -0.00000397, -0.00000543, -0.00000551, -0.00000402, -0.00000128, 0.00000195, 0.00000472, 0.00000623, 0.00000609, 0.00000442, 0.00000188, -0.00000065, -0.00000227, -0.00000243, -0.00000107, 0.00000141, 0.00000429, 0.00000679, 0.00000834, 0.00000870, 0.00000803, 0.00000671, 0.00000520, 0.00000378, 0.00000250, 0.00000115, -0.00000057, -0.00000294, -0.00000599, -0.00000944, -0.00001273, -0.00001520, -0.00001620, -0.00001537, -0.00001269, -0.00000851, -0.00000345, },
		{ 0.00001140, 0.00000914, 0.00000634, 0.00000348, 0.00000100, -0.00000075, -0.00000152, -0.00000125, -0.00000011, 0.00000154, 0.00000320, 0.00000435, 0.00000460, 0.00000382, 0.00000223, 0.00000033, -0.00000124, -0.00000193, -0.00000150, -0.00000012, 0.00000164, 0.00000295, 0.00000307, 0.00000155, -0.00000151, -0.00000546, -0.00000927, -0.00001187, -0.00001245, -0.00001078, -0.00000727, -0.00000287, 0.00000130, 0.00000427, 0.00000556, 0.00000529, 0.00000411, 0.00000289, 0.00000233, 0.00000272, 0.00000372, 0.00000455, 0.00000428, 0.00000222, -0.00000171, -0.00000690, -0.00001210, -0.00001588, -0.00001701, -0.00001494, -0.00000997, -0.00000321, 0.00000376, 0.00000942, 0.00001274, 0.00001347, 0.00001217, 0.00000996, 0.00000801, 0.00000718, 0.00000767, 0.00000895, 0.00001005, 0.00000989, 0.00000778, 0.00000363, -0.00000188, -0.00000759, -0.00001220, -0.00001468, -0.00001461, -0.00001228, -0.00000854, -0.00000451, -0.00000119, 0.00000084, 0.00000152, 0.00000123, 0.00000057, 0.00000007, -0.00000004, 0.00000017, 0.00000034, 0.00000011, -0.00000068, -0.00000190, -0.00000309, -0.00000367, -0.00000316, -0.00000143, 0.00000119, 0.00000399, 0.00000613, 0.00000686, 0.00000590, 0.00000350, 0.00000039, -0.00000240, -0.00000395, -0.00000369, -0.00000164, 0.00000161, 0.00000515, 0.00000798, 0.00000937, 0.00000907, 0.00000733, 0.00000476, 0.00000209, -0.00000011, -0.00000161, -0.00000257, -0.00000338, -0.00000452, -0.00000627, -0.00000859, -0.00001107, -0.00001309, -0.00001396, -0.00001321, -0.00001070, -0.00000668, -0.00000176, 0.00000329, 0.00000771, 0.00001091, 0.00001258, 0.00001267, },
		{ -0.00000012, -0.00000125, -0.00000153, -0.00000105, -0.00000001, 0.00000121, 0.00000220, 0.00000259, 0.00000219, 0.00000103, -0.00000058, -0.00000218, -0.00000326, -0.00000350, -0.00000284, -0.00000157, -0.00000022, 0.00000062, 0.00000051, -0.00000065, -0.00000257, -0.00000464, -0.00000617, -0.00000661, -0.00000579, -0.00000398, -0.00000182, -0.00000003, 0.00000079, 0.00000051, -0.00000058, -0.00000176, -0.00000224, -0.00000146, 0.00000063, 0.00000352, 0.00000621, 0.00000757, 0.00000678, 0.00000361, -0.00000135, -0.00000689, -0.00001146, -0.00001370, -0.00001287, -0.00000907, -0.00000323, 0.00000319, 0.00000867, 0.00001210, 0.00001311, 0.00001209, 0.00000999, 0.00000790, 0.00000662, 0.00000637, 0.00000672, 0.00000682, 0.00000577, 0.00000300, -0.00000141, -0.00000669, -0.00001162, -0.00001493, -0.00001570, -0.00001373, -0.00000955, -0.00000433, 0.00000055, 0.00000395, 0.00000525, 0.00000453, 0.00000248, 0.00000004, -0.00000187, -0.00000273, -0.00000252, -0.00000160, -0.00000058, 0.00000006, 0.00000014, -0.00000018, -0.00000046, -0.00000023, 0.00000075, 0.00000238, 0.00000415, 0.00000538, 0.00000549, 0.00000421, 0.00000180, -0.00000105, -0.00000339, -0.00000438, -0.00000354, -0.00000100, 0.00000255, 0.00000606, 0.00000847, 0.00000906, 0.00000766, 0.00000472, 0.00000109, -0.00000225, -0.00000453, -0.00000543, -0.00000518, -0.00000439, -0.00000383, -0.00000409, -0.00000538, -0.00000742, -0.00000957, -0.00001100, -0.00001101, -0.00000924, -0.00000581, -0.00000123, 0.00000371, 0.00000817, 0.00001146, 0.00001321, 0.00001335, 0.00001211, 0.00000988, 0.00000713, 0.00000431, 0.00000180, },
		{ -0.00000188, -0.00000100, -0.00000035, -0.00000017, -0.00000058, -0.00000147, -0.00000255, -0.00000342, -0.00000374, -0.00000331, -0.00000222, -0.00000078, 0.00000055, 0.00000132, 0.00000128, 0.00000044, -0.00000092, -0.00000237, -0.00000349, -0.00000405, -0.00000409, -0.00000388, -0.00000376, -0.00000400, -0.00000455, -0.00000508, -0.00000504, -0.00000386, -0.00000128, 0.00000251, 0.00000677, 0.00001040, 0.00001224, 0.00001149, 0.00000799, 0.00000240, -0.00000398, -0.00000949, -0.00001266, -0.00001262, -0.00000938, -0.00000382, 0.00000259, 0.00000825, 0.00001193, 0.00001309, 0.00001197, 0.00000943, 0.00000655, 0.00000424, 0.00000289, 0.00000228, 0.00000177, 0.00000062, -0.00000168, -0.00000505, -0.00000885, -0.00001204, -0.00001356, -0.00001274, -0.00000955, -0.00000472, 0.00000051, 0.00000477, 0.00000698, 0.00000675, 0.00000442, 0.00000095, -0.00000242, -0.00000460, -0.00000502, -0.00000374, -0.00000139, 0.00000111, 0.00000293, 0.00000363, 0.00000323, 0.00000223, 0.00000125, 0.00000085, 0.00000121, 0.00000209, 0.00000297, 0.00000325, 0.00000257, 0.00000099, -0.00000100, -0.00000265, -0.00000323, -0.00000233, -0.00000009, 0.00000285, 0.00000551, 0.00000693, 0.00000649, 0.00000417, 0.00000056, -0.00000331, -0.00000634, -0.00000772, -0.00000722, -0.00000521, -0.00000261, -0.00000047, 0.00000033, -0.00000059, -0.00000297, -0.00000602, -0.00000867, -0.00000993, -0.00000917, -0.00000632, -0.00000187, 0.00000331, 0.00000822, 0.00001200, 0.00001412, 0.00001445, 0.00001322, 0.00001086, 0.00000791, 0.00000484, 0.00000202, -0.00000029, -0.00000194, -0.00000286, -0.00000306, -0.00000267, },
		{ -0.00000105, -0.00000135, -0.00000181, -0.00000211, -0.00000199, -0.00000134, -0.00000029, 0.00000086, 0.00000173, 0.00000200, 0.00000150, 0.00000030, -0.00000134, -0.00000308, -0.00000461, -0.00000576, -0.00000652, -0.00000696, -0.00000713, -0.00000697, -0.00000626, -0.00000467, -0.00000196, 0.00000188, 0.00000645, 0.00001095, 0.00001434, 0.00001560, 0.00001409, 0.00000983, 0.00000358, -0.00000329, -0.00000913, -0.00001250, -0.00001255, -0.00000935, -0.00000379, 0.00000258, 0.00000813, 0.00001155, 0.00001223, 0.00001040, 0.00000695, 0.00000307, -0.00000015, -0.00000213, -0.00000289, -0.00000294, -0.00000303, -0.00000374, -0.00000522, -0.00000711, -0.00000864, -0.00000901, -0.00000769, -0.00000470, -0.00000068, 0.00000328, 0.00000605, 0.00000680, 0.00000536, 0.00000225, -0.00000144, -0.00000444, -0.00000577, -0.00000499, -0.00000242, 0.00000105, 0.00000427, 0.00000623, 0.00000642, 0.00000498, 0.00000258, 0.00000018, -0.00000138, -0.00000166, -0.00000077, 0.00000072, 0.00000206, 0.00000261, 0.00000212, 0.00000082, -0.00000069, -0.00000169, -0.00000171, -0.00000068, 0.00000097, 0.00000247, 0.00000301, 0.00000210, -0.00000025, -0.00000341, -0.00000639, -0.00000816, -0.00000801, -0.00000584, -0.00000225, 0.00000164, 0.00000457, 0.00000551, 0.00000408, 0.00000063, -0.00000383, -0.00000794, -0.00001040, -0.00001037, -0.00000768, -0.00000289, 0.00000297, 0.00000865, 0.00001305, 0.00001548, 0.00001576, 0.00001416, 0.00001124, 0.00000767, 0.00000404, 0.00000076, -0.00000193, -0.00000393, -0.00000521, -0.00000577, -0.00000566, -0.00000497, -0.00000391, -0.00000274, -0.00000174, -0.00000115, },
		},
		{ { -0.00000136, -0.00000138, -0.00000136, -0.00000117, -0.00000074, -0.00000013, 0.00000049, 0.00000102, 0.00000151, 0.00000214, 0.00000304, 0.00000401, 0.00000452, 0.00000392, 0.00000184, -0.00000143, -0.00000500, -0.00000769, -0.00000871, -0.00000804, -0.00000638, -0.00000470, -0.00000348, -0.00000237, -0.00000045, 0.00000308, 0.00000806, 0.00001308, 0.00001601, 0.00001510, 0.00001014, 0.00000281, -0.00000398, -0.00000747, -0.00000648, -0.00000198, 0.00000349, 0.00000720, 0.00000765, 0.00000516, 0.00000138, -0.00000193, -0.00000396, -0.00000519, -0.00000661, -0.00000865, -0.00001047, -0.00001024, -0.00000639, 0.00000099, 0.00000963, 0.00001590, 0.00001667, 0.00001118, 0.00000175, -0.00000722, -0.00001153, -0.00000942, -0.00000255, 0.00000493, 0.00000878, 0.00000705, 0.00000115, -0.00000507, -0.00000770, -0.00000507, 0.00000115, 0.00000705, 0.00000878, 0.00000493, -0.00000255, -0.00000942, -0.00001153, -0.00000722, 0.00000175, 0.00001118, 0.00001667, 0.00001590, 0.00000963, 0.00000099, -0.00000639, -0.00001024, -0.00001047, -0.00000865, -0.00000661, -0.00000519, -0.00000396, -0.00000193, 0.00000138, 0.00000516, 0.00000765, 0.00000720, 0.00000349, -0.00000198, -0.00000648, -0.00000747, -0.00000398, 0.00000281, 0.00001014, 0.00001510, 0.00001601, 0.00001308, 0.00000806, 0.00000308, -0.00000045, -0.00000237, -0.00000348, -0.00000470, -0.00000638, -0.00000804, -0.00000871, -0.00000769, -0.00000500, -0.00000143, 0.00000184, 0.00000392, 0.00000452, 0.00000401, 0.00000304, 0.00000214, 0.00000151, 0.00000102, 0.00000049, -0.00000013, -0.00000074, -0.00000117, -0.00000136, -0.00000138, },
		{ 0.00000010, 0.00000029, 0.00000050, 0.00000065, 0.00000079, 0.00000111, 0.00000178, 0.00000275, 0.00000353, 0.00000340, 0.00000176, -0.00000146, -0.00000553, -0.00000919, -0.00001118, -0.00001085, -0.00000845, -0.00000484, -0.00000100, 0.00000257, 0.00000595, 0.00000943, 0.00001289, 0.00001547, 0.00001581, 0.00001279, 0.00000654, -0.00000125, -0.00000782, -0.00001059, -0.00000849, -0.00000265, 0.00000412, 0.00000869, 0.00000922, 0.00000598, 0.00000099, -0.00000328, -0.00000532, -0.00000533, -0.00000472, -0.00000489, -0.00000605, -0.00000696, -0.00000576, -0.00000143, 0.00000515, 0.00001132, 0.00001399, 0.00001149, 0.00000474, -0.00000303, -0.00000797, -0.00000779, -0.00000308, 0.00000301, 0.00000669, 0.00000572, 0.00000085, -0.00000469, -0.00000715, -0.00000465, 0.00000155, 0.00000768, 0.00000968, 0.00000570, -0.00000263, -0.00001089, -0.00001429, -0.00001032, -0.00000028, 0.00001136, 0.00001927, 0.00001990, 0.00001324, 0.00000247, -0.00000780, -0.00001397, -0.00001487, -0.00001187, -0.00000751, -0.00000388, -0.00000166, -0.00000023, 0.00000132, 0.00000324, 0.00000480, 0.00000489, 0.00000300, -0.00000007, -0.00000247, -0.00000229, 0.00000118, 0.00000687, 0.00001231, 0.00001491, 0.00001335, 0.00000822, 0.00000173, -0.00000358, -0.00000608, -0.00000579, -0.00000413, -0.00000287, -0.00000310, -0.00000463, -0.00000628, -0.00000667, -0.00000508, -0.00000182, 0.00000195, 0.00000497, 0.00000640, 0.00000618, 0.00000484, 0.00000307, 0.00000141, 0.00000005, -0.00000102, -0.00000182, -0.00000229, -0.00000233, -0.00000195, -0.00000131, -0.00000068, -0.00000026, -0.00000007, -0.00000000, },
		{ 0.00000017, 0.00000077, 0.00000155, 0.00000245, 0.00000308, 0.00000281, 0.00000104, -0.00000236, -0.00000673, -0.00001076, -0.00001292, -0.00001222, -0.00000864, -0.00000314, 0.00000284, 0.00000804, 0.00001185, 0.00001426, 0.00001545, 0.00001531, 0.00001333, 0.00000906, 0.00000272, -0.00000431, -0.00000983, -0.00001170, -0.00000903, -0.00000287, 0.00000412, 0.00000886, 0.00000946, 0.00000607, 0.00000081, -0.00000360, -0.00000529, -0.00000430, -0.00000230, -0.00000135, -0.00000242, -0.00000467, -0.00000596, -0.00000431, 0.00000061, 0.00000700, 0.00001178, 0.00001236, 0.00000828, 0.00000157, -0.00000431, -0.00000649, -0.00000433, 0.00000022, 0.00000390, 0.00000421, 0.00000100, -0.00000339, -0.00000567, -0.00000379, 0.00000159, 0.00000723, 0.00000921, 0.00000542, -0.00000293, -0.00001163, -0.00001563, -0.00001192, -0.00000134, 0.00001162, 0.00002104, 0.00002253, 0.00001541, 0.00000293, -0.00000958, -0.00001737, -0.00001842, -0.00001399, -0.00000736, -0.00000185, 0.00000083, 0.00000105, 0.00000039, 0.00000031, 0.00000126, 0.00000263, 0.00000351, 0.00000360, 0.00000353, 0.00000434, 0.00000659, 0.00000969, 0.00001198, 0.00001167, 0.00000794, 0.00000167, -0.00000489, -0.00000922, -0.00000987, -0.00000720, -0.00000314, -0.00000008, 0.00000046, -0.00000146, -0.00000429, -0.00000600, -0.00000521, -0.00000189, 0.00000273, 0.00000683, 0.00000895, 0.00000860, 0.00000626, 0.00000299, -0.00000014, -0.00000246, -0.00000372, -0.00000402, -0.00000353, -0.00000249, -0.00000116, 0.00000012, 0.00000099, 0.00000125, 0.00000091, 0.00000025, -0.00000036, -0.00000067, -0.00000061, -0.00000030, },
		{ 0.00000252, 0.00000154, -0.00000065, -0.00000408, -0.00000817, -0.00001170, -0.00001319, -0.00001159, -0.00000684, -0.00000000, 0.00000714, 0.00001286, 0.00001613, 0.00001688, 0.00001567, 0.00001315, 0.00000962, 0.00000506, -0.00000038, -0.00000591, -0.00001008, -0.00001132, -0.00000883, -0.00000332, 0.00000309, 0.00000772, 0.00000871, 0.00000601, 0.00000140, -0.00000252, -0.00000384, -0.00000244, 0.00000003, 0.00000126, -0.00000011, -0.00000351, -0.00000667, -0.00000696, -0.00000316, 0.00000360, 0.00001028, 0.00001354, 0.00001171, 0.00000577, -0.00000120, -0.00000575, -0.00000607, -0.00000291, 0.00000103, 0.00000295, 0.00000171, -0.00000142, -0.00000373, -0.00000298, 0.00000093, 0.00000561, 0.00000757, 0.00000452, -0.00000294, -0.00001115, -0.00001528, -0.00001204, -0.00000177, 0.00001136, 0.00002133, 0.00002326, 0.00001601, 0.00000266, -0.00001105, -0.00001965, -0.00002052, -0.00001484, -0.00000652, 0.00000015, 0.00000270, 0.00000140, -0.00000143, -0.00000308, -0.00000203, 0.00000140, 0.00000565, 0.00000907, 0.00001086, 0.00001125, 0.00001091, 0.00001016, 0.00000861, 0.00000551, 0.00000061, -0.00000531, -0.00001043, -0.00001277, -0.00001133, -0.00000677, -0.00000123, 0.00000272, 0.00000343, 0.00000106, -0.00000261, -0.00000509, -0.00000460, -0.00000095, 0.00000441, 0.00000918, 0.00001137, 0.00001021, 0.00000632, 0.00000128, -0.00000321, -0.00000598, -0.00000668, -0.00000563, -0.00000353, -0.00000111, 0.00000101, 0.00000239, 0.00000275, 0.00000210, 0.00000071, -0.00000087, -0.00000207, -0.00000246, -0.00000199, -0.00000092, 0.00000033, 0.00000143, 0.00000222, 0.00000264, },
		{ -0.00000913, -0.00001133, -0.00001156, -0.00000896, -0.00000352, 0.00000369, 0.00001083, 0.00001604, 0.00001817, 0.00001720, 0.00001405, 0.00000995, 0.00000575, 0.00000170, -0.00000229, -0.00000612, -0.00000913, -0.00001022, -0.00000859, -0.00000436, 0.00000110, 0.00000566, 0.00000756, 0.00000630, 0.00000305, -0.00000010, -0.00000134, -0.00000032, 0.00000167, 0.00000251, 0.00000076, -0.00000326, -0.00000736, -0.00000871, -0.00000557, 0.00000143, 0.00000937, 0.00001451, 0.00001431, 0.00000891, 0.00000104, -0.00000546, -0.00000782, -0.00000574, -0.00000139, 0.00000214, 0.00000278, 0.00000073, -0.00000190, -0.00000263, -0.00000052, 0.00000308, 0.00000526, 0.00000355, -0.00000221, -0.00000929, -0.00001338, -0.00001108, -0.00000211, 0.00001006, 0.00001977, 0.00002206, 0.00001532, 0.00000222, -0.00001157, -0.00002030, -0.00002100, -0.00001471, -0.00000563, 0.00000137, 0.00000338, 0.00000070, -0.00000366, -0.00000600, -0.00000405, 0.00000192, 0.00000952, 0.00001569, 0.00001830, 0.00001701, 0.00001294, 0.00000768, 0.00000236, -0.00000263, -0.00000728, -0.00001130, -0.00001382, -0.00001377, -0.00001067, -0.00000523, 0.00000065, 0.00000469, 0.00000542, 0.00000304, -0.00000065, -0.00000314, -0.00000258, 0.00000120, 0.00000662, 0.00001112, 0.00001252, 0.00000999, 0.00000446, -0.00000198, -0.00000710, -0.00000941, -0.00000868, -0.00000570, -0.00000183, 0.00000165, 0.00000386, 0.00000446, 0.00000352, 0.00000149, -0.00000098, -0.00000314, -0.00000430, -0.00000412, -0.00000270, -0.00000059, 0.00000145, 0.00000281, 0.00000318, 0.00000262, 0.00000131, -0.00000059, -0.00000309, -0.00000608, },
		{ 0.00000042, 0.00000695, 0.00001309, 0.00001712, 0.00001798, 0.00001572, 0.00001145, 0.00000667, 0.00000255, -0.00000058, -0.00000308, -0.00000545, -0.00000768, -0.00000907, -0.00000866, -0.00000594, -0.00000145, 0.00000330, 0.00000653, 0.00000721, 0.00000564, 0.00000324, 0.00000162, 0.00000152, 0.00000229, 0.00000237, 0.00000043, -0.00000345, -0.00000750, -0.00000910, -0.00000638, 0.00000042, 0.00000871, 0.00001474, 0.00001555, 0.00001065, 0.00000238, -0.00000532, -0.00000905, -0.00000773, -0.00000306, 0.00000172, 0.00000380, 0.00000248, -0.00000065, -0.00000291, -0.00000251, 0.00000021, 0.00000294, 0.00000303, -0.00000058, -0.00000627, -0.00001045, -0.00000968, -0.00000291, 0.00000743, 0.00001645, 0.00001931, 0.00001391, 0.00000220, -0.00001071, -0.00001922, -0.00002013, -0.00001416, -0.00000538, 0.00000128, 0.00000274, -0.00000063, -0.00000545, -0.00000747, -0.00000404, 0.00000439, 0.00001457, 0.00002227, 0.00002438, 0.00002033, 0.00001199, 0.00000239, -0.00000582, -0.00001135, -0.00001426, -0.00001520, -0.00001462, -0.00001249, -0.00000867, -0.00000355, 0.00000170, 0.00000548, 0.00000661, 0.00000510, 0.00000235, 0.00000044, 0.00000096, 0.00000409, 0.00000839, 0.00001148, 0.00001135, 0.00000740, 0.00000084, -0.00000596, -0.00001053, -0.00001143, -0.00000874, -0.00000387, 0.00000123, 0.00000489, 0.00000621, 0.00000522, 0.00000258, -0.00000075, -0.00000378, -0.00000572, -0.00000605, -0.00000474, -0.00000219, 0.00000075, 0.00000313, 0.00000418, 0.00000364, 0.00000181, -0.00000070, -0.00000331, -0.00000564, -0.00000757, -0.00000893, -0.00000935, -0.00000819, -0.00000493, },
		{ 0.00001608, 0.00001324, 0.00000871, 0.00000402, 0.00000040, -0.00000181, -0.00000315, -0.00000451, -0.00000634, -0.00000823, -0.00000904, -0.00000769, -0.00000396, 0.00000117, 0.00000592, 0.00000865, 0.00000875, 0.00000689, 0.00000450, 0.00000278, 0.00000194, 0.00000117, -0.00000060, -0.00000361, -0.00000677, -0.00000806, -0.00000576, 0.00000023, 0.00000790, 0.00001389, 0.00001522, 0.00001099, 0.00000304, -0.00000494, -0.00000936, -0.00000863, -0.00000397, 0.00000141, 0.00000431, 0.00000337, -0.00000022, -0.00000370, -0.00000459, -0.00000236, 0.00000118, 0.00000319, 0.00000175, -0.00000266, -0.00000723, -0.00000846, -0.00000445, 0.00000363, 0.00001187, 0.00001568, 0.00001241, 0.00000297, -0.00000848, -0.00001678, -0.00001852, -0.00001384, -0.00000619, -0.00000016, 0.00000121, -0.00000181, -0.00000590, -0.00000677, -0.00000178, 0.00000832, 0.00001970, 0.00002737, 0.00002776, 0.00002043, 0.00000812, -0.00000479, -0.00001445, -0.00001907, -0.00001915, -0.00001656, -0.00001313, -0.00000976, -0.00000635, -0.00000252, 0.00000164, 0.00000537, 0.00000767, 0.00000803, 0.00000694, 0.00000566, 0.00000544, 0.00000667, 0.00000849, 0.00000923, 0.00000741, 0.00000273, -0.00000359, -0.00000928, -0.00001213, -0.00001105, -0.00000660, -0.00000061, 0.00000460, 0.00000731, 0.00000692, 0.00000402, -0.00000011, -0.00000405, -0.00000670, -0.00000751, -0.00000642, -0.00000385, -0.00000052, 0.00000259, 0.00000455, 0.00000473, 0.00000307, 0.00000013, -0.00000316, -0.00000588, -0.00000747, -0.00000788, -0.00000732, -0.00000596, -0.00000373, -0.00000042, 0.00000402, 0.00000905, 0.00001355, 0.00001620, },
		{ -0.00000079, -0.00000235, -0.00000300, -0.00000379, -0.00000544, -0.00000772, -0.00000945, -0.00000912, -0.00000594, -0.00000041, 0.00000570, 0.00001025, 0.00001178, 0.00001026, 0.00000694, 0.00000347, 0.00000097, -0.00000055, -0.00000180, -0.00000343, -0.00000518, -0.00000587, -0.00000413, 0.00000046, 0.00000666, 0.00001188, 0.00001344, 0.00001019, 0.00000332, -0.00000408, -0.00000862, -0.00000847, -0.00000435, 0.00000087, 0.00000396, 0.00000322, -0.00000055, -0.00000464, -0.00000619, -0.00000409, 0.00000029, 0.00000398, 0.00000432, 0.00000081, -0.00000443, -0.00000784, -0.00000669, -0.00000086, 0.00000678, 0.00001192, 0.00001126, 0.00000451, -0.00000533, -0.00001367, -0.00001685, -0.00001416, -0.00000805, -0.00000250, -0.00000045, -0.00000200, -0.00000438, -0.00000374, 0.00000224, 0.00001268, 0.00002356, 0.00002975, 0.00002769, 0.00001733, 0.00000218, -0.00001244, -0.00002197, -0.00002455, -0.00002140, -0.00001562, -0.00001022, -0.00000667, -0.00000456, -0.00000246, 0.00000077, 0.00000512, 0.00000946, 0.00001237, 0.00001305, 0.00001181, 0.00000968, 0.00000762, 0.00000590, 0.00000399, 0.00000111, -0.00000292, -0.00000731, -0.00001048, -0.00001092, -0.00000808, -0.00000283, 0.00000288, 0.00000687, 0.00000772, 0.00000536, 0.00000093, -0.00000380, -0.00000727, -0.00000857, -0.00000765, -0.00000506, -0.00000166, 0.00000165, 0.00000405, 0.00000491, 0.00000397, 0.00000141, -0.00000203, -0.00000532, -0.00000746, -0.00000792, -0.00000675, -0.00000452, -0.00000187, 0.00000084, 0.00000361, 0.00000660, 0.00000973, 0.00001245, 0.00001388, 0.00001326, 0.00001051, 0.00000641, 0.00000226, },
		{ -0.00000494, -0.00000733, -0.00000953, -0.00000988, -0.00000714, -0.00000142, 0.00000563, 0.00001155, 0.00001423, 0.00001301, 0.00000888, 0.00000386, -0.00000012, -0.00000222, -0.00000282, -0.00000288, -0.00000304, -0.00000305, -0.00000200, 0.00000081, 0.00000500, 0.00000897, 0.00001062, 0.00000863, 0.00000344, -0.00000272, -0.00000701, -0.00000752, -0.00000449, -0.00000012, 0.00000268, 0.00000213, -0.00000134, -0.00000530, -0.00000690, -0.00000470, 0.00000027, 0.00000503, 0.00000651, 0.00000348, -0.00000251, -0.00000790, -0.00000923, -0.00000531, 0.00000197, 0.00000856, 0.00001055, 0.00000645, -0.00000195, -0.00001063, -0.00001560, -0.00001513, -0.00001050, -0.00000492, -0.00000136, -0.00000058, -0.00000078, 0.00000109, 0.00000701, 0.00001618, 0.00002496, 0.00002866, 0.00002416, 0.00001180, -0.00000449, -0.00001903, -0.00002712, -0.00002720, -0.00002121, -0.00001322, -0.00000699, -0.00000415, -0.00000369, -0.00000307, -0.00000010, 0.00000565, 0.00001257, 0.00001803, 0.00001988, 0.00001757, 0.00001223, 0.00000589, 0.00000032, -0.00000367, -0.00000624, -0.00000790, -0.00000881, -0.00000863, -0.00000680, -0.00000326, 0.00000121, 0.00000509, 0.00000680, 0.00000555, 0.00000175, -0.00000311, -0.00000721, -0.00000916, -0.00000854, -0.00000591, -0.00000235, 0.00000101, 0.00000337, 0.00000432, 0.00000380, 0.00000196, -0.00000082, -0.00000388, -0.00000637, -0.00000748, -0.00000678, -0.00000447, -0.00000124, 0.00000200, 0.00000461, 0.00000649, 0.00000792, 0.00000922, 0.00001038, 0.00001093, 0.00001023, 0.00000802, 0.00000470, 0.00000123, -0.00000137, -0.00000264, -0.00000301, -0.00000349, },
		{ -0.00000760, -0.00000208, 0.00000537, 0.00001219, 0.00001582, 0.00001503, 0.00001048, 0.00000434, -0.00000084, -0.00000348, -0.00000354, -0.00000218, -0.00000082, -0.00000014, 0.00000023, 0.00000116, 0.00000314, 0.00000563, 0.00000724, 0.00000661, 0.00000344, -0.00000109, -0.00000489, -0.00000617, -0.00000459, -0.00000154, 0.00000070, 0.00000045, -0.00000221, -0.00000540, -0.00000657, -0.00000426, 0.00000080, 0.00000588, 0.00000781, 0.00000494, -0.00000160, -0.00000838, -0.00001154, -0.00000903, -0.00000199, 0.00000584, 0.00001007, 0.00000819, 0.00000092, -0.00000821, -0.00001488, -0.00001635, -0.00001272, -0.00000652, -0.00000088, 0.00000255, 0.00000436, 0.00000667, 0.00001121, 0.00001765, 0.00002324, 0.00002418, 0.00001795, 0.00000513, -0.00001045, -0.00002337, -0.00002931, -0.00002714, -0.00001933, -0.00001041, -0.00000437, -0.00000262, -0.00000350, -0.00000359, -0.00000005, 0.00000758, 0.00001696, 0.00002422, 0.00002605, 0.00002140, 0.00001196, 0.00000114, -0.00000761, -0.00001229, -0.00001283, -0.00001055, -0.00000712, -0.00000369, -0.00000067, 0.00000191, 0.00000389, 0.00000474, 0.00000388, 0.00000113, -0.00000283, -0.00000669, -0.00000898, -0.00000886, -0.00000646, -0.00000280, 0.00000074, 0.00000308, 0.00000380, 0.00000307, 0.00000142, -0.00000067, -0.00000280, -0.00000466, -0.00000582, -0.00000583, -0.00000442, -0.00000172, 0.00000166, 0.00000482, 0.00000704, 0.00000807, 0.00000824, 0.00000812, 0.00000808, 0.00000801, 0.00000743, 0.00000588, 0.00000338, 0.00000056, -0.00000173, -0.00000293, -0.00000324, -0.00000350, -0.00000461, -0.00000676, -0.00000906, -0.00000983, },
		{ 0.00001654, 0.00001645, 0.00001198, 0.00000522, -0.00000092, -0.00000418, -0.00000403, -0.00000164, 0.00000101, 0.00000242, 0.00000232, 0.00000160, 0.00000145, 0.00000235, 0.00000373, 0.00000432, 0.00000318, 0.00000043, -0.00000271, -0.00000470, -0.00000468, -0.00000315, -0.00000157, -0.00000137, -0.00000286, -0.00000485, -0.00000538, -0.00000312, 0.00000148, 0.00000615, 0.00000796, 0.00000516, -0.00000150, -0.00000890, -0.00001309, -0.00001158, -0.00000487, 0.00000364, 0.00000937, 0.00000911, 0.00000277, -0.00000659, -0.00001441, -0.00001712, -0.00001386, -0.00000665, 0.00000117, 0.00000692, 0.00001001, 0.00001170, 0.00001370, 0.00001645, 0.00001847, 0.00001708, 0.00001035, -0.00000125, -0.00001454, -0.00002489, -0.00002865, -0.00002508, -0.00001675, -0.00000807, -0.00000280, -0.00000192, -0.00000331, -0.00000315, 0.00000156, 0.00001095, 0.00002189, 0.00002957, 0.00002998, 0.00002210, 0.00000849, -0.00000596, -0.00001637, -0.00001994, -0.00001690, -0.00000987, -0.00000226, 0.00000330, 0.00000580, 0.00000565, 0.00000388, 0.00000136, -0.00000144, -0.00000430, -0.00000683, -0.00000838, -0.00000828, -0.00000632, -0.00000300, 0.00000051, 0.00000299, 0.00000367, 0.00000263, 0.00000061, -0.00000146, -0.00000295, -0.00000372, -0.00000392, -0.00000372, -0.00000302, -0.00000159, 0.00000068, 0.00000351, 0.00000622, 0.00000806, 0.00000862, 0.00000808, 0.00000704, 0.00000610, 0.00000548, 0.00000491, 0.00000388, 0.00000212, -0.00000009, -0.00000208, -0.00000324, -0.00000352, -0.00000354, -0.00000420, -0.00000589, -0.00000804, -0.00000913, -0.00000755, -0.00000262, 0.00000472, 0.00001204, },
		{ -0.00000030, -0.00000438, -0.00000449, -0.00000155, 0.00000215, 0.00000438, 0.00000421, 0.00000231, 0.00000029, -0.00000045, 0.00000036, 0.00000178, 0.00000243, 0.00000146, -0.00000082, -0.00000325, -0.00000463, -0.00000456, -0.00000366, -0.00000301, -0.00000322, -0.00000387, -0.00000373, -0.00000174, 0.00000195, 0.00000567, 0.00000704, 0.00000440, -0.00000185, -0.00000907, -0.00001362, -0.00001285, -0.00000679, 0.00000165, 0.00000805, 0.00000888, 0.00000347, -0.00000555, -0.00001364, -0.00001674, -0.00001334, -0.00000509, 0.00000439, 0.00001160, 0.00001492, 0.00001501, 0.00001376, 0.00001260, 0.00001142, 0.00000867, 0.00000280, -0.00000620, -0.00001619, -0.00002370, -0.00002582, -0.00002196, -0.00001430, -0.00000663, -0.00000213, -0.00000145, -0.00000236, -0.00000119, 0.00000473, 0.00001503, 0.00002607, 0.00003257, 0.00003049, 0.00001927, 0.00000245, -0.00001389, -0.00002393, -0.00002474, -0.00001733, -0.00000584, 0.00000467, 0.00001053, 0.00001069, 0.00000654, 0.00000069, -0.00000450, -0.00000778, -0.00000903, -0.00000872, -0.00000733, -0.00000511, -0.00000232, 0.00000055, 0.00000273, 0.00000346, 0.00000249, 0.00000028, -0.00000215, -0.00000376, -0.00000404, -0.00000318, -0.00000185, -0.00000071, 0.00000008, 0.00000087, 0.00000214, 0.00000407, 0.00000630, 0.00000806, 0.00000867, 0.00000795, 0.00000636, 0.00000463, 0.00000333, 0.00000250, 0.00000176, 0.00000066, -0.00000087, -0.00000240, -0.00000339, -0.00000359, -0.00000339, -0.00000359, -0.00000477, -0.00000668, -0.00000806, -0.00000725, -0.00000318, 0.00000366, 0.00001120, 0.00001654, 0.00001745, 0.00001354, 0.00000661, },
		{ 0.00000249, 0.00000571, 0.00000594, 0.00000341, -0.00000013, -0.00000255, -0.00000271, -0.00000102, 0.00000098, 0.00000173, 0.00000058, -0.00000183, -0.00000419, -0.00000540, -0.00000522, -0.00000429, -0.00000341, -0.00000283, -0.00000210, -0.00000051, 0.00000203, 0.00000458, 0.00000539, 0.00000308, -0.00000228, -0.00000872, -0.00001315, -0.00001306, -0.00000805, -0.00000041, 0.00000594, 0.00000753, 0.00000330, -0.00000462, -0.00001202, -0.00001476, -0.00001103, -0.00000221, 0.00000797, 0.00001547, 0.00001801, 0.00001589, 0.00001134, 0.00000678, 0.00000331, 0.00000035, -0.00000353, -0.00000910, -0.00001550, -0.00002049, -0.00002178, -0.00001860, -0.00001238, -0.00000595, -0.00000181, -0.00000051, -0.00000017, 0.00000224, 0.00000874, 0.00001859, 0.00002806, 0.00003213, 0.00002723, 0.00001353, -0.00000466, -0.00002065, -0.00002843, -0.00002555, -0.00001409, 0.00000044, 0.00001183, 0.00001600, 0.00001248, 0.00000406, -0.00000506, -0.00001135, -0.00001328, -0.00001146, -0.00000765, -0.00000361, -0.00000040, 0.00000170, 0.00000278, 0.00000287, 0.00000191, 0.00000004, -0.00000224, -0.00000405, -0.00000458, -0.00000354, -0.00000141, 0.00000081, 0.00000224, 0.00000261, 0.00000241, 0.00000254, 0.00000364, 0.00000560, 0.00000761, 0.00000864, 0.00000807, 0.00000609, 0.00000354, 0.00000134, 0.00000001, -0.00000060, -0.00000102, -0.00000166, -0.00000249, -0.00000313, -0.00000324, -0.00000295, -0.00000287, -0.00000365, -0.00000531, -0.00000690, -0.00000682, -0.00000373, 0.00000240, 0.00000994, 0.00001606, 0.00001813, 0.00001516, 0.00000840, 0.00000087, -0.00000422, -0.00000503, -0.00000202, },
		{ 0.00000023, -0.00000389, -0.00000543, -0.00000407, -0.00000118, 0.00000117, 0.00000147, -0.00000037, -0.00000320, -0.00000544, -0.00000611, -0.00000525, -0.00000366, -0.00000212, -0.00000088, 0.00000034, 0.00000180, 0.00000318, 0.00000347, 0.00000160, -0.00000262, -0.00000794, -0.00001201, -0.00001259, -0.00000898, -0.00000267, 0.00000317, 0.00000541, 0.00000276, -0.00000332, -0.00000923, -0.00001119, -0.00000731, 0.00000126, 0.00001094, 0.00001758, 0.00001863, 0.00001426, 0.00000701, 0.00000007, -0.00000451, -0.00000669, -0.00000798, -0.00000997, -0.00001307, -0.00001615, -0.00001734, -0.00001545, -0.00001091, -0.00000550, -0.00000118, 0.00000135, 0.00000317, 0.00000639, 0.00001237, 0.00002029, 0.00002689, 0.00002798, 0.00002087, 0.00000639, -0.00001088, -0.00002440, -0.00002875, -0.00002232, -0.00000825, 0.00000710, 0.00001714, 0.00001814, 0.00001064, -0.00000117, -0.00001183, -0.00001723, -0.00001626, -0.00001068, -0.00000372, 0.00000173, 0.00000429, 0.00000423, 0.00000267, 0.00000071, -0.00000109, -0.00000259, -0.00000371, -0.00000416, -0.00000354, -0.00000175, 0.00000076, 0.00000304, 0.00000421, 0.00000400, 0.00000300, 0.00000232, 0.00000288, 0.00000478, 0.00000715, 0.00000865, 0.00000827, 0.00000592, 0.00000250, -0.00000066, -0.00000257, -0.00000306, -0.00000268, -0.00000223, -0.00000216, -0.00000238, -0.00000250, -0.00000236, -0.00000227, -0.00000278, -0.00000414, -0.00000576, -0.00000623, -0.00000408, 0.00000121, 0.00000850, 0.00001521, 0.00001846, 0.00001664, 0.00001037, 0.00000239, -0.00000378, -0.00000564, -0.00000298, 0.00000210, 0.00000641, 0.00000750, 0.00000490, },
		{ -0.00000391, -0.00000012, 0.00000188, 0.00000114, -0.00000163, -0.00000467, -0.00000634, -0.00000601, -0.00000421, -0.00000201, -0.00000028, 0.00000080, 0.00000146, 0.00000187, 0.00000167, 0.00000021, -0.00000287, -0.00000701, -0.00001062, -0.00001185, -0.00000973, -0.00000496, 0.00000021, 0.00000313, 0.00000236, -0.00000137, -0.00000538, -0.00000650, -0.00000296, 0.00000442, 0.00001250, 0.00001743, 0.00001674, 0.00001061, 0.00000171, -0.00000636, -0.00001099, -0.00001179, -0.00001043, -0.00000925, -0.00000969, -0.00001148, -0.00001298, -0.00001250, -0.00000945, -0.00000470, 0.00000016, 0.00000398, 0.00000692, 0.00001006, 0.00001433, 0.00001918, 0.00002235, 0.00002086, 0.00001296, -0.00000021, -0.00001445, -0.00002411, -0.00002489, -0.00001618, -0.00000169, 0.00001210, 0.00001907, 0.00001638, 0.00000575, -0.00000759, -0.00001761, -0.00002039, -0.00001576, -0.00000682, 0.00000193, 0.00000701, 0.00000733, 0.00000419, 0.00000003, -0.00000308, -0.00000431, -0.00000406, -0.00000315, -0.00000214, -0.00000102, 0.00000044, 0.00000224, 0.00000388, 0.00000469, 0.00000431, 0.00000315, 0.00000222, 0.00000251, 0.00000428, 0.00000673, 0.00000842, 0.00000808, 0.00000541, 0.00000131, -0.00000262, -0.00000496, -0.00000523, -0.00000399, -0.00000241, -0.00000142, -0.00000127, -0.00000158, -0.00000185, -0.00000197, -0.00000230, -0.00000324, -0.00000460, -0.00000534, -0.00000401, 0.00000029, 0.00000701, 0.00001398, 0.00001828, 0.00001774, 0.00001227, 0.00000413, -0.00000305, -0.00000616, -0.00000422, 0.00000112, 0.00000651, 0.00000877, 0.00000662, 0.00000126, -0.00000446, -0.00000769, -0.00000721, },
		{ 0.00000033, -0.00000326, -0.00000605, -0.00000666, -0.00000512, -0.00000255, -0.00000026, 0.00000099, 0.00000124, 0.00000091, 0.00000023, -0.00000102, -0.00000322, -0.00000628, -0.00000936, -0.00001105, -0.00001023, -0.00000689, -0.00000238, 0.00000124, 0.00000243, 0.00000114, -0.00000095, -0.00000148, 0.00000115, 0.00000651, 0.00001221, 0.00001505, 0.00001285, 0.00000579, -0.00000356, -0.00001157, -0.00001549, -0.00001477, -0.00001119, -0.00000756, -0.00000605, -0.00000696, -0.00000881, -0.00000947, -0.00000754, -0.00000321, 0.00000209, 0.00000674, 0.00000995, 0.00001203, 0.00001372, 0.00001514, 0.00001525, 0.00001235, 0.00000544, -0.00000453, -0.00001435, -0.00001985, -0.00001802, -0.00000898, 0.00000357, 0.00001392, 0.00001707, 0.00001137, -0.00000057, -0.00001316, -0.00002061, -0.00001987, -0.00001193, -0.00000107, 0.00000754, 0.00001051, 0.00000759, 0.00000142, -0.00000438, -0.00000714, -0.00000635, -0.00000338, -0.00000029, 0.00000153, 0.00000197, 0.00000186, 0.00000210, 0.00000292, 0.00000383, 0.00000414, 0.00000367, 0.00000297, 0.00000294, 0.00000408, 0.00000598, 0.00000740, 0.00000703, 0.00000433, 0.00000005, -0.00000417, -0.00000666, -0.00000668, -0.00000474, -0.00000223, -0.00000052, -0.00000016, -0.00000080, -0.00000160, -0.00000201, -0.00000212, -0.00000246, -0.00000328, -0.00000404, -0.00000346, -0.00000033, 0.00000544, 0.00001227, 0.00001740, 0.00001825, 0.00001395, 0.00000608, -0.00000190, -0.00000637, -0.00000553, -0.00000035, 0.00000593, 0.00000955, 0.00000834, 0.00000280, -0.00000424, -0.00000928, -0.00001013, -0.00000694, -0.00000200, 0.00000181, 0.00000258, },
		{ -0.00000629, -0.00000359, -0.00000072, 0.00000101, 0.00000123, 0.00000040, -0.00000084, -0.00000219, -0.00000383, -0.00000598, -0.00000840, -0.00001018, -0.00001023, -0.00000802, -0.00000409, 0.00000009, 0.00000296, 0.00000382, 0.00000332, 0.00000300, 0.00000425, 0.00000715, 0.00001016, 0.00001095, 0.00000776, 0.00000072, -0.00000797, -0.00001505, -0.00001788, -0.00001588, -0.00001074, -0.00000546, -0.00000255, -0.00000273, -0.00000469, -0.00000607, -0.00000500, -0.00000119, 0.00000402, 0.00000861, 0.00001117, 0.00001151, 0.00001045, 0.00000893, 0.00000711, 0.00000437, 0.00000003, -0.00000558, -0.00001072, -0.00001281, -0.00001004, -0.00000277, 0.00000605, 0.00001208, 0.00001184, 0.00000472, -0.00000632, -0.00001612, -0.00001992, -0.00001586, -0.00000601, 0.00000474, 0.00001131, 0.00001105, 0.00000491, -0.00000328, -0.00000911, -0.00000994, -0.00000610, -0.00000027, 0.00000430, 0.00000570, 0.00000416, 0.00000156, -0.00000002, 0.00000044, 0.00000238, 0.00000440, 0.00000530, 0.00000491, 0.00000405, 0.00000372, 0.00000429, 0.00000507, 0.00000486, 0.00000280, -0.00000085, -0.00000474, -0.00000717, -0.00000716, -0.00000499, -0.00000205, 0.00000007, 0.00000054, -0.00000035, -0.00000156, -0.00000215, -0.00000195, -0.00000158, -0.00000172, -0.00000238, -0.00000255, -0.00000079, 0.00000367, 0.00001001, 0.00001579, 0.00001816, 0.00001541, 0.00000827, -0.00000021, -0.00000613, -0.00000680, -0.00000229, 0.00000455, 0.00000962, 0.00000981, 0.00000474, -0.00000315, -0.00000995, -0.00001248, -0.00000998, -0.00000434, 0.00000114, 0.00000367, 0.00000236, -0.00000146, -0.00000534, -0.00000716, },
		{ 0.00000136, 0.00000021, -0.00000163, -0.00000335, -0.00000475, -0.00000613, -0.00000769, -0.00000909, -0.00000947, -0.00000803, -0.00000468, -0.00000029, 0.00000369, 0.00000609, 0.00000667, 0.00000625, 0.00000596, 0.00000634, 0.00000681, 0.00000591, 0.00000236, -0.00000385, -0.00001110, -0.00001671, -0.00001837, -0.00001547, -0.00000948, -0.00000321, 0.00000070, 0.00000126, -0.00000051, -0.00000230, -0.00000207, 0.00000078, 0.00000507, 0.00000870, 0.00000994, 0.00000845, 0.00000527, 0.00000201, -0.00000023, -0.00000142, -0.00000230, -0.00000348, -0.00000470, -0.00000489, -0.00000298, 0.00000096, 0.00000525, 0.00000728, 0.00000501, -0.00000152, -0.00000972, -0.00001556, -0.00001573, -0.00000961, 0.00000013, 0.00000878, 0.00001205, 0.00000845, 0.00000011, -0.00000838, -0.00001252, -0.00001033, -0.00000329, 0.00000464, 0.00000937, 0.00000897, 0.00000450, -0.00000082, -0.00000371, -0.00000265, 0.00000140, 0.00000588, 0.00000832, 0.00000778, 0.00000519, 0.00000246, 0.00000110, 0.00000126, 0.00000183, 0.00000136, -0.00000074, -0.00000380, -0.00000626, -0.00000679, -0.00000515, -0.00000239, -0.00000010, 0.00000063, -0.00000015, -0.00000140, -0.00000198, -0.00000148, -0.00000049, -0.00000006, -0.00000066, -0.00000161, -0.00000130, 0.00000168, 0.00000729, 0.00001361, 0.00001759, 0.00001673, 0.00001072, 0.00000199, -0.00000545, -0.00000803, -0.00000473, 0.00000230, 0.00000886, 0.00001095, 0.00000706, -0.00000108, -0.00000944, -0.00001398, -0.00001283, -0.00000715, -0.00000032, 0.00000407, 0.00000414, 0.00000055, -0.00000420, -0.00000734, -0.00000743, -0.00000491, -0.00000152, 0.00000087, },
		{ -0.00000584, -0.00000653, -0.00000705, -0.00000762, -0.00000783, -0.00000687, -0.00000420, -0.00000013, 0.00000420, 0.00000740, 0.00000859, 0.00000791, 0.00000624, 0.00000446, 0.00000281, 0.00000071, -0.00000266, -0.00000751, -0.00001283, -0.00001672, -0.00001729, -0.00001388, -0.00000761, -0.00000090, 0.00000372, 0.00000501, 0.00000359, 0.00000145, 0.00000061, 0.00000191, 0.00000453, 0.00000658, 0.00000639, 0.00000364, -0.00000048, -0.00000397, -0.00000528, -0.00000416, -0.00000165, 0.00000070, 0.00000189, 0.00000196, 0.00000169, 0.00000172, 0.00000187, 0.00000115, -0.00000138, -0.00000559, -0.00000986, -0.00001173, -0.00000938, -0.00000306, 0.00000463, 0.00000982, 0.00000956, 0.00000358, -0.00000522, -0.00001217, -0.00001339, -0.00000795, 0.00000146, 0.00001002, 0.00001340, 0.00001020, 0.00000262, -0.00000481, -0.00000791, -0.00000511, 0.00000181, 0.00000888, 0.00001234, 0.00001073, 0.00000544, -0.00000035, -0.00000373, -0.00000367, -0.00000137, 0.00000079, 0.00000093, -0.00000117, -0.00000415, -0.00000608, -0.00000583, -0.00000374, -0.00000125, 0.00000021, 0.00000018, -0.00000065, -0.00000111, -0.00000054, 0.00000065, 0.00000133, 0.00000066, -0.00000098, -0.00000198, -0.00000041, 0.00000443, 0.00001119, 0.00001676, 0.00001792, 0.00001330, 0.00000451, -0.00000446, -0.00000926, -0.00000762, -0.00000072, 0.00000733, 0.00001178, 0.00000978, 0.00000204, -0.00000766, -0.00001453, -0.00001544, -0.00001049, -0.00000277, 0.00000353, 0.00000544, 0.00000268, -0.00000251, -0.00000693, -0.00000828, -0.00000632, -0.00000266, 0.00000042, 0.00000139, 0.00000016, -0.00000222, -0.00000445, },
		{ -0.00000545, -0.00000478, -0.00000297, 0.00000023, 0.00000415, 0.00000745, 0.00000887, 0.00000799, 0.00000532, 0.00000196, -0.00000122, -0.00000404, -0.00000688, -0.00001006, -0.00001326, -0.00001533, -0.00001491, -0.00001130, -0.00000520, 0.00000147, 0.00000644, 0.00000828, 0.00000713, 0.00000450, 0.00000231, 0.00000164, 0.00000218, 0.00000255, 0.00000139, -0.00000160, -0.00000529, -0.00000773, -0.00000736, -0.00000400, 0.00000095, 0.00000530, 0.00000726, 0.00000637, 0.00000350, 0.00000014, -0.00000256, -0.00000435, -0.00000560, -0.00000660, -0.00000700, -0.00000593, -0.00000278, 0.00000192, 0.00000624, 0.00000770, 0.00000474, -0.00000195, -0.00000933, -0.00001339, -0.00001134, -0.00000344, 0.00000678, 0.00001428, 0.00001516, 0.00000894, -0.00000102, -0.00000927, -0.00001126, -0.00000588, 0.00000396, 0.00001296, 0.00001639, 0.00001265, 0.00000401, -0.00000490, -0.00000974, -0.00000881, -0.00000376, 0.00000176, 0.00000435, 0.00000280, -0.00000151, -0.00000577, -0.00000756, -0.00000629, -0.00000318, -0.00000023, 0.00000119, 0.00000113, 0.00000062, 0.00000070, 0.00000142, 0.00000191, 0.00000114, -0.00000087, -0.00000270, -0.00000221, 0.00000191, 0.00000890, 0.00001579, 0.00001882, 0.00001566, 0.00000700, -0.00000338, -0.00001045, -0.00001073, -0.00000424, 0.00000528, 0.00001237, 0.00001284, 0.00000604, -0.00000475, -0.00001419, -0.00001783, -0.00001437, -0.00000625, 0.00000198, 0.00000622, 0.00000499, -0.00000012, -0.00000571, -0.00000863, -0.00000775, -0.00000425, -0.00000060, 0.00000106, 0.00000009, -0.00000257, -0.00000526, -0.00000675, -0.00000686, -0.00000627, -0.00000575, },
		{ 0.00000341, 0.00000625, 0.00000765, 0.00000670, 0.00000356, -0.00000072, -0.00000483, -0.00000796, -0.00001005, -0.00001149, -0.00001252, -0.00001277, -0.00001142, -0.00000786, -0.00000235, 0.00000376, 0.00000857, 0.00001055, 0.00000941, 0.00000616, 0.00000253, -0.00000013, -0.00000153, -0.00000240, -0.00000375, -0.00000592, -0.00000811, -0.00000881, -0.00000676, -0.00000198, 0.00000397, 0.00000871, 0.00001023, 0.00000796, 0.00000305, -0.00000234, -0.00000621, -0.00000760, -0.00000680, -0.00000477, -0.00000241, -0.00000007, 0.00000218, 0.00000408, 0.00000480, 0.00000330, -0.00000078, -0.00000636, -0.00001092, -0.00001163, -0.00000699, 0.00000182, 0.00001106, 0.00001616, 0.00001416, 0.00000563, -0.00000525, -0.00001274, -0.00001259, -0.00000452, 0.00000745, 0.00001707, 0.00001918, 0.00001256, 0.00000058, -0.00001071, -0.00001588, -0.00001299, -0.00000447, 0.00000453, 0.00000912, 0.00000728, 0.00000072, -0.00000653, -0.00001055, -0.00000973, -0.00000526, -0.00000003, 0.00000339, 0.00000409, 0.00000300, 0.00000173, 0.00000124, 0.00000119, 0.00000057, -0.00000114, -0.00000306, -0.00000320, 0.00000018, 0.00000695, 0.00001457, 0.00001908, 0.00001732, 0.00000908, -0.00000230, -0.00001141, -0.00001363, -0.00000777, 0.00000300, 0.00001276, 0.00001599, 0.00001056, -0.00000106, -0.00001318, -0.00002000, -0.00001867, -0.00001059, -0.00000045, 0.00000658, 0.00000753, 0.00000300, -0.00000365, -0.00000849, -0.00000928, -0.00000643, -0.00000236, 0.00000024, 0.00000002, -0.00000248, -0.00000544, -0.00000711, -0.00000687, -0.00000533, -0.00000369, -0.00000272, -0.00000223, -0.00000139, 0.00000052, },
		{ 0.00000122, -0.00000333, -0.00000779, -0.00001086, -0.00001205, -0.00001174, -0.00001065, -0.00000915, -0.00000702, -0.00000380, 0.00000063, 0.00000555, 0.00000954, 0.00001118, 0.00000984, 0.00000608, 0.00000134, -0.00000280, -0.00000553, -0.00000695, -0.00000775, -0.00000842, -0.00000869, -0.00000767, -0.00000459, 0.00000043, 0.00000601, 0.00001000, 0.00001060, 0.00000741, 0.00000171, -0.00000408, -0.00000765, -0.00000788, -0.00000525, -0.00000135, 0.00000210, 0.00000405, 0.00000432, 0.00000324, 0.00000116, -0.00000176, -0.00000521, -0.00000832, -0.00000959, -0.00000752, -0.00000174, 0.00000619, 0.00001301, 0.00001516, 0.00001085, 0.00000147, -0.00000859, -0.00001402, -0.00001140, -0.00000140, 0.00001120, 0.00001984, 0.00001959, 0.00001002, -0.00000444, -0.00001671, -0.00002083, -0.00001515, -0.00000310, 0.00000874, 0.00001428, 0.00001114, 0.00000167, -0.00000865, -0.00001445, -0.00001329, -0.00000664, 0.00000145, 0.00000691, 0.00000790, 0.00000537, 0.00000186, -0.00000043, -0.00000100, -0.00000089, -0.00000137, -0.00000252, -0.00000289, -0.00000052, 0.00000529, 0.00001277, 0.00001819, 0.00001785, 0.00001056, -0.00000110, -0.00001170, -0.00001576, -0.00001086, 0.00000067, 0.00001276, 0.00001875, 0.00001501, 0.00000296, -0.00001165, -0.00002182, -0.00002300, -0.00001537, -0.00000343, 0.00000661, 0.00001021, 0.00000668, -0.00000094, -0.00000802, -0.00001102, -0.00000928, -0.00000488, -0.00000100, 0.00000011, -0.00000167, -0.00000466, -0.00000667, -0.00000647, -0.00000440, -0.00000184, -0.00000014, 0.00000032, 0.00000024, 0.00000067, 0.00000213, 0.00000407, 0.00000522, 0.00000436, },
		{ -0.00001268, -0.00001070, -0.00000767, -0.00000465, -0.00000205, 0.00000041, 0.00000316, 0.00000620, 0.00000877, 0.00000973, 0.00000826, 0.00000447, -0.00000056, -0.00000530, -0.00000854, -0.00000992, -0.00000986, -0.00000900, -0.00000763, -0.00000552, -0.00000229, 0.00000197, 0.00000631, 0.00000917, 0.00000918, 0.00000606, 0.00000099, -0.00000386, -0.00000636, -0.00000558, -0.00000218, 0.00000197, 0.00000487, 0.00000535, 0.00000347, 0.00000026, -0.00000309, -0.00000572, -0.00000725, -0.00000743, -0.00000593, -0.00000244, 0.00000278, 0.00000839, 0.00001211, 0.00001172, 0.00000643, -0.00000204, -0.00000982, -0.00001260, -0.00000807, 0.00000237, 0.00001382, 0.00002011, 0.00001715, 0.00000541, -0.00000995, -0.00002149, -0.00002341, -0.00001474, 0.00000007, 0.00001341, 0.00001858, 0.00001334, 0.00000093, -0.00001184, -0.00001841, -0.00001595, -0.00000654, 0.00000440, 0.00001137, 0.00001177, 0.00000688, 0.00000048, -0.00000374, -0.00000438, -0.00000264, -0.00000090, -0.00000060, -0.00000112, -0.00000033, 0.00000352, 0.00000990, 0.00001575, 0.00001705, 0.00001153, 0.00000059, -0.00001079, -0.00001660, -0.00001319, -0.00000175, 0.00001196, 0.00002050, 0.00001877, 0.00000692, -0.00000960, -0.00002290, -0.00002681, -0.00002004, -0.00000667, 0.00000630, 0.00001275, 0.00001052, 0.00000206, -0.00000739, -0.00001295, -0.00001262, -0.00000792, -0.00000242, 0.00000061, 0.00000006, -0.00000279, -0.00000540, -0.00000583, -0.00000379, -0.00000063, 0.00000184, 0.00000256, 0.00000183, 0.00000082, 0.00000061, 0.00000130, 0.00000192, 0.00000118, -0.00000157, -0.00000579, -0.00000998, -0.00001255, },
		{ 0.00000284, 0.00000396, 0.00000447, 0.00000510, 0.00000593, 0.00000625, 0.00000515, 0.00000220, -0.00000207, -0.00000644, -0.00000959, -0.00001078, -0.00001016, -0.00000840, -0.00000618, -0.00000376, -0.00000104, 0.00000205, 0.00000509, 0.00000717, 0.00000736, 0.00000536, 0.00000193, -0.00000136, -0.00000289, -0.00000197, 0.00000076, 0.00000366, 0.00000496, 0.00000378, 0.00000054, -0.00000333, -0.00000628, -0.00000735, -0.00000650, -0.00000427, -0.00000131, 0.00000198, 0.00000525, 0.00000788, 0.00000884, 0.00000707, 0.00000244, -0.00000366, -0.00000841, -0.00000886, -0.00000371, 0.00000538, 0.00001409, 0.00001735, 0.00001219, -0.00000021, -0.00001454, -0.00002381, -0.00002299, -0.00001196, 0.00000411, 0.00001726, 0.00002088, 0.00001330, -0.00000133, -0.00001524, -0.00002129, -0.00001676, -0.00000459, 0.00000850, 0.00001590, 0.00001466, 0.00000674, -0.00000264, -0.00000833, -0.00000818, -0.00000385, 0.00000086, 0.00000286, 0.00000183, 0.00000020, 0.00000106, 0.00000556, 0.00001166, 0.00001513, 0.00001240, 0.00000329, -0.00000824, -0.00001591, -0.00001482, -0.00000461, 0.00000981, 0.00002064, 0.00002140, 0.00001071, -0.00000678, -0.00002272, -0.00002947, -0.00002416, -0.00001005, 0.00000541, 0.00001467, 0.00001398, 0.00000500, -0.00000667, -0.00001484, -0.00001605, -0.00001105, -0.00000369, 0.00000166, 0.00000267, 0.00000001, -0.00000356, -0.00000522, -0.00000382, -0.00000036, 0.00000299, 0.00000440, 0.00000347, 0.00000123, -0.00000079, -0.00000177, -0.00000204, -0.00000274, -0.00000477, -0.00000801, -0.00001123, -0.00001280, -0.00001171, -0.00000823, -0.00000369, 0.00000032, },
		{ 0.00000130, 0.00000150, 0.00000158, 0.00000049, -0.00000206, -0.00000542, -0.00000836, -0.00000983, -0.00000952, -0.00000788, -0.00000566, -0.00000341, -0.00000124, 0.00000101, 0.00000336, 0.00000542, 0.00000653, 0.00000622, 0.00000467, 0.00000271, 0.00000139, 0.00000131, 0.00000217, 0.00000294, 0.00000247, 0.00000030, -0.00000297, -0.00000594, -0.00000726, -0.00000631, -0.00000355, -0.00000017, 0.00000269, 0.00000445, 0.00000516, 0.00000512, 0.00000440, 0.00000277, 0.00000016, -0.00000283, -0.00000476, -0.00000398, 0.00000015, 0.00000635, 0.00001147, 0.00001192, 0.00000582, -0.00000536, -0.00001692, -0.00002301, -0.00001975, -0.00000775, 0.00000771, 0.00001909, 0.00002054, 0.00001116, -0.00000426, -0.00001767, -0.00002206, -0.00001525, -0.00000108, 0.00001283, 0.00001933, 0.00001561, 0.00000458, -0.00000717, -0.00001331, -0.00001137, -0.00000376, 0.00000411, 0.00000748, 0.00000523, 0.00000030, -0.00000258, -0.00000029, 0.00000628, 0.00001269, 0.00001381, 0.00000740, -0.00000392, -0.00001387, -0.00001613, -0.00000839, 0.00000587, 0.00001886, 0.00002283, 0.00001451, -0.00000279, -0.00002077, -0.00003056, -0.00002752, -0.00001369, 0.00000352, 0.00001541, 0.00001662, 0.00000770, -0.00000568, -0.00001622, -0.00001900, -0.00001381, -0.00000462, 0.00000316, 0.00000586, 0.00000334, -0.00000150, -0.00000491, -0.00000463, -0.00000106, 0.00000337, 0.00000595, 0.00000532, 0.00000210, -0.00000190, -0.00000497, -0.00000656, -0.00000732, -0.00000829, -0.00000984, -0.00001128, -0.00001130, -0.00000894, -0.00000439, 0.00000090, 0.00000503, 0.00000670, 0.00000592, 0.00000385, 0.00000201, },
		{ 0.00000010, -0.00000223, -0.00000546, -0.00000813, -0.00000923, -0.00000862, -0.00000687, -0.00000467, -0.00000240, -0.00000005, 0.00000252, 0.00000519, 0.00000745, 0.00000863, 0.00000838, 0.00000690, 0.00000485, 0.00000292, 0.00000141, 0.00000007, -0.00000154, -0.00000356, -0.00000552, -0.00000652, -0.00000577, -0.00000322, 0.00000028, 0.00000335, 0.00000485, 0.00000449, 0.00000294, 0.00000129, 0.00000033, 0.00000012, 0.00000018, 0.00000008, -0.00000004, 0.00000050, 0.00000221, 0.00000469, 0.00000630, 0.00000497, -0.00000040, -0.00000867, -0.00001639, -0.00001927, -0.00001465, -0.00000346, 0.00000964, 0.00001824, 0.00001767, 0.00000774, -0.00000669, -0.00001807, -0.00002017, -0.00001163, 0.00000312, 0.00001619, 0.00002058, 0.00001412, 0.00000066, -0.00001220, -0.00001752, -0.00001296, -0.00000196, 0.00000853, 0.00001240, 0.00000805, -0.00000078, -0.00000756, -0.00000722, 0.00000046, 0.00001060, 0.00001628, 0.00001299, 0.00000176, -0.00001113, -0.00001778, -0.00001352, -0.00000001, 0.00001531, 0.00002339, 0.00001871, 0.00000273, -0.00001681, -0.00003001, -0.00003028, -0.00001792, 0.00000022, 0.00001460, 0.00001824, 0.00001028, -0.00000402, -0.00001654, -0.00002095, -0.00001591, -0.00000525, 0.00000474, 0.00000908, 0.00000668, 0.00000045, -0.00000496, -0.00000605, -0.00000247, 0.00000323, 0.00000737, 0.00000745, 0.00000340, -0.00000274, -0.00000829, -0.00001155, -0.00001240, -0.00001184, -0.00001091, -0.00000979, -0.00000786, -0.00000441, 0.00000041, 0.00000532, 0.00000846, 0.00000849, 0.00000551, 0.00000105, -0.00000270, -0.00000418, -0.00000321, -0.00000106, 0.00000047, },
		{ -0.00001041, -0.00001123, -0.00000974, -0.00000681, -0.00000334, 0.00000020, 0.00000370, 0.00000709, 0.00000997, 0.00001167, 0.00001161, 0.00000963, 0.00000625, 0.00000236, -0.00000117, -0.00000382, -0.00000551, -0.00000630, -0.00000616, -0.00000492, -0.00000252, 0.00000061, 0.00000355, 0.00000514, 0.00000472, 0.00000252, -0.00000026, -0.00000214, -0.00000211, -0.00000031, 0.00000217, 0.00000394, 0.00000421, 0.00000320, 0.00000177, 0.00000065, -0.00000027, -0.00000189, -0.00000502, -0.00000936, -0.00001309, -0.00001359, -0.00000912, -0.00000039, 0.00000917, 0.00001482, 0.00001313, 0.00000428, -0.00000750, -0.00001586, -0.00001583, -0.00000679, 0.00000676, 0.00001743, 0.00001906, 0.00001036, -0.00000413, -0.00001647, -0.00001981, -0.00001239, 0.00000133, 0.00001325, 0.00001646, 0.00000933, -0.00000339, -0.00001350, -0.00001427, -0.00000472, 0.00000964, 0.00002000, 0.00001959, 0.00000792, -0.00000862, -0.00002036, -0.00002009, -0.00000743, 0.00001066, 0.00002370, 0.00002372, 0.00000984, -0.00001108, -0.00002822, -0.00003286, -0.00002309, -0.00000470, 0.00001224, 0.00001906, 0.00001311, -0.00000120, -0.00001533, -0.00002162, -0.00001738, -0.00000593, 0.00000585, 0.00001175, 0.00000963, 0.00000219, -0.00000514, -0.00000768, -0.00000417, 0.00000281, 0.00000867, 0.00000966, 0.00000488, -0.00000348, -0.00001173, -0.00001677, -0.00001752, -0.00001492, -0.00001078, -0.00000652, -0.00000255, 0.00000139, 0.00000532, 0.00000847, 0.00000950, 0.00000741, 0.00000250, -0.00000345, -0.00000793, -0.00000897, -0.00000632, -0.00000157, 0.00000267, 0.00000418, 0.00000222, -0.00000221, -0.00000707, },
		{ -0.00000269, 0.00000275, 0.00000725, 0.00001069, 0.00001300, 0.00001386, 0.00001284, 0.00000981, 0.00000521, 0.00000008, -0.00000437, -0.00000718, -0.00000800, -0.00000705, -0.00000482, -0.00000189, 0.00000118, 0.00000376, 0.00000517, 0.00000486, 0.00000278, -0.00000032, -0.00000306, -0.00000398, -0.00000238, 0.00000120, 0.00000507, 0.00000727, 0.00000664, 0.00000340, -0.00000097, -0.00000467, -0.00000666, -0.00000715, -0.00000714, -0.00000748, -0.00000797, -0.00000742, -0.00000457, 0.00000062, 0.00000636, 0.00000969, 0.00000823, 0.00000200, -0.00000603, -0.00001122, -0.00000994, -0.00000202, 0.00000864, 0.00001590, 0.00001492, 0.00000521, -0.00000854, -0.00001881, -0.00001952, -0.00000975, 0.00000530, 0.00001707, 0.00001854, 0.00000852, -0.00000731, -0.00001947, -0.00002019, -0.00000819, 0.00001024, 0.00002459, 0.00002617, 0.00001334, -0.00000723, -0.00002412, -0.00002763, -0.00001544, 0.00000589, 0.00002444, 0.00002962, 0.00001808, -0.00000434, -0.00002598, -0.00003577, -0.00002931, -0.00001098, 0.00000880, 0.00001962, 0.00001669, 0.00000309, -0.00001252, -0.00002116, -0.00001858, -0.00000716, 0.00000598, 0.00001352, 0.00001211, 0.00000396, -0.00000501, -0.00000900, -0.00000578, 0.00000222, 0.00000966, 0.00001156, 0.00000617, -0.00000428, -0.00001509, -0.00002171, -0.00002203, -0.00001690, -0.00000911, -0.00000157, 0.00000405, 0.00000748, 0.00000914, 0.00000922, 0.00000738, 0.00000329, -0.00000255, -0.00000845, -0.00001203, -0.00001145, -0.00000661, 0.00000052, 0.00000670, 0.00000903, 0.00000635, -0.00000018, -0.00000780, -0.00001347, -0.00001534, -0.00001326, -0.00000847, },
		{ 0.00001501, 0.00001378, 0.00001119, 0.00000727, 0.00000236, -0.00000267, -0.00000667, -0.00000866, -0.00000825, -0.00000582, -0.00000229, 0.00000126, 0.00000395, 0.00000523, 0.00000487, 0.00000300, 0.00000015, -0.00000271, -0.00000436, -0.00000377, -0.00000079, 0.00000360, 0.00000748, 0.00000881, 0.00000654, 0.00000125, -0.00000502, -0.00000978, -0.00001139, -0.00000986, -0.00000670, -0.00000386, -0.00000246, -0.00000225, -0.00000196, -0.00000049, 0.00000204, 0.00000421, 0.00000430, 0.00000165, -0.00000239, -0.00000506, -0.00000390, 0.00000136, 0.00000803, 0.00001173, 0.00000903, -0.00000004, -0.00001131, -0.00001848, -0.00001667, -0.00000580, 0.00000878, 0.00001885, 0.00001799, 0.00000570, -0.00001171, -0.00002424, -0.00002384, -0.00000935, 0.00001221, 0.00002911, 0.00003144, 0.00001688, -0.00000749, -0.00002874, -0.00003512, -0.00002274, 0.00000205, 0.00002596, 0.00003597, 0.00002645, 0.00000223, -0.00002413, -0.00003927, -0.00003622, -0.00001785, 0.00000518, 0.00002060, 0.00002129, 0.00000874, -0.00000849, -0.00002011, -0.00002002, -0.00000933, 0.00000493, 0.00001439, 0.00001435, 0.00000615, -0.00000411, -0.00000966, -0.00000721, 0.00000125, 0.00000989, 0.00001269, 0.00000699, -0.00000505, -0.00001787, -0.00002561, -0.00002513, -0.00001730, -0.00000593, 0.00000445, 0.00001083, 0.00001253, 0.00001066, 0.00000681, 0.00000203, -0.00000317, -0.00000830, -0.00001222, -0.00001332, -0.00001039, -0.00000366, 0.00000468, 0.00001123, 0.00001287, 0.00000845, -0.00000055, -0.00001061, -0.00001783, -0.00001965, -0.00001583, -0.00000822, 0.00000040, 0.00000769, 0.00001247, 0.00001472, },
		{ -0.00000091, -0.00000437, -0.00000686, -0.00000765, -0.00000640, -0.00000348, 0.00000012, 0.00000322, 0.00000491, 0.00000484, 0.00000322, 0.00000070, -0.00000189, -0.00000364, -0.00000378, -0.00000191, 0.00000160, 0.00000558, 0.00000824, 0.00000796, 0.00000412, -0.00000230, -0.00000900, -0.00001330, -0.00001353, -0.00000990, -0.00000443, 0.00000017, 0.00000200, 0.00000093, -0.00000146, -0.00000304, -0.00000253, -0.00000025, 0.00000221, 0.00000327, 0.00000256, 0.00000120, 0.00000092, 0.00000253, 0.00000497, 0.00000578, 0.00000276, -0.00000404, -0.00001165, -0.00001547, -0.00001205, -0.00000176, 0.00001058, 0.00001787, 0.00001485, 0.00000164, -0.00001541, -0.00002666, -0.00002451, -0.00000826, 0.00001471, 0.00003229, 0.00003414, 0.00001780, -0.00000930, -0.00003329, -0.00004117, -0.00002805, -0.00000018, 0.00002807, 0.00004178, 0.00003354, 0.00000741, -0.00002327, -0.00004313, -0.00004290, -0.00002408, 0.00000239, 0.00002248, 0.00002678, 0.00001511, -0.00000408, -0.00001920, -0.00002214, -0.00001249, 0.00000295, 0.00001483, 0.00001685, 0.00000919, -0.00000218, -0.00000964, -0.00000866, -0.00000048, 0.00000895, 0.00001276, 0.00000737, -0.00000538, -0.00001936, -0.00002764, -0.00002625, -0.00001603, -0.00000181, 0.00001039, 0.00001635, 0.00001522, 0.00000908, 0.00000122, -0.00000570, -0.00001044, -0.00001286, -0.00001301, -0.00001064, -0.00000550, 0.00000183, 0.00000942, 0.00001432, 0.00001389, 0.00000732, -0.00000354, -0.00001475, -0.00002195, -0.00002233, -0.00001589, -0.00000522, 0.00000585, 0.00001403, 0.00001778, 0.00001746, 0.00001459, 0.00001072, 0.00000675, 0.00000288, },
		{ -0.00000345, -0.00000134, 0.00000120, 0.00000323, 0.00000398, 0.00000321, 0.00000135, -0.00000082, -0.00000243, -0.00000284, -0.00000179, 0.00000055, 0.00000352, 0.00000605, 0.00000685, 0.00000497, 0.00000031, -0.00000594, -0.00001161, -0.00001440, -0.00001300, -0.00000796, -0.00000154, 0.00000330, 0.00000447, 0.00000187, -0.00000253, -0.00000587, -0.00000602, -0.00000277, 0.00000213, 0.00000615, 0.00000751, 0.00000617, 0.00000356, 0.00000137, 0.00000025, -0.00000057, -0.00000246, -0.00000588, -0.00000948, -0.00001055, -0.00000689, 0.00000114, 0.00000996, 0.00001419, 0.00000988, -0.00000249, -0.00001726, -0.00002603, -0.00002227, -0.00000568, 0.00001656, 0.00003290, 0.00003351, 0.00001610, -0.00001189, -0.00003650, -0.00004445, -0.00003048, -0.00000074, 0.00002994, 0.00004567, 0.00003793, 0.00001028, -0.00002348, -0.00004656, -0.00004802, -0.00002837, 0.00000123, 0.00002530, 0.00003244, 0.00002111, -0.00000035, -0.00001904, -0.00002494, -0.00001616, 0.00000080, 0.00001553, 0.00002006, 0.00001318, 0.00000055, -0.00000935, -0.00001062, -0.00000336, 0.00000666, 0.00001189, 0.00000774, -0.00000457, -0.00001879, -0.00002727, -0.00002530, -0.00001364, 0.00000213, 0.00001480, 0.00001924, 0.00001468, 0.00000438, -0.00000660, -0.00001413, -0.00001650, -0.00001443, -0.00000969, -0.00000384, 0.00000230, 0.00000814, 0.00001264, 0.00001409, 0.00001090, 0.00000282, -0.00000824, -0.00001850, -0.00002383, -0.00002172, -0.00001259, 0.00000032, 0.00001245, 0.00001995, 0.00002128, 0.00001748, 0.00001116, 0.00000503, 0.00000067, -0.00000179, -0.00000307, -0.00000393, -0.00000453, -0.00000451, },
		{ 0.00000189, 0.00000149, 0.00000040, -0.00000075, -0.00000128, -0.00000083, 0.00000053, 0.00000235, 0.00000395, 0.00000460, 0.00000365, 0.00000082, -0.00000357, -0.00000845, -0.00001215, -0.00001308, -0.00001050, -0.00000516, 0.00000078, 0.00000467, 0.00000473, 0.00000107, -0.00000414, -0.00000782, -0.00000759, -0.00000315, 0.00000349, 0.00000913, 0.00001113, 0.00000885, 0.00000378, -0.00000147, -0.00000484, -0.00000589, -0.00000563, -0.00000538, -0.00000549, -0.00000500, -0.00000256, 0.00000205, 0.00000689, 0.00000857, 0.00000436, -0.00000544, -0.00001655, -0.00002244, -0.00001792, -0.00000287, 0.00001654, 0.00003024, 0.00002957, 0.00001254, -0.00001402, -0.00003709, -0.00004410, -0.00002991, -0.00000029, 0.00003036, 0.00004626, 0.00003861, 0.00001056, -0.00002416, -0.00004833, -0.00005023, -0.00002972, 0.00000195, 0.00002844, 0.00003711, 0.00002548, 0.00000188, -0.00001977, -0.00002788, -0.00001936, -0.00000054, 0.00001714, 0.00002406, 0.00001770, 0.00000339, -0.00000950, -0.00001352, -0.00000747, 0.00000333, 0.00001069, 0.00000888, -0.00000192, -0.00001574, -0.00002450, -0.00002284, -0.00001121, 0.00000450, 0.00001633, 0.00001866, 0.00001094, -0.00000245, -0.00001489, -0.00002111, -0.00001944, -0.00001185, -0.00000223, 0.00000598, 0.00001107, 0.00001301, 0.00001238, 0.00000942, 0.00000392, -0.00000394, -0.00001275, -0.00001972, -0.00002173, -0.00001697, -0.00000624, 0.00000703, 0.00001815, 0.00002336, 0.00002146, 0.00001421, 0.00000524, -0.00000190, -0.00000527, -0.00000508, -0.00000304, -0.00000107, -0.00000023, -0.00000039, -0.00000074, -0.00000053, 0.00000035, 0.00000139, },
		{ 0.00000092, 0.00000129, 0.00000197, 0.00000256, 0.00000256, 0.00000162, -0.00000035, -0.00000317, -0.00000636, -0.00000917, -0.00001069, -0.00001009, -0.00000716, -0.00000265, 0.00000178, 0.00000417, 0.00000331, -0.00000050, -0.00000525, -0.00000815, -0.00000708, -0.00000191, 0.00000525, 0.00001102, 0.00001253, 0.00000899, 0.00000214, -0.00000481, -0.00000893, -0.00000911, -0.00000642, -0.00000308, -0.00000090, -0.00000021, -0.00000000, 0.00000081, 0.00000211, 0.00000234, -0.00000038, -0.00000640, -0.00001332, -0.00001676, -0.00001282, -0.00000110, 0.00001392, 0.00002440, 0.00002319, 0.00000840, -0.00001444, -0.00003428, -0.00004008, -0.00002702, -0.00000001, 0.00002807, 0.00004266, 0.00003535, 0.00000878, -0.00002423, -0.00004719, -0.00004853, -0.00002782, 0.00000406, 0.00003083, 0.00003951, 0.00002723, 0.00000227, -0.00002092, -0.00002990, -0.00002087, -0.00000019, 0.00001989, 0.00002842, 0.00002184, 0.00000536, -0.00001074, -0.00001747, -0.00001232, -0.00000017, 0.00001016, 0.00001155, 0.00000292, -0.00001040, -0.00002009, -0.00002007, -0.00001011, 0.00000407, 0.00001430, 0.00001470, 0.00000499, -0.00000963, -0.00002149, -0.00002471, -0.00001807, -0.00000513, 0.00000815, 0.00001665, 0.00001832, 0.00001433, 0.00000752, 0.00000043, -0.00000572, -0.00001082, -0.00001476, -0.00001661, -0.00001487, -0.00000853, 0.00000161, 0.00001269, 0.00002076, 0.00002268, 0.00001778, 0.00000832, -0.00000161, -0.00000809, -0.00000925, -0.00000587, -0.00000066, 0.00000343, 0.00000468, 0.00000322, 0.00000057, -0.00000156, -0.00000219, -0.00000144, -0.00000017, 0.00000078, 0.00000107, 0.00000095, },
		{ -0.00000021, -0.00000187, -0.00000389, -0.00000589, -0.00000745, -0.00000823, -0.00000797, -0.00000656, -0.00000411, -0.00000113, 0.00000139, 0.00000238, 0.00000117, -0.00000187, -0.00000522, -0.00000680, -0.00000506, -0.00000004, 0.00000634, 0.00001104, 0.00001151, 0.00000711, -0.00000034, -0.00000744, -0.00001104, -0.00000990, -0.00000527, -0.00000007, 0.00000297, 0.00000281, 0.00000047, -0.00000201, -0.00000315, -0.00000312, -0.00000338, -0.00000525, -0.00000837, -0.00001040, -0.00000842, -0.00000126, 0.00000875, 0.00001630, 0.00001580, 0.00000509, -0.00001233, -0.00002813, -0.00003324, -0.00002314, -0.00000118, 0.00002226, 0.00003479, 0.00002880, 0.00000606, -0.00002252, -0.00004230, -0.00004273, -0.00002318, 0.00000652, 0.00003123, 0.00003867, 0.00002602, 0.00000126, -0.00002146, -0.00002973, -0.00001969, 0.00000225, 0.00002338, 0.00003212, 0.00002441, 0.00000556, -0.00001332, -0.00002198, -0.00001685, -0.00000257, 0.00001128, 0.00001617, 0.00000962, -0.00000375, -0.00001543, -0.00001842, -0.00001150, 0.00000030, 0.00000896, 0.00000851, -0.00000135, -0.00001505, -0.00002451, -0.00002380, -0.00001246, 0.00000443, 0.00001916, 0.00002555, 0.00002183, 0.00001091, -0.00000177, -0.00001135, -0.00001565, -0.00001531, -0.00001236, -0.00000842, -0.00000383, 0.00000184, 0.00000857, 0.00001499, 0.00001867, 0.00001744, 0.00001101, 0.00000152, -0.00000722, -0.00001161, -0.00001013, -0.00000409, 0.00000316, 0.00000805, 0.00000858, 0.00000517, 0.00000008, -0.00000400, -0.00000537, -0.00000399, -0.00000112, 0.00000164, 0.00000324, 0.00000354, 0.00000305, 0.00000235, 0.00000170, 0.00000094, },
		{ -0.00000699, -0.00000635, -0.00000511, -0.00000356, -0.00000199, -0.00000065, 0.00000020, 0.00000023, -0.00000073, -0.00000244, -0.00000406, -0.00000439, -0.00000252, 0.00000145, 0.00000609, 0.00000913, 0.00000863, 0.00000420, -0.00000252, -0.00000850, -0.00001090, -0.00000865, -0.00000315, 0.00000247, 0.00000517, 0.00000370, -0.00000072, -0.00000532, -0.00000756, -0.00000676, -0.00000434, -0.00000263, -0.00000305, -0.00000488, -0.00000576, -0.00000350, 0.00000187, 0.00000749, 0.00000902, 0.00000361, -0.00000768, -0.00001958, -0.00002512, -0.00001976, -0.00000469, 0.00001297, 0.00002356, 0.00002034, 0.00000373, -0.00001820, -0.00003365, -0.00003360, -0.00001704, 0.00000804, 0.00002876, 0.00003438, 0.00002240, -0.00000004, -0.00002007, -0.00002632, -0.00001539, 0.00000638, 0.00002658, 0.00003387, 0.00002434, 0.00000357, -0.00001682, -0.00002591, -0.00001962, -0.00000266, 0.00001462, 0.00002242, 0.00001706, 0.00000258, -0.00001216, -0.00001912, -0.00001579, -0.00000629, 0.00000169, 0.00000202, -0.00000595, -0.00001692, -0.00002298, -0.00001856, -0.00000398, 0.00001452, 0.00002815, 0.00003038, 0.00002032, 0.00000294, -0.00001386, -0.00002337, -0.00002313, -0.00001533, -0.00000476, 0.00000416, 0.00000941, 0.00001153, 0.00001215, 0.00001226, 0.00001147, 0.00000859, 0.00000308, -0.00000388, -0.00000961, -0.00001134, -0.00000792, -0.00000076, 0.00000677, 0.00001109, 0.00001029, 0.00000508, -0.00000169, -0.00000669, -0.00000783, -0.00000515, -0.00000047, 0.00000382, 0.00000603, 0.00000582, 0.00000398, 0.00000164, -0.00000038, -0.00000195, -0.00000330, -0.00000466, -0.00000595, -0.00000683, },
		{ -0.00000084, -0.00000066, -0.00000092, -0.00000135, -0.00000181, -0.00000221, -0.00000236, -0.00000189, -0.00000045, 0.00000192, 0.00000447, 0.00000587, 0.00000492, 0.00000135, -0.00000363, -0.00000778, -0.00000892, -0.00000627, -0.00000113, 0.00000369, 0.00000538, 0.00000282, -0.00000264, -0.00000795, -0.00001016, -0.00000824, -0.00000372, 0.00000031, 0.00000123, -0.00000133, -0.00000520, -0.00000717, -0.00000525, -0.00000036, 0.00000407, 0.00000423, -0.00000134, -0.00001029, -0.00001749, -0.00001806, -0.00001065, 0.00000118, 0.00001064, 0.00001166, 0.00000283, -0.00001126, -0.00002221, -0.00002269, -0.00001096, 0.00000761, 0.00002326, 0.00002738, 0.00001768, -0.00000021, -0.00001569, -0.00001934, -0.00000845, 0.00001112, 0.00002817, 0.00003256, 0.00002116, -0.00000024, -0.00002012, -0.00002775, -0.00001927, 0.00000028, 0.00001995, 0.00002915, 0.00002348, 0.00000679, -0.00001159, -0.00002256, -0.00002234, -0.00001416, -0.00000541, -0.00000262, -0.00000715, -0.00001447, -0.00001728, -0.00001051, 0.00000501, 0.00002249, 0.00003285, 0.00002990, 0.00001399, -0.00000797, -0.00002614, -0.00003277, -0.00002594, -0.00000997, 0.00000736, 0.00001904, 0.00002204, 0.00001784, 0.00001057, 0.00000415, 0.00000026, -0.00000177, -0.00000355, -0.00000580, -0.00000767, -0.00000736, -0.00000378, 0.00000233, 0.00000838, 0.00001132, 0.00000947, 0.00000359, -0.00000343, -0.00000815, -0.00000844, -0.00000450, 0.00000139, 0.00000629, 0.00000815, 0.00000666, 0.00000309, -0.00000076, -0.00000357, -0.00000502, -0.00000558, -0.00000587, -0.00000615, -0.00000622, -0.00000572, -0.00000456, -0.00000303, -0.00000166, },
		{ -0.00000213, -0.00000176, -0.00000104, -0.00000024, 0.00000055, 0.00000136, 0.00000208, 0.00000229, 0.00000143, -0.00000069, -0.00000352, -0.00000576, -0.00000603, -0.00000379, 0.00000004, 0.00000336, 0.00000398, 0.00000103, -0.00000426, -0.00000910, -0.00001067, -0.00000795, -0.00000249, 0.00000236, 0.00000344, -0.00000014, -0.00000628, -0.00001104, -0.00001108, -0.00000595, 0.00000143, 0.00000634, 0.00000520, -0.00000209, -0.00001179, -0.00001847, -0.00001827, -0.00001138, -0.00000200, 0.00000420, 0.00000367, -0.00000263, -0.00000978, -0.00001196, -0.00000632, 0.00000492, 0.00001557, 0.00001926, 0.00001360, 0.00000202, -0.00000799, -0.00000938, -0.00000016, 0.00001495, 0.00002695, 0.00002775, 0.00001533, -0.00000465, -0.00002164, -0.00002610, -0.00001503, 0.00000607, 0.00002614, 0.00003457, 0.00002700, 0.00000748, -0.00001418, -0.00002804, -0.00002947, -0.00002102, -0.00001009, -0.00000376, -0.00000435, -0.00000832, -0.00000913, -0.00000208, 0.00001190, 0.00002619, 0.00003214, 0.00002437, 0.00000449, -0.00001919, -0.00003572, -0.00003724, -0.00002305, 0.00000013, 0.00002186, 0.00003326, 0.00003115, 0.00001882, 0.00000341, -0.00000818, -0.00001264, -0.00001093, -0.00000662, -0.00000309, -0.00000157, -0.00000103, 0.00000033, 0.00000328, 0.00000674, 0.00000851, 0.00000690, 0.00000212, -0.00000359, -0.00000724, -0.00000682, -0.00000254, 0.00000329, 0.00000765, 0.00000843, 0.00000552, 0.00000060, -0.00000401, -0.00000667, -0.00000713, -0.00000628, -0.00000531, -0.00000489, -0.00000485, -0.00000458, -0.00000364, -0.00000218, -0.00000080, -0.00000014, -0.00000039, -0.00000118, -0.00000192, },
		{ 0.00000049, 0.00000032, -0.00000015, -0.00000063, -0.00000112, -0.00000178, -0.00000263, -0.00000332, -0.00000321, -0.00000191, 0.00000019, 0.00000191, 0.00000181, -0.00000075, -0.00000492, -0.00000859, -0.00000950, -0.00000675, -0.00000167, 0.00000269, 0.00000328, -0.00000093, -0.00000799, -0.00001381, -0.00001447, -0.00000877, 0.00000071, 0.00000872, 0.00001025, 0.00000353, -0.00000870, -0.00002052, -0.00002604, -0.00002270, -0.00001263, -0.00000133, 0.00000567, 0.00000598, 0.00000149, -0.00000317, -0.00000380, 0.00000061, 0.00000736, 0.00001205, 0.00001174, 0.00000720, 0.00000245, 0.00000206, 0.00000771, 0.00001651, 0.00002242, 0.00001998, 0.00000820, -0.00000796, -0.00001991, -0.00002019, -0.00000714, 0.00001353, 0.00003139, 0.00003681, 0.00002626, 0.00000423, -0.00001917, -0.00003382, -0.00003482, -0.00002461, -0.00001075, -0.00000092, 0.00000165, -0.00000039, -0.00000111, 0.00000408, 0.00001464, 0.00002467, 0.00002643, 0.00001551, -0.00000559, -0.00002790, -0.00004033, -0.00003575, -0.00001505, 0.00001292, 0.00003578, 0.00004381, 0.00003465, 0.00001392, -0.00000819, -0.00002231, -0.00002431, -0.00001641, -0.00000485, 0.00000404, 0.00000716, 0.00000547, 0.00000232, 0.00000078, 0.00000168, 0.00000357, 0.00000418, 0.00000237, -0.00000099, -0.00000366, -0.00000363, -0.00000060, 0.00000372, 0.00000669, 0.00000639, 0.00000274, -0.00000246, -0.00000676, -0.00000848, -0.00000757, -0.00000537, -0.00000356, -0.00000308, -0.00000366, -0.00000423, -0.00000382, -0.00000230, -0.00000041, 0.00000081, 0.00000076, -0.00000033, -0.00000161, -0.00000222, -0.00000186, -0.00000086, 0.00000009, },
		{ -0.00000263, -0.00000231, -0.00000177, -0.00000136, -0.00000112, -0.00000079, -0.00000027, 0.00000014, -0.00000024, -0.00000190, -0.00000454, -0.00000693, -0.00000754, -0.00000561, -0.00000194, 0.00000117, 0.00000124, -0.00000273, -0.00000921, -0.00001458, -0.00001506, -0.00000917, 0.00000090, 0.00000998, 0.00001254, 0.00000588, -0.00000800, -0.00002306, -0.00003222, -0.00003109, -0.00002029, -0.00000511, 0.00000747, 0.00001255, 0.00000967, 0.00000262, -0.00000315, -0.00000385, 0.00000071, 0.00000760, 0.00001302, 0.00001494, 0.00001411, 0.00001300, 0.00001348, 0.00001508, 0.00001510, 0.00001074, 0.00000172, -0.00000851, -0.00001411, -0.00001033, 0.00000309, 0.00002074, 0.00003379, 0.00003457, 0.00002095, -0.00000207, -0.00002473, -0.00003754, -0.00003619, -0.00002343, -0.00000707, 0.00000491, 0.00000877, 0.00000663, 0.00000412, 0.00000602, 0.00001246, 0.00001855, 0.00001759, 0.00000597, -0.00001347, -0.00003192, -0.00003903, -0.00002888, -0.00000394, 0.00002549, 0.00004615, 0.00004856, 0.00003187, 0.00000417, -0.00002181, -0.00003521, -0.00003212, -0.00001676, 0.00000162, 0.00001402, 0.00001617, 0.00000983, 0.00000070, -0.00000539, -0.00000576, -0.00000170, 0.00000318, 0.00000561, 0.00000464, 0.00000183, -0.00000035, -0.00000034, 0.00000140, 0.00000298, 0.00000251, -0.00000045, -0.00000459, -0.00000775, -0.00000836, -0.00000648, -0.00000363, -0.00000171, -0.00000175, -0.00000327, -0.00000472, -0.00000464, -0.00000267, 0.00000024, 0.00000248, 0.00000284, 0.00000131, -0.00000105, -0.00000281, -0.00000310, -0.00000211, -0.00000077, -0.00000008, -0.00000041, -0.00000141, -0.00000233, },
		{ 0.00000015, -0.00000006, -0.00000062, -0.00000114, -0.00000162, -0.00000238, -0.00000362, -0.00000501, -0.00000573, -0.00000511, -0.00000331, -0.00000159, -0.00000168, -0.00000450, -0.00000922, -0.00001319, -0.00001334, -0.00000815, 0.00000080, 0.00000911, 0.00001158, 0.00000521, -0.00000872, -0.00002468, -0.00003542, -0.00003565, -0.00002496, -0.00000808, 0.00000747, 0.00001535, 0.00001358, 0.00000529, -0.00000335, -0.00000668, -0.00000251, 0.00000706, 0.00001736, 0.00002397, 0.00002501, 0.00002155, 0.00001616, 0.00001096, 0.00000649, 0.00000217, -0.00000219, -0.00000526, -0.00000449, 0.00000205, 0.00001355, 0.00002565, 0.00003204, 0.00002771, 0.00001219, -0.00000941, -0.00002844, -0.00003706, -0.00003225, -0.00001739, -0.00000023, 0.00001150, 0.00001425, 0.00001011, 0.00000474, 0.00000314, 0.00000616, 0.00000984, 0.00000834, -0.00000152, -0.00001711, -0.00003045, -0.00003248, -0.00001865, 0.00000746, 0.00003497, 0.00005086, 0.00004679, 0.00002376, -0.00000809, -0.00003446, -0.00004415, -0.00003444, -0.00001197, 0.00001127, 0.00002439, 0.00002293, 0.00001035, -0.00000483, -0.00001415, -0.00001364, -0.00000520, 0.00000535, 0.00001215, 0.00001245, 0.00000753, 0.00000116, -0.00000313, -0.00000408, -0.00000288, -0.00000182, -0.00000242, -0.00000445, -0.00000633, -0.00000653, -0.00000480, -0.00000233, -0.00000089, -0.00000149, -0.00000357, -0.00000538, -0.00000516, -0.00000245, 0.00000156, 0.00000472, 0.00000526, 0.00000290, -0.00000100, -0.00000433, -0.00000552, -0.00000440, -0.00000215, -0.00000041, -0.00000019, -0.00000134, -0.00000281, -0.00000350, -0.00000297, -0.00000165, -0.00000039, },
		{ -0.00000235, -0.00000269, -0.00000301, -0.00000368, -0.00000460, -0.00000523, -0.00000513, -0.00000456, -0.00000440, -0.00000556, -0.00000799, -0.00001031, -0.00001042, -0.00000690, -0.00000044, 0.00000587, 0.00000783, 0.00000250, -0.00000965, -0.00002431, -0.00003504, -0.00003650, -0.00002742, -0.00001149, 0.00000433, 0.00001346, 0.00001308, 0.00000557, -0.00000301, -0.00000632, -0.00000126, 0.00001054, 0.00002379, 0.00003254, 0.00003333, 0.00002656, 0.00001580, 0.00000546, -0.00000132, -0.00000364, -0.00000225, 0.00000171, 0.00000754, 0.00001472, 0.00002200, 0.00002685, 0.00002602, 0.00001748, 0.00000221, -0.00001519, -0.00002815, -0.00003125, -0.00002324, -0.00000802, 0.00000728, 0.00001601, 0.00001555, 0.00000842, 0.00000036, -0.00000353, -0.00000210, 0.00000130, 0.00000137, -0.00000501, -0.00001581, -0.00002426, -0.00002273, -0.00000783, 0.00001637, 0.00003935, 0.00004926, 0.00003942, 0.00001254, -0.00001995, -0.00004343, -0.00004745, -0.00003113, -0.00000343, 0.00002168, 0.00003254, 0.00002558, 0.00000653, -0.00001333, -0.00002351, -0.00001968, -0.00000529, 0.00001115, 0.00002113, 0.00002060, 0.00001140, -0.00000064, -0.00000947, -0.00001206, -0.00000935, -0.00000477, -0.00000168, -0.00000136, -0.00000282, -0.00000411, -0.00000396, -0.00000271, -0.00000172, -0.00000213, -0.00000375, -0.00000509, -0.00000443, -0.00000122, 0.00000331, 0.00000676, 0.00000699, 0.00000357, -0.00000189, -0.00000666, -0.00000848, -0.00000687, -0.00000325, 0.00000001, 0.00000110, -0.00000022, -0.00000265, -0.00000439, -0.00000430, -0.00000256, -0.00000036, 0.00000100, 0.00000092, -0.00000021, -0.00000152, },
		{ -0.00000408, -0.00000535, -0.00000643, -0.00000671, -0.00000628, -0.00000588, -0.00000619, -0.00000710, -0.00000754, -0.00000625, -0.00000295, 0.00000090, 0.00000244, -0.00000097, -0.00000985, -0.00002163, -0.00003147, -0.00003455, -0.00002881, -0.00001632, -0.00000243, 0.00000703, 0.00000893, 0.00000460, -0.00000099, -0.00000208, 0.00000445, 0.00001719, 0.00003084, 0.00003910, 0.00003801, 0.00002799, 0.00001347, 0.00000047, -0.00000627, -0.00000528, 0.00000172, 0.00001122, 0.00001983, 0.00002536, 0.00002687, 0.00002409, 0.00001702, 0.00000631, -0.00000619, -0.00001717, -0.00002281, -0.00002059, -0.00001104, 0.00000192, 0.00001249, 0.00001597, 0.00001123, 0.00000147, -0.00000774, -0.00001172, -0.00000954, -0.00000445, -0.00000152, -0.00000397, -0.00001046, -0.00001543, -0.00001252, 0.00000089, 0.00002089, 0.00003807, 0.00004229, 0.00002864, 0.00000105, -0.00002880, -0.00004704, -0.00004490, -0.00002347, 0.00000651, 0.00003024, 0.00003652, 0.00002345, -0.00000082, -0.00002283, -0.00003111, -0.00002204, -0.00000131, 0.00001981, 0.00003069, 0.00002687, 0.00001168, -0.00000638, -0.00001868, -0.00002085, -0.00001437, -0.00000467, 0.00000252, 0.00000430, 0.00000151, -0.00000271, -0.00000536, -0.00000543, -0.00000397, -0.00000271, -0.00000253, -0.00000278, -0.00000198, 0.00000071, 0.00000454, 0.00000736, 0.00000701, 0.00000287, -0.00000346, -0.00000895, -0.00001090, -0.00000857, -0.00000357, 0.00000111, 0.00000290, 0.00000118, -0.00000255, -0.00000571, -0.00000634, -0.00000418, -0.00000070, 0.00000201, 0.00000259, 0.00000111, -0.00000120, -0.00000296, -0.00000355, -0.00000338, -0.00000336, },
		{ -0.00000733, -0.00000612, -0.00000495, -0.00000477, -0.00000556, -0.00000637, -0.00000607, -0.00000458, -0.00000323, -0.00000422, -0.00000906, -0.00001723, -0.00002587, -0.00003108, -0.00003006, -0.00002282, -0.00001233, -0.00000290, 0.00000234, 0.00000332, 0.00000306, 0.00000567, 0.00001356, 0.00002558, 0.00003719, 0.00004286, 0.00003916, 0.00002687, 0.00001089, -0.00000214, -0.00000705, -0.00000249, 0.00000873, 0.00002133, 0.00003018, 0.00003236, 0.00002780, 0.00001855, 0.00000745, -0.00000295, -0.00001066, -0.00001428, -0.00001306, -0.00000726, 0.00000124, 0.00000918, 0.00001288, 0.00001013, 0.00000158, -0.00000911, -0.00001702, -0.00001856, -0.00001366, -0.00000580, 0.00000002, 0.00000062, -0.00000314, -0.00000663, -0.00000443, 0.00000573, 0.00002052, 0.00003205, 0.00003207, 0.00001717, -0.00000818, -0.00003299, -0.00004507, -0.00003768, -0.00001370, 0.00001534, 0.00003499, 0.00003554, 0.00001720, -0.00000985, -0.00003100, -0.00003501, -0.00001992, 0.00000613, 0.00002952, 0.00003853, 0.00002936, 0.00000756, -0.00001550, -0.00002903, -0.00002824, -0.00001609, -0.00000065, 0.00000979, 0.00001112, 0.00000483, -0.00000389, -0.00000972, -0.00001022, -0.00000652, -0.00000177, 0.00000131, 0.00000208, 0.00000184, 0.00000229, 0.00000389, 0.00000538, 0.00000479, 0.00000108, -0.00000473, -0.00000992, -0.00001167, -0.00000894, -0.00000326, 0.00000213, 0.00000420, 0.00000196, -0.00000294, -0.00000734, -0.00000848, -0.00000572, -0.00000075, 0.00000354, 0.00000488, 0.00000291, -0.00000084, -0.00000412, -0.00000546, -0.00000491, -0.00000380, -0.00000354, -0.00000463, -0.00000636, -0.00000752, },
		{ -0.00000462, -0.00000668, -0.00000870, -0.00000923, -0.00000812, -0.00000691, -0.00000784, -0.00001232, -0.00001967, -0.00002718, -0.00003147, -0.00003035, -0.00002404, -0.00001485, -0.00000563, 0.00000200, 0.00000847, 0.00001553, 0.00002445, 0.00003439, 0.00004225, 0.00004418, 0.00003797, 0.00002490, 0.00000974, -0.00000128, -0.00000349, 0.00000383, 0.00001708, 0.00003011, 0.00003716, 0.00003541, 0.00002588, 0.00001251, 0.00000004, -0.00000796, -0.00001013, -0.00000720, -0.00000121, 0.00000538, 0.00001015, 0.00001115, 0.00000732, -0.00000094, -0.00001141, -0.00002047, -0.00002451, -0.00002166, -0.00001309, -0.00000270, 0.00000482, 0.00000664, 0.00000358, -0.00000025, -0.00000006, 0.00000627, 0.00001619, 0.00002335, 0.00002122, 0.00000747, -0.00001356, -0.00003227, -0.00003863, -0.00002796, -0.00000427, 0.00002108, 0.00003509, 0.00003015, 0.00000856, -0.00001831, -0.00003590, -0.00003430, -0.00001375, 0.00001535, 0.00003801, 0.00004263, 0.00002705, -0.00000054, -0.00002630, -0.00003822, -0.00003217, -0.00001342, 0.00000701, 0.00001867, 0.00001709, 0.00000537, -0.00000853, -0.00001676, -0.00001581, -0.00000771, 0.00000207, 0.00000833, 0.00000904, 0.00000579, 0.00000195, 0.00000007, 0.00000029, 0.00000072, -0.00000067, -0.00000427, -0.00000832, -0.00001016, -0.00000820, -0.00000323, 0.00000187, 0.00000394, 0.00000164, -0.00000362, -0.00000846, -0.00000971, -0.00000633, -0.00000015, 0.00000533, 0.00000706, 0.00000428, -0.00000120, -0.00000621, -0.00000830, -0.00000708, -0.00000423, -0.00000219, -0.00000244, -0.00000467, -0.00000717, -0.00000820, -0.00000722, -0.00000526, -0.00000403, },
		{ -0.00001192, -0.00000944, -0.00000723, -0.00000826, -0.00001402, -0.00002329, -0.00003251, -0.00003760, -0.00003594, -0.00002750, -0.00001453, -0.00000002, 0.00001370, 0.00002565, 0.00003557, 0.00004296, 0.00004646, 0.00004443, 0.00003628, 0.00002377, 0.00001103, 0.00000308, 0.00000342, 0.00001204, 0.00002506, 0.00003642, 0.00004067, 0.00003563, 0.00002322, 0.00000845, -0.00000311, -0.00000785, -0.00000539, 0.00000167, 0.00000931, 0.00001379, 0.00001300, 0.00000675, -0.00000345, -0.00001494, -0.00002458, -0.00002945, -0.00002781, -0.00001985, -0.00000808, 0.00000342, 0.00001063, 0.00001160, 0.00000755, 0.00000235, 0.00000033, 0.00000347, 0.00000984, 0.00001440, 0.00001203, 0.00000104, -0.00001481, -0.00002768, -0.00002976, -0.00001809, 0.00000288, 0.00002286, 0.00003098, 0.00002196, -0.00000031, -0.00002426, -0.00003650, -0.00002928, -0.00000510, 0.00002412, 0.00004321, 0.00004185, 0.00002022, -0.00001102, -0.00003645, -0.00004404, -0.00003139, -0.00000653, 0.00001685, 0.00002702, 0.00002023, 0.00000205, -0.00001642, -0.00002514, -0.00002033, -0.00000593, 0.00000954, 0.00001812, 0.00001673, 0.00000809, -0.00000177, -0.00000752, -0.00000749, -0.00000391, -0.00000074, -0.00000069, -0.00000352, -0.00000667, -0.00000739, -0.00000487, -0.00000088, 0.00000157, 0.00000048, -0.00000361, -0.00000785, -0.00000898, -0.00000560, 0.00000074, 0.00000641, 0.00000795, 0.00000426, -0.00000264, -0.00000891, -0.00001124, -0.00000883, -0.00000377, 0.00000045, 0.00000115, -0.00000189, -0.00000651, -0.00000977, -0.00000993, -0.00000752, -0.00000483, -0.00000416, -0.00000626, -0.00000974, -0.00001217, },
		{ -0.00000935, -0.00001909, -0.00003219, -0.00004316, -0.00004672, -0.00004024, -0.00002473, -0.00000407, 0.00001691, 0.00003431, 0.00004603, 0.00005153, 0.00005112, 0.00004550, 0.00003589, 0.00002455, 0.00001472, 0.00000987, 0.00001205, 0.00002057, 0.00003179, 0.00004044, 0.00004205, 0.00003515, 0.00002214, 0.00000813, -0.00000150, -0.00000368, 0.00000103, 0.00000891, 0.00001500, 0.00001541, 0.00000884, -0.00000312, -0.00001692, -0.00002850, -0.00003460, -0.00003361, -0.00002589, -0.00001365, -0.00000037, 0.00001013, 0.00001500, 0.00001355, 0.00000769, 0.00000120, -0.00000219, -0.00000084, 0.00000366, 0.00000725, 0.00000584, -0.00000184, -0.00001284, -0.00002105, -0.00002065, -0.00000995, 0.00000681, 0.00002098, 0.00002411, 0.00001303, -0.00000749, -0.00002661, -0.00003299, -0.00002136, 0.00000391, 0.00003038, 0.00004390, 0.00003633, 0.00001032, -0.00002164, -0.00004371, -0.00004510, -0.00002586, 0.00000324, 0.00002670, 0.00003266, 0.00001921, -0.00000507, -0.00002623, -0.00003280, -0.00002189, -0.00000021, 0.00002027, 0.00002915, 0.00002312, 0.00000703, -0.00000962, -0.00001841, -0.00001646, -0.00000722, 0.00000244, 0.00000673, 0.00000411, -0.00000246, -0.00000813, -0.00000953, -0.00000670, -0.00000256, -0.00000040, -0.00000150, -0.00000433, -0.00000584, -0.00000395, 0.00000080, 0.00000539, 0.00000642, 0.00000245, -0.00000473, -0.00001110, -0.00001290, -0.00000906, -0.00000194, 0.00000405, 0.00000522, 0.00000081, -0.00000656, -0.00001265, -0.00001429, -0.00001128, -0.00000636, -0.00000329, -0.00000431, -0.00000869, -0.00001335, -0.00001503, -0.00001265, -0.00000827, -0.00000597, },
		{ -0.00005587, -0.00005322, -0.00003714, -0.00001158, 0.00001660, 0.00004055, 0.00005578, 0.00006107, 0.00005789, 0.00004910, 0.00003792, 0.00002732, 0.00001991, 0.00001758, 0.00002094, 0.00002870, 0.00003767, 0.00004371, 0.00004350, 0.00003619, 0.00002407, 0.00001161, 0.00000338, 0.00000173, 0.00000561, 0.00001115, 0.00001356, 0.00000954, -0.00000118, -0.00001575, -0.00002956, -0.00003819, -0.00003902, -0.00003203, -0.00001948, -0.00000504, 0.00000743, 0.00001490, 0.00001608, 0.00001169, 0.00000434, -0.00000246, -0.00000578, -0.00000464, -0.00000060, 0.00000296, 0.00000282, -0.00000201, -0.00000924, -0.00001431, -0.00001302, -0.00000447, 0.00000768, 0.00001667, 0.00001634, 0.00000520, -0.00001184, -0.00002540, -0.00002661, -0.00001254, 0.00001125, 0.00003284, 0.00004004, 0.00002733, -0.00000048, -0.00003010, -0.00004652, -0.00004116, -0.00001677, 0.00001376, 0.00003427, 0.00003409, 0.00001387, -0.00001479, -0.00003585, -0.00003766, -0.00001926, 0.00000930, 0.00003279, 0.00003922, 0.00002617, 0.00000156, -0.00002125, -0.00003100, -0.00002444, -0.00000728, 0.00000981, 0.00001758, 0.00001311, 0.00000057, -0.00001202, -0.00001790, -0.00001518, -0.00000712, 0.00000068, 0.00000407, 0.00000264, -0.00000078, -0.00000268, -0.00000161, 0.00000096, 0.00000196, -0.00000065, -0.00000613, -0.00001112, -0.00001193, -0.00000721, 0.00000081, 0.00000739, 0.00000821, 0.00000212, -0.00000793, -0.00001661, -0.00001938, -0.00001532, -0.00000757, -0.00000137, -0.00000074, -0.00000595, -0.00001344, -0.00001822, -0.00001720, -0.00001139, -0.00000552, -0.00000531, -0.00001388, -0.00002952, -0.00004610, },
		{ 0.00001219, 0.00004431, 0.00006542, 0.00007268, 0.00006798, 0.00005612, 0.00004251, 0.00003144, 0.00002539, 0.00002505, 0.00002960, 0.00003700, 0.00004435, 0.00004850, 0.00004716, 0.00003991, 0.00002866, 0.00001693, 0.00000831, 0.00000465, 0.00000505, 0.00000635, 0.00000470, -0.00000225, -0.00001406, -0.00002749, -0.00003798, -0.00004169, -0.00003712, -0.00002563, -0.00001076, 0.00000319, 0.00001270, 0.00001598, 0.00001330, 0.00000665, -0.00000100, -0.00000669, -0.00000855, -0.00000647, -0.00000220, 0.00000146, 0.00000210, -0.00000088, -0.00000567, -0.00000892, -0.00000770, -0.00000158, 0.00000649, 0.00001153, 0.00000936, -0.00000040, -0.00001323, -0.00002161, -0.00001913, -0.00000475, 0.00001558, 0.00003129, 0.00003273, 0.00001689, -0.00000995, -0.00003478, -0.00004449, -0.00003323, -0.00000616, 0.00002272, 0.00003790, 0.00003089, 0.00000525, -0.00002505, -0.00004305, -0.00003820, -0.00001232, 0.00002124, 0.00004482, 0.00004604, 0.00002450, -0.00000822, -0.00003508, -0.00004289, -0.00002916, -0.00000287, 0.00002110, 0.00003030, 0.00002125, 0.00000046, -0.00001999, -0.00002956, -0.00002476, -0.00001011, 0.00000535, 0.00001375, 0.00001247, 0.00000466, -0.00000371, -0.00000801, -0.00000748, -0.00000481, -0.00000350, -0.00000493, -0.00000739, -0.00000766, -0.00000377, 0.00000302, 0.00000858, 0.00000851, 0.00000127, -0.00001031, -0.00002040, -0.00002358, -0.00001822, -0.00000773, 0.00000142, 0.00000355, -0.00000270, -0.00001347, -0.00002204, -0.00002307, -0.00001610, -0.00000619, -0.00000120, -0.00000710, -0.00002423, -0.00004643, -0.00006373, -0.00006708, -0.00005264, -0.00002343, },
		{ 0.00008162, 0.00006641, 0.00004915, 0.00003620, 0.00003064, 0.00003232, 0.00003891, 0.00004720, 0.00005397, 0.00005668, 0.00005387, 0.00004560, 0.00003354, 0.00002048, 0.00000931, 0.00000174, -0.00000247, -0.00000535, -0.00000951, -0.00001655, -0.00002590, -0.00003492, -0.00004003, -0.00003855, -0.00003009, -0.00001682, -0.00000263, 0.00000846, 0.00001382, 0.00001299, 0.00000750, 0.00000012, -0.00000628, -0.00000963, -0.00000924, -0.00000590, -0.00000148, 0.00000179, 0.00000237, 0.00000018, -0.00000325, -0.00000548, -0.00000457, -0.00000048, 0.00000454, 0.00000697, 0.00000418, -0.00000355, -0.00001241, -0.00001675, -0.00001230, 0.00000067, 0.00001653, 0.00002658, 0.00002374, 0.00000713, -0.00001640, -0.00003510, -0.00003842, -0.00002320, 0.00000370, 0.00002832, 0.00003696, 0.00002388, -0.00000474, -0.00003356, -0.00004612, -0.00003403, -0.00000216, 0.00003343, 0.00005392, 0.00004792, 0.00001791, -0.00002091, -0.00004867, -0.00005154, -0.00002901, 0.00000607, 0.00003480, 0.00004245, 0.00002618, -0.00000411, -0.00003180, -0.00004287, -0.00003310, -0.00000948, 0.00001456, 0.00002699, 0.00002338, 0.00000819, -0.00000891, -0.00001915, -0.00001924, -0.00001213, -0.00000386, 0.00000074, 0.00000088, -0.00000064, -0.00000041, 0.00000252, 0.00000571, 0.00000516, -0.00000141, -0.00001224, -0.00002203, -0.00002506, -0.00001905, -0.00000706, 0.00000386, 0.00000687, -0.00000042, -0.00001414, -0.00002620, -0.00002919, -0.00002105, -0.00000690, 0.00000346, 0.00000101, -0.00001706, -0.00004510, -0.00007109, -0.00008232, -0.00007132, -0.00003942, 0.00000408, 0.00004613, 0.00007549, 0.00008681, },
		{ 0.00003656, 0.00004097, 0.00005092, 0.00006134, 0.00006802, 0.00006856, 0.00006237, 0.00005042, 0.00003477, 0.00001810, 0.00000303, -0.00000862, -0.00001642, -0.00002138, -0.00002517, -0.00002911, -0.00003321, -0.00003605, -0.00003557, -0.00003034, -0.00002062, -0.00000862, 0.00000235, 0.00000926, 0.00001061, 0.00000697, 0.00000056, -0.00000577, -0.00000970, -0.00001020, -0.00000765, -0.00000350, 0.00000043, 0.00000261, 0.00000242, 0.00000031, -0.00000231, -0.00000374, -0.00000292, -0.00000012, 0.00000289, 0.00000375, 0.00000096, -0.00000482, -0.00001057, -0.00001226, -0.00000731, 0.00000331, 0.00001471, 0.00002020, 0.00001501, -0.00000030, -0.00001914, -0.00003163, -0.00002994, -0.00001320, 0.00001097, 0.00002978, 0.00003202, 0.00001481, -0.00001386, -0.00003852, -0.00004442, -0.00002598, 0.00000917, 0.00004340, 0.00005820, 0.00004434, 0.00000749, -0.00003417, -0.00005933, -0.00005500, -0.00002357, 0.00001825, 0.00004842, 0.00005136, 0.00002616, -0.00001323, -0.00004589, -0.00005526, -0.00003773, -0.00000389, 0.00002795, 0.00004193, 0.00003276, 0.00000752, -0.00001933, -0.00003441, -0.00003234, -0.00001736, 0.00000049, 0.00001185, 0.00001313, 0.00000725, 0.00000035, -0.00000291, -0.00000233, -0.00000148, -0.00000427, -0.00001141, -0.00001938, -0.00002269, -0.00001800, -0.00000701, 0.00000387, 0.00000734, -0.00000007, -0.00001512, -0.00002928, -0.00003364, -0.00002461, -0.00000694, 0.00000818, 0.00000904, -0.00000977, -0.00004349, -0.00007852, -0.00009852, -0.00009211, -0.00005826, -0.00000674, 0.00004638, 0.00008574, 0.00010272, 0.00009786, 0.00007913, 0.00005743, 0.00004180, },
		{ 0.00008605, 0.00008205, 0.00006916, 0.00005005, 0.00002810, 0.00000658, -0.00001185, -0.00002550, -0.00003382, -0.00003742, -0.00003774, -0.00003631, -0.00003408, -0.00003104, -0.00002656, -0.00002009, -0.00001195, -0.00000360, 0.00000288, 0.00000571, 0.00000439, -0.00000003, -0.00000537, -0.00000929, -0.00001034, -0.00000839, -0.00000456, -0.00000054, 0.00000218, 0.00000283, 0.00000161, -0.00000057, -0.00000244, -0.00000297, -0.00000185, 0.00000030, 0.00000205, 0.00000192, -0.00000078, -0.00000514, -0.00000882, -0.00000908, -0.00000452, 0.00000360, 0.00001132, 0.00001378, 0.00000802, -0.00000466, -0.00001847, -0.00002571, -0.00002100, -0.00000504, 0.00001472, 0.00002740, 0.00002459, 0.00000575, -0.00002025, -0.00003908, -0.00003851, -0.00001586, 0.00001929, 0.00004915, 0.00005687, 0.00003609, -0.00000461, -0.00004532, -0.00006491, -0.00005250, -0.00001381, 0.00003131, 0.00005918, 0.00005494, 0.00002068, -0.00002562, -0.00005968, -0.00006395, -0.00003685, 0.00000673, 0.00004379, 0.00005572, 0.00003784, 0.00000111, -0.00003463, -0.00005173, -0.00004377, -0.00001779, 0.00001092, 0.00002794, 0.00002723, 0.00001306, -0.00000408, -0.00001460, -0.00001516, -0.00000953, -0.00000467, -0.00000540, -0.00001110, -0.00001666, -0.00001669, -0.00000996, -0.00000076, 0.00000373, -0.00000139, -0.00001482, -0.00002895, -0.00003430, -0.00002573, -0.00000676, 0.00001120, 0.00001460, -0.00000444, -0.00004258, -0.00008551, -0.00011398, -0.00011280, -0.00007814, -0.00001953, 0.00004449, 0.00009454, 0.00011840, 0.00011511, 0.00009374, 0.00006807, 0.00005013, 0.00004578, 0.00005375, 0.00006783, 0.00008050, },
		{ 0.00001057, -0.00001488, -0.00003382, -0.00004540, -0.00004995, -0.00004870, -0.00004349, -0.00003631, -0.00002888, -0.00002217, -0.00001639, -0.00001129, -0.00000670, -0.00000293, -0.00000068, -0.00000058, -0.00000259, -0.00000579, -0.00000865, -0.00000974, -0.00000842, -0.00000518, -0.00000132, 0.00000166, 0.00000279, 0.00000200, 0.00000005, -0.00000190, -0.00000285, -0.00000237, -0.00000077, 0.00000104, 0.00000190, 0.00000096, -0.00000180, -0.00000534, -0.00000782, -0.00000744, -0.00000357, 0.00000248, 0.00000766, 0.00000855, 0.00000345, -0.00000611, -0.00001549, -0.00001900, -0.00001324, 0.00000030, 0.00001506, 0.00002242, 0.00001655, -0.00000145, -0.00002294, -0.00003556, -0.00002997, -0.00000592, 0.00002619, 0.00004971, 0.00005047, 0.00002508, -0.00001582, -0.00005211, -0.00006438, -0.00004472, -0.00000186, 0.00004248, 0.00006477, 0.00005234, 0.00001069, -0.00003884, -0.00007025, -0.00006675, -0.00002995, 0.00002091, 0.00005922, 0.00006526, 0.00003658, -0.00001113, -0.00005287, -0.00006788, -0.00005035, -0.00001156, 0.00002724, 0.00004681, 0.00003987, 0.00001369, -0.00001545, -0.00003227, -0.00003067, -0.00001589, 0.00000028, 0.00000739, 0.00000257, -0.00000886, -0.00001789, -0.00001847, -0.00001147, -0.00000369, -0.00000259, -0.00001058, -0.00002273, -0.00002983, -0.00002474, -0.00000828, 0.00000969, 0.00001501, -0.00000251, -0.00004202, -0.00009006, -0.00012579, -0.00013072, -0.00009771, -0.00003480, 0.00003828, 0.00009890, 0.00013123, 0.00013218, 0.00011105, 0.00008377, 0.00006503, 0.00006220, 0.00007351, 0.00009062, 0.00010348, 0.00010485, 0.00009257, 0.00006921, 0.00004003, },
		{ -0.00005983, -0.00005157, -0.00004084, -0.00002981, -0.00002021, -0.00001308, -0.00000870, -0.00000660, -0.00000595, -0.00000603, -0.00000644, -0.00000712, -0.00000800, -0.00000874, -0.00000878, -0.00000760, -0.00000515, -0.00000201, 0.00000079, 0.00000226, 0.00000200, 0.00000039, -0.00000160, -0.00000289, -0.00000283, -0.00000149, 0.00000041, 0.00000184, 0.00000189, 0.00000022, -0.00000274, -0.00000585, -0.00000763, -0.00000695, -0.00000367, 0.00000098, 0.00000466, 0.00000506, 0.00000121, -0.00000550, -0.00001154, -0.00001294, -0.00000761, 0.00000276, 0.00001292, 0.00001645, 0.00000961, -0.00000576, -0.00002201, -0.00002924, -0.00002083, 0.00000186, 0.00002873, 0.00004535, 0.00004063, 0.00001375, -0.00002385, -0.00005332, -0.00005813, -0.00003352, 0.00000961, 0.00004936, 0.00006405, 0.00004415, -0.00000162, -0.00005001, -0.00007519, -0.00006277, -0.00001810, 0.00003598, 0.00007100, 0.00006811, 0.00002840, -0.00002760, -0.00007083, -0.00007930, -0.00004963, 0.00000163, 0.00004741, 0.00006490, 0.00004738, 0.00000683, -0.00003381, -0.00005372, -0.00004532, -0.00001709, 0.00001262, 0.00002686, 0.00001949, -0.00000253, -0.00002466, -0.00003426, -0.00002799, -0.00001284, -0.00000058, 0.00000050, -0.00000904, -0.00002055, -0.00002380, -0.00001473, 0.00000056, 0.00000849, -0.00000381, -0.00003968, -0.00008882, -0.00013069, -0.00014396, -0.00011719, -0.00005489, 0.00002413, 0.00009531, 0.00013927, 0.00014962, 0.00013417, 0.00010940, 0.00009166, 0.00008965, 0.00010154, 0.00011759, 0.00012595, 0.00011850, 0.00009400, 0.00005755, 0.00001756, -0.00001797, -0.00004389, -0.00005857, -0.00006307, },
		{ -0.00001194, -0.00000750, -0.00000611, -0.00000698, -0.00000893, -0.00001078, -0.00001173, -0.00001151, -0.00001034, -0.00000859, -0.00000653, -0.00000432, -0.00000209, -0.00000014, 0.00000110, 0.00000125, 0.00000031, -0.00000123, -0.00000258, -0.00000297, -0.00000214, -0.00000043, 0.00000132, 0.00000214, 0.00000144, -0.00000078, -0.00000382, -0.00000655, -0.00000782, -0.00000692, -0.00000401, -0.00000022, 0.00000271, 0.00000316, 0.00000058, -0.00000390, -0.00000778, -0.00000831, -0.00000422, 0.00000306, 0.00000964, 0.00001100, 0.00000478, -0.00000707, -0.00001844, -0.00002188, -0.00001295, 0.00000626, 0.00002693, 0.00003742, 0.00002959, 0.00000436, -0.00002731, -0.00004906, -0.00004776, -0.00002145, 0.00001815, 0.00005059, 0.00005733, 0.00003232, -0.00001345, -0.00005657, -0.00007329, -0.00005273, -0.00000374, 0.00004872, 0.00007643, 0.00006325, 0.00001449, -0.00004526, -0.00008482, -0.00008314, -0.00004081, 0.00002010, 0.00006788, 0.00007816, 0.00004686, -0.00000810, -0.00005712, -0.00007514, -0.00005502, -0.00001044, 0.00003262, 0.00005077, 0.00003589, -0.00000147, -0.00003928, -0.00005721, -0.00004791, -0.00002004, 0.00000814, 0.00002035, 0.00001179, -0.00000900, -0.00002666, -0.00002966, -0.00001827, -0.00000478, -0.00000592, -0.00003177, -0.00007805, -0.00012635, -0.00015256, -0.00013908, -0.00008394, -0.00000220, 0.00008112, 0.00014266, 0.00017070, 0.00016870, 0.00015121, 0.00013496, 0.00013009, 0.00013607, 0.00014356, 0.00014063, 0.00011951, 0.00008045, 0.00003120, -0.00001696, -0.00005410, -0.00007504, -0.00008005, -0.00007323, -0.00005996, -0.00004487, -0.00003098, -0.00001980, },
		{ -0.00001330, -0.00001442, -0.00001412, -0.00001232, -0.00000946, -0.00000630, -0.00000356, -0.00000162, -0.00000053, -0.00000011, -0.00000019, -0.00000065, -0.00000137, -0.00000209, -0.00000243, -0.00000210, -0.00000105, 0.00000035, 0.00000142, 0.00000151, 0.00000028, -0.00000207, -0.00000482, -0.00000701, -0.00000778, -0.00000672, -0.00000411, -0.00000090, 0.00000159, 0.00000225, 0.00000071, -0.00000223, -0.00000483, -0.00000522, -0.00000253, 0.00000226, 0.00000646, 0.00000698, 0.00000227, -0.00000612, -0.00001372, -0.00001518, -0.00000757, 0.00000710, 0.00002186, 0.00002790, 0.00001953, -0.00000161, -0.00002606, -0.00004066, -0.00003564, -0.00001095, 0.00002221, 0.00004620, 0.00004629, 0.00001951, -0.00002220, -0.00005709, -0.00006495, -0.00003877, 0.00001001, 0.00005627, 0.00007414, 0.00005147, -0.00000241, -0.00006043, -0.00009165, -0.00007807, -0.00002508, 0.00004052, 0.00008440, 0.00008315, 0.00003711, -0.00002951, -0.00008155, -0.00009196, -0.00005633, 0.00000500, 0.00005830, 0.00007505, 0.00004728, -0.00000873, -0.00006212, -0.00008489, -0.00006698, -0.00002095, 0.00002621, 0.00004885, 0.00003689, 0.00000039, -0.00003764, -0.00005567, -0.00004664, -0.00002198, -0.00000439, -0.00001390, -0.00005504, -0.00011281, -0.00015941, -0.00016813, -0.00012686, -0.00004391, 0.00005625, 0.00014490, 0.00020167, 0.00022189, 0.00021519, 0.00019763, 0.00018190, 0.00017132, 0.00016003, 0.00013853, 0.00010071, 0.00004833, -0.00000919, -0.00005912, -0.00009109, -0.00010112, -0.00009230, -0.00007233, -0.00004967, -0.00003053, -0.00001756, -0.00001055, -0.00000788, -0.00000785, -0.00000927, -0.00001134, },
		{ -0.00000655, -0.00000369, -0.00000167, -0.00000079, -0.00000091, -0.00000152, -0.00000211, -0.00000234, -0.00000220, -0.00000182, -0.00000133, -0.00000078, -0.00000021, 0.00000026, 0.00000039, -0.00000012, -0.00000140, -0.00000328, -0.00000527, -0.00000674, -0.00000709, -0.00000609, -0.00000394, -0.00000134, 0.00000080, 0.00000165, 0.00000091, -0.00000094, -0.00000276, -0.00000324, -0.00000171, 0.00000133, 0.00000414, 0.00000459, 0.00000154, -0.00000408, -0.00000925, -0.00001026, -0.00000500, 0.00000514, 0.00001517, 0.00001876, 0.00001196, -0.00000375, -0.00002109, -0.00003018, -0.00002412, -0.00000372, 0.00002151, 0.00003752, 0.00003339, 0.00000836, -0.00002619, -0.00005162, -0.00005202, -0.00002389, 0.00002021, 0.00005699, 0.00006455, 0.00003521, -0.00001870, -0.00006978, -0.00008965, -0.00006484, -0.00000548, 0.00005864, 0.00009321, 0.00007811, 0.00001917, -0.00005378, -0.00010234, -0.00010014, -0.00004758, 0.00002775, 0.00008558, 0.00009465, 0.00004971, -0.00002548, -0.00009106, -0.00011269, -0.00008003, -0.00001198, 0.00005412, 0.00008314, 0.00006115, 0.00000224, -0.00006023, -0.00009294, -0.00008138, -0.00003775, 0.00000622, 0.00001741, -0.00001967, -0.00009345, -0.00017016, -0.00021027, -0.00018759, -0.00010131, 0.00002453, 0.00015304, 0.00025086, 0.00030076, 0.00030478, 0.00027821, 0.00023878, 0.00019768, 0.00015661, 0.00011121, 0.00005775, -0.00000171, -0.00005845, -0.00010091, -0.00012034, -0.00011513, -0.00009143, -0.00006004, -0.00003170, -0.00001311, -0.00000555, -0.00000609, -0.00001025, -0.00001433, -0.00001662, -0.00001702, -0.00001617, -0.00001458, -0.00001239, -0.00000963, },
		{ -0.00000266, -0.00000327, -0.00000344, -0.00000293, -0.00000187, -0.00000073, 0.00000002, 0.00000015, -0.00000026, -0.00000093, -0.00000164, -0.00000232, -0.00000304, -0.00000388, -0.00000479, -0.00000552, -0.00000573, -0.00000516, -0.00000380, -0.00000195, -0.00000018, 0.00000090, 0.00000091, 0.00000000, -0.00000118, -0.00000172, -0.00000100, 0.00000086, 0.00000282, 0.00000343, 0.00000167, -0.00000210, -0.00000602, -0.00000747, -0.00000471, 0.00000168, 0.00000853, 0.00001145, 0.00000745, -0.00000270, -0.00001409, -0.00001980, -0.00001496, -0.00000032, 0.00001700, 0.00002672, 0.00002119, 0.00000072, -0.00002510, -0.00004162, -0.00003726, -0.00001105, 0.00002495, 0.00005089, 0.00004974, 0.00001787, -0.00003097, -0.00007127, -0.00007911, -0.00004613, 0.00001389, 0.00007050, 0.00009207, 0.00006355, -0.00000372, -0.00007618, -0.00011503, -0.00009732, -0.00002959, 0.00005407, 0.00010923, 0.00010495, 0.00004114, -0.00005025, -0.00012160, -0.00013488, -0.00008246, 0.00000834, 0.00008947, 0.00011778, 0.00007822, -0.00000812, -0.00009549, -0.00013840, -0.00011638, -0.00004536, 0.00003063, 0.00006292, 0.00002445, -0.00007499, -0.00019222, -0.00027072, -0.00026705, -0.00017032, -0.00000595, 0.00017685, 0.00032669, 0.00040967, 0.00041871, 0.00036987, 0.00029001, 0.00020335, 0.00012348, 0.00005336, -0.00000898, -0.00006397, -0.00010740, -0.00013231, -0.00013364, -0.00011254, -0.00007720, -0.00003985, -0.00001168, 0.00000147, 0.00000052, -0.00000863, -0.00001877, -0.00002472, -0.00002492, -0.00002086, -0.00001526, -0.00001034, -0.00000698, -0.00000493, -0.00000358, -0.00000262, -0.00000208, -0.00000213, },
		{ -0.00000080, 0.00000030, 0.00000063, -0.00000004, -0.00000144, -0.00000297, -0.00000402, -0.00000431, -0.00000400, -0.00000352, -0.00000332, -0.00000357, -0.00000406, -0.00000436, -0.00000407, -0.00000305, -0.00000154, -0.00000007, 0.00000082, 0.00000090, 0.00000037, -0.00000014, -0.00000003, 0.00000091, 0.00000219, 0.00000280, 0.00000179, -0.00000094, -0.00000431, -0.00000644, -0.00000571, -0.00000190, 0.00000320, 0.00000655, 0.00000563, 0.00000021, -0.00000690, -0.00001114, -0.00000890, -0.00000020, 0.00001044, 0.00001612, 0.00001159, -0.00000278, -0.00001998, -0.00002953, -0.00002353, -0.00000237, 0.00002383, 0.00003962, 0.00003286, 0.00000291, -0.00003687, -0.00006476, -0.00006229, -0.00002583, 0.00002896, 0.00007349, 0.00008101, 0.00004227, -0.00002699, -0.00009205, -0.00011676, -0.00008368, -0.00000560, 0.00007880, 0.00012420, 0.00010311, 0.00002221, -0.00007901, -0.00014782, -0.00014609, -0.00007173, 0.00003847, 0.00012701, 0.00014580, 0.00008223, -0.00003299, -0.00014101, -0.00018566, -0.00014374, -0.00003852, 0.00007090, 0.00011941, 0.00006985, -0.00006642, -0.00023261, -0.00035171, -0.00036142, -0.00024191, -0.00002367, 0.00022669, 0.00043496, 0.00054790, 0.00054964, 0.00046053, 0.00032204, 0.00017691, 0.00005429, -0.00003489, -0.00009325, -0.00012814, -0.00014462, -0.00014381, -0.00012563, -0.00009298, -0.00005363, -0.00001836, 0.00000344, 0.00000791, -0.00000179, -0.00001756, -0.00003054, -0.00003507, -0.00003054, -0.00002049, -0.00000991, -0.00000263, 0.00000016, -0.00000047, -0.00000246, -0.00000413, -0.00000486, -0.00000486, -0.00000453, -0.00000405, -0.00000330, -0.00000216, },
		{ -0.00000340, -0.00000507, -0.00000584, -0.00000536, -0.00000391, -0.00000227, -0.00000131, -0.00000148, -0.00000263, -0.00000407, -0.00000494, -0.00000464, -0.00000316, -0.00000102, 0.00000094, 0.00000207, 0.00000219, 0.00000166, 0.00000117, 0.00000120, 0.00000171, 0.00000204, 0.00000136, -0.00000078, -0.00000383, -0.00000641, -0.00000692, -0.00000462, -0.00000030, 0.00000387, 0.00000551, 0.00000354, -0.00000094, -0.00000501, -0.00000568, -0.00000204, 0.00000378, 0.00000751, 0.00000543, -0.00000272, -0.00001276, -0.00001791, -0.00001302, 0.00000138, 0.00001798, 0.00002591, 0.00001723, -0.00000713, -0.00003582, -0.00005201, -0.00004276, -0.00000802, 0.00003680, 0.00006713, 0.00006230, 0.00001857, -0.00004589, -0.00009810, -0.00010709, -0.00006191, 0.00001945, 0.00009668, 0.00012704, 0.00008902, -0.00000363, -0.00010599, -0.00016391, -0.00014287, -0.00004842, 0.00007373, 0.00015965, 0.00016045, 0.00006957, -0.00007109, -0.00019079, -0.00022621, -0.00015571, -0.00001348, 0.00012494, 0.00017942, 0.00010631, -0.00007683, -0.00029559, -0.00045055, -0.00046146, -0.00030303, -0.00001599, 0.00031057, 0.00057626, 0.00070808, 0.00068423, 0.00053464, 0.00032138, 0.00011049, -0.00005138, -0.00014686, -0.00018355, -0.00018125, -0.00015918, -0.00012911, -0.00009567, -0.00006087, -0.00002822, -0.00000360, 0.00000745, 0.00000336, -0.00001172, -0.00002935, -0.00004079, -0.00004105, -0.00003082, -0.00001541, -0.00000154, 0.00000603, 0.00000633, 0.00000171, -0.00000406, -0.00000797, -0.00000890, -0.00000749, -0.00000517, -0.00000306, -0.00000158, -0.00000061, 0.00000004, 0.00000026, -0.00000023, -0.00000155, },
		{ -0.00000290, -0.00000068, 0.00000052, 0.00000005, -0.00000190, -0.00000441, -0.00000622, -0.00000635, -0.00000461, -0.00000162, 0.00000149, 0.00000361, 0.00000423, 0.00000353, 0.00000229, 0.00000131, 0.00000094, 0.00000085, 0.00000029, -0.00000136, -0.00000398, -0.00000655, -0.00000760, -0.00000606, -0.00000211, 0.00000267, 0.00000598, 0.00000613, 0.00000309, -0.00000129, -0.00000435, -0.00000429, -0.00000146, 0.00000175, 0.00000244, -0.00000059, -0.00000563, -0.00000882, -0.00000669, 0.00000089, 0.00000963, 0.00001277, 0.00000542, -0.00001126, -0.00002901, -0.00003612, -0.00002444, 0.00000418, 0.00003643, 0.00005312, 0.00003989, -0.00000283, -0.00005681, -0.00009333, -0.00008816, -0.00003664, 0.00004027, 0.00010385, 0.00011699, 0.00006548, -0.00003094, -0.00012519, -0.00016578, -0.00012492, -0.00001645, 0.00010728, 0.00018017, 0.00015708, 0.00004034, -0.00011722, -0.00023625, -0.00025126, -0.00014692, 0.00002923, 0.00018605, 0.00023222, 0.00012305, -0.00011271, -0.00038040, -0.00055875, -0.00055327, -0.00033888, 0.00002778, 0.00043124, 0.00074408, 0.00087613, 0.00080490, 0.00057616, 0.00027800, 0.00000255, -0.00018667, -0.00026977, -0.00026507, -0.00021061, -0.00014358, -0.00008722, -0.00004862, -0.00002467, -0.00001015, -0.00000273, -0.00000301, -0.00001142, -0.00002527, -0.00003853, -0.00004453, -0.00003960, -0.00002527, -0.00000744, 0.00000672, 0.00001240, 0.00000923, 0.00000078, -0.00000782, -0.00001259, -0.00001236, -0.00000857, -0.00000387, -0.00000044, 0.00000090, 0.00000060, -0.00000042, -0.00000155, -0.00000267, -0.00000392, -0.00000527, -0.00000626, -0.00000631, -0.00000508, },
		{ -0.00000196, -0.00000510, -0.00000740, -0.00000764, -0.00000549, -0.00000166, 0.00000242, 0.00000531, 0.00000618, 0.00000512, 0.00000303, 0.00000100, -0.00000023, -0.00000072, -0.00000112, -0.00000219, -0.00000415, -0.00000635, -0.00000749, -0.00000639, -0.00000281, 0.00000208, 0.00000618, 0.00000748, 0.00000529, 0.00000075, -0.00000371, -0.00000575, -0.00000454, -0.00000130, 0.00000152, 0.00000189, -0.00000033, -0.00000323, -0.00000422, -0.00000217, 0.00000131, 0.00000262, -0.00000134, -0.00001015, -0.00001896, -0.00002064, -0.00001048, 0.00000935, 0.00002894, 0.00003488, 0.00001826, -0.00001812, -0.00005822, -0.00007926, -0.00006414, -0.00001308, 0.00005264, 0.00009891, 0.00009621, 0.00003754, -0.00005370, -0.00013205, -0.00015238, -0.00009540, 0.00001775, 0.00013173, 0.00018318, 0.00013472, -0.00000125, -0.00016320, -0.00026817, -0.00025421, -0.00011630, 0.00008415, 0.00024377, 0.00026605, 0.00011158, -0.00017559, -0.00048013, -0.00066258, -0.00062060, -0.00033616, 0.00011319, 0.00058411, 0.00092464, 0.00103303, 0.00089298, 0.00057225, 0.00018836, -0.00014058, -0.00033785, -0.00038686, -0.00032295, -0.00020707, -0.00009612, -0.00002327, 0.00000603, 0.00000513, -0.00000768, -0.00001973, -0.00002720, -0.00003194, -0.00003621, -0.00003934, -0.00003818, -0.00003015, -0.00001600, -0.00000026, 0.00001105, 0.00001379, 0.00000797, -0.00000240, -0.00001171, -0.00001573, -0.00001354, -0.00000742, -0.00000104, 0.00000265, 0.00000275, 0.00000031, -0.00000277, -0.00000502, -0.00000601, -0.00000616, -0.00000603, -0.00000579, -0.00000513, -0.00000369, -0.00000159, 0.00000045, 0.00000140, 0.00000055, },
		{ -0.00000562, -0.00000128, 0.00000341, 0.00000673, 0.00000763, 0.00000614, 0.00000325, 0.00000032, -0.00000162, -0.00000235, -0.00000246, -0.00000285, -0.00000403, -0.00000570, -0.00000678, -0.00000607, -0.00000307, 0.00000148, 0.00000574, 0.00000769, 0.00000625, 0.00000204, -0.00000284, -0.00000586, -0.00000559, -0.00000249, 0.00000130, 0.00000332, 0.00000231, -0.00000096, -0.00000431, -0.00000573, -0.00000488, -0.00000335, -0.00000338, -0.00000577, -0.00000869, -0.00000850, -0.00000251, 0.00000805, 0.00001710, 0.00001649, 0.00000127, -0.00002531, -0.00005096, -0.00005947, -0.00004002, 0.00000432, 0.00005457, 0.00008332, 0.00006924, 0.00001114, -0.00006697, -0.00012480, -0.00012609, -0.00006026, 0.00004680, 0.00014120, 0.00016682, 0.00009664, -0.00004756, -0.00019968, -0.00027907, -0.00023263, -0.00006797, 0.00014158, 0.00028605, 0.00027101, 0.00006829, -0.00026052, -0.00058209, -0.00074534, -0.00064824, -0.00028658, 0.00023821, 0.00075675, 0.00109848, 0.00115803, 0.00093269, 0.00051664, 0.00005726, -0.00030500, -0.00048649, -0.00048081, -0.00034573, -0.00016755, -0.00002185, 0.00005222, 0.00005625, 0.00001881, -0.00002569, -0.00005399, -0.00006010, -0.00005111, -0.00003795, -0.00002769, -0.00002108, -0.00001509, -0.00000723, 0.00000174, 0.00000835, 0.00000917, 0.00000360, -0.00000540, -0.00001299, -0.00001534, -0.00001171, -0.00000465, 0.00000179, 0.00000444, 0.00000258, -0.00000205, -0.00000667, -0.00000916, -0.00000900, -0.00000714, -0.00000502, -0.00000354, -0.00000271, -0.00000194, -0.00000084, 0.00000038, 0.00000091, -0.00000000, -0.00000245, -0.00000557, -0.00000788, -0.00000806, },
		{ 0.00000832, 0.00000654, 0.00000310, -0.00000046, -0.00000286, -0.00000369, -0.00000346, -0.00000318, -0.00000363, -0.00000481, -0.00000586, -0.00000561, -0.00000332, 0.00000065, 0.00000481, 0.00000726, 0.00000669, 0.00000327, -0.00000135, -0.00000478, -0.00000533, -0.00000294, 0.00000067, 0.00000303, 0.00000244, -0.00000096, -0.00000528, -0.00000806, -0.00000788, -0.00000527, -0.00000221, -0.00000060, -0.00000079, -0.00000134, -0.00000036, 0.00000241, 0.00000442, 0.00000153, -0.00000890, -0.00002466, -0.00003791, -0.00003849, -0.00002025, 0.00001308, 0.00004675, 0.00006094, 0.00004172, -0.00000853, -0.00006828, -0.00010502, -0.00009225, -0.00002674, 0.00006437, 0.00013294, 0.00013344, 0.00004984, -0.00008934, -0.00021861, -0.00026542, -0.00018939, -0.00001074, 0.00018954, 0.00030209, 0.00024183, -0.00000379, -0.00035627, -0.00066988, -0.00079070, -0.00062561, -0.00018949, 0.00039246, 0.00093031, 0.00124372, 0.00123280, 0.00091475, 0.00041179, -0.00010223, -0.00047137, -0.00061301, -0.00053748, -0.00032824, -0.00009629, 0.00006812, 0.00012564, 0.00009046, 0.00001010, -0.00006407, -0.00010009, -0.00009365, -0.00006126, -0.00002564, -0.00000269, 0.00000406, 0.00000071, -0.00000405, -0.00000503, -0.00000254, -0.00000017, -0.00000099, -0.00000496, -0.00000925, -0.00001055, -0.00000762, -0.00000215, 0.00000243, 0.00000325, -0.00000021, -0.00000595, -0.00001080, -0.00001238, -0.00001033, -0.00000621, -0.00000231, -0.00000020, -0.00000003, -0.00000080, -0.00000133, -0.00000117, -0.00000080, -0.00000115, -0.00000276, -0.00000521, -0.00000725, -0.00000747, -0.00000517, -0.00000085, 0.00000395, 0.00000743, },
		{ -0.00000363, -0.00000454, -0.00000408, -0.00000331, -0.00000320, -0.00000400, -0.00000507, -0.00000526, -0.00000364, -0.00000024, 0.00000381, 0.00000676, 0.00000715, 0.00000470, 0.00000059, -0.00000314, -0.00000468, -0.00000354, -0.00000088, 0.00000121, 0.00000097, -0.00000185, -0.00000574, -0.00000831, -0.00000783, -0.00000439, 0.00000012, 0.00000319, 0.00000331, 0.00000085, -0.00000237, -0.00000465, -0.00000591, -0.00000780, -0.00001207, -0.00001834, -0.00002297, -0.00002051, -0.00000759, 0.00001318, 0.00003224, 0.00003688, 0.00001887, -0.00001843, -0.00005829, -0.00007718, -0.00005772, -0.00000143, 0.00006694, 0.00010830, 0.00008920, 0.00000332, -0.00011807, -0.00021543, -0.00022885, -0.00013233, 0.00004373, 0.00021670, 0.00028532, 0.00017984, -0.00009492, -0.00044713, -0.00072671, -0.00078647, -0.00054976, -0.00005304, 0.00055842, 0.00108274, 0.00134032, 0.00124540, 0.00083892, 0.00026904, -0.00027084, -0.00061853, -0.00070068, -0.00054913, -0.00027311, -0.00000424, 0.00015897, 0.00018343, 0.00010050, -0.00002197, -0.00011738, -0.00014903, -0.00011884, -0.00005640, 0.00000213, 0.00003274, 0.00003166, 0.00001130, -0.00001070, -0.00002195, -0.00002013, -0.00001086, -0.00000199, 0.00000176, 0.00000055, -0.00000227, -0.00000354, -0.00000257, -0.00000115, -0.00000166, -0.00000500, -0.00000976, -0.00001319, -0.00001309, -0.00000926, -0.00000361, 0.00000110, 0.00000292, 0.00000173, -0.00000096, -0.00000320, -0.00000382, -0.00000307, -0.00000217, -0.00000232, -0.00000374, -0.00000552, -0.00000616, -0.00000459, -0.00000088, 0.00000366, 0.00000719, 0.00000826, 0.00000654, 0.00000294, -0.00000094, },
		{ -0.00000301, -0.00000351, -0.00000453, -0.00000498, -0.00000384, -0.00000086, 0.00000313, 0.00000652, 0.00000774, 0.00000616, 0.00000244, -0.00000167, -0.00000435, -0.00000468, -0.00000315, -0.00000136, -0.00000094, -0.00000245, -0.00000496, -0.00000659, -0.00000578, -0.00000244, 0.00000177, 0.00000437, 0.00000355, -0.00000060, -0.00000611, -0.00001039, -0.00001191, -0.00001113, -0.00000987, -0.00000960, -0.00000987, -0.00000829, -0.00000256, 0.00000687, 0.00001539, 0.00001604, 0.00000405, -0.00001836, -0.00004050, -0.00004746, -0.00002904, 0.00001149, 0.00005474, 0.00007238, 0.00004256, -0.00003417, -0.00012823, -0.00019047, -0.00017608, -0.00007261, 0.00008374, 0.00021532, 0.00023544, 0.00009325, -0.00019030, -0.00051621, -0.00073921, -0.00072774, -0.00042672, 0.00010680, 0.00071443, 0.00119301, 0.00137421, 0.00119301, 0.00071443, 0.00010680, -0.00042672, -0.00072774, -0.00073921, -0.00051621, -0.00019030, 0.00009325, 0.00023544, 0.00021532, 0.00008374, -0.00007261, -0.00017608, -0.00019047, -0.00012823, -0.00003417, 0.00004256, 0.00007238, 0.00005474, 0.00001149, -0.00002904, -0.00004746, -0.00004050, -0.00001836, 0.00000405, 0.00001604, 0.00001539, 0.00000687, -0.00000256, -0.00000829, -0.00000987, -0.00000960, -0.00000987, -0.00001113, -0.00001191, -0.00001039, -0.00000611, -0.00000060, 0.00000355, 0.00000437, 0.00000177, -0.00000244, -0.00000578, -0.00000659, -0.00000496, -0.00000245, -0.00000094, -0.00000136, -0.00000315, -0.00000468, -0.00000435, -0.00000167, 0.00000244, 0.00000616, 0.00000774, 0.00000652, 0.00000313, -0.00000086, -0.00000384, -0.00000498, -0.00000453, -0.00000351, },
		{ -0.00000363, -0.00000094, 0.00000294, 0.00000654, 0.00000826, 0.00000719, 0.00000366, -0.00000088, -0.00000459, -0.00000616, -0.00000552, -0.00000374, -0.00000232, -0.00000217, -0.00000307, -0.00000382, -0.00000320, -0.00000096, 0.00000173, 0.00000292, 0.00000110, -0.00000361, -0.00000926, -0.00001309, -0.00001319, -0.00000976, -0.00000500, -0.00000166, -0.00000115, -0.00000257, -0.00000354, -0.00000227, 0.00000055, 0.00000176, -0.00000199, -0.00001086, -0.00002013, -0.00002195, -0.00001070, 0.00001130, 0.00003166, 0.00003274, 0.00000213, -0.00005640, -0.00011884, -0.00014903, -0.00011738, -0.00002197, 0.00010050, 0.00018343, 0.00015897, -0.00000424, -0.00027311, -0.00054913, -0.00070068, -0.00061853, -0.00027084, 0.00026904, 0.00083892, 0.00124540, 0.00134032, 0.00108274, 0.00055842, -0.00005304, -0.00054976, -0.00078647, -0.00072671, -0.00044713, -0.00009492, 0.00017984, 0.00028532, 0.00021670, 0.00004373, -0.00013233, -0.00022885, -0.00021543, -0.00011807, 0.00000332, 0.00008920, 0.00010830, 0.00006694, -0.00000143, -0.00005772, -0.00007718, -0.00005829, -0.00001843, 0.00001887, 0.00003688, 0.00003224, 0.00001318, -0.00000759, -0.00002051, -0.00002297, -0.00001834, -0.00001207, -0.00000780, -0.00000591, -0.00000465, -0.00000237, 0.00000085, 0.00000331, 0.00000319, 0.00000012, -0.00000439, -0.00000783, -0.00000831, -0.00000574, -0.00000185, 0.00000097, 0.00000121, -0.00000088, -0.00000354, -0.00000468, -0.00000314, 0.00000059, 0.00000470, 0.00000715, 0.00000676, 0.00000381, -0.00000024, -0.00000364, -0.00000526, -0.00000507, -0.00000400, -0.00000320, -0.00000331, -0.00000408, -0.00000454, },
		{ 0.00000832, 0.00000743, 0.00000395, -0.00000085, -0.00000517, -0.00000747, -0.00000725, -0.00000521, -0.00000276, -0.00000115, -0.00000080, -0.00000117, -0.00000133, -0.00000080, -0.00000003, -0.00000020, -0.00000231, -0.00000621, -0.00001033, -0.00001238, -0.00001080, -0.00000595, -0.00000021, 0.00000325, 0.00000243, -0.00000215, -0.00000762, -0.00001055, -0.00000925, -0.00000496, -0.00000099, -0.00000017, -0.00000254, -0.00000503, -0.00000405, 0.00000071, 0.00000406, -0.00000269, -0.00002564, -0.00006126, -0.00009365, -0.00010009, -0.00006407, 0.00001010, 0.00009046, 0.00012564, 0.00006812, -0.00009629, -0.00032824, -0.00053748, -0.00061301, -0.00047137, -0.00010223, 0.00041179, 0.00091475, 0.00123280, 0.00124372, 0.00093031, 0.00039246, -0.00018949, -0.00062561, -0.00079070, -0.00066988, -0.00035627, -0.00000379, 0.00024183, 0.00030209, 0.00018954, -0.00001074, -0.00018939, -0.00026542, -0.00021861, -0.00008934, 0.00004984, 0.00013344, 0.00013294, 0.00006437, -0.00002674, -0.00009225, -0.00010502, -0.00006828, -0.00000853, 0.00004172, 0.00006094, 0.00004675, 0.00001308, -0.00002025, -0.00003849, -0.00003791, -0.00002466, -0.00000890, 0.00000153, 0.00000442, 0.00000241, -0.00000036, -0.00000134, -0.00000079, -0.00000060, -0.00000221, -0.00000527, -0.00000788, -0.00000806, -0.00000528, -0.00000096, 0.00000244, 0.00000303, 0.00000067, -0.00000294, -0.00000533, -0.00000478, -0.00000135, 0.00000327, 0.00000669, 0.00000726, 0.00000481, 0.00000065, -0.00000332, -0.00000561, -0.00000586, -0.00000481, -0.00000363, -0.00000318, -0.00000346, -0.00000369, -0.00000286, -0.00000046, 0.00000310, 0.00000654, },
		{ -0.00000562, -0.00000806, -0.00000788, -0.00000557, -0.00000245, -0.00000000, 0.00000091, 0.00000038, -0.00000084, -0.00000194, -0.00000271, -0.00000354, -0.00000502, -0.00000714, -0.00000900, -0.00000916, -0.00000667, -0.00000205, 0.00000258, 0.00000444, 0.00000179, -0.00000465, -0.00001171, -0.00001534, -0.00001299, -0.00000540, 0.00000360, 0.00000917, 0.00000835, 0.00000174, -0.00000723, -0.00001509, -0.00002108, -0.00002769, -0.00003795, -0.00005111, -0.00006010, -0.00005399, -0.00002569, 0.00001881, 0.00005625, 0.00005222, -0.00002185, -0.00016755, -0.00034573, -0.00048081, -0.00048649, -0.00030500, 0.00005726, 0.00051664, 0.00093269, 0.00115803, 0.00109848, 0.00075675, 0.00023821, -0.00028658, -0.00064824, -0.00074534, -0.00058209, -0.00026052, 0.00006829, 0.00027101, 0.00028605, 0.00014158, -0.00006797, -0.00023263, -0.00027907, -0.00019968, -0.00004756, 0.00009664, 0.00016682, 0.00014120, 0.00004680, -0.00006026, -0.00012609, -0.00012480, -0.00006697, 0.00001114, 0.00006924, 0.00008332, 0.00005457, 0.00000432, -0.00004002, -0.00005947, -0.00005096, -0.00002531, 0.00000127, 0.00001649, 0.00001710, 0.00000805, -0.00000251, -0.00000850, -0.00000869, -0.00000577, -0.00000338, -0.00000335, -0.00000488, -0.00000573, -0.00000431, -0.00000096, 0.00000231, 0.00000332, 0.00000130, -0.00000249, -0.00000559, -0.00000586, -0.00000284, 0.00000204, 0.00000625, 0.00000769, 0.00000574, 0.00000148, -0.00000307, -0.00000607, -0.00000678, -0.00000570, -0.00000403, -0.00000285, -0.00000246, -0.00000235, -0.00000162, 0.00000032, 0.00000325, 0.00000614, 0.00000763, 0.00000673, 0.00000341, -0.00000128, },
		{ -0.00000196, 0.00000055, 0.00000140, 0.00000045, -0.00000159, -0.00000369, -0.00000513, -0.00000579, -0.00000603, -0.00000616, -0.00000601, -0.00000502, -0.00000277, 0.00000031, 0.00000275, 0.00000265, -0.00000104, -0.00000742, -0.00001354, -0.00001573, -0.00001171, -0.00000240, 0.00000797, 0.00001379, 0.00001105, -0.00000026, -0.00001600, -0.00003015, -0.00003818, -0.00003934, -0.00003621, -0.00003194, -0.00002720, -0.00001973, -0.00000768, 0.00000513, 0.00000603, -0.00002327, -0.00009612, -0.00020707, -0.00032295, -0.00038686, -0.00033785, -0.00014058, 0.00018836, 0.00057225, 0.00089298, 0.00103303, 0.00092464, 0.00058411, 0.00011319, -0.00033616, -0.00062060, -0.00066258, -0.00048013, -0.00017559, 0.00011158, 0.00026605, 0.00024377, 0.00008415, -0.00011630, -0.00025421, -0.00026817, -0.00016320, -0.00000125, 0.00013472, 0.00018318, 0.00013173, 0.00001775, -0.00009540, -0.00015238, -0.00013205, -0.00005370, 0.00003754, 0.00009621, 0.00009891, 0.00005264, -0.00001308, -0.00006414, -0.00007926, -0.00005822, -0.00001812, 0.00001826, 0.00003488, 0.00002894, 0.00000935, -0.00001048, -0.00002064, -0.00001896, -0.00001015, -0.00000134, 0.00000262, 0.00000131, -0.00000217, -0.00000422, -0.00000323, -0.00000033, 0.00000189, 0.00000152, -0.00000130, -0.00000454, -0.00000575, -0.00000371, 0.00000075, 0.00000529, 0.00000748, 0.00000618, 0.00000208, -0.00000281, -0.00000639, -0.00000749, -0.00000635, -0.00000415, -0.00000219, -0.00000112, -0.00000072, -0.00000023, 0.00000100, 0.00000303, 0.00000512, 0.00000618, 0.00000531, 0.00000242, -0.00000166, -0.00000549, -0.00000764, -0.00000740, -0.00000510, },
		{ -0.00000290, -0.00000508, -0.00000631, -0.00000626, -0.00000527, -0.00000392, -0.00000267, -0.00000155, -0.00000042, 0.00000060, 0.00000090, -0.00000044, -0.00000387, -0.00000857, -0.00001236, -0.00001259, -0.00000782, 0.00000078, 0.00000923, 0.00001240, 0.00000672, -0.00000744, -0.00002527, -0.00003960, -0.00004453, -0.00003853, -0.00002527, -0.00001142, -0.00000301, -0.00000273, -0.00001015, -0.00002467, -0.00004862, -0.00008722, -0.00014358, -0.00021061, -0.00026507, -0.00026977, -0.00018667, 0.00000255, 0.00027800, 0.00057616, 0.00080490, 0.00087613, 0.00074408, 0.00043124, 0.00002778, -0.00033888, -0.00055327, -0.00055875, -0.00038040, -0.00011271, 0.00012305, 0.00023222, 0.00018605, 0.00002923, -0.00014692, -0.00025126, -0.00023625, -0.00011722, 0.00004034, 0.00015708, 0.00018017, 0.00010728, -0.00001645, -0.00012492, -0.00016578, -0.00012519, -0.00003094, 0.00006548, 0.00011699, 0.00010385, 0.00004027, -0.00003664, -0.00008816, -0.00009333, -0.00005681, -0.00000283, 0.00003989, 0.00005312, 0.00003643, 0.00000418, -0.00002444, -0.00003612, -0.00002901, -0.00001126, 0.00000542, 0.00001277, 0.00000963, 0.00000089, -0.00000669, -0.00000882, -0.00000563, -0.00000059, 0.00000244, 0.00000175, -0.00000146, -0.00000429, -0.00000435, -0.00000129, 0.00000309, 0.00000613, 0.00000598, 0.00000267, -0.00000211, -0.00000606, -0.00000760, -0.00000655, -0.00000398, -0.00000136, 0.00000029, 0.00000085, 0.00000094, 0.00000131, 0.00000229, 0.00000353, 0.00000423, 0.00000361, 0.00000149, -0.00000162, -0.00000461, -0.00000635, -0.00000622, -0.00000441, -0.00000190, 0.00000005, 0.00000052, -0.00000068, },
		{ -0.00000340, -0.00000155, -0.00000023, 0.00000026, 0.00000004, -0.00000061, -0.00000158, -0.00000306, -0.00000517, -0.00000749, -0.00000890, -0.00000797, -0.00000406, 0.00000171, 0.00000633, 0.00000603, -0.00000154, -0.00001541, -0.00003082, -0.00004105, -0.00004079, -0.00002935, -0.00001172, 0.00000336, 0.00000745, -0.00000360, -0.00002822, -0.00006087, -0.00009567, -0.00012911, -0.00015918, -0.00018125, -0.00018355, -0.00014686, -0.00005138, 0.00011049, 0.00032138, 0.00053464, 0.00068423, 0.00070808, 0.00057626, 0.00031057, -0.00001599, -0.00030303, -0.00046146, -0.00045055, -0.00029559, -0.00007683, 0.00010631, 0.00017942, 0.00012494, -0.00001348, -0.00015571, -0.00022621, -0.00019079, -0.00007109, 0.00006957, 0.00016045, 0.00015965, 0.00007373, -0.00004842, -0.00014287, -0.00016391, -0.00010599, -0.00000363, 0.00008902, 0.00012704, 0.00009668, 0.00001945, -0.00006191, -0.00010709, -0.00009810, -0.00004589, 0.00001857, 0.00006230, 0.00006713, 0.00003680, -0.00000802, -0.00004276, -0.00005201, -0.00003582, -0.00000713, 0.00001723, 0.00002591, 0.00001798, 0.00000138, -0.00001302, -0.00001791, -0.00001276, -0.00000272, 0.00000543, 0.00000751, 0.00000378, -0.00000204, -0.00000568, -0.00000501, -0.00000094, 0.00000354, 0.00000551, 0.00000387, -0.00000030, -0.00000462, -0.00000692, -0.00000641, -0.00000383, -0.00000078, 0.00000136, 0.00000204, 0.00000171, 0.00000120, 0.00000117, 0.00000166, 0.00000219, 0.00000207, 0.00000094, -0.00000102, -0.00000316, -0.00000464, -0.00000494, -0.00000407, -0.00000263, -0.00000148, -0.00000131, -0.00000227, -0.00000391, -0.00000536, -0.00000584, -0.00000507, },
		{ -0.00000080, -0.00000216, -0.00000330, -0.00000405, -0.00000453, -0.00000486, -0.00000486, -0.00000413, -0.00000246, -0.00000047, 0.00000016, -0.00000263, -0.00000991, -0.00002049, -0.00003054, -0.00003507, -0.00003054, -0.00001756, -0.00000179, 0.00000791, 0.00000344, -0.00001836, -0.00005363, -0.00009298, -0.00012563, -0.00014381, -0.00014462, -0.00012814, -0.00009325, -0.00003489, 0.00005429, 0.00017691, 0.00032204, 0.00046053, 0.00054964, 0.00054790, 0.00043496, 0.00022669, -0.00002367, -0.00024191, -0.00036142, -0.00035171, -0.00023261, -0.00006642, 0.00006985, 0.00011941, 0.00007090, -0.00003852, -0.00014374, -0.00018566, -0.00014101, -0.00003299, 0.00008223, 0.00014580, 0.00012701, 0.00003847, -0.00007173, -0.00014609, -0.00014782, -0.00007901, 0.00002221, 0.00010311, 0.00012420, 0.00007880, -0.00000560, -0.00008368, -0.00011676, -0.00009205, -0.00002699, 0.00004227, 0.00008101, 0.00007349, 0.00002896, -0.00002583, -0.00006229, -0.00006476, -0.00003687, 0.00000291, 0.00003286, 0.00003962, 0.00002383, -0.00000237, -0.00002353, -0.00002953, -0.00001998, -0.00000278, 0.00001159, 0.00001612, 0.00001044, -0.00000020, -0.00000890, -0.00001114, -0.00000690, 0.00000021, 0.00000563, 0.00000655, 0.00000320, -0.00000190, -0.00000571, -0.00000644, -0.00000431, -0.00000094, 0.00000179, 0.00000280, 0.00000219, 0.00000091, -0.00000003, -0.00000014, 0.00000037, 0.00000090, 0.00000082, -0.00000007, -0.00000154, -0.00000305, -0.00000407, -0.00000436, -0.00000406, -0.00000357, -0.00000332, -0.00000352, -0.00000400, -0.00000431, -0.00000402, -0.00000297, -0.00000144, -0.00000004, 0.00000063, 0.00000030, },
		{ -0.00000266, -0.00000213, -0.00000208, -0.00000262, -0.00000358, -0.00000493, -0.00000698, -0.00001034, -0.00001526, -0.00002086, -0.00002492, -0.00002472, -0.00001877, -0.00000863, 0.00000052, 0.00000147, -0.00001168, -0.00003985, -0.00007720, -0.00011254, -0.00013364, -0.00013231, -0.00010740, -0.00006397, -0.00000898, 0.00005336, 0.00012348, 0.00020335, 0.00029001, 0.00036987, 0.00041871, 0.00040967, 0.00032669, 0.00017685, -0.00000595, -0.00017032, -0.00026705, -0.00027072, -0.00019222, -0.00007499, 0.00002445, 0.00006292, 0.00003063, -0.00004536, -0.00011638, -0.00013840, -0.00009549, -0.00000812, 0.00007822, 0.00011778, 0.00008947, 0.00000834, -0.00008246, -0.00013488, -0.00012160, -0.00005025, 0.00004114, 0.00010495, 0.00010923, 0.00005407, -0.00002959, -0.00009732, -0.00011503, -0.00007618, -0.00000372, 0.00006355, 0.00009207, 0.00007050, 0.00001389, -0.00004613, -0.00007911, -0.00007127, -0.00003097, 0.00001787, 0.00004974, 0.00005089, 0.00002495, -0.00001105, -0.00003726, -0.00004162, -0.00002510, 0.00000072, 0.00002119, 0.00002672, 0.00001700, -0.00000032, -0.00001496, -0.00001980, -0.00001409, -0.00000270, 0.00000745, 0.00001145, 0.00000853, 0.00000168, -0.00000471, -0.00000747, -0.00000602, -0.00000210, 0.00000167, 0.00000343, 0.00000282, 0.00000086, -0.00000100, -0.00000172, -0.00000118, 0.00000000, 0.00000091, 0.00000090, -0.00000018, -0.00000195, -0.00000380, -0.00000516, -0.00000573, -0.00000552, -0.00000479, -0.00000388, -0.00000304, -0.00000232, -0.00000164, -0.00000093, -0.00000026, 0.00000015, 0.00000002, -0.00000073, -0.00000187, -0.00000293, -0.00000344, -0.00000327, },
		{ -0.00000655, -0.00000963, -0.00001239, -0.00001458, -0.00001617, -0.00001702, -0.00001662, -0.00001433, -0.00001025, -0.00000609, -0.00000555, -0.00001311, -0.00003170, -0.00006004, -0.00009143, -0.00011513, -0.00012034, -0.00010091, -0.00005845, -0.00000171, 0.00005775, 0.00011121, 0.00015661, 0.00019768, 0.00023878, 0.00027821, 0.00030478, 0.00030076, 0.00025086, 0.00015304, 0.00002453, -0.00010131, -0.00018759, -0.00021027, -0.00017016, -0.00009345, -0.00001967, 0.00001741, 0.00000622, -0.00003775, -0.00008138, -0.00009294, -0.00006023, 0.00000224, 0.00006115, 0.00008314, 0.00005412, -0.00001198, -0.00008003, -0.00011269, -0.00009106, -0.00002548, 0.00004971, 0.00009465, 0.00008558, 0.00002775, -0.00004758, -0.00010014, -0.00010234, -0.00005378, 0.00001917, 0.00007811, 0.00009321, 0.00005864, -0.00000548, -0.00006484, -0.00008965, -0.00006978, -0.00001870, 0.00003521, 0.00006455, 0.00005699, 0.00002021, -0.00002389, -0.00005202, -0.00005162, -0.00002619, 0.00000836, 0.00003339, 0.00003752, 0.00002151, -0.00000372, -0.00002412, -0.00003018, -0.00002109, -0.00000375, 0.00001196, 0.00001876, 0.00001517, 0.00000514, -0.00000500, -0.00001026, -0.00000925, -0.00000408, 0.00000154, 0.00000459, 0.00000414, 0.00000133, -0.00000171, -0.00000324, -0.00000276, -0.00000094, 0.00000091, 0.00000165, 0.00000080, -0.00000134, -0.00000394, -0.00000609, -0.00000709, -0.00000674, -0.00000527, -0.00000328, -0.00000140, -0.00000012, 0.00000039, 0.00000026, -0.00000021, -0.00000078, -0.00000133, -0.00000182, -0.00000220, -0.00000234, -0.00000211, -0.00000152, -0.00000091, -0.00000079, -0.00000167, -0.00000369, },
		{ -0.00001330, -0.00001134, -0.00000927, -0.00000785, -0.00000788, -0.00001055, -0.00001756, -0.00003053, -0.00004967, -0.00007233, -0.00009230, -0.00010112, -0.00009109, -0.00005912, -0.00000919, 0.00004833, 0.00010071, 0.00013853, 0.00016003, 0.00017132, 0.00018190, 0.00019763, 0.00021519, 0.00022189, 0.00020167, 0.00014490, 0.00005625, -0.00004391, -0.00012686, -0.00016813, -0.00015941, -0.00011281, -0.00005504, -0.00001390, -0.00000439, -0.00002198, -0.00004664, -0.00005567, -0.00003764, 0.00000039, 0.00003689, 0.00004885, 0.00002621, -0.00002095, -0.00006698, -0.00008489, -0.00006212, -0.00000873, 0.00004728, 0.00007505, 0.00005830, 0.00000500, -0.00005633, -0.00009196, -0.00008155, -0.00002951, 0.00003711, 0.00008315, 0.00008440, 0.00004052, -0.00002508, -0.00007807, -0.00009165, -0.00006043, -0.00000241, 0.00005147, 0.00007414, 0.00005627, 0.00001001, -0.00003877, -0.00006495, -0.00005709, -0.00002220, 0.00001951, 0.00004629, 0.00004620, 0.00002221, -0.00001095, -0.00003564, -0.00004066, -0.00002606, -0.00000161, 0.00001953, 0.00002790, 0.00002186, 0.00000710, -0.00000757, -0.00001518, -0.00001372, -0.00000612, 0.00000227, 0.00000698, 0.00000646, 0.00000226, -0.00000253, -0.00000522, -0.00000483, -0.00000223, 0.00000071, 0.00000225, 0.00000159, -0.00000090, -0.00000411, -0.00000672, -0.00000778, -0.00000701, -0.00000482, -0.00000207, 0.00000028, 0.00000151, 0.00000142, 0.00000035, -0.00000105, -0.00000210, -0.00000243, -0.00000209, -0.00000137, -0.00000065, -0.00000019, -0.00000011, -0.00000053, -0.00000162, -0.00000356, -0.00000630, -0.00000946, -0.00001232, -0.00001412, -0.00001442, },
		{ -0.00001194, -0.00001980, -0.00003098, -0.00004487, -0.00005996, -0.00007323, -0.00008005, -0.00007504, -0.00005410, -0.00001696, 0.00003120, 0.00008045, 0.00011951, 0.00014063, 0.00014356, 0.00013607, 0.00013009, 0.00013496, 0.00015121, 0.00016870, 0.00017070, 0.00014266, 0.00008112, -0.00000220, -0.00008394, -0.00013908, -0.00015256, -0.00012635, -0.00007805, -0.00003177, -0.00000592, -0.00000478, -0.00001827, -0.00002966, -0.00002666, -0.00000900, 0.00001179, 0.00002035, 0.00000814, -0.00002004, -0.00004791, -0.00005721, -0.00003928, -0.00000147, 0.00003589, 0.00005077, 0.00003262, -0.00001044, -0.00005502, -0.00007514, -0.00005712, -0.00000810, 0.00004686, 0.00007816, 0.00006788, 0.00002010, -0.00004081, -0.00008314, -0.00008482, -0.00004526, 0.00001449, 0.00006325, 0.00007643, 0.00004872, -0.00000374, -0.00005273, -0.00007329, -0.00005657, -0.00001345, 0.00003232, 0.00005733, 0.00005059, 0.00001815, -0.00002145, -0.00004776, -0.00004906, -0.00002731, 0.00000436, 0.00002959, 0.00003742, 0.00002693, 0.00000626, -0.00001295, -0.00002188, -0.00001844, -0.00000707, 0.00000478, 0.00001100, 0.00000964, 0.00000306, -0.00000422, -0.00000831, -0.00000778, -0.00000390, 0.00000058, 0.00000316, 0.00000271, -0.00000022, -0.00000401, -0.00000692, -0.00000782, -0.00000655, -0.00000382, -0.00000078, 0.00000144, 0.00000214, 0.00000132, -0.00000043, -0.00000214, -0.00000297, -0.00000258, -0.00000123, 0.00000031, 0.00000125, 0.00000110, -0.00000014, -0.00000209, -0.00000432, -0.00000653, -0.00000859, -0.00001034, -0.00001151, -0.00001173, -0.00001078, -0.00000893, -0.00000698, -0.00000611, -0.00000750, },
		{ -0.00005983, -0.00006307, -0.00005857, -0.00004389, -0.00001797, 0.00001756, 0.00005755, 0.00009400, 0.00011850, 0.00012595, 0.00011759, 0.00010154, 0.00008965, 0.00009166, 0.00010940, 0.00013417, 0.00014962, 0.00013927, 0.00009531, 0.00002413, -0.00005489, -0.00011719, -0.00014396, -0.00013069, -0.00008882, -0.00003968, -0.00000381, 0.00000849, 0.00000056, -0.00001473, -0.00002380, -0.00002055, -0.00000904, 0.00000050, -0.00000058, -0.00001284, -0.00002799, -0.00003426, -0.00002466, -0.00000253, 0.00001949, 0.00002686, 0.00001262, -0.00001709, -0.00004532, -0.00005372, -0.00003381, 0.00000683, 0.00004738, 0.00006490, 0.00004741, 0.00000163, -0.00004963, -0.00007930, -0.00007083, -0.00002760, 0.00002840, 0.00006811, 0.00007100, 0.00003598, -0.00001810, -0.00006277, -0.00007519, -0.00005001, -0.00000162, 0.00004415, 0.00006405, 0.00004936, 0.00000961, -0.00003352, -0.00005813, -0.00005332, -0.00002385, 0.00001375, 0.00004063, 0.00004535, 0.00002873, 0.00000186, -0.00002083, -0.00002924, -0.00002201, -0.00000576, 0.00000961, 0.00001645, 0.00001292, 0.00000276, -0.00000761, -0.00001294, -0.00001154, -0.00000550, 0.00000121, 0.00000506, 0.00000466, 0.00000098, -0.00000367, -0.00000695, -0.00000763, -0.00000585, -0.00000274, 0.00000022, 0.00000189, 0.00000184, 0.00000041, -0.00000149, -0.00000283, -0.00000289, -0.00000160, 0.00000039, 0.00000200, 0.00000226, 0.00000079, -0.00000201, -0.00000515, -0.00000760, -0.00000878, -0.00000874, -0.00000800, -0.00000712, -0.00000644, -0.00000603, -0.00000595, -0.00000660, -0.00000870, -0.00001308, -0.00002021, -0.00002981, -0.00004084, -0.00005157, },
		{ 0.00001057, 0.00004003, 0.00006921, 0.00009257, 0.00010485, 0.00010348, 0.00009062, 0.00007351, 0.00006220, 0.00006503, 0.00008377, 0.00011105, 0.00013218, 0.00013123, 0.00009890, 0.00003828, -0.00003480, -0.00009771, -0.00013072, -0.00012579, -0.00009006, -0.00004202, -0.00000251, 0.00001501, 0.00000969, -0.00000828, -0.00002474, -0.00002983, -0.00002273, -0.00001058, -0.00000259, -0.00000369, -0.00001147, -0.00001847, -0.00001789, -0.00000886, 0.00000257, 0.00000739, 0.00000028, -0.00001589, -0.00003067, -0.00003227, -0.00001545, 0.00001369, 0.00003987, 0.00004681, 0.00002724, -0.00001156, -0.00005035, -0.00006788, -0.00005287, -0.00001113, 0.00003658, 0.00006526, 0.00005922, 0.00002091, -0.00002995, -0.00006675, -0.00007025, -0.00003884, 0.00001069, 0.00005234, 0.00006477, 0.00004248, -0.00000186, -0.00004472, -0.00006438, -0.00005211, -0.00001582, 0.00002508, 0.00005047, 0.00004971, 0.00002619, -0.00000592, -0.00002997, -0.00003556, -0.00002294, -0.00000145, 0.00001655, 0.00002242, 0.00001506, 0.00000030, -0.00001324, -0.00001900, -0.00001549, -0.00000611, 0.00000345, 0.00000855, 0.00000766, 0.00000248, -0.00000357, -0.00000744, -0.00000782, -0.00000534, -0.00000180, 0.00000096, 0.00000190, 0.00000104, -0.00000077, -0.00000237, -0.00000285, -0.00000190, 0.00000005, 0.00000200, 0.00000279, 0.00000166, -0.00000132, -0.00000518, -0.00000842, -0.00000974, -0.00000865, -0.00000579, -0.00000259, -0.00000058, -0.00000068, -0.00000293, -0.00000670, -0.00001129, -0.00001639, -0.00002217, -0.00002888, -0.00003631, -0.00004349, -0.00004870, -0.00004995, -0.00004540, -0.00003382, -0.00001488, },
		{ 0.00008605, 0.00008050, 0.00006783, 0.00005375, 0.00004578, 0.00005013, 0.00006807, 0.00009374, 0.00011511, 0.00011840, 0.00009454, 0.00004449, -0.00001953, -0.00007814, -0.00011280, -0.00011398, -0.00008551, -0.00004258, -0.00000444, 0.00001460, 0.00001120, -0.00000676, -0.00002573, -0.00003430, -0.00002895, -0.00001482, -0.00000139, 0.00000373, -0.00000076, -0.00000996, -0.00001669, -0.00001666, -0.00001110, -0.00000540, -0.00000467, -0.00000953, -0.00001516, -0.00001460, -0.00000408, 0.00001306, 0.00002723, 0.00002794, 0.00001092, -0.00001779, -0.00004377, -0.00005173, -0.00003463, 0.00000111, 0.00003784, 0.00005572, 0.00004379, 0.00000673, -0.00003685, -0.00006395, -0.00005968, -0.00002562, 0.00002068, 0.00005494, 0.00005918, 0.00003131, -0.00001381, -0.00005250, -0.00006491, -0.00004532, -0.00000461, 0.00003609, 0.00005687, 0.00004915, 0.00001929, -0.00001586, -0.00003851, -0.00003908, -0.00002025, 0.00000575, 0.00002459, 0.00002740, 0.00001472, -0.00000504, -0.00002100, -0.00002571, -0.00001847, -0.00000466, 0.00000802, 0.00001378, 0.00001132, 0.00000360, -0.00000452, -0.00000908, -0.00000882, -0.00000514, -0.00000078, 0.00000192, 0.00000205, 0.00000030, -0.00000185, -0.00000297, -0.00000244, -0.00000057, 0.00000161, 0.00000283, 0.00000218, -0.00000054, -0.00000456, -0.00000839, -0.00001034, -0.00000929, -0.00000537, -0.00000003, 0.00000439, 0.00000571, 0.00000288, -0.00000360, -0.00001195, -0.00002009, -0.00002656, -0.00003104, -0.00003408, -0.00003631, -0.00003774, -0.00003742, -0.00003382, -0.00002550, -0.00001185, 0.00000658, 0.00002810, 0.00005005, 0.00006916, 0.00008205, },
		{ 0.00003656, 0.00004180, 0.00005743, 0.00007913, 0.00009786, 0.00010272, 0.00008574, 0.00004638, -0.00000674, -0.00005826, -0.00009211, -0.00009852, -0.00007852, -0.00004349, -0.00000977, 0.00000904, 0.00000818, -0.00000694, -0.00002461, -0.00003364, -0.00002928, -0.00001512, -0.00000007, 0.00000734, 0.00000387, -0.00000701, -0.00001800, -0.00002269, -0.00001938, -0.00001141, -0.00000427, -0.00000148, -0.00000233, -0.00000291, 0.00000035, 0.00000725, 0.00001313, 0.00001185, 0.00000049, -0.00001736, -0.00003234, -0.00003441, -0.00001933, 0.00000752, 0.00003276, 0.00004193, 0.00002795, -0.00000389, -0.00003773, -0.00005526, -0.00004589, -0.00001323, 0.00002616, 0.00005136, 0.00004842, 0.00001825, -0.00002357, -0.00005500, -0.00005933, -0.00003417, 0.00000749, 0.00004434, 0.00005820, 0.00004340, 0.00000917, -0.00002598, -0.00004442, -0.00003852, -0.00001386, 0.00001481, 0.00003202, 0.00002978, 0.00001097, -0.00001320, -0.00002994, -0.00003163, -0.00001914, -0.00000030, 0.00001501, 0.00002020, 0.00001471, 0.00000331, -0.00000731, -0.00001226, -0.00001057, -0.00000482, 0.00000096, 0.00000375, 0.00000289, -0.00000012, -0.00000292, -0.00000374, -0.00000231, 0.00000031, 0.00000242, 0.00000261, 0.00000043, -0.00000350, -0.00000765, -0.00001020, -0.00000970, -0.00000577, 0.00000056, 0.00000697, 0.00001061, 0.00000926, 0.00000235, -0.00000862, -0.00002062, -0.00003034, -0.00003557, -0.00003605, -0.00003321, -0.00002911, -0.00002517, -0.00002138, -0.00001642, -0.00000862, 0.00000303, 0.00001810, 0.00003477, 0.00005042, 0.00006237, 0.00006856, 0.00006802, 0.00006134, 0.00005092, 0.00004097, },
		{ 0.00008162, 0.00008681, 0.00007549, 0.00004613, 0.00000408, -0.00003942, -0.00007132, -0.00008232, -0.00007109, -0.00004510, -0.00001706, 0.00000101, 0.00000346, -0.00000690, -0.00002105, -0.00002919, -0.00002620, -0.00001414, -0.00000042, 0.00000687, 0.00000386, -0.00000706, -0.00001905, -0.00002506, -0.00002203, -0.00001224, -0.00000141, 0.00000516, 0.00000571, 0.00000252, -0.00000041, -0.00000064, 0.00000088, 0.00000074, -0.00000386, -0.00001213, -0.00001924, -0.00001915, -0.00000891, 0.00000819, 0.00002338, 0.00002699, 0.00001456, -0.00000948, -0.00003310, -0.00004287, -0.00003180, -0.00000411, 0.00002618, 0.00004245, 0.00003480, 0.00000607, -0.00002901, -0.00005154, -0.00004867, -0.00002091, 0.00001791, 0.00004792, 0.00005392, 0.00003343, -0.00000216, -0.00003403, -0.00004612, -0.00003356, -0.00000474, 0.00002388, 0.00003696, 0.00002832, 0.00000370, -0.00002320, -0.00003842, -0.00003510, -0.00001640, 0.00000713, 0.00002374, 0.00002658, 0.00001653, 0.00000067, -0.00001230, -0.00001675, -0.00001241, -0.00000355, 0.00000418, 0.00000697, 0.00000454, -0.00000048, -0.00000457, -0.00000548, -0.00000325, 0.00000018, 0.00000237, 0.00000179, -0.00000148, -0.00000590, -0.00000924, -0.00000963, -0.00000628, 0.00000012, 0.00000750, 0.00001299, 0.00001382, 0.00000846, -0.00000263, -0.00001682, -0.00003009, -0.00003855, -0.00004003, -0.00003492, -0.00002590, -0.00001655, -0.00000951, -0.00000535, -0.00000247, 0.00000174, 0.00000931, 0.00002048, 0.00003354, 0.00004560, 0.00005387, 0.00005668, 0.00005397, 0.00004720, 0.00003891, 0.00003232, 0.00003064, 0.00003620, 0.00004915, 0.00006641, },
		{ 0.00001219, -0.00002343, -0.00005264, -0.00006708, -0.00006373, -0.00004643, -0.00002423, -0.00000710, -0.00000120, -0.00000619, -0.00001610, -0.00002307, -0.00002204, -0.00001347, -0.00000270, 0.00000355, 0.00000142, -0.00000773, -0.00001822, -0.00002358, -0.00002040, -0.00001031, 0.00000127, 0.00000851, 0.00000858, 0.00000302, -0.00000377, -0.00000766, -0.00000739, -0.00000493, -0.00000350, -0.00000481, -0.00000748, -0.00000801, -0.00000371, 0.00000466, 0.00001247, 0.00001375, 0.00000535, -0.00001011, -0.00002476, -0.00002956, -0.00001999, 0.00000046, 0.00002125, 0.00003030, 0.00002110, -0.00000287, -0.00002916, -0.00004289, -0.00003508, -0.00000822, 0.00002450, 0.00004604, 0.00004482, 0.00002124, -0.00001232, -0.00003820, -0.00004305, -0.00002505, 0.00000525, 0.00003089, 0.00003790, 0.00002272, -0.00000616, -0.00003323, -0.00004449, -0.00003478, -0.00000995, 0.00001689, 0.00003273, 0.00003129, 0.00001558, -0.00000475, -0.00001913, -0.00002161, -0.00001323, -0.00000040, 0.00000936, 0.00001153, 0.00000649, -0.00000158, -0.00000770, -0.00000892, -0.00000567, -0.00000088, 0.00000210, 0.00000146, -0.00000220, -0.00000647, -0.00000855, -0.00000669, -0.00000100, 0.00000665, 0.00001330, 0.00001598, 0.00001270, 0.00000319, -0.00001076, -0.00002563, -0.00003712, -0.00004169, -0.00003798, -0.00002749, -0.00001406, -0.00000225, 0.00000470, 0.00000635, 0.00000505, 0.00000465, 0.00000831, 0.00001693, 0.00002866, 0.00003991, 0.00004716, 0.00004850, 0.00004435, 0.00003700, 0.00002960, 0.00002505, 0.00002539, 0.00003144, 0.00004251, 0.00005612, 0.00006798, 0.00007268, 0.00006542, 0.00004431, },
		{ -0.00005587, -0.00004610, -0.00002952, -0.00001388, -0.00000531, -0.00000552, -0.00001139, -0.00001720, -0.00001822, -0.00001344, -0.00000595, -0.00000074, -0.00000137, -0.00000757, -0.00001532, -0.00001938, -0.00001661, -0.00000793, 0.00000212, 0.00000821, 0.00000739, 0.00000081, -0.00000721, -0.00001193, -0.00001112, -0.00000613, -0.00000065, 0.00000196, 0.00000096, -0.00000161, -0.00000268, -0.00000078, 0.00000264, 0.00000407, 0.00000068, -0.00000712, -0.00001518, -0.00001790, -0.00001202, 0.00000057, 0.00001311, 0.00001758, 0.00000981, -0.00000728, -0.00002444, -0.00003100, -0.00002125, 0.00000156, 0.00002617, 0.00003922, 0.00003279, 0.00000930, -0.00001926, -0.00003766, -0.00003585, -0.00001479, 0.00001387, 0.00003409, 0.00003427, 0.00001376, -0.00001677, -0.00004116, -0.00004652, -0.00003010, -0.00000048, 0.00002733, 0.00004004, 0.00003284, 0.00001125, -0.00001254, -0.00002661, -0.00002540, -0.00001184, 0.00000520, 0.00001634, 0.00001667, 0.00000768, -0.00000447, -0.00001302, -0.00001431, -0.00000924, -0.00000201, 0.00000282, 0.00000296, -0.00000060, -0.00000464, -0.00000578, -0.00000246, 0.00000434, 0.00001169, 0.00001608, 0.00001490, 0.00000743, -0.00000504, -0.00001948, -0.00003203, -0.00003902, -0.00003819, -0.00002956, -0.00001575, -0.00000118, 0.00000954, 0.00001356, 0.00001115, 0.00000561, 0.00000173, 0.00000338, 0.00001161, 0.00002407, 0.00003619, 0.00004350, 0.00004371, 0.00003767, 0.00002870, 0.00002094, 0.00001758, 0.00001991, 0.00002732, 0.00003792, 0.00004910, 0.00005789, 0.00006107, 0.00005578, 0.00004055, 0.00001660, -0.00001158, -0.00003714, -0.00005322, },
		{ -0.00000935, -0.00000597, -0.00000827, -0.00001265, -0.00001503, -0.00001335, -0.00000869, -0.00000431, -0.00000329, -0.00000636, -0.00001128, -0.00001429, -0.00001265, -0.00000656, 0.00000081, 0.00000522, 0.00000405, -0.00000194, -0.00000906, -0.00001290, -0.00001110, -0.00000473, 0.00000245, 0.00000642, 0.00000539, 0.00000080, -0.00000395, -0.00000584, -0.00000433, -0.00000150, -0.00000040, -0.00000256, -0.00000670, -0.00000953, -0.00000813, -0.00000246, 0.00000411, 0.00000673, 0.00000244, -0.00000722, -0.00001646, -0.00001841, -0.00000962, 0.00000703, 0.00002312, 0.00002915, 0.00002027, -0.00000021, -0.00002189, -0.00003280, -0.00002623, -0.00000507, 0.00001921, 0.00003266, 0.00002670, 0.00000324, -0.00002586, -0.00004510, -0.00004371, -0.00002164, 0.00001032, 0.00003633, 0.00004390, 0.00003038, 0.00000391, -0.00002136, -0.00003299, -0.00002661, -0.00000749, 0.00001303, 0.00002411, 0.00002098, 0.00000681, -0.00000995, -0.00002065, -0.00002105, -0.00001284, -0.00000184, 0.00000584, 0.00000725, 0.00000366, -0.00000084, -0.00000219, 0.00000120, 0.00000769, 0.00001355, 0.00001500, 0.00001013, -0.00000037, -0.00001365, -0.00002589, -0.00003361, -0.00003460, -0.00002850, -0.00001692, -0.00000312, 0.00000884, 0.00001541, 0.00001500, 0.00000891, 0.00000103, -0.00000368, -0.00000150, 0.00000813, 0.00002214, 0.00003515, 0.00004205, 0.00004044, 0.00003179, 0.00002057, 0.00001205, 0.00000987, 0.00001472, 0.00002455, 0.00003589, 0.00004550, 0.00005112, 0.00005153, 0.00004603, 0.00003431, 0.00001691, -0.00000407, -0.00002473, -0.00004024, -0.00004672, -0.00004316, -0.00003219, -0.00001909, },
		{ -0.00001192, -0.00001217, -0.00000974, -0.00000626, -0.00000416, -0.00000483, -0.00000752, -0.00000993, -0.00000977, -0.00000651, -0.00000189, 0.00000115, 0.00000045, -0.00000377, -0.00000883, -0.00001124, -0.00000891, -0.00000264, 0.00000426, 0.00000795, 0.00000641, 0.00000074, -0.00000560, -0.00000898, -0.00000785, -0.00000361, 0.00000048, 0.00000157, -0.00000088, -0.00000487, -0.00000739, -0.00000667, -0.00000352, -0.00000069, -0.00000074, -0.00000391, -0.00000749, -0.00000752, -0.00000177, 0.00000809, 0.00001673, 0.00001812, 0.00000954, -0.00000593, -0.00002033, -0.00002514, -0.00001642, 0.00000205, 0.00002023, 0.00002702, 0.00001685, -0.00000653, -0.00003139, -0.00004404, -0.00003645, -0.00001102, 0.00002022, 0.00004185, 0.00004321, 0.00002412, -0.00000510, -0.00002928, -0.00003650, -0.00002426, -0.00000031, 0.00002196, 0.00003098, 0.00002286, 0.00000288, -0.00001809, -0.00002976, -0.00002768, -0.00001481, 0.00000104, 0.00001203, 0.00001440, 0.00000984, 0.00000347, 0.00000033, 0.00000235, 0.00000755, 0.00001160, 0.00001063, 0.00000342, -0.00000808, -0.00001985, -0.00002781, -0.00002945, -0.00002458, -0.00001494, -0.00000345, 0.00000675, 0.00001300, 0.00001379, 0.00000931, 0.00000167, -0.00000539, -0.00000785, -0.00000311, 0.00000845, 0.00002322, 0.00003563, 0.00004067, 0.00003642, 0.00002506, 0.00001204, 0.00000342, 0.00000308, 0.00001103, 0.00002377, 0.00003628, 0.00004443, 0.00004646, 0.00004296, 0.00003557, 0.00002565, 0.00001370, -0.00000002, -0.00001453, -0.00002750, -0.00003594, -0.00003760, -0.00003251, -0.00002329, -0.00001402, -0.00000826, -0.00000723, -0.00000944, },
		{ -0.00000462, -0.00000403, -0.00000526, -0.00000722, -0.00000820, -0.00000717, -0.00000467, -0.00000244, -0.00000219, -0.00000423, -0.00000708, -0.00000830, -0.00000621, -0.00000120, 0.00000428, 0.00000706, 0.00000533, -0.00000015, -0.00000633, -0.00000971, -0.00000846, -0.00000362, 0.00000164, 0.00000394, 0.00000187, -0.00000323, -0.00000820, -0.00001016, -0.00000832, -0.00000427, -0.00000067, 0.00000072, 0.00000029, 0.00000007, 0.00000195, 0.00000579, 0.00000904, 0.00000833, 0.00000207, -0.00000771, -0.00001581, -0.00001676, -0.00000853, 0.00000537, 0.00001709, 0.00001867, 0.00000701, -0.00001342, -0.00003217, -0.00003822, -0.00002630, -0.00000054, 0.00002705, 0.00004263, 0.00003801, 0.00001535, -0.00001375, -0.00003430, -0.00003590, -0.00001831, 0.00000856, 0.00003015, 0.00003509, 0.00002108, -0.00000427, -0.00002796, -0.00003863, -0.00003227, -0.00001356, 0.00000747, 0.00002122, 0.00002335, 0.00001619, 0.00000627, -0.00000006, -0.00000025, 0.00000358, 0.00000664, 0.00000482, -0.00000270, -0.00001309, -0.00002166, -0.00002451, -0.00002047, -0.00001141, -0.00000094, 0.00000732, 0.00001115, 0.00001015, 0.00000538, -0.00000121, -0.00000720, -0.00001013, -0.00000796, 0.00000004, 0.00001251, 0.00002588, 0.00003541, 0.00003716, 0.00003011, 0.00001708, 0.00000383, -0.00000349, -0.00000128, 0.00000974, 0.00002490, 0.00003797, 0.00004418, 0.00004225, 0.00003439, 0.00002445, 0.00001553, 0.00000847, 0.00000200, -0.00000563, -0.00001485, -0.00002404, -0.00003035, -0.00003147, -0.00002718, -0.00001967, -0.00001232, -0.00000784, -0.00000691, -0.00000812, -0.00000923, -0.00000870, -0.00000668, },
		{ -0.00000733, -0.00000752, -0.00000636, -0.00000463, -0.00000354, -0.00000380, -0.00000491, -0.00000546, -0.00000412, -0.00000084, 0.00000291, 0.00000488, 0.00000354, -0.00000075, -0.00000572, -0.00000848, -0.00000734, -0.00000294, 0.00000196, 0.00000420, 0.00000213, -0.00000326, -0.00000894, -0.00001167, -0.00000992, -0.00000473, 0.00000108, 0.00000479, 0.00000538, 0.00000389, 0.00000229, 0.00000184, 0.00000208, 0.00000131, -0.00000177, -0.00000652, -0.00001022, -0.00000972, -0.00000389, 0.00000483, 0.00001112, 0.00000979, -0.00000065, -0.00001609, -0.00002824, -0.00002903, -0.00001550, 0.00000756, 0.00002936, 0.00003853, 0.00002952, 0.00000613, -0.00001992, -0.00003501, -0.00003100, -0.00000985, 0.00001720, 0.00003554, 0.00003499, 0.00001534, -0.00001370, -0.00003768, -0.00004507, -0.00003299, -0.00000818, 0.00001717, 0.00003207, 0.00003205, 0.00002052, 0.00000573, -0.00000443, -0.00000663, -0.00000314, 0.00000062, 0.00000002, -0.00000580, -0.00001366, -0.00001856, -0.00001702, -0.00000911, 0.00000158, 0.00001013, 0.00001288, 0.00000918, 0.00000124, -0.00000726, -0.00001306, -0.00001428, -0.00001066, -0.00000295, 0.00000745, 0.00001855, 0.00002780, 0.00003236, 0.00003018, 0.00002133, 0.00000873, -0.00000249, -0.00000705, -0.00000214, 0.00001089, 0.00002687, 0.00003916, 0.00004286, 0.00003719, 0.00002558, 0.00001356, 0.00000567, 0.00000306, 0.00000332, 0.00000234, -0.00000290, -0.00001233, -0.00002282, -0.00003006, -0.00003108, -0.00002587, -0.00001723, -0.00000906, -0.00000422, -0.00000323, -0.00000458, -0.00000607, -0.00000637, -0.00000556, -0.00000477, -0.00000495, -0.00000612, },
		{ -0.00000408, -0.00000336, -0.00000338, -0.00000355, -0.00000296, -0.00000120, 0.00000111, 0.00000259, 0.00000201, -0.00000070, -0.00000418, -0.00000634, -0.00000571, -0.00000255, 0.00000118, 0.00000290, 0.00000111, -0.00000357, -0.00000857, -0.00001090, -0.00000895, -0.00000346, 0.00000287, 0.00000701, 0.00000736, 0.00000454, 0.00000071, -0.00000198, -0.00000278, -0.00000253, -0.00000271, -0.00000397, -0.00000543, -0.00000536, -0.00000271, 0.00000151, 0.00000430, 0.00000252, -0.00000467, -0.00001437, -0.00002085, -0.00001868, -0.00000638, 0.00001168, 0.00002687, 0.00003069, 0.00001981, -0.00000131, -0.00002204, -0.00003111, -0.00002283, -0.00000082, 0.00002345, 0.00003652, 0.00003024, 0.00000651, -0.00002347, -0.00004490, -0.00004704, -0.00002880, 0.00000105, 0.00002864, 0.00004229, 0.00003807, 0.00002089, 0.00000089, -0.00001252, -0.00001543, -0.00001046, -0.00000397, -0.00000152, -0.00000445, -0.00000954, -0.00001172, -0.00000774, 0.00000147, 0.00001123, 0.00001597, 0.00001249, 0.00000192, -0.00001104, -0.00002059, -0.00002281, -0.00001717, -0.00000619, 0.00000631, 0.00001702, 0.00002409, 0.00002687, 0.00002536, 0.00001983, 0.00001122, 0.00000172, -0.00000528, -0.00000627, 0.00000047, 0.00001347, 0.00002799, 0.00003801, 0.00003910, 0.00003084, 0.00001719, 0.00000445, -0.00000208, -0.00000099, 0.00000460, 0.00000893, 0.00000703, -0.00000243, -0.00001632, -0.00002881, -0.00003455, -0.00003147, -0.00002163, -0.00000985, -0.00000097, 0.00000244, 0.00000090, -0.00000295, -0.00000625, -0.00000754, -0.00000710, -0.00000619, -0.00000588, -0.00000628, -0.00000671, -0.00000643, -0.00000535, },
		{ -0.00000235, -0.00000152, -0.00000021, 0.00000092, 0.00000100, -0.00000036, -0.00000256, -0.00000430, -0.00000439, -0.00000265, -0.00000022, 0.00000110, 0.00000001, -0.00000325, -0.00000687, -0.00000848, -0.00000666, -0.00000189, 0.00000357, 0.00000699, 0.00000676, 0.00000331, -0.00000122, -0.00000443, -0.00000509, -0.00000375, -0.00000213, -0.00000172, -0.00000271, -0.00000396, -0.00000411, -0.00000282, -0.00000136, -0.00000168, -0.00000477, -0.00000935, -0.00001206, -0.00000947, -0.00000064, 0.00001140, 0.00002060, 0.00002113, 0.00001115, -0.00000529, -0.00001968, -0.00002351, -0.00001333, 0.00000653, 0.00002558, 0.00003254, 0.00002168, -0.00000343, -0.00003113, -0.00004745, -0.00004343, -0.00001995, 0.00001254, 0.00003942, 0.00004926, 0.00003935, 0.00001637, -0.00000783, -0.00002273, -0.00002426, -0.00001581, -0.00000501, 0.00000137, 0.00000130, -0.00000210, -0.00000353, 0.00000036, 0.00000842, 0.00001555, 0.00001601, 0.00000728, -0.00000802, -0.00002324, -0.00003125, -0.00002815, -0.00001519, 0.00000221, 0.00001748, 0.00002602, 0.00002685, 0.00002200, 0.00001472, 0.00000754, 0.00000171, -0.00000225, -0.00000364, -0.00000132, 0.00000546, 0.00001580, 0.00002656, 0.00003333, 0.00003254, 0.00002379, 0.00001054, -0.00000126, -0.00000632, -0.00000301, 0.00000557, 0.00001308, 0.00001346, 0.00000433, -0.00001149, -0.00002742, -0.00003650, -0.00003504, -0.00002431, -0.00000965, 0.00000250, 0.00000783, 0.00000587, -0.00000044, -0.00000690, -0.00001042, -0.00001031, -0.00000799, -0.00000556, -0.00000440, -0.00000456, -0.00000513, -0.00000523, -0.00000460, -0.00000368, -0.00000301, -0.00000269, },
		{ 0.00000015, -0.00000039, -0.00000165, -0.00000297, -0.00000350, -0.00000281, -0.00000134, -0.00000019, -0.00000041, -0.00000215, -0.00000440, -0.00000552, -0.00000433, -0.00000100, 0.00000290, 0.00000526, 0.00000472, 0.00000156, -0.00000245, -0.00000516, -0.00000538, -0.00000357, -0.00000149, -0.00000089, -0.00000233, -0.00000480, -0.00000653, -0.00000633, -0.00000445, -0.00000242, -0.00000182, -0.00000288, -0.00000408, -0.00000313, 0.00000116, 0.00000753, 0.00001245, 0.00001215, 0.00000535, -0.00000520, -0.00001364, -0.00001415, -0.00000483, 0.00001035, 0.00002293, 0.00002439, 0.00001127, -0.00001197, -0.00003444, -0.00004415, -0.00003446, -0.00000809, 0.00002376, 0.00004679, 0.00005086, 0.00003497, 0.00000746, -0.00001865, -0.00003248, -0.00003045, -0.00001711, -0.00000152, 0.00000834, 0.00000984, 0.00000616, 0.00000314, 0.00000474, 0.00001011, 0.00001425, 0.00001150, -0.00000023, -0.00001739, -0.00003225, -0.00003706, -0.00002844, -0.00000941, 0.00001219, 0.00002771, 0.00003204, 0.00002565, 0.00001355, 0.00000205, -0.00000449, -0.00000526, -0.00000219, 0.00000217, 0.00000649, 0.00001096, 0.00001616, 0.00002155, 0.00002501, 0.00002397, 0.00001736, 0.00000706, -0.00000251, -0.00000668, -0.00000335, 0.00000529, 0.00001358, 0.00001535, 0.00000747, -0.00000808, -0.00002496, -0.00003565, -0.00003542, -0.00002468, -0.00000872, 0.00000521, 0.00001158, 0.00000911, 0.00000080, -0.00000815, -0.00001334, -0.00001319, -0.00000922, -0.00000450, -0.00000168, -0.00000159, -0.00000331, -0.00000511, -0.00000573, -0.00000501, -0.00000362, -0.00000238, -0.00000162, -0.00000114, -0.00000062, -0.00000006, },
		{ -0.00000263, -0.00000233, -0.00000141, -0.00000041, -0.00000008, -0.00000077, -0.00000211, -0.00000310, -0.00000281, -0.00000105, 0.00000131, 0.00000284, 0.00000248, 0.00000024, -0.00000267, -0.00000464, -0.00000472, -0.00000327, -0.00000175, -0.00000171, -0.00000363, -0.00000648, -0.00000836, -0.00000775, -0.00000459, -0.00000045, 0.00000251, 0.00000298, 0.00000140, -0.00000034, -0.00000035, 0.00000183, 0.00000464, 0.00000561, 0.00000318, -0.00000170, -0.00000576, -0.00000539, 0.00000070, 0.00000983, 0.00001617, 0.00001402, 0.00000162, -0.00001676, -0.00003212, -0.00003521, -0.00002181, 0.00000417, 0.00003187, 0.00004856, 0.00004615, 0.00002549, -0.00000394, -0.00002888, -0.00003903, -0.00003192, -0.00001347, 0.00000597, 0.00001759, 0.00001855, 0.00001246, 0.00000602, 0.00000412, 0.00000663, 0.00000877, 0.00000491, -0.00000707, -0.00002343, -0.00003619, -0.00003754, -0.00002473, -0.00000207, 0.00002095, 0.00003457, 0.00003379, 0.00002074, 0.00000309, -0.00001033, -0.00001411, -0.00000851, 0.00000172, 0.00001074, 0.00001510, 0.00001508, 0.00001348, 0.00001300, 0.00001411, 0.00001494, 0.00001302, 0.00000760, 0.00000071, -0.00000385, -0.00000315, 0.00000262, 0.00000967, 0.00001255, 0.00000747, -0.00000511, -0.00002029, -0.00003109, -0.00003222, -0.00002306, -0.00000800, 0.00000588, 0.00001254, 0.00000998, 0.00000090, -0.00000917, -0.00001506, -0.00001458, -0.00000921, -0.00000273, 0.00000124, 0.00000117, -0.00000194, -0.00000561, -0.00000754, -0.00000693, -0.00000454, -0.00000190, -0.00000024, 0.00000014, -0.00000027, -0.00000079, -0.00000112, -0.00000136, -0.00000177, -0.00000231, },
		{ 0.00000049, 0.00000009, -0.00000086, -0.00000186, -0.00000222, -0.00000161, -0.00000033, 0.00000076, 0.00000081, -0.00000041, -0.00000230, -0.00000382, -0.00000423, -0.00000366, -0.00000308, -0.00000356, -0.00000537, -0.00000757, -0.00000848, -0.00000676, -0.00000246, 0.00000274, 0.00000639, 0.00000669, 0.00000372, -0.00000060, -0.00000363, -0.00000366, -0.00000099, 0.00000237, 0.00000418, 0.00000357, 0.00000168, 0.00000078, 0.00000232, 0.00000547, 0.00000716, 0.00000404, -0.00000485, -0.00001641, -0.00002431, -0.00002231, -0.00000819, 0.00001392, 0.00003465, 0.00004381, 0.00003578, 0.00001292, -0.00001505, -0.00003575, -0.00004033, -0.00002790, -0.00000559, 0.00001551, 0.00002643, 0.00002467, 0.00001464, 0.00000408, -0.00000111, -0.00000039, 0.00000165, -0.00000092, -0.00001075, -0.00002461, -0.00003482, -0.00003382, -0.00001917, 0.00000423, 0.00002626, 0.00003681, 0.00003139, 0.00001353, -0.00000714, -0.00002019, -0.00001991, -0.00000796, 0.00000820, 0.00001998, 0.00002242, 0.00001651, 0.00000771, 0.00000206, 0.00000245, 0.00000720, 0.00001174, 0.00001205, 0.00000736, 0.00000061, -0.00000380, -0.00000317, 0.00000149, 0.00000598, 0.00000567, -0.00000133, -0.00001263, -0.00002270, -0.00002604, -0.00002052, -0.00000870, 0.00000353, 0.00001025, 0.00000872, 0.00000071, -0.00000877, -0.00001447, -0.00001381, -0.00000799, -0.00000093, 0.00000328, 0.00000269, -0.00000167, -0.00000675, -0.00000950, -0.00000859, -0.00000492, -0.00000075, 0.00000181, 0.00000191, 0.00000019, -0.00000191, -0.00000321, -0.00000332, -0.00000263, -0.00000178, -0.00000112, -0.00000063, -0.00000015, 0.00000032, },
		{ -0.00000213, -0.00000192, -0.00000118, -0.00000039, -0.00000014, -0.00000080, -0.00000218, -0.00000364, -0.00000458, -0.00000485, -0.00000489, -0.00000531, -0.00000628, -0.00000713, -0.00000667, -0.00000401, 0.00000060, 0.00000552, 0.00000843, 0.00000765, 0.00000329, -0.00000254, -0.00000682, -0.00000724, -0.00000359, 0.00000212, 0.00000690, 0.00000851, 0.00000674, 0.00000328, 0.00000033, -0.00000103, -0.00000157, -0.00000309, -0.00000662, -0.00001093, -0.00001264, -0.00000818, 0.00000341, 0.00001882, 0.00003115, 0.00003326, 0.00002186, 0.00000013, -0.00002305, -0.00003724, -0.00003572, -0.00001919, 0.00000449, 0.00002437, 0.00003214, 0.00002619, 0.00001190, -0.00000208, -0.00000913, -0.00000832, -0.00000435, -0.00000376, -0.00001009, -0.00002102, -0.00002947, -0.00002804, -0.00001418, 0.00000748, 0.00002700, 0.00003457, 0.00002614, 0.00000607, -0.00001503, -0.00002610, -0.00002164, -0.00000465, 0.00001533, 0.00002775, 0.00002695, 0.00001495, -0.00000016, -0.00000938, -0.00000799, 0.00000202, 0.00001360, 0.00001926, 0.00001557, 0.00000492, -0.00000632, -0.00001196, -0.00000978, -0.00000263, 0.00000367, 0.00000420, -0.00000200, -0.00001138, -0.00001827, -0.00001847, -0.00001179, -0.00000209, 0.00000520, 0.00000634, 0.00000143, -0.00000595, -0.00001108, -0.00001104, -0.00000628, -0.00000014, 0.00000344, 0.00000236, -0.00000249, -0.00000795, -0.00001067, -0.00000910, -0.00000426, 0.00000103, 0.00000398, 0.00000336, 0.00000004, -0.00000379, -0.00000603, -0.00000576, -0.00000352, -0.00000069, 0.00000143, 0.00000229, 0.00000208, 0.00000136, 0.00000055, -0.00000024, -0.00000104, -0.00000176, },
		{ -0.00000084, -0.00000166, -0.00000303, -0.00000456, -0.00000572, -0.00000622, -0.00000615, -0.00000587, -0.00000558, -0.00000502, -0.00000357, -0.00000076, 0.00000309, 0.00000666, 0.00000815, 0.00000629, 0.00000139, -0.00000450, -0.00000844, -0.00000815, -0.00000343, 0.00000359, 0.00000947, 0.00001132, 0.00000838, 0.00000233, -0.00000378, -0.00000736, -0.00000767, -0.00000580, -0.00000355, -0.00000177, 0.00000026, 0.00000415, 0.00001057, 0.00001784, 0.00002204, 0.00001904, 0.00000736, -0.00000997, -0.00002594, -0.00003277, -0.00002614, -0.00000797, 0.00001399, 0.00002990, 0.00003285, 0.00002249, 0.00000501, -0.00001051, -0.00001728, -0.00001447, -0.00000715, -0.00000262, -0.00000541, -0.00001416, -0.00002234, -0.00002256, -0.00001159, 0.00000679, 0.00002348, 0.00002915, 0.00001995, 0.00000028, -0.00001927, -0.00002775, -0.00002012, -0.00000024, 0.00002116, 0.00003256, 0.00002817, 0.00001112, -0.00000845, -0.00001934, -0.00001569, -0.00000021, 0.00001768, 0.00002738, 0.00002326, 0.00000761, -0.00001096, -0.00002269, -0.00002221, -0.00001126, 0.00000283, 0.00001166, 0.00001064, 0.00000118, -0.00001065, -0.00001806, -0.00001749, -0.00001029, -0.00000134, 0.00000423, 0.00000407, -0.00000036, -0.00000525, -0.00000717, -0.00000520, -0.00000133, 0.00000123, 0.00000031, -0.00000372, -0.00000824, -0.00001016, -0.00000795, -0.00000264, 0.00000282, 0.00000538, 0.00000369, -0.00000113, -0.00000627, -0.00000892, -0.00000778, -0.00000363, 0.00000135, 0.00000492, 0.00000587, 0.00000447, 0.00000192, -0.00000045, -0.00000189, -0.00000236, -0.00000221, -0.00000181, -0.00000135, -0.00000092, -0.00000066, },
		{ -0.00000699, -0.00000683, -0.00000595, -0.00000466, -0.00000330, -0.00000195, -0.00000038, 0.00000164, 0.00000398, 0.00000582, 0.00000603, 0.00000382, -0.00000047, -0.00000515, -0.00000783, -0.00000669, -0.00000169, 0.00000508, 0.00001029, 0.00001109, 0.00000677, -0.00000076, -0.00000792, -0.00001134, -0.00000961, -0.00000388, 0.00000308, 0.00000859, 0.00001147, 0.00001226, 0.00001215, 0.00001153, 0.00000941, 0.00000416, -0.00000476, -0.00001533, -0.00002313, -0.00002337, -0.00001386, 0.00000294, 0.00002032, 0.00003038, 0.00002815, 0.00001452, -0.00000398, -0.00001856, -0.00002298, -0.00001692, -0.00000595, 0.00000202, 0.00000169, -0.00000629, -0.00001579, -0.00001912, -0.00001216, 0.00000258, 0.00001706, 0.00002242, 0.00001462, -0.00000266, -0.00001962, -0.00002591, -0.00001682, 0.00000357, 0.00002434, 0.00003387, 0.00002658, 0.00000638, -0.00001539, -0.00002632, -0.00002007, -0.00000004, 0.00002240, 0.00003438, 0.00002876, 0.00000804, -0.00001704, -0.00003360, -0.00003365, -0.00001820, 0.00000373, 0.00002034, 0.00002356, 0.00001297, -0.00000469, -0.00001976, -0.00002512, -0.00001958, -0.00000768, 0.00000361, 0.00000902, 0.00000749, 0.00000187, -0.00000350, -0.00000576, -0.00000488, -0.00000305, -0.00000263, -0.00000434, -0.00000676, -0.00000756, -0.00000532, -0.00000072, 0.00000370, 0.00000517, 0.00000247, -0.00000315, -0.00000865, -0.00001090, -0.00000850, -0.00000252, 0.00000420, 0.00000863, 0.00000913, 0.00000609, 0.00000145, -0.00000252, -0.00000439, -0.00000406, -0.00000244, -0.00000073, 0.00000023, 0.00000020, -0.00000065, -0.00000199, -0.00000356, -0.00000511, -0.00000635, },
		{ -0.00000021, 0.00000094, 0.00000170, 0.00000235, 0.00000305, 0.00000354, 0.00000324, 0.00000164, -0.00000112, -0.00000399, -0.00000537, -0.00000400, 0.00000008, 0.00000517, 0.00000858, 0.00000805, 0.00000316, -0.00000409, -0.00001013, -0.00001161, -0.00000722, 0.00000152, 0.00001101, 0.00001744, 0.00001867, 0.00001499, 0.00000857, 0.00000184, -0.00000383, -0.00000842, -0.00001236, -0.00001531, -0.00001565, -0.00001135, -0.00000177, 0.00001091, 0.00002183, 0.00002555, 0.00001916, 0.00000443, -0.00001246, -0.00002380, -0.00002451, -0.00001505, -0.00000135, 0.00000851, 0.00000896, 0.00000030, -0.00001150, -0.00001842, -0.00001543, -0.00000375, 0.00000962, 0.00001617, 0.00001128, -0.00000257, -0.00001685, -0.00002198, -0.00001332, 0.00000556, 0.00002441, 0.00003212, 0.00002338, 0.00000225, -0.00001969, -0.00002973, -0.00002146, 0.00000126, 0.00002602, 0.00003867, 0.00003123, 0.00000652, -0.00002318, -0.00004273, -0.00004230, -0.00002252, 0.00000606, 0.00002880, 0.00003479, 0.00002226, -0.00000118, -0.00002314, -0.00003324, -0.00002813, -0.00001233, 0.00000509, 0.00001580, 0.00001630, 0.00000875, -0.00000126, -0.00000842, -0.00001040, -0.00000837, -0.00000525, -0.00000338, -0.00000312, -0.00000315, -0.00000201, 0.00000047, 0.00000281, 0.00000297, -0.00000007, -0.00000527, -0.00000990, -0.00001104, -0.00000744, -0.00000034, 0.00000711, 0.00001151, 0.00001104, 0.00000634, -0.00000004, -0.00000506, -0.00000680, -0.00000522, -0.00000187, 0.00000117, 0.00000238, 0.00000139, -0.00000113, -0.00000411, -0.00000656, -0.00000797, -0.00000823, -0.00000745, -0.00000589, -0.00000389, -0.00000187, },
		{ 0.00000092, 0.00000095, 0.00000107, 0.00000078, -0.00000017, -0.00000144, -0.00000219, -0.00000156, 0.00000057, 0.00000322, 0.00000468, 0.00000343, -0.00000066, -0.00000587, -0.00000925, -0.00000809, -0.00000161, 0.00000832, 0.00001778, 0.00002268, 0.00002076, 0.00001269, 0.00000161, -0.00000853, -0.00001487, -0.00001661, -0.00001476, -0.00001082, -0.00000572, 0.00000043, 0.00000752, 0.00001433, 0.00001832, 0.00001665, 0.00000815, -0.00000513, -0.00001807, -0.00002471, -0.00002149, -0.00000963, 0.00000499, 0.00001470, 0.00001430, 0.00000407, -0.00001011, -0.00002007, -0.00002009, -0.00001040, 0.00000292, 0.00001155, 0.00001016, -0.00000017, -0.00001232, -0.00001747, -0.00001074, 0.00000536, 0.00002184, 0.00002842, 0.00001989, -0.00000019, -0.00002087, -0.00002990, -0.00002092, 0.00000227, 0.00002723, 0.00003951, 0.00003083, 0.00000406, -0.00002782, -0.00004853, -0.00004719, -0.00002423, 0.00000878, 0.00003535, 0.00004266, 0.00002807, -0.00000001, -0.00002702, -0.00004008, -0.00003428, -0.00001444, 0.00000840, 0.00002319, 0.00002440, 0.00001392, -0.00000110, -0.00001282, -0.00001676, -0.00001332, -0.00000640, -0.00000038, 0.00000234, 0.00000211, 0.00000081, -0.00000000, -0.00000021, -0.00000090, -0.00000308, -0.00000642, -0.00000911, -0.00000893, -0.00000481, 0.00000214, 0.00000899, 0.00001253, 0.00001102, 0.00000525, -0.00000191, -0.00000708, -0.00000815, -0.00000525, -0.00000050, 0.00000331, 0.00000417, 0.00000178, -0.00000265, -0.00000716, -0.00001009, -0.00001069, -0.00000917, -0.00000636, -0.00000317, -0.00000035, 0.00000162, 0.00000256, 0.00000256, 0.00000197, 0.00000129, },
		{ 0.00000189, 0.00000139, 0.00000035, -0.00000053, -0.00000074, -0.00000039, -0.00000023, -0.00000107, -0.00000304, -0.00000508, -0.00000527, -0.00000190, 0.00000524, 0.00001421, 0.00002146, 0.00002336, 0.00001815, 0.00000703, -0.00000624, -0.00001697, -0.00002173, -0.00001972, -0.00001275, -0.00000394, 0.00000392, 0.00000942, 0.00001238, 0.00001301, 0.00001107, 0.00000598, -0.00000223, -0.00001185, -0.00001944, -0.00002111, -0.00001489, -0.00000245, 0.00001094, 0.00001866, 0.00001633, 0.00000450, -0.00001121, -0.00002284, -0.00002450, -0.00001574, -0.00000192, 0.00000888, 0.00001069, 0.00000333, -0.00000747, -0.00001352, -0.00000950, 0.00000339, 0.00001770, 0.00002406, 0.00001714, -0.00000054, -0.00001936, -0.00002788, -0.00001977, 0.00000188, 0.00002548, 0.00003711, 0.00002844, 0.00000195, -0.00002972, -0.00005023, -0.00004833, -0.00002416, 0.00001056, 0.00003861, 0.00004626, 0.00003036, -0.00000029, -0.00002991, -0.00004410, -0.00003709, -0.00001402, 0.00001254, 0.00002957, 0.00003024, 0.00001654, -0.00000287, -0.00001792, -0.00002244, -0.00001655, -0.00000544, 0.00000436, 0.00000857, 0.00000689, 0.00000205, -0.00000256, -0.00000500, -0.00000549, -0.00000538, -0.00000563, -0.00000589, -0.00000484, -0.00000147, 0.00000378, 0.00000885, 0.00001113, 0.00000913, 0.00000349, -0.00000315, -0.00000759, -0.00000782, -0.00000414, 0.00000107, 0.00000473, 0.00000467, 0.00000078, -0.00000516, -0.00001050, -0.00001308, -0.00001215, -0.00000845, -0.00000357, 0.00000082, 0.00000365, 0.00000460, 0.00000395, 0.00000235, 0.00000053, -0.00000083, -0.00000128, -0.00000075, 0.00000040, 0.00000149, },
		{ -0.00000345, -0.00000451, -0.00000453, -0.00000393, -0.00000307, -0.00000179, 0.00000067, 0.00000503, 0.00001116, 0.00001748, 0.00002128, 0.00001995, 0.00001245, 0.00000032, -0.00001259, -0.00002172, -0.00002383, -0.00001850, -0.00000824, 0.00000282, 0.00001090, 0.00001409, 0.00001264, 0.00000814, 0.00000230, -0.00000384, -0.00000969, -0.00001443, -0.00001650, -0.00001413, -0.00000660, 0.00000438, 0.00001468, 0.00001924, 0.00001480, 0.00000213, -0.00001364, -0.00002530, -0.00002727, -0.00001879, -0.00000457, 0.00000774, 0.00001189, 0.00000666, -0.00000336, -0.00001062, -0.00000935, 0.00000055, 0.00001318, 0.00002006, 0.00001553, 0.00000080, -0.00001616, -0.00002494, -0.00001904, -0.00000035, 0.00002111, 0.00003244, 0.00002530, 0.00000123, -0.00002837, -0.00004802, -0.00004656, -0.00002348, 0.00001028, 0.00003793, 0.00004567, 0.00002994, -0.00000074, -0.00003048, -0.00004445, -0.00003650, -0.00001189, 0.00001610, 0.00003351, 0.00003290, 0.00001656, -0.00000568, -0.00002227, -0.00002603, -0.00001726, -0.00000249, 0.00000988, 0.00001419, 0.00000996, 0.00000114, -0.00000689, -0.00001055, -0.00000948, -0.00000588, -0.00000246, -0.00000057, 0.00000025, 0.00000137, 0.00000356, 0.00000617, 0.00000751, 0.00000615, 0.00000213, -0.00000277, -0.00000602, -0.00000587, -0.00000253, 0.00000187, 0.00000447, 0.00000330, -0.00000154, -0.00000796, -0.00001300, -0.00001440, -0.00001161, -0.00000594, 0.00000031, 0.00000497, 0.00000685, 0.00000605, 0.00000352, 0.00000055, -0.00000179, -0.00000284, -0.00000243, -0.00000082, 0.00000135, 0.00000321, 0.00000398, 0.00000323, 0.00000120, -0.00000134, },
		{ -0.00000091, 0.00000288, 0.00000675, 0.00001072, 0.00001459, 0.00001746, 0.00001778, 0.00001403, 0.00000585, -0.00000522, -0.00001589, -0.00002233, -0.00002195, -0.00001475, -0.00000354, 0.00000732, 0.00001389, 0.00001432, 0.00000942, 0.00000183, -0.00000550, -0.00001064, -0.00001301, -0.00001286, -0.00001044, -0.00000570, 0.00000122, 0.00000908, 0.00001522, 0.00001635, 0.00001039, -0.00000181, -0.00001603, -0.00002625, -0.00002764, -0.00001936, -0.00000538, 0.00000737, 0.00001276, 0.00000895, -0.00000048, -0.00000866, -0.00000964, -0.00000218, 0.00000919, 0.00001685, 0.00001483, 0.00000295, -0.00001249, -0.00002214, -0.00001920, -0.00000408, 0.00001511, 0.00002678, 0.00002248, 0.00000239, -0.00002408, -0.00004290, -0.00004313, -0.00002327, 0.00000741, 0.00003354, 0.00004178, 0.00002807, -0.00000018, -0.00002805, -0.00004117, -0.00003329, -0.00000930, 0.00001780, 0.00003414, 0.00003229, 0.00001471, -0.00000826, -0.00002451, -0.00002666, -0.00001541, 0.00000164, 0.00001485, 0.00001787, 0.00001058, -0.00000176, -0.00001205, -0.00001547, -0.00001165, -0.00000404, 0.00000276, 0.00000578, 0.00000497, 0.00000253, 0.00000092, 0.00000120, 0.00000256, 0.00000327, 0.00000221, -0.00000025, -0.00000253, -0.00000304, -0.00000146, 0.00000093, 0.00000200, 0.00000017, -0.00000443, -0.00000990, -0.00001353, -0.00001330, -0.00000900, -0.00000230, 0.00000412, 0.00000796, 0.00000824, 0.00000558, 0.00000160, -0.00000191, -0.00000378, -0.00000364, -0.00000189, 0.00000070, 0.00000322, 0.00000484, 0.00000491, 0.00000322, 0.00000012, -0.00000348, -0.00000640, -0.00000765, -0.00000686, -0.00000437, },
		{ 0.00001501, 0.00001472, 0.00001247, 0.00000769, 0.00000040, -0.00000822, -0.00001583, -0.00001965, -0.00001783, -0.00001061, -0.00000055, 0.00000845, 0.00001287, 0.00001123, 0.00000468, -0.00000366, -0.00001039, -0.00001332, -0.00001222, -0.00000830, -0.00000317, 0.00000203, 0.00000681, 0.00001066, 0.00001253, 0.00001083, 0.00000445, -0.00000593, -0.00001730, -0.00002513, -0.00002561, -0.00001787, -0.00000505, 0.00000699, 0.00001269, 0.00000989, 0.00000125, -0.00000721, -0.00000966, -0.00000411, 0.00000615, 0.00001435, 0.00001439, 0.00000493, -0.00000933, -0.00002002, -0.00002011, -0.00000849, 0.00000874, 0.00002129, 0.00002060, 0.00000518, -0.00001785, -0.00003622, -0.00003927, -0.00002413, 0.00000223, 0.00002645, 0.00003597, 0.00002596, 0.00000205, -0.00002274, -0.00003512, -0.00002874, -0.00000749, 0.00001688, 0.00003144, 0.00002911, 0.00001221, -0.00000935, -0.00002384, -0.00002424, -0.00001171, 0.00000570, 0.00001799, 0.00001885, 0.00000878, -0.00000580, -0.00001667, -0.00001848, -0.00001131, -0.00000004, 0.00000903, 0.00001173, 0.00000803, 0.00000136, -0.00000390, -0.00000506, -0.00000239, 0.00000165, 0.00000430, 0.00000421, 0.00000204, -0.00000049, -0.00000196, -0.00000225, -0.00000246, -0.00000386, -0.00000670, -0.00000986, -0.00001139, -0.00000978, -0.00000502, 0.00000125, 0.00000654, 0.00000881, 0.00000748, 0.00000360, -0.00000079, -0.00000377, -0.00000436, -0.00000271, 0.00000015, 0.00000300, 0.00000487, 0.00000523, 0.00000395, 0.00000126, -0.00000229, -0.00000582, -0.00000825, -0.00000866, -0.00000667, -0.00000267, 0.00000236, 0.00000727, 0.00001119, 0.00001378, },
		{ -0.00000269, -0.00000847, -0.00001326, -0.00001534, -0.00001347, -0.00000780, -0.00000018, 0.00000635, 0.00000903, 0.00000670, 0.00000052, -0.00000661, -0.00001145, -0.00001203, -0.00000845, -0.00000255, 0.00000329, 0.00000738, 0.00000922, 0.00000914, 0.00000748, 0.00000405, -0.00000157, -0.00000911, -0.00001690, -0.00002203, -0.00002171, -0.00001509, -0.00000428, 0.00000617, 0.00001156, 0.00000966, 0.00000222, -0.00000578, -0.00000900, -0.00000501, 0.00000396, 0.00001211, 0.00001352, 0.00000598, -0.00000716, -0.00001858, -0.00002116, -0.00001252, 0.00000309, 0.00001669, 0.00001962, 0.00000880, -0.00001098, -0.00002931, -0.00003577, -0.00002598, -0.00000434, 0.00001808, 0.00002962, 0.00002444, 0.00000589, -0.00001544, -0.00002763, -0.00002412, -0.00000723, 0.00001334, 0.00002617, 0.00002459, 0.00001024, -0.00000819, -0.00002019, -0.00001947, -0.00000731, 0.00000852, 0.00001854, 0.00001707, 0.00000530, -0.00000975, -0.00001952, -0.00001881, -0.00000854, 0.00000521, 0.00001492, 0.00001590, 0.00000864, -0.00000202, -0.00000994, -0.00001122, -0.00000603, 0.00000200, 0.00000823, 0.00000969, 0.00000636, 0.00000062, -0.00000457, -0.00000742, -0.00000797, -0.00000748, -0.00000714, -0.00000715, -0.00000666, -0.00000467, -0.00000097, 0.00000340, 0.00000664, 0.00000727, 0.00000507, 0.00000120, -0.00000238, -0.00000398, -0.00000306, -0.00000032, 0.00000278, 0.00000486, 0.00000517, 0.00000376, 0.00000118, -0.00000189, -0.00000482, -0.00000705, -0.00000800, -0.00000718, -0.00000437, 0.00000008, 0.00000521, 0.00000981, 0.00001284, 0.00001386, 0.00001300, 0.00001069, 0.00000725, 0.00000275, },
		{ -0.00001041, -0.00000707, -0.00000221, 0.00000222, 0.00000418, 0.00000267, -0.00000157, -0.00000632, -0.00000897, -0.00000793, -0.00000345, 0.00000250, 0.00000741, 0.00000950, 0.00000847, 0.00000532, 0.00000139, -0.00000255, -0.00000652, -0.00001078, -0.00001492, -0.00001752, -0.00001677, -0.00001173, -0.00000348, 0.00000488, 0.00000966, 0.00000867, 0.00000281, -0.00000417, -0.00000768, -0.00000514, 0.00000219, 0.00000963, 0.00001175, 0.00000585, -0.00000593, -0.00001738, -0.00002162, -0.00001533, -0.00000120, 0.00001311, 0.00001906, 0.00001224, -0.00000470, -0.00002309, -0.00003286, -0.00002822, -0.00001108, 0.00000984, 0.00002372, 0.00002370, 0.00001066, -0.00000743, -0.00002009, -0.00002036, -0.00000862, 0.00000792, 0.00001959, 0.00002000, 0.00000964, -0.00000472, -0.00001427, -0.00001350, -0.00000339, 0.00000933, 0.00001646, 0.00001325, 0.00000133, -0.00001239, -0.00001981, -0.00001647, -0.00000413, 0.00001036, 0.00001906, 0.00001743, 0.00000676, -0.00000679, -0.00001583, -0.00001586, -0.00000750, 0.00000428, 0.00001313, 0.00001482, 0.00000917, -0.00000039, -0.00000912, -0.00001359, -0.00001309, -0.00000936, -0.00000502, -0.00000189, -0.00000027, 0.00000065, 0.00000177, 0.00000320, 0.00000421, 0.00000394, 0.00000217, -0.00000031, -0.00000211, -0.00000214, -0.00000026, 0.00000252, 0.00000472, 0.00000514, 0.00000355, 0.00000061, -0.00000252, -0.00000492, -0.00000616, -0.00000630, -0.00000551, -0.00000382, -0.00000117, 0.00000236, 0.00000625, 0.00000963, 0.00001161, 0.00001167, 0.00000997, 0.00000709, 0.00000370, 0.00000020, -0.00000334, -0.00000681, -0.00000974, -0.00001123, },
		{ 0.00000010, 0.00000047, -0.00000106, -0.00000321, -0.00000418, -0.00000270, 0.00000105, 0.00000551, 0.00000849, 0.00000846, 0.00000532, 0.00000041, -0.00000441, -0.00000786, -0.00000979, -0.00001091, -0.00001184, -0.00001240, -0.00001155, -0.00000829, -0.00000274, 0.00000340, 0.00000745, 0.00000737, 0.00000323, -0.00000247, -0.00000605, -0.00000496, 0.00000045, 0.00000668, 0.00000908, 0.00000474, -0.00000525, -0.00001591, -0.00002095, -0.00001654, -0.00000402, 0.00001028, 0.00001824, 0.00001460, 0.00000022, -0.00001792, -0.00003028, -0.00003001, -0.00001681, 0.00000273, 0.00001871, 0.00002339, 0.00001531, -0.00000001, -0.00001352, -0.00001778, -0.00001113, 0.00000176, 0.00001299, 0.00001628, 0.00001060, 0.00000046, -0.00000722, -0.00000756, -0.00000078, 0.00000805, 0.00001240, 0.00000853, -0.00000196, -0.00001296, -0.00001752, -0.00001220, 0.00000066, 0.00001412, 0.00002058, 0.00001619, 0.00000312, -0.00001163, -0.00002017, -0.00001807, -0.00000669, 0.00000774, 0.00001767, 0.00001824, 0.00000964, -0.00000346, -0.00001465, -0.00001927, -0.00001639, -0.00000867, -0.00000040, 0.00000497, 0.00000630, 0.00000469, 0.00000221, 0.00000050, -0.00000004, 0.00000008, 0.00000018, 0.00000012, 0.00000033, 0.00000129, 0.00000294, 0.00000449, 0.00000485, 0.00000335, 0.00000028, -0.00000322, -0.00000577, -0.00000652, -0.00000552, -0.00000356, -0.00000154, 0.00000007, 0.00000141, 0.00000292, 0.00000485, 0.00000690, 0.00000838, 0.00000863, 0.00000745, 0.00000519, 0.00000252, -0.00000005, -0.00000240, -0.00000467, -0.00000687, -0.00000862, -0.00000923, -0.00000813, -0.00000546, -0.00000223, },
		{ 0.00000130, 0.00000201, 0.00000385, 0.00000592, 0.00000670, 0.00000503, 0.00000090, -0.00000439, -0.00000894, -0.00001130, -0.00001128, -0.00000984, -0.00000829, -0.00000732, -0.00000656, -0.00000497, -0.00000190, 0.00000210, 0.00000532, 0.00000595, 0.00000337, -0.00000106, -0.00000463, -0.00000491, -0.00000150, 0.00000334, 0.00000586, 0.00000316, -0.00000462, -0.00001381, -0.00001900, -0.00001622, -0.00000568, 0.00000770, 0.00001662, 0.00001541, 0.00000352, -0.00001369, -0.00002752, -0.00003056, -0.00002077, -0.00000279, 0.00001451, 0.00002283, 0.00001886, 0.00000587, -0.00000839, -0.00001613, -0.00001387, -0.00000392, 0.00000740, 0.00001381, 0.00001269, 0.00000628, -0.00000029, -0.00000258, 0.00000030, 0.00000523, 0.00000748, 0.00000411, -0.00000376, -0.00001137, -0.00001331, -0.00000717, 0.00000458, 0.00001561, 0.00001933, 0.00001283, -0.00000108, -0.00001525, -0.00002206, -0.00001767, -0.00000426, 0.00001116, 0.00002054, 0.00001909, 0.00000771, -0.00000775, -0.00001975, -0.00002301, -0.00001692, -0.00000536, 0.00000582, 0.00001192, 0.00001147, 0.00000635, 0.00000015, -0.00000398, -0.00000476, -0.00000283, 0.00000016, 0.00000277, 0.00000440, 0.00000512, 0.00000516, 0.00000445, 0.00000269, -0.00000017, -0.00000355, -0.00000631, -0.00000726, -0.00000594, -0.00000297, 0.00000030, 0.00000247, 0.00000294, 0.00000217, 0.00000131, 0.00000139, 0.00000271, 0.00000467, 0.00000622, 0.00000653, 0.00000542, 0.00000336, 0.00000101, -0.00000124, -0.00000341, -0.00000566, -0.00000788, -0.00000952, -0.00000983, -0.00000836, -0.00000542, -0.00000206, 0.00000049, 0.00000158, 0.00000150, },
		{ 0.00000284, 0.00000032, -0.00000369, -0.00000823, -0.00001171, -0.00001280, -0.00001123, -0.00000801, -0.00000477, -0.00000274, -0.00000204, -0.00000177, -0.00000079, 0.00000123, 0.00000347, 0.00000440, 0.00000299, -0.00000036, -0.00000382, -0.00000522, -0.00000356, 0.00000001, 0.00000267, 0.00000166, -0.00000369, -0.00001105, -0.00001605, -0.00001484, -0.00000667, 0.00000500, 0.00001398, 0.00001467, 0.00000541, -0.00001005, -0.00002416, -0.00002947, -0.00002272, -0.00000678, 0.00001071, 0.00002140, 0.00002064, 0.00000981, -0.00000461, -0.00001482, -0.00001591, -0.00000824, 0.00000329, 0.00001240, 0.00001513, 0.00001166, 0.00000556, 0.00000106, 0.00000020, 0.00000183, 0.00000286, 0.00000086, -0.00000385, -0.00000818, -0.00000833, -0.00000264, 0.00000674, 0.00001466, 0.00001590, 0.00000850, -0.00000459, -0.00001676, -0.00002129, -0.00001524, -0.00000133, 0.00001330, 0.00002088, 0.00001726, 0.00000411, -0.00001196, -0.00002299, -0.00002381, -0.00001454, -0.00000021, 0.00001219, 0.00001735, 0.00001409, 0.00000538, -0.00000371, -0.00000886, -0.00000841, -0.00000366, 0.00000244, 0.00000707, 0.00000884, 0.00000788, 0.00000525, 0.00000198, -0.00000131, -0.00000427, -0.00000650, -0.00000735, -0.00000628, -0.00000333, 0.00000054, 0.00000378, 0.00000496, 0.00000366, 0.00000076, -0.00000197, -0.00000289, -0.00000136, 0.00000193, 0.00000536, 0.00000736, 0.00000717, 0.00000509, 0.00000205, -0.00000104, -0.00000376, -0.00000618, -0.00000840, -0.00001016, -0.00001078, -0.00000959, -0.00000644, -0.00000207, 0.00000220, 0.00000515, 0.00000625, 0.00000593, 0.00000510, 0.00000447, 0.00000396, },
		{ -0.00001268, -0.00001255, -0.00000998, -0.00000579, -0.00000157, 0.00000118, 0.00000192, 0.00000130, 0.00000061, 0.00000082, 0.00000183, 0.00000256, 0.00000184, -0.00000063, -0.00000379, -0.00000583, -0.00000540, -0.00000279, 0.00000006, 0.00000061, -0.00000242, -0.00000792, -0.00001262, -0.00001295, -0.00000739, 0.00000206, 0.00001052, 0.00001275, 0.00000630, -0.00000667, -0.00002004, -0.00002681, -0.00002290, -0.00000960, 0.00000692, 0.00001877, 0.00002050, 0.00001196, -0.00000175, -0.00001319, -0.00001660, -0.00001079, 0.00000059, 0.00001153, 0.00001705, 0.00001575, 0.00000990, 0.00000352, -0.00000033, -0.00000112, -0.00000060, -0.00000090, -0.00000264, -0.00000438, -0.00000374, 0.00000048, 0.00000688, 0.00001177, 0.00001137, 0.00000440, -0.00000654, -0.00001595, -0.00001841, -0.00001184, 0.00000093, 0.00001334, 0.00001858, 0.00001341, 0.00000007, -0.00001474, -0.00002341, -0.00002149, -0.00000995, 0.00000541, 0.00001715, 0.00002011, 0.00001382, 0.00000237, -0.00000807, -0.00001260, -0.00000982, -0.00000204, 0.00000643, 0.00001172, 0.00001211, 0.00000839, 0.00000278, -0.00000244, -0.00000593, -0.00000743, -0.00000725, -0.00000572, -0.00000309, 0.00000026, 0.00000347, 0.00000535, 0.00000487, 0.00000197, -0.00000218, -0.00000558, -0.00000636, -0.00000386, 0.00000099, 0.00000606, 0.00000918, 0.00000917, 0.00000631, 0.00000197, -0.00000229, -0.00000552, -0.00000763, -0.00000900, -0.00000986, -0.00000992, -0.00000854, -0.00000530, -0.00000056, 0.00000447, 0.00000826, 0.00000973, 0.00000877, 0.00000620, 0.00000316, 0.00000041, -0.00000205, -0.00000465, -0.00000767, -0.00001070, },
		{ 0.00000122, 0.00000436, 0.00000522, 0.00000407, 0.00000213, 0.00000067, 0.00000024, 0.00000032, -0.00000014, -0.00000184, -0.00000440, -0.00000647, -0.00000667, -0.00000466, -0.00000167, 0.00000011, -0.00000100, -0.00000488, -0.00000928, -0.00001102, -0.00000802, -0.00000094, 0.00000668, 0.00001021, 0.00000661, -0.00000343, -0.00001537, -0.00002300, -0.00002182, -0.00001165, 0.00000296, 0.00001501, 0.00001875, 0.00001276, 0.00000067, -0.00001086, -0.00001576, -0.00001170, -0.00000110, 0.00001056, 0.00001785, 0.00001819, 0.00001277, 0.00000529, -0.00000052, -0.00000289, -0.00000252, -0.00000137, -0.00000089, -0.00000100, -0.00000043, 0.00000186, 0.00000537, 0.00000790, 0.00000691, 0.00000145, -0.00000664, -0.00001329, -0.00001445, -0.00000865, 0.00000167, 0.00001114, 0.00001428, 0.00000874, -0.00000310, -0.00001515, -0.00002083, -0.00001671, -0.00000444, 0.00001002, 0.00001959, 0.00001984, 0.00001120, -0.00000140, -0.00001140, -0.00001402, -0.00000859, 0.00000147, 0.00001085, 0.00001516, 0.00001301, 0.00000619, -0.00000174, -0.00000752, -0.00000959, -0.00000832, -0.00000521, -0.00000176, 0.00000116, 0.00000324, 0.00000432, 0.00000405, 0.00000210, -0.00000135, -0.00000525, -0.00000788, -0.00000765, -0.00000408, 0.00000171, 0.00000741, 0.00001060, 0.00001000, 0.00000601, 0.00000043, -0.00000459, -0.00000767, -0.00000869, -0.00000842, -0.00000775, -0.00000695, -0.00000553, -0.00000280, 0.00000134, 0.00000608, 0.00000984, 0.00001118, 0.00000954, 0.00000555, 0.00000063, -0.00000380, -0.00000702, -0.00000915, -0.00001065, -0.00001174, -0.00001205, -0.00001086, -0.00000779, -0.00000333, },
		{ 0.00000341, 0.00000052, -0.00000139, -0.00000223, -0.00000272, -0.00000369, -0.00000533, -0.00000687, -0.00000711, -0.00000544, -0.00000248, 0.00000002, 0.00000024, -0.00000236, -0.00000643, -0.00000928, -0.00000849, -0.00000365, 0.00000300, 0.00000753, 0.00000658, -0.00000045, -0.00001059, -0.00001867, -0.00002000, -0.00001318, -0.00000106, 0.00001056, 0.00001599, 0.00001276, 0.00000300, -0.00000777, -0.00001363, -0.00001141, -0.00000230, 0.00000908, 0.00001732, 0.00001908, 0.00001457, 0.00000695, 0.00000018, -0.00000320, -0.00000306, -0.00000114, 0.00000057, 0.00000119, 0.00000124, 0.00000173, 0.00000300, 0.00000409, 0.00000339, -0.00000003, -0.00000526, -0.00000973, -0.00001055, -0.00000653, 0.00000072, 0.00000728, 0.00000912, 0.00000453, -0.00000447, -0.00001299, -0.00001588, -0.00001071, 0.00000058, 0.00001256, 0.00001918, 0.00001707, 0.00000745, -0.00000452, -0.00001259, -0.00001274, -0.00000525, 0.00000563, 0.00001416, 0.00001616, 0.00001106, 0.00000182, -0.00000699, -0.00001163, -0.00001092, -0.00000636, -0.00000078, 0.00000330, 0.00000480, 0.00000408, 0.00000218, -0.00000007, -0.00000241, -0.00000477, -0.00000680, -0.00000760, -0.00000621, -0.00000234, 0.00000305, 0.00000796, 0.00001023, 0.00000871, 0.00000397, -0.00000198, -0.00000676, -0.00000881, -0.00000811, -0.00000592, -0.00000375, -0.00000240, -0.00000153, -0.00000013, 0.00000253, 0.00000616, 0.00000941, 0.00001055, 0.00000857, 0.00000376, -0.00000235, -0.00000786, -0.00001142, -0.00001277, -0.00001252, -0.00001149, -0.00001005, -0.00000796, -0.00000483, -0.00000072, 0.00000356, 0.00000670, 0.00000765, 0.00000625, },
		{ -0.00000545, -0.00000575, -0.00000627, -0.00000686, -0.00000675, -0.00000526, -0.00000257, 0.00000009, 0.00000106, -0.00000060, -0.00000425, -0.00000775, -0.00000863, -0.00000571, -0.00000012, 0.00000499, 0.00000622, 0.00000198, -0.00000625, -0.00001437, -0.00001783, -0.00001419, -0.00000475, 0.00000604, 0.00001284, 0.00001237, 0.00000528, -0.00000424, -0.00001073, -0.00001045, -0.00000338, 0.00000700, 0.00001566, 0.00001882, 0.00001579, 0.00000890, 0.00000191, -0.00000221, -0.00000270, -0.00000087, 0.00000114, 0.00000191, 0.00000142, 0.00000070, 0.00000062, 0.00000113, 0.00000119, -0.00000023, -0.00000318, -0.00000629, -0.00000756, -0.00000577, -0.00000151, 0.00000280, 0.00000435, 0.00000176, -0.00000376, -0.00000881, -0.00000974, -0.00000490, 0.00000401, 0.00001265, 0.00001639, 0.00001296, 0.00000396, -0.00000588, -0.00001126, -0.00000927, -0.00000102, 0.00000894, 0.00001516, 0.00001428, 0.00000678, -0.00000344, -0.00001134, -0.00001339, -0.00000933, -0.00000195, 0.00000474, 0.00000770, 0.00000624, 0.00000192, -0.00000278, -0.00000593, -0.00000700, -0.00000660, -0.00000560, -0.00000435, -0.00000256, 0.00000014, 0.00000350, 0.00000637, 0.00000726, 0.00000530, 0.00000095, -0.00000400, -0.00000736, -0.00000773, -0.00000529, -0.00000160, 0.00000139, 0.00000255, 0.00000218, 0.00000164, 0.00000231, 0.00000450, 0.00000713, 0.00000828, 0.00000644, 0.00000147, -0.00000520, -0.00001130, -0.00001491, -0.00001533, -0.00001326, -0.00001006, -0.00000688, -0.00000404, -0.00000122, 0.00000196, 0.00000532, 0.00000799, 0.00000887, 0.00000745, 0.00000415, 0.00000023, -0.00000297, -0.00000478, },
		{ -0.00000584, -0.00000445, -0.00000222, 0.00000016, 0.00000139, 0.00000042, -0.00000266, -0.00000632, -0.00000828, -0.00000693, -0.00000251, 0.00000268, 0.00000544, 0.00000353, -0.00000277, -0.00001049, -0.00001544, -0.00001453, -0.00000766, 0.00000204, 0.00000978, 0.00001178, 0.00000733, -0.00000072, -0.00000762, -0.00000926, -0.00000446, 0.00000451, 0.00001330, 0.00001792, 0.00001676, 0.00001119, 0.00000443, -0.00000041, -0.00000198, -0.00000098, 0.00000066, 0.00000133, 0.00000065, -0.00000054, -0.00000111, -0.00000065, 0.00000018, 0.00000021, -0.00000125, -0.00000374, -0.00000583, -0.00000608, -0.00000415, -0.00000117, 0.00000093, 0.00000079, -0.00000137, -0.00000367, -0.00000373, -0.00000035, 0.00000544, 0.00001073, 0.00001234, 0.00000888, 0.00000181, -0.00000511, -0.00000791, -0.00000481, 0.00000262, 0.00001020, 0.00001340, 0.00001002, 0.00000146, -0.00000795, -0.00001339, -0.00001217, -0.00000522, 0.00000358, 0.00000956, 0.00000982, 0.00000463, -0.00000306, -0.00000938, -0.00001173, -0.00000986, -0.00000559, -0.00000138, 0.00000115, 0.00000187, 0.00000172, 0.00000169, 0.00000196, 0.00000189, 0.00000070, -0.00000165, -0.00000416, -0.00000528, -0.00000397, -0.00000048, 0.00000364, 0.00000639, 0.00000658, 0.00000453, 0.00000191, 0.00000061, 0.00000145, 0.00000359, 0.00000501, 0.00000372, -0.00000090, -0.00000761, -0.00001388, -0.00001729, -0.00001672, -0.00001283, -0.00000751, -0.00000266, 0.00000071, 0.00000281, 0.00000446, 0.00000624, 0.00000791, 0.00000859, 0.00000740, 0.00000420, -0.00000013, -0.00000420, -0.00000687, -0.00000783, -0.00000762, -0.00000705, -0.00000653, },
		{ 0.00000136, 0.00000087, -0.00000152, -0.00000491, -0.00000743, -0.00000734, -0.00000420, 0.00000055, 0.00000414, 0.00000407, -0.00000032, -0.00000715, -0.00001283, -0.00001398, -0.00000944, -0.00000108, 0.00000706, 0.00001095, 0.00000886, 0.00000230, -0.00000473, -0.00000803, -0.00000545, 0.00000199, 0.00001072, 0.00001673, 0.00001759, 0.00001361, 0.00000729, 0.00000168, -0.00000130, -0.00000161, -0.00000066, -0.00000006, -0.00000049, -0.00000148, -0.00000198, -0.00000140, -0.00000015, 0.00000063, -0.00000010, -0.00000239, -0.00000515, -0.00000679, -0.00000626, -0.00000380, -0.00000074, 0.00000136, 0.00000183, 0.00000126, 0.00000110, 0.00000246, 0.00000519, 0.00000778, 0.00000832, 0.00000588, 0.00000140, -0.00000265, -0.00000371, -0.00000082, 0.00000450, 0.00000897, 0.00000937, 0.00000464, -0.00000329, -0.00001033, -0.00001252, -0.00000838, 0.00000011, 0.00000845, 0.00001205, 0.00000878, 0.00000013, -0.00000961, -0.00001573, -0.00001556, -0.00000972, -0.00000152, 0.00000501, 0.00000728, 0.00000525, 0.00000096, -0.00000298, -0.00000489, -0.00000470, -0.00000348, -0.00000230, -0.00000142, -0.00000023, 0.00000201, 0.00000527, 0.00000845, 0.00000994, 0.00000870, 0.00000507, 0.00000078, -0.00000207, -0.00000230, -0.00000051, 0.00000126, 0.00000070, -0.00000321, -0.00000948, -0.00001547, -0.00001837, -0.00001671, -0.00001110, -0.00000385, 0.00000236, 0.00000591, 0.00000681, 0.00000634, 0.00000596, 0.00000625, 0.00000667, 0.00000609, 0.00000369, -0.00000029, -0.00000468, -0.00000803, -0.00000947, -0.00000909, -0.00000769, -0.00000613, -0.00000475, -0.00000335, -0.00000163, 0.00000021, },
		{ -0.00000629, -0.00000716, -0.00000534, -0.00000146, 0.00000236, 0.00000367, 0.00000114, -0.00000434, -0.00000998, -0.00001248, -0.00000995, -0.00000315, 0.00000474, 0.00000981, 0.00000962, 0.00000455, -0.00000229, -0.00000680, -0.00000613, -0.00000021, 0.00000827, 0.00001541, 0.00001816, 0.00001579, 0.00001001, 0.00000367, -0.00000079, -0.00000255, -0.00000238, -0.00000172, -0.00000158, -0.00000195, -0.00000215, -0.00000156, -0.00000035, 0.00000054, 0.00000007, -0.00000205, -0.00000499, -0.00000716, -0.00000717, -0.00000474, -0.00000085, 0.00000280, 0.00000486, 0.00000507, 0.00000429, 0.00000372, 0.00000405, 0.00000491, 0.00000530, 0.00000440, 0.00000238, 0.00000044, -0.00000002, 0.00000156, 0.00000416, 0.00000570, 0.00000430, -0.00000027, -0.00000610, -0.00000994, -0.00000911, -0.00000328, 0.00000491, 0.00001105, 0.00001131, 0.00000474, -0.00000601, -0.00001586, -0.00001992, -0.00001612, -0.00000632, 0.00000472, 0.00001184, 0.00001208, 0.00000605, -0.00000277, -0.00001004, -0.00001281, -0.00001072, -0.00000558, 0.00000003, 0.00000437, 0.00000711, 0.00000893, 0.00001045, 0.00001151, 0.00001117, 0.00000861, 0.00000402, -0.00000119, -0.00000500, -0.00000607, -0.00000469, -0.00000273, -0.00000255, -0.00000546, -0.00001074, -0.00001588, -0.00001788, -0.00001505, -0.00000797, 0.00000072, 0.00000776, 0.00001095, 0.00001016, 0.00000715, 0.00000425, 0.00000300, 0.00000332, 0.00000382, 0.00000296, 0.00000009, -0.00000409, -0.00000802, -0.00001023, -0.00001018, -0.00000840, -0.00000598, -0.00000383, -0.00000219, -0.00000084, 0.00000040, 0.00000123, 0.00000101, -0.00000072, -0.00000359, },
		{ 0.00000033, 0.00000258, 0.00000181, -0.00000200, -0.00000694, -0.00001013, -0.00000928, -0.00000424, 0.00000280, 0.00000834, 0.00000955, 0.00000593, -0.00000035, -0.00000553, -0.00000637, -0.00000190, 0.00000608, 0.00001395, 0.00001825, 0.00001740, 0.00001227, 0.00000544, -0.00000033, -0.00000346, -0.00000404, -0.00000328, -0.00000246, -0.00000212, -0.00000201, -0.00000160, -0.00000080, -0.00000016, -0.00000052, -0.00000223, -0.00000474, -0.00000668, -0.00000666, -0.00000417, 0.00000005, 0.00000433, 0.00000703, 0.00000740, 0.00000598, 0.00000408, 0.00000294, 0.00000297, 0.00000367, 0.00000414, 0.00000383, 0.00000292, 0.00000210, 0.00000186, 0.00000197, 0.00000153, -0.00000029, -0.00000338, -0.00000635, -0.00000714, -0.00000438, 0.00000142, 0.00000759, 0.00001051, 0.00000754, -0.00000107, -0.00001193, -0.00001987, -0.00002061, -0.00001316, -0.00000057, 0.00001137, 0.00001707, 0.00001392, 0.00000357, -0.00000898, -0.00001802, -0.00001985, -0.00001435, -0.00000453, 0.00000544, 0.00001235, 0.00001525, 0.00001514, 0.00001372, 0.00001203, 0.00000995, 0.00000674, 0.00000209, -0.00000321, -0.00000754, -0.00000947, -0.00000881, -0.00000696, -0.00000605, -0.00000756, -0.00001119, -0.00001477, -0.00001549, -0.00001157, -0.00000356, 0.00000579, 0.00001285, 0.00001505, 0.00001221, 0.00000651, 0.00000115, -0.00000148, -0.00000095, 0.00000114, 0.00000243, 0.00000124, -0.00000238, -0.00000689, -0.00001023, -0.00001105, -0.00000936, -0.00000628, -0.00000322, -0.00000102, 0.00000023, 0.00000091, 0.00000124, 0.00000099, -0.00000026, -0.00000255, -0.00000512, -0.00000666, -0.00000605, -0.00000326, },
		{ -0.00000391, -0.00000721, -0.00000769, -0.00000446, 0.00000126, 0.00000662, 0.00000877, 0.00000651, 0.00000112, -0.00000422, -0.00000616, -0.00000305, 0.00000413, 0.00001227, 0.00001774, 0.00001828, 0.00001398, 0.00000701, 0.00000029, -0.00000401, -0.00000534, -0.00000460, -0.00000324, -0.00000230, -0.00000197, -0.00000185, -0.00000158, -0.00000127, -0.00000142, -0.00000241, -0.00000399, -0.00000523, -0.00000496, -0.00000262, 0.00000131, 0.00000541, 0.00000808, 0.00000842, 0.00000673, 0.00000428, 0.00000251, 0.00000222, 0.00000315, 0.00000431, 0.00000469, 0.00000388, 0.00000224, 0.00000044, -0.00000102, -0.00000214, -0.00000315, -0.00000406, -0.00000431, -0.00000308, 0.00000003, 0.00000419, 0.00000733, 0.00000701, 0.00000193, -0.00000682, -0.00001576, -0.00002039, -0.00001761, -0.00000759, 0.00000575, 0.00001638, 0.00001907, 0.00001210, -0.00000169, -0.00001618, -0.00002489, -0.00002411, -0.00001445, -0.00000021, 0.00001296, 0.00002086, 0.00002235, 0.00001918, 0.00001433, 0.00001006, 0.00000692, 0.00000398, 0.00000016, -0.00000470, -0.00000945, -0.00001250, -0.00001298, -0.00001148, -0.00000969, -0.00000925, -0.00001043, -0.00001179, -0.00001099, -0.00000636, 0.00000171, 0.00001061, 0.00001674, 0.00001743, 0.00001250, 0.00000442, -0.00000296, -0.00000650, -0.00000538, -0.00000137, 0.00000236, 0.00000313, 0.00000021, -0.00000496, -0.00000973, -0.00001185, -0.00001062, -0.00000701, -0.00000287, 0.00000021, 0.00000167, 0.00000187, 0.00000146, 0.00000080, -0.00000028, -0.00000201, -0.00000421, -0.00000601, -0.00000634, -0.00000467, -0.00000163, 0.00000114, 0.00000188, -0.00000012, },
		{ 0.00000023, 0.00000490, 0.00000750, 0.00000641, 0.00000210, -0.00000298, -0.00000564, -0.00000378, 0.00000239, 0.00001037, 0.00001664, 0.00001846, 0.00001521, 0.00000850, 0.00000121, -0.00000408, -0.00000623, -0.00000576, -0.00000414, -0.00000278, -0.00000227, -0.00000236, -0.00000250, -0.00000238, -0.00000216, -0.00000223, -0.00000268, -0.00000306, -0.00000257, -0.00000066, 0.00000250, 0.00000592, 0.00000827, 0.00000865, 0.00000715, 0.00000478, 0.00000288, 0.00000232, 0.00000300, 0.00000400, 0.00000421, 0.00000304, 0.00000076, -0.00000175, -0.00000354, -0.00000416, -0.00000371, -0.00000259, -0.00000109, 0.00000071, 0.00000267, 0.00000423, 0.00000429, 0.00000173, -0.00000372, -0.00001068, -0.00001626, -0.00001723, -0.00001183, -0.00000117, 0.00001064, 0.00001814, 0.00001714, 0.00000710, -0.00000825, -0.00002232, -0.00002875, -0.00002440, -0.00001088, 0.00000639, 0.00002087, 0.00002798, 0.00002689, 0.00002029, 0.00001237, 0.00000639, 0.00000317, 0.00000135, -0.00000118, -0.00000550, -0.00001091, -0.00001545, -0.00001734, -0.00001615, -0.00001307, -0.00000997, -0.00000798, -0.00000669, -0.00000451, 0.00000007, 0.00000701, 0.00001426, 0.00001863, 0.00001758, 0.00001094, 0.00000126, -0.00000731, -0.00001119, -0.00000923, -0.00000332, 0.00000276, 0.00000541, 0.00000317, -0.00000267, -0.00000898, -0.00001259, -0.00001201, -0.00000794, -0.00000262, 0.00000160, 0.00000347, 0.00000318, 0.00000180, 0.00000034, -0.00000088, -0.00000212, -0.00000366, -0.00000525, -0.00000611, -0.00000544, -0.00000320, -0.00000037, 0.00000147, 0.00000117, -0.00000118, -0.00000407, -0.00000543, -0.00000389, },
		{ 0.00000249, -0.00000202, -0.00000503, -0.00000422, 0.00000087, 0.00000840, 0.00001516, 0.00001813, 0.00001606, 0.00000994, 0.00000240, -0.00000373, -0.00000682, -0.00000690, -0.00000531, -0.00000365, -0.00000287, -0.00000295, -0.00000324, -0.00000313, -0.00000249, -0.00000166, -0.00000102, -0.00000060, 0.00000001, 0.00000134, 0.00000354, 0.00000609, 0.00000807, 0.00000864, 0.00000761, 0.00000560, 0.00000364, 0.00000254, 0.00000241, 0.00000261, 0.00000224, 0.00000081, -0.00000141, -0.00000354, -0.00000458, -0.00000405, -0.00000224, 0.00000004, 0.00000191, 0.00000287, 0.00000278, 0.00000170, -0.00000040, -0.00000361, -0.00000765, -0.00001146, -0.00001328, -0.00001135, -0.00000506, 0.00000406, 0.00001248, 0.00001600, 0.00001183, 0.00000044, -0.00001409, -0.00002555, -0.00002843, -0.00002065, -0.00000466, 0.00001353, 0.00002723, 0.00003213, 0.00002806, 0.00001859, 0.00000874, 0.00000224, -0.00000017, -0.00000051, -0.00000181, -0.00000595, -0.00001238, -0.00001860, -0.00002178, -0.00002049, -0.00001550, -0.00000910, -0.00000353, 0.00000035, 0.00000331, 0.00000678, 0.00001134, 0.00001589, 0.00001801, 0.00001547, 0.00000797, -0.00000221, -0.00001103, -0.00001476, -0.00001202, -0.00000462, 0.00000330, 0.00000753, 0.00000594, -0.00000041, -0.00000805, -0.00001306, -0.00001315, -0.00000872, -0.00000228, 0.00000308, 0.00000539, 0.00000458, 0.00000203, -0.00000051, -0.00000210, -0.00000283, -0.00000341, -0.00000429, -0.00000522, -0.00000540, -0.00000419, -0.00000183, 0.00000058, 0.00000173, 0.00000098, -0.00000102, -0.00000271, -0.00000255, -0.00000013, 0.00000341, 0.00000594, 0.00000571, },
		{ -0.00000030, 0.00000661, 0.00001354, 0.00001745, 0.00001654, 0.00001120, 0.00000366, -0.00000318, -0.00000725, -0.00000806, -0.00000668, -0.00000477, -0.00000359, -0.00000339, -0.00000359, -0.00000339, -0.00000240, -0.00000087, 0.00000066, 0.00000176, 0.00000250, 0.00000333, 0.00000463, 0.00000636, 0.00000795, 0.00000867, 0.00000806, 0.00000630, 0.00000407, 0.00000214, 0.00000087, 0.00000008, -0.00000071, -0.00000185, -0.00000318, -0.00000404, -0.00000376, -0.00000215, 0.00000028, 0.00000249, 0.00000346, 0.00000273, 0.00000055, -0.00000232, -0.00000511, -0.00000733, -0.00000872, -0.00000903, -0.00000778, -0.00000450, 0.00000069, 0.00000654, 0.00001069, 0.00001053, 0.00000467, -0.00000584, -0.00001733, -0.00002474, -0.00002393, -0.00001389, 0.00000245, 0.00001927, 0.00003049, 0.00003257, 0.00002607, 0.00001503, 0.00000473, -0.00000119, -0.00000236, -0.00000145, -0.00000213, -0.00000663, -0.00001430, -0.00002196, -0.00002582, -0.00002370, -0.00001619, -0.00000620, 0.00000280, 0.00000867, 0.00001142, 0.00001260, 0.00001376, 0.00001501, 0.00001492, 0.00001160, 0.00000439, -0.00000509, -0.00001334, -0.00001674, -0.00001364, -0.00000555, 0.00000347, 0.00000888, 0.00000805, 0.00000165, -0.00000679, -0.00001285, -0.00001362, -0.00000907, -0.00000185, 0.00000440, 0.00000704, 0.00000567, 0.00000195, -0.00000174, -0.00000373, -0.00000387, -0.00000322, -0.00000301, -0.00000366, -0.00000456, -0.00000463, -0.00000325, -0.00000082, 0.00000146, 0.00000243, 0.00000178, 0.00000036, -0.00000045, 0.00000029, 0.00000231, 0.00000421, 0.00000438, 0.00000215, -0.00000155, -0.00000449, -0.00000438, },
		{ 0.00001654, 0.00001204, 0.00000472, -0.00000262, -0.00000755, -0.00000913, -0.00000804, -0.00000589, -0.00000420, -0.00000354, -0.00000352, -0.00000324, -0.00000208, -0.00000009, 0.00000212, 0.00000388, 0.00000491, 0.00000548, 0.00000610, 0.00000704, 0.00000808, 0.00000862, 0.00000806, 0.00000622, 0.00000351, 0.00000068, -0.00000159, -0.00000302, -0.00000372, -0.00000392, -0.00000372, -0.00000295, -0.00000146, 0.00000061, 0.00000263, 0.00000367, 0.00000299, 0.00000051, -0.00000300, -0.00000632, -0.00000828, -0.00000838, -0.00000683, -0.00000430, -0.00000144, 0.00000136, 0.00000388, 0.00000565, 0.00000580, 0.00000330, -0.00000226, -0.00000987, -0.00001690, -0.00001994, -0.00001637, -0.00000596, 0.00000849, 0.00002210, 0.00002998, 0.00002957, 0.00002189, 0.00001095, 0.00000156, -0.00000315, -0.00000331, -0.00000192, -0.00000280, -0.00000807, -0.00001675, -0.00002508, -0.00002865, -0.00002489, -0.00001454, -0.00000125, 0.00001035, 0.00001708, 0.00001847, 0.00001645, 0.00001370, 0.00001170, 0.00001001, 0.00000692, 0.00000117, -0.00000665, -0.00001386, -0.00001712, -0.00001441, -0.00000659, 0.00000277, 0.00000911, 0.00000937, 0.00000364, -0.00000487, -0.00001158, -0.00001309, -0.00000890, -0.00000150, 0.00000516, 0.00000796, 0.00000615, 0.00000148, -0.00000312, -0.00000538, -0.00000485, -0.00000286, -0.00000137, -0.00000157, -0.00000315, -0.00000468, -0.00000470, -0.00000271, 0.00000043, 0.00000318, 0.00000432, 0.00000373, 0.00000235, 0.00000145, 0.00000160, 0.00000232, 0.00000242, 0.00000101, -0.00000164, -0.00000403, -0.00000418, -0.00000092, 0.00000522, 0.00001198, 0.00001645, },
		{ -0.00000760, -0.00000983, -0.00000906, -0.00000676, -0.00000461, -0.00000350, -0.00000324, -0.00000293, -0.00000173, 0.00000056, 0.00000338, 0.00000588, 0.00000743, 0.00000801, 0.00000808, 0.00000812, 0.00000824, 0.00000807, 0.00000704, 0.00000482, 0.00000166, -0.00000172, -0.00000442, -0.00000583, -0.00000582, -0.00000466, -0.00000280, -0.00000067, 0.00000142, 0.00000307, 0.00000380, 0.00000308, 0.00000074, -0.00000280, -0.00000646, -0.00000886, -0.00000898, -0.00000669, -0.00000283, 0.00000113, 0.00000388, 0.00000474, 0.00000389, 0.00000191, -0.00000067, -0.00000369, -0.00000712, -0.00001055, -0.00001283, -0.00001229, -0.00000761, 0.00000114, 0.00001196, 0.00002140, 0.00002605, 0.00002422, 0.00001696, 0.00000758, -0.00000005, -0.00000359, -0.00000350, -0.00000262, -0.00000437, -0.00001041, -0.00001933, -0.00002714, -0.00002931, -0.00002337, -0.00001045, 0.00000513, 0.00001795, 0.00002418, 0.00002324, 0.00001765, 0.00001121, 0.00000667, 0.00000436, 0.00000255, -0.00000088, -0.00000652, -0.00001272, -0.00001635, -0.00001488, -0.00000821, 0.00000092, 0.00000819, 0.00001007, 0.00000584, -0.00000199, -0.00000903, -0.00001154, -0.00000838, -0.00000160, 0.00000494, 0.00000781, 0.00000588, 0.00000080, -0.00000426, -0.00000657, -0.00000540, -0.00000221, 0.00000045, 0.00000070, -0.00000154, -0.00000459, -0.00000617, -0.00000489, -0.00000109, 0.00000344, 0.00000661, 0.00000724, 0.00000563, 0.00000314, 0.00000116, 0.00000023, -0.00000014, -0.00000082, -0.00000218, -0.00000354, -0.00000348, -0.00000084, 0.00000434, 0.00001048, 0.00001503, 0.00001582, 0.00001219, 0.00000537, -0.00000208, },
		{ -0.00000494, -0.00000349, -0.00000301, -0.00000264, -0.00000137, 0.00000123, 0.00000470, 0.00000802, 0.00001023, 0.00001093, 0.00001038, 0.00000922, 0.00000792, 0.00000649, 0.00000461, 0.00000200, -0.00000124, -0.00000447, -0.00000678, -0.00000748, -0.00000637, -0.00000388, -0.00000082, 0.00000196, 0.00000380, 0.00000432, 0.00000337, 0.00000101, -0.00000235, -0.00000591, -0.00000854, -0.00000916, -0.00000721, -0.00000311, 0.00000175, 0.00000555, 0.00000680, 0.00000509, 0.00000121, -0.00000326, -0.00000680, -0.00000863, -0.00000881, -0.00000790, -0.00000624, -0.00000367, 0.00000032, 0.00000589, 0.00001223, 0.00001757, 0.00001988, 0.00001803, 0.00001257, 0.00000565, -0.00000010, -0.00000307, -0.00000369, -0.00000415, -0.00000699, -0.00001322, -0.00002121, -0.00002720, -0.00002712, -0.00001903, -0.00000449, 0.00001180, 0.00002416, 0.00002866, 0.00002496, 0.00001618, 0.00000701, 0.00000109, -0.00000078, -0.00000058, -0.00000136, -0.00000492, -0.00001050, -0.00001513, -0.00001560, -0.00001063, -0.00000195, 0.00000645, 0.00001055, 0.00000856, 0.00000197, -0.00000531, -0.00000923, -0.00000790, -0.00000251, 0.00000348, 0.00000651, 0.00000503, 0.00000027, -0.00000470, -0.00000690, -0.00000530, -0.00000134, 0.00000213, 0.00000268, -0.00000012, -0.00000449, -0.00000752, -0.00000701, -0.00000272, 0.00000344, 0.00000863, 0.00001062, 0.00000897, 0.00000500, 0.00000081, -0.00000200, -0.00000305, -0.00000304, -0.00000288, -0.00000282, -0.00000222, -0.00000012, 0.00000386, 0.00000888, 0.00001301, 0.00001423, 0.00001155, 0.00000563, -0.00000142, -0.00000714, -0.00000988, -0.00000953, -0.00000733, },
		{ -0.00000079, 0.00000226, 0.00000641, 0.00001051, 0.00001326, 0.00001388, 0.00001245, 0.00000973, 0.00000660, 0.00000361, 0.00000084, -0.00000187, -0.00000452, -0.00000675, -0.00000792, -0.00000746, -0.00000532, -0.00000203, 0.00000141, 0.00000397, 0.00000491, 0.00000405, 0.00000165, -0.00000166, -0.00000506, -0.00000765, -0.00000857, -0.00000727, -0.00000380, 0.00000093, 0.00000536, 0.00000772, 0.00000687, 0.00000288, -0.00000283, -0.00000808, -0.00001092, -0.00001048, -0.00000731, -0.00000292, 0.00000111, 0.00000399, 0.00000590, 0.00000762, 0.00000968, 0.00001181, 0.00001305, 0.00001237, 0.00000946, 0.00000512, 0.00000077, -0.00000246, -0.00000456, -0.00000667, -0.00001022, -0.00001562, -0.00002140, -0.00002455, -0.00002197, -0.00001244, 0.00000218, 0.00001733, 0.00002769, 0.00002975, 0.00002356, 0.00001268, 0.00000224, -0.00000374, -0.00000438, -0.00000200, -0.00000045, -0.00000250, -0.00000805, -0.00001416, -0.00001685, -0.00001367, -0.00000533, 0.00000451, 0.00001126, 0.00001192, 0.00000678, -0.00000086, -0.00000669, -0.00000784, -0.00000443, 0.00000081, 0.00000432, 0.00000398, 0.00000029, -0.00000409, -0.00000619, -0.00000464, -0.00000055, 0.00000322, 0.00000396, 0.00000087, -0.00000435, -0.00000847, -0.00000862, -0.00000408, 0.00000332, 0.00001019, 0.00001344, 0.00001188, 0.00000666, 0.00000046, -0.00000413, -0.00000587, -0.00000518, -0.00000343, -0.00000180, -0.00000055, 0.00000097, 0.00000347, 0.00000694, 0.00001026, 0.00001178, 0.00001025, 0.00000570, -0.00000041, -0.00000594, -0.00000912, -0.00000945, -0.00000772, -0.00000544, -0.00000379, -0.00000300, -0.00000235, },
		{ 0.00001608, 0.00001620, 0.00001355, 0.00000905, 0.00000402, -0.00000042, -0.00000373, -0.00000596, -0.00000732, -0.00000788, -0.00000747, -0.00000588, -0.00000316, 0.00000013, 0.00000307, 0.00000473, 0.00000455, 0.00000259, -0.00000052, -0.00000385, -0.00000642, -0.00000751, -0.00000670, -0.00000405, -0.00000011, 0.00000402, 0.00000692, 0.00000731, 0.00000460, -0.00000061, -0.00000660, -0.00001105, -0.00001213, -0.00000928, -0.00000359, 0.00000273, 0.00000741, 0.00000923, 0.00000849, 0.00000667, 0.00000544, 0.00000566, 0.00000694, 0.00000803, 0.00000767, 0.00000537, 0.00000164, -0.00000252, -0.00000635, -0.00000976, -0.00001313, -0.00001656, -0.00001915, -0.00001907, -0.00001445, -0.00000479, 0.00000812, 0.00002043, 0.00002776, 0.00002737, 0.00001970, 0.00000832, -0.00000178, -0.00000677, -0.00000590, -0.00000181, 0.00000121, -0.00000016, -0.00000619, -0.00001384, -0.00001852, -0.00001678, -0.00000848, 0.00000297, 0.00001241, 0.00001568, 0.00001187, 0.00000363, -0.00000445, -0.00000846, -0.00000723, -0.00000266, 0.00000175, 0.00000319, 0.00000118, -0.00000236, -0.00000459, -0.00000370, -0.00000022, 0.00000337, 0.00000431, 0.00000141, -0.00000397, -0.00000863, -0.00000936, -0.00000494, 0.00000304, 0.00001099, 0.00001522, 0.00001389, 0.00000790, 0.00000023, -0.00000576, -0.00000806, -0.00000677, -0.00000361, -0.00000060, 0.00000117, 0.00000194, 0.00000278, 0.00000450, 0.00000689, 0.00000875, 0.00000865, 0.00000592, 0.00000117, -0.00000396, -0.00000769, -0.00000904, -0.00000823, -0.00000634, -0.00000451, -0.00000315, -0.00000181, 0.00000040, 0.00000402, 0.00000871, 0.00001324, },
		{ 0.00000042, -0.00000493, -0.00000819, -0.00000935, -0.00000893, -0.00000757, -0.00000564, -0.00000331, -0.00000070, 0.00000181, 0.00000364, 0.00000418, 0.00000313, 0.00000075, -0.00000219, -0.00000474, -0.00000605, -0.00000572, -0.00000378, -0.00000075, 0.00000258, 0.00000522, 0.00000621, 0.00000489, 0.00000123, -0.00000387, -0.00000874, -0.00001143, -0.00001053, -0.00000596, 0.00000084, 0.00000740, 0.00001135, 0.00001148, 0.00000839, 0.00000409, 0.00000096, 0.00000044, 0.00000235, 0.00000510, 0.00000661, 0.00000548, 0.00000170, -0.00000355, -0.00000867, -0.00001249, -0.00001462, -0.00001520, -0.00001426, -0.00001135, -0.00000582, 0.00000239, 0.00001199, 0.00002033, 0.00002438, 0.00002227, 0.00001457, 0.00000439, -0.00000404, -0.00000747, -0.00000545, -0.00000063, 0.00000274, 0.00000128, -0.00000538, -0.00001416, -0.00002013, -0.00001922, -0.00001071, 0.00000220, 0.00001391, 0.00001931, 0.00001645, 0.00000743, -0.00000291, -0.00000968, -0.00001045, -0.00000627, -0.00000058, 0.00000303, 0.00000294, 0.00000021, -0.00000251, -0.00000291, -0.00000065, 0.00000248, 0.00000380, 0.00000172, -0.00000306, -0.00000773, -0.00000905, -0.00000532, 0.00000238, 0.00001065, 0.00001555, 0.00001474, 0.00000871, 0.00000042, -0.00000638, -0.00000910, -0.00000750, -0.00000345, 0.00000043, 0.00000237, 0.00000229, 0.00000152, 0.00000162, 0.00000324, 0.00000564, 0.00000721, 0.00000653, 0.00000330, -0.00000145, -0.00000594, -0.00000866, -0.00000907, -0.00000768, -0.00000545, -0.00000308, -0.00000058, 0.00000255, 0.00000667, 0.00001145, 0.00001572, 0.00001798, 0.00001712, 0.00001309, 0.00000695, },
		{ -0.00000913, -0.00000608, -0.00000309, -0.00000059, 0.00000131, 0.00000262, 0.00000318, 0.00000281, 0.00000145, -0.00000059, -0.00000270, -0.00000412, -0.00000430, -0.00000314, -0.00000098, 0.00000149, 0.00000352, 0.00000446, 0.00000386, 0.00000165, -0.00000183, -0.00000570, -0.00000868, -0.00000941, -0.00000710, -0.00000198, 0.00000446, 0.00000999, 0.00001252, 0.00001112, 0.00000662, 0.00000120, -0.00000258, -0.00000314, -0.00000065, 0.00000304, 0.00000542, 0.00000469, 0.00000065, -0.00000523, -0.00001067, -0.00001377, -0.00001382, -0.00001130, -0.00000728, -0.00000263, 0.00000236, 0.00000768, 0.00001294, 0.00001701, 0.00001830, 0.00001569, 0.00000952, 0.00000192, -0.00000405, -0.00000600, -0.00000366, 0.00000070, 0.00000338, 0.00000137, -0.00000563, -0.00001471, -0.00002100, -0.00002030, -0.00001157, 0.00000222, 0.00001532, 0.00002206, 0.00001977, 0.00001006, -0.00000211, -0.00001108, -0.00001338, -0.00000929, -0.00000221, 0.00000355, 0.00000526, 0.00000308, -0.00000052, -0.00000263, -0.00000190, 0.00000073, 0.00000278, 0.00000214, -0.00000139, -0.00000574, -0.00000782, -0.00000546, 0.00000104, 0.00000891, 0.00001431, 0.00001451, 0.00000937, 0.00000143, -0.00000557, -0.00000871, -0.00000736, -0.00000326, 0.00000076, 0.00000251, 0.00000167, -0.00000032, -0.00000134, -0.00000010, 0.00000305, 0.00000630, 0.00000756, 0.00000566, 0.00000110, -0.00000436, -0.00000859, -0.00001022, -0.00000913, -0.00000612, -0.00000229, 0.00000170, 0.00000575, 0.00000995, 0.00001405, 0.00001720, 0.00001817, 0.00001604, 0.00001083, 0.00000369, -0.00000352, -0.00000896, -0.00001156, -0.00001133, },
		{ 0.00000252, 0.00000264, 0.00000222, 0.00000143, 0.00000033, -0.00000092, -0.00000199, -0.00000246, -0.00000207, -0.00000087, 0.00000071, 0.00000210, 0.00000275, 0.00000239, 0.00000101, -0.00000111, -0.00000353, -0.00000563, -0.00000668, -0.00000598, -0.00000321, 0.00000128, 0.00000632, 0.00001021, 0.00001137, 0.00000918, 0.00000441, -0.00000095, -0.00000460, -0.00000509, -0.00000261, 0.00000106, 0.00000343, 0.00000272, -0.00000123, -0.00000677, -0.00001133, -0.00001277, -0.00001043, -0.00000531, 0.00000061, 0.00000551, 0.00000861, 0.00001016, 0.00001091, 0.00001125, 0.00001086, 0.00000907, 0.00000565, 0.00000140, -0.00000203, -0.00000308, -0.00000143, 0.00000140, 0.00000270, 0.00000015, -0.00000652, -0.00001484, -0.00002052, -0.00001965, -0.00001105, 0.00000266, 0.00001601, 0.00002326, 0.00002133, 0.00001136, -0.00000177, -0.00001204, -0.00001528, -0.00001115, -0.00000294, 0.00000452, 0.00000757, 0.00000561, 0.00000093, -0.00000298, -0.00000373, -0.00000142, 0.00000171, 0.00000295, 0.00000103, -0.00000291, -0.00000607, -0.00000575, -0.00000120, 0.00000577, 0.00001171, 0.00001354, 0.00001028, 0.00000360, -0.00000316, -0.00000696, -0.00000667, -0.00000351, -0.00000011, 0.00000126, 0.00000003, -0.00000244, -0.00000384, -0.00000252, 0.00000140, 0.00000601, 0.00000871, 0.00000772, 0.00000309, -0.00000332, -0.00000883, -0.00001132, -0.00001008, -0.00000591, -0.00000038, 0.00000506, 0.00000962, 0.00001315, 0.00001567, 0.00001688, 0.00001613, 0.00001286, 0.00000714, -0.00000000, -0.00000684, -0.00001159, -0.00001319, -0.00001170, -0.00000817, -0.00000408, -0.00000065, 0.00000154, },
		{ 0.00000017, -0.00000030, -0.00000061, -0.00000067, -0.00000036, 0.00000025, 0.00000091, 0.00000125, 0.00000099, 0.00000012, -0.00000116, -0.00000249, -0.00000353, -0.00000402, -0.00000372, -0.00000246, -0.00000014, 0.00000299, 0.00000626, 0.00000860, 0.00000895, 0.00000683, 0.00000273, -0.00000189, -0.00000521, -0.00000600, -0.00000429, -0.00000146, 0.00000046, -0.00000008, -0.00000314, -0.00000720, -0.00000987, -0.00000922, -0.00000489, 0.00000167, 0.00000794, 0.00001167, 0.00001198, 0.00000969, 0.00000659, 0.00000434, 0.00000353, 0.00000360, 0.00000351, 0.00000263, 0.00000126, 0.00000031, 0.00000039, 0.00000105, 0.00000083, -0.00000185, -0.00000736, -0.00001399, -0.00001842, -0.00001737, -0.00000958, 0.00000293, 0.00001541, 0.00002253, 0.00002104, 0.00001162, -0.00000134, -0.00001192, -0.00001563, -0.00001163, -0.00000293, 0.00000542, 0.00000921, 0.00000723, 0.00000159, -0.00000379, -0.00000567, -0.00000339, 0.00000100, 0.00000421, 0.00000390, 0.00000022, -0.00000433, -0.00000649, -0.00000431, 0.00000157, 0.00000828, 0.00001236, 0.00001178, 0.00000700, 0.00000061, -0.00000431, -0.00000596, -0.00000467, -0.00000242, -0.00000135, -0.00000230, -0.00000430, -0.00000529, -0.00000360, 0.00000081, 0.00000607, 0.00000946, 0.00000886, 0.00000412, -0.00000287, -0.00000903, -0.00001170, -0.00000983, -0.00000431, 0.00000272, 0.00000906, 0.00001333, 0.00001531, 0.00001545, 0.00001426, 0.00001185, 0.00000804, 0.00000284, -0.00000314, -0.00000864, -0.00001222, -0.00001292, -0.00001076, -0.00000673, -0.00000236, 0.00000104, 0.00000281, 0.00000308, 0.00000245, 0.00000155, 0.00000077, },
		{ 0.00000010, -0.00000000, -0.00000007, -0.00000026, -0.00000068, -0.00000131, -0.00000195, -0.00000233, -0.00000229, -0.00000182, -0.00000102, 0.00000005, 0.00000141, 0.00000307, 0.00000484, 0.00000618, 0.00000640, 0.00000497, 0.00000195, -0.00000182, -0.00000508, -0.00000667, -0.00000628, -0.00000463, -0.00000310, -0.00000287, -0.00000413, -0.00000579, -0.00000608, -0.00000358, 0.00000173, 0.00000822, 0.00001335, 0.00001491, 0.00001231, 0.00000687, 0.00000118, -0.00000229, -0.00000247, -0.00000007, 0.00000300, 0.00000489, 0.00000480, 0.00000324, 0.00000132, -0.00000023, -0.00000166, -0.00000388, -0.00000751, -0.00001187, -0.00001487, -0.00001397, -0.00000780, 0.00000247, 0.00001324, 0.00001990, 0.00001927, 0.00001136, -0.00000028, -0.00001032, -0.00001429, -0.00001089, -0.00000263, 0.00000570, 0.00000968, 0.00000768, 0.00000155, -0.00000465, -0.00000715, -0.00000469, 0.00000085, 0.00000572, 0.00000669, 0.00000301, -0.00000308, -0.00000779, -0.00000797, -0.00000303, 0.00000474, 0.00001149, 0.00001399, 0.00001132, 0.00000515, -0.00000143, -0.00000576, -0.00000696, -0.00000605, -0.00000489, -0.00000472, -0.00000533, -0.00000532, -0.00000328, 0.00000099, 0.00000598, 0.00000922, 0.00000869, 0.00000412, -0.00000265, -0.00000849, -0.00001059, -0.00000782, -0.00000125, 0.00000654, 0.00001279, 0.00001581, 0.00001547, 0.00001289, 0.00000943, 0.00000595, 0.00000257, -0.00000100, -0.00000484, -0.00000845, -0.00001085, -0.00001118, -0.00000919, -0.00000553, -0.00000146, 0.00000176, 0.00000340, 0.00000353, 0.00000275, 0.00000178, 0.00000111, 0.00000079, 0.00000065, 0.00000050, 0.00000029, },
		},
		{ { -0.00000148, -0.00000279, -0.00000482, -0.00000368, 0.00000263, 0.00001189, 0.00001879, 0.00001877, 0.00001129, 0.00000006, -0.00000973, -0.00001464, -0.00001422, -0.00001007, -0.00000409, 0.00000211, 0.00000694, 0.00000849, 0.00000547, -0.00000138, -0.00000898, -0.00001360, -0.00001345, -0.00000996, -0.00000648, -0.00000534, -0.00000583, -0.00000506, -0.00000086, 0.00000593, 0.00001179, 0.00001329, 0.00000984, 0.00000396, -0.00000097, -0.00000335, -0.00000404, -0.00000477, -0.00000610, -0.00000698, -0.00000614, -0.00000376, -0.00000143, -0.00000040, -0.00000009, 0.00000148, 0.00000534, 0.00000971, 0.00001067, 0.00000535, -0.00000500, -0.00001500, -0.00001858, -0.00001338, -0.00000257, 0.00000747, 0.00001156, 0.00000892, 0.00000288, -0.00000215, -0.00000374, -0.00000229, 0.00000027, 0.00000228, 0.00000299, 0.00000228, 0.00000027, -0.00000229, -0.00000374, -0.00000215, 0.00000288, 0.00000892, 0.00001156, 0.00000747, -0.00000257, -0.00001338, -0.00001858, -0.00001500, -0.00000500, 0.00000535, 0.00001067, 0.00000971, 0.00000534, 0.00000148, -0.00000009, -0.00000040, -0.00000143, -0.00000376, -0.00000614, -0.00000698, -0.00000610, -0.00000477, -0.00000404, -0.00000335, -0.00000097, 0.00000396, 0.00000984, 0.00001329, 0.00001179, 0.00000593, -0.00000086, -0.00000506, -0.00000583, -0.00000534, -0.00000648, -0.00000996, -0.00001345, -0.00001360, -0.00000898, -0.00000138, 0.00000547, 0.00000849, 0.00000694, 0.00000211, -0.00000409, -0.00001007, -0.00001422, -0.00001464, -0.00000973, 0.00000006, 0.00001129, 0.00001877, 0.00001879, 0.00001189, 0.00000263, -0.00000368, -0.00000482, -0.00000279, },
		{ -0.00000281, 0.00000197, 0.00001022, 0.00001717, 0.00001791, 0.00001107, -0.00000020, -0.00001052, -0.00001565, -0.00001458, -0.00000907, -0.00000198, 0.00000412, 0.00000735, 0.00000656, 0.00000177, -0.00000528, -0.00001144, -0.00001380, -0.00001175, -0.00000751, -0.00000445, -0.00000423, -0.00000539, -0.00000461, 0.00000017, 0.00000753, 0.00001330, 0.00001378, 0.00000860, 0.00000099, -0.00000489, -0.00000691, -0.00000607, -0.00000483, -0.00000468, -0.00000516, -0.00000488, -0.00000326, -0.00000102, 0.00000089, 0.00000251, 0.00000465, 0.00000737, 0.00000882, 0.00000622, -0.00000145, -0.00001142, -0.00001806, -0.00001670, -0.00000734, 0.00000479, 0.00001267, 0.00001234, 0.00000550, -0.00000227, -0.00000582, -0.00000389, 0.00000077, 0.00000419, 0.00000424, 0.00000159, -0.00000153, -0.00000327, -0.00000292, -0.00000061, 0.00000310, 0.00000706, 0.00000917, 0.00000710, 0.00000013, -0.00000932, -0.00001633, -0.00001639, -0.00000886, 0.00000222, 0.00001063, 0.00001234, 0.00000803, 0.00000198, -0.00000160, -0.00000177, -0.00000083, -0.00000159, -0.00000459, -0.00000776, -0.00000861, -0.00000658, -0.00000332, -0.00000084, 0.00000055, 0.00000227, 0.00000560, 0.00000973, 0.00001205, 0.00001030, 0.00000484, -0.00000147, -0.00000560, -0.00000688, -0.00000718, -0.00000881, -0.00001187, -0.00001388, -0.00001193, -0.00000551, 0.00000263, 0.00000837, 0.00000917, 0.00000546, -0.00000034, -0.00000596, -0.00001041, -0.00001322, -0.00001327, -0.00000893, 0.00000014, 0.00001131, 0.00001950, 0.00002033, 0.00001346, 0.00000314, -0.00000469, -0.00000669, -0.00000420, -0.00000140, -0.00000128, -0.00000297, },
		{ 0.00001540, 0.00001693, 0.00001077, -0.00000044, -0.00001106, -0.00001619, -0.00001439, -0.00000776, -0.00000009, 0.00000532, 0.00000665, 0.00000376, -0.00000208, -0.00000845, -0.00001256, -0.00001267, -0.00000932, -0.00000516, -0.00000308, -0.00000382, -0.00000519, -0.00000381, 0.00000178, 0.00000935, 0.00001429, 0.00001303, 0.00000593, -0.00000294, -0.00000878, -0.00000949, -0.00000668, -0.00000371, -0.00000278, -0.00000352, -0.00000397, -0.00000273, -0.00000001, 0.00000298, 0.00000533, 0.00000682, 0.00000716, 0.00000529, 0.00000007, -0.00000789, -0.00001522, -0.00001728, -0.00001153, -0.00000016, 0.00001057, 0.00001446, 0.00000979, 0.00000061, -0.00000636, -0.00000673, -0.00000139, 0.00000467, 0.00000665, 0.00000345, -0.00000199, -0.00000546, -0.00000468, -0.00000046, 0.00000457, 0.00000793, 0.00000835, 0.00000558, -0.00000001, -0.00000716, -0.00001327, -0.00001510, -0.00001071, -0.00000143, 0.00000820, 0.00001310, 0.00001115, 0.00000469, -0.00000145, -0.00000370, -0.00000237, -0.00000080, -0.00000202, -0.00000592, -0.00000947, -0.00000955, -0.00000575, -0.00000058, 0.00000297, 0.00000404, 0.00000432, 0.00000585, 0.00000869, 0.00001063, 0.00000921, 0.00000422, -0.00000202, -0.00000664, -0.00000867, -0.00000946, -0.00001080, -0.00001263, -0.00001276, -0.00000901, -0.00000168, 0.00000598, 0.00000997, 0.00000856, 0.00000328, -0.00000269, -0.00000714, -0.00000992, -0.00001167, -0.00001182, -0.00000839, -0.00000012, 0.00001111, 0.00002016, 0.00002187, 0.00001496, 0.00000351, -0.00000583, -0.00000859, -0.00000552, -0.00000117, 0.00000033, -0.00000117, -0.00000210, 0.00000111, 0.00000837, },
		{ -0.00000047, -0.00001126, -0.00001633, -0.00001393, -0.00000655, 0.00000124, 0.00000562, 0.00000512, 0.00000064, -0.00000546, -0.00001049, -0.00001239, -0.00001067, -0.00000678, -0.00000336, -0.00000245, -0.00000380, -0.00000478, -0.00000243, 0.00000379, 0.00001091, 0.00001423, 0.00001082, 0.00000206, -0.00000718, -0.00001195, -0.00001069, -0.00000593, -0.00000182, -0.00000082, -0.00000217, -0.00000318, -0.00000183, 0.00000163, 0.00000533, 0.00000739, 0.00000708, 0.00000451, -0.00000003, -0.00000609, -0.00001219, -0.00001549, -0.00001314, -0.00000481, 0.00000598, 0.00001331, 0.00001288, 0.00000536, -0.00000375, -0.00000800, -0.00000491, 0.00000235, 0.00000757, 0.00000654, 0.00000025, -0.00000616, -0.00000764, -0.00000301, 0.00000446, 0.00000992, 0.00001028, 0.00000578, -0.00000111, -0.00000769, -0.00001206, -0.00001306, -0.00001000, -0.00000323, 0.00000511, 0.00001138, 0.00001244, 0.00000796, 0.00000104, -0.00000390, -0.00000453, -0.00000228, -0.00000092, -0.00000302, -0.00000755, -0.00001071, -0.00000930, -0.00000356, 0.00000300, 0.00000665, 0.00000653, 0.00000500, 0.00000503, 0.00000728, 0.00000947, 0.00000858, 0.00000375, -0.00000296, -0.00000840, -0.00001100, -0.00001157, -0.00001176, -0.00001183, -0.00001025, -0.00000548, 0.00000172, 0.00000808, 0.00001014, 0.00000699, 0.00000097, -0.00000442, -0.00000739, -0.00000871, -0.00000990, -0.00001062, -0.00000831, -0.00000074, 0.00001079, 0.00002089, 0.00002347, 0.00001640, 0.00000371, -0.00000712, -0.00001052, -0.00000671, -0.00000076, 0.00000201, 0.00000055, -0.00000162, -0.00000000, 0.00000634, 0.00001356, 0.00001596, 0.00001055, },
		{ -0.00001351, -0.00000571, 0.00000191, 0.00000521, 0.00000319, -0.00000226, -0.00000799, -0.00001137, -0.00001127, -0.00000835, -0.00000460, -0.00000225, -0.00000233, -0.00000375, -0.00000388, -0.00000054, 0.00000579, 0.00001166, 0.00001275, 0.00000724, -0.00000247, -0.00001103, -0.00001388, -0.00001042, -0.00000411, 0.00000043, 0.00000097, -0.00000105, -0.00000229, -0.00000055, 0.00000350, 0.00000714, 0.00000790, 0.00000517, 0.00000009, -0.00000565, -0.00001058, -0.00001332, -0.00001244, -0.00000714, 0.00000140, 0.00000951, 0.00001276, 0.00000911, 0.00000094, -0.00000605, -0.00000696, -0.00000151, 0.00000554, 0.00000814, 0.00000387, -0.00000400, -0.00000905, -0.00000688, 0.00000142, 0.00001012, 0.00001328, 0.00000887, -0.00000026, -0.00000900, -0.00001350, -0.00001293, -0.00000867, -0.00000258, 0.00000379, 0.00000898, 0.00001125, 0.00000943, 0.00000419, -0.00000168, -0.00000493, -0.00000444, -0.00000220, -0.00000164, -0.00000451, -0.00000896, -0.00001093, -0.00000768, -0.00000036, 0.00000670, 0.00000953, 0.00000773, 0.00000453, 0.00000367, 0.00000598, 0.00000872, 0.00000815, 0.00000299, -0.00000457, -0.00001075, -0.00001331, -0.00001282, -0.00001123, -0.00000953, -0.00000689, -0.00000216, 0.00000396, 0.00000866, 0.00000912, 0.00000506, -0.00000083, -0.00000517, -0.00000673, -0.00000710, -0.00000831, -0.00000994, -0.00000872, -0.00000159, 0.00001052, 0.00002179, 0.00002514, 0.00001775, 0.00000371, -0.00000856, -0.00001244, -0.00000775, -0.00000021, 0.00000372, 0.00000209, -0.00000141, -0.00000137, 0.00000418, 0.00001176, 0.00001515, 0.00001057, -0.00000024, -0.00001121, -0.00001629, },
		{ 0.00000445, 0.00000134, -0.00000455, -0.00000956, -0.00001133, -0.00000968, -0.00000617, -0.00000300, -0.00000171, -0.00000236, -0.00000335, -0.00000242, 0.00000156, 0.00000726, 0.00001115, 0.00000982, 0.00000268, -0.00000698, -0.00001386, -0.00001429, -0.00000879, -0.00000158, 0.00000273, 0.00000250, -0.00000007, -0.00000128, 0.00000085, 0.00000495, 0.00000770, 0.00000655, 0.00000160, -0.00000484, -0.00001003, -0.00001231, -0.00001131, -0.00000734, -0.00000118, 0.00000548, 0.00000996, 0.00000983, 0.00000489, -0.00000187, -0.00000582, -0.00000409, 0.00000179, 0.00000676, 0.00000616, -0.00000020, -0.00000743, -0.00000925, -0.00000318, 0.00000712, 0.00001438, 0.00001312, 0.00000372, -0.00000811, -0.00001562, -0.00001553, -0.00000942, -0.00000144, 0.00000488, 0.00000836, 0.00000939, 0.00000842, 0.00000553, 0.00000124, -0.00000284, -0.00000481, -0.00000419, -0.00000272, -0.00000303, -0.00000601, -0.00000950, -0.00000979, -0.00000492, 0.00000318, 0.00000980, 0.00001121, 0.00000774, 0.00000339, 0.00000228, 0.00000500, 0.00000819, 0.00000753, 0.00000163, -0.00000679, -0.00001321, -0.00001492, -0.00001266, -0.00000915, -0.00000621, -0.00000351, 0.00000017, 0.00000469, 0.00000790, 0.00000752, 0.00000346, -0.00000169, -0.00000490, -0.00000550, -0.00000553, -0.00000722, -0.00000987, -0.00000952, -0.00000247, 0.00001046, 0.00002289, 0.00002684, 0.00001892, 0.00000348, -0.00001013, -0.00001431, -0.00000864, 0.00000045, 0.00000535, 0.00000342, -0.00000147, -0.00000290, 0.00000205, 0.00001014, 0.00001460, 0.00001083, 0.00000017, -0.00001108, -0.00001634, -0.00001337, -0.00000533, 0.00000205, },
		{ -0.00001033, -0.00001076, -0.00000805, -0.00000441, -0.00000194, -0.00000146, -0.00000220, -0.00000243, -0.00000059, 0.00000343, 0.00000772, 0.00000920, 0.00000566, -0.00000226, -0.00001080, -0.00001524, -0.00001313, -0.00000609, 0.00000129, 0.00000482, 0.00000370, 0.00000073, -0.00000036, 0.00000187, 0.00000543, 0.00000675, 0.00000373, -0.00000257, -0.00000887, -0.00001207, -0.00001116, -0.00000717, -0.00000190, 0.00000312, 0.00000677, 0.00000797, 0.00000609, 0.00000183, -0.00000239, -0.00000365, -0.00000094, 0.00000350, 0.00000553, 0.00000250, -0.00000394, -0.00000853, -0.00000646, 0.00000226, 0.00001203, 0.00001548, 0.00000912, -0.00000376, -0.00001534, -0.00001885, -0.00001309, -0.00000260, 0.00000637, 0.00001019, 0.00000940, 0.00000676, 0.00000434, 0.00000230, -0.00000006, -0.00000262, -0.00000433, -0.00000452, -0.00000406, -0.00000465, -0.00000682, -0.00000873, -0.00000741, -0.00000163, 0.00000630, 0.00001181, 0.00001172, 0.00000700, 0.00000210, 0.00000112, 0.00000424, 0.00000751, 0.00000637, -0.00000033, -0.00000919, -0.00001508, -0.00001523, -0.00001094, -0.00000589, -0.00000264, -0.00000093, 0.00000103, 0.00000395, 0.00000633, 0.00000603, 0.00000272, -0.00000146, -0.00000389, -0.00000414, -0.00000441, -0.00000682, -0.00001032, -0.00001047, -0.00000317, 0.00001069, 0.00002415, 0.00002848, 0.00001988, 0.00000302, -0.00001179, -0.00001612, -0.00000943, 0.00000113, 0.00000686, 0.00000452, -0.00000170, -0.00000444, 0.00000009, 0.00000879, 0.00001431, 0.00001124, 0.00000057, -0.00001109, -0.00001662, -0.00001356, -0.00000529, 0.00000195, 0.00000372, -0.00000010, -0.00000611, },
		{ -0.00000309, -0.00000122, -0.00000117, -0.00000164, -0.00000109, 0.00000123, 0.00000456, 0.00000686, 0.00000588, 0.00000075, -0.00000690, -0.00001336, -0.00001494, -0.00001055, -0.00000268, 0.00000413, 0.00000647, 0.00000443, 0.00000117, 0.00000013, 0.00000213, 0.00000475, 0.00000461, 0.00000034, -0.00000611, -0.00001099, -0.00001157, -0.00000793, -0.00000243, 0.00000232, 0.00000499, 0.00000568, 0.00000490, 0.00000307, 0.00000074, -0.00000098, -0.00000090, 0.00000102, 0.00000296, 0.00000246, -0.00000126, -0.00000576, -0.00000672, -0.00000158, 0.00000750, 0.00001427, 0.00001279, 0.00000235, -0.00001131, -0.00001970, -0.00001767, -0.00000696, 0.00000540, 0.00001246, 0.00001201, 0.00000708, 0.00000248, 0.00000069, 0.00000077, 0.00000029, -0.00000179, -0.00000436, -0.00000581, -0.00000587, -0.00000577, -0.00000635, -0.00000660, -0.00000431, 0.00000141, 0.00000843, 0.00001266, 0.00001142, 0.00000602, 0.00000096, 0.00000012, 0.00000332, 0.00000633, 0.00000463, -0.00000250, -0.00001111, -0.00001578, -0.00001402, -0.00000798, -0.00000221, 0.00000032, 0.00000027, 0.00000037, 0.00000221, 0.00000467, 0.00000528, 0.00000308, -0.00000034, -0.00000258, -0.00000314, -0.00000399, -0.00000706, -0.00001107, -0.00001133, -0.00000356, 0.00001120, 0.00002549, 0.00002998, 0.00002059, 0.00000238, -0.00001351, -0.00001791, -0.00001020, 0.00000174, 0.00000823, 0.00000550, -0.00000194, -0.00000585, -0.00000164, 0.00000768, 0.00001414, 0.00001161, 0.00000081, -0.00001133, -0.00001716, -0.00001395, -0.00000538, 0.00000190, 0.00000328, -0.00000101, -0.00000703, -0.00001058, -0.00001001, -0.00000664, },
		{ -0.00000067, 0.00000042, 0.00000260, 0.00000457, 0.00000459, 0.00000150, -0.00000429, -0.00001058, -0.00001422, -0.00001290, -0.00000683, 0.00000099, 0.00000652, 0.00000738, 0.00000445, 0.00000100, -0.00000003, 0.00000156, 0.00000322, 0.00000203, -0.00000259, -0.00000813, -0.00001095, -0.00000915, -0.00000394, 0.00000149, 0.00000450, 0.00000465, 0.00000335, 0.00000219, 0.00000168, 0.00000148, 0.00000122, 0.00000103, 0.00000099, 0.00000056, -0.00000104, -0.00000352, -0.00000488, -0.00000264, 0.00000361, 0.00001061, 0.00001293, 0.00000705, -0.00000513, -0.00001660, -0.00001989, -0.00001253, 0.00000094, 0.00001220, 0.00001522, 0.00001024, 0.00000268, -0.00000170, -0.00000127, 0.00000112, 0.00000154, -0.00000134, -0.00000547, -0.00000783, -0.00000742, -0.00000568, -0.00000442, -0.00000357, -0.00000130, 0.00000355, 0.00000941, 0.00001267, 0.00001084, 0.00000516, -0.00000005, -0.00000104, 0.00000190, 0.00000458, 0.00000266, -0.00000425, -0.00001193, -0.00001501, -0.00001153, -0.00000445, 0.00000105, 0.00000206, -0.00000003, -0.00000144, 0.00000016, 0.00000356, 0.00000560, 0.00000445, 0.00000125, -0.00000149, -0.00000282, -0.00000432, -0.00000776, -0.00001185, -0.00001189, -0.00000359, 0.00001192, 0.00002681, 0.00003129, 0.00002106, 0.00000155, -0.00001534, -0.00001974, -0.00001101, 0.00000230, 0.00000958, 0.00000653, -0.00000204, -0.00000705, -0.00000318, 0.00000663, 0.00001388, 0.00001178, 0.00000081, -0.00001177, -0.00001780, -0.00001436, -0.00000536, 0.00000209, 0.00000323, -0.00000143, -0.00000751, -0.00001061, -0.00000936, -0.00000557, -0.00000211, -0.00000058, -0.00000061, },
		{ 0.00000324, 0.00000107, -0.00000340, -0.00000875, -0.00001271, -0.00001317, -0.00000932, -0.00000248, 0.00000433, 0.00000797, 0.00000722, 0.00000358, 0.00000016, -0.00000076, 0.00000051, 0.00000150, -0.00000014, -0.00000432, -0.00000840, -0.00000925, -0.00000593, -0.00000047, 0.00000376, 0.00000471, 0.00000304, 0.00000110, 0.00000076, 0.00000193, 0.00000309, 0.00000289, 0.00000126, -0.00000084, -0.00000252, -0.00000344, -0.00000336, -0.00000165, 0.00000221, 0.00000725, 0.00001049, 0.00000840, 0.00000005, -0.00001101, -0.00001811, -0.00001600, -0.00000510, 0.00000823, 0.00001596, 0.00001424, 0.00000594, -0.00000201, -0.00000430, -0.00000114, 0.00000273, 0.00000261, -0.00000203, -0.00000759, -0.00000988, -0.00000789, -0.00000404, -0.00000138, -0.00000043, 0.00000088, 0.00000449, 0.00000948, 0.00001237, 0.00001038, 0.00000443, -0.00000126, -0.00000277, -0.00000015, 0.00000257, 0.00000110, -0.00000493, -0.00001133, -0.00001299, -0.00000842, -0.00000119, 0.00000327, 0.00000238, -0.00000154, -0.00000376, -0.00000155, 0.00000342, 0.00000699, 0.00000648, 0.00000281, -0.00000101, -0.00000332, -0.00000526, -0.00000863, -0.00001240, -0.00001204, -0.00000328, 0.00001275, 0.00002803, 0.00003236, 0.00002129, 0.00000052, -0.00001731, -0.00002168, -0.00001186, 0.00000292, 0.00001108, 0.00000778, -0.00000193, -0.00000813, -0.00000475, 0.00000542, 0.00001340, 0.00001174, 0.00000069, -0.00001223, -0.00001837, -0.00001460, -0.00000511, 0.00000259, 0.00000354, -0.00000153, -0.00000779, -0.00001066, -0.00000892, -0.00000478, -0.00000125, 0.00000022, 0.00000032, 0.00000056, 0.00000171, 0.00000308, },
		{ -0.00001189, -0.00001285, -0.00001022, -0.00000466, 0.00000182, 0.00000667, 0.00000802, 0.00000579, 0.00000188, -0.00000108, -0.00000158, -0.00000036, 0.00000030, -0.00000132, -0.00000472, -0.00000737, -0.00000685, -0.00000303, 0.00000172, 0.00000436, 0.00000372, 0.00000130, -0.00000019, 0.00000072, 0.00000305, 0.00000440, 0.00000315, -0.00000018, -0.00000348, -0.00000484, -0.00000373, -0.00000089, 0.00000265, 0.00000601, 0.00000807, 0.00000720, 0.00000221, -0.00000600, -0.00001359, -0.00001558, -0.00000941, 0.00000233, 0.00001291, 0.00001600, 0.00001038, 0.00000089, -0.00000537, -0.00000468, 0.00000060, 0.00000446, 0.00000259, -0.00000399, -0.00001008, -0.00001106, -0.00000674, -0.00000103, 0.00000203, 0.00000199, 0.00000183, 0.00000443, 0.00000919, 0.00001222, 0.00001012, 0.00000351, -0.00000309, -0.00000522, -0.00000252, 0.00000096, 0.00000064, -0.00000417, -0.00000948, -0.00001032, -0.00000550, 0.00000115, 0.00000418, 0.00000150, -0.00000371, -0.00000596, -0.00000249, 0.00000432, 0.00000918, 0.00000870, 0.00000392, -0.00000135, -0.00000458, -0.00000658, -0.00000940, -0.00001253, -0.00001175, -0.00000271, 0.00001360, 0.00002905, 0.00003316, 0.00002123, -0.00000076, -0.00001948, -0.00002370, -0.00001264, 0.00000377, 0.00001291, 0.00000932, -0.00000170, -0.00000931, -0.00000659, 0.00000391, 0.00001269, 0.00001164, 0.00000065, -0.00001250, -0.00001871, -0.00001463, -0.00000463, 0.00000333, 0.00000409, -0.00000144, -0.00000804, -0.00001084, -0.00000872, -0.00000412, -0.00000030, 0.00000133, 0.00000154, 0.00000175, 0.00000234, 0.00000236, 0.00000057, -0.00000330, -0.00000811, },
		{ 0.00000031, 0.00000519, 0.00000734, 0.00000638, 0.00000322, -0.00000021, -0.00000208, -0.00000180, -0.00000047, -0.00000001, -0.00000154, -0.00000423, -0.00000587, -0.00000469, -0.00000108, 0.00000260, 0.00000394, 0.00000251, 0.00000024, -0.00000035, 0.00000148, 0.00000401, 0.00000456, 0.00000192, -0.00000247, -0.00000562, -0.00000536, -0.00000185, 0.00000283, 0.00000635, 0.00000744, 0.00000589, 0.00000200, -0.00000360, -0.00000931, -0.00001237, -0.00001010, -0.00000226, 0.00000767, 0.00001403, 0.00001289, 0.00000528, -0.00000317, -0.00000651, -0.00000314, 0.00000285, 0.00000525, 0.00000111, -0.00000671, -0.00001202, -0.00001067, -0.00000400, 0.00000266, 0.00000493, 0.00000316, 0.00000162, 0.00000388, 0.00000906, 0.00001238, 0.00000977, 0.00000195, -0.00000579, -0.00000817, -0.00000458, 0.00000047, 0.00000169, -0.00000210, -0.00000698, -0.00000784, -0.00000343, 0.00000231, 0.00000399, -0.00000007, -0.00000594, -0.00000761, -0.00000249, 0.00000609, 0.00001180, 0.00001070, 0.00000430, -0.00000256, -0.00000645, -0.00000799, -0.00000981, -0.00001212, -0.00001102, -0.00000198, 0.00001432, 0.00002980, 0.00003362, 0.00002084, -0.00000233, -0.00002181, -0.00002568, -0.00001315, 0.00000504, 0.00001514, 0.00001104, -0.00000160, -0.00001088, -0.00000887, 0.00000209, 0.00001195, 0.00001174, 0.00000094, -0.00001247, -0.00001884, -0.00001451, -0.00000401, 0.00000425, 0.00000480, -0.00000129, -0.00000838, -0.00001122, -0.00000866, -0.00000344, 0.00000090, 0.00000277, 0.00000287, 0.00000251, 0.00000194, 0.00000028, -0.00000320, -0.00000787, -0.00001184, -0.00001311, -0.00001083, -0.00000573, },
		{ 0.00000320, 0.00000015, -0.00000183, -0.00000204, -0.00000087, 0.00000040, 0.00000038, -0.00000133, -0.00000364, -0.00000466, -0.00000324, -0.00000006, 0.00000273, 0.00000331, 0.00000176, 0.00000004, 0.00000013, 0.00000214, 0.00000404, 0.00000342, -0.00000015, -0.00000445, -0.00000625, -0.00000388, 0.00000137, 0.00000631, 0.00000811, 0.00000612, 0.00000167, -0.00000329, -0.00000723, -0.00000915, -0.00000810, -0.00000360, 0.00000335, 0.00000974, 0.00001187, 0.00000811, 0.00000074, -0.00000515, -0.00000549, -0.00000068, 0.00000442, 0.00000452, -0.00000148, -0.00000926, -0.00001254, -0.00000851, -0.00000022, 0.00000612, 0.00000662, 0.00000301, 0.00000072, 0.00000343, 0.00000935, 0.00001262, 0.00000882, -0.00000061, -0.00000921, -0.00001095, -0.00000551, 0.00000163, 0.00000420, 0.00000067, -0.00000469, -0.00000624, -0.00000250, 0.00000247, 0.00000316, -0.00000177, -0.00000781, -0.00000854, -0.00000168, 0.00000845, 0.00001447, 0.00001218, 0.00000380, -0.00000459, -0.00000869, -0.00000922, -0.00000966, -0.00001110, -0.00000993, -0.00000125, 0.00001478, 0.00003014, 0.00003367, 0.00002010, -0.00000411, -0.00002415, -0.00002742, -0.00001320, 0.00000682, 0.00001767, 0.00001268, -0.00000195, -0.00001305, -0.00001162, 0.00000017, 0.00001149, 0.00001233, 0.00000168, -0.00001217, -0.00001889, -0.00001436, -0.00000332, 0.00000528, 0.00000561, -0.00000116, -0.00000886, -0.00001177, -0.00000864, -0.00000255, 0.00000246, 0.00000444, 0.00000396, 0.00000239, 0.00000019, -0.00000309, -0.00000752, -0.00001180, -0.00001378, -0.00001200, -0.00000687, -0.00000051, 0.00000456, 0.00000674, 0.00000593, },
		{ -0.00000038, 0.00000129, 0.00000185, 0.00000078, -0.00000147, -0.00000358, -0.00000406, -0.00000237, 0.00000054, 0.00000278, 0.00000302, 0.00000160, 0.00000026, 0.00000049, 0.00000203, 0.00000297, 0.00000153, -0.00000202, -0.00000517, -0.00000508, -0.00000106, 0.00000461, 0.00000823, 0.00000743, 0.00000277, -0.00000297, -0.00000691, -0.00000780, -0.00000600, -0.00000249, 0.00000187, 0.00000608, 0.00000857, 0.00000772, 0.00000344, -0.00000197, -0.00000487, -0.00000317, 0.00000147, 0.00000456, 0.00000241, -0.00000433, -0.00001068, -0.00001126, -0.00000499, 0.00000370, 0.00000854, 0.00000687, 0.00000192, -0.00000024, 0.00000345, 0.00000991, 0.00001242, 0.00000684, -0.00000412, -0.00001273, -0.00001272, -0.00000469, 0.00000447, 0.00000757, 0.00000324, -0.00000337, -0.00000584, -0.00000253, 0.00000211, 0.00000225, -0.00000324, -0.00000920, -0.00000883, -0.00000027, 0.00001110, 0.00001692, 0.00001294, 0.00000238, -0.00000729, -0.00001103, -0.00000998, -0.00000876, -0.00000946, -0.00000862, -0.00000073, 0.00001475, 0.00002996, 0.00003330, 0.00001913, -0.00000592, -0.00002628, -0.00002873, -0.00001272, 0.00000899, 0.00002024, 0.00001392, -0.00000298, -0.00001586, -0.00001460, -0.00000147, 0.00001166, 0.00001358, 0.00000283, -0.00001179, -0.00001904, -0.00001431, -0.00000260, 0.00000645, 0.00000650, -0.00000111, -0.00000953, -0.00001244, -0.00000851, -0.00000132, 0.00000437, 0.00000609, 0.00000439, 0.00000101, -0.00000297, -0.00000733, -0.00001156, -0.00001416, -0.00001333, -0.00000860, -0.00000171, 0.00000431, 0.00000698, 0.00000587, 0.00000252, -0.00000077, -0.00000240, -0.00000202, },
		{ 0.00000031, -0.00000254, -0.00000418, -0.00000377, -0.00000151, 0.00000135, 0.00000318, 0.00000309, 0.00000160, 0.00000024, 0.00000015, 0.00000101, 0.00000129, -0.00000020, -0.00000282, -0.00000431, -0.00000263, 0.00000195, 0.00000664, 0.00000803, 0.00000480, -0.00000123, -0.00000638, -0.00000789, -0.00000554, -0.00000131, 0.00000254, 0.00000490, 0.00000572, 0.00000520, 0.00000327, 0.00000027, -0.00000244, -0.00000300, -0.00000078, 0.00000239, 0.00000317, -0.00000037, -0.00000645, -0.00001041, -0.00000836, -0.00000091, 0.00000689, 0.00000954, 0.00000594, 0.00000051, -0.00000079, 0.00000393, 0.00001029, 0.00001127, 0.00000374, -0.00000805, -0.00001547, -0.00001273, -0.00000196, 0.00000844, 0.00001086, 0.00000476, -0.00000342, -0.00000647, -0.00000299, 0.00000180, 0.00000159, -0.00000439, -0.00001021, -0.00000866, 0.00000156, 0.00001385, 0.00001895, 0.00001284, 0.00000005, -0.00001048, -0.00001318, -0.00000998, -0.00000697, -0.00000728, -0.00000735, -0.00000072, 0.00001404, 0.00002922, 0.00003266, 0.00001816, -0.00000749, -0.00002801, -0.00002956, -0.00001184, 0.00001130, 0.00002250, 0.00001449, -0.00000477, -0.00001909, -0.00001745, -0.00000246, 0.00001267, 0.00001546, 0.00000417, -0.00001158, -0.00001946, -0.00001438, -0.00000177, 0.00000780, 0.00000744, -0.00000120, -0.00001040, -0.00001312, -0.00000809, 0.00000035, 0.00000648, 0.00000732, 0.00000368, -0.00000188, -0.00000731, -0.00001167, -0.00001431, -0.00001414, -0.00001031, -0.00000356, 0.00000342, 0.00000739, 0.00000672, 0.00000254, -0.00000208, -0.00000430, -0.00000322, -0.00000008, 0.00000293, 0.00000409, 0.00000298, },
		{ -0.00000307, -0.00000004, 0.00000266, 0.00000379, 0.00000301, 0.00000112, -0.00000050, -0.00000093, -0.00000043, -0.00000018, -0.00000099, -0.00000226, -0.00000228, 0.00000010, 0.00000404, 0.00000687, 0.00000604, 0.00000140, -0.00000436, -0.00000757, -0.00000629, -0.00000168, 0.00000314, 0.00000555, 0.00000507, 0.00000310, 0.00000120, -0.00000003, -0.00000081, -0.00000117, -0.00000069, 0.00000067, 0.00000176, 0.00000079, -0.00000283, -0.00000710, -0.00000840, -0.00000453, 0.00000282, 0.00000878, 0.00000917, 0.00000438, -0.00000067, -0.00000080, 0.00000449, 0.00000993, 0.00000895, -0.00000010, -0.00001155, -0.00001659, -0.00001066, 0.00000225, 0.00001260, 0.00001313, 0.00000474, -0.00000467, -0.00000755, -0.00000326, 0.00000189, 0.00000124, -0.00000537, -0.00001100, -0.00000814, 0.00000371, 0.00001656, 0.00002038, 0.00001175, -0.00000315, -0.00001388, -0.00001474, -0.00000892, -0.00000427, -0.00000482, -0.00000653, -0.00000157, 0.00001254, 0.00002808, 0.00003206, 0.00001754, -0.00000861, -0.00002931, -0.00003006, -0.00001083, 0.00001339, 0.00002421, 0.00001432, -0.00000712, -0.00002239, -0.00001976, -0.00000253, 0.00001453, 0.00001774, 0.00000539, -0.00001178, -0.00002022, -0.00001447, -0.00000073, 0.00000933, 0.00000836, -0.00000149, -0.00001141, -0.00001366, -0.00000725, 0.00000240, 0.00000845, 0.00000762, 0.00000144, -0.00000622, -0.00001215, -0.00001506, -0.00001476, -0.00001131, -0.00000519, 0.00000189, 0.00000701, 0.00000763, 0.00000356, -0.00000245, -0.00000645, -0.00000592, -0.00000141, 0.00000402, 0.00000710, 0.00000639, 0.00000274, -0.00000160, -0.00000451, -0.00000493, },
		{ 0.00000399, 0.00000215, -0.00000022, -0.00000189, -0.00000221, -0.00000151, -0.00000076, -0.00000061, -0.00000071, 0.00000001, 0.00000210, 0.00000464, 0.00000553, 0.00000327, -0.00000141, -0.00000566, -0.00000647, -0.00000311, 0.00000220, 0.00000594, 0.00000598, 0.00000301, -0.00000039, -0.00000198, -0.00000145, -0.00000006, 0.00000085, 0.00000098, 0.00000068, -0.00000005, -0.00000167, -0.00000415, -0.00000606, -0.00000517, -0.00000063, 0.00000552, 0.00000927, 0.00000790, 0.00000277, -0.00000137, -0.00000053, 0.00000458, 0.00000849, 0.00000570, -0.00000387, -0.00001370, -0.00001561, -0.00000681, 0.00000701, 0.00001590, 0.00001373, 0.00000324, -0.00000656, -0.00000838, -0.00000291, 0.00000242, 0.00000095, -0.00000643, -0.00001168, -0.00000725, 0.00000615, 0.00001906, 0.00002097, 0.00000957, -0.00000702, -0.00001709, -0.00001528, -0.00000662, -0.00000083, -0.00000255, -0.00000663, -0.00000351, 0.00001036, 0.00002693, 0.00003195, 0.00001756, -0.00000924, -0.00003037, -0.00003057, -0.00001006, 0.00001503, 0.00002532, 0.00001359, -0.00000970, -0.00002536, -0.00002124, -0.00000163, 0.00001705, 0.00002010, 0.00000622, -0.00001249, -0.00002122, -0.00001445, 0.00000059, 0.00001103, 0.00000917, -0.00000199, -0.00001245, -0.00001388, -0.00000593, 0.00000458, 0.00000978, 0.00000651, -0.00000242, -0.00001148, -0.00001647, -0.00001634, -0.00001225, -0.00000594, 0.00000085, 0.00000610, 0.00000768, 0.00000456, -0.00000179, -0.00000756, -0.00000887, -0.00000450, 0.00000311, 0.00000938, 0.00001076, 0.00000689, 0.00000048, -0.00000483, -0.00000666, -0.00000495, -0.00000130, 0.00000223, 0.00000417, },
		{ -0.00000331, -0.00000293, -0.00000163, -0.00000028, 0.00000055, 0.00000100, 0.00000168, 0.00000289, 0.00000393, 0.00000348, 0.00000090, -0.00000274, -0.00000500, -0.00000390, 0.00000024, 0.00000475, 0.00000647, 0.00000427, -0.00000007, -0.00000335, -0.00000350, -0.00000106, 0.00000162, 0.00000253, 0.00000140, -0.00000058, -0.00000226, -0.00000334, -0.00000393, -0.00000363, -0.00000150, 0.00000260, 0.00000693, 0.00000865, 0.00000626, 0.00000147, -0.00000171, -0.00000046, 0.00000380, 0.00000607, 0.00000218, -0.00000667, -0.00001390, -0.00001264, -0.00000200, 0.00001123, 0.00001757, 0.00001261, 0.00000082, -0.00000830, -0.00000845, -0.00000189, 0.00000309, 0.00000039, -0.00000768, -0.00001214, -0.00000589, 0.00000881, 0.00002108, 0.00002048, 0.00000631, -0.00001123, -0.00001955, -0.00001441, -0.00000309, 0.00000288, -0.00000107, -0.00000808, -0.00000652, 0.00000795, 0.00002636, 0.00003277, 0.00001830, -0.00000965, -0.00003166, -0.00003152, -0.00000976, 0.00001622, 0.00002602, 0.00001263, -0.00001213, -0.00002774, -0.00002183, 0.00000009, 0.00001993, 0.00002225, 0.00000648, -0.00001369, -0.00002234, -0.00001420, 0.00000221, 0.00001281, 0.00000984, -0.00000263, -0.00001333, -0.00001367, -0.00000430, 0.00000645, 0.00000993, 0.00000371, -0.00000759, -0.00001670, -0.00001901, -0.00001460, -0.00000676, 0.00000093, 0.00000606, 0.00000742, 0.00000467, -0.00000128, -0.00000770, -0.00001079, -0.00000791, 0.00000023, 0.00000941, 0.00001436, 0.00001232, 0.00000478, -0.00000366, -0.00000846, -0.00000786, -0.00000340, 0.00000172, 0.00000486, 0.00000508, 0.00000305, 0.00000017, -0.00000223, },
		{ 0.00000079, 0.00000169, 0.00000208, 0.00000234, 0.00000266, 0.00000269, 0.00000176, -0.00000028, -0.00000245, -0.00000305, -0.00000109, 0.00000254, 0.00000534, 0.00000502, 0.00000149, -0.00000286, -0.00000490, -0.00000332, 0.00000039, 0.00000323, 0.00000316, 0.00000056, -0.00000247, -0.00000401, -0.00000370, -0.00000230, -0.00000046, 0.00000187, 0.00000474, 0.00000717, 0.00000741, 0.00000467, 0.00000047, -0.00000205, -0.00000099, 0.00000213, 0.00000318, -0.00000073, -0.00000780, -0.00001209, -0.00000838, 0.00000269, 0.00001403, 0.00001738, 0.00001023, -0.00000171, -0.00000924, -0.00000760, -0.00000051, 0.00000347, -0.00000065, -0.00000903, -0.00001218, -0.00000400, 0.00001145, 0.00002227, 0.00001871, 0.00000219, -0.00001519, -0.00002070, -0.00001194, 0.00000129, 0.00000617, -0.00000102, -0.00001104, -0.00001021, 0.00000600, 0.00002697, 0.00003470, 0.00001946, -0.00001041, -0.00003370, -0.00003317, -0.00000988, 0.00001726, 0.00002671, 0.00001179, -0.00001423, -0.00002948, -0.00002165, 0.00000238, 0.00002290, 0.00002401, 0.00000616, -0.00001527, -0.00002344, -0.00001365, 0.00000410, 0.00001463, 0.00001040, -0.00000327, -0.00001394, -0.00001312, -0.00000274, 0.00000744, 0.00000849, -0.00000064, -0.00001321, -0.00002060, -0.00001864, -0.00000956, 0.00000084, 0.00000752, 0.00000855, 0.00000476, -0.00000171, -0.00000825, -0.00001196, -0.00001039, -0.00000305, 0.00000736, 0.00001564, 0.00001708, 0.00001069, 0.00000014, -0.00000859, -0.00001107, -0.00000705, -0.00000006, 0.00000543, 0.00000685, 0.00000452, 0.00000061, -0.00000266, -0.00000416, -0.00000390, -0.00000249, -0.00000072, },
		{ 0.00000224, 0.00000216, 0.00000183, 0.00000103, -0.00000018, -0.00000114, -0.00000090, 0.00000087, 0.00000318, 0.00000415, 0.00000249, -0.00000120, -0.00000446, -0.00000485, -0.00000198, 0.00000200, 0.00000402, 0.00000259, -0.00000108, -0.00000420, -0.00000472, -0.00000269, 0.00000027, 0.00000273, 0.00000443, 0.00000577, 0.00000659, 0.00000594, 0.00000323, -0.00000045, -0.00000278, -0.00000225, -0.00000007, 0.00000060, -0.00000230, -0.00000704, -0.00000878, -0.00000384, 0.00000629, 0.00001501, 0.00001567, 0.00000742, -0.00000359, -0.00000914, -0.00000611, 0.00000072, 0.00000322, -0.00000211, -0.00001017, -0.00001157, -0.00000171, 0.00001368, 0.00002226, 0.00001572, -0.00000226, -0.00001821, -0.00002013, -0.00000806, 0.00000583, 0.00000822, -0.00000280, -0.00001523, -0.00001381, 0.00000532, 0.00002914, 0.00003750, 0.00002037, -0.00001220, -0.00003686, -0.00003545, -0.00001002, 0.00001861, 0.00002773, 0.00001120, -0.00001602, -0.00003074, -0.00002089, 0.00000505, 0.00002583, 0.00002531, 0.00000529, -0.00001713, -0.00002443, -0.00001276, 0.00000627, 0.00001653, 0.00001096, -0.00000378, -0.00001431, -0.00001252, -0.00000177, 0.00000708, 0.00000546, -0.00000584, -0.00001804, -0.00002191, -0.00001482, -0.00000183, 0.00000884, 0.00001177, 0.00000690, -0.00000189, -0.00000995, -0.00001399, -0.00001247, -0.00000550, 0.00000494, 0.00001488, 0.00001956, 0.00001610, 0.00000584, -0.00000582, -0.00001260, -0.00001132, -0.00000387, 0.00000443, 0.00000861, 0.00000710, 0.00000202, -0.00000299, -0.00000544, -0.00000504, -0.00000306, -0.00000090, 0.00000068, 0.00000161, 0.00000206, 0.00000222, },
		{ 0.00000130, 0.00000061, 0.00000021, 0.00000049, 0.00000144, 0.00000230, 0.00000197, -0.00000007, -0.00000293, -0.00000463, -0.00000363, -0.00000031, 0.00000299, 0.00000370, 0.00000114, -0.00000281, -0.00000516, -0.00000419, -0.00000066, 0.00000315, 0.00000542, 0.00000604, 0.00000592, 0.00000556, 0.00000441, 0.00000182, -0.00000162, -0.00000403, -0.00000397, -0.00000212, -0.00000095, -0.00000222, -0.00000475, -0.00000488, 0.00000001, 0.00000827, 0.00001436, 0.00001320, 0.00000496, -0.00000445, -0.00000817, -0.00000451, 0.00000137, 0.00000231, -0.00000368, -0.00001078, -0.00001031, 0.00000063, 0.00001502, 0.00002089, 0.00001186, -0.00000633, -0.00001966, -0.00001776, -0.00000336, 0.00000960, 0.00000838, -0.00000635, -0.00001993, -0.00001634, 0.00000652, 0.00003276, 0.00004044, 0.00002015, -0.00001561, -0.00004113, -0.00003786, -0.00000953, 0.00002073, 0.00002918, 0.00001069, -0.00001775, -0.00003172, -0.00001967, 0.00000805, 0.00002866, 0.00002613, 0.00000387, -0.00001926, -0.00002528, -0.00001149, 0.00000879, 0.00001864, 0.00001162, -0.00000420, -0.00001472, -0.00001240, -0.00000190, 0.00000523, 0.00000137, -0.00001073, -0.00002073, -0.00001983, -0.00000786, 0.00000704, 0.00001510, 0.00001195, 0.00000075, -0.00001120, -0.00001757, -0.00001603, -0.00000797, 0.00000332, 0.00001397, 0.00002024, 0.00001936, 0.00001114, -0.00000101, -0.00001113, -0.00001399, -0.00000860, 0.00000105, 0.00000867, 0.00000989, 0.00000498, -0.00000206, -0.00000671, -0.00000701, -0.00000416, -0.00000081, 0.00000118, 0.00000161, 0.00000142, 0.00000142, 0.00000177, 0.00000214, 0.00000222, 0.00000191, },
		{ 0.00000071, 0.00000079, 0.00000061, -0.00000028, -0.00000189, -0.00000343, -0.00000368, -0.00000194, 0.00000099, 0.00000305, 0.00000247, -0.00000062, -0.00000403, -0.00000504, -0.00000260, 0.00000190, 0.00000581, 0.00000735, 0.00000676, 0.00000544, 0.00000426, 0.00000279, 0.00000021, -0.00000315, -0.00000561, -0.00000558, -0.00000336, -0.00000110, -0.00000071, -0.00000170, -0.00000134, 0.00000253, 0.00000869, 0.00001273, 0.00001077, 0.00000335, -0.00000435, -0.00000683, -0.00000332, 0.00000127, 0.00000099, -0.00000497, -0.00001068, -0.00000864, 0.00000250, 0.00001512, 0.00001834, 0.00000781, -0.00000924, -0.00001924, -0.00001402, 0.00000121, 0.00001173, 0.00000643, -0.00001107, -0.00002405, -0.00001692, 0.00000975, 0.00003717, 0.00004247, 0.00001798, -0.00002075, -0.00004595, -0.00003956, -0.00000778, 0.00002376, 0.00003081, 0.00000987, -0.00001973, -0.00003249, -0.00001793, 0.00001144, 0.00003138, 0.00002638, 0.00000183, -0.00002163, -0.00002585, -0.00000964, 0.00001187, 0.00002105, 0.00001231, -0.00000484, -0.00001567, -0.00001325, -0.00000333, 0.00000232, -0.00000274, -0.00001395, -0.00002030, -0.00001441, 0.00000090, 0.00001491, 0.00001762, 0.00000731, -0.00000889, -0.00002073, -0.00002184, -0.00001270, 0.00000132, 0.00001402, 0.00002120, 0.00002117, 0.00001441, 0.00000335, -0.00000766, -0.00001370, -0.00001186, -0.00000349, 0.00000612, 0.00001098, 0.00000841, 0.00000077, -0.00000655, -0.00000916, -0.00000650, -0.00000153, 0.00000206, 0.00000265, 0.00000126, 0.00000003, 0.00000025, 0.00000160, 0.00000290, 0.00000330, 0.00000280, 0.00000193, 0.00000118, 0.00000077, },
		{ -0.00000194, -0.00000271, -0.00000295, -0.00000218, -0.00000041, 0.00000153, 0.00000222, 0.00000077, -0.00000214, -0.00000439, -0.00000389, -0.00000030, 0.00000453, 0.00000794, 0.00000847, 0.00000675, 0.00000449, 0.00000270, 0.00000095, -0.00000165, -0.00000488, -0.00000707, -0.00000653, -0.00000344, 0.00000001, 0.00000159, 0.00000124, 0.00000121, 0.00000367, 0.00000803, 0.00001086, 0.00000893, 0.00000265, -0.00000372, -0.00000567, -0.00000281, 0.00000060, -0.00000030, -0.00000566, -0.00000998, -0.00000704, 0.00000343, 0.00001396, 0.00001514, 0.00000439, -0.00001046, -0.00001710, -0.00000976, 0.00000463, 0.00001170, 0.00000274, -0.00001590, -0.00002650, -0.00001512, 0.00001450, 0.00004124, 0.00004252, 0.00001347, -0.00002714, -0.00005036, -0.00003968, -0.00000447, 0.00002741, 0.00003200, 0.00000823, -0.00002207, -0.00003287, -0.00001541, 0.00001530, 0.00003386, 0.00002586, -0.00000093, -0.00002413, -0.00002588, -0.00000691, 0.00001565, 0.00002368, 0.00001269, -0.00000623, -0.00001769, -0.00001531, -0.00000573, -0.00000071, -0.00000553, -0.00001442, -0.00001656, -0.00000672, 0.00000941, 0.00001961, 0.00001528, -0.00000152, -0.00001971, -0.00002755, -0.00002064, -0.00000379, 0.00001352, 0.00002357, 0.00002390, 0.00001659, 0.00000565, -0.00000487, -0.00001160, -0.00001220, -0.00000656, 0.00000234, 0.00000937, 0.00001017, 0.00000429, -0.00000426, -0.00000985, -0.00000926, -0.00000384, 0.00000197, 0.00000429, 0.00000257, -0.00000058, -0.00000199, -0.00000047, 0.00000259, 0.00000491, 0.00000517, 0.00000377, 0.00000199, 0.00000078, 0.00000026, -0.00000001, -0.00000042, -0.00000110, },
		{ -0.00000067, 0.00000062, 0.00000133, 0.00000082, -0.00000096, -0.00000308, -0.00000382, -0.00000191, 0.00000233, 0.00000689, 0.00000938, 0.00000881, 0.00000611, 0.00000312, 0.00000086, -0.00000110, -0.00000361, -0.00000643, -0.00000791, -0.00000647, -0.00000241, 0.00000188, 0.00000393, 0.00000344, 0.00000259, 0.00000379, 0.00000699, 0.00000934, 0.00000784, 0.00000252, -0.00000308, -0.00000503, -0.00000289, -0.00000025, -0.00000120, -0.00000575, -0.00000910, -0.00000604, 0.00000322, 0.00001192, 0.00001209, 0.00000230, -0.00000990, -0.00001392, -0.00000603, 0.00000615, 0.00000959, -0.00000177, -0.00001960, -0.00002653, -0.00001118, 0.00001966, 0.00004366, 0.00003991, 0.00000694, -0.00003375, -0.00005320, -0.00003766, 0.00000015, 0.00003090, 0.00003203, 0.00000552, -0.00002454, -0.00003241, -0.00001186, 0.00001952, 0.00003576, 0.00002426, -0.00000442, -0.00002645, -0.00002490, -0.00000298, 0.00002013, 0.00002613, 0.00001211, -0.00000899, -0.00002105, -0.00001828, -0.00000817, -0.00000254, -0.00000585, -0.00001176, -0.00001035, 0.00000133, 0.00001544, 0.00001972, 0.00000834, -0.00001248, -0.00002872, -0.00002915, -0.00001328, 0.00000911, 0.00002549, 0.00002883, 0.00002048, 0.00000714, -0.00000445, -0.00001067, -0.00001102, -0.00000670, 0.00000018, 0.00000660, 0.00000920, 0.00000613, -0.00000117, -0.00000826, -0.00001057, -0.00000679, 0.00000005, 0.00000491, 0.00000473, 0.00000064, -0.00000328, -0.00000349, 0.00000027, 0.00000517, 0.00000782, 0.00000694, 0.00000382, 0.00000080, -0.00000073, -0.00000098, -0.00000098, -0.00000145, -0.00000227, -0.00000284, -0.00000274, -0.00000194, },
		{ -0.00000083, -0.00000241, -0.00000327, -0.00000238, 0.00000065, 0.00000499, 0.00000875, 0.00001003, 0.00000834, 0.00000485, 0.00000135, -0.00000119, -0.00000316, -0.00000532, -0.00000738, -0.00000785, -0.00000545, -0.00000072, 0.00000383, 0.00000571, 0.00000469, 0.00000308, 0.00000347, 0.00000611, 0.00000836, 0.00000723, 0.00000247, -0.00000280, -0.00000491, -0.00000321, -0.00000086, -0.00000158, -0.00000554, -0.00000855, -0.00000597, 0.00000207, 0.00000967, 0.00000995, 0.00000183, -0.00000803, -0.00001072, -0.00000376, 0.00000555, 0.00000614, -0.00000582, -0.00002119, -0.00002407, -0.00000605, 0.00002384, 0.00004345, 0.00003463, -0.00000063, -0.00003922, -0.00005362, -0.00003355, 0.00000529, 0.00003326, 0.00003036, 0.00000189, -0.00002654, -0.00003061, -0.00000721, 0.00002373, 0.00003656, 0.00002138, -0.00000835, -0.00002798, -0.00002238, 0.00000227, 0.00002488, 0.00002763, 0.00000982, -0.00001348, -0.00002548, -0.00002122, -0.00000935, -0.00000199, -0.00000323, -0.00000665, -0.00000343, 0.00000752, 0.00001736, 0.00001513, -0.00000148, -0.00002261, -0.00003312, -0.00002437, -0.00000092, 0.00002296, 0.00003378, 0.00002762, 0.00001112, -0.00000483, -0.00001272, -0.00001168, -0.00000556, 0.00000118, 0.00000581, 0.00000725, 0.00000517, 0.00000009, -0.00000590, -0.00000952, -0.00000824, -0.00000263, 0.00000353, 0.00000587, 0.00000294, -0.00000246, -0.00000552, -0.00000328, 0.00000298, 0.00000890, 0.00001060, 0.00000749, 0.00000225, -0.00000172, -0.00000288, -0.00000217, -0.00000149, -0.00000187, -0.00000283, -0.00000320, -0.00000239, -0.00000083, 0.00000058, 0.00000111, 0.00000055, },
		{ -0.00000018, 0.00000356, 0.00000742, 0.00000988, 0.00000976, 0.00000705, 0.00000303, -0.00000065, -0.00000317, -0.00000487, -0.00000639, -0.00000750, -0.00000697, -0.00000386, 0.00000106, 0.00000533, 0.00000672, 0.00000521, 0.00000318, 0.00000323, 0.00000560, 0.00000774, 0.00000669, 0.00000214, -0.00000294, -0.00000501, -0.00000339, -0.00000106, -0.00000167, -0.00000552, -0.00000872, -0.00000676, 0.00000056, 0.00000800, 0.00000918, 0.00000274, -0.00000572, -0.00000850, -0.00000341, 0.00000331, 0.00000251, -0.00000826, -0.00002027, -0.00001983, -0.00000112, 0.00002577, 0.00004028, 0.00002750, -0.00000781, -0.00004238, -0.00005134, -0.00002808, 0.00000975, 0.00003365, 0.00002695, -0.00000202, -0.00002732, -0.00002716, -0.00000180, 0.00002724, 0.00003580, 0.00001735, -0.00001208, -0.00002797, -0.00001795, 0.00000852, 0.00002905, 0.00002722, 0.00000526, -0.00001949, -0.00003001, -0.00002275, -0.00000798, 0.00000151, 0.00000179, -0.00000070, 0.00000205, 0.00001011, 0.00001476, 0.00000724, -0.00001139, -0.00002898, -0.00003130, -0.00001387, 0.00001359, 0.00003408, 0.00003568, 0.00001958, -0.00000208, -0.00001591, -0.00001624, -0.00000698, 0.00000342, 0.00000871, 0.00000784, 0.00000339, -0.00000169, -0.00000571, -0.00000783, -0.00000724, -0.00000367, 0.00000141, 0.00000492, 0.00000420, -0.00000042, -0.00000525, -0.00000584, -0.00000071, 0.00000718, 0.00001257, 0.00001191, 0.00000601, -0.00000096, -0.00000490, -0.00000474, -0.00000256, -0.00000119, -0.00000174, -0.00000309, -0.00000339, -0.00000197, 0.00000022, 0.00000155, 0.00000114, -0.00000060, -0.00000250, -0.00000340, -0.00000266, },
		{ 0.00001001, 0.00000841, 0.00000498, 0.00000090, -0.00000251, -0.00000467, -0.00000589, -0.00000669, -0.00000691, -0.00000563, -0.00000221, 0.00000248, 0.00000623, 0.00000715, 0.00000541, 0.00000332, 0.00000325, 0.00000532, 0.00000710, 0.00000588, 0.00000147, -0.00000325, -0.00000497, -0.00000321, -0.00000099, -0.00000188, -0.00000609, -0.00000966, -0.00000798, -0.00000059, 0.00000746, 0.00000972, 0.00000431, -0.00000397, -0.00000787, -0.00000476, 0.00000044, -0.00000007, -0.00000845, -0.00001726, -0.00001509, 0.00000218, 0.00002481, 0.00003466, 0.00001992, -0.00001316, -0.00004262, -0.00004692, -0.00002250, 0.00001240, 0.00003171, 0.00002237, -0.00000526, -0.00002625, -0.00002217, 0.00000365, 0.00002933, 0.00003333, 0.00001272, -0.00001467, -0.00002574, -0.00001173, 0.00001490, 0.00003145, 0.00002408, -0.00000151, -0.00002606, -0.00003313, -0.00002146, -0.00000339, 0.00000749, 0.00000767, 0.00000397, 0.00000427, 0.00000847, 0.00000874, -0.00000139, -0.00001841, -0.00002965, -0.00002347, -0.00000016, 0.00002656, 0.00003934, 0.00003021, 0.00000639, -0.00001568, -0.00002273, -0.00001347, 0.00000246, 0.00001313, 0.00001295, 0.00000480, -0.00000427, -0.00000905, -0.00000880, -0.00000570, -0.00000200, 0.00000123, 0.00000331, 0.00000323, 0.00000043, -0.00000374, -0.00000600, -0.00000341, 0.00000379, 0.00001153, 0.00001465, 0.00001083, 0.00000251, -0.00000500, -0.00000767, -0.00000549, -0.00000185, -0.00000026, -0.00000149, -0.00000346, -0.00000365, -0.00000151, 0.00000124, 0.00000229, 0.00000079, -0.00000203, -0.00000404, -0.00000378, -0.00000120, 0.00000271, 0.00000659, 0.00000930, },
		{ -0.00000107, -0.00000375, -0.00000534, -0.00000611, -0.00000640, -0.00000603, -0.00000431, -0.00000088, 0.00000344, 0.00000671, 0.00000736, 0.00000565, 0.00000364, 0.00000338, 0.00000494, 0.00000616, 0.00000474, 0.00000068, -0.00000336, -0.00000454, -0.00000275, -0.00000103, -0.00000264, -0.00000739, -0.00001104, -0.00000894, -0.00000080, 0.00000812, 0.00001101, 0.00000560, -0.00000352, -0.00000885, -0.00000699, -0.00000187, -0.00000079, -0.00000650, -0.00001324, -0.00001130, 0.00000294, 0.00002115, 0.00002786, 0.00001346, -0.00001577, -0.00004017, -0.00004153, -0.00001819, 0.00001255, 0.00002779, 0.00001761, -0.00000692, -0.00002318, -0.00001631, 0.00000821, 0.00002952, 0.00002952, 0.00000853, -0.00001514, -0.00002109, -0.00000446, 0.00002007, 0.00003096, 0.00001799, -0.00000962, -0.00003155, -0.00003320, -0.00001646, 0.00000405, 0.00001445, 0.00001227, 0.00000548, 0.00000248, 0.00000348, 0.00000164, -0.00000785, -0.00002041, -0.00002440, -0.00001170, 0.00001318, 0.00003456, 0.00003717, 0.00001852, -0.00000866, -0.00002595, -0.00002308, -0.00000480, 0.00001400, 0.00002028, 0.00001183, -0.00000306, -0.00001347, -0.00001410, -0.00000744, 0.00000023, 0.00000428, 0.00000412, 0.00000167, -0.00000123, -0.00000374, -0.00000500, -0.00000356, 0.00000150, 0.00000874, 0.00001421, 0.00001394, 0.00000728, -0.00000207, -0.00000857, -0.00000900, -0.00000478, -0.00000036, 0.00000069, -0.00000160, -0.00000411, -0.00000389, -0.00000086, 0.00000231, 0.00000276, 0.00000002, -0.00000367, -0.00000528, -0.00000336, 0.00000119, 0.00000609, 0.00000936, 0.00001020, 0.00000886, 0.00000599, 0.00000241, },
		{ -0.00000583, -0.00000600, -0.00000532, -0.00000331, 0.00000013, 0.00000415, 0.00000709, 0.00000763, 0.00000601, 0.00000397, 0.00000333, 0.00000421, 0.00000489, 0.00000349, 0.00000011, -0.00000307, -0.00000380, -0.00000236, -0.00000158, -0.00000411, -0.00000915, -0.00001225, -0.00000907, 0.00000015, 0.00000958, 0.00001220, 0.00000584, -0.00000453, -0.00001083, -0.00000901, -0.00000270, 0.00000049, -0.00000328, -0.00000960, -0.00000956, 0.00000106, 0.00001582, 0.00002145, 0.00000928, -0.00001552, -0.00003606, -0.00003666, -0.00001615, 0.00001024, 0.00002285, 0.00001381, -0.00000650, -0.00001854, -0.00001058, 0.00001110, 0.00002789, 0.00002532, 0.00000595, -0.00001292, -0.00001451, 0.00000249, 0.00002262, 0.00002699, 0.00000963, -0.00001735, -0.00003406, -0.00002908, -0.00000797, 0.00001284, 0.00002016, 0.00001360, 0.00000295, -0.00000260, -0.00000268, -0.00000370, -0.00000989, -0.00001685, -0.00001492, 0.00000063, 0.00002258, 0.00003553, 0.00002815, 0.00000359, -0.00002166, -0.00003010, -0.00001670, 0.00000743, 0.00002415, 0.00002221, 0.00000498, -0.00001375, -0.00002117, -0.00001455, -0.00000137, 0.00000818, 0.00000895, 0.00000312, -0.00000353, -0.00000678, -0.00000599, -0.00000264, 0.00000202, 0.00000734, 0.00001206, 0.00001372, 0.00001024, 0.00000228, -0.00000620, -0.00001043, -0.00000853, -0.00000306, 0.00000114, 0.00000104, -0.00000225, -0.00000483, -0.00000380, 0.00000010, 0.00000332, 0.00000277, -0.00000125, -0.00000535, -0.00000589, -0.00000187, 0.00000446, 0.00000960, 0.00001131, 0.00000967, 0.00000630, 0.00000280, -0.00000008, -0.00000228, -0.00000394, -0.00000513, },
		{ -0.00000256, 0.00000110, 0.00000498, 0.00000761, 0.00000797, 0.00000626, 0.00000402, 0.00000288, 0.00000314, 0.00000351, 0.00000248, -0.00000003, -0.00000240, -0.00000307, -0.00000244, -0.00000285, -0.00000611, -0.00001084, -0.00001271, -0.00000808, 0.00000193, 0.00001109, 0.00001253, 0.00000473, -0.00000658, -0.00001282, -0.00000986, -0.00000173, 0.00000313, -0.00000006, -0.00000754, -0.00001023, -0.00000271, 0.00001030, 0.00001676, 0.00000781, -0.00001319, -0.00003174, -0.00003353, -0.00001664, 0.00000625, 0.00001809, 0.00001176, -0.00000418, -0.00001330, -0.00000600, 0.00001204, 0.00002517, 0.00002202, 0.00000593, -0.00000814, -0.00000725, 0.00000748, 0.00002153, 0.00001987, 0.00000064, -0.00002260, -0.00003209, -0.00002069, 0.00000260, 0.00002067, 0.00002246, 0.00001061, -0.00000303, -0.00000892, -0.00000725, -0.00000501, -0.00000678, -0.00000911, -0.00000432, 0.00000997, 0.00002568, 0.00002951, 0.00001482, -0.00001069, -0.00002937, -0.00002718, -0.00000541, 0.00001977, 0.00002974, 0.00001785, -0.00000607, -0.00002415, -0.00002454, -0.00000953, 0.00000793, 0.00001533, 0.00000964, -0.00000229, -0.00001079, -0.00001084, -0.00000430, 0.00000366, 0.00000926, 0.00001176, 0.00001185, 0.00000961, 0.00000459, -0.00000238, -0.00000835, -0.00001004, -0.00000671, -0.00000130, 0.00000183, 0.00000056, -0.00000313, -0.00000515, -0.00000314, 0.00000126, 0.00000391, 0.00000208, -0.00000292, -0.00000668, -0.00000543, 0.00000080, 0.00000829, 0.00001259, 0.00001181, 0.00000744, 0.00000259, -0.00000052, -0.00000172, -0.00000220, -0.00000305, -0.00000439, -0.00000561, -0.00000599, -0.00000503, },
		{ 0.00000819, 0.00000805, 0.00000606, 0.00000358, 0.00000208, 0.00000199, 0.00000238, 0.00000190, 0.00000022, -0.00000169, -0.00000273, -0.00000320, -0.00000469, -0.00000815, -0.00001190, -0.00001212, -0.00000620, 0.00000391, 0.00001190, 0.00001161, 0.00000256, -0.00000882, -0.00001388, -0.00000911, 0.00000060, 0.00000602, 0.00000197, -0.00000759, -0.00001279, -0.00000703, 0.00000598, 0.00001444, 0.00000853, -0.00001013, -0.00002858, -0.00003272, -0.00001915, 0.00000177, 0.00001454, 0.00001160, -0.00000075, -0.00000860, -0.00000316, 0.00001149, 0.00002262, 0.00002083, 0.00000873, -0.00000181, -0.00000105, 0.00000912, 0.00001671, 0.00001103, -0.00000674, -0.00002347, -0.00002518, -0.00000927, 0.00001283, 0.00002513, 0.00002006, 0.00000375, -0.00001044, -0.00001369, -0.00000789, -0.00000141, 0.00000033, -0.00000002, 0.00000397, 0.00001377, 0.00002205, 0.00001858, 0.00000088, -0.00002082, -0.00003022, -0.00001838, 0.00000754, 0.00002870, 0.00002902, 0.00000813, -0.00001832, -0.00003104, -0.00002205, -0.00000003, 0.00001764, 0.00001885, 0.00000522, -0.00001095, -0.00001740, -0.00001092, 0.00000215, 0.00001262, 0.00001559, 0.00001242, 0.00000729, 0.00000283, -0.00000098, -0.00000479, -0.00000772, -0.00000789, -0.00000477, -0.00000059, 0.00000133, -0.00000037, -0.00000356, -0.00000463, -0.00000199, 0.00000212, 0.00000365, 0.00000063, -0.00000456, -0.00000704, -0.00000357, 0.00000445, 0.00001197, 0.00001424, 0.00001038, 0.00000365, -0.00000156, -0.00000304, -0.00000186, -0.00000074, -0.00000148, -0.00000382, -0.00000606, -0.00000663, -0.00000500, -0.00000164, 0.00000245, 0.00000611, },
		{ 0.00000277, 0.00000128, 0.00000117, 0.00000172, 0.00000171, 0.00000053, -0.00000131, -0.00000301, -0.00000453, -0.00000666, -0.00000970, -0.00001200, -0.00001060, -0.00000398, 0.00000538, 0.00001159, 0.00000962, 0.00000005, -0.00001034, -0.00001351, -0.00000701, 0.00000335, 0.00000801, 0.00000215, -0.00000947, -0.00001610, -0.00001053, 0.00000370, 0.00001426, 0.00001030, -0.00000772, -0.00002738, -0.00003397, -0.00002259, -0.00000206, 0.00001262, 0.00001276, 0.00000265, -0.00000527, -0.00000194, 0.00001055, 0.00002161, 0.00002240, 0.00001370, 0.00000443, 0.00000244, 0.00000685, 0.00000925, 0.00000273, -0.00001027, -0.00001904, -0.00001425, 0.00000279, 0.00002008, 0.00002460, 0.00001316, -0.00000503, -0.00001647, -0.00001448, -0.00000361, 0.00000610, 0.00000878, 0.00000705, 0.00000729, 0.00001130, 0.00001336, 0.00000615, -0.00001008, -0.00002475, -0.00002478, -0.00000659, 0.00001853, 0.00003187, 0.00002221, -0.00000433, -0.00002833, -0.00003222, -0.00001404, 0.00001137, 0.00002459, 0.00001694, -0.00000367, -0.00002021, -0.00002044, -0.00000558, 0.00001246, 0.00002157, 0.00001821, 0.00000797, -0.00000093, -0.00000443, -0.00000418, -0.00000380, -0.00000469, -0.00000537, -0.00000412, -0.00000153, 0.00000003, -0.00000085, -0.00000286, -0.00000326, -0.00000095, 0.00000202, 0.00000229, -0.00000115, -0.00000542, -0.00000585, -0.00000033, 0.00000839, 0.00001452, 0.00001382, 0.00000701, -0.00000098, -0.00000508, -0.00000391, -0.00000028, 0.00000167, -0.00000008, -0.00000416, -0.00000747, -0.00000769, -0.00000470, -0.00000005, 0.00000439, 0.00000738, 0.00000839, 0.00000747, 0.00000522, },
		{ 0.00000143, 0.00000153, 0.00000050, -0.00000150, -0.00000379, -0.00000599, -0.00000824, -0.00001042, -0.00001123, -0.00000866, -0.00000208, 0.00000590, 0.00001024, 0.00000716, -0.00000197, -0.00001063, -0.00001184, -0.00000437, 0.00000542, 0.00000837, 0.00000060, -0.00001221, -0.00001886, -0.00001229, 0.00000342, 0.00001524, 0.00001173, -0.00000693, -0.00002811, -0.00003635, -0.00002578, -0.00000457, 0.00001200, 0.00001418, 0.00000500, -0.00000352, -0.00000146, 0.00001061, 0.00002309, 0.00002646, 0.00001939, 0.00000875, 0.00000224, 0.00000132, 0.00000122, -0.00000261, -0.00000856, -0.00000979, -0.00000154, 0.00001264, 0.00002235, 0.00001894, 0.00000358, -0.00001286, -0.00001851, -0.00001014, 0.00000472, 0.00001494, 0.00001527, 0.00000944, 0.00000475, 0.00000391, 0.00000271, -0.00000425, -0.00001570, -0.00002250, -0.00001543, 0.00000469, 0.00002480, 0.00002883, 0.00001140, -0.00001611, -0.00003342, -0.00002732, -0.00000264, 0.00002137, 0.00002638, 0.00000956, -0.00001469, -0.00002685, -0.00001781, 0.00000484, 0.00002410, 0.00002735, 0.00001517, -0.00000130, -0.00001058, -0.00000947, -0.00000333, 0.00000048, -0.00000079, -0.00000428, -0.00000569, -0.00000381, -0.00000095, 0.00000012, -0.00000081, -0.00000166, -0.00000092, 0.00000044, 0.00000014, -0.00000240, -0.00000462, -0.00000290, 0.00000365, 0.00001148, 0.00001496, 0.00001112, 0.00000237, -0.00000525, -0.00000691, -0.00000275, 0.00000258, 0.00000405, 0.00000026, -0.00000581, -0.00000958, -0.00000849, -0.00000346, 0.00000249, 0.00000669, 0.00000824, 0.00000769, 0.00000599, 0.00000390, 0.00000203, 0.00000095, 0.00000091, },
		{ -0.00000212, -0.00000464, -0.00000709, -0.00000912, -0.00001041, -0.00001005, -0.00000689, -0.00000090, 0.00000553, 0.00000838, 0.00000498, -0.00000299, -0.00000971, -0.00000954, -0.00000217, 0.00000610, 0.00000709, -0.00000190, -0.00001460, -0.00002010, -0.00001215, 0.00000435, 0.00001610, 0.00001179, -0.00000794, -0.00003003, -0.00003871, -0.00002790, -0.00000587, 0.00001179, 0.00001479, 0.00000587, -0.00000273, -0.00000042, 0.00001283, 0.00002718, 0.00003186, 0.00002398, 0.00000984, -0.00000151, -0.00000571, -0.00000487, -0.00000316, -0.00000158, 0.00000234, 0.00001004, 0.00001792, 0.00001908, 0.00000972, -0.00000579, -0.00001691, -0.00001516, -0.00000139, 0.00001455, 0.00002181, 0.00001708, 0.00000613, -0.00000253, -0.00000554, -0.00000648, -0.00001012, -0.00001559, -0.00001594, -0.00000544, 0.00001248, 0.00002528, 0.00002103, -0.00000023, -0.00002418, -0.00003247, -0.00001784, 0.00000897, 0.00002707, 0.00002221, -0.00000144, -0.00002452, -0.00002818, -0.00000913, 0.00001818, 0.00003365, 0.00002734, 0.00000614, -0.00001335, -0.00001862, -0.00001002, 0.00000186, 0.00000651, 0.00000181, -0.00000602, -0.00000941, -0.00000619, -0.00000026, 0.00000306, 0.00000202, -0.00000100, -0.00000277, -0.00000253, -0.00000187, -0.00000196, -0.00000164, 0.00000134, 0.00000715, 0.00001246, 0.00001280, 0.00000671, -0.00000224, -0.00000786, -0.00000645, 0.00000006, 0.00000566, 0.00000530, -0.00000103, -0.00000852, -0.00001158, -0.00000818, -0.00000088, 0.00000574, 0.00000867, 0.00000799, 0.00000575, 0.00000379, 0.00000263, 0.00000191, 0.00000136, 0.00000109, 0.00000110, 0.00000091, -0.00000011, },
		{ -0.00000940, -0.00001009, -0.00000901, -0.00000567, -0.00000042, 0.00000472, 0.00000666, 0.00000355, -0.00000300, -0.00000817, -0.00000748, -0.00000111, 0.00000528, 0.00000481, -0.00000426, -0.00001569, -0.00001955, -0.00001074, 0.00000534, 0.00001576, 0.00001014, -0.00001020, -0.00003207, -0.00004013, -0.00002884, -0.00000668, 0.00001088, 0.00001393, 0.00000558, -0.00000175, 0.00000244, 0.00001765, 0.00003307, 0.00003692, 0.00002596, 0.00000743, -0.00000742, -0.00001181, -0.00000690, 0.00000163, 0.00000917, 0.00001454, 0.00001783, 0.00001760, 0.00001145, -0.00000024, -0.00001191, -0.00001548, -0.00000693, 0.00000929, 0.00002253, 0.00002389, 0.00001309, -0.00000187, -0.00001182, -0.00001375, -0.00001167, -0.00001078, -0.00001120, -0.00000803, 0.00000220, 0.00001530, 0.00002085, 0.00001124, -0.00000958, -0.00002689, -0.00002633, -0.00000663, 0.00001763, 0.00002700, 0.00001325, -0.00001295, -0.00003005, -0.00002305, 0.00000400, 0.00003100, 0.00003792, 0.00002048, -0.00000725, -0.00002495, -0.00002217, -0.00000517, 0.00001008, 0.00001184, 0.00000117, -0.00001091, -0.00001404, -0.00000694, 0.00000308, 0.00000762, 0.00000430, -0.00000251, -0.00000674, -0.00000600, -0.00000233, 0.00000107, 0.00000329, 0.00000559, 0.00000861, 0.00001050, 0.00000840, 0.00000192, -0.00000526, -0.00000793, -0.00000394, 0.00000338, 0.00000763, 0.00000467, -0.00000377, -0.00001138, -0.00001243, -0.00000621, 0.00000281, 0.00000891, 0.00000947, 0.00000619, 0.00000275, 0.00000151, 0.00000212, 0.00000280, 0.00000241, 0.00000124, 0.00000020, -0.00000040, -0.00000115, -0.00000270, -0.00000504, -0.00000753, },
		{ -0.00000494, -0.00000019, 0.00000406, 0.00000551, 0.00000286, -0.00000252, -0.00000677, -0.00000631, -0.00000132, 0.00000351, 0.00000251, -0.00000557, -0.00001515, -0.00001769, -0.00000910, 0.00000532, 0.00001378, 0.00000717, -0.00001271, -0.00003327, -0.00004033, -0.00002913, -0.00000802, 0.00000852, 0.00001168, 0.00000506, 0.00000067, 0.00000781, 0.00002460, 0.00003929, 0.00004002, 0.00002467, 0.00000244, -0.00001329, -0.00001472, -0.00000387, 0.00001075, 0.00002111, 0.00002398, 0.00002036, 0.00001241, 0.00000209, -0.00000779, -0.00001278, -0.00000879, 0.00000383, 0.00001847, 0.00002565, 0.00001988, 0.00000421, -0.00001191, -0.00001987, -0.00001815, -0.00001201, -0.00000740, -0.00000521, -0.00000166, 0.00000578, 0.00001352, 0.00001378, 0.00000246, -0.00001463, -0.00002442, -0.00001732, 0.00000317, 0.00002134, 0.00002158, 0.00000219, -0.00002166, -0.00002941, -0.00001231, 0.00001842, 0.00003973, 0.00003507, 0.00000776, -0.00002184, -0.00003272, -0.00001959, 0.00000420, 0.00001875, 0.00001395, -0.00000362, -0.00001791, -0.00001755, -0.00000454, 0.00000910, 0.00001244, 0.00000439, -0.00000692, -0.00001216, -0.00000836, 0.00000012, 0.00000687, 0.00000898, 0.00000804, 0.00000677, 0.00000564, 0.00000310, -0.00000141, -0.00000548, -0.00000550, -0.00000057, 0.00000566, 0.00000746, 0.00000218, -0.00000697, -0.00001312, -0.00001134, -0.00000269, 0.00000676, 0.00001093, 0.00000843, 0.00000293, -0.00000059, 0.00000000, 0.00000287, 0.00000461, 0.00000353, 0.00000071, -0.00000160, -0.00000228, -0.00000212, -0.00000269, -0.00000468, -0.00000741, -0.00000955, -0.00001004, -0.00000847, },
		{ 0.00000487, 0.00000242, -0.00000225, -0.00000608, -0.00000611, -0.00000230, 0.00000165, 0.00000100, -0.00000549, -0.00001335, -0.00001543, -0.00000821, 0.00000380, 0.00001041, 0.00000373, -0.00001455, -0.00003322, -0.00003966, -0.00002969, -0.00001072, 0.00000456, 0.00000875, 0.00000550, 0.00000535, 0.00001551, 0.00003248, 0.00004426, 0.00004027, 0.00002063, -0.00000324, -0.00001687, -0.00001307, 0.00000374, 0.00002200, 0.00003142, 0.00002884, 0.00001785, 0.00000462, -0.00000588, -0.00001067, -0.00000826, 0.00000102, 0.00001364, 0.00002294, 0.00002243, 0.00001060, -0.00000693, -0.00002068, -0.00002405, -0.00001781, -0.00000854, -0.00000243, -0.00000033, 0.00000138, 0.00000532, 0.00000898, 0.00000681, -0.00000316, -0.00001502, -0.00001852, -0.00000836, 0.00000921, 0.00001980, 0.00001289, -0.00000773, -0.00002515, -0.00002265, 0.00000142, 0.00003045, 0.00004182, 0.00002528, -0.00000803, -0.00003393, -0.00003425, -0.00001110, 0.00001563, 0.00002487, 0.00001122, -0.00001197, -0.00002488, -0.00001787, 0.00000160, 0.00001660, 0.00001554, 0.00000098, -0.00001396, -0.00001736, -0.00000781, 0.00000618, 0.00001447, 0.00001343, 0.00000696, 0.00000126, -0.00000089, -0.00000106, -0.00000166, -0.00000261, -0.00000171, 0.00000188, 0.00000552, 0.00000499, -0.00000113, -0.00000913, -0.00001269, -0.00000829, 0.00000144, 0.00000966, 0.00001092, 0.00000556, -0.00000102, -0.00000325, -0.00000005, 0.00000486, 0.00000673, 0.00000397, -0.00000094, -0.00000418, -0.00000413, -0.00000235, -0.00000172, -0.00000370, -0.00000728, -0.00001014, -0.00001057, -0.00000831, -0.00000422, 0.00000038, 0.00000393, },
		{ -0.00000612, -0.00000640, -0.00000328, 0.00000041, 0.00000057, -0.00000434, -0.00001109, -0.00001365, -0.00000858, 0.00000090, 0.00000635, 0.00000068, -0.00001524, -0.00003210, -0.00003889, -0.00003134, -0.00001510, -0.00000054, 0.00000615, 0.00000776, 0.00001240, 0.00002466, 0.00003988, 0.00004696, 0.00003784, 0.00001536, -0.00000747, -0.00001656, -0.00000689, 0.00001417, 0.00003271, 0.00003805, 0.00002887, 0.00001217, -0.00000280, -0.00000999, -0.00000834, -0.00000017, 0.00001054, 0.00001913, 0.00002095, 0.00001347, -0.00000135, -0.00001691, -0.00002545, -0.00002334, -0.00001366, -0.00000355, 0.00000151, 0.00000162, 0.00000092, 0.00000231, 0.00000412, 0.00000210, -0.00000491, -0.00001200, -0.00001174, -0.00000187, 0.00001069, 0.00001446, 0.00000387, -0.00001382, -0.00002274, -0.00001170, 0.00001460, 0.00003702, 0.00003660, 0.00001073, -0.00002315, -0.00004045, -0.00002893, 0.00000128, 0.00002587, 0.00002609, 0.00000358, -0.00002186, -0.00002927, -0.00001369, 0.00001071, 0.00002348, 0.00001502, -0.00000624, -0.00002213, -0.00002017, -0.00000302, 0.00001534, 0.00002188, 0.00001445, 0.00000140, -0.00000702, -0.00000690, -0.00000196, 0.00000191, 0.00000237, 0.00000143, 0.00000162, 0.00000248, 0.00000119, -0.00000352, -0.00000883, -0.00000982, -0.00000423, 0.00000467, 0.00001036, 0.00000870, 0.00000165, -0.00000442, -0.00000433, 0.00000145, 0.00000745, 0.00000824, 0.00000313, -0.00000363, -0.00000691, -0.00000519, -0.00000141, 0.00000005, -0.00000270, -0.00000779, -0.00001154, -0.00001153, -0.00000798, -0.00000289, 0.00000162, 0.00000420, 0.00000418, 0.00000153, -0.00000268, },
		{ 0.00000000, 0.00000090, -0.00000286, -0.00000914, -0.00001281, -0.00001011, -0.00000276, 0.00000236, -0.00000150, -0.00001484, -0.00003053, -0.00003874, -0.00003445, -0.00002085, -0.00000593, 0.00000472, 0.00001210, 0.00002118, 0.00003404, 0.00004574, 0.00004730, 0.00003387, 0.00001077, -0.00000861, -0.00001202, 0.00000252, 0.00002505, 0.00004079, 0.00004039, 0.00002546, 0.00000598, -0.00000733, -0.00000943, -0.00000231, 0.00000820, 0.00001630, 0.00001828, 0.00001288, 0.00000133, -0.00001247, -0.00002284, -0.00002505, -0.00001859, -0.00000796, 0.00000034, 0.00000262, 0.00000040, -0.00000177, -0.00000106, 0.00000096, 0.00000048, -0.00000364, -0.00000769, -0.00000641, 0.00000099, 0.00000844, 0.00000779, -0.00000277, -0.00001484, -0.00001563, 0.00000032, 0.00002395, 0.00003671, 0.00002548, -0.00000511, -0.00003408, -0.00003992, -0.00001812, 0.00001430, 0.00003195, 0.00002155, -0.00000736, -0.00003049, -0.00002908, -0.00000514, 0.00002080, 0.00002732, 0.00000985, -0.00001621, -0.00002908, -0.00001862, 0.00000617, 0.00002579, 0.00002650, 0.00001040, -0.00000825, -0.00001587, -0.00000984, 0.00000171, 0.00000865, 0.00000732, 0.00000170, -0.00000227, -0.00000264, -0.00000199, -0.00000318, -0.00000551, -0.00000541, -0.00000089, 0.00000549, 0.00000845, 0.00000509, -0.00000187, -0.00000614, -0.00000358, 0.00000383, 0.00000956, 0.00000840, 0.00000102, -0.00000668, -0.00000889, -0.00000492, 0.00000058, 0.00000197, -0.00000245, -0.00000936, -0.00001356, -0.00001229, -0.00000678, -0.00000061, 0.00000332, 0.00000421, 0.00000271, -0.00000025, -0.00000374, -0.00000633, -0.00000640, -0.00000360, },
		{ -0.00000785, -0.00001277, -0.00001227, -0.00000656, -0.00000113, -0.00000283, -0.00001383, -0.00002909, -0.00003958, -0.00003886, -0.00002738, -0.00001095, 0.00000472, 0.00001808, 0.00003071, 0.00004271, 0.00004986, 0.00004612, 0.00002999, 0.00000844, -0.00000607, -0.00000414, 0.00001317, 0.00003435, 0.00004537, 0.00003940, 0.00002093, 0.00000170, -0.00000796, -0.00000537, 0.00000439, 0.00001339, 0.00001604, 0.00001119, 0.00000113, -0.00001043, -0.00001964, -0.00002332, -0.00002022, -0.00001208, -0.00000338, 0.00000133, 0.00000060, -0.00000284, -0.00000466, -0.00000289, 0.00000038, 0.00000138, -0.00000113, -0.00000422, -0.00000380, 0.00000062, 0.00000444, 0.00000233, -0.00000545, -0.00001130, -0.00000636, 0.00001019, 0.00002748, 0.00003010, 0.00001139, -0.00001856, -0.00003860, -0.00003289, -0.00000471, 0.00002462, 0.00003216, 0.00001214, -0.00001886, -0.00003525, -0.00002359, 0.00000619, 0.00002918, 0.00002634, 0.00000041, -0.00002668, -0.00003230, -0.00001176, 0.00001842, 0.00003479, 0.00002601, 0.00000096, -0.00002007, -0.00002241, -0.00000768, 0.00000985, 0.00001646, 0.00000957, -0.00000244, -0.00000934, -0.00000777, -0.00000226, 0.00000095, 0.00000015, -0.00000139, -0.00000019, 0.00000316, 0.00000466, 0.00000168, -0.00000355, -0.00000570, -0.00000165, 0.00000584, 0.00001022, 0.00000705, -0.00000166, -0.00000903, -0.00000941, -0.00000338, 0.00000287, 0.00000310, -0.00000348, -0.00001180, -0.00001547, -0.00001200, -0.00000431, 0.00000237, 0.00000473, 0.00000315, 0.00000001, -0.00000279, -0.00000477, -0.00000587, -0.00000551, -0.00000318, 0.00000001, 0.00000124, -0.00000172, },
		{ -0.00001019, -0.00000422, -0.00000369, -0.00001265, -0.00002804, -0.00004128, -0.00004413, -0.00003416, -0.00001536, 0.00000586, 0.00002496, 0.00004023, 0.00005045, 0.00005285, 0.00004476, 0.00002762, 0.00000904, -0.00000041, 0.00000549, 0.00002327, 0.00004118, 0.00004708, 0.00003698, 0.00001745, 0.00000039, -0.00000561, -0.00000057, 0.00000837, 0.00001301, 0.00000962, 0.00000009, -0.00001072, -0.00001850, -0.00002121, -0.00001895, -0.00001323, -0.00000656, -0.00000185, -0.00000100, -0.00000343, -0.00000613, -0.00000594, -0.00000233, 0.00000194, 0.00000326, 0.00000074, -0.00000284, -0.00000379, -0.00000144, 0.00000093, -0.00000026, -0.00000420, -0.00000516, 0.00000212, 0.00001561, 0.00002504, 0.00001946, -0.00000214, -0.00002690, -0.00003641, -0.00002164, 0.00000787, 0.00002982, 0.00002658, 0.00000026, -0.00002794, -0.00003460, -0.00001373, 0.00001768, 0.00003343, 0.00002007, -0.00001149, -0.00003483, -0.00003001, -0.00000018, 0.00003111, 0.00003943, 0.00001923, -0.00001250, -0.00003101, -0.00002390, 0.00000022, 0.00002068, 0.00002235, 0.00000688, -0.00001109, -0.00001756, -0.00001019, 0.00000206, 0.00000855, 0.00000611, -0.00000013, -0.00000336, -0.00000182, 0.00000075, 0.00000034, -0.00000257, -0.00000367, 0.00000001, 0.00000621, 0.00000907, 0.00000492, -0.00000367, -0.00000978, -0.00000843, -0.00000137, 0.00000436, 0.00000278, -0.00000563, -0.00001428, -0.00001626, -0.00001016, -0.00000083, 0.00000517, 0.00000487, 0.00000046, -0.00000373, -0.00000526, -0.00000480, -0.00000414, -0.00000372, -0.00000256, -0.00000038, 0.00000095, -0.00000117, -0.00000698, -0.00001299, -0.00001452, },
		{ -0.00001140, -0.00002719, -0.00004345, -0.00004991, -0.00004115, -0.00001951, 0.00000751, 0.00003217, 0.00004963, 0.00005779, 0.00005582, 0.00004439, 0.00002735, 0.00001223, 0.00000716, 0.00001543, 0.00003201, 0.00004592, 0.00004744, 0.00003496, 0.00001596, 0.00000126, -0.00000280, 0.00000175, 0.00000722, 0.00000673, -0.00000094, -0.00001150, -0.00001920, -0.00002097, -0.00001757, -0.00001195, -0.00000693, -0.00000411, -0.00000386, -0.00000553, -0.00000751, -0.00000768, -0.00000476, 0.00000021, 0.00000420, 0.00000440, 0.00000079, -0.00000356, -0.00000516, -0.00000327, -0.00000049, 0.00000031, -0.00000043, 0.00000105, 0.00000749, 0.00001592, 0.00001822, 0.00000790, -0.00001216, -0.00002910, -0.00002904, -0.00000938, 0.00001674, 0.00002914, 0.00001696, -0.00001109, -0.00003238, -0.00002862, -0.00000174, 0.00002652, 0.00003222, 0.00000963, -0.00002305, -0.00003821, -0.00002190, 0.00001398, 0.00004104, 0.00003767, 0.00000672, -0.00002712, -0.00003780, -0.00001881, 0.00001268, 0.00003108, 0.00002340, -0.00000161, -0.00002256, -0.00002389, -0.00000749, 0.00001123, 0.00001767, 0.00000956, -0.00000351, -0.00001031, -0.00000760, -0.00000097, 0.00000235, 0.00000070, -0.00000166, -0.00000025, 0.00000427, 0.00000679, 0.00000346, -0.00000380, -0.00000870, -0.00000682, -0.00000022, 0.00000416, 0.00000105, -0.00000798, -0.00001558, -0.00001523, -0.00000698, 0.00000266, 0.00000660, 0.00000305, -0.00000364, -0.00000757, -0.00000655, -0.00000314, -0.00000118, -0.00000174, -0.00000269, -0.00000184, -0.00000010, -0.00000081, -0.00000591, -0.00001298, -0.00001670, -0.00001386, -0.00000729, -0.00000444, },
		{ -0.00005635, -0.00004881, -0.00002398, 0.00000928, 0.00003977, 0.00005950, 0.00006568, 0.00005968, 0.00004545, 0.00002891, 0.00001706, 0.00001555, 0.00002508, 0.00003976, 0.00004984, 0.00004793, 0.00003404, 0.00001562, 0.00000193, -0.00000256, -0.00000073, 0.00000053, -0.00000359, -0.00001218, -0.00002002, -0.00002221, -0.00001802, -0.00001087, -0.00000526, -0.00000363, -0.00000543, -0.00000841, -0.00001030, -0.00000978, -0.00000657, -0.00000153, 0.00000334, 0.00000555, 0.00000372, -0.00000100, -0.00000533, -0.00000623, -0.00000342, 0.00000058, 0.00000300, 0.00000374, 0.00000515, 0.00000877, 0.00001208, 0.00000959, -0.00000167, -0.00001714, -0.00002593, -0.00001921, 0.00000089, 0.00002042, 0.00002352, 0.00000602, -0.00001923, -0.00003140, -0.00001891, 0.00000951, 0.00003064, 0.00002580, -0.00000257, -0.00003140, -0.00003556, -0.00000933, 0.00002756, 0.00004542, 0.00002905, -0.00000924, -0.00003935, -0.00003798, -0.00000743, 0.00002685, 0.00003759, 0.00001778, -0.00001488, -0.00003372, -0.00002521, 0.00000142, 0.00002367, 0.00002518, 0.00000787, -0.00001205, -0.00001927, -0.00001144, 0.00000142, 0.00000775, 0.00000458, -0.00000175, -0.00000353, 0.00000055, 0.00000504, 0.00000426, -0.00000153, -0.00000665, -0.00000614, -0.00000115, 0.00000215, -0.00000106, -0.00000904, -0.00001475, -0.00001252, -0.00000361, 0.00000472, 0.00000578, -0.00000055, -0.00000805, -0.00001014, -0.00000578, 0.00000018, 0.00000216, -0.00000071, -0.00000435, -0.00000445, -0.00000136, 0.00000014, -0.00000409, -0.00001267, -0.00001917, -0.00001807, -0.00001063, -0.00000500, -0.00000972, -0.00002610, -0.00004589, },
		{ 0.00001145, 0.00004852, 0.00007076, 0.00007487, 0.00006466, 0.00004757, 0.00003150, 0.00002281, 0.00002455, 0.00003500, 0.00004765, 0.00005407, 0.00004884, 0.00003317, 0.00001410, -0.00000045, -0.00000706, -0.00000840, -0.00000988, -0.00001454, -0.00002052, -0.00002321, -0.00001970, -0.00001166, -0.00000406, -0.00000133, -0.00000418, -0.00000961, -0.00001349, -0.00001332, -0.00000924, -0.00000319, 0.00000238, 0.00000545, 0.00000492, 0.00000119, -0.00000366, -0.00000663, -0.00000571, -0.00000147, 0.00000335, 0.00000607, 0.00000637, 0.00000606, 0.00000648, 0.00000615, 0.00000182, -0.00000745, -0.00001720, -0.00001949, -0.00000980, 0.00000720, 0.00001910, 0.00001519, -0.00000346, -0.00002270, -0.00002581, -0.00000805, 0.00001752, 0.00002942, 0.00001589, -0.00001360, -0.00003450, -0.00002731, 0.00000494, 0.00003732, 0.00004279, 0.00001502, -0.00002515, -0.00004598, -0.00003080, 0.00000799, 0.00003897, 0.00003746, 0.00000561, -0.00003005, -0.00004089, -0.00001944, 0.00001565, 0.00003621, 0.00002773, -0.00000032, -0.00002441, -0.00002725, -0.00001069, 0.00000858, 0.00001496, 0.00000654, -0.00000570, -0.00000972, -0.00000322, 0.00000586, 0.00000831, 0.00000242, -0.00000543, -0.00000805, -0.00000457, -0.00000062, -0.00000178, -0.00000750, -0.00001175, -0.00000938, -0.00000177, 0.00000418, 0.00000279, -0.00000471, -0.00001117, -0.00001036, -0.00000296, 0.00000413, 0.00000450, -0.00000159, -0.00000761, -0.00000748, -0.00000182, 0.00000235, -0.00000153, -0.00001261, -0.00002261, -0.00002315, -0.00001407, -0.00000480, -0.00000704, -0.00002460, -0.00004892, -0.00006414, -0.00005782, -0.00002907, },
		{ 0.00008531, 0.00007015, 0.00004996, 0.00003468, 0.00002966, 0.00003502, 0.00004638, 0.00005656, 0.00005848, 0.00004875, 0.00002980, 0.00000854, -0.00000791, -0.00001651, -0.00001934, -0.00002062, -0.00002254, -0.00002363, -0.00002090, -0.00001359, -0.00000486, 0.00000025, -0.00000126, -0.00000785, -0.00001467, -0.00001703, -0.00001348, -0.00000617, 0.00000109, 0.00000531, 0.00000545, 0.00000229, -0.00000224, -0.00000576, -0.00000627, -0.00000322, 0.00000191, 0.00000632, 0.00000781, 0.00000643, 0.00000404, 0.00000223, 0.00000043, -0.00000326, -0.00000912, -0.00001383, -0.00001234, -0.00000297, 0.00000909, 0.00001434, 0.00000673, -0.00000954, -0.00002151, -0.00001763, 0.00000137, 0.00002094, 0.00002377, 0.00000515, -0.00002093, -0.00003183, -0.00001546, 0.00001764, 0.00004098, 0.00003356, -0.00000144, -0.00003740, -0.00004520, -0.00001756, 0.00002385, 0.00004555, 0.00002969, -0.00001092, -0.00004320, -0.00004103, -0.00000650, 0.00003234, 0.00004494, 0.00002293, -0.00001444, -0.00003755, -0.00003084, -0.00000382, 0.00001936, 0.00002116, 0.00000412, -0.00001398, -0.00001713, -0.00000450, 0.00001090, 0.00001526, 0.00000609, -0.00000724, -0.00001346, -0.00000959, -0.00000211, 0.00000072, -0.00000296, -0.00000781, -0.00000786, -0.00000306, 0.00000089, -0.00000106, -0.00000746, -0.00001154, -0.00000808, 0.00000077, 0.00000704, 0.00000461, -0.00000451, -0.00001160, -0.00000967, -0.00000062, 0.00000582, 0.00000104, -0.00001374, -0.00002751, -0.00002886, -0.00001681, -0.00000306, -0.00000311, -0.00002315, -0.00005348, -0.00007416, -0.00006851, -0.00003435, 0.00001496, 0.00005943, 0.00008404, },
		{ 0.00003905, 0.00003897, 0.00004843, 0.00006000, 0.00006597, 0.00006112, 0.00004472, 0.00002096, -0.00000293, -0.00002037, -0.00002878, -0.00003005, -0.00002805, -0.00002522, -0.00002125, -0.00001491, -0.00000682, -0.00000031, 0.00000081, -0.00000444, -0.00001272, -0.00001848, -0.00001777, -0.00001092, -0.00000191, 0.00000458, 0.00000610, 0.00000324, -0.00000142, -0.00000514, -0.00000611, -0.00000389, 0.00000061, 0.00000533, 0.00000796, 0.00000728, 0.00000404, 0.00000038, -0.00000202, -0.00000336, -0.00000508, -0.00000766, -0.00000915, -0.00000659, 0.00000038, 0.00000749, 0.00000833, 0.00000025, -0.00001160, -0.00001701, -0.00000928, 0.00000749, 0.00001983, 0.00001574, -0.00000379, -0.00002326, -0.00002451, -0.00000298, 0.00002596, 0.00003791, 0.00001990, -0.00001668, -0.00004330, -0.00003719, -0.00000131, 0.00003625, 0.00004450, 0.00001549, -0.00002804, -0.00005053, -0.00003288, 0.00001150, 0.00004736, 0.00004647, 0.00001034, -0.00003179, -0.00004755, -0.00002751, 0.00000880, 0.00003094, 0.00002316, -0.00000408, -0.00002536, -0.00002293, -0.00000089, 0.00002042, 0.00002318, 0.00000675, -0.00001371, -0.00002186, -0.00001394, 0.00000012, 0.00000738, 0.00000347, -0.00000537, -0.00001013, -0.00000802, -0.00000383, -0.00000343, -0.00000693, -0.00000880, -0.00000456, 0.00000337, 0.00000733, 0.00000220, -0.00000844, -0.00001473, -0.00000988, 0.00000231, 0.00000964, 0.00000241, -0.00001674, -0.00003361, -0.00003421, -0.00001779, 0.00000066, 0.00000157, -0.00002291, -0.00006071, -0.00008684, -0.00008034, -0.00003863, 0.00002105, 0.00007315, 0.00009909, 0.00009611, 0.00007524, 0.00005232, },
		{ 0.00007504, 0.00007331, 0.00005842, 0.00003341, 0.00000478, -0.00001997, -0.00003536, -0.00004005, -0.00003674, -0.00002979, -0.00002235, -0.00001524, -0.00000824, -0.00000216, 0.00000062, -0.00000192, -0.00000903, -0.00001661, -0.00001964, -0.00001573, -0.00000692, 0.00000181, 0.00000601, 0.00000455, -0.00000027, -0.00000477, -0.00000631, -0.00000436, -0.00000010, 0.00000447, 0.00000741, 0.00000744, 0.00000459, 0.00000031, -0.00000325, -0.00000475, -0.00000457, -0.00000428, -0.00000478, -0.00000509, -0.00000335, 0.00000066, 0.00000411, 0.00000310, -0.00000321, -0.00001033, -0.00001120, -0.00000275, 0.00000974, 0.00001557, 0.00000779, -0.00000910, -0.00002080, -0.00001485, 0.00000720, 0.00002841, 0.00002927, 0.00000506, -0.00002749, -0.00004193, -0.00002410, 0.00001408, 0.00004220, 0.00003584, -0.00000203, -0.00004157, -0.00004962, -0.00001749, 0.00003064, 0.00005655, 0.00003912, -0.00000818, -0.00004829, -0.00005084, -0.00001654, 0.00002473, 0.00003956, 0.00001853, -0.00001757, -0.00003692, -0.00002392, 0.00000882, 0.00003292, 0.00002896, 0.00000192, -0.00002498, -0.00003101, -0.00001459, 0.00000782, 0.00001752, 0.00000915, -0.00000724, -0.00001737, -0.00001550, -0.00000726, -0.00000180, -0.00000226, -0.00000419, -0.00000225, 0.00000274, 0.00000443, -0.00000158, -0.00001142, -0.00001545, -0.00000786, 0.00000591, 0.00001226, 0.00000158, -0.00002144, -0.00003972, -0.00003779, -0.00001632, 0.00000584, 0.00000558, -0.00002535, -0.00007125, -0.00010161, -0.00009186, -0.00004039, 0.00003044, 0.00008921, 0.00011467, 0.00010609, 0.00007964, 0.00005563, 0.00004653, 0.00005268, 0.00006569, },
		{ 0.00001274, -0.00001816, -0.00003958, -0.00004855, -0.00004625, -0.00003695, -0.00002569, -0.00001592, -0.00000869, -0.00000371, -0.00000105, -0.00000168, -0.00000615, -0.00001285, -0.00001804, -0.00001804, -0.00001207, -0.00000319, 0.00000365, 0.00000509, 0.00000150, -0.00000366, -0.00000653, -0.00000535, -0.00000107, 0.00000383, 0.00000703, 0.00000734, 0.00000486, 0.00000073, -0.00000328, -0.00000551, -0.00000533, -0.00000364, -0.00000222, -0.00000213, -0.00000277, -0.00000254, -0.00000086, 0.00000074, -0.00000014, -0.00000383, -0.00000723, -0.00000599, 0.00000094, 0.00000881, 0.00001019, 0.00000193, -0.00001028, -0.00001513, -0.00000559, 0.00001303, 0.00002520, 0.00001762, -0.00000766, -0.00003207, -0.00003435, -0.00000932, 0.00002512, 0.00004056, 0.00002171, -0.00001872, -0.00004815, -0.00004026, 0.00000194, 0.00004639, 0.00005703, 0.00002361, -0.00002881, -0.00005959, -0.00004546, 0.00000049, 0.00004006, 0.00004178, 0.00000663, -0.00003384, -0.00004492, -0.00001782, 0.00002390, 0.00004521, 0.00002928, -0.00000954, -0.00003923, -0.00003743, -0.00000883, 0.00002105, 0.00002853, 0.00001058, -0.00001550, -0.00002891, -0.00002257, -0.00000616, 0.00000538, 0.00000551, -0.00000053, -0.00000393, -0.00000219, -0.00000054, -0.00000431, -0.00001133, -0.00001321, -0.00000480, 0.00000818, 0.00001223, -0.00000143, -0.00002644, -0.00004404, -0.00003857, -0.00001285, 0.00001085, 0.00000708, -0.00003140, -0.00008455, -0.00011677, -0.00010124, -0.00003877, 0.00004253, 0.00010601, 0.00012917, 0.00011482, 0.00008464, 0.00006245, 0.00005973, 0.00007205, 0.00008581, 0.00008829, 0.00007422, 0.00004645, },
		{ -0.00005534, -0.00004471, -0.00003065, -0.00001797, -0.00000924, -0.00000457, -0.00000281, -0.00000312, -0.00000544, -0.00000967, -0.00001438, -0.00001685, -0.00001479, -0.00000850, -0.00000113, 0.00000323, 0.00000256, -0.00000167, -0.00000567, -0.00000621, -0.00000275, 0.00000253, 0.00000659, 0.00000751, 0.00000525, 0.00000116, -0.00000294, -0.00000554, -0.00000583, -0.00000410, -0.00000169, -0.00000029, -0.00000075, -0.00000232, -0.00000330, -0.00000273, -0.00000144, -0.00000120, -0.00000262, -0.00000394, -0.00000256, 0.00000192, 0.00000617, 0.00000558, -0.00000088, -0.00000815, -0.00000850, 0.00000103, 0.00001392, 0.00001800, 0.00000600, -0.00001585, -0.00003040, -0.00002318, 0.00000357, 0.00002972, 0.00003231, 0.00000571, -0.00003074, -0.00004636, -0.00002458, 0.00002098, 0.00005491, 0.00004815, 0.00000297, -0.00004688, -0.00006249, -0.00003177, 0.00001999, 0.00005057, 0.00003568, -0.00001088, -0.00004897, -0.00004599, -0.00000419, 0.00004154, 0.00005330, 0.00002179, -0.00002665, -0.00005288, -0.00003746, 0.00000451, 0.00003768, 0.00003644, 0.00000462, -0.00003035, -0.00004189, -0.00002523, 0.00000204, 0.00001763, 0.00001331, -0.00000144, -0.00001147, -0.00001057, -0.00000475, -0.00000314, -0.00000704, -0.00000918, -0.00000319, 0.00000705, 0.00000921, -0.00000510, -0.00002942, -0.00004502, -0.00003682, -0.00000926, 0.00001335, 0.00000474, -0.00004055, -0.00009858, -0.00013002, -0.00010730, -0.00003444, 0.00005524, 0.00012135, 0.00014188, 0.00012378, 0.00009347, 0.00007627, 0.00008055, 0.00009614, 0.00010500, 0.00009443, 0.00006384, 0.00002264, -0.00001659, -0.00004432, -0.00005677, },
		{ -0.00001035, -0.00000554, -0.00000444, -0.00000514, -0.00000649, -0.00000840, -0.00001096, -0.00001341, -0.00001406, -0.00001155, -0.00000640, -0.00000118, 0.00000120, -0.00000026, -0.00000369, -0.00000576, -0.00000425, 0.00000030, 0.00000514, 0.00000739, 0.00000599, 0.00000202, -0.00000237, -0.00000531, -0.00000596, -0.00000453, -0.00000204, 0.00000009, 0.00000058, -0.00000083, -0.00000308, -0.00000443, -0.00000387, -0.00000207, -0.00000073, -0.00000088, -0.00000163, -0.00000114, 0.00000119, 0.00000347, 0.00000292, -0.00000086, -0.00000440, -0.00000305, 0.00000389, 0.00001074, 0.00000933, -0.00000300, -0.00001867, -0.00002416, -0.00001165, 0.00001191, 0.00002775, 0.00002020, -0.00000825, -0.00003565, -0.00003707, -0.00000639, 0.00003529, 0.00005428, 0.00003180, -0.00001832, -0.00005822, -0.00005595, -0.00001249, 0.00003743, 0.00005308, 0.00002162, -0.00003044, -0.00005876, -0.00003833, 0.00001506, 0.00005753, 0.00005367, 0.00000623, -0.00004625, -0.00006154, -0.00002859, 0.00002423, 0.00005360, 0.00003700, -0.00001017, -0.00004946, -0.00005178, -0.00001975, 0.00001782, 0.00003201, 0.00001675, -0.00000988, -0.00002466, -0.00001931, -0.00000444, 0.00000397, 0.00000056, -0.00000630, -0.00000595, 0.00000154, 0.00000475, -0.00000652, -0.00002812, -0.00004253, -0.00003461, -0.00000850, 0.00001140, -0.00000113, -0.00005047, -0.00011043, -0.00013969, -0.00011071, -0.00003001, 0.00006561, 0.00013396, 0.00015429, 0.00013693, 0.00011072, 0.00009993, 0.00010838, 0.00012069, 0.00011686, 0.00008742, 0.00003883, -0.00001241, -0.00005048, -0.00006783, -0.00006616, -0.00005269, -0.00003546, -0.00002038, },
		{ -0.00000864, -0.00000880, -0.00000896, -0.00000978, -0.00001099, -0.00001136, -0.00000972, -0.00000626, -0.00000273, -0.00000116, -0.00000210, -0.00000398, -0.00000429, -0.00000174, 0.00000257, 0.00000598, 0.00000629, 0.00000332, -0.00000112, -0.00000467, -0.00000592, -0.00000485, -0.00000251, -0.00000021, 0.00000089, 0.00000023, -0.00000186, -0.00000413, -0.00000503, -0.00000389, -0.00000153, 0.00000027, 0.00000036, -0.00000070, -0.00000115, 0.00000002, 0.00000187, 0.00000239, 0.00000087, -0.00000085, -0.00000012, 0.00000319, 0.00000533, 0.00000164, -0.00000799, -0.00001688, -0.00001597, -0.00000259, 0.00001467, 0.00002092, 0.00000772, -0.00001716, -0.00003314, -0.00002314, 0.00000995, 0.00004193, 0.00004502, 0.00001175, -0.00003562, -0.00006047, -0.00004146, 0.00000806, 0.00004867, 0.00004651, 0.00000231, -0.00004750, -0.00006011, -0.00002254, 0.00003605, 0.00006747, 0.00004460, -0.00001519, -0.00006365, -0.00006140, -0.00001066, 0.00004667, 0.00006378, 0.00002734, -0.00003253, -0.00006819, -0.00005368, -0.00000424, 0.00003888, 0.00004356, 0.00001156, -0.00002675, -0.00004057, -0.00002380, 0.00000364, 0.00001684, 0.00000828, -0.00000850, -0.00001507, -0.00000712, 0.00000228, -0.00000241, -0.00002165, -0.00003865, -0.00003561, -0.00001344, 0.00000481, -0.00000794, -0.00005739, -0.00011759, -0.00014620, -0.00011463, -0.00002949, 0.00007141, 0.00014510, 0.00017104, 0.00016014, 0.00014054, 0.00013351, 0.00013866, 0.00013817, 0.00011387, 0.00006274, -0.00000060, -0.00005413, -0.00008205, -0.00008224, -0.00006404, -0.00004053, -0.00002142, -0.00001040, -0.00000643, -0.00000652, -0.00000780, },
		{ -0.00000732, -0.00000888, -0.00001016, -0.00000974, -0.00000745, -0.00000452, -0.00000263, -0.00000242, -0.00000288, -0.00000230, 0.00000010, 0.00000327, 0.00000509, 0.00000409, 0.00000068, -0.00000314, -0.00000533, -0.00000507, -0.00000306, -0.00000073, 0.00000067, 0.00000057, -0.00000091, -0.00000304, -0.00000467, -0.00000475, -0.00000305, -0.00000059, 0.00000095, 0.00000065, -0.00000081, -0.00000161, -0.00000055, 0.00000175, 0.00000331, 0.00000286, 0.00000113, 0.00000001, 0.00000016, -0.00000022, -0.00000344, -0.00000899, -0.00001235, -0.00000848, 0.00000240, 0.00001257, 0.00001205, -0.00000195, -0.00001975, -0.00002496, -0.00000846, 0.00002091, 0.00004012, 0.00002984, -0.00000742, -0.00004551, -0.00005369, -0.00002255, 0.00002530, 0.00005125, 0.00003227, -0.00001797, -0.00005799, -0.00005210, -0.00000157, 0.00005401, 0.00006820, 0.00002689, -0.00003814, -0.00007407, -0.00005067, 0.00001374, 0.00006652, 0.00006388, 0.00000715, -0.00005836, -0.00008079, -0.00004406, 0.00002021, 0.00006021, 0.00004673, -0.00000472, -0.00004987, -0.00005401, -0.00001936, 0.00002059, 0.00003250, 0.00001123, -0.00001915, -0.00003016, -0.00001521, 0.00000626, 0.00000931, -0.00001170, -0.00003772, -0.00004393, -0.00002519, -0.00000380, -0.00001098, -0.00005760, -0.00011981, -0.00015306, -0.00012429, -0.00003650, 0.00007320, 0.00015982, 0.00019945, 0.00019935, 0.00018421, 0.00017249, 0.00016273, 0.00013926, 0.00008992, 0.00001984, -0.00004958, -0.00009451, -0.00010353, -0.00008290, -0.00005010, -0.00002219, -0.00000753, -0.00000493, -0.00000822, -0.00001160, -0.00001234, -0.00001068, -0.00000831, -0.00000694, },
		{ -0.00001095, -0.00000869, -0.00000537, -0.00000269, -0.00000155, -0.00000141, -0.00000094, 0.00000057, 0.00000246, 0.00000322, 0.00000189, -0.00000098, -0.00000367, -0.00000463, -0.00000356, -0.00000150, 0.00000015, 0.00000050, -0.00000046, -0.00000214, -0.00000373, -0.00000443, -0.00000377, -0.00000194, 0.00000004, 0.00000092, 0.00000019, -0.00000126, -0.00000169, -0.00000009, 0.00000272, 0.00000457, 0.00000388, 0.00000106, -0.00000192, -0.00000357, -0.00000427, -0.00000544, -0.00000718, -0.00000732, -0.00000350, 0.00000327, 0.00000765, 0.00000407, -0.00000696, -0.00001672, -0.00001421, 0.00000332, 0.00002472, 0.00003138, 0.00001278, -0.00002172, -0.00004672, -0.00004024, -0.00000369, 0.00003573, 0.00004510, 0.00001380, -0.00003469, -0.00005953, -0.00003633, 0.00002003, 0.00006468, 0.00005890, 0.00000389, -0.00005739, -0.00007415, -0.00003040, 0.00003968, 0.00007834, 0.00005182, -0.00002067, -0.00008161, -0.00008214, -0.00002227, 0.00004924, 0.00007533, 0.00003727, -0.00003131, -0.00007403, -0.00005890, -0.00000312, 0.00004429, 0.00004545, 0.00000445, -0.00003931, -0.00004773, -0.00001718, 0.00002040, 0.00002796, -0.00000294, -0.00004526, -0.00006219, -0.00004168, -0.00000892, -0.00000490, -0.00004949, -0.00012048, -0.00016678, -0.00014525, -0.00005193, 0.00007600, 0.00018695, 0.00024772, 0.00025772, 0.00023782, 0.00020722, 0.00016922, 0.00011571, 0.00004313, -0.00003640, -0.00009897, -0.00012418, -0.00010892, -0.00006893, -0.00002794, -0.00000339, 0.00000114, -0.00000644, -0.00001514, -0.00001816, -0.00001505, -0.00000933, -0.00000487, -0.00000366, -0.00000545, -0.00000855, -0.00001087, },
		{ -0.00000176, -0.00000054, -0.00000066, -0.00000086, -0.00000026, 0.00000087, 0.00000142, 0.00000058, -0.00000132, -0.00000299, -0.00000332, -0.00000225, -0.00000075, 0.00000005, -0.00000033, -0.00000155, -0.00000288, -0.00000366, -0.00000358, -0.00000262, -0.00000116, 0.00000005, 0.00000030, -0.00000049, -0.00000137, -0.00000098, 0.00000119, 0.00000397, 0.00000518, 0.00000339, -0.00000071, -0.00000471, -0.00000655, -0.00000612, -0.00000496, -0.00000427, -0.00000356, -0.00000160, 0.00000126, 0.00000234, -0.00000093, -0.00000715, -0.00001026, -0.00000427, 0.00000981, 0.00002173, 0.00001886, -0.00000212, -0.00002884, -0.00004032, -0.00002397, 0.00001109, 0.00003790, 0.00003245, -0.00000450, -0.00004440, -0.00005220, -0.00001647, 0.00003754, 0.00006573, 0.00004147, -0.00001944, -0.00006841, -0.00006314, -0.00000443, 0.00006124, 0.00007830, 0.00002874, -0.00005052, -0.00009597, -0.00006958, 0.00000870, 0.00007611, 0.00007825, 0.00001399, -0.00006378, -0.00009220, -0.00005039, 0.00002438, 0.00006929, 0.00004917, -0.00001514, -0.00006652, -0.00006157, -0.00000743, 0.00004575, 0.00004919, -0.00000219, -0.00006594, -0.00008963, -0.00005707, -0.00000340, 0.00001397, -0.00003554, -0.00012705, -0.00019521, -0.00018068, -0.00007161, 0.00008977, 0.00023730, 0.00032173, 0.00033277, 0.00029115, 0.00022416, 0.00014716, 0.00006366, -0.00002207, -0.00009577, -0.00013746, -0.00013532, -0.00009639, -0.00004407, -0.00000419, 0.00000971, 0.00000142, -0.00001447, -0.00002418, -0.00002265, -0.00001363, -0.00000436, 0.00000002, -0.00000126, -0.00000581, -0.00001040, -0.00001270, -0.00001193, -0.00000874, -0.00000478, },
		{ -0.00000188, -0.00000158, -0.00000047, 0.00000050, 0.00000043, -0.00000064, -0.00000182, -0.00000221, -0.00000166, -0.00000087, -0.00000065, -0.00000124, -0.00000220, -0.00000291, -0.00000303, -0.00000258, -0.00000180, -0.00000098, -0.00000047, -0.00000048, -0.00000081, -0.00000078, 0.00000040, 0.00000266, 0.00000468, 0.00000457, 0.00000151, -0.00000328, -0.00000713, -0.00000797, -0.00000596, -0.00000323, -0.00000180, -0.00000188, -0.00000217, -0.00000175, -0.00000136, -0.00000227, -0.00000390, -0.00000322, 0.00000228, 0.00001047, 0.00001410, 0.00000622, -0.00001195, -0.00002866, -0.00002917, -0.00000892, 0.00001929, 0.00003268, 0.00001702, -0.00001864, -0.00004572, -0.00003838, 0.00000288, 0.00004744, 0.00005739, 0.00001996, -0.00003797, -0.00006875, -0.00004329, 0.00002161, 0.00007331, 0.00006557, -0.00000099, -0.00007561, -0.00009686, -0.00004411, 0.00004272, 0.00009355, 0.00006554, -0.00002030, -0.00009489, -0.00009761, -0.00002676, 0.00005864, 0.00008804, 0.00003834, -0.00004693, -0.00009465, -0.00006432, 0.00001744, 0.00007924, 0.00006544, -0.00001638, -0.00010118, -0.00012097, -0.00006265, 0.00001909, 0.00004509, -0.00002334, -0.00014964, -0.00024461, -0.00022841, -0.00008520, 0.00012800, 0.00032030, 0.00042156, 0.00041453, 0.00032855, 0.00020900, 0.00008978, -0.00001366, -0.00009466, -0.00014483, -0.00015492, -0.00012465, -0.00006954, -0.00001590, 0.00001365, 0.00001292, -0.00000654, -0.00002567, -0.00003116, -0.00002225, -0.00000770, 0.00000252, 0.00000388, -0.00000163, -0.00000885, -0.00001351, -0.00001408, -0.00001136, -0.00000716, -0.00000328, -0.00000096, -0.00000050, -0.00000122, },
		{ 0.00000063, 0.00000106, 0.00000033, -0.00000090, -0.00000172, -0.00000179, -0.00000150, -0.00000142, -0.00000175, -0.00000222, -0.00000242, -0.00000223, -0.00000186, -0.00000154, -0.00000132, -0.00000115, -0.00000093, -0.00000049, 0.00000043, 0.00000197, 0.00000366, 0.00000434, 0.00000278, -0.00000111, -0.00000569, -0.00000843, -0.00000780, -0.00000463, -0.00000150, -0.00000066, -0.00000210, -0.00000386, -0.00000402, -0.00000248, -0.00000062, 0.00000063, 0.00000194, 0.00000431, 0.00000652, 0.00000486, -0.00000345, -0.00001564, -0.00002293, -0.00001687, 0.00000184, 0.00002052, 0.00002274, 0.00000277, -0.00002617, -0.00003964, -0.00002196, 0.00001762, 0.00004844, 0.00004235, -0.00000065, -0.00004801, -0.00005886, -0.00001924, 0.00004196, 0.00007315, 0.00004271, -0.00003094, -0.00008993, -0.00008305, -0.00001044, 0.00007202, 0.00009602, 0.00003777, -0.00005907, -0.00011649, -0.00008604, 0.00000932, 0.00009196, 0.00009311, 0.00001101, -0.00008592, -0.00011521, -0.00004995, 0.00005660, 0.00011349, 0.00006770, -0.00004988, -0.00014741, -0.00014671, -0.00004924, 0.00006102, 0.00008198, -0.00002474, -0.00019815, -0.00031614, -0.00027903, -0.00007707, 0.00020424, 0.00043986, 0.00053876, 0.00048584, 0.00033206, 0.00015109, -0.00000193, -0.00010503, -0.00015895, -0.00017051, -0.00014592, -0.00009519, -0.00003684, 0.00000650, 0.00002001, 0.00000581, -0.00001910, -0.00003505, -0.00003252, -0.00001627, 0.00000088, 0.00000839, 0.00000443, -0.00000532, -0.00001348, -0.00001602, -0.00001347, -0.00000877, -0.00000459, -0.00000215, -0.00000147, -0.00000196, -0.00000280, -0.00000312, -0.00000238, -0.00000083, },
		{ -0.00000057, -0.00000181, -0.00000229, -0.00000214, -0.00000185, -0.00000175, -0.00000177, -0.00000175, -0.00000166, -0.00000165, -0.00000177, -0.00000185, -0.00000158, -0.00000079, 0.00000043, 0.00000186, 0.00000315, 0.00000375, 0.00000293, 0.00000027, -0.00000369, -0.00000725, -0.00000844, -0.00000658, -0.00000308, -0.00000059, -0.00000090, -0.00000332, -0.00000529, -0.00000456, -0.00000125, 0.00000244, 0.00000426, 0.00000385, 0.00000230, 0.00000026, -0.00000309, -0.00000850, -0.00001395, -0.00001453, -0.00000661, 0.00000697, 0.00001632, 0.00001154, -0.00000729, -0.00002679, -0.00002895, -0.00000700, 0.00002532, 0.00004172, 0.00002484, -0.00001601, -0.00004831, -0.00004186, 0.00000343, 0.00005251, 0.00006131, 0.00001502, -0.00005443, -0.00009016, -0.00005721, 0.00002412, 0.00008953, 0.00008158, -0.00000022, -0.00009377, -0.00012208, -0.00005758, 0.00005109, 0.00011545, 0.00007972, -0.00003001, -0.00012354, -0.00011977, -0.00001626, 0.00010413, 0.00013827, 0.00004847, -0.00010196, -0.00019561, -0.00015527, -0.00001060, 0.00011852, 0.00011230, -0.00005312, -0.00027815, -0.00040298, -0.00031596, -0.00002952, 0.00032746, 0.00059075, 0.00065582, 0.00052511, 0.00028617, 0.00004778, -0.00011755, -0.00019271, -0.00019842, -0.00016432, -0.00011213, -0.00005605, -0.00000924, 0.00001565, 0.00001346, -0.00000782, -0.00003038, -0.00003808, -0.00002702, -0.00000666, 0.00000867, 0.00001047, 0.00000087, -0.00001118, -0.00001754, -0.00001612, -0.00001031, -0.00000493, -0.00000254, -0.00000275, -0.00000384, -0.00000452, -0.00000437, -0.00000347, -0.00000196, -0.00000015, 0.00000131, 0.00000173, 0.00000091, },
		{ -0.00000249, -0.00000189, -0.00000138, -0.00000118, -0.00000124, -0.00000148, -0.00000187, -0.00000223, -0.00000222, -0.00000151, -0.00000009, 0.00000164, 0.00000304, 0.00000356, 0.00000284, 0.00000078, -0.00000234, -0.00000563, -0.00000772, -0.00000748, -0.00000503, -0.00000202, -0.00000069, -0.00000195, -0.00000448, -0.00000553, -0.00000332, 0.00000125, 0.00000513, 0.00000560, 0.00000246, -0.00000213, -0.00000580, -0.00000786, -0.00000892, -0.00000894, -0.00000639, -0.00000027, 0.00000697, 0.00000935, 0.00000219, -0.00001187, -0.00002207, -0.00001718, 0.00000345, 0.00002579, 0.00003035, 0.00000921, -0.00002362, -0.00004036, -0.00002254, 0.00001985, 0.00005193, 0.00004144, -0.00001102, -0.00006657, -0.00007671, -0.00002530, 0.00005183, 0.00009099, 0.00005273, -0.00004048, -0.00011620, -0.00010891, -0.00001706, 0.00008926, 0.00012176, 0.00004773, -0.00007727, -0.00014976, -0.00010286, 0.00003348, 0.00014982, 0.00014328, 0.00000475, -0.00016563, -0.00023293, -0.00013650, 0.00005336, 0.00018046, 0.00012007, -0.00011909, -0.00038722, -0.00048947, -0.00031815, 0.00007225, 0.00049749, 0.00075694, 0.00074836, 0.00051141, 0.00018305, -0.00009285, -0.00023808, -0.00025640, -0.00020051, -0.00012560, -0.00006282, -0.00002016, 0.00000336, 0.00000857, -0.00000236, -0.00002112, -0.00003453, -0.00003271, -0.00001660, 0.00000257, 0.00001216, 0.00000750, -0.00000559, -0.00001663, -0.00001881, -0.00001304, -0.00000546, -0.00000161, -0.00000259, -0.00000568, -0.00000755, -0.00000692, -0.00000467, -0.00000223, -0.00000037, 0.00000093, 0.00000172, 0.00000176, 0.00000086, -0.00000067, -0.00000207, -0.00000269, },
		{ -0.00000070, -0.00000100, -0.00000165, -0.00000232, -0.00000264, -0.00000224, -0.00000095, 0.00000095, 0.00000275, 0.00000362, 0.00000308, 0.00000116, -0.00000163, -0.00000453, -0.00000668, -0.00000729, -0.00000608, -0.00000370, -0.00000168, -0.00000144, -0.00000305, -0.00000481, -0.00000443, -0.00000103, 0.00000362, 0.00000618, 0.00000433, -0.00000116, -0.00000695, -0.00000979, -0.00000891, -0.00000599, -0.00000306, -0.00000061, 0.00000182, 0.00000362, 0.00000240, -0.00000349, -0.00001148, -0.00001473, -0.00000747, 0.00000816, 0.00002082, 0.00001812, -0.00000140, -0.00002346, -0.00002768, -0.00000575, 0.00002732, 0.00004200, 0.00001906, -0.00002973, -0.00006570, -0.00005360, 0.00000519, 0.00006671, 0.00007632, 0.00001585, -0.00007416, -0.00012107, -0.00007912, 0.00002705, 0.00011470, 0.00010754, 0.00000226, -0.00012044, -0.00015619, -0.00006425, 0.00009056, 0.00018164, 0.00012156, -0.00005992, -0.00022839, -0.00024596, -0.00008548, 0.00013465, 0.00022995, 0.00008991, -0.00022601, -0.00051282, -0.00055287, -0.00026502, 0.00023472, 0.00070208, 0.00091290, 0.00079007, 0.00043035, 0.00002649, -0.00025192, -0.00033962, -0.00027857, -0.00016085, -0.00006275, -0.00001272, -0.00000041, -0.00000439, -0.00001179, -0.00001965, -0.00002664, -0.00002827, -0.00002055, -0.00000585, 0.00000700, 0.00000948, 0.00000066, -0.00001198, -0.00001891, -0.00001616, -0.00000768, -0.00000100, -0.00000074, -0.00000554, -0.00001035, -0.00001112, -0.00000770, -0.00000290, 0.00000044, 0.00000155, 0.00000140, 0.00000115, 0.00000101, 0.00000047, -0.00000070, -0.00000204, -0.00000276, -0.00000254, -0.00000172, -0.00000095, },
		{ -0.00000281, -0.00000285, -0.00000188, -0.00000004, 0.00000204, 0.00000347, 0.00000346, 0.00000182, -0.00000094, -0.00000386, -0.00000602, -0.00000686, -0.00000628, -0.00000466, -0.00000286, -0.00000188, -0.00000230, -0.00000357, -0.00000412, -0.00000247, 0.00000124, 0.00000477, 0.00000515, 0.00000112, -0.00000541, -0.00001045, -0.00001091, -0.00000694, -0.00000165, 0.00000166, 0.00000200, 0.00000068, -0.00000081, -0.00000251, -0.00000517, -0.00000793, -0.00000759, -0.00000161, 0.00000779, 0.00001333, 0.00000830, -0.00000582, -0.00001767, -0.00001451, 0.00000506, 0.00002596, 0.00002687, -0.00000017, -0.00003790, -0.00005376, -0.00002722, 0.00002750, 0.00006640, 0.00004970, -0.00002065, -0.00009413, -0.00010762, -0.00003930, 0.00006500, 0.00012122, 0.00007480, -0.00004742, -0.00014917, -0.00013888, -0.00000977, 0.00014273, 0.00018924, 0.00007235, -0.00013517, -0.00027501, -0.00022493, -0.00000546, 0.00021760, 0.00024808, 0.00001211, -0.00036664, -0.00063297, -0.00056797, -0.00014255, 0.00045181, 0.00091697, 0.00102805, 0.00075924, 0.00027933, -0.00016684, -0.00040290, -0.00039909, -0.00024965, -0.00008553, 0.00000831, 0.00002194, -0.00000543, -0.00003106, -0.00003703, -0.00002910, -0.00001927, -0.00001249, -0.00000672, -0.00000035, 0.00000375, 0.00000150, -0.00000667, -0.00001478, -0.00001636, -0.00001043, -0.00000245, 0.00000067, -0.00000341, -0.00001076, -0.00001500, -0.00001275, -0.00000591, 0.00000070, 0.00000351, 0.00000259, 0.00000051, -0.00000047, -0.00000009, 0.00000036, -0.00000019, -0.00000155, -0.00000261, -0.00000258, -0.00000162, -0.00000063, -0.00000036, -0.00000097, -0.00000200, },
		{ 0.00000102, 0.00000296, 0.00000365, 0.00000258, 0.00000001, -0.00000311, -0.00000563, -0.00000675, -0.00000638, -0.00000503, -0.00000347, -0.00000241, -0.00000224, -0.00000281, -0.00000333, -0.00000272, -0.00000045, 0.00000260, 0.00000423, 0.00000242, -0.00000284, -0.00000883, -0.00001172, -0.00000942, -0.00000337, 0.00000233, 0.00000417, 0.00000183, -0.00000203, -0.00000442, -0.00000458, -0.00000388, -0.00000347, -0.00000261, 0.00000019, 0.00000433, 0.00000628, 0.00000268, -0.00000495, -0.00000957, -0.00000447, 0.00000874, 0.00001835, 0.00001150, -0.00001220, -0.00003559, -0.00003550, -0.00000428, 0.00003752, 0.00005266, 0.00001821, -0.00004879, -0.00009704, -0.00007988, 0.00000138, 0.00008859, 0.00010750, 0.00003058, -0.00009034, -0.00015624, -0.00009996, 0.00004974, 0.00017752, 0.00016738, 0.00000230, -0.00020593, -0.00029180, -0.00016749, 0.00009132, 0.00028191, 0.00021927, -0.00011300, -0.00052248, -0.00071999, -0.00051344, 0.00005124, 0.00070344, 0.00110948, 0.00107349, 0.00064508, 0.00007023, -0.00036962, -0.00051736, -0.00040040, -0.00017151, 0.00000963, 0.00006811, 0.00002831, -0.00003652, -0.00006834, -0.00005585, -0.00002317, 0.00000146, 0.00000750, 0.00000192, -0.00000432, -0.00000690, -0.00000804, -0.00001009, -0.00001158, -0.00000958, -0.00000438, -0.00000035, -0.00000199, -0.00000906, -0.00001621, -0.00001744, -0.00001124, -0.00000177, 0.00000486, 0.00000550, 0.00000185, -0.00000181, -0.00000252, -0.00000071, 0.00000111, 0.00000097, -0.00000086, -0.00000263, -0.00000288, -0.00000170, -0.00000036, -0.00000006, -0.00000100, -0.00000242, -0.00000326, -0.00000286, -0.00000123, },
		{ 0.00000108, -0.00000211, -0.00000511, -0.00000680, -0.00000677, -0.00000543, -0.00000372, -0.00000252, -0.00000217, -0.00000247, -0.00000283, -0.00000255, -0.00000121, 0.00000090, 0.00000258, 0.00000222, -0.00000109, -0.00000629, -0.00001052, -0.00001077, -0.00000630, 0.00000032, 0.00000484, 0.00000444, 0.00000001, -0.00000469, -0.00000621, -0.00000421, -0.00000125, 0.00000005, -0.00000049, -0.00000093, 0.00000014, 0.00000163, 0.00000128, -0.00000110, -0.00000241, 0.00000060, 0.00000650, 0.00000842, 0.00000011, -0.00001567, -0.00002579, -0.00001682, 0.00001025, 0.00003491, 0.00003068, -0.00001035, -0.00006336, -0.00008421, -0.00004539, 0.00003392, 0.00009364, 0.00007760, -0.00001500, -0.00011714, -0.00013996, -0.00004746, 0.00010131, 0.00018590, 0.00011810, -0.00007562, -0.00025621, -0.00027088, -0.00008069, 0.00018564, 0.00030768, 0.00013647, -0.00027280, -0.00066629, -0.00074673, -0.00037827, 0.00030331, 0.00095784, 0.00124496, 0.00102934, 0.00045204, -0.00017154, -0.00054931, -0.00057190, -0.00033918, -0.00005791, 0.00010272, 0.00009899, 0.00000104, -0.00008662, -0.00010332, -0.00005779, 0.00000088, 0.00003067, 0.00002356, -0.00000070, -0.00001867, -0.00002115, -0.00001376, -0.00000606, -0.00000241, -0.00000157, -0.00000173, -0.00000360, -0.00000863, -0.00001537, -0.00001918, -0.00001606, -0.00000673, 0.00000322, 0.00000777, 0.00000523, -0.00000077, -0.00000474, -0.00000398, -0.00000016, 0.00000279, 0.00000242, -0.00000049, -0.00000316, -0.00000351, -0.00000179, 0.00000007, 0.00000036, -0.00000108, -0.00000295, -0.00000372, -0.00000271, -0.00000038, 0.00000213, 0.00000359, 0.00000324, },
		{ -0.00000726, -0.00000610, -0.00000412, -0.00000243, -0.00000174, -0.00000198, -0.00000248, -0.00000249, -0.00000164, -0.00000016, 0.00000119, 0.00000132, -0.00000059, -0.00000437, -0.00000844, -0.00001032, -0.00000825, -0.00000275, 0.00000310, 0.00000559, 0.00000315, -0.00000218, -0.00000627, -0.00000615, -0.00000248, 0.00000119, 0.00000173, -0.00000075, -0.00000323, -0.00000292, -0.00000002, 0.00000273, 0.00000310, 0.00000171, 0.00000086, 0.00000114, -0.00000000, -0.00000513, -0.00001201, -0.00001319, -0.00000285, 0.00001443, 0.00002285, 0.00000728, -0.00002986, -0.00006330, -0.00006097, -0.00001285, 0.00005209, 0.00008108, 0.00003968, -0.00005174, -0.00012278, -0.00010477, 0.00000680, 0.00013352, 0.00016529, 0.00005074, -0.00014524, -0.00027374, -0.00021332, 0.00001970, 0.00025573, 0.00028119, 0.00000473, -0.00044362, -0.00076771, -0.00069362, -0.00016643, 0.00058617, 0.00117726, 0.00129475, 0.00089055, 0.00020052, -0.00041161, -0.00067546, -0.00055424, -0.00022470, 0.00006840, 0.00017130, 0.00009022, -0.00005572, -0.00014211, -0.00012238, -0.00003688, 0.00003952, 0.00005928, 0.00002745, -0.00001731, -0.00004004, -0.00003268, -0.00001064, 0.00000637, 0.00000971, 0.00000313, -0.00000564, -0.00001234, -0.00001676, -0.00001901, -0.00001747, -0.00001078, -0.00000095, 0.00000677, 0.00000788, 0.00000255, -0.00000426, -0.00000697, -0.00000392, 0.00000166, 0.00000487, 0.00000342, -0.00000089, -0.00000420, -0.00000416, -0.00000155, 0.00000082, 0.00000083, -0.00000135, -0.00000370, -0.00000421, -0.00000237, 0.00000070, 0.00000322, 0.00000382, 0.00000220, -0.00000095, -0.00000436, -0.00000672, },
		{ -0.00000123, -0.00000115, -0.00000183, -0.00000237, -0.00000208, -0.00000102, 0.00000015, 0.00000047, -0.00000071, -0.00000340, -0.00000671, -0.00000900, -0.00000863, -0.00000506, 0.00000030, 0.00000453, 0.00000495, 0.00000127, -0.00000384, -0.00000654, -0.00000491, -0.00000059, 0.00000251, 0.00000164, -0.00000220, -0.00000515, -0.00000404, 0.00000063, 0.00000504, 0.00000565, 0.00000233, -0.00000190, -0.00000419, -0.00000446, -0.00000449, -0.00000460, -0.00000245, 0.00000359, 0.00000965, 0.00000669, -0.00001133, -0.00003748, -0.00005161, -0.00003480, 0.00001055, 0.00005412, 0.00005646, 0.00000344, -0.00007237, -0.00010808, -0.00006003, 0.00005063, 0.00013989, 0.00012054, -0.00002035, -0.00019171, -0.00025385, -0.00012972, 0.00011266, 0.00028299, 0.00019954, -0.00015837, -0.00059509, -0.00080042, -0.00055466, 0.00010204, 0.00086250, 0.00132604, 0.00124332, 0.00066942, -0.00007653, -0.00061399, -0.00072694, -0.00046664, -0.00007808, 0.00018123, 0.00019844, 0.00004124, -0.00012917, -0.00018659, -0.00011553, 0.00000595, 0.00008359, 0.00007685, 0.00001391, -0.00004577, -0.00006148, -0.00003476, 0.00000378, 0.00002459, 0.00001949, 0.00000014, -0.00001711, -0.00002408, -0.00002237, -0.00001696, -0.00001055, -0.00000346, 0.00000332, 0.00000685, 0.00000482, -0.00000140, -0.00000694, -0.00000727, -0.00000217, 0.00000407, 0.00000641, 0.00000337, -0.00000207, -0.00000534, -0.00000432, -0.00000074, 0.00000180, 0.00000107, -0.00000206, -0.00000473, -0.00000461, -0.00000168, 0.00000205, 0.00000418, 0.00000350, 0.00000046, -0.00000342, -0.00000650, -0.00000771, -0.00000690, -0.00000479, -0.00000255, },
		{ -0.00000230, -0.00000184, -0.00000087, -0.00000029, -0.00000091, -0.00000290, -0.00000560, -0.00000780, -0.00000820, -0.00000612, -0.00000201, 0.00000240, 0.00000475, 0.00000357, -0.00000047, -0.00000460, -0.00000583, -0.00000333, 0.00000067, 0.00000257, 0.00000053, -0.00000364, -0.00000586, -0.00000345, 0.00000220, 0.00000653, 0.00000581, 0.00000045, -0.00000528, -0.00000716, -0.00000449, -0.00000016, 0.00000268, 0.00000330, 0.00000256, -0.00000024, -0.00000793, -0.00002138, -0.00003422, -0.00003408, -0.00001287, 0.00002114, 0.00004290, 0.00002799, -0.00002285, -0.00007460, -0.00007925, -0.00001714, 0.00007505, 0.00012088, 0.00006273, -0.00007979, -0.00020576, -0.00020126, -0.00003799, 0.00017727, 0.00025748, 0.00007267, -0.00032571, -0.00069707, -0.00074911, -0.00034033, 0.00039393, 0.00109267, 0.00137871, 0.00109267, 0.00039393, -0.00034033, -0.00074911, -0.00069707, -0.00032571, 0.00007267, 0.00025748, 0.00017727, -0.00003799, -0.00020126, -0.00020576, -0.00007979, 0.00006273, 0.00012088, 0.00007505, -0.00001714, -0.00007925, -0.00007460, -0.00002285, 0.00002799, 0.00004290, 0.00002114, -0.00001287, -0.00003408, -0.00003422, -0.00002138, -0.00000793, -0.00000024, 0.00000256, 0.00000330, 0.00000268, -0.00000016, -0.00000449, -0.00000716, -0.00000528, 0.00000045, 0.00000581, 0.00000653, 0.00000220, -0.00000345, -0.00000586, -0.00000364, 0.00000053, 0.00000257, 0.00000067, -0.00000333, -0.00000583, -0.00000460, -0.00000047, 0.00000357, 0.00000475, 0.00000240, -0.00000201, -0.00000612, -0.00000820, -0.00000780, -0.00000560, -0.00000290, -0.00000091, -0.00000029, -0.00000087, -0.00000184, },
		{ -0.00000123, -0.00000255, -0.00000479, -0.00000690, -0.00000771, -0.00000650, -0.00000342, 0.00000046, 0.00000350, 0.00000418, 0.00000205, -0.00000168, -0.00000461, -0.00000473, -0.00000206, 0.00000107, 0.00000180, -0.00000074, -0.00000432, -0.00000534, -0.00000207, 0.00000337, 0.00000641, 0.00000407, -0.00000217, -0.00000727, -0.00000694, -0.00000140, 0.00000482, 0.00000685, 0.00000332, -0.00000346, -0.00001055, -0.00001696, -0.00002237, -0.00002408, -0.00001711, 0.00000014, 0.00001949, 0.00002459, 0.00000378, -0.00003476, -0.00006148, -0.00004577, 0.00001391, 0.00007685, 0.00008359, 0.00000595, -0.00011553, -0.00018659, -0.00012917, 0.00004124, 0.00019844, 0.00018123, -0.00007808, -0.00046664, -0.00072694, -0.00061399, -0.00007653, 0.00066942, 0.00124332, 0.00132604, 0.00086250, 0.00010204, -0.00055466, -0.00080042, -0.00059509, -0.00015837, 0.00019954, 0.00028299, 0.00011266, -0.00012972, -0.00025385, -0.00019171, -0.00002035, 0.00012054, 0.00013989, 0.00005063, -0.00006003, -0.00010808, -0.00007237, 0.00000344, 0.00005646, 0.00005412, 0.00001055, -0.00003480, -0.00005161, -0.00003748, -0.00001133, 0.00000669, 0.00000965, 0.00000359, -0.00000245, -0.00000460, -0.00000449, -0.00000446, -0.00000419, -0.00000190, 0.00000233, 0.00000565, 0.00000504, 0.00000063, -0.00000404, -0.00000515, -0.00000220, 0.00000164, 0.00000251, -0.00000059, -0.00000491, -0.00000654, -0.00000384, 0.00000127, 0.00000495, 0.00000453, 0.00000030, -0.00000506, -0.00000863, -0.00000900, -0.00000671, -0.00000340, -0.00000071, 0.00000047, 0.00000015, -0.00000102, -0.00000208, -0.00000237, -0.00000183, -0.00000115, },
		{ -0.00000726, -0.00000672, -0.00000436, -0.00000095, 0.00000220, 0.00000382, 0.00000322, 0.00000070, -0.00000237, -0.00000421, -0.00000370, -0.00000135, 0.00000083, 0.00000082, -0.00000155, -0.00000416, -0.00000420, -0.00000089, 0.00000342, 0.00000487, 0.00000166, -0.00000392, -0.00000697, -0.00000426, 0.00000255, 0.00000788, 0.00000677, -0.00000095, -0.00001078, -0.00001747, -0.00001901, -0.00001676, -0.00001234, -0.00000564, 0.00000313, 0.00000971, 0.00000637, -0.00001064, -0.00003268, -0.00004004, -0.00001731, 0.00002745, 0.00005928, 0.00003952, -0.00003688, -0.00012238, -0.00014211, -0.00005572, 0.00009022, 0.00017130, 0.00006840, -0.00022470, -0.00055424, -0.00067546, -0.00041161, 0.00020052, 0.00089055, 0.00129475, 0.00117726, 0.00058617, -0.00016643, -0.00069362, -0.00076771, -0.00044362, 0.00000473, 0.00028119, 0.00025573, 0.00001970, -0.00021332, -0.00027374, -0.00014524, 0.00005074, 0.00016529, 0.00013352, 0.00000680, -0.00010477, -0.00012278, -0.00005174, 0.00003968, 0.00008108, 0.00005209, -0.00001285, -0.00006097, -0.00006330, -0.00002986, 0.00000728, 0.00002285, 0.00001443, -0.00000285, -0.00001319, -0.00001201, -0.00000513, -0.00000000, 0.00000114, 0.00000086, 0.00000171, 0.00000310, 0.00000273, -0.00000002, -0.00000292, -0.00000323, -0.00000075, 0.00000173, 0.00000119, -0.00000248, -0.00000615, -0.00000627, -0.00000218, 0.00000315, 0.00000559, 0.00000310, -0.00000275, -0.00000825, -0.00001032, -0.00000844, -0.00000437, -0.00000059, 0.00000132, 0.00000119, -0.00000016, -0.00000164, -0.00000249, -0.00000248, -0.00000198, -0.00000174, -0.00000243, -0.00000412, -0.00000610, },
		{ 0.00000108, 0.00000324, 0.00000359, 0.00000213, -0.00000038, -0.00000271, -0.00000372, -0.00000295, -0.00000108, 0.00000036, 0.00000007, -0.00000179, -0.00000351, -0.00000316, -0.00000049, 0.00000242, 0.00000279, -0.00000016, -0.00000398, -0.00000474, -0.00000077, 0.00000523, 0.00000777, 0.00000322, -0.00000673, -0.00001606, -0.00001918, -0.00001537, -0.00000863, -0.00000360, -0.00000173, -0.00000157, -0.00000241, -0.00000606, -0.00001376, -0.00002115, -0.00001867, -0.00000070, 0.00002356, 0.00003067, 0.00000088, -0.00005779, -0.00010332, -0.00008662, 0.00000104, 0.00009899, 0.00010272, -0.00005791, -0.00033918, -0.00057190, -0.00054931, -0.00017154, 0.00045204, 0.00102934, 0.00124496, 0.00095784, 0.00030331, -0.00037827, -0.00074673, -0.00066629, -0.00027280, 0.00013647, 0.00030768, 0.00018564, -0.00008069, -0.00027088, -0.00025621, -0.00007562, 0.00011810, 0.00018590, 0.00010131, -0.00004746, -0.00013996, -0.00011714, -0.00001500, 0.00007760, 0.00009364, 0.00003392, -0.00004539, -0.00008421, -0.00006336, -0.00001035, 0.00003068, 0.00003491, 0.00001025, -0.00001682, -0.00002579, -0.00001567, 0.00000011, 0.00000842, 0.00000650, 0.00000060, -0.00000241, -0.00000110, 0.00000128, 0.00000163, 0.00000014, -0.00000093, -0.00000049, 0.00000005, -0.00000125, -0.00000421, -0.00000621, -0.00000469, 0.00000001, 0.00000444, 0.00000484, 0.00000032, -0.00000630, -0.00001077, -0.00001052, -0.00000629, -0.00000109, 0.00000222, 0.00000258, 0.00000090, -0.00000121, -0.00000255, -0.00000283, -0.00000247, -0.00000217, -0.00000252, -0.00000372, -0.00000543, -0.00000677, -0.00000680, -0.00000511, -0.00000211, },
		{ 0.00000102, -0.00000123, -0.00000286, -0.00000326, -0.00000242, -0.00000100, -0.00000006, -0.00000036, -0.00000170, -0.00000288, -0.00000263, -0.00000086, 0.00000097, 0.00000111, -0.00000071, -0.00000252, -0.00000181, 0.00000185, 0.00000550, 0.00000486, -0.00000177, -0.00001124, -0.00001744, -0.00001621, -0.00000906, -0.00000199, -0.00000035, -0.00000438, -0.00000958, -0.00001158, -0.00001009, -0.00000804, -0.00000690, -0.00000432, 0.00000192, 0.00000750, 0.00000146, -0.00002317, -0.00005585, -0.00006834, -0.00003652, 0.00002831, 0.00006811, 0.00000963, -0.00017151, -0.00040040, -0.00051736, -0.00036962, 0.00007023, 0.00064508, 0.00107349, 0.00110948, 0.00070344, 0.00005124, -0.00051344, -0.00071999, -0.00052248, -0.00011300, 0.00021927, 0.00028191, 0.00009132, -0.00016749, -0.00029180, -0.00020593, 0.00000230, 0.00016738, 0.00017752, 0.00004974, -0.00009996, -0.00015624, -0.00009034, 0.00003058, 0.00010750, 0.00008859, 0.00000138, -0.00007988, -0.00009704, -0.00004879, 0.00001821, 0.00005266, 0.00003752, -0.00000428, -0.00003550, -0.00003559, -0.00001220, 0.00001150, 0.00001835, 0.00000874, -0.00000447, -0.00000957, -0.00000495, 0.00000268, 0.00000628, 0.00000433, 0.00000019, -0.00000261, -0.00000347, -0.00000388, -0.00000458, -0.00000442, -0.00000203, 0.00000183, 0.00000417, 0.00000233, -0.00000337, -0.00000942, -0.00001172, -0.00000883, -0.00000284, 0.00000242, 0.00000423, 0.00000260, -0.00000045, -0.00000272, -0.00000333, -0.00000281, -0.00000224, -0.00000241, -0.00000347, -0.00000503, -0.00000638, -0.00000675, -0.00000563, -0.00000311, 0.00000001, 0.00000258, 0.00000365, 0.00000296, },
		{ -0.00000281, -0.00000200, -0.00000097, -0.00000036, -0.00000063, -0.00000162, -0.00000258, -0.00000261, -0.00000155, -0.00000019, 0.00000036, -0.00000009, -0.00000047, 0.00000051, 0.00000259, 0.00000351, 0.00000070, -0.00000591, -0.00001275, -0.00001500, -0.00001076, -0.00000341, 0.00000067, -0.00000245, -0.00001043, -0.00001636, -0.00001478, -0.00000667, 0.00000150, 0.00000375, -0.00000035, -0.00000672, -0.00001249, -0.00001927, -0.00002910, -0.00003703, -0.00003106, -0.00000543, 0.00002194, 0.00000831, -0.00008553, -0.00024965, -0.00039909, -0.00040290, -0.00016684, 0.00027933, 0.00075924, 0.00102805, 0.00091697, 0.00045181, -0.00014255, -0.00056797, -0.00063297, -0.00036664, 0.00001211, 0.00024808, 0.00021760, -0.00000546, -0.00022493, -0.00027501, -0.00013517, 0.00007235, 0.00018924, 0.00014273, -0.00000977, -0.00013888, -0.00014917, -0.00004742, 0.00007480, 0.00012122, 0.00006500, -0.00003930, -0.00010762, -0.00009413, -0.00002065, 0.00004970, 0.00006640, 0.00002750, -0.00002722, -0.00005376, -0.00003790, -0.00000017, 0.00002687, 0.00002596, 0.00000506, -0.00001451, -0.00001767, -0.00000582, 0.00000830, 0.00001333, 0.00000779, -0.00000161, -0.00000759, -0.00000793, -0.00000517, -0.00000251, -0.00000081, 0.00000068, 0.00000200, 0.00000166, -0.00000165, -0.00000694, -0.00001091, -0.00001045, -0.00000541, 0.00000112, 0.00000515, 0.00000477, 0.00000124, -0.00000247, -0.00000412, -0.00000357, -0.00000230, -0.00000188, -0.00000286, -0.00000466, -0.00000628, -0.00000686, -0.00000602, -0.00000386, -0.00000094, 0.00000182, 0.00000346, 0.00000347, 0.00000204, -0.00000004, -0.00000188, -0.00000285, },
		{ -0.00000070, -0.00000095, -0.00000172, -0.00000254, -0.00000276, -0.00000204, -0.00000070, 0.00000047, 0.00000101, 0.00000115, 0.00000140, 0.00000155, 0.00000044, -0.00000290, -0.00000770, -0.00001112, -0.00001035, -0.00000554, -0.00000074, -0.00000100, -0.00000768, -0.00001616, -0.00001891, -0.00001198, 0.00000066, 0.00000948, 0.00000700, -0.00000585, -0.00002055, -0.00002827, -0.00002664, -0.00001965, -0.00001179, -0.00000439, -0.00000041, -0.00001272, -0.00006275, -0.00016085, -0.00027857, -0.00033962, -0.00025192, 0.00002649, 0.00043035, 0.00079007, 0.00091290, 0.00070208, 0.00023472, -0.00026502, -0.00055287, -0.00051282, -0.00022601, 0.00008991, 0.00022995, 0.00013465, -0.00008548, -0.00024596, -0.00022839, -0.00005992, 0.00012156, 0.00018164, 0.00009056, -0.00006425, -0.00015619, -0.00012044, 0.00000226, 0.00010754, 0.00011470, 0.00002705, -0.00007912, -0.00012107, -0.00007416, 0.00001585, 0.00007632, 0.00006671, 0.00000519, -0.00005360, -0.00006570, -0.00002973, 0.00001906, 0.00004200, 0.00002732, -0.00000575, -0.00002768, -0.00002346, -0.00000140, 0.00001812, 0.00002082, 0.00000816, -0.00000747, -0.00001473, -0.00001148, -0.00000349, 0.00000240, 0.00000362, 0.00000182, -0.00000061, -0.00000306, -0.00000599, -0.00000891, -0.00000979, -0.00000695, -0.00000116, 0.00000433, 0.00000618, 0.00000362, -0.00000103, -0.00000443, -0.00000481, -0.00000305, -0.00000144, -0.00000168, -0.00000370, -0.00000608, -0.00000729, -0.00000668, -0.00000453, -0.00000163, 0.00000116, 0.00000308, 0.00000362, 0.00000275, 0.00000095, -0.00000095, -0.00000224, -0.00000264, -0.00000232, -0.00000165, -0.00000100, },
		{ -0.00000249, -0.00000269, -0.00000207, -0.00000067, 0.00000086, 0.00000176, 0.00000172, 0.00000093, -0.00000037, -0.00000223, -0.00000467, -0.00000692, -0.00000755, -0.00000568, -0.00000259, -0.00000161, -0.00000546, -0.00001304, -0.00001881, -0.00001663, -0.00000559, 0.00000750, 0.00001216, 0.00000257, -0.00001660, -0.00003271, -0.00003453, -0.00002112, -0.00000236, 0.00000857, 0.00000336, -0.00002016, -0.00006282, -0.00012560, -0.00020051, -0.00025640, -0.00023808, -0.00009285, 0.00018305, 0.00051141, 0.00074836, 0.00075694, 0.00049749, 0.00007225, -0.00031815, -0.00048947, -0.00038722, -0.00011909, 0.00012007, 0.00018046, 0.00005336, -0.00013650, -0.00023293, -0.00016563, 0.00000475, 0.00014328, 0.00014982, 0.00003348, -0.00010286, -0.00014976, -0.00007727, 0.00004773, 0.00012176, 0.00008926, -0.00001706, -0.00010891, -0.00011620, -0.00004048, 0.00005273, 0.00009099, 0.00005183, -0.00002530, -0.00007671, -0.00006657, -0.00001102, 0.00004144, 0.00005193, 0.00001985, -0.00002254, -0.00004036, -0.00002362, 0.00000921, 0.00003035, 0.00002579, 0.00000345, -0.00001718, -0.00002207, -0.00001187, 0.00000219, 0.00000935, 0.00000697, -0.00000027, -0.00000639, -0.00000894, -0.00000892, -0.00000786, -0.00000580, -0.00000213, 0.00000246, 0.00000560, 0.00000513, 0.00000125, -0.00000332, -0.00000553, -0.00000448, -0.00000195, -0.00000069, -0.00000202, -0.00000503, -0.00000748, -0.00000772, -0.00000563, -0.00000234, 0.00000078, 0.00000284, 0.00000356, 0.00000304, 0.00000164, -0.00000009, -0.00000151, -0.00000222, -0.00000223, -0.00000187, -0.00000148, -0.00000124, -0.00000118, -0.00000138, -0.00000189, },
		{ -0.00000057, 0.00000091, 0.00000173, 0.00000131, -0.00000015, -0.00000196, -0.00000347, -0.00000437, -0.00000452, -0.00000384, -0.00000275, -0.00000254, -0.00000493, -0.00001031, -0.00001612, -0.00001754, -0.00001118, 0.00000087, 0.00001047, 0.00000867, -0.00000666, -0.00002702, -0.00003808, -0.00003038, -0.00000782, 0.00001346, 0.00001565, -0.00000924, -0.00005605, -0.00011213, -0.00016432, -0.00019842, -0.00019271, -0.00011755, 0.00004778, 0.00028617, 0.00052511, 0.00065582, 0.00059075, 0.00032746, -0.00002952, -0.00031596, -0.00040298, -0.00027815, -0.00005312, 0.00011230, 0.00011852, -0.00001060, -0.00015527, -0.00019561, -0.00010196, 0.00004847, 0.00013827, 0.00010413, -0.00001626, -0.00011977, -0.00012354, -0.00003001, 0.00007972, 0.00011545, 0.00005109, -0.00005758, -0.00012208, -0.00009377, -0.00000022, 0.00008158, 0.00008953, 0.00002412, -0.00005721, -0.00009016, -0.00005443, 0.00001502, 0.00006131, 0.00005251, 0.00000343, -0.00004186, -0.00004831, -0.00001601, 0.00002484, 0.00004172, 0.00002532, -0.00000700, -0.00002895, -0.00002679, -0.00000729, 0.00001154, 0.00001632, 0.00000697, -0.00000661, -0.00001453, -0.00001395, -0.00000850, -0.00000309, 0.00000026, 0.00000230, 0.00000385, 0.00000426, 0.00000244, -0.00000125, -0.00000456, -0.00000529, -0.00000332, -0.00000090, -0.00000059, -0.00000308, -0.00000658, -0.00000844, -0.00000725, -0.00000369, 0.00000027, 0.00000293, 0.00000375, 0.00000315, 0.00000186, 0.00000043, -0.00000079, -0.00000158, -0.00000185, -0.00000177, -0.00000165, -0.00000166, -0.00000175, -0.00000177, -0.00000175, -0.00000185, -0.00000214, -0.00000229, -0.00000181, },
		{ 0.00000063, -0.00000083, -0.00000238, -0.00000312, -0.00000280, -0.00000196, -0.00000147, -0.00000215, -0.00000459, -0.00000877, -0.00001347, -0.00001602, -0.00001348, -0.00000532, 0.00000443, 0.00000839, 0.00000088, -0.00001627, -0.00003252, -0.00003505, -0.00001910, 0.00000581, 0.00002001, 0.00000650, -0.00003684, -0.00009519, -0.00014592, -0.00017051, -0.00015895, -0.00010503, -0.00000193, 0.00015109, 0.00033206, 0.00048584, 0.00053876, 0.00043986, 0.00020424, -0.00007707, -0.00027903, -0.00031614, -0.00019815, -0.00002474, 0.00008198, 0.00006102, -0.00004924, -0.00014671, -0.00014741, -0.00004988, 0.00006770, 0.00011349, 0.00005660, -0.00004995, -0.00011521, -0.00008592, 0.00001101, 0.00009311, 0.00009196, 0.00000932, -0.00008604, -0.00011649, -0.00005907, 0.00003777, 0.00009602, 0.00007202, -0.00001044, -0.00008305, -0.00008993, -0.00003094, 0.00004271, 0.00007315, 0.00004196, -0.00001924, -0.00005886, -0.00004801, -0.00000065, 0.00004235, 0.00004844, 0.00001762, -0.00002196, -0.00003964, -0.00002617, 0.00000277, 0.00002274, 0.00002052, 0.00000184, -0.00001687, -0.00002293, -0.00001564, -0.00000345, 0.00000486, 0.00000652, 0.00000431, 0.00000194, 0.00000063, -0.00000062, -0.00000248, -0.00000402, -0.00000386, -0.00000210, -0.00000066, -0.00000150, -0.00000463, -0.00000780, -0.00000843, -0.00000569, -0.00000111, 0.00000278, 0.00000434, 0.00000366, 0.00000197, 0.00000043, -0.00000049, -0.00000093, -0.00000115, -0.00000132, -0.00000154, -0.00000186, -0.00000223, -0.00000242, -0.00000222, -0.00000175, -0.00000142, -0.00000150, -0.00000179, -0.00000172, -0.00000090, 0.00000033, 0.00000106, },
		{ -0.00000188, -0.00000122, -0.00000050, -0.00000096, -0.00000328, -0.00000716, -0.00001136, -0.00001408, -0.00001351, -0.00000885, -0.00000163, 0.00000388, 0.00000252, -0.00000770, -0.00002225, -0.00003116, -0.00002567, -0.00000654, 0.00001292, 0.00001365, -0.00001590, -0.00006954, -0.00012465, -0.00015492, -0.00014483, -0.00009466, -0.00001366, 0.00008978, 0.00020900, 0.00032855, 0.00041453, 0.00042156, 0.00032030, 0.00012800, -0.00008520, -0.00022841, -0.00024461, -0.00014964, -0.00002334, 0.00004509, 0.00001909, -0.00006265, -0.00012097, -0.00010118, -0.00001638, 0.00006544, 0.00007924, 0.00001744, -0.00006432, -0.00009465, -0.00004693, 0.00003834, 0.00008804, 0.00005864, -0.00002676, -0.00009761, -0.00009489, -0.00002030, 0.00006554, 0.00009355, 0.00004272, -0.00004411, -0.00009686, -0.00007561, -0.00000099, 0.00006557, 0.00007331, 0.00002161, -0.00004329, -0.00006875, -0.00003797, 0.00001996, 0.00005739, 0.00004744, 0.00000288, -0.00003838, -0.00004572, -0.00001864, 0.00001702, 0.00003268, 0.00001929, -0.00000892, -0.00002917, -0.00002866, -0.00001195, 0.00000622, 0.00001410, 0.00001047, 0.00000228, -0.00000322, -0.00000390, -0.00000227, -0.00000136, -0.00000175, -0.00000217, -0.00000188, -0.00000180, -0.00000323, -0.00000596, -0.00000797, -0.00000713, -0.00000328, 0.00000151, 0.00000457, 0.00000468, 0.00000266, 0.00000040, -0.00000078, -0.00000081, -0.00000048, -0.00000047, -0.00000098, -0.00000180, -0.00000258, -0.00000303, -0.00000291, -0.00000220, -0.00000124, -0.00000065, -0.00000087, -0.00000166, -0.00000221, -0.00000182, -0.00000064, 0.00000043, 0.00000050, -0.00000047, -0.00000158, },
		{ -0.00000176, -0.00000478, -0.00000874, -0.00001193, -0.00001270, -0.00001040, -0.00000581, -0.00000126, 0.00000002, -0.00000436, -0.00001363, -0.00002265, -0.00002418, -0.00001447, 0.00000142, 0.00000971, -0.00000419, -0.00004407, -0.00009639, -0.00013532, -0.00013746, -0.00009577, -0.00002207, 0.00006366, 0.00014716, 0.00022416, 0.00029115, 0.00033277, 0.00032173, 0.00023730, 0.00008977, -0.00007161, -0.00018068, -0.00019521, -0.00012705, -0.00003554, 0.00001397, -0.00000340, -0.00005707, -0.00008963, -0.00006594, -0.00000219, 0.00004919, 0.00004575, -0.00000743, -0.00006157, -0.00006652, -0.00001514, 0.00004917, 0.00006929, 0.00002438, -0.00005039, -0.00009220, -0.00006378, 0.00001399, 0.00007825, 0.00007611, 0.00000870, -0.00006958, -0.00009597, -0.00005052, 0.00002874, 0.00007830, 0.00006124, -0.00000443, -0.00006314, -0.00006841, -0.00001944, 0.00004147, 0.00006573, 0.00003754, -0.00001647, -0.00005220, -0.00004440, -0.00000450, 0.00003245, 0.00003790, 0.00001109, -0.00002397, -0.00004032, -0.00002884, -0.00000212, 0.00001886, 0.00002173, 0.00000981, -0.00000427, -0.00001026, -0.00000715, -0.00000093, 0.00000234, 0.00000126, -0.00000160, -0.00000356, -0.00000427, -0.00000496, -0.00000612, -0.00000655, -0.00000471, -0.00000071, 0.00000339, 0.00000518, 0.00000397, 0.00000119, -0.00000098, -0.00000137, -0.00000049, 0.00000030, 0.00000005, -0.00000116, -0.00000262, -0.00000358, -0.00000366, -0.00000288, -0.00000155, -0.00000033, 0.00000005, -0.00000075, -0.00000225, -0.00000332, -0.00000299, -0.00000132, 0.00000058, 0.00000142, 0.00000087, -0.00000026, -0.00000086, -0.00000066, -0.00000054, },
		{ -0.00001095, -0.00001087, -0.00000855, -0.00000545, -0.00000366, -0.00000487, -0.00000933, -0.00001505, -0.00001816, -0.00001514, -0.00000644, 0.00000114, -0.00000339, -0.00002794, -0.00006893, -0.00010892, -0.00012418, -0.00009897, -0.00003640, 0.00004313, 0.00011571, 0.00016922, 0.00020722, 0.00023782, 0.00025772, 0.00024772, 0.00018695, 0.00007600, -0.00005193, -0.00014525, -0.00016678, -0.00012048, -0.00004949, -0.00000490, -0.00000892, -0.00004168, -0.00006219, -0.00004526, -0.00000294, 0.00002796, 0.00002040, -0.00001718, -0.00004773, -0.00003931, 0.00000445, 0.00004545, 0.00004429, -0.00000312, -0.00005890, -0.00007403, -0.00003131, 0.00003727, 0.00007533, 0.00004924, -0.00002227, -0.00008214, -0.00008161, -0.00002067, 0.00005182, 0.00007834, 0.00003968, -0.00003040, -0.00007415, -0.00005739, 0.00000389, 0.00005890, 0.00006468, 0.00002003, -0.00003633, -0.00005953, -0.00003469, 0.00001380, 0.00004510, 0.00003573, -0.00000369, -0.00004024, -0.00004672, -0.00002172, 0.00001278, 0.00003138, 0.00002472, 0.00000332, -0.00001421, -0.00001672, -0.00000696, 0.00000407, 0.00000765, 0.00000327, -0.00000350, -0.00000732, -0.00000718, -0.00000544, -0.00000427, -0.00000357, -0.00000192, 0.00000106, 0.00000388, 0.00000457, 0.00000272, -0.00000009, -0.00000169, -0.00000126, 0.00000019, 0.00000092, 0.00000004, -0.00000194, -0.00000377, -0.00000443, -0.00000373, -0.00000214, -0.00000046, 0.00000050, 0.00000015, -0.00000150, -0.00000356, -0.00000463, -0.00000367, -0.00000098, 0.00000189, 0.00000322, 0.00000246, 0.00000057, -0.00000094, -0.00000141, -0.00000155, -0.00000269, -0.00000537, -0.00000869, },
		{ -0.00000732, -0.00000694, -0.00000831, -0.00001068, -0.00001234, -0.00001160, -0.00000822, -0.00000493, -0.00000753, -0.00002219, -0.00005010, -0.00008290, -0.00010353, -0.00009451, -0.00004958, 0.00001984, 0.00008992, 0.00013926, 0.00016273, 0.00017249, 0.00018421, 0.00019935, 0.00019945, 0.00015982, 0.00007320, -0.00003650, -0.00012429, -0.00015306, -0.00011981, -0.00005760, -0.00001098, -0.00000380, -0.00002519, -0.00004393, -0.00003772, -0.00001170, 0.00000931, 0.00000626, -0.00001521, -0.00003016, -0.00001915, 0.00001123, 0.00003250, 0.00002059, -0.00001936, -0.00005401, -0.00004987, -0.00000472, 0.00004673, 0.00006021, 0.00002021, -0.00004406, -0.00008079, -0.00005836, 0.00000715, 0.00006388, 0.00006652, 0.00001374, -0.00005067, -0.00007407, -0.00003814, 0.00002689, 0.00006820, 0.00005401, -0.00000157, -0.00005210, -0.00005799, -0.00001797, 0.00003227, 0.00005125, 0.00002530, -0.00002255, -0.00005369, -0.00004551, -0.00000742, 0.00002984, 0.00004012, 0.00002091, -0.00000846, -0.00002496, -0.00001975, -0.00000195, 0.00001205, 0.00001257, 0.00000240, -0.00000848, -0.00001235, -0.00000899, -0.00000344, -0.00000022, 0.00000016, 0.00000001, 0.00000113, 0.00000286, 0.00000331, 0.00000175, -0.00000055, -0.00000161, -0.00000081, 0.00000065, 0.00000095, -0.00000059, -0.00000305, -0.00000475, -0.00000467, -0.00000304, -0.00000091, 0.00000057, 0.00000067, -0.00000073, -0.00000306, -0.00000507, -0.00000533, -0.00000314, 0.00000068, 0.00000409, 0.00000509, 0.00000327, 0.00000010, -0.00000230, -0.00000288, -0.00000242, -0.00000263, -0.00000452, -0.00000745, -0.00000974, -0.00001016, -0.00000888, },
		{ -0.00000864, -0.00000780, -0.00000652, -0.00000643, -0.00001040, -0.00002142, -0.00004053, -0.00006404, -0.00008224, -0.00008205, -0.00005413, -0.00000060, 0.00006274, 0.00011387, 0.00013817, 0.00013866, 0.00013351, 0.00014054, 0.00016014, 0.00017104, 0.00014510, 0.00007141, -0.00002949, -0.00011463, -0.00014620, -0.00011759, -0.00005739, -0.00000794, 0.00000481, -0.00001344, -0.00003561, -0.00003865, -0.00002165, -0.00000241, 0.00000228, -0.00000712, -0.00001507, -0.00000850, 0.00000828, 0.00001684, 0.00000364, -0.00002380, -0.00004057, -0.00002675, 0.00001156, 0.00004356, 0.00003888, -0.00000424, -0.00005368, -0.00006819, -0.00003253, 0.00002734, 0.00006378, 0.00004667, -0.00001066, -0.00006140, -0.00006365, -0.00001519, 0.00004460, 0.00006747, 0.00003605, -0.00002254, -0.00006011, -0.00004750, 0.00000231, 0.00004651, 0.00004867, 0.00000806, -0.00004146, -0.00006047, -0.00003562, 0.00001175, 0.00004502, 0.00004193, 0.00000995, -0.00002314, -0.00003314, -0.00001716, 0.00000772, 0.00002092, 0.00001467, -0.00000259, -0.00001597, -0.00001688, -0.00000799, 0.00000164, 0.00000533, 0.00000319, -0.00000012, -0.00000085, 0.00000087, 0.00000239, 0.00000187, 0.00000002, -0.00000115, -0.00000070, 0.00000036, 0.00000027, -0.00000153, -0.00000389, -0.00000503, -0.00000413, -0.00000186, 0.00000023, 0.00000089, -0.00000021, -0.00000251, -0.00000485, -0.00000592, -0.00000467, -0.00000112, 0.00000332, 0.00000629, 0.00000598, 0.00000257, -0.00000174, -0.00000429, -0.00000398, -0.00000210, -0.00000116, -0.00000273, -0.00000626, -0.00000972, -0.00001136, -0.00001099, -0.00000978, -0.00000896, -0.00000880, },
		{ -0.00001035, -0.00002038, -0.00003546, -0.00005269, -0.00006616, -0.00006783, -0.00005048, -0.00001241, 0.00003883, 0.00008742, 0.00011686, 0.00012069, 0.00010838, 0.00009993, 0.00011072, 0.00013693, 0.00015429, 0.00013396, 0.00006561, -0.00003001, -0.00011071, -0.00013969, -0.00011043, -0.00005047, -0.00000113, 0.00001140, -0.00000850, -0.00003461, -0.00004253, -0.00002812, -0.00000652, 0.00000475, 0.00000154, -0.00000595, -0.00000630, 0.00000056, 0.00000397, -0.00000444, -0.00001931, -0.00002466, -0.00000988, 0.00001675, 0.00003201, 0.00001782, -0.00001975, -0.00005178, -0.00004946, -0.00001017, 0.00003700, 0.00005360, 0.00002423, -0.00002859, -0.00006154, -0.00004625, 0.00000623, 0.00005367, 0.00005753, 0.00001506, -0.00003833, -0.00005876, -0.00003044, 0.00002162, 0.00005308, 0.00003743, -0.00001249, -0.00005595, -0.00005822, -0.00001832, 0.00003180, 0.00005428, 0.00003529, -0.00000639, -0.00003707, -0.00003565, -0.00000825, 0.00002020, 0.00002775, 0.00001191, -0.00001165, -0.00002416, -0.00001867, -0.00000300, 0.00000933, 0.00001074, 0.00000389, -0.00000305, -0.00000440, -0.00000086, 0.00000292, 0.00000347, 0.00000119, -0.00000114, -0.00000163, -0.00000088, -0.00000073, -0.00000207, -0.00000387, -0.00000443, -0.00000308, -0.00000083, 0.00000058, 0.00000009, -0.00000204, -0.00000453, -0.00000596, -0.00000531, -0.00000237, 0.00000202, 0.00000599, 0.00000739, 0.00000514, 0.00000030, -0.00000425, -0.00000576, -0.00000369, -0.00000026, 0.00000120, -0.00000118, -0.00000640, -0.00001155, -0.00001406, -0.00001341, -0.00001096, -0.00000840, -0.00000649, -0.00000514, -0.00000444, -0.00000554, },
		{ -0.00005534, -0.00005677, -0.00004432, -0.00001659, 0.00002264, 0.00006384, 0.00009443, 0.00010500, 0.00009614, 0.00008055, 0.00007627, 0.00009347, 0.00012378, 0.00014188, 0.00012135, 0.00005524, -0.00003444, -0.00010730, -0.00013002, -0.00009858, -0.00004055, 0.00000474, 0.00001335, -0.00000926, -0.00003682, -0.00004502, -0.00002942, -0.00000510, 0.00000921, 0.00000705, -0.00000319, -0.00000918, -0.00000704, -0.00000314, -0.00000475, -0.00001057, -0.00001147, -0.00000144, 0.00001331, 0.00001763, 0.00000204, -0.00002523, -0.00004189, -0.00003035, 0.00000462, 0.00003644, 0.00003768, 0.00000451, -0.00003746, -0.00005288, -0.00002665, 0.00002179, 0.00005330, 0.00004154, -0.00000419, -0.00004599, -0.00004897, -0.00001088, 0.00003568, 0.00005057, 0.00001999, -0.00003177, -0.00006249, -0.00004688, 0.00000297, 0.00004815, 0.00005491, 0.00002098, -0.00002458, -0.00004636, -0.00003074, 0.00000571, 0.00003231, 0.00002972, 0.00000357, -0.00002318, -0.00003040, -0.00001585, 0.00000600, 0.00001800, 0.00001392, 0.00000103, -0.00000850, -0.00000815, -0.00000088, 0.00000558, 0.00000617, 0.00000192, -0.00000256, -0.00000394, -0.00000262, -0.00000120, -0.00000144, -0.00000273, -0.00000330, -0.00000232, -0.00000075, -0.00000029, -0.00000169, -0.00000410, -0.00000583, -0.00000554, -0.00000294, 0.00000116, 0.00000525, 0.00000751, 0.00000659, 0.00000253, -0.00000275, -0.00000621, -0.00000567, -0.00000167, 0.00000256, 0.00000323, -0.00000113, -0.00000850, -0.00001479, -0.00001685, -0.00001438, -0.00000967, -0.00000544, -0.00000312, -0.00000281, -0.00000457, -0.00000924, -0.00001797, -0.00003065, -0.00004471, },
		{ 0.00001274, 0.00004645, 0.00007422, 0.00008829, 0.00008581, 0.00007205, 0.00005973, 0.00006245, 0.00008464, 0.00011482, 0.00012917, 0.00010601, 0.00004253, -0.00003877, -0.00010124, -0.00011677, -0.00008455, -0.00003140, 0.00000708, 0.00001085, -0.00001285, -0.00003857, -0.00004404, -0.00002644, -0.00000143, 0.00001223, 0.00000818, -0.00000480, -0.00001321, -0.00001133, -0.00000431, -0.00000054, -0.00000219, -0.00000393, -0.00000053, 0.00000551, 0.00000538, -0.00000616, -0.00002257, -0.00002891, -0.00001550, 0.00001058, 0.00002853, 0.00002105, -0.00000883, -0.00003743, -0.00003923, -0.00000954, 0.00002928, 0.00004521, 0.00002390, -0.00001782, -0.00004492, -0.00003384, 0.00000663, 0.00004178, 0.00004006, 0.00000049, -0.00004546, -0.00005959, -0.00002881, 0.00002361, 0.00005703, 0.00004639, 0.00000194, -0.00004026, -0.00004815, -0.00001872, 0.00002171, 0.00004056, 0.00002512, -0.00000932, -0.00003435, -0.00003207, -0.00000766, 0.00001762, 0.00002520, 0.00001303, -0.00000559, -0.00001513, -0.00001028, 0.00000193, 0.00001019, 0.00000881, 0.00000094, -0.00000599, -0.00000723, -0.00000383, -0.00000014, 0.00000074, -0.00000086, -0.00000254, -0.00000277, -0.00000213, -0.00000222, -0.00000364, -0.00000533, -0.00000551, -0.00000328, 0.00000073, 0.00000486, 0.00000734, 0.00000703, 0.00000383, -0.00000107, -0.00000535, -0.00000653, -0.00000366, 0.00000150, 0.00000509, 0.00000365, -0.00000319, -0.00001207, -0.00001804, -0.00001804, -0.00001285, -0.00000615, -0.00000168, -0.00000105, -0.00000371, -0.00000869, -0.00001592, -0.00002569, -0.00003695, -0.00004625, -0.00004855, -0.00003958, -0.00001816, },
		{ 0.00007504, 0.00006569, 0.00005268, 0.00004653, 0.00005563, 0.00007964, 0.00010609, 0.00011467, 0.00008921, 0.00003044, -0.00004039, -0.00009186, -0.00010161, -0.00007125, -0.00002535, 0.00000558, 0.00000584, -0.00001632, -0.00003779, -0.00003972, -0.00002144, 0.00000158, 0.00001226, 0.00000591, -0.00000786, -0.00001545, -0.00001142, -0.00000158, 0.00000443, 0.00000274, -0.00000225, -0.00000419, -0.00000226, -0.00000180, -0.00000726, -0.00001550, -0.00001737, -0.00000724, 0.00000915, 0.00001752, 0.00000782, -0.00001459, -0.00003101, -0.00002498, 0.00000192, 0.00002896, 0.00003292, 0.00000882, -0.00002392, -0.00003692, -0.00001757, 0.00001853, 0.00003956, 0.00002473, -0.00001654, -0.00005084, -0.00004829, -0.00000818, 0.00003912, 0.00005655, 0.00003064, -0.00001749, -0.00004962, -0.00004157, -0.00000203, 0.00003584, 0.00004220, 0.00001408, -0.00002410, -0.00004193, -0.00002749, 0.00000506, 0.00002927, 0.00002841, 0.00000720, -0.00001485, -0.00002080, -0.00000910, 0.00000779, 0.00001557, 0.00000974, -0.00000275, -0.00001120, -0.00001033, -0.00000321, 0.00000310, 0.00000411, 0.00000066, -0.00000335, -0.00000509, -0.00000478, -0.00000428, -0.00000457, -0.00000475, -0.00000325, 0.00000031, 0.00000459, 0.00000744, 0.00000741, 0.00000447, -0.00000010, -0.00000436, -0.00000631, -0.00000477, -0.00000027, 0.00000455, 0.00000601, 0.00000181, -0.00000692, -0.00001573, -0.00001964, -0.00001661, -0.00000903, -0.00000192, 0.00000062, -0.00000216, -0.00000824, -0.00001524, -0.00002235, -0.00002979, -0.00003674, -0.00004005, -0.00003536, -0.00001997, 0.00000478, 0.00003341, 0.00005842, 0.00007331, },
		{ 0.00003905, 0.00005232, 0.00007524, 0.00009611, 0.00009909, 0.00007315, 0.00002105, -0.00003863, -0.00008034, -0.00008684, -0.00006071, -0.00002291, 0.00000157, 0.00000066, -0.00001779, -0.00003421, -0.00003361, -0.00001674, 0.00000241, 0.00000964, 0.00000231, -0.00000988, -0.00001473, -0.00000844, 0.00000220, 0.00000733, 0.00000337, -0.00000456, -0.00000880, -0.00000693, -0.00000343, -0.00000383, -0.00000802, -0.00001013, -0.00000537, 0.00000347, 0.00000738, 0.00000012, -0.00001394, -0.00002186, -0.00001371, 0.00000675, 0.00002318, 0.00002042, -0.00000089, -0.00002293, -0.00002536, -0.00000408, 0.00002316, 0.00003094, 0.00000880, -0.00002751, -0.00004755, -0.00003179, 0.00001034, 0.00004647, 0.00004736, 0.00001150, -0.00003288, -0.00005053, -0.00002804, 0.00001549, 0.00004450, 0.00003625, -0.00000131, -0.00003719, -0.00004330, -0.00001668, 0.00001990, 0.00003791, 0.00002596, -0.00000298, -0.00002451, -0.00002326, -0.00000379, 0.00001574, 0.00001983, 0.00000749, -0.00000928, -0.00001701, -0.00001160, 0.00000025, 0.00000833, 0.00000749, 0.00000038, -0.00000659, -0.00000915, -0.00000766, -0.00000508, -0.00000336, -0.00000202, 0.00000038, 0.00000404, 0.00000728, 0.00000796, 0.00000533, 0.00000061, -0.00000389, -0.00000611, -0.00000514, -0.00000142, 0.00000324, 0.00000610, 0.00000458, -0.00000191, -0.00001092, -0.00001777, -0.00001848, -0.00001272, -0.00000444, 0.00000081, -0.00000031, -0.00000682, -0.00001491, -0.00002125, -0.00002522, -0.00002805, -0.00003005, -0.00002878, -0.00002037, -0.00000293, 0.00002096, 0.00004472, 0.00006112, 0.00006597, 0.00006000, 0.00004843, 0.00003897, },
		{ 0.00008531, 0.00008404, 0.00005943, 0.00001496, -0.00003435, -0.00006851, -0.00007416, -0.00005348, -0.00002315, -0.00000311, -0.00000306, -0.00001681, -0.00002886, -0.00002751, -0.00001374, 0.00000104, 0.00000582, -0.00000062, -0.00000967, -0.00001160, -0.00000451, 0.00000461, 0.00000704, 0.00000077, -0.00000808, -0.00001154, -0.00000746, -0.00000106, 0.00000089, -0.00000306, -0.00000786, -0.00000781, -0.00000296, 0.00000072, -0.00000211, -0.00000959, -0.00001346, -0.00000724, 0.00000609, 0.00001526, 0.00001090, -0.00000450, -0.00001713, -0.00001398, 0.00000412, 0.00002116, 0.00001936, -0.00000382, -0.00003084, -0.00003755, -0.00001444, 0.00002293, 0.00004494, 0.00003234, -0.00000650, -0.00004103, -0.00004320, -0.00001092, 0.00002969, 0.00004555, 0.00002385, -0.00001756, -0.00004520, -0.00003740, -0.00000144, 0.00003356, 0.00004098, 0.00001764, -0.00001546, -0.00003183, -0.00002093, 0.00000515, 0.00002377, 0.00002094, 0.00000137, -0.00001763, -0.00002151, -0.00000954, 0.00000673, 0.00001434, 0.00000909, -0.00000297, -0.00001234, -0.00001383, -0.00000912, -0.00000326, 0.00000043, 0.00000223, 0.00000404, 0.00000643, 0.00000781, 0.00000632, 0.00000191, -0.00000322, -0.00000627, -0.00000576, -0.00000224, 0.00000229, 0.00000545, 0.00000531, 0.00000109, -0.00000617, -0.00001348, -0.00001703, -0.00001467, -0.00000785, -0.00000126, 0.00000025, -0.00000486, -0.00001359, -0.00002090, -0.00002363, -0.00002254, -0.00002062, -0.00001934, -0.00001651, -0.00000791, 0.00000854, 0.00002980, 0.00004875, 0.00005848, 0.00005656, 0.00004638, 0.00003502, 0.00002966, 0.00003468, 0.00004996, 0.00007015, },
		{ 0.00001145, -0.00002907, -0.00005782, -0.00006414, -0.00004892, -0.00002460, -0.00000704, -0.00000480, -0.00001407, -0.00002315, -0.00002261, -0.00001261, -0.00000153, 0.00000235, -0.00000182, -0.00000748, -0.00000761, -0.00000159, 0.00000450, 0.00000413, -0.00000296, -0.00001036, -0.00001117, -0.00000471, 0.00000279, 0.00000418, -0.00000177, -0.00000938, -0.00001175, -0.00000750, -0.00000178, -0.00000062, -0.00000457, -0.00000805, -0.00000543, 0.00000242, 0.00000831, 0.00000586, -0.00000322, -0.00000972, -0.00000570, 0.00000654, 0.00001496, 0.00000858, -0.00001069, -0.00002725, -0.00002441, -0.00000032, 0.00002773, 0.00003621, 0.00001565, -0.00001944, -0.00004089, -0.00003005, 0.00000561, 0.00003746, 0.00003897, 0.00000799, -0.00003080, -0.00004598, -0.00002515, 0.00001502, 0.00004279, 0.00003732, 0.00000494, -0.00002731, -0.00003450, -0.00001360, 0.00001589, 0.00002942, 0.00001752, -0.00000805, -0.00002581, -0.00002270, -0.00000346, 0.00001519, 0.00001910, 0.00000720, -0.00000980, -0.00001949, -0.00001720, -0.00000745, 0.00000182, 0.00000615, 0.00000648, 0.00000606, 0.00000637, 0.00000607, 0.00000335, -0.00000147, -0.00000571, -0.00000663, -0.00000366, 0.00000119, 0.00000492, 0.00000545, 0.00000238, -0.00000319, -0.00000924, -0.00001332, -0.00001349, -0.00000961, -0.00000418, -0.00000133, -0.00000406, -0.00001166, -0.00001970, -0.00002321, -0.00002052, -0.00001454, -0.00000988, -0.00000840, -0.00000706, -0.00000045, 0.00001410, 0.00003317, 0.00004884, 0.00005407, 0.00004765, 0.00003500, 0.00002455, 0.00002281, 0.00003150, 0.00004757, 0.00006466, 0.00007487, 0.00007076, 0.00004852, },
		{ -0.00005635, -0.00004589, -0.00002610, -0.00000972, -0.00000500, -0.00001063, -0.00001807, -0.00001917, -0.00001267, -0.00000409, 0.00000014, -0.00000136, -0.00000445, -0.00000435, -0.00000071, 0.00000216, 0.00000018, -0.00000578, -0.00001014, -0.00000805, -0.00000055, 0.00000578, 0.00000472, -0.00000361, -0.00001252, -0.00001475, -0.00000904, -0.00000106, 0.00000215, -0.00000115, -0.00000614, -0.00000665, -0.00000153, 0.00000426, 0.00000504, 0.00000055, -0.00000353, -0.00000175, 0.00000458, 0.00000775, 0.00000142, -0.00001144, -0.00001927, -0.00001205, 0.00000787, 0.00002518, 0.00002367, 0.00000142, -0.00002521, -0.00003372, -0.00001488, 0.00001778, 0.00003759, 0.00002685, -0.00000743, -0.00003798, -0.00003935, -0.00000924, 0.00002905, 0.00004542, 0.00002756, -0.00000933, -0.00003556, -0.00003140, -0.00000257, 0.00002580, 0.00003064, 0.00000951, -0.00001891, -0.00003140, -0.00001923, 0.00000602, 0.00002352, 0.00002042, 0.00000089, -0.00001921, -0.00002593, -0.00001714, -0.00000167, 0.00000959, 0.00001208, 0.00000877, 0.00000515, 0.00000374, 0.00000300, 0.00000058, -0.00000342, -0.00000623, -0.00000533, -0.00000100, 0.00000372, 0.00000555, 0.00000334, -0.00000153, -0.00000657, -0.00000978, -0.00001030, -0.00000841, -0.00000543, -0.00000363, -0.00000526, -0.00001087, -0.00001802, -0.00002221, -0.00002002, -0.00001218, -0.00000359, 0.00000053, -0.00000073, -0.00000256, 0.00000193, 0.00001562, 0.00003404, 0.00004793, 0.00004984, 0.00003976, 0.00002508, 0.00001555, 0.00001706, 0.00002891, 0.00004545, 0.00005968, 0.00006568, 0.00005950, 0.00003977, 0.00000928, -0.00002398, -0.00004881, },
		{ -0.00001140, -0.00000444, -0.00000729, -0.00001386, -0.00001670, -0.00001298, -0.00000591, -0.00000081, -0.00000010, -0.00000184, -0.00000269, -0.00000174, -0.00000118, -0.00000314, -0.00000655, -0.00000757, -0.00000364, 0.00000305, 0.00000660, 0.00000266, -0.00000698, -0.00001523, -0.00001558, -0.00000798, 0.00000105, 0.00000416, -0.00000022, -0.00000682, -0.00000870, -0.00000380, 0.00000346, 0.00000679, 0.00000427, -0.00000025, -0.00000166, 0.00000070, 0.00000235, -0.00000097, -0.00000760, -0.00001031, -0.00000351, 0.00000956, 0.00001767, 0.00001123, -0.00000749, -0.00002389, -0.00002256, -0.00000161, 0.00002340, 0.00003108, 0.00001268, -0.00001881, -0.00003780, -0.00002712, 0.00000672, 0.00003767, 0.00004104, 0.00001398, -0.00002190, -0.00003821, -0.00002305, 0.00000963, 0.00003222, 0.00002652, -0.00000174, -0.00002862, -0.00003238, -0.00001109, 0.00001696, 0.00002914, 0.00001674, -0.00000938, -0.00002904, -0.00002910, -0.00001216, 0.00000790, 0.00001822, 0.00001592, 0.00000749, 0.00000105, -0.00000043, 0.00000031, -0.00000049, -0.00000327, -0.00000516, -0.00000356, 0.00000079, 0.00000440, 0.00000420, 0.00000021, -0.00000476, -0.00000768, -0.00000751, -0.00000553, -0.00000386, -0.00000411, -0.00000693, -0.00001195, -0.00001757, -0.00002097, -0.00001920, -0.00001150, -0.00000094, 0.00000673, 0.00000722, 0.00000175, -0.00000280, 0.00000126, 0.00001596, 0.00003496, 0.00004744, 0.00004592, 0.00003201, 0.00001543, 0.00000716, 0.00001223, 0.00002735, 0.00004439, 0.00005582, 0.00005779, 0.00004963, 0.00003217, 0.00000751, -0.00001951, -0.00004115, -0.00004991, -0.00004345, -0.00002719, },
		{ -0.00001019, -0.00001452, -0.00001299, -0.00000698, -0.00000117, 0.00000095, -0.00000038, -0.00000256, -0.00000372, -0.00000414, -0.00000480, -0.00000526, -0.00000373, 0.00000046, 0.00000487, 0.00000517, -0.00000083, -0.00001016, -0.00001626, -0.00001428, -0.00000563, 0.00000278, 0.00000436, -0.00000137, -0.00000843, -0.00000978, -0.00000367, 0.00000492, 0.00000907, 0.00000621, 0.00000001, -0.00000367, -0.00000257, 0.00000034, 0.00000075, -0.00000182, -0.00000336, -0.00000013, 0.00000611, 0.00000855, 0.00000206, -0.00001019, -0.00001756, -0.00001109, 0.00000688, 0.00002235, 0.00002068, 0.00000022, -0.00002390, -0.00003101, -0.00001250, 0.00001923, 0.00003943, 0.00003111, -0.00000018, -0.00003001, -0.00003483, -0.00001149, 0.00002007, 0.00003343, 0.00001768, -0.00001373, -0.00003460, -0.00002794, 0.00000026, 0.00002658, 0.00002982, 0.00000787, -0.00002164, -0.00003641, -0.00002690, -0.00000214, 0.00001946, 0.00002504, 0.00001561, 0.00000212, -0.00000516, -0.00000420, -0.00000026, 0.00000093, -0.00000144, -0.00000379, -0.00000284, 0.00000074, 0.00000326, 0.00000194, -0.00000233, -0.00000594, -0.00000613, -0.00000343, -0.00000100, -0.00000185, -0.00000656, -0.00001323, -0.00001895, -0.00002121, -0.00001850, -0.00001072, 0.00000009, 0.00000962, 0.00001301, 0.00000837, -0.00000057, -0.00000561, 0.00000039, 0.00001745, 0.00003698, 0.00004708, 0.00004118, 0.00002327, 0.00000549, -0.00000041, 0.00000904, 0.00002762, 0.00004476, 0.00005285, 0.00005045, 0.00004023, 0.00002496, 0.00000586, -0.00001536, -0.00003416, -0.00004413, -0.00004128, -0.00002804, -0.00001265, -0.00000369, -0.00000422, },
		{ -0.00000785, -0.00000172, 0.00000124, 0.00000001, -0.00000318, -0.00000551, -0.00000587, -0.00000477, -0.00000279, 0.00000001, 0.00000315, 0.00000473, 0.00000237, -0.00000431, -0.00001200, -0.00001547, -0.00001180, -0.00000348, 0.00000310, 0.00000287, -0.00000338, -0.00000941, -0.00000903, -0.00000166, 0.00000705, 0.00001022, 0.00000584, -0.00000165, -0.00000570, -0.00000355, 0.00000168, 0.00000466, 0.00000316, -0.00000019, -0.00000139, 0.00000015, 0.00000095, -0.00000226, -0.00000777, -0.00000934, -0.00000244, 0.00000957, 0.00001646, 0.00000985, -0.00000768, -0.00002241, -0.00002007, 0.00000096, 0.00002601, 0.00003479, 0.00001842, -0.00001176, -0.00003230, -0.00002668, 0.00000041, 0.00002634, 0.00002918, 0.00000619, -0.00002359, -0.00003525, -0.00001886, 0.00001214, 0.00003216, 0.00002462, -0.00000471, -0.00003289, -0.00003860, -0.00001856, 0.00001139, 0.00003010, 0.00002748, 0.00001019, -0.00000636, -0.00001130, -0.00000545, 0.00000233, 0.00000444, 0.00000062, -0.00000380, -0.00000422, -0.00000113, 0.00000138, 0.00000038, -0.00000289, -0.00000466, -0.00000284, 0.00000060, 0.00000133, -0.00000338, -0.00001208, -0.00002022, -0.00002332, -0.00001964, -0.00001043, 0.00000113, 0.00001119, 0.00001604, 0.00001339, 0.00000439, -0.00000537, -0.00000796, 0.00000170, 0.00002093, 0.00003940, 0.00004537, 0.00003435, 0.00001317, -0.00000414, -0.00000607, 0.00000844, 0.00002999, 0.00004612, 0.00004986, 0.00004271, 0.00003071, 0.00001808, 0.00000472, -0.00001095, -0.00002738, -0.00003886, -0.00003958, -0.00002909, -0.00001383, -0.00000283, -0.00000113, -0.00000656, -0.00001227, -0.00001277, },
		{ 0.00000000, -0.00000360, -0.00000640, -0.00000633, -0.00000374, -0.00000025, 0.00000271, 0.00000421, 0.00000332, -0.00000061, -0.00000678, -0.00001229, -0.00001356, -0.00000936, -0.00000245, 0.00000197, 0.00000058, -0.00000492, -0.00000889, -0.00000668, 0.00000102, 0.00000840, 0.00000956, 0.00000383, -0.00000358, -0.00000614, -0.00000187, 0.00000509, 0.00000845, 0.00000549, -0.00000089, -0.00000541, -0.00000551, -0.00000318, -0.00000199, -0.00000264, -0.00000227, 0.00000170, 0.00000732, 0.00000865, 0.00000171, -0.00000984, -0.00001587, -0.00000825, 0.00001040, 0.00002650, 0.00002579, 0.00000617, -0.00001862, -0.00002908, -0.00001621, 0.00000985, 0.00002732, 0.00002080, -0.00000514, -0.00002908, -0.00003049, -0.00000736, 0.00002155, 0.00003195, 0.00001430, -0.00001812, -0.00003992, -0.00003408, -0.00000511, 0.00002548, 0.00003671, 0.00002395, 0.00000032, -0.00001563, -0.00001484, -0.00000277, 0.00000779, 0.00000844, 0.00000099, -0.00000641, -0.00000769, -0.00000364, 0.00000048, 0.00000096, -0.00000106, -0.00000177, 0.00000040, 0.00000262, 0.00000034, -0.00000796, -0.00001859, -0.00002505, -0.00002284, -0.00001247, 0.00000133, 0.00001288, 0.00001828, 0.00001630, 0.00000820, -0.00000231, -0.00000943, -0.00000733, 0.00000598, 0.00002546, 0.00004039, 0.00004079, 0.00002505, 0.00000252, -0.00001202, -0.00000861, 0.00001077, 0.00003387, 0.00004730, 0.00004574, 0.00003404, 0.00002118, 0.00001210, 0.00000472, -0.00000593, -0.00002085, -0.00003445, -0.00003874, -0.00003053, -0.00001484, -0.00000150, 0.00000236, -0.00000276, -0.00001011, -0.00001281, -0.00000914, -0.00000286, 0.00000090, },
		{ -0.00000612, -0.00000268, 0.00000153, 0.00000418, 0.00000420, 0.00000162, -0.00000289, -0.00000798, -0.00001153, -0.00001154, -0.00000779, -0.00000270, 0.00000005, -0.00000141, -0.00000519, -0.00000691, -0.00000363, 0.00000313, 0.00000824, 0.00000745, 0.00000145, -0.00000433, -0.00000442, 0.00000165, 0.00000870, 0.00001036, 0.00000467, -0.00000423, -0.00000982, -0.00000883, -0.00000352, 0.00000119, 0.00000248, 0.00000162, 0.00000143, 0.00000237, 0.00000191, -0.00000196, -0.00000690, -0.00000702, 0.00000140, 0.00001445, 0.00002188, 0.00001534, -0.00000302, -0.00002017, -0.00002213, -0.00000624, 0.00001502, 0.00002348, 0.00001071, -0.00001369, -0.00002927, -0.00002186, 0.00000358, 0.00002609, 0.00002587, 0.00000128, -0.00002893, -0.00004045, -0.00002315, 0.00001073, 0.00003660, 0.00003702, 0.00001460, -0.00001170, -0.00002274, -0.00001382, 0.00000387, 0.00001446, 0.00001069, -0.00000187, -0.00001174, -0.00001200, -0.00000491, 0.00000210, 0.00000412, 0.00000231, 0.00000092, 0.00000162, 0.00000151, -0.00000355, -0.00001366, -0.00002334, -0.00002545, -0.00001691, -0.00000135, 0.00001347, 0.00002095, 0.00001913, 0.00001054, -0.00000017, -0.00000834, -0.00000999, -0.00000280, 0.00001217, 0.00002887, 0.00003805, 0.00003271, 0.00001417, -0.00000689, -0.00001656, -0.00000747, 0.00001536, 0.00003784, 0.00004696, 0.00003988, 0.00002466, 0.00001240, 0.00000776, 0.00000615, -0.00000054, -0.00001510, -0.00003134, -0.00003889, -0.00003210, -0.00001524, 0.00000068, 0.00000635, 0.00000090, -0.00000858, -0.00001365, -0.00001109, -0.00000434, 0.00000057, 0.00000041, -0.00000328, -0.00000640, },
		{ 0.00000487, 0.00000393, 0.00000038, -0.00000422, -0.00000831, -0.00001057, -0.00001014, -0.00000728, -0.00000370, -0.00000172, -0.00000235, -0.00000413, -0.00000418, -0.00000094, 0.00000397, 0.00000673, 0.00000486, -0.00000005, -0.00000325, -0.00000102, 0.00000556, 0.00001092, 0.00000966, 0.00000144, -0.00000829, -0.00001269, -0.00000913, -0.00000113, 0.00000499, 0.00000552, 0.00000188, -0.00000171, -0.00000261, -0.00000166, -0.00000106, -0.00000089, 0.00000126, 0.00000696, 0.00001343, 0.00001447, 0.00000618, -0.00000781, -0.00001736, -0.00001396, 0.00000098, 0.00001554, 0.00001660, 0.00000160, -0.00001787, -0.00002488, -0.00001197, 0.00001122, 0.00002487, 0.00001563, -0.00001110, -0.00003425, -0.00003393, -0.00000803, 0.00002528, 0.00004182, 0.00003045, 0.00000142, -0.00002265, -0.00002515, -0.00000773, 0.00001289, 0.00001980, 0.00000921, -0.00000836, -0.00001852, -0.00001502, -0.00000316, 0.00000681, 0.00000898, 0.00000532, 0.00000138, -0.00000033, -0.00000243, -0.00000854, -0.00001781, -0.00002405, -0.00002068, -0.00000693, 0.00001060, 0.00002243, 0.00002294, 0.00001364, 0.00000102, -0.00000826, -0.00001067, -0.00000588, 0.00000462, 0.00001785, 0.00002884, 0.00003142, 0.00002200, 0.00000374, -0.00001307, -0.00001687, -0.00000324, 0.00002063, 0.00004027, 0.00004426, 0.00003248, 0.00001551, 0.00000535, 0.00000550, 0.00000875, 0.00000456, -0.00001072, -0.00002969, -0.00003966, -0.00003322, -0.00001455, 0.00000373, 0.00001041, 0.00000380, -0.00000821, -0.00001543, -0.00001335, -0.00000549, 0.00000100, 0.00000165, -0.00000230, -0.00000611, -0.00000608, -0.00000225, 0.00000242, },
		{ -0.00000494, -0.00000847, -0.00001004, -0.00000955, -0.00000741, -0.00000468, -0.00000269, -0.00000212, -0.00000228, -0.00000160, 0.00000071, 0.00000353, 0.00000461, 0.00000287, 0.00000000, -0.00000059, 0.00000293, 0.00000843, 0.00001093, 0.00000676, -0.00000269, -0.00001134, -0.00001312, -0.00000697, 0.00000218, 0.00000746, 0.00000566, -0.00000057, -0.00000550, -0.00000548, -0.00000141, 0.00000310, 0.00000564, 0.00000677, 0.00000804, 0.00000898, 0.00000687, 0.00000012, -0.00000836, -0.00001216, -0.00000692, 0.00000439, 0.00001244, 0.00000910, -0.00000454, -0.00001755, -0.00001791, -0.00000362, 0.00001395, 0.00001875, 0.00000420, -0.00001959, -0.00003272, -0.00002184, 0.00000776, 0.00003507, 0.00003973, 0.00001842, -0.00001231, -0.00002941, -0.00002166, 0.00000219, 0.00002158, 0.00002134, 0.00000317, -0.00001732, -0.00002442, -0.00001463, 0.00000246, 0.00001378, 0.00001352, 0.00000578, -0.00000166, -0.00000521, -0.00000740, -0.00001201, -0.00001815, -0.00001987, -0.00001191, 0.00000421, 0.00001988, 0.00002565, 0.00001847, 0.00000383, -0.00000879, -0.00001278, -0.00000779, 0.00000209, 0.00001241, 0.00002036, 0.00002398, 0.00002111, 0.00001075, -0.00000387, -0.00001472, -0.00001329, 0.00000244, 0.00002467, 0.00004002, 0.00003929, 0.00002460, 0.00000781, 0.00000067, 0.00000506, 0.00001168, 0.00000852, -0.00000802, -0.00002913, -0.00004033, -0.00003327, -0.00001271, 0.00000717, 0.00001378, 0.00000532, -0.00000910, -0.00001769, -0.00001515, -0.00000557, 0.00000251, 0.00000351, -0.00000132, -0.00000631, -0.00000677, -0.00000252, 0.00000286, 0.00000551, 0.00000406, -0.00000019, },
		{ -0.00000940, -0.00000753, -0.00000504, -0.00000270, -0.00000115, -0.00000040, 0.00000020, 0.00000124, 0.00000241, 0.00000280, 0.00000212, 0.00000151, 0.00000275, 0.00000619, 0.00000947, 0.00000891, 0.00000281, -0.00000621, -0.00001243, -0.00001138, -0.00000377, 0.00000467, 0.00000763, 0.00000338, -0.00000394, -0.00000793, -0.00000526, 0.00000192, 0.00000840, 0.00001050, 0.00000861, 0.00000559, 0.00000329, 0.00000107, -0.00000233, -0.00000600, -0.00000674, -0.00000251, 0.00000430, 0.00000762, 0.00000308, -0.00000694, -0.00001404, -0.00001091, 0.00000117, 0.00001184, 0.00001008, -0.00000517, -0.00002217, -0.00002495, -0.00000725, 0.00002048, 0.00003792, 0.00003100, 0.00000400, -0.00002305, -0.00003005, -0.00001295, 0.00001325, 0.00002700, 0.00001763, -0.00000663, -0.00002633, -0.00002689, -0.00000958, 0.00001124, 0.00002085, 0.00001530, 0.00000220, -0.00000803, -0.00001120, -0.00001078, -0.00001167, -0.00001375, -0.00001182, -0.00000187, 0.00001309, 0.00002389, 0.00002253, 0.00000929, -0.00000693, -0.00001548, -0.00001191, -0.00000024, 0.00001145, 0.00001760, 0.00001783, 0.00001454, 0.00000917, 0.00000163, -0.00000690, -0.00001181, -0.00000742, 0.00000743, 0.00002596, 0.00003692, 0.00003307, 0.00001765, 0.00000244, -0.00000175, 0.00000558, 0.00001393, 0.00001088, -0.00000668, -0.00002884, -0.00004013, -0.00003207, -0.00001020, 0.00001014, 0.00001576, 0.00000534, -0.00001074, -0.00001955, -0.00001569, -0.00000426, 0.00000481, 0.00000528, -0.00000111, -0.00000748, -0.00000817, -0.00000300, 0.00000355, 0.00000666, 0.00000472, -0.00000042, -0.00000567, -0.00000901, -0.00001009, },
		{ -0.00000212, -0.00000011, 0.00000091, 0.00000110, 0.00000109, 0.00000136, 0.00000191, 0.00000263, 0.00000379, 0.00000575, 0.00000799, 0.00000867, 0.00000574, -0.00000088, -0.00000818, -0.00001158, -0.00000852, -0.00000103, 0.00000530, 0.00000566, 0.00000006, -0.00000645, -0.00000786, -0.00000224, 0.00000671, 0.00001280, 0.00001246, 0.00000715, 0.00000134, -0.00000164, -0.00000196, -0.00000187, -0.00000253, -0.00000277, -0.00000100, 0.00000202, 0.00000306, -0.00000026, -0.00000619, -0.00000941, -0.00000602, 0.00000181, 0.00000651, 0.00000186, -0.00001002, -0.00001862, -0.00001335, 0.00000614, 0.00002734, 0.00003365, 0.00001818, -0.00000913, -0.00002818, -0.00002452, -0.00000144, 0.00002221, 0.00002707, 0.00000897, -0.00001784, -0.00003247, -0.00002418, -0.00000023, 0.00002103, 0.00002528, 0.00001248, -0.00000544, -0.00001594, -0.00001559, -0.00001012, -0.00000648, -0.00000554, -0.00000253, 0.00000613, 0.00001708, 0.00002181, 0.00001455, -0.00000139, -0.00001516, -0.00001691, -0.00000579, 0.00000972, 0.00001908, 0.00001792, 0.00001004, 0.00000234, -0.00000158, -0.00000316, -0.00000487, -0.00000571, -0.00000151, 0.00000984, 0.00002398, 0.00003186, 0.00002718, 0.00001283, -0.00000042, -0.00000273, 0.00000587, 0.00001479, 0.00001179, -0.00000587, -0.00002790, -0.00003871, -0.00003003, -0.00000794, 0.00001179, 0.00001610, 0.00000435, -0.00001215, -0.00002010, -0.00001460, -0.00000190, 0.00000709, 0.00000610, -0.00000217, -0.00000954, -0.00000971, -0.00000299, 0.00000498, 0.00000838, 0.00000553, -0.00000090, -0.00000689, -0.00001005, -0.00001041, -0.00000912, -0.00000709, -0.00000464, },
		{ 0.00000143, 0.00000091, 0.00000095, 0.00000203, 0.00000390, 0.00000599, 0.00000769, 0.00000824, 0.00000669, 0.00000249, -0.00000346, -0.00000849, -0.00000958, -0.00000581, 0.00000026, 0.00000405, 0.00000258, -0.00000275, -0.00000691, -0.00000525, 0.00000237, 0.00001112, 0.00001496, 0.00001148, 0.00000365, -0.00000290, -0.00000462, -0.00000240, 0.00000014, 0.00000044, -0.00000092, -0.00000166, -0.00000081, 0.00000012, -0.00000095, -0.00000381, -0.00000569, -0.00000428, -0.00000079, 0.00000048, -0.00000333, -0.00000947, -0.00001058, -0.00000130, 0.00001517, 0.00002735, 0.00002410, 0.00000484, -0.00001781, -0.00002685, -0.00001469, 0.00000956, 0.00002638, 0.00002137, -0.00000264, -0.00002732, -0.00003342, -0.00001611, 0.00001140, 0.00002883, 0.00002480, 0.00000469, -0.00001543, -0.00002250, -0.00001570, -0.00000425, 0.00000271, 0.00000391, 0.00000475, 0.00000944, 0.00001527, 0.00001494, 0.00000472, -0.00001014, -0.00001851, -0.00001286, 0.00000358, 0.00001894, 0.00002235, 0.00001264, -0.00000154, -0.00000979, -0.00000856, -0.00000261, 0.00000122, 0.00000132, 0.00000224, 0.00000875, 0.00001939, 0.00002646, 0.00002309, 0.00001061, -0.00000146, -0.00000352, 0.00000500, 0.00001418, 0.00001200, -0.00000457, -0.00002578, -0.00003635, -0.00002811, -0.00000693, 0.00001173, 0.00001524, 0.00000342, -0.00001229, -0.00001886, -0.00001221, 0.00000060, 0.00000837, 0.00000542, -0.00000437, -0.00001184, -0.00001063, -0.00000197, 0.00000716, 0.00001024, 0.00000590, -0.00000208, -0.00000866, -0.00001123, -0.00001042, -0.00000824, -0.00000599, -0.00000379, -0.00000150, 0.00000050, 0.00000153, },
		{ 0.00000277, 0.00000522, 0.00000747, 0.00000839, 0.00000738, 0.00000439, -0.00000005, -0.00000470, -0.00000769, -0.00000747, -0.00000416, -0.00000008, 0.00000167, -0.00000028, -0.00000391, -0.00000508, -0.00000098, 0.00000701, 0.00001382, 0.00001452, 0.00000839, -0.00000033, -0.00000585, -0.00000542, -0.00000115, 0.00000229, 0.00000202, -0.00000095, -0.00000326, -0.00000286, -0.00000085, 0.00000003, -0.00000153, -0.00000412, -0.00000537, -0.00000469, -0.00000380, -0.00000418, -0.00000443, -0.00000093, 0.00000797, 0.00001821, 0.00002157, 0.00001246, -0.00000558, -0.00002044, -0.00002021, -0.00000367, 0.00001694, 0.00002459, 0.00001137, -0.00001404, -0.00003222, -0.00002833, -0.00000433, 0.00002221, 0.00003187, 0.00001853, -0.00000659, -0.00002478, -0.00002475, -0.00001008, 0.00000615, 0.00001336, 0.00001130, 0.00000729, 0.00000705, 0.00000878, 0.00000610, -0.00000361, -0.00001448, -0.00001647, -0.00000503, 0.00001316, 0.00002460, 0.00002008, 0.00000279, -0.00001425, -0.00001904, -0.00001027, 0.00000273, 0.00000925, 0.00000685, 0.00000244, 0.00000443, 0.00001370, 0.00002240, 0.00002161, 0.00001055, -0.00000194, -0.00000527, 0.00000265, 0.00001276, 0.00001262, -0.00000206, -0.00002259, -0.00003397, -0.00002738, -0.00000772, 0.00001030, 0.00001426, 0.00000370, -0.00001053, -0.00001610, -0.00000947, 0.00000215, 0.00000801, 0.00000335, -0.00000701, -0.00001351, -0.00001034, 0.00000005, 0.00000962, 0.00001159, 0.00000538, -0.00000398, -0.00001060, -0.00001200, -0.00000970, -0.00000666, -0.00000453, -0.00000301, -0.00000131, 0.00000053, 0.00000171, 0.00000172, 0.00000117, 0.00000128, },
		{ 0.00000819, 0.00000611, 0.00000245, -0.00000164, -0.00000500, -0.00000663, -0.00000606, -0.00000382, -0.00000148, -0.00000074, -0.00000186, -0.00000304, -0.00000156, 0.00000365, 0.00001038, 0.00001424, 0.00001197, 0.00000445, -0.00000357, -0.00000704, -0.00000456, 0.00000063, 0.00000365, 0.00000212, -0.00000199, -0.00000463, -0.00000356, -0.00000037, 0.00000133, -0.00000059, -0.00000477, -0.00000789, -0.00000772, -0.00000479, -0.00000098, 0.00000283, 0.00000729, 0.00001242, 0.00001559, 0.00001262, 0.00000215, -0.00001092, -0.00001740, -0.00001095, 0.00000522, 0.00001885, 0.00001764, -0.00000003, -0.00002205, -0.00003104, -0.00001832, 0.00000813, 0.00002902, 0.00002870, 0.00000754, -0.00001838, -0.00003022, -0.00002082, 0.00000088, 0.00001858, 0.00002205, 0.00001377, 0.00000397, -0.00000002, 0.00000033, -0.00000141, -0.00000789, -0.00001369, -0.00001044, 0.00000375, 0.00002006, 0.00002513, 0.00001283, -0.00000927, -0.00002518, -0.00002347, -0.00000674, 0.00001103, 0.00001671, 0.00000912, -0.00000105, -0.00000181, 0.00000873, 0.00002083, 0.00002262, 0.00001149, -0.00000316, -0.00000860, -0.00000075, 0.00001160, 0.00001454, 0.00000177, -0.00001915, -0.00003272, -0.00002858, -0.00001013, 0.00000853, 0.00001444, 0.00000598, -0.00000703, -0.00001279, -0.00000759, 0.00000197, 0.00000602, 0.00000060, -0.00000911, -0.00001388, -0.00000882, 0.00000256, 0.00001161, 0.00001190, 0.00000391, -0.00000620, -0.00001212, -0.00001190, -0.00000815, -0.00000469, -0.00000320, -0.00000273, -0.00000169, 0.00000022, 0.00000190, 0.00000238, 0.00000199, 0.00000208, 0.00000358, 0.00000606, 0.00000805, },
		{ -0.00000256, -0.00000503, -0.00000599, -0.00000561, -0.00000439, -0.00000305, -0.00000220, -0.00000172, -0.00000052, 0.00000259, 0.00000744, 0.00001181, 0.00001259, 0.00000829, 0.00000080, -0.00000543, -0.00000668, -0.00000292, 0.00000208, 0.00000391, 0.00000126, -0.00000314, -0.00000515, -0.00000313, 0.00000056, 0.00000183, -0.00000130, -0.00000671, -0.00001004, -0.00000835, -0.00000238, 0.00000459, 0.00000961, 0.00001185, 0.00001176, 0.00000926, 0.00000366, -0.00000430, -0.00001084, -0.00001079, -0.00000229, 0.00000964, 0.00001533, 0.00000793, -0.00000953, -0.00002454, -0.00002415, -0.00000607, 0.00001785, 0.00002974, 0.00001977, -0.00000541, -0.00002718, -0.00002937, -0.00001069, 0.00001482, 0.00002951, 0.00002568, 0.00000997, -0.00000432, -0.00000911, -0.00000678, -0.00000501, -0.00000725, -0.00000892, -0.00000303, 0.00001061, 0.00002246, 0.00002067, 0.00000260, -0.00002069, -0.00003209, -0.00002260, 0.00000064, 0.00001987, 0.00002153, 0.00000748, -0.00000725, -0.00000814, 0.00000593, 0.00002202, 0.00002517, 0.00001204, -0.00000600, -0.00001330, -0.00000418, 0.00001176, 0.00001809, 0.00000625, -0.00001664, -0.00003353, -0.00003174, -0.00001319, 0.00000781, 0.00001676, 0.00001030, -0.00000271, -0.00001023, -0.00000754, -0.00000006, 0.00000313, -0.00000173, -0.00000986, -0.00001282, -0.00000658, 0.00000473, 0.00001253, 0.00001109, 0.00000193, -0.00000808, -0.00001271, -0.00001084, -0.00000611, -0.00000285, -0.00000244, -0.00000307, -0.00000240, -0.00000003, 0.00000248, 0.00000351, 0.00000314, 0.00000288, 0.00000402, 0.00000626, 0.00000797, 0.00000761, 0.00000498, 0.00000110, },
		{ -0.00000583, -0.00000513, -0.00000394, -0.00000228, -0.00000008, 0.00000280, 0.00000630, 0.00000967, 0.00001131, 0.00000960, 0.00000446, -0.00000187, -0.00000589, -0.00000535, -0.00000125, 0.00000277, 0.00000332, 0.00000010, -0.00000380, -0.00000483, -0.00000225, 0.00000104, 0.00000114, -0.00000306, -0.00000853, -0.00001043, -0.00000620, 0.00000228, 0.00001024, 0.00001372, 0.00001206, 0.00000734, 0.00000202, -0.00000264, -0.00000599, -0.00000678, -0.00000353, 0.00000312, 0.00000895, 0.00000818, -0.00000137, -0.00001455, -0.00002117, -0.00001375, 0.00000498, 0.00002221, 0.00002415, 0.00000743, -0.00001670, -0.00003010, -0.00002166, 0.00000359, 0.00002815, 0.00003553, 0.00002258, 0.00000063, -0.00001492, -0.00001685, -0.00000989, -0.00000370, -0.00000268, -0.00000260, 0.00000295, 0.00001360, 0.00002016, 0.00001284, -0.00000797, -0.00002908, -0.00003406, -0.00001735, 0.00000963, 0.00002699, 0.00002262, 0.00000249, -0.00001451, -0.00001292, 0.00000595, 0.00002532, 0.00002789, 0.00001110, -0.00001058, -0.00001854, -0.00000650, 0.00001381, 0.00002285, 0.00001024, -0.00001615, -0.00003666, -0.00003606, -0.00001552, 0.00000928, 0.00002145, 0.00001582, 0.00000106, -0.00000956, -0.00000960, -0.00000328, 0.00000049, -0.00000270, -0.00000901, -0.00001083, -0.00000453, 0.00000584, 0.00001220, 0.00000958, 0.00000015, -0.00000907, -0.00001225, -0.00000915, -0.00000411, -0.00000158, -0.00000236, -0.00000380, -0.00000307, 0.00000011, 0.00000349, 0.00000489, 0.00000421, 0.00000333, 0.00000397, 0.00000601, 0.00000763, 0.00000709, 0.00000415, 0.00000013, -0.00000331, -0.00000532, -0.00000600, },
		{ -0.00000107, 0.00000241, 0.00000599, 0.00000886, 0.00001020, 0.00000936, 0.00000609, 0.00000119, -0.00000336, -0.00000528, -0.00000367, 0.00000002, 0.00000276, 0.00000231, -0.00000086, -0.00000389, -0.00000411, -0.00000160, 0.00000069, -0.00000036, -0.00000478, -0.00000900, -0.00000857, -0.00000207, 0.00000728, 0.00001394, 0.00001421, 0.00000874, 0.00000150, -0.00000356, -0.00000500, -0.00000374, -0.00000123, 0.00000167, 0.00000412, 0.00000428, 0.00000023, -0.00000744, -0.00001410, -0.00001347, -0.00000306, 0.00001183, 0.00002028, 0.00001400, -0.00000480, -0.00002308, -0.00002595, -0.00000866, 0.00001852, 0.00003717, 0.00003456, 0.00001318, -0.00001170, -0.00002440, -0.00002041, -0.00000785, 0.00000164, 0.00000348, 0.00000248, 0.00000548, 0.00001227, 0.00001445, 0.00000405, -0.00001646, -0.00003320, -0.00003155, -0.00000962, 0.00001799, 0.00003096, 0.00002007, -0.00000446, -0.00002109, -0.00001514, 0.00000853, 0.00002952, 0.00002952, 0.00000821, -0.00001631, -0.00002318, -0.00000692, 0.00001761, 0.00002779, 0.00001255, -0.00001819, -0.00004153, -0.00004017, -0.00001577, 0.00001346, 0.00002786, 0.00002115, 0.00000294, -0.00001130, -0.00001324, -0.00000650, -0.00000079, -0.00000187, -0.00000699, -0.00000885, -0.00000352, 0.00000560, 0.00001101, 0.00000812, -0.00000080, -0.00000894, -0.00001104, -0.00000739, -0.00000264, -0.00000103, -0.00000275, -0.00000454, -0.00000336, 0.00000068, 0.00000474, 0.00000616, 0.00000494, 0.00000338, 0.00000364, 0.00000565, 0.00000736, 0.00000671, 0.00000344, -0.00000088, -0.00000431, -0.00000603, -0.00000640, -0.00000611, -0.00000534, -0.00000375, },
		{ 0.00001001, 0.00000930, 0.00000659, 0.00000271, -0.00000120, -0.00000378, -0.00000404, -0.00000203, 0.00000079, 0.00000229, 0.00000124, -0.00000151, -0.00000365, -0.00000346, -0.00000149, -0.00000026, -0.00000185, -0.00000549, -0.00000767, -0.00000500, 0.00000251, 0.00001083, 0.00001465, 0.00001153, 0.00000379, -0.00000341, -0.00000600, -0.00000374, 0.00000043, 0.00000323, 0.00000331, 0.00000123, -0.00000200, -0.00000570, -0.00000880, -0.00000905, -0.00000427, 0.00000480, 0.00001295, 0.00001313, 0.00000246, -0.00001347, -0.00002273, -0.00001568, 0.00000639, 0.00003021, 0.00003934, 0.00002656, -0.00000016, -0.00002347, -0.00002965, -0.00001841, -0.00000139, 0.00000874, 0.00000847, 0.00000427, 0.00000397, 0.00000767, 0.00000749, -0.00000339, -0.00002146, -0.00003313, -0.00002606, -0.00000151, 0.00002408, 0.00003145, 0.00001490, -0.00001173, -0.00002574, -0.00001467, 0.00001272, 0.00003333, 0.00002933, 0.00000365, -0.00002217, -0.00002625, -0.00000526, 0.00002237, 0.00003171, 0.00001240, -0.00002250, -0.00004692, -0.00004262, -0.00001316, 0.00001992, 0.00003466, 0.00002481, 0.00000218, -0.00001509, -0.00001726, -0.00000845, -0.00000007, 0.00000044, -0.00000476, -0.00000787, -0.00000397, 0.00000431, 0.00000972, 0.00000746, -0.00000059, -0.00000798, -0.00000966, -0.00000609, -0.00000188, -0.00000099, -0.00000321, -0.00000497, -0.00000325, 0.00000147, 0.00000588, 0.00000710, 0.00000532, 0.00000325, 0.00000332, 0.00000541, 0.00000715, 0.00000623, 0.00000248, -0.00000221, -0.00000563, -0.00000691, -0.00000669, -0.00000589, -0.00000467, -0.00000251, 0.00000090, 0.00000498, 0.00000841, },
		{ -0.00000018, -0.00000266, -0.00000340, -0.00000250, -0.00000060, 0.00000114, 0.00000155, 0.00000022, -0.00000197, -0.00000339, -0.00000309, -0.00000174, -0.00000119, -0.00000256, -0.00000474, -0.00000490, -0.00000096, 0.00000601, 0.00001191, 0.00001257, 0.00000718, -0.00000071, -0.00000584, -0.00000525, -0.00000042, 0.00000420, 0.00000492, 0.00000141, -0.00000367, -0.00000724, -0.00000783, -0.00000571, -0.00000169, 0.00000339, 0.00000784, 0.00000871, 0.00000342, -0.00000698, -0.00001624, -0.00001591, -0.00000208, 0.00001958, 0.00003568, 0.00003408, 0.00001359, -0.00001387, -0.00003130, -0.00002898, -0.00001139, 0.00000724, 0.00001476, 0.00001011, 0.00000205, -0.00000070, 0.00000179, 0.00000151, -0.00000798, -0.00002275, -0.00003001, -0.00001949, 0.00000526, 0.00002722, 0.00002905, 0.00000852, -0.00001795, -0.00002797, -0.00001208, 0.00001735, 0.00003580, 0.00002724, -0.00000180, -0.00002716, -0.00002732, -0.00000202, 0.00002695, 0.00003365, 0.00000975, -0.00002808, -0.00005134, -0.00004238, -0.00000781, 0.00002750, 0.00004028, 0.00002577, -0.00000112, -0.00001983, -0.00002027, -0.00000826, 0.00000251, 0.00000331, -0.00000341, -0.00000850, -0.00000572, 0.00000274, 0.00000918, 0.00000800, 0.00000056, -0.00000676, -0.00000872, -0.00000552, -0.00000167, -0.00000106, -0.00000339, -0.00000501, -0.00000294, 0.00000214, 0.00000669, 0.00000774, 0.00000560, 0.00000323, 0.00000318, 0.00000521, 0.00000672, 0.00000533, 0.00000106, -0.00000386, -0.00000697, -0.00000750, -0.00000639, -0.00000487, -0.00000317, -0.00000065, 0.00000303, 0.00000705, 0.00000976, 0.00000988, 0.00000742, 0.00000356, },
		{ -0.00000083, 0.00000055, 0.00000111, 0.00000058, -0.00000083, -0.00000239, -0.00000320, -0.00000283, -0.00000187, -0.00000149, -0.00000217, -0.00000288, -0.00000172, 0.00000225, 0.00000749, 0.00001060, 0.00000890, 0.00000298, -0.00000328, -0.00000552, -0.00000246, 0.00000294, 0.00000587, 0.00000353, -0.00000263, -0.00000824, -0.00000952, -0.00000590, 0.00000009, 0.00000517, 0.00000725, 0.00000581, 0.00000118, -0.00000556, -0.00001168, -0.00001272, -0.00000483, 0.00001112, 0.00002762, 0.00003378, 0.00002296, -0.00000092, -0.00002437, -0.00003312, -0.00002261, -0.00000148, 0.00001513, 0.00001736, 0.00000752, -0.00000343, -0.00000665, -0.00000323, -0.00000199, -0.00000935, -0.00002122, -0.00002548, -0.00001348, 0.00000982, 0.00002763, 0.00002488, 0.00000227, -0.00002238, -0.00002798, -0.00000835, 0.00002138, 0.00003656, 0.00002373, -0.00000721, -0.00003061, -0.00002654, 0.00000189, 0.00003036, 0.00003326, 0.00000529, -0.00003355, -0.00005362, -0.00003922, -0.00000063, 0.00003463, 0.00004345, 0.00002384, -0.00000605, -0.00002407, -0.00002119, -0.00000582, 0.00000614, 0.00000555, -0.00000376, -0.00001072, -0.00000803, 0.00000183, 0.00000995, 0.00000967, 0.00000207, -0.00000597, -0.00000855, -0.00000554, -0.00000158, -0.00000086, -0.00000321, -0.00000491, -0.00000280, 0.00000247, 0.00000723, 0.00000836, 0.00000611, 0.00000347, 0.00000308, 0.00000469, 0.00000571, 0.00000383, -0.00000072, -0.00000545, -0.00000785, -0.00000738, -0.00000532, -0.00000316, -0.00000119, 0.00000135, 0.00000485, 0.00000834, 0.00001003, 0.00000875, 0.00000499, 0.00000065, -0.00000238, -0.00000327, -0.00000241, },
		{ -0.00000067, -0.00000194, -0.00000274, -0.00000284, -0.00000227, -0.00000145, -0.00000098, -0.00000098, -0.00000073, 0.00000080, 0.00000382, 0.00000694, 0.00000782, 0.00000517, 0.00000027, -0.00000349, -0.00000328, 0.00000064, 0.00000473, 0.00000491, 0.00000005, -0.00000679, -0.00001057, -0.00000826, -0.00000117, 0.00000613, 0.00000920, 0.00000660, 0.00000018, -0.00000670, -0.00001102, -0.00001067, -0.00000445, 0.00000714, 0.00002048, 0.00002883, 0.00002549, 0.00000911, -0.00001328, -0.00002915, -0.00002872, -0.00001248, 0.00000834, 0.00001972, 0.00001544, 0.00000133, -0.00001035, -0.00001176, -0.00000585, -0.00000254, -0.00000817, -0.00001828, -0.00002105, -0.00000899, 0.00001211, 0.00002613, 0.00002013, -0.00000298, -0.00002490, -0.00002645, -0.00000442, 0.00002426, 0.00003576, 0.00001952, -0.00001186, -0.00003241, -0.00002454, 0.00000552, 0.00003203, 0.00003090, 0.00000015, -0.00003766, -0.00005320, -0.00003375, 0.00000694, 0.00003991, 0.00004366, 0.00001966, -0.00001118, -0.00002653, -0.00001960, -0.00000177, 0.00000959, 0.00000615, -0.00000603, -0.00001392, -0.00000990, 0.00000230, 0.00001209, 0.00001192, 0.00000322, -0.00000604, -0.00000910, -0.00000575, -0.00000120, -0.00000025, -0.00000289, -0.00000503, -0.00000308, 0.00000252, 0.00000784, 0.00000934, 0.00000699, 0.00000379, 0.00000259, 0.00000344, 0.00000393, 0.00000188, -0.00000241, -0.00000647, -0.00000791, -0.00000643, -0.00000361, -0.00000110, 0.00000086, 0.00000312, 0.00000611, 0.00000881, 0.00000938, 0.00000689, 0.00000233, -0.00000191, -0.00000382, -0.00000308, -0.00000096, 0.00000082, 0.00000133, 0.00000062, },
		{ -0.00000194, -0.00000110, -0.00000042, -0.00000001, 0.00000026, 0.00000078, 0.00000199, 0.00000377, 0.00000517, 0.00000491, 0.00000259, -0.00000047, -0.00000199, -0.00000058, 0.00000257, 0.00000429, 0.00000197, -0.00000384, -0.00000926, -0.00000985, -0.00000426, 0.00000429, 0.00001017, 0.00000937, 0.00000234, -0.00000656, -0.00001220, -0.00001160, -0.00000487, 0.00000565, 0.00001659, 0.00002390, 0.00002357, 0.00001352, -0.00000379, -0.00002064, -0.00002755, -0.00001971, -0.00000152, 0.00001528, 0.00001961, 0.00000941, -0.00000672, -0.00001656, -0.00001442, -0.00000553, -0.00000071, -0.00000573, -0.00001531, -0.00001769, -0.00000623, 0.00001269, 0.00002368, 0.00001565, -0.00000691, -0.00002588, -0.00002413, -0.00000093, 0.00002586, 0.00003386, 0.00001530, -0.00001541, -0.00003287, -0.00002207, 0.00000823, 0.00003200, 0.00002741, -0.00000447, -0.00003968, -0.00005036, -0.00002714, 0.00001347, 0.00004252, 0.00004124, 0.00001450, -0.00001512, -0.00002650, -0.00001590, 0.00000274, 0.00001170, 0.00000463, -0.00000976, -0.00001710, -0.00001046, 0.00000439, 0.00001514, 0.00001396, 0.00000343, -0.00000704, -0.00000998, -0.00000566, -0.00000030, 0.00000060, -0.00000281, -0.00000567, -0.00000372, 0.00000265, 0.00000893, 0.00001086, 0.00000803, 0.00000367, 0.00000121, 0.00000124, 0.00000159, 0.00000001, -0.00000344, -0.00000653, -0.00000707, -0.00000488, -0.00000165, 0.00000095, 0.00000270, 0.00000449, 0.00000675, 0.00000847, 0.00000794, 0.00000453, -0.00000030, -0.00000389, -0.00000439, -0.00000214, 0.00000077, 0.00000222, 0.00000153, -0.00000041, -0.00000218, -0.00000295, -0.00000271, },
		{ 0.00000071, 0.00000077, 0.00000118, 0.00000193, 0.00000280, 0.00000330, 0.00000290, 0.00000160, 0.00000025, 0.00000003, 0.00000126, 0.00000265, 0.00000206, -0.00000153, -0.00000650, -0.00000916, -0.00000655, 0.00000077, 0.00000841, 0.00001098, 0.00000612, -0.00000349, -0.00001186, -0.00001370, -0.00000766, 0.00000335, 0.00001441, 0.00002117, 0.00002120, 0.00001402, 0.00000132, -0.00001270, -0.00002184, -0.00002073, -0.00000889, 0.00000731, 0.00001762, 0.00001491, 0.00000090, -0.00001441, -0.00002030, -0.00001395, -0.00000274, 0.00000232, -0.00000333, -0.00001325, -0.00001567, -0.00000484, 0.00001231, 0.00002105, 0.00001187, -0.00000964, -0.00002585, -0.00002163, 0.00000183, 0.00002638, 0.00003138, 0.00001144, -0.00001793, -0.00003249, -0.00001973, 0.00000987, 0.00003081, 0.00002376, -0.00000778, -0.00003956, -0.00004595, -0.00002075, 0.00001798, 0.00004247, 0.00003717, 0.00000975, -0.00001692, -0.00002405, -0.00001107, 0.00000643, 0.00001173, 0.00000121, -0.00001402, -0.00001924, -0.00000924, 0.00000781, 0.00001834, 0.00001512, 0.00000250, -0.00000864, -0.00001068, -0.00000497, 0.00000099, 0.00000127, -0.00000332, -0.00000683, -0.00000435, 0.00000335, 0.00001077, 0.00001273, 0.00000869, 0.00000253, -0.00000134, -0.00000170, -0.00000071, -0.00000110, -0.00000336, -0.00000558, -0.00000561, -0.00000315, 0.00000021, 0.00000279, 0.00000426, 0.00000544, 0.00000676, 0.00000735, 0.00000581, 0.00000190, -0.00000260, -0.00000504, -0.00000403, -0.00000062, 0.00000247, 0.00000305, 0.00000099, -0.00000194, -0.00000368, -0.00000343, -0.00000189, -0.00000028, 0.00000061, 0.00000079, },
		{ 0.00000130, 0.00000191, 0.00000222, 0.00000214, 0.00000177, 0.00000142, 0.00000142, 0.00000161, 0.00000118, -0.00000081, -0.00000416, -0.00000701, -0.00000671, -0.00000206, 0.00000498, 0.00000989, 0.00000867, 0.00000105, -0.00000860, -0.00001399, -0.00001113, -0.00000101, 0.00001114, 0.00001936, 0.00002024, 0.00001397, 0.00000332, -0.00000797, -0.00001603, -0.00001757, -0.00001120, 0.00000075, 0.00001195, 0.00001510, 0.00000704, -0.00000786, -0.00001983, -0.00002073, -0.00001073, 0.00000137, 0.00000523, -0.00000190, -0.00001240, -0.00001472, -0.00000420, 0.00001162, 0.00001864, 0.00000879, -0.00001149, -0.00002528, -0.00001926, 0.00000387, 0.00002613, 0.00002866, 0.00000805, -0.00001967, -0.00003172, -0.00001775, 0.00001069, 0.00002918, 0.00002073, -0.00000953, -0.00003786, -0.00004113, -0.00001561, 0.00002015, 0.00004044, 0.00003276, 0.00000652, -0.00001634, -0.00001993, -0.00000635, 0.00000838, 0.00000960, -0.00000336, -0.00001776, -0.00001966, -0.00000633, 0.00001186, 0.00002089, 0.00001502, 0.00000063, -0.00001031, -0.00001078, -0.00000368, 0.00000231, 0.00000137, -0.00000451, -0.00000817, -0.00000445, 0.00000496, 0.00001320, 0.00001436, 0.00000827, 0.00000001, -0.00000488, -0.00000475, -0.00000222, -0.00000095, -0.00000212, -0.00000397, -0.00000403, -0.00000162, 0.00000182, 0.00000441, 0.00000556, 0.00000592, 0.00000604, 0.00000542, 0.00000315, -0.00000066, -0.00000419, -0.00000516, -0.00000281, 0.00000114, 0.00000370, 0.00000299, -0.00000031, -0.00000363, -0.00000463, -0.00000293, -0.00000007, 0.00000197, 0.00000230, 0.00000144, 0.00000049, 0.00000021, 0.00000061, },
		{ 0.00000224, 0.00000222, 0.00000206, 0.00000161, 0.00000068, -0.00000090, -0.00000306, -0.00000504, -0.00000544, -0.00000299, 0.00000202, 0.00000710, 0.00000861, 0.00000443, -0.00000387, -0.00001132, -0.00001260, -0.00000582, 0.00000584, 0.00001610, 0.00001956, 0.00001488, 0.00000494, -0.00000550, -0.00001247, -0.00001399, -0.00000995, -0.00000189, 0.00000690, 0.00001177, 0.00000884, -0.00000183, -0.00001482, -0.00002191, -0.00001804, -0.00000584, 0.00000546, 0.00000708, -0.00000177, -0.00001252, -0.00001431, -0.00000378, 0.00001096, 0.00001653, 0.00000627, -0.00001276, -0.00002443, -0.00001713, 0.00000529, 0.00002531, 0.00002583, 0.00000505, -0.00002089, -0.00003074, -0.00001602, 0.00001120, 0.00002773, 0.00001861, -0.00001002, -0.00003545, -0.00003686, -0.00001220, 0.00002037, 0.00003750, 0.00002914, 0.00000532, -0.00001381, -0.00001523, -0.00000280, 0.00000822, 0.00000583, -0.00000806, -0.00002013, -0.00001821, -0.00000226, 0.00001572, 0.00002226, 0.00001368, -0.00000171, -0.00001157, -0.00001017, -0.00000211, 0.00000322, 0.00000072, -0.00000611, -0.00000914, -0.00000359, 0.00000742, 0.00001567, 0.00001501, 0.00000629, -0.00000384, -0.00000878, -0.00000704, -0.00000230, 0.00000060, -0.00000007, -0.00000225, -0.00000278, -0.00000045, 0.00000323, 0.00000594, 0.00000659, 0.00000577, 0.00000443, 0.00000273, 0.00000027, -0.00000269, -0.00000472, -0.00000420, -0.00000108, 0.00000259, 0.00000402, 0.00000200, -0.00000198, -0.00000485, -0.00000446, -0.00000120, 0.00000249, 0.00000415, 0.00000318, 0.00000087, -0.00000090, -0.00000114, -0.00000018, 0.00000103, 0.00000183, 0.00000216, },
		{ 0.00000079, -0.00000072, -0.00000249, -0.00000390, -0.00000416, -0.00000266, 0.00000061, 0.00000452, 0.00000685, 0.00000543, -0.00000006, -0.00000705, -0.00001107, -0.00000859, 0.00000014, 0.00001069, 0.00001708, 0.00001564, 0.00000736, -0.00000305, -0.00001039, -0.00001196, -0.00000825, -0.00000171, 0.00000476, 0.00000855, 0.00000752, 0.00000084, -0.00000956, -0.00001864, -0.00002060, -0.00001321, -0.00000064, 0.00000849, 0.00000744, -0.00000274, -0.00001312, -0.00001394, -0.00000327, 0.00001040, 0.00001463, 0.00000410, -0.00001365, -0.00002344, -0.00001527, 0.00000616, 0.00002401, 0.00002290, 0.00000238, -0.00002165, -0.00002948, -0.00001423, 0.00001179, 0.00002671, 0.00001726, -0.00000988, -0.00003317, -0.00003370, -0.00001041, 0.00001946, 0.00003470, 0.00002697, 0.00000600, -0.00001021, -0.00001104, -0.00000102, 0.00000617, 0.00000129, -0.00001194, -0.00002070, -0.00001519, 0.00000219, 0.00001871, 0.00002227, 0.00001145, -0.00000400, -0.00001218, -0.00000903, -0.00000065, 0.00000347, -0.00000051, -0.00000760, -0.00000924, -0.00000171, 0.00001023, 0.00001738, 0.00001403, 0.00000269, -0.00000838, -0.00001209, -0.00000780, -0.00000073, 0.00000318, 0.00000213, -0.00000099, -0.00000205, 0.00000047, 0.00000467, 0.00000741, 0.00000717, 0.00000474, 0.00000187, -0.00000046, -0.00000230, -0.00000370, -0.00000401, -0.00000247, 0.00000056, 0.00000316, 0.00000323, 0.00000039, -0.00000332, -0.00000490, -0.00000286, 0.00000149, 0.00000502, 0.00000534, 0.00000254, -0.00000109, -0.00000305, -0.00000245, -0.00000028, 0.00000176, 0.00000269, 0.00000266, 0.00000234, 0.00000208, 0.00000169, },
		{ -0.00000331, -0.00000223, 0.00000017, 0.00000305, 0.00000508, 0.00000486, 0.00000172, -0.00000340, -0.00000786, -0.00000846, -0.00000366, 0.00000478, 0.00001232, 0.00001436, 0.00000941, 0.00000023, -0.00000791, -0.00001079, -0.00000770, -0.00000128, 0.00000467, 0.00000742, 0.00000606, 0.00000093, -0.00000676, -0.00001460, -0.00001901, -0.00001670, -0.00000759, 0.00000371, 0.00000993, 0.00000645, -0.00000430, -0.00001367, -0.00001333, -0.00000263, 0.00000984, 0.00001281, 0.00000221, -0.00001420, -0.00002234, -0.00001369, 0.00000648, 0.00002225, 0.00001993, 0.00000009, -0.00002183, -0.00002774, -0.00001213, 0.00001263, 0.00002602, 0.00001622, -0.00000976, -0.00003152, -0.00003166, -0.00000965, 0.00001830, 0.00003277, 0.00002636, 0.00000795, -0.00000652, -0.00000808, -0.00000107, 0.00000288, -0.00000309, -0.00001441, -0.00001955, -0.00001123, 0.00000631, 0.00002048, 0.00002108, 0.00000881, -0.00000589, -0.00001214, -0.00000768, 0.00000039, 0.00000309, -0.00000189, -0.00000845, -0.00000830, 0.00000082, 0.00001261, 0.00001757, 0.00001123, -0.00000200, -0.00001264, -0.00001390, -0.00000667, 0.00000218, 0.00000607, 0.00000380, -0.00000046, -0.00000171, 0.00000147, 0.00000626, 0.00000865, 0.00000693, 0.00000260, -0.00000150, -0.00000363, -0.00000393, -0.00000334, -0.00000226, -0.00000058, 0.00000140, 0.00000253, 0.00000162, -0.00000106, -0.00000350, -0.00000335, -0.00000007, 0.00000427, 0.00000647, 0.00000475, 0.00000024, -0.00000390, -0.00000500, -0.00000274, 0.00000090, 0.00000348, 0.00000393, 0.00000289, 0.00000168, 0.00000100, 0.00000055, -0.00000028, -0.00000163, -0.00000293, },
		{ 0.00000399, 0.00000417, 0.00000223, -0.00000130, -0.00000495, -0.00000666, -0.00000483, 0.00000048, 0.00000689, 0.00001076, 0.00000938, 0.00000311, -0.00000450, -0.00000887, -0.00000756, -0.00000179, 0.00000456, 0.00000768, 0.00000610, 0.00000085, -0.00000594, -0.00001225, -0.00001634, -0.00001647, -0.00001148, -0.00000242, 0.00000651, 0.00000978, 0.00000458, -0.00000593, -0.00001388, -0.00001245, -0.00000199, 0.00000917, 0.00001103, 0.00000059, -0.00001445, -0.00002122, -0.00001249, 0.00000622, 0.00002010, 0.00001705, -0.00000163, -0.00002124, -0.00002536, -0.00000970, 0.00001359, 0.00002532, 0.00001503, -0.00001006, -0.00003057, -0.00003037, -0.00000924, 0.00001756, 0.00003195, 0.00002693, 0.00001036, -0.00000351, -0.00000663, -0.00000255, -0.00000083, -0.00000662, -0.00001528, -0.00001709, -0.00000702, 0.00000957, 0.00002097, 0.00001906, 0.00000615, -0.00000725, -0.00001168, -0.00000643, 0.00000095, 0.00000242, -0.00000291, -0.00000838, -0.00000656, 0.00000324, 0.00001373, 0.00001590, 0.00000701, -0.00000681, -0.00001561, -0.00001370, -0.00000387, 0.00000570, 0.00000849, 0.00000458, -0.00000053, -0.00000137, 0.00000277, 0.00000790, 0.00000927, 0.00000552, -0.00000063, -0.00000517, -0.00000606, -0.00000415, -0.00000167, -0.00000005, 0.00000068, 0.00000098, 0.00000085, -0.00000006, -0.00000145, -0.00000198, -0.00000039, 0.00000301, 0.00000598, 0.00000594, 0.00000220, -0.00000311, -0.00000647, -0.00000566, -0.00000141, 0.00000327, 0.00000553, 0.00000464, 0.00000210, 0.00000001, -0.00000071, -0.00000061, -0.00000076, -0.00000151, -0.00000221, -0.00000189, -0.00000022, 0.00000215, },
		{ -0.00000307, -0.00000493, -0.00000451, -0.00000160, 0.00000274, 0.00000639, 0.00000710, 0.00000402, -0.00000141, -0.00000592, -0.00000645, -0.00000245, 0.00000356, 0.00000763, 0.00000701, 0.00000189, -0.00000519, -0.00001131, -0.00001476, -0.00001506, -0.00001215, -0.00000622, 0.00000144, 0.00000762, 0.00000845, 0.00000240, -0.00000725, -0.00001366, -0.00001141, -0.00000149, 0.00000836, 0.00000933, -0.00000073, -0.00001447, -0.00002022, -0.00001178, 0.00000539, 0.00001774, 0.00001453, -0.00000253, -0.00001976, -0.00002239, -0.00000712, 0.00001432, 0.00002421, 0.00001339, -0.00001083, -0.00003006, -0.00002931, -0.00000861, 0.00001754, 0.00003206, 0.00002808, 0.00001254, -0.00000157, -0.00000653, -0.00000482, -0.00000427, -0.00000892, -0.00001474, -0.00001388, -0.00000315, 0.00001175, 0.00002038, 0.00001656, 0.00000371, -0.00000814, -0.00001100, -0.00000537, 0.00000124, 0.00000189, -0.00000326, -0.00000755, -0.00000467, 0.00000474, 0.00001313, 0.00001260, 0.00000225, -0.00001066, -0.00001659, -0.00001155, -0.00000010, 0.00000895, 0.00000993, 0.00000449, -0.00000080, -0.00000067, 0.00000438, 0.00000917, 0.00000878, 0.00000282, -0.00000453, -0.00000840, -0.00000710, -0.00000283, 0.00000079, 0.00000176, 0.00000067, -0.00000069, -0.00000117, -0.00000081, -0.00000003, 0.00000120, 0.00000310, 0.00000507, 0.00000555, 0.00000314, -0.00000168, -0.00000629, -0.00000757, -0.00000436, 0.00000140, 0.00000604, 0.00000687, 0.00000404, 0.00000010, -0.00000228, -0.00000226, -0.00000099, -0.00000018, -0.00000043, -0.00000093, -0.00000050, 0.00000112, 0.00000301, 0.00000379, 0.00000266, -0.00000004, },
		{ 0.00000031, 0.00000298, 0.00000409, 0.00000293, -0.00000008, -0.00000322, -0.00000430, -0.00000208, 0.00000254, 0.00000672, 0.00000739, 0.00000342, -0.00000356, -0.00001031, -0.00001414, -0.00001431, -0.00001167, -0.00000731, -0.00000188, 0.00000368, 0.00000732, 0.00000648, 0.00000035, -0.00000809, -0.00001312, -0.00001040, -0.00000120, 0.00000744, 0.00000780, -0.00000177, -0.00001438, -0.00001946, -0.00001158, 0.00000417, 0.00001546, 0.00001267, -0.00000246, -0.00001745, -0.00001909, -0.00000477, 0.00001449, 0.00002250, 0.00001130, -0.00001184, -0.00002956, -0.00002801, -0.00000749, 0.00001816, 0.00003266, 0.00002922, 0.00001404, -0.00000072, -0.00000735, -0.00000728, -0.00000697, -0.00000998, -0.00001318, -0.00001048, 0.00000005, 0.00001284, 0.00001895, 0.00001385, 0.00000156, -0.00000866, -0.00001021, -0.00000439, 0.00000159, 0.00000180, -0.00000299, -0.00000647, -0.00000342, 0.00000476, 0.00001086, 0.00000844, -0.00000196, -0.00001273, -0.00001547, -0.00000805, 0.00000374, 0.00001127, 0.00001029, 0.00000393, -0.00000079, 0.00000051, 0.00000594, 0.00000954, 0.00000689, -0.00000091, -0.00000836, -0.00001041, -0.00000645, -0.00000037, 0.00000317, 0.00000239, -0.00000078, -0.00000300, -0.00000244, 0.00000027, 0.00000327, 0.00000520, 0.00000572, 0.00000490, 0.00000254, -0.00000131, -0.00000554, -0.00000789, -0.00000638, -0.00000123, 0.00000480, 0.00000803, 0.00000664, 0.00000195, -0.00000263, -0.00000431, -0.00000282, -0.00000020, 0.00000129, 0.00000101, 0.00000015, 0.00000024, 0.00000160, 0.00000309, 0.00000318, 0.00000135, -0.00000151, -0.00000377, -0.00000418, -0.00000254, },
		{ -0.00000038, -0.00000202, -0.00000240, -0.00000077, 0.00000252, 0.00000587, 0.00000698, 0.00000431, -0.00000171, -0.00000860, -0.00001333, -0.00001416, -0.00001156, -0.00000733, -0.00000297, 0.00000101, 0.00000439, 0.00000609, 0.00000437, -0.00000132, -0.00000851, -0.00001244, -0.00000953, -0.00000111, 0.00000650, 0.00000645, -0.00000260, -0.00001431, -0.00001904, -0.00001179, 0.00000283, 0.00001358, 0.00001166, -0.00000147, -0.00001460, -0.00001586, -0.00000298, 0.00001392, 0.00002024, 0.00000899, -0.00001272, -0.00002873, -0.00002628, -0.00000592, 0.00001913, 0.00003330, 0.00002996, 0.00001475, -0.00000073, -0.00000862, -0.00000946, -0.00000876, -0.00000998, -0.00001103, -0.00000729, 0.00000238, 0.00001294, 0.00001692, 0.00001110, -0.00000027, -0.00000883, -0.00000920, -0.00000324, 0.00000225, 0.00000211, -0.00000253, -0.00000584, -0.00000337, 0.00000324, 0.00000757, 0.00000447, -0.00000469, -0.00001272, -0.00001273, -0.00000412, 0.00000684, 0.00001242, 0.00000991, 0.00000345, -0.00000024, 0.00000192, 0.00000687, 0.00000854, 0.00000370, -0.00000499, -0.00001126, -0.00001068, -0.00000433, 0.00000241, 0.00000456, 0.00000147, -0.00000317, -0.00000487, -0.00000197, 0.00000344, 0.00000772, 0.00000857, 0.00000608, 0.00000187, -0.00000249, -0.00000600, -0.00000780, -0.00000691, -0.00000297, 0.00000277, 0.00000743, 0.00000823, 0.00000461, -0.00000106, -0.00000508, -0.00000517, -0.00000202, 0.00000153, 0.00000297, 0.00000203, 0.00000049, 0.00000026, 0.00000160, 0.00000302, 0.00000278, 0.00000054, -0.00000237, -0.00000406, -0.00000358, -0.00000147, 0.00000078, 0.00000185, 0.00000129, },
		{ 0.00000320, 0.00000593, 0.00000674, 0.00000456, -0.00000051, -0.00000687, -0.00001200, -0.00001378, -0.00001180, -0.00000752, -0.00000309, 0.00000019, 0.00000239, 0.00000396, 0.00000444, 0.00000246, -0.00000255, -0.00000864, -0.00001177, -0.00000886, -0.00000116, 0.00000561, 0.00000528, -0.00000332, -0.00001436, -0.00001889, -0.00001217, 0.00000168, 0.00001233, 0.00001149, 0.00000017, -0.00001162, -0.00001305, -0.00000195, 0.00001268, 0.00001767, 0.00000682, -0.00001320, -0.00002742, -0.00002415, -0.00000411, 0.00002010, 0.00003367, 0.00003014, 0.00001478, -0.00000125, -0.00000993, -0.00001110, -0.00000966, -0.00000922, -0.00000869, -0.00000459, 0.00000380, 0.00001218, 0.00001447, 0.00000845, -0.00000168, -0.00000854, -0.00000781, -0.00000177, 0.00000316, 0.00000247, -0.00000250, -0.00000624, -0.00000469, 0.00000067, 0.00000420, 0.00000163, -0.00000551, -0.00001095, -0.00000921, -0.00000061, 0.00000882, 0.00001262, 0.00000935, 0.00000343, 0.00000072, 0.00000301, 0.00000662, 0.00000612, -0.00000022, -0.00000851, -0.00001254, -0.00000926, -0.00000148, 0.00000452, 0.00000442, -0.00000068, -0.00000549, -0.00000515, 0.00000074, 0.00000811, 0.00001187, 0.00000974, 0.00000335, -0.00000360, -0.00000810, -0.00000915, -0.00000723, -0.00000329, 0.00000167, 0.00000612, 0.00000811, 0.00000631, 0.00000137, -0.00000388, -0.00000625, -0.00000445, -0.00000015, 0.00000342, 0.00000404, 0.00000214, 0.00000013, 0.00000004, 0.00000176, 0.00000331, 0.00000273, -0.00000006, -0.00000324, -0.00000466, -0.00000364, -0.00000133, 0.00000038, 0.00000040, -0.00000087, -0.00000204, -0.00000183, 0.00000015, },
		{ 0.00000031, -0.00000573, -0.00001083, -0.00001311, -0.00001184, -0.00000787, -0.00000320, 0.00000028, 0.00000194, 0.00000251, 0.00000287, 0.00000277, 0.00000090, -0.00000344, -0.00000866, -0.00001122, -0.00000838, -0.00000129, 0.00000480, 0.00000425, -0.00000401, -0.00001451, -0.00001884, -0.00001247, 0.00000094, 0.00001174, 0.00001195, 0.00000209, -0.00000887, -0.00001088, -0.00000160, 0.00001104, 0.00001514, 0.00000504, -0.00001315, -0.00002568, -0.00002181, -0.00000233, 0.00002084, 0.00003362, 0.00002980, 0.00001432, -0.00000198, -0.00001102, -0.00001212, -0.00000981, -0.00000799, -0.00000645, -0.00000256, 0.00000430, 0.00001070, 0.00001180, 0.00000609, -0.00000249, -0.00000761, -0.00000594, -0.00000007, 0.00000399, 0.00000231, -0.00000343, -0.00000784, -0.00000698, -0.00000210, 0.00000169, 0.00000047, -0.00000458, -0.00000817, -0.00000579, 0.00000195, 0.00000977, 0.00001238, 0.00000906, 0.00000388, 0.00000162, 0.00000316, 0.00000493, 0.00000266, -0.00000400, -0.00001067, -0.00001202, -0.00000671, 0.00000111, 0.00000525, 0.00000285, -0.00000314, -0.00000651, -0.00000317, 0.00000528, 0.00001289, 0.00001403, 0.00000767, -0.00000226, -0.00001010, -0.00001237, -0.00000931, -0.00000360, 0.00000200, 0.00000589, 0.00000744, 0.00000635, 0.00000283, -0.00000185, -0.00000536, -0.00000562, -0.00000247, 0.00000192, 0.00000456, 0.00000401, 0.00000148, -0.00000035, 0.00000024, 0.00000251, 0.00000394, 0.00000260, -0.00000108, -0.00000469, -0.00000587, -0.00000423, -0.00000154, -0.00000001, -0.00000047, -0.00000180, -0.00000208, -0.00000021, 0.00000322, 0.00000638, 0.00000734, 0.00000519, },
		{ -0.00001189, -0.00000811, -0.00000330, 0.00000057, 0.00000236, 0.00000234, 0.00000175, 0.00000154, 0.00000133, -0.00000030, -0.00000412, -0.00000872, -0.00001084, -0.00000804, -0.00000144, 0.00000409, 0.00000333, -0.00000463, -0.00001463, -0.00001871, -0.00001250, 0.00000065, 0.00001164, 0.00001269, 0.00000391, -0.00000659, -0.00000931, -0.00000170, 0.00000932, 0.00001291, 0.00000377, -0.00001264, -0.00002370, -0.00001948, -0.00000076, 0.00002123, 0.00003316, 0.00002905, 0.00001360, -0.00000271, -0.00001175, -0.00001253, -0.00000940, -0.00000658, -0.00000458, -0.00000135, 0.00000392, 0.00000870, 0.00000918, 0.00000432, -0.00000249, -0.00000596, -0.00000371, 0.00000150, 0.00000418, 0.00000115, -0.00000550, -0.00001032, -0.00000948, -0.00000417, 0.00000064, 0.00000096, -0.00000252, -0.00000522, -0.00000309, 0.00000351, 0.00001012, 0.00001222, 0.00000919, 0.00000443, 0.00000183, 0.00000199, 0.00000203, -0.00000103, -0.00000674, -0.00001106, -0.00001008, -0.00000399, 0.00000259, 0.00000446, 0.00000060, -0.00000468, -0.00000537, 0.00000089, 0.00001038, 0.00001600, 0.00001291, 0.00000233, -0.00000941, -0.00001558, -0.00001359, -0.00000600, 0.00000221, 0.00000720, 0.00000807, 0.00000601, 0.00000265, -0.00000089, -0.00000373, -0.00000484, -0.00000348, -0.00000018, 0.00000315, 0.00000440, 0.00000305, 0.00000072, -0.00000019, 0.00000130, 0.00000372, 0.00000436, 0.00000172, -0.00000303, -0.00000685, -0.00000737, -0.00000472, -0.00000132, 0.00000030, -0.00000036, -0.00000158, -0.00000108, 0.00000188, 0.00000579, 0.00000802, 0.00000667, 0.00000182, -0.00000466, -0.00001022, -0.00001285, },
		{ 0.00000324, 0.00000308, 0.00000171, 0.00000056, 0.00000032, 0.00000022, -0.00000125, -0.00000478, -0.00000892, -0.00001066, -0.00000779, -0.00000153, 0.00000354, 0.00000259, -0.00000511, -0.00001460, -0.00001837, -0.00001223, 0.00000069, 0.00001174, 0.00001340, 0.00000542, -0.00000475, -0.00000813, -0.00000193, 0.00000778, 0.00001108, 0.00000292, -0.00001186, -0.00002168, -0.00001731, 0.00000052, 0.00002129, 0.00003236, 0.00002803, 0.00001275, -0.00000328, -0.00001204, -0.00001240, -0.00000863, -0.00000526, -0.00000332, -0.00000101, 0.00000281, 0.00000648, 0.00000699, 0.00000342, -0.00000155, -0.00000376, -0.00000154, 0.00000238, 0.00000327, -0.00000119, -0.00000842, -0.00001299, -0.00001133, -0.00000493, 0.00000110, 0.00000257, -0.00000015, -0.00000277, -0.00000126, 0.00000443, 0.00001038, 0.00001237, 0.00000948, 0.00000449, 0.00000088, -0.00000043, -0.00000138, -0.00000404, -0.00000789, -0.00000988, -0.00000759, -0.00000203, 0.00000261, 0.00000273, -0.00000114, -0.00000430, -0.00000201, 0.00000594, 0.00001424, 0.00001596, 0.00000823, -0.00000510, -0.00001600, -0.00001811, -0.00001101, 0.00000005, 0.00000840, 0.00001049, 0.00000725, 0.00000221, -0.00000165, -0.00000336, -0.00000344, -0.00000252, -0.00000084, 0.00000126, 0.00000289, 0.00000309, 0.00000193, 0.00000076, 0.00000110, 0.00000304, 0.00000471, 0.00000376, -0.00000047, -0.00000593, -0.00000925, -0.00000840, -0.00000432, -0.00000014, 0.00000150, 0.00000051, -0.00000076, 0.00000016, 0.00000358, 0.00000722, 0.00000797, 0.00000433, -0.00000248, -0.00000932, -0.00001317, -0.00001271, -0.00000875, -0.00000340, 0.00000107, },
		{ -0.00000067, -0.00000061, -0.00000058, -0.00000211, -0.00000557, -0.00000936, -0.00001061, -0.00000751, -0.00000143, 0.00000323, 0.00000209, -0.00000536, -0.00001436, -0.00001780, -0.00001177, 0.00000081, 0.00001178, 0.00001388, 0.00000663, -0.00000318, -0.00000705, -0.00000204, 0.00000653, 0.00000958, 0.00000230, -0.00001101, -0.00001974, -0.00001534, 0.00000155, 0.00002106, 0.00003129, 0.00002681, 0.00001192, -0.00000359, -0.00001189, -0.00001185, -0.00000776, -0.00000432, -0.00000282, -0.00000149, 0.00000125, 0.00000445, 0.00000560, 0.00000356, 0.00000016, -0.00000144, -0.00000003, 0.00000206, 0.00000105, -0.00000445, -0.00001153, -0.00001501, -0.00001193, -0.00000425, 0.00000266, 0.00000458, 0.00000190, -0.00000104, -0.00000005, 0.00000516, 0.00001084, 0.00001267, 0.00000941, 0.00000355, -0.00000130, -0.00000357, -0.00000442, -0.00000568, -0.00000742, -0.00000783, -0.00000547, -0.00000134, 0.00000154, 0.00000112, -0.00000127, -0.00000170, 0.00000268, 0.00001024, 0.00001522, 0.00001220, 0.00000094, -0.00001253, -0.00001989, -0.00001660, -0.00000513, 0.00000705, 0.00001293, 0.00001061, 0.00000361, -0.00000264, -0.00000488, -0.00000352, -0.00000104, 0.00000056, 0.00000099, 0.00000103, 0.00000122, 0.00000148, 0.00000168, 0.00000219, 0.00000335, 0.00000465, 0.00000450, 0.00000149, -0.00000394, -0.00000915, -0.00001095, -0.00000813, -0.00000259, 0.00000203, 0.00000322, 0.00000156, -0.00000003, 0.00000100, 0.00000445, 0.00000738, 0.00000652, 0.00000099, -0.00000683, -0.00001290, -0.00001422, -0.00001058, -0.00000429, 0.00000150, 0.00000459, 0.00000457, 0.00000260, 0.00000042, },
		{ -0.00000309, -0.00000664, -0.00001001, -0.00001058, -0.00000703, -0.00000101, 0.00000328, 0.00000190, -0.00000538, -0.00001395, -0.00001716, -0.00001133, 0.00000081, 0.00001161, 0.00001414, 0.00000768, -0.00000164, -0.00000585, -0.00000194, 0.00000550, 0.00000823, 0.00000174, -0.00001020, -0.00001791, -0.00001351, 0.00000238, 0.00002059, 0.00002998, 0.00002549, 0.00001120, -0.00000356, -0.00001133, -0.00001107, -0.00000706, -0.00000399, -0.00000314, -0.00000258, -0.00000034, 0.00000308, 0.00000528, 0.00000467, 0.00000221, 0.00000037, 0.00000027, 0.00000032, -0.00000221, -0.00000798, -0.00001402, -0.00001578, -0.00001111, -0.00000250, 0.00000463, 0.00000633, 0.00000332, 0.00000012, 0.00000096, 0.00000602, 0.00001142, 0.00001266, 0.00000843, 0.00000141, -0.00000431, -0.00000660, -0.00000635, -0.00000577, -0.00000587, -0.00000581, -0.00000436, -0.00000179, 0.00000029, 0.00000077, 0.00000069, 0.00000248, 0.00000708, 0.00001201, 0.00001246, 0.00000540, -0.00000696, -0.00001767, -0.00001970, -0.00001131, 0.00000235, 0.00001279, 0.00001427, 0.00000750, -0.00000158, -0.00000672, -0.00000576, -0.00000126, 0.00000246, 0.00000296, 0.00000102, -0.00000090, -0.00000098, 0.00000074, 0.00000307, 0.00000490, 0.00000568, 0.00000499, 0.00000232, -0.00000243, -0.00000793, -0.00001157, -0.00001099, -0.00000611, 0.00000034, 0.00000461, 0.00000475, 0.00000213, 0.00000013, 0.00000117, 0.00000443, 0.00000647, 0.00000413, -0.00000268, -0.00001055, -0.00001494, -0.00001336, -0.00000690, 0.00000075, 0.00000588, 0.00000686, 0.00000456, 0.00000123, -0.00000109, -0.00000164, -0.00000117, -0.00000122, },
		{ -0.00001033, -0.00000611, -0.00000010, 0.00000372, 0.00000195, -0.00000529, -0.00001356, -0.00001662, -0.00001109, 0.00000057, 0.00001124, 0.00001431, 0.00000879, 0.00000009, -0.00000444, -0.00000170, 0.00000452, 0.00000686, 0.00000113, -0.00000943, -0.00001612, -0.00001179, 0.00000302, 0.00001988, 0.00002848, 0.00002415, 0.00001069, -0.00000317, -0.00001047, -0.00001032, -0.00000682, -0.00000441, -0.00000414, -0.00000389, -0.00000146, 0.00000272, 0.00000603, 0.00000633, 0.00000395, 0.00000103, -0.00000093, -0.00000264, -0.00000589, -0.00001094, -0.00001523, -0.00001508, -0.00000919, -0.00000033, 0.00000637, 0.00000751, 0.00000424, 0.00000112, 0.00000210, 0.00000700, 0.00001172, 0.00001181, 0.00000630, -0.00000163, -0.00000741, -0.00000873, -0.00000682, -0.00000465, -0.00000406, -0.00000452, -0.00000433, -0.00000262, -0.00000006, 0.00000230, 0.00000434, 0.00000676, 0.00000940, 0.00001019, 0.00000637, -0.00000260, -0.00001309, -0.00001885, -0.00001534, -0.00000376, 0.00000912, 0.00001548, 0.00001203, 0.00000226, -0.00000646, -0.00000853, -0.00000394, 0.00000250, 0.00000553, 0.00000350, -0.00000094, -0.00000365, -0.00000239, 0.00000183, 0.00000609, 0.00000797, 0.00000677, 0.00000312, -0.00000190, -0.00000717, -0.00001116, -0.00001207, -0.00000887, -0.00000257, 0.00000373, 0.00000675, 0.00000543, 0.00000187, -0.00000036, 0.00000073, 0.00000370, 0.00000482, 0.00000129, -0.00000609, -0.00001313, -0.00001524, -0.00001080, -0.00000226, 0.00000566, 0.00000920, 0.00000772, 0.00000343, -0.00000059, -0.00000243, -0.00000220, -0.00000146, -0.00000194, -0.00000441, -0.00000805, -0.00001076, },
		{ 0.00000445, 0.00000205, -0.00000533, -0.00001337, -0.00001634, -0.00001108, 0.00000017, 0.00001083, 0.00001460, 0.00001014, 0.00000205, -0.00000290, -0.00000147, 0.00000342, 0.00000535, 0.00000045, -0.00000864, -0.00001431, -0.00001013, 0.00000348, 0.00001892, 0.00002684, 0.00002289, 0.00001046, -0.00000247, -0.00000952, -0.00000987, -0.00000722, -0.00000553, -0.00000550, -0.00000490, -0.00000169, 0.00000346, 0.00000752, 0.00000790, 0.00000469, 0.00000017, -0.00000351, -0.00000621, -0.00000915, -0.00001266, -0.00001492, -0.00001321, -0.00000679, 0.00000163, 0.00000753, 0.00000819, 0.00000500, 0.00000228, 0.00000339, 0.00000774, 0.00001121, 0.00000980, 0.00000318, -0.00000492, -0.00000979, -0.00000950, -0.00000601, -0.00000303, -0.00000272, -0.00000419, -0.00000481, -0.00000284, 0.00000124, 0.00000553, 0.00000842, 0.00000939, 0.00000836, 0.00000488, -0.00000144, -0.00000942, -0.00001553, -0.00001562, -0.00000811, 0.00000372, 0.00001312, 0.00001438, 0.00000712, -0.00000318, -0.00000925, -0.00000743, -0.00000020, 0.00000616, 0.00000676, 0.00000179, -0.00000409, -0.00000582, -0.00000187, 0.00000489, 0.00000983, 0.00000996, 0.00000548, -0.00000118, -0.00000734, -0.00001131, -0.00001231, -0.00001003, -0.00000484, 0.00000160, 0.00000655, 0.00000770, 0.00000495, 0.00000085, -0.00000128, -0.00000007, 0.00000250, 0.00000273, -0.00000158, -0.00000879, -0.00001429, -0.00001386, -0.00000698, 0.00000268, 0.00000982, 0.00001115, 0.00000726, 0.00000156, -0.00000242, -0.00000335, -0.00000236, -0.00000171, -0.00000300, -0.00000617, -0.00000968, -0.00001133, -0.00000956, -0.00000455, 0.00000134, },
		{ -0.00001351, -0.00001629, -0.00001121, -0.00000024, 0.00001057, 0.00001515, 0.00001176, 0.00000418, -0.00000137, -0.00000141, 0.00000209, 0.00000372, -0.00000021, -0.00000775, -0.00001244, -0.00000856, 0.00000371, 0.00001775, 0.00002514, 0.00002179, 0.00001052, -0.00000159, -0.00000872, -0.00000994, -0.00000831, -0.00000710, -0.00000673, -0.00000517, -0.00000083, 0.00000506, 0.00000912, 0.00000866, 0.00000396, -0.00000216, -0.00000689, -0.00000953, -0.00001123, -0.00001282, -0.00001331, -0.00001075, -0.00000457, 0.00000299, 0.00000815, 0.00000872, 0.00000598, 0.00000367, 0.00000453, 0.00000773, 0.00000953, 0.00000670, -0.00000036, -0.00000768, -0.00001093, -0.00000896, -0.00000451, -0.00000164, -0.00000220, -0.00000444, -0.00000493, -0.00000168, 0.00000419, 0.00000943, 0.00001125, 0.00000898, 0.00000379, -0.00000258, -0.00000867, -0.00001293, -0.00001350, -0.00000900, -0.00000026, 0.00000887, 0.00001328, 0.00001012, 0.00000142, -0.00000688, -0.00000905, -0.00000400, 0.00000387, 0.00000814, 0.00000554, -0.00000151, -0.00000696, -0.00000605, 0.00000094, 0.00000911, 0.00001276, 0.00000951, 0.00000140, -0.00000714, -0.00001244, -0.00001332, -0.00001058, -0.00000565, 0.00000009, 0.00000517, 0.00000790, 0.00000714, 0.00000350, -0.00000055, -0.00000229, -0.00000105, 0.00000097, 0.00000043, -0.00000411, -0.00001042, -0.00001388, -0.00001103, -0.00000247, 0.00000724, 0.00001275, 0.00001166, 0.00000579, -0.00000054, -0.00000388, -0.00000375, -0.00000233, -0.00000225, -0.00000460, -0.00000835, -0.00001127, -0.00001137, -0.00000799, -0.00000226, 0.00000319, 0.00000521, 0.00000191, -0.00000571, },
		{ -0.00000047, 0.00001055, 0.00001596, 0.00001356, 0.00000634, -0.00000000, -0.00000162, 0.00000055, 0.00000201, -0.00000076, -0.00000671, -0.00001052, -0.00000712, 0.00000371, 0.00001640, 0.00002347, 0.00002089, 0.00001079, -0.00000074, -0.00000831, -0.00001062, -0.00000990, -0.00000871, -0.00000739, -0.00000442, 0.00000097, 0.00000699, 0.00001014, 0.00000808, 0.00000172, -0.00000548, -0.00001025, -0.00001183, -0.00001176, -0.00001157, -0.00001100, -0.00000840, -0.00000296, 0.00000375, 0.00000858, 0.00000947, 0.00000728, 0.00000503, 0.00000500, 0.00000653, 0.00000665, 0.00000300, -0.00000356, -0.00000930, -0.00001071, -0.00000755, -0.00000302, -0.00000092, -0.00000228, -0.00000453, -0.00000390, 0.00000104, 0.00000796, 0.00001244, 0.00001138, 0.00000511, -0.00000323, -0.00001000, -0.00001306, -0.00001206, -0.00000769, -0.00000111, 0.00000578, 0.00001028, 0.00000992, 0.00000446, -0.00000301, -0.00000764, -0.00000616, 0.00000025, 0.00000654, 0.00000757, 0.00000235, -0.00000491, -0.00000800, -0.00000375, 0.00000536, 0.00001288, 0.00001331, 0.00000598, -0.00000481, -0.00001314, -0.00001549, -0.00001219, -0.00000609, -0.00000003, 0.00000451, 0.00000708, 0.00000739, 0.00000533, 0.00000163, -0.00000183, -0.00000318, -0.00000217, -0.00000082, -0.00000182, -0.00000593, -0.00001069, -0.00001195, -0.00000718, 0.00000206, 0.00001082, 0.00001423, 0.00001091, 0.00000379, -0.00000243, -0.00000478, -0.00000380, -0.00000245, -0.00000336, -0.00000678, -0.00001067, -0.00001239, -0.00001049, -0.00000546, 0.00000064, 0.00000512, 0.00000562, 0.00000124, -0.00000655, -0.00001393, -0.00001633, -0.00001126, },
		{ 0.00001540, 0.00000837, 0.00000111, -0.00000210, -0.00000117, 0.00000033, -0.00000117, -0.00000552, -0.00000859, -0.00000583, 0.00000351, 0.00001496, 0.00002187, 0.00002016, 0.00001111, -0.00000012, -0.00000839, -0.00001182, -0.00001167, -0.00000992, -0.00000714, -0.00000269, 0.00000328, 0.00000856, 0.00000997, 0.00000598, -0.00000168, -0.00000901, -0.00001276, -0.00001263, -0.00001080, -0.00000946, -0.00000867, -0.00000664, -0.00000202, 0.00000422, 0.00000921, 0.00001063, 0.00000869, 0.00000585, 0.00000432, 0.00000404, 0.00000297, -0.00000058, -0.00000575, -0.00000955, -0.00000947, -0.00000592, -0.00000202, -0.00000080, -0.00000237, -0.00000370, -0.00000145, 0.00000469, 0.00001115, 0.00001310, 0.00000820, -0.00000143, -0.00001071, -0.00001510, -0.00001327, -0.00000716, -0.00000001, 0.00000558, 0.00000835, 0.00000793, 0.00000457, -0.00000046, -0.00000468, -0.00000546, -0.00000199, 0.00000345, 0.00000665, 0.00000467, -0.00000139, -0.00000673, -0.00000636, 0.00000061, 0.00000979, 0.00001446, 0.00001057, -0.00000016, -0.00001153, -0.00001728, -0.00001522, -0.00000789, 0.00000007, 0.00000529, 0.00000716, 0.00000682, 0.00000533, 0.00000298, -0.00000001, -0.00000273, -0.00000397, -0.00000352, -0.00000278, -0.00000371, -0.00000668, -0.00000949, -0.00000878, -0.00000294, 0.00000593, 0.00001303, 0.00001429, 0.00000935, 0.00000178, -0.00000381, -0.00000519, -0.00000382, -0.00000308, -0.00000516, -0.00000932, -0.00001267, -0.00001256, -0.00000845, -0.00000208, 0.00000376, 0.00000665, 0.00000532, -0.00000009, -0.00000776, -0.00001439, -0.00001619, -0.00001106, -0.00000044, 0.00001077, 0.00001693, },
		{ -0.00000281, -0.00000297, -0.00000128, -0.00000140, -0.00000420, -0.00000669, -0.00000469, 0.00000314, 0.00001346, 0.00002033, 0.00001950, 0.00001131, 0.00000014, -0.00000893, -0.00001327, -0.00001322, -0.00001041, -0.00000596, -0.00000034, 0.00000546, 0.00000917, 0.00000837, 0.00000263, -0.00000551, -0.00001193, -0.00001388, -0.00001187, -0.00000881, -0.00000718, -0.00000688, -0.00000560, -0.00000147, 0.00000484, 0.00001030, 0.00001205, 0.00000973, 0.00000560, 0.00000227, 0.00000055, -0.00000084, -0.00000332, -0.00000658, -0.00000861, -0.00000776, -0.00000459, -0.00000159, -0.00000083, -0.00000177, -0.00000160, 0.00000198, 0.00000803, 0.00001234, 0.00001063, 0.00000222, -0.00000886, -0.00001639, -0.00001633, -0.00000932, 0.00000013, 0.00000710, 0.00000917, 0.00000706, 0.00000310, -0.00000061, -0.00000292, -0.00000327, -0.00000153, 0.00000159, 0.00000424, 0.00000419, 0.00000077, -0.00000389, -0.00000582, -0.00000227, 0.00000550, 0.00001234, 0.00001267, 0.00000479, -0.00000734, -0.00001670, -0.00001806, -0.00001142, -0.00000145, 0.00000622, 0.00000882, 0.00000737, 0.00000465, 0.00000251, 0.00000089, -0.00000102, -0.00000326, -0.00000488, -0.00000516, -0.00000468, -0.00000483, -0.00000607, -0.00000691, -0.00000489, 0.00000099, 0.00000860, 0.00001378, 0.00001330, 0.00000753, 0.00000017, -0.00000461, -0.00000539, -0.00000423, -0.00000445, -0.00000751, -0.00001175, -0.00001380, -0.00001144, -0.00000528, 0.00000177, 0.00000656, 0.00000735, 0.00000412, -0.00000198, -0.00000907, -0.00001458, -0.00001565, -0.00001052, -0.00000020, 0.00001107, 0.00001791, 0.00001717, 0.00001022, 0.00000197, },
		},
		{ { -0.00000082, 0.00000059, 0.00000298, 0.00000256, -0.00000275, -0.00001099, -0.00001722, -0.00001718, -0.00001030, 0.00000019, 0.00000961, 0.00001435, 0.00001294, 0.00000634, -0.00000203, -0.00000731, -0.00000639, -0.00000125, 0.00000172, -0.00000231, -0.00001051, -0.00001383, -0.00000628, 0.00000720, 0.00001415, 0.00000750, -0.00000597, -0.00001213, -0.00000501, 0.00000610, 0.00000701, -0.00000469, -0.00001600, -0.00001302, 0.00000279, 0.00001566, 0.00001311, -0.00000029, -0.00000895, -0.00000450, 0.00000531, 0.00000731, -0.00000059, -0.00000732, -0.00000308, 0.00000797, 0.00001256, 0.00000465, -0.00000710, -0.00000965, -0.00000111, 0.00000751, 0.00000599, -0.00000318, -0.00000852, -0.00000367, 0.00000556, 0.00000898, 0.00000394, -0.00000288, -0.00000462, -0.00000174, 0.00000061, 0.00000016, -0.00000066, 0.00000016, 0.00000061, -0.00000174, -0.00000462, -0.00000288, 0.00000394, 0.00000898, 0.00000556, -0.00000367, -0.00000852, -0.00000318, 0.00000599, 0.00000751, -0.00000111, -0.00000965, -0.00000710, 0.00000465, 0.00001256, 0.00000797, -0.00000308, -0.00000732, -0.00000059, 0.00000731, 0.00000531, -0.00000450, -0.00000895, -0.00000029, 0.00001311, 0.00001566, 0.00000279, -0.00001302, -0.00001600, -0.00000469, 0.00000701, 0.00000610, -0.00000501, -0.00001213, -0.00000597, 0.00000750, 0.00001415, 0.00000720, -0.00000628, -0.00001383, -0.00001051, -0.00000231, 0.00000172, -0.00000125, -0.00000639, -0.00000731, -0.00000203, 0.00000634, 0.00001294, 0.00001435, 0.00000961, 0.00000019, -0.00001030, -0.00001718, -0.00001722, -0.00001099, -0.00000275, 0.00000256, 0.00000298, 0.00000059, },
		{ 0.00000106, -0.00000289, -0.00000990, -0.00001567, -0.00001599, -0.00000986, -0.00000000, 0.00000915, 0.00001403, 0.00001316, 0.00000718, -0.00000124, -0.00000774, -0.00000839, -0.00000312, 0.00000276, 0.00000248, -0.00000519, -0.00001312, -0.00001198, -0.00000045, 0.00001170, 0.00001270, 0.00000145, -0.00001056, -0.00001130, -0.00000147, 0.00000633, 0.00000220, -0.00000924, -0.00001395, -0.00000490, 0.00000938, 0.00001442, 0.00000627, -0.00000450, -0.00000586, 0.00000166, 0.00000683, 0.00000252, -0.00000524, -0.00000527, 0.00000405, 0.00001198, 0.00000844, -0.00000329, -0.00001014, -0.00000487, 0.00000539, 0.00000795, -0.00000004, -0.00000852, -0.00000698, 0.00000308, 0.00001041, 0.00000752, -0.00000132, -0.00000633, -0.00000383, 0.00000104, 0.00000222, -0.00000011, -0.00000166, -0.00000091, -0.00000052, -0.00000212, -0.00000277, 0.00000074, 0.00000594, 0.00000635, 0.00000000, -0.00000653, -0.00000529, 0.00000271, 0.00000760, 0.00000256, -0.00000700, -0.00000915, -0.00000002, 0.00001067, 0.00001083, 0.00000069, -0.00000734, -0.00000389, 0.00000565, 0.00000789, -0.00000132, -0.00001065, -0.00000675, 0.00000864, 0.00001886, 0.00001107, -0.00000839, -0.00002019, -0.00001317, 0.00000366, 0.00001175, 0.00000368, -0.00000954, -0.00001187, -0.00000024, 0.00001284, 0.00001363, 0.00000155, -0.00001162, -0.00001491, -0.00000845, -0.00000104, 0.00000026, -0.00000347, -0.00000594, -0.00000275, 0.00000490, 0.00001218, 0.00001459, 0.00001034, 0.00000064, -0.00001069, -0.00001855, -0.00001904, -0.00001220, -0.00000244, 0.00000441, 0.00000536, 0.00000213, -0.00000095, -0.00000104, 0.00000083, },
		{ -0.00001443, -0.00001513, -0.00000949, 0.00000004, 0.00000902, 0.00001381, 0.00001301, 0.00000738, -0.00000071, -0.00000758, -0.00000934, -0.00000470, 0.00000262, 0.00000555, -0.00000011, -0.00001027, -0.00001488, -0.00000760, 0.00000646, 0.00001439, 0.00000832, -0.00000583, -0.00001397, -0.00000883, 0.00000231, 0.00000603, -0.00000181, -0.00001133, -0.00001028, 0.00000156, 0.00001215, 0.00001122, 0.00000173, -0.00000458, -0.00000166, 0.00000440, 0.00000429, -0.00000203, -0.00000534, 0.00000039, 0.00000933, 0.00001036, 0.00000113, -0.00000836, -0.00000752, 0.00000191, 0.00000796, 0.00000306, -0.00000668, -0.00000896, -0.00000028, 0.00000989, 0.00001047, 0.00000155, -0.00000661, -0.00000610, 0.00000022, 0.00000383, 0.00000136, -0.00000265, -0.00000300, -0.00000055, 0.00000033, -0.00000116, -0.00000125, 0.00000211, 0.00000515, 0.00000294, -0.00000304, -0.00000556, -0.00000084, 0.00000557, 0.00000498, -0.00000301, -0.00000880, -0.00000423, 0.00000664, 0.00001157, 0.00000491, -0.00000504, -0.00000610, 0.00000242, 0.00000839, 0.00000233, -0.00000950, -0.00001142, 0.00000227, 0.00001825, 0.00001793, -0.00000101, -0.00002051, -0.00002114, -0.00000325, 0.00001372, 0.00001272, -0.00000291, -0.00001437, -0.00000873, 0.00000765, 0.00001716, 0.00001009, -0.00000632, -0.00001713, -0.00001506, -0.00000573, 0.00000063, -0.00000005, -0.00000340, -0.00000296, 0.00000305, 0.00001074, 0.00001455, 0.00001122, 0.00000137, -0.00001095, -0.00001997, -0.00002102, -0.00001355, -0.00000210, 0.00000648, 0.00000797, 0.00000368, -0.00000131, -0.00000287, -0.00000121, -0.00000005, -0.00000271, -0.00000883, },
		{ 0.00000022, 0.00000925, 0.00001382, 0.00001269, 0.00000703, -0.00000064, -0.00000719, -0.00000938, -0.00000558, 0.00000194, 0.00000697, 0.00000388, -0.00000635, -0.00001503, -0.00001294, -0.00000007, 0.00001257, 0.00001294, 0.00000040, -0.00001285, -0.00001409, -0.00000350, 0.00000629, 0.00000438, -0.00000623, -0.00001238, -0.00000597, 0.00000689, 0.00001328, 0.00000821, -0.00000069, -0.00000334, 0.00000092, 0.00000410, 0.00000108, -0.00000350, -0.00000191, 0.00000559, 0.00001002, 0.00000491, -0.00000484, -0.00000827, -0.00000182, 0.00000600, 0.00000507, -0.00000368, -0.00000909, -0.00000350, 0.00000762, 0.00001196, 0.00000493, -0.00000524, -0.00000771, -0.00000155, 0.00000442, 0.00000318, -0.00000251, -0.00000495, -0.00000174, 0.00000197, 0.00000137, -0.00000158, -0.00000149, 0.00000223, 0.00000427, 0.00000100, -0.00000378, -0.00000348, 0.00000202, 0.00000533, 0.00000111, -0.00000610, -0.00000673, 0.00000150, 0.00000966, 0.00000819, -0.00000091, -0.00000616, -0.00000120, 0.00000652, 0.00000494, -0.00000627, -0.00001321, -0.00000419, 0.00001413, 0.00002166, 0.00000734, -0.00001666, -0.00002647, -0.00001218, 0.00001116, 0.00001975, 0.00000650, -0.00001232, -0.00001570, -0.00000044, 0.00001662, 0.00001751, 0.00000144, -0.00001622, -0.00002086, -0.00001202, -0.00000081, 0.00000315, 0.00000024, -0.00000220, 0.00000114, 0.00000863, 0.00001401, 0.00001209, 0.00000233, -0.00001103, -0.00002134, -0.00002302, -0.00001497, -0.00000183, 0.00000858, 0.00001074, 0.00000535, -0.00000174, -0.00000487, -0.00000323, -0.00000084, -0.00000212, -0.00000762, -0.00001345, -0.00001468, -0.00000934, },
		{ 0.00001237, 0.00000626, -0.00000109, -0.00000687, -0.00000878, -0.00000561, 0.00000128, 0.00000717, 0.00000639, -0.00000242, -0.00001306, -0.00001575, -0.00000633, 0.00000807, 0.00001438, 0.00000622, -0.00000879, -0.00001612, -0.00000923, 0.00000352, 0.00000795, -0.00000033, -0.00001120, -0.00001156, -0.00000002, 0.00001197, 0.00001329, 0.00000497, -0.00000255, -0.00000231, 0.00000208, 0.00000300, -0.00000056, -0.00000230, 0.00000198, 0.00000772, 0.00000704, -0.00000062, -0.00000690, -0.00000468, 0.00000267, 0.00000532, -0.00000055, -0.00000742, -0.00000560, 0.00000433, 0.00001157, 0.00000787, -0.00000250, -0.00000801, -0.00000364, 0.00000377, 0.00000468, -0.00000139, -0.00000613, -0.00000362, 0.00000237, 0.00000396, -0.00000031, -0.00000395, -0.00000163, 0.00000358, 0.00000447, -0.00000033, -0.00000434, -0.00000201, 0.00000345, 0.00000411, -0.00000180, -0.00000675, -0.00000339, 0.00000538, 0.00000935, 0.00000393, -0.00000371, -0.00000382, 0.00000282, 0.00000537, -0.00000237, -0.00001188, -0.00000892, 0.00000775, 0.00002133, 0.00001449, -0.00000952, -0.00002762, -0.00002088, 0.00000442, 0.00002276, 0.00001649, -0.00000572, -0.00001905, -0.00000956, 0.00001177, 0.00002191, 0.00001025, -0.00001184, -0.00002443, -0.00001905, -0.00000442, 0.00000523, 0.00000458, -0.00000016, -0.00000029, 0.00000599, 0.00001277, 0.00001272, 0.00000345, -0.00001090, -0.00002259, -0.00002494, -0.00001634, -0.00000162, 0.00001057, 0.00001348, 0.00000712, -0.00000210, -0.00000693, -0.00000532, -0.00000145, -0.00000114, -0.00000613, -0.00001258, -0.00001466, -0.00000957, 0.00000038, 0.00000979, 0.00001417, },
		{ -0.00000682, -0.00000785, -0.00000488, 0.00000103, 0.00000667, 0.00000741, 0.00000075, -0.00000990, -0.00001603, -0.00001108, 0.00000224, 0.00001259, 0.00001010, -0.00000329, -0.00001474, -0.00001315, -0.00000092, 0.00000839, 0.00000465, -0.00000764, -0.00001425, -0.00000689, 0.00000775, 0.00001573, 0.00001099, 0.00000076, -0.00000402, -0.00000096, 0.00000311, 0.00000233, -0.00000091, -0.00000043, 0.00000436, 0.00000712, 0.00000311, -0.00000386, -0.00000586, -0.00000104, 0.00000373, 0.00000172, -0.00000453, -0.00000602, 0.00000099, 0.00000943, 0.00000953, 0.00000090, -0.00000672, -0.00000525, 0.00000210, 0.00000527, 0.00000024, -0.00000620, -0.00000550, 0.00000151, 0.00000581, 0.00000199, -0.00000474, -0.00000540, 0.00000107, 0.00000641, 0.00000386, -0.00000308, -0.00000530, -0.00000016, 0.00000506, 0.00000286, -0.00000424, -0.00000665, -0.00000029, 0.00000774, 0.00000800, 0.00000083, -0.00000428, -0.00000152, 0.00000323, 0.00000060, -0.00000818, -0.00001066, 0.00000101, 0.00001712, 0.00001853, -0.00000092, -0.00002411, -0.00002698, -0.00000491, 0.00002069, 0.00002449, 0.00000417, -0.00001743, -0.00001727, 0.00000352, 0.00002193, 0.00001810, -0.00000454, -0.00002462, -0.00002544, -0.00001000, 0.00000534, 0.00000885, 0.00000322, -0.00000071, 0.00000320, 0.00001075, 0.00001285, 0.00000459, -0.00001055, -0.00002366, -0.00002671, -0.00001756, -0.00000139, 0.00001236, 0.00001600, 0.00000886, -0.00000230, -0.00000886, -0.00000740, -0.00000202, 0.00000009, -0.00000432, -0.00001165, -0.00001496, -0.00001024, 0.00000035, 0.00001053, 0.00001488, 0.00001221, 0.00000526, -0.00000203, },
		{ 0.00000134, 0.00000596, 0.00000726, 0.00000274, -0.00000646, -0.00001427, -0.00001364, -0.00000352, 0.00000830, 0.00001115, 0.00000194, -0.00001068, -0.00001424, -0.00000535, 0.00000616, 0.00000745, -0.00000318, -0.00001396, -0.00001220, 0.00000179, 0.00001502, 0.00001585, 0.00000584, -0.00000341, -0.00000382, 0.00000147, 0.00000412, 0.00000152, -0.00000109, 0.00000108, 0.00000538, 0.00000540, -0.00000010, -0.00000510, -0.00000406, 0.00000078, 0.00000242, -0.00000137, -0.00000471, -0.00000143, 0.00000619, 0.00000944, 0.00000404, -0.00000404, -0.00000569, -0.00000002, 0.00000467, 0.00000170, -0.00000522, -0.00000672, -0.00000018, 0.00000643, 0.00000451, -0.00000386, -0.00000812, -0.00000254, 0.00000632, 0.00000764, 0.00000000, -0.00000686, -0.00000446, 0.00000366, 0.00000654, 0.00000012, -0.00000739, -0.00000591, 0.00000354, 0.00000998, 0.00000633, -0.00000207, -0.00000502, -0.00000104, 0.00000136, -0.00000353, -0.00000906, -0.00000415, 0.00001025, 0.00001833, 0.00000683, -0.00001672, -0.00002867, -0.00001444, 0.00001378, 0.00002823, 0.00001505, -0.00001069, -0.00002127, -0.00000620, 0.00001715, 0.00002295, 0.00000428, -0.00002090, -0.00002962, -0.00001671, 0.00000295, 0.00001203, 0.00000757, 0.00000036, 0.00000083, 0.00000805, 0.00001221, 0.00000554, -0.00000998, -0.00002450, -0.00002830, -0.00001857, -0.00000103, 0.00001401, 0.00001815, 0.00001038, -0.00000237, -0.00001049, -0.00000932, -0.00000261, 0.00000137, -0.00000228, -0.00001051, -0.00001538, -0.00001133, -0.00000004, 0.00001129, 0.00001592, 0.00001234, 0.00000420, -0.00000333, -0.00000713, -0.00000685, -0.00000357, },
		{ 0.00000351, -0.00000348, -0.00001127, -0.00001395, -0.00000807, 0.00000270, 0.00000932, 0.00000548, -0.00000534, -0.00001234, -0.00000825, 0.00000244, 0.00000761, 0.00000067, -0.00001144, -0.00001501, -0.00000439, 0.00001148, 0.00001832, 0.00001141, -0.00000039, -0.00000529, -0.00000117, 0.00000420, 0.00000392, -0.00000008, -0.00000113, 0.00000264, 0.00000583, 0.00000328, -0.00000278, -0.00000560, -0.00000261, 0.00000141, 0.00000109, -0.00000219, -0.00000235, 0.00000276, 0.00000766, 0.00000608, -0.00000061, -0.00000465, -0.00000182, 0.00000303, 0.00000240, -0.00000361, -0.00000688, -0.00000203, 0.00000573, 0.00000642, -0.00000178, -0.00000920, -0.00000625, 0.00000429, 0.00001001, 0.00000404, -0.00000625, -0.00000820, 0.00000033, 0.00000821, 0.00000509, -0.00000543, -0.00001003, -0.00000232, 0.00000912, 0.00001128, 0.00000248, -0.00000639, -0.00000623, -0.00000066, 0.00000034, -0.00000476, -0.00000628, 0.00000267, 0.00001390, 0.00001160, -0.00000728, -0.00002527, -0.00002156, 0.00000359, 0.00002641, 0.00002412, -0.00000004, -0.00001994, -0.00001483, 0.00000836, 0.00002319, 0.00001256, -0.00001366, -0.00003025, -0.00002314, -0.00000188, 0.00001313, 0.00001213, 0.00000307, -0.00000047, 0.00000502, 0.00001062, 0.00000599, -0.00000927, -0.00002503, -0.00002966, -0.00001937, -0.00000047, 0.00001561, 0.00001992, 0.00001148, -0.00000246, -0.00001173, -0.00001085, -0.00000313, 0.00000253, -0.00000019, -0.00000915, -0.00001570, -0.00001268, -0.00000085, 0.00001183, 0.00001713, 0.00001280, 0.00000328, -0.00000485, -0.00000779, -0.00000597, -0.00000194, 0.00000219, 0.00000537, 0.00000637, },
		{ -0.00001241, -0.00001077, -0.00000292, 0.00000525, 0.00000655, -0.00000024, -0.00000816, -0.00000873, -0.00000127, 0.00000552, 0.00000279, -0.00000796, -0.00001519, -0.00000937, 0.00000615, 0.00001787, 0.00001602, 0.00000441, -0.00000460, -0.00000371, 0.00000261, 0.00000531, 0.00000191, -0.00000172, -0.00000006, 0.00000458, 0.00000540, 0.00000027, -0.00000537, -0.00000546, -0.00000090, 0.00000217, 0.00000068, -0.00000165, 0.00000007, 0.00000475, 0.00000651, 0.00000270, -0.00000226, -0.00000262, 0.00000087, 0.00000202, -0.00000200, -0.00000598, -0.00000338, 0.00000408, 0.00000715, 0.00000070, -0.00000853, -0.00000906, 0.00000096, 0.00001035, 0.00000789, -0.00000365, -0.00001032, -0.00000394, 0.00000744, 0.00000928, -0.00000136, -0.00001163, -0.00000841, 0.00000544, 0.00001425, 0.00000830, -0.00000497, -0.00001078, -0.00000508, 0.00000200, 0.00000079, -0.00000485, -0.00000353, 0.00000639, 0.00001203, 0.00000175, -0.00001739, -0.00002420, -0.00000736, 0.00001909, 0.00002887, 0.00001214, -0.00001296, -0.00001982, -0.00000249, 0.00001819, 0.00001803, -0.00000427, -0.00002656, -0.00002760, -0.00000841, 0.00001141, 0.00001579, 0.00000718, -0.00000011, 0.00000224, 0.00000810, 0.00000560, -0.00000863, -0.00002517, -0.00003072, -0.00001999, 0.00000030, 0.00001729, 0.00002138, 0.00001204, -0.00000279, -0.00001264, -0.00001181, -0.00000339, 0.00000354, 0.00000170, -0.00000768, -0.00001577, -0.00001403, -0.00000204, 0.00001199, 0.00001830, 0.00001357, 0.00000265, -0.00000637, -0.00000873, -0.00000533, -0.00000022, 0.00000345, 0.00000507, 0.00000512, 0.00000328, -0.00000133, -0.00000784, },
		{ 0.00000002, 0.00000509, 0.00000334, -0.00000294, -0.00000668, -0.00000364, 0.00000224, 0.00000277, -0.00000487, -0.00001338, -0.00001228, 0.00000042, 0.00001476, 0.00001854, 0.00000978, -0.00000161, -0.00000501, 0.00000004, 0.00000510, 0.00000389, -0.00000075, -0.00000177, 0.00000240, 0.00000582, 0.00000303, -0.00000368, -0.00000700, -0.00000371, 0.00000167, 0.00000300, 0.00000023, -0.00000127, 0.00000155, 0.00000534, 0.00000505, 0.00000091, -0.00000206, -0.00000109, 0.00000064, -0.00000100, -0.00000438, -0.00000378, 0.00000209, 0.00000656, 0.00000274, -0.00000653, -0.00001030, -0.00000267, 0.00000867, 0.00001049, 0.00000013, -0.00001025, -0.00000790, 0.00000455, 0.00001156, 0.00000362, -0.00001037, -0.00001319, -0.00000021, 0.00001431, 0.00001383, -0.00000094, -0.00001329, -0.00001076, 0.00000072, 0.00000577, -0.00000039, -0.00000670, -0.00000221, 0.00000792, 0.00000801, -0.00000688, -0.00002139, -0.00001624, 0.00000778, 0.00002771, 0.00002289, -0.00000144, -0.00001931, -0.00001275, 0.00000859, 0.00001887, 0.00000516, -0.00001878, -0.00002852, -0.00001523, 0.00000676, 0.00001742, 0.00001193, 0.00000224, 0.00000046, 0.00000495, 0.00000411, -0.00000834, -0.00002493, -0.00003134, -0.00002043, 0.00000118, 0.00001910, 0.00002272, 0.00001212, -0.00000357, -0.00001342, -0.00001214, -0.00000314, 0.00000447, 0.00000321, -0.00000634, -0.00001557, -0.00001512, -0.00000339, 0.00001166, 0.00001917, 0.00001448, 0.00000240, -0.00000770, -0.00000981, -0.00000496, 0.00000141, 0.00000493, 0.00000508, 0.00000384, 0.00000238, -0.00000011, -0.00000463, -0.00000970, -0.00001153, -0.00000755, },
		{ 0.00000191, -0.00000277, -0.00000390, -0.00000089, 0.00000098, -0.00000313, -0.00001068, -0.00001296, -0.00000443, 0.00000991, 0.00001841, 0.00001439, 0.00000307, -0.00000431, -0.00000248, 0.00000333, 0.00000494, 0.00000112, -0.00000200, 0.00000024, 0.00000474, 0.00000469, -0.00000128, -0.00000696, -0.00000611, -0.00000009, 0.00000409, 0.00000254, -0.00000111, -0.00000107, 0.00000301, 0.00000594, 0.00000406, -0.00000021, -0.00000220, -0.00000130, -0.00000095, -0.00000266, -0.00000317, 0.00000045, 0.00000495, 0.00000373, -0.00000396, -0.00000981, -0.00000562, 0.00000554, 0.00001116, 0.00000399, -0.00000808, -0.00001041, 0.00000043, 0.00001137, 0.00000811, -0.00000674, -0.00001550, -0.00000641, 0.00001135, 0.00001755, 0.00000470, -0.00001291, -0.00001609, -0.00000329, 0.00000859, 0.00000568, -0.00000610, -0.00000967, 0.00000033, 0.00000990, 0.00000373, -0.00001358, -0.00002068, -0.00000492, 0.00002048, 0.00002940, 0.00001223, -0.00001274, -0.00001964, -0.00000374, 0.00001418, 0.00001223, -0.00000816, -0.00002491, -0.00002052, -0.00000021, 0.00001609, 0.00001620, 0.00000643, 0.00000038, 0.00000172, 0.00000144, -0.00000880, -0.00002444, -0.00003138, -0.00002062, 0.00000207, 0.00002099, 0.00002410, 0.00001187, -0.00000488, -0.00001435, -0.00001195, -0.00000219, 0.00000558, 0.00000429, -0.00000541, -0.00001524, -0.00001578, -0.00000461, 0.00001093, 0.00001951, 0.00001527, 0.00000249, -0.00000864, -0.00001082, -0.00000484, 0.00000281, 0.00000644, 0.00000535, 0.00000267, 0.00000112, 0.00000031, -0.00000198, -0.00000651, -0.00001068, -0.00001060, -0.00000523, 0.00000170, 0.00000470, },
		{ -0.00000271, -0.00000159, -0.00000308, -0.00000827, -0.00001196, -0.00000759, 0.00000457, 0.00001584, 0.00001713, 0.00000835, -0.00000148, -0.00000386, 0.00000063, 0.00000450, 0.00000297, -0.00000083, -0.00000101, 0.00000282, 0.00000486, 0.00000091, -0.00000559, -0.00000739, -0.00000237, 0.00000375, 0.00000444, 0.00000022, -0.00000247, 0.00000032, 0.00000529, 0.00000639, 0.00000245, -0.00000205, -0.00000312, -0.00000185, -0.00000142, -0.00000183, -0.00000039, 0.00000289, 0.00000346, -0.00000165, -0.00000791, -0.00000715, 0.00000187, 0.00000980, 0.00000683, -0.00000451, -0.00001081, -0.00000366, 0.00000889, 0.00001087, -0.00000185, -0.00001489, -0.00001161, 0.00000609, 0.00001841, 0.00001043, -0.00000966, -0.00001954, -0.00000895, 0.00000833, 0.00001147, -0.00000211, -0.00001414, -0.00000880, 0.00000697, 0.00001187, -0.00000251, -0.00001942, -0.00001598, 0.00000857, 0.00002979, 0.00002494, -0.00000105, -0.00002104, -0.00001607, 0.00000439, 0.00001484, 0.00000319, -0.00001677, -0.00002254, -0.00000811, 0.00001147, 0.00001869, 0.00001179, 0.00000250, -0.00000077, -0.00000222, -0.00001036, -0.00002399, -0.00003077, -0.00002040, 0.00000288, 0.00002280, 0.00002560, 0.00001158, -0.00000662, -0.00001567, -0.00001152, -0.00000049, 0.00000717, 0.00000508, -0.00000511, -0.00001509, -0.00001599, -0.00000538, 0.00001005, 0.00001923, 0.00001564, 0.00000275, -0.00000908, -0.00001155, -0.00000485, 0.00000391, 0.00000782, 0.00000575, 0.00000169, -0.00000026, 0.00000018, -0.00000003, -0.00000340, -0.00000877, -0.00001196, -0.00000958, -0.00000265, 0.00000383, 0.00000527, 0.00000184, -0.00000203, },
		{ -0.00000697, -0.00001027, -0.00000892, -0.00000009, 0.00001167, 0.00001748, 0.00001294, 0.00000293, -0.00000338, -0.00000199, 0.00000262, 0.00000392, 0.00000112, -0.00000091, 0.00000098, 0.00000372, 0.00000214, -0.00000357, -0.00000728, -0.00000436, 0.00000224, 0.00000528, 0.00000203, -0.00000249, -0.00000192, 0.00000350, 0.00000738, 0.00000516, -0.00000066, -0.00000419, -0.00000330, -0.00000103, -0.00000031, -0.00000027, 0.00000103, 0.00000215, -0.00000023, -0.00000532, -0.00000702, -0.00000132, 0.00000688, 0.00000791, -0.00000061, -0.00000908, -0.00000656, 0.00000496, 0.00001125, 0.00000288, -0.00001169, -0.00001453, -0.00000011, 0.00001611, 0.00001465, -0.00000441, -0.00002009, -0.00001466, 0.00000501, 0.00001524, 0.00000392, -0.00001467, -0.00001712, 0.00000010, 0.00001564, 0.00000920, -0.00001265, -0.00002274, -0.00000539, 0.00002364, 0.00003364, 0.00001342, -0.00001610, -0.00002544, -0.00000875, 0.00001182, 0.00001273, -0.00000517, -0.00002004, -0.00001510, 0.00000402, 0.00001832, 0.00001720, 0.00000681, -0.00000169, -0.00000632, -0.00001322, -0.00002399, -0.00002955, -0.00001959, 0.00000364, 0.00002432, 0.00002714, 0.00001150, -0.00000854, -0.00001749, -0.00001124, 0.00000176, 0.00000946, 0.00000591, -0.00000544, -0.00001545, -0.00001596, -0.00000552, 0.00000937, 0.00001843, 0.00001534, 0.00000290, -0.00000904, -0.00001177, -0.00000479, 0.00000470, 0.00000892, 0.00000615, 0.00000088, -0.00000160, -0.00000030, 0.00000129, -0.00000069, -0.00000640, -0.00001191, -0.00001251, -0.00000693, 0.00000129, 0.00000654, 0.00000586, 0.00000131, -0.00000260, -0.00000376, -0.00000431, },
		{ -0.00000336, 0.00000701, 0.00001564, 0.00001584, 0.00000787, -0.00000091, -0.00000350, -0.00000007, 0.00000346, 0.00000297, 0.00000042, 0.00000001, 0.00000192, 0.00000209, -0.00000175, -0.00000605, -0.00000542, 0.00000022, 0.00000487, 0.00000355, -0.00000140, -0.00000312, 0.00000126, 0.00000692, 0.00000718, 0.00000154, -0.00000410, -0.00000472, -0.00000152, 0.00000083, 0.00000057, -0.00000011, 0.00000033, 0.00000001, -0.00000284, -0.00000549, -0.00000328, 0.00000329, 0.00000705, 0.00000253, -0.00000582, -0.00000753, 0.00000075, 0.00000929, 0.00000618, -0.00000693, -0.00001457, -0.00000569, 0.00001124, 0.00001617, 0.00000135, -0.00001760, -0.00001880, -0.00000040, 0.00001589, 0.00001010, -0.00001142, -0.00002259, -0.00000882, 0.00001434, 0.00001882, -0.00000199, -0.00002363, -0.00001824, 0.00001216, 0.00003613, 0.00002748, -0.00000553, -0.00002948, -0.00002253, 0.00000332, 0.00001815, 0.00000783, -0.00001274, -0.00001923, -0.00000500, 0.00001451, 0.00002127, 0.00001277, -0.00000042, -0.00001005, -0.00001729, -0.00002489, -0.00002798, -0.00001800, 0.00000451, 0.00002539, 0.00002851, 0.00001176, -0.00001024, -0.00001968, -0.00001143, 0.00000420, 0.00001249, 0.00000716, -0.00000622, -0.00001655, -0.00001605, -0.00000501, 0.00000926, 0.00001740, 0.00001425, 0.00000256, -0.00000871, -0.00001134, -0.00000441, 0.00000527, 0.00000963, 0.00000640, 0.00000021, -0.00000280, -0.00000093, 0.00000212, 0.00000150, -0.00000401, -0.00001092, -0.00001396, -0.00001035, -0.00000201, 0.00000578, 0.00000831, 0.00000499, -0.00000064, -0.00000453, -0.00000592, -0.00000700, -0.00000888, -0.00000891, },
		{ 0.00001654, 0.00001214, 0.00000300, -0.00000320, -0.00000259, 0.00000166, 0.00000386, 0.00000236, 0.00000029, 0.00000033, 0.00000096, -0.00000079, -0.00000432, -0.00000534, -0.00000158, 0.00000344, 0.00000421, 0.00000018, -0.00000309, -0.00000066, 0.00000535, 0.00000801, 0.00000387, -0.00000286, -0.00000549, -0.00000259, 0.00000121, 0.00000165, -0.00000037, -0.00000137, -0.00000077, -0.00000114, -0.00000321, -0.00000369, -0.00000001, 0.00000462, 0.00000412, -0.00000203, -0.00000643, -0.00000254, 0.00000572, 0.00000727, -0.00000198, -0.00001194, -0.00000929, 0.00000506, 0.00001455, 0.00000602, -0.00001278, -0.00002020, -0.00000635, 0.00001323, 0.00001456, -0.00000564, -0.00002403, -0.00001740, 0.00000862, 0.00002419, 0.00000997, -0.00001854, -0.00002705, -0.00000210, 0.00003172, 0.00003788, 0.00000845, -0.00002702, -0.00003388, -0.00000913, 0.00001797, 0.00001964, -0.00000149, -0.00001902, -0.00001371, 0.00000752, 0.00002273, 0.00001931, 0.00000324, -0.00001249, -0.00002211, -0.00002699, -0.00002647, -0.00001560, 0.00000573, 0.00002596, 0.00002943, 0.00001231, -0.00001138, -0.00002193, -0.00001226, 0.00000631, 0.00001605, 0.00000914, -0.00000702, -0.00001842, -0.00001668, -0.00000413, 0.00000995, 0.00001658, 0.00001249, 0.00000145, -0.00000844, -0.00001027, -0.00000346, 0.00000581, 0.00000992, 0.00000636, -0.00000042, -0.00000383, -0.00000160, 0.00000259, 0.00000317, -0.00000187, -0.00000945, -0.00001415, -0.00001250, -0.00000518, 0.00000356, 0.00000874, 0.00000787, 0.00000246, -0.00000344, -0.00000684, -0.00000793, -0.00000849, -0.00000843, -0.00000512, 0.00000293, 0.00001234, },
		{ -0.00000101, -0.00000395, -0.00000085, 0.00000333, 0.00000403, 0.00000167, -0.00000034, -0.00000059, -0.00000095, -0.00000285, -0.00000436, -0.00000260, 0.00000159, 0.00000383, 0.00000156, -0.00000209, -0.00000173, 0.00000330, 0.00000752, 0.00000566, -0.00000081, -0.00000523, -0.00000369, 0.00000077, 0.00000244, 0.00000008, -0.00000246, -0.00000210, -0.00000052, -0.00000094, -0.00000275, -0.00000225, 0.00000143, 0.00000389, 0.00000122, -0.00000375, -0.00000410, 0.00000165, 0.00000609, 0.00000185, -0.00000755, -0.00001020, -0.00000088, 0.00001023, 0.00000830, -0.00000702, -0.00001851, -0.00001115, 0.00000807, 0.00001600, 0.00000082, -0.00002140, -0.00002345, 0.00000026, 0.00002434, 0.00002035, -0.00000890, -0.00002993, -0.00001587, 0.00002135, 0.00004216, 0.00002266, -0.00001851, -0.00004020, -0.00002289, 0.00001203, 0.00002776, 0.00001181, -0.00001389, -0.00002007, -0.00000157, 0.00002079, 0.00002499, 0.00000892, -0.00001286, -0.00002689, -0.00003035, -0.00002553, -0.00001257, 0.00000757, 0.00002619, 0.00002964, 0.00001291, -0.00001173, -0.00002375, -0.00001366, 0.00000765, 0.00001967, 0.00001195, -0.00000733, -0.00002081, -0.00001816, -0.00000331, 0.00001145, 0.00001642, 0.00001040, -0.00000060, -0.00000864, -0.00000878, -0.00000177, 0.00000662, 0.00000985, 0.00000588, -0.00000111, -0.00000468, -0.00000220, 0.00000281, 0.00000439, -0.00000010, -0.00000784, -0.00001347, -0.00001332, -0.00000755, 0.00000070, 0.00000728, 0.00000909, 0.00000562, -0.00000069, -0.00000621, -0.00000886, -0.00000918, -0.00000837, -0.00000584, 0.00000006, 0.00000864, 0.00001523, 0.00001478, 0.00000736, },
		{ 0.00000150, 0.00000462, 0.00000350, 0.00000025, -0.00000173, -0.00000196, -0.00000226, -0.00000310, -0.00000267, -0.00000001, 0.00000264, 0.00000223, -0.00000065, -0.00000177, 0.00000146, 0.00000602, 0.00000641, 0.00000143, -0.00000392, -0.00000425, -0.00000023, 0.00000258, 0.00000083, -0.00000272, -0.00000339, -0.00000091, 0.00000065, -0.00000103, -0.00000306, -0.00000161, 0.00000214, 0.00000311, -0.00000043, -0.00000372, -0.00000174, 0.00000324, 0.00000371, -0.00000270, -0.00000848, -0.00000521, 0.00000432, 0.00000760, -0.00000191, -0.00001430, -0.00001352, 0.00000190, 0.00001407, 0.00000600, -0.00001576, -0.00002556, -0.00000835, 0.00001966, 0.00002680, 0.00000275, -0.00002649, -0.00002603, 0.00000748, 0.00003931, 0.00003374, -0.00000596, -0.00004004, -0.00003489, 0.00000158, 0.00003046, 0.00002458, -0.00000439, -0.00002237, -0.00001095, 0.00001543, 0.00002841, 0.00001563, -0.00001075, -0.00003067, -0.00003464, -0.00002559, -0.00000931, 0.00001017, 0.00002635, 0.00002902, 0.00001320, -0.00001130, -0.00002469, -0.00001525, 0.00000794, 0.00002275, 0.00001539, -0.00000673, -0.00002323, -0.00002056, -0.00000310, 0.00001346, 0.00001722, 0.00000853, -0.00000346, -0.00000971, -0.00000725, 0.00000069, 0.00000799, 0.00000962, 0.00000489, -0.00000204, -0.00000536, -0.00000263, 0.00000286, 0.00000518, 0.00000129, -0.00000631, -0.00001229, -0.00001306, -0.00000878, -0.00000198, 0.00000452, 0.00000833, 0.00000777, 0.00000294, -0.00000373, -0.00000880, -0.00001043, -0.00000923, -0.00000634, -0.00000157, 0.00000547, 0.00001261, 0.00001536, 0.00001103, 0.00000255, -0.00000349, -0.00000308, },
		{ 0.00000187, -0.00000178, -0.00000315, -0.00000269, -0.00000227, -0.00000208, -0.00000088, 0.00000120, 0.00000205, 0.00000059, -0.00000101, 0.00000032, 0.00000408, 0.00000602, 0.00000321, -0.00000188, -0.00000393, -0.00000130, 0.00000199, 0.00000139, -0.00000230, -0.00000417, -0.00000187, 0.00000124, 0.00000076, -0.00000256, -0.00000372, -0.00000043, 0.00000333, 0.00000253, -0.00000179, -0.00000361, -0.00000024, 0.00000343, 0.00000132, -0.00000490, -0.00000708, -0.00000166, 0.00000420, 0.00000118, -0.00000881, -0.00001293, -0.00000359, 0.00000937, 0.00000846, -0.00000893, -0.00002347, -0.00001490, 0.00001179, 0.00002805, 0.00001350, -0.00001792, -0.00003042, -0.00000663, 0.00003004, 0.00003901, 0.00000754, -0.00003356, -0.00004238, -0.00001093, 0.00002723, 0.00003424, 0.00000777, -0.00001974, -0.00001854, 0.00000755, 0.00002859, 0.00002200, -0.00000637, -0.00003254, -0.00003915, -0.00002687, -0.00000634, 0.00001346, 0.00002680, 0.00002767, 0.00001283, -0.00001035, -0.00002443, -0.00001651, 0.00000719, 0.00002470, 0.00001896, -0.00000500, -0.00002504, -0.00002365, -0.00000393, 0.00001544, 0.00001905, 0.00000743, -0.00000675, -0.00001188, -0.00000621, 0.00000371, 0.00001017, 0.00000955, 0.00000342, -0.00000338, -0.00000598, -0.00000282, 0.00000287, 0.00000558, 0.00000226, -0.00000493, -0.00001083, -0.00001202, -0.00000891, -0.00000391, 0.00000129, 0.00000583, 0.00000819, 0.00000636, 0.00000025, -0.00000703, -0.00001134, -0.00001099, -0.00000738, -0.00000245, 0.00000335, 0.00000967, 0.00001406, 0.00001311, 0.00000650, -0.00000116, -0.00000406, -0.00000098, 0.00000370, 0.00000486, },
		{ -0.00000384, -0.00000241, -0.00000145, -0.00000093, 0.00000011, 0.00000127, 0.00000118, 0.00000004, 0.00000004, 0.00000233, 0.00000477, 0.00000410, 0.00000032, -0.00000270, -0.00000192, 0.00000091, 0.00000141, -0.00000157, -0.00000423, -0.00000288, 0.00000088, 0.00000198, -0.00000127, -0.00000452, -0.00000293, 0.00000208, 0.00000431, 0.00000088, -0.00000368, -0.00000325, 0.00000148, 0.00000360, -0.00000068, -0.00000645, -0.00000635, -0.00000091, 0.00000155, -0.00000360, -0.00000976, -0.00000702, 0.00000329, 0.00000762, -0.00000288, -0.00001807, -0.00001780, 0.00000306, 0.00002427, 0.00002080, -0.00000663, -0.00002835, -0.00001773, 0.00001661, 0.00003718, 0.00001857, -0.00002253, -0.00004367, -0.00002248, 0.00001903, 0.00003883, 0.00002016, -0.00001245, -0.00002246, -0.00000121, 0.00002525, 0.00002655, -0.00000052, -0.00003195, -0.00004294, -0.00002919, -0.00000421, 0.00001715, 0.00002782, 0.00002591, 0.00001157, -0.00000929, -0.00002291, -0.00001687, 0.00000577, 0.00002511, 0.00002198, -0.00000227, -0.00002566, -0.00002687, -0.00000599, 0.00001673, 0.00002160, 0.00000753, -0.00000984, -0.00001506, -0.00000616, 0.00000685, 0.00001325, 0.00001004, 0.00000167, -0.00000528, -0.00000671, -0.00000274, 0.00000298, 0.00000564, 0.00000275, -0.00000376, -0.00000923, -0.00001048, -0.00000818, -0.00000490, -0.00000170, 0.00000225, 0.00000668, 0.00000857, 0.00000485, -0.00000341, -0.00001097, -0.00001302, -0.00000934, -0.00000334, 0.00000225, 0.00000724, 0.00001160, 0.00001320, 0.00000973, 0.00000252, -0.00000318, -0.00000310, 0.00000144, 0.00000498, 0.00000381, -0.00000050, -0.00000371, },
		{ -0.00000050, -0.00000020, 0.00000044, 0.00000105, 0.00000081, 0.00000037, 0.00000122, 0.00000320, 0.00000401, 0.00000208, -0.00000088, -0.00000178, -0.00000020, 0.00000084, -0.00000095, -0.00000368, -0.00000349, -0.00000006, 0.00000231, 0.00000015, -0.00000409, -0.00000465, 0.00000001, 0.00000457, 0.00000332, -0.00000224, -0.00000493, -0.00000119, 0.00000389, 0.00000292, -0.00000395, -0.00000886, -0.00000635, -0.00000068, -0.00000005, -0.00000511, -0.00000771, -0.00000253, 0.00000391, 0.00000079, -0.00001107, -0.00001659, -0.00000415, 0.00001690, 0.00002316, 0.00000444, -0.00002078, -0.00002342, 0.00000224, 0.00002871, 0.00002432, -0.00000989, -0.00003866, -0.00003026, 0.00000806, 0.00003757, 0.00003017, -0.00000190, -0.00002161, -0.00000880, 0.00001904, 0.00002812, 0.00000552, -0.00002889, -0.00004507, -0.00003200, -0.00000323, 0.00002071, 0.00002949, 0.00002415, 0.00000946, -0.00000860, -0.00002039, -0.00001591, 0.00000423, 0.00002393, 0.00002373, 0.00000097, -0.00002473, -0.00002949, -0.00000907, 0.00001675, 0.00002428, 0.00000899, -0.00001205, -0.00001886, -0.00000744, 0.00000949, 0.00001706, 0.00001142, -0.00000003, -0.00000776, -0.00000780, -0.00000244, 0.00000335, 0.00000549, 0.00000272, -0.00000291, -0.00000752, -0.00000856, -0.00000686, -0.00000506, -0.00000399, -0.00000160, 0.00000361, 0.00000893, 0.00000896, 0.00000157, -0.00000870, -0.00001436, -0.00001204, -0.00000495, 0.00000162, 0.00000576, 0.00000890, 0.00001157, 0.00001138, 0.00000653, -0.00000039, -0.00000389, -0.00000145, 0.00000346, 0.00000513, 0.00000182, -0.00000297, -0.00000496, -0.00000360, -0.00000147, },
		{ 0.00000048, 0.00000091, 0.00000081, 0.00000085, 0.00000190, 0.00000322, 0.00000301, 0.00000099, -0.00000086, -0.00000087, -0.00000009, -0.00000077, -0.00000286, -0.00000351, -0.00000110, 0.00000178, 0.00000112, -0.00000290, -0.00000521, -0.00000209, 0.00000345, 0.00000476, -0.00000002, -0.00000495, -0.00000354, 0.00000256, 0.00000502, -0.00000069, -0.00000900, -0.00001076, -0.00000469, 0.00000104, -0.00000045, -0.00000586, -0.00000663, -0.00000143, 0.00000136, -0.00000444, -0.00001193, -0.00000810, 0.00000817, 0.00002049, 0.00001263, -0.00000996, -0.00002276, -0.00000974, 0.00001567, 0.00002329, 0.00000098, -0.00002885, -0.00003244, -0.00000280, 0.00003110, 0.00003571, 0.00000963, -0.00001596, -0.00001327, 0.00001143, 0.00002627, 0.00001027, -0.00002398, -0.00004485, -0.00003443, -0.00000342, 0.00002360, 0.00003165, 0.00002279, 0.00000675, -0.00000863, -0.00001734, -0.00001350, 0.00000326, 0.00002145, 0.00002369, 0.00000399, -0.00002229, -0.00003079, -0.00001264, 0.00001517, 0.00002632, 0.00001156, -0.00001279, -0.00002263, -0.00001006, 0.00001102, 0.00002115, 0.00001389, -0.00000119, -0.00001068, -0.00000948, -0.00000212, 0.00000410, 0.00000533, 0.00000215, -0.00000254, -0.00000580, -0.00000628, -0.00000510, -0.00000463, -0.00000549, -0.00000504, -0.00000027, 0.00000739, 0.00001156, 0.00000689, -0.00000452, -0.00001406, -0.00001472, -0.00000753, 0.00000068, 0.00000508, 0.00000677, 0.00000893, 0.00001114, 0.00000978, 0.00000370, -0.00000275, -0.00000393, 0.00000051, 0.00000503, 0.00000438, -0.00000075, -0.00000521, -0.00000540, -0.00000259, -0.00000023, 0.00000028, 0.00000013, },
		{ 0.00000094, 0.00000125, 0.00000227, 0.00000307, 0.00000237, 0.00000050, -0.00000083, -0.00000095, -0.00000107, -0.00000218, -0.00000305, -0.00000183, 0.00000073, 0.00000132, -0.00000154, -0.00000467, -0.00000353, 0.00000155, 0.00000484, 0.00000208, -0.00000356, -0.00000477, 0.00000045, 0.00000530, 0.00000216, -0.00000731, -0.00001323, -0.00000918, -0.00000032, 0.00000288, -0.00000240, -0.00000824, -0.00000681, -0.00000094, 0.00000016, -0.00000542, -0.00000811, 0.00000049, 0.00001403, 0.00001626, 0.00000122, -0.00001647, -0.00001677, 0.00000119, 0.00001572, 0.00000720, -0.00001701, -0.00002868, -0.00001077, 0.00002135, 0.00003576, 0.00001961, -0.00000666, -0.00001332, 0.00000437, 0.00002138, 0.00001250, -0.00001838, -0.00004212, -0.00003555, -0.00000437, 0.00002540, 0.00003387, 0.00002208, 0.00000388, -0.00000949, -0.00001434, -0.00000982, 0.00000339, 0.00001830, 0.00002168, 0.00000598, -0.00001875, -0.00003025, -0.00001590, 0.00001208, 0.00002699, 0.00001464, -0.00001179, -0.00002555, -0.00001371, 0.00001096, 0.00002487, 0.00001739, -0.00000136, -0.00001370, -0.00001189, -0.00000207, 0.00000523, 0.00000542, 0.00000116, -0.00000284, -0.00000425, -0.00000363, -0.00000290, -0.00000386, -0.00000643, -0.00000775, -0.00000413, 0.00000447, 0.00001214, 0.00001137, 0.00000093, -0.00001158, -0.00001633, -0.00001075, -0.00000122, 0.00000456, 0.00000552, 0.00000622, 0.00000931, 0.00001146, 0.00000806, 0.00000031, -0.00000500, -0.00000307, 0.00000313, 0.00000618, 0.00000255, -0.00000395, -0.00000703, -0.00000483, -0.00000082, 0.00000112, 0.00000069, 0.00000004, 0.00000033, 0.00000084, },
		{ 0.00000256, 0.00000290, 0.00000179, -0.00000009, -0.00000130, -0.00000160, -0.00000191, -0.00000244, -0.00000207, -0.00000037, 0.00000082, -0.00000057, -0.00000345, -0.00000398, -0.00000041, 0.00000372, 0.00000335, -0.00000143, -0.00000454, -0.00000148, 0.00000406, 0.00000374, -0.00000478, -0.00001353, -0.00001290, -0.00000346, 0.00000403, 0.00000130, -0.00000738, -0.00001081, -0.00000505, 0.00000184, 0.00000104, -0.00000468, -0.00000429, 0.00000589, 0.00001499, 0.00001003, -0.00000660, -0.00001770, -0.00001140, 0.00000346, 0.00000713, -0.00000640, -0.00002024, -0.00001392, 0.00001100, 0.00003068, 0.00002587, 0.00000427, -0.00000869, -0.00000019, 0.00001474, 0.00001158, -0.00001350, -0.00003733, -0.00003464, -0.00000534, 0.00002599, 0.00003566, 0.00002198, 0.00000132, -0.00001103, -0.00001188, -0.00000538, 0.00000488, 0.00001520, 0.00001794, 0.00000628, -0.00001487, -0.00002778, -0.00001801, 0.00000797, 0.00002581, 0.00001739, -0.00000918, -0.00002689, -0.00001769, 0.00000912, 0.00002747, 0.00002151, -0.00000020, -0.00001631, -0.00001498, -0.00000263, 0.00000660, 0.00000602, 0.00000002, -0.00000389, -0.00000317, -0.00000068, -0.00000016, -0.00000280, -0.00000710, -0.00000978, -0.00000734, 0.00000102, 0.00001077, 0.00001400, 0.00000646, -0.00000703, -0.00001590, -0.00001373, -0.00000427, 0.00000339, 0.00000491, 0.00000420, 0.00000668, 0.00001128, 0.00001158, 0.00000465, -0.00000404, -0.00000617, -0.00000029, 0.00000637, 0.00000603, -0.00000105, -0.00000756, -0.00000766, -0.00000270, 0.00000161, 0.00000204, 0.00000028, -0.00000057, 0.00000013, 0.00000100, 0.00000128, 0.00000168, },
		{ 0.00000100, -0.00000090, -0.00000193, -0.00000205, -0.00000206, -0.00000197, -0.00000117, -0.00000006, -0.00000023, -0.00000214, -0.00000351, -0.00000181, 0.00000193, 0.00000347, 0.00000065, -0.00000307, -0.00000247, 0.00000205, 0.00000371, -0.00000249, -0.00001210, -0.00001495, -0.00000721, 0.00000290, 0.00000398, -0.00000485, -0.00001255, -0.00000948, 0.00000068, 0.00000575, 0.00000077, -0.00000534, -0.00000155, 0.00000987, 0.00001463, 0.00000401, -0.00001302, -0.00001944, -0.00001048, 0.00000079, 0.00000009, -0.00000966, -0.00001175, 0.00000272, 0.00002212, 0.00002725, 0.00001440, -0.00000023, -0.00000083, 0.00000815, 0.00000774, -0.00001059, -0.00003151, -0.00003141, -0.00000546, 0.00002564, 0.00003659, 0.00002225, -0.00000058, -0.00001287, -0.00001025, -0.00000080, 0.00000765, 0.00001285, 0.00001306, 0.00000458, -0.00001153, -0.00002370, -0.00001832, 0.00000365, 0.00002271, 0.00001895, -0.00000551, -0.00002623, -0.00002115, 0.00000575, 0.00002834, 0.00002559, 0.00000236, -0.00001792, -0.00001846, -0.00000408, 0.00000787, 0.00000725, -0.00000091, -0.00000564, -0.00000287, 0.00000229, 0.00000325, -0.00000131, -0.00000773, -0.00001147, -0.00000969, -0.00000208, 0.00000811, 0.00001430, 0.00001078, -0.00000123, -0.00001290, -0.00001527, -0.00000799, 0.00000090, 0.00000418, 0.00000312, 0.00000422, 0.00000962, 0.00001343, 0.00000917, -0.00000109, -0.00000773, -0.00000436, 0.00000456, 0.00000860, 0.00000302, -0.00000634, -0.00001015, -0.00000580, 0.00000102, 0.00000367, 0.00000152, -0.00000119, -0.00000125, 0.00000050, 0.00000169, 0.00000182, 0.00000201, 0.00000259, 0.00000249, },
		{ -0.00000235, -0.00000214, -0.00000185, -0.00000153, -0.00000088, -0.00000047, -0.00000121, -0.00000253, -0.00000235, 0.00000014, 0.00000258, 0.00000201, -0.00000096, -0.00000224, 0.00000016, 0.00000235, -0.00000124, -0.00000983, -0.00001509, -0.00001040, 0.00000009, 0.00000480, -0.00000186, -0.00001197, -0.00001284, -0.00000242, 0.00000773, 0.00000642, -0.00000297, -0.00000640, 0.00000288, 0.00001452, 0.00001264, -0.00000461, -0.00002162, -0.00002281, -0.00001010, 0.00000075, -0.00000007, -0.00000531, -0.00000155, 0.00001250, 0.00002385, 0.00002158, 0.00001025, 0.00000293, 0.00000353, 0.00000201, -0.00001041, -0.00002597, -0.00002620, -0.00000399, 0.00002498, 0.00003647, 0.00002246, -0.00000170, -0.00001456, -0.00000945, 0.00000329, 0.00001128, 0.00001166, 0.00000785, 0.00000100, -0.00000950, -0.00001872, -0.00001654, 0.00000005, 0.00001809, 0.00001869, -0.00000165, -0.00002355, -0.00002323, 0.00000147, 0.00002714, 0.00002885, 0.00000603, -0.00001807, -0.00002183, -0.00000652, 0.00000861, 0.00000902, -0.00000122, -0.00000779, -0.00000362, 0.00000486, 0.00000728, 0.00000091, -0.00000831, -0.00001325, -0.00001137, -0.00000419, 0.00000519, 0.00001248, 0.00001281, 0.00000447, -0.00000753, -0.00001431, -0.00001133, -0.00000304, 0.00000243, 0.00000261, 0.00000268, 0.00000737, 0.00001339, 0.00001268, 0.00000312, -0.00000710, -0.00000786, 0.00000109, 0.00000930, 0.00000722, -0.00000325, -0.00001130, -0.00000947, -0.00000107, 0.00000478, 0.00000360, -0.00000094, -0.00000287, -0.00000083, 0.00000193, 0.00000263, 0.00000195, 0.00000179, 0.00000228, 0.00000191, 0.00000018, -0.00000165, },
		{ -0.00000163, -0.00000132, -0.00000092, -0.00000088, -0.00000160, -0.00000214, -0.00000108, 0.00000122, 0.00000238, 0.00000102, -0.00000099, -0.00000093, 0.00000036, -0.00000133, -0.00000769, -0.00001368, -0.00001220, -0.00000338, 0.00000360, 0.00000039, -0.00000972, -0.00001424, -0.00000609, 0.00000693, 0.00001059, 0.00000154, -0.00000771, -0.00000370, 0.00001054, 0.00001736, 0.00000481, -0.00001825, -0.00003085, -0.00002271, -0.00000451, 0.00000573, 0.00000307, -0.00000120, 0.00000434, 0.00001701, 0.00002454, 0.00002050, 0.00001048, 0.00000237, -0.00000392, -0.00001292, -0.00002198, -0.00001989, -0.00000059, 0.00002480, 0.00003551, 0.00002223, -0.00000220, -0.00001570, -0.00000922, 0.00000647, 0.00001512, 0.00001170, 0.00000307, -0.00000394, -0.00000927, -0.00001376, -0.00001288, -0.00000205, 0.00001271, 0.00001635, 0.00000147, -0.00001929, -0.00002331, -0.00000284, 0.00002395, 0.00003051, 0.00001022, -0.00001653, -0.00002446, -0.00000976, 0.00000841, 0.00001105, -0.00000063, -0.00000988, -0.00000544, 0.00000653, 0.00001166, 0.00000412, -0.00000851, -0.00001537, -0.00001293, -0.00000520, 0.00000301, 0.00000941, 0.00001211, 0.00000860, -0.00000079, -0.00001029, -0.00001293, -0.00000784, -0.00000109, 0.00000181, 0.00000229, 0.00000558, 0.00001189, 0.00001434, 0.00000741, -0.00000435, -0.00000963, -0.00000308, 0.00000775, 0.00001033, 0.00000114, -0.00001038, -0.00001266, -0.00000454, 0.00000454, 0.00000589, 0.00000047, -0.00000404, -0.00000287, 0.00000145, 0.00000374, 0.00000268, 0.00000107, 0.00000122, 0.00000211, 0.00000156, -0.00000054, -0.00000235, -0.00000266, -0.00000209, },
		{ -0.00000103, -0.00000114, -0.00000161, -0.00000149, 0.00000001, 0.00000194, 0.00000229, 0.00000069, -0.00000097, -0.00000145, -0.00000249, -0.00000643, -0.00001155, -0.00001239, -0.00000643, 0.00000092, 0.00000112, -0.00000689, -0.00001354, -0.00000902, 0.00000412, 0.00001227, 0.00000649, -0.00000566, -0.00000797, 0.00000454, 0.00001753, 0.00001254, -0.00001107, -0.00003320, -0.00003383, -0.00001412, 0.00000605, 0.00001048, 0.00000287, -0.00000051, 0.00000891, 0.00002317, 0.00002847, 0.00002027, 0.00000531, -0.00000822, -0.00001723, -0.00002037, -0.00001370, 0.00000453, 0.00002585, 0.00003424, 0.00002136, -0.00000238, -0.00001612, -0.00000919, 0.00000860, 0.00001854, 0.00001267, -0.00000068, -0.00000947, -0.00001087, -0.00000970, -0.00000795, -0.00000220, 0.00000753, 0.00001219, 0.00000304, -0.00001429, -0.00002123, -0.00000623, 0.00001929, 0.00003009, 0.00001410, -0.00001342, -0.00002576, -0.00001339, 0.00000700, 0.00001286, 0.00000093, -0.00001137, -0.00000810, 0.00000686, 0.00001583, 0.00000841, -0.00000787, -0.00001777, -0.00001493, -0.00000547, 0.00000222, 0.00000627, 0.00000901, 0.00000997, 0.00000577, -0.00000348, -0.00001156, -0.00001227, -0.00000648, -0.00000038, 0.00000257, 0.00000506, 0.00000996, 0.00001391, 0.00001049, -0.00000025, -0.00000895, -0.00000661, 0.00000429, 0.00001127, 0.00000567, -0.00000728, -0.00001423, -0.00000865, 0.00000251, 0.00000747, 0.00000284, -0.00000409, -0.00000505, -0.00000000, 0.00000452, 0.00000411, 0.00000097, -0.00000026, 0.00000134, 0.00000266, 0.00000128, -0.00000163, -0.00000330, -0.00000290, -0.00000182, -0.00000129, -0.00000117, },
		{ -0.00000127, -0.00000052, 0.00000123, 0.00000264, 0.00000207, -0.00000023, -0.00000248, -0.00000406, -0.00000628, -0.00000960, -0.00001131, -0.00000830, -0.00000231, 0.00000018, -0.00000458, -0.00001132, -0.00001033, 0.00000055, 0.00001132, 0.00001024, -0.00000139, -0.00000892, -0.00000130, 0.00001387, 0.00001663, -0.00000267, -0.00003007, -0.00004081, -0.00002536, 0.00000078, 0.00001426, 0.00000854, -0.00000143, 0.00000182, 0.00001857, 0.00003282, 0.00003028, 0.00001196, -0.00000939, -0.00002187, -0.00002123, -0.00000889, 0.00001061, 0.00002851, 0.00003341, 0.00002001, -0.00000258, -0.00001588, -0.00000899, 0.00000979, 0.00002101, 0.00001404, -0.00000320, -0.00001474, -0.00001391, -0.00000716, -0.00000265, -0.00000041, 0.00000345, 0.00000687, 0.00000265, -0.00000952, -0.00001727, -0.00000794, 0.00001400, 0.00002750, 0.00001685, -0.00000923, -0.00002532, -0.00001677, 0.00000438, 0.00001390, 0.00000325, -0.00001179, -0.00001109, 0.00000565, 0.00001914, 0.00001352, -0.00000589, -0.00002002, -0.00001776, -0.00000574, 0.00000292, 0.00000427, 0.00000452, 0.00000809, 0.00001045, 0.00000500, -0.00000657, -0.00001477, -0.00001310, -0.00000480, 0.00000249, 0.00000602, 0.00000876, 0.00001194, 0.00001142, 0.00000389, -0.00000587, -0.00000816, -0.00000002, 0.00000958, 0.00000893, -0.00000263, -0.00001334, -0.00001217, -0.00000118, 0.00000740, 0.00000541, -0.00000274, -0.00000660, -0.00000223, 0.00000438, 0.00000576, 0.00000179, -0.00000153, -0.00000030, 0.00000295, 0.00000335, 0.00000001, -0.00000355, -0.00000419, -0.00000249, -0.00000102, -0.00000090, -0.00000127, -0.00000133, -0.00000130, },
		{ 0.00000232, 0.00000264, 0.00000071, -0.00000256, -0.00000526, -0.00000693, -0.00000848, -0.00000975, -0.00000883, -0.00000511, -0.00000195, -0.00000351, -0.00000855, -0.00000987, -0.00000253, 0.00000843, 0.00001179, 0.00000347, -0.00000667, -0.00000511, 0.00000811, 0.00001640, 0.00000428, -0.00002310, -0.00004225, -0.00003513, -0.00000858, 0.00001283, 0.00001310, 0.00000076, -0.00000257, 0.00001261, 0.00003334, 0.00003857, 0.00002103, -0.00000670, -0.00002511, -0.00002387, -0.00000629, 0.00001652, 0.00003268, 0.00003372, 0.00001864, -0.00000300, -0.00001524, -0.00000844, 0.00001031, 0.00002230, 0.00001519, -0.00000460, -0.00001910, -0.00001770, -0.00000633, 0.00000212, 0.00000283, 0.00000106, 0.00000132, 0.00000035, -0.00000588, -0.00001217, -0.00000762, 0.00000907, 0.00002309, 0.00001782, -0.00000472, -0.00002310, -0.00001923, 0.00000088, 0.00001377, 0.00000586, -0.00001088, -0.00001375, 0.00000302, 0.00002093, 0.00001888, -0.00000231, -0.00002143, -0.00002138, -0.00000681, 0.00000461, 0.00000419, 0.00000010, 0.00000333, 0.00001188, 0.00001338, 0.00000179, -0.00001388, -0.00001956, -0.00001165, 0.00000077, 0.00000795, 0.00000924, 0.00000954, 0.00000995, 0.00000661, -0.00000124, -0.00000693, -0.00000366, 0.00000559, 0.00000969, 0.00000227, -0.00000982, -0.00001379, -0.00000568, 0.00000513, 0.00000711, -0.00000021, -0.00000674, -0.00000459, 0.00000303, 0.00000690, 0.00000337, -0.00000199, -0.00000229, 0.00000214, 0.00000503, 0.00000239, -0.00000298, -0.00000558, -0.00000384, -0.00000093, -0.00000009, -0.00000115, -0.00000196, -0.00000162, -0.00000089, -0.00000022, 0.00000089, },
		{ -0.00000199, -0.00000557, -0.00000771, -0.00000834, -0.00000855, -0.00000839, -0.00000683, -0.00000439, -0.00000380, -0.00000625, -0.00000816, -0.00000430, 0.00000475, 0.00001097, 0.00000737, -0.00000225, -0.00000587, 0.00000235, 0.00001247, 0.00000788, -0.00001482, -0.00003834, -0.00004086, -0.00001939, 0.00000615, 0.00001409, 0.00000409, -0.00000375, 0.00000730, 0.00003090, 0.00004388, 0.00003073, -0.00000033, -0.00002550, -0.00002698, -0.00000613, 0.00002109, 0.00003771, 0.00003563, 0.00001795, -0.00000353, -0.00001450, -0.00000757, 0.00001047, 0.00002241, 0.00001559, -0.00000531, -0.00002220, -0.00002145, -0.00000701, 0.00000565, 0.00000675, 0.00000053, -0.00000354, -0.00000335, -0.00000397, -0.00000686, -0.00000536, 0.00000536, 0.00001761, 0.00001673, -0.00000076, -0.00001941, -0.00002022, -0.00000296, 0.00001226, 0.00000817, -0.00000871, -0.00001542, -0.00000055, 0.00002077, 0.00002367, 0.00000274, -0.00002130, -0.00002531, -0.00000925, 0.00000633, 0.00000616, -0.00000283, -0.00000312, 0.00000943, 0.00001970, 0.00001232, -0.00000876, -0.00002406, -0.00002027, -0.00000362, 0.00000965, 0.00001169, 0.00000810, 0.00000665, 0.00000683, 0.00000349, -0.00000287, -0.00000512, 0.00000046, 0.00000734, 0.00000581, -0.00000435, -0.00001251, -0.00000961, 0.00000080, 0.00000684, 0.00000265, -0.00000509, -0.00000613, 0.00000072, 0.00000683, 0.00000513, -0.00000132, -0.00000391, 0.00000041, 0.00000572, 0.00000498, -0.00000137, -0.00000649, -0.00000574, -0.00000145, 0.00000095, -0.00000037, -0.00000258, -0.00000267, -0.00000091, 0.00000067, 0.00000137, 0.00000191, 0.00000226, 0.00000113, },
		{ -0.00000881, -0.00000827, -0.00000776, -0.00000735, -0.00000624, -0.00000495, -0.00000506, -0.00000610, -0.00000462, 0.00000146, 0.00000841, 0.00000932, 0.00000276, -0.00000368, -0.00000159, 0.00000650, 0.00000741, -0.00000781, -0.00003070, -0.00004120, -0.00002861, -0.00000424, 0.00001012, 0.00000602, -0.00000254, 0.00000407, 0.00002710, 0.00004592, 0.00003933, 0.00000864, -0.00002225, -0.00002899, -0.00000785, 0.00002348, 0.00004261, 0.00003909, 0.00001862, -0.00000376, -0.00001386, -0.00000661, 0.00001044, 0.00002155, 0.00001498, -0.00000587, -0.00002404, -0.00002448, -0.00000862, 0.00000759, 0.00001045, 0.00000153, -0.00000701, -0.00000761, -0.00000390, -0.00000229, -0.00000175, 0.00000339, 0.00001200, 0.00001377, 0.00000191, -0.00001492, -0.00001953, -0.00000645, 0.00000952, 0.00000962, -0.00000564, -0.00001566, -0.00000434, 0.00001862, 0.00002703, 0.00000872, -0.00001911, -0.00002870, -0.00001317, 0.00000703, 0.00000959, -0.00000328, -0.00000959, 0.00000344, 0.00002233, 0.00002308, 0.00000045, -0.00002488, -0.00002919, -0.00001108, 0.00000963, 0.00001558, 0.00000873, 0.00000276, 0.00000418, 0.00000668, 0.00000312, -0.00000337, -0.00000411, 0.00000215, 0.00000652, 0.00000165, -0.00000804, -0.00001144, -0.00000471, 0.00000390, 0.00000459, -0.00000190, -0.00000594, -0.00000182, 0.00000521, 0.00000621, 0.00000029, -0.00000445, -0.00000165, 0.00000513, 0.00000703, 0.00000107, -0.00000640, -0.00000771, -0.00000278, 0.00000172, 0.00000103, -0.00000268, -0.00000415, -0.00000173, 0.00000159, 0.00000276, 0.00000205, 0.00000128, 0.00000058, -0.00000144, -0.00000491, -0.00000789, },
		{ -0.00000714, -0.00000693, -0.00000611, -0.00000501, -0.00000463, -0.00000400, -0.00000068, 0.00000522, 0.00000918, 0.00000688, 0.00000046, -0.00000273, 0.00000056, 0.00000352, -0.00000398, -0.00002184, -0.00003642, -0.00003377, -0.00001566, 0.00000133, 0.00000432, -0.00000087, 0.00000324, 0.00002366, 0.00004544, 0.00004568, 0.00001867, -0.00001545, -0.00002853, -0.00001030, 0.00002343, 0.00004631, 0.00004359, 0.00002109, -0.00000307, -0.00001326, -0.00000584, 0.00001021, 0.00001998, 0.00001331, -0.00000673, -0.00002493, -0.00002638, -0.00001046, 0.00000803, 0.00001322, 0.00000344, -0.00000880, -0.00001156, -0.00000534, 0.00000086, 0.00000239, 0.00000324, 0.00000716, 0.00000951, 0.00000286, -0.00001044, -0.00001728, -0.00000896, 0.00000597, 0.00000984, -0.00000227, -0.00001431, -0.00000755, 0.00001484, 0.00002832, 0.00001473, -0.00001480, -0.00003062, -0.00001808, 0.00000589, 0.00001338, -0.00000098, -0.00001439, -0.00000483, 0.00002040, 0.00003184, 0.00001257, -0.00002092, -0.00003639, -0.00002104, 0.00000659, 0.00001964, 0.00001192, -0.00000012, -0.00000079, 0.00000702, 0.00000947, 0.00000174, -0.00000636, -0.00000465, 0.00000358, 0.00000634, -0.00000103, -0.00000993, -0.00000984, -0.00000166, 0.00000430, 0.00000193, -0.00000352, -0.00000347, 0.00000227, 0.00000577, 0.00000217, -0.00000352, -0.00000317, 0.00000343, 0.00000783, 0.00000370, -0.00000508, -0.00000907, -0.00000477, 0.00000170, 0.00000266, -0.00000196, -0.00000552, -0.00000338, 0.00000191, 0.00000471, 0.00000319, 0.00000035, -0.00000099, -0.00000161, -0.00000368, -0.00000700, -0.00000917, -0.00000896, -0.00000774, },
		{ -0.00000558, -0.00000429, -0.00000331, -0.00000159, 0.00000251, 0.00000757, 0.00000918, 0.00000502, -0.00000117, -0.00000357, -0.00000216, -0.00000398, -0.00001434, -0.00002819, -0.00003362, -0.00002513, -0.00001060, -0.00000217, -0.00000116, 0.00000385, 0.00002170, 0.00004380, 0.00004953, 0.00002826, -0.00000596, -0.00002484, -0.00001208, 0.00002139, 0.00004801, 0.00004825, 0.00002539, -0.00000085, -0.00001237, -0.00000546, 0.00000957, 0.00001785, 0.00001073, -0.00000820, -0.00002529, -0.00002707, -0.00001190, 0.00000739, 0.00001465, 0.00000547, -0.00000905, -0.00001447, -0.00000763, 0.00000231, 0.00000616, 0.00000454, 0.00000373, 0.00000479, 0.00000207, -0.00000675, -0.00001396, -0.00001010, 0.00000225, 0.00000876, 0.00000075, -0.00001160, -0.00000955, 0.00001014, 0.00002725, 0.00001977, -0.00000879, -0.00003028, -0.00002306, 0.00000261, 0.00001623, 0.00000351, -0.00001623, -0.00001352, 0.00001413, 0.00003665, 0.00002554, -0.00001208, -0.00003989, -0.00003200, -0.00000011, 0.00002217, 0.00001723, -0.00000063, -0.00000675, 0.00000393, 0.00001432, 0.00000939, -0.00000501, -0.00001122, -0.00000287, 0.00000798, 0.00000700, -0.00000459, -0.00001282, -0.00000894, 0.00000088, 0.00000499, 0.00000100, -0.00000313, -0.00000118, 0.00000333, 0.00000329, -0.00000131, -0.00000330, 0.00000131, 0.00000702, 0.00000564, -0.00000273, -0.00000918, -0.00000684, 0.00000064, 0.00000389, -0.00000048, -0.00000616, -0.00000547, 0.00000123, 0.00000654, 0.00000535, 0.00000018, -0.00000314, -0.00000295, -0.00000263, -0.00000504, -0.00000867, -0.00001014, -0.00000880, -0.00000700, -0.00000648, -0.00000643, },
		{ -0.00000180, 0.00000095, 0.00000557, 0.00000954, 0.00000856, 0.00000206, -0.00000499, -0.00000757, -0.00000713, -0.00001005, -0.00001909, -0.00002855, -0.00003021, -0.00002287, -0.00001305, -0.00000544, 0.00000392, 0.00002126, 0.00004243, 0.00005156, 0.00003642, 0.00000488, -0.00001795, -0.00001194, 0.00001838, 0.00004748, 0.00005212, 0.00003106, 0.00000332, -0.00001065, -0.00000545, 0.00000826, 0.00001516, 0.00000748, -0.00001031, -0.00002557, -0.00002677, -0.00001253, 0.00000627, 0.00001471, 0.00000693, -0.00000826, -0.00001598, -0.00000996, 0.00000223, 0.00000883, 0.00000660, 0.00000193, 0.00000046, -0.00000005, -0.00000437, -0.00001027, -0.00000982, -0.00000101, 0.00000664, 0.00000292, -0.00000804, -0.00000994, 0.00000538, 0.00002401, 0.00002296, -0.00000195, -0.00002738, -0.00002699, -0.00000247, 0.00001704, 0.00000900, -0.00001464, -0.00002065, 0.00000475, 0.00003637, 0.00003693, 0.00000057, -0.00003824, -0.00004185, -0.00001013, 0.00002159, 0.00002338, 0.00000199, -0.00001194, -0.00000216, 0.00001606, 0.00001791, 0.00000035, -0.00001552, -0.00001164, 0.00000543, 0.00001400, 0.00000403, -0.00001213, -0.00001629, -0.00000583, 0.00000571, 0.00000666, -0.00000007, -0.00000386, -0.00000095, 0.00000263, 0.00000145, -0.00000153, -0.00000020, 0.00000476, 0.00000605, -0.00000008, -0.00000770, -0.00000816, -0.00000129, 0.00000407, 0.00000127, -0.00000571, -0.00000729, -0.00000045, 0.00000754, 0.00000805, 0.00000124, -0.00000507, -0.00000542, -0.00000254, -0.00000257, -0.00000688, -0.00001100, -0.00001089, -0.00000774, -0.00000545, -0.00000545, -0.00000587, -0.00000502, -0.00000343, },
		{ 0.00000864, 0.00001020, 0.00000544, -0.00000383, -0.00001097, -0.00001173, -0.00000956, -0.00001172, -0.00002040, -0.00002980, -0.00003242, -0.00002636, -0.00001459, 0.00000108, 0.00002114, 0.00004213, 0.00005299, 0.00004304, 0.00001580, -0.00000858, -0.00000916, 0.00001567, 0.00004511, 0.00005449, 0.00003734, 0.00000945, -0.00000757, -0.00000553, 0.00000611, 0.00001175, 0.00000371, -0.00001294, -0.00002601, -0.00002586, -0.00001223, 0.00000523, 0.00001372, 0.00000739, -0.00000710, -0.00001612, -0.00001167, 0.00000115, 0.00001006, 0.00000861, 0.00000155, -0.00000281, -0.00000285, -0.00000348, -0.00000695, -0.00000842, -0.00000332, 0.00000396, 0.00000397, -0.00000426, -0.00000876, 0.00000138, 0.00001919, 0.00002375, 0.00000462, -0.00002213, -0.00002886, -0.00000843, 0.00001523, 0.00001398, -0.00001003, -0.00002466, -0.00000578, 0.00003095, 0.00004449, 0.00001501, -0.00003105, -0.00004834, -0.00002210, 0.00001692, 0.00002853, 0.00000750, -0.00001469, -0.00000995, 0.00001384, 0.00002525, 0.00000905, -0.00001593, -0.00002078, -0.00000140, 0.00001791, 0.00001442, -0.00000698, -0.00002168, -0.00001500, 0.00000290, 0.00001194, 0.00000573, -0.00000443, -0.00000624, -0.00000051, 0.00000356, 0.00000205, -0.00000006, 0.00000180, 0.00000440, 0.00000183, -0.00000485, -0.00000794, -0.00000344, 0.00000285, 0.00000249, -0.00000426, -0.00000812, -0.00000261, 0.00000722, 0.00001046, 0.00000348, -0.00000596, -0.00000843, -0.00000385, -0.00000046, -0.00000391, -0.00001072, -0.00001339, -0.00000978, -0.00000474, -0.00000330, -0.00000507, -0.00000619, -0.00000480, -0.00000220, 0.00000055, 0.00000425, },
		{ -0.00000122, -0.00001150, -0.00001570, -0.00001210, -0.00000791, -0.00001187, -0.00002446, -0.00003686, -0.00003911, -0.00002789, -0.00000653, 0.00001919, 0.00004258, 0.00005496, 0.00004880, 0.00002607, 0.00000213, -0.00000375, 0.00001442, 0.00004186, 0.00005516, 0.00004336, 0.00001713, -0.00000280, -0.00000526, 0.00000314, 0.00000742, -0.00000056, -0.00001589, -0.00002670, -0.00002471, -0.00001117, 0.00000468, 0.00001219, 0.00000680, -0.00000619, -0.00001526, -0.00001235, -0.00000027, 0.00000990, 0.00000991, 0.00000207, -0.00000471, -0.00000556, -0.00000385, -0.00000454, -0.00000644, -0.00000449, 0.00000131, 0.00000396, -0.00000087, -0.00000634, -0.00000128, 0.00001365, 0.00002211, 0.00000985, -0.00001529, -0.00002805, -0.00001405, 0.00001093, 0.00001708, -0.00000359, -0.00002476, -0.00001527, 0.00002149, 0.00004672, 0.00002865, -0.00001912, -0.00004970, -0.00003385, 0.00000821, 0.00003085, 0.00001473, -0.00001395, -0.00001753, 0.00000783, 0.00002951, 0.00001944, -0.00001171, -0.00002801, -0.00001148, 0.00001724, 0.00002436, 0.00000233, -0.00002327, -0.00002496, -0.00000387, 0.00001503, 0.00001341, -0.00000186, -0.00001115, -0.00000614, 0.00000377, 0.00000667, 0.00000246, -0.00000069, 0.00000077, 0.00000195, -0.00000138, -0.00000567, -0.00000480, 0.00000036, 0.00000235, -0.00000238, -0.00000745, -0.00000438, 0.00000561, 0.00001170, 0.00000630, -0.00000529, -0.00001102, -0.00000647, 0.00000039, -0.00000046, -0.00000886, -0.00001532, -0.00001295, -0.00000521, -0.00000071, -0.00000277, -0.00000683, -0.00000724, -0.00000365, 0.00000070, 0.00000420, 0.00000758, 0.00000993, 0.00000754, },
		{ -0.00001583, -0.00000817, -0.00000566, -0.00001628, -0.00003522, -0.00004806, -0.00004302, -0.00001954, 0.00001301, 0.00004223, 0.00005791, 0.00005480, 0.00003572, 0.00001316, 0.00000365, 0.00001524, 0.00003889, 0.00005453, 0.00004849, 0.00002561, 0.00000365, -0.00000418, -0.00000041, 0.00000204, -0.00000545, -0.00001903, -0.00002752, -0.00002357, -0.00000968, 0.00000471, 0.00001062, 0.00000538, -0.00000593, -0.00001398, -0.00001198, -0.00000142, 0.00000874, 0.00001013, 0.00000282, -0.00000533, -0.00000757, -0.00000498, -0.00000330, -0.00000449, -0.00000464, -0.00000082, 0.00000318, 0.00000171, -0.00000325, -0.00000239, 0.00000830, 0.00001844, 0.00001297, -0.00000797, -0.00002454, -0.00001810, 0.00000496, 0.00001745, 0.00000306, -0.00002114, -0.00002185, 0.00000995, 0.00004326, 0.00003899, -0.00000436, -0.00004516, -0.00004298, -0.00000339, 0.00002904, 0.00002184, -0.00000962, -0.00002301, -0.00000078, 0.00002946, 0.00002934, -0.00000326, -0.00003133, -0.00002284, 0.00001154, 0.00003152, 0.00001438, -0.00001990, -0.00003349, -0.00001398, 0.00001442, 0.00002139, 0.00000411, -0.00001406, -0.00001351, 0.00000109, 0.00001104, 0.00000744, -0.00000140, -0.00000410, -0.00000047, 0.00000152, -0.00000146, -0.00000435, -0.00000268, 0.00000032, -0.00000105, -0.00000533, -0.00000482, 0.00000333, 0.00001116, 0.00000868, -0.00000311, -0.00001220, -0.00000967, -0.00000055, 0.00000239, -0.00000564, -0.00001576, -0.00001644, -0.00000727, 0.00000127, 0.00000074, -0.00000607, -0.00000996, -0.00000651, 0.00000041, 0.00000532, 0.00000742, 0.00000851, 0.00000755, 0.00000128, -0.00000942, -0.00001728, },
		{ -0.00000818, -0.00002831, -0.00005080, -0.00005661, -0.00003681, 0.00000081, 0.00003864, 0.00006110, 0.00006197, 0.00004554, 0.00002403, 0.00001210, 0.00001812, 0.00003722, 0.00005349, 0.00005259, 0.00003411, 0.00001132, -0.00000202, -0.00000413, -0.00000436, -0.00001111, -0.00002232, -0.00002832, -0.00002252, -0.00000809, 0.00000521, 0.00000936, 0.00000358, -0.00000640, -0.00001281, -0.00001087, -0.00000193, 0.00000715, 0.00000925, 0.00000320, -0.00000506, -0.00000854, -0.00000625, -0.00000313, -0.00000309, -0.00000413, -0.00000215, 0.00000212, 0.00000337, -0.00000010, -0.00000209, 0.00000388, 0.00001351, 0.00001368, -0.00000132, -0.00001890, -0.00001973, -0.00000146, 0.00001493, 0.00000835, -0.00001487, -0.00002442, -0.00000137, 0.00003497, 0.00004421, 0.00001062, -0.00003520, -0.00004746, -0.00001585, 0.00002277, 0.00002680, -0.00000265, -0.00002499, -0.00001010, 0.00002494, 0.00003656, 0.00000792, -0.00002963, -0.00003302, 0.00000157, 0.00003401, 0.00002692, -0.00001153, -0.00003844, -0.00002585, 0.00000930, 0.00002767, 0.00001285, -0.00001366, -0.00002117, -0.00000476, 0.00001364, 0.00001421, 0.00000066, -0.00000889, -0.00000554, 0.00000251, 0.00000394, -0.00000135, -0.00000511, -0.00000358, -0.00000135, -0.00000242, -0.00000325, 0.00000145, 0.00000884, 0.00000951, -0.00000016, -0.00001131, -0.00001226, -0.00000313, 0.00000358, -0.00000189, -0.00001419, -0.00001908, -0.00001065, 0.00000166, 0.00000457, -0.00000354, -0.00001183, -0.00001047, -0.00000120, 0.00000685, 0.00000872, 0.00000714, 0.00000557, 0.00000221, -0.00000597, -0.00001569, -0.00001856, -0.00001154, -0.00000360, },
		{ -0.00006519, -0.00005554, -0.00001773, 0.00002923, 0.00006250, 0.00007037, 0.00005662, 0.00003507, 0.00002084, 0.00002250, 0.00003741, 0.00005313, 0.00005606, 0.00004205, 0.00001951, 0.00000113, -0.00000760, -0.00001149, -0.00001761, -0.00002586, -0.00002900, -0.00002149, -0.00000661, 0.00000590, 0.00000853, 0.00000183, -0.00000741, -0.00001211, -0.00000947, -0.00000174, 0.00000566, 0.00000759, 0.00000287, -0.00000446, -0.00000850, -0.00000713, -0.00000368, -0.00000250, -0.00000345, -0.00000268, 0.00000118, 0.00000424, 0.00000265, -0.00000081, 0.00000078, 0.00000822, 0.00001217, 0.00000371, -0.00001211, -0.00001862, -0.00000700, 0.00001013, 0.00001112, -0.00000756, -0.00002286, -0.00001035, 0.00002366, 0.00004358, 0.00002319, -0.00002149, -0.00004613, -0.00002671, 0.00001287, 0.00002803, 0.00000521, -0.00002302, -0.00001805, 0.00001689, 0.00003954, 0.00001964, -0.00002292, -0.00003974, -0.00001083, 0.00003095, 0.00003740, 0.00000067, -0.00003826, -0.00003720, -0.00000004, 0.00003043, 0.00002284, -0.00000930, -0.00002733, -0.00001322, 0.00001317, 0.00002142, 0.00000588, -0.00001207, -0.00001262, 0.00000056, 0.00000921, 0.00000432, -0.00000562, -0.00000864, -0.00000408, 0.00000013, 0.00000042, 0.00000118, 0.00000543, 0.00000797, 0.00000233, -0.00000837, -0.00001301, -0.00000648, 0.00000242, 0.00000105, -0.00001087, -0.00001973, -0.00001436, -0.00000002, 0.00000742, 0.00000036, -0.00001187, -0.00001452, -0.00000448, 0.00000764, 0.00001122, 0.00000696, 0.00000254, 0.00000081, -0.00000288, -0.00001135, -0.00001849, -0.00001599, -0.00000611, -0.00000306, -0.00001853, -0.00004649, },
		{ 0.00001224, 0.00005923, 0.00007872, 0.00006968, 0.00004724, 0.00002975, 0.00002761, 0.00003938, 0.00005433, 0.00005974, 0.00004933, 0.00002743, 0.00000479, -0.00001062, -0.00001895, -0.00002485, -0.00002977, -0.00002956, -0.00002038, -0.00000533, 0.00000650, 0.00000802, 0.00000043, -0.00000860, -0.00001198, -0.00000827, -0.00000109, 0.00000460, 0.00000561, 0.00000179, -0.00000403, -0.00000776, -0.00000731, -0.00000445, -0.00000271, -0.00000302, -0.00000264, 0.00000068, 0.00000463, 0.00000473, 0.00000094, -0.00000091, 0.00000337, 0.00000903, 0.00000658, -0.00000533, -0.00001509, -0.00001060, 0.00000418, 0.00001096, -0.00000092, -0.00001803, -0.00001559, 0.00001166, 0.00003765, 0.00003123, -0.00000650, -0.00003912, -0.00003375, 0.00000114, 0.00002484, 0.00001189, -0.00001770, -0.00002295, 0.00000707, 0.00003771, 0.00002954, -0.00001242, -0.00004145, -0.00002316, 0.00002271, 0.00004357, 0.00001456, -0.00003247, -0.00004565, -0.00001226, 0.00002847, 0.00003204, -0.00000128, -0.00003035, -0.00002296, 0.00000897, 0.00002738, 0.00001379, -0.00001229, -0.00002036, -0.00000475, 0.00001277, 0.00001199, -0.00000316, -0.00001357, -0.00000945, 0.00000095, 0.00000554, 0.00000350, 0.00000219, 0.00000394, 0.00000294, -0.00000420, -0.00001104, -0.00000920, -0.00000097, 0.00000184, -0.00000681, -0.00001776, -0.00001696, -0.00000343, 0.00000811, 0.00000451, -0.00000968, -0.00001732, -0.00000896, 0.00000669, 0.00001396, 0.00000848, -0.00000015, -0.00000268, -0.00000170, -0.00000570, -0.00001488, -0.00001911, -0.00001198, -0.00000282, -0.00000920, -0.00003623, -0.00006619, -0.00007186, -0.00004102, },
		{ 0.00008439, 0.00006173, 0.00003949, 0.00003299, 0.00004258, 0.00005750, 0.00006461, 0.00005636, 0.00003458, 0.00000823, -0.00001334, -0.00002635, -0.00003253, -0.00003408, -0.00003010, -0.00001913, -0.00000421, 0.00000683, 0.00000763, -0.00000051, -0.00000960, -0.00001224, -0.00000756, -0.00000038, 0.00000402, 0.00000377, 0.00000021, -0.00000407, -0.00000676, -0.00000679, -0.00000499, -0.00000343, -0.00000308, -0.00000238, 0.00000066, 0.00000485, 0.00000614, 0.00000269, -0.00000145, -0.00000048, 0.00000506, 0.00000724, 0.00000037, -0.00000996, -0.00001168, -0.00000158, 0.00000818, 0.00000369, -0.00001143, -0.00001666, 0.00000122, 0.00002798, 0.00003370, 0.00000709, -0.00002773, -0.00003552, -0.00001007, 0.00001766, 0.00001562, -0.00001056, -0.00002398, -0.00000240, 0.00003165, 0.00003573, -0.00000020, -0.00003773, -0.00003293, 0.00001087, 0.00004413, 0.00002751, -0.00002186, -0.00004927, -0.00002516, 0.00002163, 0.00003836, 0.00000912, -0.00002918, -0.00003210, 0.00000126, 0.00003046, 0.00002315, -0.00000880, -0.00002705, -0.00001301, 0.00001323, 0.00002034, 0.00000272, -0.00001680, -0.00001687, -0.00000112, 0.00001082, 0.00000874, 0.00000066, -0.00000193, 0.00000056, -0.00000024, -0.00000622, -0.00000970, -0.00000564, -0.00000044, -0.00000360, -0.00001336, -0.00001700, -0.00000739, 0.00000601, 0.00000733, -0.00000570, -0.00001767, -0.00001342, 0.00000361, 0.00001554, 0.00001134, -0.00000122, -0.00000709, -0.00000319, -0.00000074, -0.00000833, -0.00001896, -0.00001888, -0.00000776, -0.00000352, -0.00002289, -0.00005887, -0.00008185, -0.00006565, -0.00001241, 0.00004839, 0.00008441, },
		{ 0.00003891, 0.00004641, 0.00006245, 0.00007139, 0.00006394, 0.00004090, 0.00001074, -0.00001626, -0.00003352, -0.00004021, -0.00003864, -0.00003073, -0.00001779, -0.00000317, 0.00000683, 0.00000711, -0.00000108, -0.00001015, -0.00001260, -0.00000742, -0.00000000, 0.00000374, 0.00000235, -0.00000148, -0.00000463, -0.00000594, -0.00000582, -0.00000506, -0.00000428, -0.00000362, -0.00000222, 0.00000099, 0.00000517, 0.00000703, 0.00000412, -0.00000123, -0.00000311, 0.00000106, 0.00000609, 0.00000431, -0.00000430, -0.00001032, -0.00000597, 0.00000375, 0.00000557, -0.00000479, -0.00001412, -0.00000602, 0.00001679, 0.00003075, 0.00001700, -0.00001422, -0.00003176, -0.00001845, 0.00000798, 0.00001538, -0.00000358, -0.00002136, -0.00000964, 0.00002286, 0.00003724, 0.00001135, -0.00002942, -0.00003821, -0.00000218, 0.00003900, 0.00003703, -0.00000834, -0.00004716, -0.00003626, 0.00001092, 0.00004024, 0.00001991, -0.00002371, -0.00003876, -0.00000884, 0.00002960, 0.00003220, -0.00000171, -0.00003100, -0.00002299, 0.00000975, 0.00002760, 0.00001163, -0.00001689, -0.00002503, -0.00000664, 0.00001459, 0.00001638, 0.00000221, -0.00000818, -0.00000522, 0.00000172, 0.00000064, -0.00000677, -0.00000991, -0.00000583, -0.00000297, -0.00000767, -0.00001360, -0.00001013, 0.00000145, 0.00000732, -0.00000130, -0.00001500, -0.00001621, -0.00000105, 0.00001474, 0.00001432, -0.00000009, -0.00001074, -0.00000697, 0.00000180, -0.00000059, -0.00001478, -0.00002403, -0.00001642, -0.00000356, -0.00001029, -0.00004468, -0.00008268, -0.00008704, -0.00004262, 0.00002807, 0.00008390, 0.00009891, 0.00007915, 0.00005140, },
		{ 0.00008027, 0.00007292, 0.00004685, 0.00001186, -0.00002008, -0.00004066, -0.00004753, -0.00004309, -0.00003146, -0.00001649, -0.00000222, 0.00000655, 0.00000636, -0.00000145, -0.00001019, -0.00001275, -0.00000771, -0.00000019, 0.00000342, 0.00000140, -0.00000289, -0.00000550, -0.00000556, -0.00000479, -0.00000465, -0.00000491, -0.00000443, -0.00000234, 0.00000143, 0.00000565, 0.00000764, 0.00000512, -0.00000067, -0.00000458, -0.00000234, 0.00000378, 0.00000623, 0.00000084, -0.00000711, -0.00000825, -0.00000109, 0.00000479, 0.00000043, -0.00000929, -0.00000934, 0.00000636, 0.00002368, 0.00002190, -0.00000117, -0.00002340, -0.00002237, -0.00000207, 0.00001123, 0.00000138, -0.00001630, -0.00001351, 0.00001338, 0.00003424, 0.00002011, -0.00001831, -0.00003816, -0.00001387, 0.00002935, 0.00004140, 0.00000556, -0.00003964, -0.00004337, -0.00000174, 0.00003708, 0.00002893, -0.00001485, -0.00004152, -0.00001958, 0.00002459, 0.00003908, 0.00000801, -0.00003105, -0.00003293, 0.00000237, 0.00003202, 0.00002236, -0.00001294, -0.00003214, -0.00001532, 0.00001529, 0.00002511, 0.00000755, -0.00001297, -0.00001388, 0.00000016, 0.00000789, 0.00000000, -0.00001183, -0.00001336, -0.00000619, -0.00000259, -0.00000679, -0.00000984, -0.00000424, 0.00000358, 0.00000161, -0.00000978, -0.00001578, -0.00000584, 0.00001101, 0.00001570, 0.00000279, -0.00001208, -0.00001160, 0.00000106, 0.00000600, -0.00000735, -0.00002522, -0.00002601, -0.00000960, -0.00000214, -0.00002704, -0.00007360, -0.00010062, -0.00007423, -0.00000164, 0.00007373, 0.00010983, 0.00009885, 0.00006691, 0.00004655, 0.00005070, 0.00006858, },
		{ 0.00001164, -0.00002550, -0.00004828, -0.00005434, -0.00004700, -0.00003209, -0.00001534, -0.00000145, 0.00000604, 0.00000534, -0.00000179, -0.00000979, -0.00001251, -0.00000815, -0.00000099, 0.00000275, 0.00000075, -0.00000381, -0.00000634, -0.00000562, -0.00000404, -0.00000399, -0.00000512, -0.00000520, -0.00000273, 0.00000175, 0.00000621, 0.00000813, 0.00000576, -0.00000006, -0.00000515, -0.00000482, 0.00000103, 0.00000634, 0.00000467, -0.00000297, -0.00000826, -0.00000512, 0.00000208, 0.00000332, -0.00000379, -0.00000902, -0.00000154, 0.00001449, 0.00002163, 0.00000910, -0.00001240, -0.00002123, -0.00001024, 0.00000429, 0.00000310, -0.00001056, -0.00001387, 0.00000514, 0.00002790, 0.00002478, -0.00000671, -0.00003316, -0.00002203, 0.00001732, 0.00004011, 0.00001724, -0.00002822, -0.00004518, -0.00001392, 0.00002942, 0.00003434, -0.00000432, -0.00003982, -0.00002897, 0.00001615, 0.00004234, 0.00001869, -0.00002684, -0.00004089, -0.00000796, 0.00003232, 0.00003309, -0.00000494, -0.00003641, -0.00002596, 0.00001188, 0.00003304, 0.00001648, -0.00001450, -0.00002393, -0.00000576, 0.00001343, 0.00000996, -0.00000971, -0.00002117, -0.00001364, -0.00000026, 0.00000224, -0.00000520, -0.00000899, -0.00000371, 0.00000109, -0.00000353, -0.00001129, -0.00000871, 0.00000489, 0.00001388, 0.00000595, -0.00001034, -0.00001499, -0.00000251, 0.00000933, 0.00000121, -0.00002158, -0.00003324, -0.00001986, -0.00000088, -0.00001049, -0.00005639, -0.00010305, -0.00010175, -0.00003814, 0.00005165, 0.00011277, 0.00011833, 0.00008671, 0.00005780, 0.00005618, 0.00007532, 0.00009083, 0.00008383, 0.00005321, },
		{ -0.00006090, -0.00005001, -0.00003224, -0.00001435, -0.00000102, 0.00000524, 0.00000412, -0.00000219, -0.00000913, -0.00001183, -0.00000852, -0.00000225, 0.00000155, 0.00000011, -0.00000421, -0.00000686, -0.00000596, -0.00000377, -0.00000339, -0.00000494, -0.00000566, -0.00000322, 0.00000182, 0.00000669, 0.00000856, 0.00000615, 0.00000044, -0.00000515, -0.00000630, -0.00000153, 0.00000513, 0.00000682, 0.00000110, -0.00000637, -0.00000751, -0.00000145, 0.00000372, 0.00000086, -0.00000613, -0.00000598, 0.00000535, 0.00001715, 0.00001508, -0.00000113, -0.00001565, -0.00001482, -0.00000359, 0.00000132, -0.00000595, -0.00001156, -0.00000051, 0.00002002, 0.00002513, 0.00000321, -0.00002465, -0.00002543, 0.00000541, 0.00003387, 0.00002471, -0.00001524, -0.00004157, -0.00002333, 0.00001883, 0.00003514, 0.00000576, -0.00003411, -0.00003531, 0.00000579, 0.00004132, 0.00002839, -0.00001894, -0.00004532, -0.00001959, 0.00002805, 0.00004181, 0.00000619, -0.00003645, -0.00003674, 0.00000421, 0.00003819, 0.00002781, -0.00001152, -0.00003328, -0.00001583, 0.00001523, 0.00002142, -0.00000270, -0.00002691, -0.00002449, -0.00000248, 0.00001140, 0.00000407, -0.00001047, -0.00001318, -0.00000420, 0.00000146, -0.00000305, -0.00000765, -0.00000195, 0.00000799, 0.00000725, -0.00000594, -0.00001518, -0.00000714, 0.00000830, 0.00000806, -0.00001400, -0.00003544, -0.00003101, -0.00000684, 0.00000053, -0.00003510, -0.00009342, -0.00011963, -0.00007641, 0.00001773, 0.00010341, 0.00013331, 0.00010989, 0.00007455, 0.00006467, 0.00008277, 0.00010227, 0.00009655, 0.00006073, 0.00001064, -0.00003284, -0.00005712, },
		{ -0.00001326, -0.00000106, 0.00000403, 0.00000273, -0.00000260, -0.00000832, -0.00001085, -0.00000867, -0.00000373, -0.00000016, -0.00000076, -0.00000428, -0.00000689, -0.00000628, -0.00000397, -0.00000313, -0.00000455, -0.00000566, -0.00000354, 0.00000169, 0.00000693, 0.00000888, 0.00000639, 0.00000079, -0.00000486, -0.00000694, -0.00000350, 0.00000322, 0.00000730, 0.00000432, -0.00000335, -0.00000793, -0.00000466, 0.00000213, 0.00000367, -0.00000213, -0.00000696, -0.00000192, 0.00001021, 0.00001635, 0.00000816, -0.00000722, -0.00001501, -0.00001035, -0.00000318, -0.00000380, -0.00000806, -0.00000312, 0.00001247, 0.00002195, 0.00000990, -0.00001470, -0.00002403, -0.00000410, 0.00002445, 0.00002705, -0.00000320, -0.00003359, -0.00002836, 0.00000749, 0.00003140, 0.00001346, -0.00002574, -0.00003765, -0.00000458, 0.00003626, 0.00003539, -0.00000875, -0.00004537, -0.00003052, 0.00001972, 0.00004684, 0.00001878, -0.00003173, -0.00004557, -0.00000686, 0.00003903, 0.00003959, -0.00000377, -0.00003969, -0.00002873, 0.00001189, 0.00003200, 0.00000891, -0.00002836, -0.00003681, -0.00001009, 0.00001808, 0.00001695, -0.00000682, -0.00002248, -0.00001441, 0.00000271, 0.00000736, -0.00000136, -0.00000709, -0.00000158, 0.00000456, -0.00000060, -0.00001100, -0.00001029, 0.00000336, 0.00001054, -0.00000500, -0.00003144, -0.00003909, -0.00001801, 0.00000297, -0.00001520, -0.00007371, -0.00012378, -0.00010980, -0.00002501, 0.00007883, 0.00013827, 0.00013340, 0.00009780, 0.00007880, 0.00009213, 0.00011395, 0.00011029, 0.00006972, 0.00000970, -0.00004190, -0.00006779, -0.00006782, -0.00005206, -0.00003145, },
		{ 0.00000108, -0.00000295, -0.00000738, -0.00000968, -0.00000863, -0.00000523, -0.00000222, -0.00000197, -0.00000426, -0.00000647, -0.00000635, -0.00000446, -0.00000335, -0.00000424, -0.00000527, -0.00000350, 0.00000151, 0.00000688, 0.00000898, 0.00000652, 0.00000105, -0.00000443, -0.00000695, -0.00000475, 0.00000118, 0.00000648, 0.00000621, -0.00000008, -0.00000659, -0.00000667, -0.00000051, 0.00000430, 0.00000152, -0.00000533, -0.00000626, 0.00000282, 0.00001360, 0.00001388, 0.00000194, -0.00001113, -0.00001430, -0.00000887, -0.00000456, -0.00000502, -0.00000318, 0.00000667, 0.00001672, 0.00001281, -0.00000542, -0.00001886, -0.00000975, 0.00001415, 0.00002455, 0.00000578, -0.00002322, -0.00002849, -0.00000237, 0.00002424, 0.00001748, -0.00001657, -0.00003596, -0.00001317, 0.00002828, 0.00003862, 0.00000191, -0.00004113, -0.00003891, 0.00000868, 0.00004720, 0.00003073, -0.00002268, -0.00005064, -0.00001963, 0.00003482, 0.00004957, 0.00000792, -0.00004137, -0.00004228, 0.00000313, 0.00003919, 0.00002358, -0.00002405, -0.00004797, -0.00002266, 0.00001991, 0.00003124, 0.00000277, -0.00002879, -0.00002833, -0.00000169, 0.00001733, 0.00001022, -0.00000791, -0.00001310, -0.00000350, 0.00000310, -0.00000271, -0.00000933, -0.00000348, 0.00000713, 0.00000196, -0.00002212, -0.00004068, -0.00003046, -0.00000381, -0.00000215, -0.00004868, -0.00011278, -0.00013152, -0.00007064, 0.00003880, 0.00012771, 0.00015192, 0.00012665, 0.00010125, 0.00010597, 0.00012607, 0.00012388, 0.00007970, 0.00000956, -0.00005192, -0.00008050, -0.00007585, -0.00005350, -0.00002940, -0.00001165, -0.00000153, 0.00000219, },
		{ -0.00000841, -0.00000853, -0.00000668, -0.00000440, -0.00000345, -0.00000434, -0.00000579, -0.00000609, -0.00000500, -0.00000399, -0.00000422, -0.00000468, -0.00000303, 0.00000150, 0.00000660, 0.00000877, 0.00000647, 0.00000123, -0.00000396, -0.00000657, -0.00000531, -0.00000057, 0.00000486, 0.00000670, 0.00000269, -0.00000409, -0.00000707, -0.00000314, 0.00000307, 0.00000378, -0.00000237, -0.00000750, -0.00000326, 0.00000830, 0.00001548, 0.00000971, -0.00000448, -0.00001454, -0.00001393, -0.00000773, -0.00000373, -0.00000188, 0.00000328, 0.00001113, 0.00001237, 0.00000157, -0.00001168, -0.00001112, 0.00000518, 0.00001856, 0.00001050, -0.00001275, -0.00002435, -0.00000901, 0.00001548, 0.00001749, -0.00000849, -0.00003113, -0.00001873, 0.00001903, 0.00003784, 0.00001122, -0.00003355, -0.00004349, -0.00000321, 0.00004286, 0.00004008, -0.00001066, -0.00005084, -0.00003194, 0.00002594, 0.00005571, 0.00002175, -0.00003744, -0.00005393, -0.00001013, 0.00004100, 0.00003886, -0.00001373, -0.00005521, -0.00003844, 0.00001535, 0.00004403, 0.00001769, -0.00002960, -0.00004349, -0.00001250, 0.00002383, 0.00002566, -0.00000244, -0.00002377, -0.00001696, 0.00000237, 0.00000790, -0.00000253, -0.00000910, -0.00000198, 0.00000345, -0.00001035, -0.00003407, -0.00003933, -0.00001767, 0.00000006, -0.00002487, -0.00008865, -0.00013624, -0.00011119, -0.00001344, 0.00009762, 0.00015862, 0.00015755, 0.00013364, 0.00012787, 0.00014027, 0.00013638, 0.00008938, 0.00001032, -0.00006187, -0.00009481, -0.00008561, -0.00005506, -0.00002611, -0.00000901, -0.00000212, -0.00000040, -0.00000102, -0.00000324, -0.00000622, },
		{ -0.00000654, -0.00000501, -0.00000457, -0.00000508, -0.00000557, -0.00000539, -0.00000485, -0.00000456, -0.00000414, -0.00000223, 0.00000179, 0.00000626, 0.00000825, 0.00000619, 0.00000134, -0.00000346, -0.00000595, -0.00000531, -0.00000183, 0.00000295, 0.00000601, 0.00000444, -0.00000118, -0.00000593, -0.00000492, 0.00000076, 0.00000423, 0.00000057, -0.00000628, -0.00000694, 0.00000226, 0.00001339, 0.00001455, 0.00000308, -0.00001121, -0.00001677, -0.00001207, -0.00000474, -0.00000062, 0.00000207, 0.00000657, 0.00000973, 0.00000549, -0.00000439, -0.00000887, -0.00000087, 0.00001107, 0.00001090, -0.00000426, -0.00001748, -0.00001164, 0.00000715, 0.00001416, -0.00000292, -0.00002467, -0.00002084, 0.00001022, 0.00003369, 0.00001780, -0.00002417, -0.00004383, -0.00001399, 0.00003471, 0.00004538, 0.00000238, -0.00004604, -0.00004165, 0.00001371, 0.00005670, 0.00003537, -0.00002818, -0.00006137, -0.00002586, 0.00003646, 0.00005188, 0.00000148, -0.00005640, -0.00005467, 0.00000425, 0.00005232, 0.00003590, -0.00002335, -0.00005669, -0.00002893, 0.00002422, 0.00004222, 0.00000999, -0.00003038, -0.00003423, -0.00000492, 0.00001784, 0.00001034, -0.00001014, -0.00001465, -0.00000279, -0.00000032, -0.00001988, -0.00004020, -0.00003409, -0.00000985, -0.00000907, -0.00005675, -0.00012166, -0.00013822, -0.00007114, 0.00004704, 0.00014639, 0.00018422, 0.00017539, 0.00016160, 0.00015989, 0.00014793, 0.00009701, 0.00001113, -0.00007094, -0.00010968, -0.00009716, -0.00005765, -0.00002209, -0.00000503, -0.00000230, -0.00000363, -0.00000381, -0.00000359, -0.00000477, -0.00000700, -0.00000849, -0.00000817, },
		{ -0.00000449, -0.00000502, -0.00000562, -0.00000571, -0.00000514, -0.00000380, -0.00000129, 0.00000240, 0.00000604, 0.00000753, 0.00000562, 0.00000129, -0.00000297, -0.00000520, -0.00000492, -0.00000256, 0.00000115, 0.00000454, 0.00000503, 0.00000142, -0.00000371, -0.00000539, -0.00000170, 0.00000304, 0.00000248, -0.00000370, -0.00000792, -0.00000286, 0.00000887, 0.00001583, 0.00000965, -0.00000536, -0.00001650, -0.00001599, -0.00000774, -0.00000061, 0.00000213, 0.00000377, 0.00000635, 0.00000649, 0.00000141, -0.00000440, -0.00000337, 0.00000411, 0.00000793, 0.00000092, -0.00000989, -0.00001051, 0.00000093, 0.00000890, -0.00000046, -0.00001824, -0.00001993, 0.00000321, 0.00002743, 0.00002104, -0.00001472, -0.00004038, -0.00002211, 0.00002430, 0.00004605, 0.00001436, -0.00003708, -0.00004716, 0.00000017, 0.00005228, 0.00004640, -0.00001501, -0.00006306, -0.00004142, 0.00002592, 0.00006007, 0.00001928, -0.00005056, -0.00006821, -0.00001209, 0.00005385, 0.00005435, -0.00001001, -0.00006474, -0.00004866, 0.00001693, 0.00005640, 0.00002841, -0.00003014, -0.00005236, -0.00001950, 0.00002361, 0.00002767, -0.00000395, -0.00002727, -0.00001711, 0.00000361, -0.00000117, -0.00003041, -0.00004720, -0.00002993, -0.00000664, -0.00002479, -0.00008943, -0.00014469, -0.00012487, -0.00002096, 0.00010966, 0.00019829, 0.00022278, 0.00020985, 0.00018964, 0.00016041, 0.00010109, 0.00001008, -0.00007909, -0.00012373, -0.00010982, -0.00006192, -0.00001829, 0.00000018, -0.00000146, -0.00000711, -0.00000746, -0.00000430, -0.00000304, -0.00000530, -0.00000853, -0.00000982, -0.00000865, -0.00000645, -0.00000484, },
		{ -0.00000643, -0.00000577, -0.00000367, -0.00000046, 0.00000317, 0.00000604, 0.00000681, 0.00000484, 0.00000105, -0.00000255, -0.00000439, -0.00000429, -0.00000281, -0.00000030, 0.00000272, 0.00000458, 0.00000322, -0.00000105, -0.00000452, -0.00000351, 0.00000084, 0.00000287, -0.00000099, -0.00000668, -0.00000598, 0.00000355, 0.00001383, 0.00001376, 0.00000139, -0.00001316, -0.00001809, -0.00001173, -0.00000238, 0.00000228, 0.00000269, 0.00000348, 0.00000542, 0.00000485, 0.00000062, -0.00000266, -0.00000076, 0.00000325, 0.00000251, -0.00000346, -0.00000679, -0.00000225, 0.00000342, -0.00000080, -0.00001321, -0.00001708, -0.00000129, 0.00002053, 0.00002108, -0.00000667, -0.00003426, -0.00002672, 0.00001346, 0.00004243, 0.00002353, -0.00002559, -0.00004775, -0.00001243, 0.00004324, 0.00005301, -0.00000013, -0.00005863, -0.00005410, 0.00001098, 0.00006177, 0.00003670, -0.00003817, -0.00007630, -0.00003107, 0.00004766, 0.00006961, 0.00000887, -0.00006527, -0.00006831, 0.00000206, 0.00006482, 0.00005022, -0.00002149, -0.00006761, -0.00004029, 0.00002216, 0.00004626, 0.00001049, -0.00003561, -0.00003756, -0.00000194, 0.00001721, -0.00000996, -0.00005126, -0.00005522, -0.00002000, -0.00000117, -0.00004526, -0.00012685, -0.00016423, -0.00009906, 0.00004612, 0.00019073, 0.00026871, 0.00027285, 0.00023474, 0.00017787, 0.00010136, 0.00000464, -0.00008752, -0.00013587, -0.00012213, -0.00006788, -0.00001584, 0.00000592, 0.00000075, -0.00001012, -0.00001180, -0.00000571, -0.00000124, -0.00000320, -0.00000850, -0.00001174, -0.00001085, -0.00000767, -0.00000493, -0.00000400, -0.00000465, -0.00000582, },
		{ 0.00000017, 0.00000389, 0.00000621, 0.00000624, 0.00000401, 0.00000062, -0.00000228, -0.00000362, -0.00000352, -0.00000269, -0.00000131, 0.00000093, 0.00000339, 0.00000401, 0.00000141, -0.00000266, -0.00000418, -0.00000153, 0.00000184, 0.00000089, -0.00000419, -0.00000677, -0.00000108, 0.00000964, 0.00001475, 0.00000731, -0.00000761, -0.00001751, -0.00001532, -0.00000574, 0.00000149, 0.00000262, 0.00000184, 0.00000349, 0.00000590, 0.00000472, 0.00000016, -0.00000279, -0.00000137, 0.00000115, 0.00000050, -0.00000212, -0.00000241, -0.00000077, -0.00000295, -0.00001030, -0.00001364, -0.00000336, 0.00001431, 0.00001875, -0.00000093, -0.00002693, -0.00002774, 0.00000386, 0.00003562, 0.00002885, -0.00001353, -0.00004369, -0.00002218, 0.00003127, 0.00005429, 0.00001396, -0.00004892, -0.00006182, -0.00000584, 0.00005660, 0.00005076, -0.00002109, -0.00007718, -0.00004942, 0.00003438, 0.00007870, 0.00003037, -0.00005737, -0.00008420, -0.00001857, 0.00006501, 0.00007173, -0.00000465, -0.00007636, -0.00006448, 0.00001177, 0.00006199, 0.00003214, -0.00003595, -0.00006026, -0.00001824, 0.00002981, 0.00001826, -0.00004227, -0.00007882, -0.00004786, 0.00000692, 0.00000207, -0.00008538, -0.00017997, -0.00017645, -0.00004195, 0.00015378, 0.00030326, 0.00034718, 0.00029943, 0.00020625, 0.00009964, -0.00000759, -0.00009880, -0.00014598, -0.00013221, -0.00007458, -0.00001559, 0.00001094, 0.00000417, -0.00001192, -0.00001627, -0.00000801, 0.00000019, -0.00000075, -0.00000820, -0.00001386, -0.00001331, -0.00000873, -0.00000466, -0.00000345, -0.00000455, -0.00000622, -0.00000704, -0.00000621, -0.00000360, },
		{ 0.00000587, 0.00000331, 0.00000011, -0.00000222, -0.00000300, -0.00000273, -0.00000230, -0.00000185, -0.00000059, 0.00000181, 0.00000385, 0.00000320, -0.00000035, -0.00000366, -0.00000334, -0.00000010, 0.00000145, -0.00000155, -0.00000559, -0.00000397, 0.00000470, 0.00001286, 0.00001107, -0.00000129, -0.00001425, -0.00001726, -0.00000983, -0.00000075, 0.00000258, 0.00000146, 0.00000181, 0.00000519, 0.00000705, 0.00000367, -0.00000214, -0.00000447, -0.00000182, 0.00000160, 0.00000184, -0.00000042, -0.00000278, -0.00000553, -0.00000945, -0.00001078, -0.00000367, 0.00000954, 0.00001517, 0.00000230, -0.00001976, -0.00002584, -0.00000337, 0.00002714, 0.00003011, -0.00000274, -0.00003608, -0.00002782, 0.00001848, 0.00005042, 0.00002506, -0.00003574, -0.00006351, -0.00002175, 0.00004553, 0.00005915, -0.00000214, -0.00007055, -0.00006391, 0.00001612, 0.00007983, 0.00005092, -0.00004187, -0.00009317, -0.00004175, 0.00005613, 0.00008896, 0.00001833, -0.00007605, -0.00008809, -0.00000735, 0.00007087, 0.00005796, -0.00002610, -0.00008023, -0.00004400, 0.00003190, 0.00004898, -0.00001904, -0.00009353, -0.00008522, -0.00000470, 0.00004238, -0.00002578, -0.00016613, -0.00024061, -0.00014719, 0.00008306, 0.00031521, 0.00042534, 0.00038538, 0.00025245, 0.00010046, -0.00002771, -0.00011647, -0.00015545, -0.00013847, -0.00008008, -0.00001758, 0.00001374, 0.00000789, -0.00001198, -0.00001997, -0.00001102, 0.00000077, 0.00000174, -0.00000742, -0.00001593, -0.00001612, -0.00000987, -0.00000400, -0.00000258, -0.00000464, -0.00000704, -0.00000775, -0.00000639, -0.00000335, 0.00000064, 0.00000435, 0.00000635, },
		{ -0.00000231, -0.00000264, -0.00000206, -0.00000175, -0.00000198, -0.00000170, 0.00000018, 0.00000295, 0.00000408, 0.00000185, -0.00000218, -0.00000414, -0.00000223, 0.00000071, 0.00000035, -0.00000328, -0.00000478, 0.00000037, 0.00000900, 0.00001205, 0.00000425, -0.00000907, -0.00001680, -0.00001346, -0.00000428, 0.00000169, 0.00000180, 0.00000101, 0.00000373, 0.00000749, 0.00000659, 0.00000032, -0.00000533, -0.00000486, 0.00000029, 0.00000394, 0.00000240, -0.00000254, -0.00000730, -0.00000995, -0.00000918, -0.00000321, 0.00000640, 0.00001144, 0.00000341, -0.00001376, -0.00002207, -0.00000777, 0.00001856, 0.00002789, 0.00000542, -0.00002657, -0.00002907, 0.00000691, 0.00004254, 0.00003174, -0.00002145, -0.00005932, -0.00003425, 0.00003066, 0.00006077, 0.00001558, -0.00005758, -0.00007210, -0.00000401, 0.00007275, 0.00006701, -0.00002114, -0.00009337, -0.00006357, 0.00003916, 0.00009860, 0.00004393, -0.00006582, -0.00010685, -0.00003300, 0.00007000, 0.00008354, -0.00000609, -0.00009259, -0.00007549, 0.00002083, 0.00007575, 0.00001623, -0.00009354, -0.00012444, -0.00003605, 0.00006628, 0.00004243, -0.00012166, -0.00027962, -0.00025781, -0.00002077, 0.00029412, 0.00049614, 0.00049022, 0.00032281, 0.00011085, -0.00005443, -0.00014415, -0.00016744, -0.00014038, -0.00008184, -0.00002069, 0.00001313, 0.00001031, -0.00001044, -0.00002184, -0.00001404, 0.00000018, 0.00000365, -0.00000623, -0.00001757, -0.00001910, -0.00001132, -0.00000314, -0.00000124, -0.00000465, -0.00000829, -0.00000882, -0.00000642, -0.00000272, 0.00000116, 0.00000447, 0.00000622, 0.00000560, 0.00000292, -0.00000027, },
		{ -0.00000117, -0.00000175, -0.00000232, -0.00000127, 0.00000158, 0.00000406, 0.00000351, -0.00000020, -0.00000380, -0.00000387, -0.00000088, 0.00000104, -0.00000084, -0.00000381, -0.00000237, 0.00000449, 0.00001042, 0.00000781, -0.00000326, -0.00001392, -0.00001549, -0.00000834, -0.00000049, 0.00000201, 0.00000114, 0.00000243, 0.00000653, 0.00000812, 0.00000340, -0.00000401, -0.00000670, -0.00000230, 0.00000377, 0.00000465, -0.00000070, -0.00000750, -0.00001077, -0.00000889, -0.00000289, 0.00000453, 0.00000828, 0.00000314, -0.00000936, -0.00001759, -0.00000953, 0.00001111, 0.00002330, 0.00001032, -0.00001691, -0.00002646, -0.00000188, 0.00003243, 0.00003355, -0.00000836, -0.00005052, -0.00004170, 0.00001463, 0.00005583, 0.00002932, -0.00004063, -0.00007286, -0.00002261, 0.00005885, 0.00007608, 0.00000137, -0.00008468, -0.00008031, 0.00001678, 0.00009873, 0.00006790, -0.00004685, -0.00011719, -0.00006134, 0.00005840, 0.00010408, 0.00002168, -0.00009376, -0.00010730, -0.00000331, 0.00009241, 0.00005837, -0.00007598, -0.00015682, -0.00008279, 0.00006724, 0.00010746, -0.00005103, -0.00028464, -0.00035929, -0.00015120, 0.00023275, 0.00054616, 0.00060677, 0.00042114, 0.00013954, -0.00008326, -0.00018429, -0.00018658, -0.00013922, -0.00007743, -0.00002248, 0.00000880, 0.00000942, -0.00000837, -0.00002110, -0.00001587, -0.00000143, 0.00000426, -0.00000507, -0.00001849, -0.00002185, -0.00001314, -0.00000244, 0.00000052, -0.00000429, -0.00000978, -0.00001042, -0.00000657, -0.00000169, 0.00000200, 0.00000439, 0.00000564, 0.00000518, 0.00000282, -0.00000032, -0.00000241, -0.00000259, -0.00000166, },
		{ -0.00000237, 0.00000002, 0.00000327, 0.00000439, 0.00000179, -0.00000253, -0.00000461, -0.00000270, 0.00000051, 0.00000093, -0.00000180, -0.00000320, 0.00000057, 0.00000702, 0.00000880, 0.00000180, -0.00000932, -0.00001528, -0.00001185, -0.00000379, 0.00000134, 0.00000174, 0.00000187, 0.00000501, 0.00000813, 0.00000596, -0.00000122, -0.00000667, -0.00000488, 0.00000172, 0.00000536, 0.00000164, -0.00000608, -0.00001093, -0.00000939, -0.00000324, 0.00000325, 0.00000597, 0.00000230, -0.00000651, -0.00001331, -0.00000931, 0.00000549, 0.00001763, 0.00001204, -0.00000854, -0.00002119, -0.00000710, 0.00002202, 0.00003104, 0.00000172, -0.00003907, -0.00004358, 0.00000008, 0.00004582, 0.00003734, -0.00002267, -0.00006656, -0.00003675, 0.00004075, 0.00007699, 0.00002198, -0.00006868, -0.00008926, -0.00000726, 0.00008931, 0.00008626, -0.00002207, -0.00011710, -0.00008779, 0.00003728, 0.00011555, 0.00005297, -0.00008241, -0.00013359, -0.00003750, 0.00009456, 0.00010027, -0.00004164, -0.00017460, -0.00013706, 0.00004312, 0.00015755, 0.00003625, -0.00025194, -0.00043687, -0.00029622, 0.00012914, 0.00056201, 0.00072319, 0.00054701, 0.00019523, -0.00010646, -0.00023680, -0.00021800, -0.00013852, -0.00006554, -0.00001961, 0.00000188, 0.00000344, -0.00000776, -0.00001772, -0.00001505, -0.00000323, 0.00000303, -0.00000477, -0.00001869, -0.00002378, -0.00001509, -0.00000225, 0.00000235, -0.00000340, -0.00001113, -0.00001252, -0.00000716, -0.00000046, 0.00000335, 0.00000440, 0.00000458, 0.00000436, 0.00000285, 0.00000010, -0.00000227, -0.00000276, -0.00000167, -0.00000076, -0.00000126, -0.00000244, },
		{ 0.00000443, 0.00000342, -0.00000068, -0.00000430, -0.00000416, -0.00000086, 0.00000157, 0.00000036, -0.00000239, -0.00000190, 0.00000303, 0.00000741, 0.00000500, -0.00000417, -0.00001283, -0.00001381, -0.00000750, -0.00000056, 0.00000207, 0.00000206, 0.00000371, 0.00000703, 0.00000733, 0.00000200, -0.00000482, -0.00000632, -0.00000124, 0.00000420, 0.00000331, -0.00000359, -0.00000985, -0.00000989, -0.00000429, 0.00000194, 0.00000431, 0.00000146, -0.00000478, -0.00000975, -0.00000793, 0.00000185, 0.00001203, 0.00001125, -0.00000238, -0.00001472, -0.00000880, 0.00001297, 0.00002549, 0.00000786, -0.00002717, -0.00004050, -0.00001097, 0.00003298, 0.00003912, -0.00000659, -0.00005489, -0.00004461, 0.00002168, 0.00007024, 0.00003757, -0.00004831, -0.00008930, -0.00002900, 0.00007214, 0.00009614, 0.00000445, -0.00010658, -0.00010804, 0.00000981, 0.00011563, 0.00008260, -0.00005974, -0.00014939, -0.00007647, 0.00008067, 0.00013448, 0.00000508, -0.00017271, -0.00018906, -0.00000325, 0.00018341, 0.00012732, -0.00018394, -0.00047833, -0.00043988, -0.00001215, 0.00053299, 0.00082433, 0.00069458, 0.00028455, -0.00011376, -0.00029797, -0.00026597, -0.00014382, -0.00004687, -0.00000857, -0.00000484, -0.00000849, -0.00001123, -0.00001275, -0.00001036, -0.00000373, -0.00000005, -0.00000637, -0.00001871, -0.00002439, -0.00001656, -0.00000267, 0.00000371, -0.00000212, -0.00001194, -0.00001478, -0.00000837, 0.00000058, 0.00000516, 0.00000481, 0.00000329, 0.00000295, 0.00000270, 0.00000094, -0.00000166, -0.00000295, -0.00000213, -0.00000071, -0.00000067, -0.00000209, -0.00000299, -0.00000150, 0.00000192, },
		{ -0.00000310, -0.00000486, -0.00000253, 0.00000108, 0.00000192, -0.00000060, -0.00000261, -0.00000040, 0.00000441, 0.00000585, 0.00000025, -0.00000879, -0.00001366, -0.00001065, -0.00000351, 0.00000148, 0.00000257, 0.00000311, 0.00000554, 0.00000739, 0.00000466, -0.00000180, -0.00000606, -0.00000396, 0.00000156, 0.00000351, -0.00000100, -0.00000756, -0.00000964, -0.00000562, 0.00000025, 0.00000288, 0.00000084, -0.00000366, -0.00000704, -0.00000612, -0.00000009, 0.00000729, 0.00000888, 0.00000131, -0.00000841, -0.00000767, 0.00000635, 0.00001854, 0.00001005, -0.00001674, -0.00003391, -0.00001746, 0.00001981, 0.00003544, 0.00000545, -0.00004038, -0.00004575, 0.00000476, 0.00005772, 0.00004623, -0.00002709, -0.00008107, -0.00004512, 0.00005043, 0.00009642, 0.00002842, -0.00008761, -0.00011898, -0.00001959, 0.00010420, 0.00010564, -0.00002930, -0.00015178, -0.00011383, 0.00005240, 0.00015488, 0.00005706, -0.00014994, -0.00022919, -0.00006465, 0.00018008, 0.00020833, -0.00008907, -0.00047644, -0.00056486, -0.00018005, 0.00045357, 0.00089397, 0.00085240, 0.00040992, -0.00009405, -0.00036011, -0.00033231, -0.00016183, -0.00002481, 0.00001321, -0.00000722, -0.00002555, -0.00002140, -0.00000852, -0.00000135, -0.00000097, -0.00000413, -0.00001080, -0.00001962, -0.00002355, -0.00001666, -0.00000340, 0.00000407, -0.00000097, -0.00001200, -0.00001665, -0.00001008, 0.00000100, 0.00000710, 0.00000581, 0.00000214, 0.00000101, 0.00000206, 0.00000195, -0.00000045, -0.00000286, -0.00000291, -0.00000118, -0.00000022, -0.00000135, -0.00000306, -0.00000278, 0.00000023, 0.00000373, 0.00000445, 0.00000133, },
		{ -0.00000022, 0.00000248, 0.00000135, -0.00000181, -0.00000251, 0.00000088, 0.00000452, 0.00000297, -0.00000427, -0.00001144, -0.00001234, -0.00000683, -0.00000031, 0.00000274, 0.00000318, 0.00000436, 0.00000651, 0.00000618, 0.00000147, -0.00000419, -0.00000548, -0.00000168, 0.00000205, 0.00000071, -0.00000462, -0.00000825, -0.00000657, -0.00000176, 0.00000129, 0.00000033, -0.00000281, -0.00000499, -0.00000435, -0.00000085, 0.00000374, 0.00000588, 0.00000278, -0.00000325, -0.00000480, 0.00000254, 0.00001176, 0.00000905, -0.00000901, -0.00002569, -0.00001941, 0.00000845, 0.00002798, 0.00001236, -0.00002575, -0.00004108, -0.00000771, 0.00004222, 0.00004756, -0.00000834, -0.00006669, -0.00005364, 0.00002798, 0.00008781, 0.00004630, -0.00006367, -0.00011938, -0.00004624, 0.00008335, 0.00011850, 0.00000390, -0.00014042, -0.00014353, 0.00001423, 0.00015796, 0.00010609, -0.00010912, -0.00025004, -0.00013092, 0.00014794, 0.00026710, 0.00001967, -0.00043050, -0.00065548, -0.00035825, 0.00032515, 0.00091772, 0.00100448, 0.00056795, -0.00003757, -0.00041206, -0.00041496, -0.00019891, -0.00000539, 0.00004616, -0.00000051, -0.00004494, -0.00003999, -0.00000828, 0.00001115, 0.00000694, -0.00000745, -0.00001840, -0.00002290, -0.00002179, -0.00001458, -0.00000360, 0.00000317, -0.00000075, -0.00001152, -0.00001754, -0.00001179, 0.00000060, 0.00000860, 0.00000727, 0.00000156, -0.00000110, 0.00000076, 0.00000272, 0.00000123, -0.00000222, -0.00000374, -0.00000218, -0.00000014, -0.00000041, -0.00000258, -0.00000365, -0.00000157, 0.00000242, 0.00000479, 0.00000315, -0.00000130, -0.00000467, -0.00000397, },
		{ -0.00000013, -0.00000304, -0.00000213, 0.00000177, 0.00000357, -0.00000043, -0.00000778, -0.00001207, -0.00000960, -0.00000308, 0.00000200, 0.00000351, 0.00000385, 0.00000535, 0.00000647, 0.00000412, -0.00000131, -0.00000532, -0.00000446, -0.00000068, 0.00000089, -0.00000191, -0.00000586, -0.00000658, -0.00000367, -0.00000064, -0.00000035, -0.00000210, -0.00000335, -0.00000279, -0.00000089, 0.00000137, 0.00000296, 0.00000261, 0.00000025, -0.00000134, 0.00000123, 0.00000630, 0.00000610, -0.00000442, -0.00001763, -0.00001770, 0.00000027, 0.00001886, 0.00001422, -0.00001334, -0.00003251, -0.00001464, 0.00002668, 0.00004263, 0.00000548, -0.00004919, -0.00005421, 0.00000835, 0.00007267, 0.00005598, -0.00003888, -0.00011002, -0.00006625, 0.00005686, 0.00011963, 0.00003448, -0.00011762, -0.00016108, -0.00002757, 0.00014356, 0.00014455, -0.00005654, -0.00024789, -0.00019112, 0.00009259, 0.00029544, 0.00012706, -0.00034664, -0.00070052, -0.00052765, 0.00015653, 0.00088591, 0.00113247, 0.00074888, 0.00006158, -0.00044075, -0.00050730, -0.00025933, 0.00000346, 0.00008776, 0.00001940, -0.00006197, -0.00006700, -0.00001562, 0.00002477, 0.00002115, -0.00000752, -0.00002850, -0.00003000, -0.00002042, -0.00000989, -0.00000204, 0.00000133, -0.00000230, -0.00001122, -0.00001714, -0.00001272, -0.00000045, 0.00000912, 0.00000873, 0.00000176, -0.00000285, -0.00000107, 0.00000286, 0.00000302, -0.00000092, -0.00000424, -0.00000356, -0.00000063, 0.00000044, -0.00000162, -0.00000400, -0.00000327, 0.00000067, 0.00000445, 0.00000457, 0.00000075, -0.00000368, -0.00000481, -0.00000184, 0.00000203, 0.00000279, },
		{ -0.00000139, 0.00000224, 0.00000185, -0.00000374, -0.00000992, -0.00001100, -0.00000616, 0.00000013, 0.00000347, 0.00000397, 0.00000451, 0.00000592, 0.00000564, 0.00000171, -0.00000363, -0.00000593, -0.00000378, -0.00000055, -0.00000030, -0.00000307, -0.00000542, -0.00000491, -0.00000272, -0.00000144, -0.00000166, -0.00000197, -0.00000141, -0.00000055, -0.00000004, 0.00000053, 0.00000148, 0.00000207, 0.00000180, 0.00000175, 0.00000277, 0.00000252, -0.00000267, -0.00001106, -0.00001377, -0.00000425, 0.00001011, 0.00001211, -0.00000459, -0.00002239, -0.00001621, 0.00001359, 0.00003354, 0.00001325, -0.00003178, -0.00004805, -0.00000588, 0.00005426, 0.00005711, -0.00001707, -0.00009343, -0.00007729, 0.00002922, 0.00010978, 0.00005790, -0.00008767, -0.00016449, -0.00006634, 0.00011472, 0.00016701, -0.00000047, -0.00022338, -0.00023570, 0.00002354, 0.00029060, 0.00021848, -0.00023674, -0.00069530, -0.00066956, -0.00003700, 0.00079575, 0.00121866, 0.00093718, 0.00020353, -0.00043352, -0.00059831, -0.00034351, -0.00000667, 0.00013220, 0.00005467, -0.00007080, -0.00010009, -0.00003347, 0.00003574, 0.00004140, -0.00000167, -0.00003927, -0.00004184, -0.00002136, -0.00000298, 0.00000258, -0.00000041, -0.00000612, -0.00001221, -0.00001566, -0.00001201, -0.00000148, 0.00000830, 0.00000941, 0.00000258, -0.00000367, -0.00000295, 0.00000216, 0.00000438, 0.00000086, -0.00000407, -0.00000494, -0.00000172, 0.00000088, -0.00000042, -0.00000375, -0.00000465, -0.00000131, 0.00000353, 0.00000546, 0.00000276, -0.00000214, -0.00000492, -0.00000331, 0.00000087, 0.00000337, 0.00000169, -0.00000224, -0.00000393, },
		{ -0.00000039, -0.00000655, -0.00001061, -0.00000868, -0.00000261, 0.00000256, 0.00000427, 0.00000430, 0.00000513, 0.00000601, 0.00000409, -0.00000100, -0.00000571, -0.00000628, -0.00000314, -0.00000032, -0.00000074, -0.00000333, -0.00000502, -0.00000450, -0.00000294, -0.00000171, -0.00000087, -0.00000014, 0.00000001, -0.00000075, -0.00000127, -0.00000004, 0.00000250, 0.00000403, 0.00000321, 0.00000119, -0.00000060, -0.00000292, -0.00000669, -0.00000914, -0.00000553, 0.00000321, 0.00000768, 0.00000008, -0.00001290, -0.00001366, 0.00000444, 0.00002284, 0.00001525, -0.00001708, -0.00003747, -0.00001356, 0.00003601, 0.00005102, -0.00000103, -0.00007313, -0.00007892, 0.00000468, 0.00009163, 0.00007135, -0.00005579, -0.00015448, -0.00009644, 0.00007690, 0.00017123, 0.00005060, -0.00018113, -0.00025841, -0.00004780, 0.00025554, 0.00028252, -0.00011625, -0.00064251, -0.00076883, -0.00023587, 0.00065247, 0.00124910, 0.00111358, 0.00038144, -0.00038071, -0.00067393, -0.00044697, -0.00004267, 0.00017099, 0.00010435, -0.00006564, -0.00013456, -0.00006306, 0.00003956, 0.00006565, 0.00001221, -0.00004781, -0.00005823, -0.00002672, 0.00000477, 0.00001115, -0.00000046, -0.00001201, -0.00001573, -0.00001396, -0.00000912, -0.00000142, 0.00000634, 0.00000850, 0.00000337, -0.00000325, -0.00000413, 0.00000081, 0.00000478, 0.00000261, -0.00000310, -0.00000581, -0.00000317, 0.00000067, 0.00000067, -0.00000299, -0.00000550, -0.00000327, 0.00000212, 0.00000577, 0.00000450, -0.00000033, -0.00000438, -0.00000431, -0.00000058, 0.00000308, 0.00000304, -0.00000070, -0.00000432, -0.00000402, -0.00000029, 0.00000221, },
		{ -0.00000988, -0.00000555, 0.00000062, 0.00000417, 0.00000462, 0.00000465, 0.00000560, 0.00000539, 0.00000172, -0.00000399, -0.00000739, -0.00000597, -0.00000194, 0.00000034, -0.00000098, -0.00000391, -0.00000543, -0.00000457, -0.00000238, -0.00000023, 0.00000103, 0.00000077, -0.00000093, -0.00000245, -0.00000157, 0.00000194, 0.00000516, 0.00000485, 0.00000118, -0.00000257, -0.00000415, -0.00000451, -0.00000507, -0.00000454, -0.00000111, 0.00000266, 0.00000121, -0.00000561, -0.00000881, -0.00000038, 0.00001285, 0.00001283, -0.00000668, -0.00002521, -0.00001502, 0.00002070, 0.00004017, 0.00000800, -0.00005283, -0.00007251, -0.00001363, 0.00006905, 0.00007415, -0.00002696, -0.00013410, -0.00011431, 0.00003667, 0.00015838, 0.00008969, -0.00012858, -0.00025725, -0.00011041, 0.00019806, 0.00031294, -0.00000137, -0.00055158, -0.00081639, -0.00041939, 0.00046886, 0.00121634, 0.00125794, 0.00058174, -0.00027814, -0.00071926, -0.00056003, -0.00010799, 0.00019425, 0.00016377, -0.00004231, -0.00016386, -0.00010315, 0.00003199, 0.00009005, 0.00003486, -0.00005064, -0.00007758, -0.00003822, 0.00001110, 0.00002371, 0.00000307, -0.00001892, -0.00002272, -0.00001351, -0.00000414, 0.00000096, 0.00000405, 0.00000547, 0.00000311, -0.00000175, -0.00000384, -0.00000056, 0.00000392, 0.00000362, -0.00000156, -0.00000574, -0.00000450, -0.00000024, 0.00000123, -0.00000198, -0.00000569, -0.00000491, 0.00000042, 0.00000549, 0.00000582, 0.00000152, -0.00000334, -0.00000472, -0.00000190, 0.00000218, 0.00000360, 0.00000088, -0.00000357, -0.00000549, -0.00000305, 0.00000091, 0.00000148, -0.00000295, -0.00000856, },
		{ 0.00000319, 0.00000500, 0.00000473, 0.00000502, 0.00000563, 0.00000381, -0.00000141, -0.00000683, -0.00000810, -0.00000461, -0.00000025, 0.00000077, -0.00000192, -0.00000514, -0.00000581, -0.00000355, -0.00000024, 0.00000192, 0.00000169, -0.00000066, -0.00000310, -0.00000289, 0.00000081, 0.00000524, 0.00000611, 0.00000213, -0.00000320, -0.00000543, -0.00000404, -0.00000233, -0.00000248, -0.00000290, -0.00000157, -0.00000006, -0.00000124, -0.00000352, -0.00000151, 0.00000523, 0.00000797, -0.00000094, -0.00001378, -0.00001177, 0.00001002, 0.00002751, 0.00001036, -0.00003556, -0.00006072, -0.00002426, 0.00004614, 0.00006768, -0.00000493, -0.00010786, -0.00011901, 0.00000026, 0.00013248, 0.00011260, -0.00007429, -0.00023455, -0.00015581, 0.00012895, 0.00030938, 0.00009387, -0.00043684, -0.00081069, -0.00056910, 0.00026341, 0.00112106, 0.00135263, 0.00078571, -0.00012868, -0.00072130, -0.00066890, -0.00020108, 0.00019270, 0.00022480, 0.00000039, -0.00018087, -0.00014967, 0.00001021, 0.00010945, 0.00006503, -0.00004459, -0.00009687, -0.00005650, 0.00001319, 0.00003906, 0.00001176, -0.00002489, -0.00003335, -0.00001607, 0.00000203, 0.00000659, 0.00000282, 0.00000037, 0.00000064, 0.00000005, -0.00000156, -0.00000096, 0.00000200, 0.00000315, -0.00000010, -0.00000451, -0.00000506, -0.00000159, 0.00000090, -0.00000116, -0.00000528, -0.00000594, -0.00000130, 0.00000467, 0.00000660, 0.00000319, -0.00000202, -0.00000455, -0.00000282, 0.00000105, 0.00000340, 0.00000197, -0.00000225, -0.00000567, -0.00000517, -0.00000135, 0.00000157, -0.00000014, -0.00000550, -0.00000940, -0.00000786, -0.00000208, },
		{ 0.00000476, 0.00000521, 0.00000488, 0.00000122, -0.00000483, -0.00000881, -0.00000742, -0.00000241, 0.00000116, 0.00000020, -0.00000363, -0.00000611, -0.00000484, -0.00000098, 0.00000231, 0.00000265, 0.00000003, -0.00000322, -0.00000391, -0.00000058, 0.00000451, 0.00000671, 0.00000345, -0.00000265, -0.00000612, -0.00000454, -0.00000111, -0.00000043, -0.00000276, -0.00000420, -0.00000237, 0.00000035, 0.00000074, -0.00000022, 0.00000069, 0.00000265, 0.00000084, -0.00000491, -0.00000593, 0.00000431, 0.00001570, 0.00000761, -0.00002316, -0.00004669, -0.00002747, 0.00002632, 0.00005482, 0.00000842, -0.00008061, -0.00011207, -0.00002762, 0.00009932, 0.00011845, -0.00002612, -0.00019611, -0.00017938, 0.00005973, 0.00027697, 0.00015986, -0.00031479, -0.00075757, -0.00067177, 0.00005738, 0.00097222, 0.00138562, 0.00097222, 0.00005738, -0.00067177, -0.00075757, -0.00031479, 0.00015986, 0.00027697, 0.00005973, -0.00017938, -0.00019611, -0.00002612, 0.00011845, 0.00009932, -0.00002762, -0.00011207, -0.00008061, 0.00000842, 0.00005482, 0.00002632, -0.00002747, -0.00004669, -0.00002316, 0.00000761, 0.00001570, 0.00000431, -0.00000593, -0.00000491, 0.00000084, 0.00000265, 0.00000069, -0.00000022, 0.00000074, 0.00000035, -0.00000237, -0.00000420, -0.00000276, -0.00000043, -0.00000111, -0.00000454, -0.00000612, -0.00000265, 0.00000345, 0.00000671, 0.00000451, -0.00000058, -0.00000391, -0.00000322, 0.00000003, 0.00000265, 0.00000231, -0.00000098, -0.00000484, -0.00000611, -0.00000363, 0.00000020, 0.00000116, -0.00000241, -0.00000742, -0.00000881, -0.00000483, 0.00000122, 0.00000488, 0.00000521, },
		{ 0.00000319, -0.00000208, -0.00000786, -0.00000940, -0.00000550, -0.00000014, 0.00000157, -0.00000135, -0.00000517, -0.00000567, -0.00000225, 0.00000197, 0.00000340, 0.00000105, -0.00000282, -0.00000455, -0.00000202, 0.00000319, 0.00000660, 0.00000467, -0.00000130, -0.00000594, -0.00000528, -0.00000116, 0.00000090, -0.00000159, -0.00000506, -0.00000451, -0.00000010, 0.00000315, 0.00000200, -0.00000096, -0.00000156, 0.00000005, 0.00000064, 0.00000037, 0.00000282, 0.00000659, 0.00000203, -0.00001607, -0.00003335, -0.00002489, 0.00001176, 0.00003906, 0.00001319, -0.00005650, -0.00009687, -0.00004459, 0.00006503, 0.00010945, 0.00001021, -0.00014967, -0.00018087, 0.00000039, 0.00022480, 0.00019270, -0.00020108, -0.00066890, -0.00072130, -0.00012868, 0.00078571, 0.00135263, 0.00112106, 0.00026341, -0.00056910, -0.00081069, -0.00043684, 0.00009387, 0.00030938, 0.00012895, -0.00015581, -0.00023455, -0.00007429, 0.00011260, 0.00013248, 0.00000026, -0.00011901, -0.00010786, -0.00000493, 0.00006768, 0.00004614, -0.00002426, -0.00006072, -0.00003556, 0.00001036, 0.00002751, 0.00001002, -0.00001177, -0.00001378, -0.00000094, 0.00000797, 0.00000523, -0.00000151, -0.00000352, -0.00000124, -0.00000006, -0.00000157, -0.00000290, -0.00000248, -0.00000233, -0.00000404, -0.00000543, -0.00000320, 0.00000213, 0.00000611, 0.00000524, 0.00000081, -0.00000289, -0.00000310, -0.00000066, 0.00000169, 0.00000192, -0.00000024, -0.00000355, -0.00000581, -0.00000514, -0.00000192, 0.00000077, -0.00000025, -0.00000461, -0.00000810, -0.00000683, -0.00000141, 0.00000381, 0.00000563, 0.00000502, 0.00000473, 0.00000500, },
		{ -0.00000988, -0.00000856, -0.00000295, 0.00000148, 0.00000091, -0.00000305, -0.00000549, -0.00000357, 0.00000088, 0.00000360, 0.00000218, -0.00000190, -0.00000472, -0.00000334, 0.00000152, 0.00000582, 0.00000549, 0.00000042, -0.00000491, -0.00000569, -0.00000198, 0.00000123, -0.00000024, -0.00000450, -0.00000574, -0.00000156, 0.00000362, 0.00000392, -0.00000056, -0.00000384, -0.00000175, 0.00000311, 0.00000547, 0.00000405, 0.00000096, -0.00000414, -0.00001351, -0.00002272, -0.00001892, 0.00000307, 0.00002371, 0.00001110, -0.00003822, -0.00007758, -0.00005064, 0.00003486, 0.00009005, 0.00003199, -0.00010315, -0.00016386, -0.00004231, 0.00016377, 0.00019425, -0.00010799, -0.00056003, -0.00071926, -0.00027814, 0.00058174, 0.00125794, 0.00121634, 0.00046886, -0.00041939, -0.00081639, -0.00055158, -0.00000137, 0.00031294, 0.00019806, -0.00011041, -0.00025725, -0.00012858, 0.00008969, 0.00015838, 0.00003667, -0.00011431, -0.00013410, -0.00002696, 0.00007415, 0.00006905, -0.00001363, -0.00007251, -0.00005283, 0.00000800, 0.00004017, 0.00002070, -0.00001502, -0.00002521, -0.00000668, 0.00001283, 0.00001285, -0.00000038, -0.00000881, -0.00000561, 0.00000121, 0.00000266, -0.00000111, -0.00000454, -0.00000507, -0.00000451, -0.00000415, -0.00000257, 0.00000118, 0.00000485, 0.00000516, 0.00000194, -0.00000157, -0.00000245, -0.00000093, 0.00000077, 0.00000103, -0.00000023, -0.00000238, -0.00000457, -0.00000543, -0.00000391, -0.00000098, 0.00000034, -0.00000194, -0.00000597, -0.00000739, -0.00000399, 0.00000172, 0.00000539, 0.00000560, 0.00000465, 0.00000462, 0.00000417, 0.00000062, -0.00000555, },
		{ -0.00000039, 0.00000221, -0.00000029, -0.00000402, -0.00000432, -0.00000070, 0.00000304, 0.00000308, -0.00000058, -0.00000431, -0.00000438, -0.00000033, 0.00000450, 0.00000577, 0.00000212, -0.00000327, -0.00000550, -0.00000299, 0.00000067, 0.00000067, -0.00000317, -0.00000581, -0.00000310, 0.00000261, 0.00000478, 0.00000081, -0.00000413, -0.00000325, 0.00000337, 0.00000850, 0.00000634, -0.00000142, -0.00000912, -0.00001396, -0.00001573, -0.00001201, -0.00000046, 0.00001115, 0.00000477, -0.00002672, -0.00005823, -0.00004781, 0.00001221, 0.00006565, 0.00003956, -0.00006306, -0.00013456, -0.00006564, 0.00010435, 0.00017099, -0.00004267, -0.00044697, -0.00067393, -0.00038071, 0.00038144, 0.00111358, 0.00124910, 0.00065247, -0.00023587, -0.00076883, -0.00064251, -0.00011625, 0.00028252, 0.00025554, -0.00004780, -0.00025841, -0.00018113, 0.00005060, 0.00017123, 0.00007690, -0.00009644, -0.00015448, -0.00005579, 0.00007135, 0.00009163, 0.00000468, -0.00007892, -0.00007313, -0.00000103, 0.00005102, 0.00003601, -0.00001356, -0.00003747, -0.00001708, 0.00001525, 0.00002284, 0.00000444, -0.00001366, -0.00001290, 0.00000008, 0.00000768, 0.00000321, -0.00000553, -0.00000914, -0.00000669, -0.00000292, -0.00000060, 0.00000119, 0.00000321, 0.00000403, 0.00000250, -0.00000004, -0.00000127, -0.00000075, 0.00000001, -0.00000014, -0.00000087, -0.00000171, -0.00000294, -0.00000450, -0.00000502, -0.00000333, -0.00000074, -0.00000032, -0.00000314, -0.00000628, -0.00000571, -0.00000100, 0.00000409, 0.00000601, 0.00000513, 0.00000430, 0.00000427, 0.00000256, -0.00000261, -0.00000868, -0.00001061, -0.00000655, },
		{ -0.00000139, -0.00000393, -0.00000224, 0.00000169, 0.00000337, 0.00000087, -0.00000331, -0.00000492, -0.00000214, 0.00000276, 0.00000546, 0.00000353, -0.00000131, -0.00000465, -0.00000375, -0.00000042, 0.00000088, -0.00000172, -0.00000494, -0.00000407, 0.00000086, 0.00000438, 0.00000216, -0.00000295, -0.00000367, 0.00000258, 0.00000941, 0.00000830, -0.00000148, -0.00001201, -0.00001566, -0.00001221, -0.00000612, -0.00000041, 0.00000258, -0.00000298, -0.00002136, -0.00004184, -0.00003927, -0.00000167, 0.00004140, 0.00003574, -0.00003347, -0.00010009, -0.00007080, 0.00005467, 0.00013220, -0.00000667, -0.00034351, -0.00059831, -0.00043352, 0.00020353, 0.00093718, 0.00121866, 0.00079575, -0.00003700, -0.00066956, -0.00069530, -0.00023674, 0.00021848, 0.00029060, 0.00002354, -0.00023570, -0.00022338, -0.00000047, 0.00016701, 0.00011472, -0.00006634, -0.00016449, -0.00008767, 0.00005790, 0.00010978, 0.00002922, -0.00007729, -0.00009343, -0.00001707, 0.00005711, 0.00005426, -0.00000588, -0.00004805, -0.00003178, 0.00001325, 0.00003354, 0.00001359, -0.00001621, -0.00002239, -0.00000459, 0.00001211, 0.00001011, -0.00000425, -0.00001377, -0.00001106, -0.00000267, 0.00000252, 0.00000277, 0.00000175, 0.00000180, 0.00000207, 0.00000148, 0.00000053, -0.00000004, -0.00000055, -0.00000141, -0.00000197, -0.00000166, -0.00000144, -0.00000272, -0.00000491, -0.00000542, -0.00000307, -0.00000030, -0.00000055, -0.00000378, -0.00000593, -0.00000363, 0.00000171, 0.00000564, 0.00000592, 0.00000451, 0.00000397, 0.00000347, 0.00000013, -0.00000616, -0.00001100, -0.00000992, -0.00000374, 0.00000185, 0.00000224, },
		{ -0.00000013, 0.00000279, 0.00000203, -0.00000184, -0.00000481, -0.00000368, 0.00000075, 0.00000457, 0.00000445, 0.00000067, -0.00000327, -0.00000400, -0.00000162, 0.00000044, -0.00000063, -0.00000356, -0.00000424, -0.00000092, 0.00000302, 0.00000286, -0.00000107, -0.00000285, 0.00000176, 0.00000873, 0.00000912, -0.00000045, -0.00001272, -0.00001714, -0.00001122, -0.00000230, 0.00000133, -0.00000204, -0.00000989, -0.00002042, -0.00003000, -0.00002850, -0.00000752, 0.00002115, 0.00002477, -0.00001562, -0.00006700, -0.00006197, 0.00001940, 0.00008776, 0.00000346, -0.00025933, -0.00050730, -0.00044075, 0.00006158, 0.00074888, 0.00113247, 0.00088591, 0.00015653, -0.00052765, -0.00070052, -0.00034664, 0.00012706, 0.00029544, 0.00009259, -0.00019112, -0.00024789, -0.00005654, 0.00014455, 0.00014356, -0.00002757, -0.00016108, -0.00011762, 0.00003448, 0.00011963, 0.00005686, -0.00006625, -0.00011002, -0.00003888, 0.00005598, 0.00007267, 0.00000835, -0.00005421, -0.00004919, 0.00000548, 0.00004263, 0.00002668, -0.00001464, -0.00003251, -0.00001334, 0.00001422, 0.00001886, 0.00000027, -0.00001770, -0.00001763, -0.00000442, 0.00000610, 0.00000630, 0.00000123, -0.00000134, 0.00000025, 0.00000261, 0.00000296, 0.00000137, -0.00000089, -0.00000279, -0.00000335, -0.00000210, -0.00000035, -0.00000064, -0.00000367, -0.00000658, -0.00000586, -0.00000191, 0.00000089, -0.00000068, -0.00000446, -0.00000532, -0.00000131, 0.00000412, 0.00000647, 0.00000535, 0.00000385, 0.00000351, 0.00000200, -0.00000308, -0.00000960, -0.00001207, -0.00000778, -0.00000043, 0.00000357, 0.00000177, -0.00000213, -0.00000304, },
		{ -0.00000022, -0.00000397, -0.00000467, -0.00000130, 0.00000315, 0.00000479, 0.00000242, -0.00000157, -0.00000365, -0.00000258, -0.00000041, -0.00000014, -0.00000218, -0.00000374, -0.00000222, 0.00000123, 0.00000272, 0.00000076, -0.00000110, 0.00000156, 0.00000727, 0.00000860, 0.00000060, -0.00001179, -0.00001754, -0.00001152, -0.00000075, 0.00000317, -0.00000360, -0.00001458, -0.00002179, -0.00002290, -0.00001840, -0.00000745, 0.00000694, 0.00001115, -0.00000828, -0.00003999, -0.00004494, -0.00000051, 0.00004616, -0.00000539, -0.00019891, -0.00041496, -0.00041206, -0.00003757, 0.00056795, 0.00100448, 0.00091772, 0.00032515, -0.00035825, -0.00065548, -0.00043050, 0.00001967, 0.00026710, 0.00014794, -0.00013092, -0.00025004, -0.00010912, 0.00010609, 0.00015796, 0.00001423, -0.00014353, -0.00014042, 0.00000390, 0.00011850, 0.00008335, -0.00004624, -0.00011938, -0.00006367, 0.00004630, 0.00008781, 0.00002798, -0.00005364, -0.00006669, -0.00000834, 0.00004756, 0.00004222, -0.00000771, -0.00004108, -0.00002575, 0.00001236, 0.00002798, 0.00000845, -0.00001941, -0.00002569, -0.00000901, 0.00000905, 0.00001176, 0.00000254, -0.00000480, -0.00000325, 0.00000278, 0.00000588, 0.00000374, -0.00000085, -0.00000435, -0.00000499, -0.00000281, 0.00000033, 0.00000129, -0.00000176, -0.00000657, -0.00000825, -0.00000462, 0.00000071, 0.00000205, -0.00000168, -0.00000548, -0.00000419, 0.00000147, 0.00000618, 0.00000651, 0.00000436, 0.00000318, 0.00000274, -0.00000031, -0.00000683, -0.00001234, -0.00001144, -0.00000427, 0.00000297, 0.00000452, 0.00000088, -0.00000251, -0.00000181, 0.00000135, 0.00000248, },
		{ -0.00000310, 0.00000133, 0.00000445, 0.00000373, 0.00000023, -0.00000278, -0.00000306, -0.00000135, -0.00000022, -0.00000118, -0.00000291, -0.00000286, -0.00000045, 0.00000195, 0.00000206, 0.00000101, 0.00000214, 0.00000581, 0.00000710, 0.00000100, -0.00001008, -0.00001665, -0.00001200, -0.00000097, 0.00000407, -0.00000340, -0.00001666, -0.00002355, -0.00001962, -0.00001080, -0.00000413, -0.00000097, -0.00000135, -0.00000852, -0.00002140, -0.00002555, -0.00000722, 0.00001321, -0.00002481, -0.00016183, -0.00033231, -0.00036011, -0.00009405, 0.00040992, 0.00085240, 0.00089397, 0.00045357, -0.00018005, -0.00056486, -0.00047644, -0.00008907, 0.00020833, 0.00018008, -0.00006465, -0.00022919, -0.00014994, 0.00005706, 0.00015488, 0.00005240, -0.00011383, -0.00015178, -0.00002930, 0.00010564, 0.00010420, -0.00001959, -0.00011898, -0.00008761, 0.00002842, 0.00009642, 0.00005043, -0.00004512, -0.00008107, -0.00002709, 0.00004623, 0.00005772, 0.00000476, -0.00004575, -0.00004038, 0.00000545, 0.00003544, 0.00001981, -0.00001746, -0.00003391, -0.00001674, 0.00001005, 0.00001854, 0.00000635, -0.00000767, -0.00000841, 0.00000131, 0.00000888, 0.00000729, -0.00000009, -0.00000612, -0.00000704, -0.00000366, 0.00000084, 0.00000288, 0.00000025, -0.00000562, -0.00000964, -0.00000756, -0.00000100, 0.00000351, 0.00000156, -0.00000396, -0.00000606, -0.00000180, 0.00000466, 0.00000739, 0.00000554, 0.00000311, 0.00000257, 0.00000148, -0.00000351, -0.00001065, -0.00001366, -0.00000879, 0.00000025, 0.00000585, 0.00000441, -0.00000040, -0.00000261, -0.00000060, 0.00000192, 0.00000108, -0.00000253, -0.00000486, },
		{ 0.00000443, 0.00000192, -0.00000150, -0.00000299, -0.00000209, -0.00000067, -0.00000071, -0.00000213, -0.00000295, -0.00000166, 0.00000094, 0.00000270, 0.00000295, 0.00000329, 0.00000481, 0.00000516, 0.00000058, -0.00000837, -0.00001478, -0.00001194, -0.00000212, 0.00000371, -0.00000267, -0.00001656, -0.00002439, -0.00001871, -0.00000637, -0.00000005, -0.00000373, -0.00001036, -0.00001275, -0.00001123, -0.00000849, -0.00000484, -0.00000857, -0.00004687, -0.00014382, -0.00026597, -0.00029797, -0.00011376, 0.00028455, 0.00069458, 0.00082433, 0.00053299, -0.00001215, -0.00043988, -0.00047833, -0.00018394, 0.00012732, 0.00018341, -0.00000325, -0.00018906, -0.00017271, 0.00000508, 0.00013448, 0.00008067, -0.00007647, -0.00014939, -0.00005974, 0.00008260, 0.00011563, 0.00000981, -0.00010804, -0.00010658, 0.00000445, 0.00009614, 0.00007214, -0.00002900, -0.00008930, -0.00004831, 0.00003757, 0.00007024, 0.00002168, -0.00004461, -0.00005489, -0.00000659, 0.00003912, 0.00003298, -0.00001097, -0.00004050, -0.00002717, 0.00000786, 0.00002549, 0.00001297, -0.00000880, -0.00001472, -0.00000238, 0.00001125, 0.00001203, 0.00000185, -0.00000793, -0.00000975, -0.00000478, 0.00000146, 0.00000431, 0.00000194, -0.00000429, -0.00000989, -0.00000985, -0.00000359, 0.00000331, 0.00000420, -0.00000124, -0.00000632, -0.00000482, 0.00000200, 0.00000733, 0.00000703, 0.00000371, 0.00000206, 0.00000207, -0.00000056, -0.00000750, -0.00001381, -0.00001283, -0.00000417, 0.00000500, 0.00000741, 0.00000303, -0.00000190, -0.00000239, 0.00000036, 0.00000157, -0.00000086, -0.00000416, -0.00000430, -0.00000068, 0.00000342, },
		{ -0.00000237, -0.00000244, -0.00000126, -0.00000076, -0.00000167, -0.00000276, -0.00000227, 0.00000010, 0.00000285, 0.00000436, 0.00000458, 0.00000440, 0.00000335, -0.00000046, -0.00000716, -0.00001252, -0.00001113, -0.00000340, 0.00000235, -0.00000225, -0.00001509, -0.00002378, -0.00001869, -0.00000477, 0.00000303, -0.00000323, -0.00001505, -0.00001772, -0.00000776, 0.00000344, 0.00000188, -0.00001961, -0.00006554, -0.00013852, -0.00021800, -0.00023680, -0.00010646, 0.00019523, 0.00054701, 0.00072319, 0.00056201, 0.00012914, -0.00029622, -0.00043687, -0.00025194, 0.00003625, 0.00015755, 0.00004312, -0.00013706, -0.00017460, -0.00004164, 0.00010027, 0.00009456, -0.00003750, -0.00013359, -0.00008241, 0.00005297, 0.00011555, 0.00003728, -0.00008779, -0.00011710, -0.00002207, 0.00008626, 0.00008931, -0.00000726, -0.00008926, -0.00006868, 0.00002198, 0.00007699, 0.00004075, -0.00003675, -0.00006656, -0.00002267, 0.00003734, 0.00004582, 0.00000008, -0.00004358, -0.00003907, 0.00000172, 0.00003104, 0.00002202, -0.00000710, -0.00002119, -0.00000854, 0.00001204, 0.00001763, 0.00000549, -0.00000931, -0.00001331, -0.00000651, 0.00000230, 0.00000597, 0.00000325, -0.00000324, -0.00000939, -0.00001093, -0.00000608, 0.00000164, 0.00000536, 0.00000172, -0.00000488, -0.00000667, -0.00000122, 0.00000596, 0.00000813, 0.00000501, 0.00000187, 0.00000174, 0.00000134, -0.00000379, -0.00001185, -0.00001528, -0.00000932, 0.00000180, 0.00000880, 0.00000702, 0.00000057, -0.00000320, -0.00000180, 0.00000093, 0.00000051, -0.00000270, -0.00000461, -0.00000253, 0.00000179, 0.00000439, 0.00000327, 0.00000002, },
		{ -0.00000117, -0.00000166, -0.00000259, -0.00000241, -0.00000032, 0.00000282, 0.00000518, 0.00000564, 0.00000439, 0.00000200, -0.00000169, -0.00000657, -0.00001042, -0.00000978, -0.00000429, 0.00000052, -0.00000244, -0.00001314, -0.00002185, -0.00001849, -0.00000507, 0.00000426, -0.00000143, -0.00001587, -0.00002110, -0.00000837, 0.00000942, 0.00000880, -0.00002248, -0.00007743, -0.00013922, -0.00018658, -0.00018429, -0.00008326, 0.00013954, 0.00042114, 0.00060677, 0.00054616, 0.00023275, -0.00015120, -0.00035929, -0.00028464, -0.00005103, 0.00010746, 0.00006724, -0.00008279, -0.00015682, -0.00007598, 0.00005837, 0.00009241, -0.00000331, -0.00010730, -0.00009376, 0.00002168, 0.00010408, 0.00005840, -0.00006134, -0.00011719, -0.00004685, 0.00006790, 0.00009873, 0.00001678, -0.00008031, -0.00008468, 0.00000137, 0.00007608, 0.00005885, -0.00002261, -0.00007286, -0.00004063, 0.00002932, 0.00005583, 0.00001463, -0.00004170, -0.00005052, -0.00000836, 0.00003355, 0.00003243, -0.00000188, -0.00002646, -0.00001691, 0.00001032, 0.00002330, 0.00001111, -0.00000953, -0.00001759, -0.00000936, 0.00000314, 0.00000828, 0.00000453, -0.00000289, -0.00000889, -0.00001077, -0.00000750, -0.00000070, 0.00000465, 0.00000377, -0.00000230, -0.00000670, -0.00000401, 0.00000340, 0.00000812, 0.00000653, 0.00000243, 0.00000114, 0.00000201, -0.00000049, -0.00000834, -0.00001549, -0.00001392, -0.00000326, 0.00000781, 0.00001042, 0.00000449, -0.00000237, -0.00000381, -0.00000084, 0.00000104, -0.00000088, -0.00000387, -0.00000380, -0.00000020, 0.00000351, 0.00000406, 0.00000158, -0.00000127, -0.00000232, -0.00000175, },
		{ -0.00000231, -0.00000027, 0.00000292, 0.00000560, 0.00000622, 0.00000447, 0.00000116, -0.00000272, -0.00000642, -0.00000882, -0.00000829, -0.00000465, -0.00000124, -0.00000314, -0.00001132, -0.00001910, -0.00001757, -0.00000623, 0.00000365, 0.00000018, -0.00001404, -0.00002184, -0.00001044, 0.00001031, 0.00001313, -0.00002069, -0.00008184, -0.00014038, -0.00016744, -0.00014415, -0.00005443, 0.00011085, 0.00032281, 0.00049022, 0.00049614, 0.00029412, -0.00002077, -0.00025781, -0.00027962, -0.00012166, 0.00004243, 0.00006628, -0.00003605, -0.00012444, -0.00009354, 0.00001623, 0.00007575, 0.00002083, -0.00007549, -0.00009259, -0.00000609, 0.00008354, 0.00007000, -0.00003300, -0.00010685, -0.00006582, 0.00004393, 0.00009860, 0.00003916, -0.00006357, -0.00009337, -0.00002114, 0.00006701, 0.00007275, -0.00000401, -0.00007210, -0.00005758, 0.00001558, 0.00006077, 0.00003066, -0.00003425, -0.00005932, -0.00002145, 0.00003174, 0.00004254, 0.00000691, -0.00002907, -0.00002657, 0.00000542, 0.00002789, 0.00001856, -0.00000777, -0.00002207, -0.00001376, 0.00000341, 0.00001144, 0.00000640, -0.00000321, -0.00000918, -0.00000995, -0.00000730, -0.00000254, 0.00000240, 0.00000394, 0.00000029, -0.00000486, -0.00000533, 0.00000032, 0.00000659, 0.00000749, 0.00000373, 0.00000101, 0.00000180, 0.00000169, -0.00000428, -0.00001346, -0.00001680, -0.00000907, 0.00000425, 0.00001205, 0.00000900, 0.00000037, -0.00000478, -0.00000328, 0.00000035, 0.00000071, -0.00000223, -0.00000414, -0.00000218, 0.00000185, 0.00000408, 0.00000295, 0.00000018, -0.00000170, -0.00000198, -0.00000175, -0.00000206, -0.00000264, },
		{ 0.00000587, 0.00000635, 0.00000435, 0.00000064, -0.00000335, -0.00000639, -0.00000775, -0.00000704, -0.00000464, -0.00000258, -0.00000400, -0.00000987, -0.00001612, -0.00001593, -0.00000742, 0.00000174, 0.00000077, -0.00001102, -0.00001997, -0.00001198, 0.00000789, 0.00001374, -0.00001758, -0.00008008, -0.00013847, -0.00015545, -0.00011647, -0.00002771, 0.00010046, 0.00025245, 0.00038538, 0.00042534, 0.00031521, 0.00008306, -0.00014719, -0.00024061, -0.00016613, -0.00002578, 0.00004238, -0.00000470, -0.00008522, -0.00009353, -0.00001904, 0.00004898, 0.00003190, -0.00004400, -0.00008023, -0.00002610, 0.00005796, 0.00007087, -0.00000735, -0.00008809, -0.00007605, 0.00001833, 0.00008896, 0.00005613, -0.00004175, -0.00009317, -0.00004187, 0.00005092, 0.00007983, 0.00001612, -0.00006391, -0.00007055, -0.00000214, 0.00005915, 0.00004553, -0.00002175, -0.00006351, -0.00003574, 0.00002506, 0.00005042, 0.00001848, -0.00002782, -0.00003608, -0.00000274, 0.00003011, 0.00002714, -0.00000337, -0.00002584, -0.00001976, 0.00000230, 0.00001517, 0.00000954, -0.00000367, -0.00001078, -0.00000945, -0.00000553, -0.00000278, -0.00000042, 0.00000184, 0.00000160, -0.00000182, -0.00000447, -0.00000214, 0.00000367, 0.00000705, 0.00000519, 0.00000181, 0.00000146, 0.00000258, -0.00000075, -0.00000983, -0.00001726, -0.00001425, -0.00000129, 0.00001107, 0.00001286, 0.00000470, -0.00000397, -0.00000559, -0.00000155, 0.00000145, -0.00000010, -0.00000334, -0.00000366, -0.00000035, 0.00000320, 0.00000385, 0.00000181, -0.00000059, -0.00000185, -0.00000230, -0.00000273, -0.00000300, -0.00000222, 0.00000011, 0.00000331, },
		{ 0.00000017, -0.00000360, -0.00000621, -0.00000704, -0.00000622, -0.00000455, -0.00000345, -0.00000466, -0.00000873, -0.00001331, -0.00001386, -0.00000820, -0.00000075, 0.00000019, -0.00000801, -0.00001627, -0.00001192, 0.00000417, 0.00001094, -0.00001559, -0.00007458, -0.00013221, -0.00014598, -0.00009880, -0.00000759, 0.00009964, 0.00020625, 0.00029943, 0.00034718, 0.00030326, 0.00015378, -0.00004195, -0.00017645, -0.00017997, -0.00008538, 0.00000207, 0.00000692, -0.00004786, -0.00007882, -0.00004227, 0.00001826, 0.00002981, -0.00001824, -0.00006026, -0.00003595, 0.00003214, 0.00006199, 0.00001177, -0.00006448, -0.00007636, -0.00000465, 0.00007173, 0.00006501, -0.00001857, -0.00008420, -0.00005737, 0.00003037, 0.00007870, 0.00003438, -0.00004942, -0.00007718, -0.00002109, 0.00005076, 0.00005660, -0.00000584, -0.00006182, -0.00004892, 0.00001396, 0.00005429, 0.00003127, -0.00002218, -0.00004369, -0.00001353, 0.00002885, 0.00003562, 0.00000386, -0.00002774, -0.00002693, -0.00000093, 0.00001875, 0.00001431, -0.00000336, -0.00001364, -0.00001030, -0.00000295, -0.00000077, -0.00000241, -0.00000212, 0.00000050, 0.00000115, -0.00000137, -0.00000279, 0.00000016, 0.00000472, 0.00000590, 0.00000349, 0.00000184, 0.00000262, 0.00000149, -0.00000574, -0.00001532, -0.00001751, -0.00000761, 0.00000731, 0.00001475, 0.00000964, -0.00000108, -0.00000677, -0.00000419, 0.00000089, 0.00000184, -0.00000153, -0.00000418, -0.00000266, 0.00000141, 0.00000401, 0.00000339, 0.00000093, -0.00000131, -0.00000269, -0.00000352, -0.00000362, -0.00000228, 0.00000062, 0.00000401, 0.00000624, 0.00000621, 0.00000389, },
		{ -0.00000643, -0.00000582, -0.00000465, -0.00000400, -0.00000493, -0.00000767, -0.00001085, -0.00001174, -0.00000850, -0.00000320, -0.00000124, -0.00000571, -0.00001180, -0.00001012, 0.00000075, 0.00000592, -0.00001584, -0.00006788, -0.00012213, -0.00013587, -0.00008752, 0.00000464, 0.00010136, 0.00017787, 0.00023474, 0.00027285, 0.00026871, 0.00019073, 0.00004612, -0.00009906, -0.00016423, -0.00012685, -0.00004526, -0.00000117, -0.00002000, -0.00005522, -0.00005126, -0.00000996, 0.00001721, -0.00000194, -0.00003756, -0.00003561, 0.00001049, 0.00004626, 0.00002216, -0.00004029, -0.00006761, -0.00002149, 0.00005022, 0.00006482, 0.00000206, -0.00006831, -0.00006527, 0.00000887, 0.00006961, 0.00004766, -0.00003107, -0.00007630, -0.00003817, 0.00003670, 0.00006177, 0.00001098, -0.00005410, -0.00005863, -0.00000013, 0.00005301, 0.00004324, -0.00001243, -0.00004775, -0.00002559, 0.00002353, 0.00004243, 0.00001346, -0.00002672, -0.00003426, -0.00000667, 0.00002108, 0.00002053, -0.00000129, -0.00001708, -0.00001321, -0.00000080, 0.00000342, -0.00000225, -0.00000679, -0.00000346, 0.00000251, 0.00000325, -0.00000076, -0.00000266, 0.00000062, 0.00000485, 0.00000542, 0.00000348, 0.00000269, 0.00000228, -0.00000238, -0.00001173, -0.00001809, -0.00001316, 0.00000139, 0.00001376, 0.00001383, 0.00000355, -0.00000598, -0.00000668, -0.00000099, 0.00000287, 0.00000084, -0.00000351, -0.00000452, -0.00000105, 0.00000322, 0.00000458, 0.00000272, -0.00000030, -0.00000281, -0.00000429, -0.00000439, -0.00000255, 0.00000105, 0.00000484, 0.00000681, 0.00000604, 0.00000317, -0.00000046, -0.00000367, -0.00000577, },
		{ -0.00000449, -0.00000484, -0.00000645, -0.00000865, -0.00000982, -0.00000853, -0.00000530, -0.00000304, -0.00000430, -0.00000746, -0.00000711, -0.00000146, 0.00000018, -0.00001829, -0.00006192, -0.00010982, -0.00012373, -0.00007909, 0.00001008, 0.00010109, 0.00016041, 0.00018964, 0.00020985, 0.00022278, 0.00019829, 0.00010966, -0.00002096, -0.00012487, -0.00014469, -0.00008943, -0.00002479, -0.00000664, -0.00002993, -0.00004720, -0.00003041, -0.00000117, 0.00000361, -0.00001711, -0.00002727, -0.00000395, 0.00002767, 0.00002361, -0.00001950, -0.00005236, -0.00003014, 0.00002841, 0.00005640, 0.00001693, -0.00004866, -0.00006474, -0.00001001, 0.00005435, 0.00005385, -0.00001209, -0.00006821, -0.00005056, 0.00001928, 0.00006007, 0.00002592, -0.00004142, -0.00006306, -0.00001501, 0.00004640, 0.00005228, 0.00000017, -0.00004716, -0.00003708, 0.00001436, 0.00004605, 0.00002430, -0.00002211, -0.00004038, -0.00001472, 0.00002104, 0.00002743, 0.00000321, -0.00001993, -0.00001824, -0.00000046, 0.00000890, 0.00000093, -0.00001051, -0.00000989, 0.00000092, 0.00000793, 0.00000411, -0.00000337, -0.00000440, 0.00000141, 0.00000649, 0.00000635, 0.00000377, 0.00000213, -0.00000061, -0.00000774, -0.00001599, -0.00001650, -0.00000536, 0.00000965, 0.00001583, 0.00000887, -0.00000286, -0.00000792, -0.00000370, 0.00000248, 0.00000304, -0.00000170, -0.00000539, -0.00000371, 0.00000142, 0.00000503, 0.00000454, 0.00000115, -0.00000256, -0.00000492, -0.00000520, -0.00000297, 0.00000129, 0.00000562, 0.00000753, 0.00000604, 0.00000240, -0.00000129, -0.00000380, -0.00000514, -0.00000571, -0.00000562, -0.00000502, },
		{ -0.00000654, -0.00000817, -0.00000849, -0.00000700, -0.00000477, -0.00000359, -0.00000381, -0.00000363, -0.00000230, -0.00000503, -0.00002209, -0.00005765, -0.00009716, -0.00010968, -0.00007094, 0.00001113, 0.00009701, 0.00014793, 0.00015989, 0.00016160, 0.00017539, 0.00018422, 0.00014639, 0.00004704, -0.00007114, -0.00013822, -0.00012166, -0.00005675, -0.00000907, -0.00000985, -0.00003409, -0.00004020, -0.00001988, -0.00000032, -0.00000279, -0.00001465, -0.00001014, 0.00001034, 0.00001784, -0.00000492, -0.00003423, -0.00003038, 0.00000999, 0.00004222, 0.00002422, -0.00002893, -0.00005669, -0.00002335, 0.00003590, 0.00005232, 0.00000425, -0.00005467, -0.00005640, 0.00000148, 0.00005188, 0.00003646, -0.00002586, -0.00006137, -0.00002818, 0.00003537, 0.00005670, 0.00001371, -0.00004165, -0.00004604, 0.00000238, 0.00004538, 0.00003471, -0.00001399, -0.00004383, -0.00002417, 0.00001780, 0.00003369, 0.00001022, -0.00002084, -0.00002467, -0.00000292, 0.00001416, 0.00000715, -0.00001164, -0.00001748, -0.00000426, 0.00001090, 0.00001107, -0.00000087, -0.00000887, -0.00000439, 0.00000549, 0.00000973, 0.00000657, 0.00000207, -0.00000062, -0.00000474, -0.00001207, -0.00001677, -0.00001121, 0.00000308, 0.00001455, 0.00001339, 0.00000226, -0.00000694, -0.00000628, 0.00000057, 0.00000423, 0.00000076, -0.00000492, -0.00000593, -0.00000118, 0.00000444, 0.00000601, 0.00000295, -0.00000183, -0.00000531, -0.00000595, -0.00000346, 0.00000134, 0.00000619, 0.00000825, 0.00000626, 0.00000179, -0.00000223, -0.00000414, -0.00000456, -0.00000485, -0.00000539, -0.00000557, -0.00000508, -0.00000457, -0.00000501, },
		{ -0.00000841, -0.00000622, -0.00000324, -0.00000102, -0.00000040, -0.00000212, -0.00000901, -0.00002611, -0.00005506, -0.00008561, -0.00009481, -0.00006187, 0.00001032, 0.00008938, 0.00013638, 0.00014027, 0.00012787, 0.00013364, 0.00015755, 0.00015862, 0.00009762, -0.00001344, -0.00011119, -0.00013624, -0.00008865, -0.00002487, 0.00000006, -0.00001767, -0.00003933, -0.00003407, -0.00001035, 0.00000345, -0.00000198, -0.00000910, -0.00000253, 0.00000790, 0.00000237, -0.00001696, -0.00002377, -0.00000244, 0.00002566, 0.00002383, -0.00001250, -0.00004349, -0.00002960, 0.00001769, 0.00004403, 0.00001535, -0.00003844, -0.00005521, -0.00001373, 0.00003886, 0.00004100, -0.00001013, -0.00005393, -0.00003744, 0.00002175, 0.00005571, 0.00002594, -0.00003194, -0.00005084, -0.00001066, 0.00004008, 0.00004286, -0.00000321, -0.00004349, -0.00003355, 0.00001122, 0.00003784, 0.00001903, -0.00001873, -0.00003113, -0.00000849, 0.00001749, 0.00001548, -0.00000901, -0.00002435, -0.00001275, 0.00001050, 0.00001856, 0.00000518, -0.00001112, -0.00001168, 0.00000157, 0.00001237, 0.00001113, 0.00000328, -0.00000188, -0.00000373, -0.00000773, -0.00001393, -0.00001454, -0.00000448, 0.00000971, 0.00001548, 0.00000830, -0.00000326, -0.00000750, -0.00000237, 0.00000378, 0.00000307, -0.00000314, -0.00000707, -0.00000409, 0.00000269, 0.00000670, 0.00000486, -0.00000057, -0.00000531, -0.00000657, -0.00000396, 0.00000123, 0.00000647, 0.00000877, 0.00000660, 0.00000150, -0.00000303, -0.00000468, -0.00000422, -0.00000399, -0.00000500, -0.00000609, -0.00000579, -0.00000434, -0.00000345, -0.00000440, -0.00000668, -0.00000853, },
		{ 0.00000108, 0.00000219, -0.00000153, -0.00001165, -0.00002940, -0.00005350, -0.00007585, -0.00008050, -0.00005192, 0.00000956, 0.00007970, 0.00012388, 0.00012607, 0.00010597, 0.00010125, 0.00012665, 0.00015192, 0.00012771, 0.00003880, -0.00007064, -0.00013152, -0.00011278, -0.00004868, -0.00000215, -0.00000381, -0.00003046, -0.00004068, -0.00002212, 0.00000196, 0.00000713, -0.00000348, -0.00000933, -0.00000271, 0.00000310, -0.00000350, -0.00001310, -0.00000791, 0.00001022, 0.00001733, -0.00000169, -0.00002833, -0.00002879, 0.00000277, 0.00003124, 0.00001991, -0.00002266, -0.00004797, -0.00002405, 0.00002358, 0.00003919, 0.00000313, -0.00004228, -0.00004137, 0.00000792, 0.00004957, 0.00003482, -0.00001963, -0.00005064, -0.00002268, 0.00003073, 0.00004720, 0.00000868, -0.00003891, -0.00004113, 0.00000191, 0.00003862, 0.00002828, -0.00001317, -0.00003596, -0.00001657, 0.00001748, 0.00002424, -0.00000237, -0.00002849, -0.00002322, 0.00000578, 0.00002455, 0.00001415, -0.00000975, -0.00001886, -0.00000542, 0.00001281, 0.00001672, 0.00000667, -0.00000318, -0.00000502, -0.00000456, -0.00000887, -0.00001430, -0.00001113, 0.00000194, 0.00001388, 0.00001360, 0.00000282, -0.00000626, -0.00000533, 0.00000152, 0.00000430, -0.00000051, -0.00000667, -0.00000659, -0.00000008, 0.00000621, 0.00000648, 0.00000118, -0.00000475, -0.00000695, -0.00000443, 0.00000105, 0.00000652, 0.00000898, 0.00000688, 0.00000151, -0.00000350, -0.00000527, -0.00000424, -0.00000335, -0.00000446, -0.00000635, -0.00000647, -0.00000426, -0.00000197, -0.00000222, -0.00000523, -0.00000863, -0.00000968, -0.00000738, -0.00000295, },
		{ -0.00001326, -0.00003145, -0.00005206, -0.00006782, -0.00006779, -0.00004190, 0.00000970, 0.00006972, 0.00011029, 0.00011395, 0.00009213, 0.00007880, 0.00009780, 0.00013340, 0.00013827, 0.00007883, -0.00002501, -0.00010980, -0.00012378, -0.00007371, -0.00001520, 0.00000297, -0.00001801, -0.00003909, -0.00003144, -0.00000500, 0.00001054, 0.00000336, -0.00001029, -0.00001100, -0.00000060, 0.00000456, -0.00000158, -0.00000709, -0.00000136, 0.00000736, 0.00000271, -0.00001441, -0.00002248, -0.00000682, 0.00001695, 0.00001808, -0.00001009, -0.00003681, -0.00002836, 0.00000891, 0.00003200, 0.00001189, -0.00002873, -0.00003969, -0.00000377, 0.00003959, 0.00003903, -0.00000686, -0.00004557, -0.00003173, 0.00001878, 0.00004684, 0.00001972, -0.00003052, -0.00004537, -0.00000875, 0.00003539, 0.00003626, -0.00000458, -0.00003765, -0.00002574, 0.00001346, 0.00003140, 0.00000749, -0.00002836, -0.00003359, -0.00000320, 0.00002705, 0.00002445, -0.00000410, -0.00002403, -0.00001470, 0.00000990, 0.00002195, 0.00001247, -0.00000312, -0.00000806, -0.00000380, -0.00000318, -0.00001035, -0.00001501, -0.00000722, 0.00000816, 0.00001635, 0.00001021, -0.00000192, -0.00000696, -0.00000213, 0.00000367, 0.00000213, -0.00000466, -0.00000793, -0.00000335, 0.00000432, 0.00000730, 0.00000322, -0.00000350, -0.00000694, -0.00000486, 0.00000079, 0.00000639, 0.00000888, 0.00000693, 0.00000169, -0.00000354, -0.00000566, -0.00000455, -0.00000313, -0.00000397, -0.00000628, -0.00000689, -0.00000428, -0.00000076, -0.00000016, -0.00000373, -0.00000867, -0.00001085, -0.00000832, -0.00000260, 0.00000273, 0.00000403, -0.00000106, },
		{ -0.00006090, -0.00005712, -0.00003284, 0.00001064, 0.00006073, 0.00009655, 0.00010227, 0.00008277, 0.00006467, 0.00007455, 0.00010989, 0.00013331, 0.00010341, 0.00001773, -0.00007641, -0.00011963, -0.00009342, -0.00003510, 0.00000053, -0.00000684, -0.00003101, -0.00003544, -0.00001400, 0.00000806, 0.00000830, -0.00000714, -0.00001518, -0.00000594, 0.00000725, 0.00000799, -0.00000195, -0.00000765, -0.00000305, 0.00000146, -0.00000420, -0.00001318, -0.00001047, 0.00000407, 0.00001140, -0.00000248, -0.00002449, -0.00002691, -0.00000270, 0.00002142, 0.00001523, -0.00001583, -0.00003328, -0.00001152, 0.00002781, 0.00003819, 0.00000421, -0.00003674, -0.00003645, 0.00000619, 0.00004181, 0.00002805, -0.00001959, -0.00004532, -0.00001894, 0.00002839, 0.00004132, 0.00000579, -0.00003531, -0.00003411, 0.00000576, 0.00003514, 0.00001883, -0.00002333, -0.00004157, -0.00001524, 0.00002471, 0.00003387, 0.00000541, -0.00002543, -0.00002465, 0.00000321, 0.00002513, 0.00002002, -0.00000051, -0.00001156, -0.00000595, 0.00000132, -0.00000359, -0.00001482, -0.00001565, -0.00000113, 0.00001508, 0.00001715, 0.00000535, -0.00000598, -0.00000613, 0.00000086, 0.00000372, -0.00000145, -0.00000751, -0.00000637, 0.00000110, 0.00000682, 0.00000513, -0.00000153, -0.00000630, -0.00000515, 0.00000044, 0.00000615, 0.00000856, 0.00000669, 0.00000182, -0.00000322, -0.00000566, -0.00000494, -0.00000339, -0.00000377, -0.00000596, -0.00000686, -0.00000421, 0.00000011, 0.00000155, -0.00000225, -0.00000852, -0.00001183, -0.00000913, -0.00000219, 0.00000412, 0.00000524, -0.00000102, -0.00001435, -0.00003224, -0.00005001, },
		{ 0.00001164, 0.00005321, 0.00008383, 0.00009083, 0.00007532, 0.00005618, 0.00005780, 0.00008671, 0.00011833, 0.00011277, 0.00005165, -0.00003814, -0.00010175, -0.00010305, -0.00005639, -0.00001049, -0.00000088, -0.00001986, -0.00003324, -0.00002158, 0.00000121, 0.00000933, -0.00000251, -0.00001499, -0.00001034, 0.00000595, 0.00001388, 0.00000489, -0.00000871, -0.00001129, -0.00000353, 0.00000109, -0.00000371, -0.00000899, -0.00000520, 0.00000224, -0.00000026, -0.00001364, -0.00002117, -0.00000971, 0.00000996, 0.00001343, -0.00000576, -0.00002393, -0.00001450, 0.00001648, 0.00003304, 0.00001188, -0.00002596, -0.00003641, -0.00000494, 0.00003309, 0.00003232, -0.00000796, -0.00004089, -0.00002684, 0.00001869, 0.00004234, 0.00001615, -0.00002897, -0.00003982, -0.00000432, 0.00003434, 0.00002942, -0.00001392, -0.00004518, -0.00002822, 0.00001724, 0.00004011, 0.00001732, -0.00002203, -0.00003316, -0.00000671, 0.00002478, 0.00002790, 0.00000514, -0.00001387, -0.00001056, 0.00000310, 0.00000429, -0.00001024, -0.00002123, -0.00001240, 0.00000910, 0.00002163, 0.00001449, -0.00000154, -0.00000902, -0.00000379, 0.00000332, 0.00000208, -0.00000512, -0.00000826, -0.00000297, 0.00000467, 0.00000634, 0.00000103, -0.00000482, -0.00000515, -0.00000006, 0.00000576, 0.00000813, 0.00000621, 0.00000175, -0.00000273, -0.00000520, -0.00000512, -0.00000399, -0.00000404, -0.00000562, -0.00000634, -0.00000381, 0.00000075, 0.00000275, -0.00000099, -0.00000815, -0.00001251, -0.00000979, -0.00000179, 0.00000534, 0.00000604, -0.00000145, -0.00001534, -0.00003209, -0.00004700, -0.00005434, -0.00004828, -0.00002550, },
		{ 0.00008027, 0.00006858, 0.00005070, 0.00004655, 0.00006691, 0.00009885, 0.00010983, 0.00007373, -0.00000164, -0.00007423, -0.00010062, -0.00007360, -0.00002704, -0.00000214, -0.00000960, -0.00002601, -0.00002522, -0.00000735, 0.00000600, 0.00000106, -0.00001160, -0.00001208, 0.00000279, 0.00001570, 0.00001101, -0.00000584, -0.00001578, -0.00000978, 0.00000161, 0.00000358, -0.00000424, -0.00000984, -0.00000679, -0.00000259, -0.00000619, -0.00001336, -0.00001183, 0.00000000, 0.00000789, 0.00000016, -0.00001388, -0.00001297, 0.00000755, 0.00002511, 0.00001529, -0.00001532, -0.00003214, -0.00001294, 0.00002236, 0.00003202, 0.00000237, -0.00003293, -0.00003105, 0.00000801, 0.00003908, 0.00002459, -0.00001958, -0.00004152, -0.00001485, 0.00002893, 0.00003708, -0.00000174, -0.00004337, -0.00003964, 0.00000556, 0.00004140, 0.00002935, -0.00001387, -0.00003816, -0.00001831, 0.00002011, 0.00003424, 0.00001338, -0.00001351, -0.00001630, 0.00000138, 0.00001123, -0.00000207, -0.00002237, -0.00002340, -0.00000117, 0.00002190, 0.00002368, 0.00000636, -0.00000934, -0.00000929, 0.00000043, 0.00000479, -0.00000109, -0.00000825, -0.00000711, 0.00000084, 0.00000623, 0.00000378, -0.00000234, -0.00000458, -0.00000067, 0.00000512, 0.00000764, 0.00000565, 0.00000143, -0.00000234, -0.00000443, -0.00000491, -0.00000465, -0.00000479, -0.00000556, -0.00000550, -0.00000289, 0.00000140, 0.00000342, -0.00000019, -0.00000771, -0.00001275, -0.00001019, -0.00000145, 0.00000636, 0.00000655, -0.00000222, -0.00001649, -0.00003146, -0.00004309, -0.00004753, -0.00004066, -0.00002008, 0.00001186, 0.00004685, 0.00007292, },
		{ 0.00003891, 0.00005140, 0.00007915, 0.00009891, 0.00008390, 0.00002807, -0.00004262, -0.00008704, -0.00008268, -0.00004468, -0.00001029, -0.00000356, -0.00001642, -0.00002403, -0.00001478, -0.00000059, 0.00000180, -0.00000697, -0.00001074, -0.00000009, 0.00001432, 0.00001474, -0.00000105, -0.00001621, -0.00001500, -0.00000130, 0.00000732, 0.00000145, -0.00001013, -0.00001360, -0.00000767, -0.00000297, -0.00000583, -0.00000991, -0.00000677, 0.00000064, 0.00000172, -0.00000522, -0.00000818, 0.00000221, 0.00001638, 0.00001459, -0.00000664, -0.00002503, -0.00001689, 0.00001163, 0.00002760, 0.00000975, -0.00002299, -0.00003100, -0.00000171, 0.00003220, 0.00002960, -0.00000884, -0.00003876, -0.00002371, 0.00001991, 0.00004024, 0.00001092, -0.00003626, -0.00004716, -0.00000834, 0.00003703, 0.00003900, -0.00000218, -0.00003821, -0.00002942, 0.00001135, 0.00003724, 0.00002286, -0.00000964, -0.00002136, -0.00000358, 0.00001538, 0.00000798, -0.00001845, -0.00003176, -0.00001422, 0.00001700, 0.00003075, 0.00001679, -0.00000602, -0.00001412, -0.00000479, 0.00000557, 0.00000375, -0.00000597, -0.00001032, -0.00000430, 0.00000431, 0.00000609, 0.00000106, -0.00000311, -0.00000123, 0.00000412, 0.00000703, 0.00000517, 0.00000099, -0.00000222, -0.00000362, -0.00000428, -0.00000506, -0.00000582, -0.00000594, -0.00000463, -0.00000148, 0.00000235, 0.00000374, -0.00000000, -0.00000742, -0.00001260, -0.00001015, -0.00000108, 0.00000711, 0.00000683, -0.00000317, -0.00001779, -0.00003073, -0.00003864, -0.00004021, -0.00003352, -0.00001626, 0.00001074, 0.00004090, 0.00006394, 0.00007139, 0.00006245, 0.00004641, },
		{ 0.00008439, 0.00008441, 0.00004839, -0.00001241, -0.00006565, -0.00008185, -0.00005887, -0.00002289, -0.00000352, -0.00000776, -0.00001888, -0.00001896, -0.00000833, -0.00000074, -0.00000319, -0.00000709, -0.00000122, 0.00001134, 0.00001554, 0.00000361, -0.00001342, -0.00001767, -0.00000570, 0.00000733, 0.00000601, -0.00000739, -0.00001700, -0.00001336, -0.00000360, -0.00000044, -0.00000564, -0.00000970, -0.00000622, -0.00000024, 0.00000056, -0.00000193, 0.00000066, 0.00000874, 0.00001082, -0.00000112, -0.00001687, -0.00001680, 0.00000272, 0.00002034, 0.00001323, -0.00001301, -0.00002705, -0.00000880, 0.00002315, 0.00003046, 0.00000126, -0.00003210, -0.00002918, 0.00000912, 0.00003836, 0.00002163, -0.00002516, -0.00004927, -0.00002186, 0.00002751, 0.00004413, 0.00001087, -0.00003293, -0.00003773, -0.00000020, 0.00003573, 0.00003165, -0.00000240, -0.00002398, -0.00001056, 0.00001562, 0.00001766, -0.00001007, -0.00003552, -0.00002773, 0.00000709, 0.00003370, 0.00002798, 0.00000122, -0.00001666, -0.00001143, 0.00000369, 0.00000818, -0.00000158, -0.00001168, -0.00000996, 0.00000037, 0.00000724, 0.00000506, -0.00000048, -0.00000145, 0.00000269, 0.00000614, 0.00000485, 0.00000066, -0.00000238, -0.00000308, -0.00000343, -0.00000499, -0.00000679, -0.00000676, -0.00000407, 0.00000021, 0.00000377, 0.00000402, -0.00000038, -0.00000756, -0.00001224, -0.00000960, -0.00000051, 0.00000763, 0.00000683, -0.00000421, -0.00001913, -0.00003010, -0.00003408, -0.00003253, -0.00002635, -0.00001334, 0.00000823, 0.00003458, 0.00005636, 0.00006461, 0.00005750, 0.00004258, 0.00003299, 0.00003949, 0.00006173, },
		{ 0.00001224, -0.00004102, -0.00007186, -0.00006619, -0.00003623, -0.00000920, -0.00000282, -0.00001198, -0.00001911, -0.00001488, -0.00000570, -0.00000170, -0.00000268, -0.00000015, 0.00000848, 0.00001396, 0.00000669, -0.00000896, -0.00001732, -0.00000968, 0.00000451, 0.00000811, -0.00000343, -0.00001696, -0.00001776, -0.00000681, 0.00000184, -0.00000097, -0.00000920, -0.00001104, -0.00000420, 0.00000294, 0.00000394, 0.00000219, 0.00000350, 0.00000554, 0.00000095, -0.00000945, -0.00001357, -0.00000316, 0.00001199, 0.00001277, -0.00000475, -0.00002036, -0.00001229, 0.00001379, 0.00002738, 0.00000897, -0.00002296, -0.00003035, -0.00000128, 0.00003204, 0.00002847, -0.00001226, -0.00004565, -0.00003247, 0.00001456, 0.00004357, 0.00002271, -0.00002316, -0.00004145, -0.00001242, 0.00002954, 0.00003771, 0.00000707, -0.00002295, -0.00001770, 0.00001189, 0.00002484, 0.00000114, -0.00003375, -0.00003912, -0.00000650, 0.00003123, 0.00003765, 0.00001166, -0.00001559, -0.00001803, -0.00000092, 0.00001096, 0.00000418, -0.00001060, -0.00001509, -0.00000533, 0.00000658, 0.00000903, 0.00000337, -0.00000091, 0.00000094, 0.00000473, 0.00000463, 0.00000068, -0.00000264, -0.00000302, -0.00000271, -0.00000445, -0.00000731, -0.00000776, -0.00000403, 0.00000179, 0.00000561, 0.00000460, -0.00000109, -0.00000827, -0.00001198, -0.00000860, 0.00000043, 0.00000802, 0.00000650, -0.00000533, -0.00002038, -0.00002956, -0.00002977, -0.00002485, -0.00001895, -0.00001062, 0.00000479, 0.00002743, 0.00004933, 0.00005974, 0.00005433, 0.00003938, 0.00002761, 0.00002975, 0.00004724, 0.00006968, 0.00007872, 0.00005923, },
		{ -0.00006519, -0.00004649, -0.00001853, -0.00000306, -0.00000611, -0.00001599, -0.00001849, -0.00001135, -0.00000288, 0.00000081, 0.00000254, 0.00000696, 0.00001122, 0.00000764, -0.00000448, -0.00001452, -0.00001187, 0.00000036, 0.00000742, -0.00000002, -0.00001436, -0.00001973, -0.00001087, 0.00000105, 0.00000242, -0.00000648, -0.00001301, -0.00000837, 0.00000233, 0.00000797, 0.00000543, 0.00000118, 0.00000042, 0.00000013, -0.00000408, -0.00000864, -0.00000562, 0.00000432, 0.00000921, 0.00000056, -0.00001262, -0.00001207, 0.00000588, 0.00002142, 0.00001317, -0.00001322, -0.00002733, -0.00000930, 0.00002284, 0.00003043, -0.00000004, -0.00003720, -0.00003826, 0.00000067, 0.00003740, 0.00003095, -0.00001083, -0.00003974, -0.00002292, 0.00001964, 0.00003954, 0.00001689, -0.00001805, -0.00002302, 0.00000521, 0.00002803, 0.00001287, -0.00002671, -0.00004613, -0.00002149, 0.00002319, 0.00004358, 0.00002366, -0.00001035, -0.00002286, -0.00000756, 0.00001112, 0.00001013, -0.00000700, -0.00001862, -0.00001211, 0.00000371, 0.00001217, 0.00000822, 0.00000078, -0.00000081, 0.00000265, 0.00000424, 0.00000118, -0.00000268, -0.00000345, -0.00000250, -0.00000368, -0.00000713, -0.00000850, -0.00000446, 0.00000287, 0.00000759, 0.00000566, -0.00000174, -0.00000947, -0.00001211, -0.00000741, 0.00000183, 0.00000853, 0.00000590, -0.00000661, -0.00002149, -0.00002900, -0.00002586, -0.00001761, -0.00001149, -0.00000760, 0.00000113, 0.00001951, 0.00004205, 0.00005606, 0.00005313, 0.00003741, 0.00002250, 0.00002084, 0.00003507, 0.00005662, 0.00007037, 0.00006250, 0.00002923, -0.00001773, -0.00005554, },
		{ -0.00000818, -0.00000360, -0.00001154, -0.00001856, -0.00001569, -0.00000597, 0.00000221, 0.00000557, 0.00000714, 0.00000872, 0.00000685, -0.00000120, -0.00001047, -0.00001183, -0.00000354, 0.00000457, 0.00000166, -0.00001065, -0.00001908, -0.00001419, -0.00000189, 0.00000358, -0.00000313, -0.00001226, -0.00001131, -0.00000016, 0.00000951, 0.00000884, 0.00000145, -0.00000325, -0.00000242, -0.00000135, -0.00000358, -0.00000511, -0.00000135, 0.00000394, 0.00000251, -0.00000554, -0.00000889, 0.00000066, 0.00001421, 0.00001364, -0.00000476, -0.00002117, -0.00001366, 0.00001285, 0.00002767, 0.00000930, -0.00002585, -0.00003844, -0.00001153, 0.00002692, 0.00003401, 0.00000157, -0.00003302, -0.00002963, 0.00000792, 0.00003656, 0.00002494, -0.00001010, -0.00002499, -0.00000265, 0.00002680, 0.00002277, -0.00001585, -0.00004746, -0.00003520, 0.00001062, 0.00004421, 0.00003497, -0.00000137, -0.00002442, -0.00001487, 0.00000835, 0.00001493, -0.00000146, -0.00001973, -0.00001890, -0.00000132, 0.00001368, 0.00001351, 0.00000388, -0.00000209, -0.00000010, 0.00000337, 0.00000212, -0.00000215, -0.00000413, -0.00000309, -0.00000313, -0.00000625, -0.00000854, -0.00000506, 0.00000320, 0.00000925, 0.00000715, -0.00000193, -0.00001087, -0.00001281, -0.00000640, 0.00000358, 0.00000936, 0.00000521, -0.00000809, -0.00002252, -0.00002832, -0.00002232, -0.00001111, -0.00000436, -0.00000413, -0.00000202, 0.00001132, 0.00003411, 0.00005259, 0.00005349, 0.00003722, 0.00001812, 0.00001210, 0.00002403, 0.00004554, 0.00006197, 0.00006110, 0.00003864, 0.00000081, -0.00003681, -0.00005661, -0.00005080, -0.00002831, },
		{ -0.00001583, -0.00001728, -0.00000942, 0.00000128, 0.00000755, 0.00000851, 0.00000742, 0.00000532, 0.00000041, -0.00000651, -0.00000996, -0.00000607, 0.00000074, 0.00000127, -0.00000727, -0.00001644, -0.00001576, -0.00000564, 0.00000239, -0.00000055, -0.00000967, -0.00001220, -0.00000311, 0.00000868, 0.00001116, 0.00000333, -0.00000482, -0.00000533, -0.00000105, 0.00000032, -0.00000268, -0.00000435, -0.00000146, 0.00000152, -0.00000047, -0.00000410, -0.00000140, 0.00000744, 0.00001104, 0.00000109, -0.00001351, -0.00001406, 0.00000411, 0.00002139, 0.00001442, -0.00001398, -0.00003349, -0.00001990, 0.00001438, 0.00003152, 0.00001154, -0.00002284, -0.00003133, -0.00000326, 0.00002934, 0.00002946, -0.00000078, -0.00002301, -0.00000962, 0.00002184, 0.00002904, -0.00000339, -0.00004298, -0.00004516, -0.00000436, 0.00003899, 0.00004326, 0.00000995, -0.00002185, -0.00002114, 0.00000306, 0.00001745, 0.00000496, -0.00001810, -0.00002454, -0.00000797, 0.00001297, 0.00001844, 0.00000830, -0.00000239, -0.00000325, 0.00000171, 0.00000318, -0.00000082, -0.00000464, -0.00000449, -0.00000330, -0.00000498, -0.00000757, -0.00000533, 0.00000282, 0.00001013, 0.00000874, -0.00000142, -0.00001198, -0.00001398, -0.00000593, 0.00000538, 0.00001062, 0.00000471, -0.00000968, -0.00002357, -0.00002752, -0.00001903, -0.00000545, 0.00000204, -0.00000041, -0.00000418, 0.00000365, 0.00002561, 0.00004849, 0.00005453, 0.00003889, 0.00001524, 0.00000365, 0.00001316, 0.00003572, 0.00005480, 0.00005791, 0.00004223, 0.00001301, -0.00001954, -0.00004302, -0.00004806, -0.00003522, -0.00001628, -0.00000566, -0.00000817, },
		{ -0.00000122, 0.00000754, 0.00000993, 0.00000758, 0.00000420, 0.00000070, -0.00000365, -0.00000724, -0.00000683, -0.00000277, -0.00000071, -0.00000521, -0.00001295, -0.00001532, -0.00000886, -0.00000046, 0.00000039, -0.00000647, -0.00001102, -0.00000529, 0.00000630, 0.00001170, 0.00000561, -0.00000438, -0.00000745, -0.00000238, 0.00000235, 0.00000036, -0.00000480, -0.00000567, -0.00000138, 0.00000195, 0.00000077, -0.00000069, 0.00000246, 0.00000667, 0.00000377, -0.00000614, -0.00001115, -0.00000186, 0.00001341, 0.00001503, -0.00000387, -0.00002496, -0.00002327, 0.00000233, 0.00002436, 0.00001724, -0.00001148, -0.00002801, -0.00001171, 0.00001944, 0.00002951, 0.00000783, -0.00001753, -0.00001395, 0.00001473, 0.00003085, 0.00000821, -0.00003385, -0.00004970, -0.00001912, 0.00002865, 0.00004672, 0.00002149, -0.00001527, -0.00002476, -0.00000359, 0.00001708, 0.00001093, -0.00001405, -0.00002805, -0.00001529, 0.00000985, 0.00002211, 0.00001365, -0.00000128, -0.00000634, -0.00000087, 0.00000396, 0.00000131, -0.00000449, -0.00000644, -0.00000454, -0.00000385, -0.00000556, -0.00000471, 0.00000207, 0.00000991, 0.00000990, -0.00000027, -0.00001235, -0.00001526, -0.00000619, 0.00000680, 0.00001219, 0.00000468, -0.00001117, -0.00002471, -0.00002670, -0.00001589, -0.00000056, 0.00000742, 0.00000314, -0.00000526, -0.00000280, 0.00001713, 0.00004336, 0.00005516, 0.00004186, 0.00001442, -0.00000375, 0.00000213, 0.00002607, 0.00004880, 0.00005496, 0.00004258, 0.00001919, -0.00000653, -0.00002789, -0.00003911, -0.00003686, -0.00002446, -0.00001187, -0.00000791, -0.00001210, -0.00001570, -0.00001150, },
		{ 0.00000864, 0.00000425, 0.00000055, -0.00000220, -0.00000480, -0.00000619, -0.00000507, -0.00000330, -0.00000474, -0.00000978, -0.00001339, -0.00001072, -0.00000391, -0.00000046, -0.00000385, -0.00000843, -0.00000596, 0.00000348, 0.00001046, 0.00000722, -0.00000261, -0.00000812, -0.00000426, 0.00000249, 0.00000285, -0.00000344, -0.00000794, -0.00000485, 0.00000183, 0.00000440, 0.00000180, -0.00000006, 0.00000205, 0.00000356, -0.00000051, -0.00000624, -0.00000443, 0.00000573, 0.00001194, 0.00000290, -0.00001500, -0.00002168, -0.00000698, 0.00001442, 0.00001791, -0.00000140, -0.00002078, -0.00001593, 0.00000905, 0.00002525, 0.00001384, -0.00000995, -0.00001469, 0.00000750, 0.00002853, 0.00001692, -0.00002210, -0.00004834, -0.00003105, 0.00001501, 0.00004449, 0.00003095, -0.00000578, -0.00002466, -0.00001003, 0.00001398, 0.00001523, -0.00000843, -0.00002886, -0.00002213, 0.00000462, 0.00002375, 0.00001919, 0.00000138, -0.00000876, -0.00000426, 0.00000397, 0.00000396, -0.00000332, -0.00000842, -0.00000695, -0.00000348, -0.00000285, -0.00000281, 0.00000155, 0.00000861, 0.00001006, 0.00000115, -0.00001167, -0.00001612, -0.00000710, 0.00000739, 0.00001372, 0.00000523, -0.00001223, -0.00002586, -0.00002601, -0.00001294, 0.00000371, 0.00001175, 0.00000611, -0.00000553, -0.00000757, 0.00000945, 0.00003734, 0.00005449, 0.00004511, 0.00001567, -0.00000916, -0.00000858, 0.00001580, 0.00004304, 0.00005299, 0.00004213, 0.00002114, 0.00000108, -0.00001459, -0.00002636, -0.00003242, -0.00002980, -0.00002040, -0.00001172, -0.00000956, -0.00001173, -0.00001097, -0.00000383, 0.00000544, 0.00001020, },
		{ -0.00000180, -0.00000343, -0.00000502, -0.00000587, -0.00000545, -0.00000545, -0.00000774, -0.00001089, -0.00001100, -0.00000688, -0.00000257, -0.00000254, -0.00000542, -0.00000507, 0.00000124, 0.00000805, 0.00000754, -0.00000045, -0.00000729, -0.00000571, 0.00000127, 0.00000407, -0.00000129, -0.00000816, -0.00000770, -0.00000008, 0.00000605, 0.00000476, -0.00000020, -0.00000153, 0.00000145, 0.00000263, -0.00000095, -0.00000386, -0.00000007, 0.00000666, 0.00000571, -0.00000583, -0.00001629, -0.00001213, 0.00000403, 0.00001400, 0.00000543, -0.00001164, -0.00001552, 0.00000035, 0.00001791, 0.00001606, -0.00000216, -0.00001194, 0.00000199, 0.00002338, 0.00002159, -0.00001013, -0.00004185, -0.00003824, 0.00000057, 0.00003693, 0.00003637, 0.00000475, -0.00002065, -0.00001464, 0.00000900, 0.00001704, -0.00000247, -0.00002699, -0.00002738, -0.00000195, 0.00002296, 0.00002401, 0.00000538, -0.00000994, -0.00000804, 0.00000292, 0.00000664, -0.00000101, -0.00000982, -0.00001027, -0.00000437, -0.00000005, 0.00000046, 0.00000193, 0.00000660, 0.00000883, 0.00000223, -0.00000996, -0.00001598, -0.00000826, 0.00000693, 0.00001471, 0.00000627, -0.00001253, -0.00002677, -0.00002557, -0.00001031, 0.00000748, 0.00001516, 0.00000826, -0.00000545, -0.00001065, 0.00000332, 0.00003106, 0.00005212, 0.00004748, 0.00001838, -0.00001194, -0.00001795, 0.00000488, 0.00003642, 0.00005156, 0.00004243, 0.00002126, 0.00000392, -0.00000544, -0.00001305, -0.00002287, -0.00003021, -0.00002855, -0.00001909, -0.00001005, -0.00000713, -0.00000757, -0.00000499, 0.00000206, 0.00000856, 0.00000954, 0.00000557, 0.00000095, },
		{ -0.00000558, -0.00000643, -0.00000648, -0.00000700, -0.00000880, -0.00001014, -0.00000867, -0.00000504, -0.00000263, -0.00000295, -0.00000314, 0.00000018, 0.00000535, 0.00000654, 0.00000123, -0.00000547, -0.00000616, -0.00000048, 0.00000389, 0.00000064, -0.00000684, -0.00000918, -0.00000273, 0.00000564, 0.00000702, 0.00000131, -0.00000330, -0.00000131, 0.00000329, 0.00000333, -0.00000118, -0.00000313, 0.00000100, 0.00000499, 0.00000088, -0.00000894, -0.00001282, -0.00000459, 0.00000700, 0.00000798, -0.00000287, -0.00001122, -0.00000501, 0.00000939, 0.00001432, 0.00000393, -0.00000675, -0.00000063, 0.00001723, 0.00002217, -0.00000011, -0.00003200, -0.00003989, -0.00001208, 0.00002554, 0.00003665, 0.00001413, -0.00001352, -0.00001623, 0.00000351, 0.00001623, 0.00000261, -0.00002306, -0.00003028, -0.00000879, 0.00001977, 0.00002725, 0.00001014, -0.00000955, -0.00001160, 0.00000075, 0.00000876, 0.00000225, -0.00001010, -0.00001396, -0.00000675, 0.00000207, 0.00000479, 0.00000373, 0.00000454, 0.00000616, 0.00000231, -0.00000763, -0.00001447, -0.00000905, 0.00000547, 0.00001465, 0.00000739, -0.00001190, -0.00002707, -0.00002529, -0.00000820, 0.00001073, 0.00001785, 0.00000957, -0.00000546, -0.00001237, -0.00000085, 0.00002539, 0.00004825, 0.00004801, 0.00002139, -0.00001208, -0.00002484, -0.00000596, 0.00002826, 0.00004953, 0.00004380, 0.00002170, 0.00000385, -0.00000116, -0.00000217, -0.00001060, -0.00002513, -0.00003362, -0.00002819, -0.00001434, -0.00000398, -0.00000216, -0.00000357, -0.00000117, 0.00000502, 0.00000918, 0.00000757, 0.00000251, -0.00000159, -0.00000331, -0.00000429, },
		{ -0.00000714, -0.00000774, -0.00000896, -0.00000917, -0.00000700, -0.00000368, -0.00000161, -0.00000099, 0.00000035, 0.00000319, 0.00000471, 0.00000191, -0.00000338, -0.00000552, -0.00000196, 0.00000266, 0.00000170, -0.00000477, -0.00000907, -0.00000508, 0.00000370, 0.00000783, 0.00000343, -0.00000317, -0.00000352, 0.00000217, 0.00000577, 0.00000227, -0.00000347, -0.00000352, 0.00000193, 0.00000430, -0.00000166, -0.00000984, -0.00000993, -0.00000103, 0.00000634, 0.00000358, -0.00000465, -0.00000636, 0.00000174, 0.00000947, 0.00000702, -0.00000079, -0.00000012, 0.00001192, 0.00001964, 0.00000659, -0.00002104, -0.00003639, -0.00002092, 0.00001257, 0.00003184, 0.00002040, -0.00000483, -0.00001439, -0.00000098, 0.00001338, 0.00000589, -0.00001808, -0.00003062, -0.00001480, 0.00001473, 0.00002832, 0.00001484, -0.00000755, -0.00001431, -0.00000227, 0.00000984, 0.00000597, -0.00000896, -0.00001728, -0.00001044, 0.00000286, 0.00000951, 0.00000716, 0.00000324, 0.00000239, 0.00000086, -0.00000534, -0.00001156, -0.00000880, 0.00000344, 0.00001322, 0.00000803, -0.00001046, -0.00002638, -0.00002493, -0.00000673, 0.00001331, 0.00001998, 0.00001021, -0.00000584, -0.00001326, -0.00000307, 0.00002109, 0.00004359, 0.00004631, 0.00002343, -0.00001030, -0.00002853, -0.00001545, 0.00001867, 0.00004568, 0.00004544, 0.00002366, 0.00000324, -0.00000087, 0.00000432, 0.00000133, -0.00001566, -0.00003377, -0.00003642, -0.00002184, -0.00000398, 0.00000352, 0.00000056, -0.00000273, 0.00000046, 0.00000688, 0.00000918, 0.00000522, -0.00000068, -0.00000400, -0.00000463, -0.00000501, -0.00000611, -0.00000693, },
		{ -0.00000881, -0.00000789, -0.00000491, -0.00000144, 0.00000058, 0.00000128, 0.00000205, 0.00000276, 0.00000159, -0.00000173, -0.00000415, -0.00000268, 0.00000103, 0.00000172, -0.00000278, -0.00000771, -0.00000640, 0.00000107, 0.00000703, 0.00000513, -0.00000165, -0.00000445, 0.00000029, 0.00000621, 0.00000521, -0.00000182, -0.00000594, -0.00000190, 0.00000459, 0.00000390, -0.00000471, -0.00001144, -0.00000804, 0.00000165, 0.00000652, 0.00000215, -0.00000411, -0.00000337, 0.00000312, 0.00000668, 0.00000418, 0.00000276, 0.00000873, 0.00001558, 0.00000963, -0.00001108, -0.00002919, -0.00002488, 0.00000045, 0.00002308, 0.00002233, 0.00000344, -0.00000959, -0.00000328, 0.00000959, 0.00000703, -0.00001317, -0.00002870, -0.00001911, 0.00000872, 0.00002703, 0.00001862, -0.00000434, -0.00001566, -0.00000564, 0.00000962, 0.00000952, -0.00000645, -0.00001953, -0.00001492, 0.00000191, 0.00001377, 0.00001200, 0.00000339, -0.00000175, -0.00000229, -0.00000390, -0.00000761, -0.00000701, 0.00000153, 0.00001045, 0.00000759, -0.00000862, -0.00002448, -0.00002404, -0.00000587, 0.00001498, 0.00002155, 0.00001044, -0.00000661, -0.00001386, -0.00000376, 0.00001862, 0.00003909, 0.00004261, 0.00002348, -0.00000785, -0.00002899, -0.00002225, 0.00000864, 0.00003933, 0.00004592, 0.00002710, 0.00000407, -0.00000254, 0.00000602, 0.00001012, -0.00000424, -0.00002861, -0.00004120, -0.00003070, -0.00000781, 0.00000741, 0.00000650, -0.00000159, -0.00000368, 0.00000276, 0.00000932, 0.00000841, 0.00000146, -0.00000462, -0.00000610, -0.00000506, -0.00000495, -0.00000624, -0.00000735, -0.00000776, -0.00000827, },
		{ -0.00000199, 0.00000113, 0.00000226, 0.00000191, 0.00000137, 0.00000067, -0.00000091, -0.00000267, -0.00000258, -0.00000037, 0.00000095, -0.00000145, -0.00000574, -0.00000649, -0.00000137, 0.00000498, 0.00000572, 0.00000041, -0.00000391, -0.00000132, 0.00000513, 0.00000683, 0.00000072, -0.00000613, -0.00000509, 0.00000265, 0.00000684, 0.00000080, -0.00000961, -0.00001251, -0.00000435, 0.00000581, 0.00000734, 0.00000046, -0.00000512, -0.00000287, 0.00000349, 0.00000683, 0.00000665, 0.00000810, 0.00001169, 0.00000965, -0.00000362, -0.00002027, -0.00002406, -0.00000876, 0.00001232, 0.00001970, 0.00000943, -0.00000312, -0.00000283, 0.00000616, 0.00000633, -0.00000925, -0.00002531, -0.00002130, 0.00000274, 0.00002367, 0.00002077, -0.00000055, -0.00001542, -0.00000871, 0.00000817, 0.00001226, -0.00000296, -0.00002022, -0.00001941, -0.00000076, 0.00001673, 0.00001761, 0.00000536, -0.00000536, -0.00000686, -0.00000397, -0.00000335, -0.00000354, 0.00000053, 0.00000675, 0.00000565, -0.00000701, -0.00002145, -0.00002220, -0.00000531, 0.00001559, 0.00002241, 0.00001047, -0.00000757, -0.00001450, -0.00000353, 0.00001795, 0.00003563, 0.00003771, 0.00002109, -0.00000613, -0.00002698, -0.00002550, -0.00000033, 0.00003073, 0.00004388, 0.00003090, 0.00000730, -0.00000375, 0.00000409, 0.00001409, 0.00000615, -0.00001939, -0.00004086, -0.00003834, -0.00001482, 0.00000788, 0.00001247, 0.00000235, -0.00000587, -0.00000225, 0.00000737, 0.00001097, 0.00000475, -0.00000430, -0.00000816, -0.00000625, -0.00000380, -0.00000439, -0.00000683, -0.00000839, -0.00000855, -0.00000834, -0.00000771, -0.00000557, },
		{ 0.00000232, 0.00000089, -0.00000022, -0.00000089, -0.00000162, -0.00000196, -0.00000115, -0.00000009, -0.00000093, -0.00000384, -0.00000558, -0.00000298, 0.00000239, 0.00000503, 0.00000214, -0.00000229, -0.00000199, 0.00000337, 0.00000690, 0.00000303, -0.00000459, -0.00000674, -0.00000021, 0.00000711, 0.00000513, -0.00000568, -0.00001379, -0.00000982, 0.00000227, 0.00000969, 0.00000559, -0.00000366, -0.00000693, -0.00000124, 0.00000661, 0.00000995, 0.00000954, 0.00000924, 0.00000795, 0.00000077, -0.00001165, -0.00001956, -0.00001388, 0.00000179, 0.00001338, 0.00001188, 0.00000333, 0.00000010, 0.00000419, 0.00000461, -0.00000681, -0.00002138, -0.00002143, -0.00000231, 0.00001888, 0.00002093, 0.00000302, -0.00001375, -0.00001088, 0.00000586, 0.00001377, 0.00000088, -0.00001923, -0.00002310, -0.00000472, 0.00001782, 0.00002309, 0.00000907, -0.00000762, -0.00001217, -0.00000588, 0.00000035, 0.00000132, 0.00000106, 0.00000283, 0.00000212, -0.00000633, -0.00001770, -0.00001910, -0.00000460, 0.00001519, 0.00002230, 0.00001031, -0.00000844, -0.00001524, -0.00000300, 0.00001864, 0.00003372, 0.00003268, 0.00001652, -0.00000629, -0.00002387, -0.00002511, -0.00000670, 0.00002103, 0.00003857, 0.00003334, 0.00001261, -0.00000257, 0.00000076, 0.00001310, 0.00001283, -0.00000858, -0.00003513, -0.00004225, -0.00002310, 0.00000428, 0.00001640, 0.00000811, -0.00000511, -0.00000667, 0.00000347, 0.00001179, 0.00000843, -0.00000253, -0.00000987, -0.00000855, -0.00000351, -0.00000195, -0.00000511, -0.00000883, -0.00000975, -0.00000848, -0.00000693, -0.00000526, -0.00000256, 0.00000071, 0.00000264, },
		{ -0.00000127, -0.00000130, -0.00000133, -0.00000127, -0.00000090, -0.00000102, -0.00000249, -0.00000419, -0.00000355, 0.00000001, 0.00000335, 0.00000295, -0.00000030, -0.00000153, 0.00000179, 0.00000576, 0.00000438, -0.00000223, -0.00000660, -0.00000274, 0.00000541, 0.00000740, -0.00000118, -0.00001217, -0.00001334, -0.00000263, 0.00000893, 0.00000958, -0.00000002, -0.00000816, -0.00000587, 0.00000389, 0.00001142, 0.00001194, 0.00000876, 0.00000602, 0.00000249, -0.00000480, -0.00001310, -0.00001477, -0.00000657, 0.00000500, 0.00001045, 0.00000809, 0.00000452, 0.00000427, 0.00000292, -0.00000574, -0.00001776, -0.00002002, -0.00000589, 0.00001352, 0.00001914, 0.00000565, -0.00001109, -0.00001179, 0.00000325, 0.00001390, 0.00000438, -0.00001677, -0.00002532, -0.00000923, 0.00001685, 0.00002750, 0.00001400, -0.00000794, -0.00001727, -0.00000952, 0.00000265, 0.00000687, 0.00000345, -0.00000041, -0.00000265, -0.00000716, -0.00001391, -0.00001474, -0.00000320, 0.00001404, 0.00002101, 0.00000979, -0.00000899, -0.00001588, -0.00000258, 0.00002001, 0.00003341, 0.00002851, 0.00001061, -0.00000889, -0.00002123, -0.00002187, -0.00000939, 0.00001196, 0.00003028, 0.00003282, 0.00001857, 0.00000182, -0.00000143, 0.00000854, 0.00001426, 0.00000078, -0.00002536, -0.00004081, -0.00003007, -0.00000267, 0.00001663, 0.00001387, -0.00000130, -0.00000892, -0.00000139, 0.00001024, 0.00001132, 0.00000055, -0.00001033, -0.00001132, -0.00000458, 0.00000018, -0.00000231, -0.00000830, -0.00001131, -0.00000960, -0.00000628, -0.00000406, -0.00000248, -0.00000023, 0.00000207, 0.00000264, 0.00000123, -0.00000052, },
		{ -0.00000103, -0.00000117, -0.00000129, -0.00000182, -0.00000290, -0.00000330, -0.00000163, 0.00000128, 0.00000266, 0.00000134, -0.00000026, 0.00000097, 0.00000411, 0.00000452, -0.00000000, -0.00000505, -0.00000409, 0.00000284, 0.00000747, 0.00000251, -0.00000865, -0.00001423, -0.00000728, 0.00000567, 0.00001127, 0.00000429, -0.00000661, -0.00000895, -0.00000025, 0.00001049, 0.00001391, 0.00000996, 0.00000506, 0.00000257, -0.00000038, -0.00000648, -0.00001227, -0.00001156, -0.00000348, 0.00000577, 0.00000997, 0.00000901, 0.00000627, 0.00000222, -0.00000547, -0.00001493, -0.00001777, -0.00000787, 0.00000841, 0.00001583, 0.00000686, -0.00000810, -0.00001137, 0.00000093, 0.00001286, 0.00000700, -0.00001339, -0.00002576, -0.00001342, 0.00001410, 0.00003009, 0.00001929, -0.00000623, -0.00002123, -0.00001429, 0.00000304, 0.00001219, 0.00000753, -0.00000220, -0.00000795, -0.00000970, -0.00001087, -0.00000947, -0.00000068, 0.00001267, 0.00001854, 0.00000860, -0.00000919, -0.00001612, -0.00000238, 0.00002136, 0.00003424, 0.00002585, 0.00000453, -0.00001370, -0.00002037, -0.00001723, -0.00000822, 0.00000531, 0.00002027, 0.00002847, 0.00002317, 0.00000891, -0.00000051, 0.00000287, 0.00001048, 0.00000605, -0.00001412, -0.00003383, -0.00003320, -0.00001107, 0.00001254, 0.00001753, 0.00000454, -0.00000797, -0.00000566, 0.00000649, 0.00001227, 0.00000412, -0.00000902, -0.00001354, -0.00000689, 0.00000112, 0.00000092, -0.00000643, -0.00001239, -0.00001155, -0.00000643, -0.00000249, -0.00000145, -0.00000097, 0.00000069, 0.00000229, 0.00000194, 0.00000001, -0.00000149, -0.00000161, -0.00000114, },
		{ -0.00000163, -0.00000209, -0.00000266, -0.00000235, -0.00000054, 0.00000156, 0.00000211, 0.00000122, 0.00000107, 0.00000268, 0.00000374, 0.00000145, -0.00000287, -0.00000404, 0.00000047, 0.00000589, 0.00000454, -0.00000454, -0.00001266, -0.00001038, 0.00000114, 0.00001033, 0.00000775, -0.00000308, -0.00000963, -0.00000435, 0.00000741, 0.00001434, 0.00001189, 0.00000558, 0.00000229, 0.00000181, -0.00000109, -0.00000784, -0.00001293, -0.00001029, -0.00000079, 0.00000860, 0.00001211, 0.00000941, 0.00000301, -0.00000520, -0.00001293, -0.00001537, -0.00000851, 0.00000412, 0.00001166, 0.00000653, -0.00000544, -0.00000988, -0.00000063, 0.00001105, 0.00000841, -0.00000976, -0.00002446, -0.00001653, 0.00001022, 0.00003051, 0.00002395, -0.00000284, -0.00002331, -0.00001929, 0.00000147, 0.00001635, 0.00001271, -0.00000205, -0.00001288, -0.00001376, -0.00000927, -0.00000394, 0.00000307, 0.00001170, 0.00001512, 0.00000647, -0.00000922, -0.00001570, -0.00000220, 0.00002223, 0.00003551, 0.00002480, -0.00000059, -0.00001989, -0.00002198, -0.00001292, -0.00000392, 0.00000237, 0.00001048, 0.00002050, 0.00002454, 0.00001701, 0.00000434, -0.00000120, 0.00000307, 0.00000573, -0.00000451, -0.00002271, -0.00003085, -0.00001825, 0.00000481, 0.00001736, 0.00001054, -0.00000370, -0.00000771, 0.00000154, 0.00001059, 0.00000693, -0.00000609, -0.00001424, -0.00000972, 0.00000039, 0.00000360, -0.00000338, -0.00001220, -0.00001368, -0.00000769, -0.00000133, 0.00000036, -0.00000093, -0.00000099, 0.00000102, 0.00000238, 0.00000122, -0.00000108, -0.00000214, -0.00000160, -0.00000088, -0.00000092, -0.00000132, },
		{ -0.00000235, -0.00000165, 0.00000018, 0.00000191, 0.00000228, 0.00000179, 0.00000195, 0.00000263, 0.00000193, -0.00000083, -0.00000287, -0.00000094, 0.00000360, 0.00000478, -0.00000107, -0.00000947, -0.00001130, -0.00000325, 0.00000722, 0.00000930, 0.00000109, -0.00000786, -0.00000710, 0.00000312, 0.00001268, 0.00001339, 0.00000737, 0.00000268, 0.00000261, 0.00000243, -0.00000304, -0.00001133, -0.00001431, -0.00000753, 0.00000447, 0.00001281, 0.00001248, 0.00000519, -0.00000419, -0.00001137, -0.00001325, -0.00000831, 0.00000091, 0.00000728, 0.00000486, -0.00000362, -0.00000779, -0.00000122, 0.00000902, 0.00000861, -0.00000652, -0.00002183, -0.00001807, 0.00000603, 0.00002885, 0.00002714, 0.00000147, -0.00002323, -0.00002355, -0.00000165, 0.00001869, 0.00001809, 0.00000005, -0.00001654, -0.00001872, -0.00000950, 0.00000100, 0.00000785, 0.00001166, 0.00001128, 0.00000329, -0.00000945, -0.00001456, -0.00000170, 0.00002246, 0.00003647, 0.00002498, -0.00000399, -0.00002620, -0.00002597, -0.00001041, 0.00000201, 0.00000353, 0.00000293, 0.00001025, 0.00002158, 0.00002385, 0.00001250, -0.00000155, -0.00000531, -0.00000007, 0.00000075, -0.00001010, -0.00002281, -0.00002162, -0.00000461, 0.00001264, 0.00001452, 0.00000288, -0.00000640, -0.00000297, 0.00000642, 0.00000773, -0.00000242, -0.00001284, -0.00001197, -0.00000186, 0.00000480, 0.00000009, -0.00001040, -0.00001509, -0.00000983, -0.00000124, 0.00000235, 0.00000016, -0.00000224, -0.00000096, 0.00000201, 0.00000258, 0.00000014, -0.00000235, -0.00000253, -0.00000121, -0.00000047, -0.00000088, -0.00000153, -0.00000185, -0.00000214, },
		{ 0.00000100, 0.00000249, 0.00000259, 0.00000201, 0.00000182, 0.00000169, 0.00000050, -0.00000125, -0.00000119, 0.00000152, 0.00000367, 0.00000102, -0.00000580, -0.00001015, -0.00000634, 0.00000302, 0.00000860, 0.00000456, -0.00000436, -0.00000773, -0.00000109, 0.00000917, 0.00001343, 0.00000962, 0.00000422, 0.00000312, 0.00000418, 0.00000090, -0.00000799, -0.00001527, -0.00001290, -0.00000123, 0.00001078, 0.00001430, 0.00000811, -0.00000208, -0.00000969, -0.00001147, -0.00000773, -0.00000131, 0.00000325, 0.00000229, -0.00000287, -0.00000564, -0.00000091, 0.00000725, 0.00000787, -0.00000408, -0.00001846, -0.00001792, 0.00000236, 0.00002559, 0.00002834, 0.00000575, -0.00002115, -0.00002623, -0.00000551, 0.00001895, 0.00002271, 0.00000365, -0.00001832, -0.00002370, -0.00001153, 0.00000458, 0.00001306, 0.00001285, 0.00000765, -0.00000080, -0.00001025, -0.00001287, -0.00000058, 0.00002225, 0.00003659, 0.00002564, -0.00000546, -0.00003141, -0.00003151, -0.00001059, 0.00000774, 0.00000815, -0.00000083, -0.00000023, 0.00001440, 0.00002725, 0.00002212, 0.00000272, -0.00001175, -0.00000966, 0.00000009, 0.00000079, -0.00001048, -0.00001944, -0.00001302, 0.00000401, 0.00001463, 0.00000987, -0.00000155, -0.00000534, 0.00000077, 0.00000575, 0.00000068, -0.00000948, -0.00001255, -0.00000485, 0.00000398, 0.00000290, -0.00000721, -0.00001495, -0.00001210, -0.00000249, 0.00000371, 0.00000205, -0.00000247, -0.00000307, 0.00000065, 0.00000347, 0.00000193, -0.00000181, -0.00000351, -0.00000214, -0.00000023, -0.00000006, -0.00000117, -0.00000197, -0.00000206, -0.00000205, -0.00000193, -0.00000090, },
		{ 0.00000256, 0.00000168, 0.00000128, 0.00000100, 0.00000013, -0.00000057, 0.00000028, 0.00000204, 0.00000161, -0.00000270, -0.00000766, -0.00000756, -0.00000105, 0.00000603, 0.00000637, -0.00000029, -0.00000617, -0.00000404, 0.00000465, 0.00001158, 0.00001128, 0.00000668, 0.00000420, 0.00000491, 0.00000339, -0.00000427, -0.00001373, -0.00001590, -0.00000703, 0.00000646, 0.00001400, 0.00001077, 0.00000102, -0.00000734, -0.00000978, -0.00000710, -0.00000280, -0.00000016, -0.00000068, -0.00000317, -0.00000389, 0.00000002, 0.00000602, 0.00000660, -0.00000263, -0.00001498, -0.00001631, -0.00000020, 0.00002151, 0.00002747, 0.00000912, -0.00001769, -0.00002689, -0.00000918, 0.00001739, 0.00002581, 0.00000797, -0.00001801, -0.00002778, -0.00001487, 0.00000628, 0.00001794, 0.00001520, 0.00000488, -0.00000538, -0.00001188, -0.00001103, 0.00000132, 0.00002198, 0.00003566, 0.00002599, -0.00000534, -0.00003464, -0.00003733, -0.00001350, 0.00001158, 0.00001474, -0.00000019, -0.00000869, 0.00000427, 0.00002587, 0.00003068, 0.00001100, -0.00001392, -0.00002024, -0.00000640, 0.00000713, 0.00000346, -0.00001140, -0.00001770, -0.00000660, 0.00001003, 0.00001499, 0.00000589, -0.00000429, -0.00000468, 0.00000104, 0.00000184, -0.00000505, -0.00001081, -0.00000738, 0.00000130, 0.00000403, -0.00000346, -0.00001290, -0.00001353, -0.00000478, 0.00000374, 0.00000406, -0.00000148, -0.00000454, -0.00000143, 0.00000335, 0.00000372, -0.00000041, -0.00000398, -0.00000345, -0.00000057, 0.00000082, -0.00000037, -0.00000207, -0.00000244, -0.00000191, -0.00000160, -0.00000130, -0.00000009, 0.00000179, 0.00000290, },
		{ 0.00000094, 0.00000084, 0.00000033, 0.00000004, 0.00000069, 0.00000112, -0.00000082, -0.00000483, -0.00000703, -0.00000395, 0.00000255, 0.00000618, 0.00000313, -0.00000307, -0.00000500, 0.00000031, 0.00000806, 0.00001146, 0.00000931, 0.00000622, 0.00000552, 0.00000456, -0.00000122, -0.00001075, -0.00001633, -0.00001158, 0.00000093, 0.00001137, 0.00001214, 0.00000447, -0.00000413, -0.00000775, -0.00000643, -0.00000386, -0.00000290, -0.00000363, -0.00000425, -0.00000284, 0.00000116, 0.00000542, 0.00000523, -0.00000207, -0.00001189, -0.00001370, -0.00000136, 0.00001739, 0.00002487, 0.00001096, -0.00001371, -0.00002555, -0.00001179, 0.00001464, 0.00002699, 0.00001208, -0.00001590, -0.00003025, -0.00001875, 0.00000598, 0.00002168, 0.00001830, 0.00000339, -0.00000982, -0.00001434, -0.00000949, 0.00000388, 0.00002208, 0.00003387, 0.00002540, -0.00000437, -0.00003555, -0.00004212, -0.00001838, 0.00001250, 0.00002138, 0.00000437, -0.00001332, -0.00000666, 0.00001961, 0.00003576, 0.00002135, -0.00001077, -0.00002868, -0.00001701, 0.00000720, 0.00001572, 0.00000119, -0.00001677, -0.00001647, 0.00000122, 0.00001626, 0.00001403, 0.00000049, -0.00000811, -0.00000542, 0.00000016, -0.00000094, -0.00000681, -0.00000824, -0.00000240, 0.00000288, -0.00000032, -0.00000918, -0.00001323, -0.00000731, 0.00000216, 0.00000530, 0.00000045, -0.00000477, -0.00000356, 0.00000208, 0.00000484, 0.00000155, -0.00000353, -0.00000467, -0.00000154, 0.00000132, 0.00000073, -0.00000183, -0.00000305, -0.00000218, -0.00000107, -0.00000095, -0.00000083, 0.00000050, 0.00000237, 0.00000307, 0.00000227, 0.00000125, },
		{ 0.00000048, 0.00000013, 0.00000028, -0.00000023, -0.00000259, -0.00000540, -0.00000521, -0.00000075, 0.00000438, 0.00000503, 0.00000051, -0.00000393, -0.00000275, 0.00000370, 0.00000978, 0.00001114, 0.00000893, 0.00000677, 0.00000508, 0.00000068, -0.00000753, -0.00001472, -0.00001406, -0.00000452, 0.00000689, 0.00001156, 0.00000739, -0.00000027, -0.00000504, -0.00000549, -0.00000463, -0.00000510, -0.00000628, -0.00000580, -0.00000254, 0.00000215, 0.00000533, 0.00000410, -0.00000212, -0.00000948, -0.00001068, -0.00000119, 0.00001389, 0.00002115, 0.00001102, -0.00001006, -0.00002263, -0.00001279, 0.00001156, 0.00002632, 0.00001517, -0.00001264, -0.00003079, -0.00002229, 0.00000399, 0.00002369, 0.00002145, 0.00000326, -0.00001350, -0.00001734, -0.00000863, 0.00000675, 0.00002279, 0.00003165, 0.00002360, -0.00000342, -0.00003443, -0.00004485, -0.00002398, 0.00001027, 0.00002627, 0.00001143, -0.00001327, -0.00001596, 0.00000963, 0.00003571, 0.00003110, -0.00000280, -0.00003244, -0.00002885, 0.00000098, 0.00002329, 0.00001567, -0.00000974, -0.00002276, -0.00000996, 0.00001263, 0.00002049, 0.00000817, -0.00000810, -0.00001193, -0.00000444, 0.00000136, -0.00000143, -0.00000663, -0.00000586, -0.00000045, 0.00000104, -0.00000469, -0.00001076, -0.00000900, -0.00000069, 0.00000502, 0.00000256, -0.00000354, -0.00000495, -0.00000002, 0.00000476, 0.00000345, -0.00000209, -0.00000521, -0.00000290, 0.00000112, 0.00000178, -0.00000110, -0.00000351, -0.00000286, -0.00000077, -0.00000009, -0.00000087, -0.00000086, 0.00000099, 0.00000301, 0.00000322, 0.00000190, 0.00000085, 0.00000081, 0.00000091, },
		{ -0.00000050, -0.00000147, -0.00000360, -0.00000496, -0.00000297, 0.00000182, 0.00000513, 0.00000346, -0.00000145, -0.00000389, -0.00000039, 0.00000653, 0.00001138, 0.00001157, 0.00000890, 0.00000576, 0.00000162, -0.00000495, -0.00001204, -0.00001436, -0.00000870, 0.00000157, 0.00000896, 0.00000893, 0.00000361, -0.00000160, -0.00000399, -0.00000506, -0.00000686, -0.00000856, -0.00000752, -0.00000291, 0.00000272, 0.00000549, 0.00000335, -0.00000244, -0.00000780, -0.00000776, -0.00000003, 0.00001142, 0.00001706, 0.00000949, -0.00000744, -0.00001886, -0.00001205, 0.00000899, 0.00002428, 0.00001675, -0.00000907, -0.00002949, -0.00002473, 0.00000097, 0.00002373, 0.00002393, 0.00000423, -0.00001591, -0.00002039, -0.00000860, 0.00000946, 0.00002415, 0.00002949, 0.00002071, -0.00000323, -0.00003200, -0.00004507, -0.00002889, 0.00000552, 0.00002812, 0.00001904, -0.00000880, -0.00002161, -0.00000190, 0.00003017, 0.00003757, 0.00000806, -0.00003026, -0.00003866, -0.00000989, 0.00002432, 0.00002871, 0.00000224, -0.00002342, -0.00002078, 0.00000444, 0.00002316, 0.00001690, -0.00000415, -0.00001659, -0.00001107, 0.00000079, 0.00000391, -0.00000253, -0.00000771, -0.00000511, -0.00000005, -0.00000068, -0.00000635, -0.00000886, -0.00000395, 0.00000292, 0.00000389, -0.00000119, -0.00000493, -0.00000224, 0.00000332, 0.00000457, 0.00000001, -0.00000465, -0.00000409, 0.00000015, 0.00000231, -0.00000006, -0.00000349, -0.00000368, -0.00000095, 0.00000084, -0.00000020, -0.00000178, -0.00000088, 0.00000208, 0.00000401, 0.00000320, 0.00000122, 0.00000037, 0.00000081, 0.00000105, 0.00000044, -0.00000020, },
		{ -0.00000384, -0.00000371, -0.00000050, 0.00000381, 0.00000498, 0.00000144, -0.00000310, -0.00000318, 0.00000252, 0.00000973, 0.00001320, 0.00001160, 0.00000724, 0.00000225, -0.00000334, -0.00000934, -0.00001302, -0.00001097, -0.00000341, 0.00000485, 0.00000857, 0.00000668, 0.00000225, -0.00000170, -0.00000490, -0.00000818, -0.00001048, -0.00000923, -0.00000376, 0.00000275, 0.00000564, 0.00000298, -0.00000274, -0.00000671, -0.00000528, 0.00000167, 0.00001004, 0.00001325, 0.00000685, -0.00000616, -0.00001506, -0.00000984, 0.00000753, 0.00002160, 0.00001673, -0.00000599, -0.00002687, -0.00002566, -0.00000227, 0.00002198, 0.00002511, 0.00000577, -0.00001687, -0.00002291, -0.00000929, 0.00001157, 0.00002591, 0.00002782, 0.00001715, -0.00000421, -0.00002919, -0.00004294, -0.00003195, -0.00000052, 0.00002655, 0.00002525, -0.00000121, -0.00002246, -0.00001245, 0.00002016, 0.00003883, 0.00001903, -0.00002248, -0.00004367, -0.00002253, 0.00001857, 0.00003718, 0.00001661, -0.00001773, -0.00002835, -0.00000663, 0.00002080, 0.00002427, 0.00000306, -0.00001780, -0.00001807, -0.00000288, 0.00000762, 0.00000329, -0.00000702, -0.00000976, -0.00000360, 0.00000155, -0.00000091, -0.00000635, -0.00000645, -0.00000068, 0.00000360, 0.00000148, -0.00000325, -0.00000368, 0.00000088, 0.00000431, 0.00000208, -0.00000293, -0.00000452, -0.00000127, 0.00000198, 0.00000088, -0.00000288, -0.00000423, -0.00000157, 0.00000141, 0.00000091, -0.00000192, -0.00000270, 0.00000032, 0.00000410, 0.00000477, 0.00000233, 0.00000004, 0.00000004, 0.00000118, 0.00000127, 0.00000011, -0.00000093, -0.00000145, -0.00000241, },
		{ 0.00000187, 0.00000486, 0.00000370, -0.00000098, -0.00000406, -0.00000116, 0.00000650, 0.00001311, 0.00001406, 0.00000967, 0.00000335, -0.00000245, -0.00000738, -0.00001099, -0.00001134, -0.00000703, 0.00000025, 0.00000636, 0.00000819, 0.00000583, 0.00000129, -0.00000391, -0.00000891, -0.00001202, -0.00001083, -0.00000493, 0.00000226, 0.00000558, 0.00000287, -0.00000282, -0.00000598, -0.00000338, 0.00000342, 0.00000955, 0.00001017, 0.00000371, -0.00000621, -0.00001188, -0.00000675, 0.00000743, 0.00001905, 0.00001544, -0.00000393, -0.00002365, -0.00002504, -0.00000500, 0.00001896, 0.00002470, 0.00000719, -0.00001651, -0.00002443, -0.00001035, 0.00001283, 0.00002767, 0.00002680, 0.00001346, -0.00000634, -0.00002687, -0.00003915, -0.00003254, -0.00000637, 0.00002200, 0.00002859, 0.00000755, -0.00001854, -0.00001974, 0.00000777, 0.00003424, 0.00002723, -0.00001093, -0.00004238, -0.00003356, 0.00000754, 0.00003901, 0.00003004, -0.00000663, -0.00003042, -0.00001792, 0.00001350, 0.00002805, 0.00001179, -0.00001490, -0.00002347, -0.00000893, 0.00000846, 0.00000937, -0.00000359, -0.00001293, -0.00000881, 0.00000118, 0.00000420, -0.00000166, -0.00000708, -0.00000490, 0.00000132, 0.00000343, -0.00000024, -0.00000361, -0.00000179, 0.00000253, 0.00000333, -0.00000043, -0.00000372, -0.00000256, 0.00000076, 0.00000124, -0.00000187, -0.00000417, -0.00000230, 0.00000139, 0.00000199, -0.00000130, -0.00000393, -0.00000188, 0.00000321, 0.00000602, 0.00000408, 0.00000032, -0.00000101, 0.00000059, 0.00000205, 0.00000120, -0.00000088, -0.00000208, -0.00000227, -0.00000269, -0.00000315, -0.00000178, },
		{ 0.00000150, -0.00000308, -0.00000349, 0.00000255, 0.00001103, 0.00001536, 0.00001261, 0.00000547, -0.00000157, -0.00000634, -0.00000923, -0.00001043, -0.00000880, -0.00000373, 0.00000294, 0.00000777, 0.00000833, 0.00000452, -0.00000198, -0.00000878, -0.00001306, -0.00001229, -0.00000631, 0.00000129, 0.00000518, 0.00000286, -0.00000263, -0.00000536, -0.00000204, 0.00000489, 0.00000962, 0.00000799, 0.00000069, -0.00000725, -0.00000971, -0.00000346, 0.00000853, 0.00001722, 0.00001346, -0.00000310, -0.00002056, -0.00002323, -0.00000673, 0.00001539, 0.00002275, 0.00000794, -0.00001525, -0.00002469, -0.00001130, 0.00001320, 0.00002902, 0.00002635, 0.00001017, -0.00000931, -0.00002559, -0.00003464, -0.00003067, -0.00001075, 0.00001563, 0.00002841, 0.00001543, -0.00001095, -0.00002237, -0.00000439, 0.00002458, 0.00003046, 0.00000158, -0.00003489, -0.00004004, -0.00000596, 0.00003374, 0.00003931, 0.00000748, -0.00002603, -0.00002649, 0.00000275, 0.00002680, 0.00001966, -0.00000835, -0.00002556, -0.00001576, 0.00000600, 0.00001407, 0.00000190, -0.00001352, -0.00001430, -0.00000191, 0.00000760, 0.00000432, -0.00000521, -0.00000848, -0.00000270, 0.00000371, 0.00000324, -0.00000174, -0.00000372, -0.00000043, 0.00000311, 0.00000214, -0.00000161, -0.00000306, -0.00000103, 0.00000065, -0.00000091, -0.00000339, -0.00000272, 0.00000083, 0.00000258, -0.00000023, -0.00000425, -0.00000392, 0.00000143, 0.00000641, 0.00000602, 0.00000146, -0.00000177, -0.00000065, 0.00000223, 0.00000264, -0.00000001, -0.00000267, -0.00000310, -0.00000226, -0.00000196, -0.00000173, 0.00000025, 0.00000350, 0.00000462, },
		{ -0.00000101, 0.00000736, 0.00001478, 0.00001523, 0.00000864, 0.00000006, -0.00000584, -0.00000837, -0.00000918, -0.00000886, -0.00000621, -0.00000069, 0.00000562, 0.00000909, 0.00000728, 0.00000070, -0.00000755, -0.00001332, -0.00001347, -0.00000784, -0.00000010, 0.00000439, 0.00000281, -0.00000220, -0.00000468, -0.00000111, 0.00000588, 0.00000985, 0.00000662, -0.00000177, -0.00000878, -0.00000864, -0.00000060, 0.00001040, 0.00001642, 0.00001145, -0.00000331, -0.00001816, -0.00002081, -0.00000733, 0.00001195, 0.00001967, 0.00000765, -0.00001366, -0.00002375, -0.00001173, 0.00001291, 0.00002964, 0.00002619, 0.00000757, -0.00001257, -0.00002553, -0.00003035, -0.00002689, -0.00001286, 0.00000892, 0.00002499, 0.00002079, -0.00000157, -0.00002007, -0.00001389, 0.00001181, 0.00002776, 0.00001203, -0.00002289, -0.00004020, -0.00001851, 0.00002266, 0.00004216, 0.00002135, -0.00001587, -0.00002993, -0.00000890, 0.00002035, 0.00002434, 0.00000026, -0.00002345, -0.00002140, 0.00000082, 0.00001600, 0.00000807, -0.00001115, -0.00001851, -0.00000702, 0.00000830, 0.00001023, -0.00000088, -0.00001020, -0.00000755, 0.00000185, 0.00000609, 0.00000165, -0.00000410, -0.00000375, 0.00000122, 0.00000389, 0.00000143, -0.00000225, -0.00000275, -0.00000094, -0.00000052, -0.00000210, -0.00000246, 0.00000008, 0.00000244, 0.00000077, -0.00000369, -0.00000523, -0.00000081, 0.00000566, 0.00000752, 0.00000330, -0.00000173, -0.00000209, 0.00000156, 0.00000383, 0.00000159, -0.00000260, -0.00000436, -0.00000285, -0.00000095, -0.00000059, -0.00000034, 0.00000167, 0.00000403, 0.00000333, -0.00000085, -0.00000395, },
		{ 0.00001654, 0.00001234, 0.00000293, -0.00000512, -0.00000843, -0.00000849, -0.00000793, -0.00000684, -0.00000344, 0.00000246, 0.00000787, 0.00000874, 0.00000356, -0.00000518, -0.00001250, -0.00001415, -0.00000945, -0.00000187, 0.00000317, 0.00000259, -0.00000160, -0.00000383, -0.00000042, 0.00000636, 0.00000992, 0.00000581, -0.00000346, -0.00001027, -0.00000844, 0.00000145, 0.00001249, 0.00001658, 0.00000995, -0.00000413, -0.00001668, -0.00001842, -0.00000702, 0.00000914, 0.00001605, 0.00000631, -0.00001226, -0.00002193, -0.00001138, 0.00001231, 0.00002943, 0.00002596, 0.00000573, -0.00001560, -0.00002647, -0.00002699, -0.00002211, -0.00001249, 0.00000324, 0.00001931, 0.00002273, 0.00000752, -0.00001371, -0.00001902, -0.00000149, 0.00001964, 0.00001797, -0.00000913, -0.00003388, -0.00002702, 0.00000845, 0.00003788, 0.00003172, -0.00000210, -0.00002705, -0.00001854, 0.00000997, 0.00002419, 0.00000862, -0.00001740, -0.00002403, -0.00000564, 0.00001456, 0.00001323, -0.00000635, -0.00002020, -0.00001278, 0.00000602, 0.00001455, 0.00000506, -0.00000929, -0.00001194, -0.00000198, 0.00000727, 0.00000572, -0.00000254, -0.00000643, -0.00000203, 0.00000412, 0.00000462, -0.00000001, -0.00000369, -0.00000321, -0.00000114, -0.00000077, -0.00000137, -0.00000037, 0.00000165, 0.00000121, -0.00000259, -0.00000549, -0.00000286, 0.00000387, 0.00000801, 0.00000535, -0.00000066, -0.00000309, 0.00000018, 0.00000421, 0.00000344, -0.00000158, -0.00000534, -0.00000432, -0.00000079, 0.00000096, 0.00000033, 0.00000029, 0.00000236, 0.00000386, 0.00000166, -0.00000259, -0.00000320, 0.00000300, 0.00001214, },
		{ -0.00000336, -0.00000891, -0.00000888, -0.00000700, -0.00000592, -0.00000453, -0.00000064, 0.00000499, 0.00000831, 0.00000578, -0.00000201, -0.00001035, -0.00001396, -0.00001092, -0.00000401, 0.00000150, 0.00000212, -0.00000093, -0.00000280, 0.00000021, 0.00000640, 0.00000963, 0.00000527, -0.00000441, -0.00001134, -0.00000871, 0.00000256, 0.00001425, 0.00001740, 0.00000926, -0.00000501, -0.00001605, -0.00001655, -0.00000622, 0.00000716, 0.00001249, 0.00000420, -0.00001143, -0.00001968, -0.00001024, 0.00001176, 0.00002851, 0.00002539, 0.00000451, -0.00001800, -0.00002798, -0.00002489, -0.00001729, -0.00001005, -0.00000042, 0.00001277, 0.00002127, 0.00001451, -0.00000500, -0.00001923, -0.00001274, 0.00000783, 0.00001815, 0.00000332, -0.00002253, -0.00002948, -0.00000553, 0.00002748, 0.00003613, 0.00001216, -0.00001824, -0.00002363, -0.00000199, 0.00001882, 0.00001434, -0.00000882, -0.00002259, -0.00001142, 0.00001010, 0.00001589, -0.00000040, -0.00001880, -0.00001760, 0.00000135, 0.00001617, 0.00001124, -0.00000569, -0.00001457, -0.00000693, 0.00000618, 0.00000929, 0.00000075, -0.00000753, -0.00000582, 0.00000253, 0.00000705, 0.00000329, -0.00000328, -0.00000549, -0.00000284, 0.00000001, 0.00000033, -0.00000011, 0.00000057, 0.00000083, -0.00000152, -0.00000472, -0.00000410, 0.00000154, 0.00000718, 0.00000692, 0.00000126, -0.00000312, -0.00000140, 0.00000355, 0.00000487, 0.00000022, -0.00000542, -0.00000605, -0.00000175, 0.00000209, 0.00000192, 0.00000001, 0.00000042, 0.00000297, 0.00000346, -0.00000007, -0.00000350, -0.00000091, 0.00000787, 0.00001584, 0.00001564, 0.00000701, },
		{ -0.00000697, -0.00000431, -0.00000376, -0.00000260, 0.00000131, 0.00000586, 0.00000654, 0.00000129, -0.00000693, -0.00001251, -0.00001191, -0.00000640, -0.00000069, 0.00000129, -0.00000030, -0.00000160, 0.00000088, 0.00000615, 0.00000892, 0.00000470, -0.00000479, -0.00001177, -0.00000904, 0.00000290, 0.00001534, 0.00001843, 0.00000937, -0.00000552, -0.00001596, -0.00001545, -0.00000544, 0.00000591, 0.00000946, 0.00000176, -0.00001124, -0.00001749, -0.00000854, 0.00001150, 0.00002714, 0.00002432, 0.00000364, -0.00001959, -0.00002955, -0.00002399, -0.00001322, -0.00000632, -0.00000169, 0.00000681, 0.00001720, 0.00001832, 0.00000402, -0.00001510, -0.00002004, -0.00000517, 0.00001273, 0.00001182, -0.00000875, -0.00002544, -0.00001610, 0.00001342, 0.00003364, 0.00002364, -0.00000539, -0.00002274, -0.00001265, 0.00000920, 0.00001564, 0.00000010, -0.00001712, -0.00001467, 0.00000392, 0.00001524, 0.00000501, -0.00001466, -0.00002009, -0.00000441, 0.00001465, 0.00001611, -0.00000011, -0.00001453, -0.00001169, 0.00000288, 0.00001125, 0.00000496, -0.00000656, -0.00000908, -0.00000061, 0.00000791, 0.00000688, -0.00000132, -0.00000702, -0.00000532, -0.00000023, 0.00000215, 0.00000103, -0.00000027, -0.00000031, -0.00000103, -0.00000330, -0.00000419, -0.00000066, 0.00000516, 0.00000738, 0.00000350, -0.00000192, -0.00000249, 0.00000203, 0.00000528, 0.00000224, -0.00000436, -0.00000728, -0.00000357, 0.00000214, 0.00000372, 0.00000098, -0.00000091, 0.00000112, 0.00000392, 0.00000262, -0.00000199, -0.00000338, 0.00000293, 0.00001294, 0.00001748, 0.00001167, -0.00000009, -0.00000892, -0.00001027, },
		{ -0.00000271, -0.00000203, 0.00000184, 0.00000527, 0.00000383, -0.00000265, -0.00000958, -0.00001196, -0.00000877, -0.00000340, -0.00000003, 0.00000018, -0.00000026, 0.00000169, 0.00000575, 0.00000782, 0.00000391, -0.00000485, -0.00001155, -0.00000908, 0.00000275, 0.00001564, 0.00001923, 0.00001005, -0.00000538, -0.00001599, -0.00001509, -0.00000511, 0.00000508, 0.00000717, -0.00000049, -0.00001152, -0.00001567, -0.00000662, 0.00001158, 0.00002560, 0.00002280, 0.00000288, -0.00002040, -0.00003077, -0.00002399, -0.00001036, -0.00000222, -0.00000077, 0.00000250, 0.00001179, 0.00001869, 0.00001147, -0.00000811, -0.00002254, -0.00001677, 0.00000319, 0.00001484, 0.00000439, -0.00001607, -0.00002104, -0.00000105, 0.00002494, 0.00002979, 0.00000857, -0.00001598, -0.00001942, -0.00000251, 0.00001187, 0.00000697, -0.00000880, -0.00001414, -0.00000211, 0.00001147, 0.00000833, -0.00000895, -0.00001954, -0.00000966, 0.00001043, 0.00001841, 0.00000609, -0.00001161, -0.00001489, -0.00000185, 0.00001087, 0.00000889, -0.00000366, -0.00001081, -0.00000451, 0.00000683, 0.00000980, 0.00000187, -0.00000715, -0.00000791, -0.00000165, 0.00000346, 0.00000289, -0.00000039, -0.00000183, -0.00000142, -0.00000185, -0.00000312, -0.00000205, 0.00000245, 0.00000639, 0.00000529, 0.00000032, -0.00000247, 0.00000022, 0.00000444, 0.00000375, -0.00000237, -0.00000739, -0.00000559, 0.00000091, 0.00000486, 0.00000282, -0.00000101, -0.00000083, 0.00000297, 0.00000450, 0.00000063, -0.00000386, -0.00000148, 0.00000835, 0.00001713, 0.00001584, 0.00000457, -0.00000759, -0.00001196, -0.00000827, -0.00000308, -0.00000159, },
		{ 0.00000191, 0.00000470, 0.00000170, -0.00000523, -0.00001060, -0.00001068, -0.00000651, -0.00000198, 0.00000031, 0.00000112, 0.00000267, 0.00000535, 0.00000644, 0.00000281, -0.00000484, -0.00001082, -0.00000864, 0.00000249, 0.00001527, 0.00001951, 0.00001093, -0.00000461, -0.00001578, -0.00001524, -0.00000541, 0.00000429, 0.00000558, -0.00000219, -0.00001195, -0.00001435, -0.00000488, 0.00001187, 0.00002410, 0.00002099, 0.00000207, -0.00002062, -0.00003138, -0.00002444, -0.00000880, 0.00000144, 0.00000172, 0.00000038, 0.00000643, 0.00001620, 0.00001609, -0.00000021, -0.00002052, -0.00002491, -0.00000816, 0.00001223, 0.00001418, -0.00000374, -0.00001964, -0.00001274, 0.00001223, 0.00002940, 0.00002048, -0.00000492, -0.00002068, -0.00001358, 0.00000373, 0.00000990, 0.00000033, -0.00000967, -0.00000610, 0.00000568, 0.00000859, -0.00000329, -0.00001609, -0.00001291, 0.00000470, 0.00001755, 0.00001135, -0.00000641, -0.00001550, -0.00000674, 0.00000811, 0.00001137, 0.00000043, -0.00001041, -0.00000808, 0.00000399, 0.00001116, 0.00000554, -0.00000562, -0.00000981, -0.00000396, 0.00000373, 0.00000495, 0.00000045, -0.00000317, -0.00000266, -0.00000095, -0.00000130, -0.00000220, -0.00000021, 0.00000406, 0.00000594, 0.00000301, -0.00000107, -0.00000111, 0.00000254, 0.00000409, -0.00000009, -0.00000611, -0.00000696, -0.00000128, 0.00000469, 0.00000474, 0.00000024, -0.00000200, 0.00000112, 0.00000494, 0.00000333, -0.00000248, -0.00000431, 0.00000307, 0.00001439, 0.00001841, 0.00000991, -0.00000443, -0.00001296, -0.00001068, -0.00000313, 0.00000098, -0.00000089, -0.00000390, -0.00000277, },
		{ 0.00000002, -0.00000755, -0.00001153, -0.00000970, -0.00000463, -0.00000011, 0.00000238, 0.00000384, 0.00000508, 0.00000493, 0.00000141, -0.00000496, -0.00000981, -0.00000770, 0.00000240, 0.00001448, 0.00001917, 0.00001166, -0.00000339, -0.00001512, -0.00001557, -0.00000634, 0.00000321, 0.00000447, -0.00000314, -0.00001214, -0.00001342, -0.00000357, 0.00001212, 0.00002272, 0.00001910, 0.00000118, -0.00002043, -0.00003134, -0.00002493, -0.00000834, 0.00000411, 0.00000495, 0.00000046, 0.00000224, 0.00001193, 0.00001742, 0.00000676, -0.00001523, -0.00002852, -0.00001878, 0.00000516, 0.00001887, 0.00000859, -0.00001275, -0.00001931, -0.00000144, 0.00002289, 0.00002771, 0.00000778, -0.00001624, -0.00002139, -0.00000688, 0.00000801, 0.00000792, -0.00000221, -0.00000670, -0.00000039, 0.00000577, 0.00000072, -0.00001076, -0.00001329, -0.00000094, 0.00001383, 0.00001431, -0.00000021, -0.00001319, -0.00001037, 0.00000362, 0.00001156, 0.00000455, -0.00000790, -0.00001025, 0.00000013, 0.00001049, 0.00000867, -0.00000267, -0.00001030, -0.00000653, 0.00000274, 0.00000656, 0.00000209, -0.00000378, -0.00000438, -0.00000100, 0.00000064, -0.00000109, -0.00000206, 0.00000091, 0.00000505, 0.00000534, 0.00000155, -0.00000127, 0.00000023, 0.00000300, 0.00000167, -0.00000371, -0.00000700, -0.00000368, 0.00000303, 0.00000582, 0.00000240, -0.00000177, -0.00000075, 0.00000389, 0.00000510, 0.00000004, -0.00000501, -0.00000161, 0.00000978, 0.00001854, 0.00001476, 0.00000042, -0.00001228, -0.00001338, -0.00000487, 0.00000277, 0.00000224, -0.00000364, -0.00000668, -0.00000294, 0.00000334, 0.00000509, },
		{ -0.00001241, -0.00000784, -0.00000133, 0.00000328, 0.00000512, 0.00000507, 0.00000345, -0.00000022, -0.00000533, -0.00000873, -0.00000637, 0.00000265, 0.00001357, 0.00001830, 0.00001199, -0.00000204, -0.00001403, -0.00001577, -0.00000768, 0.00000170, 0.00000354, -0.00000339, -0.00001181, -0.00001264, -0.00000279, 0.00001204, 0.00002138, 0.00001729, 0.00000030, -0.00001999, -0.00003072, -0.00002517, -0.00000863, 0.00000560, 0.00000810, 0.00000224, -0.00000011, 0.00000718, 0.00001579, 0.00001141, -0.00000841, -0.00002760, -0.00002656, -0.00000427, 0.00001803, 0.00001819, -0.00000249, -0.00001982, -0.00001296, 0.00001214, 0.00002887, 0.00001909, -0.00000736, -0.00002420, -0.00001739, 0.00000175, 0.00001203, 0.00000639, -0.00000353, -0.00000485, 0.00000079, 0.00000200, -0.00000508, -0.00001078, -0.00000497, 0.00000830, 0.00001425, 0.00000544, -0.00000841, -0.00001163, -0.00000136, 0.00000928, 0.00000744, -0.00000394, -0.00001032, -0.00000365, 0.00000789, 0.00001035, 0.00000096, -0.00000906, -0.00000853, 0.00000070, 0.00000715, 0.00000408, -0.00000338, -0.00000598, -0.00000200, 0.00000202, 0.00000087, -0.00000262, -0.00000226, 0.00000270, 0.00000651, 0.00000475, 0.00000007, -0.00000165, 0.00000068, 0.00000217, -0.00000090, -0.00000546, -0.00000537, 0.00000027, 0.00000540, 0.00000458, -0.00000006, -0.00000172, 0.00000191, 0.00000531, 0.00000261, -0.00000371, -0.00000460, 0.00000441, 0.00001602, 0.00001787, 0.00000615, -0.00000937, -0.00001519, -0.00000796, 0.00000279, 0.00000552, -0.00000127, -0.00000873, -0.00000816, -0.00000024, 0.00000655, 0.00000525, -0.00000292, -0.00001077, },
		{ 0.00000351, 0.00000637, 0.00000537, 0.00000219, -0.00000194, -0.00000597, -0.00000779, -0.00000485, 0.00000328, 0.00001280, 0.00001713, 0.00001183, -0.00000085, -0.00001268, -0.00001570, -0.00000915, -0.00000019, 0.00000253, -0.00000313, -0.00001085, -0.00001173, -0.00000246, 0.00001148, 0.00001992, 0.00001561, -0.00000047, -0.00001937, -0.00002966, -0.00002503, -0.00000927, 0.00000599, 0.00001062, 0.00000502, -0.00000047, 0.00000307, 0.00001213, 0.00001313, -0.00000188, -0.00002314, -0.00003025, -0.00001366, 0.00001256, 0.00002319, 0.00000836, -0.00001483, -0.00001994, -0.00000004, 0.00002412, 0.00002641, 0.00000359, -0.00002156, -0.00002527, -0.00000728, 0.00001160, 0.00001390, 0.00000267, -0.00000628, -0.00000476, 0.00000034, -0.00000066, -0.00000623, -0.00000639, 0.00000248, 0.00001128, 0.00000912, -0.00000232, -0.00001003, -0.00000543, 0.00000509, 0.00000821, 0.00000033, -0.00000820, -0.00000625, 0.00000404, 0.00001001, 0.00000429, -0.00000625, -0.00000920, -0.00000178, 0.00000642, 0.00000573, -0.00000203, -0.00000688, -0.00000361, 0.00000240, 0.00000303, -0.00000182, -0.00000465, -0.00000061, 0.00000608, 0.00000766, 0.00000276, -0.00000235, -0.00000219, 0.00000109, 0.00000141, -0.00000261, -0.00000560, -0.00000278, 0.00000328, 0.00000583, 0.00000264, -0.00000113, -0.00000008, 0.00000392, 0.00000420, -0.00000117, -0.00000529, -0.00000039, 0.00001141, 0.00001832, 0.00001148, -0.00000439, -0.00001501, -0.00001144, 0.00000067, 0.00000761, 0.00000244, -0.00000825, -0.00001234, -0.00000534, 0.00000548, 0.00000932, 0.00000270, -0.00000807, -0.00001395, -0.00001127, -0.00000348, },
		{ 0.00000134, -0.00000357, -0.00000685, -0.00000713, -0.00000333, 0.00000420, 0.00001234, 0.00001592, 0.00001129, -0.00000004, -0.00001133, -0.00001538, -0.00001051, -0.00000228, 0.00000137, -0.00000261, -0.00000932, -0.00001049, -0.00000237, 0.00001038, 0.00001815, 0.00001401, -0.00000103, -0.00001857, -0.00002830, -0.00002450, -0.00000998, 0.00000554, 0.00001221, 0.00000805, 0.00000083, 0.00000036, 0.00000757, 0.00001203, 0.00000295, -0.00001671, -0.00002962, -0.00002090, 0.00000428, 0.00002295, 0.00001715, -0.00000620, -0.00002127, -0.00001069, 0.00001505, 0.00002823, 0.00001378, -0.00001444, -0.00002867, -0.00001672, 0.00000683, 0.00001833, 0.00001025, -0.00000415, -0.00000906, -0.00000353, 0.00000136, -0.00000104, -0.00000502, -0.00000207, 0.00000633, 0.00000998, 0.00000354, -0.00000591, -0.00000739, 0.00000012, 0.00000654, 0.00000366, -0.00000446, -0.00000686, 0.00000000, 0.00000764, 0.00000632, -0.00000254, -0.00000812, -0.00000386, 0.00000451, 0.00000643, -0.00000018, -0.00000672, -0.00000522, 0.00000170, 0.00000467, -0.00000002, -0.00000569, -0.00000404, 0.00000404, 0.00000944, 0.00000619, -0.00000143, -0.00000471, -0.00000137, 0.00000242, 0.00000078, -0.00000406, -0.00000510, -0.00000010, 0.00000540, 0.00000538, 0.00000108, -0.00000109, 0.00000152, 0.00000412, 0.00000147, -0.00000382, -0.00000341, 0.00000584, 0.00001585, 0.00001502, 0.00000179, -0.00001220, -0.00001396, -0.00000318, 0.00000745, 0.00000616, -0.00000535, -0.00001424, -0.00001068, 0.00000194, 0.00001115, 0.00000830, -0.00000352, -0.00001364, -0.00001427, -0.00000646, 0.00000274, 0.00000726, 0.00000596, },
		{ -0.00000682, -0.00000203, 0.00000526, 0.00001221, 0.00001488, 0.00001053, 0.00000035, -0.00001024, -0.00001496, -0.00001165, -0.00000432, 0.00000009, -0.00000202, -0.00000740, -0.00000886, -0.00000230, 0.00000886, 0.00001600, 0.00001236, -0.00000139, -0.00001756, -0.00002671, -0.00002366, -0.00001055, 0.00000459, 0.00001285, 0.00001075, 0.00000320, -0.00000071, 0.00000322, 0.00000885, 0.00000534, -0.00001000, -0.00002544, -0.00002462, -0.00000454, 0.00001810, 0.00002193, 0.00000352, -0.00001727, -0.00001743, 0.00000417, 0.00002449, 0.00002069, -0.00000491, -0.00002698, -0.00002411, -0.00000092, 0.00001853, 0.00001712, 0.00000101, -0.00001066, -0.00000818, 0.00000060, 0.00000323, -0.00000152, -0.00000428, 0.00000083, 0.00000800, 0.00000774, -0.00000029, -0.00000665, -0.00000424, 0.00000286, 0.00000506, -0.00000016, -0.00000530, -0.00000308, 0.00000386, 0.00000641, 0.00000107, -0.00000540, -0.00000474, 0.00000199, 0.00000581, 0.00000151, -0.00000550, -0.00000620, 0.00000024, 0.00000527, 0.00000210, -0.00000525, -0.00000672, 0.00000090, 0.00000953, 0.00000943, 0.00000099, -0.00000602, -0.00000453, 0.00000172, 0.00000373, -0.00000104, -0.00000586, -0.00000386, 0.00000311, 0.00000712, 0.00000436, -0.00000043, -0.00000091, 0.00000233, 0.00000311, -0.00000096, -0.00000402, 0.00000076, 0.00001099, 0.00001573, 0.00000775, -0.00000689, -0.00001425, -0.00000764, 0.00000465, 0.00000839, -0.00000092, -0.00001315, -0.00001474, -0.00000329, 0.00001010, 0.00001259, 0.00000224, -0.00001108, -0.00001603, -0.00000990, 0.00000075, 0.00000741, 0.00000667, 0.00000103, -0.00000488, -0.00000785, },
		{ 0.00001237, 0.00001417, 0.00000979, 0.00000038, -0.00000957, -0.00001466, -0.00001258, -0.00000613, -0.00000114, -0.00000145, -0.00000532, -0.00000693, -0.00000210, 0.00000712, 0.00001348, 0.00001057, -0.00000162, -0.00001634, -0.00002494, -0.00002259, -0.00001090, 0.00000345, 0.00001272, 0.00001277, 0.00000599, -0.00000029, -0.00000016, 0.00000458, 0.00000523, -0.00000442, -0.00001905, -0.00002443, -0.00001184, 0.00001025, 0.00002191, 0.00001177, -0.00000956, -0.00001905, -0.00000572, 0.00001649, 0.00002276, 0.00000442, -0.00002088, -0.00002762, -0.00000952, 0.00001449, 0.00002133, 0.00000775, -0.00000892, -0.00001188, -0.00000237, 0.00000537, 0.00000282, -0.00000382, -0.00000371, 0.00000393, 0.00000935, 0.00000538, -0.00000339, -0.00000675, -0.00000180, 0.00000411, 0.00000345, -0.00000201, -0.00000434, -0.00000033, 0.00000447, 0.00000358, -0.00000163, -0.00000395, -0.00000031, 0.00000396, 0.00000237, -0.00000362, -0.00000613, -0.00000139, 0.00000468, 0.00000377, -0.00000364, -0.00000801, -0.00000250, 0.00000787, 0.00001157, 0.00000433, -0.00000560, -0.00000742, -0.00000055, 0.00000532, 0.00000267, -0.00000468, -0.00000690, -0.00000062, 0.00000704, 0.00000772, 0.00000198, -0.00000230, -0.00000056, 0.00000300, 0.00000208, -0.00000231, -0.00000255, 0.00000497, 0.00001329, 0.00001197, -0.00000002, -0.00001156, -0.00001120, -0.00000033, 0.00000795, 0.00000352, -0.00000923, -0.00001612, -0.00000879, 0.00000622, 0.00001438, 0.00000807, -0.00000633, -0.00001575, -0.00001306, -0.00000242, 0.00000639, 0.00000717, 0.00000128, -0.00000561, -0.00000878, -0.00000687, -0.00000109, 0.00000626, },
		{ 0.00000022, -0.00000934, -0.00001468, -0.00001345, -0.00000762, -0.00000212, -0.00000084, -0.00000323, -0.00000487, -0.00000174, 0.00000535, 0.00001074, 0.00000858, -0.00000183, -0.00001497, -0.00002302, -0.00002134, -0.00001103, 0.00000233, 0.00001209, 0.00001401, 0.00000863, 0.00000114, -0.00000220, 0.00000024, 0.00000315, -0.00000081, -0.00001202, -0.00002086, -0.00001622, 0.00000144, 0.00001751, 0.00001662, -0.00000044, -0.00001570, -0.00001232, 0.00000650, 0.00001975, 0.00001116, -0.00001218, -0.00002647, -0.00001666, 0.00000734, 0.00002166, 0.00001413, -0.00000419, -0.00001321, -0.00000627, 0.00000494, 0.00000652, -0.00000120, -0.00000616, -0.00000091, 0.00000819, 0.00000966, 0.00000150, -0.00000673, -0.00000610, 0.00000111, 0.00000533, 0.00000202, -0.00000348, -0.00000378, 0.00000100, 0.00000427, 0.00000223, -0.00000149, -0.00000158, 0.00000137, 0.00000197, -0.00000174, -0.00000495, -0.00000251, 0.00000318, 0.00000442, -0.00000155, -0.00000771, -0.00000524, 0.00000493, 0.00001196, 0.00000762, -0.00000350, -0.00000909, -0.00000368, 0.00000507, 0.00000600, -0.00000182, -0.00000827, -0.00000484, 0.00000491, 0.00001002, 0.00000559, -0.00000191, -0.00000350, 0.00000108, 0.00000410, 0.00000092, -0.00000334, -0.00000069, 0.00000821, 0.00001328, 0.00000689, -0.00000597, -0.00001238, -0.00000623, 0.00000438, 0.00000629, -0.00000350, -0.00001409, -0.00001285, 0.00000040, 0.00001294, 0.00001257, -0.00000007, -0.00001294, -0.00001503, -0.00000635, 0.00000388, 0.00000697, 0.00000194, -0.00000558, -0.00000938, -0.00000719, -0.00000064, 0.00000703, 0.00001269, 0.00001382, 0.00000925, },
		{ -0.00001443, -0.00000883, -0.00000271, -0.00000005, -0.00000121, -0.00000287, -0.00000131, 0.00000368, 0.00000797, 0.00000648, -0.00000210, -0.00001355, -0.00002102, -0.00001997, -0.00001095, 0.00000137, 0.00001122, 0.00001455, 0.00001074, 0.00000305, -0.00000296, -0.00000340, -0.00000005, 0.00000063, -0.00000573, -0.00001506, -0.00001713, -0.00000632, 0.00001009, 0.00001716, 0.00000765, -0.00000873, -0.00001437, -0.00000291, 0.00001272, 0.00001372, -0.00000325, -0.00002114, -0.00002051, -0.00000101, 0.00001793, 0.00001825, 0.00000227, -0.00001142, -0.00000950, 0.00000233, 0.00000839, 0.00000242, -0.00000610, -0.00000504, 0.00000491, 0.00001157, 0.00000664, -0.00000423, -0.00000880, -0.00000301, 0.00000498, 0.00000557, -0.00000084, -0.00000556, -0.00000304, 0.00000294, 0.00000515, 0.00000211, -0.00000125, -0.00000116, 0.00000033, -0.00000055, -0.00000300, -0.00000265, 0.00000136, 0.00000383, 0.00000022, -0.00000610, -0.00000661, 0.00000155, 0.00001047, 0.00000989, -0.00000028, -0.00000896, -0.00000668, 0.00000306, 0.00000796, 0.00000191, -0.00000752, -0.00000836, 0.00000113, 0.00001036, 0.00000933, 0.00000039, -0.00000534, -0.00000203, 0.00000429, 0.00000440, -0.00000166, -0.00000458, 0.00000173, 0.00001122, 0.00001215, 0.00000156, -0.00001028, -0.00001133, -0.00000181, 0.00000603, 0.00000231, -0.00000883, -0.00001397, -0.00000583, 0.00000832, 0.00001439, 0.00000646, -0.00000760, -0.00001488, -0.00001027, -0.00000011, 0.00000555, 0.00000262, -0.00000470, -0.00000934, -0.00000758, -0.00000071, 0.00000738, 0.00001301, 0.00001381, 0.00000902, 0.00000004, -0.00000949, -0.00001513, },
		{ 0.00000106, 0.00000083, -0.00000104, -0.00000095, 0.00000213, 0.00000536, 0.00000441, -0.00000244, -0.00001220, -0.00001904, -0.00001855, -0.00001069, 0.00000064, 0.00001034, 0.00001459, 0.00001218, 0.00000490, -0.00000275, -0.00000594, -0.00000347, 0.00000026, -0.00000104, -0.00000845, -0.00001491, -0.00001162, 0.00000155, 0.00001363, 0.00001284, -0.00000024, -0.00001187, -0.00000954, 0.00000368, 0.00001175, 0.00000366, -0.00001317, -0.00002019, -0.00000839, 0.00001107, 0.00001886, 0.00000864, -0.00000675, -0.00001065, -0.00000132, 0.00000789, 0.00000565, -0.00000389, -0.00000734, 0.00000069, 0.00001083, 0.00001067, -0.00000002, -0.00000915, -0.00000700, 0.00000256, 0.00000760, 0.00000271, -0.00000529, -0.00000653, 0.00000000, 0.00000635, 0.00000594, 0.00000074, -0.00000277, -0.00000212, -0.00000052, -0.00000091, -0.00000166, -0.00000011, 0.00000222, 0.00000104, -0.00000383, -0.00000633, -0.00000132, 0.00000752, 0.00001041, 0.00000308, -0.00000698, -0.00000852, -0.00000004, 0.00000795, 0.00000539, -0.00000487, -0.00001014, -0.00000329, 0.00000844, 0.00001198, 0.00000405, -0.00000527, -0.00000524, 0.00000252, 0.00000683, 0.00000166, -0.00000586, -0.00000450, 0.00000627, 0.00001442, 0.00000938, -0.00000490, -0.00001395, -0.00000924, 0.00000220, 0.00000633, -0.00000147, -0.00001130, -0.00001056, 0.00000145, 0.00001270, 0.00001170, -0.00000045, -0.00001198, -0.00001312, -0.00000519, 0.00000248, 0.00000276, -0.00000312, -0.00000839, -0.00000774, -0.00000124, 0.00000718, 0.00001316, 0.00001403, 0.00000915, -0.00000000, -0.00000986, -0.00001599, -0.00001567, -0.00000990, -0.00000289, },
		},
		{ { 0.00000270, 0.00000214, 0.00000126, 0.00000115, 0.00000134, 0.00000045, -0.00000195, -0.00000492, -0.00000710, -0.00000738, -0.00000516, -0.00000126, 0.00000232, 0.00000437, 0.00000540, 0.00000538, 0.00000231, -0.00000367, -0.00000643, 0.00000070, 0.00001398, 0.00001934, 0.00000823, -0.00000952, -0.00001522, -0.00000338, 0.00001155, 0.00001252, -0.00000090, -0.00001380, -0.00001312, -0.00000009, 0.00001245, 0.00001184, -0.00000334, -0.00002058, -0.00002371, -0.00000992, 0.00000581, 0.00000808, 0.00000022, -0.00000227, 0.00000389, 0.00000467, -0.00000707, -0.00001590, -0.00000573, 0.00001172, 0.00000987, -0.00001185, -0.00002231, -0.00000306, 0.00002129, 0.00001607, -0.00001228, -0.00002408, -0.00000364, 0.00001891, 0.00001317, -0.00000809, -0.00001196, 0.00000390, 0.00001157, -0.00000034, -0.00000992, -0.00000034, 0.00001157, 0.00000390, -0.00001196, -0.00000809, 0.00001317, 0.00001891, -0.00000364, -0.00002408, -0.00001228, 0.00001607, 0.00002129, -0.00000306, -0.00002231, -0.00001185, 0.00000987, 0.00001172, -0.00000573, -0.00001590, -0.00000707, 0.00000467, 0.00000389, -0.00000227, 0.00000022, 0.00000808, 0.00000581, -0.00000992, -0.00002371, -0.00002058, -0.00000334, 0.00001184, 0.00001245, -0.00000009, -0.00001312, -0.00001380, -0.00000090, 0.00001252, 0.00001155, -0.00000338, -0.00001522, -0.00000952, 0.00000823, 0.00001934, 0.00001398, 0.00000070, -0.00000643, -0.00000367, 0.00000231, 0.00000538, 0.00000540, 0.00000437, 0.00000232, -0.00000126, -0.00000516, -0.00000738, -0.00000710, -0.00000492, -0.00000195, 0.00000045, 0.00000134, 0.00000115, 0.00000126, 0.00000214, },
		{ 0.00000088, 0.00000056, 0.00000067, -0.00000015, -0.00000238, -0.00000512, -0.00000701, -0.00000697, -0.00000448, -0.00000029, 0.00000367, 0.00000581, 0.00000576, 0.00000340, -0.00000123, -0.00000533, -0.00000310, 0.00000741, 0.00001784, 0.00001566, -0.00000014, -0.00001429, -0.00001154, 0.00000436, 0.00001428, 0.00000653, -0.00000968, -0.00001609, -0.00000627, 0.00000898, 0.00001428, 0.00000406, -0.00001351, -0.00002371, -0.00001791, -0.00000235, 0.00000748, 0.00000443, -0.00000185, 0.00000049, 0.00000660, 0.00000172, -0.00001307, -0.00001702, -0.00000008, 0.00001658, 0.00000726, -0.00001814, -0.00002328, 0.00000260, 0.00002533, 0.00001258, -0.00001856, -0.00002479, 0.00000159, 0.00002286, 0.00001176, -0.00001104, -0.00001175, 0.00000604, 0.00001172, -0.00000209, -0.00001032, 0.00000115, 0.00001211, 0.00000260, -0.00001247, -0.00000664, 0.00001353, 0.00001593, -0.00000666, -0.00002258, -0.00000778, 0.00001736, 0.00001762, -0.00000604, -0.00001992, -0.00000758, 0.00000939, 0.00000672, -0.00000820, -0.00001224, -0.00000223, 0.00000460, 0.00000046, -0.00000263, 0.00000413, 0.00000984, 0.00000044, -0.00001805, -0.00002523, -0.00001240, 0.00000708, 0.00001443, 0.00000518, -0.00000931, -0.00001485, -0.00000656, 0.00000769, 0.00001382, 0.00000488, -0.00001021, -0.00001441, -0.00000164, 0.00001518, 0.00001877, 0.00000736, -0.00000506, -0.00000671, -0.00000003, 0.00000542, 0.00000568, 0.00000394, 0.00000296, 0.00000151, -0.00000192, -0.00000588, -0.00000786, -0.00000713, -0.00000454, -0.00000134, 0.00000119, 0.00000207, 0.00000169, 0.00000153, 0.00000219, 0.00000263, 0.00000195, },
		{ 0.00000008, -0.00000058, -0.00000262, -0.00000523, -0.00000696, -0.00000659, -0.00000359, 0.00000113, 0.00000526, 0.00000660, 0.00000464, 0.00000040, -0.00000383, -0.00000425, 0.00000231, 0.00001302, 0.00001783, 0.00000898, -0.00000743, -0.00001481, -0.00000501, 0.00001022, 0.00001220, -0.00000190, -0.00001548, -0.00001279, 0.00000293, 0.00001455, 0.00001021, -0.00000606, -0.00002034, -0.00002162, -0.00001035, 0.00000250, 0.00000597, 0.00000097, -0.00000129, 0.00000420, 0.00000691, -0.00000382, -0.00001809, -0.00001444, 0.00000739, 0.00001907, 0.00000118, -0.00002456, -0.00002110, 0.00001040, 0.00002773, 0.00000597, -0.00002537, -0.00002291, 0.00000913, 0.00002622, 0.00000813, -0.00001492, -0.00001047, 0.00000936, 0.00001157, -0.00000497, -0.00001104, 0.00000349, 0.00001351, 0.00000099, -0.00001424, -0.00000607, 0.00001444, 0.00001453, -0.00000873, -0.00002190, -0.00000504, 0.00001817, 0.00001559, -0.00000725, -0.00001816, -0.00000596, 0.00000749, 0.00000361, -0.00000759, -0.00000807, 0.00000007, 0.00000245, -0.00000174, -0.00000033, 0.00000796, 0.00000811, -0.00000756, -0.00002388, -0.00002112, -0.00000128, 0.00001421, 0.00001063, -0.00000498, -0.00001503, -0.00001047, 0.00000308, 0.00001284, 0.00000986, -0.00000327, -0.00001348, -0.00000905, 0.00000699, 0.00001852, 0.00001410, 0.00000001, -0.00000811, -0.00000402, 0.00000420, 0.00000696, 0.00000433, 0.00000212, 0.00000215, 0.00000120, -0.00000258, -0.00000677, -0.00000834, -0.00000697, -0.00000397, -0.00000057, 0.00000204, 0.00000283, 0.00000212, 0.00000159, 0.00000206, 0.00000246, 0.00000169, 0.00000043, -0.00000005, },
		{ -0.00000272, -0.00000535, -0.00000696, -0.00000609, -0.00000232, 0.00000289, 0.00000653, 0.00000622, 0.00000226, -0.00000250, -0.00000438, -0.00000064, 0.00000812, 0.00001585, 0.00001411, 0.00000137, -0.00001167, -0.00001144, 0.00000192, 0.00001230, 0.00000624, -0.00000974, -0.00001640, -0.00000532, 0.00001108, 0.00001471, 0.00000181, -0.00001521, -0.00002224, -0.00001587, -0.00000368, 0.00000388, 0.00000290, -0.00000064, 0.00000144, 0.00000701, 0.00000419, -0.00001034, -0.00002023, -0.00000846, 0.00001443, 0.00001773, -0.00000727, -0.00002889, -0.00001519, 0.00001858, 0.00002672, -0.00000331, -0.00003066, -0.00001743, 0.00001780, 0.00002727, 0.00000212, -0.00001837, -0.00000718, 0.00001319, 0.00000993, -0.00000900, -0.00001094, 0.00000735, 0.00001493, -0.00000221, -0.00001725, -0.00000494, 0.00001701, 0.00001403, -0.00001155, -0.00002270, -0.00000277, 0.00002016, 0.00001510, -0.00000853, -0.00001840, -0.00000614, 0.00000634, 0.00000327, -0.00000537, -0.00000539, -0.00000026, 0.00000009, -0.00000145, 0.00000352, 0.00000958, 0.00000266, -0.00001572, -0.00002493, -0.00001201, 0.00000933, 0.00001534, 0.00000151, -0.00001421, -0.00001466, -0.00000129, 0.00001148, 0.00001244, 0.00000207, -0.00000974, -0.00001171, -0.00000076, 0.00001356, 0.00001721, 0.00000685, -0.00000578, -0.00000767, 0.00000054, 0.00000746, 0.00000643, 0.00000198, 0.00000076, 0.00000214, 0.00000109, -0.00000351, -0.00000780, -0.00000866, -0.00000656, -0.00000323, 0.00000032, 0.00000301, 0.00000357, 0.00000233, 0.00000138, 0.00000179, 0.00000228, 0.00000143, -0.00000008, -0.00000068, -0.00000040, -0.00000079, },
		{ -0.00000695, -0.00000527, -0.00000062, 0.00000462, 0.00000694, 0.00000454, -0.00000063, -0.00000415, -0.00000250, 0.00000441, 0.00001252, 0.00001517, 0.00000799, -0.00000498, -0.00001212, -0.00000581, 0.00000689, 0.00001020, -0.00000132, -0.00001436, -0.00001237, 0.00000354, 0.00001533, 0.00000952, -0.00000805, -0.00002074, -0.00001943, -0.00000893, 0.00000030, 0.00000250, 0.00000019, 0.00000013, 0.00000487, 0.00000747, -0.00000092, -0.00001578, -0.00001863, -0.00000064, 0.00001889, 0.00001248, -0.00001596, -0.00002945, -0.00000647, 0.00002489, 0.00002157, -0.00001355, -0.00003251, -0.00000858, 0.00002567, 0.00002489, -0.00000527, -0.00001984, -0.00000188, 0.00001618, 0.00000610, -0.00001323, -0.00000880, 0.00001245, 0.00001491, -0.00000765, -0.00002035, -0.00000168, 0.00002122, 0.00001277, -0.00001639, -0.00002424, 0.00000087, 0.00002397, 0.00001473, -0.00001171, -0.00002060, -0.00000612, 0.00000758, 0.00000503, -0.00000366, -0.00000520, -0.00000189, -0.00000061, 0.00000125, 0.00000682, 0.00000757, -0.00000507, -0.00002107, -0.00002020, -0.00000046, 0.00001579, 0.00000976, -0.00000979, -0.00001876, -0.00000776, 0.00000950, 0.00001526, 0.00000642, -0.00000652, -0.00001179, -0.00000527, 0.00000767, 0.00001591, 0.00001176, -0.00000059, -0.00000831, -0.00000416, 0.00000526, 0.00000863, 0.00000410, -0.00000041, 0.00000040, 0.00000271, 0.00000075, -0.00000482, -0.00000881, -0.00000866, -0.00000591, -0.00000237, 0.00000135, 0.00000408, 0.00000418, 0.00000221, 0.00000089, 0.00000149, 0.00000219, 0.00000116, -0.00000069, -0.00000131, -0.00000069, -0.00000078, -0.00000276, -0.00000556, },
		{ 0.00000128, 0.00000581, 0.00000616, 0.00000197, -0.00000300, -0.00000378, 0.00000137, 0.00000933, 0.00001422, 0.00001138, 0.00000136, -0.00000853, -0.00000933, -0.00000010, 0.00000851, 0.00000505, -0.00000782, -0.00001445, -0.00000507, 0.00001068, 0.00001436, 0.00000083, -0.00001632, -0.00002153, -0.00001359, -0.00000316, 0.00000113, -0.00000007, -0.00000080, 0.00000270, 0.00000735, 0.00000522, -0.00000683, -0.00001849, -0.00001380, 0.00000679, 0.00001950, 0.00000462, -0.00002262, -0.00002587, 0.00000310, 0.00002752, 0.00001298, -0.00002247, -0.00002995, 0.00000213, 0.00003081, 0.00001916, -0.00001233, -0.00001841, 0.00000442, 0.00001698, 0.00000036, -0.00001617, -0.00000409, 0.00001752, 0.00001208, -0.00001484, -0.00002169, 0.00000455, 0.00002574, 0.00000891, -0.00002326, -0.00002468, 0.00000726, 0.00002875, 0.00001235, -0.00001759, -0.00002323, -0.00000380, 0.00001141, 0.00000702, -0.00000419, -0.00000699, -0.00000280, 0.00000127, 0.00000486, 0.00000747, 0.00000211, -0.00001236, -0.00002132, -0.00001078, 0.00000980, 0.00001569, -0.00000086, -0.00001909, -0.00001635, 0.00000368, 0.00001784, 0.00001259, -0.00000315, -0.00001239, -0.00000821, 0.00000372, 0.00001324, 0.00001326, 0.00000413, -0.00000582, -0.00000705, 0.00000094, 0.00000847, 0.00000739, 0.00000091, -0.00000192, 0.00000110, 0.00000339, -0.00000011, -0.00000639, -0.00000957, -0.00000831, -0.00000510, -0.00000141, 0.00000254, 0.00000518, 0.00000450, 0.00000166, 0.00000022, 0.00000131, 0.00000224, 0.00000083, -0.00000143, -0.00000188, -0.00000071, -0.00000058, -0.00000284, -0.00000585, -0.00000676, -0.00000402, },
		{ 0.00000428, -0.00000070, -0.00000391, -0.00000129, 0.00000609, 0.00001261, 0.00001282, 0.00000576, -0.00000394, -0.00000886, -0.00000484, 0.00000380, 0.00000673, -0.00000102, -0.00001120, -0.00001040, 0.00000282, 0.00001381, 0.00000867, -0.00000859, -0.00002084, -0.00001811, -0.00000706, -0.00000001, -0.00000037, -0.00000217, 0.00000027, 0.00000572, 0.00000776, 0.00000101, -0.00001169, -0.00001792, -0.00000734, 0.00001188, 0.00001639, -0.00000371, -0.00002575, -0.00001917, 0.00001124, 0.00002584, 0.00000276, -0.00002809, -0.00002335, 0.00001257, 0.00003208, 0.00001126, -0.00001736, -0.00001418, 0.00001011, 0.00001487, -0.00000608, -0.00001646, 0.00000265, 0.00002074, 0.00000590, -0.00002215, -0.00001954, 0.00001324, 0.00002841, 0.00000141, -0.00003071, -0.00002191, 0.00001645, 0.00003236, 0.00000626, -0.00002534, -0.00002386, 0.00000200, 0.00001647, 0.00000698, -0.00000738, -0.00000900, -0.00000129, 0.00000512, 0.00000711, 0.00000443, -0.00000495, -0.00001626, -0.00001593, 0.00000033, 0.00001521, 0.00000883, -0.00001286, -0.00002268, -0.00000713, 0.00001586, 0.00002033, 0.00000395, -0.00001267, -0.00001282, 0.00000013, 0.00001201, 0.00001384, 0.00000660, -0.00000296, -0.00000724, -0.00000281, 0.00000569, 0.00000915, 0.00000434, -0.00000199, -0.00000207, 0.00000247, 0.00000365, -0.00000158, -0.00000796, -0.00000990, -0.00000766, -0.00000420, -0.00000032, 0.00000390, 0.00000616, 0.00000435, 0.00000071, -0.00000044, 0.00000140, 0.00000240, 0.00000032, -0.00000231, -0.00000227, -0.00000042, -0.00000028, -0.00000306, -0.00000617, -0.00000624, -0.00000240, 0.00000300, 0.00000604, },
		{ -0.00000288, 0.00000269, 0.00001024, 0.00001332, 0.00000888, -0.00000001, -0.00000678, -0.00000661, -0.00000043, 0.00000497, 0.00000277, -0.00000581, -0.00001069, -0.00000398, 0.00000861, 0.00001214, 0.00000025, -0.00001603, -0.00002090, -0.00001221, -0.00000190, 0.00000031, -0.00000277, -0.00000271, 0.00000273, 0.00000782, 0.00000597, -0.00000370, -0.00001423, -0.00001470, -0.00000124, 0.00001370, 0.00001089, -0.00001047, -0.00002510, -0.00001127, 0.00001625, 0.00002055, -0.00000681, -0.00002947, -0.00001422, 0.00002079, 0.00002958, 0.00000300, -0.00001939, -0.00000829, 0.00001372, 0.00001022, -0.00001150, -0.00001356, 0.00000986, 0.00002059, -0.00000287, -0.00002735, -0.00001318, 0.00002252, 0.00002714, -0.00000930, -0.00003631, -0.00001466, 0.00002685, 0.00003229, -0.00000384, -0.00003264, -0.00002032, 0.00001077, 0.00002028, 0.00000350, -0.00001205, -0.00000900, 0.00000307, 0.00000899, 0.00000605, -0.00000160, -0.00001065, -0.00001484, -0.00000657, 0.00000927, 0.00001379, -0.00000250, -0.00002189, -0.00001870, 0.00000625, 0.00002444, 0.00001552, -0.00000790, -0.00001834, -0.00000696, 0.00001051, 0.00001643, 0.00000911, -0.00000162, -0.00000702, -0.00000469, 0.00000254, 0.00000836, 0.00000726, 0.00000066, -0.00000364, -0.00000097, 0.00000388, 0.00000317, -0.00000343, -0.00000923, -0.00000978, -0.00000684, -0.00000323, 0.00000097, 0.00000537, 0.00000681, 0.00000362, -0.00000052, -0.00000085, 0.00000183, 0.00000252, -0.00000048, -0.00000324, -0.00000234, 0.00000018, 0.00000001, -0.00000344, -0.00000639, -0.00000533, -0.00000063, 0.00000409, 0.00000515, 0.00000181, -0.00000255, },
		{ 0.00001263, 0.00001135, 0.00000362, -0.00000448, -0.00000699, -0.00000310, 0.00000248, 0.00000363, -0.00000157, -0.00000784, -0.00000698, 0.00000237, 0.00001058, 0.00000651, -0.00000833, -0.00001977, -0.00001724, -0.00000598, 0.00000077, -0.00000138, -0.00000465, -0.00000154, 0.00000545, 0.00000823, 0.00000277, -0.00000741, -0.00001420, -0.00001033, 0.00000295, 0.00001253, 0.00000491, -0.00001443, -0.00002168, -0.00000420, 0.00001762, 0.00001337, -0.00001393, -0.00002694, -0.00000459, 0.00002566, 0.00002442, -0.00000394, -0.00001849, -0.00000235, 0.00001449, 0.00000430, -0.00001447, -0.00000806, 0.00001558, 0.00001654, -0.00001234, -0.00002856, -0.00000336, 0.00002983, 0.00002091, -0.00002124, -0.00003760, -0.00000325, 0.00003571, 0.00002682, -0.00001642, -0.00003656, -0.00001181, 0.00002036, 0.00002044, -0.00000326, -0.00001584, -0.00000548, 0.00000905, 0.00001039, 0.00000111, -0.00000826, -0.00001217, -0.00000820, 0.00000339, 0.00001276, 0.00000601, -0.00001427, -0.00002447, -0.00000794, 0.00001952, 0.00002614, 0.00000449, -0.00001890, -0.00001751, 0.00000380, 0.00001929, 0.00001504, 0.00000053, -0.00000839, -0.00000668, 0.00000056, 0.00000684, 0.00000812, 0.00000371, -0.00000233, -0.00000368, 0.00000081, 0.00000472, 0.00000195, -0.00000532, -0.00001003, -0.00000932, -0.00000594, -0.00000213, 0.00000250, 0.00000677, 0.00000691, 0.00000233, -0.00000175, -0.00000081, 0.00000252, 0.00000239, -0.00000160, -0.00000406, -0.00000197, 0.00000104, 0.00000015, -0.00000398, -0.00000639, -0.00000410, 0.00000091, 0.00000426, 0.00000337, -0.00000045, -0.00000285, -0.00000006, 0.00000702, },
		{ -0.00000155, -0.00000687, -0.00000519, 0.00000028, 0.00000338, 0.00000090, -0.00000460, -0.00000682, -0.00000182, 0.00000649, 0.00000839, -0.00000115, -0.00001477, -0.00001944, -0.00001154, -0.00000116, 0.00000088, -0.00000379, -0.00000519, 0.00000071, 0.00000737, 0.00000700, -0.00000059, -0.00000929, -0.00001235, -0.00000643, 0.00000467, 0.00000965, 0.00000010, -0.00001555, -0.00001718, 0.00000066, 0.00001604, 0.00000623, -0.00001779, -0.00002184, 0.00000378, 0.00002722, 0.00001823, -0.00000861, -0.00001567, 0.00000223, 0.00001274, -0.00000125, -0.00001442, -0.00000153, 0.00001823, 0.00000937, -0.00002015, -0.00002503, 0.00000781, 0.00003287, 0.00001035, -0.00003157, -0.00003327, 0.00001026, 0.00004009, 0.00001592, -0.00002851, -0.00003477, 0.00000056, 0.00002772, 0.00001572, -0.00001138, -0.00001627, 0.00000148, 0.00001414, 0.00000757, -0.00000635, -0.00001236, -0.00000811, 0.00000138, 0.00001009, 0.00000945, -0.00000527, -0.00002214, -0.00001932, 0.00000636, 0.00002824, 0.00001995, -0.00000945, -0.00002504, -0.00000971, 0.00001618, 0.00002315, 0.00000786, -0.00000895, -0.00001117, -0.00000218, 0.00000627, 0.00000839, 0.00000527, -0.00000004, -0.00000373, -0.00000243, 0.00000250, 0.00000471, 0.00000030, -0.00000690, -0.00001036, -0.00000869, -0.00000498, -0.00000079, 0.00000421, 0.00000783, 0.00000629, 0.00000065, -0.00000266, -0.00000023, 0.00000326, 0.00000183, -0.00000296, -0.00000454, -0.00000109, 0.00000199, 0.00000002, -0.00000461, -0.00000612, -0.00000275, 0.00000188, 0.00000347, 0.00000126, -0.00000171, -0.00000136, 0.00000381, 0.00001049, 0.00001263, 0.00000733, },
		{ -0.00000245, 0.00000260, 0.00000265, -0.00000178, -0.00000541, -0.00000362, 0.00000276, 0.00000719, 0.00000306, -0.00000853, -0.00001776, -0.00001589, -0.00000570, 0.00000136, -0.00000079, -0.00000577, -0.00000434, 0.00000314, 0.00000801, 0.00000490, -0.00000305, -0.00000939, -0.00000999, -0.00000409, 0.00000439, 0.00000661, -0.00000274, -0.00001479, -0.00001318, 0.00000304, 0.00001293, 0.00000060, -0.00001869, -0.00001584, 0.00000997, 0.00002638, 0.00001244, -0.00001105, -0.00001230, 0.00000479, 0.00000959, -0.00000508, -0.00001189, 0.00000419, 0.00001728, 0.00000096, -0.00002439, -0.00001755, 0.00001764, 0.00003058, -0.00000236, -0.00003760, -0.00002375, 0.00002269, 0.00003815, 0.00000156, -0.00003671, -0.00002667, 0.00001397, 0.00003019, 0.00000683, -0.00001790, -0.00001197, 0.00000999, 0.00001563, 0.00000047, -0.00001342, -0.00001125, 0.00000072, 0.00001015, 0.00001069, 0.00000056, -0.00001576, -0.00002312, -0.00000791, 0.00001966, 0.00002946, 0.00000754, -0.00002187, -0.00002386, 0.00000315, 0.00002620, 0.00002034, -0.00000310, -0.00001604, -0.00000884, 0.00000471, 0.00001005, 0.00000661, 0.00000110, -0.00000258, -0.00000338, -0.00000068, 0.00000347, 0.00000398, -0.00000135, -0.00000804, -0.00001040, -0.00000800, -0.00000384, 0.00000086, 0.00000588, 0.00000826, 0.00000497, -0.00000108, -0.00000297, 0.00000079, 0.00000374, 0.00000074, -0.00000435, -0.00000446, 0.00000025, 0.00000283, -0.00000044, -0.00000524, -0.00000561, -0.00000156, 0.00000209, 0.00000201, -0.00000046, -0.00000147, 0.00000158, 0.00000766, 0.00001216, 0.00001034, 0.00000239, -0.00000556, -0.00000717, },
		{ 0.00000123, -0.00000331, -0.00000411, 0.00000019, 0.00000514, 0.00000452, -0.00000360, -0.00001378, -0.00001718, -0.00001066, -0.00000117, 0.00000160, -0.00000315, -0.00000670, -0.00000259, 0.00000503, 0.00000760, 0.00000289, -0.00000415, -0.00000851, -0.00000834, -0.00000346, 0.00000331, 0.00000460, -0.00000384, -0.00001353, -0.00001060, 0.00000362, 0.00000980, -0.00000304, -0.00001774, -0.00001028, 0.00001401, 0.00002438, 0.00000775, -0.00001204, -0.00000948, 0.00000560, 0.00000642, -0.00000674, -0.00000823, 0.00000768, 0.00001346, -0.00000650, -0.00002435, -0.00000816, 0.00002383, 0.00002358, -0.00001428, -0.00003795, -0.00001128, 0.00003091, 0.00002998, -0.00001290, -0.00003855, -0.00001373, 0.00002493, 0.00002666, -0.00000373, -0.00002010, -0.00000347, 0.00001700, 0.00001189, -0.00000890, -0.00001664, -0.00000412, 0.00001133, 0.00001432, 0.00000464, -0.00001045, -0.00002139, -0.00001669, 0.00000603, 0.00002770, 0.00002272, -0.00000737, -0.00002878, -0.00001525, 0.00001739, 0.00003050, 0.00001122, -0.00001448, -0.00001816, -0.00000205, 0.00001134, 0.00001026, 0.00000220, -0.00000265, -0.00000310, -0.00000187, 0.00000068, 0.00000357, 0.00000297, -0.00000266, -0.00000885, -0.00001037, -0.00000719, -0.00000235, 0.00000274, 0.00000718, 0.00000789, 0.00000319, -0.00000245, -0.00000259, 0.00000196, 0.00000369, -0.00000078, -0.00000542, -0.00000368, 0.00000189, 0.00000335, -0.00000127, -0.00000582, -0.00000499, -0.00000076, 0.00000157, 0.00000042, -0.00000113, 0.00000034, 0.00000519, 0.00001045, 0.00001167, 0.00000643, -0.00000236, -0.00000788, -0.00000594, 0.00000028, 0.00000369, },
		{ -0.00000166, 0.00000303, 0.00000467, -0.00000043, -0.00000967, -0.00001597, -0.00001388, -0.00000534, 0.00000134, 0.00000020, -0.00000531, -0.00000653, -0.00000061, 0.00000613, 0.00000676, 0.00000165, -0.00000423, -0.00000777, -0.00000800, -0.00000385, 0.00000262, 0.00000393, -0.00000413, -0.00001287, -0.00000937, 0.00000358, 0.00000755, -0.00000508, -0.00001614, -0.00000571, 0.00001662, 0.00002220, 0.00000406, -0.00001258, -0.00000763, 0.00000556, 0.00000420, -0.00000674, -0.00000499, 0.00000855, 0.00000842, -0.00001133, -0.00002076, 0.00000063, 0.00002538, 0.00001397, -0.00002279, -0.00003308, 0.00000091, 0.00003303, 0.00001768, -0.00002385, -0.00003348, 0.00000085, 0.00003059, 0.00001816, -0.00001257, -0.00001666, 0.00000680, 0.00001955, 0.00000333, -0.00001706, -0.00001357, 0.00000745, 0.00001959, 0.00001162, -0.00000599, -0.00001915, -0.00001985, -0.00000506, 0.00001790, 0.00002807, 0.00001021, -0.00002021, -0.00002801, -0.00000154, 0.00002883, 0.00002747, -0.00000174, -0.00002272, -0.00001441, 0.00000696, 0.00001513, 0.00000659, -0.00000312, -0.00000446, -0.00000172, -0.00000024, 0.00000107, 0.00000312, 0.00000219, -0.00000362, -0.00000962, -0.00001034, -0.00000605, -0.00000042, 0.00000457, 0.00000776, 0.00000675, 0.00000136, -0.00000315, -0.00000167, 0.00000289, 0.00000298, -0.00000245, -0.00000587, -0.00000220, 0.00000352, 0.00000335, -0.00000240, -0.00000630, -0.00000439, -0.00000045, 0.00000063, -0.00000068, -0.00000043, 0.00000328, 0.00000850, 0.00001150, 0.00000925, 0.00000186, -0.00000615, -0.00000861, -0.00000392, 0.00000250, 0.00000396, 0.00000009, -0.00000331, },
		{ 0.00000170, -0.00000608, -0.00001375, -0.00001514, -0.00000918, -0.00000115, 0.00000184, -0.00000198, -0.00000675, -0.00000558, 0.00000120, 0.00000671, 0.00000614, 0.00000124, -0.00000407, -0.00000795, -0.00000874, -0.00000421, 0.00000300, 0.00000411, -0.00000471, -0.00001320, -0.00000873, 0.00000392, 0.00000627, -0.00000642, -0.00001469, -0.00000192, 0.00001865, 0.00002015, 0.00000076, -0.00001332, -0.00000637, 0.00000566, 0.00000313, -0.00000621, -0.00000317, 0.00000752, 0.00000397, -0.00001302, -0.00001543, 0.00000684, 0.00002289, 0.00000443, -0.00002662, -0.00002510, 0.00000981, 0.00002918, 0.00000464, -0.00002883, -0.00002314, 0.00001339, 0.00002991, 0.00000755, -0.00001676, -0.00000844, 0.00001538, 0.00001614, -0.00000745, -0.00002033, -0.00000391, 0.00001979, 0.00002183, 0.00000236, -0.00001715, -0.00002188, -0.00001155, 0.00000763, 0.00002389, 0.00002100, -0.00000404, -0.00002728, -0.00001987, 0.00001337, 0.00003419, 0.00001782, -0.00001491, -0.00002527, -0.00000582, 0.00001517, 0.00001447, 0.00000016, -0.00000731, -0.00000367, 0.00000064, 0.00000048, 0.00000052, 0.00000275, 0.00000192, -0.00000449, -0.00001059, -0.00001013, -0.00000432, 0.00000183, 0.00000592, 0.00000743, 0.00000515, -0.00000005, -0.00000309, -0.00000057, 0.00000324, 0.00000173, -0.00000387, -0.00000549, -0.00000025, 0.00000482, 0.00000272, -0.00000377, -0.00000669, -0.00000393, -0.00000055, -0.00000027, -0.00000082, 0.00000155, 0.00000657, 0.00001067, 0.00001064, 0.00000557, -0.00000253, -0.00000869, -0.00000817, -0.00000164, 0.00000429, 0.00000410, -0.00000023, -0.00000221, 0.00000067, 0.00000389, },
		{ -0.00001524, -0.00001207, -0.00000438, 0.00000135, 0.00000078, -0.00000426, -0.00000738, -0.00000420, 0.00000276, 0.00000720, 0.00000604, 0.00000120, -0.00000452, -0.00000922, -0.00000976, -0.00000366, 0.00000438, 0.00000419, -0.00000623, -0.00001412, -0.00000771, 0.00000503, 0.00000539, -0.00000788, -0.00001351, 0.00000169, 0.00002054, 0.00001792, -0.00000279, -0.00001426, -0.00000496, 0.00000643, 0.00000260, -0.00000625, -0.00000284, 0.00000599, 0.00000132, -0.00001233, -0.00001043, 0.00000975, 0.00001822, -0.00000280, -0.00002626, -0.00001696, 0.00001376, 0.00002140, -0.00000572, -0.00002740, -0.00001074, 0.00002110, 0.00002410, -0.00000161, -0.00001511, 0.00000181, 0.00001918, 0.00000754, -0.00001655, -0.00001643, 0.00001003, 0.00002844, 0.00001637, -0.00001052, -0.00002433, -0.00001718, 0.00000049, 0.00001690, 0.00002237, 0.00000924, -0.00001564, -0.00002688, -0.00000698, 0.00002535, 0.00003211, 0.00000424, -0.00002463, -0.00002139, 0.00000505, 0.00001982, 0.00000929, -0.00000691, -0.00000874, -0.00000067, 0.00000263, -0.00000023, -0.00000038, 0.00000308, 0.00000202, -0.00000574, -0.00001179, -0.00000938, -0.00000187, 0.00000401, 0.00000635, 0.00000627, 0.00000362, -0.00000070, -0.00000252, 0.00000023, 0.00000290, 0.00000029, -0.00000463, -0.00000430, 0.00000180, 0.00000549, 0.00000149, -0.00000523, -0.00000698, -0.00000362, -0.00000080, -0.00000067, 0.00000019, 0.00000435, 0.00000940, 0.00001124, 0.00000820, 0.00000139, -0.00000621, -0.00001008, -0.00000695, 0.00000077, 0.00000601, 0.00000458, 0.00000015, -0.00000093, 0.00000181, 0.00000257, -0.00000272, -0.00001095, },
		{ -0.00000044, 0.00000207, -0.00000123, -0.00000618, -0.00000724, -0.00000256, 0.00000427, 0.00000790, 0.00000625, 0.00000081, -0.00000599, -0.00001110, -0.00001014, -0.00000192, 0.00000604, 0.00000333, -0.00000865, -0.00001476, -0.00000570, 0.00000660, 0.00000409, -0.00000975, -0.00001211, 0.00000566, 0.00002217, 0.00001492, -0.00000678, -0.00001484, -0.00000273, 0.00000766, 0.00000164, -0.00000737, -0.00000319, 0.00000529, 0.00000059, -0.00001081, -0.00000721, 0.00001003, 0.00001357, -0.00000684, -0.00002364, -0.00001120, 0.00001301, 0.00001278, -0.00001122, -0.00002119, 0.00000022, 0.00002310, 0.00001607, -0.00000646, -0.00000865, 0.00001046, 0.00001684, -0.00000327, -0.00002018, -0.00000555, 0.00002402, 0.00002987, 0.00000434, -0.00002242, -0.00002438, -0.00000649, 0.00001178, 0.00001959, 0.00001443, -0.00000302, -0.00002141, -0.00001982, 0.00000674, 0.00003140, 0.00002354, -0.00000954, -0.00002844, -0.00001233, 0.00001493, 0.00001931, 0.00000101, -0.00001229, -0.00000683, 0.00000326, 0.00000303, -0.00000217, -0.00000074, 0.00000437, 0.00000198, -0.00000772, -0.00001291, -0.00000766, 0.00000114, 0.00000552, 0.00000564, 0.00000473, 0.00000273, -0.00000056, -0.00000191, 0.00000036, 0.00000204, -0.00000083, -0.00000452, -0.00000257, 0.00000351, 0.00000537, -0.00000022, -0.00000666, -0.00000716, -0.00000335, -0.00000084, -0.00000026, 0.00000217, 0.00000727, 0.00001111, 0.00001018, 0.00000475, -0.00000270, -0.00000906, -0.00001054, -0.00000512, 0.00000354, 0.00000804, 0.00000539, 0.00000059, -0.00000038, 0.00000133, -0.00000035, -0.00000732, -0.00001408, -0.00001427, -0.00000780, },
		{ -0.00000367, -0.00000748, -0.00000639, -0.00000063, 0.00000589, 0.00000877, 0.00000625, -0.00000044, -0.00000832, -0.00001277, -0.00000931, 0.00000062, 0.00000708, 0.00000123, -0.00001129, -0.00001431, -0.00000271, 0.00000785, 0.00000184, -0.00001174, -0.00000988, 0.00001004, 0.00002296, 0.00001077, -0.00001080, -0.00001430, 0.00000040, 0.00000856, -0.00000049, -0.00000924, -0.00000298, 0.00000607, 0.00000085, -0.00001005, -0.00000600, 0.00000932, 0.00001055, -0.00000836, -0.00002120, -0.00000897, 0.00000948, 0.00000622, -0.00001168, -0.00001318, 0.00000730, 0.00002070, 0.00000926, -0.00000601, -0.00000026, 0.00001450, 0.00000943, -0.00001216, -0.00001627, 0.00000947, 0.00003341, 0.00002295, -0.00001053, -0.00002886, -0.00001674, 0.00000638, 0.00001808, 0.00001506, 0.00000343, -0.00001178, -0.00002036, -0.00000897, 0.00001749, 0.00003056, 0.00001125, -0.00002001, -0.00002574, -0.00000084, 0.00002096, 0.00001372, -0.00000792, -0.00001418, -0.00000218, 0.00000642, 0.00000130, -0.00000444, 0.00000026, 0.00000632, 0.00000108, -0.00001046, -0.00001333, -0.00000473, 0.00000416, 0.00000578, 0.00000393, 0.00000349, 0.00000282, -0.00000002, -0.00000186, -0.00000024, 0.00000118, -0.00000117, -0.00000361, -0.00000080, 0.00000453, 0.00000448, -0.00000219, -0.00000794, -0.00000717, -0.00000292, -0.00000037, 0.00000101, 0.00000464, 0.00000963, 0.00001141, 0.00000781, 0.00000085, -0.00000637, -0.00001111, -0.00001015, -0.00000249, 0.00000692, 0.00001033, 0.00000601, 0.00000030, -0.00000108, -0.00000053, -0.00000394, -0.00001108, -0.00001514, -0.00001147, -0.00000333, 0.00000212, 0.00000125, },
		{ -0.00000480, 0.00000160, 0.00000757, 0.00000943, 0.00000556, -0.00000262, -0.00001089, -0.00001352, -0.00000731, 0.00000318, 0.00000688, -0.00000179, -0.00001327, -0.00001240, 0.00000065, 0.00000801, -0.00000132, -0.00001313, -0.00000645, 0.00001433, 0.00002226, 0.00000560, -0.00001406, -0.00001222, 0.00000386, 0.00000820, -0.00000386, -0.00001083, -0.00000128, 0.00000798, 0.00000071, -0.00001085, -0.00000588, 0.00000933, 0.00000945, -0.00000916, -0.00002079, -0.00000972, 0.00000583, 0.00000331, -0.00000876, -0.00000638, 0.00001001, 0.00001671, 0.00000614, -0.00000148, 0.00000653, 0.00001285, 0.00000010, -0.00001547, -0.00000543, 0.00002408, 0.00003489, 0.00000953, -0.00002327, -0.00002719, -0.00000384, 0.00001663, 0.00001700, 0.00000537, -0.00000642, -0.00001462, -0.00001395, 0.00000192, 0.00002284, 0.00002402, -0.00000121, -0.00002494, -0.00001790, 0.00000980, 0.00002158, 0.00000464, -0.00001492, -0.00001201, 0.00000360, 0.00000736, -0.00000219, -0.00000580, 0.00000292, 0.00000811, -0.00000126, -0.00001347, -0.00001237, -0.00000076, 0.00000639, 0.00000447, 0.00000184, 0.00000328, 0.00000385, 0.00000024, -0.00000276, -0.00000122, 0.00000094, -0.00000056, -0.00000233, 0.00000050, 0.00000472, 0.00000305, -0.00000419, -0.00000898, -0.00000691, -0.00000209, 0.00000076, 0.00000289, 0.00000701, 0.00001098, 0.00001037, 0.00000459, -0.00000310, -0.00000956, -0.00001237, -0.00000874, 0.00000121, 0.00001091, 0.00001239, 0.00000566, -0.00000123, -0.00000293, -0.00000309, -0.00000728, -0.00001345, -0.00001430, -0.00000758, 0.00000057, 0.00000306, -0.00000077, -0.00000607, -0.00000794, },
		{ 0.00000897, 0.00000946, 0.00000398, -0.00000535, -0.00001300, -0.00001307, -0.00000465, 0.00000497, 0.00000537, -0.00000494, -0.00001388, -0.00000932, 0.00000349, 0.00000670, -0.00000482, -0.00001327, -0.00000201, 0.00001777, 0.00001979, 0.00000008, -0.00001577, -0.00000878, 0.00000667, 0.00000608, -0.00000777, -0.00001101, 0.00000207, 0.00000986, -0.00000100, -0.00001291, -0.00000533, 0.00001097, 0.00000927, -0.00001117, -0.00002290, -0.00001155, 0.00000425, 0.00000377, -0.00000510, -0.00000256, 0.00000984, 0.00001414, 0.00000730, 0.00000427, 0.00000908, 0.00000683, -0.00000705, -0.00001144, 0.00000917, 0.00003363, 0.00002786, -0.00000613, -0.00002951, -0.00001774, 0.00000980, 0.00002034, 0.00000898, -0.00000553, -0.00001157, -0.00001149, -0.00000528, 0.00000962, 0.00002237, 0.00001451, -0.00001073, -0.00002396, -0.00000761, 0.00001683, 0.00001690, -0.00000536, -0.00001809, -0.00000651, 0.00000852, 0.00000544, -0.00000630, -0.00000518, 0.00000684, 0.00000871, -0.00000511, -0.00001589, -0.00000958, 0.00000356, 0.00000699, 0.00000177, 0.00000028, 0.00000453, 0.00000524, -0.00000054, -0.00000459, -0.00000186, 0.00000184, 0.00000076, -0.00000133, 0.00000097, 0.00000421, 0.00000139, -0.00000607, -0.00000972, -0.00000622, -0.00000067, 0.00000248, 0.00000492, 0.00000876, 0.00001117, 0.00000828, 0.00000090, -0.00000694, -0.00001221, -0.00001262, -0.00000591, 0.00000608, 0.00001502, 0.00001337, 0.00000375, -0.00000398, -0.00000533, -0.00000557, -0.00000982, -0.00001432, -0.00001196, -0.00000331, 0.00000338, 0.00000238, -0.00000357, -0.00000795, -0.00000732, -0.00000248, 0.00000397, },
		{ 0.00000177, -0.00000803, -0.00001421, -0.00001164, -0.00000207, 0.00000560, 0.00000302, -0.00000740, -0.00001298, -0.00000579, 0.00000511, 0.00000415, -0.00000787, -0.00001185, 0.00000279, 0.00001965, 0.00001582, -0.00000485, -0.00001558, -0.00000476, 0.00000793, 0.00000240, -0.00001103, -0.00000909, 0.00000632, 0.00001037, -0.00000454, -0.00001501, -0.00000306, 0.00001391, 0.00000824, -0.00001554, -0.00002649, -0.00001223, 0.00000553, 0.00000579, -0.00000305, -0.00000163, 0.00000942, 0.00001486, 0.00001124, 0.00000803, 0.00000674, -0.00000032, -0.00000871, -0.00000108, 0.00002290, 0.00003520, 0.00001469, -0.00001890, -0.00002720, -0.00000376, 0.00001923, 0.00001597, -0.00000295, -0.00001338, -0.00001054, -0.00000455, 0.00000216, 0.00001247, 0.00001761, 0.00000519, -0.00001553, -0.00001846, 0.00000209, 0.00001878, 0.00000854, -0.00001362, -0.00001677, 0.00000058, 0.00001089, 0.00000106, -0.00000947, -0.00000212, 0.00001096, 0.00000724, -0.00000995, -0.00001672, -0.00000506, 0.00000720, 0.00000546, -0.00000156, 0.00000023, 0.00000713, 0.00000603, -0.00000286, -0.00000679, -0.00000130, 0.00000400, 0.00000212, -0.00000125, 0.00000058, 0.00000339, -0.00000022, -0.00000774, -0.00001007, -0.00000492, 0.00000139, 0.00000449, 0.00000660, 0.00000962, 0.00001033, 0.00000547, -0.00000305, -0.00001061, -0.00001409, -0.00001140, -0.00000141, 0.00001173, 0.00001832, 0.00001248, 0.00000019, -0.00000737, -0.00000756, -0.00000747, -0.00001137, -0.00001381, -0.00000858, 0.00000073, 0.00000470, 0.00000021, -0.00000659, -0.00000878, -0.00000545, 0.00000037, 0.00000606, 0.00000967, 0.00000865, },
		{ -0.00001453, -0.00000978, -0.00000009, 0.00000520, 0.00000056, -0.00000868, -0.00001092, -0.00000262, 0.00000524, 0.00000095, -0.00000980, -0.00000912, 0.00000714, 0.00001972, 0.00001109, -0.00000841, -0.00001380, -0.00000121, 0.00000724, -0.00000192, -0.00001255, -0.00000523, 0.00001013, 0.00000865, -0.00000917, -0.00001564, 0.00000128, 0.00001677, 0.00000480, -0.00002205, -0.00002958, -0.00001028, 0.00000880, 0.00000704, -0.00000355, -0.00000179, 0.00001122, 0.00001883, 0.00001512, 0.00000771, 0.00000118, -0.00000477, -0.00000362, 0.00001224, 0.00003142, 0.00002870, -0.00000007, -0.00002477, -0.00001743, 0.00000974, 0.00002094, 0.00000496, -0.00001430, -0.00001522, -0.00000428, 0.00000280, 0.00000582, 0.00001080, 0.00001119, -0.00000146, -0.00001559, -0.00001090, 0.00000875, 0.00001579, -0.00000104, -0.00001819, -0.00001167, 0.00000710, 0.00000991, -0.00000452, -0.00001035, 0.00000300, 0.00001394, 0.00000343, -0.00001469, -0.00001520, 0.00000049, 0.00000901, 0.00000191, -0.00000433, 0.00000230, 0.00001033, 0.00000517, -0.00000666, -0.00000831, 0.00000112, 0.00000694, 0.00000258, -0.00000243, -0.00000030, 0.00000271, -0.00000166, -0.00000924, -0.00000992, -0.00000286, 0.00000395, 0.00000632, 0.00000752, 0.00000955, 0.00000874, 0.00000214, -0.00000723, -0.00001392, -0.00001469, -0.00000821, 0.00000461, 0.00001725, 0.00001976, 0.00000935, -0.00000450, -0.00001056, -0.00000905, -0.00000862, -0.00001198, -0.00001212, -0.00000461, 0.00000393, 0.00000424, -0.00000310, -0.00000910, -0.00000807, -0.00000244, 0.00000319, 0.00000732, 0.00000944, 0.00000723, -0.00000058, -0.00001021, },
		{ 0.00000116, 0.00000422, -0.00000143, -0.00000875, -0.00000835, -0.00000040, 0.00000412, -0.00000215, -0.00001037, -0.00000569, 0.00001038, 0.00001824, 0.00000650, -0.00001031, -0.00001120, 0.00000104, 0.00000494, -0.00000570, -0.00001182, -0.00000043, 0.00001221, 0.00000477, -0.00001343, -0.00001381, 0.00000687, 0.00001775, -0.00000161, -0.00002918, -0.00003028, -0.00000569, 0.00001203, 0.00000583, -0.00000580, -0.00000052, 0.00001639, 0.00002415, 0.00001618, 0.00000329, -0.00000433, -0.00000374, 0.00000679, 0.00002406, 0.00003235, 0.00001690, -0.00001150, -0.00002233, -0.00000405, 0.00001788, 0.00001423, -0.00000880, -0.00002087, -0.00001048, 0.00000427, 0.00000734, 0.00000496, 0.00000655, 0.00000580, -0.00000435, -0.00001236, -0.00000388, 0.00001117, 0.00000937, -0.00000933, -0.00001837, -0.00000450, 0.00001121, 0.00000586, -0.00000953, -0.00000821, 0.00000902, 0.00001456, -0.00000227, -0.00001801, -0.00001119, 0.00000583, 0.00000822, -0.00000282, -0.00000527, 0.00000642, 0.00001283, 0.00000196, -0.00001112, -0.00000795, 0.00000546, 0.00000961, 0.00000131, -0.00000470, -0.00000098, 0.00000252, -0.00000302, -0.00001061, -0.00000909, 0.00000003, 0.00000665, 0.00000748, 0.00000752, 0.00000878, 0.00000661, -0.00000169, -0.00001155, -0.00001640, -0.00001333, -0.00000288, 0.00001140, 0.00002138, 0.00001861, 0.00000433, -0.00000937, -0.00001283, -0.00000966, -0.00000915, -0.00001169, -0.00000942, -0.00000056, 0.00000573, 0.00000198, -0.00000685, -0.00001025, -0.00000558, 0.00000115, 0.00000520, 0.00000739, 0.00000849, 0.00000569, -0.00000263, -0.00001180, -0.00001433, -0.00000797, },
		{ -0.00000269, -0.00000795, -0.00000589, 0.00000072, 0.00000223, -0.00000463, -0.00000978, -0.00000224, 0.00001230, 0.00001585, 0.00000271, -0.00001082, -0.00000872, 0.00000170, 0.00000192, -0.00000794, -0.00000918, 0.00000399, 0.00001192, -0.00000027, -0.00001590, -0.00000955, 0.00001204, 0.00001557, -0.00001020, -0.00003489, -0.00002771, 0.00000007, 0.00001310, 0.00000200, -0.00000762, 0.00000415, 0.00002418, 0.00002801, 0.00001296, -0.00000316, -0.00000634, 0.00000339, 0.00001903, 0.00003056, 0.00002616, 0.00000433, -0.00001614, -0.00001335, 0.00000775, 0.00001772, 0.00000142, -0.00002046, -0.00002027, -0.00000109, 0.00001145, 0.00000725, 0.00000074, 0.00000232, 0.00000314, -0.00000394, -0.00000800, 0.00000069, 0.00000960, 0.00000168, -0.00001451, -0.00001473, 0.00000263, 0.00001191, -0.00000003, -0.00001240, -0.00000329, 0.00001439, 0.00001225, -0.00000870, -0.00001884, -0.00000532, 0.00000954, 0.00000473, -0.00000737, -0.00000348, 0.00001173, 0.00001324, -0.00000359, -0.00001486, -0.00000480, 0.00001102, 0.00001069, -0.00000207, -0.00000732, -0.00000072, 0.00000284, -0.00000458, -0.00001180, -0.00000731, 0.00000363, 0.00000899, 0.00000759, 0.00000670, 0.00000759, 0.00000400, -0.00000606, -0.00001570, -0.00001728, -0.00000947, 0.00000415, 0.00001765, 0.00002297, 0.00001483, -0.00000164, -0.00001343, -0.00001389, -0.00000960, -0.00000918, -0.00001049, -0.00000595, 0.00000289, 0.00000561, -0.00000175, -0.00001002, -0.00000933, -0.00000160, 0.00000438, 0.00000572, 0.00000635, 0.00000744, 0.00000456, -0.00000428, -0.00001311, -0.00001399, -0.00000641, 0.00000192, 0.00000315, },
		{ -0.00000388, 0.00000090, 0.00000008, -0.00000633, -0.00000844, 0.00000078, 0.00001309, 0.00001320, -0.00000009, -0.00001057, -0.00000699, 0.00000111, -0.00000077, -0.00000833, -0.00000558, 0.00000690, 0.00000951, -0.00000500, -0.00001581, -0.00000397, 0.00001508, 0.00001010, -0.00001917, -0.00003754, -0.00002249, 0.00000478, 0.00001086, -0.00000296, -0.00000653, 0.00001269, 0.00003227, 0.00002800, 0.00000608, -0.00000837, -0.00000267, 0.00001476, 0.00002897, 0.00003007, 0.00001589, -0.00000456, -0.00001337, -0.00000204, 0.00001347, 0.00000939, -0.00001299, -0.00002601, -0.00001282, 0.00000934, 0.00001430, 0.00000269, -0.00000442, 0.00000029, 0.00000350, -0.00000182, -0.00000453, 0.00000208, 0.00000537, -0.00000512, -0.00001591, -0.00000878, 0.00000786, 0.00000930, -0.00000602, -0.00001215, 0.00000335, 0.00001763, 0.00000724, -0.00001438, -0.00001669, 0.00000116, 0.00001051, -0.00000073, -0.00001015, 0.00000117, 0.00001670, 0.00001054, -0.00001054, -0.00001623, 0.00000131, 0.00001636, 0.00000899, -0.00000719, -0.00000915, 0.00000094, 0.00000326, -0.00000666, -0.00001257, -0.00000437, 0.00000761, 0.00001039, 0.00000650, 0.00000536, 0.00000618, 0.00000082, -0.00001092, -0.00001900, -0.00001569, -0.00000306, 0.00001176, 0.00002188, 0.00002148, 0.00000916, -0.00000733, -0.00001609, -0.00001392, -0.00000917, -0.00000864, -0.00000829, -0.00000221, 0.00000495, 0.00000340, -0.00000607, -0.00001149, -0.00000612, 0.00000293, 0.00000614, 0.00000452, 0.00000487, 0.00000702, 0.00000401, -0.00000590, -0.00001456, -0.00001367, -0.00000490, 0.00000254, 0.00000226, -0.00000333, -0.00000672, },
		{ -0.00000207, -0.00000733, -0.00000672, 0.00000325, 0.00001314, 0.00001070, -0.00000209, -0.00001022, -0.00000616, 0.00000001, -0.00000243, -0.00000722, -0.00000221, 0.00000784, 0.00000601, -0.00000818, -0.00001341, 0.00000127, 0.00001496, 0.00000242, -0.00002649, -0.00003670, -0.00001646, 0.00000650, 0.00000578, -0.00000654, -0.00000094, 0.00002372, 0.00003780, 0.00002326, -0.00000200, -0.00000922, 0.00000665, 0.00002683, 0.00003344, 0.00002363, 0.00000584, -0.00000713, -0.00000554, 0.00000658, 0.00001094, -0.00000403, -0.00002404, -0.00002372, -0.00000127, 0.00001695, 0.00001167, -0.00000439, -0.00000793, 0.00000140, 0.00000582, 0.00000015, -0.00000318, 0.00000075, 0.00000026, -0.00000956, -0.00001396, -0.00000240, 0.00001013, 0.00000441, -0.00001043, -0.00000874, 0.00001014, 0.00001784, 0.00000054, -0.00001791, -0.00001192, 0.00000667, 0.00000833, -0.00000673, -0.00000993, 0.00000785, 0.00001958, 0.00000454, -0.00001720, -0.00001395, 0.00000954, 0.00001964, 0.00000398, -0.00001290, -0.00000903, 0.00000391, 0.00000309, -0.00000937, -0.00001246, -0.00000015, 0.00001139, 0.00001032, 0.00000435, 0.00000390, 0.00000460, -0.00000310, -0.00001588, -0.00002045, -0.00001106, 0.00000518, 0.00001829, 0.00002292, 0.00001721, 0.00000293, -0.00001176, -0.00001735, -0.00001342, -0.00000848, -0.00000730, -0.00000524, 0.00000102, 0.00000495, -0.00000047, -0.00000967, -0.00001039, -0.00000117, 0.00000665, 0.00000568, 0.00000209, 0.00000397, 0.00000770, 0.00000366, -0.00000807, -0.00001632, -0.00001311, -0.00000311, 0.00000317, 0.00000147, -0.00000359, -0.00000532, -0.00000236, 0.00000039, },
		{ -0.00000477, 0.00000527, 0.00001277, 0.00000841, -0.00000370, -0.00001008, -0.00000592, -0.00000089, -0.00000290, -0.00000544, 0.00000012, 0.00000718, 0.00000268, -0.00000929, -0.00000984, 0.00000465, 0.00001182, -0.00000560, -0.00003082, -0.00003338, -0.00001192, 0.00000450, -0.00000017, -0.00000630, 0.00000910, 0.00003459, 0.00003867, 0.00001494, -0.00000800, -0.00000412, 0.00001937, 0.00003585, 0.00003152, 0.00001442, -0.00000023, -0.00000359, 0.00000299, 0.00000868, 0.00000114, -0.00001774, -0.00002814, -0.00001469, 0.00001031, 0.00001912, 0.00000452, -0.00001114, -0.00000804, 0.00000514, 0.00000835, 0.00000050, -0.00000408, -0.00000210, -0.00000408, -0.00001113, -0.00000989, 0.00000277, 0.00000935, -0.00000122, -0.00001212, -0.00000301, 0.00001547, 0.00001492, -0.00000633, -0.00001849, -0.00000563, 0.00000981, 0.00000348, -0.00001151, -0.00000628, 0.00001497, 0.00001897, -0.00000390, -0.00002152, -0.00000761, 0.00001810, 0.00001921, -0.00000389, -0.00001749, -0.00000623, 0.00000743, 0.00000154, -0.00001245, -0.00001087, 0.00000514, 0.00001422, 0.00000849, 0.00000154, 0.00000266, 0.00000267, -0.00000779, -0.00002015, -0.00001906, -0.00000343, 0.00001375, 0.00002203, 0.00002046, 0.00001136, -0.00000246, -0.00001454, -0.00001773, -0.00001276, -0.00000730, -0.00000490, -0.00000188, 0.00000276, 0.00000269, -0.00000491, -0.00001119, -0.00000661, 0.00000421, 0.00000822, 0.00000298, -0.00000042, 0.00000461, 0.00000929, 0.00000270, -0.00001122, -0.00001809, -0.00001179, -0.00000089, 0.00000365, 0.00000056, -0.00000358, -0.00000378, -0.00000127, -0.00000067, -0.00000416, -0.00000778, },
		{ 0.00001202, 0.00000618, -0.00000515, -0.00001011, -0.00000579, -0.00000123, -0.00000257, -0.00000380, 0.00000122, 0.00000578, 0.00000042, -0.00000872, -0.00000657, 0.00000546, 0.00000683, -0.00001216, -0.00003207, -0.00002956, -0.00001058, -0.00000033, -0.00000433, -0.00000093, 0.00002180, 0.00004245, 0.00003444, 0.00000573, -0.00000915, 0.00000637, 0.00003201, 0.00003935, 0.00002474, 0.00000635, -0.00000065, 0.00000318, 0.00000768, 0.00000307, -0.00001227, -0.00002697, -0.00002429, -0.00000227, 0.00001808, 0.00001529, -0.00000456, -0.00001498, -0.00000449, 0.00000989, 0.00000940, -0.00000128, -0.00000649, -0.00000505, -0.00000667, -0.00001021, -0.00000519, 0.00000573, 0.00000624, -0.00000602, -0.00001081, 0.00000366, 0.00001822, 0.00000962, -0.00001189, -0.00001610, 0.00000062, 0.00000984, -0.00000277, -0.00001357, 0.00000023, 0.00002056, 0.00001434, -0.00001294, -0.00002180, 0.00000201, 0.00002468, 0.00001421, -0.00001311, -0.00001924, -0.00000081, 0.00001027, -0.00000184, -0.00001518, -0.00000730, 0.00001093, 0.00001525, 0.00000492, -0.00000135, 0.00000180, 0.00000014, -0.00001292, -0.00002264, -0.00001413, 0.00000625, 0.00002067, 0.00002186, 0.00001521, 0.00000562, -0.00000609, -0.00001606, -0.00001794, -0.00001193, -0.00000512, -0.00000157, 0.00000080, 0.00000220, -0.00000125, -0.00000833, -0.00000972, -0.00000098, 0.00000829, 0.00000688, -0.00000102, -0.00000160, 0.00000709, 0.00001092, 0.00000025, -0.00001525, -0.00001913, -0.00000931, 0.00000153, 0.00000354, -0.00000056, -0.00000314, -0.00000201, -0.00000068, -0.00000236, -0.00000618, -0.00000764, -0.00000258, 0.00000694, },
		{ -0.00000642, -0.00001005, -0.00000546, -0.00000111, -0.00000200, -0.00000273, 0.00000147, 0.00000453, -0.00000056, -0.00000749, -0.00000471, 0.00000405, 0.00000166, -0.00001631, -0.00003144, -0.00002738, -0.00001280, -0.00000585, -0.00000438, 0.00000919, 0.00003441, 0.00004545, 0.00002655, -0.00000118, -0.00000432, 0.00001981, 0.00004128, 0.00003707, 0.00001626, 0.00000233, 0.00000352, 0.00000890, 0.00000563, -0.00000841, -0.00002439, -0.00002882, -0.00001432, 0.00000925, 0.00001981, 0.00000707, -0.00001247, -0.00001459, 0.00000154, 0.00001370, 0.00000802, -0.00000474, -0.00000922, -0.00000703, -0.00000728, -0.00000772, -0.00000115, 0.00000628, 0.00000204, -0.00000886, -0.00000700, 0.00000976, 0.00001798, 0.00000321, -0.00001505, -0.00001153, 0.00000535, 0.00000688, -0.00000869, -0.00001213, 0.00000811, 0.00002288, 0.00000632, -0.00002028, -0.00001722, 0.00001298, 0.00002712, 0.00000500, -0.00002144, -0.00001696, 0.00000617, 0.00001102, -0.00000689, -0.00001648, -0.00000159, 0.00001623, 0.00001382, 0.00000004, -0.00000366, 0.00000128, -0.00000314, -0.00001779, -0.00002221, -0.00000574, 0.00001619, 0.00002402, 0.00001765, 0.00000885, 0.00000154, -0.00000792, -0.00001722, -0.00001836, -0.00001038, -0.00000157, 0.00000199, 0.00000160, -0.00000079, -0.00000550, -0.00000920, -0.00000534, 0.00000476, 0.00000956, 0.00000293, -0.00000470, -0.00000039, 0.00001085, 0.00001131, -0.00000409, -0.00001937, -0.00001860, -0.00000572, 0.00000349, 0.00000251, -0.00000163, -0.00000194, -0.00000009, -0.00000090, -0.00000476, -0.00000793, -0.00000670, -0.00000012, 0.00000813, 0.00001077, 0.00000392, },
		{ -0.00000500, -0.00000084, -0.00000155, -0.00000213, 0.00000149, 0.00000387, -0.00000075, -0.00000666, -0.00000456, 0.00000153, -0.00000233, -0.00001825, -0.00003078, -0.00002819, -0.00001751, -0.00000946, 0.00000074, 0.00002219, 0.00004426, 0.00004340, 0.00001768, -0.00000340, 0.00000565, 0.00003297, 0.00004536, 0.00003080, 0.00000941, 0.00000303, 0.00000906, 0.00000976, -0.00000327, -0.00002165, -0.00003101, -0.00002312, -0.00000201, 0.00001629, 0.00001554, -0.00000256, -0.00001676, -0.00001022, 0.00000802, 0.00001506, 0.00000431, -0.00000881, -0.00001120, -0.00000755, -0.00000632, -0.00000476, 0.00000144, 0.00000487, -0.00000199, -0.00000928, -0.00000175, 0.00001410, 0.00001512, -0.00000288, -0.00001547, -0.00000607, 0.00000754, 0.00000187, -0.00001266, -0.00000740, 0.00001541, 0.00002106, -0.00000337, -0.00002385, -0.00000832, 0.00002266, 0.00002409, -0.00000676, -0.00002653, -0.00001054, 0.00001292, 0.00000869, -0.00001270, -0.00001527, 0.00000581, 0.00001979, 0.00000964, -0.00000535, -0.00000489, 0.00000085, -0.00000711, -0.00002135, -0.00001807, 0.00000509, 0.00002413, 0.00002262, 0.00001055, 0.00000345, -0.00000007, -0.00000878, -0.00001895, -0.00001873, -0.00000732, 0.00000321, 0.00000450, -0.00000036, -0.00000537, -0.00000815, -0.00000667, 0.00000070, 0.00000858, 0.00000738, -0.00000225, -0.00000635, 0.00000333, 0.00001453, 0.00000938, -0.00000988, -0.00002238, -0.00001605, -0.00000170, 0.00000421, 0.00000065, -0.00000204, 0.00000023, 0.00000154, -0.00000236, -0.00000773, -0.00000893, -0.00000476, 0.00000237, 0.00000854, 0.00000894, 0.00000180, -0.00000725, -0.00000973, },
		{ -0.00000119, -0.00000163, 0.00000168, 0.00000367, -0.00000090, -0.00000681, -0.00000564, -0.00000083, -0.00000461, -0.00001918, -0.00003186, -0.00003206, -0.00002263, -0.00000915, 0.00001053, 0.00003550, 0.00004973, 0.00003773, 0.00001069, -0.00000010, 0.00001843, 0.00004312, 0.00004442, 0.00002343, 0.00000615, 0.00000673, 0.00001204, 0.00000395, -0.00001635, -0.00003195, -0.00003007, -0.00001235, 0.00000845, 0.00001717, 0.00000737, -0.00001046, -0.00001649, -0.00000353, 0.00001289, 0.00001342, -0.00000083, -0.00001232, -0.00001181, -0.00000672, -0.00000451, -0.00000219, 0.00000239, 0.00000236, -0.00000486, -0.00000748, 0.00000370, 0.00001607, 0.00001059, -0.00000750, -0.00001351, -0.00000113, 0.00000697, -0.00000371, -0.00001366, -0.00000055, 0.00002026, 0.00001535, -0.00001242, -0.00002244, 0.00000306, 0.00002847, 0.00001569, -0.00001844, -0.00002665, -0.00000113, 0.00001738, 0.00000314, -0.00001772, -0.00001084, 0.00001378, 0.00002038, 0.00000299, -0.00001020, -0.00000474, 0.00000018, -0.00001135, -0.00002247, -0.00001016, 0.00001647, 0.00002794, 0.00001649, 0.00000265, 0.00000078, 0.00000052, -0.00001004, -0.00002164, -0.00001811, -0.00000216, 0.00000825, 0.00000445, -0.00000513, -0.00000977, -0.00000754, -0.00000113, 0.00000625, 0.00000891, 0.00000246, -0.00000660, -0.00000491, 0.00000857, 0.00001645, 0.00000482, -0.00001591, -0.00002318, -0.00001178, 0.00000164, 0.00000328, -0.00000140, -0.00000109, 0.00000316, 0.00000214, -0.00000530, -0.00001072, -0.00000860, -0.00000191, 0.00000435, 0.00000785, 0.00000677, 0.00000023, -0.00000748, -0.00000930, -0.00000469, -0.00000068, },
		{ 0.00000196, 0.00000343, -0.00000156, -0.00000784, -0.00000703, -0.00000225, -0.00000567, -0.00002064, -0.00003556, -0.00003787, -0.00002593, -0.00000410, 0.00002332, 0.00004682, 0.00005074, 0.00003079, 0.00000756, 0.00000787, 0.00003130, 0.00004899, 0.00004022, 0.00001747, 0.00000627, 0.00001009, 0.00000953, -0.00000754, -0.00002924, -0.00003584, -0.00002233, -0.00000053, 0.00001394, 0.00001255, -0.00000155, -0.00001449, -0.00001237, 0.00000325, 0.00001486, 0.00000931, -0.00000609, -0.00001441, -0.00001099, -0.00000502, -0.00000257, -0.00000053, 0.00000195, -0.00000033, -0.00000608, -0.00000416, 0.00000826, 0.00001571, 0.00000553, -0.00001006, -0.00001012, 0.00000221, 0.00000417, -0.00000834, -0.00001152, 0.00000671, 0.00002151, 0.00000714, -0.00001859, -0.00001623, 0.00001424, 0.00002866, 0.00000350, -0.00002718, -0.00002131, 0.00000907, 0.00001791, -0.00000470, -0.00002021, -0.00000327, 0.00002065, 0.00001718, -0.00000520, -0.00001343, -0.00000328, -0.00000102, -0.00001513, -0.00002027, 0.00000066, 0.00002598, 0.00002628, 0.00000699, -0.00000352, 0.00000165, 0.00000194, -0.00001297, -0.00002475, -0.00001527, 0.00000494, 0.00001188, 0.00000080, -0.00001152, -0.00001173, -0.00000292, 0.00000566, 0.00000899, 0.00000541, -0.00000329, -0.00000822, -0.00000057, 0.00001350, 0.00001545, -0.00000161, -0.00002058, -0.00002134, -0.00000686, 0.00000326, 0.00000098, -0.00000259, 0.00000154, 0.00000603, 0.00000091, -0.00000945, -0.00001282, -0.00000656, 0.00000131, 0.00000513, 0.00000608, 0.00000486, -0.00000036, -0.00000726, -0.00000921, -0.00000478, -0.00000049, -0.00000058, -0.00000096, },
		{ -0.00000271, -0.00000916, -0.00000801, -0.00000270, -0.00000654, -0.00002382, -0.00004170, -0.00004389, -0.00002584, 0.00000524, 0.00003701, 0.00005486, 0.00004846, 0.00002486, 0.00000891, 0.00001861, 0.00004221, 0.00005097, 0.00003501, 0.00001390, 0.00000758, 0.00000982, 0.00000094, -0.00002136, -0.00003770, -0.00003226, -0.00001084, 0.00000843, 0.00001349, 0.00000487, -0.00000829, -0.00001414, -0.00000631, 0.00000814, 0.00001372, 0.00000395, -0.00001031, -0.00001483, -0.00000915, -0.00000306, -0.00000102, 0.00000011, 0.00000067, -0.00000247, -0.00000564, -0.00000021, 0.00001129, 0.00001356, 0.00000098, -0.00001059, -0.00000640, 0.00000348, 0.00000022, -0.00001091, -0.00000699, 0.00001261, 0.00001904, -0.00000151, -0.00002041, -0.00000671, 0.00002244, 0.00002293, -0.00000975, -0.00003071, -0.00001154, 0.00001745, 0.00001386, -0.00001296, -0.00001878, 0.00000637, 0.00002458, 0.00001010, -0.00001347, -0.00001419, -0.00000091, -0.00000285, -0.00001754, -0.00001444, 0.00001262, 0.00003132, 0.00001908, -0.00000339, -0.00000588, 0.00000553, 0.00000230, -0.00001805, -0.00002692, -0.00000928, 0.00001278, 0.00001217, -0.00000643, -0.00001727, -0.00000935, 0.00000489, 0.00001104, 0.00000738, -0.00000072, -0.00000743, -0.00000619, 0.00000520, 0.00001621, 0.00001142, -0.00000827, -0.00002265, -0.00001742, -0.00000276, 0.00000273, -0.00000167, -0.00000194, 0.00000544, 0.00000769, -0.00000250, -0.00001391, -0.00001308, -0.00000302, 0.00000391, 0.00000426, 0.00000381, 0.00000397, 0.00000005, -0.00000724, -0.00000993, -0.00000511, 0.00000023, 0.00000057, -0.00000029, 0.00000187, 0.00000274, },
		{ -0.00000844, -0.00000280, -0.00000826, -0.00002909, -0.00004930, -0.00004847, -0.00002179, 0.00001765, 0.00004991, 0.00005953, 0.00004457, 0.00002142, 0.00001411, 0.00003023, 0.00005043, 0.00005039, 0.00003025, 0.00001172, 0.00000706, 0.00000414, -0.00001173, -0.00003316, -0.00003915, -0.00002288, 0.00000034, 0.00001210, 0.00000835, -0.00000274, -0.00001118, -0.00001049, -0.00000055, 0.00001013, 0.00001028, -0.00000130, -0.00001287, -0.00001383, -0.00000684, -0.00000128, -0.00000009, -0.00000010, -0.00000089, -0.00000364, -0.00000393, 0.00000353, 0.00001259, 0.00001044, -0.00000240, -0.00000961, -0.00000332, 0.00000285, -0.00000364, -0.00001104, -0.00000140, 0.00001592, 0.00001379, -0.00000845, -0.00001768, 0.00000365, 0.00002562, 0.00001255, -0.00002094, -0.00002812, 0.00000034, 0.00002173, 0.00000597, -0.00001933, -0.00001296, 0.00001613, 0.00002405, 0.00000001, -0.00002009, -0.00001211, 0.00000173, -0.00000515, -0.00001771, -0.00000548, 0.00002343, 0.00003092, 0.00000776, -0.00001179, -0.00000347, 0.00001065, -0.00000013, -0.00002462, -0.00002635, -0.00000016, 0.00001923, 0.00000776, -0.00001574, -0.00001964, -0.00000200, 0.00001354, 0.00001236, 0.00000145, -0.00000694, -0.00000792, -0.00000117, 0.00001012, 0.00001556, 0.00000554, -0.00001330, -0.00002185, -0.00001289, -0.00000067, 0.00000058, -0.00000327, 0.00000088, 0.00000947, 0.00000705, -0.00000768, -0.00001738, -0.00001102, 0.00000115, 0.00000488, 0.00000188, 0.00000205, 0.00000463, 0.00000090, -0.00000824, -0.00001153, -0.00000500, 0.00000202, 0.00000206, -0.00000023, 0.00000101, 0.00000168, -0.00000385, -0.00001022, },
		{ -0.00001129, -0.00003608, -0.00005719, -0.00005064, -0.00001404, 0.00003179, 0.00006114, 0.00006149, 0.00004057, 0.00002092, 0.00002208, 0.00004153, 0.00005626, 0.00004849, 0.00002580, 0.00000864, 0.00000238, -0.00000629, -0.00002471, -0.00003939, -0.00003363, -0.00001117, 0.00000779, 0.00001028, 0.00000119, -0.00000776, -0.00001019, -0.00000556, 0.00000322, 0.00000934, 0.00000586, -0.00000541, -0.00001375, -0.00001196, -0.00000453, 0.00000007, 0.00000020, -0.00000087, -0.00000223, -0.00000375, -0.00000150, 0.00000648, 0.00001240, 0.00000713, -0.00000439, -0.00000787, -0.00000143, 0.00000097, -0.00000644, -0.00000906, 0.00000381, 0.00001627, 0.00000742, -0.00001219, -0.00001148, 0.00001217, 0.00002314, 0.00000009, -0.00002745, -0.00002010, 0.00001138, 0.00002071, -0.00000382, -0.00002174, -0.00000354, 0.00002361, 0.00001841, -0.00001136, -0.00002348, -0.00000748, 0.00000389, -0.00000749, -0.00001511, 0.00000529, 0.00003075, 0.00002445, -0.00000514, -0.00001571, 0.00000318, 0.00001458, -0.00000615, -0.00003092, -0.00002162, 0.00001078, 0.00002188, -0.00000140, -0.00002436, -0.00001641, 0.00000911, 0.00001976, 0.00000809, -0.00000692, -0.00001034, -0.00000417, 0.00000475, 0.00001212, 0.00001186, -0.00000015, -0.00001551, -0.00001914, -0.00000944, -0.00000087, -0.00000182, -0.00000273, 0.00000526, 0.00001214, 0.00000373, -0.00001342, -0.00001865, -0.00000697, 0.00000462, 0.00000370, -0.00000108, 0.00000188, 0.00000667, 0.00000108, -0.00001084, -0.00001338, -0.00000351, 0.00000496, 0.00000313, -0.00000148, -0.00000056, 0.00000090, -0.00000444, -0.00001095, -0.00000880, -0.00000330, },
		{ -0.00006457, -0.00005007, -0.00000313, 0.00004676, 0.00007048, 0.00006153, 0.00003730, 0.00002325, 0.00003198, 0.00005221, 0.00006026, 0.00004535, 0.00002008, 0.00000234, -0.00000681, -0.00001863, -0.00003402, -0.00003858, -0.00002358, -0.00000101, 0.00000986, 0.00000472, -0.00000505, -0.00000907, -0.00000672, -0.00000140, 0.00000437, 0.00000675, 0.00000175, -0.00000797, -0.00001338, -0.00000976, -0.00000250, 0.00000094, -0.00000006, -0.00000188, -0.00000306, -0.00000295, 0.00000108, 0.00000836, 0.00001119, 0.00000424, -0.00000519, -0.00000608, -0.00000081, -0.00000131, -0.00000769, -0.00000584, 0.00000755, 0.00001421, 0.00000171, -0.00001228, -0.00000386, 0.00001679, 0.00001589, -0.00001136, -0.00002797, -0.00000883, 0.00001886, 0.00001474, -0.00001284, -0.00001912, 0.00000748, 0.00002663, 0.00000823, -0.00002170, -0.00002265, -0.00000117, 0.00000496, -0.00000927, -0.00000974, 0.00001598, 0.00003282, 0.00001304, -0.00001650, -0.00001382, 0.00001213, 0.00001504, -0.00001522, -0.00003465, -0.00001238, 0.00002110, 0.00001887, -0.00001369, -0.00002901, -0.00000687, 0.00002104, 0.00002043, -0.00000134, -0.00001448, -0.00000879, 0.00000258, 0.00000882, 0.00001026, 0.00000684, -0.00000362, -0.00001500, -0.00001635, -0.00000821, -0.00000245, -0.00000284, 0.00000017, 0.00000970, 0.00001222, -0.00000166, -0.00001812, -0.00001722, -0.00000212, 0.00000614, 0.00000069, -0.00000324, 0.00000384, 0.00000913, -0.00000056, -0.00001486, -0.00001425, 0.00000004, 0.00000837, 0.00000263, -0.00000430, -0.00000205, 0.00000123, -0.00000441, -0.00001196, -0.00000972, -0.00000449, -0.00001546, -0.00004415, },
		{ 0.00001059, 0.00006214, 0.00007793, 0.00006015, 0.00003517, 0.00002834, 0.00004358, 0.00006228, 0.00006224, 0.00003970, 0.00001102, -0.00000809, -0.00001861, -0.00002903, -0.00003710, -0.00003183, -0.00001267, 0.00000483, 0.00000711, -0.00000178, -0.00000836, -0.00000717, -0.00000279, 0.00000069, 0.00000332, 0.00000366, -0.00000128, -0.00000916, -0.00001231, -0.00000757, -0.00000082, 0.00000134, -0.00000071, -0.00000284, -0.00000325, -0.00000153, 0.00000335, 0.00000921, 0.00000949, 0.00000205, -0.00000522, -0.00000475, -0.00000115, -0.00000326, -0.00000747, -0.00000242, 0.00000938, 0.00001090, -0.00000203, -0.00000943, 0.00000288, 0.00001666, 0.00000607, -0.00001914, -0.00002283, 0.00000270, 0.00002111, 0.00000560, -0.00001856, -0.00001181, 0.00001740, 0.00002390, -0.00000471, -0.00002869, -0.00001751, 0.00000548, 0.00000462, -0.00000986, -0.00000224, 0.00002451, 0.00002894, -0.00000100, -0.00002351, -0.00000639, 0.00002059, 0.00001069, -0.00002544, -0.00003376, 0.00000026, 0.00002789, 0.00000979, -0.00002598, -0.00002697, 0.00000754, 0.00002979, 0.00001393, -0.00001334, -0.00001772, -0.00000201, 0.00000948, 0.00000875, 0.00000543, 0.00000305, -0.00000387, -0.00001328, -0.00001535, -0.00000911, -0.00000364, -0.00000141, 0.00000444, 0.00001249, 0.00000948, -0.00000760, -0.00002046, -0.00001362, 0.00000196, 0.00000522, -0.00000290, -0.00000334, 0.00000754, 0.00001046, -0.00000463, -0.00001918, -0.00001277, 0.00000549, 0.00001080, -0.00000036, -0.00000802, -0.00000216, 0.00000309, -0.00000450, -0.00001417, -0.00001147, -0.00000601, -0.00002031, -0.00005293, -0.00007113, -0.00004669, },
		{ 0.00008323, 0.00005754, 0.00003473, 0.00003664, 0.00005693, 0.00007115, 0.00006071, 0.00002954, -0.00000252, -0.00002139, -0.00002951, -0.00003424, -0.00003375, -0.00002218, -0.00000436, 0.00000560, 0.00000170, -0.00000665, -0.00000826, -0.00000369, -0.00000013, 0.00000053, 0.00000118, 0.00000113, -0.00000309, -0.00000949, -0.00001097, -0.00000551, 0.00000053, 0.00000130, -0.00000161, -0.00000354, -0.00000283, 0.00000014, 0.00000506, 0.00000924, 0.00000775, 0.00000058, -0.00000493, -0.00000404, -0.00000197, -0.00000446, -0.00000627, 0.00000040, 0.00000956, 0.00000764, -0.00000332, -0.00000520, 0.00000689, 0.00001239, -0.00000356, -0.00002171, -0.00001388, 0.00001163, 0.00001806, -0.00000402, -0.00001936, -0.00000158, 0.00002353, 0.00001551, -0.00001774, -0.00003062, -0.00000897, 0.00001104, 0.00000299, -0.00000878, 0.00000623, 0.00002910, 0.00001968, -0.00001465, -0.00002438, 0.00000480, 0.00002576, 0.00000169, -0.00003418, -0.00002730, 0.00001381, 0.00002873, -0.00000388, -0.00003448, -0.00001725, 0.00002339, 0.00003171, 0.00000092, -0.00002387, -0.00001430, 0.00000806, 0.00001309, 0.00000391, 0.00000013, 0.00000270, -0.00000160, -0.00001270, -0.00001708, -0.00001064, -0.00000264, 0.00000230, 0.00000821, 0.00001248, 0.00000486, -0.00001231, -0.00002011, -0.00000933, 0.00000397, 0.00000247, -0.00000543, -0.00000090, 0.00001159, 0.00000927, -0.00001067, -0.00002205, -0.00000822, 0.00001158, 0.00001062, -0.00000582, -0.00001111, 0.00000022, 0.00000593, -0.00000610, -0.00001816, -0.00001344, -0.00000709, -0.00002566, -0.00006262, -0.00007682, -0.00004020, 0.00002718, 0.00007756, },
		{ 0.00003715, 0.00004889, 0.00007148, 0.00007688, 0.00005326, 0.00001342, -0.00001968, -0.00003438, -0.00003594, -0.00003302, -0.00002609, -0.00001324, -0.00000053, 0.00000251, -0.00000367, -0.00000849, -0.00000571, -0.00000052, 0.00000045, -0.00000116, -0.00000088, -0.00000039, -0.00000403, -0.00000941, -0.00000953, -0.00000355, 0.00000153, 0.00000085, -0.00000260, -0.00000383, -0.00000192, 0.00000174, 0.00000612, 0.00000877, 0.00000624, -0.00000035, -0.00000468, -0.00000388, -0.00000282, -0.00000487, -0.00000471, 0.00000227, 0.00000883, 0.00000541, -0.00000262, -0.00000139, 0.00000738, 0.00000573, -0.00001051, -0.00001914, -0.00000387, 0.00001604, 0.00001114, -0.00001136, -0.00001509, 0.00000883, 0.00002400, 0.00000305, -0.00002797, -0.00002688, 0.00000125, 0.00001434, 0.00000056, -0.00000589, 0.00001413, 0.00002869, 0.00000681, -0.00002496, -0.00001884, 0.00001694, 0.00002570, -0.00001034, -0.00003883, -0.00001587, 0.00002507, 0.00002262, -0.00001897, -0.00003593, -0.00000117, 0.00003612, 0.00002481, -0.00001545, -0.00002871, -0.00000412, 0.00001769, 0.00001085, -0.00000416, -0.00000241, 0.00000653, 0.00000089, -0.00001537, -0.00002082, -0.00001048, 0.00000142, 0.00000669, 0.00000958, 0.00000979, 0.00000026, -0.00001456, -0.00001798, -0.00000609, 0.00000365, -0.00000057, -0.00000560, 0.00000332, 0.00001413, 0.00000505, -0.00001707, -0.00002183, -0.00000107, 0.00001627, 0.00000679, -0.00001249, -0.00001165, 0.00000531, 0.00000813, -0.00001055, -0.00002356, -0.00001433, -0.00000720, -0.00003215, -0.00007375, -0.00008118, -0.00002976, 0.00004670, 0.00009223, 0.00008575, 0.00005416, },
		{ 0.00008514, 0.00007628, 0.00003768, -0.00000824, -0.00003738, -0.00004313, -0.00003580, -0.00002658, -0.00001756, -0.00000772, -0.00000105, -0.00000203, -0.00000693, -0.00000744, -0.00000243, 0.00000103, -0.00000081, -0.00000324, -0.00000211, -0.00000106, -0.00000458, -0.00000915, -0.00000794, -0.00000169, 0.00000211, 0.00000004, -0.00000348, -0.00000366, -0.00000075, 0.00000304, 0.00000662, 0.00000813, 0.00000507, -0.00000099, -0.00000463, -0.00000404, -0.00000340, -0.00000471, -0.00000328, 0.00000331, 0.00000804, 0.00000454, -0.00000109, 0.00000056, 0.00000477, -0.00000102, -0.00001335, -0.00001292, 0.00000447, 0.00001551, 0.00000284, -0.00001449, -0.00000717, 0.00001649, 0.00001843, -0.00001074, -0.00003305, -0.00001816, 0.00001103, 0.00001479, -0.00000189, -0.00000150, 0.00001992, 0.00002323, -0.00000712, -0.00002975, -0.00000816, 0.00002696, 0.00001988, -0.00002274, -0.00003772, -0.00000160, 0.00003115, 0.00001053, -0.00003144, -0.00002880, 0.00001766, 0.00004142, 0.00000974, -0.00003037, -0.00002505, 0.00001029, 0.00002265, 0.00000224, -0.00001205, 0.00000037, 0.00001314, 0.00000048, -0.00002194, -0.00002423, -0.00000657, 0.00000776, 0.00000936, 0.00000760, 0.00000598, -0.00000239, -0.00001446, -0.00001591, -0.00000502, 0.00000204, -0.00000207, -0.00000326, 0.00000754, 0.00001372, -0.00000133, -0.00002168, -0.00001784, 0.00000690, 0.00001756, -0.00000036, -0.00001813, -0.00000832, 0.00001189, 0.00000755, -0.00001831, -0.00002879, -0.00001273, -0.00000677, -0.00004109, -0.00008653, -0.00008294, -0.00001430, 0.00006862, 0.00010472, 0.00008510, 0.00005137, 0.00004423, 0.00006535, },
		{ 0.00001352, -0.00003247, -0.00005120, -0.00004472, -0.00002946, -0.00001802, -0.00001138, -0.00000649, -0.00000412, -0.00000565, -0.00000739, -0.00000478, -0.00000005, 0.00000062, -0.00000297, -0.00000474, -0.00000241, -0.00000133, -0.00000507, -0.00000869, -0.00000614, -0.00000003, 0.00000218, -0.00000101, -0.00000407, -0.00000307, 0.00000042, 0.00000392, 0.00000676, 0.00000753, 0.00000419, -0.00000157, -0.00000482, -0.00000426, -0.00000365, -0.00000431, -0.00000218, 0.00000397, 0.00000778, 0.00000465, -0.00000002, 0.00000007, 0.00000047, -0.00000577, -0.00001200, -0.00000541, 0.00000923, 0.00001115, -0.00000416, -0.00001290, 0.00000188, 0.00001921, 0.00000804, -0.00002267, -0.00003180, -0.00000631, 0.00001840, 0.00001253, -0.00000356, 0.00000368, 0.00002238, 0.00001371, -0.00001933, -0.00002813, 0.00000520, 0.00003238, 0.00000933, -0.00003262, -0.00003069, 0.00001243, 0.00003037, -0.00000475, -0.00003759, -0.00001401, 0.00003435, 0.00003678, -0.00001011, -0.00003888, -0.00001274, 0.00002436, 0.00001981, -0.00001057, -0.00001562, 0.00000910, 0.00001924, -0.00000532, -0.00003099, -0.00002417, 0.00000175, 0.00001393, 0.00000810, 0.00000303, 0.00000349, -0.00000213, -0.00001357, -0.00001576, -0.00000586, 0.00000114, -0.00000095, 0.00000033, 0.00000969, 0.00001019, -0.00000780, -0.00002279, -0.00001100, 0.00001317, 0.00001447, -0.00000897, -0.00002035, -0.00000120, 0.00001753, 0.00000248, -0.00002835, -0.00003162, -0.00000811, -0.00000742, -0.00005388, -0.00010013, -0.00008010, 0.00000675, 0.00009134, 0.00011340, 0.00008190, 0.00005169, 0.00005762, 0.00008457, 0.00009402, 0.00006586, },
		{ -0.00005711, -0.00003857, -0.00001964, -0.00001091, -0.00000928, -0.00000847, -0.00000733, -0.00000696, -0.00000573, -0.00000218, 0.00000059, -0.00000114, -0.00000496, -0.00000524, -0.00000216, -0.00000165, -0.00000552, -0.00000785, -0.00000417, 0.00000124, 0.00000173, -0.00000211, -0.00000426, -0.00000222, 0.00000140, 0.00000442, 0.00000677, 0.00000711, 0.00000344, -0.00000225, -0.00000516, -0.00000440, -0.00000367, -0.00000389, -0.00000129, 0.00000475, 0.00000813, 0.00000488, -0.00000040, -0.00000242, -0.00000368, -0.00000735, -0.00000770, 0.00000100, 0.00000988, 0.00000509, -0.00000791, -0.00000763, 0.00000920, 0.00001621, -0.00000459, -0.00002998, -0.00002457, 0.00000614, 0.00002204, 0.00000838, -0.00000388, 0.00000860, 0.00002093, 0.00000194, -0.00002750, -0.00002064, 0.00001827, 0.00003195, -0.00000371, -0.00003772, -0.00001914, 0.00002308, 0.00002292, -0.00001934, -0.00003531, 0.00000516, 0.00004414, 0.00002249, -0.00002948, -0.00003755, 0.00000541, 0.00003297, 0.00000855, -0.00002328, -0.00001169, 0.00002175, 0.00002075, -0.00001712, -0.00003913, -0.00001808, 0.00001297, 0.00001664, 0.00000220, -0.00000167, 0.00000454, 0.00000023, -0.00001432, -0.00001834, -0.00000682, 0.00000290, 0.00000237, 0.00000284, 0.00000852, 0.00000500, -0.00001195, -0.00002007, -0.00000363, 0.00001563, 0.00000777, -0.00001625, -0.00001783, 0.00000783, 0.00001949, -0.00000730, -0.00003819, -0.00003011, -0.00000144, -0.00001155, -0.00007101, -0.00011240, -0.00007063, 0.00003268, 0.00011239, 0.00011740, 0.00007846, 0.00005824, 0.00007746, 0.00010271, 0.00009317, 0.00004353, -0.00001682, -0.00005425, },
		{ -0.00001027, -0.00000780, -0.00001091, -0.00001142, -0.00000883, -0.00000600, -0.00000342, -0.00000085, -0.00000036, -0.00000322, -0.00000607, -0.00000496, -0.00000188, -0.00000218, -0.00000572, -0.00000654, -0.00000222, 0.00000194, 0.00000086, -0.00000301, -0.00000405, -0.00000133, 0.00000205, 0.00000469, 0.00000685, 0.00000683, 0.00000266, -0.00000308, -0.00000551, -0.00000439, -0.00000358, -0.00000351, -0.00000033, 0.00000591, 0.00000872, 0.00000431, -0.00000249, -0.00000573, -0.00000606, -0.00000576, -0.00000239, 0.00000464, 0.00000726, -0.00000030, -0.00000770, -0.00000085, 0.00001255, 0.00000838, -0.00001629, -0.00003106, -0.00001311, 0.00001662, 0.00002160, 0.00000358, -0.00000270, 0.00001217, 0.00001576, -0.00000983, -0.00003021, -0.00000904, 0.00002832, 0.00002586, -0.00001643, -0.00003704, -0.00000563, 0.00002815, 0.00001081, -0.00002942, -0.00002480, 0.00002391, 0.00004393, 0.00000174, -0.00004278, -0.00002582, 0.00002425, 0.00003227, -0.00000869, -0.00003078, 0.00000053, 0.00003404, 0.00001441, -0.00003283, -0.00004213, -0.00000534, 0.00002355, 0.00001317, -0.00000685, -0.00000317, 0.00000982, 0.00000218, -0.00001885, -0.00002260, -0.00000523, 0.00000806, 0.00000574, 0.00000203, 0.00000451, 0.00000083, -0.00001224, -0.00001510, 0.00000146, 0.00001367, -0.00000005, -0.00001964, -0.00001114, 0.00001572, 0.00001597, -0.00002010, -0.00004487, -0.00002375, 0.00000478, -0.00002127, -0.00009135, -0.00012036, -0.00005356, 0.00006119, 0.00012933, 0.00011773, 0.00007865, 0.00007327, 0.00010112, 0.00011377, 0.00007825, 0.00001025, -0.00004802, -0.00006809, -0.00005319, -0.00002688, },
		{ -0.00001431, -0.00001317, -0.00000808, -0.00000392, -0.00000187, -0.00000109, -0.00000205, -0.00000473, -0.00000624, -0.00000440, -0.00000192, -0.00000274, -0.00000540, -0.00000485, -0.00000059, 0.00000202, -0.00000017, -0.00000356, -0.00000356, -0.00000062, 0.00000240, 0.00000494, 0.00000712, 0.00000657, 0.00000170, -0.00000401, -0.00000577, -0.00000426, -0.00000346, -0.00000304, 0.00000094, 0.00000733, 0.00000884, 0.00000237, -0.00000577, -0.00000840, -0.00000586, -0.00000207, 0.00000193, 0.00000504, 0.00000315, -0.00000321, -0.00000431, 0.00000488, 0.00001097, -0.00000204, -0.00002417, -0.00002592, -0.00000006, 0.00002324, 0.00001774, -0.00000063, -0.00000039, 0.00001349, 0.00000782, -0.00001941, -0.00002725, 0.00000415, 0.00003349, 0.00001560, -0.00002631, -0.00003104, 0.00000694, 0.00002698, -0.00000269, -0.00003243, -0.00000865, 0.00003736, 0.00003339, -0.00002019, -0.00004595, -0.00000637, 0.00003794, 0.00002117, -0.00002694, -0.00002902, 0.00001872, 0.00004078, -0.00000062, -0.00004806, -0.00003660, 0.00001195, 0.00002907, 0.00000279, -0.00001547, 0.00000130, 0.00001771, 0.00000028, -0.00002768, -0.00002564, 0.00000118, 0.00001531, 0.00000606, -0.00000300, 0.00000011, 0.00000033, -0.00000908, -0.00001089, 0.00000232, 0.00000884, -0.00000580, -0.00001819, -0.00000282, 0.00001947, 0.00000720, -0.00003278, -0.00004622, -0.00001409, 0.00000739, -0.00003721, -0.00011206, -0.00012138, -0.00002987, 0.00008903, 0.00014124, 0.00011780, 0.00008664, 0.00009644, 0.00012269, 0.00011112, 0.00004769, -0.00002924, -0.00007313, -0.00007016, -0.00004070, -0.00001392, -0.00000482, -0.00000916, },
		{ -0.00000587, -0.00000221, -0.00000174, -0.00000232, -0.00000352, -0.00000532, -0.00000588, -0.00000407, -0.00000225, -0.00000301, -0.00000451, -0.00000309, 0.00000047, 0.00000159, -0.00000111, -0.00000374, -0.00000304, -0.00000020, 0.00000261, 0.00000539, 0.00000755, 0.00000616, 0.00000049, -0.00000496, -0.00000587, -0.00000403, -0.00000328, -0.00000231, 0.00000257, 0.00000857, 0.00000786, -0.00000091, -0.00000916, -0.00000922, -0.00000325, 0.00000206, 0.00000393, 0.00000292, -0.00000043, -0.00000301, 0.00000043, 0.00000747, 0.00000509, -0.00001210, -0.00002644, -0.00001606, 0.00001174, 0.00002517, 0.00001182, -0.00000334, 0.00000230, 0.00001211, -0.00000138, -0.00002516, -0.00001955, 0.00001632, 0.00003316, 0.00000347, -0.00003168, -0.00002138, 0.00001611, 0.00002060, -0.00001408, -0.00002783, 0.00000895, 0.00004203, 0.00001510, -0.00003751, -0.00003777, 0.00001556, 0.00004184, 0.00000194, -0.00004025, -0.00001665, 0.00003791, 0.00003776, -0.00002200, -0.00005735, -0.00002156, 0.00002918, 0.00002590, -0.00001239, -0.00001909, 0.00001242, 0.00002447, -0.00000819, -0.00003885, -0.00002371, 0.00001287, 0.00002127, 0.00000068, -0.00001079, -0.00000113, 0.00000465, -0.00000515, -0.00001057, -0.00000082, 0.00000441, -0.00000712, -0.00001335, 0.00000356, 0.00001763, -0.00000441, -0.00004200, -0.00004211, -0.00000427, 0.00000401, -0.00005782, -0.00012965, -0.00011451, -0.00000245, 0.00011351, 0.00014991, 0.00012272, 0.00010485, 0.00012342, 0.00013403, 0.00008996, 0.00000431, -0.00006713, -0.00008585, -0.00005995, -0.00002388, -0.00000427, -0.00000485, -0.00001336, -0.00001697, -0.00001261, },
		{ -0.00000274, -0.00000358, -0.00000419, -0.00000524, -0.00000556, -0.00000413, -0.00000257, -0.00000272, -0.00000320, -0.00000159, 0.00000091, 0.00000092, -0.00000177, -0.00000369, -0.00000270, -0.00000005, 0.00000291, 0.00000612, 0.00000797, 0.00000542, -0.00000095, -0.00000580, -0.00000577, -0.00000373, -0.00000295, -0.00000119, 0.00000434, 0.00000907, 0.00000548, -0.00000492, -0.00001145, -0.00000773, 0.00000073, 0.00000501, 0.00000333, -0.00000019, -0.00000199, -0.00000033, 0.00000434, 0.00000602, -0.00000316, -0.00001905, -0.00002288, -0.00000404, 0.00002004, 0.00002278, 0.00000549, -0.00000424, 0.00000444, 0.00000820, -0.00001008, -0.00002628, -0.00000891, 0.00002534, 0.00002794, -0.00000810, -0.00003208, -0.00001040, 0.00002061, 0.00001131, -0.00002068, -0.00001725, 0.00002347, 0.00003691, -0.00000616, -0.00004567, -0.00002039, 0.00003376, 0.00003414, -0.00002027, -0.00004367, 0.00000419, 0.00005193, 0.00002345, -0.00004456, -0.00005621, 0.00000064, 0.00004064, 0.00001291, -0.00002766, -0.00001400, 0.00002793, 0.00002536, -0.00002377, -0.00004813, -0.00001383, 0.00002753, 0.00002162, -0.00001099, -0.00001756, 0.00000392, 0.00001231, -0.00000448, -0.00001567, -0.00000517, 0.00000401, -0.00000394, -0.00000861, 0.00000513, 0.00001137, -0.00001511, -0.00004596, -0.00003479, 0.00000228, -0.00000560, -0.00007986, -0.00014155, -0.00010134, 0.00002512, 0.00013435, 0.00016010, 0.00013731, 0.00013176, 0.00014601, 0.00012713, 0.00004978, -0.00004450, -0.00009464, -0.00008291, -0.00004085, -0.00000844, -0.00000101, -0.00000939, -0.00001763, -0.00001717, -0.00001016, -0.00000371, -0.00000181, },
		{ -0.00000413, -0.00000504, -0.00000561, -0.00000437, -0.00000251, -0.00000189, -0.00000182, -0.00000060, 0.00000088, 0.00000029, -0.00000214, -0.00000363, -0.00000264, 0.00000003, 0.00000352, 0.00000709, 0.00000817, 0.00000426, -0.00000249, -0.00000641, -0.00000547, -0.00000332, -0.00000236, 0.00000023, 0.00000579, 0.00000835, 0.00000193, -0.00000863, -0.00001177, -0.00000427, 0.00000462, 0.00000580, 0.00000086, -0.00000262, -0.00000105, 0.00000329, 0.00000564, 0.00000104, -0.00001117, -0.00002117, -0.00001475, 0.00000734, 0.00002370, 0.00001737, 0.00000019, -0.00000364, 0.00000521, 0.00000252, -0.00001670, -0.00002294, 0.00000249, 0.00002996, 0.00001934, -0.00001715, -0.00002818, -0.00000032, 0.00002049, 0.00000188, -0.00002142, -0.00000390, 0.00003139, 0.00002373, -0.00002487, -0.00004283, 0.00000139, 0.00004304, 0.00001655, -0.00003904, -0.00003503, 0.00002830, 0.00005553, -0.00000005, -0.00006184, -0.00004287, 0.00002468, 0.00004152, -0.00000760, -0.00003711, 0.00000096, 0.00004287, 0.00001654, -0.00004379, -0.00005029, 0.00000450, 0.00004033, 0.00001291, -0.00002644, -0.00001837, 0.00001609, 0.00001912, -0.00001078, -0.00002476, -0.00000620, 0.00000960, 0.00000086, -0.00000808, 0.00000143, 0.00000423, -0.00002170, -0.00004555, -0.00002798, 0.00000368, -0.00001917, -0.00009982, -0.00014769, -0.00008538, 0.00005066, 0.00015490, 0.00017808, 0.00016337, 0.00016076, 0.00015393, 0.00009724, -0.00000425, -0.00008835, -0.00010502, -0.00006562, -0.00001920, 0.00000062, -0.00000440, -0.00001550, -0.00001941, -0.00001477, -0.00000731, -0.00000287, -0.00000272, -0.00000394, -0.00000415, },
		{ -0.00000594, -0.00000442, -0.00000193, -0.00000077, -0.00000071, -0.00000012, 0.00000066, -0.00000014, -0.00000238, -0.00000378, -0.00000280, 0.00000029, 0.00000453, 0.00000810, 0.00000793, 0.00000270, -0.00000397, -0.00000669, -0.00000496, -0.00000277, -0.00000149, 0.00000170, 0.00000646, 0.00000629, -0.00000213, -0.00001103, -0.00000989, 0.00000010, 0.00000711, 0.00000434, -0.00000214, -0.00000318, 0.00000177, 0.00000597, 0.00000361, -0.00000577, -0.00001651, -0.00001817, -0.00000439, 0.00001574, 0.00002284, 0.00001073, -0.00000327, -0.00000231, 0.00000422, -0.00000377, -0.00002014, -0.00001614, 0.00001261, 0.00003000, 0.00000935, -0.00002254, -0.00002145, 0.00000727, 0.00001694, -0.00000533, -0.00001697, 0.00000848, 0.00003126, 0.00000636, -0.00003646, -0.00003029, 0.00002155, 0.00004084, -0.00000619, -0.00004877, -0.00001581, 0.00004902, 0.00004625, -0.00002739, -0.00006817, -0.00001937, 0.00004373, 0.00002981, -0.00003013, -0.00003572, 0.00002329, 0.00005101, -0.00000301, -0.00006272, -0.00004112, 0.00002823, 0.00004531, -0.00000554, -0.00004026, -0.00000915, 0.00003275, 0.00001943, -0.00002527, -0.00003330, 0.00000044, 0.00001991, 0.00000226, -0.00001430, -0.00000475, 0.00000065, -0.00002341, -0.00004427, -0.00002491, 0.00000092, -0.00003289, -0.00011615, -0.00015099, -0.00007024, 0.00007563, 0.00018183, 0.00020882, 0.00019728, 0.00018089, 0.00013819, 0.00004564, -0.00006172, -0.00011693, -0.00009608, -0.00003978, -0.00000200, 0.00000098, -0.00001189, -0.00001972, -0.00001769, -0.00001120, -0.00000577, -0.00000367, -0.00000403, -0.00000444, -0.00000390, -0.00000387, -0.00000517, },
		{ -0.00000094, 0.00000030, -0.00000003, 0.00000004, 0.00000048, -0.00000041, -0.00000272, -0.00000422, -0.00000297, 0.00000096, 0.00000587, 0.00000889, 0.00000714, 0.00000089, -0.00000519, -0.00000662, -0.00000427, -0.00000204, -0.00000045, 0.00000283, 0.00000601, 0.00000322, -0.00000575, -0.00001147, -0.00000631, 0.00000408, 0.00000745, 0.00000139, -0.00000423, -0.00000159, 0.00000510, 0.00000631, -0.00000122, -0.00001215, -0.00001767, -0.00001117, 0.00000558, 0.00001993, 0.00001861, 0.00000451, -0.00000478, -0.00000114, 0.00000159, -0.00000936, -0.00002002, -0.00000747, 0.00001991, 0.00002617, -0.00000014, -0.00002407, -0.00001360, 0.00001170, 0.00001178, -0.00000902, -0.00000943, 0.00001679, 0.00002394, -0.00001045, -0.00003863, -0.00001202, 0.00003476, 0.00002803, -0.00002779, -0.00004642, 0.00000919, 0.00006030, 0.00002543, -0.00005153, -0.00006074, 0.00000880, 0.00005173, 0.00000746, -0.00004760, -0.00002139, 0.00004721, 0.00004707, -0.00003046, -0.00007369, -0.00001955, 0.00005114, 0.00003768, -0.00003043, -0.00004571, 0.00001102, 0.00004787, 0.00000846, -0.00004523, -0.00003486, 0.00001646, 0.00003003, -0.00000481, -0.00002628, -0.00000804, 0.00000356, -0.00002303, -0.00004669, -0.00002618, -0.00000212, -0.00004365, -0.00013079, -0.00015615, -0.00005670, 0.00010613, 0.00022277, 0.00025258, 0.00022913, 0.00017971, 0.00009492, -0.00001943, -0.00010972, -0.00012352, -0.00007138, -0.00001378, 0.00000566, -0.00000610, -0.00001926, -0.00001973, -0.00001347, -0.00000860, -0.00000640, -0.00000531, -0.00000459, -0.00000390, -0.00000335, -0.00000398, -0.00000569, -0.00000620, -0.00000400, },
		{ 0.00000033, 0.00000013, 0.00000042, -0.00000071, -0.00000333, -0.00000482, -0.00000286, 0.00000211, 0.00000730, 0.00000921, 0.00000583, -0.00000095, -0.00000604, -0.00000620, -0.00000342, -0.00000120, 0.00000052, 0.00000325, 0.00000443, -0.00000022, -0.00000807, -0.00000986, -0.00000203, 0.00000655, 0.00000574, -0.00000178, -0.00000448, 0.00000152, 0.00000736, 0.00000378, -0.00000733, -0.00001604, -0.00001445, -0.00000226, 0.00001299, 0.00001986, 0.00001271, -0.00000017, -0.00000487, -0.00000085, -0.00000206, -0.00001314, -0.00001672, 0.00000136, 0.00002365, 0.00001980, -0.00000770, -0.00002234, -0.00000614, 0.00001325, 0.00000674, -0.00000919, -0.00000147, 0.00001945, 0.00001223, -0.00002256, -0.00003187, 0.00000668, 0.00003798, 0.00000849, -0.00004241, -0.00003256, 0.00003341, 0.00005871, -0.00000203, -0.00006594, -0.00004073, 0.00003423, 0.00004552, -0.00001990, -0.00005363, 0.00000383, 0.00006520, 0.00002896, -0.00005939, -0.00007093, 0.00001126, 0.00006551, 0.00001607, -0.00005486, -0.00003725, 0.00003866, 0.00005376, -0.00001518, -0.00006402, -0.00002378, 0.00003917, 0.00003271, -0.00002272, -0.00003886, -0.00000274, 0.00001214, -0.00002624, -0.00005579, -0.00002841, -0.00000047, -0.00005169, -0.00014932, -0.00016684, -0.00004031, 0.00015182, 0.00028250, 0.00030229, 0.00024449, 0.00014777, 0.00002814, -0.00008419, -0.00013676, -0.00010765, -0.00003929, 0.00000435, 0.00000259, -0.00001636, -0.00002261, -0.00001548, -0.00000918, -0.00000864, -0.00000870, -0.00000633, -0.00000372, -0.00000279, -0.00000325, -0.00000466, -0.00000623, -0.00000601, -0.00000314, 0.00000010, 0.00000106, },
		{ 0.00000035, -0.00000125, -0.00000420, -0.00000533, -0.00000229, 0.00000362, 0.00000854, 0.00000894, 0.00000415, -0.00000260, -0.00000643, -0.00000549, -0.00000246, -0.00000036, 0.00000111, 0.00000275, 0.00000206, -0.00000319, -0.00000860, -0.00000676, 0.00000176, 0.00000697, 0.00000274, -0.00000396, -0.00000277, 0.00000489, 0.00000741, -0.00000102, -0.00001275, -0.00001621, -0.00000791, 0.00000621, 0.00001667, 0.00001655, 0.00000682, -0.00000294, -0.00000436, -0.00000171, -0.00000578, -0.00001446, -0.00001121, 0.00000885, 0.00002391, 0.00001246, -0.00001255, -0.00001840, -0.00000005, 0.00001273, 0.00000297, -0.00000690, 0.00000456, 0.00001673, -0.00000020, -0.00002760, -0.00001912, 0.00002085, 0.00003120, -0.00001216, -0.00004645, -0.00001106, 0.00005053, 0.00004459, -0.00002920, -0.00006677, -0.00001314, 0.00004993, 0.00002614, -0.00004452, -0.00004486, 0.00003397, 0.00007050, -0.00000102, -0.00008157, -0.00005211, 0.00004418, 0.00006481, -0.00001611, -0.00007014, -0.00001313, 0.00006605, 0.00004393, -0.00004785, -0.00007285, 0.00000198, 0.00006129, 0.00002120, -0.00004897, -0.00004403, 0.00001410, 0.00002089, -0.00003927, -0.00007033, -0.00002490, 0.00000881, -0.00006228, -0.00017877, -0.00018198, -0.00001066, 0.00022244, 0.00035876, 0.00034343, 0.00022850, 0.00008298, -0.00004992, -0.00013315, -0.00013689, -0.00007564, -0.00001019, 0.00000992, -0.00000813, -0.00002445, -0.00002011, -0.00000920, -0.00000750, -0.00001157, -0.00001101, -0.00000547, -0.00000167, -0.00000209, -0.00000406, -0.00000557, -0.00000627, -0.00000527, -0.00000216, 0.00000088, 0.00000153, 0.00000061, 0.00000031, },
		{ -0.00000516, -0.00000550, -0.00000124, 0.00000525, 0.00000936, 0.00000812, 0.00000234, -0.00000390, -0.00000640, -0.00000458, -0.00000149, 0.00000028, 0.00000108, 0.00000140, -0.00000051, -0.00000499, -0.00000737, -0.00000311, 0.00000411, 0.00000549, -0.00000039, -0.00000439, 0.00000028, 0.00000718, 0.00000494, -0.00000670, -0.00001576, -0.00001262, 0.00000007, 0.00001233, 0.00001655, 0.00001155, 0.00000213, -0.00000411, -0.00000404, -0.00000349, -0.00000863, -0.00001328, -0.00000477, 0.00001402, 0.00002139, 0.00000554, -0.00001460, -0.00001339, 0.00000428, 0.00001113, 0.00000082, -0.00000377, 0.00000732, 0.00001040, -0.00001002, -0.00002550, -0.00000462, 0.00002737, 0.00001727, -0.00002845, -0.00003955, 0.00001220, 0.00005648, 0.00002184, -0.00004926, -0.00005408, 0.00001475, 0.00005161, -0.00000132, -0.00005901, -0.00002231, 0.00006082, 0.00005956, -0.00003633, -0.00008958, -0.00001979, 0.00007004, 0.00004634, -0.00005099, -0.00006884, 0.00002300, 0.00008335, 0.00001604, -0.00008100, -0.00006401, 0.00003875, 0.00007278, -0.00000741, -0.00007571, -0.00003391, 0.00004007, 0.00002088, -0.00006535, -0.00008367, -0.00000833, 0.00002336, -0.00008518, -0.00022363, -0.00019303, 0.00004656, 0.00032274, 0.00043972, 0.00035684, 0.00017150, -0.00000669, -0.00012139, -0.00015413, -0.00011216, -0.00003870, 0.00000730, 0.00000329, -0.00002054, -0.00002631, -0.00001294, -0.00000435, -0.00000986, -0.00001589, -0.00001137, -0.00000248, 0.00000042, -0.00000269, -0.00000558, -0.00000600, -0.00000554, -0.00000431, -0.00000150, 0.00000133, 0.00000195, 0.00000100, 0.00000049, 0.00000006, -0.00000211, },
		{ 0.00000015, 0.00000674, 0.00000967, 0.00000691, 0.00000061, -0.00000479, -0.00000602, -0.00000355, -0.00000065, 0.00000049, 0.00000032, -0.00000048, -0.00000259, -0.00000526, -0.00000489, 0.00000006, 0.00000460, 0.00000286, -0.00000260, -0.00000300, 0.00000354, 0.00000746, 0.00000054, -0.00001155, -0.00001541, -0.00000630, 0.00000742, 0.00001510, 0.00001357, 0.00000644, -0.00000089, -0.00000439, -0.00000437, -0.00000559, -0.00000993, -0.00001012, 0.00000132, 0.00001656, 0.00001722, 0.00000003, -0.00001430, -0.00000830, 0.00000692, 0.00000921, -0.00000008, -0.00000127, 0.00000680, 0.00000302, -0.00001519, -0.00001822, 0.00000748, 0.00002565, 0.00000072, -0.00003664, -0.00002445, 0.00003129, 0.00005059, -0.00000336, -0.00005766, -0.00003182, 0.00003564, 0.00003912, -0.00002923, -0.00005876, 0.00000866, 0.00007651, 0.00003354, -0.00006801, -0.00007953, 0.00001885, 0.00008056, 0.00001279, -0.00007833, -0.00004773, 0.00006239, 0.00008190, -0.00002614, -0.00010336, -0.00003430, 0.00007709, 0.00006441, -0.00004962, -0.00009159, -0.00000467, 0.00006660, 0.00000309, -0.00010162, -0.00008499, 0.00002504, 0.00003377, -0.00013143, -0.00028125, -0.00018367, 0.00014463, 0.00044752, 0.00050451, 0.00032441, 0.00007401, -0.00010371, -0.00016803, -0.00014276, -0.00007238, -0.00000904, 0.00000943, -0.00001036, -0.00002837, -0.00002102, -0.00000468, -0.00000390, -0.00001561, -0.00001909, -0.00000860, 0.00000167, 0.00000112, -0.00000478, -0.00000696, -0.00000535, -0.00000426, -0.00000370, -0.00000134, 0.00000173, 0.00000262, 0.00000149, 0.00000045, -0.00000062, -0.00000316, -0.00000592, -0.00000522, },
		{ 0.00000955, 0.00000552, -0.00000091, -0.00000531, -0.00000539, -0.00000253, -0.00000008, 0.00000013, -0.00000105, -0.00000237, -0.00000363, -0.00000410, -0.00000201, 0.00000195, 0.00000343, 0.00000009, -0.00000324, -0.00000039, 0.00000586, 0.00000553, -0.00000451, -0.00001413, -0.00001180, 0.00000104, 0.00001251, 0.00001453, 0.00000913, 0.00000231, -0.00000241, -0.00000450, -0.00000535, -0.00000727, -0.00000943, -0.00000587, 0.00000611, 0.00001672, 0.00001253, -0.00000359, -0.00001235, -0.00000381, 0.00000819, 0.00000738, -0.00000045, -0.00000022, 0.00000406, -0.00000305, -0.00001540, -0.00000887, 0.00001440, 0.00001754, -0.00001378, -0.00003575, -0.00000589, 0.00004200, 0.00003556, -0.00002460, -0.00005343, -0.00000647, 0.00004450, 0.00001644, -0.00004985, -0.00004350, 0.00003995, 0.00007608, -0.00000175, -0.00008747, -0.00005265, 0.00005393, 0.00007134, -0.00002794, -0.00008881, -0.00000967, 0.00009346, 0.00005770, -0.00007288, -0.00010473, 0.00001255, 0.00010410, 0.00003178, -0.00009518, -0.00008568, 0.00004037, 0.00008079, -0.00003713, -0.00013785, -0.00006312, 0.00007130, 0.00002470, -0.00020795, -0.00033863, -0.00013305, 0.00028929, 0.00057910, 0.00052755, 0.00023599, -0.00005076, -0.00018510, -0.00017725, -0.00010463, -0.00003199, 0.00000457, -0.00000123, -0.00002331, -0.00002783, -0.00001120, 0.00000034, -0.00000874, -0.00002220, -0.00001883, -0.00000303, 0.00000510, -0.00000049, -0.00000762, -0.00000715, -0.00000353, -0.00000314, -0.00000392, -0.00000149, 0.00000250, 0.00000363, 0.00000181, -0.00000008, -0.00000161, -0.00000414, -0.00000630, -0.00000455, 0.00000162, 0.00000794, },
		{ -0.00000220, -0.00000551, -0.00000459, -0.00000164, 0.00000004, -0.00000084, -0.00000273, -0.00000372, -0.00000340, -0.00000198, 0.00000039, 0.00000225, 0.00000123, -0.00000190, -0.00000227, 0.00000248, 0.00000648, 0.00000197, -0.00000876, -0.00001374, -0.00000593, 0.00000759, 0.00001453, 0.00001149, 0.00000462, -0.00000035, -0.00000303, -0.00000494, -0.00000659, -0.00000795, -0.00000740, -0.00000150, 0.00000913, 0.00001521, 0.00000828, -0.00000534, -0.00000956, -0.00000028, 0.00000844, 0.00000563, -0.00000096, -0.00000067, 0.00000068, -0.00000634, -0.00001192, -0.00000059, 0.00001535, 0.00000641, -0.00002288, -0.00002745, 0.00001108, 0.00004298, 0.00001636, -0.00003721, -0.00003925, 0.00001508, 0.00004009, -0.00000964, -0.00005773, -0.00001737, 0.00006327, 0.00005923, -0.00003754, -0.00008929, -0.00001528, 0.00007594, 0.00004364, -0.00006488, -0.00007737, 0.00003654, 0.00010546, 0.00001363, -0.00011095, -0.00007994, 0.00006581, 0.00010761, -0.00002170, -0.00012939, -0.00005213, 0.00008999, 0.00006974, -0.00009622, -0.00015836, -0.00001197, 0.00011694, -0.00002066, -0.00031176, -0.00037255, -0.00002234, 0.00047289, 0.00068867, 0.00048596, 0.00009511, -0.00017850, -0.00022916, -0.00014703, -0.00005402, -0.00000506, 0.00000013, -0.00001697, -0.00002868, -0.00001934, -0.00000189, -0.00000085, -0.00001714, -0.00002628, -0.00001420, 0.00000348, 0.00000599, -0.00000420, -0.00000981, -0.00000549, -0.00000119, -0.00000302, -0.00000500, -0.00000136, 0.00000398, 0.00000473, 0.00000155, -0.00000115, -0.00000264, -0.00000477, -0.00000632, -0.00000372, 0.00000299, 0.00000883, 0.00000914, 0.00000408, },
		{ -0.00000372, -0.00000100, -0.00000039, -0.00000228, -0.00000427, -0.00000412, -0.00000202, 0.00000036, 0.00000167, 0.00000112, -0.00000111, -0.00000257, -0.00000019, 0.00000466, 0.00000528, -0.00000214, -0.00001104, -0.00001057, 0.00000067, 0.00001199, 0.00001360, 0.00000726, 0.00000104, -0.00000169, -0.00000340, -0.00000579, -0.00000755, -0.00000733, -0.00000443, 0.00000217, 0.00001039, 0.00001287, 0.00000506, -0.00000558, -0.00000661, 0.00000217, 0.00000791, 0.00000382, -0.00000196, -0.00000196, -0.00000186, -0.00000665, -0.00000688, 0.00000449, 0.00001146, -0.00000417, -0.00002535, -0.00001522, 0.00002265, 0.00003584, -0.00000158, -0.00003957, -0.00002033, 0.00002772, 0.00002505, -0.00003178, -0.00005140, 0.00001242, 0.00007276, 0.00003040, -0.00006483, -0.00007308, 0.00002299, 0.00007886, 0.00000440, -0.00008741, -0.00004531, 0.00007852, 0.00009233, -0.00004055, -0.00012778, -0.00003167, 0.00011034, 0.00008086, -0.00008437, -0.00013774, 0.00000612, 0.00012733, 0.00002599, -0.00016110, -0.00014690, 0.00006450, 0.00014053, -0.00011366, -0.00042620, -0.00035402, 0.00015718, 0.00067168, 0.00074229, 0.00036800, -0.00007890, -0.00027960, -0.00022273, -0.00008757, -0.00000936, -0.00000006, -0.00001684, -0.00002823, -0.00002347, -0.00000743, 0.00000189, -0.00000824, -0.00002533, -0.00002534, -0.00000636, 0.00000833, 0.00000353, -0.00000862, -0.00000996, -0.00000221, 0.00000052, -0.00000440, -0.00000636, -0.00000029, 0.00000610, 0.00000539, 0.00000046, -0.00000249, -0.00000331, -0.00000498, -0.00000619, -0.00000294, 0.00000424, 0.00000956, 0.00000857, 0.00000262, -0.00000330, -0.00000544, },
		{ -0.00000137, -0.00000389, -0.00000522, -0.00000343, 0.00000006, 0.00000221, 0.00000159, -0.00000087, -0.00000282, -0.00000187, 0.00000222, 0.00000553, 0.00000274, -0.00000558, -0.00001082, -0.00000553, 0.00000645, 0.00001361, 0.00001053, 0.00000306, -0.00000121, -0.00000221, -0.00000397, -0.00000681, -0.00000773, -0.00000556, -0.00000126, 0.00000468, 0.00001031, 0.00001047, 0.00000302, -0.00000486, -0.00000399, 0.00000356, 0.00000676, 0.00000180, -0.00000329, -0.00000316, -0.00000276, -0.00000483, -0.00000242, 0.00000568, 0.00000504, -0.00001150, -0.00002211, -0.00000298, 0.00002728, 0.00002419, -0.00001406, -0.00003321, -0.00000253, 0.00002949, 0.00000487, -0.00004440, -0.00003359, 0.00003804, 0.00006676, -0.00000270, -0.00007724, -0.00004364, 0.00005230, 0.00006223, -0.00003576, -0.00008876, -0.00000038, 0.00010412, 0.00005526, -0.00009072, -0.00011588, 0.00002935, 0.00013123, 0.00002593, -0.00013891, -0.00011142, 0.00007638, 0.00013509, -0.00004781, -0.00021140, -0.00009330, 0.00014990, 0.00011831, -0.00025316, -0.00052131, -0.00025650, 0.00039700, 0.00084805, 0.00071027, 0.00017956, -0.00025359, -0.00032761, -0.00016648, -0.00001882, 0.00001306, -0.00001656, -0.00003544, -0.00002771, -0.00001010, 0.00000133, -0.00000243, -0.00001866, -0.00002926, -0.00001906, 0.00000188, 0.00000947, -0.00000148, -0.00001176, -0.00000747, 0.00000154, 0.00000048, -0.00000705, -0.00000706, 0.00000208, 0.00000834, 0.00000503, -0.00000138, -0.00000360, -0.00000338, -0.00000497, -0.00000622, -0.00000228, 0.00000554, 0.00001027, 0.00000783, 0.00000109, -0.00000424, -0.00000514, -0.00000284, -0.00000077, },
		{ -0.00000527, -0.00000179, 0.00000221, 0.00000305, 0.00000030, -0.00000299, -0.00000345, -0.00000018, 0.00000421, 0.00000500, -0.00000027, -0.00000746, -0.00000833, 0.00000005, 0.00001029, 0.00001263, 0.00000647, -0.00000024, -0.00000227, -0.00000250, -0.00000484, -0.00000755, -0.00000689, -0.00000305, 0.00000148, 0.00000595, 0.00000947, 0.00000853, 0.00000203, -0.00000374, -0.00000203, 0.00000399, 0.00000507, -0.00000033, -0.00000449, -0.00000350, -0.00000202, -0.00000232, 0.00000000, 0.00000372, -0.00000133, -0.00001451, -0.00001554, 0.00000627, 0.00002579, 0.00001223, -0.00001930, -0.00002196, 0.00000954, 0.00002190, -0.00001417, -0.00004527, -0.00001005, 0.00005344, 0.00004812, -0.00003157, -0.00007291, -0.00000936, 0.00006565, 0.00003132, -0.00006590, -0.00006838, 0.00004544, 0.00010544, 0.00000284, -0.00012272, -0.00007597, 0.00008697, 0.00011899, -0.00004555, -0.00016759, -0.00005161, 0.00013883, 0.00010200, -0.00013724, -0.00022540, 0.00000043, 0.00021742, 0.00003262, -0.00042083, -0.00055935, -0.00006588, 0.00066845, 0.00095819, 0.00057724, -0.00005383, -0.00039107, -0.00030790, -0.00007601, 0.00003566, 0.00000511, -0.00004516, -0.00004385, -0.00001375, 0.00000449, 0.00000131, -0.00001346, -0.00002699, -0.00002655, -0.00000969, 0.00000726, 0.00000644, -0.00000694, -0.00001201, -0.00000298, 0.00000414, -0.00000176, -0.00001002, -0.00000620, 0.00000548, 0.00000985, 0.00000339, -0.00000345, -0.00000395, -0.00000291, -0.00000517, -0.00000660, -0.00000153, 0.00000711, 0.00001096, 0.00000679, -0.00000060, -0.00000496, -0.00000457, -0.00000210, -0.00000105, -0.00000277, -0.00000529, },
		{ 0.00000383, 0.00000268, -0.00000171, -0.00000461, -0.00000292, 0.00000191, 0.00000528, 0.00000341, -0.00000282, -0.00000738, -0.00000434, 0.00000489, 0.00001171, 0.00000978, 0.00000248, -0.00000230, -0.00000257, -0.00000294, -0.00000582, -0.00000762, -0.00000511, -0.00000036, 0.00000338, 0.00000622, 0.00000843, 0.00000724, 0.00000174, -0.00000268, -0.00000085, 0.00000360, 0.00000304, -0.00000228, -0.00000501, -0.00000267, -0.00000037, -0.00000051, 0.00000004, 0.00000021, -0.00000579, -0.00001373, -0.00000841, 0.00001118, 0.00002065, 0.00000326, -0.00001807, -0.00001035, 0.00001381, 0.00000898, -0.00002703, -0.00003585, 0.00001249, 0.00005611, 0.00002297, -0.00004968, -0.00005490, 0.00002054, 0.00006102, -0.00000462, -0.00007826, -0.00003221, 0.00007967, 0.00008161, -0.00005126, -0.00012703, -0.00001749, 0.00012449, 0.00007348, -0.00011515, -0.00015832, 0.00002926, 0.00017227, 0.00002814, -0.00021865, -0.00018773, 0.00011720, 0.00023637, -0.00011961, -0.00058162, -0.00050469, 0.00021115, 0.00092558, 0.00096347, 0.00034970, -0.00029097, -0.00045835, -0.00022332, 0.00002193, 0.00005646, -0.00002988, -0.00007085, -0.00003431, 0.00000856, 0.00001177, -0.00000926, -0.00002607, -0.00002855, -0.00001790, -0.00000116, 0.00000769, 0.00000081, -0.00001029, -0.00000899, 0.00000177, 0.00000427, -0.00000567, -0.00001195, -0.00000345, 0.00000906, 0.00000987, 0.00000073, -0.00000499, -0.00000334, -0.00000237, -0.00000599, -0.00000722, -0.00000033, 0.00000909, 0.00001143, 0.00000525, -0.00000239, -0.00000534, -0.00000379, -0.00000165, -0.00000189, -0.00000432, -0.00000607, -0.00000432, 0.00000037, },
		{ -0.00000384, -0.00000533, -0.00000150, 0.00000382, 0.00000534, 0.00000142, -0.00000424, -0.00000556, 0.00000005, 0.00000810, 0.00001089, 0.00000606, -0.00000068, -0.00000325, -0.00000262, -0.00000365, -0.00000656, -0.00000680, -0.00000269, 0.00000199, 0.00000434, 0.00000589, 0.00000757, 0.00000654, 0.00000173, -0.00000197, -0.00000044, 0.00000256, 0.00000094, -0.00000365, -0.00000451, -0.00000095, 0.00000119, -0.00000021, -0.00000171, -0.00000312, -0.00000760, -0.00001076, -0.00000284, 0.00001221, 0.00001480, -0.00000123, -0.00001301, -0.00000206, 0.00001101, -0.00000435, -0.00003130, -0.00002037, 0.00002851, 0.00004750, -0.00000147, -0.00005428, -0.00002981, 0.00003897, 0.00004180, -0.00003540, -0.00007064, 0.00000926, 0.00009335, 0.00003927, -0.00009256, -0.00010213, 0.00004394, 0.00013014, 0.00000479, -0.00016303, -0.00010949, 0.00011015, 0.00016131, -0.00007281, -0.00026560, -0.00009655, 0.00022681, 0.00018188, -0.00032124, -0.00069000, -0.00033553, 0.00054353, 0.00111433, 0.00084236, 0.00005804, -0.00048439, -0.00043752, -0.00009493, 0.00009729, 0.00003520, -0.00007786, -0.00007875, -0.00000678, 0.00002909, 0.00000547, -0.00002700, -0.00003351, -0.00002142, -0.00000704, 0.00000265, 0.00000343, -0.00000440, -0.00000986, -0.00000395, 0.00000474, 0.00000166, -0.00000980, -0.00001179, 0.00000067, 0.00001166, 0.00000819, -0.00000214, -0.00000537, -0.00000202, -0.00000242, -0.00000759, -0.00000761, 0.00000171, 0.00001134, 0.00001129, 0.00000310, -0.00000408, -0.00000523, -0.00000290, -0.00000169, -0.00000319, -0.00000564, -0.00000597, -0.00000255, 0.00000249, 0.00000450, 0.00000128, },
		{ 0.00000039, 0.00000517, 0.00000461, -0.00000032, -0.00000425, -0.00000261, 0.00000386, 0.00000932, 0.00000845, 0.00000239, -0.00000274, -0.00000346, -0.00000279, -0.00000449, -0.00000672, -0.00000511, -0.00000008, 0.00000366, 0.00000451, 0.00000535, 0.00000707, 0.00000621, 0.00000168, -0.00000178, -0.00000070, 0.00000117, -0.00000084, -0.00000407, -0.00000302, 0.00000099, 0.00000178, -0.00000146, -0.00000408, -0.00000510, -0.00000720, -0.00000744, 0.00000028, 0.00001099, 0.00001055, -0.00000167, -0.00000739, 0.00000119, 0.00000396, -0.00001402, -0.00002764, -0.00000410, 0.00003536, 0.00003207, -0.00001928, -0.00004691, -0.00000543, 0.00004301, 0.00001532, -0.00005320, -0.00004696, 0.00004443, 0.00008390, -0.00000959, -0.00011042, -0.00005526, 0.00009129, 0.00010124, -0.00006923, -0.00017441, -0.00003166, 0.00016738, 0.00010246, -0.00017605, -0.00025753, 0.00003281, 0.00029311, 0.00004480, -0.00053502, -0.00070123, -0.00005374, 0.00087911, 0.00118579, 0.00059920, -0.00024843, -0.00059276, -0.00033213, 0.00004314, 0.00012612, -0.00002210, -0.00011871, -0.00006014, 0.00003040, 0.00003672, -0.00001520, -0.00004424, -0.00003035, -0.00000778, 0.00000067, -0.00000008, -0.00000278, -0.00000613, -0.00000581, 0.00000068, 0.00000455, -0.00000268, -0.00001248, -0.00000934, 0.00000493, 0.00001242, 0.00000530, -0.00000423, -0.00000443, -0.00000075, -0.00000360, -0.00000966, -0.00000715, 0.00000470, 0.00001339, 0.00001017, 0.00000046, -0.00000535, -0.00000458, -0.00000216, -0.00000234, -0.00000470, -0.00000635, -0.00000491, -0.00000037, 0.00000404, 0.00000408, -0.00000075, -0.00000556, -0.00000506, },
		{ 0.00000351, -0.00000134, -0.00000303, 0.00000064, 0.00000638, 0.00000867, 0.00000523, -0.00000062, -0.00000375, -0.00000337, -0.00000320, -0.00000517, -0.00000609, -0.00000279, 0.00000229, 0.00000453, 0.00000412, 0.00000487, 0.00000688, 0.00000598, 0.00000136, -0.00000210, -0.00000144, -0.00000026, -0.00000194, -0.00000344, -0.00000096, 0.00000236, 0.00000100, -0.00000366, -0.00000591, -0.00000538, -0.00000569, -0.00000510, 0.00000131, 0.00000948, 0.00000886, 0.00000010, -0.00000380, -0.00000018, -0.00000374, -0.00001801, -0.00001898, 0.00000850, 0.00003356, 0.00001541, -0.00002758, -0.00003234, 0.00001188, 0.00003444, -0.00000983, -0.00005493, -0.00001559, 0.00006436, 0.00005572, -0.00005150, -0.00010145, -0.00000016, 0.00011159, 0.00004542, -0.00012837, -0.00014470, 0.00005489, 0.00018231, 0.00000681, -0.00025238, -0.00018772, 0.00016961, 0.00028466, -0.00016172, -0.00071020, -0.00058465, 0.00031070, 0.00115488, 0.00110997, 0.00026651, -0.00051250, -0.00059263, -0.00016754, 0.00015329, 0.00009845, -0.00009650, -0.00013347, -0.00001634, 0.00006318, 0.00002405, -0.00004440, -0.00005200, -0.00001530, 0.00000659, 0.00000064, -0.00000801, -0.00000678, -0.00000286, -0.00000030, 0.00000243, 0.00000134, -0.00000675, -0.00001259, -0.00000551, 0.00000792, 0.00001126, 0.00000232, -0.00000481, -0.00000267, -0.00000041, -0.00000599, -0.00001143, -0.00000531, 0.00000834, 0.00001460, 0.00000792, -0.00000227, -0.00000587, -0.00000354, -0.00000187, -0.00000355, -0.00000602, -0.00000615, -0.00000309, 0.00000168, 0.00000466, 0.00000273, -0.00000289, -0.00000652, -0.00000396, 0.00000232, 0.00000584, },
		{ -0.00000101, 0.00000348, 0.00000734, 0.00000667, 0.00000195, -0.00000267, -0.00000402, -0.00000331, -0.00000376, -0.00000538, -0.00000463, -0.00000019, 0.00000409, 0.00000462, 0.00000346, 0.00000460, 0.00000688, 0.00000563, 0.00000066, -0.00000286, -0.00000241, -0.00000134, -0.00000212, -0.00000195, 0.00000100, 0.00000259, -0.00000096, -0.00000588, -0.00000652, -0.00000441, -0.00000431, -0.00000416, 0.00000143, 0.00000910, 0.00000930, 0.00000186, -0.00000335, -0.00000416, -0.00000917, -0.00001662, -0.00000917, 0.00001515, 0.00002618, 0.00000228, -0.00002692, -0.00001657, 0.00001908, 0.00001857, -0.00002672, -0.00004276, 0.00001365, 0.00006550, 0.00001848, -0.00007591, -0.00007047, 0.00004730, 0.00010012, -0.00002156, -0.00015648, -0.00008159, 0.00012611, 0.00014778, -0.00010208, -0.00027729, -0.00006770, 0.00027497, 0.00018557, -0.00040171, -0.00079473, -0.00033490, 0.00070301, 0.00131186, 0.00088591, -0.00010058, -0.00068337, -0.00048572, 0.00001531, 0.00020544, 0.00002190, -0.00016214, -0.00011129, 0.00004097, 0.00007694, -0.00000891, -0.00007132, -0.00004372, 0.00000759, 0.00001439, -0.00000858, -0.00001672, -0.00000489, 0.00000438, 0.00000332, 0.00000006, -0.00000313, -0.00000856, -0.00001032, -0.00000197, 0.00000880, 0.00000903, 0.00000046, -0.00000390, -0.00000115, -0.00000163, -0.00000904, -0.00001203, -0.00000199, 0.00001189, 0.00001442, 0.00000477, -0.00000454, -0.00000551, -0.00000245, -0.00000227, -0.00000510, -0.00000672, -0.00000498, -0.00000092, 0.00000310, 0.00000426, 0.00000085, -0.00000467, -0.00000662, -0.00000233, 0.00000398, 0.00000594, 0.00000246, -0.00000148, },
		{ 0.00000683, 0.00000391, -0.00000084, -0.00000378, -0.00000392, -0.00000343, -0.00000426, -0.00000491, -0.00000248, 0.00000232, 0.00000514, 0.00000407, 0.00000272, 0.00000454, 0.00000691, 0.00000502, -0.00000040, -0.00000388, -0.00000328, -0.00000178, -0.00000136, -0.00000008, 0.00000222, 0.00000152, -0.00000346, -0.00000731, -0.00000585, -0.00000308, -0.00000381, -0.00000420, 0.00000190, 0.00001028, 0.00001050, 0.00000196, -0.00000561, -0.00000824, -0.00001091, -0.00001192, -0.00000145, 0.00001602, 0.00001724, -0.00000487, -0.00002055, -0.00000464, 0.00001696, 0.00000205, -0.00003216, -0.00002289, 0.00003286, 0.00005041, -0.00001639, -0.00007846, -0.00002861, 0.00007437, 0.00006188, -0.00008087, -0.00014670, -0.00000289, 0.00016174, 0.00007128, -0.00019479, -0.00023944, 0.00007408, 0.00031290, 0.00000312, -0.00062292, -0.00074962, 0.00002271, 0.00105439, 0.00131045, 0.00054459, -0.00043662, -0.00072996, -0.00029929, 0.00017098, 0.00018616, -0.00008047, -0.00019484, -0.00005393, 0.00009334, 0.00006230, -0.00005386, -0.00008398, -0.00001855, 0.00002990, 0.00000998, -0.00002416, -0.00002045, 0.00000417, 0.00001206, 0.00000204, -0.00000557, -0.00000609, -0.00000712, -0.00000719, -0.00000035, 0.00000783, 0.00000722, 0.00000038, -0.00000235, -0.00000096, -0.00000432, -0.00001169, -0.00001082, 0.00000230, 0.00001446, 0.00001267, 0.00000131, -0.00000585, -0.00000445, -0.00000177, -0.00000340, -0.00000654, -0.00000646, -0.00000303, 0.00000104, 0.00000358, 0.00000302, -0.00000110, -0.00000579, -0.00000595, -0.00000051, 0.00000524, 0.00000569, 0.00000174, -0.00000083, 0.00000125, 0.00000538, },
		{ -0.00000285, -0.00000414, -0.00000372, -0.00000372, -0.00000443, -0.00000364, 0.00000010, 0.00000441, 0.00000534, 0.00000305, 0.00000207, 0.00000466, 0.00000685, 0.00000409, -0.00000175, -0.00000494, -0.00000375, -0.00000141, 0.00000007, 0.00000157, 0.00000230, -0.00000054, -0.00000573, -0.00000752, -0.00000438, -0.00000221, -0.00000430, -0.00000432, 0.00000353, 0.00001246, 0.00001095, -0.00000014, -0.00000904, -0.00001048, -0.00000930, -0.00000660, 0.00000257, 0.00001322, 0.00001016, -0.00000625, -0.00001279, 0.00000092, 0.00000919, -0.00000966, -0.00002726, -0.00000295, 0.00003843, 0.00002632, -0.00003958, -0.00006197, 0.00001060, 0.00007537, 0.00000968, -0.00011665, -0.00010370, 0.00006898, 0.00015195, -0.00002676, -0.00024505, -0.00014560, 0.00019921, 0.00026182, -0.00023109, -0.00076930, -0.00056125, 0.00043499, 0.00129788, 0.00114184, 0.00014320, -0.00068162, -0.00064931, -0.00007914, 0.00026225, 0.00010257, -0.00017720, -0.00018030, 0.00002424, 0.00012190, 0.00001926, -0.00009642, -0.00007400, 0.00001752, 0.00004179, -0.00000790, -0.00003955, -0.00001453, 0.00001817, 0.00001549, -0.00000515, -0.00001154, -0.00000512, -0.00000326, -0.00000549, -0.00000135, 0.00000643, 0.00000723, 0.00000173, -0.00000156, -0.00000268, -0.00000761, -0.00001284, -0.00000782, 0.00000655, 0.00001538, 0.00000971, -0.00000170, -0.00000601, -0.00000316, -0.00000186, -0.00000502, -0.00000735, -0.00000514, -0.00000077, 0.00000227, 0.00000303, 0.00000135, -0.00000267, -0.00000613, -0.00000475, 0.00000127, 0.00000609, 0.00000530, 0.00000147, 0.00000035, 0.00000327, 0.00000614, 0.00000516, 0.00000100, },
		{ -0.00000364, -0.00000403, -0.00000402, -0.00000162, 0.00000278, 0.00000576, 0.00000470, 0.00000173, 0.00000161, 0.00000493, 0.00000660, 0.00000283, -0.00000327, -0.00000578, -0.00000358, -0.00000027, 0.00000171, 0.00000245, 0.00000119, -0.00000299, -0.00000712, -0.00000660, -0.00000284, -0.00000226, -0.00000529, -0.00000363, 0.00000635, 0.00001452, 0.00000967, -0.00000382, -0.00001186, -0.00001010, -0.00000593, -0.00000284, 0.00000311, 0.00000954, 0.00000665, -0.00000409, -0.00000721, 0.00000053, 0.00000053, -0.00001395, -0.00001641, 0.00001081, 0.00003180, 0.00000221, -0.00004676, -0.00003488, 0.00003590, 0.00005314, -0.00004030, -0.00012106, -0.00004195, 0.00011421, 0.00010066, -0.00011911, -0.00023839, -0.00001991, 0.00027025, 0.00012269, -0.00046723, -0.00079586, -0.00024774, 0.00083173, 0.00138495, 0.00083173, -0.00024774, -0.00079586, -0.00046723, 0.00012269, 0.00027025, -0.00001991, -0.00023839, -0.00011911, 0.00010066, 0.00011421, -0.00004195, -0.00012106, -0.00004030, 0.00005314, 0.00003590, -0.00003488, -0.00004676, 0.00000221, 0.00003180, 0.00001081, -0.00001641, -0.00001395, 0.00000053, 0.00000053, -0.00000721, -0.00000409, 0.00000665, 0.00000954, 0.00000311, -0.00000284, -0.00000593, -0.00001010, -0.00001186, -0.00000382, 0.00000967, 0.00001452, 0.00000635, -0.00000363, -0.00000529, -0.00000226, -0.00000284, -0.00000660, -0.00000712, -0.00000299, 0.00000119, 0.00000245, 0.00000171, -0.00000027, -0.00000358, -0.00000578, -0.00000327, 0.00000283, 0.00000660, 0.00000493, 0.00000161, 0.00000173, 0.00000470, 0.00000576, 0.00000278, -0.00000162, -0.00000402, -0.00000403, },
		{ -0.00000285, 0.00000100, 0.00000516, 0.00000614, 0.00000327, 0.00000035, 0.00000147, 0.00000530, 0.00000609, 0.00000127, -0.00000475, -0.00000613, -0.00000267, 0.00000135, 0.00000303, 0.00000227, -0.00000077, -0.00000514, -0.00000735, -0.00000502, -0.00000186, -0.00000316, -0.00000601, -0.00000170, 0.00000971, 0.00001538, 0.00000655, -0.00000782, -0.00001284, -0.00000761, -0.00000268, -0.00000156, 0.00000173, 0.00000723, 0.00000643, -0.00000135, -0.00000549, -0.00000326, -0.00000512, -0.00001154, -0.00000515, 0.00001549, 0.00001817, -0.00001453, -0.00003955, -0.00000790, 0.00004179, 0.00001752, -0.00007400, -0.00009642, 0.00001926, 0.00012190, 0.00002424, -0.00018030, -0.00017720, 0.00010257, 0.00026225, -0.00007914, -0.00064931, -0.00068162, 0.00014320, 0.00114184, 0.00129788, 0.00043499, -0.00056125, -0.00076930, -0.00023109, 0.00026182, 0.00019921, -0.00014560, -0.00024505, -0.00002676, 0.00015195, 0.00006898, -0.00010370, -0.00011665, 0.00000968, 0.00007537, 0.00001060, -0.00006197, -0.00003958, 0.00002632, 0.00003843, -0.00000295, -0.00002726, -0.00000966, 0.00000919, 0.00000092, -0.00001279, -0.00000625, 0.00001016, 0.00001322, 0.00000257, -0.00000660, -0.00000930, -0.00001048, -0.00000904, -0.00000014, 0.00001095, 0.00001246, 0.00000353, -0.00000432, -0.00000430, -0.00000221, -0.00000438, -0.00000752, -0.00000573, -0.00000054, 0.00000230, 0.00000157, 0.00000007, -0.00000141, -0.00000375, -0.00000494, -0.00000175, 0.00000409, 0.00000685, 0.00000466, 0.00000207, 0.00000305, 0.00000534, 0.00000441, 0.00000010, -0.00000364, -0.00000443, -0.00000372, -0.00000372, -0.00000414, },
		{ 0.00000683, 0.00000538, 0.00000125, -0.00000083, 0.00000174, 0.00000569, 0.00000524, -0.00000051, -0.00000595, -0.00000579, -0.00000110, 0.00000302, 0.00000358, 0.00000104, -0.00000303, -0.00000646, -0.00000654, -0.00000340, -0.00000177, -0.00000445, -0.00000585, 0.00000131, 0.00001267, 0.00001446, 0.00000230, -0.00001082, -0.00001169, -0.00000432, -0.00000096, -0.00000235, 0.00000038, 0.00000722, 0.00000783, -0.00000035, -0.00000719, -0.00000712, -0.00000609, -0.00000557, 0.00000204, 0.00001206, 0.00000417, -0.00002045, -0.00002416, 0.00000998, 0.00002990, -0.00001855, -0.00008398, -0.00005386, 0.00006230, 0.00009334, -0.00005393, -0.00019484, -0.00008047, 0.00018616, 0.00017098, -0.00029929, -0.00072996, -0.00043662, 0.00054459, 0.00131045, 0.00105439, 0.00002271, -0.00074962, -0.00062292, 0.00000312, 0.00031290, 0.00007408, -0.00023944, -0.00019479, 0.00007128, 0.00016174, -0.00000289, -0.00014670, -0.00008087, 0.00006188, 0.00007437, -0.00002861, -0.00007846, -0.00001639, 0.00005041, 0.00003286, -0.00002289, -0.00003216, 0.00000205, 0.00001696, -0.00000464, -0.00002055, -0.00000487, 0.00001724, 0.00001602, -0.00000145, -0.00001192, -0.00001091, -0.00000824, -0.00000561, 0.00000196, 0.00001050, 0.00001028, 0.00000190, -0.00000420, -0.00000381, -0.00000308, -0.00000585, -0.00000731, -0.00000346, 0.00000152, 0.00000222, -0.00000008, -0.00000136, -0.00000178, -0.00000328, -0.00000388, -0.00000040, 0.00000502, 0.00000691, 0.00000454, 0.00000272, 0.00000407, 0.00000514, 0.00000232, -0.00000248, -0.00000491, -0.00000426, -0.00000343, -0.00000392, -0.00000378, -0.00000084, 0.00000391, },
		{ -0.00000101, -0.00000148, 0.00000246, 0.00000594, 0.00000398, -0.00000233, -0.00000662, -0.00000467, 0.00000085, 0.00000426, 0.00000310, -0.00000092, -0.00000498, -0.00000672, -0.00000510, -0.00000227, -0.00000245, -0.00000551, -0.00000454, 0.00000477, 0.00001442, 0.00001189, -0.00000199, -0.00001203, -0.00000904, -0.00000163, -0.00000115, -0.00000390, 0.00000046, 0.00000903, 0.00000880, -0.00000197, -0.00001032, -0.00000856, -0.00000313, 0.00000006, 0.00000332, 0.00000438, -0.00000489, -0.00001672, -0.00000858, 0.00001439, 0.00000759, -0.00004372, -0.00007132, -0.00000891, 0.00007694, 0.00004097, -0.00011129, -0.00016214, 0.00002190, 0.00020544, 0.00001531, -0.00048572, -0.00068337, -0.00010058, 0.00088591, 0.00131186, 0.00070301, -0.00033490, -0.00079473, -0.00040171, 0.00018557, 0.00027497, -0.00006770, -0.00027729, -0.00010208, 0.00014778, 0.00012611, -0.00008159, -0.00015648, -0.00002156, 0.00010012, 0.00004730, -0.00007047, -0.00007591, 0.00001848, 0.00006550, 0.00001365, -0.00004276, -0.00002672, 0.00001857, 0.00001908, -0.00001657, -0.00002692, 0.00000228, 0.00002618, 0.00001515, -0.00000917, -0.00001662, -0.00000917, -0.00000416, -0.00000335, 0.00000186, 0.00000930, 0.00000910, 0.00000143, -0.00000416, -0.00000431, -0.00000441, -0.00000652, -0.00000588, -0.00000096, 0.00000259, 0.00000100, -0.00000195, -0.00000212, -0.00000134, -0.00000241, -0.00000286, 0.00000066, 0.00000563, 0.00000688, 0.00000460, 0.00000346, 0.00000462, 0.00000409, -0.00000019, -0.00000463, -0.00000538, -0.00000376, -0.00000331, -0.00000402, -0.00000267, 0.00000195, 0.00000667, 0.00000734, 0.00000348, },
		{ 0.00000351, 0.00000584, 0.00000232, -0.00000396, -0.00000652, -0.00000289, 0.00000273, 0.00000466, 0.00000168, -0.00000309, -0.00000615, -0.00000602, -0.00000355, -0.00000187, -0.00000354, -0.00000587, -0.00000227, 0.00000792, 0.00001460, 0.00000834, -0.00000531, -0.00001143, -0.00000599, -0.00000041, -0.00000267, -0.00000481, 0.00000232, 0.00001126, 0.00000792, -0.00000551, -0.00001259, -0.00000675, 0.00000134, 0.00000243, -0.00000030, -0.00000286, -0.00000678, -0.00000801, 0.00000064, 0.00000659, -0.00001530, -0.00005200, -0.00004440, 0.00002405, 0.00006318, -0.00001634, -0.00013347, -0.00009650, 0.00009845, 0.00015329, -0.00016754, -0.00059263, -0.00051250, 0.00026651, 0.00110997, 0.00115488, 0.00031070, -0.00058465, -0.00071020, -0.00016172, 0.00028466, 0.00016961, -0.00018772, -0.00025238, 0.00000681, 0.00018231, 0.00005489, -0.00014470, -0.00012837, 0.00004542, 0.00011159, -0.00000016, -0.00010145, -0.00005150, 0.00005572, 0.00006436, -0.00001559, -0.00005493, -0.00000983, 0.00003444, 0.00001188, -0.00003234, -0.00002758, 0.00001541, 0.00003356, 0.00000850, -0.00001898, -0.00001801, -0.00000374, -0.00000018, -0.00000380, 0.00000010, 0.00000886, 0.00000948, 0.00000131, -0.00000510, -0.00000569, -0.00000538, -0.00000591, -0.00000366, 0.00000100, 0.00000236, -0.00000096, -0.00000344, -0.00000194, -0.00000026, -0.00000144, -0.00000210, 0.00000136, 0.00000598, 0.00000688, 0.00000487, 0.00000412, 0.00000453, 0.00000229, -0.00000279, -0.00000609, -0.00000517, -0.00000320, -0.00000337, -0.00000375, -0.00000062, 0.00000523, 0.00000867, 0.00000638, 0.00000064, -0.00000303, -0.00000134, },
		{ 0.00000039, -0.00000506, -0.00000556, -0.00000075, 0.00000408, 0.00000404, -0.00000037, -0.00000491, -0.00000635, -0.00000470, -0.00000234, -0.00000216, -0.00000458, -0.00000535, 0.00000046, 0.00001017, 0.00001339, 0.00000470, -0.00000715, -0.00000966, -0.00000360, -0.00000075, -0.00000443, -0.00000423, 0.00000530, 0.00001242, 0.00000493, -0.00000934, -0.00001248, -0.00000268, 0.00000455, 0.00000068, -0.00000581, -0.00000613, -0.00000278, -0.00000008, 0.00000067, -0.00000778, -0.00003035, -0.00004424, -0.00001520, 0.00003672, 0.00003040, -0.00006014, -0.00011871, -0.00002210, 0.00012612, 0.00004314, -0.00033213, -0.00059276, -0.00024843, 0.00059920, 0.00118579, 0.00087911, -0.00005374, -0.00070123, -0.00053502, 0.00004480, 0.00029311, 0.00003281, -0.00025753, -0.00017605, 0.00010246, 0.00016738, -0.00003166, -0.00017441, -0.00006923, 0.00010124, 0.00009129, -0.00005526, -0.00011042, -0.00000959, 0.00008390, 0.00004443, -0.00004696, -0.00005320, 0.00001532, 0.00004301, -0.00000543, -0.00004691, -0.00001928, 0.00003207, 0.00003536, -0.00000410, -0.00002764, -0.00001402, 0.00000396, 0.00000119, -0.00000739, -0.00000167, 0.00001055, 0.00001099, 0.00000028, -0.00000744, -0.00000720, -0.00000510, -0.00000408, -0.00000146, 0.00000178, 0.00000099, -0.00000302, -0.00000407, -0.00000084, 0.00000117, -0.00000070, -0.00000178, 0.00000168, 0.00000621, 0.00000707, 0.00000535, 0.00000451, 0.00000366, -0.00000008, -0.00000511, -0.00000672, -0.00000449, -0.00000279, -0.00000346, -0.00000274, 0.00000239, 0.00000845, 0.00000932, 0.00000386, -0.00000261, -0.00000425, -0.00000032, 0.00000461, 0.00000517, },
		{ -0.00000384, 0.00000128, 0.00000450, 0.00000249, -0.00000255, -0.00000597, -0.00000564, -0.00000319, -0.00000169, -0.00000290, -0.00000523, -0.00000408, 0.00000310, 0.00001129, 0.00001134, 0.00000171, -0.00000761, -0.00000759, -0.00000242, -0.00000202, -0.00000537, -0.00000214, 0.00000819, 0.00001166, 0.00000067, -0.00001179, -0.00000980, 0.00000166, 0.00000474, -0.00000395, -0.00000986, -0.00000440, 0.00000343, 0.00000265, -0.00000704, -0.00002142, -0.00003351, -0.00002700, 0.00000547, 0.00002909, -0.00000678, -0.00007875, -0.00007786, 0.00003520, 0.00009729, -0.00009493, -0.00043752, -0.00048439, 0.00005804, 0.00084236, 0.00111433, 0.00054353, -0.00033553, -0.00069000, -0.00032124, 0.00018188, 0.00022681, -0.00009655, -0.00026560, -0.00007281, 0.00016131, 0.00011015, -0.00010949, -0.00016303, 0.00000479, 0.00013014, 0.00004394, -0.00010213, -0.00009256, 0.00003927, 0.00009335, 0.00000926, -0.00007064, -0.00003540, 0.00004180, 0.00003897, -0.00002981, -0.00005428, -0.00000147, 0.00004750, 0.00002851, -0.00002037, -0.00003130, -0.00000435, 0.00001101, -0.00000206, -0.00001301, -0.00000123, 0.00001480, 0.00001221, -0.00000284, -0.00001076, -0.00000760, -0.00000312, -0.00000171, -0.00000021, 0.00000119, -0.00000095, -0.00000451, -0.00000365, 0.00000094, 0.00000256, -0.00000044, -0.00000197, 0.00000173, 0.00000654, 0.00000757, 0.00000589, 0.00000434, 0.00000199, -0.00000269, -0.00000680, -0.00000656, -0.00000365, -0.00000262, -0.00000325, -0.00000068, 0.00000606, 0.00001089, 0.00000810, 0.00000005, -0.00000556, -0.00000424, 0.00000142, 0.00000534, 0.00000382, -0.00000150, -0.00000533, },
		{ 0.00000383, 0.00000037, -0.00000432, -0.00000607, -0.00000432, -0.00000189, -0.00000165, -0.00000379, -0.00000534, -0.00000239, 0.00000525, 0.00001143, 0.00000909, -0.00000033, -0.00000722, -0.00000599, -0.00000237, -0.00000334, -0.00000499, 0.00000073, 0.00000987, 0.00000906, -0.00000345, -0.00001195, -0.00000567, 0.00000427, 0.00000177, -0.00000899, -0.00001029, 0.00000081, 0.00000769, -0.00000116, -0.00001790, -0.00002855, -0.00002607, -0.00000926, 0.00001177, 0.00000856, -0.00003431, -0.00007085, -0.00002988, 0.00005646, 0.00002193, -0.00022332, -0.00045835, -0.00029097, 0.00034970, 0.00096347, 0.00092558, 0.00021115, -0.00050469, -0.00058162, -0.00011961, 0.00023637, 0.00011720, -0.00018773, -0.00021865, 0.00002814, 0.00017227, 0.00002926, -0.00015832, -0.00011515, 0.00007348, 0.00012449, -0.00001749, -0.00012703, -0.00005126, 0.00008161, 0.00007967, -0.00003221, -0.00007826, -0.00000462, 0.00006102, 0.00002054, -0.00005490, -0.00004968, 0.00002297, 0.00005611, 0.00001249, -0.00003585, -0.00002703, 0.00000898, 0.00001381, -0.00001035, -0.00001807, 0.00000326, 0.00002065, 0.00001118, -0.00000841, -0.00001373, -0.00000579, 0.00000021, 0.00000004, -0.00000051, -0.00000037, -0.00000267, -0.00000501, -0.00000228, 0.00000304, 0.00000360, -0.00000085, -0.00000268, 0.00000174, 0.00000724, 0.00000843, 0.00000622, 0.00000338, -0.00000036, -0.00000511, -0.00000762, -0.00000582, -0.00000294, -0.00000257, -0.00000230, 0.00000248, 0.00000978, 0.00001171, 0.00000489, -0.00000434, -0.00000738, -0.00000282, 0.00000341, 0.00000528, 0.00000191, -0.00000292, -0.00000461, -0.00000171, 0.00000268, },
		{ -0.00000527, -0.00000529, -0.00000277, -0.00000105, -0.00000210, -0.00000457, -0.00000496, -0.00000060, 0.00000679, 0.00001096, 0.00000711, -0.00000153, -0.00000660, -0.00000517, -0.00000291, -0.00000395, -0.00000345, 0.00000339, 0.00000985, 0.00000548, -0.00000620, -0.00001002, -0.00000176, 0.00000414, -0.00000298, -0.00001201, -0.00000694, 0.00000644, 0.00000726, -0.00000969, -0.00002655, -0.00002699, -0.00001346, 0.00000131, 0.00000449, -0.00001375, -0.00004385, -0.00004516, 0.00000511, 0.00003566, -0.00007601, -0.00030790, -0.00039107, -0.00005383, 0.00057724, 0.00095819, 0.00066845, -0.00006588, -0.00055935, -0.00042083, 0.00003262, 0.00021742, 0.00000043, -0.00022540, -0.00013724, 0.00010200, 0.00013883, -0.00005161, -0.00016759, -0.00004555, 0.00011899, 0.00008697, -0.00007597, -0.00012272, 0.00000284, 0.00010544, 0.00004544, -0.00006838, -0.00006590, 0.00003132, 0.00006565, -0.00000936, -0.00007291, -0.00003157, 0.00004812, 0.00005344, -0.00001005, -0.00004527, -0.00001417, 0.00002190, 0.00000954, -0.00002196, -0.00001930, 0.00001223, 0.00002579, 0.00000627, -0.00001554, -0.00001451, -0.00000133, 0.00000372, 0.00000000, -0.00000232, -0.00000202, -0.00000350, -0.00000449, -0.00000033, 0.00000507, 0.00000399, -0.00000203, -0.00000374, 0.00000203, 0.00000853, 0.00000947, 0.00000595, 0.00000148, -0.00000305, -0.00000689, -0.00000755, -0.00000484, -0.00000250, -0.00000227, -0.00000024, 0.00000647, 0.00001263, 0.00001029, 0.00000005, -0.00000833, -0.00000746, -0.00000027, 0.00000500, 0.00000421, -0.00000018, -0.00000345, -0.00000299, 0.00000030, 0.00000305, 0.00000221, -0.00000179, },
		{ -0.00000137, -0.00000077, -0.00000284, -0.00000514, -0.00000424, 0.00000109, 0.00000783, 0.00001027, 0.00000554, -0.00000228, -0.00000622, -0.00000497, -0.00000338, -0.00000360, -0.00000138, 0.00000503, 0.00000834, 0.00000208, -0.00000706, -0.00000705, 0.00000048, 0.00000154, -0.00000747, -0.00001176, -0.00000148, 0.00000947, 0.00000188, -0.00001906, -0.00002926, -0.00001866, -0.00000243, 0.00000133, -0.00001010, -0.00002771, -0.00003544, -0.00001656, 0.00001306, -0.00001882, -0.00016648, -0.00032761, -0.00025359, 0.00017956, 0.00071027, 0.00084805, 0.00039700, -0.00025650, -0.00052131, -0.00025316, 0.00011831, 0.00014990, -0.00009330, -0.00021140, -0.00004781, 0.00013509, 0.00007638, -0.00011142, -0.00013891, 0.00002593, 0.00013123, 0.00002935, -0.00011588, -0.00009072, 0.00005526, 0.00010412, -0.00000038, -0.00008876, -0.00003576, 0.00006223, 0.00005230, -0.00004364, -0.00007724, -0.00000270, 0.00006676, 0.00003804, -0.00003359, -0.00004440, 0.00000487, 0.00002949, -0.00000253, -0.00003321, -0.00001406, 0.00002419, 0.00002728, -0.00000298, -0.00002211, -0.00001150, 0.00000504, 0.00000568, -0.00000242, -0.00000483, -0.00000276, -0.00000316, -0.00000329, 0.00000180, 0.00000676, 0.00000356, -0.00000399, -0.00000486, 0.00000302, 0.00001047, 0.00001031, 0.00000468, -0.00000126, -0.00000556, -0.00000773, -0.00000681, -0.00000397, -0.00000221, -0.00000121, 0.00000306, 0.00001053, 0.00001361, 0.00000645, -0.00000553, -0.00001082, -0.00000558, 0.00000274, 0.00000553, 0.00000222, -0.00000187, -0.00000282, -0.00000087, 0.00000159, 0.00000221, 0.00000006, -0.00000343, -0.00000522, -0.00000389, },
		{ -0.00000372, -0.00000544, -0.00000330, 0.00000262, 0.00000857, 0.00000956, 0.00000424, -0.00000294, -0.00000619, -0.00000498, -0.00000331, -0.00000249, 0.00000046, 0.00000539, 0.00000610, -0.00000029, -0.00000636, -0.00000440, 0.00000052, -0.00000221, -0.00000996, -0.00000862, 0.00000353, 0.00000833, -0.00000636, -0.00002534, -0.00002533, -0.00000824, 0.00000189, -0.00000743, -0.00002347, -0.00002823, -0.00001684, -0.00000006, -0.00000936, -0.00008757, -0.00022273, -0.00027960, -0.00007890, 0.00036800, 0.00074229, 0.00067168, 0.00015718, -0.00035402, -0.00042620, -0.00011366, 0.00014053, 0.00006450, -0.00014690, -0.00016110, 0.00002599, 0.00012733, 0.00000612, -0.00013774, -0.00008437, 0.00008086, 0.00011034, -0.00003167, -0.00012778, -0.00004055, 0.00009233, 0.00007852, -0.00004531, -0.00008741, 0.00000440, 0.00007886, 0.00002299, -0.00007308, -0.00006483, 0.00003040, 0.00007276, 0.00001242, -0.00005140, -0.00003178, 0.00002505, 0.00002772, -0.00002033, -0.00003957, -0.00000158, 0.00003584, 0.00002265, -0.00001522, -0.00002535, -0.00000417, 0.00001146, 0.00000449, -0.00000688, -0.00000665, -0.00000186, -0.00000196, -0.00000196, 0.00000382, 0.00000791, 0.00000217, -0.00000661, -0.00000558, 0.00000506, 0.00001287, 0.00001039, 0.00000217, -0.00000443, -0.00000733, -0.00000755, -0.00000579, -0.00000340, -0.00000169, 0.00000104, 0.00000726, 0.00001360, 0.00001199, 0.00000067, -0.00001057, -0.00001104, -0.00000214, 0.00000528, 0.00000466, -0.00000019, -0.00000257, -0.00000111, 0.00000112, 0.00000167, 0.00000036, -0.00000202, -0.00000412, -0.00000427, -0.00000228, -0.00000039, -0.00000100, },
		{ -0.00000220, 0.00000408, 0.00000914, 0.00000883, 0.00000299, -0.00000372, -0.00000632, -0.00000477, -0.00000264, -0.00000115, 0.00000155, 0.00000473, 0.00000398, -0.00000136, -0.00000500, -0.00000302, -0.00000119, -0.00000549, -0.00000981, -0.00000420, 0.00000599, 0.00000348, -0.00001420, -0.00002628, -0.00001714, -0.00000085, -0.00000189, -0.00001934, -0.00002868, -0.00001697, 0.00000013, -0.00000506, -0.00005402, -0.00014703, -0.00022916, -0.00017850, 0.00009511, 0.00048596, 0.00068867, 0.00047289, -0.00002234, -0.00037255, -0.00031176, -0.00002066, 0.00011694, -0.00001197, -0.00015836, -0.00009622, 0.00006974, 0.00008999, -0.00005213, -0.00012939, -0.00002170, 0.00010761, 0.00006581, -0.00007994, -0.00011095, 0.00001363, 0.00010546, 0.00003654, -0.00007737, -0.00006488, 0.00004364, 0.00007594, -0.00001528, -0.00008929, -0.00003754, 0.00005923, 0.00006327, -0.00001737, -0.00005773, -0.00000964, 0.00004009, 0.00001508, -0.00003925, -0.00003721, 0.00001636, 0.00004298, 0.00001108, -0.00002745, -0.00002288, 0.00000641, 0.00001535, -0.00000059, -0.00001192, -0.00000634, 0.00000068, -0.00000067, -0.00000096, 0.00000563, 0.00000844, -0.00000028, -0.00000956, -0.00000534, 0.00000828, 0.00001521, 0.00000913, -0.00000150, -0.00000740, -0.00000795, -0.00000659, -0.00000494, -0.00000303, -0.00000035, 0.00000462, 0.00001149, 0.00001453, 0.00000759, -0.00000593, -0.00001374, -0.00000876, 0.00000197, 0.00000648, 0.00000248, -0.00000227, -0.00000190, 0.00000123, 0.00000225, 0.00000039, -0.00000198, -0.00000340, -0.00000372, -0.00000273, -0.00000084, 0.00000004, -0.00000164, -0.00000459, -0.00000551, },
		{ 0.00000955, 0.00000794, 0.00000162, -0.00000455, -0.00000630, -0.00000414, -0.00000161, -0.00000008, 0.00000181, 0.00000363, 0.00000250, -0.00000149, -0.00000392, -0.00000314, -0.00000353, -0.00000715, -0.00000762, -0.00000049, 0.00000510, -0.00000303, -0.00001883, -0.00002220, -0.00000874, 0.00000034, -0.00001120, -0.00002783, -0.00002331, -0.00000123, 0.00000457, -0.00003199, -0.00010463, -0.00017725, -0.00018510, -0.00005076, 0.00023599, 0.00052755, 0.00057910, 0.00028929, -0.00013305, -0.00033863, -0.00020795, 0.00002470, 0.00007130, -0.00006312, -0.00013785, -0.00003713, 0.00008079, 0.00004037, -0.00008568, -0.00009518, 0.00003178, 0.00010410, 0.00001255, -0.00010473, -0.00007288, 0.00005770, 0.00009346, -0.00000967, -0.00008881, -0.00002794, 0.00007134, 0.00005393, -0.00005265, -0.00008747, -0.00000175, 0.00007608, 0.00003995, -0.00004350, -0.00004985, 0.00001644, 0.00004450, -0.00000647, -0.00005343, -0.00002460, 0.00003556, 0.00004200, -0.00000589, -0.00003575, -0.00001378, 0.00001754, 0.00001440, -0.00000887, -0.00001540, -0.00000305, 0.00000406, -0.00000022, -0.00000045, 0.00000738, 0.00000819, -0.00000381, -0.00001235, -0.00000359, 0.00001253, 0.00001672, 0.00000611, -0.00000587, -0.00000943, -0.00000727, -0.00000535, -0.00000450, -0.00000241, 0.00000231, 0.00000913, 0.00001453, 0.00001251, 0.00000104, -0.00001180, -0.00001413, -0.00000451, 0.00000553, 0.00000586, -0.00000039, -0.00000324, 0.00000009, 0.00000343, 0.00000195, -0.00000201, -0.00000410, -0.00000363, -0.00000237, -0.00000105, 0.00000013, -0.00000008, -0.00000253, -0.00000539, -0.00000531, -0.00000091, 0.00000552, },
		{ 0.00000015, -0.00000522, -0.00000592, -0.00000316, -0.00000062, 0.00000045, 0.00000149, 0.00000262, 0.00000173, -0.00000134, -0.00000370, -0.00000426, -0.00000535, -0.00000696, -0.00000478, 0.00000112, 0.00000167, -0.00000860, -0.00001909, -0.00001561, -0.00000390, -0.00000468, -0.00002102, -0.00002837, -0.00001036, 0.00000943, -0.00000904, -0.00007238, -0.00014276, -0.00016803, -0.00010371, 0.00007401, 0.00032441, 0.00050451, 0.00044752, 0.00014463, -0.00018367, -0.00028125, -0.00013143, 0.00003377, 0.00002504, -0.00008499, -0.00010162, 0.00000309, 0.00006660, -0.00000467, -0.00009159, -0.00004962, 0.00006441, 0.00007709, -0.00003430, -0.00010336, -0.00002614, 0.00008190, 0.00006239, -0.00004773, -0.00007833, 0.00001279, 0.00008056, 0.00001885, -0.00007953, -0.00006801, 0.00003354, 0.00007651, 0.00000866, -0.00005876, -0.00002923, 0.00003912, 0.00003564, -0.00003182, -0.00005766, -0.00000336, 0.00005059, 0.00003129, -0.00002445, -0.00003664, 0.00000072, 0.00002565, 0.00000748, -0.00001822, -0.00001519, 0.00000302, 0.00000680, -0.00000127, -0.00000008, 0.00000921, 0.00000692, -0.00000830, -0.00001430, 0.00000003, 0.00001722, 0.00001656, 0.00000132, -0.00001012, -0.00000993, -0.00000559, -0.00000437, -0.00000439, -0.00000089, 0.00000644, 0.00001357, 0.00001510, 0.00000742, -0.00000630, -0.00001541, -0.00001155, 0.00000054, 0.00000746, 0.00000354, -0.00000300, -0.00000260, 0.00000286, 0.00000460, 0.00000006, -0.00000489, -0.00000526, -0.00000259, -0.00000048, 0.00000032, 0.00000049, -0.00000065, -0.00000355, -0.00000602, -0.00000479, 0.00000061, 0.00000691, 0.00000967, 0.00000674, },
		{ -0.00000516, -0.00000211, 0.00000006, 0.00000049, 0.00000100, 0.00000195, 0.00000133, -0.00000150, -0.00000431, -0.00000554, -0.00000600, -0.00000558, -0.00000269, 0.00000042, -0.00000248, -0.00001137, -0.00001589, -0.00000986, -0.00000435, -0.00001294, -0.00002631, -0.00002054, 0.00000329, 0.00000730, -0.00003870, -0.00011216, -0.00015413, -0.00012139, -0.00000669, 0.00017150, 0.00035684, 0.00043972, 0.00032274, 0.00004656, -0.00019303, -0.00022363, -0.00008518, 0.00002336, -0.00000833, -0.00008367, -0.00006535, 0.00002088, 0.00004007, -0.00003391, -0.00007571, -0.00000741, 0.00007278, 0.00003875, -0.00006401, -0.00008100, 0.00001604, 0.00008335, 0.00002300, -0.00006884, -0.00005099, 0.00004634, 0.00007004, -0.00001979, -0.00008958, -0.00003633, 0.00005956, 0.00006082, -0.00002231, -0.00005901, -0.00000132, 0.00005161, 0.00001475, -0.00005408, -0.00004926, 0.00002184, 0.00005648, 0.00001220, -0.00003955, -0.00002845, 0.00001727, 0.00002737, -0.00000462, -0.00002550, -0.00001002, 0.00001040, 0.00000732, -0.00000377, 0.00000082, 0.00001113, 0.00000428, -0.00001339, -0.00001460, 0.00000554, 0.00002139, 0.00001402, -0.00000477, -0.00001328, -0.00000863, -0.00000349, -0.00000404, -0.00000411, 0.00000213, 0.00001155, 0.00001655, 0.00001233, 0.00000007, -0.00001262, -0.00001576, -0.00000670, 0.00000494, 0.00000718, 0.00000028, -0.00000439, -0.00000039, 0.00000549, 0.00000411, -0.00000311, -0.00000737, -0.00000499, -0.00000051, 0.00000140, 0.00000108, 0.00000028, -0.00000149, -0.00000458, -0.00000640, -0.00000390, 0.00000234, 0.00000812, 0.00000936, 0.00000525, -0.00000124, -0.00000550, },
		{ 0.00000035, 0.00000031, 0.00000061, 0.00000153, 0.00000088, -0.00000216, -0.00000527, -0.00000627, -0.00000557, -0.00000406, -0.00000209, -0.00000167, -0.00000547, -0.00001101, -0.00001157, -0.00000750, -0.00000920, -0.00002011, -0.00002445, -0.00000813, 0.00000992, -0.00001019, -0.00007564, -0.00013689, -0.00013315, -0.00004992, 0.00008298, 0.00022850, 0.00034343, 0.00035876, 0.00022244, -0.00001066, -0.00018198, -0.00017877, -0.00006228, 0.00000881, -0.00002490, -0.00007033, -0.00003927, 0.00002089, 0.00001410, -0.00004403, -0.00004897, 0.00002120, 0.00006129, 0.00000198, -0.00007285, -0.00004785, 0.00004393, 0.00006605, -0.00001313, -0.00007014, -0.00001611, 0.00006481, 0.00004418, -0.00005211, -0.00008157, -0.00000102, 0.00007050, 0.00003397, -0.00004486, -0.00004452, 0.00002614, 0.00004993, -0.00001314, -0.00006677, -0.00002920, 0.00004459, 0.00005053, -0.00001106, -0.00004645, -0.00001216, 0.00003120, 0.00002085, -0.00001912, -0.00002760, -0.00000020, 0.00001673, 0.00000456, -0.00000690, 0.00000297, 0.00001273, -0.00000005, -0.00001840, -0.00001255, 0.00001246, 0.00002391, 0.00000885, -0.00001121, -0.00001446, -0.00000578, -0.00000171, -0.00000436, -0.00000294, 0.00000682, 0.00001655, 0.00001667, 0.00000621, -0.00000791, -0.00001621, -0.00001275, -0.00000102, 0.00000741, 0.00000489, -0.00000277, -0.00000396, 0.00000274, 0.00000697, 0.00000176, -0.00000676, -0.00000860, -0.00000319, 0.00000206, 0.00000275, 0.00000111, -0.00000036, -0.00000246, -0.00000549, -0.00000643, -0.00000260, 0.00000415, 0.00000894, 0.00000854, 0.00000362, -0.00000229, -0.00000533, -0.00000420, -0.00000125, },
		{ 0.00000033, 0.00000106, 0.00000010, -0.00000314, -0.00000601, -0.00000623, -0.00000466, -0.00000325, -0.00000279, -0.00000372, -0.00000633, -0.00000870, -0.00000864, -0.00000918, -0.00001548, -0.00002261, -0.00001636, 0.00000259, 0.00000435, -0.00003929, -0.00010765, -0.00013676, -0.00008419, 0.00002814, 0.00014777, 0.00024449, 0.00030229, 0.00028250, 0.00015182, -0.00004031, -0.00016684, -0.00014932, -0.00005169, -0.00000047, -0.00002841, -0.00005579, -0.00002624, 0.00001214, -0.00000274, -0.00003886, -0.00002272, 0.00003271, 0.00003917, -0.00002378, -0.00006402, -0.00001518, 0.00005376, 0.00003866, -0.00003725, -0.00005486, 0.00001607, 0.00006551, 0.00001126, -0.00007093, -0.00005939, 0.00002896, 0.00006520, 0.00000383, -0.00005363, -0.00001990, 0.00004552, 0.00003423, -0.00004073, -0.00006594, -0.00000203, 0.00005871, 0.00003341, -0.00003256, -0.00004241, 0.00000849, 0.00003798, 0.00000668, -0.00003187, -0.00002256, 0.00001223, 0.00001945, -0.00000147, -0.00000919, 0.00000674, 0.00001325, -0.00000614, -0.00002234, -0.00000770, 0.00001980, 0.00002365, 0.00000136, -0.00001672, -0.00001314, -0.00000206, -0.00000085, -0.00000487, -0.00000017, 0.00001271, 0.00001986, 0.00001299, -0.00000226, -0.00001445, -0.00001604, -0.00000733, 0.00000378, 0.00000736, 0.00000152, -0.00000448, -0.00000178, 0.00000574, 0.00000655, -0.00000203, -0.00000986, -0.00000807, -0.00000022, 0.00000443, 0.00000325, 0.00000052, -0.00000120, -0.00000342, -0.00000620, -0.00000604, -0.00000095, 0.00000583, 0.00000921, 0.00000730, 0.00000211, -0.00000286, -0.00000482, -0.00000333, -0.00000071, 0.00000042, 0.00000013, },
		{ -0.00000094, -0.00000400, -0.00000620, -0.00000569, -0.00000398, -0.00000335, -0.00000390, -0.00000459, -0.00000531, -0.00000640, -0.00000860, -0.00001347, -0.00001973, -0.00001926, -0.00000610, 0.00000566, -0.00001378, -0.00007138, -0.00012352, -0.00010972, -0.00001943, 0.00009492, 0.00017971, 0.00022913, 0.00025258, 0.00022277, 0.00010613, -0.00005670, -0.00015615, -0.00013079, -0.00004365, -0.00000212, -0.00002618, -0.00004669, -0.00002303, 0.00000356, -0.00000804, -0.00002628, -0.00000481, 0.00003003, 0.00001646, -0.00003486, -0.00004523, 0.00000846, 0.00004787, 0.00001102, -0.00004571, -0.00003043, 0.00003768, 0.00005114, -0.00001955, -0.00007369, -0.00003046, 0.00004707, 0.00004721, -0.00002139, -0.00004760, 0.00000746, 0.00005173, 0.00000880, -0.00006074, -0.00005153, 0.00002543, 0.00006030, 0.00000919, -0.00004642, -0.00002779, 0.00002803, 0.00003476, -0.00001202, -0.00003863, -0.00001045, 0.00002394, 0.00001679, -0.00000943, -0.00000902, 0.00001178, 0.00001170, -0.00001360, -0.00002407, -0.00000014, 0.00002617, 0.00001991, -0.00000747, -0.00002002, -0.00000936, 0.00000159, -0.00000114, -0.00000478, 0.00000451, 0.00001861, 0.00001993, 0.00000558, -0.00001117, -0.00001767, -0.00001215, -0.00000122, 0.00000631, 0.00000510, -0.00000159, -0.00000423, 0.00000139, 0.00000745, 0.00000408, -0.00000631, -0.00001147, -0.00000575, 0.00000322, 0.00000601, 0.00000283, -0.00000045, -0.00000204, -0.00000427, -0.00000662, -0.00000519, 0.00000089, 0.00000714, 0.00000889, 0.00000587, 0.00000096, -0.00000297, -0.00000422, -0.00000272, -0.00000041, 0.00000048, 0.00000004, -0.00000003, 0.00000030, },
		{ -0.00000594, -0.00000517, -0.00000387, -0.00000390, -0.00000444, -0.00000403, -0.00000367, -0.00000577, -0.00001120, -0.00001769, -0.00001972, -0.00001189, 0.00000098, -0.00000200, -0.00003978, -0.00009608, -0.00011693, -0.00006172, 0.00004564, 0.00013819, 0.00018089, 0.00019728, 0.00020882, 0.00018183, 0.00007563, -0.00007024, -0.00015099, -0.00011615, -0.00003289, 0.00000092, -0.00002491, -0.00004427, -0.00002341, 0.00000065, -0.00000475, -0.00001430, 0.00000226, 0.00001991, 0.00000044, -0.00003330, -0.00002527, 0.00001943, 0.00003275, -0.00000915, -0.00004026, -0.00000554, 0.00004531, 0.00002823, -0.00004112, -0.00006272, -0.00000301, 0.00005101, 0.00002329, -0.00003572, -0.00003013, 0.00002981, 0.00004373, -0.00001937, -0.00006817, -0.00002739, 0.00004625, 0.00004902, -0.00001581, -0.00004877, -0.00000619, 0.00004084, 0.00002155, -0.00003029, -0.00003646, 0.00000636, 0.00003126, 0.00000848, -0.00001697, -0.00000533, 0.00001694, 0.00000727, -0.00002145, -0.00002254, 0.00000935, 0.00003000, 0.00001261, -0.00001614, -0.00002014, -0.00000377, 0.00000422, -0.00000231, -0.00000327, 0.00001073, 0.00002284, 0.00001574, -0.00000439, -0.00001817, -0.00001651, -0.00000577, 0.00000361, 0.00000597, 0.00000177, -0.00000318, -0.00000214, 0.00000434, 0.00000711, 0.00000010, -0.00000989, -0.00001103, -0.00000213, 0.00000629, 0.00000646, 0.00000170, -0.00000149, -0.00000277, -0.00000496, -0.00000669, -0.00000397, 0.00000270, 0.00000793, 0.00000810, 0.00000453, 0.00000029, -0.00000280, -0.00000378, -0.00000238, -0.00000014, 0.00000066, -0.00000012, -0.00000071, -0.00000077, -0.00000193, -0.00000442, },
		{ -0.00000413, -0.00000415, -0.00000394, -0.00000272, -0.00000287, -0.00000731, -0.00001477, -0.00001941, -0.00001550, -0.00000440, 0.00000062, -0.00001920, -0.00006562, -0.00010502, -0.00008835, -0.00000425, 0.00009724, 0.00015393, 0.00016076, 0.00016337, 0.00017808, 0.00015490, 0.00005066, -0.00008538, -0.00014769, -0.00009982, -0.00001917, 0.00000368, -0.00002798, -0.00004555, -0.00002170, 0.00000423, 0.00000143, -0.00000808, 0.00000086, 0.00000960, -0.00000620, -0.00002476, -0.00001078, 0.00001912, 0.00001609, -0.00001837, -0.00002644, 0.00001291, 0.00004033, 0.00000450, -0.00005029, -0.00004379, 0.00001654, 0.00004287, 0.00000096, -0.00003711, -0.00000760, 0.00004152, 0.00002468, -0.00004287, -0.00006184, -0.00000005, 0.00005553, 0.00002830, -0.00003503, -0.00003904, 0.00001655, 0.00004304, 0.00000139, -0.00004283, -0.00002487, 0.00002373, 0.00003139, -0.00000390, -0.00002142, 0.00000188, 0.00002049, -0.00000032, -0.00002818, -0.00001715, 0.00001934, 0.00002996, 0.00000249, -0.00002294, -0.00001670, 0.00000252, 0.00000521, -0.00000364, 0.00000019, 0.00001737, 0.00002370, 0.00000734, -0.00001475, -0.00002117, -0.00001117, 0.00000104, 0.00000564, 0.00000329, -0.00000105, -0.00000262, 0.00000086, 0.00000580, 0.00000462, -0.00000427, -0.00001177, -0.00000863, 0.00000193, 0.00000835, 0.00000579, 0.00000023, -0.00000236, -0.00000332, -0.00000547, -0.00000641, -0.00000249, 0.00000426, 0.00000817, 0.00000709, 0.00000352, 0.00000003, -0.00000264, -0.00000363, -0.00000214, 0.00000029, 0.00000088, -0.00000060, -0.00000182, -0.00000189, -0.00000251, -0.00000437, -0.00000561, -0.00000504, },
		{ -0.00000274, -0.00000181, -0.00000371, -0.00001016, -0.00001717, -0.00001763, -0.00000939, -0.00000101, -0.00000844, -0.00004085, -0.00008291, -0.00009464, -0.00004450, 0.00004978, 0.00012713, 0.00014601, 0.00013176, 0.00013731, 0.00016010, 0.00013435, 0.00002512, -0.00010134, -0.00014155, -0.00007986, -0.00000560, 0.00000228, -0.00003479, -0.00004596, -0.00001511, 0.00001137, 0.00000513, -0.00000861, -0.00000394, 0.00000401, -0.00000517, -0.00001567, -0.00000448, 0.00001231, 0.00000392, -0.00001756, -0.00001099, 0.00002162, 0.00002753, -0.00001383, -0.00004813, -0.00002377, 0.00002536, 0.00002793, -0.00001400, -0.00002766, 0.00001291, 0.00004064, 0.00000064, -0.00005621, -0.00004456, 0.00002345, 0.00005193, 0.00000419, -0.00004367, -0.00002027, 0.00003414, 0.00003376, -0.00002039, -0.00004567, -0.00000616, 0.00003691, 0.00002347, -0.00001725, -0.00002068, 0.00001131, 0.00002061, -0.00001040, -0.00003208, -0.00000810, 0.00002794, 0.00002534, -0.00000891, -0.00002628, -0.00001008, 0.00000820, 0.00000444, -0.00000424, 0.00000549, 0.00002278, 0.00002004, -0.00000404, -0.00002288, -0.00001905, -0.00000316, 0.00000602, 0.00000434, -0.00000033, -0.00000199, -0.00000019, 0.00000333, 0.00000501, 0.00000073, -0.00000773, -0.00001145, -0.00000492, 0.00000548, 0.00000907, 0.00000434, -0.00000119, -0.00000295, -0.00000373, -0.00000577, -0.00000580, -0.00000095, 0.00000542, 0.00000797, 0.00000612, 0.00000291, -0.00000005, -0.00000270, -0.00000369, -0.00000177, 0.00000092, 0.00000091, -0.00000159, -0.00000320, -0.00000272, -0.00000257, -0.00000413, -0.00000556, -0.00000524, -0.00000419, -0.00000358, },
		{ -0.00000587, -0.00001261, -0.00001697, -0.00001336, -0.00000485, -0.00000427, -0.00002388, -0.00005995, -0.00008585, -0.00006713, 0.00000431, 0.00008996, 0.00013403, 0.00012342, 0.00010485, 0.00012272, 0.00014991, 0.00011351, -0.00000245, -0.00011451, -0.00012965, -0.00005782, 0.00000401, -0.00000427, -0.00004211, -0.00004200, -0.00000441, 0.00001763, 0.00000356, -0.00001335, -0.00000712, 0.00000441, -0.00000082, -0.00001057, -0.00000515, 0.00000465, -0.00000113, -0.00001079, 0.00000068, 0.00002127, 0.00001287, -0.00002371, -0.00003885, -0.00000819, 0.00002447, 0.00001242, -0.00001909, -0.00001239, 0.00002590, 0.00002918, -0.00002156, -0.00005735, -0.00002200, 0.00003776, 0.00003791, -0.00001665, -0.00004025, 0.00000194, 0.00004184, 0.00001556, -0.00003777, -0.00003751, 0.00001510, 0.00004203, 0.00000895, -0.00002783, -0.00001408, 0.00002060, 0.00001611, -0.00002138, -0.00003168, 0.00000347, 0.00003316, 0.00001632, -0.00001955, -0.00002516, -0.00000138, 0.00001211, 0.00000230, -0.00000334, 0.00001182, 0.00002517, 0.00001174, -0.00001606, -0.00002644, -0.00001210, 0.00000509, 0.00000747, 0.00000043, -0.00000301, -0.00000043, 0.00000292, 0.00000393, 0.00000206, -0.00000325, -0.00000922, -0.00000916, -0.00000091, 0.00000786, 0.00000857, 0.00000257, -0.00000231, -0.00000328, -0.00000403, -0.00000587, -0.00000496, 0.00000049, 0.00000616, 0.00000755, 0.00000539, 0.00000261, -0.00000020, -0.00000304, -0.00000374, -0.00000111, 0.00000159, 0.00000047, -0.00000309, -0.00000451, -0.00000301, -0.00000225, -0.00000407, -0.00000588, -0.00000532, -0.00000352, -0.00000232, -0.00000174, -0.00000221, },
		{ -0.00001431, -0.00000916, -0.00000482, -0.00001392, -0.00004070, -0.00007016, -0.00007313, -0.00002924, 0.00004769, 0.00011112, 0.00012269, 0.00009644, 0.00008664, 0.00011780, 0.00014124, 0.00008903, -0.00002987, -0.00012138, -0.00011206, -0.00003721, 0.00000739, -0.00001409, -0.00004622, -0.00003278, 0.00000720, 0.00001947, -0.00000282, -0.00001819, -0.00000580, 0.00000884, 0.00000232, -0.00001089, -0.00000908, 0.00000033, 0.00000011, -0.00000300, 0.00000606, 0.00001531, 0.00000118, -0.00002564, -0.00002768, 0.00000028, 0.00001771, 0.00000130, -0.00001547, 0.00000279, 0.00002907, 0.00001195, -0.00003660, -0.00004806, -0.00000062, 0.00004078, 0.00001872, -0.00002902, -0.00002694, 0.00002117, 0.00003794, -0.00000637, -0.00004595, -0.00002019, 0.00003339, 0.00003736, -0.00000865, -0.00003243, -0.00000269, 0.00002698, 0.00000694, -0.00003104, -0.00002631, 0.00001560, 0.00003349, 0.00000415, -0.00002725, -0.00001941, 0.00000782, 0.00001349, -0.00000039, -0.00000063, 0.00001774, 0.00002324, -0.00000006, -0.00002592, -0.00002417, -0.00000204, 0.00001097, 0.00000488, -0.00000431, -0.00000321, 0.00000315, 0.00000504, 0.00000193, -0.00000207, -0.00000586, -0.00000840, -0.00000577, 0.00000237, 0.00000884, 0.00000733, 0.00000094, -0.00000304, -0.00000346, -0.00000426, -0.00000577, -0.00000401, 0.00000170, 0.00000657, 0.00000712, 0.00000494, 0.00000240, -0.00000062, -0.00000356, -0.00000356, -0.00000017, 0.00000202, -0.00000059, -0.00000485, -0.00000540, -0.00000274, -0.00000192, -0.00000440, -0.00000624, -0.00000473, -0.00000205, -0.00000109, -0.00000187, -0.00000392, -0.00000808, -0.00001317, },
		{ -0.00001027, -0.00002688, -0.00005319, -0.00006809, -0.00004802, 0.00001025, 0.00007825, 0.00011377, 0.00010112, 0.00007327, 0.00007865, 0.00011773, 0.00012933, 0.00006119, -0.00005356, -0.00012036, -0.00009135, -0.00002127, 0.00000478, -0.00002375, -0.00004487, -0.00002010, 0.00001597, 0.00001572, -0.00001114, -0.00001964, -0.00000005, 0.00001367, 0.00000146, -0.00001510, -0.00001224, 0.00000083, 0.00000451, 0.00000203, 0.00000574, 0.00000806, -0.00000523, -0.00002260, -0.00001885, 0.00000218, 0.00000982, -0.00000317, -0.00000685, 0.00001317, 0.00002355, -0.00000534, -0.00004213, -0.00003283, 0.00001441, 0.00003404, 0.00000053, -0.00003078, -0.00000869, 0.00003227, 0.00002425, -0.00002582, -0.00004278, 0.00000174, 0.00004393, 0.00002391, -0.00002480, -0.00002942, 0.00001081, 0.00002815, -0.00000563, -0.00003704, -0.00001643, 0.00002586, 0.00002832, -0.00000904, -0.00003021, -0.00000983, 0.00001576, 0.00001217, -0.00000270, 0.00000358, 0.00002160, 0.00001662, -0.00001311, -0.00003106, -0.00001629, 0.00000838, 0.00001255, -0.00000085, -0.00000770, -0.00000030, 0.00000726, 0.00000464, -0.00000239, -0.00000576, -0.00000606, -0.00000573, -0.00000249, 0.00000431, 0.00000872, 0.00000591, -0.00000033, -0.00000351, -0.00000358, -0.00000439, -0.00000551, -0.00000308, 0.00000266, 0.00000683, 0.00000685, 0.00000469, 0.00000205, -0.00000133, -0.00000405, -0.00000301, 0.00000086, 0.00000194, -0.00000222, -0.00000654, -0.00000572, -0.00000218, -0.00000188, -0.00000496, -0.00000607, -0.00000322, -0.00000036, -0.00000085, -0.00000342, -0.00000600, -0.00000883, -0.00001142, -0.00001091, -0.00000780, },
		{ -0.00005711, -0.00005425, -0.00001682, 0.00004353, 0.00009317, 0.00010271, 0.00007746, 0.00005824, 0.00007846, 0.00011740, 0.00011239, 0.00003268, -0.00007063, -0.00011240, -0.00007101, -0.00001155, -0.00000144, -0.00003011, -0.00003819, -0.00000730, 0.00001949, 0.00000783, -0.00001783, -0.00001625, 0.00000777, 0.00001563, -0.00000363, -0.00002007, -0.00001195, 0.00000500, 0.00000852, 0.00000284, 0.00000237, 0.00000290, -0.00000682, -0.00001834, -0.00001432, 0.00000023, 0.00000454, -0.00000167, 0.00000220, 0.00001664, 0.00001297, -0.00001808, -0.00003913, -0.00001712, 0.00002075, 0.00002175, -0.00001169, -0.00002328, 0.00000855, 0.00003297, 0.00000541, -0.00003755, -0.00002948, 0.00002249, 0.00004414, 0.00000516, -0.00003531, -0.00001934, 0.00002292, 0.00002308, -0.00001914, -0.00003772, -0.00000371, 0.00003195, 0.00001827, -0.00002064, -0.00002750, 0.00000194, 0.00002093, 0.00000860, -0.00000388, 0.00000838, 0.00002204, 0.00000614, -0.00002457, -0.00002998, -0.00000459, 0.00001621, 0.00000920, -0.00000763, -0.00000791, 0.00000509, 0.00000988, 0.00000100, -0.00000770, -0.00000735, -0.00000368, -0.00000242, -0.00000040, 0.00000488, 0.00000813, 0.00000475, -0.00000129, -0.00000389, -0.00000367, -0.00000440, -0.00000516, -0.00000225, 0.00000344, 0.00000711, 0.00000677, 0.00000442, 0.00000140, -0.00000222, -0.00000426, -0.00000211, 0.00000173, 0.00000124, -0.00000417, -0.00000785, -0.00000552, -0.00000165, -0.00000216, -0.00000524, -0.00000496, -0.00000114, 0.00000059, -0.00000218, -0.00000573, -0.00000696, -0.00000733, -0.00000847, -0.00000928, -0.00001091, -0.00001964, -0.00003857, },
		{ 0.00001352, 0.00006586, 0.00009402, 0.00008457, 0.00005762, 0.00005169, 0.00008190, 0.00011340, 0.00009134, 0.00000675, -0.00008010, -0.00010013, -0.00005388, -0.00000742, -0.00000811, -0.00003162, -0.00002835, 0.00000248, 0.00001753, -0.00000120, -0.00002035, -0.00000897, 0.00001447, 0.00001317, -0.00001100, -0.00002279, -0.00000780, 0.00001019, 0.00000969, 0.00000033, -0.00000095, 0.00000114, -0.00000586, -0.00001576, -0.00001357, -0.00000213, 0.00000349, 0.00000303, 0.00000810, 0.00001393, 0.00000175, -0.00002417, -0.00003099, -0.00000532, 0.00001924, 0.00000910, -0.00001562, -0.00001057, 0.00001981, 0.00002436, -0.00001274, -0.00003888, -0.00001011, 0.00003678, 0.00003435, -0.00001401, -0.00003759, -0.00000475, 0.00003037, 0.00001243, -0.00003069, -0.00003262, 0.00000933, 0.00003238, 0.00000520, -0.00002813, -0.00001933, 0.00001371, 0.00002238, 0.00000368, -0.00000356, 0.00001253, 0.00001840, -0.00000631, -0.00003180, -0.00002267, 0.00000804, 0.00001921, 0.00000188, -0.00001290, -0.00000416, 0.00001115, 0.00000923, -0.00000541, -0.00001200, -0.00000577, 0.00000047, 0.00000007, -0.00000002, 0.00000465, 0.00000778, 0.00000397, -0.00000218, -0.00000431, -0.00000365, -0.00000426, -0.00000482, -0.00000157, 0.00000419, 0.00000753, 0.00000676, 0.00000392, 0.00000042, -0.00000307, -0.00000407, -0.00000101, 0.00000218, -0.00000003, -0.00000614, -0.00000869, -0.00000507, -0.00000133, -0.00000241, -0.00000474, -0.00000297, 0.00000062, -0.00000005, -0.00000478, -0.00000739, -0.00000565, -0.00000412, -0.00000649, -0.00001138, -0.00001802, -0.00002946, -0.00004472, -0.00005120, -0.00003247, },
		{ 0.00008514, 0.00006535, 0.00004423, 0.00005137, 0.00008510, 0.00010472, 0.00006862, -0.00001430, -0.00008294, -0.00008653, -0.00004109, -0.00000677, -0.00001273, -0.00002879, -0.00001831, 0.00000755, 0.00001189, -0.00000832, -0.00001813, -0.00000036, 0.00001756, 0.00000690, -0.00001784, -0.00002168, -0.00000133, 0.00001372, 0.00000754, -0.00000326, -0.00000207, 0.00000204, -0.00000502, -0.00001591, -0.00001446, -0.00000239, 0.00000598, 0.00000760, 0.00000936, 0.00000776, -0.00000657, -0.00002423, -0.00002194, 0.00000048, 0.00001314, 0.00000037, -0.00001205, 0.00000224, 0.00002265, 0.00001029, -0.00002505, -0.00003037, 0.00000974, 0.00004142, 0.00001766, -0.00002880, -0.00003144, 0.00001053, 0.00003115, -0.00000160, -0.00003772, -0.00002274, 0.00001988, 0.00002696, -0.00000816, -0.00002975, -0.00000712, 0.00002323, 0.00001992, -0.00000150, -0.00000189, 0.00001479, 0.00001103, -0.00001816, -0.00003305, -0.00001074, 0.00001843, 0.00001649, -0.00000717, -0.00001449, 0.00000284, 0.00001551, 0.00000447, -0.00001292, -0.00001335, -0.00000102, 0.00000477, 0.00000056, -0.00000109, 0.00000454, 0.00000804, 0.00000331, -0.00000328, -0.00000471, -0.00000340, -0.00000404, -0.00000463, -0.00000099, 0.00000507, 0.00000813, 0.00000662, 0.00000304, -0.00000075, -0.00000366, -0.00000348, 0.00000004, 0.00000211, -0.00000169, -0.00000794, -0.00000915, -0.00000458, -0.00000106, -0.00000211, -0.00000324, -0.00000081, 0.00000103, -0.00000243, -0.00000744, -0.00000693, -0.00000203, -0.00000105, -0.00000772, -0.00001756, -0.00002658, -0.00003580, -0.00004313, -0.00003738, -0.00000824, 0.00003768, 0.00007628, },
		{ 0.00003715, 0.00005416, 0.00008575, 0.00009223, 0.00004670, -0.00002976, -0.00008118, -0.00007375, -0.00003215, -0.00000720, -0.00001433, -0.00002356, -0.00001055, 0.00000813, 0.00000531, -0.00001165, -0.00001249, 0.00000679, 0.00001627, -0.00000107, -0.00002183, -0.00001707, 0.00000505, 0.00001413, 0.00000332, -0.00000560, -0.00000057, 0.00000365, -0.00000609, -0.00001798, -0.00001456, 0.00000026, 0.00000979, 0.00000958, 0.00000669, 0.00000142, -0.00001048, -0.00002082, -0.00001537, 0.00000089, 0.00000653, -0.00000241, -0.00000416, 0.00001085, 0.00001769, -0.00000412, -0.00002871, -0.00001545, 0.00002481, 0.00003612, -0.00000117, -0.00003593, -0.00001897, 0.00002262, 0.00002507, -0.00001587, -0.00003883, -0.00001034, 0.00002570, 0.00001694, -0.00001884, -0.00002496, 0.00000681, 0.00002869, 0.00001413, -0.00000589, 0.00000056, 0.00001434, 0.00000125, -0.00002688, -0.00002797, 0.00000305, 0.00002400, 0.00000883, -0.00001509, -0.00001136, 0.00001114, 0.00001604, -0.00000387, -0.00001914, -0.00001051, 0.00000573, 0.00000738, -0.00000139, -0.00000262, 0.00000541, 0.00000883, 0.00000227, -0.00000471, -0.00000487, -0.00000282, -0.00000388, -0.00000468, -0.00000035, 0.00000624, 0.00000877, 0.00000612, 0.00000174, -0.00000192, -0.00000383, -0.00000260, 0.00000085, 0.00000153, -0.00000355, -0.00000953, -0.00000941, -0.00000403, -0.00000039, -0.00000088, -0.00000116, 0.00000045, -0.00000052, -0.00000571, -0.00000849, -0.00000367, 0.00000251, -0.00000053, -0.00001324, -0.00002609, -0.00003302, -0.00003594, -0.00003438, -0.00001968, 0.00001342, 0.00005326, 0.00007688, 0.00007148, 0.00004889, },
		{ 0.00008323, 0.00007756, 0.00002718, -0.00004020, -0.00007682, -0.00006262, -0.00002566, -0.00000709, -0.00001344, -0.00001816, -0.00000610, 0.00000593, 0.00000022, -0.00001111, -0.00000582, 0.00001062, 0.00001158, -0.00000822, -0.00002205, -0.00001067, 0.00000927, 0.00001159, -0.00000090, -0.00000543, 0.00000247, 0.00000397, -0.00000933, -0.00002011, -0.00001231, 0.00000486, 0.00001248, 0.00000821, 0.00000230, -0.00000264, -0.00001064, -0.00001708, -0.00001270, -0.00000160, 0.00000270, 0.00000013, 0.00000391, 0.00001309, 0.00000806, -0.00001430, -0.00002387, 0.00000092, 0.00003171, 0.00002339, -0.00001725, -0.00003448, -0.00000388, 0.00002873, 0.00001381, -0.00002730, -0.00003418, 0.00000169, 0.00002576, 0.00000480, -0.00002438, -0.00001465, 0.00001968, 0.00002910, 0.00000623, -0.00000878, 0.00000299, 0.00001104, -0.00000897, -0.00003062, -0.00001774, 0.00001551, 0.00002353, -0.00000158, -0.00001936, -0.00000402, 0.00001806, 0.00001163, -0.00001388, -0.00002171, -0.00000356, 0.00001239, 0.00000689, -0.00000520, -0.00000332, 0.00000764, 0.00000956, 0.00000040, -0.00000627, -0.00000446, -0.00000197, -0.00000404, -0.00000493, 0.00000058, 0.00000775, 0.00000924, 0.00000506, 0.00000014, -0.00000283, -0.00000354, -0.00000161, 0.00000130, 0.00000053, -0.00000551, -0.00001097, -0.00000949, -0.00000309, 0.00000113, 0.00000118, 0.00000053, -0.00000013, -0.00000369, -0.00000826, -0.00000665, 0.00000170, 0.00000560, -0.00000436, -0.00002218, -0.00003375, -0.00003424, -0.00002951, -0.00002139, -0.00000252, 0.00002954, 0.00006071, 0.00007115, 0.00005693, 0.00003664, 0.00003473, 0.00005754, },
		{ 0.00001059, -0.00004669, -0.00007113, -0.00005293, -0.00002031, -0.00000601, -0.00001147, -0.00001417, -0.00000450, 0.00000309, -0.00000216, -0.00000802, -0.00000036, 0.00001080, 0.00000549, -0.00001277, -0.00001918, -0.00000463, 0.00001046, 0.00000754, -0.00000334, -0.00000290, 0.00000522, 0.00000196, -0.00001362, -0.00002046, -0.00000760, 0.00000948, 0.00001249, 0.00000444, -0.00000141, -0.00000364, -0.00000911, -0.00001535, -0.00001328, -0.00000387, 0.00000305, 0.00000543, 0.00000875, 0.00000948, -0.00000201, -0.00001772, -0.00001334, 0.00001393, 0.00002979, 0.00000754, -0.00002697, -0.00002598, 0.00000979, 0.00002789, 0.00000026, -0.00003376, -0.00002544, 0.00001069, 0.00002059, -0.00000639, -0.00002351, -0.00000100, 0.00002894, 0.00002451, -0.00000224, -0.00000986, 0.00000462, 0.00000548, -0.00001751, -0.00002869, -0.00000471, 0.00002390, 0.00001740, -0.00001181, -0.00001856, 0.00000560, 0.00002111, 0.00000270, -0.00002283, -0.00001914, 0.00000607, 0.00001666, 0.00000288, -0.00000943, -0.00000203, 0.00001090, 0.00000938, -0.00000242, -0.00000747, -0.00000326, -0.00000115, -0.00000475, -0.00000522, 0.00000205, 0.00000949, 0.00000921, 0.00000335, -0.00000153, -0.00000325, -0.00000284, -0.00000071, 0.00000134, -0.00000082, -0.00000757, -0.00001231, -0.00000916, -0.00000128, 0.00000366, 0.00000332, 0.00000069, -0.00000279, -0.00000717, -0.00000836, -0.00000178, 0.00000711, 0.00000483, -0.00001267, -0.00003183, -0.00003710, -0.00002903, -0.00001861, -0.00000809, 0.00001102, 0.00003970, 0.00006224, 0.00006228, 0.00004358, 0.00002834, 0.00003517, 0.00006015, 0.00007793, 0.00006214, },
		{ -0.00006457, -0.00004415, -0.00001546, -0.00000449, -0.00000972, -0.00001196, -0.00000441, 0.00000123, -0.00000205, -0.00000430, 0.00000263, 0.00000837, 0.00000004, -0.00001425, -0.00001486, -0.00000056, 0.00000913, 0.00000384, -0.00000324, 0.00000069, 0.00000614, -0.00000212, -0.00001722, -0.00001812, -0.00000166, 0.00001222, 0.00000970, 0.00000017, -0.00000284, -0.00000245, -0.00000821, -0.00001635, -0.00001500, -0.00000362, 0.00000684, 0.00001026, 0.00000882, 0.00000258, -0.00000879, -0.00001448, -0.00000134, 0.00002043, 0.00002104, -0.00000687, -0.00002901, -0.00001369, 0.00001887, 0.00002110, -0.00001238, -0.00003465, -0.00001522, 0.00001504, 0.00001213, -0.00001382, -0.00001650, 0.00001304, 0.00003282, 0.00001598, -0.00000974, -0.00000927, 0.00000496, -0.00000117, -0.00002265, -0.00002170, 0.00000823, 0.00002663, 0.00000748, -0.00001912, -0.00001284, 0.00001474, 0.00001886, -0.00000883, -0.00002797, -0.00001136, 0.00001589, 0.00001679, -0.00000386, -0.00001228, 0.00000171, 0.00001421, 0.00000755, -0.00000584, -0.00000769, -0.00000131, -0.00000081, -0.00000608, -0.00000519, 0.00000424, 0.00001119, 0.00000836, 0.00000108, -0.00000295, -0.00000306, -0.00000188, -0.00000006, 0.00000094, -0.00000250, -0.00000976, -0.00001338, -0.00000797, 0.00000175, 0.00000675, 0.00000437, -0.00000140, -0.00000672, -0.00000907, -0.00000505, 0.00000472, 0.00000986, -0.00000101, -0.00002358, -0.00003858, -0.00003402, -0.00001863, -0.00000681, 0.00000234, 0.00002008, 0.00004535, 0.00006026, 0.00005221, 0.00003198, 0.00002325, 0.00003730, 0.00006153, 0.00007048, 0.00004676, -0.00000313, -0.00005007, },
		{ -0.00001129, -0.00000330, -0.00000880, -0.00001095, -0.00000444, 0.00000090, -0.00000056, -0.00000148, 0.00000313, 0.00000496, -0.00000351, -0.00001338, -0.00001084, 0.00000108, 0.00000667, 0.00000188, -0.00000108, 0.00000370, 0.00000462, -0.00000697, -0.00001865, -0.00001342, 0.00000373, 0.00001214, 0.00000526, -0.00000273, -0.00000182, -0.00000087, -0.00000944, -0.00001914, -0.00001551, -0.00000015, 0.00001186, 0.00001212, 0.00000475, -0.00000417, -0.00001034, -0.00000692, 0.00000809, 0.00001976, 0.00000911, -0.00001641, -0.00002436, -0.00000140, 0.00002188, 0.00001078, -0.00002162, -0.00003092, -0.00000615, 0.00001458, 0.00000318, -0.00001571, -0.00000514, 0.00002445, 0.00003075, 0.00000529, -0.00001511, -0.00000749, 0.00000389, -0.00000748, -0.00002348, -0.00001136, 0.00001841, 0.00002361, -0.00000354, -0.00002174, -0.00000382, 0.00002071, 0.00001138, -0.00002010, -0.00002745, 0.00000009, 0.00002314, 0.00001217, -0.00001148, -0.00001219, 0.00000742, 0.00001627, 0.00000381, -0.00000906, -0.00000644, 0.00000097, -0.00000143, -0.00000787, -0.00000439, 0.00000713, 0.00001240, 0.00000648, -0.00000150, -0.00000375, -0.00000223, -0.00000087, 0.00000020, 0.00000007, -0.00000453, -0.00001196, -0.00001375, -0.00000541, 0.00000586, 0.00000934, 0.00000322, -0.00000556, -0.00001019, -0.00000776, 0.00000119, 0.00001028, 0.00000779, -0.00001117, -0.00003363, -0.00003939, -0.00002471, -0.00000629, 0.00000238, 0.00000864, 0.00002580, 0.00004849, 0.00005626, 0.00004153, 0.00002208, 0.00002092, 0.00004057, 0.00006149, 0.00006114, 0.00003179, -0.00001404, -0.00005064, -0.00005719, -0.00003608, },
		{ -0.00000844, -0.00001022, -0.00000385, 0.00000168, 0.00000101, -0.00000023, 0.00000206, 0.00000202, -0.00000500, -0.00001153, -0.00000824, 0.00000090, 0.00000463, 0.00000205, 0.00000188, 0.00000488, 0.00000115, -0.00001102, -0.00001738, -0.00000768, 0.00000705, 0.00000947, 0.00000088, -0.00000327, 0.00000058, -0.00000067, -0.00001289, -0.00002185, -0.00001330, 0.00000554, 0.00001556, 0.00001012, -0.00000117, -0.00000792, -0.00000694, 0.00000145, 0.00001236, 0.00001354, -0.00000200, -0.00001964, -0.00001574, 0.00000776, 0.00001923, -0.00000016, -0.00002635, -0.00002462, -0.00000013, 0.00001065, -0.00000347, -0.00001179, 0.00000776, 0.00003092, 0.00002343, -0.00000548, -0.00001771, -0.00000515, 0.00000173, -0.00001211, -0.00002009, 0.00000001, 0.00002405, 0.00001613, -0.00001296, -0.00001933, 0.00000597, 0.00002173, 0.00000034, -0.00002812, -0.00002094, 0.00001255, 0.00002562, 0.00000365, -0.00001768, -0.00000845, 0.00001379, 0.00001592, -0.00000140, -0.00001104, -0.00000364, 0.00000285, -0.00000332, -0.00000961, -0.00000240, 0.00001044, 0.00001259, 0.00000353, -0.00000393, -0.00000364, -0.00000089, -0.00000010, -0.00000009, -0.00000128, -0.00000684, -0.00001383, -0.00001287, -0.00000130, 0.00001028, 0.00001013, -0.00000055, -0.00001049, -0.00001118, -0.00000274, 0.00000835, 0.00001210, 0.00000034, -0.00002288, -0.00003915, -0.00003316, -0.00001173, 0.00000414, 0.00000706, 0.00001172, 0.00003025, 0.00005039, 0.00005043, 0.00003023, 0.00001411, 0.00002142, 0.00004457, 0.00005953, 0.00004991, 0.00001765, -0.00002179, -0.00004847, -0.00004930, -0.00002909, -0.00000826, -0.00000280, },
		{ -0.00000271, 0.00000274, 0.00000187, -0.00000029, 0.00000057, 0.00000023, -0.00000511, -0.00000993, -0.00000724, 0.00000005, 0.00000397, 0.00000381, 0.00000426, 0.00000391, -0.00000302, -0.00001308, -0.00001391, -0.00000250, 0.00000769, 0.00000544, -0.00000194, -0.00000167, 0.00000273, -0.00000276, -0.00001742, -0.00002265, -0.00000827, 0.00001142, 0.00001621, 0.00000520, -0.00000619, -0.00000743, -0.00000072, 0.00000738, 0.00001104, 0.00000489, -0.00000935, -0.00001727, -0.00000643, 0.00001217, 0.00001278, -0.00000928, -0.00002692, -0.00001805, 0.00000230, 0.00000553, -0.00000588, -0.00000339, 0.00001908, 0.00003132, 0.00001262, -0.00001444, -0.00001754, -0.00000285, -0.00000091, -0.00001419, -0.00001347, 0.00001010, 0.00002458, 0.00000637, -0.00001878, -0.00001296, 0.00001386, 0.00001745, -0.00001154, -0.00003071, -0.00000975, 0.00002293, 0.00002244, -0.00000671, -0.00002041, -0.00000151, 0.00001904, 0.00001261, -0.00000699, -0.00001091, 0.00000022, 0.00000348, -0.00000640, -0.00001059, 0.00000098, 0.00001356, 0.00001129, -0.00000021, -0.00000564, -0.00000247, 0.00000067, 0.00000011, -0.00000102, -0.00000306, -0.00000915, -0.00001483, -0.00001031, 0.00000395, 0.00001372, 0.00000814, -0.00000631, -0.00001414, -0.00000829, 0.00000487, 0.00001349, 0.00000843, -0.00001084, -0.00003226, -0.00003770, -0.00002136, 0.00000094, 0.00000982, 0.00000758, 0.00001390, 0.00003501, 0.00005097, 0.00004221, 0.00001861, 0.00000891, 0.00002486, 0.00004846, 0.00005486, 0.00003701, 0.00000524, -0.00002584, -0.00004389, -0.00004170, -0.00002382, -0.00000654, -0.00000270, -0.00000801, -0.00000916, },
		{ 0.00000196, -0.00000096, -0.00000058, -0.00000049, -0.00000478, -0.00000921, -0.00000726, -0.00000036, 0.00000486, 0.00000608, 0.00000513, 0.00000131, -0.00000656, -0.00001282, -0.00000945, 0.00000091, 0.00000603, 0.00000154, -0.00000259, 0.00000098, 0.00000326, -0.00000686, -0.00002134, -0.00002058, -0.00000161, 0.00001545, 0.00001350, -0.00000057, -0.00000822, -0.00000329, 0.00000541, 0.00000899, 0.00000566, -0.00000292, -0.00001173, -0.00001152, 0.00000080, 0.00001188, 0.00000494, -0.00001527, -0.00002475, -0.00001297, 0.00000194, 0.00000165, -0.00000352, 0.00000699, 0.00002628, 0.00002598, 0.00000066, -0.00002027, -0.00001513, -0.00000102, -0.00000328, -0.00001343, -0.00000520, 0.00001718, 0.00002065, -0.00000327, -0.00002021, -0.00000470, 0.00001791, 0.00000907, -0.00002131, -0.00002718, 0.00000350, 0.00002866, 0.00001424, -0.00001623, -0.00001859, 0.00000714, 0.00002151, 0.00000671, -0.00001152, -0.00000834, 0.00000417, 0.00000221, -0.00001012, -0.00001006, 0.00000553, 0.00001571, 0.00000826, -0.00000416, -0.00000608, -0.00000033, 0.00000195, -0.00000053, -0.00000257, -0.00000502, -0.00001099, -0.00001441, -0.00000609, 0.00000931, 0.00001486, 0.00000325, -0.00001237, -0.00001449, -0.00000155, 0.00001255, 0.00001394, -0.00000053, -0.00002233, -0.00003584, -0.00002924, -0.00000754, 0.00000953, 0.00001009, 0.00000627, 0.00001747, 0.00004022, 0.00004899, 0.00003130, 0.00000787, 0.00000756, 0.00003079, 0.00005074, 0.00004682, 0.00002332, -0.00000410, -0.00002593, -0.00003787, -0.00003556, -0.00002064, -0.00000567, -0.00000225, -0.00000703, -0.00000784, -0.00000156, 0.00000343, },
		{ -0.00000119, -0.00000068, -0.00000469, -0.00000930, -0.00000748, 0.00000023, 0.00000677, 0.00000785, 0.00000435, -0.00000191, -0.00000860, -0.00001072, -0.00000530, 0.00000214, 0.00000316, -0.00000109, -0.00000140, 0.00000328, 0.00000164, -0.00001178, -0.00002318, -0.00001591, 0.00000482, 0.00001645, 0.00000857, -0.00000491, -0.00000660, 0.00000246, 0.00000891, 0.00000625, -0.00000113, -0.00000754, -0.00000977, -0.00000513, 0.00000445, 0.00000825, -0.00000216, -0.00001811, -0.00002164, -0.00001004, 0.00000052, 0.00000078, 0.00000265, 0.00001649, 0.00002794, 0.00001647, -0.00001016, -0.00002247, -0.00001135, 0.00000018, -0.00000474, -0.00001020, 0.00000299, 0.00002038, 0.00001378, -0.00001084, -0.00001772, 0.00000314, 0.00001738, -0.00000113, -0.00002665, -0.00001844, 0.00001569, 0.00002847, 0.00000306, -0.00002244, -0.00001242, 0.00001535, 0.00002026, -0.00000055, -0.00001366, -0.00000371, 0.00000697, -0.00000113, -0.00001351, -0.00000750, 0.00001059, 0.00001607, 0.00000370, -0.00000748, -0.00000486, 0.00000236, 0.00000239, -0.00000219, -0.00000451, -0.00000672, -0.00001181, -0.00001232, -0.00000083, 0.00001342, 0.00001289, -0.00000353, -0.00001649, -0.00001046, 0.00000737, 0.00001717, 0.00000845, -0.00001235, -0.00003007, -0.00003195, -0.00001635, 0.00000395, 0.00001204, 0.00000673, 0.00000615, 0.00002343, 0.00004442, 0.00004312, 0.00001843, -0.00000010, 0.00001069, 0.00003773, 0.00004973, 0.00003550, 0.00001053, -0.00000915, -0.00002263, -0.00003206, -0.00003186, -0.00001918, -0.00000461, -0.00000083, -0.00000564, -0.00000681, -0.00000090, 0.00000367, 0.00000168, -0.00000163, },
		{ -0.00000500, -0.00000973, -0.00000725, 0.00000180, 0.00000894, 0.00000854, 0.00000237, -0.00000476, -0.00000893, -0.00000773, -0.00000236, 0.00000154, 0.00000023, -0.00000204, 0.00000065, 0.00000421, -0.00000170, -0.00001605, -0.00002238, -0.00000988, 0.00000938, 0.00001453, 0.00000333, -0.00000635, -0.00000225, 0.00000738, 0.00000858, 0.00000070, -0.00000667, -0.00000815, -0.00000537, -0.00000036, 0.00000450, 0.00000321, -0.00000732, -0.00001873, -0.00001895, -0.00000878, -0.00000007, 0.00000345, 0.00001055, 0.00002262, 0.00002413, 0.00000509, -0.00001807, -0.00002135, -0.00000711, 0.00000085, -0.00000489, -0.00000535, 0.00000964, 0.00001979, 0.00000581, -0.00001527, -0.00001270, 0.00000869, 0.00001292, -0.00001054, -0.00002653, -0.00000676, 0.00002409, 0.00002266, -0.00000832, -0.00002385, -0.00000337, 0.00002106, 0.00001541, -0.00000740, -0.00001266, 0.00000187, 0.00000754, -0.00000607, -0.00001547, -0.00000288, 0.00001512, 0.00001410, -0.00000175, -0.00000928, -0.00000199, 0.00000487, 0.00000144, -0.00000476, -0.00000632, -0.00000755, -0.00001120, -0.00000881, 0.00000431, 0.00001506, 0.00000802, -0.00001022, -0.00001676, -0.00000256, 0.00001554, 0.00001629, -0.00000201, -0.00002312, -0.00003101, -0.00002165, -0.00000327, 0.00000976, 0.00000906, 0.00000303, 0.00000941, 0.00003080, 0.00004536, 0.00003297, 0.00000565, -0.00000340, 0.00001768, 0.00004340, 0.00004426, 0.00002219, 0.00000074, -0.00000946, -0.00001751, -0.00002819, -0.00003078, -0.00001825, -0.00000233, 0.00000153, -0.00000456, -0.00000666, -0.00000075, 0.00000387, 0.00000149, -0.00000213, -0.00000155, -0.00000084, },
		{ -0.00000642, 0.00000392, 0.00001077, 0.00000813, -0.00000012, -0.00000670, -0.00000793, -0.00000476, -0.00000090, -0.00000009, -0.00000194, -0.00000163, 0.00000251, 0.00000349, -0.00000572, -0.00001860, -0.00001937, -0.00000409, 0.00001131, 0.00001085, -0.00000039, -0.00000470, 0.00000293, 0.00000956, 0.00000476, -0.00000534, -0.00000920, -0.00000550, -0.00000079, 0.00000160, 0.00000199, -0.00000157, -0.00001038, -0.00001836, -0.00001722, -0.00000792, 0.00000154, 0.00000885, 0.00001765, 0.00002402, 0.00001619, -0.00000574, -0.00002221, -0.00001779, -0.00000314, 0.00000128, -0.00000366, 0.00000004, 0.00001382, 0.00001623, -0.00000159, -0.00001648, -0.00000689, 0.00001102, 0.00000617, -0.00001696, -0.00002144, 0.00000500, 0.00002712, 0.00001298, -0.00001722, -0.00002028, 0.00000632, 0.00002288, 0.00000811, -0.00001213, -0.00000869, 0.00000688, 0.00000535, -0.00001153, -0.00001505, 0.00000321, 0.00001798, 0.00000976, -0.00000700, -0.00000886, 0.00000204, 0.00000628, -0.00000115, -0.00000772, -0.00000728, -0.00000703, -0.00000922, -0.00000474, 0.00000802, 0.00001370, 0.00000154, -0.00001459, -0.00001247, 0.00000707, 0.00001981, 0.00000925, -0.00001432, -0.00002882, -0.00002439, -0.00000841, 0.00000563, 0.00000890, 0.00000352, 0.00000233, 0.00001626, 0.00003707, 0.00004128, 0.00001981, -0.00000432, -0.00000118, 0.00002655, 0.00004545, 0.00003441, 0.00000919, -0.00000438, -0.00000585, -0.00001280, -0.00002738, -0.00003144, -0.00001631, 0.00000166, 0.00000405, -0.00000471, -0.00000749, -0.00000056, 0.00000453, 0.00000147, -0.00000273, -0.00000200, -0.00000111, -0.00000546, -0.00001005, },
		{ 0.00001202, 0.00000694, -0.00000258, -0.00000764, -0.00000618, -0.00000236, -0.00000068, -0.00000201, -0.00000314, -0.00000056, 0.00000354, 0.00000153, -0.00000931, -0.00001913, -0.00001525, 0.00000025, 0.00001092, 0.00000709, -0.00000160, -0.00000102, 0.00000688, 0.00000829, -0.00000098, -0.00000972, -0.00000833, -0.00000125, 0.00000220, 0.00000080, -0.00000157, -0.00000512, -0.00001193, -0.00001794, -0.00001606, -0.00000609, 0.00000562, 0.00001521, 0.00002186, 0.00002067, 0.00000625, -0.00001413, -0.00002264, -0.00001292, 0.00000014, 0.00000180, -0.00000135, 0.00000492, 0.00001525, 0.00001093, -0.00000730, -0.00001518, -0.00000184, 0.00001027, -0.00000081, -0.00001924, -0.00001311, 0.00001421, 0.00002468, 0.00000201, -0.00002180, -0.00001294, 0.00001434, 0.00002056, 0.00000023, -0.00001357, -0.00000277, 0.00000984, 0.00000062, -0.00001610, -0.00001189, 0.00000962, 0.00001822, 0.00000366, -0.00001081, -0.00000602, 0.00000624, 0.00000573, -0.00000519, -0.00001021, -0.00000667, -0.00000505, -0.00000649, -0.00000128, 0.00000940, 0.00000989, -0.00000449, -0.00001498, -0.00000456, 0.00001529, 0.00001808, -0.00000227, -0.00002429, -0.00002697, -0.00001227, 0.00000307, 0.00000768, 0.00000318, -0.00000065, 0.00000635, 0.00002474, 0.00003935, 0.00003201, 0.00000637, -0.00000915, 0.00000573, 0.00003444, 0.00004245, 0.00002180, -0.00000093, -0.00000433, -0.00000033, -0.00001058, -0.00002956, -0.00003207, -0.00001216, 0.00000683, 0.00000546, -0.00000657, -0.00000872, 0.00000042, 0.00000578, 0.00000122, -0.00000380, -0.00000257, -0.00000123, -0.00000579, -0.00001011, -0.00000515, 0.00000618, },
		{ -0.00000477, -0.00000778, -0.00000416, -0.00000067, -0.00000127, -0.00000378, -0.00000358, 0.00000056, 0.00000365, -0.00000089, -0.00001179, -0.00001809, -0.00001122, 0.00000270, 0.00000929, 0.00000461, -0.00000042, 0.00000298, 0.00000822, 0.00000421, -0.00000661, -0.00001119, -0.00000491, 0.00000269, 0.00000276, -0.00000188, -0.00000490, -0.00000730, -0.00001276, -0.00001773, -0.00001454, -0.00000246, 0.00001136, 0.00002046, 0.00002203, 0.00001375, -0.00000343, -0.00001906, -0.00002015, -0.00000779, 0.00000267, 0.00000266, 0.00000154, 0.00000849, 0.00001422, 0.00000514, -0.00001087, -0.00001245, 0.00000154, 0.00000743, -0.00000623, -0.00001749, -0.00000389, 0.00001921, 0.00001810, -0.00000761, -0.00002152, -0.00000390, 0.00001897, 0.00001497, -0.00000628, -0.00001151, 0.00000348, 0.00000981, -0.00000563, -0.00001849, -0.00000633, 0.00001492, 0.00001547, -0.00000301, -0.00001212, -0.00000122, 0.00000935, 0.00000277, -0.00000989, -0.00001113, -0.00000408, -0.00000210, -0.00000408, 0.00000050, 0.00000835, 0.00000514, -0.00000804, -0.00001114, 0.00000452, 0.00001912, 0.00001031, -0.00001469, -0.00002814, -0.00001774, 0.00000114, 0.00000868, 0.00000299, -0.00000359, -0.00000023, 0.00001442, 0.00003152, 0.00003585, 0.00001937, -0.00000412, -0.00000800, 0.00001494, 0.00003867, 0.00003459, 0.00000910, -0.00000630, -0.00000017, 0.00000450, -0.00001192, -0.00003338, -0.00003082, -0.00000560, 0.00001182, 0.00000465, -0.00000984, -0.00000929, 0.00000268, 0.00000718, 0.00000012, -0.00000544, -0.00000290, -0.00000089, -0.00000592, -0.00001008, -0.00000370, 0.00000841, 0.00001277, 0.00000527, },
		{ -0.00000207, 0.00000039, -0.00000236, -0.00000532, -0.00000359, 0.00000147, 0.00000317, -0.00000311, -0.00001311, -0.00001632, -0.00000807, 0.00000366, 0.00000770, 0.00000397, 0.00000209, 0.00000568, 0.00000665, -0.00000117, -0.00001039, -0.00000967, -0.00000047, 0.00000495, 0.00000102, -0.00000524, -0.00000730, -0.00000848, -0.00001342, -0.00001735, -0.00001176, 0.00000293, 0.00001721, 0.00002292, 0.00001829, 0.00000518, -0.00001106, -0.00002045, -0.00001588, -0.00000310, 0.00000460, 0.00000390, 0.00000435, 0.00001032, 0.00001139, -0.00000015, -0.00001246, -0.00000937, 0.00000309, 0.00000391, -0.00000903, -0.00001290, 0.00000398, 0.00001964, 0.00000954, -0.00001395, -0.00001720, 0.00000454, 0.00001958, 0.00000785, -0.00000993, -0.00000673, 0.00000833, 0.00000667, -0.00001192, -0.00001791, 0.00000054, 0.00001784, 0.00001014, -0.00000874, -0.00001043, 0.00000441, 0.00001013, -0.00000240, -0.00001396, -0.00000956, 0.00000026, 0.00000075, -0.00000318, 0.00000015, 0.00000582, 0.00000140, -0.00000793, -0.00000439, 0.00001167, 0.00001695, -0.00000127, -0.00002372, -0.00002404, -0.00000403, 0.00001094, 0.00000658, -0.00000554, -0.00000713, 0.00000584, 0.00002363, 0.00003344, 0.00002683, 0.00000665, -0.00000922, -0.00000200, 0.00002326, 0.00003780, 0.00002372, -0.00000094, -0.00000654, 0.00000578, 0.00000650, -0.00001646, -0.00003670, -0.00002649, 0.00000242, 0.00001496, 0.00000127, -0.00001341, -0.00000818, 0.00000601, 0.00000784, -0.00000221, -0.00000722, -0.00000243, 0.00000001, -0.00000616, -0.00001022, -0.00000209, 0.00001070, 0.00001314, 0.00000325, -0.00000672, -0.00000733, },
		{ -0.00000388, -0.00000672, -0.00000333, 0.00000226, 0.00000254, -0.00000490, -0.00001367, -0.00001456, -0.00000590, 0.00000401, 0.00000702, 0.00000487, 0.00000452, 0.00000614, 0.00000293, -0.00000612, -0.00001149, -0.00000607, 0.00000340, 0.00000495, -0.00000221, -0.00000829, -0.00000864, -0.00000917, -0.00001392, -0.00001609, -0.00000733, 0.00000916, 0.00002148, 0.00002188, 0.00001176, -0.00000306, -0.00001569, -0.00001900, -0.00001092, 0.00000082, 0.00000618, 0.00000536, 0.00000650, 0.00001039, 0.00000761, -0.00000437, -0.00001257, -0.00000666, 0.00000326, 0.00000094, -0.00000915, -0.00000719, 0.00000899, 0.00001636, 0.00000131, -0.00001623, -0.00001054, 0.00001054, 0.00001670, 0.00000117, -0.00001015, -0.00000073, 0.00001051, 0.00000116, -0.00001669, -0.00001438, 0.00000724, 0.00001763, 0.00000335, -0.00001215, -0.00000602, 0.00000930, 0.00000786, -0.00000878, -0.00001591, -0.00000512, 0.00000537, 0.00000208, -0.00000453, -0.00000182, 0.00000350, 0.00000029, -0.00000442, 0.00000269, 0.00001430, 0.00000934, -0.00001282, -0.00002601, -0.00001299, 0.00000939, 0.00001347, -0.00000204, -0.00001337, -0.00000456, 0.00001589, 0.00003007, 0.00002897, 0.00001476, -0.00000267, -0.00000837, 0.00000608, 0.00002800, 0.00003227, 0.00001269, -0.00000653, -0.00000296, 0.00001086, 0.00000478, -0.00002249, -0.00003754, -0.00001917, 0.00001010, 0.00001508, -0.00000397, -0.00001581, -0.00000500, 0.00000951, 0.00000690, -0.00000558, -0.00000833, -0.00000077, 0.00000111, -0.00000699, -0.00001057, -0.00000009, 0.00001320, 0.00001309, 0.00000078, -0.00000844, -0.00000633, 0.00000008, 0.00000090, },
		{ -0.00000269, 0.00000315, 0.00000192, -0.00000641, -0.00001399, -0.00001311, -0.00000428, 0.00000456, 0.00000744, 0.00000635, 0.00000572, 0.00000438, -0.00000160, -0.00000933, -0.00001002, -0.00000175, 0.00000561, 0.00000289, -0.00000595, -0.00001049, -0.00000918, -0.00000960, -0.00001389, -0.00001343, -0.00000164, 0.00001483, 0.00002297, 0.00001765, 0.00000415, -0.00000947, -0.00001728, -0.00001570, -0.00000606, 0.00000400, 0.00000759, 0.00000670, 0.00000759, 0.00000899, 0.00000363, -0.00000731, -0.00001180, -0.00000458, 0.00000284, -0.00000072, -0.00000732, -0.00000207, 0.00001069, 0.00001102, -0.00000480, -0.00001486, -0.00000359, 0.00001324, 0.00001173, -0.00000348, -0.00000737, 0.00000473, 0.00000954, -0.00000532, -0.00001884, -0.00000870, 0.00001225, 0.00001439, -0.00000329, -0.00001240, -0.00000003, 0.00001191, 0.00000263, -0.00001473, -0.00001451, 0.00000168, 0.00000960, 0.00000069, -0.00000800, -0.00000394, 0.00000314, 0.00000232, 0.00000074, 0.00000725, 0.00001145, -0.00000109, -0.00002027, -0.00002046, 0.00000142, 0.00001772, 0.00000775, -0.00001335, -0.00001614, 0.00000433, 0.00002616, 0.00003056, 0.00001903, 0.00000339, -0.00000634, -0.00000316, 0.00001296, 0.00002801, 0.00002418, 0.00000415, -0.00000762, 0.00000200, 0.00001310, 0.00000007, -0.00002771, -0.00003489, -0.00001020, 0.00001557, 0.00001204, -0.00000955, -0.00001590, -0.00000027, 0.00001192, 0.00000399, -0.00000918, -0.00000794, 0.00000192, 0.00000170, -0.00000872, -0.00001082, 0.00000271, 0.00001585, 0.00001230, -0.00000224, -0.00000978, -0.00000463, 0.00000223, 0.00000072, -0.00000589, -0.00000795, },
		{ 0.00000116, -0.00000797, -0.00001433, -0.00001180, -0.00000263, 0.00000569, 0.00000849, 0.00000739, 0.00000520, 0.00000115, -0.00000558, -0.00001025, -0.00000685, 0.00000198, 0.00000573, -0.00000056, -0.00000942, -0.00001169, -0.00000915, -0.00000966, -0.00001283, -0.00000937, 0.00000433, 0.00001861, 0.00002138, 0.00001140, -0.00000288, -0.00001333, -0.00001640, -0.00001155, -0.00000169, 0.00000661, 0.00000878, 0.00000752, 0.00000748, 0.00000665, 0.00000003, -0.00000909, -0.00001061, -0.00000302, 0.00000252, -0.00000098, -0.00000470, 0.00000131, 0.00000961, 0.00000546, -0.00000795, -0.00001112, 0.00000196, 0.00001283, 0.00000642, -0.00000527, -0.00000282, 0.00000822, 0.00000583, -0.00001119, -0.00001801, -0.00000227, 0.00001456, 0.00000902, -0.00000821, -0.00000953, 0.00000586, 0.00001121, -0.00000450, -0.00001837, -0.00000933, 0.00000937, 0.00001117, -0.00000388, -0.00001236, -0.00000435, 0.00000580, 0.00000655, 0.00000496, 0.00000734, 0.00000427, -0.00001048, -0.00002087, -0.00000880, 0.00001423, 0.00001788, -0.00000405, -0.00002233, -0.00001150, 0.00001690, 0.00003235, 0.00002406, 0.00000679, -0.00000374, -0.00000433, 0.00000329, 0.00001618, 0.00002415, 0.00001639, -0.00000052, -0.00000580, 0.00000583, 0.00001203, -0.00000569, -0.00003028, -0.00002918, -0.00000161, 0.00001775, 0.00000687, -0.00001381, -0.00001343, 0.00000477, 0.00001221, -0.00000043, -0.00001182, -0.00000570, 0.00000494, 0.00000104, -0.00001120, -0.00001031, 0.00000650, 0.00001824, 0.00001038, -0.00000569, -0.00001037, -0.00000215, 0.00000412, -0.00000040, -0.00000835, -0.00000875, -0.00000143, 0.00000422, },
		{ -0.00001453, -0.00001021, -0.00000058, 0.00000723, 0.00000944, 0.00000732, 0.00000319, -0.00000244, -0.00000807, -0.00000910, -0.00000310, 0.00000424, 0.00000393, -0.00000461, -0.00001212, -0.00001198, -0.00000862, -0.00000905, -0.00001056, -0.00000450, 0.00000935, 0.00001976, 0.00001725, 0.00000461, -0.00000821, -0.00001469, -0.00001392, -0.00000723, 0.00000214, 0.00000874, 0.00000955, 0.00000752, 0.00000632, 0.00000395, -0.00000286, -0.00000992, -0.00000924, -0.00000166, 0.00000271, -0.00000030, -0.00000243, 0.00000258, 0.00000694, 0.00000112, -0.00000831, -0.00000666, 0.00000517, 0.00001033, 0.00000230, -0.00000433, 0.00000191, 0.00000901, 0.00000049, -0.00001520, -0.00001469, 0.00000343, 0.00001394, 0.00000300, -0.00001035, -0.00000452, 0.00000991, 0.00000710, -0.00001167, -0.00001819, -0.00000104, 0.00001579, 0.00000875, -0.00001090, -0.00001559, -0.00000146, 0.00001119, 0.00001080, 0.00000582, 0.00000280, -0.00000428, -0.00001522, -0.00001430, 0.00000496, 0.00002094, 0.00000974, -0.00001743, -0.00002477, -0.00000007, 0.00002870, 0.00003142, 0.00001224, -0.00000362, -0.00000477, 0.00000118, 0.00000771, 0.00001512, 0.00001883, 0.00001122, -0.00000179, -0.00000355, 0.00000704, 0.00000880, -0.00001028, -0.00002958, -0.00002205, 0.00000480, 0.00001677, 0.00000128, -0.00001564, -0.00000917, 0.00000865, 0.00001013, -0.00000523, -0.00001255, -0.00000192, 0.00000724, -0.00000121, -0.00001380, -0.00000841, 0.00001109, 0.00001972, 0.00000714, -0.00000912, -0.00000980, 0.00000095, 0.00000524, -0.00000262, -0.00001092, -0.00000868, 0.00000056, 0.00000520, -0.00000009, -0.00000978, },
		{ 0.00000177, 0.00000865, 0.00000967, 0.00000606, 0.00000037, -0.00000545, -0.00000878, -0.00000659, 0.00000021, 0.00000470, 0.00000073, -0.00000858, -0.00001381, -0.00001137, -0.00000747, -0.00000756, -0.00000737, 0.00000019, 0.00001248, 0.00001832, 0.00001173, -0.00000141, -0.00001140, -0.00001409, -0.00001061, -0.00000305, 0.00000547, 0.00001033, 0.00000962, 0.00000660, 0.00000449, 0.00000139, -0.00000492, -0.00001007, -0.00000774, -0.00000022, 0.00000339, 0.00000058, -0.00000125, 0.00000212, 0.00000400, -0.00000130, -0.00000679, -0.00000286, 0.00000603, 0.00000713, 0.00000023, -0.00000156, 0.00000546, 0.00000720, -0.00000506, -0.00001672, -0.00000995, 0.00000724, 0.00001096, -0.00000212, -0.00000947, 0.00000106, 0.00001089, 0.00000058, -0.00001677, -0.00001362, 0.00000854, 0.00001878, 0.00000209, -0.00001846, -0.00001553, 0.00000519, 0.00001761, 0.00001247, 0.00000216, -0.00000455, -0.00001054, -0.00001338, -0.00000295, 0.00001597, 0.00001923, -0.00000376, -0.00002720, -0.00001890, 0.00001469, 0.00003520, 0.00002290, -0.00000108, -0.00000871, -0.00000032, 0.00000674, 0.00000803, 0.00001124, 0.00001486, 0.00000942, -0.00000163, -0.00000305, 0.00000579, 0.00000553, -0.00001223, -0.00002649, -0.00001554, 0.00000824, 0.00001391, -0.00000306, -0.00001501, -0.00000454, 0.00001037, 0.00000632, -0.00000909, -0.00001103, 0.00000240, 0.00000793, -0.00000476, -0.00001558, -0.00000485, 0.00001582, 0.00001965, 0.00000279, -0.00001185, -0.00000787, 0.00000415, 0.00000511, -0.00000579, -0.00001298, -0.00000740, 0.00000302, 0.00000560, -0.00000207, -0.00001164, -0.00001421, -0.00000803, },
		{ 0.00000897, 0.00000397, -0.00000248, -0.00000732, -0.00000795, -0.00000357, 0.00000238, 0.00000338, -0.00000331, -0.00001196, -0.00001432, -0.00000982, -0.00000557, -0.00000533, -0.00000398, 0.00000375, 0.00001337, 0.00001502, 0.00000608, -0.00000591, -0.00001262, -0.00001221, -0.00000694, 0.00000090, 0.00000828, 0.00001117, 0.00000876, 0.00000492, 0.00000248, -0.00000067, -0.00000622, -0.00000972, -0.00000607, 0.00000139, 0.00000421, 0.00000097, -0.00000133, 0.00000076, 0.00000184, -0.00000186, -0.00000459, -0.00000054, 0.00000524, 0.00000453, 0.00000028, 0.00000177, 0.00000699, 0.00000356, -0.00000958, -0.00001589, -0.00000511, 0.00000871, 0.00000684, -0.00000518, -0.00000630, 0.00000544, 0.00000852, -0.00000651, -0.00001809, -0.00000536, 0.00001690, 0.00001683, -0.00000761, -0.00002396, -0.00001073, 0.00001451, 0.00002237, 0.00000962, -0.00000528, -0.00001149, -0.00001157, -0.00000553, 0.00000898, 0.00002034, 0.00000980, -0.00001774, -0.00002951, -0.00000613, 0.00002786, 0.00003363, 0.00000917, -0.00001144, -0.00000705, 0.00000683, 0.00000908, 0.00000427, 0.00000730, 0.00001414, 0.00000984, -0.00000256, -0.00000510, 0.00000377, 0.00000425, -0.00001155, -0.00002290, -0.00001117, 0.00000927, 0.00001097, -0.00000533, -0.00001291, -0.00000100, 0.00000986, 0.00000207, -0.00001101, -0.00000777, 0.00000608, 0.00000667, -0.00000878, -0.00001577, 0.00000008, 0.00001979, 0.00001777, -0.00000201, -0.00001327, -0.00000482, 0.00000670, 0.00000349, -0.00000932, -0.00001388, -0.00000494, 0.00000537, 0.00000497, -0.00000465, -0.00001307, -0.00001300, -0.00000535, 0.00000398, 0.00000946, },
		{ -0.00000480, -0.00000794, -0.00000607, -0.00000077, 0.00000306, 0.00000057, -0.00000758, -0.00001430, -0.00001345, -0.00000728, -0.00000309, -0.00000293, -0.00000123, 0.00000566, 0.00001239, 0.00001091, 0.00000121, -0.00000874, -0.00001237, -0.00000956, -0.00000310, 0.00000459, 0.00001037, 0.00001098, 0.00000701, 0.00000289, 0.00000076, -0.00000209, -0.00000691, -0.00000898, -0.00000419, 0.00000305, 0.00000472, 0.00000050, -0.00000233, -0.00000056, 0.00000094, -0.00000122, -0.00000276, 0.00000024, 0.00000385, 0.00000328, 0.00000184, 0.00000447, 0.00000639, -0.00000076, -0.00001237, -0.00001347, -0.00000126, 0.00000811, 0.00000292, -0.00000580, -0.00000219, 0.00000736, 0.00000360, -0.00001201, -0.00001492, 0.00000464, 0.00002158, 0.00000980, -0.00001790, -0.00002494, -0.00000121, 0.00002402, 0.00002284, 0.00000192, -0.00001395, -0.00001462, -0.00000642, 0.00000537, 0.00001700, 0.00001663, -0.00000384, -0.00002719, -0.00002327, 0.00000953, 0.00003489, 0.00002408, -0.00000543, -0.00001547, 0.00000010, 0.00001285, 0.00000653, -0.00000148, 0.00000614, 0.00001671, 0.00001001, -0.00000638, -0.00000876, 0.00000331, 0.00000583, -0.00000972, -0.00002079, -0.00000916, 0.00000945, 0.00000933, -0.00000588, -0.00001085, 0.00000071, 0.00000798, -0.00000128, -0.00001083, -0.00000386, 0.00000820, 0.00000386, -0.00001222, -0.00001406, 0.00000560, 0.00002226, 0.00001433, -0.00000645, -0.00001313, -0.00000132, 0.00000801, 0.00000065, -0.00001240, -0.00001327, -0.00000179, 0.00000688, 0.00000318, -0.00000731, -0.00001352, -0.00001089, -0.00000262, 0.00000556, 0.00000943, 0.00000757, 0.00000160, },
		{ -0.00000367, 0.00000125, 0.00000212, -0.00000333, -0.00001147, -0.00001514, -0.00001108, -0.00000394, -0.00000053, -0.00000108, 0.00000030, 0.00000601, 0.00001033, 0.00000692, -0.00000249, -0.00001015, -0.00001111, -0.00000637, 0.00000085, 0.00000781, 0.00001141, 0.00000963, 0.00000464, 0.00000101, -0.00000037, -0.00000292, -0.00000717, -0.00000794, -0.00000219, 0.00000448, 0.00000453, -0.00000080, -0.00000361, -0.00000117, 0.00000118, -0.00000024, -0.00000186, -0.00000002, 0.00000282, 0.00000349, 0.00000393, 0.00000578, 0.00000416, -0.00000473, -0.00001333, -0.00001046, 0.00000108, 0.00000632, 0.00000026, -0.00000444, 0.00000130, 0.00000642, -0.00000218, -0.00001418, -0.00000792, 0.00001372, 0.00002096, -0.00000084, -0.00002574, -0.00002001, 0.00001125, 0.00003056, 0.00001749, -0.00000897, -0.00002036, -0.00001178, 0.00000343, 0.00001506, 0.00001808, 0.00000638, -0.00001674, -0.00002886, -0.00001053, 0.00002295, 0.00003341, 0.00000947, -0.00001627, -0.00001216, 0.00000943, 0.00001450, -0.00000026, -0.00000601, 0.00000926, 0.00002070, 0.00000730, -0.00001318, -0.00001168, 0.00000622, 0.00000948, -0.00000897, -0.00002120, -0.00000836, 0.00001055, 0.00000932, -0.00000600, -0.00001005, 0.00000085, 0.00000607, -0.00000298, -0.00000924, -0.00000049, 0.00000856, 0.00000040, -0.00001430, -0.00001080, 0.00001077, 0.00002296, 0.00001004, -0.00000988, -0.00001174, 0.00000184, 0.00000785, -0.00000271, -0.00001431, -0.00001129, 0.00000123, 0.00000708, 0.00000062, -0.00000931, -0.00001277, -0.00000832, -0.00000044, 0.00000625, 0.00000877, 0.00000589, -0.00000063, -0.00000639, -0.00000748, },
		{ -0.00000044, -0.00000780, -0.00001427, -0.00001408, -0.00000732, -0.00000035, 0.00000133, -0.00000038, 0.00000059, 0.00000539, 0.00000804, 0.00000354, -0.00000512, -0.00001054, -0.00000906, -0.00000270, 0.00000475, 0.00001018, 0.00001111, 0.00000727, 0.00000217, -0.00000026, -0.00000084, -0.00000335, -0.00000716, -0.00000666, -0.00000022, 0.00000537, 0.00000351, -0.00000257, -0.00000452, -0.00000083, 0.00000204, 0.00000036, -0.00000191, -0.00000056, 0.00000273, 0.00000473, 0.00000564, 0.00000552, 0.00000114, -0.00000766, -0.00001291, -0.00000772, 0.00000198, 0.00000437, -0.00000074, -0.00000217, 0.00000303, 0.00000326, -0.00000683, -0.00001229, 0.00000101, 0.00001931, 0.00001493, -0.00001233, -0.00002844, -0.00000954, 0.00002354, 0.00003140, 0.00000674, -0.00001982, -0.00002141, -0.00000302, 0.00001443, 0.00001959, 0.00001178, -0.00000649, -0.00002438, -0.00002242, 0.00000434, 0.00002987, 0.00002402, -0.00000555, -0.00002018, -0.00000327, 0.00001684, 0.00001046, -0.00000865, -0.00000646, 0.00001607, 0.00002310, 0.00000022, -0.00002119, -0.00001122, 0.00001278, 0.00001301, -0.00001120, -0.00002364, -0.00000684, 0.00001357, 0.00001003, -0.00000721, -0.00001081, 0.00000059, 0.00000529, -0.00000319, -0.00000737, 0.00000164, 0.00000766, -0.00000273, -0.00001484, -0.00000678, 0.00001492, 0.00002217, 0.00000566, -0.00001211, -0.00000975, 0.00000409, 0.00000660, -0.00000570, -0.00001476, -0.00000865, 0.00000333, 0.00000604, -0.00000192, -0.00001014, -0.00001110, -0.00000599, 0.00000081, 0.00000625, 0.00000790, 0.00000427, -0.00000256, -0.00000724, -0.00000618, -0.00000123, 0.00000207, },
		{ -0.00001524, -0.00001095, -0.00000272, 0.00000257, 0.00000181, -0.00000093, 0.00000015, 0.00000458, 0.00000601, 0.00000077, -0.00000695, -0.00001008, -0.00000621, 0.00000139, 0.00000820, 0.00001124, 0.00000940, 0.00000435, 0.00000019, -0.00000067, -0.00000080, -0.00000362, -0.00000698, -0.00000523, 0.00000149, 0.00000549, 0.00000180, -0.00000430, -0.00000463, 0.00000029, 0.00000290, 0.00000023, -0.00000252, -0.00000070, 0.00000362, 0.00000627, 0.00000635, 0.00000401, -0.00000187, -0.00000938, -0.00001179, -0.00000574, 0.00000202, 0.00000308, -0.00000038, -0.00000023, 0.00000263, -0.00000067, -0.00000874, -0.00000691, 0.00000929, 0.00001982, 0.00000505, -0.00002139, -0.00002463, 0.00000424, 0.00003211, 0.00002535, -0.00000698, -0.00002688, -0.00001564, 0.00000924, 0.00002237, 0.00001690, 0.00000049, -0.00001718, -0.00002433, -0.00001052, 0.00001637, 0.00002844, 0.00001003, -0.00001643, -0.00001655, 0.00000754, 0.00001918, 0.00000181, -0.00001511, -0.00000161, 0.00002410, 0.00002110, -0.00001074, -0.00002740, -0.00000572, 0.00002140, 0.00001376, -0.00001696, -0.00002626, -0.00000280, 0.00001822, 0.00000975, -0.00001043, -0.00001233, 0.00000132, 0.00000599, -0.00000284, -0.00000625, 0.00000260, 0.00000643, -0.00000496, -0.00001426, -0.00000279, 0.00001792, 0.00002054, 0.00000169, -0.00001351, -0.00000788, 0.00000539, 0.00000503, -0.00000771, -0.00001412, -0.00000623, 0.00000419, 0.00000438, -0.00000366, -0.00000976, -0.00000922, -0.00000452, 0.00000120, 0.00000604, 0.00000720, 0.00000276, -0.00000420, -0.00000738, -0.00000426, 0.00000078, 0.00000135, -0.00000438, -0.00001207, },
		{ 0.00000170, 0.00000389, 0.00000067, -0.00000221, -0.00000023, 0.00000410, 0.00000429, -0.00000164, -0.00000817, -0.00000869, -0.00000253, 0.00000557, 0.00001064, 0.00001067, 0.00000657, 0.00000155, -0.00000082, -0.00000027, -0.00000055, -0.00000393, -0.00000669, -0.00000377, 0.00000272, 0.00000482, -0.00000025, -0.00000549, -0.00000387, 0.00000173, 0.00000324, -0.00000057, -0.00000309, -0.00000005, 0.00000515, 0.00000743, 0.00000592, 0.00000183, -0.00000432, -0.00001013, -0.00001059, -0.00000449, 0.00000192, 0.00000275, 0.00000052, 0.00000048, 0.00000064, -0.00000367, -0.00000731, 0.00000016, 0.00001447, 0.00001517, -0.00000582, -0.00002527, -0.00001491, 0.00001782, 0.00003419, 0.00001337, -0.00001987, -0.00002728, -0.00000404, 0.00002100, 0.00002389, 0.00000763, -0.00001155, -0.00002188, -0.00001715, 0.00000236, 0.00002183, 0.00001979, -0.00000391, -0.00002033, -0.00000745, 0.00001614, 0.00001538, -0.00000844, -0.00001676, 0.00000755, 0.00002991, 0.00001339, -0.00002314, -0.00002883, 0.00000464, 0.00002918, 0.00000981, -0.00002510, -0.00002662, 0.00000443, 0.00002289, 0.00000684, -0.00001543, -0.00001302, 0.00000397, 0.00000752, -0.00000317, -0.00000621, 0.00000313, 0.00000566, -0.00000637, -0.00001332, 0.00000076, 0.00002015, 0.00001865, -0.00000192, -0.00001469, -0.00000642, 0.00000627, 0.00000392, -0.00000873, -0.00001320, -0.00000471, 0.00000411, 0.00000300, -0.00000421, -0.00000874, -0.00000795, -0.00000407, 0.00000124, 0.00000614, 0.00000671, 0.00000120, -0.00000558, -0.00000675, -0.00000198, 0.00000184, -0.00000115, -0.00000918, -0.00001514, -0.00001375, -0.00000608, },
		{ -0.00000166, -0.00000331, 0.00000009, 0.00000396, 0.00000250, -0.00000392, -0.00000861, -0.00000615, 0.00000186, 0.00000925, 0.00001150, 0.00000850, 0.00000328, -0.00000043, -0.00000068, 0.00000063, -0.00000045, -0.00000439, -0.00000630, -0.00000240, 0.00000335, 0.00000352, -0.00000220, -0.00000587, -0.00000245, 0.00000298, 0.00000289, -0.00000167, -0.00000315, 0.00000136, 0.00000675, 0.00000776, 0.00000457, -0.00000042, -0.00000605, -0.00001034, -0.00000962, -0.00000362, 0.00000219, 0.00000312, 0.00000107, -0.00000024, -0.00000172, -0.00000446, -0.00000312, 0.00000659, 0.00001513, 0.00000696, -0.00001441, -0.00002272, -0.00000174, 0.00002747, 0.00002883, -0.00000154, -0.00002801, -0.00002021, 0.00001021, 0.00002807, 0.00001790, -0.00000506, -0.00001985, -0.00001915, -0.00000599, 0.00001162, 0.00001959, 0.00000745, -0.00001357, -0.00001706, 0.00000333, 0.00001955, 0.00000680, -0.00001666, -0.00001257, 0.00001816, 0.00003059, 0.00000085, -0.00003348, -0.00002385, 0.00001768, 0.00003303, 0.00000091, -0.00003308, -0.00002279, 0.00001397, 0.00002538, 0.00000063, -0.00002076, -0.00001133, 0.00000842, 0.00000855, -0.00000499, -0.00000674, 0.00000420, 0.00000556, -0.00000763, -0.00001258, 0.00000406, 0.00002220, 0.00001662, -0.00000571, -0.00001614, -0.00000508, 0.00000755, 0.00000358, -0.00000937, -0.00001287, -0.00000413, 0.00000393, 0.00000262, -0.00000385, -0.00000800, -0.00000777, -0.00000423, 0.00000165, 0.00000676, 0.00000613, -0.00000061, -0.00000653, -0.00000531, 0.00000020, 0.00000134, -0.00000534, -0.00001388, -0.00001597, -0.00000967, -0.00000043, 0.00000467, 0.00000303, },
		{ 0.00000123, 0.00000369, 0.00000028, -0.00000594, -0.00000788, -0.00000236, 0.00000643, 0.00001167, 0.00001045, 0.00000519, 0.00000034, -0.00000113, 0.00000042, 0.00000157, -0.00000076, -0.00000499, -0.00000582, -0.00000127, 0.00000335, 0.00000189, -0.00000368, -0.00000542, -0.00000078, 0.00000369, 0.00000196, -0.00000259, -0.00000245, 0.00000319, 0.00000789, 0.00000718, 0.00000274, -0.00000235, -0.00000719, -0.00001037, -0.00000885, -0.00000266, 0.00000297, 0.00000357, 0.00000068, -0.00000187, -0.00000310, -0.00000265, 0.00000220, 0.00001026, 0.00001134, -0.00000205, -0.00001816, -0.00001448, 0.00001122, 0.00003050, 0.00001739, -0.00001525, -0.00002878, -0.00000737, 0.00002272, 0.00002770, 0.00000603, -0.00001669, -0.00002139, -0.00001045, 0.00000464, 0.00001432, 0.00001133, -0.00000412, -0.00001664, -0.00000890, 0.00001189, 0.00001700, -0.00000347, -0.00002010, -0.00000373, 0.00002666, 0.00002493, -0.00001373, -0.00003855, -0.00001290, 0.00002998, 0.00003091, -0.00001128, -0.00003795, -0.00001428, 0.00002358, 0.00002383, -0.00000816, -0.00002435, -0.00000650, 0.00001346, 0.00000768, -0.00000823, -0.00000674, 0.00000642, 0.00000560, -0.00000948, -0.00001204, 0.00000775, 0.00002438, 0.00001401, -0.00001028, -0.00001774, -0.00000304, 0.00000980, 0.00000362, -0.00001060, -0.00001353, -0.00000384, 0.00000460, 0.00000331, -0.00000346, -0.00000834, -0.00000851, -0.00000415, 0.00000289, 0.00000760, 0.00000503, -0.00000259, -0.00000670, -0.00000315, 0.00000160, -0.00000117, -0.00001066, -0.00001718, -0.00001378, -0.00000360, 0.00000452, 0.00000514, 0.00000019, -0.00000411, -0.00000331, },
		{ -0.00000245, -0.00000717, -0.00000556, 0.00000239, 0.00001034, 0.00001216, 0.00000766, 0.00000158, -0.00000147, -0.00000046, 0.00000201, 0.00000209, -0.00000156, -0.00000561, -0.00000524, -0.00000044, 0.00000283, 0.00000025, -0.00000446, -0.00000435, 0.00000074, 0.00000374, 0.00000079, -0.00000297, -0.00000108, 0.00000497, 0.00000826, 0.00000588, 0.00000086, -0.00000384, -0.00000800, -0.00001040, -0.00000804, -0.00000135, 0.00000398, 0.00000347, -0.00000068, -0.00000338, -0.00000258, 0.00000110, 0.00000661, 0.00001005, 0.00000471, -0.00000884, -0.00001604, -0.00000310, 0.00002034, 0.00002620, 0.00000315, -0.00002386, -0.00002187, 0.00000754, 0.00002946, 0.00001966, -0.00000791, -0.00002312, -0.00001576, 0.00000056, 0.00001069, 0.00001015, 0.00000072, -0.00001125, -0.00001342, 0.00000047, 0.00001563, 0.00000999, -0.00001197, -0.00001790, 0.00000683, 0.00003019, 0.00001397, -0.00002667, -0.00003671, 0.00000156, 0.00003815, 0.00002269, -0.00002375, -0.00003760, -0.00000236, 0.00003058, 0.00001764, -0.00001755, -0.00002439, 0.00000096, 0.00001728, 0.00000419, -0.00001189, -0.00000508, 0.00000959, 0.00000479, -0.00001230, -0.00001105, 0.00001244, 0.00002638, 0.00000997, -0.00001584, -0.00001869, 0.00000060, 0.00001293, 0.00000304, -0.00001318, -0.00001479, -0.00000274, 0.00000661, 0.00000439, -0.00000409, -0.00000999, -0.00000939, -0.00000305, 0.00000490, 0.00000801, 0.00000314, -0.00000434, -0.00000577, -0.00000079, 0.00000136, -0.00000570, -0.00001589, -0.00001776, -0.00000853, 0.00000306, 0.00000719, 0.00000276, -0.00000362, -0.00000541, -0.00000178, 0.00000265, 0.00000260, },
		{ -0.00000155, 0.00000733, 0.00001263, 0.00001049, 0.00000381, -0.00000136, -0.00000171, 0.00000126, 0.00000347, 0.00000188, -0.00000275, -0.00000612, -0.00000461, 0.00000002, 0.00000199, -0.00000109, -0.00000454, -0.00000296, 0.00000183, 0.00000326, -0.00000023, -0.00000266, 0.00000065, 0.00000629, 0.00000783, 0.00000421, -0.00000079, -0.00000498, -0.00000869, -0.00001036, -0.00000690, 0.00000030, 0.00000471, 0.00000250, -0.00000243, -0.00000373, -0.00000004, 0.00000527, 0.00000839, 0.00000627, -0.00000218, -0.00001117, -0.00000895, 0.00000786, 0.00002315, 0.00001618, -0.00000971, -0.00002504, -0.00000945, 0.00001995, 0.00002824, 0.00000636, -0.00001932, -0.00002214, -0.00000527, 0.00000945, 0.00001009, 0.00000138, -0.00000811, -0.00001236, -0.00000635, 0.00000757, 0.00001414, 0.00000148, -0.00001627, -0.00001138, 0.00001572, 0.00002772, 0.00000056, -0.00003477, -0.00002851, 0.00001592, 0.00004009, 0.00001026, -0.00003327, -0.00003157, 0.00001035, 0.00003287, 0.00000781, -0.00002503, -0.00002015, 0.00000937, 0.00001823, -0.00000153, -0.00001442, -0.00000125, 0.00001274, 0.00000223, -0.00001567, -0.00000861, 0.00001823, 0.00002722, 0.00000378, -0.00002184, -0.00001779, 0.00000623, 0.00001604, 0.00000066, -0.00001718, -0.00001555, 0.00000010, 0.00000965, 0.00000467, -0.00000643, -0.00001235, -0.00000929, -0.00000059, 0.00000700, 0.00000737, 0.00000071, -0.00000519, -0.00000379, 0.00000088, -0.00000116, -0.00001154, -0.00001944, -0.00001477, -0.00000115, 0.00000839, 0.00000649, -0.00000182, -0.00000682, -0.00000460, 0.00000090, 0.00000338, 0.00000028, -0.00000519, -0.00000687, },
		{ 0.00001263, 0.00000702, -0.00000006, -0.00000285, -0.00000045, 0.00000337, 0.00000426, 0.00000091, -0.00000410, -0.00000639, -0.00000398, 0.00000015, 0.00000104, -0.00000197, -0.00000406, -0.00000160, 0.00000239, 0.00000252, -0.00000081, -0.00000175, 0.00000233, 0.00000691, 0.00000677, 0.00000250, -0.00000213, -0.00000594, -0.00000932, -0.00001003, -0.00000532, 0.00000195, 0.00000472, 0.00000081, -0.00000368, -0.00000233, 0.00000371, 0.00000812, 0.00000684, 0.00000056, -0.00000668, -0.00000839, 0.00000053, 0.00001504, 0.00001929, 0.00000380, -0.00001751, -0.00001890, 0.00000449, 0.00002614, 0.00001952, -0.00000794, -0.00002447, -0.00001427, 0.00000601, 0.00001276, 0.00000339, -0.00000820, -0.00001217, -0.00000826, 0.00000111, 0.00001039, 0.00000905, -0.00000548, -0.00001584, -0.00000326, 0.00002044, 0.00002036, -0.00001181, -0.00003656, -0.00001642, 0.00002682, 0.00003571, -0.00000325, -0.00003760, -0.00002124, 0.00002091, 0.00002983, -0.00000336, -0.00002856, -0.00001234, 0.00001654, 0.00001558, -0.00000806, -0.00001447, 0.00000430, 0.00001449, -0.00000235, -0.00001849, -0.00000394, 0.00002442, 0.00002566, -0.00000459, -0.00002694, -0.00001393, 0.00001337, 0.00001762, -0.00000420, -0.00002168, -0.00001443, 0.00000491, 0.00001253, 0.00000295, -0.00001033, -0.00001420, -0.00000741, 0.00000277, 0.00000823, 0.00000545, -0.00000154, -0.00000465, -0.00000138, 0.00000077, -0.00000598, -0.00001724, -0.00001977, -0.00000833, 0.00000651, 0.00001058, 0.00000237, -0.00000698, -0.00000784, -0.00000157, 0.00000363, 0.00000248, -0.00000310, -0.00000699, -0.00000448, 0.00000362, 0.00001135, },
		{ -0.00000288, -0.00000255, 0.00000181, 0.00000515, 0.00000409, -0.00000063, -0.00000533, -0.00000639, -0.00000344, 0.00000001, 0.00000018, -0.00000234, -0.00000324, -0.00000048, 0.00000252, 0.00000183, -0.00000085, -0.00000052, 0.00000362, 0.00000681, 0.00000537, 0.00000097, -0.00000323, -0.00000684, -0.00000978, -0.00000923, -0.00000343, 0.00000317, 0.00000388, -0.00000097, -0.00000364, 0.00000066, 0.00000726, 0.00000836, 0.00000254, -0.00000469, -0.00000702, -0.00000162, 0.00000911, 0.00001643, 0.00001051, -0.00000696, -0.00001834, -0.00000790, 0.00001552, 0.00002444, 0.00000625, -0.00001870, -0.00002189, -0.00000250, 0.00001379, 0.00000927, -0.00000657, -0.00001484, -0.00001065, -0.00000160, 0.00000605, 0.00000899, 0.00000307, -0.00000900, -0.00001205, 0.00000350, 0.00002028, 0.00001077, -0.00002032, -0.00003264, -0.00000384, 0.00003229, 0.00002685, -0.00001466, -0.00003631, -0.00000930, 0.00002714, 0.00002252, -0.00001318, -0.00002735, -0.00000287, 0.00002059, 0.00000986, -0.00001356, -0.00001150, 0.00001022, 0.00001372, -0.00000829, -0.00001939, 0.00000300, 0.00002958, 0.00002079, -0.00001422, -0.00002947, -0.00000681, 0.00002055, 0.00001625, -0.00001127, -0.00002510, -0.00001047, 0.00001089, 0.00001370, -0.00000124, -0.00001470, -0.00001423, -0.00000370, 0.00000597, 0.00000782, 0.00000273, -0.00000271, -0.00000277, 0.00000031, -0.00000190, -0.00001221, -0.00002090, -0.00001603, 0.00000025, 0.00001214, 0.00000861, -0.00000398, -0.00001069, -0.00000581, 0.00000277, 0.00000497, -0.00000043, -0.00000661, -0.00000678, -0.00000001, 0.00000888, 0.00001332, 0.00001024, 0.00000269, },
		{ 0.00000428, 0.00000604, 0.00000300, -0.00000240, -0.00000624, -0.00000617, -0.00000306, -0.00000028, -0.00000042, -0.00000227, -0.00000231, 0.00000032, 0.00000240, 0.00000140, -0.00000044, 0.00000071, 0.00000435, 0.00000616, 0.00000390, -0.00000032, -0.00000420, -0.00000766, -0.00000990, -0.00000796, -0.00000158, 0.00000365, 0.00000247, -0.00000207, -0.00000199, 0.00000434, 0.00000915, 0.00000569, -0.00000281, -0.00000724, -0.00000296, 0.00000660, 0.00001384, 0.00001201, 0.00000013, -0.00001282, -0.00001267, 0.00000395, 0.00002033, 0.00001586, -0.00000713, -0.00002268, -0.00001286, 0.00000883, 0.00001521, 0.00000033, -0.00001593, -0.00001626, -0.00000495, 0.00000443, 0.00000711, 0.00000512, -0.00000129, -0.00000900, -0.00000738, 0.00000698, 0.00001647, 0.00000200, -0.00002386, -0.00002534, 0.00000626, 0.00003236, 0.00001645, -0.00002191, -0.00003071, 0.00000141, 0.00002841, 0.00001324, -0.00001954, -0.00002215, 0.00000590, 0.00002074, 0.00000265, -0.00001646, -0.00000608, 0.00001487, 0.00001011, -0.00001418, -0.00001736, 0.00001126, 0.00003208, 0.00001257, -0.00002335, -0.00002809, 0.00000276, 0.00002584, 0.00001124, -0.00001917, -0.00002575, -0.00000371, 0.00001639, 0.00001188, -0.00000734, -0.00001792, -0.00001169, 0.00000101, 0.00000776, 0.00000572, 0.00000027, -0.00000217, -0.00000037, -0.00000001, -0.00000706, -0.00001811, -0.00002084, -0.00000859, 0.00000867, 0.00001381, 0.00000282, -0.00001040, -0.00001120, -0.00000102, 0.00000673, 0.00000380, -0.00000484, -0.00000886, -0.00000394, 0.00000576, 0.00001282, 0.00001261, 0.00000609, -0.00000129, -0.00000391, -0.00000070, },
		{ 0.00000128, -0.00000402, -0.00000676, -0.00000585, -0.00000284, -0.00000058, -0.00000071, -0.00000188, -0.00000143, 0.00000083, 0.00000224, 0.00000131, 0.00000022, 0.00000166, 0.00000450, 0.00000518, 0.00000254, -0.00000141, -0.00000510, -0.00000831, -0.00000957, -0.00000639, -0.00000011, 0.00000339, 0.00000110, -0.00000192, 0.00000091, 0.00000739, 0.00000847, 0.00000094, -0.00000705, -0.00000582, 0.00000413, 0.00001326, 0.00001324, 0.00000372, -0.00000821, -0.00001239, -0.00000315, 0.00001259, 0.00001784, 0.00000368, -0.00001635, -0.00001909, -0.00000086, 0.00001569, 0.00000980, -0.00001078, -0.00002132, -0.00001236, 0.00000211, 0.00000747, 0.00000486, 0.00000127, -0.00000280, -0.00000699, -0.00000419, 0.00000702, 0.00001141, -0.00000380, -0.00002323, -0.00001759, 0.00001235, 0.00002875, 0.00000726, -0.00002468, -0.00002326, 0.00000891, 0.00002574, 0.00000455, -0.00002169, -0.00001484, 0.00001208, 0.00001752, -0.00000409, -0.00001617, 0.00000036, 0.00001698, 0.00000442, -0.00001841, -0.00001233, 0.00001916, 0.00003081, 0.00000213, -0.00002995, -0.00002247, 0.00001298, 0.00002752, 0.00000310, -0.00002587, -0.00002262, 0.00000462, 0.00001950, 0.00000679, -0.00001380, -0.00001849, -0.00000683, 0.00000522, 0.00000735, 0.00000270, -0.00000080, -0.00000007, 0.00000113, -0.00000316, -0.00001359, -0.00002153, -0.00001632, 0.00000083, 0.00001436, 0.00001068, -0.00000507, -0.00001445, -0.00000782, 0.00000505, 0.00000851, -0.00000010, -0.00000933, -0.00000853, 0.00000136, 0.00001138, 0.00001422, 0.00000933, 0.00000137, -0.00000378, -0.00000300, 0.00000197, 0.00000616, 0.00000581, },
		{ -0.00000695, -0.00000556, -0.00000276, -0.00000078, -0.00000069, -0.00000131, -0.00000069, 0.00000116, 0.00000219, 0.00000149, 0.00000089, 0.00000221, 0.00000418, 0.00000408, 0.00000135, -0.00000237, -0.00000591, -0.00000866, -0.00000881, -0.00000482, 0.00000075, 0.00000271, 0.00000040, -0.00000041, 0.00000410, 0.00000863, 0.00000526, -0.00000416, -0.00000831, -0.00000059, 0.00001176, 0.00001591, 0.00000767, -0.00000527, -0.00001179, -0.00000652, 0.00000642, 0.00001526, 0.00000950, -0.00000776, -0.00001876, -0.00000979, 0.00000976, 0.00001579, -0.00000046, -0.00002020, -0.00002107, -0.00000507, 0.00000757, 0.00000682, 0.00000125, -0.00000061, -0.00000189, -0.00000520, -0.00000366, 0.00000503, 0.00000758, -0.00000612, -0.00002060, -0.00001171, 0.00001473, 0.00002397, 0.00000087, -0.00002424, -0.00001639, 0.00001277, 0.00002122, -0.00000168, -0.00002035, -0.00000765, 0.00001491, 0.00001245, -0.00000880, -0.00001323, 0.00000610, 0.00001618, -0.00000188, -0.00001984, -0.00000527, 0.00002489, 0.00002567, -0.00000858, -0.00003251, -0.00001355, 0.00002157, 0.00002489, -0.00000647, -0.00002945, -0.00001596, 0.00001248, 0.00001889, -0.00000064, -0.00001863, -0.00001578, -0.00000092, 0.00000747, 0.00000487, 0.00000013, 0.00000019, 0.00000250, 0.00000030, -0.00000893, -0.00001943, -0.00002074, -0.00000805, 0.00000952, 0.00001533, 0.00000354, -0.00001237, -0.00001436, -0.00000132, 0.00001020, 0.00000689, -0.00000581, -0.00001212, -0.00000498, 0.00000799, 0.00001517, 0.00001252, 0.00000441, -0.00000250, -0.00000415, -0.00000063, 0.00000454, 0.00000694, 0.00000462, -0.00000062, -0.00000527, },
		{ -0.00000272, -0.00000079, -0.00000040, -0.00000068, -0.00000008, 0.00000143, 0.00000228, 0.00000179, 0.00000138, 0.00000233, 0.00000357, 0.00000301, 0.00000032, -0.00000323, -0.00000656, -0.00000866, -0.00000780, -0.00000351, 0.00000109, 0.00000214, 0.00000076, 0.00000198, 0.00000643, 0.00000746, 0.00000054, -0.00000767, -0.00000578, 0.00000685, 0.00001721, 0.00001356, -0.00000076, -0.00001171, -0.00000974, 0.00000207, 0.00001244, 0.00001148, -0.00000129, -0.00001466, -0.00001421, 0.00000151, 0.00001534, 0.00000933, -0.00001201, -0.00002493, -0.00001572, 0.00000266, 0.00000958, 0.00000352, -0.00000145, 0.00000009, -0.00000026, -0.00000539, -0.00000537, 0.00000327, 0.00000634, -0.00000614, -0.00001840, -0.00000853, 0.00001510, 0.00002016, -0.00000277, -0.00002270, -0.00001155, 0.00001403, 0.00001701, -0.00000494, -0.00001725, -0.00000221, 0.00001493, 0.00000735, -0.00001094, -0.00000900, 0.00000993, 0.00001319, -0.00000718, -0.00001837, 0.00000212, 0.00002727, 0.00001780, -0.00001743, -0.00003066, -0.00000331, 0.00002672, 0.00001858, -0.00001519, -0.00002889, -0.00000727, 0.00001773, 0.00001443, -0.00000846, -0.00002023, -0.00001034, 0.00000419, 0.00000701, 0.00000144, -0.00000064, 0.00000290, 0.00000388, -0.00000368, -0.00001587, -0.00002224, -0.00001521, 0.00000181, 0.00001471, 0.00001108, -0.00000532, -0.00001640, -0.00000974, 0.00000624, 0.00001230, 0.00000192, -0.00001144, -0.00001167, 0.00000137, 0.00001411, 0.00001585, 0.00000812, -0.00000064, -0.00000438, -0.00000250, 0.00000226, 0.00000622, 0.00000653, 0.00000289, -0.00000232, -0.00000609, -0.00000696, -0.00000535, },
		{ 0.00000008, -0.00000005, 0.00000043, 0.00000169, 0.00000246, 0.00000206, 0.00000159, 0.00000212, 0.00000283, 0.00000204, -0.00000057, -0.00000397, -0.00000697, -0.00000834, -0.00000677, -0.00000258, 0.00000120, 0.00000215, 0.00000212, 0.00000433, 0.00000696, 0.00000420, -0.00000402, -0.00000811, 0.00000001, 0.00001410, 0.00001852, 0.00000699, -0.00000905, -0.00001348, -0.00000327, 0.00000986, 0.00001284, 0.00000308, -0.00001047, -0.00001503, -0.00000498, 0.00001063, 0.00001421, -0.00000128, -0.00002112, -0.00002388, -0.00000756, 0.00000811, 0.00000796, -0.00000033, -0.00000174, 0.00000245, 0.00000007, -0.00000807, -0.00000759, 0.00000361, 0.00000749, -0.00000596, -0.00001816, -0.00000725, 0.00001559, 0.00001817, -0.00000504, -0.00002190, -0.00000873, 0.00001453, 0.00001444, -0.00000607, -0.00001424, 0.00000099, 0.00001351, 0.00000349, -0.00001104, -0.00000497, 0.00001157, 0.00000936, -0.00001047, -0.00001492, 0.00000813, 0.00002622, 0.00000913, -0.00002291, -0.00002537, 0.00000597, 0.00002773, 0.00001040, -0.00002110, -0.00002456, 0.00000118, 0.00001907, 0.00000739, -0.00001444, -0.00001809, -0.00000382, 0.00000691, 0.00000420, -0.00000129, 0.00000097, 0.00000597, 0.00000250, -0.00001035, -0.00002162, -0.00002034, -0.00000606, 0.00001021, 0.00001455, 0.00000293, -0.00001279, -0.00001548, -0.00000190, 0.00001220, 0.00001022, -0.00000501, -0.00001481, -0.00000743, 0.00000898, 0.00001783, 0.00001302, 0.00000231, -0.00000425, -0.00000383, 0.00000040, 0.00000464, 0.00000660, 0.00000526, 0.00000113, -0.00000359, -0.00000659, -0.00000696, -0.00000523, -0.00000262, -0.00000058, },
		{ 0.00000088, 0.00000195, 0.00000263, 0.00000219, 0.00000153, 0.00000169, 0.00000207, 0.00000119, -0.00000134, -0.00000454, -0.00000713, -0.00000786, -0.00000588, -0.00000192, 0.00000151, 0.00000296, 0.00000394, 0.00000568, 0.00000542, -0.00000003, -0.00000671, -0.00000506, 0.00000736, 0.00001877, 0.00001518, -0.00000164, -0.00001441, -0.00001021, 0.00000488, 0.00001382, 0.00000769, -0.00000656, -0.00001485, -0.00000931, 0.00000518, 0.00001443, 0.00000708, -0.00001240, -0.00002523, -0.00001805, 0.00000044, 0.00000984, 0.00000413, -0.00000263, 0.00000046, 0.00000460, -0.00000223, -0.00001224, -0.00000820, 0.00000672, 0.00000939, -0.00000758, -0.00001992, -0.00000604, 0.00001762, 0.00001736, -0.00000778, -0.00002258, -0.00000666, 0.00001593, 0.00001353, -0.00000664, -0.00001247, 0.00000260, 0.00001211, 0.00000115, -0.00001032, -0.00000209, 0.00001172, 0.00000604, -0.00001175, -0.00001104, 0.00001176, 0.00002286, 0.00000159, -0.00002479, -0.00001856, 0.00001258, 0.00002533, 0.00000260, -0.00002328, -0.00001814, 0.00000726, 0.00001658, -0.00000008, -0.00001702, -0.00001307, 0.00000172, 0.00000660, 0.00000049, -0.00000185, 0.00000443, 0.00000748, -0.00000235, -0.00001791, -0.00002371, -0.00001351, 0.00000406, 0.00001428, 0.00000898, -0.00000627, -0.00001609, -0.00000968, 0.00000653, 0.00001428, 0.00000436, -0.00001154, -0.00001429, -0.00000014, 0.00001566, 0.00001784, 0.00000741, -0.00000310, -0.00000533, -0.00000123, 0.00000340, 0.00000576, 0.00000581, 0.00000367, -0.00000029, -0.00000448, -0.00000697, -0.00000701, -0.00000512, -0.00000238, -0.00000015, 0.00000067, 0.00000056, },
		},
		{ { -0.00000107, -0.00000018, 0.00000019, -0.00000195, -0.00000372, -0.00000258, -0.00000010, 0.00000246, 0.00000567, 0.00000716, 0.00000310, -0.00000392, -0.00000675, -0.00000401, -0.00000092, -0.00000030, -0.00000028, 0.00000012, 0.00000006, -0.00000022, -0.00000039, -0.00000096, -0.00000068, 0.00000158, 0.00000213, -0.00000203, -0.00000493, 0.00000002, 0.00000636, 0.00000350, -0.00000454, -0.00000579, -0.00000108, 0.00000024, -0.00000036, 0.00000396, 0.00000780, 0.00000269, -0.00000408, -0.00000216, 0.00000086, -0.00000543, -0.00001046, -0.00000031, 0.00001304, 0.00000795, -0.00001093, -0.00001843, -0.00000380, 0.00001845, 0.00002540, 0.00000591, -0.00002318, -0.00002458, 0.00000998, 0.00003115, -0.00000208, -0.00004488, -0.00002449, 0.00003432, 0.00003585, -0.00002656, -0.00004870, 0.00001130, 0.00005621, 0.00001130, -0.00004870, -0.00002656, 0.00003585, 0.00003432, -0.00002449, -0.00004488, -0.00000208, 0.00003115, 0.00000998, -0.00002458, -0.00002318, 0.00000591, 0.00002540, 0.00001845, -0.00000380, -0.00001843, -0.00001093, 0.00000795, 0.00001304, -0.00000031, -0.00001046, -0.00000543, 0.00000086, -0.00000216, -0.00000408, 0.00000269, 0.00000780, 0.00000396, -0.00000036, 0.00000024, -0.00000108, -0.00000579, -0.00000454, 0.00000350, 0.00000636, 0.00000002, -0.00000493, -0.00000203, 0.00000213, 0.00000158, -0.00000068, -0.00000096, -0.00000039, -0.00000022, 0.00000006, 0.00000012, -0.00000028, -0.00000030, -0.00000092, -0.00000401, -0.00000675, -0.00000392, 0.00000310, 0.00000716, 0.00000567, 0.00000246, -0.00000010, -0.00000258, -0.00000372, -0.00000195, 0.00000019, -0.00000018, },
		{ 0.00000010, -0.00000177, -0.00000365, -0.00000249, 0.00000046, 0.00000332, 0.00000608, 0.00000676, 0.00000223, -0.00000473, -0.00000710, -0.00000376, -0.00000046, -0.00000016, -0.00000036, 0.00000032, 0.00000041, -0.00000029, -0.00000071, -0.00000087, -0.00000049, 0.00000098, 0.00000121, -0.00000183, -0.00000359, 0.00000071, 0.00000549, 0.00000231, -0.00000462, -0.00000500, -0.00000046, 0.00000041, -0.00000034, 0.00000392, 0.00000753, 0.00000219, -0.00000465, -0.00000264, 0.00000053, -0.00000544, -0.00000963, 0.00000134, 0.00001394, 0.00000638, -0.00001346, -0.00001779, 0.00000055, 0.00002073, 0.00002113, -0.00000029, -0.00002298, -0.00001832, 0.00001316, 0.00002651, -0.00000847, -0.00004566, -0.00002103, 0.00003585, 0.00003400, -0.00002800, -0.00004768, 0.00001254, 0.00005546, 0.00000987, -0.00004814, -0.00002431, 0.00003635, 0.00003138, -0.00002716, -0.00004219, 0.00000454, 0.00003404, 0.00000559, -0.00002989, -0.00002169, 0.00001232, 0.00002834, 0.00001508, -0.00000780, -0.00001809, -0.00000813, 0.00000892, 0.00001159, -0.00000186, -0.00001087, -0.00000526, 0.00000111, -0.00000172, -0.00000346, 0.00000319, 0.00000798, 0.00000395, -0.00000042, -0.00000002, -0.00000174, -0.00000640, -0.00000423, 0.00000467, 0.00000694, -0.00000091, -0.00000617, -0.00000192, 0.00000323, 0.00000210, -0.00000101, -0.00000111, -0.00000006, -0.00000016, -0.00000032, -0.00000005, -0.00000016, -0.00000050, -0.00000148, -0.00000421, -0.00000619, -0.00000297, 0.00000387, 0.00000735, 0.00000516, 0.00000161, -0.00000065, -0.00000269, -0.00000376, -0.00000202, 0.00000034, 0.00000010, -0.00000112, -0.00000050, },
		{ -0.00000356, -0.00000241, 0.00000105, 0.00000419, 0.00000636, 0.00000616, 0.00000132, -0.00000536, -0.00000725, -0.00000351, -0.00000012, -0.00000006, -0.00000039, 0.00000055, 0.00000072, -0.00000039, -0.00000103, -0.00000085, -0.00000043, 0.00000039, 0.00000050, -0.00000136, -0.00000224, 0.00000113, 0.00000436, 0.00000118, -0.00000445, -0.00000408, 0.00000009, 0.00000049, -0.00000033, 0.00000385, 0.00000719, 0.00000167, -0.00000518, -0.00000317, 0.00000013, -0.00000521, -0.00000839, 0.00000298, 0.00001419, 0.00000432, -0.00001551, -0.00001619, 0.00000493, 0.00002176, 0.00001578, -0.00000583, -0.00002109, -0.00001160, 0.00001485, 0.00002044, -0.00001417, -0.00004456, -0.00001720, 0.00003579, 0.00003111, -0.00002836, -0.00004524, 0.00001328, 0.00005325, 0.00000860, -0.00004593, -0.00002157, 0.00003533, 0.00002732, -0.00002865, -0.00003770, 0.00001090, 0.00003502, 0.00000039, -0.00003384, -0.00001862, 0.00001851, 0.00002974, 0.00001089, -0.00001119, -0.00001681, -0.00000527, 0.00000922, 0.00000972, -0.00000318, -0.00001088, -0.00000498, 0.00000125, -0.00000131, -0.00000278, 0.00000367, 0.00000809, 0.00000390, -0.00000054, -0.00000039, -0.00000240, -0.00000680, -0.00000370, 0.00000575, 0.00000717, -0.00000203, -0.00000722, -0.00000149, 0.00000445, 0.00000249, -0.00000149, -0.00000127, 0.00000029, -0.00000013, -0.00000073, -0.00000019, -0.00000002, -0.00000078, -0.00000213, -0.00000431, -0.00000541, -0.00000192, 0.00000450, 0.00000734, 0.00000458, 0.00000079, -0.00000119, -0.00000280, -0.00000373, -0.00000198, 0.00000054, 0.00000029, -0.00000128, -0.00000082, 0.00000010, -0.00000151, },
		{ 0.00000168, 0.00000503, 0.00000649, 0.00000538, 0.00000040, -0.00000579, -0.00000722, -0.00000327, 0.00000012, 0.00000001, -0.00000035, 0.00000079, 0.00000098, -0.00000051, -0.00000137, -0.00000092, -0.00000048, -0.00000015, 0.00000005, -0.00000068, -0.00000097, 0.00000125, 0.00000306, 0.00000016, -0.00000404, -0.00000310, 0.00000053, 0.00000048, -0.00000031, 0.00000375, 0.00000677, 0.00000114, -0.00000568, -0.00000374, -0.00000026, -0.00000470, -0.00000679, 0.00000447, 0.00001376, 0.00000192, -0.00001692, -0.00001374, 0.00000903, 0.00002143, 0.00000972, -0.00001037, -0.00001766, -0.00000490, 0.00001487, 0.00001331, -0.00001878, -0.00004173, -0.00001342, 0.00003407, 0.00002749, -0.00002748, -0.00004166, 0.00001327, 0.00004977, 0.00000780, -0.00004214, -0.00001871, 0.00003269, 0.00002254, -0.00002864, -0.00003167, 0.00001654, 0.00003406, -0.00000517, -0.00003612, -0.00001421, 0.00002401, 0.00002956, 0.00000620, -0.00001374, -0.00001471, -0.00000257, 0.00000883, 0.00000756, -0.00000421, -0.00001054, -0.00000467, 0.00000127, -0.00000091, -0.00000202, 0.00000415, 0.00000812, 0.00000377, -0.00000073, -0.00000083, -0.00000299, -0.00000695, -0.00000300, 0.00000665, 0.00000705, -0.00000324, -0.00000800, -0.00000076, 0.00000569, 0.00000270, -0.00000209, -0.00000142, 0.00000068, -0.00000013, -0.00000116, -0.00000029, 0.00000014, -0.00000113, -0.00000283, -0.00000430, -0.00000441, -0.00000082, 0.00000496, 0.00000714, 0.00000396, 0.00000001, -0.00000174, -0.00000290, -0.00000359, -0.00000182, 0.00000072, 0.00000037, -0.00000150, -0.00000110, 0.00000020, -0.00000120, -0.00000347, -0.00000233, },
		{ 0.00000644, 0.00000444, -0.00000047, -0.00000601, -0.00000705, -0.00000309, 0.00000024, 0.00000009, -0.00000025, 0.00000103, 0.00000119, -0.00000066, -0.00000173, -0.00000109, -0.00000059, -0.00000058, -0.00000015, 0.00000013, 0.00000015, 0.00000110, 0.00000167, -0.00000068, -0.00000344, -0.00000211, 0.00000085, 0.00000043, -0.00000027, 0.00000362, 0.00000625, 0.00000058, -0.00000616, -0.00000435, -0.00000061, -0.00000387, -0.00000489, 0.00000569, 0.00001263, -0.00000064, -0.00001755, -0.00001059, 0.00001255, 0.00001978, 0.00000334, -0.00001360, -0.00001295, 0.00000130, 0.00001320, 0.00000560, -0.00002201, -0.00003743, -0.00001010, 0.00003074, 0.00002353, -0.00002528, -0.00003725, 0.00001233, 0.00004526, 0.00000775, -0.00003692, -0.00001610, 0.00002846, 0.00001746, -0.00002695, -0.00002445, 0.00002101, 0.00003125, -0.00001058, -0.00003653, -0.00000883, 0.00002842, 0.00002783, 0.00000134, -0.00001529, -0.00001197, -0.00000024, 0.00000780, 0.00000527, -0.00000489, -0.00000990, -0.00000437, 0.00000117, -0.00000051, -0.00000120, 0.00000462, 0.00000806, 0.00000356, -0.00000101, -0.00000132, -0.00000346, -0.00000684, -0.00000221, 0.00000730, 0.00000659, -0.00000448, -0.00000845, 0.00000026, 0.00000690, 0.00000269, -0.00000280, -0.00000152, 0.00000110, -0.00000017, -0.00000161, -0.00000035, 0.00000029, -0.00000158, -0.00000355, -0.00000413, -0.00000321, 0.00000030, 0.00000522, 0.00000677, 0.00000332, -0.00000074, -0.00000230, -0.00000298, -0.00000334, -0.00000156, 0.00000085, 0.00000031, -0.00000177, -0.00000129, 0.00000040, -0.00000086, -0.00000341, -0.00000222, 0.00000234, 0.00000583, },
		{ -0.00000125, -0.00000601, -0.00000678, -0.00000299, 0.00000027, 0.00000019, -0.00000008, 0.00000126, 0.00000135, -0.00000085, -0.00000214, -0.00000135, -0.00000073, -0.00000082, -0.00000009, 0.00000100, 0.00000106, 0.00000071, 0.00000027, -0.00000131, -0.00000267, -0.00000116, 0.00000105, 0.00000035, -0.00000019, 0.00000345, 0.00000565, -0.00000000, -0.00000663, -0.00000498, -0.00000084, -0.00000270, -0.00000279, 0.00000651, 0.00001085, -0.00000318, -0.00001733, -0.00000698, 0.00001525, 0.00001691, -0.00000294, -0.00001538, -0.00000731, 0.00000658, 0.00000994, -0.00000220, -0.00002369, -0.00003203, -0.00000756, 0.00002598, 0.00001957, -0.00002182, -0.00003238, 0.00001037, 0.00004008, 0.00000866, -0.00003055, -0.00001407, 0.00002281, 0.00001255, -0.00002347, -0.00001648, 0.00002397, 0.00002684, -0.00001535, -0.00003500, -0.00000290, 0.00003141, 0.00002470, -0.00000331, -0.00001576, -0.00000881, 0.00000159, 0.00000620, 0.00000300, -0.00000519, -0.00000906, -0.00000414, 0.00000099, -0.00000008, -0.00000032, 0.00000507, 0.00000792, 0.00000324, -0.00000139, -0.00000182, -0.00000377, -0.00000649, -0.00000138, 0.00000766, 0.00000583, -0.00000564, -0.00000853, 0.00000152, 0.00000800, 0.00000246, -0.00000358, -0.00000155, 0.00000156, -0.00000026, -0.00000208, -0.00000038, 0.00000042, -0.00000210, -0.00000427, -0.00000378, -0.00000183, 0.00000137, 0.00000527, 0.00000625, 0.00000268, -0.00000146, -0.00000286, -0.00000301, -0.00000295, -0.00000123, 0.00000089, 0.00000011, -0.00000204, -0.00000136, 0.00000067, -0.00000055, -0.00000336, -0.00000208, 0.00000304, 0.00000654, 0.00000620, 0.00000338, },
		{ -0.00000645, -0.00000299, 0.00000021, 0.00000033, 0.00000015, 0.00000147, 0.00000143, -0.00000107, -0.00000261, -0.00000168, -0.00000084, -0.00000086, 0.00000018, 0.00000184, 0.00000172, 0.00000012, -0.00000106, -0.00000174, -0.00000181, -0.00000030, 0.00000115, 0.00000027, -0.00000008, 0.00000323, 0.00000495, -0.00000063, -0.00000710, -0.00000557, -0.00000090, -0.00000120, -0.00000060, 0.00000685, 0.00000851, -0.00000549, -0.00001627, -0.00000315, 0.00001694, 0.00001303, -0.00000873, -0.00001564, -0.00000116, 0.00001059, 0.00000530, -0.00000964, -0.00002375, -0.00002593, -0.00000605, 0.00002006, 0.00001592, -0.00001724, -0.00002742, 0.00000742, 0.00003459, 0.00001065, -0.00002338, -0.00001289, 0.00001601, 0.00000823, -0.00001828, -0.00000826, 0.00002520, 0.00002120, -0.00001905, -0.00003163, 0.00000310, 0.00003275, 0.00002045, -0.00000742, -0.00001517, -0.00000546, 0.00000279, 0.00000416, 0.00000089, -0.00000514, -0.00000809, -0.00000399, 0.00000073, 0.00000040, 0.00000062, 0.00000549, 0.00000768, 0.00000280, -0.00000185, -0.00000228, -0.00000387, -0.00000593, -0.00000061, 0.00000769, 0.00000483, -0.00000664, -0.00000823, 0.00000294, 0.00000891, 0.00000199, -0.00000438, -0.00000147, 0.00000204, -0.00000040, -0.00000257, -0.00000038, 0.00000053, -0.00000270, -0.00000495, -0.00000324, -0.00000031, 0.00000235, 0.00000510, 0.00000561, 0.00000206, -0.00000215, -0.00000343, -0.00000296, -0.00000240, -0.00000085, 0.00000079, -0.00000021, -0.00000226, -0.00000129, 0.00000100, -0.00000029, -0.00000334, -0.00000187, 0.00000376, 0.00000713, 0.00000577, 0.00000226, -0.00000190, -0.00000582, },
		{ 0.00000011, 0.00000053, 0.00000042, 0.00000163, 0.00000143, -0.00000134, -0.00000312, -0.00000206, -0.00000087, -0.00000067, 0.00000062, 0.00000258, 0.00000210, -0.00000058, -0.00000225, -0.00000197, -0.00000091, 0.00000046, 0.00000116, 0.00000020, 0.00000007, 0.00000297, 0.00000416, -0.00000130, -0.00000756, -0.00000610, -0.00000073, 0.00000059, 0.00000155, 0.00000666, 0.00000574, -0.00000744, -0.00001443, 0.00000063, 0.00001751, 0.00000842, -0.00001366, -0.00001446, 0.00000505, 0.00001311, -0.00000037, -0.00001626, -0.00002227, -0.00001957, -0.00000571, 0.00001331, 0.00001284, -0.00001181, -0.00002270, 0.00000362, 0.00002920, 0.00001374, -0.00001582, -0.00001272, 0.00000843, 0.00000491, -0.00001156, -0.00000029, 0.00002459, 0.00001478, -0.00002131, -0.00002663, 0.00000869, 0.00003233, 0.00001539, -0.00001072, -0.00001360, -0.00000220, 0.00000331, 0.00000183, -0.00000094, -0.00000478, -0.00000709, -0.00000394, 0.00000045, 0.00000092, 0.00000159, 0.00000588, 0.00000733, 0.00000224, -0.00000238, -0.00000265, -0.00000375, -0.00000521, 0.00000004, 0.00000738, 0.00000367, -0.00000740, -0.00000755, 0.00000444, 0.00000957, 0.00000129, -0.00000516, -0.00000128, 0.00000253, -0.00000062, -0.00000307, -0.00000036, 0.00000059, -0.00000336, -0.00000555, -0.00000250, 0.00000131, 0.00000319, 0.00000470, 0.00000488, 0.00000149, -0.00000282, -0.00000400, -0.00000281, -0.00000172, -0.00000048, 0.00000052, -0.00000063, -0.00000238, -0.00000108, 0.00000133, -0.00000011, -0.00000331, -0.00000158, 0.00000448, 0.00000756, 0.00000516, 0.00000110, -0.00000239, -0.00000545, -0.00000609, -0.00000308, },
		{ 0.00000073, 0.00000172, 0.00000136, -0.00000166, -0.00000368, -0.00000246, -0.00000078, -0.00000025, 0.00000116, 0.00000313, 0.00000223, -0.00000133, -0.00000325, -0.00000203, -0.00000003, 0.00000109, 0.00000111, 0.00000018, 0.00000025, 0.00000265, 0.00000328, -0.00000203, -0.00000800, -0.00000650, -0.00000030, 0.00000261, 0.00000354, 0.00000591, 0.00000268, -0.00000887, -0.00001192, 0.00000410, 0.00001695, 0.00000341, -0.00001747, -0.00001201, 0.00001086, 0.00001403, -0.00000670, -0.00002173, -0.00001945, -0.00001339, -0.00000655, 0.00000612, 0.00001046, -0.00000584, -0.00001850, -0.00000083, 0.00002427, 0.00001783, -0.00000828, -0.00001364, 0.00000050, 0.00000286, -0.00000361, 0.00000696, 0.00002221, 0.00000809, -0.00002188, -0.00002036, 0.00001343, 0.00003016, 0.00000991, -0.00001299, -0.00001122, 0.00000076, 0.00000316, -0.00000060, -0.00000243, -0.00000418, -0.00000612, -0.00000398, 0.00000018, 0.00000150, 0.00000258, 0.00000622, 0.00000687, 0.00000153, -0.00000294, -0.00000289, -0.00000339, -0.00000438, 0.00000049, 0.00000675, 0.00000243, -0.00000786, -0.00000652, 0.00000596, 0.00000995, 0.00000040, -0.00000587, -0.00000094, 0.00000302, -0.00000091, -0.00000358, -0.00000032, 0.00000062, -0.00000405, -0.00000604, -0.00000155, 0.00000299, 0.00000387, 0.00000410, 0.00000409, 0.00000096, -0.00000348, -0.00000454, -0.00000254, -0.00000090, -0.00000017, 0.00000007, -0.00000112, -0.00000234, -0.00000071, 0.00000162, -0.00000003, -0.00000325, -0.00000119, 0.00000517, 0.00000781, 0.00000439, -0.00000002, -0.00000270, -0.00000494, -0.00000575, -0.00000327, -0.00000002, 0.00000078, },
		{ 0.00000119, -0.00000203, -0.00000427, -0.00000283, -0.00000052, 0.00000039, 0.00000172, 0.00000345, 0.00000212, -0.00000203, -0.00000403, -0.00000197, 0.00000077, 0.00000159, 0.00000104, 0.00000020, 0.00000043, 0.00000226, 0.00000230, -0.00000281, -0.00000841, -0.00000672, 0.00000042, 0.00000478, 0.00000524, 0.00000463, -0.00000049, -0.00000972, -0.00000892, 0.00000703, 0.00001531, -0.00000164, -0.00001995, -0.00000858, 0.00001587, 0.00001337, -0.00001322, -0.00002579, -0.00001556, -0.00000776, -0.00000850, -0.00000112, 0.00000886, 0.00000030, -0.00001501, -0.00000560, 0.00002011, 0.00002272, -0.00000118, -0.00001560, -0.00000730, 0.00000229, 0.00000515, 0.00001307, 0.00001823, 0.00000163, -0.00002068, -0.00001325, 0.00001691, 0.00002637, 0.00000441, -0.00001409, -0.00000826, 0.00000319, 0.00000240, -0.00000298, -0.00000352, -0.00000343, -0.00000524, -0.00000407, -0.00000005, 0.00000212, 0.00000358, 0.00000651, 0.00000628, 0.00000069, -0.00000351, -0.00000294, -0.00000280, -0.00000354, 0.00000069, 0.00000584, 0.00000120, -0.00000798, -0.00000519, 0.00000739, 0.00001002, -0.00000066, -0.00000647, -0.00000047, 0.00000348, -0.00000130, -0.00000411, -0.00000025, 0.00000060, -0.00000477, -0.00000640, -0.00000042, 0.00000469, 0.00000434, 0.00000329, 0.00000327, 0.00000048, -0.00000411, -0.00000504, -0.00000214, 0.00000003, 0.00000005, -0.00000058, -0.00000161, -0.00000212, -0.00000023, 0.00000183, -0.00000004, -0.00000312, -0.00000069, 0.00000580, 0.00000785, 0.00000349, -0.00000105, -0.00000283, -0.00000433, -0.00000545, -0.00000354, -0.00000016, 0.00000110, 0.00000106, 0.00000175, },
		{ -0.00000486, -0.00000312, -0.00000008, 0.00000121, 0.00000223, 0.00000350, 0.00000184, -0.00000263, -0.00000457, -0.00000183, 0.00000145, 0.00000199, 0.00000097, 0.00000028, 0.00000061, 0.00000181, 0.00000124, -0.00000365, -0.00000876, -0.00000671, 0.00000142, 0.00000698, 0.00000656, 0.00000288, -0.00000359, -0.00000994, -0.00000563, 0.00000925, 0.00001275, -0.00000639, -0.00002102, -0.00000450, 0.00001975, 0.00001127, -0.00001952, -0.00002833, -0.00001095, -0.00000298, -0.00001135, -0.00000806, 0.00000797, 0.00000622, -0.00001234, -0.00001036, 0.00001697, 0.00002815, 0.00000513, -0.00001844, -0.00001449, 0.00000324, 0.00001427, 0.00001775, 0.00001299, -0.00000411, -0.00001773, -0.00000580, 0.00001886, 0.00002123, -0.00000073, -0.00001401, -0.00000498, 0.00000497, 0.00000113, -0.00000514, -0.00000422, -0.00000262, -0.00000450, -0.00000419, -0.00000022, 0.00000277, 0.00000456, 0.00000674, 0.00000556, -0.00000027, -0.00000404, -0.00000276, -0.00000203, -0.00000276, 0.00000061, 0.00000471, 0.00000009, -0.00000773, -0.00000364, 0.00000868, 0.00000977, -0.00000182, -0.00000691, 0.00000014, 0.00000390, -0.00000178, -0.00000463, -0.00000017, 0.00000054, -0.00000548, -0.00000660, 0.00000089, 0.00000635, 0.00000459, 0.00000231, 0.00000245, 0.00000007, -0.00000473, -0.00000549, -0.00000158, 0.00000102, 0.00000013, -0.00000141, -0.00000206, -0.00000168, 0.00000033, 0.00000190, -0.00000015, -0.00000290, -0.00000009, 0.00000631, 0.00000765, 0.00000252, -0.00000195, -0.00000277, -0.00000367, -0.00000521, -0.00000386, -0.00000027, 0.00000148, 0.00000137, 0.00000168, 0.00000093, -0.00000246, },
		{ 0.00000057, 0.00000215, 0.00000262, 0.00000328, 0.00000145, -0.00000305, -0.00000489, -0.00000169, 0.00000198, 0.00000230, 0.00000094, 0.00000039, 0.00000077, 0.00000128, 0.00000008, -0.00000454, -0.00000902, -0.00000642, 0.00000270, 0.00000911, 0.00000742, 0.00000076, -0.00000646, -0.00000955, -0.00000228, 0.00001063, 0.00000949, -0.00001052, -0.00002072, -0.00000016, 0.00002224, 0.00000797, -0.00002518, -0.00002931, -0.00000602, 0.00000073, -0.00001486, -0.00001438, 0.00000768, 0.00001156, -0.00001049, -0.00001473, 0.00001498, 0.00003378, 0.00001038, -0.00002191, -0.00002065, 0.00000566, 0.00002325, 0.00002081, 0.00000689, -0.00000871, -0.00001322, 0.00000145, 0.00001912, 0.00001507, -0.00000516, -0.00001281, -0.00000165, 0.00000598, -0.00000048, -0.00000695, -0.00000456, -0.00000186, -0.00000391, -0.00000428, -0.00000028, 0.00000344, 0.00000549, 0.00000688, 0.00000471, -0.00000133, -0.00000447, -0.00000235, -0.00000112, -0.00000211, 0.00000024, 0.00000344, -0.00000082, -0.00000714, -0.00000194, 0.00000975, 0.00000920, -0.00000303, -0.00000717, 0.00000087, 0.00000424, -0.00000236, -0.00000516, -0.00000006, 0.00000044, -0.00000616, -0.00000663, 0.00000235, 0.00000793, 0.00000458, 0.00000118, 0.00000166, -0.00000028, -0.00000532, -0.00000585, -0.00000087, 0.00000205, 0.00000003, -0.00000240, -0.00000241, -0.00000101, 0.00000092, 0.00000180, -0.00000033, -0.00000253, 0.00000060, 0.00000666, 0.00000723, 0.00000152, -0.00000267, -0.00000257, -0.00000300, -0.00000504, -0.00000421, -0.00000033, 0.00000189, 0.00000165, 0.00000152, 0.00000057, -0.00000293, -0.00000543, -0.00000329, },
		{ 0.00000283, 0.00000281, 0.00000101, -0.00000325, -0.00000501, -0.00000160, 0.00000235, 0.00000256, 0.00000097, 0.00000053, 0.00000088, 0.00000068, -0.00000115, -0.00000546, -0.00000913, -0.00000580, 0.00000419, 0.00001104, 0.00000778, -0.00000161, -0.00000896, -0.00000864, 0.00000091, 0.00001113, 0.00000580, -0.00001378, -0.00001918, 0.00000405, 0.00002322, 0.00000378, -0.00002985, -0.00002885, -0.00000115, 0.00000326, -0.00001871, -0.00001985, 0.00000776, 0.00001601, -0.00000938, -0.00001837, 0.00001419, 0.00003925, 0.00001438, -0.00002565, -0.00002542, 0.00000934, 0.00003161, 0.00002221, 0.00000041, -0.00001188, -0.00000748, 0.00000800, 0.00001764, 0.00000832, -0.00000861, -0.00001062, 0.00000145, 0.00000621, -0.00000226, -0.00000832, -0.00000461, -0.00000122, -0.00000347, -0.00000431, -0.00000025, 0.00000409, 0.00000635, 0.00000693, 0.00000371, -0.00000245, -0.00000475, -0.00000168, -0.00000013, -0.00000168, -0.00000042, 0.00000212, -0.00000148, -0.00000622, -0.00000019, 0.00001054, 0.00000835, -0.00000425, -0.00000722, 0.00000170, 0.00000448, -0.00000305, -0.00000567, 0.00000007, 0.00000032, -0.00000679, -0.00000648, 0.00000393, 0.00000938, 0.00000433, -0.00000008, 0.00000092, -0.00000057, -0.00000589, -0.00000612, -0.00000003, 0.00000306, -0.00000028, -0.00000351, -0.00000259, -0.00000011, 0.00000148, 0.00000150, -0.00000053, -0.00000199, 0.00000135, 0.00000680, 0.00000658, 0.00000056, -0.00000319, -0.00000225, -0.00000237, -0.00000493, -0.00000455, -0.00000035, 0.00000233, 0.00000189, 0.00000126, 0.00000011, -0.00000343, -0.00000593, -0.00000329, 0.00000140, 0.00000313, },
		{ 0.00000062, -0.00000321, -0.00000499, -0.00000160, 0.00000256, 0.00000280, 0.00000106, 0.00000068, 0.00000093, 0.00000002, -0.00000245, -0.00000641, -0.00000907, -0.00000485, 0.00000584, 0.00001267, 0.00000763, -0.00000407, -0.00001098, -0.00000733, 0.00000372, 0.00001078, 0.00000196, -0.00001598, -0.00001660, 0.00000777, 0.00002266, -0.00000092, -0.00003328, -0.00002715, 0.00000331, 0.00000462, -0.00002258, -0.00002434, 0.00000795, 0.00001931, -0.00000885, -0.00002096, 0.00001454, 0.00004421, 0.00001708, -0.00002930, -0.00002855, 0.00001400, 0.00003893, 0.00002206, -0.00000593, -0.00001345, -0.00000094, 0.00001341, 0.00001454, 0.00000141, -0.00001089, -0.00000769, 0.00000408, 0.00000570, -0.00000404, -0.00000917, -0.00000446, -0.00000077, -0.00000314, -0.00000423, -0.00000012, 0.00000471, 0.00000713, 0.00000689, 0.00000260, -0.00000360, -0.00000485, -0.00000077, 0.00000086, -0.00000152, -0.00000131, 0.00000086, -0.00000184, -0.00000503, 0.00000152, 0.00001104, 0.00000726, -0.00000540, -0.00000704, 0.00000261, 0.00000460, -0.00000385, -0.00000617, 0.00000022, 0.00000019, -0.00000735, -0.00000614, 0.00000560, 0.00001066, 0.00000381, -0.00000142, 0.00000027, -0.00000078, -0.00000641, -0.00000628, 0.00000094, 0.00000400, -0.00000081, -0.00000469, -0.00000256, 0.00000099, 0.00000196, 0.00000099, -0.00000073, -0.00000124, 0.00000212, 0.00000669, 0.00000574, -0.00000030, -0.00000349, -0.00000187, -0.00000183, -0.00000487, -0.00000484, -0.00000031, 0.00000276, 0.00000205, 0.00000089, -0.00000045, -0.00000395, -0.00000631, -0.00000310, 0.00000239, 0.00000409, 0.00000280, 0.00000212, },
		{ -0.00000488, -0.00000173, 0.00000264, 0.00000305, 0.00000122, 0.00000081, 0.00000092, -0.00000070, -0.00000381, -0.00000734, -0.00000877, -0.00000354, 0.00000755, 0.00001388, 0.00000700, -0.00000649, -0.00001247, -0.00000575, 0.00000599, 0.00000968, -0.00000171, -0.00001703, -0.00001329, 0.00001070, 0.00002067, -0.00000575, -0.00003532, -0.00002447, 0.00000705, 0.00000489, -0.00002615, -0.00002780, 0.00000798, 0.00002130, -0.00000868, -0.00002226, 0.00001588, 0.00004834, 0.00001852, -0.00003246, -0.00002989, 0.00001926, 0.00004487, 0.00002057, -0.00001167, -0.00001339, 0.00000591, 0.00001731, 0.00001004, -0.00000521, -0.00001192, -0.00000428, 0.00000605, 0.00000455, -0.00000563, -0.00000952, -0.00000420, -0.00000056, -0.00000290, -0.00000400, 0.00000011, 0.00000526, 0.00000780, 0.00000673, 0.00000138, -0.00000472, -0.00000472, 0.00000035, 0.00000176, -0.00000165, -0.00000238, -0.00000026, -0.00000186, -0.00000364, 0.00000312, 0.00001122, 0.00000597, -0.00000645, -0.00000664, 0.00000357, 0.00000458, -0.00000474, -0.00000663, 0.00000041, 0.00000004, -0.00000783, -0.00000561, 0.00000731, 0.00001173, 0.00000304, -0.00000282, -0.00000029, -0.00000092, -0.00000690, -0.00000632, 0.00000201, 0.00000484, -0.00000156, -0.00000588, -0.00000228, 0.00000226, 0.00000228, 0.00000026, -0.00000086, -0.00000029, 0.00000285, 0.00000629, 0.00000474, -0.00000100, -0.00000356, -0.00000147, -0.00000139, -0.00000483, -0.00000506, -0.00000024, 0.00000314, 0.00000214, 0.00000041, -0.00000110, -0.00000445, -0.00000654, -0.00000268, 0.00000350, 0.00000493, 0.00000252, 0.00000128, 0.00000033, -0.00000293, },
		{ 0.00000261, 0.00000334, 0.00000143, 0.00000090, 0.00000082, -0.00000148, -0.00000522, -0.00000823, -0.00000822, -0.00000190, 0.00000924, 0.00001459, 0.00000596, -0.00000869, -0.00001341, -0.00000411, 0.00000760, 0.00000801, -0.00000494, -0.00001695, -0.00000957, 0.00001259, 0.00001744, -0.00001035, -0.00003592, -0.00002115, 0.00000987, 0.00000425, -0.00002917, -0.00003028, 0.00000757, 0.00002194, -0.00000861, -0.00002212, 0.00001800, 0.00005141, 0.00001886, -0.00003477, -0.00002944, 0.00002474, 0.00004923, 0.00001809, -0.00001638, -0.00001183, 0.00001253, 0.00001949, 0.00000448, -0.00001114, -0.00001172, -0.00000070, 0.00000722, 0.00000290, -0.00000689, -0.00000938, -0.00000395, -0.00000060, -0.00000268, -0.00000360, 0.00000039, 0.00000572, 0.00000835, 0.00000647, 0.00000007, -0.00000576, -0.00000434, 0.00000161, 0.00000249, -0.00000210, -0.00000355, -0.00000115, -0.00000156, -0.00000214, 0.00000454, 0.00001110, 0.00000455, -0.00000734, -0.00000603, 0.00000453, 0.00000440, -0.00000572, -0.00000704, 0.00000064, -0.00000010, -0.00000820, -0.00000491, 0.00000904, 0.00001256, 0.00000203, -0.00000423, -0.00000072, -0.00000099, -0.00000734, -0.00000625, 0.00000314, 0.00000553, -0.00000252, -0.00000701, -0.00000171, 0.00000364, 0.00000239, -0.00000065, -0.00000086, 0.00000088, 0.00000350, 0.00000558, 0.00000365, -0.00000147, -0.00000344, -0.00000112, -0.00000107, -0.00000476, -0.00000517, -0.00000014, 0.00000346, 0.00000213, -0.00000016, -0.00000183, -0.00000492, -0.00000655, -0.00000202, 0.00000465, 0.00000559, 0.00000200, 0.00000036, 0.00000020, -0.00000243, -0.00000474, -0.00000201, },
		{ 0.00000168, 0.00000093, 0.00000064, -0.00000230, -0.00000665, -0.00000903, -0.00000736, 0.00000003, 0.00001079, 0.00001476, 0.00000460, -0.00001057, -0.00001382, -0.00000256, 0.00000849, 0.00000596, -0.00000751, -0.00001583, -0.00000579, 0.00001332, 0.00001324, -0.00001440, -0.00003516, -0.00001749, 0.00001163, 0.00000292, -0.00003141, -0.00003190, 0.00000647, 0.00002124, -0.00000840, -0.00002049, 0.00002064, 0.00005325, 0.00001833, -0.00003592, -0.00002731, 0.00003002, 0.00005190, 0.00001502, -0.00001975, -0.00000904, 0.00001839, 0.00001984, -0.00000172, -0.00001602, -0.00001041, 0.00000273, 0.00000755, 0.00000095, -0.00000770, -0.00000884, -0.00000380, -0.00000088, -0.00000243, -0.00000302, 0.00000072, 0.00000606, 0.00000877, 0.00000609, -0.00000128, -0.00000667, -0.00000369, 0.00000297, 0.00000298, -0.00000286, -0.00000473, -0.00000174, -0.00000097, -0.00000061, 0.00000571, 0.00001069, 0.00000307, -0.00000804, -0.00000521, 0.00000547, 0.00000404, -0.00000677, -0.00000739, 0.00000091, -0.00000023, -0.00000847, -0.00000405, 0.00001073, 0.00001313, 0.00000080, -0.00000561, -0.00000102, -0.00000100, -0.00000772, -0.00000606, 0.00000431, 0.00000605, -0.00000366, -0.00000804, -0.00000086, 0.00000507, 0.00000224, -0.00000173, -0.00000067, 0.00000225, 0.00000399, 0.00000455, 0.00000254, -0.00000167, -0.00000314, -0.00000088, -0.00000087, -0.00000462, -0.00000516, -0.00000005, 0.00000368, 0.00000203, -0.00000084, -0.00000264, -0.00000531, -0.00000631, -0.00000115, 0.00000576, 0.00000599, 0.00000125, -0.00000056, 0.00000028, -0.00000176, -0.00000463, -0.00000244, 0.00000251, 0.00000369, },
		{ 0.00000038, -0.00000316, -0.00000808, -0.00000970, -0.00000618, 0.00000219, 0.00001209, 0.00001435, 0.00000302, -0.00001200, -0.00001378, -0.00000130, 0.00000865, 0.00000378, -0.00000925, -0.00001385, -0.00000227, 0.00001286, 0.00000841, -0.00001765, -0.00003318, -0.00001384, 0.00001231, 0.00000116, -0.00003276, -0.00003283, 0.00000451, 0.00001932, -0.00000784, -0.00001743, 0.00002353, 0.00005381, 0.00001724, -0.00003569, -0.00002372, 0.00003473, 0.00005296, 0.00001179, -0.00002158, -0.00000539, 0.00002303, 0.00001841, -0.00000811, -0.00001962, -0.00000820, 0.00000575, 0.00000704, -0.00000110, -0.00000799, -0.00000800, -0.00000383, -0.00000137, -0.00000209, -0.00000226, 0.00000105, 0.00000627, 0.00000905, 0.00000560, -0.00000264, -0.00000739, -0.00000279, 0.00000433, 0.00000318, -0.00000388, -0.00000582, -0.00000199, -0.00000014, 0.00000086, 0.00000660, 0.00001003, 0.00000157, -0.00000854, -0.00000422, 0.00000634, 0.00000350, -0.00000788, -0.00000766, 0.00000124, -0.00000035, -0.00000862, -0.00000305, 0.00001235, 0.00001340, -0.00000063, -0.00000693, -0.00000116, -0.00000093, -0.00000805, -0.00000577, 0.00000547, 0.00000636, -0.00000493, -0.00000888, 0.00000028, 0.00000648, 0.00000180, -0.00000291, -0.00000025, 0.00000378, 0.00000425, 0.00000322, 0.00000147, -0.00000156, -0.00000274, -0.00000079, -0.00000078, -0.00000437, -0.00000503, -0.00000000, 0.00000378, 0.00000185, -0.00000159, -0.00000350, -0.00000557, -0.00000578, -0.00000008, 0.00000676, 0.00000610, 0.00000034, -0.00000138, 0.00000057, -0.00000099, -0.00000459, -0.00000299, 0.00000238, 0.00000410, 0.00000195, 0.00000088, },
		{ -0.00000948, -0.00001019, -0.00000468, 0.00000450, 0.00001304, 0.00001339, 0.00000137, -0.00001292, -0.00001340, -0.00000045, 0.00000816, 0.00000171, -0.00001006, -0.00001126, 0.00000068, 0.00001127, 0.00000329, -0.00001994, -0.00003023, -0.00001045, 0.00001198, -0.00000077, -0.00003317, -0.00003331, 0.00000159, 0.00001637, -0.00000674, -0.00001308, 0.00002643, 0.00005315, 0.00001591, -0.00003396, -0.00001896, 0.00003861, 0.00005257, 0.00000884, -0.00002183, -0.00000135, 0.00002607, 0.00001541, -0.00001422, -0.00002179, -0.00000536, 0.00000810, 0.00000578, -0.00000305, -0.00000775, -0.00000698, -0.00000408, -0.00000199, -0.00000161, -0.00000134, 0.00000134, 0.00000632, 0.00000918, 0.00000502, -0.00000396, -0.00000788, -0.00000167, 0.00000561, 0.00000303, -0.00000510, -0.00000672, -0.00000187, 0.00000087, 0.00000218, 0.00000720, 0.00000917, 0.00000013, -0.00000880, -0.00000310, 0.00000710, 0.00000277, -0.00000901, -0.00000782, 0.00000161, -0.00000044, -0.00000865, -0.00000192, 0.00001386, 0.00001337, -0.00000223, -0.00000814, -0.00000113, -0.00000081, -0.00000834, -0.00000539, 0.00000659, 0.00000646, -0.00000630, -0.00000950, 0.00000167, 0.00000778, 0.00000104, -0.00000412, 0.00000044, 0.00000541, 0.00000423, 0.00000161, 0.00000053, -0.00000114, -0.00000228, -0.00000091, -0.00000075, -0.00000395, -0.00000478, -0.00000005, 0.00000374, 0.00000159, -0.00000242, -0.00000440, -0.00000568, -0.00000493, 0.00000113, 0.00000754, 0.00000588, -0.00000067, -0.00000202, 0.00000105, -0.00000018, -0.00000465, -0.00000364, 0.00000227, 0.00000457, 0.00000220, 0.00000074, 0.00000004, -0.00000405, },
		{ -0.00000286, 0.00000686, 0.00001357, 0.00001193, -0.00000024, -0.00001330, -0.00001281, -0.00000012, 0.00000715, -0.00000001, -0.00000995, -0.00000834, 0.00000287, 0.00000873, -0.00000177, -0.00002122, -0.00002658, -0.00000753, 0.00001078, -0.00000265, -0.00003270, -0.00003354, -0.00000231, 0.00001260, -0.00000503, -0.00000765, 0.00002913, 0.00005144, 0.00001466, -0.00003073, -0.00001339, 0.00004145, 0.00005104, 0.00000653, -0.00002062, 0.00000257, 0.00002726, 0.00001113, -0.00001960, -0.00002251, -0.00000222, 0.00000962, 0.00000393, -0.00000472, -0.00000700, -0.00000592, -0.00000460, -0.00000268, -0.00000096, -0.00000031, 0.00000155, 0.00000622, 0.00000917, 0.00000436, -0.00000519, -0.00000810, -0.00000037, 0.00000673, 0.00000253, -0.00000644, -0.00000736, -0.00000140, 0.00000198, 0.00000330, 0.00000750, 0.00000816, -0.00000121, -0.00000884, -0.00000187, 0.00000773, 0.00000187, -0.00001015, -0.00000787, 0.00000204, -0.00000052, -0.00000856, -0.00000069, 0.00001522, 0.00001304, -0.00000395, -0.00000920, -0.00000093, -0.00000063, -0.00000858, -0.00000494, 0.00000763, 0.00000634, -0.00000770, -0.00000984, 0.00000327, 0.00000891, -0.00000001, -0.00000531, 0.00000143, 0.00000710, 0.00000388, -0.00000023, -0.00000019, -0.00000040, -0.00000185, -0.00000127, -0.00000075, -0.00000333, -0.00000445, -0.00000023, 0.00000356, 0.00000128, -0.00000330, -0.00000532, -0.00000557, -0.00000374, 0.00000242, 0.00000801, 0.00000533, -0.00000167, -0.00000242, 0.00000170, 0.00000059, -0.00000482, -0.00000433, 0.00000218, 0.00000506, 0.00000241, 0.00000051, -0.00000037, -0.00000497, -0.00001081, -0.00001045, },
		{ 0.00001361, 0.00001003, -0.00000167, -0.00001315, -0.00001215, -0.00000036, 0.00000577, -0.00000119, -0.00000899, -0.00000539, 0.00000414, 0.00000545, -0.00000647, -0.00002151, -0.00002254, -0.00000522, 0.00000889, -0.00000429, -0.00003145, -0.00003373, -0.00000714, 0.00000825, -0.00000270, -0.00000141, 0.00003152, 0.00004892, 0.00001378, -0.00002612, -0.00000736, 0.00004319, 0.00004871, 0.00000514, -0.00001819, 0.00000587, 0.00002648, 0.00000597, -0.00002388, -0.00002186, 0.00000090, 0.00001020, 0.00000166, -0.00000596, -0.00000583, -0.00000495, -0.00000535, -0.00000335, -0.00000013, 0.00000078, 0.00000167, 0.00000596, 0.00000902, 0.00000365, -0.00000628, -0.00000805, 0.00000105, 0.00000761, 0.00000168, -0.00000782, -0.00000767, -0.00000060, 0.00000308, 0.00000415, 0.00000753, 0.00000707, -0.00000241, -0.00000866, -0.00000059, 0.00000820, 0.00000079, -0.00001125, -0.00000779, 0.00000252, -0.00000057, -0.00000836, 0.00000061, 0.00001639, 0.00001239, -0.00000576, -0.00001008, -0.00000056, -0.00000043, -0.00000879, -0.00000444, 0.00000858, 0.00000603, -0.00000908, -0.00000990, 0.00000503, 0.00000980, -0.00000135, -0.00000638, 0.00000273, 0.00000876, 0.00000316, -0.00000222, -0.00000064, 0.00000062, -0.00000152, -0.00000186, -0.00000071, -0.00000248, -0.00000407, -0.00000060, 0.00000322, 0.00000093, -0.00000424, -0.00000623, -0.00000522, -0.00000222, 0.00000370, 0.00000811, 0.00000449, -0.00000257, -0.00000254, 0.00000245, 0.00000125, -0.00000510, -0.00000501, 0.00000215, 0.00000555, 0.00000255, 0.00000019, -0.00000085, -0.00000589, -0.00001202, -0.00001043, -0.00000077, 0.00000916, },
		{ -0.00000284, -0.00001253, -0.00001155, -0.00000117, 0.00000424, -0.00000170, -0.00000733, -0.00000272, 0.00000446, 0.00000174, -0.00001058, -0.00002094, -0.00001839, -0.00000355, 0.00000655, -0.00000559, -0.00002961, -0.00003404, -0.00001274, 0.00000354, 0.00000018, 0.00000535, 0.00003354, 0.00004590, 0.00001348, -0.00002032, -0.00000121, 0.00004390, 0.00004601, 0.00000484, -0.00001493, 0.00000809, 0.00002378, 0.00000041, -0.00002678, -0.00002003, 0.00000368, 0.00000983, -0.00000079, -0.00000668, -0.00000434, -0.00000416, -0.00000630, -0.00000391, 0.00000090, 0.00000189, 0.00000165, 0.00000555, 0.00000873, 0.00000290, -0.00000717, -0.00000770, 0.00000251, 0.00000818, 0.00000051, -0.00000912, -0.00000759, 0.00000046, 0.00000411, 0.00000471, 0.00000731, 0.00000596, -0.00000343, -0.00000827, 0.00000071, 0.00000847, -0.00000043, -0.00001229, -0.00000756, 0.00000304, -0.00000060, -0.00000805, 0.00000194, 0.00001735, 0.00001144, -0.00000760, -0.00001074, -0.00000001, -0.00000020, -0.00000898, -0.00000391, 0.00000941, 0.00000555, -0.00001038, -0.00000965, 0.00000688, 0.00001039, -0.00000294, -0.00000727, 0.00000432, 0.00001030, 0.00000205, -0.00000430, -0.00000074, 0.00000188, -0.00000136, -0.00000269, -0.00000057, -0.00000138, -0.00000369, -0.00000118, 0.00000276, 0.00000057, -0.00000520, -0.00000710, -0.00000457, -0.00000039, 0.00000489, 0.00000777, 0.00000341, -0.00000326, -0.00000236, 0.00000320, 0.00000173, -0.00000544, -0.00000561, 0.00000217, 0.00000600, 0.00000262, -0.00000020, -0.00000140, -0.00000681, -0.00001306, -0.00001009, 0.00000155, 0.00001129, 0.00001314, 0.00000781, },
		{ -0.00001114, -0.00000248, 0.00000274, -0.00000147, -0.00000518, -0.00000055, 0.00000387, -0.00000211, -0.00001391, -0.00001969, -0.00001441, -0.00000250, 0.00000399, -0.00000649, -0.00002738, -0.00003456, -0.00001892, -0.00000135, 0.00000343, 0.00001234, 0.00003524, 0.00004271, 0.00001390, -0.00001363, 0.00000479, 0.00004375, 0.00004330, 0.00000563, -0.00001128, 0.00000887, 0.00001935, -0.00000508, -0.00002813, -0.00001730, 0.00000586, 0.00000859, -0.00000320, -0.00000683, -0.00000268, -0.00000364, -0.00000738, -0.00000429, 0.00000209, 0.00000294, 0.00000150, 0.00000500, 0.00000832, 0.00000216, -0.00000783, -0.00000708, 0.00000393, 0.00000838, -0.00000091, -0.00001025, -0.00000712, 0.00000172, 0.00000497, 0.00000496, 0.00000691, 0.00000488, -0.00000425, -0.00000770, 0.00000197, 0.00000853, -0.00000177, -0.00001323, -0.00000719, 0.00000360, -0.00000061, -0.00000763, 0.00000329, 0.00001805, 0.00001021, -0.00000943, -0.00001115, 0.00000068, 0.00000001, -0.00000917, -0.00000337, 0.00001011, 0.00000492, -0.00001155, -0.00000909, 0.00000874, 0.00001064, -0.00000470, -0.00000790, 0.00000618, 0.00001164, 0.00000055, -0.00000636, -0.00000045, 0.00000331, -0.00000146, -0.00000372, -0.00000025, -0.00000004, -0.00000339, -0.00000201, 0.00000220, 0.00000024, -0.00000619, -0.00000790, -0.00000360, 0.00000171, 0.00000588, 0.00000695, 0.00000218, -0.00000364, -0.00000188, 0.00000388, 0.00000200, -0.00000580, -0.00000609, 0.00000222, 0.00000637, 0.00000261, -0.00000066, -0.00000202, -0.00000771, -0.00001388, -0.00000942, 0.00000400, 0.00001312, 0.00001215, 0.00000539, -0.00000366, -0.00001152, },
		{ 0.00000147, -0.00000050, -0.00000280, 0.00000090, 0.00000250, -0.00000580, -0.00001639, -0.00001799, -0.00001078, -0.00000197, 0.00000142, -0.00000704, -0.00002499, -0.00003533, -0.00002544, -0.00000630, 0.00000685, 0.00001931, 0.00003674, 0.00003971, 0.00001508, -0.00000637, 0.00001046, 0.00004302, 0.00004096, 0.00000738, -0.00000777, 0.00000798, 0.00001352, -0.00001004, -0.00002788, -0.00001399, 0.00000723, 0.00000659, -0.00000537, -0.00000642, -0.00000101, -0.00000345, -0.00000850, -0.00000442, 0.00000340, 0.00000389, 0.00000119, 0.00000433, 0.00000780, 0.00000145, -0.00000822, -0.00000621, 0.00000523, 0.00000820, -0.00000250, -0.00001112, -0.00000626, 0.00000308, 0.00000561, 0.00000491, 0.00000636, 0.00000387, -0.00000486, -0.00000699, 0.00000315, 0.00000836, -0.00000321, -0.00001403, -0.00000666, 0.00000417, -0.00000061, -0.00000712, 0.00000462, 0.00001849, 0.00000873, -0.00001119, -0.00001128, 0.00000151, 0.00000020, -0.00000936, -0.00000283, 0.00001068, 0.00000418, -0.00001255, -0.00000827, 0.00001054, 0.00001053, -0.00000659, -0.00000823, 0.00000825, 0.00001271, -0.00000131, -0.00000831, 0.00000024, 0.00000483, -0.00000185, -0.00000490, 0.00000031, 0.00000153, -0.00000322, -0.00000311, 0.00000158, -0.00000005, -0.00000719, -0.00000859, -0.00000229, 0.00000403, 0.00000658, 0.00000565, 0.00000089, -0.00000361, -0.00000117, 0.00000437, 0.00000203, -0.00000610, -0.00000639, 0.00000227, 0.00000661, 0.00000252, -0.00000116, -0.00000272, -0.00000858, -0.00001441, -0.00000839, 0.00000650, 0.00001456, 0.00001068, 0.00000291, -0.00000410, -0.00001025, -0.00001100, -0.00000419, },
		{ -0.00000044, 0.00000153, 0.00000054, -0.00000907, -0.00001802, -0.00001609, -0.00000765, -0.00000180, -0.00000100, -0.00000732, -0.00002263, -0.00003629, -0.00003204, -0.00001123, 0.00001017, 0.00002608, 0.00003825, 0.00003719, 0.00001695, 0.00000112, 0.00001573, 0.00004205, 0.00003922, 0.00000981, -0.00000490, 0.00000530, 0.00000671, -0.00001407, -0.00002613, -0.00001044, 0.00000766, 0.00000405, -0.00000711, -0.00000551, 0.00000053, -0.00000359, -0.00000955, -0.00000426, 0.00000477, 0.00000468, 0.00000074, 0.00000357, 0.00000719, 0.00000081, -0.00000832, -0.00000515, 0.00000632, 0.00000763, -0.00000417, -0.00001165, -0.00000504, 0.00000445, 0.00000597, 0.00000460, 0.00000574, 0.00000298, -0.00000527, -0.00000616, 0.00000421, 0.00000797, -0.00000470, -0.00001467, -0.00000598, 0.00000475, -0.00000061, -0.00000653, 0.00000589, 0.00001863, 0.00000702, -0.00001282, -0.00001113, 0.00000245, 0.00000032, -0.00000957, -0.00000230, 0.00001111, 0.00000336, -0.00001334, -0.00000719, 0.00001220, 0.00001006, -0.00000850, -0.00000819, 0.00001046, 0.00001344, -0.00000349, -0.00001004, 0.00000135, 0.00000634, -0.00000259, -0.00000616, 0.00000116, 0.00000327, -0.00000327, -0.00000447, 0.00000093, -0.00000026, -0.00000819, -0.00000915, -0.00000063, 0.00000648, 0.00000689, 0.00000386, -0.00000033, -0.00000313, -0.00000028, 0.00000458, 0.00000184, -0.00000625, -0.00000651, 0.00000225, 0.00000669, 0.00000238, -0.00000169, -0.00000349, -0.00000937, -0.00001460, -0.00000702, 0.00000892, 0.00001551, 0.00000880, 0.00000049, -0.00000415, -0.00000885, -0.00001118, -0.00000614, 0.00000059, 0.00000112, },
		{ -0.00000176, -0.00001170, -0.00001890, -0.00001423, -0.00000507, -0.00000182, -0.00000314, -0.00000746, -0.00002046, -0.00003735, -0.00003849, -0.00001618, 0.00001310, 0.00003253, 0.00004001, 0.00003540, 0.00001939, 0.00000855, 0.00002063, 0.00004121, 0.00003822, 0.00001252, -0.00000313, 0.00000091, -0.00000058, -0.00001688, -0.00002306, -0.00000701, 0.00000712, 0.00000118, -0.00000831, -0.00000422, 0.00000179, -0.00000402, -0.00001043, -0.00000380, 0.00000613, 0.00000526, 0.00000017, 0.00000273, 0.00000651, 0.00000028, -0.00000812, -0.00000395, 0.00000715, 0.00000670, -0.00000582, -0.00001180, -0.00000354, 0.00000573, 0.00000603, 0.00000406, 0.00000508, 0.00000223, -0.00000547, -0.00000526, 0.00000511, 0.00000736, -0.00000619, -0.00001510, -0.00000516, 0.00000531, -0.00000060, -0.00000586, 0.00000708, 0.00001848, 0.00000514, -0.00001427, -0.00001069, 0.00000346, 0.00000036, -0.00000980, -0.00000178, 0.00001143, 0.00000252, -0.00001391, -0.00000593, 0.00001366, 0.00000924, -0.00001036, -0.00000779, 0.00001274, 0.00001375, -0.00000591, -0.00001148, 0.00000283, 0.00000775, -0.00000368, -0.00000741, 0.00000236, 0.00000512, -0.00000359, -0.00000608, 0.00000032, -0.00000039, -0.00000918, -0.00000954, 0.00000138, 0.00000899, 0.00000674, 0.00000166, -0.00000136, -0.00000215, 0.00000068, 0.00000443, 0.00000148, -0.00000618, -0.00000643, 0.00000212, 0.00000659, 0.00000222, -0.00000223, -0.00000435, -0.00001007, -0.00001438, -0.00000534, 0.00001114, 0.00001592, 0.00000660, -0.00000173, -0.00000387, -0.00000744, -0.00001167, -0.00000814, 0.00000022, 0.00000325, 0.00000165, 0.00000131, },
		{ -0.00001916, -0.00001261, -0.00000302, -0.00000186, -0.00000497, -0.00000765, -0.00001859, -0.00003836, -0.00004459, -0.00002121, 0.00001539, 0.00003861, 0.00004229, 0.00003451, 0.00002220, 0.00001567, 0.00002532, 0.00004087, 0.00003795, 0.00001504, -0.00000284, -0.00000497, -0.00000782, -0.00001826, -0.00001897, -0.00000399, 0.00000566, -0.00000176, -0.00000888, -0.00000270, 0.00000267, -0.00000469, -0.00001103, -0.00000304, 0.00000741, 0.00000560, -0.00000051, 0.00000184, 0.00000578, -0.00000012, -0.00000762, -0.00000268, 0.00000764, 0.00000544, -0.00000735, -0.00001153, -0.00000182, 0.00000684, 0.00000579, 0.00000335, 0.00000445, 0.00000164, -0.00000548, -0.00000433, 0.00000582, 0.00000654, -0.00000765, -0.00001531, -0.00000422, 0.00000583, -0.00000059, -0.00000514, 0.00000815, 0.00001802, 0.00000312, -0.00001548, -0.00000996, 0.00000450, 0.00000029, -0.00001006, -0.00000128, 0.00001164, 0.00000168, -0.00001423, -0.00000452, 0.00001487, 0.00000810, -0.00001209, -0.00000701, 0.00001498, 0.00001363, -0.00000850, -0.00001252, 0.00000465, 0.00000895, -0.00000511, -0.00000857, 0.00000391, 0.00000700, -0.00000425, -0.00000789, -0.00000019, -0.00000043, -0.00001015, -0.00000974, 0.00000371, 0.00001144, 0.00000605, -0.00000089, -0.00000206, -0.00000068, 0.00000160, 0.00000384, 0.00000100, -0.00000581, -0.00000618, 0.00000179, 0.00000630, 0.00000208, -0.00000279, -0.00000531, -0.00001062, -0.00001370, -0.00000339, 0.00001303, 0.00001576, 0.00000422, -0.00000367, -0.00000331, -0.00000615, -0.00001242, -0.00001001, 0.00000041, 0.00000568, 0.00000324, 0.00000028, -0.00000415, -0.00001356, },
		{ -0.00000144, -0.00000176, -0.00000650, -0.00000803, -0.00001706, -0.00003913, -0.00005022, -0.00002647, 0.00001681, 0.00004435, 0.00004533, 0.00003461, 0.00002516, 0.00002235, 0.00003002, 0.00004130, 0.00003824, 0.00001684, -0.00000427, -0.00001198, -0.00001450, -0.00001817, -0.00001422, -0.00000164, 0.00000340, -0.00000455, -0.00000883, -0.00000111, 0.00000311, -0.00000551, -0.00001128, -0.00000203, 0.00000852, 0.00000570, -0.00000127, 0.00000092, 0.00000502, -0.00000035, -0.00000684, -0.00000142, 0.00000778, 0.00000394, -0.00000865, -0.00001084, -0.00000000, 0.00000769, 0.00000525, 0.00000254, 0.00000389, 0.00000121, -0.00000534, -0.00000340, 0.00000630, 0.00000552, -0.00000903, -0.00001528, -0.00000318, 0.00000629, -0.00000058, -0.00000438, 0.00000907, 0.00001726, 0.00000104, -0.00001642, -0.00000896, 0.00000551, 0.00000009, -0.00001033, -0.00000078, 0.00001177, 0.00000087, -0.00001432, -0.00000304, 0.00001577, 0.00000670, -0.00001361, -0.00000587, 0.00001711, 0.00001305, -0.00001116, -0.00001312, 0.00000673, 0.00000985, -0.00000686, -0.00000954, 0.00000583, 0.00000882, -0.00000529, -0.00000987, -0.00000056, -0.00000036, -0.00001110, -0.00000971, 0.00000633, 0.00001375, 0.00000481, -0.00000367, -0.00000233, 0.00000126, 0.00000234, 0.00000279, 0.00000051, -0.00000506, -0.00000581, 0.00000121, 0.00000583, 0.00000199, -0.00000336, -0.00000638, -0.00001097, -0.00001252, -0.00000127, 0.00001446, 0.00001502, 0.00000180, -0.00000525, -0.00000260, -0.00000509, -0.00001333, -0.00001156, 0.00000113, 0.00000817, 0.00000418, -0.00000141, -0.00000637, -0.00001460, -0.00001900, -0.00001139, },
		{ -0.00000780, -0.00000875, -0.00001585, -0.00003950, -0.00005529, -0.00003210, 0.00001721, 0.00004985, 0.00004935, 0.00003568, 0.00002806, 0.00002855, 0.00003502, 0.00004270, 0.00003881, 0.00001740, -0.00000750, -0.00001966, -0.00002019, -0.00001668, -0.00000919, -0.00000010, 0.00000057, -0.00000698, -0.00000821, 0.00000039, 0.00000308, -0.00000636, -0.00001112, -0.00000082, 0.00000941, 0.00000554, -0.00000207, -0.00000000, 0.00000427, -0.00000041, -0.00000583, -0.00000023, 0.00000754, 0.00000227, -0.00000965, -0.00000977, 0.00000183, 0.00000824, 0.00000447, 0.00000169, 0.00000342, 0.00000094, -0.00000505, -0.00000251, 0.00000655, 0.00000435, -0.00001027, -0.00001499, -0.00000206, 0.00000669, -0.00000057, -0.00000359, 0.00000981, 0.00001622, -0.00000105, -0.00001705, -0.00000773, 0.00000647, -0.00000025, -0.00001060, -0.00000028, 0.00001182, 0.00000011, -0.00001419, -0.00000153, 0.00001635, 0.00000510, -0.00001485, -0.00000441, 0.00001901, 0.00001201, -0.00001377, -0.00001323, 0.00000899, 0.00001039, -0.00000887, -0.00001022, 0.00000808, 0.00001047, -0.00000674, -0.00001193, -0.00000073, -0.00000019, -0.00001202, -0.00000944, 0.00000920, 0.00001580, 0.00000299, -0.00000657, -0.00000207, 0.00000358, 0.00000278, 0.00000128, 0.00000011, -0.00000389, -0.00000538, 0.00000031, 0.00000522, 0.00000202, -0.00000396, -0.00000756, -0.00001106, -0.00001081, 0.00000094, 0.00001531, 0.00001374, -0.00000053, -0.00000643, -0.00000185, -0.00000432, -0.00001425, -0.00001263, 0.00000231, 0.00001046, 0.00000438, -0.00000356, -0.00000819, -0.00001485, -0.00001860, -0.00001064, -0.00000021, -0.00000141, },
		{ -0.00001487, -0.00003930, -0.00005981, -0.00003828, 0.00001650, 0.00005522, 0.00005447, 0.00003764, 0.00003074, 0.00003436, 0.00004062, 0.00004512, 0.00003920, 0.00001626, -0.00001244, -0.00002745, -0.00002453, -0.00001400, -0.00000428, 0.00000055, -0.00000261, -0.00000889, -0.00000713, 0.00000164, 0.00000260, -0.00000712, -0.00001052, 0.00000051, 0.00001001, 0.00000515, -0.00000287, -0.00000090, 0.00000354, -0.00000028, -0.00000462, 0.00000082, 0.00000694, 0.00000054, -0.00001028, -0.00000838, 0.00000356, 0.00000845, 0.00000348, 0.00000086, 0.00000307, 0.00000083, -0.00000466, -0.00000171, 0.00000655, 0.00000305, -0.00001132, -0.00001444, -0.00000090, 0.00000698, -0.00000056, -0.00000279, 0.00001035, 0.00001492, -0.00000309, -0.00001733, -0.00000629, 0.00000730, -0.00000074, -0.00001086, 0.00000023, 0.00001181, -0.00000056, -0.00001386, -0.00000006, 0.00001660, 0.00000337, -0.00001575, -0.00000270, 0.00002062, 0.00001056, -0.00001623, -0.00001284, 0.00001130, 0.00001049, -0.00001104, -0.00001053, 0.00001061, 0.00001185, -0.00000860, -0.00001400, -0.00000066, 0.00000006, -0.00001291, -0.00000891, 0.00001223, 0.00001749, 0.00000062, -0.00000944, -0.00000120, 0.00000619, 0.00000279, -0.00000066, -0.00000009, -0.00000229, -0.00000500, -0.00000095, 0.00000452, 0.00000219, -0.00000460, -0.00000886, -0.00001084, -0.00000856, 0.00000312, 0.00001547, 0.00001199, -0.00000260, -0.00000721, -0.00000121, -0.00000385, -0.00001501, -0.00001311, 0.00000379, 0.00001232, 0.00000383, -0.00000594, -0.00000947, -0.00001443, -0.00001817, -0.00001037, 0.00000083, -0.00000077, -0.00000901, -0.00000991, },
		{ -0.00006385, -0.00004514, 0.00001470, 0.00006062, 0.00006074, 0.00004036, 0.00003312, 0.00003999, 0.00004710, 0.00004846, 0.00003893, 0.00001307, -0.00001885, -0.00003478, -0.00002730, -0.00001044, 0.00000016, 0.00000034, -0.00000586, -0.00001019, -0.00000574, 0.00000252, 0.00000173, -0.00000769, -0.00000948, 0.00000188, 0.00001029, 0.00000456, -0.00000365, -0.00000176, 0.00000285, 0.00000003, -0.00000329, 0.00000167, 0.00000601, -0.00000116, -0.00001051, -0.00000674, 0.00000509, 0.00000832, 0.00000237, 0.00000011, 0.00000285, 0.00000084, -0.00000418, -0.00000103, 0.00000629, 0.00000167, -0.00001215, -0.00001365, 0.00000027, 0.00000718, -0.00000055, -0.00000200, 0.00001067, 0.00001340, -0.00000501, -0.00001726, -0.00000471, 0.00000797, -0.00000137, -0.00001108, 0.00000077, 0.00001175, -0.00000116, -0.00001336, 0.00000131, 0.00001651, 0.00000159, -0.00001629, -0.00000080, 0.00002186, 0.00000874, -0.00001843, -0.00001195, 0.00001357, 0.00001013, -0.00001330, -0.00001040, 0.00001336, 0.00001285, -0.00001085, -0.00001597, -0.00000031, 0.00000037, -0.00001375, -0.00000812, 0.00001537, 0.00001875, -0.00000225, -0.00001212, 0.00000030, 0.00000895, 0.00000227, -0.00000296, 0.00000006, -0.00000027, -0.00000477, -0.00000260, 0.00000379, 0.00000254, -0.00000534, -0.00001025, -0.00001022, -0.00000578, 0.00000512, 0.00001487, 0.00000988, -0.00000428, -0.00000764, -0.00000080, -0.00000368, -0.00001542, -0.00001293, 0.00000535, 0.00001353, 0.00000263, -0.00000828, -0.00001011, -0.00001349, -0.00001785, -0.00001053, 0.00000179, 0.00000016, -0.00001025, -0.00001155, -0.00001399, -0.00003843, },
		{ 0.00001187, 0.00006615, 0.00006812, 0.00004368, 0.00003524, 0.00004572, 0.00005463, 0.00005243, 0.00003745, 0.00000762, -0.00002633, -0.00004111, -0.00002841, -0.00000637, 0.00000382, -0.00000059, -0.00000891, -0.00001085, -0.00000422, 0.00000293, 0.00000057, -0.00000796, -0.00000805, 0.00000320, 0.00001023, 0.00000379, -0.00000437, -0.00000256, 0.00000223, 0.00000050, -0.00000188, 0.00000229, 0.00000482, -0.00000272, -0.00001033, -0.00000495, 0.00000634, 0.00000785, 0.00000120, -0.00000052, 0.00000275, 0.00000097, -0.00000367, -0.00000050, 0.00000577, 0.00000027, -0.00001271, -0.00001262, 0.00000140, 0.00000726, -0.00000053, -0.00000125, 0.00001075, 0.00001170, -0.00000675, -0.00001683, -0.00000303, 0.00000845, -0.00000212, -0.00001124, 0.00000133, 0.00001165, -0.00000167, -0.00001272, 0.00000255, 0.00001612, -0.00000016, -0.00001645, 0.00000120, 0.00002268, 0.00000663, -0.00002028, -0.00001061, 0.00001567, 0.00000929, -0.00001551, -0.00000980, 0.00001621, 0.00001339, -0.00001344, -0.00001776, 0.00000032, 0.00000072, -0.00001453, -0.00000708, 0.00001851, 0.00001949, -0.00000554, -0.00001448, 0.00000244, 0.00001171, 0.00000115, -0.00000551, 0.00000064, 0.00000213, -0.00000479, -0.00000465, 0.00000313, 0.00000308, -0.00000619, -0.00001171, -0.00000914, -0.00000252, 0.00000682, 0.00001346, 0.00000755, -0.00000546, -0.00000779, -0.00000075, -0.00000372, -0.00001530, -0.00001210, 0.00000676, 0.00001392, 0.00000092, -0.00001033, -0.00001010, -0.00001225, -0.00001775, -0.00001098, 0.00000279, 0.00000127, -0.00001165, -0.00001361, -0.00001306, -0.00003686, -0.00006753, -0.00005275, },
		{ 0.00007644, 0.00004748, 0.00003726, 0.00005188, 0.00006328, 0.00005660, 0.00003421, -0.00000012, -0.00003433, -0.00004595, -0.00002795, -0.00000223, 0.00000652, -0.00000207, -0.00001154, -0.00001092, -0.00000276, 0.00000286, -0.00000074, -0.00000787, -0.00000629, 0.00000438, 0.00000983, 0.00000290, -0.00000500, -0.00000329, 0.00000170, 0.00000112, -0.00000047, 0.00000265, 0.00000342, -0.00000408, -0.00000976, -0.00000311, 0.00000725, 0.00000710, 0.00000005, -0.00000100, 0.00000277, 0.00000119, -0.00000313, -0.00000014, 0.00000503, -0.00000112, -0.00001298, -0.00001139, 0.00000246, 0.00000722, -0.00000051, -0.00000055, 0.00001060, 0.00000986, -0.00000825, -0.00001606, -0.00000133, 0.00000869, -0.00000298, -0.00001131, 0.00000193, 0.00001152, -0.00000209, -0.00001199, 0.00000362, 0.00001545, -0.00000182, -0.00001623, 0.00000321, 0.00002304, 0.00000432, -0.00002171, -0.00000887, 0.00001751, 0.00000802, -0.00001758, -0.00000870, 0.00001906, 0.00001340, -0.00001630, -0.00001927, 0.00000125, 0.00000107, -0.00001525, -0.00000579, 0.00002158, 0.00001967, -0.00000914, -0.00001637, 0.00000518, 0.00001431, -0.00000062, -0.00000817, 0.00000176, 0.00000481, -0.00000518, -0.00000704, 0.00000260, 0.00000381, -0.00000721, -0.00001321, -0.00000754, 0.00000116, 0.00000806, 0.00001124, 0.00000515, -0.00000605, -0.00000776, -0.00000116, -0.00000386, -0.00001448, -0.00001070, 0.00000774, 0.00001343, -0.00000106, -0.00001187, -0.00000953, -0.00001093, -0.00001788, -0.00001156, 0.00000389, 0.00000245, -0.00001331, -0.00001601, -0.00001195, -0.00003462, -0.00007100, -0.00006115, 0.00000816, 0.00007186, },
		{ 0.00003948, 0.00005883, 0.00007293, 0.00006041, 0.00002878, -0.00000997, -0.00004227, -0.00004893, -0.00002612, 0.00000155, 0.00000815, -0.00000382, -0.00001354, -0.00001049, -0.00000153, 0.00000230, -0.00000207, -0.00000737, -0.00000429, 0.00000535, 0.00000911, 0.00000194, -0.00000554, -0.00000393, 0.00000125, 0.00000184, 0.00000089, 0.00000276, 0.00000193, -0.00000515, -0.00000884, -0.00000133, 0.00000777, 0.00000611, -0.00000101, -0.00000130, 0.00000288, 0.00000146, -0.00000261, 0.00000001, 0.00000407, -0.00000243, -0.00001294, -0.00000998, 0.00000342, 0.00000707, -0.00000047, 0.00000008, 0.00001021, 0.00000794, -0.00000946, -0.00001496, 0.00000032, 0.00000868, -0.00000389, -0.00001126, 0.00000256, 0.00001134, -0.00000245, -0.00001119, 0.00000451, 0.00001456, -0.00000332, -0.00001567, 0.00000515, 0.00002293, 0.00000191, -0.00002264, -0.00000683, 0.00001899, 0.00000634, -0.00001938, -0.00000711, 0.00002177, 0.00001283, -0.00001934, -0.00002042, 0.00000245, 0.00000138, -0.00001590, -0.00000429, 0.00002447, 0.00001927, -0.00001292, -0.00001768, 0.00000843, 0.00001658, -0.00000303, -0.00001080, 0.00000348, 0.00000766, -0.00000603, -0.00000974, 0.00000232, 0.00000472, -0.00000844, -0.00001469, -0.00000537, 0.00000516, 0.00000870, 0.00000823, 0.00000284, -0.00000601, -0.00000770, -0.00000209, -0.00000395, -0.00001283, -0.00000888, 0.00000805, 0.00001205, -0.00000304, -0.00001274, -0.00000855, -0.00000973, -0.00001820, -0.00001210, 0.00000508, 0.00000349, -0.00001528, -0.00001856, -0.00001054, -0.00003185, -0.00007441, -0.00007024, 0.00000375, 0.00007773, 0.00008548, 0.00005170, },
		{ 0.00008332, 0.00006319, 0.00002080, -0.00002158, -0.00004949, -0.00004983, -0.00002325, 0.00000460, 0.00000872, -0.00000558, -0.00001482, -0.00000973, -0.00000068, 0.00000133, -0.00000329, -0.00000647, -0.00000216, 0.00000607, 0.00000810, 0.00000094, -0.00000596, -0.00000447, 0.00000089, 0.00000262, 0.00000213, 0.00000263, 0.00000041, -0.00000590, -0.00000765, 0.00000029, 0.00000790, 0.00000496, -0.00000193, -0.00000142, 0.00000307, 0.00000178, -0.00000215, -0.00000004, 0.00000293, -0.00000361, -0.00001257, -0.00000845, 0.00000424, 0.00000680, -0.00000043, 0.00000061, 0.00000959, 0.00000600, -0.00001035, -0.00001360, 0.00000187, 0.00000842, -0.00000483, -0.00001107, 0.00000322, 0.00001110, -0.00000274, -0.00001035, 0.00000519, 0.00001350, -0.00000462, -0.00001479, 0.00000693, 0.00002237, -0.00000050, -0.00002306, -0.00000457, 0.00002002, 0.00000434, -0.00002082, -0.00000507, 0.00002423, 0.00001166, -0.00002244, -0.00002112, 0.00000389, 0.00000162, -0.00001646, -0.00000259, 0.00002710, 0.00001828, -0.00001675, -0.00001832, 0.00001208, 0.00001836, -0.00000604, -0.00001322, 0.00000582, 0.00001055, -0.00000742, -0.00001265, 0.00000234, 0.00000576, -0.00000991, -0.00001608, -0.00000258, 0.00000938, 0.00000861, 0.00000451, 0.00000080, -0.00000532, -0.00000773, -0.00000356, -0.00000381, -0.00001026, -0.00000684, 0.00000743, 0.00000986, -0.00000472, -0.00001285, -0.00000740, -0.00000882, -0.00001857, -0.00001242, 0.00000628, 0.00000422, -0.00001753, -0.00002106, -0.00000879, -0.00002874, -0.00007786, -0.00007985, -0.00000116, 0.00008362, 0.00009496, 0.00005635, 0.00004233, 0.00006687, },
		{ 0.00001014, -0.00003440, -0.00005539, -0.00004858, -0.00001977, 0.00000660, 0.00000834, -0.00000705, -0.00001531, -0.00000882, -0.00000031, 0.00000005, -0.00000427, -0.00000519, -0.00000002, 0.00000649, 0.00000687, -0.00000005, -0.00000627, -0.00000490, 0.00000062, 0.00000342, 0.00000322, 0.00000230, -0.00000103, -0.00000631, -0.00000628, 0.00000166, 0.00000765, 0.00000373, -0.00000266, -0.00000137, 0.00000332, 0.00000210, -0.00000175, -0.00000030, 0.00000166, -0.00000461, -0.00001189, -0.00000684, 0.00000490, 0.00000644, -0.00000037, 0.00000103, 0.00000877, 0.00000411, -0.00001090, -0.00001201, 0.00000325, 0.00000791, -0.00000575, -0.00001071, 0.00000388, 0.00001081, -0.00000298, -0.00000950, 0.00000568, 0.00001231, -0.00000566, -0.00001365, 0.00000849, 0.00002138, -0.00000280, -0.00002295, -0.00000222, 0.00002057, 0.00000211, -0.00002179, -0.00000267, 0.00002629, 0.00000990, -0.00002549, -0.00002133, 0.00000552, 0.00000176, -0.00001692, -0.00000074, 0.00002939, 0.00001674, -0.00002046, -0.00001824, 0.00001599, 0.00001952, -0.00000957, -0.00001526, 0.00000877, 0.00001331, -0.00000939, -0.00001566, 0.00000275, 0.00000687, -0.00001164, -0.00001730, 0.00000083, 0.00001365, 0.00000770, 0.00000019, -0.00000081, -0.00000401, -0.00000801, -0.00000553, -0.00000325, -0.00000675, -0.00000483, 0.00000573, 0.00000705, -0.00000581, -0.00001222, -0.00000632, -0.00000830, -0.00001880, -0.00001238, 0.00000732, 0.00000443, -0.00001995, -0.00002327, -0.00000671, -0.00002554, -0.00008143, -0.00008974, -0.00000644, 0.00008931, 0.00010459, 0.00006162, 0.00004632, 0.00007619, 0.00009393, 0.00006423, },
		{ -0.00005945, -0.00004531, -0.00001614, 0.00000733, 0.00000723, -0.00000801, -0.00001506, -0.00000795, -0.00000048, -0.00000140, -0.00000491, -0.00000360, 0.00000204, 0.00000662, 0.00000548, -0.00000099, -0.00000645, -0.00000522, 0.00000043, 0.00000419, 0.00000412, 0.00000184, -0.00000231, -0.00000638, -0.00000482, 0.00000272, 0.00000706, 0.00000249, -0.00000316, -0.00000116, 0.00000359, 0.00000242, -0.00000147, -0.00000078, 0.00000030, -0.00000538, -0.00001090, -0.00000519, 0.00000538, 0.00000598, -0.00000031, 0.00000132, 0.00000776, 0.00000232, -0.00001109, -0.00001027, 0.00000441, 0.00000717, -0.00000659, -0.00001020, 0.00000452, 0.00001045, -0.00000317, -0.00000866, 0.00000598, 0.00001106, -0.00000644, -0.00001232, 0.00000975, 0.00002003, -0.00000491, -0.00002233, 0.00000013, 0.00002060, -0.00000024, -0.00002225, 0.00000001, 0.00002786, 0.00000760, -0.00002834, -0.00002101, 0.00000728, 0.00000177, -0.00001726, 0.00000120, 0.00003125, 0.00001471, -0.00002391, -0.00001741, 0.00001998, 0.00001997, -0.00001349, -0.00001678, 0.00001228, 0.00001577, -0.00001197, -0.00001863, 0.00000358, 0.00000798, -0.00001365, -0.00001826, 0.00000486, 0.00001783, 0.00000588, -0.00000460, -0.00000185, -0.00000216, -0.00000867, -0.00000793, -0.00000208, -0.00000237, -0.00000315, 0.00000282, 0.00000387, -0.00000605, -0.00001094, -0.00000561, -0.00000819, -0.00001867, -0.00001190, 0.00000800, 0.00000400, -0.00002235, -0.00002496, -0.00000443, -0.00002252, -0.00008506, -0.00009957, -0.00001200, 0.00009443, 0.00011410, 0.00006783, 0.00005203, 0.00008685, 0.00010412, 0.00006286, -0.00000317, -0.00004777, },
		{ -0.00001283, 0.00000669, 0.00000565, -0.00000824, -0.00001418, -0.00000733, -0.00000117, -0.00000286, -0.00000515, -0.00000181, 0.00000391, 0.00000646, 0.00000399, -0.00000186, -0.00000652, -0.00000543, 0.00000031, 0.00000487, 0.00000483, 0.00000131, -0.00000338, -0.00000616, -0.00000338, 0.00000343, 0.00000618, 0.00000131, -0.00000341, -0.00000083, 0.00000386, 0.00000269, -0.00000131, -0.00000146, -0.00000108, -0.00000588, -0.00000964, -0.00000357, 0.00000567, 0.00000546, -0.00000024, 0.00000148, 0.00000661, 0.00000068, -0.00001093, -0.00000843, 0.00000531, 0.00000625, -0.00000731, -0.00000952, 0.00000513, 0.00001000, -0.00000333, -0.00000785, 0.00000612, 0.00000980, -0.00000695, -0.00001087, 0.00001068, 0.00001838, -0.00000673, -0.00002124, 0.00000235, 0.00002012, -0.00000258, -0.00002214, 0.00000285, 0.00002882, 0.00000482, -0.00003088, -0.00002014, 0.00000910, 0.00000162, -0.00001748, 0.00000320, 0.00003265, 0.00001225, -0.00002695, -0.00001586, 0.00002389, 0.00001962, -0.00001765, -0.00001763, 0.00001624, 0.00001778, -0.00001511, -0.00002143, 0.00000486, 0.00000900, -0.00001593, -0.00001887, 0.00000946, 0.00002173, 0.00000312, -0.00000966, -0.00000219, 0.00000011, -0.00000982, -0.00001063, -0.00000014, 0.00000278, -0.00000207, -0.00000132, 0.00000063, -0.00000524, -0.00000921, -0.00000553, -0.00000842, -0.00001791, -0.00001100, 0.00000804, 0.00000283, -0.00002447, -0.00002596, -0.00000217, -0.00001997, -0.00008862, -0.00010899, -0.00001785, 0.00009854, 0.00012330, 0.00007545, 0.00006007, 0.00009871, 0.00011303, 0.00005847, -0.00001883, -0.00006089, -0.00006125, -0.00004030, },
		{ 0.00000390, -0.00000763, -0.00001283, -0.00000711, -0.00000233, -0.00000419, -0.00000497, 0.00000007, 0.00000551, 0.00000607, 0.00000247, -0.00000266, -0.00000648, -0.00000553, 0.00000022, 0.00000543, 0.00000534, 0.00000078, -0.00000417, -0.00000570, -0.00000206, 0.00000376, 0.00000508, 0.00000027, -0.00000342, -0.00000038, 0.00000412, 0.00000290, -0.00000130, -0.00000231, -0.00000243, -0.00000610, -0.00000814, -0.00000202, 0.00000576, 0.00000488, -0.00000018, 0.00000150, 0.00000535, -0.00000075, -0.00001044, -0.00000658, 0.00000593, 0.00000518, -0.00000786, -0.00000869, 0.00000565, 0.00000946, -0.00000346, -0.00000706, 0.00000611, 0.00000857, -0.00000719, -0.00000936, 0.00001126, 0.00001652, -0.00000820, -0.00001975, 0.00000435, 0.00001916, -0.00000481, -0.00002147, 0.00000572, 0.00002913, 0.00000167, -0.00003298, -0.00001875, 0.00001089, 0.00000131, -0.00001757, 0.00000518, 0.00003353, 0.00000948, -0.00002946, -0.00001365, 0.00002752, 0.00001847, -0.00002187, -0.00001772, 0.00002053, 0.00001919, -0.00001876, -0.00002391, 0.00000658, 0.00000985, -0.00001844, -0.00001903, 0.00001458, 0.00002519, -0.00000058, -0.00001481, -0.00000174, 0.00000267, -0.00001155, -0.00001348, 0.00000269, 0.00000850, -0.00000188, -0.00000662, -0.00000233, -0.00000327, -0.00000729, -0.00000629, -0.00000884, -0.00001627, -0.00000975, 0.00000716, 0.00000095, -0.00002599, -0.00002614, -0.00000023, -0.00001811, -0.00009187, -0.00011763, -0.00002414, 0.00010115, 0.00013216, 0.00008513, 0.00007095, 0.00011138, 0.00011973, 0.00005063, -0.00003632, -0.00007295, -0.00006057, -0.00003398, -0.00001029, 0.00000474, },
		{ -0.00001122, -0.00000741, -0.00000385, -0.00000523, -0.00000439, 0.00000192, 0.00000679, 0.00000549, 0.00000098, -0.00000336, -0.00000634, -0.00000551, 0.00000016, 0.00000583, 0.00000569, 0.00000033, -0.00000467, -0.00000508, -0.00000094, 0.00000373, 0.00000386, -0.00000060, -0.00000320, 0.00000014, 0.00000434, 0.00000302, -0.00000145, -0.00000332, -0.00000370, -0.00000601, -0.00000647, -0.00000060, 0.00000566, 0.00000427, -0.00000012, 0.00000138, 0.00000404, -0.00000194, -0.00000965, -0.00000479, 0.00000624, 0.00000403, -0.00000822, -0.00000775, 0.00000607, 0.00000882, -0.00000355, -0.00000631, 0.00000598, 0.00000741, -0.00000719, -0.00000787, 0.00001148, 0.00001452, -0.00000928, -0.00001795, 0.00000604, 0.00001778, -0.00000681, -0.00002026, 0.00000847, 0.00002872, -0.00000172, -0.00003454, -0.00001687, 0.00001258, 0.00000083, -0.00001751, 0.00000710, 0.00003388, 0.00000651, -0.00003135, -0.00001087, 0.00003071, 0.00001653, -0.00002594, -0.00001699, 0.00002496, 0.00001988, -0.00002281, -0.00002594, 0.00000869, 0.00001042, -0.00002113, -0.00001866, 0.00002010, 0.00002803, -0.00000518, -0.00001983, -0.00000047, 0.00000535, -0.00001390, -0.00001629, 0.00000650, 0.00001455, -0.00000282, -0.00001292, -0.00000465, -0.00000009, -0.00000551, -0.00000806, -0.00000921, -0.00001356, -0.00000835, 0.00000509, -0.00000155, -0.00002656, -0.00002550, 0.00000103, -0.00001712, -0.00009445, -0.00012517, -0.00003116, 0.00010175, 0.00014077, 0.00009762, 0.00008508, 0.00012426, 0.00012320, 0.00003908, -0.00005492, -0.00008316, -0.00005739, -0.00002688, -0.00000885, 0.00000162, 0.00000231, -0.00000616, },
		{ -0.00000557, -0.00000587, -0.00000347, 0.00000361, 0.00000773, 0.00000479, -0.00000044, -0.00000398, -0.00000612, -0.00000539, 0.00000008, 0.00000605, 0.00000590, 0.00000001, -0.00000488, -0.00000438, -0.00000009, 0.00000338, 0.00000260, -0.00000124, -0.00000278, 0.00000071, 0.00000451, 0.00000303, -0.00000178, -0.00000442, -0.00000481, -0.00000562, -0.00000466, 0.00000068, 0.00000538, 0.00000365, -0.00000008, 0.00000112, 0.00000271, -0.00000286, -0.00000862, -0.00000311, 0.00000627, 0.00000285, -0.00000835, -0.00000671, 0.00000636, 0.00000810, -0.00000361, -0.00000558, 0.00000576, 0.00000635, -0.00000698, -0.00000646, 0.00001136, 0.00001249, -0.00000994, -0.00001592, 0.00000737, 0.00001606, -0.00000848, -0.00001857, 0.00001098, 0.00002761, -0.00000523, -0.00003547, -0.00001457, 0.00001409, 0.00000020, -0.00001730, 0.00000890, 0.00003371, 0.00000346, -0.00003253, -0.00000765, 0.00003330, 0.00001388, -0.00002967, -0.00001542, 0.00002936, 0.00001974, -0.00002713, -0.00002739, 0.00001114, 0.00001064, -0.00002393, -0.00001770, 0.00002589, 0.00003011, -0.00001058, -0.00002452, 0.00000162, 0.00000799, -0.00001686, -0.00001886, 0.00001131, 0.00002064, -0.00000506, -0.00001996, -0.00000600, 0.00000420, -0.00000420, -0.00001089, -0.00000922, -0.00000962, -0.00000708, 0.00000157, -0.00000443, -0.00002583, -0.00002414, 0.00000118, -0.00001703, -0.00009592, -0.00013138, -0.00003933, 0.00009985, 0.00014946, 0.00011377, 0.00010268, 0.00013650, 0.00012248, 0.00002384, -0.00007372, -0.00009082, -0.00005187, -0.00001962, -0.00000876, -0.00000239, 0.00000114, -0.00000387, -0.00000958, -0.00000827, },
		{ -0.00000230, 0.00000504, 0.00000833, 0.00000404, -0.00000174, -0.00000452, -0.00000582, -0.00000517, -0.00000004, 0.00000608, 0.00000603, -0.00000010, -0.00000483, -0.00000371, 0.00000045, 0.00000276, 0.00000137, -0.00000166, -0.00000218, 0.00000130, 0.00000462, 0.00000290, -0.00000228, -0.00000558, -0.00000572, -0.00000493, -0.00000280, 0.00000176, 0.00000496, 0.00000303, -0.00000008, 0.00000074, 0.00000141, -0.00000350, -0.00000740, -0.00000160, 0.00000604, 0.00000170, -0.00000825, -0.00000563, 0.00000647, 0.00000729, -0.00000361, -0.00000488, 0.00000548, 0.00000540, -0.00000660, -0.00000517, 0.00001093, 0.00001049, -0.00001018, -0.00001376, 0.00000829, 0.00001409, -0.00000974, -0.00001647, 0.00001311, 0.00002583, -0.00000868, -0.00003572, -0.00001194, 0.00001534, -0.00000058, -0.00001695, 0.00001052, 0.00003303, 0.00000045, -0.00003296, -0.00000414, 0.00003517, 0.00001062, -0.00003289, -0.00001305, 0.00003351, 0.00001873, -0.00003155, -0.00002817, 0.00001383, 0.00001045, -0.00002673, -0.00001608, 0.00003181, 0.00003127, -0.00001665, -0.00002866, 0.00000450, 0.00001040, -0.00002040, -0.00002098, 0.00001708, 0.00002643, -0.00000871, -0.00002743, -0.00000609, 0.00000942, -0.00000373, -0.00001475, -0.00000854, -0.00000442, -0.00000626, -0.00000353, -0.00000738, -0.00002351, -0.00002229, -0.00000020, -0.00001775, -0.00009577, -0.00013617, -0.00004922, 0.00009509, 0.00015877, 0.00013443, 0.00012374, 0.00014703, 0.00011666, 0.00000523, -0.00009166, -0.00009539, -0.00004441, -0.00001282, -0.00001011, -0.00000694, 0.00000062, -0.00000093, -0.00000812, -0.00000968, -0.00000732, -0.00000605, },
		{ 0.00000863, 0.00000330, -0.00000291, -0.00000500, -0.00000545, -0.00000488, -0.00000022, 0.00000592, 0.00000612, 0.00000001, -0.00000457, -0.00000314, 0.00000065, 0.00000195, 0.00000025, -0.00000184, -0.00000145, 0.00000189, 0.00000463, 0.00000262, -0.00000294, -0.00000673, -0.00000639, -0.00000399, -0.00000094, 0.00000263, 0.00000440, 0.00000243, -0.00000010, 0.00000026, 0.00000019, -0.00000385, -0.00000605, -0.00000030, 0.00000558, 0.00000065, -0.00000793, -0.00000456, 0.00000640, 0.00000641, -0.00000356, -0.00000419, 0.00000516, 0.00000457, -0.00000609, -0.00000405, 0.00001023, 0.00000861, -0.00001003, -0.00001157, 0.00000878, 0.00001198, -0.00001055, -0.00001408, 0.00001476, 0.00002343, -0.00001195, -0.00003526, -0.00000908, 0.00001628, -0.00000148, -0.00001645, 0.00001191, 0.00003189, -0.00000238, -0.00003266, -0.00000050, 0.00003624, 0.00000690, -0.00003541, -0.00000997, 0.00003722, 0.00001686, -0.00003591, -0.00002821, 0.00001663, 0.00000979, -0.00002942, -0.00001380, 0.00003767, 0.00003144, -0.00002323, -0.00003209, 0.00000808, 0.00001243, -0.00002442, -0.00002245, 0.00002367, 0.00003162, -0.00001378, -0.00003495, -0.00000471, 0.00001530, -0.00000440, -0.00001947, -0.00000683, 0.00000199, -0.00000630, -0.00001029, -0.00000999, -0.00001936, -0.00002029, -0.00000347, -0.00001905, -0.00009348, -0.00013961, -0.00006146, 0.00008723, 0.00016942, 0.00016043, 0.00014794, 0.00015463, 0.00010503, -0.00001619, -0.00010765, -0.00009650, -0.00003557, -0.00000705, -0.00001284, -0.00001163, 0.00000089, 0.00000246, -0.00000704, -0.00001157, -0.00000894, -0.00000573, -0.00000101, 0.00000613, },
		{ -0.00000393, -0.00000542, -0.00000503, -0.00000452, -0.00000047, 0.00000560, 0.00000623, 0.00000037, -0.00000416, -0.00000276, 0.00000054, 0.00000104, -0.00000070, -0.00000181, -0.00000063, 0.00000247, 0.00000456, 0.00000217, -0.00000375, -0.00000782, -0.00000678, -0.00000283, 0.00000085, 0.00000328, 0.00000375, 0.00000185, -0.00000017, -0.00000029, -0.00000091, -0.00000394, -0.00000465, 0.00000075, 0.00000494, -0.00000026, -0.00000741, -0.00000355, 0.00000614, 0.00000550, -0.00000343, -0.00000352, 0.00000481, 0.00000386, -0.00000550, -0.00000312, 0.00000932, 0.00000690, -0.00000953, -0.00000945, 0.00000887, 0.00000984, -0.00001088, -0.00001151, 0.00001584, 0.00002051, -0.00001488, -0.00003411, -0.00000610, 0.00001687, -0.00000246, -0.00001583, 0.00001305, 0.00003034, -0.00000493, -0.00003164, 0.00000309, 0.00003644, 0.00000289, -0.00003712, -0.00000629, 0.00004031, 0.00001415, -0.00004000, -0.00002748, 0.00001943, 0.00000866, -0.00003188, -0.00001087, 0.00004327, 0.00003053, -0.00003009, -0.00003463, 0.00001222, 0.00001392, -0.00002879, -0.00002313, 0.00003093, 0.00003588, -0.00002020, -0.00004210, -0.00000174, 0.00002148, -0.00000648, -0.00002482, -0.00000377, 0.00000942, -0.00000758, -0.00001864, -0.00001180, -0.00001328, -0.00001861, -0.00000895, -0.00002055, -0.00008859, -0.00014195, -0.00007677, 0.00007629, 0.00018236, 0.00019244, 0.00017466, 0.00015798, 0.00008712, -0.00003952, -0.00012061, -0.00009407, -0.00002608, -0.00000279, -0.00001671, -0.00001603, 0.00000196, 0.00000602, -0.00000645, -0.00001379, -0.00001026, -0.00000495, 0.00000029, 0.00000682, 0.00000867, 0.00000264, },
		{ -0.00000457, -0.00000413, -0.00000082, 0.00000515, 0.00000641, 0.00000096, -0.00000369, -0.00000263, 0.00000014, 0.00000010, -0.00000146, -0.00000159, 0.00000026, 0.00000301, 0.00000436, 0.00000154, -0.00000468, -0.00000877, -0.00000687, -0.00000150, 0.00000252, 0.00000371, 0.00000304, 0.00000131, -0.00000029, -0.00000090, -0.00000186, -0.00000380, -0.00000324, 0.00000155, 0.00000418, -0.00000099, -0.00000671, -0.00000265, 0.00000567, 0.00000458, -0.00000321, -0.00000285, 0.00000445, 0.00000326, -0.00000487, -0.00000239, 0.00000826, 0.00000542, -0.00000873, -0.00000746, 0.00000858, 0.00000777, -0.00001073, -0.00000890, 0.00001631, 0.00001720, -0.00001735, -0.00003229, -0.00000311, 0.00001708, -0.00000349, -0.00001508, 0.00001390, 0.00002847, -0.00000710, -0.00002997, 0.00000648, 0.00003579, -0.00000121, -0.00003792, -0.00000218, 0.00004260, 0.00001070, -0.00004365, -0.00002600, 0.00002208, 0.00000705, -0.00003400, -0.00000732, 0.00004842, 0.00002855, -0.00003703, -0.00003617, 0.00001677, 0.00001477, -0.00003334, -0.00002288, 0.00003862, 0.00003895, -0.00002780, -0.00004847, 0.00000284, 0.00002755, -0.00001012, -0.00003043, 0.00000089, 0.00001758, -0.00001048, -0.00002837, -0.00001231, -0.00000530, -0.00001779, -0.00001680, -0.00002171, -0.00008069, -0.00014366, -0.00009583, 0.00006250, 0.00019866, 0.00023093, 0.00020297, 0.00015574, 0.00006278, -0.00006366, -0.00012955, -0.00008826, -0.00001674, -0.00000040, -0.00002139, -0.00001977, 0.00000376, 0.00000944, -0.00000642, -0.00001615, -0.00001115, -0.00000378, 0.00000146, 0.00000711, 0.00000851, 0.00000209, -0.00000481, -0.00000580, },
		{ -0.00000126, 0.00000461, 0.00000670, 0.00000176, -0.00000323, -0.00000278, -0.00000049, -0.00000078, -0.00000200, -0.00000123, 0.00000117, 0.00000350, 0.00000404, 0.00000072, -0.00000567, -0.00000955, -0.00000665, -0.00000008, 0.00000400, 0.00000395, 0.00000230, 0.00000081, -0.00000045, -0.00000152, -0.00000264, -0.00000345, -0.00000189, 0.00000211, 0.00000336, -0.00000152, -0.00000590, -0.00000190, 0.00000503, 0.00000368, -0.00000290, -0.00000220, 0.00000409, 0.00000276, -0.00000424, -0.00000187, 0.00000710, 0.00000419, -0.00000770, -0.00000569, 0.00000797, 0.00000586, -0.00001014, -0.00000637, 0.00001613, 0.00001364, -0.00001925, -0.00002988, -0.00000024, 0.00001690, -0.00000452, -0.00001424, 0.00001444, 0.00002634, -0.00000882, -0.00002774, 0.00000952, 0.00003432, -0.00000520, -0.00003777, 0.00000216, 0.00004396, 0.00000665, -0.00004669, -0.00002382, 0.00002445, 0.00000502, -0.00003565, -0.00000324, 0.00005294, 0.00002551, -0.00004380, -0.00003665, 0.00002154, 0.00001489, -0.00003790, -0.00002162, 0.00004645, 0.00004061, -0.00003632, -0.00005369, 0.00000892, 0.00003307, -0.00001539, -0.00003591, 0.00000733, 0.00002607, -0.00001531, -0.00003915, -0.00001106, 0.00000441, -0.00001840, -0.00002703, -0.00002189, -0.00006953, -0.00014540, -0.00011925, 0.00004640, 0.00021949, 0.00027611, 0.00023161, 0.00014665, 0.00003221, -0.00008729, -0.00013364, -0.00007950, -0.00000840, -0.00000005, -0.00002639, -0.00002252, 0.00000610, 0.00001244, -0.00000694, -0.00001844, -0.00001155, -0.00000233, 0.00000240, 0.00000699, 0.00000823, 0.00000169, -0.00000555, -0.00000614, -0.00000407, -0.00000372, },
		{ 0.00000712, 0.00000270, -0.00000287, -0.00000321, -0.00000127, -0.00000153, -0.00000231, -0.00000076, 0.00000207, 0.00000393, 0.00000358, -0.00000030, -0.00000670, -0.00001008, -0.00000615, 0.00000138, 0.00000527, 0.00000400, 0.00000156, 0.00000034, -0.00000066, -0.00000214, -0.00000324, -0.00000294, -0.00000066, 0.00000245, 0.00000255, -0.00000183, -0.00000501, -0.00000134, 0.00000423, 0.00000284, -0.00000249, -0.00000156, 0.00000373, 0.00000235, -0.00000363, -0.00000153, 0.00000592, 0.00000322, -0.00000651, -0.00000416, 0.00000711, 0.00000420, -0.00000918, -0.00000405, 0.00001534, 0.00000997, -0.00002052, -0.00002697, 0.00000242, 0.00001634, -0.00000550, -0.00001332, 0.00001467, 0.00002406, -0.00001004, -0.00002507, 0.00001207, 0.00003212, -0.00000890, -0.00003667, 0.00000653, 0.00004431, 0.00000215, -0.00004898, -0.00002103, 0.00002641, 0.00000263, -0.00003674, 0.00000125, 0.00005663, 0.00002151, -0.00005015, -0.00003603, 0.00002633, 0.00001427, -0.00004226, -0.00001936, 0.00005413, 0.00004071, -0.00004542, -0.00005741, 0.00001629, 0.00003762, -0.00002221, -0.00004080, 0.00001560, 0.00003440, -0.00002229, -0.00005051, -0.00000762, 0.00001549, -0.00002102, -0.00003948, -0.00002040, -0.00005505, -0.00014800, -0.00014749, 0.00002882, 0.00024601, 0.00032782, 0.00025909, 0.00012962, -0.00000403, -0.00010899, -0.00013230, -0.00006846, -0.00000187, -0.00000171, -0.00003117, -0.00002408, 0.00000869, 0.00001477, -0.00000789, -0.00002043, -0.00001143, -0.00000077, 0.00000303, 0.00000654, 0.00000788, 0.00000143, -0.00000617, -0.00000645, -0.00000354, -0.00000333, -0.00000177, 0.00000404, },
		{ -0.00000268, -0.00000392, -0.00000212, -0.00000209, -0.00000242, -0.00000022, 0.00000295, 0.00000430, 0.00000295, -0.00000149, -0.00000769, -0.00001032, -0.00000538, 0.00000280, 0.00000629, 0.00000390, 0.00000086, -0.00000008, -0.00000091, -0.00000272, -0.00000365, -0.00000233, 0.00000043, 0.00000259, 0.00000180, -0.00000195, -0.00000413, -0.00000099, 0.00000331, 0.00000209, -0.00000198, -0.00000093, 0.00000338, 0.00000200, -0.00000308, -0.00000136, 0.00000476, 0.00000251, -0.00000523, -0.00000291, 0.00000608, 0.00000283, -0.00000792, -0.00000205, 0.00001396, 0.00000636, -0.00002110, -0.00002367, 0.00000476, 0.00001543, -0.00000641, -0.00001235, 0.00001460, 0.00002170, -0.00001074, -0.00002209, 0.00001404, 0.00002931, -0.00001214, -0.00003469, 0.00001072, 0.00004362, -0.00000260, -0.00005040, -0.00001774, 0.00002788, -0.00000002, -0.00003718, 0.00000601, 0.00005935, 0.00001666, -0.00005585, -0.00003435, 0.00003095, 0.00001289, -0.00004623, -0.00001611, 0.00006134, 0.00003918, -0.00005472, -0.00005937, 0.00002465, 0.00004080, -0.00003041, -0.00004464, 0.00002562, 0.00004205, -0.00003151, -0.00006188, -0.00000169, 0.00002743, -0.00002618, -0.00005376, -0.00001650, -0.00003742, -0.00015241, -0.00018081, 0.00001086, 0.00027933, 0.00038554, 0.00028371, 0.00010382, -0.00004499, -0.00012729, -0.00012525, -0.00005603, 0.00000214, -0.00000516, -0.00003517, -0.00002436, 0.00001117, 0.00001622, -0.00000909, -0.00002192, -0.00001084, 0.00000077, 0.00000332, 0.00000583, 0.00000752, 0.00000130, -0.00000668, -0.00000671, -0.00000302, -0.00000296, -0.00000234, 0.00000350, 0.00000768, 0.00000371, },
		{ -0.00000295, -0.00000244, -0.00000236, 0.00000034, 0.00000379, 0.00000458, 0.00000215, -0.00000283, -0.00000859, -0.00001024, -0.00000440, 0.00000411, 0.00000705, 0.00000370, 0.00000023, -0.00000048, -0.00000120, -0.00000323, -0.00000389, -0.00000167, 0.00000136, 0.00000260, 0.00000115, -0.00000188, -0.00000329, -0.00000087, 0.00000231, 0.00000147, -0.00000138, -0.00000034, 0.00000302, 0.00000170, -0.00000260, -0.00000132, 0.00000366, 0.00000204, -0.00000394, -0.00000193, 0.00000496, 0.00000179, -0.00000645, -0.00000044, 0.00001209, 0.00000294, -0.00002098, -0.00002012, 0.00000672, 0.00001422, -0.00000720, -0.00001136, 0.00001425, 0.00001937, -0.00001091, -0.00001893, 0.00001538, 0.00002603, -0.00001477, -0.00003192, 0.00001452, 0.00004189, -0.00000741, -0.00005089, -0.00001409, 0.00002877, -0.00000281, -0.00003693, 0.00001088, 0.00006099, 0.00001113, -0.00006069, -0.00003167, 0.00003520, 0.00001082, -0.00004962, -0.00001199, 0.00006777, 0.00003607, -0.00006380, -0.00005942, 0.00003362, 0.00004231, -0.00003968, -0.00004702, 0.00003718, 0.00004848, -0.00004290, -0.00007264, 0.00000690, 0.00003960, -0.00003427, -0.00006931, -0.00000954, -0.00001706, -0.00015966, -0.00021917, -0.00000610, 0.00032034, 0.00044831, 0.00030369, 0.00006875, -0.00008936, -0.00014082, -0.00011254, -0.00004325, 0.00000308, -0.00001000, -0.00003783, -0.00002341, 0.00001314, 0.00001670, -0.00001030, -0.00002274, -0.00000989, 0.00000210, 0.00000328, 0.00000494, 0.00000717, 0.00000127, -0.00000709, -0.00000692, -0.00000251, -0.00000264, -0.00000294, 0.00000302, 0.00000835, 0.00000470, -0.00000270, -0.00000484, },
		{ -0.00000216, 0.00000090, 0.00000458, 0.00000475, 0.00000116, -0.00000428, -0.00000933, -0.00000983, -0.00000325, 0.00000524, 0.00000754, 0.00000342, -0.00000033, -0.00000085, -0.00000152, -0.00000366, -0.00000397, -0.00000100, 0.00000211, 0.00000251, 0.00000064, -0.00000168, -0.00000257, -0.00000096, 0.00000128, 0.00000100, -0.00000071, 0.00000022, 0.00000267, 0.00000143, -0.00000220, -0.00000139, 0.00000267, 0.00000178, -0.00000270, -0.00000122, 0.00000383, 0.00000108, -0.00000490, 0.00000070, 0.00000983, -0.00000014, -0.00002020, -0.00001645, 0.00000823, 0.00001275, -0.00000785, -0.00001037, 0.00001363, 0.00001712, -0.00001059, -0.00001574, 0.00001604, 0.00002245, -0.00001668, -0.00002850, 0.00001774, 0.00003920, -0.00001205, -0.00005042, -0.00001025, 0.00002904, -0.00000563, -0.00003596, 0.00001567, 0.00006146, 0.00000513, -0.00006450, -0.00002812, 0.00003889, 0.00000815, -0.00005226, -0.00000715, 0.00007315, 0.00003148, -0.00007225, -0.00005750, 0.00004273, 0.00004192, -0.00004963, -0.00004758, 0.00004993, 0.00005316, -0.00005624, -0.00008210, 0.00001815, 0.00005126, -0.00004557, -0.00008539, 0.00000107, 0.00000535, -0.00017074, -0.00026220, -0.00002053, 0.00036971, 0.00051478, 0.00031722, 0.00002435, -0.00013555, -0.00014833, -0.00009455, -0.00003123, 0.00000058, -0.00001564, -0.00003869, -0.00002145, 0.00001421, 0.00001621, -0.00001123, -0.00002279, -0.00000875, 0.00000311, 0.00000294, 0.00000398, 0.00000683, 0.00000130, -0.00000738, -0.00000707, -0.00000203, -0.00000239, -0.00000350, 0.00000266, 0.00000910, 0.00000559, -0.00000297, -0.00000592, -0.00000369, -0.00000256, },
		{ 0.00000529, 0.00000480, -0.00000001, -0.00000580, -0.00000986, -0.00000909, -0.00000202, 0.00000615, 0.00000779, 0.00000313, -0.00000079, -0.00000120, -0.00000184, -0.00000400, -0.00000392, -0.00000037, 0.00000269, 0.00000238, 0.00000029, -0.00000140, -0.00000201, -0.00000125, 0.00000027, 0.00000068, 0.00000002, 0.00000072, 0.00000231, 0.00000118, -0.00000188, -0.00000153, 0.00000181, 0.00000169, -0.00000154, -0.00000074, 0.00000277, 0.00000070, -0.00000335, 0.00000136, 0.00000729, -0.00000278, -0.00001882, -0.00001279, 0.00000928, 0.00001111, -0.00000833, -0.00000942, 0.00001280, 0.00001503, -0.00000984, -0.00001265, 0.00001606, 0.00001875, -0.00001781, -0.00002461, 0.00002024, 0.00003564, -0.00001634, -0.00004901, -0.00000638, 0.00002867, -0.00000832, -0.00003429, 0.00002020, 0.00006074, -0.00000112, -0.00006713, -0.00002384, 0.00004189, 0.00000499, -0.00005405, -0.00000179, 0.00007724, 0.00002561, -0.00007967, -0.00005367, 0.00005153, 0.00003953, -0.00005978, -0.00004606, 0.00006341, 0.00005564, -0.00007117, -0.00008963, 0.00003187, 0.00006162, -0.00006014, -0.00010111, 0.00001573, 0.00002887, -0.00018661, -0.00030920, -0.00003076, 0.00042772, 0.00058323, 0.00032261, -0.00002902, -0.00018174, -0.00014884, -0.00007201, -0.00002112, -0.00000543, -0.00002140, -0.00003738, -0.00001881, 0.00001401, 0.00001486, -0.00001157, -0.00002203, -0.00000763, 0.00000369, 0.00000242, 0.00000303, 0.00000647, 0.00000135, -0.00000754, -0.00000714, -0.00000163, -0.00000220, -0.00000399, 0.00000245, 0.00000985, 0.00000629, -0.00000348, -0.00000707, -0.00000426, -0.00000246, -0.00000186, 0.00000143, },
		{ -0.00000135, -0.00000733, -0.00001013, -0.00000803, -0.00000076, 0.00000680, 0.00000783, 0.00000284, -0.00000115, -0.00000154, -0.00000216, -0.00000423, -0.00000377, 0.00000020, 0.00000312, 0.00000227, 0.00000009, -0.00000108, -0.00000164, -0.00000171, -0.00000065, 0.00000054, 0.00000078, 0.00000116, 0.00000195, 0.00000094, -0.00000164, -0.00000171, 0.00000108, 0.00000172, -0.00000052, -0.00000044, 0.00000184, 0.00000059, -0.00000192, 0.00000153, 0.00000461, -0.00000490, -0.00001690, -0.00000927, 0.00000986, 0.00000935, -0.00000864, -0.00000852, 0.00001178, 0.00001315, -0.00000871, -0.00000976, 0.00001546, 0.00001510, -0.00001814, -0.00002045, 0.00002190, 0.00003138, -0.00002010, -0.00004673, -0.00000264, 0.00002770, -0.00001078, -0.00003198, 0.00002427, 0.00005884, -0.00000737, -0.00006849, -0.00001903, 0.00004407, 0.00000150, -0.00005489, 0.00000387, 0.00007986, 0.00001873, -0.00008570, -0.00004809, 0.00005955, 0.00003514, -0.00006961, -0.00004233, 0.00007707, 0.00005560, -0.00008715, -0.00009465, 0.00004770, 0.00006990, -0.00007782, -0.00011550, 0.00003459, 0.00005237, -0.00020800, -0.00035909, -0.00003511, 0.00049426, 0.00065162, 0.00031839, -0.00009049, -0.00022600, -0.00014169, -0.00004596, -0.00001396, -0.00001482, -0.00002650, -0.00003371, -0.00001595, 0.00001223, 0.00001287, -0.00001103, -0.00002053, -0.00000676, 0.00000378, 0.00000184, 0.00000217, 0.00000605, 0.00000139, -0.00000755, -0.00000712, -0.00000133, -0.00000208, -0.00000433, 0.00000240, 0.00001054, 0.00000672, -0.00000421, -0.00000819, -0.00000462, -0.00000216, -0.00000151, 0.00000192, 0.00000592, 0.00000469, },
		{ -0.00001008, -0.00000671, 0.00000044, 0.00000715, 0.00000768, 0.00000262, -0.00000142, -0.00000187, -0.00000247, -0.00000435, -0.00000356, 0.00000066, 0.00000343, 0.00000221, 0.00000003, -0.00000080, -0.00000149, -0.00000228, -0.00000146, 0.00000056, 0.00000154, 0.00000152, 0.00000159, 0.00000070, -0.00000146, -0.00000190, 0.00000049, 0.00000184, 0.00000035, -0.00000029, 0.00000108, 0.00000071, -0.00000069, 0.00000124, 0.00000194, -0.00000643, -0.00001457, -0.00000600, 0.00000999, 0.00000756, -0.00000877, -0.00000770, 0.00001064, 0.00001152, -0.00000730, -0.00000718, 0.00001435, 0.00001165, -0.00001769, -0.00001621, 0.00002265, 0.00002658, -0.00002317, -0.00004367, 0.00000082, 0.00002619, -0.00001287, -0.00002912, 0.00002772, 0.00005584, -0.00001339, -0.00006855, -0.00001389, 0.00004535, -0.00000216, -0.00005475, 0.00000957, 0.00008091, 0.00001116, -0.00009006, -0.00004103, 0.00006636, 0.00002891, -0.00007861, -0.00003640, 0.00009030, 0.00005283, -0.00010356, -0.00009667, 0.00006507, 0.00007535, -0.00009823, -0.00012755, 0.00005757, 0.00007454, -0.00023543, -0.00041045, -0.00003199, 0.00056873, 0.00071767, 0.00030341, -0.00015877, -0.00026636, -0.00012659, -0.00001767, -0.00001066, -0.00002717, -0.00003016, -0.00002766, -0.00001337, 0.00000870, 0.00001055, -0.00000936, -0.00001842, -0.00000637, 0.00000341, 0.00000135, 0.00000143, 0.00000549, 0.00000140, -0.00000735, -0.00000701, -0.00000118, -0.00000202, -0.00000448, 0.00000250, 0.00001106, 0.00000681, -0.00000509, -0.00000918, -0.00000475, -0.00000172, -0.00000114, 0.00000236, 0.00000646, 0.00000442, -0.00000285, -0.00000880, },
		{ 0.00000151, 0.00000722, 0.00000739, 0.00000247, -0.00000159, -0.00000220, -0.00000276, -0.00000438, -0.00000331, 0.00000103, 0.00000367, 0.00000224, 0.00000008, -0.00000061, -0.00000155, -0.00000291, -0.00000210, 0.00000072, 0.00000228, 0.00000181, 0.00000122, 0.00000047, -0.00000134, -0.00000207, 0.00000004, 0.00000200, 0.00000106, -0.00000022, 0.00000051, 0.00000099, 0.00000027, 0.00000056, -0.00000061, -0.00000737, -0.00001195, -0.00000308, 0.00000970, 0.00000579, -0.00000873, -0.00000698, 0.00000941, 0.00001015, -0.00000570, -0.00000496, 0.00001281, 0.00000856, -0.00001653, -0.00001209, 0.00002247, 0.00002146, -0.00002546, -0.00003996, 0.00000389, 0.00002421, -0.00001451, -0.00002583, 0.00003040, 0.00005186, -0.00001893, -0.00006731, -0.00000864, 0.00004571, -0.00000582, -0.00005364, 0.00001507, 0.00008037, 0.00000326, -0.00009256, -0.00003285, 0.00007160, 0.00002110, -0.00008627, -0.00002842, 0.00010246, 0.00004727, -0.00011966, -0.00009539, 0.00008327, 0.00007737, -0.00012075, -0.00013630, 0.00008429, 0.00009399, -0.00026909, -0.00046155, -0.00002000, 0.00065008, 0.00077899, 0.00027693, -0.00023212, -0.00030096, -0.00010368, 0.00001141, -0.00001190, -0.00004182, -0.00003168, -0.00001941, -0.00001166, 0.00000334, 0.00000829, -0.00000638, -0.00001593, -0.00000667, 0.00000265, 0.00000112, 0.00000081, 0.00000469, 0.00000135, -0.00000690, -0.00000681, -0.00000120, -0.00000197, -0.00000436, 0.00000272, 0.00001131, 0.00000655, -0.00000603, -0.00000994, -0.00000464, -0.00000119, -0.00000078, 0.00000276, 0.00000687, 0.00000395, -0.00000448, -0.00001013, -0.00000971, -0.00000518, },
		{ 0.00000702, 0.00000242, -0.00000167, -0.00000252, -0.00000301, -0.00000432, -0.00000306, 0.00000129, 0.00000387, 0.00000237, 0.00000019, -0.00000056, -0.00000181, -0.00000354, -0.00000256, 0.00000100, 0.00000296, 0.00000201, 0.00000086, 0.00000024, -0.00000126, -0.00000221, -0.00000030, 0.00000217, 0.00000162, -0.00000019, 0.00000015, 0.00000134, 0.00000091, -0.00000043, -0.00000292, -0.00000772, -0.00000916, -0.00000057, 0.00000906, 0.00000411, -0.00000854, -0.00000636, 0.00000815, 0.00000904, -0.00000400, -0.00000316, 0.00001098, 0.00000592, -0.00001477, -0.00000828, 0.00002140, 0.00001622, -0.00002689, -0.00003577, 0.00000647, 0.00002188, -0.00001563, -0.00002226, 0.00003220, 0.00004705, -0.00002378, -0.00006484, -0.00000348, 0.00004515, -0.00000931, -0.00005163, 0.00002012, 0.00007827, -0.00000461, -0.00009310, -0.00002395, 0.00007499, 0.00001207, -0.00009218, -0.00001867, 0.00011295, 0.00003907, -0.00013469, -0.00009067, 0.00010147, 0.00007550, -0.00014456, -0.00014091, 0.00011405, 0.00010932, -0.00030881, -0.00051046, 0.00000194, 0.00073675, 0.00083316, 0.00023872, -0.00030846, -0.00032814, -0.00007354, 0.00003972, -0.00001809, -0.00005792, -0.00003043, -0.00000932, -0.00001134, -0.00000375, 0.00000654, -0.00000201, -0.00001335, -0.00000782, 0.00000163, 0.00000129, 0.00000027, 0.00000357, 0.00000128, -0.00000614, -0.00000655, -0.00000145, -0.00000191, -0.00000394, 0.00000301, 0.00001119, 0.00000592, -0.00000693, -0.00001041, -0.00000432, -0.00000061, -0.00000043, 0.00000311, 0.00000714, 0.00000326, -0.00000617, -0.00001125, -0.00000899, -0.00000350, 0.00000240, 0.00000701, },
		{ -0.00000169, -0.00000284, -0.00000321, -0.00000420, -0.00000284, 0.00000146, 0.00000406, 0.00000261, 0.00000032, -0.00000069, -0.00000225, -0.00000411, -0.00000281, 0.00000137, 0.00000355, 0.00000214, 0.00000052, 0.00000001, -0.00000122, -0.00000232, -0.00000053, 0.00000231, 0.00000205, -0.00000014, -0.00000003, 0.00000169, 0.00000121, -0.00000164, -0.00000488, -0.00000753, -0.00000634, 0.00000148, 0.00000813, 0.00000257, -0.00000822, -0.00000584, 0.00000690, 0.00000819, -0.00000230, -0.00000177, 0.00000898, 0.00000381, -0.00001253, -0.00000494, 0.00001950, 0.00001107, -0.00002744, -0.00003126, 0.00000848, 0.00001931, -0.00001619, -0.00001855, 0.00003306, 0.00004161, -0.00002777, -0.00006127, 0.00000138, 0.00004372, -0.00001250, -0.00004880, 0.00002451, 0.00007475, -0.00001207, -0.00009168, -0.00001476, 0.00007637, 0.00000227, -0.00009600, -0.00000757, 0.00012124, 0.00002852, -0.00014792, -0.00008258, 0.00011878, 0.00006950, -0.00016867, -0.00014071, 0.00014588, 0.00011922, -0.00035399, -0.00055509, 0.00003455, 0.00082675, 0.00087791, 0.00018907, -0.00038545, -0.00034654, -0.00003712, 0.00006573, -0.00002933, -0.00007448, -0.00002597, 0.00000208, -0.00001293, -0.00001235, 0.00000572, 0.00000373, -0.00001101, -0.00000993, 0.00000056, 0.00000199, -0.00000027, 0.00000204, 0.00000123, -0.00000502, -0.00000625, -0.00000194, -0.00000178, -0.00000318, 0.00000330, 0.00001061, 0.00000495, -0.00000767, -0.00001053, -0.00000385, -0.00000006, -0.00000009, 0.00000343, 0.00000722, 0.00000235, -0.00000787, -0.00001210, -0.00000795, -0.00000178, 0.00000305, 0.00000654, 0.00000661, 0.00000248, },
		{ -0.00000338, -0.00000403, -0.00000265, 0.00000156, 0.00000429, 0.00000293, 0.00000040, -0.00000101, -0.00000280, -0.00000457, -0.00000286, 0.00000179, 0.00000404, 0.00000220, 0.00000019, -0.00000021, -0.00000119, -0.00000237, -0.00000069, 0.00000241, 0.00000237, -0.00000005, -0.00000004, 0.00000198, 0.00000118, -0.00000296, -0.00000643, -0.00000688, -0.00000360, 0.00000306, 0.00000701, 0.00000121, -0.00000780, -0.00000544, 0.00000569, 0.00000756, -0.00000067, -0.00000078, 0.00000694, 0.00000228, -0.00000999, -0.00000220, 0.00001689, 0.00000620, -0.00002713, -0.00002660, 0.00000992, 0.00001664, -0.00001619, -0.00001489, 0.00003293, 0.00003577, -0.00003075, -0.00005674, 0.00000576, 0.00004151, -0.00001524, -0.00004530, 0.00002807, 0.00006997, -0.00001882, -0.00008841, -0.00000573, 0.00007570, -0.00000781, -0.00009750, 0.00000437, 0.00012688, 0.00001608, -0.00015864, -0.00007141, 0.00013428, 0.00005937, -0.00019199, -0.00013525, 0.00017857, 0.00012253, -0.00040365, -0.00059331, 0.00007804, 0.00091774, 0.00091119, 0.00012881, -0.00046062, -0.00035519, 0.00000423, 0.00008801, -0.00004539, -0.00009041, -0.00001806, 0.00001411, -0.00001679, -0.00002208, 0.00000624, 0.00001069, -0.00000926, -0.00001303, -0.00000035, 0.00000331, -0.00000095, 0.00000004, 0.00000128, -0.00000348, -0.00000598, -0.00000271, -0.00000153, -0.00000206, 0.00000349, 0.00000948, 0.00000371, -0.00000814, -0.00001027, -0.00000328, 0.00000044, 0.00000025, 0.00000369, 0.00000707, 0.00000121, -0.00000951, -0.00001260, -0.00000663, -0.00000008, 0.00000344, 0.00000587, 0.00000621, 0.00000265, -0.00000165, -0.00000314, },
		{ -0.00000251, 0.00000164, 0.00000458, 0.00000331, 0.00000039, -0.00000154, -0.00000342, -0.00000486, -0.00000272, 0.00000221, 0.00000440, 0.00000220, -0.00000010, -0.00000042, -0.00000118, -0.00000238, -0.00000080, 0.00000247, 0.00000261, 0.00000012, 0.00000007, 0.00000215, 0.00000084, -0.00000429, -0.00000753, -0.00000585, -0.00000106, 0.00000418, 0.00000575, 0.00000005, -0.00000731, -0.00000514, 0.00000456, 0.00000713, 0.00000082, -0.00000016, 0.00000500, 0.00000131, -0.00000730, -0.00000014, 0.00001372, 0.00000177, -0.00002602, -0.00002196, 0.00001077, 0.00001398, -0.00001567, -0.00001142, 0.00003185, 0.00002974, -0.00003263, -0.00005146, 0.00000952, 0.00003862, -0.00001744, -0.00004129, 0.00003066, 0.00006418, -0.00002456, -0.00008347, 0.00000272, 0.00007304, -0.00001766, -0.00009663, 0.00001656, 0.00012958, 0.00000233, -0.00016629, -0.00005762, 0.00014711, 0.00004537, -0.00021337, -0.00012442, 0.00021071, 0.00011840, -0.00045641, -0.00062308, 0.00013214, 0.00100709, 0.00093135, 0.00005934, -0.00053145, -0.00035360, 0.00004889, 0.00010530, -0.00006571, -0.00010465, -0.00000671, 0.00002602, -0.00002318, -0.00003245, 0.00000846, 0.00001863, -0.00000846, -0.00001703, -0.00000083, 0.00000526, -0.00000192, -0.00000249, 0.00000151, -0.00000149, -0.00000579, -0.00000375, -0.00000109, -0.00000058, 0.00000347, 0.00000776, 0.00000230, -0.00000820, -0.00000964, -0.00000272, 0.00000084, 0.00000062, 0.00000390, 0.00000665, -0.00000013, -0.00001099, -0.00001271, -0.00000509, 0.00000149, 0.00000354, 0.00000505, 0.00000586, 0.00000293, -0.00000157, -0.00000344, -0.00000351, -0.00000385, },
		{ 0.00000493, 0.00000370, 0.00000022, -0.00000225, -0.00000404, -0.00000495, -0.00000241, 0.00000259, 0.00000465, 0.00000217, -0.00000036, -0.00000062, -0.00000117, -0.00000234, -0.00000087, 0.00000247, 0.00000281, 0.00000037, 0.00000025, 0.00000215, 0.00000024, -0.00000553, -0.00000815, -0.00000454, 0.00000121, 0.00000487, 0.00000445, -0.00000091, -0.00000679, -0.00000492, 0.00000353, 0.00000683, 0.00000212, 0.00000017, 0.00000325, 0.00000088, -0.00000463, 0.00000119, 0.00001015, -0.00000207, -0.00002421, -0.00001748, 0.00001109, 0.00001146, -0.00001468, -0.00000830, 0.00002987, 0.00002377, -0.00003338, -0.00004562, 0.00001255, 0.00003521, -0.00001903, -0.00003695, 0.00003222, 0.00005765, -0.00002910, -0.00007713, 0.00001024, 0.00006857, -0.00002680, -0.00009342, 0.00002841, 0.00012920, -0.00001205, -0.00017044, -0.00004187, 0.00015652, 0.00002798, -0.00023169, -0.00010837, 0.00024077, 0.00010628, -0.00051055, -0.00064259, 0.00019599, 0.00109204, 0.00093718, -0.00001748, -0.00059556, -0.00034175, 0.00009502, 0.00011665, -0.00008945, -0.00011620, 0.00000785, 0.00003698, -0.00003217, -0.00004286, 0.00001262, 0.00002717, -0.00000889, -0.00002179, -0.00000063, 0.00000781, -0.00000334, -0.00000554, 0.00000205, 0.00000095, -0.00000575, -0.00000508, -0.00000039, 0.00000122, 0.00000313, 0.00000542, 0.00000084, -0.00000776, -0.00000867, -0.00000224, 0.00000114, 0.00000104, 0.00000404, 0.00000592, -0.00000164, -0.00001223, -0.00001241, -0.00000340, 0.00000285, 0.00000338, 0.00000414, 0.00000558, 0.00000329, -0.00000145, -0.00000371, -0.00000360, -0.00000367, -0.00000240, 0.00000171, },
		{ -0.00000012, -0.00000311, -0.00000460, -0.00000483, -0.00000198, 0.00000291, 0.00000477, 0.00000211, -0.00000058, -0.00000081, -0.00000116, -0.00000227, -0.00000091, 0.00000242, 0.00000298, 0.00000070, 0.00000046, 0.00000195, -0.00000056, -0.00000658, -0.00000833, -0.00000308, 0.00000313, 0.00000518, 0.00000317, -0.00000165, -0.00000625, -0.00000478, 0.00000262, 0.00000664, 0.00000320, 0.00000026, 0.00000179, 0.00000092, -0.00000213, 0.00000180, 0.00000638, -0.00000523, -0.00002182, -0.00001331, 0.00001092, 0.00000916, -0.00001331, -0.00000564, 0.00002711, 0.00001807, -0.00003302, -0.00003946, 0.00001478, 0.00003141, -0.00001998, -0.00003246, 0.00003271, 0.00005066, -0.00003229, -0.00006972, 0.00001653, 0.00006259, -0.00003479, -0.00008808, 0.00003931, 0.00012576, -0.00002632, -0.00017085, -0.00002493, 0.00016193, 0.00000795, -0.00024594, -0.00008761, 0.00026719, 0.00008602, -0.00056409, -0.00065029, 0.00026822, 0.00116978, 0.00092805, -0.00009939, -0.00065079, -0.00032012, 0.00014067, 0.00012140, -0.00011554, -0.00012417, 0.00002515, 0.00004624, -0.00004364, -0.00005269, 0.00001882, 0.00003587, -0.00001079, -0.00002706, 0.00000047, 0.00001085, -0.00000539, -0.00000906, 0.00000302, 0.00000384, -0.00000595, -0.00000666, 0.00000063, 0.00000330, 0.00000236, 0.00000248, -0.00000054, -0.00000673, -0.00000744, -0.00000193, 0.00000135, 0.00000156, 0.00000406, 0.00000484, -0.00000328, -0.00001312, -0.00001169, -0.00000166, 0.00000395, 0.00000299, 0.00000321, 0.00000539, 0.00000371, -0.00000131, -0.00000395, -0.00000367, -0.00000350, -0.00000232, 0.00000182, 0.00000534, 0.00000404, },
		{ -0.00000504, -0.00000449, -0.00000148, 0.00000312, 0.00000477, 0.00000204, -0.00000075, -0.00000098, -0.00000116, -0.00000217, -0.00000095, 0.00000234, 0.00000315, 0.00000109, 0.00000064, 0.00000155, -0.00000147, -0.00000738, -0.00000809, -0.00000156, 0.00000467, 0.00000518, 0.00000198, -0.00000220, -0.00000574, -0.00000470, 0.00000183, 0.00000650, 0.00000403, 0.00000022, 0.00000067, 0.00000133, 0.00000004, 0.00000172, 0.00000257, -0.00000765, -0.00001899, -0.00000953, 0.00001036, 0.00000714, -0.00001168, -0.00000355, 0.00002372, 0.00001284, -0.00003159, -0.00003319, 0.00001618, 0.00002739, -0.00002030, -0.00002799, 0.00003218, 0.00004350, -0.00003408, -0.00006161, 0.00002138, 0.00005543, -0.00004128, -0.00008093, 0.00004873, 0.00011947, -0.00003976, -0.00016746, -0.00000768, 0.00016296, -0.00001382, -0.00025527, -0.00006294, 0.00028850, 0.00005792, -0.00061485, -0.00064511, 0.00034698, 0.00123764, 0.00090391, -0.00018381, -0.00069534, -0.00028969, 0.00018392, 0.00011930, -0.00014271, -0.00012793, 0.00004450, 0.00005306, -0.00005729, -0.00006130, 0.00002706, 0.00004423, -0.00001431, -0.00003254, 0.00000267, 0.00001422, -0.00000823, -0.00001295, 0.00000455, 0.00000711, -0.00000648, -0.00000846, 0.00000202, 0.00000557, 0.00000104, -0.00000100, -0.00000167, -0.00000508, -0.00000605, -0.00000186, 0.00000149, 0.00000219, 0.00000393, 0.00000339, -0.00000496, -0.00001357, -0.00001058, 0.00000001, 0.00000471, 0.00000241, 0.00000232, 0.00000529, 0.00000416, -0.00000115, -0.00000416, -0.00000374, -0.00000337, -0.00000224, 0.00000199, 0.00000580, 0.00000429, -0.00000066, -0.00000408, },
		{ -0.00000094, 0.00000321, 0.00000468, 0.00000198, -0.00000088, -0.00000115, -0.00000115, -0.00000204, -0.00000098, 0.00000224, 0.00000334, 0.00000153, 0.00000075, 0.00000097, -0.00000242, -0.00000789, -0.00000750, -0.00000008, 0.00000581, 0.00000494, 0.00000092, -0.00000259, -0.00000528, -0.00000466, 0.00000116, 0.00000637, 0.00000463, 0.00000011, -0.00000007, 0.00000200, 0.00000180, 0.00000102, -0.00000109, -0.00000931, -0.00001587, -0.00000621, 0.00000950, 0.00000547, -0.00000990, -0.00000207, 0.00001986, 0.00000824, -0.00002922, -0.00002704, 0.00001676, 0.00002330, -0.00002001, -0.00002372, 0.00003071, 0.00003645, -0.00003450, -0.00005315, 0.00002469, 0.00004753, -0.00004600, -0.00007239, 0.00005622, 0.00011067, -0.00005167, -0.00016046, 0.00000899, 0.00015945, -0.00003628, -0.00025908, -0.00003546, 0.00030337, 0.00002270, -0.00066062, -0.00062644, 0.00042998, 0.00129320, 0.00086535, -0.00026801, -0.00072783, -0.00025185, 0.00022295, 0.00011046, -0.00016963, -0.00012704, 0.00006504, 0.00005688, -0.00007264, -0.00006808, 0.00003714, 0.00005174, -0.00001948, -0.00003784, 0.00000608, 0.00001767, -0.00001198, -0.00001708, 0.00000676, 0.00001070, -0.00000741, -0.00001040, 0.00000384, 0.00000793, -0.00000091, -0.00000494, -0.00000239, -0.00000279, -0.00000460, -0.00000210, 0.00000160, 0.00000296, 0.00000360, 0.00000153, -0.00000659, -0.00001352, -0.00000912, 0.00000151, 0.00000513, 0.00000172, 0.00000154, 0.00000526, 0.00000462, -0.00000097, -0.00000432, -0.00000381, -0.00000326, -0.00000213, 0.00000224, 0.00000625, 0.00000439, -0.00000140, -0.00000507, -0.00000530, -0.00000396, },
		{ 0.00000451, 0.00000195, -0.00000096, -0.00000131, -0.00000114, -0.00000191, -0.00000100, 0.00000213, 0.00000356, 0.00000197, 0.00000076, 0.00000023, -0.00000332, -0.00000808, -0.00000665, 0.00000126, 0.00000657, 0.00000453, 0.00000001, -0.00000284, -0.00000487, -0.00000464, 0.00000060, 0.00000622, 0.00000502, 0.00000002, -0.00000044, 0.00000282, 0.00000305, -0.00000018, -0.00000446, -0.00001024, -0.00001261, -0.00000340, 0.00000843, 0.00000414, -0.00000810, -0.00000122, 0.00001572, 0.00000440, -0.00002605, -0.00002120, 0.00001658, 0.00001928, -0.00001918, -0.00001976, 0.00002842, 0.00002976, -0.00003365, -0.00004471, 0.00002646, 0.00003929, -0.00004884, -0.00006293, 0.00006143, 0.00009986, -0.00006145, -0.00015019, 0.00002421, 0.00015151, -0.00005831, -0.00025704, -0.00000647, 0.00031072, -0.00001847, -0.00069923, -0.00059425, 0.00051461, 0.00133445, 0.00081354, -0.00034918, -0.00074743, -0.00020832, 0.00025616, 0.00009539, -0.00019492, -0.00012139, 0.00008582, 0.00005729, -0.00008907, -0.00007251, 0.00004874, 0.00005789, -0.00002620, -0.00004258, 0.00001074, 0.00002095, -0.00001671, -0.00002124, 0.00000974, 0.00001449, -0.00000882, -0.00001242, 0.00000610, 0.00001026, -0.00000356, -0.00000921, -0.00000256, 0.00000010, -0.00000324, -0.00000269, 0.00000176, 0.00000387, 0.00000299, -0.00000071, -0.00000808, -0.00001290, -0.00000740, 0.00000273, 0.00000519, 0.00000100, 0.00000090, 0.00000528, 0.00000505, -0.00000076, -0.00000445, -0.00000391, -0.00000319, -0.00000196, 0.00000259, 0.00000666, 0.00000427, -0.00000232, -0.00000603, -0.00000535, -0.00000325, -0.00000042, 0.00000318, },
		{ -0.00000101, -0.00000146, -0.00000114, -0.00000176, -0.00000102, 0.00000204, 0.00000380, 0.00000239, 0.00000065, -0.00000062, -0.00000412, -0.00000796, -0.00000562, 0.00000239, 0.00000698, 0.00000404, -0.00000072, -0.00000299, -0.00000453, -0.00000463, 0.00000015, 0.00000603, 0.00000524, 0.00000001, -0.00000046, 0.00000365, 0.00000377, -0.00000177, -0.00000740, -0.00001049, -0.00000935, -0.00000111, 0.00000726, 0.00000315, -0.00000639, -0.00000099, 0.00001149, 0.00000139, -0.00002227, -0.00001582, 0.00001572, 0.00001546, -0.00001791, -0.00001624, 0.00002547, 0.00002362, -0.00003167, -0.00003661, 0.00002677, 0.00003114, -0.00004976, -0.00005309, 0.00006417, 0.00008761, -0.00006865, -0.00013720, 0.00003720, 0.00013950, -0.00007880, -0.00024914, 0.00002260, 0.00030982, -0.00006409, -0.00072869, -0.00054905, 0.00059808, 0.00135983, 0.00075020, -0.00042466, -0.00075384, -0.00016110, 0.00028225, 0.00007494, -0.00021732, -0.00011113, 0.00010581, 0.00005409, -0.00010585, -0.00007421, 0.00006141, 0.00006223, -0.00003427, -0.00004638, 0.00001658, 0.00002376, -0.00002244, -0.00002523, 0.00001354, 0.00001832, -0.00001075, -0.00001442, 0.00000881, 0.00001243, -0.00000694, -0.00001365, -0.00000204, 0.00000352, -0.00000211, -0.00000361, 0.00000202, 0.00000492, 0.00000204, -0.00000331, -0.00000930, -0.00001167, -0.00000551, 0.00000356, 0.00000492, 0.00000034, 0.00000042, 0.00000532, 0.00000544, -0.00000052, -0.00000454, -0.00000406, -0.00000314, -0.00000169, 0.00000302, 0.00000696, 0.00000391, -0.00000337, -0.00000688, -0.00000516, -0.00000242, 0.00000004, 0.00000302, 0.00000429, 0.00000194, },
		{ -0.00000114, -0.00000161, -0.00000103, 0.00000198, 0.00000405, 0.00000275, 0.00000040, -0.00000152, -0.00000475, -0.00000754, -0.00000449, 0.00000329, 0.00000709, 0.00000352, -0.00000128, -0.00000308, -0.00000426, -0.00000460, -0.00000022, 0.00000577, 0.00000532, 0.00000013, -0.00000017, 0.00000438, 0.00000392, -0.00000359, -0.00000982, -0.00001015, -0.00000621, 0.00000070, 0.00000607, 0.00000246, -0.00000486, -0.00000132, 0.00000737, -0.00000075, -0.00001807, -0.00001104, 0.00001429, 0.00001195, -0.00001628, -0.00001322, 0.00002205, 0.00001818, -0.00002879, -0.00002912, 0.00002579, 0.00002346, -0.00004887, -0.00004337, 0.00006441, 0.00007457, -0.00007294, -0.00012216, 0.00004731, 0.00012401, -0.00009670, -0.00023570, 0.00005022, 0.00030031, -0.00011231, -0.00074732, -0.00049198, 0.00067753, 0.00136840, 0.00067753, -0.00049198, -0.00074732, -0.00011231, 0.00030031, 0.00005022, -0.00023570, -0.00009670, 0.00012401, 0.00004731, -0.00012216, -0.00007294, 0.00007457, 0.00006441, -0.00004337, -0.00004887, 0.00002346, 0.00002579, -0.00002912, -0.00002879, 0.00001818, 0.00002205, -0.00001322, -0.00001628, 0.00001195, 0.00001429, -0.00001104, -0.00001807, -0.00000075, 0.00000737, -0.00000132, -0.00000486, 0.00000246, 0.00000607, 0.00000070, -0.00000621, -0.00001015, -0.00000982, -0.00000359, 0.00000392, 0.00000438, -0.00000017, 0.00000013, 0.00000532, 0.00000577, -0.00000022, -0.00000460, -0.00000426, -0.00000308, -0.00000128, 0.00000352, 0.00000709, 0.00000329, -0.00000449, -0.00000754, -0.00000475, -0.00000152, 0.00000040, 0.00000275, 0.00000405, 0.00000198, -0.00000103, -0.00000161, },
		{ -0.00000101, 0.00000194, 0.00000429, 0.00000302, 0.00000004, -0.00000242, -0.00000516, -0.00000688, -0.00000337, 0.00000391, 0.00000696, 0.00000302, -0.00000169, -0.00000314, -0.00000406, -0.00000454, -0.00000052, 0.00000544, 0.00000532, 0.00000042, 0.00000034, 0.00000492, 0.00000356, -0.00000551, -0.00001167, -0.00000930, -0.00000331, 0.00000204, 0.00000492, 0.00000202, -0.00000361, -0.00000211, 0.00000352, -0.00000204, -0.00001365, -0.00000694, 0.00001243, 0.00000881, -0.00001442, -0.00001075, 0.00001832, 0.00001354, -0.00002523, -0.00002244, 0.00002376, 0.00001658, -0.00004638, -0.00003427, 0.00006223, 0.00006141, -0.00007421, -0.00010585, 0.00005409, 0.00010581, -0.00011113, -0.00021732, 0.00007494, 0.00028225, -0.00016110, -0.00075384, -0.00042466, 0.00075020, 0.00135983, 0.00059808, -0.00054905, -0.00072869, -0.00006409, 0.00030982, 0.00002260, -0.00024914, -0.00007880, 0.00013950, 0.00003720, -0.00013720, -0.00006865, 0.00008761, 0.00006417, -0.00005309, -0.00004976, 0.00003114, 0.00002677, -0.00003661, -0.00003167, 0.00002362, 0.00002547, -0.00001624, -0.00001791, 0.00001546, 0.00001572, -0.00001582, -0.00002227, 0.00000139, 0.00001149, -0.00000099, -0.00000639, 0.00000315, 0.00000726, -0.00000111, -0.00000935, -0.00001049, -0.00000740, -0.00000177, 0.00000377, 0.00000365, -0.00000046, 0.00000001, 0.00000524, 0.00000603, 0.00000015, -0.00000463, -0.00000453, -0.00000299, -0.00000072, 0.00000404, 0.00000698, 0.00000239, -0.00000562, -0.00000796, -0.00000412, -0.00000062, 0.00000065, 0.00000239, 0.00000380, 0.00000204, -0.00000102, -0.00000176, -0.00000114, -0.00000146, },
		{ 0.00000451, 0.00000318, -0.00000042, -0.00000325, -0.00000535, -0.00000603, -0.00000232, 0.00000427, 0.00000666, 0.00000259, -0.00000196, -0.00000319, -0.00000391, -0.00000445, -0.00000076, 0.00000505, 0.00000528, 0.00000090, 0.00000100, 0.00000519, 0.00000273, -0.00000740, -0.00001290, -0.00000808, -0.00000071, 0.00000299, 0.00000387, 0.00000176, -0.00000269, -0.00000324, 0.00000010, -0.00000256, -0.00000921, -0.00000356, 0.00001026, 0.00000610, -0.00001242, -0.00000882, 0.00001449, 0.00000974, -0.00002124, -0.00001671, 0.00002095, 0.00001074, -0.00004258, -0.00002620, 0.00005789, 0.00004874, -0.00007251, -0.00008907, 0.00005729, 0.00008582, -0.00012139, -0.00019492, 0.00009539, 0.00025616, -0.00020832, -0.00074743, -0.00034918, 0.00081354, 0.00133445, 0.00051461, -0.00059425, -0.00069923, -0.00001847, 0.00031072, -0.00000647, -0.00025704, -0.00005831, 0.00015151, 0.00002421, -0.00015019, -0.00006145, 0.00009986, 0.00006143, -0.00006293, -0.00004884, 0.00003929, 0.00002646, -0.00004471, -0.00003365, 0.00002976, 0.00002842, -0.00001976, -0.00001918, 0.00001928, 0.00001658, -0.00002120, -0.00002605, 0.00000440, 0.00001572, -0.00000122, -0.00000810, 0.00000414, 0.00000843, -0.00000340, -0.00001261, -0.00001024, -0.00000446, -0.00000018, 0.00000305, 0.00000282, -0.00000044, 0.00000002, 0.00000502, 0.00000622, 0.00000060, -0.00000464, -0.00000487, -0.00000284, 0.00000001, 0.00000453, 0.00000657, 0.00000126, -0.00000665, -0.00000808, -0.00000332, 0.00000023, 0.00000076, 0.00000197, 0.00000356, 0.00000213, -0.00000100, -0.00000191, -0.00000114, -0.00000131, -0.00000096, 0.00000195, },
		{ -0.00000094, -0.00000396, -0.00000530, -0.00000507, -0.00000140, 0.00000439, 0.00000625, 0.00000224, -0.00000213, -0.00000326, -0.00000381, -0.00000432, -0.00000097, 0.00000462, 0.00000526, 0.00000154, 0.00000172, 0.00000513, 0.00000151, -0.00000912, -0.00001352, -0.00000659, 0.00000153, 0.00000360, 0.00000296, 0.00000160, -0.00000210, -0.00000460, -0.00000279, -0.00000239, -0.00000494, -0.00000091, 0.00000793, 0.00000384, -0.00001040, -0.00000741, 0.00001070, 0.00000676, -0.00001708, -0.00001198, 0.00001767, 0.00000608, -0.00003784, -0.00001948, 0.00005174, 0.00003714, -0.00006808, -0.00007264, 0.00005688, 0.00006504, -0.00012704, -0.00016963, 0.00011046, 0.00022295, -0.00025185, -0.00072783, -0.00026801, 0.00086535, 0.00129320, 0.00042998, -0.00062644, -0.00066062, 0.00002270, 0.00030337, -0.00003546, -0.00025908, -0.00003628, 0.00015945, 0.00000899, -0.00016046, -0.00005167, 0.00011067, 0.00005622, -0.00007239, -0.00004600, 0.00004753, 0.00002469, -0.00005315, -0.00003450, 0.00003645, 0.00003071, -0.00002372, -0.00002001, 0.00002330, 0.00001676, -0.00002704, -0.00002922, 0.00000824, 0.00001986, -0.00000207, -0.00000990, 0.00000547, 0.00000950, -0.00000621, -0.00001587, -0.00000931, -0.00000109, 0.00000102, 0.00000180, 0.00000200, -0.00000007, 0.00000011, 0.00000463, 0.00000637, 0.00000116, -0.00000466, -0.00000528, -0.00000259, 0.00000092, 0.00000494, 0.00000581, -0.00000008, -0.00000750, -0.00000789, -0.00000242, 0.00000097, 0.00000075, 0.00000153, 0.00000334, 0.00000224, -0.00000098, -0.00000204, -0.00000115, -0.00000115, -0.00000088, 0.00000198, 0.00000468, 0.00000321, },
		{ -0.00000504, -0.00000408, -0.00000066, 0.00000429, 0.00000580, 0.00000199, -0.00000224, -0.00000337, -0.00000374, -0.00000416, -0.00000115, 0.00000416, 0.00000529, 0.00000232, 0.00000241, 0.00000471, 0.00000001, -0.00001058, -0.00001357, -0.00000496, 0.00000339, 0.00000393, 0.00000219, 0.00000149, -0.00000186, -0.00000605, -0.00000508, -0.00000167, -0.00000100, 0.00000104, 0.00000557, 0.00000202, -0.00000846, -0.00000648, 0.00000711, 0.00000455, -0.00001295, -0.00000823, 0.00001422, 0.00000267, -0.00003254, -0.00001431, 0.00004423, 0.00002706, -0.00006130, -0.00005729, 0.00005306, 0.00004450, -0.00012793, -0.00014271, 0.00011930, 0.00018392, -0.00028969, -0.00069534, -0.00018381, 0.00090391, 0.00123764, 0.00034698, -0.00064511, -0.00061485, 0.00005792, 0.00028850, -0.00006294, -0.00025527, -0.00001382, 0.00016296, -0.00000768, -0.00016746, -0.00003976, 0.00011947, 0.00004873, -0.00008093, -0.00004128, 0.00005543, 0.00002138, -0.00006161, -0.00003408, 0.00004350, 0.00003218, -0.00002799, -0.00002030, 0.00002739, 0.00001618, -0.00003319, -0.00003159, 0.00001284, 0.00002372, -0.00000355, -0.00001168, 0.00000714, 0.00001036, -0.00000953, -0.00001899, -0.00000765, 0.00000257, 0.00000172, 0.00000004, 0.00000133, 0.00000067, 0.00000022, 0.00000403, 0.00000650, 0.00000183, -0.00000470, -0.00000574, -0.00000220, 0.00000198, 0.00000518, 0.00000467, -0.00000156, -0.00000809, -0.00000738, -0.00000147, 0.00000155, 0.00000064, 0.00000109, 0.00000315, 0.00000234, -0.00000095, -0.00000217, -0.00000116, -0.00000098, -0.00000075, 0.00000204, 0.00000477, 0.00000312, -0.00000148, -0.00000449, },
		{ -0.00000012, 0.00000404, 0.00000534, 0.00000182, -0.00000232, -0.00000350, -0.00000367, -0.00000395, -0.00000131, 0.00000371, 0.00000539, 0.00000321, 0.00000299, 0.00000395, -0.00000166, -0.00001169, -0.00001312, -0.00000328, 0.00000484, 0.00000406, 0.00000156, 0.00000135, -0.00000193, -0.00000744, -0.00000673, -0.00000054, 0.00000248, 0.00000236, 0.00000330, 0.00000063, -0.00000666, -0.00000595, 0.00000384, 0.00000302, -0.00000906, -0.00000539, 0.00001085, 0.00000047, -0.00002706, -0.00001079, 0.00003587, 0.00001882, -0.00005269, -0.00004364, 0.00004624, 0.00002515, -0.00012417, -0.00011554, 0.00012140, 0.00014067, -0.00032012, -0.00065079, -0.00009939, 0.00092805, 0.00116978, 0.00026822, -0.00065029, -0.00056409, 0.00008602, 0.00026719, -0.00008761, -0.00024594, 0.00000795, 0.00016193, -0.00002493, -0.00017085, -0.00002632, 0.00012576, 0.00003931, -0.00008808, -0.00003479, 0.00006259, 0.00001653, -0.00006972, -0.00003229, 0.00005066, 0.00003271, -0.00003246, -0.00001998, 0.00003141, 0.00001478, -0.00003946, -0.00003302, 0.00001807, 0.00002711, -0.00000564, -0.00001331, 0.00000916, 0.00001092, -0.00001331, -0.00002182, -0.00000523, 0.00000638, 0.00000180, -0.00000213, 0.00000092, 0.00000179, 0.00000026, 0.00000320, 0.00000664, 0.00000262, -0.00000478, -0.00000625, -0.00000165, 0.00000317, 0.00000518, 0.00000313, -0.00000308, -0.00000833, -0.00000658, -0.00000056, 0.00000195, 0.00000046, 0.00000070, 0.00000298, 0.00000242, -0.00000091, -0.00000227, -0.00000116, -0.00000081, -0.00000058, 0.00000211, 0.00000477, 0.00000291, -0.00000198, -0.00000483, -0.00000460, -0.00000311, },
		{ 0.00000493, 0.00000171, -0.00000240, -0.00000367, -0.00000360, -0.00000371, -0.00000145, 0.00000329, 0.00000558, 0.00000414, 0.00000338, 0.00000285, -0.00000340, -0.00001241, -0.00001223, -0.00000164, 0.00000592, 0.00000404, 0.00000104, 0.00000114, -0.00000224, -0.00000867, -0.00000776, 0.00000084, 0.00000542, 0.00000313, 0.00000122, -0.00000039, -0.00000508, -0.00000575, 0.00000095, 0.00000205, -0.00000554, -0.00000334, 0.00000781, -0.00000063, -0.00002179, -0.00000889, 0.00002717, 0.00001262, -0.00004286, -0.00003217, 0.00003698, 0.00000785, -0.00011620, -0.00008945, 0.00011665, 0.00009502, -0.00034175, -0.00059556, -0.00001748, 0.00093718, 0.00109204, 0.00019599, -0.00064259, -0.00051055, 0.00010628, 0.00024077, -0.00010837, -0.00023169, 0.00002798, 0.00015652, -0.00004187, -0.00017044, -0.00001205, 0.00012920, 0.00002841, -0.00009342, -0.00002680, 0.00006857, 0.00001024, -0.00007713, -0.00002910, 0.00005765, 0.00003222, -0.00003695, -0.00001903, 0.00003521, 0.00001255, -0.00004562, -0.00003338, 0.00002377, 0.00002987, -0.00000830, -0.00001468, 0.00001146, 0.00001109, -0.00001748, -0.00002421, -0.00000207, 0.00001015, 0.00000119, -0.00000463, 0.00000088, 0.00000325, 0.00000017, 0.00000212, 0.00000683, 0.00000353, -0.00000492, -0.00000679, -0.00000091, 0.00000445, 0.00000487, 0.00000121, -0.00000454, -0.00000815, -0.00000553, 0.00000024, 0.00000215, 0.00000025, 0.00000037, 0.00000281, 0.00000247, -0.00000087, -0.00000234, -0.00000117, -0.00000062, -0.00000036, 0.00000217, 0.00000465, 0.00000259, -0.00000241, -0.00000495, -0.00000404, -0.00000225, 0.00000022, 0.00000370, },
		{ -0.00000251, -0.00000385, -0.00000351, -0.00000344, -0.00000157, 0.00000293, 0.00000586, 0.00000505, 0.00000354, 0.00000149, -0.00000509, -0.00001271, -0.00001099, -0.00000013, 0.00000665, 0.00000390, 0.00000062, 0.00000084, -0.00000272, -0.00000964, -0.00000820, 0.00000230, 0.00000776, 0.00000347, -0.00000058, -0.00000109, -0.00000375, -0.00000579, -0.00000149, 0.00000151, -0.00000249, -0.00000192, 0.00000526, -0.00000083, -0.00001703, -0.00000846, 0.00001863, 0.00000846, -0.00003245, -0.00002318, 0.00002602, -0.00000671, -0.00010465, -0.00006571, 0.00010530, 0.00004889, -0.00035360, -0.00053145, 0.00005934, 0.00093135, 0.00100709, 0.00013214, -0.00062308, -0.00045641, 0.00011840, 0.00021071, -0.00012442, -0.00021337, 0.00004537, 0.00014711, -0.00005762, -0.00016629, 0.00000233, 0.00012958, 0.00001656, -0.00009663, -0.00001766, 0.00007304, 0.00000272, -0.00008347, -0.00002456, 0.00006418, 0.00003066, -0.00004129, -0.00001744, 0.00003862, 0.00000952, -0.00005146, -0.00003263, 0.00002974, 0.00003185, -0.00001142, -0.00001567, 0.00001398, 0.00001077, -0.00002196, -0.00002602, 0.00000177, 0.00001372, -0.00000014, -0.00000730, 0.00000131, 0.00000500, -0.00000016, 0.00000082, 0.00000713, 0.00000456, -0.00000514, -0.00000731, 0.00000005, 0.00000575, 0.00000418, -0.00000106, -0.00000585, -0.00000753, -0.00000429, 0.00000084, 0.00000215, 0.00000007, 0.00000012, 0.00000261, 0.00000247, -0.00000080, -0.00000238, -0.00000118, -0.00000042, -0.00000010, 0.00000220, 0.00000440, 0.00000221, -0.00000272, -0.00000486, -0.00000342, -0.00000154, 0.00000039, 0.00000331, 0.00000458, 0.00000164, },
		{ -0.00000338, -0.00000314, -0.00000165, 0.00000265, 0.00000621, 0.00000587, 0.00000344, -0.00000008, -0.00000663, -0.00001260, -0.00000951, 0.00000121, 0.00000707, 0.00000369, 0.00000025, 0.00000044, -0.00000328, -0.00001027, -0.00000814, 0.00000371, 0.00000948, 0.00000349, -0.00000206, -0.00000153, -0.00000271, -0.00000598, -0.00000348, 0.00000128, 0.00000004, -0.00000095, 0.00000331, -0.00000035, -0.00001303, -0.00000926, 0.00001069, 0.00000624, -0.00002208, -0.00001679, 0.00001411, -0.00001806, -0.00009041, -0.00004539, 0.00008801, 0.00000423, -0.00035519, -0.00046062, 0.00012881, 0.00091119, 0.00091774, 0.00007804, -0.00059331, -0.00040365, 0.00012253, 0.00017857, -0.00013525, -0.00019199, 0.00005937, 0.00013428, -0.00007141, -0.00015864, 0.00001608, 0.00012688, 0.00000437, -0.00009750, -0.00000781, 0.00007570, -0.00000573, -0.00008841, -0.00001882, 0.00006997, 0.00002807, -0.00004530, -0.00001524, 0.00004151, 0.00000576, -0.00005674, -0.00003075, 0.00003577, 0.00003293, -0.00001489, -0.00001619, 0.00001664, 0.00000992, -0.00002660, -0.00002713, 0.00000620, 0.00001689, -0.00000220, -0.00000999, 0.00000228, 0.00000694, -0.00000078, -0.00000067, 0.00000756, 0.00000569, -0.00000544, -0.00000780, 0.00000121, 0.00000701, 0.00000306, -0.00000360, -0.00000688, -0.00000643, -0.00000296, 0.00000118, 0.00000198, -0.00000004, -0.00000005, 0.00000237, 0.00000241, -0.00000069, -0.00000237, -0.00000119, -0.00000021, 0.00000019, 0.00000220, 0.00000404, 0.00000179, -0.00000286, -0.00000457, -0.00000280, -0.00000101, 0.00000040, 0.00000293, 0.00000429, 0.00000156, -0.00000265, -0.00000403, },
		{ -0.00000169, 0.00000248, 0.00000661, 0.00000654, 0.00000305, -0.00000178, -0.00000795, -0.00001210, -0.00000787, 0.00000235, 0.00000722, 0.00000343, -0.00000009, -0.00000006, -0.00000385, -0.00001053, -0.00000767, 0.00000495, 0.00001061, 0.00000330, -0.00000318, -0.00000178, -0.00000194, -0.00000625, -0.00000502, 0.00000123, 0.00000204, -0.00000027, 0.00000199, 0.00000056, -0.00000993, -0.00001101, 0.00000373, 0.00000572, -0.00001235, -0.00001293, 0.00000208, -0.00002597, -0.00007448, -0.00002933, 0.00006573, -0.00003712, -0.00034654, -0.00038545, 0.00018907, 0.00087791, 0.00082675, 0.00003455, -0.00055509, -0.00035399, 0.00011922, 0.00014588, -0.00014071, -0.00016867, 0.00006950, 0.00011878, -0.00008258, -0.00014792, 0.00002852, 0.00012124, -0.00000757, -0.00009600, 0.00000227, 0.00007637, -0.00001476, -0.00009168, -0.00001207, 0.00007475, 0.00002451, -0.00004880, -0.00001250, 0.00004372, 0.00000138, -0.00006127, -0.00002777, 0.00004161, 0.00003306, -0.00001855, -0.00001619, 0.00001931, 0.00000848, -0.00003126, -0.00002744, 0.00001107, 0.00001950, -0.00000494, -0.00001253, 0.00000381, 0.00000898, -0.00000177, -0.00000230, 0.00000819, 0.00000690, -0.00000584, -0.00000822, 0.00000257, 0.00000813, 0.00000148, -0.00000634, -0.00000753, -0.00000488, -0.00000164, 0.00000121, 0.00000169, -0.00000003, -0.00000014, 0.00000205, 0.00000231, -0.00000053, -0.00000232, -0.00000122, 0.00000001, 0.00000052, 0.00000214, 0.00000355, 0.00000137, -0.00000281, -0.00000411, -0.00000225, -0.00000069, 0.00000032, 0.00000261, 0.00000406, 0.00000146, -0.00000284, -0.00000420, -0.00000321, -0.00000284, },
		{ 0.00000702, 0.00000701, 0.00000240, -0.00000350, -0.00000899, -0.00001125, -0.00000617, 0.00000326, 0.00000714, 0.00000311, -0.00000043, -0.00000061, -0.00000432, -0.00001041, -0.00000693, 0.00000592, 0.00001119, 0.00000301, -0.00000394, -0.00000191, -0.00000145, -0.00000655, -0.00000614, 0.00000128, 0.00000357, 0.00000027, 0.00000129, 0.00000163, -0.00000782, -0.00001335, -0.00000201, 0.00000654, -0.00000375, -0.00001134, -0.00000932, -0.00003043, -0.00005792, -0.00001809, 0.00003972, -0.00007354, -0.00032814, -0.00030846, 0.00023872, 0.00083316, 0.00073675, 0.00000194, -0.00051046, -0.00030881, 0.00010932, 0.00011405, -0.00014091, -0.00014456, 0.00007550, 0.00010147, -0.00009067, -0.00013469, 0.00003907, 0.00011295, -0.00001867, -0.00009218, 0.00001207, 0.00007499, -0.00002395, -0.00009310, -0.00000461, 0.00007827, 0.00002012, -0.00005163, -0.00000931, 0.00004515, -0.00000348, -0.00006484, -0.00002378, 0.00004705, 0.00003220, -0.00002226, -0.00001563, 0.00002188, 0.00000647, -0.00003577, -0.00002689, 0.00001622, 0.00002140, -0.00000828, -0.00001477, 0.00000592, 0.00001098, -0.00000316, -0.00000400, 0.00000904, 0.00000815, -0.00000636, -0.00000854, 0.00000411, 0.00000906, -0.00000057, -0.00000916, -0.00000772, -0.00000292, -0.00000043, 0.00000091, 0.00000134, 0.00000015, -0.00000019, 0.00000162, 0.00000217, -0.00000030, -0.00000221, -0.00000126, 0.00000024, 0.00000086, 0.00000201, 0.00000296, 0.00000100, -0.00000256, -0.00000354, -0.00000181, -0.00000056, 0.00000019, 0.00000237, 0.00000387, 0.00000129, -0.00000306, -0.00000432, -0.00000301, -0.00000252, -0.00000167, 0.00000242, },
		{ 0.00000151, -0.00000518, -0.00000971, -0.00001013, -0.00000448, 0.00000395, 0.00000687, 0.00000276, -0.00000078, -0.00000119, -0.00000464, -0.00000994, -0.00000603, 0.00000655, 0.00001131, 0.00000272, -0.00000436, -0.00000197, -0.00000120, -0.00000681, -0.00000690, 0.00000135, 0.00000469, 0.00000081, 0.00000112, 0.00000265, -0.00000667, -0.00001593, -0.00000638, 0.00000829, 0.00000334, -0.00001166, -0.00001941, -0.00003168, -0.00004182, -0.00001190, 0.00001141, -0.00010368, -0.00030096, -0.00023212, 0.00027693, 0.00077899, 0.00065008, -0.00002000, -0.00046155, -0.00026909, 0.00009399, 0.00008429, -0.00013630, -0.00012075, 0.00007737, 0.00008327, -0.00009539, -0.00011966, 0.00004727, 0.00010246, -0.00002842, -0.00008627, 0.00002110, 0.00007160, -0.00003285, -0.00009256, 0.00000326, 0.00008037, 0.00001507, -0.00005364, -0.00000582, 0.00004571, -0.00000864, -0.00006731, -0.00001893, 0.00005186, 0.00003040, -0.00002583, -0.00001451, 0.00002421, 0.00000389, -0.00003996, -0.00002546, 0.00002146, 0.00002247, -0.00001209, -0.00001653, 0.00000856, 0.00001281, -0.00000496, -0.00000570, 0.00001015, 0.00000941, -0.00000698, -0.00000873, 0.00000579, 0.00000970, -0.00000308, -0.00001195, -0.00000737, -0.00000061, 0.00000056, 0.00000027, 0.00000099, 0.00000051, -0.00000022, 0.00000106, 0.00000200, 0.00000004, -0.00000207, -0.00000134, 0.00000047, 0.00000122, 0.00000181, 0.00000228, 0.00000072, -0.00000210, -0.00000291, -0.00000155, -0.00000061, 0.00000008, 0.00000224, 0.00000367, 0.00000103, -0.00000331, -0.00000438, -0.00000276, -0.00000220, -0.00000159, 0.00000247, 0.00000739, 0.00000722, },
		{ -0.00001008, -0.00000880, -0.00000285, 0.00000442, 0.00000646, 0.00000236, -0.00000114, -0.00000172, -0.00000475, -0.00000918, -0.00000509, 0.00000681, 0.00001106, 0.00000250, -0.00000448, -0.00000202, -0.00000118, -0.00000701, -0.00000735, 0.00000140, 0.00000549, 0.00000143, 0.00000135, 0.00000341, -0.00000637, -0.00001842, -0.00000936, 0.00001055, 0.00000870, -0.00001337, -0.00002766, -0.00003016, -0.00002717, -0.00001066, -0.00001767, -0.00012659, -0.00026636, -0.00015877, 0.00030341, 0.00071767, 0.00056873, -0.00003199, -0.00041045, -0.00023543, 0.00007454, 0.00005757, -0.00012755, -0.00009823, 0.00007535, 0.00006507, -0.00009667, -0.00010356, 0.00005283, 0.00009030, -0.00003640, -0.00007861, 0.00002891, 0.00006636, -0.00004103, -0.00009006, 0.00001116, 0.00008091, 0.00000957, -0.00005475, -0.00000216, 0.00004535, -0.00001389, -0.00006855, -0.00001339, 0.00005584, 0.00002772, -0.00002912, -0.00001287, 0.00002619, 0.00000082, -0.00004367, -0.00002317, 0.00002658, 0.00002265, -0.00001621, -0.00001769, 0.00001165, 0.00001435, -0.00000718, -0.00000730, 0.00001152, 0.00001064, -0.00000770, -0.00000877, 0.00000756, 0.00000999, -0.00000600, -0.00001457, -0.00000643, 0.00000194, 0.00000124, -0.00000069, 0.00000071, 0.00000108, -0.00000029, 0.00000035, 0.00000184, 0.00000049, -0.00000190, -0.00000146, 0.00000070, 0.00000159, 0.00000152, 0.00000154, 0.00000056, -0.00000146, -0.00000228, -0.00000149, -0.00000080, 0.00000003, 0.00000221, 0.00000343, 0.00000066, -0.00000356, -0.00000435, -0.00000247, -0.00000187, -0.00000142, 0.00000262, 0.00000768, 0.00000715, 0.00000044, -0.00000671, },
		{ -0.00000135, 0.00000469, 0.00000592, 0.00000192, -0.00000151, -0.00000216, -0.00000462, -0.00000819, -0.00000421, 0.00000672, 0.00001054, 0.00000240, -0.00000433, -0.00000208, -0.00000133, -0.00000712, -0.00000755, 0.00000139, 0.00000605, 0.00000217, 0.00000184, 0.00000378, -0.00000676, -0.00002053, -0.00001103, 0.00001287, 0.00001223, -0.00001595, -0.00003371, -0.00002650, -0.00001482, -0.00001396, -0.00004596, -0.00014169, -0.00022600, -0.00009049, 0.00031839, 0.00065162, 0.00049426, -0.00003511, -0.00035909, -0.00020800, 0.00005237, 0.00003459, -0.00011550, -0.00007782, 0.00006990, 0.00004770, -0.00009465, -0.00008715, 0.00005560, 0.00007707, -0.00004233, -0.00006961, 0.00003514, 0.00005955, -0.00004809, -0.00008570, 0.00001873, 0.00007986, 0.00000387, -0.00005489, 0.00000150, 0.00004407, -0.00001903, -0.00006849, -0.00000737, 0.00005884, 0.00002427, -0.00003198, -0.00001078, 0.00002770, -0.00000264, -0.00004673, -0.00002010, 0.00003138, 0.00002190, -0.00002045, -0.00001814, 0.00001510, 0.00001546, -0.00000976, -0.00000871, 0.00001315, 0.00001178, -0.00000852, -0.00000864, 0.00000935, 0.00000986, -0.00000927, -0.00001690, -0.00000490, 0.00000461, 0.00000153, -0.00000192, 0.00000059, 0.00000184, -0.00000044, -0.00000052, 0.00000172, 0.00000108, -0.00000171, -0.00000164, 0.00000094, 0.00000195, 0.00000116, 0.00000078, 0.00000054, -0.00000065, -0.00000171, -0.00000164, -0.00000108, 0.00000009, 0.00000227, 0.00000312, 0.00000020, -0.00000377, -0.00000423, -0.00000216, -0.00000154, -0.00000115, 0.00000284, 0.00000783, 0.00000680, -0.00000076, -0.00000803, -0.00001013, -0.00000733, },
		{ 0.00000529, 0.00000143, -0.00000186, -0.00000246, -0.00000426, -0.00000707, -0.00000348, 0.00000629, 0.00000985, 0.00000245, -0.00000399, -0.00000220, -0.00000163, -0.00000714, -0.00000754, 0.00000135, 0.00000647, 0.00000303, 0.00000242, 0.00000369, -0.00000763, -0.00002203, -0.00001157, 0.00001486, 0.00001401, -0.00001881, -0.00003738, -0.00002140, -0.00000543, -0.00002112, -0.00007201, -0.00014884, -0.00018174, -0.00002902, 0.00032261, 0.00058323, 0.00042772, -0.00003076, -0.00030920, -0.00018661, 0.00002887, 0.00001573, -0.00010111, -0.00006014, 0.00006162, 0.00003187, -0.00008963, -0.00007117, 0.00005564, 0.00006341, -0.00004606, -0.00005978, 0.00003953, 0.00005153, -0.00005367, -0.00007967, 0.00002561, 0.00007724, -0.00000179, -0.00005405, 0.00000499, 0.00004189, -0.00002384, -0.00006713, -0.00000112, 0.00006074, 0.00002020, -0.00003429, -0.00000832, 0.00002867, -0.00000638, -0.00004901, -0.00001634, 0.00003564, 0.00002024, -0.00002461, -0.00001781, 0.00001875, 0.00001606, -0.00001265, -0.00000984, 0.00001503, 0.00001280, -0.00000942, -0.00000833, 0.00001111, 0.00000928, -0.00001279, -0.00001882, -0.00000278, 0.00000729, 0.00000136, -0.00000335, 0.00000070, 0.00000277, -0.00000074, -0.00000154, 0.00000169, 0.00000181, -0.00000153, -0.00000188, 0.00000118, 0.00000231, 0.00000072, 0.00000002, 0.00000068, 0.00000027, -0.00000125, -0.00000201, -0.00000140, 0.00000029, 0.00000238, 0.00000269, -0.00000037, -0.00000392, -0.00000400, -0.00000184, -0.00000120, -0.00000079, 0.00000313, 0.00000779, 0.00000615, -0.00000202, -0.00000909, -0.00000986, -0.00000580, -0.00000001, 0.00000480, },
		{ -0.00000216, -0.00000256, -0.00000369, -0.00000592, -0.00000297, 0.00000559, 0.00000910, 0.00000266, -0.00000350, -0.00000239, -0.00000203, -0.00000707, -0.00000738, 0.00000130, 0.00000683, 0.00000398, 0.00000294, 0.00000311, -0.00000875, -0.00002279, -0.00001123, 0.00001621, 0.00001421, -0.00002145, -0.00003869, -0.00001564, 0.00000058, -0.00003123, -0.00009455, -0.00014833, -0.00013555, 0.00002435, 0.00031722, 0.00051478, 0.00036971, -0.00002053, -0.00026220, -0.00017074, 0.00000535, 0.00000107, -0.00008539, -0.00004557, 0.00005126, 0.00001815, -0.00008210, -0.00005624, 0.00005316, 0.00004993, -0.00004758, -0.00004963, 0.00004192, 0.00004273, -0.00005750, -0.00007225, 0.00003148, 0.00007315, -0.00000715, -0.00005226, 0.00000815, 0.00003889, -0.00002812, -0.00006450, 0.00000513, 0.00006146, 0.00001567, -0.00003596, -0.00000563, 0.00002904, -0.00001025, -0.00005042, -0.00001205, 0.00003920, 0.00001774, -0.00002850, -0.00001668, 0.00002245, 0.00001604, -0.00001574, -0.00001059, 0.00001712, 0.00001363, -0.00001037, -0.00000785, 0.00001275, 0.00000823, -0.00001645, -0.00002020, -0.00000014, 0.00000983, 0.00000070, -0.00000490, 0.00000108, 0.00000383, -0.00000122, -0.00000270, 0.00000178, 0.00000267, -0.00000139, -0.00000220, 0.00000143, 0.00000267, 0.00000022, -0.00000071, 0.00000100, 0.00000128, -0.00000096, -0.00000257, -0.00000168, 0.00000064, 0.00000251, 0.00000211, -0.00000100, -0.00000397, -0.00000366, -0.00000152, -0.00000085, -0.00000033, 0.00000342, 0.00000754, 0.00000524, -0.00000325, -0.00000983, -0.00000933, -0.00000428, 0.00000116, 0.00000475, 0.00000458, 0.00000090, },
		{ -0.00000295, -0.00000484, -0.00000270, 0.00000470, 0.00000835, 0.00000302, -0.00000294, -0.00000264, -0.00000251, -0.00000692, -0.00000709, 0.00000127, 0.00000717, 0.00000494, 0.00000328, 0.00000210, -0.00000989, -0.00002274, -0.00001030, 0.00001670, 0.00001314, -0.00002341, -0.00003783, -0.00001000, 0.00000308, -0.00004325, -0.00011254, -0.00014082, -0.00008936, 0.00006875, 0.00030369, 0.00044831, 0.00032034, -0.00000610, -0.00021917, -0.00015966, -0.00001706, -0.00000954, -0.00006931, -0.00003427, 0.00003960, 0.00000690, -0.00007264, -0.00004290, 0.00004848, 0.00003718, -0.00004702, -0.00003968, 0.00004231, 0.00003362, -0.00005942, -0.00006380, 0.00003607, 0.00006777, -0.00001199, -0.00004962, 0.00001082, 0.00003520, -0.00003167, -0.00006069, 0.00001113, 0.00006099, 0.00001088, -0.00003693, -0.00000281, 0.00002877, -0.00001409, -0.00005089, -0.00000741, 0.00004189, 0.00001452, -0.00003192, -0.00001477, 0.00002603, 0.00001538, -0.00001893, -0.00001091, 0.00001937, 0.00001425, -0.00001136, -0.00000720, 0.00001422, 0.00000672, -0.00002012, -0.00002098, 0.00000294, 0.00001209, -0.00000044, -0.00000645, 0.00000179, 0.00000496, -0.00000193, -0.00000394, 0.00000204, 0.00000366, -0.00000132, -0.00000260, 0.00000170, 0.00000302, -0.00000034, -0.00000138, 0.00000147, 0.00000231, -0.00000087, -0.00000329, -0.00000188, 0.00000115, 0.00000260, 0.00000136, -0.00000167, -0.00000389, -0.00000323, -0.00000120, -0.00000048, 0.00000023, 0.00000370, 0.00000705, 0.00000411, -0.00000440, -0.00001024, -0.00000859, -0.00000283, 0.00000215, 0.00000458, 0.00000379, 0.00000034, -0.00000236, -0.00000244, },
		{ -0.00000268, 0.00000371, 0.00000768, 0.00000350, -0.00000234, -0.00000296, -0.00000302, -0.00000671, -0.00000668, 0.00000130, 0.00000752, 0.00000583, 0.00000332, 0.00000077, -0.00001084, -0.00002192, -0.00000909, 0.00001622, 0.00001117, -0.00002436, -0.00003517, -0.00000516, 0.00000214, -0.00005603, -0.00012525, -0.00012729, -0.00004499, 0.00010382, 0.00028371, 0.00038554, 0.00027933, 0.00001086, -0.00018081, -0.00015241, -0.00003742, -0.00001650, -0.00005376, -0.00002618, 0.00002743, -0.00000169, -0.00006188, -0.00003151, 0.00004205, 0.00002562, -0.00004464, -0.00003041, 0.00004080, 0.00002465, -0.00005937, -0.00005472, 0.00003918, 0.00006134, -0.00001611, -0.00004623, 0.00001289, 0.00003095, -0.00003435, -0.00005585, 0.00001666, 0.00005935, 0.00000601, -0.00003718, -0.00000002, 0.00002788, -0.00001774, -0.00005040, -0.00000260, 0.00004362, 0.00001072, -0.00003469, -0.00001214, 0.00002931, 0.00001404, -0.00002209, -0.00001074, 0.00002170, 0.00001460, -0.00001235, -0.00000641, 0.00001543, 0.00000476, -0.00002367, -0.00002110, 0.00000636, 0.00001396, -0.00000205, -0.00000792, 0.00000283, 0.00000608, -0.00000291, -0.00000523, 0.00000251, 0.00000476, -0.00000136, -0.00000308, 0.00000200, 0.00000338, -0.00000093, -0.00000198, 0.00000209, 0.00000331, -0.00000099, -0.00000413, -0.00000195, 0.00000180, 0.00000259, 0.00000043, -0.00000233, -0.00000365, -0.00000272, -0.00000091, -0.00000008, 0.00000086, 0.00000390, 0.00000629, 0.00000280, -0.00000538, -0.00001032, -0.00000769, -0.00000149, 0.00000295, 0.00000430, 0.00000295, -0.00000022, -0.00000242, -0.00000209, -0.00000212, -0.00000392, },
		{ 0.00000712, 0.00000404, -0.00000177, -0.00000333, -0.00000354, -0.00000645, -0.00000617, 0.00000143, 0.00000788, 0.00000654, 0.00000303, -0.00000077, -0.00001143, -0.00002043, -0.00000789, 0.00001477, 0.00000869, -0.00002408, -0.00003117, -0.00000171, -0.00000187, -0.00006846, -0.00013230, -0.00010899, -0.00000403, 0.00012962, 0.00025909, 0.00032782, 0.00024601, 0.00002882, -0.00014749, -0.00014800, -0.00005505, -0.00002040, -0.00003948, -0.00002102, 0.00001549, -0.00000762, -0.00005051, -0.00002229, 0.00003440, 0.00001560, -0.00004080, -0.00002221, 0.00003762, 0.00001629, -0.00005741, -0.00004542, 0.00004071, 0.00005413, -0.00001936, -0.00004226, 0.00001427, 0.00002633, -0.00003603, -0.00005015, 0.00002151, 0.00005663, 0.00000125, -0.00003674, 0.00000263, 0.00002641, -0.00002103, -0.00004898, 0.00000215, 0.00004431, 0.00000653, -0.00003667, -0.00000890, 0.00003212, 0.00001207, -0.00002507, -0.00001004, 0.00002406, 0.00001467, -0.00001332, -0.00000550, 0.00001634, 0.00000242, -0.00002697, -0.00002052, 0.00000997, 0.00001534, -0.00000405, -0.00000918, 0.00000420, 0.00000711, -0.00000416, -0.00000651, 0.00000322, 0.00000592, -0.00000153, -0.00000363, 0.00000235, 0.00000373, -0.00000156, -0.00000249, 0.00000284, 0.00000423, -0.00000134, -0.00000501, -0.00000183, 0.00000255, 0.00000245, -0.00000066, -0.00000294, -0.00000324, -0.00000214, -0.00000066, 0.00000034, 0.00000156, 0.00000400, 0.00000527, 0.00000138, -0.00000615, -0.00001008, -0.00000670, -0.00000030, 0.00000358, 0.00000393, 0.00000207, -0.00000076, -0.00000231, -0.00000153, -0.00000127, -0.00000321, -0.00000287, 0.00000270, },
		{ -0.00000126, -0.00000372, -0.00000407, -0.00000614, -0.00000555, 0.00000169, 0.00000823, 0.00000699, 0.00000240, -0.00000233, -0.00001155, -0.00001844, -0.00000694, 0.00001244, 0.00000610, -0.00002252, -0.00002639, -0.00000005, -0.00000840, -0.00007950, -0.00013364, -0.00008729, 0.00003221, 0.00014665, 0.00023161, 0.00027611, 0.00021949, 0.00004640, -0.00011925, -0.00014540, -0.00006953, -0.00002189, -0.00002703, -0.00001840, 0.00000441, -0.00001106, -0.00003915, -0.00001531, 0.00002607, 0.00000733, -0.00003591, -0.00001539, 0.00003307, 0.00000892, -0.00005369, -0.00003632, 0.00004061, 0.00004645, -0.00002162, -0.00003790, 0.00001489, 0.00002154, -0.00003665, -0.00004380, 0.00002551, 0.00005294, -0.00000324, -0.00003565, 0.00000502, 0.00002445, -0.00002382, -0.00004669, 0.00000665, 0.00004396, 0.00000216, -0.00003777, -0.00000520, 0.00003432, 0.00000952, -0.00002774, -0.00000882, 0.00002634, 0.00001444, -0.00001424, -0.00000452, 0.00001690, -0.00000024, -0.00002988, -0.00001925, 0.00001364, 0.00001613, -0.00000637, -0.00001014, 0.00000586, 0.00000797, -0.00000569, -0.00000770, 0.00000419, 0.00000710, -0.00000187, -0.00000424, 0.00000276, 0.00000409, -0.00000220, -0.00000290, 0.00000368, 0.00000503, -0.00000190, -0.00000590, -0.00000152, 0.00000336, 0.00000211, -0.00000189, -0.00000345, -0.00000264, -0.00000152, -0.00000045, 0.00000081, 0.00000230, 0.00000395, 0.00000400, -0.00000008, -0.00000665, -0.00000955, -0.00000567, 0.00000072, 0.00000404, 0.00000350, 0.00000117, -0.00000123, -0.00000200, -0.00000078, -0.00000049, -0.00000278, -0.00000323, 0.00000176, 0.00000670, 0.00000461, },
		{ -0.00000457, -0.00000580, -0.00000481, 0.00000209, 0.00000851, 0.00000711, 0.00000146, -0.00000378, -0.00001115, -0.00001615, -0.00000642, 0.00000944, 0.00000376, -0.00001977, -0.00002139, -0.00000040, -0.00001674, -0.00008826, -0.00012955, -0.00006366, 0.00006278, 0.00015574, 0.00020297, 0.00023093, 0.00019866, 0.00006250, -0.00009583, -0.00014366, -0.00008069, -0.00002171, -0.00001680, -0.00001779, -0.00000530, -0.00001231, -0.00002837, -0.00001048, 0.00001758, 0.00000089, -0.00003043, -0.00001012, 0.00002755, 0.00000284, -0.00004847, -0.00002780, 0.00003895, 0.00003862, -0.00002288, -0.00003334, 0.00001477, 0.00001677, -0.00003617, -0.00003703, 0.00002855, 0.00004842, -0.00000732, -0.00003400, 0.00000705, 0.00002208, -0.00002600, -0.00004365, 0.00001070, 0.00004260, -0.00000218, -0.00003792, -0.00000121, 0.00003579, 0.00000648, -0.00002997, -0.00000710, 0.00002847, 0.00001390, -0.00001508, -0.00000349, 0.00001708, -0.00000311, -0.00003229, -0.00001735, 0.00001720, 0.00001631, -0.00000890, -0.00001073, 0.00000777, 0.00000858, -0.00000746, -0.00000873, 0.00000542, 0.00000826, -0.00000239, -0.00000487, 0.00000326, 0.00000445, -0.00000285, -0.00000321, 0.00000458, 0.00000567, -0.00000265, -0.00000671, -0.00000099, 0.00000418, 0.00000155, -0.00000324, -0.00000380, -0.00000186, -0.00000090, -0.00000029, 0.00000131, 0.00000304, 0.00000371, 0.00000252, -0.00000150, -0.00000687, -0.00000877, -0.00000468, 0.00000154, 0.00000436, 0.00000301, 0.00000026, -0.00000159, -0.00000146, 0.00000010, 0.00000014, -0.00000263, -0.00000369, 0.00000096, 0.00000641, 0.00000515, -0.00000082, -0.00000413, },
		{ -0.00000393, 0.00000264, 0.00000867, 0.00000682, 0.00000029, -0.00000495, -0.00001026, -0.00001379, -0.00000645, 0.00000602, 0.00000196, -0.00001603, -0.00001671, -0.00000279, -0.00002608, -0.00009407, -0.00012061, -0.00003952, 0.00008712, 0.00015798, 0.00017466, 0.00019244, 0.00018236, 0.00007629, -0.00007677, -0.00014195, -0.00008859, -0.00002055, -0.00000895, -0.00001861, -0.00001328, -0.00001180, -0.00001864, -0.00000758, 0.00000942, -0.00000377, -0.00002482, -0.00000648, 0.00002148, -0.00000174, -0.00004210, -0.00002020, 0.00003588, 0.00003093, -0.00002313, -0.00002879, 0.00001392, 0.00001222, -0.00003463, -0.00003009, 0.00003053, 0.00004327, -0.00001087, -0.00003188, 0.00000866, 0.00001943, -0.00002748, -0.00004000, 0.00001415, 0.00004031, -0.00000629, -0.00003712, 0.00000289, 0.00003644, 0.00000309, -0.00003164, -0.00000493, 0.00003034, 0.00001305, -0.00001583, -0.00000246, 0.00001687, -0.00000610, -0.00003411, -0.00001488, 0.00002051, 0.00001584, -0.00001151, -0.00001088, 0.00000984, 0.00000887, -0.00000945, -0.00000953, 0.00000690, 0.00000932, -0.00000312, -0.00000550, 0.00000386, 0.00000481, -0.00000352, -0.00000343, 0.00000550, 0.00000614, -0.00000355, -0.00000741, -0.00000026, 0.00000494, 0.00000075, -0.00000465, -0.00000394, -0.00000091, -0.00000029, -0.00000017, 0.00000185, 0.00000375, 0.00000328, 0.00000085, -0.00000283, -0.00000678, -0.00000782, -0.00000375, 0.00000217, 0.00000456, 0.00000247, -0.00000063, -0.00000181, -0.00000070, 0.00000104, 0.00000054, -0.00000276, -0.00000416, 0.00000037, 0.00000623, 0.00000560, -0.00000047, -0.00000452, -0.00000503, -0.00000542, },
		{ 0.00000863, 0.00000613, -0.00000101, -0.00000573, -0.00000894, -0.00001157, -0.00000704, 0.00000246, 0.00000089, -0.00001163, -0.00001284, -0.00000705, -0.00003557, -0.00009650, -0.00010765, -0.00001619, 0.00010503, 0.00015463, 0.00014794, 0.00016043, 0.00016942, 0.00008723, -0.00006146, -0.00013961, -0.00009348, -0.00001905, -0.00000347, -0.00002029, -0.00001936, -0.00000999, -0.00001029, -0.00000630, 0.00000199, -0.00000683, -0.00001947, -0.00000440, 0.00001530, -0.00000471, -0.00003495, -0.00001378, 0.00003162, 0.00002367, -0.00002245, -0.00002442, 0.00001243, 0.00000808, -0.00003209, -0.00002323, 0.00003144, 0.00003767, -0.00001380, -0.00002942, 0.00000979, 0.00001663, -0.00002821, -0.00003591, 0.00001686, 0.00003722, -0.00000997, -0.00003541, 0.00000690, 0.00003624, -0.00000050, -0.00003266, -0.00000238, 0.00003189, 0.00001191, -0.00001645, -0.00000148, 0.00001628, -0.00000908, -0.00003526, -0.00001195, 0.00002343, 0.00001476, -0.00001408, -0.00001055, 0.00001198, 0.00000878, -0.00001157, -0.00001003, 0.00000861, 0.00001023, -0.00000405, -0.00000609, 0.00000457, 0.00000516, -0.00000419, -0.00000356, 0.00000641, 0.00000640, -0.00000456, -0.00000793, 0.00000065, 0.00000558, -0.00000030, -0.00000605, -0.00000385, 0.00000019, 0.00000026, -0.00000010, 0.00000243, 0.00000440, 0.00000263, -0.00000094, -0.00000399, -0.00000639, -0.00000673, -0.00000294, 0.00000262, 0.00000463, 0.00000189, -0.00000145, -0.00000184, 0.00000025, 0.00000195, 0.00000065, -0.00000314, -0.00000457, 0.00000001, 0.00000612, 0.00000592, -0.00000022, -0.00000488, -0.00000545, -0.00000500, -0.00000291, 0.00000330, },
		{ -0.00000230, -0.00000605, -0.00000732, -0.00000968, -0.00000812, -0.00000093, 0.00000062, -0.00000694, -0.00001011, -0.00001282, -0.00004441, -0.00009539, -0.00009166, 0.00000523, 0.00011666, 0.00014703, 0.00012374, 0.00013443, 0.00015877, 0.00009509, -0.00004922, -0.00013617, -0.00009577, -0.00001775, -0.00000020, -0.00002229, -0.00002351, -0.00000738, -0.00000353, -0.00000626, -0.00000442, -0.00000854, -0.00001475, -0.00000373, 0.00000942, -0.00000609, -0.00002743, -0.00000871, 0.00002643, 0.00001708, -0.00002098, -0.00002040, 0.00001040, 0.00000450, -0.00002866, -0.00001665, 0.00003127, 0.00003181, -0.00001608, -0.00002673, 0.00001045, 0.00001383, -0.00002817, -0.00003155, 0.00001873, 0.00003351, -0.00001305, -0.00003289, 0.00001062, 0.00003517, -0.00000414, -0.00003296, 0.00000045, 0.00003303, 0.00001052, -0.00001695, -0.00000058, 0.00001534, -0.00001194, -0.00003572, -0.00000868, 0.00002583, 0.00001311, -0.00001647, -0.00000974, 0.00001409, 0.00000829, -0.00001376, -0.00001018, 0.00001049, 0.00001093, -0.00000517, -0.00000660, 0.00000540, 0.00000548, -0.00000488, -0.00000361, 0.00000729, 0.00000647, -0.00000563, -0.00000825, 0.00000170, 0.00000604, -0.00000160, -0.00000740, -0.00000350, 0.00000141, 0.00000074, -0.00000008, 0.00000303, 0.00000496, 0.00000176, -0.00000280, -0.00000493, -0.00000572, -0.00000558, -0.00000228, 0.00000290, 0.00000462, 0.00000130, -0.00000218, -0.00000166, 0.00000137, 0.00000276, 0.00000045, -0.00000371, -0.00000483, -0.00000010, 0.00000603, 0.00000608, -0.00000004, -0.00000517, -0.00000582, -0.00000452, -0.00000174, 0.00000404, 0.00000833, 0.00000504, },
		{ -0.00000557, -0.00000827, -0.00000958, -0.00000387, 0.00000114, -0.00000239, -0.00000876, -0.00001962, -0.00005187, -0.00009082, -0.00007372, 0.00002384, 0.00012248, 0.00013650, 0.00010268, 0.00011377, 0.00014946, 0.00009985, -0.00003933, -0.00013138, -0.00009592, -0.00001703, 0.00000118, -0.00002414, -0.00002583, -0.00000443, 0.00000157, -0.00000708, -0.00000962, -0.00000922, -0.00001089, -0.00000420, 0.00000420, -0.00000600, -0.00001996, -0.00000506, 0.00002064, 0.00001131, -0.00001886, -0.00001686, 0.00000799, 0.00000162, -0.00002452, -0.00001058, 0.00003011, 0.00002589, -0.00001770, -0.00002393, 0.00001064, 0.00001114, -0.00002739, -0.00002713, 0.00001974, 0.00002936, -0.00001542, -0.00002967, 0.00001388, 0.00003330, -0.00000765, -0.00003253, 0.00000346, 0.00003371, 0.00000890, -0.00001730, 0.00000020, 0.00001409, -0.00001457, -0.00003547, -0.00000523, 0.00002761, 0.00001098, -0.00001857, -0.00000848, 0.00001606, 0.00000737, -0.00001592, -0.00000994, 0.00001249, 0.00001136, -0.00000646, -0.00000698, 0.00000635, 0.00000576, -0.00000558, -0.00000361, 0.00000810, 0.00000636, -0.00000671, -0.00000835, 0.00000285, 0.00000627, -0.00000311, -0.00000862, -0.00000286, 0.00000271, 0.00000112, -0.00000008, 0.00000365, 0.00000538, 0.00000068, -0.00000466, -0.00000562, -0.00000481, -0.00000442, -0.00000178, 0.00000303, 0.00000451, 0.00000071, -0.00000278, -0.00000124, 0.00000260, 0.00000338, -0.00000009, -0.00000438, -0.00000488, 0.00000001, 0.00000590, 0.00000605, 0.00000008, -0.00000539, -0.00000612, -0.00000398, -0.00000044, 0.00000479, 0.00000773, 0.00000361, -0.00000347, -0.00000587, },
		{ -0.00001122, -0.00000616, 0.00000231, 0.00000162, -0.00000885, -0.00002688, -0.00005739, -0.00008316, -0.00005492, 0.00003908, 0.00012320, 0.00012426, 0.00008508, 0.00009762, 0.00014077, 0.00010175, -0.00003116, -0.00012517, -0.00009445, -0.00001712, 0.00000103, -0.00002550, -0.00002656, -0.00000155, 0.00000509, -0.00000835, -0.00001356, -0.00000921, -0.00000806, -0.00000551, -0.00000009, -0.00000465, -0.00001292, -0.00000282, 0.00001455, 0.00000650, -0.00001629, -0.00001390, 0.00000535, -0.00000047, -0.00001983, -0.00000518, 0.00002803, 0.00002010, -0.00001866, -0.00002113, 0.00001042, 0.00000869, -0.00002594, -0.00002281, 0.00001988, 0.00002496, -0.00001699, -0.00002594, 0.00001653, 0.00003071, -0.00001087, -0.00003135, 0.00000651, 0.00003388, 0.00000710, -0.00001751, 0.00000083, 0.00001258, -0.00001687, -0.00003454, -0.00000172, 0.00002872, 0.00000847, -0.00002026, -0.00000681, 0.00001778, 0.00000604, -0.00001795, -0.00000928, 0.00001452, 0.00001148, -0.00000787, -0.00000719, 0.00000741, 0.00000598, -0.00000631, -0.00000355, 0.00000882, 0.00000607, -0.00000775, -0.00000822, 0.00000403, 0.00000624, -0.00000479, -0.00000965, -0.00000194, 0.00000404, 0.00000138, -0.00000012, 0.00000427, 0.00000566, -0.00000060, -0.00000647, -0.00000601, -0.00000370, -0.00000332, -0.00000145, 0.00000302, 0.00000434, 0.00000014, -0.00000320, -0.00000060, 0.00000386, 0.00000373, -0.00000094, -0.00000508, -0.00000467, 0.00000033, 0.00000569, 0.00000583, 0.00000016, -0.00000551, -0.00000634, -0.00000336, 0.00000098, 0.00000549, 0.00000679, 0.00000192, -0.00000439, -0.00000523, -0.00000385, -0.00000741, },
		{ 0.00000390, 0.00000474, -0.00001029, -0.00003398, -0.00006057, -0.00007295, -0.00003632, 0.00005063, 0.00011973, 0.00011138, 0.00007095, 0.00008513, 0.00013216, 0.00010115, -0.00002414, -0.00011763, -0.00009187, -0.00001811, -0.00000023, -0.00002614, -0.00002599, 0.00000095, 0.00000716, -0.00000975, -0.00001627, -0.00000884, -0.00000629, -0.00000729, -0.00000327, -0.00000233, -0.00000662, -0.00000188, 0.00000850, 0.00000269, -0.00001348, -0.00001155, 0.00000267, -0.00000174, -0.00001481, -0.00000058, 0.00002519, 0.00001458, -0.00001903, -0.00001844, 0.00000985, 0.00000658, -0.00002391, -0.00001876, 0.00001919, 0.00002053, -0.00001772, -0.00002187, 0.00001847, 0.00002752, -0.00001365, -0.00002946, 0.00000948, 0.00003353, 0.00000518, -0.00001757, 0.00000131, 0.00001089, -0.00001875, -0.00003298, 0.00000167, 0.00002913, 0.00000572, -0.00002147, -0.00000481, 0.00001916, 0.00000435, -0.00001975, -0.00000820, 0.00001652, 0.00001126, -0.00000936, -0.00000719, 0.00000857, 0.00000611, -0.00000706, -0.00000346, 0.00000946, 0.00000565, -0.00000869, -0.00000786, 0.00000518, 0.00000593, -0.00000658, -0.00001044, -0.00000075, 0.00000535, 0.00000150, -0.00000018, 0.00000488, 0.00000576, -0.00000202, -0.00000814, -0.00000610, -0.00000243, -0.00000231, -0.00000130, 0.00000290, 0.00000412, -0.00000038, -0.00000342, 0.00000027, 0.00000508, 0.00000376, -0.00000206, -0.00000570, -0.00000417, 0.00000078, 0.00000534, 0.00000543, 0.00000022, -0.00000553, -0.00000648, -0.00000266, 0.00000247, 0.00000607, 0.00000551, 0.00000007, -0.00000497, -0.00000419, -0.00000233, -0.00000711, -0.00001283, -0.00000763, },
		{ -0.00001283, -0.00004030, -0.00006125, -0.00006089, -0.00001883, 0.00005847, 0.00011303, 0.00009871, 0.00006007, 0.00007545, 0.00012330, 0.00009854, -0.00001785, -0.00010899, -0.00008862, -0.00001997, -0.00000217, -0.00002596, -0.00002447, 0.00000283, 0.00000804, -0.00001100, -0.00001791, -0.00000842, -0.00000553, -0.00000921, -0.00000524, 0.00000063, -0.00000132, -0.00000207, 0.00000278, -0.00000014, -0.00001063, -0.00000982, 0.00000011, -0.00000219, -0.00000966, 0.00000312, 0.00002173, 0.00000946, -0.00001887, -0.00001593, 0.00000900, 0.00000486, -0.00002143, -0.00001511, 0.00001778, 0.00001624, -0.00001763, -0.00001765, 0.00001962, 0.00002389, -0.00001586, -0.00002695, 0.00001225, 0.00003265, 0.00000320, -0.00001748, 0.00000162, 0.00000910, -0.00002014, -0.00003088, 0.00000482, 0.00002882, 0.00000285, -0.00002214, -0.00000258, 0.00002012, 0.00000235, -0.00002124, -0.00000673, 0.00001838, 0.00001068, -0.00001087, -0.00000695, 0.00000980, 0.00000612, -0.00000785, -0.00000333, 0.00001000, 0.00000513, -0.00000952, -0.00000731, 0.00000625, 0.00000531, -0.00000843, -0.00001093, 0.00000068, 0.00000661, 0.00000148, -0.00000024, 0.00000546, 0.00000567, -0.00000357, -0.00000964, -0.00000588, -0.00000108, -0.00000146, -0.00000131, 0.00000269, 0.00000386, -0.00000083, -0.00000341, 0.00000131, 0.00000618, 0.00000343, -0.00000338, -0.00000616, -0.00000338, 0.00000131, 0.00000483, 0.00000487, 0.00000031, -0.00000543, -0.00000652, -0.00000186, 0.00000399, 0.00000646, 0.00000391, -0.00000181, -0.00000515, -0.00000286, -0.00000117, -0.00000733, -0.00001418, -0.00000824, 0.00000565, 0.00000669, },
		{ -0.00005945, -0.00004777, -0.00000317, 0.00006286, 0.00010412, 0.00008685, 0.00005203, 0.00006783, 0.00011410, 0.00009443, -0.00001200, -0.00009957, -0.00008506, -0.00002252, -0.00000443, -0.00002496, -0.00002235, 0.00000400, 0.00000800, -0.00001190, -0.00001867, -0.00000819, -0.00000561, -0.00001094, -0.00000605, 0.00000387, 0.00000282, -0.00000315, -0.00000237, -0.00000208, -0.00000793, -0.00000867, -0.00000216, -0.00000185, -0.00000460, 0.00000588, 0.00001783, 0.00000486, -0.00001826, -0.00001365, 0.00000798, 0.00000358, -0.00001863, -0.00001197, 0.00001577, 0.00001228, -0.00001678, -0.00001349, 0.00001997, 0.00001998, -0.00001741, -0.00002391, 0.00001471, 0.00003125, 0.00000120, -0.00001726, 0.00000177, 0.00000728, -0.00002101, -0.00002834, 0.00000760, 0.00002786, 0.00000001, -0.00002225, -0.00000024, 0.00002060, 0.00000013, -0.00002233, -0.00000491, 0.00002003, 0.00000975, -0.00001232, -0.00000644, 0.00001106, 0.00000598, -0.00000866, -0.00000317, 0.00001045, 0.00000452, -0.00001020, -0.00000659, 0.00000717, 0.00000441, -0.00001027, -0.00001109, 0.00000232, 0.00000776, 0.00000132, -0.00000031, 0.00000598, 0.00000538, -0.00000519, -0.00001090, -0.00000538, 0.00000030, -0.00000078, -0.00000147, 0.00000242, 0.00000359, -0.00000116, -0.00000316, 0.00000249, 0.00000706, 0.00000272, -0.00000482, -0.00000638, -0.00000231, 0.00000184, 0.00000412, 0.00000419, 0.00000043, -0.00000522, -0.00000645, -0.00000099, 0.00000548, 0.00000662, 0.00000204, -0.00000360, -0.00000491, -0.00000140, -0.00000048, -0.00000795, -0.00001506, -0.00000801, 0.00000723, 0.00000733, -0.00001614, -0.00004531, },
		{ 0.00001014, 0.00006423, 0.00009393, 0.00007619, 0.00004632, 0.00006162, 0.00010459, 0.00008931, -0.00000644, -0.00008974, -0.00008143, -0.00002554, -0.00000671, -0.00002327, -0.00001995, 0.00000443, 0.00000732, -0.00001238, -0.00001880, -0.00000830, -0.00000632, -0.00001222, -0.00000581, 0.00000705, 0.00000573, -0.00000483, -0.00000675, -0.00000325, -0.00000553, -0.00000801, -0.00000401, -0.00000081, 0.00000019, 0.00000770, 0.00001365, 0.00000083, -0.00001730, -0.00001164, 0.00000687, 0.00000275, -0.00001566, -0.00000939, 0.00001331, 0.00000877, -0.00001526, -0.00000957, 0.00001952, 0.00001599, -0.00001824, -0.00002046, 0.00001674, 0.00002939, -0.00000074, -0.00001692, 0.00000176, 0.00000552, -0.00002133, -0.00002549, 0.00000990, 0.00002629, -0.00000267, -0.00002179, 0.00000211, 0.00002057, -0.00000222, -0.00002295, -0.00000280, 0.00002138, 0.00000849, -0.00001365, -0.00000566, 0.00001231, 0.00000568, -0.00000950, -0.00000298, 0.00001081, 0.00000388, -0.00001071, -0.00000575, 0.00000791, 0.00000325, -0.00001201, -0.00001090, 0.00000411, 0.00000877, 0.00000103, -0.00000037, 0.00000644, 0.00000490, -0.00000684, -0.00001189, -0.00000461, 0.00000166, -0.00000030, -0.00000175, 0.00000210, 0.00000332, -0.00000137, -0.00000266, 0.00000373, 0.00000765, 0.00000166, -0.00000628, -0.00000631, -0.00000103, 0.00000230, 0.00000322, 0.00000342, 0.00000062, -0.00000490, -0.00000627, -0.00000005, 0.00000687, 0.00000649, -0.00000002, -0.00000519, -0.00000427, 0.00000005, -0.00000031, -0.00000882, -0.00001531, -0.00000705, 0.00000834, 0.00000660, -0.00001977, -0.00004858, -0.00005539, -0.00003440, },
		{ 0.00008332, 0.00006687, 0.00004233, 0.00005635, 0.00009496, 0.00008362, -0.00000116, -0.00007985, -0.00007786, -0.00002874, -0.00000879, -0.00002106, -0.00001753, 0.00000422, 0.00000628, -0.00001242, -0.00001857, -0.00000882, -0.00000740, -0.00001285, -0.00000472, 0.00000986, 0.00000743, -0.00000684, -0.00001026, -0.00000381, -0.00000356, -0.00000773, -0.00000532, 0.00000080, 0.00000451, 0.00000861, 0.00000938, -0.00000258, -0.00001608, -0.00000991, 0.00000576, 0.00000234, -0.00001265, -0.00000742, 0.00001055, 0.00000582, -0.00001322, -0.00000604, 0.00001836, 0.00001208, -0.00001832, -0.00001675, 0.00001828, 0.00002710, -0.00000259, -0.00001646, 0.00000162, 0.00000389, -0.00002112, -0.00002244, 0.00001166, 0.00002423, -0.00000507, -0.00002082, 0.00000434, 0.00002002, -0.00000457, -0.00002306, -0.00000050, 0.00002237, 0.00000693, -0.00001479, -0.00000462, 0.00001350, 0.00000519, -0.00001035, -0.00000274, 0.00001110, 0.00000322, -0.00001107, -0.00000483, 0.00000842, 0.00000187, -0.00001360, -0.00001035, 0.00000600, 0.00000959, 0.00000061, -0.00000043, 0.00000680, 0.00000424, -0.00000845, -0.00001257, -0.00000361, 0.00000293, -0.00000004, -0.00000215, 0.00000178, 0.00000307, -0.00000142, -0.00000193, 0.00000496, 0.00000790, 0.00000029, -0.00000765, -0.00000590, 0.00000041, 0.00000263, 0.00000213, 0.00000262, 0.00000089, -0.00000447, -0.00000596, 0.00000094, 0.00000810, 0.00000607, -0.00000216, -0.00000647, -0.00000329, 0.00000133, -0.00000068, -0.00000973, -0.00001482, -0.00000558, 0.00000872, 0.00000460, -0.00002325, -0.00004983, -0.00004949, -0.00002158, 0.00002080, 0.00006319, },
		{ 0.00003948, 0.00005170, 0.00008548, 0.00007773, 0.00000375, -0.00007024, -0.00007441, -0.00003185, -0.00001054, -0.00001856, -0.00001528, 0.00000349, 0.00000508, -0.00001210, -0.00001820, -0.00000973, -0.00000855, -0.00001274, -0.00000304, 0.00001205, 0.00000805, -0.00000888, -0.00001283, -0.00000395, -0.00000209, -0.00000770, -0.00000601, 0.00000284, 0.00000823, 0.00000870, 0.00000516, -0.00000537, -0.00001469, -0.00000844, 0.00000472, 0.00000232, -0.00000974, -0.00000603, 0.00000766, 0.00000348, -0.00001080, -0.00000303, 0.00001658, 0.00000843, -0.00001768, -0.00001292, 0.00001927, 0.00002447, -0.00000429, -0.00001590, 0.00000138, 0.00000245, -0.00002042, -0.00001934, 0.00001283, 0.00002177, -0.00000711, -0.00001938, 0.00000634, 0.00001899, -0.00000683, -0.00002264, 0.00000191, 0.00002293, 0.00000515, -0.00001567, -0.00000332, 0.00001456, 0.00000451, -0.00001119, -0.00000245, 0.00001134, 0.00000256, -0.00001126, -0.00000389, 0.00000868, 0.00000032, -0.00001496, -0.00000946, 0.00000794, 0.00001021, 0.00000008, -0.00000047, 0.00000707, 0.00000342, -0.00000998, -0.00001294, -0.00000243, 0.00000407, 0.00000001, -0.00000261, 0.00000146, 0.00000288, -0.00000130, -0.00000101, 0.00000611, 0.00000777, -0.00000133, -0.00000884, -0.00000515, 0.00000193, 0.00000276, 0.00000089, 0.00000184, 0.00000125, -0.00000393, -0.00000554, 0.00000194, 0.00000911, 0.00000535, -0.00000429, -0.00000737, -0.00000207, 0.00000230, -0.00000153, -0.00001049, -0.00001354, -0.00000382, 0.00000815, 0.00000155, -0.00002612, -0.00004893, -0.00004227, -0.00000997, 0.00002878, 0.00006041, 0.00007293, 0.00005883, },
		{ 0.00007644, 0.00007186, 0.00000816, -0.00006115, -0.00007100, -0.00003462, -0.00001195, -0.00001601, -0.00001331, 0.00000245, 0.00000389, -0.00001156, -0.00001788, -0.00001093, -0.00000953, -0.00001187, -0.00000106, 0.00001343, 0.00000774, -0.00001070, -0.00001448, -0.00000386, -0.00000116, -0.00000776, -0.00000605, 0.00000515, 0.00001124, 0.00000806, 0.00000116, -0.00000754, -0.00001321, -0.00000721, 0.00000381, 0.00000260, -0.00000704, -0.00000518, 0.00000481, 0.00000176, -0.00000817, -0.00000062, 0.00001431, 0.00000518, -0.00001637, -0.00000914, 0.00001967, 0.00002158, -0.00000579, -0.00001525, 0.00000107, 0.00000125, -0.00001927, -0.00001630, 0.00001340, 0.00001906, -0.00000870, -0.00001758, 0.00000802, 0.00001751, -0.00000887, -0.00002171, 0.00000432, 0.00002304, 0.00000321, -0.00001623, -0.00000182, 0.00001545, 0.00000362, -0.00001199, -0.00000209, 0.00001152, 0.00000193, -0.00001131, -0.00000298, 0.00000869, -0.00000133, -0.00001606, -0.00000825, 0.00000986, 0.00001060, -0.00000055, -0.00000051, 0.00000722, 0.00000246, -0.00001139, -0.00001298, -0.00000112, 0.00000503, -0.00000014, -0.00000313, 0.00000119, 0.00000277, -0.00000100, 0.00000005, 0.00000710, 0.00000725, -0.00000311, -0.00000976, -0.00000408, 0.00000342, 0.00000265, -0.00000047, 0.00000112, 0.00000170, -0.00000329, -0.00000500, 0.00000290, 0.00000983, 0.00000438, -0.00000629, -0.00000787, -0.00000074, 0.00000286, -0.00000276, -0.00001092, -0.00001154, -0.00000207, 0.00000652, -0.00000223, -0.00002795, -0.00004595, -0.00003433, -0.00000012, 0.00003421, 0.00005660, 0.00006328, 0.00005188, 0.00003726, 0.00004748, },
		{ 0.00001187, -0.00005275, -0.00006753, -0.00003686, -0.00001306, -0.00001361, -0.00001165, 0.00000127, 0.00000279, -0.00001098, -0.00001775, -0.00001225, -0.00001010, -0.00001033, 0.00000092, 0.00001392, 0.00000676, -0.00001210, -0.00001530, -0.00000372, -0.00000075, -0.00000779, -0.00000546, 0.00000755, 0.00001346, 0.00000682, -0.00000252, -0.00000914, -0.00001171, -0.00000619, 0.00000308, 0.00000313, -0.00000465, -0.00000479, 0.00000213, 0.00000064, -0.00000551, 0.00000115, 0.00001171, 0.00000244, -0.00001448, -0.00000554, 0.00001949, 0.00001851, -0.00000708, -0.00001453, 0.00000072, 0.00000032, -0.00001776, -0.00001344, 0.00001339, 0.00001621, -0.00000980, -0.00001551, 0.00000929, 0.00001567, -0.00001061, -0.00002028, 0.00000663, 0.00002268, 0.00000120, -0.00001645, -0.00000016, 0.00001612, 0.00000255, -0.00001272, -0.00000167, 0.00001165, 0.00000133, -0.00001124, -0.00000212, 0.00000845, -0.00000303, -0.00001683, -0.00000675, 0.00001170, 0.00001075, -0.00000125, -0.00000053, 0.00000726, 0.00000140, -0.00001262, -0.00001271, 0.00000027, 0.00000577, -0.00000050, -0.00000367, 0.00000097, 0.00000275, -0.00000052, 0.00000120, 0.00000785, 0.00000634, -0.00000495, -0.00001033, -0.00000272, 0.00000482, 0.00000229, -0.00000188, 0.00000050, 0.00000223, -0.00000256, -0.00000437, 0.00000379, 0.00001023, 0.00000320, -0.00000805, -0.00000796, 0.00000057, 0.00000293, -0.00000422, -0.00001085, -0.00000891, -0.00000059, 0.00000382, -0.00000637, -0.00002841, -0.00004111, -0.00002633, 0.00000762, 0.00003745, 0.00005243, 0.00005463, 0.00004572, 0.00003524, 0.00004368, 0.00006812, 0.00006615, },
		{ -0.00006385, -0.00003843, -0.00001399, -0.00001155, -0.00001025, 0.00000016, 0.00000179, -0.00001053, -0.00001785, -0.00001349, -0.00001011, -0.00000828, 0.00000263, 0.00001353, 0.00000535, -0.00001293, -0.00001542, -0.00000368, -0.00000080, -0.00000764, -0.00000428, 0.00000988, 0.00001487, 0.00000512, -0.00000578, -0.00001022, -0.00001025, -0.00000534, 0.00000254, 0.00000379, -0.00000260, -0.00000477, -0.00000027, 0.00000006, -0.00000296, 0.00000227, 0.00000895, 0.00000030, -0.00001212, -0.00000225, 0.00001875, 0.00001537, -0.00000812, -0.00001375, 0.00000037, -0.00000031, -0.00001597, -0.00001085, 0.00001285, 0.00001336, -0.00001040, -0.00001330, 0.00001013, 0.00001357, -0.00001195, -0.00001843, 0.00000874, 0.00002186, -0.00000080, -0.00001629, 0.00000159, 0.00001651, 0.00000131, -0.00001336, -0.00000116, 0.00001175, 0.00000077, -0.00001108, -0.00000137, 0.00000797, -0.00000471, -0.00001726, -0.00000501, 0.00001340, 0.00001067, -0.00000200, -0.00000055, 0.00000718, 0.00000027, -0.00001365, -0.00001215, 0.00000167, 0.00000629, -0.00000103, -0.00000418, 0.00000084, 0.00000285, 0.00000011, 0.00000237, 0.00000832, 0.00000509, -0.00000674, -0.00001051, -0.00000116, 0.00000601, 0.00000167, -0.00000329, 0.00000003, 0.00000285, -0.00000176, -0.00000365, 0.00000456, 0.00001029, 0.00000188, -0.00000948, -0.00000769, 0.00000173, 0.00000252, -0.00000574, -0.00001019, -0.00000586, 0.00000034, 0.00000016, -0.00001044, -0.00002730, -0.00003478, -0.00001885, 0.00001307, 0.00003893, 0.00004846, 0.00004710, 0.00003999, 0.00003312, 0.00004036, 0.00006074, 0.00006062, 0.00001470, -0.00004514, },
		{ -0.00001487, -0.00000991, -0.00000901, -0.00000077, 0.00000083, -0.00001037, -0.00001817, -0.00001443, -0.00000947, -0.00000594, 0.00000383, 0.00001232, 0.00000379, -0.00001311, -0.00001501, -0.00000385, -0.00000121, -0.00000721, -0.00000260, 0.00001199, 0.00001547, 0.00000312, -0.00000856, -0.00001084, -0.00000886, -0.00000460, 0.00000219, 0.00000452, -0.00000095, -0.00000500, -0.00000229, -0.00000009, -0.00000066, 0.00000279, 0.00000619, -0.00000120, -0.00000944, 0.00000062, 0.00001749, 0.00001223, -0.00000891, -0.00001291, 0.00000006, -0.00000066, -0.00001400, -0.00000860, 0.00001185, 0.00001061, -0.00001053, -0.00001104, 0.00001049, 0.00001130, -0.00001284, -0.00001623, 0.00001056, 0.00002062, -0.00000270, -0.00001575, 0.00000337, 0.00001660, -0.00000006, -0.00001386, -0.00000056, 0.00001181, 0.00000023, -0.00001086, -0.00000074, 0.00000730, -0.00000629, -0.00001733, -0.00000309, 0.00001492, 0.00001035, -0.00000279, -0.00000056, 0.00000698, -0.00000090, -0.00001444, -0.00001132, 0.00000305, 0.00000655, -0.00000171, -0.00000466, 0.00000083, 0.00000307, 0.00000086, 0.00000348, 0.00000845, 0.00000356, -0.00000838, -0.00001028, 0.00000054, 0.00000694, 0.00000082, -0.00000462, -0.00000028, 0.00000354, -0.00000090, -0.00000287, 0.00000515, 0.00001001, 0.00000051, -0.00001052, -0.00000712, 0.00000260, 0.00000164, -0.00000713, -0.00000889, -0.00000261, 0.00000055, -0.00000428, -0.00001400, -0.00002453, -0.00002745, -0.00001244, 0.00001626, 0.00003920, 0.00004512, 0.00004062, 0.00003436, 0.00003074, 0.00003764, 0.00005447, 0.00005522, 0.00001650, -0.00003828, -0.00005981, -0.00003930, },
		{ -0.00000780, -0.00000141, -0.00000021, -0.00001064, -0.00001860, -0.00001485, -0.00000819, -0.00000356, 0.00000438, 0.00001046, 0.00000231, -0.00001263, -0.00001425, -0.00000432, -0.00000185, -0.00000643, -0.00000053, 0.00001374, 0.00001531, 0.00000094, -0.00001081, -0.00001106, -0.00000756, -0.00000396, 0.00000202, 0.00000522, 0.00000031, -0.00000538, -0.00000389, 0.00000011, 0.00000128, 0.00000278, 0.00000358, -0.00000207, -0.00000657, 0.00000299, 0.00001580, 0.00000920, -0.00000944, -0.00001202, -0.00000019, -0.00000073, -0.00001193, -0.00000674, 0.00001047, 0.00000808, -0.00001022, -0.00000887, 0.00001039, 0.00000899, -0.00001323, -0.00001377, 0.00001201, 0.00001901, -0.00000441, -0.00001485, 0.00000510, 0.00001635, -0.00000153, -0.00001419, 0.00000011, 0.00001182, -0.00000028, -0.00001060, -0.00000025, 0.00000647, -0.00000773, -0.00001705, -0.00000105, 0.00001622, 0.00000981, -0.00000359, -0.00000057, 0.00000669, -0.00000206, -0.00001499, -0.00001027, 0.00000435, 0.00000655, -0.00000251, -0.00000505, 0.00000094, 0.00000342, 0.00000169, 0.00000447, 0.00000824, 0.00000183, -0.00000977, -0.00000965, 0.00000227, 0.00000754, -0.00000023, -0.00000583, -0.00000041, 0.00000427, -0.00000000, -0.00000207, 0.00000554, 0.00000941, -0.00000082, -0.00001112, -0.00000636, 0.00000308, 0.00000039, -0.00000821, -0.00000698, 0.00000057, -0.00000010, -0.00000919, -0.00001668, -0.00002019, -0.00001966, -0.00000750, 0.00001740, 0.00003881, 0.00004270, 0.00003502, 0.00002855, 0.00002806, 0.00003568, 0.00004935, 0.00004985, 0.00001721, -0.00003210, -0.00005529, -0.00003950, -0.00001585, -0.00000875, },
		{ -0.00000144, -0.00001139, -0.00001900, -0.00001460, -0.00000637, -0.00000141, 0.00000418, 0.00000817, 0.00000113, -0.00001156, -0.00001333, -0.00000509, -0.00000260, -0.00000525, 0.00000180, 0.00001502, 0.00001446, -0.00000127, -0.00001252, -0.00001097, -0.00000638, -0.00000336, 0.00000199, 0.00000583, 0.00000121, -0.00000581, -0.00000506, 0.00000051, 0.00000279, 0.00000234, 0.00000126, -0.00000233, -0.00000367, 0.00000481, 0.00001375, 0.00000633, -0.00000971, -0.00001110, -0.00000036, -0.00000056, -0.00000987, -0.00000529, 0.00000882, 0.00000583, -0.00000954, -0.00000686, 0.00000985, 0.00000673, -0.00001312, -0.00001116, 0.00001305, 0.00001711, -0.00000587, -0.00001361, 0.00000670, 0.00001577, -0.00000304, -0.00001432, 0.00000087, 0.00001177, -0.00000078, -0.00001033, 0.00000009, 0.00000551, -0.00000896, -0.00001642, 0.00000104, 0.00001726, 0.00000907, -0.00000438, -0.00000058, 0.00000629, -0.00000318, -0.00001528, -0.00000903, 0.00000552, 0.00000630, -0.00000340, -0.00000534, 0.00000121, 0.00000389, 0.00000254, 0.00000525, 0.00000769, -0.00000000, -0.00001084, -0.00000865, 0.00000394, 0.00000778, -0.00000142, -0.00000684, -0.00000035, 0.00000502, 0.00000092, -0.00000127, 0.00000570, 0.00000852, -0.00000203, -0.00001128, -0.00000551, 0.00000311, -0.00000111, -0.00000883, -0.00000455, 0.00000340, -0.00000164, -0.00001422, -0.00001817, -0.00001450, -0.00001198, -0.00000427, 0.00001684, 0.00003824, 0.00004130, 0.00003002, 0.00002235, 0.00002516, 0.00003461, 0.00004533, 0.00004435, 0.00001681, -0.00002647, -0.00005022, -0.00003913, -0.00001706, -0.00000803, -0.00000650, -0.00000176, },
		{ -0.00001916, -0.00001356, -0.00000415, 0.00000028, 0.00000324, 0.00000568, 0.00000041, -0.00001001, -0.00001242, -0.00000615, -0.00000331, -0.00000367, 0.00000422, 0.00001576, 0.00001303, -0.00000339, -0.00001370, -0.00001062, -0.00000531, -0.00000279, 0.00000208, 0.00000630, 0.00000179, -0.00000618, -0.00000581, 0.00000100, 0.00000384, 0.00000160, -0.00000068, -0.00000206, -0.00000089, 0.00000605, 0.00001144, 0.00000371, -0.00000974, -0.00001015, -0.00000043, -0.00000019, -0.00000789, -0.00000425, 0.00000700, 0.00000391, -0.00000857, -0.00000511, 0.00000895, 0.00000465, -0.00001252, -0.00000850, 0.00001363, 0.00001498, -0.00000701, -0.00001209, 0.00000810, 0.00001487, -0.00000452, -0.00001423, 0.00000168, 0.00001164, -0.00000128, -0.00001006, 0.00000029, 0.00000450, -0.00000996, -0.00001548, 0.00000312, 0.00001802, 0.00000815, -0.00000514, -0.00000059, 0.00000583, -0.00000422, -0.00001531, -0.00000765, 0.00000654, 0.00000582, -0.00000433, -0.00000548, 0.00000164, 0.00000445, 0.00000335, 0.00000579, 0.00000684, -0.00000182, -0.00001153, -0.00000735, 0.00000544, 0.00000764, -0.00000268, -0.00000762, -0.00000012, 0.00000578, 0.00000184, -0.00000051, 0.00000560, 0.00000741, -0.00000304, -0.00001103, -0.00000469, 0.00000267, -0.00000270, -0.00000888, -0.00000176, 0.00000566, -0.00000399, -0.00001897, -0.00001826, -0.00000782, -0.00000497, -0.00000284, 0.00001504, 0.00003795, 0.00004087, 0.00002532, 0.00001567, 0.00002220, 0.00003451, 0.00004229, 0.00003861, 0.00001539, -0.00002121, -0.00004459, -0.00003836, -0.00001859, -0.00000765, -0.00000497, -0.00000186, -0.00000302, -0.00001261, },
		{ -0.00000176, 0.00000131, 0.00000165, 0.00000325, 0.00000022, -0.00000814, -0.00001167, -0.00000744, -0.00000387, -0.00000173, 0.00000660, 0.00001592, 0.00001114, -0.00000534, -0.00001438, -0.00001007, -0.00000435, -0.00000223, 0.00000222, 0.00000659, 0.00000212, -0.00000643, -0.00000618, 0.00000148, 0.00000443, 0.00000068, -0.00000215, -0.00000136, 0.00000166, 0.00000674, 0.00000899, 0.00000138, -0.00000954, -0.00000918, -0.00000039, 0.00000032, -0.00000608, -0.00000359, 0.00000512, 0.00000236, -0.00000741, -0.00000368, 0.00000775, 0.00000283, -0.00001148, -0.00000591, 0.00001375, 0.00001274, -0.00000779, -0.00001036, 0.00000924, 0.00001366, -0.00000593, -0.00001391, 0.00000252, 0.00001143, -0.00000178, -0.00000980, 0.00000036, 0.00000346, -0.00001069, -0.00001427, 0.00000514, 0.00001848, 0.00000708, -0.00000586, -0.00000060, 0.00000531, -0.00000516, -0.00001510, -0.00000619, 0.00000736, 0.00000511, -0.00000526, -0.00000547, 0.00000223, 0.00000508, 0.00000406, 0.00000603, 0.00000573, -0.00000354, -0.00001180, -0.00000582, 0.00000670, 0.00000715, -0.00000395, -0.00000812, 0.00000028, 0.00000651, 0.00000273, 0.00000017, 0.00000526, 0.00000613, -0.00000380, -0.00001043, -0.00000402, 0.00000179, -0.00000422, -0.00000831, 0.00000118, 0.00000712, -0.00000701, -0.00002306, -0.00001688, -0.00000058, 0.00000091, -0.00000313, 0.00001252, 0.00003822, 0.00004121, 0.00002063, 0.00000855, 0.00001939, 0.00003540, 0.00004001, 0.00003253, 0.00001310, -0.00001618, -0.00003849, -0.00003735, -0.00002046, -0.00000746, -0.00000314, -0.00000182, -0.00000507, -0.00001423, -0.00001890, -0.00001170, },
		{ -0.00000044, 0.00000112, 0.00000059, -0.00000614, -0.00001118, -0.00000885, -0.00000415, 0.00000049, 0.00000880, 0.00001551, 0.00000892, -0.00000702, -0.00001460, -0.00000937, -0.00000349, -0.00000169, 0.00000238, 0.00000669, 0.00000225, -0.00000651, -0.00000625, 0.00000184, 0.00000458, -0.00000028, -0.00000313, -0.00000033, 0.00000386, 0.00000689, 0.00000648, -0.00000063, -0.00000915, -0.00000819, -0.00000026, 0.00000093, -0.00000447, -0.00000327, 0.00000327, 0.00000116, -0.00000616, -0.00000259, 0.00000634, 0.00000135, -0.00001004, -0.00000349, 0.00001344, 0.00001046, -0.00000819, -0.00000850, 0.00001006, 0.00001220, -0.00000719, -0.00001334, 0.00000336, 0.00001111, -0.00000230, -0.00000957, 0.00000032, 0.00000245, -0.00001113, -0.00001282, 0.00000702, 0.00001863, 0.00000589, -0.00000653, -0.00000061, 0.00000475, -0.00000598, -0.00001467, -0.00000470, 0.00000797, 0.00000421, -0.00000616, -0.00000527, 0.00000298, 0.00000574, 0.00000460, 0.00000597, 0.00000445, -0.00000504, -0.00001165, -0.00000417, 0.00000763, 0.00000632, -0.00000515, -0.00000832, 0.00000081, 0.00000719, 0.00000357, 0.00000074, 0.00000468, 0.00000477, -0.00000426, -0.00000955, -0.00000359, 0.00000053, -0.00000551, -0.00000711, 0.00000405, 0.00000766, -0.00001044, -0.00002613, -0.00001407, 0.00000671, 0.00000530, -0.00000490, 0.00000981, 0.00003922, 0.00004205, 0.00001573, 0.00000112, 0.00001695, 0.00003719, 0.00003825, 0.00002608, 0.00001017, -0.00001123, -0.00003204, -0.00003629, -0.00002263, -0.00000732, -0.00000100, -0.00000180, -0.00000765, -0.00001609, -0.00001802, -0.00000907, 0.00000054, 0.00000153, },
		{ 0.00000147, -0.00000419, -0.00001100, -0.00001025, -0.00000410, 0.00000291, 0.00001068, 0.00001456, 0.00000650, -0.00000839, -0.00001441, -0.00000858, -0.00000272, -0.00000116, 0.00000252, 0.00000661, 0.00000227, -0.00000639, -0.00000610, 0.00000203, 0.00000437, -0.00000117, -0.00000361, 0.00000089, 0.00000565, 0.00000658, 0.00000403, -0.00000229, -0.00000859, -0.00000719, -0.00000005, 0.00000158, -0.00000311, -0.00000322, 0.00000153, 0.00000031, -0.00000490, -0.00000185, 0.00000483, 0.00000024, -0.00000831, -0.00000131, 0.00001271, 0.00000825, -0.00000823, -0.00000659, 0.00001053, 0.00001054, -0.00000827, -0.00001255, 0.00000418, 0.00001068, -0.00000283, -0.00000936, 0.00000020, 0.00000151, -0.00001128, -0.00001119, 0.00000873, 0.00001849, 0.00000462, -0.00000712, -0.00000061, 0.00000417, -0.00000666, -0.00001403, -0.00000321, 0.00000836, 0.00000315, -0.00000699, -0.00000486, 0.00000387, 0.00000636, 0.00000491, 0.00000561, 0.00000308, -0.00000626, -0.00001112, -0.00000250, 0.00000820, 0.00000523, -0.00000621, -0.00000822, 0.00000145, 0.00000780, 0.00000433, 0.00000119, 0.00000389, 0.00000340, -0.00000442, -0.00000850, -0.00000345, -0.00000101, -0.00000642, -0.00000537, 0.00000659, 0.00000723, -0.00001399, -0.00002788, -0.00001004, 0.00001352, 0.00000798, -0.00000777, 0.00000738, 0.00004096, 0.00004302, 0.00001046, -0.00000637, 0.00001508, 0.00003971, 0.00003674, 0.00001931, 0.00000685, -0.00000630, -0.00002544, -0.00003533, -0.00002499, -0.00000704, 0.00000142, -0.00000197, -0.00001078, -0.00001799, -0.00001639, -0.00000580, 0.00000250, 0.00000090, -0.00000280, -0.00000050, },
		{ -0.00001114, -0.00001152, -0.00000366, 0.00000539, 0.00001215, 0.00001312, 0.00000400, -0.00000942, -0.00001388, -0.00000771, -0.00000202, -0.00000066, 0.00000261, 0.00000637, 0.00000222, -0.00000609, -0.00000580, 0.00000200, 0.00000388, -0.00000188, -0.00000364, 0.00000218, 0.00000695, 0.00000588, 0.00000171, -0.00000360, -0.00000790, -0.00000619, 0.00000024, 0.00000220, -0.00000201, -0.00000339, -0.00000004, -0.00000025, -0.00000372, -0.00000146, 0.00000331, -0.00000045, -0.00000636, 0.00000055, 0.00001164, 0.00000618, -0.00000790, -0.00000470, 0.00001064, 0.00000874, -0.00000909, -0.00001155, 0.00000492, 0.00001011, -0.00000337, -0.00000917, 0.00000001, 0.00000068, -0.00001115, -0.00000943, 0.00001021, 0.00001805, 0.00000329, -0.00000763, -0.00000061, 0.00000360, -0.00000719, -0.00001323, -0.00000177, 0.00000853, 0.00000197, -0.00000770, -0.00000425, 0.00000488, 0.00000691, 0.00000496, 0.00000497, 0.00000172, -0.00000712, -0.00001025, -0.00000091, 0.00000838, 0.00000393, -0.00000708, -0.00000783, 0.00000216, 0.00000832, 0.00000500, 0.00000150, 0.00000294, 0.00000209, -0.00000429, -0.00000738, -0.00000364, -0.00000268, -0.00000683, -0.00000320, 0.00000859, 0.00000586, -0.00001730, -0.00002813, -0.00000508, 0.00001935, 0.00000887, -0.00001128, 0.00000563, 0.00004330, 0.00004375, 0.00000479, -0.00001363, 0.00001390, 0.00004271, 0.00003524, 0.00001234, 0.00000343, -0.00000135, -0.00001892, -0.00003456, -0.00002738, -0.00000649, 0.00000399, -0.00000250, -0.00001441, -0.00001969, -0.00001391, -0.00000211, 0.00000387, -0.00000055, -0.00000518, -0.00000147, 0.00000274, -0.00000248, },
		{ -0.00000284, 0.00000781, 0.00001314, 0.00001129, 0.00000155, -0.00001009, -0.00001306, -0.00000681, -0.00000140, -0.00000020, 0.00000262, 0.00000600, 0.00000217, -0.00000561, -0.00000544, 0.00000173, 0.00000320, -0.00000236, -0.00000326, 0.00000341, 0.00000777, 0.00000489, -0.00000039, -0.00000457, -0.00000710, -0.00000520, 0.00000057, 0.00000276, -0.00000118, -0.00000369, -0.00000138, -0.00000057, -0.00000269, -0.00000136, 0.00000188, -0.00000074, -0.00000430, 0.00000205, 0.00001030, 0.00000432, -0.00000727, -0.00000294, 0.00001039, 0.00000688, -0.00000965, -0.00001038, 0.00000555, 0.00000941, -0.00000391, -0.00000898, -0.00000020, -0.00000001, -0.00001074, -0.00000760, 0.00001144, 0.00001735, 0.00000194, -0.00000805, -0.00000060, 0.00000304, -0.00000756, -0.00001229, -0.00000043, 0.00000847, 0.00000071, -0.00000827, -0.00000343, 0.00000596, 0.00000731, 0.00000471, 0.00000411, 0.00000046, -0.00000759, -0.00000912, 0.00000051, 0.00000818, 0.00000251, -0.00000770, -0.00000717, 0.00000290, 0.00000873, 0.00000555, 0.00000165, 0.00000189, 0.00000090, -0.00000391, -0.00000630, -0.00000416, -0.00000434, -0.00000668, -0.00000079, 0.00000983, 0.00000368, -0.00002003, -0.00002678, 0.00000041, 0.00002378, 0.00000809, -0.00001493, 0.00000484, 0.00004601, 0.00004390, -0.00000121, -0.00002032, 0.00001348, 0.00004590, 0.00003354, 0.00000535, 0.00000018, 0.00000354, -0.00001274, -0.00003404, -0.00002961, -0.00000559, 0.00000655, -0.00000355, -0.00001839, -0.00002094, -0.00001058, 0.00000174, 0.00000446, -0.00000272, -0.00000733, -0.00000170, 0.00000424, -0.00000117, -0.00001155, -0.00001253, },
		{ 0.00001361, 0.00000916, -0.00000077, -0.00001043, -0.00001202, -0.00000589, -0.00000085, 0.00000019, 0.00000255, 0.00000555, 0.00000215, -0.00000501, -0.00000510, 0.00000125, 0.00000245, -0.00000254, -0.00000257, 0.00000449, 0.00000811, 0.00000370, -0.00000222, -0.00000522, -0.00000623, -0.00000424, 0.00000093, 0.00000322, -0.00000060, -0.00000407, -0.00000248, -0.00000071, -0.00000186, -0.00000152, 0.00000062, -0.00000064, -0.00000222, 0.00000316, 0.00000876, 0.00000273, -0.00000638, -0.00000135, 0.00000980, 0.00000503, -0.00000990, -0.00000908, 0.00000603, 0.00000858, -0.00000444, -0.00000879, -0.00000043, -0.00000056, -0.00001008, -0.00000576, 0.00001239, 0.00001639, 0.00000061, -0.00000836, -0.00000057, 0.00000252, -0.00000779, -0.00001125, 0.00000079, 0.00000820, -0.00000059, -0.00000866, -0.00000241, 0.00000707, 0.00000753, 0.00000415, 0.00000308, -0.00000060, -0.00000767, -0.00000782, 0.00000168, 0.00000761, 0.00000105, -0.00000805, -0.00000628, 0.00000365, 0.00000902, 0.00000596, 0.00000167, 0.00000078, -0.00000013, -0.00000335, -0.00000535, -0.00000495, -0.00000583, -0.00000596, 0.00000166, 0.00001020, 0.00000090, -0.00002186, -0.00002388, 0.00000597, 0.00002648, 0.00000587, -0.00001819, 0.00000514, 0.00004871, 0.00004319, -0.00000736, -0.00002612, 0.00001378, 0.00004892, 0.00003152, -0.00000141, -0.00000270, 0.00000825, -0.00000714, -0.00003373, -0.00003145, -0.00000429, 0.00000889, -0.00000522, -0.00002254, -0.00002151, -0.00000647, 0.00000545, 0.00000414, -0.00000539, -0.00000899, -0.00000119, 0.00000577, -0.00000036, -0.00001215, -0.00001315, -0.00000167, 0.00001003, },
		{ -0.00000286, -0.00001045, -0.00001081, -0.00000497, -0.00000037, 0.00000051, 0.00000241, 0.00000506, 0.00000218, -0.00000433, -0.00000482, 0.00000059, 0.00000170, -0.00000242, -0.00000167, 0.00000533, 0.00000801, 0.00000242, -0.00000374, -0.00000557, -0.00000532, -0.00000330, 0.00000128, 0.00000356, -0.00000023, -0.00000445, -0.00000333, -0.00000075, -0.00000127, -0.00000185, -0.00000040, -0.00000019, -0.00000023, 0.00000388, 0.00000710, 0.00000143, -0.00000531, -0.00000001, 0.00000891, 0.00000327, -0.00000984, -0.00000770, 0.00000634, 0.00000763, -0.00000494, -0.00000858, -0.00000063, -0.00000093, -0.00000920, -0.00000395, 0.00001304, 0.00001522, -0.00000069, -0.00000856, -0.00000052, 0.00000204, -0.00000787, -0.00001015, 0.00000187, 0.00000773, -0.00000187, -0.00000884, -0.00000121, 0.00000816, 0.00000750, 0.00000330, 0.00000198, -0.00000140, -0.00000736, -0.00000644, 0.00000253, 0.00000673, -0.00000037, -0.00000810, -0.00000519, 0.00000436, 0.00000917, 0.00000622, 0.00000155, -0.00000031, -0.00000096, -0.00000268, -0.00000460, -0.00000592, -0.00000700, -0.00000472, 0.00000393, 0.00000962, -0.00000222, -0.00002251, -0.00001960, 0.00001113, 0.00002726, 0.00000257, -0.00002062, 0.00000653, 0.00005104, 0.00004145, -0.00001339, -0.00003073, 0.00001466, 0.00005144, 0.00002913, -0.00000765, -0.00000503, 0.00001260, -0.00000231, -0.00003354, -0.00003270, -0.00000265, 0.00001078, -0.00000753, -0.00002658, -0.00002122, -0.00000177, 0.00000873, 0.00000287, -0.00000834, -0.00000995, -0.00000001, 0.00000715, -0.00000012, -0.00001281, -0.00001330, -0.00000024, 0.00001193, 0.00001357, 0.00000686, },
		{ -0.00000948, -0.00000405, 0.00000004, 0.00000074, 0.00000220, 0.00000457, 0.00000227, -0.00000364, -0.00000465, -0.00000018, 0.00000105, -0.00000202, -0.00000067, 0.00000588, 0.00000754, 0.00000113, -0.00000493, -0.00000568, -0.00000440, -0.00000242, 0.00000159, 0.00000374, -0.00000005, -0.00000478, -0.00000395, -0.00000075, -0.00000091, -0.00000228, -0.00000114, 0.00000053, 0.00000161, 0.00000423, 0.00000541, 0.00000044, -0.00000412, 0.00000104, 0.00000778, 0.00000167, -0.00000950, -0.00000630, 0.00000646, 0.00000659, -0.00000539, -0.00000834, -0.00000081, -0.00000113, -0.00000814, -0.00000223, 0.00001337, 0.00001386, -0.00000192, -0.00000865, -0.00000044, 0.00000161, -0.00000782, -0.00000901, 0.00000277, 0.00000710, -0.00000310, -0.00000880, 0.00000013, 0.00000917, 0.00000720, 0.00000218, 0.00000087, -0.00000187, -0.00000672, -0.00000510, 0.00000303, 0.00000561, -0.00000167, -0.00000788, -0.00000396, 0.00000502, 0.00000918, 0.00000632, 0.00000134, -0.00000134, -0.00000161, -0.00000199, -0.00000408, -0.00000698, -0.00000775, -0.00000305, 0.00000578, 0.00000810, -0.00000536, -0.00002179, -0.00001422, 0.00001541, 0.00002607, -0.00000135, -0.00002183, 0.00000884, 0.00005257, 0.00003861, -0.00001896, -0.00003396, 0.00001591, 0.00005315, 0.00002643, -0.00001308, -0.00000674, 0.00001637, 0.00000159, -0.00003331, -0.00003317, -0.00000077, 0.00001198, -0.00001045, -0.00003023, -0.00001994, 0.00000329, 0.00001127, 0.00000068, -0.00001126, -0.00001006, 0.00000171, 0.00000816, -0.00000045, -0.00001340, -0.00001292, 0.00000137, 0.00001339, 0.00001304, 0.00000450, -0.00000468, -0.00001019, },
		{ 0.00000038, 0.00000088, 0.00000195, 0.00000410, 0.00000238, -0.00000299, -0.00000459, -0.00000099, 0.00000057, -0.00000138, 0.00000034, 0.00000610, 0.00000676, -0.00000008, -0.00000578, -0.00000557, -0.00000350, -0.00000159, 0.00000185, 0.00000378, -0.00000000, -0.00000503, -0.00000437, -0.00000078, -0.00000079, -0.00000274, -0.00000156, 0.00000147, 0.00000322, 0.00000425, 0.00000378, -0.00000025, -0.00000291, 0.00000180, 0.00000648, 0.00000028, -0.00000888, -0.00000493, 0.00000636, 0.00000547, -0.00000577, -0.00000805, -0.00000093, -0.00000116, -0.00000693, -0.00000063, 0.00001340, 0.00001235, -0.00000305, -0.00000862, -0.00000035, 0.00000124, -0.00000766, -0.00000788, 0.00000350, 0.00000634, -0.00000422, -0.00000854, 0.00000157, 0.00001003, 0.00000660, 0.00000086, -0.00000014, -0.00000199, -0.00000582, -0.00000388, 0.00000318, 0.00000433, -0.00000279, -0.00000739, -0.00000264, 0.00000560, 0.00000905, 0.00000627, 0.00000105, -0.00000226, -0.00000209, -0.00000137, -0.00000383, -0.00000800, -0.00000799, -0.00000110, 0.00000704, 0.00000575, -0.00000820, -0.00001962, -0.00000811, 0.00001841, 0.00002303, -0.00000539, -0.00002158, 0.00001179, 0.00005296, 0.00003473, -0.00002372, -0.00003569, 0.00001724, 0.00005381, 0.00002353, -0.00001743, -0.00000784, 0.00001932, 0.00000451, -0.00003283, -0.00003276, 0.00000116, 0.00001231, -0.00001384, -0.00003318, -0.00001765, 0.00000841, 0.00001286, -0.00000227, -0.00001385, -0.00000925, 0.00000378, 0.00000865, -0.00000130, -0.00001378, -0.00001200, 0.00000302, 0.00001435, 0.00001209, 0.00000219, -0.00000618, -0.00000970, -0.00000808, -0.00000316, },
		{ 0.00000168, 0.00000369, 0.00000251, -0.00000244, -0.00000463, -0.00000176, 0.00000028, -0.00000056, 0.00000125, 0.00000599, 0.00000576, -0.00000115, -0.00000631, -0.00000531, -0.00000264, -0.00000084, 0.00000203, 0.00000368, -0.00000005, -0.00000516, -0.00000462, -0.00000087, -0.00000088, -0.00000314, -0.00000167, 0.00000254, 0.00000455, 0.00000399, 0.00000225, -0.00000067, -0.00000173, 0.00000224, 0.00000507, -0.00000086, -0.00000804, -0.00000366, 0.00000605, 0.00000431, -0.00000606, -0.00000772, -0.00000100, -0.00000102, -0.00000561, 0.00000080, 0.00001313, 0.00001073, -0.00000405, -0.00000847, -0.00000023, 0.00000091, -0.00000739, -0.00000677, 0.00000404, 0.00000547, -0.00000521, -0.00000804, 0.00000307, 0.00001069, 0.00000571, -0.00000061, -0.00000097, -0.00000174, -0.00000473, -0.00000286, 0.00000298, 0.00000297, -0.00000369, -0.00000667, -0.00000128, 0.00000609, 0.00000877, 0.00000606, 0.00000072, -0.00000302, -0.00000243, -0.00000088, -0.00000380, -0.00000884, -0.00000770, 0.00000095, 0.00000755, 0.00000273, -0.00001041, -0.00001602, -0.00000172, 0.00001984, 0.00001839, -0.00000904, -0.00001975, 0.00001502, 0.00005190, 0.00003002, -0.00002731, -0.00003592, 0.00001833, 0.00005325, 0.00002064, -0.00002049, -0.00000840, 0.00002124, 0.00000647, -0.00003190, -0.00003141, 0.00000292, 0.00001163, -0.00001749, -0.00003516, -0.00001440, 0.00001324, 0.00001332, -0.00000579, -0.00001583, -0.00000751, 0.00000596, 0.00000849, -0.00000256, -0.00001382, -0.00001057, 0.00000460, 0.00001476, 0.00001079, 0.00000003, -0.00000736, -0.00000903, -0.00000665, -0.00000230, 0.00000064, 0.00000093, },
		{ 0.00000261, -0.00000201, -0.00000474, -0.00000243, 0.00000020, 0.00000036, 0.00000200, 0.00000559, 0.00000465, -0.00000202, -0.00000655, -0.00000492, -0.00000183, -0.00000016, 0.00000213, 0.00000346, -0.00000014, -0.00000517, -0.00000476, -0.00000107, -0.00000112, -0.00000344, -0.00000147, 0.00000365, 0.00000558, 0.00000350, 0.00000088, -0.00000086, -0.00000065, 0.00000239, 0.00000364, -0.00000171, -0.00000701, -0.00000252, 0.00000553, 0.00000314, -0.00000625, -0.00000734, -0.00000099, -0.00000072, -0.00000423, 0.00000203, 0.00001256, 0.00000904, -0.00000491, -0.00000820, -0.00000010, 0.00000064, -0.00000704, -0.00000572, 0.00000440, 0.00000453, -0.00000603, -0.00000734, 0.00000455, 0.00001110, 0.00000454, -0.00000214, -0.00000156, -0.00000115, -0.00000355, -0.00000210, 0.00000249, 0.00000161, -0.00000434, -0.00000576, 0.00000007, 0.00000647, 0.00000835, 0.00000572, 0.00000039, -0.00000360, -0.00000268, -0.00000060, -0.00000395, -0.00000938, -0.00000689, 0.00000290, 0.00000722, -0.00000070, -0.00001172, -0.00001114, 0.00000448, 0.00001949, 0.00001253, -0.00001183, -0.00001638, 0.00001809, 0.00004923, 0.00002474, -0.00002944, -0.00003477, 0.00001886, 0.00005141, 0.00001800, -0.00002212, -0.00000861, 0.00002194, 0.00000757, -0.00003028, -0.00002917, 0.00000425, 0.00000987, -0.00002115, -0.00003592, -0.00001035, 0.00001744, 0.00001259, -0.00000957, -0.00001695, -0.00000494, 0.00000801, 0.00000760, -0.00000411, -0.00001341, -0.00000869, 0.00000596, 0.00001459, 0.00000924, -0.00000190, -0.00000822, -0.00000823, -0.00000522, -0.00000148, 0.00000082, 0.00000090, 0.00000143, 0.00000334, },
		{ -0.00000488, -0.00000293, 0.00000033, 0.00000128, 0.00000252, 0.00000493, 0.00000350, -0.00000268, -0.00000654, -0.00000445, -0.00000110, 0.00000041, 0.00000214, 0.00000314, -0.00000024, -0.00000506, -0.00000483, -0.00000139, -0.00000147, -0.00000356, -0.00000100, 0.00000474, 0.00000629, 0.00000285, -0.00000029, -0.00000086, 0.00000026, 0.00000228, 0.00000226, -0.00000228, -0.00000588, -0.00000156, 0.00000484, 0.00000201, -0.00000632, -0.00000690, -0.00000092, -0.00000029, -0.00000282, 0.00000304, 0.00001173, 0.00000731, -0.00000561, -0.00000783, 0.00000004, 0.00000041, -0.00000663, -0.00000474, 0.00000458, 0.00000357, -0.00000664, -0.00000645, 0.00000597, 0.00001122, 0.00000312, -0.00000364, -0.00000186, -0.00000026, -0.00000238, -0.00000165, 0.00000176, 0.00000035, -0.00000472, -0.00000472, 0.00000138, 0.00000673, 0.00000780, 0.00000526, 0.00000011, -0.00000400, -0.00000290, -0.00000056, -0.00000420, -0.00000952, -0.00000563, 0.00000455, 0.00000605, -0.00000428, -0.00001192, -0.00000521, 0.00001004, 0.00001731, 0.00000591, -0.00001339, -0.00001167, 0.00002057, 0.00004487, 0.00001926, -0.00002989, -0.00003246, 0.00001852, 0.00004834, 0.00001588, -0.00002226, -0.00000868, 0.00002130, 0.00000798, -0.00002780, -0.00002615, 0.00000489, 0.00000705, -0.00002447, -0.00003532, -0.00000575, 0.00002067, 0.00001070, -0.00001329, -0.00001703, -0.00000171, 0.00000968, 0.00000599, -0.00000575, -0.00001247, -0.00000649, 0.00000700, 0.00001388, 0.00000755, -0.00000354, -0.00000877, -0.00000734, -0.00000381, -0.00000070, 0.00000092, 0.00000081, 0.00000122, 0.00000305, 0.00000264, -0.00000173, },
		{ 0.00000062, 0.00000212, 0.00000280, 0.00000409, 0.00000239, -0.00000310, -0.00000631, -0.00000395, -0.00000045, 0.00000089, 0.00000205, 0.00000276, -0.00000031, -0.00000484, -0.00000487, -0.00000183, -0.00000187, -0.00000349, -0.00000030, 0.00000574, 0.00000669, 0.00000212, -0.00000124, -0.00000073, 0.00000099, 0.00000196, 0.00000099, -0.00000256, -0.00000469, -0.00000081, 0.00000400, 0.00000094, -0.00000628, -0.00000641, -0.00000078, 0.00000027, -0.00000142, 0.00000381, 0.00001066, 0.00000560, -0.00000614, -0.00000735, 0.00000019, 0.00000022, -0.00000617, -0.00000385, 0.00000460, 0.00000261, -0.00000704, -0.00000540, 0.00000726, 0.00001104, 0.00000152, -0.00000503, -0.00000184, 0.00000086, -0.00000131, -0.00000152, 0.00000086, -0.00000077, -0.00000485, -0.00000360, 0.00000260, 0.00000689, 0.00000713, 0.00000471, -0.00000012, -0.00000423, -0.00000314, -0.00000077, -0.00000446, -0.00000917, -0.00000404, 0.00000570, 0.00000408, -0.00000769, -0.00001089, 0.00000141, 0.00001454, 0.00001341, -0.00000094, -0.00001345, -0.00000593, 0.00002206, 0.00003893, 0.00001400, -0.00002855, -0.00002930, 0.00001708, 0.00004421, 0.00001454, -0.00002096, -0.00000885, 0.00001931, 0.00000795, -0.00002434, -0.00002258, 0.00000462, 0.00000331, -0.00002715, -0.00003328, -0.00000092, 0.00002266, 0.00000777, -0.00001660, -0.00001598, 0.00000196, 0.00001078, 0.00000372, -0.00000733, -0.00001098, -0.00000407, 0.00000763, 0.00001267, 0.00000584, -0.00000485, -0.00000907, -0.00000641, -0.00000245, 0.00000002, 0.00000093, 0.00000068, 0.00000106, 0.00000280, 0.00000256, -0.00000160, -0.00000499, -0.00000321, },
		{ 0.00000283, 0.00000313, 0.00000140, -0.00000329, -0.00000593, -0.00000343, 0.00000011, 0.00000126, 0.00000189, 0.00000233, -0.00000035, -0.00000455, -0.00000493, -0.00000237, -0.00000225, -0.00000319, 0.00000056, 0.00000658, 0.00000680, 0.00000135, -0.00000199, -0.00000053, 0.00000150, 0.00000148, -0.00000011, -0.00000259, -0.00000351, -0.00000028, 0.00000306, -0.00000003, -0.00000612, -0.00000589, -0.00000057, 0.00000092, -0.00000008, 0.00000433, 0.00000938, 0.00000393, -0.00000648, -0.00000679, 0.00000032, 0.00000007, -0.00000567, -0.00000305, 0.00000448, 0.00000170, -0.00000722, -0.00000425, 0.00000835, 0.00001054, -0.00000019, -0.00000622, -0.00000148, 0.00000212, -0.00000042, -0.00000168, -0.00000013, -0.00000168, -0.00000475, -0.00000245, 0.00000371, 0.00000693, 0.00000635, 0.00000409, -0.00000025, -0.00000431, -0.00000347, -0.00000122, -0.00000461, -0.00000832, -0.00000226, 0.00000621, 0.00000145, -0.00001062, -0.00000861, 0.00000832, 0.00001764, 0.00000800, -0.00000748, -0.00001188, 0.00000041, 0.00002221, 0.00003161, 0.00000934, -0.00002542, -0.00002565, 0.00001438, 0.00003925, 0.00001419, -0.00001837, -0.00000938, 0.00001601, 0.00000776, -0.00001985, -0.00001871, 0.00000326, -0.00000115, -0.00002885, -0.00002985, 0.00000378, 0.00002322, 0.00000405, -0.00001918, -0.00001378, 0.00000580, 0.00001113, 0.00000091, -0.00000864, -0.00000896, -0.00000161, 0.00000778, 0.00001104, 0.00000419, -0.00000580, -0.00000913, -0.00000546, -0.00000115, 0.00000068, 0.00000088, 0.00000053, 0.00000097, 0.00000256, 0.00000235, -0.00000160, -0.00000501, -0.00000325, 0.00000101, 0.00000281, },
		{ 0.00000057, -0.00000329, -0.00000543, -0.00000293, 0.00000057, 0.00000152, 0.00000165, 0.00000189, -0.00000033, -0.00000421, -0.00000504, -0.00000300, -0.00000257, -0.00000267, 0.00000152, 0.00000723, 0.00000666, 0.00000060, -0.00000253, -0.00000033, 0.00000180, 0.00000092, -0.00000101, -0.00000241, -0.00000240, 0.00000003, 0.00000205, -0.00000087, -0.00000585, -0.00000532, -0.00000028, 0.00000166, 0.00000118, 0.00000458, 0.00000793, 0.00000235, -0.00000663, -0.00000616, 0.00000044, -0.00000006, -0.00000516, -0.00000236, 0.00000424, 0.00000087, -0.00000717, -0.00000303, 0.00000920, 0.00000975, -0.00000194, -0.00000714, -0.00000082, 0.00000344, 0.00000024, -0.00000211, -0.00000112, -0.00000235, -0.00000447, -0.00000133, 0.00000471, 0.00000688, 0.00000549, 0.00000344, -0.00000028, -0.00000428, -0.00000391, -0.00000186, -0.00000456, -0.00000695, -0.00000048, 0.00000598, -0.00000165, -0.00001281, -0.00000516, 0.00001507, 0.00001912, 0.00000145, -0.00001322, -0.00000871, 0.00000689, 0.00002081, 0.00002325, 0.00000566, -0.00002065, -0.00002191, 0.00001038, 0.00003378, 0.00001498, -0.00001473, -0.00001049, 0.00001156, 0.00000768, -0.00001438, -0.00001486, 0.00000073, -0.00000602, -0.00002931, -0.00002518, 0.00000797, 0.00002224, -0.00000016, -0.00002072, -0.00001052, 0.00000949, 0.00001063, -0.00000228, -0.00000955, -0.00000646, 0.00000076, 0.00000742, 0.00000911, 0.00000270, -0.00000642, -0.00000902, -0.00000454, 0.00000008, 0.00000128, 0.00000077, 0.00000039, 0.00000094, 0.00000230, 0.00000198, -0.00000169, -0.00000489, -0.00000305, 0.00000145, 0.00000328, 0.00000262, 0.00000215, },
		{ -0.00000486, -0.00000246, 0.00000093, 0.00000168, 0.00000137, 0.00000148, -0.00000027, -0.00000386, -0.00000521, -0.00000367, -0.00000277, -0.00000195, 0.00000252, 0.00000765, 0.00000631, -0.00000009, -0.00000290, -0.00000015, 0.00000190, 0.00000033, -0.00000168, -0.00000206, -0.00000141, 0.00000013, 0.00000102, -0.00000158, -0.00000549, -0.00000473, 0.00000007, 0.00000245, 0.00000231, 0.00000459, 0.00000635, 0.00000089, -0.00000660, -0.00000548, 0.00000054, -0.00000017, -0.00000463, -0.00000178, 0.00000390, 0.00000014, -0.00000691, -0.00000182, 0.00000977, 0.00000868, -0.00000364, -0.00000773, 0.00000009, 0.00000471, 0.00000061, -0.00000276, -0.00000203, -0.00000276, -0.00000404, -0.00000027, 0.00000556, 0.00000674, 0.00000456, 0.00000277, -0.00000022, -0.00000419, -0.00000450, -0.00000262, -0.00000422, -0.00000514, 0.00000113, 0.00000497, -0.00000498, -0.00001401, -0.00000073, 0.00002123, 0.00001886, -0.00000580, -0.00001773, -0.00000411, 0.00001299, 0.00001775, 0.00001427, 0.00000324, -0.00001449, -0.00001844, 0.00000513, 0.00002815, 0.00001697, -0.00001036, -0.00001234, 0.00000622, 0.00000797, -0.00000806, -0.00001135, -0.00000298, -0.00001095, -0.00002833, -0.00001952, 0.00001127, 0.00001975, -0.00000450, -0.00002102, -0.00000639, 0.00001275, 0.00000925, -0.00000563, -0.00000994, -0.00000359, 0.00000288, 0.00000656, 0.00000698, 0.00000142, -0.00000671, -0.00000876, -0.00000365, 0.00000124, 0.00000181, 0.00000061, 0.00000028, 0.00000097, 0.00000199, 0.00000145, -0.00000183, -0.00000457, -0.00000263, 0.00000184, 0.00000350, 0.00000223, 0.00000121, -0.00000008, -0.00000312, },
		{ 0.00000119, 0.00000175, 0.00000106, 0.00000110, -0.00000016, -0.00000354, -0.00000545, -0.00000433, -0.00000283, -0.00000105, 0.00000349, 0.00000785, 0.00000580, -0.00000069, -0.00000312, -0.00000004, 0.00000183, -0.00000023, -0.00000212, -0.00000161, -0.00000058, 0.00000005, 0.00000003, -0.00000214, -0.00000504, -0.00000411, 0.00000048, 0.00000327, 0.00000329, 0.00000434, 0.00000469, -0.00000042, -0.00000640, -0.00000477, 0.00000060, -0.00000025, -0.00000411, -0.00000130, 0.00000348, -0.00000047, -0.00000647, -0.00000066, 0.00001002, 0.00000739, -0.00000519, -0.00000798, 0.00000120, 0.00000584, 0.00000069, -0.00000354, -0.00000280, -0.00000294, -0.00000351, 0.00000069, 0.00000628, 0.00000651, 0.00000358, 0.00000212, -0.00000005, -0.00000407, -0.00000524, -0.00000343, -0.00000352, -0.00000298, 0.00000240, 0.00000319, -0.00000826, -0.00001409, 0.00000441, 0.00002637, 0.00001691, -0.00001325, -0.00002068, 0.00000163, 0.00001823, 0.00001307, 0.00000515, 0.00000229, -0.00000730, -0.00001560, -0.00000118, 0.00002272, 0.00002011, -0.00000560, -0.00001501, 0.00000030, 0.00000886, -0.00000112, -0.00000850, -0.00000776, -0.00001556, -0.00002579, -0.00001322, 0.00001337, 0.00001587, -0.00000858, -0.00001995, -0.00000164, 0.00001531, 0.00000703, -0.00000892, -0.00000972, -0.00000049, 0.00000463, 0.00000524, 0.00000478, 0.00000042, -0.00000672, -0.00000841, -0.00000281, 0.00000230, 0.00000226, 0.00000043, 0.00000020, 0.00000104, 0.00000159, 0.00000077, -0.00000197, -0.00000403, -0.00000203, 0.00000212, 0.00000345, 0.00000172, 0.00000039, -0.00000052, -0.00000283, -0.00000427, -0.00000203, },
		{ 0.00000073, 0.00000078, -0.00000002, -0.00000327, -0.00000575, -0.00000494, -0.00000270, -0.00000002, 0.00000439, 0.00000781, 0.00000517, -0.00000119, -0.00000325, -0.00000003, 0.00000162, -0.00000071, -0.00000234, -0.00000112, 0.00000007, -0.00000017, -0.00000090, -0.00000254, -0.00000454, -0.00000348, 0.00000096, 0.00000409, 0.00000410, 0.00000387, 0.00000299, -0.00000155, -0.00000604, -0.00000405, 0.00000062, -0.00000032, -0.00000358, -0.00000091, 0.00000302, -0.00000094, -0.00000587, 0.00000040, 0.00000995, 0.00000596, -0.00000652, -0.00000786, 0.00000243, 0.00000675, 0.00000049, -0.00000438, -0.00000339, -0.00000289, -0.00000294, 0.00000153, 0.00000687, 0.00000622, 0.00000258, 0.00000150, 0.00000018, -0.00000398, -0.00000612, -0.00000418, -0.00000243, -0.00000060, 0.00000316, 0.00000076, -0.00001122, -0.00001299, 0.00000991, 0.00003016, 0.00001343, -0.00002036, -0.00002188, 0.00000809, 0.00002221, 0.00000696, -0.00000361, 0.00000286, 0.00000050, -0.00001364, -0.00000828, 0.00001783, 0.00002427, -0.00000083, -0.00001850, -0.00000584, 0.00001046, 0.00000612, -0.00000655, -0.00001339, -0.00001945, -0.00002173, -0.00000670, 0.00001403, 0.00001086, -0.00001201, -0.00001747, 0.00000341, 0.00001695, 0.00000410, -0.00001192, -0.00000887, 0.00000268, 0.00000591, 0.00000354, 0.00000261, -0.00000030, -0.00000650, -0.00000800, -0.00000203, 0.00000328, 0.00000265, 0.00000025, 0.00000018, 0.00000111, 0.00000109, -0.00000003, -0.00000203, -0.00000325, -0.00000133, 0.00000223, 0.00000313, 0.00000116, -0.00000025, -0.00000078, -0.00000246, -0.00000368, -0.00000166, 0.00000136, 0.00000172, },
		{ 0.00000011, -0.00000308, -0.00000609, -0.00000545, -0.00000239, 0.00000110, 0.00000516, 0.00000756, 0.00000448, -0.00000158, -0.00000331, -0.00000011, 0.00000133, -0.00000108, -0.00000238, -0.00000063, 0.00000052, -0.00000048, -0.00000172, -0.00000281, -0.00000400, -0.00000282, 0.00000149, 0.00000488, 0.00000470, 0.00000319, 0.00000131, -0.00000250, -0.00000555, -0.00000336, 0.00000059, -0.00000036, -0.00000307, -0.00000062, 0.00000253, -0.00000128, -0.00000516, 0.00000129, 0.00000957, 0.00000444, -0.00000755, -0.00000740, 0.00000367, 0.00000738, 0.00000004, -0.00000521, -0.00000375, -0.00000265, -0.00000238, 0.00000224, 0.00000733, 0.00000588, 0.00000159, 0.00000092, 0.00000045, -0.00000394, -0.00000709, -0.00000478, -0.00000094, 0.00000183, 0.00000331, -0.00000220, -0.00001360, -0.00001072, 0.00001539, 0.00003233, 0.00000869, -0.00002663, -0.00002131, 0.00001478, 0.00002459, -0.00000029, -0.00001156, 0.00000491, 0.00000843, -0.00001272, -0.00001582, 0.00001374, 0.00002920, 0.00000362, -0.00002270, -0.00001181, 0.00001284, 0.00001331, -0.00000571, -0.00001957, -0.00002227, -0.00001626, -0.00000037, 0.00001311, 0.00000505, -0.00001446, -0.00001366, 0.00000842, 0.00001751, 0.00000063, -0.00001443, -0.00000744, 0.00000574, 0.00000666, 0.00000155, 0.00000059, -0.00000073, -0.00000610, -0.00000756, -0.00000130, 0.00000416, 0.00000297, 0.00000007, 0.00000020, 0.00000116, 0.00000046, -0.00000091, -0.00000197, -0.00000225, -0.00000058, 0.00000210, 0.00000258, 0.00000062, -0.00000067, -0.00000087, -0.00000206, -0.00000312, -0.00000134, 0.00000143, 0.00000163, 0.00000042, 0.00000053, },
		{ -0.00000645, -0.00000582, -0.00000190, 0.00000226, 0.00000577, 0.00000713, 0.00000376, -0.00000187, -0.00000334, -0.00000029, 0.00000100, -0.00000129, -0.00000226, -0.00000021, 0.00000079, -0.00000085, -0.00000240, -0.00000296, -0.00000343, -0.00000215, 0.00000206, 0.00000561, 0.00000510, 0.00000235, -0.00000031, -0.00000324, -0.00000495, -0.00000270, 0.00000053, -0.00000038, -0.00000257, -0.00000040, 0.00000204, -0.00000147, -0.00000438, 0.00000199, 0.00000891, 0.00000294, -0.00000823, -0.00000664, 0.00000483, 0.00000769, -0.00000061, -0.00000593, -0.00000387, -0.00000228, -0.00000185, 0.00000280, 0.00000768, 0.00000549, 0.00000062, 0.00000040, 0.00000073, -0.00000399, -0.00000809, -0.00000514, 0.00000089, 0.00000416, 0.00000279, -0.00000546, -0.00001517, -0.00000742, 0.00002045, 0.00003275, 0.00000310, -0.00003163, -0.00001905, 0.00002120, 0.00002520, -0.00000826, -0.00001828, 0.00000823, 0.00001601, -0.00001289, -0.00002338, 0.00001065, 0.00003459, 0.00000742, -0.00002742, -0.00001724, 0.00001592, 0.00002006, -0.00000605, -0.00002593, -0.00002375, -0.00000964, 0.00000530, 0.00001059, -0.00000116, -0.00001564, -0.00000873, 0.00001303, 0.00001694, -0.00000315, -0.00001627, -0.00000549, 0.00000851, 0.00000685, -0.00000060, -0.00000120, -0.00000090, -0.00000557, -0.00000710, -0.00000063, 0.00000495, 0.00000323, -0.00000008, 0.00000027, 0.00000115, -0.00000030, -0.00000181, -0.00000174, -0.00000106, 0.00000012, 0.00000172, 0.00000184, 0.00000018, -0.00000086, -0.00000084, -0.00000168, -0.00000261, -0.00000107, 0.00000143, 0.00000147, 0.00000015, 0.00000033, 0.00000021, -0.00000299, },
		{ -0.00000125, 0.00000338, 0.00000620, 0.00000654, 0.00000304, -0.00000208, -0.00000336, -0.00000055, 0.00000067, -0.00000136, -0.00000204, 0.00000011, 0.00000089, -0.00000123, -0.00000295, -0.00000301, -0.00000286, -0.00000146, 0.00000268, 0.00000625, 0.00000527, 0.00000137, -0.00000183, -0.00000378, -0.00000427, -0.00000210, 0.00000042, -0.00000038, -0.00000208, -0.00000026, 0.00000156, -0.00000155, -0.00000358, 0.00000246, 0.00000800, 0.00000152, -0.00000853, -0.00000564, 0.00000583, 0.00000766, -0.00000138, -0.00000649, -0.00000377, -0.00000182, -0.00000139, 0.00000324, 0.00000792, 0.00000507, -0.00000032, -0.00000008, 0.00000099, -0.00000414, -0.00000906, -0.00000519, 0.00000300, 0.00000620, 0.00000159, -0.00000881, -0.00001576, -0.00000331, 0.00002470, 0.00003141, -0.00000290, -0.00003500, -0.00001535, 0.00002684, 0.00002397, -0.00001648, -0.00002347, 0.00001255, 0.00002281, -0.00001407, -0.00003055, 0.00000866, 0.00004008, 0.00001037, -0.00003238, -0.00002182, 0.00001957, 0.00002598, -0.00000756, -0.00003203, -0.00002369, -0.00000220, 0.00000994, 0.00000658, -0.00000731, -0.00001538, -0.00000294, 0.00001691, 0.00001525, -0.00000698, -0.00001733, -0.00000318, 0.00001085, 0.00000651, -0.00000279, -0.00000270, -0.00000084, -0.00000498, -0.00000663, -0.00000000, 0.00000565, 0.00000345, -0.00000019, 0.00000035, 0.00000105, -0.00000116, -0.00000267, -0.00000131, 0.00000027, 0.00000071, 0.00000106, 0.00000100, -0.00000009, -0.00000082, -0.00000073, -0.00000135, -0.00000214, -0.00000085, 0.00000135, 0.00000126, -0.00000008, 0.00000019, 0.00000027, -0.00000299, -0.00000678, -0.00000601, },
		{ 0.00000644, 0.00000583, 0.00000234, -0.00000222, -0.00000341, -0.00000086, 0.00000040, -0.00000129, -0.00000177, 0.00000031, 0.00000085, -0.00000156, -0.00000334, -0.00000298, -0.00000230, -0.00000074, 0.00000332, 0.00000677, 0.00000522, 0.00000030, -0.00000321, -0.00000413, -0.00000355, -0.00000158, 0.00000029, -0.00000035, -0.00000161, -0.00000017, 0.00000110, -0.00000152, -0.00000280, 0.00000269, 0.00000690, 0.00000026, -0.00000845, -0.00000448, 0.00000659, 0.00000730, -0.00000221, -0.00000684, -0.00000346, -0.00000132, -0.00000101, 0.00000356, 0.00000806, 0.00000462, -0.00000120, -0.00000051, 0.00000117, -0.00000437, -0.00000990, -0.00000489, 0.00000527, 0.00000780, -0.00000024, -0.00001197, -0.00001529, 0.00000134, 0.00002783, 0.00002842, -0.00000883, -0.00003653, -0.00001058, 0.00003125, 0.00002101, -0.00002445, -0.00002695, 0.00001746, 0.00002846, -0.00001610, -0.00003692, 0.00000775, 0.00004526, 0.00001233, -0.00003725, -0.00002528, 0.00002353, 0.00003074, -0.00001010, -0.00003743, -0.00002201, 0.00000560, 0.00001320, 0.00000130, -0.00001295, -0.00001360, 0.00000334, 0.00001978, 0.00001255, -0.00001059, -0.00001755, -0.00000064, 0.00001263, 0.00000569, -0.00000489, -0.00000387, -0.00000061, -0.00000435, -0.00000616, 0.00000058, 0.00000625, 0.00000362, -0.00000027, 0.00000043, 0.00000085, -0.00000211, -0.00000344, -0.00000068, 0.00000167, 0.00000110, 0.00000015, 0.00000013, -0.00000015, -0.00000058, -0.00000059, -0.00000109, -0.00000173, -0.00000066, 0.00000119, 0.00000103, -0.00000025, 0.00000009, 0.00000024, -0.00000309, -0.00000705, -0.00000601, -0.00000047, 0.00000444, },
		{ 0.00000168, -0.00000233, -0.00000347, -0.00000120, 0.00000020, -0.00000110, -0.00000150, 0.00000037, 0.00000072, -0.00000182, -0.00000359, -0.00000290, -0.00000174, 0.00000001, 0.00000396, 0.00000714, 0.00000496, -0.00000082, -0.00000441, -0.00000430, -0.00000283, -0.00000113, 0.00000014, -0.00000029, -0.00000116, -0.00000013, 0.00000068, -0.00000142, -0.00000209, 0.00000270, 0.00000569, -0.00000076, -0.00000800, -0.00000324, 0.00000705, 0.00000665, -0.00000300, -0.00000695, -0.00000299, -0.00000083, -0.00000073, 0.00000377, 0.00000812, 0.00000415, -0.00000202, -0.00000091, 0.00000127, -0.00000467, -0.00001054, -0.00000421, 0.00000756, 0.00000883, -0.00000257, -0.00001471, -0.00001374, 0.00000620, 0.00002956, 0.00002401, -0.00001421, -0.00003612, -0.00000517, 0.00003406, 0.00001654, -0.00003167, -0.00002864, 0.00002254, 0.00003269, -0.00001871, -0.00004214, 0.00000780, 0.00004977, 0.00001327, -0.00004166, -0.00002748, 0.00002749, 0.00003407, -0.00001342, -0.00004173, -0.00001878, 0.00001331, 0.00001487, -0.00000490, -0.00001766, -0.00001037, 0.00000972, 0.00002143, 0.00000903, -0.00001374, -0.00001692, 0.00000192, 0.00001376, 0.00000447, -0.00000679, -0.00000470, -0.00000026, -0.00000374, -0.00000568, 0.00000114, 0.00000677, 0.00000375, -0.00000031, 0.00000048, 0.00000053, -0.00000310, -0.00000404, 0.00000016, 0.00000306, 0.00000125, -0.00000097, -0.00000068, 0.00000005, -0.00000015, -0.00000048, -0.00000092, -0.00000137, -0.00000051, 0.00000098, 0.00000079, -0.00000035, 0.00000001, 0.00000012, -0.00000327, -0.00000722, -0.00000579, 0.00000040, 0.00000538, 0.00000649, 0.00000503, },
		{ -0.00000356, -0.00000151, 0.00000010, -0.00000082, -0.00000128, 0.00000029, 0.00000054, -0.00000198, -0.00000373, -0.00000280, -0.00000119, 0.00000079, 0.00000458, 0.00000734, 0.00000450, -0.00000192, -0.00000541, -0.00000431, -0.00000213, -0.00000078, -0.00000002, -0.00000019, -0.00000073, -0.00000013, 0.00000029, -0.00000127, -0.00000149, 0.00000249, 0.00000445, -0.00000149, -0.00000722, -0.00000203, 0.00000717, 0.00000575, -0.00000370, -0.00000680, -0.00000240, -0.00000039, -0.00000054, 0.00000390, 0.00000809, 0.00000367, -0.00000278, -0.00000131, 0.00000125, -0.00000498, -0.00001088, -0.00000318, 0.00000972, 0.00000922, -0.00000527, -0.00001681, -0.00001119, 0.00001089, 0.00002974, 0.00001851, -0.00001862, -0.00003384, 0.00000039, 0.00003502, 0.00001090, -0.00003770, -0.00002865, 0.00002732, 0.00003533, -0.00002157, -0.00004593, 0.00000860, 0.00005325, 0.00001328, -0.00004524, -0.00002836, 0.00003111, 0.00003579, -0.00001720, -0.00004456, -0.00001417, 0.00002044, 0.00001485, -0.00001160, -0.00002109, -0.00000583, 0.00001578, 0.00002176, 0.00000493, -0.00001619, -0.00001551, 0.00000432, 0.00001419, 0.00000298, -0.00000839, -0.00000521, 0.00000013, -0.00000317, -0.00000518, 0.00000167, 0.00000719, 0.00000385, -0.00000033, 0.00000049, 0.00000009, -0.00000408, -0.00000445, 0.00000118, 0.00000436, 0.00000113, -0.00000224, -0.00000136, 0.00000050, 0.00000039, -0.00000043, -0.00000085, -0.00000103, -0.00000039, 0.00000072, 0.00000055, -0.00000039, -0.00000006, -0.00000012, -0.00000351, -0.00000725, -0.00000536, 0.00000132, 0.00000616, 0.00000636, 0.00000419, 0.00000105, -0.00000241, },
		{ 0.00000010, -0.00000050, -0.00000112, 0.00000010, 0.00000034, -0.00000202, -0.00000376, -0.00000269, -0.00000065, 0.00000161, 0.00000516, 0.00000735, 0.00000387, -0.00000297, -0.00000619, -0.00000421, -0.00000148, -0.00000050, -0.00000016, -0.00000005, -0.00000032, -0.00000016, -0.00000006, -0.00000111, -0.00000101, 0.00000210, 0.00000323, -0.00000192, -0.00000617, -0.00000091, 0.00000694, 0.00000467, -0.00000423, -0.00000640, -0.00000174, -0.00000002, -0.00000042, 0.00000395, 0.00000798, 0.00000319, -0.00000346, -0.00000172, 0.00000111, -0.00000526, -0.00001087, -0.00000186, 0.00001159, 0.00000892, -0.00000813, -0.00001809, -0.00000780, 0.00001508, 0.00002834, 0.00001232, -0.00002169, -0.00002989, 0.00000559, 0.00003404, 0.00000454, -0.00004219, -0.00002716, 0.00003138, 0.00003635, -0.00002431, -0.00004814, 0.00000987, 0.00005546, 0.00001254, -0.00004768, -0.00002800, 0.00003400, 0.00003585, -0.00002103, -0.00004566, -0.00000847, 0.00002651, 0.00001316, -0.00001832, -0.00002298, -0.00000029, 0.00002113, 0.00002073, 0.00000055, -0.00001779, -0.00001346, 0.00000638, 0.00001394, 0.00000134, -0.00000963, -0.00000544, 0.00000053, -0.00000264, -0.00000465, 0.00000219, 0.00000753, 0.00000392, -0.00000034, 0.00000041, -0.00000046, -0.00000500, -0.00000462, 0.00000231, 0.00000549, 0.00000071, -0.00000359, -0.00000183, 0.00000121, 0.00000098, -0.00000049, -0.00000087, -0.00000071, -0.00000029, 0.00000041, 0.00000032, -0.00000036, -0.00000016, -0.00000046, -0.00000376, -0.00000710, -0.00000473, 0.00000223, 0.00000676, 0.00000608, 0.00000332, 0.00000046, -0.00000249, -0.00000365, -0.00000177, },
		},
		{ { 0.00000330, 0.00000324, -0.00000013, -0.00000722, -0.00000972, -0.00000374, 0.00000310, 0.00000354, 0.00000012, 0.00000018, 0.00000544, 0.00000844, 0.00000315, -0.00000407, -0.00000429, -0.00000073, -0.00000066, -0.00000274, -0.00000342, -0.00000220, 0.00000296, 0.00000927, 0.00000606, -0.00000574, -0.00000967, -0.00000203, 0.00000231, -0.00000176, -0.00000425, -0.00000123, 0.00000234, 0.00000339, 0.00000208, 0.00000164, 0.00000385, 0.00000184, -0.00000548, -0.00000427, 0.00000338, -0.00000337, -0.00001330, 0.00000239, 0.00001729, -0.00000751, -0.00002911, 0.00000099, 0.00002698, -0.00000237, -0.00002225, 0.00001074, 0.00002356, -0.00001215, -0.00001690, 0.00001985, 0.00001784, -0.00001376, 0.00000035, 0.00002967, 0.00000311, -0.00003424, -0.00001190, 0.00002919, 0.00002339, -0.00001253, -0.00003037, -0.00001253, 0.00002339, 0.00002919, -0.00001190, -0.00003424, 0.00000311, 0.00002967, 0.00000035, -0.00001376, 0.00001784, 0.00001985, -0.00001690, -0.00001215, 0.00002356, 0.00001074, -0.00002225, -0.00000237, 0.00002698, 0.00000099, -0.00002911, -0.00000751, 0.00001729, 0.00000239, -0.00001330, -0.00000337, 0.00000338, -0.00000427, -0.00000548, 0.00000184, 0.00000385, 0.00000164, 0.00000208, 0.00000339, 0.00000234, -0.00000123, -0.00000425, -0.00000176, 0.00000231, -0.00000203, -0.00000967, -0.00000574, 0.00000606, 0.00000927, 0.00000296, -0.00000220, -0.00000342, -0.00000274, -0.00000066, -0.00000073, -0.00000429, -0.00000407, 0.00000315, 0.00000844, 0.00000544, 0.00000018, 0.00000012, 0.00000354, 0.00000310, -0.00000374, -0.00000972, -0.00000722, -0.00000013, 0.00000324, },
		{ 0.00000147, -0.00000524, -0.00001029, -0.00000619, 0.00000193, 0.00000424, 0.00000082, -0.00000027, 0.00000434, 0.00000827, 0.00000464, -0.00000258, -0.00000446, -0.00000159, -0.00000068, -0.00000216, -0.00000315, -0.00000305, 0.00000089, 0.00000829, 0.00000827, -0.00000307, -0.00001017, -0.00000379, 0.00000258, -0.00000080, -0.00000465, -0.00000204, 0.00000194, 0.00000325, 0.00000252, 0.00000175, 0.00000298, 0.00000229, -0.00000370, -0.00000501, 0.00000177, -0.00000126, -0.00001242, -0.00000254, 0.00001654, 0.00000032, -0.00002834, -0.00000977, 0.00002454, 0.00000744, -0.00002139, 0.00000149, 0.00002511, -0.00000263, -0.00001845, 0.00001377, 0.00002186, -0.00000981, -0.00000437, 0.00002986, 0.00001102, -0.00003413, -0.00001832, 0.00003020, 0.00002744, -0.00001458, -0.00003153, -0.00000996, 0.00002136, 0.00002652, -0.00000702, -0.00003186, -0.00000381, 0.00002701, 0.00000528, -0.00001552, 0.00001221, 0.00002421, -0.00001232, -0.00002024, 0.00001821, 0.00001918, -0.00001891, -0.00001214, 0.00002480, 0.00001162, -0.00002594, -0.00001525, 0.00001531, 0.00000756, -0.00001254, -0.00000601, 0.00000421, -0.00000280, -0.00000691, 0.00000078, 0.00000462, 0.00000185, 0.00000164, 0.00000341, 0.00000260, -0.00000057, -0.00000355, -0.00000244, 0.00000156, -0.00000060, -0.00000842, -0.00000780, 0.00000322, 0.00000945, 0.00000507, -0.00000092, -0.00000344, -0.00000337, -0.00000093, 0.00000009, -0.00000365, -0.00000529, 0.00000130, 0.00000820, 0.00000654, 0.00000088, -0.00000033, 0.00000271, 0.00000363, -0.00000147, -0.00000835, -0.00000864, -0.00000212, 0.00000282, 0.00000335, 0.00000334, },
		{ -0.00000995, -0.00000851, 0.00000011, 0.00000462, 0.00000174, -0.00000047, 0.00000330, 0.00000781, 0.00000571, -0.00000102, -0.00000418, -0.00000232, -0.00000094, -0.00000172, -0.00000273, -0.00000349, -0.00000096, 0.00000671, 0.00000968, -0.00000011, -0.00000980, -0.00000560, 0.00000230, 0.00000029, -0.00000467, -0.00000289, 0.00000141, 0.00000299, 0.00000284, 0.00000217, 0.00000225, 0.00000214, -0.00000194, -0.00000487, -0.00000019, -0.00000016, -0.00001038, -0.00000635, 0.00001349, 0.00000686, -0.00002404, -0.00001883, 0.00001794, 0.00001536, -0.00001665, -0.00000675, 0.00002277, 0.00000669, -0.00001652, 0.00000717, 0.00002349, -0.00000431, -0.00000782, 0.00002764, 0.00001869, -0.00003114, -0.00002518, 0.00002877, 0.00003276, -0.00001504, -0.00003469, -0.00000805, 0.00002156, 0.00002330, -0.00000438, -0.00002788, -0.00000876, 0.00002234, 0.00000929, -0.00001499, 0.00000606, 0.00002608, -0.00000571, -0.00002561, 0.00000991, 0.00002514, -0.00001195, -0.00001992, 0.00001844, 0.00002024, -0.00001915, -0.00002140, 0.00001071, 0.00001192, -0.00001005, -0.00000856, 0.00000400, -0.00000094, -0.00000770, -0.00000071, 0.00000506, 0.00000234, 0.00000132, 0.00000336, 0.00000273, -0.00000013, -0.00000266, -0.00000267, 0.00000049, 0.00000027, -0.00000661, -0.00000901, 0.00000002, 0.00000870, 0.00000701, 0.00000075, -0.00000317, -0.00000396, -0.00000146, 0.00000071, -0.00000264, -0.00000604, -0.00000076, 0.00000745, 0.00000752, 0.00000180, -0.00000050, 0.00000188, 0.00000363, 0.00000040, -0.00000641, -0.00000930, -0.00000430, 0.00000197, 0.00000347, 0.00000327, 0.00000256, -0.00000299, },
		{ -0.00000229, 0.00000452, 0.00000281, -0.00000038, 0.00000235, 0.00000719, 0.00000640, 0.00000042, -0.00000354, -0.00000279, -0.00000135, -0.00000148, -0.00000226, -0.00000356, -0.00000247, 0.00000474, 0.00001020, 0.00000283, -0.00000859, -0.00000719, 0.00000148, 0.00000131, -0.00000433, -0.00000370, 0.00000077, 0.00000263, 0.00000298, 0.00000277, 0.00000186, 0.00000152, -0.00000056, -0.00000390, -0.00000197, -0.00000019, -0.00000784, -0.00000852, 0.00000901, 0.00001111, -0.00001728, -0.00002477, 0.00000840, 0.00001980, -0.00000911, -0.00001242, 0.00001723, 0.00001424, -0.00001134, 0.00000142, 0.00002247, 0.00000173, -0.00000922, 0.00002365, 0.00002500, -0.00002546, -0.00003122, 0.00002465, 0.00003823, -0.00001322, -0.00003907, -0.00000782, 0.00002362, 0.00002071, -0.00000410, -0.00002348, -0.00001117, 0.00001659, 0.00001151, -0.00001258, 0.00000052, 0.00002525, 0.00000156, -0.00002749, 0.00000007, 0.00002742, -0.00000270, -0.00002419, 0.00000915, 0.00002531, -0.00000984, -0.00002469, 0.00000420, 0.00001450, -0.00000614, -0.00001032, 0.00000277, 0.00000083, -0.00000772, -0.00000236, 0.00000504, 0.00000296, 0.00000118, 0.00000331, 0.00000277, 0.00000004, -0.00000173, -0.00000239, -0.00000067, 0.00000046, -0.00000456, -0.00000922, -0.00000320, 0.00000699, 0.00000855, 0.00000275, -0.00000254, -0.00000441, -0.00000217, 0.00000100, -0.00000143, -0.00000619, -0.00000281, 0.00000613, 0.00000822, 0.00000291, -0.00000039, 0.00000116, 0.00000323, 0.00000173, -0.00000418, -0.00000912, -0.00000638, 0.00000061, 0.00000355, 0.00000320, 0.00000311, -0.00000075, -0.00000870, -0.00001039, },
		{ 0.00000391, 0.00000000, 0.00000153, 0.00000649, 0.00000678, 0.00000162, -0.00000268, -0.00000291, -0.00000178, -0.00000147, -0.00000182, -0.00000334, -0.00000356, 0.00000262, 0.00000988, 0.00000543, -0.00000666, -0.00000832, 0.00000022, 0.00000210, -0.00000365, -0.00000438, 0.00000003, 0.00000221, 0.00000289, 0.00000339, 0.00000190, 0.00000068, 0.00000020, -0.00000238, -0.00000313, -0.00000123, -0.00000554, -0.00000893, 0.00000414, 0.00001260, -0.00000951, -0.00002683, -0.00000230, 0.00001991, -0.00000049, -0.00001454, 0.00000979, 0.00001882, -0.00000381, -0.00000221, 0.00001915, 0.00000715, -0.00000827, 0.00001886, 0.00002911, -0.00001781, -0.00003530, 0.00001817, 0.00004262, -0.00000896, -0.00004352, -0.00000990, 0.00002669, 0.00001972, -0.00000575, -0.00001984, -0.00001104, 0.00001091, 0.00001149, -0.00000914, -0.00000347, 0.00002211, 0.00000811, -0.00002584, -0.00000960, 0.00002559, 0.00000705, -0.00002415, -0.00000125, 0.00002594, 0.00000039, -0.00002434, -0.00000304, 0.00001465, -0.00000151, -0.00001074, 0.00000082, 0.00000209, -0.00000703, -0.00000383, 0.00000453, 0.00000351, 0.00000124, 0.00000334, 0.00000279, -0.00000010, -0.00000092, -0.00000159, -0.00000166, -0.00000006, -0.00000260, -0.00000839, -0.00000606, 0.00000438, 0.00000945, 0.00000497, -0.00000153, -0.00000464, -0.00000295, 0.00000092, -0.00000021, -0.00000573, -0.00000460, 0.00000429, 0.00000849, 0.00000413, 0.00000002, 0.00000061, 0.00000258, 0.00000247, -0.00000193, -0.00000811, -0.00000810, -0.00000121, 0.00000344, 0.00000323, 0.00000322, 0.00000117, -0.00000669, -0.00001152, -0.00000508, 0.00000376, },
		{ 0.00000084, 0.00000575, 0.00000698, 0.00000254, -0.00000181, -0.00000270, -0.00000209, -0.00000165, -0.00000153, -0.00000292, -0.00000420, 0.00000054, 0.00000882, 0.00000747, -0.00000422, -0.00000882, -0.00000130, 0.00000252, -0.00000272, -0.00000485, -0.00000079, 0.00000176, 0.00000261, 0.00000386, 0.00000235, -0.00000009, 0.00000027, -0.00000069, -0.00000340, -0.00000286, -0.00000404, -0.00000788, -0.00000010, 0.00001150, -0.00000221, -0.00002504, -0.00001229, 0.00001568, 0.00000731, -0.00001291, 0.00000212, 0.00001987, 0.00000473, -0.00000289, 0.00001438, 0.00001093, -0.00000522, 0.00001441, 0.00003071, -0.00000929, -0.00003665, 0.00001020, 0.00004489, -0.00000269, -0.00004685, -0.00001434, 0.00002964, 0.00002088, -0.00000844, -0.00001791, -0.00000892, 0.00000641, 0.00000930, -0.00000574, -0.00000541, 0.00001754, 0.00001281, -0.00002129, -0.00001753, 0.00002000, 0.00001543, -0.00001988, -0.00001076, 0.00002209, 0.00000967, -0.00002028, -0.00000959, 0.00001219, 0.00000289, -0.00000955, -0.00000129, 0.00000251, -0.00000591, -0.00000486, 0.00000365, 0.00000383, 0.00000142, 0.00000352, 0.00000289, -0.00000050, -0.00000042, -0.00000035, -0.00000224, -0.00000122, -0.00000107, -0.00000666, -0.00000819, 0.00000107, 0.00000949, 0.00000724, -0.00000015, -0.00000462, -0.00000368, 0.00000049, 0.00000078, -0.00000472, -0.00000588, 0.00000207, 0.00000818, 0.00000533, 0.00000074, 0.00000028, 0.00000184, 0.00000267, 0.00000006, -0.00000643, -0.00000921, -0.00000335, 0.00000299, 0.00000340, 0.00000302, 0.00000257, -0.00000417, -0.00001170, -0.00000798, 0.00000225, 0.00000486, 0.00000070, },
		{ 0.00000707, 0.00000320, -0.00000106, -0.00000225, -0.00000215, -0.00000193, -0.00000142, -0.00000239, -0.00000441, -0.00000133, 0.00000719, 0.00000879, -0.00000152, -0.00000860, -0.00000288, 0.00000252, -0.00000166, -0.00000506, -0.00000166, 0.00000129, 0.00000221, 0.00000408, 0.00000309, -0.00000053, -0.00000026, 0.00000074, -0.00000274, -0.00000456, -0.00000364, -0.00000600, -0.00000299, 0.00000846, 0.00000336, -0.00002015, -0.00001988, 0.00000797, 0.00001255, -0.00000820, -0.00000419, 0.00001755, 0.00001274, -0.00000038, 0.00000939, 0.00001245, -0.00000079, 0.00001131, 0.00003001, -0.00000110, -0.00003504, 0.00000192, 0.00004439, 0.00000468, -0.00004810, -0.00002063, 0.00003131, 0.00002416, -0.00001102, -0.00001815, -0.00000580, 0.00000393, 0.00000557, -0.00000340, -0.00000535, 0.00001263, 0.00001506, -0.00001497, -0.00002253, 0.00001174, 0.00002088, -0.00001231, -0.00001756, 0.00001459, 0.00001629, -0.00001319, -0.00001408, 0.00000752, 0.00000606, -0.00000688, -0.00000293, 0.00000203, -0.00000473, -0.00000529, 0.00000263, 0.00000379, 0.00000160, 0.00000388, 0.00000318, -0.00000111, -0.00000037, 0.00000115, -0.00000219, -0.00000280, -0.00000027, -0.00000428, -0.00000931, -0.00000265, 0.00000853, 0.00000935, 0.00000161, -0.00000432, -0.00000425, -0.00000019, 0.00000139, -0.00000336, -0.00000646, -0.00000029, 0.00000718, 0.00000636, 0.00000173, 0.00000022, 0.00000110, 0.00000241, 0.00000162, -0.00000430, -0.00000956, -0.00000559, 0.00000208, 0.00000365, 0.00000271, 0.00000335, -0.00000147, -0.00001082, -0.00001065, -0.00000002, 0.00000545, 0.00000170, 0.00000034, 0.00000498, },
		{ -0.00000056, -0.00000170, -0.00000190, -0.00000218, -0.00000154, -0.00000189, -0.00000425, -0.00000286, 0.00000518, 0.00000933, 0.00000115, -0.00000768, -0.00000429, 0.00000210, -0.00000061, -0.00000498, -0.00000254, 0.00000080, 0.00000177, 0.00000397, 0.00000392, -0.00000047, -0.00000115, 0.00000160, -0.00000138, -0.00000580, -0.00000434, -0.00000401, -0.00000419, 0.00000448, 0.00000645, -0.00001346, -0.00002398, -0.00000171, 0.00001405, -0.00000175, -0.00000789, 0.00001273, 0.00001888, 0.00000492, 0.00000547, 0.00001160, 0.00000396, 0.00001020, 0.00002775, 0.00000569, -0.00003080, -0.00000543, 0.00004104, 0.00001194, -0.00004673, -0.00002782, 0.00003083, 0.00002900, -0.00001240, -0.00002049, -0.00000282, 0.00000382, 0.00000125, -0.00000282, -0.00000381, 0.00000844, 0.00001479, -0.00000827, -0.00002404, 0.00000237, 0.00002244, -0.00000300, -0.00002043, 0.00000496, 0.00001900, -0.00000437, -0.00001548, 0.00000154, 0.00000721, -0.00000324, -0.00000351, 0.00000085, -0.00000391, -0.00000515, 0.00000175, 0.00000338, 0.00000159, 0.00000437, 0.00000377, -0.00000176, -0.00000087, 0.00000269, -0.00000139, -0.00000450, -0.00000038, -0.00000160, -0.00000923, -0.00000639, 0.00000651, 0.00001108, 0.00000366, -0.00000373, -0.00000459, -0.00000098, 0.00000150, -0.00000190, -0.00000625, -0.00000249, 0.00000549, 0.00000701, 0.00000297, 0.00000045, 0.00000046, 0.00000185, 0.00000264, -0.00000200, -0.00000907, -0.00000768, 0.00000068, 0.00000388, 0.00000245, 0.00000351, 0.00000108, -0.00000894, -0.00001271, -0.00000295, 0.00000548, 0.00000293, 0.00000009, 0.00000417, 0.00000714, 0.00000372, },
		{ -0.00000136, -0.00000228, -0.00000184, -0.00000149, -0.00000378, -0.00000397, 0.00000300, 0.00000910, 0.00000356, -0.00000616, -0.00000536, 0.00000132, 0.00000032, -0.00000461, -0.00000337, 0.00000026, 0.00000138, 0.00000359, 0.00000461, 0.00000010, -0.00000205, 0.00000171, 0.00000032, -0.00000621, -0.00000581, -0.00000258, -0.00000383, 0.00000061, 0.00000690, -0.00000646, -0.00002424, -0.00001151, 0.00001144, 0.00000470, -0.00000845, 0.00000675, 0.00002224, 0.00001203, 0.00000367, 0.00000885, 0.00000797, 0.00001123, 0.00002495, 0.00001042, -0.00002475, -0.00001080, 0.00003533, 0.00001789, -0.00004279, -0.00003471, 0.00002777, 0.00003440, -0.00001178, -0.00002430, -0.00000106, 0.00000587, -0.00000254, -0.00000428, -0.00000167, 0.00000575, 0.00001254, -0.00000250, -0.00002222, -0.00000643, 0.00001994, 0.00000618, -0.00001892, -0.00000487, 0.00001733, 0.00000446, -0.00001335, -0.00000447, 0.00000593, 0.00000055, -0.00000265, -0.00000058, -0.00000377, -0.00000465, 0.00000125, 0.00000274, 0.00000125, 0.00000486, 0.00000474, -0.00000228, -0.00000194, 0.00000398, 0.00000018, -0.00000599, -0.00000145, 0.00000095, -0.00000792, -0.00000974, 0.00000350, 0.00001219, 0.00000594, -0.00000286, -0.00000467, -0.00000170, 0.00000112, -0.00000064, -0.00000528, -0.00000420, 0.00000321, 0.00000707, 0.00000433, 0.00000101, -0.00000003, 0.00000110, 0.00000313, 0.00000022, -0.00000784, -0.00000933, -0.00000116, 0.00000392, 0.00000236, 0.00000319, 0.00000315, -0.00000626, -0.00001384, -0.00000629, 0.00000473, 0.00000426, 0.00000015, 0.00000331, 0.00000717, 0.00000420, -0.00000031, -0.00000123, },
		{ -0.00000226, -0.00000130, -0.00000311, -0.00000462, 0.00000083, 0.00000821, 0.00000550, -0.00000420, -0.00000594, 0.00000031, 0.00000101, -0.00000401, -0.00000410, -0.00000036, 0.00000107, 0.00000302, 0.00000501, 0.00000107, -0.00000266, 0.00000112, 0.00000191, -0.00000565, -0.00000755, -0.00000215, -0.00000239, -0.00000227, 0.00000514, -0.00000056, -0.00002108, -0.00001967, 0.00000519, 0.00000943, -0.00000612, 0.00000108, 0.00002257, 0.00001966, 0.00000463, 0.00000511, 0.00001037, 0.00001403, 0.00002267, 0.00001293, -0.00001801, -0.00001359, 0.00002820, 0.00002160, -0.00003686, -0.00004016, 0.00002229, 0.00003921, -0.00000883, -0.00002859, -0.00000126, 0.00000937, -0.00000482, -0.00000749, 0.00000006, 0.00000488, 0.00000921, 0.00000139, -0.00001787, -0.00001317, 0.00001403, 0.00001341, -0.00001349, -0.00001297, 0.00001163, 0.00001153, -0.00000799, -0.00000912, 0.00000233, 0.00000352, -0.00000034, -0.00000167, -0.00000443, -0.00000413, 0.00000128, 0.00000207, 0.00000049, 0.00000517, 0.00000609, -0.00000245, -0.00000351, 0.00000477, 0.00000239, -0.00000692, -0.00000337, 0.00000297, -0.00000550, -0.00001229, -0.00000032, 0.00001244, 0.00000831, -0.00000171, -0.00000450, -0.00000219, 0.00000036, 0.00000017, -0.00000375, -0.00000515, 0.00000057, 0.00000638, 0.00000567, 0.00000192, -0.00000032, 0.00000028, 0.00000314, 0.00000211, -0.00000602, -0.00001035, -0.00000329, 0.00000361, 0.00000248, 0.00000256, 0.00000455, -0.00000309, -0.00001384, -0.00000971, 0.00000311, 0.00000547, 0.00000058, 0.00000244, 0.00000713, 0.00000477, -0.00000026, -0.00000101, -0.00000063, -0.00000209, },
		{ -0.00000236, -0.00000484, -0.00000117, 0.00000677, 0.00000684, -0.00000203, -0.00000597, -0.00000077, 0.00000141, -0.00000324, -0.00000467, -0.00000107, 0.00000085, 0.00000240, 0.00000500, 0.00000223, -0.00000275, 0.00000004, 0.00000298, -0.00000425, -0.00000899, -0.00000283, -0.00000058, -0.00000365, 0.00000202, 0.00000326, -0.00001556, -0.00002483, -0.00000347, 0.00001109, -0.00000187, -0.00000300, 0.00002027, 0.00002648, 0.00000835, 0.00000155, 0.00001070, 0.00001781, 0.00002178, 0.00001360, -0.00001168, -0.00001371, 0.00002082, 0.00002258, -0.00002995, -0.00004333, 0.00001504, 0.00004236, -0.00000379, -0.00003224, -0.00000368, 0.00001331, -0.00000497, -0.00001173, 0.00000050, 0.00000564, 0.00000585, 0.00000295, -0.00001218, -0.00001687, 0.00000597, 0.00001736, -0.00000536, -0.00001776, 0.00000307, 0.00001538, -0.00000035, -0.00001126, -0.00000292, 0.00000479, 0.00000308, -0.00000183, -0.00000576, -0.00000399, 0.00000183, 0.00000167, -0.00000066, 0.00000508, 0.00000772, -0.00000207, -0.00000541, 0.00000481, 0.00000504, -0.00000700, -0.00000590, 0.00000408, -0.00000225, -0.00001369, -0.00000466, 0.00001168, 0.00001059, -0.00000031, -0.00000411, -0.00000235, -0.00000060, 0.00000035, -0.00000193, -0.00000515, -0.00000210, 0.00000485, 0.00000677, 0.00000317, -0.00000038, -0.00000051, 0.00000280, 0.00000355, -0.00000385, -0.00001058, -0.00000548, 0.00000284, 0.00000275, 0.00000187, 0.00000518, 0.00000017, -0.00001262, -0.00001280, 0.00000061, 0.00000632, 0.00000141, 0.00000163, 0.00000695, 0.00000547, -0.00000027, -0.00000114, 0.00000013, -0.00000155, -0.00000265, -0.00000135, },
		{ -0.00000289, 0.00000496, 0.00000755, 0.00000014, -0.00000546, -0.00000176, 0.00000151, -0.00000239, -0.00000503, -0.00000187, 0.00000066, 0.00000184, 0.00000461, 0.00000333, -0.00000224, -0.00000119, 0.00000329, -0.00000235, -0.00000968, -0.00000441, 0.00000089, -0.00000345, -0.00000141, 0.00000456, -0.00000911, -0.00002632, -0.00001284, 0.00000904, 0.00000284, -0.00000475, 0.00001630, 0.00003144, 0.00001430, -0.00000071, 0.00000905, 0.00002156, 0.00002276, 0.00001321, -0.00000666, -0.00001165, 0.00001434, 0.00002090, -0.00002322, -0.00004387, 0.00000707, 0.00004313, 0.00000261, -0.00003427, -0.00000802, 0.00001657, -0.00000285, -0.00001602, -0.00000085, 0.00000740, 0.00000339, 0.00000231, -0.00000651, -0.00001723, -0.00000263, 0.00001736, 0.00000375, -0.00001839, -0.00000664, 0.00001515, 0.00000808, -0.00001015, -0.00000875, 0.00000379, 0.00000689, -0.00000059, -0.00000738, -0.00000457, 0.00000269, 0.00000179, -0.00000206, 0.00000441, 0.00000945, -0.00000099, -0.00000739, 0.00000396, 0.00000780, -0.00000605, -0.00000868, 0.00000405, 0.00000142, -0.00001373, -0.00000911, 0.00000984, 0.00001259, 0.00000130, -0.00000358, -0.00000212, -0.00000152, -0.00000019, -0.00000017, -0.00000414, -0.00000445, 0.00000250, 0.00000739, 0.00000471, -0.00000016, -0.00000125, 0.00000223, 0.00000448, -0.00000163, -0.00001002, -0.00000746, 0.00000157, 0.00000301, 0.00000130, 0.00000511, 0.00000313, -0.00001030, -0.00001514, -0.00000263, 0.00000654, 0.00000256, 0.00000100, 0.00000656, 0.00000629, -0.00000018, -0.00000164, 0.00000072, -0.00000067, -0.00000288, -0.00000163, -0.00000165, -0.00000464, },
		{ 0.00000762, 0.00000213, -0.00000450, -0.00000252, 0.00000135, -0.00000156, -0.00000515, -0.00000273, 0.00000044, 0.00000145, 0.00000393, 0.00000413, -0.00000119, -0.00000222, 0.00000279, -0.00000041, -0.00000938, -0.00000645, 0.00000149, -0.00000203, -0.00000420, 0.00000352, -0.00000320, -0.00002426, -0.00002112, 0.00000344, 0.00000638, -0.00000410, 0.00001189, 0.00003399, 0.00002150, -0.00000085, 0.00000597, 0.00002435, 0.00002555, 0.00001276, -0.00000340, -0.00000826, 0.00000961, 0.00001716, -0.00001779, -0.00004197, -0.00000045, 0.00004127, 0.00000933, -0.00003406, -0.00001353, 0.00001822, 0.00000113, -0.00001933, -0.00000406, 0.00000929, 0.00000243, 0.00000011, -0.00000202, -0.00001462, -0.00001015, 0.00001357, 0.00001197, -0.00001481, -0.00001558, 0.00001076, 0.00001559, -0.00000571, -0.00001383, 0.00000039, 0.00001018, 0.00000218, -0.00000870, -0.00000603, 0.00000351, 0.00000258, -0.00000343, 0.00000304, 0.00001101, 0.00000085, -0.00000913, 0.00000218, 0.00001029, -0.00000402, -0.00001128, 0.00000279, 0.00000504, -0.00001233, -0.00001323, 0.00000693, 0.00001408, 0.00000309, -0.00000296, -0.00000152, -0.00000216, -0.00000137, 0.00000115, -0.00000223, -0.00000607, -0.00000052, 0.00000730, 0.00000644, 0.00000040, -0.00000189, 0.00000156, 0.00000493, 0.00000039, -0.00000878, -0.00000895, -0.00000014, 0.00000309, 0.00000098, 0.00000450, 0.00000546, -0.00000713, -0.00001638, -0.00000632, 0.00000590, 0.00000389, 0.00000067, 0.00000594, 0.00000715, 0.00000018, -0.00000241, 0.00000096, 0.00000046, -0.00000281, -0.00000208, -0.00000107, -0.00000412, -0.00000425, 0.00000293, },
		{ -0.00000322, -0.00000294, 0.00000102, -0.00000085, -0.00000505, -0.00000358, 0.00000013, 0.00000123, 0.00000311, 0.00000450, 0.00000020, -0.00000275, 0.00000164, 0.00000112, -0.00000811, -0.00000835, 0.00000096, -0.00000001, -0.00000567, 0.00000084, 0.00000102, -0.00001949, -0.00002683, -0.00000472, 0.00000737, -0.00000173, 0.00000821, 0.00003417, 0.00002880, 0.00000150, 0.00000239, 0.00002550, 0.00002969, 0.00001319, -0.00000185, -0.00000457, 0.00000703, 0.00001231, -0.00001440, -0.00003832, -0.00000652, 0.00003711, 0.00001526, -0.00003148, -0.00001917, 0.00001772, 0.00000617, -0.00002083, -0.00000869, 0.00001037, 0.00000311, -0.00000275, 0.00000052, -0.00000997, -0.00001530, 0.00000686, 0.00001769, -0.00000782, -0.00002198, 0.00000295, 0.00002056, 0.00000143, -0.00001688, -0.00000500, 0.00001201, 0.00000626, -0.00000911, -0.00000830, 0.00000388, 0.00000405, -0.00000445, 0.00000098, 0.00001209, 0.00000339, -0.00001029, -0.00000043, 0.00001213, -0.00000103, -0.00001327, 0.00000040, 0.00000811, -0.00000961, -0.00001659, 0.00000313, 0.00001486, 0.00000497, -0.00000229, -0.00000064, -0.00000230, -0.00000301, 0.00000171, 0.00000033, -0.00000666, -0.00000392, 0.00000631, 0.00000817, 0.00000133, -0.00000242, 0.00000087, 0.00000503, 0.00000202, -0.00000706, -0.00000974, -0.00000210, 0.00000280, 0.00000095, 0.00000362, 0.00000696, -0.00000352, -0.00001631, -0.00001006, 0.00000427, 0.00000515, 0.00000076, 0.00000514, 0.00000791, 0.00000089, -0.00000329, 0.00000069, 0.00000166, -0.00000235, -0.00000259, -0.00000070, -0.00000336, -0.00000520, 0.00000083, 0.00000712, 0.00000378, },
		{ 0.00000063, -0.00000030, -0.00000476, -0.00000434, -0.00000033, 0.00000115, 0.00000232, 0.00000439, 0.00000164, -0.00000262, 0.00000018, 0.00000191, -0.00000617, -0.00000961, -0.00000058, 0.00000189, -0.00000559, -0.00000251, 0.00000287, -0.00001331, -0.00002904, -0.00001392, 0.00000502, 0.00000114, 0.00000604, 0.00003257, 0.00003518, 0.00000619, -0.00000065, 0.00002476, 0.00003435, 0.00001518, -0.00000151, -0.00000154, 0.00000647, 0.00000741, -0.00001335, -0.00003391, -0.00001046, 0.00003140, 0.00001947, -0.00002693, -0.00002389, 0.00001503, 0.00001123, -0.00002012, -0.00001394, 0.00000996, 0.00000509, -0.00000528, 0.00000086, -0.00000450, -0.00001735, -0.00000136, 0.00001983, 0.00000112, -0.00002461, -0.00000684, 0.00002185, 0.00001006, -0.00001697, -0.00001150, 0.00001165, 0.00001107, -0.00000808, -0.00001103, 0.00000340, 0.00000602, -0.00000479, -0.00000160, 0.00001239, 0.00000643, -0.00001059, -0.00000363, 0.00001299, 0.00000265, -0.00001429, -0.00000287, 0.00001022, -0.00000584, -0.00001882, -0.00000131, 0.00001477, 0.00000684, -0.00000163, 0.00000039, -0.00000180, -0.00000482, 0.00000131, 0.00000314, -0.00000600, -0.00000734, 0.00000432, 0.00000970, 0.00000266, -0.00000282, 0.00000021, 0.00000492, 0.00000315, -0.00000515, -0.00000973, -0.00000405, 0.00000201, 0.00000110, 0.00000274, 0.00000756, 0.00000008, -0.00001489, -0.00001336, 0.00000167, 0.00000604, 0.00000131, 0.00000427, 0.00000843, 0.00000197, -0.00000406, -0.00000011, 0.00000270, -0.00000148, -0.00000302, -0.00000057, -0.00000247, -0.00000572, -0.00000120, 0.00000616, 0.00000501, -0.00000179, -0.00000300, },
		{ -0.00000438, -0.00000495, -0.00000092, 0.00000112, 0.00000169, 0.00000389, 0.00000286, -0.00000185, -0.00000122, 0.00000183, -0.00000400, -0.00000989, -0.00000275, 0.00000303, -0.00000415, -0.00000553, 0.00000231, -0.00000721, -0.00002765, -0.00002237, -0.00000068, 0.00000305, 0.00000559, 0.00003012, 0.00003993, 0.00001258, -0.00000226, 0.00002236, 0.00003863, 0.00001898, -0.00000153, 0.00000017, 0.00000736, 0.00000342, -0.00001440, -0.00002981, -0.00001211, 0.00002516, 0.00002140, -0.00002121, -0.00002686, 0.00001059, 0.00001530, -0.00001725, -0.00001884, 0.00000772, 0.00000769, -0.00000671, -0.00000069, 0.00000053, -0.00001624, -0.00000948, 0.00001806, 0.00001022, -0.00002297, -0.00001681, 0.00001897, 0.00001860, -0.00001369, -0.00001788, 0.00000874, 0.00001574, -0.00000530, -0.00001368, 0.00000182, 0.00000811, -0.00000420, -0.00000440, 0.00001169, 0.00000965, -0.00000981, -0.00000705, 0.00001266, 0.00000664, -0.00001406, -0.00000663, 0.00001106, -0.00000144, -0.00001965, -0.00000601, 0.00001372, 0.00000856, -0.00000098, 0.00000143, -0.00000063, -0.00000646, -0.00000011, 0.00000576, -0.00000407, -0.00001031, 0.00000134, 0.00001075, 0.00000436, -0.00000304, -0.00000042, 0.00000476, 0.00000380, -0.00000335, -0.00000894, -0.00000569, 0.00000067, 0.00000124, 0.00000209, 0.00000739, 0.00000321, -0.00001228, -0.00001575, -0.00000175, 0.00000624, 0.00000227, 0.00000350, 0.00000859, 0.00000332, -0.00000447, -0.00000138, 0.00000335, -0.00000025, -0.00000321, -0.00000067, -0.00000156, -0.00000584, -0.00000308, 0.00000484, 0.00000577, -0.00000036, -0.00000270, 0.00000031, 0.00000007, },
		{ -0.00000160, 0.00000105, 0.00000127, 0.00000316, 0.00000367, -0.00000062, -0.00000221, 0.00000098, -0.00000209, -0.00000913, -0.00000499, 0.00000309, -0.00000192, -0.00000742, -0.00000012, -0.00000248, -0.00002326, -0.00002850, -0.00000896, 0.00000269, 0.00000639, 0.00002778, 0.00004288, 0.00001981, -0.00000185, 0.00001892, 0.00004174, 0.00002441, -0.00000101, 0.00000036, 0.00000883, 0.00000094, -0.00001689, -0.00002693, -0.00001181, 0.00001943, 0.00002102, -0.00001534, -0.00002765, 0.00000523, 0.00001762, -0.00001277, -0.00002248, 0.00000378, 0.00001004, -0.00000660, -0.00000344, 0.00000409, -0.00001254, -0.00001601, 0.00001284, 0.00001772, -0.00001742, -0.00002511, 0.00001226, 0.00002538, -0.00000729, -0.00002282, 0.00000343, 0.00001927, -0.00000084, -0.00001556, -0.00000089, 0.00000985, -0.00000261, -0.00000702, 0.00000991, 0.00001266, -0.00000788, -0.00001027, 0.00001107, 0.00001049, -0.00001251, -0.00001039, 0.00001050, 0.00000309, -0.00001898, -0.00001056, 0.00001170, 0.00000998, -0.00000037, 0.00000231, 0.00000115, -0.00000759, -0.00000245, 0.00000773, -0.00000099, -0.00001241, -0.00000246, 0.00001107, 0.00000636, -0.00000304, -0.00000107, 0.00000467, 0.00000406, -0.00000193, -0.00000753, -0.00000672, -0.00000111, 0.00000115, 0.00000182, 0.00000670, 0.00000552, -0.00000883, -0.00001685, -0.00000561, 0.00000549, 0.00000347, 0.00000304, 0.00000835, 0.00000476, -0.00000436, -0.00000293, 0.00000340, 0.00000115, -0.00000303, -0.00000092, -0.00000074, -0.00000560, -0.00000470, 0.00000326, 0.00000608, 0.00000094, -0.00000213, 0.00000015, 0.00000028, -0.00000397, -0.00000537, },
		{ 0.00000105, 0.00000240, 0.00000400, 0.00000079, -0.00000255, -0.00000034, -0.00000085, -0.00000755, -0.00000672, 0.00000204, 0.00000039, -0.00000781, -0.00000353, 0.00000004, -0.00001709, -0.00003123, -0.00001837, -0.00000081, 0.00000739, 0.00002625, 0.00004431, 0.00002699, 0.00000070, 0.00001529, 0.00004322, 0.00003089, 0.00000086, -0.00000073, 0.00000996, 0.00000011, -0.00001993, -0.00002577, -0.00001027, 0.00001505, 0.00001874, -0.00001031, -0.00002631, -0.00000006, 0.00001784, -0.00000753, -0.00002422, -0.00000132, 0.00001136, -0.00000500, -0.00000647, 0.00000563, -0.00000728, -0.00001989, 0.00000526, 0.00002222, -0.00000904, -0.00003023, 0.00000277, 0.00002902, 0.00000133, -0.00002518, -0.00000362, 0.00002080, 0.00000484, -0.00001604, -0.00000447, 0.00001076, -0.00000013, -0.00000902, 0.00000714, 0.00001502, -0.00000491, -0.00001285, 0.00000833, 0.00001371, -0.00000972, -0.00001365, 0.00000859, 0.00000724, -0.00001688, -0.00001454, 0.00000883, 0.00001096, 0.00000020, 0.00000291, 0.00000339, -0.00000791, -0.00000546, 0.00000865, 0.00000291, -0.00001326, -0.00000678, 0.00001044, 0.00000852, -0.00000274, -0.00000178, 0.00000472, 0.00000413, -0.00000108, -0.00000576, -0.00000692, -0.00000313, 0.00000057, 0.00000192, 0.00000584, 0.00000680, -0.00000503, -0.00001645, -0.00000944, 0.00000362, 0.00000463, 0.00000304, 0.00000777, 0.00000607, -0.00000365, -0.00000448, 0.00000277, 0.00000250, -0.00000241, -0.00000120, -0.00000007, -0.00000507, -0.00000602, 0.00000154, 0.00000598, 0.00000201, -0.00000139, 0.00000023, 0.00000043, -0.00000362, -0.00000561, -0.00000228, 0.00000086, },
		{ 0.00000389, 0.00000208, -0.00000223, -0.00000173, -0.00000048, -0.00000559, -0.00000753, 0.00000024, 0.00000212, -0.00000677, -0.00000689, 0.00000011, -0.00001062, -0.00003023, -0.00002717, -0.00000753, 0.00000726, 0.00002572, 0.00004489, 0.00003345, 0.00000511, 0.00001232, 0.00004298, 0.00003764, 0.00000455, -0.00000245, 0.00001000, 0.00000060, -0.00002258, -0.00002640, -0.00000844, 0.00001243, 0.00001536, -0.00000686, -0.00002334, -0.00000437, 0.00001608, -0.00000250, -0.00002383, -0.00000676, 0.00001110, -0.00000234, -0.00000890, 0.00000509, -0.00000169, -0.00002064, -0.00000321, 0.00002297, 0.00000055, -0.00003127, -0.00000790, 0.00002868, 0.00001077, -0.00002425, -0.00001133, 0.00001972, 0.00001098, -0.00001464, -0.00000841, 0.00001045, 0.00000290, -0.00001004, 0.00000363, 0.00001637, -0.00000117, -0.00001441, 0.00000472, 0.00001591, -0.00000597, -0.00001600, 0.00000556, 0.00001056, -0.00001359, -0.00001757, 0.00000531, 0.00001139, 0.00000070, 0.00000313, 0.00000586, -0.00000722, -0.00000875, 0.00000827, 0.00000719, -0.00001263, -0.00001122, 0.00000873, 0.00001065, -0.00000208, -0.00000258, 0.00000490, 0.00000423, -0.00000086, -0.00000400, -0.00000618, -0.00000506, -0.00000065, 0.00000225, 0.00000514, 0.00000706, -0.00000145, -0.00001455, -0.00001267, 0.00000065, 0.00000537, 0.00000361, 0.00000703, 0.00000700, -0.00000237, -0.00000572, 0.00000145, 0.00000352, -0.00000138, -0.00000136, 0.00000040, -0.00000432, -0.00000699, -0.00000025, 0.00000553, 0.00000282, -0.00000061, 0.00000053, 0.00000060, -0.00000338, -0.00000571, -0.00000289, 0.00000051, 0.00000093, 0.00000175, },
		{ -0.00000139, -0.00000280, -0.00000092, -0.00000376, -0.00000729, -0.00000177, 0.00000282, -0.00000478, -0.00000933, -0.00000188, -0.00000525, -0.00002600, -0.00003368, -0.00001676, 0.00000474, 0.00002576, 0.00004532, 0.00003893, 0.00001081, 0.00001065, 0.00004136, 0.00004388, 0.00001014, -0.00000398, 0.00000853, 0.00000176, -0.00002415, -0.00002843, -0.00000719, 0.00001155, 0.00001179, -0.00000531, -0.00001956, -0.00000705, 0.00001290, 0.00000144, -0.00002154, -0.00001163, 0.00000910, 0.00000069, -0.00001011, 0.00000290, 0.00000309, -0.00001846, -0.00001100, 0.00001995, 0.00000957, -0.00002811, -0.00001794, 0.00002425, 0.00001942, -0.00001991, -0.00001840, 0.00001590, 0.00001657, -0.00001123, -0.00001202, 0.00000875, 0.00000600, -0.00000984, -0.00000018, 0.00001644, 0.00000295, -0.00001468, 0.00000064, 0.00001681, -0.00000164, -0.00001711, 0.00000180, 0.00001269, -0.00000949, -0.00001940, 0.00000144, 0.00001117, 0.00000112, 0.00000294, 0.00000829, -0.00000547, -0.00001190, 0.00000650, 0.00001131, -0.00001044, -0.00001531, 0.00000591, 0.00001250, -0.00000102, -0.00000348, 0.00000513, 0.00000458, -0.00000121, -0.00000260, -0.00000456, -0.00000652, -0.00000257, 0.00000255, 0.00000491, 0.00000649, 0.00000140, -0.00001137, -0.00001477, -0.00000319, 0.00000530, 0.00000469, 0.00000636, 0.00000741, -0.00000072, -0.00000636, -0.00000038, 0.00000394, -0.00000003, -0.00000124, 0.00000069, -0.00000347, -0.00000758, -0.00000203, 0.00000477, 0.00000335, 0.00000011, 0.00000103, 0.00000088, -0.00000324, -0.00000576, -0.00000334, 0.00000004, 0.00000079, 0.00000132, 0.00000351, 0.00000301, },
		{ -0.00000191, -0.00000252, -0.00000617, -0.00000340, 0.00000242, -0.00000255, -0.00001035, -0.00000511, -0.00000199, -0.00001966, -0.00003670, -0.00002710, -0.00000100, 0.00002538, 0.00004615, 0.00004357, 0.00001715, 0.00001059, 0.00003895, 0.00004901, 0.00001733, -0.00000450, 0.00000558, 0.00000272, -0.00002432, -0.00003115, -0.00000717, 0.00001195, 0.00000890, -0.00000553, -0.00001587, -0.00000788, 0.00000910, 0.00000374, -0.00001795, -0.00001518, 0.00000563, 0.00000328, -0.00000984, -0.00000020, 0.00000622, -0.00001405, -0.00001681, 0.00001389, 0.00001643, -0.00002143, -0.00002564, 0.00001638, 0.00002574, -0.00001268, -0.00002355, 0.00000970, 0.00002063, -0.00000605, -0.00001457, 0.00000574, 0.00000860, -0.00000839, -0.00000382, 0.00001515, 0.00000694, -0.00001357, -0.00000340, 0.00001631, 0.00000275, -0.00001684, -0.00000222, 0.00001346, -0.00000503, -0.00001990, -0.00000245, 0.00001031, 0.00000141, 0.00000232, 0.00001042, -0.00000271, -0.00001443, 0.00000343, 0.00001474, -0.00000683, -0.00001858, 0.00000210, 0.00001381, 0.00000045, -0.00000446, 0.00000529, 0.00000534, -0.00000191, -0.00000190, -0.00000227, -0.00000716, -0.00000508, 0.00000249, 0.00000529, 0.00000547, 0.00000313, -0.00000740, -0.00001530, -0.00000749, 0.00000411, 0.00000609, 0.00000606, 0.00000722, 0.00000102, -0.00000620, -0.00000243, 0.00000357, 0.00000140, -0.00000073, 0.00000086, -0.00000261, -0.00000779, -0.00000370, 0.00000375, 0.00000361, 0.00000071, 0.00000163, 0.00000134, -0.00000315, -0.00000584, -0.00000361, -0.00000047, 0.00000051, 0.00000109, 0.00000308, 0.00000347, -0.00000031, -0.00000327, },
		{ -0.00000459, -0.00000414, 0.00000120, -0.00000079, -0.00000992, -0.00000851, -0.00000123, -0.00001274, -0.00003575, -0.00003673, -0.00001008, 0.00002328, 0.00004742, 0.00004784, 0.00002363, 0.00001209, 0.00003641, 0.00005278, 0.00002554, -0.00000337, 0.00000155, 0.00000271, -0.00002318, -0.00003373, -0.00000864, 0.00001289, 0.00000726, -0.00000702, -0.00001307, -0.00000706, 0.00000557, 0.00000423, -0.00001391, -0.00001697, 0.00000129, 0.00000479, -0.00000824, -0.00000335, 0.00000732, -0.00000852, -0.00001980, 0.00000600, 0.00001999, -0.00001254, -0.00002977, 0.00000634, 0.00002858, -0.00000368, -0.00002580, 0.00000198, 0.00002232, 0.00000029, -0.00001542, 0.00000181, 0.00001019, -0.00000589, -0.00000676, 0.00001264, 0.00001031, -0.00001118, -0.00000691, 0.00001448, 0.00000672, -0.00001525, -0.00000600, 0.00001288, -0.00000070, -0.00001909, -0.00000601, 0.00000887, 0.00000154, 0.00000132, 0.00001199, 0.00000086, -0.00001596, -0.00000065, 0.00001700, -0.00000210, -0.00002061, -0.00000246, 0.00001436, 0.00000228, -0.00000543, 0.00000522, 0.00000657, -0.00000269, -0.00000210, 0.00000032, -0.00000670, -0.00000791, 0.00000173, 0.00000627, 0.00000444, 0.00000353, -0.00000323, -0.00001404, -0.00001166, 0.00000160, 0.00000746, 0.00000635, 0.00000653, 0.00000251, -0.00000518, -0.00000432, 0.00000232, 0.00000263, 0.00000024, 0.00000103, -0.00000184, -0.00000765, -0.00000518, 0.00000250, 0.00000364, 0.00000116, 0.00000225, 0.00000200, -0.00000303, -0.00000603, -0.00000375, -0.00000090, 0.00000005, 0.00000096, 0.00000279, 0.00000349, 0.00000067, -0.00000307, -0.00000303, -0.00000215, },
		{ -0.00000028, -0.00000002, -0.00000847, -0.00001114, -0.00000270, -0.00000677, -0.00003121, -0.00004384, -0.00002175, 0.00001818, 0.00004859, 0.00005228, 0.00003006, 0.00001484, 0.00003431, 0.00005523, 0.00003412, -0.00000030, -0.00000288, 0.00000118, -0.00002122, -0.00003544, -0.00001142, 0.00001355, 0.00000705, -0.00000900, -0.00001166, -0.00000515, 0.00000302, 0.00000319, -0.00001026, -0.00001694, -0.00000315, 0.00000484, -0.00000579, -0.00000580, 0.00000649, -0.00000301, -0.00001977, -0.00000222, 0.00001981, -0.00000310, -0.00002979, -0.00000419, 0.00002742, 0.00000562, -0.00002466, -0.00000606, 0.00002122, 0.00000688, -0.00001419, -0.00000242, 0.00001041, -0.00000276, -0.00000862, 0.00000922, 0.00001263, -0.00000782, -0.00000944, 0.00001160, 0.00000982, -0.00001255, -0.00000908, 0.00001112, 0.00000306, -0.00001716, -0.00000894, 0.00000698, 0.00000147, -0.00000001, 0.00001283, 0.00000494, -0.00001619, -0.00000533, 0.00001777, 0.00000328, -0.00002111, -0.00000740, 0.00001395, 0.00000437, -0.00000629, 0.00000476, 0.00000824, -0.00000319, -0.00000325, 0.00000276, -0.00000500, -0.00001065, 0.00000002, 0.00000763, 0.00000388, 0.00000266, 0.00000046, -0.00001101, -0.00001504, -0.00000223, 0.00000838, 0.00000740, 0.00000557, 0.00000343, -0.00000341, -0.00000566, 0.00000027, 0.00000334, 0.00000163, 0.00000134, -0.00000122, -0.00000721, -0.00000639, 0.00000106, 0.00000342, 0.00000147, 0.00000279, 0.00000282, -0.00000279, -0.00000634, -0.00000381, -0.00000115, -0.00000058, 0.00000078, 0.00000276, 0.00000326, 0.00000128, -0.00000232, -0.00000384, -0.00000264, -0.00000311, -0.00000381, },
		{ -0.00000673, -0.00001236, -0.00000559, -0.00000293, -0.00002423, -0.00004703, -0.00003450, 0.00000922, 0.00004855, 0.00005724, 0.00003661, 0.00001846, 0.00003299, 0.00005669, 0.00004249, 0.00000468, -0.00000695, -0.00000208, -0.00001914, -0.00003582, -0.00001501, 0.00001327, 0.00000803, -0.00001068, -0.00001173, -0.00000293, 0.00000188, 0.00000121, -0.00000766, -0.00001544, -0.00000692, 0.00000344, -0.00000317, -0.00000706, 0.00000424, 0.00000146, -0.00001707, -0.00000932, 0.00001617, 0.00000521, -0.00002589, -0.00001347, 0.00002245, 0.00001365, -0.00002026, -0.00001314, 0.00001734, 0.00001269, -0.00001086, -0.00000619, 0.00000913, 0.00000045, -0.00000918, 0.00000537, 0.00001361, -0.00000392, -0.00001071, 0.00000808, 0.00001174, -0.00000913, -0.00001116, 0.00000852, 0.00000592, -0.00001441, -0.00001101, 0.00000485, 0.00000121, -0.00000158, 0.00001280, 0.00000916, -0.00001499, -0.00001008, 0.00001687, 0.00000877, -0.00001995, -0.00001229, 0.00001250, 0.00000655, -0.00000691, 0.00000381, 0.00001021, -0.00000310, -0.00000524, 0.00000458, -0.00000212, -0.00001284, -0.00000272, 0.00000900, 0.00000416, 0.00000083, 0.00000303, -0.00000655, -0.00001698, -0.00000712, 0.00000834, 0.00000918, 0.00000468, 0.00000353, -0.00000118, -0.00000608, -0.00000234, 0.00000322, 0.00000330, 0.00000196, -0.00000077, -0.00000656, -0.00000721, -0.00000051, 0.00000293, 0.00000168, 0.00000321, 0.00000371, -0.00000236, -0.00000673, -0.00000393, -0.00000117, -0.00000128, 0.00000040, 0.00000298, 0.00000302, 0.00000136, -0.00000132, -0.00000400, -0.00000373, -0.00000226, -0.00000252, -0.00000135, -0.00000039, },
		{ -0.00000878, -0.00000178, -0.00001644, -0.00004567, -0.00004631, -0.00000370, 0.00004584, 0.00006258, 0.00004376, 0.00002268, 0.00003247, 0.00005759, 0.00005030, 0.00001126, -0.00000999, -0.00000690, -0.00001773, -0.00003479, -0.00001869, 0.00001164, 0.00000966, -0.00001145, -0.00001299, -0.00000115, 0.00000216, -0.00000098, -0.00000648, -0.00001310, -0.00000947, 0.00000093, -0.00000102, -0.00000700, 0.00000129, 0.00000423, -0.00001254, -0.00001422, 0.00000995, 0.00001100, -0.00001900, -0.00002004, 0.00001456, 0.00001907, -0.00001332, -0.00001806, 0.00001125, 0.00001673, -0.00000581, -0.00000876, 0.00000655, 0.00000315, -0.00000843, 0.00000162, 0.00001318, -0.00000000, -0.00001060, 0.00000442, 0.00001235, -0.00000545, -0.00001205, 0.00000550, 0.00000770, -0.00001121, -0.00001213, 0.00000269, 0.00000077, -0.00000329, 0.00001188, 0.00001315, -0.00001237, -0.00001438, 0.00001436, 0.00001379, -0.00001717, -0.00001667, 0.00001003, 0.00000862, -0.00000718, 0.00000232, 0.00001222, -0.00000216, -0.00000779, 0.00000536, 0.00000168, -0.00001400, -0.00000640, 0.00000992, 0.00000549, -0.00000148, 0.00000402, -0.00000121, -0.00001698, -0.00001259, 0.00000695, 0.00001149, 0.00000424, 0.00000274, 0.00000111, -0.00000534, -0.00000511, 0.00000206, 0.00000499, 0.00000303, -0.00000045, -0.00000583, -0.00000759, -0.00000212, 0.00000213, 0.00000179, 0.00000353, 0.00000458, -0.00000170, -0.00000713, -0.00000419, -0.00000098, -0.00000190, -0.00000028, 0.00000333, 0.00000302, 0.00000094, -0.00000046, -0.00000338, -0.00000497, -0.00000239, -0.00000071, -0.00000144, -0.00000164, -0.00000548, -0.00001206, },
		{ -0.00000959, -0.00004005, -0.00005509, -0.00001975, 0.00003899, 0.00006753, 0.00005201, 0.00002752, 0.00003260, 0.00005832, 0.00005748, 0.00001898, -0.00001157, -0.00001282, -0.00001760, -0.00003261, -0.00002174, 0.00000866, 0.00001123, -0.00001099, -0.00001486, -0.00000035, 0.00000353, -0.00000265, -0.00000667, -0.00001065, -0.00001053, -0.00000211, 0.00000014, -0.00000587, -0.00000158, 0.00000508, -0.00000724, -0.00001633, 0.00000247, 0.00001345, -0.00001054, -0.00002295, 0.00000513, 0.00002098, -0.00000506, -0.00002003, 0.00000391, 0.00001831, 0.00000023, -0.00000955, 0.00000311, 0.00000484, -0.00000662, -0.00000152, 0.00001147, 0.00000341, -0.00000921, 0.00000112, 0.00001171, -0.00000199, -0.00001176, 0.00000249, 0.00000836, -0.00000797, -0.00001231, 0.00000074, 0.00000023, -0.00000505, 0.00001008, 0.00001652, -0.00000852, -0.00001774, 0.00001046, 0.00001781, -0.00001303, -0.00002011, 0.00000666, 0.00001038, -0.00000699, 0.00000032, 0.00001398, -0.00000024, -0.00001049, 0.00000480, 0.00000597, -0.00001372, -0.00001072, 0.00000993, 0.00000784, -0.00000365, 0.00000322, 0.00000424, -0.00001476, -0.00001794, 0.00000391, 0.00001394, 0.00000462, 0.00000116, 0.00000300, -0.00000341, -0.00000756, -0.00000021, 0.00000638, 0.00000462, -0.00000018, -0.00000516, -0.00000748, -0.00000362, 0.00000096, 0.00000178, 0.00000380, 0.00000534, -0.00000085, -0.00000742, -0.00000464, -0.00000066, -0.00000229, -0.00000125, 0.00000362, 0.00000341, 0.00000021, -0.00000010, -0.00000210, -0.00000583, -0.00000358, 0.00000098, -0.00000026, -0.00000318, -0.00000537, -0.00001066, -0.00001117, -0.00000316, },
		{ -0.00005913, -0.00003721, 0.00002699, 0.00007070, 0.00006163, 0.00003337, 0.00003314, 0.00005906, 0.00006416, 0.00002743, -0.00001153, -0.00001923, -0.00001911, -0.00002985, -0.00002363, 0.00000466, 0.00001208, -0.00000934, -0.00001667, -0.00000080, 0.00000540, -0.00000332, -0.00000787, -0.00000877, -0.00001023, -0.00000500, 0.00000013, -0.00000413, -0.00000376, 0.00000422, -0.00000224, -0.00001562, -0.00000486, 0.00001238, -0.00000213, -0.00002199, -0.00000418, 0.00001915, 0.00000306, -0.00001876, -0.00000347, 0.00001714, 0.00000632, -0.00000829, -0.00000050, 0.00000521, -0.00000420, -0.00000365, 0.00000883, 0.00000588, -0.00000682, -0.00000138, 0.00001005, 0.00000085, -0.00001049, -0.00000011, 0.00000805, -0.00000502, -0.00001168, -0.00000084, -0.00000033, -0.00000674, 0.00000753, 0.00001898, -0.00000374, -0.00001975, 0.00000558, 0.00002042, -0.00000791, -0.00002229, 0.00000266, 0.00001161, -0.00000632, -0.00000206, 0.00001520, 0.00000263, -0.00001285, 0.00000280, 0.00001019, -0.00001181, -0.00001520, 0.00000866, 0.00001095, -0.00000505, 0.00000069, 0.00000898, -0.00001035, -0.00002241, -0.00000079, 0.00001600, 0.00000605, -0.00000094, 0.00000402, -0.00000042, -0.00000919, -0.00000350, 0.00000710, 0.00000672, 0.00000019, -0.00000467, -0.00000690, -0.00000483, -0.00000058, 0.00000159, 0.00000411, 0.00000594, 0.00000008, -0.00000749, -0.00000527, -0.00000036, -0.00000236, -0.00000240, 0.00000364, 0.00000423, -0.00000051, -0.00000044, -0.00000048, -0.00000588, -0.00000554, 0.00000189, 0.00000213, -0.00000422, -0.00000663, -0.00000893, -0.00001198, -0.00000622, -0.00000513, -0.00003138, },
		{ 0.00000964, 0.00007025, 0.00007237, 0.00004089, 0.00003404, 0.00005976, 0.00007059, 0.00003634, -0.00000996, -0.00002555, -0.00002231, -0.00002716, -0.00002411, 0.00000021, 0.00001173, -0.00000688, -0.00001778, -0.00000239, 0.00000711, -0.00000280, -0.00000949, -0.00000792, -0.00000898, -0.00000715, -0.00000095, -0.00000237, -0.00000492, 0.00000218, 0.00000162, -0.00001263, -0.00001078, 0.00000831, 0.00000472, -0.00001763, -0.00001182, 0.00001401, 0.00000962, -0.00001459, -0.00000966, 0.00001340, 0.00001143, -0.00000508, -0.00000355, 0.00000421, -0.00000172, -0.00000459, 0.00000571, 0.00000715, -0.00000386, -0.00000280, 0.00000776, 0.00000277, -0.00000855, -0.00000200, 0.00000702, -0.00000264, -0.00001046, -0.00000193, -0.00000080, -0.00000826, 0.00000438, 0.00002027, 0.00000155, -0.00002018, 0.00000023, 0.00002137, -0.00000231, -0.00002300, -0.00000168, 0.00001216, -0.00000516, -0.00000464, 0.00001560, 0.00000628, -0.00001440, -0.00000053, 0.00001371, -0.00000824, -0.00001924, 0.00000589, 0.00001436, -0.00000518, -0.00000317, 0.00001224, -0.00000412, -0.00002521, -0.00000693, 0.00001709, 0.00000859, -0.00000310, 0.00000384, 0.00000325, -0.00000953, -0.00000751, 0.00000678, 0.00000921, 0.00000084, -0.00000449, -0.00000595, -0.00000555, -0.00000248, 0.00000108, 0.00000452, 0.00000642, 0.00000095, -0.00000728, -0.00000597, -0.00000024, -0.00000210, -0.00000353, 0.00000323, 0.00000532, -0.00000089, -0.00000149, 0.00000100, -0.00000489, -0.00000779, 0.00000155, 0.00000526, -0.00000404, -0.00000905, -0.00000783, -0.00001098, -0.00000972, -0.00000386, -0.00002152, -0.00005751, -0.00005371, },
		{ 0.00008321, 0.00005077, 0.00003557, 0.00006020, 0.00007693, 0.00004567, -0.00000709, -0.00003138, -0.00002694, -0.00002513, -0.00002329, -0.00000406, 0.00001003, -0.00000420, -0.00001781, -0.00000469, 0.00000810, -0.00000124, -0.00001091, -0.00000822, -0.00000736, -0.00000821, -0.00000270, -0.00000109, -0.00000504, -0.00000036, 0.00000383, -0.00000820, -0.00001440, 0.00000225, 0.00000890, -0.00001095, -0.00001659, 0.00000661, 0.00001348, -0.00000836, -0.00001363, 0.00000776, 0.00001469, -0.00000039, -0.00000535, 0.00000210, 0.00000028, -0.00000437, 0.00000264, 0.00000714, -0.00000083, -0.00000303, 0.00000529, 0.00000365, -0.00000633, -0.00000306, 0.00000560, -0.00000096, -0.00000894, -0.00000248, -0.00000108, -0.00000948, 0.00000085, 0.00002030, 0.00000689, -0.00001896, -0.00000502, 0.00002062, 0.00000322, -0.00002220, -0.00000596, 0.00001193, -0.00000361, -0.00000719, 0.00001500, 0.00001038, -0.00001475, -0.00000487, 0.00001599, -0.00000327, -0.00002224, 0.00000165, 0.00001747, -0.00000373, -0.00000777, 0.00001343, 0.00000329, -0.00002573, -0.00001399, 0.00001667, 0.00001208, -0.00000483, 0.00000230, 0.00000709, -0.00000827, -0.00001179, 0.00000515, 0.00001185, 0.00000197, -0.00000465, -0.00000481, -0.00000558, -0.00000459, 0.00000007, 0.00000505, 0.00000689, 0.00000163, -0.00000683, -0.00000659, -0.00000038, -0.00000162, -0.00000443, 0.00000235, 0.00000644, -0.00000071, -0.00000304, 0.00000188, -0.00000294, -0.00000966, -0.00000021, 0.00000839, -0.00000217, -0.00001197, -0.00000818, -0.00000852, -0.00001231, -0.00000575, -0.00001255, -0.00005038, -0.00006666, -0.00001222, 0.00006431, },
		{ 0.00003839, 0.00006014, 0.00008316, 0.00005555, -0.00000313, -0.00003653, -0.00003263, -0.00002417, -0.00002152, -0.00000761, 0.00000716, -0.00000189, -0.00001667, -0.00000716, 0.00000803, 0.00000094, -0.00001160, -0.00000949, -0.00000600, -0.00000811, -0.00000460, -0.00000061, -0.00000437, -0.00000276, 0.00000434, -0.00000336, -0.00001539, -0.00000452, 0.00000987, -0.00000339, -0.00001788, -0.00000166, 0.00001405, -0.00000132, -0.00001478, 0.00000122, 0.00001554, 0.00000500, -0.00000543, -0.00000064, 0.00000136, -0.00000325, 0.00000007, 0.00000600, 0.00000179, -0.00000216, 0.00000308, 0.00000354, -0.00000422, -0.00000327, 0.00000413, -0.00000001, -0.00000737, -0.00000256, -0.00000108, -0.00001032, -0.00000284, 0.00001903, 0.00001180, -0.00001620, -0.00000963, 0.00001831, 0.00000817, -0.00002001, -0.00000983, 0.00001093, -0.00000178, -0.00000948, 0.00001333, 0.00001450, -0.00001364, -0.00000972, 0.00001660, 0.00000264, -0.00002366, -0.00000377, 0.00001966, -0.00000065, -0.00001231, 0.00001224, 0.00001099, -0.00002359, -0.00002123, 0.00001434, 0.00001613, -0.00000562, -0.00000054, 0.00001047, -0.00000531, -0.00001575, 0.00000208, 0.00001430, 0.00000374, -0.00000512, -0.00000371, -0.00000478, -0.00000669, -0.00000161, 0.00000563, 0.00000754, 0.00000202, -0.00000627, -0.00000694, -0.00000081, -0.00000110, -0.00000493, 0.00000114, 0.00000729, 0.00000012, -0.00000469, 0.00000184, -0.00000040, -0.00001056, -0.00000319, 0.00001070, 0.00000141, -0.00001445, -0.00001041, -0.00000551, -0.00001295, -0.00000995, -0.00000636, -0.00003898, -0.00007381, -0.00003662, 0.00005143, 0.00009240, 0.00006345, },
		{ 0.00008903, 0.00006616, 0.00000185, -0.00004099, -0.00003899, -0.00002441, -0.00001928, -0.00001013, 0.00000350, -0.00000042, -0.00001458, -0.00000923, 0.00000689, 0.00000319, -0.00001131, -0.00001131, -0.00000535, -0.00000710, -0.00000612, -0.00000099, -0.00000335, -0.00000455, 0.00000341, 0.00000097, -0.00001390, -0.00001069, 0.00000774, 0.00000358, -0.00001576, -0.00000928, 0.00001134, 0.00000519, -0.00001303, -0.00000503, 0.00001385, 0.00001016, -0.00000364, -0.00000335, 0.00000129, -0.00000165, -0.00000164, 0.00000404, 0.00000361, -0.00000047, 0.00000151, 0.00000265, -0.00000257, -0.00000280, 0.00000291, 0.00000034, -0.00000599, -0.00000227, -0.00000077, -0.00001068, -0.00000643, 0.00001657, 0.00001588, -0.00001216, -0.00001315, 0.00001475, 0.00001209, -0.00001668, -0.00001297, 0.00000921, 0.00000015, -0.00001128, 0.00001067, 0.00001819, -0.00001103, -0.00001448, 0.00001532, 0.00000885, -0.00002312, -0.00000988, 0.00002037, 0.00000381, -0.00001602, 0.00000872, 0.00001801, -0.00001874, -0.00002777, 0.00000995, 0.00002017, -0.00000507, -0.00000440, 0.00001280, -0.00000080, -0.00001876, -0.00000237, 0.00001613, 0.00000624, -0.00000578, -0.00000294, -0.00000312, -0.00000845, -0.00000408, 0.00000608, 0.00000855, 0.00000213, -0.00000581, -0.00000691, -0.00000140, -0.00000078, -0.00000498, -0.00000017, 0.00000760, 0.00000145, -0.00000602, 0.00000081, 0.00000219, -0.00001014, -0.00000680, 0.00001148, 0.00000627, -0.00001553, -0.00001440, -0.00000310, -0.00001115, -0.00001502, -0.00000414, -0.00002542, -0.00007374, -0.00006073, 0.00003105, 0.00009759, 0.00007872, 0.00004353, 0.00005958, },
		{ 0.00000800, -0.00004489, -0.00004576, -0.00002575, -0.00001704, -0.00001161, -0.00000043, -0.00000004, -0.00001198, -0.00001048, 0.00000493, 0.00000499, -0.00001005, -0.00001316, -0.00000566, -0.00000560, -0.00000687, -0.00000201, -0.00000243, -0.00000554, 0.00000154, 0.00000413, -0.00001055, -0.00001524, 0.00000319, 0.00000868, -0.00001091, -0.00001492, 0.00000598, 0.00000993, -0.00000888, -0.00000991, 0.00000996, 0.00001415, -0.00000019, -0.00000534, 0.00000011, -0.00000009, -0.00000236, 0.00000170, 0.00000441, 0.00000166, 0.00000081, 0.00000132, -0.00000160, -0.00000190, 0.00000213, 0.00000028, -0.00000495, -0.00000180, -0.00000017, -0.00001051, -0.00000968, 0.00001310, 0.00001878, -0.00000725, -0.00001524, 0.00001039, 0.00001467, -0.00001260, -0.00001513, 0.00000694, 0.00000201, -0.00001242, 0.00000720, 0.00002101, -0.00000706, -0.00001853, 0.00001219, 0.00001464, -0.00002050, -0.00001600, 0.00001925, 0.00000913, -0.00001820, 0.00000325, 0.00002342, -0.00001151, -0.00003273, 0.00000365, 0.00002353, -0.00000298, -0.00000877, 0.00001356, 0.00000489, -0.00002025, -0.00000796, 0.00001692, 0.00000945, -0.00000642, -0.00000276, -0.00000068, -0.00000948, -0.00000733, 0.00000610, 0.00001009, 0.00000211, -0.00000571, -0.00000647, -0.00000194, -0.00000086, -0.00000467, -0.00000125, 0.00000729, 0.00000296, -0.00000665, -0.00000101, 0.00000419, -0.00000836, -0.00001027, 0.00001039, 0.00001156, -0.00001449, -0.00001947, -0.00000242, -0.00000719, -0.00001928, -0.00000617, -0.00001229, -0.00006624, -0.00008129, 0.00000387, 0.00009624, 0.00009551, 0.00005210, 0.00005892, 0.00009409, 0.00007760, },
		{ -0.00005288, -0.00002796, -0.00001512, -0.00001226, -0.00000419, -0.00000069, -0.00000936, -0.00001076, 0.00000259, 0.00000601, -0.00000809, -0.00001457, -0.00000687, -0.00000412, -0.00000669, -0.00000331, -0.00000195, -0.00000577, -0.00000067, 0.00000579, -0.00000617, -0.00001752, -0.00000274, 0.00001109, -0.00000447, -0.00001772, -0.00000093, 0.00001204, -0.00000323, -0.00001259, 0.00000459, 0.00001624, 0.00000437, -0.00000603, -0.00000187, 0.00000097, -0.00000217, -0.00000057, 0.00000415, 0.00000378, 0.00000106, -0.00000009, -0.00000137, -0.00000088, 0.00000189, 0.00000004, -0.00000431, -0.00000133, 0.00000066, -0.00000978, -0.00001237, 0.00000890, 0.00002029, -0.00000194, -0.00001575, 0.00000573, 0.00001579, -0.00000820, -0.00001619, 0.00000434, 0.00000361, -0.00001281, 0.00000320, 0.00002263, -0.00000207, -0.00002134, 0.00000748, 0.00001932, -0.00001596, -0.00002140, 0.00001618, 0.00001464, -0.00001838, -0.00000348, 0.00002648, -0.00000257, -0.00003533, -0.00000411, 0.00002555, 0.00000059, -0.00001305, 0.00001245, 0.00001116, -0.00001976, -0.00001421, 0.00001629, 0.00001323, -0.00000674, -0.00000340, 0.00000228, -0.00000940, -0.00001122, 0.00000532, 0.00001221, 0.00000223, -0.00000619, -0.00000573, -0.00000212, -0.00000144, -0.00000426, -0.00000180, 0.00000645, 0.00000422, -0.00000642, -0.00000316, 0.00000512, -0.00000559, -0.00001279, 0.00000751, 0.00001625, -0.00001106, -0.00002453, -0.00000421, -0.00000199, -0.00002125, -0.00001168, -0.00000208, -0.00005243, -0.00009525, -0.00002805, 0.00008616, 0.00011174, 0.00006493, 0.00005908, 0.00009790, 0.00008970, 0.00001561, -0.00004830, },
		{ -0.00001359, -0.00001241, -0.00000752, -0.00000209, -0.00000714, -0.00001019, 0.00000034, 0.00000615, -0.00000585, -0.00001524, -0.00000870, -0.00000311, -0.00000568, -0.00000447, -0.00000205, -0.00000550, -0.00000273, 0.00000596, -0.00000168, -0.00001740, -0.00000884, 0.00001057, 0.00000219, -0.00001737, -0.00000805, 0.00001120, 0.00000273, -0.00001267, -0.00000125, 0.00001605, 0.00000925, -0.00000511, -0.00000416, 0.00000120, -0.00000134, -0.00000238, 0.00000299, 0.00000549, 0.00000213, -0.00000119, -0.00000183, -0.00000004, 0.00000214, -0.00000013, -0.00000403, -0.00000100, 0.00000159, -0.00000854, -0.00001434, 0.00000430, 0.00002030, 0.00000328, -0.00001469, 0.00000130, 0.00001548, -0.00000393, -0.00001615, 0.00000163, 0.00000482, -0.00001244, -0.00000098, 0.00002281, 0.00000346, -0.00002249, 0.00000171, 0.00002232, -0.00000991, -0.00002541, 0.00001134, 0.00001958, -0.00001638, -0.00001062, 0.00002674, 0.00000717, -0.00003506, -0.00001262, 0.00002569, 0.00000537, -0.00001659, 0.00000939, 0.00001727, -0.00001704, -0.00002050, 0.00001395, 0.00001729, -0.00000643, -0.00000495, 0.00000538, -0.00000788, -0.00001540, 0.00000336, 0.00001477, 0.00000287, -0.00000737, -0.00000498, -0.00000166, -0.00000244, -0.00000414, -0.00000167, 0.00000541, 0.00000484, -0.00000544, -0.00000504, 0.00000477, -0.00000248, -0.00001377, 0.00000344, 0.00001934, -0.00000559, -0.00002831, -0.00000860, 0.00000302, -0.00002005, -0.00001907, 0.00000336, -0.00003464, -0.00010035, -0.00006144, 0.00006612, 0.00012448, 0.00008217, 0.00006149, 0.00010029, 0.00010198, 0.00002502, -0.00005111, -0.00006046, -0.00003082, },
		{ -0.00001041, -0.00000386, -0.00000552, -0.00000907, -0.00000146, 0.00000557, -0.00000377, -0.00001511, -0.00001076, -0.00000281, -0.00000414, -0.00000511, -0.00000267, -0.00000506, -0.00000432, 0.00000495, 0.00000218, -0.00001518, -0.00001399, 0.00000745, 0.00000778, -0.00001418, -0.00001408, 0.00000764, 0.00000780, -0.00001028, -0.00000649, 0.00001360, 0.00001356, -0.00000256, -0.00000614, 0.00000048, -0.00000025, -0.00000347, 0.00000122, 0.00000648, 0.00000378, -0.00000170, -0.00000275, 0.00000038, 0.00000273, -0.00000005, -0.00000399, -0.00000093, 0.00000248, -0.00000688, -0.00001545, -0.00000036, 0.00001888, 0.00000793, -0.00001225, -0.00000243, 0.00001392, -0.00000019, -0.00001512, -0.00000092, 0.00000556, -0.00001138, -0.00000496, 0.00002153, 0.00000897, -0.00002180, -0.00000449, 0.00002327, -0.00000298, -0.00002750, 0.00000517, 0.00002323, -0.00001237, -0.00001723, 0.00002412, 0.00001664, -0.00003174, -0.00002098, 0.00002363, 0.00001081, -0.00001881, 0.00000460, 0.00002247, -0.00001214, -0.00002613, 0.00000982, 0.00002123, -0.00000521, -0.00000738, 0.00000811, -0.00000475, -0.00001936, -0.00000006, 0.00001748, 0.00000441, -0.00000920, -0.00000464, -0.00000035, -0.00000357, -0.00000467, -0.00000089, 0.00000462, 0.00000456, -0.00000409, -0.00000609, 0.00000331, 0.00000017, -0.00001301, -0.00000092, 0.00002019, 0.00000103, -0.00002971, -0.00001499, 0.00000633, -0.00001566, -0.00002624, 0.00000323, -0.00001598, -0.00009569, -0.00009228, 0.00003629, 0.00013040, 0.00010292, 0.00006779, 0.00010159, 0.00011373, 0.00003641, -0.00005297, -0.00006870, -0.00003427, -0.00001233, -0.00001239, },
		{ -0.00000447, -0.00000781, -0.00000267, 0.00000459, -0.00000213, -0.00001437, -0.00001265, -0.00000327, -0.00000246, -0.00000507, -0.00000359, -0.00000474, -0.00000533, 0.00000320, 0.00000489, -0.00001149, -0.00001739, 0.00000250, 0.00001131, -0.00000898, -0.00001798, 0.00000212, 0.00001100, -0.00000601, -0.00001021, 0.00000936, 0.00001651, 0.00000130, -0.00000721, -0.00000106, 0.00000066, -0.00000378, -0.00000077, 0.00000659, 0.00000565, -0.00000148, -0.00000387, 0.00000025, 0.00000344, 0.00000038, -0.00000401, -0.00000116, 0.00000316, -0.00000493, -0.00001566, -0.00000474, 0.00001618, 0.00001160, -0.00000875, -0.00000511, 0.00001146, 0.00000273, -0.00001332, -0.00000310, 0.00000580, -0.00000978, -0.00000842, 0.00001891, 0.00001389, -0.00001933, -0.00001040, 0.00002207, 0.00000409, -0.00002739, -0.00000166, 0.00002505, -0.00000679, -0.00002245, 0.00001893, 0.00002479, -0.00002557, -0.00002825, 0.00001934, 0.00001626, -0.00001930, -0.00000148, 0.00002605, -0.00000534, -0.00003036, 0.00000402, 0.00002454, -0.00000286, -0.00001047, 0.00000994, -0.00000004, -0.00002245, -0.00000506, 0.00001981, 0.00000718, -0.00001141, -0.00000515, 0.00000180, -0.00000439, -0.00000617, 0.00000023, 0.00000460, 0.00000341, -0.00000297, -0.00000596, 0.00000125, 0.00000160, -0.00001079, -0.00000448, 0.00001867, 0.00000755, -0.00002811, -0.00002216, 0.00000674, -0.00000896, -0.00003115, -0.00000202, 0.00000027, -0.00008199, -0.00011645, -0.00000146, 0.00012628, 0.00012499, 0.00007947, 0.00010286, 0.00012418, 0.00004964, -0.00005335, -0.00007772, -0.00003843, -0.00001110, -0.00001236, -0.00001305, -0.00000570, },
		{ -0.00000338, 0.00000359, -0.00000104, -0.00001333, -0.00001412, -0.00000429, -0.00000100, -0.00000436, -0.00000447, -0.00000472, -0.00000585, 0.00000117, 0.00000626, -0.00000712, -0.00001864, -0.00000324, 0.00001228, -0.00000291, -0.00001920, -0.00000427, 0.00001175, -0.00000077, -0.00001179, 0.00000409, 0.00001754, 0.00000586, -0.00000696, -0.00000308, 0.00000104, -0.00000342, -0.00000261, 0.00000583, 0.00000736, -0.00000050, -0.00000484, -0.00000043, 0.00000404, 0.00000115, -0.00000393, -0.00000163, 0.00000353, -0.00000288, -0.00001499, -0.00000851, 0.00001249, 0.00001401, -0.00000461, -0.00000652, 0.00000849, 0.00000463, -0.00001104, -0.00000476, 0.00000556, -0.00000783, -0.00001109, 0.00001522, 0.00001774, -0.00001534, -0.00001537, 0.00001889, 0.00001052, -0.00002508, -0.00000841, 0.00002477, -0.00000033, -0.00002564, 0.00001179, 0.00003075, -0.00001711, -0.00003357, 0.00001311, 0.00002100, -0.00001792, -0.00000821, 0.00002750, 0.00000277, -0.00003258, -0.00000313, 0.00002674, 0.00000067, -0.00001385, 0.00001038, 0.00000595, -0.00002398, -0.00001152, 0.00002114, 0.00001134, -0.00001353, -0.00000693, 0.00000455, -0.00000434, -0.00000871, 0.00000116, 0.00000577, 0.00000173, -0.00000274, -0.00000465, -0.00000060, 0.00000137, -0.00000787, -0.00000633, 0.00001529, 0.00001261, -0.00002362, -0.00002852, 0.00000368, -0.00000156, -0.00003228, -0.00001074, 0.00001133, -0.00006158, -0.00013051, -0.00004358, 0.00010981, 0.00014500, 0.00009732, 0.00010584, 0.00013280, 0.00006414, -0.00005164, -0.00008739, -0.00004366, -0.00000970, -0.00001223, -0.00001576, -0.00000743, -0.00000374, -0.00000672, },
		{ -0.00000035, -0.00001233, -0.00001510, -0.00000557, 0.00000000, -0.00000320, -0.00000506, -0.00000500, -0.00000607, -0.00000079, 0.00000639, -0.00000283, -0.00001777, -0.00000872, 0.00001070, 0.00000283, -0.00001771, -0.00001033, 0.00000996, 0.00000434, -0.00001106, -0.00000126, 0.00001645, 0.00001036, -0.00000521, -0.00000506, 0.00000067, -0.00000265, -0.00000399, 0.00000438, 0.00000859, 0.00000108, -0.00000540, -0.00000155, 0.00000431, 0.00000218, -0.00000359, -0.00000225, 0.00000350, -0.00000092, -0.00001356, -0.00001144, 0.00000817, 0.00001499, -0.00000030, -0.00000659, 0.00000545, 0.00000547, -0.00000859, -0.00000582, 0.00000495, -0.00000574, -0.00001280, 0.00001084, 0.00002017, -0.00001030, -0.00001888, 0.00001417, 0.00001563, -0.00002084, -0.00001432, 0.00002239, 0.00000621, -0.00002646, 0.00000357, 0.00003390, -0.00000721, -0.00003627, 0.00000552, 0.00002437, -0.00001474, -0.00001486, 0.00002652, 0.00001146, -0.00003234, -0.00001106, 0.00002734, 0.00000529, -0.00001702, 0.00000906, 0.00001267, -0.00002336, -0.00001902, 0.00002080, 0.00001678, -0.00001491, -0.00001023, 0.00000743, -0.00000293, -0.00001209, 0.00000123, 0.00000827, 0.00000015, -0.00000393, -0.00000256, -0.00000139, -0.00000050, -0.00000528, -0.00000597, 0.00001110, 0.00001513, -0.00001706, -0.00003249, -0.00000249, 0.00000457, -0.00002911, -0.00002045, 0.00001554, -0.00003804, -0.00013245, -0.00008532, 0.00008021, 0.00015871, 0.00012093, 0.00011279, 0.00013969, 0.00007898, -0.00004738, -0.00009729, -0.00005040, -0.00000813, -0.00001172, -0.00001881, -0.00000917, -0.00000299, -0.00000593, -0.00000388, 0.00000285, },
		{ -0.00001576, -0.00000682, 0.00000053, -0.00000186, -0.00000521, -0.00000544, -0.00000619, -0.00000250, 0.00000553, 0.00000080, -0.00001518, -0.00001303, 0.00000705, 0.00000719, -0.00001402, -0.00001502, 0.00000606, 0.00000835, -0.00000827, -0.00000575, 0.00001341, 0.00001400, -0.00000209, -0.00000648, -0.00000046, -0.00000183, -0.00000476, 0.00000252, 0.00000911, 0.00000301, -0.00000535, -0.00000290, 0.00000409, 0.00000330, -0.00000290, -0.00000285, 0.00000306, 0.00000076, -0.00001154, -0.00001335, 0.00000361, 0.00001454, 0.00000370, -0.00000544, 0.00000274, 0.00000535, -0.00000626, -0.00000626, 0.00000408, -0.00000371, -0.00001351, 0.00000621, 0.00002100, -0.00000478, -0.00002062, 0.00000849, 0.00001894, -0.00001521, -0.00001874, 0.00001821, 0.00001203, -0.00002489, -0.00000474, 0.00003399, 0.00000310, -0.00003603, -0.00000266, 0.00002588, -0.00001013, -0.00002072, 0.00002313, 0.00001986, -0.00002946, -0.00001907, 0.00002602, 0.00001069, -0.00001942, 0.00000585, 0.00001939, -0.00002015, -0.00002684, 0.00001821, 0.00002309, -0.00001486, -0.00001503, 0.00000973, 0.00000022, -0.00001580, -0.00000020, 0.00001191, -0.00000050, -0.00000678, -0.00000047, -0.00000041, -0.00000349, -0.00000414, -0.00000347, 0.00000746, 0.00001461, -0.00000988, -0.00003293, -0.00001055, 0.00000758, -0.00002223, -0.00002841, 0.00001273, -0.00001555, -0.00012215, -0.00012133, 0.00003873, 0.00016168, 0.00014832, 0.00012599, 0.00014581, 0.00009299, -0.00004054, -0.00010666, -0.00005898, -0.00000668, -0.00001042, -0.00002227, -0.00001124, -0.00000195, -0.00000530, -0.00000451, 0.00000244, 0.00000023, -0.00001159, },
		{ 0.00000073, -0.00000065, -0.00000493, -0.00000583, -0.00000633, -0.00000390, 0.00000402, 0.00000343, -0.00001147, -0.00001562, 0.00000214, 0.00000949, -0.00000902, -0.00001761, 0.00000084, 0.00001050, -0.00000410, -0.00000865, 0.00000896, 0.00001613, 0.00000199, -0.00000686, -0.00000213, -0.00000131, -0.00000490, 0.00000057, 0.00000883, 0.00000497, -0.00000460, -0.00000421, 0.00000336, 0.00000431, -0.00000185, -0.00000328, 0.00000228, 0.00000200, -0.00000916, -0.00001420, -0.00000080, 0.00001278, 0.00000698, -0.00000330, 0.00000071, 0.00000447, -0.00000431, -0.00000617, 0.00000310, -0.00000190, -0.00001329, 0.00000177, 0.00002027, 0.00000061, -0.00002049, 0.00000256, 0.00002021, -0.00000887, -0.00002125, 0.00001277, 0.00001646, -0.00002124, -0.00001222, 0.00003117, 0.00001274, -0.00003284, -0.00001054, 0.00002531, -0.00000464, -0.00002516, 0.00001760, 0.00002712, -0.00002402, -0.00002636, 0.00002262, 0.00001643, -0.00002054, 0.00000084, 0.00002522, -0.00001425, -0.00003406, 0.00001303, 0.00002958, -0.00001276, -0.00002096, 0.00001067, 0.00000514, -0.00001906, -0.00000363, 0.00001612, 0.00000059, -0.00001108, 0.00000062, 0.00000265, -0.00000661, -0.00000533, 0.00000041, 0.00000577, 0.00001131, -0.00000388, -0.00002945, -0.00001862, 0.00000628, -0.00001333, -0.00003225, 0.00000432, 0.00000194, -0.00010156, -0.00014664, -0.00001120, 0.00015010, 0.00017587, 0.00014719, 0.00015314, 0.00010521, -0.00003162, -0.00011462, -0.00006947, -0.00000591, -0.00000796, -0.00002593, -0.00001407, -0.00000054, -0.00000450, -0.00000557, 0.00000221, 0.00000106, -0.00001114, -0.00001637, -0.00000783, },
		{ -0.00000440, -0.00000599, -0.00000648, -0.00000507, 0.00000218, 0.00000495, -0.00000730, -0.00001629, -0.00000308, 0.00000950, -0.00000375, -0.00001784, -0.00000468, 0.00001050, 0.00000059, -0.00000955, 0.00000389, 0.00001637, 0.00000640, -0.00000591, -0.00000397, -0.00000133, -0.00000457, -0.00000116, 0.00000782, 0.00000664, -0.00000322, -0.00000521, 0.00000218, 0.00000501, -0.00000052, -0.00000340, 0.00000127, 0.00000270, -0.00000667, -0.00001402, -0.00000469, 0.00000997, 0.00000925, -0.00000051, -0.00000043, 0.00000311, -0.00000291, -0.00000568, 0.00000213, -0.00000043, -0.00001228, -0.00000211, 0.00001817, 0.00000533, -0.00001868, -0.00000295, 0.00001948, -0.00000257, -0.00002169, 0.00000678, 0.00001907, -0.00001610, -0.00001810, 0.00002593, 0.00002078, -0.00002708, -0.00001725, 0.00002270, 0.00000108, -0.00002778, 0.00001048, 0.00003250, -0.00001644, -0.00003217, 0.00001725, 0.00002190, -0.00001996, -0.00000555, 0.00002932, -0.00000589, -0.00003966, 0.00000524, 0.00003532, -0.00000824, -0.00002737, 0.00000953, 0.00001154, -0.00002096, -0.00000920, 0.00001999, 0.00000405, -0.00001616, -0.00000030, 0.00000757, -0.00000858, -0.00000920, 0.00000438, 0.00000707, 0.00000628, -0.00000073, -0.00002263, -0.00002453, 0.00000045, -0.00000485, -0.00003056, -0.00000696, 0.00001147, -0.00007452, -0.00015759, -0.00006424, 0.00012170, 0.00019852, 0.00017689, 0.00016451, 0.00011527, -0.00002175, -0.00012040, -0.00008148, -0.00000653, -0.00000424, -0.00002929, -0.00001803, 0.00000101, -0.00000311, -0.00000709, 0.00000178, 0.00000236, -0.00001078, -0.00001725, -0.00000863, 0.00000087, 0.00000030, },
		{ -0.00000653, -0.00000608, 0.00000019, 0.00000547, -0.00000326, -0.00001521, -0.00000775, 0.00000745, 0.00000079, -0.00001594, -0.00000952, 0.00000847, 0.00000487, -0.00000845, -0.00000094, 0.00001469, 0.00001040, -0.00000363, -0.00000551, -0.00000198, -0.00000402, -0.00000246, 0.00000626, 0.00000776, -0.00000139, -0.00000570, 0.00000069, 0.00000525, 0.00000095, -0.00000310, 0.00000021, 0.00000284, -0.00000432, -0.00001294, -0.00000777, 0.00000646, 0.00001030, 0.00000252, -0.00000059, 0.00000159, -0.00000214, -0.00000496, 0.00000130, 0.00000064, -0.00001070, -0.00000514, 0.00001506, 0.00000892, -0.00001555, -0.00000746, 0.00001703, 0.00000301, -0.00002018, 0.00000095, 0.00001968, -0.00001020, -0.00002188, 0.00001901, 0.00002648, -0.00001944, -0.00002209, 0.00001840, 0.00000636, -0.00002841, 0.00000244, 0.00003546, -0.00000734, -0.00003584, 0.00001024, 0.00002647, -0.00001750, -0.00001269, 0.00003097, 0.00000430, -0.00004265, -0.00000475, 0.00003929, -0.00000125, -0.00003331, 0.00000579, 0.00001872, -0.00002060, -0.00001662, 0.00002241, 0.00001009, -0.00002093, -0.00000401, 0.00001353, -0.00000814, -0.00001546, 0.00000686, 0.00001174, 0.00000121, -0.00000160, -0.00001399, -0.00002634, -0.00000892, 0.00000060, -0.00002333, -0.00001761, 0.00001175, -0.00004608, -0.00015260, -0.00011381, 0.00007645, 0.00021042, 0.00021380, 0.00018319, 0.00012382, -0.00001245, -0.00012363, -0.00009420, -0.00000920, 0.00000048, -0.00003171, -0.00002328, 0.00000222, -0.00000076, -0.00000886, 0.00000073, 0.00000415, -0.00001014, -0.00001856, -0.00000942, 0.00000120, 0.00000099, -0.00000389, -0.00000583, },
		{ -0.00000185, 0.00000517, 0.00000025, -0.00001279, -0.00001121, 0.00000392, 0.00000384, -0.00001253, -0.00001289, 0.00000494, 0.00000796, -0.00000576, -0.00000475, 0.00001139, 0.00001332, -0.00000026, -0.00000628, -0.00000318, -0.00000356, -0.00000323, 0.00000441, 0.00000818, 0.00000065, -0.00000554, -0.00000087, 0.00000497, 0.00000237, -0.00000238, -0.00000074, 0.00000247, -0.00000234, -0.00001119, -0.00000988, 0.00000263, 0.00001011, 0.00000539, 0.00000018, 0.00000023, -0.00000195, -0.00000420, 0.00000066, 0.00000132, -0.00000880, -0.00000717, 0.00001136, 0.00001115, -0.00001160, -0.00001058, 0.00001334, 0.00000732, -0.00001709, -0.00000406, 0.00001843, -0.00000430, -0.00002336, 0.00001132, 0.00002947, -0.00001078, -0.00002458, 0.00001296, 0.00001061, -0.00002714, -0.00000573, 0.00003573, 0.00000244, -0.00003691, 0.00000215, 0.00002954, -0.00001319, -0.00001979, 0.00002970, 0.00001538, -0.00004228, -0.00001615, 0.00004055, 0.00000783, -0.00003773, -0.00000068, 0.00002567, -0.00001734, -0.00002514, 0.00002232, 0.00001840, -0.00002409, -0.00001078, 0.00001923, -0.00000437, -0.00002308, 0.00000631, 0.00001929, -0.00000192, -0.00000679, -0.00000573, -0.00002288, -0.00001978, 0.00000082, -0.00001200, -0.00002411, 0.00000349, -0.00002166, -0.00013263, -0.00015306, 0.00001706, 0.00020587, 0.00025445, 0.00021223, 0.00013287, -0.00000532, -0.00012469, -0.00010649, -0.00001428, 0.00000554, -0.00003254, -0.00002959, 0.00000244, 0.00000266, -0.00001039, -0.00000129, 0.00000617, -0.00000880, -0.00002028, -0.00001054, 0.00000188, 0.00000164, -0.00000363, -0.00000544, -0.00000631, -0.00000697, },
		{ 0.00000301, -0.00000954, -0.00001314, -0.00000032, 0.00000502, -0.00000847, -0.00001439, 0.00000071, 0.00000940, -0.00000212, -0.00000701, 0.00000708, 0.00001468, 0.00000372, -0.00000595, -0.00000465, -0.00000344, -0.00000353, 0.00000256, 0.00000788, 0.00000259, -0.00000473, -0.00000226, 0.00000417, 0.00000356, -0.00000128, -0.00000140, 0.00000171, -0.00000088, -0.00000904, -0.00001092, -0.00000112, 0.00000877, 0.00000772, 0.00000171, -0.00000070, -0.00000225, -0.00000356, 0.00000023, 0.00000166, -0.00000682, -0.00000819, 0.00000753, 0.00001194, -0.00000739, -0.00001215, 0.00000901, 0.00001004, -0.00001297, -0.00000775, 0.00001569, 0.00000088, -0.00002269, 0.00000375, 0.00002970, -0.00000206, -0.00002457, 0.00000709, 0.00001344, -0.00002428, -0.00001330, 0.00003332, 0.00001200, -0.00003519, -0.00000628, 0.00003065, -0.00000735, -0.00002597, 0.00002542, 0.00002618, -0.00003815, -0.00002782, 0.00003843, 0.00001821, -0.00003965, -0.00000960, 0.00003122, -0.00001092, -0.00003362, 0.00001887, 0.00002813, -0.00002434, -0.00002025, 0.00002311, 0.00000309, -0.00003045, 0.00000162, 0.00002840, -0.00000125, -0.00001556, -0.00000030, -0.00001406, -0.00002938, -0.00000536, 0.00000070, -0.00002375, -0.00001064, -0.00000599, -0.00010126, -0.00017606, -0.00005108, 0.00018038, 0.00029328, 0.00025364, 0.00014573, -0.00000146, -0.00012478, -0.00011718, -0.00002158, 0.00000998, -0.00003141, -0.00003635, 0.00000104, 0.00000692, -0.00001107, -0.00000437, 0.00000794, -0.00000647, -0.00002211, -0.00001235, 0.00000283, 0.00000256, -0.00000372, -0.00000504, -0.00000569, -0.00000766, -0.00000394, 0.00000426, },
		{ -0.00001354, -0.00000448, 0.00000437, -0.00000466, -0.00001405, -0.00000337, 0.00000910, 0.00000167, -0.00000752, 0.00000249, 0.00001427, 0.00000767, -0.00000439, -0.00000601, -0.00000380, -0.00000353, 0.00000093, 0.00000694, 0.00000418, -0.00000337, -0.00000326, 0.00000299, 0.00000436, 0.00000007, -0.00000166, 0.00000072, -0.00000001, -0.00000676, -0.00001096, -0.00000444, 0.00000650, 0.00000924, 0.00000373, -0.00000103, -0.00000285, -0.00000315, -0.00000001, 0.00000174, -0.00000496, -0.00000829, 0.00000398, 0.00001142, -0.00000346, -0.00001220, 0.00000465, 0.00001108, -0.00000843, -0.00000985, 0.00001198, 0.00000483, -0.00002025, -0.00000290, 0.00002745, 0.00000581, -0.00002221, 0.00000151, 0.00001468, -0.00002031, -0.00001965, 0.00002856, 0.00002045, -0.00003081, -0.00001423, 0.00002955, -0.00000053, -0.00003045, 0.00001841, 0.00003547, -0.00003031, -0.00003842, 0.00003263, 0.00002876, -0.00003830, -0.00002019, 0.00003420, -0.00000158, -0.00004069, 0.00001172, 0.00003793, -0.00002069, -0.00003142, 0.00002364, 0.00001379, -0.00003567, -0.00000756, 0.00003708, 0.00000446, -0.00002615, 0.00000014, -0.00000108, -0.00003480, -0.00001772, 0.00001130, -0.00001527, -0.00002654, -0.00000219, -0.00006414, -0.00017890, -0.00012035, 0.00013173, 0.00032306, 0.00030749, 0.00016675, -0.00000089, -0.00012586, -0.00012551, -0.00003027, 0.00001281, -0.00002842, -0.00004267, -0.00000232, 0.00001141, -0.00001032, -0.00000826, 0.00000884, -0.00000310, -0.00002355, -0.00001509, 0.00000374, 0.00000405, -0.00000409, -0.00000494, -0.00000465, -0.00000798, -0.00000608, 0.00000285, 0.00000500, -0.00000599, },
		{ 0.00000227, -0.00000184, -0.00001228, -0.00000654, 0.00000738, 0.00000487, -0.00000642, -0.00000162, 0.00001220, 0.00001093, -0.00000171, -0.00000683, -0.00000464, -0.00000347, -0.00000033, 0.00000556, 0.00000520, -0.00000166, -0.00000370, 0.00000159, 0.00000465, 0.00000151, -0.00000147, -0.00000030, 0.00000025, -0.00000462, -0.00001013, -0.00000704, 0.00000361, 0.00000979, 0.00000591, -0.00000069, -0.00000352, -0.00000305, -0.00000013, 0.00000166, -0.00000336, -0.00000766, 0.00000102, 0.00000986, -0.00000024, -0.00001101, 0.00000080, 0.00001061, -0.00000409, -0.00001036, 0.00000793, 0.00000728, -0.00001664, -0.00000806, 0.00002330, 0.00001211, -0.00001795, -0.00000310, 0.00001440, -0.00001579, -0.00002436, 0.00002196, 0.00002704, -0.00002417, -0.00002089, 0.00002626, 0.00000654, -0.00003266, 0.00000934, 0.00004215, -0.00001934, -0.00004662, 0.00002338, 0.00003815, -0.00003337, -0.00003130, 0.00003372, 0.00000988, -0.00004498, 0.00000111, 0.00004619, -0.00001270, -0.00004270, 0.00001964, 0.00002652, -0.00003692, -0.00002065, 0.00004303, 0.00001552, -0.00003605, -0.00000578, 0.00001371, -0.00003361, -0.00003452, 0.00001632, 0.00000071, -0.00003939, -0.00001106, -0.00002821, -0.00016056, -0.00018194, 0.00006085, 0.00033590, 0.00037132, 0.00020065, -0.00000205, -0.00013034, -0.00013147, -0.00003897, 0.00001328, -0.00002423, -0.00004761, -0.00000757, 0.00001527, -0.00000786, -0.00001237, 0.00000834, 0.00000101, -0.00002402, -0.00001873, 0.00000415, 0.00000629, -0.00000442, -0.00000541, -0.00000334, -0.00000772, -0.00000821, 0.00000101, 0.00000628, -0.00000251, -0.00001264, -0.00000792, },
		{ -0.00000974, -0.00000835, 0.00000481, 0.00000694, -0.00000419, -0.00000464, 0.00000888, 0.00001296, 0.00000174, -0.00000676, -0.00000578, -0.00000357, -0.00000118, 0.00000397, 0.00000556, 0.00000015, -0.00000351, 0.00000021, 0.00000443, 0.00000284, -0.00000084, -0.00000116, -0.00000001, -0.00000285, -0.00000867, -0.00000876, 0.00000045, 0.00000931, 0.00000792, 0.00000029, -0.00000406, -0.00000323, -0.00000020, 0.00000149, -0.00000210, -0.00000653, -0.00000114, 0.00000763, 0.00000199, -0.00000896, -0.00000216, 0.00000895, -0.00000046, -0.00000946, 0.00000411, 0.00000817, -0.00001251, -0.00001141, 0.00001796, 0.00001638, -0.00001243, -0.00000627, 0.00001288, -0.00001128, -0.00002722, 0.00001422, 0.00003124, -0.00001595, -0.00002559, 0.00002108, 0.00001308, -0.00003229, -0.00000082, 0.00004538, -0.00000624, -0.00005128, 0.00001142, 0.00004509, -0.00002503, -0.00004158, 0.00002929, 0.00002220, -0.00004539, -0.00001209, 0.00005130, -0.00000067, -0.00005219, 0.00001064, 0.00003940, -0.00003283, -0.00003612, 0.00004408, 0.00003110, -0.00004251, -0.00001822, 0.00002719, -0.00002447, -0.00005270, 0.00001303, 0.00002173, -0.00004469, -0.00003083, -0.00000041, -0.00012330, -0.00022707, -0.00002784, 0.00032436, 0.00043982, 0.00025159, -0.00000155, -0.00014047, -0.00013619, -0.00004602, 0.00001128, -0.00001996, -0.00005043, -0.00001409, 0.00001757, -0.00000379, -0.00001586, 0.00000619, 0.00000526, -0.00002305, -0.00002293, 0.00000355, 0.00000921, -0.00000430, -0.00000658, -0.00000207, -0.00000673, -0.00001016, -0.00000128, 0.00000693, 0.00000067, -0.00001082, -0.00001026, -0.00000065, -0.00000045, },
		{ 0.00000211, 0.00000768, -0.00000144, -0.00000623, 0.00000493, 0.00001343, 0.00000543, -0.00000558, -0.00000690, -0.00000400, -0.00000173, 0.00000241, 0.00000528, 0.00000180, -0.00000274, -0.00000093, 0.00000374, 0.00000389, 0.00000015, -0.00000172, -0.00000064, -0.00000159, -0.00000683, -0.00000953, -0.00000261, 0.00000791, 0.00000948, 0.00000178, -0.00000428, -0.00000365, -0.00000030, 0.00000130, -0.00000120, -0.00000515, -0.00000243, 0.00000512, 0.00000314, -0.00000651, -0.00000402, 0.00000655, 0.00000214, -0.00000754, 0.00000099, 0.00000772, -0.00000847, -0.00001293, 0.00001220, 0.00001845, -0.00000641, -0.00000772, 0.00001056, -0.00000723, -0.00002826, 0.00000610, 0.00003279, -0.00000696, -0.00002787, 0.00001458, 0.00001835, -0.00002940, -0.00001099, 0.00004476, 0.00000764, -0.00005168, -0.00000205, 0.00004853, -0.00001400, -0.00004964, 0.00002100, 0.00003386, -0.00004129, -0.00002643, 0.00005194, 0.00001438, -0.00005802, -0.00000303, 0.00005020, -0.00002288, -0.00005170, 0.00003872, 0.00004928, -0.00004301, -0.00003597, 0.00003603, -0.00000752, -0.00006843, 0.00000015, 0.00004381, -0.00003917, -0.00005740, 0.00001347, -0.00007255, -0.00024852, -0.00012660, 0.00028294, 0.00050500, 0.00032206, 0.00000580, -0.00015761, -0.00014196, -0.00004993, 0.00000753, -0.00001704, -0.00005094, -0.00002081, 0.00001764, 0.00000124, -0.00001790, 0.00000255, 0.00000885, -0.00002046, -0.00002706, 0.00000154, 0.00001243, -0.00000329, -0.00000834, -0.00000127, -0.00000497, -0.00001166, -0.00000400, 0.00000697, 0.00000347, -0.00000850, -0.00001140, -0.00000373, -0.00000057, -0.00000722, -0.00000868, },
		{ 0.00000114, -0.00000638, 0.00000104, 0.00001231, 0.00000875, -0.00000330, -0.00000765, -0.00000482, -0.00000214, 0.00000105, 0.00000447, 0.00000305, -0.00000150, -0.00000165, 0.00000274, 0.00000453, 0.00000134, -0.00000188, -0.00000148, -0.00000091, -0.00000490, -0.00000940, -0.00000526, 0.00000578, 0.00001035, 0.00000356, -0.00000407, -0.00000418, -0.00000051, 0.00000112, -0.00000062, -0.00000374, -0.00000291, 0.00000273, 0.00000326, -0.00000409, -0.00000478, 0.00000390, 0.00000356, -0.00000506, -0.00000112, 0.00000627, -0.00000504, -0.00001283, 0.00000675, 0.00001841, -0.00000064, -0.00000746, 0.00000796, -0.00000398, -0.00002770, -0.00000170, 0.00003172, 0.00000189, -0.00002754, 0.00000748, 0.00002179, -0.00002439, -0.00002004, 0.00004037, 0.00002081, -0.00004759, -0.00001554, 0.00004785, -0.00000145, -0.00005435, 0.00000950, 0.00004328, -0.00003271, -0.00004015, 0.00004737, 0.00003076, -0.00005867, -0.00002011, 0.00005675, -0.00000754, -0.00006476, 0.00002641, 0.00006729, -0.00003587, -0.00005657, 0.00003736, 0.00001546, -0.00007781, -0.00002177, 0.00006220, -0.00002157, -0.00008497, 0.00000992, -0.00001618, -0.00024186, -0.00022510, 0.00020925, 0.00055698, 0.00041191, 0.00002643, -0.00018147, -0.00015208, -0.00004995, 0.00000363, -0.00001672, -0.00004967, -0.00002643, 0.00001530, 0.00000621, -0.00001790, -0.00000195, 0.00001098, -0.00001646, -0.00003036, -0.00000196, 0.00001536, -0.00000107, -0.00001032, -0.00000133, -0.00000264, -0.00001240, -0.00000707, 0.00000637, 0.00000588, -0.00000597, -0.00001147, -0.00000637, -0.00000194, -0.00000541, -0.00000780, -0.00000003, 0.00000721, },
		{ -0.00000217, 0.00000983, 0.00001112, -0.00000016, -0.00000764, -0.00000592, -0.00000261, -0.00000002, 0.00000331, 0.00000375, -0.00000002, -0.00000181, 0.00000161, 0.00000469, 0.00000257, -0.00000160, -0.00000232, -0.00000079, -0.00000313, -0.00000853, -0.00000727, 0.00000319, 0.00001043, 0.00000538, -0.00000337, -0.00000469, -0.00000085, 0.00000096, -0.00000030, -0.00000249, -0.00000273, 0.00000073, 0.00000256, -0.00000207, -0.00000458, 0.00000143, 0.00000385, -0.00000249, -0.00000214, 0.00000428, -0.00000255, -0.00001152, 0.00000215, 0.00001665, 0.00000427, -0.00000571, 0.00000558, -0.00000167, -0.00002591, -0.00000857, 0.00002832, 0.00000975, -0.00002474, 0.00000061, 0.00002307, -0.00001792, -0.00002706, 0.00003277, 0.00003187, -0.00003940, -0.00002748, 0.00004297, 0.00001117, -0.00005498, -0.00000400, 0.00004910, -0.00002034, -0.00005140, 0.00003759, 0.00004641, -0.00005331, -0.00003863, 0.00005732, 0.00001169, -0.00007278, 0.00000784, 0.00008203, -0.00002064, -0.00007664, 0.00002940, 0.00004130, -0.00007766, -0.00005025, 0.00007226, 0.00000699, -0.00010711, -0.00001153, 0.00003665, -0.00020642, -0.00031178, 0.00010497, 0.00058514, 0.00051753, 0.00006709, -0.00020951, -0.00017029, -0.00004660, 0.00000186, -0.00001969, -0.00004790, -0.00002975, 0.00001104, 0.00000986, -0.00001573, -0.00000633, 0.00001113, -0.00001170, -0.00003211, -0.00000666, 0.00001730, 0.00000236, -0.00001198, -0.00000247, -0.00000013, -0.00001210, -0.00001025, 0.00000502, 0.00000791, -0.00000343, -0.00001078, -0.00000820, -0.00000406, -0.00000475, -0.00000626, -0.00000116, 0.00000596, 0.00000302, -0.00000535, },
		{ 0.00001215, 0.00000341, -0.00000663, -0.00000707, -0.00000331, -0.00000082, 0.00000202, 0.00000385, 0.00000147, -0.00000139, 0.00000056, 0.00000437, 0.00000366, -0.00000092, -0.00000299, -0.00000113, -0.00000171, -0.00000713, -0.00000849, 0.00000045, 0.00000968, 0.00000699, -0.00000225, -0.00000503, -0.00000132, 0.00000081, -0.00000015, -0.00000148, -0.00000210, -0.00000068, 0.00000134, -0.00000068, -0.00000371, -0.00000054, 0.00000324, -0.00000025, -0.00000215, 0.00000219, -0.00000110, -0.00000948, -0.00000124, 0.00001368, 0.00000785, -0.00000288, 0.00000383, -0.00000026, -0.00002332, -0.00001408, 0.00002311, 0.00001591, -0.00001989, -0.00000524, 0.00002220, -0.00001085, -0.00003140, 0.00002291, 0.00003966, -0.00002804, -0.00003648, 0.00003443, 0.00002234, -0.00005140, -0.00001800, 0.00005042, -0.00000548, -0.00005857, 0.00002341, 0.00005917, -0.00004200, -0.00005619, 0.00005102, 0.00003249, -0.00007379, -0.00001513, 0.00009058, 0.00000169, -0.00009253, 0.00001188, 0.00006587, -0.00006620, -0.00008124, 0.00007043, 0.00004306, -0.00011795, -0.00004818, 0.00007699, -0.00014565, -0.00037543, -0.00002381, 0.00057977, 0.00063170, 0.00013364, -0.00023670, -0.00019998, -0.00004196, 0.00000475, -0.00002567, -0.00004749, -0.00003007, 0.00000603, 0.00001108, -0.00001193, -0.00000942, 0.00000925, -0.00000716, -0.00003185, -0.00001190, 0.00001762, 0.00000668, -0.00001267, -0.00000467, 0.00000201, -0.00001065, -0.00001318, 0.00000286, 0.00000949, -0.00000093, -0.00000962, -0.00000909, -0.00000633, -0.00000536, -0.00000473, -0.00000111, 0.00000451, 0.00000391, -0.00000364, -0.00000427, 0.00000649, },
		{ -0.00000454, -0.00000794, -0.00000432, -0.00000146, 0.00000079, 0.00000338, 0.00000272, -0.00000045, -0.00000018, 0.00000366, 0.00000445, 0.00000005, -0.00000336, -0.00000179, -0.00000077, -0.00000544, -0.00000890, -0.00000214, 0.00000822, 0.00000816, -0.00000081, -0.00000510, -0.00000187, 0.00000062, -0.00000012, -0.00000078, -0.00000122, -0.00000145, -0.00000010, -0.00000002, -0.00000249, -0.00000182, 0.00000202, 0.00000139, -0.00000138, 0.00000038, -0.00000066, -0.00000720, -0.00000331, 0.00001010, 0.00000988, 0.00000050, 0.00000300, 0.00000043, -0.00002036, -0.00001802, 0.00001676, 0.00001990, -0.00001365, -0.00000944, 0.00001948, -0.00000406, -0.00003283, 0.00001200, 0.00004349, -0.00001483, -0.00004150, 0.00002331, 0.00003072, -0.00004407, -0.00003089, 0.00004696, 0.00001014, -0.00006059, 0.00000636, 0.00006721, -0.00002571, -0.00007032, 0.00003797, 0.00005203, -0.00006686, -0.00003970, 0.00009080, 0.00002878, -0.00010089, -0.00001378, 0.00008484, -0.00004351, -0.00010976, 0.00005497, 0.00008139, -0.00011326, -0.00009444, 0.00009752, -0.00006683, -0.00040691, -0.00016695, 0.00053364, 0.00074382, 0.00022969, -0.00025569, -0.00024322, -0.00003979, 0.00001439, -0.00003316, -0.00005051, -0.00002754, 0.00000200, 0.00000923, -0.00000764, -0.00001022, 0.00000587, -0.00000393, -0.00002962, -0.00001669, 0.00001600, 0.00001119, -0.00001189, -0.00000757, 0.00000322, -0.00000820, -0.00001540, -0.00000007, 0.00001048, 0.00000155, -0.00000824, -0.00000916, -0.00000821, -0.00000707, -0.00000384, -0.00000005, 0.00000347, 0.00000383, -0.00000184, -0.00000510, 0.00000289, 0.00001165, 0.00000686, },
		{ -0.00000562, -0.00000209, -0.00000026, 0.00000248, 0.00000354, 0.00000086, -0.00000047, 0.00000271, 0.00000483, 0.00000119, -0.00000335, -0.00000262, -0.00000034, -0.00000371, -0.00000856, -0.00000434, 0.00000623, 0.00000875, 0.00000078, -0.00000484, -0.00000244, 0.00000039, -0.00000013, -0.00000036, -0.00000032, -0.00000161, -0.00000145, -0.00000003, -0.00000125, -0.00000239, 0.00000054, 0.00000229, -0.00000013, -0.00000088, -0.00000100, -0.00000512, -0.00000412, 0.00000648, 0.00001038, 0.00000389, 0.00000316, 0.00000067, -0.00001741, -0.00002034, 0.00000996, 0.00002152, -0.00000682, -0.00001157, 0.00001547, 0.00000166, -0.00003152, 0.00000132, 0.00004319, -0.00000135, -0.00004209, 0.00001104, 0.00003540, -0.00003400, -0.00004119, 0.00003911, 0.00002467, -0.00005709, -0.00001156, 0.00006930, -0.00000624, -0.00007894, 0.00001939, 0.00006752, -0.00005225, -0.00006260, 0.00008173, 0.00005727, -0.00009940, -0.00004459, 0.00009446, -0.00001167, -0.00013077, 0.00002643, 0.00011581, -0.00009137, -0.00014278, 0.00009385, 0.00001988, -0.00040069, -0.00031125, 0.00044357, 0.00084098, 0.00035539, -0.00025753, -0.00029978, -0.00004512, 0.00003175, -0.00003942, -0.00005865, -0.00002337, 0.00000079, 0.00000447, -0.00000447, -0.00000820, 0.00000211, -0.00000291, -0.00002601, -0.00002005, 0.00001255, 0.00001500, -0.00000942, -0.00001050, 0.00000308, -0.00000515, -0.00001648, -0.00000354, 0.00001065, 0.00000400, -0.00000675, -0.00000865, -0.00000935, -0.00000941, -0.00000405, 0.00000155, 0.00000330, 0.00000308, -0.00000045, -0.00000479, -0.00000035, 0.00000974, 0.00000962, -0.00000149, -0.00000818, },
		{ -0.00000107, 0.00000130, 0.00000381, 0.00000232, -0.00000021, 0.00000170, 0.00000475, 0.00000233, -0.00000294, -0.00000342, -0.00000041, -0.00000214, -0.00000761, -0.00000595, 0.00000393, 0.00000869, 0.00000231, -0.00000423, -0.00000293, 0.00000010, -0.00000015, -0.00000018, 0.00000046, -0.00000128, -0.00000249, -0.00000056, -0.00000025, -0.00000235, -0.00000090, 0.00000246, 0.00000126, -0.00000148, -0.00000183, -0.00000352, -0.00000393, 0.00000328, 0.00000955, 0.00000681, 0.00000423, 0.00000079, -0.00001472, -0.00002122, 0.00000339, 0.00002083, -0.00000023, -0.00001154, 0.00001089, 0.00000577, -0.00002801, -0.00000799, 0.00003915, 0.00001086, -0.00003837, -0.00000077, 0.00003602, -0.00002257, -0.00004787, 0.00002788, 0.00003638, -0.00004850, -0.00002819, 0.00006506, 0.00001408, -0.00008072, -0.00000260, 0.00007662, -0.00003146, -0.00008061, 0.00006386, 0.00008333, -0.00008717, -0.00007652, 0.00009224, 0.00002556, -0.00014007, -0.00001233, 0.00014034, -0.00005351, -0.00018495, 0.00006531, 0.00010288, -0.00035585, -0.00044208, 0.00031134, 0.00090944, 0.00050656, -0.00023287, -0.00036651, -0.00006356, 0.00005600, -0.00004089, -0.00007268, -0.00001974, 0.00000395, -0.00000212, -0.00000414, -0.00000359, -0.00000053, -0.00000454, -0.00002209, -0.00002114, 0.00000793, 0.00001719, -0.00000546, -0.00001264, 0.00000144, -0.00000216, -0.00001613, -0.00000713, 0.00000980, 0.00000632, -0.00000513, -0.00000785, -0.00000964, -0.00001182, -0.00000550, 0.00000308, 0.00000418, 0.00000215, 0.00000017, -0.00000371, -0.00000274, 0.00000677, 0.00001122, 0.00000218, -0.00000749, -0.00000704, -0.00000289, },
		{ 0.00000348, 0.00000371, 0.00000061, 0.00000085, 0.00000423, 0.00000330, -0.00000218, -0.00000407, -0.00000089, -0.00000091, -0.00000623, -0.00000689, 0.00000158, 0.00000800, 0.00000362, -0.00000335, -0.00000327, -0.00000025, -0.00000018, -0.00000016, 0.00000102, -0.00000065, -0.00000309, -0.00000140, 0.00000036, -0.00000187, -0.00000206, 0.00000201, 0.00000251, -0.00000143, -0.00000285, -0.00000257, -0.00000307, 0.00000085, 0.00000775, 0.00000888, 0.00000598, 0.00000106, -0.00001243, -0.00002093, -0.00000242, 0.00001818, 0.00000535, -0.00000955, 0.00000653, 0.00000800, -0.00002308, -0.00001507, 0.00003222, 0.00002051, -0.00003109, -0.00001065, 0.00003284, -0.00001132, -0.00005044, 0.00001473, 0.00004401, -0.00003596, -0.00004155, 0.00005502, 0.00003274, -0.00007531, -0.00002530, 0.00007793, -0.00000700, -0.00009115, 0.00003908, 0.00010332, -0.00006501, -0.00010508, 0.00007748, 0.00006326, -0.00013515, -0.00005653, 0.00015024, -0.00000371, -0.00021339, 0.00001508, 0.00017085, -0.00027643, -0.00054532, 0.00014411, 0.00093653, 0.00067450, -0.00017347, -0.00043698, -0.00010022, 0.00008414, -0.00003374, -0.00009194, -0.00001957, 0.00001218, -0.00000852, -0.00000803, 0.00000253, -0.00000051, -0.00000854, -0.00001930, -0.00001966, 0.00000321, 0.00001710, -0.00000068, -0.00001320, -0.00000143, 0.00000000, -0.00001436, -0.00001028, 0.00000784, 0.00000828, -0.00000329, -0.00000695, -0.00000920, -0.00001378, -0.00000802, 0.00000391, 0.00000601, 0.00000160, -0.00000002, -0.00000236, -0.00000402, 0.00000327, 0.00001138, 0.00000598, -0.00000573, -0.00000829, -0.00000396, -0.00000170, 0.00000006, },
		{ 0.00000191, 0.00000034, 0.00000337, 0.00000397, -0.00000114, -0.00000444, -0.00000164, -0.00000008, -0.00000464, -0.00000717, -0.00000060, 0.00000679, 0.00000455, -0.00000228, -0.00000340, -0.00000061, -0.00000020, -0.00000023, 0.00000131, 0.00000010, -0.00000323, -0.00000230, 0.00000053, -0.00000118, -0.00000282, 0.00000114, 0.00000341, -0.00000086, -0.00000378, -0.00000228, -0.00000190, -0.00000068, 0.00000540, 0.00000991, 0.00000808, 0.00000171, -0.00001055, -0.00001981, -0.00000708, 0.00001408, 0.00000936, -0.00000607, 0.00000306, 0.00000843, -0.00001763, -0.00001948, 0.00002355, 0.00002672, -0.00002144, -0.00001746, 0.00002669, -0.00000163, -0.00004904, 0.00000128, 0.00004690, -0.00002113, -0.00005014, 0.00004056, 0.00004751, -0.00006342, -0.00004587, 0.00007115, 0.00001800, -0.00009272, 0.00001039, 0.00011439, -0.00003532, -0.00012614, 0.00005145, 0.00009620, -0.00011561, -0.00010026, 0.00014291, 0.00005182, -0.00022249, -0.00005025, 0.00021447, -0.00017104, -0.00060934, -0.00004608, 0.00091262, 0.00084637, -0.00007382, -0.00050179, -0.00015854, 0.00011100, -0.00001477, -0.00011410, -0.00002589, 0.00002498, -0.00001217, -0.00001673, 0.00000829, 0.00000338, -0.00001385, -0.00001905, -0.00001595, -0.00000026, 0.00001452, 0.00000386, -0.00001164, -0.00000486, 0.00000066, -0.00001151, -0.00001234, 0.00000490, 0.00000955, -0.00000116, -0.00000602, -0.00000831, -0.00001494, -0.00001117, 0.00000357, 0.00000836, 0.00000188, -0.00000083, -0.00000123, -0.00000419, -0.00000017, 0.00001006, 0.00000936, -0.00000292, -0.00000906, -0.00000532, -0.00000226, -0.00000110, 0.00000263, 0.00000479, },
		{ 0.00000230, 0.00000425, 0.00000004, -0.00000446, -0.00000255, 0.00000029, -0.00000303, -0.00000684, -0.00000241, 0.00000521, 0.00000501, -0.00000117, -0.00000331, -0.00000095, -0.00000023, -0.00000034, 0.00000137, 0.00000081, -0.00000295, -0.00000307, 0.00000031, -0.00000048, -0.00000315, 0.00000007, 0.00000384, 0.00000004, -0.00000442, -0.00000255, -0.00000075, -0.00000131, 0.00000293, 0.00000986, 0.00001016, 0.00000282, -0.00000899, -0.00001821, -0.00001040, 0.00000917, 0.00001147, -0.00000176, 0.00000100, 0.00000746, -0.00001250, -0.00002120, 0.00001441, 0.00002918, -0.00001089, -0.00002057, 0.00001880, 0.00000542, -0.00004439, -0.00001090, 0.00004508, -0.00000593, -0.00005322, 0.00002364, 0.00005674, -0.00004671, -0.00006180, 0.00005723, 0.00004032, -0.00008516, -0.00001857, 0.00011503, -0.00000168, -0.00013663, 0.00001718, 0.00011975, -0.00008335, -0.00013743, 0.00011838, 0.00010573, -0.00020960, -0.00012173, 0.00022783, -0.00005171, -0.00062671, -0.00024355, 0.00083277, 0.00100644, 0.00006742, -0.00054949, -0.00023912, 0.00012966, 0.00001771, -0.00013520, -0.00004121, 0.00004039, -0.00001039, -0.00002971, 0.00001132, 0.00001158, -0.00001871, -0.00002239, -0.00001107, -0.00000119, 0.00000989, 0.00000693, -0.00000789, -0.00000785, -0.00000061, -0.00000827, -0.00001281, 0.00000137, 0.00000973, 0.00000119, -0.00000494, -0.00000728, -0.00001519, -0.00001439, 0.00000190, 0.00001063, 0.00000323, -0.00000181, -0.00000070, -0.00000350, -0.00000302, 0.00000750, 0.00001180, 0.00000072, -0.00000902, -0.00000688, -0.00000288, -0.00000207, 0.00000136, 0.00000539, 0.00000354, 0.00000032, },
		{ 0.00000122, -0.00000413, -0.00000346, 0.00000024, -0.00000156, -0.00000604, -0.00000376, 0.00000347, 0.00000500, -0.00000015, -0.00000302, -0.00000123, -0.00000026, -0.00000042, 0.00000123, 0.00000135, -0.00000237, -0.00000357, -0.00000017, 0.00000008, -0.00000312, -0.00000100, 0.00000381, 0.00000107, -0.00000465, -0.00000320, 0.00000014, -0.00000117, 0.00000072, 0.00000884, 0.00001188, 0.00000439, -0.00000760, -0.00001642, -0.00001239, 0.00000409, 0.00001160, 0.00000264, 0.00000055, 0.00000565, -0.00000834, -0.00002061, 0.00000598, 0.00002811, -0.00000090, -0.00001997, 0.00001059, 0.00000931, -0.00003762, -0.00002059, 0.00003926, 0.00000781, -0.00005084, 0.00000648, 0.00005965, -0.00002746, -0.00007133, 0.00003813, 0.00005720, -0.00006970, -0.00004417, 0.00010526, 0.00003168, -0.00013505, -0.00002100, 0.00013061, -0.00004224, -0.00016285, 0.00007934, 0.00015076, -0.00017541, -0.00018947, 0.00020929, 0.00006777, -0.00059536, -0.00043083, 0.00069789, 0.00113788, 0.00024650, -0.00056794, -0.00033891, 0.00013241, 0.00006308, -0.00015016, -0.00006671, 0.00005517, -0.00000105, -0.00004520, 0.00000924, 0.00002349, -0.00002096, -0.00002963, -0.00000669, 0.00000131, 0.00000430, 0.00000737, -0.00000246, -0.00000927, -0.00000378, -0.00000559, -0.00001141, -0.00000211, 0.00000854, 0.00000354, -0.00000350, -0.00000633, -0.00001468, -0.00001710, -0.00000097, 0.00001217, 0.00000557, -0.00000245, -0.00000090, -0.00000238, -0.00000493, 0.00000413, 0.00001293, 0.00000477, -0.00000796, -0.00000843, -0.00000370, -0.00000283, -0.00000015, 0.00000539, 0.00000529, 0.00000087, 0.00000120, 0.00000407, },
		{ -0.00000426, -0.00000016, -0.00000033, -0.00000492, -0.00000459, 0.00000174, 0.00000457, 0.00000066, -0.00000258, -0.00000141, -0.00000028, -0.00000047, 0.00000098, 0.00000168, -0.00000165, -0.00000373, -0.00000076, 0.00000042, -0.00000283, -0.00000194, 0.00000336, 0.00000203, -0.00000446, -0.00000402, 0.00000061, -0.00000049, -0.00000095, 0.00000712, 0.00001297, 0.00000630, -0.00000621, -0.00001466, -0.00001320, -0.00000059, 0.00000997, 0.00000642, 0.00000166, 0.00000370, -0.00000553, -0.00001839, -0.00000084, 0.00002422, 0.00000727, -0.00001620, 0.00000343, 0.00001009, -0.00003007, -0.00002708, 0.00003063, 0.00001859, -0.00004385, -0.00000878, 0.00005642, -0.00000823, -0.00007370, 0.00001651, 0.00006676, -0.00004866, -0.00006335, 0.00008666, 0.00006067, -0.00012170, -0.00005819, 0.00012739, 0.00000249, -0.00017305, 0.00003071, 0.00018092, -0.00012384, -0.00024423, 0.00016166, 0.00017379, -0.00051892, -0.00059087, 0.00051509, 0.00122487, 0.00045422, -0.00054617, -0.00045086, 0.00011200, 0.00011797, -0.00015356, -0.00010173, 0.00006516, 0.00001685, -0.00006026, 0.00000022, 0.00003739, -0.00001851, -0.00004012, -0.00000480, 0.00000744, -0.00000067, 0.00000448, 0.00000358, -0.00000817, -0.00000830, -0.00000447, -0.00000830, -0.00000470, 0.00000590, 0.00000544, -0.00000149, -0.00000550, -0.00001377, -0.00001890, -0.00000458, 0.00001247, 0.00000852, -0.00000226, -0.00000170, -0.00000130, -0.00000577, 0.00000050, 0.00001258, 0.00000872, -0.00000580, -0.00000967, -0.00000478, -0.00000343, -0.00000171, 0.00000477, 0.00000691, 0.00000201, 0.00000027, 0.00000348, 0.00000227, -0.00000346, },
		{ 0.00000058, -0.00000362, -0.00000494, 0.00000017, 0.00000383, 0.00000119, -0.00000209, -0.00000147, -0.00000029, -0.00000047, 0.00000067, 0.00000176, -0.00000092, -0.00000357, -0.00000130, 0.00000054, -0.00000241, -0.00000266, 0.00000261, 0.00000278, -0.00000391, -0.00000484, 0.00000062, 0.00000048, -0.00000195, 0.00000498, 0.00001326, 0.00000833, -0.00000467, -0.00001304, -0.00001309, -0.00000444, 0.00000701, 0.00000901, 0.00000398, 0.00000224, -0.00000409, -0.00001534, -0.00000553, 0.00001854, 0.00001277, -0.00001028, -0.00000165, 0.00000831, -0.00002299, -0.00003025, 0.00002064, 0.00002553, -0.00003366, -0.00002047, 0.00004810, 0.00000859, -0.00006923, -0.00000473, 0.00006828, -0.00002514, -0.00007413, 0.00006206, 0.00008192, -0.00009864, -0.00008975, 0.00011079, 0.00004516, -0.00016684, -0.00002123, 0.00019250, -0.00006136, -0.00027885, 0.00009170, 0.00025484, -0.00040628, -0.00070927, 0.00029716, 0.00125488, 0.00067654, -0.00047613, -0.00056429, 0.00006304, 0.00017622, -0.00014063, -0.00014341, 0.00006603, 0.00004266, -0.00007126, -0.00001651, 0.00005058, -0.00000981, -0.00005222, -0.00000726, 0.00001649, -0.00000317, -0.00000182, 0.00000872, -0.00000396, -0.00001312, -0.00000574, -0.00000410, -0.00000559, 0.00000204, 0.00000632, 0.00000116, -0.00000462, -0.00001286, -0.00001967, -0.00000831, 0.00001127, 0.00001146, -0.00000096, -0.00000272, -0.00000068, -0.00000568, -0.00000283, 0.00001085, 0.00001202, -0.00000266, -0.00001030, -0.00000609, -0.00000396, -0.00000319, 0.00000357, 0.00000820, 0.00000364, -0.00000030, 0.00000253, 0.00000305, -0.00000253, -0.00000486, -0.00000084, },
		{ -0.00000488, -0.00000112, 0.00000292, 0.00000141, -0.00000164, -0.00000143, -0.00000025, -0.00000042, 0.00000036, 0.00000165, -0.00000031, -0.00000317, -0.00000167, 0.00000051, -0.00000196, -0.00000316, 0.00000167, 0.00000325, -0.00000309, -0.00000548, 0.00000018, 0.00000148, -0.00000223, 0.00000277, 0.00001272, 0.00001023, -0.00000289, -0.00001159, -0.00001236, -0.00000721, 0.00000325, 0.00001005, 0.00000695, 0.00000174, -0.00000377, -0.00001224, -0.00000801, 0.00001221, 0.00001527, -0.00000345, -0.00000404, 0.00000496, -0.00001736, -0.00003050, 0.00001071, 0.00002836, -0.00002199, -0.00002756, 0.00003639, 0.00002118, -0.00005923, -0.00002296, 0.00006230, -0.00000237, -0.00007589, 0.00003500, 0.00009340, -0.00006924, -0.00011204, 0.00008348, 0.00008053, -0.00014550, -0.00006982, 0.00018455, 0.00000414, -0.00028927, 0.00000893, 0.00030304, -0.00027033, -0.00077615, 0.00006121, 0.00122048, 0.00089584, -0.00035443, -0.00066586, -0.00001662, 0.00022947, -0.00010843, -0.00018688, 0.00005415, 0.00007391, -0.00007447, -0.00004022, 0.00005980, 0.00000566, -0.00006343, -0.00001539, 0.00002686, -0.00000155, -0.00001074, 0.00001128, 0.00000326, -0.00001683, -0.00000981, 0.00000018, -0.00000415, -0.00000241, 0.00000563, 0.00000433, -0.00000334, -0.00001228, -0.00001957, -0.00001149, 0.00000868, 0.00001369, 0.00000145, -0.00000345, -0.00000072, -0.00000502, -0.00000542, 0.00000806, 0.00001423, 0.00000116, -0.00001003, -0.00000751, -0.00000454, -0.00000449, 0.00000193, 0.00000896, 0.00000561, -0.00000036, 0.00000135, 0.00000347, -0.00000141, -0.00000519, -0.00000169, 0.00000117, -0.00000227, },
		{ 0.00000196, 0.00000138, -0.00000132, -0.00000134, -0.00000017, -0.00000031, 0.00000010, 0.00000139, 0.00000011, -0.00000264, -0.00000184, 0.00000041, -0.00000153, -0.00000346, 0.00000065, 0.00000339, -0.00000214, -0.00000587, -0.00000058, 0.00000229, -0.00000188, 0.00000077, 0.00001142, 0.00001178, -0.00000088, -0.00001024, -0.00001131, -0.00000883, -0.00000071, 0.00000946, 0.00000994, 0.00000243, -0.00000409, -0.00000972, -0.00000857, 0.00000629, 0.00001494, 0.00000304, -0.00000371, 0.00000115, -0.00001372, -0.00002867, 0.00000204, 0.00002747, -0.00001058, -0.00002985, 0.00002332, 0.00002849, -0.00004572, -0.00003621, 0.00005041, 0.00001673, -0.00006939, 0.00000914, 0.00009463, -0.00003761, -0.00012298, 0.00004952, 0.00010469, -0.00011242, -0.00010919, 0.00015902, 0.00006463, -0.00027502, -0.00007584, 0.00031511, -0.00012616, -0.00078743, -0.00017341, 0.00112066, 0.00109294, -0.00018335, -0.00074117, -0.00012499, 0.00026819, -0.00005667, -0.00022572, 0.00002741, 0.00010647, -0.00006686, -0.00006862, 0.00006179, 0.00002710, -0.00007083, -0.00002952, 0.00003623, 0.00000529, -0.00002072, 0.00000983, 0.00001268, -0.00001795, -0.00001649, 0.00000324, -0.00000013, -0.00000654, 0.00000294, 0.00000756, -0.00000123, -0.00001215, -0.00001906, -0.00001361, 0.00000514, 0.00001459, 0.00000463, -0.00000335, -0.00000137, -0.00000425, -0.00000703, 0.00000471, 0.00001504, 0.00000518, -0.00000871, -0.00000880, -0.00000528, -0.00000558, 0.00000001, 0.00000908, 0.00000773, 0.00000017, 0.00000013, 0.00000348, -0.00000022, -0.00000522, -0.00000264, 0.00000142, -0.00000097, -0.00000450, -0.00000211, },
		{ -0.00000118, -0.00000126, -0.00000003, -0.00000015, -0.00000009, 0.00000105, 0.00000034, -0.00000211, -0.00000180, 0.00000035, -0.00000113, -0.00000359, -0.00000037, 0.00000323, -0.00000116, -0.00000594, -0.00000155, 0.00000279, -0.00000106, -0.00000077, 0.00000955, 0.00001275, 0.00000131, -0.00000889, -0.00001018, -0.00000940, -0.00000432, 0.00000741, 0.00001230, 0.00000425, -0.00000447, -0.00000810, -0.00000779, 0.00000156, 0.00001239, 0.00000813, -0.00000112, -0.00000202, -0.00001212, -0.00002575, -0.00000463, 0.00002372, -0.00000087, -0.00002784, 0.00001088, 0.00003041, -0.00003100, -0.00004347, 0.00003494, 0.00003010, -0.00005657, -0.00001233, 0.00008666, -0.00000786, -0.00012227, 0.00001368, 0.00011556, -0.00007250, -0.00013508, 0.00012023, 0.00011322, -0.00023916, -0.00015223, 0.00029262, 0.00001103, -0.00074525, -0.00038742, 0.00096120, 0.00124945, 0.00002876, -0.00077672, -0.00025551, 0.00028307, 0.00001167, -0.00025297, -0.00001402, 0.00013504, -0.00004680, -0.00009796, 0.00005402, 0.00005226, -0.00007158, -0.00004876, 0.00004195, 0.00001751, -0.00002959, 0.00000347, 0.00002279, -0.00001524, -0.00002494, 0.00000377, 0.00000616, -0.00000921, -0.00000181, 0.00001018, 0.00000206, -0.00001229, -0.00001869, -0.00001443, 0.00000138, 0.00001377, 0.00000796, -0.00000204, -0.00000233, -0.00000382, -0.00000770, 0.00000136, 0.00001442, 0.00000889, -0.00000635, -0.00000966, -0.00000621, -0.00000650, -0.00000202, 0.00000851, 0.00000975, 0.00000131, -0.00000097, 0.00000307, 0.00000092, -0.00000494, -0.00000359, 0.00000138, 0.00000022, -0.00000389, -0.00000280, 0.00000108, 0.00000115, },
		{ 0.00000013, 0.00000007, -0.00000021, 0.00000069, 0.00000038, -0.00000167, -0.00000162, 0.00000040, -0.00000073, -0.00000360, -0.00000136, 0.00000280, -0.00000026, -0.00000570, -0.00000259, 0.00000288, 0.00000005, -0.00000172, 0.00000735, 0.00001303, 0.00000352, -0.00000745, -0.00000911, -0.00000914, -0.00000718, 0.00000428, 0.00001354, 0.00000690, -0.00000437, -0.00000741, -0.00000636, -0.00000154, 0.00000853, 0.00001116, 0.00000291, -0.00000376, -0.00001214, -0.00002273, -0.00000905, 0.00001826, 0.00000619, -0.00002263, 0.00000071, 0.00002766, -0.00001730, -0.00004473, 0.00001851, 0.00003674, -0.00004012, -0.00002724, 0.00007183, 0.00001652, -0.00011137, -0.00001941, 0.00011316, -0.00003129, -0.00014550, 0.00007414, 0.00014517, -0.00018751, -0.00021179, 0.00024139, 0.00012796, -0.00065749, -0.00056406, 0.00075419, 0.00135009, 0.00026806, -0.00076196, -0.00039735, 0.00026668, 0.00009036, -0.00026226, -0.00006744, 0.00015403, -0.00001460, -0.00012363, 0.00003524, 0.00007775, -0.00006358, -0.00007100, 0.00004156, 0.00003417, -0.00003496, -0.00000779, 0.00003155, -0.00000803, -0.00003371, 0.00000073, 0.00001385, -0.00000935, -0.00000828, 0.00001133, 0.00000663, -0.00001226, -0.00001900, -0.00001409, -0.00000181, 0.00001122, 0.00001066, 0.00000060, -0.00000306, -0.00000402, -0.00000768, -0.00000147, 0.00001256, 0.00001175, -0.00000313, -0.00000982, -0.00000728, -0.00000733, -0.00000400, 0.00000731, 0.00001143, 0.00000299, -0.00000175, 0.00000228, 0.00000188, -0.00000437, -0.00000446, 0.00000106, 0.00000126, -0.00000314, -0.00000324, 0.00000033, 0.00000085, -0.00000122, -0.00000125, },
		{ -0.00000025, 0.00000036, 0.00000029, -0.00000137, -0.00000139, 0.00000058, -0.00000026, -0.00000349, -0.00000229, 0.00000213, 0.00000050, -0.00000519, -0.00000356, 0.00000257, 0.00000122, -0.00000204, 0.00000508, 0.00001257, 0.00000559, -0.00000582, -0.00000817, -0.00000834, -0.00000903, 0.00000058, 0.00001337, 0.00000986, -0.00000342, -0.00000742, -0.00000494, -0.00000297, 0.00000431, 0.00001192, 0.00000735, -0.00000367, -0.00001308, -0.00002035, -0.00001141, 0.00001224, 0.00001020, -0.00001563, -0.00000615, 0.00002158, -0.00000638, -0.00004094, 0.00000357, 0.00003683, -0.00002297, -0.00003467, 0.00005324, 0.00003328, -0.00009309, -0.00004595, 0.00009943, 0.00000599, -0.00014084, 0.00002722, 0.00015850, -0.00012762, -0.00024909, 0.00017044, 0.00021488, -0.00053652, -0.00069105, 0.00051656, 0.00138482, 0.00051656, -0.00069105, -0.00053652, 0.00021488, 0.00017044, -0.00024909, -0.00012762, 0.00015850, 0.00002722, -0.00014084, 0.00000599, 0.00009943, -0.00004595, -0.00009309, 0.00003328, 0.00005324, -0.00003467, -0.00002297, 0.00003683, 0.00000357, -0.00004094, -0.00000638, 0.00002158, -0.00000615, -0.00001563, 0.00001020, 0.00001224, -0.00001141, -0.00002035, -0.00001308, -0.00000367, 0.00000735, 0.00001192, 0.00000431, -0.00000297, -0.00000494, -0.00000742, -0.00000342, 0.00000986, 0.00001337, 0.00000058, -0.00000903, -0.00000834, -0.00000817, -0.00000582, 0.00000559, 0.00001257, 0.00000508, -0.00000204, 0.00000122, 0.00000257, -0.00000356, -0.00000519, 0.00000050, 0.00000213, -0.00000229, -0.00000349, -0.00000026, 0.00000058, -0.00000139, -0.00000137, 0.00000029, 0.00000036, },
		{ 0.00000013, -0.00000125, -0.00000122, 0.00000085, 0.00000033, -0.00000324, -0.00000314, 0.00000126, 0.00000106, -0.00000446, -0.00000437, 0.00000188, 0.00000228, -0.00000175, 0.00000299, 0.00001143, 0.00000731, -0.00000400, -0.00000733, -0.00000728, -0.00000982, -0.00000313, 0.00001175, 0.00001256, -0.00000147, -0.00000768, -0.00000402, -0.00000306, 0.00000060, 0.00001066, 0.00001122, -0.00000181, -0.00001409, -0.00001900, -0.00001226, 0.00000663, 0.00001133, -0.00000828, -0.00000935, 0.00001385, 0.00000073, -0.00003371, -0.00000803, 0.00003155, -0.00000779, -0.00003496, 0.00003417, 0.00004156, -0.00007100, -0.00006358, 0.00007775, 0.00003524, -0.00012363, -0.00001460, 0.00015403, -0.00006744, -0.00026226, 0.00009036, 0.00026668, -0.00039735, -0.00076196, 0.00026806, 0.00135009, 0.00075419, -0.00056406, -0.00065749, 0.00012796, 0.00024139, -0.00021179, -0.00018751, 0.00014517, 0.00007414, -0.00014550, -0.00003129, 0.00011316, -0.00001941, -0.00011137, 0.00001652, 0.00007183, -0.00002724, -0.00004012, 0.00003674, 0.00001851, -0.00004473, -0.00001730, 0.00002766, 0.00000071, -0.00002263, 0.00000619, 0.00001826, -0.00000905, -0.00002273, -0.00001214, -0.00000376, 0.00000291, 0.00001116, 0.00000853, -0.00000154, -0.00000636, -0.00000741, -0.00000437, 0.00000690, 0.00001354, 0.00000428, -0.00000718, -0.00000914, -0.00000911, -0.00000745, 0.00000352, 0.00001303, 0.00000735, -0.00000172, 0.00000005, 0.00000288, -0.00000259, -0.00000570, -0.00000026, 0.00000280, -0.00000136, -0.00000360, -0.00000073, 0.00000040, -0.00000162, -0.00000167, 0.00000038, 0.00000069, -0.00000021, 0.00000007, },
		{ -0.00000118, 0.00000115, 0.00000108, -0.00000280, -0.00000389, 0.00000022, 0.00000138, -0.00000359, -0.00000494, 0.00000092, 0.00000307, -0.00000097, 0.00000131, 0.00000975, 0.00000851, -0.00000202, -0.00000650, -0.00000621, -0.00000966, -0.00000635, 0.00000889, 0.00001442, 0.00000136, -0.00000770, -0.00000382, -0.00000233, -0.00000204, 0.00000796, 0.00001377, 0.00000138, -0.00001443, -0.00001869, -0.00001229, 0.00000206, 0.00001018, -0.00000181, -0.00000921, 0.00000616, 0.00000377, -0.00002494, -0.00001524, 0.00002279, 0.00000347, -0.00002959, 0.00001751, 0.00004195, -0.00004876, -0.00007158, 0.00005226, 0.00005402, -0.00009796, -0.00004680, 0.00013504, -0.00001402, -0.00025297, 0.00001167, 0.00028307, -0.00025551, -0.00077672, 0.00002876, 0.00124945, 0.00096120, -0.00038742, -0.00074525, 0.00001103, 0.00029262, -0.00015223, -0.00023916, 0.00011322, 0.00012023, -0.00013508, -0.00007250, 0.00011556, 0.00001368, -0.00012227, -0.00000786, 0.00008666, -0.00001233, -0.00005657, 0.00003010, 0.00003494, -0.00004347, -0.00003100, 0.00003041, 0.00001088, -0.00002784, -0.00000087, 0.00002372, -0.00000463, -0.00002575, -0.00001212, -0.00000202, -0.00000112, 0.00000813, 0.00001239, 0.00000156, -0.00000779, -0.00000810, -0.00000447, 0.00000425, 0.00001230, 0.00000741, -0.00000432, -0.00000940, -0.00001018, -0.00000889, 0.00000131, 0.00001275, 0.00000955, -0.00000077, -0.00000106, 0.00000279, -0.00000155, -0.00000594, -0.00000116, 0.00000323, -0.00000037, -0.00000359, -0.00000113, 0.00000035, -0.00000180, -0.00000211, 0.00000034, 0.00000105, -0.00000009, -0.00000015, -0.00000003, -0.00000126, },
		{ 0.00000196, -0.00000211, -0.00000450, -0.00000097, 0.00000142, -0.00000264, -0.00000522, -0.00000022, 0.00000348, 0.00000013, 0.00000017, 0.00000773, 0.00000908, 0.00000001, -0.00000558, -0.00000528, -0.00000880, -0.00000871, 0.00000518, 0.00001504, 0.00000471, -0.00000703, -0.00000425, -0.00000137, -0.00000335, 0.00000463, 0.00001459, 0.00000514, -0.00001361, -0.00001906, -0.00001215, -0.00000123, 0.00000756, 0.00000294, -0.00000654, -0.00000013, 0.00000324, -0.00001649, -0.00001795, 0.00001268, 0.00000983, -0.00002072, 0.00000529, 0.00003623, -0.00002952, -0.00007083, 0.00002710, 0.00006179, -0.00006862, -0.00006686, 0.00010647, 0.00002741, -0.00022572, -0.00005667, 0.00026819, -0.00012499, -0.00074117, -0.00018335, 0.00109294, 0.00112066, -0.00017341, -0.00078743, -0.00012616, 0.00031511, -0.00007584, -0.00027502, 0.00006463, 0.00015902, -0.00010919, -0.00011242, 0.00010469, 0.00004952, -0.00012298, -0.00003761, 0.00009463, 0.00000914, -0.00006939, 0.00001673, 0.00005041, -0.00003621, -0.00004572, 0.00002849, 0.00002332, -0.00002985, -0.00001058, 0.00002747, 0.00000204, -0.00002867, -0.00001372, 0.00000115, -0.00000371, 0.00000304, 0.00001494, 0.00000629, -0.00000857, -0.00000972, -0.00000409, 0.00000243, 0.00000994, 0.00000946, -0.00000071, -0.00000883, -0.00001131, -0.00001024, -0.00000088, 0.00001178, 0.00001142, 0.00000077, -0.00000188, 0.00000229, -0.00000058, -0.00000587, -0.00000214, 0.00000339, 0.00000065, -0.00000346, -0.00000153, 0.00000041, -0.00000184, -0.00000264, 0.00000011, 0.00000139, 0.00000010, -0.00000031, -0.00000017, -0.00000134, -0.00000132, 0.00000138, },
		{ -0.00000488, -0.00000227, 0.00000117, -0.00000169, -0.00000519, -0.00000141, 0.00000347, 0.00000135, -0.00000036, 0.00000561, 0.00000896, 0.00000193, -0.00000449, -0.00000454, -0.00000751, -0.00001003, 0.00000116, 0.00001423, 0.00000806, -0.00000542, -0.00000502, -0.00000072, -0.00000345, 0.00000145, 0.00001369, 0.00000868, -0.00001149, -0.00001957, -0.00001228, -0.00000334, 0.00000433, 0.00000563, -0.00000241, -0.00000415, 0.00000018, -0.00000981, -0.00001683, 0.00000326, 0.00001128, -0.00001074, -0.00000155, 0.00002686, -0.00001539, -0.00006343, 0.00000566, 0.00005980, -0.00004022, -0.00007447, 0.00007391, 0.00005415, -0.00018688, -0.00010843, 0.00022947, -0.00001662, -0.00066586, -0.00035443, 0.00089584, 0.00122048, 0.00006121, -0.00077615, -0.00027033, 0.00030304, 0.00000893, -0.00028927, 0.00000414, 0.00018455, -0.00006982, -0.00014550, 0.00008053, 0.00008348, -0.00011204, -0.00006924, 0.00009340, 0.00003500, -0.00007589, -0.00000237, 0.00006230, -0.00002296, -0.00005923, 0.00002118, 0.00003639, -0.00002756, -0.00002199, 0.00002836, 0.00001071, -0.00003050, -0.00001736, 0.00000496, -0.00000404, -0.00000345, 0.00001527, 0.00001221, -0.00000801, -0.00001224, -0.00000377, 0.00000174, 0.00000695, 0.00001005, 0.00000325, -0.00000721, -0.00001236, -0.00001159, -0.00000289, 0.00001023, 0.00001272, 0.00000277, -0.00000223, 0.00000148, 0.00000018, -0.00000548, -0.00000309, 0.00000325, 0.00000167, -0.00000316, -0.00000196, 0.00000051, -0.00000167, -0.00000317, -0.00000031, 0.00000165, 0.00000036, -0.00000042, -0.00000025, -0.00000143, -0.00000164, 0.00000141, 0.00000292, -0.00000112, },
		{ 0.00000058, -0.00000084, -0.00000486, -0.00000253, 0.00000305, 0.00000253, -0.00000030, 0.00000364, 0.00000820, 0.00000357, -0.00000319, -0.00000396, -0.00000609, -0.00001030, -0.00000266, 0.00001202, 0.00001085, -0.00000283, -0.00000568, -0.00000068, -0.00000272, -0.00000096, 0.00001146, 0.00001127, -0.00000831, -0.00001967, -0.00001286, -0.00000462, 0.00000116, 0.00000632, 0.00000204, -0.00000559, -0.00000410, -0.00000574, -0.00001312, -0.00000396, 0.00000872, -0.00000182, -0.00000317, 0.00001649, -0.00000726, -0.00005222, -0.00000981, 0.00005058, -0.00001651, -0.00007126, 0.00004266, 0.00006603, -0.00014341, -0.00014063, 0.00017622, 0.00006304, -0.00056429, -0.00047613, 0.00067654, 0.00125488, 0.00029716, -0.00070927, -0.00040628, 0.00025484, 0.00009170, -0.00027885, -0.00006136, 0.00019250, -0.00002123, -0.00016684, 0.00004516, 0.00011079, -0.00008975, -0.00009864, 0.00008192, 0.00006206, -0.00007413, -0.00002514, 0.00006828, -0.00000473, -0.00006923, 0.00000859, 0.00004810, -0.00002047, -0.00003366, 0.00002553, 0.00002064, -0.00003025, -0.00002299, 0.00000831, -0.00000165, -0.00001028, 0.00001277, 0.00001854, -0.00000553, -0.00001534, -0.00000409, 0.00000224, 0.00000398, 0.00000901, 0.00000701, -0.00000444, -0.00001309, -0.00001304, -0.00000467, 0.00000833, 0.00001326, 0.00000498, -0.00000195, 0.00000048, 0.00000062, -0.00000484, -0.00000391, 0.00000278, 0.00000261, -0.00000266, -0.00000241, 0.00000054, -0.00000130, -0.00000357, -0.00000092, 0.00000176, 0.00000067, -0.00000047, -0.00000029, -0.00000147, -0.00000209, 0.00000119, 0.00000383, 0.00000017, -0.00000494, -0.00000362, },
		{ -0.00000426, -0.00000346, 0.00000227, 0.00000348, 0.00000027, 0.00000201, 0.00000691, 0.00000477, -0.00000171, -0.00000343, -0.00000478, -0.00000967, -0.00000580, 0.00000872, 0.00001258, 0.00000050, -0.00000577, -0.00000130, -0.00000170, -0.00000226, 0.00000852, 0.00001247, -0.00000458, -0.00001890, -0.00001377, -0.00000550, -0.00000149, 0.00000544, 0.00000590, -0.00000470, -0.00000830, -0.00000447, -0.00000830, -0.00000817, 0.00000358, 0.00000448, -0.00000067, 0.00000744, -0.00000480, -0.00004012, -0.00001851, 0.00003739, 0.00000022, -0.00006026, 0.00001685, 0.00006516, -0.00010173, -0.00015356, 0.00011797, 0.00011200, -0.00045086, -0.00054617, 0.00045422, 0.00122487, 0.00051509, -0.00059087, -0.00051892, 0.00017379, 0.00016166, -0.00024423, -0.00012384, 0.00018092, 0.00003071, -0.00017305, 0.00000249, 0.00012739, -0.00005819, -0.00012170, 0.00006067, 0.00008666, -0.00006335, -0.00004866, 0.00006676, 0.00001651, -0.00007370, -0.00000823, 0.00005642, -0.00000878, -0.00004385, 0.00001859, 0.00003063, -0.00002708, -0.00003007, 0.00001009, 0.00000343, -0.00001620, 0.00000727, 0.00002422, -0.00000084, -0.00001839, -0.00000553, 0.00000370, 0.00000166, 0.00000642, 0.00000997, -0.00000059, -0.00001320, -0.00001466, -0.00000621, 0.00000630, 0.00001297, 0.00000712, -0.00000095, -0.00000049, 0.00000061, -0.00000402, -0.00000446, 0.00000203, 0.00000336, -0.00000194, -0.00000283, 0.00000042, -0.00000076, -0.00000373, -0.00000165, 0.00000168, 0.00000098, -0.00000047, -0.00000028, -0.00000141, -0.00000258, 0.00000066, 0.00000457, 0.00000174, -0.00000459, -0.00000492, -0.00000033, -0.00000016, },
		{ 0.00000122, 0.00000407, 0.00000120, 0.00000087, 0.00000529, 0.00000539, -0.00000015, -0.00000283, -0.00000370, -0.00000843, -0.00000796, 0.00000477, 0.00001293, 0.00000413, -0.00000493, -0.00000238, -0.00000090, -0.00000245, 0.00000557, 0.00001217, -0.00000097, -0.00001710, -0.00001468, -0.00000633, -0.00000350, 0.00000354, 0.00000854, -0.00000211, -0.00001141, -0.00000559, -0.00000378, -0.00000927, -0.00000246, 0.00000737, 0.00000430, 0.00000131, -0.00000669, -0.00002963, -0.00002096, 0.00002349, 0.00000924, -0.00004520, -0.00000105, 0.00005517, -0.00006671, -0.00015016, 0.00006308, 0.00013241, -0.00033891, -0.00056794, 0.00024650, 0.00113788, 0.00069789, -0.00043083, -0.00059536, 0.00006777, 0.00020929, -0.00018947, -0.00017541, 0.00015076, 0.00007934, -0.00016285, -0.00004224, 0.00013061, -0.00002100, -0.00013505, 0.00003168, 0.00010526, -0.00004417, -0.00006970, 0.00005720, 0.00003813, -0.00007133, -0.00002746, 0.00005965, 0.00000648, -0.00005084, 0.00000781, 0.00003926, -0.00002059, -0.00003762, 0.00000931, 0.00001059, -0.00001997, -0.00000090, 0.00002811, 0.00000598, -0.00002061, -0.00000834, 0.00000565, 0.00000055, 0.00000264, 0.00001160, 0.00000409, -0.00001239, -0.00001642, -0.00000760, 0.00000439, 0.00001188, 0.00000884, 0.00000072, -0.00000117, 0.00000014, -0.00000320, -0.00000465, 0.00000107, 0.00000381, -0.00000100, -0.00000312, 0.00000008, -0.00000017, -0.00000357, -0.00000237, 0.00000135, 0.00000123, -0.00000042, -0.00000026, -0.00000123, -0.00000302, -0.00000015, 0.00000500, 0.00000347, -0.00000376, -0.00000604, -0.00000156, 0.00000024, -0.00000346, -0.00000413, },
		{ 0.00000230, 0.00000032, 0.00000354, 0.00000539, 0.00000136, -0.00000207, -0.00000288, -0.00000688, -0.00000902, 0.00000072, 0.00001180, 0.00000750, -0.00000302, -0.00000350, -0.00000070, -0.00000181, 0.00000323, 0.00001063, 0.00000190, -0.00001439, -0.00001519, -0.00000728, -0.00000494, 0.00000119, 0.00000973, 0.00000137, -0.00001281, -0.00000827, -0.00000061, -0.00000785, -0.00000789, 0.00000693, 0.00000989, -0.00000119, -0.00001107, -0.00002239, -0.00001871, 0.00001158, 0.00001132, -0.00002971, -0.00001039, 0.00004039, -0.00004121, -0.00013520, 0.00001771, 0.00012966, -0.00023912, -0.00054949, 0.00006742, 0.00100644, 0.00083277, -0.00024355, -0.00062671, -0.00005171, 0.00022783, -0.00012173, -0.00020960, 0.00010573, 0.00011838, -0.00013743, -0.00008335, 0.00011975, 0.00001718, -0.00013663, -0.00000168, 0.00011503, -0.00001857, -0.00008516, 0.00004032, 0.00005723, -0.00006180, -0.00004671, 0.00005674, 0.00002364, -0.00005322, -0.00000593, 0.00004508, -0.00001090, -0.00004439, 0.00000542, 0.00001880, -0.00002057, -0.00001089, 0.00002918, 0.00001441, -0.00002120, -0.00001250, 0.00000746, 0.00000100, -0.00000176, 0.00001147, 0.00000917, -0.00001040, -0.00001821, -0.00000899, 0.00000282, 0.00001016, 0.00000986, 0.00000293, -0.00000131, -0.00000075, -0.00000255, -0.00000442, 0.00000004, 0.00000384, 0.00000007, -0.00000315, -0.00000048, 0.00000031, -0.00000307, -0.00000295, 0.00000081, 0.00000137, -0.00000034, -0.00000023, -0.00000095, -0.00000331, -0.00000117, 0.00000501, 0.00000521, -0.00000241, -0.00000684, -0.00000303, 0.00000029, -0.00000255, -0.00000446, 0.00000004, 0.00000425, },
		{ 0.00000191, 0.00000479, 0.00000263, -0.00000110, -0.00000226, -0.00000532, -0.00000906, -0.00000292, 0.00000936, 0.00001006, -0.00000017, -0.00000419, -0.00000123, -0.00000083, 0.00000188, 0.00000836, 0.00000357, -0.00001117, -0.00001494, -0.00000831, -0.00000602, -0.00000116, 0.00000955, 0.00000490, -0.00001234, -0.00001151, 0.00000066, -0.00000486, -0.00001164, 0.00000386, 0.00001452, -0.00000026, -0.00001595, -0.00001905, -0.00001385, 0.00000338, 0.00000829, -0.00001673, -0.00001217, 0.00002498, -0.00002589, -0.00011410, -0.00001477, 0.00011100, -0.00015854, -0.00050179, -0.00007382, 0.00084637, 0.00091262, -0.00004608, -0.00060934, -0.00017104, 0.00021447, -0.00005025, -0.00022249, 0.00005182, 0.00014291, -0.00010026, -0.00011561, 0.00009620, 0.00005145, -0.00012614, -0.00003532, 0.00011439, 0.00001039, -0.00009272, 0.00001800, 0.00007115, -0.00004587, -0.00006342, 0.00004751, 0.00004056, -0.00005014, -0.00002113, 0.00004690, 0.00000128, -0.00004904, -0.00000163, 0.00002669, -0.00001746, -0.00002144, 0.00002672, 0.00002355, -0.00001948, -0.00001763, 0.00000843, 0.00000306, -0.00000607, 0.00000936, 0.00001408, -0.00000708, -0.00001981, -0.00001055, 0.00000171, 0.00000808, 0.00000991, 0.00000540, -0.00000068, -0.00000190, -0.00000228, -0.00000378, -0.00000086, 0.00000341, 0.00000114, -0.00000282, -0.00000118, 0.00000053, -0.00000230, -0.00000323, 0.00000010, 0.00000131, -0.00000023, -0.00000020, -0.00000061, -0.00000340, -0.00000228, 0.00000455, 0.00000679, -0.00000060, -0.00000717, -0.00000464, -0.00000008, -0.00000164, -0.00000444, -0.00000114, 0.00000397, 0.00000337, 0.00000034, },
		{ 0.00000348, 0.00000006, -0.00000170, -0.00000396, -0.00000829, -0.00000573, 0.00000598, 0.00001138, 0.00000327, -0.00000402, -0.00000236, -0.00000002, 0.00000160, 0.00000601, 0.00000391, -0.00000802, -0.00001378, -0.00000920, -0.00000695, -0.00000329, 0.00000828, 0.00000784, -0.00001028, -0.00001436, 0.00000000, -0.00000143, -0.00001320, -0.00000068, 0.00001710, 0.00000321, -0.00001966, -0.00001930, -0.00000854, -0.00000051, 0.00000253, -0.00000803, -0.00000852, 0.00001218, -0.00001957, -0.00009194, -0.00003374, 0.00008414, -0.00010022, -0.00043698, -0.00017347, 0.00067450, 0.00093653, 0.00014411, -0.00054532, -0.00027643, 0.00017085, 0.00001508, -0.00021339, -0.00000371, 0.00015024, -0.00005653, -0.00013515, 0.00006326, 0.00007748, -0.00010508, -0.00006501, 0.00010332, 0.00003908, -0.00009115, -0.00000700, 0.00007793, -0.00002530, -0.00007531, 0.00003274, 0.00005502, -0.00004155, -0.00003596, 0.00004401, 0.00001473, -0.00005044, -0.00001132, 0.00003284, -0.00001065, -0.00003109, 0.00002051, 0.00003222, -0.00001507, -0.00002308, 0.00000800, 0.00000653, -0.00000955, 0.00000535, 0.00001818, -0.00000242, -0.00002093, -0.00001243, 0.00000106, 0.00000598, 0.00000888, 0.00000775, 0.00000085, -0.00000307, -0.00000257, -0.00000285, -0.00000143, 0.00000251, 0.00000201, -0.00000206, -0.00000187, 0.00000036, -0.00000140, -0.00000309, -0.00000065, 0.00000102, -0.00000016, -0.00000018, -0.00000025, -0.00000327, -0.00000335, 0.00000362, 0.00000800, 0.00000158, -0.00000689, -0.00000623, -0.00000091, -0.00000089, -0.00000407, -0.00000218, 0.00000330, 0.00000423, 0.00000085, 0.00000061, 0.00000371, },
		{ -0.00000107, -0.00000289, -0.00000704, -0.00000749, 0.00000218, 0.00001122, 0.00000677, -0.00000274, -0.00000371, 0.00000017, 0.00000215, 0.00000418, 0.00000308, -0.00000550, -0.00001182, -0.00000964, -0.00000785, -0.00000513, 0.00000632, 0.00000980, -0.00000713, -0.00001613, -0.00000216, 0.00000144, -0.00001264, -0.00000546, 0.00001719, 0.00000793, -0.00002114, -0.00002209, -0.00000454, -0.00000053, -0.00000359, -0.00000414, -0.00000212, 0.00000395, -0.00001974, -0.00007268, -0.00004089, 0.00005600, -0.00006356, -0.00036651, -0.00023287, 0.00050656, 0.00090944, 0.00031134, -0.00044208, -0.00035585, 0.00010288, 0.00006531, -0.00018495, -0.00005351, 0.00014034, -0.00001233, -0.00014007, 0.00002556, 0.00009224, -0.00007652, -0.00008717, 0.00008333, 0.00006386, -0.00008061, -0.00003146, 0.00007662, -0.00000260, -0.00008072, 0.00001408, 0.00006506, -0.00002819, -0.00004850, 0.00003638, 0.00002788, -0.00004787, -0.00002257, 0.00003602, -0.00000077, -0.00003837, 0.00001086, 0.00003915, -0.00000799, -0.00002801, 0.00000577, 0.00001089, -0.00001154, -0.00000023, 0.00002083, 0.00000339, -0.00002122, -0.00001472, 0.00000079, 0.00000423, 0.00000681, 0.00000955, 0.00000328, -0.00000393, -0.00000352, -0.00000183, -0.00000148, 0.00000126, 0.00000246, -0.00000090, -0.00000235, -0.00000025, -0.00000056, -0.00000249, -0.00000128, 0.00000046, -0.00000018, -0.00000015, 0.00000010, -0.00000293, -0.00000423, 0.00000231, 0.00000869, 0.00000393, -0.00000595, -0.00000761, -0.00000214, -0.00000041, -0.00000342, -0.00000294, 0.00000233, 0.00000475, 0.00000170, -0.00000021, 0.00000232, 0.00000381, 0.00000130, },
		{ -0.00000562, -0.00000818, -0.00000149, 0.00000962, 0.00000974, -0.00000035, -0.00000479, -0.00000045, 0.00000308, 0.00000330, 0.00000155, -0.00000405, -0.00000941, -0.00000935, -0.00000865, -0.00000675, 0.00000400, 0.00001065, -0.00000354, -0.00001648, -0.00000515, 0.00000308, -0.00001050, -0.00000942, 0.00001500, 0.00001255, -0.00002005, -0.00002601, -0.00000291, 0.00000211, -0.00000820, -0.00000447, 0.00000447, 0.00000079, -0.00002337, -0.00005865, -0.00003942, 0.00003175, -0.00004512, -0.00029978, -0.00025753, 0.00035539, 0.00084098, 0.00044357, -0.00031125, -0.00040069, 0.00001988, 0.00009385, -0.00014278, -0.00009137, 0.00011581, 0.00002643, -0.00013077, -0.00001167, 0.00009446, -0.00004459, -0.00009940, 0.00005727, 0.00008173, -0.00006260, -0.00005225, 0.00006752, 0.00001939, -0.00007894, -0.00000624, 0.00006930, -0.00001156, -0.00005709, 0.00002467, 0.00003911, -0.00004119, -0.00003400, 0.00003540, 0.00001104, -0.00004209, -0.00000135, 0.00004319, 0.00000132, -0.00003152, 0.00000166, 0.00001547, -0.00001157, -0.00000682, 0.00002152, 0.00000996, -0.00002034, -0.00001741, 0.00000067, 0.00000316, 0.00000389, 0.00001038, 0.00000648, -0.00000412, -0.00000512, -0.00000100, -0.00000088, -0.00000013, 0.00000229, 0.00000054, -0.00000239, -0.00000125, -0.00000003, -0.00000145, -0.00000161, -0.00000032, -0.00000036, -0.00000013, 0.00000039, -0.00000244, -0.00000484, 0.00000078, 0.00000875, 0.00000623, -0.00000434, -0.00000856, -0.00000371, -0.00000034, -0.00000262, -0.00000335, 0.00000119, 0.00000483, 0.00000271, -0.00000047, 0.00000086, 0.00000354, 0.00000248, -0.00000026, -0.00000209, },
		{ -0.00000454, 0.00000686, 0.00001165, 0.00000289, -0.00000510, -0.00000184, 0.00000383, 0.00000347, -0.00000005, -0.00000384, -0.00000707, -0.00000821, -0.00000916, -0.00000824, 0.00000155, 0.00001048, -0.00000007, -0.00001540, -0.00000820, 0.00000322, -0.00000757, -0.00001189, 0.00001119, 0.00001600, -0.00001669, -0.00002962, -0.00000393, 0.00000587, -0.00001022, -0.00000764, 0.00000923, 0.00000200, -0.00002754, -0.00005051, -0.00003316, 0.00001439, -0.00003979, -0.00024322, -0.00025569, 0.00022969, 0.00074382, 0.00053364, -0.00016695, -0.00040691, -0.00006683, 0.00009752, -0.00009444, -0.00011326, 0.00008139, 0.00005497, -0.00010976, -0.00004351, 0.00008484, -0.00001378, -0.00010089, 0.00002878, 0.00009080, -0.00003970, -0.00006686, 0.00005203, 0.00003797, -0.00007032, -0.00002571, 0.00006721, 0.00000636, -0.00006059, 0.00001014, 0.00004696, -0.00003089, -0.00004407, 0.00003072, 0.00002331, -0.00004150, -0.00001483, 0.00004349, 0.00001200, -0.00003283, -0.00000406, 0.00001948, -0.00000944, -0.00001365, 0.00001990, 0.00001676, -0.00001802, -0.00002036, 0.00000043, 0.00000300, 0.00000050, 0.00000988, 0.00001010, -0.00000331, -0.00000720, -0.00000066, 0.00000038, -0.00000138, 0.00000139, 0.00000202, -0.00000182, -0.00000249, -0.00000002, -0.00000010, -0.00000145, -0.00000122, -0.00000078, -0.00000012, 0.00000062, -0.00000187, -0.00000510, -0.00000081, 0.00000816, 0.00000822, -0.00000214, -0.00000890, -0.00000544, -0.00000077, -0.00000179, -0.00000336, 0.00000005, 0.00000445, 0.00000366, -0.00000018, -0.00000045, 0.00000272, 0.00000338, 0.00000079, -0.00000146, -0.00000432, -0.00000794, },
		{ 0.00001215, 0.00000649, -0.00000427, -0.00000364, 0.00000391, 0.00000451, -0.00000111, -0.00000473, -0.00000536, -0.00000633, -0.00000909, -0.00000962, -0.00000093, 0.00000949, 0.00000286, -0.00001318, -0.00001065, 0.00000201, -0.00000467, -0.00001267, 0.00000668, 0.00001762, -0.00001190, -0.00003185, -0.00000716, 0.00000925, -0.00000942, -0.00001193, 0.00001108, 0.00000603, -0.00003007, -0.00004749, -0.00002567, 0.00000475, -0.00004196, -0.00019998, -0.00023670, 0.00013364, 0.00063170, 0.00057977, -0.00002381, -0.00037543, -0.00014565, 0.00007699, -0.00004818, -0.00011795, 0.00004306, 0.00007043, -0.00008124, -0.00006620, 0.00006587, 0.00001188, -0.00009253, 0.00000169, 0.00009058, -0.00001513, -0.00007379, 0.00003249, 0.00005102, -0.00005619, -0.00004200, 0.00005917, 0.00002341, -0.00005857, -0.00000548, 0.00005042, -0.00001800, -0.00005140, 0.00002234, 0.00003443, -0.00003648, -0.00002804, 0.00003966, 0.00002291, -0.00003140, -0.00001085, 0.00002220, -0.00000524, -0.00001989, 0.00001591, 0.00002311, -0.00001408, -0.00002332, -0.00000026, 0.00000383, -0.00000288, 0.00000785, 0.00001368, -0.00000124, -0.00000948, -0.00000110, 0.00000219, -0.00000215, -0.00000025, 0.00000324, -0.00000054, -0.00000371, -0.00000068, 0.00000134, -0.00000068, -0.00000210, -0.00000148, -0.00000015, 0.00000081, -0.00000132, -0.00000503, -0.00000225, 0.00000699, 0.00000968, 0.00000045, -0.00000849, -0.00000713, -0.00000171, -0.00000113, -0.00000299, -0.00000092, 0.00000366, 0.00000437, 0.00000056, -0.00000139, 0.00000147, 0.00000385, 0.00000202, -0.00000082, -0.00000331, -0.00000707, -0.00000663, 0.00000341, },
		{ -0.00000217, -0.00000535, 0.00000302, 0.00000596, -0.00000116, -0.00000626, -0.00000475, -0.00000406, -0.00000820, -0.00001078, -0.00000343, 0.00000791, 0.00000502, -0.00001025, -0.00001210, -0.00000013, -0.00000247, -0.00001198, 0.00000236, 0.00001730, -0.00000666, -0.00003211, -0.00001170, 0.00001113, -0.00000633, -0.00001573, 0.00000986, 0.00001104, -0.00002975, -0.00004790, -0.00001969, 0.00000186, -0.00004660, -0.00017029, -0.00020951, 0.00006709, 0.00051753, 0.00058514, 0.00010497, -0.00031178, -0.00020642, 0.00003665, -0.00001153, -0.00010711, 0.00000699, 0.00007226, -0.00005025, -0.00007766, 0.00004130, 0.00002940, -0.00007664, -0.00002064, 0.00008203, 0.00000784, -0.00007278, 0.00001169, 0.00005732, -0.00003863, -0.00005331, 0.00004641, 0.00003759, -0.00005140, -0.00002034, 0.00004910, -0.00000400, -0.00005498, 0.00001117, 0.00004297, -0.00002748, -0.00003940, 0.00003187, 0.00003277, -0.00002706, -0.00001792, 0.00002307, 0.00000061, -0.00002474, 0.00000975, 0.00002832, -0.00000857, -0.00002591, -0.00000167, 0.00000558, -0.00000571, 0.00000427, 0.00001665, 0.00000215, -0.00001152, -0.00000255, 0.00000428, -0.00000214, -0.00000249, 0.00000385, 0.00000143, -0.00000458, -0.00000207, 0.00000256, 0.00000073, -0.00000273, -0.00000249, -0.00000030, 0.00000096, -0.00000085, -0.00000469, -0.00000337, 0.00000538, 0.00001043, 0.00000319, -0.00000727, -0.00000853, -0.00000313, -0.00000079, -0.00000232, -0.00000160, 0.00000257, 0.00000469, 0.00000161, -0.00000181, -0.00000002, 0.00000375, 0.00000331, -0.00000002, -0.00000261, -0.00000592, -0.00000764, -0.00000016, 0.00001112, 0.00000983, },
		{ 0.00000114, 0.00000721, -0.00000003, -0.00000780, -0.00000541, -0.00000194, -0.00000637, -0.00001147, -0.00000597, 0.00000588, 0.00000637, -0.00000707, -0.00001240, -0.00000264, -0.00000133, -0.00001032, -0.00000107, 0.00001536, -0.00000196, -0.00003036, -0.00001646, 0.00001098, -0.00000195, -0.00001790, 0.00000621, 0.00001530, -0.00002643, -0.00004967, -0.00001672, 0.00000363, -0.00004995, -0.00015208, -0.00018147, 0.00002643, 0.00041191, 0.00055698, 0.00020925, -0.00022510, -0.00024186, -0.00001618, 0.00000992, -0.00008497, -0.00002157, 0.00006220, -0.00002177, -0.00007781, 0.00001546, 0.00003736, -0.00005657, -0.00003587, 0.00006729, 0.00002641, -0.00006476, -0.00000754, 0.00005675, -0.00002011, -0.00005867, 0.00003076, 0.00004737, -0.00004015, -0.00003271, 0.00004328, 0.00000950, -0.00005435, -0.00000145, 0.00004785, -0.00001554, -0.00004759, 0.00002081, 0.00004037, -0.00002004, -0.00002439, 0.00002179, 0.00000748, -0.00002754, 0.00000189, 0.00003172, -0.00000170, -0.00002770, -0.00000398, 0.00000796, -0.00000746, -0.00000064, 0.00001841, 0.00000675, -0.00001283, -0.00000504, 0.00000627, -0.00000112, -0.00000506, 0.00000356, 0.00000390, -0.00000478, -0.00000409, 0.00000326, 0.00000273, -0.00000291, -0.00000374, -0.00000062, 0.00000112, -0.00000051, -0.00000418, -0.00000407, 0.00000356, 0.00001035, 0.00000578, -0.00000526, -0.00000940, -0.00000490, -0.00000091, -0.00000148, -0.00000188, 0.00000134, 0.00000453, 0.00000274, -0.00000165, -0.00000150, 0.00000305, 0.00000447, 0.00000105, -0.00000214, -0.00000482, -0.00000765, -0.00000330, 0.00000875, 0.00001231, 0.00000104, -0.00000638, },
		{ 0.00000211, -0.00000868, -0.00000722, -0.00000057, -0.00000373, -0.00001140, -0.00000850, 0.00000347, 0.00000697, -0.00000400, -0.00001166, -0.00000497, -0.00000127, -0.00000834, -0.00000329, 0.00001243, 0.00000154, -0.00002706, -0.00002046, 0.00000885, 0.00000255, -0.00001790, 0.00000124, 0.00001764, -0.00002081, -0.00005094, -0.00001704, 0.00000753, -0.00004993, -0.00014196, -0.00015761, 0.00000580, 0.00032206, 0.00050500, 0.00028294, -0.00012660, -0.00024852, -0.00007255, 0.00001347, -0.00005740, -0.00003917, 0.00004381, 0.00000015, -0.00006843, -0.00000752, 0.00003603, -0.00003597, -0.00004301, 0.00004928, 0.00003872, -0.00005170, -0.00002288, 0.00005020, -0.00000303, -0.00005802, 0.00001438, 0.00005194, -0.00002643, -0.00004129, 0.00003386, 0.00002100, -0.00004964, -0.00001400, 0.00004853, -0.00000205, -0.00005168, 0.00000764, 0.00004476, -0.00001099, -0.00002940, 0.00001835, 0.00001458, -0.00002787, -0.00000696, 0.00003279, 0.00000610, -0.00002826, -0.00000723, 0.00001056, -0.00000772, -0.00000641, 0.00001845, 0.00001220, -0.00001293, -0.00000847, 0.00000772, 0.00000099, -0.00000754, 0.00000214, 0.00000655, -0.00000402, -0.00000651, 0.00000314, 0.00000512, -0.00000243, -0.00000515, -0.00000120, 0.00000130, -0.00000030, -0.00000365, -0.00000428, 0.00000178, 0.00000948, 0.00000791, -0.00000261, -0.00000953, -0.00000683, -0.00000159, -0.00000064, -0.00000172, 0.00000015, 0.00000389, 0.00000374, -0.00000093, -0.00000274, 0.00000180, 0.00000528, 0.00000241, -0.00000173, -0.00000400, -0.00000690, -0.00000558, 0.00000543, 0.00001343, 0.00000493, -0.00000623, -0.00000144, 0.00000768, },
		{ -0.00000974, -0.00000045, -0.00000065, -0.00001026, -0.00001082, 0.00000067, 0.00000693, -0.00000128, -0.00001016, -0.00000673, -0.00000207, -0.00000658, -0.00000430, 0.00000921, 0.00000355, -0.00002293, -0.00002305, 0.00000526, 0.00000619, -0.00001586, -0.00000379, 0.00001757, -0.00001409, -0.00005043, -0.00001996, 0.00001128, -0.00004602, -0.00013619, -0.00014047, -0.00000155, 0.00025159, 0.00043982, 0.00032436, -0.00002784, -0.00022707, -0.00012330, -0.00000041, -0.00003083, -0.00004469, 0.00002173, 0.00001303, -0.00005270, -0.00002447, 0.00002719, -0.00001822, -0.00004251, 0.00003110, 0.00004408, -0.00003612, -0.00003283, 0.00003940, 0.00001064, -0.00005219, -0.00000067, 0.00005130, -0.00001209, -0.00004539, 0.00002220, 0.00002929, -0.00004158, -0.00002503, 0.00004509, 0.00001142, -0.00005128, -0.00000624, 0.00004538, -0.00000082, -0.00003229, 0.00001308, 0.00002108, -0.00002559, -0.00001595, 0.00003124, 0.00001422, -0.00002722, -0.00001128, 0.00001288, -0.00000627, -0.00001243, 0.00001638, 0.00001796, -0.00001141, -0.00001251, 0.00000817, 0.00000411, -0.00000946, -0.00000046, 0.00000895, -0.00000216, -0.00000896, 0.00000199, 0.00000763, -0.00000114, -0.00000653, -0.00000210, 0.00000149, -0.00000020, -0.00000323, -0.00000406, 0.00000029, 0.00000792, 0.00000931, 0.00000045, -0.00000876, -0.00000867, -0.00000285, -0.00000001, -0.00000116, -0.00000084, 0.00000284, 0.00000443, 0.00000021, -0.00000351, 0.00000015, 0.00000556, 0.00000397, -0.00000118, -0.00000357, -0.00000578, -0.00000676, 0.00000174, 0.00001296, 0.00000888, -0.00000464, -0.00000419, 0.00000694, 0.00000481, -0.00000835, },
		{ 0.00000227, -0.00000792, -0.00001264, -0.00000251, 0.00000628, 0.00000101, -0.00000821, -0.00000772, -0.00000334, -0.00000541, -0.00000442, 0.00000629, 0.00000415, -0.00001873, -0.00002402, 0.00000101, 0.00000834, -0.00001237, -0.00000786, 0.00001527, -0.00000757, -0.00004761, -0.00002423, 0.00001328, -0.00003897, -0.00013147, -0.00013034, -0.00000205, 0.00020065, 0.00037132, 0.00033590, 0.00006085, -0.00018194, -0.00016056, -0.00002821, -0.00001106, -0.00003939, 0.00000071, 0.00001632, -0.00003452, -0.00003361, 0.00001371, -0.00000578, -0.00003605, 0.00001552, 0.00004303, -0.00002065, -0.00003692, 0.00002652, 0.00001964, -0.00004270, -0.00001270, 0.00004619, 0.00000111, -0.00004498, 0.00000988, 0.00003372, -0.00003130, -0.00003337, 0.00003815, 0.00002338, -0.00004662, -0.00001934, 0.00004215, 0.00000934, -0.00003266, 0.00000654, 0.00002626, -0.00002089, -0.00002417, 0.00002704, 0.00002196, -0.00002436, -0.00001579, 0.00001440, -0.00000310, -0.00001795, 0.00001211, 0.00002330, -0.00000806, -0.00001664, 0.00000728, 0.00000793, -0.00001036, -0.00000409, 0.00001061, 0.00000080, -0.00001101, -0.00000024, 0.00000986, 0.00000102, -0.00000766, -0.00000336, 0.00000166, -0.00000013, -0.00000305, -0.00000352, -0.00000069, 0.00000591, 0.00000979, 0.00000361, -0.00000704, -0.00001013, -0.00000462, 0.00000025, -0.00000030, -0.00000147, 0.00000151, 0.00000465, 0.00000159, -0.00000370, -0.00000166, 0.00000520, 0.00000556, -0.00000033, -0.00000347, -0.00000464, -0.00000683, -0.00000171, 0.00001093, 0.00001220, -0.00000162, -0.00000642, 0.00000487, 0.00000738, -0.00000654, -0.00001228, -0.00000184, },
		{ -0.00001354, -0.00000599, 0.00000500, 0.00000285, -0.00000608, -0.00000798, -0.00000465, -0.00000494, -0.00000409, 0.00000405, 0.00000374, -0.00001509, -0.00002355, -0.00000310, 0.00000884, -0.00000826, -0.00001032, 0.00001141, -0.00000232, -0.00004267, -0.00002842, 0.00001281, -0.00003027, -0.00012551, -0.00012586, -0.00000089, 0.00016675, 0.00030749, 0.00032306, 0.00013173, -0.00012035, -0.00017890, -0.00006414, -0.00000219, -0.00002654, -0.00001527, 0.00001130, -0.00001772, -0.00003480, -0.00000108, 0.00000014, -0.00002615, 0.00000446, 0.00003708, -0.00000756, -0.00003567, 0.00001379, 0.00002364, -0.00003142, -0.00002069, 0.00003793, 0.00001172, -0.00004069, -0.00000158, 0.00003420, -0.00002019, -0.00003830, 0.00002876, 0.00003263, -0.00003842, -0.00003031, 0.00003547, 0.00001841, -0.00003045, -0.00000053, 0.00002955, -0.00001423, -0.00003081, 0.00002045, 0.00002856, -0.00001965, -0.00002031, 0.00001468, 0.00000151, -0.00002221, 0.00000581, 0.00002745, -0.00000290, -0.00002025, 0.00000483, 0.00001198, -0.00000985, -0.00000843, 0.00001108, 0.00000465, -0.00001220, -0.00000346, 0.00001142, 0.00000398, -0.00000829, -0.00000496, 0.00000174, -0.00000001, -0.00000315, -0.00000285, -0.00000103, 0.00000373, 0.00000924, 0.00000650, -0.00000444, -0.00001096, -0.00000676, -0.00000001, 0.00000072, -0.00000166, 0.00000007, 0.00000436, 0.00000299, -0.00000326, -0.00000337, 0.00000418, 0.00000694, 0.00000093, -0.00000353, -0.00000380, -0.00000601, -0.00000439, 0.00000767, 0.00001427, 0.00000249, -0.00000752, 0.00000167, 0.00000910, -0.00000337, -0.00001405, -0.00000466, 0.00000437, -0.00000448, },
		{ 0.00000301, 0.00000426, -0.00000394, -0.00000766, -0.00000569, -0.00000504, -0.00000372, 0.00000256, 0.00000283, -0.00001235, -0.00002211, -0.00000647, 0.00000794, -0.00000437, -0.00001107, 0.00000692, 0.00000104, -0.00003635, -0.00003141, 0.00000998, -0.00002158, -0.00011718, -0.00012478, -0.00000146, 0.00014573, 0.00025364, 0.00029328, 0.00018038, -0.00005108, -0.00017606, -0.00010126, -0.00000599, -0.00001064, -0.00002375, 0.00000070, -0.00000536, -0.00002938, -0.00001406, -0.00000030, -0.00001556, -0.00000125, 0.00002840, 0.00000162, -0.00003045, 0.00000309, 0.00002311, -0.00002025, -0.00002434, 0.00002813, 0.00001887, -0.00003362, -0.00001092, 0.00003122, -0.00000960, -0.00003965, 0.00001821, 0.00003843, -0.00002782, -0.00003815, 0.00002618, 0.00002542, -0.00002597, -0.00000735, 0.00003065, -0.00000628, -0.00003519, 0.00001200, 0.00003332, -0.00001330, -0.00002428, 0.00001344, 0.00000709, -0.00002457, -0.00000206, 0.00002970, 0.00000375, -0.00002269, 0.00000088, 0.00001569, -0.00000775, -0.00001297, 0.00001004, 0.00000901, -0.00001215, -0.00000739, 0.00001194, 0.00000753, -0.00000819, -0.00000682, 0.00000166, 0.00000023, -0.00000356, -0.00000225, -0.00000070, 0.00000171, 0.00000772, 0.00000877, -0.00000112, -0.00001092, -0.00000904, -0.00000088, 0.00000171, -0.00000140, -0.00000128, 0.00000356, 0.00000417, -0.00000226, -0.00000473, 0.00000259, 0.00000788, 0.00000256, -0.00000353, -0.00000344, -0.00000465, -0.00000595, 0.00000372, 0.00001468, 0.00000708, -0.00000701, -0.00000212, 0.00000940, 0.00000071, -0.00001439, -0.00000847, 0.00000502, -0.00000032, -0.00001314, -0.00000954, },
		{ -0.00000185, -0.00000697, -0.00000631, -0.00000544, -0.00000363, 0.00000164, 0.00000188, -0.00001054, -0.00002028, -0.00000880, 0.00000617, -0.00000129, -0.00001039, 0.00000266, 0.00000244, -0.00002959, -0.00003254, 0.00000554, -0.00001428, -0.00010649, -0.00012469, -0.00000532, 0.00013287, 0.00021223, 0.00025445, 0.00020587, 0.00001706, -0.00015306, -0.00013263, -0.00002166, 0.00000349, -0.00002411, -0.00001200, 0.00000082, -0.00001978, -0.00002288, -0.00000573, -0.00000679, -0.00000192, 0.00001929, 0.00000631, -0.00002308, -0.00000437, 0.00001923, -0.00001078, -0.00002409, 0.00001840, 0.00002232, -0.00002514, -0.00001734, 0.00002567, -0.00000068, -0.00003773, 0.00000783, 0.00004055, -0.00001615, -0.00004228, 0.00001538, 0.00002970, -0.00001979, -0.00001319, 0.00002954, 0.00000215, -0.00003691, 0.00000244, 0.00003573, -0.00000573, -0.00002714, 0.00001061, 0.00001296, -0.00002458, -0.00001078, 0.00002947, 0.00001132, -0.00002336, -0.00000430, 0.00001843, -0.00000406, -0.00001709, 0.00000732, 0.00001334, -0.00001058, -0.00001160, 0.00001115, 0.00001136, -0.00000717, -0.00000880, 0.00000132, 0.00000066, -0.00000420, -0.00000195, 0.00000023, 0.00000018, 0.00000539, 0.00001011, 0.00000263, -0.00000988, -0.00001119, -0.00000234, 0.00000247, -0.00000074, -0.00000238, 0.00000237, 0.00000497, -0.00000087, -0.00000554, 0.00000065, 0.00000818, 0.00000441, -0.00000323, -0.00000356, -0.00000318, -0.00000628, -0.00000026, 0.00001332, 0.00001139, -0.00000475, -0.00000576, 0.00000796, 0.00000494, -0.00001289, -0.00001253, 0.00000384, 0.00000392, -0.00001121, -0.00001279, 0.00000025, 0.00000517, },
		{ -0.00000653, -0.00000583, -0.00000389, 0.00000099, 0.00000120, -0.00000942, -0.00001856, -0.00001014, 0.00000415, 0.00000073, -0.00000886, -0.00000076, 0.00000222, -0.00002328, -0.00003171, 0.00000048, -0.00000920, -0.00009420, -0.00012363, -0.00001245, 0.00012382, 0.00018319, 0.00021380, 0.00021042, 0.00007645, -0.00011381, -0.00015260, -0.00004608, 0.00001175, -0.00001761, -0.00002333, 0.00000060, -0.00000892, -0.00002634, -0.00001399, -0.00000160, 0.00000121, 0.00001174, 0.00000686, -0.00001546, -0.00000814, 0.00001353, -0.00000401, -0.00002093, 0.00001009, 0.00002241, -0.00001662, -0.00002060, 0.00001872, 0.00000579, -0.00003331, -0.00000125, 0.00003929, -0.00000475, -0.00004265, 0.00000430, 0.00003097, -0.00001269, -0.00001750, 0.00002647, 0.00001024, -0.00003584, -0.00000734, 0.00003546, 0.00000244, -0.00002841, 0.00000636, 0.00001840, -0.00002209, -0.00001944, 0.00002648, 0.00001901, -0.00002188, -0.00001020, 0.00001968, 0.00000095, -0.00002018, 0.00000301, 0.00001703, -0.00000746, -0.00001555, 0.00000892, 0.00001506, -0.00000514, -0.00001070, 0.00000064, 0.00000130, -0.00000496, -0.00000214, 0.00000159, -0.00000059, 0.00000252, 0.00001030, 0.00000646, -0.00000777, -0.00001294, -0.00000432, 0.00000284, 0.00000021, -0.00000310, 0.00000095, 0.00000525, 0.00000069, -0.00000570, -0.00000139, 0.00000776, 0.00000626, -0.00000246, -0.00000402, -0.00000198, -0.00000551, -0.00000363, 0.00001040, 0.00001469, -0.00000094, -0.00000845, 0.00000487, 0.00000847, -0.00000952, -0.00001594, 0.00000079, 0.00000745, -0.00000775, -0.00001521, -0.00000326, 0.00000547, 0.00000019, -0.00000608, },
		{ -0.00000440, 0.00000030, 0.00000087, -0.00000863, -0.00001725, -0.00001078, 0.00000236, 0.00000178, -0.00000709, -0.00000311, 0.00000101, -0.00001803, -0.00002929, -0.00000424, -0.00000653, -0.00008148, -0.00012040, -0.00002175, 0.00011527, 0.00016451, 0.00017689, 0.00019852, 0.00012170, -0.00006424, -0.00015759, -0.00007452, 0.00001147, -0.00000696, -0.00003056, -0.00000485, 0.00000045, -0.00002453, -0.00002263, -0.00000073, 0.00000628, 0.00000707, 0.00000438, -0.00000920, -0.00000858, 0.00000757, -0.00000030, -0.00001616, 0.00000405, 0.00001999, -0.00000920, -0.00002096, 0.00001154, 0.00000953, -0.00002737, -0.00000824, 0.00003532, 0.00000524, -0.00003966, -0.00000589, 0.00002932, -0.00000555, -0.00001996, 0.00002190, 0.00001725, -0.00003217, -0.00001644, 0.00003250, 0.00001048, -0.00002778, 0.00000108, 0.00002270, -0.00001725, -0.00002708, 0.00002078, 0.00002593, -0.00001810, -0.00001610, 0.00001907, 0.00000678, -0.00002169, -0.00000257, 0.00001948, -0.00000295, -0.00001868, 0.00000533, 0.00001817, -0.00000211, -0.00001228, -0.00000043, 0.00000213, -0.00000568, -0.00000291, 0.00000311, -0.00000043, -0.00000051, 0.00000925, 0.00000997, -0.00000469, -0.00001402, -0.00000667, 0.00000270, 0.00000127, -0.00000340, -0.00000052, 0.00000501, 0.00000218, -0.00000521, -0.00000322, 0.00000664, 0.00000782, -0.00000116, -0.00000457, -0.00000133, -0.00000397, -0.00000591, 0.00000640, 0.00001637, 0.00000389, -0.00000955, 0.00000059, 0.00001050, -0.00000468, -0.00001784, -0.00000375, 0.00000950, -0.00000308, -0.00001629, -0.00000730, 0.00000495, 0.00000218, -0.00000507, -0.00000648, -0.00000599, },
		{ 0.00000073, -0.00000783, -0.00001637, -0.00001114, 0.00000106, 0.00000221, -0.00000557, -0.00000450, -0.00000054, -0.00001407, -0.00002593, -0.00000796, -0.00000591, -0.00006947, -0.00011462, -0.00003162, 0.00010521, 0.00015314, 0.00014719, 0.00017587, 0.00015010, -0.00001120, -0.00014664, -0.00010156, 0.00000194, 0.00000432, -0.00003225, -0.00001333, 0.00000628, -0.00001862, -0.00002945, -0.00000388, 0.00001131, 0.00000577, 0.00000041, -0.00000533, -0.00000661, 0.00000265, 0.00000062, -0.00001108, 0.00000059, 0.00001612, -0.00000363, -0.00001906, 0.00000514, 0.00001067, -0.00002096, -0.00001276, 0.00002958, 0.00001303, -0.00003406, -0.00001425, 0.00002522, 0.00000084, -0.00002054, 0.00001643, 0.00002262, -0.00002636, -0.00002402, 0.00002712, 0.00001760, -0.00002516, -0.00000464, 0.00002531, -0.00001054, -0.00003284, 0.00001274, 0.00003117, -0.00001222, -0.00002124, 0.00001646, 0.00001277, -0.00002125, -0.00000887, 0.00002021, 0.00000256, -0.00002049, 0.00000061, 0.00002027, 0.00000177, -0.00001329, -0.00000190, 0.00000310, -0.00000617, -0.00000431, 0.00000447, 0.00000071, -0.00000330, 0.00000698, 0.00001278, -0.00000080, -0.00001420, -0.00000916, 0.00000200, 0.00000228, -0.00000328, -0.00000185, 0.00000431, 0.00000336, -0.00000421, -0.00000460, 0.00000497, 0.00000883, 0.00000057, -0.00000490, -0.00000131, -0.00000213, -0.00000686, 0.00000199, 0.00001613, 0.00000896, -0.00000865, -0.00000410, 0.00001050, 0.00000084, -0.00001761, -0.00000902, 0.00000949, 0.00000214, -0.00001562, -0.00001147, 0.00000343, 0.00000402, -0.00000390, -0.00000633, -0.00000583, -0.00000493, -0.00000065, },
		{ -0.00001576, -0.00001159, 0.00000023, 0.00000244, -0.00000451, -0.00000530, -0.00000195, -0.00001124, -0.00002227, -0.00001042, -0.00000668, -0.00005898, -0.00010666, -0.00004054, 0.00009299, 0.00014581, 0.00012599, 0.00014832, 0.00016168, 0.00003873, -0.00012133, -0.00012215, -0.00001555, 0.00001273, -0.00002841, -0.00002223, 0.00000758, -0.00001055, -0.00003293, -0.00000988, 0.00001461, 0.00000746, -0.00000347, -0.00000414, -0.00000349, -0.00000041, -0.00000047, -0.00000678, -0.00000050, 0.00001191, -0.00000020, -0.00001580, 0.00000022, 0.00000973, -0.00001503, -0.00001486, 0.00002309, 0.00001821, -0.00002684, -0.00002015, 0.00001939, 0.00000585, -0.00001942, 0.00001069, 0.00002602, -0.00001907, -0.00002946, 0.00001986, 0.00002313, -0.00002072, -0.00001013, 0.00002588, -0.00000266, -0.00003603, 0.00000310, 0.00003399, -0.00000474, -0.00002489, 0.00001203, 0.00001821, -0.00001874, -0.00001521, 0.00001894, 0.00000849, -0.00002062, -0.00000478, 0.00002100, 0.00000621, -0.00001351, -0.00000371, 0.00000408, -0.00000626, -0.00000626, 0.00000535, 0.00000274, -0.00000544, 0.00000370, 0.00001454, 0.00000361, -0.00001335, -0.00001154, 0.00000076, 0.00000306, -0.00000285, -0.00000290, 0.00000330, 0.00000409, -0.00000290, -0.00000535, 0.00000301, 0.00000911, 0.00000252, -0.00000476, -0.00000183, -0.00000046, -0.00000648, -0.00000209, 0.00001400, 0.00001341, -0.00000575, -0.00000827, 0.00000835, 0.00000606, -0.00001502, -0.00001402, 0.00000719, 0.00000705, -0.00001303, -0.00001518, 0.00000080, 0.00000553, -0.00000250, -0.00000619, -0.00000544, -0.00000521, -0.00000186, 0.00000053, -0.00000682, },
		{ -0.00000035, 0.00000285, -0.00000388, -0.00000593, -0.00000299, -0.00000917, -0.00001881, -0.00001172, -0.00000813, -0.00005040, -0.00009729, -0.00004738, 0.00007898, 0.00013969, 0.00011279, 0.00012093, 0.00015871, 0.00008021, -0.00008532, -0.00013245, -0.00003804, 0.00001554, -0.00002045, -0.00002911, 0.00000457, -0.00000249, -0.00003249, -0.00001706, 0.00001513, 0.00001110, -0.00000597, -0.00000528, -0.00000050, -0.00000139, -0.00000256, -0.00000393, 0.00000015, 0.00000827, 0.00000123, -0.00001209, -0.00000293, 0.00000743, -0.00001023, -0.00001491, 0.00001678, 0.00002080, -0.00001902, -0.00002336, 0.00001267, 0.00000906, -0.00001702, 0.00000529, 0.00002734, -0.00001106, -0.00003234, 0.00001146, 0.00002652, -0.00001486, -0.00001474, 0.00002437, 0.00000552, -0.00003627, -0.00000721, 0.00003390, 0.00000357, -0.00002646, 0.00000621, 0.00002239, -0.00001432, -0.00002084, 0.00001563, 0.00001417, -0.00001888, -0.00001030, 0.00002017, 0.00001084, -0.00001280, -0.00000574, 0.00000495, -0.00000582, -0.00000859, 0.00000547, 0.00000545, -0.00000659, -0.00000030, 0.00001499, 0.00000817, -0.00001144, -0.00001356, -0.00000092, 0.00000350, -0.00000225, -0.00000359, 0.00000218, 0.00000431, -0.00000155, -0.00000540, 0.00000108, 0.00000859, 0.00000438, -0.00000399, -0.00000265, 0.00000067, -0.00000506, -0.00000521, 0.00001036, 0.00001645, -0.00000126, -0.00001106, 0.00000434, 0.00000996, -0.00001033, -0.00001771, 0.00000283, 0.00001070, -0.00000872, -0.00001777, -0.00000283, 0.00000639, -0.00000079, -0.00000607, -0.00000500, -0.00000506, -0.00000320, 0.00000000, -0.00000557, -0.00001510, -0.00001233, },
		{ -0.00000338, -0.00000672, -0.00000374, -0.00000743, -0.00001576, -0.00001223, -0.00000970, -0.00004366, -0.00008739, -0.00005164, 0.00006414, 0.00013280, 0.00010584, 0.00009732, 0.00014500, 0.00010981, -0.00004358, -0.00013051, -0.00006158, 0.00001133, -0.00001074, -0.00003228, -0.00000156, 0.00000368, -0.00002852, -0.00002362, 0.00001261, 0.00001529, -0.00000633, -0.00000787, 0.00000137, -0.00000060, -0.00000465, -0.00000274, 0.00000173, 0.00000577, 0.00000116, -0.00000871, -0.00000434, 0.00000455, -0.00000693, -0.00001353, 0.00001134, 0.00002114, -0.00001152, -0.00002398, 0.00000595, 0.00001038, -0.00001385, 0.00000067, 0.00002674, -0.00000313, -0.00003258, 0.00000277, 0.00002750, -0.00000821, -0.00001792, 0.00002100, 0.00001311, -0.00003357, -0.00001711, 0.00003075, 0.00001179, -0.00002564, -0.00000033, 0.00002477, -0.00000841, -0.00002508, 0.00001052, 0.00001889, -0.00001537, -0.00001534, 0.00001774, 0.00001522, -0.00001109, -0.00000783, 0.00000556, -0.00000476, -0.00001104, 0.00000463, 0.00000849, -0.00000652, -0.00000461, 0.00001401, 0.00001249, -0.00000851, -0.00001499, -0.00000288, 0.00000353, -0.00000163, -0.00000393, 0.00000115, 0.00000404, -0.00000043, -0.00000484, -0.00000050, 0.00000736, 0.00000583, -0.00000261, -0.00000342, 0.00000104, -0.00000308, -0.00000696, 0.00000586, 0.00001754, 0.00000409, -0.00001179, -0.00000077, 0.00001175, -0.00000427, -0.00001920, -0.00000291, 0.00001228, -0.00000324, -0.00001864, -0.00000712, 0.00000626, 0.00000117, -0.00000585, -0.00000472, -0.00000447, -0.00000436, -0.00000100, -0.00000429, -0.00001412, -0.00001333, -0.00000104, 0.00000359, },
		{ -0.00000447, -0.00000570, -0.00001305, -0.00001236, -0.00001110, -0.00003843, -0.00007772, -0.00005335, 0.00004964, 0.00012418, 0.00010286, 0.00007947, 0.00012499, 0.00012628, -0.00000146, -0.00011645, -0.00008199, 0.00000027, -0.00000202, -0.00003115, -0.00000896, 0.00000674, -0.00002216, -0.00002811, 0.00000755, 0.00001867, -0.00000448, -0.00001079, 0.00000160, 0.00000125, -0.00000596, -0.00000297, 0.00000341, 0.00000460, 0.00000023, -0.00000617, -0.00000439, 0.00000180, -0.00000515, -0.00001141, 0.00000718, 0.00001981, -0.00000506, -0.00002245, -0.00000004, 0.00000994, -0.00001047, -0.00000286, 0.00002454, 0.00000402, -0.00003036, -0.00000534, 0.00002605, -0.00000148, -0.00001930, 0.00001626, 0.00001934, -0.00002825, -0.00002557, 0.00002479, 0.00001893, -0.00002245, -0.00000679, 0.00002505, -0.00000166, -0.00002739, 0.00000409, 0.00002207, -0.00001040, -0.00001933, 0.00001389, 0.00001891, -0.00000842, -0.00000978, 0.00000580, -0.00000310, -0.00001332, 0.00000273, 0.00001146, -0.00000511, -0.00000875, 0.00001160, 0.00001618, -0.00000474, -0.00001566, -0.00000493, 0.00000316, -0.00000116, -0.00000401, 0.00000038, 0.00000344, 0.00000025, -0.00000387, -0.00000148, 0.00000565, 0.00000659, -0.00000077, -0.00000378, 0.00000066, -0.00000106, -0.00000721, 0.00000130, 0.00001651, 0.00000936, -0.00001021, -0.00000601, 0.00001100, 0.00000212, -0.00001798, -0.00000898, 0.00001131, 0.00000250, -0.00001739, -0.00001149, 0.00000489, 0.00000320, -0.00000533, -0.00000474, -0.00000359, -0.00000507, -0.00000246, -0.00000327, -0.00001265, -0.00001437, -0.00000213, 0.00000459, -0.00000267, -0.00000781, },
		{ -0.00001041, -0.00001239, -0.00001233, -0.00003427, -0.00006870, -0.00005297, 0.00003641, 0.00011373, 0.00010159, 0.00006779, 0.00010292, 0.00013040, 0.00003629, -0.00009228, -0.00009569, -0.00001598, 0.00000323, -0.00002624, -0.00001566, 0.00000633, -0.00001499, -0.00002971, 0.00000103, 0.00002019, -0.00000092, -0.00001301, 0.00000017, 0.00000331, -0.00000609, -0.00000409, 0.00000456, 0.00000462, -0.00000089, -0.00000467, -0.00000357, -0.00000035, -0.00000464, -0.00000920, 0.00000441, 0.00001748, -0.00000006, -0.00001936, -0.00000475, 0.00000811, -0.00000738, -0.00000521, 0.00002123, 0.00000982, -0.00002613, -0.00001214, 0.00002247, 0.00000460, -0.00001881, 0.00001081, 0.00002363, -0.00002098, -0.00003174, 0.00001664, 0.00002412, -0.00001723, -0.00001237, 0.00002323, 0.00000517, -0.00002750, -0.00000298, 0.00002327, -0.00000449, -0.00002180, 0.00000897, 0.00002153, -0.00000496, -0.00001138, 0.00000556, -0.00000092, -0.00001512, -0.00000019, 0.00001392, -0.00000243, -0.00001225, 0.00000793, 0.00001888, -0.00000036, -0.00001545, -0.00000688, 0.00000248, -0.00000093, -0.00000399, -0.00000005, 0.00000273, 0.00000038, -0.00000275, -0.00000170, 0.00000378, 0.00000648, 0.00000122, -0.00000347, -0.00000025, 0.00000048, -0.00000614, -0.00000256, 0.00001356, 0.00001360, -0.00000649, -0.00001028, 0.00000780, 0.00000764, -0.00001408, -0.00001418, 0.00000778, 0.00000745, -0.00001399, -0.00001518, 0.00000218, 0.00000495, -0.00000432, -0.00000506, -0.00000267, -0.00000511, -0.00000414, -0.00000281, -0.00001076, -0.00001511, -0.00000377, 0.00000557, -0.00000146, -0.00000907, -0.00000552, -0.00000386, },
		{ -0.00001359, -0.00003082, -0.00006046, -0.00005111, 0.00002502, 0.00010198, 0.00010029, 0.00006149, 0.00008217, 0.00012448, 0.00006612, -0.00006144, -0.00010035, -0.00003464, 0.00000336, -0.00001907, -0.00002005, 0.00000302, -0.00000860, -0.00002831, -0.00000559, 0.00001934, 0.00000344, -0.00001377, -0.00000248, 0.00000477, -0.00000504, -0.00000544, 0.00000484, 0.00000541, -0.00000167, -0.00000414, -0.00000244, -0.00000166, -0.00000498, -0.00000737, 0.00000287, 0.00001477, 0.00000336, -0.00001540, -0.00000788, 0.00000538, -0.00000495, -0.00000643, 0.00001729, 0.00001395, -0.00002050, -0.00001704, 0.00001727, 0.00000939, -0.00001659, 0.00000537, 0.00002569, -0.00001262, -0.00003506, 0.00000717, 0.00002674, -0.00001062, -0.00001638, 0.00001958, 0.00001134, -0.00002541, -0.00000991, 0.00002232, 0.00000171, -0.00002249, 0.00000346, 0.00002281, -0.00000098, -0.00001244, 0.00000482, 0.00000163, -0.00001615, -0.00000393, 0.00001548, 0.00000130, -0.00001469, 0.00000328, 0.00002030, 0.00000430, -0.00001434, -0.00000854, 0.00000159, -0.00000100, -0.00000403, -0.00000013, 0.00000214, -0.00000004, -0.00000183, -0.00000119, 0.00000213, 0.00000549, 0.00000299, -0.00000238, -0.00000134, 0.00000120, -0.00000416, -0.00000511, 0.00000925, 0.00001605, -0.00000125, -0.00001267, 0.00000273, 0.00001120, -0.00000805, -0.00001737, 0.00000219, 0.00001057, -0.00000884, -0.00001740, -0.00000168, 0.00000596, -0.00000273, -0.00000550, -0.00000205, -0.00000447, -0.00000568, -0.00000311, -0.00000870, -0.00001524, -0.00000585, 0.00000615, 0.00000034, -0.00001019, -0.00000714, -0.00000209, -0.00000752, -0.00001241, },
		{ -0.00005288, -0.00004830, 0.00001561, 0.00008970, 0.00009790, 0.00005908, 0.00006493, 0.00011174, 0.00008616, -0.00002805, -0.00009525, -0.00005243, -0.00000208, -0.00001168, -0.00002125, -0.00000199, -0.00000421, -0.00002453, -0.00001106, 0.00001625, 0.00000751, -0.00001279, -0.00000559, 0.00000512, -0.00000316, -0.00000642, 0.00000422, 0.00000645, -0.00000180, -0.00000426, -0.00000144, -0.00000212, -0.00000573, -0.00000619, 0.00000223, 0.00001221, 0.00000532, -0.00001122, -0.00000940, 0.00000228, -0.00000340, -0.00000674, 0.00001323, 0.00001629, -0.00001421, -0.00001976, 0.00001116, 0.00001245, -0.00001305, 0.00000059, 0.00002555, -0.00000411, -0.00003533, -0.00000257, 0.00002648, -0.00000348, -0.00001838, 0.00001464, 0.00001618, -0.00002140, -0.00001596, 0.00001932, 0.00000748, -0.00002134, -0.00000207, 0.00002263, 0.00000320, -0.00001281, 0.00000361, 0.00000434, -0.00001619, -0.00000820, 0.00001579, 0.00000573, -0.00001575, -0.00000194, 0.00002029, 0.00000890, -0.00001237, -0.00000978, 0.00000066, -0.00000133, -0.00000431, 0.00000004, 0.00000189, -0.00000088, -0.00000137, -0.00000009, 0.00000106, 0.00000378, 0.00000415, -0.00000057, -0.00000217, 0.00000097, -0.00000187, -0.00000603, 0.00000437, 0.00001624, 0.00000459, -0.00001259, -0.00000323, 0.00001204, -0.00000093, -0.00001772, -0.00000447, 0.00001109, -0.00000274, -0.00001752, -0.00000617, 0.00000579, -0.00000067, -0.00000577, -0.00000195, -0.00000331, -0.00000669, -0.00000412, -0.00000687, -0.00001457, -0.00000809, 0.00000601, 0.00000259, -0.00001076, -0.00000936, -0.00000069, -0.00000419, -0.00001226, -0.00001512, -0.00002796, },
		{ 0.00000800, 0.00007760, 0.00009409, 0.00005892, 0.00005210, 0.00009551, 0.00009624, 0.00000387, -0.00008129, -0.00006624, -0.00001229, -0.00000617, -0.00001928, -0.00000719, -0.00000242, -0.00001947, -0.00001449, 0.00001156, 0.00001039, -0.00001027, -0.00000836, 0.00000419, -0.00000101, -0.00000665, 0.00000296, 0.00000729, -0.00000125, -0.00000467, -0.00000086, -0.00000194, -0.00000647, -0.00000571, 0.00000211, 0.00001009, 0.00000610, -0.00000733, -0.00000948, -0.00000068, -0.00000276, -0.00000642, 0.00000945, 0.00001692, -0.00000796, -0.00002025, 0.00000489, 0.00001356, -0.00000877, -0.00000298, 0.00002353, 0.00000365, -0.00003273, -0.00001151, 0.00002342, 0.00000325, -0.00001820, 0.00000913, 0.00001925, -0.00001600, -0.00002050, 0.00001464, 0.00001219, -0.00001853, -0.00000706, 0.00002101, 0.00000720, -0.00001242, 0.00000201, 0.00000694, -0.00001513, -0.00001260, 0.00001467, 0.00001039, -0.00001524, -0.00000725, 0.00001878, 0.00001310, -0.00000968, -0.00001051, -0.00000017, -0.00000180, -0.00000495, 0.00000028, 0.00000213, -0.00000190, -0.00000160, 0.00000132, 0.00000081, 0.00000166, 0.00000441, 0.00000170, -0.00000236, -0.00000009, 0.00000011, -0.00000534, -0.00000019, 0.00001415, 0.00000996, -0.00000991, -0.00000888, 0.00000993, 0.00000598, -0.00001492, -0.00001091, 0.00000868, 0.00000319, -0.00001524, -0.00001055, 0.00000413, 0.00000154, -0.00000554, -0.00000243, -0.00000201, -0.00000687, -0.00000560, -0.00000566, -0.00001316, -0.00001005, 0.00000499, 0.00000493, -0.00001048, -0.00001198, -0.00000004, -0.00000043, -0.00001161, -0.00001704, -0.00002575, -0.00004576, -0.00004489, },
		{ 0.00008903, 0.00005958, 0.00004353, 0.00007872, 0.00009759, 0.00003105, -0.00006073, -0.00007374, -0.00002542, -0.00000414, -0.00001502, -0.00001115, -0.00000310, -0.00001440, -0.00001553, 0.00000627, 0.00001148, -0.00000680, -0.00001014, 0.00000219, 0.00000081, -0.00000602, 0.00000145, 0.00000760, -0.00000017, -0.00000498, -0.00000078, -0.00000140, -0.00000691, -0.00000581, 0.00000213, 0.00000855, 0.00000608, -0.00000408, -0.00000845, -0.00000312, -0.00000294, -0.00000578, 0.00000624, 0.00001613, -0.00000237, -0.00001876, -0.00000080, 0.00001280, -0.00000440, -0.00000507, 0.00002017, 0.00000995, -0.00002777, -0.00001874, 0.00001801, 0.00000872, -0.00001602, 0.00000381, 0.00002037, -0.00000988, -0.00002312, 0.00000885, 0.00001532, -0.00001448, -0.00001103, 0.00001819, 0.00001067, -0.00001128, 0.00000015, 0.00000921, -0.00001297, -0.00001668, 0.00001209, 0.00001475, -0.00001315, -0.00001216, 0.00001588, 0.00001657, -0.00000643, -0.00001068, -0.00000077, -0.00000227, -0.00000599, 0.00000034, 0.00000291, -0.00000280, -0.00000257, 0.00000265, 0.00000151, -0.00000047, 0.00000361, 0.00000404, -0.00000164, -0.00000165, 0.00000129, -0.00000335, -0.00000364, 0.00001016, 0.00001385, -0.00000503, -0.00001303, 0.00000519, 0.00001134, -0.00000928, -0.00001576, 0.00000358, 0.00000774, -0.00001069, -0.00001390, 0.00000097, 0.00000341, -0.00000455, -0.00000335, -0.00000099, -0.00000612, -0.00000710, -0.00000535, -0.00001131, -0.00001131, 0.00000319, 0.00000689, -0.00000923, -0.00001458, -0.00000042, 0.00000350, -0.00001013, -0.00001928, -0.00002441, -0.00003899, -0.00004099, 0.00000185, 0.00006616, },
		{ 0.00003839, 0.00006345, 0.00009240, 0.00005143, -0.00003662, -0.00007381, -0.00003898, -0.00000636, -0.00000995, -0.00001295, -0.00000551, -0.00001041, -0.00001445, 0.00000141, 0.00001070, -0.00000319, -0.00001056, -0.00000040, 0.00000184, -0.00000469, 0.00000012, 0.00000729, 0.00000114, -0.00000493, -0.00000110, -0.00000081, -0.00000694, -0.00000627, 0.00000202, 0.00000754, 0.00000563, -0.00000161, -0.00000669, -0.00000478, -0.00000371, -0.00000512, 0.00000374, 0.00001430, 0.00000208, -0.00001575, -0.00000531, 0.00001047, -0.00000054, -0.00000562, 0.00001613, 0.00001434, -0.00002123, -0.00002359, 0.00001099, 0.00001224, -0.00001231, -0.00000065, 0.00001966, -0.00000377, -0.00002366, 0.00000264, 0.00001660, -0.00000972, -0.00001364, 0.00001450, 0.00001333, -0.00000948, -0.00000178, 0.00001093, -0.00000983, -0.00002001, 0.00000817, 0.00001831, -0.00000963, -0.00001620, 0.00001180, 0.00001903, -0.00000284, -0.00001032, -0.00000108, -0.00000256, -0.00000737, -0.00000001, 0.00000413, -0.00000327, -0.00000422, 0.00000354, 0.00000308, -0.00000216, 0.00000179, 0.00000600, 0.00000007, -0.00000325, 0.00000136, -0.00000064, -0.00000543, 0.00000500, 0.00001554, 0.00000122, -0.00001478, -0.00000132, 0.00001405, -0.00000166, -0.00001788, -0.00000339, 0.00000987, -0.00000452, -0.00001539, -0.00000336, 0.00000434, -0.00000276, -0.00000437, -0.00000061, -0.00000460, -0.00000811, -0.00000600, -0.00000949, -0.00001160, 0.00000094, 0.00000803, -0.00000716, -0.00001667, -0.00000189, 0.00000716, -0.00000761, -0.00002152, -0.00002417, -0.00003263, -0.00003653, -0.00000313, 0.00005555, 0.00008316, 0.00006014, },
		{ 0.00008321, 0.00006431, -0.00001222, -0.00006666, -0.00005038, -0.00001255, -0.00000575, -0.00001231, -0.00000852, -0.00000818, -0.00001197, -0.00000217, 0.00000839, -0.00000021, -0.00000966, -0.00000294, 0.00000188, -0.00000304, -0.00000071, 0.00000644, 0.00000235, -0.00000443, -0.00000162, -0.00000038, -0.00000659, -0.00000683, 0.00000163, 0.00000689, 0.00000505, 0.00000007, -0.00000459, -0.00000558, -0.00000481, -0.00000465, 0.00000197, 0.00001185, 0.00000515, -0.00001179, -0.00000827, 0.00000709, 0.00000230, -0.00000483, 0.00001208, 0.00001667, -0.00001399, -0.00002573, 0.00000329, 0.00001343, -0.00000777, -0.00000373, 0.00001747, 0.00000165, -0.00002224, -0.00000327, 0.00001599, -0.00000487, -0.00001475, 0.00001038, 0.00001500, -0.00000719, -0.00000361, 0.00001193, -0.00000596, -0.00002220, 0.00000322, 0.00002062, -0.00000502, -0.00001896, 0.00000689, 0.00002030, 0.00000085, -0.00000948, -0.00000108, -0.00000248, -0.00000894, -0.00000096, 0.00000560, -0.00000306, -0.00000633, 0.00000365, 0.00000529, -0.00000303, -0.00000083, 0.00000714, 0.00000264, -0.00000437, 0.00000028, 0.00000210, -0.00000535, -0.00000039, 0.00001469, 0.00000776, -0.00001363, -0.00000836, 0.00001348, 0.00000661, -0.00001659, -0.00001095, 0.00000890, 0.00000225, -0.00001440, -0.00000820, 0.00000383, -0.00000036, -0.00000504, -0.00000109, -0.00000270, -0.00000821, -0.00000736, -0.00000822, -0.00001091, -0.00000124, 0.00000810, -0.00000469, -0.00001781, -0.00000420, 0.00001003, -0.00000406, -0.00002329, -0.00002513, -0.00002694, -0.00003138, -0.00000709, 0.00004567, 0.00007693, 0.00006020, 0.00003557, 0.00005077, },
		{ 0.00000964, -0.00005371, -0.00005751, -0.00002152, -0.00000386, -0.00000972, -0.00001098, -0.00000783, -0.00000905, -0.00000404, 0.00000526, 0.00000155, -0.00000779, -0.00000489, 0.00000100, -0.00000149, -0.00000089, 0.00000532, 0.00000323, -0.00000353, -0.00000210, -0.00000024, -0.00000597, -0.00000728, 0.00000095, 0.00000642, 0.00000452, 0.00000108, -0.00000248, -0.00000555, -0.00000595, -0.00000449, 0.00000084, 0.00000921, 0.00000678, -0.00000751, -0.00000953, 0.00000325, 0.00000384, -0.00000310, 0.00000859, 0.00001709, -0.00000693, -0.00002521, -0.00000412, 0.00001224, -0.00000317, -0.00000518, 0.00001436, 0.00000589, -0.00001924, -0.00000824, 0.00001371, -0.00000053, -0.00001440, 0.00000628, 0.00001560, -0.00000464, -0.00000516, 0.00001216, -0.00000168, -0.00002300, -0.00000231, 0.00002137, 0.00000023, -0.00002018, 0.00000155, 0.00002027, 0.00000438, -0.00000826, -0.00000080, -0.00000193, -0.00001046, -0.00000264, 0.00000702, -0.00000200, -0.00000855, 0.00000277, 0.00000776, -0.00000280, -0.00000386, 0.00000715, 0.00000571, -0.00000459, -0.00000172, 0.00000421, -0.00000355, -0.00000508, 0.00001143, 0.00001340, -0.00000966, -0.00001459, 0.00000962, 0.00001401, -0.00001182, -0.00001763, 0.00000472, 0.00000831, -0.00001078, -0.00001263, 0.00000162, 0.00000218, -0.00000492, -0.00000237, -0.00000095, -0.00000715, -0.00000898, -0.00000792, -0.00000949, -0.00000280, 0.00000711, -0.00000239, -0.00001778, -0.00000688, 0.00001173, 0.00000021, -0.00002411, -0.00002716, -0.00002231, -0.00002555, -0.00000996, 0.00003634, 0.00007059, 0.00005976, 0.00003404, 0.00004089, 0.00007237, 0.00007025, },
		{ -0.00005913, -0.00003138, -0.00000513, -0.00000622, -0.00001198, -0.00000893, -0.00000663, -0.00000422, 0.00000213, 0.00000189, -0.00000554, -0.00000588, -0.00000048, -0.00000044, -0.00000051, 0.00000423, 0.00000364, -0.00000240, -0.00000236, -0.00000036, -0.00000527, -0.00000749, 0.00000008, 0.00000594, 0.00000411, 0.00000159, -0.00000058, -0.00000483, -0.00000690, -0.00000467, 0.00000019, 0.00000672, 0.00000710, -0.00000350, -0.00000919, -0.00000042, 0.00000402, -0.00000094, 0.00000605, 0.00001600, -0.00000079, -0.00002241, -0.00001035, 0.00000898, 0.00000069, -0.00000505, 0.00001095, 0.00000866, -0.00001520, -0.00001181, 0.00001019, 0.00000280, -0.00001285, 0.00000263, 0.00001520, -0.00000206, -0.00000632, 0.00001161, 0.00000266, -0.00002229, -0.00000791, 0.00002042, 0.00000558, -0.00001975, -0.00000374, 0.00001898, 0.00000753, -0.00000674, -0.00000033, -0.00000084, -0.00001168, -0.00000502, 0.00000805, -0.00000011, -0.00001049, 0.00000085, 0.00001005, -0.00000138, -0.00000682, 0.00000588, 0.00000883, -0.00000365, -0.00000420, 0.00000521, -0.00000050, -0.00000829, 0.00000632, 0.00001714, -0.00000347, -0.00001876, 0.00000306, 0.00001915, -0.00000418, -0.00002199, -0.00000213, 0.00001238, -0.00000486, -0.00001562, -0.00000224, 0.00000422, -0.00000376, -0.00000413, 0.00000013, -0.00000500, -0.00001023, -0.00000877, -0.00000787, -0.00000332, 0.00000540, -0.00000080, -0.00001667, -0.00000934, 0.00001208, 0.00000466, -0.00002363, -0.00002985, -0.00001911, -0.00001923, -0.00001153, 0.00002743, 0.00006416, 0.00005906, 0.00003314, 0.00003337, 0.00006163, 0.00007070, 0.00002699, -0.00003721, },
		{ -0.00000959, -0.00000316, -0.00001117, -0.00001066, -0.00000537, -0.00000318, -0.00000026, 0.00000098, -0.00000358, -0.00000583, -0.00000210, -0.00000010, 0.00000021, 0.00000341, 0.00000362, -0.00000125, -0.00000229, -0.00000066, -0.00000464, -0.00000742, -0.00000085, 0.00000534, 0.00000380, 0.00000178, 0.00000096, -0.00000362, -0.00000748, -0.00000516, -0.00000018, 0.00000462, 0.00000638, -0.00000021, -0.00000756, -0.00000341, 0.00000300, 0.00000116, 0.00000462, 0.00001394, 0.00000391, -0.00001794, -0.00001476, 0.00000424, 0.00000322, -0.00000365, 0.00000784, 0.00000993, -0.00001072, -0.00001372, 0.00000597, 0.00000480, -0.00001049, -0.00000024, 0.00001398, 0.00000032, -0.00000699, 0.00001038, 0.00000666, -0.00002011, -0.00001303, 0.00001781, 0.00001046, -0.00001774, -0.00000852, 0.00001652, 0.00001008, -0.00000505, 0.00000023, 0.00000074, -0.00001231, -0.00000797, 0.00000836, 0.00000249, -0.00001176, -0.00000199, 0.00001171, 0.00000112, -0.00000921, 0.00000341, 0.00001147, -0.00000152, -0.00000662, 0.00000484, 0.00000311, -0.00000955, 0.00000023, 0.00001831, 0.00000391, -0.00002003, -0.00000506, 0.00002098, 0.00000513, -0.00002295, -0.00001054, 0.00001345, 0.00000247, -0.00001633, -0.00000724, 0.00000508, -0.00000158, -0.00000587, 0.00000014, -0.00000211, -0.00001053, -0.00001065, -0.00000667, -0.00000265, 0.00000353, -0.00000035, -0.00001486, -0.00001099, 0.00001123, 0.00000866, -0.00002174, -0.00003261, -0.00001760, -0.00001282, -0.00001157, 0.00001898, 0.00005748, 0.00005832, 0.00003260, 0.00002752, 0.00005201, 0.00006753, 0.00003899, -0.00001975, -0.00005509, -0.00004005, },
		{ -0.00000878, -0.00001206, -0.00000548, -0.00000164, -0.00000144, -0.00000071, -0.00000239, -0.00000497, -0.00000338, -0.00000046, 0.00000094, 0.00000302, 0.00000333, -0.00000028, -0.00000190, -0.00000098, -0.00000419, -0.00000713, -0.00000170, 0.00000458, 0.00000353, 0.00000179, 0.00000213, -0.00000212, -0.00000759, -0.00000583, -0.00000045, 0.00000303, 0.00000499, 0.00000206, -0.00000511, -0.00000534, 0.00000111, 0.00000274, 0.00000424, 0.00001149, 0.00000695, -0.00001259, -0.00001698, -0.00000121, 0.00000402, -0.00000148, 0.00000549, 0.00000992, -0.00000640, -0.00001400, 0.00000168, 0.00000536, -0.00000779, -0.00000216, 0.00001222, 0.00000232, -0.00000718, 0.00000862, 0.00001003, -0.00001667, -0.00001717, 0.00001379, 0.00001436, -0.00001438, -0.00001237, 0.00001315, 0.00001188, -0.00000329, 0.00000077, 0.00000269, -0.00001213, -0.00001121, 0.00000770, 0.00000550, -0.00001205, -0.00000545, 0.00001235, 0.00000442, -0.00001060, -0.00000000, 0.00001318, 0.00000162, -0.00000843, 0.00000315, 0.00000655, -0.00000876, -0.00000581, 0.00001673, 0.00001125, -0.00001806, -0.00001332, 0.00001907, 0.00001456, -0.00002004, -0.00001900, 0.00001100, 0.00000995, -0.00001422, -0.00001254, 0.00000423, 0.00000129, -0.00000700, -0.00000102, 0.00000093, -0.00000947, -0.00001310, -0.00000648, -0.00000098, 0.00000216, -0.00000115, -0.00001299, -0.00001145, 0.00000966, 0.00001164, -0.00001869, -0.00003479, -0.00001773, -0.00000690, -0.00000999, 0.00001126, 0.00005030, 0.00005759, 0.00003247, 0.00002268, 0.00004376, 0.00006258, 0.00004584, -0.00000370, -0.00004631, -0.00004567, -0.00001644, -0.00000178, },
		{ -0.00000673, -0.00000039, -0.00000135, -0.00000252, -0.00000226, -0.00000373, -0.00000400, -0.00000132, 0.00000136, 0.00000302, 0.00000298, 0.00000040, -0.00000128, -0.00000117, -0.00000393, -0.00000673, -0.00000236, 0.00000371, 0.00000321, 0.00000168, 0.00000293, -0.00000051, -0.00000721, -0.00000656, -0.00000077, 0.00000196, 0.00000330, 0.00000322, -0.00000234, -0.00000608, -0.00000118, 0.00000353, 0.00000468, 0.00000918, 0.00000834, -0.00000712, -0.00001698, -0.00000655, 0.00000303, 0.00000083, 0.00000416, 0.00000900, -0.00000272, -0.00001284, -0.00000212, 0.00000458, -0.00000524, -0.00000310, 0.00001021, 0.00000381, -0.00000691, 0.00000655, 0.00001250, -0.00001229, -0.00001995, 0.00000877, 0.00001687, -0.00001008, -0.00001499, 0.00000916, 0.00001280, -0.00000158, 0.00000121, 0.00000485, -0.00001101, -0.00001441, 0.00000592, 0.00000852, -0.00001116, -0.00000913, 0.00001174, 0.00000808, -0.00001071, -0.00000392, 0.00001361, 0.00000537, -0.00000918, 0.00000045, 0.00000913, -0.00000619, -0.00001086, 0.00001269, 0.00001734, -0.00001314, -0.00002026, 0.00001365, 0.00002245, -0.00001347, -0.00002589, 0.00000521, 0.00001617, -0.00000932, -0.00001707, 0.00000146, 0.00000424, -0.00000706, -0.00000317, 0.00000344, -0.00000692, -0.00001544, -0.00000766, 0.00000121, 0.00000188, -0.00000293, -0.00001173, -0.00001068, 0.00000803, 0.00001327, -0.00001501, -0.00003582, -0.00001914, -0.00000208, -0.00000695, 0.00000468, 0.00004249, 0.00005669, 0.00003299, 0.00001846, 0.00003661, 0.00005724, 0.00004855, 0.00000922, -0.00003450, -0.00004703, -0.00002423, -0.00000293, -0.00000559, -0.00001236, },
		{ -0.00000028, -0.00000381, -0.00000311, -0.00000264, -0.00000384, -0.00000232, 0.00000128, 0.00000326, 0.00000276, 0.00000078, -0.00000058, -0.00000115, -0.00000381, -0.00000634, -0.00000279, 0.00000282, 0.00000279, 0.00000147, 0.00000342, 0.00000106, -0.00000639, -0.00000721, -0.00000122, 0.00000134, 0.00000163, 0.00000334, 0.00000027, -0.00000566, -0.00000341, 0.00000343, 0.00000557, 0.00000740, 0.00000838, -0.00000223, -0.00001504, -0.00001101, 0.00000046, 0.00000266, 0.00000388, 0.00000763, 0.00000002, -0.00001065, -0.00000500, 0.00000276, -0.00000325, -0.00000319, 0.00000824, 0.00000476, -0.00000629, 0.00000437, 0.00001395, -0.00000740, -0.00002111, 0.00000328, 0.00001777, -0.00000533, -0.00001619, 0.00000494, 0.00001283, -0.00000001, 0.00000147, 0.00000698, -0.00000894, -0.00001716, 0.00000306, 0.00001112, -0.00000908, -0.00001255, 0.00000982, 0.00001160, -0.00000944, -0.00000782, 0.00001263, 0.00000922, -0.00000862, -0.00000276, 0.00001041, -0.00000242, -0.00001419, 0.00000688, 0.00002122, -0.00000606, -0.00002466, 0.00000562, 0.00002742, -0.00000419, -0.00002979, -0.00000310, 0.00001981, -0.00000222, -0.00001977, -0.00000301, 0.00000649, -0.00000580, -0.00000579, 0.00000484, -0.00000315, -0.00001694, -0.00001026, 0.00000319, 0.00000302, -0.00000515, -0.00001166, -0.00000900, 0.00000705, 0.00001355, -0.00001142, -0.00003544, -0.00002122, 0.00000118, -0.00000288, -0.00000030, 0.00003412, 0.00005523, 0.00003431, 0.00001484, 0.00003006, 0.00005228, 0.00004859, 0.00001818, -0.00002175, -0.00004384, -0.00003121, -0.00000677, -0.00000270, -0.00001114, -0.00000847, -0.00000002, },
		{ -0.00000459, -0.00000215, -0.00000303, -0.00000307, 0.00000067, 0.00000349, 0.00000279, 0.00000096, 0.00000005, -0.00000090, -0.00000375, -0.00000603, -0.00000303, 0.00000200, 0.00000225, 0.00000116, 0.00000364, 0.00000250, -0.00000518, -0.00000765, -0.00000184, 0.00000103, 0.00000024, 0.00000263, 0.00000232, -0.00000432, -0.00000518, 0.00000251, 0.00000653, 0.00000635, 0.00000746, 0.00000160, -0.00001166, -0.00001404, -0.00000323, 0.00000353, 0.00000444, 0.00000627, 0.00000173, -0.00000791, -0.00000670, 0.00000032, -0.00000210, -0.00000269, 0.00000657, 0.00000522, -0.00000543, 0.00000228, 0.00001436, -0.00000246, -0.00002061, -0.00000210, 0.00001700, -0.00000065, -0.00001596, 0.00000086, 0.00001199, 0.00000132, 0.00000154, 0.00000887, -0.00000601, -0.00001909, -0.00000070, 0.00001288, -0.00000600, -0.00001525, 0.00000672, 0.00001448, -0.00000691, -0.00001118, 0.00001031, 0.00001264, -0.00000676, -0.00000589, 0.00001019, 0.00000181, -0.00001542, 0.00000029, 0.00002232, 0.00000198, -0.00002580, -0.00000368, 0.00002858, 0.00000634, -0.00002977, -0.00001254, 0.00001999, 0.00000600, -0.00001980, -0.00000852, 0.00000732, -0.00000335, -0.00000824, 0.00000479, 0.00000129, -0.00001697, -0.00001391, 0.00000423, 0.00000557, -0.00000706, -0.00001307, -0.00000702, 0.00000726, 0.00001289, -0.00000864, -0.00003373, -0.00002318, 0.00000271, 0.00000155, -0.00000337, 0.00002554, 0.00005278, 0.00003641, 0.00001209, 0.00002363, 0.00004784, 0.00004742, 0.00002328, -0.00001008, -0.00003673, -0.00003575, -0.00001274, -0.00000123, -0.00000851, -0.00000992, -0.00000079, 0.00000120, -0.00000414, },
		{ -0.00000191, -0.00000327, -0.00000031, 0.00000347, 0.00000308, 0.00000109, 0.00000051, -0.00000047, -0.00000361, -0.00000584, -0.00000315, 0.00000134, 0.00000163, 0.00000071, 0.00000361, 0.00000375, -0.00000370, -0.00000779, -0.00000261, 0.00000086, -0.00000073, 0.00000140, 0.00000357, -0.00000243, -0.00000620, 0.00000102, 0.00000722, 0.00000606, 0.00000609, 0.00000411, -0.00000749, -0.00001530, -0.00000740, 0.00000313, 0.00000547, 0.00000529, 0.00000249, -0.00000508, -0.00000716, -0.00000227, -0.00000190, -0.00000191, 0.00000534, 0.00000529, -0.00000446, 0.00000045, 0.00001381, 0.00000210, -0.00001858, -0.00000683, 0.00001474, 0.00000343, -0.00001443, -0.00000271, 0.00001042, 0.00000232, 0.00000141, 0.00001031, -0.00000245, -0.00001990, -0.00000503, 0.00001346, -0.00000222, -0.00001684, 0.00000275, 0.00001631, -0.00000340, -0.00001357, 0.00000694, 0.00001515, -0.00000382, -0.00000839, 0.00000860, 0.00000574, -0.00001457, -0.00000605, 0.00002063, 0.00000970, -0.00002355, -0.00001268, 0.00002574, 0.00001638, -0.00002564, -0.00002143, 0.00001643, 0.00001389, -0.00001681, -0.00001405, 0.00000622, -0.00000020, -0.00000984, 0.00000328, 0.00000563, -0.00001518, -0.00001795, 0.00000374, 0.00000910, -0.00000788, -0.00001587, -0.00000553, 0.00000890, 0.00001195, -0.00000717, -0.00003115, -0.00002432, 0.00000272, 0.00000558, -0.00000450, 0.00001733, 0.00004901, 0.00003895, 0.00001059, 0.00001715, 0.00004357, 0.00004615, 0.00002538, -0.00000100, -0.00002710, -0.00003670, -0.00001966, -0.00000199, -0.00000511, -0.00001035, -0.00000255, 0.00000242, -0.00000340, -0.00000617, -0.00000252, },
		{ -0.00000139, 0.00000301, 0.00000351, 0.00000132, 0.00000079, 0.00000004, -0.00000334, -0.00000576, -0.00000324, 0.00000088, 0.00000103, 0.00000011, 0.00000335, 0.00000477, -0.00000203, -0.00000758, -0.00000347, 0.00000069, -0.00000124, -0.00000003, 0.00000394, -0.00000038, -0.00000636, -0.00000072, 0.00000741, 0.00000636, 0.00000469, 0.00000530, -0.00000319, -0.00001477, -0.00001137, 0.00000140, 0.00000649, 0.00000491, 0.00000255, -0.00000257, -0.00000652, -0.00000456, -0.00000260, -0.00000121, 0.00000458, 0.00000513, -0.00000348, -0.00000102, 0.00001250, 0.00000591, -0.00001531, -0.00001044, 0.00001131, 0.00000650, -0.00001190, -0.00000547, 0.00000829, 0.00000294, 0.00000112, 0.00001117, 0.00000144, -0.00001940, -0.00000949, 0.00001269, 0.00000180, -0.00001711, -0.00000164, 0.00001681, 0.00000064, -0.00001468, 0.00000295, 0.00001644, -0.00000018, -0.00000984, 0.00000600, 0.00000875, -0.00001202, -0.00001123, 0.00001657, 0.00001590, -0.00001840, -0.00001991, 0.00001942, 0.00002425, -0.00001794, -0.00002811, 0.00000957, 0.00001995, -0.00001100, -0.00001846, 0.00000309, 0.00000290, -0.00001011, 0.00000069, 0.00000910, -0.00001163, -0.00002154, 0.00000144, 0.00001290, -0.00000705, -0.00001956, -0.00000531, 0.00001179, 0.00001155, -0.00000719, -0.00002843, -0.00002415, 0.00000176, 0.00000853, -0.00000398, 0.00001014, 0.00004388, 0.00004136, 0.00001065, 0.00001081, 0.00003893, 0.00004532, 0.00002576, 0.00000474, -0.00001676, -0.00003368, -0.00002600, -0.00000525, -0.00000188, -0.00000933, -0.00000478, 0.00000282, -0.00000177, -0.00000729, -0.00000376, -0.00000092, -0.00000280, },
		{ 0.00000389, 0.00000175, 0.00000093, 0.00000051, -0.00000289, -0.00000571, -0.00000338, 0.00000060, 0.00000053, -0.00000061, 0.00000282, 0.00000553, -0.00000025, -0.00000699, -0.00000432, 0.00000040, -0.00000136, -0.00000138, 0.00000352, 0.00000145, -0.00000572, -0.00000237, 0.00000700, 0.00000703, 0.00000361, 0.00000537, 0.00000065, -0.00001267, -0.00001455, -0.00000145, 0.00000706, 0.00000514, 0.00000225, -0.00000065, -0.00000506, -0.00000618, -0.00000400, -0.00000086, 0.00000423, 0.00000490, -0.00000258, -0.00000208, 0.00001065, 0.00000873, -0.00001122, -0.00001263, 0.00000719, 0.00000827, -0.00000875, -0.00000722, 0.00000586, 0.00000313, 0.00000070, 0.00001139, 0.00000531, -0.00001757, -0.00001359, 0.00001056, 0.00000556, -0.00001600, -0.00000597, 0.00001591, 0.00000472, -0.00001441, -0.00000117, 0.00001637, 0.00000363, -0.00001004, 0.00000290, 0.00001045, -0.00000841, -0.00001464, 0.00001098, 0.00001972, -0.00001133, -0.00002425, 0.00001077, 0.00002868, -0.00000790, -0.00003127, 0.00000055, 0.00002297, -0.00000321, -0.00002064, -0.00000169, 0.00000509, -0.00000890, -0.00000234, 0.00001110, -0.00000676, -0.00002383, -0.00000250, 0.00001608, -0.00000437, -0.00002334, -0.00000686, 0.00001536, 0.00001243, -0.00000844, -0.00002640, -0.00002258, 0.00000060, 0.00001000, -0.00000245, 0.00000455, 0.00003764, 0.00004298, 0.00001232, 0.00000511, 0.00003345, 0.00004489, 0.00002572, 0.00000726, -0.00000753, -0.00002717, -0.00003023, -0.00001062, 0.00000011, -0.00000689, -0.00000677, 0.00000212, 0.00000024, -0.00000753, -0.00000559, -0.00000048, -0.00000173, -0.00000223, 0.00000208, },
		{ 0.00000105, 0.00000086, -0.00000228, -0.00000561, -0.00000362, 0.00000043, 0.00000023, -0.00000139, 0.00000201, 0.00000598, 0.00000154, -0.00000602, -0.00000507, -0.00000007, -0.00000120, -0.00000241, 0.00000250, 0.00000277, -0.00000448, -0.00000365, 0.00000607, 0.00000777, 0.00000304, 0.00000463, 0.00000362, -0.00000944, -0.00001645, -0.00000503, 0.00000680, 0.00000584, 0.00000192, 0.00000057, -0.00000313, -0.00000692, -0.00000576, -0.00000108, 0.00000413, 0.00000472, -0.00000178, -0.00000274, 0.00000852, 0.00001044, -0.00000678, -0.00001326, 0.00000291, 0.00000865, -0.00000546, -0.00000791, 0.00000339, 0.00000291, 0.00000020, 0.00001096, 0.00000883, -0.00001454, -0.00001688, 0.00000724, 0.00000859, -0.00001365, -0.00000972, 0.00001371, 0.00000833, -0.00001285, -0.00000491, 0.00001502, 0.00000714, -0.00000902, -0.00000013, 0.00001076, -0.00000447, -0.00001604, 0.00000484, 0.00002080, -0.00000362, -0.00002518, 0.00000133, 0.00002902, 0.00000277, -0.00003023, -0.00000904, 0.00002222, 0.00000526, -0.00001989, -0.00000728, 0.00000563, -0.00000647, -0.00000500, 0.00001136, -0.00000132, -0.00002422, -0.00000753, 0.00001784, -0.00000006, -0.00002631, -0.00001031, 0.00001874, 0.00001505, -0.00001027, -0.00002577, -0.00001993, 0.00000011, 0.00000996, -0.00000073, 0.00000086, 0.00003089, 0.00004322, 0.00001529, 0.00000070, 0.00002699, 0.00004431, 0.00002625, 0.00000739, -0.00000081, -0.00001837, -0.00003123, -0.00001709, 0.00000004, -0.00000353, -0.00000781, 0.00000039, 0.00000204, -0.00000672, -0.00000755, -0.00000085, -0.00000034, -0.00000255, 0.00000079, 0.00000400, 0.00000240, },
		{ -0.00000160, -0.00000537, -0.00000397, 0.00000028, 0.00000015, -0.00000213, 0.00000094, 0.00000608, 0.00000326, -0.00000470, -0.00000560, -0.00000074, -0.00000092, -0.00000303, 0.00000115, 0.00000340, -0.00000293, -0.00000436, 0.00000476, 0.00000835, 0.00000304, 0.00000347, 0.00000549, -0.00000561, -0.00001685, -0.00000883, 0.00000552, 0.00000670, 0.00000182, 0.00000115, -0.00000111, -0.00000672, -0.00000753, -0.00000193, 0.00000406, 0.00000467, -0.00000107, -0.00000304, 0.00000636, 0.00001107, -0.00000246, -0.00001241, -0.00000099, 0.00000773, -0.00000245, -0.00000759, 0.00000115, 0.00000231, -0.00000037, 0.00000998, 0.00001170, -0.00001056, -0.00001898, 0.00000309, 0.00001050, -0.00001039, -0.00001251, 0.00001049, 0.00001107, -0.00001027, -0.00000788, 0.00001266, 0.00000991, -0.00000702, -0.00000261, 0.00000985, -0.00000089, -0.00001556, -0.00000084, 0.00001927, 0.00000343, -0.00002282, -0.00000729, 0.00002538, 0.00001226, -0.00002511, -0.00001742, 0.00001772, 0.00001284, -0.00001601, -0.00001254, 0.00000409, -0.00000344, -0.00000660, 0.00001004, 0.00000378, -0.00002248, -0.00001277, 0.00001762, 0.00000523, -0.00002765, -0.00001534, 0.00002102, 0.00001943, -0.00001181, -0.00002693, -0.00001689, 0.00000094, 0.00000883, 0.00000036, -0.00000101, 0.00002441, 0.00004174, 0.00001892, -0.00000185, 0.00001981, 0.00004288, 0.00002778, 0.00000639, 0.00000269, -0.00000896, -0.00002850, -0.00002326, -0.00000248, -0.00000012, -0.00000742, -0.00000192, 0.00000309, -0.00000499, -0.00000913, -0.00000209, 0.00000098, -0.00000221, -0.00000062, 0.00000367, 0.00000316, 0.00000127, 0.00000105, },
		{ -0.00000438, 0.00000007, 0.00000031, -0.00000270, -0.00000036, 0.00000577, 0.00000484, -0.00000308, -0.00000584, -0.00000156, -0.00000067, -0.00000321, -0.00000025, 0.00000335, -0.00000138, -0.00000447, 0.00000332, 0.00000859, 0.00000350, 0.00000227, 0.00000624, -0.00000175, -0.00001575, -0.00001228, 0.00000321, 0.00000739, 0.00000209, 0.00000124, 0.00000067, -0.00000569, -0.00000894, -0.00000335, 0.00000380, 0.00000476, -0.00000042, -0.00000304, 0.00000436, 0.00001075, 0.00000134, -0.00001031, -0.00000407, 0.00000576, -0.00000011, -0.00000646, -0.00000063, 0.00000143, -0.00000098, 0.00000856, 0.00001372, -0.00000601, -0.00001965, -0.00000144, 0.00001106, -0.00000663, -0.00001406, 0.00000664, 0.00001266, -0.00000705, -0.00000981, 0.00000965, 0.00001169, -0.00000440, -0.00000420, 0.00000811, 0.00000182, -0.00001368, -0.00000530, 0.00001574, 0.00000874, -0.00001788, -0.00001369, 0.00001860, 0.00001897, -0.00001681, -0.00002297, 0.00001022, 0.00001806, -0.00000948, -0.00001624, 0.00000053, -0.00000069, -0.00000671, 0.00000769, 0.00000772, -0.00001884, -0.00001725, 0.00001530, 0.00001059, -0.00002686, -0.00002121, 0.00002140, 0.00002516, -0.00001211, -0.00002981, -0.00001440, 0.00000342, 0.00000736, 0.00000017, -0.00000153, 0.00001898, 0.00003863, 0.00002236, -0.00000226, 0.00001258, 0.00003993, 0.00003012, 0.00000559, 0.00000305, -0.00000068, -0.00002237, -0.00002765, -0.00000721, 0.00000231, -0.00000553, -0.00000415, 0.00000303, -0.00000275, -0.00000989, -0.00000400, 0.00000183, -0.00000122, -0.00000185, 0.00000286, 0.00000389, 0.00000169, 0.00000112, -0.00000092, -0.00000495, },
		{ 0.00000063, -0.00000300, -0.00000179, 0.00000501, 0.00000616, -0.00000120, -0.00000572, -0.00000247, -0.00000057, -0.00000302, -0.00000148, 0.00000270, -0.00000011, -0.00000406, 0.00000197, 0.00000843, 0.00000427, 0.00000131, 0.00000604, 0.00000167, -0.00001336, -0.00001489, 0.00000008, 0.00000756, 0.00000274, 0.00000110, 0.00000201, -0.00000405, -0.00000973, -0.00000515, 0.00000315, 0.00000492, 0.00000021, -0.00000282, 0.00000266, 0.00000970, 0.00000432, -0.00000734, -0.00000600, 0.00000314, 0.00000131, -0.00000482, -0.00000180, 0.00000039, -0.00000163, 0.00000684, 0.00001477, -0.00000131, -0.00001882, -0.00000584, 0.00001022, -0.00000287, -0.00001429, 0.00000265, 0.00001299, -0.00000363, -0.00001059, 0.00000643, 0.00001239, -0.00000160, -0.00000479, 0.00000602, 0.00000340, -0.00001103, -0.00000808, 0.00001107, 0.00001165, -0.00001150, -0.00001697, 0.00001006, 0.00002185, -0.00000684, -0.00002461, 0.00000112, 0.00001983, -0.00000136, -0.00001735, -0.00000450, 0.00000086, -0.00000528, 0.00000509, 0.00000996, -0.00001394, -0.00002012, 0.00001123, 0.00001503, -0.00002389, -0.00002693, 0.00001947, 0.00003140, -0.00001046, -0.00003391, -0.00001335, 0.00000741, 0.00000647, -0.00000154, -0.00000151, 0.00001518, 0.00003435, 0.00002476, -0.00000065, 0.00000619, 0.00003518, 0.00003257, 0.00000604, 0.00000114, 0.00000502, -0.00001392, -0.00002904, -0.00001331, 0.00000287, -0.00000251, -0.00000559, 0.00000189, -0.00000058, -0.00000961, -0.00000617, 0.00000191, 0.00000018, -0.00000262, 0.00000164, 0.00000439, 0.00000232, 0.00000115, -0.00000033, -0.00000434, -0.00000476, -0.00000030, },
		{ -0.00000322, 0.00000378, 0.00000712, 0.00000083, -0.00000520, -0.00000336, -0.00000070, -0.00000259, -0.00000235, 0.00000166, 0.00000069, -0.00000329, 0.00000089, 0.00000791, 0.00000514, 0.00000076, 0.00000515, 0.00000427, -0.00001006, -0.00001631, -0.00000352, 0.00000696, 0.00000362, 0.00000095, 0.00000280, -0.00000210, -0.00000974, -0.00000706, 0.00000202, 0.00000503, 0.00000087, -0.00000242, 0.00000133, 0.00000817, 0.00000631, -0.00000392, -0.00000666, 0.00000033, 0.00000171, -0.00000301, -0.00000230, -0.00000064, -0.00000229, 0.00000497, 0.00001486, 0.00000313, -0.00001659, -0.00000961, 0.00000811, 0.00000040, -0.00001327, -0.00000103, 0.00001213, -0.00000043, -0.00001029, 0.00000339, 0.00001209, 0.00000098, -0.00000445, 0.00000405, 0.00000388, -0.00000830, -0.00000911, 0.00000626, 0.00001201, -0.00000500, -0.00001688, 0.00000143, 0.00002056, 0.00000295, -0.00002198, -0.00000782, 0.00001769, 0.00000686, -0.00001530, -0.00000997, 0.00000052, -0.00000275, 0.00000311, 0.00001037, -0.00000869, -0.00002083, 0.00000617, 0.00001772, -0.00001917, -0.00003148, 0.00001526, 0.00003711, -0.00000652, -0.00003832, -0.00001440, 0.00001231, 0.00000703, -0.00000457, -0.00000185, 0.00001319, 0.00002969, 0.00002550, 0.00000239, 0.00000150, 0.00002880, 0.00003417, 0.00000821, -0.00000173, 0.00000737, -0.00000472, -0.00002683, -0.00001949, 0.00000102, 0.00000084, -0.00000567, -0.00000001, 0.00000096, -0.00000835, -0.00000811, 0.00000112, 0.00000164, -0.00000275, 0.00000020, 0.00000450, 0.00000311, 0.00000123, 0.00000013, -0.00000358, -0.00000505, -0.00000085, 0.00000102, -0.00000294, },
		{ 0.00000762, 0.00000293, -0.00000425, -0.00000412, -0.00000107, -0.00000208, -0.00000281, 0.00000046, 0.00000096, -0.00000241, 0.00000018, 0.00000715, 0.00000594, 0.00000067, 0.00000389, 0.00000590, -0.00000632, -0.00001638, -0.00000713, 0.00000546, 0.00000450, 0.00000098, 0.00000309, -0.00000014, -0.00000895, -0.00000878, 0.00000039, 0.00000493, 0.00000156, -0.00000189, 0.00000040, 0.00000644, 0.00000730, -0.00000052, -0.00000607, -0.00000223, 0.00000115, -0.00000137, -0.00000216, -0.00000152, -0.00000296, 0.00000309, 0.00001408, 0.00000693, -0.00001323, -0.00001233, 0.00000504, 0.00000279, -0.00001128, -0.00000402, 0.00001029, 0.00000218, -0.00000913, 0.00000085, 0.00001101, 0.00000304, -0.00000343, 0.00000258, 0.00000351, -0.00000603, -0.00000870, 0.00000218, 0.00001018, 0.00000039, -0.00001383, -0.00000571, 0.00001559, 0.00001076, -0.00001558, -0.00001481, 0.00001197, 0.00001357, -0.00001015, -0.00001462, -0.00000202, 0.00000011, 0.00000243, 0.00000929, -0.00000406, -0.00001933, 0.00000113, 0.00001822, -0.00001353, -0.00003406, 0.00000933, 0.00004127, -0.00000045, -0.00004197, -0.00001779, 0.00001716, 0.00000961, -0.00000826, -0.00000340, 0.00001276, 0.00002555, 0.00002435, 0.00000597, -0.00000085, 0.00002150, 0.00003399, 0.00001189, -0.00000410, 0.00000638, 0.00000344, -0.00002112, -0.00002426, -0.00000320, 0.00000352, -0.00000420, -0.00000203, 0.00000149, -0.00000645, -0.00000938, -0.00000041, 0.00000279, -0.00000222, -0.00000119, 0.00000413, 0.00000393, 0.00000145, 0.00000044, -0.00000273, -0.00000515, -0.00000156, 0.00000135, -0.00000252, -0.00000450, 0.00000213, },
		{ -0.00000289, -0.00000464, -0.00000165, -0.00000163, -0.00000288, -0.00000067, 0.00000072, -0.00000164, -0.00000018, 0.00000629, 0.00000656, 0.00000100, 0.00000256, 0.00000654, -0.00000263, -0.00001514, -0.00001030, 0.00000313, 0.00000511, 0.00000130, 0.00000301, 0.00000157, -0.00000746, -0.00001002, -0.00000163, 0.00000448, 0.00000223, -0.00000125, -0.00000016, 0.00000471, 0.00000739, 0.00000250, -0.00000445, -0.00000414, -0.00000017, -0.00000019, -0.00000152, -0.00000212, -0.00000358, 0.00000130, 0.00001259, 0.00000984, -0.00000911, -0.00001373, 0.00000142, 0.00000405, -0.00000868, -0.00000605, 0.00000780, 0.00000396, -0.00000739, -0.00000099, 0.00000945, 0.00000441, -0.00000206, 0.00000179, 0.00000269, -0.00000457, -0.00000738, -0.00000059, 0.00000689, 0.00000379, -0.00000875, -0.00001015, 0.00000808, 0.00001515, -0.00000664, -0.00001839, 0.00000375, 0.00001736, -0.00000263, -0.00001723, -0.00000651, 0.00000231, 0.00000339, 0.00000740, -0.00000085, -0.00001602, -0.00000285, 0.00001657, -0.00000802, -0.00003427, 0.00000261, 0.00004313, 0.00000707, -0.00004387, -0.00002322, 0.00002090, 0.00001434, -0.00001165, -0.00000666, 0.00001321, 0.00002276, 0.00002156, 0.00000905, -0.00000071, 0.00001430, 0.00003144, 0.00001630, -0.00000475, 0.00000284, 0.00000904, -0.00001284, -0.00002632, -0.00000911, 0.00000456, -0.00000141, -0.00000345, 0.00000089, -0.00000441, -0.00000968, -0.00000235, 0.00000329, -0.00000119, -0.00000224, 0.00000333, 0.00000461, 0.00000184, 0.00000066, -0.00000187, -0.00000503, -0.00000239, 0.00000151, -0.00000176, -0.00000546, 0.00000014, 0.00000755, 0.00000496, },
		{ -0.00000236, -0.00000135, -0.00000265, -0.00000155, 0.00000013, -0.00000114, -0.00000027, 0.00000547, 0.00000695, 0.00000163, 0.00000141, 0.00000632, 0.00000061, -0.00001280, -0.00001262, 0.00000017, 0.00000518, 0.00000187, 0.00000275, 0.00000284, -0.00000548, -0.00001058, -0.00000385, 0.00000355, 0.00000280, -0.00000051, -0.00000038, 0.00000317, 0.00000677, 0.00000485, -0.00000210, -0.00000515, -0.00000193, 0.00000035, -0.00000060, -0.00000235, -0.00000411, -0.00000031, 0.00001059, 0.00001168, -0.00000466, -0.00001369, -0.00000225, 0.00000408, -0.00000590, -0.00000700, 0.00000504, 0.00000481, -0.00000541, -0.00000207, 0.00000772, 0.00000508, -0.00000066, 0.00000167, 0.00000183, -0.00000399, -0.00000576, -0.00000183, 0.00000308, 0.00000479, -0.00000292, -0.00001126, -0.00000035, 0.00001538, 0.00000307, -0.00001776, -0.00000536, 0.00001736, 0.00000597, -0.00001687, -0.00001218, 0.00000295, 0.00000585, 0.00000564, 0.00000050, -0.00001173, -0.00000497, 0.00001331, -0.00000368, -0.00003224, -0.00000379, 0.00004236, 0.00001504, -0.00004333, -0.00002995, 0.00002258, 0.00002082, -0.00001371, -0.00001168, 0.00001360, 0.00002178, 0.00001781, 0.00001070, 0.00000155, 0.00000835, 0.00002648, 0.00002027, -0.00000300, -0.00000187, 0.00001109, -0.00000347, -0.00002483, -0.00001556, 0.00000326, 0.00000202, -0.00000365, -0.00000058, -0.00000283, -0.00000899, -0.00000425, 0.00000298, 0.00000004, -0.00000275, 0.00000223, 0.00000500, 0.00000240, 0.00000085, -0.00000107, -0.00000467, -0.00000324, 0.00000141, -0.00000077, -0.00000597, -0.00000203, 0.00000684, 0.00000677, -0.00000117, -0.00000484, },
		{ -0.00000226, -0.00000209, -0.00000063, -0.00000101, -0.00000026, 0.00000477, 0.00000713, 0.00000244, 0.00000058, 0.00000547, 0.00000311, -0.00000971, -0.00001384, -0.00000309, 0.00000455, 0.00000256, 0.00000248, 0.00000361, -0.00000329, -0.00001035, -0.00000602, 0.00000211, 0.00000314, 0.00000028, -0.00000032, 0.00000192, 0.00000567, 0.00000638, 0.00000057, -0.00000515, -0.00000375, 0.00000017, 0.00000036, -0.00000219, -0.00000450, -0.00000171, 0.00000831, 0.00001244, -0.00000032, -0.00001229, -0.00000550, 0.00000297, -0.00000337, -0.00000692, 0.00000239, 0.00000477, -0.00000351, -0.00000245, 0.00000609, 0.00000517, 0.00000049, 0.00000207, 0.00000128, -0.00000413, -0.00000443, -0.00000167, -0.00000034, 0.00000352, 0.00000233, -0.00000912, -0.00000799, 0.00001153, 0.00001163, -0.00001297, -0.00001349, 0.00001341, 0.00001403, -0.00001317, -0.00001787, 0.00000139, 0.00000921, 0.00000488, 0.00000006, -0.00000749, -0.00000482, 0.00000937, -0.00000126, -0.00002859, -0.00000883, 0.00003921, 0.00002229, -0.00004016, -0.00003686, 0.00002160, 0.00002820, -0.00001359, -0.00001801, 0.00001293, 0.00002267, 0.00001403, 0.00001037, 0.00000511, 0.00000463, 0.00001966, 0.00002257, 0.00000108, -0.00000612, 0.00000943, 0.00000519, -0.00001967, -0.00002108, -0.00000056, 0.00000514, -0.00000227, -0.00000239, -0.00000215, -0.00000755, -0.00000565, 0.00000191, 0.00000112, -0.00000266, 0.00000107, 0.00000501, 0.00000302, 0.00000107, -0.00000036, -0.00000410, -0.00000401, 0.00000101, 0.00000031, -0.00000594, -0.00000420, 0.00000550, 0.00000821, 0.00000083, -0.00000462, -0.00000311, -0.00000130, },
		{ -0.00000136, -0.00000123, -0.00000031, 0.00000420, 0.00000717, 0.00000331, 0.00000015, 0.00000426, 0.00000473, -0.00000629, -0.00001384, -0.00000626, 0.00000315, 0.00000319, 0.00000236, 0.00000392, -0.00000116, -0.00000933, -0.00000784, 0.00000022, 0.00000313, 0.00000110, -0.00000003, 0.00000101, 0.00000433, 0.00000707, 0.00000321, -0.00000420, -0.00000528, -0.00000064, 0.00000112, -0.00000170, -0.00000467, -0.00000286, 0.00000594, 0.00001219, 0.00000350, -0.00000974, -0.00000792, 0.00000095, -0.00000145, -0.00000599, 0.00000018, 0.00000398, -0.00000194, -0.00000228, 0.00000474, 0.00000486, 0.00000125, 0.00000274, 0.00000125, -0.00000465, -0.00000377, -0.00000058, -0.00000265, 0.00000055, 0.00000593, -0.00000447, -0.00001335, 0.00000446, 0.00001733, -0.00000487, -0.00001892, 0.00000618, 0.00001994, -0.00000643, -0.00002222, -0.00000250, 0.00001254, 0.00000575, -0.00000167, -0.00000428, -0.00000254, 0.00000587, -0.00000106, -0.00002430, -0.00001178, 0.00003440, 0.00002777, -0.00003471, -0.00004279, 0.00001789, 0.00003533, -0.00001080, -0.00002475, 0.00001042, 0.00002495, 0.00001123, 0.00000797, 0.00000885, 0.00000367, 0.00001203, 0.00002224, 0.00000675, -0.00000845, 0.00000470, 0.00001144, -0.00001151, -0.00002424, -0.00000646, 0.00000690, 0.00000061, -0.00000383, -0.00000258, -0.00000581, -0.00000621, 0.00000032, 0.00000171, -0.00000205, 0.00000010, 0.00000461, 0.00000359, 0.00000138, 0.00000026, -0.00000337, -0.00000461, 0.00000032, 0.00000132, -0.00000536, -0.00000616, 0.00000356, 0.00000910, 0.00000300, -0.00000397, -0.00000378, -0.00000149, -0.00000184, -0.00000228, },
		{ -0.00000056, 0.00000372, 0.00000714, 0.00000417, 0.00000009, 0.00000293, 0.00000548, -0.00000295, -0.00001271, -0.00000894, 0.00000108, 0.00000351, 0.00000245, 0.00000388, 0.00000068, -0.00000768, -0.00000907, -0.00000200, 0.00000264, 0.00000185, 0.00000046, 0.00000045, 0.00000297, 0.00000701, 0.00000549, -0.00000249, -0.00000625, -0.00000190, 0.00000150, -0.00000098, -0.00000459, -0.00000373, 0.00000366, 0.00001108, 0.00000651, -0.00000639, -0.00000923, -0.00000160, -0.00000038, -0.00000450, -0.00000139, 0.00000269, -0.00000087, -0.00000176, 0.00000377, 0.00000437, 0.00000159, 0.00000338, 0.00000175, -0.00000515, -0.00000391, 0.00000085, -0.00000351, -0.00000324, 0.00000721, 0.00000154, -0.00001548, -0.00000437, 0.00001900, 0.00000496, -0.00002043, -0.00000300, 0.00002244, 0.00000237, -0.00002404, -0.00000827, 0.00001479, 0.00000844, -0.00000381, -0.00000282, 0.00000125, 0.00000382, -0.00000282, -0.00002049, -0.00001240, 0.00002900, 0.00003083, -0.00002782, -0.00004673, 0.00001194, 0.00004104, -0.00000543, -0.00003080, 0.00000569, 0.00002775, 0.00001020, 0.00000396, 0.00001160, 0.00000547, 0.00000492, 0.00001888, 0.00001273, -0.00000789, -0.00000175, 0.00001405, -0.00000171, -0.00002398, -0.00001346, 0.00000645, 0.00000448, -0.00000419, -0.00000401, -0.00000434, -0.00000580, -0.00000138, 0.00000160, -0.00000115, -0.00000047, 0.00000392, 0.00000397, 0.00000177, 0.00000080, -0.00000254, -0.00000498, -0.00000061, 0.00000210, -0.00000429, -0.00000768, 0.00000115, 0.00000933, 0.00000518, -0.00000286, -0.00000425, -0.00000189, -0.00000154, -0.00000218, -0.00000190, -0.00000170, },
		{ 0.00000707, 0.00000498, 0.00000034, 0.00000170, 0.00000545, -0.00000002, -0.00001065, -0.00001082, -0.00000147, 0.00000335, 0.00000271, 0.00000365, 0.00000208, -0.00000559, -0.00000956, -0.00000430, 0.00000162, 0.00000241, 0.00000110, 0.00000022, 0.00000173, 0.00000636, 0.00000718, -0.00000029, -0.00000646, -0.00000336, 0.00000139, -0.00000019, -0.00000425, -0.00000432, 0.00000161, 0.00000935, 0.00000853, -0.00000265, -0.00000931, -0.00000428, -0.00000027, -0.00000280, -0.00000219, 0.00000115, -0.00000037, -0.00000111, 0.00000318, 0.00000388, 0.00000160, 0.00000379, 0.00000263, -0.00000529, -0.00000473, 0.00000203, -0.00000293, -0.00000688, 0.00000606, 0.00000752, -0.00001408, -0.00001319, 0.00001629, 0.00001459, -0.00001756, -0.00001231, 0.00002088, 0.00001174, -0.00002253, -0.00001497, 0.00001506, 0.00001263, -0.00000535, -0.00000340, 0.00000557, 0.00000393, -0.00000580, -0.00001815, -0.00001102, 0.00002416, 0.00003131, -0.00002063, -0.00004810, 0.00000468, 0.00004439, 0.00000192, -0.00003504, -0.00000110, 0.00003001, 0.00001131, -0.00000079, 0.00001245, 0.00000939, -0.00000038, 0.00001274, 0.00001755, -0.00000419, -0.00000820, 0.00001255, 0.00000797, -0.00001988, -0.00002015, 0.00000336, 0.00000846, -0.00000299, -0.00000600, -0.00000364, -0.00000456, -0.00000274, 0.00000074, -0.00000026, -0.00000053, 0.00000309, 0.00000408, 0.00000221, 0.00000129, -0.00000166, -0.00000506, -0.00000166, 0.00000252, -0.00000288, -0.00000860, -0.00000152, 0.00000879, 0.00000719, -0.00000133, -0.00000441, -0.00000239, -0.00000142, -0.00000193, -0.00000215, -0.00000225, -0.00000106, 0.00000320, },
		{ 0.00000084, 0.00000070, 0.00000486, 0.00000225, -0.00000798, -0.00001170, -0.00000417, 0.00000257, 0.00000302, 0.00000340, 0.00000299, -0.00000335, -0.00000921, -0.00000643, 0.00000006, 0.00000267, 0.00000184, 0.00000028, 0.00000074, 0.00000533, 0.00000818, 0.00000207, -0.00000588, -0.00000472, 0.00000078, 0.00000049, -0.00000368, -0.00000462, -0.00000015, 0.00000724, 0.00000949, 0.00000107, -0.00000819, -0.00000666, -0.00000107, -0.00000122, -0.00000224, -0.00000035, -0.00000042, -0.00000050, 0.00000289, 0.00000352, 0.00000142, 0.00000383, 0.00000365, -0.00000486, -0.00000591, 0.00000251, -0.00000129, -0.00000955, 0.00000289, 0.00001219, -0.00000959, -0.00002028, 0.00000967, 0.00002209, -0.00001076, -0.00001988, 0.00001543, 0.00002000, -0.00001753, -0.00002129, 0.00001281, 0.00001754, -0.00000541, -0.00000574, 0.00000930, 0.00000641, -0.00000892, -0.00001791, -0.00000844, 0.00002088, 0.00002964, -0.00001434, -0.00004685, -0.00000269, 0.00004489, 0.00001020, -0.00003665, -0.00000929, 0.00003071, 0.00001441, -0.00000522, 0.00001093, 0.00001438, -0.00000289, 0.00000473, 0.00001987, 0.00000212, -0.00001291, 0.00000731, 0.00001568, -0.00001229, -0.00002504, -0.00000221, 0.00001150, -0.00000010, -0.00000788, -0.00000404, -0.00000286, -0.00000340, -0.00000069, 0.00000027, -0.00000009, 0.00000235, 0.00000386, 0.00000261, 0.00000176, -0.00000079, -0.00000485, -0.00000272, 0.00000252, -0.00000130, -0.00000882, -0.00000422, 0.00000747, 0.00000882, 0.00000054, -0.00000420, -0.00000292, -0.00000153, -0.00000165, -0.00000209, -0.00000270, -0.00000181, 0.00000254, 0.00000698, 0.00000575, },
		{ 0.00000391, 0.00000376, -0.00000508, -0.00001152, -0.00000669, 0.00000117, 0.00000322, 0.00000323, 0.00000344, -0.00000121, -0.00000810, -0.00000811, -0.00000193, 0.00000247, 0.00000258, 0.00000061, 0.00000002, 0.00000413, 0.00000849, 0.00000429, -0.00000460, -0.00000573, -0.00000021, 0.00000092, -0.00000295, -0.00000464, -0.00000153, 0.00000497, 0.00000945, 0.00000438, -0.00000606, -0.00000839, -0.00000260, -0.00000006, -0.00000166, -0.00000159, -0.00000092, -0.00000010, 0.00000279, 0.00000334, 0.00000124, 0.00000351, 0.00000453, -0.00000383, -0.00000703, 0.00000209, 0.00000082, -0.00001074, -0.00000151, 0.00001465, -0.00000304, -0.00002434, 0.00000039, 0.00002594, -0.00000125, -0.00002415, 0.00000705, 0.00002559, -0.00000960, -0.00002584, 0.00000811, 0.00002211, -0.00000347, -0.00000914, 0.00001149, 0.00001091, -0.00001104, -0.00001984, -0.00000575, 0.00001972, 0.00002669, -0.00000990, -0.00004352, -0.00000896, 0.00004262, 0.00001817, -0.00003530, -0.00001781, 0.00002911, 0.00001886, -0.00000827, 0.00000715, 0.00001915, -0.00000221, -0.00000381, 0.00001882, 0.00000979, -0.00001454, -0.00000049, 0.00001991, -0.00000230, -0.00002683, -0.00000951, 0.00001260, 0.00000414, -0.00000893, -0.00000554, -0.00000123, -0.00000313, -0.00000238, 0.00000020, 0.00000068, 0.00000190, 0.00000339, 0.00000289, 0.00000221, 0.00000003, -0.00000438, -0.00000365, 0.00000210, 0.00000022, -0.00000832, -0.00000666, 0.00000543, 0.00000988, 0.00000262, -0.00000356, -0.00000334, -0.00000182, -0.00000147, -0.00000178, -0.00000291, -0.00000268, 0.00000162, 0.00000678, 0.00000649, 0.00000153, 0.00000000, },
		{ -0.00000229, -0.00001039, -0.00000870, -0.00000075, 0.00000311, 0.00000320, 0.00000355, 0.00000061, -0.00000638, -0.00000912, -0.00000418, 0.00000173, 0.00000323, 0.00000116, -0.00000039, 0.00000291, 0.00000822, 0.00000613, -0.00000281, -0.00000619, -0.00000143, 0.00000100, -0.00000217, -0.00000441, -0.00000254, 0.00000275, 0.00000855, 0.00000699, -0.00000320, -0.00000922, -0.00000456, 0.00000046, -0.00000067, -0.00000239, -0.00000173, 0.00000004, 0.00000277, 0.00000331, 0.00000118, 0.00000296, 0.00000504, -0.00000236, -0.00000772, 0.00000083, 0.00000277, -0.00001032, -0.00000614, 0.00001450, 0.00000420, -0.00002469, -0.00000984, 0.00002531, 0.00000915, -0.00002419, -0.00000270, 0.00002742, 0.00000007, -0.00002749, 0.00000156, 0.00002525, 0.00000052, -0.00001258, 0.00001151, 0.00001659, -0.00001117, -0.00002348, -0.00000410, 0.00002071, 0.00002362, -0.00000782, -0.00003907, -0.00001322, 0.00003823, 0.00002465, -0.00003122, -0.00002546, 0.00002500, 0.00002365, -0.00000922, 0.00000173, 0.00002247, 0.00000142, -0.00001134, 0.00001424, 0.00001723, -0.00001242, -0.00000911, 0.00001980, 0.00000840, -0.00002477, -0.00001728, 0.00001111, 0.00000901, -0.00000852, -0.00000784, -0.00000019, -0.00000197, -0.00000390, -0.00000056, 0.00000152, 0.00000186, 0.00000277, 0.00000298, 0.00000263, 0.00000077, -0.00000370, -0.00000433, 0.00000131, 0.00000148, -0.00000719, -0.00000859, 0.00000283, 0.00001020, 0.00000474, -0.00000247, -0.00000356, -0.00000226, -0.00000148, -0.00000135, -0.00000279, -0.00000354, 0.00000042, 0.00000640, 0.00000719, 0.00000235, -0.00000038, 0.00000281, 0.00000452, },
		{ -0.00000995, -0.00000299, 0.00000256, 0.00000327, 0.00000347, 0.00000197, -0.00000430, -0.00000930, -0.00000641, 0.00000040, 0.00000363, 0.00000188, -0.00000050, 0.00000180, 0.00000752, 0.00000745, -0.00000076, -0.00000604, -0.00000264, 0.00000071, -0.00000146, -0.00000396, -0.00000317, 0.00000075, 0.00000701, 0.00000870, 0.00000002, -0.00000901, -0.00000661, 0.00000027, 0.00000049, -0.00000267, -0.00000266, -0.00000013, 0.00000273, 0.00000336, 0.00000132, 0.00000234, 0.00000506, -0.00000071, -0.00000770, -0.00000094, 0.00000400, -0.00000856, -0.00001005, 0.00001192, 0.00001071, -0.00002140, -0.00001915, 0.00002024, 0.00001844, -0.00001992, -0.00001195, 0.00002514, 0.00000991, -0.00002561, -0.00000571, 0.00002608, 0.00000606, -0.00001499, 0.00000929, 0.00002234, -0.00000876, -0.00002788, -0.00000438, 0.00002330, 0.00002156, -0.00000805, -0.00003469, -0.00001504, 0.00003276, 0.00002877, -0.00002518, -0.00003114, 0.00001869, 0.00002764, -0.00000782, -0.00000431, 0.00002349, 0.00000717, -0.00001652, 0.00000669, 0.00002277, -0.00000675, -0.00001665, 0.00001536, 0.00001794, -0.00001883, -0.00002404, 0.00000686, 0.00001349, -0.00000635, -0.00001038, -0.00000016, -0.00000019, -0.00000487, -0.00000194, 0.00000214, 0.00000225, 0.00000217, 0.00000284, 0.00000299, 0.00000141, -0.00000289, -0.00000467, 0.00000029, 0.00000230, -0.00000560, -0.00000980, -0.00000011, 0.00000968, 0.00000671, -0.00000096, -0.00000349, -0.00000273, -0.00000172, -0.00000094, -0.00000232, -0.00000418, -0.00000102, 0.00000571, 0.00000781, 0.00000330, -0.00000047, 0.00000174, 0.00000462, 0.00000011, -0.00000851, },
		{ 0.00000147, 0.00000334, 0.00000335, 0.00000282, -0.00000212, -0.00000864, -0.00000835, -0.00000147, 0.00000363, 0.00000271, -0.00000033, 0.00000088, 0.00000654, 0.00000820, 0.00000130, -0.00000529, -0.00000365, 0.00000009, -0.00000093, -0.00000337, -0.00000344, -0.00000092, 0.00000507, 0.00000945, 0.00000322, -0.00000780, -0.00000842, -0.00000060, 0.00000156, -0.00000244, -0.00000355, -0.00000057, 0.00000260, 0.00000341, 0.00000164, 0.00000185, 0.00000462, 0.00000078, -0.00000691, -0.00000280, 0.00000421, -0.00000601, -0.00001254, 0.00000756, 0.00001531, -0.00001525, -0.00002594, 0.00001162, 0.00002480, -0.00001214, -0.00001891, 0.00001918, 0.00001821, -0.00002024, -0.00001232, 0.00002421, 0.00001221, -0.00001552, 0.00000528, 0.00002701, -0.00000381, -0.00003186, -0.00000702, 0.00002652, 0.00002136, -0.00000996, -0.00003153, -0.00001458, 0.00002744, 0.00003020, -0.00001832, -0.00003413, 0.00001102, 0.00002986, -0.00000437, -0.00000981, 0.00002186, 0.00001377, -0.00001845, -0.00000263, 0.00002511, 0.00000149, -0.00002139, 0.00000744, 0.00002454, -0.00000977, -0.00002834, 0.00000032, 0.00001654, -0.00000254, -0.00001242, -0.00000126, 0.00000177, -0.00000501, -0.00000370, 0.00000229, 0.00000298, 0.00000175, 0.00000252, 0.00000325, 0.00000194, -0.00000204, -0.00000465, -0.00000080, 0.00000258, -0.00000379, -0.00001017, -0.00000307, 0.00000827, 0.00000829, 0.00000089, -0.00000305, -0.00000315, -0.00000216, -0.00000068, -0.00000159, -0.00000446, -0.00000258, 0.00000464, 0.00000827, 0.00000434, -0.00000027, 0.00000082, 0.00000424, 0.00000193, -0.00000619, -0.00001029, -0.00000524, },
		},
		{ { -0.00000522, -0.00000512, -0.00000050, 0.00000863, 0.00001152, 0.00000513, -0.00000222, -0.00000482, -0.00000229, -0.00000023, -0.00000580, -0.00000949, 0.00000090, 0.00000904, 0.00000058, -0.00000588, -0.00000170, -0.00000029, 0.00000256, 0.00000793, -0.00000244, -0.00001508, -0.00000420, 0.00000932, 0.00000562, 0.00000194, 0.00000115, -0.00000431, -0.00000570, -0.00000658, -0.00000807, 0.00000223, 0.00000130, -0.00002300, -0.00002123, 0.00000672, 0.00000358, -0.00000684, 0.00001697, 0.00002665, 0.00000538, 0.00000037, 0.00000515, 0.00000666, 0.00002184, 0.00001307, -0.00002296, -0.00000578, 0.00003209, -0.00000462, -0.00004083, 0.00000186, 0.00002476, -0.00000910, -0.00000983, 0.00000990, -0.00000657, -0.00001108, 0.00001287, 0.00000718, -0.00001422, -0.00000331, 0.00001299, 0.00000159, -0.00001073, 0.00000159, 0.00001299, -0.00000331, -0.00001422, 0.00000718, 0.00001287, -0.00001108, -0.00000657, 0.00000990, -0.00000983, -0.00000910, 0.00002476, 0.00000186, -0.00004083, -0.00000462, 0.00003209, -0.00000578, -0.00002296, 0.00001307, 0.00002184, 0.00000666, 0.00000515, 0.00000037, 0.00000538, 0.00002665, 0.00001697, -0.00000684, 0.00000358, 0.00000672, -0.00002123, -0.00002300, 0.00000130, 0.00000223, -0.00000807, -0.00000658, -0.00000570, -0.00000431, 0.00000115, 0.00000194, 0.00000562, 0.00000932, -0.00000420, -0.00001508, -0.00000244, 0.00000793, 0.00000256, -0.00000029, -0.00000170, -0.00000588, 0.00000058, 0.00000904, 0.00000090, -0.00000949, -0.00000580, -0.00000023, -0.00000229, -0.00000482, -0.00000222, 0.00000513, 0.00001152, 0.00000863, -0.00000050, -0.00000512, },
		{ -0.00000277, 0.00000628, 0.00001227, 0.00000736, -0.00000093, -0.00000434, -0.00000308, -0.00000079, -0.00000352, -0.00000900, -0.00000365, 0.00000772, 0.00000484, -0.00000509, -0.00000377, 0.00000004, 0.00000081, 0.00000614, 0.00000254, -0.00001247, -0.00000951, 0.00000753, 0.00000787, 0.00000097, 0.00000227, -0.00000117, -0.00000670, -0.00000632, -0.00000834, -0.00000366, 0.00000496, -0.00001178, -0.00002847, -0.00000623, 0.00000995, -0.00000429, 0.00000473, 0.00002884, 0.00001690, -0.00000180, 0.00000324, 0.00000926, 0.00001749, 0.00001857, -0.00001067, -0.00001599, 0.00002217, 0.00001225, -0.00003582, -0.00001672, 0.00002416, 0.00000225, -0.00001526, 0.00000707, 0.00000125, -0.00001528, 0.00000628, 0.00001452, -0.00001081, -0.00001069, 0.00001140, 0.00000693, -0.00000987, -0.00000321, 0.00001181, 0.00000313, -0.00001308, -0.00000066, 0.00001447, -0.00000360, -0.00001177, 0.00000745, -0.00000176, -0.00001580, 0.00001649, 0.00001881, -0.00003384, -0.00002461, 0.00003167, 0.00001198, -0.00002954, 0.00000000, 0.00002699, 0.00000777, 0.00000176, 0.00000523, -0.00000003, 0.00001641, 0.00002607, -0.00000012, -0.00000626, 0.00001066, -0.00000669, -0.00002778, -0.00000879, 0.00000617, -0.00000455, -0.00000809, -0.00000439, -0.00000531, -0.00000190, 0.00000243, 0.00000482, 0.00000903, 0.00000104, -0.00001431, -0.00000858, 0.00000715, 0.00000579, -0.00000041, -0.00000073, -0.00000460, -0.00000320, 0.00000741, 0.00000556, -0.00000765, -0.00000863, -0.00000074, -0.00000080, -0.00000501, -0.00000360, 0.00000324, 0.00001023, 0.00001007, 0.00000206, -0.00000429, -0.00000520, -0.00000543, },
		{ 0.00001198, 0.00000982, 0.00000070, -0.00000398, -0.00000327, -0.00000153, -0.00000242, -0.00000722, -0.00000654, 0.00000411, 0.00000760, -0.00000203, -0.00000561, -0.00000072, 0.00000064, 0.00000363, 0.00000484, -0.00000770, -0.00001282, 0.00000325, 0.00001027, 0.00000139, 0.00000107, 0.00000222, -0.00000554, -0.00000763, -0.00000724, -0.00000763, 0.00000220, -0.00000101, -0.00002528, -0.00002106, 0.00000688, 0.00000384, -0.00000311, 0.00002184, 0.00002826, 0.00000324, -0.00000207, 0.00001050, 0.00001729, 0.00001889, 0.00000058, -0.00001655, 0.00000863, 0.00002001, -0.00002245, -0.00002973, 0.00001436, 0.00001319, -0.00001479, 0.00000008, 0.00000793, -0.00001398, -0.00000319, 0.00001790, -0.00000288, -0.00001643, 0.00000617, 0.00001186, -0.00000706, -0.00000725, 0.00000906, 0.00000742, -0.00000921, -0.00000612, 0.00001163, 0.00000364, -0.00001231, 0.00000159, 0.00000471, -0.00001574, 0.00000385, 0.00002774, -0.00001693, -0.00003928, 0.00001877, 0.00003015, -0.00002484, -0.00001781, 0.00002688, 0.00001507, -0.00000378, 0.00000666, 0.00000245, 0.00000445, 0.00002598, 0.00001295, -0.00001119, 0.00000408, 0.00000729, -0.00002207, -0.00002070, 0.00000409, 0.00000158, -0.00000869, -0.00000463, -0.00000403, -0.00000492, 0.00000090, 0.00000523, 0.00000819, 0.00000458, -0.00001066, -0.00001348, 0.00000306, 0.00000893, 0.00000110, -0.00000103, -0.00000267, -0.00000496, 0.00000365, 0.00000840, -0.00000356, -0.00001059, -0.00000286, 0.00000079, -0.00000427, -0.00000513, 0.00000136, 0.00000886, 0.00001067, 0.00000444, -0.00000289, -0.00000506, -0.00000543, -0.00000442, 0.00000334, },
		{ 0.00000302, -0.00000365, -0.00000337, -0.00000186, -0.00000225, -0.00000542, -0.00000731, -0.00000012, 0.00000768, 0.00000209, -0.00000571, -0.00000272, 0.00000089, 0.00000206, 0.00000462, -0.00000294, -0.00001300, -0.00000233, 0.00001096, 0.00000395, -0.00000104, 0.00000368, -0.00000199, -0.00000887, -0.00000711, -0.00000839, -0.00000340, 0.00000401, -0.00001466, -0.00002972, -0.00000538, 0.00000994, -0.00000267, 0.00001027, 0.00003289, 0.00001501, -0.00000536, 0.00000693, 0.00001992, 0.00001896, 0.00000705, -0.00001071, -0.00000146, 0.00001815, -0.00000783, -0.00003307, -0.00000043, 0.00001850, -0.00000840, -0.00000742, 0.00001023, -0.00000778, -0.00001162, 0.00001528, 0.00000711, -0.00001769, -0.00000213, 0.00001464, -0.00000211, -0.00001024, 0.00000540, 0.00000974, -0.00000473, -0.00000830, 0.00000682, 0.00000791, -0.00000888, -0.00000427, 0.00000693, -0.00001056, -0.00000732, 0.00002642, 0.00000323, -0.00004229, -0.00000246, 0.00004014, -0.00000850, -0.00003309, 0.00001705, 0.00002596, -0.00000554, 0.00000150, 0.00000887, -0.00000192, 0.00001633, 0.00002463, -0.00000599, -0.00000773, 0.00001268, -0.00000781, -0.00002740, -0.00000499, 0.00000682, -0.00000611, -0.00000656, -0.00000210, -0.00000587, -0.00000261, 0.00000525, 0.00000820, 0.00000619, -0.00000591, -0.00001531, -0.00000324, 0.00000982, 0.00000451, -0.00000141, -0.00000159, -0.00000450, -0.00000031, 0.00000831, 0.00000144, -0.00001034, -0.00000616, 0.00000137, -0.00000239, -0.00000627, -0.00000091, 0.00000748, 0.00001081, 0.00000633, -0.00000114, -0.00000461, -0.00000528, -0.00000539, 0.00000037, 0.00001046, 0.00001196, },
		{ -0.00000373, -0.00000186, -0.00000230, -0.00000436, -0.00000662, -0.00000334, 0.00000529, 0.00000537, -0.00000350, -0.00000484, 0.00000025, 0.00000184, 0.00000330, 0.00000010, -0.00001057, -0.00000718, 0.00000886, 0.00000765, -0.00000193, 0.00000237, 0.00000224, -0.00000801, -0.00000865, -0.00000745, -0.00000772, 0.00000232, -0.00000318, -0.00002811, -0.00002065, 0.00000764, 0.00000378, 0.00000121, 0.00002836, 0.00002820, -0.00000168, -0.00000043, 0.00002100, 0.00002216, 0.00000975, -0.00000401, -0.00000473, 0.00001114, 0.00000169, -0.00002781, -0.00001382, 0.00001589, 0.00000052, -0.00001155, 0.00000746, 0.00000017, -0.00001554, 0.00000746, 0.00001523, -0.00001315, -0.00001107, 0.00001343, 0.00000439, -0.00001141, 0.00000092, 0.00001069, -0.00000083, -0.00000813, 0.00000261, 0.00000874, -0.00000406, -0.00000745, 0.00000512, -0.00000404, -0.00001313, 0.00001743, 0.00001905, -0.00003275, -0.00002359, 0.00003673, 0.00001371, -0.00003812, -0.00000176, 0.00003374, 0.00000077, -0.00000798, 0.00001238, 0.00000070, 0.00000290, 0.00002738, 0.00000780, -0.00001583, 0.00000648, 0.00000758, -0.00002379, -0.00001740, 0.00000677, -0.00000026, -0.00000828, -0.00000163, -0.00000418, -0.00000636, 0.00000325, 0.00000919, 0.00000700, -0.00000197, -0.00001387, -0.00000940, 0.00000718, 0.00000851, -0.00000029, -0.00000184, -0.00000301, -0.00000270, 0.00000569, 0.00000533, -0.00000748, -0.00000925, 0.00000011, 0.00000004, -0.00000622, -0.00000355, 0.00000568, 0.00001080, 0.00000779, 0.00000061, -0.00000382, -0.00000497, -0.00000582, -0.00000222, 0.00000795, 0.00001317, 0.00000594, -0.00000286, },
		{ -0.00000211, -0.00000396, -0.00000556, -0.00000483, 0.00000179, 0.00000640, 0.00000015, -0.00000557, -0.00000166, 0.00000208, 0.00000246, 0.00000106, -0.00000721, -0.00000972, 0.00000451, 0.00001051, -0.00000028, -0.00000059, 0.00000476, -0.00000440, -0.00001047, -0.00000720, -0.00000891, -0.00000303, 0.00000328, -0.00001814, -0.00003087, -0.00000388, 0.00000949, -0.00000073, 0.00001786, 0.00003614, 0.00000967, -0.00000664, 0.00001683, 0.00002774, 0.00001266, -0.00000017, -0.00000273, 0.00000501, 0.00000398, -0.00001911, -0.00002096, 0.00000737, 0.00000721, -0.00001066, 0.00000182, 0.00000615, -0.00001382, -0.00000228, 0.00001814, -0.00000409, -0.00001723, 0.00000766, 0.00001060, -0.00000975, -0.00000413, 0.00001041, 0.00000251, -0.00000713, 0.00000010, 0.00000764, -0.00000034, -0.00000752, 0.00000165, 0.00000051, -0.00001322, 0.00000610, 0.00002588, -0.00001540, -0.00003639, 0.00002106, 0.00003257, -0.00002936, -0.00002308, 0.00003179, 0.00001429, -0.00001533, 0.00000805, 0.00000945, -0.00000605, 0.00001891, 0.00002260, -0.00001331, -0.00000713, 0.00001533, -0.00001037, -0.00002626, -0.00000058, 0.00000596, -0.00000727, -0.00000332, -0.00000120, -0.00000813, -0.00000109, 0.00000982, 0.00000843, 0.00000045, -0.00001055, -0.00001312, 0.00000153, 0.00001083, 0.00000302, -0.00000247, -0.00000209, -0.00000294, 0.00000225, 0.00000665, -0.00000304, -0.00001053, -0.00000287, 0.00000177, -0.00000459, -0.00000590, 0.00000305, 0.00001055, 0.00000913, 0.00000220, -0.00000288, -0.00000443, -0.00000581, -0.00000424, 0.00000491, 0.00001313, 0.00000894, -0.00000121, -0.00000420, -0.00000196, },
		{ -0.00000484, -0.00000491, -0.00000129, 0.00000501, 0.00000352, -0.00000410, -0.00000396, 0.00000153, 0.00000266, 0.00000094, -0.00000461, -0.00000978, -0.00000038, 0.00001090, 0.00000345, -0.00000292, 0.00000420, 0.00000054, -0.00001038, -0.00000876, -0.00000803, -0.00000782, 0.00000286, -0.00000602, -0.00003110, -0.00001942, 0.00000799, 0.00000349, 0.00000779, 0.00003524, 0.00002443, -0.00000626, 0.00000766, 0.00003160, 0.00001906, 0.00000119, 0.00000019, 0.00000341, 0.00000169, -0.00001267, -0.00002124, -0.00000219, 0.00000858, -0.00000630, -0.00000325, 0.00000806, -0.00000819, -0.00000995, 0.00001494, 0.00000601, -0.00001799, -0.00000116, 0.00001407, -0.00000501, -0.00000878, 0.00000846, 0.00000574, -0.00000611, -0.00000126, 0.00000657, 0.00000142, -0.00000604, -0.00000102, 0.00000208, -0.00001008, -0.00000281, 0.00002390, 0.00000238, -0.00003703, -0.00000007, 0.00004040, -0.00000988, -0.00003804, 0.00001797, 0.00002880, -0.00001432, -0.00000354, 0.00001712, -0.00000515, 0.00000390, 0.00002953, 0.00000012, -0.00001901, 0.00001075, 0.00000629, -0.00002549, -0.00001274, 0.00000810, -0.00000264, -0.00000578, 0.00000082, -0.00000679, -0.00000632, 0.00000830, 0.00001089, 0.00000209, -0.00000731, -0.00001355, -0.00000497, 0.00000965, 0.00000755, -0.00000178, -0.00000251, -0.00000193, -0.00000012, 0.00000533, 0.00000104, -0.00000921, -0.00000633, 0.00000170, -0.00000189, -0.00000697, -0.00000038, 0.00000954, 0.00001056, 0.00000379, -0.00000204, -0.00000376, -0.00000535, -0.00000564, 0.00000177, 0.00001192, 0.00001140, 0.00000128, -0.00000431, -0.00000242, -0.00000174, -0.00000373, },
		{ -0.00000307, 0.00000220, 0.00000510, -0.00000096, -0.00000518, -0.00000033, 0.00000322, 0.00000109, -0.00000340, -0.00000849, -0.00000407, 0.00000866, 0.00000740, -0.00000279, 0.00000123, 0.00000428, -0.00000729, -0.00001105, -0.00000748, -0.00000964, -0.00000203, 0.00000202, -0.00002221, -0.00003111, -0.00000250, 0.00000818, 0.00000323, 0.00002713, 0.00003605, 0.00000270, -0.00000206, 0.00002956, 0.00002857, 0.00000366, 0.00000058, 0.00000564, -0.00000025, -0.00001114, -0.00001812, -0.00000829, 0.00000508, -0.00000203, -0.00000541, 0.00000647, -0.00000188, -0.00001307, 0.00000756, 0.00001321, -0.00001306, -0.00000986, 0.00001301, 0.00000167, -0.00001141, 0.00000447, 0.00000875, -0.00000473, -0.00000272, 0.00000643, 0.00000223, -0.00000495, -0.00000202, 0.00000188, -0.00000673, -0.00000739, 0.00001687, 0.00001463, -0.00002745, -0.00001810, 0.00003499, 0.00001220, -0.00004026, -0.00000353, 0.00003630, -0.00000324, -0.00001639, 0.00001690, 0.00000462, -0.00000891, 0.00002404, 0.00001771, -0.00002068, -0.00000357, 0.00001701, -0.00001401, -0.00002334, 0.00000321, 0.00000359, -0.00000649, 0.00000053, -0.00000316, -0.00001004, 0.00000382, 0.00001332, 0.00000438, -0.00000528, -0.00001163, -0.00000976, 0.00000491, 0.00001102, 0.00000127, -0.00000358, -0.00000128, -0.00000051, 0.00000280, 0.00000305, -0.00000591, -0.00000855, -0.00000040, 0.00000061, -0.00000616, -0.00000383, 0.00000720, 0.00001179, 0.00000577, -0.00000131, -0.00000325, -0.00000448, -0.00000635, -0.00000125, 0.00000981, 0.00001295, 0.00000417, -0.00000365, -0.00000311, -0.00000152, -0.00000335, -0.00000454, -0.00000440, },
		{ 0.00000443, 0.00000235, -0.00000434, -0.00000281, 0.00000299, 0.00000209, -0.00000292, -0.00000728, -0.00000590, 0.00000498, 0.00000967, -0.00000003, -0.00000194, 0.00000497, -0.00000221, -0.00001180, -0.00000875, -0.00000905, -0.00000720, 0.00000324, -0.00000983, -0.00003335, -0.00001777, 0.00000701, 0.00000443, 0.00001719, 0.00003989, 0.00001733, -0.00000645, 0.00002082, 0.00003715, 0.00001073, -0.00000135, 0.00000797, 0.00000156, -0.00001286, -0.00001634, -0.00000964, 0.00000012, -0.00000078, -0.00000492, 0.00000375, 0.00000248, -0.00001179, -0.00000051, 0.00001523, -0.00000469, -0.00001513, 0.00000754, 0.00000788, -0.00001073, -0.00000102, 0.00001057, -0.00000217, -0.00000484, 0.00000652, 0.00000372, -0.00000496, -0.00000234, 0.00000177, -0.00000465, -0.00000869, 0.00000928, 0.00001939, -0.00001349, -0.00002714, 0.00002032, 0.00002815, -0.00002941, -0.00002426, 0.00003178, 0.00001354, -0.00002288, 0.00000696, 0.00001649, -0.00001210, 0.00000891, 0.00002968, -0.00000959, -0.00001895, 0.00001546, 0.00000296, -0.00002573, -0.00000754, 0.00000729, -0.00000385, -0.00000147, 0.00000062, -0.00001050, -0.00000264, 0.00001378, 0.00000812, -0.00000411, -0.00000929, -0.00001135, -0.00000157, 0.00001122, 0.00000616, -0.00000361, -0.00000205, 0.00000048, 0.00000097, 0.00000254, -0.00000239, -0.00000842, -0.00000351, 0.00000156, -0.00000378, -0.00000613, 0.00000352, 0.00001209, 0.00000832, -0.00000031, -0.00000319, -0.00000346, -0.00000620, -0.00000394, 0.00000702, 0.00001355, 0.00000698, -0.00000222, -0.00000359, -0.00000168, -0.00000288, -0.00000435, -0.00000392, -0.00000358, -0.00000064, },
		{ -0.00000170, -0.00000449, 0.00000134, 0.00000338, -0.00000207, -0.00000680, -0.00000647, 0.00000148, 0.00000946, 0.00000387, -0.00000313, 0.00000275, 0.00000235, -0.00000954, -0.00001112, -0.00000835, -0.00000979, -0.00000069, -0.00000040, -0.00002604, -0.00003037, -0.00000244, 0.00000693, 0.00001055, 0.00003592, 0.00003162, -0.00000209, 0.00000900, 0.00003993, 0.00002262, -0.00000204, 0.00000699, 0.00000672, -0.00001360, -0.00001850, -0.00000870, -0.00000251, -0.00000278, -0.00000416, 0.00000203, 0.00000434, -0.00000821, -0.00000625, 0.00001240, 0.00000354, -0.00001530, -0.00000016, 0.00001114, -0.00000677, -0.00000638, 0.00001006, 0.00000177, -0.00000693, 0.00000538, 0.00000661, -0.00000506, -0.00000356, 0.00000247, -0.00000328, -0.00000890, 0.00000368, 0.00001859, -0.00000112, -0.00002650, 0.00000376, 0.00003311, -0.00001111, -0.00003621, 0.00001629, 0.00002801, -0.00001860, -0.00000836, 0.00002233, -0.00000419, -0.00000745, 0.00002921, 0.00000911, -0.00002572, 0.00000223, 0.00001637, -0.00001724, -0.00001875, 0.00000495, 0.00000111, -0.00000304, 0.00000265, -0.00000768, -0.00000879, 0.00001087, 0.00001264, -0.00000238, -0.00000808, -0.00001017, -0.00000706, 0.00000743, 0.00001077, -0.00000113, -0.00000388, 0.00000126, 0.00000105, 0.00000061, -0.00000050, -0.00000618, -0.00000595, 0.00000043, -0.00000105, -0.00000639, -0.00000068, 0.00001065, 0.00001104, 0.00000156, -0.00000349, -0.00000281, -0.00000518, -0.00000595, 0.00000366, 0.00001326, 0.00000946, -0.00000037, -0.00000354, -0.00000199, -0.00000260, -0.00000411, -0.00000360, -0.00000337, -0.00000261, 0.00000218, 0.00000431, },
		{ -0.00000115, 0.00000384, -0.00000032, -0.00000661, -0.00000687, -0.00000095, 0.00000744, 0.00000694, -0.00000161, -0.00000045, 0.00000420, -0.00000484, -0.00001242, -0.00000929, -0.00000983, -0.00000580, 0.00000260, -0.00001407, -0.00003432, -0.00001693, 0.00000495, 0.00000844, 0.00002809, 0.00004025, 0.00000982, 0.00000008, 0.00003471, 0.00003561, 0.00000262, 0.00000260, 0.00001161, -0.00001003, -0.00002330, -0.00000961, -0.00000147, -0.00000544, -0.00000549, 0.00000150, 0.00000495, -0.00000464, -0.00000868, 0.00000712, 0.00000873, -0.00001113, -0.00000698, 0.00001031, -0.00000112, -0.00000971, 0.00000681, 0.00000614, -0.00000755, 0.00000217, 0.00000993, -0.00000348, -0.00000621, 0.00000302, -0.00000125, -0.00000929, -0.00000005, 0.00001568, 0.00000660, -0.00001988, -0.00000825, 0.00002808, 0.00000652, -0.00003604, -0.00000369, 0.00003322, -0.00000513, -0.00002122, 0.00001766, 0.00000963, -0.00001609, 0.00001654, 0.00002559, -0.00001901, -0.00001519, 0.00001863, -0.00000145, -0.00002371, -0.00000350, 0.00000500, -0.00000230, 0.00000265, -0.00000319, -0.00001232, 0.00000475, 0.00001610, 0.00000119, -0.00000816, -0.00000804, -0.00000949, 0.00000099, 0.00001253, 0.00000383, -0.00000512, 0.00000044, 0.00000276, -0.00000079, -0.00000102, -0.00000336, -0.00000636, -0.00000208, 0.00000050, -0.00000470, -0.00000405, 0.00000731, 0.00001292, 0.00000462, -0.00000357, -0.00000303, -0.00000366, -0.00000682, -0.00000004, 0.00001197, 0.00001163, 0.00000162, -0.00000308, -0.00000205, -0.00000258, -0.00000400, -0.00000331, -0.00000287, -0.00000354, -0.00000050, 0.00000426, 0.00000147, -0.00000430, },
		{ 0.00000178, -0.00000577, -0.00000765, -0.00000254, 0.00000493, 0.00000803, 0.00000160, -0.00000230, 0.00000296, -0.00000002, -0.00001086, -0.00001152, -0.00000938, -0.00000892, 0.00000009, -0.00000387, -0.00002870, -0.00002961, -0.00000421, 0.00000765, 0.00002100, 0.00004162, 0.00002441, -0.00000126, 0.00002374, 0.00004425, 0.00001393, -0.00000168, 0.00001246, -0.00000228, -0.00002668, -0.00001495, 0.00000112, -0.00000546, -0.00000914, 0.00000047, 0.00000584, -0.00000199, -0.00000877, 0.00000204, 0.00001004, -0.00000518, -0.00001058, 0.00000621, 0.00000385, -0.00000992, 0.00000177, 0.00000927, -0.00000569, -0.00000246, 0.00001172, 0.00000074, -0.00000909, 0.00000176, 0.00000194, -0.00000936, -0.00000338, 0.00001299, 0.00001024, -0.00001223, -0.00001335, 0.00001830, 0.00001746, -0.00002637, -0.00002017, 0.00002716, 0.00001102, -0.00002518, 0.00000431, 0.00002089, -0.00001287, -0.00000125, 0.00003118, -0.00000172, -0.00002669, 0.00000822, 0.00001342, -0.00001855, -0.00001394, 0.00000429, 0.00000061, 0.00000192, 0.00000098, -0.00001225, -0.00000280, 0.00001654, 0.00000663, -0.00000829, -0.00000696, -0.00000857, -0.00000538, 0.00001001, 0.00000950, -0.00000396, -0.00000205, 0.00000453, -0.00000017, -0.00000316, -0.00000190, -0.00000462, -0.00000436, 0.00000005, -0.00000215, -0.00000542, 0.00000282, 0.00001287, 0.00000845, -0.00000251, -0.00000410, -0.00000244, -0.00000629, -0.00000354, 0.00000947, 0.00001344, 0.00000378, -0.00000258, -0.00000171, -0.00000242, -0.00000425, -0.00000313, -0.00000218, -0.00000369, -0.00000276, 0.00000260, 0.00000371, -0.00000226, -0.00000317, 0.00000277, },
		{ -0.00000840, -0.00000406, 0.00000286, 0.00000734, 0.00000466, -0.00000165, 0.00000022, 0.00000253, -0.00000665, -0.00001301, -0.00001032, -0.00000964, -0.00000431, 0.00000058, -0.00001769, -0.00003445, -0.00001801, 0.00000330, 0.00001634, 0.00003797, 0.00003636, 0.00000569, 0.00001244, 0.00004485, 0.00002903, -0.00000106, 0.00000836, 0.00000609, -0.00002493, -0.00002369, 0.00000120, -0.00000150, -0.00001279, -0.00000301, 0.00000711, 0.00000040, -0.00000809, -0.00000145, 0.00000866, -0.00000011, -0.00001055, 0.00000110, 0.00000634, -0.00000744, -0.00000324, 0.00000984, -0.00000166, -0.00000667, 0.00001038, 0.00000673, -0.00000997, -0.00000209, 0.00000515, -0.00000769, -0.00000721, 0.00001060, 0.00001220, -0.00000659, -0.00001340, 0.00000943, 0.00002048, -0.00001342, -0.00002805, 0.00001354, 0.00002240, -0.00001892, -0.00001107, 0.00002307, -0.00000056, -0.00001467, 0.00002342, 0.00001722, -0.00002543, -0.00000914, 0.00001912, -0.00000511, -0.00002010, -0.00000196, 0.00000326, 0.00000210, 0.00000379, -0.00000918, -0.00000940, 0.00001313, 0.00001257, -0.00000680, -0.00000781, -0.00000584, -0.00000901, 0.00000384, 0.00001316, 0.00000027, -0.00000479, 0.00000466, 0.00000258, -0.00000502, -0.00000290, -0.00000201, -0.00000487, -0.00000201, -0.00000034, -0.00000457, -0.00000134, 0.00001045, 0.00001186, 0.00000028, -0.00000530, -0.00000232, -0.00000466, -0.00000600, 0.00000574, 0.00001448, 0.00000640, -0.00000224, -0.00000134, -0.00000165, -0.00000473, -0.00000343, -0.00000127, -0.00000324, -0.00000435, 0.00000021, 0.00000430, 0.00000056, -0.00000358, 0.00000062, 0.00000310, -0.00000382, },
		{ 0.00000118, 0.00000590, 0.00000623, 0.00000089, -0.00000170, 0.00000202, -0.00000187, -0.00001182, -0.00001256, -0.00000974, -0.00000736, -0.00000044, -0.00000751, -0.00003001, -0.00003020, -0.00000706, 0.00001194, 0.00003273, 0.00004294, 0.00001777, 0.00000617, 0.00003783, 0.00004245, 0.00000715, 0.00000235, 0.00001068, -0.00001750, -0.00003173, -0.00000420, 0.00000449, -0.00001313, -0.00000917, 0.00000715, 0.00000359, -0.00000733, -0.00000367, 0.00000652, 0.00000279, -0.00000832, -0.00000283, 0.00000610, -0.00000389, -0.00000659, 0.00000772, 0.00000303, -0.00000860, 0.00000594, 0.00001216, -0.00000719, -0.00000739, 0.00000626, -0.00000361, -0.00001077, 0.00000714, 0.00001427, -0.00000291, -0.00001207, 0.00000428, 0.00001871, -0.00000296, -0.00002740, -0.00000107, 0.00002496, -0.00000653, -0.00002126, 0.00001519, 0.00001326, -0.00001736, 0.00000720, 0.00002817, -0.00001185, -0.00002372, 0.00001235, 0.00000961, -0.00001753, -0.00001079, 0.00000257, 0.00000344, 0.00000561, -0.00000460, -0.00001325, 0.00000662, 0.00001693, -0.00000273, -0.00000991, -0.00000361, -0.00000874, -0.00000353, 0.00001264, 0.00000640, -0.00000574, 0.00000245, 0.00000612, -0.00000471, -0.00000597, -0.00000051, -0.00000314, -0.00000422, -0.00000037, -0.00000240, -0.00000375, 0.00000623, 0.00001347, 0.00000451, -0.00000546, -0.00000353, -0.00000287, -0.00000676, 0.00000133, 0.00001407, 0.00000961, -0.00000169, -0.00000160, -0.00000021, -0.00000478, -0.00000458, -0.00000048, -0.00000206, -0.00000532, -0.00000226, 0.00000342, 0.00000284, -0.00000223, -0.00000136, 0.00000288, -0.00000130, -0.00000824, -0.00000592, },
		{ 0.00000621, 0.00000367, -0.00000137, -0.00000027, 0.00000096, -0.00000784, -0.00001419, -0.00001112, -0.00000836, -0.00000371, -0.00000235, -0.00001996, -0.00003495, -0.00002112, 0.00000396, 0.00002736, 0.00004467, 0.00003048, 0.00000735, 0.00002727, 0.00004931, 0.00002153, -0.00000045, 0.00000944, -0.00000757, -0.00003473, -0.00001464, 0.00000831, -0.00000845, -0.00001570, 0.00000394, 0.00000753, -0.00000587, -0.00000562, 0.00000475, 0.00000399, -0.00000584, -0.00000470, 0.00000433, -0.00000094, -0.00000773, 0.00000402, 0.00000657, -0.00000754, 0.00000004, 0.00001462, -0.00000099, -0.00001159, 0.00000381, 0.00000177, -0.00001219, 0.00000172, 0.00001620, 0.00000079, -0.00001183, 0.00000189, 0.00001649, 0.00000290, -0.00002229, -0.00001131, 0.00001977, 0.00000552, -0.00002273, 0.00000181, 0.00002109, -0.00000942, -0.00000854, 0.00002634, 0.00000681, -0.00002749, -0.00000315, 0.00001745, -0.00000667, -0.00001678, -0.00000266, 0.00000430, 0.00000729, 0.00000030, -0.00001381, -0.00000118, 0.00001803, 0.00000348, -0.00001154, -0.00000363, -0.00000541, -0.00000904, 0.00000764, 0.00001187, -0.00000350, -0.00000122, 0.00000846, -0.00000156, -0.00000930, -0.00000161, -0.00000013, -0.00000494, -0.00000218, -0.00000047, -0.00000385, 0.00000168, 0.00001247, 0.00000899, -0.00000364, -0.00000543, -0.00000206, -0.00000583, -0.00000268, 0.00001173, 0.00001289, -0.00000015, -0.00000279, 0.00000122, -0.00000358, -0.00000633, -0.00000059, -0.00000011, -0.00000548, -0.00000462, 0.00000171, 0.00000389, -0.00000003, -0.00000201, 0.00000145, 0.00000061, -0.00000670, -0.00000767, -0.00000070, 0.00000452, },
		{ 0.00000089, -0.00000205, 0.00000078, -0.00000303, -0.00001313, -0.00001380, -0.00000893, -0.00000602, -0.00000236, -0.00001027, -0.00003071, -0.00003298, -0.00000934, 0.00002015, 0.00004342, 0.00004071, 0.00001484, 0.00001841, 0.00004813, 0.00003734, 0.00000393, 0.00000434, 0.00000008, -0.00003106, -0.00002632, 0.00000625, -0.00000021, -0.00001902, -0.00000306, 0.00001046, -0.00000259, -0.00000768, 0.00000312, 0.00000475, -0.00000421, -0.00000516, 0.00000259, 0.00000070, -0.00000732, 0.00000021, 0.00000796, -0.00000426, -0.00000505, 0.00001311, 0.00000650, -0.00001222, -0.00000168, 0.00000600, -0.00001012, -0.00000485, 0.00001621, 0.00000609, -0.00001239, -0.00000066, 0.00001618, 0.00000595, -0.00001730, -0.00001589, 0.00001141, 0.00001246, -0.00001704, -0.00001042, 0.00001968, 0.00000342, -0.00001617, 0.00001411, 0.00002095, -0.00001865, -0.00001871, 0.00001389, 0.00000685, -0.00001539, -0.00001026, 0.00000227, 0.00000884, 0.00000518, -0.00001147, -0.00000829, 0.00001536, 0.00001030, -0.00001100, -0.00000620, -0.00000127, -0.00001056, -0.00000012, 0.00001402, 0.00000167, -0.00000439, 0.00000820, 0.00000357, -0.00001074, -0.00000543, 0.00000223, -0.00000333, -0.00000459, -0.00000008, -0.00000224, -0.00000161, 0.00000916, 0.00001212, 0.00000017, -0.00000669, -0.00000282, -0.00000409, -0.00000517, 0.00000771, 0.00001515, 0.00000292, -0.00000438, 0.00000151, -0.00000095, -0.00000761, -0.00000226, 0.00000208, -0.00000434, -0.00000683, -0.00000047, 0.00000384, 0.00000185, -0.00000114, 0.00000000, 0.00000105, -0.00000438, -0.00000835, -0.00000308, 0.00000323, 0.00000541, 0.00000529, },
		{ -0.00000135, 0.00000001, -0.00000908, -0.00001587, -0.00001095, -0.00000662, -0.00000453, -0.00000521, -0.00002096, -0.00003707, -0.00002530, 0.00000814, 0.00003965, 0.00004771, 0.00002564, 0.00001472, 0.00004115, 0.00004926, 0.00001580, 0.00000019, 0.00000227, -0.00002295, -0.00003442, -0.00000231, 0.00000742, -0.00001664, -0.00001183, 0.00000996, 0.00000258, -0.00000896, 0.00000070, 0.00000583, -0.00000312, -0.00000555, 0.00000164, 0.00000152, -0.00000642, -0.00000277, 0.00000735, -0.00000032, -0.00000773, 0.00000845, 0.00001245, -0.00000847, -0.00000784, 0.00000679, -0.00000503, -0.00001025, 0.00001276, 0.00001251, -0.00001147, -0.00000541, 0.00001679, 0.00000954, -0.00001440, -0.00001726, 0.00000438, 0.00001368, -0.00000897, -0.00001694, 0.00001123, 0.00001380, -0.00001347, -0.00000096, 0.00002421, -0.00000245, -0.00002588, 0.00000093, 0.00001536, -0.00000625, -0.00001554, -0.00000363, 0.00000886, 0.00000998, -0.00000683, -0.00001332, 0.00000955, 0.00001582, -0.00000748, -0.00001019, 0.00000124, -0.00000784, -0.00000761, 0.00001149, 0.00000786, -0.00000514, 0.00000532, 0.00000871, -0.00000888, -0.00001056, 0.00000213, 0.00000011, -0.00000595, -0.00000154, -0.00000032, -0.00000279, 0.00000486, 0.00001278, 0.00000489, -0.00000607, -0.00000471, -0.00000286, -0.00000572, 0.00000310, 0.00001535, 0.00000725, -0.00000511, -0.00000009, 0.00000220, -0.00000712, -0.00000536, 0.00000328, -0.00000158, -0.00000841, -0.00000319, 0.00000317, 0.00000286, 0.00000043, -0.00000029, 0.00000030, -0.00000253, -0.00000750, -0.00000538, 0.00000148, 0.00000456, 0.00000548, 0.00000357, -0.00000173, },
		{ -0.00000393, -0.00001505, -0.00001450, -0.00000713, -0.00000574, -0.00000501, -0.00001166, -0.00003197, -0.00003785, -0.00000950, 0.00003136, 0.00005171, 0.00003717, 0.00001691, 0.00003254, 0.00005424, 0.00003180, 0.00000170, -0.00000045, -0.00001498, -0.00003614, -0.00001439, 0.00001011, -0.00000913, -0.00001865, 0.00000482, 0.00000799, -0.00000798, -0.00000288, 0.00000675, -0.00000148, -0.00000658, 0.00000099, 0.00000249, -0.00000560, -0.00000494, 0.00000558, 0.00000293, -0.00000776, 0.00000267, 0.00001488, -0.00000167, -0.00001174, 0.00000360, 0.00000076, -0.00001226, 0.00000613, 0.00001773, -0.00000700, -0.00001171, 0.00001532, 0.00001545, -0.00001204, -0.00001888, 0.00000018, 0.00001203, -0.00000281, -0.00001746, 0.00000100, 0.00001721, -0.00000407, -0.00001101, 0.00001675, 0.00001231, -0.00002128, -0.00001407, 0.00001378, 0.00000610, -0.00001420, -0.00001143, 0.00000536, 0.00001397, -0.00000026, -0.00001541, 0.00000186, 0.00001854, -0.00000134, -0.00001372, 0.00000060, -0.00000243, -0.00001198, 0.00000490, 0.00001242, -0.00000270, 0.00000116, 0.00001192, -0.00000375, -0.00001478, -0.00000120, 0.00000373, -0.00000511, -0.00000404, 0.00000058, -0.00000210, 0.00000112, 0.00001097, 0.00000888, -0.00000326, -0.00000642, -0.00000306, -0.00000496, -0.00000068, 0.00001321, 0.00001160, -0.00000380, -0.00000318, 0.00000423, -0.00000424, -0.00000856, 0.00000224, 0.00000228, -0.00000837, -0.00000655, 0.00000198, 0.00000324, 0.00000161, 0.00000074, -0.00000041, -0.00000206, -0.00000572, -0.00000657, -0.00000094, 0.00000362, 0.00000492, 0.00000524, 0.00000059, -0.00000315, -0.00000005, },
		{ -0.00001748, -0.00000961, -0.00000526, -0.00000664, -0.00000714, -0.00002144, -0.00004138, -0.00002920, 0.00001589, 0.00005163, 0.00004820, 0.00002388, 0.00002598, 0.00005253, 0.00004699, 0.00001081, -0.00000421, -0.00001108, -0.00003240, -0.00002522, 0.00000600, -0.00000004, -0.00002029, -0.00000372, 0.00001092, -0.00000407, -0.00000666, 0.00000627, 0.00000136, -0.00000772, -0.00000049, 0.00000399, -0.00000450, -0.00000675, 0.00000336, 0.00000500, -0.00000598, -0.00000226, 0.00001358, 0.00000554, -0.00001158, -0.00000195, 0.00000449, -0.00001026, -0.00000149, 0.00001923, 0.00000096, -0.00001665, 0.00000946, 0.00002229, -0.00000705, -0.00002211, -0.00000307, 0.00001091, 0.00000028, -0.00001505, -0.00000649, 0.00001394, 0.00000568, -0.00001233, 0.00000438, 0.00001856, -0.00000832, -0.00002281, 0.00000285, 0.00001456, -0.00000547, -0.00001676, -0.00000228, 0.00001530, 0.00000768, -0.00001398, -0.00000621, 0.00001764, 0.00000616, -0.00001499, -0.00000327, 0.00000309, -0.00001170, -0.00000345, 0.00001322, 0.00000215, -0.00000233, 0.00001211, 0.00000326, -0.00001607, -0.00000709, 0.00000561, -0.00000205, -0.00000616, -0.00000013, -0.00000061, -0.00000104, 0.00000772, 0.00001087, 0.00000084, -0.00000657, -0.00000448, -0.00000420, -0.00000281, 0.00000957, 0.00001445, -0.00000022, -0.00000625, 0.00000375, 0.00000024, -0.00000995, -0.00000131, 0.00000564, -0.00000584, -0.00000997, -0.00000022, 0.00000353, 0.00000196, 0.00000219, 0.00000012, -0.00000273, -0.00000440, -0.00000608, -0.00000337, 0.00000205, 0.00000429, 0.00000554, 0.00000345, -0.00000284, -0.00000230, -0.00000049, -0.00001083, },
		{ -0.00000501, -0.00000682, -0.00000756, -0.00001183, -0.00003472, -0.00004405, -0.00000678, 0.00004436, 0.00005766, 0.00003438, 0.00002344, 0.00004665, 0.00005748, 0.00002585, -0.00000420, -0.00001205, -0.00002701, -0.00003121, -0.00000316, 0.00000612, -0.00001612, -0.00001228, 0.00000937, 0.00000163, -0.00000879, 0.00000360, 0.00000480, -0.00000765, -0.00000338, 0.00000531, -0.00000237, -0.00000836, 0.00000080, 0.00000604, -0.00000348, -0.00000537, 0.00000980, 0.00001075, -0.00000762, -0.00000711, 0.00000462, -0.00000576, -0.00000727, 0.00001613, 0.00000998, -0.00001712, -0.00000034, 0.00002639, 0.00000236, -0.00002504, -0.00000836, 0.00001135, 0.00000228, -0.00001308, -0.00001011, 0.00000775, 0.00001094, -0.00000671, -0.00000551, 0.00001487, 0.00000529, -0.00002098, -0.00001120, 0.00001386, 0.00000668, -0.00001538, -0.00001189, 0.00001179, 0.00001546, -0.00000858, -0.00001303, 0.00001304, 0.00001334, -0.00001290, -0.00000916, 0.00000624, -0.00000706, -0.00001057, 0.00000960, 0.00000744, -0.00000365, 0.00000951, 0.00001012, -0.00001342, -0.00001373, 0.00000440, 0.00000216, -0.00000669, -0.00000207, 0.00000053, -0.00000161, 0.00000443, 0.00001066, 0.00000465, -0.00000467, -0.00000590, -0.00000451, -0.00000361, 0.00000588, 0.00001499, 0.00000458, -0.00000750, 0.00000065, 0.00000429, -0.00000823, -0.00000608, 0.00000656, -0.00000102, -0.00001200, -0.00000394, 0.00000370, 0.00000200, 0.00000287, 0.00000206, -0.00000316, -0.00000460, -0.00000458, -0.00000459, -0.00000038, 0.00000351, 0.00000510, 0.00000535, -0.00000037, -0.00000439, -0.00000039, -0.00000509, -0.00001729, -0.00001417, },
		{ -0.00000948, -0.00000782, -0.00002225, -0.00004788, -0.00003185, 0.00002674, 0.00006276, 0.00004764, 0.00002565, 0.00003954, 0.00006181, 0.00004295, 0.00000262, -0.00001528, -0.00002402, -0.00003201, -0.00001330, 0.00000654, -0.00000862, -0.00001727, 0.00000354, 0.00000658, -0.00000791, -0.00000067, 0.00000726, -0.00000538, -0.00000701, 0.00000512, 0.00000102, -0.00000915, -0.00000235, 0.00000628, -0.00000086, -0.00000672, 0.00000524, 0.00001289, -0.00000174, -0.00000957, 0.00000166, -0.00000135, -0.00000951, 0.00000980, 0.00001680, -0.00001206, -0.00001085, 0.00002444, 0.00001473, -0.00002392, -0.00001683, 0.00001125, 0.00000592, -0.00001248, -0.00001187, 0.00000247, 0.00001089, 0.00000069, -0.00000852, 0.00000552, 0.00001241, -0.00001062, -0.00002043, 0.00000391, 0.00001550, -0.00000631, -0.00001888, 0.00000281, 0.00002031, 0.00000052, -0.00001673, 0.00000531, 0.00001837, -0.00000727, -0.00001506, 0.00000546, 0.00000008, -0.00001397, 0.00000258, 0.00001098, -0.00000231, 0.00000529, 0.00001497, -0.00000712, -0.00001893, -0.00000013, 0.00000591, -0.00000517, -0.00000430, 0.00000075, -0.00000130, 0.00000195, 0.00000913, 0.00000694, -0.00000156, -0.00000591, -0.00000579, -0.00000429, 0.00000318, 0.00001365, 0.00000889, -0.00000608, -0.00000359, 0.00000585, -0.00000376, -0.00000971, 0.00000407, 0.00000440, -0.00001107, -0.00000879, 0.00000286, 0.00000258, 0.00000236, 0.00000429, -0.00000195, -0.00000607, -0.00000366, -0.00000390, -0.00000293, 0.00000192, 0.00000462, 0.00000583, 0.00000280, -0.00000427, -0.00000288, -0.00000097, -0.00001304, -0.00001849, -0.00000740, -0.00000489, },
		{ -0.00001123, -0.00003948, -0.00005093, -0.00000096, 0.00005859, 0.00006201, 0.00003324, 0.00003356, 0.00006057, 0.00005820, 0.00001626, -0.00001647, -0.00002523, -0.00003032, -0.00002063, 0.00000168, -0.00000187, -0.00001707, -0.00000405, 0.00000831, -0.00000424, -0.00000475, 0.00000733, -0.00000124, -0.00000970, 0.00000254, 0.00000477, -0.00000815, -0.00000600, 0.00000552, 0.00000175, -0.00000680, 0.00000107, 0.00001235, 0.00000382, -0.00000869, -0.00000234, 0.00000090, -0.00000853, 0.00000295, 0.00001922, -0.00000312, -0.00001793, 0.00001587, 0.00002581, -0.00001611, -0.00002626, 0.00000726, 0.00001187, -0.00001129, -0.00001439, -0.00000064, 0.00000819, 0.00000529, -0.00000525, -0.00000284, 0.00001063, 0.00000152, -0.00001998, -0.00000983, 0.00001544, 0.00000677, -0.00001863, -0.00000929, 0.00001919, 0.00001152, -0.00001549, -0.00000406, 0.00001962, 0.00000105, -0.00001876, 0.00000052, 0.00000703, -0.00001246, -0.00000558, 0.00001118, 0.00000105, 0.00000100, 0.00001667, 0.00000150, -0.00002085, -0.00000713, 0.00000767, -0.00000196, -0.00000592, 0.00000013, -0.00000081, 0.00000027, 0.00000739, 0.00000763, 0.00000121, -0.00000403, -0.00000679, -0.00000591, 0.00000132, 0.00001175, 0.00001151, -0.00000263, -0.00000683, 0.00000417, 0.00000140, -0.00001011, -0.00000093, 0.00000791, -0.00000672, -0.00001309, -0.00000009, 0.00000379, 0.00000143, 0.00000529, 0.00000116, -0.00000724, -0.00000467, -0.00000196, -0.00000412, -0.00000081, 0.00000399, 0.00000556, 0.00000501, -0.00000169, -0.00000543, -0.00000053, -0.00000663, -0.00001939, -0.00001283, -0.00000315, -0.00000915, -0.00000943, },
		{ -0.00005648, -0.00003270, 0.00004056, 0.00007319, 0.00004676, 0.00003083, 0.00005523, 0.00006905, 0.00003435, -0.00001206, -0.00002927, -0.00002983, -0.00002389, -0.00000543, 0.00000098, -0.00001296, -0.00001005, 0.00000607, 0.00000030, -0.00000677, 0.00000484, 0.00000319, -0.00000993, -0.00000197, 0.00000720, -0.00000488, -0.00000929, 0.00000327, 0.00000422, -0.00000589, -0.00000233, 0.00001020, 0.00000765, -0.00000548, -0.00000520, 0.00000059, -0.00000615, -0.00000213, 0.00001731, 0.00000637, -0.00001902, 0.00000341, 0.00003104, -0.00000232, -0.00003213, -0.00000234, 0.00001763, -0.00000668, -0.00001826, -0.00000329, 0.00000606, 0.00000584, 0.00000010, -0.00000560, 0.00000312, 0.00000839, -0.00001096, -0.00001961, 0.00000579, 0.00001706, -0.00000966, -0.00001939, 0.00001076, 0.00002081, -0.00000840, -0.00001252, 0.00001606, 0.00001023, -0.00001843, -0.00000739, 0.00001110, -0.00000644, -0.00001221, 0.00000763, 0.00000505, -0.00000202, 0.00001502, 0.00001054, -0.00001842, -0.00001497, 0.00000648, 0.00000209, -0.00000646, -0.00000092, -0.00000032, -0.00000115, 0.00000593, 0.00000767, 0.00000249, -0.00000127, -0.00000596, -0.00000832, -0.00000095, 0.00001035, 0.00001257, 0.00000118, -0.00000767, 0.00000036, 0.00000469, -0.00000699, -0.00000591, 0.00000767, -0.00000039, -0.00001461, -0.00000525, 0.00000453, 0.00000140, 0.00000443, 0.00000487, -0.00000627, -0.00000751, -0.00000065, -0.00000294, -0.00000387, 0.00000237, 0.00000536, 0.00000568, 0.00000182, -0.00000565, -0.00000325, -0.00000158, -0.00001543, -0.00001872, -0.00000476, -0.00000571, -0.00001257, -0.00000716, -0.00002401, },
		{ 0.00000844, 0.00007409, 0.00006495, 0.00003398, 0.00004758, 0.00007428, 0.00005374, -0.00000070, -0.00003273, -0.00003284, -0.00002474, -0.00001136, -0.00000059, -0.00000809, -0.00001236, 0.00000129, 0.00000336, -0.00000611, 0.00000109, 0.00000605, -0.00000736, -0.00000672, 0.00000689, -0.00000008, -0.00001084, -0.00000054, 0.00000599, -0.00000395, -0.00000498, 0.00000733, 0.00000959, -0.00000160, -0.00000587, -0.00000109, -0.00000430, -0.00000474, 0.00001289, 0.00001338, -0.00001432, -0.00000843, 0.00002822, 0.00001332, -0.00003045, -0.00001568, 0.00001903, 0.00000226, -0.00002119, -0.00000798, 0.00000560, 0.00000447, 0.00000323, -0.00000267, -0.00000400, 0.00000696, 0.00000054, -0.00002005, -0.00000852, 0.00001844, 0.00000469, -0.00002202, -0.00000291, 0.00002413, 0.00000336, -0.00001690, 0.00000797, 0.00001763, -0.00001323, -0.00001591, 0.00001036, 0.00000219, -0.00001502, 0.00000114, 0.00000803, -0.00000294, 0.00001058, 0.00001792, -0.00001154, -0.00002164, 0.00000202, 0.00000595, -0.00000576, -0.00000221, 0.00000044, -0.00000255, 0.00000426, 0.00000816, 0.00000251, 0.00000050, -0.00000286, -0.00000989, -0.00000477, 0.00000921, 0.00001334, 0.00000399, -0.00000636, -0.00000329, 0.00000470, -0.00000219, -0.00000818, 0.00000395, 0.00000509, -0.00001206, -0.00001107, 0.00000310, 0.00000279, 0.00000258, 0.00000719, -0.00000256, -0.00001036, -0.00000181, 0.00000005, -0.00000551, -0.00000080, 0.00000513, 0.00000555, 0.00000434, -0.00000301, -0.00000641, -0.00000057, -0.00000837, -0.00002118, -0.00001064, -0.00000193, -0.00001255, -0.00001019, -0.00001020, -0.00004671, -0.00005809, },
		{ 0.00008260, 0.00004561, 0.00004035, 0.00007339, 0.00007155, 0.00001695, -0.00003210, -0.00003902, -0.00002623, -0.00001443, -0.00000458, -0.00000545, -0.00001128, -0.00000337, 0.00000364, -0.00000377, -0.00000212, 0.00000642, -0.00000310, -0.00000973, 0.00000368, 0.00000448, -0.00000965, -0.00000513, 0.00000616, -0.00000104, -0.00000672, 0.00000408, 0.00001014, 0.00000170, -0.00000475, -0.00000250, -0.00000376, -0.00000570, 0.00000812, 0.00001667, -0.00000636, -0.00001585, 0.00001871, 0.00002540, -0.00002040, -0.00002795, 0.00001315, 0.00001308, -0.00001969, -0.00001534, 0.00000501, 0.00000424, 0.00000270, 0.00000224, -0.00000593, -0.00000033, 0.00000738, -0.00001146, -0.00001967, 0.00000936, 0.00001741, -0.00001464, -0.00001640, 0.00001875, 0.00001600, -0.00001460, -0.00000246, 0.00002056, -0.00000389, -0.00002206, 0.00000433, 0.00001052, -0.00001276, -0.00000645, 0.00000862, -0.00000162, 0.00000465, 0.00002182, -0.00000120, -0.00002508, -0.00000536, 0.00000853, -0.00000375, -0.00000384, 0.00000159, -0.00000336, 0.00000152, 0.00000922, 0.00000279, -0.00000004, 0.00000109, -0.00000860, -0.00000985, 0.00000673, 0.00001492, 0.00000589, -0.00000451, -0.00000503, 0.00000220, 0.00000146, -0.00000673, -0.00000083, 0.00000715, -0.00000634, -0.00001486, -0.00000130, 0.00000455, 0.00000150, 0.00000700, 0.00000255, -0.00001085, -0.00000570, 0.00000268, -0.00000422, -0.00000480, 0.00000384, 0.00000567, 0.00000509, 0.00000093, -0.00000714, -0.00000348, -0.00000215, -0.00001793, -0.00001817, -0.00000193, -0.00000791, -0.00001546, -0.00000499, -0.00002744, -0.00006738, -0.00003089, 0.00005784, },
		{ 0.00003788, 0.00006670, 0.00008520, 0.00003911, -0.00002512, -0.00004605, -0.00003074, -0.00001550, -0.00000817, -0.00000600, -0.00000902, -0.00000589, 0.00000170, -0.00000154, -0.00000365, 0.00000475, 0.00000101, -0.00000993, -0.00000116, 0.00000690, -0.00000590, -0.00000894, 0.00000414, 0.00000223, -0.00000708, 0.00000051, 0.00000977, 0.00000415, -0.00000298, -0.00000269, -0.00000395, -0.00000636, 0.00000414, 0.00001687, 0.00000182, -0.00001746, 0.00000646, 0.00003015, -0.00000505, -0.00003394, 0.00000063, 0.00002103, -0.00001190, -0.00002299, 0.00000137, 0.00000632, 0.00000050, 0.00000480, -0.00000212, -0.00000757, 0.00000588, 0.00000051, -0.00002152, -0.00000584, 0.00002148, 0.00000015, -0.00002317, 0.00000567, 0.00002422, -0.00000532, -0.00001143, 0.00001744, 0.00000710, -0.00002315, -0.00000565, 0.00001542, -0.00000572, -0.00001269, 0.00000604, 0.00000131, -0.00000105, 0.00002110, 0.00001060, -0.00002354, -0.00001447, 0.00000867, -0.00000035, -0.00000581, 0.00000258, -0.00000253, -0.00000244, 0.00000955, 0.00000496, -0.00000239, 0.00000325, -0.00000370, -0.00001400, 0.00000124, 0.00001688, 0.00000841, -0.00000362, -0.00000493, -0.00000049, 0.00000206, -0.00000309, -0.00000357, 0.00000531, -0.00000030, -0.00001452, -0.00000749, 0.00000456, 0.00000233, 0.00000500, 0.00000663, -0.00000778, -0.00001046, 0.00000251, -0.00000024, -0.00000757, 0.00000051, 0.00000609, 0.00000487, 0.00000387, -0.00000447, -0.00000746, -0.00000028, -0.00001033, -0.00002266, -0.00000761, -0.00000171, -0.00001689, -0.00000922, -0.00000921, -0.00005734, -0.00006481, 0.00002265, 0.00009047, 0.00006587, },
		{ 0.00009212, 0.00006347, -0.00001086, -0.00005097, -0.00003877, -0.00001694, -0.00000969, -0.00000850, -0.00000803, -0.00000606, -0.00000067, -0.00000072, -0.00000369, 0.00000239, 0.00000356, -0.00000766, -0.00000561, 0.00000630, -0.00000101, -0.00001045, 0.00000026, 0.00000468, -0.00000571, -0.00000304, 0.00000844, 0.00000607, -0.00000142, -0.00000191, -0.00000371, -0.00000731, 0.00000081, 0.00001538, 0.00000829, -0.00001439, -0.00000432, 0.00002712, 0.00001029, -0.00003105, -0.00001429, 0.00002182, 0.00000053, -0.00002677, -0.00000660, 0.00000889, -0.00000006, 0.00000324, 0.00000399, -0.00000956, -0.00000217, 0.00000840, -0.00001339, -0.00001915, 0.00001409, 0.00001534, -0.00001918, -0.00000997, 0.00002366, 0.00000796, -0.00001482, 0.00000899, 0.00001606, -0.00001799, -0.00001653, 0.00001449, 0.00000416, -0.00001517, 0.00000053, 0.00000450, -0.00000472, 0.00001581, 0.00002098, -0.00001619, -0.00002318, 0.00000521, 0.00000417, -0.00000747, 0.00000217, 0.00000052, -0.00000633, 0.00000731, 0.00000919, -0.00000414, 0.00000149, 0.00000307, -0.00001414, -0.00000720, 0.00001692, 0.00001308, -0.00000351, -0.00000481, -0.00000130, 0.00000013, -0.00000031, -0.00000267, 0.00000175, 0.00000289, -0.00001023, -0.00001258, 0.00000127, 0.00000433, 0.00000329, 0.00000790, -0.00000222, -0.00001315, -0.00000137, 0.00000413, -0.00000702, -0.00000439, 0.00000558, 0.00000505, 0.00000468, 0.00000005, -0.00000900, -0.00000324, -0.00000271, -0.00002071, -0.00001662, 0.00000080, -0.00001184, -0.00001738, -0.00000118, -0.00003393, -0.00008057, -0.00002458, 0.00007870, 0.00008962, 0.00004519, 0.00005651, },
		{ 0.00001079, -0.00005118, -0.00004913, -0.00002092, -0.00000944, -0.00001076, -0.00000930, -0.00000540, -0.00000184, -0.00000123, -0.00000332, 0.00000046, 0.00000428, -0.00000429, -0.00000811, 0.00000322, 0.00000301, -0.00000912, -0.00000416, 0.00000520, -0.00000288, -0.00000576, 0.00000596, 0.00000764, -0.00000015, -0.00000106, -0.00000243, -0.00000807, -0.00000256, 0.00001314, 0.00001265, -0.00000893, -0.00001110, 0.00001887, 0.00002088, -0.00002060, -0.00002595, 0.00001436, 0.00001295, -0.00002356, -0.00001701, 0.00000854, 0.00000276, -0.00000066, 0.00000770, -0.00000503, -0.00001095, 0.00000758, -0.00000042, -0.00002344, -0.00000137, 0.00002296, -0.00000562, -0.00002088, 0.00001355, 0.00001966, -0.00001039, -0.00000143, 0.00001959, -0.00000773, -0.00002440, 0.00000720, 0.00001351, -0.00001236, -0.00000643, 0.00000619, -0.00000521, 0.00000746, 0.00002694, -0.00000380, -0.00002863, -0.00000231, 0.00000866, -0.00000748, -0.00000064, 0.00000490, -0.00000783, 0.00000163, 0.00001348, -0.00000244, -0.00000392, 0.00000796, -0.00000851, -0.00001589, 0.00001224, 0.00001961, -0.00000200, -0.00000662, -0.00000021, -0.00000175, -0.00000078, 0.00000059, -0.00000007, 0.00000207, -0.00000477, -0.00001376, -0.00000468, 0.00000511, 0.00000360, 0.00000667, 0.00000307, -0.00001181, -0.00000731, 0.00000581, -0.00000279, -0.00000864, 0.00000270, 0.00000602, 0.00000413, 0.00000371, -0.00000653, -0.00000820, 0.00000051, -0.00001292, -0.00002356, -0.00000370, -0.00000321, -0.00002164, -0.00000564, -0.00000959, -0.00007239, -0.00006912, 0.00004281, 0.00010588, 0.00006483, 0.00004814, 0.00009040, 0.00008689, },
		{ -0.00005989, -0.00002846, -0.00000910, -0.00001131, -0.00001203, -0.00000575, -0.00000159, -0.00000191, -0.00000344, -0.00000078, 0.00000381, -0.00000124, -0.00000828, -0.00000076, 0.00000476, -0.00000572, -0.00000742, 0.00000344, 0.00000035, -0.00000681, 0.00000246, 0.00000855, 0.00000130, -0.00000083, -0.00000050, -0.00000767, -0.00000617, 0.00001012, 0.00001545, -0.00000294, -0.00001363, 0.00000914, 0.00002468, -0.00000685, -0.00003031, 0.00000151, 0.00002021, -0.00001358, -0.00002555, 0.00000291, 0.00000758, -0.00000322, 0.00000652, 0.00000287, -0.00001454, -0.00000104, 0.00000956, -0.00001669, -0.00001695, 0.00001862, 0.00001102, -0.00002136, -0.00000206, 0.00002403, 0.00000056, -0.00000900, 0.00001637, 0.00000415, -0.00002606, -0.00000457, 0.00001865, -0.00000445, -0.00001232, 0.00000490, -0.00000266, -0.00000120, 0.00002645, 0.00001095, -0.00002798, -0.00001295, 0.00001100, -0.00000453, -0.00000555, 0.00000824, -0.00000503, -0.00000604, 0.00001451, 0.00000384, -0.00000984, 0.00000729, 0.00000148, -0.00002034, 0.00000182, 0.00002506, 0.00000352, -0.00001035, 0.00000042, -0.00000088, -0.00000419, 0.00000274, 0.00000208, -0.00000080, -0.00000183, -0.00001049, -0.00001030, 0.00000244, 0.00000565, 0.00000518, 0.00000566, -0.00000703, -0.00001195, 0.00000308, 0.00000294, -0.00000951, -0.00000256, 0.00000649, 0.00000399, 0.00000484, -0.00000142, -0.00001100, -0.00000209, -0.00000376, -0.00002368, -0.00001362, 0.00000261, -0.00001756, -0.00001690, 0.00000359, -0.00004547, -0.00009464, -0.00001138, 0.00010154, 0.00009291, 0.00004898, 0.00008077, 0.00010503, 0.00003900, -0.00004450, },
		{ -0.00001041, -0.00001014, -0.00001457, -0.00000792, -0.00000101, -0.00000168, -0.00000402, -0.00000183, 0.00000291, 0.00000079, -0.00000680, -0.00000402, 0.00000403, -0.00000189, -0.00000843, 0.00000017, 0.00000261, -0.00000589, -0.00000127, 0.00000819, 0.00000321, -0.00000109, 0.00000105, -0.00000563, -0.00000929, 0.00000588, 0.00001700, 0.00000282, -0.00001308, 0.00000077, 0.00002264, 0.00000545, -0.00002684, -0.00001157, 0.00001957, -0.00000055, -0.00002805, -0.00000718, 0.00001077, -0.00000179, 0.00000163, 0.00000904, -0.00001065, -0.00001191, 0.00001075, -0.00000311, -0.00002450, 0.00000418, 0.00002216, -0.00001075, -0.00001582, 0.00001831, 0.00001311, -0.00000975, 0.00000823, 0.00001321, -0.00002058, -0.00001687, 0.00001697, 0.00000624, -0.00001437, 0.00000035, 0.00000132, -0.00000718, 0.00001952, 0.00002391, -0.00001979, -0.00002389, 0.00000883, 0.00000146, -0.00001050, 0.00000784, 0.00000189, -0.00001218, 0.00000977, 0.00001266, -0.00001172, 0.00000009, 0.00001116, -0.00001689, -0.00001180, 0.00002497, 0.00001375, -0.00001320, -0.00000245, 0.00000311, -0.00000736, 0.00000060, 0.00000713, -0.00000172, -0.00000332, -0.00000541, -0.00001209, -0.00000365, 0.00000690, 0.00000579, 0.00000535, -0.00000166, -0.00001254, -0.00000290, 0.00000653, -0.00000589, -0.00000805, 0.00000451, 0.00000510, 0.00000406, 0.00000323, -0.00000925, -0.00000790, 0.00000134, -0.00001642, -0.00002311, 0.00000055, -0.00000723, -0.00002534, 0.00000091, -0.00001387, -0.00009169, -0.00006774, 0.00006840, 0.00011766, 0.00006462, 0.00006839, 0.00011339, 0.00007116, -0.00002879, -0.00006890, -0.00003947, },
		{ -0.00001577, -0.00001129, -0.00000149, -0.00000050, -0.00000428, -0.00000309, 0.00000187, 0.00000188, -0.00000472, -0.00000573, 0.00000174, 0.00000081, -0.00000723, -0.00000318, 0.00000301, -0.00000361, -0.00000411, 0.00000627, 0.00000526, -0.00000110, 0.00000140, -0.00000247, -0.00001071, 0.00000046, 0.00001693, 0.00000836, -0.00001068, -0.00000517, 0.00001739, 0.00001336, -0.00001826, -0.00002026, 0.00001210, 0.00001033, -0.00002315, -0.00001796, 0.00000903, 0.00000320, -0.00000317, 0.00000996, -0.00000170, -0.00001837, 0.00000270, 0.00000931, -0.00002034, -0.00001294, 0.00002171, 0.00000576, -0.00002069, 0.00000477, 0.00002090, -0.00000294, -0.00000040, 0.00001619, -0.00001005, -0.00002516, 0.00000837, 0.00001578, -0.00001087, -0.00000601, 0.00000422, -0.00000861, 0.00000857, 0.00003089, -0.00000520, -0.00003101, 0.00000097, 0.00000853, -0.00001246, 0.00000248, 0.00001004, -0.00001290, -0.00000027, 0.00001938, -0.00000645, -0.00001056, 0.00001483, -0.00000550, -0.00002292, 0.00001629, 0.00002574, -0.00001090, -0.00000981, 0.00000732, -0.00000639, -0.00000605, 0.00001095, 0.00000251, -0.00000753, -0.00000291, -0.00000864, -0.00001017, 0.00000436, 0.00000873, 0.00000445, 0.00000121, -0.00000901, -0.00000854, 0.00000536, 0.00000041, -0.00001045, -0.00000056, 0.00000631, 0.00000343, 0.00000511, -0.00000390, -0.00001224, -0.00000018, -0.00000604, -0.00002603, -0.00000912, 0.00000217, -0.00002409, -0.00001260, 0.00000702, -0.00006347, -0.00010620, 0.00000982, 0.00012294, 0.00009367, 0.00006233, 0.00011008, 0.00010207, -0.00000256, -0.00007339, -0.00005349, -0.00001445, -0.00000825, },
		{ -0.00000353, 0.00000066, -0.00000353, -0.00000435, 0.00000051, 0.00000229, -0.00000275, -0.00000597, -0.00000081, 0.00000172, -0.00000486, -0.00000531, 0.00000166, -0.00000116, -0.00000526, 0.00000321, 0.00000662, -0.00000023, 0.00000062, 0.00000051, -0.00000972, -0.00000515, 0.00001454, 0.00001347, -0.00000691, -0.00000895, 0.00001134, 0.00001660, -0.00000850, -0.00002266, 0.00000174, 0.00001528, -0.00001310, -0.00002489, 0.00000178, 0.00000849, -0.00000436, 0.00000576, 0.00000708, -0.00001679, -0.00000968, 0.00001358, -0.00000728, -0.00002388, 0.00000966, 0.00001949, -0.00001397, -0.00000996, 0.00001948, 0.00000803, -0.00000459, 0.00001285, 0.00000121, -0.00002643, -0.00000439, 0.00002017, -0.00000229, -0.00001128, 0.00000392, -0.00000574, -0.00000236, 0.00002955, 0.00001188, -0.00003054, -0.00001124, 0.00001313, -0.00000925, -0.00000625, 0.00001501, -0.00000650, -0.00001167, 0.00001920, 0.00000514, -0.00001860, 0.00000913, 0.00000919, -0.00002543, 0.00000008, 0.00003359, -0.00000069, -0.00001886, 0.00000710, 0.00000028, -0.00001351, 0.00000890, 0.00001140, -0.00000967, -0.00000604, -0.00000236, -0.00001267, -0.00000280, 0.00001132, 0.00000603, 0.00000078, -0.00000425, -0.00001026, -0.00000007, 0.00000534, -0.00000776, -0.00000681, 0.00000522, 0.00000441, 0.00000446, 0.00000165, -0.00001175, -0.00000631, 0.00000118, -0.00002029, -0.00002063, 0.00000368, -0.00001367, -0.00002581, 0.00000904, -0.00002484, -0.00011267, -0.00005781, 0.00009620, 0.00012489, 0.00007135, 0.00009870, 0.00012487, 0.00003342, -0.00006952, -0.00006962, -0.00002200, -0.00000660, -0.00001555, -0.00001466, },
		{ -0.00000204, -0.00000486, -0.00000120, 0.00000206, -0.00000116, -0.00000527, -0.00000273, 0.00000113, -0.00000261, -0.00000576, -0.00000055, 0.00000032, -0.00000472, 0.00000005, 0.00000656, 0.00000147, -0.00000045, 0.00000218, -0.00000671, -0.00000935, 0.00000968, 0.00001731, -0.00000173, -0.00001103, 0.00000562, 0.00001662, -0.00000052, -0.00001998, -0.00000717, 0.00001377, -0.00000234, -0.00002533, -0.00000831, 0.00001031, -0.00000113, -0.00000019, 0.00001112, -0.00000839, -0.00001931, 0.00000759, 0.00000707, -0.00002323, -0.00000778, 0.00002302, 0.00000093, -0.00001818, 0.00000905, 0.00001707, -0.00000179, 0.00000632, 0.00000879, -0.00002069, -0.00001661, 0.00001715, 0.00000843, -0.00001245, 0.00000005, -0.00000092, -0.00000942, 0.00002061, 0.00002600, -0.00002106, -0.00002380, 0.00001180, -0.00000122, -0.00001436, 0.00001324, 0.00000495, -0.00001866, 0.00001033, 0.00001800, -0.00001832, -0.00000427, 0.00001969, -0.00001660, -0.00001778, 0.00003148, 0.00001585, -0.00002378, -0.00000067, 0.00000988, -0.00001612, -0.00000060, 0.00002020, -0.00000495, -0.00001371, 0.00000144, -0.00000849, -0.00001168, 0.00000956, 0.00001108, -0.00000055, -0.00000209, -0.00000723, -0.00000602, 0.00000530, -0.00000125, -0.00001042, 0.00000041, 0.00000610, 0.00000363, 0.00000438, -0.00000676, -0.00001195, 0.00000134, -0.00000952, -0.00002653, -0.00000430, -0.00000140, -0.00002910, -0.00000447, 0.00000551, -0.00008694, -0.00011106, 0.00003703, 0.00013989, 0.00009747, 0.00008876, 0.00013446, 0.00007442, -0.00005312, -0.00008565, -0.00003414, -0.00000578, -0.00001431, -0.00001738, -0.00000653, 0.00000075, },
		{ -0.00000270, 0.00000106, -0.00000008, -0.00000409, -0.00000375, -0.00000019, -0.00000128, -0.00000497, -0.00000248, 0.00000043, -0.00000329, -0.00000216, 0.00000505, 0.00000325, -0.00000086, 0.00000224, -0.00000303, -0.00001096, 0.00000329, 0.00001862, 0.00000463, -0.00001130, 0.00000033, 0.00001503, 0.00000481, -0.00001501, -0.00001215, 0.00000829, 0.00000510, -0.00001999, -0.00001701, 0.00000686, 0.00000423, -0.00000384, 0.00000922, 0.00000196, -0.00002120, -0.00000494, 0.00001477, -0.00001200, -0.00002152, 0.00001426, 0.00001600, -0.00001536, -0.00000512, 0.00001868, 0.00000647, 0.00000143, 0.00001052, -0.00001105, -0.00002387, 0.00000758, 0.00001694, -0.00000803, -0.00000552, 0.00000263, -0.00001071, 0.00000778, 0.00003241, -0.00000475, -0.00003143, 0.00000335, 0.00000826, -0.00001741, 0.00000441, 0.00001615, -0.00001683, -0.00000432, 0.00002521, -0.00000792, -0.00001891, 0.00001961, 0.00000055, -0.00002901, 0.00001758, 0.00003238, -0.00001895, -0.00001454, 0.00001620, -0.00001000, -0.00001406, 0.00002229, 0.00000766, -0.00002044, -0.00000206, 0.00000066, -0.00001673, 0.00000128, 0.00001693, 0.00000127, -0.00000418, -0.00000237, -0.00000800, -0.00000003, 0.00000470, -0.00000836, -0.00000658, 0.00000567, 0.00000472, 0.00000406, -0.00000085, -0.00001278, -0.00000435, -0.00000062, -0.00002304, -0.00001658, 0.00000395, -0.00002064, -0.00002158, 0.00001524, -0.00004347, -0.00013064, -0.00003933, 0.00012173, 0.00013132, 0.00009157, 0.00013162, 0.00011227, -0.00002204, -0.00009697, -0.00005200, -0.00000675, -0.00001203, -0.00001965, -0.00000964, -0.00000034, -0.00000084, -0.00000425, },
		{ 0.00000018, -0.00000272, -0.00000391, -0.00000156, -0.00000099, -0.00000375, -0.00000350, -0.00000047, -0.00000197, -0.00000299, 0.00000279, 0.00000422, -0.00000029, 0.00000135, -0.00000009, -0.00000988, -0.00000288, 0.00001671, 0.00001108, -0.00000921, -0.00000460, 0.00001253, 0.00000819, -0.00001016, -0.00001339, 0.00000235, 0.00000768, -0.00001216, -0.00002109, -0.00000054, 0.00000802, -0.00000320, 0.00000382, 0.00000899, -0.00001502, -0.00001704, 0.00001200, 0.00000315, -0.00002471, -0.00000222, 0.00002297, -0.00000292, -0.00001541, 0.00001122, 0.00001510, 0.00000189, 0.00000773, -0.00000204, -0.00002413, -0.00000485, 0.00001950, 0.00000077, -0.00000956, 0.00000268, -0.00000724, -0.00000389, 0.00002930, 0.00001311, -0.00003010, -0.00000987, 0.00001430, -0.00001317, -0.00000795, 0.00002118, -0.00000596, -0.00001807, 0.00002186, 0.00000872, -0.00002661, 0.00000765, 0.00001794, -0.00002735, -0.00000380, 0.00004055, -0.00000303, -0.00002817, 0.00001320, 0.00000360, -0.00002415, 0.00001353, 0.00002355, -0.00001926, -0.00001337, 0.00000871, -0.00001330, -0.00001128, 0.00001835, 0.00000857, -0.00000817, -0.00000051, -0.00000411, -0.00000688, 0.00000533, -0.00000114, -0.00001146, 0.00000077, 0.00000722, 0.00000332, 0.00000222, -0.00000841, -0.00001061, 0.00000119, -0.00001281, -0.00002476, -0.00000135, -0.00000722, -0.00002994, 0.00000486, -0.00000378, -0.00011176, -0.00010699, 0.00006535, 0.00015351, 0.00011281, 0.00012476, 0.00013893, 0.00002114, -0.00009711, -0.00007514, -0.00001178, -0.00000815, -0.00002164, -0.00001287, -0.00000177, -0.00000073, -0.00000311, -0.00000315, -0.00000048, },
		{ -0.00000332, -0.00000259, -0.00000144, -0.00000278, -0.00000361, -0.00000164, -0.00000140, -0.00000274, 0.00000073, 0.00000397, 0.00000084, 0.00000051, 0.00000137, -0.00000717, -0.00000717, 0.00001202, 0.00001597, -0.00000459, -0.00000858, 0.00000897, 0.00001057, -0.00000624, -0.00001274, -0.00000170, 0.00000669, -0.00000545, -0.00002009, -0.00000850, 0.00000772, 0.00000063, -0.00000110, 0.00001019, -0.00000475, -0.00002261, 0.00000084, 0.00001386, -0.00001637, -0.00001780, 0.00001788, 0.00001239, -0.00001597, -0.00000175, 0.00001831, 0.00000789, 0.00000440, 0.00000295, -0.00001871, -0.00001537, 0.00001492, 0.00001039, -0.00000923, -0.00000065, -0.00000217, -0.00001047, 0.00001874, 0.00002631, -0.00001926, -0.00002261, 0.00001304, -0.00000321, -0.00001804, 0.00001683, 0.00000933, -0.00002384, 0.00000826, 0.00002364, -0.00002211, -0.00001102, 0.00002638, -0.00001258, -0.00002351, 0.00003469, 0.00001909, -0.00003325, -0.00000069, 0.00001799, -0.00002386, -0.00000431, 0.00003421, -0.00000642, -0.00002710, 0.00000850, -0.00000140, -0.00002165, 0.00001104, 0.00001928, -0.00000879, -0.00000472, 0.00000290, -0.00000951, -0.00000091, 0.00000662, -0.00000991, -0.00000771, 0.00000780, 0.00000508, 0.00000168, -0.00000255, -0.00001210, -0.00000386, -0.00000312, -0.00002332, -0.00001315, 0.00000100, -0.00002489, -0.00001379, 0.00001532, -0.00006711, -0.00014146, -0.00001647, 0.00014324, 0.00014590, 0.00012653, 0.00015175, 0.00006857, -0.00008057, -0.00009966, -0.00002419, -0.00000284, -0.00002242, -0.00001713, -0.00000280, -0.00000136, -0.00000274, -0.00000233, -0.00000166, -0.00000067, -0.00000160, },
		{ -0.00000231, -0.00000231, -0.00000317, -0.00000252, -0.00000161, -0.00000212, -0.00000048, 0.00000281, 0.00000173, 0.00000033, 0.00000163, -0.00000424, -0.00000890, 0.00000606, 0.00001794, 0.00000179, -0.00001051, 0.00000418, 0.00001213, -0.00000270, -0.00001192, -0.00000368, 0.00000469, -0.00000172, -0.00001609, -0.00001382, 0.00000347, 0.00000460, -0.00000271, 0.00000671, 0.00000421, -0.00001973, -0.00001237, 0.00001479, -0.00000173, -0.00002461, 0.00000333, 0.00002207, -0.00000625, -0.00001324, 0.00001307, 0.00001569, 0.00000457, 0.00000356, -0.00001131, -0.00002049, 0.00000523, 0.00001656, -0.00000390, -0.00000494, 0.00000112, -0.00001089, 0.00000558, 0.00003055, -0.00000253, -0.00002911, 0.00000412, 0.00000761, -0.00002093, 0.00000455, 0.00002172, -0.00001814, -0.00000970, 0.00002893, -0.00000639, -0.00002702, 0.00002067, 0.00000858, -0.00003206, 0.00001567, 0.00003786, -0.00002440, -0.00002050, 0.00002432, -0.00001117, -0.00002336, 0.00003202, 0.00001521, -0.00003421, -0.00000336, 0.00001300, -0.00002265, -0.00000451, 0.00002711, -0.00000154, -0.00001335, 0.00000696, -0.00000437, -0.00001043, 0.00000867, -0.00000117, -0.00001503, 0.00000286, 0.00000945, 0.00000050, 0.00000011, -0.00000765, -0.00001031, -0.00000054, -0.00001376, -0.00002237, -0.00000193, -0.00001207, -0.00002598, 0.00001062, -0.00002023, -0.00013251, -0.00009645, 0.00009046, 0.00017159, 0.00014566, 0.00015681, 0.00011046, -0.00004655, -0.00011809, -0.00004627, 0.00000143, -0.00001988, -0.00002306, -0.00000407, -0.00000131, -0.00000391, -0.00000124, -0.00000137, -0.00000228, -0.00000140, -0.00000212, -0.00000303, },
		{ -0.00000255, -0.00000292, -0.00000227, -0.00000173, -0.00000079, 0.00000145, 0.00000183, 0.00000070, 0.00000145, -0.00000207, -0.00000854, 0.00000062, 0.00001672, 0.00000827, -0.00000951, -0.00000124, 0.00001220, 0.00000120, -0.00001137, -0.00000492, 0.00000358, -0.00000040, -0.00001201, -0.00001537, -0.00000234, 0.00000600, -0.00000095, 0.00000200, 0.00000827, -0.00001107, -0.00002098, 0.00000623, 0.00001105, -0.00001980, -0.00001304, 0.00002066, 0.00000856, -0.00001667, 0.00000124, 0.00001977, 0.00000967, 0.00000250, -0.00000573, -0.00001976, -0.00000540, 0.00001659, 0.00000430, -0.00000699, 0.00000098, -0.00000715, -0.00000496, 0.00002546, 0.00001387, -0.00002609, -0.00000880, 0.00001373, -0.00001535, -0.00001009, 0.00002506, -0.00000322, -0.00002337, 0.00002124, 0.00001340, -0.00003173, 0.00000284, 0.00002551, -0.00002488, -0.00000863, 0.00004396, -0.00000305, -0.00003657, 0.00001686, 0.00000873, -0.00003356, 0.00001523, 0.00003643, -0.00002720, -0.00002322, 0.00002049, -0.00001119, -0.00002189, 0.00002506, 0.00001330, -0.00002008, 0.00000271, 0.00000669, -0.00001609, 0.00000153, 0.00001022, -0.00001518, -0.00000791, 0.00001263, 0.00000284, -0.00000198, -0.00000122, -0.00001165, -0.00000590, -0.00000371, -0.00002175, -0.00001314, -0.00000269, -0.00002402, -0.00000693, 0.00000759, -0.00008999, -0.00014523, 0.00000466, 0.00016595, 0.00017856, 0.00016670, 0.00014116, -0.00000081, -0.00012216, -0.00007638, -0.00000025, -0.00001247, -0.00002938, -0.00000761, 0.00000051, -0.00000580, -0.00000167, 0.00000064, -0.00000338, -0.00000270, -0.00000085, -0.00000257, -0.00000333, -0.00000240, },
		{ -0.00000304, -0.00000181, -0.00000058, 0.00000054, 0.00000118, 0.00000104, 0.00000148, -0.00000080, -0.00000720, -0.00000321, 0.00001319, 0.00001312, -0.00000570, -0.00000594, 0.00001009, 0.00000538, -0.00001029, -0.00000676, 0.00000348, 0.00000041, -0.00000955, -0.00001438, -0.00000696, 0.00000417, 0.00000203, -0.00000080, 0.00000729, -0.00000164, -0.00002163, -0.00000664, 0.00001552, -0.00000682, -0.00002290, 0.00000877, 0.00002031, -0.00000977, -0.00001116, 0.00001620, 0.00001727, 0.00000357, -0.00000350, -0.00001559, -0.00001292, 0.00001078, 0.00001154, -0.00000482, -0.00000161, -0.00000277, -0.00000982, 0.00001448, 0.00002405, -0.00001457, -0.00001979, 0.00001196, -0.00000441, -0.00002038, 0.00001762, 0.00001392, -0.00002595, 0.00000386, 0.00002753, -0.00002211, -0.00001852, 0.00002924, -0.00000534, -0.00002693, 0.00003360, 0.00002234, -0.00003962, -0.00000298, 0.00002558, -0.00002843, -0.00000989, 0.00004618, -0.00000527, -0.00004113, 0.00001380, 0.00000850, -0.00003149, 0.00001036, 0.00002942, -0.00001761, -0.00001014, 0.00001659, -0.00001177, -0.00001225, 0.00001631, -0.00000529, -0.00001972, 0.00000925, 0.00001002, -0.00000550, 0.00000117, -0.00000561, -0.00001343, -0.00000129, -0.00001152, -0.00002291, -0.00000495, -0.00001200, -0.00002057, 0.00000893, -0.00003869, -0.00014666, -0.00008627, 0.00011407, 0.00020657, 0.00019220, 0.00016414, 0.00004690, -0.00010741, -0.00010783, -0.00001300, -0.00000154, -0.00003263, -0.00001522, 0.00000322, -0.00000614, -0.00000476, 0.00000302, -0.00000236, -0.00000522, -0.00000058, -0.00000105, -0.00000408, -0.00000314, -0.00000196, -0.00000285, },
		{ -0.00000027, 0.00000029, 0.00000024, 0.00000087, 0.00000179, 0.00000004, -0.00000579, -0.00000532, 0.00000882, 0.00001540, -0.00000026, -0.00000847, 0.00000597, 0.00000880, -0.00000769, -0.00000932, 0.00000309, 0.00000219, -0.00000841, -0.00001308, -0.00000908, 0.00000061, 0.00000374, -0.00000072, 0.00000385, 0.00000435, -0.00001552, -0.00001712, 0.00001021, 0.00000686, -0.00002183, -0.00000736, 0.00002247, 0.00000396, -0.00001730, 0.00000538, 0.00002243, 0.00000892, -0.00000311, -0.00001148, -0.00001556, 0.00000215, 0.00001442, 0.00000096, -0.00000393, -0.00000055, -0.00000927, 0.00000290, 0.00002525, 0.00000062, -0.00002370, 0.00000332, 0.00000610, -0.00002209, 0.00000306, 0.00002515, -0.00001619, -0.00001482, 0.00002895, -0.00000256, -0.00003268, 0.00001747, 0.00001665, -0.00003051, 0.00001118, 0.00004034, -0.00002622, -0.00002653, 0.00002954, -0.00000935, -0.00003170, 0.00003785, 0.00002386, -0.00004587, -0.00000716, 0.00002607, -0.00002614, -0.00001239, 0.00003725, -0.00000285, -0.00002529, 0.00001699, 0.00000291, -0.00002440, 0.00001069, 0.00001111, -0.00002461, -0.00000319, 0.00001816, -0.00000449, -0.00000352, 0.00000360, -0.00001494, -0.00000876, 0.00000019, -0.00002190, -0.00001740, -0.00000221, -0.00001940, -0.00000670, -0.00000438, -0.00010700, -0.00014824, 0.00002174, 0.00020297, 0.00023211, 0.00019118, 0.00008929, -0.00007645, -0.00013115, -0.00003842, 0.00000774, -0.00002909, -0.00002629, 0.00000374, -0.00000288, -0.00000951, 0.00000320, 0.00000136, -0.00000735, -0.00000237, 0.00000105, -0.00000385, -0.00000453, -0.00000166, -0.00000229, -0.00000374, -0.00000237, },
		{ -0.00000047, 0.00000010, 0.00000200, 0.00000092, -0.00000461, -0.00000631, 0.00000477, 0.00001529, 0.00000520, -0.00000821, 0.00000096, 0.00001021, -0.00000346, -0.00001148, 0.00000108, 0.00000485, -0.00000697, -0.00001291, -0.00000945, -0.00000246, 0.00000315, 0.00000096, 0.00000098, 0.00000564, -0.00000691, -0.00002082, -0.00000112, 0.00001419, -0.00001143, -0.00001950, 0.00001382, 0.00001714, -0.00001351, -0.00000782, 0.00002071, 0.00001719, -0.00000130, -0.00000957, -0.00001445, -0.00000544, 0.00001191, 0.00000747, -0.00000347, -0.00000097, -0.00000602, -0.00000496, 0.00001881, 0.00001332, -0.00001892, -0.00000754, 0.00001105, -0.00001555, -0.00001178, 0.00002551, 0.00000102, -0.00002577, 0.00001716, 0.00001731, -0.00003257, -0.00000406, 0.00002975, -0.00001806, -0.00001260, 0.00004256, -0.00000131, -0.00004216, 0.00001697, 0.00001445, -0.00003926, 0.00001366, 0.00004723, -0.00003153, -0.00003382, 0.00003032, -0.00000618, -0.00003253, 0.00002969, 0.00001976, -0.00003280, 0.00000407, 0.00002099, -0.00002557, -0.00000654, 0.00002478, -0.00001650, -0.00002086, 0.00001982, 0.00000467, -0.00001203, 0.00000794, -0.00000629, -0.00002004, 0.00000304, -0.00000868, -0.00002947, -0.00000547, -0.00000581, -0.00002046, 0.00000079, -0.00005192, -0.00015860, -0.00008278, 0.00014778, 0.00026801, 0.00023443, 0.00012805, -0.00003747, -0.00013916, -0.00007219, 0.00000828, -0.00001788, -0.00003684, -0.00000140, 0.00000359, -0.00001275, -0.00000074, 0.00000628, -0.00000683, -0.00000625, 0.00000233, -0.00000198, -0.00000617, -0.00000216, -0.00000123, -0.00000412, -0.00000338, -0.00000017, 0.00000061, },
		{ 0.00000171, 0.00000189, -0.00000344, -0.00000681, 0.00000145, 0.00001364, 0.00000944, -0.00000560, -0.00000335, 0.00000907, 0.00000137, -0.00001181, -0.00000268, 0.00000694, -0.00000389, -0.00001353, -0.00000987, -0.00000386, 0.00000116, 0.00000222, 0.00000017, 0.00000383, -0.00000014, -0.00001753, -0.00001222, 0.00001222, 0.00000196, -0.00002202, -0.00000105, 0.00002266, -0.00000159, -0.00001666, 0.00001127, 0.00002409, 0.00000443, -0.00000925, -0.00001245, -0.00000948, 0.00000578, 0.00001133, 0.00000032, -0.00000229, -0.00000319, -0.00000764, 0.00000906, 0.00001914, -0.00000810, -0.00001502, 0.00000877, -0.00000517, -0.00002056, 0.00001555, 0.00001729, -0.00002409, -0.00000162, 0.00002791, -0.00001856, -0.00002460, 0.00002731, 0.00000344, -0.00002635, 0.00002847, 0.00002404, -0.00004156, -0.00000703, 0.00003061, -0.00002858, -0.00001575, 0.00005333, -0.00000170, -0.00005276, 0.00001590, 0.00001942, -0.00003881, 0.00000728, 0.00003952, -0.00002503, -0.00001796, 0.00003141, -0.00001146, -0.00002734, 0.00002566, 0.00000375, -0.00003427, 0.00000913, 0.00001919, -0.00001629, 0.00000173, 0.00000859, -0.00002453, -0.00000757, 0.00000811, -0.00002892, -0.00002227, 0.00000662, -0.00001765, -0.00001546, -0.00001217, -0.00011922, -0.00016013, 0.00004262, 0.00026973, 0.00029448, 0.00017477, 0.00000220, -0.00013151, -0.00010549, -0.00000459, -0.00000295, -0.00004133, -0.00001319, 0.00000971, -0.00001111, -0.00000824, 0.00000912, -0.00000231, -0.00001053, 0.00000110, 0.00000134, -0.00000699, -0.00000392, 0.00000001, -0.00000376, -0.00000478, -0.00000053, 0.00000128, -0.00000066, -0.00000098, },
		{ -0.00000202, -0.00000698, -0.00000131, 0.00001127, 0.00001210, -0.00000174, -0.00000585, 0.00000594, 0.00000524, -0.00000970, -0.00000700, 0.00000687, 0.00000069, -0.00001336, -0.00001157, -0.00000411, -0.00000050, 0.00000196, 0.00000088, 0.00000157, 0.00000271, -0.00001051, -0.00001807, 0.00000321, 0.00001118, -0.00001480, -0.00001447, 0.00001769, 0.00001218, -0.00001642, -0.00000225, 0.00002495, 0.00001359, -0.00000781, -0.00001187, -0.00001012, -0.00000065, 0.00001073, 0.00000560, -0.00000210, -0.00000229, -0.00000663, 0.00000063, 0.00001742, 0.00000352, -0.00001568, 0.00000166, 0.00000342, -0.00002072, 0.00000066, 0.00002530, -0.00001140, -0.00001794, 0.00002476, 0.00000192, -0.00003421, 0.00001085, 0.00002251, -0.00002407, 0.00000574, 0.00003844, -0.00002454, -0.00003106, 0.00003040, -0.00000510, -0.00003669, 0.00003856, 0.00003147, -0.00005279, -0.00001273, 0.00003708, -0.00002607, -0.00002076, 0.00004503, -0.00000217, -0.00003795, 0.00002549, 0.00001327, -0.00003934, 0.00000953, 0.00002700, -0.00003337, -0.00001288, 0.00003002, -0.00000904, -0.00001319, 0.00001935, -0.00001478, -0.00002595, 0.00001502, -0.00001146, -0.00004111, 0.00000333, 0.00000176, -0.00003176, -0.00000580, -0.00005657, -0.00017972, -0.00008391, 0.00021089, 0.00035241, 0.00024345, 0.00004402, -0.00011455, -0.00013017, -0.00002958, 0.00000820, -0.00003633, -0.00002833, 0.00001045, -0.00000369, -0.00001570, 0.00000674, 0.00000506, -0.00001225, -0.00000337, 0.00000460, -0.00000573, -0.00000678, 0.00000056, -0.00000226, -0.00000623, -0.00000167, 0.00000208, -0.00000024, -0.00000199, 0.00000084, 0.00000266, },
		{ -0.00000374, 0.00000856, 0.00001339, 0.00000233, -0.00000628, 0.00000208, 0.00000706, -0.00000577, -0.00001010, 0.00000414, 0.00000522, -0.00001095, -0.00001420, -0.00000484, -0.00000079, 0.00000070, 0.00000157, 0.00000057, 0.00000233, -0.00000389, -0.00001730, -0.00000742, 0.00001219, -0.00000277, -0.00002008, 0.00000510, 0.00002055, -0.00000723, -0.00001350, 0.00001782, 0.00002258, -0.00000259, -0.00001276, -0.00000955, -0.00000464, 0.00000643, 0.00000940, 0.00000077, -0.00000255, -0.00000467, -0.00000389, 0.00001095, 0.00001103, -0.00000995, -0.00000549, 0.00000660, -0.00001427, -0.00001226, 0.00002240, 0.00000543, -0.00002421, 0.00001063, 0.00001860, -0.00002902, -0.00001099, 0.00002937, -0.00000832, -0.00001392, 0.00003632, 0.00000086, -0.00004354, 0.00001357, 0.00001904, -0.00003964, 0.00000962, 0.00005323, -0.00003145, -0.00004272, 0.00003599, 0.00000086, -0.00004069, 0.00003107, 0.00002658, -0.00004379, 0.00000294, 0.00003640, -0.00003302, -0.00001808, 0.00003987, -0.00001437, -0.00003679, 0.00002707, 0.00001002, -0.00002719, 0.00001612, 0.00000711, -0.00003868, 0.00000334, 0.00001453, -0.00004568, -0.00001959, 0.00002268, -0.00002806, -0.00002782, -0.00000759, -0.00013732, -0.00018595, 0.00008733, 0.00037300, 0.00033708, 0.00010107, -0.00009494, -0.00014441, -0.00005973, 0.00000894, -0.00002360, -0.00004047, 0.00000280, 0.00000638, -0.00001839, -0.00000134, 0.00001157, -0.00000903, -0.00000969, 0.00000548, -0.00000192, -0.00000954, -0.00000065, 0.00000024, -0.00000692, -0.00000383, 0.00000261, 0.00000086, -0.00000267, -0.00000044, 0.00000291, -0.00000039, -0.00000665, },
		{ 0.00001362, 0.00000605, -0.00000505, -0.00000135, 0.00000668, -0.00000144, -0.00001084, -0.00000028, 0.00000784, -0.00000622, -0.00001614, -0.00000724, -0.00000017, -0.00000014, 0.00000124, 0.00000082, 0.00000083, -0.00000028, -0.00001209, -0.00001443, 0.00000575, 0.00000724, -0.00001622, -0.00000840, 0.00001939, 0.00000595, -0.00001702, 0.00000505, 0.00002661, 0.00000673, -0.00001286, -0.00001011, -0.00000566, 0.00000114, 0.00000955, 0.00000530, -0.00000200, -0.00000368, -0.00000478, 0.00000389, 0.00001235, -0.00000151, -0.00000857, 0.00000447, -0.00000597, -0.00001848, 0.00001138, 0.00001827, -0.00001854, -0.00000645, 0.00002401, -0.00001292, -0.00002737, 0.00002110, 0.00001150, -0.00002115, 0.00002073, 0.00002275, -0.00003895, -0.00001153, 0.00003159, -0.00002437, -0.00001992, 0.00005401, 0.00000303, -0.00005927, 0.00001447, 0.00002894, -0.00004172, 0.00000250, 0.00004722, -0.00002958, -0.00002675, 0.00004447, -0.00000809, -0.00004355, 0.00003244, 0.00001626, -0.00004892, 0.00000638, 0.00003200, -0.00002844, -0.00000292, 0.00002913, -0.00003326, -0.00002345, 0.00003197, -0.00002666, -0.00005054, 0.00002522, -0.00000007, -0.00005388, 0.00000131, -0.00005949, -0.00022221, -0.00007297, 0.00032069, 0.00043632, 0.00019144, -0.00007044, -0.00015389, -0.00008742, -0.00000180, -0.00000972, -0.00004425, -0.00001153, 0.00001326, -0.00001373, -0.00001182, 0.00001295, -0.00000111, -0.00001455, 0.00000229, 0.00000321, -0.00001032, -0.00000404, 0.00000271, -0.00000590, -0.00000678, 0.00000220, 0.00000260, -0.00000279, -0.00000192, 0.00000257, 0.00000112, -0.00000567, -0.00000573, 0.00000550, },
		{ -0.00000272, -0.00000370, 0.00000476, 0.00000195, -0.00000923, -0.00000464, 0.00000755, -0.00000056, -0.00001573, -0.00001108, -0.00000019, 0.00000038, 0.00000020, 0.00000121, -0.00000003, 0.00000027, -0.00000617, -0.00001554, -0.00000367, 0.00001046, -0.00000634, -0.00001618, 0.00000988, 0.00001616, -0.00001139, -0.00000767, 0.00002279, 0.00001724, -0.00000935, -0.00001230, -0.00000537, -0.00000238, 0.00000613, 0.00000895, 0.00000072, -0.00000362, -0.00000410, -0.00000073, 0.00000881, 0.00000514, -0.00000643, 0.00000022, -0.00000020, -0.00001749, -0.00000173, 0.00002199, -0.00000512, -0.00001782, 0.00001717, 0.00000510, -0.00003115, 0.00000279, 0.00002453, -0.00001391, 0.00000127, 0.00003192, -0.00002055, -0.00003266, 0.00002674, 0.00000016, -0.00003670, 0.00003441, 0.00003573, -0.00005356, -0.00001846, 0.00004355, -0.00002235, -0.00002750, 0.00004861, 0.00000008, -0.00004873, 0.00003072, 0.00002500, -0.00005216, 0.00000489, 0.00004373, -0.00003931, -0.00002547, 0.00004297, -0.00001119, -0.00002857, 0.00003627, -0.00000734, -0.00004992, 0.00002593, 0.00001004, -0.00006843, -0.00000062, 0.00003656, -0.00005637, -0.00003009, 0.00000924, -0.00017873, -0.00021667, 0.00018175, 0.00049953, 0.00032322, -0.00002465, -0.00016582, -0.00011113, -0.00001873, -0.00000177, -0.00003925, -0.00002657, 0.00001220, -0.00000363, -0.00001918, 0.00000726, 0.00000804, -0.00001454, -0.00000438, 0.00000699, -0.00000777, -0.00000864, 0.00000346, -0.00000278, -0.00000949, 0.00000012, 0.00000464, -0.00000200, -0.00000346, 0.00000174, 0.00000226, -0.00000421, -0.00000697, 0.00000217, 0.00001282, 0.00000925, },
		{ 0.00000212, 0.00000377, -0.00000623, -0.00000746, 0.00000482, 0.00000405, -0.00001249, -0.00001483, -0.00000213, 0.00000188, -0.00000036, 0.00000083, 0.00000008, -0.00000067, -0.00000236, -0.00001195, -0.00001098, 0.00000641, 0.00000340, -0.00001537, -0.00000249, 0.00001839, -0.00000010, -0.00001449, 0.00001217, 0.00002430, -0.00000122, -0.00001432, -0.00000609, -0.00000314, 0.00000134, 0.00000947, 0.00000513, -0.00000296, -0.00000375, -0.00000235, 0.00000371, 0.00000730, -0.00000116, -0.00000238, 0.00000146, -0.00001232, -0.00001129, 0.00001660, 0.00000865, -0.00001881, 0.00000344, 0.00001621, -0.00002232, -0.00001581, 0.00002445, 0.00000201, -0.00001155, 0.00002657, 0.00000190, -0.00004041, 0.00000787, 0.00002097, -0.00003455, 0.00000474, 0.00005275, -0.00002754, -0.00004750, 0.00003646, 0.00000823, -0.00004425, 0.00002899, 0.00003183, -0.00005085, -0.00000031, 0.00004991, -0.00003668, -0.00003106, 0.00005244, -0.00000871, -0.00005323, 0.00003238, 0.00001893, -0.00004495, 0.00002034, 0.00002776, -0.00005738, -0.00000543, 0.00004462, -0.00005624, -0.00004583, 0.00005551, -0.00002318, -0.00007543, 0.00003043, -0.00007939, -0.00028804, -0.00001985, 0.00047829, 0.00047911, 0.00006781, -0.00017802, -0.00013814, -0.00003497, -0.00000179, -0.00003074, -0.00003603, 0.00000311, 0.00000617, -0.00001931, -0.00000316, 0.00001338, -0.00000863, -0.00001141, 0.00000681, -0.00000236, -0.00001223, 0.00000122, 0.00000156, -0.00001039, -0.00000381, 0.00000603, 0.00000006, -0.00000480, 0.00000042, 0.00000300, -0.00000263, -0.00000731, -0.00000105, 0.00001088, 0.00001173, 0.00000038, -0.00000481, },
		{ -0.00000299, -0.00000823, 0.00000107, 0.00000627, -0.00000745, -0.00001667, -0.00000601, 0.00000292, 0.00000047, -0.00000011, 0.00000041, -0.00000145, -0.00000135, -0.00000683, -0.00001327, -0.00000151, 0.00000791, -0.00000799, -0.00001127, 0.00001226, 0.00001059, -0.00001270, -0.00000054, 0.00002424, 0.00000951, -0.00001328, -0.00000879, -0.00000235, -0.00000211, 0.00000643, 0.00000907, -0.00000029, -0.00000408, -0.00000239, -0.00000001, 0.00000522, 0.00000360, -0.00000144, 0.00000051, -0.00000700, -0.00001484, 0.00000631, 0.00001644, -0.00001067, -0.00000908, 0.00001641, -0.00000726, -0.00002596, 0.00001240, 0.00001676, -0.00001213, 0.00001255, 0.00001789, -0.00003293, -0.00001491, 0.00002813, -0.00001715, -0.00002077, 0.00004823, 0.00000705, -0.00005909, 0.00001040, 0.00003480, -0.00003949, -0.00000252, 0.00005046, -0.00003069, -0.00003445, 0.00005321, -0.00000223, -0.00005759, 0.00003509, 0.00003006, -0.00006086, 0.00000066, 0.00004641, -0.00003906, -0.00001332, 0.00005244, -0.00003596, -0.00004744, 0.00005453, -0.00001360, -0.00008545, 0.00003721, 0.00003267, -0.00009761, -0.00000625, 0.00002372, -0.00025751, -0.00022436, 0.00034333, 0.00061185, 0.00022492, -0.00017066, -0.00017914, -0.00004994, -0.00000468, -0.00002554, -0.00003823, -0.00000892, 0.00000991, -0.00001264, -0.00001273, 0.00001167, 0.00000046, -0.00001472, 0.00000213, 0.00000340, -0.00001258, -0.00000371, 0.00000505, -0.00000832, -0.00000863, 0.00000549, 0.00000331, -0.00000538, -0.00000149, 0.00000341, -0.00000113, -0.00000698, -0.00000360, 0.00000795, 0.00001306, 0.00000401, -0.00000464, -0.00000061, 0.00000403, },
		{ -0.00000226, 0.00000603, -0.00000236, -0.00001579, -0.00001054, 0.00000212, 0.00000246, -0.00000061, 0.00000019, -0.00000148, -0.00000207, -0.00000307, -0.00001086, -0.00000856, 0.00000570, 0.00000073, -0.00001274, 0.00000191, 0.00001521, -0.00000423, -0.00000949, 0.00001682, 0.00001854, -0.00000748, -0.00001216, -0.00000230, -0.00000279, 0.00000160, 0.00001018, 0.00000423, -0.00000385, -0.00000259, -0.00000135, 0.00000139, 0.00000514, 0.00000212, -0.00000018, -0.00000393, -0.00001353, -0.00000349, 0.00001602, 0.00000097, -0.00001424, 0.00000802, 0.00000553, -0.00002446, -0.00000434, 0.00002227, -0.00000232, -0.00000067, 0.00002186, -0.00001608, -0.00003071, 0.00001967, 0.00000472, -0.00003146, 0.00002692, 0.00003506, -0.00004840, -0.00002237, 0.00004418, -0.00001600, -0.00003067, 0.00004715, 0.00000275, -0.00005526, 0.00003192, 0.00003562, -0.00006020, -0.00000233, 0.00005798, -0.00004107, -0.00003907, 0.00005422, -0.00000973, -0.00004733, 0.00005028, 0.00000770, -0.00007592, 0.00002885, 0.00004020, -0.00009285, -0.00001661, 0.00007926, -0.00007199, -0.00007622, 0.00007632, -0.00014012, -0.00035874, 0.00010764, 0.00065758, 0.00043724, -0.00010876, -0.00023556, -0.00007424, -0.00000259, -0.00002558, -0.00003785, -0.00001742, 0.00000627, -0.00000433, -0.00001636, 0.00000404, 0.00000760, -0.00001218, -0.00000460, 0.00000632, -0.00000912, -0.00000925, 0.00000559, -0.00000356, -0.00001236, 0.00000217, 0.00000683, -0.00000437, -0.00000396, 0.00000335, 0.00000040, -0.00000639, -0.00000526, 0.00000465, 0.00001282, 0.00000769, -0.00000304, -0.00000300, 0.00000302, -0.00000030, -0.00000732, },
		{ 0.00000134, -0.00001268, -0.00001400, -0.00000084, 0.00000434, 0.00000020, -0.00000050, -0.00000118, -0.00000303, -0.00000183, -0.00000634, -0.00001147, -0.00000085, 0.00000567, -0.00000765, -0.00000675, 0.00001211, 0.00000562, -0.00001101, 0.00000559, 0.00002176, 0.00000201, -0.00001342, -0.00000444, -0.00000147, -0.00000232, 0.00000762, 0.00000873, -0.00000162, -0.00000344, -0.00000131, -0.00000152, 0.00000315, 0.00000553, 0.00000137, -0.00000288, -0.00001041, -0.00000931, 0.00000956, 0.00000981, -0.00001085, -0.00000250, 0.00001031, -0.00001463, -0.00001714, 0.00001649, 0.00001046, -0.00000575, 0.00001559, 0.00000068, -0.00003372, 0.00000168, 0.00001954, -0.00002534, 0.00000080, 0.00004570, -0.00002144, -0.00004667, 0.00003218, 0.00001401, -0.00004217, 0.00002426, 0.00003374, -0.00005278, -0.00000449, 0.00005876, -0.00003594, -0.00004270, 0.00005991, -0.00000059, -0.00006709, 0.00003414, 0.00003008, -0.00006168, 0.00001870, 0.00005238, -0.00007120, -0.00002311, 0.00007567, -0.00005680, -0.00008085, 0.00008472, -0.00000307, -0.00013324, 0.00005106, 0.00000797, -0.00037061, -0.00016790, 0.00056467, 0.00065673, 0.00004103, -0.00028725, -0.00012546, 0.00000565, -0.00002420, -0.00004220, -0.00002028, -0.00000017, -0.00000057, -0.00001338, -0.00000428, 0.00000866, -0.00000549, -0.00000922, 0.00000487, -0.00000373, -0.00001263, 0.00000250, 0.00000191, -0.00001304, -0.00000346, 0.00000891, -0.00000128, -0.00000636, 0.00000233, 0.00000216, -0.00000571, -0.00000631, 0.00000177, 0.00001105, 0.00001053, -0.00000002, -0.00000453, 0.00000101, 0.00000145, -0.00000541, -0.00000441, 0.00000420, },
		{ -0.00001534, -0.00000501, 0.00000474, 0.00000224, -0.00000086, -0.00000116, -0.00000345, -0.00000253, -0.00000253, -0.00000982, -0.00000750, 0.00000463, -0.00000023, -0.00000952, 0.00000407, 0.00001136, -0.00000558, -0.00000398, 0.00001777, 0.00001151, -0.00001031, -0.00000808, -0.00000034, -0.00000350, 0.00000271, 0.00001076, 0.00000277, -0.00000389, -0.00000146, -0.00000236, -0.00000047, 0.00000633, 0.00000483, -0.00000191, -0.00000794, -0.00001081, 0.00000146, 0.00001229, -0.00000244, -0.00000881, 0.00000683, -0.00000349, -0.00002086, 0.00000372, 0.00001814, -0.00000133, 0.00000601, 0.00001001, -0.00002550, -0.00001601, 0.00002119, -0.00000896, -0.00001772, 0.00003760, 0.00000870, -0.00005234, 0.00000539, 0.00003565, -0.00003271, -0.00000617, 0.00004821, -0.00002881, -0.00003902, 0.00005615, 0.00000460, -0.00006646, 0.00003320, 0.00004223, -0.00006771, -0.00000727, 0.00005945, -0.00004559, -0.00002802, 0.00007365, -0.00003067, -0.00007514, 0.00007013, 0.00000944, -0.00011986, 0.00003772, 0.00007591, -0.00013599, -0.00003674, 0.00011516, -0.00025650, -0.00039351, 0.00032501, 0.00080530, 0.00028555, -0.00029061, -0.00021414, 0.00000679, -0.00000983, -0.00005360, -0.00002288, -0.00000242, -0.00000328, -0.00000860, -0.00000766, 0.00000386, 0.00000042, -0.00000884, 0.00000062, 0.00000037, -0.00001235, -0.00000275, 0.00000545, -0.00001003, -0.00000945, 0.00000796, 0.00000332, -0.00000751, -0.00000003, 0.00000396, -0.00000461, -0.00000728, -0.00000033, 0.00000849, 0.00001170, 0.00000389, -0.00000450, -0.00000166, 0.00000200, -0.00000307, -0.00000526, 0.00000187, 0.00000323, -0.00000870, },
		{ 0.00000317, 0.00000446, -0.00000011, -0.00000143, -0.00000344, -0.00000390, -0.00000095, -0.00000558, -0.00001078, -0.00000094, 0.00000463, -0.00000613, -0.00000376, 0.00001043, 0.00000264, -0.00000766, 0.00000890, 0.00001679, -0.00000299, -0.00001073, -0.00000126, -0.00000218, -0.00000190, 0.00000903, 0.00000770, -0.00000242, -0.00000244, -0.00000187, -0.00000332, 0.00000398, 0.00000809, 0.00000065, -0.00000659, -0.00000992, -0.00000455, 0.00000886, 0.00000548, -0.00000814, -0.00000048, 0.00000279, -0.00001617, -0.00000864, 0.00001669, 0.00000801, 0.00000041, 0.00001037, -0.00001285, -0.00002566, 0.00001125, 0.00000727, -0.00002219, 0.00001827, 0.00002918, -0.00003901, -0.00002292, 0.00003923, -0.00000873, -0.00002918, 0.00004111, 0.00000444, -0.00005613, 0.00002949, 0.00004269, -0.00006193, -0.00001030, 0.00006674, -0.00003841, -0.00005104, 0.00006116, -0.00000473, -0.00006573, 0.00005795, 0.00002809, -0.00009804, 0.00002198, 0.00007342, -0.00010802, -0.00004319, 0.00012096, -0.00007280, -0.00013638, 0.00013008, -0.00006801, -0.00048826, -0.00000992, 0.00080436, 0.00058559, -0.00019362, -0.00032657, -0.00002533, 0.00002320, -0.00006347, -0.00003491, 0.00000224, -0.00000750, -0.00000884, -0.00000481, -0.00000172, 0.00000091, -0.00000449, -0.00000247, 0.00000100, -0.00000946, -0.00000735, 0.00000547, -0.00000475, -0.00001325, 0.00000364, 0.00000771, -0.00000629, -0.00000341, 0.00000504, -0.00000256, -0.00000838, -0.00000199, 0.00000618, 0.00001100, 0.00000751, -0.00000245, -0.00000422, 0.00000108, -0.00000079, -0.00000497, -0.00000028, 0.00000369, -0.00000502, -0.00001469, -0.00000894, },
		{ 0.00000174, -0.00000139, -0.00000336, -0.00000507, -0.00000154, -0.00000142, -0.00000969, -0.00000736, 0.00000426, 0.00000017, -0.00000704, 0.00000439, 0.00000848, -0.00000474, -0.00000007, 0.00001558, 0.00000570, -0.00000984, -0.00000421, -0.00000030, -0.00000401, 0.00000440, 0.00001068, 0.00000147, -0.00000334, -0.00000158, -0.00000417, -0.00000000, 0.00000896, 0.00000475, -0.00000514, -0.00000875, -0.00000721, 0.00000276, 0.00000887, -0.00000228, -0.00000590, 0.00000226, -0.00000818, -0.00001484, 0.00000786, 0.00001534, 0.00000202, 0.00000580, -0.00000298, -0.00002535, -0.00000321, 0.00001517, -0.00001438, -0.00000097, 0.00003372, -0.00001521, -0.00004016, 0.00002484, 0.00001676, -0.00003490, 0.00001834, 0.00003128, -0.00004964, -0.00000803, 0.00006119, -0.00003138, -0.00005024, 0.00006089, 0.00000804, -0.00007553, 0.00003179, 0.00004101, -0.00007374, 0.00001116, 0.00007595, -0.00007635, -0.00004651, 0.00010079, -0.00004589, -0.00011781, 0.00010289, 0.00002990, -0.00018897, 0.00004829, 0.00010957, -0.00041871, -0.00034040, 0.00061383, 0.00085618, 0.00003543, -0.00041640, -0.00011551, 0.00006301, -0.00005432, -0.00006160, 0.00000727, -0.00000431, -0.00001678, -0.00000090, -0.00000171, -0.00000427, -0.00000105, -0.00000113, -0.00000081, -0.00000696, -0.00000924, 0.00000265, 0.00000010, -0.00001338, -0.00000256, 0.00000974, -0.00000256, -0.00000660, 0.00000445, 0.00000048, -0.00000909, -0.00000399, 0.00000469, 0.00000919, 0.00000954, 0.00000128, -0.00000549, -0.00000133, 0.00000083, -0.00000366, -0.00000198, 0.00000326, -0.00000212, -0.00001288, -0.00001167, 0.00000022, 0.00000568, },
		{ -0.00000322, -0.00000580, -0.00000333, 0.00000085, -0.00000558, -0.00001115, -0.00000070, 0.00000489, -0.00000472, -0.00000247, 0.00000882, 0.00000180, -0.00000470, 0.00000909, 0.00001169, -0.00000475, -0.00000719, 0.00000000, -0.00000327, -0.00000069, 0.00001001, 0.00000648, -0.00000257, -0.00000227, -0.00000356, -0.00000337, 0.00000676, 0.00000868, -0.00000227, -0.00000807, -0.00000751, -0.00000256, 0.00000699, 0.00000438, -0.00000598, -0.00000246, -0.00000261, -0.00001371, -0.00000268, 0.00001592, 0.00000865, 0.00000234, 0.00000072, -0.00001879, -0.00001455, 0.00001269, -0.00000183, -0.00001149, 0.00002477, 0.00000715, -0.00004077, 0.00000129, 0.00003154, -0.00002312, -0.00000724, 0.00004082, -0.00002513, -0.00003911, 0.00005318, 0.00001048, -0.00006828, 0.00002765, 0.00004998, -0.00006819, -0.00001572, 0.00006807, -0.00004665, -0.00004391, 0.00008803, -0.00001793, -0.00010041, 0.00007483, 0.00003818, -0.00014591, 0.00002545, 0.00012092, -0.00015976, -0.00008493, 0.00019708, -0.00021832, -0.00055626, 0.00026097, 0.00099291, 0.00037902, -0.00041487, -0.00026532, 0.00007713, -0.00001131, -0.00009471, -0.00000129, 0.00001138, -0.00002667, -0.00000448, 0.00000593, -0.00000962, -0.00000387, 0.00000394, -0.00000116, -0.00000723, -0.00000894, -0.00000049, 0.00000238, -0.00001047, -0.00000799, 0.00000825, 0.00000233, -0.00000796, 0.00000192, 0.00000372, -0.00000846, -0.00000676, 0.00000366, 0.00000757, 0.00000941, 0.00000536, -0.00000444, -0.00000452, 0.00000096, -0.00000152, -0.00000296, 0.00000222, 0.00000010, -0.00001064, -0.00001316, -0.00000303, 0.00000540, 0.00000385, -0.00000046, },
		{ -0.00000545, 0.00000081, -0.00000083, -0.00001077, -0.00000716, 0.00000507, 0.00000085, -0.00000593, 0.00000419, 0.00000730, -0.00000340, 0.00000123, 0.00001221, 0.00000254, -0.00000760, -0.00000186, -0.00000125, -0.00000375, 0.00000599, 0.00001016, 0.00000071, -0.00000322, -0.00000299, -0.00000482, 0.00000273, 0.00001040, 0.00000211, -0.00000712, -0.00000716, -0.00000532, 0.00000205, 0.00000778, -0.00000121, -0.00000658, -0.00000228, -0.00000861, -0.00000936, 0.00000976, 0.00001495, 0.00000379, -0.00000011, -0.00001179, -0.00001892, 0.00000367, 0.00000740, -0.00001121, 0.00001056, 0.00001932, -0.00002796, -0.00001960, 0.00003039, -0.00000231, -0.00002326, 0.00003200, 0.00000408, -0.00005131, 0.00002468, 0.00004450, -0.00005736, -0.00001680, 0.00006817, -0.00003230, -0.00005869, 0.00006214, 0.00000242, -0.00007963, 0.00005756, 0.00004953, -0.00011096, 0.00000663, 0.00010244, -0.00010997, -0.00007500, 0.00015224, -0.00005777, -0.00019803, 0.00015964, 0.00002460, -0.00058470, -0.00015862, 0.00091624, 0.00075905, -0.00026055, -0.00043626, 0.00002466, 0.00006183, -0.00011133, -0.00003515, 0.00003334, -0.00002619, -0.00002039, 0.00001539, -0.00000706, -0.00001384, 0.00000730, 0.00000380, -0.00000949, -0.00000931, -0.00000171, 0.00000220, -0.00000695, -0.00001061, 0.00000407, 0.00000611, -0.00000666, -0.00000167, 0.00000578, -0.00000592, -0.00000979, 0.00000199, 0.00000706, 0.00000780, 0.00000807, -0.00000097, -0.00000698, -0.00000098, 0.00000073, -0.00000277, 0.00000069, 0.00000170, -0.00000814, -0.00001386, -0.00000595, 0.00000403, 0.00000531, 0.00000126, -0.00000258, -0.00000614, },
		{ 0.00000257, -0.00000697, -0.00001173, 0.00000067, 0.00000564, -0.00000435, -0.00000196, 0.00000839, 0.00000190, -0.00000363, 0.00000746, 0.00000838, -0.00000418, -0.00000447, -0.00000014, -0.00000387, 0.00000080, 0.00001047, 0.00000559, -0.00000280, -0.00000339, -0.00000469, -0.00000103, 0.00000914, 0.00000655, -0.00000483, -0.00000701, -0.00000586, -0.00000286, 0.00000649, 0.00000497, -0.00000648, -0.00000588, -0.00000447, -0.00001003, 0.00000093, 0.00001623, 0.00000954, -0.00000104, -0.00000821, -0.00001747, -0.00000572, 0.00000924, -0.00000425, -0.00000038, 0.00001959, -0.00001054, -0.00002950, 0.00001675, 0.00001605, -0.00002416, 0.00001276, 0.00002464, -0.00004264, -0.00000960, 0.00005692, -0.00002454, -0.00005180, 0.00005556, 0.00001486, -0.00007677, 0.00002596, 0.00004888, -0.00007855, -0.00000037, 0.00009317, -0.00007143, -0.00007013, 0.00011425, -0.00002566, -0.00014832, 0.00010568, 0.00006862, -0.00023048, 0.00002314, 0.00020607, -0.00042506, -0.00051315, 0.00061362, 0.00105478, 0.00006320, -0.00055163, -0.00011877, 0.00013327, -0.00008452, -0.00009131, 0.00004335, -0.00000502, -0.00004326, 0.00001510, 0.00000745, -0.00002470, 0.00000200, 0.00001376, -0.00000926, -0.00001316, -0.00000136, 0.00000192, -0.00000512, -0.00001048, -0.00000043, 0.00000710, -0.00000341, -0.00000460, 0.00000572, -0.00000196, -0.00001176, -0.00000117, 0.00000741, 0.00000624, 0.00000834, 0.00000363, -0.00000710, -0.00000461, 0.00000187, -0.00000114, -0.00000092, 0.00000242, -0.00000523, -0.00001399, -0.00000862, 0.00000229, 0.00000580, 0.00000305, -0.00000113, -0.00000592, -0.00000738, -0.00000101, },
		{ -0.00001249, -0.00000586, 0.00000657, 0.00000067, -0.00000567, 0.00000478, 0.00000720, -0.00000327, 0.00000054, 0.00000971, 0.00000184, -0.00000537, -0.00000105, -0.00000216, -0.00000297, 0.00000719, 0.00000972, 0.00000010, -0.00000426, -0.00000440, -0.00000312, 0.00000587, 0.00000920, -0.00000103, -0.00000664, -0.00000549, -0.00000573, 0.00000187, 0.00000870, -0.00000182, -0.00000948, -0.00000442, -0.00000671, -0.00000555, 0.00001152, 0.00001557, 0.00000154, -0.00000789, -0.00001427, -0.00001118, 0.00000486, 0.00000288, -0.00000385, 0.00001254, 0.00000275, -0.00002721, -0.00000059, 0.00002393, -0.00001292, -0.00000548, 0.00002984, -0.00002102, -0.00003454, 0.00004551, 0.00001373, -0.00006279, 0.00002019, 0.00005139, -0.00006253, -0.00002254, 0.00007005, -0.00004243, -0.00005708, 0.00009248, -0.00000075, -0.00011713, 0.00006776, 0.00006600, -0.00015750, 0.00000332, 0.00015746, -0.00016403, -0.00013987, 0.00025411, -0.00015151, -0.00068774, 0.00015700, 0.00114755, 0.00050286, -0.00052272, -0.00033530, 0.00014989, -0.00000276, -0.00014576, 0.00002046, 0.00003462, -0.00005723, -0.00000431, 0.00002842, -0.00002556, -0.00001458, 0.00002242, -0.00000132, -0.00001988, -0.00000276, 0.00000401, -0.00000504, -0.00000982, -0.00000301, 0.00000557, -0.00000028, -0.00000555, 0.00000384, 0.00000197, -0.00001150, -0.00000558, 0.00000731, 0.00000612, 0.00000654, 0.00000729, -0.00000419, -0.00000836, 0.00000067, 0.00000149, -0.00000172, 0.00000188, -0.00000207, -0.00001319, -0.00001141, 0.00000039, 0.00000579, 0.00000421, 0.00000082, -0.00000474, -0.00000879, -0.00000390, 0.00000373, -0.00000173, },
		{ 0.00000303, 0.00000567, -0.00000494, -0.00000089, 0.00000902, 0.00000127, -0.00000445, 0.00000599, 0.00000730, -0.00000293, -0.00000311, -0.00000072, -0.00000391, 0.00000211, 0.00001081, 0.00000491, -0.00000398, -0.00000500, -0.00000372, 0.00000249, 0.00000918, 0.00000311, -0.00000520, -0.00000511, -0.00000636, -0.00000336, 0.00000805, 0.00000476, -0.00000951, -0.00000804, -0.00000355, -0.00000705, 0.00000370, 0.00001759, 0.00000782, -0.00000769, -0.00001278, -0.00001229, -0.00000144, 0.00000562, -0.00000117, 0.00000506, 0.00000785, -0.00001776, -0.00001316, 0.00002026, 0.00000208, -0.00001440, 0.00002182, 0.00000140, -0.00004205, 0.00001907, 0.00004064, -0.00004790, -0.00001999, 0.00006212, -0.00002457, -0.00006015, 0.00005699, 0.00000945, -0.00008582, 0.00004965, 0.00006684, -0.00011165, -0.00001310, 0.00012053, -0.00009841, -0.00010427, 0.00016613, -0.00003112, -0.00024751, 0.00016091, 0.00011900, -0.00063410, -0.00031713, 0.00097398, 0.00093906, -0.00029163, -0.00055493, 0.00006204, 0.00011296, -0.00016080, -0.00004335, 0.00007330, -0.00004442, -0.00004086, 0.00004113, -0.00000828, -0.00003632, 0.00002016, 0.00001541, -0.00002411, -0.00000982, 0.00000825, -0.00000406, -0.00001110, -0.00000354, 0.00000368, 0.00000092, -0.00000474, 0.00000162, 0.00000445, -0.00000895, -0.00000988, 0.00000538, 0.00000752, 0.00000438, 0.00000836, 0.00000081, -0.00001011, -0.00000305, 0.00000372, -0.00000080, 0.00000029, 0.00000052, -0.00001088, -0.00001418, -0.00000215, 0.00000586, 0.00000474, 0.00000242, -0.00000250, -0.00000914, -0.00000722, 0.00000263, 0.00000297, -0.00000956, -0.00001134, },
		{ -0.00000066, -0.00000502, 0.00000648, 0.00000664, -0.00000490, -0.00000045, 0.00000908, 0.00000226, -0.00000413, -0.00000102, -0.00000265, -0.00000213, 0.00000819, 0.00000947, -0.00000123, -0.00000613, -0.00000415, 0.00000040, 0.00000722, 0.00000594, -0.00000248, -0.00000462, -0.00000566, -0.00000702, 0.00000367, 0.00000961, -0.00000499, -0.00001213, -0.00000377, -0.00000460, -0.00000280, 0.00001404, 0.00001472, -0.00000443, -0.00001349, -0.00001172, -0.00000577, 0.00000353, 0.00000309, 0.00000179, 0.00000637, -0.00000818, -0.00001721, 0.00001001, 0.00001266, -0.00001243, 0.00000855, 0.00001514, -0.00003361, -0.00000861, 0.00004719, -0.00001728, -0.00004695, 0.00004551, 0.00001799, -0.00007063, 0.00001827, 0.00005134, -0.00007537, -0.00001263, 0.00010007, -0.00005793, -0.00008788, 0.00011328, -0.00000131, -0.00016499, 0.00009319, 0.00010344, -0.00024937, -0.00001674, 0.00027899, -0.00039210, -0.00065939, 0.00056194, 0.00122489, 0.00012950, -0.00067304, -0.00014346, 0.00020722, -0.00010556, -0.00013063, 0.00008118, 0.00000177, -0.00007757, 0.00002960, 0.00002460, -0.00004928, 0.00000067, 0.00003448, -0.00001821, -0.00002303, 0.00001049, 0.00000125, -0.00001447, -0.00000445, 0.00000381, 0.00000023, -0.00000399, 0.00000049, 0.00000527, -0.00000532, -0.00001242, 0.00000135, 0.00000913, 0.00000367, 0.00000674, 0.00000563, -0.00000847, -0.00000796, 0.00000376, 0.00000193, -0.00000123, 0.00000153, -0.00000713, -0.00001599, -0.00000598, 0.00000587, 0.00000542, 0.00000304, 0.00000015, -0.00000787, -0.00001025, -0.00000036, 0.00000587, -0.00000439, -0.00001361, -0.00000317, 0.00000756, },
		{ 0.00000144, 0.00000946, -0.00000105, -0.00000588, 0.00000605, 0.00000758, -0.00000227, -0.00000268, -0.00000113, -0.00000375, 0.00000326, 0.00001131, 0.00000374, -0.00000620, -0.00000549, -0.00000050, 0.00000501, 0.00000663, 0.00000059, -0.00000344, -0.00000452, -0.00000837, -0.00000214, 0.00001037, 0.00000213, -0.00001318, -0.00000765, -0.00000169, -0.00000504, 0.00000691, 0.00001828, 0.00000250, -0.00001406, -0.00001232, -0.00000712, -0.00000058, 0.00000485, 0.00000304, 0.00000306, -0.00000305, -0.00001461, 0.00000002, 0.00001517, -0.00000407, -0.00000168, 0.00001756, -0.00001751, -0.00002638, 0.00003514, 0.00001354, -0.00005138, 0.00001286, 0.00004609, -0.00005232, -0.00002587, 0.00006474, -0.00003460, -0.00006412, 0.00008638, 0.00001635, -0.00012115, 0.00005141, 0.00008599, -0.00015217, -0.00002242, 0.00017693, -0.00014954, -0.00018813, 0.00027722, -0.00007184, -0.00076842, 0.00002897, 0.00124485, 0.00064504, -0.00059274, -0.00042029, 0.00021028, 0.00001977, -0.00019938, 0.00003520, 0.00006661, -0.00008895, -0.00001164, 0.00005722, -0.00003932, -0.00003233, 0.00004331, 0.00000209, -0.00003666, 0.00000467, 0.00001149, -0.00001653, -0.00000901, 0.00000619, 0.00000002, -0.00000522, 0.00000041, 0.00000561, -0.00000212, -0.00001261, -0.00000359, 0.00000919, 0.00000502, 0.00000404, 0.00000810, -0.00000382, -0.00001165, 0.00000069, 0.00000519, -0.00000121, 0.00000064, -0.00000302, -0.00001557, -0.00001093, 0.00000477, 0.00000698, 0.00000298, 0.00000204, -0.00000492, -0.00001200, -0.00000472, 0.00000639, 0.00000124, -0.00001223, -0.00000923, 0.00000531, 0.00000429, -0.00000538, },
		{ 0.00000439, -0.00000740, -0.00000017, 0.00000991, 0.00000245, -0.00000381, -0.00000103, -0.00000291, -0.00000138, 0.00000932, 0.00000893, -0.00000368, -0.00000748, -0.00000148, 0.00000382, 0.00000577, 0.00000265, -0.00000148, -0.00000312, -0.00000788, -0.00000710, 0.00000699, 0.00000848, -0.00000964, -0.00001265, -0.00000160, -0.00000328, 0.00000012, 0.00001648, 0.00001067, -0.00001161, -0.00001480, -0.00000726, -0.00000352, 0.00000303, 0.00000583, 0.00000203, -0.00000252, -0.00001016, -0.00000536, 0.00001137, 0.00000422, -0.00000483, 0.00001240, -0.00000295, -0.00003047, 0.00001399, 0.00003211, -0.00003582, -0.00001901, 0.00005011, -0.00001726, -0.00005496, 0.00004700, 0.00001405, -0.00008292, 0.00003676, 0.00007567, -0.00010021, -0.00003162, 0.00012353, -0.00007681, -0.00012330, 0.00016019, -0.00000080, -0.00027404, 0.00013733, 0.00019569, -0.00063240, -0.00045969, 0.00096787, 0.00110064, -0.00027485, -0.00066631, 0.00007425, 0.00017120, -0.00020114, -0.00006210, 0.00011595, -0.00005607, -0.00006936, 0.00006683, -0.00000245, -0.00006361, 0.00003012, 0.00003221, -0.00004028, -0.00001234, 0.00002208, -0.00001203, -0.00001803, 0.00000769, 0.00000332, -0.00000840, -0.00000058, 0.00000691, 0.00000020, -0.00001130, -0.00000768, 0.00000691, 0.00000734, 0.00000238, 0.00000751, 0.00000168, -0.00001208, -0.00000467, 0.00000687, 0.00000114, -0.00000118, -0.00000022, -0.00001247, -0.00001561, 0.00000133, 0.00000917, 0.00000345, 0.00000232, -0.00000124, -0.00001146, -0.00000952, 0.00000434, 0.00000594, -0.00000805, -0.00001296, 0.00000010, 0.00000700, -0.00000229, -0.00000294, 0.00000833, },
		{ -0.00000624, 0.00000767, 0.00000787, -0.00000246, -0.00000244, -0.00000134, -0.00000364, 0.00000454, 0.00001170, 0.00000147, -0.00000847, -0.00000368, 0.00000354, 0.00000490, 0.00000301, 0.00000048, -0.00000126, -0.00000623, -0.00000999, 0.00000120, 0.00001136, -0.00000272, -0.00001533, -0.00000521, -0.00000035, -0.00000316, 0.00001050, 0.00001635, -0.00000512, -0.00001729, -0.00000867, -0.00000409, -0.00000049, 0.00000675, 0.00000419, -0.00000350, -0.00000783, -0.00000609, 0.00000587, 0.00000819, -0.00000212, 0.00000595, 0.00000487, -0.00002439, -0.00000542, 0.00003428, -0.00001119, -0.00003682, 0.00003316, 0.00001662, -0.00005858, 0.00001073, 0.00004740, -0.00006541, -0.00002217, 0.00009516, -0.00003961, -0.00009495, 0.00009884, 0.00002064, -0.00016381, 0.00006976, 0.00012524, -0.00024317, -0.00005960, 0.00031550, -0.00033111, -0.00075619, 0.00046785, 0.00133989, 0.00022976, -0.00075868, -0.00019191, 0.00027144, -0.00010965, -0.00017482, 0.00011213, 0.00001804, -0.00011408, 0.00003730, 0.00004852, -0.00007276, -0.00000707, 0.00005804, -0.00002477, -0.00003628, 0.00002445, 0.00000194, -0.00002770, 0.00000346, 0.00001093, -0.00001057, -0.00000491, 0.00000910, 0.00000282, -0.00000998, -0.00000998, 0.00000315, 0.00000872, 0.00000285, 0.00000515, 0.00000551, -0.00000893, -0.00000993, 0.00000532, 0.00000500, -0.00000209, 0.00000019, -0.00000771, -0.00001800, -0.00000464, 0.00001043, 0.00000566, 0.00000139, 0.00000152, -0.00000836, -0.00001329, -0.00000013, 0.00000867, -0.00000239, -0.00001350, -0.00000577, 0.00000617, 0.00000199, -0.00000437, 0.00000445, 0.00000823, -0.00000461, },
		{ 0.00001097, 0.00000181, -0.00000379, -0.00000089, -0.00000323, -0.00000052, 0.00001052, 0.00000741, -0.00000667, -0.00000687, 0.00000278, 0.00000518, 0.00000235, 0.00000127, 0.00000091, -0.00000377, -0.00001069, -0.00000482, 0.00000997, 0.00000464, -0.00001368, -0.00001067, 0.00000051, -0.00000243, 0.00000373, 0.00001708, 0.00000358, -0.00001699, -0.00001224, -0.00000362, -0.00000296, 0.00000459, 0.00000743, -0.00000270, -0.00000838, -0.00000519, 0.00000241, 0.00000770, 0.00000226, 0.00000267, 0.00000627, -0.00001497, -0.00001651, 0.00002424, 0.00001020, -0.00003662, 0.00000729, 0.00003530, -0.00003996, -0.00002479, 0.00005323, -0.00002557, -0.00006315, 0.00007168, 0.00002883, -0.00011150, 0.00003068, 0.00009314, -0.00013181, -0.00004438, 0.00017499, -0.00012206, -0.00021786, 0.00026566, 0.00000207, -0.00078865, -0.00010061, 0.00126921, 0.00078550, -0.00060826, -0.00050933, 0.00024362, 0.00005552, -0.00024521, 0.00003716, 0.00010312, -0.00011438, -0.00002784, 0.00008704, -0.00004631, -0.00005539, 0.00006200, 0.00001017, -0.00005514, 0.00001117, 0.00002199, -0.00003048, -0.00000935, 0.00001931, -0.00000723, -0.00001301, 0.00000968, 0.00000752, -0.00000909, -0.00001123, -0.00000030, 0.00000801, 0.00000468, 0.00000327, 0.00000641, -0.00000398, -0.00001256, 0.00000069, 0.00000819, -0.00000062, -0.00000128, -0.00000340, -0.00001685, -0.00001164, 0.00000881, 0.00000957, 0.00000091, 0.00000204, -0.00000375, -0.00001442, -0.00000617, 0.00000860, 0.00000348, -0.00001113, -0.00001029, 0.00000246, 0.00000485, -0.00000274, 0.00000045, 0.00000875, 0.00000044, -0.00000912, 0.00000183, },
		{ -0.00000304, -0.00000206, -0.00000162, -0.00000347, 0.00000603, 0.00001132, -0.00000170, -0.00000931, 0.00000027, 0.00000633, 0.00000222, 0.00000046, 0.00000248, -0.00000061, -0.00000950, -0.00000950, 0.00000535, 0.00000960, -0.00000810, -0.00001486, -0.00000214, 0.00000018, -0.00000055, 0.00001318, 0.00001106, -0.00001243, -0.00001650, -0.00000443, -0.00000308, 0.00000075, 0.00000883, 0.00000089, -0.00000965, -0.00000573, 0.00000171, 0.00000552, 0.00000472, 0.00000317, 0.00000492, -0.00000766, -0.00001887, 0.00001015, 0.00002096, -0.00002355, -0.00001425, 0.00003482, -0.00001197, -0.00004427, 0.00003449, 0.00001469, -0.00007178, 0.00002246, 0.00007385, -0.00007978, -0.00004372, 0.00011119, -0.00005114, -0.00012698, 0.00013701, 0.00002426, -0.00027264, 0.00009824, 0.00023955, -0.00058600, -0.00056223, 0.00090037, 0.00121767, -0.00020811, -0.00075047, 0.00005398, 0.00022637, -0.00022222, -0.00009015, 0.00015271, -0.00005651, -0.00010184, 0.00008596, 0.00001122, -0.00009065, 0.00003363, 0.00005217, -0.00005445, -0.00001766, 0.00003799, -0.00001966, -0.00002792, 0.00002151, 0.00000420, -0.00002144, 0.00000492, 0.00001452, -0.00000707, -0.00001312, -0.00000238, 0.00000592, 0.00000588, 0.00000326, 0.00000532, 0.00000011, -0.00001163, -0.00000495, 0.00000856, 0.00000307, -0.00000274, -0.00000136, -0.00001271, -0.00001707, 0.00000339, 0.00001341, 0.00000272, 0.00000051, 0.00000024, -0.00001211, -0.00001201, 0.00000527, 0.00000831, -0.00000658, -0.00001246, -0.00000235, 0.00000497, 0.00000022, -0.00000151, 0.00000656, 0.00000485, -0.00000787, -0.00000470, 0.00000982, 0.00000733, },
		{ -0.00000080, -0.00000359, 0.00000064, 0.00001124, 0.00000478, -0.00000887, -0.00000390, 0.00000676, 0.00000375, -0.00000106, 0.00000235, 0.00000271, -0.00000662, -0.00001214, -0.00000071, 0.00001074, -0.00000089, -0.00001542, -0.00000722, 0.00000167, -0.00000119, 0.00000736, 0.00001451, -0.00000466, -0.00001876, -0.00000777, -0.00000180, -0.00000223, 0.00000703, 0.00000557, -0.00000886, -0.00000847, 0.00000177, 0.00000449, 0.00000448, 0.00000512, 0.00000444, -0.00000402, -0.00001632, -0.00000114, 0.00002083, -0.00000709, -0.00002375, 0.00002100, 0.00001123, -0.00004326, 0.00000502, 0.00003772, -0.00005132, -0.00002647, 0.00007990, -0.00002127, -0.00008942, 0.00007524, 0.00003451, -0.00014527, 0.00004234, 0.00012833, -0.00021558, -0.00009342, 0.00031139, -0.00025871, -0.00078982, 0.00034958, 0.00138056, 0.00034958, -0.00078982, -0.00025871, 0.00031139, -0.00009342, -0.00021558, 0.00012833, 0.00004234, -0.00014527, 0.00003451, 0.00007524, -0.00008942, -0.00002127, 0.00007990, -0.00002647, -0.00005132, 0.00003772, 0.00000502, -0.00004326, 0.00001123, 0.00002100, -0.00002375, -0.00000709, 0.00002083, -0.00000114, -0.00001632, -0.00000402, 0.00000444, 0.00000512, 0.00000448, 0.00000449, 0.00000177, -0.00000847, -0.00000886, 0.00000557, 0.00000703, -0.00000223, -0.00000180, -0.00000777, -0.00001876, -0.00000466, 0.00001451, 0.00000736, -0.00000119, 0.00000167, -0.00000722, -0.00001542, -0.00000089, 0.00001074, -0.00000071, -0.00001214, -0.00000662, 0.00000271, 0.00000235, -0.00000106, 0.00000375, 0.00000676, -0.00000390, -0.00000887, 0.00000478, 0.00001124, 0.00000064, -0.00000359, },
		{ -0.00000304, 0.00000733, 0.00000982, -0.00000470, -0.00000787, 0.00000485, 0.00000656, -0.00000151, 0.00000022, 0.00000497, -0.00000235, -0.00001246, -0.00000658, 0.00000831, 0.00000527, -0.00001201, -0.00001211, 0.00000024, 0.00000051, 0.00000272, 0.00001341, 0.00000339, -0.00001707, -0.00001271, -0.00000136, -0.00000274, 0.00000307, 0.00000856, -0.00000495, -0.00001163, 0.00000011, 0.00000532, 0.00000326, 0.00000588, 0.00000592, -0.00000238, -0.00001312, -0.00000707, 0.00001452, 0.00000492, -0.00002144, 0.00000420, 0.00002151, -0.00002792, -0.00001966, 0.00003799, -0.00001766, -0.00005445, 0.00005217, 0.00003363, -0.00009065, 0.00001122, 0.00008596, -0.00010184, -0.00005651, 0.00015271, -0.00009015, -0.00022222, 0.00022637, 0.00005398, -0.00075047, -0.00020811, 0.00121767, 0.00090037, -0.00056223, -0.00058600, 0.00023955, 0.00009824, -0.00027264, 0.00002426, 0.00013701, -0.00012698, -0.00005114, 0.00011119, -0.00004372, -0.00007978, 0.00007385, 0.00002246, -0.00007178, 0.00001469, 0.00003449, -0.00004427, -0.00001197, 0.00003482, -0.00001425, -0.00002355, 0.00002096, 0.00001015, -0.00001887, -0.00000766, 0.00000492, 0.00000317, 0.00000472, 0.00000552, 0.00000171, -0.00000573, -0.00000965, 0.00000089, 0.00000883, 0.00000075, -0.00000308, -0.00000443, -0.00001650, -0.00001243, 0.00001106, 0.00001318, -0.00000055, 0.00000018, -0.00000214, -0.00001486, -0.00000810, 0.00000960, 0.00000535, -0.00000950, -0.00000950, -0.00000061, 0.00000248, 0.00000046, 0.00000222, 0.00000633, 0.00000027, -0.00000931, -0.00000170, 0.00001132, 0.00000603, -0.00000347, -0.00000162, -0.00000206, },
		{ 0.00001097, 0.00000183, -0.00000912, 0.00000044, 0.00000875, 0.00000045, -0.00000274, 0.00000485, 0.00000246, -0.00001029, -0.00001113, 0.00000348, 0.00000860, -0.00000617, -0.00001442, -0.00000375, 0.00000204, 0.00000091, 0.00000957, 0.00000881, -0.00001164, -0.00001685, -0.00000340, -0.00000128, -0.00000062, 0.00000819, 0.00000069, -0.00001256, -0.00000398, 0.00000641, 0.00000327, 0.00000468, 0.00000801, -0.00000030, -0.00001123, -0.00000909, 0.00000752, 0.00000968, -0.00001301, -0.00000723, 0.00001931, -0.00000935, -0.00003048, 0.00002199, 0.00001117, -0.00005514, 0.00001017, 0.00006200, -0.00005539, -0.00004631, 0.00008704, -0.00002784, -0.00011438, 0.00010312, 0.00003716, -0.00024521, 0.00005552, 0.00024362, -0.00050933, -0.00060826, 0.00078550, 0.00126921, -0.00010061, -0.00078865, 0.00000207, 0.00026566, -0.00021786, -0.00012206, 0.00017499, -0.00004438, -0.00013181, 0.00009314, 0.00003068, -0.00011150, 0.00002883, 0.00007168, -0.00006315, -0.00002557, 0.00005323, -0.00002479, -0.00003996, 0.00003530, 0.00000729, -0.00003662, 0.00001020, 0.00002424, -0.00001651, -0.00001497, 0.00000627, 0.00000267, 0.00000226, 0.00000770, 0.00000241, -0.00000519, -0.00000838, -0.00000270, 0.00000743, 0.00000459, -0.00000296, -0.00000362, -0.00001224, -0.00001699, 0.00000358, 0.00001708, 0.00000373, -0.00000243, 0.00000051, -0.00001067, -0.00001368, 0.00000464, 0.00000997, -0.00000482, -0.00001069, -0.00000377, 0.00000091, 0.00000127, 0.00000235, 0.00000518, 0.00000278, -0.00000687, -0.00000667, 0.00000741, 0.00001052, -0.00000052, -0.00000323, -0.00000089, -0.00000379, 0.00000181, },
		{ -0.00000624, -0.00000461, 0.00000823, 0.00000445, -0.00000437, 0.00000199, 0.00000617, -0.00000577, -0.00001350, -0.00000239, 0.00000867, -0.00000013, -0.00001329, -0.00000836, 0.00000152, 0.00000139, 0.00000566, 0.00001043, -0.00000464, -0.00001800, -0.00000771, 0.00000019, -0.00000209, 0.00000500, 0.00000532, -0.00000993, -0.00000893, 0.00000551, 0.00000515, 0.00000285, 0.00000872, 0.00000315, -0.00000998, -0.00000998, 0.00000282, 0.00000910, -0.00000491, -0.00001057, 0.00001093, 0.00000346, -0.00002770, 0.00000194, 0.00002445, -0.00003628, -0.00002477, 0.00005804, -0.00000707, -0.00007276, 0.00004852, 0.00003730, -0.00011408, 0.00001804, 0.00011213, -0.00017482, -0.00010965, 0.00027144, -0.00019191, -0.00075868, 0.00022976, 0.00133989, 0.00046785, -0.00075619, -0.00033111, 0.00031550, -0.00005960, -0.00024317, 0.00012524, 0.00006976, -0.00016381, 0.00002064, 0.00009884, -0.00009495, -0.00003961, 0.00009516, -0.00002217, -0.00006541, 0.00004740, 0.00001073, -0.00005858, 0.00001662, 0.00003316, -0.00003682, -0.00001119, 0.00003428, -0.00000542, -0.00002439, 0.00000487, 0.00000595, -0.00000212, 0.00000819, 0.00000587, -0.00000609, -0.00000783, -0.00000350, 0.00000419, 0.00000675, -0.00000049, -0.00000409, -0.00000867, -0.00001729, -0.00000512, 0.00001635, 0.00001050, -0.00000316, -0.00000035, -0.00000521, -0.00001533, -0.00000272, 0.00001136, 0.00000120, -0.00000999, -0.00000623, -0.00000126, 0.00000048, 0.00000301, 0.00000490, 0.00000354, -0.00000368, -0.00000847, 0.00000147, 0.00001170, 0.00000454, -0.00000364, -0.00000134, -0.00000244, -0.00000246, 0.00000787, 0.00000767, },
		{ 0.00000439, 0.00000833, -0.00000294, -0.00000229, 0.00000700, 0.00000010, -0.00001296, -0.00000805, 0.00000594, 0.00000434, -0.00000952, -0.00001146, -0.00000124, 0.00000232, 0.00000345, 0.00000917, 0.00000133, -0.00001561, -0.00001247, -0.00000022, -0.00000118, 0.00000114, 0.00000687, -0.00000467, -0.00001208, 0.00000168, 0.00000751, 0.00000238, 0.00000734, 0.00000691, -0.00000768, -0.00001130, 0.00000020, 0.00000691, -0.00000058, -0.00000840, 0.00000332, 0.00000769, -0.00001803, -0.00001203, 0.00002208, -0.00001234, -0.00004028, 0.00003221, 0.00003012, -0.00006361, -0.00000245, 0.00006683, -0.00006936, -0.00005607, 0.00011595, -0.00006210, -0.00020114, 0.00017120, 0.00007425, -0.00066631, -0.00027485, 0.00110064, 0.00096787, -0.00045969, -0.00063240, 0.00019569, 0.00013733, -0.00027404, -0.00000080, 0.00016019, -0.00012330, -0.00007681, 0.00012353, -0.00003162, -0.00010021, 0.00007567, 0.00003676, -0.00008292, 0.00001405, 0.00004700, -0.00005496, -0.00001726, 0.00005011, -0.00001901, -0.00003582, 0.00003211, 0.00001399, -0.00003047, -0.00000295, 0.00001240, -0.00000483, 0.00000422, 0.00001137, -0.00000536, -0.00001016, -0.00000252, 0.00000203, 0.00000583, 0.00000303, -0.00000352, -0.00000726, -0.00001480, -0.00001161, 0.00001067, 0.00001648, 0.00000012, -0.00000328, -0.00000160, -0.00001265, -0.00000964, 0.00000848, 0.00000699, -0.00000710, -0.00000788, -0.00000312, -0.00000148, 0.00000265, 0.00000577, 0.00000382, -0.00000148, -0.00000748, -0.00000368, 0.00000893, 0.00000932, -0.00000138, -0.00000291, -0.00000103, -0.00000381, 0.00000245, 0.00000991, -0.00000017, -0.00000740, },
		{ 0.00000144, -0.00000538, 0.00000429, 0.00000531, -0.00000923, -0.00001223, 0.00000124, 0.00000639, -0.00000472, -0.00001200, -0.00000492, 0.00000204, 0.00000298, 0.00000698, 0.00000477, -0.00001093, -0.00001557, -0.00000302, 0.00000064, -0.00000121, 0.00000519, 0.00000069, -0.00001165, -0.00000382, 0.00000810, 0.00000404, 0.00000502, 0.00000919, -0.00000359, -0.00001261, -0.00000212, 0.00000561, 0.00000041, -0.00000522, 0.00000002, 0.00000619, -0.00000901, -0.00001653, 0.00001149, 0.00000467, -0.00003666, 0.00000209, 0.00004331, -0.00003233, -0.00003932, 0.00005722, -0.00001164, -0.00008895, 0.00006661, 0.00003520, -0.00019938, 0.00001977, 0.00021028, -0.00042029, -0.00059274, 0.00064504, 0.00124485, 0.00002897, -0.00076842, -0.00007184, 0.00027722, -0.00018813, -0.00014954, 0.00017693, -0.00002242, -0.00015217, 0.00008599, 0.00005141, -0.00012115, 0.00001635, 0.00008638, -0.00006412, -0.00003460, 0.00006474, -0.00002587, -0.00005232, 0.00004609, 0.00001286, -0.00005138, 0.00001354, 0.00003514, -0.00002638, -0.00001751, 0.00001756, -0.00000168, -0.00000407, 0.00001517, 0.00000002, -0.00001461, -0.00000305, 0.00000306, 0.00000304, 0.00000485, -0.00000058, -0.00000712, -0.00001232, -0.00001406, 0.00000250, 0.00001828, 0.00000691, -0.00000504, -0.00000169, -0.00000765, -0.00001318, 0.00000213, 0.00001037, -0.00000214, -0.00000837, -0.00000452, -0.00000344, 0.00000059, 0.00000663, 0.00000501, -0.00000050, -0.00000549, -0.00000620, 0.00000374, 0.00001131, 0.00000326, -0.00000375, -0.00000113, -0.00000268, -0.00000227, 0.00000758, 0.00000605, -0.00000588, -0.00000105, 0.00000946, },
		{ -0.00000066, 0.00000756, -0.00000317, -0.00001361, -0.00000439, 0.00000587, -0.00000036, -0.00001025, -0.00000787, 0.00000015, 0.00000304, 0.00000542, 0.00000587, -0.00000598, -0.00001599, -0.00000713, 0.00000153, -0.00000123, 0.00000193, 0.00000376, -0.00000796, -0.00000847, 0.00000563, 0.00000674, 0.00000367, 0.00000913, 0.00000135, -0.00001242, -0.00000532, 0.00000527, 0.00000049, -0.00000399, 0.00000023, 0.00000381, -0.00000445, -0.00001447, 0.00000125, 0.00001049, -0.00002303, -0.00001821, 0.00003448, 0.00000067, -0.00004928, 0.00002460, 0.00002960, -0.00007757, 0.00000177, 0.00008118, -0.00013063, -0.00010556, 0.00020722, -0.00014346, -0.00067304, 0.00012950, 0.00122489, 0.00056194, -0.00065939, -0.00039210, 0.00027899, -0.00001674, -0.00024937, 0.00010344, 0.00009319, -0.00016499, -0.00000131, 0.00011328, -0.00008788, -0.00005793, 0.00010007, -0.00001263, -0.00007537, 0.00005134, 0.00001827, -0.00007063, 0.00001799, 0.00004551, -0.00004695, -0.00001728, 0.00004719, -0.00000861, -0.00003361, 0.00001514, 0.00000855, -0.00001243, 0.00001266, 0.00001001, -0.00001721, -0.00000818, 0.00000637, 0.00000179, 0.00000309, 0.00000353, -0.00000577, -0.00001172, -0.00001349, -0.00000443, 0.00001472, 0.00001404, -0.00000280, -0.00000460, -0.00000377, -0.00001213, -0.00000499, 0.00000961, 0.00000367, -0.00000702, -0.00000566, -0.00000462, -0.00000248, 0.00000594, 0.00000722, 0.00000040, -0.00000415, -0.00000613, -0.00000123, 0.00000947, 0.00000819, -0.00000213, -0.00000265, -0.00000102, -0.00000413, 0.00000226, 0.00000908, -0.00000045, -0.00000490, 0.00000664, 0.00000648, -0.00000502, },
		{ 0.00000303, -0.00001134, -0.00000956, 0.00000297, 0.00000263, -0.00000722, -0.00000914, -0.00000250, 0.00000242, 0.00000474, 0.00000586, -0.00000215, -0.00001418, -0.00001088, 0.00000052, 0.00000029, -0.00000080, 0.00000372, -0.00000305, -0.00001011, 0.00000081, 0.00000836, 0.00000438, 0.00000752, 0.00000538, -0.00000988, -0.00000895, 0.00000445, 0.00000162, -0.00000474, 0.00000092, 0.00000368, -0.00000354, -0.00001110, -0.00000406, 0.00000825, -0.00000982, -0.00002411, 0.00001541, 0.00002016, -0.00003632, -0.00000828, 0.00004113, -0.00004086, -0.00004442, 0.00007330, -0.00004335, -0.00016080, 0.00011296, 0.00006204, -0.00055493, -0.00029163, 0.00093906, 0.00097398, -0.00031713, -0.00063410, 0.00011900, 0.00016091, -0.00024751, -0.00003112, 0.00016613, -0.00010427, -0.00009841, 0.00012053, -0.00001310, -0.00011165, 0.00006684, 0.00004965, -0.00008582, 0.00000945, 0.00005699, -0.00006015, -0.00002457, 0.00006212, -0.00001999, -0.00004790, 0.00004064, 0.00001907, -0.00004205, 0.00000140, 0.00002182, -0.00001440, 0.00000208, 0.00002026, -0.00001316, -0.00001776, 0.00000785, 0.00000506, -0.00000117, 0.00000562, -0.00000144, -0.00001229, -0.00001278, -0.00000769, 0.00000782, 0.00001759, 0.00000370, -0.00000705, -0.00000355, -0.00000804, -0.00000951, 0.00000476, 0.00000805, -0.00000336, -0.00000636, -0.00000511, -0.00000520, 0.00000311, 0.00000918, 0.00000249, -0.00000372, -0.00000500, -0.00000398, 0.00000491, 0.00001081, 0.00000211, -0.00000391, -0.00000072, -0.00000311, -0.00000293, 0.00000730, 0.00000599, -0.00000445, 0.00000127, 0.00000902, -0.00000089, -0.00000494, 0.00000567, },
		{ -0.00001249, -0.00000173, 0.00000373, -0.00000390, -0.00000879, -0.00000474, 0.00000082, 0.00000421, 0.00000579, 0.00000039, -0.00001141, -0.00001319, -0.00000207, 0.00000188, -0.00000172, 0.00000149, 0.00000067, -0.00000836, -0.00000419, 0.00000729, 0.00000654, 0.00000612, 0.00000731, -0.00000558, -0.00001150, 0.00000197, 0.00000384, -0.00000555, -0.00000028, 0.00000557, -0.00000301, -0.00000982, -0.00000504, 0.00000401, -0.00000276, -0.00001988, -0.00000132, 0.00002242, -0.00001458, -0.00002556, 0.00002842, -0.00000431, -0.00005723, 0.00003462, 0.00002046, -0.00014576, -0.00000276, 0.00014989, -0.00033530, -0.00052272, 0.00050286, 0.00114755, 0.00015700, -0.00068774, -0.00015151, 0.00025411, -0.00013987, -0.00016403, 0.00015746, 0.00000332, -0.00015750, 0.00006600, 0.00006776, -0.00011713, -0.00000075, 0.00009248, -0.00005708, -0.00004243, 0.00007005, -0.00002254, -0.00006253, 0.00005139, 0.00002019, -0.00006279, 0.00001373, 0.00004551, -0.00003454, -0.00002102, 0.00002984, -0.00000548, -0.00001292, 0.00002393, -0.00000059, -0.00002721, 0.00000275, 0.00001254, -0.00000385, 0.00000288, 0.00000486, -0.00001118, -0.00001427, -0.00000789, 0.00000154, 0.00001557, 0.00001152, -0.00000555, -0.00000671, -0.00000442, -0.00000948, -0.00000182, 0.00000870, 0.00000187, -0.00000573, -0.00000549, -0.00000664, -0.00000103, 0.00000920, 0.00000587, -0.00000312, -0.00000440, -0.00000426, 0.00000010, 0.00000972, 0.00000719, -0.00000297, -0.00000216, -0.00000105, -0.00000537, 0.00000184, 0.00000971, 0.00000054, -0.00000327, 0.00000720, 0.00000478, -0.00000567, 0.00000067, 0.00000657, -0.00000586, },
		{ 0.00000257, -0.00000101, -0.00000738, -0.00000592, -0.00000113, 0.00000305, 0.00000580, 0.00000229, -0.00000862, -0.00001399, -0.00000523, 0.00000242, -0.00000092, -0.00000114, 0.00000187, -0.00000461, -0.00000710, 0.00000363, 0.00000834, 0.00000624, 0.00000741, -0.00000117, -0.00001176, -0.00000196, 0.00000572, -0.00000460, -0.00000341, 0.00000710, -0.00000043, -0.00001048, -0.00000512, 0.00000192, -0.00000136, -0.00001316, -0.00000926, 0.00001376, 0.00000200, -0.00002470, 0.00000745, 0.00001510, -0.00004326, -0.00000502, 0.00004335, -0.00009131, -0.00008452, 0.00013327, -0.00011877, -0.00055163, 0.00006320, 0.00105478, 0.00061362, -0.00051315, -0.00042506, 0.00020607, 0.00002314, -0.00023048, 0.00006862, 0.00010568, -0.00014832, -0.00002566, 0.00011425, -0.00007013, -0.00007143, 0.00009317, -0.00000037, -0.00007855, 0.00004888, 0.00002596, -0.00007677, 0.00001486, 0.00005556, -0.00005180, -0.00002454, 0.00005692, -0.00000960, -0.00004264, 0.00002464, 0.00001276, -0.00002416, 0.00001605, 0.00001675, -0.00002950, -0.00001054, 0.00001959, -0.00000038, -0.00000425, 0.00000924, -0.00000572, -0.00001747, -0.00000821, -0.00000104, 0.00000954, 0.00001623, 0.00000093, -0.00001003, -0.00000447, -0.00000588, -0.00000648, 0.00000497, 0.00000649, -0.00000286, -0.00000586, -0.00000701, -0.00000483, 0.00000655, 0.00000914, -0.00000103, -0.00000469, -0.00000339, -0.00000280, 0.00000559, 0.00001047, 0.00000080, -0.00000387, -0.00000014, -0.00000447, -0.00000418, 0.00000838, 0.00000746, -0.00000363, 0.00000190, 0.00000839, -0.00000196, -0.00000435, 0.00000564, 0.00000067, -0.00001173, -0.00000697, },
		{ -0.00000545, -0.00000614, -0.00000258, 0.00000126, 0.00000531, 0.00000403, -0.00000595, -0.00001386, -0.00000814, 0.00000170, 0.00000069, -0.00000277, 0.00000073, -0.00000098, -0.00000698, -0.00000097, 0.00000807, 0.00000780, 0.00000706, 0.00000199, -0.00000979, -0.00000592, 0.00000578, -0.00000167, -0.00000666, 0.00000611, 0.00000407, -0.00001061, -0.00000695, 0.00000220, -0.00000171, -0.00000931, -0.00000949, 0.00000380, 0.00000730, -0.00001384, -0.00000706, 0.00001539, -0.00002039, -0.00002619, 0.00003334, -0.00003515, -0.00011133, 0.00006183, 0.00002466, -0.00043626, -0.00026055, 0.00075905, 0.00091624, -0.00015862, -0.00058470, 0.00002460, 0.00015964, -0.00019803, -0.00005777, 0.00015224, -0.00007500, -0.00010997, 0.00010244, 0.00000663, -0.00011096, 0.00004953, 0.00005756, -0.00007963, 0.00000242, 0.00006214, -0.00005869, -0.00003230, 0.00006817, -0.00001680, -0.00005736, 0.00004450, 0.00002468, -0.00005131, 0.00000408, 0.00003200, -0.00002326, -0.00000231, 0.00003039, -0.00001960, -0.00002796, 0.00001932, 0.00001056, -0.00001121, 0.00000740, 0.00000367, -0.00001892, -0.00001179, -0.00000011, 0.00000379, 0.00001495, 0.00000976, -0.00000936, -0.00000861, -0.00000228, -0.00000658, -0.00000121, 0.00000778, 0.00000205, -0.00000532, -0.00000716, -0.00000712, 0.00000211, 0.00001040, 0.00000273, -0.00000482, -0.00000299, -0.00000322, 0.00000071, 0.00001016, 0.00000599, -0.00000375, -0.00000125, -0.00000186, -0.00000760, 0.00000254, 0.00001221, 0.00000123, -0.00000340, 0.00000730, 0.00000419, -0.00000593, 0.00000085, 0.00000507, -0.00000716, -0.00001077, -0.00000083, 0.00000081, },
		{ -0.00000322, -0.00000046, 0.00000385, 0.00000540, -0.00000303, -0.00001316, -0.00001064, 0.00000010, 0.00000222, -0.00000296, -0.00000152, 0.00000096, -0.00000452, -0.00000444, 0.00000536, 0.00000941, 0.00000757, 0.00000366, -0.00000676, -0.00000846, 0.00000372, 0.00000192, -0.00000796, 0.00000233, 0.00000825, -0.00000799, -0.00001047, 0.00000238, -0.00000049, -0.00000894, -0.00000723, -0.00000116, 0.00000394, -0.00000387, -0.00000962, 0.00000593, -0.00000448, -0.00002667, 0.00001138, -0.00000129, -0.00009471, -0.00001131, 0.00007713, -0.00026532, -0.00041487, 0.00037902, 0.00099291, 0.00026097, -0.00055626, -0.00021832, 0.00019708, -0.00008493, -0.00015976, 0.00012092, 0.00002545, -0.00014591, 0.00003818, 0.00007483, -0.00010041, -0.00001793, 0.00008803, -0.00004391, -0.00004665, 0.00006807, -0.00001572, -0.00006819, 0.00004998, 0.00002765, -0.00006828, 0.00001048, 0.00005318, -0.00003911, -0.00002513, 0.00004082, -0.00000724, -0.00002312, 0.00003154, 0.00000129, -0.00004077, 0.00000715, 0.00002477, -0.00001149, -0.00000183, 0.00001269, -0.00001455, -0.00001879, 0.00000072, 0.00000234, 0.00000865, 0.00001592, -0.00000268, -0.00001371, -0.00000261, -0.00000246, -0.00000598, 0.00000438, 0.00000699, -0.00000256, -0.00000751, -0.00000807, -0.00000227, 0.00000868, 0.00000676, -0.00000337, -0.00000356, -0.00000227, -0.00000257, 0.00000648, 0.00001001, -0.00000069, -0.00000327, 0.00000000, -0.00000719, -0.00000475, 0.00001169, 0.00000909, -0.00000470, 0.00000180, 0.00000882, -0.00000247, -0.00000472, 0.00000489, -0.00000070, -0.00001115, -0.00000558, 0.00000085, -0.00000333, -0.00000580, },
		{ 0.00000174, 0.00000568, 0.00000022, -0.00001167, -0.00001288, -0.00000212, 0.00000326, -0.00000198, -0.00000366, 0.00000083, -0.00000133, -0.00000549, 0.00000128, 0.00000954, 0.00000919, 0.00000469, -0.00000399, -0.00000909, 0.00000048, 0.00000445, -0.00000660, -0.00000256, 0.00000974, -0.00000256, -0.00001338, 0.00000010, 0.00000265, -0.00000924, -0.00000696, -0.00000081, -0.00000113, -0.00000105, -0.00000427, -0.00000171, -0.00000090, -0.00001678, -0.00000431, 0.00000727, -0.00006160, -0.00005432, 0.00006301, -0.00011551, -0.00041640, 0.00003543, 0.00085618, 0.00061383, -0.00034040, -0.00041871, 0.00010957, 0.00004829, -0.00018897, 0.00002990, 0.00010289, -0.00011781, -0.00004589, 0.00010079, -0.00004651, -0.00007635, 0.00007595, 0.00001116, -0.00007374, 0.00004101, 0.00003179, -0.00007553, 0.00000804, 0.00006089, -0.00005024, -0.00003138, 0.00006119, -0.00000803, -0.00004964, 0.00003128, 0.00001834, -0.00003490, 0.00001676, 0.00002484, -0.00004016, -0.00001521, 0.00003372, -0.00000097, -0.00001438, 0.00001517, -0.00000321, -0.00002535, -0.00000298, 0.00000580, 0.00000202, 0.00001534, 0.00000786, -0.00001484, -0.00000818, 0.00000226, -0.00000590, -0.00000228, 0.00000887, 0.00000276, -0.00000721, -0.00000875, -0.00000514, 0.00000475, 0.00000896, -0.00000000, -0.00000417, -0.00000158, -0.00000334, 0.00000147, 0.00001068, 0.00000440, -0.00000401, -0.00000030, -0.00000421, -0.00000984, 0.00000570, 0.00001558, -0.00000007, -0.00000474, 0.00000848, 0.00000439, -0.00000704, 0.00000017, 0.00000426, -0.00000736, -0.00000969, -0.00000142, -0.00000154, -0.00000507, -0.00000336, -0.00000139, },
		{ 0.00000317, -0.00000894, -0.00001469, -0.00000502, 0.00000369, -0.00000028, -0.00000497, -0.00000079, 0.00000108, -0.00000422, -0.00000245, 0.00000751, 0.00001100, 0.00000618, -0.00000199, -0.00000838, -0.00000256, 0.00000504, -0.00000341, -0.00000629, 0.00000771, 0.00000364, -0.00001325, -0.00000475, 0.00000547, -0.00000735, -0.00000946, 0.00000100, -0.00000247, -0.00000449, 0.00000091, -0.00000172, -0.00000481, -0.00000884, -0.00000750, 0.00000224, -0.00003491, -0.00006347, 0.00002320, -0.00002533, -0.00032657, -0.00019362, 0.00058559, 0.00080436, -0.00000992, -0.00048826, -0.00006801, 0.00013008, -0.00013638, -0.00007280, 0.00012096, -0.00004319, -0.00010802, 0.00007342, 0.00002198, -0.00009804, 0.00002809, 0.00005795, -0.00006573, -0.00000473, 0.00006116, -0.00005104, -0.00003841, 0.00006674, -0.00001030, -0.00006193, 0.00004269, 0.00002949, -0.00005613, 0.00000444, 0.00004111, -0.00002918, -0.00000873, 0.00003923, -0.00002292, -0.00003901, 0.00002918, 0.00001827, -0.00002219, 0.00000727, 0.00001125, -0.00002566, -0.00001285, 0.00001037, 0.00000041, 0.00000801, 0.00001669, -0.00000864, -0.00001617, 0.00000279, -0.00000048, -0.00000814, 0.00000548, 0.00000886, -0.00000455, -0.00000992, -0.00000659, 0.00000065, 0.00000809, 0.00000398, -0.00000332, -0.00000187, -0.00000244, -0.00000242, 0.00000770, 0.00000903, -0.00000190, -0.00000218, -0.00000126, -0.00001073, -0.00000299, 0.00001679, 0.00000890, -0.00000766, 0.00000264, 0.00001043, -0.00000376, -0.00000613, 0.00000463, -0.00000094, -0.00001078, -0.00000558, -0.00000095, -0.00000390, -0.00000344, -0.00000143, -0.00000011, 0.00000446, },
		{ -0.00001534, -0.00000870, 0.00000323, 0.00000187, -0.00000526, -0.00000307, 0.00000200, -0.00000166, -0.00000450, 0.00000389, 0.00001170, 0.00000849, -0.00000033, -0.00000728, -0.00000461, 0.00000396, -0.00000003, -0.00000751, 0.00000332, 0.00000796, -0.00000945, -0.00001003, 0.00000545, -0.00000275, -0.00001235, 0.00000037, 0.00000062, -0.00000884, 0.00000042, 0.00000386, -0.00000766, -0.00000860, -0.00000328, -0.00000242, -0.00002288, -0.00005360, -0.00000983, 0.00000679, -0.00021414, -0.00029061, 0.00028555, 0.00080530, 0.00032501, -0.00039351, -0.00025650, 0.00011516, -0.00003674, -0.00013599, 0.00007591, 0.00003772, -0.00011986, 0.00000944, 0.00007013, -0.00007514, -0.00003067, 0.00007365, -0.00002802, -0.00004559, 0.00005945, -0.00000727, -0.00006771, 0.00004223, 0.00003320, -0.00006646, 0.00000460, 0.00005615, -0.00003902, -0.00002881, 0.00004821, -0.00000617, -0.00003271, 0.00003565, 0.00000539, -0.00005234, 0.00000870, 0.00003760, -0.00001772, -0.00000896, 0.00002119, -0.00001601, -0.00002550, 0.00001001, 0.00000601, -0.00000133, 0.00001814, 0.00000372, -0.00002086, -0.00000349, 0.00000683, -0.00000881, -0.00000244, 0.00001229, 0.00000146, -0.00001081, -0.00000794, -0.00000191, 0.00000483, 0.00000633, -0.00000047, -0.00000236, -0.00000146, -0.00000389, 0.00000277, 0.00001076, 0.00000271, -0.00000350, -0.00000034, -0.00000808, -0.00001031, 0.00001151, 0.00001777, -0.00000398, -0.00000558, 0.00001136, 0.00000407, -0.00000952, -0.00000023, 0.00000463, -0.00000750, -0.00000982, -0.00000253, -0.00000253, -0.00000345, -0.00000116, -0.00000086, 0.00000224, 0.00000474, -0.00000501, },
		{ 0.00000134, 0.00000420, -0.00000441, -0.00000541, 0.00000145, 0.00000101, -0.00000453, -0.00000002, 0.00001053, 0.00001105, 0.00000177, -0.00000631, -0.00000571, 0.00000216, 0.00000233, -0.00000636, -0.00000128, 0.00000891, -0.00000346, -0.00001304, 0.00000191, 0.00000250, -0.00001263, -0.00000373, 0.00000487, -0.00000922, -0.00000549, 0.00000866, -0.00000428, -0.00001338, -0.00000057, -0.00000017, -0.00002028, -0.00004220, -0.00002420, 0.00000565, -0.00012546, -0.00028725, 0.00004103, 0.00065673, 0.00056467, -0.00016790, -0.00037061, 0.00000797, 0.00005106, -0.00013324, -0.00000307, 0.00008472, -0.00008085, -0.00005680, 0.00007567, -0.00002311, -0.00007120, 0.00005238, 0.00001870, -0.00006168, 0.00003008, 0.00003414, -0.00006709, -0.00000059, 0.00005991, -0.00004270, -0.00003594, 0.00005876, -0.00000449, -0.00005278, 0.00003374, 0.00002426, -0.00004217, 0.00001401, 0.00003218, -0.00004667, -0.00002144, 0.00004570, 0.00000080, -0.00002534, 0.00001954, 0.00000168, -0.00003372, 0.00000068, 0.00001559, -0.00000575, 0.00001046, 0.00001649, -0.00001714, -0.00001463, 0.00001031, -0.00000250, -0.00001085, 0.00000981, 0.00000956, -0.00000931, -0.00001041, -0.00000288, 0.00000137, 0.00000553, 0.00000315, -0.00000152, -0.00000131, -0.00000344, -0.00000162, 0.00000873, 0.00000762, -0.00000232, -0.00000147, -0.00000444, -0.00001342, 0.00000201, 0.00002176, 0.00000559, -0.00001101, 0.00000562, 0.00001211, -0.00000675, -0.00000765, 0.00000567, -0.00000085, -0.00001147, -0.00000634, -0.00000183, -0.00000303, -0.00000118, -0.00000050, 0.00000020, 0.00000434, -0.00000084, -0.00001400, -0.00001268, },
		{ -0.00000226, -0.00000732, -0.00000030, 0.00000302, -0.00000300, -0.00000304, 0.00000769, 0.00001282, 0.00000465, -0.00000526, -0.00000639, 0.00000040, 0.00000335, -0.00000396, -0.00000437, 0.00000683, 0.00000217, -0.00001236, -0.00000356, 0.00000559, -0.00000925, -0.00000912, 0.00000632, -0.00000460, -0.00001218, 0.00000760, 0.00000404, -0.00001636, -0.00000433, 0.00000627, -0.00001742, -0.00003785, -0.00002558, -0.00000259, -0.00007424, -0.00023556, -0.00010876, 0.00043724, 0.00065758, 0.00010764, -0.00035874, -0.00014012, 0.00007632, -0.00007622, -0.00007199, 0.00007926, -0.00001661, -0.00009285, 0.00004020, 0.00002885, -0.00007592, 0.00000770, 0.00005028, -0.00004733, -0.00000973, 0.00005422, -0.00003907, -0.00004107, 0.00005798, -0.00000233, -0.00006020, 0.00003562, 0.00003192, -0.00005526, 0.00000275, 0.00004715, -0.00003067, -0.00001600, 0.00004418, -0.00002237, -0.00004840, 0.00003506, 0.00002692, -0.00003146, 0.00000472, 0.00001967, -0.00003071, -0.00001608, 0.00002186, -0.00000067, -0.00000232, 0.00002227, -0.00000434, -0.00002446, 0.00000553, 0.00000802, -0.00001424, 0.00000097, 0.00001602, -0.00000349, -0.00001353, -0.00000393, -0.00000018, 0.00000212, 0.00000514, 0.00000139, -0.00000135, -0.00000259, -0.00000385, 0.00000423, 0.00001018, 0.00000160, -0.00000279, -0.00000230, -0.00001216, -0.00000748, 0.00001854, 0.00001682, -0.00000949, -0.00000423, 0.00001521, 0.00000191, -0.00001274, 0.00000073, 0.00000570, -0.00000856, -0.00001086, -0.00000307, -0.00000207, -0.00000148, 0.00000019, -0.00000061, 0.00000246, 0.00000212, -0.00001054, -0.00001579, -0.00000236, 0.00000603, },
		{ -0.00000299, 0.00000403, -0.00000061, -0.00000464, 0.00000401, 0.00001306, 0.00000795, -0.00000360, -0.00000698, -0.00000113, 0.00000341, -0.00000149, -0.00000538, 0.00000331, 0.00000549, -0.00000863, -0.00000832, 0.00000505, -0.00000371, -0.00001258, 0.00000340, 0.00000213, -0.00001472, 0.00000046, 0.00001167, -0.00001273, -0.00001264, 0.00000991, -0.00000892, -0.00003823, -0.00002554, -0.00000468, -0.00004994, -0.00017914, -0.00017066, 0.00022492, 0.00061185, 0.00034333, -0.00022436, -0.00025751, 0.00002372, -0.00000625, -0.00009761, 0.00003267, 0.00003721, -0.00008545, -0.00001360, 0.00005453, -0.00004744, -0.00003596, 0.00005244, -0.00001332, -0.00003906, 0.00004641, 0.00000066, -0.00006086, 0.00003006, 0.00003509, -0.00005759, -0.00000223, 0.00005321, -0.00003445, -0.00003069, 0.00005046, -0.00000252, -0.00003949, 0.00003480, 0.00001040, -0.00005909, 0.00000705, 0.00004823, -0.00002077, -0.00001715, 0.00002813, -0.00001491, -0.00003293, 0.00001789, 0.00001255, -0.00001213, 0.00001676, 0.00001240, -0.00002596, -0.00000726, 0.00001641, -0.00000908, -0.00001067, 0.00001644, 0.00000631, -0.00001484, -0.00000700, 0.00000051, -0.00000144, 0.00000360, 0.00000522, -0.00000001, -0.00000239, -0.00000408, -0.00000029, 0.00000907, 0.00000643, -0.00000211, -0.00000235, -0.00000879, -0.00001328, 0.00000951, 0.00002424, -0.00000054, -0.00001270, 0.00001059, 0.00001226, -0.00001127, -0.00000799, 0.00000791, -0.00000151, -0.00001327, -0.00000683, -0.00000135, -0.00000145, 0.00000041, -0.00000011, 0.00000047, 0.00000292, -0.00000601, -0.00001667, -0.00000745, 0.00000627, 0.00000107, -0.00000823, },
		{ 0.00000212, -0.00000481, 0.00000038, 0.00001173, 0.00001088, -0.00000105, -0.00000731, -0.00000263, 0.00000300, 0.00000042, -0.00000480, 0.00000006, 0.00000603, -0.00000381, -0.00001039, 0.00000156, 0.00000122, -0.00001223, -0.00000236, 0.00000681, -0.00001141, -0.00000863, 0.00001338, -0.00000316, -0.00001931, 0.00000617, 0.00000311, -0.00003603, -0.00003074, -0.00000179, -0.00003497, -0.00013814, -0.00017802, 0.00006781, 0.00047911, 0.00047829, -0.00001985, -0.00028804, -0.00007939, 0.00003043, -0.00007543, -0.00002318, 0.00005551, -0.00004583, -0.00005624, 0.00004462, -0.00000543, -0.00005738, 0.00002776, 0.00002034, -0.00004495, 0.00001893, 0.00003238, -0.00005323, -0.00000871, 0.00005244, -0.00003106, -0.00003668, 0.00004991, -0.00000031, -0.00005085, 0.00003183, 0.00002899, -0.00004425, 0.00000823, 0.00003646, -0.00004750, -0.00002754, 0.00005275, 0.00000474, -0.00003455, 0.00002097, 0.00000787, -0.00004041, 0.00000190, 0.00002657, -0.00001155, 0.00000201, 0.00002445, -0.00001581, -0.00002232, 0.00001621, 0.00000344, -0.00001881, 0.00000865, 0.00001660, -0.00001129, -0.00001232, 0.00000146, -0.00000238, -0.00000116, 0.00000730, 0.00000371, -0.00000235, -0.00000375, -0.00000296, 0.00000513, 0.00000947, 0.00000134, -0.00000314, -0.00000609, -0.00001432, -0.00000122, 0.00002430, 0.00001217, -0.00001449, -0.00000010, 0.00001839, -0.00000249, -0.00001537, 0.00000340, 0.00000641, -0.00001098, -0.00001195, -0.00000236, -0.00000067, 0.00000008, 0.00000083, -0.00000036, 0.00000188, -0.00000213, -0.00001483, -0.00001249, 0.00000405, 0.00000482, -0.00000746, -0.00000623, 0.00000377, },
		{ -0.00000272, 0.00000925, 0.00001282, 0.00000217, -0.00000697, -0.00000421, 0.00000226, 0.00000174, -0.00000346, -0.00000200, 0.00000464, 0.00000012, -0.00000949, -0.00000278, 0.00000346, -0.00000864, -0.00000777, 0.00000699, -0.00000438, -0.00001454, 0.00000804, 0.00000726, -0.00001918, -0.00000363, 0.00001220, -0.00002657, -0.00003925, -0.00000177, -0.00001873, -0.00011113, -0.00016582, -0.00002465, 0.00032322, 0.00049953, 0.00018175, -0.00021667, -0.00017873, 0.00000924, -0.00003009, -0.00005637, 0.00003656, -0.00000062, -0.00006843, 0.00001004, 0.00002593, -0.00004992, -0.00000734, 0.00003627, -0.00002857, -0.00001119, 0.00004297, -0.00002547, -0.00003931, 0.00004373, 0.00000489, -0.00005216, 0.00002500, 0.00003072, -0.00004873, 0.00000008, 0.00004861, -0.00002750, -0.00002235, 0.00004355, -0.00001846, -0.00005356, 0.00003573, 0.00003441, -0.00003670, 0.00000016, 0.00002674, -0.00003266, -0.00002055, 0.00003192, 0.00000127, -0.00001391, 0.00002453, 0.00000279, -0.00003115, 0.00000510, 0.00001717, -0.00001782, -0.00000512, 0.00002199, -0.00000173, -0.00001749, -0.00000020, 0.00000022, -0.00000643, 0.00000514, 0.00000881, -0.00000073, -0.00000410, -0.00000362, 0.00000072, 0.00000895, 0.00000613, -0.00000238, -0.00000537, -0.00001230, -0.00000935, 0.00001724, 0.00002279, -0.00000767, -0.00001139, 0.00001616, 0.00000988, -0.00001618, -0.00000634, 0.00001046, -0.00000367, -0.00001554, -0.00000617, 0.00000027, -0.00000003, 0.00000121, 0.00000020, 0.00000038, -0.00000019, -0.00001108, -0.00001573, -0.00000056, 0.00000755, -0.00000464, -0.00000923, 0.00000195, 0.00000476, -0.00000370, },
		{ 0.00001362, 0.00000550, -0.00000573, -0.00000567, 0.00000112, 0.00000257, -0.00000192, -0.00000279, 0.00000260, 0.00000220, -0.00000678, -0.00000590, 0.00000271, -0.00000404, -0.00001032, 0.00000321, 0.00000229, -0.00001455, -0.00000111, 0.00001295, -0.00001182, -0.00001373, 0.00001326, -0.00001153, -0.00004425, -0.00000972, -0.00000180, -0.00008742, -0.00015389, -0.00007044, 0.00019144, 0.00043632, 0.00032069, -0.00007297, -0.00022221, -0.00005949, 0.00000131, -0.00005388, -0.00000007, 0.00002522, -0.00005054, -0.00002666, 0.00003197, -0.00002345, -0.00003326, 0.00002913, -0.00000292, -0.00002844, 0.00003200, 0.00000638, -0.00004892, 0.00001626, 0.00003244, -0.00004355, -0.00000809, 0.00004447, -0.00002675, -0.00002958, 0.00004722, 0.00000250, -0.00004172, 0.00002894, 0.00001447, -0.00005927, 0.00000303, 0.00005401, -0.00001992, -0.00002437, 0.00003159, -0.00001153, -0.00003895, 0.00002275, 0.00002073, -0.00002115, 0.00001150, 0.00002110, -0.00002737, -0.00001292, 0.00002401, -0.00000645, -0.00001854, 0.00001827, 0.00001138, -0.00001848, -0.00000597, 0.00000447, -0.00000857, -0.00000151, 0.00001235, 0.00000389, -0.00000478, -0.00000368, -0.00000200, 0.00000530, 0.00000955, 0.00000114, -0.00000566, -0.00001011, -0.00001286, 0.00000673, 0.00002661, 0.00000505, -0.00001702, 0.00000595, 0.00001939, -0.00000840, -0.00001622, 0.00000724, 0.00000575, -0.00001443, -0.00001209, -0.00000028, 0.00000083, 0.00000082, 0.00000124, -0.00000014, -0.00000017, -0.00000724, -0.00001614, -0.00000622, 0.00000784, -0.00000028, -0.00001084, -0.00000144, 0.00000668, -0.00000135, -0.00000505, 0.00000605, },
		{ -0.00000374, -0.00000665, -0.00000039, 0.00000291, -0.00000044, -0.00000267, 0.00000086, 0.00000261, -0.00000383, -0.00000692, 0.00000024, -0.00000065, -0.00000954, -0.00000192, 0.00000548, -0.00000969, -0.00000903, 0.00001157, -0.00000134, -0.00001839, 0.00000638, 0.00000280, -0.00004047, -0.00002360, 0.00000894, -0.00005973, -0.00014441, -0.00009494, 0.00010107, 0.00033708, 0.00037300, 0.00008733, -0.00018595, -0.00013732, -0.00000759, -0.00002782, -0.00002806, 0.00002268, -0.00001959, -0.00004568, 0.00001453, 0.00000334, -0.00003868, 0.00000711, 0.00001612, -0.00002719, 0.00001002, 0.00002707, -0.00003679, -0.00001437, 0.00003987, -0.00001808, -0.00003302, 0.00003640, 0.00000294, -0.00004379, 0.00002658, 0.00003107, -0.00004069, 0.00000086, 0.00003599, -0.00004272, -0.00003145, 0.00005323, 0.00000962, -0.00003964, 0.00001904, 0.00001357, -0.00004354, 0.00000086, 0.00003632, -0.00001392, -0.00000832, 0.00002937, -0.00001099, -0.00002902, 0.00001860, 0.00001063, -0.00002421, 0.00000543, 0.00002240, -0.00001226, -0.00001427, 0.00000660, -0.00000549, -0.00000995, 0.00001103, 0.00001095, -0.00000389, -0.00000467, -0.00000255, 0.00000077, 0.00000940, 0.00000643, -0.00000464, -0.00000955, -0.00001276, -0.00000259, 0.00002258, 0.00001782, -0.00001350, -0.00000723, 0.00002055, 0.00000510, -0.00002008, -0.00000277, 0.00001219, -0.00000742, -0.00001730, -0.00000389, 0.00000233, 0.00000057, 0.00000157, 0.00000070, -0.00000079, -0.00000484, -0.00001420, -0.00001095, 0.00000522, 0.00000414, -0.00001010, -0.00000577, 0.00000706, 0.00000208, -0.00000628, 0.00000233, 0.00001339, 0.00000856, },
		{ -0.00000202, 0.00000266, 0.00000084, -0.00000199, -0.00000024, 0.00000208, -0.00000167, -0.00000623, -0.00000226, 0.00000056, -0.00000678, -0.00000573, 0.00000460, -0.00000337, -0.00001225, 0.00000506, 0.00000674, -0.00001570, -0.00000369, 0.00001045, -0.00002833, -0.00003633, 0.00000820, -0.00002958, -0.00013017, -0.00011455, 0.00004402, 0.00024345, 0.00035241, 0.00021089, -0.00008391, -0.00017972, -0.00005657, -0.00000580, -0.00003176, 0.00000176, 0.00000333, -0.00004111, -0.00001146, 0.00001502, -0.00002595, -0.00001478, 0.00001935, -0.00001319, -0.00000904, 0.00003002, -0.00001288, -0.00003337, 0.00002700, 0.00000953, -0.00003934, 0.00001327, 0.00002549, -0.00003795, -0.00000217, 0.00004503, -0.00002076, -0.00002607, 0.00003708, -0.00001273, -0.00005279, 0.00003147, 0.00003856, -0.00003669, -0.00000510, 0.00003040, -0.00003106, -0.00002454, 0.00003844, 0.00000574, -0.00002407, 0.00002251, 0.00001085, -0.00003421, 0.00000192, 0.00002476, -0.00001794, -0.00001140, 0.00002530, 0.00000066, -0.00002072, 0.00000342, 0.00000166, -0.00001568, 0.00000352, 0.00001742, 0.00000063, -0.00000663, -0.00000229, -0.00000210, 0.00000560, 0.00001073, -0.00000065, -0.00001012, -0.00001187, -0.00000781, 0.00001359, 0.00002495, -0.00000225, -0.00001642, 0.00001218, 0.00001769, -0.00001447, -0.00001480, 0.00001118, 0.00000321, -0.00001807, -0.00001051, 0.00000271, 0.00000157, 0.00000088, 0.00000196, -0.00000050, -0.00000411, -0.00001157, -0.00001336, 0.00000069, 0.00000687, -0.00000700, -0.00000970, 0.00000524, 0.00000594, -0.00000585, -0.00000174, 0.00001210, 0.00001127, -0.00000131, -0.00000698, },
		{ 0.00000171, -0.00000098, -0.00000066, 0.00000128, -0.00000053, -0.00000478, -0.00000376, 0.00000001, -0.00000392, -0.00000699, 0.00000134, 0.00000110, -0.00001053, -0.00000231, 0.00000912, -0.00000824, -0.00001111, 0.00000971, -0.00001319, -0.00004133, -0.00000295, -0.00000459, -0.00010549, -0.00013151, 0.00000220, 0.00017477, 0.00029448, 0.00026973, 0.00004262, -0.00016013, -0.00011922, -0.00001217, -0.00001546, -0.00001765, 0.00000662, -0.00002227, -0.00002892, 0.00000811, -0.00000757, -0.00002453, 0.00000859, 0.00000173, -0.00001629, 0.00001919, 0.00000913, -0.00003427, 0.00000375, 0.00002566, -0.00002734, -0.00001146, 0.00003141, -0.00001796, -0.00002503, 0.00003952, 0.00000728, -0.00003881, 0.00001942, 0.00001590, -0.00005276, -0.00000170, 0.00005333, -0.00001575, -0.00002858, 0.00003061, -0.00000703, -0.00004156, 0.00002404, 0.00002847, -0.00002635, 0.00000344, 0.00002731, -0.00002460, -0.00001856, 0.00002791, -0.00000162, -0.00002409, 0.00001729, 0.00001555, -0.00002056, -0.00000517, 0.00000877, -0.00001502, -0.00000810, 0.00001914, 0.00000906, -0.00000764, -0.00000319, -0.00000229, 0.00000032, 0.00001133, 0.00000578, -0.00000948, -0.00001245, -0.00000925, 0.00000443, 0.00002409, 0.00001127, -0.00001666, -0.00000159, 0.00002266, -0.00000105, -0.00002202, 0.00000196, 0.00001222, -0.00001222, -0.00001753, -0.00000014, 0.00000383, 0.00000017, 0.00000222, 0.00000116, -0.00000386, -0.00000987, -0.00001353, -0.00000389, 0.00000694, -0.00000268, -0.00001181, 0.00000137, 0.00000907, -0.00000335, -0.00000560, 0.00000944, 0.00001364, 0.00000145, -0.00000681, -0.00000344, 0.00000189, },
		{ -0.00000047, 0.00000061, -0.00000017, -0.00000338, -0.00000412, -0.00000123, -0.00000216, -0.00000617, -0.00000198, 0.00000233, -0.00000625, -0.00000683, 0.00000628, -0.00000074, -0.00001275, 0.00000359, -0.00000140, -0.00003684, -0.00001788, 0.00000828, -0.00007219, -0.00013916, -0.00003747, 0.00012805, 0.00023443, 0.00026801, 0.00014778, -0.00008278, -0.00015860, -0.00005192, 0.00000079, -0.00002046, -0.00000581, -0.00000547, -0.00002947, -0.00000868, 0.00000304, -0.00002004, -0.00000629, 0.00000794, -0.00001203, 0.00000467, 0.00001982, -0.00002086, -0.00001650, 0.00002478, -0.00000654, -0.00002557, 0.00002099, 0.00000407, -0.00003280, 0.00001976, 0.00002969, -0.00003253, -0.00000618, 0.00003032, -0.00003382, -0.00003153, 0.00004723, 0.00001366, -0.00003926, 0.00001445, 0.00001697, -0.00004216, -0.00000131, 0.00004256, -0.00001260, -0.00001806, 0.00002975, -0.00000406, -0.00003257, 0.00001731, 0.00001716, -0.00002577, 0.00000102, 0.00002551, -0.00001178, -0.00001555, 0.00001105, -0.00000754, -0.00001892, 0.00001332, 0.00001881, -0.00000496, -0.00000602, -0.00000097, -0.00000347, 0.00000747, 0.00001191, -0.00000544, -0.00001445, -0.00000957, -0.00000130, 0.00001719, 0.00002071, -0.00000782, -0.00001351, 0.00001714, 0.00001382, -0.00001950, -0.00001143, 0.00001419, -0.00000112, -0.00002082, -0.00000691, 0.00000564, 0.00000098, 0.00000096, 0.00000315, -0.00000246, -0.00000945, -0.00001291, -0.00000697, 0.00000485, 0.00000108, -0.00001148, -0.00000346, 0.00001021, 0.00000096, -0.00000821, 0.00000520, 0.00001529, 0.00000477, -0.00000631, -0.00000461, 0.00000092, 0.00000200, 0.00000010, },
		{ -0.00000027, -0.00000237, -0.00000374, -0.00000229, -0.00000166, -0.00000453, -0.00000385, 0.00000105, -0.00000237, -0.00000735, 0.00000136, 0.00000320, -0.00000951, -0.00000288, 0.00000374, -0.00002629, -0.00002909, 0.00000774, -0.00003842, -0.00013115, -0.00007645, 0.00008929, 0.00019118, 0.00023211, 0.00020297, 0.00002174, -0.00014824, -0.00010700, -0.00000438, -0.00000670, -0.00001940, -0.00000221, -0.00001740, -0.00002190, 0.00000019, -0.00000876, -0.00001494, 0.00000360, -0.00000352, -0.00000449, 0.00001816, -0.00000319, -0.00002461, 0.00001111, 0.00001069, -0.00002440, 0.00000291, 0.00001699, -0.00002529, -0.00000285, 0.00003725, -0.00001239, -0.00002614, 0.00002607, -0.00000716, -0.00004587, 0.00002386, 0.00003785, -0.00003170, -0.00000935, 0.00002954, -0.00002653, -0.00002622, 0.00004034, 0.00001118, -0.00003051, 0.00001665, 0.00001747, -0.00003268, -0.00000256, 0.00002895, -0.00001482, -0.00001619, 0.00002515, 0.00000306, -0.00002209, 0.00000610, 0.00000332, -0.00002370, 0.00000062, 0.00002525, 0.00000290, -0.00000927, -0.00000055, -0.00000393, 0.00000096, 0.00001442, 0.00000215, -0.00001556, -0.00001148, -0.00000311, 0.00000892, 0.00002243, 0.00000538, -0.00001730, 0.00000396, 0.00002247, -0.00000736, -0.00002183, 0.00000686, 0.00001021, -0.00001712, -0.00001552, 0.00000435, 0.00000385, -0.00000072, 0.00000374, 0.00000061, -0.00000908, -0.00001308, -0.00000841, 0.00000219, 0.00000309, -0.00000932, -0.00000769, 0.00000880, 0.00000597, -0.00000847, -0.00000026, 0.00001540, 0.00000882, -0.00000532, -0.00000579, 0.00000004, 0.00000179, 0.00000087, 0.00000024, 0.00000029, },
		{ -0.00000304, -0.00000285, -0.00000196, -0.00000314, -0.00000408, -0.00000105, -0.00000058, -0.00000522, -0.00000236, 0.00000302, -0.00000476, -0.00000614, 0.00000322, -0.00001522, -0.00003263, -0.00000154, -0.00001300, -0.00010783, -0.00010741, 0.00004690, 0.00016414, 0.00019220, 0.00020657, 0.00011407, -0.00008627, -0.00014666, -0.00003869, 0.00000893, -0.00002057, -0.00001200, -0.00000495, -0.00002291, -0.00001152, -0.00000129, -0.00001343, -0.00000561, 0.00000117, -0.00000550, 0.00001002, 0.00000925, -0.00001972, -0.00000529, 0.00001631, -0.00001225, -0.00001177, 0.00001659, -0.00001014, -0.00001761, 0.00002942, 0.00001036, -0.00003149, 0.00000850, 0.00001380, -0.00004113, -0.00000527, 0.00004618, -0.00000989, -0.00002843, 0.00002558, -0.00000298, -0.00003962, 0.00002234, 0.00003360, -0.00002693, -0.00000534, 0.00002924, -0.00001852, -0.00002211, 0.00002753, 0.00000386, -0.00002595, 0.00001392, 0.00001762, -0.00002038, -0.00000441, 0.00001196, -0.00001979, -0.00001457, 0.00002405, 0.00001448, -0.00000982, -0.00000277, -0.00000161, -0.00000482, 0.00001154, 0.00001078, -0.00001292, -0.00001559, -0.00000350, 0.00000357, 0.00001727, 0.00001620, -0.00001116, -0.00000977, 0.00002031, 0.00000877, -0.00002290, -0.00000682, 0.00001552, -0.00000664, -0.00002163, -0.00000164, 0.00000729, -0.00000080, 0.00000203, 0.00000417, -0.00000696, -0.00001438, -0.00000955, 0.00000041, 0.00000348, -0.00000676, -0.00001029, 0.00000538, 0.00001009, -0.00000594, -0.00000570, 0.00001312, 0.00001319, -0.00000321, -0.00000720, -0.00000080, 0.00000148, 0.00000104, 0.00000118, 0.00000054, -0.00000058, -0.00000181, },
		{ -0.00000255, -0.00000240, -0.00000333, -0.00000257, -0.00000085, -0.00000270, -0.00000338, 0.00000064, -0.00000167, -0.00000580, 0.00000051, -0.00000761, -0.00002938, -0.00001247, -0.00000025, -0.00007638, -0.00012216, -0.00000081, 0.00014116, 0.00016670, 0.00017856, 0.00016595, 0.00000466, -0.00014523, -0.00008999, 0.00000759, -0.00000693, -0.00002402, -0.00000269, -0.00001314, -0.00002175, -0.00000371, -0.00000590, -0.00001165, -0.00000122, -0.00000198, 0.00000284, 0.00001263, -0.00000791, -0.00001518, 0.00001022, 0.00000153, -0.00001609, 0.00000669, 0.00000271, -0.00002008, 0.00001330, 0.00002506, -0.00002189, -0.00001119, 0.00002049, -0.00002322, -0.00002720, 0.00003643, 0.00001523, -0.00003356, 0.00000873, 0.00001686, -0.00003657, -0.00000305, 0.00004396, -0.00000863, -0.00002488, 0.00002551, 0.00000284, -0.00003173, 0.00001340, 0.00002124, -0.00002337, -0.00000322, 0.00002506, -0.00001009, -0.00001535, 0.00001373, -0.00000880, -0.00002609, 0.00001387, 0.00002546, -0.00000496, -0.00000715, 0.00000098, -0.00000699, 0.00000430, 0.00001659, -0.00000540, -0.00001976, -0.00000573, 0.00000250, 0.00000967, 0.00001977, 0.00000124, -0.00001667, 0.00000856, 0.00002066, -0.00001304, -0.00001980, 0.00001105, 0.00000623, -0.00002098, -0.00001107, 0.00000827, 0.00000200, -0.00000095, 0.00000600, -0.00000234, -0.00001537, -0.00001201, -0.00000040, 0.00000358, -0.00000492, -0.00001137, 0.00000120, 0.00001220, -0.00000124, -0.00000951, 0.00000827, 0.00001672, 0.00000062, -0.00000854, -0.00000207, 0.00000145, 0.00000070, 0.00000183, 0.00000145, -0.00000079, -0.00000173, -0.00000227, -0.00000292, },
		{ -0.00000231, -0.00000303, -0.00000212, -0.00000140, -0.00000228, -0.00000137, -0.00000124, -0.00000391, -0.00000131, -0.00000407, -0.00002306, -0.00001988, 0.00000143, -0.00004627, -0.00011809, -0.00004655, 0.00011046, 0.00015681, 0.00014566, 0.00017159, 0.00009046, -0.00009645, -0.00013251, -0.00002023, 0.00001062, -0.00002598, -0.00001207, -0.00000193, -0.00002237, -0.00001376, -0.00000054, -0.00001031, -0.00000765, 0.00000011, 0.00000050, 0.00000945, 0.00000286, -0.00001503, -0.00000117, 0.00000867, -0.00001043, -0.00000437, 0.00000696, -0.00001335, -0.00000154, 0.00002711, -0.00000451, -0.00002265, 0.00001300, -0.00000336, -0.00003421, 0.00001521, 0.00003202, -0.00002336, -0.00001117, 0.00002432, -0.00002050, -0.00002440, 0.00003786, 0.00001567, -0.00003206, 0.00000858, 0.00002067, -0.00002702, -0.00000639, 0.00002893, -0.00000970, -0.00001814, 0.00002172, 0.00000455, -0.00002093, 0.00000761, 0.00000412, -0.00002911, -0.00000253, 0.00003055, 0.00000558, -0.00001089, 0.00000112, -0.00000494, -0.00000390, 0.00001656, 0.00000523, -0.00002049, -0.00001131, 0.00000356, 0.00000457, 0.00001569, 0.00001307, -0.00001324, -0.00000625, 0.00002207, 0.00000333, -0.00002461, -0.00000173, 0.00001479, -0.00001237, -0.00001973, 0.00000421, 0.00000671, -0.00000271, 0.00000460, 0.00000347, -0.00001382, -0.00001609, -0.00000172, 0.00000469, -0.00000368, -0.00001192, -0.00000270, 0.00001213, 0.00000418, -0.00001051, 0.00000179, 0.00001794, 0.00000606, -0.00000890, -0.00000424, 0.00000163, 0.00000033, 0.00000173, 0.00000281, -0.00000048, -0.00000212, -0.00000161, -0.00000252, -0.00000317, -0.00000231, },
		{ -0.00000332, -0.00000160, -0.00000067, -0.00000166, -0.00000233, -0.00000274, -0.00000136, -0.00000280, -0.00001713, -0.00002242, -0.00000284, -0.00002419, -0.00009966, -0.00008057, 0.00006857, 0.00015175, 0.00012653, 0.00014590, 0.00014324, -0.00001647, -0.00014146, -0.00006711, 0.00001532, -0.00001379, -0.00002489, 0.00000100, -0.00001315, -0.00002332, -0.00000312, -0.00000386, -0.00001210, -0.00000255, 0.00000168, 0.00000508, 0.00000780, -0.00000771, -0.00000991, 0.00000662, -0.00000091, -0.00000951, 0.00000290, -0.00000472, -0.00000879, 0.00001928, 0.00001104, -0.00002165, -0.00000140, 0.00000850, -0.00002710, -0.00000642, 0.00003421, -0.00000431, -0.00002386, 0.00001799, -0.00000069, -0.00003325, 0.00001909, 0.00003469, -0.00002351, -0.00001258, 0.00002638, -0.00001102, -0.00002211, 0.00002364, 0.00000826, -0.00002384, 0.00000933, 0.00001683, -0.00001804, -0.00000321, 0.00001304, -0.00002261, -0.00001926, 0.00002631, 0.00001874, -0.00001047, -0.00000217, -0.00000065, -0.00000923, 0.00001039, 0.00001492, -0.00001537, -0.00001871, 0.00000295, 0.00000440, 0.00000789, 0.00001831, -0.00000175, -0.00001597, 0.00001239, 0.00001788, -0.00001780, -0.00001637, 0.00001386, 0.00000084, -0.00002261, -0.00000475, 0.00001019, -0.00000110, 0.00000063, 0.00000772, -0.00000850, -0.00002009, -0.00000545, 0.00000669, -0.00000170, -0.00001274, -0.00000624, 0.00001057, 0.00000897, -0.00000858, -0.00000459, 0.00001597, 0.00001202, -0.00000717, -0.00000717, 0.00000137, 0.00000051, 0.00000084, 0.00000397, 0.00000073, -0.00000274, -0.00000140, -0.00000164, -0.00000361, -0.00000278, -0.00000144, -0.00000259, },
		{ 0.00000018, -0.00000048, -0.00000315, -0.00000311, -0.00000073, -0.00000177, -0.00001287, -0.00002164, -0.00000815, -0.00001178, -0.00007514, -0.00009711, 0.00002114, 0.00013893, 0.00012476, 0.00011281, 0.00015351, 0.00006535, -0.00010699, -0.00011176, -0.00000378, 0.00000486, -0.00002994, -0.00000722, -0.00000135, -0.00002476, -0.00001281, 0.00000119, -0.00001061, -0.00000841, 0.00000222, 0.00000332, 0.00000722, 0.00000077, -0.00001146, -0.00000114, 0.00000533, -0.00000688, -0.00000411, -0.00000051, -0.00000817, 0.00000857, 0.00001835, -0.00001128, -0.00001330, 0.00000871, -0.00001337, -0.00001926, 0.00002355, 0.00001353, -0.00002415, 0.00000360, 0.00001320, -0.00002817, -0.00000303, 0.00004055, -0.00000380, -0.00002735, 0.00001794, 0.00000765, -0.00002661, 0.00000872, 0.00002186, -0.00001807, -0.00000596, 0.00002118, -0.00000795, -0.00001317, 0.00001430, -0.00000987, -0.00003010, 0.00001311, 0.00002930, -0.00000389, -0.00000724, 0.00000268, -0.00000956, 0.00000077, 0.00001950, -0.00000485, -0.00002413, -0.00000204, 0.00000773, 0.00000189, 0.00001510, 0.00001122, -0.00001541, -0.00000292, 0.00002297, -0.00000222, -0.00002471, 0.00000315, 0.00001200, -0.00001704, -0.00001502, 0.00000899, 0.00000382, -0.00000320, 0.00000802, -0.00000054, -0.00002109, -0.00001216, 0.00000768, 0.00000235, -0.00001339, -0.00001016, 0.00000819, 0.00001253, -0.00000460, -0.00000921, 0.00001108, 0.00001671, -0.00000288, -0.00000988, -0.00000009, 0.00000135, -0.00000029, 0.00000422, 0.00000279, -0.00000299, -0.00000197, -0.00000047, -0.00000350, -0.00000375, -0.00000099, -0.00000156, -0.00000391, -0.00000272, },
		{ -0.00000270, -0.00000425, -0.00000084, -0.00000034, -0.00000964, -0.00001965, -0.00001203, -0.00000675, -0.00005200, -0.00009697, -0.00002204, 0.00011227, 0.00013162, 0.00009157, 0.00013132, 0.00012173, -0.00003933, -0.00013064, -0.00004347, 0.00001524, -0.00002158, -0.00002064, 0.00000395, -0.00001658, -0.00002304, -0.00000062, -0.00000435, -0.00001278, -0.00000085, 0.00000406, 0.00000472, 0.00000567, -0.00000658, -0.00000836, 0.00000470, -0.00000003, -0.00000800, -0.00000237, -0.00000418, 0.00000127, 0.00001693, 0.00000128, -0.00001673, 0.00000066, -0.00000206, -0.00002044, 0.00000766, 0.00002229, -0.00001406, -0.00001000, 0.00001620, -0.00001454, -0.00001895, 0.00003238, 0.00001758, -0.00002901, 0.00000055, 0.00001961, -0.00001891, -0.00000792, 0.00002521, -0.00000432, -0.00001683, 0.00001615, 0.00000441, -0.00001741, 0.00000826, 0.00000335, -0.00003143, -0.00000475, 0.00003241, 0.00000778, -0.00001071, 0.00000263, -0.00000552, -0.00000803, 0.00001694, 0.00000758, -0.00002387, -0.00001105, 0.00001052, 0.00000143, 0.00000647, 0.00001868, -0.00000512, -0.00001536, 0.00001600, 0.00001426, -0.00002152, -0.00001200, 0.00001477, -0.00000494, -0.00002120, 0.00000196, 0.00000922, -0.00000384, 0.00000423, 0.00000686, -0.00001701, -0.00001999, 0.00000510, 0.00000829, -0.00001215, -0.00001501, 0.00000481, 0.00001503, 0.00000033, -0.00001130, 0.00000463, 0.00001862, 0.00000329, -0.00001096, -0.00000303, 0.00000224, -0.00000086, 0.00000325, 0.00000505, -0.00000216, -0.00000329, 0.00000043, -0.00000248, -0.00000497, -0.00000128, -0.00000019, -0.00000375, -0.00000409, -0.00000008, 0.00000106, },
		{ -0.00000204, 0.00000075, -0.00000653, -0.00001738, -0.00001431, -0.00000578, -0.00003414, -0.00008565, -0.00005312, 0.00007442, 0.00013446, 0.00008876, 0.00009747, 0.00013989, 0.00003703, -0.00011106, -0.00008694, 0.00000551, -0.00000447, -0.00002910, -0.00000140, -0.00000430, -0.00002653, -0.00000952, 0.00000134, -0.00001195, -0.00000676, 0.00000438, 0.00000363, 0.00000610, 0.00000041, -0.00001042, -0.00000125, 0.00000530, -0.00000602, -0.00000723, -0.00000209, -0.00000055, 0.00001108, 0.00000956, -0.00001168, -0.00000849, 0.00000144, -0.00001371, -0.00000495, 0.00002020, -0.00000060, -0.00001612, 0.00000988, -0.00000067, -0.00002378, 0.00001585, 0.00003148, -0.00001778, -0.00001660, 0.00001969, -0.00000427, -0.00001832, 0.00001800, 0.00001033, -0.00001866, 0.00000495, 0.00001324, -0.00001436, -0.00000122, 0.00001180, -0.00002380, -0.00002106, 0.00002600, 0.00002061, -0.00000942, -0.00000092, 0.00000005, -0.00001245, 0.00000843, 0.00001715, -0.00001661, -0.00002069, 0.00000879, 0.00000632, -0.00000179, 0.00001707, 0.00000905, -0.00001818, 0.00000093, 0.00002302, -0.00000778, -0.00002323, 0.00000707, 0.00000759, -0.00001931, -0.00000839, 0.00001112, -0.00000019, -0.00000113, 0.00001031, -0.00000831, -0.00002533, -0.00000234, 0.00001377, -0.00000717, -0.00001998, -0.00000052, 0.00001662, 0.00000562, -0.00001103, -0.00000173, 0.00001731, 0.00000968, -0.00000935, -0.00000671, 0.00000218, -0.00000045, 0.00000147, 0.00000656, 0.00000005, -0.00000472, 0.00000032, -0.00000055, -0.00000576, -0.00000261, 0.00000113, -0.00000273, -0.00000527, -0.00000116, 0.00000206, -0.00000120, -0.00000486, },
		{ -0.00000353, -0.00001466, -0.00001555, -0.00000660, -0.00002200, -0.00006962, -0.00006952, 0.00003342, 0.00012487, 0.00009870, 0.00007135, 0.00012489, 0.00009620, -0.00005781, -0.00011267, -0.00002484, 0.00000904, -0.00002581, -0.00001367, 0.00000368, -0.00002063, -0.00002029, 0.00000118, -0.00000631, -0.00001175, 0.00000165, 0.00000446, 0.00000441, 0.00000522, -0.00000681, -0.00000776, 0.00000534, -0.00000007, -0.00001026, -0.00000425, 0.00000078, 0.00000603, 0.00001132, -0.00000280, -0.00001267, -0.00000236, -0.00000604, -0.00000967, 0.00001140, 0.00000890, -0.00001351, 0.00000028, 0.00000710, -0.00001886, -0.00000069, 0.00003359, 0.00000008, -0.00002543, 0.00000919, 0.00000913, -0.00001860, 0.00000514, 0.00001920, -0.00001167, -0.00000650, 0.00001501, -0.00000625, -0.00000925, 0.00001313, -0.00001124, -0.00003054, 0.00001188, 0.00002955, -0.00000236, -0.00000574, 0.00000392, -0.00001128, -0.00000229, 0.00002017, -0.00000439, -0.00002643, 0.00000121, 0.00001285, -0.00000459, 0.00000803, 0.00001948, -0.00000996, -0.00001397, 0.00001949, 0.00000966, -0.00002388, -0.00000728, 0.00001358, -0.00000968, -0.00001679, 0.00000708, 0.00000576, -0.00000436, 0.00000849, 0.00000178, -0.00002489, -0.00001310, 0.00001528, 0.00000174, -0.00002266, -0.00000850, 0.00001660, 0.00001134, -0.00000895, -0.00000691, 0.00001347, 0.00001454, -0.00000515, -0.00000972, 0.00000051, 0.00000062, -0.00000023, 0.00000662, 0.00000321, -0.00000526, -0.00000116, 0.00000166, -0.00000531, -0.00000486, 0.00000172, -0.00000081, -0.00000597, -0.00000275, 0.00000229, 0.00000051, -0.00000435, -0.00000353, 0.00000066, },
		{ -0.00001577, -0.00000825, -0.00001445, -0.00005349, -0.00007339, -0.00000256, 0.00010207, 0.00011008, 0.00006233, 0.00009367, 0.00012294, 0.00000982, -0.00010620, -0.00006347, 0.00000702, -0.00001260, -0.00002409, 0.00000217, -0.00000912, -0.00002603, -0.00000604, -0.00000018, -0.00001224, -0.00000390, 0.00000511, 0.00000343, 0.00000631, -0.00000056, -0.00001045, 0.00000041, 0.00000536, -0.00000854, -0.00000901, 0.00000121, 0.00000445, 0.00000873, 0.00000436, -0.00001017, -0.00000864, -0.00000291, -0.00000753, 0.00000251, 0.00001095, -0.00000605, -0.00000639, 0.00000732, -0.00000981, -0.00001090, 0.00002574, 0.00001629, -0.00002292, -0.00000550, 0.00001483, -0.00001056, -0.00000645, 0.00001938, -0.00000027, -0.00001290, 0.00001004, 0.00000248, -0.00001246, 0.00000853, 0.00000097, -0.00003101, -0.00000520, 0.00003089, 0.00000857, -0.00000861, 0.00000422, -0.00000601, -0.00001087, 0.00001578, 0.00000837, -0.00002516, -0.00001005, 0.00001619, -0.00000040, -0.00000294, 0.00002090, 0.00000477, -0.00002069, 0.00000576, 0.00002171, -0.00001294, -0.00002034, 0.00000931, 0.00000270, -0.00001837, -0.00000170, 0.00000996, -0.00000317, 0.00000320, 0.00000903, -0.00001796, -0.00002315, 0.00001033, 0.00001210, -0.00002026, -0.00001826, 0.00001336, 0.00001739, -0.00000517, -0.00001068, 0.00000836, 0.00001693, 0.00000046, -0.00001071, -0.00000247, 0.00000140, -0.00000110, 0.00000526, 0.00000627, -0.00000411, -0.00000361, 0.00000301, -0.00000318, -0.00000723, 0.00000081, 0.00000174, -0.00000573, -0.00000472, 0.00000188, 0.00000187, -0.00000309, -0.00000428, -0.00000050, -0.00000149, -0.00001129, },
		{ -0.00001041, -0.00003947, -0.00006890, -0.00002879, 0.00007116, 0.00011339, 0.00006839, 0.00006462, 0.00011766, 0.00006840, -0.00006774, -0.00009169, -0.00001387, 0.00000091, -0.00002534, -0.00000723, 0.00000055, -0.00002311, -0.00001642, 0.00000134, -0.00000790, -0.00000925, 0.00000323, 0.00000406, 0.00000510, 0.00000451, -0.00000805, -0.00000589, 0.00000653, -0.00000290, -0.00001254, -0.00000166, 0.00000535, 0.00000579, 0.00000690, -0.00000365, -0.00001209, -0.00000541, -0.00000332, -0.00000172, 0.00000713, 0.00000060, -0.00000736, 0.00000311, -0.00000245, -0.00001320, 0.00001375, 0.00002497, -0.00001180, -0.00001689, 0.00001116, 0.00000009, -0.00001172, 0.00001266, 0.00000977, -0.00001218, 0.00000189, 0.00000784, -0.00001050, 0.00000146, 0.00000883, -0.00002389, -0.00001979, 0.00002391, 0.00001952, -0.00000718, 0.00000132, 0.00000035, -0.00001437, 0.00000624, 0.00001697, -0.00001687, -0.00002058, 0.00001321, 0.00000823, -0.00000975, 0.00001311, 0.00001831, -0.00001582, -0.00001075, 0.00002216, 0.00000418, -0.00002450, -0.00000311, 0.00001075, -0.00001191, -0.00001065, 0.00000904, 0.00000163, -0.00000179, 0.00001077, -0.00000718, -0.00002805, -0.00000055, 0.00001957, -0.00001157, -0.00002684, 0.00000545, 0.00002264, 0.00000077, -0.00001308, 0.00000282, 0.00001700, 0.00000588, -0.00000929, -0.00000563, 0.00000105, -0.00000109, 0.00000321, 0.00000819, -0.00000127, -0.00000589, 0.00000261, 0.00000017, -0.00000843, -0.00000189, 0.00000403, -0.00000402, -0.00000680, 0.00000079, 0.00000291, -0.00000183, -0.00000402, -0.00000168, -0.00000101, -0.00000792, -0.00001457, -0.00001014, },
		{ -0.00005989, -0.00004450, 0.00003900, 0.00010503, 0.00008077, 0.00004898, 0.00009291, 0.00010154, -0.00001138, -0.00009464, -0.00004547, 0.00000359, -0.00001690, -0.00001756, 0.00000261, -0.00001362, -0.00002368, -0.00000376, -0.00000209, -0.00001100, -0.00000142, 0.00000484, 0.00000399, 0.00000649, -0.00000256, -0.00000951, 0.00000294, 0.00000308, -0.00001195, -0.00000703, 0.00000566, 0.00000518, 0.00000565, 0.00000244, -0.00001030, -0.00001049, -0.00000183, -0.00000080, 0.00000208, 0.00000274, -0.00000419, -0.00000088, 0.00000042, -0.00001035, 0.00000352, 0.00002506, 0.00000182, -0.00002034, 0.00000148, 0.00000729, -0.00000984, 0.00000384, 0.00001451, -0.00000604, -0.00000503, 0.00000824, -0.00000555, -0.00000453, 0.00001100, -0.00001295, -0.00002798, 0.00001095, 0.00002645, -0.00000120, -0.00000266, 0.00000490, -0.00001232, -0.00000445, 0.00001865, -0.00000457, -0.00002606, 0.00000415, 0.00001637, -0.00000900, 0.00000056, 0.00002403, -0.00000206, -0.00002136, 0.00001102, 0.00001862, -0.00001695, -0.00001669, 0.00000956, -0.00000104, -0.00001454, 0.00000287, 0.00000652, -0.00000322, 0.00000758, 0.00000291, -0.00002555, -0.00001358, 0.00002021, 0.00000151, -0.00003031, -0.00000685, 0.00002468, 0.00000914, -0.00001363, -0.00000294, 0.00001545, 0.00001012, -0.00000617, -0.00000767, -0.00000050, -0.00000083, 0.00000130, 0.00000855, 0.00000246, -0.00000681, 0.00000035, 0.00000344, -0.00000742, -0.00000572, 0.00000476, -0.00000076, -0.00000828, -0.00000124, 0.00000381, -0.00000078, -0.00000344, -0.00000191, -0.00000159, -0.00000575, -0.00001203, -0.00001131, -0.00000910, -0.00002846, },
		{ 0.00001079, 0.00008689, 0.00009040, 0.00004814, 0.00006483, 0.00010588, 0.00004281, -0.00006912, -0.00007239, -0.00000959, -0.00000564, -0.00002164, -0.00000321, -0.00000370, -0.00002356, -0.00001292, 0.00000051, -0.00000820, -0.00000653, 0.00000371, 0.00000413, 0.00000602, 0.00000270, -0.00000864, -0.00000279, 0.00000581, -0.00000731, -0.00001181, 0.00000307, 0.00000667, 0.00000360, 0.00000511, -0.00000468, -0.00001376, -0.00000477, 0.00000207, -0.00000007, 0.00000059, -0.00000078, -0.00000175, -0.00000021, -0.00000662, -0.00000200, 0.00001961, 0.00001224, -0.00001589, -0.00000851, 0.00000796, -0.00000392, -0.00000244, 0.00001348, 0.00000163, -0.00000783, 0.00000490, -0.00000064, -0.00000748, 0.00000866, -0.00000231, -0.00002863, -0.00000380, 0.00002694, 0.00000746, -0.00000521, 0.00000619, -0.00000643, -0.00001236, 0.00001351, 0.00000720, -0.00002440, -0.00000773, 0.00001959, -0.00000143, -0.00001039, 0.00001966, 0.00001355, -0.00002088, -0.00000562, 0.00002296, -0.00000137, -0.00002344, -0.00000042, 0.00000758, -0.00001095, -0.00000503, 0.00000770, -0.00000066, 0.00000276, 0.00000854, -0.00001701, -0.00002356, 0.00001295, 0.00001436, -0.00002595, -0.00002060, 0.00002088, 0.00001887, -0.00001110, -0.00000893, 0.00001265, 0.00001314, -0.00000256, -0.00000807, -0.00000243, -0.00000106, -0.00000015, 0.00000764, 0.00000596, -0.00000576, -0.00000288, 0.00000520, -0.00000416, -0.00000912, 0.00000301, 0.00000322, -0.00000811, -0.00000429, 0.00000428, 0.00000046, -0.00000332, -0.00000123, -0.00000184, -0.00000540, -0.00000930, -0.00001076, -0.00000944, -0.00002092, -0.00004913, -0.00005118, },
		{ 0.00009212, 0.00005651, 0.00004519, 0.00008962, 0.00007870, -0.00002458, -0.00008057, -0.00003393, -0.00000118, -0.00001738, -0.00001184, 0.00000080, -0.00001662, -0.00002071, -0.00000271, -0.00000324, -0.00000900, 0.00000005, 0.00000468, 0.00000505, 0.00000558, -0.00000439, -0.00000702, 0.00000413, -0.00000137, -0.00001315, -0.00000222, 0.00000790, 0.00000329, 0.00000433, 0.00000127, -0.00001258, -0.00001023, 0.00000289, 0.00000175, -0.00000267, -0.00000031, 0.00000013, -0.00000130, -0.00000481, -0.00000351, 0.00001308, 0.00001692, -0.00000720, -0.00001414, 0.00000307, 0.00000149, -0.00000414, 0.00000919, 0.00000731, -0.00000633, 0.00000052, 0.00000217, -0.00000747, 0.00000417, 0.00000521, -0.00002318, -0.00001619, 0.00002098, 0.00001581, -0.00000472, 0.00000450, 0.00000053, -0.00001517, 0.00000416, 0.00001449, -0.00001653, -0.00001799, 0.00001606, 0.00000899, -0.00001482, 0.00000796, 0.00002366, -0.00000997, -0.00001918, 0.00001534, 0.00001409, -0.00001915, -0.00001339, 0.00000840, -0.00000217, -0.00000956, 0.00000399, 0.00000324, -0.00000006, 0.00000889, -0.00000660, -0.00002677, 0.00000053, 0.00002182, -0.00001429, -0.00003105, 0.00001029, 0.00002712, -0.00000432, -0.00001439, 0.00000829, 0.00001538, 0.00000081, -0.00000731, -0.00000371, -0.00000191, -0.00000142, 0.00000607, 0.00000844, -0.00000304, -0.00000571, 0.00000468, 0.00000026, -0.00001045, -0.00000101, 0.00000630, -0.00000561, -0.00000766, 0.00000356, 0.00000239, -0.00000369, -0.00000072, -0.00000067, -0.00000606, -0.00000803, -0.00000850, -0.00000969, -0.00001694, -0.00003877, -0.00005097, -0.00001086, 0.00006347, },
		{ 0.00003788, 0.00006587, 0.00009047, 0.00002265, -0.00006481, -0.00005734, -0.00000921, -0.00000922, -0.00001689, -0.00000171, -0.00000761, -0.00002266, -0.00001033, -0.00000028, -0.00000746, -0.00000447, 0.00000387, 0.00000487, 0.00000609, 0.00000051, -0.00000757, -0.00000024, 0.00000251, -0.00001046, -0.00000778, 0.00000663, 0.00000500, 0.00000233, 0.00000456, -0.00000749, -0.00001452, -0.00000030, 0.00000531, -0.00000357, -0.00000309, 0.00000206, -0.00000049, -0.00000493, -0.00000362, 0.00000841, 0.00001688, 0.00000124, -0.00001400, -0.00000370, 0.00000325, -0.00000239, 0.00000496, 0.00000955, -0.00000244, -0.00000253, 0.00000258, -0.00000581, -0.00000035, 0.00000867, -0.00001447, -0.00002354, 0.00001060, 0.00002110, -0.00000105, 0.00000131, 0.00000604, -0.00001269, -0.00000572, 0.00001542, -0.00000565, -0.00002315, 0.00000710, 0.00001744, -0.00001143, -0.00000532, 0.00002422, 0.00000567, -0.00002317, 0.00000015, 0.00002148, -0.00000584, -0.00002152, 0.00000051, 0.00000588, -0.00000757, -0.00000212, 0.00000480, 0.00000050, 0.00000632, 0.00000137, -0.00002299, -0.00001190, 0.00002103, 0.00000063, -0.00003394, -0.00000505, 0.00003015, 0.00000646, -0.00001746, 0.00000182, 0.00001687, 0.00000414, -0.00000636, -0.00000395, -0.00000269, -0.00000298, 0.00000415, 0.00000977, 0.00000051, -0.00000708, 0.00000223, 0.00000414, -0.00000894, -0.00000590, 0.00000690, -0.00000116, -0.00000993, 0.00000101, 0.00000475, -0.00000365, -0.00000154, 0.00000170, -0.00000589, -0.00000902, -0.00000600, -0.00000817, -0.00001550, -0.00003074, -0.00004605, -0.00002512, 0.00003911, 0.00008520, 0.00006670, },
		{ 0.00008260, 0.00005784, -0.00003089, -0.00006738, -0.00002744, -0.00000499, -0.00001546, -0.00000791, -0.00000193, -0.00001817, -0.00001793, -0.00000215, -0.00000348, -0.00000714, 0.00000093, 0.00000509, 0.00000567, 0.00000384, -0.00000480, -0.00000422, 0.00000268, -0.00000570, -0.00001085, 0.00000255, 0.00000700, 0.00000150, 0.00000455, -0.00000130, -0.00001486, -0.00000634, 0.00000715, -0.00000083, -0.00000673, 0.00000146, 0.00000220, -0.00000503, -0.00000451, 0.00000589, 0.00001492, 0.00000673, -0.00000985, -0.00000860, 0.00000109, -0.00000004, 0.00000279, 0.00000922, 0.00000152, -0.00000336, 0.00000159, -0.00000384, -0.00000375, 0.00000853, -0.00000536, -0.00002508, -0.00000120, 0.00002182, 0.00000465, -0.00000162, 0.00000862, -0.00000645, -0.00001276, 0.00001052, 0.00000433, -0.00002206, -0.00000389, 0.00002056, -0.00000246, -0.00001460, 0.00001600, 0.00001875, -0.00001640, -0.00001464, 0.00001741, 0.00000936, -0.00001967, -0.00001146, 0.00000738, -0.00000033, -0.00000593, 0.00000224, 0.00000270, 0.00000424, 0.00000501, -0.00001534, -0.00001969, 0.00001308, 0.00001315, -0.00002795, -0.00002040, 0.00002540, 0.00001871, -0.00001585, -0.00000636, 0.00001667, 0.00000812, -0.00000570, -0.00000376, -0.00000250, -0.00000475, 0.00000170, 0.00001014, 0.00000408, -0.00000672, -0.00000104, 0.00000616, -0.00000513, -0.00000965, 0.00000448, 0.00000368, -0.00000973, -0.00000310, 0.00000642, -0.00000212, -0.00000377, 0.00000364, -0.00000337, -0.00001128, -0.00000545, -0.00000458, -0.00001443, -0.00002623, -0.00003902, -0.00003210, 0.00001695, 0.00007155, 0.00007339, 0.00004035, 0.00004561, },
		{ 0.00000844, -0.00005809, -0.00004671, -0.00001020, -0.00001019, -0.00001255, -0.00000193, -0.00001064, -0.00002118, -0.00000837, -0.00000057, -0.00000641, -0.00000301, 0.00000434, 0.00000555, 0.00000513, -0.00000080, -0.00000551, 0.00000005, -0.00000181, -0.00001036, -0.00000256, 0.00000719, 0.00000258, 0.00000279, 0.00000310, -0.00001107, -0.00001206, 0.00000509, 0.00000395, -0.00000818, -0.00000219, 0.00000470, -0.00000329, -0.00000636, 0.00000399, 0.00001334, 0.00000921, -0.00000477, -0.00000989, -0.00000286, 0.00000050, 0.00000251, 0.00000816, 0.00000426, -0.00000255, 0.00000044, -0.00000221, -0.00000576, 0.00000595, 0.00000202, -0.00002164, -0.00001154, 0.00001792, 0.00001058, -0.00000294, 0.00000803, 0.00000114, -0.00001502, 0.00000219, 0.00001036, -0.00001591, -0.00001323, 0.00001763, 0.00000797, -0.00001690, 0.00000336, 0.00002413, -0.00000291, -0.00002202, 0.00000469, 0.00001844, -0.00000852, -0.00002005, 0.00000054, 0.00000696, -0.00000400, -0.00000267, 0.00000323, 0.00000447, 0.00000560, -0.00000798, -0.00002119, 0.00000226, 0.00001903, -0.00001568, -0.00003045, 0.00001332, 0.00002822, -0.00000843, -0.00001432, 0.00001338, 0.00001289, -0.00000474, -0.00000430, -0.00000109, -0.00000587, -0.00000160, 0.00000959, 0.00000733, -0.00000498, -0.00000395, 0.00000599, -0.00000054, -0.00001084, -0.00000008, 0.00000689, -0.00000672, -0.00000736, 0.00000605, 0.00000109, -0.00000611, 0.00000336, 0.00000129, -0.00001236, -0.00000809, -0.00000059, -0.00001136, -0.00002474, -0.00003284, -0.00003273, -0.00000070, 0.00005374, 0.00007428, 0.00004758, 0.00003398, 0.00006495, 0.00007409, },
		{ -0.00005648, -0.00002401, -0.00000716, -0.00001257, -0.00000571, -0.00000476, -0.00001872, -0.00001543, -0.00000158, -0.00000325, -0.00000565, 0.00000182, 0.00000568, 0.00000536, 0.00000237, -0.00000387, -0.00000294, -0.00000065, -0.00000751, -0.00000627, 0.00000487, 0.00000443, 0.00000140, 0.00000453, -0.00000525, -0.00001461, -0.00000039, 0.00000767, -0.00000591, -0.00000699, 0.00000469, 0.00000036, -0.00000767, 0.00000118, 0.00001257, 0.00001035, -0.00000095, -0.00000832, -0.00000596, -0.00000127, 0.00000249, 0.00000767, 0.00000593, -0.00000115, -0.00000032, -0.00000092, -0.00000646, 0.00000209, 0.00000648, -0.00001497, -0.00001842, 0.00001054, 0.00001502, -0.00000202, 0.00000505, 0.00000763, -0.00001221, -0.00000644, 0.00001110, -0.00000739, -0.00001843, 0.00001023, 0.00001606, -0.00001252, -0.00000840, 0.00002081, 0.00001076, -0.00001939, -0.00000966, 0.00001706, 0.00000579, -0.00001961, -0.00001096, 0.00000839, 0.00000312, -0.00000560, 0.00000010, 0.00000584, 0.00000606, -0.00000329, -0.00001826, -0.00000668, 0.00001763, -0.00000234, -0.00003213, -0.00000232, 0.00003104, 0.00000341, -0.00001902, 0.00000637, 0.00001731, -0.00000213, -0.00000615, 0.00000059, -0.00000520, -0.00000548, 0.00000765, 0.00001020, -0.00000233, -0.00000589, 0.00000422, 0.00000327, -0.00000929, -0.00000488, 0.00000720, -0.00000197, -0.00000993, 0.00000319, 0.00000484, -0.00000677, 0.00000030, 0.00000607, -0.00001005, -0.00001296, 0.00000098, -0.00000543, -0.00002389, -0.00002983, -0.00002927, -0.00001206, 0.00003435, 0.00006905, 0.00005523, 0.00003083, 0.00004676, 0.00007319, 0.00004056, -0.00003270, },
		{ -0.00001123, -0.00000943, -0.00000915, -0.00000315, -0.00001283, -0.00001939, -0.00000663, -0.00000053, -0.00000543, -0.00000169, 0.00000501, 0.00000556, 0.00000399, -0.00000081, -0.00000412, -0.00000196, -0.00000467, -0.00000724, 0.00000116, 0.00000529, 0.00000143, 0.00000379, -0.00000009, -0.00001309, -0.00000672, 0.00000791, -0.00000093, -0.00001011, 0.00000140, 0.00000417, -0.00000683, -0.00000263, 0.00001151, 0.00001175, 0.00000132, -0.00000591, -0.00000679, -0.00000403, 0.00000121, 0.00000763, 0.00000739, 0.00000027, -0.00000081, 0.00000013, -0.00000592, -0.00000196, 0.00000767, -0.00000713, -0.00002085, 0.00000150, 0.00001667, 0.00000100, 0.00000105, 0.00001118, -0.00000558, -0.00001246, 0.00000703, 0.00000052, -0.00001876, 0.00000105, 0.00001962, -0.00000406, -0.00001549, 0.00001152, 0.00001919, -0.00000929, -0.00001863, 0.00000677, 0.00001544, -0.00000983, -0.00001998, 0.00000152, 0.00001063, -0.00000284, -0.00000525, 0.00000529, 0.00000819, -0.00000064, -0.00001439, -0.00001129, 0.00001187, 0.00000726, -0.00002626, -0.00001611, 0.00002581, 0.00001587, -0.00001793, -0.00000312, 0.00001922, 0.00000295, -0.00000853, 0.00000090, -0.00000234, -0.00000869, 0.00000382, 0.00001235, 0.00000107, -0.00000680, 0.00000175, 0.00000552, -0.00000600, -0.00000815, 0.00000477, 0.00000254, -0.00000970, -0.00000124, 0.00000733, -0.00000475, -0.00000424, 0.00000831, -0.00000405, -0.00001707, -0.00000187, 0.00000168, -0.00002063, -0.00003032, -0.00002523, -0.00001647, 0.00001626, 0.00005820, 0.00006057, 0.00003356, 0.00003324, 0.00006201, 0.00005859, -0.00000096, -0.00005093, -0.00003948, },
		{ -0.00000948, -0.00000489, -0.00000740, -0.00001849, -0.00001304, -0.00000097, -0.00000288, -0.00000427, 0.00000280, 0.00000583, 0.00000462, 0.00000192, -0.00000293, -0.00000390, -0.00000366, -0.00000607, -0.00000195, 0.00000429, 0.00000236, 0.00000258, 0.00000286, -0.00000879, -0.00001107, 0.00000440, 0.00000407, -0.00000971, -0.00000376, 0.00000585, -0.00000359, -0.00000608, 0.00000889, 0.00001365, 0.00000318, -0.00000429, -0.00000579, -0.00000591, -0.00000156, 0.00000694, 0.00000913, 0.00000195, -0.00000130, 0.00000075, -0.00000430, -0.00000517, 0.00000591, -0.00000013, -0.00001893, -0.00000712, 0.00001497, 0.00000529, -0.00000231, 0.00001098, 0.00000258, -0.00001397, 0.00000008, 0.00000546, -0.00001506, -0.00000727, 0.00001837, 0.00000531, -0.00001673, 0.00000052, 0.00002031, 0.00000281, -0.00001888, -0.00000631, 0.00001550, 0.00000391, -0.00002043, -0.00001062, 0.00001241, 0.00000552, -0.00000852, 0.00000069, 0.00001089, 0.00000247, -0.00001187, -0.00001248, 0.00000592, 0.00001125, -0.00001683, -0.00002392, 0.00001473, 0.00002444, -0.00001085, -0.00001206, 0.00001680, 0.00000980, -0.00000951, -0.00000135, 0.00000166, -0.00000957, -0.00000174, 0.00001289, 0.00000524, -0.00000672, -0.00000086, 0.00000628, -0.00000235, -0.00000915, 0.00000102, 0.00000512, -0.00000701, -0.00000538, 0.00000726, -0.00000067, -0.00000791, 0.00000658, 0.00000354, -0.00001727, -0.00000862, 0.00000654, -0.00001330, -0.00003201, -0.00002402, -0.00001528, 0.00000262, 0.00004295, 0.00006181, 0.00003954, 0.00002565, 0.00004764, 0.00006276, 0.00002674, -0.00003185, -0.00004788, -0.00002225, -0.00000782, },
		{ -0.00000501, -0.00001417, -0.00001729, -0.00000509, -0.00000039, -0.00000439, -0.00000037, 0.00000535, 0.00000510, 0.00000351, -0.00000038, -0.00000459, -0.00000458, -0.00000460, -0.00000316, 0.00000206, 0.00000287, 0.00000200, 0.00000370, -0.00000394, -0.00001200, -0.00000102, 0.00000656, -0.00000608, -0.00000823, 0.00000429, 0.00000065, -0.00000750, 0.00000458, 0.00001499, 0.00000588, -0.00000361, -0.00000451, -0.00000590, -0.00000467, 0.00000465, 0.00001066, 0.00000443, -0.00000161, 0.00000053, -0.00000207, -0.00000669, 0.00000216, 0.00000440, -0.00001373, -0.00001342, 0.00001012, 0.00000951, -0.00000365, 0.00000744, 0.00000960, -0.00001057, -0.00000706, 0.00000624, -0.00000916, -0.00001290, 0.00001334, 0.00001304, -0.00001303, -0.00000858, 0.00001546, 0.00001179, -0.00001189, -0.00001538, 0.00000668, 0.00001386, -0.00001120, -0.00002098, 0.00000529, 0.00001487, -0.00000551, -0.00000671, 0.00001094, 0.00000775, -0.00001011, -0.00001308, 0.00000228, 0.00001135, -0.00000836, -0.00002504, 0.00000236, 0.00002639, -0.00000034, -0.00001712, 0.00000998, 0.00001613, -0.00000727, -0.00000576, 0.00000462, -0.00000711, -0.00000762, 0.00001075, 0.00000980, -0.00000537, -0.00000348, 0.00000604, 0.00000080, -0.00000836, -0.00000237, 0.00000531, -0.00000338, -0.00000765, 0.00000480, 0.00000360, -0.00000879, 0.00000163, 0.00000937, -0.00001228, -0.00001612, 0.00000612, -0.00000316, -0.00003121, -0.00002701, -0.00001205, -0.00000420, 0.00002585, 0.00005748, 0.00004665, 0.00002344, 0.00003438, 0.00005766, 0.00004436, -0.00000678, -0.00004405, -0.00003472, -0.00001183, -0.00000756, -0.00000682, },
		{ -0.00001748, -0.00001083, -0.00000049, -0.00000230, -0.00000284, 0.00000345, 0.00000554, 0.00000429, 0.00000205, -0.00000337, -0.00000608, -0.00000440, -0.00000273, 0.00000012, 0.00000219, 0.00000196, 0.00000353, -0.00000022, -0.00000997, -0.00000584, 0.00000564, -0.00000131, -0.00000995, 0.00000024, 0.00000375, -0.00000625, -0.00000022, 0.00001445, 0.00000957, -0.00000281, -0.00000420, -0.00000448, -0.00000657, 0.00000084, 0.00001087, 0.00000772, -0.00000104, -0.00000061, -0.00000013, -0.00000616, -0.00000205, 0.00000561, -0.00000709, -0.00001607, 0.00000326, 0.00001211, -0.00000233, 0.00000215, 0.00001322, -0.00000345, -0.00001170, 0.00000309, -0.00000327, -0.00001499, 0.00000616, 0.00001764, -0.00000621, -0.00001398, 0.00000768, 0.00001530, -0.00000228, -0.00001676, -0.00000547, 0.00001456, 0.00000285, -0.00002281, -0.00000832, 0.00001856, 0.00000438, -0.00001233, 0.00000568, 0.00001394, -0.00000649, -0.00001505, 0.00000028, 0.00001091, -0.00000307, -0.00002211, -0.00000705, 0.00002229, 0.00000946, -0.00001665, 0.00000096, 0.00001923, -0.00000149, -0.00001026, 0.00000449, -0.00000195, -0.00001158, 0.00000554, 0.00001358, -0.00000226, -0.00000598, 0.00000500, 0.00000336, -0.00000675, -0.00000450, 0.00000399, -0.00000049, -0.00000772, 0.00000136, 0.00000627, -0.00000666, -0.00000407, 0.00001092, -0.00000372, -0.00002029, -0.00000004, 0.00000600, -0.00002522, -0.00003240, -0.00001108, -0.00000421, 0.00001081, 0.00004699, 0.00005253, 0.00002598, 0.00002388, 0.00004820, 0.00005163, 0.00001589, -0.00002920, -0.00004138, -0.00002144, -0.00000714, -0.00000664, -0.00000526, -0.00000961, },
		{ -0.00000393, -0.00000005, -0.00000315, 0.00000059, 0.00000524, 0.00000492, 0.00000362, -0.00000094, -0.00000657, -0.00000572, -0.00000206, -0.00000041, 0.00000074, 0.00000161, 0.00000324, 0.00000198, -0.00000655, -0.00000837, 0.00000228, 0.00000224, -0.00000856, -0.00000424, 0.00000423, -0.00000318, -0.00000380, 0.00001160, 0.00001321, -0.00000068, -0.00000496, -0.00000306, -0.00000642, -0.00000326, 0.00000888, 0.00001097, 0.00000112, -0.00000210, 0.00000058, -0.00000404, -0.00000511, 0.00000373, -0.00000120, -0.00001478, -0.00000375, 0.00001192, 0.00000116, -0.00000270, 0.00001242, 0.00000490, -0.00001198, -0.00000243, 0.00000060, -0.00001372, -0.00000134, 0.00001854, 0.00000186, -0.00001541, -0.00000026, 0.00001397, 0.00000536, -0.00001143, -0.00001420, 0.00000610, 0.00001378, -0.00001407, -0.00002128, 0.00001231, 0.00001675, -0.00001101, -0.00000407, 0.00001721, 0.00000100, -0.00001746, -0.00000281, 0.00001203, 0.00000018, -0.00001888, -0.00001204, 0.00001545, 0.00001532, -0.00001171, -0.00000700, 0.00001773, 0.00000613, -0.00001226, 0.00000076, 0.00000360, -0.00001174, -0.00000167, 0.00001488, 0.00000267, -0.00000776, 0.00000293, 0.00000558, -0.00000494, -0.00000560, 0.00000249, 0.00000099, -0.00000658, -0.00000148, 0.00000675, -0.00000288, -0.00000798, 0.00000799, 0.00000482, -0.00001865, -0.00000913, 0.00001011, -0.00001439, -0.00003614, -0.00001498, -0.00000045, 0.00000170, 0.00003180, 0.00005424, 0.00003254, 0.00001691, 0.00003717, 0.00005171, 0.00003136, -0.00000950, -0.00003785, -0.00003197, -0.00001166, -0.00000501, -0.00000574, -0.00000713, -0.00001450, -0.00001505, },
		{ -0.00000135, -0.00000173, 0.00000357, 0.00000548, 0.00000456, 0.00000148, -0.00000538, -0.00000750, -0.00000253, 0.00000030, -0.00000029, 0.00000043, 0.00000286, 0.00000317, -0.00000319, -0.00000841, -0.00000158, 0.00000328, -0.00000536, -0.00000712, 0.00000220, -0.00000009, -0.00000511, 0.00000725, 0.00001535, 0.00000310, -0.00000572, -0.00000286, -0.00000471, -0.00000607, 0.00000489, 0.00001278, 0.00000486, -0.00000279, -0.00000032, -0.00000154, -0.00000595, 0.00000011, 0.00000213, -0.00001056, -0.00000888, 0.00000871, 0.00000532, -0.00000514, 0.00000786, 0.00001149, -0.00000761, -0.00000784, 0.00000124, -0.00001019, -0.00000748, 0.00001582, 0.00000955, -0.00001332, -0.00000683, 0.00000998, 0.00000886, -0.00000363, -0.00001554, -0.00000625, 0.00001536, 0.00000093, -0.00002588, -0.00000245, 0.00002421, -0.00000096, -0.00001347, 0.00001380, 0.00001123, -0.00001694, -0.00000897, 0.00001368, 0.00000438, -0.00001726, -0.00001440, 0.00000954, 0.00001679, -0.00000541, -0.00001147, 0.00001251, 0.00001276, -0.00001025, -0.00000503, 0.00000679, -0.00000784, -0.00000847, 0.00001245, 0.00000845, -0.00000773, -0.00000032, 0.00000735, -0.00000277, -0.00000642, 0.00000152, 0.00000164, -0.00000555, -0.00000312, 0.00000583, 0.00000070, -0.00000896, 0.00000258, 0.00000996, -0.00001183, -0.00001664, 0.00000742, -0.00000231, -0.00003442, -0.00002295, 0.00000227, 0.00000019, 0.00001580, 0.00004926, 0.00004115, 0.00001472, 0.00002564, 0.00004771, 0.00003965, 0.00000814, -0.00002530, -0.00003707, -0.00002096, -0.00000521, -0.00000453, -0.00000662, -0.00001095, -0.00001587, -0.00000908, 0.00000001, },
		{ 0.00000089, 0.00000529, 0.00000541, 0.00000323, -0.00000308, -0.00000835, -0.00000438, 0.00000105, 0.00000000, -0.00000114, 0.00000185, 0.00000384, -0.00000047, -0.00000683, -0.00000434, 0.00000208, -0.00000226, -0.00000761, -0.00000095, 0.00000151, -0.00000438, 0.00000292, 0.00001515, 0.00000771, -0.00000517, -0.00000409, -0.00000282, -0.00000669, 0.00000017, 0.00001212, 0.00000916, -0.00000161, -0.00000224, -0.00000008, -0.00000459, -0.00000333, 0.00000223, -0.00000543, -0.00001074, 0.00000357, 0.00000820, -0.00000439, 0.00000167, 0.00001402, -0.00000012, -0.00001056, -0.00000127, -0.00000620, -0.00001100, 0.00001030, 0.00001536, -0.00000829, -0.00001147, 0.00000518, 0.00000884, 0.00000227, -0.00001026, -0.00001539, 0.00000685, 0.00001389, -0.00001871, -0.00001865, 0.00002095, 0.00001411, -0.00001617, 0.00000342, 0.00001968, -0.00001042, -0.00001704, 0.00001246, 0.00001141, -0.00001589, -0.00001730, 0.00000595, 0.00001618, -0.00000066, -0.00001239, 0.00000609, 0.00001621, -0.00000485, -0.00001012, 0.00000600, -0.00000168, -0.00001222, 0.00000650, 0.00001311, -0.00000505, -0.00000426, 0.00000796, 0.00000021, -0.00000732, 0.00000070, 0.00000259, -0.00000516, -0.00000421, 0.00000475, 0.00000312, -0.00000768, -0.00000259, 0.00001046, -0.00000306, -0.00001902, -0.00000021, 0.00000625, -0.00002632, -0.00003106, 0.00000008, 0.00000434, 0.00000393, 0.00003734, 0.00004813, 0.00001841, 0.00001484, 0.00004071, 0.00004342, 0.00002015, -0.00000934, -0.00003298, -0.00003071, -0.00001027, -0.00000236, -0.00000602, -0.00000893, -0.00001380, -0.00001313, -0.00000303, 0.00000078, -0.00000205, },
		{ 0.00000621, 0.00000452, -0.00000070, -0.00000767, -0.00000670, 0.00000061, 0.00000145, -0.00000201, -0.00000003, 0.00000389, 0.00000171, -0.00000462, -0.00000548, -0.00000011, -0.00000059, -0.00000633, -0.00000358, 0.00000122, -0.00000279, -0.00000015, 0.00001289, 0.00001173, -0.00000268, -0.00000583, -0.00000206, -0.00000543, -0.00000364, 0.00000899, 0.00001247, 0.00000168, -0.00000385, -0.00000047, -0.00000218, -0.00000494, -0.00000013, -0.00000161, -0.00000930, -0.00000156, 0.00000846, -0.00000122, -0.00000350, 0.00001187, 0.00000764, -0.00000904, -0.00000541, -0.00000363, -0.00001154, 0.00000348, 0.00001803, -0.00000118, -0.00001381, 0.00000030, 0.00000729, 0.00000430, -0.00000266, -0.00001678, -0.00000667, 0.00001745, -0.00000315, -0.00002749, 0.00000681, 0.00002634, -0.00000854, -0.00000942, 0.00002109, 0.00000181, -0.00002273, 0.00000552, 0.00001977, -0.00001131, -0.00002229, 0.00000290, 0.00001649, 0.00000189, -0.00001183, 0.00000079, 0.00001620, 0.00000172, -0.00001219, 0.00000177, 0.00000381, -0.00001159, -0.00000099, 0.00001462, 0.00000004, -0.00000754, 0.00000657, 0.00000402, -0.00000773, -0.00000094, 0.00000433, -0.00000470, -0.00000584, 0.00000399, 0.00000475, -0.00000562, -0.00000587, 0.00000753, 0.00000394, -0.00001570, -0.00000845, 0.00000831, -0.00001464, -0.00003473, -0.00000757, 0.00000944, -0.00000045, 0.00002153, 0.00004931, 0.00002727, 0.00000735, 0.00003048, 0.00004467, 0.00002736, 0.00000396, -0.00002112, -0.00003495, -0.00001996, -0.00000235, -0.00000371, -0.00000836, -0.00001112, -0.00001419, -0.00000784, 0.00000096, -0.00000027, -0.00000137, 0.00000367, },
		{ 0.00000118, -0.00000592, -0.00000824, -0.00000130, 0.00000288, -0.00000136, -0.00000223, 0.00000284, 0.00000342, -0.00000226, -0.00000532, -0.00000206, -0.00000048, -0.00000458, -0.00000478, -0.00000021, -0.00000160, -0.00000169, 0.00000961, 0.00001407, 0.00000133, -0.00000676, -0.00000287, -0.00000353, -0.00000546, 0.00000451, 0.00001347, 0.00000623, -0.00000375, -0.00000240, -0.00000037, -0.00000422, -0.00000314, -0.00000051, -0.00000597, -0.00000471, 0.00000612, 0.00000245, -0.00000574, 0.00000640, 0.00001264, -0.00000353, -0.00000874, -0.00000361, -0.00000991, -0.00000273, 0.00001693, 0.00000662, -0.00001325, -0.00000460, 0.00000561, 0.00000344, 0.00000257, -0.00001079, -0.00001753, 0.00000961, 0.00001235, -0.00002372, -0.00001185, 0.00002817, 0.00000720, -0.00001736, 0.00001326, 0.00001519, -0.00002126, -0.00000653, 0.00002496, -0.00000107, -0.00002740, -0.00000296, 0.00001871, 0.00000428, -0.00001207, -0.00000291, 0.00001427, 0.00000714, -0.00001077, -0.00000361, 0.00000626, -0.00000739, -0.00000719, 0.00001216, 0.00000594, -0.00000860, 0.00000303, 0.00000772, -0.00000659, -0.00000389, 0.00000610, -0.00000283, -0.00000832, 0.00000279, 0.00000652, -0.00000367, -0.00000733, 0.00000359, 0.00000715, -0.00000917, -0.00001313, 0.00000449, -0.00000420, -0.00003173, -0.00001750, 0.00001068, 0.00000235, 0.00000715, 0.00004245, 0.00003783, 0.00000617, 0.00001777, 0.00004294, 0.00003273, 0.00001194, -0.00000706, -0.00003020, -0.00003001, -0.00000751, -0.00000044, -0.00000736, -0.00000974, -0.00001256, -0.00001182, -0.00000187, 0.00000202, -0.00000170, 0.00000089, 0.00000623, 0.00000590, },
		{ -0.00000840, -0.00000382, 0.00000310, 0.00000062, -0.00000358, 0.00000056, 0.00000430, 0.00000021, -0.00000435, -0.00000324, -0.00000127, -0.00000343, -0.00000473, -0.00000165, -0.00000134, -0.00000224, 0.00000640, 0.00001448, 0.00000574, -0.00000600, -0.00000466, -0.00000232, -0.00000530, 0.00000028, 0.00001186, 0.00001045, -0.00000134, -0.00000457, -0.00000034, -0.00000201, -0.00000487, -0.00000201, -0.00000290, -0.00000502, 0.00000258, 0.00000466, -0.00000479, 0.00000027, 0.00001316, 0.00000384, -0.00000901, -0.00000584, -0.00000781, -0.00000680, 0.00001257, 0.00001313, -0.00000940, -0.00000918, 0.00000379, 0.00000210, 0.00000326, -0.00000196, -0.00002010, -0.00000511, 0.00001912, -0.00000914, -0.00002543, 0.00001722, 0.00002342, -0.00001467, -0.00000056, 0.00002307, -0.00001107, -0.00001892, 0.00002240, 0.00001354, -0.00002805, -0.00001342, 0.00002048, 0.00000943, -0.00001340, -0.00000659, 0.00001220, 0.00001060, -0.00000721, -0.00000769, 0.00000515, -0.00000209, -0.00000997, 0.00000673, 0.00001038, -0.00000667, -0.00000166, 0.00000984, -0.00000324, -0.00000744, 0.00000634, 0.00000110, -0.00001055, -0.00000011, 0.00000866, -0.00000145, -0.00000809, 0.00000040, 0.00000711, -0.00000301, -0.00001279, -0.00000150, 0.00000120, -0.00002369, -0.00002493, 0.00000609, 0.00000836, -0.00000106, 0.00002903, 0.00004485, 0.00001244, 0.00000569, 0.00003636, 0.00003797, 0.00001634, 0.00000330, -0.00001801, -0.00003445, -0.00001769, 0.00000058, -0.00000431, -0.00000964, -0.00001032, -0.00001301, -0.00000665, 0.00000253, 0.00000022, -0.00000165, 0.00000466, 0.00000734, 0.00000286, -0.00000406, },
		{ 0.00000178, 0.00000277, -0.00000317, -0.00000226, 0.00000371, 0.00000260, -0.00000276, -0.00000369, -0.00000218, -0.00000313, -0.00000425, -0.00000242, -0.00000171, -0.00000258, 0.00000378, 0.00001344, 0.00000947, -0.00000354, -0.00000629, -0.00000244, -0.00000410, -0.00000251, 0.00000845, 0.00001287, 0.00000282, -0.00000542, -0.00000215, 0.00000005, -0.00000436, -0.00000462, -0.00000190, -0.00000316, -0.00000017, 0.00000453, -0.00000205, -0.00000396, 0.00000950, 0.00001001, -0.00000538, -0.00000857, -0.00000696, -0.00000829, 0.00000663, 0.00001654, -0.00000280, -0.00001225, 0.00000098, 0.00000192, 0.00000061, 0.00000429, -0.00001394, -0.00001855, 0.00001342, 0.00000822, -0.00002669, -0.00000172, 0.00003118, -0.00000125, -0.00001287, 0.00002089, 0.00000431, -0.00002518, 0.00001102, 0.00002716, -0.00002017, -0.00002637, 0.00001746, 0.00001830, -0.00001335, -0.00001223, 0.00001024, 0.00001299, -0.00000338, -0.00000936, 0.00000194, 0.00000176, -0.00000909, 0.00000074, 0.00001172, -0.00000246, -0.00000569, 0.00000927, 0.00000177, -0.00000992, 0.00000385, 0.00000621, -0.00001058, -0.00000518, 0.00001004, 0.00000204, -0.00000877, -0.00000199, 0.00000584, 0.00000047, -0.00000914, -0.00000546, 0.00000112, -0.00001495, -0.00002668, -0.00000228, 0.00001246, -0.00000168, 0.00001393, 0.00004425, 0.00002374, -0.00000126, 0.00002441, 0.00004162, 0.00002100, 0.00000765, -0.00000421, -0.00002961, -0.00002870, -0.00000387, 0.00000009, -0.00000892, -0.00000938, -0.00001152, -0.00001086, -0.00000002, 0.00000296, -0.00000230, 0.00000160, 0.00000803, 0.00000493, -0.00000254, -0.00000765, -0.00000577, },
		{ -0.00000115, -0.00000430, 0.00000147, 0.00000426, -0.00000050, -0.00000354, -0.00000287, -0.00000331, -0.00000400, -0.00000258, -0.00000205, -0.00000308, 0.00000162, 0.00001163, 0.00001197, -0.00000004, -0.00000682, -0.00000366, -0.00000303, -0.00000357, 0.00000462, 0.00001292, 0.00000731, -0.00000405, -0.00000470, 0.00000050, -0.00000208, -0.00000636, -0.00000336, -0.00000102, -0.00000079, 0.00000276, 0.00000044, -0.00000512, 0.00000383, 0.00001253, 0.00000099, -0.00000949, -0.00000804, -0.00000816, 0.00000119, 0.00001610, 0.00000475, -0.00001232, -0.00000319, 0.00000265, -0.00000230, 0.00000500, -0.00000350, -0.00002371, -0.00000145, 0.00001863, -0.00001519, -0.00001901, 0.00002559, 0.00001654, -0.00001609, 0.00000963, 0.00001766, -0.00002122, -0.00000513, 0.00003322, -0.00000369, -0.00003604, 0.00000652, 0.00002808, -0.00000825, -0.00001988, 0.00000660, 0.00001568, -0.00000005, -0.00000929, -0.00000125, 0.00000302, -0.00000621, -0.00000348, 0.00000993, 0.00000217, -0.00000755, 0.00000614, 0.00000681, -0.00000971, -0.00000112, 0.00001031, -0.00000698, -0.00001113, 0.00000873, 0.00000712, -0.00000868, -0.00000464, 0.00000495, 0.00000150, -0.00000549, -0.00000544, -0.00000147, -0.00000961, -0.00002330, -0.00001003, 0.00001161, 0.00000260, 0.00000262, 0.00003561, 0.00003471, 0.00000008, 0.00000982, 0.00004025, 0.00002809, 0.00000844, 0.00000495, -0.00001693, -0.00003432, -0.00001407, 0.00000260, -0.00000580, -0.00000983, -0.00000929, -0.00001242, -0.00000484, 0.00000420, -0.00000045, -0.00000161, 0.00000694, 0.00000744, -0.00000095, -0.00000687, -0.00000661, -0.00000032, 0.00000384, },
		{ -0.00000170, 0.00000431, 0.00000218, -0.00000261, -0.00000337, -0.00000360, -0.00000411, -0.00000260, -0.00000199, -0.00000354, -0.00000037, 0.00000946, 0.00001326, 0.00000366, -0.00000595, -0.00000518, -0.00000281, -0.00000349, 0.00000156, 0.00001104, 0.00001065, -0.00000068, -0.00000639, -0.00000105, 0.00000043, -0.00000595, -0.00000618, -0.00000050, 0.00000061, 0.00000105, 0.00000126, -0.00000388, -0.00000113, 0.00001077, 0.00000743, -0.00000706, -0.00001017, -0.00000808, -0.00000238, 0.00001264, 0.00001087, -0.00000879, -0.00000768, 0.00000265, -0.00000304, 0.00000111, 0.00000495, -0.00001875, -0.00001724, 0.00001637, 0.00000223, -0.00002572, 0.00000911, 0.00002921, -0.00000745, -0.00000419, 0.00002233, -0.00000836, -0.00001860, 0.00002801, 0.00001629, -0.00003621, -0.00001111, 0.00003311, 0.00000376, -0.00002650, -0.00000112, 0.00001859, 0.00000368, -0.00000890, -0.00000328, 0.00000247, -0.00000356, -0.00000506, 0.00000661, 0.00000538, -0.00000693, 0.00000177, 0.00001006, -0.00000638, -0.00000677, 0.00001114, -0.00000016, -0.00001530, 0.00000354, 0.00001240, -0.00000625, -0.00000821, 0.00000434, 0.00000203, -0.00000416, -0.00000278, -0.00000251, -0.00000870, -0.00001850, -0.00001360, 0.00000672, 0.00000699, -0.00000204, 0.00002262, 0.00003993, 0.00000900, -0.00000209, 0.00003162, 0.00003592, 0.00001055, 0.00000693, -0.00000244, -0.00003037, -0.00002604, -0.00000040, -0.00000069, -0.00000979, -0.00000835, -0.00001112, -0.00000954, 0.00000235, 0.00000275, -0.00000313, 0.00000387, 0.00000946, 0.00000148, -0.00000647, -0.00000680, -0.00000207, 0.00000338, 0.00000134, -0.00000449, },
		{ 0.00000443, -0.00000064, -0.00000358, -0.00000392, -0.00000435, -0.00000288, -0.00000168, -0.00000359, -0.00000222, 0.00000698, 0.00001355, 0.00000702, -0.00000394, -0.00000620, -0.00000346, -0.00000319, -0.00000031, 0.00000832, 0.00001209, 0.00000352, -0.00000613, -0.00000378, 0.00000156, -0.00000351, -0.00000842, -0.00000239, 0.00000254, 0.00000097, 0.00000048, -0.00000205, -0.00000361, 0.00000616, 0.00001122, -0.00000157, -0.00001135, -0.00000929, -0.00000411, 0.00000812, 0.00001378, -0.00000264, -0.00001050, 0.00000062, -0.00000147, -0.00000385, 0.00000729, -0.00000754, -0.00002573, 0.00000296, 0.00001546, -0.00001895, -0.00000959, 0.00002968, 0.00000891, -0.00001210, 0.00001649, 0.00000696, -0.00002288, 0.00001354, 0.00003178, -0.00002426, -0.00002941, 0.00002815, 0.00002032, -0.00002714, -0.00001349, 0.00001939, 0.00000928, -0.00000869, -0.00000465, 0.00000177, -0.00000234, -0.00000496, 0.00000372, 0.00000652, -0.00000484, -0.00000217, 0.00001057, -0.00000102, -0.00001073, 0.00000788, 0.00000754, -0.00001513, -0.00000469, 0.00001523, -0.00000051, -0.00001179, 0.00000248, 0.00000375, -0.00000492, -0.00000078, 0.00000012, -0.00000964, -0.00001634, -0.00001286, 0.00000156, 0.00000797, -0.00000135, 0.00001073, 0.00003715, 0.00002082, -0.00000645, 0.00001733, 0.00003989, 0.00001719, 0.00000443, 0.00000701, -0.00001777, -0.00003335, -0.00000983, 0.00000324, -0.00000720, -0.00000905, -0.00000875, -0.00001180, -0.00000221, 0.00000497, -0.00000194, -0.00000003, 0.00000967, 0.00000498, -0.00000590, -0.00000728, -0.00000292, 0.00000209, 0.00000299, -0.00000281, -0.00000434, 0.00000235, },
		{ -0.00000307, -0.00000440, -0.00000454, -0.00000335, -0.00000152, -0.00000311, -0.00000365, 0.00000417, 0.00001295, 0.00000981, -0.00000125, -0.00000635, -0.00000448, -0.00000325, -0.00000131, 0.00000577, 0.00001179, 0.00000720, -0.00000383, -0.00000616, 0.00000061, -0.00000040, -0.00000855, -0.00000591, 0.00000305, 0.00000280, -0.00000051, -0.00000128, -0.00000358, 0.00000127, 0.00001102, 0.00000491, -0.00000976, -0.00001163, -0.00000528, 0.00000438, 0.00001332, 0.00000382, -0.00001004, -0.00000316, 0.00000053, -0.00000649, 0.00000359, 0.00000321, -0.00002334, -0.00001401, 0.00001701, -0.00000357, -0.00002068, 0.00001771, 0.00002404, -0.00000891, 0.00000462, 0.00001690, -0.00001639, -0.00000324, 0.00003630, -0.00000353, -0.00004026, 0.00001220, 0.00003499, -0.00001810, -0.00002745, 0.00001463, 0.00001687, -0.00000739, -0.00000673, 0.00000188, -0.00000202, -0.00000495, 0.00000223, 0.00000643, -0.00000272, -0.00000473, 0.00000875, 0.00000447, -0.00001141, 0.00000167, 0.00001301, -0.00000986, -0.00001306, 0.00001321, 0.00000756, -0.00001307, -0.00000188, 0.00000647, -0.00000541, -0.00000203, 0.00000508, -0.00000829, -0.00001812, -0.00001114, -0.00000025, 0.00000564, 0.00000058, 0.00000366, 0.00002857, 0.00002956, -0.00000206, 0.00000270, 0.00003605, 0.00002713, 0.00000323, 0.00000818, -0.00000250, -0.00003111, -0.00002221, 0.00000202, -0.00000203, -0.00000964, -0.00000748, -0.00001105, -0.00000729, 0.00000428, 0.00000123, -0.00000279, 0.00000740, 0.00000866, -0.00000407, -0.00000849, -0.00000340, 0.00000109, 0.00000322, -0.00000033, -0.00000518, -0.00000096, 0.00000510, 0.00000220, },
		{ -0.00000484, -0.00000373, -0.00000174, -0.00000242, -0.00000431, 0.00000128, 0.00001140, 0.00001192, 0.00000177, -0.00000564, -0.00000535, -0.00000376, -0.00000204, 0.00000379, 0.00001056, 0.00000954, -0.00000038, -0.00000697, -0.00000189, 0.00000170, -0.00000633, -0.00000921, 0.00000104, 0.00000533, -0.00000012, -0.00000193, -0.00000251, -0.00000178, 0.00000755, 0.00000965, -0.00000497, -0.00001355, -0.00000731, 0.00000209, 0.00001089, 0.00000830, -0.00000632, -0.00000679, 0.00000082, -0.00000578, -0.00000264, 0.00000810, -0.00001274, -0.00002549, 0.00000629, 0.00001075, -0.00001901, 0.00000012, 0.00002953, 0.00000390, -0.00000515, 0.00001712, -0.00000354, -0.00001432, 0.00002880, 0.00001797, -0.00003804, -0.00000988, 0.00004040, -0.00000007, -0.00003703, 0.00000238, 0.00002390, -0.00000281, -0.00001008, 0.00000208, -0.00000102, -0.00000604, 0.00000142, 0.00000657, -0.00000126, -0.00000611, 0.00000574, 0.00000846, -0.00000878, -0.00000501, 0.00001407, -0.00000116, -0.00001799, 0.00000601, 0.00001494, -0.00000995, -0.00000819, 0.00000806, -0.00000325, -0.00000630, 0.00000858, -0.00000219, -0.00002124, -0.00001267, 0.00000169, 0.00000341, 0.00000019, 0.00000119, 0.00001906, 0.00003160, 0.00000766, -0.00000626, 0.00002443, 0.00003524, 0.00000779, 0.00000349, 0.00000799, -0.00001942, -0.00003110, -0.00000602, 0.00000286, -0.00000782, -0.00000803, -0.00000876, -0.00001038, 0.00000054, 0.00000420, -0.00000292, 0.00000345, 0.00001090, -0.00000038, -0.00000978, -0.00000461, 0.00000094, 0.00000266, 0.00000153, -0.00000396, -0.00000410, 0.00000352, 0.00000501, -0.00000129, -0.00000491, },
		{ -0.00000211, -0.00000196, -0.00000420, -0.00000121, 0.00000894, 0.00001313, 0.00000491, -0.00000424, -0.00000581, -0.00000443, -0.00000288, 0.00000220, 0.00000913, 0.00001055, 0.00000305, -0.00000590, -0.00000459, 0.00000177, -0.00000287, -0.00001053, -0.00000304, 0.00000665, 0.00000225, -0.00000294, -0.00000209, -0.00000247, 0.00000302, 0.00001083, 0.00000153, -0.00001312, -0.00001055, 0.00000045, 0.00000843, 0.00000982, -0.00000109, -0.00000813, -0.00000120, -0.00000332, -0.00000727, 0.00000596, -0.00000058, -0.00002626, -0.00001037, 0.00001533, -0.00000713, -0.00001331, 0.00002260, 0.00001891, -0.00000605, 0.00000945, 0.00000805, -0.00001533, 0.00001429, 0.00003179, -0.00002308, -0.00002936, 0.00003257, 0.00002106, -0.00003639, -0.00001540, 0.00002588, 0.00000610, -0.00001322, 0.00000051, 0.00000165, -0.00000752, -0.00000034, 0.00000764, 0.00000010, -0.00000713, 0.00000251, 0.00001041, -0.00000413, -0.00000975, 0.00001060, 0.00000766, -0.00001723, -0.00000409, 0.00001814, -0.00000228, -0.00001382, 0.00000615, 0.00000182, -0.00001066, 0.00000721, 0.00000737, -0.00002096, -0.00001911, 0.00000398, 0.00000501, -0.00000273, -0.00000017, 0.00001266, 0.00002774, 0.00001683, -0.00000664, 0.00000967, 0.00003614, 0.00001786, -0.00000073, 0.00000949, -0.00000388, -0.00003087, -0.00001814, 0.00000328, -0.00000303, -0.00000891, -0.00000720, -0.00001047, -0.00000440, 0.00000476, -0.00000059, -0.00000028, 0.00001051, 0.00000451, -0.00000972, -0.00000721, 0.00000106, 0.00000246, 0.00000208, -0.00000166, -0.00000557, 0.00000015, 0.00000640, 0.00000179, -0.00000483, -0.00000556, -0.00000396, },
		{ -0.00000373, -0.00000286, 0.00000594, 0.00001317, 0.00000795, -0.00000222, -0.00000582, -0.00000497, -0.00000382, 0.00000061, 0.00000779, 0.00001080, 0.00000568, -0.00000355, -0.00000622, 0.00000004, 0.00000011, -0.00000925, -0.00000748, 0.00000533, 0.00000569, -0.00000270, -0.00000301, -0.00000184, -0.00000029, 0.00000851, 0.00000718, -0.00000940, -0.00001387, -0.00000197, 0.00000700, 0.00000919, 0.00000325, -0.00000636, -0.00000418, -0.00000163, -0.00000828, -0.00000026, 0.00000677, -0.00001740, -0.00002379, 0.00000758, 0.00000648, -0.00001583, 0.00000780, 0.00002738, 0.00000290, 0.00000070, 0.00001238, -0.00000798, 0.00000077, 0.00003374, -0.00000176, -0.00003812, 0.00001371, 0.00003673, -0.00002359, -0.00003275, 0.00001905, 0.00001743, -0.00001313, -0.00000404, 0.00000512, -0.00000745, -0.00000406, 0.00000874, 0.00000261, -0.00000813, -0.00000083, 0.00001069, 0.00000092, -0.00001141, 0.00000439, 0.00001343, -0.00001107, -0.00001315, 0.00001523, 0.00000746, -0.00001554, 0.00000017, 0.00000746, -0.00001155, 0.00000052, 0.00001589, -0.00001382, -0.00002781, 0.00000169, 0.00001114, -0.00000473, -0.00000401, 0.00000975, 0.00002216, 0.00002100, -0.00000043, -0.00000168, 0.00002820, 0.00002836, 0.00000121, 0.00000378, 0.00000764, -0.00002065, -0.00002811, -0.00000318, 0.00000232, -0.00000772, -0.00000745, -0.00000865, -0.00000801, 0.00000224, 0.00000237, -0.00000193, 0.00000765, 0.00000886, -0.00000718, -0.00001057, 0.00000010, 0.00000330, 0.00000184, 0.00000025, -0.00000484, -0.00000350, 0.00000537, 0.00000529, -0.00000334, -0.00000662, -0.00000436, -0.00000230, -0.00000186, },
		{ 0.00000302, 0.00001196, 0.00001046, 0.00000037, -0.00000539, -0.00000528, -0.00000461, -0.00000114, 0.00000633, 0.00001081, 0.00000748, -0.00000091, -0.00000627, -0.00000239, 0.00000137, -0.00000616, -0.00001034, 0.00000144, 0.00000831, -0.00000031, -0.00000450, -0.00000159, -0.00000141, 0.00000451, 0.00000982, -0.00000324, -0.00001531, -0.00000591, 0.00000619, 0.00000820, 0.00000525, -0.00000261, -0.00000587, -0.00000210, -0.00000656, -0.00000611, 0.00000682, -0.00000499, -0.00002740, -0.00000781, 0.00001268, -0.00000773, -0.00000599, 0.00002463, 0.00001633, -0.00000192, 0.00000887, 0.00000150, -0.00000554, 0.00002596, 0.00001705, -0.00003309, -0.00000850, 0.00004014, -0.00000246, -0.00004229, 0.00000323, 0.00002642, -0.00000732, -0.00001056, 0.00000693, -0.00000427, -0.00000888, 0.00000791, 0.00000682, -0.00000830, -0.00000473, 0.00000974, 0.00000540, -0.00001024, -0.00000211, 0.00001464, -0.00000213, -0.00001769, 0.00000711, 0.00001528, -0.00001162, -0.00000778, 0.00001023, -0.00000742, -0.00000840, 0.00001850, -0.00000043, -0.00003307, -0.00000783, 0.00001815, -0.00000146, -0.00001071, 0.00000705, 0.00001896, 0.00001992, 0.00000693, -0.00000536, 0.00001501, 0.00003289, 0.00001027, -0.00000267, 0.00000994, -0.00000538, -0.00002972, -0.00001466, 0.00000401, -0.00000340, -0.00000839, -0.00000711, -0.00000887, -0.00000199, 0.00000368, -0.00000104, 0.00000395, 0.00001096, -0.00000233, -0.00001300, -0.00000294, 0.00000462, 0.00000206, 0.00000089, -0.00000272, -0.00000571, 0.00000209, 0.00000768, -0.00000012, -0.00000731, -0.00000542, -0.00000225, -0.00000186, -0.00000337, -0.00000365, },
		{ 0.00001198, 0.00000334, -0.00000442, -0.00000543, -0.00000506, -0.00000289, 0.00000444, 0.00001067, 0.00000886, 0.00000136, -0.00000513, -0.00000427, 0.00000079, -0.00000286, -0.00001059, -0.00000356, 0.00000840, 0.00000365, -0.00000496, -0.00000267, -0.00000103, 0.00000110, 0.00000893, 0.00000306, -0.00001348, -0.00001066, 0.00000458, 0.00000819, 0.00000523, 0.00000090, -0.00000492, -0.00000403, -0.00000463, -0.00000869, 0.00000158, 0.00000409, -0.00002070, -0.00002207, 0.00000729, 0.00000408, -0.00001119, 0.00001295, 0.00002598, 0.00000445, 0.00000245, 0.00000666, -0.00000378, 0.00001507, 0.00002688, -0.00001781, -0.00002484, 0.00003015, 0.00001877, -0.00003928, -0.00001693, 0.00002774, 0.00000385, -0.00001574, 0.00000471, 0.00000159, -0.00001231, 0.00000364, 0.00001163, -0.00000612, -0.00000921, 0.00000742, 0.00000906, -0.00000725, -0.00000706, 0.00001186, 0.00000617, -0.00001643, -0.00000288, 0.00001790, -0.00000319, -0.00001398, 0.00000793, 0.00000008, -0.00001479, 0.00001319, 0.00001436, -0.00002973, -0.00002245, 0.00002001, 0.00000863, -0.00001655, 0.00000058, 0.00001889, 0.00001729, 0.00001050, -0.00000207, 0.00000324, 0.00002826, 0.00002184, -0.00000311, 0.00000384, 0.00000688, -0.00002106, -0.00002528, -0.00000101, 0.00000220, -0.00000763, -0.00000724, -0.00000763, -0.00000554, 0.00000222, 0.00000107, 0.00000139, 0.00001027, 0.00000325, -0.00001282, -0.00000770, 0.00000484, 0.00000363, 0.00000064, -0.00000072, -0.00000561, -0.00000203, 0.00000760, 0.00000411, -0.00000654, -0.00000722, -0.00000242, -0.00000153, -0.00000327, -0.00000398, 0.00000070, 0.00000982, },
		{ -0.00000277, -0.00000543, -0.00000520, -0.00000429, 0.00000206, 0.00001007, 0.00001023, 0.00000324, -0.00000360, -0.00000501, -0.00000080, -0.00000074, -0.00000863, -0.00000765, 0.00000556, 0.00000741, -0.00000320, -0.00000460, -0.00000073, -0.00000041, 0.00000579, 0.00000715, -0.00000858, -0.00001431, 0.00000104, 0.00000903, 0.00000482, 0.00000243, -0.00000190, -0.00000531, -0.00000439, -0.00000809, -0.00000455, 0.00000617, -0.00000879, -0.00002778, -0.00000669, 0.00001066, -0.00000626, -0.00000012, 0.00002607, 0.00001641, -0.00000003, 0.00000523, 0.00000176, 0.00000777, 0.00002699, 0.00000000, -0.00002954, 0.00001198, 0.00003167, -0.00002461, -0.00003384, 0.00001881, 0.00001649, -0.00001580, -0.00000176, 0.00000745, -0.00001177, -0.00000360, 0.00001447, -0.00000066, -0.00001308, 0.00000313, 0.00001181, -0.00000321, -0.00000987, 0.00000693, 0.00001140, -0.00001069, -0.00001081, 0.00001452, 0.00000628, -0.00001528, 0.00000125, 0.00000707, -0.00001526, 0.00000225, 0.00002416, -0.00001672, -0.00003582, 0.00001225, 0.00002217, -0.00001599, -0.00001067, 0.00001857, 0.00001749, 0.00000926, 0.00000324, -0.00000180, 0.00001690, 0.00002884, 0.00000473, -0.00000429, 0.00000995, -0.00000623, -0.00002847, -0.00001178, 0.00000496, -0.00000366, -0.00000834, -0.00000632, -0.00000670, -0.00000117, 0.00000227, 0.00000097, 0.00000787, 0.00000753, -0.00000951, -0.00001247, 0.00000254, 0.00000614, 0.00000081, 0.00000004, -0.00000377, -0.00000509, 0.00000484, 0.00000772, -0.00000365, -0.00000900, -0.00000352, -0.00000079, -0.00000308, -0.00000434, -0.00000093, 0.00000736, 0.00001227, 0.00000628, },
		},
		{ { -0.00000043, -0.00000198, 0.00000053, 0.00000374, -0.00000251, -0.00000619, -0.00000275, -0.00000174, -0.00000099, -0.00000153, -0.00000196, 0.00000407, 0.00000117, -0.00000722, 0.00000192, 0.00000671, -0.00000386, -0.00000441, -0.00000541, -0.00000723, 0.00000461, -0.00000108, -0.00001167, 0.00000152, -0.00001013, -0.00003030, -0.00000485, 0.00000997, 0.00000725, 0.00003328, 0.00002573, -0.00001420, 0.00000469, 0.00003767, 0.00001613, -0.00000717, 0.00000287, 0.00000696, -0.00001571, -0.00002046, 0.00000828, 0.00000467, -0.00001719, 0.00000285, 0.00000604, -0.00000918, 0.00001958, 0.00001352, -0.00003173, -0.00000761, 0.00002135, -0.00000188, -0.00000193, 0.00000038, -0.00001406, 0.00000219, 0.00000822, -0.00000776, -0.00000030, 0.00000625, 0.00000476, 0.00000430, -0.00000997, -0.00000546, 0.00000967, -0.00000546, -0.00000997, 0.00000430, 0.00000476, 0.00000625, -0.00000030, -0.00000776, 0.00000822, 0.00000219, -0.00001406, 0.00000038, -0.00000193, -0.00000188, 0.00002135, -0.00000761, -0.00003173, 0.00001352, 0.00001958, -0.00000918, 0.00000604, 0.00000285, -0.00001719, 0.00000467, 0.00000828, -0.00002046, -0.00001571, 0.00000696, 0.00000287, -0.00000717, 0.00001613, 0.00003767, 0.00000469, -0.00001420, 0.00002573, 0.00003328, 0.00000725, 0.00000997, -0.00000485, -0.00003030, -0.00001013, 0.00000152, -0.00001167, -0.00000108, 0.00000461, -0.00000723, -0.00000541, -0.00000441, -0.00000386, 0.00000671, 0.00000192, -0.00000722, 0.00000117, 0.00000407, -0.00000196, -0.00000153, -0.00000099, -0.00000174, -0.00000275, -0.00000619, -0.00000251, 0.00000374, 0.00000053, -0.00000198, },
		{ -0.00000198, 0.00000246, 0.00000178, -0.00000559, -0.00000476, -0.00000099, -0.00000168, -0.00000157, -0.00000181, 0.00000080, 0.00000351, -0.00000401, -0.00000385, 0.00000693, 0.00000175, -0.00000574, -0.00000385, -0.00000845, -0.00000196, 0.00000638, -0.00000953, -0.00000799, 0.00000252, -0.00002311, -0.00002530, 0.00000646, 0.00000801, 0.00001796, 0.00003903, 0.00000569, -0.00001340, 0.00002780, 0.00003433, -0.00000033, -0.00000322, 0.00000915, -0.00000671, -0.00002459, -0.00000460, 0.00001201, -0.00001048, -0.00000935, 0.00001070, -0.00000654, 0.00000167, 0.00002794, -0.00001191, -0.00002966, 0.00001460, 0.00001133, -0.00000835, 0.00000430, -0.00000746, -0.00001097, 0.00001059, -0.00000001, -0.00000800, 0.00000604, 0.00000511, 0.00000462, -0.00000088, -0.00001199, 0.00000320, 0.00000620, -0.00001188, -0.00000357, 0.00000707, 0.00000465, 0.00000434, -0.00000557, -0.00000150, 0.00000965, -0.00000751, -0.00000957, 0.00000192, -0.00000500, 0.00001203, 0.00001417, -0.00002792, -0.00001317, 0.00002528, 0.00000205, -0.00000390, 0.00001088, -0.00001240, -0.00000896, 0.00001437, -0.00000889, -0.00002272, 0.00000138, 0.00000746, -0.00000630, 0.00000118, 0.00003016, 0.00002508, -0.00001268, 0.00000031, 0.00003679, 0.00001959, 0.00000410, 0.00000868, -0.00001763, -0.00002575, -0.00000073, -0.00000366, -0.00000891, 0.00000398, -0.00000128, -0.00000748, -0.00000333, -0.00000560, 0.00000143, 0.00000701, -0.00000521, -0.00000461, 0.00000581, 0.00000085, -0.00000319, -0.00000002, -0.00000134, -0.00000272, -0.00000454, -0.00000553, 0.00000106, 0.00000363, -0.00000119, -0.00000122, -0.00000088, },
		{ 0.00000331, -0.00000173, -0.00000662, -0.00000199, -0.00000046, -0.00000252, -0.00000175, -0.00000047, 0.00000226, -0.00000013, -0.00000540, 0.00000230, 0.00000653, -0.00000333, -0.00000495, -0.00000603, -0.00000806, 0.00000502, 0.00000066, -0.00001460, -0.00000026, -0.00000513, -0.00003227, -0.00001234, 0.00001139, 0.00000913, 0.00003217, 0.00003062, -0.00001115, 0.00000466, 0.00004203, 0.00001748, -0.00000694, 0.00000656, 0.00000247, -0.00002219, -0.00001712, 0.00000916, 0.00000116, -0.00001520, 0.00000461, 0.00000391, -0.00001048, 0.00002034, 0.00001570, -0.00003148, -0.00000869, 0.00002157, -0.00000433, -0.00000310, 0.00000406, -0.00001494, -0.00000021, 0.00001059, -0.00000855, -0.00000129, 0.00000858, 0.00000276, 0.00000400, -0.00000695, -0.00000823, 0.00000910, -0.00000228, -0.00001233, 0.00000401, 0.00000695, 0.00000378, 0.00000042, -0.00000616, 0.00000526, 0.00000408, -0.00001209, -0.00000370, -0.00000105, 0.00000138, 0.00001995, -0.00000662, -0.00003007, 0.00001038, 0.00001756, -0.00000669, 0.00000806, 0.00000045, -0.00001826, 0.00000817, 0.00000643, -0.00002337, -0.00000817, 0.00001054, -0.00000252, -0.00000637, 0.00001589, 0.00003184, 0.00000531, -0.00001544, 0.00001964, 0.00003459, 0.00000674, 0.00000605, 0.00000168, -0.00002586, -0.00001567, 0.00000232, -0.00000728, -0.00000331, 0.00000315, -0.00000556, -0.00000481, -0.00000405, -0.00000383, 0.00000648, 0.00000132, -0.00000857, 0.00000219, 0.00000579, -0.00000326, -0.00000157, 0.00000072, -0.00000286, -0.00000383, -0.00000542, -0.00000310, 0.00000363, 0.00000200, -0.00000188, -0.00000060, -0.00000183, -0.00000080, },
		{ -0.00000513, -0.00000498, 0.00000041, -0.00000164, -0.00000312, -0.00000066, 0.00000093, 0.00000099, -0.00000336, -0.00000221, 0.00000634, 0.00000216, -0.00000550, -0.00000477, -0.00000879, -0.00000250, 0.00000733, -0.00000933, -0.00001151, 0.00000403, -0.00001929, -0.00003115, 0.00000159, 0.00001096, 0.00001798, 0.00004023, 0.00001118, -0.00001213, 0.00002956, 0.00003806, 0.00000001, -0.00000082, 0.00000894, -0.00001426, -0.00002540, -0.00000092, 0.00000899, -0.00001117, -0.00000617, 0.00000965, -0.00000821, 0.00000156, 0.00002868, -0.00000994, -0.00002985, 0.00001371, 0.00001052, -0.00001092, 0.00000575, -0.00000487, -0.00001368, 0.00001042, 0.00000172, -0.00000984, 0.00000715, 0.00000579, 0.00000199, 0.00000151, -0.00001111, -0.00000025, 0.00000841, -0.00001078, -0.00000625, 0.00000903, 0.00000451, 0.00000224, -0.00000288, -0.00000255, 0.00000774, -0.00000371, -0.00001097, -0.00000120, -0.00000122, 0.00001270, 0.00001252, -0.00002486, -0.00001356, 0.00002083, 0.00000255, -0.00000037, 0.00001023, -0.00001511, -0.00000670, 0.00001583, -0.00001342, -0.00001977, 0.00000949, 0.00000413, -0.00000973, 0.00000434, 0.00002564, 0.00002130, -0.00001011, -0.00000461, 0.00003320, 0.00002266, 0.00000010, 0.00000811, -0.00000896, -0.00002673, -0.00000539, 0.00000098, -0.00000737, 0.00000033, -0.00000063, -0.00000579, -0.00000325, -0.00000512, 0.00000133, 0.00000644, -0.00000623, -0.00000498, 0.00000760, 0.00000110, -0.00000450, 0.00000125, -0.00000072, -0.00000449, -0.00000444, -0.00000502, 0.00000014, 0.00000439, -0.00000001, -0.00000161, -0.00000076, -0.00000238, 0.00000107, 0.00000211, },
		{ -0.00000162, 0.00000077, -0.00000357, -0.00000223, 0.00000090, 0.00000060, -0.00000160, -0.00000339, 0.00000260, 0.00000595, -0.00000237, -0.00000553, -0.00000671, -0.00000809, 0.00000453, 0.00000164, -0.00001590, -0.00000267, -0.00000128, -0.00003175, -0.00001992, 0.00000984, 0.00001178, 0.00003177, 0.00003431, -0.00000655, 0.00000543, 0.00004497, 0.00002001, -0.00000574, 0.00000813, -0.00000267, -0.00002722, -0.00001370, 0.00000904, -0.00000206, -0.00001244, 0.00000563, 0.00000172, -0.00001052, 0.00001994, 0.00001670, -0.00002936, -0.00000968, 0.00002044, -0.00000585, -0.00000439, 0.00000696, -0.00001460, -0.00000285, 0.00001204, -0.00000843, -0.00000237, 0.00001045, 0.00000115, 0.00000316, -0.00000363, -0.00001028, 0.00000746, 0.00000095, -0.00001374, 0.00000314, 0.00000904, 0.00000147, 0.00000078, -0.00000396, 0.00000212, 0.00000524, -0.00000925, -0.00000775, -0.00000060, 0.00000521, 0.00001767, -0.00000585, -0.00002665, 0.00000659, 0.00001428, -0.00000323, 0.00000995, -0.00000244, -0.00001834, 0.00001139, 0.00000373, -0.00002549, -0.00000002, 0.00001302, -0.00000904, -0.00000463, 0.00001660, 0.00002473, 0.00000623, -0.00001480, 0.00001259, 0.00003556, 0.00000787, 0.00000036, 0.00000653, -0.00001931, -0.00002101, 0.00000198, -0.00000184, -0.00000570, 0.00000068, -0.00000318, -0.00000412, -0.00000396, -0.00000321, 0.00000586, 0.00000042, -0.00000928, 0.00000289, 0.00000717, -0.00000411, -0.00000183, 0.00000227, -0.00000365, -0.00000508, -0.00000451, -0.00000337, 0.00000303, 0.00000340, -0.00000140, -0.00000107, -0.00000162, -0.00000182, 0.00000236, -0.00000091, -0.00000652, },
		{ -0.00000120, -0.00000437, 0.00000001, 0.00000115, -0.00000140, -0.00000284, -0.00000043, 0.00000544, 0.00000253, -0.00000487, -0.00000594, -0.00000867, -0.00000282, 0.00000686, -0.00000855, -0.00001383, 0.00000378, -0.00001496, -0.00003459, -0.00000540, 0.00001210, 0.00001955, 0.00004092, 0.00001651, -0.00000895, 0.00003063, 0.00004136, 0.00000220, 0.00000061, 0.00000670, -0.00002078, -0.00002519, 0.00000183, 0.00000558, -0.00001098, -0.00000303, 0.00000780, -0.00000892, 0.00000184, 0.00002753, -0.00000753, -0.00002841, 0.00001158, 0.00000945, -0.00001245, 0.00000617, -0.00000208, -0.00001533, 0.00000913, 0.00000344, -0.00001081, 0.00000755, 0.00000656, -0.00000047, 0.00000335, -0.00000938, -0.00000374, 0.00000957, -0.00000881, -0.00000860, 0.00001025, 0.00000467, -0.00000001, -0.00000026, -0.00000303, 0.00000515, -0.00000002, -0.00001152, -0.00000472, 0.00000261, 0.00001342, 0.00000987, -0.00002081, -0.00001299, 0.00001491, 0.00000298, 0.00000384, 0.00000889, -0.00001733, -0.00000388, 0.00001625, -0.00001781, -0.00001581, 0.00001706, -0.00000049, -0.00001313, 0.00000876, 0.00002087, 0.00001650, -0.00000602, -0.00000873, 0.00002817, 0.00002670, -0.00000327, 0.00000495, -0.00000020, -0.00002579, -0.00001094, 0.00000538, -0.00000494, -0.00000421, -0.00000016, -0.00000344, -0.00000353, -0.00000439, 0.00000150, 0.00000519, -0.00000689, -0.00000499, 0.00000869, 0.00000153, -0.00000544, 0.00000206, 0.00000004, -0.00000601, -0.00000454, -0.00000418, -0.00000079, 0.00000463, 0.00000146, -0.00000185, -0.00000099, -0.00000242, -0.00000035, 0.00000194, -0.00000419, -0.00000502, 0.00000125, },
		{ -0.00000288, 0.00000180, -0.00000070, -0.00000309, -0.00000147, 0.00000295, 0.00000506, -0.00000120, -0.00000596, -0.00000747, -0.00000748, 0.00000336, 0.00000170, -0.00001567, -0.00000527, 0.00000097, -0.00002902, -0.00002667, 0.00000522, 0.00001435, 0.00003247, 0.00003707, -0.00000080, 0.00000720, 0.00004670, 0.00002364, -0.00000375, 0.00000732, -0.00000822, -0.00003053, -0.00001057, 0.00000796, -0.00000474, -0.00000911, 0.00000590, -0.00000047, -0.00000936, 0.00001858, 0.00001646, -0.00002560, -0.00001040, 0.00001793, -0.00000643, -0.00000556, 0.00000882, -0.00001310, -0.00000542, 0.00001234, -0.00000743, -0.00000333, 0.00001161, 0.00000009, 0.00000197, -0.00000038, -0.00001140, 0.00000496, 0.00000383, -0.00001400, 0.00000180, 0.00001076, -0.00000041, 0.00000074, -0.00000144, -0.00000084, 0.00000559, -0.00000590, -0.00001132, -0.00000063, 0.00000922, 0.00001485, -0.00000535, -0.00002180, 0.00000251, 0.00000987, 0.00000086, 0.00001171, -0.00000556, -0.00001745, 0.00001408, 0.00000033, -0.00002653, 0.00000824, 0.00001426, -0.00001608, -0.00000209, 0.00001799, 0.00001668, 0.00000717, -0.00001238, 0.00000500, 0.00003572, 0.00001060, -0.00000622, 0.00000924, -0.00001130, -0.00002543, 0.00000052, 0.00000409, -0.00000792, -0.00000262, -0.00000037, -0.00000326, -0.00000411, -0.00000207, 0.00000490, -0.00000081, -0.00000926, 0.00000327, 0.00000800, -0.00000435, -0.00000218, 0.00000340, -0.00000403, -0.00000630, -0.00000368, -0.00000329, 0.00000206, 0.00000454, -0.00000047, -0.00000172, -0.00000154, -0.00000243, 0.00000101, -0.00000023, -0.00000577, -0.00000163, 0.00000168, -0.00000389, },
		{ 0.00000130, -0.00000334, -0.00000257, 0.00000135, 0.00000450, 0.00000268, -0.00000392, -0.00000709, -0.00000832, -0.00000288, 0.00000519, -0.00000753, -0.00001473, 0.00000202, -0.00001080, -0.00003532, -0.00001370, 0.00001071, 0.00002221, 0.00004174, 0.00002154, -0.00000415, 0.00003140, 0.00004433, 0.00000599, 0.00000083, 0.00000247, -0.00002590, -0.00002409, 0.00000340, 0.00000207, -0.00000995, -0.00000015, 0.00000532, -0.00000879, 0.00000253, 0.00002475, -0.00000494, -0.00002542, 0.00000842, 0.00000801, -0.00001284, 0.00000564, 0.00000060, -0.00001570, 0.00000684, 0.00000485, -0.00001074, 0.00000726, 0.00000723, -0.00000249, 0.00000452, -0.00000703, -0.00000688, 0.00000955, -0.00000622, -0.00001030, 0.00001057, 0.00000506, -0.00000212, 0.00000198, -0.00000291, 0.00000220, 0.00000316, -0.00001121, -0.00000825, 0.00000608, 0.00001415, 0.00000660, -0.00001616, -0.00001150, 0.00000800, 0.00000313, 0.00000847, 0.00000708, -0.00001890, -0.00000071, 0.00001554, -0.00002170, -0.00001095, 0.00002357, -0.00000596, -0.00001616, 0.00001394, 0.00001601, 0.00001088, -0.00000090, -0.00001177, 0.00002185, 0.00003107, -0.00000544, -0.00000025, 0.00000772, -0.00002292, -0.00001666, 0.00000904, -0.00000187, -0.00000914, -0.00000003, -0.00000058, -0.00000400, -0.00000350, 0.00000189, 0.00000334, -0.00000722, -0.00000456, 0.00000897, 0.00000205, -0.00000581, 0.00000231, 0.00000074, -0.00000706, -0.00000482, -0.00000317, -0.00000157, 0.00000433, 0.00000306, -0.00000178, -0.00000159, -0.00000226, -0.00000159, 0.00000122, -0.00000288, -0.00000472, 0.00000136, -0.00000046, -0.00000501, 0.00000000, },
		{ -0.00000441, 0.00000006, 0.00000345, 0.00000400, -0.00000009, -0.00000603, -0.00000824, -0.00000655, 0.00000179, 0.00000081, -0.00001417, -0.00000758, 0.00000141, -0.00002469, -0.00003173, -0.00000220, 0.00001583, 0.00003443, 0.00003941, 0.00000562, 0.00001001, 0.00004770, 0.00002817, -0.00000129, 0.00000410, -0.00001387, -0.00003198, -0.00000804, 0.00000598, -0.00000667, -0.00000549, 0.00000548, -0.00000257, -0.00000723, 0.00001651, 0.00001501, -0.00002057, -0.00001069, 0.00001415, -0.00000624, -0.00000643, 0.00000949, -0.00001062, -0.00000760, 0.00001137, -0.00000569, -0.00000394, 0.00001190, -0.00000041, 0.00000064, 0.00000245, -0.00001151, 0.00000188, 0.00000604, -0.00001308, 0.00000019, 0.00001183, -0.00000168, 0.00000036, 0.00000108, -0.00000326, 0.00000516, -0.00000246, -0.00001403, -0.00000107, 0.00001302, 0.00001185, -0.00000510, -0.00001597, -0.00000150, 0.00000459, 0.00000513, 0.00001332, -0.00000862, -0.00001569, 0.00001601, -0.00000354, -0.00002631, 0.00001614, 0.00001417, -0.00002297, 0.00000103, 0.00001970, 0.00000815, 0.00000793, -0.00000849, -0.00000258, 0.00003472, 0.00001459, -0.00001276, 0.00000964, -0.00000266, -0.00002831, -0.00000182, 0.00000983, -0.00000969, -0.00000645, 0.00000252, -0.00000218, -0.00000437, -0.00000060, 0.00000367, -0.00000231, -0.00000853, 0.00000337, 0.00000818, -0.00000391, -0.00000248, 0.00000388, -0.00000399, -0.00000728, -0.00000305, -0.00000288, 0.00000090, 0.00000526, 0.00000089, -0.00000240, -0.00000173, -0.00000261, -0.00000057, 0.00000015, -0.00000436, -0.00000193, 0.00000208, -0.00000336, -0.00000356, 0.00000195, -0.00000151, },
		{ 0.00000291, 0.00000379, 0.00000249, -0.00000278, -0.00000794, -0.00000798, -0.00000278, 0.00000269, -0.00000660, -0.00001420, -0.00000079, -0.00000748, -0.00003340, -0.00002225, 0.00000630, 0.00002520, 0.00004336, 0.00002631, 0.00000183, 0.00003236, 0.00004717, 0.00001093, -0.00000032, -0.00000353, -0.00002938, -0.00002227, 0.00000363, -0.00000125, -0.00000818, 0.00000232, 0.00000245, -0.00000795, 0.00000356, 0.00002069, -0.00000246, -0.00002110, 0.00000458, 0.00000610, -0.00001210, 0.00000434, 0.00000287, -0.00001476, 0.00000382, 0.00000570, -0.00000958, 0.00000639, 0.00000761, -0.00000383, 0.00000497, -0.00000435, -0.00000935, 0.00000838, -0.00000329, -0.00001109, 0.00000995, 0.00000555, -0.00000380, 0.00000361, -0.00000226, -0.00000071, 0.00000551, -0.00001011, -0.00001134, 0.00000886, 0.00001481, 0.00000312, -0.00001129, -0.00000923, 0.00000066, 0.00000284, 0.00001312, 0.00000509, -0.00001971, 0.00000253, 0.00001372, -0.00002479, -0.00000540, 0.00002861, -0.00001186, -0.00001849, 0.00001934, 0.00001119, 0.00000478, 0.00000472, -0.00001353, 0.00001461, 0.00003509, -0.00000612, -0.00000673, 0.00001400, -0.00001837, -0.00002175, 0.00001159, 0.00000148, -0.00001387, -0.00000031, 0.00000247, -0.00000440, -0.00000249, 0.00000238, 0.00000104, -0.00000722, -0.00000372, 0.00000846, 0.00000256, -0.00000548, 0.00000201, 0.00000119, -0.00000749, -0.00000519, -0.00000221, -0.00000210, 0.00000355, 0.00000461, -0.00000129, -0.00000252, -0.00000204, -0.00000252, 0.00000001, -0.00000135, -0.00000397, 0.00000070, 0.00000046, -0.00000482, -0.00000075, 0.00000121, -0.00000455, -0.00000264, },
		{ 0.00000293, 0.00000069, -0.00000561, -0.00000889, -0.00000565, 0.00000010, -0.00000088, -0.00001188, -0.00000917, 0.00000011, -0.00001954, -0.00003442, -0.00001175, 0.00001517, 0.00003756, 0.00004200, 0.00001228, 0.00001383, 0.00004857, 0.00003334, 0.00000117, -0.00000139, -0.00001934, -0.00003160, -0.00000633, 0.00000326, -0.00000771, -0.00000183, 0.00000448, -0.00000445, -0.00000441, 0.00001401, 0.00001254, -0.00001477, -0.00001041, 0.00000938, -0.00000549, -0.00000682, 0.00000900, -0.00000744, -0.00000912, 0.00000918, -0.00000347, -0.00000404, 0.00001125, -0.00000039, -0.00000060, 0.00000459, -0.00001067, -0.00000141, 0.00000732, -0.00001106, -0.00000144, 0.00001205, -0.00000224, -0.00000024, 0.00000328, -0.00000491, 0.00000406, 0.00000068, -0.00001561, -0.00000180, 0.00001616, 0.00000898, -0.00000498, -0.00000970, -0.00000510, -0.00000120, 0.00000912, 0.00001475, -0.00001129, -0.00001323, 0.00001701, -0.00000764, -0.00002473, 0.00002325, 0.00001277, -0.00002913, 0.00000452, 0.00002134, -0.00000035, 0.00000833, -0.00000356, -0.00000958, 0.00003239, 0.00001928, -0.00001840, 0.00000785, 0.00000565, -0.00002920, -0.00000463, 0.00001466, -0.00001083, -0.00001033, 0.00000511, -0.00000087, -0.00000455, 0.00000098, 0.00000222, -0.00000395, -0.00000721, 0.00000323, 0.00000769, -0.00000287, -0.00000259, 0.00000363, -0.00000365, -0.00000781, -0.00000271, -0.00000228, -0.00000028, 0.00000549, 0.00000255, -0.00000294, -0.00000228, -0.00000246, -0.00000217, 0.00000015, -0.00000236, -0.00000238, 0.00000181, -0.00000196, -0.00000408, 0.00000114, -0.00000179, -0.00000533, 0.00000094, 0.00000416, },
		{ -0.00000161, -0.00000818, -0.00000785, -0.00000268, -0.00000018, -0.00000605, -0.00001243, -0.00000403, -0.00000554, -0.00002922, -0.00002991, -0.00000131, 0.00002751, 0.00004631, 0.00003103, 0.00000843, 0.00003399, 0.00005018, 0.00001636, -0.00000296, -0.00001089, -0.00003114, -0.00001995, 0.00000253, -0.00000417, -0.00000580, 0.00000423, -0.00000056, -0.00000660, 0.00000482, 0.00001583, -0.00000031, -0.00001584, 0.00000047, 0.00000372, -0.00001042, 0.00000254, 0.00000453, -0.00001261, 0.00000039, 0.00000582, -0.00000744, 0.00000512, 0.00000754, -0.00000437, 0.00000479, -0.00000168, -0.00001090, 0.00000620, -0.00000040, -0.00001083, 0.00000848, 0.00000596, -0.00000485, 0.00000451, -0.00000124, -0.00000321, 0.00000680, -0.00000843, -0.00001359, 0.00001071, 0.00001525, -0.00000011, -0.00000656, -0.00000643, -0.00000651, 0.00000200, 0.00001739, 0.00000321, -0.00001966, 0.00000556, 0.00001090, -0.00002683, 0.00000059, 0.00003187, -0.00001773, -0.00001986, 0.00002441, 0.00000654, -0.00000140, 0.00001025, -0.00001396, 0.00000698, 0.00003809, -0.00000528, -0.00001356, 0.00001804, -0.00001263, -0.00002545, 0.00001278, 0.00000461, -0.00001784, -0.00000098, 0.00000535, -0.00000448, -0.00000141, 0.00000277, -0.00000149, -0.00000691, -0.00000256, 0.00000727, 0.00000301, -0.00000446, 0.00000130, 0.00000123, -0.00000728, -0.00000551, -0.00000144, -0.00000234, 0.00000244, 0.00000591, -0.00000036, -0.00000362, -0.00000196, -0.00000305, -0.00000153, 0.00000019, -0.00000274, -0.00000064, 0.00000139, -0.00000367, -0.00000218, 0.00000068, -0.00000483, -0.00000304, 0.00000407, 0.00000350, 0.00000190, },
		{ -0.00000925, -0.00000516, -0.00000146, -0.00000310, -0.00000935, -0.00000972, -0.00000261, -0.00001448, -0.00003434, -0.00002248, 0.00001144, 0.00004144, 0.00004560, 0.00001883, 0.00001850, 0.00005002, 0.00003881, 0.00000305, -0.00000876, -0.00002434, -0.00002960, -0.00000557, -0.00000001, -0.00000783, 0.00000160, 0.00000306, -0.00000599, -0.00000128, 0.00001134, 0.00000936, -0.00000875, -0.00000955, 0.00000402, -0.00000448, -0.00000667, 0.00000752, -0.00000394, -0.00000980, 0.00000599, -0.00000108, -0.00000350, 0.00000974, 0.00000001, -0.00000153, 0.00000589, -0.00000903, -0.00000453, 0.00000753, -0.00000816, -0.00000283, 0.00001133, -0.00000213, -0.00000090, 0.00000489, -0.00000566, 0.00000251, 0.00000312, -0.00001592, -0.00000262, 0.00001827, 0.00000650, -0.00000482, -0.00000354, -0.00000802, -0.00000707, 0.00001236, 0.00001595, -0.00001322, -0.00001028, 0.00001697, -0.00001169, -0.00002182, 0.00002917, 0.00001016, -0.00003407, 0.00000813, 0.00002256, -0.00000832, 0.00000831, 0.00000186, -0.00001546, 0.00002877, 0.00002399, -0.00002247, 0.00000429, 0.00001275, -0.00002790, -0.00000739, 0.00001798, -0.00001128, -0.00001377, 0.00000700, 0.00000059, -0.00000436, 0.00000242, 0.00000056, -0.00000550, -0.00000549, 0.00000287, 0.00000663, -0.00000138, -0.00000242, 0.00000270, -0.00000316, -0.00000777, -0.00000261, -0.00000165, -0.00000132, 0.00000524, 0.00000432, -0.00000317, -0.00000317, -0.00000212, -0.00000360, -0.00000023, 0.00000001, -0.00000283, 0.00000084, 0.00000016, -0.00000425, -0.00000070, -0.00000164, -0.00000568, 0.00000104, 0.00000504, 0.00000204, 0.00000091, -0.00000460, },
		{ -0.00000282, -0.00000291, -0.00000607, -0.00000981, -0.00000703, -0.00000528, -0.00002344, -0.00003559, -0.00001191, 0.00002797, 0.00005093, 0.00003610, 0.00001510, 0.00003673, 0.00005369, 0.00002150, -0.00000716, -0.00001905, -0.00003126, -0.00001738, 0.00000019, -0.00000651, -0.00000302, 0.00000549, -0.00000341, -0.00000495, 0.00000612, 0.00001068, 0.00000133, -0.00001012, -0.00000348, 0.00000093, -0.00000809, 0.00000055, 0.00000543, -0.00000951, -0.00000304, 0.00000511, -0.00000454, 0.00000368, 0.00000692, -0.00000412, 0.00000412, 0.00000069, -0.00001142, 0.00000330, 0.00000215, -0.00000950, 0.00000633, 0.00000611, -0.00000513, 0.00000468, -0.00000005, -0.00000503, 0.00000698, -0.00000643, -0.00001470, 0.00001151, 0.00001530, -0.00000270, -0.00000224, -0.00000341, -0.00001294, 0.00000060, 0.00002085, 0.00000173, -0.00001869, 0.00000811, 0.00000726, -0.00002766, 0.00000674, 0.00003318, -0.00002319, -0.00002007, 0.00002867, 0.00000215, -0.00000723, 0.00001519, -0.00001314, -0.00000043, 0.00003956, -0.00000313, -0.00001984, 0.00001953, -0.00000635, -0.00002712, 0.00001259, 0.00000699, -0.00002056, -0.00000194, 0.00000763, -0.00000400, -0.00000022, 0.00000280, -0.00000399, -0.00000628, -0.00000130, 0.00000560, 0.00000333, -0.00000288, 0.00000034, 0.00000078, -0.00000653, -0.00000557, -0.00000097, -0.00000237, 0.00000121, 0.00000683, 0.00000093, -0.00000471, -0.00000214, -0.00000324, -0.00000318, 0.00000157, -0.00000108, -0.00000246, 0.00000214, -0.00000158, -0.00000412, -0.00000044, -0.00000415, -0.00000365, 0.00000463, 0.00000379, 0.00000092, -0.00000060, -0.00000755, -0.00000806, },
		{ -0.00000548, -0.00000713, -0.00000910, -0.00000621, -0.00001036, -0.00003142, -0.00003319, 0.00000396, 0.00004533, 0.00005092, 0.00002509, 0.00002382, 0.00005276, 0.00004421, 0.00000367, -0.00001750, -0.00002856, -0.00002634, -0.00000578, -0.00000358, -0.00000706, 0.00000457, 0.00000138, -0.00000710, 0.00000181, 0.00000873, 0.00000583, -0.00000307, -0.00000813, -0.00000148, -0.00000349, -0.00000598, 0.00000532, -0.00000048, -0.00000958, 0.00000211, 0.00000116, -0.00000232, 0.00000755, 0.00000060, -0.00000198, 0.00000631, -0.00000686, -0.00000716, 0.00000668, -0.00000471, -0.00000376, 0.00000968, -0.00000145, -0.00000143, 0.00000575, -0.00000552, 0.00000077, 0.00000461, -0.00001495, -0.00000329, 0.00001908, 0.00000455, -0.00000444, 0.00000202, -0.00001012, -0.00001256, 0.00001447, 0.00001683, -0.00001412, -0.00000706, 0.00001585, -0.00001538, -0.00001772, 0.00003364, 0.00000650, -0.00003740, 0.00001163, 0.00002301, -0.00001533, 0.00000787, 0.00000723, -0.00001978, 0.00002408, 0.00002798, -0.00002454, -0.00000043, 0.00001785, -0.00002449, -0.00000952, 0.00001930, -0.00001114, -0.00001623, 0.00000793, 0.00000205, -0.00000351, 0.00000350, -0.00000132, -0.00000671, -0.00000357, 0.00000230, 0.00000521, 0.00000036, -0.00000196, 0.00000126, -0.00000274, -0.00000711, -0.00000261, -0.00000119, -0.00000213, 0.00000464, 0.00000596, -0.00000300, -0.00000429, -0.00000183, -0.00000470, -0.00000088, 0.00000249, -0.00000313, -0.00000073, 0.00000275, -0.00000392, -0.00000348, -0.00000122, -0.00000532, 0.00000034, 0.00000599, 0.00000150, 0.00000043, -0.00000302, -0.00000914, -0.00000542, -0.00000277, },
		{ -0.00000682, -0.00000918, -0.00000678, -0.00001696, -0.00003842, -0.00002488, 0.00002543, 0.00005721, 0.00004207, 0.00002141, 0.00004095, 0.00005806, 0.00002548, -0.00001292, -0.00002732, -0.00002995, -0.00001486, -0.00000310, -0.00000816, -0.00000005, 0.00000609, -0.00000585, -0.00000320, 0.00000727, 0.00000574, 0.00000238, -0.00000446, -0.00000689, -0.00000213, -0.00000546, -0.00000132, 0.00000558, -0.00000581, -0.00000612, 0.00000359, -0.00000122, 0.00000234, 0.00000576, -0.00000318, 0.00000319, 0.00000251, -0.00001095, 0.00000004, 0.00000406, -0.00000722, 0.00000380, 0.00000586, -0.00000462, 0.00000419, 0.00000111, -0.00000597, 0.00000611, -0.00000438, -0.00001447, 0.00001127, 0.00001478, -0.00000434, 0.00000150, -0.00000051, -0.00001816, -0.00000131, 0.00002312, 0.00000087, -0.00001677, 0.00000993, 0.00000304, -0.00002721, 0.00001271, 0.00003254, -0.00002788, -0.00001907, 0.00003172, -0.00000193, -0.00001231, 0.00001914, -0.00001128, -0.00000702, 0.00003916, -0.00000011, -0.00002474, 0.00001847, -0.00000025, -0.00002635, 0.00001119, 0.00000811, -0.00002169, -0.00000298, 0.00000893, -0.00000280, 0.00000115, 0.00000223, -0.00000625, -0.00000531, -0.00000016, 0.00000365, 0.00000354, -0.00000098, -0.00000064, -0.00000013, -0.00000545, -0.00000521, -0.00000084, -0.00000235, 0.00000008, 0.00000729, 0.00000239, -0.00000553, -0.00000267, -0.00000320, -0.00000466, 0.00000266, 0.00000082, -0.00000447, 0.00000258, 0.00000127, -0.00000628, -0.00000222, -0.00000263, -0.00000431, 0.00000449, 0.00000464, -0.00000034, 0.00000006, -0.00000593, -0.00000862, -0.00000344, -0.00000507, -0.00000673, },
		{ -0.00001002, -0.00000790, -0.00002595, -0.00004263, -0.00000755, 0.00004825, 0.00005846, 0.00003116, 0.00002958, 0.00005739, 0.00004917, 0.00000236, -0.00002696, -0.00003172, -0.00002232, -0.00000687, -0.00000718, -0.00000558, 0.00000694, -0.00000035, -0.00000769, 0.00000450, 0.00000635, 0.00000233, 0.00000182, -0.00000631, -0.00000661, -0.00000278, -0.00000484, 0.00000277, 0.00000257, -0.00000850, -0.00000203, 0.00000294, -0.00000057, 0.00000497, 0.00000120, -0.00000188, 0.00000595, -0.00000447, -0.00000904, 0.00000491, -0.00000110, -0.00000408, 0.00000725, -0.00000041, -0.00000170, 0.00000579, -0.00000461, -0.00000090, 0.00000500, -0.00001287, -0.00000358, 0.00001845, 0.00000317, -0.00000364, 0.00000658, -0.00001135, -0.00001723, 0.00001519, 0.00001730, -0.00001376, -0.00000379, 0.00001368, -0.00001848, -0.00001269, 0.00003645, 0.00000207, -0.00003892, 0.00001484, 0.00002248, -0.00002107, 0.00000710, 0.00001201, -0.00002230, 0.00001867, 0.00003062, -0.00002447, -0.00000559, 0.00002044, -0.00001927, -0.00001048, 0.00001839, -0.00001062, -0.00001726, 0.00000773, 0.00000331, -0.00000177, 0.00000404, -0.00000345, -0.00000732, -0.00000167, 0.00000146, 0.00000368, 0.00000210, -0.00000124, -0.00000041, -0.00000257, -0.00000590, -0.00000251, -0.00000109, -0.00000268, 0.00000389, 0.00000722, -0.00000241, -0.00000543, -0.00000181, -0.00000537, -0.00000162, 0.00000475, -0.00000320, -0.00000266, 0.00000547, -0.00000303, -0.00000694, -0.00000078, -0.00000421, -0.00000104, 0.00000686, 0.00000143, -0.00000074, -0.00000099, -0.00000839, -0.00000663, -0.00000379, -0.00000761, -0.00000569, -0.00000735, },
		{ -0.00001074, -0.00003783, -0.00003928, 0.00001894, 0.00006476, 0.00004951, 0.00002714, 0.00004689, 0.00006356, 0.00002750, -0.00002023, -0.00003494, -0.00002753, -0.00001268, -0.00000700, -0.00000911, 0.00000284, 0.00000609, -0.00000769, -0.00000153, 0.00000809, 0.00000143, 0.00000284, 0.00000065, -0.00000945, -0.00000525, -0.00000294, -0.00000282, 0.00000505, -0.00000196, -0.00000851, 0.00000142, 0.00000213, 0.00000130, 0.00000413, -0.00000176, 0.00000222, 0.00000362, -0.00000964, -0.00000322, 0.00000516, -0.00000423, 0.00000120, 0.00000512, -0.00000341, 0.00000324, 0.00000203, -0.00000596, 0.00000439, -0.00000257, -0.00001288, 0.00001016, 0.00001354, -0.00000485, 0.00000459, 0.00000197, -0.00002182, -0.00000359, 0.00002390, 0.00000076, -0.00001390, 0.00001085, -0.00000149, -0.00002549, 0.00001817, 0.00003008, -0.00003153, -0.00001689, 0.00003333, -0.00000566, -0.00001631, 0.00002181, -0.00000868, -0.00001227, 0.00003679, 0.00000323, -0.00002765, 0.00001515, 0.00000494, -0.00002299, 0.00000890, 0.00000757, -0.00002109, -0.00000382, 0.00000893, -0.00000083, 0.00000281, 0.00000084, -0.00000806, -0.00000392, 0.00000057, 0.00000168, 0.00000371, 0.00000093, -0.00000144, -0.00000139, -0.00000432, -0.00000425, -0.00000097, -0.00000250, -0.00000072, 0.00000731, 0.00000376, -0.00000586, -0.00000352, -0.00000317, -0.00000569, 0.00000338, 0.00000269, -0.00000634, 0.00000266, 0.00000456, -0.00000831, -0.00000464, -0.00000056, -0.00000481, 0.00000364, 0.00000595, -0.00000171, 0.00000027, -0.00000333, -0.00000948, -0.00000476, -0.00000636, -0.00000792, -0.00000404, -0.00001001, -0.00000980, },
		{ -0.00004964, -0.00002292, 0.00004907, 0.00006842, 0.00003737, 0.00003564, 0.00006434, 0.00005335, -0.00000150, -0.00003646, -0.00003356, -0.00001813, -0.00000870, -0.00001050, -0.00000363, 0.00000859, -0.00000193, -0.00000776, 0.00000654, 0.00000432, -0.00000078, 0.00000554, -0.00000427, -0.00001093, -0.00000253, -0.00000344, 0.00000024, 0.00000494, -0.00000674, -0.00000599, 0.00000402, 0.00000155, 0.00000232, 0.00000163, -0.00000124, 0.00000498, -0.00000217, -0.00001003, 0.00000248, 0.00000229, -0.00000370, 0.00000429, 0.00000076, -0.00000159, 0.00000508, -0.00000314, -0.00000225, 0.00000429, -0.00000994, -0.00000333, 0.00001642, 0.00000230, -0.00000228, 0.00000992, -0.00001176, -0.00002075, 0.00001439, 0.00001723, -0.00001203, -0.00000066, 0.00001055, -0.00002075, -0.00000704, 0.00003752, -0.00000283, -0.00003857, 0.00001757, 0.00002085, -0.00002531, 0.00000611, 0.00001580, -0.00002294, 0.00001300, 0.00003142, -0.00002240, -0.00001041, 0.00002027, -0.00001279, -0.00000986, 0.00001525, -0.00001001, -0.00001655, 0.00000640, 0.00000414, 0.00000104, 0.00000397, -0.00000591, -0.00000710, 0.00000001, 0.00000031, 0.00000236, 0.00000364, -0.00000043, -0.00000198, -0.00000279, -0.00000429, -0.00000205, -0.00000148, -0.00000305, 0.00000327, 0.00000788, -0.00000148, -0.00000629, -0.00000228, -0.00000564, -0.00000215, 0.00000649, -0.00000302, -0.00000462, 0.00000794, -0.00000160, -0.00001065, -0.00000061, -0.00000243, -0.00000282, 0.00000744, 0.00000183, -0.00000249, 0.00000131, -0.00000690, -0.00000883, -0.00000460, -0.00000910, -0.00000533, -0.00000474, -0.00001334, -0.00000754, -0.00001857, },
		{ 0.00000784, 0.00007274, 0.00005893, 0.00003244, 0.00005464, 0.00007033, 0.00002683, -0.00002899, -0.00004117, -0.00002438, -0.00001116, -0.00001104, -0.00000941, 0.00000539, 0.00000561, -0.00000878, -0.00000010, 0.00000845, -0.00000194, 0.00000278, 0.00000475, -0.00001098, -0.00000816, -0.00000086, -0.00000379, 0.00000402, 0.00000163, -0.00001000, -0.00000118, 0.00000510, 0.00000071, 0.00000220, -0.00000011, 0.00000140, 0.00000401, -0.00000774, -0.00000609, 0.00000539, -0.00000086, -0.00000115, 0.00000388, -0.00000169, 0.00000204, 0.00000257, -0.00000507, 0.00000211, -0.00000121, -0.00001005, 0.00000840, 0.00001152, -0.00000424, 0.00000705, 0.00000378, -0.00002377, -0.00000605, 0.00002303, 0.00000144, -0.00001016, 0.00001077, -0.00000604, -0.00002261, 0.00002281, 0.00002604, -0.00003395, -0.00001367, 0.00003339, -0.00000903, -0.00001901, 0.00002310, -0.00000566, -0.00001580, 0.00003259, 0.00000632, -0.00002822, 0.00001012, 0.00000864, -0.00001721, 0.00000617, 0.00000514, -0.00001885, -0.00000417, 0.00000746, 0.00000184, 0.00000487, -0.00000151, -0.00000932, -0.00000206, 0.00000067, -0.00000015, 0.00000396, 0.00000250, -0.00000189, -0.00000279, -0.00000346, -0.00000265, -0.00000120, -0.00000307, -0.00000105, 0.00000701, 0.00000472, -0.00000552, -0.00000458, -0.00000341, -0.00000603, 0.00000378, 0.00000419, -0.00000772, 0.00000242, 0.00000782, -0.00000986, -0.00000753, 0.00000167, -0.00000491, 0.00000217, 0.00000745, -0.00000298, -0.00000000, 0.00000008, -0.00001045, -0.00000694, -0.00000664, -0.00000942, -0.00000197, -0.00000928, -0.00001388, -0.00000570, -0.00003370, -0.00005387, },
		{ 0.00008053, 0.00004434, 0.00004206, 0.00007373, 0.00005641, -0.00000837, -0.00004534, -0.00003387, -0.00001438, -0.00001105, -0.00001322, -0.00000153, 0.00000949, -0.00000318, -0.00000735, 0.00000775, 0.00000271, -0.00000323, 0.00000790, -0.00000224, -0.00001408, -0.00000280, -0.00000199, -0.00000193, 0.00000647, -0.00000454, -0.00000935, 0.00000429, 0.00000378, -0.00000006, 0.00000178, -0.00000019, 0.00000365, -0.00000023, -0.00001011, -0.00000027, 0.00000510, -0.00000268, 0.00000110, 0.00000182, -0.00000107, 0.00000376, -0.00000137, -0.00000312, 0.00000264, -0.00000653, -0.00000243, 0.00001319, 0.00000179, -0.00000032, 0.00001194, -0.00001148, -0.00002287, 0.00001213, 0.00001654, -0.00000893, 0.00000219, 0.00000662, -0.00002204, -0.00000112, 0.00003687, -0.00000783, -0.00003643, 0.00001970, 0.00001809, -0.00002800, 0.00000506, 0.00001829, -0.00002179, 0.00000754, 0.00003012, -0.00001872, -0.00001421, 0.00001744, -0.00000574, -0.00000743, 0.00001016, -0.00000965, -0.00001396, 0.00000413, 0.00000431, 0.00000494, 0.00000331, -0.00000876, -0.00000589, 0.00000137, -0.00000124, 0.00000153, 0.00000480, 0.00000028, -0.00000309, -0.00000348, -0.00000255, -0.00000099, -0.00000244, -0.00000339, 0.00000304, 0.00000781, -0.00000039, -0.00000658, -0.00000338, -0.00000561, -0.00000218, 0.00000743, -0.00000266, -0.00000620, 0.00000980, 0.00000019, -0.00001411, -0.00000095, -0.00000021, -0.00000461, 0.00000754, 0.00000264, -0.00000452, 0.00000361, -0.00000471, -0.00001187, -0.00000537, -0.00000965, -0.00000612, -0.00000168, -0.00001558, -0.00000939, -0.00001016, -0.00005329, -0.00004143, 0.00004661, },
		{ 0.00003782, 0.00006418, 0.00007827, 0.00002296, -0.00003906, -0.00004534, -0.00002091, -0.00001059, -0.00001474, -0.00000913, 0.00000732, 0.00000484, -0.00000907, 0.00000096, 0.00000828, -0.00000420, 0.00000229, 0.00000755, -0.00001144, -0.00001061, 0.00000051, -0.00000414, 0.00000272, 0.00000459, -0.00001049, -0.00000389, 0.00000733, 0.00000065, 0.00000021, 0.00000148, 0.00000089, 0.00000373, -0.00000555, -0.00000828, 0.00000480, 0.00000250, -0.00000300, 0.00000223, 0.00000028, 0.00000083, 0.00000266, -0.00000349, -0.00000040, -0.00000045, -0.00000628, 0.00000631, 0.00000876, -0.00000264, 0.00000898, 0.00000479, -0.00002403, -0.00000847, 0.00002051, 0.00000282, -0.00000566, 0.00000972, -0.00001033, -0.00001876, 0.00002636, 0.00002079, -0.00003501, -0.00000966, 0.00003191, -0.00001204, -0.00002033, 0.00002304, -0.00000260, -0.00001741, 0.00002688, 0.00000858, -0.00002642, 0.00000411, 0.00001044, -0.00000947, 0.00000349, 0.00000081, -0.00001524, -0.00000375, 0.00000452, 0.00000500, 0.00000742, -0.00000483, -0.00001000, 0.00000034, -0.00000005, -0.00000168, 0.00000445, 0.00000341, -0.00000192, -0.00000403, -0.00000317, -0.00000042, -0.00000130, -0.00000427, -0.00000084, 0.00000658, 0.00000497, -0.00000444, -0.00000564, -0.00000421, -0.00000552, 0.00000399, 0.00000494, -0.00000834, 0.00000198, 0.00001054, -0.00001063, -0.00001061, 0.00000356, -0.00000444, 0.00000030, 0.00000879, -0.00000393, -0.00000069, 0.00000396, -0.00001133, -0.00000997, -0.00000601, -0.00001090, -0.00000100, -0.00000707, -0.00001837, -0.00000255, -0.00002637, -0.00006730, -0.00000803, 0.00007992, 0.00007067, },
		{ 0.00008537, 0.00005801, -0.00001850, -0.00005302, -0.00003250, -0.00001165, -0.00001369, -0.00001507, 0.00000036, 0.00000966, -0.00000391, -0.00000655, 0.00000807, 0.00000155, -0.00000485, 0.00000882, -0.00000045, -0.00001586, -0.00000357, -0.00000072, -0.00000346, 0.00000710, -0.00000219, -0.00001175, 0.00000372, 0.00000582, -0.00000186, 0.00000161, 0.00000111, 0.00000223, 0.00000117, -0.00000940, -0.00000298, 0.00000709, -0.00000116, -0.00000195, 0.00000258, -0.00000021, 0.00000208, 0.00000042, -0.00000342, 0.00000032, -0.00000300, -0.00000089, 0.00000910, 0.00000148, 0.00000222, 0.00001274, -0.00001069, -0.00002352, 0.00000862, 0.00001519, -0.00000462, 0.00000468, 0.00000212, -0.00002228, 0.00000467, 0.00003460, -0.00001257, -0.00003274, 0.00002111, 0.00001434, -0.00002920, 0.00000407, 0.00001936, -0.00001913, 0.00000273, 0.00002670, -0.00001400, -0.00001647, 0.00001236, 0.00000116, -0.00000320, 0.00000362, -0.00000984, -0.00000958, 0.00000121, 0.00000366, 0.00000983, 0.00000218, -0.00001201, -0.00000365, 0.00000234, -0.00000328, 0.00000143, 0.00000550, 0.00000065, -0.00000344, -0.00000460, -0.00000100, 0.00000088, -0.00000390, -0.00000390, 0.00000347, 0.00000695, 0.00000060, -0.00000601, -0.00000515, -0.00000549, -0.00000140, 0.00000741, -0.00000231, -0.00000696, 0.00001076, 0.00000208, -0.00001675, -0.00000201, 0.00000212, -0.00000596, 0.00000700, 0.00000365, -0.00000645, 0.00000565, -0.00000194, -0.00001544, -0.00000633, -0.00000914, -0.00000790, 0.00000128, -0.00001628, -0.00001317, -0.00000179, -0.00005303, -0.00006188, 0.00003982, 0.00009400, 0.00005286, 0.00004914, },
		{ 0.00001561, -0.00005022, -0.00004698, -0.00001749, -0.00001120, -0.00001763, -0.00000842, 0.00000840, 0.00000401, -0.00000859, 0.00000154, 0.00000761, -0.00000530, 0.00000150, 0.00000892, -0.00001091, -0.00001239, 0.00000101, -0.00000393, 0.00000139, 0.00000664, -0.00001001, -0.00000640, 0.00000853, 0.00000107, -0.00000156, 0.00000279, 0.00000073, 0.00000296, -0.00000340, -0.00000962, 0.00000357, 0.00000549, -0.00000416, 0.00000033, 0.00000223, -0.00000018, 0.00000233, -0.00000150, -0.00000280, -0.00000030, -0.00000195, 0.00000420, 0.00000538, -0.00000032, 0.00001049, 0.00000498, -0.00002279, -0.00001064, 0.00001649, 0.00000475, -0.00000058, 0.00000780, -0.00001414, -0.00001419, 0.00002861, 0.00001474, -0.00003470, -0.00000516, 0.00002903, -0.00001467, -0.00002030, 0.00002178, 0.00000020, -0.00001712, 0.00002015, 0.00000960, -0.00002247, -0.00000209, 0.00001015, -0.00000047, 0.00000133, -0.00000521, -0.00001072, -0.00000236, 0.00000032, 0.00000838, 0.00001050, -0.00000900, -0.00001015, 0.00000330, -0.00000166, -0.00000286, 0.00000531, 0.00000341, -0.00000152, -0.00000480, -0.00000368, 0.00000231, -0.00000100, -0.00000627, -0.00000010, 0.00000627, 0.00000425, -0.00000262, -0.00000642, -0.00000583, -0.00000414, 0.00000420, 0.00000463, -0.00000801, 0.00000159, 0.00001218, -0.00001041, -0.00001348, 0.00000457, -0.00000328, -0.00000162, 0.00000953, -0.00000435, -0.00000158, 0.00000783, -0.00001187, -0.00001370, -0.00000476, -0.00001198, -0.00000134, -0.00000370, -0.00002250, -0.00000173, -0.00001664, -0.00007820, -0.00002828, 0.00008477, 0.00008478, 0.00004425, 0.00007540, 0.00008699, },
		{ -0.00005906, -0.00002941, -0.00001041, -0.00001638, -0.00001579, 0.00000172, 0.00000916, -0.00000397, -0.00000554, 0.00000752, 0.00000085, -0.00000560, 0.00000842, 0.00000092, -0.00001621, -0.00000471, 0.00000019, -0.00000422, 0.00000689, 0.00000000, -0.00001299, 0.00000242, 0.00000738, -0.00000289, 0.00000118, 0.00000241, 0.00000098, 0.00000194, -0.00000810, -0.00000532, 0.00000810, 0.00000064, -0.00000453, 0.00000290, 0.00000090, 0.00000032, 0.00000198, -0.00000317, -0.00000233, 0.00000026, 0.00000118, 0.00000457, 0.00000118, 0.00000517, 0.00001253, -0.00000959, -0.00002277, 0.00000417, 0.00001318, 0.00000060, 0.00000678, -0.00000265, -0.00002149, 0.00000997, 0.00003092, -0.00001671, -0.00002780, 0.00002175, 0.00000979, -0.00002906, 0.00000327, 0.00001903, -0.00001534, -0.00000108, 0.00002137, -0.00000890, -0.00001686, 0.00000568, 0.00000719, 0.00000260, -0.00000369, -0.00001082, -0.00000372, -0.00000193, 0.00000210, 0.00001543, 0.00000078, -0.00001565, -0.00000046, 0.00000294, -0.00000583, 0.00000217, 0.00000571, 0.00000044, -0.00000282, -0.00000603, 0.00000004, 0.00000367, -0.00000573, -0.00000483, 0.00000474, 0.00000538, 0.00000120, -0.00000435, -0.00000750, -0.00000553, 0.00000042, 0.00000637, -0.00000218, -0.00000650, 0.00001063, 0.00000371, -0.00001801, -0.00000385, 0.00000414, -0.00000640, 0.00000573, 0.00000457, -0.00000783, 0.00000718, 0.00000112, -0.00001903, -0.00000769, -0.00000762, -0.00001028, 0.00000358, -0.00001529, -0.00001827, 0.00000547, -0.00004882, -0.00008266, 0.00002793, 0.00010751, 0.00006385, 0.00005751, 0.00009871, 0.00005781, -0.00003182, },
		{ -0.00001443, -0.00001314, -0.00001934, -0.00000738, 0.00000843, 0.00000335, -0.00000741, 0.00000158, 0.00000653, -0.00000535, 0.00000053, 0.00000887, -0.00000958, -0.00001334, 0.00000065, -0.00000327, 0.00000026, 0.00000765, -0.00000870, -0.00000840, 0.00000858, 0.00000185, -0.00000286, 0.00000362, 0.00000091, 0.00000191, -0.00000155, -0.00001004, 0.00000194, 0.00000777, -0.00000455, -0.00000160, 0.00000387, -0.00000085, 0.00000166, 0.00000062, -0.00000480, -0.00000071, 0.00000247, 0.00000234, 0.00000163, 0.00000235, 0.00001171, 0.00000447, -0.00002037, -0.00001240, 0.00001131, 0.00000699, 0.00000481, 0.00000523, -0.00001726, -0.00000920, 0.00002944, 0.00000834, -0.00003304, -0.00000050, 0.00002500, -0.00001692, -0.00001909, 0.00001957, 0.00000250, -0.00001511, 0.00001296, 0.00000909, -0.00001687, -0.00000768, 0.00000788, 0.00000895, 0.00000007, -0.00001246, -0.00000585, 0.00000008, -0.00000479, 0.00001160, 0.00001404, -0.00001378, -0.00000990, 0.00000676, -0.00000412, -0.00000369, 0.00000668, 0.00000236, -0.00000078, -0.00000481, -0.00000510, 0.00000531, -0.00000005, -0.00000913, 0.00000102, 0.00000635, 0.00000240, -0.00000021, -0.00000662, -0.00000842, -0.00000197, 0.00000467, 0.00000298, -0.00000668, 0.00000155, 0.00001229, -0.00000912, -0.00001568, 0.00000424, -0.00000142, -0.00000318, 0.00000925, -0.00000411, -0.00000238, 0.00001114, -0.00001186, -0.00001782, -0.00000335, -0.00001232, -0.00000293, 0.00000021, -0.00002554, -0.00000331, -0.00000571, -0.00008538, -0.00005193, 0.00008555, 0.00010089, 0.00005308, 0.00008820, 0.00009576, 0.00000479, -0.00006211, -0.00004583, },
		{ -0.00001892, -0.00001521, 0.00000224, 0.00000810, -0.00000326, -0.00000448, 0.00000621, 0.00000058, -0.00000557, 0.00000690, 0.00000176, -0.00001526, -0.00000603, 0.00000061, -0.00000418, 0.00000602, 0.00000180, -0.00001299, 0.00000058, 0.00000820, -0.00000305, 0.00000059, 0.00000351, 0.00000008, 0.00000209, -0.00000645, -0.00000701, 0.00000814, 0.00000247, -0.00000639, 0.00000275, 0.00000207, -0.00000126, 0.00000313, -0.00000248, -0.00000493, 0.00000295, 0.00000358, 0.00000010, 0.00000075, 0.00000830, 0.00001162, -0.00000836, -0.00002086, -0.00000078, 0.00001059, 0.00000634, 0.00000847, -0.00000738, -0.00001977, 0.00001446, 0.00002609, -0.00001995, -0.00002198, 0.00002157, 0.00000473, -0.00002782, 0.00000271, 0.00001748, -0.00001086, -0.00000368, 0.00001456, -0.00000410, -0.00001533, -0.00000178, 0.00001177, 0.00000948, -0.00001100, -0.00001269, 0.00000313, -0.00000483, -0.00000032, 0.00002134, -0.00000065, -0.00001955, 0.00000346, 0.00000326, -0.00000889, 0.00000379, 0.00000552, -0.00000054, -0.00000112, -0.00000756, 0.00000028, 0.00000736, -0.00000770, -0.00000637, 0.00000691, 0.00000328, 0.00000111, -0.00000151, -0.00001025, -0.00000604, 0.00000341, 0.00000442, -0.00000255, -0.00000454, 0.00000940, 0.00000465, -0.00001744, -0.00000644, 0.00000536, -0.00000553, 0.00000375, 0.00000503, -0.00000817, 0.00000801, 0.00000405, -0.00002207, -0.00000964, -0.00000537, -0.00001269, 0.00000473, -0.00001274, -0.00002377, 0.00001063, -0.00004111, -0.00010187, 0.00001056, 0.00011924, 0.00007821, 0.00006816, 0.00011286, 0.00005538, -0.00004793, -0.00006316, -0.00002470, -0.00001093, },
		{ -0.00000616, 0.00000728, 0.00000306, -0.00000566, 0.00000103, 0.00000522, -0.00000453, -0.00000055, 0.00000761, -0.00000772, -0.00001344, -0.00000043, -0.00000233, -0.00000053, 0.00000763, -0.00000680, -0.00000967, 0.00000749, 0.00000280, -0.00000350, 0.00000391, 0.00000133, 0.00000082, -0.00000022, -0.00000960, 0.00000019, 0.00000912, -0.00000423, -0.00000333, 0.00000498, -0.00000110, 0.00000081, 0.00000254, -0.00000620, -0.00000151, 0.00000651, 0.00000097, -0.00000217, 0.00000497, 0.00001273, 0.00000347, -0.00001719, -0.00001364, 0.00000540, 0.00000926, 0.00001023, 0.00000230, -0.00001953, -0.00000414, 0.00002882, 0.00000201, -0.00003017, 0.00000396, 0.00002010, -0.00001876, -0.00001693, 0.00001671, 0.00000415, -0.00001173, 0.00000588, 0.00000700, -0.00001029, -0.00001199, 0.00000395, 0.00001786, -0.00000008, -0.00002033, -0.00000121, 0.00000354, -0.00001027, 0.00001431, 0.00001790, -0.00001879, -0.00000940, 0.00001062, -0.00000728, -0.00000424, 0.00000859, 0.00000024, 0.00000013, -0.00000383, -0.00000742, 0.00000827, 0.00000175, -0.00001279, 0.00000232, 0.00000706, -0.00000062, 0.00000254, -0.00000594, -0.00001203, 0.00000077, 0.00000567, -0.00000013, -0.00000444, 0.00000216, 0.00001053, -0.00000683, -0.00001671, 0.00000219, 0.00000103, -0.00000395, 0.00000758, -0.00000318, -0.00000270, 0.00001329, -0.00001116, -0.00002185, -0.00000237, -0.00001163, -0.00000547, 0.00000387, -0.00002684, -0.00000693, 0.00000493, -0.00008798, -0.00007739, 0.00008051, 0.00011815, 0.00006599, 0.00010252, 0.00010347, -0.00000926, -0.00007426, -0.00004192, -0.00001197, -0.00001642, -0.00001961, },
		{ 0.00000657, -0.00000172, -0.00000358, 0.00000434, 0.00000073, -0.00000498, 0.00000458, 0.00000204, -0.00001331, -0.00000731, 0.00000051, -0.00000348, 0.00000470, 0.00000303, -0.00001185, -0.00000153, 0.00000812, -0.00000238, 0.00000001, 0.00000420, -0.00000038, 0.00000174, -0.00000475, -0.00000790, 0.00000731, 0.00000404, -0.00000735, 0.00000217, 0.00000313, -0.00000242, 0.00000378, -0.00000150, -0.00000712, 0.00000487, 0.00000602, -0.00000386, 0.00000012, 0.00001128, 0.00001037, -0.00000711, -0.00001814, -0.00000579, 0.00000760, 0.00001213, 0.00000978, -0.00001173, -0.00001729, 0.00001785, 0.00002043, -0.00002209, -0.00001570, 0.00002058, -0.00000052, -0.00002574, 0.00000241, 0.00001497, -0.00000616, -0.00000499, 0.00000682, -0.00000020, -0.00001205, -0.00000911, 0.00001449, 0.00001681, -0.00001755, -0.00001541, 0.00001034, -0.00000704, -0.00000344, 0.00002704, -0.00000187, -0.00002354, 0.00000781, 0.00000339, -0.00001236, 0.00000616, 0.00000504, -0.00000240, 0.00000160, -0.00000892, -0.00000049, 0.00001180, -0.00000953, -0.00000869, 0.00000991, 0.00000091, 0.00000008, 0.00000249, -0.00001310, -0.00000725, 0.00000757, 0.00000176, -0.00000366, -0.00000092, 0.00000717, 0.00000450, -0.00001475, -0.00000958, 0.00000531, -0.00000305, 0.00000117, 0.00000462, -0.00000706, 0.00000805, 0.00000636, -0.00002393, -0.00001227, -0.00000286, -0.00001441, 0.00000435, -0.00000907, -0.00002856, 0.00001294, -0.00003089, -0.00011750, -0.00001214, 0.00012697, 0.00009668, 0.00008248, 0.00012671, 0.00005021, -0.00006606, -0.00006520, -0.00001864, -0.00001331, -0.00002116, -0.00001329, 0.00000176, },
		{ -0.00000350, -0.00000011, 0.00000387, -0.00000310, -0.00000169, 0.00000547, -0.00000563, -0.00001276, -0.00000199, -0.00000132, -0.00000091, 0.00000673, -0.00000458, -0.00001010, 0.00000541, 0.00000370, -0.00000334, 0.00000366, 0.00000185, -0.00000009, 0.00000052, -0.00000849, -0.00000140, 0.00000943, -0.00000333, -0.00000464, 0.00000545, -0.00000095, -0.00000005, 0.00000401, -0.00000688, -0.00000250, 0.00000978, 0.00000018, -0.00000566, 0.00000718, 0.00001358, 0.00000227, -0.00001366, -0.00001433, -0.00000070, 0.00001131, 0.00001532, -0.00000068, -0.00002086, 0.00000065, 0.00002681, -0.00000384, -0.00002625, 0.00000792, 0.00001466, -0.00002015, -0.00001415, 0.00001351, 0.00000508, -0.00000744, -0.00000056, 0.00000344, -0.00000348, -0.00001453, -0.00000109, 0.00002539, 0.00000096, -0.00002807, 0.00000264, 0.00000780, -0.00001554, 0.00001618, 0.00002184, -0.00002358, -0.00000881, 0.00001468, -0.00001087, -0.00000460, 0.00001104, -0.00000281, 0.00000101, -0.00000172, -0.00001050, 0.00001086, 0.00000449, -0.00001706, 0.00000353, 0.00000855, -0.00000470, 0.00000531, -0.00000417, -0.00001654, 0.00000375, 0.00000740, -0.00000466, -0.00000149, 0.00000370, 0.00000673, -0.00000375, -0.00001613, -0.00000177, 0.00000383, -0.00000349, 0.00000426, -0.00000161, -0.00000212, 0.00001376, -0.00000975, -0.00002519, -0.00000242, -0.00000979, -0.00000843, 0.00000637, -0.00002597, -0.00001186, 0.00001370, -0.00008556, -0.00010263, 0.00006809, 0.00013527, 0.00008483, 0.00011845, 0.00010873, -0.00002606, -0.00008600, -0.00003560, -0.00001028, -0.00002092, -0.00001836, -0.00000486, 0.00000493, 0.00000330, },
		{ 0.00000211, 0.00000127, -0.00000409, 0.00000179, 0.00000184, -0.00001073, -0.00000835, -0.00000002, -0.00000234, 0.00000318, 0.00000362, -0.00000980, -0.00000363, 0.00000712, -0.00000104, -0.00000039, 0.00000436, -0.00000040, 0.00000107, -0.00000324, -0.00000794, 0.00000584, 0.00000512, -0.00000737, 0.00000129, 0.00000390, -0.00000302, 0.00000391, -0.00000045, -0.00000863, 0.00000593, 0.00000821, -0.00000690, -0.00000071, 0.00001380, 0.00000912, -0.00000590, -0.00001501, -0.00001042, 0.00000443, 0.00001746, 0.00001074, -0.00001535, -0.00001426, 0.00001995, 0.00001429, -0.00002300, -0.00000935, 0.00001883, -0.00000563, -0.00002311, 0.00000235, 0.00001184, -0.00000166, -0.00000508, -0.00000119, 0.00000237, -0.00000740, -0.00001546, 0.00001520, 0.00002383, -0.00002269, -0.00001880, 0.00001719, -0.00000817, -0.00000698, 0.00003199, -0.00000267, -0.00002733, 0.00001219, 0.00000346, -0.00001608, 0.00000906, 0.00000442, -0.00000516, 0.00000514, -0.00000983, -0.00000236, 0.00001672, -0.00001090, -0.00001181, 0.00001357, -0.00000142, -0.00000205, 0.00000745, -0.00001568, -0.00000935, 0.00001272, -0.00000128, -0.00000571, 0.00000431, 0.00000419, 0.00000293, -0.00000985, -0.00001297, 0.00000360, 0.00000116, -0.00000179, 0.00000298, -0.00000418, 0.00000732, 0.00000751, -0.00002404, -0.00001553, -0.00000069, -0.00001474, 0.00000225, -0.00000501, -0.00003148, 0.00001200, -0.00001956, -0.00012762, -0.00003944, 0.00012827, 0.00011970, 0.00010211, 0.00013900, 0.00004168, -0.00008512, -0.00006519, -0.00001169, -0.00001737, -0.00002299, -0.00001010, 0.00000020, 0.00000469, 0.00000062, -0.00000298, },
		{ -0.00000135, -0.00000290, 0.00000288, -0.00000360, -0.00001148, -0.00000369, -0.00000043, -0.00000093, 0.00000522, -0.00000234, -0.00000968, 0.00000262, 0.00000432, -0.00000239, 0.00000301, 0.00000229, -0.00000066, 0.00000067, -0.00000695, -0.00000262, 0.00000877, -0.00000208, -0.00000536, 0.00000525, -0.00000048, -0.00000074, 0.00000487, -0.00000685, -0.00000345, 0.00001195, -0.00000002, -0.00000849, 0.00000868, 0.00001425, 0.00000120, -0.00001019, -0.00001455, -0.00000645, 0.00001292, 0.00001971, -0.00000338, -0.00002117, 0.00000484, 0.00002356, -0.00000889, -0.00002153, 0.00001112, 0.00000901, -0.00002107, -0.00001104, 0.00001027, 0.00000534, -0.00000273, -0.00000595, -0.00000128, 0.00000282, -0.00001506, -0.00000659, 0.00003083, 0.00000303, -0.00003493, 0.00000525, 0.00001255, -0.00001996, 0.00001698, 0.00002551, -0.00002769, -0.00000828, 0.00001867, -0.00001455, -0.00000487, 0.00001390, -0.00000655, 0.00000165, 0.00000149, -0.00001403, 0.00001276, 0.00000817, -0.00002161, 0.00000437, 0.00001089, -0.00000960, 0.00000774, -0.00000118, -0.00002170, 0.00000661, 0.00001000, -0.00001039, 0.00000182, 0.00000638, 0.00000097, -0.00000023, -0.00001363, -0.00000762, 0.00000666, -0.00000146, -0.00000081, 0.00000042, -0.00000026, 0.00001212, -0.00000769, -0.00002722, -0.00000407, -0.00000683, -0.00001110, 0.00000684, -0.00002278, -0.00001707, 0.00001904, -0.00007817, -0.00012529, 0.00004707, 0.00015048, 0.00011144, 0.00013628, 0.00010989, -0.00004497, -0.00009653, -0.00002750, -0.00000942, -0.00002632, -0.00001569, -0.00000356, 0.00000147, 0.00000409, -0.00000109, -0.00000179, 0.00000269, },
		{ -0.00000111, 0.00000132, -0.00000797, -0.00000899, -0.00000082, -0.00000104, 0.00000168, 0.00000363, -0.00000717, -0.00000546, 0.00000528, 0.00000072, -0.00000044, 0.00000398, -0.00000006, 0.00000030, -0.00000210, -0.00000724, 0.00000401, 0.00000558, -0.00000655, 0.00000028, 0.00000426, -0.00000301, 0.00000360, 0.00000049, -0.00000927, 0.00000616, 0.00000986, -0.00000877, -0.00000167, 0.00001553, 0.00000813, -0.00000474, -0.00001191, -0.00001431, 0.00000136, 0.00002188, 0.00001136, -0.00001795, -0.00001093, 0.00002066, 0.00000804, -0.00002263, -0.00000329, 0.00001640, -0.00001033, -0.00002019, 0.00000245, 0.00000845, 0.00000231, -0.00000416, -0.00000885, 0.00000335, -0.00000192, -0.00002011, 0.00001394, 0.00002981, -0.00002592, -0.00002253, 0.00002300, -0.00000800, -0.00001055, 0.00003566, -0.00000291, -0.00003061, 0.00001619, 0.00000356, -0.00001978, 0.00001220, 0.00000381, -0.00000867, 0.00000921, -0.00001004, -0.00000524, 0.00002175, -0.00001154, -0.00001568, 0.00001757, -0.00000340, -0.00000530, 0.00001306, -0.00001763, -0.00001241, 0.00001855, -0.00000431, -0.00000878, 0.00001092, 0.00000082, -0.00000027, -0.00000289, -0.00001622, -0.00000003, 0.00000704, -0.00000482, -0.00000018, 0.00000061, 0.00000595, 0.00000699, -0.00002195, -0.00001926, 0.00000044, -0.00001305, -0.00000150, -0.00000146, -0.00003145, 0.00000777, -0.00000884, -0.00013063, -0.00007007, 0.00012071, 0.00014729, 0.00012895, 0.00014847, 0.00002902, -0.00010375, -0.00006325, -0.00000450, -0.00002271, -0.00002431, -0.00000590, -0.00000234, 0.00000258, 0.00000363, -0.00000279, -0.00000025, 0.00000216, -0.00000322, },
		{ -0.00000185, -0.00000989, -0.00000519, 0.00000022, -0.00000071, 0.00000341, -0.00000032, -0.00000855, -0.00000049, 0.00000450, -0.00000078, 0.00000216, 0.00000249, -0.00000082, 0.00000035, -0.00000527, -0.00000332, 0.00000729, -0.00000072, -0.00000541, 0.00000445, 0.00000019, -0.00000112, 0.00000505, -0.00000621, -0.00000413, 0.00001286, 0.00000030, -0.00001037, 0.00000927, 0.00001466, 0.00000054, -0.00000711, -0.00001441, -0.00001134, 0.00001396, 0.00002309, -0.00000551, -0.00002047, 0.00000816, 0.00001928, -0.00001287, -0.00001627, 0.00001338, 0.00000343, -0.00002149, -0.00000793, 0.00000722, 0.00000504, 0.00000191, -0.00001003, -0.00000674, 0.00000798, -0.00001355, -0.00001184, 0.00003368, 0.00000585, -0.00004021, 0.00000636, 0.00001735, -0.00002297, 0.00001659, 0.00002854, -0.00003067, -0.00000788, 0.00002227, -0.00001796, -0.00000515, 0.00001695, -0.00001063, 0.00000188, 0.00000566, -0.00001766, 0.00001370, 0.00001260, -0.00002608, 0.00000460, 0.00001400, -0.00001493, 0.00000948, 0.00000299, -0.00002714, 0.00000894, 0.00001348, -0.00001697, 0.00000513, 0.00001027, -0.00000649, 0.00000333, -0.00000904, -0.00001513, 0.00000914, 0.00000233, -0.00000754, 0.00000267, 0.00000319, 0.00000815, -0.00000519, -0.00002730, -0.00000777, -0.00000296, -0.00001266, 0.00000454, -0.00001741, -0.00002131, 0.00001963, -0.00006642, -0.00014293, 0.00001679, 0.00016171, 0.00014747, 0.00015654, 0.00010518, -0.00006533, -0.00010488, -0.00001853, -0.00000945, -0.00003213, -0.00001192, -0.00000236, -0.00000286, 0.00000540, 0.00000138, -0.00000390, 0.00000185, 0.00000042, -0.00000420, 0.00000027, },
		{ -0.00000914, -0.00000165, 0.00000011, 0.00000038, 0.00000322, -0.00000435, -0.00000682, 0.00000282, 0.00000257, -0.00000004, 0.00000314, 0.00000046, -0.00000035, -0.00000142, -0.00000601, 0.00000212, 0.00000535, -0.00000509, -0.00000065, 0.00000415, -0.00000244, 0.00000299, 0.00000116, -0.00000902, 0.00000567, 0.00001073, -0.00000935, -0.00000262, 0.00001623, 0.00000758, -0.00000357, -0.00000922, -0.00001723, -0.00000131, 0.00002501, 0.00001163, -0.00001930, -0.00000755, 0.00001994, 0.00000201, -0.00002105, 0.00000217, 0.00001339, -0.00001435, -0.00001721, 0.00000262, 0.00000513, 0.00000550, -0.00000250, -0.00001560, 0.00000268, 0.00000379, -0.00002258, 0.00001102, 0.00003409, -0.00002698, -0.00002617, 0.00002718, -0.00000644, -0.00001372, 0.00003761, -0.00000257, -0.00003302, 0.00001941, 0.00000375, -0.00002318, 0.00001522, 0.00000334, -0.00001272, 0.00001341, -0.00000937, -0.00000895, 0.00002643, -0.00001126, -0.00002008, 0.00002151, -0.00000474, -0.00000956, 0.00001888, -0.00001860, -0.00001635, 0.00002462, -0.00000695, -0.00001285, 0.00001849, -0.00000254, -0.00000516, 0.00000575, -0.00001891, -0.00000565, 0.00001432, -0.00000758, -0.00000502, 0.00000727, 0.00000415, 0.00000439, -0.00001739, -0.00002315, -0.00000014, -0.00000886, -0.00000662, 0.00000058, -0.00002762, 0.00000062, -0.00000057, -0.00012535, -0.00010226, 0.00010210, 0.00017900, 0.00016487, 0.00015403, 0.00001137, -0.00012046, -0.00005953, 0.00000210, -0.00002872, -0.00002507, -0.00000108, -0.00000562, 0.00000034, 0.00000709, -0.00000302, -0.00000252, 0.00000332, -0.00000264, -0.00000383, 0.00000069, -0.00000548, },
		{ 0.00000056, -0.00000045, 0.00000166, 0.00000130, -0.00000696, -0.00000353, 0.00000414, 0.00000127, 0.00000133, 0.00000235, -0.00000058, -0.00000023, -0.00000372, -0.00000348, 0.00000527, 0.00000049, -0.00000482, 0.00000325, 0.00000090, -0.00000112, 0.00000460, -0.00000513, -0.00000440, 0.00001251, 0.00000096, -0.00001110, 0.00000891, 0.00001472, 0.00000052, -0.00000462, -0.00001408, -0.00001496, 0.00001436, 0.00002516, -0.00000688, -0.00001878, 0.00001039, 0.00001423, -0.00001563, -0.00001076, 0.00001459, -0.00000181, -0.00002142, -0.00000507, 0.00000453, 0.00000434, 0.00000604, -0.00001269, -0.00001243, 0.00001153, -0.00001025, -0.00001620, 0.00003372, 0.00000900, -0.00004336, 0.00000590, 0.00002174, -0.00002414, 0.00001506, 0.00003055, -0.00003218, -0.00000765, 0.00002515, -0.00002074, -0.00000548, 0.00001992, -0.00001466, 0.00000157, 0.00001049, -0.00002096, 0.00001354, 0.00001749, -0.00003001, 0.00000405, 0.00001767, -0.00002026, 0.00001027, 0.00000816, -0.00003237, 0.00001040, 0.00001772, -0.00002388, 0.00000803, 0.00001528, -0.00001516, 0.00000647, -0.00000240, -0.00002386, 0.00001091, 0.00000789, -0.00001565, 0.00000484, 0.00000840, 0.00000182, -0.00000254, -0.00002494, -0.00001374, 0.00000144, -0.00001234, -0.00000103, -0.00001030, -0.00002330, 0.00001453, -0.00005137, -0.00015322, -0.00002278, 0.00016668, 0.00019410, 0.00017998, 0.00009291, -0.00008648, -0.00010995, -0.00000973, -0.00001039, -0.00003767, -0.00000750, -0.00000133, -0.00000771, 0.00000705, 0.00000375, -0.00000627, 0.00000146, 0.00000196, -0.00000562, -0.00000194, -0.00000055, -0.00000828, -0.00000617, },
		{ -0.00000064, 0.00000260, -0.00000175, -0.00000760, 0.00000006, 0.00000418, 0.00000084, 0.00000200, 0.00000099, -0.00000074, -0.00000120, -0.00000452, 0.00000044, 0.00000450, -0.00000327, -0.00000134, 0.00000361, -0.00000144, 0.00000224, 0.00000145, -0.00000795, 0.00000467, 0.00001071, -0.00000869, -0.00000338, 0.00001580, 0.00000752, -0.00000229, -0.00000723, -0.00001903, -0.00000330, 0.00002662, 0.00001153, -0.00001925, -0.00000431, 0.00001787, -0.00000349, -0.00001841, 0.00000680, 0.00000995, -0.00001754, -0.00001436, 0.00000278, 0.00000216, 0.00000777, -0.00000046, -0.00002102, 0.00000052, 0.00000911, -0.00002263, 0.00000688, 0.00003616, -0.00002585, -0.00002929, 0.00002931, -0.00000357, -0.00001607, 0.00003754, -0.00000169, -0.00003423, 0.00002155, 0.00000403, -0.00002596, 0.00001777, 0.00000307, -0.00001696, 0.00001732, -0.00000776, -0.00001313, 0.00003032, -0.00000996, -0.00002469, 0.00002499, -0.00000525, -0.00001456, 0.00002441, -0.00001835, -0.00002097, 0.00003039, -0.00000884, -0.00001771, 0.00002649, -0.00000548, -0.00001160, 0.00001551, -0.00002064, -0.00001313, 0.00002252, -0.00000973, -0.00001152, 0.00001553, 0.00000221, -0.00000054, -0.00001035, -0.00002682, -0.00000301, -0.00000196, -0.00001261, 0.00000016, -0.00001946, -0.00000869, 0.00000351, -0.00011127, -0.00013391, 0.00007077, 0.00021383, 0.00021164, 0.00015491, -0.00001221, -0.00013372, -0.00005420, 0.00000715, -0.00003465, -0.00002518, 0.00000377, -0.00000926, -0.00000189, 0.00001068, -0.00000362, -0.00000450, 0.00000465, -0.00000258, -0.00000613, 0.00000019, -0.00000362, -0.00000881, -0.00000225, 0.00000084, },
		{ 0.00000242, -0.00000519, -0.00000612, 0.00000324, 0.00000344, 0.00000076, 0.00000185, -0.00000006, -0.00000086, -0.00000253, -0.00000318, 0.00000306, 0.00000134, -0.00000369, 0.00000189, 0.00000150, -0.00000074, 0.00000366, -0.00000383, -0.00000419, 0.00001104, 0.00000177, -0.00001061, 0.00000771, 0.00001430, 0.00000124, -0.00000280, -0.00001373, -0.00001707, 0.00001419, 0.00002574, -0.00000738, -0.00001618, 0.00001142, 0.00000870, -0.00001711, -0.00000527, 0.00001476, -0.00000651, -0.00002085, -0.00000265, 0.00000231, 0.00000340, 0.00000934, -0.00001397, -0.00001786, 0.00001321, -0.00000558, -0.00001913, 0.00003105, 0.00001201, -0.00004404, 0.00000398, 0.00002526, -0.00002324, 0.00001254, 0.00003120, -0.00003199, -0.00000752, 0.00002699, -0.00002260, -0.00000586, 0.00002247, -0.00001825, 0.00000070, 0.00001559, -0.00002352, 0.00001223, 0.00002245, -0.00003300, 0.00000266, 0.00002160, -0.00002511, 0.00000993, 0.00001402, -0.00003690, 0.00001074, 0.00002245, -0.00003055, 0.00001020, 0.00002118, -0.00002441, 0.00000881, 0.00000602, -0.00003318, 0.00001162, 0.00001505, -0.00002469, 0.00000664, 0.00001536, -0.00000667, -0.00000006, -0.00001979, -0.00002196, 0.00000587, -0.00000946, -0.00001007, -0.00000215, -0.00002188, 0.00000330, -0.00003445, -0.00015422, -0.00007092, 0.00016318, 0.00025191, 0.00020757, 0.00007160, -0.00010793, -0.00011063, -0.00000220, -0.00001231, -0.00004216, -0.00000300, -0.00000060, -0.00001256, 0.00000876, 0.00000586, -0.00000864, 0.00000154, 0.00000306, -0.00000719, -0.00000343, 0.00000023, -0.00000699, -0.00000640, 0.00000063, -0.00000035, 0.00000026, },
		{ -0.00000779, -0.00000262, 0.00000527, 0.00000211, 0.00000080, 0.00000131, -0.00000075, -0.00000134, -0.00000309, -0.00000082, 0.00000320, -0.00000142, -0.00000164, 0.00000276, -0.00000022, 0.00000153, 0.00000136, -0.00000628, 0.00000339, 0.00000979, -0.00000700, -0.00000380, 0.00001424, 0.00000788, -0.00000085, -0.00000610, -0.00001974, -0.00000440, 0.00002663, 0.00001099, -0.00001779, -0.00000139, 0.00001459, -0.00000824, -0.00001491, 0.00001042, 0.00000623, -0.00001980, -0.00001175, 0.00000284, -0.00000027, 0.00000911, 0.00000163, -0.00002482, -0.00000280, 0.00001353, -0.00002033, 0.00000210, 0.00003576, -0.00002277, -0.00003145, 0.00002919, 0.00000033, -0.00001722, 0.00003537, -0.00000045, -0.00003399, 0.00002239, 0.00000433, -0.00002780, 0.00001956, 0.00000300, -0.00002101, 0.00002056, -0.00000523, -0.00001736, 0.00003303, -0.00000768, -0.00002911, 0.00002760, -0.00000486, -0.00001992, 0.00002916, -0.00001674, -0.00002591, 0.00003533, -0.00000974, -0.00002305, 0.00003426, -0.00000767, -0.00001923, 0.00002565, -0.00002107, -0.00002209, 0.00003107, -0.00001098, -0.00001946, 0.00002498, 0.00000042, -0.00000786, -0.00000102, -0.00002981, -0.00000857, 0.00000759, -0.00001881, -0.00000349, -0.00000689, -0.00001915, 0.00000191, -0.00008854, -0.00016272, 0.00002574, 0.00025027, 0.00027061, 0.00015079, -0.00004261, -0.00014215, -0.00004734, 0.00000960, -0.00003972, -0.00002452, 0.00000800, -0.00001282, -0.00000396, 0.00001397, -0.00000440, -0.00000602, 0.00000598, -0.00000315, -0.00000785, 0.00000000, -0.00000265, -0.00000810, -0.00000238, 0.00000098, -0.00000134, 0.00000202, 0.00000031, },
		{ 0.00000192, 0.00000540, 0.00000064, 0.00000102, 0.00000057, -0.00000130, -0.00000184, -0.00000259, 0.00000101, 0.00000172, -0.00000225, 0.00000062, 0.00000190, -0.00000006, 0.00000245, -0.00000251, -0.00000353, 0.00000873, 0.00000250, -0.00000898, 0.00000590, 0.00001332, 0.00000266, -0.00000161, -0.00001350, -0.00001757, 0.00001356, 0.00002476, -0.00000704, -0.00001278, 0.00001126, 0.00000299, -0.00001735, -0.00000008, 0.00001394, -0.00001051, -0.00001984, -0.00000080, 0.00000060, 0.00000241, 0.00001159, -0.00001406, -0.00002255, 0.00001300, -0.00000012, -0.00002030, 0.00002602, 0.00001443, -0.00004218, 0.00000092, 0.00002752, -0.00002023, 0.00000934, 0.00003027, -0.00003007, -0.00000737, 0.00002756, -0.00002333, -0.00000623, 0.00002430, -0.00002107, -0.00000067, 0.00002051, -0.00002499, 0.00000988, 0.00002702, -0.00003470, 0.00000050, 0.00002537, -0.00002905, 0.00000842, 0.00002011, -0.00004024, 0.00000985, 0.00002729, -0.00003639, 0.00001139, 0.00002756, -0.00003354, 0.00001003, 0.00001574, -0.00004237, 0.00001105, 0.00002340, -0.00003406, 0.00000780, 0.00002385, -0.00001689, 0.00000188, -0.00001177, -0.00003218, 0.00000977, -0.00000358, -0.00002240, 0.00000616, -0.00001613, -0.00001394, -0.00001736, -0.00014461, -0.00012619, 0.00014924, 0.00032068, 0.00024038, 0.00004020, -0.00012939, -0.00010583, 0.00000299, -0.00001532, -0.00004473, 0.00000094, -0.00000030, -0.00001683, 0.00001021, 0.00000752, -0.00001069, 0.00000200, 0.00000356, -0.00000886, -0.00000401, 0.00000052, -0.00000626, -0.00000578, 0.00000049, -0.00000062, -0.00000056, 0.00000306, -0.00000359, -0.00000794, },
		{ 0.00000363, -0.00000020, 0.00000128, -0.00000039, -0.00000166, -0.00000201, -0.00000157, 0.00000170, 0.00000014, -0.00000152, 0.00000178, 0.00000100, 0.00000098, 0.00000094, -0.00000427, 0.00000205, 0.00000807, -0.00000458, -0.00000373, 0.00001172, 0.00000848, 0.00000081, -0.00000582, -0.00001947, -0.00000451, 0.00002512, 0.00000997, -0.00001504, 0.00000114, 0.00001034, -0.00001206, -0.00001083, 0.00001297, 0.00000242, -0.00002112, -0.00000947, 0.00000273, -0.00000205, 0.00000958, 0.00000347, -0.00002694, -0.00000684, 0.00001663, -0.00001601, -0.00000273, 0.00003289, -0.00001818, -0.00003230, 0.00002686, 0.00000489, -0.00001694, 0.00003120, 0.00000092, -0.00003214, 0.00002189, 0.00000456, -0.00002847, 0.00002038, 0.00000306, -0.00002446, 0.00002281, -0.00000197, -0.00002117, 0.00003426, -0.00000459, -0.00003290, 0.00002901, -0.00000363, -0.00002519, 0.00003269, -0.00001381, -0.00003072, 0.00003895, -0.00000952, -0.00002842, 0.00004116, -0.00000886, -0.00002755, 0.00003540, -0.00002004, -0.00003198, 0.00003926, -0.00001114, -0.00002843, 0.00003501, -0.00000095, -0.00001737, 0.00001012, -0.00003170, -0.00001693, 0.00001942, -0.00002451, -0.00001087, 0.00000974, -0.00002957, -0.00000637, -0.00005804, -0.00018646, -0.00003304, 0.00028642, 0.00034255, 0.00014189, -0.00008063, -0.00014468, -0.00003893, 0.00000845, -0.00004315, -0.00002293, 0.00001083, -0.00001580, -0.00000569, 0.00001647, -0.00000515, -0.00000694, 0.00000709, -0.00000435, -0.00000888, 0.00000026, -0.00000267, -0.00000716, -0.00000187, 0.00000045, -0.00000180, 0.00000172, 0.00000159, -0.00000748, -0.00000487, 0.00000564, },
		{ 0.00000001, 0.00000112, -0.00000137, -0.00000168, -0.00000192, -0.00000058, 0.00000160, -0.00000076, -0.00000029, 0.00000205, 0.00000078, 0.00000121, -0.00000136, -0.00000253, 0.00000593, 0.00000299, -0.00000643, 0.00000377, 0.00001177, 0.00000463, -0.00000093, -0.00001347, -0.00001658, 0.00001264, 0.00002231, -0.00000600, -0.00000876, 0.00001002, -0.00000263, -0.00001648, 0.00000459, 0.00001227, -0.00001375, -0.00001845, 0.00000046, -0.00000064, 0.00000151, 0.00001273, -0.00001323, -0.00002616, 0.00001110, 0.00000551, -0.00001960, 0.00001925, 0.00001586, -0.00003795, -0.00000285, 0.00002827, -0.00001533, 0.00000581, 0.00002764, -0.00002655, -0.00000706, 0.00002671, -0.00002288, -0.00000650, 0.00002513, -0.00002288, -0.00000237, 0.00002478, -0.00002515, 0.00000674, 0.00003073, -0.00003489, -0.00000227, 0.00002856, -0.00003171, 0.00000586, 0.00002594, -0.00004199, 0.00000776, 0.00003178, -0.00004086, 0.00001148, 0.00003392, -0.00004180, 0.00000998, 0.00002610, -0.00005067, 0.00000913, 0.00003239, -0.00004310, 0.00000814, 0.00003348, -0.00002823, 0.00000298, -0.00000104, -0.00004386, 0.00001260, 0.00000546, -0.00003750, 0.00001373, -0.00000556, -0.00003649, -0.00000181, -0.00012380, -0.00018652, 0.00012341, 0.00039933, 0.00027945, -0.00000172, -0.00015083, -0.00009464, 0.00000501, -0.00001961, -0.00004451, 0.00000373, -0.00000065, -0.00001985, 0.00001103, 0.00000857, -0.00001206, 0.00000268, 0.00000336, -0.00001051, -0.00000364, 0.00000039, -0.00000624, -0.00000435, 0.00000030, -0.00000137, -0.00000072, 0.00000328, -0.00000244, -0.00000882, 0.00000035, 0.00000681, 0.00000107, },
		{ 0.00000027, -0.00000197, -0.00000149, -0.00000179, 0.00000028, 0.00000116, -0.00000101, 0.00000090, 0.00000200, 0.00000067, 0.00000034, -0.00000226, 0.00000084, 0.00000579, -0.00000185, -0.00000310, 0.00000854, 0.00000908, 0.00000267, -0.00000623, -0.00001844, -0.00000362, 0.00002236, 0.00000845, -0.00001124, 0.00000325, 0.00000541, -0.00001491, -0.00000643, 0.00001443, -0.00000131, -0.00002154, -0.00000754, 0.00000244, -0.00000315, 0.00000931, 0.00000481, -0.00002743, -0.00001110, 0.00001820, -0.00001022, -0.00000703, 0.00002784, -0.00001267, -0.00003166, 0.00002261, 0.00000964, -0.00001510, 0.00002534, 0.00000218, -0.00002868, 0.00002014, 0.00000458, -0.00002783, 0.00002013, 0.00000315, -0.00002695, 0.00002386, 0.00000175, -0.00002411, 0.00003386, -0.00000098, -0.00003564, 0.00002903, -0.00000172, -0.00002987, 0.00003466, -0.00000975, -0.00003493, 0.00004086, -0.00000821, -0.00003334, 0.00004660, -0.00000896, -0.00003592, 0.00004400, -0.00001751, -0.00004208, 0.00004644, -0.00001017, -0.00003789, 0.00004497, -0.00000170, -0.00002865, 0.00002244, -0.00003212, -0.00002794, 0.00003291, -0.00002899, -0.00002211, 0.00002966, -0.00003874, -0.00002175, -0.00002129, -0.00020318, -0.00010464, 0.00032009, 0.00042745, 0.00012897, -0.00012678, -0.00014067, -0.00002884, 0.00000280, -0.00004433, -0.00002016, 0.00001153, -0.00001773, -0.00000685, 0.00001764, -0.00000560, -0.00000712, 0.00000770, -0.00000606, -0.00000920, 0.00000098, -0.00000363, -0.00000616, -0.00000067, -0.00000066, -0.00000209, 0.00000185, 0.00000198, -0.00000681, -0.00000636, 0.00000519, 0.00000517, -0.00000074, 0.00000084, },
		{ -0.00000198, -0.00000144, -0.00000149, 0.00000105, 0.00000050, -0.00000068, 0.00000199, 0.00000160, 0.00000017, -0.00000052, -0.00000138, 0.00000304, 0.00000311, -0.00000331, 0.00000170, 0.00000969, 0.00000686, -0.00000054, -0.00001363, -0.00001436, 0.00001163, 0.00001861, -0.00000448, -0.00000432, 0.00000793, -0.00000791, -0.00001470, 0.00000855, 0.00000994, -0.00001617, -0.00001676, 0.00000113, -0.00000146, 0.00000079, 0.00001277, -0.00001180, -0.00002845, 0.00000786, 0.00001071, -0.00001716, 0.00001149, 0.00001606, -0.00003181, -0.00000681, 0.00002741, -0.00000895, 0.00000237, 0.00002340, -0.00002172, -0.00000642, 0.00002444, -0.00002131, -0.00000655, 0.00002477, -0.00002355, -0.00000419, 0.00002800, -0.00002389, 0.00000313, 0.00003317, -0.00003348, -0.00000535, 0.00003077, -0.00003287, 0.00000249, 0.00003099, -0.00004191, 0.00000466, 0.00003546, -0.00004354, 0.00001050, 0.00003971, -0.00004853, 0.00000864, 0.00003634, -0.00005740, 0.00000592, 0.00004131, -0.00005116, 0.00000758, 0.00004366, -0.00003996, 0.00000303, 0.00001193, -0.00005629, 0.00001389, 0.00001748, -0.00005451, 0.00001966, 0.00000985, -0.00006314, 0.00001059, -0.00009212, -0.00024931, 0.00008494, 0.00048587, 0.00032566, -0.00005383, -0.00017248, -0.00007642, 0.00000330, -0.00002544, -0.00004072, 0.00000490, -0.00000190, -0.00002099, 0.00001088, 0.00000879, -0.00001233, 0.00000336, 0.00000238, -0.00001196, -0.00000244, -0.00000006, -0.00000695, -0.00000230, 0.00000019, -0.00000263, -0.00000025, 0.00000321, -0.00000195, -0.00000869, -0.00000123, 0.00000740, 0.00000206, -0.00000099, 0.00000139, -0.00000099, },
		{ -0.00000159, -0.00000078, 0.00000147, -0.00000021, 0.00000034, 0.00000263, 0.00000063, -0.00000029, -0.00000055, -0.00000011, 0.00000326, 0.00000079, -0.00000193, 0.00000509, 0.00000943, 0.00000464, -0.00000705, -0.00001688, -0.00000188, 0.00001873, 0.00000643, -0.00000678, 0.00000496, 0.00000015, -0.00001678, -0.00000203, 0.00001489, -0.00000482, -0.00002118, -0.00000595, 0.00000195, -0.00000364, 0.00000849, 0.00000550, -0.00002653, -0.00001508, 0.00001819, -0.00000365, -0.00001035, 0.00002109, -0.00000689, -0.00002949, 0.00001693, 0.00001410, -0.00001181, 0.00001829, 0.00000305, -0.00002375, 0.00001737, 0.00000429, -0.00002585, 0.00001885, 0.00000313, -0.00002822, 0.00002364, 0.00000559, -0.00002580, 0.00003184, 0.00000279, -0.00003698, 0.00002760, 0.00000059, -0.00003351, 0.00003490, -0.00000487, -0.00003806, 0.00004083, -0.00000598, -0.00003731, 0.00005012, -0.00000803, -0.00004366, 0.00005076, -0.00001363, -0.00005159, 0.00005198, -0.00000815, -0.00004714, 0.00005414, -0.00000173, -0.00004105, 0.00003514, -0.00003084, -0.00004111, 0.00004718, -0.00003168, -0.00003690, 0.00005171, -0.00004558, -0.00004397, 0.00001964, -0.00021139, -0.00018710, 0.00034902, 0.00052434, 0.00011337, -0.00018117, -0.00013004, -0.00001685, -0.00000802, -0.00004285, -0.00001594, 0.00000941, -0.00001824, -0.00000718, 0.00001697, -0.00000549, -0.00000644, 0.00000749, -0.00000805, -0.00000880, 0.00000204, -0.00000530, -0.00000529, 0.00000114, -0.00000218, -0.00000235, 0.00000247, 0.00000153, -0.00000600, -0.00000688, 0.00000396, 0.00000646, -0.00000065, 0.00000000, 0.00000106, -0.00000207, -0.00000167, },
		{ 0.00000023, 0.00000126, -0.00000043, 0.00000180, 0.00000223, -0.00000049, -0.00000003, -0.00000031, 0.00000043, 0.00000284, -0.00000002, 0.00000003, 0.00000721, 0.00000897, -0.00000023, -0.00001388, -0.00001132, 0.00001071, 0.00001403, -0.00000280, 0.00000028, 0.00000529, -0.00001264, -0.00001227, 0.00001169, 0.00000716, -0.00001781, -0.00001485, 0.00000130, -0.00000194, 0.00000030, 0.00001186, -0.00001010, -0.00002930, 0.00000380, 0.00001492, -0.00001330, 0.00000356, 0.00001492, -0.00002437, -0.00001045, 0.00002503, -0.00000168, -0.00000061, 0.00001778, -0.00001604, -0.00000532, 0.00002089, -0.00001879, -0.00000628, 0.00002317, -0.00002308, -0.00000589, 0.00002985, -0.00002128, -0.00000053, 0.00003405, -0.00003057, -0.00000840, 0.00003166, -0.00003246, -0.00000133, 0.00003479, -0.00003992, 0.00000089, 0.00003789, -0.00004420, 0.00000862, 0.00004434, -0.00005323, 0.00000616, 0.00004564, -0.00006198, 0.00000167, 0.00004942, -0.00005765, 0.00000617, 0.00005368, -0.00005131, 0.00000193, 0.00002646, -0.00006863, 0.00001335, 0.00003192, -0.00007236, 0.00002325, 0.00002961, -0.00009228, 0.00001857, -0.00005077, -0.00031167, 0.00003396, 0.00057747, 0.00037954, -0.00011499, -0.00019476, -0.00005093, -0.00000231, -0.00003312, -0.00003279, 0.00000413, -0.00000434, -0.00001966, 0.00000949, 0.00000798, -0.00001111, 0.00000377, 0.00000057, -0.00001294, -0.00000064, -0.00000071, -0.00000825, 0.00000006, 0.00000029, -0.00000428, 0.00000070, 0.00000300, -0.00000220, -0.00000764, -0.00000256, 0.00000703, 0.00000349, -0.00000176, 0.00000125, -0.00000015, -0.00000261, -0.00000136, -0.00000160, },
		{ 0.00000068, 0.00000029, 0.00000281, 0.00000080, -0.00000076, 0.00000062, -0.00000076, 0.00000082, 0.00000296, -0.00000032, 0.00000183, 0.00000930, 0.00000656, -0.00000792, -0.00001505, 0.00000045, 0.00001468, 0.00000397, -0.00000210, 0.00000634, -0.00000504, -0.00001778, 0.00000210, 0.00001445, -0.00000795, -0.00002017, -0.00000469, 0.00000129, -0.00000362, 0.00000728, 0.00000546, -0.00002453, -0.00001832, 0.00001672, 0.00000295, -0.00001239, 0.00001334, -0.00000148, -0.00002592, 0.00001045, 0.00001782, -0.00000729, 0.00001065, 0.00000334, -0.00001766, 0.00001392, 0.00000359, -0.00002265, 0.00001667, 0.00000285, -0.00002810, 0.00002224, 0.00000919, -0.00002601, 0.00002836, 0.00000631, -0.00003669, 0.00002481, 0.00000295, -0.00003574, 0.00003341, 0.00000039, -0.00003971, 0.00003883, -0.00000314, -0.00003989, 0.00005144, -0.00000625, -0.00005009, 0.00005518, -0.00000872, -0.00005973, 0.00005542, -0.00000529, -0.00005548, 0.00006189, -0.00000107, -0.00005378, 0.00004739, -0.00002779, -0.00005571, 0.00006128, -0.00003216, -0.00005455, 0.00007450, -0.00004927, -0.00007210, 0.00006234, -0.00021026, -0.00027752, 0.00037109, 0.00063128, 0.00009684, -0.00024341, -0.00011333, -0.00000266, -0.00002432, -0.00003861, -0.00000996, 0.00000396, -0.00001712, -0.00000639, 0.00001402, -0.00000460, -0.00000478, 0.00000612, -0.00001004, -0.00000768, 0.00000322, -0.00000736, -0.00000468, 0.00000331, -0.00000382, -0.00000265, 0.00000350, 0.00000040, -0.00000522, -0.00000637, 0.00000213, 0.00000727, 0.00000016, -0.00000112, 0.00000181, -0.00000183, -0.00000252, -0.00000114, -0.00000124, 0.00000104, },
		{ 0.00000162, 0.00000252, -0.00000064, 0.00000011, 0.00000043, -0.00000158, 0.00000222, 0.00000298, -0.00000096, 0.00000454, 0.00001057, 0.00000019, -0.00001403, -0.00000792, 0.00001000, 0.00000902, -0.00000126, 0.00000474, 0.00000248, -0.00001663, -0.00000951, 0.00001393, 0.00000415, -0.00001871, -0.00001284, 0.00000108, -0.00000219, 0.00000006, 0.00001018, -0.00000844, -0.00002876, -0.00000056, 0.00001775, -0.00000851, -0.00000374, 0.00001255, -0.00001638, -0.00001332, 0.00002138, 0.00000580, -0.00000283, 0.00001119, -0.00001002, -0.00000368, 0.00001635, -0.00001563, -0.00000561, 0.00002038, -0.00002160, -0.00000721, 0.00003017, -0.00001754, -0.00000385, 0.00003320, -0.00002639, -0.00001106, 0.00003103, -0.00003059, -0.00000519, 0.00003697, -0.00003611, -0.00000314, 0.00003872, -0.00004278, 0.00000611, 0.00004735, -0.00005555, 0.00000286, 0.00005326, -0.00006403, -0.00000324, 0.00005601, -0.00006211, 0.00000410, 0.00006280, -0.00006153, -0.00000025, 0.00004163, -0.00007999, 0.00001085, 0.00004796, -0.00008983, 0.00002402, 0.00005271, -0.00012204, 0.00002129, -0.00000181, -0.00037061, -0.00002845, 0.00067065, 0.00044112, -0.00018320, -0.00021827, -0.00001845, -0.00001155, -0.00004296, -0.00002042, 0.00000136, -0.00000828, -0.00001543, 0.00000673, 0.00000592, -0.00000806, 0.00000367, -0.00000211, -0.00001312, 0.00000143, -0.00000151, -0.00000984, 0.00000234, 0.00000064, -0.00000608, 0.00000188, 0.00000277, -0.00000312, -0.00000590, -0.00000343, 0.00000567, 0.00000514, -0.00000209, 0.00000065, 0.00000103, -0.00000337, -0.00000181, -0.00000097, -0.00000064, 0.00000136, 0.00000041, },
		{ 0.00000110, -0.00000093, 0.00000108, -0.00000112, -0.00000120, 0.00000438, 0.00000155, -0.00000081, 0.00000851, 0.00000820, -0.00000846, -0.00001313, 0.00000302, 0.00001071, 0.00000122, 0.00000229, 0.00000748, -0.00000977, -0.00001806, 0.00000571, 0.00001330, -0.00001062, -0.00001867, -0.00000370, 0.00000051, -0.00000324, 0.00000587, 0.00000471, -0.00002179, -0.00002049, 0.00001406, 0.00000884, -0.00001300, 0.00000536, 0.00000300, -0.00002127, 0.00000388, 0.00002047, -0.00000197, 0.00000308, 0.00000293, -0.00001086, 0.00001021, 0.00000246, -0.00001848, 0.00001384, 0.00000223, -0.00002660, 0.00001984, 0.00001220, -0.00002464, 0.00002377, 0.00000921, -0.00003471, 0.00002092, 0.00000500, -0.00003632, 0.00003035, 0.00000554, -0.00003963, 0.00003502, -0.00000005, -0.00004074, 0.00005048, -0.00000392, -0.00005462, 0.00005699, -0.00000322, -0.00006577, 0.00005649, -0.00000196, -0.00006222, 0.00006768, 0.00000017, -0.00006597, 0.00005837, -0.00002308, -0.00007078, 0.00007418, -0.00003028, -0.00007400, 0.00009650, -0.00004933, -0.00010457, 0.00010425, -0.00019973, -0.00037219, 0.00038453, 0.00074531, 0.00008143, -0.00031248, -0.00009168, 0.00001399, -0.00004602, -0.00003183, -0.00000193, -0.00000508, -0.00001433, -0.00000420, 0.00000846, -0.00000280, -0.00000205, 0.00000329, -0.00001171, -0.00000585, 0.00000417, -0.00000941, -0.00000436, 0.00000549, -0.00000524, -0.00000301, 0.00000473, -0.00000115, -0.00000463, -0.00000492, -0.00000003, 0.00000738, 0.00000163, -0.00000231, 0.00000232, -0.00000121, -0.00000394, -0.00000065, -0.00000100, -0.00000004, 0.00000146, 0.00000086, 0.00000255, },
		{ -0.00000002, 0.00000068, -0.00000279, 0.00000136, 0.00000532, -0.00000109, 0.00000196, 0.00001130, 0.00000085, -0.00001385, -0.00000464, 0.00000955, 0.00000410, -0.00000017, 0.00000874, -0.00000012, -0.00001976, -0.00000674, 0.00001527, 0.00000112, -0.00001896, -0.00001080, 0.00000060, -0.00000229, 0.00000002, 0.00000796, -0.00000704, -0.00002695, -0.00000468, 0.00001895, -0.00000337, -0.00000973, 0.00000917, -0.00000862, -0.00001508, 0.00001687, 0.00001280, -0.00000409, 0.00000415, -0.00000420, -0.00000152, 0.00001120, -0.00001215, -0.00000451, 0.00001659, -0.00001933, -0.00000796, 0.00002894, -0.00001299, -0.00000645, 0.00003067, -0.00002132, -0.00001299, 0.00002884, -0.00002749, -0.00000867, 0.00003734, -0.00003080, -0.00000698, 0.00003778, -0.00003945, 0.00000334, 0.00004838, -0.00005541, -0.00000088, 0.00005855, -0.00006341, -0.00000832, 0.00006045, -0.00006426, 0.00000162, 0.00007029, -0.00006996, -0.00000332, 0.00005645, -0.00008953, 0.00000653, 0.00006451, -0.00010571, 0.00002183, 0.00007774, -0.00015046, 0.00001846, 0.00005202, -0.00042332, -0.00010023, 0.00076145, 0.00050981, -0.00025566, -0.00024360, 0.00002021, -0.00002373, -0.00005519, -0.00000366, -0.00000324, -0.00001399, -0.00000808, 0.00000262, 0.00000243, -0.00000293, 0.00000285, -0.00000569, -0.00001215, 0.00000341, -0.00000244, -0.00001131, 0.00000415, 0.00000122, -0.00000769, 0.00000298, 0.00000256, -0.00000451, -0.00000377, -0.00000369, 0.00000349, 0.00000675, -0.00000182, -0.00000037, 0.00000235, -0.00000406, -0.00000284, 0.00000026, -0.00000135, 0.00000073, 0.00000173, 0.00000161, 0.00000238, -0.00000030, },
		{ -0.00000126, -0.00000253, 0.00000488, 0.00000344, -0.00000246, 0.00000702, 0.00000931, -0.00000831, -0.00001127, 0.00000543, 0.00000723, -0.00000164, 0.00000591, 0.00000844, -0.00001365, -0.00001779, 0.00000858, 0.00001162, -0.00001275, -0.00001684, -0.00000292, -0.00000034, -0.00000263, 0.00000439, 0.00000333, -0.00001867, -0.00002137, 0.00001055, 0.00001341, -0.00001223, -0.00000207, 0.00000614, -0.00001596, -0.00000211, 0.00002181, 0.00000366, -0.00000377, 0.00000181, -0.00000386, 0.00000664, 0.00000092, -0.00001369, 0.00001065, 0.00000119, -0.00002386, 0.00001675, 0.00001433, -0.00002178, 0.00001847, 0.00001116, -0.00003115, 0.00001628, 0.00000644, -0.00003518, 0.00002606, 0.00001012, -0.00003770, 0.00002974, 0.00000288, -0.00003972, 0.00004738, -0.00000141, -0.00005685, 0.00005614, 0.00000234, -0.00006918, 0.00005513, 0.00000145, -0.00006679, 0.00007114, 0.00000173, -0.00007672, 0.00006737, -0.00001701, -0.00008522, 0.00008495, -0.00002615, -0.00009392, 0.00011622, -0.00004571, -0.00013933, 0.00014287, -0.00018053, -0.00046688, 0.00038809, 0.00086258, 0.00006931, -0.00038670, -0.00006678, 0.00003329, -0.00007255, -0.00002303, 0.00000835, -0.00001771, -0.00001006, -0.00000036, 0.00000016, -0.00000008, 0.00000185, -0.00000127, -0.00001278, -0.00000327, 0.00000449, -0.00001106, -0.00000428, 0.00000721, -0.00000610, -0.00000340, 0.00000584, -0.00000278, -0.00000429, -0.00000276, -0.00000221, 0.00000669, 0.00000360, -0.00000332, 0.00000245, -0.00000023, -0.00000569, -0.00000034, -0.00000004, -0.00000155, 0.00000194, 0.00000205, 0.00000198, 0.00000141, -0.00000071, 0.00000080, },
		{ 0.00000041, 0.00000667, -0.00000032, -0.00000025, 0.00001088, 0.00000178, -0.00001307, -0.00000190, 0.00000932, -0.00000023, 0.00000026, 0.00001196, -0.00000215, -0.00002192, -0.00000426, 0.00001574, -0.00000173, -0.00001867, -0.00000883, -0.00000004, -0.00000232, 0.00000013, 0.00000542, -0.00000608, -0.00002411, -0.00000812, 0.00001845, 0.00000154, -0.00001389, 0.00000517, -0.00000178, -0.00001556, 0.00001196, 0.00001867, -0.00000433, -0.00000274, 0.00000091, 0.00000108, 0.00000593, -0.00000870, -0.00000304, 0.00001213, -0.00001656, -0.00000802, 0.00002633, -0.00000806, -0.00000807, 0.00002663, -0.00001581, -0.00001396, 0.00002519, -0.00002352, -0.00001138, 0.00003585, -0.00002440, -0.00001017, 0.00003505, -0.00003455, 0.00000067, 0.00004726, -0.00005292, -0.00000458, 0.00006110, -0.00006023, -0.00001307, 0.00006231, -0.00006403, -0.00000093, 0.00007551, -0.00007610, -0.00000696, 0.00006988, -0.00009653, 0.00000070, 0.00008035, -0.00011887, 0.00001686, 0.00010299, -0.00017571, 0.00001037, 0.00010750, -0.00046738, -0.00017842, 0.00084575, 0.00058430, -0.00032894, -0.00027124, 0.00006361, -0.00003779, -0.00006994, 0.00001704, -0.00000924, -0.00002167, 0.00000237, -0.00000266, -0.00000265, 0.00000436, 0.00000121, -0.00001020, -0.00000972, 0.00000497, -0.00000358, -0.00001217, 0.00000511, 0.00000187, -0.00000867, 0.00000371, 0.00000235, -0.00000603, -0.00000159, -0.00000330, 0.00000076, 0.00000800, -0.00000089, -0.00000163, 0.00000357, -0.00000451, -0.00000437, 0.00000181, -0.00000169, -0.00000058, 0.00000330, 0.00000190, 0.00000182, 0.00000047, -0.00000034, 0.00000032, -0.00000313, },
		{ 0.00000511, -0.00000288, 0.00000487, 0.00000965, -0.00000722, -0.00000949, 0.00000727, 0.00000458, -0.00000436, 0.00000835, 0.00000926, -0.00001636, -0.00001718, 0.00001054, 0.00000962, -0.00001429, -0.00001484, -0.00000230, -0.00000119, -0.00000195, 0.00000295, 0.00000145, -0.00001547, -0.00002089, 0.00000658, 0.00001623, -0.00001027, -0.00000827, 0.00000774, -0.00001048, -0.00000695, 0.00002180, 0.00000903, -0.00000935, 0.00000008, 0.00000276, 0.00000359, -0.00000090, -0.00000871, 0.00000741, -0.00000021, -0.00002014, 0.00001331, 0.00001541, -0.00001769, 0.00001296, 0.00001198, -0.00002627, 0.00001132, 0.00000703, -0.00003243, 0.00002096, 0.00001370, -0.00003401, 0.00002348, 0.00000526, -0.00003686, 0.00004247, 0.00000091, -0.00005655, 0.00005283, 0.00000743, -0.00006962, 0.00005151, 0.00000449, -0.00006879, 0.00007211, 0.00000332, -0.00008525, 0.00007384, -0.00001002, -0.00009794, 0.00009281, -0.00002014, -0.00011285, 0.00013233, -0.00003877, -0.00017404, 0.00017598, -0.00015414, -0.00055713, 0.00038123, 0.00097858, 0.00006249, -0.00046372, -0.00004073, 0.00005521, -0.00010285, -0.00001303, 0.00002095, -0.00003356, -0.00000469, 0.00000532, -0.00001080, 0.00000343, 0.00000696, -0.00000767, -0.00001304, 0.00000010, 0.00000377, -0.00001196, -0.00000427, 0.00000796, -0.00000612, -0.00000366, 0.00000645, -0.00000419, -0.00000416, -0.00000030, -0.00000405, 0.00000523, 0.00000576, -0.00000388, 0.00000212, 0.00000093, -0.00000748, -0.00000038, 0.00000151, -0.00000322, 0.00000197, 0.00000375, 0.00000126, 0.00000161, -0.00000011, -0.00000017, -0.00000130, -0.00000300, 0.00000443, },
		{ -0.00000183, 0.00000918, 0.00000294, -0.00001145, 0.00000001, 0.00000919, -0.00000351, -0.00000009, 0.00001411, -0.00000332, -0.00002303, -0.00000235, 0.00001540, -0.00000423, -0.00001793, -0.00000697, -0.00000071, -0.00000233, 0.00000031, 0.00000275, -0.00000562, -0.00002049, -0.00001056, 0.00001636, 0.00000569, -0.00001596, 0.00000099, 0.00000357, -0.00001473, 0.00000716, 0.00002291, -0.00000363, -0.00000891, 0.00000492, 0.00000392, 0.00000101, -0.00000559, -0.00000131, 0.00000737, -0.00001358, -0.00000735, 0.00002263, -0.00000318, -0.00000857, 0.00002146, -0.00001034, -0.00001382, 0.00002038, -0.00001911, -0.00001305, 0.00003268, -0.00001747, -0.00001237, 0.00003071, -0.00002857, -0.00000156, 0.00004404, -0.00004842, -0.00000782, 0.00006072, -0.00005481, -0.00001698, 0.00006137, -0.00006156, -0.00000321, 0.00007802, -0.00007966, -0.00001078, 0.00008100, -0.00010050, -0.00000610, 0.00009427, -0.00012842, 0.00000960, 0.00012662, -0.00019620, -0.00000213, 0.00016117, -0.00050094, -0.00025935, 0.00091955, 0.00066257, -0.00039921, -0.00030145, 0.00010980, -0.00005237, -0.00008710, 0.00004085, -0.00001600, -0.00003139, 0.00001559, -0.00000874, -0.00000942, 0.00001372, -0.00000123, -0.00001565, -0.00000557, 0.00000582, -0.00000506, -0.00001195, 0.00000496, 0.00000235, -0.00000860, 0.00000380, 0.00000202, -0.00000727, 0.00000029, -0.00000236, -0.00000211, 0.00000859, 0.00000064, -0.00000289, 0.00000440, -0.00000462, -0.00000620, 0.00000336, -0.00000153, -0.00000238, 0.00000485, 0.00000257, 0.00000094, 0.00000149, -0.00000074, -0.00000067, -0.00000265, -0.00000044, 0.00000687, 0.00000128, },
		{ 0.00000903, -0.00000508, -0.00000776, 0.00000821, 0.00000294, -0.00000666, 0.00000930, 0.00000994, -0.00001763, -0.00001638, 0.00001146, 0.00000747, -0.00001524, -0.00001279, -0.00000180, -0.00000198, -0.00000130, 0.00000159, -0.00000078, -0.00001244, -0.00001911, 0.00000254, 0.00001706, -0.00000742, -0.00001270, 0.00000777, -0.00000533, -0.00001021, 0.00002051, 0.00001362, -0.00001327, -0.00000204, 0.00000848, 0.00000135, -0.00000282, -0.00000397, 0.00000441, -0.00000190, -0.00001580, 0.00000987, 0.00001539, -0.00001273, 0.00000772, 0.00001160, -0.00002048, 0.00000649, 0.00000664, -0.00002833, 0.00001556, 0.00001600, -0.00002884, 0.00001681, 0.00000680, -0.00003237, 0.00003623, 0.00000271, -0.00005375, 0.00004747, 0.00001156, -0.00006706, 0.00004600, 0.00000677, -0.00006804, 0.00007062, 0.00000464, -0.00009092, 0.00007746, -0.00000268, -0.00010794, 0.00009725, -0.00001285, -0.00012935, 0.00014380, -0.00002925, -0.00020624, 0.00020183, -0.00012269, -0.00063852, 0.00036419, 0.00108834, 0.00006266, -0.00054063, -0.00001589, 0.00007949, -0.00013544, -0.00000281, 0.00003577, -0.00005194, 0.00000125, 0.00001287, -0.00002408, 0.00000745, 0.00001328, -0.00001590, -0.00001237, 0.00000433, 0.00000164, -0.00001186, -0.00000407, 0.00000728, -0.00000514, -0.00000358, 0.00000615, -0.00000508, -0.00000410, 0.00000204, -0.00000528, 0.00000313, 0.00000774, -0.00000376, 0.00000135, 0.00000203, -0.00000899, -0.00000083, 0.00000338, -0.00000470, 0.00000148, 0.00000575, 0.00000063, 0.00000163, 0.00000079, -0.00000165, -0.00000134, -0.00000257, 0.00000315, 0.00000632, -0.00000198, 0.00000222, },
		{ -0.00000884, 0.00000094, 0.00000898, -0.00000544, -0.00000119, 0.00001495, -0.00000348, -0.00002304, -0.00000120, 0.00001433, -0.00000626, -0.00001686, -0.00000528, -0.00000133, -0.00000236, 0.00000048, 0.00000011, -0.00000565, -0.00001637, -0.00001183, 0.00001289, 0.00000869, -0.00001590, -0.00000293, 0.00000706, -0.00001273, 0.00000295, 0.00002520, -0.00000217, -0.00001381, 0.00000757, 0.00000676, -0.00000310, -0.00000307, 0.00000050, 0.00000273, -0.00001068, -0.00000604, 0.00001822, 0.00000125, -0.00000795, 0.00001561, -0.00000536, -0.00001259, 0.00001479, -0.00001470, -0.00001352, 0.00002813, -0.00001057, -0.00001332, 0.00002510, -0.00002207, -0.00000309, 0.00003895, -0.00004238, -0.00001022, 0.00005748, -0.00004767, -0.00001965, 0.00005769, -0.00005716, -0.00000490, 0.00007756, -0.00008053, -0.00001434, 0.00008901, -0.00010116, -0.00001325, 0.00010514, -0.00013377, 0.00000080, 0.00014684, -0.00021075, -0.00001776, 0.00020965, -0.00052288, -0.00033888, 0.00097929, 0.00074192, -0.00046246, -0.00033409, 0.00015637, -0.00006600, -0.00010636, 0.00006655, -0.00002276, -0.00004306, 0.00003098, -0.00001513, -0.00001784, 0.00002484, -0.00000435, -0.00002201, 0.00000040, 0.00000584, -0.00000708, -0.00001022, 0.00000357, 0.00000236, -0.00000707, 0.00000310, 0.00000137, -0.00000779, 0.00000162, -0.00000108, -0.00000466, 0.00000833, 0.00000262, -0.00000382, 0.00000462, -0.00000437, -0.00000803, 0.00000457, -0.00000085, -0.00000437, 0.00000607, 0.00000364, -0.00000006, 0.00000255, -0.00000112, -0.00000216, -0.00000152, -0.00000105, 0.00000591, 0.00000349, -0.00000258, 0.00000621, 0.00000419, },
		{ 0.00000802, 0.00000233, -0.00000828, 0.00000864, 0.00001040, -0.00001733, -0.00001552, 0.00001130, 0.00000532, -0.00001562, -0.00001080, -0.00000138, -0.00000268, -0.00000076, 0.00000033, -0.00000320, -0.00000971, -0.00001621, -0.00000128, 0.00001590, -0.00000406, -0.00001506, 0.00000642, -0.00000095, -0.00001168, 0.00001815, 0.00001697, -0.00001530, -0.00000429, 0.00001286, 0.00000006, -0.00000462, 0.00000012, 0.00000186, -0.00000370, -0.00001125, 0.00000674, 0.00001436, -0.00000738, 0.00000316, 0.00001010, -0.00001427, 0.00000218, 0.00000528, -0.00002329, 0.00001033, 0.00001686, -0.00002259, 0.00001034, 0.00000733, -0.00002665, 0.00002924, 0.00000375, -0.00004871, 0.00004063, 0.00001436, -0.00006171, 0.00003913, 0.00000800, -0.00006460, 0.00006689, 0.00000540, -0.00009334, 0.00007812, 0.00000441, -0.00011439, 0.00009801, -0.00000501, -0.00014210, 0.00014997, -0.00001819, -0.00023362, 0.00021926, -0.00008872, -0.00070707, 0.00033796, 0.00118681, 0.00007097, -0.00061403, 0.00000539, 0.00010551, -0.00016847, 0.00000655, 0.00005246, -0.00007186, 0.00000708, 0.00002221, -0.00003911, 0.00001162, 0.00002074, -0.00002577, -0.00001080, 0.00000947, -0.00000215, -0.00001069, -0.00000336, 0.00000481, -0.00000312, -0.00000288, 0.00000457, -0.00000528, -0.00000386, 0.00000375, -0.00000567, 0.00000069, 0.00000912, -0.00000280, 0.00000027, 0.00000278, -0.00000992, -0.00000169, 0.00000524, -0.00000569, 0.00000050, 0.00000774, 0.00000032, 0.00000145, 0.00000179, -0.00000340, -0.00000145, -0.00000135, 0.00000124, 0.00000690, 0.00000015, -0.00000066, 0.00000735, -0.00000193, -0.00000597, },
		{ -0.00000586, -0.00000289, 0.00001435, -0.00000258, -0.00002193, -0.00000092, 0.00001262, -0.00000775, -0.00001556, -0.00000379, -0.00000185, -0.00000242, 0.00000058, -0.00000242, -0.00000609, -0.00001200, -0.00001192, 0.00000838, 0.00001030, -0.00001388, -0.00000618, 0.00000859, -0.00000985, -0.00000034, 0.00002543, -0.00000024, -0.00001707, 0.00000876, 0.00000932, -0.00000606, -0.00000128, 0.00000219, -0.00000141, -0.00000809, -0.00000424, 0.00001355, 0.00000489, -0.00000634, 0.00000961, -0.00000124, -0.00001041, 0.00000891, -0.00001068, -0.00001280, 0.00002265, -0.00000426, -0.00001294, 0.00001871, -0.00001565, -0.00000379, 0.00003242, -0.00003541, -0.00001153, 0.00005174, -0.00003945, -0.00002080, 0.00005155, -0.00005129, -0.00000579, 0.00007415, -0.00007883, -0.00001724, 0.00009340, -0.00009852, -0.00002005, 0.00011208, -0.00013466, -0.00000862, 0.00016207, -0.00021868, -0.00003490, 0.00024990, -0.00053285, -0.00041273, 0.00102211, 0.00081910, -0.00051487, -0.00036862, 0.00020058, -0.00007723, -0.00012709, 0.00009266, -0.00002868, -0.00005637, 0.00004768, -0.00002126, -0.00002776, 0.00003723, -0.00000789, -0.00002920, 0.00000815, 0.00000498, -0.00000983, -0.00000669, 0.00000098, 0.00000155, -0.00000382, 0.00000159, 0.00000015, -0.00000718, 0.00000223, 0.00000023, -0.00000642, 0.00000708, 0.00000475, -0.00000409, 0.00000405, -0.00000385, -0.00000952, 0.00000515, 0.00000027, -0.00000617, 0.00000671, 0.00000502, -0.00000094, 0.00000340, -0.00000145, -0.00000396, 0.00000001, -0.00000133, 0.00000394, 0.00000603, -0.00000238, 0.00000219, 0.00000534, -0.00000519, 0.00000091, 0.00000847, },
		{ 0.00000643, 0.00001055, -0.00001542, -0.00001463, 0.00001004, 0.00000329, -0.00001546, -0.00000893, -0.00000101, -0.00000326, -0.00000038, -0.00000084, -0.00000567, -0.00000734, -0.00001245, -0.00000464, 0.00001295, -0.00000055, -0.00001529, 0.00000402, 0.00000236, -0.00001139, 0.00001501, 0.00001878, -0.00001545, -0.00000635, 0.00001559, -0.00000024, -0.00000605, 0.00000324, -0.00000010, -0.00000541, -0.00000690, 0.00000415, 0.00001251, -0.00000214, -0.00000040, 0.00000771, -0.00000817, -0.00000128, 0.00000312, -0.00001776, 0.00000572, 0.00001630, -0.00001580, 0.00000459, 0.00000682, -0.00002021, 0.00002207, 0.00000392, -0.00004189, 0.00003294, 0.00001562, -0.00005402, 0.00003149, 0.00000801, -0.00005877, 0.00006129, 0.00000539, -0.00009235, 0.00007596, 0.00001070, -0.00011675, 0.00009515, 0.00000259, -0.00015008, 0.00015061, -0.00000678, -0.00025421, 0.00022775, -0.00005501, -0.00075949, 0.00030429, 0.00126921, 0.00008787, -0.00068031, 0.00002093, 0.00013234, -0.00019988, 0.00001403, 0.00007045, -0.00009210, 0.00001213, 0.00003306, -0.00005509, 0.00001546, 0.00002913, -0.00003688, -0.00000847, 0.00001553, -0.00000767, -0.00000850, -0.00000183, 0.00000032, -0.00000020, -0.00000126, 0.00000147, -0.00000474, -0.00000318, 0.00000442, -0.00000514, -0.00000173, 0.00000952, -0.00000095, -0.00000089, 0.00000287, -0.00001007, -0.00000282, 0.00000668, -0.00000593, -0.00000082, 0.00000936, 0.00000053, 0.00000111, 0.00000267, -0.00000515, -0.00000168, 0.00000042, -0.00000098, 0.00000679, 0.00000324, -0.00000348, 0.00000465, 0.00000201, -0.00000399, 0.00000663, 0.00000260, -0.00000900, },
		{ -0.00000075, -0.00001973, -0.00000153, 0.00001034, -0.00000867, -0.00001408, -0.00000252, -0.00000224, -0.00000250, 0.00000056, -0.00000476, -0.00000680, -0.00000760, -0.00001098, 0.00000322, 0.00001048, -0.00001029, -0.00000846, 0.00000826, -0.00000642, -0.00000247, 0.00002368, 0.00000185, -0.00001844, 0.00000860, 0.00001132, -0.00000765, -0.00000023, 0.00000355, -0.00000472, -0.00000596, -0.00000218, 0.00000905, 0.00000751, -0.00000401, 0.00000398, 0.00000178, -0.00000754, 0.00000322, -0.00000736, -0.00001102, 0.00001675, 0.00000102, -0.00001130, 0.00001211, -0.00000983, -0.00000365, 0.00002500, -0.00002810, -0.00001165, 0.00004403, -0.00003082, -0.00002036, 0.00004346, -0.00004449, -0.00000577, 0.00006804, -0.00007485, -0.00001915, 0.00009392, -0.00009283, -0.00002583, 0.00011451, -0.00013118, -0.00001769, 0.00017113, -0.00021982, -0.00005181, 0.00027947, -0.00053128, -0.00047674, 0.00104608, 0.00089056, -0.00055307, -0.00040399, 0.00023963, -0.00008478, -0.00014839, 0.00011751, -0.00003300, -0.00007081, 0.00006462, -0.00002653, -0.00003889, 0.00005018, -0.00001153, -0.00003707, 0.00001745, 0.00000338, -0.00001348, -0.00000126, -0.00000260, -0.00000040, 0.00000127, -0.00000062, -0.00000189, -0.00000516, 0.00000206, 0.00000119, -0.00000698, 0.00000488, 0.00000668, -0.00000342, 0.00000263, -0.00000325, -0.00001030, 0.00000488, 0.00000164, -0.00000740, 0.00000658, 0.00000656, -0.00000141, 0.00000385, -0.00000173, -0.00000577, 0.00000165, -0.00000126, 0.00000126, 0.00000855, -0.00000128, -0.00000251, 0.00000618, -0.00000061, 0.00000011, 0.00000748, -0.00000484, -0.00000494, 0.00001230, },
		{ -0.00001370, 0.00000775, 0.00000142, -0.00001478, -0.00000723, -0.00000069, -0.00000368, -0.00000018, -0.00000198, -0.00000804, -0.00000532, -0.00000816, -0.00000744, 0.00000860, 0.00000276, -0.00001357, 0.00000099, 0.00000446, -0.00000955, 0.00001145, 0.00001892, -0.00001389, -0.00000792, 0.00001656, 0.00000033, -0.00000692, 0.00000519, -0.00000141, -0.00000682, -0.00000311, 0.00000225, 0.00001011, 0.00000255, -0.00000280, 0.00000476, -0.00000267, -0.00000371, 0.00000040, -0.00001226, 0.00000205, 0.00001448, -0.00000902, 0.00000002, 0.00000538, -0.00001365, 0.00001528, 0.00000325, -0.00003389, 0.00002506, 0.00001531, -0.00004466, 0.00002373, 0.00000679, -0.00005105, 0.00005431, 0.00000451, -0.00008809, 0.00007130, 0.00001569, -0.00011480, 0.00008900, 0.00000916, -0.00015264, 0.00014590, 0.00000374, -0.00026655, 0.00022743, -0.00002429, -0.00079342, 0.00026548, 0.00133135, 0.00011307, -0.00073585, 0.00002901, 0.00015870, -0.00022753, 0.00001874, 0.00008890, -0.00011130, 0.00001576, 0.00004497, -0.00007106, 0.00001852, 0.00003818, -0.00004869, -0.00000563, 0.00002246, -0.00001482, -0.00000551, 0.00000080, -0.00000618, 0.00000335, 0.00000150, -0.00000327, -0.00000352, -0.00000173, 0.00000370, -0.00000375, -0.00000370, 0.00000863, 0.00000169, -0.00000185, 0.00000203, -0.00000932, -0.00000401, 0.00000732, -0.00000526, -0.00000224, 0.00001027, 0.00000139, 0.00000073, 0.00000314, -0.00000664, -0.00000197, 0.00000241, -0.00000320, 0.00000600, 0.00000687, -0.00000594, 0.00000110, 0.00000639, -0.00000167, 0.00000414, 0.00000345, -0.00000870, 0.00000293, 0.00001025, -0.00001206, },
		{ 0.00000755, -0.00000902, -0.00001247, -0.00000147, -0.00000249, -0.00000258, 0.00000036, -0.00000692, -0.00000762, -0.00000334, -0.00000927, -0.00000220, 0.00000937, -0.00000564, -0.00000961, 0.00000644, -0.00000283, -0.00000342, 0.00002026, 0.00000379, -0.00001790, 0.00000732, 0.00001254, -0.00000781, 0.00000014, 0.00000439, -0.00000697, -0.00000434, -0.00000013, 0.00000512, 0.00000904, -0.00000129, -0.00000082, 0.00000360, -0.00000433, -0.00000181, -0.00000493, -0.00000845, 0.00001097, 0.00000495, -0.00000864, 0.00000587, -0.00000504, -0.00000282, 0.00001732, -0.00002101, -0.00001063, 0.00003505, -0.00002246, -0.00001841, 0.00003411, -0.00003731, -0.00000485, 0.00005969, -0.00006901, -0.00001987, 0.00009065, -0.00008455, -0.00003004, 0.00011222, -0.00012377, -0.00002550, 0.00017330, -0.00021450, -0.00006680, 0.00029667, -0.00051928, -0.00052727, 0.00105034, 0.00095262, -0.00057446, -0.00043871, 0.00027080, -0.00008773, -0.00016911, 0.00013941, -0.00003509, -0.00008568, 0.00008065, -0.00003038, -0.00005076, 0.00006289, -0.00001489, -0.00004538, 0.00002789, 0.00000127, -0.00001814, 0.00000596, -0.00000681, -0.00000374, 0.00000811, -0.00000330, -0.00000496, -0.00000156, 0.00000123, 0.00000142, -0.00000607, 0.00000188, 0.00000800, -0.00000158, 0.00000044, -0.00000280, -0.00001007, 0.00000369, 0.00000298, -0.00000769, 0.00000560, 0.00000801, -0.00000119, 0.00000374, -0.00000204, -0.00000727, 0.00000311, -0.00000090, -0.00000171, 0.00001070, 0.00000057, -0.00000742, 0.00000653, 0.00000465, -0.00000113, 0.00000585, -0.00000261, -0.00000699, 0.00000893, 0.00000177, -0.00001649, -0.00000292, },
		{ -0.00001365, -0.00000572, -0.00000042, -0.00000393, -0.00000013, -0.00000313, -0.00001022, -0.00000355, -0.00000367, -0.00000965, 0.00000336, 0.00000561, -0.00001030, -0.00000221, 0.00000536, -0.00000656, 0.00000777, 0.00001744, -0.00001098, -0.00000876, 0.00001584, 0.00000154, -0.00000707, 0.00000592, -0.00000212, -0.00000774, -0.00000013, 0.00000106, 0.00000751, 0.00000630, -0.00000402, 0.00000162, 0.00000181, -0.00000507, -0.00000254, -0.00000722, -0.00000051, 0.00001171, -0.00000280, -0.00000311, 0.00000325, -0.00000754, 0.00000929, 0.00000189, -0.00002539, 0.00001759, 0.00001357, -0.00003441, 0.00001640, 0.00000449, -0.00004208, 0.00004645, 0.00000277, -0.00008095, 0.00006460, 0.00001903, -0.00010868, 0.00008014, 0.00001409, -0.00014960, 0.00013642, 0.00001226, -0.00026984, 0.00021902, 0.00000103, -0.00080759, 0.00022421, 0.00137001, 0.00014548, -0.00077731, 0.00002850, 0.00018305, -0.00024944, 0.00002011, 0.00010680, -0.00012808, 0.00001746, 0.00005730, -0.00008598, 0.00002034, 0.00004745, -0.00006049, -0.00000258, 0.00003012, -0.00002329, -0.00000207, 0.00000472, -0.00001447, 0.00000714, 0.00000558, -0.00000955, -0.00000181, 0.00000074, 0.00000143, -0.00000168, -0.00000481, 0.00000627, 0.00000490, -0.00000229, 0.00000009, -0.00000769, -0.00000498, 0.00000684, -0.00000368, -0.00000347, 0.00001018, 0.00000291, 0.00000046, 0.00000295, -0.00000765, -0.00000224, 0.00000425, -0.00000510, 0.00000470, 0.00001054, -0.00000774, -0.00000298, 0.00001075, 0.00000107, 0.00000081, 0.00000448, -0.00000736, -0.00000141, 0.00000938, -0.00000752, -0.00001265, 0.00000456, -0.00000029, },
		{ -0.00000064, -0.00000261, -0.00000263, -0.00000003, -0.00000891, -0.00000838, 0.00000066, -0.00000711, -0.00000750, 0.00000727, -0.00000051, -0.00000963, 0.00000361, 0.00000058, -0.00000330, 0.00001560, 0.00000532, -0.00001563, 0.00000527, 0.00001285, -0.00000667, 0.00000001, 0.00000460, -0.00000810, -0.00000321, 0.00000166, 0.00000202, 0.00000951, 0.00000144, -0.00000446, 0.00000426, -0.00000116, -0.00000583, -0.00000342, -0.00000543, 0.00000576, 0.00000738, -0.00000528, 0.00000051, -0.00000154, -0.00000151, 0.00001000, -0.00001460, -0.00000868, 0.00002558, -0.00001491, -0.00001521, 0.00002425, -0.00003027, -0.00000318, 0.00004973, -0.00006181, -0.00001932, 0.00008396, -0.00007430, -0.00003229, 0.00010541, -0.00011312, -0.00003129, 0.00016839, -0.00020351, -0.00007842, 0.00030072, -0.00049848, -0.00056144, 0.00103514, 0.00100181, -0.00057741, -0.00047091, 0.00029176, -0.00008559, -0.00018794, 0.00015680, -0.00003455, -0.00010010, 0.00009460, -0.00003239, -0.00006279, 0.00007444, -0.00001759, -0.00005377, 0.00003888, -0.00000103, -0.00002378, 0.00001464, -0.00001118, -0.00000860, 0.00001638, -0.00000611, -0.00000920, 0.00000357, -0.00000006, 0.00000059, -0.00000355, -0.00000163, 0.00000834, 0.00000150, -0.00000235, -0.00000278, -0.00000861, 0.00000161, 0.00000397, -0.00000675, 0.00000383, 0.00000909, -0.00000009, 0.00000300, -0.00000249, -0.00000814, 0.00000415, -0.00000041, -0.00000451, 0.00001220, 0.00000288, -0.00001197, 0.00000625, 0.00001022, -0.00000241, 0.00000353, 0.00000039, -0.00000867, 0.00000453, 0.00000459, -0.00001235, -0.00000488, 0.00000431, -0.00000889, -0.00001075, },
		{ -0.00000402, -0.00000022, -0.00000436, -0.00001211, -0.00000192, 0.00000067, -0.00001134, -0.00000221, 0.00000784, -0.00000604, -0.00000511, 0.00000526, -0.00000292, 0.00000425, 0.00001459, -0.00000717, -0.00000871, 0.00001370, 0.00000311, -0.00000645, 0.00000553, -0.00000234, -0.00000803, 0.00000189, 0.00000051, 0.00000502, 0.00000887, -0.00000416, -0.00000131, 0.00000502, -0.00000542, -0.00000535, -0.00000303, -0.00000192, 0.00000838, 0.00000240, -0.00000469, 0.00000077, -0.00000239, 0.00000440, 0.00000010, -0.00001708, 0.00001099, 0.00001071, -0.00002409, 0.00000996, 0.00000141, -0.00003259, 0.00003823, 0.00000032, -0.00007151, 0.00005646, 0.00002058, -0.00009883, 0.00006929, 0.00001692, -0.00014118, 0.00012303, 0.00001790, -0.00026394, 0.00020372, 0.00001903, -0.00080189, 0.00018331, 0.00138313, 0.00018331, -0.00080189, 0.00001903, 0.00020372, -0.00026394, 0.00001790, 0.00012303, -0.00014118, 0.00001692, 0.00006929, -0.00009883, 0.00002058, 0.00005646, -0.00007151, 0.00000032, 0.00003823, -0.00003259, 0.00000141, 0.00000996, -0.00002409, 0.00001071, 0.00001099, -0.00001708, 0.00000010, 0.00000440, -0.00000239, 0.00000077, -0.00000469, 0.00000240, 0.00000838, -0.00000192, -0.00000303, -0.00000535, -0.00000542, 0.00000502, -0.00000131, -0.00000416, 0.00000887, 0.00000502, 0.00000051, 0.00000189, -0.00000803, -0.00000234, 0.00000553, -0.00000645, 0.00000311, 0.00001370, -0.00000871, -0.00000717, 0.00001459, 0.00000425, -0.00000292, 0.00000526, -0.00000511, -0.00000604, 0.00000784, -0.00000221, -0.00001134, 0.00000067, -0.00000192, -0.00001211, -0.00000436, -0.00000022, },
		{ -0.00000064, -0.00001075, -0.00000889, 0.00000431, -0.00000488, -0.00001235, 0.00000459, 0.00000453, -0.00000867, 0.00000039, 0.00000353, -0.00000241, 0.00001022, 0.00000625, -0.00001197, 0.00000288, 0.00001220, -0.00000451, -0.00000041, 0.00000415, -0.00000814, -0.00000249, 0.00000300, -0.00000009, 0.00000909, 0.00000383, -0.00000675, 0.00000397, 0.00000161, -0.00000861, -0.00000278, -0.00000235, 0.00000150, 0.00000834, -0.00000163, -0.00000355, 0.00000059, -0.00000006, 0.00000357, -0.00000920, -0.00000611, 0.00001638, -0.00000860, -0.00001118, 0.00001464, -0.00002378, -0.00000103, 0.00003888, -0.00005377, -0.00001759, 0.00007444, -0.00006279, -0.00003239, 0.00009460, -0.00010010, -0.00003455, 0.00015680, -0.00018794, -0.00008559, 0.00029176, -0.00047091, -0.00057741, 0.00100181, 0.00103514, -0.00056144, -0.00049848, 0.00030072, -0.00007842, -0.00020351, 0.00016839, -0.00003129, -0.00011312, 0.00010541, -0.00003229, -0.00007430, 0.00008396, -0.00001932, -0.00006181, 0.00004973, -0.00000318, -0.00003027, 0.00002425, -0.00001521, -0.00001491, 0.00002558, -0.00000868, -0.00001460, 0.00001000, -0.00000151, -0.00000154, 0.00000051, -0.00000528, 0.00000738, 0.00000576, -0.00000543, -0.00000342, -0.00000583, -0.00000116, 0.00000426, -0.00000446, 0.00000144, 0.00000951, 0.00000202, 0.00000166, -0.00000321, -0.00000810, 0.00000460, 0.00000001, -0.00000667, 0.00001285, 0.00000527, -0.00001563, 0.00000532, 0.00001560, -0.00000330, 0.00000058, 0.00000361, -0.00000963, -0.00000051, 0.00000727, -0.00000750, -0.00000711, 0.00000066, -0.00000838, -0.00000891, -0.00000003, -0.00000263, -0.00000261, },
		{ -0.00001365, -0.00000029, 0.00000456, -0.00001265, -0.00000752, 0.00000938, -0.00000141, -0.00000736, 0.00000448, 0.00000081, 0.00000107, 0.00001075, -0.00000298, -0.00000774, 0.00001054, 0.00000470, -0.00000510, 0.00000425, -0.00000224, -0.00000765, 0.00000295, 0.00000046, 0.00000291, 0.00001018, -0.00000347, -0.00000368, 0.00000684, -0.00000498, -0.00000769, 0.00000009, -0.00000229, 0.00000490, 0.00000627, -0.00000481, -0.00000168, 0.00000143, 0.00000074, -0.00000181, -0.00000955, 0.00000558, 0.00000714, -0.00001447, 0.00000472, -0.00000207, -0.00002329, 0.00003012, -0.00000258, -0.00006049, 0.00004745, 0.00002034, -0.00008598, 0.00005730, 0.00001746, -0.00012808, 0.00010680, 0.00002011, -0.00024944, 0.00018305, 0.00002850, -0.00077731, 0.00014548, 0.00137001, 0.00022421, -0.00080759, 0.00000103, 0.00021902, -0.00026984, 0.00001226, 0.00013642, -0.00014960, 0.00001409, 0.00008014, -0.00010868, 0.00001903, 0.00006460, -0.00008095, 0.00000277, 0.00004645, -0.00004208, 0.00000449, 0.00001640, -0.00003441, 0.00001357, 0.00001759, -0.00002539, 0.00000189, 0.00000929, -0.00000754, 0.00000325, -0.00000311, -0.00000280, 0.00001171, -0.00000051, -0.00000722, -0.00000254, -0.00000507, 0.00000181, 0.00000162, -0.00000402, 0.00000630, 0.00000751, 0.00000106, -0.00000013, -0.00000774, -0.00000212, 0.00000592, -0.00000707, 0.00000154, 0.00001584, -0.00000876, -0.00001098, 0.00001744, 0.00000777, -0.00000656, 0.00000536, -0.00000221, -0.00001030, 0.00000561, 0.00000336, -0.00000965, -0.00000367, -0.00000355, -0.00001022, -0.00000313, -0.00000013, -0.00000393, -0.00000042, -0.00000572, },
		{ 0.00000755, -0.00000292, -0.00001649, 0.00000177, 0.00000893, -0.00000699, -0.00000261, 0.00000585, -0.00000113, 0.00000465, 0.00000653, -0.00000742, 0.00000057, 0.00001070, -0.00000171, -0.00000090, 0.00000311, -0.00000727, -0.00000204, 0.00000374, -0.00000119, 0.00000801, 0.00000560, -0.00000769, 0.00000298, 0.00000369, -0.00001007, -0.00000280, 0.00000044, -0.00000158, 0.00000800, 0.00000188, -0.00000607, 0.00000142, 0.00000123, -0.00000156, -0.00000496, -0.00000330, 0.00000811, -0.00000374, -0.00000681, 0.00000596, -0.00001814, 0.00000127, 0.00002789, -0.00004538, -0.00001489, 0.00006289, -0.00005076, -0.00003038, 0.00008065, -0.00008568, -0.00003509, 0.00013941, -0.00016911, -0.00008773, 0.00027080, -0.00043871, -0.00057446, 0.00095262, 0.00105034, -0.00052727, -0.00051928, 0.00029667, -0.00006680, -0.00021450, 0.00017330, -0.00002550, -0.00012377, 0.00011222, -0.00003004, -0.00008455, 0.00009065, -0.00001987, -0.00006901, 0.00005969, -0.00000485, -0.00003731, 0.00003411, -0.00001841, -0.00002246, 0.00003505, -0.00001063, -0.00002101, 0.00001732, -0.00000282, -0.00000504, 0.00000587, -0.00000864, 0.00000495, 0.00001097, -0.00000845, -0.00000493, -0.00000181, -0.00000433, 0.00000360, -0.00000082, -0.00000129, 0.00000904, 0.00000512, -0.00000013, -0.00000434, -0.00000697, 0.00000439, 0.00000014, -0.00000781, 0.00001254, 0.00000732, -0.00001790, 0.00000379, 0.00002026, -0.00000342, -0.00000283, 0.00000644, -0.00000961, -0.00000564, 0.00000937, -0.00000220, -0.00000927, -0.00000334, -0.00000762, -0.00000692, 0.00000036, -0.00000258, -0.00000249, -0.00000147, -0.00001247, -0.00000902, },
		{ -0.00001370, -0.00001206, 0.00001025, 0.00000293, -0.00000870, 0.00000345, 0.00000414, -0.00000167, 0.00000639, 0.00000110, -0.00000594, 0.00000687, 0.00000600, -0.00000320, 0.00000241, -0.00000197, -0.00000664, 0.00000314, 0.00000073, 0.00000139, 0.00001027, -0.00000224, -0.00000526, 0.00000732, -0.00000401, -0.00000932, 0.00000203, -0.00000185, 0.00000169, 0.00000863, -0.00000370, -0.00000375, 0.00000370, -0.00000173, -0.00000352, -0.00000327, 0.00000150, 0.00000335, -0.00000618, 0.00000080, -0.00000551, -0.00001482, 0.00002246, -0.00000563, -0.00004869, 0.00003818, 0.00001852, -0.00007106, 0.00004497, 0.00001576, -0.00011130, 0.00008890, 0.00001874, -0.00022753, 0.00015870, 0.00002901, -0.00073585, 0.00011307, 0.00133135, 0.00026548, -0.00079342, -0.00002429, 0.00022743, -0.00026655, 0.00000374, 0.00014590, -0.00015264, 0.00000916, 0.00008900, -0.00011480, 0.00001569, 0.00007130, -0.00008809, 0.00000451, 0.00005431, -0.00005105, 0.00000679, 0.00002373, -0.00004466, 0.00001531, 0.00002506, -0.00003389, 0.00000325, 0.00001528, -0.00001365, 0.00000538, 0.00000002, -0.00000902, 0.00001448, 0.00000205, -0.00001226, 0.00000040, -0.00000371, -0.00000267, 0.00000476, -0.00000280, 0.00000255, 0.00001011, 0.00000225, -0.00000311, -0.00000682, -0.00000141, 0.00000519, -0.00000692, 0.00000033, 0.00001656, -0.00000792, -0.00001389, 0.00001892, 0.00001145, -0.00000955, 0.00000446, 0.00000099, -0.00001357, 0.00000276, 0.00000860, -0.00000744, -0.00000816, -0.00000532, -0.00000804, -0.00000198, -0.00000018, -0.00000368, -0.00000069, -0.00000723, -0.00001478, 0.00000142, 0.00000775, },
		{ -0.00000075, 0.00001230, -0.00000494, -0.00000484, 0.00000748, 0.00000011, -0.00000061, 0.00000618, -0.00000251, -0.00000128, 0.00000855, 0.00000126, -0.00000126, 0.00000165, -0.00000577, -0.00000173, 0.00000385, -0.00000141, 0.00000656, 0.00000658, -0.00000740, 0.00000164, 0.00000488, -0.00001030, -0.00000325, 0.00000263, -0.00000342, 0.00000668, 0.00000488, -0.00000698, 0.00000119, 0.00000206, -0.00000516, -0.00000189, -0.00000062, 0.00000127, -0.00000040, -0.00000260, -0.00000126, -0.00001348, 0.00000338, 0.00001745, -0.00003707, -0.00001153, 0.00005018, -0.00003889, -0.00002653, 0.00006462, -0.00007081, -0.00003300, 0.00011751, -0.00014839, -0.00008478, 0.00023963, -0.00040399, -0.00055307, 0.00089056, 0.00104608, -0.00047674, -0.00053128, 0.00027947, -0.00005181, -0.00021982, 0.00017113, -0.00001769, -0.00013118, 0.00011451, -0.00002583, -0.00009283, 0.00009392, -0.00001915, -0.00007485, 0.00006804, -0.00000577, -0.00004449, 0.00004346, -0.00002036, -0.00003082, 0.00004403, -0.00001165, -0.00002810, 0.00002500, -0.00000365, -0.00000983, 0.00001211, -0.00001130, 0.00000102, 0.00001675, -0.00001102, -0.00000736, 0.00000322, -0.00000754, 0.00000178, 0.00000398, -0.00000401, 0.00000751, 0.00000905, -0.00000218, -0.00000596, -0.00000472, 0.00000355, -0.00000023, -0.00000765, 0.00001132, 0.00000860, -0.00001844, 0.00000185, 0.00002368, -0.00000247, -0.00000642, 0.00000826, -0.00000846, -0.00001029, 0.00001048, 0.00000322, -0.00001098, -0.00000760, -0.00000680, -0.00000476, 0.00000056, -0.00000250, -0.00000224, -0.00000252, -0.00001408, -0.00000867, 0.00001034, -0.00000153, -0.00001973, },
		{ 0.00000643, -0.00000900, 0.00000260, 0.00000663, -0.00000399, 0.00000201, 0.00000465, -0.00000348, 0.00000324, 0.00000679, -0.00000098, 0.00000042, -0.00000168, -0.00000515, 0.00000267, 0.00000111, 0.00000053, 0.00000936, -0.00000082, -0.00000593, 0.00000668, -0.00000282, -0.00001007, 0.00000287, -0.00000089, -0.00000095, 0.00000952, -0.00000173, -0.00000514, 0.00000442, -0.00000318, -0.00000474, 0.00000147, -0.00000126, -0.00000020, 0.00000032, -0.00000183, -0.00000850, -0.00000767, 0.00001553, -0.00000847, -0.00003688, 0.00002913, 0.00001546, -0.00005509, 0.00003306, 0.00001213, -0.00009210, 0.00007045, 0.00001403, -0.00019988, 0.00013234, 0.00002093, -0.00068031, 0.00008787, 0.00126921, 0.00030429, -0.00075949, -0.00005501, 0.00022775, -0.00025421, -0.00000678, 0.00015061, -0.00015008, 0.00000259, 0.00009515, -0.00011675, 0.00001070, 0.00007596, -0.00009235, 0.00000539, 0.00006129, -0.00005877, 0.00000801, 0.00003149, -0.00005402, 0.00001562, 0.00003294, -0.00004189, 0.00000392, 0.00002207, -0.00002021, 0.00000682, 0.00000459, -0.00001580, 0.00001630, 0.00000572, -0.00001776, 0.00000312, -0.00000128, -0.00000817, 0.00000771, -0.00000040, -0.00000214, 0.00001251, 0.00000415, -0.00000690, -0.00000541, -0.00000010, 0.00000324, -0.00000605, -0.00000024, 0.00001559, -0.00000635, -0.00001545, 0.00001878, 0.00001501, -0.00001139, 0.00000236, 0.00000402, -0.00001529, -0.00000055, 0.00001295, -0.00000464, -0.00001245, -0.00000734, -0.00000567, -0.00000084, -0.00000038, -0.00000326, -0.00000101, -0.00000893, -0.00001546, 0.00000329, 0.00001004, -0.00001463, -0.00001542, 0.00001055, },
		{ -0.00000586, 0.00000847, 0.00000091, -0.00000519, 0.00000534, 0.00000219, -0.00000238, 0.00000603, 0.00000394, -0.00000133, 0.00000001, -0.00000396, -0.00000145, 0.00000340, -0.00000094, 0.00000502, 0.00000671, -0.00000617, 0.00000027, 0.00000515, -0.00000952, -0.00000385, 0.00000405, -0.00000409, 0.00000475, 0.00000708, -0.00000642, 0.00000023, 0.00000223, -0.00000718, 0.00000015, 0.00000159, -0.00000382, 0.00000155, 0.00000098, -0.00000669, -0.00000983, 0.00000498, 0.00000815, -0.00002920, -0.00000789, 0.00003723, -0.00002776, -0.00002126, 0.00004768, -0.00005637, -0.00002868, 0.00009266, -0.00012709, -0.00007723, 0.00020058, -0.00036862, -0.00051487, 0.00081910, 0.00102211, -0.00041273, -0.00053285, 0.00024990, -0.00003490, -0.00021868, 0.00016207, -0.00000862, -0.00013466, 0.00011208, -0.00002005, -0.00009852, 0.00009340, -0.00001724, -0.00007883, 0.00007415, -0.00000579, -0.00005129, 0.00005155, -0.00002080, -0.00003945, 0.00005174, -0.00001153, -0.00003541, 0.00003242, -0.00000379, -0.00001565, 0.00001871, -0.00001294, -0.00000426, 0.00002265, -0.00001280, -0.00001068, 0.00000891, -0.00001041, -0.00000124, 0.00000961, -0.00000634, 0.00000489, 0.00001355, -0.00000424, -0.00000809, -0.00000141, 0.00000219, -0.00000128, -0.00000606, 0.00000932, 0.00000876, -0.00001707, -0.00000024, 0.00002543, -0.00000034, -0.00000985, 0.00000859, -0.00000618, -0.00001388, 0.00001030, 0.00000838, -0.00001192, -0.00001200, -0.00000609, -0.00000242, 0.00000058, -0.00000242, -0.00000185, -0.00000379, -0.00001556, -0.00000775, 0.00001262, -0.00000092, -0.00002193, -0.00000258, 0.00001435, -0.00000289, },
		{ 0.00000802, -0.00000597, -0.00000193, 0.00000735, -0.00000066, 0.00000015, 0.00000690, 0.00000124, -0.00000135, -0.00000145, -0.00000340, 0.00000179, 0.00000145, 0.00000032, 0.00000774, 0.00000050, -0.00000569, 0.00000524, -0.00000169, -0.00000992, 0.00000278, 0.00000027, -0.00000280, 0.00000912, 0.00000069, -0.00000567, 0.00000375, -0.00000386, -0.00000528, 0.00000457, -0.00000288, -0.00000312, 0.00000481, -0.00000336, -0.00001069, -0.00000215, 0.00000947, -0.00001080, -0.00002577, 0.00002074, 0.00001162, -0.00003911, 0.00002221, 0.00000708, -0.00007186, 0.00005246, 0.00000655, -0.00016847, 0.00010551, 0.00000539, -0.00061403, 0.00007097, 0.00118681, 0.00033796, -0.00070707, -0.00008872, 0.00021926, -0.00023362, -0.00001819, 0.00014997, -0.00014210, -0.00000501, 0.00009801, -0.00011439, 0.00000441, 0.00007812, -0.00009334, 0.00000540, 0.00006689, -0.00006460, 0.00000800, 0.00003913, -0.00006171, 0.00001436, 0.00004063, -0.00004871, 0.00000375, 0.00002924, -0.00002665, 0.00000733, 0.00001034, -0.00002259, 0.00001686, 0.00001033, -0.00002329, 0.00000528, 0.00000218, -0.00001427, 0.00001010, 0.00000316, -0.00000738, 0.00001436, 0.00000674, -0.00001125, -0.00000370, 0.00000186, 0.00000012, -0.00000462, 0.00000006, 0.00001286, -0.00000429, -0.00001530, 0.00001697, 0.00001815, -0.00001168, -0.00000095, 0.00000642, -0.00001506, -0.00000406, 0.00001590, -0.00000128, -0.00001621, -0.00000971, -0.00000320, 0.00000033, -0.00000076, -0.00000268, -0.00000138, -0.00001080, -0.00001562, 0.00000532, 0.00001130, -0.00001552, -0.00001733, 0.00001040, 0.00000864, -0.00000828, 0.00000233, },
		{ -0.00000884, 0.00000419, 0.00000621, -0.00000258, 0.00000349, 0.00000591, -0.00000105, -0.00000152, -0.00000216, -0.00000112, 0.00000255, -0.00000006, 0.00000364, 0.00000607, -0.00000437, -0.00000085, 0.00000457, -0.00000803, -0.00000437, 0.00000462, -0.00000382, 0.00000262, 0.00000833, -0.00000466, -0.00000108, 0.00000162, -0.00000779, 0.00000137, 0.00000310, -0.00000707, 0.00000236, 0.00000357, -0.00001022, -0.00000708, 0.00000584, 0.00000040, -0.00002201, -0.00000435, 0.00002484, -0.00001784, -0.00001513, 0.00003098, -0.00004306, -0.00002276, 0.00006655, -0.00010636, -0.00006600, 0.00015637, -0.00033409, -0.00046246, 0.00074192, 0.00097929, -0.00033888, -0.00052288, 0.00020965, -0.00001776, -0.00021075, 0.00014684, 0.00000080, -0.00013377, 0.00010514, -0.00001325, -0.00010116, 0.00008901, -0.00001434, -0.00008053, 0.00007756, -0.00000490, -0.00005716, 0.00005769, -0.00001965, -0.00004767, 0.00005748, -0.00001022, -0.00004238, 0.00003895, -0.00000309, -0.00002207, 0.00002510, -0.00001332, -0.00001057, 0.00002813, -0.00001352, -0.00001470, 0.00001479, -0.00001259, -0.00000536, 0.00001561, -0.00000795, 0.00000125, 0.00001822, -0.00000604, -0.00001068, 0.00000273, 0.00000050, -0.00000307, -0.00000310, 0.00000676, 0.00000757, -0.00001381, -0.00000217, 0.00002520, 0.00000295, -0.00001273, 0.00000706, -0.00000293, -0.00001590, 0.00000869, 0.00001289, -0.00001183, -0.00001637, -0.00000565, 0.00000011, 0.00000048, -0.00000236, -0.00000133, -0.00000528, -0.00001686, -0.00000626, 0.00001433, -0.00000120, -0.00002304, -0.00000348, 0.00001495, -0.00000119, -0.00000544, 0.00000898, 0.00000094, },
		{ 0.00000903, 0.00000222, -0.00000198, 0.00000632, 0.00000315, -0.00000257, -0.00000134, -0.00000165, 0.00000079, 0.00000163, 0.00000063, 0.00000575, 0.00000148, -0.00000470, 0.00000338, -0.00000083, -0.00000899, 0.00000203, 0.00000135, -0.00000376, 0.00000774, 0.00000313, -0.00000528, 0.00000204, -0.00000410, -0.00000508, 0.00000615, -0.00000358, -0.00000514, 0.00000728, -0.00000407, -0.00001186, 0.00000164, 0.00000433, -0.00001237, -0.00001590, 0.00001328, 0.00000745, -0.00002408, 0.00001287, 0.00000125, -0.00005194, 0.00003577, -0.00000281, -0.00013544, 0.00007949, -0.00001589, -0.00054063, 0.00006266, 0.00108834, 0.00036419, -0.00063852, -0.00012269, 0.00020183, -0.00020624, -0.00002925, 0.00014380, -0.00012935, -0.00001285, 0.00009725, -0.00010794, -0.00000268, 0.00007746, -0.00009092, 0.00000464, 0.00007062, -0.00006804, 0.00000677, 0.00004600, -0.00006706, 0.00001156, 0.00004747, -0.00005375, 0.00000271, 0.00003623, -0.00003237, 0.00000680, 0.00001681, -0.00002884, 0.00001600, 0.00001556, -0.00002833, 0.00000664, 0.00000649, -0.00002048, 0.00001160, 0.00000772, -0.00001273, 0.00001539, 0.00000987, -0.00001580, -0.00000190, 0.00000441, -0.00000397, -0.00000282, 0.00000135, 0.00000848, -0.00000204, -0.00001327, 0.00001362, 0.00002051, -0.00001021, -0.00000533, 0.00000777, -0.00001270, -0.00000742, 0.00001706, 0.00000254, -0.00001911, -0.00001244, -0.00000078, 0.00000159, -0.00000130, -0.00000198, -0.00000180, -0.00001279, -0.00001524, 0.00000747, 0.00001146, -0.00001638, -0.00001763, 0.00000994, 0.00000930, -0.00000666, 0.00000294, 0.00000821, -0.00000776, -0.00000508, },
		{ -0.00000183, 0.00000128, 0.00000687, -0.00000044, -0.00000265, -0.00000067, -0.00000074, 0.00000149, 0.00000094, 0.00000257, 0.00000485, -0.00000238, -0.00000153, 0.00000336, -0.00000620, -0.00000462, 0.00000440, -0.00000289, 0.00000064, 0.00000859, -0.00000211, -0.00000236, 0.00000029, -0.00000727, 0.00000202, 0.00000380, -0.00000860, 0.00000235, 0.00000496, -0.00001195, -0.00000506, 0.00000582, -0.00000557, -0.00001565, -0.00000123, 0.00001372, -0.00000942, -0.00000874, 0.00001559, -0.00003139, -0.00001600, 0.00004085, -0.00008710, -0.00005237, 0.00010980, -0.00030145, -0.00039921, 0.00066257, 0.00091955, -0.00025935, -0.00050094, 0.00016117, -0.00000213, -0.00019620, 0.00012662, 0.00000960, -0.00012842, 0.00009427, -0.00000610, -0.00010050, 0.00008100, -0.00001078, -0.00007966, 0.00007802, -0.00000321, -0.00006156, 0.00006137, -0.00001698, -0.00005481, 0.00006072, -0.00000782, -0.00004842, 0.00004404, -0.00000156, -0.00002857, 0.00003071, -0.00001237, -0.00001747, 0.00003268, -0.00001305, -0.00001911, 0.00002038, -0.00001382, -0.00001034, 0.00002146, -0.00000857, -0.00000318, 0.00002263, -0.00000735, -0.00001358, 0.00000737, -0.00000131, -0.00000559, 0.00000101, 0.00000392, 0.00000492, -0.00000891, -0.00000363, 0.00002291, 0.00000716, -0.00001473, 0.00000357, 0.00000099, -0.00001596, 0.00000569, 0.00001636, -0.00001056, -0.00002049, -0.00000562, 0.00000275, 0.00000031, -0.00000233, -0.00000071, -0.00000697, -0.00001793, -0.00000423, 0.00001540, -0.00000235, -0.00002303, -0.00000332, 0.00001411, -0.00000009, -0.00000351, 0.00000919, 0.00000001, -0.00001145, 0.00000294, 0.00000918, },
		{ 0.00000511, 0.00000443, -0.00000300, -0.00000130, -0.00000017, -0.00000011, 0.00000161, 0.00000126, 0.00000375, 0.00000197, -0.00000322, 0.00000151, -0.00000038, -0.00000748, 0.00000093, 0.00000212, -0.00000388, 0.00000576, 0.00000523, -0.00000405, -0.00000030, -0.00000416, -0.00000419, 0.00000645, -0.00000366, -0.00000612, 0.00000796, -0.00000427, -0.00001196, 0.00000377, 0.00000010, -0.00001304, -0.00000767, 0.00000696, 0.00000343, -0.00001080, 0.00000532, -0.00000469, -0.00003356, 0.00002095, -0.00001303, -0.00010285, 0.00005521, -0.00004073, -0.00046372, 0.00006249, 0.00097858, 0.00038123, -0.00055713, -0.00015414, 0.00017598, -0.00017404, -0.00003877, 0.00013233, -0.00011285, -0.00002014, 0.00009281, -0.00009794, -0.00001002, 0.00007384, -0.00008525, 0.00000332, 0.00007211, -0.00006879, 0.00000449, 0.00005151, -0.00006962, 0.00000743, 0.00005283, -0.00005655, 0.00000091, 0.00004247, -0.00003686, 0.00000526, 0.00002348, -0.00003401, 0.00001370, 0.00002096, -0.00003243, 0.00000703, 0.00001132, -0.00002627, 0.00001198, 0.00001296, -0.00001769, 0.00001541, 0.00001331, -0.00002014, -0.00000021, 0.00000741, -0.00000871, -0.00000090, 0.00000359, 0.00000276, 0.00000008, -0.00000935, 0.00000903, 0.00002180, -0.00000695, -0.00001048, 0.00000774, -0.00000827, -0.00001027, 0.00001623, 0.00000658, -0.00002089, -0.00001547, 0.00000145, 0.00000295, -0.00000195, -0.00000119, -0.00000230, -0.00001484, -0.00001429, 0.00000962, 0.00001054, -0.00001718, -0.00001636, 0.00000926, 0.00000835, -0.00000436, 0.00000458, 0.00000727, -0.00000949, -0.00000722, 0.00000965, 0.00000487, -0.00000288, },
		{ 0.00000041, -0.00000313, 0.00000032, -0.00000034, 0.00000047, 0.00000182, 0.00000190, 0.00000330, -0.00000058, -0.00000169, 0.00000181, -0.00000437, -0.00000451, 0.00000357, -0.00000163, -0.00000089, 0.00000800, 0.00000076, -0.00000330, -0.00000159, -0.00000603, 0.00000235, 0.00000371, -0.00000867, 0.00000187, 0.00000511, -0.00001217, -0.00000358, 0.00000497, -0.00000972, -0.00001020, 0.00000121, 0.00000436, -0.00000265, -0.00000266, 0.00000237, -0.00002167, -0.00000924, 0.00001704, -0.00006994, -0.00003779, 0.00006361, -0.00027124, -0.00032894, 0.00058430, 0.00084575, -0.00017842, -0.00046738, 0.00010750, 0.00001037, -0.00017571, 0.00010299, 0.00001686, -0.00011887, 0.00008035, 0.00000070, -0.00009653, 0.00006988, -0.00000696, -0.00007610, 0.00007551, -0.00000093, -0.00006403, 0.00006231, -0.00001307, -0.00006023, 0.00006110, -0.00000458, -0.00005292, 0.00004726, 0.00000067, -0.00003455, 0.00003505, -0.00001017, -0.00002440, 0.00003585, -0.00001138, -0.00002352, 0.00002519, -0.00001396, -0.00001581, 0.00002663, -0.00000807, -0.00000806, 0.00002633, -0.00000802, -0.00001656, 0.00001213, -0.00000304, -0.00000870, 0.00000593, 0.00000108, 0.00000091, -0.00000274, -0.00000433, 0.00001867, 0.00001196, -0.00001556, -0.00000178, 0.00000517, -0.00001389, 0.00000154, 0.00001845, -0.00000812, -0.00002411, -0.00000608, 0.00000542, 0.00000013, -0.00000232, -0.00000004, -0.00000883, -0.00001867, -0.00000173, 0.00001574, -0.00000426, -0.00002192, -0.00000215, 0.00001196, 0.00000026, -0.00000023, 0.00000932, -0.00000190, -0.00001307, 0.00000178, 0.00001088, -0.00000025, -0.00000032, 0.00000667, },
		{ -0.00000126, 0.00000080, -0.00000071, 0.00000141, 0.00000198, 0.00000205, 0.00000194, -0.00000155, -0.00000004, -0.00000034, -0.00000569, -0.00000023, 0.00000245, -0.00000332, 0.00000360, 0.00000669, -0.00000221, -0.00000276, -0.00000429, -0.00000278, 0.00000584, -0.00000340, -0.00000610, 0.00000721, -0.00000428, -0.00001106, 0.00000449, -0.00000327, -0.00001278, -0.00000127, 0.00000185, -0.00000008, 0.00000016, -0.00000036, -0.00001006, -0.00001771, 0.00000835, -0.00002303, -0.00007255, 0.00003329, -0.00006678, -0.00038670, 0.00006931, 0.00086258, 0.00038809, -0.00046688, -0.00018053, 0.00014287, -0.00013933, -0.00004571, 0.00011622, -0.00009392, -0.00002615, 0.00008495, -0.00008522, -0.00001701, 0.00006737, -0.00007672, 0.00000173, 0.00007114, -0.00006679, 0.00000145, 0.00005513, -0.00006918, 0.00000234, 0.00005614, -0.00005685, -0.00000141, 0.00004738, -0.00003972, 0.00000288, 0.00002974, -0.00003770, 0.00001012, 0.00002606, -0.00003518, 0.00000644, 0.00001628, -0.00003115, 0.00001116, 0.00001847, -0.00002178, 0.00001433, 0.00001675, -0.00002386, 0.00000119, 0.00001065, -0.00001369, 0.00000092, 0.00000664, -0.00000386, 0.00000181, -0.00000377, 0.00000366, 0.00002181, -0.00000211, -0.00001596, 0.00000614, -0.00000207, -0.00001223, 0.00001341, 0.00001055, -0.00002137, -0.00001867, 0.00000333, 0.00000439, -0.00000263, -0.00000034, -0.00000292, -0.00001684, -0.00001275, 0.00001162, 0.00000858, -0.00001779, -0.00001365, 0.00000844, 0.00000591, -0.00000164, 0.00000723, 0.00000543, -0.00001127, -0.00000831, 0.00000931, 0.00000702, -0.00000246, 0.00000344, 0.00000488, -0.00000253, },
		{ -0.00000002, -0.00000030, 0.00000238, 0.00000161, 0.00000173, 0.00000073, -0.00000135, 0.00000026, -0.00000284, -0.00000406, 0.00000235, -0.00000037, -0.00000182, 0.00000675, 0.00000349, -0.00000369, -0.00000377, -0.00000451, 0.00000256, 0.00000298, -0.00000769, 0.00000122, 0.00000415, -0.00001131, -0.00000244, 0.00000341, -0.00001215, -0.00000569, 0.00000285, -0.00000293, 0.00000243, 0.00000262, -0.00000808, -0.00001399, -0.00000324, -0.00000366, -0.00005519, -0.00002373, 0.00002021, -0.00024360, -0.00025566, 0.00050981, 0.00076145, -0.00010023, -0.00042332, 0.00005202, 0.00001846, -0.00015046, 0.00007774, 0.00002183, -0.00010571, 0.00006451, 0.00000653, -0.00008953, 0.00005645, -0.00000332, -0.00006996, 0.00007029, 0.00000162, -0.00006426, 0.00006045, -0.00000832, -0.00006341, 0.00005855, -0.00000088, -0.00005541, 0.00004838, 0.00000334, -0.00003945, 0.00003778, -0.00000698, -0.00003080, 0.00003734, -0.00000867, -0.00002749, 0.00002884, -0.00001299, -0.00002132, 0.00003067, -0.00000645, -0.00001299, 0.00002894, -0.00000796, -0.00001933, 0.00001659, -0.00000451, -0.00001215, 0.00001120, -0.00000152, -0.00000420, 0.00000415, -0.00000409, 0.00001280, 0.00001687, -0.00001508, -0.00000862, 0.00000917, -0.00000973, -0.00000337, 0.00001895, -0.00000468, -0.00002695, -0.00000704, 0.00000796, 0.00000002, -0.00000229, 0.00000060, -0.00001080, -0.00001896, 0.00000112, 0.00001527, -0.00000674, -0.00001976, -0.00000012, 0.00000874, -0.00000017, 0.00000410, 0.00000955, -0.00000464, -0.00001385, 0.00000085, 0.00001130, 0.00000196, -0.00000109, 0.00000532, 0.00000136, -0.00000279, 0.00000068, },
		{ 0.00000110, 0.00000255, 0.00000086, 0.00000146, -0.00000004, -0.00000100, -0.00000065, -0.00000394, -0.00000121, 0.00000232, -0.00000231, 0.00000163, 0.00000738, -0.00000003, -0.00000492, -0.00000463, -0.00000115, 0.00000473, -0.00000301, -0.00000524, 0.00000549, -0.00000436, -0.00000941, 0.00000417, -0.00000585, -0.00001171, 0.00000329, -0.00000205, -0.00000280, 0.00000846, -0.00000420, -0.00001433, -0.00000508, -0.00000193, -0.00003183, -0.00004602, 0.00001399, -0.00009168, -0.00031248, 0.00008143, 0.00074531, 0.00038453, -0.00037219, -0.00019973, 0.00010425, -0.00010457, -0.00004933, 0.00009650, -0.00007400, -0.00003028, 0.00007418, -0.00007078, -0.00002308, 0.00005837, -0.00006597, 0.00000017, 0.00006768, -0.00006222, -0.00000196, 0.00005649, -0.00006577, -0.00000322, 0.00005699, -0.00005462, -0.00000392, 0.00005048, -0.00004074, -0.00000005, 0.00003502, -0.00003963, 0.00000554, 0.00003035, -0.00003632, 0.00000500, 0.00002092, -0.00003471, 0.00000921, 0.00002377, -0.00002464, 0.00001220, 0.00001984, -0.00002660, 0.00000223, 0.00001384, -0.00001848, 0.00000246, 0.00001021, -0.00001086, 0.00000293, 0.00000308, -0.00000197, 0.00002047, 0.00000388, -0.00002127, 0.00000300, 0.00000536, -0.00001300, 0.00000884, 0.00001406, -0.00002049, -0.00002179, 0.00000471, 0.00000587, -0.00000324, 0.00000051, -0.00000370, -0.00001867, -0.00001062, 0.00001330, 0.00000571, -0.00001806, -0.00000977, 0.00000748, 0.00000229, 0.00000122, 0.00001071, 0.00000302, -0.00001313, -0.00000846, 0.00000820, 0.00000851, -0.00000081, 0.00000155, 0.00000438, -0.00000120, -0.00000112, 0.00000108, -0.00000093, },
		{ 0.00000162, 0.00000041, 0.00000136, -0.00000064, -0.00000097, -0.00000181, -0.00000337, 0.00000103, 0.00000065, -0.00000209, 0.00000514, 0.00000567, -0.00000343, -0.00000590, -0.00000312, 0.00000277, 0.00000188, -0.00000608, 0.00000064, 0.00000234, -0.00000984, -0.00000151, 0.00000143, -0.00001312, -0.00000211, 0.00000367, -0.00000806, 0.00000592, 0.00000673, -0.00001543, -0.00000828, 0.00000136, -0.00002042, -0.00004296, -0.00001155, -0.00001845, -0.00021827, -0.00018320, 0.00044112, 0.00067065, -0.00002845, -0.00037061, -0.00000181, 0.00002129, -0.00012204, 0.00005271, 0.00002402, -0.00008983, 0.00004796, 0.00001085, -0.00007999, 0.00004163, -0.00000025, -0.00006153, 0.00006280, 0.00000410, -0.00006211, 0.00005601, -0.00000324, -0.00006403, 0.00005326, 0.00000286, -0.00005555, 0.00004735, 0.00000611, -0.00004278, 0.00003872, -0.00000314, -0.00003611, 0.00003697, -0.00000519, -0.00003059, 0.00003103, -0.00001106, -0.00002639, 0.00003320, -0.00000385, -0.00001754, 0.00003017, -0.00000721, -0.00002160, 0.00002038, -0.00000561, -0.00001563, 0.00001635, -0.00000368, -0.00001002, 0.00001119, -0.00000283, 0.00000580, 0.00002138, -0.00001332, -0.00001638, 0.00001255, -0.00000374, -0.00000851, 0.00001775, -0.00000056, -0.00002876, -0.00000844, 0.00001018, 0.00000006, -0.00000219, 0.00000108, -0.00001284, -0.00001871, 0.00000415, 0.00001393, -0.00000951, -0.00001663, 0.00000248, 0.00000474, -0.00000126, 0.00000902, 0.00001000, -0.00000792, -0.00001403, 0.00000019, 0.00001057, 0.00000454, -0.00000096, 0.00000298, 0.00000222, -0.00000158, 0.00000043, 0.00000011, -0.00000064, 0.00000252, },
		{ 0.00000068, 0.00000104, -0.00000124, -0.00000114, -0.00000252, -0.00000183, 0.00000181, -0.00000112, 0.00000016, 0.00000727, 0.00000213, -0.00000637, -0.00000522, 0.00000040, 0.00000350, -0.00000265, -0.00000382, 0.00000331, -0.00000468, -0.00000736, 0.00000322, -0.00000768, -0.00001004, 0.00000612, -0.00000478, -0.00000460, 0.00001402, -0.00000639, -0.00001712, 0.00000396, -0.00000996, -0.00003861, -0.00002432, -0.00000266, -0.00011333, -0.00024341, 0.00009684, 0.00063128, 0.00037109, -0.00027752, -0.00021026, 0.00006234, -0.00007210, -0.00004927, 0.00007450, -0.00005455, -0.00003216, 0.00006128, -0.00005571, -0.00002779, 0.00004739, -0.00005378, -0.00000107, 0.00006189, -0.00005548, -0.00000529, 0.00005542, -0.00005973, -0.00000872, 0.00005518, -0.00005009, -0.00000625, 0.00005144, -0.00003989, -0.00000314, 0.00003883, -0.00003971, 0.00000039, 0.00003341, -0.00003574, 0.00000295, 0.00002481, -0.00003669, 0.00000631, 0.00002836, -0.00002601, 0.00000919, 0.00002224, -0.00002810, 0.00000285, 0.00001667, -0.00002265, 0.00000359, 0.00001392, -0.00001766, 0.00000334, 0.00001065, -0.00000729, 0.00001782, 0.00001045, -0.00002592, -0.00000148, 0.00001334, -0.00001239, 0.00000295, 0.00001672, -0.00001832, -0.00002453, 0.00000546, 0.00000728, -0.00000362, 0.00000129, -0.00000469, -0.00002017, -0.00000795, 0.00001445, 0.00000210, -0.00001778, -0.00000504, 0.00000634, -0.00000210, 0.00000397, 0.00001468, 0.00000045, -0.00001505, -0.00000792, 0.00000656, 0.00000930, 0.00000183, -0.00000032, 0.00000296, 0.00000082, -0.00000076, 0.00000062, -0.00000076, 0.00000080, 0.00000281, 0.00000029, },
		{ 0.00000023, -0.00000160, -0.00000136, -0.00000261, -0.00000015, 0.00000125, -0.00000176, 0.00000349, 0.00000703, -0.00000256, -0.00000764, -0.00000220, 0.00000300, 0.00000070, -0.00000428, 0.00000029, 0.00000006, -0.00000825, -0.00000071, -0.00000064, -0.00001294, 0.00000057, 0.00000377, -0.00001111, 0.00000798, 0.00000949, -0.00001966, -0.00000434, 0.00000413, -0.00003279, -0.00003312, -0.00000231, -0.00005093, -0.00019476, -0.00011499, 0.00037954, 0.00057747, 0.00003396, -0.00031167, -0.00005077, 0.00001857, -0.00009228, 0.00002961, 0.00002325, -0.00007236, 0.00003192, 0.00001335, -0.00006863, 0.00002646, 0.00000193, -0.00005131, 0.00005368, 0.00000617, -0.00005765, 0.00004942, 0.00000167, -0.00006198, 0.00004564, 0.00000616, -0.00005323, 0.00004434, 0.00000862, -0.00004420, 0.00003789, 0.00000089, -0.00003992, 0.00003479, -0.00000133, -0.00003246, 0.00003166, -0.00000840, -0.00003057, 0.00003405, -0.00000053, -0.00002128, 0.00002985, -0.00000589, -0.00002308, 0.00002317, -0.00000628, -0.00001879, 0.00002089, -0.00000532, -0.00001604, 0.00001778, -0.00000061, -0.00000168, 0.00002503, -0.00001045, -0.00002437, 0.00001492, 0.00000356, -0.00001330, 0.00001492, 0.00000380, -0.00002930, -0.00001010, 0.00001186, 0.00000030, -0.00000194, 0.00000130, -0.00001485, -0.00001781, 0.00000716, 0.00001169, -0.00001227, -0.00001264, 0.00000529, 0.00000028, -0.00000280, 0.00001403, 0.00001071, -0.00001132, -0.00001388, -0.00000023, 0.00000897, 0.00000721, 0.00000003, -0.00000002, 0.00000284, 0.00000043, -0.00000031, -0.00000003, -0.00000049, 0.00000223, 0.00000180, -0.00000043, 0.00000126, },
		{ -0.00000159, -0.00000167, -0.00000207, 0.00000106, 0.00000000, -0.00000065, 0.00000646, 0.00000396, -0.00000688, -0.00000600, 0.00000153, 0.00000247, -0.00000235, -0.00000218, 0.00000114, -0.00000529, -0.00000530, 0.00000204, -0.00000880, -0.00000805, 0.00000749, -0.00000644, -0.00000549, 0.00001697, -0.00000718, -0.00001824, 0.00000941, -0.00001594, -0.00004285, -0.00000802, -0.00001685, -0.00013004, -0.00018117, 0.00011337, 0.00052434, 0.00034902, -0.00018710, -0.00021139, 0.00001964, -0.00004397, -0.00004558, 0.00005171, -0.00003690, -0.00003168, 0.00004718, -0.00004111, -0.00003084, 0.00003514, -0.00004105, -0.00000173, 0.00005414, -0.00004714, -0.00000815, 0.00005198, -0.00005159, -0.00001363, 0.00005076, -0.00004366, -0.00000803, 0.00005012, -0.00003731, -0.00000598, 0.00004083, -0.00003806, -0.00000487, 0.00003490, -0.00003351, 0.00000059, 0.00002760, -0.00003698, 0.00000279, 0.00003184, -0.00002580, 0.00000559, 0.00002364, -0.00002822, 0.00000313, 0.00001885, -0.00002585, 0.00000429, 0.00001737, -0.00002375, 0.00000305, 0.00001829, -0.00001181, 0.00001410, 0.00001693, -0.00002949, -0.00000689, 0.00002109, -0.00001035, -0.00000365, 0.00001819, -0.00001508, -0.00002653, 0.00000550, 0.00000849, -0.00000364, 0.00000195, -0.00000595, -0.00002118, -0.00000482, 0.00001489, -0.00000203, -0.00001678, 0.00000015, 0.00000496, -0.00000678, 0.00000643, 0.00001873, -0.00000188, -0.00001688, -0.00000705, 0.00000464, 0.00000943, 0.00000509, -0.00000193, 0.00000079, 0.00000326, -0.00000011, -0.00000055, -0.00000029, 0.00000063, 0.00000263, 0.00000034, -0.00000021, 0.00000147, -0.00000078, },
		{ -0.00000198, -0.00000099, 0.00000139, -0.00000099, 0.00000206, 0.00000740, -0.00000123, -0.00000869, -0.00000195, 0.00000321, -0.00000025, -0.00000263, 0.00000019, -0.00000230, -0.00000695, -0.00000006, -0.00000244, -0.00001196, 0.00000238, 0.00000336, -0.00001233, 0.00000879, 0.00001088, -0.00002099, -0.00000190, 0.00000490, -0.00004072, -0.00002544, 0.00000330, -0.00007642, -0.00017248, -0.00005383, 0.00032566, 0.00048587, 0.00008494, -0.00024931, -0.00009212, 0.00001059, -0.00006314, 0.00000985, 0.00001966, -0.00005451, 0.00001748, 0.00001389, -0.00005629, 0.00001193, 0.00000303, -0.00003996, 0.00004366, 0.00000758, -0.00005116, 0.00004131, 0.00000592, -0.00005740, 0.00003634, 0.00000864, -0.00004853, 0.00003971, 0.00001050, -0.00004354, 0.00003546, 0.00000466, -0.00004191, 0.00003099, 0.00000249, -0.00003287, 0.00003077, -0.00000535, -0.00003348, 0.00003317, 0.00000313, -0.00002389, 0.00002800, -0.00000419, -0.00002355, 0.00002477, -0.00000655, -0.00002131, 0.00002444, -0.00000642, -0.00002172, 0.00002340, 0.00000237, -0.00000895, 0.00002741, -0.00000681, -0.00003181, 0.00001606, 0.00001149, -0.00001716, 0.00001071, 0.00000786, -0.00002845, -0.00001180, 0.00001277, 0.00000079, -0.00000146, 0.00000113, -0.00001676, -0.00001617, 0.00000994, 0.00000855, -0.00001470, -0.00000791, 0.00000793, -0.00000432, -0.00000448, 0.00001861, 0.00001163, -0.00001436, -0.00001363, -0.00000054, 0.00000686, 0.00000969, 0.00000170, -0.00000331, 0.00000311, 0.00000304, -0.00000138, -0.00000052, 0.00000017, 0.00000160, 0.00000199, -0.00000068, 0.00000050, 0.00000105, -0.00000149, -0.00000144, },
		{ 0.00000027, 0.00000084, -0.00000074, 0.00000517, 0.00000519, -0.00000636, -0.00000681, 0.00000198, 0.00000185, -0.00000209, -0.00000066, -0.00000067, -0.00000616, -0.00000363, 0.00000098, -0.00000920, -0.00000606, 0.00000770, -0.00000712, -0.00000560, 0.00001764, -0.00000685, -0.00001773, 0.00001153, -0.00002016, -0.00004433, 0.00000280, -0.00002884, -0.00014067, -0.00012678, 0.00012897, 0.00042745, 0.00032009, -0.00010464, -0.00020318, -0.00002129, -0.00002175, -0.00003874, 0.00002966, -0.00002211, -0.00002899, 0.00003291, -0.00002794, -0.00003212, 0.00002244, -0.00002865, -0.00000170, 0.00004497, -0.00003789, -0.00001017, 0.00004644, -0.00004208, -0.00001751, 0.00004400, -0.00003592, -0.00000896, 0.00004660, -0.00003334, -0.00000821, 0.00004086, -0.00003493, -0.00000975, 0.00003466, -0.00002987, -0.00000172, 0.00002903, -0.00003564, -0.00000098, 0.00003386, -0.00002411, 0.00000175, 0.00002386, -0.00002695, 0.00000315, 0.00002013, -0.00002783, 0.00000458, 0.00002014, -0.00002868, 0.00000218, 0.00002534, -0.00001510, 0.00000964, 0.00002261, -0.00003166, -0.00001267, 0.00002784, -0.00000703, -0.00001022, 0.00001820, -0.00001110, -0.00002743, 0.00000481, 0.00000931, -0.00000315, 0.00000244, -0.00000754, -0.00002154, -0.00000131, 0.00001443, -0.00000643, -0.00001491, 0.00000541, 0.00000325, -0.00001124, 0.00000845, 0.00002236, -0.00000362, -0.00001844, -0.00000623, 0.00000267, 0.00000908, 0.00000854, -0.00000310, -0.00000185, 0.00000579, 0.00000084, -0.00000226, 0.00000034, 0.00000067, 0.00000200, 0.00000090, -0.00000101, 0.00000116, 0.00000028, -0.00000179, -0.00000149, -0.00000197, },
		{ 0.00000001, 0.00000107, 0.00000681, 0.00000035, -0.00000882, -0.00000244, 0.00000328, -0.00000072, -0.00000137, 0.00000030, -0.00000435, -0.00000624, 0.00000039, -0.00000364, -0.00001051, 0.00000336, 0.00000268, -0.00001206, 0.00000857, 0.00001103, -0.00001985, -0.00000065, 0.00000373, -0.00004451, -0.00001961, 0.00000501, -0.00009464, -0.00015083, -0.00000172, 0.00027945, 0.00039933, 0.00012341, -0.00018652, -0.00012380, -0.00000181, -0.00003649, -0.00000556, 0.00001373, -0.00003750, 0.00000546, 0.00001260, -0.00004386, -0.00000104, 0.00000298, -0.00002823, 0.00003348, 0.00000814, -0.00004310, 0.00003239, 0.00000913, -0.00005067, 0.00002610, 0.00000998, -0.00004180, 0.00003392, 0.00001148, -0.00004086, 0.00003178, 0.00000776, -0.00004199, 0.00002594, 0.00000586, -0.00003171, 0.00002856, -0.00000227, -0.00003489, 0.00003073, 0.00000674, -0.00002515, 0.00002478, -0.00000237, -0.00002288, 0.00002513, -0.00000650, -0.00002288, 0.00002671, -0.00000706, -0.00002655, 0.00002764, 0.00000581, -0.00001533, 0.00002827, -0.00000285, -0.00003795, 0.00001586, 0.00001925, -0.00001960, 0.00000551, 0.00001110, -0.00002616, -0.00001323, 0.00001273, 0.00000151, -0.00000064, 0.00000046, -0.00001845, -0.00001375, 0.00001227, 0.00000459, -0.00001648, -0.00000263, 0.00001002, -0.00000876, -0.00000600, 0.00002231, 0.00001264, -0.00001658, -0.00001347, -0.00000093, 0.00000463, 0.00001177, 0.00000377, -0.00000643, 0.00000299, 0.00000593, -0.00000253, -0.00000136, 0.00000121, 0.00000078, 0.00000205, -0.00000029, -0.00000076, 0.00000160, -0.00000058, -0.00000192, -0.00000168, -0.00000137, 0.00000112, },
		{ 0.00000363, 0.00000564, -0.00000487, -0.00000748, 0.00000159, 0.00000172, -0.00000180, 0.00000045, -0.00000187, -0.00000716, -0.00000267, 0.00000026, -0.00000888, -0.00000435, 0.00000709, -0.00000694, -0.00000515, 0.00001647, -0.00000569, -0.00001580, 0.00001083, -0.00002293, -0.00004315, 0.00000845, -0.00003893, -0.00014468, -0.00008063, 0.00014189, 0.00034255, 0.00028642, -0.00003304, -0.00018646, -0.00005804, -0.00000637, -0.00002957, 0.00000974, -0.00001087, -0.00002451, 0.00001942, -0.00001693, -0.00003170, 0.00001012, -0.00001737, -0.00000095, 0.00003501, -0.00002843, -0.00001114, 0.00003926, -0.00003198, -0.00002004, 0.00003540, -0.00002755, -0.00000886, 0.00004116, -0.00002842, -0.00000952, 0.00003895, -0.00003072, -0.00001381, 0.00003269, -0.00002519, -0.00000363, 0.00002901, -0.00003290, -0.00000459, 0.00003426, -0.00002117, -0.00000197, 0.00002281, -0.00002446, 0.00000306, 0.00002038, -0.00002847, 0.00000456, 0.00002189, -0.00003214, 0.00000092, 0.00003120, -0.00001694, 0.00000489, 0.00002686, -0.00003230, -0.00001818, 0.00003289, -0.00000273, -0.00001601, 0.00001663, -0.00000684, -0.00002694, 0.00000347, 0.00000958, -0.00000205, 0.00000273, -0.00000947, -0.00002112, 0.00000242, 0.00001297, -0.00001083, -0.00001206, 0.00001034, 0.00000114, -0.00001504, 0.00000997, 0.00002512, -0.00000451, -0.00001947, -0.00000582, 0.00000081, 0.00000848, 0.00001172, -0.00000373, -0.00000458, 0.00000807, 0.00000205, -0.00000427, 0.00000094, 0.00000098, 0.00000100, 0.00000178, -0.00000152, 0.00000014, 0.00000170, -0.00000157, -0.00000201, -0.00000166, -0.00000039, 0.00000128, -0.00000020, },
		{ 0.00000192, -0.00000794, -0.00000359, 0.00000306, -0.00000056, -0.00000062, 0.00000049, -0.00000578, -0.00000626, 0.00000052, -0.00000401, -0.00000886, 0.00000356, 0.00000200, -0.00001069, 0.00000752, 0.00001021, -0.00001683, -0.00000030, 0.00000094, -0.00004473, -0.00001532, 0.00000299, -0.00010583, -0.00012939, 0.00004020, 0.00024038, 0.00032068, 0.00014924, -0.00012619, -0.00014461, -0.00001736, -0.00001394, -0.00001613, 0.00000616, -0.00002240, -0.00000358, 0.00000977, -0.00003218, -0.00001177, 0.00000188, -0.00001689, 0.00002385, 0.00000780, -0.00003406, 0.00002340, 0.00001105, -0.00004237, 0.00001574, 0.00001003, -0.00003354, 0.00002756, 0.00001139, -0.00003639, 0.00002729, 0.00000985, -0.00004024, 0.00002011, 0.00000842, -0.00002905, 0.00002537, 0.00000050, -0.00003470, 0.00002702, 0.00000988, -0.00002499, 0.00002051, -0.00000067, -0.00002107, 0.00002430, -0.00000623, -0.00002333, 0.00002756, -0.00000737, -0.00003007, 0.00003027, 0.00000934, -0.00002023, 0.00002752, 0.00000092, -0.00004218, 0.00001443, 0.00002602, -0.00002030, -0.00000012, 0.00001300, -0.00002255, -0.00001406, 0.00001159, 0.00000241, 0.00000060, -0.00000080, -0.00001984, -0.00001051, 0.00001394, -0.00000008, -0.00001735, 0.00000299, 0.00001126, -0.00001278, -0.00000704, 0.00002476, 0.00001356, -0.00001757, -0.00001350, -0.00000161, 0.00000266, 0.00001332, 0.00000590, -0.00000898, 0.00000250, 0.00000873, -0.00000353, -0.00000251, 0.00000245, -0.00000006, 0.00000190, 0.00000062, -0.00000225, 0.00000172, 0.00000101, -0.00000259, -0.00000184, -0.00000130, 0.00000057, 0.00000102, 0.00000064, 0.00000540, },
		{ -0.00000779, 0.00000031, 0.00000202, -0.00000134, 0.00000098, -0.00000238, -0.00000810, -0.00000265, 0.00000000, -0.00000785, -0.00000315, 0.00000598, -0.00000602, -0.00000440, 0.00001397, -0.00000396, -0.00001282, 0.00000800, -0.00002452, -0.00003972, 0.00000960, -0.00004734, -0.00014215, -0.00004261, 0.00015079, 0.00027061, 0.00025027, 0.00002574, -0.00016272, -0.00008854, 0.00000191, -0.00001915, -0.00000689, -0.00000349, -0.00001881, 0.00000759, -0.00000857, -0.00002981, -0.00000102, -0.00000786, 0.00000042, 0.00002498, -0.00001946, -0.00001098, 0.00003107, -0.00002209, -0.00002107, 0.00002565, -0.00001923, -0.00000767, 0.00003426, -0.00002305, -0.00000974, 0.00003533, -0.00002591, -0.00001674, 0.00002916, -0.00001992, -0.00000486, 0.00002760, -0.00002911, -0.00000768, 0.00003303, -0.00001736, -0.00000523, 0.00002056, -0.00002101, 0.00000300, 0.00001956, -0.00002780, 0.00000433, 0.00002239, -0.00003399, -0.00000045, 0.00003537, -0.00001722, 0.00000033, 0.00002919, -0.00003145, -0.00002277, 0.00003576, 0.00000210, -0.00002033, 0.00001353, -0.00000280, -0.00002482, 0.00000163, 0.00000911, -0.00000027, 0.00000284, -0.00001175, -0.00001980, 0.00000623, 0.00001042, -0.00001491, -0.00000824, 0.00001459, -0.00000139, -0.00001779, 0.00001099, 0.00002663, -0.00000440, -0.00001974, -0.00000610, -0.00000085, 0.00000788, 0.00001424, -0.00000380, -0.00000700, 0.00000979, 0.00000339, -0.00000628, 0.00000136, 0.00000153, -0.00000022, 0.00000276, -0.00000164, -0.00000142, 0.00000320, -0.00000082, -0.00000309, -0.00000134, -0.00000075, 0.00000131, 0.00000080, 0.00000211, 0.00000527, -0.00000262, },
		{ 0.00000242, 0.00000026, -0.00000035, 0.00000063, -0.00000640, -0.00000699, 0.00000023, -0.00000343, -0.00000719, 0.00000306, 0.00000154, -0.00000864, 0.00000586, 0.00000876, -0.00001256, -0.00000060, -0.00000300, -0.00004216, -0.00001231, -0.00000220, -0.00011063, -0.00010793, 0.00007160, 0.00020757, 0.00025191, 0.00016318, -0.00007092, -0.00015422, -0.00003445, 0.00000330, -0.00002188, -0.00000215, -0.00001007, -0.00000946, 0.00000587, -0.00002196, -0.00001979, -0.00000006, -0.00000667, 0.00001536, 0.00000664, -0.00002469, 0.00001505, 0.00001162, -0.00003318, 0.00000602, 0.00000881, -0.00002441, 0.00002118, 0.00001020, -0.00003055, 0.00002245, 0.00001074, -0.00003690, 0.00001402, 0.00000993, -0.00002511, 0.00002160, 0.00000266, -0.00003300, 0.00002245, 0.00001223, -0.00002352, 0.00001559, 0.00000070, -0.00001825, 0.00002247, -0.00000586, -0.00002260, 0.00002699, -0.00000752, -0.00003199, 0.00003120, 0.00001254, -0.00002324, 0.00002526, 0.00000398, -0.00004404, 0.00001201, 0.00003105, -0.00001913, -0.00000558, 0.00001321, -0.00001786, -0.00001397, 0.00000934, 0.00000340, 0.00000231, -0.00000265, -0.00002085, -0.00000651, 0.00001476, -0.00000527, -0.00001711, 0.00000870, 0.00001142, -0.00001618, -0.00000738, 0.00002574, 0.00001419, -0.00001707, -0.00001373, -0.00000280, 0.00000124, 0.00001430, 0.00000771, -0.00001061, 0.00000177, 0.00001104, -0.00000419, -0.00000383, 0.00000366, -0.00000074, 0.00000150, 0.00000189, -0.00000369, 0.00000134, 0.00000306, -0.00000318, -0.00000253, -0.00000086, -0.00000006, 0.00000185, 0.00000076, 0.00000344, 0.00000324, -0.00000612, -0.00000519, },
		{ -0.00000064, 0.00000084, -0.00000225, -0.00000881, -0.00000362, 0.00000019, -0.00000613, -0.00000258, 0.00000465, -0.00000450, -0.00000362, 0.00001068, -0.00000189, -0.00000926, 0.00000377, -0.00002518, -0.00003465, 0.00000715, -0.00005420, -0.00013372, -0.00001221, 0.00015491, 0.00021164, 0.00021383, 0.00007077, -0.00013391, -0.00011127, 0.00000351, -0.00000869, -0.00001946, 0.00000016, -0.00001261, -0.00000196, -0.00000301, -0.00002682, -0.00001035, -0.00000054, 0.00000221, 0.00001553, -0.00001152, -0.00000973, 0.00002252, -0.00001313, -0.00002064, 0.00001551, -0.00001160, -0.00000548, 0.00002649, -0.00001771, -0.00000884, 0.00003039, -0.00002097, -0.00001835, 0.00002441, -0.00001456, -0.00000525, 0.00002499, -0.00002469, -0.00000996, 0.00003032, -0.00001313, -0.00000776, 0.00001732, -0.00001696, 0.00000307, 0.00001777, -0.00002596, 0.00000403, 0.00002155, -0.00003423, -0.00000169, 0.00003754, -0.00001607, -0.00000357, 0.00002931, -0.00002929, -0.00002585, 0.00003616, 0.00000688, -0.00002263, 0.00000911, 0.00000052, -0.00002102, -0.00000046, 0.00000777, 0.00000216, 0.00000278, -0.00001436, -0.00001754, 0.00000995, 0.00000680, -0.00001841, -0.00000349, 0.00001787, -0.00000431, -0.00001925, 0.00001153, 0.00002662, -0.00000330, -0.00001903, -0.00000723, -0.00000229, 0.00000752, 0.00001580, -0.00000338, -0.00000869, 0.00001071, 0.00000467, -0.00000795, 0.00000145, 0.00000224, -0.00000144, 0.00000361, -0.00000134, -0.00000327, 0.00000450, 0.00000044, -0.00000452, -0.00000120, -0.00000074, 0.00000099, 0.00000200, 0.00000084, 0.00000418, 0.00000006, -0.00000760, -0.00000175, 0.00000260, },
		{ 0.00000056, -0.00000617, -0.00000828, -0.00000055, -0.00000194, -0.00000562, 0.00000196, 0.00000146, -0.00000627, 0.00000375, 0.00000705, -0.00000771, -0.00000133, -0.00000750, -0.00003767, -0.00001039, -0.00000973, -0.00010995, -0.00008648, 0.00009291, 0.00017998, 0.00019410, 0.00016668, -0.00002278, -0.00015322, -0.00005137, 0.00001453, -0.00002330, -0.00001030, -0.00000103, -0.00001234, 0.00000144, -0.00001374, -0.00002494, -0.00000254, 0.00000182, 0.00000840, 0.00000484, -0.00001565, 0.00000789, 0.00001091, -0.00002386, -0.00000240, 0.00000647, -0.00001516, 0.00001528, 0.00000803, -0.00002388, 0.00001772, 0.00001040, -0.00003237, 0.00000816, 0.00001027, -0.00002026, 0.00001767, 0.00000405, -0.00003001, 0.00001749, 0.00001354, -0.00002096, 0.00001049, 0.00000157, -0.00001466, 0.00001992, -0.00000548, -0.00002074, 0.00002515, -0.00000765, -0.00003218, 0.00003055, 0.00001506, -0.00002414, 0.00002174, 0.00000590, -0.00004336, 0.00000900, 0.00003372, -0.00001620, -0.00001025, 0.00001153, -0.00001243, -0.00001269, 0.00000604, 0.00000434, 0.00000453, -0.00000507, -0.00002142, -0.00000181, 0.00001459, -0.00001076, -0.00001563, 0.00001423, 0.00001039, -0.00001878, -0.00000688, 0.00002516, 0.00001436, -0.00001496, -0.00001408, -0.00000462, 0.00000052, 0.00001472, 0.00000891, -0.00001110, 0.00000096, 0.00001251, -0.00000440, -0.00000513, 0.00000460, -0.00000112, 0.00000090, 0.00000325, -0.00000482, 0.00000049, 0.00000527, -0.00000348, -0.00000372, -0.00000023, -0.00000058, 0.00000235, 0.00000133, 0.00000127, 0.00000414, -0.00000353, -0.00000696, 0.00000130, 0.00000166, -0.00000045, },
		{ -0.00000914, -0.00000548, 0.00000069, -0.00000383, -0.00000264, 0.00000332, -0.00000252, -0.00000302, 0.00000709, 0.00000034, -0.00000562, -0.00000108, -0.00002507, -0.00002872, 0.00000210, -0.00005953, -0.00012046, 0.00001137, 0.00015403, 0.00016487, 0.00017900, 0.00010210, -0.00010226, -0.00012535, -0.00000057, 0.00000062, -0.00002762, 0.00000058, -0.00000662, -0.00000886, -0.00000014, -0.00002315, -0.00001739, 0.00000439, 0.00000415, 0.00000727, -0.00000502, -0.00000758, 0.00001432, -0.00000565, -0.00001891, 0.00000575, -0.00000516, -0.00000254, 0.00001849, -0.00001285, -0.00000695, 0.00002462, -0.00001635, -0.00001860, 0.00001888, -0.00000956, -0.00000474, 0.00002151, -0.00002008, -0.00001126, 0.00002643, -0.00000895, -0.00000937, 0.00001341, -0.00001272, 0.00000334, 0.00001522, -0.00002318, 0.00000375, 0.00001941, -0.00003302, -0.00000257, 0.00003761, -0.00001372, -0.00000644, 0.00002718, -0.00002617, -0.00002698, 0.00003409, 0.00001102, -0.00002258, 0.00000379, 0.00000268, -0.00001560, -0.00000250, 0.00000550, 0.00000513, 0.00000262, -0.00001721, -0.00001435, 0.00001339, 0.00000217, -0.00002105, 0.00000201, 0.00001994, -0.00000755, -0.00001930, 0.00001163, 0.00002501, -0.00000131, -0.00001723, -0.00000922, -0.00000357, 0.00000758, 0.00001623, -0.00000262, -0.00000935, 0.00001073, 0.00000567, -0.00000902, 0.00000116, 0.00000299, -0.00000244, 0.00000415, -0.00000065, -0.00000509, 0.00000535, 0.00000212, -0.00000601, -0.00000142, -0.00000035, 0.00000046, 0.00000314, -0.00000004, 0.00000257, 0.00000282, -0.00000682, -0.00000435, 0.00000322, 0.00000038, 0.00000011, -0.00000165, },
		{ -0.00000185, 0.00000027, -0.00000420, 0.00000042, 0.00000185, -0.00000390, 0.00000138, 0.00000540, -0.00000286, -0.00000236, -0.00001192, -0.00003213, -0.00000945, -0.00001853, -0.00010488, -0.00006533, 0.00010518, 0.00015654, 0.00014747, 0.00016171, 0.00001679, -0.00014293, -0.00006642, 0.00001963, -0.00002131, -0.00001741, 0.00000454, -0.00001266, -0.00000296, -0.00000777, -0.00002730, -0.00000519, 0.00000815, 0.00000319, 0.00000267, -0.00000754, 0.00000233, 0.00000914, -0.00001513, -0.00000904, 0.00000333, -0.00000649, 0.00001027, 0.00000513, -0.00001697, 0.00001348, 0.00000894, -0.00002714, 0.00000299, 0.00000948, -0.00001493, 0.00001400, 0.00000460, -0.00002608, 0.00001260, 0.00001370, -0.00001766, 0.00000566, 0.00000188, -0.00001063, 0.00001695, -0.00000515, -0.00001796, 0.00002227, -0.00000788, -0.00003067, 0.00002854, 0.00001659, -0.00002297, 0.00001735, 0.00000636, -0.00004021, 0.00000585, 0.00003368, -0.00001184, -0.00001355, 0.00000798, -0.00000674, -0.00001003, 0.00000191, 0.00000504, 0.00000722, -0.00000793, -0.00002149, 0.00000343, 0.00001338, -0.00001627, -0.00001287, 0.00001928, 0.00000816, -0.00002047, -0.00000551, 0.00002309, 0.00001396, -0.00001134, -0.00001441, -0.00000711, 0.00000054, 0.00001466, 0.00000927, -0.00001037, 0.00000030, 0.00001286, -0.00000413, -0.00000621, 0.00000505, -0.00000112, 0.00000019, 0.00000445, -0.00000541, -0.00000072, 0.00000729, -0.00000332, -0.00000527, 0.00000035, -0.00000082, 0.00000249, 0.00000216, -0.00000078, 0.00000450, -0.00000049, -0.00000855, -0.00000032, 0.00000341, -0.00000071, 0.00000022, -0.00000519, -0.00000989, },
		{ -0.00000111, -0.00000322, 0.00000216, -0.00000025, -0.00000279, 0.00000363, 0.00000258, -0.00000234, -0.00000590, -0.00002431, -0.00002271, -0.00000450, -0.00006325, -0.00010375, 0.00002902, 0.00014847, 0.00012895, 0.00014729, 0.00012071, -0.00007007, -0.00013063, -0.00000884, 0.00000777, -0.00003145, -0.00000146, -0.00000150, -0.00001305, 0.00000044, -0.00001926, -0.00002195, 0.00000699, 0.00000595, 0.00000061, -0.00000018, -0.00000482, 0.00000704, -0.00000003, -0.00001622, -0.00000289, -0.00000027, 0.00000082, 0.00001092, -0.00000878, -0.00000431, 0.00001855, -0.00001241, -0.00001763, 0.00001306, -0.00000530, -0.00000340, 0.00001757, -0.00001568, -0.00001154, 0.00002175, -0.00000524, -0.00001004, 0.00000921, -0.00000867, 0.00000381, 0.00001220, -0.00001978, 0.00000356, 0.00001619, -0.00003061, -0.00000291, 0.00003566, -0.00001055, -0.00000800, 0.00002300, -0.00002253, -0.00002592, 0.00002981, 0.00001394, -0.00002011, -0.00000192, 0.00000335, -0.00000885, -0.00000416, 0.00000231, 0.00000845, 0.00000245, -0.00002019, -0.00001033, 0.00001640, -0.00000329, -0.00002263, 0.00000804, 0.00002066, -0.00001093, -0.00001795, 0.00001136, 0.00002188, 0.00000136, -0.00001431, -0.00001191, -0.00000474, 0.00000813, 0.00001553, -0.00000167, -0.00000877, 0.00000986, 0.00000616, -0.00000927, 0.00000049, 0.00000360, -0.00000301, 0.00000426, 0.00000028, -0.00000655, 0.00000558, 0.00000401, -0.00000724, -0.00000210, 0.00000030, -0.00000006, 0.00000398, -0.00000044, 0.00000072, 0.00000528, -0.00000546, -0.00000717, 0.00000363, 0.00000168, -0.00000104, -0.00000082, -0.00000899, -0.00000797, 0.00000132, },
		{ -0.00000135, 0.00000269, -0.00000179, -0.00000109, 0.00000409, 0.00000147, -0.00000356, -0.00001569, -0.00002632, -0.00000942, -0.00002750, -0.00009653, -0.00004497, 0.00010989, 0.00013628, 0.00011144, 0.00015048, 0.00004707, -0.00012529, -0.00007817, 0.00001904, -0.00001707, -0.00002278, 0.00000684, -0.00001110, -0.00000683, -0.00000407, -0.00002722, -0.00000769, 0.00001212, -0.00000026, 0.00000042, -0.00000081, -0.00000146, 0.00000666, -0.00000762, -0.00001363, -0.00000023, 0.00000097, 0.00000638, 0.00000182, -0.00001039, 0.00001000, 0.00000661, -0.00002170, -0.00000118, 0.00000774, -0.00000960, 0.00001089, 0.00000437, -0.00002161, 0.00000817, 0.00001276, -0.00001403, 0.00000149, 0.00000165, -0.00000655, 0.00001390, -0.00000487, -0.00001455, 0.00001867, -0.00000828, -0.00002769, 0.00002551, 0.00001698, -0.00001996, 0.00001255, 0.00000525, -0.00003493, 0.00000303, 0.00003083, -0.00000659, -0.00001506, 0.00000282, -0.00000128, -0.00000595, -0.00000273, 0.00000534, 0.00001027, -0.00001104, -0.00002107, 0.00000901, 0.00001112, -0.00002153, -0.00000889, 0.00002356, 0.00000484, -0.00002117, -0.00000338, 0.00001971, 0.00001292, -0.00000645, -0.00001455, -0.00001019, 0.00000120, 0.00001425, 0.00000868, -0.00000849, -0.00000002, 0.00001195, -0.00000345, -0.00000685, 0.00000487, -0.00000074, -0.00000048, 0.00000525, -0.00000536, -0.00000208, 0.00000877, -0.00000262, -0.00000695, 0.00000067, -0.00000066, 0.00000229, 0.00000301, -0.00000239, 0.00000432, 0.00000262, -0.00000968, -0.00000234, 0.00000522, -0.00000093, -0.00000043, -0.00000369, -0.00001148, -0.00000360, 0.00000288, -0.00000290, },
		{ 0.00000211, -0.00000298, 0.00000062, 0.00000469, 0.00000020, -0.00001010, -0.00002299, -0.00001737, -0.00001169, -0.00006519, -0.00008512, 0.00004168, 0.00013900, 0.00010211, 0.00011970, 0.00012827, -0.00003944, -0.00012762, -0.00001956, 0.00001200, -0.00003148, -0.00000501, 0.00000225, -0.00001474, -0.00000069, -0.00001553, -0.00002404, 0.00000751, 0.00000732, -0.00000418, 0.00000298, -0.00000179, 0.00000116, 0.00000360, -0.00001297, -0.00000985, 0.00000293, 0.00000419, 0.00000431, -0.00000571, -0.00000128, 0.00001272, -0.00000935, -0.00001568, 0.00000745, -0.00000205, -0.00000142, 0.00001357, -0.00001181, -0.00001090, 0.00001672, -0.00000236, -0.00000983, 0.00000514, -0.00000516, 0.00000442, 0.00000906, -0.00001608, 0.00000346, 0.00001219, -0.00002733, -0.00000267, 0.00003199, -0.00000698, -0.00000817, 0.00001719, -0.00001880, -0.00002269, 0.00002383, 0.00001520, -0.00001546, -0.00000740, 0.00000237, -0.00000119, -0.00000508, -0.00000166, 0.00001184, 0.00000235, -0.00002311, -0.00000563, 0.00001883, -0.00000935, -0.00002300, 0.00001429, 0.00001995, -0.00001426, -0.00001535, 0.00001074, 0.00001746, 0.00000443, -0.00001042, -0.00001501, -0.00000590, 0.00000912, 0.00001380, -0.00000071, -0.00000690, 0.00000821, 0.00000593, -0.00000863, -0.00000045, 0.00000391, -0.00000302, 0.00000390, 0.00000129, -0.00000737, 0.00000512, 0.00000584, -0.00000794, -0.00000324, 0.00000107, -0.00000040, 0.00000436, -0.00000039, -0.00000104, 0.00000712, -0.00000363, -0.00000980, 0.00000362, 0.00000318, -0.00000234, -0.00000002, -0.00000835, -0.00001073, 0.00000184, 0.00000179, -0.00000409, 0.00000127, },
		{ -0.00000350, 0.00000330, 0.00000493, -0.00000486, -0.00001836, -0.00002092, -0.00001028, -0.00003560, -0.00008600, -0.00002606, 0.00010873, 0.00011845, 0.00008483, 0.00013527, 0.00006809, -0.00010263, -0.00008556, 0.00001370, -0.00001186, -0.00002597, 0.00000637, -0.00000843, -0.00000979, -0.00000242, -0.00002519, -0.00000975, 0.00001376, -0.00000212, -0.00000161, 0.00000426, -0.00000349, 0.00000383, -0.00000177, -0.00001613, -0.00000375, 0.00000673, 0.00000370, -0.00000149, -0.00000466, 0.00000740, 0.00000375, -0.00001654, -0.00000417, 0.00000531, -0.00000470, 0.00000855, 0.00000353, -0.00001706, 0.00000449, 0.00001086, -0.00001050, -0.00000172, 0.00000101, -0.00000281, 0.00001104, -0.00000460, -0.00001087, 0.00001468, -0.00000881, -0.00002358, 0.00002184, 0.00001618, -0.00001554, 0.00000780, 0.00000264, -0.00002807, 0.00000096, 0.00002539, -0.00000109, -0.00001453, -0.00000348, 0.00000344, -0.00000056, -0.00000744, 0.00000508, 0.00001351, -0.00001415, -0.00002015, 0.00001466, 0.00000792, -0.00002625, -0.00000384, 0.00002681, 0.00000065, -0.00002086, -0.00000068, 0.00001532, 0.00001131, -0.00000070, -0.00001433, -0.00001366, 0.00000227, 0.00001358, 0.00000718, -0.00000566, 0.00000018, 0.00000978, -0.00000250, -0.00000688, 0.00000401, -0.00000005, -0.00000095, 0.00000545, -0.00000464, -0.00000333, 0.00000943, -0.00000140, -0.00000849, 0.00000052, -0.00000009, 0.00000185, 0.00000366, -0.00000334, 0.00000370, 0.00000541, -0.00001010, -0.00000458, 0.00000673, -0.00000091, -0.00000132, -0.00000199, -0.00001276, -0.00000563, 0.00000547, -0.00000169, -0.00000310, 0.00000387, -0.00000011, },
		{ 0.00000657, 0.00000176, -0.00001329, -0.00002116, -0.00001331, -0.00001864, -0.00006520, -0.00006606, 0.00005021, 0.00012671, 0.00008248, 0.00009668, 0.00012697, -0.00001214, -0.00011750, -0.00003089, 0.00001294, -0.00002856, -0.00000907, 0.00000435, -0.00001441, -0.00000286, -0.00001227, -0.00002393, 0.00000636, 0.00000805, -0.00000706, 0.00000462, 0.00000117, -0.00000305, 0.00000531, -0.00000958, -0.00001475, 0.00000450, 0.00000717, -0.00000092, -0.00000366, 0.00000176, 0.00000757, -0.00000725, -0.00001310, 0.00000249, 0.00000008, 0.00000091, 0.00000991, -0.00000869, -0.00000953, 0.00001180, -0.00000049, -0.00000892, 0.00000160, -0.00000240, 0.00000504, 0.00000616, -0.00001236, 0.00000339, 0.00000781, -0.00002354, -0.00000187, 0.00002704, -0.00000344, -0.00000704, 0.00001034, -0.00001541, -0.00001755, 0.00001681, 0.00001449, -0.00000911, -0.00001205, -0.00000020, 0.00000682, -0.00000499, -0.00000616, 0.00001497, 0.00000241, -0.00002574, -0.00000052, 0.00002058, -0.00001570, -0.00002209, 0.00002043, 0.00001785, -0.00001729, -0.00001173, 0.00000978, 0.00001213, 0.00000760, -0.00000579, -0.00001814, -0.00000711, 0.00001037, 0.00001128, 0.00000012, -0.00000386, 0.00000602, 0.00000487, -0.00000712, -0.00000150, 0.00000378, -0.00000242, 0.00000313, 0.00000217, -0.00000735, 0.00000404, 0.00000731, -0.00000790, -0.00000475, 0.00000174, -0.00000038, 0.00000420, 0.00000001, -0.00000238, 0.00000812, -0.00000153, -0.00001185, 0.00000303, 0.00000470, -0.00000348, 0.00000051, -0.00000731, -0.00001331, 0.00000204, 0.00000458, -0.00000498, 0.00000073, 0.00000434, -0.00000358, -0.00000172, },
		{ -0.00000616, -0.00001961, -0.00001642, -0.00001197, -0.00004192, -0.00007426, -0.00000926, 0.00010347, 0.00010252, 0.00006599, 0.00011815, 0.00008051, -0.00007739, -0.00008798, 0.00000493, -0.00000693, -0.00002684, 0.00000387, -0.00000547, -0.00001163, -0.00000237, -0.00002185, -0.00001116, 0.00001329, -0.00000270, -0.00000318, 0.00000758, -0.00000395, 0.00000103, 0.00000219, -0.00001671, -0.00000683, 0.00001053, 0.00000216, -0.00000444, -0.00000013, 0.00000567, 0.00000077, -0.00001203, -0.00000594, 0.00000254, -0.00000062, 0.00000706, 0.00000232, -0.00001279, 0.00000175, 0.00000827, -0.00000742, -0.00000383, 0.00000013, 0.00000024, 0.00000859, -0.00000424, -0.00000728, 0.00001062, -0.00000940, -0.00001879, 0.00001790, 0.00001431, -0.00001027, 0.00000354, -0.00000121, -0.00002033, -0.00000008, 0.00001786, 0.00000395, -0.00001199, -0.00001029, 0.00000700, 0.00000588, -0.00001173, 0.00000415, 0.00001671, -0.00001693, -0.00001876, 0.00002010, 0.00000396, -0.00003017, 0.00000201, 0.00002882, -0.00000414, -0.00001953, 0.00000230, 0.00001023, 0.00000926, 0.00000540, -0.00001364, -0.00001719, 0.00000347, 0.00001273, 0.00000497, -0.00000217, 0.00000097, 0.00000651, -0.00000151, -0.00000620, 0.00000254, 0.00000081, -0.00000110, 0.00000498, -0.00000333, -0.00000423, 0.00000912, 0.00000019, -0.00000960, -0.00000022, 0.00000082, 0.00000133, 0.00000391, -0.00000350, 0.00000280, 0.00000749, -0.00000967, -0.00000680, 0.00000763, -0.00000053, -0.00000233, -0.00000043, -0.00001344, -0.00000772, 0.00000761, -0.00000055, -0.00000453, 0.00000522, 0.00000103, -0.00000566, 0.00000306, 0.00000728, },
		{ -0.00001892, -0.00001093, -0.00002470, -0.00006316, -0.00004793, 0.00005538, 0.00011286, 0.00006816, 0.00007821, 0.00011924, 0.00001056, -0.00010187, -0.00004111, 0.00001063, -0.00002377, -0.00001274, 0.00000473, -0.00001269, -0.00000537, -0.00000964, -0.00002207, 0.00000405, 0.00000801, -0.00000817, 0.00000503, 0.00000375, -0.00000553, 0.00000536, -0.00000644, -0.00001744, 0.00000465, 0.00000940, -0.00000454, -0.00000255, 0.00000442, 0.00000341, -0.00000604, -0.00001025, -0.00000151, 0.00000111, 0.00000328, 0.00000691, -0.00000637, -0.00000770, 0.00000736, 0.00000028, -0.00000756, -0.00000112, -0.00000054, 0.00000552, 0.00000379, -0.00000889, 0.00000326, 0.00000346, -0.00001955, -0.00000065, 0.00002134, -0.00000032, -0.00000483, 0.00000313, -0.00001269, -0.00001100, 0.00000948, 0.00001177, -0.00000178, -0.00001533, -0.00000410, 0.00001456, -0.00000368, -0.00001086, 0.00001748, 0.00000271, -0.00002782, 0.00000473, 0.00002157, -0.00002198, -0.00001995, 0.00002609, 0.00001446, -0.00001977, -0.00000738, 0.00000847, 0.00000634, 0.00001059, -0.00000078, -0.00002086, -0.00000836, 0.00001162, 0.00000830, 0.00000075, 0.00000010, 0.00000358, 0.00000295, -0.00000493, -0.00000248, 0.00000313, -0.00000126, 0.00000207, 0.00000275, -0.00000639, 0.00000247, 0.00000814, -0.00000701, -0.00000645, 0.00000209, 0.00000008, 0.00000351, 0.00000059, -0.00000305, 0.00000820, 0.00000058, -0.00001299, 0.00000180, 0.00000602, -0.00000418, 0.00000061, -0.00000603, -0.00001526, 0.00000176, 0.00000690, -0.00000557, 0.00000058, 0.00000621, -0.00000448, -0.00000326, 0.00000810, 0.00000224, -0.00001521, },
		{ -0.00001443, -0.00004583, -0.00006211, 0.00000479, 0.00009576, 0.00008820, 0.00005308, 0.00010089, 0.00008555, -0.00005193, -0.00008538, -0.00000571, -0.00000331, -0.00002554, 0.00000021, -0.00000293, -0.00001232, -0.00000335, -0.00001782, -0.00001186, 0.00001114, -0.00000238, -0.00000411, 0.00000925, -0.00000318, -0.00000142, 0.00000424, -0.00001568, -0.00000912, 0.00001229, 0.00000155, -0.00000668, 0.00000298, 0.00000467, -0.00000197, -0.00000842, -0.00000662, -0.00000021, 0.00000240, 0.00000635, 0.00000102, -0.00000913, -0.00000005, 0.00000531, -0.00000510, -0.00000481, -0.00000078, 0.00000236, 0.00000668, -0.00000369, -0.00000412, 0.00000676, -0.00000990, -0.00001378, 0.00001404, 0.00001160, -0.00000479, 0.00000008, -0.00000585, -0.00001246, 0.00000007, 0.00000895, 0.00000788, -0.00000768, -0.00001687, 0.00000909, 0.00001296, -0.00001511, 0.00000250, 0.00001957, -0.00001909, -0.00001692, 0.00002500, -0.00000050, -0.00003304, 0.00000834, 0.00002944, -0.00000920, -0.00001726, 0.00000523, 0.00000481, 0.00000699, 0.00001131, -0.00001240, -0.00002037, 0.00000447, 0.00001171, 0.00000235, 0.00000163, 0.00000234, 0.00000247, -0.00000071, -0.00000480, 0.00000062, 0.00000166, -0.00000085, 0.00000387, -0.00000160, -0.00000455, 0.00000777, 0.00000194, -0.00001004, -0.00000155, 0.00000191, 0.00000091, 0.00000362, -0.00000286, 0.00000185, 0.00000858, -0.00000840, -0.00000870, 0.00000765, 0.00000026, -0.00000327, 0.00000065, -0.00001334, -0.00000958, 0.00000887, 0.00000053, -0.00000535, 0.00000653, 0.00000158, -0.00000741, 0.00000335, 0.00000843, -0.00000738, -0.00001934, -0.00001314, },
		{ -0.00005906, -0.00003182, 0.00005781, 0.00009871, 0.00005751, 0.00006385, 0.00010751, 0.00002793, -0.00008266, -0.00004882, 0.00000547, -0.00001827, -0.00001529, 0.00000358, -0.00001028, -0.00000762, -0.00000769, -0.00001903, 0.00000112, 0.00000718, -0.00000783, 0.00000457, 0.00000573, -0.00000640, 0.00000414, -0.00000385, -0.00001801, 0.00000371, 0.00001063, -0.00000650, -0.00000218, 0.00000637, 0.00000042, -0.00000553, -0.00000750, -0.00000435, 0.00000120, 0.00000538, 0.00000474, -0.00000483, -0.00000573, 0.00000367, 0.00000004, -0.00000603, -0.00000282, 0.00000044, 0.00000571, 0.00000217, -0.00000583, 0.00000294, -0.00000046, -0.00001565, 0.00000078, 0.00001543, 0.00000210, -0.00000193, -0.00000372, -0.00001082, -0.00000369, 0.00000260, 0.00000719, 0.00000568, -0.00001686, -0.00000890, 0.00002137, -0.00000108, -0.00001534, 0.00001903, 0.00000327, -0.00002906, 0.00000979, 0.00002175, -0.00002780, -0.00001671, 0.00003092, 0.00000997, -0.00002149, -0.00000265, 0.00000678, 0.00000060, 0.00001318, 0.00000417, -0.00002277, -0.00000959, 0.00001253, 0.00000517, 0.00000118, 0.00000457, 0.00000118, 0.00000026, -0.00000233, -0.00000317, 0.00000198, 0.00000032, 0.00000090, 0.00000290, -0.00000453, 0.00000064, 0.00000810, -0.00000532, -0.00000810, 0.00000194, 0.00000098, 0.00000241, 0.00000118, -0.00000289, 0.00000738, 0.00000242, -0.00001299, 0.00000000, 0.00000689, -0.00000422, 0.00000019, -0.00000471, -0.00001621, 0.00000092, 0.00000842, -0.00000560, 0.00000085, 0.00000752, -0.00000554, -0.00000397, 0.00000916, 0.00000172, -0.00001579, -0.00001638, -0.00001041, -0.00002941, },
		{ 0.00001561, 0.00008699, 0.00007540, 0.00004425, 0.00008478, 0.00008477, -0.00002828, -0.00007820, -0.00001664, -0.00000173, -0.00002250, -0.00000370, -0.00000134, -0.00001198, -0.00000476, -0.00001370, -0.00001187, 0.00000783, -0.00000158, -0.00000435, 0.00000953, -0.00000162, -0.00000328, 0.00000457, -0.00001348, -0.00001041, 0.00001218, 0.00000159, -0.00000801, 0.00000463, 0.00000420, -0.00000414, -0.00000583, -0.00000642, -0.00000262, 0.00000425, 0.00000627, -0.00000010, -0.00000627, -0.00000100, 0.00000231, -0.00000368, -0.00000480, -0.00000152, 0.00000341, 0.00000531, -0.00000286, -0.00000166, 0.00000330, -0.00001015, -0.00000900, 0.00001050, 0.00000838, 0.00000032, -0.00000236, -0.00001072, -0.00000521, 0.00000133, -0.00000047, 0.00001015, -0.00000209, -0.00002247, 0.00000960, 0.00002015, -0.00001712, 0.00000020, 0.00002178, -0.00002030, -0.00001467, 0.00002903, -0.00000516, -0.00003470, 0.00001474, 0.00002861, -0.00001419, -0.00001414, 0.00000780, -0.00000058, 0.00000475, 0.00001649, -0.00001064, -0.00002279, 0.00000498, 0.00001049, -0.00000032, 0.00000538, 0.00000420, -0.00000195, -0.00000030, -0.00000280, -0.00000150, 0.00000233, -0.00000018, 0.00000223, 0.00000033, -0.00000416, 0.00000549, 0.00000357, -0.00000962, -0.00000340, 0.00000296, 0.00000073, 0.00000279, -0.00000156, 0.00000107, 0.00000853, -0.00000640, -0.00001001, 0.00000664, 0.00000139, -0.00000393, 0.00000101, -0.00001239, -0.00001091, 0.00000892, 0.00000150, -0.00000530, 0.00000761, 0.00000154, -0.00000859, 0.00000401, 0.00000840, -0.00000842, -0.00001763, -0.00001120, -0.00001749, -0.00004698, -0.00005022, },
		{ 0.00008537, 0.00004914, 0.00005286, 0.00009400, 0.00003982, -0.00006188, -0.00005303, -0.00000179, -0.00001317, -0.00001628, 0.00000128, -0.00000790, -0.00000914, -0.00000633, -0.00001544, -0.00000194, 0.00000565, -0.00000645, 0.00000365, 0.00000700, -0.00000596, 0.00000212, -0.00000201, -0.00001675, 0.00000208, 0.00001076, -0.00000696, -0.00000231, 0.00000741, -0.00000140, -0.00000549, -0.00000515, -0.00000601, 0.00000060, 0.00000695, 0.00000347, -0.00000390, -0.00000390, 0.00000088, -0.00000100, -0.00000460, -0.00000344, 0.00000065, 0.00000550, 0.00000143, -0.00000328, 0.00000234, -0.00000365, -0.00001201, 0.00000218, 0.00000983, 0.00000366, 0.00000121, -0.00000958, -0.00000984, 0.00000362, -0.00000320, 0.00000116, 0.00001236, -0.00001647, -0.00001400, 0.00002670, 0.00000273, -0.00001913, 0.00001936, 0.00000407, -0.00002920, 0.00001434, 0.00002111, -0.00003274, -0.00001257, 0.00003460, 0.00000467, -0.00002228, 0.00000212, 0.00000468, -0.00000462, 0.00001519, 0.00000862, -0.00002352, -0.00001069, 0.00001274, 0.00000222, 0.00000148, 0.00000910, -0.00000089, -0.00000300, 0.00000032, -0.00000342, 0.00000042, 0.00000208, -0.00000021, 0.00000258, -0.00000195, -0.00000116, 0.00000709, -0.00000298, -0.00000940, 0.00000117, 0.00000223, 0.00000111, 0.00000161, -0.00000186, 0.00000582, 0.00000372, -0.00001175, -0.00000219, 0.00000710, -0.00000346, -0.00000072, -0.00000357, -0.00001586, -0.00000045, 0.00000882, -0.00000485, 0.00000155, 0.00000807, -0.00000655, -0.00000391, 0.00000966, 0.00000036, -0.00001507, -0.00001369, -0.00001165, -0.00003250, -0.00005302, -0.00001850, 0.00005801, },
		{ 0.00003782, 0.00007067, 0.00007992, -0.00000803, -0.00006730, -0.00002637, -0.00000255, -0.00001837, -0.00000707, -0.00000100, -0.00001090, -0.00000601, -0.00000997, -0.00001133, 0.00000396, -0.00000069, -0.00000393, 0.00000879, 0.00000030, -0.00000444, 0.00000356, -0.00001061, -0.00001063, 0.00001054, 0.00000198, -0.00000834, 0.00000494, 0.00000399, -0.00000552, -0.00000421, -0.00000564, -0.00000444, 0.00000497, 0.00000658, -0.00000084, -0.00000427, -0.00000130, -0.00000042, -0.00000317, -0.00000403, -0.00000192, 0.00000341, 0.00000445, -0.00000168, -0.00000005, 0.00000034, -0.00001000, -0.00000483, 0.00000742, 0.00000500, 0.00000452, -0.00000375, -0.00001524, 0.00000081, 0.00000349, -0.00000947, 0.00001044, 0.00000411, -0.00002642, 0.00000858, 0.00002688, -0.00001741, -0.00000260, 0.00002304, -0.00002033, -0.00001204, 0.00003191, -0.00000966, -0.00003501, 0.00002079, 0.00002636, -0.00001876, -0.00001033, 0.00000972, -0.00000566, 0.00000282, 0.00002051, -0.00000847, -0.00002403, 0.00000479, 0.00000898, -0.00000264, 0.00000876, 0.00000631, -0.00000628, -0.00000045, -0.00000040, -0.00000349, 0.00000266, 0.00000083, 0.00000028, 0.00000223, -0.00000300, 0.00000250, 0.00000480, -0.00000828, -0.00000555, 0.00000373, 0.00000089, 0.00000148, 0.00000021, 0.00000065, 0.00000733, -0.00000389, -0.00001049, 0.00000459, 0.00000272, -0.00000414, 0.00000051, -0.00001061, -0.00001144, 0.00000755, 0.00000229, -0.00000420, 0.00000828, 0.00000096, -0.00000907, 0.00000484, 0.00000732, -0.00000913, -0.00001474, -0.00001059, -0.00002091, -0.00004534, -0.00003906, 0.00002296, 0.00007827, 0.00006418, },
		{ 0.00008053, 0.00004661, -0.00004143, -0.00005329, -0.00001016, -0.00000939, -0.00001558, -0.00000168, -0.00000612, -0.00000965, -0.00000537, -0.00001187, -0.00000471, 0.00000361, -0.00000452, 0.00000264, 0.00000754, -0.00000461, -0.00000021, -0.00000095, -0.00001411, 0.00000019, 0.00000980, -0.00000620, -0.00000266, 0.00000743, -0.00000218, -0.00000561, -0.00000338, -0.00000658, -0.00000039, 0.00000781, 0.00000304, -0.00000339, -0.00000244, -0.00000099, -0.00000255, -0.00000348, -0.00000309, 0.00000028, 0.00000480, 0.00000153, -0.00000124, 0.00000137, -0.00000589, -0.00000876, 0.00000331, 0.00000494, 0.00000431, 0.00000413, -0.00001396, -0.00000965, 0.00001016, -0.00000743, -0.00000574, 0.00001744, -0.00001421, -0.00001872, 0.00003012, 0.00000754, -0.00002179, 0.00001829, 0.00000506, -0.00002800, 0.00001809, 0.00001970, -0.00003643, -0.00000783, 0.00003687, -0.00000112, -0.00002204, 0.00000662, 0.00000219, -0.00000893, 0.00001654, 0.00001213, -0.00002287, -0.00001148, 0.00001194, -0.00000032, 0.00000179, 0.00001319, -0.00000243, -0.00000653, 0.00000264, -0.00000312, -0.00000137, 0.00000376, -0.00000107, 0.00000182, 0.00000110, -0.00000268, 0.00000510, -0.00000027, -0.00001011, -0.00000023, 0.00000365, -0.00000019, 0.00000178, -0.00000006, 0.00000378, 0.00000429, -0.00000935, -0.00000454, 0.00000647, -0.00000193, -0.00000199, -0.00000280, -0.00001408, -0.00000224, 0.00000790, -0.00000323, 0.00000271, 0.00000775, -0.00000735, -0.00000318, 0.00000949, -0.00000153, -0.00001322, -0.00001105, -0.00001438, -0.00003387, -0.00004534, -0.00000837, 0.00005641, 0.00007373, 0.00004206, 0.00004434, },
		{ 0.00000784, -0.00005387, -0.00003370, -0.00000570, -0.00001388, -0.00000928, -0.00000197, -0.00000942, -0.00000664, -0.00000694, -0.00001045, 0.00000008, -0.00000000, -0.00000298, 0.00000745, 0.00000217, -0.00000491, 0.00000167, -0.00000753, -0.00000986, 0.00000782, 0.00000242, -0.00000772, 0.00000419, 0.00000378, -0.00000603, -0.00000341, -0.00000458, -0.00000552, 0.00000472, 0.00000701, -0.00000105, -0.00000307, -0.00000120, -0.00000265, -0.00000346, -0.00000279, -0.00000189, 0.00000250, 0.00000396, -0.00000015, 0.00000067, -0.00000206, -0.00000932, -0.00000151, 0.00000487, 0.00000184, 0.00000746, -0.00000417, -0.00001885, 0.00000514, 0.00000617, -0.00001721, 0.00000864, 0.00001012, -0.00002822, 0.00000632, 0.00003259, -0.00001580, -0.00000566, 0.00002310, -0.00001901, -0.00000903, 0.00003339, -0.00001367, -0.00003395, 0.00002604, 0.00002281, -0.00002261, -0.00000604, 0.00001077, -0.00001016, 0.00000144, 0.00002303, -0.00000605, -0.00002377, 0.00000378, 0.00000705, -0.00000424, 0.00001152, 0.00000840, -0.00001005, -0.00000121, 0.00000211, -0.00000507, 0.00000257, 0.00000204, -0.00000169, 0.00000388, -0.00000115, -0.00000086, 0.00000539, -0.00000609, -0.00000774, 0.00000401, 0.00000140, -0.00000011, 0.00000220, 0.00000071, 0.00000510, -0.00000118, -0.00001000, 0.00000163, 0.00000402, -0.00000379, -0.00000086, -0.00000816, -0.00001098, 0.00000475, 0.00000278, -0.00000194, 0.00000845, -0.00000010, -0.00000878, 0.00000561, 0.00000539, -0.00000941, -0.00001104, -0.00001116, -0.00002438, -0.00004117, -0.00002899, 0.00002683, 0.00007033, 0.00005464, 0.00003244, 0.00005893, 0.00007274, },
		{ -0.00004964, -0.00001857, -0.00000754, -0.00001334, -0.00000474, -0.00000533, -0.00000910, -0.00000460, -0.00000883, -0.00000690, 0.00000131, -0.00000249, 0.00000183, 0.00000744, -0.00000282, -0.00000243, -0.00000061, -0.00001065, -0.00000160, 0.00000794, -0.00000462, -0.00000302, 0.00000649, -0.00000215, -0.00000564, -0.00000228, -0.00000629, -0.00000148, 0.00000788, 0.00000327, -0.00000305, -0.00000148, -0.00000205, -0.00000429, -0.00000279, -0.00000198, -0.00000043, 0.00000364, 0.00000236, 0.00000031, 0.00000001, -0.00000710, -0.00000591, 0.00000397, 0.00000104, 0.00000414, 0.00000640, -0.00001655, -0.00001001, 0.00001525, -0.00000986, -0.00001279, 0.00002027, -0.00001041, -0.00002240, 0.00003142, 0.00001300, -0.00002294, 0.00001580, 0.00000611, -0.00002531, 0.00002085, 0.00001757, -0.00003857, -0.00000283, 0.00003752, -0.00000704, -0.00002075, 0.00001055, -0.00000066, -0.00001203, 0.00001723, 0.00001439, -0.00002075, -0.00001176, 0.00000992, -0.00000228, 0.00000230, 0.00001642, -0.00000333, -0.00000994, 0.00000429, -0.00000225, -0.00000314, 0.00000508, -0.00000159, 0.00000076, 0.00000429, -0.00000370, 0.00000229, 0.00000248, -0.00001003, -0.00000217, 0.00000498, -0.00000124, 0.00000163, 0.00000232, 0.00000155, 0.00000402, -0.00000599, -0.00000674, 0.00000494, 0.00000024, -0.00000344, -0.00000253, -0.00001093, -0.00000427, 0.00000554, -0.00000078, 0.00000432, 0.00000654, -0.00000776, -0.00000193, 0.00000859, -0.00000363, -0.00001050, -0.00000870, -0.00001813, -0.00003356, -0.00003646, -0.00000150, 0.00005335, 0.00006434, 0.00003564, 0.00003737, 0.00006842, 0.00004907, -0.00002292, },
		{ -0.00001074, -0.00000980, -0.00001001, -0.00000404, -0.00000792, -0.00000636, -0.00000476, -0.00000948, -0.00000333, 0.00000027, -0.00000171, 0.00000595, 0.00000364, -0.00000481, -0.00000056, -0.00000464, -0.00000831, 0.00000456, 0.00000266, -0.00000634, 0.00000269, 0.00000338, -0.00000569, -0.00000317, -0.00000352, -0.00000586, 0.00000376, 0.00000731, -0.00000072, -0.00000250, -0.00000097, -0.00000425, -0.00000432, -0.00000139, -0.00000144, 0.00000093, 0.00000371, 0.00000168, 0.00000057, -0.00000392, -0.00000806, 0.00000084, 0.00000281, -0.00000083, 0.00000893, -0.00000382, -0.00002109, 0.00000757, 0.00000890, -0.00002299, 0.00000494, 0.00001515, -0.00002765, 0.00000323, 0.00003679, -0.00001227, -0.00000868, 0.00002181, -0.00001631, -0.00000566, 0.00003333, -0.00001689, -0.00003153, 0.00003008, 0.00001817, -0.00002549, -0.00000149, 0.00001085, -0.00001390, 0.00000076, 0.00002390, -0.00000359, -0.00002182, 0.00000197, 0.00000459, -0.00000485, 0.00001354, 0.00001016, -0.00001288, -0.00000257, 0.00000439, -0.00000596, 0.00000203, 0.00000324, -0.00000341, 0.00000512, 0.00000120, -0.00000423, 0.00000516, -0.00000322, -0.00000964, 0.00000362, 0.00000222, -0.00000176, 0.00000413, 0.00000130, 0.00000213, 0.00000142, -0.00000851, -0.00000196, 0.00000505, -0.00000282, -0.00000294, -0.00000525, -0.00000945, 0.00000065, 0.00000284, 0.00000143, 0.00000809, -0.00000153, -0.00000769, 0.00000609, 0.00000284, -0.00000911, -0.00000700, -0.00001268, -0.00002753, -0.00003494, -0.00002023, 0.00002750, 0.00006356, 0.00004689, 0.00002714, 0.00004951, 0.00006476, 0.00001894, -0.00003928, -0.00003783, },
		{ -0.00001002, -0.00000735, -0.00000569, -0.00000761, -0.00000379, -0.00000663, -0.00000839, -0.00000099, -0.00000074, 0.00000143, 0.00000686, -0.00000104, -0.00000421, -0.00000078, -0.00000694, -0.00000303, 0.00000547, -0.00000266, -0.00000320, 0.00000475, -0.00000162, -0.00000537, -0.00000181, -0.00000543, -0.00000241, 0.00000722, 0.00000389, -0.00000268, -0.00000109, -0.00000251, -0.00000590, -0.00000257, -0.00000041, -0.00000124, 0.00000210, 0.00000368, 0.00000146, -0.00000167, -0.00000732, -0.00000345, 0.00000404, -0.00000177, 0.00000331, 0.00000773, -0.00001726, -0.00001062, 0.00001839, -0.00001048, -0.00001927, 0.00002044, -0.00000559, -0.00002447, 0.00003062, 0.00001867, -0.00002230, 0.00001201, 0.00000710, -0.00002107, 0.00002248, 0.00001484, -0.00003892, 0.00000207, 0.00003645, -0.00001269, -0.00001848, 0.00001368, -0.00000379, -0.00001376, 0.00001730, 0.00001519, -0.00001723, -0.00001135, 0.00000658, -0.00000364, 0.00000317, 0.00001845, -0.00000358, -0.00001287, 0.00000500, -0.00000090, -0.00000461, 0.00000579, -0.00000170, -0.00000041, 0.00000725, -0.00000408, -0.00000110, 0.00000491, -0.00000904, -0.00000447, 0.00000595, -0.00000188, 0.00000120, 0.00000497, -0.00000057, 0.00000294, -0.00000203, -0.00000850, 0.00000257, 0.00000277, -0.00000484, -0.00000278, -0.00000661, -0.00000631, 0.00000182, 0.00000233, 0.00000635, 0.00000450, -0.00000769, -0.00000035, 0.00000694, -0.00000558, -0.00000718, -0.00000687, -0.00002232, -0.00003172, -0.00002696, 0.00000236, 0.00004917, 0.00005739, 0.00002958, 0.00003116, 0.00005846, 0.00004825, -0.00000755, -0.00004263, -0.00002595, -0.00000790, },
		{ -0.00000682, -0.00000673, -0.00000507, -0.00000344, -0.00000862, -0.00000593, 0.00000006, -0.00000034, 0.00000464, 0.00000449, -0.00000431, -0.00000263, -0.00000222, -0.00000628, 0.00000127, 0.00000258, -0.00000447, 0.00000082, 0.00000266, -0.00000466, -0.00000320, -0.00000267, -0.00000553, 0.00000239, 0.00000729, 0.00000008, -0.00000235, -0.00000084, -0.00000521, -0.00000545, -0.00000013, -0.00000064, -0.00000098, 0.00000354, 0.00000365, -0.00000016, -0.00000531, -0.00000625, 0.00000223, 0.00000115, -0.00000280, 0.00000893, -0.00000298, -0.00002169, 0.00000811, 0.00001119, -0.00002635, -0.00000025, 0.00001847, -0.00002474, -0.00000011, 0.00003916, -0.00000702, -0.00001128, 0.00001914, -0.00001231, -0.00000193, 0.00003172, -0.00001907, -0.00002788, 0.00003254, 0.00001271, -0.00002721, 0.00000304, 0.00000993, -0.00001677, 0.00000087, 0.00002312, -0.00000131, -0.00001816, -0.00000051, 0.00000150, -0.00000434, 0.00001478, 0.00001127, -0.00001447, -0.00000438, 0.00000611, -0.00000597, 0.00000111, 0.00000419, -0.00000462, 0.00000586, 0.00000380, -0.00000722, 0.00000406, 0.00000004, -0.00001095, 0.00000251, 0.00000319, -0.00000318, 0.00000576, 0.00000234, -0.00000122, 0.00000359, -0.00000612, -0.00000581, 0.00000558, -0.00000132, -0.00000546, -0.00000213, -0.00000689, -0.00000446, 0.00000238, 0.00000574, 0.00000727, -0.00000320, -0.00000585, 0.00000609, -0.00000005, -0.00000816, -0.00000310, -0.00001486, -0.00002995, -0.00002732, -0.00001292, 0.00002548, 0.00005806, 0.00004095, 0.00002141, 0.00004207, 0.00005721, 0.00002543, -0.00002488, -0.00003842, -0.00001696, -0.00000678, -0.00000918, },
		{ -0.00000548, -0.00000277, -0.00000542, -0.00000914, -0.00000302, 0.00000043, 0.00000150, 0.00000599, 0.00000034, -0.00000532, -0.00000122, -0.00000348, -0.00000392, 0.00000275, -0.00000073, -0.00000313, 0.00000249, -0.00000088, -0.00000470, -0.00000183, -0.00000429, -0.00000300, 0.00000596, 0.00000464, -0.00000213, -0.00000119, -0.00000261, -0.00000711, -0.00000274, 0.00000126, -0.00000196, 0.00000036, 0.00000521, 0.00000230, -0.00000357, -0.00000671, -0.00000132, 0.00000350, -0.00000351, 0.00000205, 0.00000793, -0.00001623, -0.00001114, 0.00001930, -0.00000952, -0.00002449, 0.00001785, -0.00000043, -0.00002454, 0.00002798, 0.00002408, -0.00001978, 0.00000723, 0.00000787, -0.00001533, 0.00002301, 0.00001163, -0.00003740, 0.00000650, 0.00003364, -0.00001772, -0.00001538, 0.00001585, -0.00000706, -0.00001412, 0.00001683, 0.00001447, -0.00001256, -0.00001012, 0.00000202, -0.00000444, 0.00000455, 0.00001908, -0.00000329, -0.00001495, 0.00000461, 0.00000077, -0.00000552, 0.00000575, -0.00000143, -0.00000145, 0.00000968, -0.00000376, -0.00000471, 0.00000668, -0.00000716, -0.00000686, 0.00000631, -0.00000198, 0.00000060, 0.00000755, -0.00000232, 0.00000116, 0.00000211, -0.00000958, -0.00000048, 0.00000532, -0.00000598, -0.00000349, -0.00000148, -0.00000813, -0.00000307, 0.00000583, 0.00000873, 0.00000181, -0.00000710, 0.00000138, 0.00000457, -0.00000706, -0.00000358, -0.00000578, -0.00002634, -0.00002856, -0.00001750, 0.00000367, 0.00004421, 0.00005276, 0.00002382, 0.00002509, 0.00005092, 0.00004533, 0.00000396, -0.00003319, -0.00003142, -0.00001036, -0.00000621, -0.00000910, -0.00000713, },
		{ -0.00000282, -0.00000806, -0.00000755, -0.00000060, 0.00000092, 0.00000379, 0.00000463, -0.00000365, -0.00000415, -0.00000044, -0.00000412, -0.00000158, 0.00000214, -0.00000246, -0.00000108, 0.00000157, -0.00000318, -0.00000324, -0.00000214, -0.00000471, 0.00000093, 0.00000683, 0.00000121, -0.00000237, -0.00000097, -0.00000557, -0.00000653, 0.00000078, 0.00000034, -0.00000288, 0.00000333, 0.00000560, -0.00000130, -0.00000628, -0.00000399, 0.00000280, -0.00000022, -0.00000400, 0.00000763, -0.00000194, -0.00002056, 0.00000699, 0.00001259, -0.00002712, -0.00000635, 0.00001953, -0.00001984, -0.00000313, 0.00003956, -0.00000043, -0.00001314, 0.00001519, -0.00000723, 0.00000215, 0.00002867, -0.00002007, -0.00002319, 0.00003318, 0.00000674, -0.00002766, 0.00000726, 0.00000811, -0.00001869, 0.00000173, 0.00002085, 0.00000060, -0.00001294, -0.00000341, -0.00000224, -0.00000270, 0.00001530, 0.00001151, -0.00001470, -0.00000643, 0.00000698, -0.00000503, -0.00000005, 0.00000468, -0.00000513, 0.00000611, 0.00000633, -0.00000950, 0.00000215, 0.00000330, -0.00001142, 0.00000069, 0.00000412, -0.00000412, 0.00000692, 0.00000368, -0.00000454, 0.00000511, -0.00000304, -0.00000951, 0.00000543, 0.00000055, -0.00000809, 0.00000093, -0.00000348, -0.00001012, 0.00000133, 0.00001068, 0.00000612, -0.00000495, -0.00000341, 0.00000549, -0.00000302, -0.00000651, 0.00000019, -0.00001738, -0.00003126, -0.00001905, -0.00000716, 0.00002150, 0.00005369, 0.00003673, 0.00001510, 0.00003610, 0.00005093, 0.00002797, -0.00001191, -0.00003559, -0.00002344, -0.00000528, -0.00000703, -0.00000981, -0.00000607, -0.00000291, },
		{ -0.00000925, -0.00000460, 0.00000091, 0.00000204, 0.00000504, 0.00000104, -0.00000568, -0.00000164, -0.00000070, -0.00000425, 0.00000016, 0.00000084, -0.00000283, 0.00000001, -0.00000023, -0.00000360, -0.00000212, -0.00000317, -0.00000317, 0.00000432, 0.00000524, -0.00000132, -0.00000165, -0.00000261, -0.00000777, -0.00000316, 0.00000270, -0.00000242, -0.00000138, 0.00000663, 0.00000287, -0.00000549, -0.00000550, 0.00000056, 0.00000242, -0.00000436, 0.00000059, 0.00000700, -0.00001377, -0.00001128, 0.00001798, -0.00000739, -0.00002790, 0.00001275, 0.00000429, -0.00002247, 0.00002399, 0.00002877, -0.00001546, 0.00000186, 0.00000831, -0.00000832, 0.00002256, 0.00000813, -0.00003407, 0.00001016, 0.00002917, -0.00002182, -0.00001169, 0.00001697, -0.00001028, -0.00001322, 0.00001595, 0.00001236, -0.00000707, -0.00000802, -0.00000354, -0.00000482, 0.00000650, 0.00001827, -0.00000262, -0.00001592, 0.00000312, 0.00000251, -0.00000566, 0.00000489, -0.00000090, -0.00000213, 0.00001133, -0.00000283, -0.00000816, 0.00000753, -0.00000453, -0.00000903, 0.00000589, -0.00000153, 0.00000001, 0.00000974, -0.00000350, -0.00000108, 0.00000599, -0.00000980, -0.00000394, 0.00000752, -0.00000667, -0.00000448, 0.00000402, -0.00000955, -0.00000875, 0.00000936, 0.00001134, -0.00000128, -0.00000599, 0.00000306, 0.00000160, -0.00000783, -0.00000001, -0.00000557, -0.00002960, -0.00002434, -0.00000876, 0.00000305, 0.00003881, 0.00005002, 0.00001850, 0.00001883, 0.00004560, 0.00004144, 0.00001144, -0.00002248, -0.00003434, -0.00001448, -0.00000261, -0.00000972, -0.00000935, -0.00000310, -0.00000146, -0.00000516, },
		{ -0.00000161, 0.00000190, 0.00000350, 0.00000407, -0.00000304, -0.00000483, 0.00000068, -0.00000218, -0.00000367, 0.00000139, -0.00000064, -0.00000274, 0.00000019, -0.00000153, -0.00000305, -0.00000196, -0.00000362, -0.00000036, 0.00000591, 0.00000244, -0.00000234, -0.00000144, -0.00000551, -0.00000728, 0.00000123, 0.00000130, -0.00000446, 0.00000301, 0.00000727, -0.00000256, -0.00000691, -0.00000149, 0.00000277, -0.00000141, -0.00000448, 0.00000535, -0.00000098, -0.00001784, 0.00000461, 0.00001278, -0.00002545, -0.00001263, 0.00001804, -0.00001356, -0.00000528, 0.00003809, 0.00000698, -0.00001396, 0.00001025, -0.00000140, 0.00000654, 0.00002441, -0.00001986, -0.00001773, 0.00003187, 0.00000059, -0.00002683, 0.00001090, 0.00000556, -0.00001966, 0.00000321, 0.00001739, 0.00000200, -0.00000651, -0.00000643, -0.00000656, -0.00000011, 0.00001525, 0.00001071, -0.00001359, -0.00000843, 0.00000680, -0.00000321, -0.00000124, 0.00000451, -0.00000485, 0.00000596, 0.00000848, -0.00001083, -0.00000040, 0.00000620, -0.00001090, -0.00000168, 0.00000479, -0.00000437, 0.00000754, 0.00000512, -0.00000744, 0.00000582, 0.00000039, -0.00001261, 0.00000453, 0.00000254, -0.00001042, 0.00000372, 0.00000047, -0.00001584, -0.00000031, 0.00001583, 0.00000482, -0.00000660, -0.00000056, 0.00000423, -0.00000580, -0.00000417, 0.00000253, -0.00001995, -0.00003114, -0.00001089, -0.00000296, 0.00001636, 0.00005018, 0.00003399, 0.00000843, 0.00003103, 0.00004631, 0.00002751, -0.00000131, -0.00002991, -0.00002922, -0.00000554, -0.00000403, -0.00001243, -0.00000605, -0.00000018, -0.00000268, -0.00000785, -0.00000818, },
		{ 0.00000293, 0.00000416, 0.00000094, -0.00000533, -0.00000179, 0.00000114, -0.00000408, -0.00000196, 0.00000181, -0.00000238, -0.00000236, 0.00000015, -0.00000217, -0.00000246, -0.00000228, -0.00000294, 0.00000255, 0.00000549, -0.00000028, -0.00000228, -0.00000271, -0.00000781, -0.00000365, 0.00000363, -0.00000259, -0.00000287, 0.00000769, 0.00000323, -0.00000721, -0.00000395, 0.00000222, 0.00000098, -0.00000455, -0.00000087, 0.00000511, -0.00001033, -0.00001083, 0.00001466, -0.00000463, -0.00002920, 0.00000565, 0.00000785, -0.00001840, 0.00001928, 0.00003239, -0.00000958, -0.00000356, 0.00000833, -0.00000035, 0.00002134, 0.00000452, -0.00002913, 0.00001277, 0.00002325, -0.00002473, -0.00000764, 0.00001701, -0.00001323, -0.00001129, 0.00001475, 0.00000912, -0.00000120, -0.00000510, -0.00000970, -0.00000498, 0.00000898, 0.00001616, -0.00000180, -0.00001561, 0.00000068, 0.00000406, -0.00000491, 0.00000328, -0.00000024, -0.00000224, 0.00001205, -0.00000144, -0.00001106, 0.00000732, -0.00000141, -0.00001067, 0.00000459, -0.00000060, -0.00000039, 0.00001125, -0.00000404, -0.00000347, 0.00000918, -0.00000912, -0.00000744, 0.00000900, -0.00000682, -0.00000549, 0.00000938, -0.00001041, -0.00001477, 0.00001254, 0.00001401, -0.00000441, -0.00000445, 0.00000448, -0.00000183, -0.00000771, 0.00000326, -0.00000633, -0.00003160, -0.00001934, -0.00000139, 0.00000117, 0.00003334, 0.00004857, 0.00001383, 0.00001228, 0.00004200, 0.00003756, 0.00001517, -0.00001175, -0.00003442, -0.00001954, 0.00000011, -0.00000917, -0.00001188, -0.00000088, 0.00000010, -0.00000565, -0.00000889, -0.00000561, 0.00000069, },
		{ 0.00000291, -0.00000264, -0.00000455, 0.00000121, -0.00000075, -0.00000482, 0.00000046, 0.00000070, -0.00000397, -0.00000135, 0.00000001, -0.00000252, -0.00000204, -0.00000252, -0.00000129, 0.00000461, 0.00000355, -0.00000210, -0.00000221, -0.00000519, -0.00000749, 0.00000119, 0.00000201, -0.00000548, 0.00000256, 0.00000846, -0.00000372, -0.00000722, 0.00000104, 0.00000238, -0.00000249, -0.00000440, 0.00000247, -0.00000031, -0.00001387, 0.00000148, 0.00001159, -0.00002175, -0.00001837, 0.00001400, -0.00000673, -0.00000612, 0.00003509, 0.00001461, -0.00001353, 0.00000472, 0.00000478, 0.00001119, 0.00001934, -0.00001849, -0.00001186, 0.00002861, -0.00000540, -0.00002479, 0.00001372, 0.00000253, -0.00001971, 0.00000509, 0.00001312, 0.00000284, 0.00000066, -0.00000923, -0.00001129, 0.00000312, 0.00001481, 0.00000886, -0.00001134, -0.00001011, 0.00000551, -0.00000071, -0.00000226, 0.00000361, -0.00000380, 0.00000555, 0.00000995, -0.00001109, -0.00000329, 0.00000838, -0.00000935, -0.00000435, 0.00000497, -0.00000383, 0.00000761, 0.00000639, -0.00000958, 0.00000570, 0.00000382, -0.00001476, 0.00000287, 0.00000434, -0.00001210, 0.00000610, 0.00000458, -0.00002110, -0.00000246, 0.00002069, 0.00000356, -0.00000795, 0.00000245, 0.00000232, -0.00000818, -0.00000125, 0.00000363, -0.00002227, -0.00002938, -0.00000353, -0.00000032, 0.00001093, 0.00004717, 0.00003236, 0.00000183, 0.00002631, 0.00004336, 0.00002520, 0.00000630, -0.00002225, -0.00003340, -0.00000748, -0.00000079, -0.00001420, -0.00000660, 0.00000269, -0.00000278, -0.00000798, -0.00000794, -0.00000278, 0.00000249, 0.00000379, },
		{ -0.00000441, -0.00000151, 0.00000195, -0.00000356, -0.00000336, 0.00000208, -0.00000193, -0.00000436, 0.00000015, -0.00000057, -0.00000261, -0.00000173, -0.00000240, 0.00000089, 0.00000526, 0.00000090, -0.00000288, -0.00000305, -0.00000728, -0.00000399, 0.00000388, -0.00000248, -0.00000391, 0.00000818, 0.00000337, -0.00000853, -0.00000231, 0.00000367, -0.00000060, -0.00000437, -0.00000218, 0.00000252, -0.00000645, -0.00000969, 0.00000983, -0.00000182, -0.00002831, -0.00000266, 0.00000964, -0.00001276, 0.00001459, 0.00003472, -0.00000258, -0.00000849, 0.00000793, 0.00000815, 0.00001970, 0.00000103, -0.00002297, 0.00001417, 0.00001614, -0.00002631, -0.00000354, 0.00001601, -0.00001569, -0.00000862, 0.00001332, 0.00000513, 0.00000459, -0.00000150, -0.00001597, -0.00000510, 0.00001185, 0.00001302, -0.00000107, -0.00001403, -0.00000246, 0.00000516, -0.00000326, 0.00000108, 0.00000036, -0.00000168, 0.00001183, 0.00000019, -0.00001308, 0.00000604, 0.00000188, -0.00001151, 0.00000245, 0.00000064, -0.00000041, 0.00001190, -0.00000394, -0.00000569, 0.00001137, -0.00000760, -0.00001062, 0.00000949, -0.00000643, -0.00000624, 0.00001415, -0.00001069, -0.00002057, 0.00001501, 0.00001651, -0.00000723, -0.00000257, 0.00000548, -0.00000549, -0.00000667, 0.00000598, -0.00000804, -0.00003198, -0.00001387, 0.00000410, -0.00000129, 0.00002817, 0.00004770, 0.00001001, 0.00000562, 0.00003941, 0.00003443, 0.00001583, -0.00000220, -0.00003173, -0.00002469, 0.00000141, -0.00000758, -0.00001417, 0.00000081, 0.00000179, -0.00000655, -0.00000824, -0.00000603, -0.00000009, 0.00000400, 0.00000345, 0.00000006, },
		{ 0.00000130, 0.00000000, -0.00000501, -0.00000046, 0.00000136, -0.00000472, -0.00000288, 0.00000122, -0.00000159, -0.00000226, -0.00000159, -0.00000178, 0.00000306, 0.00000433, -0.00000157, -0.00000317, -0.00000482, -0.00000706, 0.00000074, 0.00000231, -0.00000581, 0.00000205, 0.00000897, -0.00000456, -0.00000722, 0.00000334, 0.00000189, -0.00000350, -0.00000400, -0.00000058, -0.00000003, -0.00000914, -0.00000187, 0.00000904, -0.00001666, -0.00002292, 0.00000772, -0.00000025, -0.00000544, 0.00003107, 0.00002185, -0.00001177, -0.00000090, 0.00001088, 0.00001601, 0.00001394, -0.00001616, -0.00000596, 0.00002357, -0.00001095, -0.00002170, 0.00001554, -0.00000071, -0.00001890, 0.00000708, 0.00000847, 0.00000313, 0.00000800, -0.00001150, -0.00001616, 0.00000660, 0.00001415, 0.00000608, -0.00000825, -0.00001121, 0.00000316, 0.00000220, -0.00000291, 0.00000198, -0.00000212, 0.00000506, 0.00001057, -0.00001030, -0.00000622, 0.00000955, -0.00000688, -0.00000703, 0.00000452, -0.00000249, 0.00000723, 0.00000726, -0.00001074, 0.00000485, 0.00000684, -0.00001570, 0.00000060, 0.00000564, -0.00001284, 0.00000801, 0.00000842, -0.00002542, -0.00000494, 0.00002475, 0.00000253, -0.00000879, 0.00000532, -0.00000015, -0.00000995, 0.00000207, 0.00000340, -0.00002409, -0.00002590, 0.00000247, 0.00000083, 0.00000599, 0.00004433, 0.00003140, -0.00000415, 0.00002154, 0.00004174, 0.00002221, 0.00001071, -0.00001370, -0.00003532, -0.00001080, 0.00000202, -0.00001473, -0.00000753, 0.00000519, -0.00000288, -0.00000832, -0.00000709, -0.00000392, 0.00000268, 0.00000450, 0.00000135, -0.00000257, -0.00000334, },
		{ -0.00000288, -0.00000389, 0.00000168, -0.00000163, -0.00000577, -0.00000023, 0.00000101, -0.00000243, -0.00000154, -0.00000172, -0.00000047, 0.00000454, 0.00000206, -0.00000329, -0.00000368, -0.00000630, -0.00000403, 0.00000340, -0.00000218, -0.00000435, 0.00000800, 0.00000327, -0.00000926, -0.00000081, 0.00000490, -0.00000207, -0.00000411, -0.00000326, -0.00000037, -0.00000262, -0.00000792, 0.00000409, 0.00000052, -0.00002543, -0.00001130, 0.00000924, -0.00000622, 0.00001060, 0.00003572, 0.00000500, -0.00001238, 0.00000717, 0.00001668, 0.00001799, -0.00000209, -0.00001608, 0.00001426, 0.00000824, -0.00002653, 0.00000033, 0.00001408, -0.00001745, -0.00000556, 0.00001171, 0.00000086, 0.00000987, 0.00000251, -0.00002180, -0.00000535, 0.00001485, 0.00000922, -0.00000063, -0.00001132, -0.00000590, 0.00000559, -0.00000084, -0.00000144, 0.00000074, -0.00000041, 0.00001076, 0.00000180, -0.00001400, 0.00000383, 0.00000496, -0.00001140, -0.00000038, 0.00000197, 0.00000009, 0.00001161, -0.00000333, -0.00000743, 0.00001234, -0.00000542, -0.00001310, 0.00000882, -0.00000556, -0.00000643, 0.00001793, -0.00001040, -0.00002560, 0.00001646, 0.00001858, -0.00000936, -0.00000047, 0.00000590, -0.00000911, -0.00000474, 0.00000796, -0.00001057, -0.00003053, -0.00000822, 0.00000732, -0.00000375, 0.00002364, 0.00004670, 0.00000720, -0.00000080, 0.00003707, 0.00003247, 0.00001435, 0.00000522, -0.00002667, -0.00002902, 0.00000097, -0.00000527, -0.00001567, 0.00000170, 0.00000336, -0.00000748, -0.00000747, -0.00000596, -0.00000120, 0.00000506, 0.00000295, -0.00000147, -0.00000309, -0.00000070, 0.00000180, },
		{ -0.00000120, 0.00000125, -0.00000502, -0.00000419, 0.00000194, -0.00000035, -0.00000242, -0.00000099, -0.00000185, 0.00000146, 0.00000463, -0.00000079, -0.00000418, -0.00000454, -0.00000601, 0.00000004, 0.00000206, -0.00000544, 0.00000153, 0.00000869, -0.00000499, -0.00000689, 0.00000519, 0.00000150, -0.00000439, -0.00000353, -0.00000344, -0.00000016, -0.00000421, -0.00000494, 0.00000538, -0.00001094, -0.00002579, -0.00000020, 0.00000495, -0.00000327, 0.00002670, 0.00002817, -0.00000873, -0.00000602, 0.00001650, 0.00002087, 0.00000876, -0.00001313, -0.00000049, 0.00001706, -0.00001581, -0.00001781, 0.00001625, -0.00000388, -0.00001733, 0.00000889, 0.00000384, 0.00000298, 0.00001491, -0.00001299, -0.00002081, 0.00000987, 0.00001342, 0.00000261, -0.00000472, -0.00001152, -0.00000002, 0.00000515, -0.00000303, -0.00000026, -0.00000001, 0.00000467, 0.00001025, -0.00000860, -0.00000881, 0.00000957, -0.00000374, -0.00000938, 0.00000335, -0.00000047, 0.00000656, 0.00000755, -0.00001081, 0.00000344, 0.00000913, -0.00001533, -0.00000208, 0.00000617, -0.00001245, 0.00000945, 0.00001158, -0.00002841, -0.00000753, 0.00002753, 0.00000184, -0.00000892, 0.00000780, -0.00000303, -0.00001098, 0.00000558, 0.00000183, -0.00002519, -0.00002078, 0.00000670, 0.00000061, 0.00000220, 0.00004136, 0.00003063, -0.00000895, 0.00001651, 0.00004092, 0.00001955, 0.00001210, -0.00000540, -0.00003459, -0.00001496, 0.00000378, -0.00001383, -0.00000855, 0.00000686, -0.00000282, -0.00000867, -0.00000594, -0.00000487, 0.00000253, 0.00000544, -0.00000043, -0.00000284, -0.00000140, 0.00000115, 0.00000001, -0.00000437, },
		{ -0.00000162, -0.00000652, -0.00000091, 0.00000236, -0.00000182, -0.00000162, -0.00000107, -0.00000140, 0.00000340, 0.00000303, -0.00000337, -0.00000451, -0.00000508, -0.00000365, 0.00000227, -0.00000183, -0.00000411, 0.00000717, 0.00000289, -0.00000928, 0.00000042, 0.00000586, -0.00000321, -0.00000396, -0.00000412, -0.00000318, 0.00000068, -0.00000570, -0.00000184, 0.00000198, -0.00002101, -0.00001931, 0.00000653, 0.00000036, 0.00000787, 0.00003556, 0.00001259, -0.00001480, 0.00000623, 0.00002473, 0.00001660, -0.00000463, -0.00000904, 0.00001302, -0.00000002, -0.00002549, 0.00000373, 0.00001139, -0.00001834, -0.00000244, 0.00000995, -0.00000323, 0.00001428, 0.00000659, -0.00002665, -0.00000585, 0.00001767, 0.00000521, -0.00000060, -0.00000775, -0.00000925, 0.00000524, 0.00000212, -0.00000396, 0.00000078, 0.00000147, 0.00000904, 0.00000314, -0.00001374, 0.00000095, 0.00000746, -0.00001028, -0.00000363, 0.00000316, 0.00000115, 0.00001045, -0.00000237, -0.00000843, 0.00001204, -0.00000285, -0.00001460, 0.00000696, -0.00000439, -0.00000585, 0.00002044, -0.00000968, -0.00002936, 0.00001670, 0.00001994, -0.00001052, 0.00000172, 0.00000563, -0.00001244, -0.00000206, 0.00000904, -0.00001370, -0.00002722, -0.00000267, 0.00000813, -0.00000574, 0.00002001, 0.00004497, 0.00000543, -0.00000655, 0.00003431, 0.00003177, 0.00001178, 0.00000984, -0.00001992, -0.00003175, -0.00000128, -0.00000267, -0.00001590, 0.00000164, 0.00000453, -0.00000809, -0.00000671, -0.00000553, -0.00000237, 0.00000595, 0.00000260, -0.00000339, -0.00000160, 0.00000060, 0.00000090, -0.00000223, -0.00000357, 0.00000077, },
		{ -0.00000513, 0.00000211, 0.00000107, -0.00000238, -0.00000076, -0.00000161, -0.00000001, 0.00000439, 0.00000014, -0.00000502, -0.00000444, -0.00000449, -0.00000072, 0.00000125, -0.00000450, 0.00000110, 0.00000760, -0.00000498, -0.00000623, 0.00000644, 0.00000133, -0.00000512, -0.00000325, -0.00000579, -0.00000063, 0.00000033, -0.00000737, 0.00000098, -0.00000539, -0.00002673, -0.00000896, 0.00000811, 0.00000010, 0.00002266, 0.00003320, -0.00000461, -0.00001011, 0.00002130, 0.00002564, 0.00000434, -0.00000973, 0.00000413, 0.00000949, -0.00001977, -0.00001342, 0.00001583, -0.00000670, -0.00001511, 0.00001023, -0.00000037, 0.00000255, 0.00002083, -0.00001356, -0.00002486, 0.00001252, 0.00001270, -0.00000122, -0.00000120, -0.00001097, -0.00000371, 0.00000774, -0.00000255, -0.00000288, 0.00000224, 0.00000451, 0.00000903, -0.00000625, -0.00001078, 0.00000841, -0.00000025, -0.00001111, 0.00000151, 0.00000199, 0.00000579, 0.00000715, -0.00000984, 0.00000172, 0.00001042, -0.00001368, -0.00000487, 0.00000575, -0.00001092, 0.00001052, 0.00001371, -0.00002985, -0.00000994, 0.00002868, 0.00000156, -0.00000821, 0.00000965, -0.00000617, -0.00001117, 0.00000899, -0.00000092, -0.00002540, -0.00001426, 0.00000894, -0.00000082, 0.00000001, 0.00003806, 0.00002956, -0.00001213, 0.00001118, 0.00004023, 0.00001798, 0.00001096, 0.00000159, -0.00003115, -0.00001929, 0.00000403, -0.00001151, -0.00000933, 0.00000733, -0.00000250, -0.00000879, -0.00000477, -0.00000550, 0.00000216, 0.00000634, -0.00000221, -0.00000336, 0.00000099, 0.00000093, -0.00000066, -0.00000312, -0.00000164, 0.00000041, -0.00000498, },
		{ 0.00000331, -0.00000080, -0.00000183, -0.00000060, -0.00000188, 0.00000200, 0.00000363, -0.00000310, -0.00000542, -0.00000383, -0.00000286, 0.00000072, -0.00000157, -0.00000326, 0.00000579, 0.00000219, -0.00000857, 0.00000132, 0.00000648, -0.00000383, -0.00000405, -0.00000481, -0.00000556, 0.00000315, -0.00000331, -0.00000728, 0.00000232, -0.00001567, -0.00002586, 0.00000168, 0.00000605, 0.00000674, 0.00003459, 0.00001964, -0.00001544, 0.00000531, 0.00003184, 0.00001589, -0.00000637, -0.00000252, 0.00001054, -0.00000817, -0.00002337, 0.00000643, 0.00000817, -0.00001826, 0.00000045, 0.00000806, -0.00000669, 0.00001756, 0.00001038, -0.00003007, -0.00000662, 0.00001995, 0.00000138, -0.00000105, -0.00000370, -0.00001209, 0.00000408, 0.00000526, -0.00000616, 0.00000042, 0.00000378, 0.00000695, 0.00000401, -0.00001233, -0.00000228, 0.00000910, -0.00000823, -0.00000695, 0.00000400, 0.00000276, 0.00000858, -0.00000129, -0.00000855, 0.00001059, -0.00000021, -0.00001494, 0.00000406, -0.00000310, -0.00000433, 0.00002157, -0.00000869, -0.00003148, 0.00001570, 0.00002034, -0.00001048, 0.00000391, 0.00000461, -0.00001520, 0.00000116, 0.00000916, -0.00001712, -0.00002219, 0.00000247, 0.00000656, -0.00000694, 0.00001748, 0.00004203, 0.00000466, -0.00001115, 0.00003062, 0.00003217, 0.00000913, 0.00001139, -0.00001234, -0.00003227, -0.00000513, -0.00000026, -0.00001460, 0.00000066, 0.00000502, -0.00000806, -0.00000603, -0.00000495, -0.00000333, 0.00000653, 0.00000230, -0.00000540, -0.00000013, 0.00000226, -0.00000047, -0.00000175, -0.00000252, -0.00000046, -0.00000199, -0.00000662, -0.00000173, },
		{ -0.00000198, -0.00000088, -0.00000122, -0.00000119, 0.00000363, 0.00000106, -0.00000553, -0.00000454, -0.00000272, -0.00000134, -0.00000002, -0.00000319, 0.00000085, 0.00000581, -0.00000461, -0.00000521, 0.00000701, 0.00000143, -0.00000560, -0.00000333, -0.00000748, -0.00000128, 0.00000398, -0.00000891, -0.00000366, -0.00000073, -0.00002575, -0.00001763, 0.00000868, 0.00000410, 0.00001959, 0.00003679, 0.00000031, -0.00001268, 0.00002508, 0.00003016, 0.00000118, -0.00000630, 0.00000746, 0.00000138, -0.00002272, -0.00000889, 0.00001437, -0.00000896, -0.00001240, 0.00001088, -0.00000390, 0.00000205, 0.00002528, -0.00001317, -0.00002792, 0.00001417, 0.00001203, -0.00000500, 0.00000192, -0.00000957, -0.00000751, 0.00000965, -0.00000150, -0.00000557, 0.00000434, 0.00000465, 0.00000707, -0.00000357, -0.00001188, 0.00000620, 0.00000320, -0.00001199, -0.00000088, 0.00000462, 0.00000511, 0.00000604, -0.00000800, -0.00000001, 0.00001059, -0.00001097, -0.00000746, 0.00000430, -0.00000835, 0.00001133, 0.00001460, -0.00002966, -0.00001191, 0.00002794, 0.00000167, -0.00000654, 0.00001070, -0.00000935, -0.00001048, 0.00001201, -0.00000460, -0.00002459, -0.00000671, 0.00000915, -0.00000322, -0.00000033, 0.00003433, 0.00002780, -0.00001340, 0.00000569, 0.00003903, 0.00001796, 0.00000801, 0.00000646, -0.00002530, -0.00002311, 0.00000252, -0.00000799, -0.00000953, 0.00000638, -0.00000196, -0.00000845, -0.00000385, -0.00000574, 0.00000175, 0.00000693, -0.00000385, -0.00000401, 0.00000351, 0.00000080, -0.00000181, -0.00000157, -0.00000168, -0.00000099, -0.00000476, -0.00000559, 0.00000178, 0.00000246, },
		},
		{ { -0.00000082, 0.00000111, -0.00000638, -0.00001160, 0.00000037, 0.00000315, -0.00000207, 0.00000178, 0.00000326, -0.00000212, -0.00000189, -0.00000410, -0.00000889, -0.00000014, -0.00000506, -0.00001204, 0.00000419, -0.00001150, -0.00002581, -0.00000368, -0.00000004, 0.00002818, 0.00004160, -0.00000635, 0.00001750, 0.00004529, -0.00000195, 0.00000659, 0.00000740, -0.00003711, -0.00001103, 0.00001536, -0.00000636, -0.00000809, 0.00000327, 0.00000382, -0.00000526, -0.00000445, 0.00000392, -0.00000161, -0.00000433, -0.00000163, 0.00000244, 0.00000270, -0.00000460, -0.00000212, -0.00001191, -0.00000844, 0.00002505, 0.00000872, -0.00000936, 0.00000731, -0.00001370, -0.00000805, 0.00002043, -0.00000014, 0.00000230, 0.00000232, -0.00002369, 0.00000539, 0.00001460, -0.00001504, 0.00000726, 0.00000920, -0.00001633, 0.00000920, 0.00000726, -0.00001504, 0.00001460, 0.00000539, -0.00002369, 0.00000232, 0.00000230, -0.00000014, 0.00002043, -0.00000805, -0.00001370, 0.00000731, -0.00000936, 0.00000872, 0.00002505, -0.00000844, -0.00001191, -0.00000212, -0.00000460, 0.00000270, 0.00000244, -0.00000163, -0.00000433, -0.00000161, 0.00000392, -0.00000445, -0.00000526, 0.00000382, 0.00000327, -0.00000809, -0.00000636, 0.00001536, -0.00001103, -0.00003711, 0.00000740, 0.00000659, -0.00000195, 0.00004529, 0.00001750, -0.00000635, 0.00004160, 0.00002818, -0.00000004, -0.00000368, -0.00002581, -0.00001150, 0.00000419, -0.00001204, -0.00000506, -0.00000014, -0.00000889, -0.00000410, -0.00000189, -0.00000212, 0.00000326, 0.00000178, -0.00000207, 0.00000315, 0.00000037, -0.00001160, -0.00000638, 0.00000111, },
		{ -0.00000006, -0.00001017, -0.00000827, 0.00000371, 0.00000068, -0.00000180, 0.00000339, 0.00000148, -0.00000255, -0.00000188, -0.00000687, -0.00000658, 0.00000088, -0.00001003, -0.00000744, 0.00000413, -0.00002041, -0.00002245, -0.00000275, 0.00000759, 0.00004259, 0.00002651, -0.00000723, 0.00003801, 0.00003193, -0.00000679, 0.00001481, -0.00001288, -0.00003736, 0.00000717, 0.00000811, -0.00001225, -0.00000161, 0.00000425, 0.00000069, -0.00000560, -0.00000202, 0.00000294, -0.00000347, -0.00000404, 0.00000013, 0.00000346, -0.00000036, -0.00000408, -0.00000474, -0.00001420, 0.00000584, 0.00002437, -0.00000370, -0.00000230, 0.00000326, -0.00001825, 0.00000592, 0.00001539, -0.00000431, 0.00000864, -0.00000881, -0.00002026, 0.00001618, 0.00000305, -0.00001166, 0.00001505, -0.00000350, -0.00001129, 0.00001696, -0.00000540, -0.00000905, 0.00002055, -0.00000877, -0.00001802, 0.00000756, -0.00000421, 0.00001018, 0.00001583, -0.00001899, -0.00000312, 0.00000332, -0.00000992, 0.00002304, 0.00001447, -0.00001773, -0.00000531, -0.00000305, -0.00000299, 0.00000486, -0.00000015, -0.00000210, -0.00000448, 0.00000094, 0.00000380, -0.00000781, -0.00000256, 0.00000655, -0.00000187, -0.00001066, 0.00000430, 0.00001163, -0.00002809, -0.00002257, 0.00001783, -0.00000488, 0.00001340, 0.00004601, -0.00000337, 0.00000778, 0.00004622, 0.00001068, -0.00000012, -0.00000708, -0.00002637, -0.00000241, -0.00000012, -0.00001299, -0.00000040, -0.00000374, -0.00000865, -0.00000162, -0.00000288, -0.00000055, 0.00000407, -0.00000016, -0.00000068, 0.00000403, -0.00000448, -0.00001203, -0.00000233, 0.00000078, -0.00000047, },
		{ -0.00001202, -0.00000322, 0.00000437, -0.00000179, -0.00000004, 0.00000378, -0.00000037, -0.00000230, -0.00000322, -0.00000834, -0.00000321, -0.00000125, -0.00001271, -0.00000156, -0.00000062, -0.00002692, -0.00001864, -0.00000078, 0.00002179, 0.00004708, 0.00000947, 0.00000553, 0.00004853, 0.00001432, -0.00000158, 0.00001158, -0.00003247, -0.00002320, 0.00001659, -0.00000425, -0.00001026, 0.00000357, 0.00000223, -0.00000082, -0.00000529, -0.00000050, 0.00000156, -0.00000504, -0.00000274, 0.00000211, 0.00000239, -0.00000224, -0.00000389, -0.00000916, -0.00000911, 0.00001756, 0.00001477, -0.00000831, 0.00000413, -0.00000566, -0.00001364, 0.00001522, 0.00000577, -0.00000081, 0.00000880, -0.00001966, -0.00000901, 0.00001862, -0.00000719, -0.00000170, 0.00001361, -0.00001343, 0.00000077, 0.00001503, -0.00001541, 0.00000335, 0.00001705, -0.00001924, -0.00000735, 0.00000507, -0.00000463, 0.00002009, 0.00000246, -0.00002063, 0.00000616, -0.00000564, -0.00000037, 0.00003041, -0.00000289, -0.00001761, 0.00000005, -0.00000562, 0.00000116, 0.00000410, -0.00000221, -0.00000182, -0.00000461, 0.00000448, 0.00000114, -0.00001034, 0.00000291, 0.00000561, -0.00000851, -0.00000612, 0.00001247, -0.00000203, -0.00003438, -0.00000212, 0.00001466, -0.00000945, 0.00003202, 0.00003212, -0.00001409, 0.00002746, 0.00003748, -0.00000187, 0.00000368, -0.00001364, -0.00002229, 0.00000369, -0.00000640, -0.00000986, 0.00000136, -0.00000770, -0.00000590, -0.00000086, -0.00000364, 0.00000178, 0.00000348, -0.00000134, 0.00000148, 0.00000248, -0.00000897, -0.00000970, 0.00000052, -0.00000014, 0.00000020, -0.00000297, },
		{ 0.00000157, 0.00000250, -0.00000280, 0.00000222, 0.00000282, -0.00000159, -0.00000222, -0.00000532, -0.00000781, -0.00000075, -0.00000535, -0.00001189, 0.00000280, -0.00000859, -0.00003054, -0.00001509, 0.00000591, 0.00003719, 0.00004010, -0.00000021, 0.00002524, 0.00004559, 0.00000144, 0.00000627, -0.00000368, -0.00004065, -0.00000282, 0.00001326, -0.00001336, -0.00000294, 0.00000434, 0.00000023, -0.00000115, -0.00000533, 0.00000078, -0.00000043, -0.00000583, -0.00000030, 0.00000276, 0.00000049, -0.00000274, -0.00000588, -0.00001121, 0.00000145, 0.00002106, 0.00000316, -0.00000533, 0.00000471, -0.00001284, -0.00000287, 0.00001583, -0.00000132, 0.00000630, 0.00000120, -0.00002400, 0.00000434, 0.00001285, -0.00001060, 0.00000823, 0.00000433, -0.00001475, 0.00001257, 0.00000450, -0.00001688, 0.00001529, 0.00000574, -0.00002136, 0.00000166, -0.00000150, 0.00000297, 0.00002227, -0.00001296, -0.00001252, 0.00000772, -0.00001184, 0.00001596, 0.00002487, -0.00001814, -0.00000971, 0.00000038, -0.00000613, 0.00000548, 0.00000052, -0.00000203, -0.00000236, -0.00000349, 0.00000786, -0.00000463, -0.00000918, 0.00000866, -0.00000053, -0.00001145, 0.00000352, 0.00001178, -0.00001793, -0.00002678, 0.00001312, 0.00000295, -0.00000189, 0.00004342, 0.00000988, -0.00000941, 0.00004195, 0.00001970, -0.00000492, 0.00000580, -0.00002085, -0.00001404, 0.00000468, -0.00001118, -0.00000455, -0.00000063, -0.00000952, -0.00000228, -0.00000205, -0.00000304, 0.00000384, 0.00000186, -0.00000111, 0.00000302, -0.00000113, -0.00001144, -0.00000583, 0.00000159, -0.00000077, 0.00000027, -0.00000681, -0.00001090, },
		{ -0.00000054, -0.00000178, 0.00000374, 0.00000105, -0.00000205, -0.00000292, -0.00000707, -0.00000575, -0.00000046, -0.00000933, -0.00000816, 0.00000374, -0.00001783, -0.00003186, -0.00001030, 0.00001827, 0.00004710, 0.00002618, 0.00000235, 0.00004233, 0.00003285, -0.00000274, 0.00000792, -0.00002342, -0.00003327, 0.00001259, 0.00000112, -0.00001399, 0.00000400, 0.00000158, 0.00000024, -0.00000198, -0.00000520, 0.00000180, -0.00000312, -0.00000490, 0.00000214, 0.00000165, -0.00000062, -0.00000351, -0.00000922, -0.00000776, 0.00001183, 0.00001607, -0.00000394, 0.00000021, -0.00000049, -0.00001332, 0.00000742, 0.00000964, -0.00000165, 0.00001060, -0.00001043, -0.00001950, 0.00001340, 0.00000362, -0.00000642, 0.00001203, -0.00000668, -0.00000707, 0.00001723, -0.00000850, -0.00000905, 0.00002035, -0.00000745, -0.00001563, 0.00000429, -0.00000561, 0.00001454, 0.00001374, -0.00002207, -0.00000054, 0.00000090, -0.00000837, 0.00003000, 0.00000776, -0.00002357, -0.00000039, -0.00000358, -0.00000248, 0.00000658, -0.00000324, -0.00000005, -0.00000443, 0.00000039, 0.00000835, -0.00001154, -0.00000279, 0.00001023, -0.00000929, -0.00000691, 0.00001171, 0.00000163, -0.00002676, -0.00001051, 0.00001637, -0.00000728, 0.00001469, 0.00004067, -0.00000963, 0.00000757, 0.00004360, 0.00000161, 0.00000032, 0.00000254, -0.00002479, -0.00000461, 0.00000090, -0.00001199, -0.00000026, -0.00000493, -0.00000800, -0.00000002, -0.00000405, -0.00000079, 0.00000462, 0.00000021, 0.00000025, 0.00000281, -0.00000543, -0.00001126, -0.00000199, 0.00000122, -0.00000081, -0.00000102, -0.00001006, -0.00000706, 0.00000422, },
		{ 0.00000069, 0.00000376, -0.00000067, -0.00000217, -0.00000434, -0.00000762, -0.00000342, -0.00000234, -0.00001144, -0.00000327, 0.00000071, -0.00002691, -0.00003112, -0.00000178, 0.00003346, 0.00004750, 0.00001325, 0.00001548, 0.00004986, 0.00001768, -0.00000109, -0.00000069, -0.00003689, -0.00001516, 0.00001575, -0.00001117, -0.00000706, 0.00000619, -0.00000129, 0.00000163, -0.00000413, -0.00000396, 0.00000169, -0.00000558, -0.00000208, 0.00000295, -0.00000001, -0.00000074, -0.00000611, -0.00001073, 0.00000058, 0.00001658, 0.00000721, -0.00000448, 0.00000283, -0.00000693, -0.00000714, 0.00001179, 0.00000251, 0.00000363, 0.00000795, -0.00001981, -0.00000895, 0.00001509, -0.00000313, 0.00000125, 0.00000798, -0.00001266, 0.00000462, 0.00001237, -0.00001674, 0.00000356, 0.00001636, -0.00001617, -0.00000696, 0.00000081, -0.00000256, 0.00002254, -0.00000182, -0.00002040, 0.00000718, -0.00000836, 0.00000555, 0.00003252, -0.00001255, -0.00001759, 0.00000392, -0.00000736, 0.00000368, 0.00000313, -0.00000406, 0.00000122, -0.00000649, 0.00000669, 0.00000352, -0.00001527, 0.00000665, 0.00000454, -0.00001463, 0.00000372, 0.00001182, -0.00001188, -0.00002389, 0.00000466, 0.00000906, -0.00000799, 0.00003107, 0.00002477, -0.00001696, 0.00002735, 0.00003183, -0.00000875, 0.00000821, -0.00000578, -0.00002300, 0.00000222, -0.00000483, -0.00000889, 0.00000055, -0.00000859, -0.00000409, -0.00000041, -0.00000506, 0.00000222, 0.00000388, -0.00000050, 0.00000166, 0.00000068, -0.00000885, -0.00000888, 0.00000066, 0.00000023, -0.00000062, -0.00000368, -0.00001121, -0.00000200, 0.00000392, -0.00000290, },
		{ 0.00000229, -0.00000175, -0.00000252, -0.00000593, -0.00000686, -0.00000204, -0.00000533, -0.00001102, 0.00000083, -0.00000587, -0.00003495, -0.00002730, 0.00001166, 0.00004623, 0.00003928, 0.00000810, 0.00003261, 0.00004644, 0.00000597, -0.00000019, -0.00001611, -0.00003703, 0.00000330, 0.00000726, -0.00001619, 0.00000189, 0.00000340, -0.00000145, 0.00000213, -0.00000665, -0.00000158, -0.00000022, -0.00000624, 0.00000128, 0.00000163, -0.00000062, -0.00000140, -0.00000994, -0.00000748, 0.00000952, 0.00001414, 0.00000036, -0.00000131, 0.00000067, -0.00000955, 0.00000120, 0.00000945, -0.00000025, 0.00000940, -0.00000091, -0.00002232, 0.00000207, 0.00001105, -0.00000427, 0.00000667, -0.00000053, -0.00001047, 0.00001334, 0.00000122, -0.00001622, 0.00001422, 0.00000620, -0.00001758, -0.00000103, -0.00000400, 0.00000703, 0.00002106, -0.00001630, -0.00001029, 0.00000598, -0.00001153, 0.00002309, 0.00002034, -0.00002568, -0.00000553, 0.00000121, -0.00000658, 0.00000799, -0.00000265, -0.00000110, -0.00000057, -0.00000550, 0.00001220, -0.00000607, -0.00001185, 0.00001346, -0.00000675, -0.00001117, 0.00001369, 0.00000244, -0.00002005, -0.00001241, 0.00001113, -0.00000102, 0.00000185, 0.00003809, 0.00000390, -0.00000958, 0.00003965, 0.00001291, -0.00000833, 0.00001248, -0.00001541, -0.00001613, 0.00000418, -0.00000906, -0.00000428, -0.00000220, -0.00000916, -0.00000029, -0.00000294, -0.00000389, 0.00000458, 0.00000231, -0.00000008, 0.00000202, -0.00000259, -0.00001038, -0.00000543, 0.00000177, -0.00000068, -0.00000091, -0.00000687, -0.00000957, 0.00000222, 0.00000129, -0.00000319, 0.00000328, },
		{ -0.00000206, -0.00000344, -0.00000698, -0.00000537, -0.00000214, -0.00000806, -0.00000847, 0.00000262, -0.00001527, -0.00004095, -0.00001861, 0.00002845, 0.00005197, 0.00002755, 0.00001307, 0.00004637, 0.00003544, -0.00000107, -0.00000501, -0.00002996, -0.00002468, 0.00001275, -0.00000568, -0.00001214, 0.00000705, -0.00000099, 0.00000109, 0.00000009, -0.00000772, 0.00000058, -0.00000326, -0.00000427, 0.00000310, -0.00000058, 0.00000016, -0.00000427, -0.00001208, 0.00000044, 0.00001408, 0.00000757, -0.00000155, 0.00000093, -0.00000372, -0.00000668, 0.00000638, 0.00000421, 0.00000260, 0.00001060, -0.00001133, -0.00001769, 0.00000901, 0.00000571, -0.00000109, 0.00000644, -0.00000763, -0.00000218, 0.00001448, -0.00000987, -0.00000808, 0.00001817, -0.00000456, -0.00001345, -0.00000036, -0.00000460, 0.00001754, 0.00000980, -0.00002240, 0.00000089, -0.00000195, -0.00000366, 0.00003372, -0.00000115, -0.00002570, 0.00000423, -0.00000493, -0.00000065, 0.00000686, -0.00000633, 0.00000309, -0.00000507, 0.00000043, 0.00001231, -0.00001581, -0.00000129, 0.00001184, -0.00001685, 0.00000107, 0.00001514, -0.00001078, -0.00001789, -0.00000058, 0.00000817, -0.00000565, 0.00001625, 0.00003229, -0.00001173, 0.00000735, 0.00003890, -0.00000414, 0.00000003, 0.00000985, -0.00002184, -0.00000735, 0.00000177, -0.00000980, -0.00000128, -0.00000628, -0.00000625, 0.00000102, -0.00000562, -0.00000076, 0.00000530, 0.00000098, 0.00000080, 0.00000086, -0.00000574, -0.00000990, -0.00000209, 0.00000158, -0.00000126, -0.00000214, -0.00000927, -0.00000570, 0.00000398, -0.00000195, -0.00000117, 0.00000458, 0.00000008, },
		{ -0.00000486, -0.00000708, -0.00000402, -0.00000346, -0.00000956, -0.00000484, 0.00000115, -0.00002658, -0.00004320, -0.00000401, 0.00004461, 0.00004928, 0.00001870, 0.00002566, 0.00005212, 0.00002160, -0.00000618, -0.00001521, -0.00003462, -0.00000753, 0.00001027, -0.00001462, -0.00000306, 0.00000624, -0.00000305, 0.00000381, -0.00000397, -0.00000638, 0.00000094, -0.00000551, -0.00000064, 0.00000225, -0.00000162, 0.00000072, -0.00000911, -0.00000943, 0.00000934, 0.00001221, 0.00000192, 0.00000006, -0.00000025, -0.00000611, -0.00000094, 0.00000618, 0.00000112, 0.00000791, 0.00000562, -0.00001821, -0.00000933, 0.00001095, 0.00000275, 0.00000236, 0.00000154, -0.00000908, 0.00000673, 0.00000822, -0.00001545, 0.00000282, 0.00001489, -0.00001142, -0.00000821, -0.00000299, 0.00000124, 0.00002211, -0.00000559, -0.00001825, 0.00000597, -0.00000907, 0.00001258, 0.00002996, -0.00002105, -0.00001466, 0.00000605, -0.00000843, 0.00000650, 0.00000061, -0.00000477, 0.00000437, -0.00000867, 0.00000949, 0.00000456, -0.00001925, 0.00001096, 0.00000071, -0.00001778, 0.00001517, 0.00000521, -0.00001864, -0.00000789, 0.00000442, 0.00000151, -0.00000173, 0.00002689, 0.00001743, -0.00001570, 0.00002459, 0.00002652, -0.00001256, 0.00001012, 0.00000132, -0.00002255, -0.00000029, -0.00000245, -0.00000760, -0.00000156, -0.00000864, -0.00000181, -0.00000083, -0.00000637, 0.00000288, 0.00000447, 0.00000056, 0.00000111, -0.00000137, -0.00000783, -0.00000796, 0.00000040, 0.00000062, -0.00000169, -0.00000419, -0.00000975, -0.00000114, 0.00000283, -0.00000383, 0.00000215, 0.00000382, -0.00000180, -0.00000201, },
		{ -0.00000632, -0.00000346, -0.00000524, -0.00000946, -0.00000142, -0.00000411, -0.00003839, -0.00003947, 0.00001551, 0.00005561, 0.00004058, 0.00001718, 0.00004050, 0.00004893, 0.00000789, -0.00001256, -0.00002568, -0.00002819, 0.00000514, 0.00000009, -0.00001506, 0.00000480, 0.00000169, -0.00000135, 0.00000394, -0.00000778, -0.00000351, -0.00000084, -0.00000526, 0.00000233, -0.00000038, -0.00000054, -0.00000118, -0.00001307, -0.00000146, 0.00001430, 0.00000617, 0.00000039, 0.00000135, -0.00000352, -0.00000424, 0.00000318, 0.00000268, 0.00000267, 0.00001094, -0.00000301, -0.00001926, -0.00000145, 0.00001013, 0.00000262, 0.00000265, -0.00000393, -0.00000475, 0.00001123, -0.00000128, -0.00001382, 0.00001129, 0.00000747, -0.00001334, -0.00000537, -0.00000447, 0.00001078, 0.00001730, -0.00001730, -0.00000822, 0.00000295, -0.00000796, 0.00002809, 0.00001242, -0.00002964, -0.00000075, 0.00000060, -0.00000557, 0.00000941, -0.00000632, 0.00000111, 0.00000055, -0.00000700, 0.00001622, -0.00000848, -0.00001263, 0.00001668, -0.00001405, -0.00000641, 0.00002109, -0.00001074, -0.00001515, 0.00000176, 0.00000160, -0.00000176, 0.00000743, 0.00002865, 0.00000134, -0.00000819, 0.00003416, 0.00000903, -0.00001069, 0.00001610, -0.00000925, -0.00001799, 0.00000307, -0.00000577, -0.00000479, -0.00000433, -0.00000757, 0.00000128, -0.00000441, -0.00000433, 0.00000540, 0.00000302, 0.00000080, 0.00000024, -0.00000368, -0.00000856, -0.00000531, 0.00000171, -0.00000063, -0.00000226, -0.00000640, -0.00000796, 0.00000223, -0.00000026, -0.00000309, 0.00000499, 0.00000129, -0.00000259, -0.00000224, -0.00000631, },
		{ -0.00000383, -0.00000666, -0.00000802, 0.00000057, -0.00001326, -0.00004844, -0.00002804, 0.00003667, 0.00005867, 0.00003074, 0.00002393, 0.00005214, 0.00003834, -0.00000508, -0.00002062, -0.00003041, -0.00001519, 0.00000813, -0.00001007, -0.00000845, 0.00000733, -0.00000234, 0.00000239, 0.00000072, -0.00000923, -0.00000114, -0.00000333, -0.00000250, 0.00000266, -0.00000244, 0.00000138, -0.00000616, -0.00001242, 0.00000860, 0.00001264, 0.00000073, 0.00000220, -0.00000030, -0.00000529, 0.00000032, 0.00000299, 0.00000018, 0.00000731, 0.00000893, -0.00001095, -0.00001573, 0.00000389, 0.00000935, 0.00000298, -0.00000043, -0.00000600, 0.00000183, 0.00000964, -0.00000901, -0.00000717, 0.00001471, -0.00000019, -0.00001233, -0.00000525, -0.00000148, 0.00001826, 0.00000528, -0.00002033, 0.00000048, -0.00000379, 0.00000318, 0.00003300, -0.00001024, -0.00002416, 0.00000740, -0.00000625, 0.00000205, 0.00000543, -0.00000858, 0.00000666, -0.00000624, 0.00000129, 0.00001493, -0.00001961, 0.00000143, 0.00001053, -0.00002232, 0.00001184, 0.00001277, -0.00002185, -0.00000252, 0.00000347, -0.00000355, 0.00000133, 0.00001551, 0.00002220, -0.00000914, 0.00000540, 0.00003293, -0.00000607, -0.00000165, 0.00001516, -0.00001766, -0.00001068, 0.00000294, -0.00000699, -0.00000356, -0.00000709, -0.00000377, 0.00000112, -0.00000718, -0.00000036, 0.00000602, 0.00000198, 0.00000081, -0.00000152, -0.00000530, -0.00000819, -0.00000258, 0.00000184, -0.00000187, -0.00000315, -0.00000789, -0.00000451, 0.00000309, -0.00000334, 0.00000015, 0.00000568, -0.00000176, -0.00000218, -0.00000326, -0.00000720, -0.00000521, },
		{ -0.00000723, -0.00000591, 0.00000000, -0.00002569, -0.00005377, -0.00000885, 0.00005496, 0.00005397, 0.00002480, 0.00003656, 0.00005670, 0.00002251, -0.00001743, -0.00002749, -0.00002690, -0.00000311, 0.00000264, -0.00001426, -0.00000005, 0.00000474, -0.00000290, 0.00000495, -0.00000421, -0.00000797, -0.00000069, -0.00000440, 0.00000085, 0.00000035, -0.00000207, 0.00000141, -0.00001189, -0.00000550, 0.00001547, 0.00000596, -0.00000046, 0.00000377, -0.00000409, -0.00000311, 0.00000365, -0.00000041, 0.00000164, 0.00001107, 0.00000277, -0.00001523, -0.00001055, 0.00000734, 0.00000937, 0.00000122, -0.00000410, -0.00000366, 0.00000639, 0.00000399, -0.00001212, 0.00000072, 0.00001355, -0.00000587, -0.00001104, -0.00000533, 0.00000561, 0.00001892, -0.00000779, -0.00001530, 0.00000299, -0.00000691, 0.00001879, 0.00002316, -0.00002665, -0.00001002, 0.00000630, -0.00000818, 0.00000875, -0.00000293, -0.00000408, 0.00000696, -0.00001054, 0.00001240, 0.00000417, -0.00002137, 0.00001441, -0.00000512, -0.00001647, 0.00002523, -0.00000639, -0.00001926, 0.00000971, -0.00000366, -0.00000396, 0.00000780, 0.00001828, 0.00001219, -0.00001147, 0.00001840, 0.00002286, -0.00001378, 0.00000924, 0.00000801, -0.00002137, -0.00000338, 0.00000083, -0.00000675, -0.00000445, -0.00000752, 0.00000027, -0.00000202, -0.00000722, 0.00000369, 0.00000516, 0.00000159, -0.00000014, -0.00000329, -0.00000603, -0.00000708, -0.00000020, 0.00000093, -0.00000287, -0.00000431, -0.00000793, -0.00000079, 0.00000129, -0.00000437, 0.00000431, 0.00000356, -0.00000377, -0.00000131, -0.00000507, -0.00000709, -0.00000447, -0.00000474, },
		{ -0.00000396, -0.00000397, -0.00003985, -0.00005145, 0.00001595, 0.00006630, 0.00004455, 0.00002608, 0.00005067, 0.00005220, 0.00000349, -0.00002803, -0.00002946, -0.00001765, 0.00000247, -0.00000570, -0.00001144, 0.00000521, 0.00000043, -0.00000017, 0.00000420, -0.00000827, -0.00000545, -0.00000180, -0.00000304, 0.00000233, -0.00000250, 0.00000036, -0.00000222, -0.00001421, 0.00000542, 0.00001531, -0.00000089, 0.00000242, 0.00000202, -0.00000654, 0.00000202, 0.00000276, -0.00000317, 0.00000649, 0.00001080, -0.00000428, -0.00001576, -0.00000565, 0.00001057, 0.00000859, -0.00000292, -0.00000506, 0.00000077, 0.00000683, -0.00000212, -0.00001082, 0.00000704, 0.00000981, -0.00000968, -0.00001040, -0.00000282, 0.00001303, 0.00001211, -0.00001590, -0.00000727, -0.00000004, -0.00000168, 0.00002938, 0.00000291, -0.00002956, 0.00000330, -0.00000073, -0.00000326, 0.00000918, -0.00000960, 0.00000414, 0.00000089, -0.00000754, 0.00001915, -0.00001110, -0.00001154, 0.00001710, -0.00002038, 0.00000221, 0.00002321, -0.00002434, -0.00000314, 0.00001139, -0.00001279, 0.00000290, 0.00001194, 0.00001618, 0.00000327, -0.00000721, 0.00002619, 0.00000869, -0.00001286, 0.00001715, -0.00000251, -0.00001990, 0.00000204, -0.00000186, -0.00000638, -0.00000616, -0.00000510, 0.00000197, -0.00000602, -0.00000426, 0.00000622, 0.00000387, 0.00000130, -0.00000200, -0.00000424, -0.00000624, -0.00000545, 0.00000140, -0.00000069, -0.00000352, -0.00000544, -0.00000635, 0.00000161, -0.00000192, -0.00000230, 0.00000708, -0.00000052, -0.00000372, -0.00000116, -0.00000703, -0.00000605, -0.00000459, -0.00000547, -0.00000698, },
		{ -0.00001183, -0.00005319, -0.00003947, 0.00004225, 0.00006852, 0.00003519, 0.00003517, 0.00006115, 0.00003828, -0.00001616, -0.00003428, -0.00002516, -0.00000813, 0.00000087, -0.00001112, -0.00000486, 0.00000563, -0.00000205, 0.00000331, 0.00000042, -0.00000985, -0.00000350, -0.00000280, -0.00000026, 0.00000093, -0.00000338, 0.00000228, -0.00000844, -0.00001003, 0.00001497, 0.00000854, -0.00000332, 0.00000578, -0.00000285, -0.00000463, 0.00000625, -0.00000188, -0.00000195, 0.00001130, 0.00000639, -0.00000930, -0.00001434, -0.00000084, 0.00001386, 0.00000491, -0.00000696, -0.00000231, 0.00000384, 0.00000414, -0.00000616, -0.00000707, 0.00001101, 0.00000499, -0.00001261, -0.00000921, 0.00000280, 0.00001640, 0.00000142, -0.00001680, -0.00000170, -0.00000348, 0.00001040, 0.00002771, -0.00001753, -0.00001984, 0.00000860, -0.00000671, 0.00000482, 0.00000252, -0.00000944, 0.00000991, -0.00000748, 0.00000279, 0.00001583, -0.00002191, 0.00000427, 0.00000642, -0.00002375, 0.00002278, 0.00000454, -0.00002794, 0.00001609, -0.00000029, -0.00001446, 0.00001267, 0.00001053, 0.00001280, -0.00000272, 0.00000051, 0.00002718, -0.00000470, -0.00000509, 0.00001914, -0.00001292, -0.00001424, 0.00000478, -0.00000449, -0.00000664, -0.00000683, -0.00000130, 0.00000037, -0.00000834, 0.00000032, 0.00000675, 0.00000292, 0.00000027, -0.00000391, -0.00000421, -0.00000629, -0.00000339, 0.00000187, -0.00000249, -0.00000387, -0.00000614, -0.00000367, 0.00000165, -0.00000442, 0.00000223, 0.00000657, -0.00000448, -0.00000187, -0.00000262, -0.00000812, -0.00000479, -0.00000553, -0.00000535, -0.00000646, -0.00000292, },
		{ -0.00006238, -0.00001756, 0.00006495, 0.00006223, 0.00003084, 0.00004974, 0.00006314, 0.00001646, -0.00003273, -0.00003383, -0.00001689, -0.00000295, -0.00000436, -0.00001117, 0.00000101, 0.00000297, -0.00000138, 0.00000482, -0.00000437, -0.00000921, -0.00000278, -0.00000220, 0.00000168, -0.00000204, -0.00000150, 0.00000095, -0.00001337, 0.00000005, 0.00001800, -0.00000030, -0.00000027, 0.00000562, -0.00000736, 0.00000142, 0.00000587, -0.00000658, 0.00000349, 0.00001261, 0.00000038, -0.00001165, -0.00001242, 0.00000509, 0.00001532, -0.00000176, -0.00000767, 0.00000208, 0.00000372, 0.00000090, -0.00000785, -0.00000252, 0.00001297, -0.00000067, -0.00001480, -0.00000571, 0.00000925, 0.00001379, -0.00000798, -0.00001268, -0.00000060, -0.00000197, 0.00002234, 0.00001357, -0.00002838, -0.00000511, 0.00000515, -0.00000647, 0.00000967, -0.00000667, -0.00000215, 0.00000858, -0.00001153, 0.00001475, 0.00000279, -0.00002125, 0.00001564, -0.00001130, -0.00001052, 0.00003095, -0.00002021, -0.00001246, 0.00002425, -0.00001777, -0.00000428, 0.00001751, 0.00000573, 0.00001082, -0.00000642, 0.00000923, 0.00002183, -0.00001359, 0.00000613, 0.00001441, -0.00002011, -0.00000622, 0.00000459, -0.00000693, -0.00000711, -0.00000545, 0.00000154, -0.00000348, -0.00000753, 0.00000457, 0.00000589, 0.00000227, -0.00000177, -0.00000485, -0.00000375, -0.00000623, -0.00000117, 0.00000106, -0.00000388, -0.00000405, -0.00000601, -0.00000101, -0.00000050, -0.00000421, 0.00000694, 0.00000254, -0.00000612, 0.00000020, -0.00000553, -0.00000757, -0.00000436, -0.00000647, -0.00000435, -0.00000640, -0.00000333, -0.00002353, },
		{ 0.00001223, 0.00007921, 0.00005092, 0.00003508, 0.00006471, 0.00005325, -0.00000932, -0.00004208, -0.00002689, -0.00000909, -0.00000313, -0.00000862, -0.00000729, 0.00000360, 0.00000030, 0.00000121, 0.00000337, -0.00000812, -0.00000766, -0.00000249, -0.00000015, 0.00000113, -0.00000402, 0.00000141, -0.00000411, -0.00001285, 0.00001147, 0.00001304, -0.00000558, 0.00000510, 0.00000085, -0.00000746, 0.00000787, 0.00000035, -0.00000721, 0.00001008, 0.00000948, -0.00000435, -0.00001269, -0.00000956, 0.00001217, 0.00001243, -0.00000861, -0.00000390, 0.00000467, 0.00000138, -0.00000113, -0.00000826, 0.00000244, 0.00001293, -0.00000733, -0.00001491, 0.00000036, 0.00001311, 0.00000679, -0.00001275, -0.00000775, -0.00000165, 0.00000585, 0.00002630, -0.00000609, -0.00002597, 0.00000571, -0.00000187, -0.00000027, 0.00000722, -0.00001174, 0.00000730, 0.00000065, -0.00000700, 0.00002040, -0.00001297, -0.00000934, 0.00001433, -0.00002360, 0.00001265, 0.00001898, -0.00003446, 0.00001371, 0.00001346, -0.00002772, 0.00001284, 0.00001264, 0.00000314, 0.00000943, -0.00000840, 0.00001742, 0.00001153, -0.00001561, 0.00001667, 0.00000414, -0.00002181, 0.00000182, 0.00000163, -0.00000882, -0.00000689, -0.00000255, 0.00000176, -0.00000729, -0.00000386, 0.00000709, 0.00000459, 0.00000129, -0.00000422, -0.00000433, -0.00000362, -0.00000574, 0.00000068, -0.00000074, -0.00000448, -0.00000420, -0.00000492, 0.00000036, -0.00000339, -0.00000064, 0.00000901, -0.00000322, -0.00000450, 0.00000047, -0.00000849, -0.00000563, -0.00000543, -0.00000621, -0.00000333, -0.00000719, -0.00000559, -0.00003840, -0.00006373, },
		{ 0.00008200, 0.00004044, 0.00004839, 0.00007319, 0.00003113, -0.00003292, -0.00004172, -0.00001683, -0.00000535, -0.00000622, -0.00000942, -0.00000269, 0.00000323, -0.00000045, 0.00000343, -0.00000020, -0.00001003, -0.00000619, -0.00000141, 0.00000167, -0.00000154, -0.00000325, 0.00000232, -0.00001016, -0.00000552, 0.00001828, 0.00000341, -0.00000443, 0.00000792, -0.00000561, -0.00000193, 0.00001002, -0.00000699, -0.00000235, 0.00001381, 0.00000396, -0.00000685, -0.00001375, -0.00000415, 0.00001808, 0.00000458, -0.00001166, 0.00000222, 0.00000349, -0.00000058, -0.00000221, -0.00000798, 0.00000797, 0.00000996, -0.00001406, -0.00001140, 0.00000700, 0.00001241, -0.00000096, -0.00001297, -0.00000481, -0.00000064, 0.00001604, 0.00001888, -0.00002159, -0.00001418, 0.00000799, -0.00000585, 0.00000678, -0.00000120, -0.00000878, 0.00001226, -0.00000824, 0.00000445, 0.00001512, -0.00002209, 0.00000588, 0.00000069, -0.00002026, 0.00003083, -0.00000769, -0.00002669, 0.00003340, -0.00001121, -0.00002062, 0.00002548, 0.00000093, 0.00000656, 0.00000542, -0.00000740, 0.00002318, -0.00000145, -0.00000993, 0.00002237, -0.00000857, -0.00001723, 0.00000722, -0.00000325, -0.00000949, -0.00000547, 0.00000035, -0.00000072, -0.00000895, 0.00000113, 0.00000755, 0.00000342, -0.00000059, -0.00000596, -0.00000278, -0.00000420, -0.00000453, 0.00000155, -0.00000284, -0.00000430, -0.00000427, -0.00000324, -0.00000025, -0.00000478, 0.00000490, 0.00000667, -0.00000762, -0.00000077, -0.00000214, -0.00000966, -0.00000375, -0.00000750, -0.00000402, -0.00000378, -0.00000836, -0.00001046, -0.00005478, -0.00005382, 0.00004535, },
		{ 0.00003733, 0.00006693, 0.00006861, 0.00000076, -0.00004779, -0.00003276, -0.00000845, -0.00000613, -0.00000873, -0.00000723, 0.00000024, 0.00000187, 0.00000063, 0.00000375, -0.00000427, -0.00001047, -0.00000468, 0.00000068, 0.00000150, -0.00000416, -0.00000039, -0.00000057, -0.00001284, 0.00000575, 0.00001699, -0.00000493, 0.00000142, 0.00000534, -0.00000895, 0.00000633, 0.00000568, -0.00001121, 0.00000568, 0.00001265, -0.00000087, -0.00000836, -0.00001432, 0.00000444, 0.00001925, -0.00000550, -0.00000870, 0.00000658, -0.00000037, -0.00000041, -0.00000386, -0.00000617, 0.00001328, 0.00000320, -0.00001845, -0.00000429, 0.00001117, 0.00000794, -0.00000640, -0.00001107, -0.00000345, 0.00000472, 0.00002207, 0.00000327, -0.00002629, -0.00000115, 0.00000347, -0.00000365, 0.00000887, -0.00000973, 0.00000048, 0.00000919, -0.00001133, 0.00001591, 0.00000119, -0.00001926, 0.00001385, -0.00001574, -0.00000122, 0.00003044, -0.00003248, 0.00000069, 0.00003182, -0.00003401, 0.00000161, 0.00002396, -0.00000858, 0.00001390, -0.00000291, -0.00000129, 0.00002359, -0.00001330, 0.00000199, 0.00002014, -0.00001918, -0.00000778, 0.00000790, -0.00000825, -0.00000849, -0.00000320, 0.00000179, -0.00000460, -0.00000750, 0.00000560, 0.00000650, 0.00000231, -0.00000327, -0.00000603, -0.00000124, -0.00000525, -0.00000267, 0.00000111, -0.00000447, -0.00000366, -0.00000410, -0.00000183, -0.00000235, -0.00000298, 0.00000944, 0.00000053, -0.00000811, 0.00000227, -0.00000670, -0.00000805, -0.00000373, -0.00000880, -0.00000066, -0.00000675, -0.00000891, -0.00001933, -0.00006940, -0.00003069, 0.00007515, 0.00007373, },
		{ 0.00008261, 0.00004777, -0.00002982, -0.00005000, -0.00001993, -0.00000508, -0.00000897, -0.00000870, -0.00000426, 0.00000118, 0.00000121, 0.00000216, 0.00000208, -0.00000773, -0.00001001, -0.00000243, 0.00000259, -0.00000092, -0.00000454, 0.00000197, -0.00000606, -0.00000921, 0.00001528, 0.00000882, -0.00000732, 0.00000705, -0.00000131, -0.00000606, 0.00001206, -0.00000307, -0.00000919, 0.00001246, 0.00000780, -0.00000337, -0.00001061, -0.00001219, 0.00001417, 0.00001359, -0.00001291, -0.00000125, 0.00000614, -0.00000364, 0.00000119, -0.00000677, -0.00000147, 0.00001606, -0.00000649, -0.00001787, 0.00000402, 0.00001102, 0.00000242, -0.00000889, -0.00000926, -0.00000110, 0.00001270, 0.00001935, -0.00001280, -0.00002012, 0.00000627, -0.00000212, 0.00000252, 0.00000394, -0.00001229, 0.00000993, 0.00000029, -0.00000566, 0.00001985, -0.00001335, -0.00000719, 0.00000900, -0.00002230, 0.00002206, 0.00000925, -0.00003795, 0.00003137, 0.00000732, -0.00003865, 0.00002578, 0.00000808, -0.00000747, 0.00001808, -0.00001263, 0.00000989, 0.00001623, -0.00001910, 0.00001575, 0.00000966, -0.00002320, 0.00000276, 0.00000351, -0.00001122, -0.00000617, -0.00000092, 0.00000112, -0.00000798, -0.00000336, 0.00000819, 0.00000477, 0.00000084, -0.00000594, -0.00000427, -0.00000073, -0.00000613, -0.00000065, -0.00000047, -0.00000511, -0.00000294, -0.00000365, -0.00000156, -0.00000424, 0.00000191, 0.00000998, -0.00000638, -0.00000442, 0.00000193, -0.00001061, -0.00000466, -0.00000628, -0.00000724, 0.00000145, -0.00001163, -0.00000837, -0.00003373, -0.00007704, 0.00000460, 0.00009444, 0.00005925, 0.00004603, },
		{ 0.00001362, -0.00005135, -0.00004061, -0.00000925, -0.00000667, -0.00001075, -0.00000662, -0.00000223, 0.00000128, 0.00000159, 0.00000288, -0.00000081, -0.00001030, -0.00000858, 0.00000066, 0.00000269, -0.00000397, -0.00000235, 0.00000132, -0.00001049, -0.00000009, 0.00001818, -0.00000107, -0.00000331, 0.00000801, -0.00000738, 0.00000195, 0.00001119, -0.00001113, -0.00000170, 0.00001458, 0.00000245, -0.00000437, -0.00001369, -0.00000550, 0.00002079, 0.00000264, -0.00001366, 0.00000610, 0.00000123, -0.00000352, 0.00000167, -0.00000953, 0.00000612, 0.00001365, -0.00001594, -0.00001156, 0.00000994, 0.00000726, -0.00000170, -0.00000978, -0.00000763, 0.00000419, 0.00001854, 0.00000834, -0.00002195, -0.00000862, 0.00000629, -0.00000373, 0.00000730, -0.00000484, -0.00000687, 0.00001344, -0.00000809, 0.00000566, 0.00001336, -0.00002016, 0.00000522, -0.00000482, -0.00001240, 0.00003353, -0.00002062, -0.00001794, 0.00004494, -0.00002625, -0.00001979, 0.00003596, -0.00001180, 0.00000503, 0.00001257, -0.00001711, 0.00002178, 0.00000194, -0.00001510, 0.00002498, -0.00000543, -0.00001854, 0.00000960, -0.00000378, -0.00001088, -0.00000364, 0.00000069, -0.00000141, -0.00000925, 0.00000208, 0.00000843, 0.00000304, -0.00000131, -0.00000752, -0.00000138, -0.00000173, -0.00000616, 0.00000088, -0.00000258, -0.00000470, -0.00000237, -0.00000323, -0.00000258, -0.00000390, 0.00000764, 0.00000547, -0.00001027, 0.00000095, -0.00000243, -0.00001125, -0.00000215, -0.00000991, -0.00000226, -0.00000054, -0.00001580, -0.00000831, -0.00005370, -0.00007140, 0.00004653, 0.00009837, 0.00004715, 0.00006569, 0.00008549, },
		{ -0.00005756, -0.00002549, -0.00000470, -0.00001026, -0.00000991, -0.00000429, -0.00000137, 0.00000164, 0.00000231, 0.00000238, -0.00000419, -0.00001188, -0.00000574, 0.00000349, 0.00000038, -0.00000539, 0.00000059, -0.00000257, -0.00001018, 0.00001012, 0.00001337, -0.00000703, 0.00000330, 0.00000345, -0.00000830, 0.00001036, 0.00000370, -0.00001375, 0.00000725, 0.00001175, -0.00000082, -0.00000604, -0.00001542, 0.00000501, 0.00002039, -0.00000866, -0.00000752, 0.00000875, -0.00000458, -0.00000004, -0.00000103, -0.00000927, 0.00001410, 0.00000523, -0.00002092, -0.00000181, 0.00001096, 0.00000257, -0.00000386, -0.00001050, -0.00000449, 0.00001146, 0.00001790, -0.00000571, -0.00002133, 0.00000124, 0.00000215, -0.00000049, 0.00000648, -0.00001144, 0.00000316, 0.00000909, -0.00001000, 0.00001554, 0.00000020, -0.00001635, 0.00000919, -0.00001668, 0.00000895, 0.00002356, -0.00003960, 0.00001700, 0.00003053, -0.00004769, 0.00001234, 0.00002468, -0.00002149, 0.00002044, -0.00000262, -0.00001089, 0.00002721, -0.00001377, -0.00000175, 0.00002431, -0.00001848, -0.00000733, 0.00000949, -0.00001014, -0.00000785, -0.00000188, 0.00000135, -0.00000494, -0.00000754, 0.00000699, 0.00000666, 0.00000160, -0.00000405, -0.00000712, 0.00000137, -0.00000402, -0.00000499, 0.00000138, -0.00000454, -0.00000346, -0.00000211, -0.00000338, -0.00000392, -0.00000043, 0.00001088, -0.00000220, -0.00000893, 0.00000410, -0.00000858, -0.00000800, -0.00000304, -0.00001139, 0.00000389, -0.00000771, -0.00001601, -0.00001280, -0.00007569, -0.00004763, 0.00008548, 0.00008763, 0.00004652, 0.00008856, 0.00006840, -0.00002453, },
		{ -0.00001224, -0.00000613, -0.00001243, -0.00000725, -0.00000303, -0.00000079, 0.00000245, 0.00000259, 0.00000086, -0.00000770, -0.00001200, -0.00000154, 0.00000443, -0.00000313, -0.00000413, 0.00000165, -0.00000720, -0.00000419, 0.00001616, 0.00000425, -0.00000646, 0.00000741, -0.00000333, -0.00000286, 0.00001386, -0.00000637, -0.00000949, 0.00001316, 0.00000657, -0.00000184, -0.00000953, -0.00001289, 0.00001553, 0.00001235, -0.00001475, 0.00000162, 0.00000520, -0.00000705, 0.00000385, -0.00000636, -0.00000396, 0.00001828, -0.00000676, -0.00001870, 0.00000701, 0.00000738, -0.00000046, -0.00000536, -0.00001083, 0.00000158, 0.00001710, 0.00000999, -0.00001620, -0.00001356, 0.00000542, -0.00000119, 0.00000429, 0.00000013, -0.00001128, 0.00001156, 0.00000028, -0.00000403, 0.00001782, -0.00001193, -0.00000627, 0.00000265, -0.00001632, 0.00002763, -0.00000337, -0.00003352, 0.00004524, -0.00000510, -0.00004232, 0.00003798, -0.00000125, -0.00001322, 0.00002653, -0.00001940, 0.00000515, 0.00002087, -0.00002275, 0.00001506, 0.00001297, -0.00002312, 0.00000459, 0.00000290, -0.00001229, -0.00000432, -0.00000098, 0.00000087, -0.00000823, -0.00000292, 0.00000968, 0.00000391, 0.00000035, -0.00000686, -0.00000450, 0.00000261, -0.00000674, -0.00000270, 0.00000058, -0.00000569, -0.00000180, -0.00000245, -0.00000426, -0.00000388, 0.00000501, 0.00000921, -0.00000903, -0.00000341, 0.00000236, -0.00001272, -0.00000317, -0.00000750, -0.00000791, 0.00000671, -0.00001753, -0.00001120, -0.00002655, -0.00009169, -0.00000598, 0.00011082, 0.00006996, 0.00006178, 0.00010174, 0.00003168, -0.00005426, -0.00004868, },
		{ -0.00001010, -0.00001164, -0.00000465, -0.00000277, 0.00000029, 0.00000317, 0.00000224, -0.00000149, -0.00001096, -0.00000996, 0.00000280, 0.00000259, -0.00000562, -0.00000125, -0.00000039, -0.00000902, 0.00000482, 0.00001523, -0.00000373, -0.00000119, 0.00000621, -0.00000738, 0.00000587, 0.00001006, -0.00001352, -0.00000088, 0.00001392, 0.00000223, -0.00000253, -0.00001356, -0.00000502, 0.00002108, 0.00000021, -0.00001294, 0.00000812, -0.00000191, -0.00000442, 0.00000446, -0.00001100, 0.00000541, 0.00001537, -0.00001729, -0.00001008, 0.00001093, 0.00000209, -0.00000123, -0.00000761, -0.00000899, 0.00000968, 0.00001728, -0.00000177, -0.00001912, -0.00000418, 0.00000443, -0.00000091, 0.00000617, -0.00000760, -0.00000430, 0.00001353, -0.00000690, 0.00000597, 0.00001136, -0.00001679, 0.00000197, -0.00000816, -0.00000208, 0.00002981, -0.00003050, -0.00000383, 0.00004779, -0.00004092, -0.00001218, 0.00004081, -0.00002476, 0.00000822, 0.00001634, -0.00002648, 0.00002278, 0.00000400, -0.00001916, 0.00002648, -0.00000382, -0.00001723, 0.00001076, -0.00000587, -0.00000976, -0.00000243, -0.00000016, -0.00000104, -0.00000973, 0.00000338, 0.00000918, 0.00000149, -0.00000121, -0.00000878, -0.00000021, 0.00000133, -0.00000863, 0.00000014, -0.00000150, -0.00000550, -0.00000029, -0.00000375, -0.00000517, -0.00000127, 0.00000952, 0.00000294, -0.00001134, 0.00000247, -0.00000371, -0.00001209, -0.00000064, -0.00001248, 0.00000026, 0.00000216, -0.00002445, -0.00000487, -0.00005124, -0.00009132, 0.00004596, 0.00011608, 0.00005782, 0.00008836, 0.00009288, -0.00001503, -0.00006598, -0.00003149, -0.00000596, },
		{ -0.00000885, -0.00000338, -0.00000251, 0.00000180, 0.00000332, 0.00000134, -0.00000467, -0.00001310, -0.00000564, 0.00000539, -0.00000117, -0.00000553, 0.00000089, -0.00000424, -0.00000596, 0.00001220, 0.00000864, -0.00000657, 0.00000427, 0.00000095, -0.00000559, 0.00001259, 0.00000094, -0.00001425, 0.00000755, 0.00001064, 0.00000024, -0.00000492, -0.00001507, 0.00000596, 0.00001873, -0.00001010, -0.00000512, 0.00000828, -0.00000767, 0.00000134, 0.00000028, -0.00001102, 0.00001470, 0.00000547, -0.00002146, 0.00000072, 0.00000887, -0.00000152, -0.00000130, -0.00001034, -0.00000334, 0.00001644, 0.00001080, -0.00001190, -0.00001497, 0.00000211, 0.00000176, 0.00000210, 0.00000306, -0.00001155, 0.00000528, 0.00000870, -0.00000792, 0.00001376, 0.00000034, -0.00001371, 0.00000277, -0.00001332, 0.00001712, 0.00001205, -0.00003931, 0.00003211, 0.00002108, -0.00005475, 0.00002467, 0.00001905, -0.00002995, 0.00002817, -0.00000595, -0.00001747, 0.00003101, -0.00001503, -0.00000399, 0.00002588, -0.00001731, -0.00000478, 0.00000824, -0.00001139, -0.00000519, -0.00000253, 0.00000104, -0.00000450, -0.00000795, 0.00000907, 0.00000585, 0.00000038, -0.00000365, -0.00000859, 0.00000425, -0.00000259, -0.00000836, 0.00000242, -0.00000433, -0.00000371, 0.00000019, -0.00000589, -0.00000466, 0.00000342, 0.00001021, -0.00000478, -0.00000819, 0.00000455, -0.00001051, -0.00000721, -0.00000297, -0.00001337, 0.00000874, -0.00000992, -0.00002316, -0.00000469, -0.00008196, -0.00006685, 0.00009457, 0.00010367, 0.00006214, 0.00011245, 0.00005741, -0.00005596, -0.00005805, -0.00001556, -0.00000679, -0.00001263, },
		{ -0.00000321, -0.00000153, 0.00000310, 0.00000290, -0.00000018, -0.00000845, -0.00001292, -0.00000021, 0.00000491, -0.00000471, -0.00000320, 0.00000050, -0.00000694, 0.00000091, 0.00001429, 0.00000072, -0.00000405, 0.00000602, -0.00000425, 0.00000116, 0.00001324, -0.00000863, -0.00000883, 0.00001223, 0.00000628, -0.00000038, -0.00000906, -0.00001159, 0.00001534, 0.00000960, -0.00001384, 0.00000349, 0.00000261, -0.00000830, 0.00000588, -0.00000657, -0.00000481, 0.00001882, -0.00000739, -0.00001746, 0.00000845, 0.00000315, -0.00000191, -0.00000278, -0.00001136, 0.00000548, 0.00001799, 0.00000020, -0.00001626, -0.00000759, 0.00000399, 0.00000064, 0.00000456, -0.00000334, -0.00000913, 0.00001206, 0.00000090, -0.00000264, 0.00001502, -0.00000906, -0.00000730, -0.00000278, -0.00000692, 0.00002755, -0.00001533, -0.00002214, 0.00005162, -0.00001990, -0.00003772, 0.00004565, -0.00001306, -0.00001312, 0.00003202, -0.00002702, 0.00000379, 0.00002365, -0.00002537, 0.00001461, 0.00001346, -0.00002051, 0.00000634, -0.00000035, -0.00001078, -0.00000236, -0.00000290, 0.00000187, -0.00000856, -0.00000240, 0.00001154, 0.00000155, 0.00000046, -0.00000704, -0.00000534, 0.00000670, -0.00000797, -0.00000527, 0.00000277, -0.00000674, -0.00000065, -0.00000132, -0.00000783, -0.00000164, 0.00000787, 0.00000626, -0.00000987, -0.00000214, 0.00000120, -0.00001375, -0.00000181, -0.00000906, -0.00000744, 0.00001109, -0.00002429, -0.00001303, -0.00001859, -0.00010686, -0.00001867, 0.00012600, 0.00008538, 0.00008488, 0.00011656, 0.00000306, -0.00007667, -0.00003787, -0.00000778, -0.00001066, -0.00001202, -0.00000615, },
		{ 0.00000002, 0.00000377, 0.00000214, -0.00000267, -0.00001190, -0.00000981, 0.00000421, 0.00000168, -0.00000606, -0.00000057, -0.00000216, -0.00000600, 0.00000805, 0.00001070, -0.00000418, 0.00000073, 0.00000337, -0.00000545, 0.00000877, 0.00000742, -0.00001398, -0.00000061, 0.00001243, 0.00000321, -0.00000167, -0.00001274, -0.00000334, 0.00001878, -0.00000157, -0.00001024, 0.00000773, -0.00000474, -0.00000366, 0.00000562, -0.00001162, 0.00000534, 0.00001495, -0.00001732, -0.00000775, 0.00000989, -0.00000218, -0.00000030, -0.00000621, -0.00000823, 0.00001415, 0.00001267, -0.00000969, -0.00001429, -0.00000122, 0.00000320, 0.00000178, 0.00000371, -0.00000895, -0.00000173, 0.00001288, -0.00000487, 0.00000524, 0.00000981, -0.00001308, -0.00000325, -0.00000799, 0.00000796, 0.00002048, -0.00003437, 0.00001183, 0.00004140, -0.00005073, -0.00000026, 0.00003835, -0.00003464, 0.00001450, 0.00001552, -0.00003316, 0.00002495, 0.00000409, -0.00002083, 0.00002600, -0.00000336, -0.00001287, 0.00000944, -0.00000853, -0.00000577, -0.00000307, -0.00000148, 0.00000060, -0.00001087, 0.00000542, 0.00000934, -0.00000117, 0.00000030, -0.00001022, 0.00000081, 0.00000505, -0.00001232, -0.00000010, 0.00000032, -0.00000714, 0.00000232, -0.00000505, -0.00000765, 0.00000341, 0.00000931, -0.00000029, -0.00001001, 0.00000248, -0.00000548, -0.00001162, -0.00000020, -0.00001426, 0.00000335, 0.00000313, -0.00003257, -0.00000052, -0.00004880, -0.00011153, 0.00004253, 0.00013375, 0.00007687, 0.00011526, 0.00008898, -0.00005192, -0.00007176, -0.00001767, -0.00000880, -0.00001281, -0.00000947, -0.00000481, -0.00000306, },
		{ 0.00000389, 0.00000090, -0.00000622, -0.00001350, -0.00000455, 0.00000577, -0.00000239, -0.00000482, 0.00000033, -0.00000489, -0.00000128, 0.00001177, 0.00000428, -0.00000434, 0.00000385, -0.00000107, -0.00000162, 0.00001269, -0.00000164, -0.00001328, 0.00000645, 0.00000990, 0.00000177, -0.00000474, -0.00001308, 0.00000639, 0.00001497, -0.00000932, -0.00000271, 0.00000566, -0.00000890, 0.00000281, 0.00000033, -0.00001104, 0.00001435, 0.00000450, -0.00001992, 0.00000240, 0.00000561, -0.00000409, 0.00000058, -0.00000968, -0.00000050, 0.00001830, 0.00000244, -0.00001474, -0.00000865, 0.00000207, 0.00000236, 0.00000345, -0.00000054, -0.00001024, 0.00000651, 0.00000845, -0.00000563, 0.00001106, 0.00000163, -0.00001233, -0.00000366, -0.00000618, 0.00002090, -0.00000099, -0.00003144, 0.00004185, 0.00000652, -0.00005308, 0.00003461, 0.00000852, -0.00003208, 0.00003422, -0.00001186, -0.00001979, 0.00003308, -0.00001631, -0.00000436, 0.00002406, -0.00001465, -0.00000085, 0.00000365, -0.00001061, -0.00000134, -0.00000561, 0.00000189, -0.00000376, -0.00000876, 0.00001200, 0.00000350, -0.00000076, -0.00000197, -0.00001090, 0.00000777, -0.00000131, -0.00001277, 0.00000478, -0.00000439, -0.00000438, 0.00000299, -0.00000965, -0.00000384, 0.00000806, 0.00000658, -0.00000580, -0.00000622, 0.00000263, -0.00001121, -0.00000618, -0.00000393, -0.00001354, 0.00001252, -0.00001333, -0.00002861, 0.00000307, -0.00008793, -0.00008655, 0.00009957, 0.00012334, 0.00008821, 0.00013317, 0.00003208, -0.00008714, -0.00004800, -0.00000741, -0.00001356, -0.00001132, -0.00000739, -0.00000438, -0.00000225, 0.00000148, },
		{ -0.00000136, -0.00001004, -0.00001215, 0.00000083, 0.00000410, -0.00000507, -0.00000236, -0.00000103, -0.00000521, 0.00000483, 0.00001061, -0.00000105, -0.00000133, 0.00000343, -0.00000359, 0.00000487, 0.00001077, -0.00000944, -0.00000796, 0.00001017, 0.00000701, 0.00000060, -0.00000861, -0.00000881, 0.00001324, 0.00000649, -0.00001071, 0.00000362, -0.00000052, -0.00000753, 0.00000659, -0.00000670, -0.00000434, 0.00001739, -0.00000760, -0.00001476, 0.00000803, -0.00000076, -0.00000243, -0.00000116, -0.00001008, 0.00000916, 0.00001528, -0.00000810, -0.00001380, -0.00000297, 0.00000285, 0.00000271, 0.00000333, -0.00000574, -0.00000644, 0.00001161, 0.00000215, -0.00000185, 0.00001222, -0.00000559, -0.00001016, -0.00000561, 0.00000357, 0.00002164, -0.00002321, -0.00000685, 0.00004882, -0.00003245, -0.00002639, 0.00004616, -0.00002405, -0.00000793, 0.00003261, -0.00003320, 0.00000514, 0.00002343, -0.00002566, 0.00001370, 0.00001129, -0.00001471, 0.00000667, -0.00000549, -0.00000596, -0.00000163, -0.00000611, 0.00000459, -0.00000963, -0.00000139, 0.00001345, -0.00000247, 0.00000191, -0.00000692, -0.00000682, 0.00001177, -0.00001033, -0.00000784, 0.00000622, -0.00000890, 0.00000094, -0.00000043, -0.00001213, 0.00000311, 0.00000928, 0.00000136, -0.00000765, -0.00000196, -0.00000141, -0.00001268, -0.00000181, -0.00001021, -0.00000546, 0.00001289, -0.00003026, -0.00001357, -0.00001202, -0.00012037, -0.00003375, 0.00013737, 0.00011005, 0.00011581, 0.00011954, -0.00003632, -0.00009073, -0.00002056, -0.00000922, -0.00001572, -0.00000795, -0.00000697, -0.00000362, -0.00000110, 0.00000253, 0.00000372, },
		{ -0.00001255, -0.00000812, 0.00000416, 0.00000059, -0.00000531, -0.00000069, -0.00000326, -0.00000223, 0.00000906, 0.00000607, -0.00000286, 0.00000166, 0.00000055, -0.00000210, 0.00001004, 0.00000426, -0.00001286, -0.00000121, 0.00001075, 0.00000495, -0.00000165, -0.00001093, -0.00000140, 0.00001440, -0.00000199, -0.00000673, 0.00000516, -0.00000636, -0.00000206, 0.00000522, -0.00001092, 0.00000522, 0.00001280, -0.00001569, -0.00000547, 0.00000729, -0.00000511, 0.00000040, -0.00000486, -0.00000551, 0.00001615, 0.00000601, -0.00001443, -0.00000890, 0.00000053, 0.00000296, 0.00000354, 0.00000061, -0.00000879, 0.00000033, 0.00001193, -0.00000244, 0.00000371, 0.00000905, -0.00001012, -0.00000903, -0.00000412, 0.00001502, 0.00000796, -0.00003096, 0.00002472, 0.00002780, -0.00005259, 0.00001208, 0.00002906, -0.00003880, 0.00002121, 0.00001040, -0.00003542, 0.00002648, 0.00000255, -0.00001945, 0.00002267, -0.00000302, -0.00000595, 0.00000477, -0.00001024, 0.00000052, -0.00000639, -0.00000211, 0.00000314, -0.00001291, 0.00000859, 0.00000829, -0.00000449, 0.00000353, -0.00001245, 0.00000209, 0.00000917, -0.00001745, 0.00000092, 0.00000233, -0.00000979, 0.00000587, -0.00000757, -0.00000936, 0.00000999, 0.00000583, -0.00000291, -0.00000617, -0.00000040, -0.00000647, -0.00000997, -0.00000178, -0.00001377, 0.00000592, 0.00000169, -0.00003792, 0.00000300, -0.00004699, -0.00012936, 0.00003348, 0.00015152, 0.00010967, 0.00014172, 0.00006828, -0.00009059, -0.00006601, -0.00000418, -0.00001709, -0.00001244, -0.00000609, -0.00000721, -0.00000202, -0.00000036, 0.00000348, 0.00000297, -0.00000490, },
		{ -0.00000308, 0.00000447, -0.00000260, -0.00000377, -0.00000088, -0.00000418, 0.00000265, 0.00000944, 0.00000131, -0.00000151, 0.00000229, -0.00000189, 0.00000256, 0.00001094, -0.00000340, -0.00001157, 0.00000428, 0.00000974, 0.00000330, -0.00000494, -0.00000982, 0.00000575, 0.00001028, -0.00000660, -0.00000122, 0.00000190, -0.00000822, 0.00000360, -0.00000025, -0.00000953, 0.00001263, 0.00000308, -0.00001667, 0.00000271, 0.00000205, -0.00000543, 0.00000157, -0.00000772, 0.00000289, 0.00001660, -0.00000529, -0.00001462, -0.00000342, 0.00000191, 0.00000354, 0.00000326, -0.00000349, -0.00000798, 0.00000683, 0.00000852, -0.00000360, 0.00000814, 0.00000354, -0.00001258, -0.00000826, 0.00000299, 0.00001921, -0.00001199, -0.00001800, 0.00004351, -0.00000870, -0.00004311, 0.00003877, -0.00000392, -0.00002774, 0.00003616, -0.00001829, -0.00001792, 0.00003197, -0.00001639, -0.00000334, 0.00001878, -0.00000958, 0.00000309, -0.00000387, -0.00000669, 0.00000228, -0.00001081, 0.00000465, -0.00000353, -0.00000960, 0.00001565, -0.00000080, -0.00000099, 0.00000065, -0.00001425, 0.00001234, -0.00000081, -0.00001780, 0.00000886, -0.00000554, -0.00000508, 0.00000629, -0.00001482, -0.00000064, 0.00001240, -0.00000022, -0.00000371, -0.00000423, -0.00000198, -0.00000927, -0.00000619, -0.00000559, -0.00001088, 0.00001322, -0.00001672, -0.00003121, 0.00000777, -0.00009178, -0.00010585, 0.00009689, 0.00015066, 0.00012723, 0.00014247, -0.00000758, -0.00011041, -0.00002970, -0.00000473, -0.00002194, -0.00000613, -0.00000736, -0.00000614, -0.00000048, -0.00000005, 0.00000465, 0.00000081, -0.00000900, -0.00001240, },
		{ 0.00000253, -0.00000404, -0.00000212, -0.00000228, -0.00000251, 0.00000679, 0.00000649, -0.00000111, 0.00000068, 0.00000062, -0.00000155, 0.00000744, 0.00000735, -0.00000890, -0.00000738, 0.00000759, 0.00000842, 0.00000107, -0.00000774, -0.00000545, 0.00000953, 0.00000396, -0.00000643, 0.00000196, -0.00000318, -0.00000548, 0.00000588, -0.00000618, -0.00000321, 0.00001421, -0.00000686, -0.00001144, 0.00000598, -0.00000376, -0.00000266, 0.00000002, -0.00000690, 0.00001119, 0.00000988, -0.00001363, -0.00001009, 0.00000022, 0.00000256, 0.00000424, 0.00000107, -0.00000672, -0.00000385, 0.00001061, 0.00000379, -0.00000168, 0.00000995, -0.00000258, -0.00001394, -0.00000500, 0.00001241, 0.00001145, -0.00002480, 0.00000825, 0.00003768, -0.00003887, -0.00001178, 0.00003893, -0.00003105, 0.00000007, 0.00002785, -0.00003584, 0.00000764, 0.00002018, -0.00002275, 0.00001137, 0.00000741, -0.00000587, 0.00000437, -0.00001102, 0.00000197, -0.00000334, -0.00000941, 0.00000886, -0.00001195, 0.00000067, 0.00001477, -0.00000788, 0.00000524, -0.00000721, -0.00000857, 0.00001775, -0.00001431, -0.00000952, 0.00001056, -0.00001259, 0.00000358, -0.00000027, -0.00001672, 0.00001059, 0.00000782, -0.00000439, -0.00000175, -0.00000456, -0.00000436, -0.00000913, -0.00000458, -0.00000944, -0.00000252, 0.00001066, -0.00003295, -0.00001391, -0.00000848, -0.00012904, -0.00005334, 0.00014350, 0.00015026, 0.00015144, 0.00010312, -0.00008052, -0.00009189, -0.00000220, -0.00001668, -0.00001832, -0.00000206, -0.00001011, -0.00000299, -0.00000032, 0.00000073, 0.00000549, -0.00000326, -0.00001196, -0.00000952, 0.00000080, },
		{ -0.00000377, -0.00000162, -0.00000322, 0.00000118, 0.00000808, 0.00000254, -0.00000077, 0.00000147, -0.00000133, 0.00000174, 0.00000964, 0.00000137, -0.00001082, -0.00000255, 0.00000923, 0.00000691, -0.00000202, -0.00000829, 0.00000005, 0.00000901, -0.00000100, -0.00000352, 0.00000130, -0.00000649, -0.00000049, 0.00000379, -0.00000890, 0.00000449, 0.00000967, -0.00001257, -0.00000390, 0.00000387, -0.00000678, 0.00000055, -0.00000281, -0.00000173, 0.00001495, -0.00000096, -0.00001595, -0.00000412, 0.00000172, 0.00000353, 0.00000388, -0.00000232, -0.00000742, 0.00000164, 0.00001104, -0.00000011, 0.00000189, 0.00000892, -0.00000863, -0.00001365, 0.00000241, 0.00001730, -0.00000443, -0.00002114, 0.00003134, 0.00001092, -0.00004573, 0.00002092, 0.00001536, -0.00003618, 0.00002555, 0.00000266, -0.00003274, 0.00002569, 0.00000052, -0.00001532, 0.00001610, -0.00000159, 0.00000223, -0.00000326, -0.00000953, 0.00000772, -0.00001241, -0.00000097, 0.00000570, -0.00001564, 0.00001298, 0.00000539, -0.00000761, 0.00000831, -0.00001589, 0.00000424, 0.00001305, -0.00002386, 0.00000389, 0.00000365, -0.00001319, 0.00001063, -0.00001174, -0.00000940, 0.00001769, -0.00000164, -0.00000314, -0.00000071, -0.00000707, -0.00000492, -0.00000820, -0.00000562, -0.00000945, 0.00000596, -0.00000167, -0.00003845, 0.00000282, -0.00004484, -0.00014273, 0.00001470, 0.00017225, 0.00016077, 0.00015953, 0.00002841, -0.00012226, -0.00004974, 0.00000408, -0.00002784, -0.00000815, -0.00000369, -0.00001076, 0.00000064, -0.00000170, 0.00000296, 0.00000450, -0.00000845, -0.00001217, -0.00000540, 0.00000233, 0.00000009, },
		{ -0.00000218, -0.00000236, 0.00000489, 0.00000634, 0.00000002, 0.00000079, 0.00000032, -0.00000136, 0.00000597, 0.00000807, -0.00000418, -0.00000975, 0.00000161, 0.00001001, 0.00000464, -0.00000500, -0.00000610, 0.00000402, 0.00000575, -0.00000283, -0.00000101, -0.00000186, -0.00000618, 0.00000340, -0.00000081, -0.00000706, 0.00000958, 0.00000189, -0.00001249, 0.00000166, -0.00000118, -0.00000606, 0.00000201, -0.00000423, 0.00000543, 0.00001189, -0.00001095, -0.00001255, 0.00000037, 0.00000230, 0.00000459, 0.00000178, -0.00000524, -0.00000542, 0.00000646, 0.00000889, -0.00000208, 0.00000564, 0.00000524, -0.00001404, -0.00000976, 0.00001170, 0.00001263, -0.00001810, -0.00000257, 0.00003670, -0.00002017, -0.00002757, 0.00003547, -0.00001475, -0.00001862, 0.00003276, -0.00002291, -0.00001311, 0.00002723, -0.00001417, -0.00000203, 0.00001086, -0.00000174, 0.00000544, -0.00001291, 0.00000066, 0.00000412, -0.00001699, 0.00000949, -0.00000460, -0.00000980, 0.00001946, -0.00000703, 0.00000048, 0.00000342, -0.00001840, 0.00001817, -0.00000186, -0.00002259, 0.00001457, -0.00000851, -0.00000506, 0.00000972, -0.00002126, 0.00000557, 0.00001494, -0.00000952, 0.00000269, -0.00000409, -0.00000856, -0.00000369, -0.00000895, -0.00000666, -0.00000522, 0.00000876, -0.00001770, -0.00003116, 0.00000684, -0.00009030, -0.00012580, 0.00008357, 0.00019211, 0.00017822, 0.00013100, -0.00005697, -0.00011908, -0.00000862, -0.00000888, -0.00002824, 0.00000124, -0.00000963, -0.00000714, 0.00000201, -0.00000292, 0.00000617, 0.00000040, -0.00001258, -0.00000945, -0.00000211, 0.00000191, -0.00000143, -0.00000296, },
		{ 0.00000025, 0.00000668, 0.00000319, -0.00000019, 0.00000155, -0.00000144, 0.00000121, 0.00000856, 0.00000383, -0.00000752, -0.00000717, 0.00000499, 0.00001002, 0.00000135, -0.00000638, -0.00000242, 0.00000507, 0.00000240, -0.00000221, -0.00000083, -0.00000478, -0.00000307, 0.00000416, -0.00000482, -0.00000212, 0.00000997, -0.00000511, -0.00000828, 0.00000291, -0.00000576, -0.00000296, 0.00000145, -0.00000249, 0.00001064, 0.00000333, -0.00001601, -0.00000633, 0.00000244, 0.00000317, 0.00000461, -0.00000147, -0.00000634, -0.00000177, 0.00000946, 0.00000542, -0.00000186, 0.00000838, -0.00000088, -0.00001719, -0.00000127, 0.00001734, -0.00000022, -0.00001983, 0.00001919, 0.00002128, -0.00003720, 0.00000190, 0.00002557, -0.00003205, 0.00000791, 0.00001888, -0.00003380, 0.00000940, 0.00001487, -0.00001667, 0.00000684, 0.00000330, 0.00000496, -0.00000117, -0.00001509, 0.00001181, -0.00000800, -0.00001139, 0.00001385, -0.00001554, 0.00000419, 0.00001466, -0.00001389, 0.00001056, -0.00000865, -0.00000985, 0.00002412, -0.00002012, -0.00000935, 0.00001492, -0.00001785, 0.00000783, -0.00000145, -0.00002075, 0.00002043, 0.00000227, -0.00000919, 0.00000746, -0.00001142, -0.00000576, -0.00000374, -0.00001105, -0.00000478, -0.00000031, 0.00000403, -0.00002965, -0.00001668, -0.00000806, -0.00012975, -0.00008151, 0.00014565, 0.00021232, 0.00018441, 0.00006219, -0.00012023, -0.00007925, 0.00001106, -0.00002755, -0.00001675, 0.00000302, -0.00001459, -0.00000072, -0.00000024, -0.00000159, 0.00000810, -0.00000633, -0.00001336, -0.00000544, -0.00000082, 0.00000099, -0.00000188, -0.00000266, -0.00000268, },
		{ 0.00000597, 0.00000070, 0.00000100, 0.00000065, -0.00000184, 0.00000510, 0.00000806, -0.00000086, -0.00000851, -0.00000419, 0.00000793, 0.00000877, -0.00000221, -0.00000543, 0.00000069, 0.00000380, 0.00000078, -0.00000129, -0.00000271, -0.00000542, 0.00000048, 0.00000208, -0.00000605, 0.00000294, 0.00000635, -0.00000857, -0.00000328, 0.00000033, -0.00000760, 0.00000037, 0.00000034, 0.00000181, 0.00001077, -0.00000670, -0.00001500, -0.00000056, 0.00000301, 0.00000436, 0.00000280, -0.00000436, -0.00000537, 0.00000225, 0.00001037, 0.00000176, 0.00000035, 0.00000886, -0.00000862, -0.00001562, 0.00000961, 0.00001445, -0.00001346, -0.00000765, 0.00003010, -0.00000466, -0.00003188, 0.00002349, 0.00000081, -0.00002763, 0.00002554, -0.00000528, -0.00002593, 0.00002165, -0.00000065, -0.00000957, 0.00000677, 0.00000186, 0.00000983, -0.00001361, -0.00000547, 0.00001398, -0.00002024, 0.00000255, 0.00000720, -0.00001834, 0.00001827, 0.00000029, -0.00000948, 0.00001391, -0.00002060, 0.00000781, 0.00001575, -0.00003085, 0.00000911, 0.00000324, -0.00001661, 0.00001647, -0.00001778, -0.00000684, 0.00002508, -0.00001296, 0.00000073, 0.00000454, -0.00001736, 0.00000076, -0.00000819, -0.00001091, -0.00000036, 0.00000099, -0.00000498, -0.00003338, -0.00000429, -0.00003948, -0.00015141, -0.00001799, 0.00020199, 0.00023167, 0.00015814, -0.00002817, -0.00013817, -0.00002782, 0.00000415, -0.00003673, -0.00000121, -0.00000378, -0.00001363, 0.00000420, -0.00000437, 0.00000300, 0.00000599, -0.00001300, -0.00001017, -0.00000260, -0.00000103, 0.00000070, -0.00000206, -0.00000294, -0.00000202, 0.00000325, },
		{ -0.00000000, 0.00000191, -0.00000115, 0.00000014, 0.00000806, 0.00000489, -0.00000425, -0.00000809, -0.00000103, 0.00001031, 0.00000590, -0.00000464, -0.00000282, 0.00000177, 0.00000211, 0.00000081, -0.00000177, -0.00000478, -0.00000368, 0.00000237, -0.00000090, -0.00000442, 0.00000568, 0.00000127, -0.00000821, -0.00000035, -0.00000380, -0.00000640, 0.00000257, 0.00000036, 0.00000607, 0.00000547, -0.00001385, -0.00000972, 0.00000293, 0.00000345, 0.00000485, -0.00000040, -0.00000561, -0.00000308, 0.00000578, 0.00000933, -0.00000104, 0.00000388, 0.00000589, -0.00001566, -0.00000788, 0.00001754, 0.00000319, -0.00001798, 0.00001082, 0.00002337, -0.00002484, -0.00001058, 0.00002523, -0.00002112, -0.00000756, 0.00002441, -0.00002395, -0.00000736, 0.00001954, -0.00000918, -0.00000167, 0.00000184, 0.00000830, 0.00000497, -0.00002150, 0.00001070, 0.00000305, -0.00002251, 0.00001579, -0.00000738, -0.00000862, 0.00002253, -0.00001462, 0.00000415, 0.00000537, -0.00002265, 0.00002506, -0.00000505, -0.00002599, 0.00002124, -0.00001371, -0.00000346, 0.00001257, -0.00002834, 0.00001491, 0.00001408, -0.00001970, 0.00001366, -0.00000779, -0.00001529, 0.00000551, -0.00001606, -0.00000502, 0.00000228, -0.00000213, -0.00001318, -0.00003051, -0.00000096, -0.00007965, -0.00014996, 0.00005909, 0.00025537, 0.00023541, 0.00009092, -0.00010706, -0.00010983, 0.00000808, -0.00001813, -0.00002963, 0.00000784, -0.00001354, -0.00000631, 0.00000385, -0.00000642, 0.00000856, -0.00000104, -0.00001600, -0.00000489, -0.00000236, -0.00000110, 0.00000098, -0.00000275, -0.00000300, -0.00000012, 0.00000509, 0.00000370, },
		{ 0.00000148, -0.00000215, 0.00000382, 0.00000840, 0.00000080, -0.00000586, -0.00000709, 0.00000259, 0.00001142, 0.00000191, -0.00000491, -0.00000026, 0.00000096, 0.00000159, 0.00000101, -0.00000376, -0.00000531, -0.00000105, 0.00000211, -0.00000283, -0.00000153, 0.00000546, -0.00000272, -0.00000574, -0.00000050, -0.00000703, -0.00000328, 0.00000365, 0.00000196, 0.00000744, -0.00000279, -0.00001573, -0.00000330, 0.00000432, 0.00000425, 0.00000362, -0.00000353, -0.00000499, -0.00000036, 0.00000844, 0.00000666, -0.00000198, 0.00000721, -0.00000080, -0.00001846, 0.00000417, 0.00001725, -0.00001021, -0.00000996, 0.00002337, 0.00000397, -0.00002791, 0.00001102, 0.00000929, -0.00002686, 0.00001293, 0.00000797, -0.00002728, 0.00000892, 0.00000907, -0.00000833, -0.00000002, 0.00000046, 0.00001596, -0.00000954, -0.00001617, 0.00002165, -0.00001518, -0.00001093, 0.00001827, -0.00001986, 0.00000919, 0.00001241, -0.00001934, 0.00001738, -0.00001167, -0.00000981, 0.00002995, -0.00002739, -0.00000657, 0.00001809, -0.00002411, 0.00001394, -0.00000459, -0.00002310, 0.00003152, -0.00000787, -0.00001093, 0.00001839, -0.00002310, -0.00000348, 0.00000172, -0.00002105, 0.00000545, -0.00000149, -0.00000573, -0.00001855, -0.00002552, -0.00000863, -0.00012094, -0.00012332, 0.00014882, 0.00030032, 0.00020371, -0.00000356, -0.00014493, -0.00005636, 0.00001441, -0.00003720, -0.00001147, 0.00000502, -0.00001837, 0.00000271, -0.00000162, -0.00000336, 0.00001077, -0.00001052, -0.00001337, -0.00000090, -0.00000399, 0.00000020, 0.00000078, -0.00000375, -0.00000204, 0.00000209, 0.00000512, 0.00000148, 0.00000073, },
		{ -0.00000112, 0.00000732, 0.00000596, -0.00000245, -0.00000634, -0.00000566, 0.00000660, 0.00001049, -0.00000187, -0.00000315, 0.00000074, -0.00000032, 0.00000239, -0.00000009, -0.00000588, -0.00000388, 0.00000075, 0.00000070, -0.00000304, 0.00000077, 0.00000340, -0.00000446, -0.00000343, -0.00000292, -0.00000799, 0.00000039, 0.00000434, 0.00000393, 0.00000471, -0.00001030, -0.00001271, 0.00000183, 0.00000476, 0.00000476, 0.00000076, -0.00000515, -0.00000324, 0.00000238, 0.00000985, 0.00000306, -0.00000052, 0.00000818, -0.00000941, -0.00001424, 0.00001517, 0.00000766, -0.00001706, 0.00000583, 0.00002170, -0.00001509, -0.00001461, 0.00001907, -0.00001106, -0.00001555, 0.00002064, -0.00001111, -0.00001687, 0.00001450, 0.00000009, -0.00000378, -0.00000402, 0.00000748, 0.00001508, -0.00002443, 0.00000182, 0.00001764, -0.00002823, 0.00000825, 0.00000677, -0.00002006, 0.00002364, -0.00000681, -0.00000920, 0.00001922, -0.00002607, 0.00001299, 0.00001631, -0.00003726, 0.00001630, 0.00000026, -0.00001898, 0.00002278, -0.00002549, -0.00000100, 0.00003035, -0.00002694, 0.00000973, 0.00000719, -0.00002982, 0.00001134, -0.00001209, -0.00001564, 0.00001315, -0.00001115, -0.00000501, -0.00002389, -0.00002073, -0.00002669, -0.00015805, -0.00006702, 0.00024917, 0.00031889, 0.00012718, -0.00009360, -0.00013221, -0.00000791, -0.00000371, -0.00003973, 0.00000535, -0.00000633, -0.00001416, 0.00000710, -0.00000782, 0.00000408, 0.00000641, -0.00001759, -0.00000651, -0.00000072, -0.00000512, 0.00000250, -0.00000071, -0.00000409, -0.00000018, 0.00000348, 0.00000387, 0.00000037, 0.00000168, -0.00000005, },
		{ 0.00000880, 0.00000203, -0.00000405, -0.00000655, -0.00000333, 0.00001028, 0.00000742, -0.00000402, -0.00000066, -0.00000013, -0.00000052, 0.00000334, -0.00000276, -0.00000656, -0.00000155, 0.00000104, -0.00000047, -0.00000222, 0.00000161, 0.00000115, -0.00000441, -0.00000273, -0.00000591, -0.00000652, 0.00000380, 0.00000502, 0.00000444, -0.00000107, -0.00001423, -0.00000703, 0.00000487, 0.00000509, 0.00000395, -0.00000249, -0.00000482, -0.00000130, 0.00000510, 0.00000954, -0.00000021, 0.00000277, 0.00000505, -0.00001611, -0.00000345, 0.00001891, -0.00000626, -0.00001222, 0.00001892, 0.00000720, -0.00002187, 0.00000368, 0.00001050, -0.00002165, 0.00000233, 0.00001291, -0.00002079, -0.00000265, 0.00001044, -0.00000189, -0.00000309, -0.00000643, 0.00001910, 0.00000131, -0.00002760, 0.00002171, -0.00000113, -0.00002576, 0.00002227, -0.00001167, -0.00000560, 0.00002382, -0.00002250, 0.00000994, 0.00000563, -0.00002596, 0.00003224, -0.00001050, -0.00002684, 0.00002766, -0.00002096, 0.00000041, 0.00001394, -0.00003501, 0.00002677, 0.00000863, -0.00002843, 0.00002821, -0.00001675, -0.00001967, 0.00001695, -0.00002820, 0.00000166, 0.00000911, -0.00001911, -0.00000033, -0.00003296, -0.00001451, -0.00005688, -0.00018391, 0.00002685, 0.00034703, 0.00028822, 0.00001864, -0.00014616, -0.00008412, 0.00001377, -0.00002804, -0.00002542, 0.00001061, -0.00001706, -0.00000368, 0.00000385, -0.00000925, 0.00001116, -0.00000383, -0.00001809, 0.00000036, -0.00000410, -0.00000363, 0.00000407, -0.00000310, -0.00000296, 0.00000162, 0.00000366, 0.00000239, 0.00000038, 0.00000183, -0.00000144, 0.00000190, },
		{ -0.00000160, -0.00000437, -0.00000680, 0.00000030, 0.00001247, 0.00000311, -0.00000392, 0.00000097, -0.00000189, 0.00000084, 0.00000292, -0.00000584, -0.00000523, 0.00000020, 0.00000040, -0.00000074, -0.00000153, 0.00000137, -0.00000024, -0.00000389, -0.00000379, -0.00000793, -0.00000320, 0.00000658, 0.00000520, 0.00000240, -0.00000735, -0.00001379, -0.00000114, 0.00000631, 0.00000514, 0.00000156, -0.00000458, -0.00000322, 0.00000048, 0.00000763, 0.00000735, -0.00000169, 0.00000591, -0.00000195, -0.00001678, 0.00000929, 0.00001254, -0.00001595, 0.00000175, 0.00002031, -0.00001002, -0.00001363, 0.00001358, -0.00000613, -0.00001703, 0.00001360, -0.00000224, -0.00001770, 0.00000564, 0.00000436, 0.00000064, -0.00000848, -0.00000016, 0.00002508, -0.00001932, -0.00001358, 0.00002940, -0.00002351, -0.00000762, 0.00002084, -0.00002388, 0.00001514, 0.00000772, -0.00002298, 0.00002466, -0.00001617, -0.00000779, 0.00003404, -0.00003518, -0.00000096, 0.00001882, -0.00003029, 0.00002162, -0.00000997, -0.00002269, 0.00004207, -0.00002209, -0.00000780, 0.00002855, -0.00003882, 0.00000414, 0.00000472, -0.00003301, 0.00002185, -0.00000895, -0.00001560, 0.00000030, -0.00004391, -0.00000588, -0.00010394, -0.00018314, 0.00016162, 0.00041366, 0.00019692, -0.00008837, -0.00014570, -0.00003064, 0.00000623, -0.00004057, -0.00000520, 0.00000309, -0.00001920, 0.00000570, -0.00000428, -0.00000364, 0.00001225, -0.00001501, -0.00001154, 0.00000288, -0.00000803, 0.00000052, 0.00000328, -0.00000485, -0.00000066, 0.00000238, 0.00000315, 0.00000140, 0.00000083, 0.00000107, -0.00000145, 0.00000558, 0.00000747, },
		{ -0.00000438, -0.00000666, 0.00000500, 0.00001226, -0.00000100, -0.00000208, 0.00000083, -0.00000303, 0.00000293, 0.00000052, -0.00000771, -0.00000266, 0.00000060, -0.00000007, -0.00000054, -0.00000157, 0.00000104, -0.00000084, -0.00000398, -0.00000582, -0.00000810, 0.00000121, 0.00000837, 0.00000402, -0.00000175, -0.00001170, -0.00001007, 0.00000355, 0.00000684, 0.00000421, -0.00000153, -0.00000472, -0.00000145, 0.00000235, 0.00000931, 0.00000392, -0.00000070, 0.00000638, -0.00000989, -0.00000985, 0.00001721, -0.00000077, -0.00001482, 0.00001578, 0.00000888, -0.00001831, 0.00000183, 0.00000902, -0.00001785, -0.00000307, 0.00001183, -0.00001333, -0.00000782, 0.00000543, 0.00000301, 0.00000061, -0.00001444, 0.00001455, 0.00001682, -0.00003349, 0.00001119, 0.00001774, -0.00003448, 0.00001511, 0.00000411, -0.00001990, 0.00002792, -0.00001513, -0.00000633, 0.00002303, -0.00003132, 0.00001945, 0.00001408, -0.00004169, 0.00002449, -0.00000561, -0.00001920, 0.00002853, -0.00003407, 0.00000813, 0.00003172, -0.00004139, 0.00002368, 0.00000500, -0.00004187, 0.00002618, -0.00002153, -0.00001703, 0.00002903, -0.00003126, 0.00000188, -0.00001338, -0.00004671, -0.00000215, -0.00016824, -0.00013112, 0.00032272, 0.00041318, 0.00005951, -0.00015580, -0.00010312, 0.00000175, -0.00001578, -0.00003476, 0.00000782, -0.00000958, -0.00001171, 0.00000763, -0.00001070, 0.00000587, 0.00000498, -0.00002085, -0.00000188, -0.00000041, -0.00000885, 0.00000506, 0.00000010, -0.00000455, 0.00000146, 0.00000204, 0.00000273, 0.00000085, 0.00000107, 0.00000006, 0.00000030, 0.00000820, 0.00000395, -0.00000366, },
		{ -0.00000529, 0.00000981, 0.00000950, -0.00000358, 0.00000024, -0.00000087, -0.00000257, 0.00000427, -0.00000314, -0.00000740, -0.00000031, -0.00000000, 0.00000022, -0.00000077, -0.00000199, 0.00000124, -0.00000137, -0.00000507, -0.00000774, -0.00000612, 0.00000584, 0.00000852, 0.00000105, -0.00000644, -0.00001294, -0.00000483, 0.00000672, 0.00000659, 0.00000198, -0.00000386, -0.00000334, -0.00000011, 0.00000460, 0.00000934, 0.00000070, 0.00000190, 0.00000285, -0.00001433, 0.00000186, 0.00001551, -0.00001302, -0.00000306, 0.00002022, -0.00000758, -0.00001270, 0.00001221, -0.00000514, -0.00001681, 0.00000868, 0.00000084, -0.00001412, -0.00000032, 0.00000179, 0.00000644, -0.00000636, -0.00001238, 0.00002871, -0.00000485, -0.00002981, 0.00003151, -0.00000760, -0.00002573, 0.00002737, -0.00001663, -0.00000081, 0.00002256, -0.00002925, 0.00001712, 0.00000378, -0.00002731, 0.00003855, -0.00001769, -0.00002436, 0.00003236, -0.00002936, 0.00000673, 0.00001302, -0.00003991, 0.00003972, -0.00000175, -0.00003329, 0.00004405, -0.00003114, -0.00001921, 0.00002808, -0.00004461, 0.00001508, 0.00001185, -0.00003980, 0.00002212, -0.00004298, -0.00002991, -0.00002156, -0.00023381, -0.00000518, 0.00046990, 0.00032280, -0.00008307, -0.00016269, -0.00004836, 0.00000456, -0.00003278, -0.00001757, 0.00000765, -0.00001744, -0.00000061, 0.00000158, -0.00001003, 0.00001265, -0.00000753, -0.00001793, 0.00000492, -0.00000692, -0.00000502, 0.00000706, -0.00000358, -0.00000216, 0.00000209, 0.00000151, 0.00000263, 0.00000034, 0.00000104, -0.00000036, 0.00000324, 0.00000851, -0.00000026, -0.00000393, -0.00000491, },
		{ 0.00001334, 0.00000496, -0.00000397, 0.00000173, -0.00000307, -0.00000058, 0.00000373, -0.00000652, -0.00000518, 0.00000076, -0.00000056, 0.00000092, -0.00000191, -0.00000195, 0.00000180, -0.00000256, -0.00000680, -0.00000860, -0.00000218, 0.00000962, 0.00000646, -0.00000311, -0.00001000, -0.00001129, 0.00000053, 0.00000847, 0.00000521, -0.00000093, -0.00000447, -0.00000154, 0.00000098, 0.00000695, 0.00000752, -0.00000089, 0.00000395, -0.00000343, -0.00001210, 0.00001213, 0.00000490, -0.00001628, 0.00001189, 0.00001161, -0.00001771, 0.00000176, 0.00000969, -0.00001751, -0.00000531, 0.00000991, -0.00000957, -0.00000724, 0.00000007, 0.00000173, 0.00000848, -0.00001713, 0.00000150, 0.00003063, -0.00002847, -0.00000775, 0.00003347, -0.00003104, -0.00000202, 0.00002070, -0.00002643, 0.00002100, 0.00000091, -0.00002385, 0.00003105, -0.00002145, -0.00000362, 0.00003531, -0.00004208, 0.00000697, 0.00001625, -0.00003499, 0.00002999, -0.00001734, -0.00001879, 0.00005000, -0.00003869, 0.00000106, 0.00003507, -0.00005637, 0.00001751, 0.00000273, -0.00004410, 0.00004331, -0.00002482, -0.00002137, 0.00002427, -0.00007374, 0.00000567, -0.00008368, -0.00026238, 0.00019480, 0.00054547, 0.00015323, -0.00018072, -0.00011801, -0.00001152, -0.00001114, -0.00003398, -0.00000204, -0.00000160, -0.00001563, 0.00000613, -0.00000703, -0.00000224, 0.00001128, -0.00001838, -0.00000816, 0.00000481, -0.00001179, 0.00000174, 0.00000515, -0.00000537, 0.00000070, 0.00000109, 0.00000178, 0.00000234, -0.00000021, 0.00000102, 0.00000022, 0.00000620, 0.00000629, -0.00000352, -0.00000313, -0.00000607, -0.00000205, },
		{ -0.00000000, -0.00000241, 0.00000164, -0.00000458, 0.00000193, 0.00000122, -0.00000823, -0.00000228, 0.00000047, -0.00000029, 0.00000109, -0.00000358, -0.00000083, 0.00000197, -0.00000461, -0.00000851, -0.00000769, 0.00000304, 0.00001130, 0.00000230, -0.00000719, -0.00001154, -0.00000764, 0.00000521, 0.00000872, 0.00000263, -0.00000331, -0.00000344, -0.00000019, 0.00000237, 0.00000856, 0.00000457, -0.00000049, 0.00000340, -0.00000892, -0.00000375, 0.00001492, -0.00000819, -0.00000810, 0.00001995, -0.00000459, -0.00001456, 0.00001377, -0.00000376, -0.00001884, 0.00000690, 0.00000124, -0.00001173, -0.00000070, -0.00000379, 0.00000747, 0.00000277, -0.00002266, 0.00002154, 0.00001496, -0.00003889, 0.00002067, 0.00001439, -0.00003740, 0.00002156, -0.00000034, -0.00001741, 0.00002998, -0.00002335, -0.00000109, 0.00002444, -0.00003514, 0.00002624, 0.00000903, -0.00004289, 0.00003217, -0.00001392, -0.00001648, 0.00003256, -0.00004226, 0.00001979, 0.00002800, -0.00005356, 0.00004101, -0.00000333, -0.00005043, 0.00004316, -0.00003689, -0.00001237, 0.00004380, -0.00005809, 0.00001854, -0.00000692, -0.00007941, 0.00003674, -0.00018957, -0.00020402, 0.00042920, 0.00050025, -0.00004599, -0.00020095, -0.00005544, -0.00000512, -0.00002593, -0.00002260, 0.00000318, -0.00001053, -0.00000716, 0.00000468, -0.00001124, 0.00000719, 0.00000164, -0.00002140, 0.00000236, -0.00000163, -0.00001124, 0.00000756, 0.00000061, -0.00000420, 0.00000204, -0.00000025, 0.00000293, 0.00000124, -0.00000040, 0.00000127, 0.00000175, 0.00000796, 0.00000227, -0.00000479, -0.00000248, -0.00000721, 0.00000303, 0.00001433, },
		{ 0.00000022, -0.00000008, -0.00000462, 0.00000376, -0.00000237, -0.00000778, -0.00000001, -0.00000036, 0.00000077, -0.00000002, -0.00000476, 0.00000118, 0.00000096, -0.00000720, -0.00000942, -0.00000471, 0.00000823, 0.00000998, -0.00000293, -0.00001002, -0.00001105, -0.00000289, 0.00000870, 0.00000725, -0.00000052, -0.00000416, -0.00000172, 0.00000062, 0.00000427, 0.00000871, 0.00000183, 0.00000077, 0.00000001, -0.00000993, 0.00000612, 0.00000842, -0.00001534, 0.00000646, 0.00001526, -0.00001751, -0.00000048, 0.00001391, -0.00001818, -0.00000856, 0.00001044, -0.00000926, -0.00000562, -0.00000068, -0.00000486, 0.00001412, -0.00001071, -0.00001523, 0.00003524, -0.00001193, -0.00002778, 0.00003809, -0.00001468, -0.00002231, 0.00002976, -0.00002101, 0.00000504, 0.00001851, -0.00003354, 0.00002445, 0.00000007, -0.00002600, 0.00004265, -0.00002548, -0.00001847, 0.00003395, -0.00003748, 0.00001505, 0.00000944, -0.00004180, 0.00005173, -0.00001619, -0.00003241, 0.00005809, -0.00004957, -0.00001228, 0.00003573, -0.00006289, 0.00003541, 0.00000717, -0.00005993, 0.00005301, -0.00006427, -0.00004104, 0.00002297, -0.00030415, -0.00002529, 0.00062064, 0.00032492, -0.00020446, -0.00014843, -0.00001253, -0.00001914, -0.00002711, -0.00001025, -0.00000081, -0.00001237, 0.00000055, -0.00000216, -0.00000792, 0.00001140, -0.00001085, -0.00001531, 0.00000736, -0.00000975, -0.00000513, 0.00000925, -0.00000356, -0.00000121, 0.00000104, -0.00000027, 0.00000391, -0.00000057, 0.00000026, 0.00000171, 0.00000383, 0.00000768, -0.00000209, -0.00000396, -0.00000300, -0.00000709, 0.00000905, 0.00001213, -0.00000383, },
		{ -0.00000281, -0.00000310, 0.00000403, -0.00000576, -0.00000563, 0.00000094, -0.00000076, 0.00000181, -0.00000236, -0.00000447, 0.00000318, -0.00000156, -0.00000962, -0.00000884, 0.00000005, 0.00001174, 0.00000572, -0.00000771, -0.00001115, -0.00000891, 0.00000231, 0.00001039, 0.00000418, -0.00000308, -0.00000341, -0.00000026, 0.00000136, 0.00000625, 0.00000736, 0.00000023, 0.00000114, -0.00000412, -0.00000530, 0.00001144, -0.00000328, -0.00001170, 0.00001775, 0.00000041, -0.00001824, 0.00001437, 0.00000132, -0.00002310, 0.00000532, 0.00000318, -0.00001295, 0.00000194, -0.00000647, 0.00000132, 0.00001384, -0.00002424, 0.00000459, 0.00003182, -0.00003490, -0.00000014, 0.00003325, -0.00003586, 0.00000456, 0.00001775, -0.00002657, 0.00002549, -0.00000706, -0.00002162, 0.00003526, -0.00002636, 0.00000222, 0.00003316, -0.00004656, 0.00001597, 0.00001020, -0.00003690, 0.00003773, -0.00002580, -0.00001138, 0.00005342, -0.00005507, 0.00001508, 0.00003555, -0.00007250, 0.00003540, -0.00000591, -0.00005087, 0.00006685, -0.00004940, -0.00001876, 0.00004840, -0.00011398, 0.00003092, -0.00006828, -0.00035762, 0.00025864, 0.00068243, 0.00006684, -0.00026533, -0.00006349, -0.00000775, -0.00003180, -0.00001668, -0.00000608, -0.00000611, -0.00000779, 0.00000216, -0.00000790, 0.00000000, 0.00000721, -0.00001893, -0.00000439, 0.00000447, -0.00001404, 0.00000314, 0.00000640, -0.00000494, 0.00000114, -0.00000114, 0.00000159, 0.00000336, -0.00000205, 0.00000173, 0.00000221, 0.00000587, 0.00000512, -0.00000520, -0.00000199, -0.00000496, -0.00000454, 0.00001429, 0.00000707, -0.00000525, 0.00000262, },
		{ -0.00000047, 0.00000251, -0.00000786, -0.00000282, 0.00000071, -0.00000025, 0.00000181, -0.00000501, -0.00000239, 0.00000398, -0.00000526, -0.00001104, -0.00000632, 0.00000552, 0.00001218, -0.00000027, -0.00001083, -0.00001083, -0.00000539, 0.00000717, 0.00000969, 0.00000036, -0.00000408, -0.00000183, 0.00000059, 0.00000251, 0.00000756, 0.00000523, -0.00000038, -0.00000023, -0.00000585, 0.00000228, 0.00000873, -0.00001252, 0.00000059, 0.00001793, -0.00001504, -0.00000602, 0.00001914, -0.00001582, -0.00001501, 0.00001252, -0.00000865, -0.00000727, 0.00000346, -0.00001141, 0.00001219, 0.00000274, -0.00002743, 0.00002669, 0.00001046, -0.00003962, 0.00002811, 0.00000877, -0.00003624, 0.00002596, -0.00000551, -0.00001280, 0.00002907, -0.00002999, 0.00000560, 0.00002311, -0.00003645, 0.00003204, 0.00000185, -0.00004015, 0.00003767, -0.00002349, -0.00001070, 0.00003386, -0.00004854, 0.00003238, 0.00001903, -0.00006083, 0.00005888, -0.00001761, -0.00005276, 0.00005896, -0.00005674, 0.00000002, 0.00005326, -0.00008806, 0.00004571, -0.00000988, -0.00011294, 0.00008913, -0.00022949, -0.00027442, 0.00056989, 0.00056322, -0.00018282, -0.00021694, 0.00000122, -0.00003073, -0.00002746, -0.00000699, -0.00000953, -0.00000602, -0.00000284, -0.00000159, -0.00000796, 0.00000623, -0.00000269, -0.00001848, 0.00000455, -0.00000357, -0.00001179, 0.00000912, 0.00000143, -0.00000341, 0.00000106, -0.00000226, 0.00000415, 0.00000093, -0.00000197, 0.00000329, 0.00000282, 0.00000708, 0.00000086, -0.00000585, -0.00000046, -0.00000752, 0.00000092, 0.00001674, 0.00000059, -0.00000378, 0.00000351, -0.00000556, },
		{ -0.00000045, -0.00000814, -0.00000038, -0.00000001, 0.00000091, 0.00000022, -0.00000668, 0.00000086, 0.00000268, -0.00000927, -0.00001069, -0.00000200, 0.00001002, 0.00000919, -0.00000616, -0.00001200, -0.00000938, -0.00000069, 0.00001059, 0.00000663, -0.00000283, -0.00000343, -0.00000036, 0.00000111, 0.00000401, 0.00000781, 0.00000311, -0.00000099, -0.00000240, -0.00000338, 0.00000778, -0.00000035, -0.00001298, 0.00001344, 0.00000635, -0.00002079, 0.00001105, 0.00000970, -0.00002614, 0.00000038, 0.00000806, -0.00001569, 0.00000264, -0.00000304, -0.00000874, 0.00001962, -0.00001484, -0.00001517, 0.00003744, -0.00001794, -0.00002235, 0.00004020, -0.00002038, -0.00001637, 0.00002877, -0.00002357, 0.00001078, 0.00001212, -0.00003449, 0.00003049, -0.00000467, -0.00002200, 0.00004347, -0.00003227, -0.00000993, 0.00003161, -0.00004368, 0.00002424, 0.00000344, -0.00003988, 0.00006057, -0.00003268, -0.00002514, 0.00006719, -0.00006920, 0.00000117, 0.00003707, -0.00007941, 0.00006082, -0.00000702, -0.00007431, 0.00008830, -0.00009823, -0.00004104, 0.00006927, -0.00039536, -0.00002021, 0.00078869, 0.00028363, -0.00033005, -0.00010107, 0.00001240, -0.00005246, -0.00000830, -0.00000847, -0.00001219, -0.00000101, -0.00000274, -0.00000541, -0.00000322, 0.00000618, -0.00001195, -0.00001099, 0.00000676, -0.00001120, -0.00000450, 0.00001035, -0.00000248, -0.00000100, -0.00000110, -0.00000085, 0.00000529, -0.00000217, 0.00000005, 0.00000411, 0.00000378, 0.00000660, -0.00000370, -0.00000394, -0.00000091, -0.00000885, 0.00000837, 0.00001492, -0.00000501, -0.00000005, 0.00000201, -0.00000713, 0.00000238, },
		{ -0.00000666, 0.00000105, -0.00000044, 0.00000191, -0.00000269, -0.00000633, 0.00000396, -0.00000093, -0.00001233, -0.00000813, 0.00000320, 0.00001196, 0.00000374, -0.00001046, -0.00001168, -0.00000681, 0.00000463, 0.00001142, 0.00000219, -0.00000425, -0.00000186, 0.00000055, 0.00000177, 0.00000543, 0.00000714, 0.00000123, -0.00000226, -0.00000312, 0.00000203, 0.00000710, -0.00000971, -0.00000410, 0.00001814, -0.00000977, -0.00001280, 0.00002145, -0.00000851, -0.00002292, 0.00001250, -0.00000446, -0.00001253, 0.00000830, -0.00001234, 0.00000242, 0.00001619, -0.00002840, 0.00000767, 0.00002900, -0.00003713, 0.00000720, 0.00002928, -0.00003667, 0.00001040, 0.00001270, -0.00002398, 0.00002748, -0.00001477, -0.00001667, 0.00003646, -0.00002962, 0.00000862, 0.00002772, -0.00004738, 0.00002431, 0.00000137, -0.00003523, 0.00004343, -0.00003396, -0.00000132, 0.00005121, -0.00006826, 0.00003224, 0.00002887, -0.00008371, 0.00005501, -0.00002119, -0.00005028, 0.00008807, -0.00008054, -0.00000491, 0.00006630, -0.00016012, 0.00007218, -0.00006742, -0.00045971, 0.00035965, 0.00080647, -0.00006053, -0.00032649, 0.00000984, -0.00002471, -0.00004796, 0.00000862, -0.00001981, -0.00000657, 0.00000233, -0.00000686, -0.00000481, 0.00000129, 0.00000043, -0.00001535, -0.00000196, 0.00000230, -0.00001387, 0.00000372, 0.00000750, -0.00000369, -0.00000007, -0.00000317, 0.00000249, 0.00000366, -0.00000392, 0.00000306, 0.00000396, 0.00000505, 0.00000389, -0.00000667, -0.00000070, -0.00000379, -0.00000695, 0.00001546, 0.00000881, -0.00000740, 0.00000409, -0.00000177, -0.00000657, 0.00000435, -0.00000401, },
		{ 0.00000132, -0.00000019, 0.00000183, -0.00000587, -0.00000370, 0.00000532, -0.00000615, -0.00001317, -0.00000358, 0.00000774, 0.00001070, -0.00000241, -0.00001267, -0.00001041, -0.00000288, 0.00000932, 0.00000920, -0.00000194, -0.00000372, -0.00000036, 0.00000106, 0.00000269, 0.00000639, 0.00000592, -0.00000083, -0.00000362, -0.00000074, 0.00000651, 0.00000021, -0.00001282, 0.00000839, 0.00001117, -0.00002004, 0.00000397, 0.00001785, -0.00002429, -0.00000853, 0.00001340, -0.00001580, -0.00000163, 0.00000466, -0.00001625, 0.00001576, 0.00000134, -0.00002842, 0.00002855, 0.00000501, -0.00003588, 0.00003182, 0.00000274, -0.00003132, 0.00002713, -0.00001002, -0.00000694, 0.00002508, -0.00003377, 0.00001239, 0.00001945, -0.00003466, 0.00003551, -0.00000609, -0.00003362, 0.00003958, -0.00003260, -0.00000250, 0.00003197, -0.00005156, 0.00004375, 0.00000590, -0.00006142, 0.00007389, -0.00003594, -0.00004736, 0.00006998, -0.00007799, 0.00001993, 0.00005370, -0.00011582, 0.00008119, -0.00002619, -0.00013941, 0.00014978, -0.00029272, -0.00032757, 0.00073875, 0.00058625, -0.00033706, -0.00019665, 0.00005435, -0.00007222, -0.00001537, 0.00000642, -0.00002862, 0.00000597, -0.00000183, -0.00000950, -0.00000043, 0.00000118, -0.00000620, -0.00001233, 0.00000330, -0.00000482, -0.00001071, 0.00000916, 0.00000324, -0.00000302, -0.00000132, -0.00000294, 0.00000545, -0.00000012, -0.00000291, 0.00000546, 0.00000349, 0.00000593, -0.00000060, -0.00000661, 0.00000162, -0.00000787, -0.00000090, 0.00001913, 0.00000041, -0.00000542, 0.00000625, -0.00000652, -0.00000363, 0.00000443, -0.00000711, -0.00000399, },
		{ 0.00000067, 0.00000014, -0.00000791, 0.00000043, 0.00000375, -0.00001143, -0.00001097, 0.00000182, 0.00001016, 0.00000683, -0.00000765, -0.00001315, -0.00000819, 0.00000226, 0.00001178, 0.00000472, -0.00000423, -0.00000204, 0.00000055, 0.00000155, 0.00000366, 0.00000687, 0.00000421, -0.00000333, -0.00000339, 0.00000406, 0.00000617, -0.00000850, -0.00000713, 0.00001611, -0.00000330, -0.00001769, 0.00001875, 0.00000197, -0.00002816, 0.00000749, 0.00000310, -0.00001791, 0.00000889, -0.00000610, -0.00001001, 0.00002204, -0.00001736, -0.00001324, 0.00003514, -0.00002114, -0.00001524, 0.00003775, -0.00002298, -0.00000949, 0.00002461, -0.00002346, 0.00001510, 0.00000451, -0.00003193, 0.00003400, -0.00000918, -0.00001596, 0.00004055, -0.00003651, -0.00000025, 0.00002537, -0.00004654, 0.00003275, -0.00000401, -0.00003415, 0.00006440, -0.00004842, -0.00001238, 0.00006898, -0.00008635, 0.00001940, 0.00003054, -0.00009025, 0.00008754, -0.00003057, -0.00007817, 0.00012163, -0.00014273, -0.00002442, 0.00010692, -0.00050212, 0.00002154, 0.00095687, 0.00019464, -0.00044154, -0.00002547, 0.00001752, -0.00008720, 0.00002265, -0.00001640, -0.00002222, 0.00001539, -0.00001212, -0.00000563, 0.00000253, -0.00000330, -0.00000899, -0.00000664, 0.00000309, -0.00001002, -0.00000412, 0.00001062, -0.00000016, -0.00000244, -0.00000330, -0.00000006, 0.00000573, -0.00000376, 0.00000048, 0.00000614, 0.00000354, 0.00000525, -0.00000522, -0.00000349, 0.00000100, -0.00001045, 0.00000808, 0.00001714, -0.00000685, 0.00000003, 0.00000464, -0.00001005, 0.00000082, 0.00000216, -0.00000864, -0.00000106, 0.00000079, },
		{ -0.00000297, -0.00000759, 0.00000436, -0.00000097, -0.00001472, -0.00000594, 0.00000636, 0.00000985, 0.00000180, -0.00001127, -0.00001246, -0.00000462, 0.00000752, 0.00001094, -0.00000017, -0.00000417, -0.00000033, 0.00000101, 0.00000219, 0.00000449, 0.00000697, 0.00000163, -0.00000556, -0.00000014, 0.00000815, 0.00000008, -0.00001294, 0.00000393, 0.00001367, -0.00001607, -0.00000411, 0.00002202, -0.00001679, -0.00001844, 0.00001514, -0.00001090, -0.00000951, 0.00001129, -0.00001619, 0.00000386, 0.00001588, -0.00002893, 0.00000926, 0.00002348, -0.00003476, 0.00001234, 0.00002308, -0.00003317, 0.00001396, 0.00000682, -0.00001905, 0.00002633, -0.00002077, -0.00001007, 0.00003453, -0.00003025, 0.00001420, 0.00001991, -0.00004402, 0.00003023, -0.00000878, -0.00002994, 0.00004597, -0.00004023, 0.00000972, 0.00004338, -0.00007567, 0.00004940, 0.00001560, -0.00008727, 0.00007255, -0.00004118, -0.00004076, 0.00010225, -0.00011375, 0.00002043, 0.00007189, -0.00020459, 0.00012754, -0.00008969, -0.00055448, 0.00049771, 0.00089806, -0.00021870, -0.00035184, 0.00008915, -0.00006262, -0.00005229, 0.00003623, -0.00004294, 0.00000039, 0.00001134, -0.00001990, 0.00000353, -0.00000061, -0.00000745, -0.00000722, -0.00000270, -0.00000039, -0.00001114, 0.00000260, 0.00000923, -0.00000227, -0.00000309, -0.00000379, 0.00000368, 0.00000290, -0.00000493, 0.00000443, 0.00000532, 0.00000422, 0.00000227, -0.00000771, 0.00000082, -0.00000296, -0.00000879, 0.00001665, 0.00000963, -0.00000961, 0.00000607, -0.00000070, -0.00001032, 0.00000489, -0.00000189, -0.00000797, 0.00000116, 0.00000010, 0.00000149, },
		{ -0.00000456, 0.00000600, -0.00000760, -0.00001426, 0.00000045, 0.00000856, 0.00000741, -0.00000304, -0.00001340, -0.00001061, 0.00000041, 0.00001115, 0.00000711, -0.00000352, -0.00000247, 0.00000068, 0.00000143, 0.00000279, 0.00000525, 0.00000647, -0.00000200, -0.00000576, 0.00000556, 0.00000784, -0.00000856, -0.00000955, 0.00001313, 0.00000235, -0.00001897, 0.00001213, 0.00001158, -0.00002759, -0.00000190, 0.00001045, -0.00001912, 0.00000342, 0.00000379, -0.00001801, 0.00001708, -0.00000022, -0.00002628, 0.00002650, 0.00000048, -0.00002896, 0.00003110, -0.00000179, -0.00002389, 0.00002474, -0.00001266, -0.00000109, 0.00001866, -0.00003399, 0.00001786, 0.00001447, -0.00002996, 0.00003568, -0.00001313, -0.00002431, 0.00003725, -0.00003946, 0.00000679, 0.00002715, -0.00005054, 0.00005174, -0.00000920, -0.00005496, 0.00008283, -0.00005506, -0.00003457, 0.00007331, -0.00009651, 0.00004495, 0.00004317, -0.00013551, 0.00011983, -0.00005678, -0.00015089, 0.00020926, -0.00037888, -0.00034820, 0.00092197, 0.00055850, -0.00048964, -0.00014017, 0.00009166, -0.00012206, 0.00001182, 0.00001180, -0.00004993, 0.00002541, -0.00000689, -0.00001626, 0.00001034, -0.00000902, -0.00000663, -0.00000435, -0.00000202, -0.00000380, -0.00000892, 0.00000762, 0.00000653, -0.00000404, -0.00000427, -0.00000180, 0.00000568, -0.00000133, -0.00000290, 0.00000699, 0.00000417, 0.00000459, -0.00000234, -0.00000660, 0.00000343, -0.00000828, -0.00000193, 0.00002079, -0.00000048, -0.00000654, 0.00000910, -0.00000764, -0.00000668, 0.00000654, -0.00000624, -0.00000534, 0.00000205, -0.00000013, 0.00000139, -0.00000643, },
		{ 0.00000383, -0.00001363, -0.00000958, 0.00000588, 0.00000803, 0.00000416, -0.00000716, -0.00001437, -0.00000714, 0.00000605, 0.00001161, 0.00000199, -0.00000431, -0.00000047, 0.00000106, 0.00000200, 0.00000319, 0.00000609, 0.00000470, -0.00000583, -0.00000236, 0.00001080, 0.00000170, -0.00001423, 0.00000009, 0.00001424, -0.00001068, -0.00001020, 0.00002089, -0.00000634, -0.00002501, 0.00001115, -0.00000245, -0.00001655, 0.00001219, -0.00000851, -0.00000952, 0.00002135, -0.00001737, -0.00001088, 0.00002924, -0.00002060, -0.00000849, 0.00003179, -0.00002162, -0.00000341, 0.00001827, -0.00002056, 0.00001696, -0.00000288, -0.00002649, 0.00003437, -0.00001221, -0.00000914, 0.00003427, -0.00003708, 0.00000871, 0.00001620, -0.00004533, 0.00003904, -0.00001148, -0.00002552, 0.00006226, -0.00006046, 0.00000354, 0.00006265, -0.00009745, 0.00003912, 0.00001651, -0.00009228, 0.00011062, -0.00006067, -0.00006876, 0.00014576, -0.00019194, 0.00001082, 0.00012471, -0.00061301, 0.00010628, 0.00110409, 0.00006238, -0.00052174, 0.00006702, -0.00000146, -0.00011406, 0.00006095, -0.00003577, -0.00002627, 0.00003359, -0.00002760, -0.00000057, 0.00000684, -0.00001594, -0.00000084, -0.00000441, -0.00000263, -0.00000558, -0.00000518, 0.00001081, 0.00000301, -0.00000603, -0.00000425, 0.00000153, 0.00000464, -0.00000432, 0.00000110, 0.00000746, 0.00000366, 0.00000328, -0.00000648, -0.00000240, 0.00000219, -0.00001145, 0.00000817, 0.00001808, -0.00000859, 0.00000066, 0.00000688, -0.00001278, -0.00000051, 0.00000474, -0.00000902, -0.00000184, 0.00000166, 0.00000030, -0.00000031, -0.00000863, 0.00000023, },
		{ -0.00001619, -0.00000213, 0.00000828, 0.00000580, 0.00000111, -0.00001078, -0.00001389, -0.00000177, 0.00001049, 0.00000868, -0.00000217, -0.00000294, 0.00000082, 0.00000132, 0.00000252, 0.00000358, 0.00000680, 0.00000114, -0.00000784, 0.00000439, 0.00001170, -0.00000786, -0.00001273, 0.00001004, 0.00000635, -0.00001713, 0.00000459, 0.00001696, -0.00002139, -0.00001194, 0.00001354, -0.00001467, -0.00000549, 0.00001137, -0.00001739, 0.00000448, 0.00001393, -0.00002610, 0.00000835, 0.00001713, -0.00002857, 0.00001402, 0.00001663, -0.00002614, 0.00001435, 0.00000156, -0.00001283, 0.00002218, -0.00002391, -0.00000325, 0.00003008, -0.00002787, 0.00001763, 0.00001127, -0.00003688, 0.00003242, -0.00001835, -0.00002186, 0.00004488, -0.00004329, 0.00001968, 0.00003123, -0.00007581, 0.00006306, -0.00000194, -0.00008197, 0.00008422, -0.00006230, -0.00002290, 0.00010560, -0.00014303, 0.00005426, 0.00006126, -0.00023833, 0.00019030, -0.00013982, -0.00062499, 0.00066445, 0.00094039, -0.00039019, -0.00033554, 0.00015975, -0.00011612, -0.00004020, 0.00005908, -0.00007216, 0.00001637, 0.00001546, -0.00003431, 0.00001699, -0.00000764, -0.00001400, 0.00000454, -0.00000787, -0.00000180, -0.00000653, -0.00000065, 0.00001233, -0.00000171, -0.00000739, -0.00000218, 0.00000393, 0.00000145, -0.00000450, 0.00000497, 0.00000657, 0.00000357, -0.00000011, -0.00000781, 0.00000231, -0.00000272, -0.00000946, 0.00001722, 0.00000933, -0.00001092, 0.00000794, -0.00000006, -0.00001339, 0.00000535, 0.00000013, -0.00000905, 0.00000108, 0.00000074, 0.00000085, -0.00000348, -0.00000811, 0.00000453, -0.00000211, },
		{ 0.00000507, 0.00000709, 0.00000360, -0.00000177, -0.00001404, -0.00001114, 0.00000468, 0.00001191, 0.00000378, -0.00000388, -0.00000079, 0.00000124, 0.00000182, 0.00000271, 0.00000432, 0.00000652, -0.00000366, -0.00000590, 0.00001172, 0.00000613, -0.00001562, -0.00000427, 0.00001389, -0.00000556, -0.00001303, 0.00001612, 0.00000296, -0.00002574, 0.00000290, 0.00000540, -0.00001872, 0.00000663, 0.00000209, -0.00001724, 0.00001580, -0.00000084, -0.00002233, 0.00002097, -0.00000173, -0.00002082, 0.00002641, -0.00000342, -0.00001571, 0.00001935, -0.00001277, 0.00000346, 0.00001105, -0.00003075, 0.00002095, 0.00000948, -0.00002322, 0.00003232, -0.00001772, -0.00001394, 0.00003094, -0.00004269, 0.00001550, 0.00002041, -0.00004555, 0.00005468, -0.00002343, -0.00004270, 0.00008366, -0.00007115, -0.00001659, 0.00006765, -0.00010819, 0.00007081, 0.00002230, -0.00014228, 0.00015446, -0.00009861, -0.00014177, 0.00025563, -0.00048107, -0.00032428, 0.00109949, 0.00047731, -0.00062000, -0.00005533, 0.00010435, -0.00016999, 0.00005135, 0.00000506, -0.00006791, 0.00004981, -0.00001940, -0.00001878, 0.00002195, -0.00002408, -0.00000205, 0.00000317, -0.00001084, 0.00000068, -0.00000781, 0.00000511, 0.00001125, -0.00000729, -0.00000648, 0.00000088, 0.00000394, -0.00000169, -0.00000218, 0.00000733, 0.00000547, 0.00000273, -0.00000430, -0.00000542, 0.00000438, -0.00000846, -0.00000196, 0.00002088, -0.00000145, -0.00000667, 0.00001105, -0.00000852, -0.00000898, 0.00000813, -0.00000517, -0.00000653, 0.00000246, 0.00000015, 0.00000054, -0.00000693, -0.00000458, 0.00000577, -0.00000970, -0.00001348, },
		{ 0.00000380, 0.00000259, -0.00000531, -0.00001621, -0.00000557, 0.00001024, 0.00000968, -0.00000080, -0.00000308, 0.00000083, 0.00000133, 0.00000240, 0.00000265, 0.00000550, 0.00000425, -0.00000772, 0.00000058, 0.00001536, -0.00000437, -0.00001673, 0.00000637, 0.00000913, -0.00001384, -0.00000126, 0.00001753, -0.00001271, -0.00001855, 0.00001099, -0.00000686, -0.00001299, 0.00001254, -0.00000935, -0.00000842, 0.00001832, -0.00001473, -0.00000940, 0.00002142, -0.00001647, -0.00000381, 0.00002413, -0.00001650, 0.00000047, 0.00001123, -0.00001551, 0.00001594, -0.00000864, -0.00001939, 0.00003172, -0.00001291, -0.00000299, 0.00002579, -0.00003369, 0.00001518, 0.00000579, -0.00004019, 0.00004198, -0.00001741, -0.00001565, 0.00005445, -0.00006647, 0.00001932, 0.00004923, -0.00009999, 0.00005632, -0.00000263, -0.00008421, 0.00012519, -0.00009221, -0.00004641, 0.00015445, -0.00023724, 0.00006178, 0.00011364, -0.00071189, 0.00023239, 0.00120924, -0.00009988, -0.00055855, 0.00016095, -0.00004280, -0.00012529, 0.00009915, -0.00006500, -0.00002082, 0.00004914, -0.00004718, 0.00001081, 0.00000701, -0.00002938, 0.00001239, -0.00000626, -0.00000851, 0.00000185, -0.00000853, 0.00001192, 0.00000607, -0.00001160, -0.00000281, 0.00000274, 0.00000208, -0.00000306, 0.00000107, 0.00000821, 0.00000448, 0.00000024, -0.00000697, -0.00000075, 0.00000224, -0.00001121, 0.00000818, 0.00001731, -0.00000921, 0.00000143, 0.00000803, -0.00001439, -0.00000166, 0.00000666, -0.00000879, -0.00000283, 0.00000236, 0.00000015, -0.00000137, -0.00000875, 0.00000051, 0.00000242, -0.00001537, -0.00000610, 0.00000887, },
		{ 0.00000230, -0.00001011, -0.00001574, 0.00000210, 0.00001272, 0.00000492, -0.00000317, -0.00000102, 0.00000145, 0.00000164, 0.00000265, 0.00000284, 0.00000640, -0.00000016, -0.00000834, 0.00000944, 0.00001210, -0.00001487, -0.00001001, 0.00001280, -0.00000115, -0.00001333, 0.00001060, 0.00000845, -0.00002157, -0.00000565, 0.00000902, -0.00001533, -0.00000194, 0.00000960, -0.00001611, 0.00000363, 0.00001153, -0.00002098, 0.00000482, 0.00001168, -0.00002022, 0.00001205, 0.00001171, -0.00001718, 0.00001157, -0.00000197, -0.00000666, 0.00001585, -0.00002377, 0.00000245, 0.00002428, -0.00002289, 0.00001811, 0.00000352, -0.00002724, 0.00003041, -0.00002551, -0.00001250, 0.00004049, -0.00004244, 0.00002652, 0.00001706, -0.00006870, 0.00007030, -0.00002026, -0.00006864, 0.00008735, -0.00008020, 0.00000041, 0.00009654, -0.00016235, 0.00009070, 0.00003418, -0.00025310, 0.00024973, -0.00021639, -0.00065515, 0.00084359, 0.00092338, -0.00055338, -0.00028000, 0.00020859, -0.00017547, -0.00001144, 0.00007076, -0.00010184, 0.00004038, 0.00001177, -0.00004643, 0.00003374, -0.00002057, -0.00001654, 0.00001779, -0.00001766, -0.00000016, -0.00000148, -0.00000573, 0.00001721, -0.00000312, -0.00001175, 0.00000178, 0.00000210, 0.00000026, -0.00000264, 0.00000398, 0.00000833, 0.00000286, -0.00000332, -0.00000641, 0.00000319, -0.00000292, -0.00000857, 0.00001633, 0.00000835, -0.00001063, 0.00000869, 0.00000028, -0.00001497, 0.00000510, 0.00000198, -0.00000947, 0.00000041, 0.00000161, 0.00000013, -0.00000454, -0.00000752, 0.00000431, -0.00000464, -0.00001586, 0.00000279, 0.00000781, 0.00000103, },
		{ -0.00001533, -0.00001123, 0.00000957, 0.00001103, -0.00000001, -0.00000287, 0.00000083, 0.00000148, 0.00000221, 0.00000245, 0.00000370, 0.00000578, -0.00000524, -0.00000387, 0.00001631, 0.00000215, -0.00001978, 0.00000098, 0.00001122, -0.00001025, -0.00000469, 0.00001522, -0.00000528, -0.00002006, 0.00000488, 0.00000011, -0.00001560, 0.00000757, 0.00000071, -0.00001498, 0.00001235, 0.00000002, -0.00001806, 0.00001327, -0.00000113, -0.00001343, 0.00001920, -0.00000170, -0.00000853, 0.00001222, -0.00001045, 0.00000579, 0.00000371, -0.00002488, 0.00002125, 0.00000568, -0.00001583, 0.00002604, -0.00001892, -0.00000444, 0.00002187, -0.00004168, 0.00002202, 0.00001319, -0.00003755, 0.00005193, -0.00003405, -0.00002724, 0.00007604, -0.00008080, 0.00000296, 0.00005385, -0.00011014, 0.00009237, -0.00000554, -0.00013387, 0.00017761, -0.00014489, -0.00011076, 0.00027716, -0.00058626, -0.00025047, 0.00124851, 0.00034960, -0.00071045, 0.00004391, 0.00008936, -0.00020566, 0.00009664, -0.00001474, -0.00007730, 0.00007475, -0.00003861, -0.00001467, 0.00003121, -0.00004217, 0.00000844, 0.00000764, -0.00002146, 0.00000894, -0.00000873, 0.00000274, 0.00001673, -0.00001304, -0.00000654, 0.00000409, -0.00000004, -0.00000017, -0.00000155, 0.00000647, 0.00000784, -0.00000017, -0.00000602, -0.00000300, 0.00000388, -0.00000778, -0.00000126, 0.00001880, -0.00000151, -0.00000597, 0.00001121, -0.00000836, -0.00001037, 0.00000848, -0.00000350, -0.00000753, 0.00000229, 0.00000097, -0.00000092, -0.00000739, -0.00000345, 0.00000410, -0.00001210, -0.00001021, 0.00000895, 0.00000321, 0.00000084, 0.00000092, },
		{ -0.00000290, 0.00001396, 0.00000608, -0.00000287, -0.00000092, 0.00000166, 0.00000158, 0.00000263, 0.00000217, 0.00000493, 0.00000283, -0.00000825, 0.00000473, 0.00001691, -0.00001039, -0.00001632, 0.00001002, 0.00000296, -0.00001240, 0.00000627, 0.00001021, -0.00001574, -0.00001104, 0.00000780, -0.00000918, -0.00000877, 0.00001074, -0.00000843, -0.00000766, 0.00001413, -0.00001012, -0.00000948, 0.00001357, -0.00000992, -0.00000204, 0.00001677, -0.00000879, 0.00000155, 0.00000492, -0.00000948, 0.00001234, -0.00001185, -0.00001206, 0.00002687, -0.00001114, 0.00000124, 0.00001675, -0.00002697, 0.00001803, -0.00000388, -0.00003215, 0.00004116, -0.00002054, -0.00000651, 0.00004247, -0.00006537, 0.00003163, 0.00003140, -0.00009329, 0.00006734, -0.00002299, -0.00006714, 0.00012782, -0.00011901, -0.00001487, 0.00014437, -0.00026922, 0.00012058, 0.00006974, -0.00078097, 0.00038942, 0.00125579, -0.00027214, -0.00054773, 0.00024031, -0.00009911, -0.00011693, 0.00012914, -0.00009937, -0.00000480, 0.00005771, -0.00006724, 0.00002776, 0.00000102, -0.00004051, 0.00002908, -0.00001329, -0.00001227, 0.00001098, -0.00001441, 0.00001478, 0.00000781, -0.00001822, 0.00000156, 0.00000219, -0.00000113, 0.00000032, -0.00000050, 0.00000901, 0.00000586, -0.00000412, -0.00000601, 0.00000093, 0.00000109, -0.00000916, 0.00000729, 0.00001503, -0.00000783, 0.00000142, 0.00000796, -0.00001401, -0.00000316, 0.00000770, -0.00000753, -0.00000432, 0.00000296, 0.00000042, -0.00000336, -0.00000795, 0.00000119, -0.00000087, -0.00001573, -0.00000081, 0.00000942, -0.00000146, 0.00000304, -0.00000332, -0.00001858, },
		{ 0.00001340, 0.00000038, -0.00000285, 0.00000107, 0.00000168, 0.00000203, 0.00000254, 0.00000240, 0.00000546, -0.00000184, -0.00000670, 0.00001372, 0.00000971, -0.00001954, -0.00000639, 0.00001213, -0.00000640, -0.00000648, 0.00001247, -0.00000086, -0.00001798, -0.00000107, 0.00000334, -0.00001341, 0.00000029, 0.00000722, -0.00001320, 0.00000130, 0.00000962, -0.00001506, -0.00000054, 0.00000824, -0.00001173, 0.00000729, 0.00000939, -0.00000813, 0.00000646, -0.00000328, -0.00000178, 0.00000866, -0.00002065, 0.00000613, 0.00001849, -0.00001636, 0.00001563, -0.00000191, -0.00001690, 0.00002475, -0.00002899, -0.00000361, 0.00003386, -0.00003786, 0.00002888, 0.00000363, -0.00005597, 0.00006951, -0.00003556, -0.00004992, 0.00008110, -0.00009092, 0.00002461, 0.00007634, -0.00016733, 0.00012224, -0.00000538, -0.00024391, 0.00029332, -0.00031102, -0.00063363, 0.00101337, 0.00084644, -0.00068678, -0.00019542, 0.00022746, -0.00022902, 0.00002969, 0.00006744, -0.00012550, 0.00006885, -0.00000075, -0.00005270, 0.00005058, -0.00003858, -0.00001305, 0.00002959, -0.00003088, 0.00000566, 0.00000223, -0.00001166, 0.00002362, -0.00000721, -0.00001465, 0.00000732, -0.00000249, 0.00000054, -0.00000001, 0.00000122, 0.00001124, 0.00000150, -0.00000700, -0.00000323, 0.00000276, -0.00000301, -0.00000616, 0.00001320, 0.00000761, -0.00000866, 0.00000735, 0.00000101, -0.00001475, 0.00000345, 0.00000394, -0.00000913, -0.00000113, 0.00000300, -0.00000083, -0.00000619, -0.00000542, 0.00000317, -0.00000833, -0.00001288, 0.00000760, 0.00000449, -0.00000274, 0.00000511, -0.00001030, -0.00001717, 0.00000683, },
		{ -0.00000333, -0.00000079, 0.00000206, 0.00000160, 0.00000253, 0.00000210, 0.00000328, 0.00000413, -0.00000602, -0.00000014, 0.00001833, -0.00000274, -0.00002087, 0.00000448, 0.00000677, -0.00001076, 0.00000331, 0.00001022, -0.00001119, -0.00001258, 0.00000417, -0.00000407, -0.00001120, 0.00000661, 0.00000025, -0.00001236, 0.00000773, 0.00000221, -0.00001462, 0.00000517, 0.00000178, -0.00000821, 0.00001138, 0.00000280, -0.00000352, 0.00000490, -0.00000649, 0.00000566, -0.00000208, -0.00001770, 0.00001910, 0.00000376, -0.00000922, 0.00001814, -0.00001667, 0.00000255, 0.00001187, -0.00003680, 0.00002529, 0.00000702, -0.00002819, 0.00004405, -0.00003909, -0.00001189, 0.00006149, -0.00008196, 0.00002007, 0.00003472, -0.00010159, 0.00010501, -0.00003490, -0.00011145, 0.00018354, -0.00018651, -0.00006194, 0.00026550, -0.00067732, -0.00013032, 0.00134791, 0.00019113, -0.00075000, 0.00014051, 0.00005014, -0.00022142, 0.00013901, -0.00004485, -0.00007470, 0.00009498, -0.00006167, -0.00000320, 0.00003501, -0.00006018, 0.00002419, 0.00000696, -0.00003137, 0.00002021, -0.00001255, 0.00000173, 0.00002178, -0.00002078, -0.00000348, 0.00000632, -0.00000571, 0.00000385, -0.00000203, 0.00000495, 0.00001132, -0.00000457, -0.00000672, 0.00000026, 0.00000166, -0.00000554, -0.00000060, 0.00001452, 0.00000036, -0.00000523, 0.00000928, -0.00000626, -0.00001127, 0.00000734, -0.00000081, -0.00000860, 0.00000156, 0.00000252, -0.00000331, -0.00000746, -0.00000115, 0.00000047, -0.00001392, -0.00000444, 0.00001056, -0.00000223, 0.00000038, 0.00000378, -0.00001742, -0.00001004, 0.00001422, 0.00000835, },
		{ 0.00000158, 0.00000204, 0.00000189, 0.00000263, 0.00000186, 0.00000413, 0.00000076, -0.00000697, 0.00000895, 0.00001542, -0.00001491, -0.00001418, 0.00001048, -0.00000194, -0.00000748, 0.00001043, 0.00000127, -0.00001529, -0.00000436, 0.00000304, -0.00000937, -0.00000506, 0.00000793, -0.00000620, -0.00000768, 0.00000995, -0.00000476, -0.00001099, 0.00000720, -0.00000271, -0.00000302, 0.00001125, -0.00000020, 0.00000010, 0.00000034, -0.00000380, 0.00000712, -0.00001226, -0.00000577, 0.00002108, -0.00000744, 0.00000294, 0.00000883, -0.00001836, 0.00001705, -0.00001106, -0.00002284, 0.00003701, -0.00002031, 0.00000010, 0.00002867, -0.00005764, 0.00003788, 0.00001281, -0.00007867, 0.00006990, -0.00004023, -0.00004439, 0.00011758, -0.00013535, 0.00001946, 0.00011632, -0.00028012, 0.00017574, -0.00000402, -0.00080466, 0.00055919, 0.00123553, -0.00043156, -0.00049392, 0.00029222, -0.00015914, -0.00008992, 0.00014439, -0.00013219, 0.00001992, 0.00005639, -0.00008346, 0.00004771, -0.00001163, -0.00004630, 0.00004639, -0.00002525, -0.00001196, 0.00001988, -0.00002220, 0.00001975, 0.00000708, -0.00002438, 0.00000851, -0.00000119, -0.00000370, 0.00000535, -0.00000411, 0.00001065, 0.00000719, -0.00000962, -0.00000306, 0.00000181, -0.00000083, -0.00000514, 0.00000457, 0.00001209, -0.00000414, -0.00000042, 0.00000726, -0.00001123, -0.00000568, 0.00000814, -0.00000511, -0.00000662, 0.00000377, 0.00000091, -0.00000637, -0.00000576, 0.00000174, -0.00000599, -0.00001372, 0.00000522, 0.00000671, -0.00000599, 0.00000555, -0.00000246, -0.00002067, 0.00000099, 0.00001618, 0.00000145, -0.00000371, },
		{ 0.00000174, 0.00000239, 0.00000225, 0.00000220, 0.00000383, -0.00000324, -0.00000317, 0.00001615, 0.00000556, -0.00002134, -0.00000352, 0.00000914, -0.00000803, 0.00000095, 0.00001026, -0.00000869, -0.00001219, 0.00000127, -0.00000191, -0.00001002, 0.00000109, 0.00000515, -0.00000974, -0.00000191, 0.00000857, -0.00000968, -0.00000638, 0.00000698, -0.00000481, 0.00000133, 0.00000978, -0.00000059, 0.00000042, -0.00000263, 0.00000108, 0.00000200, -0.00001551, 0.00000756, 0.00001377, -0.00000967, 0.00001094, -0.00000426, -0.00000774, 0.00001683, -0.00002841, 0.00000323, 0.00002645, -0.00003058, 0.00002641, -0.00000657, -0.00004041, 0.00006090, -0.00004476, -0.00002961, 0.00006684, -0.00009206, 0.00004466, 0.00004892, -0.00015657, 0.00014174, -0.00004985, -0.00021075, 0.00030978, -0.00040938, -0.00055706, 0.00115063, 0.00071908, -0.00077349, -0.00009756, 0.00021478, -0.00026619, 0.00007536, 0.00004908, -0.00013765, 0.00009646, -0.00002066, -0.00005075, 0.00006372, -0.00005916, -0.00000290, 0.00003699, -0.00004520, 0.00001567, 0.00000305, -0.00001701, 0.00003068, -0.00001404, -0.00001468, 0.00001298, -0.00000965, 0.00000327, 0.00000225, -0.00000291, 0.00001558, -0.00000122, -0.00001033, 0.00000153, 0.00000046, -0.00000218, -0.00000290, 0.00000754, 0.00000827, -0.00000570, 0.00000345, 0.00000306, -0.00001312, 0.00000007, 0.00000650, -0.00000835, -0.00000349, 0.00000514, -0.00000243, -0.00000806, -0.00000174, 0.00000044, -0.00001234, -0.00000697, 0.00001065, -0.00000113, -0.00000399, 0.00000837, -0.00001179, -0.00001714, 0.00001166, 0.00001209, -0.00000389, -0.00000137, 0.00000286, },
		{ 0.00000265, 0.00000180, 0.00000282, 0.00000177, -0.00000545, 0.00000431, 0.00001747, -0.00000712, -0.00001966, 0.00000533, 0.00000260, -0.00000756, 0.00000875, 0.00000295, -0.00001396, -0.00000528, 0.00000174, -0.00000655, -0.00000682, 0.00000461, 0.00000068, -0.00001015, 0.00000305, 0.00000498, -0.00001244, -0.00000168, 0.00000574, -0.00000562, 0.00000473, 0.00000874, -0.00000102, -0.00000128, -0.00000206, 0.00000350, -0.00000558, -0.00001058, 0.00001541, 0.00000372, -0.00000448, 0.00001025, -0.00001178, 0.00000612, 0.00000291, -0.00002927, 0.00002503, 0.00000305, -0.00001933, 0.00003270, -0.00003792, 0.00000024, 0.00004299, -0.00007450, 0.00003135, 0.00001431, -0.00008422, 0.00010593, -0.00005966, -0.00007951, 0.00016998, -0.00021427, -0.00000431, 0.00021829, -0.00073661, 0.00002347, 0.00138284, 0.00002347, -0.00073661, 0.00021829, -0.00000431, -0.00021427, 0.00016998, -0.00007951, -0.00005966, 0.00010593, -0.00008422, 0.00001431, 0.00003135, -0.00007450, 0.00004299, 0.00000024, -0.00003792, 0.00003270, -0.00001933, 0.00000305, 0.00002503, -0.00002927, 0.00000291, 0.00000612, -0.00001178, 0.00001025, -0.00000448, 0.00000372, 0.00001541, -0.00001058, -0.00000558, 0.00000350, -0.00000206, -0.00000128, -0.00000102, 0.00000874, 0.00000473, -0.00000562, 0.00000574, -0.00000168, -0.00001244, 0.00000498, 0.00000305, -0.00001015, 0.00000068, 0.00000461, -0.00000682, -0.00000655, 0.00000174, -0.00000528, -0.00001396, 0.00000295, 0.00000875, -0.00000756, 0.00000260, 0.00000533, -0.00001966, -0.00000712, 0.00001747, 0.00000431, -0.00000545, 0.00000177, 0.00000282, 0.00000180, },
		{ 0.00000174, 0.00000286, -0.00000137, -0.00000389, 0.00001209, 0.00001166, -0.00001714, -0.00001179, 0.00000837, -0.00000399, -0.00000113, 0.00001065, -0.00000697, -0.00001234, 0.00000044, -0.00000174, -0.00000806, -0.00000243, 0.00000514, -0.00000349, -0.00000835, 0.00000650, 0.00000007, -0.00001312, 0.00000306, 0.00000345, -0.00000570, 0.00000827, 0.00000754, -0.00000290, -0.00000218, 0.00000046, 0.00000153, -0.00001033, -0.00000122, 0.00001558, -0.00000291, 0.00000225, 0.00000327, -0.00000965, 0.00001298, -0.00001468, -0.00001404, 0.00003068, -0.00001701, 0.00000305, 0.00001567, -0.00004520, 0.00003699, -0.00000290, -0.00005916, 0.00006372, -0.00005075, -0.00002066, 0.00009646, -0.00013765, 0.00004908, 0.00007536, -0.00026619, 0.00021478, -0.00009756, -0.00077349, 0.00071908, 0.00115063, -0.00055706, -0.00040938, 0.00030978, -0.00021075, -0.00004985, 0.00014174, -0.00015657, 0.00004892, 0.00004466, -0.00009206, 0.00006684, -0.00002961, -0.00004476, 0.00006090, -0.00004041, -0.00000657, 0.00002641, -0.00003058, 0.00002645, 0.00000323, -0.00002841, 0.00001683, -0.00000774, -0.00000426, 0.00001094, -0.00000967, 0.00001377, 0.00000756, -0.00001551, 0.00000200, 0.00000108, -0.00000263, 0.00000042, -0.00000059, 0.00000978, 0.00000133, -0.00000481, 0.00000698, -0.00000638, -0.00000968, 0.00000857, -0.00000191, -0.00000974, 0.00000515, 0.00000109, -0.00001002, -0.00000191, 0.00000127, -0.00001219, -0.00000869, 0.00001026, 0.00000095, -0.00000803, 0.00000914, -0.00000352, -0.00002134, 0.00000556, 0.00001615, -0.00000317, -0.00000324, 0.00000383, 0.00000220, 0.00000225, 0.00000239, },
		{ 0.00000158, -0.00000371, 0.00000145, 0.00001618, 0.00000099, -0.00002067, -0.00000246, 0.00000555, -0.00000599, 0.00000671, 0.00000522, -0.00001372, -0.00000599, 0.00000174, -0.00000576, -0.00000637, 0.00000091, 0.00000377, -0.00000662, -0.00000511, 0.00000814, -0.00000568, -0.00001123, 0.00000726, -0.00000042, -0.00000414, 0.00001209, 0.00000457, -0.00000514, -0.00000083, 0.00000181, -0.00000306, -0.00000962, 0.00000719, 0.00001065, -0.00000411, 0.00000535, -0.00000370, -0.00000119, 0.00000851, -0.00002438, 0.00000708, 0.00001975, -0.00002220, 0.00001988, -0.00001196, -0.00002525, 0.00004639, -0.00004630, -0.00001163, 0.00004771, -0.00008346, 0.00005639, 0.00001992, -0.00013219, 0.00014439, -0.00008992, -0.00015914, 0.00029222, -0.00049392, -0.00043156, 0.00123553, 0.00055919, -0.00080466, -0.00000402, 0.00017574, -0.00028012, 0.00011632, 0.00001946, -0.00013535, 0.00011758, -0.00004439, -0.00004023, 0.00006990, -0.00007867, 0.00001281, 0.00003788, -0.00005764, 0.00002867, 0.00000010, -0.00002031, 0.00003701, -0.00002284, -0.00001106, 0.00001705, -0.00001836, 0.00000883, 0.00000294, -0.00000744, 0.00002108, -0.00000577, -0.00001226, 0.00000712, -0.00000380, 0.00000034, 0.00000010, -0.00000020, 0.00001125, -0.00000302, -0.00000271, 0.00000720, -0.00001099, -0.00000476, 0.00000995, -0.00000768, -0.00000620, 0.00000793, -0.00000506, -0.00000937, 0.00000304, -0.00000436, -0.00001529, 0.00000127, 0.00001043, -0.00000748, -0.00000194, 0.00001048, -0.00001418, -0.00001491, 0.00001542, 0.00000895, -0.00000697, 0.00000076, 0.00000413, 0.00000186, 0.00000263, 0.00000189, 0.00000204, },
		{ -0.00000333, 0.00000835, 0.00001422, -0.00001004, -0.00001742, 0.00000378, 0.00000038, -0.00000223, 0.00001056, -0.00000444, -0.00001392, 0.00000047, -0.00000115, -0.00000746, -0.00000331, 0.00000252, 0.00000156, -0.00000860, -0.00000081, 0.00000734, -0.00001127, -0.00000626, 0.00000928, -0.00000523, 0.00000036, 0.00001452, -0.00000060, -0.00000554, 0.00000166, 0.00000026, -0.00000672, -0.00000457, 0.00001132, 0.00000495, -0.00000203, 0.00000385, -0.00000571, 0.00000632, -0.00000348, -0.00002078, 0.00002178, 0.00000173, -0.00001255, 0.00002021, -0.00003137, 0.00000696, 0.00002419, -0.00006018, 0.00003501, -0.00000320, -0.00006167, 0.00009498, -0.00007470, -0.00004485, 0.00013901, -0.00022142, 0.00005014, 0.00014051, -0.00075000, 0.00019113, 0.00134791, -0.00013032, -0.00067732, 0.00026550, -0.00006194, -0.00018651, 0.00018354, -0.00011145, -0.00003490, 0.00010501, -0.00010159, 0.00003472, 0.00002007, -0.00008196, 0.00006149, -0.00001189, -0.00003909, 0.00004405, -0.00002819, 0.00000702, 0.00002529, -0.00003680, 0.00001187, 0.00000255, -0.00001667, 0.00001814, -0.00000922, 0.00000376, 0.00001910, -0.00001770, -0.00000208, 0.00000566, -0.00000649, 0.00000490, -0.00000352, 0.00000280, 0.00001138, -0.00000821, 0.00000178, 0.00000517, -0.00001462, 0.00000221, 0.00000773, -0.00001236, 0.00000025, 0.00000661, -0.00001120, -0.00000407, 0.00000417, -0.00001258, -0.00001119, 0.00001022, 0.00000331, -0.00001076, 0.00000677, 0.00000448, -0.00002087, -0.00000274, 0.00001833, -0.00000014, -0.00000602, 0.00000413, 0.00000328, 0.00000210, 0.00000253, 0.00000160, 0.00000206, -0.00000079, },
		{ 0.00001340, 0.00000683, -0.00001717, -0.00001030, 0.00000511, -0.00000274, 0.00000449, 0.00000760, -0.00001288, -0.00000833, 0.00000317, -0.00000542, -0.00000619, -0.00000083, 0.00000300, -0.00000113, -0.00000913, 0.00000394, 0.00000345, -0.00001475, 0.00000101, 0.00000735, -0.00000866, 0.00000761, 0.00001320, -0.00000616, -0.00000301, 0.00000276, -0.00000323, -0.00000700, 0.00000150, 0.00001124, 0.00000122, -0.00000001, 0.00000054, -0.00000249, 0.00000732, -0.00001465, -0.00000721, 0.00002362, -0.00001166, 0.00000223, 0.00000566, -0.00003088, 0.00002959, -0.00001305, -0.00003858, 0.00005058, -0.00005270, -0.00000075, 0.00006885, -0.00012550, 0.00006744, 0.00002969, -0.00022902, 0.00022746, -0.00019542, -0.00068678, 0.00084644, 0.00101337, -0.00063363, -0.00031102, 0.00029332, -0.00024391, -0.00000538, 0.00012224, -0.00016733, 0.00007634, 0.00002461, -0.00009092, 0.00008110, -0.00004992, -0.00003556, 0.00006951, -0.00005597, 0.00000363, 0.00002888, -0.00003786, 0.00003386, -0.00000361, -0.00002899, 0.00002475, -0.00001690, -0.00000191, 0.00001563, -0.00001636, 0.00001849, 0.00000613, -0.00002065, 0.00000866, -0.00000178, -0.00000328, 0.00000646, -0.00000813, 0.00000939, 0.00000729, -0.00001173, 0.00000824, -0.00000054, -0.00001506, 0.00000962, 0.00000130, -0.00001320, 0.00000722, 0.00000029, -0.00001341, 0.00000334, -0.00000107, -0.00001798, -0.00000086, 0.00001247, -0.00000648, -0.00000640, 0.00001213, -0.00000639, -0.00001954, 0.00000971, 0.00001372, -0.00000670, -0.00000184, 0.00000546, 0.00000240, 0.00000254, 0.00000203, 0.00000168, 0.00000107, -0.00000285, 0.00000038, },
		{ -0.00000290, -0.00001858, -0.00000332, 0.00000304, -0.00000146, 0.00000942, -0.00000081, -0.00001573, -0.00000087, 0.00000119, -0.00000795, -0.00000336, 0.00000042, 0.00000296, -0.00000432, -0.00000753, 0.00000770, -0.00000316, -0.00001401, 0.00000796, 0.00000142, -0.00000783, 0.00001503, 0.00000729, -0.00000916, 0.00000109, 0.00000093, -0.00000601, -0.00000412, 0.00000586, 0.00000901, -0.00000050, 0.00000032, -0.00000113, 0.00000219, 0.00000156, -0.00001822, 0.00000781, 0.00001478, -0.00001441, 0.00001098, -0.00001227, -0.00001329, 0.00002908, -0.00004051, 0.00000102, 0.00002776, -0.00006724, 0.00005771, -0.00000480, -0.00009937, 0.00012914, -0.00011693, -0.00009911, 0.00024031, -0.00054773, -0.00027214, 0.00125579, 0.00038942, -0.00078097, 0.00006974, 0.00012058, -0.00026922, 0.00014437, -0.00001487, -0.00011901, 0.00012782, -0.00006714, -0.00002299, 0.00006734, -0.00009329, 0.00003140, 0.00003163, -0.00006537, 0.00004247, -0.00000651, -0.00002054, 0.00004116, -0.00003215, -0.00000388, 0.00001803, -0.00002697, 0.00001675, 0.00000124, -0.00001114, 0.00002687, -0.00001206, -0.00001185, 0.00001234, -0.00000948, 0.00000492, 0.00000155, -0.00000879, 0.00001677, -0.00000204, -0.00000992, 0.00001357, -0.00000948, -0.00001012, 0.00001413, -0.00000766, -0.00000843, 0.00001074, -0.00000877, -0.00000918, 0.00000780, -0.00001104, -0.00001574, 0.00001021, 0.00000627, -0.00001240, 0.00000296, 0.00001002, -0.00001632, -0.00001039, 0.00001691, 0.00000473, -0.00000825, 0.00000283, 0.00000493, 0.00000217, 0.00000263, 0.00000158, 0.00000166, -0.00000092, -0.00000287, 0.00000608, 0.00001396, },
		{ -0.00001533, 0.00000092, 0.00000084, 0.00000321, 0.00000895, -0.00001021, -0.00001210, 0.00000410, -0.00000345, -0.00000739, -0.00000092, 0.00000097, 0.00000229, -0.00000753, -0.00000350, 0.00000848, -0.00001037, -0.00000836, 0.00001121, -0.00000597, -0.00000151, 0.00001880, -0.00000126, -0.00000778, 0.00000388, -0.00000300, -0.00000602, -0.00000017, 0.00000784, 0.00000647, -0.00000155, -0.00000017, -0.00000004, 0.00000409, -0.00000654, -0.00001304, 0.00001673, 0.00000274, -0.00000873, 0.00000894, -0.00002146, 0.00000764, 0.00000844, -0.00004217, 0.00003121, -0.00001467, -0.00003861, 0.00007475, -0.00007730, -0.00001474, 0.00009664, -0.00020566, 0.00008936, 0.00004391, -0.00071045, 0.00034960, 0.00124851, -0.00025047, -0.00058626, 0.00027716, -0.00011076, -0.00014489, 0.00017761, -0.00013387, -0.00000554, 0.00009237, -0.00011014, 0.00005385, 0.00000296, -0.00008080, 0.00007604, -0.00002724, -0.00003405, 0.00005193, -0.00003755, 0.00001319, 0.00002202, -0.00004168, 0.00002187, -0.00000444, -0.00001892, 0.00002604, -0.00001583, 0.00000568, 0.00002125, -0.00002488, 0.00000371, 0.00000579, -0.00001045, 0.00001222, -0.00000853, -0.00000170, 0.00001920, -0.00001343, -0.00000113, 0.00001327, -0.00001806, 0.00000002, 0.00001235, -0.00001498, 0.00000071, 0.00000757, -0.00001560, 0.00000011, 0.00000488, -0.00002006, -0.00000528, 0.00001522, -0.00000469, -0.00001025, 0.00001122, 0.00000098, -0.00001978, 0.00000215, 0.00001631, -0.00000387, -0.00000524, 0.00000578, 0.00000370, 0.00000245, 0.00000221, 0.00000148, 0.00000083, -0.00000287, -0.00000001, 0.00001103, 0.00000957, -0.00001123, },
		{ 0.00000230, 0.00000103, 0.00000781, 0.00000279, -0.00001586, -0.00000464, 0.00000431, -0.00000752, -0.00000454, 0.00000013, 0.00000161, 0.00000041, -0.00000947, 0.00000198, 0.00000510, -0.00001497, 0.00000028, 0.00000869, -0.00001063, 0.00000835, 0.00001633, -0.00000857, -0.00000292, 0.00000319, -0.00000641, -0.00000332, 0.00000286, 0.00000833, 0.00000398, -0.00000264, 0.00000026, 0.00000210, 0.00000178, -0.00001175, -0.00000312, 0.00001721, -0.00000573, -0.00000148, -0.00000016, -0.00001766, 0.00001779, -0.00001654, -0.00002057, 0.00003374, -0.00004643, 0.00001177, 0.00004038, -0.00010184, 0.00007076, -0.00001144, -0.00017547, 0.00020859, -0.00028000, -0.00055338, 0.00092338, 0.00084359, -0.00065515, -0.00021639, 0.00024973, -0.00025310, 0.00003418, 0.00009070, -0.00016235, 0.00009654, 0.00000041, -0.00008020, 0.00008735, -0.00006864, -0.00002026, 0.00007030, -0.00006870, 0.00001706, 0.00002652, -0.00004244, 0.00004049, -0.00001250, -0.00002551, 0.00003041, -0.00002724, 0.00000352, 0.00001811, -0.00002289, 0.00002428, 0.00000245, -0.00002377, 0.00001585, -0.00000666, -0.00000197, 0.00001157, -0.00001718, 0.00001171, 0.00001205, -0.00002022, 0.00001168, 0.00000482, -0.00002098, 0.00001153, 0.00000363, -0.00001611, 0.00000960, -0.00000194, -0.00001533, 0.00000902, -0.00000565, -0.00002157, 0.00000845, 0.00001060, -0.00001333, -0.00000115, 0.00001280, -0.00001001, -0.00001487, 0.00001210, 0.00000944, -0.00000834, -0.00000016, 0.00000640, 0.00000284, 0.00000265, 0.00000164, 0.00000145, -0.00000102, -0.00000317, 0.00000492, 0.00001272, 0.00000210, -0.00001574, -0.00001011, },
		{ 0.00000380, 0.00000887, -0.00000610, -0.00001537, 0.00000242, 0.00000051, -0.00000875, -0.00000137, 0.00000015, 0.00000236, -0.00000283, -0.00000879, 0.00000666, -0.00000166, -0.00001439, 0.00000803, 0.00000143, -0.00000921, 0.00001731, 0.00000818, -0.00001121, 0.00000224, -0.00000075, -0.00000697, 0.00000024, 0.00000448, 0.00000821, 0.00000107, -0.00000306, 0.00000208, 0.00000274, -0.00000281, -0.00001160, 0.00000607, 0.00001192, -0.00000853, 0.00000185, -0.00000851, -0.00000626, 0.00001239, -0.00002938, 0.00000701, 0.00001081, -0.00004718, 0.00004914, -0.00002082, -0.00006500, 0.00009915, -0.00012529, -0.00004280, 0.00016095, -0.00055855, -0.00009988, 0.00120924, 0.00023239, -0.00071189, 0.00011364, 0.00006178, -0.00023724, 0.00015445, -0.00004641, -0.00009221, 0.00012519, -0.00008421, -0.00000263, 0.00005632, -0.00009999, 0.00004923, 0.00001932, -0.00006647, 0.00005445, -0.00001565, -0.00001741, 0.00004198, -0.00004019, 0.00000579, 0.00001518, -0.00003369, 0.00002579, -0.00000299, -0.00001291, 0.00003172, -0.00001939, -0.00000864, 0.00001594, -0.00001551, 0.00001123, 0.00000047, -0.00001650, 0.00002413, -0.00000381, -0.00001647, 0.00002142, -0.00000940, -0.00001473, 0.00001832, -0.00000842, -0.00000935, 0.00001254, -0.00001299, -0.00000686, 0.00001099, -0.00001855, -0.00001271, 0.00001753, -0.00000126, -0.00001384, 0.00000913, 0.00000637, -0.00001673, -0.00000437, 0.00001536, 0.00000058, -0.00000772, 0.00000425, 0.00000550, 0.00000265, 0.00000240, 0.00000133, 0.00000083, -0.00000308, -0.00000080, 0.00000968, 0.00001024, -0.00000557, -0.00001621, -0.00000531, 0.00000259, },
		{ 0.00000507, -0.00001348, -0.00000970, 0.00000577, -0.00000458, -0.00000693, 0.00000054, 0.00000015, 0.00000246, -0.00000653, -0.00000517, 0.00000813, -0.00000898, -0.00000852, 0.00001105, -0.00000667, -0.00000145, 0.00002088, -0.00000196, -0.00000846, 0.00000438, -0.00000542, -0.00000430, 0.00000273, 0.00000547, 0.00000733, -0.00000218, -0.00000169, 0.00000394, 0.00000088, -0.00000648, -0.00000729, 0.00001125, 0.00000511, -0.00000781, 0.00000068, -0.00001084, 0.00000317, -0.00000205, -0.00002408, 0.00002195, -0.00001878, -0.00001940, 0.00004981, -0.00006791, 0.00000506, 0.00005135, -0.00016999, 0.00010435, -0.00005533, -0.00062000, 0.00047731, 0.00109949, -0.00032428, -0.00048107, 0.00025563, -0.00014177, -0.00009861, 0.00015446, -0.00014228, 0.00002230, 0.00007081, -0.00010819, 0.00006765, -0.00001659, -0.00007115, 0.00008366, -0.00004270, -0.00002343, 0.00005468, -0.00004555, 0.00002041, 0.00001550, -0.00004269, 0.00003094, -0.00001394, -0.00001772, 0.00003232, -0.00002322, 0.00000948, 0.00002095, -0.00003075, 0.00001105, 0.00000346, -0.00001277, 0.00001935, -0.00001571, -0.00000342, 0.00002641, -0.00002082, -0.00000173, 0.00002097, -0.00002233, -0.00000084, 0.00001580, -0.00001724, 0.00000209, 0.00000663, -0.00001872, 0.00000540, 0.00000290, -0.00002574, 0.00000296, 0.00001612, -0.00001303, -0.00000556, 0.00001389, -0.00000427, -0.00001562, 0.00000613, 0.00001172, -0.00000590, -0.00000366, 0.00000652, 0.00000432, 0.00000271, 0.00000182, 0.00000124, -0.00000079, -0.00000388, 0.00000378, 0.00001191, 0.00000468, -0.00001114, -0.00001404, -0.00000177, 0.00000360, 0.00000709, },
		{ -0.00001619, -0.00000211, 0.00000453, -0.00000811, -0.00000348, 0.00000085, 0.00000074, 0.00000108, -0.00000905, 0.00000013, 0.00000535, -0.00001339, -0.00000006, 0.00000794, -0.00001092, 0.00000933, 0.00001722, -0.00000946, -0.00000272, 0.00000231, -0.00000781, -0.00000011, 0.00000357, 0.00000657, 0.00000497, -0.00000450, 0.00000145, 0.00000393, -0.00000218, -0.00000739, -0.00000171, 0.00001233, -0.00000065, -0.00000653, -0.00000180, -0.00000787, 0.00000454, -0.00001400, -0.00000764, 0.00001699, -0.00003431, 0.00001546, 0.00001637, -0.00007216, 0.00005908, -0.00004020, -0.00011612, 0.00015975, -0.00033554, -0.00039019, 0.00094039, 0.00066445, -0.00062499, -0.00013982, 0.00019030, -0.00023833, 0.00006126, 0.00005426, -0.00014303, 0.00010560, -0.00002290, -0.00006230, 0.00008422, -0.00008197, -0.00000194, 0.00006306, -0.00007581, 0.00003123, 0.00001968, -0.00004329, 0.00004488, -0.00002186, -0.00001835, 0.00003242, -0.00003688, 0.00001127, 0.00001763, -0.00002787, 0.00003008, -0.00000325, -0.00002391, 0.00002218, -0.00001283, 0.00000156, 0.00001435, -0.00002614, 0.00001663, 0.00001402, -0.00002857, 0.00001713, 0.00000835, -0.00002610, 0.00001393, 0.00000448, -0.00001739, 0.00001137, -0.00000549, -0.00001467, 0.00001354, -0.00001194, -0.00002139, 0.00001696, 0.00000459, -0.00001713, 0.00000635, 0.00001004, -0.00001273, -0.00000786, 0.00001170, 0.00000439, -0.00000784, 0.00000114, 0.00000680, 0.00000358, 0.00000252, 0.00000132, 0.00000082, -0.00000294, -0.00000217, 0.00000868, 0.00001049, -0.00000177, -0.00001389, -0.00001078, 0.00000111, 0.00000580, 0.00000828, -0.00000213, },
		{ 0.00000383, 0.00000023, -0.00000863, -0.00000031, 0.00000030, 0.00000166, -0.00000184, -0.00000902, 0.00000474, -0.00000051, -0.00001278, 0.00000688, 0.00000066, -0.00000859, 0.00001808, 0.00000817, -0.00001145, 0.00000219, -0.00000240, -0.00000648, 0.00000328, 0.00000366, 0.00000746, 0.00000110, -0.00000432, 0.00000464, 0.00000153, -0.00000425, -0.00000603, 0.00000301, 0.00001081, -0.00000518, -0.00000558, -0.00000263, -0.00000441, -0.00000084, -0.00001594, 0.00000684, -0.00000057, -0.00002760, 0.00003359, -0.00002627, -0.00003577, 0.00006095, -0.00011406, -0.00000146, 0.00006702, -0.00052174, 0.00006238, 0.00110409, 0.00010628, -0.00061301, 0.00012471, 0.00001082, -0.00019194, 0.00014576, -0.00006876, -0.00006067, 0.00011062, -0.00009228, 0.00001651, 0.00003912, -0.00009745, 0.00006265, 0.00000354, -0.00006046, 0.00006226, -0.00002552, -0.00001148, 0.00003904, -0.00004533, 0.00001620, 0.00000871, -0.00003708, 0.00003427, -0.00000914, -0.00001221, 0.00003437, -0.00002649, -0.00000288, 0.00001696, -0.00002056, 0.00001827, -0.00000341, -0.00002162, 0.00003179, -0.00000849, -0.00002060, 0.00002924, -0.00001088, -0.00001737, 0.00002135, -0.00000952, -0.00000851, 0.00001219, -0.00001655, -0.00000245, 0.00001115, -0.00002501, -0.00000634, 0.00002089, -0.00001020, -0.00001068, 0.00001424, 0.00000009, -0.00001423, 0.00000170, 0.00001080, -0.00000236, -0.00000583, 0.00000470, 0.00000609, 0.00000319, 0.00000200, 0.00000106, -0.00000047, -0.00000431, 0.00000199, 0.00001161, 0.00000605, -0.00000714, -0.00001437, -0.00000716, 0.00000416, 0.00000803, 0.00000588, -0.00000958, -0.00001363, },
		{ -0.00000456, -0.00000643, 0.00000139, -0.00000013, 0.00000205, -0.00000534, -0.00000624, 0.00000654, -0.00000668, -0.00000764, 0.00000910, -0.00000654, -0.00000048, 0.00002079, -0.00000193, -0.00000828, 0.00000343, -0.00000660, -0.00000234, 0.00000459, 0.00000417, 0.00000699, -0.00000290, -0.00000133, 0.00000568, -0.00000180, -0.00000427, -0.00000404, 0.00000653, 0.00000762, -0.00000892, -0.00000380, -0.00000202, -0.00000435, -0.00000663, -0.00000902, 0.00001034, -0.00001626, -0.00000689, 0.00002541, -0.00004993, 0.00001180, 0.00001182, -0.00012206, 0.00009166, -0.00014017, -0.00048964, 0.00055850, 0.00092197, -0.00034820, -0.00037888, 0.00020926, -0.00015089, -0.00005678, 0.00011983, -0.00013551, 0.00004317, 0.00004495, -0.00009651, 0.00007331, -0.00003457, -0.00005506, 0.00008283, -0.00005496, -0.00000920, 0.00005174, -0.00005054, 0.00002715, 0.00000679, -0.00003946, 0.00003725, -0.00002431, -0.00001313, 0.00003568, -0.00002996, 0.00001447, 0.00001786, -0.00003399, 0.00001866, -0.00000109, -0.00001266, 0.00002474, -0.00002389, -0.00000179, 0.00003110, -0.00002896, 0.00000048, 0.00002650, -0.00002628, -0.00000022, 0.00001708, -0.00001801, 0.00000379, 0.00000342, -0.00001912, 0.00001045, -0.00000190, -0.00002759, 0.00001158, 0.00001213, -0.00001897, 0.00000235, 0.00001313, -0.00000955, -0.00000856, 0.00000784, 0.00000556, -0.00000576, -0.00000200, 0.00000647, 0.00000525, 0.00000279, 0.00000143, 0.00000068, -0.00000247, -0.00000352, 0.00000711, 0.00001115, 0.00000041, -0.00001061, -0.00001340, -0.00000304, 0.00000741, 0.00000856, 0.00000045, -0.00001426, -0.00000760, 0.00000600, },
		{ -0.00000297, 0.00000149, 0.00000010, 0.00000116, -0.00000797, -0.00000189, 0.00000489, -0.00001032, -0.00000070, 0.00000607, -0.00000961, 0.00000963, 0.00001665, -0.00000879, -0.00000296, 0.00000082, -0.00000771, 0.00000227, 0.00000422, 0.00000532, 0.00000443, -0.00000493, 0.00000290, 0.00000368, -0.00000379, -0.00000309, -0.00000227, 0.00000923, 0.00000260, -0.00001114, -0.00000039, -0.00000270, -0.00000722, -0.00000745, -0.00000061, 0.00000353, -0.00001990, 0.00001134, 0.00000039, -0.00004294, 0.00003623, -0.00005229, -0.00006262, 0.00008915, -0.00035184, -0.00021870, 0.00089806, 0.00049771, -0.00055448, -0.00008969, 0.00012754, -0.00020459, 0.00007189, 0.00002043, -0.00011375, 0.00010225, -0.00004076, -0.00004118, 0.00007255, -0.00008727, 0.00001560, 0.00004940, -0.00007567, 0.00004338, 0.00000972, -0.00004023, 0.00004597, -0.00002994, -0.00000878, 0.00003023, -0.00004402, 0.00001991, 0.00001420, -0.00003025, 0.00003453, -0.00001007, -0.00002077, 0.00002633, -0.00001905, 0.00000682, 0.00001396, -0.00003317, 0.00002308, 0.00001234, -0.00003476, 0.00002348, 0.00000926, -0.00002893, 0.00001588, 0.00000386, -0.00001619, 0.00001129, -0.00000951, -0.00001090, 0.00001514, -0.00001844, -0.00001679, 0.00002202, -0.00000411, -0.00001607, 0.00001367, 0.00000393, -0.00001294, 0.00000008, 0.00000815, -0.00000014, -0.00000556, 0.00000163, 0.00000697, 0.00000449, 0.00000219, 0.00000101, -0.00000033, -0.00000417, -0.00000017, 0.00001094, 0.00000752, -0.00000462, -0.00001246, -0.00001127, 0.00000180, 0.00000985, 0.00000636, -0.00000594, -0.00001472, -0.00000097, 0.00000436, -0.00000759, },
		{ 0.00000067, 0.00000079, -0.00000106, -0.00000864, 0.00000216, 0.00000082, -0.00001005, 0.00000464, 0.00000003, -0.00000685, 0.00001714, 0.00000808, -0.00001045, 0.00000100, -0.00000349, -0.00000522, 0.00000525, 0.00000354, 0.00000614, 0.00000048, -0.00000376, 0.00000573, -0.00000006, -0.00000330, -0.00000244, -0.00000016, 0.00001062, -0.00000412, -0.00001002, 0.00000309, -0.00000664, -0.00000899, -0.00000330, 0.00000253, -0.00000563, -0.00001212, 0.00001539, -0.00002222, -0.00001640, 0.00002265, -0.00008720, 0.00001752, -0.00002547, -0.00044154, 0.00019464, 0.00095687, 0.00002154, -0.00050212, 0.00010692, -0.00002442, -0.00014273, 0.00012163, -0.00007817, -0.00003057, 0.00008754, -0.00009025, 0.00003054, 0.00001940, -0.00008635, 0.00006898, -0.00001238, -0.00004842, 0.00006440, -0.00003415, -0.00000401, 0.00003275, -0.00004654, 0.00002537, -0.00000025, -0.00003651, 0.00004055, -0.00001596, -0.00000918, 0.00003400, -0.00003193, 0.00000451, 0.00001510, -0.00002346, 0.00002461, -0.00000949, -0.00002298, 0.00003775, -0.00001524, -0.00002114, 0.00003514, -0.00001324, -0.00001736, 0.00002204, -0.00001001, -0.00000610, 0.00000889, -0.00001791, 0.00000310, 0.00000749, -0.00002816, 0.00000197, 0.00001875, -0.00001769, -0.00000330, 0.00001611, -0.00000713, -0.00000850, 0.00000617, 0.00000406, -0.00000339, -0.00000333, 0.00000421, 0.00000687, 0.00000366, 0.00000155, 0.00000055, -0.00000204, -0.00000423, 0.00000472, 0.00001178, 0.00000226, -0.00000819, -0.00001315, -0.00000765, 0.00000683, 0.00001016, 0.00000182, -0.00001097, -0.00001143, 0.00000375, 0.00000043, -0.00000791, 0.00000014, },
		{ 0.00000132, -0.00000399, -0.00000711, 0.00000443, -0.00000363, -0.00000652, 0.00000625, -0.00000542, 0.00000041, 0.00001913, -0.00000090, -0.00000787, 0.00000162, -0.00000661, -0.00000060, 0.00000593, 0.00000349, 0.00000546, -0.00000291, -0.00000012, 0.00000545, -0.00000294, -0.00000132, -0.00000302, 0.00000324, 0.00000916, -0.00001071, -0.00000482, 0.00000330, -0.00001233, -0.00000620, 0.00000118, -0.00000043, -0.00000950, -0.00000183, 0.00000597, -0.00002862, 0.00000642, -0.00001537, -0.00007222, 0.00005435, -0.00019665, -0.00033706, 0.00058625, 0.00073875, -0.00032757, -0.00029272, 0.00014978, -0.00013941, -0.00002619, 0.00008119, -0.00011582, 0.00005370, 0.00001993, -0.00007799, 0.00006998, -0.00004736, -0.00003594, 0.00007389, -0.00006142, 0.00000590, 0.00004375, -0.00005156, 0.00003197, -0.00000250, -0.00003260, 0.00003958, -0.00003362, -0.00000609, 0.00003551, -0.00003466, 0.00001945, 0.00001239, -0.00003377, 0.00002508, -0.00000694, -0.00001002, 0.00002713, -0.00003132, 0.00000274, 0.00003182, -0.00003588, 0.00000501, 0.00002855, -0.00002842, 0.00000134, 0.00001576, -0.00001625, 0.00000466, -0.00000163, -0.00001580, 0.00001340, -0.00000853, -0.00002429, 0.00001785, 0.00000397, -0.00002004, 0.00001117, 0.00000839, -0.00001282, 0.00000021, 0.00000651, -0.00000074, -0.00000362, -0.00000083, 0.00000592, 0.00000639, 0.00000269, 0.00000106, -0.00000036, -0.00000372, -0.00000194, 0.00000920, 0.00000932, -0.00000288, -0.00001041, -0.00001267, -0.00000241, 0.00001070, 0.00000774, -0.00000358, -0.00001317, -0.00000615, 0.00000532, -0.00000370, -0.00000587, 0.00000183, -0.00000019, },
		{ -0.00000666, -0.00000401, 0.00000435, -0.00000657, -0.00000177, 0.00000409, -0.00000740, 0.00000881, 0.00001546, -0.00000695, -0.00000379, -0.00000070, -0.00000667, 0.00000389, 0.00000505, 0.00000396, 0.00000306, -0.00000392, 0.00000366, 0.00000249, -0.00000317, -0.00000007, -0.00000369, 0.00000750, 0.00000372, -0.00001387, 0.00000230, -0.00000196, -0.00001535, 0.00000043, 0.00000129, -0.00000481, -0.00000686, 0.00000233, -0.00000657, -0.00001981, 0.00000862, -0.00004796, -0.00002471, 0.00000984, -0.00032649, -0.00006053, 0.00080647, 0.00035965, -0.00045971, -0.00006742, 0.00007218, -0.00016012, 0.00006630, -0.00000491, -0.00008054, 0.00008807, -0.00005028, -0.00002119, 0.00005501, -0.00008371, 0.00002887, 0.00003224, -0.00006826, 0.00005121, -0.00000132, -0.00003396, 0.00004343, -0.00003523, 0.00000137, 0.00002431, -0.00004738, 0.00002772, 0.00000862, -0.00002962, 0.00003646, -0.00001667, -0.00001477, 0.00002748, -0.00002398, 0.00001270, 0.00001040, -0.00003667, 0.00002928, 0.00000720, -0.00003713, 0.00002900, 0.00000767, -0.00002840, 0.00001619, 0.00000242, -0.00001234, 0.00000830, -0.00001253, -0.00000446, 0.00001250, -0.00002292, -0.00000851, 0.00002145, -0.00001280, -0.00000977, 0.00001814, -0.00000410, -0.00000971, 0.00000710, 0.00000203, -0.00000312, -0.00000226, 0.00000123, 0.00000714, 0.00000543, 0.00000177, 0.00000055, -0.00000186, -0.00000425, 0.00000219, 0.00001142, 0.00000463, -0.00000681, -0.00001168, -0.00001046, 0.00000374, 0.00001196, 0.00000320, -0.00000813, -0.00001233, -0.00000093, 0.00000396, -0.00000633, -0.00000269, 0.00000191, -0.00000044, 0.00000105, },
		{ -0.00000045, 0.00000238, -0.00000713, 0.00000201, -0.00000005, -0.00000501, 0.00001492, 0.00000837, -0.00000885, -0.00000091, -0.00000394, -0.00000370, 0.00000660, 0.00000378, 0.00000411, 0.00000005, -0.00000217, 0.00000529, -0.00000085, -0.00000110, -0.00000100, -0.00000248, 0.00001035, -0.00000450, -0.00001120, 0.00000676, -0.00001099, -0.00001195, 0.00000618, -0.00000322, -0.00000541, -0.00000274, -0.00000101, -0.00001219, -0.00000847, -0.00000830, -0.00005246, 0.00001240, -0.00010107, -0.00033005, 0.00028363, 0.00078869, -0.00002021, -0.00039536, 0.00006927, -0.00004104, -0.00009823, 0.00008830, -0.00007431, -0.00000702, 0.00006082, -0.00007941, 0.00003707, 0.00000117, -0.00006920, 0.00006719, -0.00002514, -0.00003268, 0.00006057, -0.00003988, 0.00000344, 0.00002424, -0.00004368, 0.00003161, -0.00000993, -0.00003227, 0.00004347, -0.00002200, -0.00000467, 0.00003049, -0.00003449, 0.00001212, 0.00001078, -0.00002357, 0.00002877, -0.00001637, -0.00002038, 0.00004020, -0.00002235, -0.00001794, 0.00003744, -0.00001517, -0.00001484, 0.00001962, -0.00000874, -0.00000304, 0.00000264, -0.00001569, 0.00000806, 0.00000038, -0.00002614, 0.00000970, 0.00001105, -0.00002079, 0.00000635, 0.00001344, -0.00001298, -0.00000035, 0.00000778, -0.00000338, -0.00000240, -0.00000099, 0.00000311, 0.00000781, 0.00000401, 0.00000111, -0.00000036, -0.00000343, -0.00000283, 0.00000663, 0.00001059, -0.00000069, -0.00000938, -0.00001200, -0.00000616, 0.00000919, 0.00001002, -0.00000200, -0.00001069, -0.00000927, 0.00000268, 0.00000086, -0.00000668, 0.00000022, 0.00000091, -0.00000001, -0.00000038, -0.00000814, },
		{ -0.00000047, -0.00000556, 0.00000351, -0.00000378, 0.00000059, 0.00001674, 0.00000092, -0.00000752, -0.00000046, -0.00000585, 0.00000086, 0.00000708, 0.00000282, 0.00000329, -0.00000197, 0.00000093, 0.00000415, -0.00000226, 0.00000106, -0.00000341, 0.00000143, 0.00000912, -0.00001179, -0.00000357, 0.00000455, -0.00001848, -0.00000269, 0.00000623, -0.00000796, -0.00000159, -0.00000284, -0.00000602, -0.00000953, -0.00000699, -0.00002746, -0.00003073, 0.00000122, -0.00021694, -0.00018282, 0.00056322, 0.00056989, -0.00027442, -0.00022949, 0.00008913, -0.00011294, -0.00000988, 0.00004571, -0.00008806, 0.00005326, 0.00000002, -0.00005674, 0.00005896, -0.00005276, -0.00001761, 0.00005888, -0.00006083, 0.00001903, 0.00003238, -0.00004854, 0.00003386, -0.00001070, -0.00002349, 0.00003767, -0.00004015, 0.00000185, 0.00003204, -0.00003645, 0.00002311, 0.00000560, -0.00002999, 0.00002907, -0.00001280, -0.00000551, 0.00002596, -0.00003624, 0.00000877, 0.00002811, -0.00003962, 0.00001046, 0.00002669, -0.00002743, 0.00000274, 0.00001219, -0.00001141, 0.00000346, -0.00000727, -0.00000865, 0.00001252, -0.00001501, -0.00001582, 0.00001914, -0.00000602, -0.00001504, 0.00001793, 0.00000059, -0.00001252, 0.00000873, 0.00000228, -0.00000585, -0.00000023, -0.00000038, 0.00000523, 0.00000756, 0.00000251, 0.00000059, -0.00000183, -0.00000408, 0.00000036, 0.00000969, 0.00000717, -0.00000539, -0.00001083, -0.00001083, -0.00000027, 0.00001218, 0.00000552, -0.00000632, -0.00001104, -0.00000526, 0.00000398, -0.00000239, -0.00000501, 0.00000181, -0.00000025, 0.00000071, -0.00000282, -0.00000786, 0.00000251, },
		{ -0.00000281, 0.00000262, -0.00000525, 0.00000707, 0.00001429, -0.00000454, -0.00000496, -0.00000199, -0.00000520, 0.00000512, 0.00000587, 0.00000221, 0.00000173, -0.00000205, 0.00000336, 0.00000159, -0.00000114, 0.00000114, -0.00000494, 0.00000640, 0.00000314, -0.00001404, 0.00000447, -0.00000439, -0.00001893, 0.00000721, 0.00000000, -0.00000790, 0.00000216, -0.00000779, -0.00000611, -0.00000608, -0.00001668, -0.00003180, -0.00000775, -0.00006349, -0.00026533, 0.00006684, 0.00068243, 0.00025864, -0.00035762, -0.00006828, 0.00003092, -0.00011398, 0.00004840, -0.00001876, -0.00004940, 0.00006685, -0.00005087, -0.00000591, 0.00003540, -0.00007250, 0.00003555, 0.00001508, -0.00005507, 0.00005342, -0.00001138, -0.00002580, 0.00003773, -0.00003690, 0.00001020, 0.00001597, -0.00004656, 0.00003316, 0.00000222, -0.00002636, 0.00003526, -0.00002162, -0.00000706, 0.00002549, -0.00002657, 0.00001775, 0.00000456, -0.00003586, 0.00003325, -0.00000014, -0.00003490, 0.00003182, 0.00000459, -0.00002424, 0.00001384, 0.00000132, -0.00000647, 0.00000194, -0.00001295, 0.00000318, 0.00000532, -0.00002310, 0.00000132, 0.00001437, -0.00001824, 0.00000041, 0.00001775, -0.00001170, -0.00000328, 0.00001144, -0.00000530, -0.00000412, 0.00000114, 0.00000023, 0.00000736, 0.00000625, 0.00000136, -0.00000026, -0.00000341, -0.00000308, 0.00000418, 0.00001039, 0.00000231, -0.00000891, -0.00001115, -0.00000771, 0.00000572, 0.00001174, 0.00000005, -0.00000884, -0.00000962, -0.00000156, 0.00000318, -0.00000447, -0.00000236, 0.00000181, -0.00000076, 0.00000094, -0.00000563, -0.00000576, 0.00000403, -0.00000310, },
		{ 0.00000022, -0.00000383, 0.00001213, 0.00000905, -0.00000709, -0.00000300, -0.00000396, -0.00000209, 0.00000768, 0.00000383, 0.00000171, 0.00000026, -0.00000057, 0.00000391, -0.00000027, 0.00000104, -0.00000121, -0.00000356, 0.00000925, -0.00000513, -0.00000975, 0.00000736, -0.00001531, -0.00001085, 0.00001140, -0.00000792, -0.00000216, 0.00000055, -0.00001237, -0.00000081, -0.00001025, -0.00002711, -0.00001914, -0.00001253, -0.00014843, -0.00020446, 0.00032492, 0.00062064, -0.00002529, -0.00030415, 0.00002297, -0.00004104, -0.00006427, 0.00005301, -0.00005993, 0.00000717, 0.00003541, -0.00006289, 0.00003573, -0.00001228, -0.00004957, 0.00005809, -0.00003241, -0.00001619, 0.00005173, -0.00004180, 0.00000944, 0.00001505, -0.00003748, 0.00003395, -0.00001847, -0.00002548, 0.00004265, -0.00002600, 0.00000007, 0.00002445, -0.00003354, 0.00001851, 0.00000504, -0.00002101, 0.00002976, -0.00002231, -0.00001468, 0.00003809, -0.00002778, -0.00001193, 0.00003524, -0.00001523, -0.00001071, 0.00001412, -0.00000486, -0.00000068, -0.00000562, -0.00000926, 0.00001044, -0.00000856, -0.00001818, 0.00001391, -0.00000048, -0.00001751, 0.00001526, 0.00000646, -0.00001534, 0.00000842, 0.00000612, -0.00000993, 0.00000001, 0.00000077, 0.00000183, 0.00000871, 0.00000427, 0.00000062, -0.00000172, -0.00000416, -0.00000052, 0.00000725, 0.00000870, -0.00000289, -0.00001105, -0.00001002, -0.00000293, 0.00000998, 0.00000823, -0.00000471, -0.00000942, -0.00000720, 0.00000096, 0.00000118, -0.00000476, -0.00000002, 0.00000077, -0.00000036, -0.00000001, -0.00000778, -0.00000237, 0.00000376, -0.00000462, -0.00000008, },
		{ -0.00000000, 0.00001433, 0.00000303, -0.00000721, -0.00000248, -0.00000479, 0.00000227, 0.00000796, 0.00000175, 0.00000127, -0.00000040, 0.00000124, 0.00000293, -0.00000025, 0.00000204, -0.00000420, 0.00000061, 0.00000756, -0.00001124, -0.00000163, 0.00000236, -0.00002140, 0.00000164, 0.00000719, -0.00001124, 0.00000468, -0.00000716, -0.00001053, 0.00000318, -0.00002260, -0.00002593, -0.00000512, -0.00005544, -0.00020095, -0.00004599, 0.00050025, 0.00042920, -0.00020402, -0.00018957, 0.00003674, -0.00007941, -0.00000692, 0.00001854, -0.00005809, 0.00004380, -0.00001237, -0.00003689, 0.00004316, -0.00005043, -0.00000333, 0.00004101, -0.00005356, 0.00002800, 0.00001979, -0.00004226, 0.00003256, -0.00001648, -0.00001392, 0.00003217, -0.00004289, 0.00000903, 0.00002624, -0.00003514, 0.00002444, -0.00000109, -0.00002335, 0.00002998, -0.00001741, -0.00000034, 0.00002156, -0.00003740, 0.00001439, 0.00002067, -0.00003889, 0.00001496, 0.00002154, -0.00002266, 0.00000277, 0.00000747, -0.00000379, -0.00000070, -0.00001173, 0.00000124, 0.00000690, -0.00001884, -0.00000376, 0.00001377, -0.00001456, -0.00000459, 0.00001995, -0.00000810, -0.00000819, 0.00001492, -0.00000375, -0.00000892, 0.00000340, -0.00000049, 0.00000457, 0.00000856, 0.00000237, -0.00000019, -0.00000344, -0.00000331, 0.00000263, 0.00000872, 0.00000521, -0.00000764, -0.00001154, -0.00000719, 0.00000230, 0.00001130, 0.00000304, -0.00000769, -0.00000851, -0.00000461, 0.00000197, -0.00000083, -0.00000358, 0.00000109, -0.00000029, 0.00000047, -0.00000228, -0.00000823, 0.00000122, 0.00000193, -0.00000458, 0.00000164, -0.00000241, },
		{ 0.00001334, -0.00000205, -0.00000607, -0.00000313, -0.00000352, 0.00000629, 0.00000620, 0.00000022, 0.00000102, -0.00000021, 0.00000234, 0.00000178, 0.00000109, 0.00000070, -0.00000537, 0.00000515, 0.00000174, -0.00001179, 0.00000481, -0.00000816, -0.00001838, 0.00001128, -0.00000224, -0.00000703, 0.00000613, -0.00001563, -0.00000160, -0.00000204, -0.00003398, -0.00001114, -0.00001152, -0.00011801, -0.00018072, 0.00015323, 0.00054547, 0.00019480, -0.00026238, -0.00008368, 0.00000567, -0.00007374, 0.00002427, -0.00002137, -0.00002482, 0.00004331, -0.00004410, 0.00000273, 0.00001751, -0.00005637, 0.00003507, 0.00000106, -0.00003869, 0.00005000, -0.00001879, -0.00001734, 0.00002999, -0.00003499, 0.00001625, 0.00000697, -0.00004208, 0.00003531, -0.00000362, -0.00002145, 0.00003105, -0.00002385, 0.00000091, 0.00002100, -0.00002643, 0.00002070, -0.00000202, -0.00003104, 0.00003347, -0.00000775, -0.00002847, 0.00003063, 0.00000150, -0.00001713, 0.00000848, 0.00000173, 0.00000007, -0.00000724, -0.00000957, 0.00000991, -0.00000531, -0.00001751, 0.00000969, 0.00000176, -0.00001771, 0.00001161, 0.00001189, -0.00001628, 0.00000490, 0.00001213, -0.00001210, -0.00000343, 0.00000395, -0.00000089, 0.00000752, 0.00000695, 0.00000098, -0.00000154, -0.00000447, -0.00000093, 0.00000521, 0.00000847, 0.00000053, -0.00001129, -0.00001000, -0.00000311, 0.00000646, 0.00000962, -0.00000218, -0.00000860, -0.00000680, -0.00000256, 0.00000180, -0.00000195, -0.00000191, 0.00000092, -0.00000056, 0.00000076, -0.00000518, -0.00000652, 0.00000373, -0.00000058, -0.00000307, 0.00000173, -0.00000397, 0.00000496, },
		{ -0.00000529, -0.00000491, -0.00000393, -0.00000026, 0.00000851, 0.00000324, -0.00000036, 0.00000104, 0.00000034, 0.00000263, 0.00000151, 0.00000209, -0.00000216, -0.00000358, 0.00000706, -0.00000502, -0.00000692, 0.00000492, -0.00001793, -0.00000753, 0.00001265, -0.00001003, 0.00000158, -0.00000061, -0.00001744, 0.00000765, -0.00001757, -0.00003278, 0.00000456, -0.00004836, -0.00016269, -0.00008307, 0.00032280, 0.00046990, -0.00000518, -0.00023381, -0.00002156, -0.00002991, -0.00004298, 0.00002212, -0.00003980, 0.00001185, 0.00001508, -0.00004461, 0.00002808, -0.00001921, -0.00003114, 0.00004405, -0.00003329, -0.00000175, 0.00003972, -0.00003991, 0.00001302, 0.00000673, -0.00002936, 0.00003236, -0.00002436, -0.00001769, 0.00003855, -0.00002731, 0.00000378, 0.00001712, -0.00002925, 0.00002256, -0.00000081, -0.00001663, 0.00002737, -0.00002573, -0.00000760, 0.00003151, -0.00002981, -0.00000485, 0.00002871, -0.00001238, -0.00000636, 0.00000644, 0.00000179, -0.00000032, -0.00001412, 0.00000084, 0.00000868, -0.00001681, -0.00000514, 0.00001221, -0.00001270, -0.00000758, 0.00002022, -0.00000306, -0.00001302, 0.00001551, 0.00000186, -0.00001433, 0.00000285, 0.00000190, 0.00000070, 0.00000934, 0.00000460, -0.00000011, -0.00000334, -0.00000386, 0.00000198, 0.00000659, 0.00000672, -0.00000483, -0.00001294, -0.00000644, 0.00000105, 0.00000852, 0.00000584, -0.00000612, -0.00000774, -0.00000507, -0.00000137, 0.00000124, -0.00000199, -0.00000077, 0.00000022, -0.00000000, -0.00000031, -0.00000740, -0.00000314, 0.00000427, -0.00000257, -0.00000087, 0.00000024, -0.00000358, 0.00000950, 0.00000981, },
		{ -0.00000438, -0.00000366, 0.00000395, 0.00000820, 0.00000030, 0.00000006, 0.00000107, 0.00000085, 0.00000273, 0.00000204, 0.00000146, -0.00000455, 0.00000010, 0.00000506, -0.00000885, -0.00000041, -0.00000188, -0.00002085, 0.00000498, 0.00000587, -0.00001070, 0.00000763, -0.00001171, -0.00000958, 0.00000782, -0.00003476, -0.00001578, 0.00000175, -0.00010312, -0.00015580, 0.00005951, 0.00041318, 0.00032272, -0.00013112, -0.00016824, -0.00000215, -0.00004671, -0.00001338, 0.00000188, -0.00003126, 0.00002903, -0.00001703, -0.00002153, 0.00002618, -0.00004187, 0.00000500, 0.00002368, -0.00004139, 0.00003172, 0.00000813, -0.00003407, 0.00002853, -0.00001920, -0.00000561, 0.00002449, -0.00004169, 0.00001408, 0.00001945, -0.00003132, 0.00002303, -0.00000633, -0.00001513, 0.00002792, -0.00001990, 0.00000411, 0.00001511, -0.00003448, 0.00001774, 0.00001119, -0.00003349, 0.00001682, 0.00001455, -0.00001444, 0.00000061, 0.00000301, 0.00000543, -0.00000782, -0.00001333, 0.00001183, -0.00000307, -0.00001785, 0.00000902, 0.00000183, -0.00001831, 0.00000888, 0.00001578, -0.00001482, -0.00000077, 0.00001721, -0.00000985, -0.00000989, 0.00000638, -0.00000070, 0.00000392, 0.00000931, 0.00000235, -0.00000145, -0.00000472, -0.00000153, 0.00000421, 0.00000684, 0.00000355, -0.00001007, -0.00001170, -0.00000175, 0.00000402, 0.00000837, 0.00000121, -0.00000810, -0.00000582, -0.00000398, -0.00000084, 0.00000104, -0.00000157, -0.00000054, -0.00000007, 0.00000060, -0.00000266, -0.00000771, 0.00000052, 0.00000293, -0.00000303, 0.00000083, -0.00000208, -0.00000100, 0.00001226, 0.00000500, -0.00000666, },
		{ -0.00000160, 0.00000747, 0.00000558, -0.00000145, 0.00000107, 0.00000083, 0.00000140, 0.00000315, 0.00000238, -0.00000066, -0.00000485, 0.00000328, 0.00000052, -0.00000803, 0.00000288, -0.00001154, -0.00001501, 0.00001225, -0.00000364, -0.00000428, 0.00000570, -0.00001920, 0.00000309, -0.00000520, -0.00004057, 0.00000623, -0.00003064, -0.00014570, -0.00008837, 0.00019692, 0.00041366, 0.00016162, -0.00018314, -0.00010394, -0.00000588, -0.00004391, 0.00000030, -0.00001560, -0.00000895, 0.00002185, -0.00003301, 0.00000472, 0.00000414, -0.00003882, 0.00002855, -0.00000780, -0.00002209, 0.00004207, -0.00002269, -0.00000997, 0.00002162, -0.00003029, 0.00001882, -0.00000096, -0.00003518, 0.00003404, -0.00000779, -0.00001617, 0.00002466, -0.00002298, 0.00000772, 0.00001514, -0.00002388, 0.00002084, -0.00000762, -0.00002351, 0.00002940, -0.00001358, -0.00001932, 0.00002508, -0.00000016, -0.00000848, 0.00000064, 0.00000436, 0.00000564, -0.00001770, -0.00000224, 0.00001360, -0.00001703, -0.00000613, 0.00001358, -0.00001363, -0.00001002, 0.00002031, 0.00000175, -0.00001595, 0.00001254, 0.00000929, -0.00001678, -0.00000195, 0.00000591, -0.00000169, 0.00000735, 0.00000763, 0.00000048, -0.00000322, -0.00000458, 0.00000156, 0.00000514, 0.00000631, -0.00000114, -0.00001379, -0.00000735, 0.00000240, 0.00000520, 0.00000658, -0.00000320, -0.00000793, -0.00000379, -0.00000389, -0.00000024, 0.00000137, -0.00000153, -0.00000074, 0.00000040, 0.00000020, -0.00000523, -0.00000584, 0.00000292, 0.00000084, -0.00000189, 0.00000097, -0.00000392, 0.00000311, 0.00001247, 0.00000030, -0.00000680, -0.00000437, },
		{ 0.00000880, 0.00000190, -0.00000144, 0.00000183, 0.00000038, 0.00000239, 0.00000366, 0.00000162, -0.00000296, -0.00000310, 0.00000407, -0.00000363, -0.00000410, 0.00000036, -0.00001809, -0.00000383, 0.00001116, -0.00000925, 0.00000385, -0.00000368, -0.00001706, 0.00001061, -0.00002542, -0.00002804, 0.00001377, -0.00008412, -0.00014616, 0.00001864, 0.00028822, 0.00034703, 0.00002685, -0.00018391, -0.00005688, -0.00001451, -0.00003296, -0.00000033, -0.00001911, 0.00000911, 0.00000166, -0.00002820, 0.00001695, -0.00001967, -0.00001675, 0.00002821, -0.00002843, 0.00000863, 0.00002677, -0.00003501, 0.00001394, 0.00000041, -0.00002096, 0.00002766, -0.00002684, -0.00001050, 0.00003224, -0.00002596, 0.00000563, 0.00000994, -0.00002250, 0.00002382, -0.00000560, -0.00001167, 0.00002227, -0.00002576, -0.00000113, 0.00002171, -0.00002760, 0.00000131, 0.00001910, -0.00000643, -0.00000309, -0.00000189, 0.00001044, -0.00000265, -0.00002079, 0.00001291, 0.00000233, -0.00002165, 0.00001050, 0.00000368, -0.00002187, 0.00000720, 0.00001892, -0.00001222, -0.00000626, 0.00001891, -0.00000345, -0.00001611, 0.00000505, 0.00000277, -0.00000021, 0.00000954, 0.00000510, -0.00000130, -0.00000482, -0.00000249, 0.00000395, 0.00000509, 0.00000487, -0.00000703, -0.00001423, -0.00000107, 0.00000444, 0.00000502, 0.00000380, -0.00000652, -0.00000591, -0.00000273, -0.00000441, 0.00000115, 0.00000161, -0.00000222, -0.00000047, 0.00000104, -0.00000155, -0.00000656, -0.00000276, 0.00000334, -0.00000052, -0.00000013, -0.00000066, -0.00000402, 0.00000742, 0.00001028, -0.00000333, -0.00000655, -0.00000405, 0.00000203, },
		{ -0.00000112, -0.00000005, 0.00000168, 0.00000037, 0.00000387, 0.00000348, -0.00000018, -0.00000409, -0.00000071, 0.00000250, -0.00000512, -0.00000072, -0.00000651, -0.00001759, 0.00000641, 0.00000408, -0.00000782, 0.00000710, -0.00001416, -0.00000633, 0.00000535, -0.00003973, -0.00000371, -0.00000791, -0.00013221, -0.00009360, 0.00012718, 0.00031889, 0.00024917, -0.00006702, -0.00015805, -0.00002669, -0.00002073, -0.00002389, -0.00000501, -0.00001115, 0.00001315, -0.00001564, -0.00001209, 0.00001134, -0.00002982, 0.00000719, 0.00000973, -0.00002694, 0.00003035, -0.00000100, -0.00002549, 0.00002278, -0.00001898, 0.00000026, 0.00001630, -0.00003726, 0.00001631, 0.00001299, -0.00002607, 0.00001922, -0.00000920, -0.00000681, 0.00002364, -0.00002006, 0.00000677, 0.00000825, -0.00002823, 0.00001764, 0.00000182, -0.00002443, 0.00001508, 0.00000748, -0.00000402, -0.00000378, 0.00000009, 0.00001450, -0.00001687, -0.00001111, 0.00002064, -0.00001555, -0.00001106, 0.00001907, -0.00001461, -0.00001509, 0.00002170, 0.00000583, -0.00001706, 0.00000766, 0.00001517, -0.00001424, -0.00000941, 0.00000818, -0.00000052, 0.00000306, 0.00000985, 0.00000238, -0.00000324, -0.00000515, 0.00000076, 0.00000476, 0.00000476, 0.00000183, -0.00001271, -0.00001030, 0.00000471, 0.00000393, 0.00000434, 0.00000039, -0.00000799, -0.00000292, -0.00000343, -0.00000446, 0.00000340, 0.00000077, -0.00000304, 0.00000070, 0.00000075, -0.00000388, -0.00000588, -0.00000009, 0.00000239, -0.00000032, 0.00000074, -0.00000315, -0.00000187, 0.00001049, 0.00000660, -0.00000566, -0.00000634, -0.00000245, 0.00000596, 0.00000732, },
		{ 0.00000148, 0.00000073, 0.00000148, 0.00000512, 0.00000209, -0.00000204, -0.00000375, 0.00000078, 0.00000020, -0.00000399, -0.00000090, -0.00001337, -0.00001052, 0.00001077, -0.00000336, -0.00000162, 0.00000271, -0.00001837, 0.00000502, -0.00001147, -0.00003720, 0.00001441, -0.00005636, -0.00014493, -0.00000356, 0.00020371, 0.00030032, 0.00014882, -0.00012332, -0.00012094, -0.00000863, -0.00002552, -0.00001855, -0.00000573, -0.00000149, 0.00000545, -0.00002105, 0.00000172, -0.00000348, -0.00002310, 0.00001839, -0.00001093, -0.00000787, 0.00003152, -0.00002310, -0.00000459, 0.00001394, -0.00002411, 0.00001809, -0.00000657, -0.00002739, 0.00002995, -0.00000981, -0.00001167, 0.00001738, -0.00001934, 0.00001241, 0.00000919, -0.00001986, 0.00001827, -0.00001093, -0.00001518, 0.00002165, -0.00001617, -0.00000954, 0.00001596, 0.00000046, -0.00000002, -0.00000833, 0.00000907, 0.00000892, -0.00002728, 0.00000797, 0.00001293, -0.00002686, 0.00000929, 0.00001102, -0.00002791, 0.00000397, 0.00002337, -0.00000996, -0.00001021, 0.00001725, 0.00000417, -0.00001846, -0.00000080, 0.00000721, -0.00000198, 0.00000666, 0.00000844, -0.00000036, -0.00000499, -0.00000353, 0.00000362, 0.00000425, 0.00000432, -0.00000330, -0.00001573, -0.00000279, 0.00000744, 0.00000196, 0.00000365, -0.00000328, -0.00000703, -0.00000050, -0.00000574, -0.00000272, 0.00000546, -0.00000153, -0.00000283, 0.00000211, -0.00000105, -0.00000531, -0.00000376, 0.00000101, 0.00000159, 0.00000096, -0.00000026, -0.00000491, 0.00000191, 0.00001142, 0.00000259, -0.00000709, -0.00000586, 0.00000080, 0.00000840, 0.00000382, -0.00000215, },
		{ -0.00000000, 0.00000370, 0.00000509, -0.00000012, -0.00000300, -0.00000275, 0.00000098, -0.00000110, -0.00000236, -0.00000489, -0.00001600, -0.00000104, 0.00000856, -0.00000642, 0.00000385, -0.00000631, -0.00001354, 0.00000784, -0.00002963, -0.00001813, 0.00000808, -0.00010983, -0.00010706, 0.00009092, 0.00023541, 0.00025537, 0.00005909, -0.00014996, -0.00007965, -0.00000096, -0.00003051, -0.00001318, -0.00000213, 0.00000228, -0.00000502, -0.00001606, 0.00000551, -0.00001529, -0.00000779, 0.00001366, -0.00001970, 0.00001408, 0.00001491, -0.00002834, 0.00001257, -0.00000346, -0.00001371, 0.00002124, -0.00002599, -0.00000505, 0.00002506, -0.00002265, 0.00000537, 0.00000415, -0.00001462, 0.00002253, -0.00000862, -0.00000738, 0.00001579, -0.00002251, 0.00000305, 0.00001070, -0.00002150, 0.00000497, 0.00000830, 0.00000184, -0.00000167, -0.00000918, 0.00001954, -0.00000736, -0.00002395, 0.00002441, -0.00000756, -0.00002112, 0.00002523, -0.00001058, -0.00002484, 0.00002337, 0.00001082, -0.00001798, 0.00000319, 0.00001754, -0.00000788, -0.00001566, 0.00000589, 0.00000388, -0.00000104, 0.00000933, 0.00000578, -0.00000308, -0.00000561, -0.00000040, 0.00000485, 0.00000345, 0.00000293, -0.00000972, -0.00001385, 0.00000547, 0.00000607, 0.00000036, 0.00000257, -0.00000640, -0.00000380, -0.00000035, -0.00000821, 0.00000127, 0.00000568, -0.00000442, -0.00000090, 0.00000237, -0.00000368, -0.00000478, -0.00000177, 0.00000081, 0.00000211, 0.00000177, -0.00000282, -0.00000464, 0.00000590, 0.00001031, -0.00000103, -0.00000809, -0.00000425, 0.00000489, 0.00000806, 0.00000014, -0.00000115, 0.00000191, },
		{ 0.00000597, 0.00000325, -0.00000202, -0.00000294, -0.00000206, 0.00000070, -0.00000103, -0.00000260, -0.00001017, -0.00001300, 0.00000599, 0.00000300, -0.00000437, 0.00000420, -0.00001363, -0.00000378, -0.00000121, -0.00003673, 0.00000415, -0.00002782, -0.00013817, -0.00002817, 0.00015814, 0.00023167, 0.00020199, -0.00001799, -0.00015141, -0.00003948, -0.00000429, -0.00003338, -0.00000498, 0.00000099, -0.00000036, -0.00001091, -0.00000819, 0.00000076, -0.00001736, 0.00000454, 0.00000073, -0.00001296, 0.00002508, -0.00000684, -0.00001778, 0.00001647, -0.00001661, 0.00000324, 0.00000911, -0.00003085, 0.00001575, 0.00000781, -0.00002060, 0.00001391, -0.00000948, 0.00000029, 0.00001827, -0.00001834, 0.00000720, 0.00000255, -0.00002024, 0.00001398, -0.00000547, -0.00001361, 0.00000983, 0.00000186, 0.00000677, -0.00000957, -0.00000065, 0.00002165, -0.00002593, -0.00000528, 0.00002554, -0.00002763, 0.00000081, 0.00002349, -0.00003188, -0.00000466, 0.00003010, -0.00000765, -0.00001346, 0.00001445, 0.00000961, -0.00001562, -0.00000862, 0.00000886, 0.00000035, 0.00000176, 0.00001037, 0.00000225, -0.00000537, -0.00000436, 0.00000280, 0.00000436, 0.00000301, -0.00000056, -0.00001500, -0.00000670, 0.00001077, 0.00000181, 0.00000034, 0.00000037, -0.00000760, 0.00000033, -0.00000328, -0.00000857, 0.00000635, 0.00000294, -0.00000605, 0.00000208, 0.00000048, -0.00000542, -0.00000271, -0.00000129, 0.00000078, 0.00000380, 0.00000069, -0.00000543, -0.00000221, 0.00000877, 0.00000793, -0.00000419, -0.00000851, -0.00000086, 0.00000806, 0.00000510, -0.00000184, 0.00000065, 0.00000100, 0.00000070, },
		{ 0.00000025, -0.00000268, -0.00000266, -0.00000188, 0.00000099, -0.00000082, -0.00000544, -0.00001336, -0.00000633, 0.00000810, -0.00000159, -0.00000024, -0.00000072, -0.00001459, 0.00000302, -0.00001675, -0.00002755, 0.00001106, -0.00007925, -0.00012023, 0.00006219, 0.00018441, 0.00021232, 0.00014565, -0.00008151, -0.00012975, -0.00000806, -0.00001668, -0.00002965, 0.00000403, -0.00000031, -0.00000478, -0.00001105, -0.00000374, -0.00000576, -0.00001142, 0.00000746, -0.00000919, 0.00000227, 0.00002043, -0.00002075, -0.00000145, 0.00000783, -0.00001785, 0.00001492, -0.00000935, -0.00002012, 0.00002412, -0.00000985, -0.00000865, 0.00001056, -0.00001389, 0.00001466, 0.00000419, -0.00001554, 0.00001385, -0.00001139, -0.00000800, 0.00001181, -0.00001509, -0.00000117, 0.00000496, 0.00000330, 0.00000684, -0.00001667, 0.00001487, 0.00000940, -0.00003380, 0.00001888, 0.00000791, -0.00003205, 0.00002557, 0.00000190, -0.00003720, 0.00002128, 0.00001919, -0.00001983, -0.00000022, 0.00001734, -0.00000127, -0.00001719, -0.00000088, 0.00000838, -0.00000186, 0.00000542, 0.00000946, -0.00000177, -0.00000634, -0.00000147, 0.00000461, 0.00000317, 0.00000244, -0.00000633, -0.00001601, 0.00000333, 0.00001064, -0.00000249, 0.00000145, -0.00000296, -0.00000576, 0.00000291, -0.00000828, -0.00000511, 0.00000997, -0.00000212, -0.00000482, 0.00000416, -0.00000307, -0.00000478, -0.00000083, -0.00000221, 0.00000240, 0.00000507, -0.00000242, -0.00000638, 0.00000135, 0.00001002, 0.00000499, -0.00000717, -0.00000752, 0.00000383, 0.00000856, 0.00000121, -0.00000144, 0.00000155, -0.00000019, 0.00000319, 0.00000668, },
		{ -0.00000218, -0.00000296, -0.00000143, 0.00000191, -0.00000211, -0.00000945, -0.00001258, 0.00000040, 0.00000617, -0.00000292, 0.00000201, -0.00000714, -0.00000963, 0.00000124, -0.00002824, -0.00000888, -0.00000862, -0.00011908, -0.00005697, 0.00013100, 0.00017822, 0.00019211, 0.00008357, -0.00012580, -0.00009030, 0.00000684, -0.00003116, -0.00001770, 0.00000876, -0.00000522, -0.00000666, -0.00000895, -0.00000369, -0.00000856, -0.00000409, 0.00000269, -0.00000952, 0.00001494, 0.00000557, -0.00002126, 0.00000972, -0.00000506, -0.00000851, 0.00001457, -0.00002259, -0.00000186, 0.00001817, -0.00001840, 0.00000342, 0.00000048, -0.00000703, 0.00001946, -0.00000980, -0.00000460, 0.00000949, -0.00001699, 0.00000412, 0.00000066, -0.00001291, 0.00000544, -0.00000174, 0.00001086, -0.00000203, -0.00001417, 0.00002723, -0.00001311, -0.00002291, 0.00003276, -0.00001862, -0.00001475, 0.00003547, -0.00002757, -0.00002017, 0.00003670, -0.00000257, -0.00001810, 0.00001263, 0.00001170, -0.00000976, -0.00001404, 0.00000524, 0.00000564, -0.00000208, 0.00000889, 0.00000646, -0.00000542, -0.00000524, 0.00000178, 0.00000459, 0.00000230, 0.00000037, -0.00001255, -0.00001095, 0.00001189, 0.00000543, -0.00000423, 0.00000201, -0.00000606, -0.00000118, 0.00000166, -0.00001249, 0.00000189, 0.00000958, -0.00000706, -0.00000081, 0.00000340, -0.00000618, -0.00000186, -0.00000101, -0.00000283, 0.00000575, 0.00000402, -0.00000610, -0.00000500, 0.00000464, 0.00001001, 0.00000161, -0.00000975, -0.00000418, 0.00000807, 0.00000597, -0.00000136, 0.00000032, 0.00000079, 0.00000002, 0.00000634, 0.00000489, -0.00000236, },
		{ -0.00000377, 0.00000009, 0.00000233, -0.00000540, -0.00001217, -0.00000845, 0.00000450, 0.00000296, -0.00000170, 0.00000064, -0.00001076, -0.00000369, -0.00000815, -0.00002784, 0.00000408, -0.00004974, -0.00012226, 0.00002841, 0.00015953, 0.00016077, 0.00017225, 0.00001470, -0.00014273, -0.00004484, 0.00000282, -0.00003845, -0.00000167, 0.00000596, -0.00000945, -0.00000562, -0.00000820, -0.00000492, -0.00000707, -0.00000071, -0.00000314, -0.00000164, 0.00001769, -0.00000940, -0.00001174, 0.00001063, -0.00001319, 0.00000365, 0.00000389, -0.00002386, 0.00001305, 0.00000424, -0.00001589, 0.00000831, -0.00000761, 0.00000539, 0.00001298, -0.00001564, 0.00000570, -0.00000097, -0.00001241, 0.00000772, -0.00000953, -0.00000326, 0.00000223, -0.00000159, 0.00001610, -0.00001532, 0.00000052, 0.00002569, -0.00003274, 0.00000266, 0.00002555, -0.00003618, 0.00001536, 0.00002092, -0.00004573, 0.00001092, 0.00003134, -0.00002114, -0.00000443, 0.00001730, 0.00000241, -0.00001365, -0.00000863, 0.00000892, 0.00000189, -0.00000011, 0.00001104, 0.00000164, -0.00000742, -0.00000232, 0.00000388, 0.00000353, 0.00000172, -0.00000412, -0.00001595, -0.00000096, 0.00001495, -0.00000173, -0.00000281, 0.00000055, -0.00000678, 0.00000387, -0.00000390, -0.00001257, 0.00000967, 0.00000449, -0.00000890, 0.00000379, -0.00000049, -0.00000649, 0.00000130, -0.00000352, -0.00000100, 0.00000901, 0.00000005, -0.00000829, -0.00000202, 0.00000691, 0.00000923, -0.00000255, -0.00001082, 0.00000137, 0.00000964, 0.00000174, -0.00000133, 0.00000147, -0.00000077, 0.00000254, 0.00000808, 0.00000118, -0.00000322, -0.00000162, },
		{ 0.00000253, 0.00000080, -0.00000952, -0.00001196, -0.00000326, 0.00000549, 0.00000073, -0.00000032, -0.00000299, -0.00001011, -0.00000206, -0.00001832, -0.00001668, -0.00000220, -0.00009189, -0.00008052, 0.00010312, 0.00015144, 0.00015026, 0.00014350, -0.00005334, -0.00012904, -0.00000848, -0.00001391, -0.00003295, 0.00001066, -0.00000252, -0.00000944, -0.00000458, -0.00000913, -0.00000436, -0.00000456, -0.00000175, -0.00000439, 0.00000782, 0.00001059, -0.00001672, -0.00000027, 0.00000358, -0.00001259, 0.00001056, -0.00000952, -0.00001431, 0.00001775, -0.00000857, -0.00000721, 0.00000524, -0.00000788, 0.00001477, 0.00000067, -0.00001195, 0.00000886, -0.00000941, -0.00000334, 0.00000197, -0.00001102, 0.00000437, -0.00000587, 0.00000741, 0.00001137, -0.00002275, 0.00002018, 0.00000764, -0.00003584, 0.00002785, 0.00000007, -0.00003105, 0.00003893, -0.00001178, -0.00003887, 0.00003768, 0.00000825, -0.00002480, 0.00001145, 0.00001241, -0.00000500, -0.00001394, -0.00000258, 0.00000995, -0.00000168, 0.00000379, 0.00001061, -0.00000385, -0.00000672, 0.00000107, 0.00000424, 0.00000256, 0.00000022, -0.00001009, -0.00001363, 0.00000988, 0.00001119, -0.00000690, 0.00000002, -0.00000266, -0.00000376, 0.00000598, -0.00001144, -0.00000686, 0.00001421, -0.00000321, -0.00000618, 0.00000588, -0.00000548, -0.00000318, 0.00000196, -0.00000643, 0.00000396, 0.00000953, -0.00000545, -0.00000774, 0.00000107, 0.00000842, 0.00000759, -0.00000738, -0.00000890, 0.00000735, 0.00000744, -0.00000155, 0.00000062, 0.00000068, -0.00000111, 0.00000649, 0.00000679, -0.00000251, -0.00000228, -0.00000212, -0.00000404, },
		{ -0.00000308, -0.00001240, -0.00000900, 0.00000081, 0.00000465, -0.00000005, -0.00000048, -0.00000614, -0.00000736, -0.00000613, -0.00002194, -0.00000473, -0.00002970, -0.00011041, -0.00000758, 0.00014247, 0.00012723, 0.00015066, 0.00009689, -0.00010585, -0.00009178, 0.00000777, -0.00003121, -0.00001672, 0.00001322, -0.00001088, -0.00000559, -0.00000619, -0.00000927, -0.00000198, -0.00000423, -0.00000371, -0.00000022, 0.00001240, -0.00000064, -0.00001482, 0.00000629, -0.00000508, -0.00000554, 0.00000886, -0.00001780, -0.00000081, 0.00001234, -0.00001425, 0.00000065, -0.00000099, -0.00000080, 0.00001565, -0.00000960, -0.00000353, 0.00000465, -0.00001081, 0.00000228, -0.00000669, -0.00000387, 0.00000309, -0.00000958, 0.00001878, -0.00000334, -0.00001639, 0.00003197, -0.00001792, -0.00001829, 0.00003616, -0.00002774, -0.00000392, 0.00003877, -0.00004311, -0.00000870, 0.00004351, -0.00001800, -0.00001199, 0.00001921, 0.00000299, -0.00000826, -0.00001258, 0.00000354, 0.00000814, -0.00000360, 0.00000852, 0.00000683, -0.00000798, -0.00000349, 0.00000326, 0.00000354, 0.00000191, -0.00000342, -0.00001462, -0.00000529, 0.00001660, 0.00000289, -0.00000772, 0.00000157, -0.00000543, 0.00000205, 0.00000271, -0.00001667, 0.00000308, 0.00001263, -0.00000953, -0.00000025, 0.00000360, -0.00000822, 0.00000190, -0.00000122, -0.00000660, 0.00001028, 0.00000575, -0.00000982, -0.00000494, 0.00000330, 0.00000974, 0.00000428, -0.00001157, -0.00000340, 0.00001094, 0.00000256, -0.00000189, 0.00000229, -0.00000151, 0.00000131, 0.00000944, 0.00000265, -0.00000418, -0.00000088, -0.00000377, -0.00000260, 0.00000447, },
		{ -0.00001255, -0.00000490, 0.00000297, 0.00000348, -0.00000036, -0.00000202, -0.00000721, -0.00000609, -0.00001244, -0.00001709, -0.00000418, -0.00006601, -0.00009059, 0.00006828, 0.00014172, 0.00010967, 0.00015152, 0.00003348, -0.00012936, -0.00004699, 0.00000300, -0.00003792, 0.00000169, 0.00000592, -0.00001377, -0.00000178, -0.00000997, -0.00000647, -0.00000040, -0.00000617, -0.00000291, 0.00000583, 0.00000999, -0.00000936, -0.00000757, 0.00000587, -0.00000979, 0.00000233, 0.00000092, -0.00001745, 0.00000917, 0.00000209, -0.00001245, 0.00000353, -0.00000449, 0.00000829, 0.00000859, -0.00001291, 0.00000314, -0.00000211, -0.00000639, 0.00000052, -0.00001024, 0.00000477, -0.00000595, -0.00000302, 0.00002267, -0.00001945, 0.00000255, 0.00002648, -0.00003542, 0.00001040, 0.00002121, -0.00003880, 0.00002906, 0.00001208, -0.00005259, 0.00002780, 0.00002472, -0.00003096, 0.00000796, 0.00001502, -0.00000412, -0.00000903, -0.00001012, 0.00000905, 0.00000371, -0.00000244, 0.00001193, 0.00000033, -0.00000879, 0.00000061, 0.00000354, 0.00000296, 0.00000053, -0.00000890, -0.00001443, 0.00000601, 0.00001615, -0.00000551, -0.00000486, 0.00000040, -0.00000511, 0.00000729, -0.00000547, -0.00001569, 0.00001280, 0.00000522, -0.00001092, 0.00000522, -0.00000206, -0.00000636, 0.00000516, -0.00000673, -0.00000199, 0.00001440, -0.00000140, -0.00001093, -0.00000165, 0.00000495, 0.00001075, -0.00000121, -0.00001286, 0.00000426, 0.00001004, -0.00000210, 0.00000055, 0.00000166, -0.00000286, 0.00000607, 0.00000906, -0.00000223, -0.00000326, -0.00000069, -0.00000531, 0.00000059, 0.00000416, -0.00000812, },
		{ -0.00000136, 0.00000372, 0.00000253, -0.00000110, -0.00000362, -0.00000697, -0.00000795, -0.00001572, -0.00000922, -0.00002056, -0.00009073, -0.00003632, 0.00011954, 0.00011581, 0.00011005, 0.00013737, -0.00003375, -0.00012037, -0.00001202, -0.00001357, -0.00003026, 0.00001289, -0.00000546, -0.00001021, -0.00000181, -0.00001268, -0.00000141, -0.00000196, -0.00000765, 0.00000136, 0.00000928, 0.00000311, -0.00001213, -0.00000043, 0.00000094, -0.00000890, 0.00000622, -0.00000784, -0.00001033, 0.00001177, -0.00000682, -0.00000692, 0.00000191, -0.00000247, 0.00001345, -0.00000139, -0.00000963, 0.00000459, -0.00000611, -0.00000163, -0.00000596, -0.00000549, 0.00000667, -0.00001471, 0.00001129, 0.00001370, -0.00002566, 0.00002343, 0.00000514, -0.00003320, 0.00003261, -0.00000793, -0.00002405, 0.00004616, -0.00002639, -0.00003245, 0.00004882, -0.00000685, -0.00002321, 0.00002164, 0.00000357, -0.00000561, -0.00001016, -0.00000559, 0.00001222, -0.00000185, 0.00000215, 0.00001161, -0.00000644, -0.00000574, 0.00000333, 0.00000271, 0.00000285, -0.00000297, -0.00001380, -0.00000810, 0.00001528, 0.00000916, -0.00001008, -0.00000116, -0.00000243, -0.00000076, 0.00000803, -0.00001476, -0.00000760, 0.00001739, -0.00000434, -0.00000670, 0.00000659, -0.00000753, -0.00000052, 0.00000362, -0.00001071, 0.00000649, 0.00001324, -0.00000881, -0.00000861, 0.00000060, 0.00000701, 0.00001017, -0.00000796, -0.00000944, 0.00001077, 0.00000487, -0.00000359, 0.00000343, -0.00000133, -0.00000105, 0.00001061, 0.00000483, -0.00000521, -0.00000103, -0.00000236, -0.00000507, 0.00000410, 0.00000083, -0.00001215, -0.00001004, },
		{ 0.00000389, 0.00000148, -0.00000225, -0.00000438, -0.00000739, -0.00001132, -0.00001356, -0.00000741, -0.00004800, -0.00008714, 0.00003208, 0.00013317, 0.00008821, 0.00012334, 0.00009957, -0.00008655, -0.00008793, 0.00000307, -0.00002861, -0.00001333, 0.00001252, -0.00001354, -0.00000393, -0.00000618, -0.00001121, 0.00000263, -0.00000622, -0.00000580, 0.00000658, 0.00000806, -0.00000384, -0.00000965, 0.00000299, -0.00000438, -0.00000439, 0.00000478, -0.00001277, -0.00000131, 0.00000777, -0.00001090, -0.00000197, -0.00000076, 0.00000350, 0.00001200, -0.00000876, -0.00000376, 0.00000189, -0.00000561, -0.00000134, -0.00001061, 0.00000365, -0.00000085, -0.00001465, 0.00002406, -0.00000436, -0.00001631, 0.00003308, -0.00001979, -0.00001186, 0.00003422, -0.00003208, 0.00000852, 0.00003461, -0.00005308, 0.00000652, 0.00004185, -0.00003144, -0.00000099, 0.00002090, -0.00000618, -0.00000366, -0.00001233, 0.00000163, 0.00001106, -0.00000563, 0.00000845, 0.00000651, -0.00001024, -0.00000054, 0.00000345, 0.00000236, 0.00000207, -0.00000865, -0.00001474, 0.00000244, 0.00001830, -0.00000050, -0.00000968, 0.00000058, -0.00000409, 0.00000561, 0.00000240, -0.00001992, 0.00000450, 0.00001435, -0.00001104, 0.00000033, 0.00000281, -0.00000890, 0.00000566, -0.00000271, -0.00000932, 0.00001497, 0.00000639, -0.00001308, -0.00000474, 0.00000177, 0.00000990, 0.00000645, -0.00001328, -0.00000164, 0.00001269, -0.00000162, -0.00000107, 0.00000385, -0.00000434, 0.00000428, 0.00001177, -0.00000128, -0.00000489, 0.00000033, -0.00000482, -0.00000239, 0.00000577, -0.00000455, -0.00001350, -0.00000622, 0.00000090, },
		{ 0.00000002, -0.00000306, -0.00000481, -0.00000947, -0.00001281, -0.00000880, -0.00001767, -0.00007176, -0.00005192, 0.00008898, 0.00011526, 0.00007687, 0.00013375, 0.00004253, -0.00011153, -0.00004880, -0.00000052, -0.00003257, 0.00000313, 0.00000335, -0.00001426, -0.00000020, -0.00001162, -0.00000548, 0.00000248, -0.00001001, -0.00000029, 0.00000931, 0.00000341, -0.00000765, -0.00000505, 0.00000232, -0.00000714, 0.00000032, -0.00000010, -0.00001232, 0.00000505, 0.00000081, -0.00001022, 0.00000030, -0.00000117, 0.00000934, 0.00000542, -0.00001087, 0.00000060, -0.00000148, -0.00000307, -0.00000577, -0.00000853, 0.00000944, -0.00001287, -0.00000336, 0.00002600, -0.00002083, 0.00000409, 0.00002495, -0.00003316, 0.00001552, 0.00001450, -0.00003464, 0.00003835, -0.00000026, -0.00005073, 0.00004140, 0.00001183, -0.00003437, 0.00002048, 0.00000796, -0.00000799, -0.00000325, -0.00001308, 0.00000981, 0.00000524, -0.00000487, 0.00001288, -0.00000173, -0.00000895, 0.00000371, 0.00000178, 0.00000320, -0.00000122, -0.00001429, -0.00000969, 0.00001267, 0.00001415, -0.00000823, -0.00000621, -0.00000030, -0.00000218, 0.00000989, -0.00000775, -0.00001732, 0.00001495, 0.00000534, -0.00001162, 0.00000562, -0.00000366, -0.00000474, 0.00000773, -0.00001024, -0.00000157, 0.00001878, -0.00000334, -0.00001274, -0.00000167, 0.00000321, 0.00001243, -0.00000061, -0.00001398, 0.00000742, 0.00000877, -0.00000545, 0.00000337, 0.00000073, -0.00000418, 0.00001070, 0.00000805, -0.00000600, -0.00000216, -0.00000057, -0.00000606, 0.00000168, 0.00000421, -0.00000981, -0.00001190, -0.00000267, 0.00000214, 0.00000377, },
		{ -0.00000321, -0.00000615, -0.00001202, -0.00001066, -0.00000778, -0.00003787, -0.00007667, 0.00000306, 0.00011656, 0.00008488, 0.00008538, 0.00012600, -0.00001867, -0.00010686, -0.00001859, -0.00001303, -0.00002429, 0.00001109, -0.00000744, -0.00000906, -0.00000181, -0.00001375, 0.00000120, -0.00000214, -0.00000987, 0.00000626, 0.00000787, -0.00000164, -0.00000783, -0.00000132, -0.00000065, -0.00000674, 0.00000277, -0.00000527, -0.00000797, 0.00000670, -0.00000534, -0.00000704, 0.00000046, 0.00000155, 0.00001154, -0.00000240, -0.00000856, 0.00000187, -0.00000290, -0.00000236, -0.00001078, -0.00000035, 0.00000634, -0.00002051, 0.00001346, 0.00001461, -0.00002537, 0.00002365, 0.00000379, -0.00002702, 0.00003202, -0.00001312, -0.00001306, 0.00004565, -0.00003772, -0.00001990, 0.00005162, -0.00002214, -0.00001533, 0.00002755, -0.00000692, -0.00000278, -0.00000730, -0.00000906, 0.00001502, -0.00000264, 0.00000090, 0.00001206, -0.00000913, -0.00000334, 0.00000456, 0.00000064, 0.00000399, -0.00000759, -0.00001626, 0.00000020, 0.00001799, 0.00000548, -0.00001136, -0.00000278, -0.00000191, 0.00000315, 0.00000845, -0.00001746, -0.00000739, 0.00001882, -0.00000481, -0.00000657, 0.00000588, -0.00000830, 0.00000261, 0.00000349, -0.00001384, 0.00000960, 0.00001534, -0.00001159, -0.00000906, -0.00000038, 0.00000628, 0.00001223, -0.00000883, -0.00000863, 0.00001324, 0.00000116, -0.00000425, 0.00000602, -0.00000405, 0.00000072, 0.00001429, 0.00000091, -0.00000694, 0.00000050, -0.00000320, -0.00000471, 0.00000491, -0.00000021, -0.00001292, -0.00000845, -0.00000018, 0.00000290, 0.00000310, -0.00000153, },
		{ -0.00000885, -0.00001263, -0.00000679, -0.00001556, -0.00005805, -0.00005596, 0.00005741, 0.00011245, 0.00006214, 0.00010367, 0.00009457, -0.00006685, -0.00008196, -0.00000469, -0.00002316, -0.00000992, 0.00000874, -0.00001337, -0.00000297, -0.00000721, -0.00001051, 0.00000455, -0.00000819, -0.00000478, 0.00001021, 0.00000342, -0.00000466, -0.00000589, 0.00000019, -0.00000371, -0.00000433, 0.00000242, -0.00000836, -0.00000259, 0.00000425, -0.00000859, -0.00000365, 0.00000038, 0.00000585, 0.00000907, -0.00000795, -0.00000450, 0.00000104, -0.00000253, -0.00000519, -0.00001139, 0.00000824, -0.00000478, -0.00001731, 0.00002588, -0.00000399, -0.00001503, 0.00003101, -0.00001747, -0.00000595, 0.00002817, -0.00002995, 0.00001905, 0.00002467, -0.00005475, 0.00002108, 0.00003211, -0.00003931, 0.00001205, 0.00001712, -0.00001332, 0.00000277, -0.00001371, 0.00000034, 0.00001376, -0.00000792, 0.00000870, 0.00000528, -0.00001155, 0.00000306, 0.00000210, 0.00000176, 0.00000211, -0.00001497, -0.00001190, 0.00001080, 0.00001644, -0.00000334, -0.00001034, -0.00000130, -0.00000152, 0.00000887, 0.00000072, -0.00002146, 0.00000547, 0.00001470, -0.00001102, 0.00000028, 0.00000134, -0.00000767, 0.00000828, -0.00000512, -0.00001010, 0.00001873, 0.00000596, -0.00001507, -0.00000492, 0.00000024, 0.00001064, 0.00000755, -0.00001425, 0.00000094, 0.00001259, -0.00000559, 0.00000095, 0.00000427, -0.00000657, 0.00000864, 0.00001220, -0.00000596, -0.00000424, 0.00000089, -0.00000553, -0.00000117, 0.00000539, -0.00000564, -0.00001310, -0.00000467, 0.00000134, 0.00000332, 0.00000180, -0.00000251, -0.00000338, },
		{ -0.00001010, -0.00000596, -0.00003149, -0.00006598, -0.00001503, 0.00009288, 0.00008836, 0.00005782, 0.00011608, 0.00004596, -0.00009132, -0.00005124, -0.00000487, -0.00002445, 0.00000216, 0.00000026, -0.00001248, -0.00000064, -0.00001209, -0.00000371, 0.00000247, -0.00001134, 0.00000294, 0.00000952, -0.00000127, -0.00000517, -0.00000375, -0.00000029, -0.00000550, -0.00000150, 0.00000014, -0.00000863, 0.00000133, -0.00000021, -0.00000878, -0.00000121, 0.00000149, 0.00000918, 0.00000338, -0.00000973, -0.00000104, -0.00000016, -0.00000243, -0.00000976, -0.00000587, 0.00001076, -0.00001723, -0.00000382, 0.00002648, -0.00001916, 0.00000400, 0.00002278, -0.00002648, 0.00001634, 0.00000822, -0.00002476, 0.00004081, -0.00001218, -0.00004092, 0.00004779, -0.00000383, -0.00003050, 0.00002981, -0.00000208, -0.00000816, 0.00000197, -0.00001679, 0.00001136, 0.00000597, -0.00000690, 0.00001353, -0.00000430, -0.00000760, 0.00000617, -0.00000091, 0.00000443, -0.00000418, -0.00001912, -0.00000177, 0.00001728, 0.00000968, -0.00000899, -0.00000761, -0.00000123, 0.00000209, 0.00001093, -0.00001008, -0.00001729, 0.00001537, 0.00000541, -0.00001100, 0.00000446, -0.00000442, -0.00000191, 0.00000812, -0.00001294, 0.00000021, 0.00002108, -0.00000502, -0.00001356, -0.00000253, 0.00000223, 0.00001392, -0.00000088, -0.00001352, 0.00001006, 0.00000587, -0.00000738, 0.00000621, -0.00000119, -0.00000373, 0.00001523, 0.00000482, -0.00000902, -0.00000039, -0.00000125, -0.00000562, 0.00000259, 0.00000280, -0.00000996, -0.00001096, -0.00000149, 0.00000224, 0.00000317, 0.00000029, -0.00000277, -0.00000465, -0.00001164, },
		{ -0.00001224, -0.00004868, -0.00005426, 0.00003168, 0.00010174, 0.00006178, 0.00006996, 0.00011082, -0.00000598, -0.00009169, -0.00002655, -0.00001120, -0.00001753, 0.00000671, -0.00000791, -0.00000750, -0.00000317, -0.00001272, 0.00000236, -0.00000341, -0.00000903, 0.00000921, 0.00000501, -0.00000388, -0.00000426, -0.00000245, -0.00000180, -0.00000569, 0.00000058, -0.00000270, -0.00000674, 0.00000261, -0.00000450, -0.00000686, 0.00000035, 0.00000391, 0.00000968, -0.00000292, -0.00000823, 0.00000087, -0.00000098, -0.00000432, -0.00001229, 0.00000290, 0.00000459, -0.00002312, 0.00001297, 0.00001506, -0.00002275, 0.00002087, 0.00000515, -0.00001940, 0.00002653, -0.00001322, -0.00000125, 0.00003798, -0.00004232, -0.00000510, 0.00004524, -0.00003352, -0.00000337, 0.00002763, -0.00001632, 0.00000265, -0.00000627, -0.00001193, 0.00001782, -0.00000403, 0.00000028, 0.00001156, -0.00001128, 0.00000013, 0.00000429, -0.00000119, 0.00000542, -0.00001356, -0.00001620, 0.00000999, 0.00001710, 0.00000158, -0.00001083, -0.00000536, -0.00000046, 0.00000738, 0.00000701, -0.00001870, -0.00000676, 0.00001828, -0.00000396, -0.00000636, 0.00000385, -0.00000705, 0.00000520, 0.00000162, -0.00001475, 0.00001235, 0.00001553, -0.00001289, -0.00000953, -0.00000184, 0.00000657, 0.00001316, -0.00000949, -0.00000637, 0.00001386, -0.00000286, -0.00000333, 0.00000741, -0.00000646, 0.00000425, 0.00001616, -0.00000419, -0.00000720, 0.00000165, -0.00000413, -0.00000313, 0.00000443, -0.00000154, -0.00001200, -0.00000770, 0.00000086, 0.00000259, 0.00000245, -0.00000079, -0.00000303, -0.00000725, -0.00001243, -0.00000613, },
		{ -0.00005756, -0.00002453, 0.00006840, 0.00008856, 0.00004652, 0.00008763, 0.00008548, -0.00004763, -0.00007569, -0.00001280, -0.00001601, -0.00000771, 0.00000389, -0.00001139, -0.00000304, -0.00000800, -0.00000858, 0.00000410, -0.00000893, -0.00000220, 0.00001088, -0.00000043, -0.00000392, -0.00000338, -0.00000211, -0.00000346, -0.00000454, 0.00000138, -0.00000499, -0.00000402, 0.00000137, -0.00000712, -0.00000405, 0.00000160, 0.00000666, 0.00000699, -0.00000754, -0.00000494, 0.00000135, -0.00000188, -0.00000785, -0.00001014, 0.00000949, -0.00000733, -0.00001848, 0.00002431, -0.00000175, -0.00001377, 0.00002721, -0.00001089, -0.00000262, 0.00002044, -0.00002149, 0.00002468, 0.00001234, -0.00004769, 0.00003053, 0.00001700, -0.00003960, 0.00002356, 0.00000895, -0.00001668, 0.00000919, -0.00001635, 0.00000020, 0.00001554, -0.00001000, 0.00000909, 0.00000316, -0.00001144, 0.00000648, -0.00000049, 0.00000215, 0.00000124, -0.00002133, -0.00000571, 0.00001790, 0.00001146, -0.00000449, -0.00001050, -0.00000386, 0.00000257, 0.00001096, -0.00000181, -0.00002092, 0.00000523, 0.00001410, -0.00000927, -0.00000103, -0.00000004, -0.00000458, 0.00000875, -0.00000752, -0.00000866, 0.00002039, 0.00000501, -0.00001542, -0.00000604, -0.00000082, 0.00001175, 0.00000725, -0.00001375, 0.00000370, 0.00001036, -0.00000830, 0.00000345, 0.00000330, -0.00000703, 0.00001337, 0.00001012, -0.00001018, -0.00000257, 0.00000059, -0.00000539, 0.00000038, 0.00000349, -0.00000574, -0.00001188, -0.00000419, 0.00000238, 0.00000231, 0.00000164, -0.00000137, -0.00000429, -0.00000991, -0.00001026, -0.00000470, -0.00002549, },
		{ 0.00001362, 0.00008549, 0.00006569, 0.00004715, 0.00009837, 0.00004653, -0.00007140, -0.00005370, -0.00000831, -0.00001580, -0.00000054, -0.00000226, -0.00000991, -0.00000215, -0.00001125, -0.00000243, 0.00000095, -0.00001027, 0.00000547, 0.00000764, -0.00000390, -0.00000258, -0.00000323, -0.00000237, -0.00000470, -0.00000258, 0.00000088, -0.00000616, -0.00000173, -0.00000138, -0.00000752, -0.00000131, 0.00000304, 0.00000843, 0.00000208, -0.00000925, -0.00000141, 0.00000069, -0.00000364, -0.00001088, -0.00000378, 0.00000960, -0.00001854, -0.00000543, 0.00002498, -0.00001510, 0.00000194, 0.00002178, -0.00001711, 0.00001257, 0.00000503, -0.00001180, 0.00003596, -0.00001979, -0.00002625, 0.00004494, -0.00001794, -0.00002062, 0.00003353, -0.00001240, -0.00000482, 0.00000522, -0.00002016, 0.00001336, 0.00000566, -0.00000809, 0.00001344, -0.00000687, -0.00000484, 0.00000730, -0.00000373, 0.00000629, -0.00000862, -0.00002195, 0.00000834, 0.00001854, 0.00000419, -0.00000763, -0.00000978, -0.00000170, 0.00000726, 0.00000994, -0.00001156, -0.00001594, 0.00001365, 0.00000612, -0.00000953, 0.00000167, -0.00000352, 0.00000123, 0.00000610, -0.00001366, 0.00000264, 0.00002079, -0.00000550, -0.00001369, -0.00000437, 0.00000245, 0.00001458, -0.00000170, -0.00001113, 0.00001119, 0.00000195, -0.00000738, 0.00000801, -0.00000331, -0.00000107, 0.00001818, -0.00000009, -0.00001049, 0.00000132, -0.00000235, -0.00000397, 0.00000269, 0.00000066, -0.00000858, -0.00001030, -0.00000081, 0.00000288, 0.00000159, 0.00000128, -0.00000223, -0.00000662, -0.00001075, -0.00000667, -0.00000925, -0.00004061, -0.00005135, },
		{ 0.00008261, 0.00004603, 0.00005925, 0.00009444, 0.00000460, -0.00007704, -0.00003373, -0.00000837, -0.00001163, 0.00000145, -0.00000724, -0.00000628, -0.00000466, -0.00001061, 0.00000193, -0.00000442, -0.00000638, 0.00000998, 0.00000191, -0.00000424, -0.00000156, -0.00000365, -0.00000294, -0.00000511, -0.00000047, -0.00000065, -0.00000613, -0.00000073, -0.00000427, -0.00000594, 0.00000084, 0.00000477, 0.00000819, -0.00000336, -0.00000798, 0.00000112, -0.00000092, -0.00000617, -0.00001122, 0.00000351, 0.00000276, -0.00002320, 0.00000966, 0.00001575, -0.00001910, 0.00001623, 0.00000989, -0.00001263, 0.00001808, -0.00000747, 0.00000808, 0.00002578, -0.00003865, 0.00000732, 0.00003137, -0.00003795, 0.00000925, 0.00002206, -0.00002230, 0.00000900, -0.00000719, -0.00001335, 0.00001985, -0.00000566, 0.00000029, 0.00000993, -0.00001229, 0.00000394, 0.00000252, -0.00000212, 0.00000627, -0.00002012, -0.00001280, 0.00001935, 0.00001270, -0.00000110, -0.00000926, -0.00000889, 0.00000242, 0.00001102, 0.00000402, -0.00001787, -0.00000649, 0.00001606, -0.00000147, -0.00000677, 0.00000119, -0.00000364, 0.00000614, -0.00000125, -0.00001291, 0.00001359, 0.00001417, -0.00001219, -0.00001061, -0.00000337, 0.00000780, 0.00001246, -0.00000919, -0.00000307, 0.00001206, -0.00000606, -0.00000131, 0.00000705, -0.00000732, 0.00000882, 0.00001528, -0.00000921, -0.00000606, 0.00000197, -0.00000454, -0.00000092, 0.00000259, -0.00000243, -0.00001001, -0.00000773, 0.00000208, 0.00000216, 0.00000121, 0.00000118, -0.00000426, -0.00000870, -0.00000897, -0.00000508, -0.00001993, -0.00005000, -0.00002982, 0.00004777, },
		{ 0.00003733, 0.00007373, 0.00007515, -0.00003069, -0.00006940, -0.00001933, -0.00000891, -0.00000675, -0.00000066, -0.00000880, -0.00000373, -0.00000805, -0.00000670, 0.00000227, -0.00000811, 0.00000053, 0.00000944, -0.00000298, -0.00000235, -0.00000183, -0.00000410, -0.00000366, -0.00000447, 0.00000111, -0.00000267, -0.00000525, -0.00000124, -0.00000603, -0.00000327, 0.00000231, 0.00000650, 0.00000560, -0.00000750, -0.00000460, 0.00000179, -0.00000320, -0.00000849, -0.00000825, 0.00000790, -0.00000778, -0.00001918, 0.00002014, 0.00000199, -0.00001330, 0.00002359, -0.00000129, -0.00000291, 0.00001390, -0.00000858, 0.00002396, 0.00000161, -0.00003401, 0.00003182, 0.00000069, -0.00003248, 0.00003044, -0.00000122, -0.00001574, 0.00001385, -0.00001926, 0.00000119, 0.00001591, -0.00001133, 0.00000919, 0.00000048, -0.00000973, 0.00000887, -0.00000365, 0.00000347, -0.00000115, -0.00002629, 0.00000327, 0.00002207, 0.00000472, -0.00000345, -0.00001107, -0.00000640, 0.00000794, 0.00001117, -0.00000429, -0.00001845, 0.00000320, 0.00001328, -0.00000617, -0.00000386, -0.00000041, -0.00000037, 0.00000658, -0.00000870, -0.00000550, 0.00001925, 0.00000444, -0.00001432, -0.00000836, -0.00000087, 0.00001265, 0.00000568, -0.00001121, 0.00000568, 0.00000633, -0.00000895, 0.00000534, 0.00000142, -0.00000493, 0.00001699, 0.00000575, -0.00001284, -0.00000057, -0.00000039, -0.00000416, 0.00000150, 0.00000068, -0.00000468, -0.00001047, -0.00000427, 0.00000375, 0.00000063, 0.00000187, 0.00000024, -0.00000723, -0.00000873, -0.00000613, -0.00000845, -0.00003276, -0.00004779, 0.00000076, 0.00006861, 0.00006693, },
		{ 0.00008200, 0.00004535, -0.00005382, -0.00005478, -0.00001046, -0.00000836, -0.00000378, -0.00000402, -0.00000750, -0.00000375, -0.00000966, -0.00000214, -0.00000077, -0.00000762, 0.00000667, 0.00000490, -0.00000478, -0.00000025, -0.00000324, -0.00000427, -0.00000430, -0.00000284, 0.00000155, -0.00000453, -0.00000420, -0.00000278, -0.00000596, -0.00000059, 0.00000342, 0.00000755, 0.00000113, -0.00000895, -0.00000072, 0.00000035, -0.00000547, -0.00000949, -0.00000325, 0.00000722, -0.00001723, -0.00000857, 0.00002237, -0.00000993, -0.00000145, 0.00002318, -0.00000740, 0.00000542, 0.00000656, 0.00000093, 0.00002548, -0.00002062, -0.00001121, 0.00003340, -0.00002669, -0.00000769, 0.00003083, -0.00002026, 0.00000069, 0.00000588, -0.00002209, 0.00001512, 0.00000445, -0.00000824, 0.00001226, -0.00000878, -0.00000120, 0.00000678, -0.00000585, 0.00000799, -0.00001418, -0.00002159, 0.00001888, 0.00001604, -0.00000064, -0.00000481, -0.00001297, -0.00000096, 0.00001241, 0.00000700, -0.00001140, -0.00001406, 0.00000996, 0.00000797, -0.00000798, -0.00000221, -0.00000058, 0.00000349, 0.00000222, -0.00001166, 0.00000458, 0.00001808, -0.00000415, -0.00001375, -0.00000685, 0.00000396, 0.00001381, -0.00000235, -0.00000699, 0.00001002, -0.00000193, -0.00000561, 0.00000792, -0.00000443, 0.00000341, 0.00001828, -0.00000552, -0.00001016, 0.00000232, -0.00000325, -0.00000154, 0.00000167, -0.00000141, -0.00000619, -0.00001003, -0.00000020, 0.00000343, -0.00000045, 0.00000323, -0.00000269, -0.00000942, -0.00000622, -0.00000535, -0.00001683, -0.00004172, -0.00003292, 0.00003113, 0.00007319, 0.00004839, 0.00004044, },
		{ 0.00001223, -0.00006373, -0.00003840, -0.00000559, -0.00000719, -0.00000333, -0.00000621, -0.00000543, -0.00000563, -0.00000849, 0.00000047, -0.00000450, -0.00000322, 0.00000901, -0.00000064, -0.00000339, 0.00000036, -0.00000492, -0.00000420, -0.00000448, -0.00000074, 0.00000068, -0.00000574, -0.00000362, -0.00000433, -0.00000422, 0.00000129, 0.00000459, 0.00000709, -0.00000386, -0.00000729, 0.00000176, -0.00000255, -0.00000689, -0.00000882, 0.00000163, 0.00000182, -0.00002181, 0.00000414, 0.00001667, -0.00001561, 0.00001153, 0.00001742, -0.00000840, 0.00000943, 0.00000314, 0.00001264, 0.00001284, -0.00002772, 0.00001346, 0.00001371, -0.00003446, 0.00001898, 0.00001265, -0.00002360, 0.00001433, -0.00000934, -0.00001297, 0.00002040, -0.00000700, 0.00000065, 0.00000730, -0.00001174, 0.00000722, -0.00000027, -0.00000187, 0.00000571, -0.00002597, -0.00000609, 0.00002630, 0.00000585, -0.00000165, -0.00000775, -0.00001275, 0.00000679, 0.00001311, 0.00000036, -0.00001491, -0.00000733, 0.00001293, 0.00000244, -0.00000826, -0.00000113, 0.00000138, 0.00000467, -0.00000390, -0.00000861, 0.00001243, 0.00001217, -0.00000956, -0.00001269, -0.00000435, 0.00000948, 0.00001008, -0.00000721, 0.00000035, 0.00000787, -0.00000746, 0.00000085, 0.00000510, -0.00000558, 0.00001304, 0.00001147, -0.00001285, -0.00000411, 0.00000141, -0.00000402, 0.00000113, -0.00000015, -0.00000249, -0.00000766, -0.00000812, 0.00000337, 0.00000121, 0.00000030, 0.00000360, -0.00000729, -0.00000862, -0.00000313, -0.00000909, -0.00002689, -0.00004208, -0.00000932, 0.00005325, 0.00006471, 0.00003508, 0.00005092, 0.00007921, },
		{ -0.00006238, -0.00002353, -0.00000333, -0.00000640, -0.00000435, -0.00000647, -0.00000436, -0.00000757, -0.00000553, 0.00000020, -0.00000612, 0.00000254, 0.00000694, -0.00000421, -0.00000050, -0.00000101, -0.00000601, -0.00000405, -0.00000388, 0.00000106, -0.00000117, -0.00000623, -0.00000375, -0.00000485, -0.00000177, 0.00000227, 0.00000589, 0.00000457, -0.00000753, -0.00000348, 0.00000154, -0.00000545, -0.00000711, -0.00000693, 0.00000459, -0.00000622, -0.00002011, 0.00001441, 0.00000613, -0.00001359, 0.00002183, 0.00000923, -0.00000642, 0.00001082, 0.00000573, 0.00001751, -0.00000428, -0.00001777, 0.00002425, -0.00001246, -0.00002021, 0.00003095, -0.00001052, -0.00001130, 0.00001564, -0.00002125, 0.00000279, 0.00001475, -0.00001153, 0.00000858, -0.00000215, -0.00000667, 0.00000967, -0.00000647, 0.00000515, -0.00000511, -0.00002838, 0.00001357, 0.00002234, -0.00000197, -0.00000060, -0.00001268, -0.00000798, 0.00001379, 0.00000925, -0.00000571, -0.00001480, -0.00000067, 0.00001297, -0.00000252, -0.00000785, 0.00000090, 0.00000372, 0.00000208, -0.00000767, -0.00000176, 0.00001532, 0.00000509, -0.00001242, -0.00001165, 0.00000038, 0.00001261, 0.00000349, -0.00000658, 0.00000587, 0.00000142, -0.00000736, 0.00000562, -0.00000027, -0.00000030, 0.00001800, 0.00000005, -0.00001337, 0.00000095, -0.00000150, -0.00000204, 0.00000168, -0.00000220, -0.00000278, -0.00000921, -0.00000437, 0.00000482, -0.00000138, 0.00000297, 0.00000101, -0.00001117, -0.00000436, -0.00000295, -0.00001689, -0.00003383, -0.00003273, 0.00001646, 0.00006314, 0.00004974, 0.00003084, 0.00006223, 0.00006495, -0.00001756, },
		{ -0.00001183, -0.00000292, -0.00000646, -0.00000535, -0.00000553, -0.00000479, -0.00000812, -0.00000262, -0.00000187, -0.00000448, 0.00000657, 0.00000223, -0.00000442, 0.00000165, -0.00000367, -0.00000614, -0.00000387, -0.00000249, 0.00000187, -0.00000339, -0.00000629, -0.00000421, -0.00000391, 0.00000027, 0.00000292, 0.00000675, 0.00000032, -0.00000834, 0.00000037, -0.00000130, -0.00000683, -0.00000664, -0.00000449, 0.00000478, -0.00001424, -0.00001292, 0.00001914, -0.00000509, -0.00000470, 0.00002718, 0.00000051, -0.00000272, 0.00001280, 0.00001053, 0.00001267, -0.00001446, -0.00000029, 0.00001609, -0.00002794, 0.00000454, 0.00002278, -0.00002375, 0.00000642, 0.00000427, -0.00002191, 0.00001583, 0.00000279, -0.00000748, 0.00000991, -0.00000944, 0.00000252, 0.00000482, -0.00000671, 0.00000860, -0.00001984, -0.00001753, 0.00002771, 0.00001040, -0.00000348, -0.00000170, -0.00001680, 0.00000142, 0.00001640, 0.00000280, -0.00000921, -0.00001261, 0.00000499, 0.00001101, -0.00000707, -0.00000616, 0.00000414, 0.00000384, -0.00000231, -0.00000696, 0.00000491, 0.00001386, -0.00000084, -0.00001434, -0.00000930, 0.00000639, 0.00001130, -0.00000195, -0.00000188, 0.00000625, -0.00000463, -0.00000285, 0.00000578, -0.00000332, 0.00000854, 0.00001497, -0.00001003, -0.00000844, 0.00000228, -0.00000338, 0.00000093, -0.00000026, -0.00000280, -0.00000350, -0.00000985, 0.00000042, 0.00000331, -0.00000205, 0.00000563, -0.00000486, -0.00001112, 0.00000087, -0.00000813, -0.00002516, -0.00003428, -0.00001616, 0.00003828, 0.00006115, 0.00003517, 0.00003519, 0.00006852, 0.00004225, -0.00003947, -0.00005319, },
		{ -0.00000396, -0.00000698, -0.00000547, -0.00000459, -0.00000605, -0.00000703, -0.00000116, -0.00000372, -0.00000052, 0.00000708, -0.00000230, -0.00000192, 0.00000161, -0.00000635, -0.00000544, -0.00000352, -0.00000069, 0.00000140, -0.00000545, -0.00000624, -0.00000424, -0.00000200, 0.00000130, 0.00000387, 0.00000622, -0.00000426, -0.00000602, 0.00000197, -0.00000510, -0.00000616, -0.00000638, -0.00000186, 0.00000204, -0.00001990, -0.00000251, 0.00001715, -0.00001286, 0.00000869, 0.00002619, -0.00000721, 0.00000327, 0.00001618, 0.00001194, 0.00000290, -0.00001279, 0.00001139, -0.00000314, -0.00002434, 0.00002321, 0.00000221, -0.00002038, 0.00001710, -0.00001154, -0.00001110, 0.00001915, -0.00000754, 0.00000089, 0.00000414, -0.00000960, 0.00000918, -0.00000326, -0.00000073, 0.00000330, -0.00002956, 0.00000291, 0.00002938, -0.00000168, -0.00000004, -0.00000727, -0.00001590, 0.00001211, 0.00001303, -0.00000282, -0.00001040, -0.00000968, 0.00000981, 0.00000704, -0.00001082, -0.00000212, 0.00000683, 0.00000077, -0.00000506, -0.00000292, 0.00000859, 0.00001057, -0.00000565, -0.00001576, -0.00000428, 0.00001080, 0.00000649, -0.00000317, 0.00000276, 0.00000202, -0.00000654, 0.00000202, 0.00000242, -0.00000089, 0.00001531, 0.00000542, -0.00001421, -0.00000222, 0.00000036, -0.00000250, 0.00000233, -0.00000304, -0.00000180, -0.00000545, -0.00000827, 0.00000420, -0.00000017, 0.00000043, 0.00000521, -0.00001144, -0.00000570, 0.00000247, -0.00001765, -0.00002946, -0.00002803, 0.00000349, 0.00005220, 0.00005067, 0.00002608, 0.00004455, 0.00006630, 0.00001595, -0.00005145, -0.00003985, -0.00000397, },
		{ -0.00000723, -0.00000474, -0.00000447, -0.00000709, -0.00000507, -0.00000131, -0.00000377, 0.00000356, 0.00000431, -0.00000437, 0.00000129, -0.00000079, -0.00000793, -0.00000431, -0.00000287, 0.00000093, -0.00000020, -0.00000708, -0.00000603, -0.00000329, -0.00000014, 0.00000159, 0.00000516, 0.00000369, -0.00000722, -0.00000202, 0.00000027, -0.00000752, -0.00000445, -0.00000675, 0.00000083, -0.00000338, -0.00002137, 0.00000801, 0.00000924, -0.00001378, 0.00002286, 0.00001840, -0.00001147, 0.00001219, 0.00001828, 0.00000780, -0.00000396, -0.00000366, 0.00000971, -0.00001926, -0.00000639, 0.00002523, -0.00001647, -0.00000512, 0.00001441, -0.00002137, 0.00000417, 0.00001240, -0.00001054, 0.00000696, -0.00000408, -0.00000293, 0.00000875, -0.00000818, 0.00000630, -0.00001002, -0.00002665, 0.00002316, 0.00001879, -0.00000691, 0.00000299, -0.00001530, -0.00000779, 0.00001892, 0.00000561, -0.00000533, -0.00001104, -0.00000587, 0.00001355, 0.00000072, -0.00001212, 0.00000399, 0.00000639, -0.00000366, -0.00000410, 0.00000122, 0.00000937, 0.00000734, -0.00001055, -0.00001523, 0.00000277, 0.00001107, 0.00000164, -0.00000041, 0.00000365, -0.00000311, -0.00000409, 0.00000377, -0.00000046, 0.00000596, 0.00001547, -0.00000550, -0.00001189, 0.00000141, -0.00000207, 0.00000035, 0.00000085, -0.00000440, -0.00000069, -0.00000797, -0.00000421, 0.00000495, -0.00000290, 0.00000474, -0.00000005, -0.00001426, 0.00000264, -0.00000311, -0.00002690, -0.00002749, -0.00001743, 0.00002251, 0.00005670, 0.00003656, 0.00002480, 0.00005397, 0.00005496, -0.00000885, -0.00005377, -0.00002569, 0.00000000, -0.00000591, },
		{ -0.00000383, -0.00000521, -0.00000720, -0.00000326, -0.00000218, -0.00000176, 0.00000568, 0.00000015, -0.00000334, 0.00000309, -0.00000451, -0.00000789, -0.00000315, -0.00000187, 0.00000184, -0.00000258, -0.00000819, -0.00000530, -0.00000152, 0.00000081, 0.00000198, 0.00000602, -0.00000036, -0.00000718, 0.00000112, -0.00000377, -0.00000709, -0.00000356, -0.00000699, 0.00000294, -0.00001068, -0.00001766, 0.00001516, -0.00000165, -0.00000607, 0.00003293, 0.00000540, -0.00000914, 0.00002220, 0.00001551, 0.00000133, -0.00000355, 0.00000347, -0.00000252, -0.00002185, 0.00001277, 0.00001184, -0.00002232, 0.00001053, 0.00000143, -0.00001961, 0.00001493, 0.00000129, -0.00000624, 0.00000666, -0.00000858, 0.00000543, 0.00000205, -0.00000625, 0.00000740, -0.00002416, -0.00001024, 0.00003300, 0.00000318, -0.00000379, 0.00000048, -0.00002033, 0.00000528, 0.00001826, -0.00000148, -0.00000525, -0.00001233, -0.00000019, 0.00001471, -0.00000717, -0.00000901, 0.00000964, 0.00000183, -0.00000600, -0.00000043, 0.00000298, 0.00000935, 0.00000389, -0.00001573, -0.00001095, 0.00000893, 0.00000731, 0.00000018, 0.00000299, 0.00000032, -0.00000529, -0.00000030, 0.00000220, 0.00000073, 0.00001264, 0.00000860, -0.00001242, -0.00000616, 0.00000138, -0.00000244, 0.00000266, -0.00000250, -0.00000333, -0.00000114, -0.00000923, 0.00000072, 0.00000239, -0.00000234, 0.00000733, -0.00000845, -0.00001007, 0.00000813, -0.00001519, -0.00003041, -0.00002062, -0.00000508, 0.00003834, 0.00005214, 0.00002393, 0.00003074, 0.00005867, 0.00003667, -0.00002804, -0.00004844, -0.00001326, 0.00000057, -0.00000802, -0.00000666, },
		{ -0.00000632, -0.00000631, -0.00000224, -0.00000259, 0.00000129, 0.00000499, -0.00000309, -0.00000026, 0.00000223, -0.00000796, -0.00000640, -0.00000226, -0.00000063, 0.00000171, -0.00000531, -0.00000856, -0.00000368, 0.00000024, 0.00000080, 0.00000302, 0.00000540, -0.00000433, -0.00000441, 0.00000128, -0.00000757, -0.00000433, -0.00000479, -0.00000577, 0.00000307, -0.00001799, -0.00000925, 0.00001610, -0.00001069, 0.00000903, 0.00003416, -0.00000819, 0.00000134, 0.00002865, 0.00000743, -0.00000176, 0.00000160, 0.00000176, -0.00001515, -0.00001074, 0.00002109, -0.00000641, -0.00001405, 0.00001668, -0.00001263, -0.00000848, 0.00001622, -0.00000700, 0.00000055, 0.00000111, -0.00000632, 0.00000941, -0.00000557, 0.00000060, -0.00000075, -0.00002964, 0.00001242, 0.00002809, -0.00000796, 0.00000295, -0.00000822, -0.00001730, 0.00001730, 0.00001078, -0.00000447, -0.00000537, -0.00001334, 0.00000747, 0.00001129, -0.00001382, -0.00000128, 0.00001123, -0.00000475, -0.00000393, 0.00000265, 0.00000262, 0.00001013, -0.00000145, -0.00001926, -0.00000301, 0.00001094, 0.00000267, 0.00000268, 0.00000318, -0.00000424, -0.00000352, 0.00000135, 0.00000039, 0.00000617, 0.00001430, -0.00000146, -0.00001307, -0.00000118, -0.00000054, -0.00000038, 0.00000233, -0.00000526, -0.00000084, -0.00000351, -0.00000778, 0.00000394, -0.00000135, 0.00000169, 0.00000480, -0.00001506, 0.00000009, 0.00000514, -0.00002819, -0.00002568, -0.00001256, 0.00000789, 0.00004893, 0.00004050, 0.00001718, 0.00004058, 0.00005561, 0.00001551, -0.00003947, -0.00003839, -0.00000411, -0.00000142, -0.00000946, -0.00000524, -0.00000346, },
		{ -0.00000486, -0.00000201, -0.00000180, 0.00000382, 0.00000215, -0.00000383, 0.00000283, -0.00000114, -0.00000975, -0.00000419, -0.00000169, 0.00000062, 0.00000040, -0.00000796, -0.00000783, -0.00000137, 0.00000111, 0.00000056, 0.00000447, 0.00000288, -0.00000637, -0.00000083, -0.00000181, -0.00000864, -0.00000156, -0.00000760, -0.00000245, -0.00000029, -0.00002255, 0.00000132, 0.00001012, -0.00001256, 0.00002652, 0.00002459, -0.00001570, 0.00001743, 0.00002689, -0.00000173, 0.00000151, 0.00000442, -0.00000789, -0.00001864, 0.00000521, 0.00001517, -0.00001778, 0.00000071, 0.00001096, -0.00001925, 0.00000456, 0.00000949, -0.00000867, 0.00000437, -0.00000477, 0.00000061, 0.00000650, -0.00000843, 0.00000605, -0.00001466, -0.00002105, 0.00002996, 0.00001258, -0.00000907, 0.00000597, -0.00001825, -0.00000559, 0.00002211, 0.00000124, -0.00000299, -0.00000821, -0.00001142, 0.00001489, 0.00000282, -0.00001545, 0.00000822, 0.00000673, -0.00000908, 0.00000154, 0.00000236, 0.00000275, 0.00001095, -0.00000933, -0.00001821, 0.00000562, 0.00000791, 0.00000112, 0.00000618, -0.00000094, -0.00000611, -0.00000025, 0.00000006, 0.00000192, 0.00001221, 0.00000934, -0.00000943, -0.00000911, 0.00000072, -0.00000162, 0.00000225, -0.00000064, -0.00000551, 0.00000094, -0.00000638, -0.00000397, 0.00000381, -0.00000305, 0.00000624, -0.00000306, -0.00001462, 0.00001027, -0.00000753, -0.00003462, -0.00001521, -0.00000618, 0.00002160, 0.00005212, 0.00002566, 0.00001870, 0.00004928, 0.00004461, -0.00000401, -0.00004320, -0.00002658, 0.00000115, -0.00000484, -0.00000956, -0.00000346, -0.00000402, -0.00000708, },
		{ -0.00000206, 0.00000008, 0.00000458, -0.00000117, -0.00000195, 0.00000398, -0.00000570, -0.00000927, -0.00000214, -0.00000126, 0.00000158, -0.00000209, -0.00000990, -0.00000574, 0.00000086, 0.00000080, 0.00000098, 0.00000530, -0.00000076, -0.00000562, 0.00000102, -0.00000625, -0.00000628, -0.00000128, -0.00000980, 0.00000177, -0.00000735, -0.00002184, 0.00000985, 0.00000003, -0.00000414, 0.00003890, 0.00000735, -0.00001173, 0.00003229, 0.00001625, -0.00000565, 0.00000817, -0.00000058, -0.00001789, -0.00001078, 0.00001514, 0.00000107, -0.00001685, 0.00001184, -0.00000129, -0.00001581, 0.00001231, 0.00000043, -0.00000507, 0.00000309, -0.00000633, 0.00000686, -0.00000065, -0.00000493, 0.00000423, -0.00002570, -0.00000115, 0.00003372, -0.00000366, -0.00000195, 0.00000089, -0.00002240, 0.00000980, 0.00001754, -0.00000460, -0.00000036, -0.00001345, -0.00000456, 0.00001817, -0.00000808, -0.00000987, 0.00001448, -0.00000218, -0.00000763, 0.00000644, -0.00000109, 0.00000571, 0.00000901, -0.00001769, -0.00001133, 0.00001060, 0.00000260, 0.00000421, 0.00000638, -0.00000668, -0.00000372, 0.00000093, -0.00000155, 0.00000757, 0.00001408, 0.00000044, -0.00001208, -0.00000427, 0.00000016, -0.00000058, 0.00000310, -0.00000427, -0.00000326, 0.00000058, -0.00000772, 0.00000009, 0.00000109, -0.00000099, 0.00000705, -0.00001214, -0.00000568, 0.00001275, -0.00002468, -0.00002996, -0.00000501, -0.00000107, 0.00003544, 0.00004637, 0.00001307, 0.00002755, 0.00005197, 0.00002845, -0.00001861, -0.00004095, -0.00001527, 0.00000262, -0.00000847, -0.00000806, -0.00000214, -0.00000537, -0.00000698, -0.00000344, },
		{ 0.00000229, 0.00000328, -0.00000319, 0.00000129, 0.00000222, -0.00000957, -0.00000687, -0.00000091, -0.00000068, 0.00000177, -0.00000543, -0.00001038, -0.00000259, 0.00000202, -0.00000008, 0.00000231, 0.00000458, -0.00000389, -0.00000294, -0.00000029, -0.00000916, -0.00000220, -0.00000428, -0.00000906, 0.00000418, -0.00001613, -0.00001541, 0.00001248, -0.00000833, 0.00001291, 0.00003965, -0.00000958, 0.00000390, 0.00003809, 0.00000185, -0.00000102, 0.00001113, -0.00001241, -0.00002005, 0.00000244, 0.00001369, -0.00001117, -0.00000675, 0.00001346, -0.00001185, -0.00000607, 0.00001220, -0.00000550, -0.00000057, -0.00000110, -0.00000265, 0.00000799, -0.00000658, 0.00000121, -0.00000553, -0.00002568, 0.00002034, 0.00002309, -0.00001153, 0.00000598, -0.00001029, -0.00001630, 0.00002106, 0.00000703, -0.00000400, -0.00000103, -0.00001758, 0.00000620, 0.00001422, -0.00001622, 0.00000122, 0.00001334, -0.00001047, -0.00000053, 0.00000667, -0.00000427, 0.00001105, 0.00000207, -0.00002232, -0.00000091, 0.00000940, -0.00000025, 0.00000945, 0.00000120, -0.00000955, 0.00000067, -0.00000131, 0.00000036, 0.00001414, 0.00000952, -0.00000748, -0.00000994, -0.00000140, -0.00000062, 0.00000163, 0.00000128, -0.00000624, -0.00000022, -0.00000158, -0.00000665, 0.00000213, -0.00000145, 0.00000340, 0.00000189, -0.00001619, 0.00000726, 0.00000330, -0.00003703, -0.00001611, -0.00000019, 0.00000597, 0.00004644, 0.00003261, 0.00000810, 0.00003928, 0.00004623, 0.00001166, -0.00002730, -0.00003495, -0.00000587, 0.00000083, -0.00001102, -0.00000533, -0.00000204, -0.00000686, -0.00000593, -0.00000252, -0.00000175, },
		{ 0.00000069, -0.00000290, 0.00000392, -0.00000200, -0.00001121, -0.00000368, -0.00000062, 0.00000023, 0.00000066, -0.00000888, -0.00000885, 0.00000068, 0.00000166, -0.00000050, 0.00000388, 0.00000222, -0.00000506, -0.00000041, -0.00000409, -0.00000859, 0.00000055, -0.00000889, -0.00000483, 0.00000222, -0.00002300, -0.00000578, 0.00000821, -0.00000875, 0.00003183, 0.00002735, -0.00001696, 0.00002477, 0.00003107, -0.00000799, 0.00000906, 0.00000466, -0.00002389, -0.00001188, 0.00001182, 0.00000372, -0.00001463, 0.00000454, 0.00000665, -0.00001527, 0.00000352, 0.00000669, -0.00000649, 0.00000122, -0.00000406, 0.00000313, 0.00000368, -0.00000736, 0.00000392, -0.00001759, -0.00001255, 0.00003252, 0.00000555, -0.00000836, 0.00000718, -0.00002040, -0.00000182, 0.00002254, -0.00000256, 0.00000081, -0.00000696, -0.00001617, 0.00001636, 0.00000356, -0.00001674, 0.00001237, 0.00000462, -0.00001266, 0.00000798, 0.00000125, -0.00000313, 0.00001509, -0.00000895, -0.00001981, 0.00000795, 0.00000363, 0.00000251, 0.00001179, -0.00000714, -0.00000693, 0.00000283, -0.00000448, 0.00000721, 0.00001658, 0.00000058, -0.00001073, -0.00000611, -0.00000074, -0.00000001, 0.00000295, -0.00000208, -0.00000558, 0.00000169, -0.00000396, -0.00000413, 0.00000163, -0.00000129, 0.00000619, -0.00000706, -0.00001117, 0.00001575, -0.00001516, -0.00003689, -0.00000069, -0.00000109, 0.00001768, 0.00004986, 0.00001548, 0.00001325, 0.00004750, 0.00003346, -0.00000178, -0.00003112, -0.00002691, 0.00000071, -0.00000327, -0.00001144, -0.00000234, -0.00000342, -0.00000762, -0.00000434, -0.00000217, -0.00000067, 0.00000376, },
		{ -0.00000054, 0.00000422, -0.00000706, -0.00001006, -0.00000102, -0.00000081, 0.00000122, -0.00000199, -0.00001126, -0.00000543, 0.00000281, 0.00000025, 0.00000021, 0.00000462, -0.00000079, -0.00000405, -0.00000002, -0.00000800, -0.00000493, -0.00000026, -0.00001199, 0.00000090, -0.00000461, -0.00002479, 0.00000254, 0.00000032, 0.00000161, 0.00004360, 0.00000757, -0.00000963, 0.00004067, 0.00001469, -0.00000728, 0.00001637, -0.00001051, -0.00002676, 0.00000163, 0.00001171, -0.00000691, -0.00000929, 0.00001023, -0.00000279, -0.00001154, 0.00000835, 0.00000039, -0.00000443, -0.00000005, -0.00000324, 0.00000658, -0.00000248, -0.00000358, -0.00000039, -0.00002357, 0.00000776, 0.00003000, -0.00000837, 0.00000090, -0.00000054, -0.00002207, 0.00001374, 0.00001454, -0.00000561, 0.00000429, -0.00001563, -0.00000745, 0.00002035, -0.00000905, -0.00000850, 0.00001723, -0.00000707, -0.00000668, 0.00001203, -0.00000642, 0.00000362, 0.00001340, -0.00001950, -0.00001043, 0.00001060, -0.00000165, 0.00000964, 0.00000742, -0.00001332, -0.00000049, 0.00000021, -0.00000394, 0.00001607, 0.00001183, -0.00000776, -0.00000922, -0.00000351, -0.00000062, 0.00000165, 0.00000214, -0.00000490, -0.00000312, 0.00000180, -0.00000520, -0.00000198, 0.00000024, 0.00000158, 0.00000400, -0.00001399, 0.00000112, 0.00001259, -0.00003327, -0.00002342, 0.00000792, -0.00000274, 0.00003285, 0.00004233, 0.00000235, 0.00002618, 0.00004710, 0.00001827, -0.00001030, -0.00003186, -0.00001783, 0.00000374, -0.00000816, -0.00000933, -0.00000046, -0.00000575, -0.00000707, -0.00000292, -0.00000205, 0.00000105, 0.00000374, -0.00000178, },
		{ 0.00000157, -0.00001090, -0.00000681, 0.00000027, -0.00000077, 0.00000159, -0.00000583, -0.00001144, -0.00000113, 0.00000302, -0.00000111, 0.00000186, 0.00000384, -0.00000304, -0.00000205, -0.00000228, -0.00000952, -0.00000063, -0.00000455, -0.00001118, 0.00000468, -0.00001404, -0.00002085, 0.00000580, -0.00000492, 0.00001970, 0.00004195, -0.00000941, 0.00000988, 0.00004342, -0.00000189, 0.00000295, 0.00001312, -0.00002678, -0.00001793, 0.00001178, 0.00000352, -0.00001145, -0.00000053, 0.00000866, -0.00000918, -0.00000463, 0.00000786, -0.00000349, -0.00000236, -0.00000203, 0.00000052, 0.00000548, -0.00000613, 0.00000038, -0.00000971, -0.00001814, 0.00002487, 0.00001596, -0.00001184, 0.00000772, -0.00001252, -0.00001296, 0.00002227, 0.00000297, -0.00000150, 0.00000166, -0.00002136, 0.00000574, 0.00001529, -0.00001688, 0.00000450, 0.00001257, -0.00001475, 0.00000433, 0.00000823, -0.00001060, 0.00001285, 0.00000434, -0.00002400, 0.00000120, 0.00000630, -0.00000132, 0.00001583, -0.00000287, -0.00001284, 0.00000471, -0.00000533, 0.00000316, 0.00002106, 0.00000145, -0.00001121, -0.00000588, -0.00000274, 0.00000049, 0.00000276, -0.00000030, -0.00000583, -0.00000043, 0.00000078, -0.00000533, -0.00000115, 0.00000023, 0.00000434, -0.00000294, -0.00001336, 0.00001326, -0.00000282, -0.00004065, -0.00000368, 0.00000627, 0.00000144, 0.00004559, 0.00002524, -0.00000021, 0.00004010, 0.00003719, 0.00000591, -0.00001509, -0.00003054, -0.00000859, 0.00000280, -0.00001189, -0.00000535, -0.00000075, -0.00000781, -0.00000532, -0.00000222, -0.00000159, 0.00000282, 0.00000222, -0.00000280, 0.00000250, },
		{ -0.00001202, -0.00000297, 0.00000020, -0.00000014, 0.00000052, -0.00000970, -0.00000897, 0.00000248, 0.00000148, -0.00000134, 0.00000348, 0.00000178, -0.00000364, -0.00000086, -0.00000590, -0.00000770, 0.00000136, -0.00000986, -0.00000640, 0.00000369, -0.00002229, -0.00001364, 0.00000368, -0.00000187, 0.00003748, 0.00002746, -0.00001409, 0.00003212, 0.00003202, -0.00000945, 0.00001466, -0.00000212, -0.00003438, -0.00000203, 0.00001247, -0.00000612, -0.00000851, 0.00000561, 0.00000291, -0.00001034, 0.00000114, 0.00000448, -0.00000461, -0.00000182, -0.00000221, 0.00000410, 0.00000116, -0.00000562, 0.00000005, -0.00001761, -0.00000289, 0.00003041, -0.00000037, -0.00000564, 0.00000616, -0.00002063, 0.00000246, 0.00002009, -0.00000463, 0.00000507, -0.00000735, -0.00001924, 0.00001705, 0.00000335, -0.00001541, 0.00001503, 0.00000077, -0.00001343, 0.00001361, -0.00000170, -0.00000719, 0.00001862, -0.00000901, -0.00001966, 0.00000880, -0.00000081, 0.00000577, 0.00001522, -0.00001364, -0.00000566, 0.00000413, -0.00000831, 0.00001477, 0.00001756, -0.00000911, -0.00000916, -0.00000389, -0.00000224, 0.00000239, 0.00000211, -0.00000274, -0.00000504, 0.00000156, -0.00000050, -0.00000529, -0.00000082, 0.00000223, 0.00000357, -0.00001026, -0.00000425, 0.00001659, -0.00002320, -0.00003247, 0.00001158, -0.00000158, 0.00001432, 0.00004853, 0.00000553, 0.00000947, 0.00004708, 0.00002179, -0.00000078, -0.00001864, -0.00002692, -0.00000062, -0.00000156, -0.00001271, -0.00000125, -0.00000321, -0.00000834, -0.00000322, -0.00000230, -0.00000037, 0.00000378, -0.00000004, -0.00000179, 0.00000437, -0.00000322, },
		{ -0.00000006, -0.00000047, 0.00000078, -0.00000233, -0.00001203, -0.00000448, 0.00000403, -0.00000068, -0.00000016, 0.00000407, -0.00000055, -0.00000288, -0.00000162, -0.00000865, -0.00000374, -0.00000040, -0.00001299, -0.00000012, -0.00000241, -0.00002637, -0.00000708, -0.00000012, 0.00001068, 0.00004622, 0.00000778, -0.00000337, 0.00004601, 0.00001340, -0.00000488, 0.00001783, -0.00002257, -0.00002809, 0.00001163, 0.00000430, -0.00001066, -0.00000187, 0.00000655, -0.00000256, -0.00000781, 0.00000380, 0.00000094, -0.00000448, -0.00000210, -0.00000015, 0.00000486, -0.00000299, -0.00000305, -0.00000531, -0.00001773, 0.00001447, 0.00002304, -0.00000992, 0.00000332, -0.00000312, -0.00001899, 0.00001583, 0.00001018, -0.00000421, 0.00000756, -0.00001802, -0.00000877, 0.00002055, -0.00000905, -0.00000540, 0.00001696, -0.00001129, -0.00000350, 0.00001505, -0.00001166, 0.00000305, 0.00001618, -0.00002026, -0.00000881, 0.00000864, -0.00000431, 0.00001539, 0.00000592, -0.00001825, 0.00000326, -0.00000230, -0.00000370, 0.00002437, 0.00000584, -0.00001420, -0.00000474, -0.00000408, -0.00000036, 0.00000346, 0.00000013, -0.00000404, -0.00000347, 0.00000294, -0.00000202, -0.00000560, 0.00000069, 0.00000425, -0.00000161, -0.00001225, 0.00000811, 0.00000717, -0.00003736, -0.00001288, 0.00001481, -0.00000679, 0.00003193, 0.00003801, -0.00000723, 0.00002651, 0.00004259, 0.00000759, -0.00000275, -0.00002245, -0.00002041, 0.00000413, -0.00000744, -0.00001003, 0.00000088, -0.00000658, -0.00000687, -0.00000188, -0.00000255, 0.00000148, 0.00000339, -0.00000180, 0.00000068, 0.00000371, -0.00000827, -0.00001017, },
		},
		{ { -0.00000158, -0.00000148, -0.00000011, 0.00000210, -0.00000351, -0.00000686, -0.00000402, -0.00000437, -0.00000760, -0.00000696, 0.00000046, -0.00001976, -0.00003415, -0.00000743, 0.00003050, 0.00004857, 0.00001385, 0.00002130, 0.00004612, 0.00000321, -0.00000130, -0.00001297, -0.00002622, 0.00000934, -0.00000488, -0.00000902, 0.00000708, -0.00000556, -0.00000207, 0.00000439, -0.00000304, -0.00000712, 0.00000211, -0.00000096, -0.00000282, -0.00000005, 0.00000097, 0.00001741, -0.00000736, -0.00001851, 0.00000834, 0.00000439, 0.00000271, -0.00000869, -0.00000579, 0.00000502, -0.00000593, 0.00000516, -0.00000394, -0.00000441, 0.00000692, -0.00000504, 0.00000819, 0.00000208, -0.00001108, -0.00000358, -0.00000966, 0.00000700, 0.00002352, -0.00000302, -0.00002280, -0.00000416, 0.00000936, 0.00000268, -0.00000073, 0.00000268, 0.00000936, -0.00000416, -0.00002280, -0.00000302, 0.00002352, 0.00000700, -0.00000966, -0.00000358, -0.00001108, 0.00000208, 0.00000819, -0.00000504, 0.00000692, -0.00000441, -0.00000394, 0.00000516, -0.00000593, 0.00000502, -0.00000579, -0.00000869, 0.00000271, 0.00000439, 0.00000834, -0.00001851, -0.00000736, 0.00001741, 0.00000097, -0.00000005, -0.00000282, -0.00000096, 0.00000211, -0.00000712, -0.00000304, 0.00000439, -0.00000207, -0.00000556, 0.00000708, -0.00000902, -0.00000488, 0.00000934, -0.00002622, -0.00001297, -0.00000130, 0.00000321, 0.00004612, 0.00002130, 0.00001385, 0.00004857, 0.00003050, -0.00000743, -0.00003415, -0.00001976, 0.00000046, -0.00000696, -0.00000760, -0.00000437, -0.00000402, -0.00000686, -0.00000351, 0.00000210, -0.00000011, -0.00000148, },
		{ -0.00000116, 0.00000073, 0.00000114, -0.00000526, -0.00000622, -0.00000396, -0.00000516, -0.00000773, -0.00000531, -0.00000163, -0.00002541, -0.00003395, -0.00000002, 0.00004181, 0.00004424, 0.00001008, 0.00003411, 0.00004016, -0.00000273, -0.00000272, -0.00002318, -0.00001620, 0.00001126, -0.00001239, -0.00000187, 0.00000479, -0.00000776, 0.00000242, 0.00000190, -0.00000534, -0.00000455, 0.00000205, -0.00000194, -0.00000204, -0.00000095, 0.00000624, 0.00001473, -0.00001473, -0.00001221, 0.00000912, 0.00000357, 0.00000154, -0.00001071, -0.00000172, 0.00000188, -0.00000390, 0.00000544, -0.00000793, 0.00000062, 0.00000406, -0.00000394, 0.00001105, -0.00000394, -0.00001017, -0.00000415, -0.00000850, 0.00001464, 0.00001943, -0.00001222, -0.00001910, 0.00000103, 0.00000739, 0.00000276, -0.00000145, 0.00000415, 0.00001035, -0.00001171, -0.00002229, 0.00000726, 0.00002317, -0.00000038, -0.00000880, -0.00000412, -0.00001004, 0.00000699, 0.00000353, -0.00000331, 0.00000724, -0.00000825, 0.00000140, 0.00000191, -0.00000529, 0.00000676, -0.00001015, -0.00000510, 0.00000243, 0.00000647, 0.00000513, -0.00002301, 0.00000246, 0.00001580, -0.00000302, 0.00000170, -0.00000450, 0.00000121, 0.00000058, -0.00000905, 0.00000126, 0.00000386, -0.00000591, -0.00000072, 0.00000488, -0.00001299, 0.00000410, 0.00000067, -0.00002985, -0.00000159, -0.00000352, 0.00001245, 0.00004641, 0.00000942, 0.00002227, 0.00004770, 0.00001942, -0.00001226, -0.00003382, -0.00001368, 0.00000116, -0.00000880, -0.00000683, -0.00000362, -0.00000455, -0.00000703, -0.00000137, 0.00000231, -0.00000104, -0.00000156, -0.00000159, },
		{ 0.00000121, -0.00000033, -0.00000640, -0.00000542, -0.00000436, -0.00000568, -0.00000742, -0.00000426, -0.00000463, -0.00003085, -0.00003271, 0.00001063, 0.00005088, 0.00003646, 0.00001220, 0.00004457, 0.00003069, -0.00000627, -0.00000836, -0.00002822, -0.00000398, 0.00000651, -0.00001514, 0.00000493, -0.00000057, -0.00000617, 0.00000510, -0.00000196, -0.00000530, -0.00000270, 0.00000146, -0.00000220, -0.00000218, -0.00000003, 0.00001080, 0.00000889, -0.00001856, -0.00000594, 0.00000853, 0.00000350, -0.00000084, -0.00001077, 0.00000088, -0.00000099, -0.00000053, 0.00000279, -0.00000910, 0.00000484, 0.00000007, -0.00000020, 0.00001089, -0.00000947, -0.00000801, -0.00000564, -0.00000471, 0.00002062, 0.00001150, -0.00001808, -0.00001308, 0.00000318, 0.00000622, 0.00000295, -0.00000297, 0.00000764, 0.00000847, -0.00001959, -0.00001702, 0.00001610, 0.00001902, -0.00000609, -0.00000691, -0.00000534, -0.00000704, 0.00000956, -0.00000117, 0.00000034, 0.00000469, -0.00000927, 0.00000595, -0.00000303, -0.00000174, 0.00000576, -0.00001317, -0.00000108, 0.00000128, 0.00000958, -0.00000106, -0.00002370, 0.00001245, 0.00001015, -0.00000411, 0.00000270, -0.00000615, 0.00000418, -0.00000305, -0.00000863, 0.00000582, 0.00000015, -0.00000677, 0.00000405, -0.00000089, -0.00001162, 0.00001021, -0.00001178, -0.00002534, 0.00000656, -0.00000674, 0.00002405, 0.00004008, 0.00000188, 0.00003248, 0.00004159, 0.00001064, -0.00001581, -0.00003271, -0.00000740, 0.00000015, -0.00001021, -0.00000554, -0.00000320, -0.00000538, -0.00000649, 0.00000079, 0.00000172, -0.00000180, -0.00000148, -0.00000157, -0.00000060, },
		{ -0.00000201, -0.00000688, -0.00000478, -0.00000504, -0.00000572, -0.00000707, -0.00000384, -0.00000828, -0.00003635, -0.00002933, 0.00002405, 0.00005567, 0.00002801, 0.00001967, 0.00005044, 0.00001993, -0.00000937, -0.00001641, -0.00002626, 0.00000570, -0.00000207, -0.00001223, 0.00000828, -0.00000618, -0.00000160, 0.00000457, -0.00000511, -0.00000381, -0.00000201, 0.00000116, -0.00000254, -0.00000246, 0.00000267, 0.00001312, 0.00000173, -0.00001900, -0.00000082, 0.00000755, 0.00000343, -0.00000364, -0.00000931, 0.00000165, -0.00000233, 0.00000236, -0.00000153, -0.00000719, 0.00000675, -0.00000321, 0.00000491, 0.00000742, -0.00001314, -0.00000567, -0.00000721, 0.00000151, 0.00002310, 0.00000156, -0.00001945, -0.00000717, 0.00000309, 0.00000664, 0.00000183, -0.00000370, 0.00001222, 0.00000272, -0.00002516, -0.00000804, 0.00002160, 0.00001255, -0.00000948, -0.00000494, -0.00000652, -0.00000273, 0.00000929, -0.00000431, 0.00000426, 0.00000011, -0.00000690, 0.00000774, -0.00000753, 0.00000361, 0.00000168, -0.00001355, 0.00000193, 0.00000046, 0.00001259, -0.00000965, -0.00001922, 0.00001972, 0.00000224, -0.00000192, 0.00000161, -0.00000634, 0.00000660, -0.00000813, -0.00000481, 0.00000823, -0.00000513, -0.00000384, 0.00000588, -0.00000734, -0.00000529, 0.00001010, -0.00002312, -0.00001419, 0.00000863, -0.00000747, 0.00003521, 0.00002821, 0.00000091, 0.00004096, 0.00003188, 0.00000502, -0.00001930, -0.00003015, -0.00000168, -0.00000242, -0.00001061, -0.00000407, -0.00000324, -0.00000626, -0.00000513, 0.00000255, 0.00000046, -0.00000215, -0.00000138, -0.00000144, 0.00000008, 0.00000115, },
		{ -0.00000683, -0.00000452, -0.00000568, -0.00000535, -0.00000703, -0.00000384, -0.00001269, -0.00004185, -0.00002251, 0.00003868, 0.00005534, 0.00002173, 0.00003052, 0.00005091, 0.00000924, -0.00001366, -0.00002360, -0.00001833, 0.00000959, -0.00001033, -0.00000539, 0.00000690, -0.00000921, 0.00000364, 0.00000109, -0.00000612, -0.00000230, -0.00000201, 0.00000128, -0.00000342, -0.00000191, 0.00000620, 0.00001263, -0.00000499, -0.00001696, 0.00000285, 0.00000668, 0.00000281, -0.00000605, -0.00000725, 0.00000113, -0.00000182, 0.00000337, -0.00000567, -0.00000316, 0.00000589, -0.00000424, 0.00000951, 0.00000146, -0.00001424, -0.00000413, -0.00000769, 0.00000901, 0.00002103, -0.00000783, -0.00001672, -0.00000332, 0.00000250, 0.00000810, -0.00000117, -0.00000203, 0.00001583, -0.00000629, -0.00002624, 0.00000245, 0.00002301, 0.00000545, -0.00001063, -0.00000347, -0.00000693, 0.00000176, 0.00000662, -0.00000495, 0.00000676, -0.00000474, -0.00000193, 0.00000587, -0.00000944, 0.00000870, -0.00000451, -0.00001096, 0.00000285, 0.00000130, 0.00001368, -0.00001876, -0.00000971, 0.00002185, -0.00000510, 0.00000240, -0.00000186, -0.00000396, 0.00000668, -0.00001275, 0.00000193, 0.00000660, -0.00000917, 0.00000159, 0.00000331, -0.00001095, 0.00000299, 0.00000308, -0.00002864, -0.00000062, 0.00000462, -0.00000317, 0.00004236, 0.00001386, 0.00000673, 0.00004476, 0.00002126, 0.00000205, -0.00002323, -0.00002560, 0.00000244, -0.00000587, -0.00000979, -0.00000284, -0.00000371, -0.00000692, -0.00000302, 0.00000347, -0.00000113, -0.00000200, -0.00000138, -0.00000109, 0.00000067, 0.00000054, -0.00000357, },
		{ -0.00000467, -0.00000596, -0.00000485, -0.00000743, -0.00000401, -0.00001826, -0.00004661, -0.00001139, 0.00005222, 0.00005051, 0.00001975, 0.00004211, 0.00004625, -0.00000104, -0.00001943, -0.00002683, -0.00000778, 0.00000711, -0.00001475, 0.00000209, 0.00000183, -0.00000820, 0.00000698, -0.00000362, -0.00000484, -0.00000183, -0.00000183, 0.00000134, -0.00000464, 0.00000003, 0.00000925, 0.00000979, -0.00001013, -0.00001353, 0.00000527, 0.00000595, 0.00000156, -0.00000761, -0.00000542, 0.00000034, -0.00000013, 0.00000204, -0.00000798, 0.00000126, 0.00000300, -0.00000243, 0.00001183, -0.00000541, -0.00001298, -0.00000384, -0.00000603, 0.00001591, 0.00001463, -0.00001426, -0.00001166, -0.00000221, 0.00000310, 0.00000906, -0.00000523, 0.00000271, 0.00001613, -0.00001642, -0.00002204, 0.00001188, 0.00002076, -0.00000095, -0.00001006, -0.00000264, -0.00000621, 0.00000530, 0.00000267, -0.00000316, 0.00000675, -0.00000793, 0.00000374, 0.00000089, -0.00000755, 0.00001129, -0.00001085, -0.00000623, 0.00000161, 0.00000449, 0.00001117, -0.00002557, 0.00000286, 0.00001789, -0.00000908, 0.00000654, -0.00000657, 0.00000091, 0.00000318, -0.00001438, 0.00000936, 0.00000074, -0.00000935, 0.00000658, -0.00000275, -0.00000940, 0.00000898, -0.00000832, -0.00002587, 0.00001030, -0.00000268, 0.00000640, 0.00004272, 0.00000120, 0.00001725, 0.00004255, 0.00001239, 0.00000024, -0.00002715, -0.00001913, 0.00000414, -0.00000917, -0.00000797, -0.00000214, -0.00000449, -0.00000706, -0.00000041, 0.00000325, -0.00000256, -0.00000148, -0.00000155, -0.00000044, 0.00000091, -0.00000048, -0.00000479, -0.00000651, },
		{ -0.00000574, -0.00000462, -0.00000810, -0.00000432, -0.00002541, -0.00004928, 0.00000398, 0.00006242, 0.00004298, 0.00002292, 0.00005190, 0.00003729, -0.00001117, -0.00002539, -0.00002460, 0.00000131, 0.00000031, -0.00001386, 0.00000694, -0.00000430, -0.00000371, 0.00000683, -0.00000741, -0.00000229, -0.00000260, -0.00000090, 0.00000066, -0.00000544, 0.00000318, 0.00001080, 0.00000561, -0.00001335, -0.00000955, 0.00000676, 0.00000517, -0.00000003, -0.00000837, -0.00000419, 0.00000018, 0.00000151, -0.00000101, -0.00000768, 0.00000439, -0.00000039, 0.00000165, 0.00001082, -0.00001143, -0.00001027, -0.00000449, -0.00000184, 0.00002014, 0.00000550, -0.00001632, -0.00000666, -0.00000306, 0.00000564, 0.00000791, -0.00000847, 0.00000966, 0.00001164, -0.00002467, -0.00001348, 0.00001826, 0.00001608, -0.00000589, -0.00000838, -0.00000231, -0.00000441, 0.00000709, -0.00000115, 0.00000011, 0.00000421, -0.00000814, 0.00000785, -0.00000534, -0.00000211, 0.00000993, -0.00001515, -0.00000112, -0.00000077, 0.00000954, 0.00000434, -0.00002736, 0.00001512, 0.00000900, -0.00000813, 0.00000802, -0.00001031, 0.00000668, -0.00000367, -0.00001112, 0.00001423, -0.00000734, -0.00000473, 0.00000793, -0.00000926, -0.00000298, 0.00000915, -0.00001939, -0.00001576, 0.00001480, -0.00000902, 0.00001894, 0.00003560, -0.00000604, 0.00002888, 0.00003503, 0.00000695, -0.00000207, -0.00002994, -0.00001162, 0.00000319, -0.00001136, -0.00000571, -0.00000202, -0.00000544, -0.00000642, 0.00000221, 0.00000185, -0.00000334, -0.00000089, -0.00000176, 0.00000044, 0.00000059, -0.00000160, -0.00000560, -0.00000616, -0.00000504, },
		{ -0.00000499, -0.00000867, -0.00000516, -0.00003422, -0.00004824, 0.00002251, 0.00006764, 0.00003524, 0.00003083, 0.00005785, 0.00002486, -0.00002116, -0.00002933, -0.00001770, 0.00000587, -0.00000725, -0.00000866, 0.00000741, -0.00000857, 0.00000212, 0.00000327, -0.00000871, -0.00000005, -0.00000393, 0.00000069, -0.00000105, -0.00000495, 0.00000665, 0.00001050, 0.00000113, -0.00001480, -0.00000550, 0.00000752, 0.00000420, -0.00000163, -0.00000870, -0.00000335, 0.00000092, 0.00000195, -0.00000441, -0.00000512, 0.00000525, -0.00000258, 0.00000650, 0.00000653, -0.00001522, -0.00000732, -0.00000514, 0.00000425, 0.00002021, -0.00000385, -0.00001415, -0.00000378, -0.00000407, 0.00000947, 0.00000389, -0.00000879, 0.00001651, 0.00000256, -0.00002839, -0.00000280, 0.00002079, 0.00001040, -0.00000916, -0.00000611, -0.00000223, -0.00000195, 0.00000690, -0.00000365, 0.00000343, 0.00000012, -0.00000519, 0.00000865, -0.00001028, 0.00000511, 0.00000460, -0.00001589, 0.00000245, -0.00000251, 0.00001472, -0.00000591, -0.00002257, 0.00002329, -0.00000177, -0.00000271, 0.00000544, -0.00001082, 0.00001074, -0.00001175, -0.00000290, 0.00001370, -0.00001404, 0.00000316, 0.00000399, -0.00001240, 0.00000530, 0.00000262, -0.00002532, -0.00000223, 0.00001201, -0.00001061, 0.00003047, 0.00002292, -0.00000591, 0.00003781, 0.00002458, 0.00000506, -0.00000595, -0.00003040, -0.00000447, 0.00000015, -0.00001191, -0.00000362, -0.00000239, -0.00000640, -0.00000479, 0.00000420, -0.00000043, -0.00000317, -0.00000062, -0.00000175, 0.00000132, -0.00000033, -0.00000251, -0.00000617, -0.00000593, -0.00000541, -0.00000512, },
		{ -0.00000877, -0.00000721, -0.00004415, -0.00004202, 0.00004219, 0.00006733, 0.00002981, 0.00004198, 0.00005848, 0.00000978, -0.00003021, -0.00002931, -0.00000880, 0.00000505, -0.00001223, -0.00000189, 0.00000391, -0.00000913, 0.00000667, -0.00000210, -0.00000722, 0.00000069, -0.00000475, 0.00000223, -0.00000343, -0.00000275, 0.00000937, 0.00000871, -0.00000294, -0.00001482, -0.00000161, 0.00000756, 0.00000313, -0.00000308, -0.00000895, -0.00000234, 0.00000216, 0.00000070, -0.00000671, -0.00000152, 0.00000395, -0.00000248, 0.00001025, 0.00000005, -0.00001627, -0.00000507, -0.00000465, 0.00001074, 0.00001585, -0.00001083, -0.00000936, -0.00000385, -0.00000337, 0.00001285, -0.00000226, -0.00000497, 0.00002041, -0.00000907, -0.00002626, 0.00000721, 0.00001986, 0.00000479, -0.00001085, -0.00000360, -0.00000218, 0.00000060, 0.00000509, -0.00000424, 0.00000550, -0.00000389, -0.00000016, 0.00000561, -0.00001178, 0.00001152, -0.00000322, -0.00001287, 0.00000325, -0.00000187, 0.00001761, -0.00001702, -0.00001172, 0.00002468, -0.00001054, 0.00000476, -0.00000073, -0.00000701, 0.00001056, -0.00001769, 0.00000798, 0.00000685, -0.00001574, 0.00001067, -0.00000422, -0.00000964, 0.00001099, -0.00000824, -0.00002335, 0.00000953, 0.00000424, -0.00000588, 0.00003707, 0.00000848, 0.00000104, 0.00004134, 0.00001420, 0.00000549, -0.00001135, -0.00002794, 0.00000089, -0.00000383, -0.00001088, -0.00000212, -0.00000309, -0.00000714, -0.00000218, 0.00000491, -0.00000289, -0.00000211, -0.00000092, -0.00000122, 0.00000177, -0.00000161, -0.00000304, -0.00000672, -0.00000574, -0.00000559, -0.00000434, -0.00000603, },
		{ -0.00001126, -0.00005397, -0.00002974, 0.00006052, 0.00006214, 0.00002871, 0.00005416, 0.00005285, -0.00000689, -0.00003674, -0.00002479, -0.00000119, 0.00000028, -0.00001291, 0.00000344, -0.00000148, -0.00000597, 0.00000800, -0.00000706, -0.00000393, -0.00000038, -0.00000432, 0.00000282, -0.00000554, 0.00000082, 0.00001057, 0.00000616, -0.00000638, -0.00001364, 0.00000190, 0.00000688, 0.00000219, -0.00000455, -0.00000913, -0.00000070, 0.00000300, -0.00000186, -0.00000703, 0.00000163, 0.00000149, 0.00000002, 0.00001142, -0.00000693, -0.00001490, -0.00000384, -0.00000228, 0.00001564, 0.00000818, -0.00001369, -0.00000440, -0.00000616, 0.00000003, 0.00001375, -0.00000847, 0.00000255, 0.00001914, -0.00002000, -0.00001885, 0.00001434, 0.00001671, -0.00000021, -0.00001107, -0.00000115, -0.00000209, 0.00000269, 0.00000239, -0.00000304, 0.00000558, -0.00000625, 0.00000501, -0.00000028, -0.00000888, 0.00001463, -0.00001104, -0.00000731, 0.00000133, 0.00000190, 0.00001613, -0.00002555, 0.00000246, 0.00001878, -0.00001409, 0.00001083, -0.00000818, 0.00000026, 0.00000511, -0.00001822, 0.00001745, -0.00000438, -0.00001067, 0.00001365, -0.00001305, -0.00000128, 0.00001037, -0.00001870, -0.00001419, 0.00001540, -0.00000428, 0.00000390, 0.00003653, -0.00000342, 0.00001216, 0.00003875, 0.00000637, 0.00000636, -0.00001728, -0.00002289, 0.00000360, -0.00000749, -0.00000882, -0.00000132, -0.00000405, -0.00000731, 0.00000105, 0.00000391, -0.00000468, -0.00000065, -0.00000173, -0.00000002, 0.00000139, -0.00000278, -0.00000329, -0.00000742, -0.00000539, -0.00000557, -0.00000369, -0.00000761, -0.00000825, },
		{ -0.00006194, -0.00001144, 0.00007501, 0.00005381, 0.00003311, 0.00006461, 0.00004065, -0.00002347, -0.00003901, -0.00001697, 0.00000270, -0.00000565, -0.00000968, 0.00000538, -0.00000608, -0.00000074, 0.00000573, -0.00000975, -0.00000051, -0.00000259, -0.00000259, 0.00000197, -0.00000637, 0.00000483, 0.00001011, 0.00000348, -0.00000920, -0.00001136, 0.00000470, 0.00000570, 0.00000152, -0.00000631, -0.00000885, 0.00000156, 0.00000261, -0.00000470, -0.00000541, 0.00000323, -0.00000066, 0.00000399, 0.00000938, -0.00001274, -0.00001210, -0.00000325, 0.00000194, 0.00001726, -0.00000057, -0.00001216, -0.00000155, -0.00000885, 0.00000564, 0.00001090, -0.00001217, 0.00001163, 0.00001212, -0.00002699, -0.00000838, 0.00001758, 0.00001270, -0.00000447, -0.00000987, 0.00000094, -0.00000191, 0.00000397, -0.00000033, -0.00000072, 0.00000374, -0.00000600, 0.00000826, -0.00000688, -0.00000226, 0.00001304, -0.00001636, -0.00000134, -0.00000197, 0.00000814, 0.00000951, -0.00002836, 0.00001582, 0.00000763, -0.00001122, 0.00001251, -0.00001363, 0.00000824, -0.00000434, -0.00001181, 0.00002122, -0.00001569, -0.00000011, 0.00000953, -0.00001782, 0.00000925, 0.00000249, -0.00002380, -0.00000162, 0.00001409, -0.00000938, 0.00001528, 0.00002922, -0.00000963, 0.00002374, 0.00003126, 0.00000227, 0.00000605, -0.00002231, -0.00001637, 0.00000365, -0.00000992, -0.00000644, -0.00000111, -0.00000524, -0.00000646, 0.00000415, 0.00000130, -0.00000507, 0.00000048, -0.00000259, 0.00000161, -0.00000004, -0.00000332, -0.00000368, -0.00000812, -0.00000475, -0.00000550, -0.00000337, -0.00000945, -0.00000724, -0.00001794, },
		{ 0.00001182, 0.00008363, 0.00004494, 0.00004299, 0.00007034, 0.00002250, -0.00003762, -0.00003613, -0.00000838, 0.00000222, -0.00000992, -0.00000454, 0.00000375, -0.00000790, 0.00000418, 0.00000099, -0.00000946, 0.00000155, -0.00000475, -0.00000024, -0.00000015, -0.00000531, 0.00000813, 0.00000846, 0.00000098, -0.00001136, -0.00000810, 0.00000639, 0.00000442, 0.00000095, -0.00000848, -0.00000758, 0.00000385, 0.00000067, -0.00000662, -0.00000269, 0.00000303, -0.00000130, 0.00000788, 0.00000455, -0.00001628, -0.00000891, -0.00000249, 0.00000703, 0.00001485, -0.00000777, -0.00000755, -0.00000196, -0.00000973, 0.00001166, 0.00000457, -0.00001137, 0.00001910, 0.00000086, -0.00002798, 0.00000217, 0.00001734, 0.00000873, -0.00000799, -0.00000732, 0.00000229, -0.00000161, 0.00000431, -0.00000237, 0.00000177, 0.00000082, -0.00000323, 0.00000821, -0.00001168, 0.00000598, 0.00000698, -0.00001754, 0.00000284, -0.00000456, 0.00001481, -0.00000114, -0.00002393, 0.00002419, -0.00000478, -0.00000326, 0.00000861, -0.00001435, 0.00001334, -0.00001432, 0.00000023, 0.00001679, -0.00002204, 0.00001172, -0.00000103, -0.00001505, 0.00001670, -0.00001002, -0.00002077, 0.00000917, 0.00000753, -0.00000881, 0.00002431, 0.00001777, -0.00000909, 0.00003246, 0.00002130, 0.00000165, 0.00000376, -0.00002530, -0.00000981, 0.00000176, -0.00001081, -0.00000428, -0.00000136, -0.00000652, -0.00000431, 0.00000616, -0.00000219, -0.00000385, 0.00000059, -0.00000282, 0.00000304, -0.00000225, -0.00000293, -0.00000468, -0.00000842, -0.00000395, -0.00000552, -0.00000344, -0.00001125, -0.00000617, -0.00002754, -0.00006598, },
		{ 0.00008525, 0.00003857, 0.00005689, 0.00006857, 0.00000025, -0.00004679, -0.00002868, -0.00000172, -0.00000140, -0.00001099, 0.00000005, -0.00000002, -0.00000644, 0.00000678, -0.00000423, -0.00000679, 0.00000157, -0.00000580, 0.00000172, -0.00000269, -0.00000251, 0.00000992, 0.00000638, -0.00000144, -0.00001265, -0.00000422, 0.00000680, 0.00000355, 0.00000002, -0.00001074, -0.00000496, 0.00000523, -0.00000236, -0.00000685, -0.00000007, 0.00000161, 0.00000005, 0.00001017, -0.00000186, -0.00001722, -0.00000608, -0.00000079, 0.00001142, 0.00000898, -0.00001141, -0.00000222, -0.00000527, -0.00000736, 0.00001569, -0.00000338, -0.00000569, 0.00002200, -0.00001150, -0.00002286, 0.00001013, 0.00001498, 0.00000507, -0.00001057, -0.00000368, 0.00000260, -0.00000109, 0.00000377, -0.00000331, 0.00000356, -0.00000198, 0.00000100, 0.00000470, -0.00001272, 0.00001304, -0.00000178, -0.00001445, 0.00000387, -0.00000453, 0.00001923, -0.00001295, -0.00001312, 0.00002497, -0.00001404, 0.00000635, 0.00000041, -0.00000953, 0.00001278, -0.00002044, 0.00001382, 0.00000490, -0.00001992, 0.00001930, -0.00001399, -0.00000454, 0.00001646, -0.00002178, -0.00001048, 0.00001408, -0.00000025, -0.00000312, 0.00002820, 0.00000586, -0.00000295, 0.00003639, 0.00001142, 0.00000328, -0.00000033, -0.00002580, -0.00000439, -0.00000101, -0.00001041, -0.00000258, -0.00000205, -0.00000752, -0.00000094, 0.00000620, -0.00000539, -0.00000149, -0.00000051, -0.00000188, 0.00000345, -0.00000450, -0.00000183, -0.00000656, -0.00000780, -0.00000340, -0.00000557, -0.00000396, -0.00001273, -0.00000570, -0.00003980, -0.00006403, 0.00003794, },
		{ 0.00003760, 0.00007181, 0.00005741, -0.00002285, -0.00004916, -0.00001876, 0.00000126, -0.00000590, -0.00000899, 0.00000232, -0.00000380, -0.00000270, 0.00000624, -0.00000800, -0.00000320, -0.00000014, -0.00000531, 0.00000248, -0.00000458, 0.00000120, 0.00000999, 0.00000447, -0.00000397, -0.00001271, -0.00000036, 0.00000611, 0.00000335, -0.00000178, -0.00001239, -0.00000124, 0.00000488, -0.00000549, -0.00000537, 0.00000146, 0.00000010, 0.00000295, 0.00000987, -0.00000829, -0.00001599, -0.00000378, 0.00000210, 0.00001349, 0.00000143, -0.00001076, 0.00000139, -0.00000971, -0.00000174, 0.00001585, -0.00001016, 0.00000334, 0.00001882, -0.00002133, -0.00001351, 0.00001411, 0.00001206, 0.00000145, -0.00001176, 0.00000044, 0.00000181, -0.00000029, 0.00000252, -0.00000310, 0.00000415, -0.00000358, 0.00000505, -0.00000116, -0.00000933, 0.00001644, -0.00001052, -0.00000848, 0.00000181, -0.00000099, 0.00001918, -0.00002228, 0.00000094, 0.00001814, -0.00001695, 0.00001352, -0.00000885, -0.00000080, 0.00000603, -0.00001932, 0.00002358, -0.00001033, -0.00000929, 0.00001834, -0.00002337, 0.00000999, 0.00000702, -0.00002710, 0.00000280, 0.00001194, -0.00000521, 0.00000492, 0.00002642, -0.00000331, 0.00000629, 0.00003527, 0.00000349, 0.00000571, -0.00000543, -0.00002412, -0.00000066, -0.00000376, -0.00000919, -0.00000134, -0.00000324, -0.00000766, 0.00000299, 0.00000396, -0.00000707, 0.00000095, -0.00000238, 0.00000028, 0.00000220, -0.00000581, -0.00000077, -0.00000904, -0.00000598, -0.00000365, -0.00000535, -0.00000511, -0.00001358, -0.00000668, -0.00005385, -0.00005435, 0.00006394, 0.00008011, },
		{ 0.00008342, 0.00003675, -0.00004270, -0.00004451, -0.00000925, 0.00000046, -0.00000909, -0.00000534, 0.00000188, -0.00000583, 0.00000146, 0.00000311, -0.00000929, -0.00000022, -0.00000256, -0.00000360, 0.00000186, -0.00000501, 0.00000471, 0.00000882, 0.00000289, -0.00000663, -0.00001126, 0.00000272, 0.00000491, 0.00000364, -0.00000472, -0.00001255, 0.00000276, 0.00000258, -0.00000755, -0.00000292, 0.00000154, -0.00000043, 0.00000627, 0.00000686, -0.00001341, -0.00001336, -0.00000172, 0.00000572, 0.00001230, -0.00000544, -0.00000663, 0.00000168, -0.00001294, 0.00000562, 0.00001162, -0.00001317, 0.00001272, 0.00001011, -0.00002579, -0.00000296, 0.00001431, 0.00000960, -0.00000241, -0.00001103, 0.00000415, 0.00000021, 0.00000077, 0.00000083, -0.00000195, 0.00000354, -0.00000337, 0.00000732, -0.00000734, -0.00000238, 0.00001490, -0.00001660, -0.00000190, -0.00000196, 0.00000539, 0.00001386, -0.00002605, 0.00001392, 0.00000633, -0.00001280, 0.00001518, -0.00001532, 0.00000825, -0.00000462, -0.00001029, 0.00002502, -0.00002280, 0.00000602, 0.00000800, -0.00002395, 0.00002217, -0.00000886, -0.00002266, 0.00001329, 0.00000501, -0.00000523, 0.00001187, 0.00002054, -0.00000806, 0.00001590, 0.00003008, -0.00000163, 0.00000782, -0.00001071, -0.00002093, 0.00000146, -0.00000599, -0.00000756, -0.00000057, -0.00000489, -0.00000634, 0.00000632, -0.00000004, -0.00000651, 0.00000229, -0.00000402, 0.00000301, -0.00000071, -0.00000541, -0.00000078, -0.00001123, -0.00000327, -0.00000499, -0.00000450, -0.00000719, -0.00001345, -0.00001016, -0.00006802, -0.00003597, 0.00008629, 0.00007008, 0.00004401, },
		{ 0.00000888, -0.00005540, -0.00003454, -0.00000264, -0.00000274, -0.00000985, -0.00000192, -0.00000038, -0.00000544, 0.00000430, -0.00000114, -0.00000817, 0.00000114, -0.00000460, -0.00000150, 0.00000031, -0.00000380, 0.00000710, 0.00000716, 0.00000139, -0.00000910, -0.00000844, 0.00000444, 0.00000395, 0.00000379, -0.00000850, -0.00001062, 0.00000582, -0.00000117, -0.00000779, -0.00000056, 0.00000058, 0.00000058, 0.00000869, 0.00000185, -0.00001645, -0.00001012, 0.00000057, 0.00000895, 0.00000799, -0.00000952, -0.00000108, -0.00000148, -0.00001302, 0.00001229, 0.00000423, -0.00001107, 0.00001914, -0.00000158, -0.00002390, 0.00000573, 0.00001214, 0.00000776, -0.00000643, -0.00000813, 0.00000649, -0.00000158, 0.00000186, -0.00000097, -0.00000027, 0.00000213, -0.00000145, 0.00000681, -0.00001160, 0.00000598, 0.00000874, -0.00001836, 0.00000297, -0.00000532, 0.00001264, 0.00000437, -0.00002291, 0.00002180, -0.00000613, -0.00000362, 0.00001056, -0.00001624, 0.00001367, -0.00001484, 0.00000392, 0.00001672, -0.00002702, 0.00001958, -0.00000821, -0.00001396, 0.00002568, -0.00002485, -0.00000943, 0.00001647, -0.00000224, -0.00000104, 0.00001537, 0.00001316, -0.00000845, 0.00002390, 0.00002232, -0.00000383, 0.00000895, -0.00001559, -0.00001688, 0.00000230, -0.00000760, -0.00000570, -0.00000041, -0.00000664, -0.00000335, 0.00000782, -0.00000452, -0.00000392, 0.00000179, -0.00000432, 0.00000510, -0.00000449, -0.00000327, -0.00000266, -0.00001195, -0.00000063, -0.00000716, -0.00000290, -0.00001044, -0.00001199, -0.00001734, -0.00007978, -0.00000917, 0.00010157, 0.00005867, 0.00005786, 0.00008688, },
		{ -0.00005855, -0.00002249, -0.00000007, -0.00000632, -0.00000835, -0.00000004, -0.00000294, -0.00000315, 0.00000489, -0.00000486, -0.00000562, 0.00000077, -0.00000552, 0.00000017, -0.00000131, -0.00000146, 0.00000804, 0.00000567, -0.00000042, -0.00001077, -0.00000484, 0.00000469, 0.00000383, 0.00000298, -0.00001223, -0.00000668, 0.00000684, -0.00000519, -0.00000620, 0.00000077, -0.00000049, 0.00000293, 0.00000918, -0.00000399, -0.00001728, -0.00000673, 0.00000328, 0.00001054, 0.00000182, -0.00000972, 0.00000346, -0.00000674, -0.00000929, 0.00001590, -0.00000382, -0.00000445, 0.00002022, -0.00001264, -0.00001681, 0.00001062, 0.00000946, 0.00000596, -0.00000994, -0.00000341, 0.00000680, -0.00000284, 0.00000263, -0.00000248, 0.00000149, 0.00000056, 0.00000140, 0.00000354, -0.00001233, 0.00001309, -0.00000019, -0.00001563, 0.00000466, -0.00000631, 0.00001823, -0.00000663, -0.00001381, 0.00002236, -0.00001488, 0.00000672, 0.00000151, -0.00001120, 0.00001289, -0.00001997, 0.00001801, 0.00000127, -0.00002058, 0.00002507, -0.00002355, 0.00000362, 0.00001736, -0.00003368, 0.00000750, 0.00001127, -0.00000553, 0.00000443, 0.00001515, 0.00000651, -0.00000559, 0.00002929, 0.00001332, -0.00000336, 0.00000877, -0.00001978, -0.00001232, 0.00000217, -0.00000862, -0.00000371, -0.00000111, -0.00000776, 0.00000083, 0.00000668, -0.00000787, -0.00000036, -0.00000041, -0.00000266, 0.00000530, -0.00000772, -0.00000024, -0.00000641, -0.00001029, 0.00000059, -0.00000917, -0.00000107, -0.00001470, -0.00000920, -0.00002920, -0.00008582, 0.00002398, 0.00010740, 0.00005043, 0.00007659, 0.00007821, -0.00002152, },
		{ -0.00001191, -0.00000100, -0.00000862, -0.00000576, -0.00000001, -0.00000449, -0.00000019, 0.00000334, -0.00000698, -0.00000290, -0.00000077, -0.00000522, 0.00000097, -0.00000224, 0.00000115, 0.00000780, 0.00000457, -0.00000276, -0.00001105, -0.00000134, 0.00000394, 0.00000462, 0.00000058, -0.00001471, -0.00000158, 0.00000531, -0.00000817, -0.00000352, 0.00000073, -0.00000072, 0.00000578, 0.00000740, -0.00000942, -0.00001624, -0.00000333, 0.00000616, 0.00000963, -0.00000426, -0.00000638, 0.00000508, -0.00001186, -0.00000263, 0.00001517, -0.00000967, 0.00000439, 0.00001545, -0.00001979, -0.00000724, 0.00001140, 0.00000767, 0.00000340, -0.00001193, 0.00000211, 0.00000504, -0.00000302, 0.00000268, -0.00000332, 0.00000290, -0.00000056, 0.00000408, -0.00000152, -0.00000904, 0.00001664, -0.00000926, -0.00000969, 0.00000302, -0.00000390, 0.00002000, -0.00001586, -0.00000173, 0.00001594, -0.00001714, 0.00001405, -0.00000833, -0.00000235, 0.00000603, -0.00001718, 0.00002619, -0.00001549, -0.00000558, 0.00001921, -0.00003082, 0.00002180, -0.00000088, -0.00003046, 0.00002092, 0.00000075, -0.00000294, 0.00000781, 0.00001281, 0.00000155, -0.00000073, 0.00003172, 0.00000417, -0.00000050, 0.00000690, -0.00002302, -0.00000739, 0.00000113, -0.00000896, -0.00000180, -0.00000273, -0.00000744, 0.00000504, 0.00000306, -0.00000882, 0.00000258, -0.00000329, 0.00000062, 0.00000297, -0.00000893, 0.00000211, -0.00001099, -0.00000623, -0.00000065, -0.00000970, -0.00000018, -0.00001904, -0.00000586, -0.00004599, -0.00008263, 0.00005958, 0.00010349, 0.00004981, 0.00009496, 0.00005586, -0.00004830, -0.00005233, },
		{ -0.00000365, -0.00000903, -0.00000337, -0.00000122, -0.00000449, 0.00000214, 0.00000057, -0.00000723, -0.00000102, -0.00000263, -0.00000411, 0.00000097, -0.00000216, 0.00000321, 0.00000702, 0.00000362, -0.00000543, -0.00000972, 0.00000117, 0.00000302, 0.00000578, -0.00000343, -0.00001490, 0.00000326, 0.00000164, -0.00000915, -0.00000086, -0.00000032, 0.00000042, 0.00000800, 0.00000366, -0.00001355, -0.00001384, 0.00000012, 0.00000848, 0.00000615, -0.00000830, -0.00000106, 0.00000313, -0.00001465, 0.00000490, 0.00001045, -0.00001131, 0.00001229, 0.00000644, -0.00002117, 0.00000161, 0.00000929, 0.00000717, -0.00000032, -0.00001144, 0.00000699, 0.00000186, -0.00000202, 0.00000179, -0.00000322, 0.00000367, -0.00000081, 0.00000549, -0.00000671, -0.00000259, 0.00001545, -0.00001586, -0.00000276, -0.00000078, 0.00000157, 0.00001708, -0.00002059, 0.00000944, 0.00000528, -0.00001269, 0.00001561, -0.00001508, 0.00000651, -0.00000402, -0.00000688, 0.00002473, -0.00002676, 0.00001197, 0.00000357, -0.00002569, 0.00003237, -0.00002232, -0.00001521, 0.00002482, -0.00000934, 0.00000405, 0.00000721, 0.00001061, -0.00000205, 0.00000544, 0.00003099, -0.00000409, 0.00000433, 0.00000289, -0.00002476, -0.00000238, -0.00000084, -0.00000839, -0.00000042, -0.00000489, -0.00000519, 0.00000782, -0.00000190, -0.00000707, 0.00000360, -0.00000533, 0.00000428, -0.00000144, -0.00000740, 0.00000220, -0.00001457, -0.00000094, -0.00000447, -0.00000777, -0.00000167, -0.00002189, -0.00000382, -0.00006645, -0.00006740, 0.00009234, 0.00009234, 0.00005940, 0.00010599, 0.00002199, -0.00006562, -0.00003963, -0.00000536, },
		{ -0.00000802, -0.00000198, -0.00000268, -0.00000320, 0.00000310, -0.00000227, -0.00000609, -0.00000038, -0.00000409, -0.00000274, 0.00000061, -0.00000126, 0.00000434, 0.00000629, 0.00000237, -0.00000781, -0.00000709, 0.00000223, 0.00000278, 0.00000637, -0.00000834, -0.00001243, 0.00000638, -0.00000302, -0.00000794, 0.00000075, -0.00000150, 0.00000279, 0.00000868, -0.00000117, -0.00001594, -0.00001054, 0.00000359, 0.00000934, 0.00000101, -0.00000908, 0.00000406, -0.00000164, -0.00001386, 0.00001086, 0.00000353, -0.00000832, 0.00001646, -0.00000378, -0.00001703, 0.00000727, 0.00000631, 0.00000732, -0.00000482, -0.00000815, 0.00000987, -0.00000161, -0.00000024, 0.00000009, -0.00000210, 0.00000371, -0.00000013, 0.00000497, -0.00001027, 0.00000513, 0.00000981, -0.00001828, 0.00000284, -0.00000483, 0.00000858, 0.00001016, -0.00001964, 0.00001631, -0.00000563, -0.00000379, 0.00001105, -0.00001632, 0.00001167, -0.00001278, 0.00000719, 0.00001368, -0.00002771, 0.00002454, -0.00001569, -0.00000904, 0.00002957, -0.00003795, 0.00000668, 0.00001726, -0.00001331, 0.00001129, 0.00000333, 0.00000991, -0.00000505, 0.00001270, 0.00002673, -0.00001015, 0.00001022, -0.00000350, -0.00002420, 0.00000208, -0.00000350, -0.00000677, -0.00000015, -0.00000674, -0.00000130, 0.00000802, -0.00000645, -0.00000346, 0.00000229, -0.00000530, 0.00000662, -0.00000647, -0.00000356, -0.00000078, -0.00001531, 0.00000352, -0.00000966, -0.00000341, -0.00000651, -0.00002150, -0.00000586, -0.00008739, -0.00003928, 0.00011687, 0.00007911, 0.00007835, 0.00010295, -0.00001745, -0.00006997, -0.00002511, -0.00000342, -0.00000605, },
		{ -0.00000159, -0.00000363, -0.00000137, 0.00000264, -0.00000432, -0.00000439, -0.00000083, -0.00000480, -0.00000153, 0.00000036, -0.00000006, 0.00000466, 0.00000587, 0.00000048, -0.00000922, -0.00000395, 0.00000200, 0.00000363, 0.00000545, -0.00001287, -0.00000778, 0.00000681, -0.00000717, -0.00000518, 0.00000084, -0.00000184, 0.00000566, 0.00000740, -0.00000611, -0.00001654, -0.00000664, 0.00000672, 0.00000811, -0.00000425, -0.00000655, 0.00000701, -0.00000739, -0.00000961, 0.00001349, -0.00000303, -0.00000205, 0.00001554, -0.00001188, -0.00000941, 0.00000881, 0.00000429, 0.00000694, -0.00000896, -0.00000262, 0.00000996, -0.00000417, 0.00000154, -0.00000193, -0.00000017, 0.00000312, 0.00000118, 0.00000249, -0.00001091, 0.00001179, 0.00000137, -0.00001620, 0.00000546, -0.00000717, 0.00001499, 0.00000124, -0.00001374, 0.00001716, -0.00001305, 0.00000581, 0.00000245, -0.00001200, 0.00001114, -0.00001614, 0.00001942, -0.00000288, -0.00001769, 0.00002633, -0.00003026, 0.00001295, 0.00001303, -0.00004033, 0.00002640, 0.00000154, -0.00000843, 0.00001419, -0.00000096, 0.00001030, -0.00000761, 0.00002072, 0.00001867, -0.00001243, 0.00001552, -0.00001180, -0.00002070, 0.00000501, -0.00000619, -0.00000437, -0.00000131, -0.00000720, 0.00000310, 0.00000544, -0.00000898, 0.00000041, -0.00000062, -0.00000291, 0.00000631, -0.00001015, 0.00000093, -0.00000623, -0.00001227, 0.00000497, -0.00001400, 0.00000190, -0.00001438, -0.00001695, -0.00001484, -0.00010379, -0.00000033, 0.00012944, 0.00007034, 0.00010166, 0.00008187, -0.00005396, -0.00006159, -0.00001347, -0.00000469, -0.00000703, -0.00000661, },
		{ -0.00000384, 0.00000027, 0.00000123, -0.00000524, -0.00000286, -0.00000191, -0.00000479, -0.00000059, 0.00000047, 0.00000087, 0.00000462, 0.00000561, -0.00000203, -0.00000915, -0.00000125, 0.00000116, 0.00000535, 0.00000255, -0.00001567, -0.00000223, 0.00000451, -0.00000952, -0.00000204, -0.00000033, -0.00000074, 0.00000799, 0.00000434, -0.00001033, -0.00001556, -0.00000237, 0.00000893, 0.00000478, -0.00000797, -0.00000180, 0.00000675, -0.00001204, -0.00000330, 0.00001230, -0.00000697, 0.00000498, 0.00001010, -0.00001552, -0.00000130, 0.00000698, 0.00000409, 0.00000503, -0.00001131, 0.00000373, 0.00000736, -0.00000503, 0.00000250, -0.00000355, 0.00000207, 0.00000217, 0.00000256, -0.00000129, -0.00000824, 0.00001536, -0.00000746, -0.00001065, 0.00000463, -0.00000653, 0.00001885, -0.00000714, -0.00000518, 0.00001244, -0.00001481, 0.00001246, -0.00000676, -0.00000432, 0.00000554, -0.00001228, 0.00002469, -0.00001837, -0.00000075, 0.00001600, -0.00003324, 0.00003107, -0.00001158, -0.00002713, 0.00003552, -0.00001506, 0.00000356, 0.00001020, -0.00000225, 0.00000983, -0.00000868, 0.00002828, 0.00000733, -0.00000972, 0.00001803, -0.00002053, -0.00001432, 0.00000554, -0.00000791, -0.00000198, -0.00000358, -0.00000554, 0.00000642, 0.00000101, -0.00000876, 0.00000289, -0.00000362, 0.00000093, 0.00000308, -0.00001096, 0.00000394, -0.00001227, -0.00000607, 0.00000223, -0.00001517, 0.00000560, -0.00002333, -0.00000904, -0.00003245, -0.00010991, 0.00004435, 0.00012961, 0.00007163, 0.00012085, 0.00004362, -0.00007898, -0.00004457, -0.00000765, -0.00000669, -0.00000667, -0.00000560, -0.00000171, },
		{ 0.00000133, -0.00000053, -0.00000516, -0.00000197, -0.00000314, -0.00000426, 0.00000018, 0.00000086, 0.00000126, 0.00000470, 0.00000507, -0.00000470, -0.00000761, 0.00000032, 0.00000060, 0.00000716, -0.00000206, -0.00001585, 0.00000260, 0.00000032, -0.00000950, 0.00000029, -0.00000188, 0.00000175, 0.00000887, 0.00000018, -0.00001330, -0.00001326, 0.00000197, 0.00000952, 0.00000008, -0.00000901, 0.00000339, 0.00000346, -0.00001400, 0.00000301, 0.00000823, -0.00000720, 0.00001010, 0.00000232, -0.00001408, 0.00000460, 0.00000365, 0.00000541, 0.00000137, -0.00001079, 0.00000913, 0.00000307, -0.00000408, 0.00000214, -0.00000410, 0.00000401, 0.00000124, 0.00000339, -0.00000525, -0.00000290, 0.00001474, -0.00001423, -0.00000366, 0.00000109, -0.00000268, 0.00001905, -0.00001282, 0.00000312, 0.00000444, -0.00001102, 0.00001395, -0.00001319, 0.00000333, -0.00000222, -0.00000253, 0.00002062, -0.00002653, 0.00001595, -0.00000248, -0.00002240, 0.00003701, -0.00003443, -0.00000257, 0.00002974, -0.00002448, 0.00001702, 0.00000040, 0.00000132, 0.00000639, -0.00000639, 0.00003314, -0.00000536, -0.00000195, 0.00001574, -0.00002736, -0.00000621, 0.00000350, -0.00000786, -0.00000072, -0.00000586, -0.00000193, 0.00000731, -0.00000356, -0.00000629, 0.00000316, -0.00000517, 0.00000453, -0.00000202, -0.00000857, 0.00000378, -0.00001639, 0.00000112, -0.00000415, -0.00001188, 0.00000508, -0.00003019, -0.00000072, -0.00005786, -0.00010100, 0.00008767, 0.00012099, 0.00008513, 0.00012633, -0.00000533, -0.00008687, -0.00002553, -0.00000771, -0.00000723, -0.00000601, -0.00000510, -0.00000181, -0.00000354, },
		{ -0.00000215, -0.00000445, -0.00000182, -0.00000414, -0.00000338, 0.00000092, 0.00000121, 0.00000120, 0.00000514, 0.00000386, -0.00000686, -0.00000512, 0.00000054, 0.00000101, 0.00000803, -0.00000740, -0.00001329, 0.00000537, -0.00000431, -0.00000743, 0.00000107, -0.00000277, 0.00000494, 0.00000784, -0.00000423, -0.00001485, -0.00000989, 0.00000588, 0.00000807, -0.00000472, -0.00000712, 0.00000722, -0.00000160, -0.00001281, 0.00000750, 0.00000316, -0.00000417, 0.00001157, -0.00000492, -0.00000881, 0.00000680, 0.00000089, 0.00000700, -0.00000330, -0.00000719, 0.00001206, -0.00000144, -0.00000195, 0.00000055, -0.00000323, 0.00000511, 0.00000068, 0.00000318, -0.00000811, 0.00000364, 0.00001006, -0.00001722, 0.00000252, -0.00000351, 0.00000350, 0.00001561, -0.00001468, 0.00000867, -0.00000370, -0.00000374, 0.00001020, -0.00001493, 0.00000797, -0.00000830, 0.00000910, 0.00000871, -0.00002410, 0.00002517, -0.00002131, -0.00000152, 0.00002712, -0.00004563, 0.00002397, 0.00001108, -0.00002163, 0.00002506, -0.00001077, 0.00000871, -0.00000070, 0.00000068, 0.00003263, -0.00001618, 0.00000911, 0.00000785, -0.00002990, 0.00000151, -0.00000024, -0.00000595, -0.00000139, -0.00000669, 0.00000239, 0.00000544, -0.00000657, -0.00000301, 0.00000154, -0.00000450, 0.00000621, -0.00000708, -0.00000410, 0.00000014, -0.00001666, 0.00000642, -0.00001188, -0.00000477, -0.00000106, -0.00003181, 0.00000346, -0.00008711, -0.00007516, 0.00012268, 0.00011040, 0.00010772, 0.00011082, -0.00005428, -0.00007705, -0.00001108, -0.00001109, -0.00000556, -0.00000619, -0.00000455, -0.00000179, -0.00000304, 0.00000178, },
		{ -0.00000349, -0.00000234, -0.00000471, -0.00000229, 0.00000169, 0.00000123, 0.00000104, 0.00000583, 0.00000185, -0.00000789, -0.00000250, -0.00000020, 0.00000256, 0.00000714, -0.00001215, -0.00000879, 0.00000542, -0.00000792, -0.00000430, 0.00000030, -0.00000221, 0.00000779, 0.00000501, -0.00000813, -0.00001497, -0.00000571, 0.00000868, 0.00000467, -0.00000824, -0.00000303, 0.00000850, -0.00000670, -0.00000912, 0.00000919, -0.00000092, 0.00000047, 0.00000918, -0.00000919, -0.00000214, 0.00000545, 0.00000009, 0.00000747, -0.00000751, -0.00000135, 0.00001182, -0.00000476, 0.00000032, -0.00000166, -0.00000109, 0.00000510, 0.00000064, 0.00000175, -0.00000887, 0.00000948, 0.00000265, -0.00001589, 0.00000606, -0.00000726, 0.00001044, 0.00000954, -0.00001291, 0.00001023, -0.00000924, 0.00000400, 0.00000307, -0.00001208, 0.00000831, -0.00000976, 0.00001770, -0.00000622, -0.00001230, 0.00002279, -0.00003216, 0.00002093, 0.00000441, -0.00003951, 0.00004167, -0.00001265, -0.00000695, 0.00002282, -0.00001759, 0.00001623, -0.00000969, 0.00001218, 0.00002506, -0.00002147, 0.00001993, -0.00000441, -0.00002686, 0.00000657, -0.00000393, -0.00000315, -0.00000386, -0.00000499, 0.00000559, 0.00000172, -0.00000720, -0.00000050, -0.00000071, -0.00000206, 0.00000519, -0.00001020, 0.00000039, -0.00000569, -0.00001265, 0.00000744, -0.00001775, 0.00000355, -0.00001202, -0.00002657, -0.00000110, -0.00011362, -0.00003454, 0.00014507, 0.00010547, 0.00013103, 0.00007257, -0.00009138, -0.00005445, -0.00000531, -0.00001396, -0.00000273, -0.00000759, -0.00000330, -0.00000199, -0.00000246, 0.00000176, -0.00000340, },
		{ -0.00000333, -0.00000471, -0.00000105, 0.00000234, 0.00000079, 0.00000116, 0.00000641, -0.00000070, -0.00000756, -0.00000053, -0.00000117, 0.00000483, 0.00000427, -0.00001509, -0.00000368, 0.00000305, -0.00000954, -0.00000137, -0.00000133, -0.00000003, 0.00000928, 0.00000101, -0.00001102, -0.00001379, -0.00000114, 0.00000973, 0.00000003, -0.00000945, 0.00000187, 0.00000702, -0.00001032, -0.00000435, 0.00000829, -0.00000270, 0.00000458, 0.00000423, -0.00000938, 0.00000333, 0.00000201, 0.00000143, 0.00000593, -0.00000973, 0.00000511, 0.00000872, -0.00000606, 0.00000172, -0.00000356, 0.00000171, 0.00000410, 0.00000105, -0.00000064, -0.00000712, 0.00001290, -0.00000540, -0.00001099, 0.00000610, -0.00000855, 0.00001636, 0.00000244, -0.00000872, 0.00000807, -0.00001069, 0.00000938, -0.00000465, -0.00000644, 0.00000512, -0.00000587, 0.00001959, -0.00001810, 0.00000362, 0.00000975, -0.00002990, 0.00003525, -0.00002229, -0.00001742, 0.00004291, -0.00003113, 0.00001355, 0.00001005, -0.00001591, 0.00001922, -0.00001671, 0.00002529, 0.00001101, -0.00001871, 0.00002629, -0.00001784, -0.00001882, 0.00000760, -0.00000576, -0.00000118, -0.00000682, -0.00000083, 0.00000607, -0.00000201, -0.00000590, 0.00000035, -0.00000214, 0.00000077, 0.00000197, -0.00001047, 0.00000287, -0.00001134, -0.00000580, 0.00000352, -0.00001908, 0.00000947, -0.00002472, -0.00001557, -0.00001719, -0.00012998, 0.00001502, 0.00015491, 0.00011133, 0.00014371, 0.00001705, -0.00010767, -0.00002789, -0.00000827, -0.00001327, -0.00000086, -0.00000942, -0.00000121, -0.00000283, -0.00000161, 0.00000131, -0.00000418, -0.00000260, },
		{ -0.00000408, 0.00000020, 0.00000270, -0.00000001, 0.00000178, 0.00000648, -0.00000331, -0.00000604, 0.00000034, -0.00000164, 0.00000700, -0.00000017, -0.00001558, 0.00000060, -0.00000073, -0.00000897, 0.00000038, -0.00000277, 0.00000327, 0.00000879, -0.00000333, -0.00001268, -0.00001145, 0.00000325, 0.00000866, -0.00000471, -0.00000807, 0.00000605, 0.00000351, -0.00001165, -0.00000001, 0.00000593, -0.00000205, 0.00000648, -0.00000108, -0.00000602, 0.00000581, -0.00000149, 0.00000400, 0.00000249, -0.00000899, 0.00001034, 0.00000396, -0.00000534, 0.00000170, -0.00000429, 0.00000431, 0.00000257, 0.00000158, -0.00000338, -0.00000322, 0.00001289, -0.00001190, -0.00000427, 0.00000299, -0.00000655, 0.00001982, -0.00000409, -0.00000381, 0.00000363, -0.00000823, 0.00001081, -0.00001039, -0.00000049, 0.00000072, 0.00000157, 0.00001401, -0.00002224, 0.00001668, -0.00000828, -0.00001498, 0.00003476, -0.00004187, 0.00001249, 0.00002674, -0.00003593, 0.00003075, -0.00000845, -0.00000547, 0.00001458, -0.00001753, 0.00003533, -0.00000612, -0.00000798, 0.00002497, -0.00002824, -0.00000836, 0.00000493, -0.00000483, -0.00000162, -0.00000819, 0.00000430, 0.00000348, -0.00000397, -0.00000414, -0.00000011, -0.00000204, 0.00000248, -0.00000196, -0.00000837, 0.00000238, -0.00001450, 0.00000113, -0.00000393, -0.00001497, 0.00000985, -0.00003489, -0.00000273, -0.00004415, -0.00013029, 0.00006582, 0.00015680, 0.00012776, 0.00013538, -0.00004389, -0.00010035, -0.00000676, -0.00001612, -0.00000826, -0.00000182, -0.00001013, 0.00000104, -0.00000434, -0.00000024, 0.00000036, -0.00000439, -0.00000209, -0.00000452, },
		{ 0.00000128, 0.00000264, -0.00000091, 0.00000290, 0.00000580, -0.00000542, -0.00000383, 0.00000010, -0.00000113, 0.00000822, -0.00000524, -0.00001372, 0.00000302, -0.00000461, -0.00000678, 0.00000056, -0.00000304, 0.00000669, 0.00000627, -0.00000718, -0.00001312, -0.00000815, 0.00000673, 0.00000559, -0.00000832, -0.00000462, 0.00000840, -0.00000081, -0.00001082, 0.00000288, 0.00000352, 0.00000009, 0.00000561, -0.00000461, -0.00000089, 0.00000492, -0.00000337, 0.00000636, -0.00000184, -0.00000531, 0.00001293, -0.00000090, -0.00000334, 0.00000037, -0.00000344, 0.00000591, 0.00000109, 0.00000181, -0.00000562, 0.00000181, 0.00000939, -0.00001515, 0.00000210, -0.00000195, -0.00000142, 0.00002000, -0.00000887, 0.00000030, -0.00000113, -0.00000333, 0.00000829, -0.00001271, 0.00000372, -0.00000230, 0.00000912, 0.00000340, -0.00001738, 0.00002125, -0.00002331, 0.00000647, 0.00001892, -0.00004562, 0.00003837, -0.00000048, -0.00002438, 0.00003652, -0.00002483, 0.00000963, 0.00000273, -0.00000978, 0.00003763, -0.00002093, 0.00000747, 0.00001535, -0.00003213, 0.00000091, 0.00000061, -0.00000185, -0.00000489, -0.00000617, 0.00000789, -0.00000092, -0.00000346, -0.00000345, -0.00000063, -0.00000085, 0.00000236, -0.00000497, -0.00000544, -0.00000051, -0.00001406, 0.00000541, -0.00001194, -0.00000689, 0.00000354, -0.00003875, 0.00000639, -0.00007764, -0.00011222, 0.00011108, 0.00015783, 0.00014819, 0.00010098, -0.00009547, -0.00007398, 0.00000245, -0.00002310, -0.00000093, -0.00000600, -0.00000833, 0.00000215, -0.00000592, 0.00000156, -0.00000115, -0.00000395, -0.00000223, -0.00000554, -0.00000285, },
		{ 0.00000218, -0.00000165, 0.00000435, 0.00000433, -0.00000663, -0.00000155, -0.00000090, 0.00000039, 0.00000790, -0.00000984, -0.00001022, 0.00000325, -0.00000744, -0.00000401, -0.00000048, -0.00000168, 0.00000909, 0.00000224, -0.00000996, -0.00001245, -0.00000423, 0.00000861, 0.00000118, -0.00000990, -0.00000019, 0.00000852, -0.00000468, -0.00000859, 0.00000415, 0.00000214, 0.00000225, 0.00000268, -0.00000512, 0.00000376, 0.00000162, -0.00000291, 0.00000719, -0.00000551, 0.00000026, 0.00001240, -0.00000449, -0.00000113, -0.00000157, -0.00000121, 0.00000612, 0.00000016, 0.00000139, -0.00000659, 0.00000655, 0.00000341, -0.00001448, 0.00000618, -0.00000679, 0.00000576, 0.00001692, -0.00001133, 0.00000272, -0.00000456, 0.00000194, 0.00000315, -0.00001162, 0.00000543, -0.00000236, 0.00001333, -0.00000769, -0.00000616, 0.00001555, -0.00002861, 0.00002527, -0.00000614, -0.00003110, 0.00004962, -0.00002756, -0.00000087, 0.00002741, -0.00003168, 0.00002260, -0.00001213, 0.00000532, 0.00002977, -0.00002803, 0.00002213, 0.00000002, -0.00002837, 0.00000595, -0.00000249, 0.00000097, -0.00000961, -0.00000041, 0.00000771, -0.00000467, -0.00000138, -0.00000443, 0.00000011, 0.00000006, 0.00000090, -0.00000627, -0.00000334, -0.00000407, -0.00001063, 0.00000555, -0.00001740, 0.00000195, -0.00000798, -0.00003477, 0.00000657, -0.00011072, -0.00007765, 0.00014749, 0.00016412, 0.00016134, 0.00004368, -0.00012487, -0.00003897, -0.00000140, -0.00002428, 0.00000504, -0.00001177, -0.00000383, 0.00000098, -0.00000647, 0.00000325, -0.00000303, -0.00000291, -0.00000320, -0.00000594, -0.00000125, 0.00000202, },
		{ -0.00000201, 0.00000587, 0.00000222, -0.00000678, 0.00000025, -0.00000212, 0.00000258, 0.00000589, -0.00001309, -0.00000616, 0.00000162, -0.00000859, -0.00000167, -0.00000193, 0.00000115, 0.00000961, -0.00000243, -0.00001139, -0.00001085, -0.00000021, 0.00000850, -0.00000353, -0.00000914, 0.00000398, 0.00000675, -0.00000733, -0.00000597, 0.00000432, 0.00000209, 0.00000313, -0.00000075, -0.00000270, 0.00000614, -0.00000236, -0.00000053, 0.00000596, -0.00000717, 0.00000611, 0.00000924, -0.00000612, 0.00000031, -0.00000319, 0.00000169, 0.00000503, -0.00000000, 0.00000021, -0.00000584, 0.00000961, -0.00000337, -0.00001038, 0.00000688, -0.00000964, 0.00001334, 0.00001124, -0.00001153, 0.00000332, -0.00000577, 0.00000573, -0.00000272, -0.00000819, 0.00000519, 0.00000042, 0.00001237, -0.00001473, 0.00000609, 0.00000243, -0.00002170, 0.00003306, -0.00003005, -0.00000350, 0.00004139, -0.00004315, 0.00002484, 0.00000643, -0.00002553, 0.00002703, -0.00002354, 0.00002293, 0.00001316, -0.00002443, 0.00003025, -0.00001594, -0.00001882, 0.00000575, -0.00000226, 0.00000099, -0.00001294, 0.00000734, 0.00000306, -0.00000536, 0.00000012, -0.00000616, 0.00000241, -0.00000059, -0.00000049, -0.00000636, -0.00000279, -0.00000648, -0.00000621, 0.00000201, -0.00001846, 0.00000803, -0.00002120, -0.00002445, -0.00000500, -0.00013613, -0.00003153, 0.00017616, 0.00017730, 0.00015534, -0.00002512, -0.00012581, -0.00000777, -0.00001396, -0.00001787, 0.00000618, -0.00001618, 0.00000197, -0.00000260, -0.00000506, 0.00000393, -0.00000470, -0.00000165, -0.00000496, -0.00000530, 0.00000031, 0.00000236, 0.00000139, },
		{ 0.00000721, -0.00000028, -0.00000591, 0.00000125, -0.00000306, 0.00000488, 0.00000249, -0.00001452, -0.00000256, -0.00000106, -0.00000809, -0.00000043, -0.00000282, 0.00000469, 0.00000788, -0.00000676, -0.00001151, -0.00000852, 0.00000330, 0.00000640, -0.00000741, -0.00000641, 0.00000693, 0.00000390, -0.00000858, -0.00000368, 0.00000424, 0.00000283, 0.00000220, -0.00000298, 0.00000138, 0.00000556, -0.00000529, 0.00000257, 0.00000308, -0.00000616, 0.00001059, 0.00000464, -0.00000588, 0.00000054, -0.00000369, 0.00000433, 0.00000318, 0.00000043, -0.00000150, -0.00000345, 0.00001008, -0.00000904, -0.00000437, 0.00000421, -0.00000918, 0.00001949, 0.00000413, -0.00000995, 0.00000255, -0.00000481, 0.00000701, -0.00000756, -0.00000387, 0.00000426, 0.00000439, 0.00000681, -0.00001518, 0.00001385, -0.00001195, -0.00000557, 0.00002615, -0.00004248, 0.00002638, 0.00001678, -0.00004057, 0.00004181, -0.00001815, -0.00000860, 0.00002007, -0.00002577, 0.00003627, -0.00000698, -0.00001117, 0.00002832, -0.00002701, -0.00000768, 0.00000198, 0.00000140, -0.00000327, -0.00001183, 0.00001362, -0.00000443, -0.00000207, -0.00000110, -0.00000659, 0.00000512, -0.00000304, -0.00000049, -0.00000666, -0.00000314, -0.00000693, -0.00000295, -0.00000319, -0.00001536, 0.00000908, -0.00003215, -0.00001171, -0.00002771, -0.00014858, 0.00002054, 0.00020131, 0.00019257, 0.00012273, -0.00008899, -0.00010092, 0.00000980, -0.00002754, -0.00000608, 0.00000121, -0.00001642, 0.00000655, -0.00000742, -0.00000155, 0.00000287, -0.00000539, -0.00000085, -0.00000703, -0.00000348, 0.00000132, 0.00000246, 0.00000033, -0.00000184, },
		{ -0.00000289, -0.00000419, 0.00000132, -0.00000335, 0.00000673, -0.00000174, -0.00001413, -0.00000010, -0.00000386, -0.00000647, -0.00000039, -0.00000243, 0.00000785, 0.00000416, -0.00000993, -0.00001055, -0.00000574, 0.00000569, 0.00000283, -0.00000959, -0.00000257, 0.00000821, 0.00000082, -0.00000879, -0.00000193, 0.00000456, 0.00000337, -0.00000007, -0.00000294, 0.00000531, 0.00000252, -0.00000612, 0.00000505, -0.00000033, -0.00000275, 0.00001261, 0.00000002, -0.00000445, -0.00000027, -0.00000279, 0.00000592, 0.00000131, 0.00000104, -0.00000323, -0.00000005, 0.00000782, -0.00001210, 0.00000158, -0.00000068, -0.00000504, 0.00002271, -0.00000306, -0.00000728, 0.00000115, -0.00000243, 0.00000569, -0.00001032, 0.00000017, 0.00000382, 0.00000731, -0.00000073, -0.00000957, 0.00001391, -0.00002095, 0.00001258, 0.00000722, -0.00003770, 0.00004647, -0.00001432, -0.00002071, 0.00004258, -0.00003629, 0.00001203, 0.00000395, -0.00001661, 0.00003954, -0.00002349, 0.00000676, 0.00001681, -0.00002964, 0.00000040, -0.00000183, 0.00000614, -0.00001092, -0.00000479, 0.00001473, -0.00001116, 0.00000375, -0.00000574, -0.00000377, 0.00000600, -0.00000595, 0.00000105, -0.00000839, -0.00000280, -0.00000622, -0.00000191, -0.00000768, -0.00001007, 0.00000501, -0.00003807, -0.00000132, -0.00005801, -0.00014595, 0.00007498, 0.00022705, 0.00019974, 0.00006448, -0.00013195, -0.00006082, 0.00001005, -0.00003462, 0.00000594, -0.00000816, -0.00001139, 0.00000744, -0.00001128, 0.00000304, 0.00000003, -0.00000459, -0.00000129, -0.00000853, -0.00000084, 0.00000150, 0.00000256, -0.00000104, -0.00000099, 0.00000810, },
		{ -0.00000189, 0.00000046, -0.00000280, 0.00000767, -0.00000613, -0.00001228, 0.00000095, -0.00000604, -0.00000447, -0.00000114, -0.00000055, 0.00000955, -0.00000079, -0.00001144, -0.00000886, -0.00000287, 0.00000656, -0.00000135, -0.00000972, 0.00000138, 0.00000794, -0.00000188, -0.00000844, -0.00000045, 0.00000539, 0.00000285, -0.00000253, -0.00000058, 0.00000748, -0.00000161, -0.00000481, 0.00000598, -0.00000295, 0.00000198, 0.00001193, -0.00000352, -0.00000275, -0.00000152, -0.00000074, 0.00000610, 0.00000004, 0.00000134, -0.00000438, 0.00000339, 0.00000352, -0.00001188, 0.00000560, -0.00000594, 0.00000206, 0.00002213, -0.00000905, -0.00000420, -0.00000021, 0.00000030, 0.00000246, -0.00001069, 0.00000341, 0.00000426, 0.00000759, -0.00000691, -0.00000112, 0.00000675, -0.00002048, 0.00002460, -0.00001571, -0.00001711, 0.00004855, -0.00003930, 0.00000831, 0.00002644, -0.00004058, 0.00002749, -0.00001472, 0.00000124, 0.00003065, -0.00003024, 0.00002226, 0.00000009, -0.00002388, 0.00000262, -0.00000212, 0.00000823, -0.00001856, 0.00000680, 0.00000875, -0.00001315, 0.00000866, -0.00001217, 0.00000269, 0.00000306, -0.00000691, 0.00000260, -0.00001144, -0.00000046, -0.00000620, -0.00000243, -0.00001015, -0.00000505, -0.00000240, -0.00003848, 0.00000306, -0.00009109, -0.00012867, 0.00013143, 0.00025370, 0.00018705, -0.00000906, -0.00014420, -0.00002011, -0.00000359, -0.00003120, 0.00001267, -0.00001776, -0.00000249, 0.00000359, -0.00001201, 0.00000689, -0.00000370, -0.00000244, -0.00000340, -0.00000851, 0.00000180, 0.00000094, 0.00000279, -0.00000272, 0.00000075, 0.00000822, -0.00000527, },
		{ -0.00000116, -0.00000134, 0.00000739, -0.00001008, -0.00000948, 0.00000069, -0.00000721, -0.00000276, -0.00000198, 0.00000247, 0.00000909, -0.00000589, -0.00001121, -0.00000687, -0.00000028, 0.00000582, -0.00000520, -0.00000798, 0.00000459, 0.00000661, -0.00000402, -0.00000791, 0.00000124, 0.00000625, 0.00000098, -0.00000384, 0.00000316, 0.00000711, -0.00000526, -0.00000211, 0.00000522, -0.00000380, 0.00000665, 0.00000910, -0.00000545, -0.00000150, -0.00000245, 0.00000177, 0.00000504, -0.00000035, 0.00000104, -0.00000452, 0.00000585, -0.00000153, -0.00000872, 0.00000654, -0.00000954, 0.00001050, 0.00001780, -0.00001289, -0.00000131, -0.00000107, 0.00000242, -0.00000161, -0.00000900, 0.00000582, 0.00000509, 0.00000503, -0.00000927, 0.00000590, -0.00000376, -0.00001090, 0.00002499, -0.00003262, 0.00001120, 0.00003170, -0.00004812, 0.00003457, -0.00000025, -0.00002950, 0.00003109, -0.00002785, 0.00002104, 0.00001246, -0.00002488, 0.00002913, -0.00001527, -0.00001344, -0.00000057, 0.00000263, 0.00000447, -0.00002152, 0.00001868, -0.00000307, -0.00000814, 0.00000876, -0.00001680, 0.00001075, -0.00000395, -0.00000390, 0.00000163, -0.00001392, 0.00000366, -0.00000844, -0.00000263, -0.00001108, -0.00000175, -0.00001079, -0.00003486, -0.00000009, -0.00012286, -0.00009774, 0.00019170, 0.00027554, 0.00014662, -0.00008091, -0.00012566, 0.00000807, -0.00002246, -0.00001841, 0.00001085, -0.00002298, 0.00000675, -0.00000388, -0.00000857, 0.00000813, -0.00000675, 0.00000008, -0.00000684, -0.00000637, 0.00000341, 0.00000022, 0.00000296, -0.00000458, 0.00000346, 0.00000721, -0.00000700, 0.00000061, },
		{ 0.00000093, 0.00000571, -0.00001307, -0.00000632, -0.00000056, -0.00000735, -0.00000172, -0.00000217, 0.00000581, 0.00000628, -0.00001001, -0.00000952, -0.00000496, 0.00000169, 0.00000374, -0.00000793, -0.00000498, 0.00000664, 0.00000478, -0.00000574, -0.00000717, 0.00000340, 0.00000637, -0.00000176, -0.00000316, 0.00000680, 0.00000442, -0.00000730, 0.00000081, 0.00000334, -0.00000262, 0.00001001, 0.00000515, -0.00000587, -0.00000101, -0.00000252, 0.00000394, 0.00000329, 0.00000004, 0.00000010, -0.00000354, 0.00000661, -0.00000585, -0.00000389, 0.00000433, -0.00000993, 0.00001823, 0.00001058, -0.00001412, 0.00000097, -0.00000130, 0.00000330, -0.00000545, -0.00000580, 0.00000756, 0.00000536, 0.00000092, -0.00000735, 0.00000825, -0.00001232, 0.00000317, 0.00001347, -0.00003595, 0.00003575, 0.00000270, -0.00003742, 0.00004737, -0.00002686, -0.00000800, 0.00002135, -0.00002952, 0.00003484, -0.00000817, -0.00001006, 0.00002481, -0.00002349, -0.00000392, -0.00000558, 0.00001068, -0.00000565, -0.00001621, 0.00002517, -0.00001615, 0.00000274, 0.00000193, -0.00001572, 0.00001640, -0.00001266, 0.00000313, -0.00000387, -0.00001288, 0.00000725, -0.00001279, -0.00000086, -0.00001221, 0.00000012, -0.00001881, -0.00002926, -0.00001048, -0.00015083, -0.00005246, 0.00025667, 0.00028155, 0.00007914, -0.00013319, -0.00008595, 0.00001684, -0.00003679, -0.00000184, 0.00000119, -0.00002094, 0.00001228, -0.00001197, -0.00000181, 0.00000587, -0.00000763, 0.00000146, -0.00001044, -0.00000233, 0.00000325, 0.00000010, 0.00000256, -0.00000611, 0.00000696, 0.00000477, -0.00000757, 0.00000279, -0.00000323, },
		{ 0.00000262, -0.00001469, -0.00000332, -0.00000235, -0.00000667, -0.00000140, -0.00000125, 0.00000841, 0.00000157, -0.00001229, -0.00000697, -0.00000343, 0.00000286, 0.00000086, -0.00000910, -0.00000151, 0.00000751, 0.00000282, -0.00000724, -0.00000591, 0.00000586, 0.00000512, -0.00000434, -0.00000045, 0.00000894, 0.00000044, -0.00000737, 0.00000299, 0.00000132, 0.00000016, 0.00001137, 0.00000114, -0.00000529, -0.00000113, -0.00000161, 0.00000517, 0.00000155, 0.00000082, -0.00000120, -0.00000175, 0.00000552, -0.00000826, 0.00000093, -0.00000001, -0.00000655, 0.00002329, 0.00000199, -0.00001283, 0.00000239, -0.00000100, 0.00000275, -0.00000819, -0.00000170, 0.00000857, 0.00000434, -0.00000272, -0.00000281, 0.00000521, -0.00001472, 0.00001502, -0.00000484, -0.00002382, 0.00004651, -0.00002676, -0.00001184, 0.00004169, -0.00004299, 0.00001474, 0.00000262, -0.00001869, 0.00003699, -0.00002329, 0.00000763, 0.00001161, -0.00002220, -0.00000006, -0.00000748, 0.00001751, -0.00001902, -0.00000233, 0.00002185, -0.00002427, 0.00001489, -0.00001048, -0.00000699, 0.00001547, -0.00001867, 0.00001145, -0.00001363, -0.00000610, 0.00000685, -0.00001683, 0.00000274, -0.00001502, 0.00000204, -0.00002675, -0.00002276, -0.00002763, -0.00017327, 0.00001039, 0.00032248, 0.00025938, -0.00000454, -0.00015370, -0.00004056, 0.00000756, -0.00003989, 0.00001131, -0.00001196, -0.00001200, 0.00001140, -0.00001708, 0.00000580, 0.00000076, -0.00000574, 0.00000042, -0.00001251, 0.00000243, 0.00000134, 0.00000113, 0.00000094, -0.00000649, 0.00001061, 0.00000097, -0.00000654, 0.00000403, -0.00000523, 0.00000372, },
		{ -0.00001468, -0.00000096, -0.00000427, -0.00000551, -0.00000150, 0.00000084, 0.00000930, -0.00000406, -0.00001227, -0.00000436, -0.00000242, 0.00000319, -0.00000222, -0.00000867, 0.00000174, 0.00000746, 0.00000082, -0.00000859, -0.00000378, 0.00000799, 0.00000240, -0.00000559, 0.00000344, 0.00000883, -0.00000349, -0.00000585, 0.00000398, 0.00000007, 0.00000362, 0.00001070, -0.00000213, -0.00000431, -0.00000144, -0.00000001, 0.00000527, 0.00000037, 0.00000150, -0.00000245, 0.00000029, 0.00000304, -0.00000826, 0.00000418, -0.00000478, 0.00000005, 0.00002438, -0.00000620, -0.00000961, 0.00000289, -0.00000042, 0.00000094, -0.00000927, 0.00000270, 0.00000858, 0.00000207, -0.00000426, 0.00000153, -0.00000114, -0.00000989, 0.00001898, -0.00002166, -0.00000102, 0.00003912, -0.00004498, 0.00001801, 0.00002012, -0.00004291, 0.00002934, -0.00001686, 0.00000018, 0.00002675, -0.00002733, 0.00002045, -0.00000429, -0.00001358, -0.00000326, -0.00000275, 0.00001789, -0.00002965, 0.00001610, 0.00000806, -0.00002238, 0.00002191, -0.00002353, 0.00000762, 0.00000604, -0.00001755, 0.00001621, -0.00002424, 0.00000600, -0.00000023, -0.00001675, 0.00000558, -0.00001909, 0.00000501, -0.00003590, -0.00001487, -0.00005247, -0.00018686, 0.00009496, 0.00037826, 0.00020137, -0.00008592, -0.00014048, -0.00000483, -0.00001157, -0.00003084, 0.00001559, -0.00002262, 0.00000026, 0.00000418, -0.00001666, 0.00001116, -0.00000511, -0.00000187, -0.00000337, -0.00001161, 0.00000618, -0.00000143, 0.00000320, -0.00000222, -0.00000481, 0.00001332, -0.00000365, -0.00000381, 0.00000372, -0.00000646, 0.00000650, -0.00000168, },
		{ 0.00000036, -0.00000590, -0.00000422, -0.00000157, 0.00000372, 0.00000788, -0.00000933, -0.00001009, -0.00000241, -0.00000186, 0.00000279, -0.00000494, -0.00000694, 0.00000433, 0.00000681, -0.00000130, -0.00000954, -0.00000066, 0.00000896, -0.00000119, -0.00000478, 0.00000720, 0.00000657, -0.00000630, -0.00000357, 0.00000393, 0.00000006, 0.00000676, 0.00000849, -0.00000432, -0.00000336, -0.00000153, 0.00000176, 0.00000445, -0.00000001, 0.00000170, -0.00000328, 0.00000196, 0.00000005, -0.00000618, 0.00000496, -0.00000809, 0.00000832, 0.00002120, -0.00001241, -0.00000538, 0.00000263, 0.00000010, -0.00000166, -0.00000841, 0.00000674, 0.00000726, -0.00000058, -0.00000328, 0.00000324, -0.00000710, -0.00000038, 0.00001307, -0.00002925, 0.00002308, 0.00001666, -0.00004509, 0.00004013, -0.00000843, -0.00002772, 0.00003031, -0.00002874, 0.00001917, 0.00000865, -0.00001963, 0.00002309, -0.00001573, -0.00000330, -0.00001071, 0.00000848, 0.00000879, -0.00003125, 0.00003181, -0.00001197, -0.00000945, 0.00001861, -0.00003049, 0.00002260, -0.00000981, -0.00000739, 0.00001266, -0.00003006, 0.00001937, -0.00001364, -0.00000960, 0.00000366, -0.00002143, 0.00000782, -0.00004656, -0.00000468, -0.00008759, -0.00018429, 0.00020218, 0.00040728, 0.00010988, -0.00014504, -0.00010255, 0.00001175, -0.00002963, -0.00001439, 0.00000982, -0.00002599, 0.00001067, -0.00000644, -0.00001049, 0.00001198, -0.00000920, 0.00000203, -0.00000887, -0.00000733, 0.00000733, -0.00000351, 0.00000535, -0.00000650, -0.00000060, 0.00001387, -0.00000796, 0.00000018, 0.00000160, -0.00000623, 0.00000848, -0.00000665, -0.00001299, },
		{ -0.00000694, -0.00000309, -0.00000114, 0.00000660, 0.00000408, -0.00001293, -0.00000644, -0.00000164, -0.00000150, 0.00000182, -0.00000691, -0.00000437, 0.00000609, 0.00000578, -0.00000367, -0.00000961, 0.00000312, 0.00000813, -0.00000457, -0.00000193, 0.00000959, 0.00000299, -0.00000752, -0.00000136, 0.00000343, 0.00000124, 0.00000884, 0.00000545, -0.00000546, -0.00000260, -0.00000116, 0.00000318, 0.00000316, 0.00000032, 0.00000129, -0.00000352, 0.00000283, -0.00000251, -0.00000299, 0.00000328, -0.00000855, 0.00001618, 0.00001448, -0.00001565, -0.00000115, 0.00000185, 0.00000025, -0.00000438, -0.00000569, 0.00000969, 0.00000461, -0.00000236, -0.00000080, 0.00000149, -0.00000925, 0.00000892, -0.00000002, -0.00002385, 0.00003848, -0.00001171, -0.00002771, 0.00004607, -0.00003271, -0.00000461, 0.00001826, -0.00002825, 0.00003047, -0.00000951, -0.00000471, 0.00001506, -0.00001793, 0.00000229, -0.00001665, 0.00002189, -0.00000845, -0.00002039, 0.00003718, -0.00003025, 0.00001031, 0.00000394, -0.00002602, 0.00003062, -0.00002597, 0.00000928, -0.00000095, -0.00002569, 0.00002737, -0.00002897, 0.00000426, -0.00000599, -0.00001774, 0.00000648, -0.00005613, 0.00000661, -0.00013527, -0.00015374, 0.00032500, 0.00039193, -0.00000065, -0.00016795, -0.00005620, 0.00000866, -0.00003803, 0.00000174, -0.00000257, -0.00002070, 0.00001485, -0.00001595, -0.00000090, 0.00000794, -0.00000969, 0.00000368, -0.00001390, -0.00000069, 0.00000523, -0.00000351, 0.00000611, -0.00001063, 0.00000579, 0.00001139, -0.00001058, 0.00000438, -0.00000203, -0.00000412, 0.00000880, -0.00001146, -0.00000990, 0.00000043, },
		{ -0.00000225, 0.00000012, 0.00000850, -0.00000147, -0.00001391, -0.00000249, -0.00000214, -0.00000105, 0.00000041, -0.00000790, -0.00000141, 0.00000707, 0.00000440, -0.00000619, -0.00000836, 0.00000676, 0.00000543, -0.00000656, 0.00000216, 0.00000987, -0.00000078, -0.00000732, 0.00000032, 0.00000308, 0.00000312, 0.00000954, 0.00000226, -0.00000577, -0.00000199, -0.00000036, 0.00000395, 0.00000192, 0.00000103, 0.00000034, -0.00000321, 0.00000283, -0.00000394, 0.00000009, 0.00000005, -0.00000567, 0.00002167, 0.00000577, -0.00001567, 0.00000223, 0.00000086, -0.00000019, -0.00000648, -0.00000155, 0.00001085, 0.00000124, -0.00000239, 0.00000130, -0.00000249, -0.00000626, 0.00001314, -0.00001415, -0.00000740, 0.00003882, -0.00003466, -0.00000046, 0.00003423, -0.00004383, 0.00001652, -0.00000069, -0.00001623, 0.00002986, -0.00002025, 0.00000984, 0.00000090, -0.00001084, -0.00000054, -0.00001528, 0.00003059, -0.00002785, 0.00000113, 0.00002785, -0.00003819, 0.00002858, -0.00001757, -0.00000924, 0.00002600, -0.00003450, 0.00002563, -0.00002143, -0.00000921, 0.00002364, -0.00003879, 0.00001999, -0.00002267, -0.00000512, -0.00000421, -0.00005870, 0.00001326, -0.00019344, -0.00008171, 0.00044551, 0.00032125, -0.00010658, -0.00015232, -0.00001841, -0.00000665, -0.00003401, 0.00001070, -0.00001535, -0.00000937, 0.00001125, -0.00002031, 0.00000831, 0.00000086, -0.00000648, 0.00000166, -0.00001606, 0.00000605, 0.00000068, -0.00000088, 0.00000418, -0.00001276, 0.00001294, 0.00000587, -0.00001039, 0.00000740, -0.00000623, -0.00000029, 0.00000686, -0.00001509, -0.00000607, -0.00000070, -0.00000720, },
		{ 0.00000225, 0.00000849, -0.00000754, -0.00001198, 0.00000053, -0.00000349, -0.00000030, -0.00000136, -0.00000784, 0.00000151, 0.00000737, 0.00000260, -0.00000854, -0.00000561, 0.00000930, 0.00000143, -0.00000637, 0.00000622, 0.00000812, -0.00000380, -0.00000626, 0.00000146, 0.00000329, 0.00000507, 0.00000898, -0.00000062, -0.00000553, -0.00000138, 0.00000063, 0.00000404, 0.00000109, 0.00000175, -0.00000091, -0.00000254, 0.00000220, -0.00000405, 0.00000203, -0.00000331, -0.00000003, 0.00002346, -0.00000303, -0.00001295, 0.00000427, -0.00000012, -0.00000126, -0.00000726, 0.00000319, 0.00000982, -0.00000177, -0.00000072, 0.00000141, -0.00000612, 0.00000044, 0.00000984, -0.00002251, 0.00001326, 0.00002405, -0.00004342, 0.00002547, 0.00001019, -0.00003883, 0.00002744, -0.00001786, 0.00000145, 0.00001857, -0.00001982, 0.00001692, -0.00001195, 0.00000078, -0.00001076, -0.00000407, 0.00002841, -0.00004087, 0.00002612, 0.00000542, -0.00003062, 0.00003625, -0.00003710, 0.00001481, 0.00000793, -0.00002936, 0.00003299, -0.00004082, 0.00001562, 0.00000569, -0.00003584, 0.00002933, -0.00004060, 0.00001483, -0.00002726, -0.00004727, 0.00000456, -0.00025158, 0.00004076, 0.00053632, 0.00019790, -0.00018237, -0.00010865, -0.00000014, -0.00002291, -0.00002118, 0.00000973, -0.00002262, 0.00000279, 0.00000181, -0.00001778, 0.00001358, -0.00000613, -0.00000132, -0.00000378, -0.00001382, 0.00001035, -0.00000432, 0.00000356, -0.00000082, -0.00001123, 0.00001871, -0.00000156, -0.00000703, 0.00000796, -0.00000961, 0.00000439, 0.00000258, -0.00001665, -0.00000238, -0.00000267, -0.00000669, -0.00000165, },
		{ 0.00000602, -0.00001257, -0.00000768, 0.00000161, -0.00000493, 0.00000066, -0.00000339, -0.00000670, 0.00000409, 0.00000708, 0.00000030, -0.00001015, -0.00000162, 0.00000992, -0.00000274, -0.00000394, 0.00000903, 0.00000504, -0.00000558, -0.00000495, 0.00000237, 0.00000405, 0.00000655, 0.00000749, -0.00000296, -0.00000488, -0.00000072, 0.00000156, 0.00000363, 0.00000081, 0.00000216, -0.00000225, -0.00000172, 0.00000139, -0.00000316, 0.00000238, -0.00000538, 0.00000698, 0.00002121, -0.00001018, -0.00000853, 0.00000488, -0.00000102, -0.00000277, -0.00000628, 0.00000741, 0.00000683, -0.00000328, 0.00000159, -0.00000090, -0.00000688, 0.00000729, 0.00000030, -0.00002080, 0.00002942, 0.00000028, -0.00003527, 0.00004002, -0.00001606, -0.00002144, 0.00002495, -0.00002608, 0.00001677, 0.00000253, -0.00000987, 0.00001331, -0.00001682, 0.00000966, -0.00002278, 0.00001443, 0.00001328, -0.00004019, 0.00004446, -0.00002237, -0.00000872, 0.00002774, -0.00004501, 0.00003652, -0.00001803, -0.00000988, 0.00002491, -0.00004921, 0.00003953, -0.00002245, -0.00001716, 0.00002427, -0.00005031, 0.00003458, -0.00005985, -0.00001821, -0.00003201, -0.00028912, 0.00021061, 0.00056728, 0.00004128, -0.00020989, -0.00005631, -0.00000246, -0.00003068, -0.00000706, 0.00000113, -0.00002162, 0.00001050, -0.00000904, -0.00000968, 0.00001318, -0.00001013, 0.00000301, -0.00001065, -0.00000744, 0.00001046, -0.00000736, 0.00000783, -0.00000784, -0.00000536, 0.00002096, -0.00000887, -0.00000126, 0.00000547, -0.00001080, 0.00000852, -0.00000334, -0.00001572, 0.00000026, -0.00000488, -0.00000561, -0.00000105, 0.00000486, },
		{ -0.00001511, -0.00000231, 0.00000041, -0.00000556, 0.00000147, -0.00000544, -0.00000457, 0.00000605, 0.00000623, -0.00000241, -0.00001046, 0.00000284, 0.00000833, -0.00000581, 0.00000001, 0.00000983, 0.00000167, -0.00000625, -0.00000375, 0.00000343, 0.00000507, 0.00000731, 0.00000545, -0.00000468, -0.00000393, -0.00000004, 0.00000224, 0.00000302, 0.00000100, 0.00000206, -0.00000347, -0.00000083, 0.00000073, -0.00000191, 0.00000136, -0.00000523, 0.00001358, 0.00001559, -0.00001454, -0.00000359, 0.00000427, -0.00000189, -0.00000423, -0.00000356, 0.00001000, 0.00000277, -0.00000269, 0.00000296, -0.00000450, -0.00000372, 0.00001042, -0.00001105, -0.00000920, 0.00003423, -0.00002284, -0.00001439, 0.00003829, -0.00003446, -0.00000000, 0.00001186, -0.00002271, 0.00002335, -0.00001057, 0.00000347, 0.00000121, -0.00001109, 0.00000966, -0.00002901, 0.00003305, -0.00001097, -0.00002339, 0.00004731, -0.00004431, 0.00001958, 0.00000420, -0.00003543, 0.00004550, -0.00004140, 0.00001767, 0.00000103, -0.00003931, 0.00005097, -0.00005045, 0.00001297, 0.00000160, -0.00004276, 0.00004237, -0.00009160, 0.00002376, -0.00010438, -0.00027864, 0.00040728, 0.00051585, -0.00011568, -0.00018576, -0.00001579, -0.00001708, -0.00002663, 0.00000118, -0.00000929, -0.00001384, 0.00001099, -0.00001657, 0.00000039, 0.00000797, -0.00000981, 0.00000406, -0.00001599, 0.00000105, 0.00000633, -0.00000678, 0.00000956, -0.00001459, 0.00000399, 0.00001838, -0.00001375, 0.00000516, 0.00000042, -0.00000905, 0.00001062, -0.00000960, -0.00001255, 0.00000123, -0.00000668, -0.00000429, -0.00000011, 0.00000719, 0.00000123, },
		{ 0.00000239, -0.00000258, -0.00000478, 0.00000160, -0.00000712, -0.00000169, 0.00000722, 0.00000484, -0.00000522, -0.00000913, 0.00000675, 0.00000491, -0.00000681, 0.00000426, 0.00000859, -0.00000109, -0.00000626, -0.00000261, 0.00000483, 0.00000592, 0.00000738, 0.00000315, -0.00000575, -0.00000276, 0.00000056, 0.00000265, 0.00000243, 0.00000149, 0.00000140, -0.00000443, 0.00000015, 0.00000037, -0.00000093, -0.00000026, -0.00000270, 0.00001817, 0.00000799, -0.00001572, 0.00000081, 0.00000285, -0.00000280, -0.00000504, 0.00000033, 0.00001023, -0.00000099, -0.00000035, 0.00000217, -0.00000729, 0.00000221, 0.00000762, -0.00001863, 0.00000761, 0.00002565, -0.00003631, 0.00001020, 0.00002222, -0.00003904, 0.00001646, -0.00000481, -0.00001055, 0.00001929, -0.00001494, 0.00001259, -0.00001277, 0.00000234, -0.00000097, -0.00002361, 0.00004297, -0.00003547, 0.00000506, 0.00003140, -0.00005032, 0.00004242, -0.00002607, -0.00000966, 0.00003546, -0.00005101, 0.00004177, -0.00003129, -0.00001080, 0.00004130, -0.00006517, 0.00004307, -0.00003364, -0.00001466, 0.00002717, -0.00010676, 0.00006309, -0.00020820, -0.00019437, 0.00059284, 0.00037764, -0.00023567, -0.00012353, -0.00000039, -0.00003115, -0.00001454, 0.00000038, -0.00001552, -0.00000398, 0.00000511, -0.00001795, 0.00000836, 0.00000079, -0.00000604, 0.00000095, -0.00001727, 0.00000855, -0.00000025, -0.00000249, 0.00000710, -0.00001825, 0.00001445, 0.00001121, -0.00001452, 0.00001008, -0.00000572, -0.00000460, 0.00000970, -0.00001464, -0.00000807, 0.00000046, -0.00000759, -0.00000303, 0.00000146, 0.00000824, -0.00000499, -0.00001426, },
		{ -0.00000615, -0.00000260, 0.00000057, -0.00000790, 0.00000157, 0.00000747, 0.00000297, -0.00000768, -0.00000623, 0.00000915, 0.00000063, -0.00000545, 0.00000752, 0.00000597, -0.00000281, -0.00000612, -0.00000123, 0.00000644, 0.00000629, 0.00000692, 0.00000080, -0.00000614, -0.00000150, 0.00000098, 0.00000284, 0.00000201, 0.00000205, 0.00000018, -0.00000496, 0.00000127, 0.00000016, -0.00000055, -0.00000153, 0.00000156, 0.00001979, 0.00000006, -0.00001409, 0.00000397, 0.00000097, -0.00000372, -0.00000466, 0.00000435, 0.00000811, -0.00000312, 0.00000248, -0.00000089, -0.00000723, 0.00000805, -0.00000043, -0.00001833, 0.00002253, 0.00000723, -0.00003541, 0.00002862, -0.00000078, -0.00002994, 0.00002212, -0.00001728, 0.00000384, 0.00000784, -0.00000943, 0.00001215, -0.00002085, 0.00001631, -0.00001829, -0.00000576, 0.00003787, -0.00004972, 0.00003489, 0.00000116, -0.00003617, 0.00004891, -0.00005051, 0.00002336, 0.00000785, -0.00004013, 0.00005024, -0.00005894, 0.00002784, 0.00000970, -0.00005650, 0.00005834, -0.00006821, 0.00002752, -0.00001481, -0.00009047, 0.00007576, -0.00032166, -0.00002385, 0.00071952, 0.00017253, -0.00028952, -0.00004926, -0.00001099, -0.00003386, -0.00000269, -0.00000711, -0.00001454, 0.00000274, -0.00000333, -0.00001344, 0.00001157, -0.00000507, -0.00000134, -0.00000512, -0.00001359, 0.00001230, -0.00000639, 0.00000378, 0.00000043, -0.00001671, 0.00002291, 0.00000126, -0.00001083, 0.00001171, -0.00001094, 0.00000129, 0.00000567, -0.00001719, -0.00000358, -0.00000157, -0.00000748, -0.00000192, 0.00000365, 0.00000711, -0.00001107, -0.00001012, 0.00000481, },
		{ 0.00000033, -0.00000168, -0.00000727, 0.00000461, 0.00000683, 0.00000076, -0.00000927, -0.00000224, 0.00000946, -0.00000324, -0.00000220, 0.00000891, 0.00000297, -0.00000361, -0.00000608, 0.00000072, 0.00000786, 0.00000613, 0.00000612, -0.00000147, -0.00000583, -0.00000034, 0.00000122, 0.00000293, 0.00000181, 0.00000245, -0.00000151, -0.00000487, 0.00000249, -0.00000020, -0.00000067, -0.00000171, 0.00000644, 0.00001827, -0.00000671, -0.00001049, 0.00000561, -0.00000107, -0.00000443, -0.00000289, 0.00000732, 0.00000447, -0.00000291, 0.00000417, -0.00000502, -0.00000357, 0.00001055, -0.00001028, -0.00000955, 0.00002912, -0.00001348, -0.00002149, 0.00003420, -0.00002128, -0.00001280, 0.00001646, -0.00002059, 0.00001352, -0.00000443, 0.00000199, 0.00000179, -0.00001792, 0.00002276, -0.00003419, 0.00001924, 0.00001731, -0.00004629, 0.00005395, -0.00003232, -0.00000599, 0.00003427, -0.00005753, 0.00005011, -0.00002764, -0.00001052, 0.00003598, -0.00006814, 0.00006150, -0.00003456, -0.00002306, 0.00004728, -0.00008506, 0.00006756, -0.00007517, -0.00003731, 0.00003745, -0.00040651, 0.00022149, 0.00074357, -0.00005726, -0.00026710, 0.00000827, -0.00003630, -0.00002185, 0.00000099, -0.00001502, -0.00000782, 0.00000337, -0.00000975, -0.00000597, 0.00000992, -0.00000751, 0.00000148, -0.00001142, -0.00000619, 0.00001110, -0.00000938, 0.00000920, -0.00000859, -0.00000948, 0.00002659, -0.00000858, -0.00000385, 0.00000940, -0.00001350, 0.00000673, -0.00000050, -0.00001675, -0.00000029, -0.00000403, -0.00000659, -0.00000079, 0.00000602, 0.00000343, -0.00001520, -0.00000390, 0.00000405, -0.00000877, },
		{ -0.00000468, -0.00000505, 0.00000683, 0.00000548, -0.00000157, -0.00000963, 0.00000202, 0.00000766, -0.00000557, 0.00000187, 0.00000823, 0.00000054, -0.00000398, -0.00000599, 0.00000335, 0.00000861, 0.00000561, 0.00000507, -0.00000351, -0.00000489, 0.00000050, 0.00000138, 0.00000298, 0.00000183, 0.00000251, -0.00000347, -0.00000395, 0.00000359, -0.00000095, -0.00000086, -0.00000056, 0.00001071, 0.00001420, -0.00001135, -0.00000594, 0.00000573, -0.00000306, -0.00000453, -0.00000003, 0.00000832, 0.00000071, -0.00000068, 0.00000348, -0.00000812, 0.00000258, 0.00000777, -0.00001729, 0.00000446, 0.00002447, -0.00002828, -0.00000102, 0.00002598, -0.00003188, 0.00000426, 0.00000397, -0.00001463, 0.00001447, -0.00001096, 0.00001228, -0.00001361, -0.00000418, 0.00001670, -0.00004001, 0.00004181, -0.00001228, -0.00002450, 0.00005333, -0.00005551, 0.00002874, 0.00000254, -0.00004192, 0.00005811, -0.00005646, 0.00002720, 0.00000124, -0.00005092, 0.00007419, -0.00007448, 0.00002515, 0.00000816, -0.00007089, 0.00008462, -0.00013305, 0.00004183, -0.00006468, -0.00041693, 0.00050207, 0.00064121, -0.00025776, -0.00018224, 0.00002878, -0.00005851, -0.00000082, -0.00000647, -0.00001691, -0.00000032, -0.00000128, -0.00001125, 0.00000064, 0.00000558, -0.00000656, 0.00000082, -0.00001505, 0.00000212, 0.00000585, -0.00000794, 0.00001103, -0.00001681, 0.00000190, 0.00002415, -0.00001542, 0.00000411, 0.00000387, -0.00001255, 0.00000996, -0.00000713, -0.00001378, 0.00000113, -0.00000610, -0.00000533, 0.00000069, 0.00000768, -0.00000235, -0.00001592, 0.00000242, 0.00000036, -0.00000919, 0.00000284, },
		{ -0.00000151, 0.00000774, 0.00000375, -0.00000377, -0.00000859, 0.00000563, 0.00000436, -0.00000574, 0.00000547, 0.00000598, -0.00000082, -0.00000449, -0.00000538, 0.00000639, 0.00000842, 0.00000502, 0.00000374, -0.00000510, -0.00000349, 0.00000093, 0.00000161, 0.00000299, 0.00000202, 0.00000201, -0.00000539, -0.00000212, 0.00000415, -0.00000203, -0.00000057, 0.00000164, 0.00001346, 0.00000861, -0.00001348, -0.00000139, 0.00000455, -0.00000477, -0.00000366, 0.00000310, 0.00000713, -0.00000174, 0.00000231, 0.00000033, -0.00000828, 0.00000858, 0.00000020, -0.00001773, 0.00001794, 0.00001040, -0.00003173, 0.00001739, 0.00000858, -0.00003008, 0.00001435, -0.00000852, -0.00000367, 0.00000733, -0.00000848, 0.00001486, -0.00002618, 0.00001464, -0.00000110, -0.00003071, 0.00005210, -0.00003979, 0.00000853, 0.00003163, -0.00005815, 0.00005358, -0.00003429, -0.00000795, 0.00004174, -0.00006515, 0.00005713, -0.00004185, -0.00000990, 0.00005634, -0.00009165, 0.00006912, -0.00004774, -0.00002378, 0.00006225, -0.00016127, 0.00011841, -0.00022101, -0.00031504, 0.00075561, 0.00042071, -0.00037977, -0.00006878, 0.00000847, -0.00006191, 0.00001754, -0.00002122, -0.00000996, 0.00000249, -0.00000728, -0.00000804, 0.00000377, 0.00000152, -0.00000416, -0.00000287, -0.00001437, 0.00000822, -0.00000087, -0.00000283, 0.00000791, -0.00002103, 0.00001432, 0.00001623, -0.00001740, 0.00001049, -0.00000308, -0.00000855, 0.00000997, -0.00001249, -0.00000943, 0.00000070, -0.00000728, -0.00000400, 0.00000273, 0.00000760, -0.00000888, -0.00001275, 0.00000667, -0.00000492, -0.00000695, 0.00000375, -0.00000746, },
		{ 0.00000719, 0.00000202, -0.00000562, -0.00000629, 0.00000786, 0.00000055, -0.00000382, 0.00000754, 0.00000313, -0.00000119, -0.00000543, -0.00000375, 0.00000924, 0.00000729, 0.00000461, 0.00000209, -0.00000603, -0.00000194, 0.00000099, 0.00000200, 0.00000291, 0.00000233, 0.00000081, -0.00000681, 0.00000047, 0.00000377, -0.00000309, 0.00000054, 0.00000423, 0.00001429, 0.00000263, -0.00001324, 0.00000243, 0.00000238, -0.00000586, -0.00000169, 0.00000546, 0.00000437, -0.00000201, 0.00000434, -0.00000409, -0.00000476, 0.00001143, -0.00000926, -0.00001064, 0.00002514, -0.00000737, -0.00002329, 0.00002666, -0.00001023, -0.00001869, 0.00001460, -0.00001515, 0.00000604, -0.00000314, 0.00000141, 0.00000692, -0.00002879, 0.00002976, -0.00002402, -0.00000762, 0.00004450, -0.00005388, 0.00004032, -0.00000386, -0.00003790, 0.00005727, -0.00006095, 0.00003166, 0.00000548, -0.00004769, 0.00006471, -0.00007542, 0.00004109, 0.00001035, -0.00007464, 0.00008790, -0.00009863, 0.00004284, -0.00000222, -0.00013752, 0.00015319, -0.00039464, -0.00008812, 0.00091278, 0.00012507, -0.00039562, 0.00003109, -0.00003857, -0.00004064, 0.00002172, -0.00003428, 0.00000308, -0.00000210, -0.00000991, -0.00000310, 0.00000334, -0.00000028, -0.00000279, -0.00000744, -0.00000979, 0.00001016, -0.00000606, 0.00000349, 0.00000055, -0.00001929, 0.00002420, 0.00000517, -0.00001433, 0.00001340, -0.00000925, -0.00000294, 0.00000689, -0.00001536, -0.00000513, -0.00000095, -0.00000753, -0.00000265, 0.00000516, 0.00000501, -0.00001423, -0.00000650, 0.00000729, -0.00000973, -0.00000272, 0.00000240, -0.00000888, 0.00000258, },
		{ 0.00000056, -0.00000694, -0.00000310, 0.00000829, -0.00000266, -0.00000052, 0.00000759, 0.00000072, -0.00000110, -0.00000667, -0.00000084, 0.00001120, 0.00000557, 0.00000443, 0.00000016, -0.00000617, -0.00000058, 0.00000086, 0.00000253, 0.00000276, 0.00000262, -0.00000111, -0.00000722, 0.00000333, 0.00000224, -0.00000353, 0.00000241, 0.00000649, 0.00001329, -0.00000279, -0.00001111, 0.00000503, -0.00000042, -0.00000595, 0.00000107, 0.00000622, 0.00000125, -0.00000024, 0.00000411, -0.00000772, 0.00000143, 0.00000925, -0.00001648, 0.00000158, 0.00002289, -0.00002177, -0.00000729, 0.00002407, -0.00002301, -0.00000408, 0.00000693, -0.00001369, 0.00000949, -0.00001064, 0.00001303, -0.00000903, -0.00001858, 0.00003332, -0.00004227, 0.00002164, 0.00002051, -0.00004801, 0.00005799, -0.00003963, -0.00000192, 0.00003685, -0.00006533, 0.00006057, -0.00003703, -0.00000900, 0.00004358, -0.00008316, 0.00008157, -0.00004850, -0.00002538, 0.00006810, -0.00012024, 0.00010283, -0.00009247, -0.00005666, 0.00011034, -0.00052691, 0.00023934, 0.00091876, -0.00017733, -0.00030959, 0.00008236, -0.00008667, -0.00000180, 0.00000675, -0.00003630, 0.00001501, -0.00001215, -0.00000659, -0.00000026, 0.00000153, 0.00000007, -0.00000394, -0.00001025, -0.00000355, 0.00000800, -0.00000769, 0.00000806, -0.00000853, -0.00001168, 0.00002881, -0.00000585, -0.00000763, 0.00001231, -0.00001297, 0.00000239, 0.00000184, -0.00001550, -0.00000195, -0.00000297, -0.00000714, -0.00000099, 0.00000728, -0.00000007, -0.00001652, 0.00000085, 0.00000398, -0.00001210, 0.00000198, -0.00000102, -0.00000808, 0.00000613, 0.00000549, },
		{ -0.00000759, 0.00000043, 0.00000699, -0.00000441, 0.00000305, 0.00000583, -0.00000055, -0.00000128, -0.00000762, 0.00000315, 0.00001169, 0.00000380, 0.00000433, -0.00000191, -0.00000553, 0.00000035, 0.00000081, 0.00000309, 0.00000259, 0.00000268, -0.00000355, -0.00000623, 0.00000575, -0.00000024, -0.00000277, 0.00000461, 0.00000792, 0.00001094, -0.00000703, -0.00000771, 0.00000612, -0.00000336, -0.00000474, 0.00000387, 0.00000510, -0.00000083, 0.00000246, 0.00000136, -0.00000857, 0.00000781, 0.00000228, -0.00001795, 0.00001411, 0.00001194, -0.00002735, 0.00000923, 0.00001195, -0.00002576, 0.00000706, -0.00000347, -0.00000619, 0.00000532, -0.00001057, 0.00001945, -0.00002614, 0.00000145, 0.00002215, -0.00004721, 0.00004610, -0.00001143, -0.00002351, 0.00005374, -0.00006137, 0.00003569, -0.00000056, -0.00004406, 0.00006572, -0.00006824, 0.00003612, -0.00000054, -0.00005796, 0.00009262, -0.00009700, 0.00003949, 0.00001139, -0.00009670, 0.00012664, -0.00017580, 0.00006148, -0.00002541, -0.00055357, 0.00060460, 0.00075378, -0.00041215, -0.00015764, 0.00007024, -0.00010983, 0.00003747, -0.00002245, -0.00002284, 0.00001794, -0.00002182, 0.00000144, -0.00000177, 0.00000129, 0.00000057, -0.00000711, -0.00000974, 0.00000157, 0.00000362, -0.00000563, 0.00000873, -0.00001606, -0.00000029, 0.00002720, -0.00001398, 0.00000038, 0.00000799, -0.00001363, 0.00000594, -0.00000359, -0.00001352, -0.00000032, -0.00000471, -0.00000643, 0.00000125, 0.00000800, -0.00000660, -0.00001469, 0.00000676, -0.00000206, -0.00001092, 0.00000524, -0.00000535, -0.00000495, 0.00000812, 0.00000329, -0.00000056, },
		{ 0.00000368, 0.00000443, -0.00000430, 0.00000576, 0.00000305, -0.00000055, -0.00000226, -0.00000753, 0.00000758, 0.00001054, 0.00000253, 0.00000406, -0.00000383, -0.00000431, 0.00000076, 0.00000105, 0.00000350, 0.00000250, 0.00000224, -0.00000606, -0.00000374, 0.00000692, -0.00000301, -0.00000059, 0.00000643, 0.00000838, 0.00000777, -0.00000977, -0.00000369, 0.00000561, -0.00000582, -0.00000219, 0.00000577, 0.00000262, -0.00000100, 0.00000447, -0.00000290, -0.00000577, 0.00001147, -0.00000695, -0.00001248, 0.00002179, -0.00000333, -0.00002257, 0.00001957, -0.00000374, -0.00001914, 0.00001063, -0.00001092, 0.00000245, -0.00000368, -0.00000236, 0.00001589, -0.00003633, 0.00002352, -0.00000056, -0.00003523, 0.00005603, -0.00003919, 0.00001054, 0.00002843, -0.00005997, 0.00005955, -0.00004025, -0.00000451, 0.00004335, -0.00007410, 0.00006865, -0.00005108, -0.00000631, 0.00006556, -0.00011321, 0.00009384, -0.00006364, -0.00002958, 0.00009429, -0.00021401, 0.00017375, -0.00023570, -0.00042745, 0.00091971, 0.00044436, -0.00052257, 0.00000454, 0.00000563, -0.00009336, 0.00005792, -0.00005266, 0.00000257, 0.00000777, -0.00002434, 0.00000953, -0.00000672, 0.00000419, -0.00000125, -0.00001008, -0.00000632, 0.00000370, -0.00000035, -0.00000155, 0.00000521, -0.00001937, 0.00001167, 0.00002032, -0.00001764, 0.00000738, 0.00000206, -0.00001170, 0.00000707, -0.00000803, -0.00001052, -0.00000002, -0.00000595, -0.00000547, 0.00000408, 0.00000634, -0.00001271, -0.00000907, 0.00000908, -0.00000854, -0.00000649, 0.00000565, -0.00000890, -0.00000021, 0.00000799, 0.00000139, -0.00000148, -0.00000744, },
		{ 0.00000136, -0.00000249, 0.00000682, 0.00000029, 0.00000030, -0.00000413, -0.00000577, 0.00001148, 0.00000803, 0.00000209, 0.00000332, -0.00000530, -0.00000283, 0.00000078, 0.00000164, 0.00000363, 0.00000255, 0.00000104, -0.00000798, -0.00000014, 0.00000631, -0.00000510, 0.00000273, 0.00000729, 0.00000808, 0.00000426, -0.00001089, 0.00000028, 0.00000362, -0.00000713, 0.00000126, 0.00000597, -0.00000009, 0.00000071, 0.00000444, -0.00000672, -0.00000006, 0.00001041, -0.00001468, -0.00000182, 0.00002136, -0.00001685, -0.00001018, 0.00002020, -0.00001631, -0.00000763, 0.00000658, -0.00001203, 0.00000713, -0.00001215, 0.00001020, 0.00000219, -0.00003398, 0.00003838, -0.00002638, -0.00000960, 0.00004720, -0.00005193, 0.00004083, -0.00000846, -0.00003555, 0.00005952, -0.00006593, 0.00003778, 0.00000118, -0.00005050, 0.00007345, -0.00008688, 0.00005300, 0.00000711, -0.00008639, 0.00011243, -0.00012706, 0.00005950, 0.00000663, -0.00017958, 0.00022611, -0.00046610, -0.00014004, 0.00109603, 0.00006008, -0.00048835, 0.00012087, -0.00007967, -0.00004069, 0.00004754, -0.00006825, 0.00002923, -0.00001293, -0.00001592, 0.00001206, -0.00001138, 0.00000916, -0.00000652, -0.00001022, -0.00000221, 0.00000286, -0.00000202, 0.00000206, -0.00000097, -0.00001745, 0.00002113, 0.00001046, -0.00001679, 0.00001187, -0.00000383, -0.00000835, 0.00000610, -0.00001081, -0.00000752, -0.00000050, -0.00000681, -0.00000398, 0.00000687, 0.00000193, -0.00001627, -0.00000141, 0.00000690, -0.00001291, -0.00000042, 0.00000285, -0.00001011, 0.00000471, 0.00000595, 0.00000037, -0.00000246, -0.00000640, 0.00000609, },
		{ 0.00000034, 0.00000598, -0.00000151, 0.00000118, -0.00000641, -0.00000217, 0.00001388, 0.00000486, 0.00000247, 0.00000197, -0.00000606, -0.00000144, 0.00000065, 0.00000246, 0.00000348, 0.00000266, -0.00000099, -0.00000863, 0.00000378, 0.00000387, -0.00000559, 0.00000641, 0.00000691, 0.00000739, 0.00000072, -0.00001042, 0.00000356, 0.00000055, -0.00000674, 0.00000473, 0.00000428, -0.00000176, 0.00000332, 0.00000199, -0.00000813, 0.00000632, 0.00000456, -0.00001748, 0.00000993, 0.00001290, -0.00002349, 0.00000415, 0.00001190, -0.00002118, 0.00000292, -0.00000160, -0.00000704, 0.00000517, -0.00001493, 0.00002073, -0.00001691, -0.00001864, 0.00003936, -0.00004513, 0.00002071, 0.00002279, -0.00004473, 0.00005530, -0.00004255, 0.00000237, 0.00003512, -0.00006660, 0.00006541, -0.00004424, -0.00000573, 0.00004622, -0.00009105, 0.00009576, -0.00006179, -0.00002313, 0.00008225, -0.00014939, 0.00013496, -0.00011068, -0.00007059, 0.00017404, -0.00063769, 0.00026703, 0.00107279, -0.00030487, -0.00033319, 0.00015487, -0.00014608, 0.00002764, 0.00000789, -0.00005881, 0.00004396, -0.00003533, 0.00000163, 0.00000577, -0.00001120, 0.00001273, -0.00001384, -0.00000631, -0.00000008, 0.00000085, -0.00000120, 0.00000325, -0.00000725, -0.00001117, 0.00002620, 0.00000022, -0.00001254, 0.00001344, -0.00000860, -0.00000474, 0.00000390, -0.00001201, -0.00000502, -0.00000129, -0.00000748, -0.00000158, 0.00000852, -0.00000448, -0.00001575, 0.00000561, 0.00000101, -0.00001335, 0.00000492, -0.00000216, -0.00000818, 0.00000820, 0.00000287, 0.00000037, -0.00000369, -0.00000444, 0.00000724, -0.00000138, },
		{ 0.00000365, -0.00000184, 0.00000128, -0.00000824, 0.00000285, 0.00001414, 0.00000191, 0.00000337, 0.00000009, -0.00000604, -0.00000038, 0.00000066, 0.00000329, 0.00000317, 0.00000263, -0.00000362, -0.00000754, 0.00000696, 0.00000020, -0.00000391, 0.00000941, 0.00000553, 0.00000660, -0.00000263, -0.00000851, 0.00000557, -0.00000290, -0.00000448, 0.00000713, 0.00000124, -0.00000151, 0.00000533, -0.00000206, -0.00000607, 0.00001058, -0.00000403, -0.00001387, 0.00001805, -0.00000023, -0.00002113, 0.00001440, -0.00000096, -0.00001760, 0.00000799, -0.00000899, 0.00000063, -0.00000248, -0.00000966, 0.00002334, -0.00003378, 0.00000461, 0.00002526, -0.00004915, 0.00004472, -0.00000785, -0.00002071, 0.00004819, -0.00006047, 0.00003866, -0.00000406, -0.00004143, 0.00006648, -0.00007425, 0.00004242, -0.00000375, -0.00005885, 0.00010243, -0.00011358, 0.00005354, 0.00000991, -0.00011479, 0.00016083, -0.00021364, 0.00008236, 0.00000371, -0.00067093, 0.00070657, 0.00083992, -0.00055986, -0.00011707, 0.00010234, -0.00016267, 0.00008271, -0.00004554, -0.00002488, 0.00003754, -0.00004813, 0.00002106, -0.00000792, -0.00000369, 0.00001091, -0.00001964, 0.00000045, -0.00000147, 0.00000012, 0.00000056, 0.00000165, -0.00001136, -0.00000262, 0.00002657, -0.00000843, -0.00000643, 0.00001247, -0.00001193, -0.00000153, 0.00000133, -0.00001218, -0.00000304, -0.00000230, -0.00000786, 0.00000182, 0.00000780, -0.00001112, -0.00001100, 0.00000940, -0.00000630, -0.00000961, 0.00000737, -0.00000741, -0.00000352, 0.00000912, 0.00000002, 0.00000101, -0.00000511, -0.00000168, 0.00000694, -0.00000308, 0.00000326, },
		{ -0.00000081, 0.00000012, -0.00000865, 0.00000832, 0.00001215, -0.00000003, 0.00000425, -0.00000209, -0.00000531, 0.00000025, 0.00000101, 0.00000389, 0.00000286, 0.00000215, -0.00000633, -0.00000469, 0.00000841, -0.00000359, -0.00000017, 0.00001080, 0.00000373, 0.00000575, -0.00000553, -0.00000547, 0.00000588, -0.00000580, -0.00000069, 0.00000756, -0.00000197, 0.00000064, 0.00000539, -0.00000592, -0.00000104, 0.00001062, -0.00001193, -0.00000502, 0.00001926, -0.00001280, -0.00001145, 0.00001673, -0.00001257, -0.00000848, 0.00000625, -0.00001173, 0.00000638, -0.00001178, 0.00000197, 0.00001540, -0.00004119, 0.00002721, 0.00000098, -0.00003651, 0.00005393, -0.00003307, 0.00001029, 0.00002247, -0.00005538, 0.00005886, -0.00004222, -0.00000034, 0.00003987, -0.00007601, 0.00007382, -0.00005677, -0.00000076, 0.00006697, -0.00012536, 0.00011241, -0.00007885, -0.00003009, 0.00011662, -0.00025577, 0.00022199, -0.00025488, -0.00051457, 0.00106747, 0.00044586, -0.00064646, 0.00008436, -0.00000813, -0.00011990, 0.00009987, -0.00008955, 0.00002128, 0.00000980, -0.00004303, 0.00003274, -0.00002290, 0.00000956, 0.00000173, -0.00001989, 0.00000664, -0.00000556, 0.00000213, 0.00000105, -0.00000145, -0.00001233, 0.00000593, 0.00002315, -0.00001451, 0.00000027, 0.00000957, -0.00001396, 0.00000123, -0.00000125, -0.00001179, -0.00000131, -0.00000372, -0.00000746, 0.00000563, 0.00000406, -0.00001569, -0.00000346, 0.00000849, -0.00001213, -0.00000310, 0.00000590, -0.00001069, 0.00000237, 0.00000725, -0.00000148, 0.00000153, -0.00000636, 0.00000152, 0.00000532, -0.00000333, 0.00000534, 0.00000064, },
		{ -0.00000224, -0.00000695, 0.00001300, 0.00000844, -0.00000051, 0.00000458, -0.00000417, -0.00000411, 0.00000053, 0.00000170, 0.00000411, 0.00000266, 0.00000095, -0.00000838, -0.00000062, 0.00000760, -0.00000617, 0.00000482, 0.00001009, 0.00000225, 0.00000469, -0.00000767, -0.00000181, 0.00000443, -0.00000721, 0.00000366, 0.00000569, -0.00000400, 0.00000375, 0.00000304, -0.00000768, 0.00000504, 0.00000604, -0.00001584, 0.00000563, 0.00001306, -0.00001999, 0.00000098, 0.00001093, -0.00001833, 0.00000129, -0.00000029, -0.00000876, 0.00000679, -0.00001775, 0.00001503, -0.00000111, -0.00003550, 0.00004067, -0.00002456, -0.00001172, 0.00004573, -0.00004375, 0.00003582, -0.00001149, -0.00002979, 0.00005520, -0.00006402, 0.00004002, -0.00000404, -0.00004816, 0.00007466, -0.00009059, 0.00006093, 0.00000066, -0.00008918, 0.00012680, -0.00014818, 0.00007504, 0.00000887, -0.00020808, 0.00028279, -0.00052961, -0.00016917, 0.00124747, -0.00001348, -0.00055723, 0.00020590, -0.00012714, -0.00003363, 0.00006988, -0.00010336, 0.00006029, -0.00002926, -0.00001934, 0.00002941, -0.00003102, 0.00002293, -0.00001299, -0.00001256, 0.00000830, -0.00000942, 0.00000627, -0.00000116, -0.00000396, -0.00001071, 0.00001305, 0.00001722, -0.00001781, 0.00000679, 0.00000513, -0.00001481, 0.00000376, -0.00000395, -0.00001089, 0.00000020, -0.00000570, -0.00000565, 0.00000862, -0.00000225, -0.00001629, 0.00000417, 0.00000321, -0.00001408, 0.00000359, 0.00000108, -0.00001050, 0.00000744, 0.00000346, -0.00000112, 0.00000118, -0.00000684, 0.00000459, 0.00000280, -0.00000216, 0.00000593, -0.00000204, 0.00000097, },
		{ -0.00000303, 0.00001572, 0.00000399, 0.00000044, 0.00000396, -0.00000576, -0.00000272, 0.00000068, 0.00000262, 0.00000395, 0.00000255, -0.00000102, -0.00000906, 0.00000369, 0.00000468, -0.00000647, 0.00000973, 0.00000749, 0.00000162, 0.00000315, -0.00000867, 0.00000174, 0.00000158, -0.00000653, 0.00000731, 0.00000203, -0.00000388, 0.00000631, -0.00000102, -0.00000618, 0.00000956, -0.00000157, -0.00001401, 0.00001382, 0.00000203, -0.00001940, 0.00001080, 0.00000018, -0.00001661, 0.00000728, -0.00000765, -0.00000206, 0.00000129, -0.00001686, 0.00002357, -0.00002038, -0.00001821, 0.00004006, -0.00004207, 0.00001626, 0.00002427, -0.00003694, 0.00004603, -0.00004023, 0.00000567, 0.00002971, -0.00006091, 0.00006324, -0.00004757, -0.00000170, 0.00004359, -0.00008996, 0.00010089, -0.00007150, -0.00001672, 0.00008655, -0.00016664, 0.00015806, -0.00012749, -0.00007477, 0.00021645, -0.00072482, 0.00030446, 0.00118660, -0.00042247, -0.00033754, 0.00021489, -0.00020430, 0.00006158, 0.00000358, -0.00007822, 0.00007414, -0.00006312, 0.00001474, 0.00001026, -0.00002617, 0.00002908, -0.00002768, 0.00000076, 0.00000322, -0.00000948, 0.00000994, -0.00000563, -0.00000430, -0.00000796, 0.00001838, 0.00000974, -0.00001835, 0.00001261, -0.00000077, -0.00001420, 0.00000603, -0.00000704, -0.00000918, 0.00000114, -0.00000795, -0.00000211, 0.00000925, -0.00000944, -0.00001233, 0.00000900, -0.00000432, -0.00001132, 0.00000776, -0.00000511, -0.00000668, 0.00000988, -0.00000067, 0.00000077, -0.00000041, -0.00000603, 0.00000684, 0.00000004, 0.00000004, 0.00000483, -0.00000354, 0.00000258, -0.00000515, },
		{ 0.00001576, -0.00000001, 0.00000236, 0.00000231, -0.00000657, -0.00000141, 0.00000089, 0.00000351, 0.00000353, 0.00000236, -0.00000354, -0.00000796, 0.00000708, 0.00000051, -0.00000405, 0.00001316, 0.00000378, 0.00000197, 0.00000099, -0.00000824, 0.00000438, -0.00000185, -0.00000372, 0.00000903, -0.00000219, -0.00000151, 0.00000691, -0.00000514, -0.00000169, 0.00001039, -0.00000920, -0.00000707, 0.00001626, -0.00000938, -0.00001190, 0.00001414, -0.00001059, -0.00000915, 0.00000733, -0.00001194, 0.00000460, -0.00000768, -0.00000878, 0.00002332, -0.00003525, 0.00000470, 0.00002586, -0.00004560, 0.00003778, -0.00000158, -0.00001680, 0.00003763, -0.00005285, 0.00003674, -0.00000677, -0.00003472, 0.00006002, -0.00007311, 0.00004443, -0.00000727, -0.00005346, 0.00010141, -0.00012049, 0.00006428, 0.00000382, -0.00012125, 0.00018036, -0.00023995, 0.00010249, 0.00001493, -0.00075192, 0.00079699, 0.00088930, -0.00068108, -0.00006952, 0.00012117, -0.00020801, 0.00012700, -0.00007214, -0.00002201, 0.00005456, -0.00007588, 0.00004511, -0.00001786, -0.00000787, 0.00002271, -0.00003527, 0.00001483, -0.00000739, -0.00000367, 0.00000987, -0.00001004, -0.00000248, -0.00000533, 0.00002224, 0.00000115, -0.00001595, 0.00001699, -0.00000792, -0.00001151, 0.00000746, -0.00001045, -0.00000631, 0.00000083, -0.00000964, 0.00000269, 0.00000648, -0.00001503, -0.00000498, 0.00000908, -0.00001098, -0.00000493, 0.00000774, -0.00001006, -0.00000052, 0.00000891, -0.00000344, 0.00000313, -0.00000299, -0.00000368, 0.00000765, -0.00000222, 0.00000254, 0.00000237, -0.00000346, 0.00000317, -0.00000757, 0.00000253, },
		{ -0.00000252, 0.00000445, -0.00000015, -0.00000649, -0.00000035, 0.00000133, 0.00000414, 0.00000303, 0.00000182, -0.00000610, -0.00000512, 0.00000857, -0.00000355, 0.00000068, 0.00001410, 0.00000012, 0.00000297, -0.00000165, -0.00000634, 0.00000549, -0.00000478, 0.00000055, 0.00000817, -0.00000546, 0.00000224, 0.00000494, -0.00000741, 0.00000414, 0.00000685, -0.00001372, 0.00000218, 0.00001204, -0.00001671, -0.00000115, 0.00001013, -0.00001693, 0.00000014, 0.00000225, -0.00001123, 0.00000762, -0.00001591, 0.00000353, 0.00001362, -0.00004015, 0.00002524, 0.00000356, -0.00003464, 0.00004631, -0.00002209, 0.00000755, 0.00001494, -0.00004537, 0.00005154, -0.00003933, 0.00000332, 0.00003205, -0.00007053, 0.00007134, -0.00005715, 0.00000524, 0.00006036, -0.00012535, 0.00012070, -0.00009025, -0.00002499, 0.00012440, -0.00027842, 0.00025545, -0.00027865, -0.00056213, 0.00118050, 0.00042653, -0.00073585, 0.00015648, -0.00002890, -0.00013734, 0.00013590, -0.00012548, 0.00004334, 0.00000746, -0.00005968, 0.00005781, -0.00004280, 0.00001759, 0.00000369, -0.00003060, 0.00002291, -0.00001889, 0.00000698, 0.00000409, -0.00001135, -0.00000020, -0.00000290, 0.00002473, -0.00000820, -0.00001021, 0.00001872, -0.00001549, -0.00000636, 0.00000705, -0.00001344, -0.00000245, -0.00000123, -0.00000959, 0.00000732, 0.00000042, -0.00001673, 0.00000310, 0.00000434, -0.00001396, 0.00000247, 0.00000356, -0.00001164, 0.00000572, 0.00000507, -0.00000378, 0.00000465, -0.00000570, -0.00000010, 0.00000674, -0.00000332, 0.00000452, -0.00000072, -0.00000193, 0.00000228, -0.00000842, 0.00000859, 0.00001309, },
		{ 0.00000587, -0.00000295, -0.00000560, 0.00000041, 0.00000202, 0.00000437, 0.00000260, 0.00000072, -0.00000801, -0.00000115, 0.00000774, -0.00000610, 0.00000657, 0.00001222, -0.00000243, 0.00000393, -0.00000430, -0.00000333, 0.00000479, -0.00000619, 0.00000505, 0.00000488, -0.00000653, 0.00000581, 0.00000094, -0.00000655, 0.00000887, 0.00000015, -0.00001322, 0.00000992, 0.00000303, -0.00001737, 0.00000812, 0.00000100, -0.00001668, 0.00000706, -0.00000494, -0.00000619, 0.00000530, -0.00001954, 0.00001514, -0.00000238, -0.00003337, 0.00003664, -0.00001880, -0.00001392, 0.00004071, -0.00003063, 0.00002602, -0.00001226, -0.00002197, 0.00004533, -0.00005530, 0.00003746, -0.00000890, -0.00004141, 0.00006816, -0.00008534, 0.00006273, -0.00000729, -0.00008247, 0.00012779, -0.00015746, 0.00008631, 0.00000364, -0.00021684, 0.00031286, -0.00057732, -0.00016908, 0.00134748, -0.00008386, -0.00059554, 0.00027236, -0.00017161, -0.00002138, 0.00008480, -0.00013324, 0.00009084, -0.00004880, -0.00001889, 0.00004559, -0.00005246, 0.00003930, -0.00002186, -0.00001346, 0.00001989, -0.00002507, 0.00001822, -0.00000649, -0.00000761, 0.00000018, 0.00000046, 0.00002512, -0.00001741, -0.00000115, 0.00001637, -0.00002181, 0.00000082, 0.00000393, -0.00001478, 0.00000148, -0.00000483, -0.00000698, 0.00000986, -0.00000732, -0.00001350, 0.00000873, -0.00000327, -0.00001193, 0.00000782, -0.00000295, -0.00000906, 0.00000973, -0.00000001, -0.00000165, 0.00000433, -0.00000742, 0.00000389, 0.00000430, -0.00000296, 0.00000528, -0.00000351, 0.00000041, 0.00000000, -0.00000698, 0.00001368, 0.00000840, -0.00000297, },
		{ -0.00000548, -0.00000413, 0.00000095, 0.00000285, 0.00000418, 0.00000226, -0.00000101, -0.00000866, 0.00000297, 0.00000486, -0.00000610, 0.00001201, 0.00000805, -0.00000322, 0.00000416, -0.00000630, 0.00000010, 0.00000258, -0.00000550, 0.00000837, 0.00000020, -0.00000497, 0.00000764, -0.00000357, -0.00000258, 0.00001034, -0.00000710, -0.00000789, 0.00001300, -0.00000702, -0.00001163, 0.00001211, -0.00000908, -0.00001064, 0.00000899, -0.00001068, 0.00000046, -0.00000145, -0.00001673, 0.00002139, -0.00001908, -0.00001747, 0.00003584, -0.00003385, 0.00000894, 0.00002509, -0.00002607, 0.00003184, -0.00003308, 0.00000711, 0.00002207, -0.00004940, 0.00005415, -0.00004610, 0.00000183, 0.00003664, -0.00007998, 0.00009544, -0.00007505, -0.00000784, 0.00008009, -0.00016823, 0.00016717, -0.00013964, -0.00006760, 0.00022876, -0.00077535, 0.00034809, 0.00124574, -0.00051271, -0.00032592, 0.00025392, -0.00025073, 0.00009331, -0.00000515, -0.00009090, 0.00009994, -0.00009077, 0.00003099, 0.00001179, -0.00004054, 0.00004647, -0.00004362, 0.00001048, 0.00000530, -0.00002116, 0.00002424, -0.00001795, 0.00000060, -0.00000287, 0.00000627, 0.00002187, -0.00002455, 0.00001015, 0.00000902, -0.00002463, 0.00000845, -0.00000186, -0.00001337, 0.00000401, -0.00000882, -0.00000196, 0.00000867, -0.00001408, -0.00000623, 0.00000954, -0.00001052, -0.00000564, 0.00000891, -0.00000901, -0.00000324, 0.00001010, -0.00000424, 0.00000193, 0.00000197, -0.00000721, 0.00000717, 0.00000101, -0.00000126, 0.00000448, -0.00000519, 0.00000265, -0.00000296, -0.00000314, 0.00001650, 0.00000294, -0.00000141, 0.00000596, },
		{ -0.00000242, 0.00000137, 0.00000365, 0.00000367, 0.00000191, -0.00000316, -0.00000773, 0.00000618, 0.00000090, -0.00000327, 0.00001548, 0.00000277, -0.00000224, 0.00000324, -0.00000704, 0.00000308, -0.00000039, -0.00000278, 0.00000943, -0.00000433, -0.00000137, 0.00000682, -0.00000663, 0.00000309, 0.00000767, -0.00001195, 0.00000001, 0.00001021, -0.00001405, -0.00000239, 0.00000950, -0.00001592, -0.00000187, 0.00000576, -0.00001251, 0.00000545, -0.00000974, -0.00000846, 0.00001983, -0.00003082, 0.00000192, 0.00002428, -0.00003757, 0.00002663, 0.00000654, -0.00001256, 0.00002403, -0.00004009, 0.00003021, -0.00000774, -0.00002552, 0.00004754, -0.00006493, 0.00004137, -0.00000972, -0.00004316, 0.00008966, -0.00011579, 0.00006888, -0.00000529, -0.00011478, 0.00018093, -0.00024919, 0.00011841, 0.00000462, -0.00078340, 0.00086394, 0.00089693, -0.00075907, -0.00002569, 0.00012588, -0.00023833, 0.00016199, -0.00009703, -0.00001482, 0.00006564, -0.00009996, 0.00006954, -0.00003037, -0.00001004, 0.00003362, -0.00005140, 0.00003147, -0.00001580, -0.00000660, 0.00002053, -0.00002481, 0.00000988, -0.00000863, 0.00001506, 0.00001357, -0.00002696, 0.00002111, -0.00000278, -0.00002210, 0.00001405, -0.00000905, -0.00000897, 0.00000381, -0.00001143, 0.00000405, 0.00000339, -0.00001708, 0.00000242, 0.00000506, -0.00001419, 0.00000231, 0.00000531, -0.00001200, 0.00000357, 0.00000689, -0.00000596, 0.00000532, -0.00000173, -0.00000477, 0.00000872, -0.00000215, 0.00000119, 0.00000225, -0.00000530, 0.00000393, -0.00000553, 0.00000243, 0.00001630, -0.00000182, 0.00000150, 0.00000447, -0.00000718, },
		{ 0.00000182, 0.00000423, 0.00000302, 0.00000137, -0.00000533, -0.00000530, 0.00000763, -0.00000281, 0.00000179, 0.00001601, -0.00000217, -0.00000011, 0.00000121, -0.00000621, 0.00000484, -0.00000308, 0.00000119, 0.00000783, -0.00000721, 0.00000286, 0.00000355, -0.00000678, 0.00000817, 0.00000168, -0.00001240, 0.00000713, 0.00000283, -0.00001540, 0.00000620, 0.00000178, -0.00001706, 0.00000586, -0.00000058, -0.00001005, 0.00000652, -0.00001612, 0.00000206, 0.00001110, -0.00003403, 0.00001852, 0.00000685, -0.00003032, 0.00003485, -0.00000811, 0.00000279, 0.00000725, -0.00003201, 0.00003912, -0.00003184, 0.00000538, 0.00002156, -0.00005891, 0.00006157, -0.00005152, 0.00000992, 0.00004736, -0.00011301, 0.00011644, -0.00009489, -0.00001574, 0.00011596, -0.00027690, 0.00026743, -0.00030412, -0.00056173, 0.00124317, 0.00039144, -0.00078011, 0.00020768, -0.00005077, -0.00014374, 0.00015961, -0.00015356, 0.00006448, 0.00000103, -0.00007103, 0.00008014, -0.00006312, 0.00002724, 0.00000380, -0.00004002, 0.00003979, -0.00003441, 0.00001396, 0.00000644, -0.00002269, 0.00001541, -0.00001407, 0.00002529, 0.00000018, -0.00002239, 0.00002810, -0.00001659, -0.00001383, 0.00001526, -0.00001530, -0.00000267, 0.00000051, -0.00001102, 0.00000870, -0.00000463, -0.00001475, 0.00000896, -0.00000290, -0.00001251, 0.00000856, -0.00000138, -0.00001061, 0.00000877, 0.00000155, -0.00000460, 0.00000697, -0.00000539, -0.00000061, 0.00000801, -0.00000421, 0.00000356, -0.00000084, -0.00000387, 0.00000377, -0.00000663, 0.00000846, 0.00001311, -0.00000467, 0.00000470, 0.00000163, -0.00000770, -0.00000080, },
		{ 0.00000445, 0.00000237, 0.00000049, -0.00000700, -0.00000191, 0.00000703, -0.00000495, 0.00000778, 0.00001343, -0.00000553, 0.00000222, -0.00000143, -0.00000395, 0.00000497, -0.00000453, 0.00000521, 0.00000409, -0.00000754, 0.00000605, -0.00000088, -0.00000370, 0.00001045, -0.00000535, -0.00000828, 0.00001044, -0.00000597, -0.00001093, 0.00001052, -0.00000759, -0.00001254, 0.00000968, -0.00000704, -0.00000486, 0.00000331, -0.00001813, 0.00001095, -0.00000156, -0.00002840, 0.00002775, -0.00001036, -0.00001596, 0.00003337, -0.00001495, 0.00001328, -0.00001074, -0.00001373, 0.00003209, -0.00004141, 0.00003031, -0.00001226, -0.00003196, 0.00005545, -0.00007180, 0.00005741, -0.00001448, -0.00006789, 0.00011496, -0.00015236, 0.00009023, -0.00000756, -0.00020379, 0.00030973, -0.00060135, -0.00013896, 0.00138244, -0.00013896, -0.00060135, 0.00030973, -0.00020379, -0.00000756, 0.00009023, -0.00015236, 0.00011496, -0.00006789, -0.00001448, 0.00005741, -0.00007180, 0.00005545, -0.00003196, -0.00001226, 0.00003031, -0.00004141, 0.00003209, -0.00001373, -0.00001074, 0.00001328, -0.00001495, 0.00003337, -0.00001596, -0.00001036, 0.00002775, -0.00002840, -0.00000156, 0.00001095, -0.00001813, 0.00000331, -0.00000486, -0.00000704, 0.00000968, -0.00001254, -0.00000759, 0.00001052, -0.00001093, -0.00000597, 0.00001044, -0.00000828, -0.00000535, 0.00001045, -0.00000370, -0.00000088, 0.00000605, -0.00000754, 0.00000409, 0.00000521, -0.00000453, 0.00000497, -0.00000395, -0.00000143, 0.00000222, -0.00000553, 0.00001343, 0.00000778, -0.00000495, 0.00000703, -0.00000191, -0.00000700, 0.00000049, 0.00000237, },
		{ 0.00000182, -0.00000080, -0.00000770, 0.00000163, 0.00000470, -0.00000467, 0.00001311, 0.00000846, -0.00000663, 0.00000377, -0.00000387, -0.00000084, 0.00000356, -0.00000421, 0.00000801, -0.00000061, -0.00000539, 0.00000697, -0.00000460, 0.00000155, 0.00000877, -0.00001061, -0.00000138, 0.00000856, -0.00001251, -0.00000290, 0.00000896, -0.00001475, -0.00000463, 0.00000870, -0.00001102, 0.00000051, -0.00000267, -0.00001530, 0.00001526, -0.00001383, -0.00001659, 0.00002810, -0.00002239, 0.00000018, 0.00002529, -0.00001407, 0.00001541, -0.00002269, 0.00000644, 0.00001396, -0.00003441, 0.00003979, -0.00004002, 0.00000380, 0.00002724, -0.00006312, 0.00008014, -0.00007103, 0.00000103, 0.00006448, -0.00015356, 0.00015961, -0.00014374, -0.00005077, 0.00020768, -0.00078011, 0.00039144, 0.00124317, -0.00056173, -0.00030412, 0.00026743, -0.00027690, 0.00011596, -0.00001574, -0.00009489, 0.00011644, -0.00011301, 0.00004736, 0.00000992, -0.00005152, 0.00006157, -0.00005891, 0.00002156, 0.00000538, -0.00003184, 0.00003912, -0.00003201, 0.00000725, 0.00000279, -0.00000811, 0.00003485, -0.00003032, 0.00000685, 0.00001852, -0.00003403, 0.00001110, 0.00000206, -0.00001612, 0.00000652, -0.00001005, -0.00000058, 0.00000586, -0.00001706, 0.00000178, 0.00000620, -0.00001540, 0.00000283, 0.00000713, -0.00001240, 0.00000168, 0.00000817, -0.00000678, 0.00000355, 0.00000286, -0.00000721, 0.00000783, 0.00000119, -0.00000308, 0.00000484, -0.00000621, 0.00000121, -0.00000011, -0.00000217, 0.00001601, 0.00000179, -0.00000281, 0.00000763, -0.00000530, -0.00000533, 0.00000137, 0.00000302, 0.00000423, },
		{ -0.00000242, -0.00000718, 0.00000447, 0.00000150, -0.00000182, 0.00001630, 0.00000243, -0.00000553, 0.00000393, -0.00000530, 0.00000225, 0.00000119, -0.00000215, 0.00000872, -0.00000477, -0.00000173, 0.00000532, -0.00000596, 0.00000689, 0.00000357, -0.00001200, 0.00000531, 0.00000231, -0.00001419, 0.00000506, 0.00000242, -0.00001708, 0.00000339, 0.00000405, -0.00001143, 0.00000381, -0.00000897, -0.00000905, 0.00001405, -0.00002210, -0.00000278, 0.00002111, -0.00002696, 0.00001357, 0.00001506, -0.00000863, 0.00000988, -0.00002481, 0.00002053, -0.00000660, -0.00001580, 0.00003147, -0.00005140, 0.00003362, -0.00001004, -0.00003037, 0.00006954, -0.00009996, 0.00006564, -0.00001482, -0.00009703, 0.00016199, -0.00023833, 0.00012588, -0.00002569, -0.00075907, 0.00089693, 0.00086394, -0.00078340, 0.00000462, 0.00011841, -0.00024919, 0.00018093, -0.00011478, -0.00000529, 0.00006888, -0.00011579, 0.00008966, -0.00004316, -0.00000972, 0.00004137, -0.00006493, 0.00004754, -0.00002552, -0.00000774, 0.00003021, -0.00004009, 0.00002403, -0.00001256, 0.00000654, 0.00002663, -0.00003757, 0.00002428, 0.00000192, -0.00003082, 0.00001983, -0.00000846, -0.00000974, 0.00000545, -0.00001251, 0.00000576, -0.00000187, -0.00001592, 0.00000950, -0.00000239, -0.00001405, 0.00001021, 0.00000001, -0.00001195, 0.00000767, 0.00000309, -0.00000663, 0.00000682, -0.00000137, -0.00000433, 0.00000943, -0.00000278, -0.00000039, 0.00000308, -0.00000704, 0.00000324, -0.00000224, 0.00000277, 0.00001548, -0.00000327, 0.00000090, 0.00000618, -0.00000773, -0.00000316, 0.00000191, 0.00000367, 0.00000365, 0.00000137, },
		{ -0.00000548, 0.00000596, -0.00000141, 0.00000294, 0.00001650, -0.00000314, -0.00000296, 0.00000265, -0.00000519, 0.00000448, -0.00000126, 0.00000101, 0.00000717, -0.00000721, 0.00000197, 0.00000193, -0.00000424, 0.00001010, -0.00000324, -0.00000901, 0.00000891, -0.00000564, -0.00001052, 0.00000954, -0.00000623, -0.00001408, 0.00000867, -0.00000196, -0.00000882, 0.00000401, -0.00001337, -0.00000186, 0.00000845, -0.00002463, 0.00000902, 0.00001015, -0.00002455, 0.00002187, 0.00000627, -0.00000287, 0.00000060, -0.00001795, 0.00002424, -0.00002116, 0.00000530, 0.00001048, -0.00004362, 0.00004647, -0.00004054, 0.00001179, 0.00003099, -0.00009077, 0.00009994, -0.00009090, -0.00000515, 0.00009331, -0.00025073, 0.00025392, -0.00032592, -0.00051271, 0.00124574, 0.00034809, -0.00077535, 0.00022876, -0.00006760, -0.00013964, 0.00016717, -0.00016823, 0.00008009, -0.00000784, -0.00007505, 0.00009544, -0.00007998, 0.00003664, 0.00000183, -0.00004610, 0.00005415, -0.00004940, 0.00002207, 0.00000711, -0.00003308, 0.00003184, -0.00002607, 0.00002509, 0.00000894, -0.00003385, 0.00003584, -0.00001747, -0.00001908, 0.00002139, -0.00001673, -0.00000145, 0.00000046, -0.00001068, 0.00000899, -0.00001064, -0.00000908, 0.00001211, -0.00001163, -0.00000702, 0.00001300, -0.00000789, -0.00000710, 0.00001034, -0.00000258, -0.00000357, 0.00000764, -0.00000497, 0.00000020, 0.00000837, -0.00000550, 0.00000258, 0.00000010, -0.00000630, 0.00000416, -0.00000322, 0.00000805, 0.00001201, -0.00000610, 0.00000486, 0.00000297, -0.00000866, -0.00000101, 0.00000226, 0.00000418, 0.00000285, 0.00000095, -0.00000413, },
		{ 0.00000587, -0.00000297, 0.00000840, 0.00001368, -0.00000698, 0.00000000, 0.00000041, -0.00000351, 0.00000528, -0.00000296, 0.00000430, 0.00000389, -0.00000742, 0.00000433, -0.00000165, -0.00000001, 0.00000973, -0.00000906, -0.00000295, 0.00000782, -0.00001193, -0.00000327, 0.00000873, -0.00001350, -0.00000732, 0.00000986, -0.00000698, -0.00000483, 0.00000148, -0.00001478, 0.00000393, 0.00000082, -0.00002181, 0.00001637, -0.00000115, -0.00001741, 0.00002512, 0.00000046, 0.00000018, -0.00000761, -0.00000649, 0.00001822, -0.00002507, 0.00001989, -0.00001346, -0.00002186, 0.00003930, -0.00005246, 0.00004559, -0.00001889, -0.00004880, 0.00009084, -0.00013324, 0.00008480, -0.00002138, -0.00017161, 0.00027236, -0.00059554, -0.00008386, 0.00134748, -0.00016908, -0.00057732, 0.00031286, -0.00021684, 0.00000364, 0.00008631, -0.00015746, 0.00012779, -0.00008247, -0.00000729, 0.00006273, -0.00008534, 0.00006816, -0.00004141, -0.00000890, 0.00003746, -0.00005530, 0.00004533, -0.00002197, -0.00001226, 0.00002602, -0.00003063, 0.00004071, -0.00001392, -0.00001880, 0.00003664, -0.00003337, -0.00000238, 0.00001514, -0.00001954, 0.00000530, -0.00000619, -0.00000494, 0.00000706, -0.00001668, 0.00000100, 0.00000812, -0.00001737, 0.00000303, 0.00000992, -0.00001322, 0.00000015, 0.00000887, -0.00000655, 0.00000094, 0.00000581, -0.00000653, 0.00000488, 0.00000505, -0.00000619, 0.00000479, -0.00000333, -0.00000430, 0.00000393, -0.00000243, 0.00001222, 0.00000657, -0.00000610, 0.00000774, -0.00000115, -0.00000801, 0.00000072, 0.00000260, 0.00000437, 0.00000202, 0.00000041, -0.00000560, -0.00000295, },
		{ -0.00000252, 0.00001309, 0.00000859, -0.00000842, 0.00000228, -0.00000193, -0.00000072, 0.00000452, -0.00000332, 0.00000674, -0.00000010, -0.00000570, 0.00000465, -0.00000378, 0.00000507, 0.00000572, -0.00001164, 0.00000356, 0.00000247, -0.00001396, 0.00000434, 0.00000310, -0.00001673, 0.00000042, 0.00000732, -0.00000959, -0.00000123, -0.00000245, -0.00001344, 0.00000705, -0.00000636, -0.00001549, 0.00001872, -0.00001021, -0.00000820, 0.00002473, -0.00000290, -0.00000020, -0.00001135, 0.00000409, 0.00000698, -0.00001889, 0.00002291, -0.00003060, 0.00000369, 0.00001759, -0.00004280, 0.00005781, -0.00005968, 0.00000746, 0.00004334, -0.00012548, 0.00013590, -0.00013734, -0.00002890, 0.00015648, -0.00073585, 0.00042653, 0.00118050, -0.00056213, -0.00027865, 0.00025545, -0.00027842, 0.00012440, -0.00002499, -0.00009025, 0.00012070, -0.00012535, 0.00006036, 0.00000524, -0.00005715, 0.00007134, -0.00007053, 0.00003205, 0.00000332, -0.00003933, 0.00005154, -0.00004537, 0.00001494, 0.00000755, -0.00002209, 0.00004631, -0.00003464, 0.00000356, 0.00002524, -0.00004015, 0.00001362, 0.00000353, -0.00001591, 0.00000762, -0.00001123, 0.00000225, 0.00000014, -0.00001693, 0.00001013, -0.00000115, -0.00001671, 0.00001204, 0.00000218, -0.00001372, 0.00000685, 0.00000414, -0.00000741, 0.00000494, 0.00000224, -0.00000546, 0.00000817, 0.00000055, -0.00000478, 0.00000549, -0.00000634, -0.00000165, 0.00000297, 0.00000012, 0.00001410, 0.00000068, -0.00000355, 0.00000857, -0.00000512, -0.00000610, 0.00000182, 0.00000303, 0.00000414, 0.00000133, -0.00000035, -0.00000649, -0.00000015, 0.00000445, },
		{ 0.00001576, 0.00000253, -0.00000757, 0.00000317, -0.00000346, 0.00000237, 0.00000254, -0.00000222, 0.00000765, -0.00000368, -0.00000299, 0.00000313, -0.00000344, 0.00000891, -0.00000052, -0.00001006, 0.00000774, -0.00000493, -0.00001098, 0.00000908, -0.00000498, -0.00001503, 0.00000648, 0.00000269, -0.00000964, 0.00000083, -0.00000631, -0.00001045, 0.00000746, -0.00001151, -0.00000792, 0.00001699, -0.00001595, 0.00000115, 0.00002224, -0.00000533, -0.00000248, -0.00001004, 0.00000987, -0.00000367, -0.00000739, 0.00001483, -0.00003527, 0.00002271, -0.00000787, -0.00001786, 0.00004511, -0.00007588, 0.00005456, -0.00002201, -0.00007214, 0.00012700, -0.00020801, 0.00012117, -0.00006952, -0.00068108, 0.00088930, 0.00079699, -0.00075192, 0.00001493, 0.00010249, -0.00023995, 0.00018036, -0.00012125, 0.00000382, 0.00006428, -0.00012049, 0.00010141, -0.00005346, -0.00000727, 0.00004443, -0.00007311, 0.00006002, -0.00003472, -0.00000677, 0.00003674, -0.00005285, 0.00003763, -0.00001680, -0.00000158, 0.00003778, -0.00004560, 0.00002586, 0.00000470, -0.00003525, 0.00002332, -0.00000878, -0.00000768, 0.00000460, -0.00001194, 0.00000733, -0.00000915, -0.00001059, 0.00001414, -0.00001190, -0.00000938, 0.00001626, -0.00000707, -0.00000920, 0.00001039, -0.00000169, -0.00000514, 0.00000691, -0.00000151, -0.00000219, 0.00000903, -0.00000372, -0.00000185, 0.00000438, -0.00000824, 0.00000099, 0.00000197, 0.00000378, 0.00001316, -0.00000405, 0.00000051, 0.00000708, -0.00000796, -0.00000354, 0.00000236, 0.00000353, 0.00000351, 0.00000089, -0.00000141, -0.00000657, 0.00000231, 0.00000236, -0.00000001, },
		{ -0.00000303, -0.00000515, 0.00000258, -0.00000354, 0.00000483, 0.00000004, 0.00000004, 0.00000684, -0.00000603, -0.00000041, 0.00000077, -0.00000067, 0.00000988, -0.00000668, -0.00000511, 0.00000776, -0.00001132, -0.00000432, 0.00000900, -0.00001233, -0.00000944, 0.00000925, -0.00000211, -0.00000795, 0.00000114, -0.00000918, -0.00000704, 0.00000603, -0.00001420, -0.00000077, 0.00001261, -0.00001835, 0.00000974, 0.00001838, -0.00000796, -0.00000430, -0.00000563, 0.00000994, -0.00000948, 0.00000322, 0.00000076, -0.00002768, 0.00002908, -0.00002617, 0.00001026, 0.00001474, -0.00006312, 0.00007414, -0.00007822, 0.00000358, 0.00006158, -0.00020430, 0.00021489, -0.00033754, -0.00042247, 0.00118660, 0.00030446, -0.00072482, 0.00021645, -0.00007477, -0.00012749, 0.00015806, -0.00016664, 0.00008655, -0.00001672, -0.00007150, 0.00010089, -0.00008996, 0.00004359, -0.00000170, -0.00004757, 0.00006324, -0.00006091, 0.00002971, 0.00000567, -0.00004023, 0.00004603, -0.00003694, 0.00002427, 0.00001626, -0.00004207, 0.00004006, -0.00001821, -0.00002038, 0.00002357, -0.00001686, 0.00000129, -0.00000206, -0.00000765, 0.00000728, -0.00001661, 0.00000018, 0.00001080, -0.00001940, 0.00000203, 0.00001382, -0.00001401, -0.00000157, 0.00000956, -0.00000618, -0.00000102, 0.00000631, -0.00000388, 0.00000203, 0.00000731, -0.00000653, 0.00000158, 0.00000174, -0.00000867, 0.00000315, 0.00000162, 0.00000749, 0.00000973, -0.00000647, 0.00000468, 0.00000369, -0.00000906, -0.00000102, 0.00000255, 0.00000395, 0.00000262, 0.00000068, -0.00000272, -0.00000576, 0.00000396, 0.00000044, 0.00000399, 0.00001572, },
		{ -0.00000224, 0.00000097, -0.00000204, 0.00000593, -0.00000216, 0.00000280, 0.00000459, -0.00000684, 0.00000118, -0.00000112, 0.00000346, 0.00000744, -0.00001050, 0.00000108, 0.00000359, -0.00001408, 0.00000321, 0.00000417, -0.00001629, -0.00000225, 0.00000862, -0.00000565, -0.00000570, 0.00000020, -0.00001089, -0.00000395, 0.00000376, -0.00001481, 0.00000513, 0.00000679, -0.00001781, 0.00001722, 0.00001305, -0.00001071, -0.00000396, -0.00000116, 0.00000627, -0.00000942, 0.00000830, -0.00001256, -0.00001299, 0.00002293, -0.00003102, 0.00002941, -0.00001934, -0.00002926, 0.00006029, -0.00010336, 0.00006988, -0.00003363, -0.00012714, 0.00020590, -0.00055723, -0.00001348, 0.00124747, -0.00016917, -0.00052961, 0.00028279, -0.00020808, 0.00000887, 0.00007504, -0.00014818, 0.00012680, -0.00008918, 0.00000066, 0.00006093, -0.00009059, 0.00007466, -0.00004816, -0.00000404, 0.00004002, -0.00006402, 0.00005520, -0.00002979, -0.00001149, 0.00003582, -0.00004375, 0.00004573, -0.00001172, -0.00002456, 0.00004067, -0.00003550, -0.00000111, 0.00001503, -0.00001775, 0.00000679, -0.00000876, -0.00000029, 0.00000129, -0.00001833, 0.00001093, 0.00000098, -0.00001999, 0.00001306, 0.00000563, -0.00001584, 0.00000604, 0.00000504, -0.00000768, 0.00000304, 0.00000375, -0.00000400, 0.00000569, 0.00000366, -0.00000721, 0.00000443, -0.00000181, -0.00000767, 0.00000469, 0.00000225, 0.00001009, 0.00000482, -0.00000617, 0.00000760, -0.00000062, -0.00000838, 0.00000095, 0.00000266, 0.00000411, 0.00000170, 0.00000053, -0.00000411, -0.00000417, 0.00000458, -0.00000051, 0.00000844, 0.00001300, -0.00000695, },
		{ -0.00000081, 0.00000064, 0.00000534, -0.00000333, 0.00000532, 0.00000152, -0.00000636, 0.00000153, -0.00000148, 0.00000725, 0.00000237, -0.00001069, 0.00000590, -0.00000310, -0.00001213, 0.00000849, -0.00000346, -0.00001569, 0.00000406, 0.00000563, -0.00000746, -0.00000372, -0.00000131, -0.00001179, -0.00000125, 0.00000123, -0.00001396, 0.00000957, 0.00000027, -0.00001451, 0.00002315, 0.00000593, -0.00001233, -0.00000145, 0.00000105, 0.00000213, -0.00000556, 0.00000664, -0.00001989, 0.00000173, 0.00000956, -0.00002290, 0.00003274, -0.00004303, 0.00000980, 0.00002128, -0.00008955, 0.00009987, -0.00011990, -0.00000813, 0.00008436, -0.00064646, 0.00044586, 0.00106747, -0.00051457, -0.00025488, 0.00022199, -0.00025577, 0.00011662, -0.00003009, -0.00007885, 0.00011241, -0.00012536, 0.00006697, -0.00000076, -0.00005677, 0.00007382, -0.00007601, 0.00003987, -0.00000034, -0.00004222, 0.00005886, -0.00005538, 0.00002247, 0.00001029, -0.00003307, 0.00005393, -0.00003651, 0.00000098, 0.00002721, -0.00004119, 0.00001540, 0.00000197, -0.00001178, 0.00000638, -0.00001173, 0.00000625, -0.00000848, -0.00001257, 0.00001673, -0.00001145, -0.00001280, 0.00001926, -0.00000502, -0.00001193, 0.00001062, -0.00000104, -0.00000592, 0.00000539, 0.00000064, -0.00000197, 0.00000756, -0.00000069, -0.00000580, 0.00000588, -0.00000547, -0.00000553, 0.00000575, 0.00000373, 0.00001080, -0.00000017, -0.00000359, 0.00000841, -0.00000469, -0.00000633, 0.00000215, 0.00000286, 0.00000389, 0.00000101, 0.00000025, -0.00000531, -0.00000209, 0.00000425, -0.00000003, 0.00001215, 0.00000832, -0.00000865, 0.00000012, },
		{ 0.00000365, 0.00000326, -0.00000308, 0.00000694, -0.00000168, -0.00000511, 0.00000101, 0.00000002, 0.00000912, -0.00000352, -0.00000741, 0.00000737, -0.00000961, -0.00000630, 0.00000940, -0.00001100, -0.00001112, 0.00000780, 0.00000182, -0.00000786, -0.00000230, -0.00000304, -0.00001218, 0.00000133, -0.00000153, -0.00001193, 0.00001247, -0.00000643, -0.00000843, 0.00002657, -0.00000262, -0.00001136, 0.00000165, 0.00000056, 0.00000012, -0.00000147, 0.00000045, -0.00001964, 0.00001091, -0.00000369, -0.00000792, 0.00002106, -0.00004813, 0.00003754, -0.00002488, -0.00004554, 0.00008271, -0.00016267, 0.00010234, -0.00011707, -0.00055986, 0.00083992, 0.00070657, -0.00067093, 0.00000371, 0.00008236, -0.00021364, 0.00016083, -0.00011479, 0.00000991, 0.00005354, -0.00011358, 0.00010243, -0.00005885, -0.00000375, 0.00004242, -0.00007425, 0.00006648, -0.00004143, -0.00000406, 0.00003866, -0.00006047, 0.00004819, -0.00002071, -0.00000785, 0.00004472, -0.00004915, 0.00002526, 0.00000461, -0.00003378, 0.00002334, -0.00000966, -0.00000248, 0.00000063, -0.00000899, 0.00000799, -0.00001760, -0.00000096, 0.00001440, -0.00002113, -0.00000023, 0.00001805, -0.00001387, -0.00000403, 0.00001058, -0.00000607, -0.00000206, 0.00000533, -0.00000151, 0.00000124, 0.00000713, -0.00000448, -0.00000290, 0.00000557, -0.00000851, -0.00000263, 0.00000660, 0.00000553, 0.00000941, -0.00000391, 0.00000020, 0.00000696, -0.00000754, -0.00000362, 0.00000263, 0.00000317, 0.00000329, 0.00000066, -0.00000038, -0.00000604, 0.00000009, 0.00000337, 0.00000191, 0.00001414, 0.00000285, -0.00000824, 0.00000128, -0.00000184, },
		{ 0.00000034, -0.00000138, 0.00000724, -0.00000444, -0.00000369, 0.00000037, 0.00000287, 0.00000820, -0.00000818, -0.00000216, 0.00000492, -0.00001335, 0.00000101, 0.00000561, -0.00001575, -0.00000448, 0.00000852, -0.00000158, -0.00000748, -0.00000129, -0.00000502, -0.00001201, 0.00000390, -0.00000474, -0.00000860, 0.00001344, -0.00001254, 0.00000022, 0.00002620, -0.00001117, -0.00000725, 0.00000325, -0.00000120, 0.00000085, -0.00000008, -0.00000631, -0.00001384, 0.00001273, -0.00001120, 0.00000577, 0.00000163, -0.00003533, 0.00004396, -0.00005881, 0.00000789, 0.00002764, -0.00014608, 0.00015487, -0.00033319, -0.00030487, 0.00107279, 0.00026703, -0.00063769, 0.00017404, -0.00007059, -0.00011068, 0.00013496, -0.00014939, 0.00008225, -0.00002313, -0.00006179, 0.00009576, -0.00009105, 0.00004622, -0.00000573, -0.00004424, 0.00006541, -0.00006660, 0.00003512, 0.00000237, -0.00004255, 0.00005530, -0.00004473, 0.00002279, 0.00002071, -0.00004513, 0.00003936, -0.00001864, -0.00001691, 0.00002073, -0.00001493, 0.00000517, -0.00000704, -0.00000160, 0.00000292, -0.00002118, 0.00001190, 0.00000415, -0.00002349, 0.00001290, 0.00000993, -0.00001748, 0.00000456, 0.00000632, -0.00000813, 0.00000199, 0.00000332, -0.00000176, 0.00000428, 0.00000473, -0.00000674, 0.00000055, 0.00000356, -0.00001042, 0.00000072, 0.00000739, 0.00000691, 0.00000641, -0.00000559, 0.00000387, 0.00000378, -0.00000863, -0.00000099, 0.00000266, 0.00000348, 0.00000246, 0.00000065, -0.00000144, -0.00000606, 0.00000197, 0.00000247, 0.00000486, 0.00001388, -0.00000217, -0.00000641, 0.00000118, -0.00000151, 0.00000598, },
		{ 0.00000136, 0.00000609, -0.00000640, -0.00000246, 0.00000037, 0.00000595, 0.00000471, -0.00001011, 0.00000285, -0.00000042, -0.00001291, 0.00000690, -0.00000141, -0.00001627, 0.00000193, 0.00000687, -0.00000398, -0.00000681, -0.00000050, -0.00000752, -0.00001081, 0.00000610, -0.00000835, -0.00000383, 0.00001187, -0.00001679, 0.00001046, 0.00002113, -0.00001745, -0.00000097, 0.00000206, -0.00000202, 0.00000286, -0.00000221, -0.00001022, -0.00000652, 0.00000916, -0.00001138, 0.00001206, -0.00001592, -0.00001293, 0.00002923, -0.00006825, 0.00004754, -0.00004069, -0.00007967, 0.00012087, -0.00048835, 0.00006008, 0.00109603, -0.00014004, -0.00046610, 0.00022611, -0.00017958, 0.00000663, 0.00005950, -0.00012706, 0.00011243, -0.00008639, 0.00000711, 0.00005300, -0.00008688, 0.00007345, -0.00005050, 0.00000118, 0.00003778, -0.00006593, 0.00005952, -0.00003555, -0.00000846, 0.00004083, -0.00005193, 0.00004720, -0.00000960, -0.00002638, 0.00003838, -0.00003398, 0.00000219, 0.00001020, -0.00001215, 0.00000713, -0.00001203, 0.00000658, -0.00000763, -0.00001631, 0.00002020, -0.00001018, -0.00001685, 0.00002136, -0.00000182, -0.00001468, 0.00001041, -0.00000006, -0.00000672, 0.00000444, 0.00000071, -0.00000009, 0.00000597, 0.00000126, -0.00000713, 0.00000362, 0.00000028, -0.00001089, 0.00000426, 0.00000808, 0.00000729, 0.00000273, -0.00000510, 0.00000631, -0.00000014, -0.00000798, 0.00000104, 0.00000255, 0.00000363, 0.00000164, 0.00000078, -0.00000283, -0.00000530, 0.00000332, 0.00000209, 0.00000803, 0.00001148, -0.00000577, -0.00000413, 0.00000030, 0.00000029, 0.00000682, -0.00000249, },
		{ 0.00000368, -0.00000744, -0.00000148, 0.00000139, 0.00000799, -0.00000021, -0.00000890, 0.00000565, -0.00000649, -0.00000854, 0.00000908, -0.00000907, -0.00001271, 0.00000634, 0.00000408, -0.00000547, -0.00000595, -0.00000002, -0.00001052, -0.00000803, 0.00000707, -0.00001170, 0.00000206, 0.00000738, -0.00001764, 0.00002032, 0.00001167, -0.00001937, 0.00000521, -0.00000155, -0.00000035, 0.00000370, -0.00000632, -0.00001008, -0.00000125, 0.00000419, -0.00000672, 0.00000953, -0.00002434, 0.00000777, 0.00000257, -0.00005266, 0.00005792, -0.00009336, 0.00000563, 0.00000454, -0.00052257, 0.00044436, 0.00091971, -0.00042745, -0.00023570, 0.00017375, -0.00021401, 0.00009429, -0.00002958, -0.00006364, 0.00009384, -0.00011321, 0.00006556, -0.00000631, -0.00005108, 0.00006865, -0.00007410, 0.00004335, -0.00000451, -0.00004025, 0.00005955, -0.00005997, 0.00002843, 0.00001054, -0.00003919, 0.00005603, -0.00003523, -0.00000056, 0.00002352, -0.00003633, 0.00001589, -0.00000236, -0.00000368, 0.00000245, -0.00001092, 0.00001063, -0.00001914, -0.00000374, 0.00001957, -0.00002257, -0.00000333, 0.00002179, -0.00001248, -0.00000695, 0.00001147, -0.00000577, -0.00000290, 0.00000447, -0.00000100, 0.00000262, 0.00000577, -0.00000219, -0.00000582, 0.00000561, -0.00000369, -0.00000977, 0.00000777, 0.00000838, 0.00000643, -0.00000059, -0.00000301, 0.00000692, -0.00000374, -0.00000606, 0.00000224, 0.00000250, 0.00000350, 0.00000105, 0.00000076, -0.00000431, -0.00000383, 0.00000406, 0.00000253, 0.00001054, 0.00000758, -0.00000753, -0.00000226, -0.00000055, 0.00000305, 0.00000576, -0.00000430, 0.00000443, },
		{ -0.00000759, -0.00000056, 0.00000329, 0.00000812, -0.00000495, -0.00000535, 0.00000524, -0.00001092, -0.00000206, 0.00000676, -0.00001469, -0.00000660, 0.00000800, 0.00000125, -0.00000643, -0.00000471, -0.00000032, -0.00001352, -0.00000359, 0.00000594, -0.00001363, 0.00000799, 0.00000038, -0.00001398, 0.00002720, -0.00000029, -0.00001606, 0.00000873, -0.00000563, 0.00000362, 0.00000157, -0.00000974, -0.00000711, 0.00000057, 0.00000129, -0.00000177, 0.00000144, -0.00002182, 0.00001794, -0.00002284, -0.00002245, 0.00003747, -0.00010983, 0.00007024, -0.00015764, -0.00041215, 0.00075378, 0.00060460, -0.00055357, -0.00002541, 0.00006148, -0.00017580, 0.00012664, -0.00009670, 0.00001139, 0.00003949, -0.00009700, 0.00009262, -0.00005796, -0.00000054, 0.00003612, -0.00006824, 0.00006572, -0.00004406, -0.00000056, 0.00003569, -0.00006137, 0.00005374, -0.00002351, -0.00001143, 0.00004610, -0.00004721, 0.00002215, 0.00000145, -0.00002614, 0.00001945, -0.00001057, 0.00000532, -0.00000619, -0.00000347, 0.00000706, -0.00002576, 0.00001195, 0.00000923, -0.00002735, 0.00001194, 0.00001411, -0.00001795, 0.00000228, 0.00000781, -0.00000857, 0.00000136, 0.00000246, -0.00000083, 0.00000510, 0.00000387, -0.00000474, -0.00000336, 0.00000612, -0.00000771, -0.00000703, 0.00001094, 0.00000792, 0.00000461, -0.00000277, -0.00000024, 0.00000575, -0.00000623, -0.00000355, 0.00000268, 0.00000259, 0.00000309, 0.00000081, 0.00000035, -0.00000553, -0.00000191, 0.00000433, 0.00000380, 0.00001169, 0.00000315, -0.00000762, -0.00000128, -0.00000055, 0.00000583, 0.00000305, -0.00000441, 0.00000699, 0.00000043, },
		{ 0.00000056, 0.00000549, 0.00000613, -0.00000808, -0.00000102, 0.00000198, -0.00001210, 0.00000398, 0.00000085, -0.00001652, -0.00000007, 0.00000728, -0.00000099, -0.00000714, -0.00000297, -0.00000195, -0.00001550, 0.00000184, 0.00000239, -0.00001297, 0.00001231, -0.00000763, -0.00000585, 0.00002881, -0.00001168, -0.00000853, 0.00000806, -0.00000769, 0.00000800, -0.00000355, -0.00001025, -0.00000394, 0.00000007, 0.00000153, -0.00000026, -0.00000659, -0.00001215, 0.00001501, -0.00003630, 0.00000675, -0.00000180, -0.00008667, 0.00008236, -0.00030959, -0.00017733, 0.00091876, 0.00023934, -0.00052691, 0.00011034, -0.00005666, -0.00009247, 0.00010283, -0.00012024, 0.00006810, -0.00002538, -0.00004850, 0.00008157, -0.00008316, 0.00004358, -0.00000900, -0.00003703, 0.00006057, -0.00006533, 0.00003685, -0.00000192, -0.00003963, 0.00005799, -0.00004801, 0.00002051, 0.00002164, -0.00004227, 0.00003332, -0.00001858, -0.00000903, 0.00001303, -0.00001064, 0.00000949, -0.00001369, 0.00000693, -0.00000408, -0.00002301, 0.00002407, -0.00000729, -0.00002177, 0.00002289, 0.00000158, -0.00001648, 0.00000925, 0.00000143, -0.00000772, 0.00000411, -0.00000024, 0.00000125, 0.00000622, 0.00000107, -0.00000595, -0.00000042, 0.00000503, -0.00001111, -0.00000279, 0.00001329, 0.00000649, 0.00000241, -0.00000353, 0.00000224, 0.00000333, -0.00000722, -0.00000111, 0.00000262, 0.00000276, 0.00000253, 0.00000086, -0.00000058, -0.00000617, 0.00000016, 0.00000443, 0.00000557, 0.00001120, -0.00000084, -0.00000667, -0.00000110, 0.00000072, 0.00000759, -0.00000052, -0.00000266, 0.00000829, -0.00000310, -0.00000694, },
		{ 0.00000719, 0.00000258, -0.00000888, 0.00000240, -0.00000272, -0.00000973, 0.00000729, -0.00000650, -0.00001423, 0.00000501, 0.00000516, -0.00000265, -0.00000753, -0.00000095, -0.00000513, -0.00001536, 0.00000689, -0.00000294, -0.00000925, 0.00001340, -0.00001433, 0.00000517, 0.00002420, -0.00001929, 0.00000055, 0.00000349, -0.00000606, 0.00001016, -0.00000979, -0.00000744, -0.00000279, -0.00000028, 0.00000334, -0.00000310, -0.00000991, -0.00000210, 0.00000308, -0.00003428, 0.00002172, -0.00004064, -0.00003857, 0.00003109, -0.00039562, 0.00012507, 0.00091278, -0.00008812, -0.00039464, 0.00015319, -0.00013752, -0.00000222, 0.00004284, -0.00009863, 0.00008790, -0.00007464, 0.00001035, 0.00004109, -0.00007542, 0.00006471, -0.00004769, 0.00000548, 0.00003166, -0.00006095, 0.00005727, -0.00003790, -0.00000386, 0.00004032, -0.00005388, 0.00004450, -0.00000762, -0.00002402, 0.00002976, -0.00002879, 0.00000692, 0.00000141, -0.00000314, 0.00000604, -0.00001515, 0.00001460, -0.00001869, -0.00001023, 0.00002666, -0.00002329, -0.00000737, 0.00002514, -0.00001064, -0.00000926, 0.00001143, -0.00000476, -0.00000409, 0.00000434, -0.00000201, 0.00000437, 0.00000546, -0.00000169, -0.00000586, 0.00000238, 0.00000243, -0.00001324, 0.00000263, 0.00001429, 0.00000423, 0.00000054, -0.00000309, 0.00000377, 0.00000047, -0.00000681, 0.00000081, 0.00000233, 0.00000291, 0.00000200, 0.00000099, -0.00000194, -0.00000603, 0.00000209, 0.00000461, 0.00000729, 0.00000924, -0.00000375, -0.00000543, -0.00000119, 0.00000313, 0.00000754, -0.00000382, 0.00000055, 0.00000786, -0.00000629, -0.00000562, 0.00000202, },
		{ -0.00000151, -0.00000746, 0.00000375, -0.00000695, -0.00000492, 0.00000667, -0.00001275, -0.00000888, 0.00000760, 0.00000273, -0.00000400, -0.00000728, 0.00000070, -0.00000943, -0.00001249, 0.00000997, -0.00000855, -0.00000308, 0.00001049, -0.00001740, 0.00001623, 0.00001432, -0.00002103, 0.00000791, -0.00000283, -0.00000087, 0.00000822, -0.00001437, -0.00000287, -0.00000416, 0.00000152, 0.00000377, -0.00000804, -0.00000728, 0.00000249, -0.00000996, -0.00002122, 0.00001754, -0.00006191, 0.00000847, -0.00006878, -0.00037977, 0.00042071, 0.00075561, -0.00031504, -0.00022101, 0.00011841, -0.00016127, 0.00006225, -0.00002378, -0.00004774, 0.00006912, -0.00009165, 0.00005634, -0.00000990, -0.00004185, 0.00005713, -0.00006515, 0.00004174, -0.00000795, -0.00003429, 0.00005358, -0.00005815, 0.00003163, 0.00000853, -0.00003979, 0.00005210, -0.00003071, -0.00000110, 0.00001464, -0.00002618, 0.00001486, -0.00000848, 0.00000733, -0.00000367, -0.00000852, 0.00001435, -0.00003008, 0.00000858, 0.00001739, -0.00003173, 0.00001040, 0.00001794, -0.00001773, 0.00000020, 0.00000858, -0.00000828, 0.00000033, 0.00000231, -0.00000174, 0.00000713, 0.00000310, -0.00000366, -0.00000477, 0.00000455, -0.00000139, -0.00001348, 0.00000861, 0.00001346, 0.00000164, -0.00000057, -0.00000203, 0.00000415, -0.00000212, -0.00000539, 0.00000201, 0.00000202, 0.00000299, 0.00000161, 0.00000093, -0.00000349, -0.00000510, 0.00000374, 0.00000502, 0.00000842, 0.00000639, -0.00000538, -0.00000449, -0.00000082, 0.00000598, 0.00000547, -0.00000574, 0.00000436, 0.00000563, -0.00000859, -0.00000377, 0.00000375, 0.00000774, },
		{ -0.00000468, 0.00000284, -0.00000919, 0.00000036, 0.00000242, -0.00001592, -0.00000235, 0.00000768, 0.00000069, -0.00000533, -0.00000610, 0.00000113, -0.00001378, -0.00000713, 0.00000996, -0.00001255, 0.00000387, 0.00000411, -0.00001542, 0.00002415, 0.00000190, -0.00001681, 0.00001103, -0.00000794, 0.00000585, 0.00000212, -0.00001505, 0.00000082, -0.00000656, 0.00000558, 0.00000064, -0.00001125, -0.00000128, -0.00000032, -0.00001691, -0.00000647, -0.00000082, -0.00005851, 0.00002878, -0.00018224, -0.00025776, 0.00064121, 0.00050207, -0.00041693, -0.00006468, 0.00004183, -0.00013305, 0.00008462, -0.00007089, 0.00000816, 0.00002515, -0.00007448, 0.00007419, -0.00005092, 0.00000124, 0.00002720, -0.00005646, 0.00005811, -0.00004192, 0.00000254, 0.00002874, -0.00005551, 0.00005333, -0.00002450, -0.00001228, 0.00004181, -0.00004001, 0.00001670, -0.00000418, -0.00001361, 0.00001228, -0.00001096, 0.00001447, -0.00001463, 0.00000397, 0.00000426, -0.00003188, 0.00002598, -0.00000102, -0.00002828, 0.00002447, 0.00000446, -0.00001729, 0.00000777, 0.00000258, -0.00000812, 0.00000348, -0.00000068, 0.00000071, 0.00000832, -0.00000003, -0.00000453, -0.00000306, 0.00000573, -0.00000594, -0.00001135, 0.00001420, 0.00001071, -0.00000056, -0.00000086, -0.00000095, 0.00000359, -0.00000395, -0.00000347, 0.00000251, 0.00000183, 0.00000298, 0.00000138, 0.00000050, -0.00000489, -0.00000351, 0.00000507, 0.00000561, 0.00000861, 0.00000335, -0.00000599, -0.00000398, 0.00000054, 0.00000823, 0.00000187, -0.00000557, 0.00000766, 0.00000202, -0.00000963, -0.00000157, 0.00000548, 0.00000683, -0.00000505, },
		{ 0.00000033, -0.00000877, 0.00000405, -0.00000390, -0.00001520, 0.00000343, 0.00000602, -0.00000079, -0.00000659, -0.00000403, -0.00000029, -0.00001675, -0.00000050, 0.00000673, -0.00001350, 0.00000940, -0.00000385, -0.00000858, 0.00002659, -0.00000948, -0.00000859, 0.00000920, -0.00000938, 0.00001110, -0.00000619, -0.00001142, 0.00000148, -0.00000751, 0.00000992, -0.00000597, -0.00000975, 0.00000337, -0.00000782, -0.00001502, 0.00000099, -0.00002185, -0.00003630, 0.00000827, -0.00026710, -0.00005726, 0.00074357, 0.00022149, -0.00040651, 0.00003745, -0.00003731, -0.00007517, 0.00006756, -0.00008506, 0.00004728, -0.00002306, -0.00003456, 0.00006150, -0.00006814, 0.00003598, -0.00001052, -0.00002764, 0.00005011, -0.00005753, 0.00003427, -0.00000599, -0.00003232, 0.00005395, -0.00004629, 0.00001731, 0.00001924, -0.00003419, 0.00002276, -0.00001792, 0.00000179, 0.00000199, -0.00000443, 0.00001352, -0.00002059, 0.00001646, -0.00001280, -0.00002128, 0.00003420, -0.00002149, -0.00001348, 0.00002912, -0.00000955, -0.00001028, 0.00001055, -0.00000357, -0.00000502, 0.00000417, -0.00000291, 0.00000447, 0.00000732, -0.00000289, -0.00000443, -0.00000107, 0.00000561, -0.00001049, -0.00000671, 0.00001827, 0.00000644, -0.00000171, -0.00000067, -0.00000020, 0.00000249, -0.00000487, -0.00000151, 0.00000245, 0.00000181, 0.00000293, 0.00000122, -0.00000034, -0.00000583, -0.00000147, 0.00000612, 0.00000613, 0.00000786, 0.00000072, -0.00000608, -0.00000361, 0.00000297, 0.00000891, -0.00000220, -0.00000324, 0.00000946, -0.00000224, -0.00000927, 0.00000076, 0.00000683, 0.00000461, -0.00000727, -0.00000168, },
		{ -0.00000615, 0.00000481, -0.00001012, -0.00001107, 0.00000711, 0.00000365, -0.00000192, -0.00000748, -0.00000157, -0.00000358, -0.00001719, 0.00000567, 0.00000129, -0.00001094, 0.00001171, -0.00001083, 0.00000126, 0.00002291, -0.00001671, 0.00000043, 0.00000378, -0.00000639, 0.00001230, -0.00001359, -0.00000512, -0.00000134, -0.00000507, 0.00001157, -0.00001344, -0.00000333, 0.00000274, -0.00001454, -0.00000711, -0.00000269, -0.00003386, -0.00001099, -0.00004926, -0.00028952, 0.00017253, 0.00071952, -0.00002385, -0.00032166, 0.00007576, -0.00009047, -0.00001481, 0.00002752, -0.00006821, 0.00005834, -0.00005650, 0.00000970, 0.00002784, -0.00005894, 0.00005024, -0.00004013, 0.00000785, 0.00002336, -0.00005051, 0.00004891, -0.00003617, 0.00000116, 0.00003489, -0.00004972, 0.00003787, -0.00000576, -0.00001829, 0.00001631, -0.00002085, 0.00001215, -0.00000943, 0.00000784, 0.00000384, -0.00001728, 0.00002212, -0.00002994, -0.00000078, 0.00002862, -0.00003541, 0.00000723, 0.00002253, -0.00001833, -0.00000043, 0.00000805, -0.00000723, -0.00000089, 0.00000248, -0.00000312, 0.00000811, 0.00000435, -0.00000466, -0.00000372, 0.00000097, 0.00000397, -0.00001409, 0.00000006, 0.00001979, 0.00000156, -0.00000153, -0.00000055, 0.00000016, 0.00000127, -0.00000496, 0.00000018, 0.00000205, 0.00000201, 0.00000284, 0.00000098, -0.00000150, -0.00000614, 0.00000080, 0.00000692, 0.00000629, 0.00000644, -0.00000123, -0.00000612, -0.00000281, 0.00000597, 0.00000752, -0.00000545, 0.00000063, 0.00000915, -0.00000623, -0.00000768, 0.00000297, 0.00000747, 0.00000157, -0.00000790, 0.00000057, -0.00000260, },
		{ 0.00000239, -0.00001426, -0.00000499, 0.00000824, 0.00000146, -0.00000303, -0.00000759, 0.00000046, -0.00000807, -0.00001464, 0.00000970, -0.00000460, -0.00000572, 0.00001008, -0.00001452, 0.00001121, 0.00001445, -0.00001825, 0.00000710, -0.00000249, -0.00000025, 0.00000855, -0.00001727, 0.00000095, -0.00000604, 0.00000079, 0.00000836, -0.00001795, 0.00000511, -0.00000398, -0.00001552, 0.00000038, -0.00001454, -0.00003115, -0.00000039, -0.00012353, -0.00023567, 0.00037764, 0.00059284, -0.00019437, -0.00020820, 0.00006309, -0.00010676, 0.00002717, -0.00001466, -0.00003364, 0.00004307, -0.00006517, 0.00004130, -0.00001080, -0.00003129, 0.00004177, -0.00005101, 0.00003546, -0.00000966, -0.00002607, 0.00004242, -0.00005032, 0.00003140, 0.00000506, -0.00003547, 0.00004297, -0.00002361, -0.00000097, 0.00000234, -0.00001277, 0.00001259, -0.00001494, 0.00001929, -0.00001055, -0.00000481, 0.00001646, -0.00003904, 0.00002222, 0.00001020, -0.00003631, 0.00002565, 0.00000761, -0.00001863, 0.00000762, 0.00000221, -0.00000729, 0.00000217, -0.00000035, -0.00000099, 0.00001023, 0.00000033, -0.00000504, -0.00000280, 0.00000285, 0.00000081, -0.00001572, 0.00000799, 0.00001817, -0.00000270, -0.00000026, -0.00000093, 0.00000037, 0.00000015, -0.00000443, 0.00000140, 0.00000149, 0.00000243, 0.00000265, 0.00000056, -0.00000276, -0.00000575, 0.00000315, 0.00000738, 0.00000592, 0.00000483, -0.00000261, -0.00000626, -0.00000109, 0.00000859, 0.00000426, -0.00000681, 0.00000491, 0.00000675, -0.00000913, -0.00000522, 0.00000484, 0.00000722, -0.00000169, -0.00000712, 0.00000160, -0.00000478, -0.00000258, },
		{ -0.00001511, 0.00000123, 0.00000719, -0.00000011, -0.00000429, -0.00000668, 0.00000123, -0.00001255, -0.00000960, 0.00001062, -0.00000905, 0.00000042, 0.00000516, -0.00001375, 0.00001838, 0.00000399, -0.00001459, 0.00000956, -0.00000678, 0.00000633, 0.00000105, -0.00001599, 0.00000406, -0.00000981, 0.00000797, 0.00000039, -0.00001657, 0.00001099, -0.00001384, -0.00000929, 0.00000118, -0.00002663, -0.00001708, -0.00001579, -0.00018576, -0.00011568, 0.00051585, 0.00040728, -0.00027864, -0.00010438, 0.00002376, -0.00009160, 0.00004237, -0.00004276, 0.00000160, 0.00001297, -0.00005045, 0.00005097, -0.00003931, 0.00000103, 0.00001767, -0.00004140, 0.00004550, -0.00003543, 0.00000420, 0.00001958, -0.00004431, 0.00004731, -0.00002339, -0.00001097, 0.00003305, -0.00002901, 0.00000966, -0.00001109, 0.00000121, 0.00000347, -0.00001057, 0.00002335, -0.00002271, 0.00001186, -0.00000000, -0.00003446, 0.00003829, -0.00001439, -0.00002284, 0.00003423, -0.00000920, -0.00001105, 0.00001042, -0.00000372, -0.00000450, 0.00000296, -0.00000269, 0.00000277, 0.00001000, -0.00000356, -0.00000423, -0.00000189, 0.00000427, -0.00000359, -0.00001454, 0.00001559, 0.00001358, -0.00000523, 0.00000136, -0.00000191, 0.00000073, -0.00000083, -0.00000347, 0.00000206, 0.00000100, 0.00000302, 0.00000224, -0.00000004, -0.00000393, -0.00000468, 0.00000545, 0.00000731, 0.00000507, 0.00000343, -0.00000375, -0.00000625, 0.00000167, 0.00000983, 0.00000001, -0.00000581, 0.00000833, 0.00000284, -0.00001046, -0.00000241, 0.00000623, 0.00000605, -0.00000457, -0.00000544, 0.00000147, -0.00000556, 0.00000041, -0.00000231, },
		{ 0.00000602, 0.00000486, -0.00000105, -0.00000561, -0.00000488, 0.00000026, -0.00001572, -0.00000334, 0.00000852, -0.00001080, 0.00000547, -0.00000126, -0.00000887, 0.00002096, -0.00000536, -0.00000784, 0.00000783, -0.00000736, 0.00001046, -0.00000744, -0.00001065, 0.00000301, -0.00001013, 0.00001318, -0.00000968, -0.00000904, 0.00001050, -0.00002162, 0.00000113, -0.00000706, -0.00003068, -0.00000246, -0.00005631, -0.00020989, 0.00004128, 0.00056728, 0.00021061, -0.00028912, -0.00003201, -0.00001821, -0.00005985, 0.00003458, -0.00005031, 0.00002427, -0.00001716, -0.00002245, 0.00003953, -0.00004921, 0.00002491, -0.00000988, -0.00001803, 0.00003652, -0.00004501, 0.00002774, -0.00000872, -0.00002237, 0.00004446, -0.00004019, 0.00001328, 0.00001443, -0.00002278, 0.00000966, -0.00001682, 0.00001331, -0.00000987, 0.00000253, 0.00001677, -0.00002608, 0.00002495, -0.00002144, -0.00001606, 0.00004002, -0.00003527, 0.00000028, 0.00002942, -0.00002080, 0.00000030, 0.00000729, -0.00000688, -0.00000090, 0.00000159, -0.00000328, 0.00000683, 0.00000741, -0.00000628, -0.00000277, -0.00000102, 0.00000488, -0.00000853, -0.00001018, 0.00002121, 0.00000698, -0.00000538, 0.00000238, -0.00000316, 0.00000139, -0.00000172, -0.00000225, 0.00000216, 0.00000081, 0.00000363, 0.00000156, -0.00000072, -0.00000488, -0.00000296, 0.00000749, 0.00000655, 0.00000405, 0.00000237, -0.00000495, -0.00000558, 0.00000504, 0.00000903, -0.00000394, -0.00000274, 0.00000992, -0.00000162, -0.00001015, 0.00000030, 0.00000708, 0.00000409, -0.00000670, -0.00000339, 0.00000066, -0.00000493, 0.00000161, -0.00000768, -0.00001257, },
		{ 0.00000225, -0.00000165, -0.00000669, -0.00000267, -0.00000238, -0.00001665, 0.00000258, 0.00000439, -0.00000961, 0.00000796, -0.00000703, -0.00000156, 0.00001871, -0.00001123, -0.00000082, 0.00000356, -0.00000432, 0.00001035, -0.00001382, -0.00000378, -0.00000132, -0.00000613, 0.00001358, -0.00001778, 0.00000181, 0.00000279, -0.00002262, 0.00000973, -0.00002118, -0.00002291, -0.00000014, -0.00010865, -0.00018237, 0.00019790, 0.00053632, 0.00004076, -0.00025158, 0.00000456, -0.00004727, -0.00002726, 0.00001483, -0.00004060, 0.00002933, -0.00003584, 0.00000569, 0.00001562, -0.00004082, 0.00003299, -0.00002936, 0.00000793, 0.00001481, -0.00003710, 0.00003625, -0.00003062, 0.00000542, 0.00002612, -0.00004087, 0.00002841, -0.00000407, -0.00001076, 0.00000078, -0.00001195, 0.00001692, -0.00001982, 0.00001857, 0.00000145, -0.00001786, 0.00002744, -0.00003883, 0.00001019, 0.00002547, -0.00004342, 0.00002405, 0.00001326, -0.00002251, 0.00000984, 0.00000044, -0.00000612, 0.00000141, -0.00000072, -0.00000177, 0.00000982, 0.00000319, -0.00000726, -0.00000126, -0.00000012, 0.00000427, -0.00001295, -0.00000303, 0.00002346, -0.00000003, -0.00000331, 0.00000203, -0.00000405, 0.00000220, -0.00000254, -0.00000091, 0.00000175, 0.00000109, 0.00000404, 0.00000063, -0.00000138, -0.00000553, -0.00000062, 0.00000898, 0.00000507, 0.00000329, 0.00000146, -0.00000626, -0.00000380, 0.00000812, 0.00000622, -0.00000637, 0.00000143, 0.00000930, -0.00000561, -0.00000854, 0.00000260, 0.00000737, 0.00000151, -0.00000784, -0.00000136, -0.00000030, -0.00000349, 0.00000053, -0.00001198, -0.00000754, 0.00000849, },
		{ -0.00000225, -0.00000720, -0.00000070, -0.00000607, -0.00001509, 0.00000686, -0.00000029, -0.00000623, 0.00000740, -0.00001039, 0.00000587, 0.00001294, -0.00001276, 0.00000418, -0.00000088, 0.00000068, 0.00000605, -0.00001606, 0.00000166, -0.00000648, 0.00000086, 0.00000831, -0.00002031, 0.00001125, -0.00000937, -0.00001535, 0.00001070, -0.00003401, -0.00000665, -0.00001841, -0.00015232, -0.00010658, 0.00032125, 0.00044551, -0.00008171, -0.00019344, 0.00001326, -0.00005870, -0.00000421, -0.00000512, -0.00002267, 0.00001999, -0.00003879, 0.00002364, -0.00000921, -0.00002143, 0.00002563, -0.00003450, 0.00002600, -0.00000924, -0.00001757, 0.00002858, -0.00003819, 0.00002785, 0.00000113, -0.00002785, 0.00003059, -0.00001528, -0.00000054, -0.00001084, 0.00000090, 0.00000984, -0.00002025, 0.00002986, -0.00001623, -0.00000069, 0.00001652, -0.00004383, 0.00003423, -0.00000046, -0.00003466, 0.00003882, -0.00000740, -0.00001415, 0.00001314, -0.00000626, -0.00000249, 0.00000130, -0.00000239, 0.00000124, 0.00001085, -0.00000155, -0.00000648, -0.00000019, 0.00000086, 0.00000223, -0.00001567, 0.00000577, 0.00002167, -0.00000567, 0.00000005, 0.00000009, -0.00000394, 0.00000283, -0.00000321, 0.00000034, 0.00000103, 0.00000192, 0.00000395, -0.00000036, -0.00000199, -0.00000577, 0.00000226, 0.00000954, 0.00000312, 0.00000308, 0.00000032, -0.00000732, -0.00000078, 0.00000987, 0.00000216, -0.00000656, 0.00000543, 0.00000676, -0.00000836, -0.00000619, 0.00000440, 0.00000707, -0.00000141, -0.00000790, 0.00000041, -0.00000105, -0.00000214, -0.00000249, -0.00001391, -0.00000147, 0.00000850, 0.00000012, },
		{ -0.00000694, 0.00000043, -0.00000990, -0.00001146, 0.00000880, -0.00000412, -0.00000203, 0.00000438, -0.00001058, 0.00001139, 0.00000579, -0.00001063, 0.00000611, -0.00000351, 0.00000523, -0.00000069, -0.00001390, 0.00000368, -0.00000969, 0.00000794, -0.00000090, -0.00001595, 0.00001485, -0.00002070, -0.00000257, 0.00000174, -0.00003803, 0.00000866, -0.00005620, -0.00016795, -0.00000065, 0.00039193, 0.00032500, -0.00015374, -0.00013527, 0.00000661, -0.00005613, 0.00000648, -0.00001774, -0.00000599, 0.00000426, -0.00002897, 0.00002737, -0.00002569, -0.00000095, 0.00000928, -0.00002597, 0.00003062, -0.00002602, 0.00000394, 0.00001031, -0.00003025, 0.00003718, -0.00002039, -0.00000845, 0.00002189, -0.00001665, 0.00000229, -0.00001793, 0.00001506, -0.00000471, -0.00000951, 0.00003047, -0.00002825, 0.00001826, -0.00000461, -0.00003271, 0.00004607, -0.00002771, -0.00001171, 0.00003848, -0.00002385, -0.00000002, 0.00000892, -0.00000925, 0.00000149, -0.00000080, -0.00000236, 0.00000461, 0.00000969, -0.00000569, -0.00000438, 0.00000025, 0.00000185, -0.00000115, -0.00001565, 0.00001448, 0.00001618, -0.00000855, 0.00000328, -0.00000299, -0.00000251, 0.00000283, -0.00000352, 0.00000129, 0.00000032, 0.00000316, 0.00000318, -0.00000116, -0.00000260, -0.00000546, 0.00000545, 0.00000884, 0.00000124, 0.00000343, -0.00000136, -0.00000752, 0.00000299, 0.00000959, -0.00000193, -0.00000457, 0.00000813, 0.00000312, -0.00000961, -0.00000367, 0.00000578, 0.00000609, -0.00000437, -0.00000691, 0.00000182, -0.00000150, -0.00000164, -0.00000644, -0.00001293, 0.00000408, 0.00000660, -0.00000114, -0.00000309, },
		{ 0.00000036, -0.00001299, -0.00000665, 0.00000848, -0.00000623, 0.00000160, 0.00000018, -0.00000796, 0.00001387, -0.00000060, -0.00000650, 0.00000535, -0.00000351, 0.00000733, -0.00000733, -0.00000887, 0.00000203, -0.00000920, 0.00001198, -0.00001049, -0.00000644, 0.00001067, -0.00002599, 0.00000982, -0.00001439, -0.00002963, 0.00001175, -0.00010255, -0.00014504, 0.00010988, 0.00040728, 0.00020218, -0.00018429, -0.00008759, -0.00000468, -0.00004656, 0.00000782, -0.00002143, 0.00000366, -0.00000960, -0.00001364, 0.00001937, -0.00003006, 0.00001266, -0.00000739, -0.00000981, 0.00002260, -0.00003049, 0.00001861, -0.00000945, -0.00001197, 0.00003181, -0.00003125, 0.00000879, 0.00000848, -0.00001071, -0.00000330, -0.00001573, 0.00002309, -0.00001963, 0.00000865, 0.00001917, -0.00002874, 0.00003031, -0.00002772, -0.00000843, 0.00004013, -0.00004509, 0.00001666, 0.00002308, -0.00002925, 0.00001307, -0.00000038, -0.00000710, 0.00000324, -0.00000328, -0.00000058, 0.00000726, 0.00000674, -0.00000841, -0.00000166, 0.00000010, 0.00000263, -0.00000538, -0.00001241, 0.00002120, 0.00000832, -0.00000809, 0.00000496, -0.00000618, 0.00000005, 0.00000196, -0.00000328, 0.00000170, -0.00000001, 0.00000445, 0.00000176, -0.00000153, -0.00000336, -0.00000432, 0.00000849, 0.00000676, 0.00000006, 0.00000393, -0.00000357, -0.00000630, 0.00000657, 0.00000720, -0.00000478, -0.00000119, 0.00000896, -0.00000066, -0.00000954, -0.00000130, 0.00000681, 0.00000433, -0.00000694, -0.00000494, 0.00000279, -0.00000186, -0.00000241, -0.00001009, -0.00000933, 0.00000788, 0.00000372, -0.00000157, -0.00000422, -0.00000590, },
		{ -0.00001468, -0.00000168, 0.00000650, -0.00000646, 0.00000372, -0.00000381, -0.00000365, 0.00001332, -0.00000481, -0.00000222, 0.00000320, -0.00000143, 0.00000618, -0.00001161, -0.00000337, -0.00000187, -0.00000511, 0.00001116, -0.00001666, 0.00000418, 0.00000026, -0.00002262, 0.00001559, -0.00003084, -0.00001157, -0.00000483, -0.00014048, -0.00008592, 0.00020137, 0.00037826, 0.00009496, -0.00018686, -0.00005247, -0.00001487, -0.00003590, 0.00000501, -0.00001909, 0.00000558, -0.00001675, -0.00000023, 0.00000600, -0.00002424, 0.00001621, -0.00001755, 0.00000604, 0.00000762, -0.00002353, 0.00002191, -0.00002238, 0.00000806, 0.00001610, -0.00002965, 0.00001789, -0.00000275, -0.00000326, -0.00001358, -0.00000429, 0.00002045, -0.00002733, 0.00002675, 0.00000018, -0.00001686, 0.00002934, -0.00004291, 0.00002012, 0.00001801, -0.00004498, 0.00003912, -0.00000102, -0.00002166, 0.00001898, -0.00000989, -0.00000114, 0.00000153, -0.00000426, 0.00000207, 0.00000858, 0.00000270, -0.00000927, 0.00000094, -0.00000042, 0.00000289, -0.00000961, -0.00000620, 0.00002438, 0.00000005, -0.00000478, 0.00000418, -0.00000826, 0.00000304, 0.00000029, -0.00000245, 0.00000150, 0.00000037, 0.00000527, -0.00000001, -0.00000144, -0.00000431, -0.00000213, 0.00001070, 0.00000362, 0.00000007, 0.00000398, -0.00000585, -0.00000349, 0.00000883, 0.00000344, -0.00000559, 0.00000240, 0.00000799, -0.00000378, -0.00000859, 0.00000082, 0.00000746, 0.00000174, -0.00000867, -0.00000222, 0.00000319, -0.00000242, -0.00000436, -0.00001227, -0.00000406, 0.00000930, 0.00000084, -0.00000150, -0.00000551, -0.00000427, -0.00000096, },
		{ 0.00000262, 0.00000372, -0.00000523, 0.00000403, -0.00000654, 0.00000097, 0.00001061, -0.00000649, 0.00000094, 0.00000113, 0.00000134, 0.00000243, -0.00001251, 0.00000042, -0.00000574, 0.00000076, 0.00000580, -0.00001708, 0.00001140, -0.00001200, -0.00001196, 0.00001131, -0.00003989, 0.00000756, -0.00004056, -0.00015370, -0.00000454, 0.00025938, 0.00032248, 0.00001039, -0.00017327, -0.00002763, -0.00002276, -0.00002675, 0.00000204, -0.00001502, 0.00000274, -0.00001683, 0.00000685, -0.00000610, -0.00001363, 0.00001145, -0.00001867, 0.00001547, -0.00000699, -0.00001048, 0.00001489, -0.00002427, 0.00002185, -0.00000233, -0.00001902, 0.00001751, -0.00000748, -0.00000006, -0.00002220, 0.00001161, 0.00000763, -0.00002329, 0.00003699, -0.00001869, 0.00000262, 0.00001474, -0.00004299, 0.00004169, -0.00001184, -0.00002676, 0.00004651, -0.00002382, -0.00000484, 0.00001502, -0.00001472, 0.00000521, -0.00000281, -0.00000272, 0.00000434, 0.00000857, -0.00000170, -0.00000819, 0.00000275, -0.00000100, 0.00000239, -0.00001283, 0.00000199, 0.00002329, -0.00000655, -0.00000001, 0.00000093, -0.00000826, 0.00000552, -0.00000175, -0.00000120, 0.00000082, 0.00000155, 0.00000517, -0.00000161, -0.00000113, -0.00000529, 0.00000114, 0.00001137, 0.00000016, 0.00000132, 0.00000299, -0.00000737, 0.00000044, 0.00000894, -0.00000045, -0.00000434, 0.00000512, 0.00000586, -0.00000591, -0.00000724, 0.00000282, 0.00000751, -0.00000151, -0.00000910, 0.00000086, 0.00000286, -0.00000343, -0.00000697, -0.00001229, 0.00000157, 0.00000841, -0.00000125, -0.00000140, -0.00000667, -0.00000235, -0.00000332, -0.00001469, },
		{ 0.00000093, -0.00000323, 0.00000279, -0.00000757, 0.00000477, 0.00000696, -0.00000611, 0.00000256, 0.00000010, 0.00000325, -0.00000233, -0.00001044, 0.00000146, -0.00000763, 0.00000587, -0.00000181, -0.00001197, 0.00001228, -0.00002094, 0.00000119, -0.00000184, -0.00003679, 0.00001684, -0.00008595, -0.00013319, 0.00007914, 0.00028155, 0.00025667, -0.00005246, -0.00015083, -0.00001048, -0.00002926, -0.00001881, 0.00000012, -0.00001221, -0.00000086, -0.00001279, 0.00000725, -0.00001288, -0.00000387, 0.00000313, -0.00001266, 0.00001640, -0.00001572, 0.00000193, 0.00000274, -0.00001615, 0.00002517, -0.00001621, -0.00000565, 0.00001068, -0.00000558, -0.00000392, -0.00002349, 0.00002481, -0.00001006, -0.00000817, 0.00003484, -0.00002952, 0.00002135, -0.00000800, -0.00002686, 0.00004737, -0.00003742, 0.00000270, 0.00003575, -0.00003595, 0.00001347, 0.00000317, -0.00001232, 0.00000825, -0.00000735, 0.00000092, 0.00000536, 0.00000756, -0.00000580, -0.00000545, 0.00000330, -0.00000130, 0.00000097, -0.00001412, 0.00001058, 0.00001823, -0.00000993, 0.00000433, -0.00000389, -0.00000585, 0.00000661, -0.00000354, 0.00000010, 0.00000004, 0.00000329, 0.00000394, -0.00000252, -0.00000101, -0.00000587, 0.00000515, 0.00001001, -0.00000262, 0.00000334, 0.00000081, -0.00000730, 0.00000442, 0.00000680, -0.00000316, -0.00000176, 0.00000637, 0.00000340, -0.00000717, -0.00000574, 0.00000478, 0.00000664, -0.00000498, -0.00000793, 0.00000374, 0.00000169, -0.00000496, -0.00000952, -0.00001001, 0.00000628, 0.00000581, -0.00000217, -0.00000172, -0.00000735, -0.00000056, -0.00000632, -0.00001307, 0.00000571, },
		{ -0.00000116, 0.00000061, -0.00000700, 0.00000721, 0.00000346, -0.00000458, 0.00000296, 0.00000022, 0.00000341, -0.00000637, -0.00000684, 0.00000008, -0.00000675, 0.00000813, -0.00000857, -0.00000388, 0.00000675, -0.00002298, 0.00001085, -0.00001841, -0.00002246, 0.00000807, -0.00012566, -0.00008091, 0.00014662, 0.00027554, 0.00019170, -0.00009774, -0.00012286, -0.00000009, -0.00003486, -0.00001079, -0.00000175, -0.00001108, -0.00000263, -0.00000844, 0.00000366, -0.00001392, 0.00000163, -0.00000390, -0.00000395, 0.00001075, -0.00001680, 0.00000876, -0.00000814, -0.00000307, 0.00001868, -0.00002152, 0.00000447, 0.00000263, -0.00000057, -0.00001344, -0.00001527, 0.00002913, -0.00002488, 0.00001246, 0.00002104, -0.00002785, 0.00003109, -0.00002950, -0.00000025, 0.00003457, -0.00004812, 0.00003170, 0.00001120, -0.00003262, 0.00002499, -0.00001090, -0.00000376, 0.00000590, -0.00000927, 0.00000503, 0.00000509, 0.00000582, -0.00000900, -0.00000161, 0.00000242, -0.00000107, -0.00000131, -0.00001289, 0.00001780, 0.00001050, -0.00000954, 0.00000654, -0.00000872, -0.00000153, 0.00000585, -0.00000452, 0.00000104, -0.00000035, 0.00000504, 0.00000177, -0.00000245, -0.00000150, -0.00000545, 0.00000910, 0.00000665, -0.00000380, 0.00000522, -0.00000211, -0.00000526, 0.00000711, 0.00000316, -0.00000384, 0.00000098, 0.00000625, 0.00000124, -0.00000791, -0.00000402, 0.00000661, 0.00000459, -0.00000798, -0.00000520, 0.00000582, -0.00000028, -0.00000687, -0.00001121, -0.00000589, 0.00000909, 0.00000247, -0.00000198, -0.00000276, -0.00000721, 0.00000069, -0.00000948, -0.00001008, 0.00000739, -0.00000134, },
		{ -0.00000189, -0.00000527, 0.00000822, 0.00000075, -0.00000272, 0.00000279, 0.00000094, 0.00000180, -0.00000851, -0.00000340, -0.00000244, -0.00000370, 0.00000689, -0.00001201, 0.00000359, -0.00000249, -0.00001776, 0.00001267, -0.00003120, -0.00000359, -0.00002011, -0.00014420, -0.00000906, 0.00018705, 0.00025370, 0.00013143, -0.00012867, -0.00009109, 0.00000306, -0.00003848, -0.00000240, -0.00000505, -0.00001015, -0.00000243, -0.00000620, -0.00000046, -0.00001144, 0.00000260, -0.00000691, 0.00000306, 0.00000269, -0.00001217, 0.00000866, -0.00001315, 0.00000875, 0.00000680, -0.00001856, 0.00000823, -0.00000212, 0.00000262, -0.00002388, 0.00000009, 0.00002226, -0.00003024, 0.00003065, 0.00000124, -0.00001472, 0.00002749, -0.00004058, 0.00002644, 0.00000831, -0.00003930, 0.00004855, -0.00001711, -0.00001571, 0.00002460, -0.00002048, 0.00000675, -0.00000112, -0.00000691, 0.00000759, 0.00000426, 0.00000341, -0.00001069, 0.00000246, 0.00000030, -0.00000021, -0.00000420, -0.00000905, 0.00002213, 0.00000206, -0.00000594, 0.00000560, -0.00001188, 0.00000352, 0.00000339, -0.00000438, 0.00000134, 0.00000004, 0.00000610, -0.00000074, -0.00000152, -0.00000275, -0.00000352, 0.00001193, 0.00000198, -0.00000295, 0.00000598, -0.00000481, -0.00000161, 0.00000748, -0.00000058, -0.00000253, 0.00000285, 0.00000539, -0.00000045, -0.00000844, -0.00000188, 0.00000794, 0.00000138, -0.00000972, -0.00000135, 0.00000656, -0.00000287, -0.00000886, -0.00001144, -0.00000079, 0.00000955, -0.00000055, -0.00000114, -0.00000447, -0.00000604, 0.00000095, -0.00001228, -0.00000613, 0.00000767, -0.00000280, 0.00000046, },
		{ -0.00000289, 0.00000810, -0.00000099, -0.00000104, 0.00000256, 0.00000150, -0.00000084, -0.00000853, -0.00000129, -0.00000459, 0.00000003, 0.00000304, -0.00001128, 0.00000744, -0.00001139, -0.00000816, 0.00000594, -0.00003462, 0.00001005, -0.00006082, -0.00013195, 0.00006448, 0.00019974, 0.00022705, 0.00007498, -0.00014595, -0.00005801, -0.00000132, -0.00003807, 0.00000501, -0.00001007, -0.00000768, -0.00000191, -0.00000622, -0.00000280, -0.00000839, 0.00000105, -0.00000595, 0.00000600, -0.00000377, -0.00000574, 0.00000375, -0.00001116, 0.00001473, -0.00000479, -0.00001092, 0.00000614, -0.00000183, 0.00000040, -0.00002964, 0.00001681, 0.00000676, -0.00002349, 0.00003954, -0.00001661, 0.00000395, 0.00001203, -0.00003629, 0.00004258, -0.00002071, -0.00001432, 0.00004647, -0.00003770, 0.00000722, 0.00001258, -0.00002095, 0.00001391, -0.00000957, -0.00000073, 0.00000731, 0.00000382, 0.00000017, -0.00001032, 0.00000569, -0.00000243, 0.00000115, -0.00000728, -0.00000306, 0.00002271, -0.00000504, -0.00000068, 0.00000158, -0.00001210, 0.00000782, -0.00000005, -0.00000323, 0.00000104, 0.00000131, 0.00000592, -0.00000279, -0.00000027, -0.00000445, 0.00000002, 0.00001261, -0.00000275, -0.00000033, 0.00000505, -0.00000612, 0.00000252, 0.00000531, -0.00000294, -0.00000007, 0.00000337, 0.00000456, -0.00000193, -0.00000879, 0.00000082, 0.00000821, -0.00000257, -0.00000959, 0.00000283, 0.00000569, -0.00000574, -0.00001055, -0.00000993, 0.00000416, 0.00000785, -0.00000243, -0.00000039, -0.00000647, -0.00000386, -0.00000010, -0.00001413, -0.00000174, 0.00000673, -0.00000335, 0.00000132, -0.00000419, },
		{ 0.00000721, -0.00000184, 0.00000033, 0.00000246, 0.00000132, -0.00000348, -0.00000703, -0.00000085, -0.00000539, 0.00000287, -0.00000155, -0.00000742, 0.00000655, -0.00001642, 0.00000121, -0.00000608, -0.00002754, 0.00000980, -0.00010092, -0.00008899, 0.00012273, 0.00019257, 0.00020131, 0.00002054, -0.00014858, -0.00002771, -0.00001171, -0.00003215, 0.00000908, -0.00001536, -0.00000319, -0.00000295, -0.00000693, -0.00000314, -0.00000666, -0.00000049, -0.00000304, 0.00000512, -0.00000659, -0.00000110, -0.00000207, -0.00000443, 0.00001362, -0.00001183, -0.00000327, 0.00000140, 0.00000198, -0.00000768, -0.00002701, 0.00002832, -0.00001117, -0.00000698, 0.00003627, -0.00002577, 0.00002007, -0.00000860, -0.00001815, 0.00004181, -0.00004057, 0.00001678, 0.00002638, -0.00004248, 0.00002615, -0.00000557, -0.00001195, 0.00001385, -0.00001518, 0.00000681, 0.00000439, 0.00000426, -0.00000387, -0.00000756, 0.00000701, -0.00000481, 0.00000255, -0.00000995, 0.00000413, 0.00001949, -0.00000918, 0.00000421, -0.00000437, -0.00000904, 0.00001008, -0.00000345, -0.00000150, 0.00000043, 0.00000318, 0.00000433, -0.00000369, 0.00000054, -0.00000588, 0.00000464, 0.00001059, -0.00000616, 0.00000308, 0.00000257, -0.00000529, 0.00000556, 0.00000138, -0.00000298, 0.00000220, 0.00000283, 0.00000424, -0.00000368, -0.00000858, 0.00000390, 0.00000693, -0.00000641, -0.00000741, 0.00000640, 0.00000330, -0.00000852, -0.00001151, -0.00000676, 0.00000788, 0.00000469, -0.00000282, -0.00000043, -0.00000809, -0.00000106, -0.00000256, -0.00001452, 0.00000249, 0.00000488, -0.00000306, 0.00000125, -0.00000591, -0.00000028, },
		{ -0.00000201, 0.00000139, 0.00000236, 0.00000031, -0.00000530, -0.00000496, -0.00000165, -0.00000470, 0.00000393, -0.00000506, -0.00000260, 0.00000197, -0.00001618, 0.00000618, -0.00001787, -0.00001396, -0.00000777, -0.00012581, -0.00002512, 0.00015534, 0.00017730, 0.00017616, -0.00003153, -0.00013613, -0.00000500, -0.00002445, -0.00002120, 0.00000803, -0.00001846, 0.00000201, -0.00000621, -0.00000648, -0.00000279, -0.00000636, -0.00000049, -0.00000059, 0.00000241, -0.00000616, 0.00000012, -0.00000536, 0.00000306, 0.00000734, -0.00001294, 0.00000099, -0.00000226, 0.00000575, -0.00001882, -0.00001594, 0.00003025, -0.00002443, 0.00001316, 0.00002293, -0.00002354, 0.00002703, -0.00002553, 0.00000643, 0.00002484, -0.00004315, 0.00004139, -0.00000350, -0.00003005, 0.00003306, -0.00002170, 0.00000243, 0.00000609, -0.00001473, 0.00001237, 0.00000042, 0.00000519, -0.00000819, -0.00000272, 0.00000573, -0.00000577, 0.00000332, -0.00001153, 0.00001124, 0.00001334, -0.00000964, 0.00000688, -0.00001038, -0.00000337, 0.00000961, -0.00000584, 0.00000021, -0.00000000, 0.00000503, 0.00000169, -0.00000319, 0.00000031, -0.00000612, 0.00000924, 0.00000611, -0.00000717, 0.00000596, -0.00000053, -0.00000236, 0.00000614, -0.00000270, -0.00000075, 0.00000313, 0.00000209, 0.00000432, -0.00000597, -0.00000733, 0.00000675, 0.00000398, -0.00000914, -0.00000353, 0.00000850, -0.00000021, -0.00001085, -0.00001139, -0.00000243, 0.00000961, 0.00000115, -0.00000193, -0.00000167, -0.00000859, 0.00000162, -0.00000616, -0.00001309, 0.00000589, 0.00000258, -0.00000212, 0.00000025, -0.00000678, 0.00000222, 0.00000587, },
		{ 0.00000218, 0.00000202, -0.00000125, -0.00000594, -0.00000320, -0.00000291, -0.00000303, 0.00000325, -0.00000647, 0.00000098, -0.00000383, -0.00001177, 0.00000504, -0.00002428, -0.00000140, -0.00003897, -0.00012487, 0.00004368, 0.00016134, 0.00016412, 0.00014749, -0.00007765, -0.00011072, 0.00000657, -0.00003477, -0.00000798, 0.00000195, -0.00001740, 0.00000555, -0.00001063, -0.00000407, -0.00000334, -0.00000627, 0.00000090, 0.00000006, 0.00000011, -0.00000443, -0.00000138, -0.00000467, 0.00000771, -0.00000041, -0.00000961, 0.00000097, -0.00000249, 0.00000595, -0.00002837, 0.00000002, 0.00002213, -0.00002803, 0.00002977, 0.00000532, -0.00001213, 0.00002260, -0.00003168, 0.00002741, -0.00000087, -0.00002756, 0.00004962, -0.00003110, -0.00000614, 0.00002527, -0.00002861, 0.00001555, -0.00000616, -0.00000769, 0.00001333, -0.00000236, 0.00000543, -0.00001162, 0.00000315, 0.00000194, -0.00000456, 0.00000272, -0.00001133, 0.00001692, 0.00000576, -0.00000679, 0.00000618, -0.00001448, 0.00000341, 0.00000655, -0.00000659, 0.00000139, 0.00000016, 0.00000612, -0.00000121, -0.00000157, -0.00000113, -0.00000449, 0.00001240, 0.00000026, -0.00000551, 0.00000719, -0.00000291, 0.00000162, 0.00000376, -0.00000512, 0.00000268, 0.00000225, 0.00000214, 0.00000415, -0.00000859, -0.00000468, 0.00000852, -0.00000019, -0.00000990, 0.00000118, 0.00000861, -0.00000423, -0.00001245, -0.00000996, 0.00000224, 0.00000909, -0.00000168, -0.00000048, -0.00000401, -0.00000744, 0.00000325, -0.00001022, -0.00000984, 0.00000790, 0.00000039, -0.00000090, -0.00000155, -0.00000663, 0.00000433, 0.00000435, -0.00000165, },
		{ 0.00000128, -0.00000285, -0.00000554, -0.00000223, -0.00000395, -0.00000115, 0.00000156, -0.00000592, 0.00000215, -0.00000833, -0.00000600, -0.00000093, -0.00002310, 0.00000245, -0.00007398, -0.00009547, 0.00010098, 0.00014819, 0.00015783, 0.00011108, -0.00011222, -0.00007764, 0.00000639, -0.00003875, 0.00000354, -0.00000689, -0.00001194, 0.00000541, -0.00001406, -0.00000051, -0.00000544, -0.00000497, 0.00000236, -0.00000085, -0.00000063, -0.00000345, -0.00000346, -0.00000092, 0.00000789, -0.00000617, -0.00000489, -0.00000185, 0.00000061, 0.00000091, -0.00003213, 0.00001535, 0.00000747, -0.00002093, 0.00003763, -0.00000978, 0.00000273, 0.00000963, -0.00002483, 0.00003652, -0.00002438, -0.00000048, 0.00003837, -0.00004562, 0.00001892, 0.00000647, -0.00002331, 0.00002125, -0.00001738, 0.00000340, 0.00000912, -0.00000230, 0.00000372, -0.00001271, 0.00000829, -0.00000333, -0.00000113, 0.00000030, -0.00000887, 0.00002000, -0.00000142, -0.00000195, 0.00000210, -0.00001515, 0.00000939, 0.00000181, -0.00000562, 0.00000181, 0.00000109, 0.00000591, -0.00000344, 0.00000037, -0.00000334, -0.00000090, 0.00001293, -0.00000531, -0.00000184, 0.00000636, -0.00000337, 0.00000492, -0.00000089, -0.00000461, 0.00000561, 0.00000009, 0.00000352, 0.00000288, -0.00001082, -0.00000081, 0.00000840, -0.00000462, -0.00000832, 0.00000559, 0.00000673, -0.00000815, -0.00001312, -0.00000718, 0.00000627, 0.00000669, -0.00000304, 0.00000056, -0.00000678, -0.00000461, 0.00000302, -0.00001372, -0.00000524, 0.00000822, -0.00000113, 0.00000010, -0.00000383, -0.00000542, 0.00000580, 0.00000290, -0.00000091, 0.00000264, },
		{ -0.00000408, -0.00000452, -0.00000209, -0.00000439, 0.00000036, -0.00000024, -0.00000434, 0.00000104, -0.00001013, -0.00000182, -0.00000826, -0.00001612, -0.00000676, -0.00010035, -0.00004389, 0.00013538, 0.00012776, 0.00015680, 0.00006582, -0.00013029, -0.00004415, -0.00000273, -0.00003489, 0.00000985, -0.00001497, -0.00000393, 0.00000113, -0.00001450, 0.00000238, -0.00000837, -0.00000196, 0.00000248, -0.00000204, -0.00000011, -0.00000414, -0.00000397, 0.00000348, 0.00000430, -0.00000819, -0.00000162, -0.00000483, 0.00000493, -0.00000836, -0.00002824, 0.00002497, -0.00000798, -0.00000612, 0.00003533, -0.00001753, 0.00001458, -0.00000547, -0.00000845, 0.00003075, -0.00003593, 0.00002674, 0.00001249, -0.00004187, 0.00003476, -0.00001498, -0.00000828, 0.00001668, -0.00002224, 0.00001401, 0.00000157, 0.00000072, -0.00000049, -0.00001039, 0.00001081, -0.00000823, 0.00000363, -0.00000381, -0.00000409, 0.00001982, -0.00000655, 0.00000299, -0.00000427, -0.00001190, 0.00001289, -0.00000322, -0.00000338, 0.00000158, 0.00000257, 0.00000431, -0.00000429, 0.00000170, -0.00000534, 0.00000396, 0.00001034, -0.00000899, 0.00000249, 0.00000400, -0.00000149, 0.00000581, -0.00000602, -0.00000108, 0.00000648, -0.00000205, 0.00000593, -0.00000001, -0.00001165, 0.00000351, 0.00000605, -0.00000807, -0.00000471, 0.00000866, 0.00000325, -0.00001145, -0.00001268, -0.00000333, 0.00000879, 0.00000327, -0.00000277, 0.00000038, -0.00000897, -0.00000073, 0.00000060, -0.00001558, -0.00000017, 0.00000700, -0.00000164, 0.00000034, -0.00000604, -0.00000331, 0.00000648, 0.00000178, -0.00000001, 0.00000270, 0.00000020, },
		{ -0.00000333, -0.00000260, -0.00000418, 0.00000131, -0.00000161, -0.00000283, -0.00000121, -0.00000942, -0.00000086, -0.00001327, -0.00000827, -0.00002789, -0.00010767, 0.00001705, 0.00014371, 0.00011133, 0.00015491, 0.00001502, -0.00012998, -0.00001719, -0.00001557, -0.00002472, 0.00000947, -0.00001908, 0.00000352, -0.00000580, -0.00001134, 0.00000287, -0.00001047, 0.00000197, 0.00000077, -0.00000214, 0.00000035, -0.00000590, -0.00000201, 0.00000607, -0.00000083, -0.00000682, -0.00000118, -0.00000576, 0.00000760, -0.00001882, -0.00001784, 0.00002629, -0.00001871, 0.00001101, 0.00002529, -0.00001671, 0.00001922, -0.00001591, 0.00001005, 0.00001355, -0.00003113, 0.00004291, -0.00001742, -0.00002229, 0.00003525, -0.00002990, 0.00000975, 0.00000362, -0.00001810, 0.00001959, -0.00000587, 0.00000512, -0.00000644, -0.00000465, 0.00000938, -0.00001069, 0.00000807, -0.00000872, 0.00000244, 0.00001636, -0.00000855, 0.00000610, -0.00001099, -0.00000540, 0.00001290, -0.00000712, -0.00000064, 0.00000105, 0.00000410, 0.00000171, -0.00000356, 0.00000172, -0.00000606, 0.00000872, 0.00000511, -0.00000973, 0.00000593, 0.00000143, 0.00000201, 0.00000333, -0.00000938, 0.00000423, 0.00000458, -0.00000270, 0.00000829, -0.00000435, -0.00001032, 0.00000702, 0.00000187, -0.00000945, 0.00000003, 0.00000973, -0.00000114, -0.00001379, -0.00001102, 0.00000101, 0.00000928, -0.00000003, -0.00000133, -0.00000137, -0.00000954, 0.00000305, -0.00000368, -0.00001509, 0.00000427, 0.00000483, -0.00000117, -0.00000053, -0.00000756, -0.00000070, 0.00000641, 0.00000116, 0.00000079, 0.00000234, -0.00000105, -0.00000471, },
		{ -0.00000349, -0.00000340, 0.00000176, -0.00000246, -0.00000199, -0.00000330, -0.00000759, -0.00000273, -0.00001396, -0.00000531, -0.00005445, -0.00009138, 0.00007257, 0.00013103, 0.00010547, 0.00014507, -0.00003454, -0.00011362, -0.00000110, -0.00002657, -0.00001202, 0.00000355, -0.00001775, 0.00000744, -0.00001265, -0.00000569, 0.00000039, -0.00001020, 0.00000519, -0.00000206, -0.00000071, -0.00000050, -0.00000720, 0.00000172, 0.00000559, -0.00000499, -0.00000386, -0.00000315, -0.00000393, 0.00000657, -0.00002686, -0.00000441, 0.00001993, -0.00002147, 0.00002506, 0.00001218, -0.00000969, 0.00001623, -0.00001759, 0.00002282, -0.00000695, -0.00001265, 0.00004167, -0.00003951, 0.00000441, 0.00002093, -0.00003216, 0.00002279, -0.00001230, -0.00000622, 0.00001770, -0.00000976, 0.00000831, -0.00001208, 0.00000307, 0.00000400, -0.00000924, 0.00001023, -0.00001291, 0.00000954, 0.00001044, -0.00000726, 0.00000606, -0.00001589, 0.00000265, 0.00000948, -0.00000887, 0.00000175, 0.00000064, 0.00000510, -0.00000109, -0.00000166, 0.00000032, -0.00000476, 0.00001182, -0.00000135, -0.00000751, 0.00000747, 0.00000009, 0.00000545, -0.00000214, -0.00000919, 0.00000918, 0.00000047, -0.00000092, 0.00000919, -0.00000912, -0.00000670, 0.00000850, -0.00000303, -0.00000824, 0.00000467, 0.00000868, -0.00000571, -0.00001497, -0.00000813, 0.00000501, 0.00000779, -0.00000221, 0.00000030, -0.00000430, -0.00000792, 0.00000542, -0.00000879, -0.00001215, 0.00000714, 0.00000256, -0.00000020, -0.00000250, -0.00000789, 0.00000185, 0.00000583, 0.00000104, 0.00000123, 0.00000169, -0.00000229, -0.00000471, -0.00000234, },
		{ -0.00000215, 0.00000178, -0.00000304, -0.00000179, -0.00000455, -0.00000619, -0.00000556, -0.00001109, -0.00001108, -0.00007705, -0.00005428, 0.00011082, 0.00010772, 0.00011040, 0.00012268, -0.00007516, -0.00008711, 0.00000346, -0.00003181, -0.00000106, -0.00000477, -0.00001188, 0.00000642, -0.00001666, 0.00000014, -0.00000410, -0.00000708, 0.00000621, -0.00000450, 0.00000154, -0.00000301, -0.00000657, 0.00000544, 0.00000239, -0.00000669, -0.00000139, -0.00000595, -0.00000024, 0.00000151, -0.00002990, 0.00000785, 0.00000911, -0.00001618, 0.00003263, 0.00000068, -0.00000070, 0.00000871, -0.00001077, 0.00002506, -0.00002163, 0.00001108, 0.00002397, -0.00004563, 0.00002712, -0.00000152, -0.00002131, 0.00002517, -0.00002410, 0.00000871, 0.00000910, -0.00000830, 0.00000797, -0.00001493, 0.00001020, -0.00000374, -0.00000370, 0.00000867, -0.00001468, 0.00001561, 0.00000350, -0.00000351, 0.00000252, -0.00001722, 0.00001006, 0.00000364, -0.00000811, 0.00000318, 0.00000068, 0.00000511, -0.00000323, 0.00000055, -0.00000195, -0.00000144, 0.00001206, -0.00000719, -0.00000330, 0.00000700, 0.00000089, 0.00000680, -0.00000881, -0.00000492, 0.00001157, -0.00000417, 0.00000316, 0.00000750, -0.00001281, -0.00000160, 0.00000722, -0.00000712, -0.00000472, 0.00000807, 0.00000588, -0.00000989, -0.00001485, -0.00000423, 0.00000784, 0.00000494, -0.00000277, 0.00000107, -0.00000743, -0.00000431, 0.00000537, -0.00001329, -0.00000740, 0.00000803, 0.00000101, 0.00000054, -0.00000512, -0.00000686, 0.00000386, 0.00000514, 0.00000120, 0.00000121, 0.00000092, -0.00000338, -0.00000414, -0.00000182, -0.00000445, },
		{ 0.00000133, -0.00000354, -0.00000181, -0.00000510, -0.00000601, -0.00000723, -0.00000771, -0.00002553, -0.00008687, -0.00000533, 0.00012633, 0.00008513, 0.00012099, 0.00008767, -0.00010100, -0.00005786, -0.00000072, -0.00003019, 0.00000508, -0.00001188, -0.00000415, 0.00000112, -0.00001639, 0.00000378, -0.00000857, -0.00000202, 0.00000453, -0.00000517, 0.00000316, -0.00000629, -0.00000356, 0.00000731, -0.00000193, -0.00000586, -0.00000072, -0.00000786, 0.00000350, -0.00000621, -0.00002736, 0.00001574, -0.00000195, -0.00000536, 0.00003314, -0.00000639, 0.00000639, 0.00000132, 0.00000040, 0.00001702, -0.00002448, 0.00002974, -0.00000257, -0.00003443, 0.00003701, -0.00002240, -0.00000248, 0.00001595, -0.00002653, 0.00002062, -0.00000253, -0.00000222, 0.00000333, -0.00001319, 0.00001395, -0.00001102, 0.00000444, 0.00000312, -0.00001282, 0.00001905, -0.00000268, 0.00000109, -0.00000366, -0.00001423, 0.00001474, -0.00000290, -0.00000525, 0.00000339, 0.00000124, 0.00000401, -0.00000410, 0.00000214, -0.00000408, 0.00000307, 0.00000913, -0.00001079, 0.00000137, 0.00000541, 0.00000365, 0.00000460, -0.00001408, 0.00000232, 0.00001010, -0.00000720, 0.00000823, 0.00000301, -0.00001400, 0.00000346, 0.00000339, -0.00000901, 0.00000008, 0.00000952, 0.00000197, -0.00001326, -0.00001330, 0.00000018, 0.00000887, 0.00000175, -0.00000188, 0.00000029, -0.00000950, 0.00000032, 0.00000260, -0.00001585, -0.00000206, 0.00000716, 0.00000060, 0.00000032, -0.00000761, -0.00000470, 0.00000507, 0.00000470, 0.00000126, 0.00000086, 0.00000018, -0.00000426, -0.00000314, -0.00000197, -0.00000516, -0.00000053, },
		{ -0.00000384, -0.00000171, -0.00000560, -0.00000667, -0.00000669, -0.00000765, -0.00004457, -0.00007898, 0.00004362, 0.00012085, 0.00007163, 0.00012961, 0.00004435, -0.00010991, -0.00003245, -0.00000904, -0.00002333, 0.00000560, -0.00001517, 0.00000223, -0.00000607, -0.00001227, 0.00000394, -0.00001096, 0.00000308, 0.00000093, -0.00000362, 0.00000289, -0.00000876, 0.00000101, 0.00000642, -0.00000554, -0.00000358, -0.00000198, -0.00000791, 0.00000554, -0.00001432, -0.00002053, 0.00001803, -0.00000972, 0.00000733, 0.00002828, -0.00000868, 0.00000983, -0.00000225, 0.00001020, 0.00000356, -0.00001506, 0.00003552, -0.00002713, -0.00001158, 0.00003107, -0.00003324, 0.00001600, -0.00000075, -0.00001837, 0.00002469, -0.00001228, 0.00000554, -0.00000432, -0.00000676, 0.00001246, -0.00001481, 0.00001244, -0.00000518, -0.00000714, 0.00001885, -0.00000653, 0.00000463, -0.00001065, -0.00000746, 0.00001536, -0.00000824, -0.00000129, 0.00000256, 0.00000217, 0.00000207, -0.00000355, 0.00000250, -0.00000503, 0.00000736, 0.00000373, -0.00001131, 0.00000503, 0.00000409, 0.00000698, -0.00000130, -0.00001552, 0.00001010, 0.00000498, -0.00000697, 0.00001230, -0.00000330, -0.00001204, 0.00000675, -0.00000180, -0.00000797, 0.00000478, 0.00000893, -0.00000237, -0.00001556, -0.00001033, 0.00000434, 0.00000799, -0.00000074, -0.00000033, -0.00000204, -0.00000952, 0.00000451, -0.00000223, -0.00001567, 0.00000255, 0.00000535, 0.00000116, -0.00000125, -0.00000915, -0.00000203, 0.00000561, 0.00000462, 0.00000087, 0.00000047, -0.00000059, -0.00000479, -0.00000191, -0.00000286, -0.00000524, 0.00000123, 0.00000027, },
		{ -0.00000159, -0.00000661, -0.00000703, -0.00000469, -0.00001347, -0.00006159, -0.00005396, 0.00008187, 0.00010166, 0.00007034, 0.00012944, -0.00000033, -0.00010379, -0.00001484, -0.00001695, -0.00001438, 0.00000190, -0.00001400, 0.00000497, -0.00001227, -0.00000623, 0.00000093, -0.00001015, 0.00000631, -0.00000291, -0.00000062, 0.00000041, -0.00000898, 0.00000544, 0.00000310, -0.00000720, -0.00000131, -0.00000437, -0.00000619, 0.00000501, -0.00002070, -0.00001180, 0.00001552, -0.00001243, 0.00001867, 0.00002072, -0.00000761, 0.00001030, -0.00000096, 0.00001419, -0.00000843, 0.00000154, 0.00002640, -0.00004033, 0.00001303, 0.00001295, -0.00003026, 0.00002633, -0.00001769, -0.00000288, 0.00001942, -0.00001614, 0.00001114, -0.00001200, 0.00000245, 0.00000581, -0.00001305, 0.00001716, -0.00001374, 0.00000124, 0.00001499, -0.00000717, 0.00000546, -0.00001620, 0.00000137, 0.00001179, -0.00001091, 0.00000249, 0.00000118, 0.00000312, -0.00000017, -0.00000193, 0.00000154, -0.00000417, 0.00000996, -0.00000262, -0.00000896, 0.00000694, 0.00000429, 0.00000881, -0.00000941, -0.00001188, 0.00001554, -0.00000205, -0.00000303, 0.00001349, -0.00000961, -0.00000739, 0.00000701, -0.00000655, -0.00000425, 0.00000811, 0.00000672, -0.00000664, -0.00001654, -0.00000611, 0.00000740, 0.00000566, -0.00000184, 0.00000084, -0.00000518, -0.00000717, 0.00000681, -0.00000778, -0.00001287, 0.00000545, 0.00000363, 0.00000200, -0.00000395, -0.00000922, 0.00000048, 0.00000587, 0.00000466, -0.00000006, 0.00000036, -0.00000153, -0.00000480, -0.00000083, -0.00000439, -0.00000432, 0.00000264, -0.00000137, -0.00000363, },
		{ -0.00000802, -0.00000605, -0.00000342, -0.00002511, -0.00006997, -0.00001745, 0.00010295, 0.00007835, 0.00007911, 0.00011687, -0.00003928, -0.00008739, -0.00000586, -0.00002150, -0.00000651, -0.00000341, -0.00000966, 0.00000352, -0.00001531, -0.00000078, -0.00000356, -0.00000647, 0.00000662, -0.00000530, 0.00000229, -0.00000346, -0.00000645, 0.00000802, -0.00000130, -0.00000674, -0.00000015, -0.00000677, -0.00000350, 0.00000208, -0.00002420, -0.00000350, 0.00001022, -0.00001015, 0.00002673, 0.00001270, -0.00000505, 0.00000991, 0.00000333, 0.00001129, -0.00001331, 0.00001726, 0.00000668, -0.00003795, 0.00002957, -0.00000904, -0.00001569, 0.00002454, -0.00002771, 0.00001368, 0.00000719, -0.00001278, 0.00001167, -0.00001632, 0.00001105, -0.00000379, -0.00000563, 0.00001631, -0.00001964, 0.00001016, 0.00000858, -0.00000483, 0.00000284, -0.00001828, 0.00000981, 0.00000513, -0.00001027, 0.00000497, -0.00000013, 0.00000371, -0.00000210, 0.00000009, -0.00000024, -0.00000161, 0.00000987, -0.00000815, -0.00000482, 0.00000732, 0.00000631, 0.00000727, -0.00001703, -0.00000378, 0.00001646, -0.00000832, 0.00000353, 0.00001086, -0.00001386, -0.00000164, 0.00000406, -0.00000908, 0.00000101, 0.00000934, 0.00000359, -0.00001054, -0.00001594, -0.00000117, 0.00000868, 0.00000279, -0.00000150, 0.00000075, -0.00000794, -0.00000302, 0.00000638, -0.00001243, -0.00000834, 0.00000637, 0.00000278, 0.00000223, -0.00000709, -0.00000781, 0.00000237, 0.00000629, 0.00000434, -0.00000126, 0.00000061, -0.00000274, -0.00000409, -0.00000038, -0.00000609, -0.00000227, 0.00000310, -0.00000320, -0.00000268, -0.00000198, },
		{ -0.00000365, -0.00000536, -0.00003963, -0.00006562, 0.00002199, 0.00010599, 0.00005940, 0.00009234, 0.00009234, -0.00006740, -0.00006645, -0.00000382, -0.00002189, -0.00000167, -0.00000777, -0.00000447, -0.00000094, -0.00001457, 0.00000220, -0.00000740, -0.00000144, 0.00000428, -0.00000533, 0.00000360, -0.00000707, -0.00000190, 0.00000782, -0.00000519, -0.00000489, -0.00000042, -0.00000839, -0.00000084, -0.00000238, -0.00002476, 0.00000289, 0.00000433, -0.00000409, 0.00003099, 0.00000544, -0.00000205, 0.00001061, 0.00000721, 0.00000405, -0.00000934, 0.00002482, -0.00001521, -0.00002232, 0.00003237, -0.00002569, 0.00000357, 0.00001197, -0.00002676, 0.00002473, -0.00000688, -0.00000402, 0.00000651, -0.00001508, 0.00001561, -0.00001269, 0.00000528, 0.00000944, -0.00002059, 0.00001708, 0.00000157, -0.00000078, -0.00000276, -0.00001586, 0.00001545, -0.00000259, -0.00000671, 0.00000549, -0.00000081, 0.00000367, -0.00000322, 0.00000179, -0.00000202, 0.00000186, 0.00000699, -0.00001144, -0.00000032, 0.00000717, 0.00000929, 0.00000161, -0.00002117, 0.00000644, 0.00001229, -0.00001131, 0.00001045, 0.00000490, -0.00001465, 0.00000313, -0.00000106, -0.00000830, 0.00000615, 0.00000848, 0.00000012, -0.00001384, -0.00001355, 0.00000366, 0.00000800, 0.00000042, -0.00000032, -0.00000086, -0.00000915, 0.00000164, 0.00000326, -0.00001490, -0.00000343, 0.00000578, 0.00000302, 0.00000117, -0.00000972, -0.00000543, 0.00000362, 0.00000702, 0.00000321, -0.00000216, 0.00000097, -0.00000411, -0.00000263, -0.00000102, -0.00000723, 0.00000057, 0.00000214, -0.00000449, -0.00000122, -0.00000337, -0.00000903, },
		{ -0.00001191, -0.00005233, -0.00004830, 0.00005586, 0.00009496, 0.00004981, 0.00010349, 0.00005958, -0.00008263, -0.00004599, -0.00000586, -0.00001904, -0.00000018, -0.00000970, -0.00000065, -0.00000623, -0.00001099, 0.00000211, -0.00000893, 0.00000297, 0.00000062, -0.00000329, 0.00000258, -0.00000882, 0.00000306, 0.00000504, -0.00000744, -0.00000273, -0.00000180, -0.00000896, 0.00000113, -0.00000739, -0.00002302, 0.00000690, -0.00000050, 0.00000417, 0.00003172, -0.00000073, 0.00000155, 0.00001281, 0.00000781, -0.00000294, 0.00000075, 0.00002092, -0.00003046, -0.00000088, 0.00002180, -0.00003082, 0.00001921, -0.00000558, -0.00001549, 0.00002619, -0.00001718, 0.00000603, -0.00000235, -0.00000833, 0.00001405, -0.00001714, 0.00001594, -0.00000173, -0.00001586, 0.00002000, -0.00000390, 0.00000302, -0.00000969, -0.00000926, 0.00001664, -0.00000904, -0.00000152, 0.00000408, -0.00000056, 0.00000290, -0.00000332, 0.00000268, -0.00000302, 0.00000504, 0.00000211, -0.00001193, 0.00000340, 0.00000767, 0.00001140, -0.00000724, -0.00001979, 0.00001545, 0.00000439, -0.00000967, 0.00001517, -0.00000263, -0.00001186, 0.00000508, -0.00000638, -0.00000426, 0.00000963, 0.00000616, -0.00000333, -0.00001624, -0.00000942, 0.00000740, 0.00000578, -0.00000072, 0.00000073, -0.00000352, -0.00000817, 0.00000531, -0.00000158, -0.00001471, 0.00000058, 0.00000462, 0.00000394, -0.00000134, -0.00001105, -0.00000276, 0.00000457, 0.00000780, 0.00000115, -0.00000224, 0.00000097, -0.00000522, -0.00000077, -0.00000290, -0.00000698, 0.00000334, -0.00000019, -0.00000449, -0.00000001, -0.00000576, -0.00000862, -0.00000100, },
		{ -0.00005855, -0.00002152, 0.00007821, 0.00007659, 0.00005043, 0.00010740, 0.00002398, -0.00008582, -0.00002920, -0.00000920, -0.00001470, -0.00000107, -0.00000917, 0.00000059, -0.00001029, -0.00000641, -0.00000024, -0.00000772, 0.00000530, -0.00000266, -0.00000041, -0.00000036, -0.00000787, 0.00000668, 0.00000083, -0.00000776, -0.00000111, -0.00000371, -0.00000862, 0.00000217, -0.00001232, -0.00001978, 0.00000877, -0.00000336, 0.00001332, 0.00002929, -0.00000559, 0.00000651, 0.00001515, 0.00000443, -0.00000553, 0.00001127, 0.00000750, -0.00003368, 0.00001736, 0.00000362, -0.00002355, 0.00002507, -0.00002058, 0.00000127, 0.00001801, -0.00001997, 0.00001289, -0.00001120, 0.00000151, 0.00000672, -0.00001488, 0.00002236, -0.00001381, -0.00000663, 0.00001823, -0.00000631, 0.00000466, -0.00001563, -0.00000019, 0.00001309, -0.00001233, 0.00000354, 0.00000140, 0.00000056, 0.00000149, -0.00000248, 0.00000263, -0.00000284, 0.00000680, -0.00000341, -0.00000994, 0.00000596, 0.00000946, 0.00001062, -0.00001681, -0.00001264, 0.00002022, -0.00000445, -0.00000382, 0.00001590, -0.00000929, -0.00000674, 0.00000346, -0.00000972, 0.00000182, 0.00001054, 0.00000328, -0.00000673, -0.00001728, -0.00000399, 0.00000918, 0.00000293, -0.00000049, 0.00000077, -0.00000620, -0.00000519, 0.00000684, -0.00000668, -0.00001223, 0.00000298, 0.00000383, 0.00000469, -0.00000484, -0.00001077, -0.00000042, 0.00000567, 0.00000804, -0.00000146, -0.00000131, 0.00000017, -0.00000552, 0.00000077, -0.00000562, -0.00000486, 0.00000489, -0.00000315, -0.00000294, -0.00000004, -0.00000835, -0.00000632, -0.00000007, -0.00002249, },
		{ 0.00000888, 0.00008688, 0.00005786, 0.00005867, 0.00010157, -0.00000917, -0.00007978, -0.00001734, -0.00001199, -0.00001044, -0.00000290, -0.00000716, -0.00000063, -0.00001195, -0.00000266, -0.00000327, -0.00000449, 0.00000510, -0.00000432, 0.00000179, -0.00000392, -0.00000452, 0.00000782, -0.00000335, -0.00000664, -0.00000041, -0.00000570, -0.00000760, 0.00000230, -0.00001688, -0.00001559, 0.00000895, -0.00000383, 0.00002232, 0.00002390, -0.00000845, 0.00001316, 0.00001537, -0.00000104, -0.00000224, 0.00001647, -0.00000943, -0.00002485, 0.00002568, -0.00001396, -0.00000821, 0.00001958, -0.00002702, 0.00001672, 0.00000392, -0.00001484, 0.00001367, -0.00001624, 0.00001056, -0.00000362, -0.00000613, 0.00002180, -0.00002291, 0.00000437, 0.00001264, -0.00000532, 0.00000297, -0.00001836, 0.00000874, 0.00000598, -0.00001160, 0.00000681, -0.00000145, 0.00000213, -0.00000027, -0.00000097, 0.00000186, -0.00000158, 0.00000649, -0.00000813, -0.00000643, 0.00000776, 0.00001214, 0.00000573, -0.00002390, -0.00000158, 0.00001914, -0.00001107, 0.00000423, 0.00001229, -0.00001302, -0.00000148, -0.00000108, -0.00000952, 0.00000799, 0.00000895, 0.00000057, -0.00001012, -0.00001645, 0.00000185, 0.00000869, 0.00000058, 0.00000058, -0.00000056, -0.00000779, -0.00000117, 0.00000582, -0.00001062, -0.00000850, 0.00000379, 0.00000395, 0.00000444, -0.00000844, -0.00000910, 0.00000139, 0.00000716, 0.00000710, -0.00000380, 0.00000031, -0.00000150, -0.00000460, 0.00000114, -0.00000817, -0.00000114, 0.00000430, -0.00000544, -0.00000038, -0.00000192, -0.00000985, -0.00000274, -0.00000264, -0.00003454, -0.00005540, },
		{ 0.00008342, 0.00004401, 0.00007008, 0.00008629, -0.00003597, -0.00006802, -0.00001016, -0.00001345, -0.00000719, -0.00000450, -0.00000499, -0.00000327, -0.00001123, -0.00000078, -0.00000541, -0.00000071, 0.00000301, -0.00000402, 0.00000229, -0.00000651, -0.00000004, 0.00000632, -0.00000634, -0.00000489, -0.00000057, -0.00000756, -0.00000599, 0.00000146, -0.00002093, -0.00001071, 0.00000782, -0.00000163, 0.00003008, 0.00001590, -0.00000806, 0.00002054, 0.00001187, -0.00000523, 0.00000501, 0.00001329, -0.00002266, -0.00000886, 0.00002217, -0.00002395, 0.00000800, 0.00000602, -0.00002280, 0.00002502, -0.00001029, -0.00000462, 0.00000825, -0.00001532, 0.00001518, -0.00001280, 0.00000633, 0.00001392, -0.00002605, 0.00001386, 0.00000539, -0.00000196, -0.00000190, -0.00001660, 0.00001490, -0.00000238, -0.00000734, 0.00000732, -0.00000337, 0.00000354, -0.00000195, 0.00000083, 0.00000077, 0.00000021, 0.00000415, -0.00001103, -0.00000241, 0.00000960, 0.00001431, -0.00000296, -0.00002579, 0.00001011, 0.00001272, -0.00001317, 0.00001162, 0.00000562, -0.00001294, 0.00000168, -0.00000663, -0.00000544, 0.00001230, 0.00000572, -0.00000172, -0.00001336, -0.00001341, 0.00000686, 0.00000627, -0.00000043, 0.00000154, -0.00000292, -0.00000755, 0.00000258, 0.00000276, -0.00001255, -0.00000472, 0.00000364, 0.00000491, 0.00000272, -0.00001126, -0.00000663, 0.00000289, 0.00000882, 0.00000471, -0.00000501, 0.00000186, -0.00000360, -0.00000256, -0.00000022, -0.00000929, 0.00000311, 0.00000146, -0.00000583, 0.00000188, -0.00000534, -0.00000909, 0.00000046, -0.00000925, -0.00004451, -0.00004270, 0.00003675, },
		{ 0.00003760, 0.00008011, 0.00006394, -0.00005435, -0.00005385, -0.00000668, -0.00001358, -0.00000511, -0.00000535, -0.00000365, -0.00000598, -0.00000904, -0.00000077, -0.00000581, 0.00000220, 0.00000028, -0.00000238, 0.00000095, -0.00000707, 0.00000396, 0.00000299, -0.00000766, -0.00000324, -0.00000134, -0.00000919, -0.00000376, -0.00000066, -0.00002412, -0.00000543, 0.00000571, 0.00000349, 0.00003527, 0.00000629, -0.00000331, 0.00002642, 0.00000492, -0.00000521, 0.00001194, 0.00000280, -0.00002710, 0.00000702, 0.00000999, -0.00002337, 0.00001834, -0.00000929, -0.00001033, 0.00002358, -0.00001932, 0.00000603, -0.00000080, -0.00000885, 0.00001352, -0.00001695, 0.00001814, 0.00000094, -0.00002228, 0.00001918, -0.00000099, 0.00000181, -0.00000848, -0.00001052, 0.00001644, -0.00000933, -0.00000116, 0.00000505, -0.00000358, 0.00000415, -0.00000310, 0.00000252, -0.00000029, 0.00000181, 0.00000044, -0.00001176, 0.00000145, 0.00001206, 0.00001411, -0.00001351, -0.00002133, 0.00001882, 0.00000334, -0.00001016, 0.00001585, -0.00000174, -0.00000971, 0.00000139, -0.00001076, 0.00000143, 0.00001349, 0.00000210, -0.00000378, -0.00001599, -0.00000829, 0.00000987, 0.00000295, 0.00000010, 0.00000146, -0.00000537, -0.00000549, 0.00000488, -0.00000124, -0.00001239, -0.00000178, 0.00000335, 0.00000611, -0.00000036, -0.00001271, -0.00000397, 0.00000447, 0.00000999, 0.00000120, -0.00000458, 0.00000248, -0.00000531, -0.00000014, -0.00000320, -0.00000800, 0.00000624, -0.00000270, -0.00000380, 0.00000232, -0.00000899, -0.00000590, 0.00000126, -0.00001876, -0.00004916, -0.00002285, 0.00005741, 0.00007181, },
		{ 0.00008525, 0.00003794, -0.00006403, -0.00003980, -0.00000570, -0.00001273, -0.00000396, -0.00000557, -0.00000340, -0.00000780, -0.00000656, -0.00000183, -0.00000450, 0.00000345, -0.00000188, -0.00000051, -0.00000149, -0.00000539, 0.00000620, -0.00000094, -0.00000752, -0.00000205, -0.00000258, -0.00001041, -0.00000101, -0.00000439, -0.00002580, -0.00000033, 0.00000328, 0.00001142, 0.00003639, -0.00000295, 0.00000586, 0.00002820, -0.00000312, -0.00000025, 0.00001408, -0.00001048, -0.00002178, 0.00001646, -0.00000454, -0.00001399, 0.00001930, -0.00001992, 0.00000490, 0.00001382, -0.00002044, 0.00001278, -0.00000953, 0.00000041, 0.00000635, -0.00001404, 0.00002497, -0.00001312, -0.00001295, 0.00001923, -0.00000453, 0.00000387, -0.00001445, -0.00000178, 0.00001304, -0.00001272, 0.00000470, 0.00000100, -0.00000198, 0.00000356, -0.00000331, 0.00000377, -0.00000109, 0.00000260, -0.00000368, -0.00001057, 0.00000507, 0.00001498, 0.00001013, -0.00002286, -0.00001150, 0.00002200, -0.00000569, -0.00000338, 0.00001569, -0.00000736, -0.00000527, -0.00000222, -0.00001141, 0.00000898, 0.00001142, -0.00000079, -0.00000608, -0.00001722, -0.00000186, 0.00001017, 0.00000005, 0.00000161, -0.00000007, -0.00000685, -0.00000236, 0.00000523, -0.00000496, -0.00001074, 0.00000002, 0.00000355, 0.00000680, -0.00000422, -0.00001265, -0.00000144, 0.00000638, 0.00000992, -0.00000251, -0.00000269, 0.00000172, -0.00000580, 0.00000157, -0.00000679, -0.00000423, 0.00000678, -0.00000644, -0.00000002, 0.00000005, -0.00001099, -0.00000140, -0.00000172, -0.00002868, -0.00004679, 0.00000025, 0.00006857, 0.00005689, 0.00003857, },
		{ 0.00001182, -0.00006598, -0.00002754, -0.00000617, -0.00001125, -0.00000344, -0.00000552, -0.00000395, -0.00000842, -0.00000468, -0.00000293, -0.00000225, 0.00000304, -0.00000282, 0.00000059, -0.00000385, -0.00000219, 0.00000616, -0.00000431, -0.00000652, -0.00000136, -0.00000428, -0.00001081, 0.00000176, -0.00000981, -0.00002530, 0.00000376, 0.00000165, 0.00002130, 0.00003246, -0.00000909, 0.00001777, 0.00002431, -0.00000881, 0.00000753, 0.00000917, -0.00002077, -0.00001002, 0.00001670, -0.00001505, -0.00000103, 0.00001172, -0.00002204, 0.00001679, 0.00000023, -0.00001432, 0.00001334, -0.00001435, 0.00000861, -0.00000326, -0.00000478, 0.00002419, -0.00002393, -0.00000114, 0.00001481, -0.00000456, 0.00000284, -0.00001754, 0.00000698, 0.00000598, -0.00001168, 0.00000821, -0.00000323, 0.00000082, 0.00000177, -0.00000237, 0.00000431, -0.00000161, 0.00000229, -0.00000732, -0.00000799, 0.00000873, 0.00001734, 0.00000217, -0.00002798, 0.00000086, 0.00001910, -0.00001137, 0.00000457, 0.00001166, -0.00000973, -0.00000196, -0.00000755, -0.00000777, 0.00001485, 0.00000703, -0.00000249, -0.00000891, -0.00001628, 0.00000455, 0.00000788, -0.00000130, 0.00000303, -0.00000269, -0.00000662, 0.00000067, 0.00000385, -0.00000758, -0.00000848, 0.00000095, 0.00000442, 0.00000639, -0.00000810, -0.00001136, 0.00000098, 0.00000846, 0.00000813, -0.00000531, -0.00000015, -0.00000024, -0.00000475, 0.00000155, -0.00000946, 0.00000099, 0.00000418, -0.00000790, 0.00000375, -0.00000454, -0.00000992, 0.00000222, -0.00000838, -0.00003613, -0.00003762, 0.00002250, 0.00007034, 0.00004299, 0.00004494, 0.00008363, },
		{ -0.00006194, -0.00001794, -0.00000724, -0.00000945, -0.00000337, -0.00000550, -0.00000475, -0.00000812, -0.00000368, -0.00000332, -0.00000004, 0.00000161, -0.00000259, 0.00000048, -0.00000507, 0.00000130, 0.00000415, -0.00000646, -0.00000524, -0.00000111, -0.00000644, -0.00000992, 0.00000365, -0.00001637, -0.00002231, 0.00000605, 0.00000227, 0.00003126, 0.00002374, -0.00000963, 0.00002922, 0.00001528, -0.00000938, 0.00001409, -0.00000162, -0.00002380, 0.00000249, 0.00000925, -0.00001782, 0.00000953, -0.00000011, -0.00001569, 0.00002122, -0.00001181, -0.00000434, 0.00000824, -0.00001363, 0.00001251, -0.00001122, 0.00000763, 0.00001582, -0.00002836, 0.00000951, 0.00000814, -0.00000197, -0.00000134, -0.00001636, 0.00001304, -0.00000226, -0.00000688, 0.00000826, -0.00000600, 0.00000374, -0.00000072, -0.00000033, 0.00000397, -0.00000191, 0.00000094, -0.00000987, -0.00000447, 0.00001270, 0.00001758, -0.00000838, -0.00002699, 0.00001212, 0.00001163, -0.00001217, 0.00001090, 0.00000564, -0.00000885, -0.00000155, -0.00001216, -0.00000057, 0.00001726, 0.00000194, -0.00000325, -0.00001210, -0.00001274, 0.00000938, 0.00000399, -0.00000066, 0.00000323, -0.00000541, -0.00000470, 0.00000261, 0.00000156, -0.00000885, -0.00000631, 0.00000152, 0.00000570, 0.00000470, -0.00001136, -0.00000920, 0.00000348, 0.00001011, 0.00000483, -0.00000637, 0.00000197, -0.00000259, -0.00000259, -0.00000051, -0.00000975, 0.00000573, -0.00000074, -0.00000608, 0.00000538, -0.00000968, -0.00000565, 0.00000270, -0.00001697, -0.00003901, -0.00002347, 0.00004065, 0.00006461, 0.00003311, 0.00005381, 0.00007501, -0.00001144, },
		{ -0.00001126, -0.00000825, -0.00000761, -0.00000369, -0.00000557, -0.00000539, -0.00000742, -0.00000329, -0.00000278, 0.00000139, -0.00000002, -0.00000173, -0.00000065, -0.00000468, 0.00000391, 0.00000105, -0.00000731, -0.00000405, -0.00000132, -0.00000882, -0.00000749, 0.00000360, -0.00002289, -0.00001728, 0.00000636, 0.00000637, 0.00003875, 0.00001216, -0.00000342, 0.00003653, 0.00000390, -0.00000428, 0.00001540, -0.00001419, -0.00001870, 0.00001037, -0.00000128, -0.00001305, 0.00001365, -0.00001067, -0.00000438, 0.00001745, -0.00001822, 0.00000511, 0.00000026, -0.00000818, 0.00001083, -0.00001409, 0.00001878, 0.00000246, -0.00002555, 0.00001613, 0.00000190, 0.00000133, -0.00000731, -0.00001104, 0.00001463, -0.00000888, -0.00000028, 0.00000501, -0.00000625, 0.00000558, -0.00000304, 0.00000239, 0.00000269, -0.00000209, -0.00000115, -0.00001107, -0.00000021, 0.00001671, 0.00001434, -0.00001885, -0.00002000, 0.00001914, 0.00000255, -0.00000847, 0.00001375, 0.00000003, -0.00000616, -0.00000440, -0.00001369, 0.00000818, 0.00001564, -0.00000228, -0.00000384, -0.00001490, -0.00000693, 0.00001142, 0.00000002, 0.00000149, 0.00000163, -0.00000703, -0.00000186, 0.00000300, -0.00000070, -0.00000913, -0.00000455, 0.00000219, 0.00000688, 0.00000190, -0.00001364, -0.00000638, 0.00000616, 0.00001057, 0.00000082, -0.00000554, 0.00000282, -0.00000432, -0.00000038, -0.00000393, -0.00000706, 0.00000800, -0.00000597, -0.00000148, 0.00000344, -0.00001291, 0.00000028, -0.00000119, -0.00002479, -0.00003674, -0.00000689, 0.00005285, 0.00005416, 0.00002871, 0.00006214, 0.00006052, -0.00002974, -0.00005397, },
		{ -0.00000877, -0.00000603, -0.00000434, -0.00000559, -0.00000574, -0.00000672, -0.00000304, -0.00000161, 0.00000177, -0.00000122, -0.00000092, -0.00000211, -0.00000289, 0.00000491, -0.00000218, -0.00000714, -0.00000309, -0.00000212, -0.00001088, -0.00000383, 0.00000089, -0.00002794, -0.00001135, 0.00000549, 0.00001420, 0.00004134, 0.00000104, 0.00000848, 0.00003707, -0.00000588, 0.00000424, 0.00000953, -0.00002335, -0.00000824, 0.00001099, -0.00000964, -0.00000422, 0.00001067, -0.00001574, 0.00000685, 0.00000798, -0.00001769, 0.00001056, -0.00000701, -0.00000073, 0.00000476, -0.00001054, 0.00002468, -0.00001172, -0.00001702, 0.00001761, -0.00000187, 0.00000325, -0.00001287, -0.00000322, 0.00001152, -0.00001178, 0.00000561, -0.00000016, -0.00000389, 0.00000550, -0.00000424, 0.00000509, 0.00000060, -0.00000218, -0.00000360, -0.00001085, 0.00000479, 0.00001986, 0.00000721, -0.00002626, -0.00000907, 0.00002041, -0.00000497, -0.00000226, 0.00001285, -0.00000337, -0.00000385, -0.00000936, -0.00001083, 0.00001585, 0.00001074, -0.00000465, -0.00000507, -0.00001627, 0.00000005, 0.00001025, -0.00000248, 0.00000395, -0.00000152, -0.00000671, 0.00000070, 0.00000216, -0.00000234, -0.00000895, -0.00000308, 0.00000313, 0.00000756, -0.00000161, -0.00001482, -0.00000294, 0.00000871, 0.00000937, -0.00000275, -0.00000343, 0.00000223, -0.00000475, 0.00000069, -0.00000722, -0.00000210, 0.00000667, -0.00000913, 0.00000391, -0.00000189, -0.00001223, 0.00000505, -0.00000880, -0.00002931, -0.00003021, 0.00000978, 0.00005848, 0.00004198, 0.00002981, 0.00006733, 0.00004219, -0.00004202, -0.00004415, -0.00000721, },
		{ -0.00000499, -0.00000512, -0.00000541, -0.00000593, -0.00000617, -0.00000251, -0.00000033, 0.00000132, -0.00000175, -0.00000062, -0.00000317, -0.00000043, 0.00000420, -0.00000479, -0.00000640, -0.00000239, -0.00000362, -0.00001191, 0.00000015, -0.00000447, -0.00003040, -0.00000595, 0.00000506, 0.00002458, 0.00003781, -0.00000591, 0.00002292, 0.00003047, -0.00001061, 0.00001201, -0.00000223, -0.00002532, 0.00000262, 0.00000530, -0.00001240, 0.00000399, 0.00000316, -0.00001404, 0.00001370, -0.00000290, -0.00001175, 0.00001074, -0.00001082, 0.00000544, -0.00000271, -0.00000177, 0.00002329, -0.00002257, -0.00000591, 0.00001472, -0.00000251, 0.00000245, -0.00001589, 0.00000460, 0.00000511, -0.00001028, 0.00000865, -0.00000519, 0.00000012, 0.00000343, -0.00000365, 0.00000690, -0.00000195, -0.00000223, -0.00000611, -0.00000916, 0.00001040, 0.00002079, -0.00000280, -0.00002839, 0.00000256, 0.00001651, -0.00000879, 0.00000389, 0.00000947, -0.00000407, -0.00000378, -0.00001415, -0.00000385, 0.00002021, 0.00000425, -0.00000514, -0.00000732, -0.00001522, 0.00000653, 0.00000650, -0.00000258, 0.00000525, -0.00000512, -0.00000441, 0.00000195, 0.00000092, -0.00000335, -0.00000870, -0.00000163, 0.00000420, 0.00000752, -0.00000550, -0.00001480, 0.00000113, 0.00001050, 0.00000665, -0.00000495, -0.00000105, 0.00000069, -0.00000393, -0.00000005, -0.00000871, 0.00000327, 0.00000212, -0.00000857, 0.00000741, -0.00000866, -0.00000725, 0.00000587, -0.00001770, -0.00002933, -0.00002116, 0.00002486, 0.00005785, 0.00003083, 0.00003524, 0.00006764, 0.00002251, -0.00004824, -0.00003422, -0.00000516, -0.00000867, },
		{ -0.00000574, -0.00000504, -0.00000616, -0.00000560, -0.00000160, 0.00000059, 0.00000044, -0.00000176, -0.00000089, -0.00000334, 0.00000185, 0.00000221, -0.00000642, -0.00000544, -0.00000202, -0.00000571, -0.00001136, 0.00000319, -0.00001162, -0.00002994, -0.00000207, 0.00000695, 0.00003503, 0.00002888, -0.00000604, 0.00003560, 0.00001894, -0.00000902, 0.00001480, -0.00001576, -0.00001939, 0.00000915, -0.00000298, -0.00000926, 0.00000793, -0.00000473, -0.00000734, 0.00001423, -0.00001112, -0.00000367, 0.00000668, -0.00001031, 0.00000802, -0.00000813, 0.00000900, 0.00001512, -0.00002736, 0.00000434, 0.00000954, -0.00000077, -0.00000112, -0.00001515, 0.00000993, -0.00000211, -0.00000534, 0.00000785, -0.00000814, 0.00000421, 0.00000011, -0.00000115, 0.00000709, -0.00000441, -0.00000231, -0.00000838, -0.00000589, 0.00001608, 0.00001826, -0.00001348, -0.00002467, 0.00001164, 0.00000966, -0.00000847, 0.00000791, 0.00000564, -0.00000306, -0.00000666, -0.00001632, 0.00000550, 0.00002014, -0.00000184, -0.00000449, -0.00001027, -0.00001143, 0.00001082, 0.00000165, -0.00000039, 0.00000439, -0.00000768, -0.00000101, 0.00000151, 0.00000018, -0.00000419, -0.00000837, -0.00000003, 0.00000517, 0.00000676, -0.00000955, -0.00001335, 0.00000561, 0.00001080, 0.00000318, -0.00000544, 0.00000066, -0.00000090, -0.00000260, -0.00000229, -0.00000741, 0.00000683, -0.00000371, -0.00000430, 0.00000694, -0.00001386, 0.00000031, 0.00000131, -0.00002460, -0.00002539, -0.00001117, 0.00003729, 0.00005190, 0.00002292, 0.00004298, 0.00006242, 0.00000398, -0.00004928, -0.00002541, -0.00000432, -0.00000810, -0.00000462, },
		{ -0.00000467, -0.00000651, -0.00000479, -0.00000048, 0.00000091, -0.00000044, -0.00000155, -0.00000148, -0.00000256, 0.00000325, -0.00000041, -0.00000706, -0.00000449, -0.00000214, -0.00000797, -0.00000917, 0.00000414, -0.00001913, -0.00002715, 0.00000024, 0.00001239, 0.00004255, 0.00001725, 0.00000120, 0.00004272, 0.00000640, -0.00000268, 0.00001030, -0.00002587, -0.00000832, 0.00000898, -0.00000940, -0.00000275, 0.00000658, -0.00000935, 0.00000074, 0.00000936, -0.00001438, 0.00000318, 0.00000091, -0.00000657, 0.00000654, -0.00000908, 0.00001789, 0.00000286, -0.00002557, 0.00001117, 0.00000449, 0.00000161, -0.00000623, -0.00001085, 0.00001129, -0.00000755, 0.00000089, 0.00000374, -0.00000793, 0.00000675, -0.00000316, 0.00000267, 0.00000530, -0.00000621, -0.00000264, -0.00001006, -0.00000095, 0.00002076, 0.00001188, -0.00002204, -0.00001642, 0.00001613, 0.00000271, -0.00000523, 0.00000906, 0.00000310, -0.00000221, -0.00001166, -0.00001426, 0.00001463, 0.00001591, -0.00000603, -0.00000384, -0.00001298, -0.00000541, 0.00001183, -0.00000243, 0.00000300, 0.00000126, -0.00000798, 0.00000204, -0.00000013, 0.00000034, -0.00000542, -0.00000761, 0.00000156, 0.00000595, 0.00000527, -0.00001353, -0.00001013, 0.00000979, 0.00000925, 0.00000003, -0.00000464, 0.00000134, -0.00000183, -0.00000183, -0.00000484, -0.00000362, 0.00000698, -0.00000820, 0.00000183, 0.00000209, -0.00001475, 0.00000711, -0.00000778, -0.00002683, -0.00001943, -0.00000104, 0.00004625, 0.00004211, 0.00001975, 0.00005051, 0.00005222, -0.00001139, -0.00004661, -0.00001826, -0.00000401, -0.00000743, -0.00000485, -0.00000596, },
		{ -0.00000683, -0.00000357, 0.00000054, 0.00000067, -0.00000109, -0.00000138, -0.00000200, -0.00000113, 0.00000347, -0.00000302, -0.00000692, -0.00000371, -0.00000284, -0.00000979, -0.00000587, 0.00000244, -0.00002560, -0.00002323, 0.00000205, 0.00002126, 0.00004476, 0.00000673, 0.00001386, 0.00004236, -0.00000317, 0.00000462, -0.00000062, -0.00002864, 0.00000308, 0.00000299, -0.00001095, 0.00000331, 0.00000159, -0.00000917, 0.00000660, 0.00000193, -0.00001275, 0.00000668, -0.00000396, -0.00000186, 0.00000240, -0.00000510, 0.00002185, -0.00000971, -0.00001876, 0.00001368, 0.00000130, 0.00000285, -0.00001096, -0.00000451, 0.00000870, -0.00000944, 0.00000587, -0.00000193, -0.00000474, 0.00000676, -0.00000495, 0.00000662, 0.00000176, -0.00000693, -0.00000347, -0.00001063, 0.00000545, 0.00002301, 0.00000245, -0.00002624, -0.00000629, 0.00001583, -0.00000203, -0.00000117, 0.00000810, 0.00000250, -0.00000332, -0.00001672, -0.00000783, 0.00002103, 0.00000901, -0.00000769, -0.00000413, -0.00001424, 0.00000146, 0.00000951, -0.00000424, 0.00000589, -0.00000316, -0.00000567, 0.00000337, -0.00000182, 0.00000113, -0.00000725, -0.00000605, 0.00000281, 0.00000668, 0.00000285, -0.00001696, -0.00000499, 0.00001263, 0.00000620, -0.00000191, -0.00000342, 0.00000128, -0.00000201, -0.00000230, -0.00000612, 0.00000109, 0.00000364, -0.00000921, 0.00000690, -0.00000539, -0.00001033, 0.00000959, -0.00001833, -0.00002360, -0.00001366, 0.00000924, 0.00005091, 0.00003052, 0.00002173, 0.00005534, 0.00003868, -0.00002251, -0.00004185, -0.00001269, -0.00000384, -0.00000703, -0.00000535, -0.00000568, -0.00000452, },
		{ -0.00000201, 0.00000115, 0.00000008, -0.00000144, -0.00000138, -0.00000215, 0.00000046, 0.00000255, -0.00000513, -0.00000626, -0.00000324, -0.00000407, -0.00001061, -0.00000242, -0.00000168, -0.00003015, -0.00001930, 0.00000502, 0.00003188, 0.00004096, 0.00000091, 0.00002821, 0.00003521, -0.00000747, 0.00000863, -0.00001419, -0.00002312, 0.00001010, -0.00000529, -0.00000734, 0.00000588, -0.00000384, -0.00000513, 0.00000823, -0.00000481, -0.00000813, 0.00000660, -0.00000634, 0.00000161, -0.00000192, 0.00000224, 0.00001972, -0.00001922, -0.00000965, 0.00001259, 0.00000046, 0.00000193, -0.00001355, 0.00000168, 0.00000361, -0.00000753, 0.00000774, -0.00000690, 0.00000011, 0.00000426, -0.00000431, 0.00000929, -0.00000273, -0.00000652, -0.00000494, -0.00000948, 0.00001255, 0.00002160, -0.00000804, -0.00002516, 0.00000272, 0.00001222, -0.00000370, 0.00000183, 0.00000664, 0.00000309, -0.00000717, -0.00001945, 0.00000156, 0.00002310, 0.00000151, -0.00000721, -0.00000567, -0.00001314, 0.00000742, 0.00000491, -0.00000321, 0.00000675, -0.00000719, -0.00000153, 0.00000236, -0.00000233, 0.00000165, -0.00000931, -0.00000364, 0.00000343, 0.00000755, -0.00000082, -0.00001900, 0.00000173, 0.00001312, 0.00000267, -0.00000246, -0.00000254, 0.00000116, -0.00000201, -0.00000381, -0.00000511, 0.00000457, -0.00000160, -0.00000618, 0.00000828, -0.00001223, -0.00000207, 0.00000570, -0.00002626, -0.00001641, -0.00000937, 0.00001993, 0.00005044, 0.00001967, 0.00002801, 0.00005567, 0.00002405, -0.00002933, -0.00003635, -0.00000828, -0.00000384, -0.00000707, -0.00000572, -0.00000504, -0.00000478, -0.00000688, },
		{ 0.00000121, -0.00000060, -0.00000157, -0.00000148, -0.00000180, 0.00000172, 0.00000079, -0.00000649, -0.00000538, -0.00000320, -0.00000554, -0.00001021, 0.00000015, -0.00000740, -0.00003271, -0.00001581, 0.00001064, 0.00004159, 0.00003248, 0.00000188, 0.00004008, 0.00002405, -0.00000674, 0.00000656, -0.00002534, -0.00001178, 0.00001021, -0.00001162, -0.00000089, 0.00000405, -0.00000677, 0.00000015, 0.00000582, -0.00000863, -0.00000305, 0.00000418, -0.00000615, 0.00000270, -0.00000411, 0.00001015, 0.00001245, -0.00002370, -0.00000106, 0.00000958, 0.00000128, -0.00000108, -0.00001317, 0.00000576, -0.00000174, -0.00000303, 0.00000595, -0.00000927, 0.00000469, 0.00000034, -0.00000117, 0.00000956, -0.00000704, -0.00000534, -0.00000691, -0.00000609, 0.00001902, 0.00001610, -0.00001702, -0.00001959, 0.00000847, 0.00000764, -0.00000297, 0.00000295, 0.00000622, 0.00000318, -0.00001308, -0.00001808, 0.00001150, 0.00002062, -0.00000471, -0.00000564, -0.00000801, -0.00000947, 0.00001089, -0.00000020, 0.00000007, 0.00000484, -0.00000910, 0.00000279, -0.00000053, -0.00000099, 0.00000088, -0.00001077, -0.00000084, 0.00000350, 0.00000853, -0.00000594, -0.00001856, 0.00000889, 0.00001080, -0.00000003, -0.00000218, -0.00000220, 0.00000146, -0.00000270, -0.00000530, -0.00000196, 0.00000510, -0.00000617, -0.00000057, 0.00000493, -0.00001514, 0.00000651, -0.00000398, -0.00002822, -0.00000836, -0.00000627, 0.00003069, 0.00004457, 0.00001220, 0.00003646, 0.00005088, 0.00001063, -0.00003271, -0.00003085, -0.00000463, -0.00000426, -0.00000742, -0.00000568, -0.00000436, -0.00000542, -0.00000640, -0.00000033, },
		{ -0.00000116, -0.00000159, -0.00000156, -0.00000104, 0.00000231, -0.00000137, -0.00000703, -0.00000455, -0.00000362, -0.00000683, -0.00000880, 0.00000116, -0.00001368, -0.00003382, -0.00001226, 0.00001942, 0.00004770, 0.00002227, 0.00000942, 0.00004641, 0.00001245, -0.00000352, -0.00000159, -0.00002985, 0.00000067, 0.00000410, -0.00001299, 0.00000488, -0.00000072, -0.00000591, 0.00000386, 0.00000126, -0.00000905, 0.00000058, 0.00000121, -0.00000450, 0.00000170, -0.00000302, 0.00001580, 0.00000246, -0.00002301, 0.00000513, 0.00000647, 0.00000243, -0.00000510, -0.00001015, 0.00000676, -0.00000529, 0.00000191, 0.00000140, -0.00000825, 0.00000724, -0.00000331, 0.00000353, 0.00000699, -0.00001004, -0.00000412, -0.00000880, -0.00000038, 0.00002317, 0.00000726, -0.00002229, -0.00001171, 0.00001035, 0.00000415, -0.00000145, 0.00000276, 0.00000739, 0.00000103, -0.00001910, -0.00001222, 0.00001943, 0.00001464, -0.00000850, -0.00000415, -0.00001017, -0.00000394, 0.00001105, -0.00000394, 0.00000406, 0.00000062, -0.00000793, 0.00000544, -0.00000390, 0.00000188, -0.00000172, -0.00001071, 0.00000154, 0.00000357, 0.00000912, -0.00001221, -0.00001473, 0.00001473, 0.00000624, -0.00000095, -0.00000204, -0.00000194, 0.00000205, -0.00000455, -0.00000534, 0.00000190, 0.00000242, -0.00000776, 0.00000479, -0.00000187, -0.00001239, 0.00001126, -0.00001620, -0.00002318, -0.00000272, -0.00000273, 0.00004016, 0.00003411, 0.00001008, 0.00004424, 0.00004181, -0.00000002, -0.00003395, -0.00002541, -0.00000163, -0.00000531, -0.00000773, -0.00000516, -0.00000396, -0.00000622, -0.00000526, 0.00000114, 0.00000073, },
		},
		{ { -0.00000424, -0.00000474, -0.00000426, -0.00000504, -0.00001177, -0.00000276, -0.00002655, -0.00004177, 0.00001614, 0.00005269, 0.00002944, 0.00002810, 0.00005570, 0.00001386, -0.00001986, -0.00002832, -0.00000918, 0.00000597, -0.00001958, 0.00000585, 0.00000632, 0.00000062, -0.00001035, -0.00001144, 0.00000972, -0.00000268, -0.00000005, -0.00000971, -0.00000264, 0.00000665, -0.00000532, 0.00000316, 0.00000016, -0.00000911, -0.00000154, -0.00000241, -0.00000269, -0.00000745, 0.00000936, 0.00002270, -0.00001740, -0.00000714, -0.00000422, -0.00000036, 0.00001268, -0.00000723, 0.00000470, 0.00000011, 0.00000368, -0.00000458, -0.00000425, 0.00000294, -0.00000091, 0.00000981, -0.00000622, 0.00000042, -0.00000016, -0.00000697, 0.00000087, 0.00000164, 0.00000513, 0.00000752, -0.00000555, -0.00000765, 0.00000686, -0.00000765, -0.00000555, 0.00000752, 0.00000513, 0.00000164, 0.00000087, -0.00000697, -0.00000016, 0.00000042, -0.00000622, 0.00000981, -0.00000091, 0.00000294, -0.00000425, -0.00000458, 0.00000368, 0.00000011, 0.00000470, -0.00000723, 0.00001268, -0.00000036, -0.00000422, -0.00000714, -0.00001740, 0.00002270, 0.00000936, -0.00000745, -0.00000269, -0.00000241, -0.00000154, -0.00000911, 0.00000016, 0.00000316, -0.00000532, 0.00000665, -0.00000264, -0.00000971, -0.00000005, -0.00000268, 0.00000972, -0.00001144, -0.00001035, 0.00000062, 0.00000632, 0.00000585, -0.00001958, 0.00000597, -0.00000918, -0.00002832, -0.00001986, 0.00001386, 0.00005570, 0.00002810, 0.00002944, 0.00005269, 0.00001614, -0.00004177, -0.00002655, -0.00000276, -0.00001177, -0.00000504, -0.00000426, -0.00000474, },
		{ -0.00000475, -0.00000403, -0.00000632, -0.00001118, -0.00000300, -0.00003224, -0.00004034, 0.00002582, 0.00005469, 0.00002704, 0.00003619, 0.00005549, 0.00000468, -0.00002574, -0.00002749, -0.00000251, 0.00000079, -0.00001743, 0.00000833, 0.00000355, 0.00000119, -0.00001321, -0.00000707, 0.00000836, -0.00000317, 0.00000022, -0.00001185, 0.00000118, 0.00000440, -0.00000517, 0.00000438, -0.00000264, -0.00000752, -0.00000183, -0.00000238, -0.00000328, -0.00000799, 0.00001591, 0.00001684, -0.00002024, -0.00000373, -0.00000558, 0.00000421, 0.00000897, -0.00000689, 0.00000595, -0.00000061, 0.00000403, -0.00000684, -0.00000171, 0.00000140, 0.00000123, 0.00000923, -0.00000788, 0.00000219, -0.00000166, -0.00000685, 0.00000130, 0.00000274, 0.00000634, 0.00000582, -0.00000869, -0.00000379, 0.00000587, -0.00001039, -0.00000170, 0.00000820, 0.00000373, 0.00000073, 0.00000073, -0.00000702, 0.00000108, -0.00000130, -0.00000404, 0.00000933, -0.00000217, 0.00000399, -0.00000661, -0.00000187, 0.00000270, 0.00000129, 0.00000252, -0.00000588, 0.00001496, -0.00000497, -0.00000229, -0.00001135, -0.00001196, 0.00002587, 0.00000283, -0.00000541, -0.00000329, -0.00000151, -0.00000224, -0.00000998, 0.00000316, 0.00000090, -0.00000403, 0.00000754, -0.00000601, -0.00000694, -0.00000142, -0.00000078, 0.00000949, -0.00001499, -0.00000731, -0.00000021, 0.00000989, 0.00000123, -0.00001921, 0.00000974, -0.00001705, -0.00002602, -0.00001406, 0.00002205, 0.00005354, 0.00002111, 0.00003276, 0.00004904, 0.00000738, -0.00004186, -0.00002126, -0.00000324, -0.00001189, -0.00000385, -0.00000471, -0.00000453, -0.00000433, },
		{ -0.00000411, -0.00000749, -0.00001027, -0.00000404, -0.00003817, -0.00003729, 0.00003606, 0.00005499, 0.00002605, 0.00004481, 0.00005267, -0.00000550, -0.00003064, -0.00002388, 0.00000190, -0.00000444, -0.00001361, 0.00000873, 0.00000202, 0.00000091, -0.00001521, -0.00000265, 0.00000591, -0.00000237, -0.00000089, -0.00001271, 0.00000466, 0.00000139, -0.00000378, 0.00000428, -0.00000463, -0.00000589, -0.00000262, -0.00000174, -0.00000498, -0.00000644, 0.00002119, 0.00000932, -0.00002055, -0.00000152, -0.00000583, 0.00000791, 0.00000462, -0.00000511, 0.00000615, -0.00000075, 0.00000360, -0.00000831, 0.00000060, -0.00000026, 0.00000398, 0.00000752, -0.00000875, 0.00000381, -0.00000344, -0.00000654, 0.00000200, 0.00000392, 0.00000723, 0.00000313, -0.00001051, 0.00000039, 0.00000310, -0.00001145, 0.00000218, 0.00000803, 0.00000222, 0.00000013, 0.00000076, -0.00000709, 0.00000208, -0.00000280, -0.00000164, 0.00000795, -0.00000248, 0.00000425, -0.00000838, 0.00000091, 0.00000135, 0.00000265, -0.00000026, -0.00000286, 0.00001527, -0.00000880, -0.00000050, -0.00001557, -0.00000441, 0.00002575, -0.00000240, -0.00000281, -0.00000470, 0.00000027, -0.00000417, -0.00000953, 0.00000558, -0.00000180, -0.00000156, 0.00000683, -0.00000822, -0.00000435, -0.00000327, 0.00000223, 0.00000752, -0.00001718, -0.00000476, -0.00000058, 0.00001341, -0.00000503, -0.00001596, 0.00001080, -0.00002458, -0.00002097, -0.00000912, 0.00002945, 0.00004916, 0.00001573, 0.00003643, 0.00004392, -0.00000016, -0.00004087, -0.00001651, -0.00000427, -0.00001147, -0.00000295, -0.00000522, -0.00000421, -0.00000457, -0.00000455, },
		{ -0.00000838, -0.00000924, -0.00000592, -0.00004413, -0.00003241, 0.00004645, 0.00005370, 0.00002688, 0.00005336, 0.00004694, -0.00001635, -0.00003352, -0.00001843, 0.00000365, -0.00000855, -0.00000923, 0.00000761, 0.00000173, -0.00000045, -0.00001586, 0.00000107, 0.00000312, -0.00000072, -0.00000328, -0.00001194, 0.00000706, -0.00000162, -0.00000163, 0.00000296, -0.00000547, -0.00000478, -0.00000334, -0.00000104, -0.00000728, -0.00000277, 0.00002418, 0.00000143, -0.00001887, -0.00000053, -0.00000477, 0.00001012, 0.00000048, -0.00000243, 0.00000543, -0.00000030, 0.00000236, -0.00000879, 0.00000230, -0.00000160, 0.00000695, 0.00000483, -0.00000866, 0.00000503, -0.00000539, -0.00000592, 0.00000293, 0.00000514, 0.00000759, -0.00000029, -0.00001063, 0.00000399, -0.00000084, -0.00001064, 0.00000550, 0.00000722, 0.00000063, -0.00000001, 0.00000079, -0.00000717, 0.00000290, -0.00000401, 0.00000071, 0.00000596, -0.00000190, 0.00000360, -0.00000924, 0.00000334, -0.00000005, 0.00000386, -0.00000313, 0.00000141, 0.00001343, -0.00001121, 0.00000043, -0.00001882, 0.00000426, 0.00002238, -0.00000539, -0.00000069, -0.00000620, 0.00000246, -0.00000715, -0.00000747, 0.00000673, -0.00000413, 0.00000149, 0.00000467, -0.00000892, -0.00000261, -0.00000484, 0.00000569, 0.00000405, -0.00001780, -0.00000313, 0.00000018, 0.00001586, -0.00001191, -0.00001017, 0.00000839, -0.00003009, -0.00001422, -0.00000532, 0.00003618, 0.00004272, 0.00001241, 0.00003980, 0.00003765, -0.00000635, -0.00003910, -0.00001241, -0.00000567, -0.00001051, -0.00000247, -0.00000561, -0.00000389, -0.00000485, -0.00000420, -0.00000456, },
		{ -0.00000831, -0.00000862, -0.00004994, -0.00002550, 0.00005649, 0.00005103, 0.00002989, 0.00006109, 0.00003804, -0.00002710, -0.00003368, -0.00001243, 0.00000304, -0.00001087, -0.00000530, 0.00000577, 0.00000230, -0.00000277, -0.00001499, 0.00000353, 0.00000070, 0.00000108, -0.00000647, -0.00000959, 0.00000795, -0.00000394, 0.00000060, 0.00000087, -0.00000523, -0.00000448, -0.00000356, -0.00000086, -0.00000941, 0.00000252, 0.00002438, -0.00000568, -0.00001599, -0.00000040, -0.00000257, 0.00001053, -0.00000271, 0.00000052, 0.00000410, 0.00000054, 0.00000047, -0.00000824, 0.00000316, -0.00000224, 0.00000967, 0.00000145, -0.00000755, 0.00000559, -0.00000737, -0.00000492, 0.00000400, 0.00000636, 0.00000717, -0.00000397, -0.00000899, 0.00000623, -0.00000510, -0.00000815, 0.00000788, 0.00000602, -0.00000102, 0.00000046, 0.00000058, -0.00000716, 0.00000355, -0.00000488, 0.00000277, 0.00000369, -0.00000064, 0.00000208, -0.00000901, 0.00000510, -0.00000122, 0.00000460, -0.00000550, 0.00000627, 0.00000964, -0.00001189, -0.00000003, -0.00002014, 0.00001268, 0.00001649, -0.00000573, 0.00000008, -0.00000697, 0.00000432, -0.00001062, -0.00000394, 0.00000620, -0.00000533, 0.00000429, 0.00000163, -0.00000816, -0.00000214, -0.00000545, 0.00000883, -0.00000030, -0.00001704, -0.00000250, 0.00000243, 0.00001627, -0.00001806, -0.00000289, 0.00000247, -0.00003221, -0.00000723, -0.00000235, 0.00004215, 0.00003451, 0.00001152, 0.00004218, 0.00003073, -0.00001116, -0.00003678, -0.00000900, -0.00000722, -0.00000909, -0.00000249, -0.00000574, -0.00000372, -0.00000505, -0.00000385, -0.00000530, -0.00000889, },
		{ -0.00001210, -0.00005539, -0.00001639, 0.00006565, 0.00004734, 0.00003533, 0.00006705, 0.00002591, -0.00003665, -0.00003099, -0.00000714, 0.00000085, -0.00001132, -0.00000252, 0.00000399, 0.00000314, -0.00000563, -0.00001279, 0.00000451, -0.00000078, 0.00000233, -0.00000976, -0.00000610, 0.00000726, -0.00000510, 0.00000225, -0.00000137, -0.00000424, -0.00000496, -0.00000309, -0.00000159, -0.00001053, 0.00000852, 0.00002184, -0.00001116, -0.00001273, -0.00000058, 0.00000026, 0.00000925, -0.00000448, 0.00000311, 0.00000257, 0.00000149, -0.00000179, -0.00000683, 0.00000314, -0.00000192, 0.00001167, -0.00000221, -0.00000552, 0.00000532, -0.00000914, -0.00000354, 0.00000514, 0.00000750, 0.00000580, -0.00000732, -0.00000591, 0.00000663, -0.00000879, -0.00000450, 0.00000916, 0.00000459, -0.00000261, 0.00000155, -0.00000012, -0.00000689, 0.00000398, -0.00000539, 0.00000436, 0.00000146, 0.00000098, -0.00000012, -0.00000769, 0.00000597, -0.00000196, 0.00000467, -0.00000687, 0.00001092, 0.00000449, -0.00001095, -0.00000200, -0.00001887, 0.00001943, 0.00000933, -0.00000372, -0.00000095, -0.00000632, 0.00000503, -0.00001371, 0.00000048, 0.00000403, -0.00000495, 0.00000606, -0.00000147, -0.00000644, -0.00000290, -0.00000470, 0.00001091, -0.00000478, -0.00001541, -0.00000252, 0.00000606, 0.00001408, -0.00002214, 0.00000435, -0.00000614, -0.00003021, -0.00000150, 0.00000048, 0.00004692, 0.00002509, 0.00001313, 0.00004298, 0.00002374, -0.00001475, -0.00003410, -0.00000633, -0.00000871, -0.00000738, -0.00000298, -0.00000552, -0.00000381, -0.00000501, -0.00000368, -0.00000617, -0.00000902, -0.00000764, },
		{ -0.00006022, -0.00000498, 0.00007330, 0.00004321, 0.00004327, 0.00007010, 0.00001094, -0.00004378, -0.00002594, -0.00000347, -0.00000190, -0.00001029, -0.00000112, 0.00000287, 0.00000364, -0.00000842, -0.00000974, 0.00000414, -0.00000109, 0.00000253, -0.00001241, -0.00000214, 0.00000531, -0.00000498, 0.00000290, -0.00000315, -0.00000304, -0.00000592, -0.00000208, -0.00000332, -0.00001001, 0.00001413, 0.00001716, -0.00001463, -0.00000973, -0.00000052, 0.00000306, 0.00000671, -0.00000469, 0.00000490, 0.00000122, 0.00000223, -0.00000404, -0.00000485, 0.00000235, -0.00000054, 0.00001254, -0.00000564, -0.00000282, 0.00000413, -0.00001048, -0.00000186, 0.00000632, 0.00000840, 0.00000344, -0.00000970, -0.00000200, 0.00000509, -0.00001115, -0.00000034, 0.00000944, 0.00000299, -0.00000396, 0.00000317, -0.00000144, -0.00000616, 0.00000410, -0.00000552, 0.00000535, -0.00000041, 0.00000261, -0.00000269, -0.00000541, 0.00000588, -0.00000216, 0.00000400, -0.00000687, 0.00001455, -0.00000122, -0.00000884, -0.00000518, -0.00001484, 0.00002332, 0.00000242, -0.00000021, -0.00000363, -0.00000397, 0.00000393, -0.00001548, 0.00000487, 0.00000074, -0.00000295, 0.00000625, -0.00000379, -0.00000451, -0.00000447, -0.00000261, 0.00001147, -0.00000867, -0.00001357, -0.00000256, 0.00001045, 0.00000932, -0.00002318, 0.00000985, -0.00001588, -0.00002426, 0.00000188, 0.00000407, 0.00004979, 0.00001531, 0.00001700, 0.00004179, 0.00001727, -0.00001737, -0.00003118, -0.00000441, -0.00000993, -0.00000561, -0.00000382, -0.00000496, -0.00000422, -0.00000463, -0.00000388, -0.00000695, -0.00000891, -0.00000733, -0.00001632, },
		{ 0.00000875, 0.00007881, 0.00003945, 0.00005342, 0.00006904, -0.00000594, -0.00004746, -0.00001954, -0.00000182, -0.00000427, -0.00000850, -0.00000092, 0.00000264, 0.00000339, -0.00001060, -0.00000648, 0.00000285, -0.00000035, 0.00000145, -0.00001385, 0.00000154, 0.00000265, -0.00000383, 0.00000243, -0.00000406, -0.00000216, -0.00000692, -0.00000092, -0.00000576, -0.00000758, 0.00001832, 0.00001126, -0.00001616, -0.00000732, 0.00000017, 0.00000516, 0.00000362, -0.00000354, 0.00000567, 0.00000034, 0.00000248, -0.00000592, -0.00000269, 0.00000111, 0.00000177, 0.00001207, -0.00000837, 0.00000015, 0.00000208, -0.00001118, 0.00000001, 0.00000754, 0.00000886, 0.00000023, -0.00001062, 0.00000190, 0.00000196, -0.00001171, 0.00000361, 0.00000894, 0.00000119, -0.00000481, 0.00000505, -0.00000337, -0.00000483, 0.00000381, -0.00000526, 0.00000570, -0.00000170, 0.00000392, -0.00000524, -0.00000248, 0.00000488, -0.00000184, 0.00000268, -0.00000540, 0.00001654, -0.00000664, -0.00000624, -0.00000888, -0.00000844, 0.00002371, -0.00000277, 0.00000352, -0.00000725, -0.00000015, 0.00000081, -0.00001517, 0.00000819, -0.00000279, 0.00000021, 0.00000476, -0.00000475, -0.00000316, -0.00000610, 0.00000039, 0.00001046, -0.00001151, -0.00001208, -0.00000187, 0.00001461, 0.00000266, -0.00002089, 0.00001221, -0.00002478, -0.00001541, 0.00000246, 0.00000914, 0.00005000, 0.00000628, 0.00002251, 0.00003851, 0.00001182, -0.00001935, -0.00002806, -0.00000325, -0.00001067, -0.00000402, -0.00000482, -0.00000417, -0.00000492, -0.00000388, -0.00000461, -0.00000736, -0.00000879, -0.00000734, -0.00002131, -0.00006410, },
		{ 0.00008159, 0.00003712, 0.00006504, 0.00006284, -0.00002325, -0.00004714, -0.00001305, -0.00000198, -0.00000564, -0.00000669, -0.00000145, 0.00000321, 0.00000224, -0.00001176, -0.00000360, 0.00000122, 0.00000104, -0.00000078, -0.00001379, 0.00000430, -0.00000003, -0.00000211, 0.00000104, -0.00000395, -0.00000198, -0.00000753, -0.00000011, -0.00000830, -0.00000347, 0.00002041, 0.00000517, -0.00001618, -0.00000549, 0.00000157, 0.00000608, 0.00000069, -0.00000150, 0.00000546, 0.00000004, 0.00000209, -0.00000713, -0.00000075, -0.00000017, 0.00000467, 0.00001024, -0.00000999, 0.00000293, -0.00000062, -0.00001112, 0.00000195, 0.00000880, 0.00000860, -0.00000344, -0.00000984, 0.00000495, -0.00000209, -0.00001040, 0.00000682, 0.00000795, -0.00000084, -0.00000487, 0.00000678, -0.00000572, -0.00000288, 0.00000302, -0.00000461, 0.00000540, -0.00000226, 0.00000460, -0.00000736, 0.00000074, 0.00000315, -0.00000108, 0.00000095, -0.00000262, 0.00001653, -0.00001103, -0.00000391, -0.00001218, -0.00000066, 0.00002064, -0.00000523, 0.00000617, -0.00001074, 0.00000439, -0.00000404, -0.00001246, 0.00000959, -0.00000552, 0.00000368, 0.00000200, -0.00000418, -0.00000298, -0.00000704, 0.00000358, 0.00000825, -0.00001324, -0.00001121, 0.00000011, 0.00001742, -0.00000469, -0.00001577, 0.00001068, -0.00003094, -0.00000538, 0.00000061, 0.00001595, 0.00004692, -0.00000084, 0.00002873, 0.00003339, 0.00000768, -0.00002098, -0.00002475, -0.00000282, -0.00001083, -0.00000287, -0.00000571, -0.00000334, -0.00000576, -0.00000285, -0.00000590, -0.00000719, -0.00000895, -0.00000754, -0.00002719, -0.00006656, 0.00002463, },
		{ 0.00003737, 0.00007678, 0.00005092, -0.00003911, -0.00004298, -0.00000767, -0.00000331, -0.00000586, -0.00000544, -0.00000214, 0.00000422, 0.00000034, -0.00001180, -0.00000151, -0.00000021, 0.00000254, -0.00000372, -0.00001231, 0.00000580, -0.00000218, -0.00000040, -0.00000083, -0.00000299, -0.00000261, -0.00000749, -0.00000005, -0.00001022, 0.00000170, 0.00002017, -0.00000025, -0.00001529, -0.00000397, 0.00000347, 0.00000564, -0.00000142, 0.00000084, 0.00000453, 0.00000026, 0.00000107, -0.00000751, 0.00000065, -0.00000107, 0.00000772, 0.00000727, -0.00001029, 0.00000509, -0.00000366, -0.00001029, 0.00000387, 0.00001005, 0.00000740, -0.00000703, -0.00000746, 0.00000649, -0.00000620, -0.00000752, 0.00000896, 0.00000668, -0.00000309, -0.00000391, 0.00000789, -0.00000814, -0.00000043, 0.00000171, -0.00000360, 0.00000453, -0.00000207, 0.00000447, -0.00000868, 0.00000378, 0.00000097, -0.00000004, -0.00000087, 0.00000108, 0.00001451, -0.00001391, -0.00000242, -0.00001414, 0.00000718, 0.00001487, -0.00000457, 0.00000669, -0.00001294, 0.00000852, -0.00000977, -0.00000761, 0.00000862, -0.00000655, 0.00000642, -0.00000122, -0.00000242, -0.00000415, -0.00000670, 0.00000616, 0.00000548, -0.00001419, -0.00001078, 0.00000353, 0.00001798, -0.00001125, -0.00000905, 0.00000541, -0.00003296, 0.00000385, -0.00000259, 0.00002414, 0.00004039, -0.00000504, 0.00003451, 0.00002699, 0.00000488, -0.00002246, -0.00002131, -0.00000302, -0.00001036, -0.00000232, -0.00000625, -0.00000271, -0.00000650, -0.00000175, -0.00000764, -0.00000633, -0.00000969, -0.00000772, -0.00003418, -0.00006692, 0.00004217, 0.00008126, },
		{ 0.00008677, 0.00003345, -0.00005158, -0.00003584, -0.00000420, -0.00000498, -0.00000515, -0.00000502, -0.00000249, 0.00000524, -0.00000198, -0.00001086, -0.00000037, -0.00000105, 0.00000362, -0.00000681, -0.00000977, 0.00000599, -0.00000345, 0.00000080, -0.00000264, -0.00000155, -0.00000392, -0.00000680, -0.00000093, -0.00001086, 0.00000701, 0.00001785, -0.00000443, -0.00001406, -0.00000237, 0.00000540, 0.00000401, -0.00000233, 0.00000291, 0.00000325, 0.00000078, -0.00000042, -0.00000709, 0.00000136, -0.00000126, 0.00001037, 0.00000356, -0.00000927, 0.00000628, -0.00000666, -0.00000880, 0.00000572, 0.00001116, 0.00000515, -0.00000993, -0.00000393, 0.00000619, -0.00000951, -0.00000364, 0.00000998, 0.00000526, -0.00000543, -0.00000186, 0.00000798, -0.00001017, 0.00000224, -0.00000002, -0.00000230, 0.00000322, -0.00000119, 0.00000347, -0.00000895, 0.00000624, -0.00000135, 0.00000109, -0.00000244, 0.00000514, 0.00001079, -0.00001513, -0.00000207, -0.00001408, 0.00001366, 0.00000776, -0.00000119, 0.00000462, -0.00001290, 0.00001107, -0.00001516, -0.00000145, 0.00000545, -0.00000541, 0.00000753, -0.00000387, -0.00000023, -0.00000635, -0.00000491, 0.00000755, 0.00000286, -0.00001491, -0.00001028, 0.00000801, 0.00001588, -0.00001570, -0.00000241, -0.00000251, -0.00003033, 0.00001053, -0.00000565, 0.00003277, 0.00003080, -0.00000568, 0.00003879, 0.00002011, 0.00000321, -0.00002387, -0.00001780, -0.00000371, -0.00000933, -0.00000244, -0.00000624, -0.00000254, -0.00000685, -0.00000092, -0.00000954, -0.00000490, -0.00001113, -0.00000771, -0.00004251, -0.00006436, 0.00006053, 0.00007785, 0.00004127, },
		{ 0.00001155, -0.00005906, -0.00002715, -0.00000288, -0.00000622, -0.00000402, -0.00000537, -0.00000225, 0.00000589, -0.00000431, -0.00000926, -0.00000008, -0.00000118, 0.00000394, -0.00000949, -0.00000671, 0.00000512, -0.00000376, 0.00000119, -0.00000393, -0.00000012, -0.00000558, -0.00000569, -0.00000265, -0.00000984, 0.00001153, 0.00001409, -0.00000723, -0.00001283, -0.00000036, 0.00000682, 0.00000165, -0.00000199, 0.00000430, 0.00000201, 0.00000133, -0.00000212, -0.00000603, 0.00000140, -0.00000054, 0.00001217, -0.00000035, -0.00000717, 0.00000630, -0.00000924, -0.00000680, 0.00000754, 0.00001192, 0.00000191, -0.00001155, 0.00000005, 0.00000409, -0.00001136, 0.00000054, 0.00001005, 0.00000366, -0.00000759, 0.00000114, 0.00000676, -0.00001133, 0.00000473, -0.00000196, -0.00000085, 0.00000166, 0.00000019, 0.00000169, -0.00000803, 0.00000774, -0.00000346, 0.00000215, -0.00000348, 0.00000897, 0.00000591, -0.00001487, -0.00000280, -0.00001172, 0.00001760, 0.00000092, 0.00000386, 0.00000019, -0.00001023, 0.00001109, -0.00001891, 0.00000475, 0.00000083, -0.00000220, 0.00000650, -0.00000504, 0.00000141, -0.00000883, -0.00000198, 0.00000752, 0.00000091, -0.00001598, -0.00000893, 0.00001268, 0.00001140, -0.00001722, 0.00000248, -0.00001131, -0.00002357, 0.00001353, -0.00000704, 0.00004044, 0.00001916, -0.00000269, 0.00004070, 0.00001359, 0.00000227, -0.00002514, -0.00001438, -0.00000468, -0.00000793, -0.00000316, -0.00000562, -0.00000300, -0.00000657, -0.00000073, -0.00001114, -0.00000324, -0.00001320, -0.00000749, -0.00005232, -0.00005798, 0.00007849, 0.00007192, 0.00004949, 0.00009277, },
		{ -0.00006067, -0.00001857, -0.00000336, -0.00000658, -0.00000301, -0.00000619, -0.00000138, 0.00000595, -0.00000630, -0.00000741, -0.00000037, -0.00000068, 0.00000337, -0.00001134, -0.00000369, 0.00000363, -0.00000332, 0.00000074, -0.00000439, 0.00000088, -0.00000721, -0.00000454, -0.00000477, -0.00000718, 0.00001452, 0.00000973, -0.00000884, -0.00001171, 0.00000217, 0.00000724, -0.00000080, -0.00000062, 0.00000481, 0.00000109, 0.00000163, -0.00000372, -0.00000463, 0.00000095, 0.00000106, 0.00001277, -0.00000388, -0.00000437, 0.00000517, -0.00001113, -0.00000449, 0.00000935, 0.00001205, -0.00000203, -0.00001154, 0.00000366, 0.00000065, -0.00001140, 0.00000435, 0.00000944, 0.00000179, -0.00000919, 0.00000472, 0.00000423, -0.00001126, 0.00000664, -0.00000385, 0.00000054, 0.00000011, 0.00000180, -0.00000065, -0.00000598, 0.00000804, -0.00000504, 0.00000297, -0.00000378, 0.00001201, 0.00000050, -0.00001354, -0.00000418, -0.00000733, 0.00001829, -0.00000412, 0.00000910, -0.00000569, -0.00000518, 0.00000817, -0.00001996, 0.00000964, -0.00000404, 0.00000236, 0.00000340, -0.00000419, 0.00000160, -0.00001063, 0.00000136, 0.00000635, -0.00000024, -0.00001771, -0.00000605, 0.00001637, 0.00000550, -0.00001578, 0.00000438, -0.00001897, -0.00001409, 0.00001267, -0.00000566, 0.00004570, 0.00000692, 0.00000343, 0.00003980, 0.00000820, 0.00000158, -0.00002610, -0.00001120, -0.00000568, -0.00000642, -0.00000426, -0.00000448, -0.00000408, -0.00000549, -0.00000151, -0.00001196, -0.00000186, -0.00001557, -0.00000732, -0.00006349, -0.00004695, 0.00009456, 0.00006468, 0.00006199, 0.00009256, -0.00001273, },
		{ -0.00001162, -0.00000485, -0.00000600, -0.00000254, -0.00000707, -0.00000007, 0.00000539, -0.00000775, -0.00000561, -0.00000093, 0.00000017, 0.00000204, -0.00001219, -0.00000116, 0.00000197, -0.00000250, -0.00000035, -0.00000400, 0.00000118, -0.00000849, -0.00000374, -0.00000670, -0.00000328, 0.00001558, 0.00000556, -0.00000967, -0.00001052, 0.00000508, 0.00000640, -0.00000271, 0.00000132, 0.00000449, 0.00000064, 0.00000152, -0.00000494, -0.00000320, 0.00000033, 0.00000331, 0.00001207, -0.00000657, -0.00000138, 0.00000307, -0.00001214, -0.00000203, 0.00001117, 0.00001126, -0.00000617, -0.00000981, 0.00000615, -0.00000340, -0.00000967, 0.00000731, 0.00000846, -0.00000046, -0.00000983, 0.00000831, 0.00000062, -0.00000982, 0.00000759, -0.00000537, 0.00000167, -0.00000119, 0.00000332, -0.00000319, -0.00000303, 0.00000706, -0.00000586, 0.00000345, -0.00000328, 0.00001386, -0.00000477, -0.00001163, -0.00000559, -0.00000167, 0.00001574, -0.00000624, 0.00001298, -0.00001166, 0.00000132, 0.00000261, -0.00001781, 0.00001205, -0.00000779, 0.00000710, -0.00000107, -0.00000139, -0.00000017, -0.00001085, 0.00000413, 0.00000471, -0.00000092, -0.00001998, -0.00000129, 0.00001795, -0.00000042, -0.00001216, 0.00000286, -0.00002372, -0.00000389, 0.00000869, -0.00000111, 0.00004735, -0.00000429, 0.00001163, 0.00003612, 0.00000444, 0.00000070, -0.00002658, -0.00000844, -0.00000647, -0.00000512, -0.00000541, -0.00000310, -0.00000563, -0.00000369, -0.00000342, -0.00001154, -0.00000142, -0.00001765, -0.00000776, -0.00007547, -0.00003082, 0.00010720, 0.00005792, 0.00007779, 0.00008437, -0.00003662, -0.00005652, },
		{ -0.00000640, -0.00000482, -0.00000273, -0.00000770, 0.00000142, 0.00000435, -0.00000863, -0.00000406, -0.00000149, 0.00000107, 0.00000023, -0.00001210, 0.00000066, 0.00000053, -0.00000167, -0.00000171, -0.00000292, 0.00000074, -0.00000929, -0.00000354, -0.00000784, 0.00000114, 0.00001478, 0.00000215, -0.00001018, -0.00000895, 0.00000797, 0.00000435, -0.00000362, 0.00000330, 0.00000358, 0.00000063, 0.00000096, -0.00000563, -0.00000202, -0.00000011, 0.00000583, 0.00001018, -0.00000810, 0.00000129, 0.00000034, -0.00001226, 0.00000047, 0.00001290, 0.00000935, -0.00000990, -0.00000664, 0.00000698, -0.00000723, -0.00000656, 0.00000917, 0.00000731, -0.00000310, -0.00000917, 0.00001127, -0.00000355, -0.00000712, 0.00000738, -0.00000627, 0.00000234, -0.00000200, 0.00000442, -0.00000550, 0.00000044, 0.00000490, -0.00000575, 0.00000353, -0.00000201, 0.00001428, -0.00000940, -0.00000961, -0.00000636, 0.00000417, 0.00001063, -0.00000502, 0.00001428, -0.00001623, 0.00000789, -0.00000464, -0.00001267, 0.00001138, -0.00000927, 0.00001067, -0.00000572, 0.00000270, -0.00000384, -0.00000899, 0.00000546, 0.00000349, -0.00000184, -0.00002217, 0.00000511, 0.00001671, -0.00000488, -0.00000782, -0.00000156, -0.00002450, 0.00000498, 0.00000306, 0.00000620, 0.00004480, -0.00001296, 0.00002059, 0.00003015, 0.00000248, -0.00000069, -0.00002644, -0.00000619, -0.00000691, -0.00000429, -0.00000626, -0.00000184, -0.00000728, -0.00000145, -0.00000635, -0.00000966, -0.00000248, -0.00001872, -0.00000970, -0.00008721, -0.00000968, 0.00011515, 0.00005374, 0.00009480, 0.00006745, -0.00005707, -0.00004774, -0.00000729, },
		{ -0.00000355, -0.00000345, -0.00000791, 0.00000287, 0.00000300, -0.00000903, -0.00000280, -0.00000192, 0.00000178, -0.00000171, -0.00001132, 0.00000176, -0.00000052, -0.00000115, -0.00000297, -0.00000146, -0.00000032, -0.00000965, -0.00000395, -0.00000774, 0.00000526, 0.00001255, -0.00000027, -0.00001068, -0.00000669, 0.00001029, 0.00000144, -0.00000331, 0.00000483, 0.00000242, 0.00000090, 0.00000002, -0.00000576, -0.00000123, -0.00000009, 0.00000818, 0.00000743, -0.00000838, 0.00000323, -0.00000261, -0.00001158, 0.00000298, 0.00001440, 0.00000627, -0.00001259, -0.00000259, 0.00000598, -0.00001004, -0.00000267, 0.00000995, 0.00000606, -0.00000599, -0.00000708, 0.00001299, -0.00000762, -0.00000351, 0.00000603, -0.00000641, 0.00000243, -0.00000214, 0.00000482, -0.00000717, 0.00000389, 0.00000186, -0.00000470, 0.00000318, -0.00000013, 0.00001325, -0.00001303, -0.00000778, -0.00000599, 0.00000898, 0.00000421, -0.00000084, 0.00001239, -0.00001817, 0.00001300, -0.00001212, -0.00000548, 0.00000774, -0.00000783, 0.00001189, -0.00000918, 0.00000687, -0.00000871, -0.00000511, 0.00000484, 0.00000347, -0.00000384, -0.00002324, 0.00001224, 0.00001264, -0.00000676, -0.00000444, -0.00000748, -0.00002124, 0.00001092, -0.00000243, 0.00001510, 0.00003819, -0.00001801, 0.00002891, 0.00002270, 0.00000222, -0.00000272, -0.00002562, -0.00000448, -0.00000696, -0.00000406, -0.00000651, -0.00000109, -0.00000860, 0.00000076, -0.00000991, -0.00000641, -0.00000539, -0.00001813, -0.00001416, -0.00009717, 0.00001563, 0.00011777, 0.00005416, 0.00011005, 0.00004235, -0.00007133, -0.00003632, -0.00000583, -0.00000720, },
		{ -0.00000438, -0.00000773, 0.00000416, 0.00000148, -0.00000907, -0.00000176, -0.00000223, 0.00000220, -0.00000352, -0.00001015, 0.00000230, -0.00000116, -0.00000107, -0.00000382, 0.00000001, -0.00000177, -0.00000977, -0.00000472, -0.00000630, 0.00000834, 0.00000957, -0.00000180, -0.00001125, -0.00000358, 0.00001151, -0.00000173, -0.00000187, 0.00000561, 0.00000134, 0.00000123, -0.00000109, -0.00000544, -0.00000084, 0.00000056, 0.00000993, 0.00000427, -0.00000756, 0.00000420, -0.00000541, -0.00001026, 0.00000553, 0.00001537, 0.00000219, -0.00001373, 0.00000157, 0.00000336, -0.00001132, 0.00000132, 0.00000988, 0.00000463, -0.00000883, -0.00000366, 0.00001306, -0.00001091, 0.00000044, 0.00000375, -0.00000582, 0.00000193, -0.00000158, 0.00000437, -0.00000782, 0.00000679, -0.00000162, -0.00000285, 0.00000246, 0.00000216, 0.00001088, -0.00001552, -0.00000624, -0.00000428, 0.00001178, -0.00000202, 0.00000526, 0.00000745, -0.00001682, 0.00001536, -0.00001820, 0.00000229, 0.00000201, -0.00000359, 0.00001012, -0.00001027, 0.00000976, -0.00001358, 0.00000007, 0.00000235, 0.00000504, -0.00000761, -0.00002207, 0.00001871, 0.00000654, -0.00000566, -0.00000348, -0.00001302, -0.00001491, 0.00001317, -0.00000615, 0.00002402, 0.00002838, -0.00001896, 0.00003540, 0.00001472, 0.00000327, -0.00000537, -0.00002415, -0.00000323, -0.00000672, -0.00000441, -0.00000603, -0.00000115, -0.00000913, 0.00000232, -0.00001343, -0.00000229, -0.00001008, -0.00001554, -0.00002209, -0.00010351, 0.00004352, 0.00011535, 0.00006060, 0.00012005, 0.00001115, -0.00007751, -0.00002466, -0.00000669, -0.00000683, -0.00000273, },
		{ -0.00000731, 0.00000530, -0.00000017, -0.00000880, -0.00000087, -0.00000253, 0.00000238, -0.00000505, -0.00000885, 0.00000249, -0.00000154, -0.00000139, -0.00000412, 0.00000124, -0.00000339, -0.00000989, -0.00000543, -0.00000374, 0.00000993, 0.00000657, -0.00000279, -0.00001170, 0.00000025, 0.00001127, -0.00000452, 0.00000034, 0.00000553, 0.00000057, 0.00000141, -0.00000217, -0.00000486, -0.00000071, 0.00000183, 0.00001079, 0.00000121, -0.00000594, 0.00000413, -0.00000774, -0.00000847, 0.00000813, 0.00001551, -0.00000249, -0.00001306, 0.00000502, -0.00000033, -0.00001084, 0.00000481, 0.00000926, 0.00000289, -0.00001117, 0.00000069, 0.00001134, -0.00001283, 0.00000410, 0.00000094, -0.00000465, 0.00000099, -0.00000041, 0.00000307, -0.00000725, 0.00000863, -0.00000499, -0.00000044, 0.00000146, 0.00000459, 0.00000743, -0.00001685, -0.00000487, -0.00000140, 0.00001197, -0.00000660, 0.00001176, 0.00000038, -0.00001225, 0.00001424, -0.00002148, 0.00000897, -0.00000438, 0.00000260, 0.00000548, -0.00000831, 0.00001014, -0.00001702, 0.00000533, -0.00000130, 0.00000797, -0.00001335, -0.00001779, 0.00002291, -0.00000006, -0.00000215, -0.00000570, -0.00001630, -0.00000726, 0.00001195, -0.00000705, 0.00003133, 0.00001676, -0.00001602, 0.00003925, 0.00000713, 0.00000516, -0.00000849, -0.00002214, -0.00000232, -0.00000639, -0.00000512, -0.00000491, -0.00000210, -0.00000858, 0.00000269, -0.00001613, 0.00000188, -0.00001600, -0.00001107, -0.00003405, -0.00010436, 0.00007181, 0.00010924, 0.00007335, 0.00012144, -0.00002275, -0.00007507, -0.00001505, -0.00000874, -0.00000539, -0.00000264, -0.00000520, },
		{ 0.00000639, -0.00000198, -0.00000821, -0.00000012, -0.00000290, 0.00000245, -0.00000632, -0.00000756, 0.00000252, -0.00000185, -0.00000193, -0.00000389, 0.00000207, -0.00000505, -0.00001018, -0.00000560, -0.00000062, 0.00000995, 0.00000410, -0.00000366, -0.00001169, 0.00000443, 0.00000948, -0.00000634, 0.00000281, 0.00000473, 0.00000021, 0.00000132, -0.00000303, -0.00000424, -0.00000061, 0.00000352, 0.00001064, -0.00000134, -0.00000395, 0.00000316, -0.00000941, -0.00000635, 0.00001075, 0.00001451, -0.00000717, -0.00001066, 0.00000711, -0.00000433, -0.00000877, 0.00000739, 0.00000837, 0.00000068, -0.00001255, 0.00000536, 0.00000804, -0.00001307, 0.00000688, -0.00000191, -0.00000321, -0.00000018, 0.00000113, 0.00000111, -0.00000545, 0.00000902, -0.00000770, 0.00000215, 0.00000035, 0.00000690, 0.00000322, -0.00001711, -0.00000344, 0.00000211, 0.00000955, -0.00000843, 0.00001705, -0.00000737, -0.00000533, 0.00000971, -0.00002113, 0.00001305, -0.00000974, 0.00000929, -0.00000115, -0.00000344, 0.00000735, -0.00001778, 0.00000922, -0.00000487, 0.00001140, -0.00002054, -0.00001013, 0.00002350, -0.00000532, 0.00000236, -0.00001082, -0.00001603, -0.00000029, 0.00000840, -0.00000489, 0.00003578, 0.00000492, -0.00000991, 0.00004013, 0.00000066, 0.00000736, -0.00001187, -0.00001966, -0.00000164, -0.00000621, -0.00000588, -0.00000342, -0.00000379, -0.00000690, 0.00000151, -0.00001726, 0.00000508, -0.00002219, -0.00000547, -0.00004997, -0.00009820, 0.00009805, 0.00010171, 0.00009126, 0.00011170, -0.00005510, -0.00006499, -0.00000915, -0.00001060, -0.00000345, -0.00000327, -0.00000570, -0.00000681, },
		{ -0.00000406, -0.00000718, 0.00000038, -0.00000337, 0.00000256, -0.00000744, -0.00000630, 0.00000248, -0.00000223, -0.00000248, -0.00000326, 0.00000249, -0.00000674, -0.00001062, -0.00000487, 0.00000239, 0.00000871, 0.00000243, -0.00000475, -0.00001081, 0.00000836, 0.00000639, -0.00000683, 0.00000503, 0.00000347, 0.00000022, 0.00000094, -0.00000357, -0.00000374, -0.00000031, 0.00000533, 0.00000958, -0.00000308, -0.00000202, 0.00000154, -0.00001038, -0.00000398, 0.00001327, 0.00001222, -0.00001119, -0.00000693, 0.00000745, -0.00000784, -0.00000554, 0.00000892, 0.00000738, -0.00000201, -0.00001259, 0.00000962, 0.00000366, -0.00001159, 0.00000837, -0.00000435, -0.00000183, -0.00000126, 0.00000271, -0.00000118, -0.00000265, 0.00000784, -0.00000926, 0.00000449, -0.00000067, 0.00000882, -0.00000140, -0.00001643, -0.00000173, 0.00000549, 0.00000506, -0.00000708, 0.00001978, -0.00001419, 0.00000251, 0.00000259, -0.00001711, 0.00001362, -0.00001258, 0.00001476, -0.00000823, 0.00000344, 0.00000151, -0.00001519, 0.00001049, -0.00000695, 0.00001399, -0.00002798, 0.00000030, 0.00001983, -0.00000752, 0.00000588, -0.00001755, -0.00001191, 0.00000424, 0.00000418, -0.00000032, 0.00003676, -0.00000568, -0.00000171, 0.00003811, -0.00000421, 0.00000942, -0.00001533, -0.00001680, -0.00000117, -0.00000632, -0.00000634, -0.00000199, -0.00000583, -0.00000436, -0.00000124, -0.00001633, 0.00000631, -0.00002740, -0.00000004, -0.00006892, -0.00008428, 0.00012008, 0.00009561, 0.00011160, 0.00008983, -0.00008152, -0.00004961, -0.00000760, -0.00001108, -0.00000183, -0.00000424, -0.00000591, -0.00000629, 0.00000746, },
		{ -0.00000555, 0.00000041, -0.00000380, 0.00000276, -0.00000856, -0.00000498, 0.00000232, -0.00000273, -0.00000287, -0.00000242, 0.00000255, -0.00000856, -0.00001099, -0.00000312, 0.00000462, 0.00000677, 0.00000151, -0.00000615, -0.00000879, 0.00001139, 0.00000255, -0.00000595, 0.00000656, 0.00000210, 0.00000044, 0.00000036, -0.00000382, -0.00000341, 0.00000038, 0.00000691, 0.00000787, -0.00000394, -0.00000050, -0.00000041, -0.00001069, -0.00000136, 0.00001543, 0.00000866, -0.00001392, -0.00000255, 0.00000600, -0.00001019, -0.00000178, 0.00000948, 0.00000632, -0.00000504, -0.00001108, 0.00001278, -0.00000109, -0.00000871, 0.00000842, -0.00000604, -0.00000079, -0.00000196, 0.00000397, -0.00000334, 0.00000070, 0.00000521, -0.00000937, 0.00000620, -0.00000135, 0.00001007, -0.00000609, -0.00001489, 0.00000039, 0.00000793, -0.00000045, -0.00000280, 0.00001920, -0.00001860, 0.00000960, -0.00000566, -0.00001023, 0.00001057, -0.00001194, 0.00001748, -0.00001400, 0.00001074, -0.00000642, -0.00000936, 0.00000845, -0.00000636, 0.00001426, -0.00003392, 0.00001200, 0.00001227, -0.00000562, 0.00000644, -0.00002381, -0.00000477, 0.00000537, 0.00000097, 0.00000545, 0.00003439, -0.00001397, 0.00000741, 0.00003359, -0.00000722, 0.00001100, -0.00001868, -0.00001357, -0.00000098, -0.00000676, -0.00000625, -0.00000103, -0.00000769, -0.00000150, -0.00000519, -0.00001329, 0.00000489, -0.00003040, 0.00000353, -0.00008921, -0.00006286, 0.00013646, 0.00009370, 0.00013037, 0.00005685, -0.00009835, -0.00003224, -0.00000989, -0.00000949, -0.00000128, -0.00000502, -0.00000607, -0.00000563, 0.00000844, -0.00000641, },
		{ -0.00000024, -0.00000394, 0.00000296, -0.00000972, -0.00000349, 0.00000193, -0.00000327, -0.00000303, -0.00000149, 0.00000228, -0.00001062, -0.00001092, -0.00000058, 0.00000564, 0.00000478, 0.00000101, -0.00000771, -0.00000559, 0.00001298, -0.00000136, -0.00000392, 0.00000719, 0.00000090, 0.00000068, -0.00000028, -0.00000389, -0.00000319, 0.00000150, 0.00000797, 0.00000588, -0.00000401, 0.00000042, -0.00000239, -0.00001043, 0.00000150, 0.00001692, 0.00000409, -0.00001492, 0.00000169, 0.00000312, -0.00001097, 0.00000190, 0.00000932, 0.00000507, -0.00000810, -0.00000810, 0.00001434, -0.00000543, -0.00000499, 0.00000714, -0.00000683, -0.00000029, -0.00000208, 0.00000461, -0.00000491, 0.00000402, 0.00000153, -0.00000797, 0.00000696, -0.00000150, 0.00001044, -0.00001049, -0.00001257, 0.00000289, 0.00000878, -0.00000575, 0.00000350, 0.00001528, -0.00001968, 0.00001440, -0.00001329, -0.00000194, 0.00000464, -0.00000773, 0.00001648, -0.00001684, 0.00001659, -0.00001480, -0.00000128, 0.00000324, -0.00000250, 0.00001102, -0.00003654, 0.00002285, 0.00000224, 0.00000035, 0.00000266, -0.00002728, 0.00000354, 0.00000321, -0.00000002, 0.00001095, 0.00002940, -0.00001943, 0.00001644, 0.00002712, -0.00000830, 0.00001178, -0.00002174, -0.00000998, -0.00000125, -0.00000735, -0.00000559, -0.00000084, -0.00000885, 0.00000099, -0.00000960, -0.00000861, 0.00000066, -0.00003030, 0.00000353, -0.00010851, -0.00003523, 0.00014698, 0.00009798, 0.00014293, 0.00001585, -0.00010344, -0.00001640, -0.00001451, -0.00000595, -0.00000225, -0.00000506, -0.00000658, -0.00000455, 0.00000908, -0.00000889, -0.00000328, },
		{ -0.00000352, 0.00000294, -0.00001085, -0.00000181, 0.00000119, -0.00000370, -0.00000300, -0.00000052, 0.00000167, -0.00001293, -0.00001000, 0.00000227, 0.00000535, 0.00000328, 0.00000048, -0.00000903, -0.00000150, 0.00001283, -0.00000463, -0.00000122, 0.00000692, 0.00000006, 0.00000080, -0.00000084, -0.00000391, -0.00000293, 0.00000297, 0.00000832, 0.00000398, -0.00000356, 0.00000069, -0.00000418, -0.00000969, 0.00000461, 0.00001738, -0.00000100, -0.00001405, 0.00000503, -0.00000059, -0.00001013, 0.00000499, 0.00000873, 0.00000348, -0.00001075, -0.00000402, 0.00001407, -0.00000869, -0.00000109, 0.00000490, -0.00000676, -0.00000036, -0.00000154, 0.00000445, -0.00000552, 0.00000667, -0.00000260, -0.00000527, 0.00000665, -0.00000097, 0.00000972, -0.00001424, -0.00000953, 0.00000551, 0.00000772, -0.00000961, 0.00001048, 0.00000872, -0.00001724, 0.00001588, -0.00001867, 0.00000595, -0.00000270, -0.00000070, 0.00001166, -0.00001574, 0.00001930, -0.00002166, 0.00000736, -0.00000410, 0.00000435, 0.00000381, -0.00003437, 0.00003050, -0.00000797, 0.00000911, -0.00000574, -0.00002607, 0.00001064, -0.00000107, 0.00000156, 0.00001496, 0.00002288, -0.00002201, 0.00002457, 0.00001929, -0.00000748, 0.00001150, -0.00002425, -0.00000610, -0.00000218, -0.00000778, -0.00000452, -0.00000145, -0.00000901, 0.00000245, -0.00001356, -0.00000317, -0.00000592, -0.00002685, -0.00000137, -0.00012431, -0.00000355, 0.00015270, 0.00010896, 0.00014493, -0.00002837, -0.00009662, -0.00000508, -0.00001939, -0.00000126, -0.00000466, -0.00000410, -0.00000775, -0.00000269, 0.00000900, -0.00001119, -0.00000051, -0.00000172, },
		{ 0.00000241, -0.00001175, -0.00000003, 0.00000007, -0.00000386, -0.00000290, 0.00000050, 0.00000057, -0.00001535, -0.00000792, 0.00000477, 0.00000404, 0.00000255, -0.00000047, -0.00000960, 0.00000291, 0.00001099, -0.00000671, 0.00000160, 0.00000595, -0.00000040, 0.00000074, -0.00000121, -0.00000404, -0.00000244, 0.00000454, 0.00000796, 0.00000246, -0.00000287, 0.00000041, -0.00000565, -0.00000852, 0.00000784, 0.00001654, -0.00000595, -0.00001149, 0.00000690, -0.00000436, -0.00000790, 0.00000719, 0.00000793, 0.00000142, -0.00001251, 0.00000058, 0.00001210, -0.00001044, 0.00000231, 0.00000218, -0.00000606, -0.00000084, -0.00000045, 0.00000351, -0.00000497, 0.00000811, -0.00000645, -0.00000172, 0.00000534, 0.00000027, 0.00000782, -0.00001702, -0.00000589, 0.00000784, 0.00000483, -0.00001115, 0.00001666, 0.00000080, -0.00001184, 0.00001375, -0.00002064, 0.00001173, -0.00000962, 0.00000763, 0.00000385, -0.00001055, 0.00001777, -0.00002521, 0.00001453, -0.00001184, 0.00001296, -0.00000686, -0.00002682, 0.00003301, -0.00001574, 0.00001837, -0.00001770, -0.00001932, 0.00001426, -0.00000556, 0.00000515, 0.00001681, 0.00001590, -0.00002205, 0.00003129, 0.00001066, -0.00000486, 0.00000990, -0.00002590, -0.00000213, -0.00000385, -0.00000771, -0.00000345, -0.00000262, -0.00000812, 0.00000249, -0.00001622, 0.00000184, -0.00001382, -0.00002052, -0.00001182, -0.00013438, 0.00002963, 0.00015587, 0.00012528, 0.00013322, -0.00007002, -0.00007981, -0.00000012, -0.00002245, 0.00000325, -0.00000788, -0.00000226, -0.00000968, 0.00000017, 0.00000777, -0.00001279, 0.00000243, -0.00000399, -0.00000230, },
		{ -0.00001211, 0.00000161, -0.00000136, -0.00000367, -0.00000282, 0.00000162, -0.00000122, -0.00001750, -0.00000466, 0.00000627, 0.00000224, 0.00000252, -0.00000200, -0.00000901, 0.00000695, 0.00000785, -0.00000733, 0.00000402, 0.00000461, -0.00000055, 0.00000053, -0.00000141, -0.00000431, -0.00000153, 0.00000594, 0.00000705, 0.00000146, -0.00000225, -0.00000024, -0.00000678, -0.00000691, 0.00001101, 0.00001428, -0.00001009, -0.00000774, 0.00000703, -0.00000747, -0.00000477, 0.00000845, 0.00000705, -0.00000112, -0.00001303, 0.00000501, 0.00000886, -0.00001055, 0.00000473, -0.00000052, -0.00000507, -0.00000150, 0.00000096, 0.00000198, -0.00000332, 0.00000800, -0.00000933, 0.00000205, 0.00000332, 0.00000207, 0.00000478, -0.00001851, -0.00000182, 0.00000940, 0.00000063, -0.00000997, 0.00002077, -0.00000695, -0.00000469, 0.00000850, -0.00001884, 0.00001415, -0.00001437, 0.00001535, -0.00000535, -0.00000216, 0.00001185, -0.00002429, 0.00001837, -0.00001789, 0.00002143, -0.00001944, -0.00001443, 0.00002936, -0.00001869, 0.00002524, -0.00003086, -0.00000756, 0.00001291, -0.00000814, 0.00000940, 0.00001660, 0.00000924, -0.00001995, 0.00003623, 0.00000179, -0.00000063, 0.00000676, -0.00002625, 0.00000151, -0.00000619, -0.00000687, -0.00000282, -0.00000389, -0.00000653, 0.00000108, -0.00001703, 0.00000531, -0.00002165, -0.00001256, -0.00002747, -0.00013724, 0.00006200, 0.00015935, 0.00014369, 0.00010677, -0.00010343, -0.00005668, -0.00000177, -0.00002217, 0.00000616, -0.00001092, -0.00000006, -0.00001212, 0.00000403, 0.00000508, -0.00001316, 0.00000498, -0.00000674, -0.00000023, 0.00000110, },
		{ 0.00000283, -0.00000287, -0.00000318, -0.00000278, 0.00000285, -0.00000389, -0.00001887, -0.00000051, 0.00000636, 0.00000064, 0.00000283, -0.00000400, -0.00000703, 0.00000991, 0.00000406, -0.00000656, 0.00000567, 0.00000321, -0.00000056, 0.00000028, -0.00000152, -0.00000467, -0.00000016, 0.00000688, 0.00000587, 0.00000097, -0.00000188, -0.00000106, -0.00000758, -0.00000481, 0.00001377, 0.00001071, -0.00001284, -0.00000351, 0.00000553, -0.00000938, -0.00000132, 0.00000888, 0.00000607, -0.00000396, -0.00001211, 0.00000860, 0.00000499, -0.00000924, 0.00000590, -0.00000277, -0.00000412, -0.00000204, 0.00000234, 0.00000025, -0.00000087, 0.00000629, -0.00001070, 0.00000538, 0.00000103, 0.00000413, 0.00000076, -0.00001850, 0.00000237, 0.00000979, -0.00000410, -0.00000625, 0.00002199, -0.00001307, 0.00000267, 0.00000132, -0.00001369, 0.00001276, -0.00001565, 0.00002060, -0.00001391, 0.00000774, 0.00000241, -0.00001867, 0.00001767, -0.00002038, 0.00002752, -0.00003164, 0.00000101, 0.00001987, -0.00001544, 0.00002702, -0.00004210, 0.00000719, 0.00000640, -0.00000716, 0.00001261, 0.00001508, 0.00000319, -0.00001607, 0.00003908, -0.00000671, 0.00000487, 0.00000199, -0.00002489, 0.00000425, -0.00000879, -0.00000525, -0.00000296, -0.00000472, -0.00000483, -0.00000134, -0.00001592, 0.00000645, -0.00002801, -0.00000466, -0.00004709, -0.00013244, 0.00009204, 0.00016583, 0.00015944, 0.00006712, -0.00012408, -0.00003189, -0.00000871, -0.00001804, 0.00000638, -0.00001265, 0.00000165, -0.00001448, 0.00000854, 0.00000092, -0.00001187, 0.00000651, -0.00000946, 0.00000249, -0.00000108, -0.00001170, },
		{ -0.00000415, -0.00000257, -0.00000271, 0.00000402, -0.00000748, -0.00001888, 0.00000393, 0.00000498, -0.00000021, 0.00000298, -0.00000606, -0.00000379, 0.00001128, 0.00000037, -0.00000473, 0.00000641, 0.00000201, -0.00000058, 0.00000013, -0.00000170, -0.00000495, 0.00000161, 0.00000716, 0.00000472, 0.00000084, -0.00000184, -0.00000188, -0.00000808, -0.00000220, 0.00001576, 0.00000618, -0.00001385, 0.00000042, 0.00000283, -0.00000983, 0.00000192, 0.00000874, 0.00000486, -0.00000680, -0.00000985, 0.00001087, 0.00000119, -0.00000695, 0.00000583, -0.00000434, -0.00000343, -0.00000221, 0.00000338, -0.00000127, 0.00000186, 0.00000326, -0.00001032, 0.00000769, -0.00000101, 0.00000603, -0.00000387, -0.00001687, 0.00000629, 0.00000878, -0.00000843, -0.00000067, 0.00002014, -0.00001653, 0.00000872, -0.00000621, -0.00000641, 0.00000797, -0.00001298, 0.00002204, -0.00001982, 0.00001700, -0.00000876, -0.00000925, 0.00001220, -0.00001810, 0.00002927, -0.00004084, 0.00001683, 0.00000627, -0.00000602, 0.00002191, -0.00004818, 0.00002184, -0.00000402, -0.00000199, 0.00001330, 0.00001336, -0.00000236, -0.00001056, 0.00003949, -0.00001411, 0.00001107, -0.00000429, -0.00002152, 0.00000547, -0.00001106, -0.00000317, -0.00000393, -0.00000471, -0.00000367, -0.00000405, -0.00001336, 0.00000505, -0.00003186, 0.00000139, -0.00006877, -0.00012059, 0.00011950, 0.00017696, 0.00016723, 0.00001841, -0.00012959, -0.00001016, -0.00001839, -0.00001066, 0.00000352, -0.00001226, 0.00000199, -0.00001591, 0.00001297, -0.00000436, -0.00000879, 0.00000647, -0.00001148, 0.00000541, -0.00000402, -0.00001037, 0.00000333, },
		{ -0.00000206, -0.00000240, 0.00000483, -0.00001181, -0.00001711, 0.00000784, 0.00000244, -0.00000000, 0.00000245, -0.00000760, 0.00000025, 0.00001088, -0.00000255, -0.00000237, 0.00000634, 0.00000111, -0.00000070, 0.00000018, -0.00000210, -0.00000494, 0.00000357, 0.00000675, 0.00000388, 0.00000083, -0.00000207, -0.00000264, -0.00000827, 0.00000088, 0.00001660, 0.00000124, -0.00001308, 0.00000338, -0.00000045, -0.00000886, 0.00000456, 0.00000827, 0.00000328, -0.00000922, -0.00000659, 0.00001161, -0.00000192, -0.00000427, 0.00000479, -0.00000518, -0.00000312, -0.00000189, 0.00000384, -0.00000218, 0.00000427, -0.00000055, -0.00000825, 0.00000860, -0.00000235, 0.00000730, -0.00000859, -0.00001370, 0.00000952, 0.00000639, -0.00001152, 0.00000572, 0.00001563, -0.00001686, 0.00001227, -0.00001249, 0.00000139, 0.00000099, -0.00000678, 0.00001913, -0.00002161, 0.00002344, -0.00001939, 0.00000215, 0.00000286, -0.00001091, 0.00002550, -0.00004475, 0.00002994, -0.00000858, 0.00000794, 0.00000959, -0.00004664, 0.00003287, -0.00001581, 0.00000673, 0.00001070, 0.00001252, -0.00000780, -0.00000344, 0.00003709, -0.00001956, 0.00001712, -0.00001159, -0.00001612, 0.00000464, -0.00001222, -0.00000125, -0.00000542, -0.00000378, -0.00000355, -0.00000618, -0.00001025, 0.00000155, -0.00003283, 0.00000415, -0.00009039, -0.00010308, 0.00014548, 0.00019259, 0.00016228, -0.00003327, -0.00012019, 0.00000476, -0.00002774, -0.00000161, -0.00000197, -0.00000946, 0.00000033, -0.00001557, 0.00001639, -0.00001002, -0.00000416, 0.00000457, -0.00001219, 0.00000792, -0.00000733, -0.00000822, 0.00000299, -0.00000494, },
		{ -0.00000164, 0.00000488, -0.00001640, -0.00001342, 0.00001042, -0.00000058, 0.00000119, 0.00000095, -0.00000802, 0.00000440, 0.00000891, -0.00000423, -0.00000004, 0.00000570, 0.00000049, -0.00000090, 0.00000043, -0.00000280, -0.00000441, 0.00000539, 0.00000578, 0.00000349, 0.00000071, -0.00000243, -0.00000333, -0.00000807, 0.00000425, 0.00001604, -0.00000344, -0.00001081, 0.00000493, -0.00000360, -0.00000678, 0.00000642, 0.00000762, 0.00000125, -0.00001080, -0.00000286, 0.00001093, -0.00000395, -0.00000176, 0.00000315, -0.00000544, -0.00000312, -0.00000109, 0.00000366, -0.00000222, 0.00000580, -0.00000441, -0.00000493, 0.00000804, -0.00000267, 0.00000748, -0.00001280, -0.00000926, 0.00001162, 0.00000291, -0.00001278, 0.00001177, 0.00000939, -0.00001429, 0.00001277, -0.00001623, 0.00000801, -0.00000651, 0.00000167, 0.00001233, -0.00001869, 0.00002546, -0.00002716, 0.00001316, -0.00000847, 0.00000017, 0.00001616, -0.00004200, 0.00003757, -0.00002125, 0.00002344, -0.00000842, -0.00003653, 0.00003723, -0.00002574, 0.00001711, 0.00000507, 0.00001316, -0.00001343, 0.00000525, 0.00003154, -0.00002217, 0.00002195, -0.00001903, -0.00000912, 0.00000157, -0.00001162, -0.00000029, -0.00000676, -0.00000230, -0.00000458, -0.00000711, -0.00000764, -0.00000308, -0.00003130, 0.00000291, -0.00011017, -0.00008148, 0.00017208, 0.00021037, 0.00014162, -0.00008089, -0.00009872, 0.00001085, -0.00003391, 0.00000695, -0.00000883, -0.00000465, -0.00000348, -0.00001289, 0.00001787, -0.00001505, 0.00000134, 0.00000091, -0.00001119, 0.00000941, -0.00001051, -0.00000555, 0.00000183, -0.00000506, -0.00000188, },
		{ 0.00000372, -0.00002054, -0.00000810, 0.00001108, -0.00000325, 0.00000293, -0.00000147, -0.00000691, 0.00000785, 0.00000589, -0.00000453, 0.00000181, 0.00000482, 0.00000003, -0.00000109, 0.00000078, -0.00000377, -0.00000323, 0.00000676, 0.00000452, 0.00000357, 0.00000030, -0.00000276, -0.00000399, -0.00000736, 0.00000763, 0.00001404, -0.00000723, -0.00000760, 0.00000495, -0.00000603, -0.00000408, 0.00000752, 0.00000683, -0.00000117, -0.00001125, 0.00000072, 0.00000921, -0.00000479, 0.00000012, 0.00000135, -0.00000535, -0.00000327, 0.00000002, 0.00000295, -0.00000137, 0.00000603, -0.00000757, -0.00000100, 0.00000623, -0.00000190, 0.00000631, -0.00001586, -0.00000403, 0.00001232, -0.00000113, -0.00001199, 0.00001646, 0.00000262, -0.00000962, 0.00001031, -0.00001677, 0.00001206, -0.00001273, 0.00001059, 0.00000298, -0.00001150, 0.00002236, -0.00003031, 0.00002136, -0.00001932, 0.00001309, 0.00000252, -0.00003253, 0.00003794, -0.00002855, 0.00003672, -0.00002900, -0.00001880, 0.00003315, -0.00003069, 0.00002648, -0.00000227, 0.00001511, -0.00001920, 0.00001523, 0.00002279, -0.00002121, 0.00002439, -0.00002541, -0.00000144, -0.00000344, -0.00000894, -0.00000099, -0.00000710, -0.00000105, -0.00000632, -0.00000666, -0.00000637, -0.00000764, -0.00002830, -0.00000216, -0.00012707, -0.00005687, 0.00020168, 0.00022593, 0.00010495, -0.00011786, -0.00007003, 0.00000817, -0.00003499, 0.00001293, -0.00001536, 0.00000110, -0.00000895, -0.00000778, 0.00001671, -0.00001841, 0.00000676, -0.00000398, -0.00000841, 0.00000943, -0.00001298, -0.00000283, 0.00000014, -0.00000456, -0.00000209, -0.00000030, },
		{ -0.00002339, -0.00000182, 0.00000962, -0.00000482, 0.00000455, -0.00000442, -0.00000421, 0.00000992, 0.00000256, -0.00000366, 0.00000290, 0.00000399, -0.00000044, -0.00000111, 0.00000106, -0.00000485, -0.00000141, 0.00000740, 0.00000330, 0.00000394, -0.00000042, -0.00000295, -0.00000468, -0.00000602, 0.00001061, 0.00001079, -0.00000963, -0.00000414, 0.00000366, -0.00000735, -0.00000125, 0.00000801, 0.00000583, -0.00000376, -0.00001048, 0.00000363, 0.00000696, -0.00000460, 0.00000117, -0.00000026, -0.00000516, -0.00000333, 0.00000117, 0.00000197, 0.00000017, 0.00000486, -0.00000941, 0.00000278, 0.00000368, -0.00000025, 0.00000373, -0.00001726, 0.00000136, 0.00001151, -0.00000513, -0.00000931, 0.00001912, -0.00000350, -0.00000399, 0.00000565, -0.00001416, 0.00001279, -0.00001620, 0.00001805, -0.00000698, -0.00000144, 0.00001453, -0.00002802, 0.00002487, -0.00002719, 0.00002524, -0.00001309, -0.00001774, 0.00003076, -0.00002829, 0.00004410, -0.00004797, 0.00000367, 0.00002075, -0.00002850, 0.00003203, -0.00000925, 0.00001743, -0.00002452, 0.00002578, 0.00001126, -0.00001633, 0.00002344, -0.00002940, 0.00000559, -0.00000949, -0.00000442, -0.00000371, -0.00000568, -0.00000099, -0.00000787, -0.00000530, -0.00000672, -0.00001114, -0.00002515, -0.00001015, -0.00014106, -0.00002920, 0.00023591, 0.00023355, 0.00005509, -0.00013934, -0.00003995, -0.00000129, -0.00003049, 0.00001483, -0.00001984, 0.00000634, -0.00001505, -0.00000075, 0.00001273, -0.00001928, 0.00001101, -0.00000925, -0.00000420, 0.00000780, -0.00001430, -0.00000056, -0.00000169, -0.00000364, -0.00000261, 0.00000158, 0.00000104, },
		{ 0.00000443, 0.00000633, -0.00000479, 0.00000532, -0.00000722, -0.00000022, 0.00001022, -0.00000033, -0.00000205, 0.00000323, 0.00000337, -0.00000102, -0.00000082, 0.00000105, -0.00000578, 0.00000088, 0.00000719, 0.00000244, 0.00000436, -0.00000138, -0.00000297, -0.00000538, -0.00000398, 0.00001277, 0.00000671, -0.00001042, -0.00000111, 0.00000151, -0.00000745, 0.00000130, 0.00000810, 0.00000450, -0.00000618, -0.00000864, 0.00000550, 0.00000476, -0.00000374, 0.00000140, -0.00000150, -0.00000508, -0.00000312, 0.00000210, 0.00000108, 0.00000199, 0.00000247, -0.00000960, 0.00000569, 0.00000101, 0.00000181, -0.00000000, -0.00001669, 0.00000622, 0.00000934, -0.00000848, -0.00000525, 0.00001948, -0.00000803, 0.00000135, -0.00000005, -0.00000915, 0.00001020, -0.00001610, 0.00002247, -0.00001547, 0.00000941, 0.00000342, -0.00002063, 0.00002277, -0.00003014, 0.00003399, -0.00002773, -0.00000028, 0.00001743, -0.00001994, 0.00004292, -0.00006101, 0.00002661, 0.00000213, -0.00001863, 0.00003160, -0.00001360, 0.00001864, -0.00002827, 0.00003564, -0.00000202, -0.00000783, 0.00001858, -0.00002984, 0.00001044, -0.00001519, 0.00000106, -0.00000817, -0.00000216, -0.00000298, -0.00000807, -0.00000406, -0.00000828, -0.00001319, -0.00002298, -0.00001985, -0.00015285, 0.00000298, 0.00027471, 0.00022741, -0.00000222, -0.00014324, -0.00001405, -0.00001414, -0.00002141, 0.00001219, -0.00002104, 0.00000962, -0.00002040, 0.00000714, 0.00000634, -0.00001729, 0.00001319, -0.00001389, 0.00000074, 0.00000473, -0.00001428, 0.00000086, -0.00000321, -0.00000266, -0.00000313, 0.00000369, -0.00000315, -0.00002420, },
		{ 0.00000195, -0.00000312, 0.00000471, -0.00000908, 0.00000438, 0.00000872, -0.00000217, -0.00000026, 0.00000299, 0.00000298, -0.00000175, -0.00000016, 0.00000057, -0.00000625, 0.00000331, 0.00000620, 0.00000216, 0.00000454, -0.00000241, -0.00000287, -0.00000601, -0.00000130, 0.00001372, 0.00000240, -0.00000969, 0.00000096, -0.00000089, -0.00000648, 0.00000333, 0.00000794, 0.00000272, -0.00000804, -0.00000612, 0.00000624, 0.00000306, -0.00000267, 0.00000100, -0.00000232, -0.00000515, -0.00000253, 0.00000262, 0.00000061, 0.00000356, -0.00000066, -0.00000815, 0.00000725, -0.00000114, 0.00000369, -0.00000436, -0.00001413, 0.00000991, 0.00000616, -0.00001073, -0.00000050, 0.00001774, -0.00001046, 0.00000538, -0.00000550, -0.00000295, 0.00000504, -0.00001239, 0.00002290, -0.00002072, 0.00001880, -0.00000881, -0.00000957, 0.00001537, -0.00002723, 0.00003732, -0.00003850, 0.00001650, 0.00000079, -0.00000484, 0.00003226, -0.00006466, 0.00004525, -0.00001890, -0.00000244, 0.00002436, -0.00001350, 0.00001714, -0.00002905, 0.00004317, -0.00001540, 0.00000323, 0.00001000, -0.00002608, 0.00001178, -0.00001893, 0.00000609, -0.00001343, 0.00000312, -0.00000737, -0.00000593, -0.00000430, -0.00000990, -0.00001422, -0.00002223, -0.00003029, -0.00016341, 0.00004239, 0.00031551, 0.00020302, -0.00005932, -0.00013070, 0.00000352, -0.00002652, -0.00000990, 0.00000571, -0.00001852, 0.00000990, -0.00002361, 0.00001447, -0.00000149, -0.00001266, 0.00001276, -0.00001697, 0.00000553, 0.00000073, -0.00001303, 0.00000126, -0.00000406, -0.00000200, -0.00000329, 0.00000551, -0.00000851, -0.00002251, 0.00000955, },
		{ -0.00000022, 0.00000253, -0.00000932, 0.00000873, 0.00000582, -0.00000264, 0.00000117, 0.00000248, 0.00000271, -0.00000252, 0.00000081, -0.00000043, -0.00000600, 0.00000550, 0.00000466, 0.00000253, 0.00000429, -0.00000327, -0.00000279, -0.00000639, 0.00000180, 0.00001328, -0.00000149, -0.00000781, 0.00000181, -0.00000298, -0.00000481, 0.00000478, 0.00000760, 0.00000050, -0.00000901, -0.00000341, 0.00000603, 0.00000210, -0.00000181, 0.00000028, -0.00000285, -0.00000531, -0.00000160, 0.00000266, 0.00000075, 0.00000439, -0.00000388, -0.00000542, 0.00000729, -0.00000230, 0.00000479, -0.00000862, -0.00000991, 0.00001197, 0.00000250, -0.00001164, 0.00000422, 0.00001445, -0.00001075, 0.00000744, -0.00000957, 0.00000307, -0.00000143, -0.00000586, 0.00001930, -0.00002168, 0.00002479, -0.00001973, 0.00000292, 0.00000411, -0.00001877, 0.00003432, -0.00004318, 0.00002929, -0.00001546, 0.00001398, 0.00001331, -0.00005725, 0.00005534, -0.00003768, 0.00001696, 0.00001117, -0.00000821, 0.00001177, -0.00002560, 0.00004663, -0.00002681, 0.00001508, -0.00000123, -0.00001831, 0.00000886, -0.00001928, 0.00000900, -0.00001793, 0.00000905, -0.00001375, -0.00000101, -0.00000724, -0.00001007, -0.00001551, -0.00002235, -0.00004139, -0.00017313, 0.00009235, 0.00035309, 0.00015863, -0.00010798, -0.00010580, 0.00001087, -0.00003509, 0.00000130, -0.00000295, -0.00001285, 0.00000683, -0.00002370, 0.00001984, -0.00000940, -0.00000617, 0.00000975, -0.00001785, 0.00000932, -0.00000349, -0.00001097, 0.00000075, -0.00000412, -0.00000190, -0.00000272, 0.00000640, -0.00001431, -0.00001832, 0.00001261, -0.00000251, },
		{ -0.00000097, -0.00000758, 0.00001193, 0.00000222, -0.00000179, 0.00000187, 0.00000206, 0.00000237, -0.00000314, 0.00000191, -0.00000189, -0.00000490, 0.00000707, 0.00000294, 0.00000341, 0.00000349, -0.00000380, -0.00000286, -0.00000630, 0.00000497, 0.00001144, -0.00000439, -0.00000537, 0.00000147, -0.00000431, -0.00000284, 0.00000576, 0.00000702, -0.00000201, -0.00000889, -0.00000103, 0.00000523, 0.00000187, -0.00000144, -0.00000045, -0.00000327, -0.00000542, -0.00000050, 0.00000234, 0.00000153, 0.00000414, -0.00000647, -0.00000204, 0.00000598, -0.00000227, 0.00000463, -0.00001201, -0.00000465, 0.00001221, -0.00000109, -0.00001125, 0.00000827, 0.00001032, -0.00000927, 0.00000741, -0.00001152, 0.00000763, -0.00000770, 0.00000214, 0.00001245, -0.00001821, 0.00002614, -0.00002721, 0.00001432, -0.00000873, -0.00000627, 0.00002541, -0.00004084, 0.00003557, -0.00002759, 0.00003239, -0.00001072, -0.00003940, 0.00005426, -0.00004971, 0.00003539, -0.00000553, 0.00000162, 0.00000234, -0.00001727, 0.00004460, -0.00003413, 0.00002546, -0.00001322, -0.00000769, 0.00000182, -0.00001540, 0.00000828, -0.00001989, 0.00001384, -0.00002077, 0.00000620, -0.00001342, -0.00000739, -0.00001876, -0.00002190, -0.00005433, -0.00018093, 0.00015557, 0.00038006, 0.00009613, -0.00014118, -0.00007468, 0.00000859, -0.00003781, 0.00000964, -0.00001159, -0.00000539, 0.00000095, -0.00002034, 0.00002215, -0.00001597, 0.00000099, 0.00000471, -0.00001640, 0.00001146, -0.00000719, -0.00000866, -0.00000031, -0.00000352, -0.00000240, -0.00000125, 0.00000575, -0.00001956, -0.00001217, 0.00001308, -0.00000597, 0.00000313, },
		{ -0.00000392, 0.00001331, -0.00000126, 0.00000002, 0.00000172, 0.00000195, 0.00000176, -0.00000341, 0.00000287, -0.00000356, -0.00000300, 0.00000772, 0.00000144, 0.00000454, 0.00000221, -0.00000392, -0.00000319, -0.00000554, 0.00000776, 0.00000851, -0.00000594, -0.00000301, 0.00000028, -0.00000470, -0.00000096, 0.00000643, 0.00000604, -0.00000450, -0.00000774, 0.00000061, 0.00000434, 0.00000213, -0.00000164, -0.00000097, -0.00000374, -0.00000528, 0.00000056, 0.00000189, 0.00000273, 0.00000274, -0.00000789, 0.00000126, 0.00000381, -0.00000119, 0.00000303, -0.00001383, 0.00000083, 0.00001075, -0.00000411, -0.00000979, 0.00001126, 0.00000609, -0.00000668, 0.00000560, -0.00001116, 0.00000992, -0.00001241, 0.00001000, 0.00000381, -0.00001112, 0.00002262, -0.00002987, 0.00002235, -0.00002054, 0.00000791, 0.00001225, -0.00003197, 0.00003420, -0.00003267, 0.00004610, -0.00003524, -0.00001405, 0.00004170, -0.00005174, 0.00004853, -0.00002218, 0.00001400, -0.00001021, -0.00000438, 0.00003648, -0.00003570, 0.00003210, -0.00002362, 0.00000374, -0.00000813, -0.00000747, 0.00000315, -0.00001773, 0.00001548, -0.00002636, 0.00001413, -0.00002232, -0.00000126, -0.00002550, -0.00001888, -0.00007148, -0.00018366, 0.00023282, 0.00038809, 0.00002128, -0.00015455, -0.00004414, -0.00000052, -0.00003440, 0.00001345, -0.00001812, 0.00000203, -0.00000650, -0.00001400, 0.00002096, -0.00002002, 0.00000746, -0.00000136, -0.00001297, 0.00001173, -0.00000986, -0.00000667, -0.00000140, -0.00000263, -0.00000325, 0.00000099, 0.00000321, -0.00002321, -0.00000505, 0.00001097, -0.00000761, 0.00000596, -0.00000510, },
		{ 0.00001259, -0.00000381, 0.00000220, 0.00000086, 0.00000224, 0.00000079, -0.00000314, 0.00000342, -0.00000511, -0.00000051, 0.00000739, 0.00000052, 0.00000558, 0.00000066, -0.00000366, -0.00000376, -0.00000402, 0.00000970, 0.00000497, -0.00000608, -0.00000128, -0.00000128, -0.00000421, 0.00000058, 0.00000693, 0.00000454, -0.00000660, -0.00000582, 0.00000133, 0.00000377, 0.00000253, -0.00000228, -0.00000121, -0.00000437, -0.00000476, 0.00000135, 0.00000160, 0.00000398, 0.00000039, -0.00000784, 0.00000378, 0.00000142, 0.00000052, 0.00000015, -0.00001365, 0.00000565, 0.00000801, -0.00000623, -0.00000764, 0.00001305, 0.00000237, -0.00000374, 0.00000266, -0.00000880, 0.00000965, -0.00001461, 0.00001619, -0.00000486, -0.00000193, 0.00001500, -0.00002733, 0.00002548, -0.00002895, 0.00002107, -0.00000261, -0.00001850, 0.00002574, -0.00002934, 0.00005164, -0.00005522, 0.00001403, 0.00001982, -0.00004260, 0.00005295, -0.00003492, 0.00002594, -0.00002376, 0.00001156, 0.00002281, -0.00003080, 0.00003323, -0.00003005, 0.00001345, -0.00001883, 0.00000319, -0.00000613, -0.00001067, 0.00001223, -0.00002816, 0.00002034, -0.00003214, 0.00000770, -0.00003638, -0.00001163, -0.00009587, -0.00017595, 0.00032166, 0.00036969, -0.00005692, -0.00014744, -0.00002015, -0.00001236, -0.00002629, 0.00001235, -0.00002109, 0.00000760, -0.00001385, -0.00000582, 0.00001654, -0.00002096, 0.00001211, -0.00000729, -0.00000831, 0.00001032, -0.00001130, -0.00000540, -0.00000206, -0.00000192, -0.00000399, 0.00000355, -0.00000115, -0.00002442, 0.00000178, 0.00000686, -0.00000705, 0.00000739, -0.00000893, 0.00000116, },
		{ -0.00000491, 0.00000413, -0.00000040, 0.00000285, -0.00000050, -0.00000229, 0.00000334, -0.00000618, 0.00000217, 0.00000616, 0.00000036, 0.00000620, -0.00000090, -0.00000318, -0.00000444, -0.00000179, 0.00001043, 0.00000144, -0.00000502, -0.00000053, -0.00000268, -0.00000313, 0.00000172, 0.00000731, 0.00000247, -0.00000791, -0.00000358, 0.00000120, 0.00000383, 0.00000268, -0.00000312, -0.00000125, -0.00000512, -0.00000383, 0.00000173, 0.00000172, 0.00000483, -0.00000244, -0.00000635, 0.00000505, -0.00000053, 0.00000223, -0.00000352, -0.00001142, 0.00000908, 0.00000460, -0.00000734, -0.00000519, 0.00001372, -0.00000049, -0.00000111, -0.00000059, -0.00000512, 0.00000710, -0.00001394, 0.00001963, -0.00001191, 0.00000747, 0.00000485, -0.00002028, 0.00002323, -0.00003234, 0.00003079, -0.00001631, -0.00000326, 0.00001231, -0.00001814, 0.00004721, -0.00006634, 0.00003927, -0.00000704, -0.00002363, 0.00004701, -0.00004054, 0.00003421, -0.00003550, 0.00002805, 0.00000538, -0.00001995, 0.00002810, -0.00003074, 0.00001890, -0.00002753, 0.00001425, -0.00001803, 0.00000086, 0.00000328, -0.00002417, 0.00002202, -0.00004005, 0.00001734, -0.00005047, 0.00000033, -0.00013013, -0.00015075, 0.00041566, 0.00032012, -0.00012772, -0.00012313, -0.00000650, -0.00002268, -0.00001611, 0.00000732, -0.00002011, 0.00001009, -0.00001946, 0.00000260, 0.00000989, -0.00001885, 0.00001425, -0.00001203, -0.00000337, 0.00000779, -0.00001166, -0.00000492, -0.00000199, -0.00000181, -0.00000411, 0.00000570, -0.00000679, -0.00002278, 0.00000706, 0.00000182, -0.00000446, 0.00000684, -0.00001150, 0.00000677, 0.00000990, },
		{ 0.00000526, -0.00000162, 0.00000355, -0.00000196, -0.00000093, 0.00000255, -0.00000647, 0.00000462, 0.00000436, 0.00000098, 0.00000618, -0.00000219, -0.00000269, -0.00000500, 0.00000092, 0.00000979, -0.00000145, -0.00000325, -0.00000081, -0.00000349, -0.00000181, 0.00000255, 0.00000751, -0.00000010, -0.00000816, -0.00000153, 0.00000053, 0.00000454, 0.00000227, -0.00000385, -0.00000125, -0.00000585, -0.00000260, 0.00000170, 0.00000239, 0.00000486, -0.00000514, -0.00000381, 0.00000491, -0.00000152, 0.00000329, -0.00000724, -0.00000747, 0.00001062, 0.00000121, -0.00000757, -0.00000277, 0.00001351, -0.00000239, 0.00000077, -0.00000342, -0.00000104, 0.00000301, -0.00001065, 0.00001988, -0.00001610, 0.00001524, -0.00000581, -0.00001031, 0.00001624, -0.00003016, 0.00003544, -0.00002635, 0.00001058, -0.00000288, -0.00000144, 0.00003318, -0.00006598, 0.00005648, -0.00003328, 0.00000158, 0.00003134, -0.00003732, 0.00003616, -0.00004252, 0.00004200, -0.00001300, -0.00000501, 0.00001732, -0.00002502, 0.00001825, -0.00003155, 0.00002277, -0.00002990, 0.00001502, -0.00001078, -0.00001345, 0.00001679, -0.00004274, 0.00002427, -0.00006500, 0.00001535, -0.00017514, -0.00010067, 0.00050453, 0.00023910, -0.00018066, -0.00008821, -0.00000406, -0.00002818, -0.00000688, 0.00000034, -0.00001588, 0.00000912, -0.00002217, 0.00000970, 0.00000239, -0.00001442, 0.00001379, -0.00001490, 0.00000096, 0.00000483, -0.00001139, -0.00000502, -0.00000121, -0.00000247, -0.00000319, 0.00000665, -0.00001277, -0.00001850, 0.00000987, -0.00000291, -0.00000054, 0.00000426, -0.00001205, 0.00001190, 0.00000581, -0.00000435, },
		{ -0.00000242, 0.00000406, -0.00000333, 0.00000071, 0.00000113, -0.00000583, 0.00000640, 0.00000239, 0.00000220, 0.00000546, -0.00000298, -0.00000242, -0.00000517, 0.00000371, 0.00000789, -0.00000322, -0.00000137, -0.00000189, -0.00000348, -0.00000063, 0.00000326, 0.00000733, -0.00000289, -0.00000729, -0.00000015, -0.00000023, 0.00000573, 0.00000119, -0.00000424, -0.00000142, -0.00000636, -0.00000125, 0.00000141, 0.00000354, 0.00000384, -0.00000706, -0.00000083, 0.00000357, -0.00000137, 0.00000319, -0.00001018, -0.00000249, 0.00001018, -0.00000155, -0.00000722, -0.00000054, 0.00001272, -0.00000344, 0.00000172, -0.00000529, 0.00000260, -0.00000164, -0.00000552, 0.00001716, -0.00001687, 0.00001999, -0.00001501, 0.00000052, 0.00000604, -0.00002300, 0.00003450, -0.00003112, 0.00002032, -0.00001626, 0.00001710, 0.00001207, -0.00005392, 0.00006202, -0.00005330, 0.00002788, 0.00000880, -0.00002552, 0.00003046, -0.00004261, 0.00005044, -0.00002902, 0.00001110, 0.00000288, -0.00001371, 0.00001088, -0.00002898, 0.00002589, -0.00003851, 0.00002877, -0.00002776, 0.00000327, 0.00000350, -0.00003737, 0.00002451, -0.00007563, 0.00002926, -0.00022871, -0.00001987, 0.00057507, 0.00013179, -0.00020773, -0.00005116, -0.00001073, -0.00002742, -0.00000109, -0.00000619, -0.00001000, 0.00000528, -0.00002157, 0.00001428, -0.00000449, -0.00000884, 0.00001124, -0.00001572, 0.00000408, 0.00000213, -0.00001105, -0.00000520, -0.00000004, -0.00000379, -0.00000112, 0.00000573, -0.00001788, -0.00001236, 0.00000981, -0.00000614, 0.00000365, 0.00000009, -0.00001024, 0.00001557, 0.00000118, -0.00000234, 0.00000528, },
		{ 0.00000413, -0.00000433, 0.00000232, -0.00000067, -0.00000429, 0.00000725, 0.00000068, 0.00000369, 0.00000415, -0.00000322, -0.00000249, -0.00000471, 0.00000609, 0.00000508, -0.00000367, 0.00000003, -0.00000335, -0.00000268, 0.00000018, 0.00000402, 0.00000658, -0.00000551, -0.00000548, 0.00000029, -0.00000059, 0.00000704, -0.00000047, -0.00000415, -0.00000190, -0.00000645, -0.00000005, 0.00000111, 0.00000492, 0.00000179, -0.00000776, 0.00000187, 0.00000153, -0.00000020, 0.00000171, -0.00001162, 0.00000260, 0.00000802, -0.00000327, -0.00000667, 0.00000145, 0.00001157, -0.00000387, 0.00000184, -0.00000597, 0.00000513, -0.00000587, 0.00000040, 0.00001224, -0.00001441, 0.00002105, -0.00002117, 0.00001011, -0.00000526, -0.00001237, 0.00002861, -0.00003022, 0.00002427, -0.00002475, 0.00003331, -0.00001197, -0.00003240, 0.00005466, -0.00006267, 0.00004968, -0.00001613, -0.00000742, 0.00001751, -0.00003488, 0.00005122, -0.00003953, 0.00002499, -0.00001217, 0.00000094, -0.00000220, -0.00001928, 0.00002164, -0.00004075, 0.00003839, -0.00004406, 0.00002340, -0.00001707, -0.00002252, 0.00001460, -0.00007726, 0.00003559, -0.00028464, 0.00009391, 0.00061327, 0.00000872, -0.00020522, -0.00002042, -0.00002209, -0.00002110, 0.00000002, -0.00001025, -0.00000442, -0.00000006, -0.00001817, 0.00001587, -0.00000959, -0.00000335, 0.00000753, -0.00001484, 0.00000584, 0.00000007, -0.00001112, -0.00000493, 0.00000097, -0.00000530, 0.00000182, 0.00000264, -0.00002098, -0.00000562, 0.00000717, -0.00000710, 0.00000695, -0.00000478, -0.00000625, 0.00001712, -0.00000306, 0.00000061, 0.00000418, -0.00000256, },
		{ -0.00000475, 0.00000359, -0.00000248, -0.00000210, 0.00000708, -0.00000039, 0.00000506, 0.00000250, -0.00000298, -0.00000290, -0.00000350, 0.00000761, 0.00000192, -0.00000287, 0.00000054, -0.00000465, -0.00000135, 0.00000056, 0.00000492, 0.00000508, -0.00000751, -0.00000311, -0.00000024, -0.00000021, 0.00000802, -0.00000244, -0.00000365, -0.00000273, -0.00000602, 0.00000081, 0.00000109, 0.00000612, -0.00000101, -0.00000703, 0.00000367, -0.00000056, 0.00000154, -0.00000101, -0.00001112, 0.00000688, 0.00000476, -0.00000384, -0.00000622, 0.00000330, 0.00001022, -0.00000394, 0.00000136, -0.00000554, 0.00000623, -0.00000891, 0.00000602, 0.00000625, -0.00000951, 0.00001851, -0.00002347, 0.00001677, -0.00001552, -0.00000035, 0.00001935, -0.00002450, 0.00002215, -0.00002647, 0.00004353, -0.00003406, -0.00000577, 0.00003594, -0.00005922, 0.00006225, -0.00003827, 0.00001313, -0.00000046, -0.00002015, 0.00004364, -0.00004234, 0.00003352, -0.00002434, 0.00001557, -0.00001841, -0.00000365, 0.00000957, -0.00003453, 0.00004042, -0.00005539, 0.00004270, -0.00004190, 0.00000089, -0.00000717, -0.00006538, 0.00002655, -0.00033266, 0.00023745, 0.00060687, -0.00011543, -0.00017476, -0.00000255, -0.00003271, -0.00001180, -0.00000317, -0.00001074, -0.00000085, -0.00000520, -0.00001313, 0.00001478, -0.00001239, 0.00000102, 0.00000368, -0.00001298, 0.00000651, -0.00000133, -0.00001180, -0.00000379, 0.00000124, -0.00000635, 0.00000493, -0.00000235, -0.00002129, 0.00000029, 0.00000284, -0.00000561, 0.00000844, -0.00000925, -0.00000074, 0.00001633, -0.00000609, 0.00000381, 0.00000219, -0.00000199, 0.00000364, },
		{ 0.00000427, -0.00000395, 0.00000037, 0.00000601, -0.00000065, 0.00000598, 0.00000084, -0.00000245, -0.00000352, -0.00000161, 0.00000795, -0.00000098, -0.00000118, 0.00000001, -0.00000534, 0.00000013, 0.00000062, 0.00000588, 0.00000281, -0.00000849, -0.00000074, -0.00000150, 0.00000108, 0.00000827, -0.00000437, -0.00000291, -0.00000381, -0.00000505, 0.00000120, 0.00000156, 0.00000671, -0.00000403, -0.00000501, 0.00000416, -0.00000204, 0.00000321, -0.00000446, -0.00000862, 0.00000956, 0.00000120, -0.00000345, -0.00000608, 0.00000515, 0.00000871, -0.00000383, 0.00000061, -0.00000427, 0.00000593, -0.00001032, 0.00001047, 0.00000037, -0.00000337, 0.00001316, -0.00002191, 0.00001954, -0.00002296, 0.00001093, 0.00000880, -0.00001578, 0.00001510, -0.00002119, 0.00004543, -0.00004967, 0.00002051, 0.00000980, -0.00004352, 0.00006272, -0.00005297, 0.00003162, -0.00001992, -0.00000089, 0.00002874, -0.00003682, 0.00003461, -0.00003054, 0.00002645, -0.00003404, 0.00001498, -0.00000883, -0.00001949, 0.00003252, -0.00005768, 0.00005608, -0.00006601, 0.00002933, -0.00003973, -0.00003769, -0.00000519, -0.00035936, 0.00040103, 0.00054821, -0.00022414, -0.00012328, -0.00000065, -0.00003768, -0.00000312, -0.00000880, -0.00000788, -0.00000024, -0.00000864, -0.00000797, 0.00001196, -0.00001303, 0.00000378, 0.00000052, -0.00001093, 0.00000665, -0.00000240, -0.00001291, -0.00000168, 0.00000038, -0.00000626, 0.00000727, -0.00000844, -0.00001860, 0.00000410, -0.00000185, -0.00000212, 0.00000764, -0.00001226, 0.00000531, 0.00001345, -0.00000742, 0.00000655, -0.00000025, -0.00000083, 0.00000259, -0.00000448, },
		{ -0.00000476, 0.00000269, 0.00000435, -0.00000012, 0.00000622, -0.00000053, -0.00000189, -0.00000409, 0.00000073, 0.00000700, -0.00000303, 0.00000083, -0.00000144, -0.00000512, 0.00000134, 0.00000063, 0.00000671, -0.00000004, -0.00000824, 0.00000110, -0.00000305, 0.00000317, 0.00000757, -0.00000591, -0.00000219, -0.00000492, -0.00000371, 0.00000116, 0.00000259, 0.00000634, -0.00000666, -0.00000216, 0.00000331, -0.00000246, 0.00000415, -0.00000791, -0.00000451, 0.00001020, -0.00000186, -0.00000254, -0.00000621, 0.00000708, 0.00000699, -0.00000364, -0.00000013, -0.00000256, 0.00000450, -0.00001006, 0.00001321, -0.00000443, 0.00000274, 0.00000626, -0.00001722, 0.00001830, -0.00002653, 0.00001975, -0.00000096, -0.00000636, 0.00000534, -0.00001031, 0.00003860, -0.00005568, 0.00004106, -0.00001819, -0.00001880, 0.00005084, -0.00005714, 0.00004386, -0.00003673, 0.00001924, 0.00000923, -0.00002411, 0.00002779, -0.00002882, 0.00003033, -0.00004500, 0.00003248, -0.00003021, 0.00000254, 0.00001441, -0.00004827, 0.00005871, -0.00008335, 0.00005681, -0.00007865, 0.00000447, -0.00006435, -0.00035021, 0.00056858, 0.00043656, -0.00030195, -0.00006178, -0.00001366, -0.00003395, 0.00000139, -0.00001407, -0.00000312, -0.00000241, -0.00000953, -0.00000407, 0.00000871, -0.00001228, 0.00000504, -0.00000154, -0.00000935, 0.00000688, -0.00000370, -0.00001394, 0.00000109, -0.00000166, -0.00000463, 0.00000788, -0.00001439, -0.00001344, 0.00000504, -0.00000552, 0.00000237, 0.00000464, -0.00001309, 0.00001081, 0.00000914, -0.00000693, 0.00000827, -0.00000265, 0.00000065, 0.00000114, -0.00000358, 0.00000425, },
		{ 0.00000447, 0.00000251, 0.00000102, 0.00000575, -0.00000141, -0.00000153, -0.00000433, 0.00000310, 0.00000494, -0.00000386, 0.00000254, -0.00000341, -0.00000398, 0.00000200, 0.00000084, 0.00000710, -0.00000312, -0.00000678, 0.00000196, -0.00000434, 0.00000572, 0.00000589, -0.00000679, -0.00000175, -0.00000582, -0.00000222, 0.00000087, 0.00000407, 0.00000483, -0.00000831, 0.00000086, 0.00000140, -0.00000168, 0.00000388, -0.00001051, 0.00000047, 0.00000878, -0.00000379, -0.00000164, -0.00000642, 0.00000906, 0.00000502, -0.00000333, -0.00000069, -0.00000082, 0.00000238, -0.00000839, 0.00001413, -0.00000753, 0.00000774, -0.00000082, -0.00001064, 0.00001372, -0.00002602, 0.00002509, -0.00000834, 0.00000160, -0.00000443, 0.00000346, 0.00002457, -0.00005108, 0.00005173, -0.00004210, 0.00000975, 0.00002900, -0.00004999, 0.00004704, -0.00004719, 0.00003614, -0.00001109, -0.00000704, 0.00001448, -0.00001906, 0.00002534, -0.00004783, 0.00004436, -0.00004992, 0.00002767, -0.00001173, -0.00002682, 0.00004734, -0.00008809, 0.00007599, -0.00011644, 0.00005527, -0.00015083, -0.00029238, 0.00071921, 0.00027939, -0.00033750, -0.00000324, -0.00003661, -0.00002129, -0.00000060, -0.00001619, 0.00000129, -0.00000616, -0.00000800, -0.00000219, 0.00000620, -0.00001118, 0.00000544, -0.00000267, -0.00000845, 0.00000762, -0.00000580, -0.00001418, 0.00000388, -0.00000448, -0.00000146, 0.00000608, -0.00001881, -0.00000694, 0.00000305, -0.00000705, 0.00000660, 0.00000013, -0.00001151, 0.00001480, 0.00000432, -0.00000488, 0.00000869, -0.00000457, 0.00000212, -0.00000046, -0.00000223, 0.00000359, -0.00000477, },
		{ 0.00000090, 0.00000244, 0.00000468, -0.00000172, -0.00000150, -0.00000400, 0.00000504, 0.00000218, -0.00000338, 0.00000344, -0.00000535, -0.00000214, 0.00000199, 0.00000145, 0.00000676, -0.00000594, -0.00000439, 0.00000167, -0.00000488, 0.00000823, 0.00000343, -0.00000695, -0.00000177, -0.00000626, -0.00000085, 0.00000061, 0.00000568, 0.00000230, -0.00000857, 0.00000334, -0.00000101, 0.00000011, 0.00000218, -0.00001158, 0.00000538, 0.00000573, -0.00000428, -0.00000127, -0.00000636, 0.00001095, 0.00000281, -0.00000286, -0.00000103, 0.00000068, 0.00000006, -0.00000576, 0.00001345, -0.00000879, 0.00001091, -0.00000687, -0.00000362, 0.00000699, -0.00002198, 0.00002676, -0.00001254, 0.00000656, -0.00001170, 0.00001681, 0.00000650, -0.00003721, 0.00005058, -0.00005693, 0.00003615, 0.00000168, -0.00003319, 0.00004036, -0.00004891, 0.00004630, -0.00002807, 0.00001050, -0.00000225, -0.00000308, 0.00001158, -0.00004059, 0.00004685, -0.00006293, 0.00005056, -0.00004136, 0.00000410, 0.00002148, -0.00007624, 0.00007968, -0.00014369, 0.00010450, -0.00025816, -0.00017795, 0.00083014, 0.00009226, -0.00032601, 0.00004015, -0.00006188, -0.00000240, -0.00000939, -0.00001323, 0.00000305, -0.00000957, -0.00000510, -0.00000227, 0.00000510, -0.00001069, 0.00000587, -0.00000346, -0.00000799, 0.00000888, -0.00000897, -0.00001301, 0.00000586, -0.00000731, 0.00000270, 0.00000173, -0.00002051, -0.00000067, -0.00000119, -0.00000589, 0.00000937, -0.00000483, -0.00000782, 0.00001672, -0.00000011, -0.00000183, 0.00000781, -0.00000572, 0.00000329, -0.00000194, -0.00000070, 0.00000248, -0.00000400, 0.00000545, },
		{ 0.00000376, 0.00000328, -0.00000156, -0.00000181, -0.00000301, 0.00000612, -0.00000072, -0.00000179, 0.00000323, -0.00000672, -0.00000001, 0.00000142, 0.00000248, 0.00000550, -0.00000799, -0.00000159, 0.00000029, -0.00000432, 0.00001019, 0.00000056, -0.00000646, -0.00000227, -0.00000612, 0.00000016, 0.00000065, 0.00000702, -0.00000089, -0.00000736, 0.00000471, -0.00000323, 0.00000240, -0.00000076, -0.00001068, 0.00000925, 0.00000179, -0.00000337, -0.00000177, -0.00000568, 0.00001246, 0.00000046, -0.00000215, -0.00000121, 0.00000179, -0.00000209, -0.00000267, 0.00001162, -0.00000842, 0.00001206, -0.00001108, 0.00000257, -0.00000045, -0.00001552, 0.00002533, -0.00001363, 0.00000793, -0.00001478, 0.00002667, -0.00001160, -0.00001740, 0.00003825, -0.00005973, 0.00005492, -0.00002550, -0.00001048, 0.00002527, -0.00004141, 0.00004754, -0.00003824, 0.00002440, -0.00001830, 0.00001553, -0.00000865, -0.00002357, 0.00003791, -0.00006510, 0.00006555, -0.00006829, 0.00003920, -0.00001600, -0.00004700, 0.00006281, -0.00015097, 0.00013885, -0.00037297, -0.00000679, 0.00088044, -0.00010300, -0.00027053, 0.00005951, -0.00008104, 0.00001774, -0.00002304, -0.00000489, 0.00000054, -0.00001067, -0.00000248, -0.00000340, 0.00000533, -0.00001130, 0.00000713, -0.00000471, -0.00000732, 0.00001024, -0.00001307, -0.00001010, 0.00000629, -0.00000921, 0.00000685, -0.00000464, -0.00001884, 0.00000378, -0.00000637, -0.00000223, 0.00000984, -0.00000907, -0.00000280, 0.00001646, -0.00000340, 0.00000151, 0.00000587, -0.00000603, 0.00000398, -0.00000308, 0.00000073, 0.00000122, -0.00000265, 0.00000556, -0.00000017, },
		{ 0.00000184, -0.00000111, -0.00000232, -0.00000143, 0.00000610, -0.00000314, 0.00000042, 0.00000192, -0.00000714, 0.00000191, 0.00000057, 0.00000380, 0.00000332, -0.00000888, 0.00000103, -0.00000181, -0.00000259, 0.00001117, -0.00000227, -0.00000558, -0.00000314, -0.00000540, 0.00000072, 0.00000119, 0.00000763, -0.00000419, -0.00000491, 0.00000464, -0.00000461, 0.00000452, -0.00000440, -0.00000784, 0.00001134, -0.00000214, -0.00000147, -0.00000318, -0.00000410, 0.00001328, -0.00000182, -0.00000122, -0.00000130, 0.00000250, -0.00000383, 0.00000042, 0.00000917, -0.00000689, 0.00001139, -0.00001315, 0.00000701, -0.00000729, -0.00000794, 0.00002179, -0.00001241, 0.00000616, -0.00001317, 0.00003089, -0.00002585, 0.00000378, 0.00001781, -0.00005019, 0.00006231, -0.00004706, 0.00001321, 0.00000507, -0.00002630, 0.00003968, -0.00003960, 0.00003132, -0.00002950, 0.00003225, -0.00003126, 0.00000057, 0.00001797, -0.00005424, 0.00006792, -0.00008595, 0.00007121, -0.00005892, -0.00000359, 0.00002411, -0.00013130, 0.00014421, -0.00047602, 0.00021163, 0.00085510, -0.00028150, -0.00018183, 0.00005132, -0.00008704, 0.00003316, -0.00003769, 0.00000730, -0.00000648, -0.00000810, -0.00000188, -0.00000415, 0.00000604, -0.00001283, 0.00000954, -0.00000711, -0.00000566, 0.00001090, -0.00001748, -0.00000559, 0.00000477, -0.00000931, 0.00000974, -0.00001185, -0.00001390, 0.00000520, -0.00001087, 0.00000308, 0.00000777, -0.00001158, 0.00000251, 0.00001437, -0.00000516, 0.00000450, 0.00000330, -0.00000556, 0.00000415, -0.00000377, 0.00000187, 0.00000009, -0.00000103, 0.00000494, -0.00000054, 0.00000469, },
		{ -0.00000063, -0.00000278, 0.00000047, 0.00000494, -0.00000460, 0.00000267, -0.00000021, -0.00000644, 0.00000321, -0.00000019, 0.00000506, 0.00000047, -0.00000843, 0.00000290, -0.00000409, 0.00000011, 0.00001093, -0.00000466, -0.00000462, -0.00000416, -0.00000423, 0.00000086, 0.00000228, 0.00000721, -0.00000694, -0.00000175, 0.00000318, -0.00000473, 0.00000579, -0.00000799, -0.00000347, 0.00001128, -0.00000518, 0.00000077, -0.00000529, -0.00000154, 0.00001312, -0.00000373, -0.00000016, -0.00000139, 0.00000286, -0.00000507, 0.00000325, 0.00000655, -0.00000478, 0.00000941, -0.00001323, 0.00000931, -0.00001257, -0.00000043, 0.00001731, -0.00001007, 0.00000250, -0.00000763, 0.00002880, -0.00003340, 0.00002160, -0.00000598, -0.00003067, 0.00005706, -0.00005889, 0.00003286, -0.00001581, -0.00000690, 0.00002454, -0.00003216, 0.00002955, -0.00003262, 0.00004257, -0.00005110, 0.00002715, -0.00000988, -0.00003092, 0.00005525, -0.00008889, 0.00009236, -0.00009873, 0.00004672, -0.00003272, -0.00008241, 0.00010888, -0.00054463, 0.00045797, 0.00074852, -0.00041934, -0.00007659, 0.00001820, -0.00007600, 0.00003855, -0.00004862, 0.00002021, -0.00001671, -0.00000164, -0.00000448, -0.00000299, 0.00000592, -0.00001439, 0.00001280, -0.00001089, -0.00000239, 0.00000996, -0.00002122, -0.00000019, 0.00000144, -0.00000719, 0.00001022, -0.00001830, -0.00000666, 0.00000308, -0.00001320, 0.00000868, 0.00000360, -0.00001180, 0.00000708, 0.00001115, -0.00000534, 0.00000665, 0.00000055, -0.00000453, 0.00000386, -0.00000401, 0.00000262, -0.00000067, 0.00000051, 0.00000383, -0.00000025, 0.00000504, 0.00000065, },
		{ -0.00000295, 0.00000232, 0.00000287, -0.00000483, 0.00000437, -0.00000265, -0.00000476, 0.00000365, -0.00000051, 0.00000590, -0.00000264, -0.00000672, 0.00000361, -0.00000594, 0.00000335, 0.00000949, -0.00000631, -0.00000386, -0.00000505, -0.00000286, 0.00000076, 0.00000377, 0.00000565, -0.00000859, 0.00000143, 0.00000071, -0.00000351, 0.00000573, -0.00001070, 0.00000164, 0.00000914, -0.00000673, 0.00000261, -0.00000760, 0.00000181, 0.00001185, -0.00000500, 0.00000085, -0.00000146, 0.00000295, -0.00000583, 0.00000566, 0.00000406, -0.00000258, 0.00000675, -0.00001182, 0.00000954, -0.00001580, 0.00000613, 0.00001284, -0.00000780, -0.00000142, 0.00000010, 0.00002130, -0.00003305, 0.00003225, -0.00002779, -0.00000564, 0.00004063, -0.00005907, 0.00004453, -0.00003287, 0.00001250, 0.00000561, -0.00001798, 0.00001944, -0.00002625, 0.00004314, -0.00006317, 0.00005040, -0.00004037, 0.00000132, 0.00002823, -0.00007428, 0.00009618, -0.00012617, 0.00009372, -0.00009891, -0.00000838, 0.00002672, -0.00055649, 0.00070473, 0.00056662, -0.00049791, 0.00002556, -0.00003156, -0.00004842, 0.00003076, -0.00005148, 0.00002990, -0.00002747, 0.00000751, -0.00001044, 0.00000109, 0.00000360, -0.00001467, 0.00001598, -0.00001565, 0.00000264, 0.00000675, -0.00002318, 0.00000502, -0.00000301, -0.00000299, 0.00000760, -0.00002235, 0.00000126, -0.00000220, -0.00001237, 0.00001308, -0.00000170, -0.00000975, 0.00001012, 0.00000766, -0.00000428, 0.00000772, -0.00000200, -0.00000318, 0.00000325, -0.00000389, 0.00000302, -0.00000098, 0.00000170, 0.00000259, 0.00000053, 0.00000479, -0.00000012, -0.00000029, },
		{ 0.00000368, 0.00000034, -0.00000385, 0.00000507, -0.00000480, -0.00000247, 0.00000324, -0.00000021, 0.00000595, -0.00000546, -0.00000413, 0.00000303, -0.00000684, 0.00000657, 0.00000710, -0.00000714, -0.00000350, -0.00000558, -0.00000153, 0.00000067, 0.00000534, 0.00000308, -0.00000881, 0.00000392, -0.00000215, -0.00000120, 0.00000416, -0.00001184, 0.00000654, 0.00000544, -0.00000653, 0.00000345, -0.00000953, 0.00000557, 0.00000952, -0.00000542, 0.00000159, -0.00000147, 0.00000281, -0.00000613, 0.00000765, 0.00000180, -0.00000060, 0.00000396, -0.00000956, 0.00000813, -0.00001696, 0.00001133, 0.00000892, -0.00000642, -0.00000408, 0.00000777, 0.00001056, -0.00002555, 0.00003371, -0.00004289, 0.00001939, 0.00001681, -0.00004824, 0.00004623, -0.00004265, 0.00002768, -0.00001281, -0.00000071, 0.00000346, -0.00001124, 0.00003266, -0.00006378, 0.00006473, -0.00006701, 0.00003648, -0.00000905, -0.00004288, 0.00007915, -0.00013332, 0.00012609, -0.00016176, 0.00007979, -0.00010005, -0.00049406, 0.00091970, 0.00032710, -0.00050740, 0.00010584, -0.00008576, -0.00000917, 0.00000992, -0.00004360, 0.00003271, -0.00003535, 0.00001698, -0.00001865, 0.00000815, -0.00000178, -0.00001236, 0.00001769, -0.00002037, 0.00000895, 0.00000110, -0.00002250, 0.00000883, -0.00000740, 0.00000243, 0.00000189, -0.00002275, 0.00000794, -0.00000940, -0.00000816, 0.00001505, -0.00000683, -0.00000599, 0.00001131, 0.00000466, -0.00000249, 0.00000775, -0.00000409, -0.00000172, 0.00000247, -0.00000355, 0.00000320, -0.00000088, 0.00000236, 0.00000152, 0.00000151, 0.00000407, -0.00000044, -0.00000022, -0.00000265, },
		{ -0.00000211, -0.00000198, 0.00000461, -0.00000617, -0.00000009, 0.00000219, 0.00000072, 0.00000503, -0.00000747, -0.00000120, 0.00000133, -0.00000649, 0.00000919, 0.00000417, -0.00000724, -0.00000361, -0.00000561, -0.00000045, 0.00000083, 0.00000658, -0.00000008, -0.00000755, 0.00000518, -0.00000468, 0.00000164, 0.00000129, -0.00001106, 0.00001029, 0.00000099, -0.00000475, 0.00000293, -0.00001048, 0.00000919, 0.00000641, -0.00000493, 0.00000189, -0.00000128, 0.00000243, -0.00000602, 0.00000922, -0.00000021, 0.00000105, 0.00000142, -0.00000704, 0.00000569, -0.00001634, 0.00001513, 0.00000566, -0.00000617, -0.00000452, 0.00001332, -0.00000052, -0.00001330, 0.00002614, -0.00004824, 0.00003912, -0.00000923, -0.00002935, 0.00003824, -0.00004343, 0.00003548, -0.00002658, 0.00001525, -0.00001439, 0.00000937, 0.00001236, -0.00005157, 0.00006602, -0.00008356, 0.00006723, -0.00004974, 0.00000063, 0.00004196, -0.00011570, 0.00013396, -0.00020682, 0.00016534, -0.00025950, -0.00034878, 0.00107091, 0.00005750, -0.00044891, 0.00014992, -0.00013101, 0.00003346, -0.00002038, -0.00002494, 0.00002632, -0.00003718, 0.00002373, -0.00002691, 0.00001709, -0.00001020, -0.00000658, 0.00001659, -0.00002362, 0.00001541, -0.00000649, -0.00001887, 0.00001023, -0.00001045, 0.00000773, -0.00000604, -0.00001899, 0.00001159, -0.00001671, -0.00000123, 0.00001392, -0.00001052, -0.00000150, 0.00001079, 0.00000266, -0.00000056, 0.00000695, -0.00000564, -0.00000028, 0.00000161, -0.00000311, 0.00000332, -0.00000055, 0.00000250, 0.00000084, 0.00000240, 0.00000310, -0.00000040, -0.00000037, -0.00000183, 0.00000424, },
		{ 0.00000027, 0.00000314, -0.00000644, 0.00000187, 0.00000093, 0.00000203, 0.00000315, -0.00000829, 0.00000143, -0.00000106, -0.00000486, 0.00001080, 0.00000115, -0.00000684, -0.00000407, -0.00000511, 0.00000027, 0.00000141, 0.00000711, -0.00000327, -0.00000512, 0.00000495, -0.00000624, 0.00000430, -0.00000234, -0.00000838, 0.00001222, -0.00000325, -0.00000193, 0.00000104, -0.00001005, 0.00001208, 0.00000295, -0.00000366, 0.00000165, -0.00000079, 0.00000175, -0.00000545, 0.00001038, -0.00000203, 0.00000241, -0.00000072, -0.00000469, 0.00000280, -0.00001438, 0.00001773, 0.00000280, -0.00000673, -0.00000259, 0.00001537, -0.00000916, 0.00000025, 0.00001179, -0.00004310, 0.00004962, -0.00003211, -0.00000689, 0.00002300, -0.00003566, 0.00003453, -0.00003272, 0.00002585, -0.00002944, 0.00003092, -0.00001411, -0.00002800, 0.00005278, -0.00008541, 0.00008649, -0.00008544, 0.00004832, -0.00001012, -0.00007376, 0.00011140, -0.00022100, 0.00022821, -0.00043038, -0.00012430, 0.00113203, -0.00020868, -0.00033450, 0.00015076, -0.00015571, 0.00006967, -0.00005368, 0.00000166, 0.00001066, -0.00003094, 0.00002492, -0.00003246, 0.00002580, -0.00002053, 0.00000262, 0.00001175, -0.00002398, 0.00002051, -0.00001483, -0.00001268, 0.00000877, -0.00001108, 0.00001132, -0.00001461, -0.00001155, 0.00001102, -0.00002216, 0.00000702, 0.00000981, -0.00001189, 0.00000261, 0.00000911, 0.00000181, 0.00000106, 0.00000562, -0.00000665, 0.00000107, 0.00000070, -0.00000263, 0.00000352, -0.00000020, 0.00000225, 0.00000062, 0.00000300, 0.00000212, -0.00000017, -0.00000060, -0.00000063, 0.00000383, -0.00000394, },
		{ 0.00000106, -0.00000561, 0.00000301, -0.00000013, 0.00000335, 0.00000059, -0.00000778, 0.00000320, -0.00000356, -0.00000220, 0.00001115, -0.00000152, -0.00000625, -0.00000472, -0.00000421, 0.00000068, 0.00000242, 0.00000665, -0.00000587, -0.00000207, 0.00000333, -0.00000644, 0.00000612, -0.00000597, -0.00000428, 0.00001201, -0.00000644, 0.00000118, -0.00000190, -0.00000812, 0.00001375, -0.00000034, -0.00000189, 0.00000094, 0.00000005, 0.00000069, -0.00000436, 0.00001103, -0.00000369, 0.00000359, -0.00000249, -0.00000272, -0.00000009, -0.00001152, 0.00001940, -0.00000006, -0.00000734, 0.00000104, 0.00001360, -0.00001333, 0.00001149, -0.00000558, -0.00002912, 0.00004914, -0.00004745, 0.00001422, 0.00000439, -0.00002173, 0.00002556, -0.00003015, 0.00002830, -0.00003755, 0.00004820, -0.00004139, 0.00000280, 0.00002660, -0.00007085, 0.00008907, -0.00010801, 0.00009042, -0.00006788, -0.00001342, 0.00005837, -0.00019579, 0.00024889, -0.00058454, 0.00016235, 0.00108742, -0.00043724, -0.00018532, 0.00011004, -0.00015271, 0.00009046, -0.00008210, 0.00003089, -0.00001190, -0.00001655, 0.00001875, -0.00003266, 0.00003163, -0.00003067, 0.00001416, 0.00000313, -0.00002048, 0.00002270, -0.00002231, -0.00000504, 0.00000476, -0.00000880, 0.00001185, -0.00002184, -0.00000185, 0.00000603, -0.00002411, 0.00001480, 0.00000358, -0.00001062, 0.00000540, 0.00000709, 0.00000192, 0.00000211, 0.00000406, -0.00000721, 0.00000234, -0.00000029, -0.00000205, 0.00000384, -0.00000004, 0.00000184, 0.00000078, 0.00000320, 0.00000132, 0.00000009, -0.00000071, 0.00000069, 0.00000252, -0.00000478, 0.00000231, },
		{ -0.00000393, 0.00000320, -0.00000063, 0.00000425, -0.00000218, -0.00000610, 0.00000378, -0.00000554, 0.00000102, 0.00001030, -0.00000358, -0.00000572, -0.00000529, -0.00000310, 0.00000092, 0.00000368, 0.00000516, -0.00000740, 0.00000092, 0.00000075, -0.00000523, 0.00000661, -0.00000881, 0.00000050, 0.00000982, -0.00000799, 0.00000377, -0.00000525, -0.00000487, 0.00001394, -0.00000296, 0.00000000, -0.00000008, 0.00000117, -0.00000073, -0.00000266, 0.00001106, -0.00000511, 0.00000468, -0.00000401, -0.00000114, -0.00000272, -0.00000806, 0.00002027, -0.00000321, -0.00000714, 0.00000512, 0.00000884, -0.00001232, 0.00001753, -0.00002169, -0.00000983, 0.00003841, -0.00005274, 0.00002973, -0.00001321, -0.00000535, 0.00001118, -0.00001991, 0.00002188, -0.00003608, 0.00005666, -0.00006353, 0.00003478, -0.00000794, -0.00004166, 0.00007299, -0.00011143, 0.00011738, -0.00011985, 0.00005456, -0.00001841, -0.00012990, 0.00021261, -0.00069111, 0.00048178, 0.00093564, -0.00059887, -0.00002783, 0.00003772, -0.00012100, 0.00008987, -0.00009821, 0.00005614, -0.00003681, 0.00000387, 0.00000517, -0.00002589, 0.00003207, -0.00003800, 0.00002596, -0.00000828, -0.00001297, 0.00002090, -0.00002726, 0.00000246, -0.00000077, -0.00000392, 0.00000860, -0.00002584, 0.00000802, -0.00000255, -0.00002167, 0.00002033, -0.00000338, -0.00000708, 0.00000631, 0.00000552, 0.00000251, 0.00000255, 0.00000245, -0.00000738, 0.00000352, -0.00000141, -0.00000127, 0.00000422, -0.00000017, 0.00000150, 0.00000114, 0.00000304, 0.00000079, 0.00000025, -0.00000052, 0.00000176, 0.00000062, -0.00000452, 0.00000365, -0.00000109, },
		{ 0.00000255, -0.00000041, 0.00000436, -0.00000462, -0.00000367, 0.00000311, -0.00000654, 0.00000422, 0.00000847, -0.00000491, -0.00000544, -0.00000561, -0.00000197, 0.00000120, 0.00000490, 0.00000282, -0.00000758, 0.00000317, -0.00000214, -0.00000290, 0.00000562, -0.00001024, 0.00000505, 0.00000619, -0.00000773, 0.00000519, -0.00000831, -0.00000078, 0.00001265, -0.00000456, 0.00000162, -0.00000110, 0.00000240, -0.00000242, -0.00000036, 0.00001029, -0.00000611, 0.00000566, -0.00000540, 0.00000013, -0.00000502, -0.00000415, 0.00002031, -0.00000665, -0.00000549, 0.00000821, 0.00000276, -0.00000686, 0.00001693, -0.00003271, 0.00001020, 0.00002032, -0.00004784, 0.00003697, -0.00002601, 0.00000939, -0.00000479, -0.00000490, 0.00000804, -0.00002462, 0.00005350, -0.00007537, 0.00006127, -0.00004411, -0.00000294, 0.00004021, -0.00009337, 0.00012201, -0.00015460, 0.00011647, -0.00010606, -0.00003078, 0.00011332, -0.00072187, 0.00079579, 0.00069074, -0.00067421, 0.00011089, -0.00005002, -0.00006604, 0.00006654, -0.00009685, 0.00007105, -0.00005836, 0.00002631, -0.00001389, -0.00001210, 0.00002555, -0.00003999, 0.00003544, -0.00002052, -0.00000233, 0.00001484, -0.00002841, 0.00000813, -0.00000626, 0.00000240, 0.00000177, -0.00002528, 0.00001572, -0.00001295, -0.00001497, 0.00002230, -0.00000944, -0.00000222, 0.00000535, 0.00000498, 0.00000307, 0.00000258, 0.00000086, -0.00000716, 0.00000456, -0.00000268, -0.00000017, 0.00000449, -0.00000058, 0.00000140, 0.00000150, 0.00000266, 0.00000049, 0.00000030, 0.00000004, 0.00000226, -0.00000141, -0.00000330, 0.00000399, -0.00000276, -0.00000187, },
		{ 0.00000045, 0.00000354, -0.00000624, -0.00000105, 0.00000143, -0.00000631, 0.00000688, 0.00000607, -0.00000561, -0.00000547, -0.00000556, -0.00000098, 0.00000169, 0.00000571, 0.00000007, -0.00000645, 0.00000422, -0.00000465, -0.00000001, 0.00000336, -0.00000994, 0.00000852, 0.00000194, -0.00000589, 0.00000510, -0.00001040, 0.00000350, 0.00001014, -0.00000500, 0.00000264, -0.00000184, 0.00000346, -0.00000415, 0.00000243, 0.00000864, -0.00000649, 0.00000639, -0.00000669, 0.00000121, -0.00000696, 0.00000014, 0.00001929, -0.00001003, -0.00000225, 0.00000914, -0.00000261, 0.00000110, 0.00000998, -0.00003621, 0.00002646, -0.00000084, -0.00003477, 0.00003538, -0.00003167, 0.00001904, -0.00001829, 0.00001084, -0.00000986, -0.00000517, 0.00003845, -0.00007370, 0.00007640, -0.00007446, 0.00003784, -0.00000354, -0.00005595, 0.00010132, -0.00016333, 0.00015835, -0.00018762, 0.00008569, -0.00004361, -0.00065683, 0.00106284, 0.00038088, -0.00065718, 0.00020813, -0.00013413, 0.00000136, 0.00002425, -0.00007659, 0.00007111, -0.00007098, 0.00004567, -0.00003474, 0.00000689, 0.00001198, -0.00003498, 0.00004004, -0.00003110, 0.00000961, 0.00000525, -0.00002532, 0.00001063, -0.00001003, 0.00000843, -0.00000743, -0.00001984, 0.00001928, -0.00002284, -0.00000512, 0.00002023, -0.00001323, 0.00000270, 0.00000302, 0.00000569, 0.00000311, 0.00000244, -0.00000075, -0.00000652, 0.00000533, -0.00000403, 0.00000134, 0.00000443, -0.00000113, 0.00000160, 0.00000168, 0.00000223, 0.00000034, 0.00000034, 0.00000093, 0.00000196, -0.00000308, -0.00000154, 0.00000333, -0.00000353, 0.00000002, 0.00000141, },
		{ 0.00000185, -0.00000672, 0.00000115, -0.00000081, -0.00000486, 0.00000860, 0.00000352, -0.00000585, -0.00000574, -0.00000514, -0.00000019, 0.00000247, 0.00000580, -0.00000258, -0.00000434, 0.00000390, -0.00000619, 0.00000277, 0.00000034, -0.00000797, 0.00001032, -0.00000204, -0.00000305, 0.00000351, -0.00001104, 0.00000727, 0.00000691, -0.00000438, 0.00000292, -0.00000203, 0.00000406, -0.00000561, 0.00000547, 0.00000612, -0.00000604, 0.00000666, -0.00000779, 0.00000211, -0.00000850, 0.00000473, 0.00001699, -0.00001269, 0.00000214, 0.00000739, -0.00000545, 0.00000895, -0.00000142, -0.00003161, 0.00003557, -0.00002042, -0.00001713, 0.00002642, -0.00002974, 0.00002161, -0.00002600, 0.00002313, -0.00002734, 0.00001826, 0.00001391, -0.00005815, 0.00007651, -0.00009243, 0.00007233, -0.00004984, -0.00000564, 0.00005753, -0.00014205, 0.00016910, -0.00024526, 0.00019733, -0.00024070, -0.00048887, 0.00124459, 0.00004442, -0.00055605, 0.00024958, -0.00019672, 0.00006736, -0.00002876, -0.00004024, 0.00005489, -0.00007066, 0.00005695, -0.00005271, 0.00002764, -0.00000697, -0.00002274, 0.00003799, -0.00003750, 0.00002044, -0.00000616, -0.00001853, 0.00000944, -0.00001076, 0.00001225, -0.00001700, -0.00001035, 0.00001758, -0.00002992, 0.00000597, 0.00001453, -0.00001388, 0.00000641, 0.00000022, 0.00000744, 0.00000241, 0.00000239, -0.00000244, -0.00000537, 0.00000566, -0.00000527, 0.00000322, 0.00000388, -0.00000160, 0.00000200, 0.00000160, 0.00000188, 0.00000018, 0.00000053, 0.00000192, 0.00000086, -0.00000397, 0.00000025, 0.00000194, -0.00000327, 0.00000127, 0.00000022, 0.00000165, },
		{ -0.00000601, 0.00000246, -0.00000300, -0.00000251, 0.00000920, 0.00000118, -0.00000588, -0.00000608, -0.00000443, 0.00000048, 0.00000348, 0.00000502, -0.00000460, -0.00000181, 0.00000236, -0.00000643, 0.00000481, -0.00000277, -0.00000478, 0.00001025, -0.00000497, 0.00000003, 0.00000080, -0.00001008, 0.00000998, 0.00000353, -0.00000298, 0.00000248, -0.00000158, 0.00000399, -0.00000645, 0.00000838, 0.00000295, -0.00000469, 0.00000625, -0.00000850, 0.00000276, -0.00000953, 0.00000941, 0.00001326, -0.00001391, 0.00000671, 0.00000329, -0.00000468, 0.00001408, -0.00001424, -0.00002031, 0.00003593, -0.00003449, 0.00000082, 0.00001309, -0.00002165, 0.00001697, -0.00002613, 0.00002867, -0.00003994, 0.00004056, -0.00001550, -0.00003133, 0.00006100, -0.00009385, 0.00009316, -0.00008923, 0.00004790, -0.00000207, -0.00009289, 0.00014319, -0.00026411, 0.00027950, -0.00044993, -0.00022650, 0.00131243, -0.00027594, -0.00039197, 0.00023175, -0.00022462, 0.00011810, -0.00008145, 0.00000552, 0.00002465, -0.00005609, 0.00005654, -0.00006320, 0.00004574, -0.00002805, -0.00000470, 0.00002884, -0.00003792, 0.00002772, -0.00001708, -0.00000955, 0.00000497, -0.00000793, 0.00001233, -0.00002459, 0.00000129, 0.00001069, -0.00003241, 0.00001607, 0.00000650, -0.00001134, 0.00000797, -0.00000205, 0.00000968, 0.00000099, 0.00000254, -0.00000423, -0.00000365, 0.00000535, -0.00000614, 0.00000532, 0.00000273, -0.00000175, 0.00000245, 0.00000132, 0.00000169, -0.00000006, 0.00000105, 0.00000269, -0.00000083, -0.00000388, 0.00000153, 0.00000032, -0.00000211, 0.00000161, -0.00000055, 0.00000276, -0.00000036, },
		{ 0.00000263, -0.00000460, 0.00000028, 0.00000875, -0.00000071, -0.00000590, -0.00000635, -0.00000358, 0.00000114, 0.00000451, 0.00000340, -0.00000558, 0.00000051, 0.00000006, -0.00000537, 0.00000568, -0.00000525, -0.00000106, 0.00000850, -0.00000639, 0.00000258, -0.00000241, -0.00000771, 0.00001127, 0.00000056, -0.00000122, 0.00000153, -0.00000054, 0.00000315, -0.00000639, 0.00001072, -0.00000053, -0.00000251, 0.00000501, -0.00000862, 0.00000300, -0.00000985, 0.00001380, 0.00000827, -0.00001310, 0.00001026, -0.00000206, -0.00000031, 0.00001464, -0.00002515, -0.00000522, 0.00002803, -0.00004065, 0.00001523, -0.00000091, -0.00001024, 0.00000682, -0.00001886, 0.00002586, -0.00004422, 0.00005663, -0.00004391, 0.00000169, 0.00003260, -0.00007793, 0.00009558, -0.00011331, 0.00009378, -0.00006632, -0.00002409, 0.00008242, -0.00023586, 0.00030984, -0.00063668, 0.00010592, 0.00125263, -0.00053993, -0.00019510, 0.00016228, -0.00021214, 0.00014280, -0.00012247, 0.00005154, -0.00001404, -0.00002914, 0.00004318, -0.00006281, 0.00005682, -0.00004706, 0.00001618, 0.00001371, -0.00003173, 0.00002963, -0.00002515, -0.00000051, -0.00000136, -0.00000206, 0.00000796, -0.00002808, 0.00001252, -0.00000010, -0.00002954, 0.00002316, -0.00000208, -0.00000628, 0.00000700, -0.00000291, 0.00001170, -0.00000085, 0.00000287, -0.00000605, -0.00000142, 0.00000427, -0.00000636, 0.00000733, 0.00000107, -0.00000142, 0.00000272, 0.00000096, 0.00000161, -0.00000037, 0.00000198, 0.00000290, -0.00000270, -0.00000286, 0.00000197, -0.00000098, -0.00000048, 0.00000104, -0.00000062, 0.00000335, -0.00000263, -0.00000435, },
		{ -0.00000522, 0.00000299, 0.00000748, -0.00000210, -0.00000603, -0.00000641, -0.00000272, 0.00000193, 0.00000527, 0.00000120, -0.00000540, 0.00000207, -0.00000240, -0.00000331, 0.00000527, -0.00000661, 0.00000242, 0.00000561, -0.00000622, 0.00000404, -0.00000541, -0.00000438, 0.00001111, -0.00000159, 0.00000046, 0.00000040, 0.00000087, 0.00000161, -0.00000530, 0.00001208, -0.00000384, 0.00000022, 0.00000296, -0.00000800, 0.00000269, -0.00000921, 0.00001740, 0.00000249, -0.00001009, 0.00001169, -0.00000705, 0.00000656, 0.00001000, -0.00003136, 0.00000996, 0.00001419, -0.00003849, 0.00002349, -0.00001218, 0.00000105, -0.00000581, -0.00000621, 0.00001517, -0.00003866, 0.00006265, -0.00006552, 0.00003438, -0.00000321, -0.00004753, 0.00007860, -0.00011659, 0.00012216, -0.00012245, 0.00005135, -0.00000382, -0.00016121, 0.00027312, -0.00076530, 0.00047093, 0.00106922, -0.00071617, 0.00000074, 0.00005810, -0.00016203, 0.00013608, -0.00014263, 0.00008816, -0.00005333, 0.00000533, 0.00001853, -0.00005035, 0.00005762, -0.00005975, 0.00003596, -0.00000480, -0.00001981, 0.00002548, -0.00002854, 0.00000636, -0.00000755, 0.00000534, -0.00000041, -0.00002616, 0.00002071, -0.00001252, -0.00002177, 0.00002589, -0.00000932, -0.00000000, 0.00000378, -0.00000187, 0.00001280, -0.00000267, 0.00000317, -0.00000773, 0.00000115, 0.00000240, -0.00000568, 0.00000891, -0.00000090, -0.00000058, 0.00000264, 0.00000069, 0.00000151, -0.00000065, 0.00000329, 0.00000231, -0.00000429, -0.00000125, 0.00000144, -0.00000150, 0.00000109, -0.00000015, 0.00000006, 0.00000311, -0.00000442, -0.00000220, 0.00000171, },
		{ 0.00000514, 0.00000575, -0.00000304, -0.00000629, -0.00000622, -0.00000190, 0.00000295, 0.00000545, -0.00000113, -0.00000418, 0.00000252, -0.00000438, -0.00000079, 0.00000378, -0.00000660, 0.00000496, 0.00000234, -0.00000475, 0.00000412, -0.00000755, -0.00000072, 0.00000970, -0.00000272, 0.00000173, -0.00000058, 0.00000234, -0.00000035, -0.00000321, 0.00001220, -0.00000650, 0.00000309, 0.00000028, -0.00000664, 0.00000176, -0.00000745, 0.00001962, -0.00000326, -0.00000523, 0.00001035, -0.00001004, 0.00001402, 0.00000092, -0.00003127, 0.00002168, -0.00000212, -0.00002945, 0.00002471, -0.00001835, 0.00000908, -0.00001738, 0.00000848, -0.00000101, -0.00002399, 0.00005689, -0.00007585, 0.00006023, -0.00003930, -0.00000863, 0.00004530, -0.00009781, 0.00012651, -0.00015872, 0.00011816, -0.00009998, -0.00005078, 0.00016523, -0.00080545, 0.00082317, 0.00078392, -0.00078715, 0.00016406, -0.00005822, -0.00008474, 0.00009916, -0.00013699, 0.00010739, -0.00008484, 0.00004043, -0.00001305, -0.00002725, 0.00004685, -0.00006287, 0.00005058, -0.00002312, -0.00000445, 0.00001597, -0.00002648, 0.00000932, -0.00001147, 0.00001208, -0.00001115, -0.00001876, 0.00002377, -0.00002385, -0.00001063, 0.00002393, -0.00001370, 0.00000595, -0.00000086, 0.00000108, 0.00001249, -0.00000402, 0.00000318, -0.00000897, 0.00000377, -0.00000008, -0.00000399, 0.00000969, -0.00000283, 0.00000066, 0.00000215, 0.00000068, 0.00000123, -0.00000068, 0.00000476, 0.00000084, -0.00000512, 0.00000047, 0.00000017, -0.00000101, 0.00000206, -0.00000152, 0.00000130, 0.00000196, -0.00000530, -0.00000013, 0.00000003, -0.00000470, },
		{ 0.00000391, -0.00000369, -0.00000660, -0.00000582, -0.00000111, 0.00000415, 0.00000485, -0.00000310, -0.00000231, 0.00000180, -0.00000542, 0.00000159, 0.00000169, -0.00000535, 0.00000615, -0.00000057, -0.00000253, 0.00000292, -0.00000843, 0.00000263, 0.00000748, -0.00000288, 0.00000240, -0.00000113, 0.00000351, -0.00000238, -0.00000038, 0.00001098, -0.00000809, 0.00000562, -0.00000268, -0.00000469, 0.00000031, -0.00000456, 0.00002003, -0.00000805, 0.00000059, 0.00000629, -0.00000980, 0.00001983, -0.00001064, -0.00002484, 0.00002738, -0.00001711, -0.00001638, 0.00001977, -0.00001863, 0.00001179, -0.00002477, 0.00002144, -0.00001891, -0.00000307, 0.00004018, -0.00007278, 0.00007412, -0.00006849, 0.00003099, 0.00000226, -0.00006034, 0.00010509, -0.00016707, 0.00016197, -0.00018716, 0.00007620, -0.00000483, -0.00073800, 0.00111626, 0.00043301, -0.00075171, 0.00027108, -0.00016306, 0.00000388, 0.00003963, -0.00010605, 0.00010473, -0.00010153, 0.00006870, -0.00004535, 0.00000254, 0.00002571, -0.00005499, 0.00005683, -0.00003751, 0.00001118, 0.00000312, -0.00001954, 0.00000766, -0.00001153, 0.00001585, -0.00002175, -0.00000715, 0.00002075, -0.00003156, 0.00000157, 0.00001800, -0.00001444, 0.00001016, -0.00000574, 0.00000546, 0.00001064, -0.00000458, 0.00000267, -0.00000950, 0.00000603, -0.00000285, -0.00000136, 0.00000944, -0.00000436, 0.00000204, 0.00000131, 0.00000097, 0.00000066, -0.00000023, 0.00000606, -0.00000137, -0.00000491, 0.00000174, -0.00000142, 0.00000043, 0.00000206, -0.00000253, 0.00000269, 0.00000009, -0.00000508, 0.00000132, -0.00000192, -0.00000318, 0.00000645, },
		{ -0.00000422, -0.00000685, -0.00000529, -0.00000023, 0.00000539, 0.00000343, -0.00000429, -0.00000035, 0.00000017, -0.00000528, 0.00000330, -0.00000044, -0.00000327, 0.00000590, -0.00000251, -0.00000027, 0.00000083, -0.00000794, 0.00000515, 0.00000498, -0.00000230, 0.00000247, -0.00000113, 0.00000411, -0.00000407, 0.00000276, 0.00000860, -0.00000841, 0.00000737, -0.00000549, -0.00000244, -0.00000143, -0.00000078, 0.00001844, -0.00001099, 0.00000615, 0.00000030, -0.00000591, 0.00002205, -0.00002199, -0.00001362, 0.00002602, -0.00002764, -0.00000271, 0.00001089, -0.00001385, 0.00000868, -0.00002599, 0.00002935, -0.00003437, 0.00001981, 0.00001567, -0.00005699, 0.00007348, -0.00008505, 0.00006337, -0.00004185, -0.00001155, 0.00006153, -0.00014498, 0.00017250, -0.00024696, 0.00019497, -0.00021520, -0.00055927, 0.00131023, 0.00006175, -0.00062476, 0.00030947, -0.00023637, 0.00008610, -0.00003026, -0.00005568, 0.00008027, -0.00009932, 0.00008384, -0.00007159, 0.00003329, -0.00000224, -0.00003694, 0.00005310, -0.00004493, 0.00002365, -0.00001014, -0.00000956, 0.00000193, -0.00000712, 0.00001486, -0.00002944, 0.00000621, 0.00001213, -0.00003392, 0.00001233, 0.00000970, -0.00001164, 0.00001166, -0.00000962, 0.00001043, 0.00000746, -0.00000422, 0.00000151, -0.00000911, 0.00000756, -0.00000547, 0.00000192, 0.00000807, -0.00000516, 0.00000326, 0.00000031, 0.00000150, -0.00000023, 0.00000087, 0.00000677, -0.00000392, -0.00000365, 0.00000213, -0.00000273, 0.00000246, 0.00000099, -0.00000277, 0.00000374, -0.00000210, -0.00000383, 0.00000178, -0.00000357, -0.00000101, 0.00000682, 0.00000223, },
		{ -0.00000694, -0.00000475, 0.00000085, 0.00000638, 0.00000135, -0.00000445, 0.00000116, -0.00000191, -0.00000406, 0.00000403, -0.00000204, -0.00000095, 0.00000450, -0.00000322, 0.00000142, -0.00000157, -0.00000629, 0.00000654, 0.00000266, -0.00000130, 0.00000209, -0.00000059, 0.00000402, -0.00000506, 0.00000573, 0.00000541, -0.00000746, 0.00000804, -0.00000773, -0.00000027, -0.00000307, 0.00000344, 0.00001505, -0.00001153, 0.00001018, -0.00000623, 0.00000106, 0.00001955, -0.00003043, -0.00000026, 0.00001826, -0.00003183, 0.00000842, 0.00000100, -0.00000610, 0.00000080, -0.00002069, 0.00003019, -0.00004377, 0.00003990, -0.00001182, -0.00003176, 0.00005875, -0.00008599, 0.00008213, -0.00007811, 0.00003872, 0.00000416, -0.00009621, 0.00014594, -0.00026544, 0.00027989, -0.00043410, -0.00028283, 0.00137806, -0.00028283, -0.00043410, 0.00027989, -0.00026544, 0.00014594, -0.00009621, 0.00000416, 0.00003872, -0.00007811, 0.00008213, -0.00008599, 0.00005875, -0.00003176, -0.00001182, 0.00003990, -0.00004377, 0.00003019, -0.00002069, 0.00000080, -0.00000610, 0.00000100, 0.00000842, -0.00003183, 0.00001826, -0.00000026, -0.00003043, 0.00001955, 0.00000106, -0.00000623, 0.00001018, -0.00001153, 0.00001505, 0.00000344, -0.00000307, -0.00000027, -0.00000773, 0.00000804, -0.00000746, 0.00000541, 0.00000573, -0.00000506, 0.00000402, -0.00000059, 0.00000209, -0.00000130, 0.00000266, 0.00000654, -0.00000629, -0.00000157, 0.00000142, -0.00000322, 0.00000450, -0.00000095, -0.00000204, 0.00000403, -0.00000406, -0.00000191, 0.00000116, -0.00000445, 0.00000135, 0.00000638, 0.00000085, -0.00000475, },
		{ -0.00000422, 0.00000223, 0.00000682, -0.00000101, -0.00000357, 0.00000178, -0.00000383, -0.00000210, 0.00000374, -0.00000277, 0.00000099, 0.00000246, -0.00000273, 0.00000213, -0.00000365, -0.00000392, 0.00000677, 0.00000087, -0.00000023, 0.00000150, 0.00000031, 0.00000326, -0.00000516, 0.00000807, 0.00000192, -0.00000547, 0.00000756, -0.00000911, 0.00000151, -0.00000422, 0.00000746, 0.00001043, -0.00000962, 0.00001166, -0.00001164, 0.00000970, 0.00001233, -0.00003392, 0.00001213, 0.00000621, -0.00002944, 0.00001486, -0.00000712, 0.00000193, -0.00000956, -0.00001014, 0.00002365, -0.00004493, 0.00005310, -0.00003694, -0.00000224, 0.00003329, -0.00007159, 0.00008384, -0.00009932, 0.00008027, -0.00005568, -0.00003026, 0.00008610, -0.00023637, 0.00030947, -0.00062476, 0.00006175, 0.00131023, -0.00055927, -0.00021520, 0.00019497, -0.00024696, 0.00017250, -0.00014498, 0.00006153, -0.00001155, -0.00004185, 0.00006337, -0.00008505, 0.00007348, -0.00005699, 0.00001567, 0.00001981, -0.00003437, 0.00002935, -0.00002599, 0.00000868, -0.00001385, 0.00001089, -0.00000271, -0.00002764, 0.00002602, -0.00001362, -0.00002199, 0.00002205, -0.00000591, 0.00000030, 0.00000615, -0.00001099, 0.00001844, -0.00000078, -0.00000143, -0.00000244, -0.00000549, 0.00000737, -0.00000841, 0.00000860, 0.00000276, -0.00000407, 0.00000411, -0.00000113, 0.00000247, -0.00000230, 0.00000498, 0.00000515, -0.00000794, 0.00000083, -0.00000027, -0.00000251, 0.00000590, -0.00000327, -0.00000044, 0.00000330, -0.00000528, 0.00000017, -0.00000035, -0.00000429, 0.00000343, 0.00000539, -0.00000023, -0.00000529, -0.00000685, },
		{ 0.00000391, 0.00000645, -0.00000318, -0.00000192, 0.00000132, -0.00000508, 0.00000009, 0.00000269, -0.00000253, 0.00000206, 0.00000043, -0.00000142, 0.00000174, -0.00000491, -0.00000137, 0.00000606, -0.00000023, 0.00000066, 0.00000097, 0.00000131, 0.00000204, -0.00000436, 0.00000944, -0.00000136, -0.00000285, 0.00000603, -0.00000950, 0.00000267, -0.00000458, 0.00001064, 0.00000546, -0.00000574, 0.00001016, -0.00001444, 0.00001800, 0.00000157, -0.00003156, 0.00002075, -0.00000715, -0.00002175, 0.00001585, -0.00001153, 0.00000766, -0.00001954, 0.00000312, 0.00001118, -0.00003751, 0.00005683, -0.00005499, 0.00002571, 0.00000254, -0.00004535, 0.00006870, -0.00010153, 0.00010473, -0.00010605, 0.00003963, 0.00000388, -0.00016306, 0.00027108, -0.00075171, 0.00043301, 0.00111626, -0.00073800, -0.00000483, 0.00007620, -0.00018716, 0.00016197, -0.00016707, 0.00010509, -0.00006034, 0.00000226, 0.00003099, -0.00006849, 0.00007412, -0.00007278, 0.00004018, -0.00000307, -0.00001891, 0.00002144, -0.00002477, 0.00001179, -0.00001863, 0.00001977, -0.00001638, -0.00001711, 0.00002738, -0.00002484, -0.00001064, 0.00001983, -0.00000980, 0.00000629, 0.00000059, -0.00000805, 0.00002003, -0.00000456, 0.00000031, -0.00000469, -0.00000268, 0.00000562, -0.00000809, 0.00001098, -0.00000038, -0.00000238, 0.00000351, -0.00000113, 0.00000240, -0.00000288, 0.00000748, 0.00000263, -0.00000843, 0.00000292, -0.00000253, -0.00000057, 0.00000615, -0.00000535, 0.00000169, 0.00000159, -0.00000542, 0.00000180, -0.00000231, -0.00000310, 0.00000485, 0.00000415, -0.00000111, -0.00000582, -0.00000660, -0.00000369, },
		{ 0.00000514, -0.00000470, 0.00000003, -0.00000013, -0.00000530, 0.00000196, 0.00000130, -0.00000152, 0.00000206, -0.00000101, 0.00000017, 0.00000047, -0.00000512, 0.00000084, 0.00000476, -0.00000068, 0.00000123, 0.00000068, 0.00000215, 0.00000066, -0.00000283, 0.00000969, -0.00000399, -0.00000008, 0.00000377, -0.00000897, 0.00000318, -0.00000402, 0.00001249, 0.00000108, -0.00000086, 0.00000595, -0.00001370, 0.00002393, -0.00001063, -0.00002385, 0.00002377, -0.00001876, -0.00001115, 0.00001208, -0.00001147, 0.00000932, -0.00002648, 0.00001597, -0.00000445, -0.00002312, 0.00005058, -0.00006287, 0.00004685, -0.00002725, -0.00001305, 0.00004043, -0.00008484, 0.00010739, -0.00013699, 0.00009916, -0.00008474, -0.00005822, 0.00016406, -0.00078715, 0.00078392, 0.00082317, -0.00080545, 0.00016523, -0.00005078, -0.00009998, 0.00011816, -0.00015872, 0.00012651, -0.00009781, 0.00004530, -0.00000863, -0.00003930, 0.00006023, -0.00007585, 0.00005689, -0.00002399, -0.00000101, 0.00000848, -0.00001738, 0.00000908, -0.00001835, 0.00002471, -0.00002945, -0.00000212, 0.00002168, -0.00003127, 0.00000092, 0.00001402, -0.00001004, 0.00001035, -0.00000523, -0.00000326, 0.00001962, -0.00000745, 0.00000176, -0.00000664, 0.00000028, 0.00000309, -0.00000650, 0.00001220, -0.00000321, -0.00000035, 0.00000234, -0.00000058, 0.00000173, -0.00000272, 0.00000970, -0.00000072, -0.00000755, 0.00000412, -0.00000475, 0.00000234, 0.00000496, -0.00000660, 0.00000378, -0.00000079, -0.00000438, 0.00000252, -0.00000418, -0.00000113, 0.00000545, 0.00000295, -0.00000190, -0.00000622, -0.00000629, -0.00000304, 0.00000575, },
		{ -0.00000522, 0.00000171, -0.00000220, -0.00000442, 0.00000311, 0.00000006, -0.00000015, 0.00000109, -0.00000150, 0.00000144, -0.00000125, -0.00000429, 0.00000231, 0.00000329, -0.00000065, 0.00000151, 0.00000069, 0.00000264, -0.00000058, -0.00000090, 0.00000891, -0.00000568, 0.00000240, 0.00000115, -0.00000773, 0.00000317, -0.00000267, 0.00001280, -0.00000187, 0.00000378, -0.00000000, -0.00000932, 0.00002589, -0.00002177, -0.00001252, 0.00002071, -0.00002616, -0.00000041, 0.00000534, -0.00000755, 0.00000636, -0.00002854, 0.00002548, -0.00001981, -0.00000480, 0.00003596, -0.00005975, 0.00005762, -0.00005035, 0.00001853, 0.00000533, -0.00005333, 0.00008816, -0.00014263, 0.00013608, -0.00016203, 0.00005810, 0.00000074, -0.00071617, 0.00106922, 0.00047093, -0.00076530, 0.00027312, -0.00016121, -0.00000382, 0.00005135, -0.00012245, 0.00012216, -0.00011659, 0.00007860, -0.00004753, -0.00000321, 0.00003438, -0.00006552, 0.00006265, -0.00003866, 0.00001517, -0.00000621, -0.00000581, 0.00000105, -0.00001218, 0.00002349, -0.00003849, 0.00001419, 0.00000996, -0.00003136, 0.00001000, 0.00000656, -0.00000705, 0.00001169, -0.00001009, 0.00000249, 0.00001740, -0.00000921, 0.00000269, -0.00000800, 0.00000296, 0.00000022, -0.00000384, 0.00001208, -0.00000530, 0.00000161, 0.00000087, 0.00000040, 0.00000046, -0.00000159, 0.00001111, -0.00000438, -0.00000541, 0.00000404, -0.00000622, 0.00000561, 0.00000242, -0.00000661, 0.00000527, -0.00000331, -0.00000240, 0.00000207, -0.00000540, 0.00000120, 0.00000527, 0.00000193, -0.00000272, -0.00000641, -0.00000603, -0.00000210, 0.00000748, 0.00000299, },
		{ 0.00000263, -0.00000435, -0.00000263, 0.00000335, -0.00000062, 0.00000104, -0.00000048, -0.00000098, 0.00000197, -0.00000286, -0.00000270, 0.00000290, 0.00000198, -0.00000037, 0.00000161, 0.00000096, 0.00000272, -0.00000142, 0.00000107, 0.00000733, -0.00000636, 0.00000427, -0.00000142, -0.00000605, 0.00000287, -0.00000085, 0.00001170, -0.00000291, 0.00000700, -0.00000628, -0.00000208, 0.00002316, -0.00002954, -0.00000010, 0.00001252, -0.00002808, 0.00000796, -0.00000206, -0.00000136, -0.00000051, -0.00002515, 0.00002963, -0.00003173, 0.00001371, 0.00001618, -0.00004706, 0.00005682, -0.00006281, 0.00004318, -0.00002914, -0.00001404, 0.00005154, -0.00012247, 0.00014280, -0.00021214, 0.00016228, -0.00019510, -0.00053993, 0.00125263, 0.00010592, -0.00063668, 0.00030984, -0.00023586, 0.00008242, -0.00002409, -0.00006632, 0.00009378, -0.00011331, 0.00009558, -0.00007793, 0.00003260, 0.00000169, -0.00004391, 0.00005663, -0.00004422, 0.00002586, -0.00001886, 0.00000682, -0.00001024, -0.00000091, 0.00001523, -0.00004065, 0.00002803, -0.00000522, -0.00002515, 0.00001464, -0.00000031, -0.00000206, 0.00001026, -0.00001310, 0.00000827, 0.00001380, -0.00000985, 0.00000300, -0.00000862, 0.00000501, -0.00000251, -0.00000053, 0.00001072, -0.00000639, 0.00000315, -0.00000054, 0.00000153, -0.00000122, 0.00000056, 0.00001127, -0.00000771, -0.00000241, 0.00000258, -0.00000639, 0.00000850, -0.00000106, -0.00000525, 0.00000568, -0.00000537, 0.00000006, 0.00000051, -0.00000558, 0.00000340, 0.00000451, 0.00000114, -0.00000358, -0.00000635, -0.00000590, -0.00000071, 0.00000875, 0.00000028, -0.00000460, },
		{ -0.00000601, -0.00000036, 0.00000276, -0.00000055, 0.00000161, -0.00000211, 0.00000032, 0.00000153, -0.00000388, -0.00000083, 0.00000269, 0.00000105, -0.00000006, 0.00000169, 0.00000132, 0.00000245, -0.00000175, 0.00000273, 0.00000532, -0.00000614, 0.00000535, -0.00000365, -0.00000423, 0.00000254, 0.00000099, 0.00000968, -0.00000205, 0.00000797, -0.00001134, 0.00000650, 0.00001607, -0.00003241, 0.00001069, 0.00000129, -0.00002459, 0.00001233, -0.00000793, 0.00000497, -0.00000955, -0.00001708, 0.00002772, -0.00003792, 0.00002884, -0.00000470, -0.00002805, 0.00004574, -0.00006320, 0.00005654, -0.00005609, 0.00002465, 0.00000552, -0.00008145, 0.00011810, -0.00022462, 0.00023175, -0.00039197, -0.00027594, 0.00131243, -0.00022650, -0.00044993, 0.00027950, -0.00026411, 0.00014319, -0.00009289, -0.00000207, 0.00004790, -0.00008923, 0.00009316, -0.00009385, 0.00006100, -0.00003133, -0.00001550, 0.00004056, -0.00003994, 0.00002867, -0.00002613, 0.00001697, -0.00002165, 0.00001309, 0.00000082, -0.00003449, 0.00003593, -0.00002031, -0.00001424, 0.00001408, -0.00000468, 0.00000329, 0.00000671, -0.00001391, 0.00001326, 0.00000941, -0.00000953, 0.00000276, -0.00000850, 0.00000625, -0.00000469, 0.00000295, 0.00000838, -0.00000645, 0.00000399, -0.00000158, 0.00000248, -0.00000298, 0.00000353, 0.00000998, -0.00001008, 0.00000080, 0.00000003, -0.00000497, 0.00001025, -0.00000478, -0.00000277, 0.00000481, -0.00000643, 0.00000236, -0.00000181, -0.00000460, 0.00000502, 0.00000348, 0.00000048, -0.00000443, -0.00000608, -0.00000588, 0.00000118, 0.00000920, -0.00000251, -0.00000300, 0.00000246, },
		{ 0.00000185, 0.00000165, 0.00000022, 0.00000127, -0.00000327, 0.00000194, 0.00000025, -0.00000397, 0.00000086, 0.00000192, 0.00000053, 0.00000018, 0.00000188, 0.00000160, 0.00000200, -0.00000160, 0.00000388, 0.00000322, -0.00000527, 0.00000566, -0.00000537, -0.00000244, 0.00000239, 0.00000241, 0.00000744, 0.00000022, 0.00000641, -0.00001388, 0.00001453, 0.00000597, -0.00002992, 0.00001758, -0.00001035, -0.00001700, 0.00001225, -0.00001076, 0.00000944, -0.00001853, -0.00000616, 0.00002044, -0.00003750, 0.00003799, -0.00002274, -0.00000697, 0.00002764, -0.00005271, 0.00005695, -0.00007066, 0.00005489, -0.00004024, -0.00002876, 0.00006736, -0.00019672, 0.00024958, -0.00055605, 0.00004442, 0.00124459, -0.00048887, -0.00024070, 0.00019733, -0.00024526, 0.00016910, -0.00014205, 0.00005753, -0.00000564, -0.00004984, 0.00007233, -0.00009243, 0.00007651, -0.00005815, 0.00001391, 0.00001826, -0.00002734, 0.00002313, -0.00002600, 0.00002161, -0.00002974, 0.00002642, -0.00001713, -0.00002042, 0.00003557, -0.00003161, -0.00000142, 0.00000895, -0.00000545, 0.00000739, 0.00000214, -0.00001269, 0.00001699, 0.00000473, -0.00000850, 0.00000211, -0.00000779, 0.00000666, -0.00000604, 0.00000612, 0.00000547, -0.00000561, 0.00000406, -0.00000203, 0.00000292, -0.00000438, 0.00000691, 0.00000727, -0.00001104, 0.00000351, -0.00000305, -0.00000204, 0.00001032, -0.00000797, 0.00000034, 0.00000277, -0.00000619, 0.00000390, -0.00000434, -0.00000258, 0.00000580, 0.00000247, -0.00000019, -0.00000514, -0.00000574, -0.00000585, 0.00000352, 0.00000860, -0.00000486, -0.00000081, 0.00000115, -0.00000672, },
		{ 0.00000045, 0.00000141, 0.00000002, -0.00000353, 0.00000333, -0.00000154, -0.00000308, 0.00000196, 0.00000093, 0.00000034, 0.00000034, 0.00000223, 0.00000168, 0.00000160, -0.00000113, 0.00000443, 0.00000134, -0.00000403, 0.00000533, -0.00000652, -0.00000075, 0.00000244, 0.00000311, 0.00000569, 0.00000302, 0.00000270, -0.00001323, 0.00002023, -0.00000512, -0.00002284, 0.00001928, -0.00001984, -0.00000743, 0.00000843, -0.00001003, 0.00001063, -0.00002532, 0.00000525, 0.00000961, -0.00003110, 0.00004004, -0.00003498, 0.00001198, 0.00000689, -0.00003474, 0.00004567, -0.00007098, 0.00007111, -0.00007659, 0.00002425, 0.00000136, -0.00013413, 0.00020813, -0.00065718, 0.00038088, 0.00106284, -0.00065683, -0.00004361, 0.00008569, -0.00018762, 0.00015835, -0.00016333, 0.00010132, -0.00005595, -0.00000354, 0.00003784, -0.00007446, 0.00007640, -0.00007370, 0.00003845, -0.00000517, -0.00000986, 0.00001084, -0.00001829, 0.00001904, -0.00003167, 0.00003538, -0.00003477, -0.00000084, 0.00002646, -0.00003621, 0.00000998, 0.00000110, -0.00000261, 0.00000914, -0.00000225, -0.00001003, 0.00001929, 0.00000014, -0.00000696, 0.00000121, -0.00000669, 0.00000639, -0.00000649, 0.00000864, 0.00000243, -0.00000415, 0.00000346, -0.00000184, 0.00000264, -0.00000500, 0.00001014, 0.00000350, -0.00001040, 0.00000510, -0.00000589, 0.00000194, 0.00000852, -0.00000994, 0.00000336, -0.00000001, -0.00000465, 0.00000422, -0.00000645, 0.00000007, 0.00000571, 0.00000169, -0.00000098, -0.00000556, -0.00000547, -0.00000561, 0.00000607, 0.00000688, -0.00000631, 0.00000143, -0.00000105, -0.00000624, 0.00000354, },
		{ 0.00000255, -0.00000187, -0.00000276, 0.00000399, -0.00000330, -0.00000141, 0.00000226, 0.00000004, 0.00000030, 0.00000049, 0.00000266, 0.00000150, 0.00000140, -0.00000058, 0.00000449, -0.00000017, -0.00000268, 0.00000456, -0.00000716, 0.00000086, 0.00000258, 0.00000307, 0.00000498, 0.00000535, -0.00000222, -0.00000944, 0.00002230, -0.00001497, -0.00001295, 0.00001572, -0.00002528, 0.00000177, 0.00000240, -0.00000626, 0.00000813, -0.00002841, 0.00001484, -0.00000233, -0.00002052, 0.00003544, -0.00003999, 0.00002555, -0.00001210, -0.00001389, 0.00002631, -0.00005836, 0.00007105, -0.00009685, 0.00006654, -0.00006604, -0.00005002, 0.00011089, -0.00067421, 0.00069074, 0.00079579, -0.00072187, 0.00011332, -0.00003078, -0.00010606, 0.00011647, -0.00015460, 0.00012201, -0.00009337, 0.00004021, -0.00000294, -0.00004411, 0.00006127, -0.00007537, 0.00005350, -0.00002462, 0.00000804, -0.00000490, -0.00000479, 0.00000939, -0.00002601, 0.00003697, -0.00004784, 0.00002032, 0.00001020, -0.00003271, 0.00001693, -0.00000686, 0.00000276, 0.00000821, -0.00000549, -0.00000665, 0.00002031, -0.00000415, -0.00000502, 0.00000013, -0.00000540, 0.00000566, -0.00000611, 0.00001029, -0.00000036, -0.00000242, 0.00000240, -0.00000110, 0.00000162, -0.00000456, 0.00001265, -0.00000078, -0.00000831, 0.00000519, -0.00000773, 0.00000619, 0.00000505, -0.00001024, 0.00000562, -0.00000290, -0.00000214, 0.00000317, -0.00000758, 0.00000282, 0.00000490, 0.00000120, -0.00000197, -0.00000561, -0.00000544, -0.00000491, 0.00000847, 0.00000422, -0.00000654, 0.00000311, -0.00000367, -0.00000462, 0.00000436, -0.00000041, },
		{ -0.00000393, -0.00000109, 0.00000365, -0.00000452, 0.00000062, 0.00000176, -0.00000052, 0.00000025, 0.00000079, 0.00000304, 0.00000114, 0.00000150, -0.00000017, 0.00000422, -0.00000127, -0.00000141, 0.00000352, -0.00000738, 0.00000245, 0.00000255, 0.00000251, 0.00000552, 0.00000631, -0.00000708, -0.00000338, 0.00002033, -0.00002167, -0.00000255, 0.00000802, -0.00002584, 0.00000860, -0.00000392, -0.00000077, 0.00000246, -0.00002726, 0.00002090, -0.00001297, -0.00000828, 0.00002596, -0.00003800, 0.00003207, -0.00002589, 0.00000517, 0.00000387, -0.00003681, 0.00005614, -0.00009821, 0.00008987, -0.00012100, 0.00003772, -0.00002783, -0.00059887, 0.00093564, 0.00048178, -0.00069111, 0.00021261, -0.00012990, -0.00001841, 0.00005456, -0.00011985, 0.00011738, -0.00011143, 0.00007299, -0.00004166, -0.00000794, 0.00003478, -0.00006353, 0.00005666, -0.00003608, 0.00002188, -0.00001991, 0.00001118, -0.00000535, -0.00001321, 0.00002973, -0.00005274, 0.00003841, -0.00000983, -0.00002169, 0.00001753, -0.00001232, 0.00000884, 0.00000512, -0.00000714, -0.00000321, 0.00002027, -0.00000806, -0.00000272, -0.00000114, -0.00000401, 0.00000468, -0.00000511, 0.00001106, -0.00000266, -0.00000073, 0.00000117, -0.00000008, 0.00000000, -0.00000296, 0.00001394, -0.00000487, -0.00000525, 0.00000377, -0.00000799, 0.00000982, 0.00000050, -0.00000881, 0.00000661, -0.00000523, 0.00000075, 0.00000092, -0.00000740, 0.00000516, 0.00000368, 0.00000092, -0.00000310, -0.00000529, -0.00000572, -0.00000358, 0.00001030, 0.00000102, -0.00000554, 0.00000378, -0.00000610, -0.00000218, 0.00000425, -0.00000063, 0.00000320, },
		{ 0.00000106, 0.00000231, -0.00000478, 0.00000252, 0.00000069, -0.00000071, 0.00000009, 0.00000132, 0.00000320, 0.00000078, 0.00000184, -0.00000004, 0.00000384, -0.00000205, -0.00000029, 0.00000234, -0.00000721, 0.00000406, 0.00000211, 0.00000192, 0.00000709, 0.00000540, -0.00001062, 0.00000358, 0.00001480, -0.00002411, 0.00000603, -0.00000185, -0.00002184, 0.00001185, -0.00000880, 0.00000476, -0.00000504, -0.00002231, 0.00002270, -0.00002048, 0.00000313, 0.00001416, -0.00003067, 0.00003163, -0.00003266, 0.00001875, -0.00001655, -0.00001190, 0.00003089, -0.00008210, 0.00009046, -0.00015271, 0.00011004, -0.00018532, -0.00043724, 0.00108742, 0.00016235, -0.00058454, 0.00024889, -0.00019579, 0.00005837, -0.00001342, -0.00006788, 0.00009042, -0.00010801, 0.00008907, -0.00007085, 0.00002660, 0.00000280, -0.00004139, 0.00004820, -0.00003755, 0.00002830, -0.00003015, 0.00002556, -0.00002173, 0.00000439, 0.00001422, -0.00004745, 0.00004914, -0.00002912, -0.00000558, 0.00001149, -0.00001333, 0.00001360, 0.00000104, -0.00000734, -0.00000006, 0.00001940, -0.00001152, -0.00000009, -0.00000272, -0.00000249, 0.00000359, -0.00000369, 0.00001103, -0.00000436, 0.00000069, 0.00000005, 0.00000094, -0.00000189, -0.00000034, 0.00001375, -0.00000812, -0.00000190, 0.00000118, -0.00000644, 0.00001201, -0.00000428, -0.00000597, 0.00000612, -0.00000644, 0.00000333, -0.00000207, -0.00000587, 0.00000665, 0.00000242, 0.00000068, -0.00000421, -0.00000472, -0.00000625, -0.00000152, 0.00001115, -0.00000220, -0.00000356, 0.00000320, -0.00000778, 0.00000059, 0.00000335, -0.00000013, 0.00000301, -0.00000561, },
		{ 0.00000027, -0.00000394, 0.00000383, -0.00000063, -0.00000060, -0.00000017, 0.00000212, 0.00000300, 0.00000062, 0.00000225, -0.00000020, 0.00000352, -0.00000263, 0.00000070, 0.00000107, -0.00000665, 0.00000562, 0.00000106, 0.00000181, 0.00000911, 0.00000261, -0.00001189, 0.00000981, 0.00000702, -0.00002216, 0.00001102, -0.00001155, -0.00001461, 0.00001132, -0.00001108, 0.00000877, -0.00001268, -0.00001483, 0.00002051, -0.00002398, 0.00001175, 0.00000262, -0.00002053, 0.00002580, -0.00003246, 0.00002492, -0.00003094, 0.00001066, 0.00000166, -0.00005368, 0.00006967, -0.00015571, 0.00015076, -0.00033450, -0.00020868, 0.00113203, -0.00012430, -0.00043038, 0.00022821, -0.00022100, 0.00011140, -0.00007376, -0.00001012, 0.00004832, -0.00008544, 0.00008649, -0.00008541, 0.00005278, -0.00002800, -0.00001411, 0.00003092, -0.00002944, 0.00002585, -0.00003272, 0.00003453, -0.00003566, 0.00002300, -0.00000689, -0.00003211, 0.00004962, -0.00004310, 0.00001179, 0.00000025, -0.00000916, 0.00001537, -0.00000259, -0.00000673, 0.00000280, 0.00001773, -0.00001438, 0.00000280, -0.00000469, -0.00000072, 0.00000241, -0.00000203, 0.00001038, -0.00000545, 0.00000175, -0.00000079, 0.00000165, -0.00000366, 0.00000295, 0.00001208, -0.00001005, 0.00000104, -0.00000193, -0.00000325, 0.00001222, -0.00000838, -0.00000234, 0.00000430, -0.00000624, 0.00000495, -0.00000512, -0.00000327, 0.00000711, 0.00000141, 0.00000027, -0.00000511, -0.00000407, -0.00000684, 0.00000115, 0.00001080, -0.00000486, -0.00000106, 0.00000143, -0.00000829, 0.00000315, 0.00000203, 0.00000093, 0.00000187, -0.00000644, 0.00000314, },
		{ -0.00000211, 0.00000424, -0.00000183, -0.00000037, -0.00000040, 0.00000310, 0.00000240, 0.00000084, 0.00000250, -0.00000055, 0.00000332, -0.00000311, 0.00000161, -0.00000028, -0.00000564, 0.00000695, -0.00000056, 0.00000266, 0.00001079, -0.00000150, -0.00001052, 0.00001392, -0.00000123, -0.00001671, 0.00001159, -0.00001899, -0.00000604, 0.00000773, -0.00001045, 0.00001023, -0.00001887, -0.00000649, 0.00001541, -0.00002362, 0.00001659, -0.00000658, -0.00001020, 0.00001709, -0.00002691, 0.00002373, -0.00003718, 0.00002632, -0.00002494, -0.00002038, 0.00003346, -0.00013101, 0.00014992, -0.00044891, 0.00005750, 0.00107091, -0.00034878, -0.00025950, 0.00016534, -0.00020682, 0.00013396, -0.00011570, 0.00004196, 0.00000063, -0.00004974, 0.00006723, -0.00008356, 0.00006602, -0.00005157, 0.00001236, 0.00000937, -0.00001439, 0.00001525, -0.00002658, 0.00003548, -0.00004343, 0.00003824, -0.00002935, -0.00000923, 0.00003912, -0.00004824, 0.00002614, -0.00001330, -0.00000052, 0.00001332, -0.00000452, -0.00000617, 0.00000566, 0.00001513, -0.00001634, 0.00000569, -0.00000704, 0.00000142, 0.00000105, -0.00000021, 0.00000922, -0.00000602, 0.00000243, -0.00000128, 0.00000189, -0.00000493, 0.00000641, 0.00000919, -0.00001048, 0.00000293, -0.00000475, 0.00000099, 0.00001029, -0.00001106, 0.00000129, 0.00000164, -0.00000468, 0.00000518, -0.00000755, -0.00000008, 0.00000658, 0.00000083, -0.00000045, -0.00000561, -0.00000361, -0.00000724, 0.00000417, 0.00000919, -0.00000649, 0.00000133, -0.00000120, -0.00000747, 0.00000503, 0.00000072, 0.00000219, -0.00000009, -0.00000617, 0.00000461, -0.00000198, },
		{ 0.00000368, -0.00000265, -0.00000022, -0.00000044, 0.00000407, 0.00000151, 0.00000152, 0.00000236, -0.00000088, 0.00000320, -0.00000355, 0.00000247, -0.00000172, -0.00000409, 0.00000775, -0.00000249, 0.00000466, 0.00001131, -0.00000599, -0.00000683, 0.00001505, -0.00000816, -0.00000940, 0.00000794, -0.00002275, 0.00000189, 0.00000243, -0.00000740, 0.00000883, -0.00002250, 0.00000110, 0.00000895, -0.00002037, 0.00001769, -0.00001236, -0.00000178, 0.00000815, -0.00001865, 0.00001698, -0.00003535, 0.00003271, -0.00004360, 0.00000992, -0.00000917, -0.00008576, 0.00010584, -0.00050740, 0.00032710, 0.00091970, -0.00049406, -0.00010005, 0.00007979, -0.00016176, 0.00012609, -0.00013332, 0.00007915, -0.00004288, -0.00000905, 0.00003648, -0.00006701, 0.00006473, -0.00006378, 0.00003266, -0.00001124, 0.00000346, -0.00000071, -0.00001281, 0.00002768, -0.00004265, 0.00004623, -0.00004824, 0.00001681, 0.00001939, -0.00004289, 0.00003371, -0.00002555, 0.00001056, 0.00000777, -0.00000408, -0.00000642, 0.00000892, 0.00001133, -0.00001696, 0.00000813, -0.00000956, 0.00000396, -0.00000060, 0.00000180, 0.00000765, -0.00000613, 0.00000281, -0.00000147, 0.00000159, -0.00000542, 0.00000952, 0.00000557, -0.00000953, 0.00000345, -0.00000653, 0.00000544, 0.00000654, -0.00001184, 0.00000416, -0.00000120, -0.00000215, 0.00000392, -0.00000881, 0.00000308, 0.00000534, 0.00000067, -0.00000153, -0.00000558, -0.00000350, -0.00000714, 0.00000710, 0.00000657, -0.00000684, 0.00000303, -0.00000413, -0.00000546, 0.00000595, -0.00000021, 0.00000324, -0.00000247, -0.00000480, 0.00000507, -0.00000385, 0.00000034, },
		{ -0.00000295, -0.00000029, -0.00000012, 0.00000479, 0.00000053, 0.00000259, 0.00000170, -0.00000098, 0.00000302, -0.00000389, 0.00000325, -0.00000318, -0.00000200, 0.00000772, -0.00000428, 0.00000766, 0.00001012, -0.00000975, -0.00000170, 0.00001308, -0.00001237, -0.00000220, 0.00000126, -0.00002235, 0.00000760, -0.00000299, -0.00000301, 0.00000502, -0.00002318, 0.00000675, 0.00000264, -0.00001565, 0.00001598, -0.00001467, 0.00000360, 0.00000109, -0.00001044, 0.00000751, -0.00002747, 0.00002990, -0.00005148, 0.00003076, -0.00004842, -0.00003156, 0.00002556, -0.00049791, 0.00056662, 0.00070473, -0.00055649, 0.00002672, -0.00000838, -0.00009891, 0.00009372, -0.00012617, 0.00009618, -0.00007428, 0.00002823, 0.00000132, -0.00004037, 0.00005040, -0.00006317, 0.00004314, -0.00002625, 0.00001944, -0.00001798, 0.00000561, 0.00001250, -0.00003287, 0.00004453, -0.00005907, 0.00004063, -0.00000564, -0.00002779, 0.00003225, -0.00003305, 0.00002130, 0.00000010, -0.00000142, -0.00000780, 0.00001284, 0.00000613, -0.00001580, 0.00000954, -0.00001182, 0.00000675, -0.00000258, 0.00000406, 0.00000566, -0.00000583, 0.00000295, -0.00000146, 0.00000085, -0.00000500, 0.00001185, 0.00000181, -0.00000760, 0.00000261, -0.00000673, 0.00000914, 0.00000164, -0.00001070, 0.00000573, -0.00000351, 0.00000071, 0.00000143, -0.00000859, 0.00000565, 0.00000377, 0.00000076, -0.00000286, -0.00000505, -0.00000386, -0.00000631, 0.00000949, 0.00000335, -0.00000594, 0.00000361, -0.00000672, -0.00000264, 0.00000590, -0.00000051, 0.00000365, -0.00000476, -0.00000265, 0.00000437, -0.00000483, 0.00000287, 0.00000232, },
		{ -0.00000063, 0.00000065, 0.00000504, -0.00000025, 0.00000383, 0.00000051, -0.00000067, 0.00000262, -0.00000401, 0.00000386, -0.00000453, 0.00000055, 0.00000665, -0.00000534, 0.00001115, 0.00000708, -0.00001180, 0.00000360, 0.00000868, -0.00001320, 0.00000308, -0.00000666, -0.00001830, 0.00001022, -0.00000719, 0.00000144, -0.00000019, -0.00002122, 0.00000996, -0.00000239, -0.00001089, 0.00001280, -0.00001439, 0.00000592, -0.00000299, -0.00000448, -0.00000164, -0.00001671, 0.00002021, -0.00004862, 0.00003855, -0.00007600, 0.00001820, -0.00007659, -0.00041934, 0.00074852, 0.00045797, -0.00054463, 0.00010888, -0.00008241, -0.00003272, 0.00004672, -0.00009873, 0.00009236, -0.00008889, 0.00005525, -0.00003092, -0.00000988, 0.00002715, -0.00005110, 0.00004257, -0.00003262, 0.00002955, -0.00003216, 0.00002454, -0.00000690, -0.00001581, 0.00003286, -0.00005889, 0.00005706, -0.00003067, -0.00000598, 0.00002160, -0.00003340, 0.00002880, -0.00000763, 0.00000250, -0.00001007, 0.00001731, -0.00000043, -0.00001257, 0.00000931, -0.00001323, 0.00000941, -0.00000478, 0.00000655, 0.00000325, -0.00000507, 0.00000286, -0.00000139, -0.00000016, -0.00000373, 0.00001312, -0.00000154, -0.00000529, 0.00000077, -0.00000518, 0.00001128, -0.00000347, -0.00000799, 0.00000579, -0.00000473, 0.00000318, -0.00000175, -0.00000694, 0.00000721, 0.00000228, 0.00000086, -0.00000423, -0.00000416, -0.00000462, -0.00000466, 0.00001093, 0.00000011, -0.00000409, 0.00000290, -0.00000843, 0.00000047, 0.00000506, -0.00000019, 0.00000321, -0.00000644, -0.00000021, 0.00000267, -0.00000460, 0.00000494, 0.00000047, -0.00000278, },
		{ 0.00000184, 0.00000469, -0.00000054, 0.00000494, -0.00000103, 0.00000009, 0.00000187, -0.00000377, 0.00000415, -0.00000556, 0.00000330, 0.00000450, -0.00000516, 0.00001437, 0.00000251, -0.00001158, 0.00000777, 0.00000308, -0.00001087, 0.00000520, -0.00001390, -0.00001185, 0.00000974, -0.00000931, 0.00000477, -0.00000559, -0.00001748, 0.00001090, -0.00000566, -0.00000711, 0.00000954, -0.00001283, 0.00000604, -0.00000415, -0.00000188, -0.00000810, -0.00000648, 0.00000730, -0.00003769, 0.00003316, -0.00008704, 0.00005132, -0.00018183, -0.00028150, 0.00085510, 0.00021163, -0.00047602, 0.00014421, -0.00013130, 0.00002411, -0.00000359, -0.00005892, 0.00007121, -0.00008595, 0.00006792, -0.00005424, 0.00001797, 0.00000057, -0.00003126, 0.00003225, -0.00002950, 0.00003132, -0.00003960, 0.00003968, -0.00002630, 0.00000507, 0.00001321, -0.00004706, 0.00006231, -0.00005019, 0.00001781, 0.00000378, -0.00002585, 0.00003089, -0.00001317, 0.00000616, -0.00001241, 0.00002179, -0.00000794, -0.00000729, 0.00000701, -0.00001315, 0.00001139, -0.00000689, 0.00000917, 0.00000042, -0.00000383, 0.00000250, -0.00000130, -0.00000122, -0.00000182, 0.00001328, -0.00000410, -0.00000318, -0.00000147, -0.00000214, 0.00001134, -0.00000784, -0.00000440, 0.00000452, -0.00000461, 0.00000464, -0.00000491, -0.00000419, 0.00000763, 0.00000119, 0.00000072, -0.00000540, -0.00000314, -0.00000558, -0.00000227, 0.00001117, -0.00000259, -0.00000181, 0.00000103, -0.00000888, 0.00000332, 0.00000380, 0.00000057, 0.00000191, -0.00000714, 0.00000192, 0.00000042, -0.00000314, 0.00000610, -0.00000143, -0.00000232, -0.00000111, },
		{ 0.00000376, -0.00000017, 0.00000556, -0.00000265, 0.00000122, 0.00000073, -0.00000308, 0.00000398, -0.00000603, 0.00000587, 0.00000151, -0.00000340, 0.00001646, -0.00000280, -0.00000907, 0.00000984, -0.00000223, -0.00000637, 0.00000378, -0.00001884, -0.00000464, 0.00000685, -0.00000921, 0.00000629, -0.00001010, -0.00001307, 0.00001024, -0.00000732, -0.00000471, 0.00000713, -0.00001130, 0.00000533, -0.00000340, -0.00000248, -0.00001067, 0.00000054, -0.00000489, -0.00002304, 0.00001774, -0.00008104, 0.00005951, -0.00027053, -0.00010300, 0.00088044, -0.00000679, -0.00037297, 0.00013885, -0.00015097, 0.00006281, -0.00004700, -0.00001600, 0.00003920, -0.00006829, 0.00006555, -0.00006510, 0.00003791, -0.00002357, -0.00000865, 0.00001553, -0.00001830, 0.00002440, -0.00003824, 0.00004754, -0.00004141, 0.00002527, -0.00001048, -0.00002550, 0.00005492, -0.00005973, 0.00003825, -0.00001740, -0.00001160, 0.00002667, -0.00001478, 0.00000793, -0.00001363, 0.00002533, -0.00001552, -0.00000045, 0.00000257, -0.00001108, 0.00001206, -0.00000842, 0.00001162, -0.00000267, -0.00000209, 0.00000179, -0.00000121, -0.00000215, 0.00000046, 0.00001246, -0.00000568, -0.00000177, -0.00000337, 0.00000179, 0.00000925, -0.00001068, -0.00000076, 0.00000240, -0.00000323, 0.00000471, -0.00000736, -0.00000089, 0.00000702, 0.00000065, 0.00000016, -0.00000612, -0.00000227, -0.00000646, 0.00000056, 0.00001019, -0.00000432, 0.00000029, -0.00000159, -0.00000799, 0.00000550, 0.00000248, 0.00000142, -0.00000001, -0.00000672, 0.00000323, -0.00000179, -0.00000072, 0.00000612, -0.00000301, -0.00000181, -0.00000156, 0.00000328, },
		{ 0.00000090, 0.00000545, -0.00000400, 0.00000248, -0.00000070, -0.00000194, 0.00000329, -0.00000572, 0.00000781, -0.00000183, -0.00000011, 0.00001672, -0.00000782, -0.00000483, 0.00000937, -0.00000589, -0.00000119, -0.00000067, -0.00002051, 0.00000173, 0.00000270, -0.00000731, 0.00000586, -0.00001301, -0.00000897, 0.00000888, -0.00000799, -0.00000346, 0.00000587, -0.00001069, 0.00000510, -0.00000227, -0.00000510, -0.00000957, 0.00000305, -0.00001323, -0.00000939, -0.00000240, -0.00006188, 0.00004015, -0.00032601, 0.00009226, 0.00083014, -0.00017795, -0.00025816, 0.00010450, -0.00014369, 0.00007968, -0.00007624, 0.00002148, 0.00000410, -0.00004136, 0.00005056, -0.00006293, 0.00004685, -0.00004059, 0.00001158, -0.00000308, -0.00000225, 0.00001050, -0.00002807, 0.00004630, -0.00004891, 0.00004036, -0.00003319, 0.00000168, 0.00003615, -0.00005693, 0.00005058, -0.00003721, 0.00000650, 0.00001681, -0.00001170, 0.00000656, -0.00001254, 0.00002676, -0.00002198, 0.00000699, -0.00000362, -0.00000687, 0.00001091, -0.00000879, 0.00001345, -0.00000576, 0.00000006, 0.00000068, -0.00000103, -0.00000286, 0.00000281, 0.00001095, -0.00000636, -0.00000127, -0.00000428, 0.00000573, 0.00000538, -0.00001158, 0.00000218, 0.00000011, -0.00000101, 0.00000334, -0.00000857, 0.00000230, 0.00000568, 0.00000061, -0.00000085, -0.00000626, -0.00000177, -0.00000695, 0.00000343, 0.00000823, -0.00000488, 0.00000167, -0.00000439, -0.00000594, 0.00000676, 0.00000145, 0.00000199, -0.00000214, -0.00000535, 0.00000344, -0.00000338, 0.00000218, 0.00000504, -0.00000400, -0.00000150, -0.00000172, 0.00000468, 0.00000244, },
		{ 0.00000447, -0.00000477, 0.00000359, -0.00000223, -0.00000046, 0.00000212, -0.00000457, 0.00000869, -0.00000488, 0.00000432, 0.00001480, -0.00001151, 0.00000013, 0.00000660, -0.00000705, 0.00000305, -0.00000694, -0.00001881, 0.00000608, -0.00000146, -0.00000448, 0.00000388, -0.00001418, -0.00000580, 0.00000762, -0.00000845, -0.00000267, 0.00000544, -0.00001118, 0.00000620, -0.00000219, -0.00000800, -0.00000616, 0.00000129, -0.00001619, -0.00000060, -0.00002129, -0.00003661, -0.00000324, -0.00033750, 0.00027939, 0.00071921, -0.00029238, -0.00015083, 0.00005527, -0.00011644, 0.00007599, -0.00008809, 0.00004734, -0.00002682, -0.00001173, 0.00002767, -0.00004992, 0.00004436, -0.00004783, 0.00002534, -0.00001906, 0.00001448, -0.00000704, -0.00001109, 0.00003614, -0.00004719, 0.00004704, -0.00004999, 0.00002900, 0.00000975, -0.00004210, 0.00005173, -0.00005108, 0.00002457, 0.00000346, -0.00000443, 0.00000160, -0.00000834, 0.00002509, -0.00002602, 0.00001372, -0.00001064, -0.00000082, 0.00000774, -0.00000753, 0.00001413, -0.00000839, 0.00000238, -0.00000082, -0.00000069, -0.00000333, 0.00000502, 0.00000906, -0.00000642, -0.00000164, -0.00000379, 0.00000878, 0.00000047, -0.00001051, 0.00000388, -0.00000168, 0.00000140, 0.00000086, -0.00000831, 0.00000483, 0.00000407, 0.00000087, -0.00000222, -0.00000582, -0.00000175, -0.00000679, 0.00000589, 0.00000572, -0.00000434, 0.00000196, -0.00000678, -0.00000312, 0.00000710, 0.00000084, 0.00000200, -0.00000398, -0.00000341, 0.00000254, -0.00000386, 0.00000494, 0.00000310, -0.00000433, -0.00000153, -0.00000141, 0.00000575, 0.00000102, 0.00000251, },
		{ -0.00000476, 0.00000425, -0.00000358, 0.00000114, 0.00000065, -0.00000265, 0.00000827, -0.00000693, 0.00000914, 0.00001081, -0.00001309, 0.00000464, 0.00000237, -0.00000552, 0.00000504, -0.00001344, -0.00001439, 0.00000788, -0.00000463, -0.00000166, 0.00000109, -0.00001394, -0.00000370, 0.00000688, -0.00000935, -0.00000154, 0.00000504, -0.00001228, 0.00000871, -0.00000407, -0.00000953, -0.00000241, -0.00000312, -0.00001407, 0.00000139, -0.00003395, -0.00001366, -0.00006178, -0.00030195, 0.00043656, 0.00056858, -0.00035021, -0.00006435, 0.00000447, -0.00007865, 0.00005681, -0.00008335, 0.00005871, -0.00004827, 0.00001441, 0.00000254, -0.00003021, 0.00003248, -0.00004500, 0.00003033, -0.00002882, 0.00002779, -0.00002411, 0.00000923, 0.00001924, -0.00003673, 0.00004386, -0.00005714, 0.00005084, -0.00001880, -0.00001819, 0.00004106, -0.00005568, 0.00003860, -0.00001031, 0.00000534, -0.00000636, -0.00000096, 0.00001975, -0.00002653, 0.00001830, -0.00001722, 0.00000626, 0.00000274, -0.00000443, 0.00001321, -0.00001006, 0.00000450, -0.00000256, -0.00000013, -0.00000364, 0.00000699, 0.00000708, -0.00000621, -0.00000254, -0.00000186, 0.00001020, -0.00000451, -0.00000791, 0.00000415, -0.00000246, 0.00000331, -0.00000216, -0.00000666, 0.00000634, 0.00000259, 0.00000116, -0.00000371, -0.00000492, -0.00000219, -0.00000591, 0.00000757, 0.00000317, -0.00000305, 0.00000110, -0.00000824, -0.00000004, 0.00000671, 0.00000063, 0.00000134, -0.00000512, -0.00000144, 0.00000083, -0.00000303, 0.00000700, 0.00000073, -0.00000409, -0.00000189, -0.00000053, 0.00000622, -0.00000012, 0.00000435, 0.00000269, },
		{ 0.00000427, -0.00000448, 0.00000259, -0.00000083, -0.00000025, 0.00000655, -0.00000742, 0.00001345, 0.00000531, -0.00001226, 0.00000764, -0.00000212, -0.00000185, 0.00000410, -0.00001860, -0.00000844, 0.00000727, -0.00000626, 0.00000038, -0.00000168, -0.00001291, -0.00000240, 0.00000665, -0.00001093, 0.00000052, 0.00000378, -0.00001303, 0.00001196, -0.00000797, -0.00000864, -0.00000024, -0.00000788, -0.00000880, -0.00000312, -0.00003768, -0.00000065, -0.00012328, -0.00022414, 0.00054821, 0.00040103, -0.00035936, -0.00000519, -0.00003769, -0.00003973, 0.00002933, -0.00006601, 0.00005608, -0.00005768, 0.00003252, -0.00001949, -0.00000883, 0.00001498, -0.00003404, 0.00002645, -0.00003054, 0.00003461, -0.00003682, 0.00002874, -0.00000089, -0.00001992, 0.00003162, -0.00005297, 0.00006272, -0.00004352, 0.00000980, 0.00002051, -0.00004967, 0.00004543, -0.00002119, 0.00001510, -0.00001578, 0.00000880, 0.00001093, -0.00002296, 0.00001954, -0.00002191, 0.00001316, -0.00000337, 0.00000037, 0.00001047, -0.00001032, 0.00000593, -0.00000427, 0.00000061, -0.00000383, 0.00000871, 0.00000515, -0.00000608, -0.00000345, 0.00000120, 0.00000956, -0.00000862, -0.00000446, 0.00000321, -0.00000204, 0.00000416, -0.00000501, -0.00000403, 0.00000671, 0.00000156, 0.00000120, -0.00000505, -0.00000381, -0.00000291, -0.00000437, 0.00000827, 0.00000108, -0.00000150, -0.00000074, -0.00000849, 0.00000281, 0.00000588, 0.00000062, 0.00000013, -0.00000534, 0.00000001, -0.00000118, -0.00000098, 0.00000795, -0.00000161, -0.00000352, -0.00000245, 0.00000084, 0.00000598, -0.00000065, 0.00000601, 0.00000037, -0.00000395, },
		{ -0.00000475, 0.00000364, -0.00000199, 0.00000219, 0.00000381, -0.00000609, 0.00001633, -0.00000074, -0.00000925, 0.00000844, -0.00000561, 0.00000284, 0.00000029, -0.00002129, -0.00000235, 0.00000493, -0.00000635, 0.00000124, -0.00000379, -0.00001180, -0.00000133, 0.00000651, -0.00001298, 0.00000368, 0.00000102, -0.00001239, 0.00001478, -0.00001313, -0.00000520, -0.00000085, -0.00001074, -0.00000317, -0.00001180, -0.00003271, -0.00000255, -0.00017476, -0.00011543, 0.00060687, 0.00023745, -0.00033266, 0.00002655, -0.00006538, -0.00000717, 0.00000089, -0.00004190, 0.00004270, -0.00005539, 0.00004042, -0.00003453, 0.00000957, -0.00000365, -0.00001841, 0.00001557, -0.00002434, 0.00003352, -0.00004234, 0.00004364, -0.00002015, -0.00000046, 0.00001313, -0.00003827, 0.00006225, -0.00005922, 0.00003594, -0.00000577, -0.00003406, 0.00004353, -0.00002647, 0.00002215, -0.00002450, 0.00001935, -0.00000035, -0.00001552, 0.00001677, -0.00002347, 0.00001851, -0.00000951, 0.00000625, 0.00000602, -0.00000891, 0.00000623, -0.00000554, 0.00000136, -0.00000394, 0.00001022, 0.00000330, -0.00000622, -0.00000384, 0.00000476, 0.00000688, -0.00001112, -0.00000101, 0.00000154, -0.00000056, 0.00000367, -0.00000703, -0.00000101, 0.00000612, 0.00000109, 0.00000081, -0.00000602, -0.00000273, -0.00000365, -0.00000244, 0.00000802, -0.00000021, -0.00000024, -0.00000311, -0.00000751, 0.00000508, 0.00000492, 0.00000056, -0.00000135, -0.00000465, 0.00000054, -0.00000287, 0.00000192, 0.00000761, -0.00000350, -0.00000290, -0.00000298, 0.00000250, 0.00000506, -0.00000039, 0.00000708, -0.00000210, -0.00000248, 0.00000359, },
		{ 0.00000413, -0.00000256, 0.00000418, 0.00000061, -0.00000306, 0.00001712, -0.00000625, -0.00000478, 0.00000695, -0.00000710, 0.00000717, -0.00000562, -0.00002098, 0.00000264, 0.00000182, -0.00000530, 0.00000097, -0.00000493, -0.00001112, 0.00000007, 0.00000584, -0.00001484, 0.00000753, -0.00000335, -0.00000959, 0.00001587, -0.00001817, -0.00000006, -0.00000442, -0.00001025, 0.00000002, -0.00002110, -0.00002209, -0.00002042, -0.00020522, 0.00000872, 0.00061327, 0.00009391, -0.00028464, 0.00003559, -0.00007726, 0.00001460, -0.00002252, -0.00001707, 0.00002340, -0.00004406, 0.00003839, -0.00004075, 0.00002164, -0.00001928, -0.00000220, 0.00000094, -0.00001217, 0.00002499, -0.00003953, 0.00005122, -0.00003488, 0.00001751, -0.00000742, -0.00001613, 0.00004968, -0.00006267, 0.00005466, -0.00003240, -0.00001197, 0.00003331, -0.00002475, 0.00002427, -0.00003022, 0.00002861, -0.00001237, -0.00000526, 0.00001011, -0.00002117, 0.00002105, -0.00001441, 0.00001224, 0.00000040, -0.00000587, 0.00000513, -0.00000597, 0.00000184, -0.00000387, 0.00001157, 0.00000145, -0.00000667, -0.00000327, 0.00000802, 0.00000260, -0.00001162, 0.00000171, -0.00000020, 0.00000153, 0.00000187, -0.00000776, 0.00000179, 0.00000492, 0.00000111, -0.00000005, -0.00000645, -0.00000190, -0.00000415, -0.00000047, 0.00000704, -0.00000059, 0.00000029, -0.00000548, -0.00000551, 0.00000658, 0.00000402, 0.00000018, -0.00000268, -0.00000335, 0.00000003, -0.00000367, 0.00000508, 0.00000609, -0.00000471, -0.00000249, -0.00000322, 0.00000415, 0.00000369, 0.00000068, 0.00000725, -0.00000429, -0.00000067, 0.00000232, -0.00000433, },
		{ -0.00000242, 0.00000528, -0.00000234, 0.00000118, 0.00001557, -0.00001024, 0.00000009, 0.00000365, -0.00000614, 0.00000981, -0.00001236, -0.00001788, 0.00000573, -0.00000112, -0.00000379, -0.00000004, -0.00000520, -0.00001105, 0.00000213, 0.00000408, -0.00001572, 0.00001124, -0.00000884, -0.00000449, 0.00001428, -0.00002157, 0.00000528, -0.00001000, -0.00000619, -0.00000109, -0.00002742, -0.00001073, -0.00005116, -0.00020773, 0.00013179, 0.00057507, -0.00001987, -0.00022871, 0.00002926, -0.00007563, 0.00002451, -0.00003737, 0.00000350, 0.00000327, -0.00002776, 0.00002877, -0.00003851, 0.00002589, -0.00002898, 0.00001088, -0.00001371, 0.00000288, 0.00001110, -0.00002902, 0.00005044, -0.00004261, 0.00003046, -0.00002552, 0.00000880, 0.00002788, -0.00005330, 0.00006202, -0.00005392, 0.00001207, 0.00001710, -0.00001626, 0.00002032, -0.00003112, 0.00003450, -0.00002300, 0.00000604, 0.00000052, -0.00001501, 0.00001999, -0.00001687, 0.00001716, -0.00000552, -0.00000164, 0.00000260, -0.00000529, 0.00000172, -0.00000344, 0.00001272, -0.00000054, -0.00000722, -0.00000155, 0.00001018, -0.00000249, -0.00001018, 0.00000319, -0.00000137, 0.00000357, -0.00000083, -0.00000706, 0.00000384, 0.00000354, 0.00000141, -0.00000125, -0.00000636, -0.00000142, -0.00000424, 0.00000119, 0.00000573, -0.00000023, -0.00000015, -0.00000729, -0.00000289, 0.00000733, 0.00000326, -0.00000063, -0.00000348, -0.00000189, -0.00000137, -0.00000322, 0.00000789, 0.00000371, -0.00000517, -0.00000242, -0.00000298, 0.00000546, 0.00000220, 0.00000239, 0.00000640, -0.00000583, 0.00000113, 0.00000071, -0.00000333, 0.00000406, },
		{ 0.00000526, -0.00000435, 0.00000581, 0.00001190, -0.00001205, 0.00000426, -0.00000054, -0.00000291, 0.00000987, -0.00001850, -0.00001277, 0.00000665, -0.00000319, -0.00000247, -0.00000121, -0.00000502, -0.00001139, 0.00000483, 0.00000096, -0.00001490, 0.00001379, -0.00001442, 0.00000239, 0.00000970, -0.00002217, 0.00000912, -0.00001588, 0.00000034, -0.00000688, -0.00002818, -0.00000406, -0.00008821, -0.00018066, 0.00023910, 0.00050453, -0.00010067, -0.00017514, 0.00001535, -0.00006500, 0.00002427, -0.00004274, 0.00001679, -0.00001345, -0.00001078, 0.00001502, -0.00002990, 0.00002277, -0.00003155, 0.00001825, -0.00002502, 0.00001732, -0.00000501, -0.00001300, 0.00004200, -0.00004252, 0.00003616, -0.00003732, 0.00003134, 0.00000158, -0.00003328, 0.00005648, -0.00006598, 0.00003318, -0.00000144, -0.00000288, 0.00001058, -0.00002635, 0.00003544, -0.00003016, 0.00001624, -0.00001031, -0.00000581, 0.00001524, -0.00001610, 0.00001988, -0.00001065, 0.00000301, -0.00000104, -0.00000342, 0.00000077, -0.00000239, 0.00001351, -0.00000277, -0.00000757, 0.00000121, 0.00001062, -0.00000747, -0.00000724, 0.00000329, -0.00000152, 0.00000491, -0.00000381, -0.00000514, 0.00000486, 0.00000239, 0.00000170, -0.00000260, -0.00000585, -0.00000125, -0.00000385, 0.00000227, 0.00000454, 0.00000053, -0.00000153, -0.00000816, -0.00000010, 0.00000751, 0.00000255, -0.00000181, -0.00000349, -0.00000081, -0.00000325, -0.00000145, 0.00000979, 0.00000092, -0.00000500, -0.00000269, -0.00000219, 0.00000618, 0.00000098, 0.00000436, 0.00000462, -0.00000647, 0.00000255, -0.00000093, -0.00000196, 0.00000355, -0.00000162, },
		{ -0.00000491, 0.00000990, 0.00000677, -0.00001150, 0.00000684, -0.00000446, 0.00000182, 0.00000706, -0.00002278, -0.00000679, 0.00000570, -0.00000411, -0.00000181, -0.00000199, -0.00000492, -0.00001166, 0.00000779, -0.00000337, -0.00001203, 0.00001425, -0.00001885, 0.00000989, 0.00000260, -0.00001946, 0.00001009, -0.00002011, 0.00000732, -0.00001611, -0.00002268, -0.00000650, -0.00012313, -0.00012772, 0.00032012, 0.00041566, -0.00015075, -0.00013013, 0.00000033, -0.00005047, 0.00001734, -0.00004005, 0.00002202, -0.00002417, 0.00000328, 0.00000086, -0.00001803, 0.00001425, -0.00002753, 0.00001890, -0.00003074, 0.00002810, -0.00001995, 0.00000538, 0.00002805, -0.00003550, 0.00003421, -0.00004054, 0.00004701, -0.00002363, -0.00000704, 0.00003927, -0.00006634, 0.00004721, -0.00001814, 0.00001231, -0.00000326, -0.00001631, 0.00003079, -0.00003234, 0.00002323, -0.00002028, 0.00000485, 0.00000747, -0.00001191, 0.00001963, -0.00001394, 0.00000710, -0.00000512, -0.00000059, -0.00000111, -0.00000049, 0.00001372, -0.00000519, -0.00000734, 0.00000460, 0.00000908, -0.00001142, -0.00000352, 0.00000223, -0.00000053, 0.00000505, -0.00000635, -0.00000244, 0.00000483, 0.00000172, 0.00000173, -0.00000383, -0.00000512, -0.00000125, -0.00000312, 0.00000268, 0.00000383, 0.00000120, -0.00000358, -0.00000791, 0.00000247, 0.00000731, 0.00000172, -0.00000313, -0.00000268, -0.00000053, -0.00000502, 0.00000144, 0.00001043, -0.00000179, -0.00000444, -0.00000318, -0.00000090, 0.00000620, 0.00000036, 0.00000616, 0.00000217, -0.00000618, 0.00000334, -0.00000229, -0.00000050, 0.00000285, -0.00000040, 0.00000413, },
		{ 0.00001259, 0.00000116, -0.00000893, 0.00000739, -0.00000705, 0.00000686, 0.00000178, -0.00002442, -0.00000115, 0.00000355, -0.00000399, -0.00000192, -0.00000206, -0.00000540, -0.00001130, 0.00001032, -0.00000831, -0.00000729, 0.00001211, -0.00002096, 0.00001654, -0.00000582, -0.00001385, 0.00000760, -0.00002109, 0.00001235, -0.00002629, -0.00001236, -0.00002015, -0.00014744, -0.00005692, 0.00036969, 0.00032166, -0.00017595, -0.00009587, -0.00001163, -0.00003638, 0.00000770, -0.00003214, 0.00002034, -0.00002816, 0.00001223, -0.00001067, -0.00000613, 0.00000319, -0.00001883, 0.00001345, -0.00003005, 0.00003323, -0.00003080, 0.00002281, 0.00001156, -0.00002376, 0.00002594, -0.00003492, 0.00005295, -0.00004260, 0.00001982, 0.00001403, -0.00005522, 0.00005164, -0.00002934, 0.00002574, -0.00001850, -0.00000261, 0.00002107, -0.00002895, 0.00002548, -0.00002733, 0.00001500, -0.00000193, -0.00000486, 0.00001619, -0.00001461, 0.00000965, -0.00000880, 0.00000266, -0.00000374, 0.00000237, 0.00001305, -0.00000764, -0.00000623, 0.00000801, 0.00000565, -0.00001365, 0.00000015, 0.00000052, 0.00000142, 0.00000378, -0.00000784, 0.00000039, 0.00000398, 0.00000160, 0.00000135, -0.00000476, -0.00000437, -0.00000121, -0.00000228, 0.00000253, 0.00000377, 0.00000133, -0.00000582, -0.00000660, 0.00000454, 0.00000693, 0.00000058, -0.00000421, -0.00000128, -0.00000128, -0.00000608, 0.00000497, 0.00000970, -0.00000402, -0.00000376, -0.00000366, 0.00000066, 0.00000558, 0.00000052, 0.00000739, -0.00000051, -0.00000511, 0.00000342, -0.00000314, 0.00000079, 0.00000224, 0.00000086, 0.00000220, -0.00000381, },
		{ -0.00000392, -0.00000510, 0.00000596, -0.00000761, 0.00001097, -0.00000505, -0.00002321, 0.00000321, 0.00000099, -0.00000325, -0.00000263, -0.00000140, -0.00000667, -0.00000986, 0.00001173, -0.00001297, -0.00000136, 0.00000746, -0.00002002, 0.00002096, -0.00001400, -0.00000650, 0.00000203, -0.00001812, 0.00001345, -0.00003440, -0.00000052, -0.00004414, -0.00015455, 0.00002128, 0.00038809, 0.00023282, -0.00018366, -0.00007148, -0.00001888, -0.00002550, -0.00000126, -0.00002232, 0.00001413, -0.00002636, 0.00001548, -0.00001773, 0.00000315, -0.00000747, -0.00000813, 0.00000374, -0.00002362, 0.00003210, -0.00003570, 0.00003648, -0.00000438, -0.00001021, 0.00001400, -0.00002218, 0.00004853, -0.00005174, 0.00004170, -0.00001405, -0.00003524, 0.00004610, -0.00003267, 0.00003420, -0.00003197, 0.00001225, 0.00000791, -0.00002054, 0.00002235, -0.00002987, 0.00002262, -0.00001112, 0.00000381, 0.00001000, -0.00001241, 0.00000992, -0.00001116, 0.00000560, -0.00000668, 0.00000609, 0.00001126, -0.00000979, -0.00000411, 0.00001075, 0.00000083, -0.00001383, 0.00000303, -0.00000119, 0.00000381, 0.00000126, -0.00000789, 0.00000274, 0.00000273, 0.00000189, 0.00000056, -0.00000528, -0.00000374, -0.00000097, -0.00000164, 0.00000213, 0.00000434, 0.00000061, -0.00000774, -0.00000450, 0.00000604, 0.00000643, -0.00000096, -0.00000470, 0.00000028, -0.00000301, -0.00000594, 0.00000851, 0.00000776, -0.00000554, -0.00000319, -0.00000392, 0.00000221, 0.00000454, 0.00000144, 0.00000772, -0.00000300, -0.00000356, 0.00000287, -0.00000341, 0.00000176, 0.00000195, 0.00000172, 0.00000002, -0.00000126, 0.00001331, },
		{ -0.00000097, 0.00000313, -0.00000597, 0.00001308, -0.00001217, -0.00001956, 0.00000575, -0.00000125, -0.00000240, -0.00000352, -0.00000031, -0.00000866, -0.00000719, 0.00001146, -0.00001640, 0.00000471, 0.00000099, -0.00001597, 0.00002215, -0.00002034, 0.00000095, -0.00000539, -0.00001159, 0.00000964, -0.00003781, 0.00000859, -0.00007468, -0.00014118, 0.00009613, 0.00038006, 0.00015557, -0.00018093, -0.00005433, -0.00002190, -0.00001876, -0.00000739, -0.00001342, 0.00000620, -0.00002077, 0.00001384, -0.00001989, 0.00000828, -0.00001540, 0.00000182, -0.00000769, -0.00001322, 0.00002546, -0.00003413, 0.00004460, -0.00001727, 0.00000234, 0.00000162, -0.00000553, 0.00003539, -0.00004971, 0.00005426, -0.00003940, -0.00001072, 0.00003239, -0.00002759, 0.00003557, -0.00004084, 0.00002541, -0.00000627, -0.00000873, 0.00001432, -0.00002721, 0.00002614, -0.00001821, 0.00001245, 0.00000214, -0.00000770, 0.00000763, -0.00001152, 0.00000741, -0.00000927, 0.00001032, 0.00000827, -0.00001125, -0.00000109, 0.00001221, -0.00000465, -0.00001201, 0.00000463, -0.00000227, 0.00000598, -0.00000204, -0.00000647, 0.00000414, 0.00000153, 0.00000234, -0.00000050, -0.00000542, -0.00000327, -0.00000045, -0.00000144, 0.00000187, 0.00000523, -0.00000103, -0.00000889, -0.00000201, 0.00000702, 0.00000576, -0.00000284, -0.00000431, 0.00000147, -0.00000537, -0.00000439, 0.00001144, 0.00000497, -0.00000630, -0.00000286, -0.00000380, 0.00000349, 0.00000341, 0.00000294, 0.00000707, -0.00000490, -0.00000189, 0.00000191, -0.00000314, 0.00000237, 0.00000206, 0.00000187, -0.00000179, 0.00000222, 0.00001193, -0.00000758, },
		{ -0.00000022, -0.00000251, 0.00001261, -0.00001832, -0.00001431, 0.00000640, -0.00000272, -0.00000190, -0.00000412, 0.00000075, -0.00001097, -0.00000349, 0.00000932, -0.00001785, 0.00000975, -0.00000617, -0.00000940, 0.00001984, -0.00002370, 0.00000683, -0.00001285, -0.00000295, 0.00000130, -0.00003509, 0.00001087, -0.00010580, -0.00010798, 0.00015863, 0.00035309, 0.00009235, -0.00017313, -0.00004139, -0.00002235, -0.00001551, -0.00001007, -0.00000724, -0.00000101, -0.00001375, 0.00000905, -0.00001793, 0.00000900, -0.00001928, 0.00000886, -0.00001831, -0.00000123, 0.00001508, -0.00002681, 0.00004663, -0.00002560, 0.00001177, -0.00000821, 0.00001117, 0.00001696, -0.00003768, 0.00005534, -0.00005725, 0.00001331, 0.00001398, -0.00001546, 0.00002929, -0.00004318, 0.00003432, -0.00001877, 0.00000411, 0.00000292, -0.00001973, 0.00002479, -0.00002168, 0.00001930, -0.00000586, -0.00000143, 0.00000307, -0.00000957, 0.00000744, -0.00001075, 0.00001445, 0.00000422, -0.00001164, 0.00000250, 0.00001197, -0.00000991, -0.00000862, 0.00000479, -0.00000230, 0.00000729, -0.00000542, -0.00000388, 0.00000439, 0.00000075, 0.00000266, -0.00000160, -0.00000531, -0.00000285, 0.00000028, -0.00000181, 0.00000210, 0.00000603, -0.00000341, -0.00000901, 0.00000050, 0.00000760, 0.00000478, -0.00000481, -0.00000298, 0.00000181, -0.00000781, -0.00000149, 0.00001328, 0.00000180, -0.00000639, -0.00000279, -0.00000327, 0.00000429, 0.00000253, 0.00000466, 0.00000550, -0.00000600, -0.00000043, 0.00000081, -0.00000252, 0.00000271, 0.00000248, 0.00000117, -0.00000264, 0.00000582, 0.00000873, -0.00000932, 0.00000253, },
		{ 0.00000195, 0.00000955, -0.00002251, -0.00000851, 0.00000551, -0.00000329, -0.00000200, -0.00000406, 0.00000126, -0.00001303, 0.00000073, 0.00000553, -0.00001697, 0.00001276, -0.00001266, -0.00000149, 0.00001447, -0.00002361, 0.00000990, -0.00001852, 0.00000571, -0.00000990, -0.00002652, 0.00000352, -0.00013070, -0.00005932, 0.00020302, 0.00031551, 0.00004239, -0.00016341, -0.00003029, -0.00002223, -0.00001422, -0.00000990, -0.00000430, -0.00000593, -0.00000737, 0.00000312, -0.00001343, 0.00000609, -0.00001893, 0.00001178, -0.00002608, 0.00001000, 0.00000323, -0.00001540, 0.00004317, -0.00002905, 0.00001714, -0.00001350, 0.00002436, -0.00000244, -0.00001890, 0.00004525, -0.00006466, 0.00003226, -0.00000484, 0.00000079, 0.00001650, -0.00003850, 0.00003732, -0.00002723, 0.00001537, -0.00000957, -0.00000881, 0.00001880, -0.00002072, 0.00002290, -0.00001239, 0.00000504, -0.00000295, -0.00000550, 0.00000538, -0.00001046, 0.00001774, -0.00000050, -0.00001073, 0.00000616, 0.00000991, -0.00001413, -0.00000436, 0.00000369, -0.00000114, 0.00000725, -0.00000815, -0.00000066, 0.00000356, 0.00000061, 0.00000262, -0.00000253, -0.00000515, -0.00000232, 0.00000100, -0.00000267, 0.00000306, 0.00000624, -0.00000612, -0.00000804, 0.00000272, 0.00000794, 0.00000333, -0.00000648, -0.00000089, 0.00000096, -0.00000969, 0.00000240, 0.00001372, -0.00000130, -0.00000601, -0.00000287, -0.00000241, 0.00000454, 0.00000216, 0.00000620, 0.00000331, -0.00000625, 0.00000057, -0.00000016, -0.00000175, 0.00000298, 0.00000299, -0.00000026, -0.00000217, 0.00000872, 0.00000438, -0.00000908, 0.00000471, -0.00000312, },
		{ 0.00000443, -0.00002420, -0.00000315, 0.00000369, -0.00000313, -0.00000266, -0.00000321, 0.00000086, -0.00001428, 0.00000473, 0.00000074, -0.00001389, 0.00001319, -0.00001729, 0.00000634, 0.00000714, -0.00002040, 0.00000962, -0.00002104, 0.00001219, -0.00002141, -0.00001414, -0.00001405, -0.00014324, -0.00000222, 0.00022741, 0.00027471, 0.00000298, -0.00015285, -0.00001985, -0.00002298, -0.00001319, -0.00000828, -0.00000406, -0.00000807, -0.00000298, -0.00000216, -0.00000817, 0.00000106, -0.00001519, 0.00001044, -0.00002984, 0.00001858, -0.00000783, -0.00000202, 0.00003564, -0.00002827, 0.00001864, -0.00001360, 0.00003160, -0.00001863, 0.00000213, 0.00002661, -0.00006101, 0.00004292, -0.00001994, 0.00001743, -0.00000028, -0.00002773, 0.00003399, -0.00003014, 0.00002277, -0.00002063, 0.00000342, 0.00000941, -0.00001547, 0.00002247, -0.00001610, 0.00001020, -0.00000915, -0.00000005, 0.00000135, -0.00000803, 0.00001948, -0.00000525, -0.00000848, 0.00000934, 0.00000622, -0.00001669, -0.00000000, 0.00000181, 0.00000101, 0.00000569, -0.00000960, 0.00000247, 0.00000199, 0.00000108, 0.00000210, -0.00000312, -0.00000508, -0.00000150, 0.00000140, -0.00000374, 0.00000476, 0.00000550, -0.00000864, -0.00000618, 0.00000450, 0.00000810, 0.00000130, -0.00000745, 0.00000151, -0.00000111, -0.00001042, 0.00000671, 0.00001277, -0.00000398, -0.00000538, -0.00000297, -0.00000138, 0.00000436, 0.00000244, 0.00000719, 0.00000088, -0.00000578, 0.00000105, -0.00000082, -0.00000102, 0.00000337, 0.00000323, -0.00000205, -0.00000033, 0.00001022, -0.00000022, -0.00000722, 0.00000532, -0.00000479, 0.00000633, },
		{ -0.00002339, 0.00000104, 0.00000158, -0.00000261, -0.00000364, -0.00000169, -0.00000056, -0.00001430, 0.00000780, -0.00000420, -0.00000925, 0.00001101, -0.00001928, 0.00001273, -0.00000075, -0.00001505, 0.00000634, -0.00001984, 0.00001483, -0.00003049, -0.00000129, -0.00003995, -0.00013934, 0.00005509, 0.00023355, 0.00023591, -0.00002920, -0.00014106, -0.00001015, -0.00002515, -0.00001114, -0.00000672, -0.00000530, -0.00000787, -0.00000099, -0.00000568, -0.00000371, -0.00000442, -0.00000949, 0.00000559, -0.00002940, 0.00002344, -0.00001633, 0.00001126, 0.00002578, -0.00002452, 0.00001743, -0.00000925, 0.00003203, -0.00002850, 0.00002075, 0.00000367, -0.00004797, 0.00004410, -0.00002829, 0.00003076, -0.00001774, -0.00001309, 0.00002524, -0.00002719, 0.00002487, -0.00002802, 0.00001453, -0.00000144, -0.00000698, 0.00001805, -0.00001620, 0.00001279, -0.00001416, 0.00000565, -0.00000399, -0.00000350, 0.00001912, -0.00000931, -0.00000513, 0.00001151, 0.00000136, -0.00001726, 0.00000373, -0.00000025, 0.00000368, 0.00000278, -0.00000941, 0.00000486, 0.00000017, 0.00000197, 0.00000117, -0.00000333, -0.00000516, -0.00000026, 0.00000117, -0.00000460, 0.00000696, 0.00000363, -0.00001048, -0.00000376, 0.00000583, 0.00000801, -0.00000125, -0.00000735, 0.00000366, -0.00000414, -0.00000963, 0.00001079, 0.00001061, -0.00000602, -0.00000468, -0.00000295, -0.00000042, 0.00000394, 0.00000330, 0.00000740, -0.00000141, -0.00000485, 0.00000106, -0.00000111, -0.00000044, 0.00000399, 0.00000290, -0.00000366, 0.00000256, 0.00000992, -0.00000421, -0.00000442, 0.00000455, -0.00000482, 0.00000962, -0.00000182, },
		{ 0.00000372, -0.00000030, -0.00000209, -0.00000456, 0.00000014, -0.00000283, -0.00001298, 0.00000943, -0.00000841, -0.00000398, 0.00000676, -0.00001841, 0.00001671, -0.00000778, -0.00000895, 0.00000110, -0.00001536, 0.00001293, -0.00003499, 0.00000817, -0.00007003, -0.00011786, 0.00010495, 0.00022593, 0.00020168, -0.00005687, -0.00012707, -0.00000216, -0.00002830, -0.00000764, -0.00000637, -0.00000666, -0.00000632, -0.00000105, -0.00000710, -0.00000099, -0.00000894, -0.00000344, -0.00000144, -0.00002541, 0.00002439, -0.00002121, 0.00002279, 0.00001523, -0.00001920, 0.00001511, -0.00000227, 0.00002648, -0.00003069, 0.00003315, -0.00001880, -0.00002900, 0.00003672, -0.00002855, 0.00003794, -0.00003253, 0.00000252, 0.00001309, -0.00001932, 0.00002136, -0.00003031, 0.00002236, -0.00001150, 0.00000298, 0.00001059, -0.00001273, 0.00001206, -0.00001677, 0.00001031, -0.00000962, 0.00000262, 0.00001646, -0.00001199, -0.00000113, 0.00001232, -0.00000403, -0.00001586, 0.00000631, -0.00000190, 0.00000623, -0.00000100, -0.00000757, 0.00000603, -0.00000137, 0.00000295, 0.00000002, -0.00000327, -0.00000535, 0.00000135, 0.00000012, -0.00000479, 0.00000921, 0.00000072, -0.00001125, -0.00000117, 0.00000683, 0.00000752, -0.00000408, -0.00000603, 0.00000495, -0.00000760, -0.00000723, 0.00001404, 0.00000763, -0.00000736, -0.00000399, -0.00000276, 0.00000030, 0.00000357, 0.00000452, 0.00000676, -0.00000323, -0.00000377, 0.00000078, -0.00000109, 0.00000003, 0.00000482, 0.00000181, -0.00000453, 0.00000589, 0.00000785, -0.00000691, -0.00000147, 0.00000293, -0.00000325, 0.00001108, -0.00000810, -0.00002054, },
		{ -0.00000164, -0.00000188, -0.00000506, 0.00000183, -0.00000555, -0.00001051, 0.00000941, -0.00001119, 0.00000091, 0.00000134, -0.00001505, 0.00001787, -0.00001289, -0.00000348, -0.00000465, -0.00000883, 0.00000695, -0.00003391, 0.00001085, -0.00009872, -0.00008089, 0.00014162, 0.00021037, 0.00017208, -0.00008148, -0.00011017, 0.00000291, -0.00003130, -0.00000308, -0.00000764, -0.00000711, -0.00000458, -0.00000230, -0.00000676, -0.00000029, -0.00001162, 0.00000157, -0.00000912, -0.00001903, 0.00002195, -0.00002217, 0.00003154, 0.00000525, -0.00001343, 0.00001316, 0.00000507, 0.00001711, -0.00002574, 0.00003723, -0.00003653, -0.00000842, 0.00002344, -0.00002125, 0.00003757, -0.00004200, 0.00001616, 0.00000017, -0.00000847, 0.00001316, -0.00002716, 0.00002546, -0.00001869, 0.00001233, 0.00000167, -0.00000651, 0.00000801, -0.00001623, 0.00001277, -0.00001429, 0.00000939, 0.00001177, -0.00001278, 0.00000291, 0.00001162, -0.00000926, -0.00001280, 0.00000748, -0.00000267, 0.00000804, -0.00000493, -0.00000441, 0.00000580, -0.00000222, 0.00000366, -0.00000109, -0.00000312, -0.00000544, 0.00000315, -0.00000176, -0.00000395, 0.00001093, -0.00000286, -0.00001080, 0.00000125, 0.00000762, 0.00000642, -0.00000678, -0.00000360, 0.00000493, -0.00001081, -0.00000344, 0.00001604, 0.00000425, -0.00000807, -0.00000333, -0.00000243, 0.00000071, 0.00000349, 0.00000578, 0.00000539, -0.00000441, -0.00000280, 0.00000043, -0.00000090, 0.00000049, 0.00000570, -0.00000004, -0.00000423, 0.00000891, 0.00000440, -0.00000802, 0.00000095, 0.00000119, -0.00000058, 0.00001042, -0.00001342, -0.00001640, 0.00000488, },
		{ -0.00000206, -0.00000494, 0.00000299, -0.00000822, -0.00000733, 0.00000792, -0.00001219, 0.00000457, -0.00000416, -0.00001002, 0.00001639, -0.00001557, 0.00000033, -0.00000946, -0.00000197, -0.00000161, -0.00002774, 0.00000476, -0.00012019, -0.00003327, 0.00016228, 0.00019259, 0.00014548, -0.00010308, -0.00009039, 0.00000415, -0.00003283, 0.00000155, -0.00001025, -0.00000618, -0.00000355, -0.00000378, -0.00000542, -0.00000125, -0.00001222, 0.00000464, -0.00001612, -0.00001159, 0.00001712, -0.00001956, 0.00003709, -0.00000344, -0.00000780, 0.00001252, 0.00001070, 0.00000673, -0.00001581, 0.00003287, -0.00004664, 0.00000959, 0.00000794, -0.00000858, 0.00002994, -0.00004475, 0.00002550, -0.00001091, 0.00000286, 0.00000215, -0.00001939, 0.00002344, -0.00002161, 0.00001913, -0.00000678, 0.00000099, 0.00000139, -0.00001249, 0.00001227, -0.00001686, 0.00001563, 0.00000572, -0.00001152, 0.00000639, 0.00000952, -0.00001370, -0.00000859, 0.00000730, -0.00000235, 0.00000860, -0.00000825, -0.00000055, 0.00000427, -0.00000218, 0.00000384, -0.00000189, -0.00000312, -0.00000518, 0.00000479, -0.00000427, -0.00000192, 0.00001161, -0.00000659, -0.00000922, 0.00000328, 0.00000827, 0.00000456, -0.00000886, -0.00000045, 0.00000338, -0.00001308, 0.00000124, 0.00001660, 0.00000088, -0.00000827, -0.00000264, -0.00000207, 0.00000083, 0.00000388, 0.00000675, 0.00000357, -0.00000494, -0.00000210, 0.00000018, -0.00000070, 0.00000111, 0.00000634, -0.00000237, -0.00000255, 0.00001088, 0.00000025, -0.00000760, 0.00000245, -0.00000000, 0.00000244, 0.00000784, -0.00001711, -0.00001181, 0.00000483, -0.00000240, },
		{ -0.00000415, 0.00000333, -0.00001037, -0.00000402, 0.00000541, -0.00001148, 0.00000647, -0.00000879, -0.00000436, 0.00001297, -0.00001591, 0.00000199, -0.00001226, 0.00000352, -0.00001066, -0.00001839, -0.00001016, -0.00012959, 0.00001841, 0.00016723, 0.00017696, 0.00011950, -0.00012059, -0.00006877, 0.00000139, -0.00003186, 0.00000505, -0.00001336, -0.00000405, -0.00000367, -0.00000471, -0.00000393, -0.00000317, -0.00001106, 0.00000547, -0.00002152, -0.00000429, 0.00001107, -0.00001411, 0.00003949, -0.00001056, -0.00000236, 0.00001336, 0.00001330, -0.00000199, -0.00000402, 0.00002184, -0.00004818, 0.00002191, -0.00000602, 0.00000627, 0.00001683, -0.00004084, 0.00002927, -0.00001810, 0.00001220, -0.00000925, -0.00000876, 0.00001700, -0.00001982, 0.00002204, -0.00001298, 0.00000797, -0.00000641, -0.00000621, 0.00000872, -0.00001653, 0.00002014, -0.00000067, -0.00000843, 0.00000878, 0.00000629, -0.00001687, -0.00000387, 0.00000603, -0.00000101, 0.00000769, -0.00001032, 0.00000326, 0.00000186, -0.00000127, 0.00000338, -0.00000221, -0.00000343, -0.00000434, 0.00000583, -0.00000695, 0.00000119, 0.00001087, -0.00000985, -0.00000680, 0.00000486, 0.00000874, 0.00000192, -0.00000983, 0.00000283, 0.00000042, -0.00001385, 0.00000618, 0.00001576, -0.00000220, -0.00000808, -0.00000188, -0.00000184, 0.00000084, 0.00000472, 0.00000716, 0.00000161, -0.00000495, -0.00000170, 0.00000013, -0.00000058, 0.00000201, 0.00000641, -0.00000473, 0.00000037, 0.00001128, -0.00000379, -0.00000606, 0.00000298, -0.00000021, 0.00000498, 0.00000393, -0.00001888, -0.00000748, 0.00000402, -0.00000271, -0.00000257, },
		{ 0.00000283, -0.00001170, -0.00000108, 0.00000249, -0.00000946, 0.00000651, -0.00001187, 0.00000092, 0.00000854, -0.00001448, 0.00000165, -0.00001265, 0.00000638, -0.00001804, -0.00000871, -0.00003189, -0.00012408, 0.00006712, 0.00015944, 0.00016583, 0.00009204, -0.00013244, -0.00004709, -0.00000466, -0.00002801, 0.00000645, -0.00001592, -0.00000134, -0.00000483, -0.00000472, -0.00000296, -0.00000525, -0.00000879, 0.00000425, -0.00002489, 0.00000199, 0.00000487, -0.00000671, 0.00003908, -0.00001607, 0.00000319, 0.00001508, 0.00001261, -0.00000716, 0.00000640, 0.00000719, -0.00004210, 0.00002702, -0.00001544, 0.00001987, 0.00000101, -0.00003164, 0.00002752, -0.00002038, 0.00001767, -0.00001867, 0.00000241, 0.00000774, -0.00001391, 0.00002060, -0.00001565, 0.00001276, -0.00001369, 0.00000132, 0.00000267, -0.00001307, 0.00002199, -0.00000625, -0.00000410, 0.00000979, 0.00000237, -0.00001850, 0.00000076, 0.00000413, 0.00000103, 0.00000538, -0.00001070, 0.00000629, -0.00000087, 0.00000025, 0.00000234, -0.00000204, -0.00000412, -0.00000277, 0.00000590, -0.00000924, 0.00000499, 0.00000860, -0.00001211, -0.00000396, 0.00000607, 0.00000888, -0.00000132, -0.00000938, 0.00000553, -0.00000351, -0.00001284, 0.00001071, 0.00001377, -0.00000481, -0.00000758, -0.00000106, -0.00000188, 0.00000097, 0.00000587, 0.00000688, -0.00000016, -0.00000467, -0.00000152, 0.00000028, -0.00000056, 0.00000321, 0.00000567, -0.00000656, 0.00000406, 0.00000991, -0.00000703, -0.00000400, 0.00000283, 0.00000064, 0.00000636, -0.00000051, -0.00001887, -0.00000389, 0.00000285, -0.00000278, -0.00000318, -0.00000287, },
		{ -0.00001211, 0.00000110, -0.00000023, -0.00000674, 0.00000498, -0.00001316, 0.00000508, 0.00000403, -0.00001212, -0.00000006, -0.00001092, 0.00000616, -0.00002217, -0.00000177, -0.00005668, -0.00010343, 0.00010677, 0.00014369, 0.00015935, 0.00006200, -0.00013724, -0.00002747, -0.00001256, -0.00002165, 0.00000531, -0.00001703, 0.00000108, -0.00000653, -0.00000389, -0.00000282, -0.00000687, -0.00000619, 0.00000151, -0.00002625, 0.00000676, -0.00000063, 0.00000179, 0.00003623, -0.00001995, 0.00000924, 0.00001660, 0.00000940, -0.00000814, 0.00001291, -0.00000756, -0.00003086, 0.00002524, -0.00001869, 0.00002936, -0.00001443, -0.00001944, 0.00002143, -0.00001789, 0.00001837, -0.00002429, 0.00001185, -0.00000216, -0.00000535, 0.00001535, -0.00001437, 0.00001415, -0.00001884, 0.00000850, -0.00000469, -0.00000695, 0.00002077, -0.00000997, 0.00000063, 0.00000940, -0.00000182, -0.00001851, 0.00000478, 0.00000207, 0.00000332, 0.00000205, -0.00000933, 0.00000800, -0.00000332, 0.00000198, 0.00000096, -0.00000150, -0.00000507, -0.00000052, 0.00000473, -0.00001055, 0.00000886, 0.00000501, -0.00001303, -0.00000112, 0.00000705, 0.00000845, -0.00000477, -0.00000747, 0.00000703, -0.00000774, -0.00001009, 0.00001428, 0.00001101, -0.00000691, -0.00000678, -0.00000024, -0.00000225, 0.00000146, 0.00000705, 0.00000594, -0.00000153, -0.00000431, -0.00000141, 0.00000053, -0.00000055, 0.00000461, 0.00000402, -0.00000733, 0.00000785, 0.00000695, -0.00000901, -0.00000200, 0.00000252, 0.00000224, 0.00000627, -0.00000466, -0.00001750, -0.00000122, 0.00000162, -0.00000282, -0.00000367, -0.00000136, 0.00000161, },
		{ 0.00000241, -0.00000230, -0.00000399, 0.00000243, -0.00001279, 0.00000777, 0.00000017, -0.00000968, -0.00000226, -0.00000788, 0.00000325, -0.00002245, -0.00000012, -0.00007981, -0.00007002, 0.00013322, 0.00012528, 0.00015587, 0.00002963, -0.00013438, -0.00001182, -0.00002052, -0.00001382, 0.00000184, -0.00001622, 0.00000249, -0.00000812, -0.00000262, -0.00000345, -0.00000771, -0.00000385, -0.00000213, -0.00002590, 0.00000990, -0.00000486, 0.00001066, 0.00003129, -0.00002205, 0.00001590, 0.00001681, 0.00000515, -0.00000556, 0.00001426, -0.00001932, -0.00001770, 0.00001837, -0.00001574, 0.00003301, -0.00002682, -0.00000686, 0.00001296, -0.00001184, 0.00001453, -0.00002521, 0.00001777, -0.00001055, 0.00000385, 0.00000763, -0.00000962, 0.00001173, -0.00002064, 0.00001375, -0.00001184, 0.00000080, 0.00001666, -0.00001115, 0.00000483, 0.00000784, -0.00000589, -0.00001702, 0.00000782, 0.00000027, 0.00000534, -0.00000172, -0.00000645, 0.00000811, -0.00000497, 0.00000351, -0.00000045, -0.00000084, -0.00000606, 0.00000218, 0.00000231, -0.00001044, 0.00001210, 0.00000058, -0.00001251, 0.00000142, 0.00000793, 0.00000719, -0.00000790, -0.00000436, 0.00000690, -0.00001149, -0.00000595, 0.00001654, 0.00000784, -0.00000852, -0.00000565, 0.00000041, -0.00000287, 0.00000246, 0.00000796, 0.00000454, -0.00000244, -0.00000404, -0.00000121, 0.00000074, -0.00000040, 0.00000595, 0.00000160, -0.00000671, 0.00001099, 0.00000291, -0.00000960, -0.00000047, 0.00000255, 0.00000404, 0.00000477, -0.00000792, -0.00001535, 0.00000057, 0.00000050, -0.00000290, -0.00000386, 0.00000007, -0.00000003, -0.00001175, },
		{ -0.00000352, -0.00000172, -0.00000051, -0.00001119, 0.00000900, -0.00000269, -0.00000775, -0.00000410, -0.00000466, -0.00000126, -0.00001939, -0.00000508, -0.00009662, -0.00002837, 0.00014493, 0.00010896, 0.00015270, -0.00000355, -0.00012431, -0.00000137, -0.00002685, -0.00000592, -0.00000317, -0.00001356, 0.00000245, -0.00000901, -0.00000145, -0.00000452, -0.00000778, -0.00000218, -0.00000610, -0.00002425, 0.00001150, -0.00000748, 0.00001929, 0.00002457, -0.00002201, 0.00002288, 0.00001496, 0.00000156, -0.00000107, 0.00001064, -0.00002607, -0.00000574, 0.00000911, -0.00000797, 0.00003050, -0.00003437, 0.00000381, 0.00000435, -0.00000410, 0.00000736, -0.00002166, 0.00001930, -0.00001574, 0.00001166, -0.00000070, -0.00000270, 0.00000595, -0.00001867, 0.00001588, -0.00001724, 0.00000872, 0.00001048, -0.00000961, 0.00000772, 0.00000551, -0.00000953, -0.00001424, 0.00000972, -0.00000097, 0.00000665, -0.00000527, -0.00000260, 0.00000667, -0.00000552, 0.00000445, -0.00000154, -0.00000036, -0.00000676, 0.00000490, -0.00000109, -0.00000869, 0.00001407, -0.00000402, -0.00001075, 0.00000348, 0.00000873, 0.00000499, -0.00001013, -0.00000059, 0.00000503, -0.00001405, -0.00000100, 0.00001738, 0.00000461, -0.00000969, -0.00000418, 0.00000069, -0.00000356, 0.00000398, 0.00000832, 0.00000297, -0.00000293, -0.00000391, -0.00000084, 0.00000080, 0.00000006, 0.00000692, -0.00000122, -0.00000463, 0.00001283, -0.00000150, -0.00000903, 0.00000048, 0.00000328, 0.00000535, 0.00000227, -0.00001000, -0.00001293, 0.00000167, -0.00000052, -0.00000300, -0.00000370, 0.00000119, -0.00000181, -0.00001085, 0.00000294, },
		{ -0.00000024, -0.00000328, -0.00000889, 0.00000908, -0.00000455, -0.00000658, -0.00000506, -0.00000225, -0.00000595, -0.00001451, -0.00001640, -0.00010344, 0.00001585, 0.00014293, 0.00009798, 0.00014698, -0.00003523, -0.00010851, 0.00000353, -0.00003030, 0.00000066, -0.00000861, -0.00000960, 0.00000099, -0.00000885, -0.00000084, -0.00000559, -0.00000735, -0.00000125, -0.00000998, -0.00002174, 0.00001178, -0.00000830, 0.00002712, 0.00001644, -0.00001943, 0.00002940, 0.00001095, -0.00000002, 0.00000321, 0.00000354, -0.00002728, 0.00000266, 0.00000035, 0.00000224, 0.00002285, -0.00003654, 0.00001102, -0.00000250, 0.00000324, -0.00000128, -0.00001480, 0.00001659, -0.00001684, 0.00001648, -0.00000773, 0.00000464, -0.00000194, -0.00001329, 0.00001440, -0.00001968, 0.00001528, 0.00000350, -0.00000575, 0.00000878, 0.00000289, -0.00001257, -0.00001049, 0.00001044, -0.00000150, 0.00000696, -0.00000797, 0.00000153, 0.00000402, -0.00000491, 0.00000461, -0.00000208, -0.00000029, -0.00000683, 0.00000714, -0.00000499, -0.00000543, 0.00001434, -0.00000810, -0.00000810, 0.00000507, 0.00000932, 0.00000190, -0.00001097, 0.00000312, 0.00000169, -0.00001492, 0.00000409, 0.00001692, 0.00000150, -0.00001043, -0.00000239, 0.00000042, -0.00000401, 0.00000588, 0.00000797, 0.00000150, -0.00000319, -0.00000389, -0.00000028, 0.00000068, 0.00000090, 0.00000719, -0.00000392, -0.00000136, 0.00001298, -0.00000559, -0.00000771, 0.00000101, 0.00000478, 0.00000564, -0.00000058, -0.00001092, -0.00001062, 0.00000228, -0.00000149, -0.00000303, -0.00000327, 0.00000193, -0.00000349, -0.00000972, 0.00000296, -0.00000394, },
		{ -0.00000555, -0.00000641, 0.00000844, -0.00000563, -0.00000607, -0.00000502, -0.00000128, -0.00000949, -0.00000989, -0.00003224, -0.00009835, 0.00005685, 0.00013037, 0.00009370, 0.00013646, -0.00006286, -0.00008921, 0.00000353, -0.00003040, 0.00000489, -0.00001329, -0.00000519, -0.00000150, -0.00000769, -0.00000103, -0.00000625, -0.00000676, -0.00000098, -0.00001357, -0.00001868, 0.00001100, -0.00000722, 0.00003359, 0.00000741, -0.00001397, 0.00003439, 0.00000545, 0.00000097, 0.00000537, -0.00000477, -0.00002381, 0.00000644, -0.00000562, 0.00001227, 0.00001200, -0.00003392, 0.00001426, -0.00000636, 0.00000845, -0.00000936, -0.00000642, 0.00001074, -0.00001400, 0.00001748, -0.00001194, 0.00001057, -0.00001023, -0.00000566, 0.00000960, -0.00001860, 0.00001920, -0.00000280, -0.00000045, 0.00000793, 0.00000039, -0.00001489, -0.00000609, 0.00001007, -0.00000135, 0.00000620, -0.00000937, 0.00000521, 0.00000070, -0.00000334, 0.00000397, -0.00000196, -0.00000079, -0.00000604, 0.00000842, -0.00000871, -0.00000109, 0.00001278, -0.00001108, -0.00000504, 0.00000632, 0.00000948, -0.00000178, -0.00001019, 0.00000600, -0.00000255, -0.00001392, 0.00000866, 0.00001543, -0.00000136, -0.00001069, -0.00000041, -0.00000050, -0.00000394, 0.00000787, 0.00000691, 0.00000038, -0.00000341, -0.00000382, 0.00000036, 0.00000044, 0.00000210, 0.00000656, -0.00000595, 0.00000255, 0.00001139, -0.00000879, -0.00000615, 0.00000151, 0.00000677, 0.00000462, -0.00000312, -0.00001099, -0.00000856, 0.00000255, -0.00000242, -0.00000287, -0.00000273, 0.00000232, -0.00000498, -0.00000856, 0.00000276, -0.00000380, 0.00000041, },
		{ -0.00000406, 0.00000746, -0.00000629, -0.00000591, -0.00000424, -0.00000183, -0.00001108, -0.00000760, -0.00004961, -0.00008152, 0.00008983, 0.00011160, 0.00009561, 0.00012008, -0.00008428, -0.00006892, -0.00000004, -0.00002740, 0.00000631, -0.00001633, -0.00000124, -0.00000436, -0.00000583, -0.00000199, -0.00000634, -0.00000632, -0.00000117, -0.00001680, -0.00001533, 0.00000942, -0.00000421, 0.00003811, -0.00000171, -0.00000568, 0.00003676, -0.00000032, 0.00000418, 0.00000424, -0.00001191, -0.00001755, 0.00000588, -0.00000752, 0.00001983, 0.00000030, -0.00002798, 0.00001399, -0.00000695, 0.00001049, -0.00001519, 0.00000151, 0.00000344, -0.00000823, 0.00001476, -0.00001258, 0.00001362, -0.00001711, 0.00000259, 0.00000251, -0.00001419, 0.00001978, -0.00000708, 0.00000506, 0.00000549, -0.00000173, -0.00001643, -0.00000140, 0.00000882, -0.00000067, 0.00000449, -0.00000926, 0.00000784, -0.00000265, -0.00000118, 0.00000271, -0.00000126, -0.00000183, -0.00000435, 0.00000837, -0.00001159, 0.00000366, 0.00000962, -0.00001259, -0.00000201, 0.00000738, 0.00000892, -0.00000554, -0.00000784, 0.00000745, -0.00000693, -0.00001119, 0.00001222, 0.00001327, -0.00000398, -0.00001038, 0.00000154, -0.00000202, -0.00000308, 0.00000958, 0.00000533, -0.00000031, -0.00000374, -0.00000357, 0.00000094, 0.00000022, 0.00000347, 0.00000503, -0.00000683, 0.00000639, 0.00000836, -0.00001081, -0.00000475, 0.00000243, 0.00000871, 0.00000239, -0.00000487, -0.00001062, -0.00000674, 0.00000249, -0.00000326, -0.00000248, -0.00000223, 0.00000248, -0.00000630, -0.00000744, 0.00000256, -0.00000337, 0.00000038, -0.00000718, },
		{ 0.00000639, -0.00000681, -0.00000570, -0.00000327, -0.00000345, -0.00001060, -0.00000915, -0.00006499, -0.00005510, 0.00011170, 0.00009126, 0.00010171, 0.00009805, -0.00009820, -0.00004997, -0.00000547, -0.00002219, 0.00000508, -0.00001726, 0.00000151, -0.00000690, -0.00000379, -0.00000342, -0.00000588, -0.00000621, -0.00000164, -0.00001966, -0.00001187, 0.00000736, 0.00000066, 0.00004013, -0.00000991, 0.00000492, 0.00003578, -0.00000489, 0.00000840, -0.00000029, -0.00001603, -0.00001082, 0.00000236, -0.00000532, 0.00002350, -0.00001013, -0.00002054, 0.00001140, -0.00000487, 0.00000922, -0.00001778, 0.00000735, -0.00000344, -0.00000115, 0.00000929, -0.00000974, 0.00001305, -0.00002113, 0.00000971, -0.00000533, -0.00000737, 0.00001705, -0.00000843, 0.00000955, 0.00000211, -0.00000344, -0.00001711, 0.00000322, 0.00000690, 0.00000035, 0.00000215, -0.00000770, 0.00000902, -0.00000545, 0.00000111, 0.00000113, -0.00000018, -0.00000321, -0.00000191, 0.00000688, -0.00001307, 0.00000804, 0.00000536, -0.00001255, 0.00000068, 0.00000837, 0.00000739, -0.00000877, -0.00000433, 0.00000711, -0.00001066, -0.00000717, 0.00001451, 0.00001075, -0.00000635, -0.00000941, 0.00000316, -0.00000395, -0.00000134, 0.00001064, 0.00000352, -0.00000061, -0.00000424, -0.00000303, 0.00000132, 0.00000021, 0.00000473, 0.00000281, -0.00000634, 0.00000948, 0.00000443, -0.00001169, -0.00000366, 0.00000410, 0.00000995, -0.00000062, -0.00000560, -0.00001018, -0.00000505, 0.00000207, -0.00000389, -0.00000193, -0.00000185, 0.00000252, -0.00000756, -0.00000632, 0.00000245, -0.00000290, -0.00000012, -0.00000821, -0.00000198, },
		{ -0.00000731, -0.00000520, -0.00000264, -0.00000539, -0.00000874, -0.00001505, -0.00007507, -0.00002275, 0.00012144, 0.00007335, 0.00010924, 0.00007181, -0.00010436, -0.00003405, -0.00001107, -0.00001600, 0.00000188, -0.00001613, 0.00000269, -0.00000858, -0.00000210, -0.00000491, -0.00000512, -0.00000639, -0.00000232, -0.00002214, -0.00000849, 0.00000516, 0.00000713, 0.00003925, -0.00001602, 0.00001676, 0.00003133, -0.00000705, 0.00001195, -0.00000726, -0.00001630, -0.00000570, -0.00000215, -0.00000006, 0.00002291, -0.00001779, -0.00001335, 0.00000797, -0.00000130, 0.00000533, -0.00001702, 0.00001014, -0.00000831, 0.00000548, 0.00000260, -0.00000438, 0.00000897, -0.00002148, 0.00001424, -0.00001225, 0.00000038, 0.00001176, -0.00000660, 0.00001197, -0.00000140, -0.00000487, -0.00001685, 0.00000743, 0.00000459, 0.00000146, -0.00000044, -0.00000499, 0.00000863, -0.00000725, 0.00000307, -0.00000041, 0.00000099, -0.00000465, 0.00000094, 0.00000410, -0.00001283, 0.00001134, 0.00000069, -0.00001117, 0.00000289, 0.00000926, 0.00000481, -0.00001084, -0.00000033, 0.00000502, -0.00001306, -0.00000249, 0.00001551, 0.00000813, -0.00000847, -0.00000774, 0.00000413, -0.00000594, 0.00000121, 0.00001079, 0.00000183, -0.00000071, -0.00000486, -0.00000217, 0.00000141, 0.00000057, 0.00000553, 0.00000034, -0.00000452, 0.00001127, 0.00000025, -0.00001170, -0.00000279, 0.00000657, 0.00000993, -0.00000374, -0.00000543, -0.00000989, -0.00000339, 0.00000124, -0.00000412, -0.00000139, -0.00000154, 0.00000249, -0.00000885, -0.00000505, 0.00000238, -0.00000253, -0.00000087, -0.00000880, -0.00000017, 0.00000530, },
		{ -0.00000438, -0.00000273, -0.00000683, -0.00000669, -0.00002466, -0.00007751, 0.00001115, 0.00012005, 0.00006060, 0.00011535, 0.00004352, -0.00010351, -0.00002209, -0.00001554, -0.00001008, -0.00000229, -0.00001343, 0.00000232, -0.00000913, -0.00000115, -0.00000603, -0.00000441, -0.00000672, -0.00000323, -0.00002415, -0.00000537, 0.00000327, 0.00001472, 0.00003540, -0.00001896, 0.00002838, 0.00002402, -0.00000615, 0.00001317, -0.00001491, -0.00001302, -0.00000348, -0.00000566, 0.00000654, 0.00001871, -0.00002207, -0.00000761, 0.00000504, 0.00000235, 0.00000007, -0.00001358, 0.00000976, -0.00001027, 0.00001012, -0.00000359, 0.00000201, 0.00000229, -0.00001820, 0.00001536, -0.00001682, 0.00000745, 0.00000526, -0.00000202, 0.00001178, -0.00000428, -0.00000624, -0.00001552, 0.00001088, 0.00000216, 0.00000246, -0.00000285, -0.00000162, 0.00000679, -0.00000782, 0.00000437, -0.00000158, 0.00000193, -0.00000582, 0.00000375, 0.00000044, -0.00001091, 0.00001306, -0.00000366, -0.00000883, 0.00000463, 0.00000988, 0.00000132, -0.00001132, 0.00000336, 0.00000157, -0.00001373, 0.00000219, 0.00001537, 0.00000553, -0.00001026, -0.00000541, 0.00000420, -0.00000756, 0.00000427, 0.00000993, 0.00000056, -0.00000084, -0.00000544, -0.00000109, 0.00000123, 0.00000134, 0.00000561, -0.00000187, -0.00000173, 0.00001151, -0.00000358, -0.00001125, -0.00000180, 0.00000957, 0.00000834, -0.00000630, -0.00000472, -0.00000977, -0.00000177, 0.00000001, -0.00000382, -0.00000107, -0.00000116, 0.00000230, -0.00001015, -0.00000352, 0.00000220, -0.00000223, -0.00000176, -0.00000907, 0.00000148, 0.00000416, -0.00000773, },
		{ -0.00000355, -0.00000720, -0.00000583, -0.00003632, -0.00007133, 0.00004235, 0.00011005, 0.00005416, 0.00011777, 0.00001563, -0.00009717, -0.00001416, -0.00001813, -0.00000539, -0.00000641, -0.00000991, 0.00000076, -0.00000860, -0.00000109, -0.00000651, -0.00000406, -0.00000696, -0.00000448, -0.00002562, -0.00000272, 0.00000222, 0.00002270, 0.00002891, -0.00001801, 0.00003819, 0.00001510, -0.00000243, 0.00001092, -0.00002124, -0.00000748, -0.00000444, -0.00000676, 0.00001264, 0.00001224, -0.00002324, -0.00000384, 0.00000347, 0.00000484, -0.00000511, -0.00000871, 0.00000687, -0.00000918, 0.00001189, -0.00000783, 0.00000774, -0.00000548, -0.00001212, 0.00001300, -0.00001817, 0.00001239, -0.00000084, 0.00000421, 0.00000898, -0.00000599, -0.00000778, -0.00001303, 0.00001325, -0.00000013, 0.00000318, -0.00000470, 0.00000186, 0.00000389, -0.00000717, 0.00000482, -0.00000214, 0.00000243, -0.00000641, 0.00000603, -0.00000351, -0.00000762, 0.00001299, -0.00000708, -0.00000599, 0.00000606, 0.00000995, -0.00000267, -0.00001004, 0.00000598, -0.00000259, -0.00001259, 0.00000627, 0.00001440, 0.00000298, -0.00001158, -0.00000261, 0.00000323, -0.00000838, 0.00000743, 0.00000818, -0.00000009, -0.00000123, -0.00000576, 0.00000002, 0.00000090, 0.00000242, 0.00000483, -0.00000331, 0.00000144, 0.00001029, -0.00000669, -0.00001068, -0.00000027, 0.00001255, 0.00000526, -0.00000774, -0.00000395, -0.00000965, -0.00000032, -0.00000146, -0.00000297, -0.00000115, -0.00000052, 0.00000176, -0.00001132, -0.00000171, 0.00000178, -0.00000192, -0.00000280, -0.00000903, 0.00000300, 0.00000287, -0.00000791, -0.00000345, },
		{ -0.00000640, -0.00000729, -0.00004774, -0.00005707, 0.00006745, 0.00009480, 0.00005374, 0.00011515, -0.00000968, -0.00008721, -0.00000970, -0.00001872, -0.00000248, -0.00000966, -0.00000635, -0.00000145, -0.00000728, -0.00000184, -0.00000626, -0.00000429, -0.00000691, -0.00000619, -0.00002644, -0.00000069, 0.00000248, 0.00003015, 0.00002059, -0.00001296, 0.00004480, 0.00000620, 0.00000306, 0.00000498, -0.00002450, -0.00000156, -0.00000782, -0.00000488, 0.00001671, 0.00000511, -0.00002217, -0.00000184, 0.00000349, 0.00000546, -0.00000899, -0.00000384, 0.00000270, -0.00000572, 0.00001067, -0.00000927, 0.00001138, -0.00001267, -0.00000464, 0.00000789, -0.00001623, 0.00001428, -0.00000502, 0.00001063, 0.00000417, -0.00000636, -0.00000961, -0.00000940, 0.00001428, -0.00000201, 0.00000353, -0.00000575, 0.00000490, 0.00000044, -0.00000550, 0.00000442, -0.00000200, 0.00000234, -0.00000627, 0.00000738, -0.00000712, -0.00000355, 0.00001127, -0.00000917, -0.00000310, 0.00000731, 0.00000917, -0.00000656, -0.00000723, 0.00000698, -0.00000664, -0.00000990, 0.00000935, 0.00001290, 0.00000047, -0.00001226, 0.00000034, 0.00000129, -0.00000810, 0.00001018, 0.00000583, -0.00000011, -0.00000202, -0.00000563, 0.00000096, 0.00000063, 0.00000358, 0.00000330, -0.00000362, 0.00000435, 0.00000797, -0.00000895, -0.00001018, 0.00000215, 0.00001478, 0.00000114, -0.00000784, -0.00000354, -0.00000929, 0.00000074, -0.00000292, -0.00000171, -0.00000167, 0.00000053, 0.00000066, -0.00001210, 0.00000023, 0.00000107, -0.00000149, -0.00000406, -0.00000863, 0.00000435, 0.00000142, -0.00000770, -0.00000273, -0.00000482, },
		{ -0.00001162, -0.00005652, -0.00003662, 0.00008437, 0.00007779, 0.00005792, 0.00010720, -0.00003082, -0.00007547, -0.00000776, -0.00001765, -0.00000142, -0.00001154, -0.00000342, -0.00000369, -0.00000563, -0.00000310, -0.00000541, -0.00000512, -0.00000647, -0.00000844, -0.00002658, 0.00000070, 0.00000444, 0.00003612, 0.00001163, -0.00000429, 0.00004735, -0.00000111, 0.00000869, -0.00000389, -0.00002372, 0.00000286, -0.00001216, -0.00000042, 0.00001795, -0.00000129, -0.00001998, -0.00000092, 0.00000471, 0.00000413, -0.00001085, -0.00000017, -0.00000139, -0.00000107, 0.00000710, -0.00000779, 0.00001205, -0.00001781, 0.00000261, 0.00000132, -0.00001166, 0.00001298, -0.00000624, 0.00001574, -0.00000167, -0.00000559, -0.00001163, -0.00000477, 0.00001386, -0.00000328, 0.00000345, -0.00000586, 0.00000706, -0.00000303, -0.00000319, 0.00000332, -0.00000119, 0.00000167, -0.00000537, 0.00000759, -0.00000982, 0.00000062, 0.00000831, -0.00000983, -0.00000046, 0.00000846, 0.00000731, -0.00000967, -0.00000340, 0.00000615, -0.00000981, -0.00000617, 0.00001126, 0.00001117, -0.00000203, -0.00001214, 0.00000307, -0.00000138, -0.00000657, 0.00001207, 0.00000331, 0.00000033, -0.00000320, -0.00000494, 0.00000152, 0.00000064, 0.00000449, 0.00000132, -0.00000271, 0.00000640, 0.00000508, -0.00001052, -0.00000967, 0.00000556, 0.00001558, -0.00000328, -0.00000670, -0.00000374, -0.00000849, 0.00000118, -0.00000400, -0.00000035, -0.00000250, 0.00000197, -0.00000116, -0.00001219, 0.00000204, 0.00000017, -0.00000093, -0.00000561, -0.00000775, 0.00000539, -0.00000007, -0.00000707, -0.00000254, -0.00000600, -0.00000485, },
		{ -0.00006067, -0.00001273, 0.00009256, 0.00006199, 0.00006468, 0.00009456, -0.00004695, -0.00006349, -0.00000732, -0.00001557, -0.00000186, -0.00001196, -0.00000151, -0.00000549, -0.00000408, -0.00000448, -0.00000426, -0.00000642, -0.00000568, -0.00001120, -0.00002610, 0.00000158, 0.00000820, 0.00003980, 0.00000343, 0.00000692, 0.00004570, -0.00000566, 0.00001267, -0.00001409, -0.00001897, 0.00000438, -0.00001578, 0.00000550, 0.00001637, -0.00000605, -0.00001771, -0.00000024, 0.00000635, 0.00000136, -0.00001063, 0.00000160, -0.00000419, 0.00000340, 0.00000236, -0.00000404, 0.00000964, -0.00001996, 0.00000817, -0.00000518, -0.00000569, 0.00000910, -0.00000412, 0.00001829, -0.00000733, -0.00000418, -0.00001354, 0.00000050, 0.00001201, -0.00000378, 0.00000297, -0.00000504, 0.00000804, -0.00000598, -0.00000065, 0.00000180, 0.00000011, 0.00000054, -0.00000385, 0.00000664, -0.00001126, 0.00000423, 0.00000472, -0.00000919, 0.00000179, 0.00000944, 0.00000435, -0.00001140, 0.00000065, 0.00000366, -0.00001154, -0.00000203, 0.00001205, 0.00000935, -0.00000449, -0.00001113, 0.00000517, -0.00000437, -0.00000388, 0.00001277, 0.00000106, 0.00000095, -0.00000463, -0.00000372, 0.00000163, 0.00000109, 0.00000481, -0.00000062, -0.00000080, 0.00000724, 0.00000217, -0.00001171, -0.00000884, 0.00000973, 0.00001452, -0.00000718, -0.00000477, -0.00000454, -0.00000721, 0.00000088, -0.00000439, 0.00000074, -0.00000332, 0.00000363, -0.00000369, -0.00001134, 0.00000337, -0.00000068, -0.00000037, -0.00000741, -0.00000630, 0.00000595, -0.00000138, -0.00000619, -0.00000301, -0.00000658, -0.00000336, -0.00001857, },
		{ 0.00001155, 0.00009277, 0.00004949, 0.00007192, 0.00007849, -0.00005798, -0.00005232, -0.00000749, -0.00001320, -0.00000324, -0.00001114, -0.00000073, -0.00000657, -0.00000300, -0.00000562, -0.00000316, -0.00000793, -0.00000468, -0.00001438, -0.00002514, 0.00000227, 0.00001359, 0.00004070, -0.00000269, 0.00001916, 0.00004044, -0.00000704, 0.00001353, -0.00002357, -0.00001131, 0.00000248, -0.00001722, 0.00001140, 0.00001268, -0.00000893, -0.00001598, 0.00000091, 0.00000752, -0.00000198, -0.00000883, 0.00000141, -0.00000504, 0.00000650, -0.00000220, 0.00000083, 0.00000475, -0.00001891, 0.00001109, -0.00001023, 0.00000019, 0.00000386, 0.00000092, 0.00001760, -0.00001172, -0.00000280, -0.00001487, 0.00000591, 0.00000897, -0.00000348, 0.00000215, -0.00000346, 0.00000774, -0.00000803, 0.00000169, 0.00000019, 0.00000166, -0.00000085, -0.00000196, 0.00000473, -0.00001133, 0.00000676, 0.00000114, -0.00000759, 0.00000366, 0.00001005, 0.00000054, -0.00001136, 0.00000409, 0.00000005, -0.00001155, 0.00000191, 0.00001192, 0.00000754, -0.00000680, -0.00000924, 0.00000630, -0.00000717, -0.00000035, 0.00001217, -0.00000054, 0.00000140, -0.00000603, -0.00000212, 0.00000133, 0.00000201, 0.00000430, -0.00000199, 0.00000165, 0.00000682, -0.00000036, -0.00001283, -0.00000723, 0.00001409, 0.00001153, -0.00000984, -0.00000265, -0.00000569, -0.00000558, -0.00000012, -0.00000393, 0.00000119, -0.00000376, 0.00000512, -0.00000671, -0.00000949, 0.00000394, -0.00000118, -0.00000008, -0.00000926, -0.00000431, 0.00000589, -0.00000225, -0.00000537, -0.00000402, -0.00000622, -0.00000288, -0.00002715, -0.00005906, },
		{ 0.00008677, 0.00004127, 0.00007785, 0.00006053, -0.00006436, -0.00004251, -0.00000771, -0.00001113, -0.00000490, -0.00000954, -0.00000092, -0.00000685, -0.00000254, -0.00000624, -0.00000244, -0.00000933, -0.00000371, -0.00001780, -0.00002387, 0.00000321, 0.00002011, 0.00003879, -0.00000568, 0.00003080, 0.00003277, -0.00000565, 0.00001053, -0.00003033, -0.00000251, -0.00000241, -0.00001570, 0.00001588, 0.00000801, -0.00001028, -0.00001491, 0.00000286, 0.00000755, -0.00000491, -0.00000635, -0.00000023, -0.00000387, 0.00000753, -0.00000541, 0.00000545, -0.00000145, -0.00001516, 0.00001107, -0.00001290, 0.00000462, -0.00000119, 0.00000776, 0.00001366, -0.00001408, -0.00000207, -0.00001513, 0.00001079, 0.00000514, -0.00000244, 0.00000109, -0.00000135, 0.00000624, -0.00000895, 0.00000347, -0.00000119, 0.00000322, -0.00000230, -0.00000002, 0.00000224, -0.00001017, 0.00000798, -0.00000186, -0.00000543, 0.00000526, 0.00000998, -0.00000364, -0.00000951, 0.00000619, -0.00000393, -0.00000993, 0.00000515, 0.00001116, 0.00000572, -0.00000880, -0.00000666, 0.00000628, -0.00000927, 0.00000356, 0.00001037, -0.00000126, 0.00000136, -0.00000709, -0.00000042, 0.00000078, 0.00000325, 0.00000291, -0.00000233, 0.00000401, 0.00000540, -0.00000237, -0.00001406, -0.00000443, 0.00001785, 0.00000701, -0.00001086, -0.00000093, -0.00000680, -0.00000392, -0.00000155, -0.00000264, 0.00000080, -0.00000345, 0.00000599, -0.00000977, -0.00000681, 0.00000362, -0.00000105, -0.00000037, -0.00001086, -0.00000198, 0.00000524, -0.00000249, -0.00000502, -0.00000515, -0.00000498, -0.00000420, -0.00003584, -0.00005158, 0.00003345, },
		{ 0.00003737, 0.00008126, 0.00004217, -0.00006692, -0.00003418, -0.00000772, -0.00000969, -0.00000633, -0.00000764, -0.00000175, -0.00000650, -0.00000271, -0.00000625, -0.00000232, -0.00001036, -0.00000302, -0.00002131, -0.00002246, 0.00000488, 0.00002699, 0.00003451, -0.00000504, 0.00004039, 0.00002414, -0.00000259, 0.00000385, -0.00003296, 0.00000541, -0.00000905, -0.00001125, 0.00001798, 0.00000353, -0.00001078, -0.00001419, 0.00000548, 0.00000616, -0.00000670, -0.00000415, -0.00000242, -0.00000122, 0.00000642, -0.00000655, 0.00000862, -0.00000761, -0.00000977, 0.00000852, -0.00001294, 0.00000669, -0.00000457, 0.00001487, 0.00000718, -0.00001414, -0.00000242, -0.00001391, 0.00001451, 0.00000108, -0.00000087, -0.00000004, 0.00000097, 0.00000378, -0.00000868, 0.00000447, -0.00000207, 0.00000453, -0.00000360, 0.00000171, -0.00000043, -0.00000814, 0.00000789, -0.00000391, -0.00000309, 0.00000668, 0.00000896, -0.00000752, -0.00000620, 0.00000649, -0.00000746, -0.00000703, 0.00000740, 0.00001005, 0.00000387, -0.00001029, -0.00000366, 0.00000509, -0.00001029, 0.00000727, 0.00000772, -0.00000107, 0.00000065, -0.00000751, 0.00000107, 0.00000026, 0.00000453, 0.00000084, -0.00000142, 0.00000564, 0.00000347, -0.00000397, -0.00001529, -0.00000025, 0.00002017, 0.00000170, -0.00001022, -0.00000005, -0.00000749, -0.00000261, -0.00000299, -0.00000083, -0.00000040, -0.00000218, 0.00000580, -0.00001231, -0.00000372, 0.00000254, -0.00000021, -0.00000151, -0.00001180, 0.00000034, 0.00000422, -0.00000214, -0.00000544, -0.00000586, -0.00000331, -0.00000767, -0.00004298, -0.00003911, 0.00005092, 0.00007678, },
		{ 0.00008159, 0.00002463, -0.00006656, -0.00002719, -0.00000754, -0.00000895, -0.00000719, -0.00000590, -0.00000285, -0.00000576, -0.00000334, -0.00000571, -0.00000287, -0.00001083, -0.00000282, -0.00002475, -0.00002098, 0.00000768, 0.00003339, 0.00002873, -0.00000084, 0.00004692, 0.00001595, 0.00000061, -0.00000538, -0.00003094, 0.00001068, -0.00001577, -0.00000469, 0.00001742, 0.00000011, -0.00001121, -0.00001324, 0.00000825, 0.00000358, -0.00000704, -0.00000298, -0.00000418, 0.00000200, 0.00000368, -0.00000552, 0.00000959, -0.00001246, -0.00000404, 0.00000439, -0.00001074, 0.00000617, -0.00000523, 0.00002064, -0.00000066, -0.00001218, -0.00000391, -0.00001103, 0.00001653, -0.00000262, 0.00000095, -0.00000108, 0.00000315, 0.00000074, -0.00000736, 0.00000460, -0.00000226, 0.00000540, -0.00000461, 0.00000302, -0.00000288, -0.00000572, 0.00000678, -0.00000487, -0.00000084, 0.00000795, 0.00000682, -0.00001040, -0.00000209, 0.00000495, -0.00000984, -0.00000344, 0.00000860, 0.00000880, 0.00000195, -0.00001112, -0.00000062, 0.00000293, -0.00000999, 0.00001024, 0.00000467, -0.00000017, -0.00000075, -0.00000713, 0.00000209, 0.00000004, 0.00000546, -0.00000150, 0.00000069, 0.00000608, 0.00000157, -0.00000549, -0.00001618, 0.00000517, 0.00002041, -0.00000347, -0.00000830, -0.00000011, -0.00000753, -0.00000198, -0.00000395, 0.00000104, -0.00000211, -0.00000003, 0.00000430, -0.00001379, -0.00000078, 0.00000104, 0.00000122, -0.00000360, -0.00001176, 0.00000224, 0.00000321, -0.00000145, -0.00000669, -0.00000564, -0.00000198, -0.00001305, -0.00004714, -0.00002325, 0.00006284, 0.00006504, 0.00003712, },
		{ 0.00000875, -0.00006410, -0.00002131, -0.00000734, -0.00000879, -0.00000736, -0.00000461, -0.00000388, -0.00000492, -0.00000417, -0.00000482, -0.00000402, -0.00001067, -0.00000325, -0.00002806, -0.00001935, 0.00001182, 0.00003851, 0.00002251, 0.00000628, 0.00005000, 0.00000914, 0.00000246, -0.00001541, -0.00002478, 0.00001221, -0.00002089, 0.00000266, 0.00001461, -0.00000187, -0.00001208, -0.00001151, 0.00001046, 0.00000039, -0.00000610, -0.00000316, -0.00000475, 0.00000476, 0.00000021, -0.00000279, 0.00000819, -0.00001517, 0.00000081, -0.00000015, -0.00000725, 0.00000352, -0.00000277, 0.00002371, -0.00000844, -0.00000888, -0.00000624, -0.00000664, 0.00001654, -0.00000540, 0.00000268, -0.00000184, 0.00000488, -0.00000248, -0.00000524, 0.00000392, -0.00000170, 0.00000570, -0.00000526, 0.00000381, -0.00000483, -0.00000337, 0.00000505, -0.00000481, 0.00000119, 0.00000894, 0.00000361, -0.00001171, 0.00000196, 0.00000190, -0.00001062, 0.00000023, 0.00000886, 0.00000754, 0.00000001, -0.00001118, 0.00000208, 0.00000015, -0.00000837, 0.00001207, 0.00000177, 0.00000111, -0.00000269, -0.00000592, 0.00000248, 0.00000034, 0.00000567, -0.00000354, 0.00000362, 0.00000516, 0.00000017, -0.00000732, -0.00001616, 0.00001126, 0.00001832, -0.00000758, -0.00000576, -0.00000092, -0.00000692, -0.00000216, -0.00000406, 0.00000243, -0.00000383, 0.00000265, 0.00000154, -0.00001385, 0.00000145, -0.00000035, 0.00000285, -0.00000648, -0.00001060, 0.00000339, 0.00000264, -0.00000092, -0.00000850, -0.00000427, -0.00000182, -0.00001954, -0.00004746, -0.00000594, 0.00006904, 0.00005342, 0.00003945, 0.00007881, },
		{ -0.00006022, -0.00001632, -0.00000733, -0.00000891, -0.00000695, -0.00000388, -0.00000463, -0.00000422, -0.00000496, -0.00000382, -0.00000561, -0.00000993, -0.00000441, -0.00003118, -0.00001737, 0.00001727, 0.00004179, 0.00001700, 0.00001531, 0.00004979, 0.00000407, 0.00000188, -0.00002426, -0.00001588, 0.00000985, -0.00002318, 0.00000932, 0.00001045, -0.00000256, -0.00001357, -0.00000867, 0.00001147, -0.00000261, -0.00000447, -0.00000451, -0.00000379, 0.00000625, -0.00000295, 0.00000074, 0.00000487, -0.00001548, 0.00000393, -0.00000397, -0.00000363, -0.00000021, 0.00000242, 0.00002332, -0.00001484, -0.00000518, -0.00000884, -0.00000122, 0.00001455, -0.00000687, 0.00000400, -0.00000216, 0.00000588, -0.00000541, -0.00000269, 0.00000261, -0.00000041, 0.00000535, -0.00000552, 0.00000410, -0.00000616, -0.00000144, 0.00000317, -0.00000396, 0.00000299, 0.00000944, -0.00000034, -0.00001115, 0.00000509, -0.00000200, -0.00000970, 0.00000344, 0.00000840, 0.00000632, -0.00000186, -0.00001048, 0.00000413, -0.00000282, -0.00000564, 0.00001254, -0.00000054, 0.00000235, -0.00000485, -0.00000404, 0.00000223, 0.00000122, 0.00000490, -0.00000469, 0.00000671, 0.00000306, -0.00000052, -0.00000973, -0.00001463, 0.00001716, 0.00001413, -0.00001001, -0.00000332, -0.00000208, -0.00000592, -0.00000304, -0.00000315, 0.00000290, -0.00000498, 0.00000531, -0.00000214, -0.00001241, 0.00000253, -0.00000109, 0.00000414, -0.00000974, -0.00000842, 0.00000364, 0.00000287, -0.00000112, -0.00001029, -0.00000190, -0.00000347, -0.00002594, -0.00004378, 0.00001094, 0.00007010, 0.00004327, 0.00004321, 0.00007330, -0.00000498, },
		{ -0.00001210, -0.00000764, -0.00000902, -0.00000617, -0.00000368, -0.00000501, -0.00000381, -0.00000552, -0.00000298, -0.00000738, -0.00000871, -0.00000633, -0.00003410, -0.00001475, 0.00002374, 0.00004298, 0.00001313, 0.00002509, 0.00004692, 0.00000048, -0.00000150, -0.00003021, -0.00000614, 0.00000435, -0.00002214, 0.00001408, 0.00000606, -0.00000252, -0.00001541, -0.00000478, 0.00001091, -0.00000470, -0.00000290, -0.00000644, -0.00000147, 0.00000606, -0.00000495, 0.00000403, 0.00000048, -0.00001371, 0.00000503, -0.00000632, -0.00000095, -0.00000372, 0.00000933, 0.00001943, -0.00001887, -0.00000200, -0.00001095, 0.00000449, 0.00001092, -0.00000687, 0.00000467, -0.00000196, 0.00000597, -0.00000769, -0.00000012, 0.00000098, 0.00000146, 0.00000436, -0.00000539, 0.00000398, -0.00000689, -0.00000012, 0.00000155, -0.00000261, 0.00000459, 0.00000916, -0.00000450, -0.00000879, 0.00000663, -0.00000591, -0.00000732, 0.00000580, 0.00000750, 0.00000514, -0.00000354, -0.00000914, 0.00000532, -0.00000552, -0.00000221, 0.00001167, -0.00000192, 0.00000314, -0.00000683, -0.00000179, 0.00000149, 0.00000257, 0.00000311, -0.00000448, 0.00000925, 0.00000026, -0.00000058, -0.00001273, -0.00001116, 0.00002184, 0.00000852, -0.00001053, -0.00000159, -0.00000309, -0.00000496, -0.00000424, -0.00000137, 0.00000225, -0.00000510, 0.00000726, -0.00000610, -0.00000976, 0.00000233, -0.00000078, 0.00000451, -0.00001279, -0.00000563, 0.00000314, 0.00000399, -0.00000252, -0.00001132, 0.00000085, -0.00000714, -0.00003099, -0.00003665, 0.00002591, 0.00006705, 0.00003533, 0.00004734, 0.00006565, -0.00001639, -0.00005539, },
		{ -0.00000831, -0.00000889, -0.00000530, -0.00000385, -0.00000505, -0.00000372, -0.00000574, -0.00000249, -0.00000909, -0.00000722, -0.00000900, -0.00003678, -0.00001116, 0.00003073, 0.00004218, 0.00001152, 0.00003451, 0.00004215, -0.00000235, -0.00000723, -0.00003221, 0.00000247, -0.00000289, -0.00001806, 0.00001627, 0.00000243, -0.00000250, -0.00001704, -0.00000030, 0.00000883, -0.00000545, -0.00000214, -0.00000816, 0.00000163, 0.00000429, -0.00000533, 0.00000620, -0.00000394, -0.00001062, 0.00000432, -0.00000697, 0.00000008, -0.00000573, 0.00001649, 0.00001268, -0.00002014, -0.00000003, -0.00001189, 0.00000964, 0.00000627, -0.00000550, 0.00000460, -0.00000122, 0.00000510, -0.00000901, 0.00000208, -0.00000064, 0.00000369, 0.00000277, -0.00000488, 0.00000355, -0.00000716, 0.00000058, 0.00000046, -0.00000102, 0.00000602, 0.00000788, -0.00000815, -0.00000510, 0.00000623, -0.00000899, -0.00000397, 0.00000717, 0.00000636, 0.00000400, -0.00000492, -0.00000737, 0.00000559, -0.00000755, 0.00000145, 0.00000967, -0.00000224, 0.00000316, -0.00000824, 0.00000047, 0.00000054, 0.00000410, 0.00000052, -0.00000271, 0.00001053, -0.00000257, -0.00000040, -0.00001599, -0.00000568, 0.00002438, 0.00000252, -0.00000941, -0.00000086, -0.00000356, -0.00000448, -0.00000523, 0.00000087, 0.00000060, -0.00000394, 0.00000795, -0.00000959, -0.00000647, 0.00000108, 0.00000070, 0.00000353, -0.00001499, -0.00000277, 0.00000230, 0.00000577, -0.00000530, -0.00001087, 0.00000304, -0.00001243, -0.00003368, -0.00002710, 0.00003804, 0.00006109, 0.00002989, 0.00005103, 0.00005649, -0.00002550, -0.00004994, -0.00000862, },
		{ -0.00000838, -0.00000456, -0.00000420, -0.00000485, -0.00000389, -0.00000561, -0.00000247, -0.00001051, -0.00000567, -0.00001241, -0.00003910, -0.00000635, 0.00003765, 0.00003980, 0.00001241, 0.00004272, 0.00003618, -0.00000532, -0.00001422, -0.00003009, 0.00000839, -0.00001017, -0.00001191, 0.00001586, 0.00000018, -0.00000313, -0.00001780, 0.00000405, 0.00000569, -0.00000484, -0.00000261, -0.00000892, 0.00000467, 0.00000149, -0.00000413, 0.00000673, -0.00000747, -0.00000715, 0.00000246, -0.00000620, -0.00000069, -0.00000539, 0.00002238, 0.00000426, -0.00001882, 0.00000043, -0.00001121, 0.00001343, 0.00000141, -0.00000313, 0.00000386, -0.00000005, 0.00000334, -0.00000924, 0.00000360, -0.00000190, 0.00000596, 0.00000071, -0.00000401, 0.00000290, -0.00000717, 0.00000079, -0.00000001, 0.00000063, 0.00000722, 0.00000550, -0.00001064, -0.00000084, 0.00000399, -0.00001063, -0.00000029, 0.00000759, 0.00000514, 0.00000293, -0.00000592, -0.00000539, 0.00000503, -0.00000866, 0.00000483, 0.00000695, -0.00000160, 0.00000230, -0.00000879, 0.00000236, -0.00000030, 0.00000543, -0.00000243, 0.00000048, 0.00001012, -0.00000477, -0.00000053, -0.00001887, 0.00000143, 0.00002418, -0.00000277, -0.00000728, -0.00000104, -0.00000334, -0.00000478, -0.00000547, 0.00000296, -0.00000163, -0.00000162, 0.00000706, -0.00001194, -0.00000328, -0.00000072, 0.00000312, 0.00000107, -0.00001586, -0.00000045, 0.00000173, 0.00000761, -0.00000923, -0.00000855, 0.00000365, -0.00001843, -0.00003352, -0.00001635, 0.00004694, 0.00005336, 0.00002688, 0.00005370, 0.00004645, -0.00003241, -0.00004413, -0.00000592, -0.00000924, },
		{ -0.00000411, -0.00000455, -0.00000457, -0.00000421, -0.00000522, -0.00000295, -0.00001147, -0.00000427, -0.00001651, -0.00004087, -0.00000016, 0.00004392, 0.00003643, 0.00001573, 0.00004916, 0.00002945, -0.00000912, -0.00002097, -0.00002458, 0.00001080, -0.00001596, -0.00000503, 0.00001341, -0.00000058, -0.00000476, -0.00001718, 0.00000752, 0.00000223, -0.00000327, -0.00000435, -0.00000822, 0.00000683, -0.00000156, -0.00000180, 0.00000558, -0.00000953, -0.00000417, 0.00000027, -0.00000470, -0.00000281, -0.00000240, 0.00002575, -0.00000441, -0.00001557, -0.00000050, -0.00000880, 0.00001527, -0.00000286, -0.00000026, 0.00000265, 0.00000135, 0.00000091, -0.00000838, 0.00000425, -0.00000248, 0.00000795, -0.00000164, -0.00000280, 0.00000208, -0.00000709, 0.00000076, 0.00000013, 0.00000222, 0.00000803, 0.00000218, -0.00001145, 0.00000310, 0.00000039, -0.00001051, 0.00000313, 0.00000723, 0.00000392, 0.00000200, -0.00000654, -0.00000344, 0.00000381, -0.00000875, 0.00000752, 0.00000398, -0.00000026, 0.00000060, -0.00000831, 0.00000360, -0.00000075, 0.00000615, -0.00000511, 0.00000462, 0.00000791, -0.00000583, -0.00000152, -0.00002055, 0.00000932, 0.00002119, -0.00000644, -0.00000498, -0.00000174, -0.00000262, -0.00000589, -0.00000463, 0.00000428, -0.00000378, 0.00000139, 0.00000466, -0.00001271, -0.00000089, -0.00000237, 0.00000591, -0.00000265, -0.00001521, 0.00000091, 0.00000202, 0.00000873, -0.00001361, -0.00000444, 0.00000190, -0.00002388, -0.00003064, -0.00000550, 0.00005267, 0.00004481, 0.00002605, 0.00005499, 0.00003606, -0.00003729, -0.00003817, -0.00000404, -0.00001027, -0.00000749, },
		{ -0.00000475, -0.00000433, -0.00000453, -0.00000471, -0.00000385, -0.00001189, -0.00000324, -0.00002126, -0.00004186, 0.00000738, 0.00004904, 0.00003276, 0.00002111, 0.00005354, 0.00002205, -0.00001406, -0.00002602, -0.00001705, 0.00000974, -0.00001921, 0.00000123, 0.00000989, -0.00000021, -0.00000731, -0.00001499, 0.00000949, -0.00000078, -0.00000142, -0.00000694, -0.00000601, 0.00000754, -0.00000403, 0.00000090, 0.00000316, -0.00000998, -0.00000224, -0.00000151, -0.00000329, -0.00000541, 0.00000283, 0.00002587, -0.00001196, -0.00001135, -0.00000229, -0.00000497, 0.00001496, -0.00000588, 0.00000252, 0.00000129, 0.00000270, -0.00000187, -0.00000661, 0.00000399, -0.00000217, 0.00000933, -0.00000404, -0.00000130, 0.00000108, -0.00000702, 0.00000073, 0.00000073, 0.00000373, 0.00000820, -0.00000170, -0.00001039, 0.00000587, -0.00000379, -0.00000869, 0.00000582, 0.00000634, 0.00000274, 0.00000130, -0.00000685, -0.00000166, 0.00000219, -0.00000788, 0.00000923, 0.00000123, 0.00000140, -0.00000171, -0.00000684, 0.00000403, -0.00000061, 0.00000595, -0.00000689, 0.00000897, 0.00000421, -0.00000558, -0.00000373, -0.00002024, 0.00001684, 0.00001591, -0.00000799, -0.00000328, -0.00000238, -0.00000183, -0.00000752, -0.00000264, 0.00000438, -0.00000517, 0.00000440, 0.00000118, -0.00001185, 0.00000022, -0.00000317, 0.00000836, -0.00000707, -0.00001321, 0.00000119, 0.00000355, 0.00000833, -0.00001743, 0.00000079, -0.00000251, -0.00002749, -0.00002574, 0.00000468, 0.00005549, 0.00003619, 0.00002704, 0.00005469, 0.00002582, -0.00004034, -0.00003224, -0.00000300, -0.00001118, -0.00000632, -0.00000403, },
		},
		{ { -0.00002178, -0.00001697, -0.00005328, -0.00000631, 0.00006686, 0.00003097, 0.00004911, 0.00006673, -0.00001239, -0.00003443, -0.00002512, 0.00000047, -0.00001535, -0.00001627, 0.00000655, 0.00000028, -0.00000502, -0.00001209, 0.00000934, -0.00000959, -0.00000541, -0.00000975, 0.00000791, 0.00000711, -0.00000684, -0.00000361, 0.00000374, 0.00000623, 0.00000423, -0.00001049, -0.00000239, 0.00000339, -0.00000030, 0.00000522, 0.00000098, -0.00000051, -0.00000058, 0.00000052, -0.00000671, -0.00000604, 0.00001207, 0.00000060, -0.00000405, 0.00000922, -0.00001139, 0.00000600, -0.00001043, 0.00000231, 0.00000120, 0.00000291, 0.00000099, -0.00000264, -0.00000303, -0.00000136, 0.00000771, -0.00000147, 0.00000237, 0.00000154, -0.00000420, -0.00001074, 0.00000850, 0.00000034, -0.00000039, 0.00000114, -0.00000219, 0.00000114, -0.00000039, 0.00000034, 0.00000850, -0.00001074, -0.00000420, 0.00000154, 0.00000237, -0.00000147, 0.00000771, -0.00000136, -0.00000303, -0.00000264, 0.00000099, 0.00000291, 0.00000120, 0.00000231, -0.00001043, 0.00000600, -0.00001139, 0.00000922, -0.00000405, 0.00000060, 0.00001207, -0.00000604, -0.00000671, 0.00000052, -0.00000058, -0.00000051, 0.00000098, 0.00000522, -0.00000030, 0.00000339, -0.00000239, -0.00001049, 0.00000423, 0.00000623, 0.00000374, -0.00000361, -0.00000684, 0.00000711, 0.00000791, -0.00000975, -0.00000541, -0.00000959, 0.00000934, -0.00001209, -0.00000502, 0.00000028, 0.00000655, -0.00001627, -0.00001535, 0.00000047, -0.00002512, -0.00003443, -0.00001239, 0.00006673, 0.00004911, 0.00003097, 0.00006686, -0.00000631, -0.00005328, -0.00001697, },
		{ -0.00001764, -0.00005616, -0.00000147, 0.00007101, 0.00002973, 0.00005659, 0.00006455, -0.00002069, -0.00003569, -0.00002191, -0.00000002, -0.00001694, -0.00001365, 0.00000521, 0.00000094, -0.00000686, -0.00000991, 0.00000892, -0.00001093, -0.00000569, -0.00000842, 0.00000966, 0.00000434, -0.00000644, -0.00000330, 0.00000477, 0.00000620, 0.00000301, -0.00001038, -0.00000163, 0.00000290, 0.00000008, 0.00000584, 0.00000020, -0.00000050, -0.00000035, -0.00000024, -0.00000733, -0.00000434, 0.00001273, -0.00000176, -0.00000158, 0.00000714, -0.00001064, 0.00000565, -0.00001055, 0.00000304, 0.00000071, 0.00000383, -0.00000038, -0.00000182, -0.00000428, 0.00000077, 0.00000685, -0.00000174, 0.00000298, 0.00000098, -0.00000563, -0.00000908, 0.00000897, -0.00000056, -0.00000011, 0.00000100, -0.00000221, 0.00000132, -0.00000068, 0.00000136, 0.00000766, -0.00001214, -0.00000256, 0.00000186, 0.00000168, -0.00000086, 0.00000810, -0.00000327, -0.00000169, -0.00000358, 0.00000256, 0.00000161, 0.00000210, 0.00000117, -0.00000988, 0.00000594, -0.00001172, 0.00001099, -0.00000655, 0.00000339, 0.00001069, -0.00000735, -0.00000602, 0.00000128, -0.00000100, -0.00000028, 0.00000161, 0.00000449, -0.00000050, 0.00000383, -0.00000323, -0.00001048, 0.00000559, 0.00000585, 0.00000297, -0.00000415, -0.00000681, 0.00000985, 0.00000549, -0.00001048, -0.00000524, -0.00000794, 0.00000912, -0.00001383, -0.00000328, -0.00000039, 0.00000802, -0.00001942, -0.00001281, -0.00000003, -0.00002805, -0.00003196, -0.00000469, 0.00006746, 0.00004226, 0.00003249, 0.00006228, -0.00001042, -0.00005053, -0.00001639, -0.00002180, },
		{ -0.00005916, 0.00000415, 0.00007466, 0.00002891, 0.00006460, 0.00006064, -0.00002927, -0.00003555, -0.00001887, -0.00000120, -0.00001764, -0.00001164, 0.00000416, 0.00000140, -0.00000859, -0.00000747, 0.00000797, -0.00001201, -0.00000604, -0.00000656, 0.00001060, 0.00000179, -0.00000579, -0.00000312, 0.00000600, 0.00000577, 0.00000199, -0.00001022, -0.00000093, 0.00000237, 0.00000065, 0.00000629, -0.00000065, -0.00000029, -0.00000031, -0.00000101, -0.00000781, -0.00000237, 0.00001271, -0.00000357, 0.00000067, 0.00000493, -0.00000959, 0.00000500, -0.00001032, 0.00000340, 0.00000064, 0.00000431, -0.00000145, -0.00000123, -0.00000532, 0.00000298, 0.00000557, -0.00000163, 0.00000346, 0.00000020, -0.00000680, -0.00000725, 0.00000909, -0.00000134, 0.00000016, 0.00000089, -0.00000228, 0.00000154, -0.00000098, 0.00000246, 0.00000644, -0.00001318, -0.00000081, 0.00000192, 0.00000102, 0.00000004, 0.00000797, -0.00000482, -0.00000040, -0.00000453, 0.00000423, 0.00000001, 0.00000338, -0.00000040, -0.00000885, 0.00000541, -0.00001151, 0.00001228, -0.00000887, 0.00000645, 0.00000865, -0.00000817, -0.00000534, 0.00000205, -0.00000161, 0.00000023, 0.00000200, 0.00000376, -0.00000057, 0.00000422, -0.00000417, -0.00001028, 0.00000698, 0.00000513, 0.00000248, -0.00000496, -0.00000621, 0.00001230, 0.00000260, -0.00001067, -0.00000520, -0.00000601, 0.00000822, -0.00001496, -0.00000184, -0.00000085, 0.00000936, -0.00002287, -0.00000937, -0.00000173, -0.00003026, -0.00002860, 0.00000226, 0.00006699, 0.00003613, 0.00003415, 0.00005735, -0.00001382, -0.00004796, -0.00001587, -0.00002186, -0.00001840, },
		{ 0.00001056, 0.00007772, 0.00002871, 0.00007299, 0.00005474, -0.00003770, -0.00003398, -0.00001638, -0.00000273, -0.00001763, -0.00001025, 0.00000348, 0.00000152, -0.00001002, -0.00000497, 0.00000661, -0.00001286, -0.00000634, -0.00000427, 0.00001069, -0.00000034, -0.00000507, -0.00000295, 0.00000732, 0.00000498, 0.00000123, -0.00001006, -0.00000027, 0.00000180, 0.00000143, 0.00000650, -0.00000147, 0.00000003, -0.00000043, -0.00000179, -0.00000806, -0.00000026, 0.00001207, -0.00000476, 0.00000258, 0.00000276, -0.00000839, 0.00000417, -0.00000983, 0.00000348, 0.00000092, 0.00000435, -0.00000216, -0.00000093, -0.00000601, 0.00000511, 0.00000399, -0.00000114, 0.00000373, -0.00000072, -0.00000766, -0.00000535, 0.00000891, -0.00000197, 0.00000041, 0.00000080, -0.00000238, 0.00000179, -0.00000126, 0.00000361, 0.00000485, -0.00001379, 0.00000096, 0.00000171, 0.00000045, 0.00000114, 0.00000732, -0.00000592, 0.00000070, -0.00000534, 0.00000585, -0.00000176, 0.00000496, -0.00000236, -0.00000734, 0.00000436, -0.00001068, 0.00001293, -0.00001077, 0.00000956, 0.00000604, -0.00000848, -0.00000471, 0.00000281, -0.00000240, 0.00000103, 0.00000207, 0.00000311, -0.00000056, 0.00000459, -0.00000524, -0.00000984, 0.00000829, 0.00000416, 0.00000223, -0.00000605, -0.00000493, 0.00001420, -0.00000053, -0.00001040, -0.00000524, -0.00000388, 0.00000667, -0.00001538, -0.00000087, -0.00000089, 0.00001030, -0.00002634, -0.00000522, -0.00000470, -0.00003135, -0.00002477, 0.00000842, 0.00006552, 0.00003074, 0.00003583, 0.00005216, -0.00001654, -0.00004559, -0.00001540, -0.00002196, -0.00001927, -0.00006224, },
		{ 0.00008010, 0.00002932, 0.00008149, 0.00004664, -0.00004548, -0.00003111, -0.00001472, -0.00000425, -0.00001714, -0.00000939, 0.00000320, 0.00000123, -0.00001101, -0.00000260, 0.00000498, -0.00001359, -0.00000646, -0.00000172, 0.00000997, -0.00000192, -0.00000446, -0.00000264, 0.00000860, 0.00000389, 0.00000072, -0.00000992, 0.00000034, 0.00000121, 0.00000240, 0.00000643, -0.00000221, 0.00000043, -0.00000070, -0.00000257, -0.00000802, 0.00000184, 0.00001093, -0.00000535, 0.00000403, 0.00000074, -0.00000715, 0.00000325, -0.00000920, 0.00000338, 0.00000147, 0.00000400, -0.00000252, -0.00000098, -0.00000627, 0.00000700, 0.00000224, -0.00000032, 0.00000373, -0.00000174, -0.00000818, -0.00000346, 0.00000846, -0.00000248, 0.00000066, 0.00000070, -0.00000249, 0.00000207, -0.00000152, 0.00000475, 0.00000293, -0.00001391, 0.00000265, 0.00000127, 0.00000005, 0.00000234, 0.00000621, -0.00000652, 0.00000151, -0.00000592, 0.00000728, -0.00000354, 0.00000671, -0.00000461, -0.00000540, 0.00000281, -0.00000919, 0.00001280, -0.00001205, 0.00001245, 0.00000304, -0.00000829, -0.00000417, 0.00000355, -0.00000334, 0.00000213, 0.00000179, 0.00000262, -0.00000053, 0.00000495, -0.00000644, -0.00000908, 0.00000940, 0.00000305, 0.00000214, -0.00000734, -0.00000295, 0.00001534, -0.00000363, -0.00000980, -0.00000526, -0.00000169, 0.00000456, -0.00001508, -0.00000050, -0.00000033, 0.00001058, -0.00002947, -0.00000067, -0.00000885, -0.00003104, -0.00002088, 0.00001385, 0.00006320, 0.00002616, 0.00003739, 0.00004681, -0.00001864, -0.00004342, -0.00001496, -0.00002211, -0.00002026, -0.00006537, 0.00001780, },
		{ 0.00003100, 0.00008973, 0.00003624, -0.00005204, -0.00002722, -0.00001402, -0.00000546, -0.00001643, -0.00000891, 0.00000324, 0.00000054, -0.00001149, -0.00000050, 0.00000322, -0.00001426, -0.00000627, 0.00000087, 0.00000855, -0.00000287, -0.00000407, -0.00000209, 0.00000970, 0.00000261, 0.00000045, -0.00000982, 0.00000091, 0.00000065, 0.00000353, 0.00000607, -0.00000279, 0.00000083, -0.00000110, -0.00000335, -0.00000766, 0.00000376, 0.00000946, -0.00000539, 0.00000501, -0.00000103, -0.00000597, 0.00000236, -0.00000854, 0.00000322, 0.00000219, 0.00000332, -0.00000255, -0.00000136, -0.00000604, 0.00000852, 0.00000045, 0.00000078, 0.00000344, -0.00000277, -0.00000838, -0.00000166, 0.00000782, -0.00000286, 0.00000091, 0.00000058, -0.00000258, 0.00000233, -0.00000171, 0.00000582, 0.00000074, -0.00001350, 0.00000416, 0.00000065, -0.00000016, 0.00000354, 0.00000470, -0.00000658, 0.00000196, -0.00000615, 0.00000838, -0.00000518, 0.00000848, -0.00000703, -0.00000313, 0.00000082, -0.00000705, 0.00001184, -0.00001252, 0.00001488, -0.00000014, -0.00000767, -0.00000372, 0.00000424, -0.00000435, 0.00000349, 0.00000113, 0.00000236, -0.00000052, 0.00000532, -0.00000778, -0.00000796, 0.00001018, 0.00000194, 0.00000209, -0.00000870, -0.00000032, 0.00001559, -0.00000647, -0.00000906, -0.00000516, 0.00000039, 0.00000207, -0.00001413, -0.00000076, 0.00000096, 0.00000995, -0.00003188, 0.00000387, -0.00001394, -0.00002920, -0.00001732, 0.00001870, 0.00006012, 0.00002244, 0.00003869, 0.00004140, -0.00002020, -0.00004145, -0.00001456, -0.00002230, -0.00002141, -0.00006852, 0.00002588, 0.00008172, },
		{ 0.00009725, 0.00002360, -0.00005684, -0.00002276, -0.00001427, -0.00000615, -0.00001577, -0.00000861, 0.00000348, -0.00000048, -0.00001145, 0.00000123, 0.00000141, -0.00001492, -0.00000565, 0.00000326, 0.00000663, -0.00000322, -0.00000397, -0.00000120, 0.00001049, 0.00000127, 0.00000037, -0.00000977, 0.00000143, 0.00000014, 0.00000477, 0.00000542, -0.00000317, 0.00000120, -0.00000160, -0.00000408, -0.00000696, 0.00000539, 0.00000781, -0.00000499, 0.00000551, -0.00000250, -0.00000492, 0.00000155, -0.00000792, 0.00000310, 0.00000293, 0.00000242, -0.00000231, -0.00000205, -0.00000530, 0.00000957, -0.00000123, 0.00000208, 0.00000284, -0.00000375, -0.00000827, -0.00000002, 0.00000702, -0.00000312, 0.00000117, 0.00000040, -0.00000263, 0.00000257, -0.00000180, 0.00000674, -0.00000164, -0.00001255, 0.00000540, -0.00000010, -0.00000012, 0.00000463, 0.00000291, -0.00000615, 0.00000200, -0.00000596, 0.00000905, -0.00000652, 0.00001011, -0.00000947, -0.00000065, -0.00000149, -0.00000436, 0.00001004, -0.00001206, 0.00001661, -0.00000327, -0.00000674, -0.00000332, 0.00000482, -0.00000536, 0.00000504, 0.00000009, 0.00000238, -0.00000060, 0.00000571, -0.00000925, -0.00000645, 0.00001055, 0.00000099, 0.00000194, -0.00000996, 0.00000281, 0.00001488, -0.00000885, -0.00000831, -0.00000479, 0.00000215, -0.00000059, -0.00001266, -0.00000157, 0.00000300, 0.00000824, -0.00003323, 0.00000792, -0.00001956, -0.00002585, -0.00001442, 0.00002314, 0.00005632, 0.00001963, 0.00003959, 0.00003610, -0.00002131, -0.00003964, -0.00001421, -0.00002250, -0.00002275, -0.00007161, 0.00003483, 0.00008249, 0.00003399, },
		{ 0.00000896, -0.00005944, -0.00001823, -0.00001532, -0.00000623, -0.00001533, -0.00000833, 0.00000378, -0.00000170, -0.00001094, 0.00000254, -0.00000038, -0.00001559, -0.00000452, 0.00000523, 0.00000442, -0.00000306, -0.00000417, 0.00000008, 0.00001085, -0.00000001, 0.00000041, -0.00000974, 0.00000188, -0.00000027, 0.00000606, 0.00000450, -0.00000333, 0.00000147, -0.00000219, -0.00000473, -0.00000596, 0.00000664, 0.00000614, -0.00000427, 0.00000558, -0.00000365, -0.00000402, 0.00000087, -0.00000741, 0.00000311, 0.00000359, 0.00000142, -0.00000191, -0.00000295, -0.00000408, 0.00001007, -0.00000267, 0.00000345, 0.00000195, -0.00000461, -0.00000788, 0.00000143, 0.00000612, -0.00000328, 0.00000144, 0.00000016, -0.00000260, 0.00000276, -0.00000177, 0.00000745, -0.00000408, -0.00001111, 0.00000630, -0.00000090, 0.00000014, 0.00000551, 0.00000096, -0.00000528, 0.00000164, -0.00000534, 0.00000919, -0.00000743, 0.00001144, -0.00001176, 0.00000187, -0.00000398, -0.00000124, 0.00000746, -0.00001060, 0.00001742, -0.00000609, -0.00000564, -0.00000292, 0.00000521, -0.00000627, 0.00000668, -0.00000128, 0.00000271, -0.00000080, 0.00000612, -0.00001083, -0.00000456, 0.00001041, 0.00000030, 0.00000154, -0.00001091, 0.00000623, 0.00001326, -0.00001063, -0.00000770, -0.00000405, 0.00000340, -0.00000314, -0.00001090, -0.00000279, 0.00000570, 0.00000541, -0.00003325, 0.00001103, -0.00002522, -0.00002122, -0.00001237, 0.00002732, 0.00005185, 0.00001780, 0.00003993, 0.00003103, -0.00002209, -0.00003794, -0.00001398, -0.00002265, -0.00002436, -0.00007455, 0.00004459, 0.00008240, 0.00003855, 0.00010347, },
		{ -0.00005954, -0.00001416, -0.00001687, -0.00000571, -0.00001525, -0.00000789, 0.00000398, -0.00000299, -0.00001007, 0.00000346, -0.00000214, -0.00001625, -0.00000287, 0.00000659, 0.00000220, -0.00000254, -0.00000462, 0.00000174, 0.00001070, -0.00000111, 0.00000050, -0.00000971, 0.00000226, -0.00000053, 0.00000732, 0.00000335, -0.00000325, 0.00000161, -0.00000285, -0.00000526, -0.00000469, 0.00000746, 0.00000458, -0.00000336, 0.00000531, -0.00000451, -0.00000328, 0.00000033, -0.00000704, 0.00000332, 0.00000407, 0.00000043, -0.00000143, -0.00000396, -0.00000247, 0.00000999, -0.00000376, 0.00000481, 0.00000081, -0.00000531, -0.00000726, 0.00000265, 0.00000515, -0.00000333, 0.00000171, -0.00000015, -0.00000247, 0.00000287, -0.00000160, 0.00000786, -0.00000649, -0.00000923, 0.00000679, -0.00000166, 0.00000063, 0.00000612, -0.00000103, -0.00000406, 0.00000092, -0.00000429, 0.00000877, -0.00000780, 0.00001232, -0.00001372, 0.00000424, -0.00000645, 0.00000210, 0.00000427, -0.00000818, 0.00001719, -0.00000839, -0.00000453, -0.00000243, 0.00000534, -0.00000695, 0.00000828, -0.00000288, 0.00000331, -0.00000112, 0.00000655, -0.00001249, -0.00000232, 0.00000975, -0.00000003, 0.00000078, -0.00001132, 0.00000965, 0.00001086, -0.00001176, -0.00000730, -0.00000288, 0.00000396, -0.00000533, -0.00000913, -0.00000419, 0.00000887, 0.00000153, -0.00003179, 0.00001278, -0.00003036, -0.00001565, -0.00001120, 0.00003136, 0.00004675, 0.00001697, 0.00003959, 0.00002637, -0.00002269, -0.00003627, -0.00001392, -0.00002268, -0.00002634, -0.00007719, 0.00005508, 0.00008146, 0.00004487, 0.00010774, -0.00000725, },
		{ -0.00001104, -0.00001857, -0.00000472, -0.00001555, -0.00000721, 0.00000395, -0.00000419, -0.00000896, 0.00000402, -0.00000387, -0.00001682, -0.00000074, 0.00000722, 0.00000017, -0.00000182, -0.00000520, 0.00000371, 0.00001003, -0.00000193, 0.00000056, -0.00000964, 0.00000254, -0.00000060, 0.00000849, 0.00000205, -0.00000294, 0.00000160, -0.00000355, -0.00000561, -0.00000325, 0.00000783, 0.00000325, -0.00000238, 0.00000479, -0.00000512, -0.00000268, -0.00000009, -0.00000679, 0.00000374, 0.00000428, -0.00000044, -0.00000099, -0.00000497, -0.00000056, 0.00000936, -0.00000441, 0.00000602, -0.00000053, -0.00000579, -0.00000646, 0.00000364, 0.00000417, -0.00000329, 0.00000197, -0.00000053, -0.00000225, 0.00000289, -0.00000127, 0.00000793, -0.00000871, -0.00000699, 0.00000686, -0.00000231, 0.00000129, 0.00000640, -0.00000293, -0.00000261, -0.00000010, -0.00000286, 0.00000780, -0.00000757, 0.00001263, -0.00001520, 0.00000630, -0.00000870, 0.00000544, 0.00000066, -0.00000489, 0.00001585, -0.00000996, -0.00000358, -0.00000177, 0.00000511, -0.00000728, 0.00000970, -0.00000462, 0.00000415, -0.00000155, 0.00000698, -0.00001415, 0.00000020, 0.00000856, 0.00000006, -0.00000045, -0.00001100, 0.00001278, 0.00000790, -0.00001229, -0.00000710, -0.00000125, 0.00000374, -0.00000692, -0.00000762, -0.00000548, 0.00001222, -0.00000320, -0.00002885, 0.00001288, -0.00003442, -0.00000965, -0.00001084, 0.00003528, 0.00004111, 0.00001713, 0.00003847, 0.00002226, -0.00002324, -0.00003455, -0.00001411, -0.00002249, -0.00002882, -0.00007935, 0.00006617, 0.00007979, 0.00005306, 0.00010937, -0.00002438, -0.00005700, },
		{ -0.00002005, -0.00000348, -0.00001618, -0.00000626, 0.00000363, -0.00000517, -0.00000774, 0.00000430, -0.00000564, -0.00001718, 0.00000175, 0.00000707, -0.00000144, -0.00000111, -0.00000577, 0.00000588, 0.00000885, -0.00000240, 0.00000053, -0.00000952, 0.00000271, -0.00000046, 0.00000949, 0.00000064, -0.00000244, 0.00000142, -0.00000427, -0.00000575, -0.00000172, 0.00000778, 0.00000222, -0.00000146, 0.00000411, -0.00000554, -0.00000217, -0.00000043, -0.00000664, 0.00000437, 0.00000419, -0.00000111, -0.00000069, -0.00000584, 0.00000150, 0.00000823, -0.00000459, 0.00000699, -0.00000199, -0.00000605, -0.00000554, 0.00000437, 0.00000319, -0.00000316, 0.00000221, -0.00000095, -0.00000192, 0.00000283, -0.00000080, 0.00000761, -0.00001063, -0.00000453, 0.00000652, -0.00000280, 0.00000207, 0.00000632, -0.00000462, -0.00000104, -0.00000131, -0.00000113, 0.00000632, -0.00000674, 0.00001231, -0.00001609, 0.00000787, -0.00001054, 0.00000853, -0.00000310, -0.00000095, 0.00001344, -0.00001069, -0.00000292, -0.00000087, 0.00000445, -0.00000714, 0.00001076, -0.00000632, 0.00000515, -0.00000207, 0.00000736, -0.00001574, 0.00000291, 0.00000691, 0.00000057, -0.00000218, -0.00000979, 0.00001529, 0.00000467, -0.00001234, -0.00000706, 0.00000074, 0.00000270, -0.00000774, -0.00000662, -0.00000634, 0.00001541, -0.00000844, -0.00002459, 0.00001117, -0.00003693, -0.00000374, -0.00001104, 0.00003904, 0.00003505, 0.00001822, 0.00003651, 0.00001880, -0.00002386, -0.00003267, -0.00001465, -0.00002198, -0.00003196, -0.00008080, 0.00007764, 0.00007759, 0.00006310, 0.00010772, -0.00004162, -0.00005195, -0.00000926, },
		{ -0.00000225, -0.00001700, -0.00000508, 0.00000299, -0.00000586, -0.00000654, 0.00000438, -0.00000748, -0.00001722, 0.00000443, 0.00000617, -0.00000251, -0.00000057, -0.00000618, 0.00000811, 0.00000722, -0.00000251, 0.00000035, -0.00000930, 0.00000277, -0.00000008, 0.00001025, -0.00000078, -0.00000177, 0.00000106, -0.00000498, -0.00000565, -0.00000020, 0.00000738, 0.00000152, -0.00000068, 0.00000336, -0.00000582, -0.00000171, -0.00000075, -0.00000652, 0.00000516, 0.00000379, -0.00000154, -0.00000059, -0.00000647, 0.00000355, 0.00000672, -0.00000428, 0.00000763, -0.00000349, -0.00000606, -0.00000455, 0.00000487, 0.00000225, -0.00000295, 0.00000240, -0.00000141, -0.00000150, 0.00000268, -0.00000018, 0.00000688, -0.00001214, -0.00000197, 0.00000580, -0.00000305, 0.00000291, 0.00000588, -0.00000601, 0.00000052, -0.00000261, 0.00000079, 0.00000443, -0.00000535, 0.00001135, -0.00001630, 0.00000885, -0.00001178, 0.00001112, -0.00000671, 0.00000338, 0.00001009, -0.00001052, -0.00000266, 0.00000032, 0.00000331, -0.00000643, 0.00001132, -0.00000784, 0.00000618, -0.00000260, 0.00000764, -0.00001718, 0.00000567, 0.00000490, 0.00000144, -0.00000436, -0.00000760, 0.00001690, 0.00000147, -0.00001208, -0.00000703, 0.00000295, 0.00000090, -0.00000770, -0.00000634, -0.00000650, 0.00001805, -0.00001377, -0.00001934, 0.00000770, -0.00003753, 0.00000154, -0.00001150, 0.00004248, 0.00002878, 0.00002011, 0.00003374, 0.00001607, -0.00002467, -0.00003058, -0.00001560, -0.00002105, -0.00003592, -0.00008126, 0.00008923, 0.00007520, 0.00007477, 0.00010224, -0.00005805, -0.00004472, -0.00000906, -0.00002094, },
		{ -0.00001783, -0.00000377, 0.00000207, -0.00000621, -0.00000543, 0.00000432, -0.00000943, -0.00001678, 0.00000708, 0.00000465, -0.00000297, -0.00000034, -0.00000628, 0.00001023, 0.00000527, -0.00000226, 0.00000001, -0.00000899, 0.00000273, 0.00000054, 0.00001071, -0.00000214, -0.00000098, 0.00000051, -0.00000563, -0.00000530, 0.00000121, 0.00000672, 0.00000113, -0.00000008, 0.00000260, -0.00000602, -0.00000126, -0.00000109, -0.00000636, 0.00000603, 0.00000312, -0.00000169, -0.00000074, -0.00000675, 0.00000546, 0.00000494, -0.00000352, 0.00000790, -0.00000495, -0.00000586, -0.00000355, 0.00000515, 0.00000138, -0.00000266, 0.00000253, -0.00000187, -0.00000102, 0.00000248, 0.00000052, 0.00000575, -0.00001313, 0.00000054, 0.00000476, -0.00000305, 0.00000373, 0.00000511, -0.00000703, 0.00000195, -0.00000387, 0.00000276, 0.00000226, -0.00000348, 0.00000977, -0.00001582, 0.00000915, -0.00001231, 0.00001301, -0.00000986, 0.00000778, 0.00000599, -0.00000948, -0.00000284, 0.00000184, 0.00000168, -0.00000507, 0.00001124, -0.00000899, 0.00000711, -0.00000307, 0.00000775, -0.00001837, 0.00000835, 0.00000265, 0.00000257, -0.00000691, -0.00000444, 0.00001740, -0.00000137, -0.00001174, -0.00000683, 0.00000516, -0.00000147, -0.00000682, -0.00000689, -0.00000573, 0.00001978, -0.00001870, -0.00001355, 0.00000271, -0.00003606, 0.00000569, -0.00001186, 0.00004542, 0.00002253, 0.00002260, 0.00003022, 0.00001408, -0.00002570, -0.00002823, -0.00001701, -0.00001962, -0.00004085, -0.00008047, 0.00010061, 0.00007304, 0.00008767, 0.00009258, -0.00007269, -0.00003587, -0.00001048, -0.00002097, -0.00000130, },
		{ -0.00000248, 0.00000097, -0.00000626, -0.00000446, 0.00000415, -0.00001151, -0.00001579, 0.00000948, 0.00000267, -0.00000284, -0.00000051, -0.00000595, 0.00001207, 0.00000312, -0.00000169, -0.00000051, -0.00000857, 0.00000260, 0.00000140, 0.00001084, -0.00000337, -0.00000013, -0.00000022, -0.00000619, -0.00000471, 0.00000242, 0.00000589, 0.00000101, 0.00000030, 0.00000189, -0.00000616, -0.00000078, -0.00000152, -0.00000609, 0.00000690, 0.00000224, -0.00000161, -0.00000115, -0.00000663, 0.00000709, 0.00000306, -0.00000239, 0.00000775, -0.00000629, -0.00000544, -0.00000257, 0.00000523, 0.00000057, -0.00000230, 0.00000257, -0.00000230, -0.00000050, 0.00000224, 0.00000127, 0.00000426, -0.00001355, 0.00000287, 0.00000351, -0.00000279, 0.00000446, 0.00000405, -0.00000764, 0.00000317, -0.00000500, 0.00000465, -0.00000006, -0.00000126, 0.00000767, -0.00001467, 0.00000877, -0.00001205, 0.00001401, -0.00001228, 0.00001190, 0.00000144, -0.00000768, -0.00000346, 0.00000366, -0.00000044, -0.00000304, 0.00001041, -0.00000963, 0.00000782, -0.00000339, 0.00000762, -0.00001920, 0.00001080, 0.00000032, 0.00000381, -0.00000966, -0.00000041, 0.00001665, -0.00000358, -0.00001155, -0.00000626, 0.00000709, -0.00000416, -0.00000524, -0.00000824, -0.00000391, 0.00002031, -0.00002278, -0.00000778, -0.00000339, -0.00003253, 0.00000837, -0.00001176, 0.00004761, 0.00001659, 0.00002544, 0.00002610, 0.00001278, -0.00002697, -0.00002563, -0.00001888, -0.00001767, -0.00004687, -0.00007815, 0.00011146, 0.00007162, 0.00010117, 0.00007859, -0.00008461, -0.00002611, -0.00001340, -0.00001998, -0.00000086, -0.00001848, },
		{ -0.00000019, -0.00000606, -0.00000363, 0.00000386, -0.00001367, -0.00001416, 0.00001139, 0.00000048, -0.00000222, -0.00000107, -0.00000509, 0.00001346, 0.00000094, -0.00000088, -0.00000118, -0.00000806, 0.00000242, 0.00000246, 0.00001061, -0.00000440, 0.00000072, -0.00000111, -0.00000662, -0.00000390, 0.00000338, 0.00000501, 0.00000111, 0.00000048, 0.00000124, -0.00000628, -0.00000027, -0.00000205, -0.00000564, 0.00000767, 0.00000124, -0.00000133, -0.00000180, -0.00000610, 0.00000831, 0.00000120, -0.00000097, 0.00000718, -0.00000745, -0.00000486, -0.00000166, 0.00000513, -0.00000013, -0.00000189, 0.00000253, -0.00000267, 0.00000004, 0.00000201, 0.00000199, 0.00000248, -0.00001335, 0.00000489, 0.00000213, -0.00000229, 0.00000504, 0.00000278, -0.00000783, 0.00000410, -0.00000590, 0.00000634, -0.00000236, 0.00000116, 0.00000517, -0.00001296, 0.00000774, -0.00001099, 0.00001403, -0.00001373, 0.00001539, -0.00000323, -0.00000530, -0.00000443, 0.00000570, -0.00000296, -0.00000036, 0.00000877, -0.00000961, 0.00000815, -0.00000345, 0.00000717, -0.00001958, 0.00001287, -0.00000190, 0.00000498, -0.00001236, 0.00000430, 0.00001463, -0.00000497, -0.00001172, -0.00000510, 0.00000845, -0.00000685, -0.00000321, -0.00001027, -0.00000106, 0.00001945, -0.00002557, -0.00000260, -0.00001003, -0.00002718, 0.00000937, -0.00001091, 0.00004883, 0.00001123, 0.00002835, 0.00002160, 0.00001204, -0.00002841, -0.00002286, -0.00002114, -0.00001524, -0.00005401, -0.00007410, 0.00012148, 0.00007149, 0.00011443, 0.00006046, -0.00009299, -0.00001628, -0.00001748, -0.00001796, -0.00000108, -0.00001878, -0.00000136, },
		{ -0.00000572, -0.00000289, 0.00000341, -0.00001585, -0.00001191, 0.00001265, -0.00000168, -0.00000125, -0.00000199, -0.00000370, 0.00001429, -0.00000112, 0.00000007, -0.00000198, -0.00000748, 0.00000221, 0.00000368, 0.00001002, -0.00000518, 0.00000150, -0.00000212, -0.00000688, -0.00000293, 0.00000404, 0.00000416, 0.00000135, 0.00000048, 0.00000067, -0.00000636, 0.00000026, -0.00000269, -0.00000497, 0.00000826, 0.00000021, -0.00000093, -0.00000261, -0.00000515, 0.00000906, -0.00000047, 0.00000061, 0.00000624, -0.00000837, -0.00000413, -0.00000084, 0.00000487, -0.00000073, -0.00000144, 0.00000238, -0.00000296, 0.00000055, 0.00000182, 0.00000262, 0.00000051, -0.00001257, 0.00000650, 0.00000074, -0.00000158, 0.00000543, 0.00000137, -0.00000762, 0.00000471, -0.00000650, 0.00000772, -0.00000449, 0.00000360, 0.00000243, -0.00001081, 0.00000618, -0.00000920, 0.00001305, -0.00001405, 0.00001794, -0.00000766, -0.00000259, -0.00000564, 0.00000785, -0.00000575, 0.00000287, 0.00000631, -0.00000885, 0.00000799, -0.00000316, 0.00000634, -0.00001943, 0.00001442, -0.00000385, 0.00000589, -0.00001476, 0.00000941, 0.00001145, -0.00000541, -0.00001240, -0.00000319, 0.00000895, -0.00000917, -0.00000106, -0.00001273, 0.00000267, 0.00001717, -0.00002679, 0.00000144, -0.00001657, -0.00002046, 0.00000871, -0.00000912, 0.00004890, 0.00000674, 0.00003104, 0.00001693, 0.00001173, -0.00002993, -0.00002003, -0.00002368, -0.00001243, -0.00006224, -0.00006818, 0.00013044, 0.00007320, 0.00012646, 0.00003866, -0.00009724, -0.00000722, -0.00002226, -0.00001503, -0.00000201, -0.00001860, -0.00000055, -0.00000125, },
		{ -0.00000218, 0.00000274, -0.00001791, -0.00000910, 0.00001312, -0.00000357, -0.00000013, -0.00000315, -0.00000181, 0.00001448, -0.00000291, 0.00000106, -0.00000285, -0.00000686, 0.00000204, 0.00000499, 0.00000910, -0.00000567, 0.00000217, -0.00000322, -0.00000694, -0.00000186, 0.00000441, 0.00000341, 0.00000166, 0.00000034, 0.00000016, -0.00000640, 0.00000081, -0.00000343, -0.00000405, 0.00000858, -0.00000075, -0.00000050, -0.00000350, -0.00000386, 0.00000930, -0.00000185, 0.00000223, 0.00000496, -0.00000901, -0.00000332, -0.00000014, 0.00000450, -0.00000121, -0.00000097, 0.00000214, -0.00000314, 0.00000101, 0.00000170, 0.00000308, -0.00000154, -0.00001123, 0.00000762, -0.00000056, -0.00000072, 0.00000560, -0.00000011, -0.00000704, 0.00000497, -0.00000676, 0.00000871, -0.00000632, 0.00000592, -0.00000038, -0.00000838, 0.00000423, -0.00000682, 0.00001114, -0.00001317, 0.00001931, -0.00001153, 0.00000019, -0.00000692, 0.00000995, -0.00000865, 0.00000651, 0.00000312, -0.00000732, 0.00000727, -0.00000244, 0.00000508, -0.00001869, 0.00001535, -0.00000538, 0.00000637, -0.00001658, 0.00001454, 0.00000732, -0.00000491, -0.00001365, -0.00000044, 0.00000836, -0.00001076, 0.00000083, -0.00001526, 0.00000701, 0.00001361, -0.00002630, 0.00000387, -0.00002232, -0.00001292, 0.00000654, -0.00000633, 0.00004773, 0.00000332, 0.00003323, 0.00001236, 0.00001163, -0.00003138, -0.00001728, -0.00002630, -0.00000944, -0.00007142, -0.00006036, 0.00013824, 0.00007717, 0.00013612, 0.00001403, -0.00009705, 0.00000029, -0.00002716, -0.00001147, -0.00000362, -0.00001789, -0.00000015, -0.00000210, -0.00000536, },
		{ 0.00000174, -0.00001969, -0.00000589, 0.00001279, -0.00000499, 0.00000094, -0.00000438, 0.00000048, 0.00001400, -0.00000434, 0.00000199, -0.00000373, -0.00000623, 0.00000195, 0.00000632, 0.00000787, -0.00000585, 0.00000266, -0.00000437, -0.00000678, -0.00000075, 0.00000449, 0.00000282, 0.00000196, 0.00000012, -0.00000031, -0.00000637, 0.00000131, -0.00000423, -0.00000291, 0.00000859, -0.00000156, -0.00000013, -0.00000438, -0.00000231, 0.00000905, -0.00000285, 0.00000376, 0.00000344, -0.00000936, -0.00000244, 0.00000044, 0.00000403, -0.00000156, -0.00000052, 0.00000182, -0.00000322, 0.00000140, 0.00000166, 0.00000331, -0.00000353, -0.00000943, 0.00000822, -0.00000166, 0.00000021, 0.00000555, -0.00000157, -0.00000617, 0.00000490, -0.00000668, 0.00000926, -0.00000775, 0.00000796, -0.00000308, -0.00000586, 0.00000207, -0.00000404, 0.00000845, -0.00001112, 0.00001933, -0.00001451, 0.00000275, -0.00000806, 0.00001181, -0.00001145, 0.00001036, -0.00000069, -0.00000503, 0.00000592, -0.00000125, 0.00000334, -0.00001733, 0.00001557, -0.00000636, 0.00000626, -0.00001754, 0.00001931, 0.00000256, -0.00000359, -0.00001544, 0.00000317, 0.00000652, -0.00001130, 0.00000208, -0.00001746, 0.00001155, 0.00000906, -0.00002416, 0.00000443, -0.00002669, -0.00000524, 0.00000323, -0.00000263, 0.00004532, 0.00000108, 0.00003470, 0.00000811, 0.00001157, -0.00003261, -0.00001482, -0.00002880, -0.00000654, -0.00008131, -0.00005073, 0.00014491, 0.00008370, 0.00014231, -0.00001234, -0.00009242, 0.00000560, -0.00003160, -0.00000762, -0.00000572, -0.00001670, -0.00000019, -0.00000264, -0.00000509, -0.00000141, },
		{ -0.00002100, -0.00000248, 0.00001168, -0.00000581, 0.00000176, -0.00000551, 0.00000303, 0.00001290, -0.00000531, 0.00000276, -0.00000456, -0.00000564, 0.00000198, 0.00000760, 0.00000641, -0.00000572, 0.00000293, -0.00000550, -0.00000639, 0.00000032, 0.00000434, 0.00000242, 0.00000221, -0.00000016, -0.00000076, -0.00000625, 0.00000172, -0.00000502, -0.00000158, 0.00000828, -0.00000216, 0.00000012, -0.00000515, -0.00000063, 0.00000834, -0.00000341, 0.00000510, 0.00000174, -0.00000942, -0.00000156, 0.00000089, 0.00000350, -0.00000178, -0.00000009, 0.00000144, -0.00000319, 0.00000171, 0.00000173, 0.00000325, -0.00000535, -0.00000728, 0.00000830, -0.00000249, 0.00000115, 0.00000526, -0.00000295, -0.00000507, 0.00000455, -0.00000626, 0.00000935, -0.00000871, 0.00000959, -0.00000551, -0.00000342, -0.00000010, -0.00000108, 0.00000520, -0.00000803, 0.00001796, -0.00001637, 0.00000483, -0.00000886, 0.00001325, -0.00001391, 0.00001416, -0.00000491, -0.00000209, 0.00000398, 0.00000043, 0.00000114, -0.00001534, 0.00001505, -0.00000673, 0.00000546, -0.00001742, 0.00002330, -0.00000245, -0.00000166, -0.00001763, 0.00000752, 0.00000340, -0.00001057, 0.00000234, -0.00001890, 0.00001580, 0.00000396, -0.00002063, 0.00000301, -0.00002918, 0.00000191, -0.00000078, 0.00000176, 0.00004179, 0.00000007, 0.00003531, 0.00000435, 0.00001139, -0.00003348, -0.00001281, -0.00003093, -0.00000402, -0.00009158, -0.00003951, 0.00015068, 0.00009283, 0.00014396, -0.00003911, -0.00008373, 0.00000828, -0.00003504, -0.00000390, -0.00000809, -0.00001514, -0.00000065, -0.00000282, -0.00000497, -0.00000054, 0.00000034, },
		{ 0.00000088, 0.00000995, -0.00000597, 0.00000217, -0.00000635, 0.00000564, 0.00001127, -0.00000580, 0.00000332, -0.00000529, -0.00000510, 0.00000219, 0.00000873, 0.00000479, -0.00000530, 0.00000295, -0.00000656, -0.00000579, 0.00000129, 0.00000400, 0.00000219, 0.00000237, -0.00000045, -0.00000120, -0.00000600, 0.00000200, -0.00000572, -0.00000014, 0.00000767, -0.00000251, 0.00000017, -0.00000569, 0.00000106, 0.00000727, -0.00000351, 0.00000615, -0.00000002, -0.00000920, -0.00000069, 0.00000119, 0.00000293, -0.00000188, 0.00000027, 0.00000103, -0.00000306, 0.00000195, 0.00000189, 0.00000288, -0.00000686, -0.00000493, 0.00000790, -0.00000300, 0.00000201, 0.00000478, -0.00000418, -0.00000381, 0.00000395, -0.00000555, 0.00000901, -0.00000917, 0.00001075, -0.00000755, -0.00000123, -0.00000207, 0.00000181, 0.00000166, -0.00000413, 0.00001525, -0.00001698, 0.00000621, -0.00000915, 0.00001407, -0.00001579, 0.00001765, -0.00000929, 0.00000135, 0.00000152, 0.00000256, -0.00000151, -0.00001277, 0.00001378, -0.00000644, 0.00000394, -0.00001605, 0.00002614, -0.00000726, 0.00000056, -0.00001998, 0.00001239, -0.00000089, -0.00000847, 0.00000134, -0.00001919, 0.00001927, -0.00000112, -0.00001615, -0.00000022, -0.00002949, 0.00000793, -0.00000494, 0.00000652, 0.00003737, 0.00000020, 0.00003500, 0.00000122, 0.00001095, -0.00003387, -0.00001141, -0.00003247, -0.00000222, -0.00010185, -0.00002702, 0.00015598, 0.00010433, 0.00014021, -0.00006481, -0.00007165, 0.00000817, -0.00003703, -0.00000073, -0.00001043, -0.00001340, -0.00000140, -0.00000269, -0.00000503, 0.00000046, -0.00000150, -0.00002168, },
		{ 0.00000780, -0.00000550, 0.00000209, -0.00000674, 0.00000811, 0.00000923, -0.00000584, 0.00000363, -0.00000590, -0.00000465, 0.00000261, 0.00000965, 0.00000309, -0.00000463, 0.00000270, -0.00000749, -0.00000500, 0.00000211, 0.00000355, 0.00000213, 0.00000242, -0.00000073, -0.00000164, -0.00000563, 0.00000210, -0.00000628, 0.00000134, 0.00000680, -0.00000261, 0.00000000, -0.00000595, 0.00000263, 0.00000595, -0.00000319, 0.00000684, -0.00000177, -0.00000873, 0.00000013, 0.00000135, 0.00000236, -0.00000187, 0.00000056, 0.00000060, -0.00000285, 0.00000213, 0.00000212, 0.00000218, -0.00000796, -0.00000252, 0.00000709, -0.00000318, 0.00000273, 0.00000414, -0.00000523, -0.00000247, 0.00000316, -0.00000462, 0.00000829, -0.00000914, 0.00001140, -0.00000911, 0.00000058, -0.00000369, 0.00000438, -0.00000185, 0.00000028, 0.00001136, -0.00001628, 0.00000675, -0.00000877, 0.00001411, -0.00001689, 0.00002054, -0.00001354, 0.00000507, -0.00000135, 0.00000507, -0.00000454, -0.00000968, 0.00001183, -0.00000553, 0.00000170, -0.00001337, 0.00002752, -0.00001144, 0.00000273, -0.00002218, 0.00001745, -0.00000613, -0.00000504, -0.00000105, -0.00001803, 0.00002149, -0.00000559, -0.00001131, -0.00000489, -0.00002758, 0.00001233, -0.00000875, 0.00001127, 0.00003237, 0.00000128, 0.00003383, -0.00000127, 0.00001022, -0.00003375, -0.00001072, -0.00003324, -0.00000145, -0.00011171, -0.00001365, 0.00016138, 0.00011762, 0.00013050, -0.00008795, -0.00005717, 0.00000537, -0.00003731, 0.00000157, -0.00001243, -0.00001171, -0.00000229, -0.00000233, -0.00000524, 0.00000158, -0.00000381, -0.00002157, 0.00000391, },
		{ -0.00000451, 0.00000148, -0.00000656, 0.00001026, 0.00000695, -0.00000546, 0.00000366, -0.00000635, -0.00000428, 0.00000324, 0.00001027, 0.00000140, -0.00000376, 0.00000217, -0.00000823, -0.00000408, 0.00000274, 0.00000305, 0.00000220, 0.00000237, -0.00000100, -0.00000208, -0.00000514, 0.00000200, -0.00000661, 0.00000276, 0.00000575, -0.00000250, -0.00000038, -0.00000587, 0.00000397, 0.00000452, -0.00000251, 0.00000715, -0.00000342, -0.00000805, 0.00000089, 0.00000137, 0.00000182, -0.00000175, 0.00000075, 0.00000020, -0.00000260, 0.00000228, 0.00000238, 0.00000119, -0.00000857, -0.00000022, 0.00000599, -0.00000303, 0.00000327, 0.00000337, -0.00000607, -0.00000112, 0.00000227, -0.00000352, 0.00000725, -0.00000867, 0.00001153, -0.00001014, 0.00000193, -0.00000481, 0.00000643, -0.00000502, 0.00000484, 0.00000657, -0.00001434, 0.00000637, -0.00000766, 0.00001328, -0.00001701, 0.00002256, -0.00001737, 0.00000882, -0.00000443, 0.00000783, -0.00000782, -0.00000620, 0.00000929, -0.00000408, -0.00000113, -0.00000940, 0.00002722, -0.00001461, 0.00000446, -0.00002389, 0.00002229, -0.00001194, -0.00000049, -0.00000481, -0.00001527, 0.00002209, -0.00000888, -0.00000681, -0.00001043, -0.00002362, 0.00001485, -0.00001175, 0.00001561, 0.00002715, 0.00000306, 0.00003192, -0.00000316, 0.00000921, -0.00003315, -0.00001073, -0.00003314, -0.00000194, -0.00012077, 0.00000020, 0.00016759, 0.00013179, 0.00011466, -0.00010717, -0.00004143, 0.00000027, -0.00003580, 0.00000273, -0.00001383, -0.00001032, -0.00000311, -0.00000189, -0.00000550, 0.00000273, -0.00000651, -0.00002061, 0.00000637, 0.00000549, },
		{ 0.00000041, -0.00000574, 0.00001191, 0.00000462, -0.00000477, 0.00000344, -0.00000664, -0.00000399, 0.00000410, 0.00001055, -0.00000017, -0.00000276, 0.00000139, -0.00000873, -0.00000307, 0.00000317, 0.00000256, 0.00000236, 0.00000222, -0.00000126, -0.00000250, -0.00000456, 0.00000169, -0.00000667, 0.00000402, 0.00000461, -0.00000221, -0.00000096, -0.00000545, 0.00000498, 0.00000310, -0.00000156, 0.00000707, -0.00000489, -0.00000719, 0.00000155, 0.00000127, 0.00000133, -0.00000156, 0.00000085, -0.00000015, -0.00000233, 0.00000241, 0.00000261, -0.00000005, -0.00000865, 0.00000186, 0.00000470, -0.00000259, 0.00000358, 0.00000254, -0.00000668, 0.00000018, 0.00000132, -0.00000235, 0.00000599, -0.00000782, 0.00001119, -0.00001065, 0.00000278, -0.00000536, 0.00000779, -0.00000758, 0.00000917, 0.00000122, -0.00001134, 0.00000509, -0.00000579, 0.00001153, -0.00001605, 0.00002349, -0.00002046, 0.00001230, -0.00000751, 0.00001068, -0.00001123, -0.00000245, 0.00000629, -0.00000220, -0.00000439, -0.00000427, 0.00002513, -0.00001643, 0.00000542, -0.00002472, 0.00002647, -0.00001787, 0.00000484, -0.00000974, -0.00001091, 0.00002084, -0.00001050, -0.00000332, -0.00001612, -0.00001805, 0.00001541, -0.00001362, 0.00001919, 0.00002207, 0.00000525, 0.00002946, -0.00000457, 0.00000803, -0.00003218, -0.00001138, -0.00003219, -0.00000385, -0.00012872, 0.00001417, 0.00017538, 0.00014564, 0.00009296, -0.00012131, -0.00002570, -0.00000650, -0.00003266, 0.00000266, -0.00001445, -0.00000943, -0.00000366, -0.00000154, -0.00000570, 0.00000378, -0.00000949, -0.00001878, 0.00000805, 0.00000331, -0.00000319, },
		{ -0.00000431, 0.00001295, 0.00000239, -0.00000386, 0.00000302, -0.00000679, -0.00000376, 0.00000516, 0.00001045, -0.00000156, -0.00000172, 0.00000040, -0.00000898, -0.00000204, 0.00000342, 0.00000211, 0.00000256, 0.00000201, -0.00000151, -0.00000286, -0.00000391, 0.00000121, -0.00000644, 0.00000504, 0.00000346, -0.00000183, -0.00000166, -0.00000471, 0.00000560, 0.00000181, -0.00000043, 0.00000662, -0.00000615, -0.00000621, 0.00000210, 0.00000105, 0.00000092, -0.00000132, 0.00000085, -0.00000043, -0.00000206, 0.00000257, 0.00000277, -0.00000145, -0.00000819, 0.00000358, 0.00000335, -0.00000192, 0.00000365, 0.00000168, -0.00000707, 0.00000139, 0.00000037, -0.00000115, 0.00000459, -0.00000669, 0.00001045, -0.00001069, 0.00000315, -0.00000532, 0.00000837, -0.00000929, 0.00001291, -0.00000429, -0.00000752, 0.00000305, -0.00000323, 0.00000891, -0.00001396, 0.00002318, -0.00002257, 0.00001524, -0.00001035, 0.00001342, -0.00001460, 0.00000137, 0.00000300, -0.00000007, -0.00000785, 0.00000176, 0.00002131, -0.00001672, 0.00000531, -0.00002437, 0.00002951, -0.00002338, 0.00001047, -0.00001550, -0.00000515, 0.00001771, -0.00001011, -0.00000150, -0.00002115, -0.00001150, 0.00001421, -0.00001423, 0.00002177, 0.00001744, 0.00000754, 0.00002668, -0.00000566, 0.00000682, -0.00003102, -0.00001249, -0.00003051, -0.00000717, -0.00013539, 0.00002804, 0.00018547, 0.00015770, 0.00006616, -0.00012959, -0.00001120, -0.00001417, -0.00002821, 0.00000142, -0.00001419, -0.00000919, -0.00000374, -0.00000146, -0.00000566, 0.00000456, -0.00001254, -0.00001618, 0.00000881, 0.00000152, -0.00000176, -0.00000100, },
		{ 0.00001329, 0.00000042, -0.00000287, 0.00000244, -0.00000684, -0.00000354, 0.00000638, 0.00000995, -0.00000268, -0.00000071, -0.00000075, -0.00000894, -0.00000106, 0.00000351, 0.00000174, 0.00000278, 0.00000174, -0.00000178, -0.00000314, -0.00000327, 0.00000059, -0.00000591, 0.00000577, 0.00000241, -0.00000144, -0.00000242, -0.00000370, 0.00000582, 0.00000077, 0.00000075, 0.00000587, -0.00000715, -0.00000513, 0.00000254, 0.00000073, 0.00000062, -0.00000106, 0.00000078, -0.00000063, -0.00000183, 0.00000276, 0.00000278, -0.00000292, -0.00000723, 0.00000486, 0.00000207, -0.00000111, 0.00000350, 0.00000084, -0.00000724, 0.00000247, -0.00000054, -0.00000001, 0.00000314, -0.00000538, 0.00000942, -0.00001032, 0.00000310, -0.00000474, 0.00000813, -0.00001000, 0.00001575, -0.00000957, -0.00000322, 0.00000041, -0.00000012, 0.00000552, -0.00001080, 0.00002155, -0.00002349, 0.00001738, -0.00001269, 0.00001584, -0.00001773, 0.00000510, -0.00000041, 0.00000214, -0.00001124, 0.00000837, 0.00001593, -0.00001539, 0.00000396, -0.00002258, 0.00003103, -0.00002792, 0.00001586, -0.00002160, 0.00000163, 0.00001288, -0.00000759, -0.00000182, -0.00002476, -0.00000474, 0.00001165, -0.00001363, 0.00002325, 0.00001349, 0.00000969, 0.00002379, -0.00000660, 0.00000576, -0.00002989, -0.00001381, -0.00002832, -0.00001179, -0.00014076, 0.00004179, 0.00019841, 0.00016639, 0.00003547, -0.00013166, 0.00000097, -0.00002188, -0.00002295, -0.00000078, -0.00001311, -0.00000964, -0.00000325, -0.00000182, -0.00000523, 0.00000488, -0.00001543, -0.00001299, 0.00000861, 0.00000034, -0.00000047, -0.00000257, -0.00000234, },
		{ -0.00000116, -0.00000190, 0.00000178, -0.00000681, -0.00000329, 0.00000769, 0.00000907, -0.00000349, 0.00000019, -0.00000199, -0.00000863, -0.00000018, 0.00000347, 0.00000145, 0.00000298, 0.00000144, -0.00000208, -0.00000330, -0.00000267, -0.00000010, -0.00000512, 0.00000617, 0.00000151, -0.00000111, -0.00000314, -0.00000251, 0.00000564, 0.00000002, 0.00000188, 0.00000487, -0.00000787, -0.00000400, 0.00000285, 0.00000034, 0.00000042, -0.00000082, 0.00000064, -0.00000076, -0.00000162, 0.00000302, 0.00000259, -0.00000434, -0.00000584, 0.00000566, 0.00000095, -0.00000025, 0.00000315, 0.00000004, -0.00000721, 0.00000340, -0.00000136, 0.00000104, 0.00000172, -0.00000400, 0.00000820, -0.00000967, 0.00000272, -0.00000370, 0.00000715, -0.00000965, 0.00001742, -0.00001421, 0.00000121, -0.00000257, 0.00000332, 0.00000156, -0.00000670, 0.00001861, -0.00002310, 0.00001850, -0.00001431, 0.00001773, -0.00002043, 0.00000852, -0.00000373, 0.00000424, -0.00001428, 0.00001515, 0.00000930, -0.00001252, 0.00000131, -0.00001921, 0.00003068, -0.00003100, 0.00002043, -0.00002746, 0.00000890, 0.00000674, -0.00000306, -0.00000456, -0.00002628, 0.00000141, 0.00000830, -0.00001206, 0.00002367, 0.00001031, 0.00001152, 0.00002096, -0.00000754, 0.00000501, -0.00002903, -0.00001505, -0.00002597, -0.00001744, -0.00014499, 0.00005563, 0.00021450, 0.00017012, 0.00000252, -0.00012762, 0.00000998, -0.00002879, -0.00001746, -0.00000360, -0.00001138, -0.00001074, -0.00000215, -0.00000270, -0.00000428, 0.00000455, -0.00001786, -0.00000946, 0.00000752, -0.00000008, 0.00000044, -0.00000409, 0.00000002, 0.00001295, },
		{ -0.00000105, 0.00000110, -0.00000676, -0.00000296, 0.00000902, 0.00000786, -0.00000393, 0.00000089, -0.00000324, -0.00000808, 0.00000056, 0.00000336, 0.00000123, 0.00000316, 0.00000112, -0.00000239, -0.00000334, -0.00000218, -0.00000079, -0.00000413, 0.00000625, 0.00000081, -0.00000090, -0.00000375, -0.00000124, 0.00000513, -0.00000038, 0.00000287, 0.00000370, -0.00000833, -0.00000285, 0.00000302, -0.00000009, 0.00000034, -0.00000061, 0.00000046, -0.00000082, -0.00000146, 0.00000334, 0.00000217, -0.00000559, -0.00000413, 0.00000597, 0.00000010, 0.00000057, 0.00000264, -0.00000068, -0.00000701, 0.00000416, -0.00000209, 0.00000195, 0.00000040, -0.00000261, 0.00000689, -0.00000883, 0.00000215, -0.00000236, 0.00000554, -0.00000826, 0.00001778, -0.00001789, 0.00000541, -0.00000560, 0.00000686, -0.00000271, -0.00000188, 0.00001449, -0.00002136, 0.00001847, -0.00001502, 0.00001889, -0.00002252, 0.00001145, -0.00000678, 0.00000604, -0.00001669, 0.00002166, 0.00000182, -0.00000831, -0.00000256, -0.00001426, 0.00002830, -0.00003220, 0.00002361, -0.00003245, 0.00001600, -0.00000011, 0.00000309, -0.00000971, -0.00002523, 0.00000618, 0.00000483, -0.00000993, 0.00002324, 0.00000787, 0.00001297, 0.00001827, -0.00000857, 0.00000469, -0.00002864, -0.00001591, -0.00002383, -0.00002374, -0.00014842, 0.00007006, 0.00023368, 0.00016745, -0.00003080, -0.00011808, 0.00001534, -0.00003417, -0.00001235, -0.00000661, -0.00000928, -0.00001231, -0.00000048, -0.00000414, -0.00000273, 0.00000341, -0.00001957, -0.00000593, 0.00000572, 0.00000032, 0.00000079, -0.00000532, 0.00000259, 0.00001198, -0.00000230, },
		{ 0.00000047, -0.00000672, -0.00000248, 0.00001028, 0.00000636, -0.00000402, 0.00000134, -0.00000441, -0.00000734, 0.00000113, 0.00000321, 0.00000107, 0.00000332, 0.00000076, -0.00000272, -0.00000323, -0.00000185, -0.00000138, -0.00000302, 0.00000604, 0.00000034, -0.00000087, -0.00000415, 0.00000000, 0.00000437, -0.00000043, 0.00000363, 0.00000244, -0.00000854, -0.00000172, 0.00000304, -0.00000054, 0.00000038, -0.00000048, 0.00000028, -0.00000083, -0.00000133, 0.00000370, 0.00000149, -0.00000658, -0.00000222, 0.00000582, -0.00000045, 0.00000127, 0.00000203, -0.00000133, -0.00000665, 0.00000476, -0.00000270, 0.00000272, -0.00000079, -0.00000130, 0.00000558, -0.00000792, 0.00000150, -0.00000088, 0.00000349, -0.00000597, 0.00001677, -0.00002035, 0.00000905, -0.00000839, 0.00001020, -0.00000700, 0.00000337, 0.00000939, -0.00001834, 0.00001722, -0.00001468, 0.00001917, -0.00002384, 0.00001371, -0.00000939, 0.00000740, -0.00001821, 0.00002745, -0.00000601, -0.00000311, -0.00000742, -0.00000789, 0.00002385, -0.00003124, 0.00002496, -0.00003598, 0.00002224, -0.00000694, 0.00001026, -0.00001698, -0.00002141, 0.00000893, 0.00000193, -0.00000773, 0.00002230, 0.00000600, 0.00001410, 0.00001571, -0.00000968, 0.00000483, -0.00002886, -0.00001615, -0.00002227, -0.00003025, -0.00015150, 0.00008587, 0.00025542, 0.00015724, -0.00006241, -0.00010411, 0.00001691, -0.00003748, -0.00000820, -0.00000934, -0.00000714, -0.00001411, 0.00000158, -0.00000606, -0.00000059, 0.00000140, -0.00002031, -0.00000276, 0.00000347, 0.00000150, 0.00000046, -0.00000605, 0.00000514, 0.00001049, -0.00000297, -0.00000039, },
		{ -0.00000673, -0.00000181, 0.00001138, 0.00000468, -0.00000377, 0.00000151, -0.00000542, -0.00000648, 0.00000153, 0.00000307, 0.00000094, 0.00000346, 0.00000037, -0.00000303, -0.00000300, -0.00000171, -0.00000182, -0.00000189, 0.00000559, 0.00000009, -0.00000102, -0.00000431, 0.00000111, 0.00000345, -0.00000017, 0.00000414, 0.00000118, -0.00000851, -0.00000061, 0.00000291, -0.00000097, 0.00000051, -0.00000042, 0.00000012, -0.00000081, -0.00000121, 0.00000410, 0.00000055, -0.00000720, -0.00000025, 0.00000528, -0.00000067, 0.00000177, 0.00000138, -0.00000190, -0.00000615, 0.00000519, -0.00000320, 0.00000333, -0.00000181, -0.00000012, 0.00000436, -0.00000704, 0.00000091, 0.00000057, 0.00000122, -0.00000298, 0.00001448, -0.00002144, 0.00001186, -0.00001067, 0.00001308, -0.00001098, 0.00000869, 0.00000363, -0.00001421, 0.00001478, -0.00001323, 0.00001846, -0.00002425, 0.00001517, -0.00001141, 0.00000819, -0.00001867, 0.00003211, -0.00001367, 0.00000266, -0.00001295, -0.00000039, 0.00001748, -0.00002804, 0.00002415, -0.00003753, 0.00002695, -0.00001297, 0.00001766, -0.00002581, -0.00001490, 0.00000924, 0.00000022, -0.00000601, 0.00002128, 0.00000441, 0.00001509, 0.00001314, -0.00001074, 0.00000535, -0.00002969, -0.00001561, -0.00002161, -0.00003655, -0.00015478, 0.00010409, 0.00027869, 0.00013884, -0.00009030, -0.00008711, 0.00001497, -0.00003846, -0.00000542, -0.00001137, -0.00000536, -0.00001582, 0.00000380, -0.00000831, 0.00000203, -0.00000148, -0.00001991, -0.00000028, 0.00000111, 0.00000328, -0.00000058, -0.00000614, 0.00000745, 0.00000863, -0.00000318, 0.00000003, -0.00000006, },
		{ -0.00000092, 0.00001222, 0.00000290, -0.00000322, 0.00000138, -0.00000620, -0.00000557, 0.00000178, 0.00000296, 0.00000083, 0.00000359, -0.00000007, -0.00000330, -0.00000267, -0.00000175, -0.00000204, -0.00000082, 0.00000499, 0.00000002, -0.00000134, -0.00000419, 0.00000199, 0.00000250, 0.00000034, 0.00000435, -0.00000003, -0.00000829, 0.00000044, 0.00000263, -0.00000134, 0.00000072, -0.00000046, 0.00000001, -0.00000080, -0.00000108, 0.00000448, -0.00000063, -0.00000738, 0.00000164, 0.00000445, -0.00000057, 0.00000204, 0.00000074, -0.00000240, -0.00000554, 0.00000545, -0.00000359, 0.00000380, -0.00000266, 0.00000092, 0.00000327, -0.00000626, 0.00000047, 0.00000182, -0.00000100, 0.00000042, 0.00001106, -0.00002112, 0.00001364, -0.00001223, 0.00001526, -0.00001436, 0.00001373, -0.00000246, -0.00000924, 0.00001129, -0.00001070, 0.00001671, -0.00002369, 0.00001571, -0.00001273, 0.00000834, -0.00001792, 0.00003529, -0.00002064, 0.00000852, -0.00001871, 0.00000780, 0.00000951, -0.00002269, 0.00002103, -0.00003677, 0.00002953, -0.00001739, 0.00002435, -0.00003538, -0.00000612, 0.00000696, 0.00000017, -0.00000527, 0.00002063, 0.00000274, 0.00001621, 0.00001034, -0.00001153, 0.00000602, -0.00003101, -0.00001430, -0.00002203, -0.00004231, -0.00015875, 0.00012594, 0.00030198, 0.00011216, -0.00011260, -0.00006872, 0.00001016, -0.00003713, -0.00000428, -0.00001235, -0.00000427, -0.00001709, 0.00000588, -0.00001064, 0.00000495, -0.00000512, -0.00001829, 0.00000121, -0.00000100, 0.00000544, -0.00000228, -0.00000548, 0.00000931, 0.00000660, -0.00000302, 0.00000020, -0.00000048, -0.00000680, },
		{ 0.00001275, 0.00000112, -0.00000244, 0.00000096, -0.00000672, -0.00000470, 0.00000192, 0.00000288, 0.00000071, 0.00000373, -0.00000059, -0.00000348, -0.00000229, -0.00000196, -0.00000201, 0.00000010, 0.00000431, 0.00000008, -0.00000181, -0.00000380, 0.00000258, 0.00000160, 0.00000101, 0.00000429, -0.00000113, -0.00000789, 0.00000142, 0.00000220, -0.00000163, 0.00000098, -0.00000058, -0.00000003, -0.00000081, -0.00000090, 0.00000479, -0.00000198, -0.00000709, 0.00000331, 0.00000344, -0.00000020, 0.00000204, 0.00000016, -0.00000283, -0.00000482, 0.00000555, -0.00000387, 0.00000412, -0.00000335, 0.00000180, 0.00000233, -0.00000562, 0.00000026, 0.00000272, -0.00000296, 0.00000394, 0.00000678, -0.00001945, 0.00001429, -0.00001291, 0.00001654, -0.00001687, 0.00001813, -0.00000847, -0.00000376, 0.00000697, -0.00000720, 0.00001396, -0.00002213, 0.00001528, -0.00001328, 0.00000784, -0.00001593, 0.00003672, -0.00002644, 0.00001397, -0.00002422, 0.00001615, 0.00000044, -0.00001549, 0.00001570, -0.00003352, 0.00002959, -0.00001951, 0.00002941, -0.00004468, 0.00000421, 0.00000226, 0.00000200, -0.00000588, 0.00002077, 0.00000063, 0.00001777, 0.00000700, -0.00001173, 0.00000651, -0.00003254, -0.00001240, -0.00002353, -0.00004735, -0.00016379, 0.00015265, 0.00032332, 0.00007776, -0.00012788, -0.00005066, 0.00000340, -0.00003379, -0.00000481, -0.00001209, -0.00000415, -0.00001761, 0.00000749, -0.00001277, 0.00000790, -0.00000929, -0.00001550, 0.00000150, -0.00000250, 0.00000767, -0.00000448, -0.00000405, 0.00001057, 0.00000457, -0.00000257, 0.00000015, -0.00000078, -0.00000693, 0.00000020, },
		{ -0.00000054, -0.00000151, 0.00000029, -0.00000693, -0.00000393, 0.00000200, 0.00000283, 0.00000059, 0.00000389, -0.00000120, -0.00000353, -0.00000192, -0.00000230, -0.00000172, 0.00000080, 0.00000365, 0.00000020, -0.00000236, -0.00000318, 0.00000285, 0.00000087, 0.00000177, 0.00000398, -0.00000210, -0.00000735, 0.00000230, 0.00000165, -0.00000180, 0.00000124, -0.00000078, 0.00000000, -0.00000087, -0.00000066, 0.00000498, -0.00000344, -0.00000634, 0.00000467, 0.00000237, 0.00000038, 0.00000178, -0.00000032, -0.00000322, -0.00000402, 0.00000549, -0.00000404, 0.00000433, -0.00000390, 0.00000254, 0.00000154, -0.00000514, 0.00000031, 0.00000319, -0.00000445, 0.00000724, 0.00000197, -0.00001662, 0.00001383, -0.00001264, 0.00001680, -0.00001829, 0.00002155, -0.00001400, 0.00000187, 0.00000209, -0.00000291, 0.00001030, -0.00001962, 0.00001387, -0.00001303, 0.00000672, -0.00001275, 0.00003626, -0.00003068, 0.00001855, -0.00002898, 0.00002410, -0.00000914, -0.00000692, 0.00000843, -0.00002787, 0.00002692, -0.00001880, 0.00003196, -0.00005265, 0.00001514, -0.00000436, 0.00000568, -0.00000804, 0.00002200, -0.00000225, 0.00002008, 0.00000280, -0.00001095, 0.00000639, -0.00003383, -0.00001027, -0.00002588, -0.00005177, -0.00017001, 0.00018538, 0.00034040, 0.00003692, -0.00013517, -0.00003456, -0.00000420, -0.00002900, -0.00000681, -0.00001055, -0.00000516, -0.00001714, 0.00000834, -0.00001437, 0.00001056, -0.00001368, -0.00001171, 0.00000048, -0.00000308, 0.00000963, -0.00000695, -0.00000193, 0.00001112, 0.00000272, -0.00000195, -0.00000010, -0.00000096, -0.00000710, 0.00000155, 0.00001288, },
		{ -0.00000052, -0.00000057, -0.00000684, -0.00000334, 0.00000207, 0.00000277, 0.00000047, 0.00000404, -0.00000189, -0.00000342, -0.00000162, -0.00000270, -0.00000119, 0.00000124, 0.00000307, 0.00000030, -0.00000292, -0.00000237, 0.00000282, 0.00000037, 0.00000251, 0.00000348, -0.00000290, -0.00000670, 0.00000307, 0.00000098, -0.00000183, 0.00000148, -0.00000103, 0.00000011, -0.00000098, -0.00000033, 0.00000499, -0.00000490, -0.00000516, 0.00000563, 0.00000136, 0.00000106, 0.00000129, -0.00000068, -0.00000357, -0.00000315, 0.00000528, -0.00000410, 0.00000443, -0.00000432, 0.00000316, 0.00000088, -0.00000478, 0.00000061, 0.00000319, -0.00000530, 0.00001003, -0.00000299, -0.00001289, 0.00001235, -0.00001144, 0.00001602, -0.00001851, 0.00002376, -0.00001868, 0.00000723, -0.00000298, 0.00000189, 0.00000592, -0.00001628, 0.00001156, -0.00001203, 0.00000507, -0.00000852, 0.00003389, -0.00003309, 0.00002186, -0.00003256, 0.00003104, -0.00001856, 0.00000242, -0.00000029, -0.00002011, 0.00002158, -0.00001499, 0.00003131, -0.00005822, 0.00002556, -0.00001216, 0.00001089, -0.00001170, 0.00002444, -0.00000611, 0.00002337, -0.00000256, -0.00000880, 0.00000516, -0.00003436, -0.00000848, -0.00002862, -0.00005593, -0.00017718, 0.00022500, 0.00035074, -0.00000843, -0.00013418, -0.00002180, -0.00001146, -0.00002350, -0.00000990, -0.00000791, -0.00000730, -0.00001553, 0.00000819, -0.00001517, 0.00001258, -0.00001792, -0.00000724, -0.00000179, -0.00000252, 0.00001101, -0.00000942, 0.00000073, 0.00001092, 0.00000119, -0.00000127, -0.00000046, -0.00000106, -0.00000728, 0.00000310, 0.00001260, -0.00000199, },
		{ -0.00000153, -0.00000648, -0.00000296, 0.00000220, 0.00000268, 0.00000039, 0.00000417, -0.00000265, -0.00000313, -0.00000147, -0.00000308, -0.00000048, 0.00000141, 0.00000263, 0.00000033, -0.00000341, -0.00000148, 0.00000250, 0.00000014, 0.00000315, 0.00000284, -0.00000355, -0.00000594, 0.00000370, 0.00000023, -0.00000172, 0.00000165, -0.00000129, 0.00000027, -0.00000116, 0.00000010, 0.00000477, -0.00000627, -0.00000362, 0.00000616, 0.00000051, 0.00000177, 0.00000062, -0.00000091, -0.00000388, -0.00000222, 0.00000493, -0.00000407, 0.00000443, -0.00000464, 0.00000370, 0.00000031, -0.00000452, 0.00000110, 0.00000272, -0.00000543, 0.00001207, -0.00000776, -0.00000857, 0.00001004, -0.00000943, 0.00001426, -0.00001749, 0.00002458, -0.00002221, 0.00001196, -0.00000788, 0.00000691, 0.00000105, -0.00001227, 0.00000845, -0.00001032, 0.00000300, -0.00000347, 0.00002972, -0.00003353, 0.00002362, -0.00003457, 0.00003644, -0.00002713, 0.00001181, -0.00000981, -0.00001079, 0.00001393, -0.00000808, 0.00002702, -0.00006049, 0.00003433, -0.00002015, 0.00001705, -0.00001659, 0.00002804, -0.00001099, 0.00002778, -0.00000926, -0.00000498, 0.00000236, -0.00003353, -0.00000770, -0.00003107, -0.00006054, -0.00018460, 0.00027195, 0.00035185, -0.00005582, -0.00012529, -0.00001339, -0.00001729, -0.00001815, -0.00001351, -0.00000449, -0.00001040, -0.00001280, 0.00000691, -0.00001493, 0.00001365, -0.00002157, -0.00000246, -0.00000513, -0.00000073, 0.00001151, -0.00001158, 0.00000373, 0.00001002, 0.00000009, -0.00000066, -0.00000089, -0.00000112, -0.00000741, 0.00000479, 0.00001188, -0.00000314, 0.00000042, },
		{ -0.00000592, -0.00000280, 0.00000244, 0.00000252, 0.00000037, 0.00000424, -0.00000345, -0.00000266, -0.00000150, -0.00000336, 0.00000033, 0.00000133, 0.00000237, 0.00000024, -0.00000377, -0.00000058, 0.00000197, 0.00000018, 0.00000363, 0.00000213, -0.00000404, -0.00000511, 0.00000417, -0.00000056, -0.00000145, 0.00000173, -0.00000154, 0.00000047, -0.00000139, 0.00000062, 0.00000430, -0.00000744, -0.00000182, 0.00000624, -0.00000010, 0.00000240, -0.00000016, -0.00000102, -0.00000416, -0.00000127, 0.00000444, -0.00000393, 0.00000434, -0.00000487, 0.00000418, -0.00000021, -0.00000430, 0.00000173, 0.00000187, -0.00000484, 0.00001315, -0.00001197, -0.00000402, 0.00000715, -0.00000677, 0.00001165, -0.00001532, 0.00002394, -0.00002435, 0.00001571, -0.00001224, 0.00001179, -0.00000402, -0.00000783, 0.00000469, -0.00000801, 0.00000068, 0.00000214, 0.00002398, -0.00003199, 0.00002368, -0.00003476, 0.00003987, -0.00003421, 0.00002050, -0.00001941, -0.00000061, 0.00000453, 0.00000158, 0.00001898, -0.00005878, 0.00004034, -0.00002728, 0.00002337, -0.00002218, 0.00003249, -0.00001673, 0.00003322, -0.00001736, 0.00000073, -0.00000240, -0.00003076, -0.00000866, -0.00003236, -0.00006658, -0.00019107, 0.00032607, 0.00034152, -0.00010235, -0.00010956, -0.00000983, -0.00002079, -0.00001377, -0.00001698, -0.00000075, -0.00001416, -0.00000909, 0.00000451, -0.00001353, 0.00001351, -0.00002423, 0.00000214, -0.00000922, 0.00000227, 0.00001096, -0.00001315, 0.00000680, 0.00000851, -0.00000055, -0.00000019, -0.00000131, -0.00000118, -0.00000745, 0.00000654, 0.00001076, -0.00000395, 0.00000124, -0.00000251, },
		{ -0.00000284, 0.00000280, 0.00000224, 0.00000045, 0.00000422, -0.00000425, -0.00000202, -0.00000175, -0.00000346, 0.00000117, 0.00000106, 0.00000228, -0.00000000, -0.00000395, 0.00000023, 0.00000130, 0.00000047, 0.00000391, 0.00000141, -0.00000440, -0.00000421, 0.00000447, -0.00000136, -0.00000105, 0.00000169, -0.00000174, 0.00000068, -0.00000164, 0.00000122, 0.00000356, -0.00000831, 0.00000013, 0.00000593, -0.00000043, 0.00000287, -0.00000099, -0.00000104, -0.00000439, -0.00000030, 0.00000384, -0.00000368, 0.00000416, -0.00000501, 0.00000462, -0.00000073, -0.00000405, 0.00000237, 0.00000074, -0.00000357, 0.00001321, -0.00001535, 0.00000042, 0.00000397, -0.00000371, 0.00000843, -0.00001215, 0.00002191, -0.00002499, 0.00001822, -0.00001574, 0.00001620, -0.00000897, -0.00000323, 0.00000051, -0.00000524, -0.00000171, 0.00000797, 0.00001700, -0.00002864, 0.00002203, -0.00003303, 0.00004102, -0.00003926, 0.00002782, -0.00002827, 0.00000962, -0.00000580, 0.00001336, 0.00000742, -0.00005274, 0.00004272, -0.00003247, 0.00002892, -0.00002776, 0.00003727, -0.00002297, 0.00003944, -0.00002670, 0.00000832, -0.00000936, -0.00002561, -0.00001210, -0.00003158, -0.00007530, -0.00019489, 0.00038645, 0.00031799, -0.00014495, -0.00008867, -0.00001108, -0.00002138, -0.00001107, -0.00001966, 0.00000275, -0.00001814, -0.00000470, 0.00000112, -0.00001099, 0.00001203, -0.00002557, 0.00000609, -0.00001365, 0.00000631, 0.00000925, -0.00001388, 0.00000968, 0.00000655, -0.00000073, 0.00000007, -0.00000166, -0.00000128, -0.00000733, 0.00000828, 0.00000928, -0.00000438, 0.00000185, -0.00000341, -0.00000523, },
		{ 0.00000331, 0.00000182, 0.00000066, 0.00000405, -0.00000497, -0.00000127, -0.00000222, -0.00000332, 0.00000193, 0.00000064, 0.00000233, -0.00000039, -0.00000391, 0.00000087, 0.00000061, 0.00000096, 0.00000396, 0.00000072, -0.00000464, -0.00000326, 0.00000457, -0.00000212, -0.00000054, 0.00000153, -0.00000186, 0.00000086, -0.00000190, 0.00000186, 0.00000256, -0.00000880, 0.00000208, 0.00000528, -0.00000047, 0.00000312, -0.00000180, -0.00000099, -0.00000456, 0.00000065, 0.00000313, -0.00000333, 0.00000388, -0.00000508, 0.00000503, -0.00000126, -0.00000372, 0.00000294, -0.00000051, -0.00000178, 0.00001223, -0.00001769, 0.00000443, 0.00000077, -0.00000052, 0.00000485, -0.00000826, 0.00001863, -0.00002413, 0.00001935, -0.00001811, 0.00001981, -0.00001351, 0.00000126, -0.00000387, -0.00000218, -0.00000399, 0.00001368, 0.00000920, -0.00002372, 0.00001879, -0.00002943, 0.00003977, -0.00004190, 0.00003314, -0.00003567, 0.00001903, -0.00001612, 0.00002636, -0.00000704, -0.00004242, 0.00004087, -0.00003477, 0.00003274, -0.00003249, 0.00004171, -0.00002915, 0.00004596, -0.00003689, 0.00001758, -0.00001850, -0.00001780, -0.00001857, -0.00002781, -0.00008806, -0.00019390, 0.00045139, 0.00028023, -0.00018060, -0.00006477, -0.00001653, -0.00001886, -0.00001055, -0.00002098, 0.00000545, -0.00002184, -0.00000005, -0.00000297, -0.00000744, 0.00000919, -0.00002535, 0.00000892, -0.00001792, 0.00001110, 0.00000643, -0.00001359, 0.00001209, 0.00000435, -0.00000050, 0.00000010, -0.00000192, -0.00000145, -0.00000701, 0.00000990, 0.00000751, -0.00000442, 0.00000221, -0.00000413, -0.00000451, -0.00000305, },
		{ 0.00000125, 0.00000104, 0.00000368, -0.00000555, -0.00000046, -0.00000288, -0.00000291, 0.00000253, 0.00000018, 0.00000249, -0.00000090, -0.00000364, 0.00000128, -0.00000002, 0.00000159, 0.00000379, 0.00000012, -0.00000477, -0.00000227, 0.00000446, -0.00000278, 0.00000004, 0.00000124, -0.00000187, 0.00000099, -0.00000212, 0.00000249, 0.00000132, -0.00000886, 0.00000392, 0.00000438, -0.00000023, 0.00000312, -0.00000251, -0.00000091, -0.00000466, 0.00000155, 0.00000236, -0.00000289, 0.00000351, -0.00000504, 0.00000541, -0.00000182, -0.00000326, 0.00000333, -0.00000172, 0.00000036, 0.00001030, -0.00001888, 0.00000775, -0.00000215, 0.00000254, 0.00000122, -0.00000393, 0.00001436, -0.00002186, 0.00001902, -0.00001917, 0.00002238, -0.00001733, 0.00000534, -0.00000819, 0.00000101, -0.00000600, 0.00001893, 0.00000101, -0.00001762, 0.00001423, -0.00002416, 0.00003617, -0.00004191, 0.00003604, -0.00004097, 0.00002679, -0.00002544, 0.00003948, -0.00002345, -0.00002827, 0.00003457, -0.00003347, 0.00003398, -0.00003551, 0.00004502, -0.00003454, 0.00005210, -0.00004732, 0.00002800, -0.00002955, -0.00000736, -0.00002841, -0.00002035, -0.00010620, -0.00018561, 0.00051835, 0.00022806, -0.00020659, -0.00004032, -0.00002511, -0.00001345, -0.00001244, -0.00002050, 0.00000685, -0.00002473, 0.00000438, -0.00000737, -0.00000317, 0.00000514, -0.00002346, 0.00001027, -0.00002154, 0.00001623, 0.00000264, -0.00001223, 0.00001382, 0.00000211, 0.00000006, -0.00000011, -0.00000206, -0.00000171, -0.00000644, 0.00001131, 0.00000556, -0.00000411, 0.00000228, -0.00000460, -0.00000386, -0.00000334, 0.00000393, },
		{ 0.00000158, 0.00000308, -0.00000593, 0.00000030, -0.00000369, -0.00000221, 0.00000289, -0.00000025, 0.00000269, -0.00000149, -0.00000318, 0.00000143, -0.00000050, 0.00000225, 0.00000345, -0.00000040, -0.00000482, -0.00000127, 0.00000415, -0.00000331, 0.00000066, 0.00000085, -0.00000178, 0.00000103, -0.00000226, 0.00000307, -0.00000008, -0.00000847, 0.00000553, 0.00000334, 0.00000024, 0.00000284, -0.00000307, -0.00000084, -0.00000467, 0.00000237, 0.00000154, -0.00000237, 0.00000304, -0.00000490, 0.00000574, -0.00000241, -0.00000267, 0.00000347, -0.00000273, 0.00000263, 0.00000760, -0.00001891, 0.00001022, -0.00000459, 0.00000522, -0.00000218, 0.00000047, 0.00000943, -0.00001842, 0.00001731, -0.00001884, 0.00002372, -0.00002018, 0.00000873, -0.00001219, 0.00000416, -0.00000759, 0.00002342, -0.00000708, -0.00001077, 0.00000872, -0.00001757, 0.00003046, -0.00003932, 0.00003627, -0.00004368, 0.00003219, -0.00003278, 0.00005155, -0.00004062, -0.00001116, 0.00002402, -0.00002821, 0.00003194, -0.00003603, 0.00004639, -0.00003836, 0.00005707, -0.00005719, 0.00003885, -0.00004191, 0.00000536, -0.00004161, -0.00000876, -0.00013089, -0.00016734, 0.00058405, 0.00016233, -0.00022078, -0.00001785, -0.00003534, -0.00000577, -0.00001663, -0.00001803, 0.00000660, -0.00002635, 0.00000811, -0.00001163, 0.00000142, 0.00000017, -0.00001999, 0.00000988, -0.00002404, 0.00002126, -0.00000185, -0.00000982, 0.00001471, 0.00000004, 0.00000081, -0.00000051, -0.00000208, -0.00000205, -0.00000560, 0.00001242, 0.00000351, -0.00000350, 0.00000208, -0.00000478, -0.00000340, -0.00000364, 0.00000463, 0.00000056, },
		{ 0.00000224, -0.00000602, 0.00000094, -0.00000455, -0.00000126, 0.00000299, -0.00000059, 0.00000287, -0.00000210, -0.00000257, 0.00000133, -0.00000077, 0.00000288, 0.00000297, -0.00000081, -0.00000478, -0.00000028, 0.00000364, -0.00000367, 0.00000125, 0.00000037, -0.00000156, 0.00000096, -0.00000231, 0.00000352, -0.00000158, -0.00000763, 0.00000682, 0.00000228, 0.00000086, 0.00000230, -0.00000346, -0.00000080, -0.00000457, 0.00000307, 0.00000071, -0.00000178, 0.00000247, -0.00000464, 0.00000600, -0.00000302, -0.00000193, 0.00000333, -0.00000341, 0.00000478, 0.00000436, -0.00001786, 0.00001174, -0.00000637, 0.00000733, -0.00000506, 0.00000460, 0.00000422, -0.00001409, 0.00001438, -0.00001714, 0.00002372, -0.00002190, 0.00001121, -0.00001563, 0.00000710, -0.00000864, 0.00002690, -0.00001464, -0.00000364, 0.00000267, -0.00001010, 0.00002304, -0.00003433, 0.00003381, -0.00004356, 0.00003471, -0.00003731, 0.00006138, -0.00005718, 0.00000774, 0.00000985, -0.00001898, 0.00002625, -0.00003343, 0.00004511, -0.00003982, 0.00006000, -0.00006555, 0.00004917, -0.00005470, 0.00001962, -0.00005773, 0.00000692, -0.00016289, -0.00013652, 0.00064458, 0.00008496, -0.00022180, 0.00000025, -0.00004551, 0.00000321, -0.00002268, -0.00001365, 0.00000452, -0.00002636, 0.00001067, -0.00001525, 0.00000586, -0.00000528, -0.00001518, 0.00000769, -0.00002503, 0.00002571, -0.00000668, -0.00000648, 0.00001467, -0.00000168, 0.00000163, -0.00000104, -0.00000201, -0.00000244, -0.00000450, 0.00001316, 0.00000150, -0.00000267, 0.00000164, -0.00000464, -0.00000319, -0.00000383, 0.00000533, -0.00000022, 0.00000227, },
		{ -0.00000580, 0.00000137, -0.00000538, -0.00000011, 0.00000281, -0.00000076, 0.00000296, -0.00000265, -0.00000188, 0.00000100, -0.00000080, 0.00000338, 0.00000242, -0.00000113, -0.00000466, 0.00000068, 0.00000295, -0.00000382, 0.00000178, -0.00000015, -0.00000125, 0.00000079, -0.00000223, 0.00000381, -0.00000308, -0.00000639, 0.00000773, 0.00000129, 0.00000155, 0.00000155, -0.00000365, -0.00000082, -0.00000434, 0.00000362, -0.00000008, -0.00000115, 0.00000182, -0.00000425, 0.00000615, -0.00000359, -0.00000110, 0.00000288, -0.00000365, 0.00000661, 0.00000085, -0.00001590, 0.00001232, -0.00000741, 0.00000873, -0.00000723, 0.00000816, -0.00000089, -0.00000926, 0.00001051, -0.00001422, 0.00002240, -0.00002239, 0.00001260, -0.00001831, 0.00000968, -0.00000908, 0.00002917, -0.00002127, 0.00000333, -0.00000344, -0.00000228, 0.00001443, -0.00002734, 0.00002886, -0.00004060, 0.00003410, -0.00003842, 0.00006792, -0.00007174, 0.00002700, -0.00000693, -0.00000622, 0.00001685, -0.00002736, 0.00004063, -0.00003822, 0.00006003, -0.00007142, 0.00005787, -0.00006679, 0.00003434, -0.00007588, 0.00002612, -0.00020239, -0.00009083, 0.00069565, -0.00000113, -0.00020921, 0.00001200, -0.00005386, 0.00001229, -0.00002988, -0.00000769, 0.00000067, -0.00002459, 0.00001172, -0.00001780, 0.00000965, -0.00001072, -0.00000943, 0.00000378, -0.00002427, 0.00002916, -0.00001146, -0.00000247, 0.00001371, -0.00000293, 0.00000241, -0.00000165, -0.00000189, -0.00000285, -0.00000317, 0.00001348, -0.00000038, -0.00000169, 0.00000102, -0.00000420, -0.00000330, -0.00000384, 0.00000595, -0.00000103, 0.00000307, 0.00000116, },
		{ 0.00000150, -0.00000606, 0.00000115, 0.00000236, -0.00000074, 0.00000291, -0.00000309, -0.00000120, 0.00000050, -0.00000057, 0.00000371, 0.00000186, -0.00000138, -0.00000444, 0.00000155, 0.00000211, -0.00000377, 0.00000220, -0.00000067, -0.00000087, 0.00000051, -0.00000201, 0.00000387, -0.00000448, -0.00000483, 0.00000822, 0.00000046, 0.00000223, 0.00000065, -0.00000366, -0.00000091, -0.00000399, 0.00000401, -0.00000080, -0.00000052, 0.00000110, -0.00000372, 0.00000618, -0.00000409, -0.00000021, 0.00000216, -0.00000341, 0.00000792, -0.00000264, -0.00001324, 0.00001204, -0.00000772, 0.00000936, -0.00000855, 0.00001087, -0.00000552, -0.00000432, 0.00000603, -0.00001030, 0.00001985, -0.00002162, 0.00001281, -0.00002005, 0.00001178, -0.00000889, 0.00003012, -0.00002667, 0.00000970, -0.00000919, 0.00000536, 0.00000525, -0.00001890, 0.00002185, -0.00003504, 0.00003037, -0.00003576, 0.00007036, -0.00008297, 0.00004507, -0.00002498, 0.00000925, 0.00000410, -0.00001780, 0.00003268, -0.00003307, 0.00005648, -0.00007385, 0.00006378, -0.00007688, 0.00004814, -0.00009471, 0.00004768, -0.00024884, -0.00002860, 0.00073284, -0.00009208, -0.00018363, 0.00001598, -0.00005877, 0.00002014, -0.00003729, -0.00000074, -0.00000464, -0.00002108, 0.00001109, -0.00001894, 0.00001233, -0.00001558, -0.00000326, -0.00000156, -0.00002166, 0.00003124, -0.00001578, 0.00000194, 0.00001192, -0.00000362, 0.00000302, -0.00000225, -0.00000176, -0.00000321, -0.00000165, 0.00001335, -0.00000203, -0.00000068, 0.00000029, -0.00000352, -0.00000376, -0.00000357, 0.00000640, -0.00000179, 0.00000392, -0.00000011, -0.00000525, },
		{ -0.00000650, 0.00000241, 0.00000170, -0.00000053, 0.00000270, -0.00000336, -0.00000060, -0.00000009, -0.00000012, 0.00000383, 0.00000134, -0.00000159, -0.00000414, 0.00000231, 0.00000118, -0.00000351, 0.00000247, -0.00000115, -0.00000044, 0.00000014, -0.00000167, 0.00000368, -0.00000568, -0.00000304, 0.00000831, -0.00000015, 0.00000280, -0.00000034, -0.00000350, -0.00000105, -0.00000352, 0.00000421, -0.00000141, 0.00000008, 0.00000034, -0.00000306, 0.00000605, -0.00000448, 0.00000066, 0.00000125, -0.00000270, 0.00000858, -0.00000584, -0.00001017, 0.00001105, -0.00000735, 0.00000925, -0.00000896, 0.00001257, -0.00000937, 0.00000035, 0.00000132, -0.00000570, 0.00001629, -0.00001971, 0.00001180, -0.00002074, 0.00001330, -0.00000807, 0.00002974, -0.00003060, 0.00001514, -0.00001417, 0.00001230, -0.00000387, -0.00000968, 0.00001334, -0.00002734, 0.00002380, -0.00002934, 0.00006819, -0.00008979, 0.00006041, -0.00004279, 0.00002627, -0.00001127, -0.00000510, 0.00002130, -0.00002419, 0.00004885, -0.00007206, 0.00006583, -0.00008362, 0.00005943, -0.00011249, 0.00006979, -0.00030081, 0.00005102, 0.00075196, -0.00018336, -0.00014667, 0.00001152, -0.00005895, 0.00002549, -0.00004385, 0.00000641, -0.00001089, -0.00001609, 0.00000878, -0.00001849, 0.00001356, -0.00001934, 0.00000273, -0.00000789, -0.00001731, 0.00003172, -0.00001927, 0.00000640, 0.00000945, -0.00000373, 0.00000340, -0.00000280, -0.00000166, -0.00000349, -0.00000002, 0.00001279, -0.00000337, 0.00000027, -0.00000043, -0.00000268, -0.00000454, -0.00000297, 0.00000660, -0.00000240, 0.00000472, -0.00000152, -0.00000438, 0.00000129, },
		{ 0.00000358, 0.00000091, -0.00000015, 0.00000232, -0.00000342, -0.00000016, -0.00000068, 0.00000050, 0.00000373, 0.00000091, -0.00000177, -0.00000372, 0.00000292, 0.00000020, -0.00000305, 0.00000257, -0.00000153, -0.00000002, -0.00000029, -0.00000121, 0.00000323, -0.00000660, -0.00000112, 0.00000803, -0.00000050, 0.00000319, -0.00000134, -0.00000322, -0.00000125, -0.00000295, 0.00000423, -0.00000189, 0.00000061, -0.00000043, -0.00000228, 0.00000576, -0.00000470, 0.00000144, 0.00000024, -0.00000156, 0.00000852, -0.00000852, -0.00000695, 0.00000954, -0.00000644, 0.00000849, -0.00000851, 0.00001317, -0.00001220, 0.00000440, -0.00000324, -0.00000075, 0.00001197, -0.00001680, 0.00000965, -0.00002034, 0.00001421, -0.00000668, 0.00002807, -0.00003295, 0.00001941, -0.00001807, 0.00001810, -0.00001230, -0.00000035, 0.00000402, -0.00001814, 0.00001495, -0.00001948, 0.00006129, -0.00009141, 0.00007167, -0.00005876, 0.00004345, -0.00002821, 0.00001003, 0.00000692, -0.00001170, 0.00003700, -0.00006555, 0.00006313, -0.00008575, 0.00006654, -0.00012722, 0.00009012, -0.00035599, 0.00014780, 0.00074938, -0.00027004, -0.00010091, -0.00000123, -0.00005356, 0.00002726, -0.00004852, 0.00001288, -0.00001739, -0.00001008, 0.00000503, -0.00001644, 0.00001311, -0.00002154, 0.00000790, -0.00001465, -0.00001148, 0.00003053, -0.00002161, 0.00001057, 0.00000651, -0.00000331, 0.00000349, -0.00000324, -0.00000164, -0.00000360, 0.00000164, 0.00001183, -0.00000437, 0.00000110, -0.00000105, -0.00000179, -0.00000560, -0.00000202, 0.00000649, -0.00000279, 0.00000538, -0.00000296, -0.00000324, 0.00000072, -0.00000660, },
		{ 0.00000007, 0.00000035, 0.00000179, -0.00000328, 0.00000006, -0.00000116, 0.00000122, 0.00000341, 0.00000059, -0.00000195, -0.00000320, 0.00000334, -0.00000077, -0.00000244, 0.00000249, -0.00000180, 0.00000036, -0.00000073, -0.00000069, 0.00000253, -0.00000717, 0.00000079, 0.00000745, -0.00000059, 0.00000336, -0.00000226, -0.00000286, -0.00000147, -0.00000231, 0.00000407, -0.00000221, 0.00000103, -0.00000115, -0.00000142, 0.00000530, -0.00000473, 0.00000204, -0.00000075, -0.00000012, 0.00000774, -0.00001051, -0.00000384, 0.00000775, -0.00000516, 0.00000724, -0.00000731, 0.00001269, -0.00001385, 0.00000755, -0.00000728, 0.00000415, 0.00000723, -0.00001316, 0.00000649, -0.00001885, 0.00001447, -0.00000482, 0.00002525, -0.00003371, 0.00002234, -0.00002068, 0.00002240, -0.00001950, 0.00000840, -0.00000537, -0.00000822, 0.00000460, -0.00000685, 0.00004993, -0.00008747, 0.00007776, -0.00007141, 0.00005927, -0.00004542, 0.00002656, -0.00000971, 0.00000389, 0.00002113, -0.00005416, 0.00005511, -0.00008223, 0.00006793, -0.00013678, 0.00010591, -0.00041116, 0.00026024, 0.00072234, -0.00034711, -0.00004971, -0.00002131, -0.00004237, 0.00002471, -0.00005039, 0.00001778, -0.00002334, -0.00000370, 0.00000024, -0.00001299, 0.00001096, -0.00002189, 0.00001171, -0.00002121, -0.00000459, 0.00002773, -0.00002262, 0.00001414, 0.00000336, -0.00000244, 0.00000331, -0.00000355, -0.00000171, -0.00000351, 0.00000322, 0.00001054, -0.00000500, 0.00000173, -0.00000149, -0.00000098, -0.00000684, -0.00000074, 0.00000602, -0.00000288, 0.00000581, -0.00000433, -0.00000192, -0.00000019, -0.00000632, 0.00000453, },
		{ 0.00000090, 0.00000117, -0.00000296, 0.00000003, -0.00000147, 0.00000195, 0.00000293, 0.00000040, -0.00000215, -0.00000257, 0.00000354, -0.00000166, -0.00000171, 0.00000224, -0.00000192, 0.00000065, -0.00000116, -0.00000013, 0.00000161, -0.00000734, 0.00000259, 0.00000665, -0.00000045, 0.00000327, -0.00000306, -0.00000247, -0.00000168, -0.00000162, 0.00000376, -0.00000237, 0.00000132, -0.00000179, -0.00000051, 0.00000469, -0.00000455, 0.00000242, -0.00000160, 0.00000148, 0.00000631, -0.00001170, -0.00000107, 0.00000589, -0.00000370, 0.00000570, -0.00000555, 0.00001124, -0.00001430, 0.00000959, -0.00001051, 0.00000865, 0.00000242, -0.00000907, 0.00000253, -0.00001637, 0.00001410, -0.00000260, 0.00002147, -0.00003296, 0.00002392, -0.00002192, 0.00002499, -0.00002504, 0.00001596, -0.00001410, 0.00000161, -0.00000635, 0.00000765, 0.00003477, -0.00007805, 0.00007801, -0.00007946, 0.00007225, -0.00006145, 0.00004323, -0.00002751, 0.00002168, 0.00000189, -0.00003818, 0.00004163, -0.00007241, 0.00006229, -0.00013915, 0.00011415, -0.00046232, 0.00038554, 0.00066932, -0.00040986, 0.00000307, -0.00004697, -0.00002578, 0.00001752, -0.00004879, 0.00002030, -0.00002792, 0.00000233, -0.00000501, -0.00000855, 0.00000730, -0.00002027, 0.00001371, -0.00002693, 0.00000285, 0.00002356, -0.00002221, 0.00001685, 0.00000024, -0.00000124, 0.00000287, -0.00000372, -0.00000188, -0.00000316, 0.00000464, 0.00000899, -0.00000527, 0.00000214, -0.00000166, -0.00000037, -0.00000813, 0.00000082, 0.00000519, -0.00000262, 0.00000591, -0.00000551, -0.00000053, -0.00000138, -0.00000563, 0.00000519, -0.00000071, },
		{ 0.00000050, -0.00000249, -0.00000023, -0.00000155, 0.00000260, 0.00000234, 0.00000034, -0.00000235, -0.00000184, 0.00000351, -0.00000242, -0.00000093, 0.00000184, -0.00000189, 0.00000083, -0.00000152, 0.00000040, 0.00000054, -0.00000711, 0.00000417, 0.00000574, -0.00000013, 0.00000293, -0.00000368, -0.00000209, -0.00000186, -0.00000095, 0.00000333, -0.00000238, 0.00000145, -0.00000231, 0.00000039, 0.00000397, -0.00000418, 0.00000253, -0.00000221, 0.00000307, 0.00000436, -0.00001208, 0.00000121, 0.00000415, -0.00000226, 0.00000408, -0.00000347, 0.00000901, -0.00001362, 0.00001046, -0.00001269, 0.00001244, -0.00000212, -0.00000488, -0.00000195, -0.00001302, 0.00001316, -0.00000015, 0.00001695, -0.00003086, 0.00002420, -0.00002182, 0.00002580, -0.00002865, 0.00002186, -0.00002154, 0.00001056, -0.00001693, 0.00002292, 0.00001684, -0.00006370, 0.00007220, -0.00008200, 0.00008108, -0.00007488, 0.00005864, -0.00004517, 0.00004050, -0.00001973, -0.00001834, 0.00002301, -0.00005612, 0.00004877, -0.00013258, 0.00011188, -0.00050490, 0.00051959, 0.00059022, -0.00045425, 0.00005331, -0.00007585, -0.00000480, 0.00000588, -0.00004342, 0.00001988, -0.00003043, 0.00000725, -0.00001004, -0.00000366, 0.00000252, -0.00001679, 0.00001361, -0.00003124, 0.00001023, 0.00001835, -0.00002045, 0.00001852, -0.00000261, 0.00000013, 0.00000223, -0.00000377, -0.00000213, -0.00000256, 0.00000581, 0.00000729, -0.00000522, 0.00000233, -0.00000152, -0.00000008, -0.00000935, 0.00000255, 0.00000404, -0.00000199, 0.00000561, -0.00000638, 0.00000080, -0.00000276, -0.00000456, 0.00000549, -0.00000136, 0.00000143, },
		{ -0.00000196, -0.00000070, -0.00000135, 0.00000309, 0.00000171, 0.00000036, -0.00000256, -0.00000104, 0.00000324, -0.00000300, -0.00000016, 0.00000133, -0.00000171, 0.00000087, -0.00000178, 0.00000084, -0.00000062, -0.00000648, 0.00000547, 0.00000480, 0.00000031, 0.00000236, -0.00000408, -0.00000177, -0.00000196, -0.00000032, 0.00000282, -0.00000225, 0.00000143, -0.00000267, 0.00000123, 0.00000318, -0.00000365, 0.00000236, -0.00000248, 0.00000448, 0.00000208, -0.00001168, 0.00000289, 0.00000268, -0.00000100, 0.00000258, -0.00000132, 0.00000627, -0.00001197, 0.00001015, -0.00001371, 0.00001527, -0.00000607, -0.00000092, -0.00000664, -0.00000901, 0.00001172, 0.00000238, 0.00001196, -0.00002765, 0.00002333, -0.00002051, 0.00002488, -0.00003020, 0.00002577, -0.00002719, 0.00001792, -0.00002620, 0.00003775, -0.00000262, -0.00004540, 0.00006065, -0.00007856, 0.00008475, -0.00008442, 0.00007140, -0.00006127, 0.00005892, -0.00004238, 0.00000420, 0.00000013, -0.00003377, 0.00002713, -0.00011591, 0.00009641, -0.00053403, 0.00065710, 0.00048649, -0.00047725, 0.00009701, -0.00010518, 0.00001899, -0.00000953, -0.00003435, 0.00001621, -0.00003030, 0.00001040, -0.00001415, 0.00000103, -0.00000284, -0.00001178, 0.00001136, -0.00003371, 0.00001698, 0.00001257, -0.00001751, 0.00001908, -0.00000501, 0.00000153, 0.00000146, -0.00000375, -0.00000243, -0.00000170, 0.00000664, 0.00000555, -0.00000492, 0.00000233, -0.00000108, -0.00000020, -0.00001033, 0.00000433, 0.00000265, -0.00000103, 0.00000489, -0.00000684, 0.00000193, -0.00000420, -0.00000316, 0.00000541, -0.00000179, 0.00000187, -0.00000014, },
		{ -0.00000131, -0.00000088, 0.00000336, 0.00000110, 0.00000044, -0.00000274, -0.00000020, 0.00000275, -0.00000336, 0.00000055, 0.00000076, -0.00000140, 0.00000077, -0.00000191, 0.00000115, -0.00000178, -0.00000551, 0.00000644, 0.00000392, 0.00000078, 0.00000162, -0.00000428, -0.00000152, -0.00000197, 0.00000023, 0.00000228, -0.00000202, 0.00000126, -0.00000285, 0.00000197, 0.00000239, -0.00000300, 0.00000194, -0.00000237, 0.00000556, -0.00000034, -0.00001064, 0.00000395, 0.00000158, -0.00000005, 0.00000138, 0.00000064, 0.00000330, -0.00000959, 0.00000881, -0.00001353, 0.00001698, -0.00000917, 0.00000250, -0.00001122, -0.00000455, 0.00000990, 0.00000485, 0.00000676, -0.00002358, 0.00002152, -0.00001821, 0.00002245, -0.00002975, 0.00002755, -0.00003072, 0.00002318, -0.00003334, 0.00005099, -0.00002217, -0.00002446, 0.00004414, -0.00006919, 0.00008267, -0.00008904, 0.00008024, -0.00007439, 0.00007542, -0.00006447, 0.00002792, -0.00002568, -0.00000639, -0.00000219, -0.00008864, 0.00006564, -0.00054488, 0.00079182, 0.00036120, -0.00047710, 0.00013060, -0.00013200, 0.00004361, -0.00002753, -0.00002212, 0.00000935, -0.00002726, 0.00001123, -0.00001669, 0.00000482, -0.00000809, -0.00000577, 0.00000712, -0.00003408, 0.00002258, 0.00000671, -0.00001369, 0.00001857, -0.00000682, 0.00000283, 0.00000063, -0.00000368, -0.00000274, -0.00000061, 0.00000710, 0.00000387, -0.00000444, 0.00000219, -0.00000035, -0.00000079, -0.00001095, 0.00000603, 0.00000111, 0.00000022, 0.00000374, -0.00000682, 0.00000274, -0.00000556, -0.00000153, 0.00000496, -0.00000197, 0.00000214, -0.00000068, -0.00000144, },
		{ -0.00000017, 0.00000336, 0.00000060, 0.00000054, -0.00000288, 0.00000064, 0.00000207, -0.00000348, 0.00000115, 0.00000017, -0.00000099, 0.00000053, -0.00000191, 0.00000128, -0.00000286, -0.00000426, 0.00000705, 0.00000318, 0.00000120, 0.00000076, -0.00000428, -0.00000137, -0.00000187, 0.00000066, 0.00000176, -0.00000173, 0.00000096, -0.00000284, 0.00000255, 0.00000163, -0.00000229, 0.00000132, -0.00000187, 0.00000618, -0.00000268, -0.00000910, 0.00000442, 0.00000090, 0.00000052, 0.00000061, 0.00000220, 0.00000040, -0.00000678, 0.00000664, -0.00001225, 0.00001752, -0.00001126, 0.00000511, -0.00001539, 0.00000010, 0.00000780, 0.00000712, 0.00000161, -0.00001895, 0.00001902, -0.00001520, 0.00001879, -0.00002749, 0.00002725, -0.00003199, 0.00002600, -0.00003774, 0.00006157, -0.00004039, -0.00000244, 0.00002388, -0.00005443, 0.00007471, -0.00008811, 0.00008418, -0.00008327, 0.00008849, -0.00008432, 0.00005102, -0.00005265, 0.00002440, -0.00003799, -0.00005121, 0.00001836, -0.00053306, 0.00091694, 0.00021894, -0.00045347, 0.00015123, -0.00015348, 0.00006686, -0.00004661, -0.00000762, -0.00000027, -0.00002133, 0.00000947, -0.00001718, 0.00000707, -0.00001250, 0.00000057, 0.00000131, -0.00003228, 0.00002662, 0.00000125, -0.00000932, 0.00001708, -0.00000797, 0.00000393, -0.00000018, -0.00000364, -0.00000299, 0.00000066, 0.00000714, 0.00000235, -0.00000387, 0.00000200, 0.00000059, -0.00000187, -0.00001110, 0.00000749, -0.00000043, 0.00000164, 0.00000222, -0.00000630, 0.00000312, -0.00000671, 0.00000021, 0.00000420, -0.00000187, 0.00000223, -0.00000105, -0.00000102, -0.00000198, },
		{ 0.00000310, 0.00000023, 0.00000062, -0.00000295, 0.00000143, 0.00000124, -0.00000335, 0.00000158, -0.00000037, -0.00000054, 0.00000018, -0.00000177, 0.00000120, -0.00000376, -0.00000282, 0.00000732, 0.00000261, 0.00000151, -0.00000013, -0.00000412, -0.00000130, -0.00000167, 0.00000094, 0.00000130, -0.00000142, 0.00000056, -0.00000265, 0.00000296, 0.00000098, -0.00000161, 0.00000058, -0.00000101, 0.00000626, -0.00000475, -0.00000727, 0.00000441, 0.00000060, 0.00000071, 0.00000033, 0.00000318, -0.00000214, -0.00000385, 0.00000392, -0.00001003, 0.00001693, -0.00001225, 0.00000673, -0.00001886, 0.00000468, 0.00000557, 0.00000910, -0.00000326, -0.00001402, 0.00001608, -0.00001179, 0.00001428, -0.00002376, 0.00002506, -0.00003109, 0.00002629, -0.00003901, 0.00006869, -0.00005592, 0.00001900, 0.00000142, -0.00003530, 0.00006125, -0.00008144, 0.00008259, -0.00008693, 0.00009682, -0.00010027, 0.00007161, -0.00007870, 0.00005648, -0.00007833, -0.00000499, -0.00004552, -0.00049497, 0.00102540, 0.00006556, -0.00040760, 0.00015707, -0.00016715, 0.00008651, -0.00006502, 0.00000790, -0.00001188, -0.00001286, 0.00000510, -0.00001531, 0.00000729, -0.00001538, 0.00000645, -0.00000551, -0.00002850, 0.00002886, -0.00000338, -0.00000481, 0.00001483, -0.00000847, 0.00000475, -0.00000092, -0.00000364, -0.00000312, 0.00000202, 0.00000678, 0.00000106, -0.00000330, 0.00000185, 0.00000164, -0.00000340, -0.00001070, 0.00000858, -0.00000183, 0.00000312, 0.00000042, -0.00000531, 0.00000301, -0.00000750, 0.00000191, 0.00000321, -0.00000153, 0.00000210, -0.00000123, -0.00000077, -0.00000262, 0.00000073, },
		{ 0.00000005, 0.00000062, -0.00000290, 0.00000212, 0.00000031, -0.00000301, 0.00000182, -0.00000082, -0.00000008, -0.00000025, -0.00000152, 0.00000091, -0.00000443, -0.00000130, 0.00000729, 0.00000223, 0.00000164, -0.00000099, -0.00000384, -0.00000131, -0.00000137, 0.00000108, 0.00000093, -0.00000114, 0.00000012, -0.00000229, 0.00000317, 0.00000048, -0.00000102, -0.00000019, 0.00000013, 0.00000577, -0.00000639, -0.00000534, 0.00000403, 0.00000061, 0.00000056, 0.00000055, 0.00000350, -0.00000412, -0.00000111, 0.00000094, -0.00000714, 0.00001535, -0.00001216, 0.00000724, -0.00002140, 0.00000895, 0.00000334, 0.00001069, -0.00000767, -0.00000907, 0.00001297, -0.00000830, 0.00000932, -0.00001895, 0.00002134, -0.00002828, 0.00002421, -0.00003709, 0.00007180, -0.00006767, 0.00003824, -0.00002152, -0.00001320, 0.00004316, -0.00006935, 0.00007530, -0.00008478, 0.00009938, -0.00011089, 0.00008783, -0.00010163, 0.00008736, -0.00012061, 0.00004764, -0.00012482, -0.00042824, 0.00111046, -0.00009209, -0.00034219, 0.00014741, -0.00017114, 0.00010053, -0.00008095, 0.00002300, -0.00002441, -0.00000254, -0.00000157, -0.00001106, 0.00000517, -0.00001615, 0.00001112, -0.00001259, -0.00002313, 0.00002927, -0.00000686, -0.00000052, 0.00001204, -0.00000838, 0.00000527, -0.00000154, -0.00000374, -0.00000308, 0.00000341, 0.00000605, 0.00000006, -0.00000279, 0.00000183, 0.00000267, -0.00000529, -0.00000973, 0.00000920, -0.00000295, 0.00000449, -0.00000154, -0.00000389, 0.00000239, -0.00000783, 0.00000344, 0.00000210, -0.00000099, 0.00000177, -0.00000120, -0.00000074, -0.00000311, 0.00000171, 0.00000259, },
		{ 0.00000055, -0.00000273, 0.00000265, -0.00000063, -0.00000247, 0.00000187, -0.00000114, 0.00000034, -0.00000072, -0.00000119, 0.00000042, -0.00000480, 0.00000021, 0.00000703, 0.00000203, 0.00000158, -0.00000175, -0.00000349, -0.00000135, -0.00000101, 0.00000107, 0.00000068, -0.00000094, -0.00000031, -0.00000181, 0.00000319, 0.00000014, -0.00000057, -0.00000089, 0.00000145, 0.00000476, -0.00000747, -0.00000351, 0.00000345, 0.00000084, 0.00000015, 0.00000119, 0.00000312, -0.00000536, 0.00000118, -0.00000196, -0.00000385, 0.00001298, -0.00001108, 0.00000661, -0.00002284, 0.00001270, 0.00000121, 0.00001182, -0.00001144, -0.00000431, 0.00000989, -0.00000500, 0.00000431, -0.00001353, 0.00001651, -0.00002394, 0.00002010, -0.00003217, 0.00007073, -0.00007482, 0.00005386, -0.00004315, 0.00001019, 0.00002170, -0.00005261, 0.00006264, -0.00007669, 0.00009557, -0.00011506, 0.00009806, -0.00011931, 0.00011437, -0.00016176, 0.00010343, -0.00021699, -0.00033196, 0.00116608, -0.00024661, -0.00026125, 0.00012277, -0.00016439, 0.00010731, -0.00009270, 0.00003616, -0.00003661, 0.00000871, -0.00000995, -0.00000472, 0.00000067, -0.00001444, 0.00001387, -0.00001915, -0.00001671, 0.00002799, -0.00000901, 0.00000322, 0.00000899, -0.00000779, 0.00000549, -0.00000204, -0.00000393, -0.00000280, 0.00000471, 0.00000499, -0.00000062, -0.00000241, 0.00000203, 0.00000352, -0.00000742, -0.00000822, 0.00000929, -0.00000368, 0.00000562, -0.00000349, -0.00000217, 0.00000127, -0.00000763, 0.00000466, 0.00000099, -0.00000033, 0.00000130, -0.00000098, -0.00000098, -0.00000338, 0.00000270, 0.00000188, 0.00000006, },
		{ -0.00000242, 0.00000300, -0.00000154, -0.00000179, 0.00000172, -0.00000130, 0.00000066, -0.00000119, -0.00000082, -0.00000023, -0.00000484, 0.00000162, 0.00000661, 0.00000199, 0.00000130, -0.00000235, -0.00000312, -0.00000141, -0.00000060, 0.00000094, 0.00000055, -0.00000084, -0.00000069, -0.00000125, 0.00000304, -0.00000001, -0.00000033, -0.00000142, 0.00000281, 0.00000328, -0.00000796, -0.00000192, 0.00000280, 0.00000115, -0.00000038, 0.00000213, 0.00000212, -0.00000581, 0.00000281, -0.00000451, -0.00000048, 0.00001011, -0.00000921, 0.00000493, -0.00002311, 0.00001575, -0.00000069, 0.00001246, -0.00001449, 0.00000009, 0.00000701, -0.00000214, -0.00000038, -0.00000794, 0.00001106, -0.00001858, 0.00001448, -0.00002473, 0.00006565, -0.00007698, 0.00006474, -0.00006182, 0.00003310, -0.00000157, -0.00003240, 0.00004539, -0.00006304, 0.00008531, -0.00011214, 0.00010107, -0.00012985, 0.00013490, -0.00019841, 0.00015837, -0.00031812, -0.00020699, 0.00118746, -0.00039049, -0.00016988, 0.00008488, -0.00014676, 0.00010580, -0.00009889, 0.00004595, -0.00004715, 0.00001976, -0.00001920, 0.00000317, -0.00000595, -0.00001010, 0.00001416, -0.00002442, -0.00000994, 0.00002534, -0.00000979, 0.00000617, 0.00000591, -0.00000684, 0.00000544, -0.00000241, -0.00000423, -0.00000227, 0.00000585, 0.00000369, -0.00000099, -0.00000219, 0.00000251, 0.00000406, -0.00000962, -0.00000625, 0.00000882, -0.00000393, 0.00000635, -0.00000527, -0.00000030, -0.00000025, -0.00000690, 0.00000547, -0.00000001, 0.00000037, 0.00000073, -0.00000061, -0.00000148, -0.00000336, 0.00000358, 0.00000104, 0.00000024, 0.00000037, },
		{ 0.00000313, -0.00000235, -0.00000103, 0.00000141, -0.00000130, 0.00000086, -0.00000160, -0.00000046, -0.00000100, -0.00000456, 0.00000284, 0.00000610, 0.00000204, 0.00000084, -0.00000276, -0.00000279, -0.00000143, -0.00000020, 0.00000072, 0.00000054, -0.00000086, -0.00000097, -0.00000067, 0.00000276, -0.00000000, -0.00000029, -0.00000170, 0.00000407, 0.00000149, -0.00000784, -0.00000068, 0.00000223, 0.00000143, -0.00000089, 0.00000322, 0.00000061, -0.00000547, 0.00000366, -0.00000646, 0.00000266, 0.00000702, -0.00000680, 0.00000235, -0.00002220, 0.00001797, -0.00000229, 0.00001260, -0.00001675, 0.00000400, 0.00000446, 0.00000014, -0.00000446, -0.00000262, 0.00000547, -0.00001274, 0.00000798, -0.00001543, 0.00005707, -0.00007416, 0.00007015, -0.00007613, 0.00005379, -0.00002491, -0.00001024, 0.00002474, -0.00004470, 0.00006904, -0.00010207, 0.00009616, -0.00013182, 0.00014662, -0.00022723, 0.00020799, -0.00042312, -0.00005601, 0.00117137, -0.00051666, -0.00007388, 0.00003652, -0.00011902, 0.00009567, -0.00009857, 0.00005117, -0.00005476, 0.00002940, -0.00002829, 0.00001181, -0.00001418, -0.00000331, 0.00001170, -0.00002772, -0.00000353, 0.00002180, -0.00000934, 0.00000819, 0.00000303, -0.00000565, 0.00000517, -0.00000269, -0.00000460, -0.00000147, 0.00000673, 0.00000223, -0.00000107, -0.00000212, 0.00000331, 0.00000416, -0.00001171, -0.00000395, 0.00000785, -0.00000366, 0.00000660, -0.00000671, 0.00000156, -0.00000205, -0.00000565, 0.00000582, -0.00000079, 0.00000101, 0.00000015, -0.00000016, -0.00000220, -0.00000301, 0.00000426, 0.00000014, 0.00000058, 0.00000010, -0.00000197, },
		{ -0.00000299, -0.00000025, 0.00000097, -0.00000114, 0.00000092, -0.00000192, -0.00000016, -0.00000181, -0.00000398, 0.00000384, 0.00000559, 0.00000213, 0.00000022, -0.00000297, -0.00000252, -0.00000140, 0.00000015, 0.00000045, 0.00000063, -0.00000101, -0.00000111, -0.00000013, 0.00000240, 0.00000015, -0.00000048, -0.00000167, 0.00000509, -0.00000048, -0.00000719, 0.00000014, 0.00000183, 0.00000157, -0.00000125, 0.00000426, -0.00000123, -0.00000444, 0.00000369, -0.00000765, 0.00000532, 0.00000402, -0.00000412, -0.00000092, -0.00002019, 0.00001930, -0.00000351, 0.00001226, -0.00001820, 0.00000734, 0.00000230, 0.00000173, -0.00000770, 0.00000211, 0.00000019, -0.00000696, 0.00000127, -0.00000510, 0.00004578, -0.00006680, 0.00006984, -0.00008504, 0.00007069, -0.00004660, 0.00001222, 0.00000221, -0.00002293, 0.00004772, -0.00008535, 0.00008325, -0.00012440, 0.00014771, -0.00024515, 0.00024766, -0.00052593, 0.00011647, 0.00111650, -0.00061890, 0.00002069, -0.00001866, -0.00008287, 0.00007733, -0.00009137, 0.00005100, -0.00005839, 0.00003649, -0.00003612, 0.00002023, -0.00002321, 0.00000546, 0.00000648, -0.00002857, 0.00000182, 0.00001794, -0.00000791, 0.00000927, 0.00000049, -0.00000436, 0.00000475, -0.00000290, -0.00000502, -0.00000040, 0.00000728, 0.00000073, -0.00000092, -0.00000217, 0.00000444, 0.00000369, -0.00001349, -0.00000149, 0.00000646, -0.00000289, 0.00000628, -0.00000769, 0.00000322, -0.00000398, -0.00000398, 0.00000568, -0.00000127, 0.00000151, -0.00000037, 0.00000028, -0.00000306, -0.00000232, 0.00000467, -0.00000072, 0.00000101, -0.00000024, -0.00000140, 0.00000303, },
		{ 0.00000049, 0.00000045, -0.00000084, 0.00000083, -0.00000213, 0.00000003, -0.00000259, -0.00000316, 0.00000457, 0.00000513, 0.00000218, -0.00000050, -0.00000299, -0.00000234, -0.00000129, 0.00000042, 0.00000017, 0.00000078, -0.00000128, -0.00000109, 0.00000032, 0.00000201, 0.00000039, -0.00000084, -0.00000130, 0.00000576, -0.00000245, -0.00000611, 0.00000054, 0.00000166, 0.00000149, -0.00000134, 0.00000509, -0.00000317, -0.00000288, 0.00000295, -0.00000800, 0.00000730, 0.00000138, -0.00000149, -0.00000458, -0.00001724, 0.00001972, -0.00000433, 0.00001147, -0.00001886, 0.00001007, 0.00000053, 0.00000263, -0.00000999, 0.00000599, -0.00000444, -0.00000169, -0.00000500, 0.00000538, 0.00003278, -0.00005569, 0.00006403, -0.00008800, 0.00008260, -0.00006504, 0.00003324, -0.00002050, 0.00000063, 0.00002278, -0.00006306, 0.00006296, -0.00010752, 0.00013710, -0.00024970, 0.00027296, -0.00061988, 0.00030427, 0.00102361, -0.00069233, 0.00010789, -0.00007648, -0.00004076, 0.00005194, -0.00007752, 0.00004509, -0.00005728, 0.00004005, -0.00004163, 0.00002738, -0.00003209, 0.00001549, -0.00000122, -0.00002673, 0.00000553, 0.00001432, -0.00000587, 0.00000951, -0.00000162, -0.00000306, 0.00000421, -0.00000306, -0.00000543, 0.00000091, 0.00000744, -0.00000071, -0.00000060, -0.00000231, 0.00000583, 0.00000262, -0.00001480, 0.00000093, 0.00000479, -0.00000167, 0.00000540, -0.00000810, 0.00000452, -0.00000585, -0.00000200, 0.00000509, -0.00000142, 0.00000181, -0.00000074, 0.00000061, -0.00000399, -0.00000134, 0.00000476, -0.00000148, 0.00000148, -0.00000063, -0.00000074, 0.00000271, -0.00000343, },
		{ -0.00000009, -0.00000045, 0.00000059, -0.00000222, 0.00000008, -0.00000327, -0.00000214, 0.00000505, 0.00000477, 0.00000214, -0.00000124, -0.00000284, -0.00000225, -0.00000108, 0.00000058, -0.00000007, 0.00000096, -0.00000164, -0.00000091, 0.00000064, 0.00000164, 0.00000066, -0.00000133, -0.00000063, 0.00000600, -0.00000426, -0.00000474, 0.00000056, 0.00000174, 0.00000117, -0.00000109, 0.00000554, -0.00000500, -0.00000101, 0.00000158, -0.00000752, 0.00000848, -0.00000070, 0.00000083, -0.00000835, -0.00001356, 0.00001928, -0.00000474, 0.00001029, -0.00001879, 0.00001220, -0.00000086, 0.00000289, -0.00001131, 0.00000887, -0.00000816, 0.00000271, -0.00001027, 0.00001516, 0.00001917, -0.00004193, 0.00005340, -0.00008495, 0.00008873, -0.00007893, 0.00005126, -0.00004166, 0.00002420, -0.00000402, -0.00003680, 0.00003652, -0.00008187, 0.00011456, -0.00023926, 0.00028004, -0.00069814, 0.00049983, 0.00089559, -0.00073370, 0.00018236, -0.00013254, 0.00000431, 0.00002130, -0.00005788, 0.00003361, -0.00005111, 0.00003941, -0.00004393, 0.00003224, -0.00003976, 0.00002580, -0.00001078, -0.00002225, 0.00000719, 0.00001151, -0.00000365, 0.00000907, -0.00000329, -0.00000180, 0.00000360, -0.00000321, -0.00000581, 0.00000242, 0.00000718, -0.00000200, -0.00000020, -0.00000243, 0.00000742, 0.00000093, -0.00001549, 0.00000313, 0.00000301, -0.00000014, 0.00000399, -0.00000788, 0.00000532, -0.00000749, 0.00000012, 0.00000411, -0.00000124, 0.00000187, -0.00000092, 0.00000076, -0.00000487, -0.00000013, 0.00000451, -0.00000205, 0.00000193, -0.00000101, -0.00000003, 0.00000218, -0.00000365, 0.00000112, },
		{ 0.00000000, 0.00000023, -0.00000219, -0.00000002, -0.00000379, -0.00000103, 0.00000529, 0.00000452, 0.00000197, -0.00000194, -0.00000257, -0.00000225, -0.00000079, 0.00000062, -0.00000023, 0.00000110, -0.00000205, -0.00000057, 0.00000081, 0.00000135, 0.00000090, -0.00000186, 0.00000032, 0.00000576, -0.00000575, -0.00000324, 0.00000030, 0.00000202, 0.00000061, -0.00000047, 0.00000550, -0.00000650, 0.00000093, -0.00000023, -0.00000630, 0.00000883, -0.00000205, 0.00000258, -0.00001192, -0.00000939, 0.00001808, -0.00000477, 0.00000880, -0.00001805, 0.00001374, -0.00000194, 0.00000260, -0.00001172, 0.00001072, -0.00001086, 0.00000601, -0.00001416, 0.00002348, 0.00000607, -0.00002680, 0.00003901, -0.00007633, 0.00008876, -0.00008736, 0.00006496, -0.00005968, 0.00004596, -0.00003072, -0.00000853, 0.00000575, -0.00004893, 0.00008085, -0.00021327, 0.00026604, -0.00075418, 0.00069460, 0.00073733, -0.00074167, 0.00023972, -0.00018261, 0.00004904, -0.00001229, -0.00003386, 0.00001728, -0.00004003, 0.00003424, -0.00004236, 0.00003394, -0.00004515, 0.00003531, -0.00002137, -0.00001547, 0.00000664, 0.00000994, -0.00000168, 0.00000822, -0.00000459, -0.00000061, 0.00000292, -0.00000336, -0.00000609, 0.00000407, 0.00000648, -0.00000305, 0.00000021, -0.00000246, 0.00000907, -0.00000133, -0.00001547, 0.00000492, 0.00000131, 0.00000155, 0.00000216, -0.00000705, 0.00000554, -0.00000872, 0.00000221, 0.00000286, -0.00000075, 0.00000169, -0.00000089, 0.00000065, -0.00000558, 0.00000122, 0.00000395, -0.00000239, 0.00000232, -0.00000134, 0.00000068, 0.00000149, -0.00000363, 0.00000161, -0.00000059, },
		{ -0.00000021, -0.00000206, -0.00000029, -0.00000409, 0.00000011, 0.00000534, 0.00000438, 0.00000162, -0.00000254, -0.00000223, -0.00000230, -0.00000042, 0.00000052, -0.00000028, 0.00000118, -0.00000246, -0.00000010, 0.00000081, 0.00000117, 0.00000105, -0.00000237, 0.00000144, 0.00000506, -0.00000682, -0.00000177, -0.00000014, 0.00000244, -0.00000013, 0.00000049, 0.00000491, -0.00000751, 0.00000272, -0.00000226, -0.00000450, 0.00000839, -0.00000262, 0.00000361, -0.00001502, -0.00000501, 0.00001626, -0.00000446, 0.00000708, -0.00001674, 0.00001475, -0.00000277, 0.00000192, -0.00001133, 0.00001160, -0.00001252, 0.00000814, -0.00001641, 0.00002980, -0.00000553, -0.00001164, 0.00002218, -0.00006301, 0.00008293, -0.00008990, 0.00007338, -0.00007321, 0.00006420, -0.00005533, 0.00001956, -0.00002707, -0.00001085, 0.00003764, -0.00017235, 0.00022933, -0.00078224, 0.00087955, 0.00055549, -0.00071685, 0.00027684, -0.00022291, 0.00009011, -0.00004622, -0.00000734, -0.00000269, -0.00002465, 0.00002466, -0.00003665, 0.00003186, -0.00004734, 0.00004291, -0.00003196, -0.00000702, 0.00000396, 0.00000987, -0.00000033, 0.00000722, -0.00000561, 0.00000054, 0.00000218, -0.00000349, -0.00000626, 0.00000580, 0.00000534, -0.00000379, 0.00000054, -0.00000230, 0.00001064, -0.00000405, -0.00001471, 0.00000617, -0.00000013, 0.00000321, 0.00000006, -0.00000568, 0.00000512, -0.00000943, 0.00000410, 0.00000148, -0.00000003, 0.00000129, -0.00000063, 0.00000024, -0.00000602, 0.00000258, 0.00000312, -0.00000250, 0.00000259, -0.00000159, 0.00000134, 0.00000069, -0.00000339, 0.00000192, -0.00000102, 0.00000045, },
		{ -0.00000185, -0.00000069, -0.00000415, 0.00000119, 0.00000525, 0.00000433, 0.00000109, -0.00000297, -0.00000188, -0.00000238, -0.00000001, 0.00000030, -0.00000020, 0.00000115, -0.00000282, 0.00000045, 0.00000065, 0.00000111, 0.00000108, -0.00000276, 0.00000265, 0.00000394, -0.00000740, -0.00000046, -0.00000065, 0.00000292, -0.00000096, 0.00000173, 0.00000377, -0.00000792, 0.00000413, -0.00000426, -0.00000234, 0.00000730, -0.00000243, 0.00000380, -0.00001742, -0.00000069, 0.00001398, -0.00000388, 0.00000522, -0.00001496, 0.00001527, -0.00000343, 0.00000098, -0.00001032, 0.00001163, -0.00001323, 0.00000916, -0.00001701, 0.00003376, -0.00001481, 0.00000227, 0.00000440, -0.00004625, 0.00007191, -0.00008662, 0.00007607, -0.00008131, 0.00007753, -0.00007601, 0.00004525, -0.00005943, 0.00002967, -0.00001249, -0.00011837, 0.00016978, -0.00077780, 0.00104573, 0.00035808, -0.00066178, 0.00029208, -0.00025048, 0.00012446, -0.00007780, 0.00001954, -0.00002472, -0.00000605, 0.00001124, -0.00002691, 0.00002571, -0.00004566, 0.00004757, -0.00004142, 0.00000225, -0.00000050, 0.00001140, 0.00000012, 0.00000633, -0.00000648, 0.00000168, 0.00000133, -0.00000358, -0.00000630, 0.00000755, 0.00000380, -0.00000415, 0.00000072, -0.00000184, 0.00001196, -0.00000707, -0.00001323, 0.00000678, -0.00000115, 0.00000466, -0.00000214, -0.00000389, 0.00000411, -0.00000953, 0.00000564, 0.00000010, 0.00000083, 0.00000073, -0.00000020, -0.00000048, -0.00000611, 0.00000385, 0.00000210, -0.00000236, 0.00000272, -0.00000172, 0.00000189, -0.00000016, -0.00000296, 0.00000205, -0.00000132, 0.00000086, -0.00000069, },
		{ -0.00000118, -0.00000396, 0.00000215, 0.00000510, 0.00000432, 0.00000041, -0.00000321, -0.00000156, -0.00000244, 0.00000042, -0.00000001, -0.00000001, 0.00000098, -0.00000307, 0.00000103, 0.00000035, 0.00000119, 0.00000094, -0.00000296, 0.00000382, 0.00000250, -0.00000747, 0.00000058, -0.00000110, 0.00000336, -0.00000177, 0.00000312, 0.00000215, -0.00000766, 0.00000500, -0.00000600, -0.00000004, 0.00000574, -0.00000156, 0.00000313, -0.00001896, 0.00000331, 0.00001143, -0.00000311, 0.00000333, -0.00001282, 0.00001535, -0.00000398, -0.00000006, -0.00000888, 0.00001099, -0.00001316, 0.00000925, -0.00001608, 0.00003528, -0.00002123, 0.00001382, -0.00001283, -0.00002751, 0.00005679, -0.00007805, 0.00007303, -0.00008350, 0.00008496, -0.00009122, 0.00006648, -0.00008878, 0.00006961, -0.00006630, -0.00005433, 0.00008889, -0.00073800, 0.00118484, 0.00015402, -0.00058074, 0.00028537, -0.00026337, 0.00014958, -0.00010451, 0.00004457, -0.00004696, 0.00001437, -0.00000507, -0.00001365, 0.00001558, -0.00003973, 0.00004846, -0.00004866, 0.00001136, -0.00000617, 0.00001438, -0.00000049, 0.00000577, -0.00000732, 0.00000287, 0.00000034, -0.00000359, -0.00000621, 0.00000924, 0.00000191, -0.00000412, 0.00000071, -0.00000103, 0.00001285, -0.00001019, -0.00001114, 0.00000672, -0.00000163, 0.00000571, -0.00000422, -0.00000185, 0.00000259, -0.00000899, 0.00000672, -0.00000114, 0.00000171, 0.00000010, 0.00000033, -0.00000146, -0.00000579, 0.00000490, 0.00000098, -0.00000202, 0.00000270, -0.00000171, 0.00000230, -0.00000099, -0.00000238, 0.00000200, -0.00000147, 0.00000118, -0.00000117, -0.00000162, },
		{ -0.00000352, 0.00000294, 0.00000495, 0.00000429, -0.00000039, -0.00000325, -0.00000131, -0.00000246, 0.00000081, -0.00000037, 0.00000029, 0.00000066, -0.00000317, 0.00000158, -0.00000004, 0.00000139, 0.00000065, -0.00000291, 0.00000484, 0.00000085, -0.00000707, 0.00000128, -0.00000141, 0.00000366, -0.00000246, 0.00000452, 0.00000017, -0.00000676, 0.00000524, -0.00000728, 0.00000215, 0.00000392, -0.00000019, 0.00000168, -0.00001954, 0.00000679, 0.00000878, -0.00000225, 0.00000148, -0.00001042, 0.00001504, -0.00000446, -0.00000108, -0.00000720, 0.00000989, -0.00001252, 0.00000869, -0.00001395, 0.00003453, -0.00002454, 0.00002218, -0.00002814, -0.00000838, 0.00003894, -0.00006516, 0.00006477, -0.00007981, 0.00008599, -0.00009990, 0.00008155, -0.00011273, 0.00010588, -0.00012006, 0.00001575, -0.00001020, -0.00066187, 0.00128974, -0.00004743, -0.00047942, 0.00025815, -0.00026079, 0.00016367, -0.00012422, 0.00006562, -0.00006745, 0.00003494, -0.00002297, 0.00000221, 0.00000198, -0.00002958, 0.00004506, -0.00005272, 0.00001927, -0.00001228, 0.00001849, -0.00000214, 0.00000567, -0.00000824, 0.00000415, -0.00000083, -0.00000347, -0.00000600, 0.00001082, -0.00000029, -0.00000369, 0.00000049, 0.00000019, 0.00001316, -0.00001318, -0.00000859, 0.00000602, -0.00000151, 0.00000623, -0.00000599, 0.00000024, 0.00000069, -0.00000785, 0.00000726, -0.00000211, 0.00000253, -0.00000052, 0.00000088, -0.00000265, -0.00000506, 0.00000565, -0.00000013, -0.00000153, 0.00000255, -0.00000156, 0.00000252, -0.00000174, -0.00000172, 0.00000179, -0.00000148, 0.00000138, -0.00000161, -0.00000140, -0.00000173, },
		{ 0.00000354, 0.00000483, 0.00000418, -0.00000125, -0.00000308, -0.00000117, -0.00000240, 0.00000113, -0.00000075, 0.00000067, 0.00000020, -0.00000310, 0.00000205, -0.00000047, 0.00000167, 0.00000020, -0.00000259, 0.00000559, -0.00000087, -0.00000627, 0.00000162, -0.00000151, 0.00000376, -0.00000290, 0.00000577, -0.00000201, -0.00000532, 0.00000480, -0.00000795, 0.00000401, 0.00000209, 0.00000148, -0.00000043, -0.00001915, 0.00000958, 0.00000621, -0.00000139, -0.00000022, -0.00000786, 0.00001439, -0.00000491, -0.00000197, -0.00000544, 0.00000851, -0.00001154, 0.00000777, -0.00001102, 0.00003188, -0.00002482, 0.00002685, -0.00004039, 0.00000963, 0.00001991, -0.00004924, 0.00005222, -0.00007079, 0.00008069, -0.00010150, 0.00008925, -0.00012930, 0.00013554, -0.00016988, 0.00008714, -0.00012288, -0.00055055, 0.00135498, -0.00023727, -0.00036451, 0.00021328, -0.00024315, 0.00016584, -0.00013536, 0.00008087, -0.00008437, 0.00005392, -0.00004094, 0.00001945, -0.00001419, -0.00001565, 0.00003721, -0.00005288, 0.00002500, -0.00001799, 0.00002320, -0.00000467, 0.00000603, -0.00000927, 0.00000554, -0.00000219, -0.00000317, -0.00000570, 0.00001223, -0.00000270, -0.00000286, 0.00000004, 0.00000182, 0.00001277, -0.00001582, -0.00000580, 0.00000481, -0.00000079, 0.00000613, -0.00000727, 0.00000217, -0.00000139, -0.00000620, 0.00000724, -0.00000275, 0.00000318, -0.00000104, 0.00000135, -0.00000393, -0.00000394, 0.00000603, -0.00000115, -0.00000095, 0.00000230, -0.00000128, 0.00000254, -0.00000237, -0.00000104, 0.00000147, -0.00000134, 0.00000143, -0.00000197, -0.00000123, -0.00000227, -0.00000289, },
		{ 0.00000479, 0.00000395, -0.00000211, -0.00000274, -0.00000114, -0.00000224, 0.00000135, -0.00000108, 0.00000108, -0.00000038, -0.00000283, 0.00000238, -0.00000090, 0.00000199, -0.00000035, -0.00000199, 0.00000600, -0.00000251, -0.00000518, 0.00000162, -0.00000139, 0.00000360, -0.00000301, 0.00000672, -0.00000420, -0.00000347, 0.00000375, -0.00000795, 0.00000537, 0.00000045, 0.00000322, -0.00000300, -0.00001787, 0.00001158, 0.00000386, -0.00000060, -0.00000169, -0.00000526, 0.00001344, -0.00000533, -0.00000268, -0.00000375, 0.00000702, -0.00001040, 0.00000679, -0.00000777, 0.00002789, -0.00002243, 0.00002771, -0.00004877, 0.00002518, 0.00000124, -0.00003174, 0.00003663, -0.00005739, 0.00006965, -0.00009611, 0.00008900, -0.00013708, 0.00015613, -0.00021200, 0.00015469, -0.00024325, -0.00040728, 0.00137712, -0.00040728, -0.00024325, 0.00015469, -0.00021200, 0.00015613, -0.00013708, 0.00008900, -0.00009611, 0.00006965, -0.00005739, 0.00003663, -0.00003174, 0.00000124, 0.00002518, -0.00004877, 0.00002771, -0.00002243, 0.00002789, -0.00000777, 0.00000679, -0.00001040, 0.00000702, -0.00000375, -0.00000268, -0.00000533, 0.00001344, -0.00000526, -0.00000169, -0.00000060, 0.00000386, 0.00001158, -0.00001787, -0.00000300, 0.00000322, 0.00000045, 0.00000537, -0.00000795, 0.00000375, -0.00000347, -0.00000420, 0.00000672, -0.00000301, 0.00000360, -0.00000139, 0.00000162, -0.00000518, -0.00000251, 0.00000600, -0.00000199, -0.00000035, 0.00000199, -0.00000090, 0.00000238, -0.00000283, -0.00000038, 0.00000108, -0.00000108, 0.00000135, -0.00000224, -0.00000114, -0.00000274, -0.00000211, 0.00000395, },
		{ 0.00000354, -0.00000289, -0.00000227, -0.00000123, -0.00000197, 0.00000143, -0.00000134, 0.00000147, -0.00000104, -0.00000237, 0.00000254, -0.00000128, 0.00000230, -0.00000095, -0.00000115, 0.00000603, -0.00000394, -0.00000393, 0.00000135, -0.00000104, 0.00000318, -0.00000275, 0.00000724, -0.00000620, -0.00000139, 0.00000217, -0.00000727, 0.00000613, -0.00000079, 0.00000481, -0.00000580, -0.00001582, 0.00001277, 0.00000182, 0.00000004, -0.00000286, -0.00000270, 0.00001223, -0.00000570, -0.00000317, -0.00000219, 0.00000554, -0.00000927, 0.00000603, -0.00000467, 0.00002320, -0.00001799, 0.00002500, -0.00005288, 0.00003721, -0.00001565, -0.00001419, 0.00001945, -0.00004094, 0.00005392, -0.00008437, 0.00008087, -0.00013536, 0.00016584, -0.00024315, 0.00021328, -0.00036451, -0.00023727, 0.00135498, -0.00055055, -0.00012288, 0.00008714, -0.00016988, 0.00013554, -0.00012930, 0.00008925, -0.00010150, 0.00008069, -0.00007079, 0.00005222, -0.00004924, 0.00001991, 0.00000963, -0.00004039, 0.00002685, -0.00002482, 0.00003188, -0.00001102, 0.00000777, -0.00001154, 0.00000851, -0.00000544, -0.00000197, -0.00000491, 0.00001439, -0.00000786, -0.00000022, -0.00000139, 0.00000621, 0.00000958, -0.00001915, -0.00000043, 0.00000148, 0.00000209, 0.00000401, -0.00000795, 0.00000480, -0.00000532, -0.00000201, 0.00000577, -0.00000290, 0.00000376, -0.00000151, 0.00000162, -0.00000627, -0.00000087, 0.00000559, -0.00000259, 0.00000020, 0.00000167, -0.00000047, 0.00000205, -0.00000310, 0.00000020, 0.00000067, -0.00000075, 0.00000113, -0.00000240, -0.00000117, -0.00000308, -0.00000125, 0.00000418, 0.00000483, },
		{ -0.00000352, -0.00000173, -0.00000140, -0.00000161, 0.00000138, -0.00000148, 0.00000179, -0.00000172, -0.00000174, 0.00000252, -0.00000156, 0.00000255, -0.00000153, -0.00000013, 0.00000565, -0.00000506, -0.00000265, 0.00000088, -0.00000052, 0.00000253, -0.00000211, 0.00000726, -0.00000785, 0.00000069, 0.00000024, -0.00000599, 0.00000623, -0.00000151, 0.00000602, -0.00000859, -0.00001318, 0.00001316, 0.00000019, 0.00000049, -0.00000369, -0.00000029, 0.00001082, -0.00000600, -0.00000347, -0.00000083, 0.00000415, -0.00000824, 0.00000567, -0.00000214, 0.00001849, -0.00001228, 0.00001927, -0.00005272, 0.00004506, -0.00002958, 0.00000198, 0.00000221, -0.00002297, 0.00003494, -0.00006745, 0.00006562, -0.00012422, 0.00016367, -0.00026079, 0.00025815, -0.00047942, -0.00004743, 0.00128974, -0.00066187, -0.00001020, 0.00001575, -0.00012006, 0.00010588, -0.00011273, 0.00008155, -0.00009990, 0.00008599, -0.00007981, 0.00006477, -0.00006516, 0.00003894, -0.00000838, -0.00002814, 0.00002218, -0.00002454, 0.00003453, -0.00001395, 0.00000869, -0.00001252, 0.00000989, -0.00000720, -0.00000108, -0.00000446, 0.00001504, -0.00001042, 0.00000148, -0.00000225, 0.00000878, 0.00000679, -0.00001954, 0.00000168, -0.00000019, 0.00000392, 0.00000215, -0.00000728, 0.00000524, -0.00000676, 0.00000017, 0.00000452, -0.00000246, 0.00000366, -0.00000141, 0.00000128, -0.00000707, 0.00000085, 0.00000484, -0.00000291, 0.00000065, 0.00000139, -0.00000004, 0.00000158, -0.00000317, 0.00000066, 0.00000029, -0.00000037, 0.00000081, -0.00000246, -0.00000131, -0.00000325, -0.00000039, 0.00000429, 0.00000495, 0.00000294, },
		{ -0.00000118, -0.00000162, -0.00000117, 0.00000118, -0.00000147, 0.00000200, -0.00000238, -0.00000099, 0.00000230, -0.00000171, 0.00000270, -0.00000202, 0.00000098, 0.00000490, -0.00000579, -0.00000146, 0.00000033, 0.00000010, 0.00000171, -0.00000114, 0.00000672, -0.00000899, 0.00000259, -0.00000185, -0.00000422, 0.00000571, -0.00000163, 0.00000672, -0.00001114, -0.00001019, 0.00001285, -0.00000103, 0.00000071, -0.00000412, 0.00000191, 0.00000924, -0.00000621, -0.00000359, 0.00000034, 0.00000287, -0.00000732, 0.00000577, -0.00000049, 0.00001438, -0.00000617, 0.00001136, -0.00004866, 0.00004846, -0.00003973, 0.00001558, -0.00001365, -0.00000507, 0.00001437, -0.00004696, 0.00004457, -0.00010451, 0.00014958, -0.00026337, 0.00028537, -0.00058074, 0.00015402, 0.00118484, -0.00073800, 0.00008889, -0.00005433, -0.00006630, 0.00006961, -0.00008878, 0.00006648, -0.00009122, 0.00008496, -0.00008350, 0.00007303, -0.00007805, 0.00005679, -0.00002751, -0.00001283, 0.00001382, -0.00002123, 0.00003528, -0.00001608, 0.00000925, -0.00001316, 0.00001099, -0.00000888, -0.00000006, -0.00000398, 0.00001535, -0.00001282, 0.00000333, -0.00000311, 0.00001143, 0.00000331, -0.00001896, 0.00000313, -0.00000156, 0.00000574, -0.00000004, -0.00000600, 0.00000500, -0.00000766, 0.00000215, 0.00000312, -0.00000177, 0.00000336, -0.00000110, 0.00000058, -0.00000747, 0.00000250, 0.00000382, -0.00000296, 0.00000094, 0.00000119, 0.00000035, 0.00000103, -0.00000307, 0.00000098, -0.00000001, -0.00000001, 0.00000042, -0.00000244, -0.00000156, -0.00000321, 0.00000041, 0.00000432, 0.00000510, 0.00000215, -0.00000396, },
		{ -0.00000185, -0.00000069, 0.00000086, -0.00000132, 0.00000205, -0.00000296, -0.00000016, 0.00000189, -0.00000172, 0.00000272, -0.00000236, 0.00000210, 0.00000385, -0.00000611, -0.00000048, -0.00000020, 0.00000073, 0.00000083, 0.00000010, 0.00000564, -0.00000953, 0.00000411, -0.00000389, -0.00000214, 0.00000466, -0.00000115, 0.00000678, -0.00001323, -0.00000707, 0.00001196, -0.00000184, 0.00000072, -0.00000415, 0.00000380, 0.00000755, -0.00000630, -0.00000358, 0.00000133, 0.00000168, -0.00000648, 0.00000633, 0.00000012, 0.00001140, -0.00000050, 0.00000225, -0.00004142, 0.00004757, -0.00004566, 0.00002571, -0.00002691, 0.00001124, -0.00000605, -0.00002472, 0.00001954, -0.00007780, 0.00012446, -0.00025048, 0.00029208, -0.00066178, 0.00035808, 0.00104573, -0.00077780, 0.00016978, -0.00011837, -0.00001249, 0.00002967, -0.00005943, 0.00004525, -0.00007601, 0.00007753, -0.00008131, 0.00007607, -0.00008662, 0.00007191, -0.00004625, 0.00000440, 0.00000227, -0.00001481, 0.00003376, -0.00001701, 0.00000916, -0.00001323, 0.00001163, -0.00001032, 0.00000098, -0.00000343, 0.00001527, -0.00001496, 0.00000522, -0.00000388, 0.00001398, -0.00000069, -0.00001742, 0.00000380, -0.00000243, 0.00000730, -0.00000234, -0.00000426, 0.00000413, -0.00000792, 0.00000377, 0.00000173, -0.00000096, 0.00000292, -0.00000065, -0.00000046, -0.00000740, 0.00000394, 0.00000265, -0.00000276, 0.00000108, 0.00000111, 0.00000065, 0.00000045, -0.00000282, 0.00000115, -0.00000020, 0.00000030, -0.00000001, -0.00000238, -0.00000188, -0.00000297, 0.00000109, 0.00000433, 0.00000525, 0.00000119, -0.00000415, -0.00000069, },
		{ -0.00000021, 0.00000045, -0.00000102, 0.00000192, -0.00000339, 0.00000069, 0.00000134, -0.00000159, 0.00000259, -0.00000250, 0.00000312, 0.00000258, -0.00000602, 0.00000024, -0.00000063, 0.00000129, -0.00000003, 0.00000148, 0.00000410, -0.00000943, 0.00000512, -0.00000568, 0.00000006, 0.00000321, -0.00000013, 0.00000617, -0.00001471, -0.00000405, 0.00001064, -0.00000230, 0.00000054, -0.00000379, 0.00000534, 0.00000580, -0.00000626, -0.00000349, 0.00000218, 0.00000054, -0.00000561, 0.00000722, -0.00000033, 0.00000987, 0.00000396, -0.00000702, -0.00003196, 0.00004291, -0.00004734, 0.00003186, -0.00003665, 0.00002466, -0.00002465, -0.00000269, -0.00000734, -0.00004622, 0.00009011, -0.00022291, 0.00027684, -0.00071685, 0.00055549, 0.00087955, -0.00078224, 0.00022933, -0.00017235, 0.00003764, -0.00001085, -0.00002707, 0.00001956, -0.00005533, 0.00006420, -0.00007321, 0.00007338, -0.00008990, 0.00008293, -0.00006301, 0.00002218, -0.00001164, -0.00000553, 0.00002980, -0.00001641, 0.00000814, -0.00001252, 0.00001160, -0.00001133, 0.00000192, -0.00000277, 0.00001475, -0.00001674, 0.00000708, -0.00000446, 0.00001626, -0.00000501, -0.00001502, 0.00000361, -0.00000262, 0.00000839, -0.00000450, -0.00000226, 0.00000272, -0.00000751, 0.00000491, 0.00000049, -0.00000013, 0.00000244, -0.00000014, -0.00000177, -0.00000682, 0.00000506, 0.00000144, -0.00000237, 0.00000105, 0.00000117, 0.00000081, -0.00000010, -0.00000246, 0.00000118, -0.00000028, 0.00000052, -0.00000042, -0.00000230, -0.00000223, -0.00000254, 0.00000162, 0.00000438, 0.00000534, 0.00000011, -0.00000409, -0.00000029, -0.00000206, },
		{ 0.00000000, -0.00000059, 0.00000161, -0.00000363, 0.00000149, 0.00000068, -0.00000134, 0.00000232, -0.00000239, 0.00000395, 0.00000122, -0.00000558, 0.00000065, -0.00000089, 0.00000169, -0.00000075, 0.00000286, 0.00000221, -0.00000872, 0.00000554, -0.00000705, 0.00000216, 0.00000155, 0.00000131, 0.00000492, -0.00001547, -0.00000133, 0.00000907, -0.00000246, 0.00000021, -0.00000305, 0.00000648, 0.00000407, -0.00000609, -0.00000336, 0.00000292, -0.00000061, -0.00000459, 0.00000822, -0.00000168, 0.00000994, 0.00000664, -0.00001547, -0.00002137, 0.00003531, -0.00004515, 0.00003394, -0.00004236, 0.00003424, -0.00004003, 0.00001728, -0.00003386, -0.00001229, 0.00004904, -0.00018261, 0.00023972, -0.00074167, 0.00073733, 0.00069460, -0.00075418, 0.00026604, -0.00021327, 0.00008085, -0.00004893, 0.00000575, -0.00000853, -0.00003072, 0.00004596, -0.00005968, 0.00006496, -0.00008736, 0.00008876, -0.00007633, 0.00003901, -0.00002680, 0.00000607, 0.00002348, -0.00001416, 0.00000601, -0.00001086, 0.00001072, -0.00001172, 0.00000260, -0.00000194, 0.00001374, -0.00001805, 0.00000880, -0.00000477, 0.00001808, -0.00000939, -0.00001192, 0.00000258, -0.00000205, 0.00000883, -0.00000630, -0.00000023, 0.00000093, -0.00000650, 0.00000550, -0.00000047, 0.00000061, 0.00000202, 0.00000030, -0.00000324, -0.00000575, 0.00000576, 0.00000032, -0.00000186, 0.00000090, 0.00000135, 0.00000081, -0.00000057, -0.00000205, 0.00000110, -0.00000023, 0.00000062, -0.00000079, -0.00000225, -0.00000257, -0.00000194, 0.00000197, 0.00000452, 0.00000529, -0.00000103, -0.00000379, -0.00000002, -0.00000219, 0.00000023, },
		{ -0.00000009, 0.00000112, -0.00000365, 0.00000218, -0.00000003, -0.00000101, 0.00000193, -0.00000205, 0.00000451, -0.00000013, -0.00000487, 0.00000076, -0.00000092, 0.00000187, -0.00000124, 0.00000411, 0.00000012, -0.00000749, 0.00000532, -0.00000788, 0.00000399, -0.00000014, 0.00000301, 0.00000313, -0.00001549, 0.00000093, 0.00000742, -0.00000243, -0.00000020, -0.00000200, 0.00000718, 0.00000242, -0.00000581, -0.00000321, 0.00000360, -0.00000180, -0.00000329, 0.00000907, -0.00000365, 0.00001151, 0.00000719, -0.00002225, -0.00001078, 0.00002580, -0.00003976, 0.00003224, -0.00004393, 0.00003941, -0.00005111, 0.00003361, -0.00005788, 0.00002130, 0.00000431, -0.00013254, 0.00018236, -0.00073370, 0.00089559, 0.00049983, -0.00069814, 0.00028004, -0.00023926, 0.00011456, -0.00008187, 0.00003652, -0.00003680, -0.00000402, 0.00002420, -0.00004166, 0.00005126, -0.00007893, 0.00008873, -0.00008495, 0.00005340, -0.00004193, 0.00001917, 0.00001516, -0.00001027, 0.00000271, -0.00000816, 0.00000887, -0.00001131, 0.00000289, -0.00000086, 0.00001220, -0.00001879, 0.00001029, -0.00000474, 0.00001928, -0.00001356, -0.00000835, 0.00000083, -0.00000070, 0.00000848, -0.00000752, 0.00000158, -0.00000101, -0.00000500, 0.00000554, -0.00000109, 0.00000117, 0.00000174, 0.00000056, -0.00000474, -0.00000426, 0.00000600, -0.00000063, -0.00000133, 0.00000066, 0.00000164, 0.00000064, -0.00000091, -0.00000164, 0.00000096, -0.00000007, 0.00000058, -0.00000108, -0.00000225, -0.00000284, -0.00000124, 0.00000214, 0.00000477, 0.00000505, -0.00000214, -0.00000327, 0.00000008, -0.00000222, 0.00000059, -0.00000045, },
		{ 0.00000049, -0.00000343, 0.00000271, -0.00000074, -0.00000063, 0.00000148, -0.00000148, 0.00000476, -0.00000134, -0.00000399, 0.00000061, -0.00000074, 0.00000181, -0.00000142, 0.00000509, -0.00000200, -0.00000585, 0.00000452, -0.00000810, 0.00000540, -0.00000167, 0.00000479, 0.00000093, -0.00001480, 0.00000262, 0.00000583, -0.00000231, -0.00000060, -0.00000071, 0.00000744, 0.00000091, -0.00000543, -0.00000306, 0.00000421, -0.00000306, -0.00000162, 0.00000951, -0.00000587, 0.00001432, 0.00000553, -0.00002673, -0.00000122, 0.00001549, -0.00003209, 0.00002738, -0.00004163, 0.00004005, -0.00005728, 0.00004509, -0.00007752, 0.00005194, -0.00004076, -0.00007648, 0.00010789, -0.00069233, 0.00102361, 0.00030427, -0.00061988, 0.00027296, -0.00024970, 0.00013710, -0.00010752, 0.00006296, -0.00006306, 0.00002278, 0.00000063, -0.00002050, 0.00003324, -0.00006504, 0.00008260, -0.00008800, 0.00006403, -0.00005569, 0.00003278, 0.00000538, -0.00000500, -0.00000169, -0.00000444, 0.00000599, -0.00000999, 0.00000263, 0.00000053, 0.00001007, -0.00001886, 0.00001147, -0.00000433, 0.00001972, -0.00001724, -0.00000458, -0.00000149, 0.00000138, 0.00000730, -0.00000800, 0.00000295, -0.00000288, -0.00000317, 0.00000509, -0.00000134, 0.00000149, 0.00000166, 0.00000054, -0.00000611, -0.00000245, 0.00000576, -0.00000130, -0.00000084, 0.00000039, 0.00000201, 0.00000032, -0.00000109, -0.00000128, 0.00000078, 0.00000017, 0.00000042, -0.00000129, -0.00000234, -0.00000299, -0.00000050, 0.00000218, 0.00000513, 0.00000457, -0.00000316, -0.00000259, 0.00000003, -0.00000213, 0.00000083, -0.00000084, 0.00000045, },
		{ -0.00000299, 0.00000303, -0.00000140, -0.00000024, 0.00000101, -0.00000072, 0.00000467, -0.00000232, -0.00000306, 0.00000028, -0.00000037, 0.00000151, -0.00000127, 0.00000568, -0.00000398, -0.00000398, 0.00000322, -0.00000769, 0.00000628, -0.00000289, 0.00000646, -0.00000149, -0.00001349, 0.00000369, 0.00000444, -0.00000217, -0.00000092, 0.00000073, 0.00000728, -0.00000040, -0.00000502, -0.00000290, 0.00000475, -0.00000436, 0.00000049, 0.00000927, -0.00000791, 0.00001794, 0.00000182, -0.00002857, 0.00000648, 0.00000546, -0.00002321, 0.00002023, -0.00003612, 0.00003649, -0.00005839, 0.00005100, -0.00009137, 0.00007733, -0.00008287, -0.00001866, 0.00002069, -0.00061890, 0.00111650, 0.00011647, -0.00052593, 0.00024766, -0.00024515, 0.00014771, -0.00012440, 0.00008325, -0.00008535, 0.00004772, -0.00002293, 0.00000221, 0.00001222, -0.00004660, 0.00007069, -0.00008504, 0.00006984, -0.00006680, 0.00004578, -0.00000510, 0.00000127, -0.00000696, 0.00000019, 0.00000211, -0.00000770, 0.00000173, 0.00000230, 0.00000734, -0.00001820, 0.00001226, -0.00000351, 0.00001930, -0.00002019, -0.00000092, -0.00000412, 0.00000402, 0.00000532, -0.00000765, 0.00000369, -0.00000444, -0.00000123, 0.00000426, -0.00000125, 0.00000157, 0.00000183, 0.00000014, -0.00000719, -0.00000048, 0.00000509, -0.00000167, -0.00000048, 0.00000015, 0.00000240, -0.00000013, -0.00000111, -0.00000101, 0.00000063, 0.00000045, 0.00000015, -0.00000140, -0.00000252, -0.00000297, 0.00000022, 0.00000213, 0.00000559, 0.00000384, -0.00000398, -0.00000181, -0.00000016, -0.00000192, 0.00000092, -0.00000114, 0.00000097, -0.00000025, },
		{ 0.00000313, -0.00000197, 0.00000010, 0.00000058, 0.00000014, 0.00000426, -0.00000301, -0.00000220, -0.00000016, 0.00000015, 0.00000101, -0.00000079, 0.00000582, -0.00000565, -0.00000205, 0.00000156, -0.00000671, 0.00000660, -0.00000366, 0.00000785, -0.00000395, -0.00001171, 0.00000416, 0.00000331, -0.00000212, -0.00000107, 0.00000223, 0.00000673, -0.00000147, -0.00000460, -0.00000269, 0.00000517, -0.00000565, 0.00000303, 0.00000819, -0.00000934, 0.00002180, -0.00000353, -0.00002772, 0.00001170, -0.00000331, -0.00001418, 0.00001181, -0.00002829, 0.00002940, -0.00005476, 0.00005117, -0.00009857, 0.00009567, -0.00011902, 0.00003652, -0.00007388, -0.00051666, 0.00117137, -0.00005601, -0.00042312, 0.00020799, -0.00022723, 0.00014662, -0.00013182, 0.00009616, -0.00010207, 0.00006904, -0.00004470, 0.00002474, -0.00001024, -0.00002491, 0.00005379, -0.00007613, 0.00007015, -0.00007416, 0.00005707, -0.00001543, 0.00000798, -0.00001274, 0.00000547, -0.00000262, -0.00000446, 0.00000014, 0.00000446, 0.00000400, -0.00001675, 0.00001260, -0.00000229, 0.00001797, -0.00002220, 0.00000235, -0.00000680, 0.00000702, 0.00000266, -0.00000646, 0.00000366, -0.00000547, 0.00000061, 0.00000322, -0.00000089, 0.00000143, 0.00000223, -0.00000068, -0.00000784, 0.00000149, 0.00000407, -0.00000170, -0.00000029, -0.00000000, 0.00000276, -0.00000067, -0.00000097, -0.00000086, 0.00000054, 0.00000072, -0.00000020, -0.00000143, -0.00000279, -0.00000276, 0.00000084, 0.00000204, 0.00000610, 0.00000284, -0.00000456, -0.00000100, -0.00000046, -0.00000160, 0.00000086, -0.00000130, 0.00000141, -0.00000103, -0.00000235, },
		{ -0.00000242, 0.00000037, 0.00000024, 0.00000104, 0.00000358, -0.00000336, -0.00000148, -0.00000061, 0.00000073, 0.00000037, -0.00000001, 0.00000547, -0.00000690, -0.00000025, -0.00000030, -0.00000527, 0.00000635, -0.00000393, 0.00000882, -0.00000625, -0.00000962, 0.00000406, 0.00000251, -0.00000219, -0.00000099, 0.00000369, 0.00000585, -0.00000227, -0.00000423, -0.00000241, 0.00000544, -0.00000684, 0.00000591, 0.00000617, -0.00000979, 0.00002534, -0.00000994, -0.00002442, 0.00001416, -0.00001010, -0.00000595, 0.00000317, -0.00001920, 0.00001976, -0.00004715, 0.00004595, -0.00009889, 0.00010580, -0.00014676, 0.00008488, -0.00016988, -0.00039049, 0.00118746, -0.00020699, -0.00031812, 0.00015837, -0.00019841, 0.00013490, -0.00012985, 0.00010107, -0.00011214, 0.00008531, -0.00006304, 0.00004539, -0.00003240, -0.00000157, 0.00003310, -0.00006182, 0.00006474, -0.00007698, 0.00006565, -0.00002473, 0.00001448, -0.00001858, 0.00001106, -0.00000794, -0.00000038, -0.00000214, 0.00000701, 0.00000009, -0.00001449, 0.00001246, -0.00000069, 0.00001575, -0.00002311, 0.00000493, -0.00000921, 0.00001011, -0.00000048, -0.00000451, 0.00000281, -0.00000581, 0.00000212, 0.00000213, -0.00000038, 0.00000115, 0.00000280, -0.00000192, -0.00000796, 0.00000328, 0.00000281, -0.00000142, -0.00000033, -0.00000001, 0.00000304, -0.00000125, -0.00000069, -0.00000084, 0.00000055, 0.00000094, -0.00000060, -0.00000141, -0.00000312, -0.00000235, 0.00000130, 0.00000199, 0.00000661, 0.00000162, -0.00000484, -0.00000023, -0.00000082, -0.00000119, 0.00000066, -0.00000130, 0.00000172, -0.00000179, -0.00000154, 0.00000300, },
		{ 0.00000055, 0.00000006, 0.00000188, 0.00000270, -0.00000338, -0.00000098, -0.00000098, 0.00000130, -0.00000033, 0.00000099, 0.00000466, -0.00000763, 0.00000127, -0.00000217, -0.00000349, 0.00000562, -0.00000368, 0.00000929, -0.00000822, -0.00000742, 0.00000352, 0.00000203, -0.00000241, -0.00000062, 0.00000499, 0.00000471, -0.00000280, -0.00000393, -0.00000204, 0.00000549, -0.00000779, 0.00000899, 0.00000322, -0.00000901, 0.00002799, -0.00001671, -0.00001915, 0.00001387, -0.00001444, 0.00000067, -0.00000472, -0.00000995, 0.00000871, -0.00003661, 0.00003616, -0.00009270, 0.00010731, -0.00016439, 0.00012277, -0.00026125, -0.00024661, 0.00116608, -0.00033196, -0.00021699, 0.00010343, -0.00016176, 0.00011437, -0.00011931, 0.00009806, -0.00011506, 0.00009557, -0.00007669, 0.00006264, -0.00005261, 0.00002170, 0.00001019, -0.00004315, 0.00005386, -0.00007482, 0.00007073, -0.00003217, 0.00002010, -0.00002394, 0.00001651, -0.00001353, 0.00000431, -0.00000500, 0.00000989, -0.00000431, -0.00001144, 0.00001182, 0.00000121, 0.00001270, -0.00002284, 0.00000661, -0.00001108, 0.00001298, -0.00000385, -0.00000196, 0.00000118, -0.00000536, 0.00000312, 0.00000119, 0.00000015, 0.00000084, 0.00000345, -0.00000351, -0.00000747, 0.00000476, 0.00000145, -0.00000089, -0.00000057, 0.00000014, 0.00000319, -0.00000181, -0.00000031, -0.00000094, 0.00000068, 0.00000107, -0.00000101, -0.00000135, -0.00000349, -0.00000175, 0.00000158, 0.00000203, 0.00000703, 0.00000021, -0.00000480, 0.00000042, -0.00000119, -0.00000072, 0.00000034, -0.00000114, 0.00000187, -0.00000247, -0.00000063, 0.00000265, -0.00000273, },
		{ 0.00000005, 0.00000259, 0.00000171, -0.00000311, -0.00000074, -0.00000120, 0.00000177, -0.00000099, 0.00000210, 0.00000344, -0.00000783, 0.00000239, -0.00000389, -0.00000154, 0.00000449, -0.00000295, 0.00000920, -0.00000973, -0.00000529, 0.00000267, 0.00000183, -0.00000279, 0.00000006, 0.00000605, 0.00000341, -0.00000308, -0.00000374, -0.00000154, 0.00000527, -0.00000838, 0.00001204, -0.00000052, -0.00000686, 0.00002927, -0.00002313, -0.00001259, 0.00001112, -0.00001615, 0.00000517, -0.00001106, -0.00000157, -0.00000254, -0.00002441, 0.00002300, -0.00008095, 0.00010053, -0.00017114, 0.00014741, -0.00034219, -0.00009209, 0.00111046, -0.00042824, -0.00012482, 0.00004764, -0.00012061, 0.00008736, -0.00010163, 0.00008783, -0.00011089, 0.00009938, -0.00008478, 0.00007530, -0.00006935, 0.00004316, -0.00001320, -0.00002152, 0.00003824, -0.00006767, 0.00007180, -0.00003709, 0.00002421, -0.00002828, 0.00002134, -0.00001895, 0.00000932, -0.00000830, 0.00001297, -0.00000907, -0.00000767, 0.00001069, 0.00000334, 0.00000895, -0.00002140, 0.00000724, -0.00001216, 0.00001535, -0.00000714, 0.00000094, -0.00000111, -0.00000412, 0.00000350, 0.00000055, 0.00000056, 0.00000061, 0.00000403, -0.00000534, -0.00000639, 0.00000577, 0.00000013, -0.00000019, -0.00000102, 0.00000048, 0.00000317, -0.00000229, 0.00000012, -0.00000114, 0.00000093, 0.00000108, -0.00000137, -0.00000131, -0.00000384, -0.00000099, 0.00000164, 0.00000223, 0.00000729, -0.00000130, -0.00000443, 0.00000091, -0.00000152, -0.00000025, -0.00000008, -0.00000082, 0.00000182, -0.00000301, 0.00000031, 0.00000212, -0.00000290, 0.00000062, },
		{ 0.00000310, 0.00000073, -0.00000262, -0.00000077, -0.00000123, 0.00000210, -0.00000153, 0.00000321, 0.00000191, -0.00000750, 0.00000301, -0.00000531, 0.00000042, 0.00000312, -0.00000183, 0.00000858, -0.00001070, -0.00000340, 0.00000164, 0.00000185, -0.00000330, 0.00000106, 0.00000678, 0.00000202, -0.00000312, -0.00000364, -0.00000092, 0.00000475, -0.00000847, 0.00001483, -0.00000481, -0.00000338, 0.00002886, -0.00002850, -0.00000551, 0.00000645, -0.00001538, 0.00000729, -0.00001531, 0.00000510, -0.00001286, -0.00001188, 0.00000790, -0.00006502, 0.00008651, -0.00016715, 0.00015707, -0.00040760, 0.00006556, 0.00102540, -0.00049497, -0.00004552, -0.00000499, -0.00007833, 0.00005648, -0.00007870, 0.00007161, -0.00010027, 0.00009682, -0.00008693, 0.00008259, -0.00008144, 0.00006125, -0.00003530, 0.00000142, 0.00001900, -0.00005592, 0.00006869, -0.00003901, 0.00002629, -0.00003109, 0.00002506, -0.00002376, 0.00001428, -0.00001179, 0.00001608, -0.00001402, -0.00000326, 0.00000910, 0.00000557, 0.00000468, -0.00001886, 0.00000673, -0.00001225, 0.00001693, -0.00001003, 0.00000392, -0.00000385, -0.00000214, 0.00000318, 0.00000033, 0.00000071, 0.00000060, 0.00000441, -0.00000727, -0.00000475, 0.00000626, -0.00000101, 0.00000058, -0.00000161, 0.00000098, 0.00000296, -0.00000265, 0.00000056, -0.00000142, 0.00000130, 0.00000094, -0.00000167, -0.00000130, -0.00000412, -0.00000013, 0.00000151, 0.00000261, 0.00000732, -0.00000282, -0.00000376, 0.00000120, -0.00000177, 0.00000018, -0.00000054, -0.00000037, 0.00000158, -0.00000335, 0.00000124, 0.00000143, -0.00000295, 0.00000062, 0.00000023, },
		{ -0.00000017, -0.00000198, -0.00000102, -0.00000105, 0.00000223, -0.00000187, 0.00000420, 0.00000021, -0.00000671, 0.00000312, -0.00000630, 0.00000222, 0.00000164, -0.00000043, 0.00000749, -0.00001110, -0.00000187, 0.00000059, 0.00000200, -0.00000387, 0.00000235, 0.00000714, 0.00000066, -0.00000299, -0.00000364, -0.00000018, 0.00000393, -0.00000797, 0.00001708, -0.00000932, 0.00000125, 0.00002662, -0.00003228, 0.00000131, 0.00000057, -0.00001250, 0.00000707, -0.00001718, 0.00000947, -0.00002133, -0.00000027, -0.00000762, -0.00004661, 0.00006686, -0.00015348, 0.00015123, -0.00045347, 0.00021894, 0.00091694, -0.00053306, 0.00001836, -0.00005121, -0.00003799, 0.00002440, -0.00005265, 0.00005102, -0.00008432, 0.00008849, -0.00008327, 0.00008418, -0.00008811, 0.00007471, -0.00005443, 0.00002388, -0.00000244, -0.00004039, 0.00006157, -0.00003774, 0.00002600, -0.00003199, 0.00002725, -0.00002749, 0.00001879, -0.00001520, 0.00001902, -0.00001895, 0.00000161, 0.00000712, 0.00000780, 0.00000010, -0.00001539, 0.00000511, -0.00001126, 0.00001752, -0.00001225, 0.00000664, -0.00000678, 0.00000040, 0.00000220, 0.00000061, 0.00000052, 0.00000090, 0.00000442, -0.00000910, -0.00000268, 0.00000618, -0.00000187, 0.00000132, -0.00000229, 0.00000163, 0.00000255, -0.00000284, 0.00000096, -0.00000173, 0.00000176, 0.00000066, -0.00000187, -0.00000137, -0.00000428, 0.00000076, 0.00000120, 0.00000318, 0.00000705, -0.00000426, -0.00000286, 0.00000128, -0.00000191, 0.00000053, -0.00000099, 0.00000017, 0.00000115, -0.00000348, 0.00000207, 0.00000064, -0.00000288, 0.00000054, 0.00000060, 0.00000336, },
		{ -0.00000131, -0.00000144, -0.00000068, 0.00000214, -0.00000197, 0.00000496, -0.00000153, -0.00000556, 0.00000274, -0.00000682, 0.00000374, 0.00000022, 0.00000111, 0.00000603, -0.00001095, -0.00000079, -0.00000035, 0.00000219, -0.00000444, 0.00000387, 0.00000710, -0.00000061, -0.00000274, -0.00000368, 0.00000063, 0.00000283, -0.00000682, 0.00001857, -0.00001369, 0.00000671, 0.00002258, -0.00003408, 0.00000712, -0.00000577, -0.00000809, 0.00000482, -0.00001669, 0.00001123, -0.00002726, 0.00000935, -0.00002212, -0.00002753, 0.00004361, -0.00013200, 0.00013060, -0.00047710, 0.00036120, 0.00079182, -0.00054488, 0.00006564, -0.00008864, -0.00000219, -0.00000639, -0.00002568, 0.00002792, -0.00006447, 0.00007542, -0.00007439, 0.00008024, -0.00008904, 0.00008267, -0.00006919, 0.00004414, -0.00002446, -0.00002217, 0.00005099, -0.00003334, 0.00002318, -0.00003072, 0.00002755, -0.00002975, 0.00002245, -0.00001821, 0.00002152, -0.00002358, 0.00000676, 0.00000485, 0.00000990, -0.00000455, -0.00001122, 0.00000250, -0.00000917, 0.00001698, -0.00001353, 0.00000881, -0.00000959, 0.00000330, 0.00000064, 0.00000138, -0.00000005, 0.00000158, 0.00000395, -0.00001064, -0.00000034, 0.00000556, -0.00000237, 0.00000194, -0.00000300, 0.00000239, 0.00000197, -0.00000285, 0.00000126, -0.00000202, 0.00000228, 0.00000023, -0.00000197, -0.00000152, -0.00000428, 0.00000162, 0.00000078, 0.00000392, 0.00000644, -0.00000551, -0.00000178, 0.00000115, -0.00000191, 0.00000077, -0.00000140, 0.00000076, 0.00000055, -0.00000336, 0.00000275, -0.00000020, -0.00000274, 0.00000044, 0.00000110, 0.00000336, -0.00000088, },
		{ -0.00000196, -0.00000014, 0.00000187, -0.00000179, 0.00000541, -0.00000316, -0.00000420, 0.00000193, -0.00000684, 0.00000489, -0.00000103, 0.00000265, 0.00000433, -0.00001033, -0.00000020, -0.00000108, 0.00000233, -0.00000492, 0.00000555, 0.00000664, -0.00000170, -0.00000243, -0.00000375, 0.00000146, 0.00000153, -0.00000501, 0.00001908, -0.00001751, 0.00001257, 0.00001698, -0.00003371, 0.00001136, -0.00001178, -0.00000284, 0.00000103, -0.00001415, 0.00001040, -0.00003030, 0.00001621, -0.00003435, -0.00000953, 0.00001899, -0.00010518, 0.00009701, -0.00047725, 0.00048649, 0.00065710, -0.00053403, 0.00009641, -0.00011591, 0.00002713, -0.00003377, 0.00000013, 0.00000420, -0.00004238, 0.00005892, -0.00006127, 0.00007140, -0.00008442, 0.00008475, -0.00007856, 0.00006065, -0.00004540, -0.00000262, 0.00003775, -0.00002620, 0.00001792, -0.00002719, 0.00002577, -0.00003020, 0.00002488, -0.00002051, 0.00002333, -0.00002765, 0.00001196, 0.00000238, 0.00001172, -0.00000901, -0.00000664, -0.00000092, -0.00000607, 0.00001527, -0.00001371, 0.00001015, -0.00001197, 0.00000627, -0.00000132, 0.00000258, -0.00000100, 0.00000268, 0.00000289, -0.00001168, 0.00000208, 0.00000448, -0.00000248, 0.00000236, -0.00000365, 0.00000318, 0.00000123, -0.00000267, 0.00000143, -0.00000225, 0.00000282, -0.00000032, -0.00000196, -0.00000177, -0.00000408, 0.00000236, 0.00000031, 0.00000480, 0.00000547, -0.00000648, -0.00000062, 0.00000084, -0.00000178, 0.00000087, -0.00000171, 0.00000133, -0.00000016, -0.00000300, 0.00000324, -0.00000104, -0.00000256, 0.00000036, 0.00000171, 0.00000309, -0.00000135, -0.00000070, },
		{ 0.00000050, 0.00000143, -0.00000136, 0.00000549, -0.00000456, -0.00000276, 0.00000080, -0.00000638, 0.00000561, -0.00000199, 0.00000404, 0.00000255, -0.00000935, -0.00000008, -0.00000152, 0.00000233, -0.00000522, 0.00000729, 0.00000581, -0.00000256, -0.00000213, -0.00000377, 0.00000223, 0.00000013, -0.00000261, 0.00001852, -0.00002045, 0.00001835, 0.00001023, -0.00003124, 0.00001361, -0.00001679, 0.00000252, -0.00000366, -0.00001004, 0.00000725, -0.00003043, 0.00001988, -0.00004342, 0.00000588, -0.00000480, -0.00007585, 0.00005331, -0.00045425, 0.00059022, 0.00051959, -0.00050490, 0.00011188, -0.00013258, 0.00004877, -0.00005612, 0.00002301, -0.00001834, -0.00001973, 0.00004050, -0.00004517, 0.00005864, -0.00007488, 0.00008108, -0.00008200, 0.00007220, -0.00006370, 0.00001684, 0.00002292, -0.00001693, 0.00001056, -0.00002154, 0.00002186, -0.00002865, 0.00002580, -0.00002182, 0.00002420, -0.00003086, 0.00001695, -0.00000015, 0.00001316, -0.00001302, -0.00000195, -0.00000488, -0.00000212, 0.00001244, -0.00001269, 0.00001046, -0.00001362, 0.00000901, -0.00000347, 0.00000408, -0.00000226, 0.00000415, 0.00000121, -0.00001208, 0.00000436, 0.00000307, -0.00000221, 0.00000253, -0.00000418, 0.00000397, 0.00000039, -0.00000231, 0.00000145, -0.00000238, 0.00000333, -0.00000095, -0.00000186, -0.00000209, -0.00000368, 0.00000293, -0.00000013, 0.00000574, 0.00000417, -0.00000711, 0.00000054, 0.00000040, -0.00000152, 0.00000083, -0.00000189, 0.00000184, -0.00000093, -0.00000242, 0.00000351, -0.00000184, -0.00000235, 0.00000034, 0.00000234, 0.00000260, -0.00000155, -0.00000023, -0.00000249, },
		{ 0.00000090, -0.00000071, 0.00000519, -0.00000563, -0.00000138, -0.00000053, -0.00000551, 0.00000591, -0.00000262, 0.00000519, 0.00000082, -0.00000813, -0.00000037, -0.00000166, 0.00000214, -0.00000527, 0.00000899, 0.00000464, -0.00000316, -0.00000188, -0.00000372, 0.00000287, -0.00000124, 0.00000024, 0.00001685, -0.00002221, 0.00002356, 0.00000285, -0.00002693, 0.00001371, -0.00002027, 0.00000730, -0.00000855, -0.00000501, 0.00000233, -0.00002792, 0.00002030, -0.00004879, 0.00001752, -0.00002578, -0.00004697, 0.00000307, -0.00040986, 0.00066932, 0.00038554, -0.00046232, 0.00011415, -0.00013915, 0.00006229, -0.00007241, 0.00004163, -0.00003818, 0.00000189, 0.00002168, -0.00002751, 0.00004323, -0.00006145, 0.00007225, -0.00007946, 0.00007801, -0.00007805, 0.00003477, 0.00000765, -0.00000635, 0.00000161, -0.00001410, 0.00001596, -0.00002504, 0.00002499, -0.00002192, 0.00002392, -0.00003296, 0.00002147, -0.00000260, 0.00001410, -0.00001637, 0.00000253, -0.00000907, 0.00000242, 0.00000865, -0.00001051, 0.00000959, -0.00001430, 0.00001124, -0.00000555, 0.00000570, -0.00000370, 0.00000589, -0.00000107, -0.00001170, 0.00000631, 0.00000148, -0.00000160, 0.00000242, -0.00000455, 0.00000469, -0.00000051, -0.00000179, 0.00000132, -0.00000237, 0.00000376, -0.00000162, -0.00000168, -0.00000247, -0.00000306, 0.00000327, -0.00000045, 0.00000665, 0.00000259, -0.00000734, 0.00000161, -0.00000013, -0.00000116, 0.00000065, -0.00000192, 0.00000224, -0.00000171, -0.00000166, 0.00000354, -0.00000257, -0.00000215, 0.00000040, 0.00000293, 0.00000195, -0.00000147, 0.00000003, -0.00000296, 0.00000117, },
		{ 0.00000007, 0.00000453, -0.00000632, -0.00000019, -0.00000192, -0.00000433, 0.00000581, -0.00000288, 0.00000602, -0.00000074, -0.00000684, -0.00000098, -0.00000149, 0.00000173, -0.00000500, 0.00001054, 0.00000322, -0.00000351, -0.00000171, -0.00000355, 0.00000331, -0.00000244, 0.00000336, 0.00001414, -0.00002262, 0.00002773, -0.00000459, -0.00002121, 0.00001171, -0.00002189, 0.00001096, -0.00001299, 0.00000024, -0.00000370, -0.00002334, 0.00001778, -0.00005039, 0.00002471, -0.00004237, -0.00002131, -0.00004971, -0.00034711, 0.00072234, 0.00026024, -0.00041116, 0.00010591, -0.00013678, 0.00006793, -0.00008223, 0.00005511, -0.00005416, 0.00002113, 0.00000389, -0.00000971, 0.00002656, -0.00004542, 0.00005927, -0.00007141, 0.00007776, -0.00008747, 0.00004993, -0.00000685, 0.00000460, -0.00000822, -0.00000537, 0.00000840, -0.00001950, 0.00002240, -0.00002068, 0.00002234, -0.00003371, 0.00002525, -0.00000482, 0.00001447, -0.00001885, 0.00000649, -0.00001316, 0.00000723, 0.00000415, -0.00000728, 0.00000755, -0.00001385, 0.00001269, -0.00000731, 0.00000724, -0.00000516, 0.00000775, -0.00000384, -0.00001051, 0.00000774, -0.00000012, -0.00000075, 0.00000204, -0.00000473, 0.00000530, -0.00000142, -0.00000115, 0.00000103, -0.00000221, 0.00000407, -0.00000231, -0.00000147, -0.00000286, -0.00000226, 0.00000336, -0.00000059, 0.00000745, 0.00000079, -0.00000717, 0.00000253, -0.00000069, -0.00000073, 0.00000036, -0.00000180, 0.00000249, -0.00000244, -0.00000077, 0.00000334, -0.00000320, -0.00000195, 0.00000059, 0.00000341, 0.00000122, -0.00000116, 0.00000006, -0.00000328, 0.00000179, 0.00000035, },
		{ 0.00000358, -0.00000660, 0.00000072, -0.00000324, -0.00000296, 0.00000538, -0.00000279, 0.00000649, -0.00000202, -0.00000560, -0.00000179, -0.00000105, 0.00000110, -0.00000437, 0.00001183, 0.00000164, -0.00000360, -0.00000164, -0.00000324, 0.00000349, -0.00000331, 0.00000651, 0.00001057, -0.00002161, 0.00003053, -0.00001148, -0.00001465, 0.00000790, -0.00002154, 0.00001311, -0.00001644, 0.00000503, -0.00001008, -0.00001739, 0.00001288, -0.00004852, 0.00002726, -0.00005356, -0.00000123, -0.00010091, -0.00027004, 0.00074938, 0.00014780, -0.00035599, 0.00009012, -0.00012722, 0.00006654, -0.00008575, 0.00006313, -0.00006555, 0.00003700, -0.00001170, 0.00000692, 0.00001003, -0.00002821, 0.00004345, -0.00005876, 0.00007167, -0.00009141, 0.00006129, -0.00001948, 0.00001495, -0.00001814, 0.00000402, -0.00000035, -0.00001230, 0.00001810, -0.00001807, 0.00001941, -0.00003295, 0.00002807, -0.00000668, 0.00001421, -0.00002034, 0.00000965, -0.00001680, 0.00001197, -0.00000075, -0.00000324, 0.00000440, -0.00001220, 0.00001317, -0.00000851, 0.00000849, -0.00000644, 0.00000954, -0.00000695, -0.00000852, 0.00000852, -0.00000156, 0.00000024, 0.00000144, -0.00000470, 0.00000576, -0.00000228, -0.00000043, 0.00000061, -0.00000189, 0.00000423, -0.00000295, -0.00000125, -0.00000322, -0.00000134, 0.00000319, -0.00000050, 0.00000803, -0.00000112, -0.00000660, 0.00000323, -0.00000121, -0.00000029, -0.00000002, -0.00000153, 0.00000257, -0.00000305, 0.00000020, 0.00000292, -0.00000372, -0.00000177, 0.00000091, 0.00000373, 0.00000050, -0.00000068, -0.00000016, -0.00000342, 0.00000232, -0.00000015, 0.00000091, },
		{ -0.00000650, 0.00000129, -0.00000438, -0.00000152, 0.00000472, -0.00000240, 0.00000660, -0.00000297, -0.00000454, -0.00000268, -0.00000043, 0.00000027, -0.00000337, 0.00001279, -0.00000002, -0.00000349, -0.00000166, -0.00000280, 0.00000340, -0.00000373, 0.00000945, 0.00000640, -0.00001927, 0.00003172, -0.00001731, -0.00000789, 0.00000273, -0.00001934, 0.00001356, -0.00001849, 0.00000878, -0.00001609, -0.00001089, 0.00000641, -0.00004385, 0.00002549, -0.00005895, 0.00001152, -0.00014667, -0.00018336, 0.00075196, 0.00005102, -0.00030081, 0.00006979, -0.00011249, 0.00005943, -0.00008362, 0.00006583, -0.00007206, 0.00004885, -0.00002419, 0.00002130, -0.00000510, -0.00001127, 0.00002627, -0.00004279, 0.00006041, -0.00008979, 0.00006819, -0.00002934, 0.00002380, -0.00002734, 0.00001334, -0.00000968, -0.00000387, 0.00001230, -0.00001417, 0.00001514, -0.00003060, 0.00002974, -0.00000807, 0.00001330, -0.00002074, 0.00001180, -0.00001971, 0.00001629, -0.00000570, 0.00000132, 0.00000035, -0.00000937, 0.00001257, -0.00000896, 0.00000925, -0.00000735, 0.00001105, -0.00001017, -0.00000584, 0.00000858, -0.00000270, 0.00000125, 0.00000066, -0.00000448, 0.00000605, -0.00000306, 0.00000034, 0.00000008, -0.00000141, 0.00000421, -0.00000352, -0.00000105, -0.00000350, -0.00000034, 0.00000280, -0.00000015, 0.00000831, -0.00000304, -0.00000568, 0.00000368, -0.00000167, 0.00000014, -0.00000044, -0.00000115, 0.00000247, -0.00000351, 0.00000118, 0.00000231, -0.00000414, -0.00000159, 0.00000134, 0.00000383, -0.00000012, -0.00000009, -0.00000060, -0.00000336, 0.00000270, -0.00000053, 0.00000170, 0.00000241, },
		{ 0.00000150, -0.00000525, -0.00000011, 0.00000392, -0.00000179, 0.00000640, -0.00000357, -0.00000376, -0.00000352, 0.00000029, -0.00000068, -0.00000203, 0.00001335, -0.00000165, -0.00000321, -0.00000176, -0.00000225, 0.00000302, -0.00000362, 0.00001192, 0.00000194, -0.00001578, 0.00003124, -0.00002166, -0.00000156, -0.00000326, -0.00001558, 0.00001233, -0.00001894, 0.00001109, -0.00002108, -0.00000464, -0.00000074, -0.00003729, 0.00002014, -0.00005877, 0.00001598, -0.00018363, -0.00009208, 0.00073284, -0.00002860, -0.00024884, 0.00004768, -0.00009471, 0.00004814, -0.00007688, 0.00006378, -0.00007385, 0.00005648, -0.00003307, 0.00003268, -0.00001780, 0.00000410, 0.00000925, -0.00002498, 0.00004507, -0.00008297, 0.00007036, -0.00003576, 0.00003037, -0.00003504, 0.00002185, -0.00001890, 0.00000525, 0.00000536, -0.00000919, 0.00000970, -0.00002667, 0.00003012, -0.00000889, 0.00001178, -0.00002005, 0.00001281, -0.00002162, 0.00001985, -0.00001030, 0.00000603, -0.00000432, -0.00000552, 0.00001087, -0.00000855, 0.00000936, -0.00000772, 0.00001204, -0.00001324, -0.00000264, 0.00000792, -0.00000341, 0.00000216, -0.00000021, -0.00000409, 0.00000618, -0.00000372, 0.00000110, -0.00000052, -0.00000080, 0.00000401, -0.00000399, -0.00000091, -0.00000366, 0.00000065, 0.00000223, 0.00000046, 0.00000822, -0.00000483, -0.00000448, 0.00000387, -0.00000201, 0.00000051, -0.00000087, -0.00000067, 0.00000220, -0.00000377, 0.00000211, 0.00000155, -0.00000444, -0.00000138, 0.00000186, 0.00000371, -0.00000057, 0.00000050, -0.00000120, -0.00000309, 0.00000291, -0.00000074, 0.00000236, 0.00000115, -0.00000606, },
		{ -0.00000580, 0.00000116, 0.00000307, -0.00000103, 0.00000595, -0.00000384, -0.00000330, -0.00000420, 0.00000102, -0.00000169, -0.00000038, 0.00001348, -0.00000317, -0.00000285, -0.00000189, -0.00000165, 0.00000241, -0.00000293, 0.00001371, -0.00000247, -0.00001146, 0.00002916, -0.00002427, 0.00000378, -0.00000943, -0.00001072, 0.00000965, -0.00001780, 0.00001172, -0.00002459, 0.00000067, -0.00000769, -0.00002988, 0.00001229, -0.00005386, 0.00001200, -0.00020921, -0.00000113, 0.00069565, -0.00009083, -0.00020239, 0.00002612, -0.00007588, 0.00003434, -0.00006679, 0.00005787, -0.00007142, 0.00006003, -0.00003822, 0.00004063, -0.00002736, 0.00001685, -0.00000622, -0.00000693, 0.00002700, -0.00007174, 0.00006792, -0.00003842, 0.00003410, -0.00004060, 0.00002886, -0.00002734, 0.00001443, -0.00000228, -0.00000344, 0.00000333, -0.00002127, 0.00002917, -0.00000908, 0.00000968, -0.00001831, 0.00001260, -0.00002239, 0.00002240, -0.00001422, 0.00001051, -0.00000926, -0.00000089, 0.00000816, -0.00000723, 0.00000873, -0.00000741, 0.00001232, -0.00001590, 0.00000085, 0.00000661, -0.00000365, 0.00000288, -0.00000110, -0.00000359, 0.00000615, -0.00000425, 0.00000182, -0.00000115, -0.00000008, 0.00000362, -0.00000434, -0.00000082, -0.00000365, 0.00000155, 0.00000155, 0.00000129, 0.00000773, -0.00000639, -0.00000308, 0.00000381, -0.00000223, 0.00000079, -0.00000125, -0.00000015, 0.00000178, -0.00000382, 0.00000295, 0.00000068, -0.00000466, -0.00000113, 0.00000242, 0.00000338, -0.00000080, 0.00000100, -0.00000188, -0.00000265, 0.00000296, -0.00000076, 0.00000281, -0.00000011, -0.00000538, 0.00000137, },
		{ 0.00000224, 0.00000227, -0.00000022, 0.00000533, -0.00000383, -0.00000319, -0.00000464, 0.00000164, -0.00000267, 0.00000150, 0.00001316, -0.00000450, -0.00000244, -0.00000201, -0.00000104, 0.00000163, -0.00000168, 0.00001467, -0.00000648, -0.00000668, 0.00002571, -0.00002503, 0.00000769, -0.00001518, -0.00000528, 0.00000586, -0.00001525, 0.00001067, -0.00002636, 0.00000452, -0.00001365, -0.00002268, 0.00000321, -0.00004551, 0.00000025, -0.00022180, 0.00008496, 0.00064458, -0.00013652, -0.00016289, 0.00000692, -0.00005773, 0.00001962, -0.00005470, 0.00004917, -0.00006555, 0.00006000, -0.00003982, 0.00004511, -0.00003343, 0.00002625, -0.00001898, 0.00000985, 0.00000774, -0.00005718, 0.00006138, -0.00003731, 0.00003471, -0.00004356, 0.00003381, -0.00003433, 0.00002304, -0.00001010, 0.00000267, -0.00000364, -0.00001464, 0.00002690, -0.00000864, 0.00000710, -0.00001563, 0.00001121, -0.00002190, 0.00002372, -0.00001714, 0.00001438, -0.00001409, 0.00000422, 0.00000460, -0.00000506, 0.00000733, -0.00000637, 0.00001174, -0.00001786, 0.00000436, 0.00000478, -0.00000341, 0.00000333, -0.00000193, -0.00000302, 0.00000600, -0.00000464, 0.00000247, -0.00000178, 0.00000071, 0.00000307, -0.00000457, -0.00000080, -0.00000346, 0.00000230, 0.00000086, 0.00000228, 0.00000682, -0.00000763, -0.00000158, 0.00000352, -0.00000231, 0.00000096, -0.00000156, 0.00000037, 0.00000125, -0.00000367, 0.00000364, -0.00000028, -0.00000478, -0.00000081, 0.00000297, 0.00000288, -0.00000077, 0.00000133, -0.00000257, -0.00000210, 0.00000287, -0.00000059, 0.00000299, -0.00000126, -0.00000455, 0.00000094, -0.00000602, },
		{ 0.00000158, 0.00000056, 0.00000463, -0.00000364, -0.00000340, -0.00000478, 0.00000208, -0.00000350, 0.00000351, 0.00001242, -0.00000560, -0.00000205, -0.00000208, -0.00000051, 0.00000081, 0.00000004, 0.00001471, -0.00000982, -0.00000185, 0.00002126, -0.00002404, 0.00000988, -0.00001999, 0.00000017, 0.00000142, -0.00001163, 0.00000811, -0.00002635, 0.00000660, -0.00001803, -0.00001663, -0.00000577, -0.00003534, -0.00001785, -0.00022078, 0.00016233, 0.00058405, -0.00016734, -0.00013089, -0.00000876, -0.00004161, 0.00000536, -0.00004191, 0.00003885, -0.00005719, 0.00005707, -0.00003836, 0.00004639, -0.00003603, 0.00003194, -0.00002821, 0.00002402, -0.00001116, -0.00004062, 0.00005155, -0.00003278, 0.00003219, -0.00004368, 0.00003627, -0.00003932, 0.00003046, -0.00001757, 0.00000872, -0.00001077, -0.00000708, 0.00002342, -0.00000759, 0.00000416, -0.00001219, 0.00000873, -0.00002018, 0.00002372, -0.00001884, 0.00001731, -0.00001842, 0.00000943, 0.00000047, -0.00000218, 0.00000522, -0.00000459, 0.00001022, -0.00001891, 0.00000760, 0.00000263, -0.00000273, 0.00000347, -0.00000267, -0.00000241, 0.00000574, -0.00000490, 0.00000304, -0.00000237, 0.00000154, 0.00000237, -0.00000467, -0.00000084, -0.00000307, 0.00000284, 0.00000024, 0.00000334, 0.00000553, -0.00000847, -0.00000008, 0.00000307, -0.00000226, 0.00000103, -0.00000178, 0.00000085, 0.00000066, -0.00000331, 0.00000415, -0.00000127, -0.00000482, -0.00000040, 0.00000345, 0.00000225, -0.00000050, 0.00000143, -0.00000318, -0.00000149, 0.00000269, -0.00000025, 0.00000289, -0.00000221, -0.00000369, 0.00000030, -0.00000593, 0.00000308, },
		{ 0.00000125, 0.00000393, -0.00000334, -0.00000386, -0.00000460, 0.00000228, -0.00000411, 0.00000556, 0.00001131, -0.00000644, -0.00000171, -0.00000206, -0.00000011, 0.00000006, 0.00000211, 0.00001382, -0.00001223, 0.00000264, 0.00001623, -0.00002154, 0.00001027, -0.00002346, 0.00000514, -0.00000317, -0.00000737, 0.00000438, -0.00002473, 0.00000685, -0.00002050, -0.00001244, -0.00001345, -0.00002511, -0.00004032, -0.00020659, 0.00022806, 0.00051835, -0.00018561, -0.00010620, -0.00002035, -0.00002841, -0.00000736, -0.00002955, 0.00002800, -0.00004732, 0.00005210, -0.00003454, 0.00004502, -0.00003551, 0.00003398, -0.00003347, 0.00003457, -0.00002827, -0.00002345, 0.00003948, -0.00002544, 0.00002679, -0.00004097, 0.00003604, -0.00004191, 0.00003617, -0.00002416, 0.00001423, -0.00001762, 0.00000101, 0.00001893, -0.00000600, 0.00000101, -0.00000819, 0.00000534, -0.00001733, 0.00002238, -0.00001917, 0.00001902, -0.00002186, 0.00001436, -0.00000393, 0.00000122, 0.00000254, -0.00000215, 0.00000775, -0.00001888, 0.00001030, 0.00000036, -0.00000172, 0.00000333, -0.00000326, -0.00000182, 0.00000541, -0.00000504, 0.00000351, -0.00000289, 0.00000236, 0.00000155, -0.00000466, -0.00000091, -0.00000251, 0.00000312, -0.00000023, 0.00000438, 0.00000392, -0.00000886, 0.00000132, 0.00000249, -0.00000212, 0.00000099, -0.00000187, 0.00000124, 0.00000004, -0.00000278, 0.00000446, -0.00000227, -0.00000477, 0.00000012, 0.00000379, 0.00000159, -0.00000002, 0.00000128, -0.00000364, -0.00000090, 0.00000249, 0.00000018, 0.00000253, -0.00000291, -0.00000288, -0.00000046, -0.00000555, 0.00000368, 0.00000104, },
		{ 0.00000331, -0.00000305, -0.00000451, -0.00000413, 0.00000221, -0.00000442, 0.00000751, 0.00000990, -0.00000701, -0.00000145, -0.00000192, 0.00000010, -0.00000050, 0.00000435, 0.00001209, -0.00001359, 0.00000643, 0.00001110, -0.00001792, 0.00000892, -0.00002535, 0.00000919, -0.00000744, -0.00000297, -0.00000005, -0.00002184, 0.00000545, -0.00002098, -0.00001055, -0.00001886, -0.00001653, -0.00006477, -0.00018060, 0.00028023, 0.00045139, -0.00019390, -0.00008806, -0.00002781, -0.00001857, -0.00001780, -0.00001850, 0.00001758, -0.00003689, 0.00004596, -0.00002915, 0.00004171, -0.00003249, 0.00003274, -0.00003477, 0.00004087, -0.00004242, -0.00000704, 0.00002636, -0.00001612, 0.00001903, -0.00003567, 0.00003314, -0.00004190, 0.00003977, -0.00002943, 0.00001879, -0.00002372, 0.00000920, 0.00001368, -0.00000399, -0.00000218, -0.00000387, 0.00000126, -0.00001351, 0.00001981, -0.00001811, 0.00001935, -0.00002413, 0.00001863, -0.00000826, 0.00000485, -0.00000052, 0.00000077, 0.00000443, -0.00001769, 0.00001223, -0.00000178, -0.00000051, 0.00000294, -0.00000372, -0.00000126, 0.00000503, -0.00000508, 0.00000388, -0.00000333, 0.00000313, 0.00000065, -0.00000456, -0.00000099, -0.00000180, 0.00000312, -0.00000047, 0.00000528, 0.00000208, -0.00000880, 0.00000256, 0.00000186, -0.00000190, 0.00000086, -0.00000186, 0.00000153, -0.00000054, -0.00000212, 0.00000457, -0.00000326, -0.00000464, 0.00000072, 0.00000396, 0.00000096, 0.00000061, 0.00000087, -0.00000391, -0.00000039, 0.00000233, 0.00000064, 0.00000193, -0.00000332, -0.00000222, -0.00000127, -0.00000497, 0.00000405, 0.00000066, 0.00000182, },
		{ -0.00000284, -0.00000523, -0.00000341, 0.00000185, -0.00000438, 0.00000928, 0.00000828, -0.00000733, -0.00000128, -0.00000166, 0.00000007, -0.00000073, 0.00000655, 0.00000968, -0.00001388, 0.00000925, 0.00000631, -0.00001365, 0.00000609, -0.00002557, 0.00001203, -0.00001099, 0.00000112, -0.00000470, -0.00001814, 0.00000275, -0.00001966, -0.00001107, -0.00002138, -0.00001108, -0.00008867, -0.00014495, 0.00031799, 0.00038645, -0.00019489, -0.00007530, -0.00003158, -0.00001210, -0.00002561, -0.00000936, 0.00000832, -0.00002670, 0.00003944, -0.00002297, 0.00003727, -0.00002776, 0.00002892, -0.00003247, 0.00004272, -0.00005274, 0.00000742, 0.00001336, -0.00000580, 0.00000962, -0.00002827, 0.00002782, -0.00003926, 0.00004102, -0.00003303, 0.00002203, -0.00002864, 0.00001700, 0.00000797, -0.00000171, -0.00000524, 0.00000051, -0.00000323, -0.00000897, 0.00001620, -0.00001574, 0.00001822, -0.00002499, 0.00002191, -0.00001215, 0.00000843, -0.00000371, 0.00000397, 0.00000042, -0.00001535, 0.00001321, -0.00000357, 0.00000074, 0.00000237, -0.00000405, -0.00000073, 0.00000462, -0.00000501, 0.00000416, -0.00000368, 0.00000384, -0.00000030, -0.00000439, -0.00000104, -0.00000099, 0.00000287, -0.00000043, 0.00000593, 0.00000013, -0.00000831, 0.00000356, 0.00000122, -0.00000164, 0.00000068, -0.00000174, 0.00000169, -0.00000105, -0.00000136, 0.00000447, -0.00000421, -0.00000440, 0.00000141, 0.00000391, 0.00000047, 0.00000130, 0.00000023, -0.00000395, -0.00000000, 0.00000228, 0.00000106, 0.00000117, -0.00000346, -0.00000175, -0.00000202, -0.00000425, 0.00000422, 0.00000045, 0.00000224, 0.00000280, },
		{ -0.00000592, -0.00000251, 0.00000124, -0.00000395, 0.00001076, 0.00000654, -0.00000745, -0.00000118, -0.00000131, -0.00000019, -0.00000055, 0.00000851, 0.00000680, -0.00001315, 0.00001096, 0.00000227, -0.00000922, 0.00000214, -0.00002423, 0.00001351, -0.00001353, 0.00000451, -0.00000909, -0.00001416, -0.00000075, -0.00001698, -0.00001377, -0.00002079, -0.00000983, -0.00010956, -0.00010235, 0.00034152, 0.00032607, -0.00019107, -0.00006658, -0.00003236, -0.00000866, -0.00003076, -0.00000240, 0.00000073, -0.00001736, 0.00003322, -0.00001673, 0.00003249, -0.00002218, 0.00002337, -0.00002728, 0.00004034, -0.00005878, 0.00001898, 0.00000158, 0.00000453, -0.00000061, -0.00001941, 0.00002050, -0.00003421, 0.00003987, -0.00003476, 0.00002368, -0.00003199, 0.00002398, 0.00000214, 0.00000068, -0.00000801, 0.00000469, -0.00000783, -0.00000402, 0.00001179, -0.00001224, 0.00001571, -0.00002435, 0.00002394, -0.00001532, 0.00001165, -0.00000677, 0.00000715, -0.00000402, -0.00001197, 0.00001315, -0.00000484, 0.00000187, 0.00000173, -0.00000430, -0.00000021, 0.00000418, -0.00000487, 0.00000434, -0.00000393, 0.00000444, -0.00000127, -0.00000416, -0.00000102, -0.00000016, 0.00000240, -0.00000010, 0.00000624, -0.00000182, -0.00000744, 0.00000430, 0.00000062, -0.00000139, 0.00000047, -0.00000154, 0.00000173, -0.00000145, -0.00000056, 0.00000417, -0.00000511, -0.00000404, 0.00000213, 0.00000363, 0.00000018, 0.00000197, -0.00000058, -0.00000377, 0.00000024, 0.00000237, 0.00000133, 0.00000033, -0.00000336, -0.00000150, -0.00000266, -0.00000345, 0.00000424, 0.00000037, 0.00000252, 0.00000244, -0.00000280, },
		{ -0.00000153, 0.00000042, -0.00000314, 0.00001188, 0.00000479, -0.00000741, -0.00000112, -0.00000089, -0.00000066, 0.00000009, 0.00001002, 0.00000373, -0.00001158, 0.00001151, -0.00000073, -0.00000513, -0.00000246, -0.00002157, 0.00001365, -0.00001493, 0.00000691, -0.00001280, -0.00001040, -0.00000449, -0.00001351, -0.00001815, -0.00001729, -0.00001339, -0.00012529, -0.00005582, 0.00035185, 0.00027195, -0.00018460, -0.00006054, -0.00003107, -0.00000770, -0.00003353, 0.00000236, -0.00000498, -0.00000926, 0.00002778, -0.00001099, 0.00002804, -0.00001659, 0.00001705, -0.00002015, 0.00003433, -0.00006049, 0.00002702, -0.00000808, 0.00001393, -0.00001079, -0.00000981, 0.00001181, -0.00002713, 0.00003644, -0.00003457, 0.00002362, -0.00003353, 0.00002972, -0.00000347, 0.00000300, -0.00001032, 0.00000845, -0.00001227, 0.00000105, 0.00000691, -0.00000788, 0.00001196, -0.00002221, 0.00002458, -0.00001749, 0.00001426, -0.00000943, 0.00001004, -0.00000857, -0.00000776, 0.00001207, -0.00000543, 0.00000272, 0.00000110, -0.00000452, 0.00000031, 0.00000370, -0.00000464, 0.00000443, -0.00000407, 0.00000493, -0.00000222, -0.00000388, -0.00000091, 0.00000062, 0.00000177, 0.00000051, 0.00000616, -0.00000362, -0.00000627, 0.00000477, 0.00000010, -0.00000116, 0.00000027, -0.00000129, 0.00000165, -0.00000172, 0.00000023, 0.00000370, -0.00000594, -0.00000355, 0.00000284, 0.00000315, 0.00000014, 0.00000250, -0.00000148, -0.00000341, 0.00000033, 0.00000263, 0.00000141, -0.00000048, -0.00000308, -0.00000147, -0.00000313, -0.00000265, 0.00000417, 0.00000039, 0.00000268, 0.00000220, -0.00000296, -0.00000648, },
		{ -0.00000052, -0.00000199, 0.00001260, 0.00000310, -0.00000728, -0.00000106, -0.00000046, -0.00000127, 0.00000119, 0.00001092, 0.00000073, -0.00000942, 0.00001101, -0.00000252, -0.00000179, -0.00000724, -0.00001792, 0.00001258, -0.00001517, 0.00000819, -0.00001553, -0.00000730, -0.00000791, -0.00000990, -0.00002350, -0.00001146, -0.00002180, -0.00013418, -0.00000843, 0.00035074, 0.00022500, -0.00017718, -0.00005593, -0.00002862, -0.00000848, -0.00003436, 0.00000516, -0.00000880, -0.00000256, 0.00002337, -0.00000611, 0.00002444, -0.00001170, 0.00001089, -0.00001216, 0.00002556, -0.00005822, 0.00003131, -0.00001499, 0.00002158, -0.00002011, -0.00000029, 0.00000242, -0.00001856, 0.00003104, -0.00003256, 0.00002186, -0.00003309, 0.00003389, -0.00000852, 0.00000507, -0.00001203, 0.00001156, -0.00001628, 0.00000592, 0.00000189, -0.00000298, 0.00000723, -0.00001868, 0.00002376, -0.00001851, 0.00001602, -0.00001144, 0.00001235, -0.00001289, -0.00000299, 0.00001003, -0.00000530, 0.00000319, 0.00000061, -0.00000478, 0.00000088, 0.00000316, -0.00000432, 0.00000443, -0.00000410, 0.00000528, -0.00000315, -0.00000357, -0.00000068, 0.00000129, 0.00000106, 0.00000136, 0.00000563, -0.00000516, -0.00000490, 0.00000499, -0.00000033, -0.00000098, 0.00000011, -0.00000103, 0.00000148, -0.00000183, 0.00000098, 0.00000307, -0.00000670, -0.00000290, 0.00000348, 0.00000251, 0.00000037, 0.00000282, -0.00000237, -0.00000292, 0.00000030, 0.00000307, 0.00000124, -0.00000119, -0.00000270, -0.00000162, -0.00000342, -0.00000189, 0.00000404, 0.00000047, 0.00000277, 0.00000207, -0.00000334, -0.00000684, -0.00000057, },
		{ -0.00000054, 0.00001288, 0.00000155, -0.00000710, -0.00000096, -0.00000010, -0.00000195, 0.00000272, 0.00001112, -0.00000193, -0.00000695, 0.00000963, -0.00000308, 0.00000048, -0.00001171, -0.00001368, 0.00001056, -0.00001437, 0.00000834, -0.00001714, -0.00000516, -0.00001055, -0.00000681, -0.00002900, -0.00000420, -0.00003456, -0.00013517, 0.00003692, 0.00034040, 0.00018538, -0.00017001, -0.00005177, -0.00002588, -0.00001027, -0.00003383, 0.00000639, -0.00001095, 0.00000280, 0.00002008, -0.00000225, 0.00002200, -0.00000804, 0.00000568, -0.00000436, 0.00001514, -0.00005265, 0.00003196, -0.00001880, 0.00002692, -0.00002787, 0.00000843, -0.00000692, -0.00000914, 0.00002410, -0.00002898, 0.00001855, -0.00003068, 0.00003626, -0.00001275, 0.00000672, -0.00001303, 0.00001387, -0.00001962, 0.00001030, -0.00000291, 0.00000209, 0.00000187, -0.00001400, 0.00002155, -0.00001829, 0.00001680, -0.00001264, 0.00001383, -0.00001662, 0.00000197, 0.00000724, -0.00000445, 0.00000319, 0.00000031, -0.00000514, 0.00000154, 0.00000254, -0.00000390, 0.00000433, -0.00000404, 0.00000549, -0.00000402, -0.00000322, -0.00000032, 0.00000178, 0.00000038, 0.00000237, 0.00000467, -0.00000634, -0.00000344, 0.00000498, -0.00000066, -0.00000087, 0.00000000, -0.00000078, 0.00000124, -0.00000180, 0.00000165, 0.00000230, -0.00000735, -0.00000210, 0.00000398, 0.00000177, 0.00000087, 0.00000285, -0.00000318, -0.00000236, 0.00000020, 0.00000365, 0.00000080, -0.00000172, -0.00000230, -0.00000192, -0.00000353, -0.00000120, 0.00000389, 0.00000059, 0.00000283, 0.00000200, -0.00000393, -0.00000693, 0.00000029, -0.00000151, },
		{ 0.00001275, 0.00000020, -0.00000693, -0.00000078, 0.00000015, -0.00000257, 0.00000457, 0.00001057, -0.00000405, -0.00000448, 0.00000767, -0.00000250, 0.00000150, -0.00001550, -0.00000929, 0.00000790, -0.00001277, 0.00000749, -0.00001761, -0.00000415, -0.00001209, -0.00000481, -0.00003379, 0.00000340, -0.00005066, -0.00012788, 0.00007776, 0.00032332, 0.00015265, -0.00016379, -0.00004735, -0.00002353, -0.00001240, -0.00003254, 0.00000651, -0.00001173, 0.00000700, 0.00001777, 0.00000063, 0.00002077, -0.00000588, 0.00000200, 0.00000226, 0.00000421, -0.00004468, 0.00002941, -0.00001951, 0.00002959, -0.00003352, 0.00001570, -0.00001549, 0.00000044, 0.00001615, -0.00002422, 0.00001397, -0.00002644, 0.00003672, -0.00001593, 0.00000784, -0.00001328, 0.00001528, -0.00002213, 0.00001396, -0.00000720, 0.00000697, -0.00000376, -0.00000847, 0.00001813, -0.00001687, 0.00001654, -0.00001291, 0.00001429, -0.00001945, 0.00000678, 0.00000394, -0.00000296, 0.00000272, 0.00000026, -0.00000562, 0.00000233, 0.00000180, -0.00000335, 0.00000412, -0.00000387, 0.00000555, -0.00000482, -0.00000283, 0.00000016, 0.00000204, -0.00000020, 0.00000344, 0.00000331, -0.00000709, -0.00000198, 0.00000479, -0.00000090, -0.00000081, -0.00000003, -0.00000058, 0.00000098, -0.00000163, 0.00000220, 0.00000142, -0.00000789, -0.00000113, 0.00000429, 0.00000101, 0.00000160, 0.00000258, -0.00000380, -0.00000181, 0.00000008, 0.00000431, 0.00000010, -0.00000201, -0.00000196, -0.00000229, -0.00000348, -0.00000059, 0.00000373, 0.00000071, 0.00000288, 0.00000192, -0.00000470, -0.00000672, 0.00000096, -0.00000244, 0.00000112, },
		{ -0.00000092, -0.00000680, -0.00000048, 0.00000020, -0.00000302, 0.00000660, 0.00000931, -0.00000548, -0.00000228, 0.00000544, -0.00000100, 0.00000121, -0.00001829, -0.00000512, 0.00000495, -0.00001064, 0.00000588, -0.00001709, -0.00000427, -0.00001235, -0.00000428, -0.00003713, 0.00001016, -0.00006872, -0.00011260, 0.00011216, 0.00030198, 0.00012594, -0.00015875, -0.00004231, -0.00002203, -0.00001430, -0.00003101, 0.00000602, -0.00001153, 0.00001034, 0.00001621, 0.00000274, 0.00002063, -0.00000527, 0.00000017, 0.00000696, -0.00000612, -0.00003538, 0.00002435, -0.00001739, 0.00002953, -0.00003677, 0.00002103, -0.00002269, 0.00000951, 0.00000780, -0.00001871, 0.00000852, -0.00002064, 0.00003529, -0.00001792, 0.00000834, -0.00001273, 0.00001571, -0.00002369, 0.00001671, -0.00001070, 0.00001129, -0.00000924, -0.00000246, 0.00001373, -0.00001436, 0.00001526, -0.00001223, 0.00001364, -0.00002112, 0.00001106, 0.00000042, -0.00000100, 0.00000182, 0.00000047, -0.00000626, 0.00000327, 0.00000092, -0.00000266, 0.00000380, -0.00000359, 0.00000545, -0.00000554, -0.00000240, 0.00000074, 0.00000204, -0.00000057, 0.00000445, 0.00000164, -0.00000738, -0.00000063, 0.00000448, -0.00000108, -0.00000080, 0.00000001, -0.00000046, 0.00000072, -0.00000134, 0.00000263, 0.00000044, -0.00000829, -0.00000003, 0.00000435, 0.00000034, 0.00000250, 0.00000199, -0.00000419, -0.00000134, 0.00000002, 0.00000499, -0.00000082, -0.00000204, -0.00000175, -0.00000267, -0.00000330, -0.00000007, 0.00000359, 0.00000083, 0.00000296, 0.00000178, -0.00000557, -0.00000620, 0.00000138, -0.00000322, 0.00000290, 0.00001222, },
		{ -0.00000673, -0.00000006, 0.00000003, -0.00000318, 0.00000863, 0.00000745, -0.00000614, -0.00000058, 0.00000328, 0.00000111, -0.00000028, -0.00001991, -0.00000148, 0.00000203, -0.00000831, 0.00000380, -0.00001582, -0.00000536, -0.00001137, -0.00000542, -0.00003846, 0.00001497, -0.00008711, -0.00009030, 0.00013884, 0.00027869, 0.00010409, -0.00015478, -0.00003655, -0.00002161, -0.00001561, -0.00002969, 0.00000535, -0.00001074, 0.00001314, 0.00001509, 0.00000441, 0.00002128, -0.00000601, 0.00000022, 0.00000924, -0.00001490, -0.00002581, 0.00001766, -0.00001297, 0.00002695, -0.00003753, 0.00002415, -0.00002804, 0.00001748, -0.00000039, -0.00001295, 0.00000266, -0.00001367, 0.00003211, -0.00001867, 0.00000819, -0.00001141, 0.00001517, -0.00002425, 0.00001846, -0.00001323, 0.00001478, -0.00001421, 0.00000363, 0.00000869, -0.00001098, 0.00001308, -0.00001067, 0.00001186, -0.00002144, 0.00001448, -0.00000298, 0.00000122, 0.00000057, 0.00000091, -0.00000704, 0.00000436, -0.00000012, -0.00000181, 0.00000333, -0.00000320, 0.00000519, -0.00000615, -0.00000190, 0.00000138, 0.00000177, -0.00000067, 0.00000528, -0.00000025, -0.00000720, 0.00000055, 0.00000410, -0.00000121, -0.00000081, 0.00000012, -0.00000042, 0.00000051, -0.00000097, 0.00000291, -0.00000061, -0.00000851, 0.00000118, 0.00000414, -0.00000017, 0.00000345, 0.00000111, -0.00000431, -0.00000102, 0.00000009, 0.00000559, -0.00000189, -0.00000182, -0.00000171, -0.00000300, -0.00000303, 0.00000037, 0.00000346, 0.00000094, 0.00000307, 0.00000153, -0.00000648, -0.00000542, 0.00000151, -0.00000377, 0.00000468, 0.00001138, -0.00000181, },
		{ 0.00000047, -0.00000039, -0.00000297, 0.00001049, 0.00000514, -0.00000605, 0.00000046, 0.00000150, 0.00000347, -0.00000276, -0.00002031, 0.00000140, -0.00000059, -0.00000606, 0.00000158, -0.00001411, -0.00000714, -0.00000934, -0.00000820, -0.00003748, 0.00001691, -0.00010411, -0.00006241, 0.00015724, 0.00025542, 0.00008587, -0.00015150, -0.00003025, -0.00002227, -0.00001615, -0.00002886, 0.00000483, -0.00000968, 0.00001571, 0.00001410, 0.00000600, 0.00002230, -0.00000773, 0.00000193, 0.00000893, -0.00002141, -0.00001698, 0.00001026, -0.00000694, 0.00002224, -0.00003598, 0.00002496, -0.00003124, 0.00002385, -0.00000789, -0.00000742, -0.00000311, -0.00000601, 0.00002745, -0.00001821, 0.00000740, -0.00000939, 0.00001371, -0.00002384, 0.00001917, -0.00001468, 0.00001722, -0.00001834, 0.00000939, 0.00000337, -0.00000700, 0.00001020, -0.00000839, 0.00000905, -0.00002035, 0.00001677, -0.00000597, 0.00000349, -0.00000088, 0.00000150, -0.00000792, 0.00000558, -0.00000130, -0.00000079, 0.00000272, -0.00000270, 0.00000476, -0.00000665, -0.00000133, 0.00000203, 0.00000127, -0.00000045, 0.00000582, -0.00000222, -0.00000658, 0.00000149, 0.00000370, -0.00000133, -0.00000083, 0.00000028, -0.00000048, 0.00000038, -0.00000054, 0.00000304, -0.00000172, -0.00000854, 0.00000244, 0.00000363, -0.00000043, 0.00000437, 0.00000000, -0.00000415, -0.00000087, 0.00000034, 0.00000604, -0.00000302, -0.00000138, -0.00000185, -0.00000323, -0.00000272, 0.00000076, 0.00000332, 0.00000107, 0.00000321, 0.00000113, -0.00000734, -0.00000441, 0.00000134, -0.00000402, 0.00000636, 0.00001028, -0.00000248, -0.00000672, },
		{ -0.00000105, -0.00000230, 0.00001198, 0.00000259, -0.00000532, 0.00000079, 0.00000032, 0.00000572, -0.00000593, -0.00001957, 0.00000341, -0.00000273, -0.00000414, -0.00000048, -0.00001231, -0.00000928, -0.00000661, -0.00001235, -0.00003417, 0.00001534, -0.00011808, -0.00003080, 0.00016745, 0.00023368, 0.00007006, -0.00014842, -0.00002374, -0.00002383, -0.00001591, -0.00002864, 0.00000469, -0.00000857, 0.00001827, 0.00001297, 0.00000787, 0.00002324, -0.00000993, 0.00000483, 0.00000618, -0.00002523, -0.00000971, 0.00000309, -0.00000011, 0.00001600, -0.00003245, 0.00002361, -0.00003220, 0.00002830, -0.00001426, -0.00000256, -0.00000831, 0.00000182, 0.00002166, -0.00001669, 0.00000604, -0.00000678, 0.00001145, -0.00002252, 0.00001889, -0.00001502, 0.00001847, -0.00002136, 0.00001449, -0.00000188, -0.00000271, 0.00000686, -0.00000560, 0.00000541, -0.00001789, 0.00001778, -0.00000826, 0.00000554, -0.00000236, 0.00000215, -0.00000883, 0.00000689, -0.00000261, 0.00000040, 0.00000195, -0.00000209, 0.00000416, -0.00000701, -0.00000068, 0.00000264, 0.00000057, 0.00000010, 0.00000597, -0.00000413, -0.00000559, 0.00000217, 0.00000334, -0.00000146, -0.00000082, 0.00000046, -0.00000061, 0.00000034, -0.00000009, 0.00000302, -0.00000285, -0.00000833, 0.00000370, 0.00000287, -0.00000038, 0.00000513, -0.00000124, -0.00000375, -0.00000090, 0.00000081, 0.00000625, -0.00000413, -0.00000079, -0.00000218, -0.00000334, -0.00000239, 0.00000112, 0.00000316, 0.00000123, 0.00000336, 0.00000056, -0.00000808, -0.00000324, 0.00000089, -0.00000393, 0.00000786, 0.00000902, -0.00000296, -0.00000676, 0.00000110, },
		{ -0.00000116, 0.00001295, 0.00000002, -0.00000409, 0.00000044, -0.00000008, 0.00000752, -0.00000946, -0.00001786, 0.00000455, -0.00000428, -0.00000270, -0.00000215, -0.00001074, -0.00001138, -0.00000360, -0.00001746, -0.00002879, 0.00000998, -0.00012762, 0.00000252, 0.00017012, 0.00021450, 0.00005563, -0.00014499, -0.00001744, -0.00002597, -0.00001505, -0.00002903, 0.00000501, -0.00000754, 0.00002096, 0.00001152, 0.00001031, 0.00002367, -0.00001206, 0.00000830, 0.00000141, -0.00002628, -0.00000456, -0.00000306, 0.00000674, 0.00000890, -0.00002746, 0.00002043, -0.00003100, 0.00003068, -0.00001921, 0.00000131, -0.00001252, 0.00000930, 0.00001515, -0.00001428, 0.00000424, -0.00000373, 0.00000852, -0.00002043, 0.00001773, -0.00001431, 0.00001850, -0.00002310, 0.00001861, -0.00000670, 0.00000156, 0.00000332, -0.00000257, 0.00000121, -0.00001421, 0.00001742, -0.00000965, 0.00000715, -0.00000370, 0.00000272, -0.00000967, 0.00000820, -0.00000400, 0.00000172, 0.00000104, -0.00000136, 0.00000340, -0.00000721, 0.00000004, 0.00000315, -0.00000025, 0.00000095, 0.00000566, -0.00000584, -0.00000434, 0.00000259, 0.00000302, -0.00000162, -0.00000076, 0.00000064, -0.00000082, 0.00000042, 0.00000034, 0.00000285, -0.00000400, -0.00000787, 0.00000487, 0.00000188, 0.00000002, 0.00000564, -0.00000251, -0.00000314, -0.00000111, 0.00000151, 0.00000617, -0.00000512, -0.00000010, -0.00000267, -0.00000330, -0.00000208, 0.00000144, 0.00000298, 0.00000145, 0.00000347, -0.00000018, -0.00000863, -0.00000199, 0.00000019, -0.00000349, 0.00000907, 0.00000769, -0.00000329, -0.00000681, 0.00000178, -0.00000190, },
		{ 0.00001329, -0.00000234, -0.00000257, -0.00000047, 0.00000034, 0.00000861, -0.00001299, -0.00001543, 0.00000488, -0.00000523, -0.00000182, -0.00000325, -0.00000964, -0.00001311, -0.00000078, -0.00002295, -0.00002188, 0.00000097, -0.00013166, 0.00003547, 0.00016639, 0.00019841, 0.00004179, -0.00014076, -0.00001179, -0.00002832, -0.00001381, -0.00002989, 0.00000576, -0.00000660, 0.00002379, 0.00000969, 0.00001349, 0.00002325, -0.00001363, 0.00001165, -0.00000474, -0.00002476, -0.00000182, -0.00000759, 0.00001288, 0.00000163, -0.00002160, 0.00001586, -0.00002792, 0.00003103, -0.00002258, 0.00000396, -0.00001539, 0.00001593, 0.00000837, -0.00001124, 0.00000214, -0.00000041, 0.00000510, -0.00001773, 0.00001584, -0.00001269, 0.00001738, -0.00002349, 0.00002155, -0.00001080, 0.00000552, -0.00000012, 0.00000041, -0.00000322, -0.00000957, 0.00001575, -0.00001000, 0.00000813, -0.00000474, 0.00000310, -0.00001032, 0.00000942, -0.00000538, 0.00000314, -0.00000001, -0.00000054, 0.00000247, -0.00000724, 0.00000084, 0.00000350, -0.00000111, 0.00000207, 0.00000486, -0.00000723, -0.00000292, 0.00000278, 0.00000276, -0.00000183, -0.00000063, 0.00000078, -0.00000106, 0.00000062, 0.00000073, 0.00000254, -0.00000513, -0.00000715, 0.00000587, 0.00000075, 0.00000077, 0.00000582, -0.00000370, -0.00000242, -0.00000144, 0.00000241, 0.00000577, -0.00000591, 0.00000059, -0.00000327, -0.00000314, -0.00000178, 0.00000174, 0.00000278, 0.00000174, 0.00000351, -0.00000106, -0.00000894, -0.00000075, -0.00000071, -0.00000268, 0.00000995, 0.00000638, -0.00000354, -0.00000684, 0.00000244, -0.00000287, 0.00000042, },
		{ -0.00000431, -0.00000100, -0.00000176, 0.00000152, 0.00000881, -0.00001618, -0.00001254, 0.00000456, -0.00000566, -0.00000146, -0.00000374, -0.00000919, -0.00001419, 0.00000142, -0.00002821, -0.00001417, -0.00001120, -0.00012959, 0.00006616, 0.00015770, 0.00018547, 0.00002804, -0.00013539, -0.00000717, -0.00003051, -0.00001249, -0.00003102, 0.00000682, -0.00000566, 0.00002668, 0.00000754, 0.00001744, 0.00002177, -0.00001423, 0.00001421, -0.00001150, -0.00002115, -0.00000150, -0.00001011, 0.00001771, -0.00000515, -0.00001550, 0.00001047, -0.00002338, 0.00002951, -0.00002437, 0.00000531, -0.00001672, 0.00002131, 0.00000176, -0.00000785, -0.00000007, 0.00000300, 0.00000137, -0.00001460, 0.00001342, -0.00001035, 0.00001524, -0.00002257, 0.00002318, -0.00001396, 0.00000891, -0.00000323, 0.00000305, -0.00000752, -0.00000429, 0.00001291, -0.00000929, 0.00000837, -0.00000532, 0.00000315, -0.00001069, 0.00001045, -0.00000669, 0.00000459, -0.00000115, 0.00000037, 0.00000139, -0.00000707, 0.00000168, 0.00000365, -0.00000192, 0.00000335, 0.00000358, -0.00000819, -0.00000145, 0.00000277, 0.00000257, -0.00000206, -0.00000043, 0.00000085, -0.00000132, 0.00000092, 0.00000105, 0.00000210, -0.00000621, -0.00000615, 0.00000662, -0.00000043, 0.00000181, 0.00000560, -0.00000471, -0.00000166, -0.00000183, 0.00000346, 0.00000504, -0.00000644, 0.00000121, -0.00000391, -0.00000286, -0.00000151, 0.00000201, 0.00000256, 0.00000211, 0.00000342, -0.00000204, -0.00000898, 0.00000040, -0.00000172, -0.00000156, 0.00001045, 0.00000516, -0.00000376, -0.00000679, 0.00000302, -0.00000386, 0.00000239, 0.00001295, },
		{ 0.00000041, -0.00000319, 0.00000331, 0.00000805, -0.00001878, -0.00000949, 0.00000378, -0.00000570, -0.00000154, -0.00000366, -0.00000943, -0.00001445, 0.00000266, -0.00003266, -0.00000650, -0.00002570, -0.00012131, 0.00009296, 0.00014564, 0.00017538, 0.00001417, -0.00012872, -0.00000385, -0.00003219, -0.00001138, -0.00003218, 0.00000803, -0.00000457, 0.00002946, 0.00000525, 0.00002207, 0.00001919, -0.00001362, 0.00001541, -0.00001805, -0.00001612, -0.00000332, -0.00001050, 0.00002084, -0.00001091, -0.00000974, 0.00000484, -0.00001787, 0.00002647, -0.00002472, 0.00000542, -0.00001643, 0.00002513, -0.00000427, -0.00000439, -0.00000220, 0.00000629, -0.00000245, -0.00001123, 0.00001068, -0.00000751, 0.00001230, -0.00002046, 0.00002349, -0.00001605, 0.00001153, -0.00000579, 0.00000509, -0.00001134, 0.00000122, 0.00000917, -0.00000758, 0.00000779, -0.00000536, 0.00000278, -0.00001065, 0.00001119, -0.00000782, 0.00000599, -0.00000235, 0.00000132, 0.00000018, -0.00000668, 0.00000254, 0.00000358, -0.00000259, 0.00000470, 0.00000186, -0.00000865, -0.00000005, 0.00000261, 0.00000241, -0.00000233, -0.00000015, 0.00000085, -0.00000156, 0.00000133, 0.00000127, 0.00000155, -0.00000719, -0.00000489, 0.00000707, -0.00000156, 0.00000310, 0.00000498, -0.00000545, -0.00000096, -0.00000221, 0.00000461, 0.00000402, -0.00000667, 0.00000169, -0.00000456, -0.00000250, -0.00000126, 0.00000222, 0.00000236, 0.00000256, 0.00000317, -0.00000307, -0.00000873, 0.00000139, -0.00000276, -0.00000017, 0.00001055, 0.00000410, -0.00000399, -0.00000664, 0.00000344, -0.00000477, 0.00000462, 0.00001191, -0.00000574, },
		{ -0.00000451, 0.00000549, 0.00000637, -0.00002061, -0.00000651, 0.00000273, -0.00000550, -0.00000189, -0.00000311, -0.00001032, -0.00001383, 0.00000273, -0.00003580, 0.00000027, -0.00004143, -0.00010717, 0.00011466, 0.00013179, 0.00016759, 0.00000020, -0.00012077, -0.00000194, -0.00003314, -0.00001073, -0.00003315, 0.00000921, -0.00000316, 0.00003192, 0.00000306, 0.00002715, 0.00001561, -0.00001175, 0.00001485, -0.00002362, -0.00001043, -0.00000681, -0.00000888, 0.00002209, -0.00001527, -0.00000481, -0.00000049, -0.00001194, 0.00002229, -0.00002389, 0.00000446, -0.00001461, 0.00002722, -0.00000940, -0.00000113, -0.00000408, 0.00000929, -0.00000620, -0.00000782, 0.00000783, -0.00000443, 0.00000882, -0.00001737, 0.00002256, -0.00001701, 0.00001328, -0.00000766, 0.00000637, -0.00001434, 0.00000657, 0.00000484, -0.00000502, 0.00000643, -0.00000481, 0.00000193, -0.00001014, 0.00001153, -0.00000867, 0.00000725, -0.00000352, 0.00000227, -0.00000112, -0.00000607, 0.00000337, 0.00000327, -0.00000303, 0.00000599, -0.00000022, -0.00000857, 0.00000119, 0.00000238, 0.00000228, -0.00000260, 0.00000020, 0.00000075, -0.00000175, 0.00000182, 0.00000137, 0.00000089, -0.00000805, -0.00000342, 0.00000715, -0.00000251, 0.00000452, 0.00000397, -0.00000587, -0.00000038, -0.00000250, 0.00000575, 0.00000276, -0.00000661, 0.00000200, -0.00000514, -0.00000208, -0.00000100, 0.00000237, 0.00000220, 0.00000305, 0.00000274, -0.00000408, -0.00000823, 0.00000217, -0.00000376, 0.00000140, 0.00001027, 0.00000324, -0.00000428, -0.00000635, 0.00000366, -0.00000546, 0.00000695, 0.00001026, -0.00000656, 0.00000148, },
		{ 0.00000780, 0.00000391, -0.00002157, -0.00000381, 0.00000158, -0.00000524, -0.00000233, -0.00000229, -0.00001171, -0.00001243, 0.00000157, -0.00003731, 0.00000537, -0.00005717, -0.00008795, 0.00013050, 0.00011762, 0.00016138, -0.00001365, -0.00011171, -0.00000145, -0.00003324, -0.00001072, -0.00003375, 0.00001022, -0.00000127, 0.00003383, 0.00000128, 0.00003237, 0.00001127, -0.00000875, 0.00001233, -0.00002758, -0.00000489, -0.00001131, -0.00000559, 0.00002149, -0.00001803, -0.00000105, -0.00000504, -0.00000613, 0.00001745, -0.00002218, 0.00000273, -0.00001144, 0.00002752, -0.00001337, 0.00000170, -0.00000553, 0.00001183, -0.00000968, -0.00000454, 0.00000507, -0.00000135, 0.00000507, -0.00001354, 0.00002054, -0.00001689, 0.00001411, -0.00000877, 0.00000675, -0.00001628, 0.00001136, 0.00000028, -0.00000185, 0.00000438, -0.00000369, 0.00000058, -0.00000911, 0.00001140, -0.00000914, 0.00000829, -0.00000462, 0.00000316, -0.00000247, -0.00000523, 0.00000414, 0.00000273, -0.00000318, 0.00000709, -0.00000252, -0.00000796, 0.00000218, 0.00000212, 0.00000213, -0.00000285, 0.00000060, 0.00000056, -0.00000187, 0.00000236, 0.00000135, 0.00000013, -0.00000873, -0.00000177, 0.00000684, -0.00000319, 0.00000595, 0.00000263, -0.00000595, 0.00000000, -0.00000261, 0.00000680, 0.00000134, -0.00000628, 0.00000210, -0.00000563, -0.00000164, -0.00000073, 0.00000242, 0.00000213, 0.00000355, 0.00000211, -0.00000500, -0.00000749, 0.00000270, -0.00000463, 0.00000309, 0.00000965, 0.00000261, -0.00000465, -0.00000590, 0.00000363, -0.00000584, 0.00000923, 0.00000811, -0.00000674, 0.00000209, -0.00000550, },
		{ 0.00000088, -0.00002168, -0.00000150, 0.00000046, -0.00000503, -0.00000269, -0.00000140, -0.00001340, -0.00001043, -0.00000073, -0.00003703, 0.00000817, -0.00007165, -0.00006481, 0.00014021, 0.00010433, 0.00015598, -0.00002702, -0.00010185, -0.00000222, -0.00003247, -0.00001141, -0.00003387, 0.00001095, 0.00000122, 0.00003500, 0.00000020, 0.00003737, 0.00000652, -0.00000494, 0.00000793, -0.00002949, -0.00000022, -0.00001615, -0.00000112, 0.00001927, -0.00001919, 0.00000134, -0.00000847, -0.00000089, 0.00001239, -0.00001998, 0.00000056, -0.00000726, 0.00002614, -0.00001605, 0.00000394, -0.00000644, 0.00001378, -0.00001277, -0.00000151, 0.00000256, 0.00000152, 0.00000135, -0.00000929, 0.00001765, -0.00001579, 0.00001407, -0.00000915, 0.00000621, -0.00001698, 0.00001525, -0.00000413, 0.00000166, 0.00000181, -0.00000207, -0.00000123, -0.00000755, 0.00001075, -0.00000917, 0.00000901, -0.00000555, 0.00000395, -0.00000381, -0.00000418, 0.00000478, 0.00000201, -0.00000300, 0.00000790, -0.00000493, -0.00000686, 0.00000288, 0.00000189, 0.00000195, -0.00000306, 0.00000103, 0.00000027, -0.00000188, 0.00000293, 0.00000119, -0.00000069, -0.00000920, -0.00000002, 0.00000615, -0.00000351, 0.00000727, 0.00000106, -0.00000569, 0.00000017, -0.00000251, 0.00000767, -0.00000014, -0.00000572, 0.00000200, -0.00000600, -0.00000120, -0.00000045, 0.00000237, 0.00000219, 0.00000400, 0.00000129, -0.00000579, -0.00000656, 0.00000295, -0.00000530, 0.00000479, 0.00000873, 0.00000219, -0.00000510, -0.00000529, 0.00000332, -0.00000580, 0.00001127, 0.00000564, -0.00000635, 0.00000217, -0.00000597, 0.00000995, },
		{ -0.00002100, 0.00000034, -0.00000054, -0.00000497, -0.00000282, -0.00000065, -0.00001514, -0.00000809, -0.00000390, -0.00003504, 0.00000828, -0.00008373, -0.00003911, 0.00014396, 0.00009283, 0.00015068, -0.00003951, -0.00009158, -0.00000402, -0.00003093, -0.00001281, -0.00003348, 0.00001139, 0.00000435, 0.00003531, 0.00000007, 0.00004179, 0.00000176, -0.00000078, 0.00000191, -0.00002918, 0.00000301, -0.00002063, 0.00000396, 0.00001580, -0.00001890, 0.00000234, -0.00001057, 0.00000340, 0.00000752, -0.00001763, -0.00000166, -0.00000245, 0.00002330, -0.00001742, 0.00000546, -0.00000673, 0.00001505, -0.00001534, 0.00000114, 0.00000043, 0.00000398, -0.00000209, -0.00000491, 0.00001416, -0.00001391, 0.00001325, -0.00000886, 0.00000483, -0.00001637, 0.00001796, -0.00000803, 0.00000520, -0.00000108, -0.00000010, -0.00000342, -0.00000551, 0.00000959, -0.00000871, 0.00000935, -0.00000626, 0.00000455, -0.00000507, -0.00000295, 0.00000526, 0.00000115, -0.00000249, 0.00000830, -0.00000728, -0.00000535, 0.00000325, 0.00000173, 0.00000171, -0.00000319, 0.00000144, -0.00000009, -0.00000178, 0.00000350, 0.00000089, -0.00000156, -0.00000942, 0.00000174, 0.00000510, -0.00000341, 0.00000834, -0.00000063, -0.00000515, 0.00000012, -0.00000216, 0.00000828, -0.00000158, -0.00000502, 0.00000172, -0.00000625, -0.00000076, -0.00000016, 0.00000221, 0.00000242, 0.00000434, 0.00000032, -0.00000639, -0.00000550, 0.00000293, -0.00000572, 0.00000641, 0.00000760, 0.00000198, -0.00000564, -0.00000456, 0.00000276, -0.00000531, 0.00001290, 0.00000303, -0.00000551, 0.00000176, -0.00000581, 0.00001168, -0.00000248, },
		{ 0.00000174, -0.00000141, -0.00000509, -0.00000264, -0.00000019, -0.00001670, -0.00000572, -0.00000762, -0.00003160, 0.00000560, -0.00009242, -0.00001234, 0.00014231, 0.00008370, 0.00014491, -0.00005073, -0.00008131, -0.00000654, -0.00002880, -0.00001482, -0.00003261, 0.00001157, 0.00000811, 0.00003470, 0.00000108, 0.00004532, -0.00000263, 0.00000323, -0.00000524, -0.00002669, 0.00000443, -0.00002416, 0.00000906, 0.00001155, -0.00001746, 0.00000208, -0.00001130, 0.00000652, 0.00000317, -0.00001544, -0.00000359, 0.00000256, 0.00001931, -0.00001754, 0.00000626, -0.00000636, 0.00001557, -0.00001733, 0.00000334, -0.00000125, 0.00000592, -0.00000503, -0.00000069, 0.00001036, -0.00001145, 0.00001181, -0.00000806, 0.00000275, -0.00001451, 0.00001933, -0.00001112, 0.00000845, -0.00000404, 0.00000207, -0.00000586, -0.00000308, 0.00000796, -0.00000775, 0.00000926, -0.00000668, 0.00000490, -0.00000617, -0.00000157, 0.00000555, 0.00000021, -0.00000166, 0.00000822, -0.00000943, -0.00000353, 0.00000331, 0.00000166, 0.00000140, -0.00000322, 0.00000182, -0.00000052, -0.00000156, 0.00000403, 0.00000044, -0.00000244, -0.00000936, 0.00000344, 0.00000376, -0.00000285, 0.00000905, -0.00000231, -0.00000438, -0.00000013, -0.00000156, 0.00000859, -0.00000291, -0.00000423, 0.00000131, -0.00000637, -0.00000031, 0.00000012, 0.00000196, 0.00000282, 0.00000449, -0.00000075, -0.00000678, -0.00000437, 0.00000266, -0.00000585, 0.00000787, 0.00000632, 0.00000195, -0.00000623, -0.00000373, 0.00000199, -0.00000434, 0.00001400, 0.00000048, -0.00000438, 0.00000094, -0.00000499, 0.00001279, -0.00000589, -0.00001969, },
		{ -0.00000218, -0.00000536, -0.00000210, -0.00000015, -0.00001789, -0.00000362, -0.00001147, -0.00002716, 0.00000029, -0.00009705, 0.00001403, 0.00013612, 0.00007717, 0.00013824, -0.00006036, -0.00007142, -0.00000944, -0.00002630, -0.00001728, -0.00003138, 0.00001163, 0.00001236, 0.00003323, 0.00000332, 0.00004773, -0.00000633, 0.00000654, -0.00001292, -0.00002232, 0.00000387, -0.00002630, 0.00001361, 0.00000701, -0.00001526, 0.00000083, -0.00001076, 0.00000836, -0.00000044, -0.00001365, -0.00000491, 0.00000732, 0.00001454, -0.00001658, 0.00000637, -0.00000538, 0.00001535, -0.00001869, 0.00000508, -0.00000244, 0.00000727, -0.00000732, 0.00000312, 0.00000651, -0.00000865, 0.00000995, -0.00000692, 0.00000019, -0.00001153, 0.00001931, -0.00001317, 0.00001114, -0.00000682, 0.00000423, -0.00000838, -0.00000038, 0.00000592, -0.00000632, 0.00000871, -0.00000676, 0.00000497, -0.00000704, -0.00000011, 0.00000560, -0.00000072, -0.00000056, 0.00000762, -0.00001123, -0.00000154, 0.00000308, 0.00000170, 0.00000101, -0.00000314, 0.00000214, -0.00000097, -0.00000121, 0.00000450, -0.00000014, -0.00000332, -0.00000901, 0.00000496, 0.00000223, -0.00000185, 0.00000930, -0.00000386, -0.00000350, -0.00000050, -0.00000075, 0.00000858, -0.00000405, -0.00000343, 0.00000081, -0.00000640, 0.00000016, 0.00000034, 0.00000166, 0.00000341, 0.00000441, -0.00000186, -0.00000694, -0.00000322, 0.00000217, -0.00000567, 0.00000910, 0.00000499, 0.00000204, -0.00000686, -0.00000285, 0.00000106, -0.00000291, 0.00001448, -0.00000181, -0.00000315, -0.00000013, -0.00000357, 0.00001312, -0.00000910, -0.00001791, 0.00000274, },
		{ -0.00000572, -0.00000125, -0.00000055, -0.00001860, -0.00000201, -0.00001503, -0.00002226, -0.00000722, -0.00009724, 0.00003866, 0.00012646, 0.00007320, 0.00013044, -0.00006818, -0.00006224, -0.00001243, -0.00002368, -0.00002003, -0.00002993, 0.00001173, 0.00001693, 0.00003104, 0.00000674, 0.00004890, -0.00000912, 0.00000871, -0.00002046, -0.00001657, 0.00000144, -0.00002679, 0.00001717, 0.00000267, -0.00001273, -0.00000106, -0.00000917, 0.00000895, -0.00000319, -0.00001240, -0.00000541, 0.00001145, 0.00000941, -0.00001476, 0.00000589, -0.00000385, 0.00001442, -0.00001943, 0.00000634, -0.00000316, 0.00000799, -0.00000885, 0.00000631, 0.00000287, -0.00000575, 0.00000785, -0.00000564, -0.00000259, -0.00000766, 0.00001794, -0.00001405, 0.00001305, -0.00000920, 0.00000618, -0.00001081, 0.00000243, 0.00000360, -0.00000449, 0.00000772, -0.00000650, 0.00000471, -0.00000762, 0.00000137, 0.00000543, -0.00000158, 0.00000074, 0.00000650, -0.00001257, 0.00000051, 0.00000262, 0.00000182, 0.00000055, -0.00000296, 0.00000238, -0.00000144, -0.00000073, 0.00000487, -0.00000084, -0.00000413, -0.00000837, 0.00000624, 0.00000061, -0.00000047, 0.00000906, -0.00000515, -0.00000261, -0.00000093, 0.00000021, 0.00000826, -0.00000497, -0.00000269, 0.00000026, -0.00000636, 0.00000067, 0.00000048, 0.00000135, 0.00000416, 0.00000404, -0.00000293, -0.00000688, -0.00000212, 0.00000150, -0.00000518, 0.00001002, 0.00000368, 0.00000221, -0.00000748, -0.00000198, 0.00000007, -0.00000112, 0.00001429, -0.00000370, -0.00000199, -0.00000125, -0.00000168, 0.00001265, -0.00001191, -0.00001585, 0.00000341, -0.00000289, },
		{ -0.00000019, -0.00000136, -0.00001878, -0.00000108, -0.00001796, -0.00001748, -0.00001628, -0.00009299, 0.00006046, 0.00011443, 0.00007149, 0.00012148, -0.00007410, -0.00005401, -0.00001524, -0.00002114, -0.00002286, -0.00002841, 0.00001204, 0.00002160, 0.00002835, 0.00001123, 0.00004883, -0.00001091, 0.00000937, -0.00002718, -0.00001003, -0.00000260, -0.00002557, 0.00001945, -0.00000106, -0.00001027, -0.00000321, -0.00000685, 0.00000845, -0.00000510, -0.00001172, -0.00000497, 0.00001463, 0.00000430, -0.00001236, 0.00000498, -0.00000190, 0.00001287, -0.00001958, 0.00000717, -0.00000345, 0.00000815, -0.00000961, 0.00000877, -0.00000036, -0.00000296, 0.00000570, -0.00000443, -0.00000530, -0.00000323, 0.00001539, -0.00001373, 0.00001403, -0.00001099, 0.00000774, -0.00001296, 0.00000517, 0.00000116, -0.00000236, 0.00000634, -0.00000590, 0.00000410, -0.00000783, 0.00000278, 0.00000504, -0.00000229, 0.00000213, 0.00000489, -0.00001335, 0.00000248, 0.00000199, 0.00000201, 0.00000004, -0.00000267, 0.00000253, -0.00000189, -0.00000013, 0.00000513, -0.00000166, -0.00000486, -0.00000745, 0.00000718, -0.00000097, 0.00000120, 0.00000831, -0.00000610, -0.00000180, -0.00000133, 0.00000124, 0.00000767, -0.00000564, -0.00000205, -0.00000027, -0.00000628, 0.00000124, 0.00000048, 0.00000111, 0.00000501, 0.00000338, -0.00000390, -0.00000662, -0.00000111, 0.00000072, -0.00000440, 0.00001061, 0.00000246, 0.00000242, -0.00000806, -0.00000118, -0.00000088, 0.00000094, 0.00001346, -0.00000509, -0.00000107, -0.00000222, 0.00000048, 0.00001139, -0.00001416, -0.00001367, 0.00000386, -0.00000363, -0.00000606, },
		{ -0.00000248, -0.00001848, -0.00000086, -0.00001998, -0.00001340, -0.00002611, -0.00008461, 0.00007859, 0.00010117, 0.00007162, 0.00011146, -0.00007815, -0.00004687, -0.00001767, -0.00001888, -0.00002563, -0.00002697, 0.00001278, 0.00002610, 0.00002544, 0.00001659, 0.00004761, -0.00001176, 0.00000837, -0.00003253, -0.00000339, -0.00000778, -0.00002278, 0.00002031, -0.00000391, -0.00000824, -0.00000524, -0.00000416, 0.00000709, -0.00000626, -0.00001155, -0.00000358, 0.00001665, -0.00000041, -0.00000966, 0.00000381, 0.00000032, 0.00001080, -0.00001920, 0.00000762, -0.00000339, 0.00000782, -0.00000963, 0.00001041, -0.00000304, -0.00000044, 0.00000366, -0.00000346, -0.00000768, 0.00000144, 0.00001190, -0.00001228, 0.00001401, -0.00001205, 0.00000877, -0.00001467, 0.00000767, -0.00000126, -0.00000006, 0.00000465, -0.00000500, 0.00000317, -0.00000764, 0.00000405, 0.00000446, -0.00000279, 0.00000351, 0.00000287, -0.00001355, 0.00000426, 0.00000127, 0.00000224, -0.00000050, -0.00000230, 0.00000257, -0.00000230, 0.00000057, 0.00000523, -0.00000257, -0.00000544, -0.00000629, 0.00000775, -0.00000239, 0.00000306, 0.00000709, -0.00000663, -0.00000115, -0.00000161, 0.00000224, 0.00000690, -0.00000609, -0.00000152, -0.00000078, -0.00000616, 0.00000189, 0.00000030, 0.00000101, 0.00000589, 0.00000242, -0.00000471, -0.00000619, -0.00000022, -0.00000013, -0.00000337, 0.00001084, 0.00000140, 0.00000260, -0.00000857, -0.00000051, -0.00000169, 0.00000312, 0.00001207, -0.00000595, -0.00000051, -0.00000284, 0.00000267, 0.00000948, -0.00001579, -0.00001151, 0.00000415, -0.00000446, -0.00000626, 0.00000097, },
		{ -0.00001783, -0.00000130, -0.00002097, -0.00001048, -0.00003587, -0.00007269, 0.00009258, 0.00008767, 0.00007304, 0.00010061, -0.00008047, -0.00004085, -0.00001962, -0.00001701, -0.00002823, -0.00002570, 0.00001408, 0.00003022, 0.00002260, 0.00002253, 0.00004542, -0.00001186, 0.00000569, -0.00003606, 0.00000271, -0.00001355, -0.00001870, 0.00001978, -0.00000573, -0.00000689, -0.00000682, -0.00000147, 0.00000516, -0.00000683, -0.00001174, -0.00000137, 0.00001740, -0.00000444, -0.00000691, 0.00000257, 0.00000265, 0.00000835, -0.00001837, 0.00000775, -0.00000307, 0.00000711, -0.00000899, 0.00001124, -0.00000507, 0.00000168, 0.00000184, -0.00000284, -0.00000948, 0.00000599, 0.00000778, -0.00000986, 0.00001301, -0.00001231, 0.00000915, -0.00001582, 0.00000977, -0.00000348, 0.00000226, 0.00000276, -0.00000387, 0.00000195, -0.00000703, 0.00000511, 0.00000373, -0.00000305, 0.00000476, 0.00000054, -0.00001313, 0.00000575, 0.00000052, 0.00000248, -0.00000102, -0.00000187, 0.00000253, -0.00000266, 0.00000138, 0.00000515, -0.00000355, -0.00000586, -0.00000495, 0.00000790, -0.00000352, 0.00000494, 0.00000546, -0.00000675, -0.00000074, -0.00000169, 0.00000312, 0.00000603, -0.00000636, -0.00000109, -0.00000126, -0.00000602, 0.00000260, -0.00000008, 0.00000113, 0.00000672, 0.00000121, -0.00000530, -0.00000563, 0.00000051, -0.00000098, -0.00000214, 0.00001071, 0.00000054, 0.00000273, -0.00000899, 0.00000001, -0.00000226, 0.00000527, 0.00001023, -0.00000628, -0.00000034, -0.00000297, 0.00000465, 0.00000708, -0.00001678, -0.00000943, 0.00000432, -0.00000543, -0.00000621, 0.00000207, -0.00000377, },
		{ -0.00000225, -0.00002094, -0.00000906, -0.00004472, -0.00005805, 0.00010224, 0.00007477, 0.00007520, 0.00008923, -0.00008126, -0.00003592, -0.00002105, -0.00001560, -0.00003058, -0.00002467, 0.00001607, 0.00003374, 0.00002011, 0.00002878, 0.00004248, -0.00001150, 0.00000154, -0.00003753, 0.00000770, -0.00001934, -0.00001377, 0.00001805, -0.00000650, -0.00000634, -0.00000770, 0.00000090, 0.00000295, -0.00000703, -0.00001208, 0.00000147, 0.00001690, -0.00000760, -0.00000436, 0.00000144, 0.00000490, 0.00000567, -0.00001718, 0.00000764, -0.00000260, 0.00000618, -0.00000784, 0.00001132, -0.00000643, 0.00000331, 0.00000032, -0.00000266, -0.00001052, 0.00001009, 0.00000338, -0.00000671, 0.00001112, -0.00001178, 0.00000885, -0.00001630, 0.00001135, -0.00000535, 0.00000443, 0.00000079, -0.00000261, 0.00000052, -0.00000601, 0.00000588, 0.00000291, -0.00000305, 0.00000580, -0.00000197, -0.00001214, 0.00000688, -0.00000018, 0.00000268, -0.00000150, -0.00000141, 0.00000240, -0.00000295, 0.00000225, 0.00000487, -0.00000455, -0.00000606, -0.00000349, 0.00000763, -0.00000428, 0.00000672, 0.00000355, -0.00000647, -0.00000059, -0.00000154, 0.00000379, 0.00000516, -0.00000652, -0.00000075, -0.00000171, -0.00000582, 0.00000336, -0.00000068, 0.00000152, 0.00000738, -0.00000020, -0.00000565, -0.00000498, 0.00000106, -0.00000177, -0.00000078, 0.00001025, -0.00000008, 0.00000277, -0.00000930, 0.00000035, -0.00000251, 0.00000722, 0.00000811, -0.00000618, -0.00000057, -0.00000251, 0.00000617, 0.00000443, -0.00001722, -0.00000748, 0.00000438, -0.00000654, -0.00000586, 0.00000299, -0.00000508, -0.00001700, },
		{ -0.00002005, -0.00000926, -0.00005195, -0.00004162, 0.00010772, 0.00006310, 0.00007759, 0.00007764, -0.00008080, -0.00003196, -0.00002198, -0.00001465, -0.00003267, -0.00002386, 0.00001880, 0.00003651, 0.00001822, 0.00003505, 0.00003904, -0.00001104, -0.00000374, -0.00003693, 0.00001117, -0.00002459, -0.00000844, 0.00001541, -0.00000634, -0.00000662, -0.00000774, 0.00000270, 0.00000074, -0.00000706, -0.00001234, 0.00000467, 0.00001529, -0.00000979, -0.00000218, 0.00000057, 0.00000691, 0.00000291, -0.00001574, 0.00000736, -0.00000207, 0.00000515, -0.00000632, 0.00001076, -0.00000714, 0.00000445, -0.00000087, -0.00000292, -0.00001069, 0.00001344, -0.00000095, -0.00000310, 0.00000853, -0.00001054, 0.00000787, -0.00001609, 0.00001231, -0.00000674, 0.00000632, -0.00000113, -0.00000131, -0.00000104, -0.00000462, 0.00000632, 0.00000207, -0.00000280, 0.00000652, -0.00000453, -0.00001063, 0.00000761, -0.00000080, 0.00000283, -0.00000192, -0.00000095, 0.00000221, -0.00000316, 0.00000319, 0.00000437, -0.00000554, -0.00000605, -0.00000199, 0.00000699, -0.00000459, 0.00000823, 0.00000150, -0.00000584, -0.00000069, -0.00000111, 0.00000419, 0.00000437, -0.00000664, -0.00000043, -0.00000217, -0.00000554, 0.00000411, -0.00000146, 0.00000222, 0.00000778, -0.00000172, -0.00000575, -0.00000427, 0.00000142, -0.00000244, 0.00000064, 0.00000949, -0.00000046, 0.00000271, -0.00000952, 0.00000053, -0.00000240, 0.00000885, 0.00000588, -0.00000577, -0.00000111, -0.00000144, 0.00000707, 0.00000175, -0.00001718, -0.00000564, 0.00000430, -0.00000774, -0.00000517, 0.00000363, -0.00000626, -0.00001618, -0.00000348, },
		{ -0.00001104, -0.00005700, -0.00002438, 0.00010937, 0.00005306, 0.00007979, 0.00006617, -0.00007935, -0.00002882, -0.00002249, -0.00001411, -0.00003455, -0.00002324, 0.00002226, 0.00003847, 0.00001713, 0.00004111, 0.00003528, -0.00001084, -0.00000965, -0.00003442, 0.00001288, -0.00002885, -0.00000320, 0.00001222, -0.00000548, -0.00000762, -0.00000692, 0.00000374, -0.00000125, -0.00000710, -0.00001229, 0.00000790, 0.00001278, -0.00001100, -0.00000045, 0.00000006, 0.00000856, 0.00000020, -0.00001415, 0.00000698, -0.00000155, 0.00000415, -0.00000462, 0.00000970, -0.00000728, 0.00000511, -0.00000177, -0.00000358, -0.00000996, 0.00001585, -0.00000489, 0.00000066, 0.00000544, -0.00000870, 0.00000630, -0.00001520, 0.00001263, -0.00000757, 0.00000780, -0.00000286, -0.00000010, -0.00000261, -0.00000293, 0.00000640, 0.00000129, -0.00000231, 0.00000686, -0.00000699, -0.00000871, 0.00000793, -0.00000127, 0.00000289, -0.00000225, -0.00000053, 0.00000197, -0.00000329, 0.00000417, 0.00000364, -0.00000646, -0.00000579, -0.00000053, 0.00000602, -0.00000441, 0.00000936, -0.00000056, -0.00000497, -0.00000099, -0.00000044, 0.00000428, 0.00000374, -0.00000679, -0.00000009, -0.00000268, -0.00000512, 0.00000479, -0.00000238, 0.00000325, 0.00000783, -0.00000325, -0.00000561, -0.00000355, 0.00000160, -0.00000294, 0.00000205, 0.00000849, -0.00000060, 0.00000254, -0.00000964, 0.00000056, -0.00000193, 0.00001003, 0.00000371, -0.00000520, -0.00000182, 0.00000017, 0.00000722, -0.00000074, -0.00001682, -0.00000387, 0.00000402, -0.00000896, -0.00000419, 0.00000395, -0.00000721, -0.00001555, -0.00000472, -0.00001857, },
		{ -0.00005954, -0.00000725, 0.00010774, 0.00004487, 0.00008146, 0.00005508, -0.00007719, -0.00002634, -0.00002268, -0.00001392, -0.00003627, -0.00002269, 0.00002637, 0.00003959, 0.00001697, 0.00004675, 0.00003136, -0.00001120, -0.00001565, -0.00003036, 0.00001278, -0.00003179, 0.00000153, 0.00000887, -0.00000419, -0.00000913, -0.00000533, 0.00000396, -0.00000288, -0.00000730, -0.00001176, 0.00001086, 0.00000965, -0.00001132, 0.00000078, -0.00000003, 0.00000975, -0.00000232, -0.00001249, 0.00000655, -0.00000112, 0.00000331, -0.00000288, 0.00000828, -0.00000695, 0.00000534, -0.00000243, -0.00000453, -0.00000839, 0.00001719, -0.00000818, 0.00000427, 0.00000210, -0.00000645, 0.00000424, -0.00001372, 0.00001232, -0.00000780, 0.00000877, -0.00000429, 0.00000092, -0.00000406, -0.00000103, 0.00000612, 0.00000063, -0.00000166, 0.00000679, -0.00000923, -0.00000649, 0.00000786, -0.00000160, 0.00000287, -0.00000247, -0.00000015, 0.00000171, -0.00000333, 0.00000515, 0.00000265, -0.00000726, -0.00000531, 0.00000081, 0.00000481, -0.00000376, 0.00000999, -0.00000247, -0.00000396, -0.00000143, 0.00000043, 0.00000407, 0.00000332, -0.00000704, 0.00000033, -0.00000328, -0.00000451, 0.00000531, -0.00000336, 0.00000458, 0.00000746, -0.00000469, -0.00000526, -0.00000285, 0.00000161, -0.00000325, 0.00000335, 0.00000732, -0.00000053, 0.00000226, -0.00000971, 0.00000050, -0.00000111, 0.00001070, 0.00000174, -0.00000462, -0.00000254, 0.00000220, 0.00000659, -0.00000287, -0.00001625, -0.00000214, 0.00000346, -0.00001007, -0.00000299, 0.00000398, -0.00000789, -0.00001525, -0.00000571, -0.00001687, -0.00001416, },
		{ 0.00000896, 0.00010347, 0.00003855, 0.00008240, 0.00004459, -0.00007455, -0.00002436, -0.00002265, -0.00001398, -0.00003794, -0.00002209, 0.00003103, 0.00003993, 0.00001780, 0.00005185, 0.00002732, -0.00001237, -0.00002122, -0.00002522, 0.00001103, -0.00003325, 0.00000541, 0.00000570, -0.00000279, -0.00001090, -0.00000314, 0.00000340, -0.00000405, -0.00000770, -0.00001063, 0.00001326, 0.00000623, -0.00001091, 0.00000154, 0.00000030, 0.00001041, -0.00000456, -0.00001083, 0.00000612, -0.00000080, 0.00000271, -0.00000128, 0.00000668, -0.00000627, 0.00000521, -0.00000292, -0.00000564, -0.00000609, 0.00001742, -0.00001060, 0.00000746, -0.00000124, -0.00000398, 0.00000187, -0.00001176, 0.00001144, -0.00000743, 0.00000919, -0.00000534, 0.00000164, -0.00000528, 0.00000096, 0.00000551, 0.00000014, -0.00000090, 0.00000630, -0.00001111, -0.00000408, 0.00000745, -0.00000177, 0.00000276, -0.00000260, 0.00000016, 0.00000144, -0.00000328, 0.00000612, 0.00000143, -0.00000788, -0.00000461, 0.00000195, 0.00000345, -0.00000267, 0.00001007, -0.00000408, -0.00000295, -0.00000191, 0.00000142, 0.00000359, 0.00000311, -0.00000741, 0.00000087, -0.00000402, -0.00000365, 0.00000558, -0.00000427, 0.00000614, 0.00000664, -0.00000596, -0.00000473, -0.00000219, 0.00000147, -0.00000333, 0.00000450, 0.00000606, -0.00000027, 0.00000188, -0.00000974, 0.00000041, -0.00000001, 0.00001085, 0.00000008, -0.00000417, -0.00000306, 0.00000442, 0.00000523, -0.00000452, -0.00001559, -0.00000038, 0.00000254, -0.00001094, -0.00000170, 0.00000378, -0.00000833, -0.00001533, -0.00000623, -0.00001532, -0.00001823, -0.00005944, },
		{ 0.00009725, 0.00003399, 0.00008249, 0.00003483, -0.00007161, -0.00002275, -0.00002250, -0.00001421, -0.00003964, -0.00002131, 0.00003610, 0.00003959, 0.00001963, 0.00005632, 0.00002314, -0.00001442, -0.00002585, -0.00001956, 0.00000792, -0.00003323, 0.00000824, 0.00000300, -0.00000157, -0.00001266, -0.00000059, 0.00000215, -0.00000479, -0.00000831, -0.00000885, 0.00001488, 0.00000281, -0.00000996, 0.00000194, 0.00000099, 0.00001055, -0.00000645, -0.00000925, 0.00000571, -0.00000060, 0.00000238, 0.00000009, 0.00000504, -0.00000536, 0.00000482, -0.00000332, -0.00000674, -0.00000327, 0.00001661, -0.00001206, 0.00001004, -0.00000436, -0.00000149, -0.00000065, -0.00000947, 0.00001011, -0.00000652, 0.00000905, -0.00000596, 0.00000200, -0.00000615, 0.00000291, 0.00000463, -0.00000012, -0.00000010, 0.00000540, -0.00001255, -0.00000164, 0.00000674, -0.00000180, 0.00000257, -0.00000263, 0.00000040, 0.00000117, -0.00000312, 0.00000702, -0.00000002, -0.00000827, -0.00000375, 0.00000284, 0.00000208, -0.00000123, 0.00000957, -0.00000530, -0.00000205, -0.00000231, 0.00000242, 0.00000293, 0.00000310, -0.00000792, 0.00000155, -0.00000492, -0.00000250, 0.00000551, -0.00000499, 0.00000781, 0.00000539, -0.00000696, -0.00000408, -0.00000160, 0.00000120, -0.00000317, 0.00000542, 0.00000477, 0.00000014, 0.00000143, -0.00000977, 0.00000037, 0.00000127, 0.00001049, -0.00000120, -0.00000397, -0.00000322, 0.00000663, 0.00000326, -0.00000565, -0.00001492, 0.00000141, 0.00000123, -0.00001145, -0.00000048, 0.00000348, -0.00000861, -0.00001577, -0.00000615, -0.00001427, -0.00002276, -0.00005684, 0.00002360, },
		{ 0.00003100, 0.00008172, 0.00002588, -0.00006852, -0.00002141, -0.00002230, -0.00001456, -0.00004145, -0.00002020, 0.00004140, 0.00003869, 0.00002244, 0.00006012, 0.00001870, -0.00001732, -0.00002920, -0.00001394, 0.00000387, -0.00003188, 0.00000995, 0.00000096, -0.00000076, -0.00001413, 0.00000207, 0.00000039, -0.00000516, -0.00000906, -0.00000647, 0.00001559, -0.00000032, -0.00000870, 0.00000209, 0.00000194, 0.00001018, -0.00000796, -0.00000778, 0.00000532, -0.00000052, 0.00000236, 0.00000113, 0.00000349, -0.00000435, 0.00000424, -0.00000372, -0.00000767, -0.00000014, 0.00001488, -0.00001252, 0.00001184, -0.00000705, 0.00000082, -0.00000313, -0.00000703, 0.00000848, -0.00000518, 0.00000838, -0.00000615, 0.00000196, -0.00000658, 0.00000470, 0.00000354, -0.00000016, 0.00000065, 0.00000416, -0.00001350, 0.00000074, 0.00000582, -0.00000171, 0.00000233, -0.00000258, 0.00000058, 0.00000091, -0.00000286, 0.00000782, -0.00000166, -0.00000838, -0.00000277, 0.00000344, 0.00000078, 0.00000045, 0.00000852, -0.00000604, -0.00000136, -0.00000255, 0.00000332, 0.00000219, 0.00000322, -0.00000854, 0.00000236, -0.00000597, -0.00000103, 0.00000501, -0.00000539, 0.00000946, 0.00000376, -0.00000766, -0.00000335, -0.00000110, 0.00000083, -0.00000279, 0.00000607, 0.00000353, 0.00000065, 0.00000091, -0.00000982, 0.00000045, 0.00000261, 0.00000970, -0.00000209, -0.00000407, -0.00000287, 0.00000855, 0.00000087, -0.00000627, -0.00001426, 0.00000322, -0.00000050, -0.00001149, 0.00000054, 0.00000324, -0.00000891, -0.00001643, -0.00000546, -0.00001402, -0.00002722, -0.00005204, 0.00003624, 0.00008973, },
		{ 0.00008010, 0.00001780, -0.00006537, -0.00002026, -0.00002211, -0.00001496, -0.00004342, -0.00001864, 0.00004681, 0.00003739, 0.00002616, 0.00006320, 0.00001385, -0.00002088, -0.00003104, -0.00000885, -0.00000067, -0.00002947, 0.00001058, -0.00000033, -0.00000050, -0.00001508, 0.00000456, -0.00000169, -0.00000526, -0.00000980, -0.00000363, 0.00001534, -0.00000295, -0.00000734, 0.00000214, 0.00000305, 0.00000940, -0.00000908, -0.00000644, 0.00000495, -0.00000053, 0.00000262, 0.00000179, 0.00000213, -0.00000334, 0.00000355, -0.00000417, -0.00000829, 0.00000304, 0.00001245, -0.00001205, 0.00001280, -0.00000919, 0.00000281, -0.00000540, -0.00000461, 0.00000671, -0.00000354, 0.00000728, -0.00000592, 0.00000151, -0.00000652, 0.00000621, 0.00000234, 0.00000005, 0.00000127, 0.00000265, -0.00001391, 0.00000293, 0.00000475, -0.00000152, 0.00000207, -0.00000249, 0.00000070, 0.00000066, -0.00000248, 0.00000846, -0.00000346, -0.00000818, -0.00000174, 0.00000373, -0.00000032, 0.00000224, 0.00000700, -0.00000627, -0.00000098, -0.00000252, 0.00000400, 0.00000147, 0.00000338, -0.00000920, 0.00000325, -0.00000715, 0.00000074, 0.00000403, -0.00000535, 0.00001093, 0.00000184, -0.00000802, -0.00000257, -0.00000070, 0.00000043, -0.00000221, 0.00000643, 0.00000240, 0.00000121, 0.00000034, -0.00000992, 0.00000072, 0.00000389, 0.00000860, -0.00000264, -0.00000446, -0.00000192, 0.00000997, -0.00000172, -0.00000646, -0.00001359, 0.00000498, -0.00000260, -0.00001101, 0.00000123, 0.00000320, -0.00000939, -0.00001714, -0.00000425, -0.00001472, -0.00003111, -0.00004548, 0.00004664, 0.00008149, 0.00002932, },
		{ 0.00001056, -0.00006224, -0.00001927, -0.00002196, -0.00001540, -0.00004559, -0.00001654, 0.00005216, 0.00003583, 0.00003074, 0.00006552, 0.00000842, -0.00002477, -0.00003135, -0.00000470, -0.00000522, -0.00002634, 0.00001030, -0.00000089, -0.00000087, -0.00001538, 0.00000667, -0.00000388, -0.00000524, -0.00001040, -0.00000053, 0.00001420, -0.00000493, -0.00000605, 0.00000223, 0.00000416, 0.00000829, -0.00000984, -0.00000524, 0.00000459, -0.00000056, 0.00000311, 0.00000207, 0.00000103, -0.00000240, 0.00000281, -0.00000471, -0.00000848, 0.00000604, 0.00000956, -0.00001077, 0.00001293, -0.00001068, 0.00000436, -0.00000734, -0.00000236, 0.00000496, -0.00000176, 0.00000585, -0.00000534, 0.00000070, -0.00000592, 0.00000732, 0.00000114, 0.00000045, 0.00000171, 0.00000096, -0.00001379, 0.00000485, 0.00000361, -0.00000126, 0.00000179, -0.00000238, 0.00000080, 0.00000041, -0.00000197, 0.00000891, -0.00000535, -0.00000766, -0.00000072, 0.00000373, -0.00000114, 0.00000399, 0.00000511, -0.00000601, -0.00000093, -0.00000216, 0.00000435, 0.00000092, 0.00000348, -0.00000983, 0.00000417, -0.00000839, 0.00000276, 0.00000258, -0.00000476, 0.00001207, -0.00000026, -0.00000806, -0.00000179, -0.00000043, 0.00000003, -0.00000147, 0.00000650, 0.00000143, 0.00000180, -0.00000027, -0.00001006, 0.00000123, 0.00000498, 0.00000732, -0.00000295, -0.00000507, -0.00000034, 0.00001069, -0.00000427, -0.00000634, -0.00001286, 0.00000661, -0.00000497, -0.00001002, 0.00000152, 0.00000348, -0.00001025, -0.00001763, -0.00000273, -0.00001638, -0.00003398, -0.00003770, 0.00005474, 0.00007299, 0.00002871, 0.00007772, },
		{ -0.00005916, -0.00001840, -0.00002186, -0.00001587, -0.00004796, -0.00001382, 0.00005735, 0.00003415, 0.00003613, 0.00006699, 0.00000226, -0.00002860, -0.00003026, -0.00000173, -0.00000937, -0.00002287, 0.00000936, -0.00000085, -0.00000184, -0.00001496, 0.00000822, -0.00000601, -0.00000520, -0.00001067, 0.00000260, 0.00001230, -0.00000621, -0.00000496, 0.00000248, 0.00000513, 0.00000698, -0.00001028, -0.00000417, 0.00000422, -0.00000057, 0.00000376, 0.00000200, 0.00000023, -0.00000161, 0.00000205, -0.00000534, -0.00000817, 0.00000865, 0.00000645, -0.00000887, 0.00001228, -0.00001151, 0.00000541, -0.00000885, -0.00000040, 0.00000338, 0.00000001, 0.00000423, -0.00000453, -0.00000040, -0.00000482, 0.00000797, 0.00000004, 0.00000102, 0.00000192, -0.00000081, -0.00001318, 0.00000644, 0.00000246, -0.00000098, 0.00000154, -0.00000228, 0.00000089, 0.00000016, -0.00000134, 0.00000909, -0.00000725, -0.00000680, 0.00000020, 0.00000346, -0.00000163, 0.00000557, 0.00000298, -0.00000532, -0.00000123, -0.00000145, 0.00000431, 0.00000064, 0.00000340, -0.00001032, 0.00000500, -0.00000959, 0.00000493, 0.00000067, -0.00000357, 0.00001271, -0.00000237, -0.00000781, -0.00000101, -0.00000031, -0.00000029, -0.00000065, 0.00000629, 0.00000065, 0.00000237, -0.00000093, -0.00001022, 0.00000199, 0.00000577, 0.00000600, -0.00000312, -0.00000579, 0.00000179, 0.00001060, -0.00000656, -0.00000604, -0.00001201, 0.00000797, -0.00000747, -0.00000859, 0.00000140, 0.00000416, -0.00001164, -0.00001764, -0.00000120, -0.00001887, -0.00003555, -0.00002927, 0.00006064, 0.00006460, 0.00002891, 0.00007466, 0.00000415, },
		{ -0.00001764, -0.00002180, -0.00001639, -0.00005053, -0.00001042, 0.00006228, 0.00003249, 0.00004226, 0.00006746, -0.00000469, -0.00003196, -0.00002805, -0.00000003, -0.00001281, -0.00001942, 0.00000802, -0.00000039, -0.00000328, -0.00001383, 0.00000912, -0.00000794, -0.00000524, -0.00001048, 0.00000549, 0.00000985, -0.00000681, -0.00000415, 0.00000297, 0.00000585, 0.00000559, -0.00001048, -0.00000323, 0.00000383, -0.00000050, 0.00000449, 0.00000161, -0.00000028, -0.00000100, 0.00000128, -0.00000602, -0.00000735, 0.00001069, 0.00000339, -0.00000655, 0.00001099, -0.00001172, 0.00000594, -0.00000988, 0.00000117, 0.00000210, 0.00000161, 0.00000256, -0.00000358, -0.00000169, -0.00000327, 0.00000810, -0.00000086, 0.00000168, 0.00000186, -0.00000256, -0.00001214, 0.00000766, 0.00000136, -0.00000068, 0.00000132, -0.00000221, 0.00000100, -0.00000011, -0.00000056, 0.00000897, -0.00000908, -0.00000563, 0.00000098, 0.00000298, -0.00000174, 0.00000685, 0.00000077, -0.00000428, -0.00000182, -0.00000038, 0.00000383, 0.00000071, 0.00000304, -0.00001055, 0.00000565, -0.00001064, 0.00000714, -0.00000158, -0.00000176, 0.00001273, -0.00000434, -0.00000733, -0.00000024, -0.00000035, -0.00000050, 0.00000020, 0.00000584, 0.00000008, 0.00000290, -0.00000163, -0.00001038, 0.00000301, 0.00000620, 0.00000477, -0.00000330, -0.00000644, 0.00000434, 0.00000966, -0.00000842, -0.00000569, -0.00001093, 0.00000892, -0.00000991, -0.00000686, 0.00000094, 0.00000521, -0.00001365, -0.00001694, -0.00000002, -0.00002191, -0.00003569, -0.00002069, 0.00006455, 0.00005659, 0.00002973, 0.00007101, -0.00000147, -0.00005616, },
		},
		{ { 0.00015577, 0.00004666, 0.00001278, 0.00000396, -0.00005515, -0.00002812, -0.00000218, -0.00002879, 0.00000619, -0.00000626, -0.00000421, -0.00000650, 0.00000305, 0.00000958, -0.00000268, -0.00001394, 0.00000415, 0.00000192, -0.00000501, 0.00000429, -0.00000198, -0.00000177, 0.00000522, -0.00000592, 0.00000053, -0.00000167, -0.00000491, -0.00000169, -0.00000291, 0.00000624, 0.00000782, -0.00000746, 0.00000043, -0.00000290, 0.00000520, -0.00000524, 0.00000660, -0.00000400, 0.00000301, -0.00000471, -0.00000099, -0.00000426, 0.00000554, -0.00000648, 0.00000965, -0.00000743, 0.00001054, -0.00000405, -0.00000114, -0.00000225, 0.00000072, -0.00000290, 0.00000266, -0.00000615, 0.00000531, -0.00000418, 0.00000398, 0.00000091, 0.00000022, 0.00000271, -0.00000224, -0.00000035, -0.00000006, -0.00000286, 0.00000198, -0.00000286, -0.00000006, -0.00000035, -0.00000224, 0.00000271, 0.00000022, 0.00000091, 0.00000398, -0.00000418, 0.00000531, -0.00000615, 0.00000266, -0.00000290, 0.00000072, -0.00000225, -0.00000114, -0.00000405, 0.00001054, -0.00000743, 0.00000965, -0.00000648, 0.00000554, -0.00000426, -0.00000099, -0.00000471, 0.00000301, -0.00000400, 0.00000660, -0.00000524, 0.00000520, -0.00000290, 0.00000043, -0.00000746, 0.00000782, 0.00000624, -0.00000291, -0.00000169, -0.00000491, -0.00000167, 0.00000053, -0.00000592, 0.00000522, -0.00000177, -0.00000198, 0.00000429, -0.00000501, 0.00000192, 0.00000415, -0.00001394, -0.00000268, 0.00000958, 0.00000305, -0.00000650, -0.00000421, -0.00000626, 0.00000619, -0.00002879, -0.00000218, -0.00002812, -0.00005515, 0.00000396, 0.00001278, 0.00004666, },
		{ 0.00004929, 0.00001992, -0.00000087, -0.00005761, -0.00002678, -0.00000300, -0.00002756, 0.00000579, -0.00000720, -0.00000257, -0.00000786, 0.00000464, 0.00000822, -0.00000278, -0.00001329, 0.00000442, 0.00000115, -0.00000413, 0.00000384, -0.00000196, -0.00000204, 0.00000536, -0.00000595, 0.00000057, -0.00000185, -0.00000463, -0.00000179, -0.00000275, 0.00000673, 0.00000693, -0.00000706, -0.00000006, -0.00000221, 0.00000471, -0.00000491, 0.00000652, -0.00000409, 0.00000290, -0.00000448, -0.00000152, -0.00000343, 0.00000492, -0.00000593, 0.00000929, -0.00000739, 0.00001085, -0.00000466, -0.00000091, -0.00000221, 0.00000068, -0.00000274, 0.00000241, -0.00000601, 0.00000543, -0.00000445, 0.00000443, 0.00000042, 0.00000061, 0.00000245, -0.00000231, -0.00000015, -0.00000028, -0.00000266, 0.00000196, -0.00000304, 0.00000015, -0.00000056, -0.00000214, 0.00000293, -0.00000016, 0.00000142, 0.00000348, -0.00000386, 0.00000511, -0.00000623, 0.00000287, -0.00000305, 0.00000076, -0.00000231, -0.00000136, -0.00000341, 0.00001014, -0.00000736, 0.00000991, -0.00000695, 0.00000613, -0.00000511, -0.00000042, -0.00000497, 0.00000315, -0.00000393, 0.00000666, -0.00000553, 0.00000568, -0.00000362, 0.00000096, -0.00000787, 0.00000877, 0.00000562, -0.00000296, -0.00000166, -0.00000511, -0.00000157, 0.00000053, -0.00000593, 0.00000515, -0.00000155, -0.00000195, 0.00000468, -0.00000584, 0.00000270, 0.00000380, -0.00001454, -0.00000248, 0.00001083, 0.00000146, -0.00000508, -0.00000596, -0.00000509, 0.00000624, -0.00002968, -0.00000159, -0.00002936, -0.00005239, 0.00000816, 0.00000586, 0.00004395, 0.00015594, },
		{ 0.00002723, -0.00000635, -0.00005969, -0.00002543, -0.00000400, -0.00002601, 0.00000505, -0.00000787, -0.00000107, -0.00000915, 0.00000621, 0.00000676, -0.00000279, -0.00001260, 0.00000460, 0.00000040, -0.00000323, 0.00000335, -0.00000193, -0.00000235, 0.00000554, -0.00000601, 0.00000065, -0.00000209, -0.00000428, -0.00000197, -0.00000249, 0.00000709, 0.00000611, -0.00000670, -0.00000050, -0.00000156, 0.00000423, -0.00000456, 0.00000642, -0.00000419, 0.00000282, -0.00000428, -0.00000199, -0.00000262, 0.00000428, -0.00000533, 0.00000886, -0.00000726, 0.00001106, -0.00000522, -0.00000067, -0.00000218, 0.00000064, -0.00000256, 0.00000213, -0.00000581, 0.00000548, -0.00000464, 0.00000483, -0.00000006, 0.00000101, 0.00000216, -0.00000234, 0.00000003, -0.00000051, -0.00000242, 0.00000190, -0.00000318, 0.00000035, -0.00000078, -0.00000201, 0.00000311, -0.00000052, 0.00000192, 0.00000295, -0.00000347, 0.00000484, -0.00000624, 0.00000305, -0.00000319, 0.00000081, -0.00000238, -0.00000156, -0.00000273, 0.00000965, -0.00000720, 0.00001007, -0.00000735, 0.00000666, -0.00000595, 0.00000019, -0.00000527, 0.00000332, -0.00000387, 0.00000670, -0.00000579, 0.00000615, -0.00000437, 0.00000153, -0.00000829, 0.00000977, 0.00000488, -0.00000290, -0.00000172, -0.00000522, -0.00000154, 0.00000059, -0.00000597, 0.00000513, -0.00000139, -0.00000187, 0.00000500, -0.00000660, 0.00000348, 0.00000336, -0.00001510, -0.00000217, 0.00001195, -0.00000009, -0.00000363, -0.00000779, -0.00000371, 0.00000595, -0.00003020, -0.00000127, -0.00003043, -0.00004942, 0.00001178, -0.00000082, 0.00004113, 0.00015645, 0.00005188, },
		{ -0.00001250, -0.00006132, -0.00002413, -0.00000512, -0.00002420, 0.00000399, -0.00000829, 0.00000027, -0.00001032, 0.00000774, 0.00000522, -0.00000270, -0.00001189, 0.00000468, -0.00000030, -0.00000231, 0.00000283, -0.00000188, -0.00000269, 0.00000576, -0.00000609, 0.00000077, -0.00000238, -0.00000386, -0.00000221, -0.00000213, 0.00000732, 0.00000535, -0.00000636, -0.00000089, -0.00000097, 0.00000376, -0.00000419, 0.00000631, -0.00000430, 0.00000277, -0.00000413, -0.00000241, -0.00000185, 0.00000364, -0.00000470, 0.00000835, -0.00000704, 0.00001119, -0.00000574, -0.00000043, -0.00000216, 0.00000061, -0.00000239, 0.00000184, -0.00000556, 0.00000546, -0.00000476, 0.00000517, -0.00000051, 0.00000142, 0.00000184, -0.00000234, 0.00000020, -0.00000074, -0.00000216, 0.00000179, -0.00000328, 0.00000052, -0.00000100, -0.00000185, 0.00000326, -0.00000084, 0.00000242, 0.00000240, -0.00000302, 0.00000450, -0.00000619, 0.00000317, -0.00000330, 0.00000084, -0.00000246, -0.00000174, -0.00000203, 0.00000909, -0.00000694, 0.00001013, -0.00000766, 0.00000713, -0.00000677, 0.00000082, -0.00000558, 0.00000351, -0.00000383, 0.00000671, -0.00000601, 0.00000660, -0.00000514, 0.00000212, -0.00000872, 0.00001080, 0.00000402, -0.00000274, -0.00000186, -0.00000524, -0.00000159, 0.00000071, -0.00000606, 0.00000519, -0.00000131, -0.00000173, 0.00000523, -0.00000728, 0.00000423, 0.00000285, -0.00001561, -0.00000176, 0.00001291, -0.00000158, -0.00000219, -0.00000966, -0.00000215, 0.00000532, -0.00003034, -0.00000127, -0.00003126, -0.00004633, 0.00001487, -0.00000724, 0.00003818, 0.00015730, 0.00005446, 0.00003467, },
		{ -0.00006241, -0.00002296, -0.00000630, -0.00002216, 0.00000263, -0.00000843, 0.00000142, -0.00001136, 0.00000920, 0.00000363, -0.00000251, -0.00001115, 0.00000466, -0.00000096, -0.00000139, 0.00000229, -0.00000182, -0.00000304, 0.00000600, -0.00000618, 0.00000091, -0.00000272, -0.00000338, -0.00000250, -0.00000169, 0.00000743, 0.00000467, -0.00000606, -0.00000121, -0.00000043, 0.00000332, -0.00000381, 0.00000618, -0.00000443, 0.00000276, -0.00000402, -0.00000276, -0.00000114, 0.00000300, -0.00000403, 0.00000777, -0.00000673, 0.00001122, -0.00000619, -0.00000019, -0.00000216, 0.00000059, -0.00000222, 0.00000153, -0.00000526, 0.00000537, -0.00000481, 0.00000546, -0.00000094, 0.00000183, 0.00000150, -0.00000230, 0.00000034, -0.00000096, -0.00000188, 0.00000165, -0.00000335, 0.00000068, -0.00000122, -0.00000166, 0.00000336, -0.00000113, 0.00000290, 0.00000182, -0.00000253, 0.00000408, -0.00000606, 0.00000325, -0.00000339, 0.00000087, -0.00000255, -0.00000191, -0.00000131, 0.00000845, -0.00000658, 0.00001008, -0.00000787, 0.00000752, -0.00000756, 0.00000146, -0.00000591, 0.00000372, -0.00000381, 0.00000671, -0.00000618, 0.00000702, -0.00000590, 0.00000273, -0.00000914, 0.00001186, 0.00000306, -0.00000246, -0.00000209, -0.00000517, -0.00000173, 0.00000090, -0.00000620, 0.00000533, -0.00000131, -0.00000151, 0.00000538, -0.00000787, 0.00000494, 0.00000228, -0.00001606, -0.00000125, 0.00001371, -0.00000299, -0.00000078, -0.00001153, -0.00000045, 0.00000438, -0.00003009, -0.00000161, -0.00003180, -0.00004321, 0.00001748, -0.00001339, 0.00003509, 0.00015846, 0.00005711, 0.00004216, -0.00001932, },
		{ -0.00002201, -0.00000746, -0.00001996, 0.00000101, -0.00000830, 0.00000237, -0.00001226, 0.00001057, 0.00000199, -0.00000224, -0.00001039, 0.00000455, -0.00000155, -0.00000048, 0.00000174, -0.00000178, -0.00000338, 0.00000626, -0.00000627, 0.00000107, -0.00000310, -0.00000286, -0.00000285, -0.00000116, 0.00000742, 0.00000406, -0.00000580, -0.00000148, 0.00000004, 0.00000290, -0.00000341, 0.00000603, -0.00000457, 0.00000278, -0.00000396, -0.00000304, -0.00000048, 0.00000238, -0.00000334, 0.00000715, -0.00000635, 0.00001116, -0.00000659, 0.00000005, -0.00000216, 0.00000059, -0.00000206, 0.00000121, -0.00000493, 0.00000522, -0.00000479, 0.00000568, -0.00000133, 0.00000223, 0.00000114, -0.00000223, 0.00000046, -0.00000118, -0.00000158, 0.00000146, -0.00000337, 0.00000081, -0.00000144, -0.00000145, 0.00000343, -0.00000139, 0.00000336, 0.00000123, -0.00000199, 0.00000360, -0.00000587, 0.00000327, -0.00000344, 0.00000089, -0.00000265, -0.00000204, -0.00000058, 0.00000775, -0.00000612, 0.00000992, -0.00000798, 0.00000783, -0.00000831, 0.00000209, -0.00000624, 0.00000395, -0.00000379, 0.00000668, -0.00000631, 0.00000740, -0.00000667, 0.00000335, -0.00000954, 0.00001293, 0.00000200, -0.00000208, -0.00000240, -0.00000501, -0.00000196, 0.00000115, -0.00000640, 0.00000556, -0.00000142, -0.00000121, 0.00000541, -0.00000834, 0.00000559, 0.00000166, -0.00001647, -0.00000063, 0.00001433, -0.00000428, 0.00000055, -0.00001335, 0.00000134, 0.00000316, -0.00002945, -0.00000231, -0.00003200, -0.00004014, 0.00001967, -0.00001929, 0.00003185, 0.00015993, 0.00005987, 0.00004964, -0.00002680, -0.00006290, },
		{ -0.00000854, -0.00001766, -0.00000084, -0.00000792, 0.00000312, -0.00001299, 0.00001184, 0.00000034, -0.00000188, -0.00000962, 0.00000436, -0.00000208, 0.00000041, 0.00000118, -0.00000175, -0.00000371, 0.00000651, -0.00000635, 0.00000123, -0.00000350, -0.00000229, -0.00000323, -0.00000057, 0.00000729, 0.00000353, -0.00000557, -0.00000169, 0.00000044, 0.00000252, -0.00000302, 0.00000586, -0.00000472, 0.00000282, -0.00000394, -0.00000324, 0.00000011, 0.00000179, -0.00000266, 0.00000649, -0.00000590, 0.00001101, -0.00000693, 0.00000028, -0.00000218, 0.00000059, -0.00000191, 0.00000089, -0.00000457, 0.00000501, -0.00000471, 0.00000584, -0.00000168, 0.00000261, 0.00000076, -0.00000213, 0.00000055, -0.00000138, -0.00000127, 0.00000125, -0.00000335, 0.00000091, -0.00000164, -0.00000121, 0.00000345, -0.00000160, 0.00000379, 0.00000063, -0.00000141, 0.00000306, -0.00000560, 0.00000324, -0.00000346, 0.00000089, -0.00000275, -0.00000216, 0.00000015, 0.00000698, -0.00000557, 0.00000965, -0.00000797, 0.00000804, -0.00000899, 0.00000271, -0.00000657, 0.00000418, -0.00000378, 0.00000663, -0.00000638, 0.00000772, -0.00000741, 0.00000397, -0.00000991, 0.00001400, 0.00000086, -0.00000159, -0.00000280, -0.00000476, -0.00000227, 0.00000147, -0.00000664, 0.00000586, -0.00000162, -0.00000082, 0.00000533, -0.00000868, 0.00000615, 0.00000101, -0.00001682, 0.00000009, 0.00001477, -0.00000545, 0.00000179, -0.00001509, 0.00000318, 0.00000170, -0.00002846, -0.00000336, -0.00003182, -0.00003721, 0.00002152, -0.00002493, 0.00002845, 0.00016167, 0.00006281, 0.00005701, -0.00003491, -0.00006276, -0.00002133, },
		{ -0.00001532, -0.00000287, -0.00000730, 0.00000367, -0.00001356, 0.00001298, -0.00000132, -0.00000144, -0.00000886, 0.00000408, -0.00000253, 0.00000127, 0.00000064, -0.00000174, -0.00000402, 0.00000676, -0.00000641, 0.00000139, -0.00000392, -0.00000170, -0.00000363, 0.00000008, 0.00000705, 0.00000307, -0.00000538, -0.00000183, 0.00000078, 0.00000218, -0.00000262, 0.00000569, -0.00000488, 0.00000290, -0.00000397, -0.00000336, 0.00000062, 0.00000124, -0.00000199, 0.00000580, -0.00000539, 0.00001078, -0.00000720, 0.00000050, -0.00000221, 0.00000061, -0.00000177, 0.00000058, -0.00000419, 0.00000475, -0.00000455, 0.00000594, -0.00000198, 0.00000298, 0.00000037, -0.00000199, 0.00000061, -0.00000156, -0.00000096, 0.00000100, -0.00000330, 0.00000097, -0.00000183, -0.00000097, 0.00000344, -0.00000177, 0.00000419, 0.00000004, -0.00000081, 0.00000247, -0.00000526, 0.00000314, -0.00000343, 0.00000087, -0.00000286, -0.00000224, 0.00000087, 0.00000617, -0.00000493, 0.00000926, -0.00000786, 0.00000814, -0.00000960, 0.00000330, -0.00000688, 0.00000441, -0.00000377, 0.00000654, -0.00000639, 0.00000799, -0.00000812, 0.00000457, -0.00001024, 0.00001505, -0.00000036, -0.00000102, -0.00000327, -0.00000441, -0.00000267, 0.00000186, -0.00000694, 0.00000625, -0.00000193, -0.00000033, 0.00000512, -0.00000887, 0.00000661, 0.00000033, -0.00001712, 0.00000091, 0.00001500, -0.00000645, 0.00000289, -0.00001669, 0.00000500, 0.00000004, -0.00002713, -0.00000477, -0.00003123, -0.00003448, 0.00002308, -0.00003034, 0.00002490, 0.00016365, 0.00006601, 0.00006417, -0.00004360, -0.00006194, -0.00002100, -0.00000948, },
		{ -0.00000502, -0.00000648, 0.00000400, -0.00001394, 0.00001398, -0.00000297, -0.00000093, -0.00000810, 0.00000372, -0.00000292, 0.00000209, 0.00000010, -0.00000177, -0.00000429, 0.00000699, -0.00000645, 0.00000153, -0.00000434, -0.00000109, -0.00000405, 0.00000078, 0.00000671, 0.00000269, -0.00000523, -0.00000192, 0.00000104, 0.00000189, -0.00000223, 0.00000549, -0.00000504, 0.00000299, -0.00000403, -0.00000341, 0.00000105, 0.00000075, -0.00000134, 0.00000510, -0.00000484, 0.00001046, -0.00000741, 0.00000071, -0.00000225, 0.00000065, -0.00000166, 0.00000029, -0.00000379, 0.00000445, -0.00000434, 0.00000597, -0.00000224, 0.00000331, -0.00000002, -0.00000183, 0.00000065, -0.00000173, -0.00000064, 0.00000073, -0.00000320, 0.00000101, -0.00000200, -0.00000071, 0.00000339, -0.00000190, 0.00000454, -0.00000053, -0.00000019, 0.00000182, -0.00000485, 0.00000299, -0.00000336, 0.00000082, -0.00000296, -0.00000229, 0.00000158, 0.00000532, -0.00000422, 0.00000877, -0.00000762, 0.00000812, -0.00001013, 0.00000385, -0.00000717, 0.00000463, -0.00000376, 0.00000644, -0.00000635, 0.00000819, -0.00000879, 0.00000514, -0.00001052, 0.00001606, -0.00000163, -0.00000036, -0.00000381, -0.00000399, -0.00000315, 0.00000231, -0.00000728, 0.00000673, -0.00000234, 0.00000027, 0.00000478, -0.00000890, 0.00000695, -0.00000034, -0.00001738, 0.00000183, 0.00001502, -0.00000729, 0.00000384, -0.00001812, 0.00000676, -0.00000176, -0.00002551, -0.00000652, -0.00003023, -0.00003201, 0.00002441, -0.00003553, 0.00002121, 0.00016587, 0.00006953, 0.00007100, -0.00005281, -0.00006044, -0.00002105, -0.00001022, -0.00001301, },
		{ -0.00000547, 0.00000414, -0.00001415, 0.00001484, -0.00000459, -0.00000035, -0.00000737, 0.00000330, -0.00000323, 0.00000287, -0.00000042, -0.00000183, -0.00000451, 0.00000720, -0.00000646, 0.00000165, -0.00000476, -0.00000047, -0.00000448, 0.00000151, 0.00000629, 0.00000237, -0.00000511, -0.00000195, 0.00000123, 0.00000165, -0.00000185, 0.00000528, -0.00000519, 0.00000311, -0.00000413, -0.00000338, 0.00000139, 0.00000032, -0.00000073, 0.00000440, -0.00000425, 0.00001008, -0.00000756, 0.00000090, -0.00000229, 0.00000070, -0.00000156, 0.00000001, -0.00000339, 0.00000412, -0.00000407, 0.00000593, -0.00000244, 0.00000362, -0.00000041, -0.00000164, 0.00000065, -0.00000187, -0.00000033, 0.00000043, -0.00000307, 0.00000101, -0.00000215, -0.00000045, 0.00000330, -0.00000198, 0.00000485, -0.00000108, 0.00000043, 0.00000114, -0.00000438, 0.00000278, -0.00000325, 0.00000075, -0.00000307, -0.00000232, 0.00000227, 0.00000444, -0.00000344, 0.00000818, -0.00000727, 0.00000799, -0.00001056, 0.00000435, -0.00000742, 0.00000484, -0.00000375, 0.00000630, -0.00000624, 0.00000832, -0.00000941, 0.00000568, -0.00001074, 0.00001702, -0.00000293, 0.00000036, -0.00000442, -0.00000349, -0.00000371, 0.00000283, -0.00000767, 0.00000728, -0.00000287, 0.00000097, 0.00000431, -0.00000876, 0.00000714, -0.00000101, -0.00001758, 0.00000284, 0.00001483, -0.00000793, 0.00000459, -0.00001933, 0.00000840, -0.00000365, -0.00002363, -0.00000856, -0.00002880, -0.00002986, 0.00002558, -0.00004051, 0.00001737, 0.00016828, 0.00007346, 0.00007737, -0.00006244, -0.00005823, -0.00002155, -0.00001069, -0.00001079, -0.00000724, },
		{ 0.00000410, -0.00001417, 0.00001554, -0.00000616, 0.00000030, -0.00000666, 0.00000283, -0.00000347, 0.00000360, -0.00000092, -0.00000193, -0.00000468, 0.00000736, -0.00000643, 0.00000174, -0.00000516, 0.00000014, -0.00000490, 0.00000225, 0.00000578, 0.00000212, -0.00000502, -0.00000193, 0.00000134, 0.00000146, -0.00000148, 0.00000506, -0.00000535, 0.00000324, -0.00000426, -0.00000328, 0.00000164, -0.00000003, -0.00000016, 0.00000371, -0.00000363, 0.00000962, -0.00000764, 0.00000107, -0.00000234, 0.00000077, -0.00000149, -0.00000025, -0.00000300, 0.00000376, -0.00000375, 0.00000584, -0.00000259, 0.00000389, -0.00000080, -0.00000143, 0.00000063, -0.00000198, -0.00000003, 0.00000012, -0.00000291, 0.00000097, -0.00000227, -0.00000019, 0.00000318, -0.00000201, 0.00000510, -0.00000160, 0.00000106, 0.00000043, -0.00000386, 0.00000251, -0.00000309, 0.00000064, -0.00000316, -0.00000231, 0.00000294, 0.00000354, -0.00000260, 0.00000749, -0.00000680, 0.00000773, -0.00001088, 0.00000478, -0.00000763, 0.00000503, -0.00000373, 0.00000613, -0.00000607, 0.00000838, -0.00000997, 0.00000618, -0.00001087, 0.00001791, -0.00000425, 0.00000114, -0.00000508, -0.00000292, -0.00000433, 0.00000339, -0.00000809, 0.00000790, -0.00000349, 0.00000178, 0.00000369, -0.00000844, 0.00000717, -0.00000165, -0.00001775, 0.00000394, 0.00001442, -0.00000836, 0.00000514, -0.00002029, 0.00000985, -0.00000555, -0.00002157, -0.00001086, -0.00002697, -0.00002805, 0.00002664, -0.00004531, 0.00001340, 0.00017086, 0.00007786, 0.00008317, -0.00007241, -0.00005535, -0.00002251, -0.00001084, -0.00000871, -0.00000948, -0.00000432, },
		{ -0.00001403, 0.00001607, -0.00000767, 0.00000100, -0.00000600, 0.00000232, -0.00000364, 0.00000429, -0.00000141, -0.00000206, -0.00000480, 0.00000749, -0.00000636, 0.00000180, -0.00000553, 0.00000075, -0.00000531, 0.00000300, 0.00000521, 0.00000192, -0.00000495, -0.00000185, 0.00000139, 0.00000133, -0.00000114, 0.00000483, -0.00000550, 0.00000338, -0.00000442, -0.00000311, 0.00000181, -0.00000032, 0.00000035, 0.00000305, -0.00000300, 0.00000911, -0.00000766, 0.00000122, -0.00000239, 0.00000085, -0.00000145, -0.00000047, -0.00000261, 0.00000338, -0.00000339, 0.00000568, -0.00000269, 0.00000412, -0.00000117, -0.00000120, 0.00000057, -0.00000206, 0.00000025, -0.00000020, -0.00000272, 0.00000090, -0.00000237, 0.00000008, 0.00000304, -0.00000199, 0.00000530, -0.00000208, 0.00000167, -0.00000031, -0.00000328, 0.00000218, -0.00000288, 0.00000051, -0.00000325, -0.00000227, 0.00000357, 0.00000264, -0.00000171, 0.00000671, -0.00000623, 0.00000735, -0.00001109, 0.00000514, -0.00000779, 0.00000519, -0.00000370, 0.00000593, -0.00000583, 0.00000834, -0.00001046, 0.00000662, -0.00001093, 0.00001872, -0.00000557, 0.00000196, -0.00000578, -0.00000230, -0.00000502, 0.00000401, -0.00000854, 0.00000858, -0.00000422, 0.00000269, 0.00000293, -0.00000794, 0.00000703, -0.00000224, -0.00001787, 0.00000512, 0.00001379, -0.00000858, 0.00000547, -0.00002097, 0.00001108, -0.00000742, -0.00001937, -0.00001336, -0.00002477, -0.00002662, 0.00002762, -0.00004993, 0.00000928, 0.00017361, 0.00008282, 0.00008824, -0.00008260, -0.00005181, -0.00002396, -0.00001065, -0.00000685, -0.00001167, -0.00000307, 0.00000390, },
		{ 0.00001645, -0.00000912, 0.00000176, -0.00000538, 0.00000178, -0.00000376, 0.00000493, -0.00000189, -0.00000223, -0.00000486, 0.00000756, -0.00000623, 0.00000181, -0.00000587, 0.00000133, -0.00000569, 0.00000375, 0.00000458, 0.00000177, -0.00000490, -0.00000174, 0.00000138, 0.00000126, -0.00000081, 0.00000458, -0.00000564, 0.00000352, -0.00000459, -0.00000288, 0.00000188, -0.00000053, 0.00000080, 0.00000242, -0.00000237, 0.00000855, -0.00000762, 0.00000135, -0.00000244, 0.00000094, -0.00000143, -0.00000067, -0.00000225, 0.00000299, -0.00000300, 0.00000547, -0.00000273, 0.00000430, -0.00000153, -0.00000095, 0.00000049, -0.00000212, 0.00000051, -0.00000052, -0.00000250, 0.00000080, -0.00000244, 0.00000033, 0.00000287, -0.00000193, 0.00000545, -0.00000252, 0.00000226, -0.00000105, -0.00000267, 0.00000180, -0.00000263, 0.00000033, -0.00000332, -0.00000220, 0.00000415, 0.00000174, -0.00000078, 0.00000586, -0.00000555, 0.00000684, -0.00001118, 0.00000542, -0.00000789, 0.00000531, -0.00000366, 0.00000570, -0.00000552, 0.00000822, -0.00001087, 0.00000699, -0.00001088, 0.00001942, -0.00000687, 0.00000281, -0.00000652, -0.00000163, -0.00000576, 0.00000466, -0.00000901, 0.00000931, -0.00000503, 0.00000370, 0.00000204, -0.00000725, 0.00000670, -0.00000277, -0.00001796, 0.00000637, 0.00001295, -0.00000857, 0.00000555, -0.00002132, 0.00001202, -0.00000918, -0.00001711, -0.00001600, -0.00002223, -0.00002556, 0.00002856, -0.00005439, 0.00000503, 0.00017649, 0.00008839, 0.00009246, -0.00009288, -0.00004768, -0.00002590, -0.00001006, -0.00000525, -0.00001375, -0.00000175, 0.00000355, -0.00001372, },
		{ -0.00001048, 0.00000255, -0.00000481, 0.00000122, -0.00000382, 0.00000552, -0.00000235, -0.00000243, -0.00000485, 0.00000759, -0.00000606, 0.00000177, -0.00000618, 0.00000189, -0.00000604, 0.00000448, 0.00000391, 0.00000166, -0.00000487, -0.00000159, 0.00000130, 0.00000124, -0.00000050, 0.00000432, -0.00000576, 0.00000366, -0.00000478, -0.00000260, 0.00000188, -0.00000065, 0.00000118, 0.00000183, -0.00000174, 0.00000795, -0.00000752, 0.00000146, -0.00000248, 0.00000105, -0.00000143, -0.00000083, -0.00000191, 0.00000260, -0.00000257, 0.00000522, -0.00000272, 0.00000444, -0.00000186, -0.00000070, 0.00000038, -0.00000213, 0.00000075, -0.00000085, -0.00000226, 0.00000066, -0.00000247, 0.00000057, 0.00000269, -0.00000182, 0.00000553, -0.00000291, 0.00000282, -0.00000180, -0.00000202, 0.00000137, -0.00000233, 0.00000013, -0.00000338, -0.00000211, 0.00000470, 0.00000085, 0.00000017, 0.00000493, -0.00000478, 0.00000622, -0.00001115, 0.00000560, -0.00000791, 0.00000539, -0.00000359, 0.00000543, -0.00000515, 0.00000801, -0.00001119, 0.00000729, -0.00001073, 0.00002000, -0.00000813, 0.00000367, -0.00000726, -0.00000092, -0.00000654, 0.00000534, -0.00000949, 0.00001008, -0.00000592, 0.00000480, 0.00000102, -0.00000637, 0.00000618, -0.00000322, -0.00001801, 0.00000769, 0.00001190, -0.00000834, 0.00000537, -0.00002132, 0.00001262, -0.00001077, -0.00001485, -0.00001872, -0.00001942, -0.00002488, 0.00002949, -0.00005869, 0.00000063, 0.00017952, 0.00009464, 0.00009568, -0.00010312, -0.00004301, -0.00002831, -0.00000908, -0.00000396, -0.00001568, -0.00000042, 0.00000309, -0.00001327, 0.00001666, },
		{ 0.00000339, -0.00000431, 0.00000067, -0.00000384, 0.00000608, -0.00000281, -0.00000266, -0.00000479, 0.00000756, -0.00000582, 0.00000168, -0.00000643, 0.00000241, -0.00000635, 0.00000518, 0.00000321, 0.00000159, -0.00000484, -0.00000141, 0.00000117, 0.00000128, -0.00000023, 0.00000405, -0.00000587, 0.00000380, -0.00000497, -0.00000226, 0.00000179, -0.00000070, 0.00000148, 0.00000130, -0.00000114, 0.00000732, -0.00000736, 0.00000154, -0.00000253, 0.00000116, -0.00000146, -0.00000095, -0.00000160, 0.00000221, -0.00000213, 0.00000492, -0.00000266, 0.00000453, -0.00000217, -0.00000044, 0.00000026, -0.00000212, 0.00000096, -0.00000117, -0.00000200, 0.00000050, -0.00000247, 0.00000080, 0.00000249, -0.00000168, 0.00000556, -0.00000324, 0.00000334, -0.00000253, -0.00000134, 0.00000089, -0.00000198, -0.00000012, -0.00000342, -0.00000198, 0.00000519, -0.00000002, 0.00000113, 0.00000396, -0.00000391, 0.00000547, -0.00001099, 0.00000569, -0.00000786, 0.00000541, -0.00000350, 0.00000513, -0.00000472, 0.00000770, -0.00001142, 0.00000751, -0.00001046, 0.00002045, -0.00000933, 0.00000452, -0.00000801, -0.00000019, -0.00000734, 0.00000603, -0.00000996, 0.00001087, -0.00000687, 0.00000597, -0.00000012, -0.00000531, 0.00000544, -0.00000359, -0.00001803, 0.00000907, 0.00001065, -0.00000788, 0.00000494, -0.00002096, 0.00001285, -0.00001213, -0.00001268, -0.00002143, -0.00001639, -0.00002456, 0.00003041, -0.00006282, -0.00000393, 0.00018269, 0.00010161, 0.00009776, -0.00011315, -0.00003788, -0.00003118, -0.00000767, -0.00000301, -0.00001740, 0.00000090, 0.00000254, -0.00001268, 0.00001671, -0.00001175, },
		{ -0.00000387, 0.00000012, -0.00000382, 0.00000660, -0.00000327, -0.00000290, -0.00000467, 0.00000747, -0.00000554, 0.00000154, -0.00000664, 0.00000289, -0.00000662, 0.00000584, 0.00000248, 0.00000155, -0.00000482, -0.00000122, 0.00000099, 0.00000137, 0.00000002, 0.00000377, -0.00000596, 0.00000392, -0.00000516, -0.00000189, 0.00000163, -0.00000067, 0.00000171, 0.00000082, -0.00000057, 0.00000666, -0.00000716, 0.00000160, -0.00000257, 0.00000127, -0.00000151, -0.00000104, -0.00000132, 0.00000184, -0.00000168, 0.00000458, -0.00000255, 0.00000457, -0.00000245, -0.00000018, 0.00000011, -0.00000208, 0.00000114, -0.00000148, -0.00000172, 0.00000030, -0.00000243, 0.00000100, 0.00000228, -0.00000150, 0.00000553, -0.00000351, 0.00000382, -0.00000325, -0.00000064, 0.00000038, -0.00000160, -0.00000040, -0.00000344, -0.00000184, 0.00000563, -0.00000085, 0.00000210, 0.00000294, -0.00000298, 0.00000462, -0.00001071, 0.00000568, -0.00000773, 0.00000538, -0.00000337, 0.00000480, -0.00000422, 0.00000730, -0.00001155, 0.00000765, -0.00001007, 0.00002075, -0.00001044, 0.00000534, -0.00000875, 0.00000054, -0.00000816, 0.00000673, -0.00001041, 0.00001167, -0.00000788, 0.00000720, -0.00000136, -0.00000407, 0.00000451, -0.00000385, -0.00001803, 0.00001049, 0.00000922, -0.00000721, 0.00000426, -0.00002023, 0.00001266, -0.00001319, -0.00001065, -0.00002405, -0.00001322, -0.00002455, 0.00003133, -0.00006678, -0.00000868, 0.00018602, 0.00010934, 0.00009857, -0.00012284, -0.00003239, -0.00003447, -0.00000586, -0.00000244, -0.00001886, 0.00000214, 0.00000193, -0.00001198, 0.00001660, -0.00001292, 0.00000425, },
		{ -0.00000040, -0.00000377, 0.00000709, -0.00000374, -0.00000314, -0.00000450, 0.00000733, -0.00000520, 0.00000135, -0.00000680, 0.00000332, -0.00000683, 0.00000645, 0.00000175, 0.00000153, -0.00000479, -0.00000101, 0.00000077, 0.00000152, 0.00000024, 0.00000348, -0.00000603, 0.00000403, -0.00000533, -0.00000149, 0.00000140, -0.00000057, 0.00000186, 0.00000042, -0.00000003, 0.00000599, -0.00000691, 0.00000163, -0.00000260, 0.00000140, -0.00000158, -0.00000108, -0.00000109, 0.00000149, -0.00000122, 0.00000421, -0.00000240, 0.00000457, -0.00000269, 0.00000008, -0.00000006, -0.00000200, 0.00000129, -0.00000177, -0.00000145, 0.00000009, -0.00000237, 0.00000119, 0.00000207, -0.00000129, 0.00000543, -0.00000371, 0.00000424, -0.00000393, 0.00000006, -0.00000016, -0.00000118, -0.00000072, -0.00000343, -0.00000167, 0.00000601, -0.00000163, 0.00000304, 0.00000189, -0.00000198, 0.00000367, -0.00001030, 0.00000556, -0.00000751, 0.00000529, -0.00000322, 0.00000443, -0.00000366, 0.00000681, -0.00001158, 0.00000769, -0.00000955, 0.00002088, -0.00001146, 0.00000612, -0.00000946, 0.00000127, -0.00000897, 0.00000742, -0.00001082, 0.00001245, -0.00000891, 0.00000849, -0.00000270, -0.00000267, 0.00000336, -0.00000401, -0.00001800, 0.00001195, 0.00000761, -0.00000632, 0.00000332, -0.00001911, 0.00001204, -0.00001390, -0.00000884, -0.00002651, -0.00000999, -0.00002483, 0.00003223, -0.00007053, -0.00001365, 0.00018954, 0.00011785, 0.00009800, -0.00013202, -0.00002665, -0.00003811, -0.00000366, -0.00000225, -0.00002003, 0.00000327, 0.00000129, -0.00001119, 0.00001635, -0.00001398, 0.00000512, -0.00000351, },
		{ -0.00000371, 0.00000755, -0.00000422, -0.00000339, -0.00000429, 0.00000715, -0.00000481, 0.00000110, -0.00000689, 0.00000371, -0.00000699, 0.00000701, 0.00000102, 0.00000152, -0.00000475, -0.00000079, 0.00000052, 0.00000171, 0.00000043, 0.00000318, -0.00000607, 0.00000412, -0.00000549, -0.00000107, 0.00000112, -0.00000039, 0.00000194, 0.00000008, 0.00000046, 0.00000532, -0.00000663, 0.00000164, -0.00000262, 0.00000152, -0.00000167, -0.00000109, -0.00000089, 0.00000117, -0.00000077, 0.00000383, -0.00000221, 0.00000451, -0.00000290, 0.00000032, -0.00000024, -0.00000189, 0.00000140, -0.00000205, -0.00000116, -0.00000015, -0.00000227, 0.00000134, 0.00000186, -0.00000105, 0.00000529, -0.00000386, 0.00000460, -0.00000458, 0.00000076, -0.00000072, -0.00000072, -0.00000107, -0.00000339, -0.00000148, 0.00000633, -0.00000237, 0.00000397, 0.00000083, -0.00000093, 0.00000262, -0.00000978, 0.00000534, -0.00000720, 0.00000513, -0.00000302, 0.00000404, -0.00000305, 0.00000623, -0.00001151, 0.00000765, -0.00000891, 0.00002085, -0.00001236, 0.00000683, -0.00001012, 0.00000197, -0.00000977, 0.00000809, -0.00001120, 0.00001320, -0.00000996, 0.00000979, -0.00000410, -0.00000112, 0.00000201, -0.00000405, -0.00001794, 0.00001342, 0.00000584, -0.00000523, 0.00000214, -0.00001761, 0.00001095, -0.00001422, -0.00000732, -0.00002871, -0.00000680, -0.00002531, 0.00003309, -0.00007406, -0.00001888, 0.00019330, 0.00012713, 0.00009592, -0.00014052, -0.00002078, -0.00004204, -0.00000109, -0.00000246, -0.00002087, 0.00000426, 0.00000064, -0.00001032, 0.00001596, -0.00001492, 0.00000601, -0.00000322, -0.00000088, },
		{ 0.00000799, -0.00000472, -0.00000362, -0.00000404, 0.00000692, -0.00000437, 0.00000081, -0.00000694, 0.00000404, -0.00000709, 0.00000750, 0.00000029, 0.00000152, -0.00000470, -0.00000058, 0.00000024, 0.00000195, 0.00000059, 0.00000288, -0.00000609, 0.00000418, -0.00000562, -0.00000065, 0.00000079, -0.00000015, 0.00000193, -0.00000019, 0.00000089, 0.00000465, -0.00000631, 0.00000163, -0.00000264, 0.00000164, -0.00000177, -0.00000107, -0.00000074, 0.00000088, -0.00000033, 0.00000342, -0.00000199, 0.00000442, -0.00000308, 0.00000056, -0.00000042, -0.00000175, 0.00000148, -0.00000230, -0.00000088, -0.00000040, -0.00000214, 0.00000147, 0.00000165, -0.00000079, 0.00000509, -0.00000394, 0.00000490, -0.00000518, 0.00000144, -0.00000129, -0.00000024, -0.00000146, -0.00000333, -0.00000127, 0.00000660, -0.00000304, 0.00000485, -0.00000023, 0.00000015, 0.00000151, -0.00000914, 0.00000502, -0.00000680, 0.00000490, -0.00000279, 0.00000361, -0.00000239, 0.00000556, -0.00001134, 0.00000751, -0.00000813, 0.00002064, -0.00001312, 0.00000747, -0.00001072, 0.00000263, -0.00001055, 0.00000873, -0.00001150, 0.00001390, -0.00001099, 0.00001111, -0.00000556, 0.00000057, 0.00000045, -0.00000396, -0.00001786, 0.00001490, 0.00000393, -0.00000397, 0.00000074, -0.00001574, 0.00000938, -0.00001410, -0.00000615, -0.00003058, -0.00000374, -0.00002595, 0.00003388, -0.00007732, -0.00002442, 0.00019734, 0.00013719, 0.00009225, -0.00014819, -0.00001490, -0.00004619, 0.00000180, -0.00000307, -0.00002136, 0.00000506, 0.00000003, -0.00000941, 0.00001545, -0.00001575, 0.00000690, -0.00000300, -0.00000132, -0.00000364, },
		{ -0.00000525, -0.00000382, -0.00000376, 0.00000665, -0.00000389, 0.00000046, -0.00000692, 0.00000432, -0.00000713, 0.00000793, -0.00000042, 0.00000153, -0.00000463, -0.00000038, -0.00000006, 0.00000222, 0.00000071, 0.00000257, -0.00000609, 0.00000421, -0.00000572, -0.00000022, 0.00000042, 0.00000014, 0.00000186, -0.00000038, 0.00000127, 0.00000400, -0.00000596, 0.00000160, -0.00000264, 0.00000175, -0.00000188, -0.00000101, -0.00000064, 0.00000063, 0.00000009, 0.00000301, -0.00000174, 0.00000427, -0.00000321, 0.00000078, -0.00000061, -0.00000159, 0.00000151, -0.00000252, -0.00000061, -0.00000067, -0.00000197, 0.00000157, 0.00000145, -0.00000052, 0.00000484, -0.00000395, 0.00000513, -0.00000572, 0.00000210, -0.00000187, 0.00000027, -0.00000188, -0.00000324, -0.00000106, 0.00000680, -0.00000365, 0.00000569, -0.00000127, 0.00000125, 0.00000032, -0.00000840, 0.00000459, -0.00000630, 0.00000459, -0.00000252, 0.00000314, -0.00000168, 0.00000480, -0.00001107, 0.00000728, -0.00000723, 0.00002025, -0.00001374, 0.00000801, -0.00001125, 0.00000324, -0.00001127, 0.00000931, -0.00001174, 0.00001454, -0.00001199, 0.00001241, -0.00000704, 0.00000236, -0.00000128, -0.00000375, -0.00001776, 0.00001637, 0.00000191, -0.00000254, -0.00000087, -0.00001352, 0.00000734, -0.00001352, -0.00000538, -0.00003204, -0.00000089, -0.00002666, 0.00003455, -0.00008027, -0.00003032, 0.00020174, 0.00014797, 0.00008691, -0.00015488, -0.00000913, -0.00005046, 0.00000496, -0.00000405, -0.00002147, 0.00000565, -0.00000053, -0.00000846, 0.00001483, -0.00001646, 0.00000777, -0.00000286, -0.00000171, -0.00000357, 0.00000841, },
		{ -0.00000399, -0.00000346, 0.00000635, -0.00000337, 0.00000008, -0.00000686, 0.00000455, -0.00000712, 0.00000829, -0.00000110, 0.00000153, -0.00000453, -0.00000019, -0.00000036, 0.00000253, 0.00000080, 0.00000226, -0.00000605, 0.00000421, -0.00000579, 0.00000020, 0.00000002, 0.00000048, 0.00000171, -0.00000050, 0.00000158, 0.00000336, -0.00000559, 0.00000155, -0.00000263, 0.00000186, -0.00000200, -0.00000092, -0.00000057, 0.00000041, 0.00000048, 0.00000261, -0.00000147, 0.00000409, -0.00000330, 0.00000098, -0.00000081, -0.00000141, 0.00000152, -0.00000271, -0.00000035, -0.00000094, -0.00000179, 0.00000164, 0.00000127, -0.00000024, 0.00000455, -0.00000390, 0.00000529, -0.00000620, 0.00000274, -0.00000245, 0.00000078, -0.00000232, -0.00000311, -0.00000084, 0.00000695, -0.00000419, 0.00000647, -0.00000229, 0.00000235, -0.00000090, -0.00000757, 0.00000407, -0.00000572, 0.00000422, -0.00000220, 0.00000265, -0.00000092, 0.00000397, -0.00001070, 0.00000697, -0.00000620, 0.00001966, -0.00001419, 0.00000844, -0.00001169, 0.00000379, -0.00001194, 0.00000983, -0.00001188, 0.00001508, -0.00001294, 0.00001367, -0.00000853, 0.00000424, -0.00000319, -0.00000342, -0.00001763, 0.00001782, -0.00000021, -0.00000097, -0.00000265, -0.00001097, 0.00000483, -0.00001246, -0.00000506, -0.00003302, 0.00000163, -0.00002736, 0.00003505, -0.00008285, -0.00003667, 0.00020658, 0.00015943, 0.00007983, -0.00016045, -0.00000363, -0.00005476, 0.00000832, -0.00000537, -0.00002122, 0.00000601, -0.00000101, -0.00000752, 0.00001411, -0.00001705, 0.00000863, -0.00000279, -0.00000203, -0.00000351, 0.00000882, -0.00000581, },
		{ -0.00000316, 0.00000602, -0.00000283, -0.00000034, -0.00000674, 0.00000474, -0.00000705, 0.00000858, -0.00000176, 0.00000152, -0.00000440, -0.00000002, -0.00000067, 0.00000287, 0.00000086, 0.00000194, -0.00000599, 0.00000418, -0.00000581, 0.00000059, -0.00000039, 0.00000086, 0.00000151, -0.00000055, 0.00000182, 0.00000276, -0.00000520, 0.00000148, -0.00000261, 0.00000196, -0.00000212, -0.00000081, -0.00000056, 0.00000025, 0.00000084, 0.00000221, -0.00000118, 0.00000388, -0.00000335, 0.00000117, -0.00000100, -0.00000120, 0.00000148, -0.00000287, -0.00000011, -0.00000122, -0.00000158, 0.00000168, 0.00000111, 0.00000005, 0.00000423, -0.00000379, 0.00000537, -0.00000661, 0.00000333, -0.00000302, 0.00000131, -0.00000279, -0.00000296, -0.00000061, 0.00000704, -0.00000466, 0.00000718, -0.00000327, 0.00000344, -0.00000217, -0.00000666, 0.00000347, -0.00000506, 0.00000377, -0.00000185, 0.00000214, -0.00000013, 0.00000307, -0.00001023, 0.00000656, -0.00000505, 0.00001889, -0.00001448, 0.00000874, -0.00001202, 0.00000425, -0.00001254, 0.00001028, -0.00001192, 0.00001552, -0.00001381, 0.00001487, -0.00000999, 0.00000617, -0.00000526, -0.00000297, -0.00001748, 0.00001922, -0.00000239, 0.00000071, -0.00000458, -0.00000813, 0.00000187, -0.00001090, -0.00000522, -0.00003346, 0.00000375, -0.00002797, 0.00003531, -0.00008500, -0.00004352, 0.00021196, 0.00017147, 0.00007099, -0.00016475, 0.00000149, -0.00005898, 0.00001181, -0.00000701, -0.00002059, 0.00000612, -0.00000139, -0.00000658, 0.00001331, -0.00001751, 0.00000945, -0.00000279, -0.00000229, -0.00000346, 0.00000923, -0.00000642, -0.00000411, },
		{ 0.00000568, -0.00000227, -0.00000079, -0.00000657, 0.00000487, -0.00000693, 0.00000880, -0.00000238, 0.00000151, -0.00000425, 0.00000011, -0.00000097, 0.00000323, 0.00000089, 0.00000163, -0.00000591, 0.00000412, -0.00000579, 0.00000096, -0.00000081, 0.00000126, 0.00000125, -0.00000054, 0.00000200, 0.00000219, -0.00000481, 0.00000140, -0.00000258, 0.00000205, -0.00000225, -0.00000067, -0.00000058, 0.00000012, 0.00000116, 0.00000182, -0.00000089, 0.00000363, -0.00000336, 0.00000132, -0.00000118, -0.00000098, 0.00000141, -0.00000300, 0.00000011, -0.00000149, -0.00000134, 0.00000169, 0.00000097, 0.00000034, 0.00000387, -0.00000363, 0.00000539, -0.00000695, 0.00000387, -0.00000357, 0.00000183, -0.00000327, -0.00000277, -0.00000038, 0.00000707, -0.00000505, 0.00000781, -0.00000419, 0.00000451, -0.00000344, -0.00000568, 0.00000278, -0.00000431, 0.00000326, -0.00000144, 0.00000160, 0.00000068, 0.00000211, -0.00000968, 0.00000608, -0.00000379, 0.00001794, -0.00001459, 0.00000892, -0.00001224, 0.00000463, -0.00001305, 0.00001065, -0.00001185, 0.00001583, -0.00001459, 0.00001598, -0.00001140, 0.00000813, -0.00000746, -0.00000240, -0.00001729, 0.00002056, -0.00000462, 0.00000247, -0.00000662, -0.00000502, -0.00000152, -0.00000885, -0.00000589, -0.00003329, 0.00000537, -0.00002840, 0.00003529, -0.00008666, -0.00005097, 0.00021800, 0.00018401, 0.00006038, -0.00016767, 0.00000608, -0.00006301, 0.00001534, -0.00000890, -0.00001961, 0.00000597, -0.00000166, -0.00000569, 0.00001244, -0.00001784, 0.00001023, -0.00000285, -0.00000247, -0.00000343, 0.00000962, -0.00000707, -0.00000417, -0.00000287, },
		{ -0.00000170, -0.00000126, -0.00000637, 0.00000496, -0.00000677, 0.00000895, -0.00000296, 0.00000148, -0.00000407, 0.00000022, -0.00000125, 0.00000361, 0.00000089, 0.00000132, -0.00000580, 0.00000402, -0.00000572, 0.00000130, -0.00000122, 0.00000168, 0.00000095, -0.00000045, 0.00000210, 0.00000166, -0.00000441, 0.00000131, -0.00000254, 0.00000213, -0.00000236, -0.00000052, -0.00000064, 0.00000004, 0.00000144, 0.00000145, -0.00000059, 0.00000336, -0.00000334, 0.00000145, -0.00000135, -0.00000075, 0.00000131, -0.00000309, 0.00000031, -0.00000176, -0.00000109, 0.00000166, 0.00000085, 0.00000062, 0.00000349, -0.00000341, 0.00000533, -0.00000720, 0.00000436, -0.00000408, 0.00000235, -0.00000377, -0.00000255, -0.00000016, 0.00000706, -0.00000536, 0.00000837, -0.00000505, 0.00000553, -0.00000471, -0.00000465, 0.00000202, -0.00000348, 0.00000267, -0.00000100, 0.00000104, 0.00000152, 0.00000110, -0.00000904, 0.00000552, -0.00000242, 0.00001680, -0.00001452, 0.00000895, -0.00001234, 0.00000489, -0.00001346, 0.00001091, -0.00001165, 0.00001601, -0.00001525, 0.00001698, -0.00001272, 0.00001009, -0.00000976, -0.00000172, -0.00001706, 0.00002183, -0.00000686, 0.00000429, -0.00000873, -0.00000171, -0.00000530, -0.00000632, -0.00000708, -0.00003249, 0.00000642, -0.00002855, 0.00003489, -0.00008776, -0.00005911, 0.00022482, 0.00019690, 0.00004800, -0.00016911, 0.00001002, -0.00006675, 0.00001883, -0.00001099, -0.00001829, 0.00000555, -0.00000179, -0.00000484, 0.00001152, -0.00001805, 0.00001095, -0.00000296, -0.00000258, -0.00000343, 0.00001001, -0.00000777, -0.00000416, -0.00000259, 0.00000534, },
		{ -0.00000174, -0.00000612, 0.00000501, -0.00000655, 0.00000903, -0.00000351, 0.00000144, -0.00000385, 0.00000029, -0.00000152, 0.00000400, 0.00000085, 0.00000101, -0.00000567, 0.00000389, -0.00000561, 0.00000159, -0.00000163, 0.00000211, 0.00000061, -0.00000032, 0.00000214, 0.00000117, -0.00000401, 0.00000121, -0.00000249, 0.00000219, -0.00000248, -0.00000036, -0.00000074, 0.00000001, 0.00000168, 0.00000111, -0.00000030, 0.00000306, -0.00000328, 0.00000156, -0.00000151, -0.00000051, 0.00000118, -0.00000314, 0.00000048, -0.00000201, -0.00000083, 0.00000161, 0.00000076, 0.00000089, 0.00000309, -0.00000315, 0.00000521, -0.00000738, 0.00000478, -0.00000457, 0.00000286, -0.00000428, -0.00000231, 0.00000006, 0.00000699, -0.00000559, 0.00000883, -0.00000584, 0.00000649, -0.00000597, -0.00000358, 0.00000120, -0.00000260, 0.00000202, -0.00000052, 0.00000046, 0.00000237, 0.00000003, -0.00000833, 0.00000490, -0.00000096, 0.00001549, -0.00001426, 0.00000883, -0.00001231, 0.00000505, -0.00001377, 0.00001108, -0.00001133, 0.00001604, -0.00001576, 0.00001785, -0.00001394, 0.00001201, -0.00001215, -0.00000095, -0.00001680, 0.00002300, -0.00000908, 0.00000613, -0.00001088, 0.00000177, -0.00000943, -0.00000334, -0.00000880, -0.00003101, 0.00000683, -0.00002836, 0.00003407, -0.00008823, -0.00006801, 0.00023255, 0.00021001, 0.00003390, -0.00016898, 0.00001320, -0.00007008, 0.00002218, -0.00001322, -0.00001667, 0.00000487, -0.00000179, -0.00000406, 0.00001056, -0.00001814, 0.00001160, -0.00000312, -0.00000263, -0.00000344, 0.00001039, -0.00000852, -0.00000407, -0.00000234, 0.00000500, -0.00000113, },
		{ -0.00000585, 0.00000502, -0.00000631, 0.00000905, -0.00000402, 0.00000139, -0.00000361, 0.00000033, -0.00000176, 0.00000439, 0.00000079, 0.00000072, -0.00000551, 0.00000373, -0.00000544, 0.00000183, -0.00000200, 0.00000254, 0.00000025, -0.00000013, 0.00000211, 0.00000074, -0.00000361, 0.00000110, -0.00000243, 0.00000223, -0.00000258, -0.00000019, -0.00000086, 0.00000001, 0.00000188, 0.00000080, -0.00000002, 0.00000275, -0.00000318, 0.00000164, -0.00000165, -0.00000027, 0.00000102, -0.00000315, 0.00000062, -0.00000225, -0.00000056, 0.00000153, 0.00000070, 0.00000114, 0.00000268, -0.00000285, 0.00000502, -0.00000748, 0.00000514, -0.00000501, 0.00000334, -0.00000479, -0.00000203, 0.00000027, 0.00000688, -0.00000575, 0.00000920, -0.00000654, 0.00000739, -0.00000719, -0.00000249, 0.00000033, -0.00000165, 0.00000132, -0.00000001, -0.00000013, 0.00000323, -0.00000106, -0.00000756, 0.00000422, 0.00000058, 0.00001402, -0.00001382, 0.00000856, -0.00001215, 0.00000509, -0.00001397, 0.00001113, -0.00001086, 0.00001589, -0.00001613, 0.00001856, -0.00001502, 0.00001386, -0.00001459, -0.00000009, -0.00001648, 0.00002406, -0.00001127, 0.00000796, -0.00001302, 0.00000536, -0.00001384, 0.00000006, -0.00001101, -0.00002885, 0.00000654, -0.00002774, 0.00003275, -0.00008799, -0.00007778, 0.00024134, 0.00022317, 0.00001816, -0.00016724, 0.00001551, -0.00007291, 0.00002530, -0.00001551, -0.00001480, 0.00000395, -0.00000165, -0.00000335, 0.00000957, -0.00001811, 0.00001217, -0.00000330, -0.00000260, -0.00000348, 0.00001076, -0.00000932, -0.00000389, -0.00000213, 0.00000467, -0.00000057, -0.00000222, },
		{ 0.00000500, -0.00000602, 0.00000900, -0.00000449, 0.00000133, -0.00000334, 0.00000032, -0.00000196, 0.00000478, 0.00000070, 0.00000043, -0.00000534, 0.00000354, -0.00000522, 0.00000203, -0.00000235, 0.00000295, -0.00000013, 0.00000011, 0.00000201, 0.00000035, -0.00000322, 0.00000100, -0.00000235, 0.00000226, -0.00000266, -0.00000002, -0.00000101, 0.00000006, 0.00000203, 0.00000053, 0.00000025, 0.00000243, -0.00000305, 0.00000168, -0.00000177, -0.00000003, 0.00000084, -0.00000313, 0.00000073, -0.00000248, -0.00000029, 0.00000142, 0.00000067, 0.00000136, 0.00000227, -0.00000252, 0.00000477, -0.00000750, 0.00000543, -0.00000540, 0.00000379, -0.00000529, -0.00000173, 0.00000046, 0.00000674, -0.00000583, 0.00000948, -0.00000715, 0.00000820, -0.00000836, -0.00000139, -0.00000057, -0.00000065, 0.00000056, 0.00000054, -0.00000073, 0.00000409, -0.00000218, -0.00000673, 0.00000350, 0.00000218, 0.00001240, -0.00001320, 0.00000814, -0.00001184, 0.00000501, -0.00001404, 0.00001107, -0.00001025, 0.00001558, -0.00001632, 0.00001909, -0.00001594, 0.00001561, -0.00001705, 0.00000083, -0.00001611, 0.00002500, -0.00001339, 0.00000975, -0.00001511, 0.00000900, -0.00001849, 0.00000384, -0.00001371, -0.00002600, 0.00000551, -0.00002663, 0.00003087, -0.00008699, -0.00008850, 0.00025133, 0.00023618, 0.00000087, -0.00016384, 0.00001685, -0.00007512, 0.00002809, -0.00001777, -0.00001271, 0.00000280, -0.00000137, -0.00000274, 0.00000857, -0.00001795, 0.00001266, -0.00000351, -0.00000251, -0.00000355, 0.00001112, -0.00001016, -0.00000362, -0.00000197, 0.00000437, -0.00000002, -0.00000270, -0.00000556, },
		{ -0.00000572, 0.00000890, -0.00000492, 0.00000126, -0.00000305, 0.00000028, -0.00000213, 0.00000517, 0.00000058, 0.00000016, -0.00000516, 0.00000333, -0.00000496, 0.00000217, -0.00000266, 0.00000334, -0.00000052, 0.00000037, 0.00000186, 0.00000001, -0.00000285, 0.00000089, -0.00000228, 0.00000228, -0.00000273, 0.00000015, -0.00000118, 0.00000015, 0.00000213, 0.00000029, 0.00000049, 0.00000210, -0.00000290, 0.00000170, -0.00000187, 0.00000020, 0.00000063, -0.00000308, 0.00000081, -0.00000267, -0.00000002, 0.00000130, 0.00000067, 0.00000156, 0.00000186, -0.00000216, 0.00000447, -0.00000744, 0.00000565, -0.00000573, 0.00000421, -0.00000578, -0.00000140, 0.00000063, 0.00000656, -0.00000584, 0.00000967, -0.00000767, 0.00000892, -0.00000946, -0.00000030, -0.00000149, 0.00000038, -0.00000024, 0.00000112, -0.00000134, 0.00000493, -0.00000331, -0.00000586, 0.00000275, 0.00000383, 0.00001064, -0.00001241, 0.00000755, -0.00001138, 0.00000479, -0.00001399, 0.00001088, -0.00000949, 0.00001508, -0.00001632, 0.00001943, -0.00001666, 0.00001722, -0.00001949, 0.00000181, -0.00001568, 0.00002578, -0.00001540, 0.00001147, -0.00001711, 0.00001263, -0.00002330, 0.00000792, -0.00001685, -0.00002248, 0.00000371, -0.00002497, 0.00002838, -0.00008516, -0.00010027, 0.00026268, 0.00024885, -0.00001783, -0.00015878, 0.00001716, -0.00007663, 0.00003045, -0.00001993, -0.00001047, 0.00000145, -0.00000096, -0.00000221, 0.00000756, -0.00001768, 0.00001304, -0.00000372, -0.00000237, -0.00000363, 0.00001145, -0.00001105, -0.00000326, -0.00000186, 0.00000409, 0.00000050, -0.00000316, -0.00000525, 0.00000496, },
		{ 0.00000876, -0.00000532, 0.00000118, -0.00000273, 0.00000021, -0.00000226, 0.00000554, 0.00000044, -0.00000010, -0.00000496, 0.00000309, -0.00000465, 0.00000225, -0.00000292, 0.00000370, -0.00000091, 0.00000066, 0.00000166, -0.00000027, -0.00000249, 0.00000078, -0.00000219, 0.00000228, -0.00000278, 0.00000030, -0.00000136, 0.00000026, 0.00000219, 0.00000009, 0.00000071, 0.00000178, -0.00000273, 0.00000170, -0.00000194, 0.00000043, 0.00000042, -0.00000300, 0.00000085, -0.00000285, 0.00000025, 0.00000115, 0.00000070, 0.00000173, 0.00000146, -0.00000179, 0.00000411, -0.00000730, 0.00000579, -0.00000601, 0.00000458, -0.00000625, -0.00000106, 0.00000078, 0.00000636, -0.00000578, 0.00000976, -0.00000808, 0.00000954, -0.00001048, 0.00000076, -0.00000241, 0.00000143, -0.00000107, 0.00000172, -0.00000195, 0.00000576, -0.00000444, -0.00000495, 0.00000197, 0.00000551, 0.00000876, -0.00001146, 0.00000682, -0.00001079, 0.00000445, -0.00001381, 0.00001058, -0.00000859, 0.00001440, -0.00001613, 0.00001956, -0.00001718, 0.00001866, -0.00002190, 0.00000281, -0.00001517, 0.00002641, -0.00001730, 0.00001308, -0.00001898, 0.00001618, -0.00002820, 0.00001226, -0.00002038, -0.00001832, 0.00000112, -0.00002270, 0.00002522, -0.00008245, -0.00011315, 0.00027554, 0.00026095, -0.00003781, -0.00015207, 0.00001636, -0.00007736, 0.00003231, -0.00002191, -0.00000813, -0.00000008, -0.00000043, -0.00000177, 0.00000656, -0.00001729, 0.00001330, -0.00000392, -0.00000218, -0.00000372, 0.00001176, -0.00001197, -0.00000281, -0.00000182, 0.00000385, 0.00000098, -0.00000360, -0.00000492, 0.00000490, -0.00000539, },
		{ -0.00000568, 0.00000109, -0.00000240, 0.00000010, -0.00000234, 0.00000590, 0.00000028, -0.00000035, -0.00000476, 0.00000283, -0.00000430, 0.00000229, -0.00000314, 0.00000403, -0.00000128, 0.00000097, 0.00000141, -0.00000051, -0.00000215, 0.00000068, -0.00000211, 0.00000226, -0.00000281, 0.00000045, -0.00000155, 0.00000040, 0.00000220, -0.00000007, 0.00000090, 0.00000146, -0.00000253, 0.00000167, -0.00000199, 0.00000064, 0.00000019, -0.00000289, 0.00000086, -0.00000299, 0.00000051, 0.00000099, 0.00000076, 0.00000187, 0.00000107, -0.00000141, 0.00000372, -0.00000710, 0.00000585, -0.00000622, 0.00000491, -0.00000670, -0.00000070, 0.00000091, 0.00000613, -0.00000566, 0.00000976, -0.00000840, 0.00001005, -0.00001140, 0.00000178, -0.00000333, 0.00000249, -0.00000194, 0.00000234, -0.00000255, 0.00000657, -0.00000557, -0.00000402, 0.00000119, 0.00000721, 0.00000678, -0.00001035, 0.00000594, -0.00001006, 0.00000398, -0.00001350, 0.00001015, -0.00000754, 0.00001353, -0.00001573, 0.00001946, -0.00001746, 0.00001989, -0.00002422, 0.00000383, -0.00001459, 0.00002687, -0.00001905, 0.00001456, -0.00002066, 0.00001960, -0.00003311, 0.00001678, -0.00002424, -0.00001356, -0.00000225, -0.00001981, 0.00002136, -0.00007882, -0.00012721, 0.00029004, 0.00027226, -0.00005886, -0.00014375, 0.00001443, -0.00007724, 0.00003356, -0.00002362, -0.00000577, -0.00000173, 0.00000020, -0.00000142, 0.00000558, -0.00001678, 0.00001344, -0.00000409, -0.00000196, -0.00000382, 0.00001203, -0.00001291, -0.00000226, -0.00000184, 0.00000366, 0.00000142, -0.00000401, -0.00000460, 0.00000482, -0.00000506, 0.00000857, },
		{ 0.00000101, -0.00000206, -0.00000004, -0.00000239, 0.00000623, 0.00000010, -0.00000057, -0.00000455, 0.00000257, -0.00000391, 0.00000226, -0.00000330, 0.00000430, -0.00000164, 0.00000128, 0.00000112, -0.00000069, -0.00000183, 0.00000058, -0.00000202, 0.00000223, -0.00000281, 0.00000057, -0.00000173, 0.00000057, 0.00000218, -0.00000018, 0.00000105, 0.00000115, -0.00000232, 0.00000161, -0.00000201, 0.00000083, -0.00000004, -0.00000275, 0.00000084, -0.00000310, 0.00000076, 0.00000082, 0.00000084, 0.00000197, 0.00000071, -0.00000102, 0.00000330, -0.00000683, 0.00000585, -0.00000637, 0.00000519, -0.00000712, -0.00000033, 0.00000102, 0.00000589, -0.00000548, 0.00000967, -0.00000861, 0.00001045, -0.00001223, 0.00000275, -0.00000423, 0.00000356, -0.00000281, 0.00000298, -0.00000314, 0.00000734, -0.00000667, -0.00000307, 0.00000040, 0.00000890, 0.00000473, -0.00000910, 0.00000492, -0.00000919, 0.00000338, -0.00001307, 0.00000961, -0.00000636, 0.00001248, -0.00001513, 0.00001913, -0.00001749, 0.00002090, -0.00002643, 0.00000483, -0.00001392, 0.00002713, -0.00002063, 0.00001587, -0.00002214, 0.00002283, -0.00003797, 0.00002140, -0.00002834, -0.00000826, -0.00000637, -0.00001628, 0.00001679, -0.00007423, -0.00014252, 0.00030635, 0.00028253, -0.00008080, -0.00013391, 0.00001135, -0.00007621, 0.00003415, -0.00002498, -0.00000344, -0.00000347, 0.00000091, -0.00000115, 0.00000461, -0.00001617, 0.00001345, -0.00000423, -0.00000171, -0.00000391, 0.00001225, -0.00001387, -0.00000162, -0.00000193, 0.00000351, 0.00000181, -0.00000438, -0.00000428, 0.00000473, -0.00000472, 0.00000834, -0.00000602, },
		{ -0.00000172, -0.00000020, -0.00000238, 0.00000654, -0.00000011, -0.00000077, -0.00000434, 0.00000229, -0.00000350, 0.00000219, -0.00000342, 0.00000453, -0.00000196, 0.00000159, 0.00000080, -0.00000083, -0.00000153, 0.00000050, -0.00000193, 0.00000219, -0.00000280, 0.00000068, -0.00000192, 0.00000075, 0.00000212, -0.00000026, 0.00000117, 0.00000086, -0.00000210, 0.00000154, -0.00000200, 0.00000101, -0.00000027, -0.00000260, 0.00000078, -0.00000318, 0.00000099, 0.00000064, 0.00000095, 0.00000204, 0.00000037, -0.00000064, 0.00000285, -0.00000649, 0.00000577, -0.00000645, 0.00000540, -0.00000751, 0.00000006, 0.00000109, 0.00000564, -0.00000525, 0.00000950, -0.00000872, 0.00001073, -0.00001293, 0.00000365, -0.00000509, 0.00000460, -0.00000369, 0.00000361, -0.00000372, 0.00000807, -0.00000774, -0.00000213, -0.00000036, 0.00001057, 0.00000261, -0.00000773, 0.00000377, -0.00000820, 0.00000267, -0.00001251, 0.00000895, -0.00000505, 0.00001125, -0.00001432, 0.00001857, -0.00001726, 0.00002165, -0.00002849, 0.00000581, -0.00001315, 0.00002720, -0.00002201, 0.00001700, -0.00002336, 0.00002580, -0.00004268, 0.00002602, -0.00003262, -0.00000251, -0.00001122, -0.00001210, 0.00001147, -0.00006867, -0.00015912, 0.00032458, 0.00029153, -0.00010340, -0.00012262, 0.00000713, -0.00007423, 0.00003400, -0.00002591, -0.00000123, -0.00000526, 0.00000167, -0.00000096, 0.00000366, -0.00001546, 0.00001332, -0.00000430, -0.00000145, -0.00000400, 0.00001241, -0.00001482, -0.00000089, -0.00000210, 0.00000342, 0.00000215, -0.00000470, -0.00000397, 0.00000463, -0.00000438, 0.00000809, -0.00000633, 0.00000093, },
		{ -0.00000038, -0.00000234, 0.00000682, -0.00000033, -0.00000095, -0.00000414, 0.00000201, -0.00000306, 0.00000207, -0.00000348, 0.00000470, -0.00000225, 0.00000190, 0.00000045, -0.00000093, -0.00000126, 0.00000042, -0.00000184, 0.00000213, -0.00000276, 0.00000075, -0.00000209, 0.00000094, 0.00000203, -0.00000030, 0.00000126, 0.00000059, -0.00000188, 0.00000145, -0.00000197, 0.00000116, -0.00000051, -0.00000242, 0.00000070, -0.00000323, 0.00000120, 0.00000046, 0.00000107, 0.00000207, 0.00000007, -0.00000028, 0.00000239, -0.00000611, 0.00000562, -0.00000646, 0.00000556, -0.00000785, 0.00000044, 0.00000115, 0.00000538, -0.00000498, 0.00000925, -0.00000874, 0.00001089, -0.00001352, 0.00000446, -0.00000590, 0.00000562, -0.00000457, 0.00000424, -0.00000427, 0.00000875, -0.00000877, -0.00000120, -0.00000110, 0.00001219, 0.00000046, -0.00000626, 0.00000250, -0.00000709, 0.00000184, -0.00001183, 0.00000818, -0.00000361, 0.00000985, -0.00001330, 0.00001776, -0.00001676, 0.00002213, -0.00003038, 0.00000673, -0.00001229, 0.00002706, -0.00002319, 0.00001791, -0.00002430, 0.00002847, -0.00004716, 0.00003058, -0.00003698, 0.00000362, -0.00001672, -0.00000730, 0.00000544, -0.00006213, -0.00017703, 0.00034485, 0.00029900, -0.00012642, -0.00011003, 0.00000178, -0.00007130, 0.00003306, -0.00002634, 0.00000081, -0.00000704, 0.00000246, -0.00000083, 0.00000274, -0.00001464, 0.00001303, -0.00000431, -0.00000119, -0.00000406, 0.00001250, -0.00001576, -0.00000010, -0.00000233, 0.00000338, 0.00000242, -0.00000497, -0.00000368, 0.00000454, -0.00000405, 0.00000781, -0.00000661, 0.00000086, -0.00000138, },
		{ -0.00000225, 0.00000707, -0.00000056, -0.00000110, -0.00000394, 0.00000173, -0.00000261, 0.00000191, -0.00000348, 0.00000482, -0.00000251, 0.00000218, 0.00000009, -0.00000098, -0.00000101, 0.00000036, -0.00000176, 0.00000207, -0.00000270, 0.00000081, -0.00000225, 0.00000114, 0.00000191, -0.00000030, 0.00000130, 0.00000034, -0.00000165, 0.00000134, -0.00000191, 0.00000129, -0.00000073, -0.00000224, 0.00000059, -0.00000325, 0.00000139, 0.00000027, 0.00000122, 0.00000206, -0.00000020, 0.00000007, 0.00000192, -0.00000568, 0.00000541, -0.00000641, 0.00000565, -0.00000815, 0.00000083, 0.00000117, 0.00000513, -0.00000468, 0.00000893, -0.00000866, 0.00001094, -0.00001397, 0.00000519, -0.00000665, 0.00000660, -0.00000542, 0.00000487, -0.00000481, 0.00000938, -0.00000975, -0.00000028, -0.00000180, 0.00001375, -0.00000170, -0.00000470, 0.00000114, -0.00000587, 0.00000090, -0.00001103, 0.00000731, -0.00000207, 0.00000829, -0.00001209, 0.00001672, -0.00001598, 0.00002230, -0.00003206, 0.00000758, -0.00001131, 0.00002670, -0.00002413, 0.00001860, -0.00002493, 0.00003077, -0.00005136, 0.00003497, -0.00004132, 0.00001003, -0.00002281, -0.00000191, -0.00000131, -0.00005465, -0.00019626, 0.00036727, 0.00030472, -0.00014959, -0.00009626, -0.00000464, -0.00006740, 0.00003130, -0.00002621, 0.00000261, -0.00000877, 0.00000324, -0.00000074, 0.00000185, -0.00001372, 0.00001259, -0.00000422, -0.00000095, -0.00000408, 0.00001249, -0.00001666, 0.00000076, -0.00000262, 0.00000339, 0.00000264, -0.00000518, -0.00000340, 0.00000444, -0.00000372, 0.00000751, -0.00000688, 0.00000079, -0.00000104, -0.00000057, },
		{ 0.00000728, -0.00000080, -0.00000123, -0.00000375, 0.00000146, -0.00000214, 0.00000171, -0.00000344, 0.00000487, -0.00000271, 0.00000243, -0.00000028, -0.00000101, -0.00000078, 0.00000030, -0.00000168, 0.00000199, -0.00000262, 0.00000083, -0.00000239, 0.00000133, 0.00000177, -0.00000027, 0.00000131, 0.00000012, -0.00000142, 0.00000122, -0.00000183, 0.00000139, -0.00000094, -0.00000204, 0.00000046, -0.00000323, 0.00000155, 0.00000010, 0.00000138, 0.00000202, -0.00000044, 0.00000039, 0.00000146, -0.00000521, 0.00000514, -0.00000630, 0.00000569, -0.00000840, 0.00000121, 0.00000117, 0.00000488, -0.00000435, 0.00000855, -0.00000849, 0.00001087, -0.00001430, 0.00000581, -0.00000732, 0.00000753, -0.00000625, 0.00000547, -0.00000531, 0.00000994, -0.00001067, 0.00000060, -0.00000244, 0.00001523, -0.00000384, -0.00000309, -0.00000031, -0.00000457, -0.00000012, -0.00001014, 0.00000635, -0.00000044, 0.00000658, -0.00001068, 0.00001544, -0.00001492, 0.00002217, -0.00003351, 0.00000834, -0.00001022, 0.00002612, -0.00002484, 0.00001904, -0.00002522, 0.00003266, -0.00005517, 0.00003911, -0.00004555, 0.00001660, -0.00002940, 0.00000402, -0.00000871, -0.00004625, -0.00021679, 0.00039191, 0.00030843, -0.00017265, -0.00008150, -0.00001205, -0.00006255, 0.00002869, -0.00002547, 0.00000410, -0.00001039, 0.00000398, -0.00000068, 0.00000097, -0.00001270, 0.00001199, -0.00000404, -0.00000074, -0.00000407, 0.00001239, -0.00001750, 0.00000166, -0.00000297, 0.00000345, 0.00000280, -0.00000534, -0.00000315, 0.00000435, -0.00000342, 0.00000720, -0.00000713, 0.00000074, -0.00000072, -0.00000078, -0.00000212, },
		{ -0.00000106, -0.00000134, -0.00000357, 0.00000121, -0.00000168, 0.00000148, -0.00000335, 0.00000487, -0.00000287, 0.00000266, -0.00000065, -0.00000100, -0.00000057, 0.00000027, -0.00000161, 0.00000192, -0.00000252, 0.00000082, -0.00000250, 0.00000151, 0.00000162, -0.00000021, 0.00000128, -0.00000007, -0.00000121, 0.00000109, -0.00000173, 0.00000146, -0.00000114, -0.00000184, 0.00000030, -0.00000319, 0.00000168, -0.00000007, 0.00000156, 0.00000195, -0.00000063, 0.00000068, 0.00000100, -0.00000472, 0.00000482, -0.00000612, 0.00000565, -0.00000859, 0.00000158, 0.00000115, 0.00000464, -0.00000400, 0.00000812, -0.00000825, 0.00001069, -0.00001449, 0.00000633, -0.00000792, 0.00000839, -0.00000704, 0.00000605, -0.00000578, 0.00001045, -0.00001152, 0.00000145, -0.00000303, 0.00001661, -0.00000595, -0.00000143, -0.00000182, -0.00000318, -0.00000123, -0.00000915, 0.00000532, 0.00000127, 0.00000475, -0.00000910, 0.00001394, -0.00001358, 0.00002171, -0.00003471, 0.00000899, -0.00000901, 0.00002531, -0.00002529, 0.00001923, -0.00002516, 0.00003410, -0.00005855, 0.00004290, -0.00004955, 0.00002322, -0.00003639, 0.00001042, -0.00001672, -0.00003698, -0.00023857, 0.00041883, 0.00030994, -0.00019533, -0.00006593, -0.00002036, -0.00005679, 0.00002522, -0.00002406, 0.00000522, -0.00001186, 0.00000465, -0.00000062, 0.00000012, -0.00001160, 0.00001122, -0.00000374, -0.00000057, -0.00000399, 0.00001216, -0.00001827, 0.00000260, -0.00000338, 0.00000356, 0.00000290, -0.00000544, -0.00000293, 0.00000426, -0.00000313, 0.00000689, -0.00000737, 0.00000071, -0.00000042, -0.00000098, -0.00000195, 0.00000745, },
		{ -0.00000141, -0.00000341, 0.00000097, -0.00000122, 0.00000124, -0.00000323, 0.00000482, -0.00000297, 0.00000284, -0.00000101, -0.00000096, -0.00000038, 0.00000024, -0.00000155, 0.00000184, -0.00000241, 0.00000079, -0.00000259, 0.00000168, 0.00000145, -0.00000012, 0.00000122, -0.00000023, -0.00000100, 0.00000096, -0.00000160, 0.00000151, -0.00000133, -0.00000164, 0.00000013, -0.00000311, 0.00000179, -0.00000023, 0.00000174, 0.00000184, -0.00000078, 0.00000093, 0.00000056, -0.00000421, 0.00000446, -0.00000589, 0.00000555, -0.00000873, 0.00000193, 0.00000110, 0.00000442, -0.00000364, 0.00000764, -0.00000793, 0.00001040, -0.00001455, 0.00000674, -0.00000842, 0.00000918, -0.00000779, 0.00000660, -0.00000621, 0.00001088, -0.00001230, 0.00000224, -0.00000355, 0.00001788, -0.00000801, 0.00000025, -0.00000339, -0.00000174, -0.00000240, -0.00000809, 0.00000422, 0.00000304, 0.00000280, -0.00000736, 0.00001223, -0.00001198, 0.00002092, -0.00003562, 0.00000951, -0.00000768, 0.00002427, -0.00002549, 0.00001915, -0.00002475, 0.00003507, -0.00006143, 0.00004628, -0.00005323, 0.00002976, -0.00004366, 0.00001719, -0.00002527, -0.00002694, -0.00026155, 0.00044805, 0.00030902, -0.00021734, -0.00004975, -0.00002946, -0.00005018, 0.00002090, -0.00002198, 0.00000593, -0.00001313, 0.00000520, -0.00000056, -0.00000072, -0.00001041, 0.00001028, -0.00000331, -0.00000046, -0.00000385, 0.00001181, -0.00001895, 0.00000354, -0.00000382, 0.00000370, 0.00000294, -0.00000547, -0.00000273, 0.00000417, -0.00000286, 0.00000657, -0.00000760, 0.00000070, -0.00000015, -0.00000118, -0.00000175, 0.00000759, -0.00000132, },
		{ -0.00000326, 0.00000075, -0.00000077, 0.00000097, -0.00000306, 0.00000471, -0.00000303, 0.00000298, -0.00000135, -0.00000091, -0.00000021, 0.00000023, -0.00000149, 0.00000176, -0.00000228, 0.00000072, -0.00000264, 0.00000184, 0.00000129, -0.00000002, 0.00000113, -0.00000036, -0.00000080, 0.00000083, -0.00000146, 0.00000153, -0.00000149, -0.00000145, -0.00000005, -0.00000301, 0.00000187, -0.00000037, 0.00000192, 0.00000171, -0.00000090, 0.00000114, 0.00000015, -0.00000369, 0.00000406, -0.00000560, 0.00000540, -0.00000881, 0.00000227, 0.00000102, 0.00000422, -0.00000327, 0.00000713, -0.00000755, 0.00001002, -0.00001448, 0.00000703, -0.00000883, 0.00000988, -0.00000847, 0.00000711, -0.00000660, 0.00001124, -0.00001299, 0.00000299, -0.00000399, 0.00001903, -0.00000997, 0.00000193, -0.00000498, -0.00000025, -0.00000362, -0.00000696, 0.00000307, 0.00000485, 0.00000077, -0.00000548, 0.00001033, -0.00001012, 0.00001980, -0.00003625, 0.00000989, -0.00000622, 0.00002301, -0.00002542, 0.00001881, -0.00002396, 0.00003553, -0.00006376, 0.00004915, -0.00005649, 0.00003611, -0.00005109, 0.00002425, -0.00003425, -0.00001620, -0.00028563, 0.00047958, 0.00030550, -0.00023841, -0.00003318, -0.00003922, -0.00004278, 0.00001577, -0.00001918, 0.00000618, -0.00001415, 0.00000561, -0.00000046, -0.00000155, -0.00000914, 0.00000918, -0.00000275, -0.00000042, -0.00000364, 0.00001131, -0.00001951, 0.00000449, -0.00000429, 0.00000388, 0.00000293, -0.00000544, -0.00000255, 0.00000409, -0.00000261, 0.00000625, -0.00000781, 0.00000071, 0.00000009, -0.00000136, -0.00000152, 0.00000768, -0.00000158, -0.00000146, },
		{ 0.00000056, -0.00000034, 0.00000070, -0.00000287, 0.00000455, -0.00000303, 0.00000308, -0.00000168, -0.00000084, -0.00000005, 0.00000023, -0.00000145, 0.00000169, -0.00000215, 0.00000063, -0.00000267, 0.00000197, 0.00000112, 0.00000010, 0.00000101, -0.00000046, -0.00000062, 0.00000070, -0.00000131, 0.00000152, -0.00000163, -0.00000126, -0.00000024, -0.00000288, 0.00000192, -0.00000049, 0.00000211, 0.00000155, -0.00000097, 0.00000131, -0.00000023, -0.00000317, 0.00000363, -0.00000527, 0.00000518, -0.00000882, 0.00000258, 0.00000093, 0.00000404, -0.00000291, 0.00000658, -0.00000712, 0.00000955, -0.00001429, 0.00000721, -0.00000914, 0.00001050, -0.00000909, 0.00000758, -0.00000694, 0.00001152, -0.00001360, 0.00000367, -0.00000434, 0.00002003, -0.00001184, 0.00000357, -0.00000658, 0.00000127, -0.00000487, -0.00000578, 0.00000189, 0.00000668, -0.00000132, -0.00000349, 0.00000827, -0.00000802, 0.00001835, -0.00003657, 0.00001012, -0.00000464, 0.00002153, -0.00002510, 0.00001820, -0.00002281, 0.00003546, -0.00006548, 0.00005145, -0.00005922, 0.00004213, -0.00005854, 0.00003147, -0.00004357, -0.00000489, -0.00031070, 0.00051338, 0.00029922, -0.00025826, -0.00001644, -0.00004950, -0.00003470, 0.00000986, -0.00001569, 0.00000594, -0.00001490, 0.00000585, -0.00000030, -0.00000237, -0.00000780, 0.00000792, -0.00000204, -0.00000046, -0.00000334, 0.00001065, -0.00001993, 0.00000540, -0.00000477, 0.00000408, 0.00000287, -0.00000536, -0.00000241, 0.00000401, -0.00000239, 0.00000592, -0.00000802, 0.00000073, 0.00000030, -0.00000153, -0.00000126, 0.00000774, -0.00000185, -0.00000149, -0.00000314, },
		{ 0.00000006, 0.00000044, -0.00000266, 0.00000435, -0.00000298, 0.00000313, -0.00000198, -0.00000076, 0.00000009, 0.00000024, -0.00000142, 0.00000162, -0.00000201, 0.00000052, -0.00000266, 0.00000207, 0.00000096, 0.00000023, 0.00000087, -0.00000052, -0.00000046, 0.00000058, -0.00000115, 0.00000148, -0.00000175, -0.00000108, -0.00000044, -0.00000273, 0.00000194, -0.00000059, 0.00000228, 0.00000137, -0.00000100, 0.00000144, -0.00000057, -0.00000265, 0.00000318, -0.00000490, 0.00000490, -0.00000878, 0.00000286, 0.00000081, 0.00000388, -0.00000255, 0.00000603, -0.00000665, 0.00000900, -0.00001398, 0.00000727, -0.00000934, 0.00001101, -0.00000964, 0.00000799, -0.00000723, 0.00001173, -0.00001412, 0.00000429, -0.00000461, 0.00002088, -0.00001358, 0.00000517, -0.00000817, 0.00000279, -0.00000614, -0.00000456, 0.00000068, 0.00000851, -0.00000346, -0.00000140, 0.00000605, -0.00000571, 0.00001659, -0.00003658, 0.00001018, -0.00000295, 0.00001984, -0.00002451, 0.00001733, -0.00002130, 0.00003487, -0.00006657, 0.00005312, -0.00006135, 0.00004769, -0.00006588, 0.00003873, -0.00005310, 0.00000685, -0.00033661, 0.00054940, 0.00029005, -0.00027664, 0.00000023, -0.00006012, -0.00002603, 0.00000324, -0.00001149, 0.00000518, -0.00001532, 0.00000588, -0.00000006, -0.00000320, -0.00000638, 0.00000649, -0.00000119, -0.00000059, -0.00000295, 0.00000982, -0.00002020, 0.00000627, -0.00000525, 0.00000430, 0.00000277, -0.00000523, -0.00000229, 0.00000394, -0.00000218, 0.00000560, -0.00000821, 0.00000079, 0.00000047, -0.00000167, -0.00000099, 0.00000775, -0.00000211, -0.00000149, -0.00000303, 0.00000039, },
		{ 0.00000018, -0.00000243, 0.00000411, -0.00000289, 0.00000312, -0.00000224, -0.00000068, 0.00000022, 0.00000026, -0.00000139, 0.00000155, -0.00000187, 0.00000038, -0.00000261, 0.00000215, 0.00000082, 0.00000036, 0.00000070, -0.00000056, -0.00000032, 0.00000046, -0.00000097, 0.00000143, -0.00000184, -0.00000092, -0.00000063, -0.00000256, 0.00000193, -0.00000067, 0.00000245, 0.00000118, -0.00000100, 0.00000152, -0.00000087, -0.00000215, 0.00000272, -0.00000450, 0.00000458, -0.00000868, 0.00000312, 0.00000068, 0.00000375, -0.00000221, 0.00000546, -0.00000615, 0.00000839, -0.00001356, 0.00000722, -0.00000943, 0.00001142, -0.00001010, 0.00000835, -0.00000748, 0.00001185, -0.00001454, 0.00000484, -0.00000479, 0.00002158, -0.00001518, 0.00000670, -0.00000972, 0.00000431, -0.00000741, -0.00000332, -0.00000053, 0.00001032, -0.00000560, 0.00000075, 0.00000372, -0.00000321, 0.00001453, -0.00003628, 0.00001008, -0.00000113, 0.00001794, -0.00002368, 0.00001621, -0.00001944, 0.00003375, -0.00006699, 0.00005410, -0.00006279, 0.00005269, -0.00007296, 0.00004590, -0.00006270, 0.00001888, -0.00036318, 0.00058752, 0.00027788, -0.00029330, 0.00001660, -0.00007093, -0.00001691, -0.00000403, -0.00000664, 0.00000390, -0.00001540, 0.00000569, 0.00000028, -0.00000404, -0.00000491, 0.00000492, -0.00000018, -0.00000081, -0.00000246, 0.00000882, -0.00002029, 0.00000706, -0.00000572, 0.00000451, 0.00000265, -0.00000505, -0.00000219, 0.00000385, -0.00000199, 0.00000528, -0.00000839, 0.00000086, 0.00000060, -0.00000177, -0.00000070, 0.00000773, -0.00000237, -0.00000146, -0.00000294, 0.00000025, 0.00000044, },
		{ -0.00000220, 0.00000384, -0.00000276, 0.00000307, -0.00000247, -0.00000059, 0.00000035, 0.00000028, -0.00000138, 0.00000150, -0.00000172, 0.00000022, -0.00000254, 0.00000219, 0.00000069, 0.00000049, 0.00000052, -0.00000057, -0.00000020, 0.00000036, -0.00000080, 0.00000135, -0.00000191, -0.00000078, -0.00000082, -0.00000238, 0.00000189, -0.00000073, 0.00000261, 0.00000097, -0.00000096, 0.00000155, -0.00000113, -0.00000168, 0.00000225, -0.00000408, 0.00000421, -0.00000852, 0.00000334, 0.00000054, 0.00000364, -0.00000189, 0.00000490, -0.00000562, 0.00000772, -0.00001304, 0.00000707, -0.00000942, 0.00001173, -0.00001048, 0.00000864, -0.00000766, 0.00001190, -0.00001486, 0.00000532, -0.00000488, 0.00002210, -0.00001663, 0.00000813, -0.00001123, 0.00000579, -0.00000867, -0.00000208, -0.00000172, 0.00001208, -0.00000774, 0.00000294, 0.00000129, -0.00000054, 0.00001218, -0.00003566, 0.00000979, 0.00000079, 0.00001585, -0.00002260, 0.00001485, -0.00001726, 0.00003211, -0.00006674, 0.00005436, -0.00006348, 0.00005700, -0.00007964, 0.00005284, -0.00007224, 0.00003103, -0.00039023, 0.00062761, 0.00026265, -0.00030800, 0.00003243, -0.00008174, -0.00000746, -0.00001183, -0.00000116, 0.00000209, -0.00001512, 0.00000524, 0.00000074, -0.00000489, -0.00000337, 0.00000319, 0.00000097, -0.00000114, -0.00000187, 0.00000763, -0.00002019, 0.00000777, -0.00000615, 0.00000471, 0.00000250, -0.00000482, -0.00000211, 0.00000376, -0.00000181, 0.00000496, -0.00000856, 0.00000095, 0.00000068, -0.00000184, -0.00000039, 0.00000766, -0.00000262, -0.00000142, -0.00000287, 0.00000014, 0.00000077, -0.00000007, },
		{ 0.00000354, -0.00000259, 0.00000298, -0.00000267, -0.00000051, 0.00000047, 0.00000032, -0.00000138, 0.00000146, -0.00000159, 0.00000005, -0.00000244, 0.00000221, 0.00000058, 0.00000061, 0.00000033, -0.00000055, -0.00000010, 0.00000027, -0.00000062, 0.00000125, -0.00000195, -0.00000065, -0.00000101, -0.00000218, 0.00000184, -0.00000076, 0.00000275, 0.00000075, -0.00000090, 0.00000154, -0.00000133, -0.00000123, 0.00000179, -0.00000364, 0.00000380, -0.00000830, 0.00000352, 0.00000039, 0.00000356, -0.00000160, 0.00000434, -0.00000509, 0.00000702, -0.00001243, 0.00000681, -0.00000931, 0.00001193, -0.00001077, 0.00000887, -0.00000780, 0.00001186, -0.00001509, 0.00000573, -0.00000488, 0.00002246, -0.00001790, 0.00000946, -0.00001266, 0.00000722, -0.00000988, -0.00000085, -0.00000289, 0.00001378, -0.00000983, 0.00000514, -0.00000119, 0.00000226, 0.00000957, -0.00003474, 0.00000934, 0.00000282, 0.00001358, -0.00002129, 0.00001327, -0.00001477, 0.00002996, -0.00006581, 0.00005386, -0.00006334, 0.00006052, -0.00008580, 0.00005941, -0.00008156, 0.00004311, -0.00041754, 0.00066951, 0.00024433, -0.00032054, 0.00004750, -0.00009237, 0.00000217, -0.00002008, 0.00000489, -0.00000025, -0.00001447, 0.00000452, 0.00000133, -0.00000577, -0.00000179, 0.00000134, 0.00000228, -0.00000157, -0.00000119, 0.00000626, -0.00001988, 0.00000837, -0.00000653, 0.00000489, 0.00000235, -0.00000457, -0.00000204, 0.00000366, -0.00000164, 0.00000464, -0.00000871, 0.00000105, 0.00000073, -0.00000187, -0.00000009, 0.00000755, -0.00000287, -0.00000135, -0.00000281, 0.00000007, 0.00000107, -0.00000028, -0.00000197, },
		{ -0.00000239, 0.00000284, -0.00000282, -0.00000044, 0.00000058, 0.00000036, -0.00000139, 0.00000143, -0.00000146, -0.00000013, -0.00000230, 0.00000219, 0.00000049, 0.00000072, 0.00000013, -0.00000052, -0.00000002, 0.00000020, -0.00000044, 0.00000114, -0.00000197, -0.00000055, -0.00000118, -0.00000198, 0.00000176, -0.00000077, 0.00000286, 0.00000054, -0.00000081, 0.00000149, -0.00000148, -0.00000082, 0.00000135, -0.00000319, 0.00000336, -0.00000802, 0.00000367, 0.00000023, 0.00000351, -0.00000133, 0.00000380, -0.00000455, 0.00000628, -0.00001175, 0.00000646, -0.00000910, 0.00001203, -0.00001097, 0.00000904, -0.00000788, 0.00001175, -0.00001522, 0.00000607, -0.00000479, 0.00002265, -0.00001900, 0.00001067, -0.00001401, 0.00000859, -0.00001105, 0.00000035, -0.00000402, 0.00001540, -0.00001186, 0.00000732, -0.00000370, 0.00000515, 0.00000674, -0.00003352, 0.00000870, 0.00000493, 0.00001115, -0.00001977, 0.00001149, -0.00001201, 0.00002735, -0.00006421, 0.00005260, -0.00006233, 0.00006316, -0.00009128, 0.00006546, -0.00009051, 0.00005494, -0.00044486, 0.00071301, 0.00022292, -0.00033075, 0.00006159, -0.00010263, 0.00001183, -0.00002866, 0.00001144, -0.00000308, -0.00001345, 0.00000352, 0.00000207, -0.00000668, -0.00000018, -0.00000064, 0.00000374, -0.00000210, -0.00000040, 0.00000470, -0.00001935, 0.00000883, -0.00000683, 0.00000502, 0.00000220, -0.00000429, -0.00000198, 0.00000354, -0.00000147, 0.00000430, -0.00000884, 0.00000117, 0.00000073, -0.00000186, 0.00000022, 0.00000740, -0.00000310, -0.00000127, -0.00000277, 0.00000002, 0.00000132, -0.00000047, -0.00000175, 0.00000323, },
		{ 0.00000265, -0.00000293, -0.00000037, 0.00000070, 0.00000040, -0.00000140, 0.00000141, -0.00000135, -0.00000032, -0.00000215, 0.00000213, 0.00000043, 0.00000082, -0.00000007, -0.00000046, 0.00000004, 0.00000014, -0.00000027, 0.00000101, -0.00000196, -0.00000047, -0.00000135, -0.00000178, 0.00000166, -0.00000075, 0.00000296, 0.00000032, -0.00000070, 0.00000140, -0.00000158, -0.00000045, 0.00000092, -0.00000273, 0.00000289, -0.00000770, 0.00000378, 0.00000007, 0.00000348, -0.00000110, 0.00000329, -0.00000403, 0.00000553, -0.00001100, 0.00000603, -0.00000880, 0.00001202, -0.00001107, 0.00000913, -0.00000790, 0.00001156, -0.00001526, 0.00000633, -0.00000461, 0.00002266, -0.00001990, 0.00001173, -0.00001525, 0.00000989, -0.00001215, 0.00000151, -0.00000508, 0.00001691, -0.00001380, 0.00000945, -0.00000620, 0.00000810, 0.00000370, -0.00003201, 0.00000791, 0.00000712, 0.00000858, -0.00001805, 0.00000954, -0.00000902, 0.00002429, -0.00006196, 0.00005056, -0.00006043, 0.00006483, -0.00009598, 0.00007086, -0.00009892, 0.00006631, -0.00047195, 0.00075788, 0.00019848, -0.00033847, 0.00007450, -0.00011235, 0.00002136, -0.00003743, 0.00001841, -0.00000637, -0.00001206, 0.00000224, 0.00000297, -0.00000762, 0.00000147, -0.00000273, 0.00000533, -0.00000274, 0.00000048, 0.00000295, -0.00001859, 0.00000915, -0.00000706, 0.00000510, 0.00000207, -0.00000400, -0.00000192, 0.00000339, -0.00000130, 0.00000396, -0.00000894, 0.00000129, 0.00000070, -0.00000179, 0.00000053, 0.00000721, -0.00000331, -0.00000117, -0.00000275, 0.00000000, 0.00000153, -0.00000063, -0.00000154, 0.00000291, -0.00000217, },
		{ -0.00000300, -0.00000032, 0.00000082, 0.00000044, -0.00000142, 0.00000140, -0.00000125, -0.00000051, -0.00000197, 0.00000205, 0.00000039, 0.00000089, -0.00000026, -0.00000038, 0.00000008, 0.00000011, -0.00000011, 0.00000087, -0.00000194, -0.00000041, -0.00000149, -0.00000157, 0.00000154, -0.00000071, 0.00000303, 0.00000011, -0.00000057, 0.00000127, -0.00000162, -0.00000012, 0.00000052, -0.00000229, 0.00000240, -0.00000733, 0.00000384, -0.00000008, 0.00000347, -0.00000090, 0.00000280, -0.00000353, 0.00000478, -0.00001020, 0.00000553, -0.00000841, 0.00001191, -0.00001109, 0.00000915, -0.00000787, 0.00001129, -0.00001520, 0.00000653, -0.00000435, 0.00002251, -0.00002061, 0.00001264, -0.00001637, 0.00001108, -0.00001317, 0.00000262, -0.00000608, 0.00001830, -0.00001562, 0.00001150, -0.00000865, 0.00001107, 0.00000050, -0.00003024, 0.00000695, 0.00000937, 0.00000589, -0.00001616, 0.00000743, -0.00000583, 0.00002085, -0.00005908, 0.00004777, -0.00005763, 0.00006547, -0.00009977, 0.00007547, -0.00010664, 0.00007701, -0.00049851, 0.00080385, 0.00017109, -0.00034357, 0.00008604, -0.00012136, 0.00003061, -0.00004627, 0.00002570, -0.00001008, -0.00001031, 0.00000068, 0.00000403, -0.00000859, 0.00000313, -0.00000490, 0.00000704, -0.00000347, 0.00000145, 0.00000102, -0.00001760, 0.00000930, -0.00000718, 0.00000511, 0.00000197, -0.00000370, -0.00000185, 0.00000322, -0.00000112, 0.00000360, -0.00000901, 0.00000140, 0.00000063, -0.00000168, 0.00000082, 0.00000699, -0.00000351, -0.00000106, -0.00000273, 0.00000001, 0.00000169, -0.00000075, -0.00000136, 0.00000258, -0.00000192, 0.00000243, },
		{ -0.00000029, 0.00000094, 0.00000048, -0.00000144, 0.00000141, -0.00000118, -0.00000070, -0.00000178, 0.00000193, 0.00000038, 0.00000095, -0.00000045, -0.00000030, 0.00000011, 0.00000009, 0.00000004, 0.00000073, -0.00000189, -0.00000038, -0.00000161, -0.00000137, 0.00000141, -0.00000064, 0.00000308, -0.00000008, -0.00000043, 0.00000111, -0.00000162, 0.00000016, 0.00000015, -0.00000186, 0.00000190, -0.00000691, 0.00000387, -0.00000024, 0.00000348, -0.00000073, 0.00000234, -0.00000305, 0.00000403, -0.00000937, 0.00000497, -0.00000795, 0.00001170, -0.00001101, 0.00000910, -0.00000779, 0.00001096, -0.00001505, 0.00000666, -0.00000402, 0.00002220, -0.00002112, 0.00001339, -0.00001736, 0.00001217, -0.00001409, 0.00000365, -0.00000699, 0.00001955, -0.00001731, 0.00001345, -0.00001103, 0.00001402, -0.00000283, -0.00002823, 0.00000584, 0.00001167, 0.00000310, -0.00001411, 0.00000520, -0.00000248, 0.00001705, -0.00005563, 0.00004425, -0.00005391, 0.00006503, -0.00010256, 0.00007917, -0.00011352, 0.00008685, -0.00052428, 0.00085061, 0.00014088, -0.00034598, 0.00009604, -0.00012947, 0.00003942, -0.00005503, 0.00003321, -0.00001415, -0.00000823, -0.00000116, 0.00000527, -0.00000960, 0.00000480, -0.00000715, 0.00000887, -0.00000429, 0.00000250, -0.00000107, -0.00001636, 0.00000928, -0.00000719, 0.00000504, 0.00000190, -0.00000341, -0.00000178, 0.00000300, -0.00000092, 0.00000322, -0.00000904, 0.00000151, 0.00000054, -0.00000153, 0.00000111, 0.00000673, -0.00000369, -0.00000094, -0.00000272, 0.00000004, 0.00000181, -0.00000082, -0.00000121, 0.00000227, -0.00000167, 0.00000218, -0.00000302, },
		{ 0.00000107, 0.00000051, -0.00000147, 0.00000143, -0.00000112, -0.00000087, -0.00000157, 0.00000180, 0.00000040, 0.00000098, -0.00000062, -0.00000020, 0.00000012, 0.00000009, 0.00000017, 0.00000058, -0.00000183, -0.00000037, -0.00000172, -0.00000118, 0.00000127, -0.00000056, 0.00000310, -0.00000027, -0.00000029, 0.00000093, -0.00000156, 0.00000039, -0.00000018, -0.00000145, 0.00000139, -0.00000646, 0.00000386, -0.00000039, 0.00000351, -0.00000059, 0.00000192, -0.00000261, 0.00000331, -0.00000851, 0.00000436, -0.00000742, 0.00001141, -0.00001084, 0.00000898, -0.00000766, 0.00001056, -0.00001481, 0.00000674, -0.00000362, 0.00002173, -0.00002143, 0.00001397, -0.00001821, 0.00001314, -0.00001491, 0.00000460, -0.00000781, 0.00002065, -0.00001883, 0.00001526, -0.00001330, 0.00001690, -0.00000624, -0.00002601, 0.00000460, 0.00001400, 0.00000024, -0.00001194, 0.00000288, 0.00000098, 0.00001297, -0.00005165, 0.00004003, -0.00004929, 0.00006348, -0.00010426, 0.00008184, -0.00011940, 0.00009561, -0.00054895, 0.00089786, 0.00010801, -0.00034565, 0.00010437, -0.00013656, 0.00004763, -0.00006358, 0.00004083, -0.00001851, -0.00000586, -0.00000326, 0.00000667, -0.00001065, 0.00000647, -0.00000945, 0.00001080, -0.00000520, 0.00000362, -0.00000333, -0.00001489, 0.00000907, -0.00000707, 0.00000488, 0.00000189, -0.00000313, -0.00000169, 0.00000275, -0.00000069, 0.00000281, -0.00000903, 0.00000160, 0.00000042, -0.00000133, 0.00000138, 0.00000643, -0.00000384, -0.00000082, -0.00000272, 0.00000009, 0.00000189, -0.00000086, -0.00000109, 0.00000196, -0.00000142, 0.00000191, -0.00000301, -0.00000027, },
		{ 0.00000054, -0.00000149, 0.00000146, -0.00000109, -0.00000104, -0.00000136, 0.00000163, 0.00000044, 0.00000098, -0.00000078, -0.00000010, 0.00000012, 0.00000011, 0.00000029, 0.00000043, -0.00000176, -0.00000038, -0.00000180, -0.00000100, 0.00000113, -0.00000046, 0.00000309, -0.00000043, -0.00000014, 0.00000072, -0.00000147, 0.00000058, -0.00000048, -0.00000106, 0.00000089, -0.00000598, 0.00000380, -0.00000053, 0.00000356, -0.00000049, 0.00000154, -0.00000221, 0.00000262, -0.00000764, 0.00000371, -0.00000683, 0.00001103, -0.00001058, 0.00000879, -0.00000747, 0.00001010, -0.00001449, 0.00000675, -0.00000317, 0.00002111, -0.00002154, 0.00001437, -0.00001891, 0.00001398, -0.00001561, 0.00000546, -0.00000853, 0.00002159, -0.00002018, 0.00001691, -0.00001544, 0.00001969, -0.00000970, -0.00002360, 0.00000324, 0.00001633, -0.00000266, -0.00000966, 0.00000049, 0.00000451, 0.00000865, -0.00004720, 0.00003516, -0.00004382, 0.00006081, -0.00010481, 0.00008337, -0.00012415, 0.00010311, -0.00057223, 0.00094524, 0.00007268, -0.00034257, 0.00011091, -0.00014246, 0.00005511, -0.00007177, 0.00004844, -0.00002311, -0.00000323, -0.00000560, 0.00000824, -0.00001173, 0.00000811, -0.00001178, 0.00001282, -0.00000617, 0.00000478, -0.00000572, -0.00001318, 0.00000866, -0.00000681, 0.00000462, 0.00000194, -0.00000287, -0.00000158, 0.00000244, -0.00000044, 0.00000237, -0.00000896, 0.00000167, 0.00000029, -0.00000108, 0.00000163, 0.00000610, -0.00000396, -0.00000071, -0.00000271, 0.00000016, 0.00000193, -0.00000085, -0.00000100, 0.00000168, -0.00000118, 0.00000162, -0.00000297, -0.00000026, 0.00000120, },
		{ -0.00000152, 0.00000150, -0.00000108, -0.00000118, -0.00000114, 0.00000145, 0.00000051, 0.00000096, -0.00000091, -0.00000001, 0.00000011, 0.00000015, 0.00000039, 0.00000029, -0.00000167, -0.00000041, -0.00000186, -0.00000083, 0.00000098, -0.00000034, 0.00000305, -0.00000058, 0.00000000, 0.00000050, -0.00000133, 0.00000071, -0.00000073, -0.00000071, 0.00000039, -0.00000547, 0.00000371, -0.00000065, 0.00000362, -0.00000043, 0.00000120, -0.00000185, 0.00000197, -0.00000678, 0.00000305, -0.00000620, 0.00001058, -0.00001025, 0.00000854, -0.00000725, 0.00000958, -0.00001410, 0.00000672, -0.00000267, 0.00002037, -0.00002146, 0.00001460, -0.00001945, 0.00001468, -0.00001618, 0.00000622, -0.00000914, 0.00002235, -0.00002133, 0.00001838, -0.00001740, 0.00002234, -0.00001315, -0.00002104, 0.00000179, 0.00001865, -0.00000559, -0.00000731, -0.00000193, 0.00000804, 0.00000417, -0.00004234, 0.00002972, -0.00003754, 0.00005703, -0.00010417, 0.00008369, -0.00012765, 0.00010915, -0.00059381, 0.00099238, 0.00003515, -0.00033675, 0.00011556, -0.00014707, 0.00006172, -0.00007946, 0.00005593, -0.00002784, -0.00000039, -0.00000816, 0.00000996, -0.00001284, 0.00000973, -0.00001412, 0.00001489, -0.00000720, 0.00000598, -0.00000824, -0.00001125, 0.00000805, -0.00000641, 0.00000424, 0.00000207, -0.00000266, -0.00000143, 0.00000209, -0.00000015, 0.00000189, -0.00000884, 0.00000170, 0.00000016, -0.00000079, 0.00000187, 0.00000573, -0.00000405, -0.00000059, -0.00000270, 0.00000023, 0.00000193, -0.00000080, -0.00000096, 0.00000141, -0.00000094, 0.00000131, -0.00000289, -0.00000028, 0.00000134, 0.00000056, },
		{ 0.00000155, -0.00000110, -0.00000131, -0.00000093, 0.00000126, 0.00000060, 0.00000090, -0.00000102, 0.00000009, 0.00000009, 0.00000020, 0.00000047, 0.00000015, -0.00000158, -0.00000045, -0.00000190, -0.00000068, 0.00000084, -0.00000020, 0.00000299, -0.00000070, 0.00000013, 0.00000028, -0.00000115, 0.00000080, -0.00000093, -0.00000040, -0.00000008, -0.00000495, 0.00000358, -0.00000077, 0.00000368, -0.00000039, 0.00000090, -0.00000154, 0.00000137, -0.00000594, 0.00000237, -0.00000554, 0.00001007, -0.00000985, 0.00000822, -0.00000698, 0.00000902, -0.00001363, 0.00000664, -0.00000213, 0.00001950, -0.00002121, 0.00001466, -0.00001983, 0.00001524, -0.00001662, 0.00000688, -0.00000964, 0.00002294, -0.00002228, 0.00001965, -0.00001917, 0.00002481, -0.00001657, -0.00001836, 0.00000026, 0.00002094, -0.00000851, -0.00000491, -0.00000435, 0.00001155, -0.00000041, -0.00003716, 0.00002377, -0.00003052, 0.00005217, -0.00010230, 0.00008274, -0.00012980, 0.00011355, -0.00061342, 0.00103892, -0.00000433, -0.00032827, 0.00011827, -0.00015029, 0.00006733, -0.00008653, 0.00006317, -0.00003264, 0.00000260, -0.00001089, 0.00001182, -0.00001398, 0.00001129, -0.00001645, 0.00001701, -0.00000826, 0.00000719, -0.00001085, -0.00000910, 0.00000725, -0.00000585, 0.00000374, 0.00000228, -0.00000248, -0.00000125, 0.00000167, 0.00000018, 0.00000137, -0.00000865, 0.00000169, 0.00000002, -0.00000047, 0.00000209, 0.00000533, -0.00000411, -0.00000048, -0.00000267, 0.00000031, 0.00000190, -0.00000071, -0.00000095, 0.00000118, -0.00000072, 0.00000101, -0.00000278, -0.00000030, 0.00000149, 0.00000057, -0.00000153, },
		{ -0.00000114, -0.00000141, -0.00000073, 0.00000106, 0.00000071, 0.00000082, -0.00000111, 0.00000018, 0.00000007, 0.00000027, 0.00000054, 0.00000002, -0.00000148, -0.00000052, -0.00000192, -0.00000054, 0.00000070, -0.00000006, 0.00000290, -0.00000080, 0.00000026, 0.00000004, -0.00000095, 0.00000084, -0.00000109, -0.00000013, -0.00000054, -0.00000441, 0.00000342, -0.00000086, 0.00000375, -0.00000037, 0.00000064, -0.00000127, 0.00000082, -0.00000512, 0.00000170, -0.00000486, 0.00000951, -0.00000938, 0.00000785, -0.00000667, 0.00000841, -0.00001310, 0.00000652, -0.00000157, 0.00001853, -0.00002077, 0.00001454, -0.00002005, 0.00001565, -0.00001692, 0.00000742, -0.00001002, 0.00002334, -0.00002301, 0.00002071, -0.00002071, 0.00002707, -0.00001990, -0.00001559, -0.00000133, 0.00002317, -0.00001140, -0.00000248, -0.00000674, 0.00001498, -0.00000502, -0.00003174, 0.00001740, -0.00002283, 0.00004626, -0.00009920, 0.00008045, -0.00013050, 0.00011617, -0.00063075, 0.00108446, -0.00004544, -0.00031722, 0.00011901, -0.00015203, 0.00007184, -0.00009283, 0.00007005, -0.00003740, 0.00000569, -0.00001377, 0.00001379, -0.00001513, 0.00001280, -0.00001874, 0.00001914, -0.00000935, 0.00000840, -0.00001354, -0.00000676, 0.00000625, -0.00000514, 0.00000311, 0.00000259, -0.00000236, -0.00000104, 0.00000119, 0.00000057, 0.00000080, -0.00000839, 0.00000162, -0.00000010, -0.00000012, 0.00000229, 0.00000490, -0.00000414, -0.00000039, -0.00000264, 0.00000038, 0.00000185, -0.00000058, -0.00000099, 0.00000098, -0.00000053, 0.00000070, -0.00000265, -0.00000034, 0.00000165, 0.00000057, -0.00000154, 0.00000161, },
		{ -0.00000148, -0.00000054, 0.00000085, 0.00000083, 0.00000072, -0.00000117, 0.00000025, 0.00000004, 0.00000035, 0.00000058, -0.00000009, -0.00000139, -0.00000059, -0.00000191, -0.00000043, 0.00000056, 0.00000008, 0.00000279, -0.00000087, 0.00000036, -0.00000019, -0.00000072, 0.00000085, -0.00000120, 0.00000010, -0.00000096, -0.00000387, 0.00000323, -0.00000094, 0.00000383, -0.00000039, 0.00000041, -0.00000106, 0.00000034, -0.00000434, 0.00000104, -0.00000417, 0.00000891, -0.00000885, 0.00000742, -0.00000633, 0.00000776, -0.00001252, 0.00000637, -0.00000099, 0.00001747, -0.00002018, 0.00001426, -0.00002012, 0.00001592, -0.00001709, 0.00000785, -0.00001028, 0.00002355, -0.00002352, 0.00002154, -0.00002202, 0.00002909, -0.00002311, -0.00001277, -0.00000295, 0.00002532, -0.00001422, -0.00000006, -0.00000907, 0.00001828, -0.00000960, -0.00002614, 0.00001070, -0.00001457, 0.00003939, -0.00009490, 0.00007682, -0.00012969, 0.00011686, -0.00064555, 0.00112861, -0.00008786, -0.00030374, 0.00011778, -0.00015224, 0.00007516, -0.00009827, 0.00007644, -0.00004204, 0.00000881, -0.00001674, 0.00001587, -0.00001629, 0.00001423, -0.00002096, 0.00002127, -0.00001044, 0.00000958, -0.00001627, -0.00000425, 0.00000505, -0.00000428, 0.00000236, 0.00000299, -0.00000229, -0.00000078, 0.00000065, 0.00000100, 0.00000017, -0.00000805, 0.00000149, -0.00000019, 0.00000025, 0.00000247, 0.00000444, -0.00000412, -0.00000031, -0.00000258, 0.00000044, 0.00000178, -0.00000043, -0.00000106, 0.00000081, -0.00000037, 0.00000040, -0.00000250, -0.00000040, 0.00000182, 0.00000056, -0.00000154, 0.00000166, -0.00000121, },
		{ -0.00000037, 0.00000065, 0.00000097, 0.00000059, -0.00000120, 0.00000031, 0.00000002, 0.00000044, 0.00000061, -0.00000020, -0.00000129, -0.00000068, -0.00000189, -0.00000033, 0.00000044, 0.00000023, 0.00000266, -0.00000091, 0.00000045, -0.00000041, -0.00000047, 0.00000081, -0.00000127, 0.00000029, -0.00000135, -0.00000333, 0.00000301, -0.00000100, 0.00000390, -0.00000042, 0.00000023, -0.00000090, -0.00000008, -0.00000361, 0.00000041, -0.00000349, 0.00000827, -0.00000828, 0.00000695, -0.00000597, 0.00000709, -0.00001188, 0.00000619, -0.00000041, 0.00001633, -0.00001944, 0.00001381, -0.00002002, 0.00001605, -0.00001711, 0.00000816, -0.00001042, 0.00002359, -0.00002381, 0.00002213, -0.00002307, 0.00003083, -0.00002615, -0.00000994, -0.00000458, 0.00002738, -0.00001695, 0.00000232, -0.00001132, 0.00002142, -0.00001407, -0.00002047, 0.00000377, -0.00000584, 0.00003164, -0.00008942, 0.00007184, -0.00012733, 0.00011552, -0.00065757, 0.00117098, -0.00013124, -0.00028799, 0.00011461, -0.00015091, 0.00007722, -0.00010273, 0.00008224, -0.00004646, 0.00001189, -0.00001976, 0.00001801, -0.00001744, 0.00001557, -0.00002310, 0.00002337, -0.00001151, 0.00001071, -0.00001902, -0.00000158, 0.00000368, -0.00000326, 0.00000148, 0.00000349, -0.00000229, -0.00000048, 0.00000004, 0.00000149, -0.00000051, -0.00000763, 0.00000130, -0.00000025, 0.00000063, 0.00000264, 0.00000394, -0.00000407, -0.00000026, -0.00000250, 0.00000049, 0.00000170, -0.00000024, -0.00000117, 0.00000067, -0.00000024, 0.00000012, -0.00000233, -0.00000046, 0.00000199, 0.00000053, -0.00000153, 0.00000171, -0.00000130, -0.00000152, },
		{ 0.00000045, 0.00000110, 0.00000044, -0.00000120, 0.00000036, 0.00000001, 0.00000054, 0.00000061, -0.00000029, -0.00000121, -0.00000078, -0.00000184, -0.00000026, 0.00000033, 0.00000039, 0.00000250, -0.00000093, 0.00000051, -0.00000061, -0.00000022, 0.00000075, -0.00000130, 0.00000043, -0.00000171, -0.00000280, 0.00000276, -0.00000103, 0.00000396, -0.00000047, 0.00000007, -0.00000078, -0.00000043, -0.00000293, -0.00000018, -0.00000282, 0.00000762, -0.00000767, 0.00000644, -0.00000558, 0.00000639, -0.00001119, 0.00000600, 0.00000016, 0.00001514, -0.00001857, 0.00001322, -0.00001978, 0.00001603, -0.00001700, 0.00000836, -0.00001045, 0.00002344, -0.00002386, 0.00002248, -0.00002385, 0.00003228, -0.00002899, -0.00000713, -0.00000620, 0.00002931, -0.00001957, 0.00000465, -0.00001345, 0.00002434, -0.00001837, -0.00001480, -0.00000330, 0.00000324, 0.00002309, -0.00008283, 0.00006554, -0.00012340, 0.00011205, -0.00066657, 0.00121118, -0.00017521, -0.00027017, 0.00010956, -0.00014801, 0.00007798, -0.00010613, 0.00008733, -0.00005057, 0.00001487, -0.00002278, 0.00002019, -0.00001856, 0.00001680, -0.00002512, 0.00002542, -0.00001255, 0.00001176, -0.00002175, 0.00000120, 0.00000214, -0.00000210, 0.00000047, 0.00000409, -0.00000235, -0.00000013, -0.00000064, 0.00000204, -0.00000125, -0.00000713, 0.00000102, -0.00000027, 0.00000103, 0.00000279, 0.00000342, -0.00000397, -0.00000022, -0.00000239, 0.00000051, 0.00000161, -0.00000004, -0.00000131, 0.00000057, -0.00000014, -0.00000014, -0.00000215, -0.00000053, 0.00000217, 0.00000049, -0.00000150, 0.00000176, -0.00000142, -0.00000153, -0.00000022, },
		{ 0.00000124, 0.00000028, -0.00000117, 0.00000039, 0.00000001, 0.00000064, 0.00000060, -0.00000036, -0.00000113, -0.00000087, -0.00000179, -0.00000020, 0.00000023, 0.00000053, 0.00000234, -0.00000093, 0.00000055, -0.00000080, 0.00000005, 0.00000066, -0.00000129, 0.00000052, -0.00000202, -0.00000227, 0.00000250, -0.00000105, 0.00000401, -0.00000053, -0.00000006, -0.00000070, -0.00000072, -0.00000231, -0.00000073, -0.00000218, 0.00000697, -0.00000703, 0.00000590, -0.00000517, 0.00000567, -0.00001047, 0.00000579, 0.00000071, 0.00001391, -0.00001758, 0.00001250, -0.00001939, 0.00001588, -0.00001676, 0.00000844, -0.00001038, 0.00002312, -0.00002370, 0.00002259, -0.00002435, 0.00003341, -0.00003159, -0.00000438, -0.00000778, 0.00003110, -0.00002205, 0.00000690, -0.00001545, 0.00002702, -0.00002244, -0.00000922, -0.00001039, 0.00001255, 0.00001388, -0.00007522, 0.00005795, -0.00011791, 0.00010640, -0.00067235, 0.00124885, -0.00021939, -0.00025050, 0.00010273, -0.00014358, 0.00007739, -0.00010840, 0.00009163, -0.00005429, 0.00001768, -0.00002575, 0.00002238, -0.00001965, 0.00001791, -0.00002701, 0.00002739, -0.00001353, 0.00001271, -0.00002442, 0.00000408, 0.00000046, -0.00000081, -0.00000066, 0.00000479, -0.00000248, 0.00000027, -0.00000138, 0.00000265, -0.00000205, -0.00000654, 0.00000067, -0.00000024, 0.00000142, 0.00000292, 0.00000286, -0.00000383, -0.00000022, -0.00000226, 0.00000051, 0.00000154, 0.00000017, -0.00000147, 0.00000050, -0.00000008, -0.00000038, -0.00000197, -0.00000060, 0.00000235, 0.00000044, -0.00000145, 0.00000180, -0.00000155, -0.00000152, -0.00000009, 0.00000026, },
		{ 0.00000010, -0.00000112, 0.00000040, 0.00000002, 0.00000075, 0.00000058, -0.00000041, -0.00000106, -0.00000097, -0.00000172, -0.00000016, 0.00000015, 0.00000067, 0.00000215, -0.00000090, 0.00000056, -0.00000097, 0.00000031, 0.00000055, -0.00000124, 0.00000056, -0.00000229, -0.00000177, 0.00000222, -0.00000105, 0.00000406, -0.00000060, -0.00000016, -0.00000067, -0.00000094, -0.00000176, -0.00000123, -0.00000158, 0.00000631, -0.00000637, 0.00000533, -0.00000475, 0.00000495, -0.00000972, 0.00000557, 0.00000123, 0.00001266, -0.00001650, 0.00001164, -0.00001887, 0.00001560, -0.00001639, 0.00000842, -0.00001020, 0.00002264, -0.00002332, 0.00002245, -0.00002457, 0.00003421, -0.00003392, -0.00000171, -0.00000930, 0.00003272, -0.00002436, 0.00000904, -0.00001729, 0.00002942, -0.00002623, -0.00000381, -0.00001741, 0.00002197, 0.00000413, -0.00006667, 0.00004916, -0.00011089, 0.00009854, -0.00067475, 0.00128363, -0.00026339, -0.00022925, 0.00009424, -0.00013766, 0.00007547, -0.00010948, 0.00009505, -0.00005754, 0.00002024, -0.00002861, 0.00002454, -0.00002069, 0.00001888, -0.00002875, 0.00002927, -0.00001444, 0.00001354, -0.00002701, 0.00000702, -0.00000135, 0.00000061, -0.00000190, 0.00000560, -0.00000268, 0.00000071, -0.00000218, 0.00000331, -0.00000291, -0.00000586, 0.00000022, -0.00000014, 0.00000180, 0.00000305, 0.00000228, -0.00000364, -0.00000024, -0.00000208, 0.00000047, 0.00000147, 0.00000038, -0.00000166, 0.00000045, -0.00000006, -0.00000059, -0.00000179, -0.00000068, 0.00000253, 0.00000038, -0.00000139, 0.00000183, -0.00000169, -0.00000147, 0.00000001, 0.00000009, 0.00000138, },
		{ -0.00000105, 0.00000039, 0.00000004, 0.00000085, 0.00000054, -0.00000045, -0.00000100, -0.00000106, -0.00000163, -0.00000014, 0.00000009, 0.00000080, 0.00000197, -0.00000085, 0.00000055, -0.00000111, 0.00000057, 0.00000042, -0.00000116, 0.00000057, -0.00000251, -0.00000129, 0.00000193, -0.00000103, 0.00000409, -0.00000067, -0.00000025, -0.00000067, -0.00000109, -0.00000127, -0.00000166, -0.00000102, 0.00000567, -0.00000571, 0.00000475, -0.00000433, 0.00000422, -0.00000894, 0.00000535, 0.00000172, 0.00001140, -0.00001534, 0.00001069, -0.00001822, 0.00001521, -0.00001590, 0.00000830, -0.00000993, 0.00002201, -0.00002273, 0.00002208, -0.00002451, 0.00003466, -0.00003595, 0.00000083, -0.00001073, 0.00003417, -0.00002649, 0.00001106, -0.00001895, 0.00003151, -0.00002968, 0.00000135, -0.00002425, 0.00003136, -0.00000603, -0.00005732, 0.00003924, -0.00010241, 0.00008848, -0.00067362, 0.00131517, -0.00030681, -0.00020666, 0.00008423, -0.00013033, 0.00007221, -0.00010935, 0.00009752, -0.00006024, 0.00002250, -0.00003130, 0.00002664, -0.00002166, 0.00001970, -0.00003030, 0.00003101, -0.00001525, 0.00001422, -0.00002948, 0.00000998, -0.00000326, 0.00000214, -0.00000325, 0.00000650, -0.00000296, 0.00000120, -0.00000304, 0.00000403, -0.00000383, -0.00000509, -0.00000031, 0.00000003, 0.00000216, 0.00000316, 0.00000167, -0.00000341, -0.00000029, -0.00000187, 0.00000039, 0.00000143, 0.00000059, -0.00000185, 0.00000043, -0.00000007, -0.00000076, -0.00000161, -0.00000075, 0.00000271, 0.00000030, -0.00000130, 0.00000184, -0.00000185, -0.00000139, 0.00000008, -0.00000006, 0.00000150, -0.00000008, },
		{ 0.00000037, 0.00000007, 0.00000094, 0.00000049, -0.00000047, -0.00000096, -0.00000115, -0.00000155, -0.00000014, 0.00000004, 0.00000092, 0.00000177, -0.00000078, 0.00000052, -0.00000122, 0.00000081, 0.00000029, -0.00000106, 0.00000053, -0.00000269, -0.00000083, 0.00000164, -0.00000099, 0.00000410, -0.00000074, -0.00000032, -0.00000070, -0.00000119, -0.00000085, -0.00000204, -0.00000051, 0.00000506, -0.00000505, 0.00000415, -0.00000391, 0.00000349, -0.00000815, 0.00000513, 0.00000216, 0.00001016, -0.00001412, 0.00000963, -0.00001746, 0.00001470, -0.00001530, 0.00000807, -0.00000957, 0.00002123, -0.00002196, 0.00002148, -0.00002418, 0.00003476, -0.00003767, 0.00000323, -0.00001206, 0.00003542, -0.00002841, 0.00001293, -0.00002043, 0.00003328, -0.00003274, 0.00000619, -0.00003080, 0.00004059, -0.00001644, -0.00004728, 0.00002832, -0.00009255, 0.00007626, -0.00066888, 0.00134318, -0.00034928, -0.00018304, 0.00007286, -0.00012168, 0.00006768, -0.00010798, 0.00009898, -0.00006231, 0.00002438, -0.00003378, 0.00002863, -0.00002253, 0.00002036, -0.00003166, 0.00003261, -0.00001595, 0.00001473, -0.00003179, 0.00001293, -0.00000525, 0.00000376, -0.00000469, 0.00000749, -0.00000330, 0.00000172, -0.00000395, 0.00000480, -0.00000480, -0.00000424, -0.00000095, 0.00000029, 0.00000249, 0.00000327, 0.00000103, -0.00000312, -0.00000038, -0.00000162, 0.00000027, 0.00000141, 0.00000079, -0.00000206, 0.00000043, -0.00000012, -0.00000091, -0.00000144, -0.00000082, 0.00000289, 0.00000021, -0.00000119, 0.00000183, -0.00000202, -0.00000128, 0.00000011, -0.00000019, 0.00000161, -0.00000027, -0.00000095, },
		{ 0.00000013, 0.00000103, 0.00000043, -0.00000048, -0.00000093, -0.00000123, -0.00000146, -0.00000015, 0.00000002, 0.00000102, 0.00000157, -0.00000069, 0.00000046, -0.00000130, 0.00000103, 0.00000015, -0.00000094, 0.00000046, -0.00000282, -0.00000041, 0.00000134, -0.00000093, 0.00000410, -0.00000081, -0.00000038, -0.00000074, -0.00000123, -0.00000050, -0.00000235, -0.00000006, 0.00000447, -0.00000439, 0.00000355, -0.00000349, 0.00000278, -0.00000734, 0.00000492, 0.00000254, 0.00000894, -0.00001287, 0.00000851, -0.00001661, 0.00001409, -0.00001460, 0.00000777, -0.00000915, 0.00002034, -0.00002100, 0.00002067, -0.00002357, 0.00003452, -0.00003905, 0.00000545, -0.00001327, 0.00003645, -0.00003011, 0.00001464, -0.00002170, 0.00003471, -0.00003538, 0.00001064, -0.00003698, 0.00004952, -0.00002696, -0.00003672, 0.00001653, -0.00008143, 0.00006194, -0.00066046, 0.00136737, -0.00039041, -0.00015867, 0.00006033, -0.00011183, 0.00006192, -0.00010538, 0.00009940, -0.00006372, 0.00002585, -0.00003599, 0.00003049, -0.00002330, 0.00002085, -0.00003280, 0.00003405, -0.00001652, 0.00001505, -0.00003391, 0.00001583, -0.00000729, 0.00000545, -0.00000620, 0.00000856, -0.00000370, 0.00000228, -0.00000491, 0.00000561, -0.00000582, -0.00000330, -0.00000167, 0.00000062, 0.00000279, 0.00000338, 0.00000036, -0.00000280, -0.00000050, -0.00000133, 0.00000009, 0.00000144, 0.00000096, -0.00000226, 0.00000044, -0.00000019, -0.00000102, -0.00000128, -0.00000089, 0.00000307, 0.00000011, -0.00000106, 0.00000180, -0.00000218, -0.00000116, 0.00000012, -0.00000029, 0.00000170, -0.00000046, -0.00000084, 0.00000032, },
		{ 0.00000110, 0.00000037, -0.00000047, -0.00000092, -0.00000130, -0.00000136, -0.00000018, 0.00000001, 0.00000111, 0.00000138, -0.00000059, 0.00000038, -0.00000134, 0.00000124, 0.00000002, -0.00000080, 0.00000035, -0.00000291, -0.00000002, 0.00000104, -0.00000086, 0.00000408, -0.00000088, -0.00000043, -0.00000081, -0.00000122, -0.00000021, -0.00000259, 0.00000033, 0.00000393, -0.00000376, 0.00000296, -0.00000308, 0.00000208, -0.00000653, 0.00000471, 0.00000286, 0.00000776, -0.00001159, 0.00000732, -0.00001567, 0.00001341, -0.00001381, 0.00000738, -0.00000866, 0.00001934, -0.00001988, 0.00001966, -0.00002271, 0.00003392, -0.00004008, 0.00000747, -0.00001434, 0.00003727, -0.00003156, 0.00001617, -0.00002277, 0.00003577, -0.00003757, 0.00001464, -0.00004270, 0.00005804, -0.00003742, -0.00002578, 0.00000400, -0.00006918, 0.00004564, -0.00064834, 0.00138751, -0.00042984, -0.00013385, 0.00004683, -0.00010091, 0.00005502, -0.00010158, 0.00009875, -0.00006440, 0.00002684, -0.00003789, 0.00003218, -0.00002394, 0.00002114, -0.00003370, 0.00003530, -0.00001695, 0.00001516, -0.00003580, 0.00001865, -0.00000935, 0.00000718, -0.00000777, 0.00000969, -0.00000415, 0.00000286, -0.00000590, 0.00000647, -0.00000688, -0.00000229, -0.00000249, 0.00000105, 0.00000304, 0.00000349, -0.00000033, -0.00000242, -0.00000066, -0.00000099, -0.00000013, 0.00000150, 0.00000110, -0.00000245, 0.00000045, -0.00000030, -0.00000109, -0.00000113, -0.00000094, 0.00000324, 0.00000001, -0.00000091, 0.00000174, -0.00000234, -0.00000101, 0.00000010, -0.00000037, 0.00000177, -0.00000064, -0.00000071, 0.00000026, 0.00000020, },
		{ 0.00000030, -0.00000045, -0.00000093, -0.00000136, -0.00000127, -0.00000021, 0.00000001, 0.00000117, 0.00000119, -0.00000048, 0.00000027, -0.00000136, 0.00000141, -0.00000010, -0.00000066, 0.00000022, -0.00000295, 0.00000033, 0.00000075, -0.00000077, 0.00000403, -0.00000093, -0.00000049, -0.00000088, -0.00000117, 0.00000002, -0.00000276, 0.00000065, 0.00000344, -0.00000315, 0.00000237, -0.00000269, 0.00000140, -0.00000572, 0.00000452, 0.00000312, 0.00000664, -0.00001030, 0.00000610, -0.00001466, 0.00001264, -0.00001295, 0.00000693, -0.00000812, 0.00001824, -0.00001863, 0.00001848, -0.00002161, 0.00003298, -0.00004075, 0.00000928, -0.00001527, 0.00003785, -0.00003277, 0.00001751, -0.00002362, 0.00003648, -0.00003929, 0.00001815, -0.00004787, 0.00006602, -0.00004766, -0.00001464, -0.00000908, -0.00005597, 0.00002749, -0.00063254, 0.00140339, -0.00046722, -0.00010890, 0.00003258, -0.00008909, 0.00004708, -0.00009664, 0.00009703, -0.00006433, 0.00002731, -0.00003942, 0.00003365, -0.00002444, 0.00002125, -0.00003437, 0.00003635, -0.00001722, 0.00001505, -0.00003745, 0.00002134, -0.00001139, 0.00000893, -0.00000938, 0.00001088, -0.00000466, 0.00000346, -0.00000693, 0.00000735, -0.00000798, -0.00000121, -0.00000340, 0.00000157, 0.00000324, 0.00000361, -0.00000104, -0.00000200, -0.00000085, -0.00000062, -0.00000041, 0.00000161, 0.00000120, -0.00000261, 0.00000047, -0.00000042, -0.00000113, -0.00000100, -0.00000098, 0.00000340, -0.00000010, -0.00000073, 0.00000164, -0.00000250, -0.00000084, 0.00000005, -0.00000041, 0.00000181, -0.00000080, -0.00000057, 0.00000018, 0.00000029, 0.00000117, },
		{ -0.00000042, -0.00000095, -0.00000140, -0.00000119, -0.00000025, 0.00000003, 0.00000122, 0.00000100, -0.00000037, 0.00000016, -0.00000134, 0.00000156, -0.00000021, -0.00000051, 0.00000006, -0.00000296, 0.00000065, 0.00000046, -0.00000067, 0.00000397, -0.00000098, -0.00000054, -0.00000096, -0.00000108, 0.00000020, -0.00000288, 0.00000090, 0.00000299, -0.00000258, 0.00000180, -0.00000231, 0.00000074, -0.00000491, 0.00000433, 0.00000331, 0.00000559, -0.00000904, 0.00000485, -0.00001360, 0.00001183, -0.00001202, 0.00000643, -0.00000756, 0.00001708, -0.00001725, 0.00001714, -0.00002030, 0.00003172, -0.00004106, 0.00001085, -0.00001603, 0.00003820, -0.00003371, 0.00001865, -0.00002426, 0.00003684, -0.00004052, 0.00002112, -0.00005242, 0.00007335, -0.00005755, -0.00000345, -0.00002255, -0.00004197, 0.00000766, -0.00061313, 0.00141485, -0.00050223, -0.00008412, 0.00001781, -0.00007653, 0.00003823, -0.00009060, 0.00009425, -0.00006348, 0.00002724, -0.00004055, 0.00003488, -0.00002477, 0.00002115, -0.00003477, 0.00003718, -0.00001733, 0.00001471, -0.00003881, 0.00002387, -0.00001339, 0.00001069, -0.00001101, 0.00001211, -0.00000520, 0.00000408, -0.00000796, 0.00000825, -0.00000910, -0.00000007, -0.00000439, 0.00000219, 0.00000339, 0.00000373, -0.00000178, -0.00000153, -0.00000108, -0.00000021, -0.00000074, 0.00000177, 0.00000126, -0.00000275, 0.00000047, -0.00000056, -0.00000114, -0.00000088, -0.00000101, 0.00000354, -0.00000022, -0.00000053, 0.00000152, -0.00000264, -0.00000066, -0.00000002, -0.00000043, 0.00000183, -0.00000096, -0.00000042, 0.00000008, 0.00000040, 0.00000121, 0.00000023, },
		{ -0.00000099, -0.00000143, -0.00000111, -0.00000030, 0.00000007, 0.00000125, 0.00000083, -0.00000025, 0.00000002, -0.00000128, 0.00000168, -0.00000030, -0.00000037, -0.00000011, -0.00000292, 0.00000092, 0.00000019, -0.00000057, 0.00000389, -0.00000101, -0.00000061, -0.00000103, -0.00000097, 0.00000032, -0.00000292, 0.00000108, 0.00000260, -0.00000205, 0.00000125, -0.00000196, 0.00000012, -0.00000411, 0.00000417, 0.00000344, 0.00000461, -0.00000780, 0.00000359, -0.00001250, 0.00001096, -0.00001105, 0.00000588, -0.00000697, 0.00001587, -0.00001578, 0.00001566, -0.00001879, 0.00003016, -0.00004101, 0.00001219, -0.00001662, 0.00003831, -0.00003438, 0.00001959, -0.00002469, 0.00003684, -0.00004125, 0.00002353, -0.00005630, 0.00007992, -0.00006692, 0.00000759, -0.00003622, -0.00002737, -0.00001365, -0.00059023, 0.00142177, -0.00053457, -0.00005980, 0.00000274, -0.00006340, 0.00002860, -0.00008357, 0.00009044, -0.00006186, 0.00002659, -0.00004125, 0.00003584, -0.00002494, 0.00002084, -0.00003491, 0.00003779, -0.00001727, 0.00001412, -0.00003988, 0.00002620, -0.00001531, 0.00001241, -0.00001263, 0.00001336, -0.00000578, 0.00000469, -0.00000900, 0.00000916, -0.00001024, 0.00000112, -0.00000546, 0.00000290, 0.00000347, 0.00000387, -0.00000254, -0.00000102, -0.00000134, 0.00000025, -0.00000113, 0.00000199, 0.00000126, -0.00000286, 0.00000045, -0.00000070, -0.00000112, -0.00000078, -0.00000102, 0.00000368, -0.00000034, -0.00000031, 0.00000136, -0.00000276, -0.00000047, -0.00000012, -0.00000041, 0.00000181, -0.00000109, -0.00000027, -0.00000002, 0.00000053, 0.00000125, 0.00000017, -0.00000039, },
		{ -0.00000143, -0.00000105, -0.00000034, 0.00000012, 0.00000126, 0.00000067, -0.00000013, -0.00000012, -0.00000120, 0.00000176, -0.00000037, -0.00000024, -0.00000029, -0.00000286, 0.00000116, -0.00000007, -0.00000045, 0.00000379, -0.00000102, -0.00000069, -0.00000108, -0.00000084, 0.00000040, -0.00000292, 0.00000120, 0.00000227, -0.00000156, 0.00000073, -0.00000164, -0.00000047, -0.00000331, 0.00000401, 0.00000349, 0.00000371, -0.00000660, 0.00000234, -0.00001137, 0.00001007, -0.00001004, 0.00000529, -0.00000637, 0.00001462, -0.00001423, 0.00001408, -0.00001712, 0.00002831, -0.00004062, 0.00001328, -0.00001703, 0.00003817, -0.00003478, 0.00002032, -0.00002491, 0.00003650, -0.00004149, 0.00002535, -0.00005946, 0.00008564, -0.00007563, 0.00001833, -0.00004990, -0.00001240, -0.00003622, -0.00056398, 0.00142408, -0.00056398, -0.00003622, -0.00001240, -0.00004990, 0.00001833, -0.00007563, 0.00008564, -0.00005946, 0.00002535, -0.00004149, 0.00003650, -0.00002491, 0.00002032, -0.00003478, 0.00003817, -0.00001703, 0.00001328, -0.00004062, 0.00002831, -0.00001712, 0.00001408, -0.00001423, 0.00001462, -0.00000637, 0.00000529, -0.00001004, 0.00001007, -0.00001137, 0.00000234, -0.00000660, 0.00000371, 0.00000349, 0.00000401, -0.00000331, -0.00000047, -0.00000164, 0.00000073, -0.00000156, 0.00000227, 0.00000120, -0.00000292, 0.00000040, -0.00000084, -0.00000108, -0.00000069, -0.00000102, 0.00000379, -0.00000045, -0.00000007, 0.00000116, -0.00000286, -0.00000029, -0.00000024, -0.00000037, 0.00000176, -0.00000120, -0.00000012, -0.00000013, 0.00000067, 0.00000126, 0.00000012, -0.00000034, -0.00000105, },
		{ -0.00000099, -0.00000039, 0.00000017, 0.00000125, 0.00000053, -0.00000002, -0.00000027, -0.00000109, 0.00000181, -0.00000041, -0.00000012, -0.00000047, -0.00000276, 0.00000136, -0.00000031, -0.00000034, 0.00000368, -0.00000102, -0.00000078, -0.00000112, -0.00000070, 0.00000045, -0.00000286, 0.00000126, 0.00000199, -0.00000113, 0.00000025, -0.00000134, -0.00000102, -0.00000254, 0.00000387, 0.00000347, 0.00000290, -0.00000546, 0.00000112, -0.00001024, 0.00000916, -0.00000900, 0.00000469, -0.00000578, 0.00001336, -0.00001263, 0.00001241, -0.00001531, 0.00002620, -0.00003988, 0.00001412, -0.00001727, 0.00003779, -0.00003491, 0.00002084, -0.00002494, 0.00003584, -0.00004125, 0.00002659, -0.00006186, 0.00009044, -0.00008357, 0.00002860, -0.00006340, 0.00000274, -0.00005980, -0.00053457, 0.00142177, -0.00059023, -0.00001365, -0.00002737, -0.00003622, 0.00000759, -0.00006692, 0.00007992, -0.00005630, 0.00002353, -0.00004125, 0.00003684, -0.00002469, 0.00001959, -0.00003438, 0.00003831, -0.00001662, 0.00001219, -0.00004101, 0.00003016, -0.00001879, 0.00001566, -0.00001578, 0.00001587, -0.00000697, 0.00000588, -0.00001105, 0.00001096, -0.00001250, 0.00000359, -0.00000780, 0.00000461, 0.00000344, 0.00000417, -0.00000411, 0.00000012, -0.00000196, 0.00000125, -0.00000205, 0.00000260, 0.00000108, -0.00000292, 0.00000032, -0.00000097, -0.00000103, -0.00000061, -0.00000101, 0.00000389, -0.00000057, 0.00000019, 0.00000092, -0.00000292, -0.00000011, -0.00000037, -0.00000030, 0.00000168, -0.00000128, 0.00000002, -0.00000025, 0.00000083, 0.00000125, 0.00000007, -0.00000030, -0.00000111, -0.00000143, },
		{ -0.00000042, 0.00000023, 0.00000121, 0.00000040, 0.00000008, -0.00000042, -0.00000096, 0.00000183, -0.00000043, -0.00000002, -0.00000066, -0.00000264, 0.00000152, -0.00000053, -0.00000022, 0.00000354, -0.00000101, -0.00000088, -0.00000114, -0.00000056, 0.00000047, -0.00000275, 0.00000126, 0.00000177, -0.00000074, -0.00000021, -0.00000108, -0.00000153, -0.00000178, 0.00000373, 0.00000339, 0.00000219, -0.00000439, -0.00000007, -0.00000910, 0.00000825, -0.00000796, 0.00000408, -0.00000520, 0.00001211, -0.00001101, 0.00001069, -0.00001339, 0.00002387, -0.00003881, 0.00001471, -0.00001733, 0.00003718, -0.00003477, 0.00002115, -0.00002477, 0.00003488, -0.00004055, 0.00002724, -0.00006348, 0.00009425, -0.00009060, 0.00003823, -0.00007653, 0.00001781, -0.00008412, -0.00050223, 0.00141485, -0.00061313, 0.00000766, -0.00004197, -0.00002255, -0.00000345, -0.00005755, 0.00007335, -0.00005242, 0.00002112, -0.00004052, 0.00003684, -0.00002426, 0.00001865, -0.00003371, 0.00003820, -0.00001603, 0.00001085, -0.00004106, 0.00003172, -0.00002030, 0.00001714, -0.00001725, 0.00001708, -0.00000756, 0.00000643, -0.00001202, 0.00001183, -0.00001360, 0.00000485, -0.00000904, 0.00000559, 0.00000331, 0.00000433, -0.00000491, 0.00000074, -0.00000231, 0.00000180, -0.00000258, 0.00000299, 0.00000090, -0.00000288, 0.00000020, -0.00000108, -0.00000096, -0.00000054, -0.00000098, 0.00000397, -0.00000067, 0.00000046, 0.00000065, -0.00000296, 0.00000006, -0.00000051, -0.00000021, 0.00000156, -0.00000134, 0.00000016, -0.00000037, 0.00000100, 0.00000122, 0.00000003, -0.00000025, -0.00000119, -0.00000140, -0.00000095, },
		{ 0.00000030, 0.00000117, 0.00000029, 0.00000018, -0.00000057, -0.00000080, 0.00000181, -0.00000041, 0.00000005, -0.00000084, -0.00000250, 0.00000164, -0.00000073, -0.00000010, 0.00000340, -0.00000098, -0.00000100, -0.00000113, -0.00000042, 0.00000047, -0.00000261, 0.00000120, 0.00000161, -0.00000041, -0.00000062, -0.00000085, -0.00000200, -0.00000104, 0.00000361, 0.00000324, 0.00000157, -0.00000340, -0.00000121, -0.00000798, 0.00000735, -0.00000693, 0.00000346, -0.00000466, 0.00001088, -0.00000938, 0.00000893, -0.00001139, 0.00002134, -0.00003745, 0.00001505, -0.00001722, 0.00003635, -0.00003437, 0.00002125, -0.00002444, 0.00003365, -0.00003942, 0.00002731, -0.00006433, 0.00009703, -0.00009664, 0.00004708, -0.00008909, 0.00003258, -0.00010890, -0.00046722, 0.00140339, -0.00063254, 0.00002749, -0.00005597, -0.00000908, -0.00001464, -0.00004766, 0.00006602, -0.00004787, 0.00001815, -0.00003929, 0.00003648, -0.00002362, 0.00001751, -0.00003277, 0.00003785, -0.00001527, 0.00000928, -0.00004075, 0.00003298, -0.00002161, 0.00001848, -0.00001863, 0.00001824, -0.00000812, 0.00000693, -0.00001295, 0.00001264, -0.00001466, 0.00000610, -0.00001030, 0.00000664, 0.00000312, 0.00000452, -0.00000572, 0.00000140, -0.00000269, 0.00000237, -0.00000315, 0.00000344, 0.00000065, -0.00000276, 0.00000002, -0.00000117, -0.00000088, -0.00000049, -0.00000093, 0.00000403, -0.00000077, 0.00000075, 0.00000033, -0.00000295, 0.00000022, -0.00000066, -0.00000010, 0.00000141, -0.00000136, 0.00000027, -0.00000048, 0.00000119, 0.00000117, 0.00000001, -0.00000021, -0.00000127, -0.00000136, -0.00000093, -0.00000045, },
		{ 0.00000110, 0.00000020, 0.00000026, -0.00000071, -0.00000064, 0.00000177, -0.00000037, 0.00000010, -0.00000101, -0.00000234, 0.00000174, -0.00000091, 0.00000001, 0.00000324, -0.00000094, -0.00000113, -0.00000109, -0.00000030, 0.00000045, -0.00000245, 0.00000110, 0.00000150, -0.00000013, -0.00000099, -0.00000066, -0.00000242, -0.00000033, 0.00000349, 0.00000304, 0.00000105, -0.00000249, -0.00000229, -0.00000688, 0.00000647, -0.00000590, 0.00000286, -0.00000415, 0.00000969, -0.00000777, 0.00000718, -0.00000935, 0.00001865, -0.00003580, 0.00001516, -0.00001695, 0.00003530, -0.00003370, 0.00002114, -0.00002394, 0.00003218, -0.00003789, 0.00002684, -0.00006440, 0.00009875, -0.00010158, 0.00005502, -0.00010091, 0.00004683, -0.00013385, -0.00042984, 0.00138751, -0.00064834, 0.00004564, -0.00006918, 0.00000400, -0.00002578, -0.00003742, 0.00005804, -0.00004270, 0.00001464, -0.00003757, 0.00003577, -0.00002277, 0.00001617, -0.00003156, 0.00003727, -0.00001434, 0.00000747, -0.00004008, 0.00003392, -0.00002271, 0.00001966, -0.00001988, 0.00001934, -0.00000866, 0.00000738, -0.00001381, 0.00001341, -0.00001567, 0.00000732, -0.00001159, 0.00000776, 0.00000286, 0.00000471, -0.00000653, 0.00000208, -0.00000308, 0.00000296, -0.00000376, 0.00000393, 0.00000033, -0.00000259, -0.00000021, -0.00000122, -0.00000081, -0.00000043, -0.00000088, 0.00000408, -0.00000086, 0.00000104, -0.00000002, -0.00000291, 0.00000035, -0.00000080, 0.00000002, 0.00000124, -0.00000134, 0.00000038, -0.00000059, 0.00000138, 0.00000111, 0.00000001, -0.00000018, -0.00000136, -0.00000130, -0.00000092, -0.00000047, 0.00000037, },
		{ 0.00000013, 0.00000032, -0.00000084, -0.00000046, 0.00000170, -0.00000029, 0.00000012, -0.00000116, -0.00000218, 0.00000180, -0.00000106, 0.00000011, 0.00000307, -0.00000089, -0.00000128, -0.00000102, -0.00000019, 0.00000044, -0.00000226, 0.00000096, 0.00000144, 0.00000009, -0.00000133, -0.00000050, -0.00000280, 0.00000036, 0.00000338, 0.00000279, 0.00000062, -0.00000167, -0.00000330, -0.00000582, 0.00000561, -0.00000491, 0.00000228, -0.00000370, 0.00000856, -0.00000620, 0.00000545, -0.00000729, 0.00001583, -0.00003391, 0.00001505, -0.00001652, 0.00003405, -0.00003280, 0.00002085, -0.00002330, 0.00003049, -0.00003599, 0.00002585, -0.00006372, 0.00009940, -0.00010538, 0.00006192, -0.00011183, 0.00006033, -0.00015867, -0.00039041, 0.00136737, -0.00066046, 0.00006194, -0.00008143, 0.00001653, -0.00003672, -0.00002696, 0.00004952, -0.00003698, 0.00001064, -0.00003538, 0.00003471, -0.00002170, 0.00001464, -0.00003011, 0.00003645, -0.00001327, 0.00000545, -0.00003905, 0.00003452, -0.00002357, 0.00002067, -0.00002100, 0.00002034, -0.00000915, 0.00000777, -0.00001460, 0.00001409, -0.00001661, 0.00000851, -0.00001287, 0.00000894, 0.00000254, 0.00000492, -0.00000734, 0.00000278, -0.00000349, 0.00000355, -0.00000439, 0.00000447, -0.00000006, -0.00000235, -0.00000050, -0.00000123, -0.00000074, -0.00000038, -0.00000081, 0.00000410, -0.00000093, 0.00000134, -0.00000041, -0.00000282, 0.00000046, -0.00000094, 0.00000015, 0.00000103, -0.00000130, 0.00000046, -0.00000069, 0.00000157, 0.00000102, 0.00000002, -0.00000015, -0.00000146, -0.00000123, -0.00000093, -0.00000048, 0.00000043, 0.00000103, },
		{ 0.00000037, -0.00000095, -0.00000027, 0.00000161, -0.00000019, 0.00000011, -0.00000128, -0.00000202, 0.00000183, -0.00000119, 0.00000021, 0.00000289, -0.00000082, -0.00000144, -0.00000091, -0.00000012, 0.00000043, -0.00000206, 0.00000079, 0.00000141, 0.00000027, -0.00000162, -0.00000038, -0.00000312, 0.00000103, 0.00000327, 0.00000249, 0.00000029, -0.00000095, -0.00000424, -0.00000480, 0.00000480, -0.00000395, 0.00000172, -0.00000330, 0.00000749, -0.00000469, 0.00000376, -0.00000525, 0.00001293, -0.00003179, 0.00001473, -0.00001595, 0.00003261, -0.00003166, 0.00002036, -0.00002253, 0.00002863, -0.00003378, 0.00002438, -0.00006231, 0.00009898, -0.00010798, 0.00006768, -0.00012168, 0.00007286, -0.00018304, -0.00034928, 0.00134318, -0.00066888, 0.00007626, -0.00009255, 0.00002832, -0.00004728, -0.00001644, 0.00004059, -0.00003080, 0.00000619, -0.00003274, 0.00003328, -0.00002043, 0.00001293, -0.00002841, 0.00003542, -0.00001206, 0.00000323, -0.00003767, 0.00003476, -0.00002418, 0.00002148, -0.00002196, 0.00002123, -0.00000957, 0.00000807, -0.00001530, 0.00001470, -0.00001746, 0.00000963, -0.00001412, 0.00001016, 0.00000216, 0.00000513, -0.00000815, 0.00000349, -0.00000391, 0.00000415, -0.00000505, 0.00000506, -0.00000051, -0.00000204, -0.00000085, -0.00000119, -0.00000070, -0.00000032, -0.00000074, 0.00000410, -0.00000099, 0.00000164, -0.00000083, -0.00000269, 0.00000053, -0.00000106, 0.00000029, 0.00000081, -0.00000122, 0.00000052, -0.00000078, 0.00000177, 0.00000092, 0.00000004, -0.00000014, -0.00000155, -0.00000115, -0.00000096, -0.00000047, 0.00000049, 0.00000094, 0.00000007, },
		{ -0.00000105, -0.00000008, 0.00000150, -0.00000006, 0.00000008, -0.00000139, -0.00000185, 0.00000184, -0.00000130, 0.00000030, 0.00000271, -0.00000075, -0.00000161, -0.00000076, -0.00000007, 0.00000043, -0.00000185, 0.00000059, 0.00000143, 0.00000039, -0.00000187, -0.00000029, -0.00000341, 0.00000167, 0.00000316, 0.00000216, 0.00000003, -0.00000031, -0.00000509, -0.00000383, 0.00000403, -0.00000304, 0.00000120, -0.00000296, 0.00000650, -0.00000325, 0.00000214, -0.00000326, 0.00000998, -0.00002948, 0.00001422, -0.00001525, 0.00003101, -0.00003030, 0.00001970, -0.00002166, 0.00002664, -0.00003130, 0.00002250, -0.00006024, 0.00009752, -0.00010935, 0.00007221, -0.00013033, 0.00008423, -0.00020666, -0.00030681, 0.00131517, -0.00067362, 0.00008848, -0.00010241, 0.00003924, -0.00005732, -0.00000603, 0.00003136, -0.00002425, 0.00000135, -0.00002968, 0.00003151, -0.00001895, 0.00001106, -0.00002649, 0.00003417, -0.00001073, 0.00000083, -0.00003595, 0.00003466, -0.00002451, 0.00002208, -0.00002273, 0.00002201, -0.00000993, 0.00000830, -0.00001590, 0.00001521, -0.00001822, 0.00001069, -0.00001534, 0.00001140, 0.00000172, 0.00000535, -0.00000894, 0.00000422, -0.00000433, 0.00000475, -0.00000571, 0.00000567, -0.00000102, -0.00000166, -0.00000127, -0.00000109, -0.00000067, -0.00000025, -0.00000067, 0.00000409, -0.00000103, 0.00000193, -0.00000129, -0.00000251, 0.00000057, -0.00000116, 0.00000042, 0.00000057, -0.00000111, 0.00000055, -0.00000085, 0.00000197, 0.00000080, 0.00000009, -0.00000014, -0.00000163, -0.00000106, -0.00000100, -0.00000045, 0.00000054, 0.00000085, 0.00000004, 0.00000039, },
		{ 0.00000010, 0.00000138, 0.00000009, 0.00000001, -0.00000147, -0.00000169, 0.00000183, -0.00000139, 0.00000038, 0.00000253, -0.00000068, -0.00000179, -0.00000059, -0.00000006, 0.00000045, -0.00000166, 0.00000038, 0.00000147, 0.00000047, -0.00000208, -0.00000024, -0.00000364, 0.00000228, 0.00000305, 0.00000180, -0.00000014, 0.00000022, -0.00000586, -0.00000291, 0.00000331, -0.00000218, 0.00000071, -0.00000268, 0.00000560, -0.00000190, 0.00000061, -0.00000135, 0.00000702, -0.00002701, 0.00001354, -0.00001444, 0.00002927, -0.00002875, 0.00001888, -0.00002069, 0.00002454, -0.00002861, 0.00002024, -0.00005754, 0.00009505, -0.00010948, 0.00007547, -0.00013766, 0.00009424, -0.00022925, -0.00026339, 0.00128363, -0.00067475, 0.00009854, -0.00011089, 0.00004916, -0.00006667, 0.00000413, 0.00002197, -0.00001741, -0.00000381, -0.00002623, 0.00002942, -0.00001729, 0.00000904, -0.00002436, 0.00003272, -0.00000930, -0.00000171, -0.00003392, 0.00003421, -0.00002457, 0.00002245, -0.00002332, 0.00002264, -0.00001020, 0.00000842, -0.00001639, 0.00001560, -0.00001887, 0.00001164, -0.00001650, 0.00001266, 0.00000123, 0.00000557, -0.00000972, 0.00000495, -0.00000475, 0.00000533, -0.00000637, 0.00000631, -0.00000158, -0.00000123, -0.00000176, -0.00000094, -0.00000067, -0.00000016, -0.00000060, 0.00000406, -0.00000105, 0.00000222, -0.00000177, -0.00000229, 0.00000056, -0.00000124, 0.00000055, 0.00000031, -0.00000097, 0.00000056, -0.00000090, 0.00000215, 0.00000067, 0.00000015, -0.00000016, -0.00000172, -0.00000097, -0.00000106, -0.00000041, 0.00000058, 0.00000075, 0.00000002, 0.00000040, -0.00000112, },
		{ 0.00000124, 0.00000026, -0.00000009, -0.00000152, -0.00000155, 0.00000180, -0.00000145, 0.00000044, 0.00000235, -0.00000060, -0.00000197, -0.00000038, -0.00000008, 0.00000050, -0.00000147, 0.00000017, 0.00000154, 0.00000051, -0.00000226, -0.00000022, -0.00000383, 0.00000286, 0.00000292, 0.00000142, -0.00000024, 0.00000067, -0.00000654, -0.00000205, 0.00000265, -0.00000138, 0.00000027, -0.00000248, 0.00000479, -0.00000066, -0.00000081, 0.00000046, 0.00000408, -0.00002442, 0.00001271, -0.00001353, 0.00002739, -0.00002701, 0.00001791, -0.00001965, 0.00002238, -0.00002575, 0.00001768, -0.00005429, 0.00009163, -0.00010840, 0.00007739, -0.00014358, 0.00010273, -0.00025050, -0.00021939, 0.00124885, -0.00067235, 0.00010640, -0.00011791, 0.00005795, -0.00007522, 0.00001388, 0.00001255, -0.00001039, -0.00000922, -0.00002244, 0.00002702, -0.00001545, 0.00000690, -0.00002205, 0.00003110, -0.00000778, -0.00000438, -0.00003159, 0.00003341, -0.00002435, 0.00002259, -0.00002370, 0.00002312, -0.00001038, 0.00000844, -0.00001676, 0.00001588, -0.00001939, 0.00001250, -0.00001758, 0.00001391, 0.00000071, 0.00000579, -0.00001047, 0.00000567, -0.00000517, 0.00000590, -0.00000703, 0.00000697, -0.00000218, -0.00000073, -0.00000231, -0.00000072, -0.00000070, -0.00000006, -0.00000053, 0.00000401, -0.00000105, 0.00000250, -0.00000227, -0.00000202, 0.00000052, -0.00000129, 0.00000066, 0.00000005, -0.00000080, 0.00000055, -0.00000093, 0.00000234, 0.00000053, 0.00000023, -0.00000020, -0.00000179, -0.00000087, -0.00000113, -0.00000036, 0.00000060, 0.00000064, 0.00000001, 0.00000039, -0.00000117, 0.00000028, },
		{ 0.00000045, -0.00000022, -0.00000153, -0.00000142, 0.00000176, -0.00000150, 0.00000049, 0.00000217, -0.00000053, -0.00000215, -0.00000014, -0.00000014, 0.00000057, -0.00000131, -0.00000004, 0.00000161, 0.00000051, -0.00000239, -0.00000022, -0.00000397, 0.00000342, 0.00000279, 0.00000103, -0.00000027, 0.00000102, -0.00000713, -0.00000125, 0.00000204, -0.00000064, -0.00000013, -0.00000235, 0.00000409, 0.00000047, -0.00000210, 0.00000214, 0.00000120, -0.00002175, 0.00001176, -0.00001255, 0.00002542, -0.00002512, 0.00001680, -0.00001856, 0.00002019, -0.00002278, 0.00001487, -0.00005057, 0.00008733, -0.00010613, 0.00007798, -0.00014801, 0.00010956, -0.00027017, -0.00017521, 0.00121118, -0.00066657, 0.00011205, -0.00012340, 0.00006554, -0.00008283, 0.00002309, 0.00000324, -0.00000330, -0.00001480, -0.00001837, 0.00002434, -0.00001345, 0.00000465, -0.00001957, 0.00002931, -0.00000620, -0.00000713, -0.00002899, 0.00003228, -0.00002385, 0.00002248, -0.00002386, 0.00002344, -0.00001045, 0.00000836, -0.00001700, 0.00001603, -0.00001978, 0.00001322, -0.00001857, 0.00001514, 0.00000016, 0.00000600, -0.00001119, 0.00000639, -0.00000558, 0.00000644, -0.00000767, 0.00000762, -0.00000282, -0.00000018, -0.00000293, -0.00000043, -0.00000078, 0.00000007, -0.00000047, 0.00000396, -0.00000103, 0.00000276, -0.00000280, -0.00000171, 0.00000043, -0.00000130, 0.00000075, -0.00000022, -0.00000061, 0.00000051, -0.00000093, 0.00000250, 0.00000039, 0.00000033, -0.00000026, -0.00000184, -0.00000078, -0.00000121, -0.00000029, 0.00000061, 0.00000054, 0.00000001, 0.00000036, -0.00000120, 0.00000044, 0.00000110, },
		{ -0.00000037, -0.00000152, -0.00000130, 0.00000171, -0.00000153, 0.00000053, 0.00000199, -0.00000046, -0.00000233, 0.00000012, -0.00000024, 0.00000067, -0.00000117, -0.00000024, 0.00000170, 0.00000049, -0.00000250, -0.00000026, -0.00000407, 0.00000394, 0.00000264, 0.00000063, -0.00000025, 0.00000130, -0.00000763, -0.00000051, 0.00000149, 0.00000004, -0.00000048, -0.00000229, 0.00000349, 0.00000148, -0.00000326, 0.00000368, -0.00000158, -0.00001902, 0.00001071, -0.00001151, 0.00002337, -0.00002310, 0.00001557, -0.00001744, 0.00001801, -0.00001976, 0.00001189, -0.00004646, 0.00008224, -0.00010273, 0.00007722, -0.00015091, 0.00011461, -0.00028799, -0.00013124, 0.00117098, -0.00065757, 0.00011552, -0.00012733, 0.00007184, -0.00008942, 0.00003164, -0.00000584, 0.00000377, -0.00002047, -0.00001407, 0.00002142, -0.00001132, 0.00000232, -0.00001695, 0.00002738, -0.00000458, -0.00000994, -0.00002615, 0.00003083, -0.00002307, 0.00002213, -0.00002381, 0.00002359, -0.00001042, 0.00000816, -0.00001711, 0.00001605, -0.00002002, 0.00001381, -0.00001944, 0.00001633, -0.00000041, 0.00000619, -0.00001188, 0.00000709, -0.00000597, 0.00000695, -0.00000828, 0.00000827, -0.00000349, 0.00000041, -0.00000361, -0.00000008, -0.00000090, 0.00000023, -0.00000042, 0.00000390, -0.00000100, 0.00000301, -0.00000333, -0.00000135, 0.00000029, -0.00000127, 0.00000081, -0.00000047, -0.00000041, 0.00000045, -0.00000091, 0.00000266, 0.00000023, 0.00000044, -0.00000033, -0.00000189, -0.00000068, -0.00000129, -0.00000020, 0.00000061, 0.00000044, 0.00000002, 0.00000031, -0.00000120, 0.00000059, 0.00000097, 0.00000065, },
		{ -0.00000148, -0.00000121, 0.00000166, -0.00000154, 0.00000056, 0.00000182, -0.00000040, -0.00000250, 0.00000040, -0.00000037, 0.00000081, -0.00000106, -0.00000043, 0.00000178, 0.00000044, -0.00000258, -0.00000031, -0.00000412, 0.00000444, 0.00000247, 0.00000025, -0.00000019, 0.00000149, -0.00000805, 0.00000017, 0.00000100, 0.00000065, -0.00000078, -0.00000229, 0.00000299, 0.00000236, -0.00000428, 0.00000505, -0.00000425, -0.00001627, 0.00000958, -0.00001044, 0.00002127, -0.00002096, 0.00001423, -0.00001629, 0.00001587, -0.00001674, 0.00000881, -0.00004204, 0.00007644, -0.00009827, 0.00007516, -0.00015224, 0.00011778, -0.00030374, -0.00008786, 0.00112861, -0.00064555, 0.00011686, -0.00012969, 0.00007682, -0.00009490, 0.00003939, -0.00001457, 0.00001070, -0.00002614, -0.00000960, 0.00001828, -0.00000907, -0.00000006, -0.00001422, 0.00002532, -0.00000295, -0.00001277, -0.00002311, 0.00002909, -0.00002202, 0.00002154, -0.00002352, 0.00002355, -0.00001028, 0.00000785, -0.00001709, 0.00001592, -0.00002012, 0.00001426, -0.00002018, 0.00001747, -0.00000099, 0.00000637, -0.00001252, 0.00000776, -0.00000633, 0.00000742, -0.00000885, 0.00000891, -0.00000417, 0.00000104, -0.00000434, 0.00000034, -0.00000106, 0.00000041, -0.00000039, 0.00000383, -0.00000094, 0.00000323, -0.00000387, -0.00000096, 0.00000010, -0.00000120, 0.00000085, -0.00000072, -0.00000019, 0.00000036, -0.00000087, 0.00000279, 0.00000008, 0.00000056, -0.00000043, -0.00000191, -0.00000059, -0.00000139, -0.00000009, 0.00000058, 0.00000035, 0.00000004, 0.00000025, -0.00000117, 0.00000072, 0.00000083, 0.00000085, -0.00000054, },
		{ -0.00000114, 0.00000161, -0.00000154, 0.00000057, 0.00000165, -0.00000034, -0.00000265, 0.00000070, -0.00000053, 0.00000098, -0.00000099, -0.00000058, 0.00000185, 0.00000038, -0.00000264, -0.00000039, -0.00000414, 0.00000490, 0.00000229, -0.00000012, -0.00000010, 0.00000162, -0.00000839, 0.00000080, 0.00000057, 0.00000119, -0.00000104, -0.00000236, 0.00000259, 0.00000311, -0.00000514, 0.00000625, -0.00000676, -0.00001354, 0.00000840, -0.00000935, 0.00001914, -0.00001874, 0.00001280, -0.00001513, 0.00001379, -0.00001377, 0.00000569, -0.00003740, 0.00007005, -0.00009283, 0.00007184, -0.00015203, 0.00011901, -0.00031722, -0.00004544, 0.00108446, -0.00063075, 0.00011617, -0.00013050, 0.00008045, -0.00009920, 0.00004626, -0.00002283, 0.00001740, -0.00003174, -0.00000502, 0.00001498, -0.00000674, -0.00000248, -0.00001140, 0.00002317, -0.00000133, -0.00001559, -0.00001990, 0.00002707, -0.00002071, 0.00002071, -0.00002301, 0.00002334, -0.00001002, 0.00000742, -0.00001692, 0.00001565, -0.00002005, 0.00001454, -0.00002077, 0.00001853, -0.00000157, 0.00000652, -0.00001310, 0.00000841, -0.00000667, 0.00000785, -0.00000938, 0.00000951, -0.00000486, 0.00000170, -0.00000512, 0.00000082, -0.00000127, 0.00000064, -0.00000037, 0.00000375, -0.00000086, 0.00000342, -0.00000441, -0.00000054, -0.00000013, -0.00000109, 0.00000084, -0.00000095, 0.00000004, 0.00000026, -0.00000080, 0.00000290, -0.00000006, 0.00000070, -0.00000054, -0.00000192, -0.00000052, -0.00000148, 0.00000002, 0.00000054, 0.00000027, 0.00000007, 0.00000018, -0.00000111, 0.00000082, 0.00000071, 0.00000106, -0.00000073, -0.00000141, },
		{ 0.00000155, -0.00000153, 0.00000057, 0.00000149, -0.00000030, -0.00000278, 0.00000101, -0.00000072, 0.00000118, -0.00000095, -0.00000071, 0.00000190, 0.00000031, -0.00000267, -0.00000048, -0.00000411, 0.00000533, 0.00000209, -0.00000047, 0.00000002, 0.00000169, -0.00000865, 0.00000137, 0.00000018, 0.00000167, -0.00000125, -0.00000248, 0.00000228, 0.00000374, -0.00000585, 0.00000725, -0.00000910, -0.00001085, 0.00000719, -0.00000826, 0.00001701, -0.00001645, 0.00001129, -0.00001398, 0.00001182, -0.00001089, 0.00000260, -0.00003264, 0.00006317, -0.00008653, 0.00006733, -0.00015029, 0.00011827, -0.00032827, -0.00000433, 0.00103892, -0.00061342, 0.00011355, -0.00012980, 0.00008274, -0.00010230, 0.00005217, -0.00003052, 0.00002377, -0.00003716, -0.00000041, 0.00001155, -0.00000435, -0.00000491, -0.00000851, 0.00002094, 0.00000026, -0.00001836, -0.00001657, 0.00002481, -0.00001917, 0.00001965, -0.00002228, 0.00002294, -0.00000964, 0.00000688, -0.00001662, 0.00001524, -0.00001983, 0.00001466, -0.00002121, 0.00001950, -0.00000213, 0.00000664, -0.00001363, 0.00000902, -0.00000698, 0.00000822, -0.00000985, 0.00001007, -0.00000554, 0.00000237, -0.00000594, 0.00000137, -0.00000154, 0.00000090, -0.00000039, 0.00000368, -0.00000077, 0.00000358, -0.00000495, -0.00000008, -0.00000040, -0.00000093, 0.00000080, -0.00000115, 0.00000028, 0.00000013, -0.00000070, 0.00000299, -0.00000020, 0.00000084, -0.00000068, -0.00000190, -0.00000045, -0.00000158, 0.00000015, 0.00000047, 0.00000020, 0.00000009, 0.00000009, -0.00000102, 0.00000090, 0.00000060, 0.00000126, -0.00000093, -0.00000131, -0.00000110, },
		{ -0.00000152, 0.00000056, 0.00000134, -0.00000028, -0.00000289, 0.00000131, -0.00000094, 0.00000141, -0.00000096, -0.00000080, 0.00000193, 0.00000023, -0.00000270, -0.00000059, -0.00000405, 0.00000573, 0.00000187, -0.00000079, 0.00000016, 0.00000170, -0.00000884, 0.00000189, -0.00000015, 0.00000209, -0.00000143, -0.00000266, 0.00000207, 0.00000424, -0.00000641, 0.00000805, -0.00001125, -0.00000824, 0.00000598, -0.00000720, 0.00001489, -0.00001412, 0.00000973, -0.00001284, 0.00000996, -0.00000816, -0.00000039, -0.00002784, 0.00005593, -0.00007946, 0.00006172, -0.00014707, 0.00011556, -0.00033675, 0.00003515, 0.00099238, -0.00059381, 0.00010915, -0.00012765, 0.00008369, -0.00010417, 0.00005703, -0.00003754, 0.00002972, -0.00004234, 0.00000417, 0.00000804, -0.00000193, -0.00000731, -0.00000559, 0.00001865, 0.00000179, -0.00002104, -0.00001315, 0.00002234, -0.00001740, 0.00001838, -0.00002133, 0.00002235, -0.00000914, 0.00000622, -0.00001618, 0.00001468, -0.00001945, 0.00001460, -0.00002146, 0.00002037, -0.00000267, 0.00000672, -0.00001410, 0.00000958, -0.00000725, 0.00000854, -0.00001025, 0.00001058, -0.00000620, 0.00000305, -0.00000678, 0.00000197, -0.00000185, 0.00000120, -0.00000043, 0.00000362, -0.00000065, 0.00000371, -0.00000547, 0.00000039, -0.00000071, -0.00000073, 0.00000071, -0.00000133, 0.00000050, 0.00000000, -0.00000058, 0.00000305, -0.00000034, 0.00000098, -0.00000083, -0.00000186, -0.00000041, -0.00000167, 0.00000029, 0.00000039, 0.00000015, 0.00000011, -0.00000001, -0.00000091, 0.00000096, 0.00000051, 0.00000145, -0.00000114, -0.00000118, -0.00000108, 0.00000150, },
		{ 0.00000054, 0.00000120, -0.00000026, -0.00000297, 0.00000162, -0.00000118, 0.00000168, -0.00000100, -0.00000085, 0.00000193, 0.00000016, -0.00000271, -0.00000071, -0.00000396, 0.00000610, 0.00000163, -0.00000108, 0.00000029, 0.00000167, -0.00000896, 0.00000237, -0.00000044, 0.00000244, -0.00000158, -0.00000287, 0.00000194, 0.00000462, -0.00000681, 0.00000866, -0.00001318, -0.00000572, 0.00000478, -0.00000617, 0.00001282, -0.00001178, 0.00000811, -0.00001173, 0.00000824, -0.00000560, -0.00000323, -0.00002311, 0.00004844, -0.00007177, 0.00005511, -0.00014246, 0.00011091, -0.00034257, 0.00007268, 0.00094524, -0.00057223, 0.00010311, -0.00012415, 0.00008337, -0.00010481, 0.00006081, -0.00004382, 0.00003516, -0.00004720, 0.00000865, 0.00000451, 0.00000049, -0.00000966, -0.00000266, 0.00001633, 0.00000324, -0.00002360, -0.00000970, 0.00001969, -0.00001544, 0.00001691, -0.00002018, 0.00002159, -0.00000853, 0.00000546, -0.00001561, 0.00001398, -0.00001891, 0.00001437, -0.00002154, 0.00002111, -0.00000317, 0.00000675, -0.00001449, 0.00001010, -0.00000747, 0.00000879, -0.00001058, 0.00001103, -0.00000683, 0.00000371, -0.00000764, 0.00000262, -0.00000221, 0.00000154, -0.00000049, 0.00000356, -0.00000053, 0.00000380, -0.00000598, 0.00000089, -0.00000106, -0.00000048, 0.00000058, -0.00000147, 0.00000072, -0.00000014, -0.00000043, 0.00000309, -0.00000046, 0.00000113, -0.00000100, -0.00000180, -0.00000038, -0.00000176, 0.00000043, 0.00000029, 0.00000011, 0.00000012, -0.00000010, -0.00000078, 0.00000098, 0.00000044, 0.00000163, -0.00000136, -0.00000104, -0.00000109, 0.00000146, -0.00000149, },
		{ 0.00000107, -0.00000027, -0.00000301, 0.00000191, -0.00000142, 0.00000196, -0.00000109, -0.00000086, 0.00000189, 0.00000009, -0.00000272, -0.00000082, -0.00000384, 0.00000643, 0.00000138, -0.00000133, 0.00000042, 0.00000160, -0.00000903, 0.00000281, -0.00000069, 0.00000275, -0.00000169, -0.00000313, 0.00000189, 0.00000488, -0.00000707, 0.00000907, -0.00001489, -0.00000333, 0.00000362, -0.00000520, 0.00001080, -0.00000945, 0.00000647, -0.00001065, 0.00000667, -0.00000326, -0.00000586, -0.00001851, 0.00004083, -0.00006358, 0.00004763, -0.00013656, 0.00010437, -0.00034565, 0.00010801, 0.00089786, -0.00054895, 0.00009561, -0.00011940, 0.00008184, -0.00010426, 0.00006348, -0.00004929, 0.00004003, -0.00005165, 0.00001297, 0.00000098, 0.00000288, -0.00001194, 0.00000024, 0.00001400, 0.00000460, -0.00002601, -0.00000624, 0.00001690, -0.00001330, 0.00001526, -0.00001883, 0.00002065, -0.00000781, 0.00000460, -0.00001491, 0.00001314, -0.00001821, 0.00001397, -0.00002143, 0.00002173, -0.00000362, 0.00000674, -0.00001481, 0.00001056, -0.00000766, 0.00000898, -0.00001084, 0.00001141, -0.00000742, 0.00000436, -0.00000851, 0.00000331, -0.00000261, 0.00000192, -0.00000059, 0.00000351, -0.00000039, 0.00000386, -0.00000646, 0.00000139, -0.00000145, -0.00000018, 0.00000039, -0.00000156, 0.00000093, -0.00000029, -0.00000027, 0.00000310, -0.00000056, 0.00000127, -0.00000118, -0.00000172, -0.00000037, -0.00000183, 0.00000058, 0.00000017, 0.00000009, 0.00000012, -0.00000020, -0.00000062, 0.00000098, 0.00000040, 0.00000180, -0.00000157, -0.00000087, -0.00000112, 0.00000143, -0.00000147, 0.00000051, },
		{ -0.00000029, -0.00000302, 0.00000218, -0.00000167, 0.00000227, -0.00000121, -0.00000082, 0.00000181, 0.00000004, -0.00000272, -0.00000094, -0.00000369, 0.00000673, 0.00000111, -0.00000153, 0.00000054, 0.00000151, -0.00000904, 0.00000322, -0.00000092, 0.00000300, -0.00000178, -0.00000341, 0.00000190, 0.00000504, -0.00000719, 0.00000928, -0.00001636, -0.00000107, 0.00000250, -0.00000429, 0.00000887, -0.00000715, 0.00000480, -0.00000960, 0.00000527, -0.00000116, -0.00000823, -0.00001415, 0.00003321, -0.00005503, 0.00003942, -0.00012947, 0.00009604, -0.00034598, 0.00014088, 0.00085061, -0.00052428, 0.00008685, -0.00011352, 0.00007917, -0.00010256, 0.00006503, -0.00005391, 0.00004425, -0.00005563, 0.00001705, -0.00000248, 0.00000520, -0.00001411, 0.00000310, 0.00001167, 0.00000584, -0.00002823, -0.00000283, 0.00001402, -0.00001103, 0.00001345, -0.00001731, 0.00001955, -0.00000699, 0.00000365, -0.00001409, 0.00001217, -0.00001736, 0.00001339, -0.00002112, 0.00002220, -0.00000402, 0.00000666, -0.00001505, 0.00001096, -0.00000779, 0.00000910, -0.00001101, 0.00001170, -0.00000795, 0.00000497, -0.00000937, 0.00000403, -0.00000305, 0.00000234, -0.00000073, 0.00000348, -0.00000024, 0.00000387, -0.00000691, 0.00000190, -0.00000186, 0.00000015, 0.00000016, -0.00000162, 0.00000111, -0.00000043, -0.00000008, 0.00000308, -0.00000064, 0.00000141, -0.00000137, -0.00000161, -0.00000038, -0.00000189, 0.00000073, 0.00000004, 0.00000009, 0.00000011, -0.00000030, -0.00000045, 0.00000095, 0.00000038, 0.00000193, -0.00000178, -0.00000070, -0.00000118, 0.00000141, -0.00000144, 0.00000048, 0.00000094, },
		{ -0.00000300, 0.00000243, -0.00000192, 0.00000258, -0.00000136, -0.00000075, 0.00000169, 0.00000001, -0.00000273, -0.00000106, -0.00000351, 0.00000699, 0.00000082, -0.00000168, 0.00000063, 0.00000140, -0.00000901, 0.00000360, -0.00000112, 0.00000322, -0.00000185, -0.00000370, 0.00000197, 0.00000511, -0.00000718, 0.00000930, -0.00001760, 0.00000102, 0.00000145, -0.00000347, 0.00000704, -0.00000490, 0.00000313, -0.00000859, 0.00000403, 0.00000068, -0.00001031, -0.00001008, 0.00002570, -0.00004627, 0.00003061, -0.00012136, 0.00008604, -0.00034357, 0.00017109, 0.00080385, -0.00049851, 0.00007701, -0.00010664, 0.00007547, -0.00009977, 0.00006547, -0.00005763, 0.00004777, -0.00005908, 0.00002085, -0.00000583, 0.00000743, -0.00001616, 0.00000589, 0.00000937, 0.00000695, -0.00003024, 0.00000050, 0.00001107, -0.00000865, 0.00001150, -0.00001562, 0.00001830, -0.00000608, 0.00000262, -0.00001317, 0.00001108, -0.00001637, 0.00001264, -0.00002061, 0.00002251, -0.00000435, 0.00000653, -0.00001520, 0.00001129, -0.00000787, 0.00000915, -0.00001109, 0.00001191, -0.00000841, 0.00000553, -0.00001020, 0.00000478, -0.00000353, 0.00000280, -0.00000090, 0.00000347, -0.00000008, 0.00000384, -0.00000733, 0.00000240, -0.00000229, 0.00000052, -0.00000012, -0.00000162, 0.00000127, -0.00000057, 0.00000011, 0.00000303, -0.00000071, 0.00000154, -0.00000157, -0.00000149, -0.00000041, -0.00000194, 0.00000087, -0.00000011, 0.00000011, 0.00000008, -0.00000038, -0.00000026, 0.00000089, 0.00000039, 0.00000205, -0.00000197, -0.00000051, -0.00000125, 0.00000140, -0.00000142, 0.00000044, 0.00000082, -0.00000032, },
		{ 0.00000265, -0.00000217, 0.00000291, -0.00000154, -0.00000063, 0.00000153, 0.00000000, -0.00000275, -0.00000117, -0.00000331, 0.00000721, 0.00000053, -0.00000179, 0.00000070, 0.00000129, -0.00000894, 0.00000396, -0.00000130, 0.00000339, -0.00000192, -0.00000400, 0.00000207, 0.00000510, -0.00000706, 0.00000915, -0.00001859, 0.00000295, 0.00000048, -0.00000274, 0.00000533, -0.00000273, 0.00000147, -0.00000762, 0.00000297, 0.00000224, -0.00001206, -0.00000637, 0.00001841, -0.00003743, 0.00002136, -0.00011235, 0.00007450, -0.00033847, 0.00019848, 0.00075788, -0.00047195, 0.00006631, -0.00009892, 0.00007086, -0.00009598, 0.00006483, -0.00006043, 0.00005056, -0.00006196, 0.00002429, -0.00000902, 0.00000954, -0.00001805, 0.00000858, 0.00000712, 0.00000791, -0.00003201, 0.00000370, 0.00000810, -0.00000620, 0.00000945, -0.00001380, 0.00001691, -0.00000508, 0.00000151, -0.00001215, 0.00000989, -0.00001525, 0.00001173, -0.00001990, 0.00002266, -0.00000461, 0.00000633, -0.00001526, 0.00001156, -0.00000790, 0.00000913, -0.00001107, 0.00001202, -0.00000880, 0.00000603, -0.00001100, 0.00000553, -0.00000403, 0.00000329, -0.00000110, 0.00000348, 0.00000007, 0.00000378, -0.00000770, 0.00000289, -0.00000273, 0.00000092, -0.00000045, -0.00000158, 0.00000140, -0.00000070, 0.00000032, 0.00000296, -0.00000075, 0.00000166, -0.00000178, -0.00000135, -0.00000047, -0.00000196, 0.00000101, -0.00000027, 0.00000014, 0.00000004, -0.00000046, -0.00000007, 0.00000082, 0.00000043, 0.00000213, -0.00000215, -0.00000032, -0.00000135, 0.00000141, -0.00000140, 0.00000040, 0.00000070, -0.00000037, -0.00000293, },
		{ -0.00000239, 0.00000323, -0.00000175, -0.00000047, 0.00000132, 0.00000002, -0.00000277, -0.00000127, -0.00000310, 0.00000740, 0.00000022, -0.00000186, 0.00000073, 0.00000117, -0.00000884, 0.00000430, -0.00000147, 0.00000354, -0.00000198, -0.00000429, 0.00000220, 0.00000502, -0.00000683, 0.00000883, -0.00001935, 0.00000470, -0.00000040, -0.00000210, 0.00000374, -0.00000064, -0.00000018, -0.00000668, 0.00000207, 0.00000352, -0.00001345, -0.00000308, 0.00001144, -0.00002866, 0.00001183, -0.00010263, 0.00006159, -0.00033075, 0.00022292, 0.00071301, -0.00044486, 0.00005494, -0.00009051, 0.00006546, -0.00009128, 0.00006316, -0.00006233, 0.00005260, -0.00006421, 0.00002735, -0.00001201, 0.00001149, -0.00001977, 0.00001115, 0.00000493, 0.00000870, -0.00003352, 0.00000674, 0.00000515, -0.00000370, 0.00000732, -0.00001186, 0.00001540, -0.00000402, 0.00000035, -0.00001105, 0.00000859, -0.00001401, 0.00001067, -0.00001900, 0.00002265, -0.00000479, 0.00000607, -0.00001522, 0.00001175, -0.00000788, 0.00000904, -0.00001097, 0.00001203, -0.00000910, 0.00000646, -0.00001175, 0.00000628, -0.00000455, 0.00000380, -0.00000133, 0.00000351, 0.00000023, 0.00000367, -0.00000802, 0.00000336, -0.00000319, 0.00000135, -0.00000082, -0.00000148, 0.00000149, -0.00000081, 0.00000054, 0.00000286, -0.00000077, 0.00000176, -0.00000198, -0.00000118, -0.00000055, -0.00000197, 0.00000114, -0.00000044, 0.00000020, -0.00000002, -0.00000052, 0.00000013, 0.00000072, 0.00000049, 0.00000219, -0.00000230, -0.00000013, -0.00000146, 0.00000143, -0.00000139, 0.00000036, 0.00000058, -0.00000044, -0.00000282, 0.00000284, },
		{ 0.00000354, -0.00000197, -0.00000028, 0.00000107, 0.00000007, -0.00000281, -0.00000135, -0.00000287, 0.00000755, -0.00000009, -0.00000187, 0.00000073, 0.00000105, -0.00000871, 0.00000464, -0.00000164, 0.00000366, -0.00000204, -0.00000457, 0.00000235, 0.00000489, -0.00000653, 0.00000837, -0.00001988, 0.00000626, -0.00000119, -0.00000157, 0.00000228, 0.00000134, -0.00000179, -0.00000577, 0.00000133, 0.00000452, -0.00001447, -0.00000025, 0.00000489, -0.00002008, 0.00000217, -0.00009237, 0.00004750, -0.00032054, 0.00024433, 0.00066951, -0.00041754, 0.00004311, -0.00008156, 0.00005941, -0.00008580, 0.00006052, -0.00006334, 0.00005386, -0.00006581, 0.00002996, -0.00001477, 0.00001327, -0.00002129, 0.00001358, 0.00000282, 0.00000934, -0.00003474, 0.00000957, 0.00000226, -0.00000119, 0.00000514, -0.00000983, 0.00001378, -0.00000289, -0.00000085, -0.00000988, 0.00000722, -0.00001266, 0.00000946, -0.00001790, 0.00002246, -0.00000488, 0.00000573, -0.00001509, 0.00001186, -0.00000780, 0.00000887, -0.00001077, 0.00001193, -0.00000931, 0.00000681, -0.00001243, 0.00000702, -0.00000509, 0.00000434, -0.00000160, 0.00000356, 0.00000039, 0.00000352, -0.00000830, 0.00000380, -0.00000364, 0.00000179, -0.00000123, -0.00000133, 0.00000154, -0.00000090, 0.00000075, 0.00000275, -0.00000076, 0.00000184, -0.00000218, -0.00000101, -0.00000065, -0.00000195, 0.00000125, -0.00000062, 0.00000027, -0.00000010, -0.00000055, 0.00000033, 0.00000061, 0.00000058, 0.00000221, -0.00000244, 0.00000005, -0.00000159, 0.00000146, -0.00000138, 0.00000032, 0.00000047, -0.00000051, -0.00000267, 0.00000298, -0.00000259, },
		{ -0.00000220, -0.00000007, 0.00000077, 0.00000014, -0.00000287, -0.00000142, -0.00000262, 0.00000766, -0.00000039, -0.00000184, 0.00000068, 0.00000095, -0.00000856, 0.00000496, -0.00000181, 0.00000376, -0.00000211, -0.00000482, 0.00000250, 0.00000471, -0.00000615, 0.00000777, -0.00002019, 0.00000763, -0.00000187, -0.00000114, 0.00000097, 0.00000319, -0.00000337, -0.00000489, 0.00000074, 0.00000524, -0.00001512, 0.00000209, -0.00000116, -0.00001183, -0.00000746, -0.00008174, 0.00003243, -0.00030800, 0.00026265, 0.00062761, -0.00039023, 0.00003103, -0.00007224, 0.00005284, -0.00007964, 0.00005700, -0.00006348, 0.00005436, -0.00006674, 0.00003211, -0.00001726, 0.00001485, -0.00002260, 0.00001585, 0.00000079, 0.00000979, -0.00003566, 0.00001218, -0.00000054, 0.00000129, 0.00000294, -0.00000774, 0.00001208, -0.00000172, -0.00000208, -0.00000867, 0.00000579, -0.00001123, 0.00000813, -0.00001663, 0.00002210, -0.00000488, 0.00000532, -0.00001486, 0.00001190, -0.00000766, 0.00000864, -0.00001048, 0.00001173, -0.00000942, 0.00000707, -0.00001304, 0.00000772, -0.00000562, 0.00000490, -0.00000189, 0.00000364, 0.00000054, 0.00000334, -0.00000852, 0.00000421, -0.00000408, 0.00000225, -0.00000168, -0.00000113, 0.00000155, -0.00000096, 0.00000097, 0.00000261, -0.00000073, 0.00000189, -0.00000238, -0.00000082, -0.00000078, -0.00000191, 0.00000135, -0.00000080, 0.00000036, -0.00000020, -0.00000057, 0.00000052, 0.00000049, 0.00000069, 0.00000219, -0.00000254, 0.00000022, -0.00000172, 0.00000150, -0.00000138, 0.00000028, 0.00000035, -0.00000059, -0.00000247, 0.00000307, -0.00000276, 0.00000384, },
		{ 0.00000018, 0.00000044, 0.00000025, -0.00000294, -0.00000146, -0.00000237, 0.00000773, -0.00000070, -0.00000177, 0.00000060, 0.00000086, -0.00000839, 0.00000528, -0.00000199, 0.00000385, -0.00000219, -0.00000505, 0.00000265, 0.00000451, -0.00000572, 0.00000706, -0.00002029, 0.00000882, -0.00000246, -0.00000081, -0.00000018, 0.00000492, -0.00000491, -0.00000404, 0.00000028, 0.00000569, -0.00001540, 0.00000390, -0.00000664, -0.00000403, -0.00001691, -0.00007093, 0.00001660, -0.00029330, 0.00027788, 0.00058752, -0.00036318, 0.00001888, -0.00006270, 0.00004590, -0.00007296, 0.00005269, -0.00006279, 0.00005410, -0.00006699, 0.00003375, -0.00001944, 0.00001621, -0.00002368, 0.00001794, -0.00000113, 0.00001008, -0.00003628, 0.00001453, -0.00000321, 0.00000372, 0.00000075, -0.00000560, 0.00001032, -0.00000053, -0.00000332, -0.00000741, 0.00000431, -0.00000972, 0.00000670, -0.00001518, 0.00002158, -0.00000479, 0.00000484, -0.00001454, 0.00001185, -0.00000748, 0.00000835, -0.00001010, 0.00001142, -0.00000943, 0.00000722, -0.00001356, 0.00000839, -0.00000615, 0.00000546, -0.00000221, 0.00000375, 0.00000068, 0.00000312, -0.00000868, 0.00000458, -0.00000450, 0.00000272, -0.00000215, -0.00000087, 0.00000152, -0.00000100, 0.00000118, 0.00000245, -0.00000067, 0.00000193, -0.00000256, -0.00000063, -0.00000092, -0.00000184, 0.00000143, -0.00000097, 0.00000046, -0.00000032, -0.00000056, 0.00000070, 0.00000036, 0.00000082, 0.00000215, -0.00000261, 0.00000038, -0.00000187, 0.00000155, -0.00000139, 0.00000026, 0.00000022, -0.00000068, -0.00000224, 0.00000312, -0.00000289, 0.00000411, -0.00000243, },
		{ 0.00000006, 0.00000039, -0.00000303, -0.00000149, -0.00000211, 0.00000775, -0.00000099, -0.00000167, 0.00000047, 0.00000079, -0.00000821, 0.00000560, -0.00000218, 0.00000394, -0.00000229, -0.00000523, 0.00000277, 0.00000430, -0.00000525, 0.00000627, -0.00002020, 0.00000982, -0.00000295, -0.00000059, -0.00000119, 0.00000649, -0.00000638, -0.00000320, -0.00000006, 0.00000588, -0.00001532, 0.00000518, -0.00001149, 0.00000324, -0.00002603, -0.00006012, 0.00000023, -0.00027664, 0.00029005, 0.00054940, -0.00033661, 0.00000685, -0.00005310, 0.00003873, -0.00006588, 0.00004769, -0.00006135, 0.00005312, -0.00006657, 0.00003487, -0.00002130, 0.00001733, -0.00002451, 0.00001984, -0.00000295, 0.00001018, -0.00003658, 0.00001659, -0.00000571, 0.00000605, -0.00000140, -0.00000346, 0.00000851, 0.00000068, -0.00000456, -0.00000614, 0.00000279, -0.00000817, 0.00000517, -0.00001358, 0.00002088, -0.00000461, 0.00000429, -0.00001412, 0.00001173, -0.00000723, 0.00000799, -0.00000964, 0.00001101, -0.00000934, 0.00000727, -0.00001398, 0.00000900, -0.00000665, 0.00000603, -0.00000255, 0.00000388, 0.00000081, 0.00000286, -0.00000878, 0.00000490, -0.00000490, 0.00000318, -0.00000265, -0.00000057, 0.00000144, -0.00000100, 0.00000137, 0.00000228, -0.00000059, 0.00000194, -0.00000273, -0.00000044, -0.00000108, -0.00000175, 0.00000148, -0.00000115, 0.00000058, -0.00000046, -0.00000052, 0.00000087, 0.00000023, 0.00000096, 0.00000207, -0.00000266, 0.00000052, -0.00000201, 0.00000162, -0.00000142, 0.00000024, 0.00000009, -0.00000076, -0.00000198, 0.00000313, -0.00000298, 0.00000435, -0.00000266, 0.00000044, },
		{ 0.00000056, -0.00000314, -0.00000149, -0.00000185, 0.00000774, -0.00000126, -0.00000153, 0.00000030, 0.00000073, -0.00000802, 0.00000592, -0.00000239, 0.00000401, -0.00000241, -0.00000536, 0.00000287, 0.00000408, -0.00000477, 0.00000540, -0.00001993, 0.00001065, -0.00000334, -0.00000046, -0.00000204, 0.00000792, -0.00000780, -0.00000237, -0.00000030, 0.00000585, -0.00001490, 0.00000594, -0.00001569, 0.00000986, -0.00003470, -0.00004950, -0.00001644, -0.00025826, 0.00029922, 0.00051338, -0.00031070, -0.00000489, -0.00004357, 0.00003147, -0.00005854, 0.00004213, -0.00005922, 0.00005145, -0.00006548, 0.00003546, -0.00002281, 0.00001820, -0.00002510, 0.00002153, -0.00000464, 0.00001012, -0.00003657, 0.00001835, -0.00000802, 0.00000827, -0.00000349, -0.00000132, 0.00000668, 0.00000189, -0.00000578, -0.00000487, 0.00000127, -0.00000658, 0.00000357, -0.00001184, 0.00002003, -0.00000434, 0.00000367, -0.00001360, 0.00001152, -0.00000694, 0.00000758, -0.00000909, 0.00001050, -0.00000914, 0.00000721, -0.00001429, 0.00000955, -0.00000712, 0.00000658, -0.00000291, 0.00000404, 0.00000093, 0.00000258, -0.00000882, 0.00000518, -0.00000527, 0.00000363, -0.00000317, -0.00000023, 0.00000131, -0.00000097, 0.00000155, 0.00000211, -0.00000049, 0.00000192, -0.00000288, -0.00000024, -0.00000126, -0.00000163, 0.00000152, -0.00000131, 0.00000070, -0.00000062, -0.00000046, 0.00000101, 0.00000010, 0.00000112, 0.00000197, -0.00000267, 0.00000063, -0.00000215, 0.00000169, -0.00000145, 0.00000023, -0.00000005, -0.00000084, -0.00000168, 0.00000308, -0.00000303, 0.00000455, -0.00000287, 0.00000070, -0.00000034, },
		{ -0.00000326, -0.00000146, -0.00000158, 0.00000768, -0.00000152, -0.00000136, 0.00000009, 0.00000071, -0.00000781, 0.00000625, -0.00000261, 0.00000409, -0.00000255, -0.00000544, 0.00000293, 0.00000388, -0.00000429, 0.00000449, -0.00001951, 0.00001131, -0.00000364, -0.00000042, -0.00000275, 0.00000918, -0.00000914, -0.00000155, -0.00000046, 0.00000561, -0.00001415, 0.00000618, -0.00001918, 0.00001577, -0.00004278, -0.00003922, -0.00003318, -0.00023841, 0.00030550, 0.00047958, -0.00028563, -0.00001620, -0.00003425, 0.00002425, -0.00005109, 0.00003611, -0.00005649, 0.00004915, -0.00006376, 0.00003553, -0.00002396, 0.00001881, -0.00002542, 0.00002301, -0.00000622, 0.00000989, -0.00003625, 0.00001980, -0.00001012, 0.00001033, -0.00000548, 0.00000077, 0.00000485, 0.00000307, -0.00000696, -0.00000362, -0.00000025, -0.00000498, 0.00000193, -0.00000997, 0.00001903, -0.00000399, 0.00000299, -0.00001299, 0.00001124, -0.00000660, 0.00000711, -0.00000847, 0.00000988, -0.00000883, 0.00000703, -0.00001448, 0.00001002, -0.00000755, 0.00000713, -0.00000327, 0.00000422, 0.00000102, 0.00000227, -0.00000881, 0.00000540, -0.00000560, 0.00000406, -0.00000369, 0.00000015, 0.00000114, -0.00000090, 0.00000171, 0.00000192, -0.00000037, 0.00000187, -0.00000301, -0.00000005, -0.00000145, -0.00000149, 0.00000153, -0.00000146, 0.00000083, -0.00000080, -0.00000036, 0.00000113, -0.00000002, 0.00000129, 0.00000184, -0.00000264, 0.00000072, -0.00000228, 0.00000176, -0.00000149, 0.00000023, -0.00000021, -0.00000091, -0.00000135, 0.00000298, -0.00000303, 0.00000471, -0.00000306, 0.00000097, -0.00000077, 0.00000075, },
		{ -0.00000141, -0.00000132, 0.00000759, -0.00000175, -0.00000118, -0.00000015, 0.00000070, -0.00000760, 0.00000657, -0.00000286, 0.00000417, -0.00000273, -0.00000547, 0.00000294, 0.00000370, -0.00000382, 0.00000354, -0.00001895, 0.00001181, -0.00000385, -0.00000046, -0.00000331, 0.00001028, -0.00001041, -0.00000072, -0.00000056, 0.00000520, -0.00001313, 0.00000593, -0.00002198, 0.00002090, -0.00005018, -0.00002946, -0.00004975, -0.00021734, 0.00030902, 0.00044805, -0.00026155, -0.00002694, -0.00002527, 0.00001719, -0.00004366, 0.00002976, -0.00005323, 0.00004628, -0.00006143, 0.00003507, -0.00002475, 0.00001915, -0.00002549, 0.00002427, -0.00000768, 0.00000951, -0.00003562, 0.00002092, -0.00001198, 0.00001223, -0.00000736, 0.00000280, 0.00000304, 0.00000422, -0.00000809, -0.00000240, -0.00000174, -0.00000339, 0.00000025, -0.00000801, 0.00001788, -0.00000355, 0.00000224, -0.00001230, 0.00001088, -0.00000621, 0.00000660, -0.00000779, 0.00000918, -0.00000842, 0.00000674, -0.00001455, 0.00001040, -0.00000793, 0.00000764, -0.00000364, 0.00000442, 0.00000110, 0.00000193, -0.00000873, 0.00000555, -0.00000589, 0.00000446, -0.00000421, 0.00000056, 0.00000093, -0.00000078, 0.00000184, 0.00000174, -0.00000023, 0.00000179, -0.00000311, 0.00000013, -0.00000164, -0.00000133, 0.00000151, -0.00000160, 0.00000096, -0.00000100, -0.00000023, 0.00000122, -0.00000012, 0.00000145, 0.00000168, -0.00000259, 0.00000079, -0.00000241, 0.00000184, -0.00000155, 0.00000024, -0.00000038, -0.00000096, -0.00000101, 0.00000284, -0.00000297, 0.00000482, -0.00000323, 0.00000124, -0.00000122, 0.00000097, -0.00000341, },
		{ -0.00000106, 0.00000745, -0.00000195, -0.00000098, -0.00000042, 0.00000071, -0.00000737, 0.00000689, -0.00000313, 0.00000426, -0.00000293, -0.00000544, 0.00000290, 0.00000356, -0.00000338, 0.00000260, -0.00001827, 0.00001216, -0.00000399, -0.00000057, -0.00000374, 0.00001122, -0.00001160, 0.00000012, -0.00000062, 0.00000465, -0.00001186, 0.00000522, -0.00002406, 0.00002522, -0.00005679, -0.00002036, -0.00006593, -0.00019533, 0.00030994, 0.00041883, -0.00023857, -0.00003698, -0.00001672, 0.00001042, -0.00003639, 0.00002322, -0.00004955, 0.00004290, -0.00005855, 0.00003410, -0.00002516, 0.00001923, -0.00002529, 0.00002531, -0.00000901, 0.00000899, -0.00003471, 0.00002171, -0.00001358, 0.00001394, -0.00000910, 0.00000475, 0.00000127, 0.00000532, -0.00000915, -0.00000123, -0.00000318, -0.00000182, -0.00000143, -0.00000595, 0.00001661, -0.00000303, 0.00000145, -0.00001152, 0.00001045, -0.00000578, 0.00000605, -0.00000704, 0.00000839, -0.00000792, 0.00000633, -0.00001449, 0.00001069, -0.00000825, 0.00000812, -0.00000400, 0.00000464, 0.00000115, 0.00000158, -0.00000859, 0.00000565, -0.00000612, 0.00000482, -0.00000472, 0.00000100, 0.00000068, -0.00000063, 0.00000195, 0.00000156, -0.00000007, 0.00000168, -0.00000319, 0.00000030, -0.00000184, -0.00000114, 0.00000146, -0.00000173, 0.00000109, -0.00000121, -0.00000007, 0.00000128, -0.00000021, 0.00000162, 0.00000151, -0.00000250, 0.00000082, -0.00000252, 0.00000192, -0.00000161, 0.00000027, -0.00000057, -0.00000100, -0.00000065, 0.00000266, -0.00000287, 0.00000487, -0.00000335, 0.00000148, -0.00000168, 0.00000121, -0.00000357, -0.00000134, },
		{ 0.00000728, -0.00000212, -0.00000078, -0.00000072, 0.00000074, -0.00000713, 0.00000720, -0.00000342, 0.00000435, -0.00000315, -0.00000534, 0.00000280, 0.00000345, -0.00000297, 0.00000166, -0.00001750, 0.00001239, -0.00000407, -0.00000074, -0.00000404, 0.00001199, -0.00001270, 0.00000097, -0.00000068, 0.00000398, -0.00001039, 0.00000410, -0.00002547, 0.00002869, -0.00006255, -0.00001205, -0.00008150, -0.00017265, 0.00030843, 0.00039191, -0.00021679, -0.00004625, -0.00000871, 0.00000402, -0.00002940, 0.00001660, -0.00004555, 0.00003911, -0.00005517, 0.00003266, -0.00002522, 0.00001904, -0.00002484, 0.00002612, -0.00001022, 0.00000834, -0.00003351, 0.00002217, -0.00001492, 0.00001544, -0.00001068, 0.00000658, -0.00000044, 0.00000635, -0.00001014, -0.00000012, -0.00000457, -0.00000031, -0.00000309, -0.00000384, 0.00001523, -0.00000244, 0.00000060, -0.00001067, 0.00000994, -0.00000531, 0.00000547, -0.00000625, 0.00000753, -0.00000732, 0.00000581, -0.00001430, 0.00001087, -0.00000849, 0.00000855, -0.00000435, 0.00000488, 0.00000117, 0.00000121, -0.00000840, 0.00000569, -0.00000630, 0.00000514, -0.00000521, 0.00000146, 0.00000039, -0.00000044, 0.00000202, 0.00000138, 0.00000010, 0.00000155, -0.00000323, 0.00000046, -0.00000204, -0.00000094, 0.00000139, -0.00000183, 0.00000122, -0.00000142, 0.00000012, 0.00000131, -0.00000027, 0.00000177, 0.00000133, -0.00000239, 0.00000083, -0.00000262, 0.00000199, -0.00000168, 0.00000030, -0.00000078, -0.00000101, -0.00000028, 0.00000243, -0.00000271, 0.00000487, -0.00000344, 0.00000171, -0.00000214, 0.00000146, -0.00000375, -0.00000123, -0.00000080, },
		{ -0.00000225, -0.00000057, -0.00000104, 0.00000079, -0.00000688, 0.00000751, -0.00000372, 0.00000444, -0.00000340, -0.00000518, 0.00000264, 0.00000339, -0.00000262, 0.00000076, -0.00001666, 0.00001249, -0.00000408, -0.00000095, -0.00000422, 0.00001259, -0.00001372, 0.00000185, -0.00000074, 0.00000324, -0.00000877, 0.00000261, -0.00002621, 0.00003130, -0.00006740, -0.00000464, -0.00009626, -0.00014959, 0.00030472, 0.00036727, -0.00019626, -0.00005465, -0.00000131, -0.00000191, -0.00002281, 0.00001003, -0.00004132, 0.00003497, -0.00005136, 0.00003077, -0.00002493, 0.00001860, -0.00002413, 0.00002670, -0.00001131, 0.00000758, -0.00003206, 0.00002230, -0.00001598, 0.00001672, -0.00001209, 0.00000829, -0.00000207, 0.00000731, -0.00001103, 0.00000090, -0.00000587, 0.00000114, -0.00000470, -0.00000170, 0.00001375, -0.00000180, -0.00000028, -0.00000975, 0.00000938, -0.00000481, 0.00000487, -0.00000542, 0.00000660, -0.00000665, 0.00000519, -0.00001397, 0.00001094, -0.00000866, 0.00000893, -0.00000468, 0.00000513, 0.00000117, 0.00000083, -0.00000815, 0.00000565, -0.00000641, 0.00000541, -0.00000568, 0.00000192, 0.00000007, -0.00000020, 0.00000206, 0.00000122, 0.00000027, 0.00000139, -0.00000325, 0.00000059, -0.00000224, -0.00000073, 0.00000129, -0.00000191, 0.00000134, -0.00000165, 0.00000034, 0.00000130, -0.00000030, 0.00000191, 0.00000114, -0.00000225, 0.00000081, -0.00000270, 0.00000207, -0.00000176, 0.00000036, -0.00000101, -0.00000098, 0.00000009, 0.00000218, -0.00000251, 0.00000482, -0.00000348, 0.00000191, -0.00000261, 0.00000173, -0.00000394, -0.00000110, -0.00000056, 0.00000707, },
		{ -0.00000038, -0.00000138, 0.00000086, -0.00000661, 0.00000781, -0.00000405, 0.00000454, -0.00000368, -0.00000497, 0.00000242, 0.00000338, -0.00000233, -0.00000010, -0.00001576, 0.00001250, -0.00000406, -0.00000119, -0.00000431, 0.00001303, -0.00001464, 0.00000274, -0.00000083, 0.00000246, -0.00000704, 0.00000081, -0.00002634, 0.00003306, -0.00007130, 0.00000178, -0.00011003, -0.00012642, 0.00029900, 0.00034485, -0.00017703, -0.00006213, 0.00000544, -0.00000730, -0.00001672, 0.00000362, -0.00003698, 0.00003058, -0.00004716, 0.00002847, -0.00002430, 0.00001791, -0.00002319, 0.00002706, -0.00001229, 0.00000673, -0.00003038, 0.00002213, -0.00001676, 0.00001776, -0.00001330, 0.00000985, -0.00000361, 0.00000818, -0.00001183, 0.00000184, -0.00000709, 0.00000250, -0.00000626, 0.00000046, 0.00001219, -0.00000110, -0.00000120, -0.00000877, 0.00000875, -0.00000427, 0.00000424, -0.00000457, 0.00000562, -0.00000590, 0.00000446, -0.00001352, 0.00001089, -0.00000874, 0.00000925, -0.00000498, 0.00000538, 0.00000115, 0.00000044, -0.00000785, 0.00000556, -0.00000646, 0.00000562, -0.00000611, 0.00000239, -0.00000028, 0.00000007, 0.00000207, 0.00000107, 0.00000046, 0.00000120, -0.00000323, 0.00000070, -0.00000242, -0.00000051, 0.00000116, -0.00000197, 0.00000145, -0.00000188, 0.00000059, 0.00000126, -0.00000030, 0.00000203, 0.00000094, -0.00000209, 0.00000075, -0.00000276, 0.00000213, -0.00000184, 0.00000042, -0.00000126, -0.00000093, 0.00000045, 0.00000190, -0.00000225, 0.00000470, -0.00000348, 0.00000207, -0.00000306, 0.00000201, -0.00000414, -0.00000095, -0.00000033, 0.00000682, -0.00000234, },
		{ -0.00000172, 0.00000093, -0.00000633, 0.00000809, -0.00000438, 0.00000463, -0.00000397, -0.00000470, 0.00000215, 0.00000342, -0.00000210, -0.00000089, -0.00001482, 0.00001241, -0.00000400, -0.00000145, -0.00000430, 0.00001332, -0.00001546, 0.00000366, -0.00000096, 0.00000167, -0.00000526, -0.00000123, -0.00002591, 0.00003400, -0.00007423, 0.00000713, -0.00012262, -0.00010340, 0.00029153, 0.00032458, -0.00015912, -0.00006867, 0.00001147, -0.00001210, -0.00001122, -0.00000251, -0.00003262, 0.00002602, -0.00004268, 0.00002580, -0.00002336, 0.00001700, -0.00002201, 0.00002720, -0.00001315, 0.00000581, -0.00002849, 0.00002165, -0.00001726, 0.00001857, -0.00001432, 0.00001125, -0.00000505, 0.00000895, -0.00001251, 0.00000267, -0.00000820, 0.00000377, -0.00000773, 0.00000261, 0.00001057, -0.00000036, -0.00000213, -0.00000774, 0.00000807, -0.00000372, 0.00000361, -0.00000369, 0.00000460, -0.00000509, 0.00000365, -0.00001293, 0.00001073, -0.00000872, 0.00000950, -0.00000525, 0.00000564, 0.00000109, 0.00000006, -0.00000751, 0.00000540, -0.00000645, 0.00000577, -0.00000649, 0.00000285, -0.00000064, 0.00000037, 0.00000204, 0.00000095, 0.00000064, 0.00000099, -0.00000318, 0.00000078, -0.00000260, -0.00000027, 0.00000101, -0.00000200, 0.00000154, -0.00000210, 0.00000086, 0.00000117, -0.00000026, 0.00000212, 0.00000075, -0.00000192, 0.00000068, -0.00000280, 0.00000219, -0.00000193, 0.00000050, -0.00000153, -0.00000083, 0.00000080, 0.00000159, -0.00000196, 0.00000453, -0.00000342, 0.00000219, -0.00000350, 0.00000229, -0.00000434, -0.00000077, -0.00000011, 0.00000654, -0.00000238, -0.00000020, },
		{ 0.00000101, -0.00000602, 0.00000834, -0.00000472, 0.00000473, -0.00000428, -0.00000438, 0.00000181, 0.00000351, -0.00000193, -0.00000162, -0.00001387, 0.00001225, -0.00000391, -0.00000171, -0.00000423, 0.00001345, -0.00001617, 0.00000461, -0.00000115, 0.00000091, -0.00000347, -0.00000344, -0.00002498, 0.00003415, -0.00007621, 0.00001135, -0.00013391, -0.00008080, 0.00028253, 0.00030635, -0.00014252, -0.00007423, 0.00001679, -0.00001628, -0.00000637, -0.00000826, -0.00002834, 0.00002140, -0.00003797, 0.00002283, -0.00002214, 0.00001587, -0.00002063, 0.00002713, -0.00001392, 0.00000483, -0.00002643, 0.00002090, -0.00001749, 0.00001913, -0.00001513, 0.00001248, -0.00000636, 0.00000961, -0.00001307, 0.00000338, -0.00000919, 0.00000492, -0.00000910, 0.00000473, 0.00000890, 0.00000040, -0.00000307, -0.00000667, 0.00000734, -0.00000314, 0.00000298, -0.00000281, 0.00000356, -0.00000423, 0.00000275, -0.00001223, 0.00001045, -0.00000861, 0.00000967, -0.00000548, 0.00000589, 0.00000102, -0.00000033, -0.00000712, 0.00000519, -0.00000637, 0.00000585, -0.00000683, 0.00000330, -0.00000102, 0.00000071, 0.00000197, 0.00000084, 0.00000082, 0.00000076, -0.00000310, 0.00000084, -0.00000275, -0.00000004, 0.00000083, -0.00000201, 0.00000161, -0.00000232, 0.00000115, 0.00000105, -0.00000018, 0.00000218, 0.00000057, -0.00000173, 0.00000057, -0.00000281, 0.00000223, -0.00000202, 0.00000058, -0.00000183, -0.00000069, 0.00000112, 0.00000128, -0.00000164, 0.00000430, -0.00000330, 0.00000226, -0.00000391, 0.00000257, -0.00000455, -0.00000057, 0.00000010, 0.00000623, -0.00000239, -0.00000004, -0.00000206, },
		{ -0.00000568, 0.00000857, -0.00000506, 0.00000482, -0.00000460, -0.00000401, 0.00000142, 0.00000366, -0.00000184, -0.00000226, -0.00001291, 0.00001203, -0.00000382, -0.00000196, -0.00000409, 0.00001344, -0.00001678, 0.00000558, -0.00000142, 0.00000020, -0.00000173, -0.00000577, -0.00002362, 0.00003356, -0.00007724, 0.00001443, -0.00014375, -0.00005886, 0.00027226, 0.00029004, -0.00012721, -0.00007882, 0.00002136, -0.00001981, -0.00000225, -0.00001356, -0.00002424, 0.00001678, -0.00003311, 0.00001960, -0.00002066, 0.00001456, -0.00001905, 0.00002687, -0.00001459, 0.00000383, -0.00002422, 0.00001989, -0.00001746, 0.00001946, -0.00001573, 0.00001353, -0.00000754, 0.00001015, -0.00001350, 0.00000398, -0.00001006, 0.00000594, -0.00001035, 0.00000678, 0.00000721, 0.00000119, -0.00000402, -0.00000557, 0.00000657, -0.00000255, 0.00000234, -0.00000194, 0.00000249, -0.00000333, 0.00000178, -0.00001140, 0.00001005, -0.00000840, 0.00000976, -0.00000566, 0.00000613, 0.00000091, -0.00000070, -0.00000670, 0.00000491, -0.00000622, 0.00000585, -0.00000710, 0.00000372, -0.00000141, 0.00000107, 0.00000187, 0.00000076, 0.00000099, 0.00000051, -0.00000299, 0.00000086, -0.00000289, 0.00000019, 0.00000064, -0.00000199, 0.00000167, -0.00000253, 0.00000146, 0.00000090, -0.00000007, 0.00000220, 0.00000040, -0.00000155, 0.00000045, -0.00000281, 0.00000226, -0.00000211, 0.00000068, -0.00000215, -0.00000051, 0.00000141, 0.00000097, -0.00000128, 0.00000403, -0.00000314, 0.00000229, -0.00000430, 0.00000283, -0.00000476, -0.00000035, 0.00000028, 0.00000590, -0.00000234, 0.00000010, -0.00000240, 0.00000109, },
		{ 0.00000876, -0.00000539, 0.00000490, -0.00000492, -0.00000360, 0.00000098, 0.00000385, -0.00000182, -0.00000281, -0.00001197, 0.00001176, -0.00000372, -0.00000218, -0.00000392, 0.00001330, -0.00001729, 0.00000656, -0.00000177, -0.00000043, -0.00000008, -0.00000813, -0.00002191, 0.00003231, -0.00007736, 0.00001636, -0.00015207, -0.00003781, 0.00026095, 0.00027554, -0.00011315, -0.00008245, 0.00002522, -0.00002270, 0.00000112, -0.00001832, -0.00002038, 0.00001226, -0.00002820, 0.00001618, -0.00001898, 0.00001308, -0.00001730, 0.00002641, -0.00001517, 0.00000281, -0.00002190, 0.00001866, -0.00001718, 0.00001956, -0.00001613, 0.00001440, -0.00000859, 0.00001058, -0.00001381, 0.00000445, -0.00001079, 0.00000682, -0.00001146, 0.00000876, 0.00000551, 0.00000197, -0.00000495, -0.00000444, 0.00000576, -0.00000195, 0.00000172, -0.00000107, 0.00000143, -0.00000241, 0.00000076, -0.00001048, 0.00000954, -0.00000808, 0.00000976, -0.00000578, 0.00000636, 0.00000078, -0.00000106, -0.00000625, 0.00000458, -0.00000601, 0.00000579, -0.00000730, 0.00000411, -0.00000179, 0.00000146, 0.00000173, 0.00000070, 0.00000115, 0.00000025, -0.00000285, 0.00000085, -0.00000300, 0.00000042, 0.00000043, -0.00000194, 0.00000170, -0.00000273, 0.00000178, 0.00000071, 0.00000009, 0.00000219, 0.00000026, -0.00000136, 0.00000030, -0.00000278, 0.00000228, -0.00000219, 0.00000078, -0.00000249, -0.00000027, 0.00000166, 0.00000066, -0.00000091, 0.00000370, -0.00000292, 0.00000225, -0.00000465, 0.00000309, -0.00000496, -0.00000010, 0.00000044, 0.00000554, -0.00000226, 0.00000021, -0.00000273, 0.00000118, -0.00000532, },
		{ -0.00000572, 0.00000496, -0.00000525, -0.00000316, 0.00000050, 0.00000409, -0.00000186, -0.00000326, -0.00001105, 0.00001145, -0.00000363, -0.00000237, -0.00000372, 0.00001304, -0.00001768, 0.00000756, -0.00000221, -0.00000096, 0.00000145, -0.00001047, -0.00001993, 0.00003045, -0.00007663, 0.00001716, -0.00015878, -0.00001783, 0.00024885, 0.00026268, -0.00010027, -0.00008516, 0.00002838, -0.00002497, 0.00000371, -0.00002248, -0.00001685, 0.00000792, -0.00002330, 0.00001263, -0.00001711, 0.00001147, -0.00001540, 0.00002578, -0.00001568, 0.00000181, -0.00001949, 0.00001722, -0.00001666, 0.00001943, -0.00001632, 0.00001508, -0.00000949, 0.00001088, -0.00001399, 0.00000479, -0.00001138, 0.00000755, -0.00001241, 0.00001064, 0.00000383, 0.00000275, -0.00000586, -0.00000331, 0.00000493, -0.00000134, 0.00000112, -0.00000024, 0.00000038, -0.00000149, -0.00000030, -0.00000946, 0.00000892, -0.00000767, 0.00000967, -0.00000584, 0.00000656, 0.00000063, -0.00000140, -0.00000578, 0.00000421, -0.00000573, 0.00000565, -0.00000744, 0.00000447, -0.00000216, 0.00000186, 0.00000156, 0.00000067, 0.00000130, -0.00000002, -0.00000267, 0.00000081, -0.00000308, 0.00000063, 0.00000020, -0.00000187, 0.00000170, -0.00000290, 0.00000210, 0.00000049, 0.00000029, 0.00000213, 0.00000015, -0.00000118, 0.00000015, -0.00000273, 0.00000228, -0.00000228, 0.00000089, -0.00000285, 0.00000001, 0.00000186, 0.00000037, -0.00000052, 0.00000334, -0.00000266, 0.00000217, -0.00000496, 0.00000333, -0.00000516, 0.00000016, 0.00000058, 0.00000517, -0.00000213, 0.00000028, -0.00000305, 0.00000126, -0.00000492, 0.00000890, },
		{ 0.00000500, -0.00000556, -0.00000270, -0.00000002, 0.00000437, -0.00000197, -0.00000362, -0.00001016, 0.00001112, -0.00000355, -0.00000251, -0.00000351, 0.00001266, -0.00001795, 0.00000857, -0.00000274, -0.00000137, 0.00000280, -0.00001271, -0.00001777, 0.00002809, -0.00007512, 0.00001685, -0.00016384, 0.00000087, 0.00023618, 0.00025133, -0.00008850, -0.00008699, 0.00003087, -0.00002663, 0.00000551, -0.00002600, -0.00001371, 0.00000384, -0.00001849, 0.00000900, -0.00001511, 0.00000975, -0.00001339, 0.00002500, -0.00001611, 0.00000083, -0.00001705, 0.00001561, -0.00001594, 0.00001909, -0.00001632, 0.00001558, -0.00001025, 0.00001107, -0.00001404, 0.00000501, -0.00001184, 0.00000814, -0.00001320, 0.00001240, 0.00000218, 0.00000350, -0.00000673, -0.00000218, 0.00000409, -0.00000073, 0.00000054, 0.00000056, -0.00000065, -0.00000057, -0.00000139, -0.00000836, 0.00000820, -0.00000715, 0.00000948, -0.00000583, 0.00000674, 0.00000046, -0.00000173, -0.00000529, 0.00000379, -0.00000540, 0.00000543, -0.00000750, 0.00000477, -0.00000252, 0.00000227, 0.00000136, 0.00000067, 0.00000142, -0.00000029, -0.00000248, 0.00000073, -0.00000313, 0.00000084, -0.00000003, -0.00000177, 0.00000168, -0.00000305, 0.00000243, 0.00000025, 0.00000053, 0.00000203, 0.00000006, -0.00000101, -0.00000002, -0.00000266, 0.00000226, -0.00000235, 0.00000100, -0.00000322, 0.00000035, 0.00000201, 0.00000011, -0.00000013, 0.00000295, -0.00000235, 0.00000203, -0.00000522, 0.00000354, -0.00000534, 0.00000043, 0.00000070, 0.00000478, -0.00000196, 0.00000032, -0.00000334, 0.00000133, -0.00000449, 0.00000900, -0.00000602, },
		{ -0.00000585, -0.00000222, -0.00000057, 0.00000467, -0.00000213, -0.00000389, -0.00000932, 0.00001076, -0.00000348, -0.00000260, -0.00000330, 0.00001217, -0.00001811, 0.00000957, -0.00000335, -0.00000165, 0.00000395, -0.00001480, -0.00001551, 0.00002530, -0.00007291, 0.00001551, -0.00016724, 0.00001816, 0.00022317, 0.00024134, -0.00007778, -0.00008799, 0.00003275, -0.00002774, 0.00000654, -0.00002885, -0.00001101, 0.00000006, -0.00001384, 0.00000536, -0.00001302, 0.00000796, -0.00001127, 0.00002406, -0.00001648, -0.00000009, -0.00001459, 0.00001386, -0.00001502, 0.00001856, -0.00001613, 0.00001589, -0.00001086, 0.00001113, -0.00001397, 0.00000509, -0.00001215, 0.00000856, -0.00001382, 0.00001402, 0.00000058, 0.00000422, -0.00000756, -0.00000106, 0.00000323, -0.00000013, -0.00000001, 0.00000132, -0.00000165, 0.00000033, -0.00000249, -0.00000719, 0.00000739, -0.00000654, 0.00000920, -0.00000575, 0.00000688, 0.00000027, -0.00000203, -0.00000479, 0.00000334, -0.00000501, 0.00000514, -0.00000748, 0.00000502, -0.00000285, 0.00000268, 0.00000114, 0.00000070, 0.00000153, -0.00000056, -0.00000225, 0.00000062, -0.00000315, 0.00000102, -0.00000027, -0.00000165, 0.00000164, -0.00000318, 0.00000275, -0.00000002, 0.00000080, 0.00000188, 0.00000001, -0.00000086, -0.00000019, -0.00000258, 0.00000223, -0.00000243, 0.00000110, -0.00000361, 0.00000074, 0.00000211, -0.00000013, 0.00000025, 0.00000254, -0.00000200, 0.00000183, -0.00000544, 0.00000373, -0.00000551, 0.00000072, 0.00000079, 0.00000439, -0.00000176, 0.00000033, -0.00000361, 0.00000139, -0.00000402, 0.00000905, -0.00000631, 0.00000502, },
		{ -0.00000174, -0.00000113, 0.00000500, -0.00000234, -0.00000407, -0.00000852, 0.00001039, -0.00000344, -0.00000263, -0.00000312, 0.00001160, -0.00001814, 0.00001056, -0.00000406, -0.00000179, 0.00000487, -0.00001667, -0.00001322, 0.00002218, -0.00007008, 0.00001320, -0.00016898, 0.00003390, 0.00021001, 0.00023255, -0.00006801, -0.00008823, 0.00003407, -0.00002836, 0.00000683, -0.00003101, -0.00000880, -0.00000334, -0.00000943, 0.00000177, -0.00001088, 0.00000613, -0.00000908, 0.00002300, -0.00001680, -0.00000095, -0.00001215, 0.00001201, -0.00001394, 0.00001785, -0.00001576, 0.00001604, -0.00001133, 0.00001108, -0.00001377, 0.00000505, -0.00001231, 0.00000883, -0.00001426, 0.00001549, -0.00000096, 0.00000490, -0.00000833, 0.00000003, 0.00000237, 0.00000046, -0.00000052, 0.00000202, -0.00000260, 0.00000120, -0.00000358, -0.00000597, 0.00000649, -0.00000584, 0.00000883, -0.00000559, 0.00000699, 0.00000006, -0.00000231, -0.00000428, 0.00000286, -0.00000457, 0.00000478, -0.00000738, 0.00000521, -0.00000315, 0.00000309, 0.00000089, 0.00000076, 0.00000161, -0.00000083, -0.00000201, 0.00000048, -0.00000314, 0.00000118, -0.00000051, -0.00000151, 0.00000156, -0.00000328, 0.00000306, -0.00000030, 0.00000111, 0.00000168, 0.00000001, -0.00000074, -0.00000036, -0.00000248, 0.00000219, -0.00000249, 0.00000121, -0.00000401, 0.00000117, 0.00000214, -0.00000032, 0.00000061, 0.00000211, -0.00000163, 0.00000159, -0.00000561, 0.00000389, -0.00000567, 0.00000101, 0.00000085, 0.00000400, -0.00000152, 0.00000029, -0.00000385, 0.00000144, -0.00000351, 0.00000903, -0.00000655, 0.00000501, -0.00000612, },
		{ -0.00000170, 0.00000534, -0.00000259, -0.00000416, -0.00000777, 0.00001001, -0.00000343, -0.00000258, -0.00000296, 0.00001095, -0.00001805, 0.00001152, -0.00000484, -0.00000179, 0.00000555, -0.00001829, -0.00001099, 0.00001883, -0.00006675, 0.00001002, -0.00016911, 0.00004800, 0.00019690, 0.00022482, -0.00005911, -0.00008776, 0.00003489, -0.00002855, 0.00000642, -0.00003249, -0.00000708, -0.00000632, -0.00000530, -0.00000171, -0.00000873, 0.00000429, -0.00000686, 0.00002183, -0.00001706, -0.00000172, -0.00000976, 0.00001009, -0.00001272, 0.00001698, -0.00001525, 0.00001601, -0.00001165, 0.00001091, -0.00001346, 0.00000489, -0.00001234, 0.00000895, -0.00001452, 0.00001680, -0.00000242, 0.00000552, -0.00000904, 0.00000110, 0.00000152, 0.00000104, -0.00000100, 0.00000267, -0.00000348, 0.00000202, -0.00000465, -0.00000471, 0.00000553, -0.00000505, 0.00000837, -0.00000536, 0.00000706, -0.00000016, -0.00000255, -0.00000377, 0.00000235, -0.00000408, 0.00000436, -0.00000720, 0.00000533, -0.00000341, 0.00000349, 0.00000062, 0.00000085, 0.00000166, -0.00000109, -0.00000176, 0.00000031, -0.00000309, 0.00000131, -0.00000075, -0.00000135, 0.00000145, -0.00000334, 0.00000336, -0.00000059, 0.00000145, 0.00000144, 0.00000004, -0.00000064, -0.00000052, -0.00000236, 0.00000213, -0.00000254, 0.00000131, -0.00000441, 0.00000166, 0.00000210, -0.00000045, 0.00000095, 0.00000168, -0.00000122, 0.00000130, -0.00000572, 0.00000402, -0.00000580, 0.00000132, 0.00000089, 0.00000361, -0.00000125, 0.00000022, -0.00000407, 0.00000148, -0.00000296, 0.00000895, -0.00000677, 0.00000496, -0.00000637, -0.00000126, },
		{ 0.00000568, -0.00000287, -0.00000417, -0.00000707, 0.00000962, -0.00000343, -0.00000247, -0.00000285, 0.00001023, -0.00001784, 0.00001244, -0.00000569, -0.00000166, 0.00000597, -0.00001961, -0.00000890, 0.00001534, -0.00006301, 0.00000608, -0.00016767, 0.00006038, 0.00018401, 0.00021800, -0.00005097, -0.00008666, 0.00003529, -0.00002840, 0.00000537, -0.00003329, -0.00000589, -0.00000885, -0.00000152, -0.00000502, -0.00000662, 0.00000247, -0.00000462, 0.00002056, -0.00001729, -0.00000240, -0.00000746, 0.00000813, -0.00001140, 0.00001598, -0.00001459, 0.00001583, -0.00001185, 0.00001065, -0.00001305, 0.00000463, -0.00001224, 0.00000892, -0.00001459, 0.00001794, -0.00000379, 0.00000608, -0.00000968, 0.00000211, 0.00000068, 0.00000160, -0.00000144, 0.00000326, -0.00000431, 0.00000278, -0.00000568, -0.00000344, 0.00000451, -0.00000419, 0.00000781, -0.00000505, 0.00000707, -0.00000038, -0.00000277, -0.00000327, 0.00000183, -0.00000357, 0.00000387, -0.00000695, 0.00000539, -0.00000363, 0.00000387, 0.00000034, 0.00000097, 0.00000169, -0.00000134, -0.00000149, 0.00000011, -0.00000300, 0.00000141, -0.00000098, -0.00000118, 0.00000132, -0.00000336, 0.00000363, -0.00000089, 0.00000182, 0.00000116, 0.00000012, -0.00000058, -0.00000067, -0.00000225, 0.00000205, -0.00000258, 0.00000140, -0.00000481, 0.00000219, 0.00000200, -0.00000054, 0.00000125, 0.00000126, -0.00000081, 0.00000096, -0.00000579, 0.00000412, -0.00000591, 0.00000163, 0.00000089, 0.00000323, -0.00000097, 0.00000011, -0.00000425, 0.00000151, -0.00000238, 0.00000880, -0.00000693, 0.00000487, -0.00000657, -0.00000079, -0.00000227, },
		{ -0.00000316, -0.00000411, -0.00000642, 0.00000923, -0.00000346, -0.00000229, -0.00000279, 0.00000945, -0.00001751, 0.00001331, -0.00000658, -0.00000139, 0.00000612, -0.00002059, -0.00000701, 0.00001181, -0.00005898, 0.00000149, -0.00016475, 0.00007099, 0.00017147, 0.00021196, -0.00004352, -0.00008500, 0.00003531, -0.00002797, 0.00000375, -0.00003346, -0.00000522, -0.00001090, 0.00000187, -0.00000813, -0.00000458, 0.00000071, -0.00000239, 0.00001922, -0.00001748, -0.00000297, -0.00000526, 0.00000617, -0.00000999, 0.00001487, -0.00001381, 0.00001552, -0.00001192, 0.00001028, -0.00001254, 0.00000425, -0.00001202, 0.00000874, -0.00001448, 0.00001889, -0.00000505, 0.00000656, -0.00001023, 0.00000307, -0.00000013, 0.00000214, -0.00000185, 0.00000377, -0.00000506, 0.00000347, -0.00000666, -0.00000217, 0.00000344, -0.00000327, 0.00000718, -0.00000466, 0.00000704, -0.00000061, -0.00000296, -0.00000279, 0.00000131, -0.00000302, 0.00000333, -0.00000661, 0.00000537, -0.00000379, 0.00000423, 0.00000005, 0.00000111, 0.00000168, -0.00000158, -0.00000122, -0.00000011, -0.00000287, 0.00000148, -0.00000120, -0.00000100, 0.00000117, -0.00000335, 0.00000388, -0.00000118, 0.00000221, 0.00000084, 0.00000025, -0.00000056, -0.00000081, -0.00000212, 0.00000196, -0.00000261, 0.00000148, -0.00000520, 0.00000276, 0.00000182, -0.00000055, 0.00000151, 0.00000086, -0.00000039, 0.00000059, -0.00000581, 0.00000418, -0.00000599, 0.00000194, 0.00000086, 0.00000287, -0.00000067, -0.00000002, -0.00000440, 0.00000152, -0.00000176, 0.00000858, -0.00000705, 0.00000474, -0.00000674, -0.00000034, -0.00000283, 0.00000602, },
		{ -0.00000399, -0.00000581, 0.00000882, -0.00000351, -0.00000203, -0.00000279, 0.00000863, -0.00001705, 0.00001411, -0.00000752, -0.00000101, 0.00000601, -0.00002122, -0.00000537, 0.00000832, -0.00005476, -0.00000363, -0.00016045, 0.00007983, 0.00015943, 0.00020658, -0.00003667, -0.00008285, 0.00003505, -0.00002736, 0.00000163, -0.00003302, -0.00000506, -0.00001246, 0.00000483, -0.00001097, -0.00000265, -0.00000097, -0.00000021, 0.00001782, -0.00001763, -0.00000342, -0.00000319, 0.00000424, -0.00000853, 0.00001367, -0.00001294, 0.00001508, -0.00001188, 0.00000983, -0.00001194, 0.00000379, -0.00001169, 0.00000844, -0.00001419, 0.00001966, -0.00000620, 0.00000697, -0.00001070, 0.00000397, -0.00000092, 0.00000265, -0.00000220, 0.00000422, -0.00000572, 0.00000407, -0.00000757, -0.00000090, 0.00000235, -0.00000229, 0.00000647, -0.00000419, 0.00000695, -0.00000084, -0.00000311, -0.00000232, 0.00000078, -0.00000245, 0.00000274, -0.00000620, 0.00000529, -0.00000390, 0.00000455, -0.00000024, 0.00000127, 0.00000164, -0.00000179, -0.00000094, -0.00000035, -0.00000271, 0.00000152, -0.00000141, -0.00000081, 0.00000098, -0.00000330, 0.00000409, -0.00000147, 0.00000261, 0.00000048, 0.00000041, -0.00000057, -0.00000092, -0.00000200, 0.00000186, -0.00000263, 0.00000155, -0.00000559, 0.00000336, 0.00000158, -0.00000050, 0.00000171, 0.00000048, 0.00000002, 0.00000020, -0.00000579, 0.00000421, -0.00000605, 0.00000226, 0.00000080, 0.00000253, -0.00000036, -0.00000019, -0.00000453, 0.00000153, -0.00000110, 0.00000829, -0.00000712, 0.00000455, -0.00000686, 0.00000008, -0.00000337, 0.00000635, -0.00000346, },
		{ -0.00000525, 0.00000841, -0.00000357, -0.00000171, -0.00000286, 0.00000777, -0.00001646, 0.00001483, -0.00000846, -0.00000053, 0.00000565, -0.00002147, -0.00000405, 0.00000496, -0.00005046, -0.00000913, -0.00015488, 0.00008691, 0.00014797, 0.00020174, -0.00003032, -0.00008027, 0.00003455, -0.00002666, -0.00000089, -0.00003204, -0.00000538, -0.00001352, 0.00000734, -0.00001352, -0.00000087, -0.00000254, 0.00000191, 0.00001637, -0.00001776, -0.00000375, -0.00000128, 0.00000236, -0.00000704, 0.00001241, -0.00001199, 0.00001454, -0.00001174, 0.00000931, -0.00001127, 0.00000324, -0.00001125, 0.00000801, -0.00001374, 0.00002025, -0.00000723, 0.00000728, -0.00001107, 0.00000480, -0.00000168, 0.00000314, -0.00000252, 0.00000459, -0.00000630, 0.00000459, -0.00000840, 0.00000032, 0.00000125, -0.00000127, 0.00000569, -0.00000365, 0.00000680, -0.00000106, -0.00000324, -0.00000188, 0.00000027, -0.00000187, 0.00000210, -0.00000572, 0.00000513, -0.00000395, 0.00000484, -0.00000052, 0.00000145, 0.00000157, -0.00000197, -0.00000067, -0.00000061, -0.00000252, 0.00000151, -0.00000159, -0.00000061, 0.00000078, -0.00000321, 0.00000427, -0.00000174, 0.00000301, 0.00000009, 0.00000063, -0.00000064, -0.00000101, -0.00000188, 0.00000175, -0.00000264, 0.00000160, -0.00000596, 0.00000400, 0.00000127, -0.00000038, 0.00000186, 0.00000014, 0.00000042, -0.00000022, -0.00000572, 0.00000421, -0.00000609, 0.00000257, 0.00000071, 0.00000222, -0.00000006, -0.00000038, -0.00000463, 0.00000153, -0.00000042, 0.00000793, -0.00000713, 0.00000432, -0.00000692, 0.00000046, -0.00000389, 0.00000665, -0.00000376, -0.00000382, },
		{ 0.00000799, -0.00000364, -0.00000132, -0.00000300, 0.00000690, -0.00001575, 0.00001545, -0.00000941, 0.00000003, 0.00000506, -0.00002136, -0.00000307, 0.00000180, -0.00004619, -0.00001490, -0.00014819, 0.00009225, 0.00013719, 0.00019734, -0.00002442, -0.00007732, 0.00003388, -0.00002595, -0.00000374, -0.00003058, -0.00000615, -0.00001410, 0.00000938, -0.00001574, 0.00000074, -0.00000397, 0.00000393, 0.00001490, -0.00001786, -0.00000396, 0.00000045, 0.00000057, -0.00000556, 0.00001111, -0.00001099, 0.00001390, -0.00001150, 0.00000873, -0.00001055, 0.00000263, -0.00001072, 0.00000747, -0.00001312, 0.00002064, -0.00000813, 0.00000751, -0.00001134, 0.00000556, -0.00000239, 0.00000361, -0.00000279, 0.00000490, -0.00000680, 0.00000502, -0.00000914, 0.00000151, 0.00000015, -0.00000023, 0.00000485, -0.00000304, 0.00000660, -0.00000127, -0.00000333, -0.00000146, -0.00000024, -0.00000129, 0.00000144, -0.00000518, 0.00000490, -0.00000394, 0.00000509, -0.00000079, 0.00000165, 0.00000147, -0.00000214, -0.00000040, -0.00000088, -0.00000230, 0.00000148, -0.00000175, -0.00000042, 0.00000056, -0.00000308, 0.00000442, -0.00000199, 0.00000342, -0.00000033, 0.00000088, -0.00000074, -0.00000107, -0.00000177, 0.00000164, -0.00000264, 0.00000163, -0.00000631, 0.00000465, 0.00000089, -0.00000019, 0.00000193, -0.00000015, 0.00000079, -0.00000065, -0.00000562, 0.00000418, -0.00000609, 0.00000288, 0.00000059, 0.00000195, 0.00000024, -0.00000058, -0.00000470, 0.00000152, 0.00000029, 0.00000750, -0.00000709, 0.00000404, -0.00000694, 0.00000081, -0.00000437, 0.00000692, -0.00000404, -0.00000362, -0.00000472, },
		{ -0.00000371, -0.00000088, -0.00000322, 0.00000601, -0.00001492, 0.00001596, -0.00001032, 0.00000064, 0.00000426, -0.00002087, -0.00000246, -0.00000109, -0.00004204, -0.00002078, -0.00014052, 0.00009592, 0.00012713, 0.00019330, -0.00001888, -0.00007406, 0.00003309, -0.00002531, -0.00000680, -0.00002871, -0.00000732, -0.00001422, 0.00001095, -0.00001761, 0.00000214, -0.00000523, 0.00000584, 0.00001342, -0.00001794, -0.00000405, 0.00000201, -0.00000112, -0.00000410, 0.00000979, -0.00000996, 0.00001320, -0.00001120, 0.00000809, -0.00000977, 0.00000197, -0.00001012, 0.00000683, -0.00001236, 0.00002085, -0.00000891, 0.00000765, -0.00001151, 0.00000623, -0.00000305, 0.00000404, -0.00000302, 0.00000513, -0.00000720, 0.00000534, -0.00000978, 0.00000262, -0.00000093, 0.00000083, 0.00000397, -0.00000237, 0.00000633, -0.00000148, -0.00000339, -0.00000107, -0.00000072, -0.00000072, 0.00000076, -0.00000458, 0.00000460, -0.00000386, 0.00000529, -0.00000105, 0.00000186, 0.00000134, -0.00000227, -0.00000015, -0.00000116, -0.00000205, 0.00000140, -0.00000189, -0.00000024, 0.00000032, -0.00000290, 0.00000451, -0.00000221, 0.00000383, -0.00000077, 0.00000117, -0.00000089, -0.00000109, -0.00000167, 0.00000152, -0.00000262, 0.00000164, -0.00000663, 0.00000532, 0.00000046, 0.00000008, 0.00000194, -0.00000039, 0.00000112, -0.00000107, -0.00000549, 0.00000412, -0.00000607, 0.00000318, 0.00000043, 0.00000171, 0.00000052, -0.00000079, -0.00000475, 0.00000152, 0.00000102, 0.00000701, -0.00000699, 0.00000371, -0.00000689, 0.00000110, -0.00000481, 0.00000715, -0.00000429, -0.00000339, -0.00000422, 0.00000755, },
		{ -0.00000040, -0.00000351, 0.00000512, -0.00001398, 0.00001635, -0.00001119, 0.00000129, 0.00000327, -0.00002003, -0.00000225, -0.00000366, -0.00003811, -0.00002665, -0.00013202, 0.00009800, 0.00011785, 0.00018954, -0.00001365, -0.00007053, 0.00003223, -0.00002483, -0.00000999, -0.00002651, -0.00000884, -0.00001390, 0.00001204, -0.00001911, 0.00000332, -0.00000632, 0.00000761, 0.00001195, -0.00001800, -0.00000401, 0.00000336, -0.00000267, -0.00000270, 0.00000849, -0.00000891, 0.00001245, -0.00001082, 0.00000742, -0.00000897, 0.00000127, -0.00000946, 0.00000612, -0.00001146, 0.00002088, -0.00000955, 0.00000769, -0.00001158, 0.00000681, -0.00000366, 0.00000443, -0.00000322, 0.00000529, -0.00000751, 0.00000556, -0.00001030, 0.00000367, -0.00000198, 0.00000189, 0.00000304, -0.00000163, 0.00000601, -0.00000167, -0.00000343, -0.00000072, -0.00000118, -0.00000016, 0.00000006, -0.00000393, 0.00000424, -0.00000371, 0.00000543, -0.00000129, 0.00000207, 0.00000119, -0.00000237, 0.00000009, -0.00000145, -0.00000177, 0.00000129, -0.00000200, -0.00000006, 0.00000008, -0.00000269, 0.00000457, -0.00000240, 0.00000421, -0.00000122, 0.00000149, -0.00000109, -0.00000108, -0.00000158, 0.00000140, -0.00000260, 0.00000163, -0.00000691, 0.00000599, -0.00000003, 0.00000042, 0.00000186, -0.00000057, 0.00000140, -0.00000149, -0.00000533, 0.00000403, -0.00000603, 0.00000348, 0.00000024, 0.00000152, 0.00000077, -0.00000101, -0.00000479, 0.00000153, 0.00000175, 0.00000645, -0.00000683, 0.00000332, -0.00000680, 0.00000135, -0.00000520, 0.00000733, -0.00000450, -0.00000314, -0.00000374, 0.00000709, -0.00000377, },
		{ -0.00000387, 0.00000425, -0.00001292, 0.00001660, -0.00001198, 0.00000193, 0.00000214, -0.00001886, -0.00000244, -0.00000586, -0.00003447, -0.00003239, -0.00012284, 0.00009857, 0.00010934, 0.00018602, -0.00000868, -0.00006678, 0.00003133, -0.00002455, -0.00001322, -0.00002405, -0.00001065, -0.00001319, 0.00001266, -0.00002023, 0.00000426, -0.00000721, 0.00000922, 0.00001049, -0.00001803, -0.00000385, 0.00000451, -0.00000407, -0.00000136, 0.00000720, -0.00000788, 0.00001167, -0.00001041, 0.00000673, -0.00000816, 0.00000054, -0.00000875, 0.00000534, -0.00001044, 0.00002075, -0.00001007, 0.00000765, -0.00001155, 0.00000730, -0.00000422, 0.00000480, -0.00000337, 0.00000538, -0.00000773, 0.00000568, -0.00001071, 0.00000462, -0.00000298, 0.00000294, 0.00000210, -0.00000085, 0.00000563, -0.00000184, -0.00000344, -0.00000040, -0.00000160, 0.00000038, -0.00000064, -0.00000325, 0.00000382, -0.00000351, 0.00000553, -0.00000150, 0.00000228, 0.00000100, -0.00000243, 0.00000030, -0.00000172, -0.00000148, 0.00000114, -0.00000208, 0.00000011, -0.00000018, -0.00000245, 0.00000457, -0.00000255, 0.00000458, -0.00000168, 0.00000184, -0.00000132, -0.00000104, -0.00000151, 0.00000127, -0.00000257, 0.00000160, -0.00000716, 0.00000666, -0.00000057, 0.00000082, 0.00000171, -0.00000067, 0.00000163, -0.00000189, -0.00000516, 0.00000392, -0.00000596, 0.00000377, 0.00000002, 0.00000137, 0.00000099, -0.00000122, -0.00000482, 0.00000155, 0.00000248, 0.00000584, -0.00000662, 0.00000289, -0.00000664, 0.00000154, -0.00000554, 0.00000747, -0.00000467, -0.00000290, -0.00000327, 0.00000660, -0.00000382, 0.00000012, },
		{ 0.00000339, -0.00001175, 0.00001671, -0.00001268, 0.00000254, 0.00000090, -0.00001740, -0.00000301, -0.00000767, -0.00003118, -0.00003788, -0.00011315, 0.00009776, 0.00010161, 0.00018269, -0.00000393, -0.00006282, 0.00003041, -0.00002456, -0.00001639, -0.00002143, -0.00001268, -0.00001213, 0.00001285, -0.00002096, 0.00000494, -0.00000788, 0.00001065, 0.00000907, -0.00001803, -0.00000359, 0.00000544, -0.00000531, -0.00000012, 0.00000597, -0.00000687, 0.00001087, -0.00000996, 0.00000603, -0.00000734, -0.00000019, -0.00000801, 0.00000452, -0.00000933, 0.00002045, -0.00001046, 0.00000751, -0.00001142, 0.00000770, -0.00000472, 0.00000513, -0.00000350, 0.00000541, -0.00000786, 0.00000569, -0.00001099, 0.00000547, -0.00000391, 0.00000396, 0.00000113, -0.00000002, 0.00000519, -0.00000198, -0.00000342, -0.00000012, -0.00000198, 0.00000089, -0.00000134, -0.00000253, 0.00000334, -0.00000324, 0.00000556, -0.00000168, 0.00000249, 0.00000080, -0.00000247, 0.00000050, -0.00000200, -0.00000117, 0.00000096, -0.00000212, 0.00000026, -0.00000044, -0.00000217, 0.00000453, -0.00000266, 0.00000492, -0.00000213, 0.00000221, -0.00000160, -0.00000095, -0.00000146, 0.00000116, -0.00000253, 0.00000154, -0.00000736, 0.00000732, -0.00000114, 0.00000130, 0.00000148, -0.00000070, 0.00000179, -0.00000226, -0.00000497, 0.00000380, -0.00000587, 0.00000405, -0.00000023, 0.00000128, 0.00000117, -0.00000141, -0.00000484, 0.00000159, 0.00000321, 0.00000518, -0.00000635, 0.00000241, -0.00000643, 0.00000168, -0.00000582, 0.00000756, -0.00000479, -0.00000266, -0.00000281, 0.00000608, -0.00000384, 0.00000067, -0.00000431, },
		{ -0.00001048, 0.00001666, -0.00001327, 0.00000309, -0.00000042, -0.00001568, -0.00000396, -0.00000908, -0.00002831, -0.00004301, -0.00010312, 0.00009568, 0.00009464, 0.00017952, 0.00000063, -0.00005869, 0.00002949, -0.00002488, -0.00001942, -0.00001872, -0.00001485, -0.00001077, 0.00001262, -0.00002132, 0.00000537, -0.00000834, 0.00001190, 0.00000769, -0.00001801, -0.00000322, 0.00000618, -0.00000637, 0.00000102, 0.00000480, -0.00000592, 0.00001008, -0.00000949, 0.00000534, -0.00000654, -0.00000092, -0.00000726, 0.00000367, -0.00000813, 0.00002000, -0.00001073, 0.00000729, -0.00001119, 0.00000801, -0.00000515, 0.00000543, -0.00000359, 0.00000539, -0.00000791, 0.00000560, -0.00001115, 0.00000622, -0.00000478, 0.00000493, 0.00000017, 0.00000085, 0.00000470, -0.00000211, -0.00000338, 0.00000013, -0.00000233, 0.00000137, -0.00000202, -0.00000180, 0.00000282, -0.00000291, 0.00000553, -0.00000182, 0.00000269, 0.00000057, -0.00000247, 0.00000066, -0.00000226, -0.00000085, 0.00000075, -0.00000213, 0.00000038, -0.00000070, -0.00000186, 0.00000444, -0.00000272, 0.00000522, -0.00000257, 0.00000260, -0.00000191, -0.00000083, -0.00000143, 0.00000105, -0.00000248, 0.00000146, -0.00000752, 0.00000795, -0.00000174, 0.00000183, 0.00000118, -0.00000065, 0.00000188, -0.00000260, -0.00000478, 0.00000366, -0.00000576, 0.00000432, -0.00000050, 0.00000124, 0.00000130, -0.00000159, -0.00000487, 0.00000166, 0.00000391, 0.00000448, -0.00000604, 0.00000189, -0.00000618, 0.00000177, -0.00000606, 0.00000759, -0.00000485, -0.00000243, -0.00000235, 0.00000552, -0.00000382, 0.00000122, -0.00000481, 0.00000255, },
		{ 0.00001645, -0.00001372, 0.00000355, -0.00000175, -0.00001375, -0.00000525, -0.00001006, -0.00002590, -0.00004768, -0.00009288, 0.00009246, 0.00008839, 0.00017649, 0.00000503, -0.00005439, 0.00002856, -0.00002556, -0.00002223, -0.00001600, -0.00001711, -0.00000918, 0.00001202, -0.00002132, 0.00000555, -0.00000857, 0.00001295, 0.00000637, -0.00001796, -0.00000277, 0.00000670, -0.00000725, 0.00000204, 0.00000370, -0.00000503, 0.00000931, -0.00000901, 0.00000466, -0.00000576, -0.00000163, -0.00000652, 0.00000281, -0.00000687, 0.00001942, -0.00001088, 0.00000699, -0.00001087, 0.00000822, -0.00000552, 0.00000570, -0.00000366, 0.00000531, -0.00000789, 0.00000542, -0.00001118, 0.00000684, -0.00000555, 0.00000586, -0.00000078, 0.00000174, 0.00000415, -0.00000220, -0.00000332, 0.00000033, -0.00000263, 0.00000180, -0.00000267, -0.00000105, 0.00000226, -0.00000252, 0.00000545, -0.00000193, 0.00000287, 0.00000033, -0.00000244, 0.00000080, -0.00000250, -0.00000052, 0.00000051, -0.00000212, 0.00000049, -0.00000095, -0.00000153, 0.00000430, -0.00000273, 0.00000547, -0.00000300, 0.00000299, -0.00000225, -0.00000067, -0.00000143, 0.00000094, -0.00000244, 0.00000135, -0.00000762, 0.00000855, -0.00000237, 0.00000242, 0.00000080, -0.00000053, 0.00000188, -0.00000288, -0.00000459, 0.00000352, -0.00000564, 0.00000458, -0.00000081, 0.00000126, 0.00000138, -0.00000174, -0.00000490, 0.00000177, 0.00000458, 0.00000375, -0.00000569, 0.00000133, -0.00000587, 0.00000181, -0.00000623, 0.00000756, -0.00000486, -0.00000223, -0.00000189, 0.00000493, -0.00000376, 0.00000178, -0.00000538, 0.00000176, -0.00000912, },
		{ -0.00001403, 0.00000390, -0.00000307, -0.00001167, -0.00000685, -0.00001065, -0.00002396, -0.00005181, -0.00008260, 0.00008824, 0.00008282, 0.00017361, 0.00000928, -0.00004993, 0.00002762, -0.00002662, -0.00002477, -0.00001336, -0.00001937, -0.00000742, 0.00001108, -0.00002097, 0.00000547, -0.00000858, 0.00001379, 0.00000512, -0.00001787, -0.00000224, 0.00000703, -0.00000794, 0.00000293, 0.00000269, -0.00000422, 0.00000858, -0.00000854, 0.00000401, -0.00000502, -0.00000230, -0.00000578, 0.00000196, -0.00000557, 0.00001872, -0.00001093, 0.00000662, -0.00001046, 0.00000834, -0.00000583, 0.00000593, -0.00000370, 0.00000519, -0.00000779, 0.00000514, -0.00001109, 0.00000735, -0.00000623, 0.00000671, -0.00000171, 0.00000264, 0.00000357, -0.00000227, -0.00000325, 0.00000051, -0.00000288, 0.00000218, -0.00000328, -0.00000031, 0.00000167, -0.00000208, 0.00000530, -0.00000199, 0.00000304, 0.00000008, -0.00000237, 0.00000090, -0.00000272, -0.00000020, 0.00000025, -0.00000206, 0.00000057, -0.00000120, -0.00000117, 0.00000412, -0.00000269, 0.00000568, -0.00000339, 0.00000338, -0.00000261, -0.00000047, -0.00000145, 0.00000085, -0.00000239, 0.00000122, -0.00000766, 0.00000911, -0.00000300, 0.00000305, 0.00000035, -0.00000032, 0.00000181, -0.00000311, -0.00000442, 0.00000338, -0.00000550, 0.00000483, -0.00000114, 0.00000133, 0.00000139, -0.00000185, -0.00000495, 0.00000192, 0.00000521, 0.00000300, -0.00000531, 0.00000075, -0.00000553, 0.00000180, -0.00000636, 0.00000749, -0.00000480, -0.00000206, -0.00000141, 0.00000429, -0.00000364, 0.00000232, -0.00000600, 0.00000100, -0.00000767, 0.00001607, },
		{ 0.00000410, -0.00000432, -0.00000948, -0.00000871, -0.00001084, -0.00002251, -0.00005535, -0.00007241, 0.00008317, 0.00007786, 0.00017086, 0.00001340, -0.00004531, 0.00002664, -0.00002805, -0.00002697, -0.00001086, -0.00002157, -0.00000555, 0.00000985, -0.00002029, 0.00000514, -0.00000836, 0.00001442, 0.00000394, -0.00001775, -0.00000165, 0.00000717, -0.00000844, 0.00000369, 0.00000178, -0.00000349, 0.00000790, -0.00000809, 0.00000339, -0.00000433, -0.00000292, -0.00000508, 0.00000114, -0.00000425, 0.00001791, -0.00001087, 0.00000618, -0.00000997, 0.00000838, -0.00000607, 0.00000613, -0.00000373, 0.00000503, -0.00000763, 0.00000478, -0.00001088, 0.00000773, -0.00000680, 0.00000749, -0.00000260, 0.00000354, 0.00000294, -0.00000231, -0.00000316, 0.00000064, -0.00000309, 0.00000251, -0.00000386, 0.00000043, 0.00000106, -0.00000160, 0.00000510, -0.00000201, 0.00000318, -0.00000019, -0.00000227, 0.00000097, -0.00000291, 0.00000012, -0.00000003, -0.00000198, 0.00000063, -0.00000143, -0.00000080, 0.00000389, -0.00000259, 0.00000584, -0.00000375, 0.00000376, -0.00000300, -0.00000025, -0.00000149, 0.00000077, -0.00000234, 0.00000107, -0.00000764, 0.00000962, -0.00000363, 0.00000371, -0.00000016, -0.00000003, 0.00000164, -0.00000328, -0.00000426, 0.00000324, -0.00000535, 0.00000506, -0.00000148, 0.00000146, 0.00000134, -0.00000193, -0.00000502, 0.00000212, 0.00000578, 0.00000225, -0.00000490, 0.00000014, -0.00000516, 0.00000174, -0.00000643, 0.00000736, -0.00000468, -0.00000193, -0.00000092, 0.00000360, -0.00000347, 0.00000283, -0.00000666, 0.00000030, -0.00000616, 0.00001554, -0.00001417, },
		{ -0.00000547, -0.00000724, -0.00001079, -0.00001069, -0.00002155, -0.00005823, -0.00006244, 0.00007737, 0.00007346, 0.00016828, 0.00001737, -0.00004051, 0.00002558, -0.00002986, -0.00002880, -0.00000856, -0.00002363, -0.00000365, 0.00000840, -0.00001933, 0.00000459, -0.00000793, 0.00001483, 0.00000284, -0.00001758, -0.00000101, 0.00000714, -0.00000876, 0.00000431, 0.00000097, -0.00000287, 0.00000728, -0.00000767, 0.00000283, -0.00000371, -0.00000349, -0.00000442, 0.00000036, -0.00000293, 0.00001702, -0.00001074, 0.00000568, -0.00000941, 0.00000832, -0.00000624, 0.00000630, -0.00000375, 0.00000484, -0.00000742, 0.00000435, -0.00001056, 0.00000799, -0.00000727, 0.00000818, -0.00000344, 0.00000444, 0.00000227, -0.00000232, -0.00000307, 0.00000075, -0.00000325, 0.00000278, -0.00000438, 0.00000114, 0.00000043, -0.00000108, 0.00000485, -0.00000198, 0.00000330, -0.00000045, -0.00000215, 0.00000101, -0.00000307, 0.00000043, -0.00000033, -0.00000187, 0.00000065, -0.00000164, -0.00000041, 0.00000362, -0.00000244, 0.00000593, -0.00000407, 0.00000412, -0.00000339, 0.00000001, -0.00000156, 0.00000070, -0.00000229, 0.00000090, -0.00000756, 0.00001008, -0.00000425, 0.00000440, -0.00000073, 0.00000032, 0.00000139, -0.00000338, -0.00000413, 0.00000311, -0.00000519, 0.00000528, -0.00000185, 0.00000165, 0.00000123, -0.00000195, -0.00000511, 0.00000237, 0.00000629, 0.00000151, -0.00000448, -0.00000047, -0.00000476, 0.00000165, -0.00000646, 0.00000720, -0.00000451, -0.00000183, -0.00000042, 0.00000287, -0.00000323, 0.00000330, -0.00000737, -0.00000035, -0.00000459, 0.00001484, -0.00001415, 0.00000414, },
		{ -0.00000502, -0.00001301, -0.00001022, -0.00002105, -0.00006044, -0.00005281, 0.00007100, 0.00006953, 0.00016587, 0.00002121, -0.00003553, 0.00002441, -0.00003201, -0.00003023, -0.00000652, -0.00002551, -0.00000176, 0.00000676, -0.00001812, 0.00000384, -0.00000729, 0.00001502, 0.00000183, -0.00001738, -0.00000034, 0.00000695, -0.00000890, 0.00000478, 0.00000027, -0.00000234, 0.00000673, -0.00000728, 0.00000231, -0.00000315, -0.00000399, -0.00000381, -0.00000036, -0.00000163, 0.00001606, -0.00001052, 0.00000514, -0.00000879, 0.00000819, -0.00000635, 0.00000644, -0.00000376, 0.00000463, -0.00000717, 0.00000385, -0.00001013, 0.00000812, -0.00000762, 0.00000877, -0.00000422, 0.00000532, 0.00000158, -0.00000229, -0.00000296, 0.00000082, -0.00000336, 0.00000299, -0.00000485, 0.00000182, -0.00000019, -0.00000053, 0.00000454, -0.00000190, 0.00000339, -0.00000071, -0.00000200, 0.00000101, -0.00000320, 0.00000073, -0.00000064, -0.00000173, 0.00000065, -0.00000183, -0.00000002, 0.00000331, -0.00000224, 0.00000597, -0.00000434, 0.00000445, -0.00000379, 0.00000029, -0.00000166, 0.00000065, -0.00000225, 0.00000071, -0.00000741, 0.00001046, -0.00000484, 0.00000510, -0.00000134, 0.00000075, 0.00000105, -0.00000341, -0.00000403, 0.00000299, -0.00000504, 0.00000549, -0.00000223, 0.00000189, 0.00000104, -0.00000192, -0.00000523, 0.00000269, 0.00000671, 0.00000078, -0.00000405, -0.00000109, -0.00000434, 0.00000153, -0.00000645, 0.00000699, -0.00000429, -0.00000177, 0.00000010, 0.00000209, -0.00000292, 0.00000372, -0.00000810, -0.00000093, -0.00000297, 0.00001398, -0.00001394, 0.00000400, -0.00000648, },
		{ -0.00001532, -0.00000948, -0.00002100, -0.00006194, -0.00004360, 0.00006417, 0.00006601, 0.00016365, 0.00002490, -0.00003034, 0.00002308, -0.00003448, -0.00003123, -0.00000477, -0.00002713, 0.00000004, 0.00000500, -0.00001669, 0.00000289, -0.00000645, 0.00001500, 0.00000091, -0.00001712, 0.00000033, 0.00000661, -0.00000887, 0.00000512, -0.00000033, -0.00000193, 0.00000625, -0.00000694, 0.00000186, -0.00000267, -0.00000441, -0.00000327, -0.00000102, -0.00000036, 0.00001505, -0.00001024, 0.00000457, -0.00000812, 0.00000799, -0.00000639, 0.00000654, -0.00000377, 0.00000441, -0.00000688, 0.00000330, -0.00000960, 0.00000814, -0.00000786, 0.00000926, -0.00000493, 0.00000617, 0.00000087, -0.00000224, -0.00000286, 0.00000087, -0.00000343, 0.00000314, -0.00000526, 0.00000247, -0.00000081, 0.00000004, 0.00000419, -0.00000177, 0.00000344, -0.00000097, -0.00000183, 0.00000097, -0.00000330, 0.00000100, -0.00000096, -0.00000156, 0.00000061, -0.00000199, 0.00000037, 0.00000298, -0.00000198, 0.00000594, -0.00000455, 0.00000475, -0.00000419, 0.00000058, -0.00000177, 0.00000061, -0.00000221, 0.00000050, -0.00000720, 0.00001078, -0.00000539, 0.00000580, -0.00000199, 0.00000124, 0.00000062, -0.00000336, -0.00000397, 0.00000290, -0.00000488, 0.00000569, -0.00000262, 0.00000218, 0.00000078, -0.00000183, -0.00000538, 0.00000307, 0.00000705, 0.00000008, -0.00000363, -0.00000170, -0.00000392, 0.00000139, -0.00000641, 0.00000676, -0.00000402, -0.00000174, 0.00000064, 0.00000127, -0.00000253, 0.00000408, -0.00000886, -0.00000144, -0.00000132, 0.00001298, -0.00001356, 0.00000367, -0.00000730, -0.00000287, },
		{ -0.00000854, -0.00002133, -0.00006276, -0.00003491, 0.00005701, 0.00006281, 0.00016167, 0.00002845, -0.00002493, 0.00002152, -0.00003721, -0.00003182, -0.00000336, -0.00002846, 0.00000170, 0.00000318, -0.00001509, 0.00000179, -0.00000545, 0.00001477, 0.00000009, -0.00001682, 0.00000101, 0.00000615, -0.00000868, 0.00000533, -0.00000082, -0.00000162, 0.00000586, -0.00000664, 0.00000147, -0.00000227, -0.00000476, -0.00000280, -0.00000159, 0.00000086, 0.00001400, -0.00000991, 0.00000397, -0.00000741, 0.00000772, -0.00000638, 0.00000663, -0.00000378, 0.00000418, -0.00000657, 0.00000271, -0.00000899, 0.00000804, -0.00000797, 0.00000965, -0.00000557, 0.00000698, 0.00000015, -0.00000216, -0.00000275, 0.00000089, -0.00000346, 0.00000324, -0.00000560, 0.00000306, -0.00000141, 0.00000063, 0.00000379, -0.00000160, 0.00000345, -0.00000121, -0.00000164, 0.00000091, -0.00000335, 0.00000125, -0.00000127, -0.00000138, 0.00000055, -0.00000213, 0.00000076, 0.00000261, -0.00000168, 0.00000584, -0.00000471, 0.00000501, -0.00000457, 0.00000089, -0.00000191, 0.00000059, -0.00000218, 0.00000028, -0.00000693, 0.00001101, -0.00000590, 0.00000649, -0.00000266, 0.00000179, 0.00000011, -0.00000324, -0.00000394, 0.00000282, -0.00000472, 0.00000586, -0.00000302, 0.00000252, 0.00000044, -0.00000169, -0.00000557, 0.00000353, 0.00000729, -0.00000057, -0.00000323, -0.00000229, -0.00000350, 0.00000123, -0.00000635, 0.00000651, -0.00000371, -0.00000175, 0.00000118, 0.00000041, -0.00000208, 0.00000436, -0.00000962, -0.00000188, 0.00000034, 0.00001184, -0.00001299, 0.00000312, -0.00000792, -0.00000084, -0.00001766, },
		{ -0.00002201, -0.00006290, -0.00002680, 0.00004964, 0.00005987, 0.00015993, 0.00003185, -0.00001929, 0.00001967, -0.00004014, -0.00003200, -0.00000231, -0.00002945, 0.00000316, 0.00000134, -0.00001335, 0.00000055, -0.00000428, 0.00001433, -0.00000063, -0.00001647, 0.00000166, 0.00000559, -0.00000834, 0.00000541, -0.00000121, -0.00000142, 0.00000556, -0.00000640, 0.00000115, -0.00000196, -0.00000501, -0.00000240, -0.00000208, 0.00000200, 0.00001293, -0.00000954, 0.00000335, -0.00000667, 0.00000740, -0.00000631, 0.00000668, -0.00000379, 0.00000395, -0.00000624, 0.00000209, -0.00000831, 0.00000783, -0.00000798, 0.00000992, -0.00000612, 0.00000775, -0.00000058, -0.00000204, -0.00000265, 0.00000089, -0.00000344, 0.00000327, -0.00000587, 0.00000360, -0.00000199, 0.00000123, 0.00000336, -0.00000139, 0.00000343, -0.00000145, -0.00000144, 0.00000081, -0.00000337, 0.00000146, -0.00000158, -0.00000118, 0.00000046, -0.00000223, 0.00000114, 0.00000223, -0.00000133, 0.00000568, -0.00000479, 0.00000522, -0.00000493, 0.00000121, -0.00000206, 0.00000059, -0.00000216, 0.00000005, -0.00000659, 0.00001116, -0.00000635, 0.00000715, -0.00000334, 0.00000238, -0.00000048, -0.00000304, -0.00000396, 0.00000278, -0.00000457, 0.00000603, -0.00000341, 0.00000290, 0.00000004, -0.00000148, -0.00000580, 0.00000406, 0.00000742, -0.00000116, -0.00000285, -0.00000286, -0.00000310, 0.00000107, -0.00000627, 0.00000626, -0.00000338, -0.00000178, 0.00000174, -0.00000048, -0.00000155, 0.00000455, -0.00001039, -0.00000224, 0.00000199, 0.00001057, -0.00001226, 0.00000237, -0.00000830, 0.00000101, -0.00001996, -0.00000746, },
		{ -0.00006241, -0.00001932, 0.00004216, 0.00005711, 0.00015846, 0.00003509, -0.00001339, 0.00001748, -0.00004321, -0.00003180, -0.00000161, -0.00003009, 0.00000438, -0.00000045, -0.00001153, -0.00000078, -0.00000299, 0.00001371, -0.00000125, -0.00001606, 0.00000228, 0.00000494, -0.00000787, 0.00000538, -0.00000151, -0.00000131, 0.00000533, -0.00000620, 0.00000090, -0.00000173, -0.00000517, -0.00000209, -0.00000246, 0.00000306, 0.00001186, -0.00000914, 0.00000273, -0.00000590, 0.00000702, -0.00000618, 0.00000671, -0.00000381, 0.00000372, -0.00000591, 0.00000146, -0.00000756, 0.00000752, -0.00000787, 0.00001008, -0.00000658, 0.00000845, -0.00000131, -0.00000191, -0.00000255, 0.00000087, -0.00000339, 0.00000325, -0.00000606, 0.00000408, -0.00000253, 0.00000182, 0.00000290, -0.00000113, 0.00000336, -0.00000166, -0.00000122, 0.00000068, -0.00000335, 0.00000165, -0.00000188, -0.00000096, 0.00000034, -0.00000230, 0.00000150, 0.00000183, -0.00000094, 0.00000546, -0.00000481, 0.00000537, -0.00000526, 0.00000153, -0.00000222, 0.00000059, -0.00000216, -0.00000019, -0.00000619, 0.00001122, -0.00000673, 0.00000777, -0.00000403, 0.00000300, -0.00000114, -0.00000276, -0.00000402, 0.00000276, -0.00000443, 0.00000618, -0.00000381, 0.00000332, -0.00000043, -0.00000121, -0.00000606, 0.00000467, 0.00000743, -0.00000169, -0.00000250, -0.00000338, -0.00000272, 0.00000091, -0.00000618, 0.00000600, -0.00000304, -0.00000182, 0.00000229, -0.00000139, -0.00000096, 0.00000466, -0.00001115, -0.00000251, 0.00000363, 0.00000920, -0.00001136, 0.00000142, -0.00000843, 0.00000263, -0.00002216, -0.00000630, -0.00002296, },
		{ -0.00001250, 0.00003467, 0.00005446, 0.00015730, 0.00003818, -0.00000724, 0.00001487, -0.00004633, -0.00003126, -0.00000127, -0.00003034, 0.00000532, -0.00000215, -0.00000966, -0.00000219, -0.00000158, 0.00001291, -0.00000176, -0.00001561, 0.00000285, 0.00000423, -0.00000728, 0.00000523, -0.00000173, -0.00000131, 0.00000519, -0.00000606, 0.00000071, -0.00000159, -0.00000524, -0.00000186, -0.00000274, 0.00000402, 0.00001080, -0.00000872, 0.00000212, -0.00000514, 0.00000660, -0.00000601, 0.00000671, -0.00000383, 0.00000351, -0.00000558, 0.00000082, -0.00000677, 0.00000713, -0.00000766, 0.00001013, -0.00000694, 0.00000909, -0.00000203, -0.00000174, -0.00000246, 0.00000084, -0.00000330, 0.00000317, -0.00000619, 0.00000450, -0.00000302, 0.00000240, 0.00000242, -0.00000084, 0.00000326, -0.00000185, -0.00000100, 0.00000052, -0.00000328, 0.00000179, -0.00000216, -0.00000074, 0.00000020, -0.00000234, 0.00000184, 0.00000142, -0.00000051, 0.00000517, -0.00000476, 0.00000546, -0.00000556, 0.00000184, -0.00000239, 0.00000061, -0.00000216, -0.00000043, -0.00000574, 0.00001119, -0.00000704, 0.00000835, -0.00000470, 0.00000364, -0.00000185, -0.00000241, -0.00000413, 0.00000277, -0.00000430, 0.00000631, -0.00000419, 0.00000376, -0.00000097, -0.00000089, -0.00000636, 0.00000535, 0.00000732, -0.00000213, -0.00000221, -0.00000386, -0.00000238, 0.00000077, -0.00000609, 0.00000576, -0.00000269, -0.00000188, 0.00000283, -0.00000231, -0.00000030, 0.00000468, -0.00001189, -0.00000270, 0.00000522, 0.00000774, -0.00001032, 0.00000027, -0.00000829, 0.00000399, -0.00002420, -0.00000512, -0.00002413, -0.00006132, },
		{ 0.00002723, 0.00005188, 0.00015645, 0.00004113, -0.00000082, 0.00001178, -0.00004942, -0.00003043, -0.00000127, -0.00003020, 0.00000595, -0.00000371, -0.00000779, -0.00000363, -0.00000009, 0.00001195, -0.00000217, -0.00001510, 0.00000336, 0.00000348, -0.00000660, 0.00000500, -0.00000187, -0.00000139, 0.00000513, -0.00000597, 0.00000059, -0.00000154, -0.00000522, -0.00000172, -0.00000290, 0.00000488, 0.00000977, -0.00000829, 0.00000153, -0.00000437, 0.00000615, -0.00000579, 0.00000670, -0.00000387, 0.00000332, -0.00000527, 0.00000019, -0.00000595, 0.00000666, -0.00000735, 0.00001007, -0.00000720, 0.00000965, -0.00000273, -0.00000156, -0.00000238, 0.00000081, -0.00000319, 0.00000305, -0.00000624, 0.00000484, -0.00000347, 0.00000295, 0.00000192, -0.00000052, 0.00000311, -0.00000201, -0.00000078, 0.00000035, -0.00000318, 0.00000190, -0.00000242, -0.00000051, 0.00000003, -0.00000234, 0.00000216, 0.00000101, -0.00000006, 0.00000483, -0.00000464, 0.00000548, -0.00000581, 0.00000213, -0.00000256, 0.00000064, -0.00000218, -0.00000067, -0.00000522, 0.00001106, -0.00000726, 0.00000886, -0.00000533, 0.00000428, -0.00000262, -0.00000199, -0.00000428, 0.00000282, -0.00000419, 0.00000642, -0.00000456, 0.00000423, -0.00000156, -0.00000050, -0.00000670, 0.00000611, 0.00000709, -0.00000249, -0.00000197, -0.00000428, -0.00000209, 0.00000065, -0.00000601, 0.00000554, -0.00000235, -0.00000193, 0.00000335, -0.00000323, 0.00000040, 0.00000460, -0.00001260, -0.00000279, 0.00000676, 0.00000621, -0.00000915, -0.00000107, -0.00000787, 0.00000505, -0.00002601, -0.00000400, -0.00002543, -0.00005969, -0.00000635, },
		{ 0.00004929, 0.00015594, 0.00004395, 0.00000586, 0.00000816, -0.00005239, -0.00002936, -0.00000159, -0.00002968, 0.00000624, -0.00000509, -0.00000596, -0.00000508, 0.00000146, 0.00001083, -0.00000248, -0.00001454, 0.00000380, 0.00000270, -0.00000584, 0.00000468, -0.00000195, -0.00000155, 0.00000515, -0.00000593, 0.00000053, -0.00000157, -0.00000511, -0.00000166, -0.00000296, 0.00000562, 0.00000877, -0.00000787, 0.00000096, -0.00000362, 0.00000568, -0.00000553, 0.00000666, -0.00000393, 0.00000315, -0.00000497, -0.00000042, -0.00000511, 0.00000613, -0.00000695, 0.00000991, -0.00000736, 0.00001014, -0.00000341, -0.00000136, -0.00000231, 0.00000076, -0.00000305, 0.00000287, -0.00000623, 0.00000511, -0.00000386, 0.00000348, 0.00000142, -0.00000016, 0.00000293, -0.00000214, -0.00000056, 0.00000015, -0.00000304, 0.00000196, -0.00000266, -0.00000028, -0.00000015, -0.00000231, 0.00000245, 0.00000061, 0.00000042, 0.00000443, -0.00000445, 0.00000543, -0.00000601, 0.00000241, -0.00000274, 0.00000068, -0.00000221, -0.00000091, -0.00000466, 0.00001085, -0.00000739, 0.00000929, -0.00000593, 0.00000492, -0.00000343, -0.00000152, -0.00000448, 0.00000290, -0.00000409, 0.00000652, -0.00000491, 0.00000471, -0.00000221, -0.00000006, -0.00000706, 0.00000693, 0.00000673, -0.00000275, -0.00000179, -0.00000463, -0.00000185, 0.00000057, -0.00000595, 0.00000536, -0.00000204, -0.00000196, 0.00000384, -0.00000413, 0.00000115, 0.00000442, -0.00001329, -0.00000278, 0.00000822, 0.00000464, -0.00000786, -0.00000257, -0.00000720, 0.00000579, -0.00002756, -0.00000300, -0.00002678, -0.00005761, -0.00000087, 0.00001992, },
		},
		{ { -0.00010590, -0.00000503, 0.00007543, 0.00003427, 0.00008027, 0.00001055, -0.00007323, -0.00000817, -0.00001432, -0.00001406, 0.00000443, -0.00000221, -0.00001921, -0.00000127, -0.00000467, 0.00000148, -0.00000192, -0.00000597, 0.00001687, -0.00000596, 0.00000536, -0.00000298, -0.00000609, 0.00000044, -0.00000026, -0.00000034, 0.00000271, 0.00000337, -0.00000307, 0.00000301, -0.00001066, 0.00000272, 0.00000088, -0.00000139, 0.00000042, -0.00000308, 0.00000158, -0.00000166, 0.00000104, 0.00000079, -0.00000002, -0.00000100, 0.00000813, -0.00000073, -0.00000392, -0.00000158, -0.00000324, -0.00000090, 0.00000013, 0.00000137, -0.00000033, 0.00000237, 0.00000321, 0.00000089, -0.00000564, 0.00000058, -0.00000008, -0.00000090, -0.00000071, -0.00000098, -0.00000096, -0.00000009, 0.00000409, -0.00000083, -0.00000113, -0.00000083, 0.00000409, -0.00000009, -0.00000096, -0.00000098, -0.00000071, -0.00000090, -0.00000008, 0.00000058, -0.00000564, 0.00000089, 0.00000321, 0.00000237, -0.00000033, 0.00000137, 0.00000013, -0.00000090, -0.00000324, -0.00000158, -0.00000392, -0.00000073, 0.00000813, -0.00000100, -0.00000002, 0.00000079, 0.00000104, -0.00000166, 0.00000158, -0.00000308, 0.00000042, -0.00000139, 0.00000088, 0.00000272, -0.00001066, 0.00000301, -0.00000307, 0.00000337, 0.00000271, -0.00000034, -0.00000026, 0.00000044, -0.00000609, -0.00000298, 0.00000536, -0.00000596, 0.00001687, -0.00000597, -0.00000192, 0.00000148, -0.00000467, -0.00000127, -0.00001921, -0.00000221, 0.00000443, -0.00001406, -0.00001432, -0.00000817, -0.00007323, 0.00001055, 0.00008027, 0.00003427, 0.00007543, -0.00000503, },
		{ -0.00000733, 0.00007132, 0.00003255, 0.00007493, 0.00001995, -0.00007198, -0.00001146, -0.00001224, -0.00001530, 0.00000416, -0.00000147, -0.00001827, -0.00000335, -0.00000340, 0.00000064, -0.00000082, -0.00000737, 0.00001697, -0.00000502, 0.00000418, -0.00000151, -0.00000713, 0.00000102, -0.00000088, 0.00000008, 0.00000231, 0.00000391, -0.00000339, 0.00000324, -0.00001022, 0.00000148, 0.00000192, -0.00000194, 0.00000106, -0.00000351, 0.00000153, -0.00000170, 0.00000093, 0.00000093, 0.00000016, -0.00000130, 0.00000768, 0.00000029, -0.00000436, -0.00000148, -0.00000323, -0.00000089, -0.00000015, 0.00000156, -0.00000039, 0.00000211, 0.00000330, 0.00000132, -0.00000552, 0.00000010, 0.00000016, -0.00000105, -0.00000051, -0.00000118, -0.00000072, -0.00000050, 0.00000408, -0.00000051, -0.00000113, -0.00000112, 0.00000401, 0.00000039, -0.00000122, -0.00000078, -0.00000091, -0.00000076, -0.00000032, 0.00000103, -0.00000567, 0.00000042, 0.00000312, 0.00000258, -0.00000020, 0.00000112, 0.00000043, -0.00000090, -0.00000321, -0.00000178, -0.00000341, -0.00000171, 0.00000842, -0.00000059, -0.00000026, 0.00000070, 0.00000107, -0.00000155, 0.00000151, -0.00000251, -0.00000030, -0.00000081, -0.00000014, 0.00000380, -0.00001083, 0.00000255, -0.00000260, 0.00000271, 0.00000323, -0.00000091, 0.00000045, -0.00000018, -0.00000504, -0.00000435, 0.00000629, -0.00000654, 0.00001635, -0.00000428, -0.00000321, 0.00000254, -0.00000609, 0.00000072, -0.00001986, -0.00000314, 0.00000472, -0.00001284, -0.00001619, -0.00000550, -0.00007305, -0.00000003, 0.00008528, 0.00003643, 0.00007949, -0.00000246, -0.00010602, },
		{ 0.00006717, 0.00003117, 0.00006944, 0.00002810, -0.00006944, -0.00001514, -0.00001014, -0.00001644, 0.00000380, -0.00000083, -0.00001715, -0.00000538, -0.00000240, 0.00000010, 0.00000004, -0.00000846, 0.00001664, -0.00000372, 0.00000281, 0.00000000, -0.00000812, 0.00000153, -0.00000138, 0.00000034, 0.00000202, 0.00000435, -0.00000354, 0.00000326, -0.00000958, 0.00000014, 0.00000293, -0.00000241, 0.00000158, -0.00000377, 0.00000135, -0.00000166, 0.00000074, 0.00000113, 0.00000029, -0.00000146, 0.00000710, 0.00000132, -0.00000470, -0.00000146, -0.00000319, -0.00000088, -0.00000042, 0.00000169, -0.00000038, 0.00000180, 0.00000338, 0.00000173, -0.00000534, -0.00000040, 0.00000039, -0.00000120, -0.00000030, -0.00000136, -0.00000052, -0.00000084, 0.00000399, -0.00000017, -0.00000111, -0.00000137, 0.00000384, 0.00000092, -0.00000147, -0.00000059, -0.00000109, -0.00000064, -0.00000055, 0.00000144, -0.00000562, -0.00000009, 0.00000304, 0.00000272, 0.00000002, 0.00000082, 0.00000074, -0.00000089, -0.00000313, -0.00000205, -0.00000286, -0.00000262, 0.00000855, -0.00000005, -0.00000054, 0.00000065, 0.00000104, -0.00000138, 0.00000133, -0.00000182, -0.00000104, -0.00000025, -0.00000107, 0.00000466, -0.00001072, 0.00000185, -0.00000196, 0.00000194, 0.00000387, -0.00000162, 0.00000123, -0.00000080, -0.00000402, -0.00000557, 0.00000695, -0.00000674, 0.00001546, -0.00000237, -0.00000463, 0.00000374, -0.00000753, 0.00000248, -0.00002011, -0.00000433, 0.00000510, -0.00001178, -0.00001770, -0.00000360, -0.00007137, -0.00001162, 0.00008973, 0.00003911, 0.00008347, 0.00000040, -0.00010638, -0.00000938, },
		{ 0.00003008, 0.00006394, 0.00003497, -0.00006579, -0.00001899, -0.00000820, -0.00001735, 0.00000326, -0.00000021, -0.00001595, -0.00000725, -0.00000176, -0.00000009, 0.00000062, -0.00000920, 0.00001590, -0.00000213, 0.00000130, 0.00000150, -0.00000902, 0.00000193, -0.00000174, 0.00000045, 0.00000182, 0.00000470, -0.00000354, 0.00000310, -0.00000876, -0.00000124, 0.00000382, -0.00000273, 0.00000195, -0.00000383, 0.00000104, -0.00000153, 0.00000047, 0.00000139, 0.00000036, -0.00000152, 0.00000642, 0.00000233, -0.00000494, -0.00000153, -0.00000311, -0.00000085, -0.00000068, 0.00000176, -0.00000030, 0.00000147, 0.00000343, 0.00000211, -0.00000509, -0.00000091, 0.00000060, -0.00000135, -0.00000011, -0.00000151, -0.00000035, -0.00000111, 0.00000382, 0.00000018, -0.00000108, -0.00000157, 0.00000358, 0.00000149, -0.00000171, -0.00000042, -0.00000123, -0.00000055, -0.00000077, 0.00000178, -0.00000546, -0.00000063, 0.00000297, 0.00000279, 0.00000031, 0.00000049, 0.00000104, -0.00000088, -0.00000299, -0.00000239, -0.00000228, -0.00000343, 0.00000851, 0.00000059, -0.00000084, 0.00000063, 0.00000095, -0.00000116, 0.00000106, -0.00000106, -0.00000175, 0.00000022, -0.00000186, 0.00000525, -0.00001030, 0.00000093, -0.00000118, 0.00000108, 0.00000462, -0.00000243, 0.00000202, -0.00000137, -0.00000308, -0.00000661, 0.00000730, -0.00000658, 0.00001425, -0.00000032, -0.00000608, 0.00000496, -0.00000887, 0.00000387, -0.00001986, -0.00000585, 0.00000567, -0.00001097, -0.00001873, -0.00000261, -0.00006819, -0.00002399, 0.00009334, 0.00004243, 0.00008733, 0.00000359, -0.00010701, -0.00001119, 0.00006301, },
		{ 0.00005852, 0.00004062, -0.00006128, -0.00002274, -0.00000658, -0.00001790, 0.00000247, 0.00000046, -0.00001477, -0.00000885, -0.00000156, 0.00000009, 0.00000093, -0.00000960, 0.00001476, -0.00000028, -0.00000028, 0.00000294, -0.00000978, 0.00000220, -0.00000194, 0.00000040, 0.00000170, 0.00000497, -0.00000342, 0.00000279, -0.00000784, -0.00000258, 0.00000455, -0.00000288, 0.00000213, -0.00000369, 0.00000061, -0.00000134, 0.00000012, 0.00000170, 0.00000037, -0.00000146, 0.00000565, 0.00000328, -0.00000505, -0.00000169, -0.00000303, -0.00000081, -0.00000090, 0.00000176, -0.00000016, 0.00000112, 0.00000345, 0.00000248, -0.00000480, -0.00000140, 0.00000078, -0.00000148, 0.00000006, -0.00000162, -0.00000023, -0.00000129, 0.00000357, 0.00000054, -0.00000104, -0.00000174, 0.00000324, 0.00000208, -0.00000192, -0.00000029, -0.00000134, -0.00000049, -0.00000095, 0.00000206, -0.00000521, -0.00000121, 0.00000292, 0.00000279, 0.00000068, 0.00000012, 0.00000133, -0.00000087, -0.00000282, -0.00000278, -0.00000172, -0.00000411, 0.00000828, 0.00000133, -0.00000115, 0.00000063, 0.00000083, -0.00000091, 0.00000074, -0.00000028, -0.00000238, 0.00000056, -0.00000246, 0.00000555, -0.00000957, -0.00000021, -0.00000028, 0.00000014, 0.00000544, -0.00000331, 0.00000277, -0.00000185, -0.00000226, -0.00000743, 0.00000734, -0.00000609, 0.00001278, 0.00000180, -0.00000749, 0.00000609, -0.00000996, 0.00000479, -0.00001906, -0.00000774, 0.00000646, -0.00001051, -0.00001918, -0.00000258, -0.00006361, -0.00003685, 0.00009583, 0.00004650, 0.00009103, 0.00000716, -0.00010793, -0.00001276, 0.00005882, 0.00002923, },
		{ 0.00004514, -0.00005617, -0.00002617, -0.00000546, -0.00001800, 0.00000137, 0.00000123, -0.00001370, -0.00001010, -0.00000184, 0.00000064, 0.00000098, -0.00000967, 0.00001327, 0.00000174, -0.00000187, 0.00000426, -0.00001038, 0.00000232, -0.00000199, 0.00000024, 0.00000163, 0.00000518, -0.00000319, 0.00000238, -0.00000687, -0.00000383, 0.00000507, -0.00000282, 0.00000211, -0.00000336, 0.00000008, -0.00000108, -0.00000031, 0.00000205, 0.00000034, -0.00000132, 0.00000484, 0.00000416, -0.00000502, -0.00000192, -0.00000294, -0.00000075, -0.00000110, 0.00000171, 0.00000002, 0.00000078, 0.00000343, 0.00000284, -0.00000446, -0.00000188, 0.00000092, -0.00000160, 0.00000019, -0.00000169, -0.00000016, -0.00000139, 0.00000327, 0.00000090, -0.00000099, -0.00000185, 0.00000283, 0.00000267, -0.00000208, -0.00000020, -0.00000140, -0.00000047, -0.00000111, 0.00000227, -0.00000485, -0.00000183, 0.00000289, 0.00000272, 0.00000110, -0.00000026, 0.00000160, -0.00000084, -0.00000260, -0.00000320, -0.00000119, -0.00000465, 0.00000788, 0.00000213, -0.00000145, 0.00000062, 0.00000069, -0.00000065, 0.00000039, 0.00000049, -0.00000287, 0.00000072, -0.00000284, 0.00000554, -0.00000855, -0.00000152, 0.00000071, -0.00000084, 0.00000632, -0.00000422, 0.00000344, -0.00000221, -0.00000157, -0.00000802, 0.00000708, -0.00000530, 0.00001112, 0.00000388, -0.00000876, 0.00000703, -0.00001070, 0.00000515, -0.00001766, -0.00001000, 0.00000750, -0.00001042, -0.00001903, -0.00000349, -0.00005779, -0.00004982, 0.00009689, 0.00005142, 0.00009454, 0.00001116, -0.00010915, -0.00001409, 0.00005460, 0.00002861, 0.00005323, },
		{ -0.00005072, -0.00002904, -0.00000493, -0.00001761, -0.00000007, 0.00000214, -0.00001277, -0.00001095, -0.00000259, 0.00000153, 0.00000081, -0.00000946, 0.00001148, 0.00000387, -0.00000338, 0.00000541, -0.00001080, 0.00000230, -0.00000190, -0.00000002, 0.00000157, 0.00000535, -0.00000287, 0.00000189, -0.00000591, -0.00000492, 0.00000535, -0.00000254, 0.00000190, -0.00000287, -0.00000054, -0.00000078, -0.00000079, 0.00000243, 0.00000028, -0.00000111, 0.00000400, 0.00000493, -0.00000486, -0.00000220, -0.00000286, -0.00000066, -0.00000128, 0.00000161, 0.00000023, 0.00000046, 0.00000335, 0.00000319, -0.00000410, -0.00000232, 0.00000102, -0.00000168, 0.00000028, -0.00000170, -0.00000015, -0.00000142, 0.00000291, 0.00000124, -0.00000093, -0.00000191, 0.00000235, 0.00000325, -0.00000217, -0.00000018, -0.00000140, -0.00000050, -0.00000123, 0.00000239, -0.00000439, -0.00000248, 0.00000287, 0.00000260, 0.00000156, -0.00000064, 0.00000183, -0.00000080, -0.00000235, -0.00000364, -0.00000073, -0.00000503, 0.00000732, 0.00000297, -0.00000170, 0.00000060, 0.00000054, -0.00000040, 0.00000003, 0.00000120, -0.00000318, 0.00000068, -0.00000298, 0.00000523, -0.00000727, -0.00000295, 0.00000173, -0.00000182, 0.00000721, -0.00000508, 0.00000398, -0.00000240, -0.00000106, -0.00000837, 0.00000653, -0.00000428, 0.00000937, 0.00000585, -0.00000981, 0.00000769, -0.00001098, 0.00000488, -0.00001565, -0.00001262, 0.00000879, -0.00001071, -0.00001828, -0.00000528, -0.00005097, -0.00006251, 0.00009620, 0.00005725, 0.00009784, 0.00001561, -0.00011069, -0.00001515, 0.00005032, 0.00002820, 0.00004810, 0.00004865, },
		{ -0.00003119, -0.00000509, -0.00001669, -0.00000184, 0.00000317, -0.00001202, -0.00001138, -0.00000380, 0.00000269, 0.00000048, -0.00000902, 0.00000948, 0.00000601, -0.00000475, 0.00000635, -0.00001100, 0.00000211, -0.00000168, -0.00000035, 0.00000150, 0.00000551, -0.00000250, 0.00000138, -0.00000499, -0.00000581, 0.00000535, -0.00000205, 0.00000151, -0.00000223, -0.00000120, -0.00000046, -0.00000131, 0.00000281, 0.00000021, -0.00000086, 0.00000318, 0.00000557, -0.00000456, -0.00000253, -0.00000282, -0.00000055, -0.00000143, 0.00000147, 0.00000045, 0.00000017, 0.00000322, 0.00000353, -0.00000370, -0.00000273, 0.00000108, -0.00000173, 0.00000032, -0.00000166, -0.00000019, -0.00000139, 0.00000252, 0.00000156, -0.00000086, -0.00000192, 0.00000183, 0.00000380, -0.00000219, -0.00000021, -0.00000135, -0.00000056, -0.00000131, 0.00000243, -0.00000385, -0.00000314, 0.00000285, 0.00000242, 0.00000204, -0.00000099, 0.00000201, -0.00000074, -0.00000208, -0.00000406, -0.00000036, -0.00000525, 0.00000662, 0.00000383, -0.00000190, 0.00000055, 0.00000042, -0.00000017, -0.00000031, 0.00000181, -0.00000328, 0.00000042, -0.00000288, 0.00000464, -0.00000578, -0.00000446, 0.00000274, -0.00000277, 0.00000807, -0.00000587, 0.00000432, -0.00000239, -0.00000072, -0.00000847, 0.00000573, -0.00000309, 0.00000761, 0.00000760, -0.00001054, 0.00000798, -0.00001075, 0.00000397, -0.00001309, -0.00001553, 0.00001027, -0.00001135, -0.00001699, -0.00000781, -0.00004344, -0.00007448, 0.00009349, 0.00006403, 0.00010094, 0.00002056, -0.00011259, -0.00001595, 0.00004596, 0.00002800, 0.00004311, 0.00005132, -0.00004520, },
		{ -0.00000594, -0.00001527, -0.00000391, 0.00000431, -0.00001145, -0.00001141, -0.00000540, 0.00000403, 0.00000007, -0.00000841, 0.00000733, 0.00000808, -0.00000592, 0.00000704, -0.00001099, 0.00000178, -0.00000135, -0.00000072, 0.00000139, 0.00000566, -0.00000209, 0.00000087, -0.00000418, -0.00000647, 0.00000508, -0.00000135, 0.00000098, -0.00000148, -0.00000188, -0.00000014, -0.00000185, 0.00000318, 0.00000013, -0.00000058, 0.00000239, 0.00000607, -0.00000413, -0.00000289, -0.00000281, -0.00000042, -0.00000155, 0.00000129, 0.00000068, -0.00000007, 0.00000302, 0.00000386, -0.00000329, -0.00000310, 0.00000108, -0.00000174, 0.00000032, -0.00000156, -0.00000028, -0.00000129, 0.00000210, 0.00000186, -0.00000079, -0.00000189, 0.00000127, 0.00000429, -0.00000212, -0.00000031, -0.00000125, -0.00000066, -0.00000135, 0.00000239, -0.00000322, -0.00000380, 0.00000283, 0.00000221, 0.00000252, -0.00000129, 0.00000213, -0.00000065, -0.00000180, -0.00000444, -0.00000010, -0.00000530, 0.00000579, 0.00000467, -0.00000201, 0.00000045, 0.00000032, 0.00000002, -0.00000060, 0.00000229, -0.00000315, -0.00000007, -0.00000256, 0.00000380, -0.00000414, -0.00000597, 0.00000367, -0.00000364, 0.00000885, -0.00000651, 0.00000444, -0.00000218, -0.00000058, -0.00000835, 0.00000470, -0.00000179, 0.00000593, 0.00000906, -0.00001091, 0.00000784, -0.00000996, 0.00000244, -0.00001004, -0.00001864, 0.00001187, -0.00001224, -0.00001529, -0.00001088, -0.00003556, -0.00008530, 0.00008851, 0.00007175, 0.00010388, 0.00002599, -0.00011483, -0.00001648, 0.00004152, 0.00002802, 0.00003827, 0.00005329, -0.00003985, -0.00003247, },
		{ -0.00001342, -0.00000618, 0.00000551, -0.00001102, -0.00001107, -0.00000731, 0.00000547, -0.00000034, -0.00000770, 0.00000513, 0.00001001, -0.00000682, 0.00000746, -0.00001077, 0.00000132, -0.00000094, -0.00000108, 0.00000122, 0.00000582, -0.00000166, 0.00000039, -0.00000349, -0.00000689, 0.00000454, -0.00000048, 0.00000034, -0.00000068, -0.00000253, 0.00000016, -0.00000239, 0.00000352, 0.00000008, -0.00000030, 0.00000167, 0.00000642, -0.00000359, -0.00000324, -0.00000285, -0.00000026, -0.00000163, 0.00000107, 0.00000090, -0.00000025, 0.00000276, 0.00000417, -0.00000285, -0.00000341, 0.00000104, -0.00000172, 0.00000026, -0.00000142, -0.00000041, -0.00000116, 0.00000168, 0.00000212, -0.00000070, -0.00000181, 0.00000069, 0.00000471, -0.00000195, -0.00000047, -0.00000110, -0.00000079, -0.00000135, 0.00000227, -0.00000252, -0.00000444, 0.00000278, 0.00000198, 0.00000299, -0.00000154, 0.00000218, -0.00000054, -0.00000152, -0.00000478, 0.00000004, -0.00000518, 0.00000486, 0.00000546, -0.00000202, 0.00000030, 0.00000025, 0.00000015, -0.00000083, 0.00000261, -0.00000277, -0.00000077, -0.00000204, 0.00000278, -0.00000242, -0.00000740, 0.00000448, -0.00000438, 0.00000950, -0.00000697, 0.00000429, -0.00000174, -0.00000061, -0.00000801, 0.00000351, -0.00000045, 0.00000441, 0.00001017, -0.00001087, 0.00000726, -0.00000863, 0.00000034, -0.00000661, -0.00002182, 0.00001348, -0.00001329, -0.00001332, -0.00001427, -0.00002769, -0.00009454, 0.00008108, 0.00008032, 0.00010674, 0.00003190, -0.00011740, -0.00001675, 0.00003697, 0.00002826, 0.00003354, 0.00005472, -0.00003484, -0.00003282, -0.00000746, },
		{ -0.00000857, 0.00000668, -0.00001068, -0.00001044, -0.00000943, 0.00000687, -0.00000067, -0.00000696, 0.00000294, 0.00001170, -0.00000740, 0.00000759, -0.00001034, 0.00000075, -0.00000048, -0.00000142, 0.00000098, 0.00000599, -0.00000122, -0.00000003, -0.00000296, -0.00000705, 0.00000376, 0.00000052, -0.00000036, 0.00000012, -0.00000311, 0.00000041, -0.00000291, 0.00000380, 0.00000007, -0.00000004, 0.00000103, 0.00000661, -0.00000295, -0.00000358, -0.00000294, -0.00000009, -0.00000168, 0.00000083, 0.00000110, -0.00000038, 0.00000245, 0.00000447, -0.00000240, -0.00000368, 0.00000095, -0.00000166, 0.00000016, -0.00000123, -0.00000058, -0.00000099, 0.00000126, 0.00000233, -0.00000061, -0.00000169, 0.00000011, 0.00000504, -0.00000168, -0.00000069, -0.00000091, -0.00000095, -0.00000133, 0.00000208, -0.00000177, -0.00000504, 0.00000270, 0.00000175, 0.00000341, -0.00000171, 0.00000216, -0.00000038, -0.00000125, -0.00000504, 0.00000004, -0.00000492, 0.00000387, 0.00000616, -0.00000192, 0.00000010, 0.00000022, 0.00000024, -0.00000100, 0.00000277, -0.00000217, -0.00000164, -0.00000139, 0.00000163, -0.00000069, -0.00000870, 0.00000510, -0.00000495, 0.00000999, -0.00000719, 0.00000385, -0.00000109, -0.00000080, -0.00000749, 0.00000221, 0.00000084, 0.00000312, 0.00001089, -0.00001040, 0.00000623, -0.00000679, -0.00000221, -0.00000296, -0.00002493, 0.00001497, -0.00001435, -0.00001126, -0.00001772, -0.00002021, -0.00010183, 0.00007111, 0.00008959, 0.00010965, 0.00003825, -0.00012027, -0.00001679, 0.00003233, 0.00002870, 0.00002891, 0.00005574, -0.00003034, -0.00003222, -0.00000956, -0.00001126, },
		{ 0.00000773, -0.00001037, -0.00000961, -0.00001163, 0.00000813, -0.00000082, -0.00000628, 0.00000086, 0.00001309, -0.00000764, 0.00000743, -0.00000973, 0.00000009, 0.00000001, -0.00000171, 0.00000065, 0.00000616, -0.00000077, -0.00000038, -0.00000258, -0.00000697, 0.00000278, 0.00000158, -0.00000106, 0.00000088, -0.00000358, 0.00000058, -0.00000337, 0.00000401, 0.00000012, 0.00000017, 0.00000049, 0.00000665, -0.00000223, -0.00000387, -0.00000310, 0.00000010, -0.00000169, 0.00000057, 0.00000127, -0.00000044, 0.00000209, 0.00000473, -0.00000193, -0.00000390, 0.00000082, -0.00000156, 0.00000001, -0.00000100, -0.00000077, -0.00000081, 0.00000086, 0.00000250, -0.00000051, -0.00000154, -0.00000046, 0.00000527, -0.00000131, -0.00000096, -0.00000069, -0.00000112, -0.00000128, 0.00000183, -0.00000100, -0.00000558, 0.00000256, 0.00000154, 0.00000377, -0.00000179, 0.00000205, -0.00000019, -0.00000101, -0.00000521, -0.00000010, -0.00000452, 0.00000284, 0.00000676, -0.00000170, -0.00000016, 0.00000023, 0.00000026, -0.00000108, 0.00000277, -0.00000135, -0.00000265, -0.00000064, 0.00000043, 0.00000095, -0.00000978, 0.00000548, -0.00000531, 0.00001027, -0.00000714, 0.00000312, -0.00000024, -0.00000113, -0.00000683, 0.00000084, 0.00000202, 0.00000211, 0.00001118, -0.00000951, 0.00000479, -0.00000452, -0.00000509, 0.00000075, -0.00002780, 0.00001620, -0.00001528, -0.00000930, -0.00002096, -0.00001348, -0.00010685, 0.00005858, 0.00009932, 0.00011276, 0.00004497, -0.00012338, -0.00001664, 0.00002763, 0.00002932, 0.00002439, 0.00005645, -0.00002644, -0.00003073, -0.00001211, -0.00000891, -0.00001095, },
		{ -0.00001001, -0.00000866, -0.00001380, 0.00000913, -0.00000073, -0.00000569, -0.00000105, 0.00001413, -0.00000751, 0.00000700, -0.00000895, -0.00000063, 0.00000050, -0.00000192, 0.00000024, 0.00000633, -0.00000031, -0.00000066, -0.00000237, -0.00000668, 0.00000165, 0.00000266, -0.00000171, 0.00000155, -0.00000392, 0.00000066, -0.00000378, 0.00000413, 0.00000024, 0.00000034, 0.00000006, 0.00000653, -0.00000146, -0.00000412, -0.00000333, 0.00000028, -0.00000166, 0.00000029, 0.00000141, -0.00000044, 0.00000169, 0.00000495, -0.00000144, -0.00000408, 0.00000065, -0.00000144, -0.00000018, -0.00000075, -0.00000097, -0.00000061, 0.00000048, 0.00000263, -0.00000040, -0.00000137, -0.00000100, 0.00000539, -0.00000084, -0.00000127, -0.00000045, -0.00000129, -0.00000122, 0.00000153, -0.00000022, -0.00000604, 0.00000236, 0.00000137, 0.00000406, -0.00000176, 0.00000185, 0.00000004, -0.00000080, -0.00000528, -0.00000037, -0.00000401, 0.00000182, 0.00000723, -0.00000135, -0.00000047, 0.00000028, 0.00000024, -0.00000109, 0.00000261, -0.00000036, -0.00000373, 0.00000012, -0.00000074, 0.00000244, -0.00001059, 0.00000558, -0.00000543, 0.00001031, -0.00000681, 0.00000209, 0.00000079, -0.00000156, -0.00000606, -0.00000054, 0.00000304, 0.00000142, 0.00001105, -0.00000824, 0.00000299, -0.00000192, -0.00000814, 0.00000432, -0.00003025, 0.00001702, -0.00001592, -0.00000764, -0.00002373, -0.00000783, -0.00010937, 0.00004362, 0.00010920, 0.00011628, 0.00005196, -0.00012666, -0.00001638, 0.00002290, 0.00003008, 0.00001998, 0.00005691, -0.00002316, -0.00002845, -0.00001492, -0.00000655, -0.00001317, 0.00000856, },
		{ -0.00000770, -0.00001581, 0.00000977, -0.00000033, -0.00000525, -0.00000275, 0.00001479, -0.00000701, 0.00000631, -0.00000805, -0.00000138, 0.00000097, -0.00000205, -0.00000024, 0.00000647, 0.00000016, -0.00000086, -0.00000229, -0.00000621, 0.00000043, 0.00000368, -0.00000225, 0.00000208, -0.00000409, 0.00000064, -0.00000411, 0.00000413, 0.00000044, 0.00000044, -0.00000025, 0.00000628, -0.00000065, -0.00000429, -0.00000361, 0.00000046, -0.00000157, 0.00000000, 0.00000152, -0.00000038, 0.00000128, 0.00000512, -0.00000092, -0.00000420, 0.00000045, -0.00000129, -0.00000039, -0.00000049, -0.00000116, -0.00000043, 0.00000015, 0.00000270, -0.00000029, -0.00000117, -0.00000149, 0.00000539, -0.00000028, -0.00000160, -0.00000022, -0.00000145, -0.00000115, 0.00000121, 0.00000054, -0.00000640, 0.00000207, 0.00000124, 0.00000425, -0.00000163, 0.00000157, 0.00000031, -0.00000063, -0.00000524, -0.00000076, -0.00000342, 0.00000084, 0.00000754, -0.00000088, -0.00000082, 0.00000035, 0.00000017, -0.00000103, 0.00000230, 0.00000078, -0.00000483, 0.00000084, -0.00000182, 0.00000369, -0.00001107, 0.00000538, -0.00000530, 0.00001011, -0.00000618, 0.00000079, 0.00000196, -0.00000206, -0.00000524, -0.00000186, 0.00000385, 0.00000107, 0.00001051, -0.00000662, 0.00000091, 0.00000088, -0.00001118, 0.00000756, -0.00003212, 0.00001729, -0.00001613, -0.00000645, -0.00002582, -0.00000351, -0.00010926, 0.00002644, 0.00011882, 0.00012046, 0.00005914, -0.00013001, -0.00001608, 0.00001820, 0.00003091, 0.00001573, 0.00005717, -0.00002048, -0.00002555, -0.00001780, -0.00000436, -0.00001510, 0.00000907, -0.00000951, },
		{ -0.00001755, 0.00000998, 0.00000042, -0.00000499, -0.00000418, 0.00001503, -0.00000615, 0.00000540, -0.00000705, -0.00000213, 0.00000139, -0.00000209, -0.00000078, 0.00000657, 0.00000066, -0.00000100, -0.00000235, -0.00000560, -0.00000081, 0.00000459, -0.00000263, 0.00000244, -0.00000409, 0.00000051, -0.00000436, 0.00000403, 0.00000073, 0.00000047, -0.00000045, 0.00000591, 0.00000018, -0.00000437, -0.00000394, 0.00000061, -0.00000143, -0.00000029, 0.00000158, -0.00000028, 0.00000085, 0.00000523, -0.00000039, -0.00000428, 0.00000021, -0.00000113, -0.00000063, -0.00000022, -0.00000134, -0.00000027, -0.00000014, 0.00000272, -0.00000017, -0.00000096, -0.00000193, 0.00000526, 0.00000035, -0.00000194, 0.00000001, -0.00000160, -0.00000108, 0.00000086, 0.00000126, -0.00000663, 0.00000170, 0.00000118, 0.00000434, -0.00000138, 0.00000122, 0.00000061, -0.00000050, -0.00000510, -0.00000127, -0.00000277, -0.00000007, 0.00000768, -0.00000029, -0.00000119, 0.00000043, 0.00000007, -0.00000092, 0.00000187, 0.00000200, -0.00000588, 0.00000144, -0.00000273, 0.00000466, -0.00001119, 0.00000486, -0.00000491, 0.00000966, -0.00000526, -0.00000076, 0.00000320, -0.00000257, -0.00000441, -0.00000308, 0.00000440, 0.00000108, 0.00000961, -0.00000475, -0.00000132, 0.00000373, -0.00001403, 0.00001029, -0.00003326, 0.00001690, -0.00001578, -0.00000590, -0.00002702, -0.00000073, -0.00010650, 0.00000740, 0.00012772, 0.00012555, 0.00006637, -0.00013333, -0.00001582, 0.00001360, 0.00003174, 0.00001169, 0.00005721, -0.00001835, -0.00002223, -0.00002051, -0.00000253, -0.00001658, 0.00000916, -0.00000881, -0.00000682, },
		{ 0.00000968, 0.00000152, -0.00000492, -0.00000534, 0.00001486, -0.00000497, 0.00000432, -0.00000601, -0.00000285, 0.00000175, -0.00000204, -0.00000135, 0.00000660, 0.00000118, -0.00000107, -0.00000251, -0.00000490, -0.00000201, 0.00000532, -0.00000281, 0.00000260, -0.00000389, 0.00000028, -0.00000452, 0.00000380, 0.00000110, 0.00000043, -0.00000054, 0.00000543, 0.00000098, -0.00000437, -0.00000432, 0.00000072, -0.00000124, -0.00000059, 0.00000161, -0.00000014, 0.00000044, 0.00000526, 0.00000017, -0.00000432, -0.00000004, -0.00000095, -0.00000088, 0.00000004, -0.00000150, -0.00000013, -0.00000039, 0.00000270, -0.00000006, -0.00000076, -0.00000230, 0.00000502, 0.00000104, -0.00000226, 0.00000021, -0.00000171, -0.00000103, 0.00000052, 0.00000191, -0.00000672, 0.00000122, 0.00000118, 0.00000433, -0.00000104, 0.00000079, 0.00000093, -0.00000041, -0.00000486, -0.00000188, -0.00000211, -0.00000088, 0.00000765, 0.00000039, -0.00000156, 0.00000050, -0.00000005, -0.00000077, 0.00000136, 0.00000325, -0.00000681, 0.00000186, -0.00000341, 0.00000529, -0.00001092, 0.00000402, -0.00000427, 0.00000898, -0.00000408, -0.00000249, 0.00000447, -0.00000304, -0.00000362, -0.00000416, 0.00000468, 0.00000143, 0.00000840, -0.00000272, -0.00000361, 0.00000647, -0.00001649, 0.00001233, -0.00003355, 0.00001576, -0.00001480, -0.00000608, -0.00002723, 0.00000043, -0.00010120, -0.00001307, 0.00013538, 0.00013182, 0.00007355, -0.00013652, -0.00001572, 0.00000920, 0.00003249, 0.00000793, 0.00005701, -0.00001664, -0.00001872, -0.00002283, -0.00000121, -0.00001751, 0.00000876, -0.00000784, -0.00000609, -0.00001893, },
		{ 0.00000298, -0.00000502, -0.00000622, 0.00001429, -0.00000351, 0.00000313, -0.00000497, -0.00000351, 0.00000204, -0.00000192, -0.00000193, 0.00000654, 0.00000176, -0.00000110, -0.00000274, -0.00000417, -0.00000312, 0.00000582, -0.00000275, 0.00000255, -0.00000352, -0.00000005, -0.00000459, 0.00000346, 0.00000155, 0.00000034, -0.00000053, 0.00000488, 0.00000174, -0.00000426, -0.00000472, 0.00000078, -0.00000099, -0.00000089, 0.00000160, 0.00000003, 0.00000005, 0.00000521, 0.00000074, -0.00000431, -0.00000032, -0.00000077, -0.00000113, 0.00000028, -0.00000162, -0.00000003, -0.00000057, 0.00000263, 0.00000006, -0.00000055, -0.00000260, 0.00000466, 0.00000177, -0.00000256, 0.00000037, -0.00000178, -0.00000100, 0.00000018, 0.00000249, -0.00000667, 0.00000065, 0.00000124, 0.00000421, -0.00000060, 0.00000031, 0.00000127, -0.00000036, -0.00000452, -0.00000254, -0.00000147, -0.00000156, 0.00000744, 0.00000116, -0.00000190, 0.00000056, -0.00000017, -0.00000059, 0.00000078, 0.00000448, -0.00000755, 0.00000205, -0.00000381, 0.00000555, -0.00001028, 0.00000289, -0.00000343, 0.00000808, -0.00000267, -0.00000435, 0.00000569, -0.00000342, -0.00000290, -0.00000507, 0.00000468, 0.00000208, 0.00000695, -0.00000063, -0.00000581, 0.00000894, -0.00001842, 0.00001356, -0.00003291, 0.00001384, -0.00001312, -0.00000705, -0.00002636, -0.00000005, -0.00009359, -0.00003439, 0.00014122, 0.00013953, 0.00008057, -0.00013947, -0.00001587, 0.00000507, 0.00003305, 0.00000453, 0.00005651, -0.00001522, -0.00001525, -0.00002454, -0.00000056, -0.00001780, 0.00000782, -0.00000658, -0.00000558, -0.00001990, 0.00000886, },
		{ -0.00000526, -0.00000684, 0.00001337, -0.00000183, 0.00000188, -0.00000397, -0.00000409, 0.00000225, -0.00000172, -0.00000249, 0.00000636, 0.00000239, -0.00000110, -0.00000301, -0.00000345, -0.00000407, 0.00000606, -0.00000246, 0.00000229, -0.00000297, -0.00000045, -0.00000459, 0.00000301, 0.00000206, 0.00000019, -0.00000044, 0.00000428, 0.00000245, -0.00000406, -0.00000514, 0.00000077, -0.00000067, -0.00000119, 0.00000156, 0.00000022, -0.00000029, 0.00000508, 0.00000133, -0.00000426, -0.00000061, -0.00000059, -0.00000138, 0.00000048, -0.00000170, 0.00000002, -0.00000071, 0.00000252, 0.00000018, -0.00000036, -0.00000281, 0.00000419, 0.00000252, -0.00000280, 0.00000047, -0.00000181, -0.00000100, -0.00000013, 0.00000296, -0.00000645, -0.00000003, 0.00000137, 0.00000401, -0.00000008, -0.00000020, 0.00000160, -0.00000033, -0.00000411, -0.00000325, -0.00000087, -0.00000210, 0.00000706, 0.00000198, -0.00000220, 0.00000057, -0.00000028, -0.00000042, 0.00000018, 0.00000564, -0.00000806, 0.00000195, -0.00000392, 0.00000545, -0.00000927, 0.00000151, -0.00000240, 0.00000702, -0.00000109, -0.00000627, 0.00000680, -0.00000367, -0.00000229, -0.00000577, 0.00000439, 0.00000299, 0.00000536, 0.00000142, -0.00000780, 0.00001101, -0.00001965, 0.00001386, -0.00003129, 0.00001114, -0.00001077, -0.00000883, -0.00002445, -0.00000208, -0.00008399, -0.00005593, 0.00014469, 0.00014890, 0.00008738, -0.00014210, -0.00001635, 0.00000130, 0.00003334, 0.00000159, 0.00005564, -0.00001394, -0.00001206, -0.00002547, -0.00000068, -0.00001739, 0.00000633, -0.00000500, -0.00000532, -0.00002041, 0.00000752, 0.00000474, },
		{ -0.00000723, 0.00001213, -0.00000000, 0.00000063, -0.00000306, -0.00000457, 0.00000238, -0.00000148, -0.00000300, 0.00000604, 0.00000307, -0.00000106, -0.00000330, -0.00000278, -0.00000482, 0.00000601, -0.00000192, 0.00000183, -0.00000229, -0.00000091, -0.00000453, 0.00000247, 0.00000262, 0.00000001, -0.00000029, 0.00000365, 0.00000307, -0.00000375, -0.00000556, 0.00000069, -0.00000031, -0.00000147, 0.00000149, 0.00000042, -0.00000059, 0.00000486, 0.00000193, -0.00000416, -0.00000090, -0.00000041, -0.00000161, 0.00000064, -0.00000172, 0.00000004, -0.00000079, 0.00000237, 0.00000030, -0.00000019, -0.00000294, 0.00000362, 0.00000327, -0.00000297, 0.00000051, -0.00000178, -0.00000102, -0.00000040, 0.00000332, -0.00000608, -0.00000078, 0.00000155, 0.00000372, 0.00000049, -0.00000072, 0.00000192, -0.00000033, -0.00000363, -0.00000396, -0.00000036, -0.00000246, 0.00000652, 0.00000282, -0.00000244, 0.00000054, -0.00000036, -0.00000026, -0.00000041, 0.00000666, -0.00000828, 0.00000156, -0.00000372, 0.00000499, -0.00000794, -0.00000007, -0.00000126, 0.00000584, 0.00000060, -0.00000817, 0.00000772, -0.00000373, -0.00000183, -0.00000626, 0.00000383, 0.00000412, 0.00000370, 0.00000331, -0.00000948, 0.00001255, -0.00002010, 0.00001319, -0.00002873, 0.00000770, -0.00000779, -0.00001134, -0.00002158, -0.00000547, -0.00007285, -0.00007700, 0.00014523, 0.00016010, 0.00009394, -0.00014435, -0.00001725, -0.00000204, 0.00003327, -0.00000083, 0.00005433, -0.00001264, -0.00000934, -0.00002547, -0.00000162, -0.00001627, 0.00000430, -0.00000313, -0.00000530, -0.00002047, 0.00000570, 0.00000674, -0.00000559, },
		{ 0.00001065, 0.00000190, -0.00000054, -0.00000226, -0.00000493, 0.00000243, -0.00000120, -0.00000342, 0.00000557, 0.00000381, -0.00000101, -0.00000358, -0.00000220, -0.00000535, 0.00000567, -0.00000115, 0.00000122, -0.00000151, -0.00000139, -0.00000443, 0.00000186, 0.00000321, -0.00000019, -0.00000010, 0.00000302, 0.00000360, -0.00000336, -0.00000596, 0.00000051, 0.00000012, -0.00000174, 0.00000138, 0.00000061, -0.00000082, 0.00000455, 0.00000253, -0.00000401, -0.00000121, -0.00000024, -0.00000181, 0.00000076, -0.00000170, 0.00000001, -0.00000082, 0.00000220, 0.00000041, -0.00000004, -0.00000297, 0.00000297, 0.00000399, -0.00000306, 0.00000048, -0.00000169, -0.00000109, -0.00000063, 0.00000357, -0.00000556, -0.00000161, 0.00000176, 0.00000337, 0.00000110, -0.00000124, 0.00000221, -0.00000033, -0.00000311, -0.00000465, 0.00000004, -0.00000265, 0.00000583, 0.00000366, -0.00000258, 0.00000044, -0.00000039, -0.00000012, -0.00000097, 0.00000751, -0.00000819, 0.00000086, -0.00000322, 0.00000420, -0.00000634, -0.00000178, -0.00000007, 0.00000460, 0.00000233, -0.00000997, 0.00000839, -0.00000358, -0.00000153, -0.00000651, 0.00000302, 0.00000538, 0.00000208, 0.00000494, -0.00001073, 0.00001347, -0.00001970, 0.00001156, -0.00002529, 0.00000366, -0.00000430, -0.00001447, -0.00001792, -0.00000995, -0.00006068, -0.00009688, 0.00014233, 0.00017322, 0.00010029, -0.00014622, -0.00001858, -0.00000491, 0.00003277, -0.00000266, 0.00005253, -0.00001120, -0.00000729, -0.00002446, -0.00000340, -0.00001448, 0.00000179, -0.00000103, -0.00000550, -0.00002011, 0.00000347, 0.00000891, -0.00000596, -0.00000743, },
		{ 0.00000381, -0.00000158, -0.00000163, -0.00000517, 0.00000241, -0.00000091, -0.00000375, 0.00000494, 0.00000460, -0.00000094, -0.00000384, -0.00000174, -0.00000562, 0.00000505, -0.00000019, 0.00000049, -0.00000067, -0.00000185, -0.00000431, 0.00000119, 0.00000380, -0.00000038, 0.00000011, 0.00000241, 0.00000402, -0.00000288, -0.00000632, 0.00000025, 0.00000058, -0.00000198, 0.00000125, 0.00000078, -0.00000099, 0.00000416, 0.00000311, -0.00000381, -0.00000152, -0.00000009, -0.00000199, 0.00000082, -0.00000161, -0.00000006, -0.00000082, 0.00000200, 0.00000053, 0.00000008, -0.00000293, 0.00000226, 0.00000465, -0.00000304, 0.00000038, -0.00000155, -0.00000118, -0.00000082, 0.00000369, -0.00000490, -0.00000248, 0.00000199, 0.00000299, 0.00000171, -0.00000171, 0.00000244, -0.00000032, -0.00000257, -0.00000528, 0.00000029, -0.00000266, 0.00000503, 0.00000446, -0.00000261, 0.00000027, -0.00000039, -0.00000003, -0.00000147, 0.00000816, -0.00000778, -0.00000013, -0.00000246, 0.00000314, -0.00000456, -0.00000353, 0.00000110, 0.00000338, 0.00000402, -0.00001157, 0.00000875, -0.00000318, -0.00000140, -0.00000655, 0.00000200, 0.00000673, 0.00000058, 0.00000624, -0.00001149, 0.00001370, -0.00001843, 0.00000902, -0.00002110, -0.00000085, -0.00000046, -0.00001804, -0.00001368, -0.00001517, -0.00004806, -0.00011483, 0.00013556, 0.00018826, 0.00010654, -0.00014771, -0.00002038, -0.00000728, 0.00003179, -0.00000386, 0.00005019, -0.00000951, -0.00000600, -0.00002240, -0.00000594, -0.00001212, -0.00000108, 0.00000122, -0.00000588, -0.00001938, 0.00000092, 0.00001112, -0.00000627, -0.00000750, 0.00000898, },
		{ -0.00000243, -0.00000119, -0.00000527, 0.00000233, -0.00000063, -0.00000396, 0.00000416, 0.00000542, -0.00000084, -0.00000406, -0.00000142, -0.00000564, 0.00000416, 0.00000092, -0.00000031, 0.00000018, -0.00000226, -0.00000419, 0.00000049, 0.00000436, -0.00000054, 0.00000032, 0.00000185, 0.00000432, -0.00000233, -0.00000663, -0.00000011, 0.00000108, -0.00000219, 0.00000110, 0.00000094, -0.00000108, 0.00000370, 0.00000367, -0.00000355, -0.00000182, 0.00000004, -0.00000213, 0.00000082, -0.00000148, -0.00000017, -0.00000077, 0.00000179, 0.00000064, 0.00000017, -0.00000280, 0.00000151, 0.00000525, -0.00000292, 0.00000020, -0.00000137, -0.00000130, -0.00000096, 0.00000370, -0.00000412, -0.00000337, 0.00000220, 0.00000259, 0.00000230, -0.00000213, 0.00000261, -0.00000029, -0.00000204, -0.00000582, 0.00000040, -0.00000250, 0.00000413, 0.00000519, -0.00000252, 0.00000002, -0.00000034, 0.00000001, -0.00000189, 0.00000859, -0.00000704, -0.00000139, -0.00000149, 0.00000189, -0.00000269, -0.00000524, 0.00000218, 0.00000224, 0.00000560, -0.00001291, 0.00000876, -0.00000252, -0.00000145, -0.00000637, 0.00000081, 0.00000809, -0.00000073, 0.00000713, -0.00001171, 0.00001323, -0.00001633, 0.00000569, -0.00001635, -0.00000560, 0.00000353, -0.00002184, -0.00000914, -0.00002075, -0.00003560, -0.00013018, 0.00012461, 0.00020508, 0.00011289, -0.00014892, -0.00002261, -0.00000916, 0.00003034, -0.00000440, 0.00004730, -0.00000749, -0.00000555, -0.00001935, -0.00000915, -0.00000932, -0.00000418, 0.00000350, -0.00000633, -0.00001837, -0.00000183, 0.00001326, -0.00000646, -0.00000750, 0.00000721, 0.00000563, },
		{ -0.00000095, -0.00000525, 0.00000220, -0.00000037, -0.00000403, 0.00000324, 0.00000624, -0.00000071, -0.00000424, -0.00000123, -0.00000543, 0.00000306, 0.00000212, -0.00000111, 0.00000097, -0.00000258, -0.00000410, -0.00000021, 0.00000488, -0.00000066, 0.00000050, 0.00000136, 0.00000450, -0.00000172, -0.00000687, -0.00000055, 0.00000159, -0.00000236, 0.00000091, 0.00000107, -0.00000111, 0.00000319, 0.00000418, -0.00000323, -0.00000213, 0.00000015, -0.00000224, 0.00000077, -0.00000130, -0.00000031, -0.00000070, 0.00000157, 0.00000074, 0.00000023, -0.00000260, 0.00000073, 0.00000574, -0.00000268, -0.00000004, -0.00000114, -0.00000145, -0.00000106, 0.00000360, -0.00000324, -0.00000425, 0.00000238, 0.00000221, 0.00000285, -0.00000247, 0.00000270, -0.00000023, -0.00000153, -0.00000626, 0.00000033, -0.00000217, 0.00000317, 0.00000583, -0.00000229, -0.00000030, -0.00000024, 0.00000001, -0.00000222, 0.00000879, -0.00000599, -0.00000285, -0.00000037, 0.00000052, -0.00000080, -0.00000682, 0.00000308, 0.00000125, 0.00000699, -0.00001392, 0.00000836, -0.00000160, -0.00000167, -0.00000601, -0.00000050, 0.00000939, -0.00000177, 0.00000756, -0.00001137, 0.00001208, -0.00001349, 0.00000173, -0.00001126, -0.00001037, 0.00000745, -0.00002562, -0.00000462, -0.00002625, -0.00002389, -0.00014233, 0.00010928, 0.00022343, 0.00011961, -0.00014998, -0.00002519, -0.00001060, 0.00002842, -0.00000433, 0.00004388, -0.00000512, -0.00000595, -0.00001541, -0.00001285, -0.00000625, -0.00000736, 0.00000570, -0.00000677, -0.00001719, -0.00000465, 0.00001521, -0.00000644, -0.00000750, 0.00000540, 0.00000728, -0.00000304, },
		{ -0.00000510, 0.00000203, -0.00000014, -0.00000398, 0.00000221, 0.00000703, -0.00000052, -0.00000437, -0.00000118, -0.00000500, 0.00000179, 0.00000334, -0.00000186, 0.00000167, -0.00000277, -0.00000405, -0.00000088, 0.00000532, -0.00000070, 0.00000063, 0.00000095, 0.00000456, -0.00000106, -0.00000703, -0.00000108, 0.00000211, -0.00000247, 0.00000071, 0.00000118, -0.00000107, 0.00000264, 0.00000463, -0.00000284, -0.00000243, 0.00000023, -0.00000230, 0.00000066, -0.00000108, -0.00000047, -0.00000061, 0.00000134, 0.00000084, 0.00000026, -0.00000233, -0.00000004, 0.00000613, -0.00000232, -0.00000035, -0.00000087, -0.00000160, -0.00000112, 0.00000339, -0.00000229, -0.00000509, 0.00000249, 0.00000187, 0.00000331, -0.00000270, 0.00000269, -0.00000013, -0.00000106, -0.00000655, 0.00000009, -0.00000170, 0.00000218, 0.00000635, -0.00000192, -0.00000068, -0.00000012, -0.00000003, -0.00000246, 0.00000876, -0.00000468, -0.00000447, 0.00000082, -0.00000088, 0.00000098, -0.00000818, 0.00000375, 0.00000047, 0.00000813, -0.00001455, 0.00000755, -0.00000043, -0.00000203, -0.00000548, -0.00000190, 0.00001058, -0.00000250, 0.00000753, -0.00001049, 0.00001031, -0.00001003, -0.00000265, -0.00000608, -0.00001489, 0.00001104, -0.00002910, -0.00000043, -0.00003127, -0.00001351, -0.00015076, 0.00008957, 0.00024289, 0.00012706, -0.00015106, -0.00002802, -0.00001167, 0.00002610, -0.00000368, 0.00003999, -0.00000243, -0.00000711, -0.00001074, -0.00001682, -0.00000312, -0.00001042, 0.00000765, -0.00000709, -0.00001596, -0.00000740, 0.00001685, -0.00000613, -0.00000753, 0.00000364, 0.00000871, -0.00000336, -0.00000093, },
		{ 0.00000183, 0.00000004, -0.00000381, 0.00000109, 0.00000776, -0.00000026, -0.00000449, -0.00000125, -0.00000440, 0.00000041, 0.00000451, -0.00000250, 0.00000222, -0.00000281, -0.00000407, -0.00000152, 0.00000566, -0.00000066, 0.00000071, 0.00000063, 0.00000450, -0.00000039, -0.00000711, -0.00000169, 0.00000260, -0.00000251, 0.00000047, 0.00000126, -0.00000097, 0.00000206, 0.00000500, -0.00000239, -0.00000272, 0.00000029, -0.00000232, 0.00000050, -0.00000083, -0.00000063, -0.00000052, 0.00000113, 0.00000093, 0.00000025, -0.00000202, -0.00000080, 0.00000639, -0.00000185, -0.00000070, -0.00000060, -0.00000177, -0.00000115, 0.00000310, -0.00000131, -0.00000585, 0.00000251, 0.00000159, 0.00000368, -0.00000281, 0.00000258, 0.00000002, -0.00000066, -0.00000670, -0.00000033, -0.00000111, 0.00000120, 0.00000672, -0.00000142, -0.00000110, 0.00000003, -0.00000011, -0.00000263, 0.00000852, -0.00000315, -0.00000615, 0.00000200, -0.00000221, 0.00000258, -0.00000924, 0.00000411, -0.00000005, 0.00000896, -0.00001474, 0.00000632, 0.00000097, -0.00000252, -0.00000482, -0.00000333, 0.00001160, -0.00000289, 0.00000703, -0.00000911, 0.00000801, -0.00000613, -0.00000721, -0.00000109, -0.00001890, 0.00001405, -0.00003205, 0.00000312, -0.00003538, -0.00000494, -0.00015512, 0.00006562, 0.00026291, 0.00013569, -0.00015239, -0.00003096, -0.00001250, 0.00002346, -0.00000255, 0.00003573, 0.00000052, -0.00000894, -0.00000558, -0.00002082, -0.00000015, -0.00001317, 0.00000921, -0.00000717, -0.00001478, -0.00000995, 0.00001807, -0.00000548, -0.00000766, 0.00000199, 0.00000984, -0.00000337, -0.00000114, -0.00000483, },
		{ 0.00000018, -0.00000353, -0.00000007, 0.00000837, 0.00000009, -0.00000458, -0.00000142, -0.00000367, -0.00000099, 0.00000557, -0.00000295, 0.00000258, -0.00000268, -0.00000417, -0.00000210, 0.00000589, -0.00000052, 0.00000072, 0.00000041, 0.00000433, 0.00000030, -0.00000709, -0.00000235, 0.00000305, -0.00000247, 0.00000021, 0.00000131, -0.00000083, 0.00000149, 0.00000528, -0.00000186, -0.00000301, 0.00000032, -0.00000230, 0.00000029, -0.00000056, -0.00000080, -0.00000042, 0.00000092, 0.00000102, 0.00000021, -0.00000166, -0.00000151, 0.00000651, -0.00000127, -0.00000108, -0.00000032, -0.00000192, -0.00000115, 0.00000275, -0.00000031, -0.00000649, 0.00000242, 0.00000140, 0.00000393, -0.00000280, 0.00000237, 0.00000023, -0.00000033, -0.00000669, -0.00000091, -0.00000043, 0.00000025, 0.00000693, -0.00000079, -0.00000155, 0.00000018, -0.00000020, -0.00000272, 0.00000810, -0.00000146, -0.00000782, 0.00000306, -0.00000338, 0.00000391, -0.00000994, 0.00000412, -0.00000027, 0.00000946, -0.00001449, 0.00000469, 0.00000255, -0.00000309, -0.00000407, -0.00000474, 0.00001242, -0.00000292, 0.00000611, -0.00000733, 0.00000533, -0.00000200, -0.00001169, 0.00000344, -0.00002214, 0.00001627, -0.00003420, 0.00000574, -0.00003826, 0.00000140, -0.00015519, 0.00003779, 0.00028281, 0.00014601, -0.00015423, -0.00003382, -0.00001323, 0.00002062, -0.00000108, 0.00003123, 0.00000361, -0.00001124, -0.00000017, -0.00002458, 0.00000241, -0.00001543, 0.00001025, -0.00000691, -0.00001378, -0.00001217, 0.00001877, -0.00000444, -0.00000792, 0.00000049, 0.00001065, -0.00000305, -0.00000155, -0.00000447, 0.00000161, },
		{ -0.00000317, -0.00000123, 0.00000884, 0.00000057, -0.00000468, -0.00000165, -0.00000286, -0.00000236, 0.00000643, -0.00000318, 0.00000271, -0.00000237, -0.00000434, -0.00000260, 0.00000599, -0.00000028, 0.00000065, 0.00000028, 0.00000406, 0.00000098, -0.00000697, -0.00000305, 0.00000344, -0.00000235, -0.00000008, 0.00000135, -0.00000066, 0.00000093, 0.00000545, -0.00000126, -0.00000327, 0.00000031, -0.00000224, 0.00000005, -0.00000028, -0.00000096, -0.00000033, 0.00000073, 0.00000109, 0.00000015, -0.00000129, -0.00000215, 0.00000648, -0.00000060, -0.00000148, -0.00000005, -0.00000205, -0.00000114, 0.00000235, 0.00000065, -0.00000700, 0.00000219, 0.00000132, 0.00000405, -0.00000265, 0.00000205, 0.00000050, -0.00000008, -0.00000651, -0.00000163, 0.00000031, -0.00000062, 0.00000697, -0.00000006, -0.00000200, 0.00000030, -0.00000029, -0.00000276, 0.00000754, 0.00000031, -0.00000939, 0.00000393, -0.00000431, 0.00000489, -0.00001021, 0.00000375, -0.00000017, 0.00000959, -0.00001378, 0.00000270, 0.00000426, -0.00000370, -0.00000326, -0.00000610, 0.00001301, -0.00000260, 0.00000481, -0.00000524, 0.00000241, 0.00000216, -0.00001584, 0.00000724, -0.00002439, 0.00001749, -0.00003536, 0.00000719, -0.00003960, 0.00000522, -0.00015096, 0.00000662, 0.00030176, 0.00015858, -0.00015687, -0.00003644, -0.00001400, 0.00001772, 0.00000060, 0.00002664, 0.00000670, -0.00001383, 0.00000519, -0.00002784, 0.00000437, -0.00001703, 0.00001065, -0.00000621, -0.00001306, -0.00001396, 0.00001889, -0.00000299, -0.00000833, -0.00000081, 0.00001108, -0.00000241, -0.00000216, -0.00000403, 0.00000139, 0.00000028, },
		{ -0.00000233, 0.00000911, 0.00000118, -0.00000479, -0.00000192, -0.00000203, -0.00000361, 0.00000705, -0.00000315, 0.00000262, -0.00000187, -0.00000460, -0.00000302, 0.00000595, 0.00000007, 0.00000051, 0.00000026, 0.00000371, 0.00000164, -0.00000677, -0.00000377, 0.00000374, -0.00000213, -0.00000039, 0.00000137, -0.00000047, 0.00000042, 0.00000551, -0.00000060, -0.00000352, 0.00000028, -0.00000215, -0.00000022, -0.00000000, -0.00000110, -0.00000026, 0.00000056, 0.00000116, 0.00000006, -0.00000090, -0.00000272, 0.00000631, 0.00000015, -0.00000186, 0.00000019, -0.00000215, -0.00000113, 0.00000193, 0.00000156, -0.00000734, 0.00000182, 0.00000134, 0.00000404, -0.00000237, 0.00000164, 0.00000083, 0.00000008, -0.00000617, -0.00000248, 0.00000108, -0.00000139, 0.00000684, 0.00000076, -0.00000242, 0.00000039, -0.00000036, -0.00000276, 0.00000686, 0.00000211, -0.00001077, 0.00000454, -0.00000492, 0.00000546, -0.00001004, 0.00000301, 0.00000025, 0.00000937, -0.00001265, 0.00000040, 0.00000603, -0.00000430, -0.00000244, -0.00000737, 0.00001334, -0.00000196, 0.00000320, -0.00000295, -0.00000057, 0.00000613, -0.00001939, 0.00001008, -0.00002547, 0.00001756, -0.00003537, 0.00000730, -0.00003923, 0.00000639, -0.00014258, -0.00002717, 0.00031885, 0.00017397, -0.00016056, -0.00003862, -0.00001498, 0.00001491, 0.00000232, 0.00002214, 0.00000960, -0.00001645, 0.00001021, -0.00003032, 0.00000552, -0.00001782, 0.00001031, -0.00000502, -0.00001268, -0.00001524, 0.00001839, -0.00000114, -0.00000887, -0.00000189, 0.00001114, -0.00000146, -0.00000293, -0.00000353, 0.00000117, 0.00000033, -0.00000275, },
		{ 0.00000915, 0.00000195, -0.00000493, -0.00000220, -0.00000122, -0.00000471, 0.00000737, -0.00000283, 0.00000228, -0.00000120, -0.00000492, -0.00000336, 0.00000577, 0.00000052, 0.00000029, 0.00000032, 0.00000329, 0.00000225, -0.00000647, -0.00000449, 0.00000394, -0.00000182, -0.00000073, 0.00000137, -0.00000027, -0.00000003, 0.00000543, 0.00000011, -0.00000373, 0.00000021, -0.00000202, -0.00000052, 0.00000026, -0.00000121, -0.00000021, 0.00000041, 0.00000122, -0.00000004, -0.00000052, -0.00000320, 0.00000599, 0.00000096, -0.00000222, 0.00000038, -0.00000221, -0.00000113, 0.00000150, 0.00000238, -0.00000749, 0.00000129, 0.00000147, 0.00000390, -0.00000197, 0.00000114, 0.00000119, 0.00000015, -0.00000569, -0.00000342, 0.00000183, -0.00000204, 0.00000653, 0.00000164, -0.00000278, 0.00000041, -0.00000040, -0.00000275, 0.00000612, 0.00000384, -0.00001190, 0.00000481, -0.00000518, 0.00000560, -0.00000942, 0.00000189, 0.00000098, 0.00000880, -0.00001114, -0.00000213, 0.00000780, -0.00000485, -0.00000163, -0.00000851, 0.00001341, -0.00000103, 0.00000139, -0.00000061, -0.00000345, 0.00000967, -0.00002213, 0.00001176, -0.00002525, 0.00001641, -0.00003417, 0.00000600, -0.00003707, 0.00000488, -0.00013040, -0.00006268, 0.00033307, 0.00019278, -0.00016555, -0.00004020, -0.00001631, 0.00001234, 0.00000391, 0.00001791, 0.00001214, -0.00001889, 0.00001463, -0.00003181, 0.00000571, -0.00001772, 0.00000919, -0.00000330, -0.00001269, -0.00001596, 0.00001726, 0.00000108, -0.00000954, -0.00000275, 0.00001083, -0.00000023, -0.00000383, -0.00000300, 0.00000096, 0.00000036, -0.00000231, -0.00000333, },
		{ 0.00000288, -0.00000511, -0.00000245, -0.00000049, -0.00000558, 0.00000737, -0.00000222, 0.00000171, -0.00000038, -0.00000529, -0.00000362, 0.00000546, 0.00000105, 0.00000002, 0.00000045, 0.00000282, 0.00000282, -0.00000610, -0.00000518, 0.00000401, -0.00000141, -0.00000110, 0.00000136, -0.00000008, -0.00000041, 0.00000523, 0.00000088, -0.00000391, 0.00000010, -0.00000186, -0.00000082, 0.00000050, -0.00000128, -0.00000018, 0.00000028, 0.00000127, -0.00000014, -0.00000016, -0.00000357, 0.00000554, 0.00000180, -0.00000252, 0.00000051, -0.00000222, -0.00000115, 0.00000108, 0.00000308, -0.00000743, 0.00000061, 0.00000170, 0.00000364, -0.00000147, 0.00000058, 0.00000159, 0.00000015, -0.00000509, -0.00000442, 0.00000253, -0.00000253, 0.00000607, 0.00000255, -0.00000306, 0.00000035, -0.00000038, -0.00000275, 0.00000535, 0.00000546, -0.00001269, 0.00000470, -0.00000504, 0.00000530, -0.00000837, 0.00000044, 0.00000196, 0.00000794, -0.00000930, -0.00000480, 0.00000947, -0.00000528, -0.00000089, -0.00000949, 0.00001323, 0.00000014, -0.00000055, 0.00000166, -0.00000605, 0.00001261, -0.00002388, 0.00001217, -0.00002366, 0.00001403, -0.00003176, 0.00000329, -0.00003315, 0.00000087, -0.00011495, -0.00009889, 0.00034338, 0.00021554, -0.00017203, -0.00004106, -0.00001813, 0.00001012, 0.00000523, 0.00001413, 0.00001414, -0.00002091, 0.00001818, -0.00003213, 0.00000485, -0.00001667, 0.00000728, -0.00000107, -0.00001311, -0.00001613, 0.00001554, 0.00000361, -0.00001029, -0.00000339, 0.00001017, 0.00000124, -0.00000481, -0.00000246, 0.00000078, 0.00000038, -0.00000188, -0.00000419, 0.00000893, },
		{ -0.00000533, -0.00000267, 0.00000014, -0.00000619, 0.00000702, -0.00000132, 0.00000094, 0.00000056, -0.00000568, -0.00000382, 0.00000503, 0.00000166, -0.00000030, 0.00000064, 0.00000233, 0.00000332, -0.00000565, -0.00000582, 0.00000395, -0.00000089, -0.00000147, 0.00000134, 0.00000009, -0.00000071, 0.00000491, 0.00000167, -0.00000404, -0.00000004, -0.00000168, -0.00000112, 0.00000071, -0.00000131, -0.00000019, 0.00000018, 0.00000131, -0.00000026, 0.00000017, -0.00000382, 0.00000497, 0.00000265, -0.00000275, 0.00000057, -0.00000218, -0.00000119, 0.00000069, 0.00000365, -0.00000718, -0.00000020, 0.00000203, 0.00000328, -0.00000088, -0.00000002, 0.00000201, 0.00000009, -0.00000437, -0.00000543, 0.00000314, -0.00000286, 0.00000545, 0.00000346, -0.00000323, 0.00000019, -0.00000030, -0.00000278, 0.00000459, 0.00000689, -0.00001310, 0.00000419, -0.00000452, 0.00000457, -0.00000693, -0.00000128, 0.00000315, 0.00000683, -0.00000722, -0.00000751, 0.00001096, -0.00000554, -0.00000025, -0.00001030, 0.00001279, 0.00000149, -0.00000250, 0.00000372, -0.00000823, 0.00001478, -0.00002449, 0.00001124, -0.00002073, 0.00001048, -0.00002821, -0.00000074, -0.00002763, -0.00000536, -0.00009692, -0.00013465, 0.00034877, 0.00024270, -0.00018012, -0.00004113, -0.00002051, 0.00000837, 0.00000614, 0.00001093, 0.00001545, -0.00002231, 0.00002067, -0.00003116, 0.00000290, -0.00001469, 0.00000463, 0.00000162, -0.00001388, -0.00001577, 0.00001330, 0.00000636, -0.00001107, -0.00000385, 0.00000920, 0.00000288, -0.00000582, -0.00000196, 0.00000062, 0.00000038, -0.00000149, -0.00000486, 0.00000843, 0.00000397, },
		{ -0.00000284, 0.00000063, -0.00000653, 0.00000634, -0.00000018, 0.00000001, 0.00000157, -0.00000606, -0.00000396, 0.00000450, 0.00000232, -0.00000063, 0.00000087, 0.00000183, 0.00000375, -0.00000514, -0.00000639, 0.00000375, -0.00000030, -0.00000185, 0.00000132, 0.00000024, -0.00000091, 0.00000446, 0.00000248, -0.00000410, -0.00000022, -0.00000147, -0.00000142, 0.00000087, -0.00000130, -0.00000022, 0.00000010, 0.00000133, -0.00000037, 0.00000046, -0.00000396, 0.00000428, 0.00000348, -0.00000289, 0.00000056, -0.00000207, -0.00000125, 0.00000035, 0.00000408, -0.00000671, -0.00000114, 0.00000242, 0.00000284, -0.00000024, -0.00000064, 0.00000241, -0.00000001, -0.00000359, -0.00000643, 0.00000362, -0.00000301, 0.00000470, 0.00000433, -0.00000326, -0.00000008, -0.00000015, -0.00000284, 0.00000388, 0.00000809, -0.00001310, 0.00000327, -0.00000362, 0.00000347, -0.00000517, -0.00000319, 0.00000448, 0.00000557, -0.00000500, -0.00001016, 0.00001219, -0.00000560, 0.00000027, -0.00001092, 0.00001213, 0.00000297, -0.00000437, 0.00000548, -0.00000986, 0.00001606, -0.00002392, 0.00000898, -0.00001656, 0.00000592, -0.00002369, -0.00000588, -0.00002077, -0.00001336, -0.00007715, -0.00016876, 0.00034823, 0.00027462, -0.00018981, -0.00004043, -0.00002347, 0.00000712, 0.00000655, 0.00000845, 0.00001596, -0.00002295, 0.00002198, -0.00002888, -0.00000011, -0.00001185, 0.00000134, 0.00000469, -0.00001496, -0.00001495, 0.00001064, 0.00000922, -0.00001180, -0.00000416, 0.00000798, 0.00000464, -0.00000680, -0.00000153, 0.00000051, 0.00000039, -0.00000116, -0.00000532, 0.00000764, 0.00000520, -0.00000558, },
		{ 0.00000097, -0.00000658, 0.00000534, 0.00000117, -0.00000102, 0.00000259, -0.00000640, -0.00000408, 0.00000389, 0.00000300, -0.00000097, 0.00000110, 0.00000135, 0.00000409, -0.00000458, -0.00000687, 0.00000341, 0.00000037, -0.00000222, 0.00000128, 0.00000036, -0.00000102, 0.00000392, 0.00000329, -0.00000409, -0.00000045, -0.00000126, -0.00000169, 0.00000099, -0.00000123, -0.00000028, 0.00000004, 0.00000135, -0.00000046, 0.00000069, -0.00000399, 0.00000350, 0.00000427, -0.00000293, 0.00000047, -0.00000192, -0.00000134, 0.00000005, 0.00000436, -0.00000606, -0.00000218, 0.00000285, 0.00000235, 0.00000042, -0.00000124, 0.00000279, -0.00000014, -0.00000276, -0.00000735, 0.00000395, -0.00000297, 0.00000383, 0.00000514, -0.00000315, -0.00000045, 0.00000007, -0.00000293, 0.00000323, 0.00000904, -0.00001269, 0.00000198, -0.00000239, 0.00000206, -0.00000318, -0.00000521, 0.00000584, 0.00000422, -0.00000273, -0.00001263, 0.00001308, -0.00000540, 0.00000065, -0.00001133, 0.00001126, 0.00000450, -0.00000609, 0.00000683, -0.00001085, 0.00001640, -0.00002215, 0.00000550, -0.00001133, 0.00000057, -0.00001846, -0.00001184, -0.00001295, -0.00002257, -0.00005656, -0.00020000, 0.00034089, 0.00031148, -0.00020099, -0.00003904, -0.00002698, 0.00000639, 0.00000641, 0.00000672, 0.00001560, -0.00002274, 0.00002204, -0.00002532, -0.00000407, -0.00000830, -0.00000244, 0.00000799, -0.00001624, -0.00001377, 0.00000769, 0.00001206, -0.00001239, -0.00000439, 0.00000656, 0.00000642, -0.00000770, -0.00000118, 0.00000044, 0.00000040, -0.00000091, -0.00000555, 0.00000658, 0.00000654, -0.00000584, -0.00000295, },
		{ -0.00000637, 0.00000407, 0.00000265, -0.00000208, 0.00000357, -0.00000666, -0.00000420, 0.00000323, 0.00000367, -0.00000127, 0.00000133, 0.00000091, 0.00000434, -0.00000398, -0.00000725, 0.00000293, 0.00000109, -0.00000257, 0.00000124, 0.00000044, -0.00000104, 0.00000328, 0.00000406, -0.00000400, -0.00000071, -0.00000103, -0.00000195, 0.00000105, -0.00000113, -0.00000036, 0.00000000, 0.00000136, -0.00000055, 0.00000087, -0.00000390, 0.00000266, 0.00000499, -0.00000285, 0.00000029, -0.00000171, -0.00000145, -0.00000020, 0.00000448, -0.00000524, -0.00000327, 0.00000329, 0.00000184, 0.00000106, -0.00000179, 0.00000311, -0.00000027, -0.00000193, -0.00000817, 0.00000409, -0.00000275, 0.00000288, 0.00000586, -0.00000288, -0.00000092, 0.00000034, -0.00000307, 0.00000267, 0.00000971, -0.00001188, 0.00000036, -0.00000091, 0.00000042, -0.00000107, -0.00000721, 0.00000714, 0.00000290, -0.00000052, -0.00001481, 0.00001355, -0.00000492, 0.00000085, -0.00001153, 0.00001020, 0.00000605, -0.00000759, 0.00000772, -0.00001118, 0.00001578, -0.00001927, 0.00000096, -0.00000530, -0.00000527, -0.00001282, -0.00001828, -0.00000461, -0.00003237, -0.00003613, -0.00022717, 0.00032601, 0.00035329, -0.00021338, -0.00003718, -0.00003092, 0.00000611, 0.00000575, 0.00000578, 0.00001436, -0.00002167, 0.00002088, -0.00002060, -0.00000878, -0.00000423, -0.00000652, 0.00001136, -0.00001757, -0.00001237, 0.00000461, 0.00001473, -0.00001276, -0.00000459, 0.00000504, 0.00000816, -0.00000845, -0.00000097, 0.00000043, 0.00000043, -0.00000076, -0.00000554, 0.00000527, 0.00000795, -0.00000610, -0.00000302, 0.00000116, },
		{ 0.00000259, 0.00000420, -0.00000310, 0.00000444, -0.00000679, -0.00000433, 0.00000256, 0.00000429, -0.00000153, 0.00000152, 0.00000052, 0.00000450, -0.00000335, -0.00000753, 0.00000232, 0.00000184, -0.00000288, 0.00000117, 0.00000049, -0.00000098, 0.00000258, 0.00000477, -0.00000381, -0.00000101, -0.00000080, -0.00000217, 0.00000105, -0.00000097, -0.00000047, -0.00000003, 0.00000135, -0.00000061, 0.00000097, -0.00000371, 0.00000177, 0.00000562, -0.00000266, 0.00000005, -0.00000147, -0.00000158, -0.00000040, 0.00000446, -0.00000428, -0.00000438, 0.00000369, 0.00000136, 0.00000165, -0.00000227, 0.00000336, -0.00000038, -0.00000112, -0.00000885, 0.00000403, -0.00000236, 0.00000187, 0.00000647, -0.00000246, -0.00000147, 0.00000065, -0.00000323, 0.00000218, 0.00001010, -0.00001070, -0.00000151, 0.00000073, -0.00000133, 0.00000105, -0.00000910, 0.00000829, 0.00000169, 0.00000150, -0.00001659, 0.00001353, -0.00000413, 0.00000086, -0.00001150, 0.00000897, 0.00000757, -0.00000881, 0.00000812, -0.00001082, 0.00001427, -0.00001542, -0.00000439, 0.00000121, -0.00001124, -0.00000715, -0.00002477, 0.00000373, -0.00004204, -0.00001689, -0.00024918, 0.00030304, 0.00039985, -0.00022654, -0.00003515, -0.00003509, 0.00000615, 0.00000465, 0.00000554, 0.00001231, -0.00001978, 0.00001861, -0.00001494, -0.00001401, 0.00000012, -0.00001066, 0.00001459, -0.00001880, -0.00001091, 0.00000157, 0.00001708, -0.00001281, -0.00000483, 0.00000348, 0.00000977, -0.00000899, -0.00000092, 0.00000049, 0.00000046, -0.00000071, -0.00000532, 0.00000376, 0.00000938, -0.00000630, -0.00000305, 0.00000122, -0.00000592, },
		{ 0.00000572, -0.00000400, 0.00000515, -0.00000676, -0.00000452, 0.00000189, 0.00000484, -0.00000170, 0.00000165, 0.00000021, 0.00000455, -0.00000271, -0.00000770, 0.00000160, 0.00000259, -0.00000312, 0.00000107, 0.00000052, -0.00000085, 0.00000185, 0.00000539, -0.00000351, -0.00000136, -0.00000056, -0.00000235, 0.00000100, -0.00000078, -0.00000058, -0.00000004, 0.00000134, -0.00000065, 0.00000101, -0.00000341, 0.00000085, 0.00000614, -0.00000234, -0.00000025, -0.00000120, -0.00000171, -0.00000055, 0.00000431, -0.00000321, -0.00000547, 0.00000402, 0.00000094, 0.00000216, -0.00000263, 0.00000351, -0.00000044, -0.00000039, -0.00000934, 0.00000375, -0.00000179, 0.00000082, 0.00000695, -0.00000189, -0.00000207, 0.00000097, -0.00000340, 0.00000177, 0.00001024, -0.00000921, -0.00000355, 0.00000242, -0.00000307, 0.00000306, -0.00001074, 0.00000919, 0.00000070, 0.00000324, -0.00001789, 0.00001298, -0.00000301, 0.00000068, -0.00001125, 0.00000758, 0.00000904, -0.00000975, 0.00000805, -0.00000985, 0.00001197, -0.00001082, -0.00001023, 0.00000783, -0.00001695, -0.00000186, -0.00003086, 0.00001152, -0.00005087, 0.00000021, -0.00026505, 0.00027168, 0.00045072, -0.00023988, -0.00003338, -0.00003920, 0.00000631, 0.00000323, 0.00000590, 0.00000959, -0.00001722, 0.00001542, -0.00000860, -0.00001945, 0.00000445, -0.00001460, 0.00001748, -0.00001977, -0.00000955, -0.00000125, 0.00001897, -0.00001245, -0.00000518, 0.00000196, 0.00001115, -0.00000927, -0.00000105, 0.00000064, 0.00000050, -0.00000075, -0.00000490, 0.00000210, 0.00001075, -0.00000641, -0.00000309, 0.00000118, -0.00000528, 0.00000097, },
		{ -0.00000471, 0.00000564, -0.00000654, -0.00000477, 0.00000127, 0.00000528, -0.00000176, 0.00000171, -0.00000000, 0.00000450, -0.00000205, -0.00000776, 0.00000081, 0.00000331, -0.00000328, 0.00000093, 0.00000053, -0.00000066, 0.00000111, 0.00000591, -0.00000310, -0.00000174, -0.00000034, -0.00000249, 0.00000090, -0.00000056, -0.00000071, -0.00000005, 0.00000131, -0.00000065, 0.00000098, -0.00000304, -0.00000006, 0.00000654, -0.00000190, -0.00000060, -0.00000093, -0.00000184, -0.00000066, 0.00000406, -0.00000208, -0.00000648, 0.00000423, 0.00000062, 0.00000255, -0.00000286, 0.00000353, -0.00000042, 0.00000025, -0.00000962, 0.00000325, -0.00000107, -0.00000024, 0.00000730, -0.00000120, -0.00000270, 0.00000127, -0.00000354, 0.00000140, 0.00001016, -0.00000748, -0.00000565, 0.00000406, -0.00000469, 0.00000482, -0.00001204, 0.00000973, 0.00000001, 0.00000459, -0.00001861, 0.00001187, -0.00000157, 0.00000029, -0.00001077, 0.00000605, 0.00001042, -0.00001040, 0.00000753, -0.00000834, 0.00000906, -0.00000573, -0.00001622, 0.00001415, -0.00002199, 0.00000265, -0.00003609, 0.00001821, -0.00005815, 0.00001425, -0.00027401, 0.00023187, 0.00050520, -0.00025264, -0.00003238, -0.00004290, 0.00000631, 0.00000172, 0.00000666, 0.00000637, -0.00001419, 0.00001159, -0.00000195, -0.00002475, 0.00000845, -0.00001807, 0.00001980, -0.00002028, -0.00000846, -0.00000368, 0.00002025, -0.00001162, -0.00000568, 0.00000058, 0.00001225, -0.00000923, -0.00000140, 0.00000089, 0.00000052, -0.00000085, -0.00000433, 0.00000036, 0.00001201, -0.00000637, -0.00000317, 0.00000107, -0.00000451, -0.00000070, 0.00000713, },
		{ 0.00000586, -0.00000609, -0.00000511, 0.00000072, 0.00000557, -0.00000170, 0.00000167, -0.00000012, 0.00000434, -0.00000139, -0.00000771, -0.00000005, 0.00000396, -0.00000334, 0.00000075, 0.00000053, -0.00000045, 0.00000039, 0.00000629, -0.00000257, -0.00000215, -0.00000012, -0.00000260, 0.00000074, -0.00000032, -0.00000083, -0.00000006, 0.00000127, -0.00000063, 0.00000088, -0.00000259, -0.00000095, 0.00000679, -0.00000135, -0.00000098, -0.00000066, -0.00000195, -0.00000075, 0.00000371, -0.00000093, -0.00000736, 0.00000430, 0.00000042, 0.00000279, -0.00000293, 0.00000342, -0.00000032, 0.00000077, -0.00000967, 0.00000252, -0.00000021, -0.00000130, 0.00000751, -0.00000040, -0.00000332, 0.00000152, -0.00000363, 0.00000106, 0.00000991, -0.00000562, -0.00000769, 0.00000551, -0.00000605, 0.00000622, -0.00001289, 0.00000984, -0.00000029, 0.00000549, -0.00001870, 0.00001019, 0.00000018, -0.00000029, -0.00001006, 0.00000438, 0.00001173, -0.00001080, 0.00000665, -0.00000643, 0.00000575, -0.00000045, -0.00002197, 0.00001975, -0.00002594, 0.00000597, -0.00004003, 0.00002329, -0.00006324, 0.00002447, -0.00027549, 0.00018387, 0.00056237, -0.00026389, -0.00003276, -0.00004578, 0.00000585, 0.00000037, 0.00000758, 0.00000293, -0.00001095, 0.00000744, 0.00000465, -0.00002956, 0.00001180, -0.00002079, 0.00002133, -0.00002017, -0.00000781, -0.00000555, 0.00002082, -0.00001023, -0.00000641, -0.00000060, 0.00001298, -0.00000884, -0.00000198, 0.00000126, 0.00000050, -0.00000098, -0.00000366, -0.00000136, 0.00001305, -0.00000613, -0.00000332, 0.00000094, -0.00000368, -0.00000232, 0.00000833, -0.00000514, },
		{ -0.00000541, -0.00000556, 0.00000027, 0.00000571, -0.00000149, 0.00000151, -0.00000011, 0.00000408, -0.00000073, -0.00000758, -0.00000092, 0.00000451, -0.00000328, 0.00000050, 0.00000054, -0.00000023, -0.00000028, 0.00000652, -0.00000191, -0.00000258, 0.00000007, -0.00000266, 0.00000055, -0.00000007, -0.00000094, -0.00000007, 0.00000123, -0.00000057, 0.00000072, -0.00000210, -0.00000179, 0.00000690, -0.00000071, -0.00000136, -0.00000042, -0.00000202, -0.00000082, 0.00000331, 0.00000020, -0.00000808, 0.00000418, 0.00000038, 0.00000286, -0.00000283, 0.00000316, -0.00000011, 0.00000113, -0.00000948, 0.00000157, 0.00000076, -0.00000233, 0.00000758, 0.00000045, -0.00000386, 0.00000166, -0.00000364, 0.00000072, 0.00000954, -0.00000371, -0.00000954, 0.00000666, -0.00000706, 0.00000716, -0.00001322, 0.00000946, -0.00000015, 0.00000585, -0.00001812, 0.00000793, 0.00000223, -0.00000107, -0.00000911, 0.00000258, 0.00001297, -0.00001098, 0.00000552, -0.00000430, 0.00000228, 0.00000466, -0.00002708, 0.00002423, -0.00002842, 0.00000772, -0.00004227, 0.00002628, -0.00006558, 0.00003024, -0.00026919, 0.00012821, 0.00062104, -0.00027260, -0.00003521, -0.00004737, 0.00000457, -0.00000054, 0.00000837, -0.00000044, -0.00000783, 0.00000334, 0.00001078, -0.00003349, 0.00001421, -0.00002253, 0.00002188, -0.00001928, -0.00000775, -0.00000672, 0.00002057, -0.00000827, -0.00000738, -0.00000151, 0.00001328, -0.00000806, -0.00000282, 0.00000176, 0.00000043, -0.00000111, -0.00000296, -0.00000299, 0.00001381, -0.00000563, -0.00000359, 0.00000084, -0.00000286, -0.00000382, 0.00000924, -0.00000525, 0.00000576, },
		{ -0.00000612, -0.00000006, 0.00000565, -0.00000112, 0.00000124, 0.00000001, 0.00000372, -0.00000007, -0.00000739, -0.00000177, 0.00000492, -0.00000307, 0.00000019, 0.00000056, -0.00000003, -0.00000086, 0.00000657, -0.00000113, -0.00000304, 0.00000025, -0.00000267, 0.00000032, 0.00000019, -0.00000104, -0.00000009, 0.00000117, -0.00000048, 0.00000050, -0.00000157, -0.00000256, 0.00000687, 0.00000001, -0.00000172, -0.00000023, -0.00000204, -0.00000091, 0.00000289, 0.00000125, -0.00000859, 0.00000385, 0.00000051, 0.00000275, -0.00000255, 0.00000275, 0.00000021, 0.00000133, -0.00000903, 0.00000042, 0.00000181, -0.00000331, 0.00000754, 0.00000133, -0.00000430, 0.00000166, -0.00000352, 0.00000032, 0.00000912, -0.00000186, -0.00001110, 0.00000740, -0.00000760, 0.00000756, -0.00001295, 0.00000854, 0.00000048, 0.00000565, -0.00001687, 0.00000514, 0.00000453, -0.00000203, -0.00000793, 0.00000064, 0.00001418, -0.00001103, 0.00000426, -0.00000213, -0.00000108, 0.00000929, -0.00003118, 0.00002721, -0.00002909, 0.00000760, -0.00004248, 0.00002683, -0.00006474, 0.00003114, -0.00025508, 0.00006576, 0.00067982, -0.00027760, -0.00004043, -0.00004719, 0.00000210, -0.00000066, 0.00000873, -0.00000340, -0.00000517, -0.00000032, 0.00001603, -0.00003619, 0.00001537, -0.00002305, 0.00002130, -0.00001751, -0.00000839, -0.00000709, 0.00001945, -0.00000570, -0.00000863, -0.00000211, 0.00001312, -0.00000686, -0.00000391, 0.00000240, 0.00000027, -0.00000119, -0.00000231, -0.00000442, 0.00001420, -0.00000483, -0.00000402, 0.00000084, -0.00000212, -0.00000510, 0.00000977, -0.00000497, 0.00000533, -0.00000448, },
		{ -0.00000026, 0.00000540, -0.00000058, 0.00000084, 0.00000025, 0.00000326, 0.00000059, -0.00000714, -0.00000257, 0.00000517, -0.00000271, -0.00000019, 0.00000062, 0.00000015, -0.00000134, 0.00000643, -0.00000024, -0.00000350, 0.00000040, -0.00000265, 0.00000006, 0.00000043, -0.00000111, -0.00000012, 0.00000111, -0.00000036, 0.00000024, -0.00000102, -0.00000325, 0.00000668, 0.00000079, -0.00000203, -0.00000010, -0.00000200, -0.00000100, 0.00000246, 0.00000220, -0.00000886, 0.00000329, 0.00000082, 0.00000245, -0.00000208, 0.00000219, 0.00000065, 0.00000135, -0.00000833, -0.00000091, 0.00000292, -0.00000423, 0.00000739, 0.00000218, -0.00000457, 0.00000147, -0.00000324, -0.00000016, 0.00000871, -0.00000017, -0.00001226, 0.00000764, -0.00000760, 0.00000735, -0.00001205, 0.00000706, 0.00000161, 0.00000487, -0.00001494, 0.00000186, 0.00000704, -0.00000314, -0.00000652, -0.00000144, 0.00001537, -0.00001103, 0.00000302, -0.00000012, -0.00000406, 0.00001308, -0.00003391, 0.00002838, -0.00002771, 0.00000538, -0.00004044, 0.00002467, -0.00006045, 0.00002696, -0.00023345, -0.00000234, 0.00073715, -0.00027770, -0.00004913, -0.00004476, -0.00000194, 0.00000030, 0.00000831, -0.00000562, -0.00000332, -0.00000315, 0.00002002, -0.00003736, 0.00001507, -0.00002219, 0.00001948, -0.00001477, -0.00000982, -0.00000657, 0.00001743, -0.00000255, -0.00001016, -0.00000235, 0.00001246, -0.00000525, -0.00000527, 0.00000321, -0.00000001, -0.00000117, -0.00000176, -0.00000558, 0.00001415, -0.00000369, -0.00000465, 0.00000096, -0.00000155, -0.00000607, 0.00000988, -0.00000428, 0.00000456, -0.00000330, -0.00000679, },
		{ 0.00000494, 0.00000012, 0.00000032, 0.00000062, 0.00000271, 0.00000125, -0.00000686, -0.00000328, 0.00000522, -0.00000218, -0.00000065, 0.00000072, 0.00000026, -0.00000168, 0.00000611, 0.00000076, -0.00000395, 0.00000052, -0.00000259, -0.00000021, 0.00000065, -0.00000115, -0.00000016, 0.00000104, -0.00000021, -0.00000006, -0.00000048, -0.00000383, 0.00000636, 0.00000158, -0.00000228, -0.00000007, -0.00000189, -0.00000113, 0.00000206, 0.00000299, -0.00000886, 0.00000249, 0.00000131, 0.00000196, -0.00000144, 0.00000148, 0.00000121, 0.00000119, -0.00000739, -0.00000239, 0.00000405, -0.00000508, 0.00000716, 0.00000296, -0.00000463, 0.00000107, -0.00000277, -0.00000077, 0.00000837, 0.00000127, -0.00001291, 0.00000730, -0.00000701, 0.00000650, -0.00001053, 0.00000504, 0.00000323, 0.00000352, -0.00001239, -0.00000184, 0.00000970, -0.00000441, -0.00000487, -0.00000367, 0.00001659, -0.00001107, 0.00000194, 0.00000153, -0.00000640, 0.00001574, -0.00003499, 0.00002750, -0.00002408, 0.00000094, -0.00003601, 0.00001967, -0.00005260, 0.00001774, -0.00020485, -0.00007469, 0.00079130, -0.00027168, -0.00006200, -0.00003966, -0.00000790, 0.00000267, 0.00000683, -0.00000676, -0.00000262, -0.00000478, 0.00002240, -0.00003673, 0.00001313, -0.00001985, 0.00001635, -0.00001102, -0.00001206, -0.00000515, 0.00001452, 0.00000115, -0.00001195, -0.00000222, 0.00001128, -0.00000322, -0.00000688, 0.00000419, -0.00000044, -0.00000101, -0.00000141, -0.00000637, 0.00001361, -0.00000217, -0.00000549, 0.00000127, -0.00000118, -0.00000669, 0.00000951, -0.00000316, 0.00000344, -0.00000191, -0.00000756, -0.00000031, },
		{ 0.00000097, -0.00000031, 0.00000109, 0.00000208, 0.00000191, -0.00000658, -0.00000386, 0.00000505, -0.00000149, -0.00000118, 0.00000087, 0.00000030, -0.00000188, 0.00000559, 0.00000183, -0.00000437, 0.00000060, -0.00000250, -0.00000048, 0.00000084, -0.00000114, -0.00000021, 0.00000096, -0.00000003, -0.00000038, 0.00000005, -0.00000431, 0.00000590, 0.00000239, -0.00000243, -0.00000012, -0.00000170, -0.00000131, 0.00000172, 0.00000360, -0.00000857, 0.00000147, 0.00000198, 0.00000130, -0.00000065, 0.00000066, 0.00000187, 0.00000086, -0.00000623, -0.00000399, 0.00000519, -0.00000585, 0.00000688, 0.00000363, -0.00000445, 0.00000042, -0.00000209, -0.00000153, 0.00000817, 0.00000237, -0.00001299, 0.00000635, -0.00000581, 0.00000502, -0.00000840, 0.00000250, 0.00000531, 0.00000163, -0.00000927, -0.00000587, 0.00001243, -0.00000578, -0.00000300, -0.00000605, 0.00001789, -0.00001125, 0.00000117, 0.00000262, -0.00000787, 0.00001702, -0.00003420, 0.00002441, -0.00001815, -0.00000574, -0.00002921, 0.00001185, -0.00004127, 0.00000373, -0.00017012, -0.00014965, 0.00084047, -0.00025833, -0.00007964, -0.00003150, -0.00001605, 0.00000670, 0.00000399, -0.00000653, -0.00000337, -0.00000491, 0.00002289, -0.00003413, 0.00000946, -0.00001597, 0.00001193, -0.00000629, -0.00001512, -0.00000284, 0.00001078, 0.00000531, -0.00001397, -0.00000172, 0.00000960, -0.00000081, -0.00000873, 0.00000534, -0.00000104, -0.00000065, -0.00000131, -0.00000673, 0.00001253, -0.00000027, -0.00000658, 0.00000178, -0.00000109, -0.00000689, 0.00000864, -0.00000162, 0.00000202, -0.00000032, -0.00000843, -0.00000022, 0.00000427, },
		{ -0.00000105, 0.00000166, 0.00000138, 0.00000258, -0.00000631, -0.00000428, 0.00000465, -0.00000063, -0.00000179, 0.00000107, 0.00000026, -0.00000192, 0.00000489, 0.00000297, -0.00000476, 0.00000063, -0.00000238, -0.00000075, 0.00000098, -0.00000110, -0.00000028, 0.00000088, 0.00000016, -0.00000072, 0.00000053, -0.00000467, 0.00000534, 0.00000316, -0.00000247, -0.00000029, -0.00000142, -0.00000153, 0.00000144, 0.00000400, -0.00000799, 0.00000022, 0.00000280, 0.00000048, 0.00000028, -0.00000028, 0.00000263, 0.00000037, -0.00000488, -0.00000566, 0.00000628, -0.00000654, 0.00000657, 0.00000413, -0.00000398, -0.00000050, -0.00000118, -0.00000245, 0.00000815, 0.00000307, -0.00001246, 0.00000477, -0.00000401, 0.00000294, -0.00000571, -0.00000048, 0.00000779, -0.00000074, -0.00000568, -0.00001012, 0.00001516, -0.00000724, -0.00000092, -0.00000859, 0.00001929, -0.00001164, 0.00000084, 0.00000299, -0.00000828, 0.00001673, -0.00003139, 0.00001909, -0.00000998, -0.00001456, -0.00002016, 0.00000139, -0.00002670, -0.00001457, -0.00013034, -0.00022540, 0.00088283, -0.00023658, -0.00010253, -0.00002002, -0.00002660, 0.00001259, -0.00000041, -0.00000469, -0.00000582, -0.00000330, 0.00002131, -0.00002949, 0.00000406, -0.00001062, 0.00000628, -0.00000066, -0.00001893, 0.00000030, 0.00000632, 0.00000984, -0.00001617, -0.00000085, 0.00000743, 0.00000195, -0.00001080, 0.00000665, -0.00000184, -0.00000006, -0.00000152, -0.00000660, 0.00001090, 0.00000201, -0.00000790, 0.00000253, -0.00000129, -0.00000665, 0.00000727, 0.00000031, 0.00000032, 0.00000142, -0.00000936, 0.00000000, 0.00000341, 0.00000196, },
		{ 0.00000232, 0.00000061, 0.00000324, -0.00000608, -0.00000453, 0.00000401, 0.00000038, -0.00000246, 0.00000133, 0.00000013, -0.00000181, 0.00000403, 0.00000415, -0.00000509, 0.00000062, -0.00000224, -0.00000099, 0.00000108, -0.00000102, -0.00000037, 0.00000079, 0.00000037, -0.00000105, 0.00000096, -0.00000490, 0.00000468, 0.00000388, -0.00000238, -0.00000057, -0.00000105, -0.00000180, 0.00000126, 0.00000419, -0.00000714, -0.00000121, 0.00000374, -0.00000046, 0.00000131, -0.00000129, 0.00000346, -0.00000026, -0.00000337, -0.00000736, 0.00000732, -0.00000714, 0.00000625, 0.00000444, -0.00000321, -0.00000168, -0.00000006, -0.00000356, 0.00000833, 0.00000333, -0.00001130, 0.00000259, -0.00000165, 0.00000032, -0.00000256, -0.00000382, 0.00001059, -0.00000349, -0.00000173, -0.00001447, 0.00001782, -0.00000875, 0.00000135, -0.00001128, 0.00002083, -0.00001232, 0.00000105, 0.00000253, -0.00000749, 0.00001476, -0.00002656, 0.00001160, 0.00000028, -0.00002527, -0.00000911, -0.00001141, -0.00000933, -0.00003647, -0.00008682, -0.00030001, 0.00091661, -0.00020548, -0.00013099, -0.00000510, -0.00003966, 0.00002044, -0.00000652, -0.00000108, -0.00001012, 0.00000022, 0.00001758, -0.00002282, -0.00000299, -0.00000391, -0.00000045, 0.00000574, -0.00002339, 0.00000416, 0.00000125, 0.00001461, -0.00001847, 0.00000034, 0.00000482, 0.00000502, -0.00001304, 0.00000813, -0.00000283, 0.00000079, -0.00000207, -0.00000596, 0.00000872, 0.00000464, -0.00000943, 0.00000350, -0.00000180, -0.00000597, 0.00000543, 0.00000259, -0.00000159, 0.00000326, -0.00001033, 0.00000035, 0.00000237, 0.00000307, -0.00000187, },
		{ -0.00000020, 0.00000391, -0.00000591, -0.00000458, 0.00000314, 0.00000151, -0.00000317, 0.00000163, -0.00000010, -0.00000155, 0.00000302, 0.00000533, -0.00000534, 0.00000057, -0.00000208, -0.00000122, 0.00000112, -0.00000089, -0.00000046, 0.00000069, 0.00000059, -0.00000137, 0.00000133, -0.00000502, 0.00000395, 0.00000452, -0.00000214, -0.00000096, -0.00000061, -0.00000213, 0.00000117, 0.00000416, -0.00000602, -0.00000279, 0.00000476, -0.00000148, 0.00000241, -0.00000234, 0.00000433, -0.00000101, -0.00000175, -0.00000906, 0.00000826, -0.00000766, 0.00000596, 0.00000453, -0.00000213, -0.00000312, 0.00000128, -0.00000484, 0.00000874, 0.00000313, -0.00000952, -0.00000016, 0.00000119, -0.00000275, 0.00000096, -0.00000741, 0.00001361, -0.00000653, 0.00000247, -0.00001880, 0.00002032, -0.00001027, 0.00000380, -0.00001410, 0.00002253, -0.00001334, 0.00000188, 0.00000115, -0.00000544, 0.00001110, -0.00001976, 0.00000213, 0.00001233, -0.00003755, 0.00000354, -0.00002609, 0.00001025, -0.00006108, -0.00004101, -0.00037149, 0.00094012, -0.00016430, -0.00016514, 0.00001326, -0.00005521, 0.00003026, -0.00001439, 0.00000438, -0.00001634, 0.00000567, 0.00001173, -0.00001424, -0.00001150, 0.00000396, -0.00000806, 0.00001273, -0.00002833, 0.00000859, -0.00000425, 0.00001948, -0.00002079, 0.00000180, 0.00000183, 0.00000832, -0.00001540, 0.00000975, -0.00000403, 0.00000191, -0.00000299, -0.00000479, 0.00000601, 0.00000758, -0.00001115, 0.00000469, -0.00000261, -0.00000486, 0.00000316, 0.00000514, -0.00000364, 0.00000515, -0.00001129, 0.00000081, 0.00000117, 0.00000427, -0.00000275, 0.00000305, },
		{ 0.00000458, -0.00000580, -0.00000443, 0.00000206, 0.00000276, -0.00000390, 0.00000198, -0.00000041, -0.00000114, 0.00000189, 0.00000648, -0.00000551, 0.00000046, -0.00000191, -0.00000140, 0.00000110, -0.00000071, -0.00000057, 0.00000060, 0.00000080, -0.00000166, 0.00000163, -0.00000502, 0.00000317, 0.00000505, -0.00000176, -0.00000145, -0.00000010, -0.00000250, 0.00000117, 0.00000391, -0.00000467, -0.00000448, 0.00000582, -0.00000254, 0.00000353, -0.00000339, 0.00000522, -0.00000183, -0.00000006, -0.00001068, 0.00000908, -0.00000810, 0.00000572, 0.00000438, -0.00000075, -0.00000479, 0.00000280, -0.00000628, 0.00000938, 0.00000247, -0.00000716, -0.00000337, 0.00000440, -0.00000613, 0.00000470, -0.00001112, 0.00001674, -0.00000974, 0.00000677, -0.00002298, 0.00002257, -0.00001177, 0.00000637, -0.00001703, 0.00002437, -0.00001474, 0.00000336, -0.00000118, -0.00000213, 0.00000580, -0.00001118, -0.00000901, 0.00002579, -0.00005095, 0.00001732, -0.00004208, 0.00003131, -0.00008739, 0.00000552, -0.00043787, 0.00095187, -0.00011256, -0.00020487, 0.00003484, -0.00007308, 0.00004194, -0.00002396, 0.00001165, -0.00002444, 0.00001301, 0.00000389, -0.00000400, -0.00002120, 0.00001272, -0.00001627, 0.00002006, -0.00003357, 0.00001340, -0.00000999, 0.00002429, -0.00002304, 0.00000349, -0.00000147, 0.00001177, -0.00001784, 0.00001147, -0.00000542, 0.00000330, -0.00000428, -0.00000312, 0.00000284, 0.00001075, -0.00001301, 0.00000607, -0.00000371, -0.00000337, 0.00000054, 0.00000787, -0.00000574, 0.00000701, -0.00001222, 0.00000134, -0.00000015, 0.00000553, -0.00000366, 0.00000382, -0.00000104, },
		{ -0.00000576, -0.00000408, 0.00000078, 0.00000406, -0.00000464, 0.00000235, -0.00000080, -0.00000061, 0.00000069, 0.00000756, -0.00000556, 0.00000030, -0.00000174, -0.00000155, 0.00000102, -0.00000050, -0.00000069, 0.00000049, 0.00000101, -0.00000192, 0.00000184, -0.00000491, 0.00000236, 0.00000546, -0.00000124, -0.00000203, 0.00000046, -0.00000290, 0.00000126, 0.00000346, -0.00000314, -0.00000623, 0.00000687, -0.00000360, 0.00000463, -0.00000442, 0.00000609, -0.00000270, 0.00000166, -0.00001220, 0.00000975, -0.00000845, 0.00000554, 0.00000398, 0.00000091, -0.00000665, 0.00000445, -0.00000783, 0.00001021, 0.00000139, -0.00000431, -0.00000694, 0.00000786, -0.00000969, 0.00000853, -0.00001481, 0.00001984, -0.00001299, 0.00001103, -0.00002687, 0.00002450, -0.00001319, 0.00000905, -0.00002003, 0.00002636, -0.00001650, 0.00000548, -0.00000442, 0.00000236, -0.00000097, -0.00000111, -0.00002143, 0.00004015, -0.00006492, 0.00003164, -0.00005872, 0.00005304, -0.00011426, 0.00005113, -0.00049727, 0.00095063, -0.00005008, -0.00024982, 0.00005927, -0.00009297, 0.00005525, -0.00003506, 0.00002059, -0.00003425, 0.00002203, -0.00000567, 0.00000756, -0.00003174, 0.00002202, -0.00002479, 0.00002749, -0.00003889, 0.00001839, -0.00001576, 0.00002886, -0.00002511, 0.00000531, -0.00000498, 0.00001529, -0.00002029, 0.00001327, -0.00000698, 0.00000494, -0.00000592, -0.00000099, -0.00000072, 0.00001405, -0.00001493, 0.00000758, -0.00000503, -0.00000155, -0.00000234, 0.00001069, -0.00000782, 0.00000878, -0.00001308, 0.00000193, -0.00000155, 0.00000680, -0.00000458, 0.00000460, -0.00000189, 0.00000524, },
		{ -0.00000354, -0.00000065, 0.00000540, -0.00000534, 0.00000274, -0.00000125, 0.00000001, -0.00000056, 0.00000855, -0.00000549, 0.00000009, -0.00000158, -0.00000165, 0.00000087, -0.00000025, -0.00000081, 0.00000039, 0.00000121, -0.00000213, 0.00000196, -0.00000470, 0.00000154, 0.00000574, -0.00000057, -0.00000267, 0.00000105, -0.00000332, 0.00000144, 0.00000284, -0.00000146, -0.00000796, 0.00000785, -0.00000460, 0.00000567, -0.00000537, 0.00000691, -0.00000357, 0.00000334, -0.00001357, 0.00001026, -0.00000872, 0.00000542, 0.00000334, 0.00000281, -0.00000865, 0.00000619, -0.00000946, 0.00001121, -0.00000006, -0.00000107, -0.00001073, 0.00001141, -0.00001328, 0.00001228, -0.00001834, 0.00002279, -0.00001614, 0.00001510, -0.00003035, 0.00002604, -0.00001450, 0.00001176, -0.00002304, 0.00002843, -0.00001860, 0.00000820, -0.00000847, 0.00000790, -0.00000897, 0.00001009, -0.00003465, 0.00005488, -0.00007887, 0.00004588, -0.00007530, 0.00007456, -0.00014053, 0.00009424, -0.00054797, 0.00093544, 0.00002296, -0.00029935, 0.00008595, -0.00011441, 0.00006983, -0.00004740, 0.00003095, -0.00004553, 0.00003245, -0.00001659, 0.00002003, -0.00004270, 0.00003148, -0.00003326, 0.00003471, -0.00004404, 0.00002333, -0.00002134, 0.00003303, -0.00002690, 0.00000720, -0.00000862, 0.00001879, -0.00002267, 0.00001509, -0.00000866, 0.00000680, -0.00000786, 0.00000153, -0.00000455, 0.00001740, -0.00001683, 0.00000916, -0.00000653, 0.00000050, -0.00000539, 0.00001348, -0.00000976, 0.00001038, -0.00001381, 0.00000254, -0.00000300, 0.00000805, -0.00000546, 0.00000538, -0.00000273, 0.00000588, -0.00000579, },
		{ -0.00000220, 0.00000671, -0.00000598, 0.00000311, -0.00000174, 0.00000070, -0.00000181, 0.00000941, -0.00000528, -0.00000017, -0.00000142, -0.00000169, 0.00000068, 0.00000002, -0.00000092, 0.00000029, 0.00000138, -0.00000229, 0.00000200, -0.00000441, 0.00000074, 0.00000586, 0.00000023, -0.00000336, 0.00000165, -0.00000374, 0.00000167, 0.00000208, 0.00000029, -0.00000961, 0.00000871, -0.00000551, 0.00000660, -0.00000621, 0.00000765, -0.00000442, 0.00000494, -0.00001473, 0.00001059, -0.00000891, 0.00000538, 0.00000249, 0.00000489, -0.00001071, 0.00000794, -0.00001110, 0.00001232, -0.00000180, 0.00000245, -0.00001459, 0.00001490, -0.00001674, 0.00001581, -0.00002157, 0.00002546, -0.00001907, 0.00001885, -0.00003329, 0.00002711, -0.00001566, 0.00001446, -0.00002600, 0.00003055, -0.00002097, 0.00001141, -0.00001320, 0.00001424, -0.00001787, 0.00002196, -0.00004811, 0.00006934, -0.00009218, 0.00005937, -0.00009106, 0.00009494, -0.00016503, 0.00013332, -0.00058848, 0.00090573, 0.00010605, -0.00035256, 0.00011412, -0.00013682, 0.00008519, -0.00006060, 0.00004237, -0.00005789, 0.00004386, -0.00002839, 0.00003290, -0.00005360, 0.00004069, -0.00004133, 0.00004143, -0.00004878, 0.00002799, -0.00002651, 0.00003663, -0.00002832, 0.00000907, -0.00001227, 0.00002214, -0.00002491, 0.00001687, -0.00001043, 0.00000880, -0.00001003, 0.00000433, -0.00000853, 0.00002065, -0.00001862, 0.00001074, -0.00000811, 0.00000270, -0.00000847, 0.00001612, -0.00001149, 0.00001177, -0.00001438, 0.00000313, -0.00000445, 0.00000922, -0.00000628, 0.00000613, -0.00000354, 0.00000650, -0.00000589, -0.00000283, },
		{ 0.00000795, -0.00000653, 0.00000345, -0.00000225, 0.00000142, -0.00000302, 0.00001011, -0.00000493, -0.00000047, -0.00000128, -0.00000169, 0.00000043, 0.00000031, -0.00000104, 0.00000019, 0.00000153, -0.00000238, 0.00000195, -0.00000404, -0.00000002, 0.00000584, 0.00000113, -0.00000407, 0.00000222, -0.00000414, 0.00000195, 0.00000122, 0.00000206, -0.00001114, 0.00000940, -0.00000627, 0.00000738, -0.00000691, 0.00000827, -0.00000519, 0.00000642, -0.00001566, 0.00001072, -0.00000902, 0.00000541, 0.00000144, 0.00000711, -0.00001276, 0.00000963, -0.00001269, 0.00001347, -0.00000374, 0.00000610, -0.00001834, 0.00001815, -0.00001989, 0.00001895, -0.00002436, 0.00002773, -0.00002165, 0.00002213, -0.00003559, 0.00002766, -0.00001662, 0.00001708, -0.00002882, 0.00003262, -0.00002352, 0.00001498, -0.00001840, 0.00002113, -0.00002728, 0.00003400, -0.00006123, 0.00008293, -0.00010420, 0.00007145, -0.00010525, 0.00011331, -0.00018664, 0.00016699, -0.00061759, 0.00086131, 0.00019830, -0.00040827, 0.00014284, -0.00015945, 0.00010076, -0.00007417, 0.00005440, -0.00007087, 0.00005578, -0.00004057, 0.00004563, -0.00006396, 0.00004921, -0.00004862, 0.00004734, -0.00005286, 0.00003213, -0.00003107, 0.00003952, -0.00002928, 0.00001084, -0.00001583, 0.00002526, -0.00002693, 0.00001855, -0.00001220, 0.00001089, -0.00001234, 0.00000731, -0.00001252, 0.00002369, -0.00002021, 0.00001224, -0.00000970, 0.00000493, -0.00001147, 0.00001851, -0.00001291, 0.00001287, -0.00001477, 0.00000368, -0.00000587, 0.00001028, -0.00000700, 0.00000681, -0.00000430, 0.00000708, -0.00000605, -0.00000198, -0.00000381, },
		{ -0.00000695, 0.00000373, -0.00000274, 0.00000213, -0.00000414, 0.00001061, -0.00000444, -0.00000080, -0.00000116, -0.00000163, 0.00000014, 0.00000061, -0.00000114, 0.00000009, 0.00000165, -0.00000241, 0.00000181, -0.00000361, -0.00000073, 0.00000566, 0.00000211, -0.00000476, 0.00000275, -0.00000449, 0.00000224, 0.00000030, 0.00000380, -0.00001246, 0.00000987, -0.00000686, 0.00000799, -0.00000743, 0.00000874, -0.00000585, 0.00000773, -0.00001633, 0.00001064, -0.00000904, 0.00000550, 0.00000026, 0.00000936, -0.00001471, 0.00001117, -0.00001414, 0.00001458, -0.00000578, 0.00000973, -0.00002183, 0.00002102, -0.00002260, 0.00002156, -0.00002659, 0.00002949, -0.00002377, 0.00002483, -0.00003717, 0.00002765, -0.00001734, 0.00001955, -0.00003141, 0.00003455, -0.00002612, 0.00001874, -0.00002383, 0.00002824, -0.00003678, 0.00004567, -0.00007341, 0.00009502, -0.00011432, 0.00008150, -0.00011718, 0.00012885, -0.00020437, 0.00019408, -0.00063443, 0.00080240, 0.00029843, -0.00046509, 0.00017101, -0.00018150, 0.00011589, -0.00008755, 0.00006651, -0.00008394, 0.00006765, -0.00005252, 0.00005765, -0.00007328, 0.00005663, -0.00005480, 0.00005217, -0.00005605, 0.00003556, -0.00003482, 0.00004156, -0.00002970, 0.00001242, -0.00001919, 0.00002803, -0.00002865, 0.00002007, -0.00001392, 0.00001298, -0.00001470, 0.00001033, -0.00001636, 0.00002638, -0.00002149, 0.00001358, -0.00001121, 0.00000710, -0.00001427, 0.00002053, -0.00001394, 0.00001365, -0.00001493, 0.00000416, -0.00000719, 0.00001119, -0.00000758, 0.00000740, -0.00000498, 0.00000760, -0.00000624, -0.00000102, -0.00000543, 0.00000907, },
		{ 0.00000394, -0.00000321, 0.00000281, -0.00000515, 0.00001090, -0.00000379, -0.00000116, -0.00000107, -0.00000153, -0.00000019, 0.00000091, -0.00000122, -0.00000001, 0.00000173, -0.00000237, 0.00000160, -0.00000314, -0.00000137, 0.00000535, 0.00000315, -0.00000540, 0.00000320, -0.00000478, 0.00000251, -0.00000064, 0.00000542, -0.00001352, 0.00001008, -0.00000724, 0.00000838, -0.00000775, 0.00000905, -0.00000638, 0.00000882, -0.00001670, 0.00001034, -0.00000899, 0.00000562, -0.00000101, 0.00001158, -0.00001646, 0.00001248, -0.00001539, 0.00001559, -0.00000779, 0.00001318, -0.00002488, 0.00002333, -0.00002471, 0.00002354, -0.00002817, 0.00003065, -0.00002532, 0.00002685, -0.00003796, 0.00002706, -0.00001779, 0.00002179, -0.00003368, 0.00003624, -0.00002864, 0.00002248, -0.00002923, 0.00003522, -0.00004592, 0.00005644, -0.00008407, 0.00010502, -0.00012199, 0.00008895, -0.00012624, 0.00014084, -0.00021735, 0.00021365, -0.00063851, 0.00072967, 0.00040485, -0.00052140, 0.00019743, -0.00020206, 0.00012984, -0.00010012, 0.00007812, -0.00009652, 0.00007889, -0.00006365, 0.00006839, -0.00008108, 0.00006257, -0.00005955, 0.00005569, -0.00005816, 0.00003809, -0.00003762, 0.00004266, -0.00002953, 0.00001374, -0.00002224, 0.00003036, -0.00002999, 0.00002135, -0.00001549, 0.00001497, -0.00001698, 0.00001326, -0.00001991, 0.00002858, -0.00002237, 0.00001466, -0.00001254, 0.00000908, -0.00001675, 0.00002209, -0.00001453, 0.00001406, -0.00001485, 0.00000453, -0.00000839, 0.00001190, -0.00000801, 0.00000787, -0.00000556, 0.00000805, -0.00000644, -0.00000000, -0.00000700, 0.00001001, -0.00000720, },
		{ -0.00000361, 0.00000342, -0.00000600, 0.00001096, -0.00000301, -0.00000153, -0.00000101, -0.00000138, -0.00000054, 0.00000120, -0.00000129, -0.00000010, 0.00000177, -0.00000226, 0.00000132, -0.00000264, -0.00000193, 0.00000489, 0.00000420, -0.00000595, 0.00000354, -0.00000497, 0.00000276, -0.00000155, 0.00000688, -0.00001428, 0.00001001, -0.00000738, 0.00000855, -0.00000786, 0.00000917, -0.00000674, 0.00000968, -0.00001676, 0.00000983, -0.00000884, 0.00000575, -0.00000232, 0.00001367, -0.00001794, 0.00001349, -0.00001636, 0.00001640, -0.00000966, 0.00001630, -0.00002734, 0.00002498, -0.00002613, 0.00002477, -0.00002900, 0.00003114, -0.00002625, 0.00002813, -0.00003791, 0.00002586, -0.00001794, 0.00002372, -0.00003552, 0.00003756, -0.00003092, 0.00002600, -0.00003430, 0.00004170, -0.00005423, 0.00006577, -0.00009266, 0.00011244, -0.00012673, 0.00009334, -0.00013191, 0.00014869, -0.00022492, 0.00022505, -0.00062971, 0.00064420, 0.00051563, -0.00057542, 0.00022084, -0.00022020, 0.00014188, -0.00011125, 0.00008863, -0.00010801, 0.00008890, -0.00007337, 0.00007731, -0.00008692, 0.00006667, -0.00006260, 0.00005768, -0.00005903, 0.00003958, -0.00003933, 0.00004275, -0.00002873, 0.00001474, -0.00002489, 0.00003215, -0.00003087, 0.00002232, -0.00001684, 0.00001677, -0.00001907, 0.00001594, -0.00002301, 0.00003019, -0.00002278, 0.00001542, -0.00001360, 0.00001078, -0.00001880, 0.00002310, -0.00001462, 0.00001408, -0.00001452, 0.00000478, -0.00000942, 0.00001238, -0.00000825, 0.00000821, -0.00000601, 0.00000840, -0.00000664, 0.00000103, -0.00000845, 0.00001073, -0.00000727, 0.00000405, },
		{ 0.00000392, -0.00000667, 0.00001079, -0.00000211, -0.00000191, -0.00000098, -0.00000120, -0.00000090, 0.00000147, -0.00000133, -0.00000019, 0.00000177, -0.00000209, 0.00000098, -0.00000214, -0.00000239, 0.00000433, 0.00000525, -0.00000639, 0.00000376, -0.00000505, 0.00000294, -0.00000238, 0.00000813, -0.00001470, 0.00000963, -0.00000728, 0.00000848, -0.00000774, 0.00000909, -0.00000691, 0.00001026, -0.00001649, 0.00000911, -0.00000861, 0.00000587, -0.00000358, 0.00001554, -0.00001904, 0.00001412, -0.00001699, 0.00001694, -0.00001129, 0.00001895, -0.00002908, 0.00002584, -0.00002676, 0.00002520, -0.00002905, 0.00003093, -0.00002649, 0.00002859, -0.00003700, 0.00002407, -0.00001777, 0.00002527, -0.00003684, 0.00003841, -0.00003280, 0.00002906, -0.00003875, 0.00004733, -0.00006129, 0.00007320, -0.00009874, 0.00011686, -0.00012819, 0.00009434, -0.00013384, 0.00015200, -0.00022664, 0.00022793, -0.00060834, 0.00054746, 0.00062860, -0.00062528, 0.00023994, -0.00023499, 0.00015128, -0.00012031, 0.00009745, -0.00011782, 0.00009710, -0.00008111, 0.00008393, -0.00009045, 0.00006869, -0.00006377, 0.00005802, -0.00005856, 0.00003992, -0.00003987, 0.00004178, -0.00002728, 0.00001537, -0.00002704, 0.00003333, -0.00003124, 0.00002291, -0.00001787, 0.00001826, -0.00002085, 0.00001825, -0.00002553, 0.00003109, -0.00002263, 0.00001578, -0.00001432, 0.00001211, -0.00002034, 0.00002351, -0.00001419, 0.00001371, -0.00001393, 0.00000488, -0.00001025, 0.00001261, -0.00000829, 0.00000838, -0.00000633, 0.00000864, -0.00000679, 0.00000203, -0.00000974, 0.00001119, -0.00000713, 0.00000404, -0.00000392, },
		{ -0.00000712, 0.00001037, -0.00000109, -0.00000227, -0.00000099, -0.00000098, -0.00000125, 0.00000170, -0.00000134, -0.00000027, 0.00000173, -0.00000185, 0.00000060, -0.00000164, -0.00000277, 0.00000367, 0.00000624, -0.00000668, 0.00000382, -0.00000501, 0.00000304, -0.00000310, 0.00000911, -0.00001475, 0.00000895, -0.00000694, 0.00000817, -0.00000740, 0.00000881, -0.00000688, 0.00001057, -0.00001591, 0.00000819, -0.00000828, 0.00000594, -0.00000474, 0.00001711, -0.00001970, 0.00001431, -0.00001721, 0.00001714, -0.00001258, 0.00002102, -0.00003000, 0.00002585, -0.00002655, 0.00002479, -0.00002830, 0.00002999, -0.00002602, 0.00002823, -0.00003525, 0.00002170, -0.00001725, 0.00002637, -0.00003753, 0.00003866, -0.00003413, 0.00003146, -0.00004232, 0.00005176, -0.00006671, 0.00007830, -0.00010193, 0.00011797, -0.00012614, 0.00009174, -0.00013180, 0.00015055, -0.00022230, 0.00022225, -0.00057506, 0.00044129, 0.00074136, -0.00066905, 0.00025350, -0.00024558, 0.00015735, -0.00012672, 0.00010402, -0.00012539, 0.00010296, -0.00008640, 0.00008786, -0.00009140, 0.00006843, -0.00006293, 0.00005664, -0.00005670, 0.00003906, -0.00003922, 0.00003977, -0.00002519, 0.00001558, -0.00002863, 0.00003382, -0.00003103, 0.00002306, -0.00001852, 0.00001935, -0.00002220, 0.00002004, -0.00002734, 0.00003122, -0.00002189, 0.00001570, -0.00001463, 0.00001298, -0.00002129, 0.00002327, -0.00001323, 0.00001295, -0.00001311, 0.00000484, -0.00001086, 0.00001255, -0.00000810, 0.00000839, -0.00000649, 0.00000875, -0.00000688, 0.00000294, -0.00001081, 0.00001135, -0.00000677, 0.00000391, -0.00000413, 0.00000428, },
		{ 0.00000972, 0.00000002, -0.00000261, -0.00000104, -0.00000074, -0.00000160, 0.00000190, -0.00000131, -0.00000035, 0.00000165, -0.00000157, 0.00000018, -0.00000117, -0.00000305, 0.00000293, 0.00000714, -0.00000680, 0.00000372, -0.00000483, 0.00000303, -0.00000368, 0.00000980, -0.00001442, 0.00000795, -0.00000636, 0.00000763, -0.00000683, 0.00000832, -0.00000665, 0.00001058, -0.00001501, 0.00000708, -0.00000786, 0.00000593, -0.00000574, 0.00001831, -0.00001985, 0.00001401, -0.00001699, 0.00001696, -0.00001344, 0.00002240, -0.00003002, 0.00002496, -0.00002549, 0.00002354, -0.00002677, 0.00002834, -0.00002484, 0.00002705, -0.00003268, 0.00001881, -0.00001638, 0.00002695, -0.00003752, 0.00003822, -0.00003476, 0.00003299, -0.00004475, 0.00005472, -0.00007016, 0.00008078, -0.00010198, 0.00011563, -0.00012049, 0.00008550, -0.00012576, 0.00014429, -0.00021192, 0.00020832, -0.00053094, 0.00032785, 0.00085139, -0.00070486, 0.00026035, -0.00025117, 0.00015950, -0.00012996, 0.00010785, -0.00013025, 0.00010603, -0.00008883, 0.00008880, -0.00008958, 0.00006580, -0.00006005, 0.00005352, -0.00005345, 0.00003700, -0.00003737, 0.00003675, -0.00002250, 0.00001537, -0.00002960, 0.00003358, -0.00003022, 0.00002273, -0.00001871, 0.00001995, -0.00002303, 0.00002122, -0.00002834, 0.00003050, -0.00002052, 0.00001515, -0.00001449, 0.00001335, -0.00002161, 0.00002238, -0.00001176, 0.00001184, -0.00001205, 0.00000463, -0.00001122, 0.00001220, -0.00000769, 0.00000821, -0.00000648, 0.00000871, -0.00000687, 0.00000373, -0.00001161, 0.00001119, -0.00000618, 0.00000364, -0.00000422, 0.00000450, -0.00000736, },
		{ 0.00000119, -0.00000290, -0.00000113, -0.00000049, -0.00000191, 0.00000204, -0.00000126, -0.00000042, 0.00000153, -0.00000124, -0.00000025, -0.00000073, -0.00000323, 0.00000215, 0.00000793, -0.00000672, 0.00000345, -0.00000452, 0.00000292, -0.00000409, 0.00001017, -0.00001371, 0.00000668, -0.00000557, 0.00000688, -0.00000606, 0.00000765, -0.00000621, 0.00001030, -0.00001381, 0.00000581, -0.00000734, 0.00000580, -0.00000651, 0.00001906, -0.00001945, 0.00001320, -0.00001631, 0.00001635, -0.00001382, 0.00002302, -0.00002911, 0.00002319, -0.00002360, 0.00002150, -0.00002449, 0.00002603, -0.00002299, 0.00002508, -0.00002935, 0.00001544, -0.00001516, 0.00002697, -0.00003674, 0.00003700, -0.00003458, 0.00003348, -0.00004584, 0.00005597, -0.00007139, 0.00008042, -0.00009878, 0.00010981, -0.00011131, 0.00007574, -0.00011583, 0.00013340, -0.00019581, 0.00018672, -0.00047735, 0.00020950, 0.00095614, -0.00073091, 0.00025950, -0.00025114, 0.00015724, -0.00012962, 0.00010855, -0.00013202, 0.00010598, -0.00008812, 0.00008659, -0.00008495, 0.00006082, -0.00005518, 0.00004876, -0.00004889, 0.00003380, -0.00003438, 0.00003279, -0.00001926, 0.00001473, -0.00002990, 0.00003258, -0.00002879, 0.00002187, -0.00001838, 0.00002000, -0.00002324, 0.00002168, -0.00002847, 0.00002893, -0.00001853, 0.00001412, -0.00001387, 0.00001318, -0.00002128, 0.00002084, -0.00000982, 0.00001041, -0.00001079, 0.00000427, -0.00001131, 0.00001155, -0.00000705, 0.00000785, -0.00000629, 0.00000851, -0.00000673, 0.00000436, -0.00001210, 0.00001068, -0.00000536, 0.00000324, -0.00000418, 0.00000455, -0.00000737, 0.00000886, },
		{ -0.00000313, -0.00000126, -0.00000024, -0.00000219, 0.00000212, -0.00000117, -0.00000048, 0.00000137, -0.00000088, -0.00000068, -0.00000033, -0.00000333, 0.00000135, 0.00000857, -0.00000643, 0.00000301, -0.00000406, 0.00000268, -0.00000432, 0.00001021, -0.00001263, 0.00000515, -0.00000460, 0.00000594, -0.00000511, 0.00000680, -0.00000558, 0.00000974, -0.00001234, 0.00000439, -0.00000672, 0.00000553, -0.00000700, 0.00001932, -0.00001848, 0.00001187, -0.00001515, 0.00001529, -0.00001366, 0.00002285, -0.00002726, 0.00002056, -0.00002094, 0.00001872, -0.00002155, 0.00002312, -0.00002052, 0.00002238, -0.00002534, 0.00001166, -0.00001361, 0.00002639, -0.00003513, 0.00003496, -0.00003348, 0.00003282, -0.00004544, 0.00005534, -0.00007026, 0.00007713, -0.00009235, 0.00010064, -0.00009882, 0.00006271, -0.00010233, 0.00011825, -0.00017445, 0.00015831, -0.00041595, 0.00008880, 0.00105308, -0.00074560, 0.00025013, -0.00024504, 0.00015026, -0.00012542, 0.00010584, -0.00013042, 0.00010257, -0.00008414, 0.00008118, -0.00007758, 0.00005361, -0.00004846, 0.00004249, -0.00004314, 0.00002954, -0.00003035, 0.00002802, -0.00001555, 0.00001365, -0.00002952, 0.00003082, -0.00002672, 0.00002048, -0.00001750, 0.00001943, -0.00002278, 0.00002138, -0.00002770, 0.00002650, -0.00001595, 0.00001262, -0.00001277, 0.00001247, -0.00002032, 0.00001870, -0.00000747, 0.00000872, -0.00000937, 0.00000376, -0.00001115, 0.00001060, -0.00000620, 0.00000730, -0.00000591, 0.00000815, -0.00000646, 0.00000478, -0.00001225, 0.00000984, -0.00000433, 0.00000270, -0.00000400, 0.00000443, -0.00000716, 0.00000780, 0.00000238, },
		{ -0.00000142, 0.00000001, -0.00000242, 0.00000215, -0.00000104, -0.00000054, 0.00000118, -0.00000050, -0.00000110, 0.00000001, -0.00000334, 0.00000055, 0.00000904, -0.00000594, 0.00000240, -0.00000347, 0.00000233, -0.00000435, 0.00000992, -0.00001122, 0.00000341, -0.00000349, 0.00000485, -0.00000400, 0.00000580, -0.00000477, 0.00000893, -0.00001063, 0.00000286, -0.00000601, 0.00000510, -0.00000717, 0.00001906, -0.00001692, 0.00001004, -0.00001354, 0.00001381, -0.00001297, 0.00002189, -0.00002452, 0.00001716, -0.00001759, 0.00001531, -0.00001805, 0.00001970, -0.00001750, 0.00001902, -0.00002075, 0.00000756, -0.00001175, 0.00002519, -0.00003268, 0.00003206, -0.00003143, 0.00003091, -0.00004346, 0.00005277, -0.00006673, 0.00007095, -0.00008286, 0.00008840, -0.00008338, 0.00004682, -0.00008569, 0.00009937, -0.00014856, 0.00012419, -0.00034859, -0.00003164, 0.00113979, -0.00074757, 0.00023171, -0.00023261, 0.00013840, -0.00011723, 0.00009958, -0.00012535, 0.00009575, -0.00007687, 0.00007269, -0.00006765, 0.00004439, -0.00004010, 0.00003492, -0.00003638, 0.00002438, -0.00002541, 0.00002255, -0.00001146, 0.00001218, -0.00002846, 0.00002829, -0.00002404, 0.00001854, -0.00001603, 0.00001821, -0.00002162, 0.00002028, -0.00002602, 0.00002327, -0.00001283, 0.00001070, -0.00001120, 0.00001123, -0.00001875, 0.00001602, -0.00000477, 0.00000684, -0.00000782, 0.00000311, -0.00001072, 0.00000937, -0.00000514, 0.00000657, -0.00000536, 0.00000760, -0.00000603, 0.00000498, -0.00001206, 0.00000866, -0.00000310, 0.00000206, -0.00000370, 0.00000414, -0.00000673, 0.00000658, 0.00000358, -0.00000329, },
		{ 0.00000023, -0.00000259, 0.00000211, -0.00000089, -0.00000058, 0.00000097, -0.00000010, -0.00000149, 0.00000029, -0.00000328, -0.00000022, 0.00000932, -0.00000523, 0.00000165, -0.00000277, 0.00000187, -0.00000420, 0.00000933, -0.00000951, 0.00000152, -0.00000228, 0.00000364, -0.00000278, 0.00000467, -0.00000380, 0.00000789, -0.00000872, 0.00000125, -0.00000522, 0.00000448, -0.00000700, 0.00001827, -0.00001481, 0.00000774, -0.00001152, 0.00001191, -0.00001175, 0.00002015, -0.00002097, 0.00001310, -0.00001369, 0.00001140, -0.00001411, 0.00001589, -0.00001402, 0.00001512, -0.00001569, 0.00000324, -0.00000962, 0.00002336, -0.00002940, 0.00002832, -0.00002840, 0.00002773, -0.00003989, 0.00004826, -0.00006084, 0.00006206, -0.00007060, 0.00007350, -0.00006549, 0.00002862, -0.00006651, 0.00007744, -0.00011903, 0.00008562, -0.00027730, -0.00014918, 0.00121406, -0.00073576, 0.00020397, -0.00021385, 0.00012170, -0.00010510, 0.00008979, -0.00011681, 0.00008556, -0.00006646, 0.00006134, -0.00005547, 0.00003347, -0.00003042, 0.00002633, -0.00002884, 0.00001850, -0.00001974, 0.00001656, -0.00000711, 0.00001035, -0.00002673, 0.00002506, -0.00002078, 0.00001608, -0.00001398, 0.00001634, -0.00001975, 0.00001840, -0.00002348, 0.00001932, -0.00000925, 0.00000840, -0.00000922, 0.00000950, -0.00001663, 0.00001288, -0.00000183, 0.00000482, -0.00000620, 0.00000235, -0.00001004, 0.00000788, -0.00000390, 0.00000569, -0.00000464, 0.00000689, -0.00000544, 0.00000495, -0.00001152, 0.00000718, -0.00000171, 0.00000132, -0.00000327, 0.00000369, -0.00000612, 0.00000523, 0.00000475, -0.00000336, -0.00000161, },
		{ -0.00000270, 0.00000201, -0.00000070, -0.00000062, 0.00000074, 0.00000029, -0.00000185, 0.00000050, -0.00000317, -0.00000093, 0.00000940, -0.00000432, 0.00000077, -0.00000198, 0.00000131, -0.00000386, 0.00000845, -0.00000757, -0.00000046, -0.00000103, 0.00000235, -0.00000147, 0.00000346, -0.00000270, 0.00000665, -0.00000665, -0.00000041, -0.00000435, 0.00000368, -0.00000648, 0.00001695, -0.00001219, 0.00000504, -0.00000915, 0.00000965, -0.00001004, 0.00001771, -0.00001670, 0.00000851, -0.00000938, 0.00000713, -0.00000986, 0.00001180, -0.00001021, 0.00001078, -0.00001031, -0.00000121, -0.00000726, 0.00002094, -0.00002533, 0.00002377, -0.00002444, 0.00002332, -0.00003477, 0.00004191, -0.00005278, 0.00005072, -0.00005599, 0.00005647, -0.00004573, 0.00000875, -0.00004546, 0.00005326, -0.00008687, 0.00004403, -0.00020415, -0.00026130, 0.00127396, -0.00070947, 0.00016699, -0.00018901, 0.00010041, -0.00008923, 0.00007666, -0.00010499, 0.00007224, -0.00005320, 0.00004751, -0.00004144, 0.00002123, -0.00001975, 0.00001701, -0.00002077, 0.00001211, -0.00001352, 0.00001021, -0.00000261, 0.00000822, -0.00002437, 0.00002116, -0.00001702, 0.00001313, -0.00001138, 0.00001384, -0.00001720, 0.00001578, -0.00002015, 0.00001475, -0.00000533, 0.00000581, -0.00000688, 0.00000735, -0.00001405, 0.00000939, 0.00000126, 0.00000277, -0.00000454, 0.00000148, -0.00000913, 0.00000615, -0.00000251, 0.00000467, -0.00000375, 0.00000601, -0.00000468, 0.00000466, -0.00001065, 0.00000543, -0.00000021, 0.00000051, -0.00000273, 0.00000310, -0.00000533, 0.00000378, 0.00000585, -0.00000333, -0.00000183, 0.00000043, },
		{ 0.00000185, -0.00000049, -0.00000064, 0.00000049, 0.00000067, -0.00000215, 0.00000065, -0.00000300, -0.00000158, 0.00000927, -0.00000324, -0.00000022, -0.00000111, 0.00000068, -0.00000336, 0.00000733, -0.00000544, -0.00000247, 0.00000022, 0.00000103, -0.00000011, 0.00000219, -0.00000150, 0.00000525, -0.00000447, -0.00000207, -0.00000342, 0.00000268, -0.00000560, 0.00001515, -0.00000912, 0.00000202, -0.00000650, 0.00000710, -0.00000791, 0.00001466, -0.00001187, 0.00000356, -0.00000481, 0.00000266, -0.00000546, 0.00000756, -0.00000617, 0.00000613, -0.00000473, -0.00000567, -0.00000472, 0.00001796, -0.00002052, 0.00001851, -0.00001960, 0.00001775, -0.00002824, 0.00003391, -0.00004281, 0.00003735, -0.00003956, 0.00003791, -0.00002480, -0.00001209, -0.00002331, 0.00002771, -0.00005316, 0.00000088, -0.00013121, -0.00036565, 0.00131793, -0.00066842, 0.00012115, -0.00015858, 0.00007497, -0.00007001, 0.00006049, -0.00009020, 0.00005615, -0.00003749, 0.00003167, -0.00002606, 0.00000814, -0.00000850, 0.00000732, -0.00001245, 0.00000543, -0.00000697, 0.00000369, 0.00000192, 0.00000584, -0.00002144, 0.00001669, -0.00001282, 0.00000976, -0.00000827, 0.00001076, -0.00001404, 0.00001249, -0.00001615, 0.00000970, -0.00000119, 0.00000302, -0.00000425, 0.00000486, -0.00001110, 0.00000567, 0.00000439, 0.00000074, -0.00000290, 0.00000055, -0.00000801, 0.00000423, -0.00000101, 0.00000353, -0.00000273, 0.00000498, -0.00000377, 0.00000415, -0.00000946, 0.00000346, 0.00000137, -0.00000033, -0.00000211, 0.00000238, -0.00000442, 0.00000230, 0.00000684, -0.00000320, -0.00000207, 0.00000058, -0.00000273, },
		{ -0.00000026, -0.00000066, 0.00000024, 0.00000103, -0.00000240, 0.00000073, -0.00000279, -0.00000214, 0.00000894, -0.00000200, -0.00000127, -0.00000021, -0.00000001, -0.00000273, 0.00000601, -0.00000321, -0.00000444, 0.00000141, -0.00000028, 0.00000125, 0.00000090, -0.00000024, 0.00000373, -0.00000225, -0.00000370, -0.00000244, 0.00000150, -0.00000440, 0.00001292, -0.00000569, -0.00000121, -0.00000367, 0.00000433, -0.00000543, 0.00001110, -0.00000664, -0.00000157, -0.00000017, -0.00000185, -0.00000106, 0.00000330, -0.00000204, 0.00000132, 0.00000090, -0.00001003, -0.00000207, 0.00001448, -0.00001508, 0.00001264, -0.00001403, 0.00001118, -0.00002049, 0.00002451, -0.00003129, 0.00002241, -0.00002191, 0.00001853, -0.00000341, -0.00003315, -0.00000086, 0.00000172, -0.00001904, -0.00004232, -0.00006048, -0.00046013, 0.00134479, -0.00061272, 0.00006720, -0.00012331, 0.00004602, -0.00004797, 0.00004177, -0.00007289, 0.00003776, -0.00001987, 0.00001438, -0.00000990, -0.00000532, 0.00000291, -0.00000239, -0.00000417, -0.00000131, -0.00000029, -0.00000282, 0.00000636, 0.00000328, -0.00001802, 0.00001175, -0.00000829, 0.00000604, -0.00000472, 0.00000718, -0.00001034, 0.00000864, -0.00001160, 0.00000433, 0.00000303, 0.00000014, -0.00000144, 0.00000212, -0.00000790, 0.00000183, 0.00000744, -0.00000118, -0.00000132, -0.00000043, -0.00000671, 0.00000216, 0.00000056, 0.00000231, -0.00000160, 0.00000383, -0.00000271, 0.00000341, -0.00000799, 0.00000131, 0.00000297, -0.00000118, -0.00000144, 0.00000157, -0.00000341, 0.00000081, 0.00000770, -0.00000295, -0.00000231, 0.00000069, -0.00000269, 0.00000163, },
		{ -0.00000066, -0.00000002, 0.00000135, -0.00000258, 0.00000074, -0.00000256, -0.00000260, 0.00000841, -0.00000064, -0.00000234, 0.00000069, -0.00000072, -0.00000199, 0.00000454, -0.00000093, -0.00000630, 0.00000248, -0.00000154, 0.00000257, -0.00000038, 0.00000104, 0.00000216, -0.00000002, -0.00000526, -0.00000144, 0.00000016, -0.00000291, 0.00001032, -0.00000200, -0.00000456, -0.00000076, 0.00000145, -0.00000271, 0.00000718, -0.00000117, -0.00000668, 0.00000437, -0.00000623, 0.00000320, -0.00000083, 0.00000208, -0.00000353, 0.00000644, -0.00001418, 0.00000062, 0.00001058, -0.00000913, 0.00000631, -0.00000786, 0.00000380, -0.00001177, 0.00001405, -0.00001863, 0.00000646, -0.00000370, -0.00000097, 0.00001768, -0.00005366, 0.00002109, -0.00002380, 0.00001437, -0.00008416, 0.00000620, -0.00054296, 0.00135383, -0.00054296, 0.00000620, -0.00008416, 0.00001437, -0.00002380, 0.00002109, -0.00005366, 0.00001768, -0.00000097, -0.00000370, 0.00000646, -0.00001863, 0.00001405, -0.00001177, 0.00000380, -0.00000786, 0.00000631, -0.00000913, 0.00001058, 0.00000062, -0.00001418, 0.00000644, -0.00000353, 0.00000208, -0.00000083, 0.00000320, -0.00000623, 0.00000437, -0.00000668, -0.00000117, 0.00000718, -0.00000271, 0.00000145, -0.00000076, -0.00000456, -0.00000200, 0.00001032, -0.00000291, 0.00000016, -0.00000144, -0.00000526, -0.00000002, 0.00000216, 0.00000104, -0.00000038, 0.00000257, -0.00000154, 0.00000248, -0.00000630, -0.00000093, 0.00000454, -0.00000199, -0.00000072, 0.00000069, -0.00000234, -0.00000064, 0.00000841, -0.00000260, -0.00000256, 0.00000074, -0.00000258, 0.00000135, -0.00000002, },
		{ -0.00000026, 0.00000163, -0.00000269, 0.00000069, -0.00000231, -0.00000295, 0.00000770, 0.00000081, -0.00000341, 0.00000157, -0.00000144, -0.00000118, 0.00000297, 0.00000131, -0.00000799, 0.00000341, -0.00000271, 0.00000383, -0.00000160, 0.00000231, 0.00000056, 0.00000216, -0.00000671, -0.00000043, -0.00000132, -0.00000118, 0.00000744, 0.00000183, -0.00000790, 0.00000212, -0.00000144, 0.00000014, 0.00000303, 0.00000433, -0.00001160, 0.00000864, -0.00001034, 0.00000718, -0.00000472, 0.00000604, -0.00000829, 0.00001175, -0.00001802, 0.00000328, 0.00000636, -0.00000282, -0.00000029, -0.00000131, -0.00000417, -0.00000239, 0.00000291, -0.00000532, -0.00000990, 0.00001438, -0.00001987, 0.00003776, -0.00007289, 0.00004177, -0.00004797, 0.00004602, -0.00012331, 0.00006720, -0.00061272, 0.00134479, -0.00046013, -0.00006048, -0.00004232, -0.00001904, 0.00000172, -0.00000086, -0.00003315, -0.00000341, 0.00001853, -0.00002191, 0.00002241, -0.00003129, 0.00002451, -0.00002049, 0.00001118, -0.00001403, 0.00001264, -0.00001508, 0.00001448, -0.00000207, -0.00001003, 0.00000090, 0.00000132, -0.00000204, 0.00000330, -0.00000106, -0.00000185, -0.00000017, -0.00000157, -0.00000664, 0.00001110, -0.00000543, 0.00000433, -0.00000367, -0.00000121, -0.00000569, 0.00001292, -0.00000440, 0.00000150, -0.00000244, -0.00000370, -0.00000225, 0.00000373, -0.00000024, 0.00000090, 0.00000125, -0.00000028, 0.00000141, -0.00000444, -0.00000321, 0.00000601, -0.00000273, -0.00000001, -0.00000021, -0.00000127, -0.00000200, 0.00000894, -0.00000214, -0.00000279, 0.00000073, -0.00000240, 0.00000103, 0.00000024, -0.00000066, },
		{ 0.00000185, -0.00000273, 0.00000058, -0.00000207, -0.00000320, 0.00000684, 0.00000230, -0.00000442, 0.00000238, -0.00000211, -0.00000033, 0.00000137, 0.00000346, -0.00000946, 0.00000415, -0.00000377, 0.00000498, -0.00000273, 0.00000353, -0.00000101, 0.00000423, -0.00000801, 0.00000055, -0.00000290, 0.00000074, 0.00000439, 0.00000567, -0.00001110, 0.00000486, -0.00000425, 0.00000302, -0.00000119, 0.00000970, -0.00001615, 0.00001249, -0.00001404, 0.00001076, -0.00000827, 0.00000976, -0.00001282, 0.00001669, -0.00002144, 0.00000584, 0.00000192, 0.00000369, -0.00000697, 0.00000543, -0.00001245, 0.00000732, -0.00000850, 0.00000814, -0.00002606, 0.00003167, -0.00003749, 0.00005615, -0.00009020, 0.00006049, -0.00007001, 0.00007497, -0.00015858, 0.00012115, -0.00066842, 0.00131793, -0.00036565, -0.00013121, 0.00000088, -0.00005316, 0.00002771, -0.00002331, -0.00001209, -0.00002480, 0.00003791, -0.00003956, 0.00003735, -0.00004281, 0.00003391, -0.00002824, 0.00001775, -0.00001960, 0.00001851, -0.00002052, 0.00001796, -0.00000472, -0.00000567, -0.00000473, 0.00000613, -0.00000617, 0.00000756, -0.00000546, 0.00000266, -0.00000481, 0.00000356, -0.00001187, 0.00001466, -0.00000791, 0.00000710, -0.00000650, 0.00000202, -0.00000912, 0.00001515, -0.00000560, 0.00000268, -0.00000342, -0.00000207, -0.00000447, 0.00000525, -0.00000150, 0.00000219, -0.00000011, 0.00000103, 0.00000022, -0.00000247, -0.00000544, 0.00000733, -0.00000336, 0.00000068, -0.00000111, -0.00000022, -0.00000324, 0.00000927, -0.00000158, -0.00000300, 0.00000065, -0.00000215, 0.00000067, 0.00000049, -0.00000064, -0.00000049, },
		{ -0.00000270, 0.00000043, -0.00000183, -0.00000333, 0.00000585, 0.00000378, -0.00000533, 0.00000310, -0.00000273, 0.00000051, -0.00000021, 0.00000543, -0.00001065, 0.00000466, -0.00000468, 0.00000601, -0.00000375, 0.00000467, -0.00000251, 0.00000615, -0.00000913, 0.00000148, -0.00000454, 0.00000277, 0.00000126, 0.00000939, -0.00001405, 0.00000735, -0.00000688, 0.00000581, -0.00000533, 0.00001475, -0.00002015, 0.00001578, -0.00001720, 0.00001384, -0.00001138, 0.00001313, -0.00001702, 0.00002116, -0.00002437, 0.00000822, -0.00000261, 0.00001021, -0.00001352, 0.00001211, -0.00002077, 0.00001701, -0.00001975, 0.00002123, -0.00004144, 0.00004751, -0.00005320, 0.00007224, -0.00010499, 0.00007666, -0.00008923, 0.00010041, -0.00018901, 0.00016699, -0.00070947, 0.00127396, -0.00026130, -0.00020415, 0.00004403, -0.00008687, 0.00005326, -0.00004546, 0.00000875, -0.00004573, 0.00005647, -0.00005599, 0.00005072, -0.00005278, 0.00004191, -0.00003477, 0.00002332, -0.00002444, 0.00002377, -0.00002533, 0.00002094, -0.00000726, -0.00000121, -0.00001031, 0.00001078, -0.00001021, 0.00001180, -0.00000986, 0.00000713, -0.00000938, 0.00000851, -0.00001670, 0.00001771, -0.00001004, 0.00000965, -0.00000915, 0.00000504, -0.00001219, 0.00001695, -0.00000648, 0.00000368, -0.00000435, -0.00000041, -0.00000665, 0.00000665, -0.00000270, 0.00000346, -0.00000147, 0.00000235, -0.00000103, -0.00000046, -0.00000757, 0.00000845, -0.00000386, 0.00000131, -0.00000198, 0.00000077, -0.00000432, 0.00000940, -0.00000093, -0.00000317, 0.00000050, -0.00000185, 0.00000029, 0.00000074, -0.00000062, -0.00000070, 0.00000201, },
		{ 0.00000023, -0.00000161, -0.00000336, 0.00000475, 0.00000523, -0.00000612, 0.00000369, -0.00000327, 0.00000132, -0.00000171, 0.00000718, -0.00001152, 0.00000495, -0.00000544, 0.00000689, -0.00000464, 0.00000569, -0.00000390, 0.00000788, -0.00001004, 0.00000235, -0.00000620, 0.00000482, -0.00000183, 0.00001288, -0.00001663, 0.00000950, -0.00000922, 0.00000840, -0.00000925, 0.00001932, -0.00002348, 0.00001840, -0.00001975, 0.00001634, -0.00001398, 0.00001608, -0.00002078, 0.00002506, -0.00002673, 0.00001035, -0.00000711, 0.00001656, -0.00001974, 0.00001850, -0.00002884, 0.00002633, -0.00003042, 0.00003347, -0.00005547, 0.00006134, -0.00006646, 0.00008556, -0.00011681, 0.00008979, -0.00010510, 0.00012170, -0.00021385, 0.00020397, -0.00073576, 0.00121406, -0.00014918, -0.00027730, 0.00008562, -0.00011903, 0.00007744, -0.00006651, 0.00002862, -0.00006549, 0.00007350, -0.00007060, 0.00006206, -0.00006084, 0.00004826, -0.00003989, 0.00002773, -0.00002840, 0.00002832, -0.00002940, 0.00002336, -0.00000962, 0.00000324, -0.00001569, 0.00001512, -0.00001402, 0.00001589, -0.00001411, 0.00001140, -0.00001369, 0.00001310, -0.00002097, 0.00002015, -0.00001175, 0.00001191, -0.00001152, 0.00000774, -0.00001481, 0.00001827, -0.00000700, 0.00000448, -0.00000522, 0.00000125, -0.00000872, 0.00000789, -0.00000380, 0.00000467, -0.00000278, 0.00000364, -0.00000228, 0.00000152, -0.00000951, 0.00000933, -0.00000420, 0.00000187, -0.00000277, 0.00000165, -0.00000523, 0.00000932, -0.00000022, -0.00000328, 0.00000029, -0.00000149, -0.00000010, 0.00000097, -0.00000058, -0.00000089, 0.00000211, -0.00000259, },
		{ -0.00000142, -0.00000329, 0.00000358, 0.00000658, -0.00000673, 0.00000414, -0.00000370, 0.00000206, -0.00000310, 0.00000866, -0.00001206, 0.00000498, -0.00000603, 0.00000760, -0.00000536, 0.00000657, -0.00000514, 0.00000937, -0.00001072, 0.00000311, -0.00000782, 0.00000684, -0.00000477, 0.00001602, -0.00001875, 0.00001123, -0.00001120, 0.00001070, -0.00001283, 0.00002327, -0.00002602, 0.00002028, -0.00002162, 0.00001821, -0.00001603, 0.00001854, -0.00002404, 0.00002829, -0.00002846, 0.00001218, -0.00001146, 0.00002255, -0.00002541, 0.00002438, -0.00003638, 0.00003492, -0.00004010, 0.00004439, -0.00006765, 0.00007269, -0.00007687, 0.00009575, -0.00012535, 0.00009958, -0.00011723, 0.00013840, -0.00023261, 0.00023171, -0.00074757, 0.00113979, -0.00003164, -0.00034859, 0.00012419, -0.00014856, 0.00009937, -0.00008569, 0.00004682, -0.00008338, 0.00008840, -0.00008286, 0.00007095, -0.00006673, 0.00005277, -0.00004346, 0.00003091, -0.00003143, 0.00003206, -0.00003268, 0.00002519, -0.00001175, 0.00000756, -0.00002075, 0.00001902, -0.00001750, 0.00001970, -0.00001805, 0.00001531, -0.00001759, 0.00001716, -0.00002452, 0.00002189, -0.00001297, 0.00001381, -0.00001354, 0.00001004, -0.00001692, 0.00001906, -0.00000717, 0.00000510, -0.00000601, 0.00000286, -0.00001063, 0.00000893, -0.00000477, 0.00000580, -0.00000400, 0.00000485, -0.00000349, 0.00000341, -0.00001122, 0.00000992, -0.00000435, 0.00000233, -0.00000347, 0.00000240, -0.00000594, 0.00000904, 0.00000055, -0.00000334, 0.00000001, -0.00000110, -0.00000050, 0.00000118, -0.00000054, -0.00000104, 0.00000215, -0.00000242, 0.00000001, },
		{ -0.00000313, 0.00000238, 0.00000780, -0.00000716, 0.00000443, -0.00000400, 0.00000270, -0.00000433, 0.00000984, -0.00001225, 0.00000478, -0.00000646, 0.00000815, -0.00000591, 0.00000730, -0.00000620, 0.00001060, -0.00001115, 0.00000376, -0.00000937, 0.00000872, -0.00000747, 0.00001870, -0.00002032, 0.00001247, -0.00001277, 0.00001262, -0.00001595, 0.00002650, -0.00002770, 0.00002138, -0.00002278, 0.00001943, -0.00001750, 0.00002048, -0.00002672, 0.00003082, -0.00002952, 0.00001365, -0.00001555, 0.00002802, -0.00003035, 0.00002954, -0.00004314, 0.00004249, -0.00004846, 0.00005361, -0.00007758, 0.00008118, -0.00008414, 0.00010257, -0.00013042, 0.00010584, -0.00012542, 0.00015026, -0.00024504, 0.00025013, -0.00074560, 0.00105308, 0.00008880, -0.00041595, 0.00015831, -0.00017445, 0.00011825, -0.00010233, 0.00006271, -0.00009882, 0.00010064, -0.00009235, 0.00007713, -0.00007026, 0.00005534, -0.00004544, 0.00003282, -0.00003348, 0.00003496, -0.00003513, 0.00002639, -0.00001361, 0.00001166, -0.00002534, 0.00002238, -0.00002052, 0.00002312, -0.00002155, 0.00001872, -0.00002094, 0.00002056, -0.00002726, 0.00002285, -0.00001366, 0.00001529, -0.00001515, 0.00001187, -0.00001848, 0.00001932, -0.00000700, 0.00000553, -0.00000672, 0.00000439, -0.00001234, 0.00000974, -0.00000558, 0.00000680, -0.00000511, 0.00000594, -0.00000460, 0.00000515, -0.00001263, 0.00001021, -0.00000432, 0.00000268, -0.00000406, 0.00000301, -0.00000643, 0.00000857, 0.00000135, -0.00000333, -0.00000033, -0.00000068, -0.00000088, 0.00000137, -0.00000048, -0.00000117, 0.00000212, -0.00000219, -0.00000024, -0.00000126, },
		{ 0.00000119, 0.00000886, -0.00000737, 0.00000455, -0.00000418, 0.00000324, -0.00000536, 0.00001068, -0.00001210, 0.00000436, -0.00000673, 0.00000851, -0.00000629, 0.00000785, -0.00000705, 0.00001155, -0.00001131, 0.00000427, -0.00001079, 0.00001041, -0.00000982, 0.00002084, -0.00002128, 0.00001318, -0.00001387, 0.00001412, -0.00001853, 0.00002893, -0.00002847, 0.00002168, -0.00002324, 0.00002000, -0.00001838, 0.00002187, -0.00002879, 0.00003258, -0.00002990, 0.00001473, -0.00001926, 0.00003279, -0.00003438, 0.00003380, -0.00004889, 0.00004876, -0.00005518, 0.00006082, -0.00008495, 0.00008659, -0.00008812, 0.00010598, -0.00013202, 0.00010855, -0.00012962, 0.00015724, -0.00025114, 0.00025950, -0.00073091, 0.00095614, 0.00020950, -0.00047735, 0.00018672, -0.00019581, 0.00013340, -0.00011583, 0.00007574, -0.00011131, 0.00010981, -0.00009878, 0.00008042, -0.00007139, 0.00005597, -0.00004584, 0.00003348, -0.00003458, 0.00003700, -0.00003674, 0.00002697, -0.00001516, 0.00001544, -0.00002935, 0.00002508, -0.00002299, 0.00002603, -0.00002449, 0.00002150, -0.00002360, 0.00002319, -0.00002911, 0.00002302, -0.00001382, 0.00001635, -0.00001631, 0.00001320, -0.00001945, 0.00001906, -0.00000651, 0.00000580, -0.00000734, 0.00000581, -0.00001381, 0.00001030, -0.00000621, 0.00000765, -0.00000606, 0.00000688, -0.00000557, 0.00000668, -0.00001371, 0.00001017, -0.00000409, 0.00000292, -0.00000452, 0.00000345, -0.00000672, 0.00000793, 0.00000215, -0.00000323, -0.00000073, -0.00000025, -0.00000124, 0.00000153, -0.00000042, -0.00000126, 0.00000204, -0.00000191, -0.00000049, -0.00000113, -0.00000290, },
		{ 0.00000972, -0.00000736, 0.00000450, -0.00000422, 0.00000364, -0.00000618, 0.00001119, -0.00001161, 0.00000373, -0.00000687, 0.00000871, -0.00000648, 0.00000821, -0.00000769, 0.00001220, -0.00001122, 0.00000463, -0.00001205, 0.00001184, -0.00001176, 0.00002238, -0.00002161, 0.00001335, -0.00001449, 0.00001515, -0.00002052, 0.00003050, -0.00002834, 0.00002122, -0.00002303, 0.00001995, -0.00001871, 0.00002273, -0.00003022, 0.00003358, -0.00002960, 0.00001537, -0.00002250, 0.00003675, -0.00003737, 0.00003700, -0.00005345, 0.00005352, -0.00006005, 0.00006580, -0.00008958, 0.00008880, -0.00008883, 0.00010603, -0.00013025, 0.00010785, -0.00012996, 0.00015950, -0.00025117, 0.00026035, -0.00070486, 0.00085139, 0.00032785, -0.00053094, 0.00020832, -0.00021192, 0.00014429, -0.00012576, 0.00008550, -0.00012049, 0.00011563, -0.00010198, 0.00008078, -0.00007016, 0.00005472, -0.00004475, 0.00003299, -0.00003476, 0.00003822, -0.00003752, 0.00002695, -0.00001638, 0.00001881, -0.00003268, 0.00002705, -0.00002484, 0.00002834, -0.00002677, 0.00002354, -0.00002549, 0.00002496, -0.00003002, 0.00002240, -0.00001344, 0.00001696, -0.00001699, 0.00001401, -0.00001985, 0.00001831, -0.00000574, 0.00000593, -0.00000786, 0.00000708, -0.00001501, 0.00001058, -0.00000665, 0.00000832, -0.00000683, 0.00000763, -0.00000636, 0.00000795, -0.00001442, 0.00000980, -0.00000368, 0.00000303, -0.00000483, 0.00000372, -0.00000680, 0.00000714, 0.00000293, -0.00000305, -0.00000117, 0.00000018, -0.00000157, 0.00000165, -0.00000035, -0.00000131, 0.00000190, -0.00000160, -0.00000074, -0.00000104, -0.00000261, 0.00000002, },
		{ -0.00000712, 0.00000428, -0.00000413, 0.00000391, -0.00000677, 0.00001135, -0.00001081, 0.00000294, -0.00000688, 0.00000875, -0.00000649, 0.00000839, -0.00000810, 0.00001255, -0.00001086, 0.00000484, -0.00001311, 0.00001295, -0.00001323, 0.00002327, -0.00002129, 0.00001298, -0.00001463, 0.00001570, -0.00002189, 0.00003122, -0.00002734, 0.00002004, -0.00002220, 0.00001935, -0.00001852, 0.00002306, -0.00003103, 0.00003382, -0.00002863, 0.00001558, -0.00002519, 0.00003977, -0.00003922, 0.00003906, -0.00005670, 0.00005664, -0.00006293, 0.00006843, -0.00009140, 0.00008786, -0.00008640, 0.00010296, -0.00012539, 0.00010402, -0.00012672, 0.00015735, -0.00024558, 0.00025350, -0.00066905, 0.00074136, 0.00044129, -0.00057506, 0.00022225, -0.00022230, 0.00015055, -0.00013180, 0.00009174, -0.00012614, 0.00011797, -0.00010193, 0.00007830, -0.00006671, 0.00005176, -0.00004232, 0.00003146, -0.00003413, 0.00003866, -0.00003753, 0.00002637, -0.00001725, 0.00002170, -0.00003525, 0.00002823, -0.00002602, 0.00002999, -0.00002830, 0.00002479, -0.00002655, 0.00002585, -0.00003000, 0.00002102, -0.00001258, 0.00001714, -0.00001721, 0.00001431, -0.00001970, 0.00001711, -0.00000474, 0.00000594, -0.00000828, 0.00000819, -0.00001591, 0.00001057, -0.00000688, 0.00000881, -0.00000740, 0.00000817, -0.00000694, 0.00000895, -0.00001475, 0.00000911, -0.00000310, 0.00000304, -0.00000501, 0.00000382, -0.00000668, 0.00000624, 0.00000367, -0.00000277, -0.00000164, 0.00000060, -0.00000185, 0.00000173, -0.00000027, -0.00000134, 0.00000170, -0.00000125, -0.00000098, -0.00000099, -0.00000227, -0.00000109, 0.00001037, },
		{ 0.00000392, -0.00000392, 0.00000404, -0.00000713, 0.00001119, -0.00000974, 0.00000203, -0.00000679, 0.00000864, -0.00000633, 0.00000838, -0.00000829, 0.00001261, -0.00001025, 0.00000488, -0.00001393, 0.00001371, -0.00001419, 0.00002351, -0.00002034, 0.00001211, -0.00001432, 0.00001578, -0.00002263, 0.00003109, -0.00002553, 0.00001825, -0.00002085, 0.00001826, -0.00001787, 0.00002291, -0.00003124, 0.00003333, -0.00002704, 0.00001537, -0.00002728, 0.00004178, -0.00003987, 0.00003992, -0.00005856, 0.00005802, -0.00006377, 0.00006869, -0.00009045, 0.00008393, -0.00008111, 0.00009710, -0.00011782, 0.00009745, -0.00012031, 0.00015128, -0.00023499, 0.00023994, -0.00062528, 0.00062860, 0.00054746, -0.00060834, 0.00022793, -0.00022664, 0.00015200, -0.00013384, 0.00009434, -0.00012819, 0.00011686, -0.00009874, 0.00007320, -0.00006129, 0.00004733, -0.00003875, 0.00002906, -0.00003280, 0.00003841, -0.00003684, 0.00002527, -0.00001777, 0.00002407, -0.00003700, 0.00002859, -0.00002649, 0.00003093, -0.00002905, 0.00002520, -0.00002676, 0.00002584, -0.00002908, 0.00001895, -0.00001129, 0.00001694, -0.00001699, 0.00001412, -0.00001904, 0.00001554, -0.00000358, 0.00000587, -0.00000861, 0.00000911, -0.00001649, 0.00001026, -0.00000691, 0.00000909, -0.00000774, 0.00000848, -0.00000728, 0.00000963, -0.00001470, 0.00000813, -0.00000238, 0.00000294, -0.00000505, 0.00000376, -0.00000639, 0.00000525, 0.00000433, -0.00000239, -0.00000214, 0.00000098, -0.00000209, 0.00000177, -0.00000019, -0.00000133, 0.00000147, -0.00000090, -0.00000120, -0.00000098, -0.00000191, -0.00000211, 0.00001079, -0.00000667, },
		{ -0.00000361, 0.00000405, -0.00000727, 0.00001073, -0.00000845, 0.00000103, -0.00000664, 0.00000840, -0.00000601, 0.00000821, -0.00000825, 0.00001238, -0.00000942, 0.00000478, -0.00001452, 0.00001408, -0.00001462, 0.00002310, -0.00001880, 0.00001078, -0.00001360, 0.00001542, -0.00002278, 0.00003019, -0.00002301, 0.00001594, -0.00001907, 0.00001677, -0.00001684, 0.00002232, -0.00003087, 0.00003215, -0.00002489, 0.00001474, -0.00002873, 0.00004275, -0.00003933, 0.00003958, -0.00005903, 0.00005768, -0.00006260, 0.00006667, -0.00008692, 0.00007731, -0.00007337, 0.00008890, -0.00010801, 0.00008863, -0.00011125, 0.00014188, -0.00022020, 0.00022084, -0.00057542, 0.00051563, 0.00064420, -0.00062971, 0.00022505, -0.00022492, 0.00014869, -0.00013191, 0.00009334, -0.00012673, 0.00011244, -0.00009266, 0.00006577, -0.00005423, 0.00004170, -0.00003430, 0.00002600, -0.00003092, 0.00003756, -0.00003552, 0.00002372, -0.00001794, 0.00002586, -0.00003791, 0.00002813, -0.00002625, 0.00003114, -0.00002900, 0.00002477, -0.00002613, 0.00002498, -0.00002734, 0.00001630, -0.00000966, 0.00001640, -0.00001636, 0.00001349, -0.00001794, 0.00001367, -0.00000232, 0.00000575, -0.00000884, 0.00000983, -0.00001676, 0.00000968, -0.00000674, 0.00000917, -0.00000786, 0.00000855, -0.00000738, 0.00001001, -0.00001428, 0.00000688, -0.00000155, 0.00000276, -0.00000497, 0.00000354, -0.00000595, 0.00000420, 0.00000489, -0.00000193, -0.00000264, 0.00000132, -0.00000226, 0.00000177, -0.00000010, -0.00000129, 0.00000120, -0.00000054, -0.00000138, -0.00000101, -0.00000153, -0.00000301, 0.00001096, -0.00000600, 0.00000342, },
		{ 0.00000394, -0.00000720, 0.00001001, -0.00000700, -0.00000000, -0.00000644, 0.00000805, -0.00000556, 0.00000787, -0.00000801, 0.00001190, -0.00000839, 0.00000453, -0.00001485, 0.00001406, -0.00001453, 0.00002209, -0.00001675, 0.00000908, -0.00001254, 0.00001466, -0.00002237, 0.00002858, -0.00001991, 0.00001326, -0.00001698, 0.00001497, -0.00001549, 0.00002135, -0.00002999, 0.00003036, -0.00002224, 0.00001374, -0.00002953, 0.00004266, -0.00003762, 0.00003809, -0.00005816, 0.00005569, -0.00005955, 0.00006257, -0.00008108, 0.00006839, -0.00006365, 0.00007889, -0.00009652, 0.00007812, -0.00010012, 0.00012984, -0.00020206, 0.00019743, -0.00052140, 0.00040485, 0.00072967, -0.00063851, 0.00021365, -0.00021735, 0.00014084, -0.00012624, 0.00008895, -0.00012199, 0.00010502, -0.00008407, 0.00005644, -0.00004592, 0.00003522, -0.00002923, 0.00002248, -0.00002864, 0.00003624, -0.00003368, 0.00002179, -0.00001779, 0.00002706, -0.00003796, 0.00002685, -0.00002532, 0.00003065, -0.00002817, 0.00002354, -0.00002471, 0.00002333, -0.00002488, 0.00001318, -0.00000779, 0.00001559, -0.00001539, 0.00001248, -0.00001646, 0.00001158, -0.00000101, 0.00000562, -0.00000899, 0.00001034, -0.00001670, 0.00000882, -0.00000638, 0.00000905, -0.00000775, 0.00000838, -0.00000724, 0.00001008, -0.00001352, 0.00000542, -0.00000064, 0.00000251, -0.00000478, 0.00000320, -0.00000540, 0.00000315, 0.00000535, -0.00000137, -0.00000314, 0.00000160, -0.00000237, 0.00000173, -0.00000001, -0.00000122, 0.00000091, -0.00000019, -0.00000153, -0.00000107, -0.00000116, -0.00000379, 0.00001090, -0.00000515, 0.00000281, -0.00000321, },
		{ -0.00000695, 0.00000907, -0.00000543, -0.00000102, -0.00000624, 0.00000760, -0.00000498, 0.00000740, -0.00000758, 0.00001119, -0.00000719, 0.00000416, -0.00001493, 0.00001365, -0.00001394, 0.00002053, -0.00001427, 0.00000710, -0.00001121, 0.00001358, -0.00002149, 0.00002638, -0.00001636, 0.00001033, -0.00001470, 0.00001298, -0.00001392, 0.00002007, -0.00002865, 0.00002803, -0.00001919, 0.00001242, -0.00002970, 0.00004156, -0.00003482, 0.00003556, -0.00005605, 0.00005217, -0.00005480, 0.00005663, -0.00007328, 0.00005765, -0.00005252, 0.00006765, -0.00008394, 0.00006651, -0.00008755, 0.00011589, -0.00018150, 0.00017101, -0.00046509, 0.00029843, 0.00080240, -0.00063443, 0.00019408, -0.00020437, 0.00012885, -0.00011718, 0.00008150, -0.00011432, 0.00009502, -0.00007341, 0.00004567, -0.00003678, 0.00002824, -0.00002383, 0.00001874, -0.00002612, 0.00003455, -0.00003141, 0.00001955, -0.00001734, 0.00002765, -0.00003717, 0.00002483, -0.00002377, 0.00002949, -0.00002659, 0.00002156, -0.00002260, 0.00002102, -0.00002183, 0.00000973, -0.00000578, 0.00001458, -0.00001414, 0.00001117, -0.00001471, 0.00000936, 0.00000026, 0.00000550, -0.00000904, 0.00001064, -0.00001633, 0.00000773, -0.00000585, 0.00000874, -0.00000743, 0.00000799, -0.00000686, 0.00000987, -0.00001246, 0.00000380, 0.00000030, 0.00000224, -0.00000449, 0.00000275, -0.00000476, 0.00000211, 0.00000566, -0.00000073, -0.00000361, 0.00000181, -0.00000241, 0.00000165, 0.00000009, -0.00000114, 0.00000061, 0.00000014, -0.00000163, -0.00000116, -0.00000080, -0.00000444, 0.00001061, -0.00000414, 0.00000213, -0.00000274, 0.00000373, },
		{ 0.00000795, -0.00000381, -0.00000198, -0.00000605, 0.00000708, -0.00000430, 0.00000681, -0.00000700, 0.00001028, -0.00000587, 0.00000368, -0.00001477, 0.00001287, -0.00001291, 0.00001851, -0.00001147, 0.00000493, -0.00000970, 0.00001224, -0.00002021, 0.00002369, -0.00001252, 0.00000731, -0.00001234, 0.00001089, -0.00001220, 0.00001855, -0.00002693, 0.00002526, -0.00001583, 0.00001084, -0.00002928, 0.00003952, -0.00003107, 0.00003213, -0.00005286, 0.00004734, -0.00004862, 0.00004921, -0.00006396, 0.00004563, -0.00004057, 0.00005578, -0.00007087, 0.00005440, -0.00007417, 0.00010076, -0.00015945, 0.00014284, -0.00040827, 0.00019830, 0.00086131, -0.00061759, 0.00016699, -0.00018664, 0.00011331, -0.00010525, 0.00007145, -0.00010420, 0.00008293, -0.00006123, 0.00003400, -0.00002728, 0.00002113, -0.00001840, 0.00001498, -0.00002352, 0.00003262, -0.00002882, 0.00001708, -0.00001662, 0.00002766, -0.00003559, 0.00002213, -0.00002165, 0.00002773, -0.00002436, 0.00001895, -0.00001989, 0.00001815, -0.00001834, 0.00000610, -0.00000374, 0.00001347, -0.00001269, 0.00000963, -0.00001276, 0.00000711, 0.00000144, 0.00000541, -0.00000902, 0.00001072, -0.00001566, 0.00000642, -0.00000519, 0.00000827, -0.00000691, 0.00000738, -0.00000627, 0.00000940, -0.00001114, 0.00000206, 0.00000122, 0.00000195, -0.00000414, 0.00000222, -0.00000407, 0.00000113, 0.00000584, -0.00000002, -0.00000404, 0.00000195, -0.00000238, 0.00000153, 0.00000019, -0.00000104, 0.00000031, 0.00000043, -0.00000169, -0.00000128, -0.00000047, -0.00000493, 0.00001011, -0.00000302, 0.00000142, -0.00000225, 0.00000345, -0.00000653, },
		{ -0.00000220, -0.00000283, -0.00000589, 0.00000650, -0.00000354, 0.00000613, -0.00000628, 0.00000922, -0.00000445, 0.00000313, -0.00001438, 0.00001177, -0.00001149, 0.00001612, -0.00000847, 0.00000270, -0.00000811, 0.00001074, -0.00001862, 0.00002065, -0.00000853, 0.00000433, -0.00001003, 0.00000880, -0.00001043, 0.00001687, -0.00002491, 0.00002214, -0.00001227, 0.00000907, -0.00002832, 0.00003663, -0.00002651, 0.00002799, -0.00004878, 0.00004143, -0.00004133, 0.00004069, -0.00005360, 0.00003290, -0.00002839, 0.00004386, -0.00005789, 0.00004237, -0.00006060, 0.00008519, -0.00013682, 0.00011412, -0.00035256, 0.00010605, 0.00090573, -0.00058848, 0.00013332, -0.00016503, 0.00009494, -0.00009106, 0.00005937, -0.00009218, 0.00006934, -0.00004811, 0.00002196, -0.00001787, 0.00001424, -0.00001320, 0.00001141, -0.00002097, 0.00003055, -0.00002600, 0.00001446, -0.00001566, 0.00002711, -0.00003329, 0.00001885, -0.00001907, 0.00002546, -0.00002157, 0.00001581, -0.00001674, 0.00001490, -0.00001459, 0.00000245, -0.00000180, 0.00001232, -0.00001110, 0.00000794, -0.00001071, 0.00000489, 0.00000249, 0.00000538, -0.00000891, 0.00001059, -0.00001473, 0.00000494, -0.00000442, 0.00000765, -0.00000621, 0.00000660, -0.00000551, 0.00000871, -0.00000961, 0.00000029, 0.00000208, 0.00000167, -0.00000374, 0.00000165, -0.00000336, 0.00000023, 0.00000586, 0.00000074, -0.00000441, 0.00000200, -0.00000229, 0.00000138, 0.00000029, -0.00000092, 0.00000002, 0.00000068, -0.00000169, -0.00000142, -0.00000017, -0.00000528, 0.00000941, -0.00000181, 0.00000070, -0.00000174, 0.00000311, -0.00000598, 0.00000671, },
		{ -0.00000354, -0.00000579, 0.00000588, -0.00000273, 0.00000538, -0.00000546, 0.00000805, -0.00000300, 0.00000254, -0.00001381, 0.00001038, -0.00000976, 0.00001348, -0.00000539, 0.00000050, -0.00000653, 0.00000916, -0.00001683, 0.00001740, -0.00000455, 0.00000153, -0.00000786, 0.00000680, -0.00000866, 0.00001509, -0.00002267, 0.00001879, -0.00000862, 0.00000720, -0.00002690, 0.00003303, -0.00002134, 0.00002333, -0.00004404, 0.00003471, -0.00003326, 0.00003148, -0.00004270, 0.00002003, -0.00001659, 0.00003245, -0.00004553, 0.00003095, -0.00004740, 0.00006983, -0.00011441, 0.00008595, -0.00029935, 0.00002296, 0.00093544, -0.00054797, 0.00009424, -0.00014053, 0.00007456, -0.00007530, 0.00004588, -0.00007887, 0.00005488, -0.00003465, 0.00001009, -0.00000897, 0.00000790, -0.00000847, 0.00000820, -0.00001860, 0.00002843, -0.00002304, 0.00001176, -0.00001450, 0.00002604, -0.00003035, 0.00001510, -0.00001614, 0.00002279, -0.00001834, 0.00001228, -0.00001328, 0.00001141, -0.00001073, -0.00000107, -0.00000006, 0.00001121, -0.00000946, 0.00000619, -0.00000865, 0.00000281, 0.00000334, 0.00000542, -0.00000872, 0.00001026, -0.00001357, 0.00000334, -0.00000357, 0.00000691, -0.00000537, 0.00000567, -0.00000460, 0.00000785, -0.00000796, -0.00000146, 0.00000284, 0.00000144, -0.00000332, 0.00000105, -0.00000267, -0.00000057, 0.00000574, 0.00000154, -0.00000470, 0.00000196, -0.00000213, 0.00000121, 0.00000039, -0.00000081, -0.00000025, 0.00000087, -0.00000165, -0.00000158, 0.00000009, -0.00000549, 0.00000855, -0.00000056, 0.00000001, -0.00000125, 0.00000274, -0.00000534, 0.00000540, -0.00000065, },
		{ -0.00000576, 0.00000524, -0.00000189, 0.00000460, -0.00000458, 0.00000680, -0.00000155, 0.00000193, -0.00001308, 0.00000878, -0.00000782, 0.00001069, -0.00000234, -0.00000155, -0.00000503, 0.00000758, -0.00001493, 0.00001405, -0.00000072, -0.00000099, -0.00000592, 0.00000494, -0.00000698, 0.00001327, -0.00002029, 0.00001529, -0.00000498, 0.00000531, -0.00002511, 0.00002886, -0.00001576, 0.00001839, -0.00003889, 0.00002749, -0.00002479, 0.00002202, -0.00003174, 0.00000756, -0.00000567, 0.00002203, -0.00003425, 0.00002059, -0.00003506, 0.00005525, -0.00009297, 0.00005927, -0.00024982, -0.00005008, 0.00095063, -0.00049727, 0.00005113, -0.00011426, 0.00005304, -0.00005872, 0.00003164, -0.00006492, 0.00004015, -0.00002143, -0.00000111, -0.00000097, 0.00000236, -0.00000442, 0.00000548, -0.00001650, 0.00002636, -0.00002003, 0.00000905, -0.00001319, 0.00002450, -0.00002687, 0.00001103, -0.00001299, 0.00001984, -0.00001481, 0.00000853, -0.00000969, 0.00000786, -0.00000694, -0.00000431, 0.00000139, 0.00001021, -0.00000783, 0.00000445, -0.00000665, 0.00000091, 0.00000398, 0.00000554, -0.00000845, 0.00000975, -0.00001220, 0.00000166, -0.00000270, 0.00000609, -0.00000442, 0.00000463, -0.00000360, 0.00000687, -0.00000623, -0.00000314, 0.00000346, 0.00000126, -0.00000290, 0.00000046, -0.00000203, -0.00000124, 0.00000546, 0.00000236, -0.00000491, 0.00000184, -0.00000192, 0.00000101, 0.00000049, -0.00000069, -0.00000050, 0.00000102, -0.00000155, -0.00000174, 0.00000030, -0.00000556, 0.00000756, 0.00000069, -0.00000061, -0.00000080, 0.00000235, -0.00000464, 0.00000406, 0.00000078, -0.00000408, },
		{ 0.00000458, -0.00000104, 0.00000382, -0.00000366, 0.00000553, -0.00000015, 0.00000134, -0.00001222, 0.00000701, -0.00000574, 0.00000787, 0.00000054, -0.00000337, -0.00000371, 0.00000607, -0.00001301, 0.00001075, 0.00000284, -0.00000312, -0.00000428, 0.00000330, -0.00000542, 0.00001147, -0.00001784, 0.00001177, -0.00000147, 0.00000349, -0.00002304, 0.00002429, -0.00000999, 0.00001340, -0.00003357, 0.00002006, -0.00001627, 0.00001272, -0.00002120, -0.00000400, 0.00000389, 0.00001301, -0.00002444, 0.00001165, -0.00002396, 0.00004194, -0.00007308, 0.00003484, -0.00020487, -0.00011256, 0.00095187, -0.00043787, 0.00000552, -0.00008739, 0.00003131, -0.00004208, 0.00001732, -0.00005095, 0.00002579, -0.00000901, -0.00001118, 0.00000580, -0.00000213, -0.00000118, 0.00000336, -0.00001474, 0.00002437, -0.00001703, 0.00000637, -0.00001177, 0.00002257, -0.00002298, 0.00000677, -0.00000974, 0.00001674, -0.00001112, 0.00000470, -0.00000613, 0.00000440, -0.00000337, -0.00000716, 0.00000247, 0.00000938, -0.00000628, 0.00000280, -0.00000479, -0.00000075, 0.00000438, 0.00000572, -0.00000810, 0.00000908, -0.00001068, -0.00000006, -0.00000183, 0.00000522, -0.00000339, 0.00000353, -0.00000254, 0.00000582, -0.00000448, -0.00000467, 0.00000391, 0.00000117, -0.00000250, -0.00000010, -0.00000145, -0.00000176, 0.00000505, 0.00000317, -0.00000502, 0.00000163, -0.00000166, 0.00000080, 0.00000060, -0.00000057, -0.00000071, 0.00000110, -0.00000140, -0.00000191, 0.00000046, -0.00000551, 0.00000648, 0.00000189, -0.00000114, -0.00000041, 0.00000198, -0.00000390, 0.00000276, 0.00000206, -0.00000443, -0.00000580, },
		{ -0.00000020, 0.00000305, -0.00000275, 0.00000427, 0.00000117, 0.00000081, -0.00001129, 0.00000515, -0.00000364, 0.00000514, 0.00000316, -0.00000486, -0.00000261, 0.00000469, -0.00001115, 0.00000758, 0.00000601, -0.00000479, -0.00000299, 0.00000191, -0.00000403, 0.00000975, -0.00001540, 0.00000832, 0.00000183, 0.00000180, -0.00002079, 0.00001948, -0.00000425, 0.00000859, -0.00002833, 0.00001273, -0.00000806, 0.00000396, -0.00001150, -0.00001424, 0.00001173, 0.00000567, -0.00001634, 0.00000438, -0.00001439, 0.00003026, -0.00005521, 0.00001326, -0.00016514, -0.00016430, 0.00094012, -0.00037149, -0.00004101, -0.00006108, 0.00001025, -0.00002609, 0.00000354, -0.00003755, 0.00001233, 0.00000213, -0.00001976, 0.00001110, -0.00000544, 0.00000115, 0.00000188, -0.00001334, 0.00002253, -0.00001410, 0.00000380, -0.00001027, 0.00002032, -0.00001880, 0.00000247, -0.00000653, 0.00001361, -0.00000741, 0.00000096, -0.00000275, 0.00000119, -0.00000016, -0.00000952, 0.00000313, 0.00000874, -0.00000484, 0.00000128, -0.00000312, -0.00000213, 0.00000453, 0.00000596, -0.00000766, 0.00000826, -0.00000906, -0.00000175, -0.00000101, 0.00000433, -0.00000234, 0.00000241, -0.00000148, 0.00000476, -0.00000279, -0.00000602, 0.00000416, 0.00000117, -0.00000213, -0.00000061, -0.00000096, -0.00000214, 0.00000452, 0.00000395, -0.00000502, 0.00000133, -0.00000137, 0.00000059, 0.00000069, -0.00000046, -0.00000089, 0.00000112, -0.00000122, -0.00000208, 0.00000057, -0.00000534, 0.00000533, 0.00000302, -0.00000155, -0.00000010, 0.00000163, -0.00000317, 0.00000151, 0.00000314, -0.00000458, -0.00000591, 0.00000391, },
		{ 0.00000232, -0.00000187, 0.00000307, 0.00000237, 0.00000035, -0.00001033, 0.00000326, -0.00000159, 0.00000259, 0.00000543, -0.00000597, -0.00000180, 0.00000350, -0.00000943, 0.00000464, 0.00000872, -0.00000596, -0.00000207, 0.00000079, -0.00000283, 0.00000813, -0.00001304, 0.00000502, 0.00000482, 0.00000034, -0.00001847, 0.00001461, 0.00000125, 0.00000416, -0.00002339, 0.00000574, -0.00000045, -0.00000391, -0.00000299, -0.00002282, 0.00001758, 0.00000022, -0.00001012, -0.00000108, -0.00000652, 0.00002044, -0.00003966, -0.00000510, -0.00013099, -0.00020548, 0.00091661, -0.00030001, -0.00008682, -0.00003647, -0.00000933, -0.00001141, -0.00000911, -0.00002527, 0.00000028, 0.00001160, -0.00002656, 0.00001476, -0.00000749, 0.00000253, 0.00000105, -0.00001232, 0.00002083, -0.00001128, 0.00000135, -0.00000875, 0.00001782, -0.00001447, -0.00000173, -0.00000349, 0.00001059, -0.00000382, -0.00000256, 0.00000032, -0.00000165, 0.00000259, -0.00001130, 0.00000333, 0.00000833, -0.00000356, -0.00000006, -0.00000168, -0.00000321, 0.00000444, 0.00000625, -0.00000714, 0.00000732, -0.00000736, -0.00000337, -0.00000026, 0.00000346, -0.00000129, 0.00000131, -0.00000046, 0.00000374, -0.00000121, -0.00000714, 0.00000419, 0.00000126, -0.00000180, -0.00000105, -0.00000057, -0.00000238, 0.00000388, 0.00000468, -0.00000490, 0.00000096, -0.00000105, 0.00000037, 0.00000079, -0.00000037, -0.00000102, 0.00000108, -0.00000099, -0.00000224, 0.00000062, -0.00000509, 0.00000415, 0.00000403, -0.00000181, 0.00000013, 0.00000133, -0.00000246, 0.00000038, 0.00000401, -0.00000453, -0.00000608, 0.00000324, 0.00000061, },
		{ -0.00000105, 0.00000196, 0.00000341, 0.00000000, -0.00000936, 0.00000142, 0.00000032, 0.00000031, 0.00000727, -0.00000665, -0.00000129, 0.00000253, -0.00000790, 0.00000201, 0.00001090, -0.00000660, -0.00000152, -0.00000006, -0.00000184, 0.00000665, -0.00001080, 0.00000195, 0.00000743, -0.00000085, -0.00001617, 0.00000984, 0.00000632, 0.00000030, -0.00001893, -0.00000066, 0.00000628, -0.00001062, 0.00000406, -0.00002949, 0.00002131, -0.00000330, -0.00000582, -0.00000469, -0.00000041, 0.00001259, -0.00002660, -0.00002002, -0.00010253, -0.00023658, 0.00088283, -0.00022540, -0.00013034, -0.00001457, -0.00002670, 0.00000139, -0.00002016, -0.00001456, -0.00000998, 0.00001909, -0.00003139, 0.00001673, -0.00000828, 0.00000299, 0.00000084, -0.00001164, 0.00001929, -0.00000859, -0.00000092, -0.00000724, 0.00001516, -0.00001012, -0.00000568, -0.00000074, 0.00000779, -0.00000048, -0.00000571, 0.00000294, -0.00000401, 0.00000477, -0.00001246, 0.00000307, 0.00000815, -0.00000245, -0.00000118, -0.00000050, -0.00000398, 0.00000413, 0.00000657, -0.00000654, 0.00000628, -0.00000566, -0.00000488, 0.00000037, 0.00000263, -0.00000028, 0.00000028, 0.00000048, 0.00000280, 0.00000022, -0.00000799, 0.00000400, 0.00000144, -0.00000153, -0.00000142, -0.00000029, -0.00000247, 0.00000316, 0.00000534, -0.00000467, 0.00000053, -0.00000072, 0.00000016, 0.00000088, -0.00000028, -0.00000110, 0.00000098, -0.00000075, -0.00000238, 0.00000063, -0.00000476, 0.00000297, 0.00000489, -0.00000192, 0.00000026, 0.00000107, -0.00000179, -0.00000063, 0.00000465, -0.00000428, -0.00000631, 0.00000258, 0.00000138, 0.00000166, },
		{ 0.00000097, 0.00000427, -0.00000022, -0.00000843, -0.00000032, 0.00000202, -0.00000162, 0.00000864, -0.00000689, -0.00000109, 0.00000178, -0.00000658, -0.00000027, 0.00001253, -0.00000673, -0.00000131, -0.00000065, -0.00000104, 0.00000534, -0.00000873, -0.00000081, 0.00000960, -0.00000172, -0.00001397, 0.00000531, 0.00001078, -0.00000284, -0.00001512, -0.00000629, 0.00001193, -0.00001597, 0.00000946, -0.00003413, 0.00002289, -0.00000491, -0.00000337, -0.00000653, 0.00000399, 0.00000670, -0.00001605, -0.00003150, -0.00007964, -0.00025833, 0.00084047, -0.00014965, -0.00017012, 0.00000373, -0.00004127, 0.00001185, -0.00002921, -0.00000574, -0.00001815, 0.00002441, -0.00003420, 0.00001702, -0.00000787, 0.00000262, 0.00000117, -0.00001125, 0.00001789, -0.00000605, -0.00000300, -0.00000578, 0.00001243, -0.00000587, -0.00000927, 0.00000163, 0.00000531, 0.00000250, -0.00000840, 0.00000502, -0.00000581, 0.00000635, -0.00001299, 0.00000237, 0.00000817, -0.00000153, -0.00000209, 0.00000042, -0.00000445, 0.00000363, 0.00000688, -0.00000585, 0.00000519, -0.00000399, -0.00000623, 0.00000086, 0.00000187, 0.00000066, -0.00000065, 0.00000130, 0.00000198, 0.00000147, -0.00000857, 0.00000360, 0.00000172, -0.00000131, -0.00000170, -0.00000012, -0.00000243, 0.00000239, 0.00000590, -0.00000431, 0.00000005, -0.00000038, -0.00000003, 0.00000096, -0.00000021, -0.00000114, 0.00000084, -0.00000048, -0.00000250, 0.00000060, -0.00000437, 0.00000183, 0.00000559, -0.00000188, 0.00000030, 0.00000087, -0.00000118, -0.00000149, 0.00000505, -0.00000386, -0.00000658, 0.00000191, 0.00000208, 0.00000109, -0.00000031, },
		{ 0.00000494, -0.00000031, -0.00000756, -0.00000191, 0.00000344, -0.00000316, 0.00000951, -0.00000669, -0.00000118, 0.00000127, -0.00000549, -0.00000217, 0.00001361, -0.00000637, -0.00000141, -0.00000101, -0.00000044, 0.00000419, -0.00000688, -0.00000322, 0.00001128, -0.00000222, -0.00001195, 0.00000115, 0.00001452, -0.00000515, -0.00001206, -0.00001102, 0.00001635, -0.00001985, 0.00001313, -0.00003673, 0.00002240, -0.00000478, -0.00000262, -0.00000676, 0.00000683, 0.00000267, -0.00000790, -0.00003966, -0.00006200, -0.00027168, 0.00079130, -0.00007469, -0.00020485, 0.00001774, -0.00005260, 0.00001967, -0.00003601, 0.00000094, -0.00002408, 0.00002750, -0.00003499, 0.00001574, -0.00000640, 0.00000153, 0.00000194, -0.00001107, 0.00001659, -0.00000367, -0.00000487, -0.00000441, 0.00000970, -0.00000184, -0.00001239, 0.00000352, 0.00000323, 0.00000504, -0.00001053, 0.00000650, -0.00000701, 0.00000730, -0.00001291, 0.00000127, 0.00000837, -0.00000077, -0.00000277, 0.00000107, -0.00000463, 0.00000296, 0.00000716, -0.00000508, 0.00000405, -0.00000239, -0.00000739, 0.00000119, 0.00000121, 0.00000148, -0.00000144, 0.00000196, 0.00000131, 0.00000249, -0.00000886, 0.00000299, 0.00000206, -0.00000113, -0.00000189, -0.00000007, -0.00000228, 0.00000158, 0.00000636, -0.00000383, -0.00000048, -0.00000006, -0.00000021, 0.00000104, -0.00000016, -0.00000115, 0.00000065, -0.00000021, -0.00000259, 0.00000052, -0.00000395, 0.00000076, 0.00000611, -0.00000168, 0.00000026, 0.00000072, -0.00000065, -0.00000218, 0.00000522, -0.00000328, -0.00000686, 0.00000125, 0.00000271, 0.00000062, 0.00000032, 0.00000012, },
		{ -0.00000026, -0.00000679, -0.00000330, 0.00000456, -0.00000428, 0.00000988, -0.00000607, -0.00000155, 0.00000096, -0.00000465, -0.00000369, 0.00001415, -0.00000558, -0.00000176, -0.00000117, -0.00000001, 0.00000321, -0.00000527, -0.00000525, 0.00001246, -0.00000235, -0.00001016, -0.00000255, 0.00001743, -0.00000657, -0.00000982, -0.00001477, 0.00001948, -0.00002219, 0.00001507, -0.00003736, 0.00002002, -0.00000315, -0.00000332, -0.00000562, 0.00000831, 0.00000030, -0.00000194, -0.00004476, -0.00004913, -0.00027770, 0.00073715, -0.00000234, -0.00023345, 0.00002696, -0.00006045, 0.00002467, -0.00004044, 0.00000538, -0.00002771, 0.00002838, -0.00003391, 0.00001308, -0.00000406, -0.00000012, 0.00000302, -0.00001103, 0.00001537, -0.00000144, -0.00000652, -0.00000314, 0.00000704, 0.00000186, -0.00001494, 0.00000487, 0.00000161, 0.00000706, -0.00001205, 0.00000735, -0.00000760, 0.00000764, -0.00001226, -0.00000017, 0.00000871, -0.00000016, -0.00000324, 0.00000147, -0.00000457, 0.00000218, 0.00000739, -0.00000423, 0.00000292, -0.00000091, -0.00000833, 0.00000135, 0.00000065, 0.00000219, -0.00000208, 0.00000245, 0.00000082, 0.00000329, -0.00000886, 0.00000220, 0.00000246, -0.00000100, -0.00000200, -0.00000010, -0.00000203, 0.00000079, 0.00000668, -0.00000325, -0.00000102, 0.00000024, -0.00000036, 0.00000111, -0.00000012, -0.00000111, 0.00000043, 0.00000006, -0.00000265, 0.00000040, -0.00000350, -0.00000024, 0.00000643, -0.00000134, 0.00000015, 0.00000062, -0.00000019, -0.00000271, 0.00000517, -0.00000257, -0.00000714, 0.00000059, 0.00000326, 0.00000025, 0.00000084, -0.00000058, 0.00000540, },
		{ -0.00000612, -0.00000448, 0.00000533, -0.00000497, 0.00000977, -0.00000510, -0.00000212, 0.00000084, -0.00000402, -0.00000483, 0.00001420, -0.00000442, -0.00000231, -0.00000119, 0.00000027, 0.00000240, -0.00000391, -0.00000686, 0.00001312, -0.00000211, -0.00000863, -0.00000570, 0.00001945, -0.00000709, -0.00000839, -0.00001751, 0.00002130, -0.00002305, 0.00001537, -0.00003619, 0.00001603, -0.00000032, -0.00000517, -0.00000340, 0.00000873, -0.00000066, 0.00000210, -0.00004719, -0.00004043, -0.00027760, 0.00067982, 0.00006576, -0.00025508, 0.00003114, -0.00006474, 0.00002683, -0.00004248, 0.00000760, -0.00002909, 0.00002721, -0.00003118, 0.00000929, -0.00000108, -0.00000213, 0.00000426, -0.00001103, 0.00001418, 0.00000064, -0.00000793, -0.00000203, 0.00000453, 0.00000514, -0.00001687, 0.00000565, 0.00000048, 0.00000854, -0.00001295, 0.00000756, -0.00000760, 0.00000740, -0.00001110, -0.00000186, 0.00000912, 0.00000032, -0.00000352, 0.00000166, -0.00000430, 0.00000133, 0.00000754, -0.00000331, 0.00000181, 0.00000042, -0.00000903, 0.00000133, 0.00000021, 0.00000275, -0.00000255, 0.00000275, 0.00000051, 0.00000385, -0.00000859, 0.00000125, 0.00000289, -0.00000091, -0.00000204, -0.00000023, -0.00000172, 0.00000001, 0.00000687, -0.00000256, -0.00000157, 0.00000050, -0.00000048, 0.00000117, -0.00000009, -0.00000104, 0.00000019, 0.00000032, -0.00000267, 0.00000025, -0.00000304, -0.00000113, 0.00000657, -0.00000086, -0.00000003, 0.00000056, 0.00000019, -0.00000307, 0.00000492, -0.00000177, -0.00000739, -0.00000007, 0.00000372, 0.00000001, 0.00000124, -0.00000112, 0.00000565, -0.00000006, },
		{ -0.00000541, 0.00000576, -0.00000525, 0.00000924, -0.00000382, -0.00000286, 0.00000084, -0.00000359, -0.00000563, 0.00001381, -0.00000299, -0.00000296, -0.00000111, 0.00000043, 0.00000176, -0.00000282, -0.00000806, 0.00001328, -0.00000151, -0.00000738, -0.00000827, 0.00002057, -0.00000672, -0.00000775, -0.00001928, 0.00002188, -0.00002253, 0.00001421, -0.00003349, 0.00001078, 0.00000334, -0.00000783, -0.00000044, 0.00000837, -0.00000054, 0.00000457, -0.00004737, -0.00003521, -0.00027260, 0.00062104, 0.00012821, -0.00026919, 0.00003024, -0.00006558, 0.00002628, -0.00004227, 0.00000772, -0.00002842, 0.00002423, -0.00002708, 0.00000466, 0.00000228, -0.00000430, 0.00000552, -0.00001098, 0.00001297, 0.00000258, -0.00000911, -0.00000107, 0.00000223, 0.00000793, -0.00001812, 0.00000585, -0.00000015, 0.00000946, -0.00001322, 0.00000716, -0.00000706, 0.00000666, -0.00000954, -0.00000371, 0.00000954, 0.00000072, -0.00000364, 0.00000166, -0.00000386, 0.00000045, 0.00000758, -0.00000233, 0.00000076, 0.00000157, -0.00000948, 0.00000113, -0.00000011, 0.00000316, -0.00000283, 0.00000286, 0.00000038, 0.00000418, -0.00000808, 0.00000020, 0.00000331, -0.00000082, -0.00000202, -0.00000042, -0.00000136, -0.00000071, 0.00000690, -0.00000179, -0.00000210, 0.00000072, -0.00000057, 0.00000123, -0.00000007, -0.00000094, -0.00000007, 0.00000055, -0.00000266, 0.00000007, -0.00000258, -0.00000191, 0.00000652, -0.00000028, -0.00000023, 0.00000054, 0.00000050, -0.00000328, 0.00000451, -0.00000092, -0.00000758, -0.00000073, 0.00000408, -0.00000011, 0.00000151, -0.00000149, 0.00000571, 0.00000027, -0.00000556, },
		{ 0.00000586, -0.00000514, 0.00000833, -0.00000232, -0.00000368, 0.00000094, -0.00000332, -0.00000613, 0.00001305, -0.00000136, -0.00000366, -0.00000098, 0.00000050, 0.00000126, -0.00000198, -0.00000884, 0.00001298, -0.00000060, -0.00000641, -0.00001023, 0.00002082, -0.00000555, -0.00000781, -0.00002017, 0.00002133, -0.00002079, 0.00001180, -0.00002956, 0.00000465, 0.00000744, -0.00001095, 0.00000293, 0.00000758, 0.00000037, 0.00000585, -0.00004578, -0.00003276, -0.00026389, 0.00056237, 0.00018387, -0.00027549, 0.00002447, -0.00006324, 0.00002329, -0.00004003, 0.00000597, -0.00002594, 0.00001975, -0.00002197, -0.00000045, 0.00000575, -0.00000643, 0.00000665, -0.00001080, 0.00001173, 0.00000438, -0.00001006, -0.00000029, 0.00000018, 0.00001019, -0.00001870, 0.00000549, -0.00000029, 0.00000984, -0.00001289, 0.00000622, -0.00000605, 0.00000551, -0.00000769, -0.00000562, 0.00000991, 0.00000106, -0.00000363, 0.00000152, -0.00000332, -0.00000040, 0.00000751, -0.00000130, -0.00000021, 0.00000252, -0.00000967, 0.00000077, -0.00000032, 0.00000342, -0.00000293, 0.00000279, 0.00000042, 0.00000430, -0.00000736, -0.00000093, 0.00000371, -0.00000075, -0.00000195, -0.00000066, -0.00000098, -0.00000135, 0.00000679, -0.00000095, -0.00000259, 0.00000088, -0.00000063, 0.00000127, -0.00000006, -0.00000083, -0.00000032, 0.00000074, -0.00000260, -0.00000012, -0.00000215, -0.00000257, 0.00000629, 0.00000039, -0.00000045, 0.00000053, 0.00000075, -0.00000334, 0.00000396, -0.00000005, -0.00000771, -0.00000139, 0.00000434, -0.00000012, 0.00000167, -0.00000170, 0.00000557, 0.00000072, -0.00000511, -0.00000609, },
		{ -0.00000471, 0.00000713, -0.00000070, -0.00000451, 0.00000107, -0.00000317, -0.00000637, 0.00001201, 0.00000036, -0.00000433, -0.00000085, 0.00000052, 0.00000089, -0.00000140, -0.00000923, 0.00001225, 0.00000058, -0.00000568, -0.00001162, 0.00002025, -0.00000368, -0.00000846, -0.00002028, 0.00001980, -0.00001807, 0.00000845, -0.00002475, -0.00000195, 0.00001159, -0.00001419, 0.00000637, 0.00000666, 0.00000172, 0.00000631, -0.00004290, -0.00003238, -0.00025264, 0.00050520, 0.00023187, -0.00027401, 0.00001425, -0.00005815, 0.00001821, -0.00003609, 0.00000265, -0.00002199, 0.00001415, -0.00001622, -0.00000573, 0.00000906, -0.00000834, 0.00000753, -0.00001040, 0.00001042, 0.00000605, -0.00001077, 0.00000029, -0.00000157, 0.00001187, -0.00001861, 0.00000459, 0.00000001, 0.00000973, -0.00001204, 0.00000482, -0.00000469, 0.00000406, -0.00000565, -0.00000748, 0.00001016, 0.00000140, -0.00000354, 0.00000127, -0.00000270, -0.00000120, 0.00000730, -0.00000024, -0.00000107, 0.00000325, -0.00000962, 0.00000025, -0.00000042, 0.00000353, -0.00000286, 0.00000255, 0.00000062, 0.00000423, -0.00000648, -0.00000208, 0.00000406, -0.00000066, -0.00000184, -0.00000093, -0.00000060, -0.00000190, 0.00000654, -0.00000006, -0.00000304, 0.00000098, -0.00000065, 0.00000131, -0.00000005, -0.00000071, -0.00000056, 0.00000090, -0.00000249, -0.00000034, -0.00000174, -0.00000310, 0.00000591, 0.00000111, -0.00000066, 0.00000053, 0.00000093, -0.00000328, 0.00000331, 0.00000081, -0.00000776, -0.00000205, 0.00000450, -0.00000000, 0.00000171, -0.00000176, 0.00000528, 0.00000127, -0.00000477, -0.00000654, 0.00000564, },
		{ 0.00000572, 0.00000097, -0.00000528, 0.00000118, -0.00000309, -0.00000641, 0.00001075, 0.00000210, -0.00000490, -0.00000075, 0.00000050, 0.00000064, -0.00000105, -0.00000927, 0.00001115, 0.00000196, -0.00000518, -0.00001245, 0.00001897, -0.00000125, -0.00000955, -0.00001977, 0.00001748, -0.00001460, 0.00000445, -0.00001945, -0.00000860, 0.00001542, -0.00001722, 0.00000959, 0.00000590, 0.00000323, 0.00000631, -0.00003920, -0.00003338, -0.00023988, 0.00045072, 0.00027168, -0.00026505, 0.00000021, -0.00005087, 0.00001152, -0.00003086, -0.00000186, -0.00001695, 0.00000783, -0.00001023, -0.00001082, 0.00001197, -0.00000985, 0.00000805, -0.00000975, 0.00000904, 0.00000758, -0.00001125, 0.00000068, -0.00000301, 0.00001298, -0.00001789, 0.00000324, 0.00000070, 0.00000919, -0.00001074, 0.00000306, -0.00000307, 0.00000242, -0.00000355, -0.00000921, 0.00001024, 0.00000177, -0.00000340, 0.00000097, -0.00000207, -0.00000189, 0.00000695, 0.00000082, -0.00000179, 0.00000375, -0.00000934, -0.00000039, -0.00000044, 0.00000351, -0.00000263, 0.00000216, 0.00000094, 0.00000402, -0.00000547, -0.00000321, 0.00000431, -0.00000055, -0.00000171, -0.00000120, -0.00000025, -0.00000234, 0.00000614, 0.00000085, -0.00000341, 0.00000101, -0.00000065, 0.00000134, -0.00000004, -0.00000058, -0.00000078, 0.00000100, -0.00000235, -0.00000056, -0.00000136, -0.00000351, 0.00000539, 0.00000185, -0.00000085, 0.00000052, 0.00000107, -0.00000312, 0.00000259, 0.00000160, -0.00000770, -0.00000271, 0.00000455, 0.00000021, 0.00000165, -0.00000170, 0.00000484, 0.00000189, -0.00000452, -0.00000676, 0.00000515, -0.00000400, },
		{ 0.00000259, -0.00000592, 0.00000122, -0.00000305, -0.00000630, 0.00000938, 0.00000376, -0.00000532, -0.00000071, 0.00000046, 0.00000049, -0.00000092, -0.00000899, 0.00000977, 0.00000348, -0.00000483, -0.00001281, 0.00001708, 0.00000157, -0.00001091, -0.00001880, 0.00001459, -0.00001066, 0.00000012, -0.00001401, -0.00001494, 0.00001861, -0.00001978, 0.00001231, 0.00000554, 0.00000465, 0.00000615, -0.00003509, -0.00003515, -0.00022654, 0.00039985, 0.00030304, -0.00024918, -0.00001689, -0.00004204, 0.00000373, -0.00002477, -0.00000715, -0.00001124, 0.00000121, -0.00000439, -0.00001542, 0.00001427, -0.00001082, 0.00000812, -0.00000881, 0.00000757, 0.00000897, -0.00001150, 0.00000086, -0.00000413, 0.00001353, -0.00001659, 0.00000150, 0.00000169, 0.00000829, -0.00000910, 0.00000105, -0.00000133, 0.00000073, -0.00000151, -0.00001070, 0.00001010, 0.00000218, -0.00000323, 0.00000065, -0.00000147, -0.00000246, 0.00000647, 0.00000187, -0.00000236, 0.00000403, -0.00000885, -0.00000112, -0.00000038, 0.00000336, -0.00000227, 0.00000165, 0.00000136, 0.00000369, -0.00000438, -0.00000428, 0.00000446, -0.00000040, -0.00000158, -0.00000147, 0.00000005, -0.00000266, 0.00000562, 0.00000177, -0.00000371, 0.00000097, -0.00000061, 0.00000135, -0.00000003, -0.00000047, -0.00000097, 0.00000105, -0.00000217, -0.00000080, -0.00000101, -0.00000381, 0.00000477, 0.00000258, -0.00000098, 0.00000049, 0.00000117, -0.00000288, 0.00000184, 0.00000232, -0.00000753, -0.00000335, 0.00000450, 0.00000052, 0.00000152, -0.00000153, 0.00000429, 0.00000256, -0.00000433, -0.00000679, 0.00000444, -0.00000310, 0.00000420, },
		{ -0.00000637, 0.00000116, -0.00000302, -0.00000610, 0.00000795, 0.00000527, -0.00000554, -0.00000076, 0.00000043, 0.00000043, -0.00000097, -0.00000845, 0.00000816, 0.00000504, -0.00000459, -0.00001276, 0.00001473, 0.00000461, -0.00001237, -0.00001757, 0.00001136, -0.00000652, -0.00000423, -0.00000878, -0.00002060, 0.00002088, -0.00002167, 0.00001436, 0.00000578, 0.00000575, 0.00000611, -0.00003092, -0.00003718, -0.00021338, 0.00035329, 0.00032601, -0.00022717, -0.00003613, -0.00003237, -0.00000461, -0.00001828, -0.00001282, -0.00000527, -0.00000530, 0.00000096, -0.00001927, 0.00001578, -0.00001118, 0.00000772, -0.00000759, 0.00000605, 0.00001020, -0.00001153, 0.00000085, -0.00000492, 0.00001355, -0.00001481, -0.00000052, 0.00000290, 0.00000714, -0.00000721, -0.00000107, 0.00000042, -0.00000091, 0.00000036, -0.00001188, 0.00000971, 0.00000267, -0.00000307, 0.00000034, -0.00000092, -0.00000288, 0.00000586, 0.00000288, -0.00000275, 0.00000409, -0.00000817, -0.00000193, -0.00000027, 0.00000311, -0.00000179, 0.00000106, 0.00000184, 0.00000329, -0.00000327, -0.00000524, 0.00000448, -0.00000020, -0.00000145, -0.00000171, 0.00000029, -0.00000285, 0.00000499, 0.00000266, -0.00000390, 0.00000087, -0.00000055, 0.00000136, 0.00000000, -0.00000036, -0.00000113, 0.00000105, -0.00000195, -0.00000103, -0.00000071, -0.00000400, 0.00000406, 0.00000328, -0.00000104, 0.00000044, 0.00000124, -0.00000257, 0.00000109, 0.00000293, -0.00000725, -0.00000398, 0.00000434, 0.00000091, 0.00000133, -0.00000127, 0.00000367, 0.00000323, -0.00000420, -0.00000666, 0.00000357, -0.00000208, 0.00000265, 0.00000407, },
		{ 0.00000097, -0.00000295, -0.00000584, 0.00000654, 0.00000658, -0.00000555, -0.00000091, 0.00000040, 0.00000044, -0.00000118, -0.00000770, 0.00000642, 0.00000656, -0.00000439, -0.00001239, 0.00001206, 0.00000769, -0.00001377, -0.00001624, 0.00000799, -0.00000244, -0.00000830, -0.00000407, -0.00002532, 0.00002204, -0.00002274, 0.00001560, 0.00000672, 0.00000641, 0.00000639, -0.00002698, -0.00003904, -0.00020099, 0.00031148, 0.00034089, -0.00020000, -0.00005656, -0.00002257, -0.00001295, -0.00001184, -0.00001846, 0.00000057, -0.00001133, 0.00000550, -0.00002215, 0.00001640, -0.00001085, 0.00000683, -0.00000609, 0.00000450, 0.00001126, -0.00001133, 0.00000065, -0.00000540, 0.00001308, -0.00001263, -0.00000273, 0.00000422, 0.00000584, -0.00000521, -0.00000318, 0.00000206, -0.00000239, 0.00000198, -0.00001269, 0.00000904, 0.00000323, -0.00000293, 0.00000007, -0.00000045, -0.00000315, 0.00000514, 0.00000383, -0.00000297, 0.00000395, -0.00000735, -0.00000276, -0.00000014, 0.00000279, -0.00000124, 0.00000042, 0.00000235, 0.00000285, -0.00000218, -0.00000606, 0.00000436, 0.00000005, -0.00000134, -0.00000192, 0.00000047, -0.00000293, 0.00000427, 0.00000350, -0.00000399, 0.00000069, -0.00000046, 0.00000135, 0.00000004, -0.00000028, -0.00000123, 0.00000099, -0.00000169, -0.00000126, -0.00000045, -0.00000409, 0.00000329, 0.00000392, -0.00000102, 0.00000036, 0.00000128, -0.00000222, 0.00000037, 0.00000341, -0.00000687, -0.00000458, 0.00000409, 0.00000135, 0.00000110, -0.00000097, 0.00000300, 0.00000389, -0.00000408, -0.00000640, 0.00000259, -0.00000102, 0.00000117, 0.00000534, -0.00000658, },
		{ -0.00000284, -0.00000558, 0.00000520, 0.00000764, -0.00000532, -0.00000116, 0.00000039, 0.00000051, -0.00000153, -0.00000680, 0.00000464, 0.00000798, -0.00000416, -0.00001180, 0.00000922, 0.00001064, -0.00001495, -0.00001496, 0.00000469, 0.00000134, -0.00001185, -0.00000011, -0.00002888, 0.00002198, -0.00002295, 0.00001596, 0.00000845, 0.00000655, 0.00000712, -0.00002347, -0.00004043, -0.00018981, 0.00027462, 0.00034823, -0.00016876, -0.00007715, -0.00001336, -0.00002077, -0.00000588, -0.00002369, 0.00000592, -0.00001656, 0.00000898, -0.00002392, 0.00001606, -0.00000986, 0.00000548, -0.00000437, 0.00000297, 0.00001213, -0.00001092, 0.00000027, -0.00000560, 0.00001219, -0.00001016, -0.00000500, 0.00000557, 0.00000448, -0.00000319, -0.00000517, 0.00000347, -0.00000362, 0.00000327, -0.00001310, 0.00000809, 0.00000388, -0.00000284, -0.00000015, -0.00000008, -0.00000326, 0.00000433, 0.00000470, -0.00000301, 0.00000362, -0.00000643, -0.00000359, -0.00000001, 0.00000241, -0.00000064, -0.00000024, 0.00000284, 0.00000242, -0.00000114, -0.00000671, 0.00000408, 0.00000035, -0.00000125, -0.00000207, 0.00000056, -0.00000289, 0.00000348, 0.00000428, -0.00000396, 0.00000046, -0.00000037, 0.00000133, 0.00000010, -0.00000022, -0.00000130, 0.00000087, -0.00000142, -0.00000147, -0.00000022, -0.00000410, 0.00000248, 0.00000446, -0.00000091, 0.00000024, 0.00000132, -0.00000185, -0.00000030, 0.00000375, -0.00000639, -0.00000514, 0.00000375, 0.00000183, 0.00000087, -0.00000063, 0.00000232, 0.00000450, -0.00000396, -0.00000606, 0.00000157, 0.00000001, -0.00000018, 0.00000634, -0.00000653, 0.00000063, },
		{ -0.00000533, 0.00000397, 0.00000843, -0.00000486, -0.00000149, 0.00000038, 0.00000062, -0.00000196, -0.00000582, 0.00000288, 0.00000920, -0.00000385, -0.00001107, 0.00000636, 0.00001330, -0.00001577, -0.00001388, 0.00000162, 0.00000463, -0.00001469, 0.00000290, -0.00003116, 0.00002067, -0.00002231, 0.00001545, 0.00001093, 0.00000614, 0.00000837, -0.00002051, -0.00004113, -0.00018012, 0.00024270, 0.00034877, -0.00013465, -0.00009692, -0.00000536, -0.00002763, -0.00000074, -0.00002821, 0.00001048, -0.00002073, 0.00001124, -0.00002449, 0.00001478, -0.00000823, 0.00000372, -0.00000250, 0.00000149, 0.00001279, -0.00001030, -0.00000025, -0.00000554, 0.00001096, -0.00000751, -0.00000722, 0.00000683, 0.00000315, -0.00000128, -0.00000693, 0.00000457, -0.00000452, 0.00000419, -0.00001310, 0.00000689, 0.00000459, -0.00000278, -0.00000030, 0.00000019, -0.00000323, 0.00000346, 0.00000545, -0.00000286, 0.00000314, -0.00000543, -0.00000437, 0.00000009, 0.00000201, -0.00000002, -0.00000088, 0.00000328, 0.00000203, -0.00000020, -0.00000718, 0.00000365, 0.00000069, -0.00000119, -0.00000218, 0.00000057, -0.00000275, 0.00000265, 0.00000497, -0.00000382, 0.00000017, -0.00000026, 0.00000131, 0.00000018, -0.00000019, -0.00000131, 0.00000071, -0.00000112, -0.00000168, -0.00000004, -0.00000404, 0.00000167, 0.00000491, -0.00000071, 0.00000009, 0.00000134, -0.00000147, -0.00000089, 0.00000395, -0.00000582, -0.00000565, 0.00000332, 0.00000233, 0.00000064, -0.00000030, 0.00000166, 0.00000503, -0.00000382, -0.00000568, 0.00000056, 0.00000094, -0.00000132, 0.00000702, -0.00000619, 0.00000014, -0.00000267, },
		{ 0.00000288, 0.00000893, -0.00000419, -0.00000188, 0.00000038, 0.00000078, -0.00000246, -0.00000481, 0.00000124, 0.00001017, -0.00000339, -0.00001029, 0.00000361, 0.00001554, -0.00001613, -0.00001311, -0.00000107, 0.00000728, -0.00001667, 0.00000485, -0.00003213, 0.00001818, -0.00002091, 0.00001414, 0.00001413, 0.00000523, 0.00001012, -0.00001813, -0.00004106, -0.00017203, 0.00021554, 0.00034338, -0.00009889, -0.00011495, 0.00000087, -0.00003315, 0.00000329, -0.00003176, 0.00001403, -0.00002366, 0.00001217, -0.00002388, 0.00001261, -0.00000605, 0.00000166, -0.00000055, 0.00000014, 0.00001323, -0.00000949, -0.00000089, -0.00000528, 0.00000947, -0.00000480, -0.00000930, 0.00000794, 0.00000196, 0.00000044, -0.00000837, 0.00000530, -0.00000504, 0.00000470, -0.00001269, 0.00000546, 0.00000535, -0.00000275, -0.00000038, 0.00000035, -0.00000306, 0.00000255, 0.00000607, -0.00000253, 0.00000253, -0.00000442, -0.00000509, 0.00000015, 0.00000159, 0.00000058, -0.00000147, 0.00000364, 0.00000170, 0.00000061, -0.00000743, 0.00000308, 0.00000108, -0.00000115, -0.00000222, 0.00000051, -0.00000252, 0.00000180, 0.00000554, -0.00000357, -0.00000016, -0.00000014, 0.00000127, 0.00000028, -0.00000018, -0.00000128, 0.00000050, -0.00000082, -0.00000186, 0.00000010, -0.00000391, 0.00000088, 0.00000523, -0.00000041, -0.00000008, 0.00000136, -0.00000110, -0.00000141, 0.00000401, -0.00000518, -0.00000610, 0.00000282, 0.00000282, 0.00000045, 0.00000002, 0.00000105, 0.00000546, -0.00000362, -0.00000529, -0.00000038, 0.00000171, -0.00000222, 0.00000737, -0.00000558, -0.00000049, -0.00000245, -0.00000511, },
		{ 0.00000915, -0.00000333, -0.00000231, 0.00000036, 0.00000096, -0.00000300, -0.00000383, -0.00000023, 0.00001083, -0.00000275, -0.00000954, 0.00000108, 0.00001726, -0.00001596, -0.00001269, -0.00000330, 0.00000919, -0.00001772, 0.00000571, -0.00003181, 0.00001463, -0.00001889, 0.00001214, 0.00001791, 0.00000391, 0.00001234, -0.00001631, -0.00004020, -0.00016555, 0.00019278, 0.00033307, -0.00006268, -0.00013040, 0.00000488, -0.00003707, 0.00000600, -0.00003417, 0.00001641, -0.00002525, 0.00001176, -0.00002213, 0.00000967, -0.00000345, -0.00000061, 0.00000139, -0.00000103, 0.00001341, -0.00000851, -0.00000163, -0.00000485, 0.00000780, -0.00000213, -0.00001114, 0.00000880, 0.00000098, 0.00000189, -0.00000942, 0.00000560, -0.00000518, 0.00000481, -0.00001190, 0.00000384, 0.00000612, -0.00000275, -0.00000040, 0.00000041, -0.00000278, 0.00000164, 0.00000653, -0.00000204, 0.00000183, -0.00000342, -0.00000569, 0.00000015, 0.00000119, 0.00000114, -0.00000197, 0.00000390, 0.00000147, 0.00000129, -0.00000749, 0.00000238, 0.00000150, -0.00000113, -0.00000221, 0.00000038, -0.00000222, 0.00000096, 0.00000599, -0.00000320, -0.00000052, -0.00000004, 0.00000122, 0.00000041, -0.00000021, -0.00000121, 0.00000026, -0.00000052, -0.00000202, 0.00000021, -0.00000373, 0.00000011, 0.00000543, -0.00000003, -0.00000027, 0.00000137, -0.00000073, -0.00000182, 0.00000394, -0.00000449, -0.00000647, 0.00000225, 0.00000329, 0.00000032, 0.00000029, 0.00000052, 0.00000577, -0.00000336, -0.00000492, -0.00000120, 0.00000228, -0.00000283, 0.00000737, -0.00000471, -0.00000122, -0.00000220, -0.00000493, 0.00000195, },
		{ -0.00000233, -0.00000275, 0.00000033, 0.00000117, -0.00000353, -0.00000293, -0.00000146, 0.00001114, -0.00000189, -0.00000887, -0.00000114, 0.00001839, -0.00001524, -0.00001268, -0.00000502, 0.00001031, -0.00001782, 0.00000552, -0.00003032, 0.00001021, -0.00001645, 0.00000960, 0.00002214, 0.00000232, 0.00001491, -0.00001498, -0.00003862, -0.00016056, 0.00017397, 0.00031885, -0.00002717, -0.00014258, 0.00000639, -0.00003923, 0.00000730, -0.00003537, 0.00001756, -0.00002547, 0.00001008, -0.00001939, 0.00000613, -0.00000057, -0.00000295, 0.00000320, -0.00000196, 0.00001334, -0.00000737, -0.00000244, -0.00000430, 0.00000603, 0.00000040, -0.00001265, 0.00000937, 0.00000025, 0.00000301, -0.00001004, 0.00000546, -0.00000492, 0.00000454, -0.00001077, 0.00000211, 0.00000686, -0.00000276, -0.00000036, 0.00000039, -0.00000242, 0.00000076, 0.00000684, -0.00000139, 0.00000108, -0.00000248, -0.00000617, 0.00000008, 0.00000083, 0.00000164, -0.00000237, 0.00000404, 0.00000134, 0.00000182, -0.00000734, 0.00000156, 0.00000193, -0.00000113, -0.00000215, 0.00000019, -0.00000186, 0.00000015, 0.00000631, -0.00000272, -0.00000090, 0.00000006, 0.00000116, 0.00000056, -0.00000026, -0.00000110, -0.00000000, -0.00000022, -0.00000215, 0.00000028, -0.00000352, -0.00000060, 0.00000551, 0.00000042, -0.00000047, 0.00000137, -0.00000039, -0.00000213, 0.00000374, -0.00000377, -0.00000677, 0.00000164, 0.00000371, 0.00000026, 0.00000051, 0.00000007, 0.00000595, -0.00000302, -0.00000460, -0.00000187, 0.00000262, -0.00000315, 0.00000705, -0.00000361, -0.00000203, -0.00000192, -0.00000479, 0.00000118, 0.00000911, },
		{ -0.00000317, 0.00000028, 0.00000139, -0.00000403, -0.00000216, -0.00000241, 0.00001108, -0.00000081, -0.00000833, -0.00000299, 0.00001889, -0.00001396, -0.00001306, -0.00000621, 0.00001065, -0.00001703, 0.00000437, -0.00002784, 0.00000519, -0.00001383, 0.00000670, 0.00002664, 0.00000060, 0.00001772, -0.00001400, -0.00003644, -0.00015687, 0.00015858, 0.00030176, 0.00000662, -0.00015096, 0.00000522, -0.00003960, 0.00000719, -0.00003536, 0.00001749, -0.00002439, 0.00000724, -0.00001584, 0.00000216, 0.00000241, -0.00000524, 0.00000481, -0.00000260, 0.00001301, -0.00000610, -0.00000326, -0.00000370, 0.00000426, 0.00000270, -0.00001378, 0.00000959, -0.00000017, 0.00000375, -0.00001021, 0.00000489, -0.00000431, 0.00000393, -0.00000939, 0.00000031, 0.00000754, -0.00000276, -0.00000029, 0.00000030, -0.00000200, -0.00000006, 0.00000697, -0.00000062, 0.00000031, -0.00000163, -0.00000651, -0.00000008, 0.00000050, 0.00000205, -0.00000265, 0.00000405, 0.00000132, 0.00000219, -0.00000700, 0.00000065, 0.00000235, -0.00000114, -0.00000205, -0.00000005, -0.00000148, -0.00000060, 0.00000648, -0.00000215, -0.00000129, 0.00000015, 0.00000109, 0.00000073, -0.00000033, -0.00000096, -0.00000028, 0.00000005, -0.00000224, 0.00000031, -0.00000327, -0.00000126, 0.00000545, 0.00000093, -0.00000066, 0.00000135, -0.00000008, -0.00000235, 0.00000344, -0.00000305, -0.00000697, 0.00000098, 0.00000406, 0.00000028, 0.00000065, -0.00000028, 0.00000599, -0.00000260, -0.00000434, -0.00000237, 0.00000271, -0.00000318, 0.00000643, -0.00000236, -0.00000286, -0.00000165, -0.00000468, 0.00000057, 0.00000884, -0.00000123, },
		{ 0.00000018, 0.00000161, -0.00000447, -0.00000155, -0.00000305, 0.00001065, 0.00000049, -0.00000792, -0.00000444, 0.00001877, -0.00001217, -0.00001378, -0.00000691, 0.00001025, -0.00001543, 0.00000241, -0.00002458, -0.00000017, -0.00001124, 0.00000361, 0.00003123, -0.00000108, 0.00002062, -0.00001323, -0.00003382, -0.00015423, 0.00014601, 0.00028281, 0.00003779, -0.00015519, 0.00000140, -0.00003826, 0.00000574, -0.00003420, 0.00001627, -0.00002214, 0.00000344, -0.00001169, -0.00000200, 0.00000533, -0.00000733, 0.00000611, -0.00000292, 0.00001242, -0.00000474, -0.00000407, -0.00000309, 0.00000255, 0.00000469, -0.00001449, 0.00000946, -0.00000027, 0.00000412, -0.00000994, 0.00000391, -0.00000338, 0.00000306, -0.00000782, -0.00000146, 0.00000810, -0.00000272, -0.00000020, 0.00000018, -0.00000155, -0.00000079, 0.00000693, 0.00000025, -0.00000043, -0.00000091, -0.00000669, -0.00000033, 0.00000023, 0.00000237, -0.00000280, 0.00000393, 0.00000140, 0.00000242, -0.00000649, -0.00000031, 0.00000275, -0.00000115, -0.00000192, -0.00000032, -0.00000108, -0.00000127, 0.00000651, -0.00000151, -0.00000166, 0.00000021, 0.00000102, 0.00000092, -0.00000042, -0.00000080, -0.00000056, 0.00000029, -0.00000230, 0.00000032, -0.00000301, -0.00000186, 0.00000528, 0.00000149, -0.00000083, 0.00000131, 0.00000021, -0.00000247, 0.00000305, -0.00000235, -0.00000709, 0.00000030, 0.00000433, 0.00000041, 0.00000072, -0.00000052, 0.00000589, -0.00000210, -0.00000417, -0.00000268, 0.00000258, -0.00000295, 0.00000557, -0.00000099, -0.00000367, -0.00000142, -0.00000458, 0.00000009, 0.00000837, -0.00000007, -0.00000353, },
		{ 0.00000183, -0.00000483, -0.00000114, -0.00000337, 0.00000984, 0.00000199, -0.00000766, -0.00000548, 0.00001807, -0.00000995, -0.00001478, -0.00000717, 0.00000921, -0.00001317, -0.00000015, -0.00002082, -0.00000558, -0.00000894, 0.00000052, 0.00003573, -0.00000255, 0.00002346, -0.00001250, -0.00003096, -0.00015239, 0.00013569, 0.00026291, 0.00006562, -0.00015512, -0.00000494, -0.00003538, 0.00000312, -0.00003205, 0.00001405, -0.00001890, -0.00000109, -0.00000721, -0.00000613, 0.00000801, -0.00000911, 0.00000703, -0.00000289, 0.00001160, -0.00000333, -0.00000482, -0.00000252, 0.00000097, 0.00000632, -0.00001474, 0.00000896, -0.00000005, 0.00000411, -0.00000924, 0.00000258, -0.00000221, 0.00000200, -0.00000615, -0.00000315, 0.00000852, -0.00000263, -0.00000011, 0.00000003, -0.00000110, -0.00000142, 0.00000672, 0.00000120, -0.00000111, -0.00000033, -0.00000670, -0.00000066, 0.00000002, 0.00000258, -0.00000281, 0.00000368, 0.00000159, 0.00000251, -0.00000585, -0.00000131, 0.00000310, -0.00000115, -0.00000177, -0.00000060, -0.00000070, -0.00000185, 0.00000639, -0.00000080, -0.00000202, 0.00000025, 0.00000093, 0.00000113, -0.00000052, -0.00000063, -0.00000083, 0.00000050, -0.00000232, 0.00000029, -0.00000272, -0.00000239, 0.00000500, 0.00000206, -0.00000097, 0.00000126, 0.00000047, -0.00000251, 0.00000260, -0.00000169, -0.00000711, -0.00000039, 0.00000450, 0.00000063, 0.00000071, -0.00000066, 0.00000566, -0.00000152, -0.00000407, -0.00000281, 0.00000222, -0.00000250, 0.00000451, 0.00000041, -0.00000440, -0.00000125, -0.00000449, -0.00000026, 0.00000776, 0.00000109, -0.00000381, 0.00000004, },
		{ -0.00000510, -0.00000093, -0.00000336, 0.00000871, 0.00000364, -0.00000753, -0.00000613, 0.00001685, -0.00000740, -0.00001596, -0.00000709, 0.00000765, -0.00001042, -0.00000312, -0.00001682, -0.00001074, -0.00000711, -0.00000243, 0.00003999, -0.00000368, 0.00002610, -0.00001167, -0.00002802, -0.00015106, 0.00012706, 0.00024289, 0.00008957, -0.00015076, -0.00001351, -0.00003127, -0.00000043, -0.00002910, 0.00001104, -0.00001489, -0.00000608, -0.00000265, -0.00001003, 0.00001031, -0.00001049, 0.00000753, -0.00000250, 0.00001058, -0.00000190, -0.00000548, -0.00000203, -0.00000043, 0.00000755, -0.00001455, 0.00000813, 0.00000047, 0.00000375, -0.00000818, 0.00000098, -0.00000088, 0.00000082, -0.00000447, -0.00000468, 0.00000876, -0.00000246, -0.00000003, -0.00000012, -0.00000068, -0.00000192, 0.00000635, 0.00000218, -0.00000170, 0.00000009, -0.00000655, -0.00000106, -0.00000013, 0.00000269, -0.00000270, 0.00000331, 0.00000187, 0.00000249, -0.00000509, -0.00000229, 0.00000339, -0.00000112, -0.00000160, -0.00000087, -0.00000035, -0.00000232, 0.00000613, -0.00000004, -0.00000233, 0.00000026, 0.00000084, 0.00000134, -0.00000061, -0.00000047, -0.00000108, 0.00000066, -0.00000230, 0.00000023, -0.00000243, -0.00000284, 0.00000463, 0.00000264, -0.00000107, 0.00000118, 0.00000071, -0.00000247, 0.00000211, -0.00000108, -0.00000703, -0.00000106, 0.00000456, 0.00000095, 0.00000063, -0.00000070, 0.00000532, -0.00000088, -0.00000405, -0.00000277, 0.00000167, -0.00000186, 0.00000334, 0.00000179, -0.00000500, -0.00000118, -0.00000437, -0.00000052, 0.00000703, 0.00000221, -0.00000398, -0.00000014, 0.00000203, },
		{ -0.00000095, -0.00000304, 0.00000728, 0.00000540, -0.00000750, -0.00000644, 0.00001521, -0.00000465, -0.00001719, -0.00000677, 0.00000570, -0.00000736, -0.00000625, -0.00001285, -0.00001541, -0.00000595, -0.00000512, 0.00004388, -0.00000433, 0.00002842, -0.00001060, -0.00002519, -0.00014998, 0.00011961, 0.00022343, 0.00010928, -0.00014233, -0.00002389, -0.00002625, -0.00000462, -0.00002562, 0.00000745, -0.00001037, -0.00001126, 0.00000173, -0.00001349, 0.00001208, -0.00001137, 0.00000756, -0.00000177, 0.00000939, -0.00000050, -0.00000601, -0.00000167, -0.00000160, 0.00000836, -0.00001392, 0.00000699, 0.00000125, 0.00000308, -0.00000682, -0.00000080, 0.00000052, -0.00000037, -0.00000285, -0.00000599, 0.00000879, -0.00000222, 0.00000001, -0.00000024, -0.00000030, -0.00000229, 0.00000583, 0.00000317, -0.00000217, 0.00000033, -0.00000626, -0.00000153, -0.00000023, 0.00000270, -0.00000247, 0.00000285, 0.00000221, 0.00000238, -0.00000425, -0.00000324, 0.00000360, -0.00000106, -0.00000145, -0.00000114, -0.00000004, -0.00000268, 0.00000574, 0.00000073, -0.00000260, 0.00000023, 0.00000074, 0.00000157, -0.00000070, -0.00000031, -0.00000130, 0.00000077, -0.00000224, 0.00000015, -0.00000213, -0.00000323, 0.00000418, 0.00000319, -0.00000111, 0.00000107, 0.00000091, -0.00000236, 0.00000159, -0.00000055, -0.00000687, -0.00000172, 0.00000450, 0.00000136, 0.00000050, -0.00000066, 0.00000488, -0.00000021, -0.00000410, -0.00000258, 0.00000097, -0.00000111, 0.00000212, 0.00000306, -0.00000543, -0.00000123, -0.00000424, -0.00000071, 0.00000624, 0.00000324, -0.00000403, -0.00000037, 0.00000220, -0.00000525, },
		{ -0.00000243, 0.00000563, 0.00000721, -0.00000750, -0.00000646, 0.00001326, -0.00000183, -0.00001837, -0.00000633, 0.00000350, -0.00000418, -0.00000932, -0.00000915, -0.00001935, -0.00000555, -0.00000749, 0.00004730, -0.00000440, 0.00003034, -0.00000916, -0.00002261, -0.00014892, 0.00011289, 0.00020508, 0.00012461, -0.00013018, -0.00003560, -0.00002075, -0.00000914, -0.00002184, 0.00000353, -0.00000560, -0.00001635, 0.00000569, -0.00001633, 0.00001323, -0.00001171, 0.00000713, -0.00000073, 0.00000809, 0.00000081, -0.00000637, -0.00000145, -0.00000252, 0.00000876, -0.00001291, 0.00000560, 0.00000224, 0.00000218, -0.00000524, -0.00000269, 0.00000189, -0.00000149, -0.00000139, -0.00000704, 0.00000859, -0.00000189, 0.00000001, -0.00000034, 0.00000002, -0.00000252, 0.00000519, 0.00000413, -0.00000250, 0.00000040, -0.00000582, -0.00000204, -0.00000029, 0.00000261, -0.00000213, 0.00000230, 0.00000259, 0.00000220, -0.00000337, -0.00000412, 0.00000370, -0.00000096, -0.00000130, -0.00000137, 0.00000020, -0.00000292, 0.00000525, 0.00000151, -0.00000280, 0.00000017, 0.00000064, 0.00000179, -0.00000077, -0.00000017, -0.00000148, 0.00000082, -0.00000213, 0.00000004, -0.00000182, -0.00000355, 0.00000367, 0.00000370, -0.00000108, 0.00000094, 0.00000110, -0.00000219, 0.00000108, -0.00000011, -0.00000663, -0.00000233, 0.00000432, 0.00000185, 0.00000032, -0.00000054, 0.00000436, 0.00000049, -0.00000419, -0.00000226, 0.00000018, -0.00000031, 0.00000092, 0.00000416, -0.00000564, -0.00000142, -0.00000406, -0.00000084, 0.00000542, 0.00000416, -0.00000396, -0.00000063, 0.00000233, -0.00000527, -0.00000119, },
		{ 0.00000381, 0.00000898, -0.00000750, -0.00000627, 0.00001112, 0.00000092, -0.00001938, -0.00000588, 0.00000122, -0.00000108, -0.00001212, -0.00000594, -0.00002240, -0.00000600, -0.00000951, 0.00005019, -0.00000386, 0.00003179, -0.00000728, -0.00002038, -0.00014771, 0.00010654, 0.00018826, 0.00013556, -0.00011483, -0.00004806, -0.00001517, -0.00001368, -0.00001804, -0.00000046, -0.00000085, -0.00002110, 0.00000902, -0.00001843, 0.00001370, -0.00001149, 0.00000624, 0.00000058, 0.00000673, 0.00000200, -0.00000655, -0.00000140, -0.00000318, 0.00000875, -0.00001157, 0.00000402, 0.00000338, 0.00000110, -0.00000353, -0.00000456, 0.00000314, -0.00000246, -0.00000013, -0.00000778, 0.00000816, -0.00000147, -0.00000003, -0.00000039, 0.00000027, -0.00000261, 0.00000446, 0.00000503, -0.00000266, 0.00000029, -0.00000528, -0.00000257, -0.00000032, 0.00000244, -0.00000171, 0.00000171, 0.00000299, 0.00000199, -0.00000248, -0.00000490, 0.00000369, -0.00000082, -0.00000118, -0.00000155, 0.00000038, -0.00000304, 0.00000465, 0.00000226, -0.00000293, 0.00000008, 0.00000053, 0.00000200, -0.00000082, -0.00000006, -0.00000161, 0.00000082, -0.00000199, -0.00000009, -0.00000152, -0.00000381, 0.00000311, 0.00000416, -0.00000099, 0.00000078, 0.00000125, -0.00000198, 0.00000058, 0.00000025, -0.00000632, -0.00000288, 0.00000402, 0.00000241, 0.00000011, -0.00000038, 0.00000380, 0.00000119, -0.00000431, -0.00000185, -0.00000067, 0.00000049, -0.00000019, 0.00000505, -0.00000562, -0.00000174, -0.00000384, -0.00000094, 0.00000460, 0.00000494, -0.00000375, -0.00000091, 0.00000241, -0.00000517, -0.00000163, -0.00000158, },
		{ 0.00001065, -0.00000743, -0.00000596, 0.00000891, 0.00000347, -0.00002011, -0.00000550, -0.00000103, 0.00000179, -0.00001448, -0.00000340, -0.00002446, -0.00000729, -0.00001120, 0.00005253, -0.00000266, 0.00003277, -0.00000491, -0.00001858, -0.00014622, 0.00010029, 0.00017322, 0.00014233, -0.00009688, -0.00006068, -0.00000995, -0.00001792, -0.00001447, -0.00000430, 0.00000366, -0.00002529, 0.00001156, -0.00001970, 0.00001347, -0.00001073, 0.00000494, 0.00000208, 0.00000538, 0.00000302, -0.00000651, -0.00000153, -0.00000358, 0.00000839, -0.00000997, 0.00000233, 0.00000460, -0.00000007, -0.00000178, -0.00000634, 0.00000420, -0.00000322, 0.00000086, -0.00000819, 0.00000751, -0.00000097, -0.00000012, -0.00000039, 0.00000044, -0.00000258, 0.00000366, 0.00000583, -0.00000265, 0.00000004, -0.00000465, -0.00000311, -0.00000033, 0.00000221, -0.00000124, 0.00000110, 0.00000337, 0.00000176, -0.00000161, -0.00000556, 0.00000357, -0.00000063, -0.00000109, -0.00000169, 0.00000048, -0.00000306, 0.00000399, 0.00000297, -0.00000297, -0.00000004, 0.00000041, 0.00000220, -0.00000082, 0.00000001, -0.00000170, 0.00000076, -0.00000181, -0.00000024, -0.00000121, -0.00000401, 0.00000253, 0.00000455, -0.00000082, 0.00000061, 0.00000138, -0.00000174, 0.00000012, 0.00000051, -0.00000596, -0.00000336, 0.00000360, 0.00000302, -0.00000010, -0.00000019, 0.00000321, 0.00000186, -0.00000443, -0.00000139, -0.00000151, 0.00000122, -0.00000115, 0.00000567, -0.00000535, -0.00000220, -0.00000358, -0.00000101, 0.00000381, 0.00000557, -0.00000342, -0.00000120, 0.00000243, -0.00000493, -0.00000226, -0.00000054, 0.00000190, },
		{ -0.00000723, -0.00000559, 0.00000674, 0.00000570, -0.00002047, -0.00000530, -0.00000313, 0.00000430, -0.00001627, -0.00000162, -0.00002547, -0.00000934, -0.00001264, 0.00005433, -0.00000083, 0.00003327, -0.00000204, -0.00001725, -0.00014435, 0.00009394, 0.00016010, 0.00014523, -0.00007700, -0.00007285, -0.00000547, -0.00002158, -0.00001134, -0.00000779, 0.00000770, -0.00002873, 0.00001319, -0.00002010, 0.00001255, -0.00000948, 0.00000331, 0.00000370, 0.00000412, 0.00000383, -0.00000626, -0.00000183, -0.00000373, 0.00000772, -0.00000817, 0.00000060, 0.00000584, -0.00000126, -0.00000007, -0.00000794, 0.00000499, -0.00000372, 0.00000156, -0.00000828, 0.00000666, -0.00000041, -0.00000026, -0.00000036, 0.00000054, -0.00000244, 0.00000282, 0.00000652, -0.00000246, -0.00000036, -0.00000396, -0.00000363, -0.00000033, 0.00000192, -0.00000072, 0.00000049, 0.00000372, 0.00000155, -0.00000078, -0.00000608, 0.00000332, -0.00000040, -0.00000102, -0.00000178, 0.00000051, -0.00000297, 0.00000327, 0.00000362, -0.00000294, -0.00000019, 0.00000030, 0.00000237, -0.00000079, 0.00000004, -0.00000172, 0.00000064, -0.00000161, -0.00000041, -0.00000090, -0.00000416, 0.00000193, 0.00000486, -0.00000059, 0.00000042, 0.00000149, -0.00000147, -0.00000031, 0.00000069, -0.00000556, -0.00000375, 0.00000307, 0.00000365, -0.00000029, 0.00000001, 0.00000262, 0.00000247, -0.00000453, -0.00000091, -0.00000229, 0.00000183, -0.00000192, 0.00000601, -0.00000482, -0.00000278, -0.00000330, -0.00000106, 0.00000307, 0.00000604, -0.00000300, -0.00000148, 0.00000238, -0.00000457, -0.00000306, 0.00000063, -0.00000000, 0.00001213, },
		{ -0.00000526, 0.00000474, 0.00000752, -0.00002041, -0.00000532, -0.00000500, 0.00000633, -0.00001739, -0.00000068, -0.00002547, -0.00001206, -0.00001394, 0.00005564, 0.00000159, 0.00003334, 0.00000130, -0.00001635, -0.00014210, 0.00008738, 0.00014890, 0.00014469, -0.00005593, -0.00008399, -0.00000208, -0.00002445, -0.00000883, -0.00001077, 0.00001114, -0.00003129, 0.00001386, -0.00001965, 0.00001101, -0.00000780, 0.00000142, 0.00000536, 0.00000299, 0.00000439, -0.00000577, -0.00000229, -0.00000367, 0.00000680, -0.00000627, -0.00000109, 0.00000702, -0.00000240, 0.00000151, -0.00000927, 0.00000545, -0.00000392, 0.00000195, -0.00000806, 0.00000564, 0.00000018, -0.00000042, -0.00000028, 0.00000057, -0.00000220, 0.00000198, 0.00000706, -0.00000210, -0.00000087, -0.00000325, -0.00000411, -0.00000033, 0.00000160, -0.00000020, -0.00000008, 0.00000401, 0.00000137, -0.00000003, -0.00000645, 0.00000296, -0.00000013, -0.00000100, -0.00000181, 0.00000047, -0.00000280, 0.00000252, 0.00000419, -0.00000281, -0.00000036, 0.00000018, 0.00000252, -0.00000071, 0.00000002, -0.00000170, 0.00000048, -0.00000138, -0.00000059, -0.00000061, -0.00000426, 0.00000133, 0.00000508, -0.00000029, 0.00000022, 0.00000156, -0.00000119, -0.00000067, 0.00000077, -0.00000514, -0.00000406, 0.00000245, 0.00000428, -0.00000044, 0.00000019, 0.00000206, 0.00000301, -0.00000459, -0.00000045, -0.00000297, 0.00000229, -0.00000246, 0.00000606, -0.00000407, -0.00000345, -0.00000301, -0.00000110, 0.00000239, 0.00000636, -0.00000249, -0.00000172, 0.00000225, -0.00000409, -0.00000397, 0.00000188, -0.00000183, 0.00001337, -0.00000684, },
		{ 0.00000298, 0.00000886, -0.00001990, -0.00000558, -0.00000658, 0.00000782, -0.00001780, -0.00000056, -0.00002454, -0.00001525, -0.00001522, 0.00005651, 0.00000453, 0.00003305, 0.00000507, -0.00001587, -0.00013947, 0.00008057, 0.00013953, 0.00014122, -0.00003439, -0.00009359, -0.00000005, -0.00002636, -0.00000705, -0.00001312, 0.00001384, -0.00003291, 0.00001356, -0.00001842, 0.00000894, -0.00000581, -0.00000063, 0.00000695, 0.00000208, 0.00000468, -0.00000507, -0.00000290, -0.00000342, 0.00000569, -0.00000435, -0.00000267, 0.00000808, -0.00000343, 0.00000289, -0.00001028, 0.00000555, -0.00000381, 0.00000205, -0.00000755, 0.00000448, 0.00000078, -0.00000059, -0.00000017, 0.00000056, -0.00000190, 0.00000116, 0.00000744, -0.00000156, -0.00000147, -0.00000254, -0.00000452, -0.00000036, 0.00000127, 0.00000031, -0.00000060, 0.00000421, 0.00000124, 0.00000065, -0.00000667, 0.00000249, 0.00000018, -0.00000100, -0.00000178, 0.00000037, -0.00000256, 0.00000177, 0.00000466, -0.00000260, -0.00000055, 0.00000006, 0.00000263, -0.00000057, -0.00000003, -0.00000162, 0.00000028, -0.00000113, -0.00000077, -0.00000032, -0.00000431, 0.00000074, 0.00000521, 0.00000005, 0.00000003, 0.00000160, -0.00000089, -0.00000099, 0.00000078, -0.00000472, -0.00000426, 0.00000174, 0.00000488, -0.00000053, 0.00000034, 0.00000155, 0.00000346, -0.00000459, -0.00000005, -0.00000352, 0.00000255, -0.00000275, 0.00000582, -0.00000312, -0.00000417, -0.00000274, -0.00000110, 0.00000176, 0.00000654, -0.00000193, -0.00000192, 0.00000204, -0.00000351, -0.00000497, 0.00000313, -0.00000351, 0.00001429, -0.00000622, -0.00000502, },
		{ 0.00000968, -0.00001893, -0.00000609, -0.00000784, 0.00000876, -0.00001751, -0.00000121, -0.00002283, -0.00001872, -0.00001664, 0.00005701, 0.00000793, 0.00003249, 0.00000920, -0.00001572, -0.00013652, 0.00007355, 0.00013182, 0.00013538, -0.00001307, -0.00010120, 0.00000043, -0.00002723, -0.00000608, -0.00001480, 0.00001576, -0.00003355, 0.00001233, -0.00001649, 0.00000647, -0.00000361, -0.00000272, 0.00000840, 0.00000143, 0.00000468, -0.00000416, -0.00000362, -0.00000304, 0.00000447, -0.00000249, -0.00000408, 0.00000898, -0.00000427, 0.00000402, -0.00001092, 0.00000529, -0.00000341, 0.00000186, -0.00000681, 0.00000325, 0.00000136, -0.00000077, -0.00000005, 0.00000050, -0.00000156, 0.00000039, 0.00000765, -0.00000088, -0.00000211, -0.00000188, -0.00000486, -0.00000041, 0.00000093, 0.00000079, -0.00000104, 0.00000433, 0.00000118, 0.00000122, -0.00000672, 0.00000191, 0.00000052, -0.00000103, -0.00000171, 0.00000021, -0.00000226, 0.00000104, 0.00000502, -0.00000230, -0.00000076, -0.00000006, 0.00000270, -0.00000039, -0.00000013, -0.00000150, 0.00000004, -0.00000088, -0.00000095, -0.00000004, -0.00000432, 0.00000017, 0.00000526, 0.00000044, -0.00000014, 0.00000161, -0.00000059, -0.00000124, 0.00000072, -0.00000432, -0.00000437, 0.00000098, 0.00000543, -0.00000054, 0.00000043, 0.00000110, 0.00000380, -0.00000452, 0.00000028, -0.00000389, 0.00000260, -0.00000281, 0.00000532, -0.00000201, -0.00000490, -0.00000251, -0.00000107, 0.00000118, 0.00000660, -0.00000135, -0.00000204, 0.00000175, -0.00000285, -0.00000601, 0.00000432, -0.00000497, 0.00001486, -0.00000534, -0.00000492, 0.00000152, },
		{ -0.00001755, -0.00000682, -0.00000881, 0.00000916, -0.00001658, -0.00000253, -0.00002051, -0.00002223, -0.00001835, 0.00005721, 0.00001169, 0.00003174, 0.00001360, -0.00001582, -0.00013333, 0.00006637, 0.00012555, 0.00012772, 0.00000740, -0.00010650, -0.00000073, -0.00002702, -0.00000590, -0.00001578, 0.00001690, -0.00003326, 0.00001029, -0.00001403, 0.00000373, -0.00000132, -0.00000475, 0.00000961, 0.00000108, 0.00000440, -0.00000308, -0.00000441, -0.00000257, 0.00000320, -0.00000076, -0.00000526, 0.00000966, -0.00000491, 0.00000486, -0.00001119, 0.00000466, -0.00000273, 0.00000144, -0.00000588, 0.00000200, 0.00000187, -0.00000092, 0.00000007, 0.00000043, -0.00000119, -0.00000029, 0.00000768, -0.00000007, -0.00000277, -0.00000127, -0.00000510, -0.00000050, 0.00000061, 0.00000122, -0.00000138, 0.00000434, 0.00000118, 0.00000170, -0.00000663, 0.00000126, 0.00000086, -0.00000108, -0.00000160, 0.00000001, -0.00000194, 0.00000035, 0.00000526, -0.00000193, -0.00000096, -0.00000017, 0.00000272, -0.00000014, -0.00000027, -0.00000134, -0.00000022, -0.00000063, -0.00000113, 0.00000021, -0.00000428, -0.00000039, 0.00000523, 0.00000085, -0.00000028, 0.00000158, -0.00000029, -0.00000143, 0.00000061, -0.00000394, -0.00000437, 0.00000018, 0.00000591, -0.00000045, 0.00000047, 0.00000073, 0.00000403, -0.00000436, 0.00000051, -0.00000409, 0.00000244, -0.00000263, 0.00000459, -0.00000081, -0.00000560, -0.00000235, -0.00000100, 0.00000066, 0.00000657, -0.00000078, -0.00000209, 0.00000139, -0.00000213, -0.00000705, 0.00000540, -0.00000615, 0.00001503, -0.00000418, -0.00000499, 0.00000042, 0.00000998, },
		{ -0.00000770, -0.00000951, 0.00000907, -0.00001510, -0.00000436, -0.00001780, -0.00002555, -0.00002048, 0.00005717, 0.00001573, 0.00003091, 0.00001820, -0.00001608, -0.00013001, 0.00005914, 0.00012046, 0.00011882, 0.00002644, -0.00010926, -0.00000351, -0.00002582, -0.00000645, -0.00001613, 0.00001729, -0.00003212, 0.00000756, -0.00001118, 0.00000088, 0.00000091, -0.00000662, 0.00001051, 0.00000107, 0.00000385, -0.00000186, -0.00000524, -0.00000206, 0.00000196, 0.00000079, -0.00000618, 0.00001011, -0.00000530, 0.00000538, -0.00001107, 0.00000369, -0.00000182, 0.00000084, -0.00000483, 0.00000078, 0.00000230, -0.00000103, 0.00000017, 0.00000035, -0.00000082, -0.00000088, 0.00000754, 0.00000084, -0.00000342, -0.00000076, -0.00000524, -0.00000063, 0.00000031, 0.00000157, -0.00000163, 0.00000425, 0.00000124, 0.00000207, -0.00000640, 0.00000054, 0.00000121, -0.00000115, -0.00000145, -0.00000022, -0.00000160, -0.00000028, 0.00000539, -0.00000149, -0.00000117, -0.00000029, 0.00000270, 0.00000015, -0.00000043, -0.00000116, -0.00000049, -0.00000039, -0.00000129, 0.00000045, -0.00000420, -0.00000092, 0.00000512, 0.00000128, -0.00000038, 0.00000152, 0.00000000, -0.00000157, 0.00000046, -0.00000361, -0.00000429, -0.00000065, 0.00000628, -0.00000025, 0.00000044, 0.00000044, 0.00000413, -0.00000411, 0.00000064, -0.00000409, 0.00000208, -0.00000225, 0.00000368, 0.00000043, -0.00000621, -0.00000229, -0.00000086, 0.00000016, 0.00000647, -0.00000024, -0.00000205, 0.00000097, -0.00000138, -0.00000805, 0.00000631, -0.00000701, 0.00001479, -0.00000275, -0.00000525, -0.00000033, 0.00000977, -0.00001581, },
		{ -0.00001001, 0.00000856, -0.00001317, -0.00000655, -0.00001492, -0.00002845, -0.00002316, 0.00005691, 0.00001998, 0.00003008, 0.00002290, -0.00001638, -0.00012666, 0.00005196, 0.00011628, 0.00010920, 0.00004362, -0.00010937, -0.00000783, -0.00002373, -0.00000764, -0.00001592, 0.00001702, -0.00003025, 0.00000432, -0.00000814, -0.00000192, 0.00000299, -0.00000824, 0.00001105, 0.00000142, 0.00000304, -0.00000054, -0.00000606, -0.00000156, 0.00000079, 0.00000209, -0.00000681, 0.00001031, -0.00000543, 0.00000558, -0.00001059, 0.00000244, -0.00000074, 0.00000012, -0.00000373, -0.00000036, 0.00000261, -0.00000109, 0.00000024, 0.00000028, -0.00000047, -0.00000135, 0.00000723, 0.00000182, -0.00000401, -0.00000037, -0.00000528, -0.00000080, 0.00000004, 0.00000185, -0.00000176, 0.00000406, 0.00000137, 0.00000236, -0.00000604, -0.00000022, 0.00000153, -0.00000122, -0.00000129, -0.00000045, -0.00000127, -0.00000084, 0.00000539, -0.00000100, -0.00000137, -0.00000040, 0.00000263, 0.00000048, -0.00000061, -0.00000097, -0.00000075, -0.00000018, -0.00000144, 0.00000065, -0.00000408, -0.00000144, 0.00000495, 0.00000169, -0.00000044, 0.00000141, 0.00000029, -0.00000166, 0.00000028, -0.00000333, -0.00000412, -0.00000146, 0.00000653, 0.00000006, 0.00000034, 0.00000024, 0.00000413, -0.00000378, 0.00000066, -0.00000392, 0.00000155, -0.00000171, 0.00000266, 0.00000165, -0.00000668, -0.00000237, -0.00000066, -0.00000031, 0.00000633, 0.00000024, -0.00000192, 0.00000050, -0.00000063, -0.00000895, 0.00000700, -0.00000751, 0.00001413, -0.00000105, -0.00000569, -0.00000073, 0.00000913, -0.00001380, -0.00000866, },
		{ 0.00000773, -0.00001095, -0.00000891, -0.00001211, -0.00003073, -0.00002644, 0.00005645, 0.00002439, 0.00002932, 0.00002763, -0.00001664, -0.00012338, 0.00004497, 0.00011276, 0.00009932, 0.00005858, -0.00010685, -0.00001348, -0.00002096, -0.00000930, -0.00001528, 0.00001620, -0.00002780, 0.00000075, -0.00000509, -0.00000452, 0.00000479, -0.00000951, 0.00001118, 0.00000211, 0.00000202, 0.00000084, -0.00000683, -0.00000113, -0.00000024, 0.00000312, -0.00000714, 0.00001027, -0.00000531, 0.00000548, -0.00000978, 0.00000095, 0.00000043, -0.00000064, -0.00000265, -0.00000135, 0.00000277, -0.00000108, 0.00000026, 0.00000023, -0.00000016, -0.00000170, 0.00000676, 0.00000284, -0.00000452, -0.00000010, -0.00000521, -0.00000101, -0.00000019, 0.00000205, -0.00000179, 0.00000377, 0.00000154, 0.00000256, -0.00000558, -0.00000100, 0.00000183, -0.00000128, -0.00000112, -0.00000069, -0.00000096, -0.00000131, 0.00000527, -0.00000046, -0.00000154, -0.00000051, 0.00000250, 0.00000086, -0.00000081, -0.00000077, -0.00000100, 0.00000001, -0.00000156, 0.00000082, -0.00000390, -0.00000193, 0.00000473, 0.00000209, -0.00000044, 0.00000127, 0.00000057, -0.00000169, 0.00000010, -0.00000310, -0.00000387, -0.00000223, 0.00000665, 0.00000049, 0.00000017, 0.00000012, 0.00000401, -0.00000337, 0.00000058, -0.00000358, 0.00000088, -0.00000106, 0.00000158, 0.00000278, -0.00000697, -0.00000258, -0.00000038, -0.00000077, 0.00000616, 0.00000065, -0.00000171, 0.00000001, 0.00000009, -0.00000973, 0.00000743, -0.00000764, 0.00001309, 0.00000086, -0.00000628, -0.00000082, 0.00000813, -0.00001163, -0.00000961, -0.00001037, },
		{ -0.00000857, -0.00001126, -0.00000956, -0.00003222, -0.00003034, 0.00005574, 0.00002891, 0.00002870, 0.00003233, -0.00001679, -0.00012027, 0.00003825, 0.00010965, 0.00008959, 0.00007111, -0.00010183, -0.00002021, -0.00001772, -0.00001126, -0.00001435, 0.00001497, -0.00002493, -0.00000296, -0.00000221, -0.00000679, 0.00000623, -0.00001040, 0.00001089, 0.00000312, 0.00000084, 0.00000221, -0.00000749, -0.00000080, -0.00000109, 0.00000385, -0.00000719, 0.00000999, -0.00000495, 0.00000510, -0.00000870, -0.00000069, 0.00000163, -0.00000139, -0.00000164, -0.00000217, 0.00000277, -0.00000100, 0.00000024, 0.00000022, 0.00000010, -0.00000192, 0.00000616, 0.00000387, -0.00000492, 0.00000004, -0.00000504, -0.00000125, -0.00000038, 0.00000216, -0.00000171, 0.00000341, 0.00000175, 0.00000270, -0.00000504, -0.00000177, 0.00000208, -0.00000133, -0.00000095, -0.00000091, -0.00000069, -0.00000168, 0.00000504, 0.00000011, -0.00000169, -0.00000061, 0.00000233, 0.00000126, -0.00000099, -0.00000058, -0.00000123, 0.00000016, -0.00000166, 0.00000095, -0.00000368, -0.00000240, 0.00000447, 0.00000245, -0.00000038, 0.00000110, 0.00000083, -0.00000168, -0.00000009, -0.00000294, -0.00000358, -0.00000295, 0.00000661, 0.00000103, -0.00000004, 0.00000007, 0.00000380, -0.00000291, 0.00000041, -0.00000311, 0.00000012, -0.00000036, 0.00000052, 0.00000376, -0.00000705, -0.00000296, -0.00000003, -0.00000122, 0.00000599, 0.00000098, -0.00000142, -0.00000048, 0.00000075, -0.00001034, 0.00000759, -0.00000740, 0.00001170, 0.00000294, -0.00000696, -0.00000067, 0.00000687, -0.00000943, -0.00001044, -0.00001068, 0.00000668, },
		{ -0.00001342, -0.00000746, -0.00003282, -0.00003484, 0.00005472, 0.00003354, 0.00002826, 0.00003697, -0.00001675, -0.00011740, 0.00003190, 0.00010674, 0.00008032, 0.00008108, -0.00009454, -0.00002769, -0.00001427, -0.00001332, -0.00001329, 0.00001348, -0.00002182, -0.00000661, 0.00000034, -0.00000863, 0.00000726, -0.00001087, 0.00001017, 0.00000441, -0.00000045, 0.00000351, -0.00000801, -0.00000061, -0.00000174, 0.00000429, -0.00000697, 0.00000950, -0.00000438, 0.00000448, -0.00000740, -0.00000242, 0.00000278, -0.00000204, -0.00000077, -0.00000277, 0.00000261, -0.00000083, 0.00000015, 0.00000025, 0.00000030, -0.00000202, 0.00000546, 0.00000486, -0.00000518, 0.00000004, -0.00000478, -0.00000152, -0.00000054, 0.00000218, -0.00000154, 0.00000299, 0.00000198, 0.00000278, -0.00000444, -0.00000252, 0.00000227, -0.00000135, -0.00000079, -0.00000110, -0.00000047, -0.00000195, 0.00000471, 0.00000069, -0.00000181, -0.00000070, 0.00000212, 0.00000168, -0.00000116, -0.00000041, -0.00000142, 0.00000026, -0.00000172, 0.00000104, -0.00000341, -0.00000285, 0.00000417, 0.00000276, -0.00000025, 0.00000090, 0.00000107, -0.00000163, -0.00000026, -0.00000285, -0.00000324, -0.00000359, 0.00000642, 0.00000167, -0.00000030, 0.00000008, 0.00000352, -0.00000239, 0.00000016, -0.00000253, -0.00000068, 0.00000034, -0.00000048, 0.00000454, -0.00000689, -0.00000349, 0.00000039, -0.00000166, 0.00000582, 0.00000122, -0.00000108, -0.00000094, 0.00000132, -0.00001077, 0.00000746, -0.00000682, 0.00001001, 0.00000513, -0.00000770, -0.00000034, 0.00000547, -0.00000731, -0.00001107, -0.00001102, 0.00000551, -0.00000618, },
		{ -0.00000594, -0.00003247, -0.00003985, 0.00005329, 0.00003827, 0.00002802, 0.00004152, -0.00001648, -0.00011483, 0.00002599, 0.00010388, 0.00007175, 0.00008851, -0.00008530, -0.00003556, -0.00001088, -0.00001529, -0.00001224, 0.00001187, -0.00001864, -0.00001004, 0.00000244, -0.00000996, 0.00000784, -0.00001091, 0.00000906, 0.00000593, -0.00000179, 0.00000470, -0.00000835, -0.00000058, -0.00000218, 0.00000444, -0.00000651, 0.00000885, -0.00000364, 0.00000367, -0.00000597, -0.00000414, 0.00000380, -0.00000256, -0.00000007, -0.00000315, 0.00000229, -0.00000060, 0.00000002, 0.00000032, 0.00000045, -0.00000201, 0.00000467, 0.00000579, -0.00000530, -0.00000010, -0.00000444, -0.00000180, -0.00000065, 0.00000213, -0.00000129, 0.00000252, 0.00000221, 0.00000283, -0.00000380, -0.00000322, 0.00000239, -0.00000135, -0.00000066, -0.00000125, -0.00000031, -0.00000212, 0.00000429, 0.00000127, -0.00000189, -0.00000079, 0.00000186, 0.00000210, -0.00000129, -0.00000028, -0.00000156, 0.00000032, -0.00000174, 0.00000108, -0.00000310, -0.00000329, 0.00000386, 0.00000302, -0.00000007, 0.00000068, 0.00000129, -0.00000155, -0.00000042, -0.00000281, -0.00000289, -0.00000413, 0.00000607, 0.00000239, -0.00000058, 0.00000013, 0.00000318, -0.00000185, -0.00000014, -0.00000188, -0.00000148, 0.00000098, -0.00000135, 0.00000508, -0.00000647, -0.00000418, 0.00000087, -0.00000209, 0.00000566, 0.00000139, -0.00000072, -0.00000135, 0.00000178, -0.00001099, 0.00000704, -0.00000592, 0.00000808, 0.00000733, -0.00000841, 0.00000007, 0.00000403, -0.00000540, -0.00001141, -0.00001145, 0.00000431, -0.00000391, -0.00001527, },
		{ -0.00003119, -0.00004520, 0.00005132, 0.00004311, 0.00002800, 0.00004596, -0.00001595, -0.00011259, 0.00002056, 0.00010094, 0.00006403, 0.00009349, -0.00007448, -0.00004344, -0.00000781, -0.00001699, -0.00001135, 0.00001027, -0.00001553, -0.00001309, 0.00000397, -0.00001075, 0.00000798, -0.00001054, 0.00000760, 0.00000761, -0.00000309, 0.00000573, -0.00000847, -0.00000072, -0.00000239, 0.00000432, -0.00000587, 0.00000807, -0.00000277, 0.00000274, -0.00000446, -0.00000578, 0.00000464, -0.00000288, 0.00000042, -0.00000328, 0.00000181, -0.00000031, -0.00000017, 0.00000042, 0.00000055, -0.00000190, 0.00000383, 0.00000662, -0.00000525, -0.00000036, -0.00000406, -0.00000208, -0.00000074, 0.00000201, -0.00000099, 0.00000204, 0.00000242, 0.00000285, -0.00000314, -0.00000385, 0.00000243, -0.00000131, -0.00000056, -0.00000135, -0.00000021, -0.00000219, 0.00000380, 0.00000183, -0.00000192, -0.00000086, 0.00000156, 0.00000252, -0.00000139, -0.00000019, -0.00000166, 0.00000032, -0.00000173, 0.00000108, -0.00000273, -0.00000370, 0.00000353, 0.00000322, 0.00000017, 0.00000045, 0.00000147, -0.00000143, -0.00000055, -0.00000282, -0.00000253, -0.00000456, 0.00000557, 0.00000318, -0.00000086, 0.00000021, 0.00000281, -0.00000131, -0.00000046, -0.00000120, -0.00000223, 0.00000151, -0.00000205, 0.00000535, -0.00000581, -0.00000499, 0.00000138, -0.00000250, 0.00000551, 0.00000150, -0.00000035, -0.00000168, 0.00000211, -0.00001100, 0.00000635, -0.00000475, 0.00000601, 0.00000948, -0.00000902, 0.00000048, 0.00000269, -0.00000380, -0.00001138, -0.00001202, 0.00000317, -0.00000184, -0.00001669, -0.00000509, },
		{ -0.00005072, 0.00004865, 0.00004810, 0.00002820, 0.00005032, -0.00001515, -0.00011069, 0.00001561, 0.00009784, 0.00005725, 0.00009620, -0.00006251, -0.00005097, -0.00000528, -0.00001828, -0.00001071, 0.00000879, -0.00001262, -0.00001565, 0.00000488, -0.00001098, 0.00000769, -0.00000981, 0.00000585, 0.00000937, -0.00000428, 0.00000653, -0.00000837, -0.00000106, -0.00000240, 0.00000398, -0.00000508, 0.00000721, -0.00000182, 0.00000173, -0.00000295, -0.00000727, 0.00000523, -0.00000298, 0.00000068, -0.00000318, 0.00000120, 0.00000003, -0.00000040, 0.00000054, 0.00000060, -0.00000170, 0.00000297, 0.00000732, -0.00000503, -0.00000073, -0.00000364, -0.00000235, -0.00000080, 0.00000183, -0.00000064, 0.00000156, 0.00000260, 0.00000287, -0.00000248, -0.00000439, 0.00000239, -0.00000123, -0.00000050, -0.00000140, -0.00000018, -0.00000217, 0.00000325, 0.00000235, -0.00000191, -0.00000093, 0.00000124, 0.00000291, -0.00000142, -0.00000015, -0.00000170, 0.00000028, -0.00000168, 0.00000102, -0.00000232, -0.00000410, 0.00000319, 0.00000335, 0.00000046, 0.00000023, 0.00000161, -0.00000128, -0.00000066, -0.00000286, -0.00000220, -0.00000486, 0.00000493, 0.00000400, -0.00000111, 0.00000028, 0.00000243, -0.00000079, -0.00000078, -0.00000054, -0.00000287, 0.00000190, -0.00000254, 0.00000535, -0.00000492, -0.00000591, 0.00000189, -0.00000287, 0.00000535, 0.00000157, -0.00000002, -0.00000190, 0.00000230, -0.00001080, 0.00000541, -0.00000338, 0.00000387, 0.00001148, -0.00000946, 0.00000081, 0.00000153, -0.00000259, -0.00001095, -0.00001277, 0.00000214, -0.00000007, -0.00001761, -0.00000493, -0.00002904, },
		{ 0.00004514, 0.00005323, 0.00002861, 0.00005460, -0.00001409, -0.00010915, 0.00001116, 0.00009454, 0.00005142, 0.00009689, -0.00004982, -0.00005779, -0.00000349, -0.00001903, -0.00001042, 0.00000750, -0.00001000, -0.00001766, 0.00000515, -0.00001070, 0.00000703, -0.00000876, 0.00000388, 0.00001112, -0.00000530, 0.00000708, -0.00000802, -0.00000157, -0.00000221, 0.00000344, -0.00000422, 0.00000632, -0.00000084, 0.00000071, -0.00000152, -0.00000855, 0.00000554, -0.00000284, 0.00000072, -0.00000287, 0.00000049, 0.00000039, -0.00000065, 0.00000069, 0.00000062, -0.00000145, 0.00000213, 0.00000788, -0.00000465, -0.00000119, -0.00000320, -0.00000260, -0.00000084, 0.00000160, -0.00000026, 0.00000110, 0.00000272, 0.00000289, -0.00000183, -0.00000485, 0.00000227, -0.00000111, -0.00000047, -0.00000140, -0.00000020, -0.00000208, 0.00000267, 0.00000283, -0.00000185, -0.00000099, 0.00000090, 0.00000327, -0.00000139, -0.00000016, -0.00000169, 0.00000019, -0.00000160, 0.00000092, -0.00000188, -0.00000446, 0.00000284, 0.00000343, 0.00000078, 0.00000002, 0.00000171, -0.00000110, -0.00000075, -0.00000294, -0.00000192, -0.00000502, 0.00000416, 0.00000484, -0.00000132, 0.00000034, 0.00000205, -0.00000031, -0.00000108, 0.00000008, -0.00000336, 0.00000211, -0.00000282, 0.00000507, -0.00000383, -0.00000687, 0.00000238, -0.00000319, 0.00000518, 0.00000163, 0.00000024, -0.00000199, 0.00000232, -0.00001038, 0.00000426, -0.00000187, 0.00000174, 0.00001327, -0.00000967, 0.00000098, 0.00000064, -0.00000184, -0.00001010, -0.00001370, 0.00000123, 0.00000137, -0.00001800, -0.00000546, -0.00002617, -0.00005617, },
		{ 0.00005852, 0.00002923, 0.00005882, -0.00001276, -0.00010793, 0.00000716, 0.00009103, 0.00004650, 0.00009583, -0.00003685, -0.00006361, -0.00000258, -0.00001918, -0.00001051, 0.00000646, -0.00000774, -0.00001906, 0.00000479, -0.00000996, 0.00000609, -0.00000749, 0.00000180, 0.00001278, -0.00000609, 0.00000734, -0.00000743, -0.00000226, -0.00000185, 0.00000277, -0.00000331, 0.00000544, 0.00000014, -0.00000028, -0.00000021, -0.00000957, 0.00000555, -0.00000246, 0.00000056, -0.00000238, -0.00000028, 0.00000074, -0.00000091, 0.00000083, 0.00000063, -0.00000115, 0.00000133, 0.00000828, -0.00000411, -0.00000172, -0.00000278, -0.00000282, -0.00000087, 0.00000133, 0.00000012, 0.00000068, 0.00000279, 0.00000292, -0.00000121, -0.00000521, 0.00000206, -0.00000095, -0.00000049, -0.00000134, -0.00000029, -0.00000192, 0.00000208, 0.00000324, -0.00000174, -0.00000104, 0.00000054, 0.00000357, -0.00000129, -0.00000023, -0.00000162, 0.00000006, -0.00000148, 0.00000078, -0.00000140, -0.00000480, 0.00000248, 0.00000345, 0.00000112, -0.00000016, 0.00000176, -0.00000090, -0.00000081, -0.00000303, -0.00000169, -0.00000505, 0.00000328, 0.00000565, -0.00000146, 0.00000037, 0.00000170, 0.00000012, -0.00000134, 0.00000061, -0.00000369, 0.00000213, -0.00000288, 0.00000455, -0.00000258, -0.00000784, 0.00000279, -0.00000342, 0.00000497, 0.00000170, 0.00000040, -0.00000194, 0.00000220, -0.00000978, 0.00000294, -0.00000028, -0.00000028, 0.00001476, -0.00000960, 0.00000093, 0.00000009, -0.00000156, -0.00000885, -0.00001477, 0.00000046, 0.00000247, -0.00001790, -0.00000658, -0.00002274, -0.00006128, 0.00004062, },
		{ 0.00003008, 0.00006301, -0.00001119, -0.00010701, 0.00000359, 0.00008733, 0.00004243, 0.00009334, -0.00002399, -0.00006819, -0.00000261, -0.00001873, -0.00001097, 0.00000567, -0.00000585, -0.00001986, 0.00000387, -0.00000887, 0.00000496, -0.00000608, -0.00000032, 0.00001425, -0.00000658, 0.00000730, -0.00000661, -0.00000308, -0.00000137, 0.00000202, -0.00000243, 0.00000462, 0.00000108, -0.00000118, 0.00000093, -0.00001030, 0.00000525, -0.00000186, 0.00000022, -0.00000175, -0.00000106, 0.00000106, -0.00000116, 0.00000095, 0.00000063, -0.00000084, 0.00000059, 0.00000851, -0.00000343, -0.00000228, -0.00000239, -0.00000299, -0.00000088, 0.00000104, 0.00000049, 0.00000031, 0.00000279, 0.00000297, -0.00000063, -0.00000546, 0.00000178, -0.00000077, -0.00000055, -0.00000123, -0.00000042, -0.00000171, 0.00000149, 0.00000358, -0.00000157, -0.00000108, 0.00000018, 0.00000382, -0.00000111, -0.00000035, -0.00000151, -0.00000011, -0.00000135, 0.00000060, -0.00000091, -0.00000509, 0.00000211, 0.00000343, 0.00000147, -0.00000030, 0.00000176, -0.00000068, -0.00000085, -0.00000311, -0.00000153, -0.00000494, 0.00000233, 0.00000642, -0.00000152, 0.00000036, 0.00000139, 0.00000047, -0.00000153, 0.00000104, -0.00000383, 0.00000195, -0.00000273, 0.00000382, -0.00000124, -0.00000876, 0.00000310, -0.00000354, 0.00000470, 0.00000182, 0.00000045, -0.00000174, 0.00000193, -0.00000902, 0.00000150, 0.00000130, -0.00000213, 0.00001590, -0.00000920, 0.00000062, -0.00000009, -0.00000176, -0.00000725, -0.00001595, -0.00000021, 0.00000326, -0.00001735, -0.00000820, -0.00001899, -0.00006579, 0.00003497, 0.00006394, },
		{ 0.00006717, -0.00000938, -0.00010638, 0.00000040, 0.00008347, 0.00003911, 0.00008973, -0.00001162, -0.00007137, -0.00000360, -0.00001770, -0.00001178, 0.00000510, -0.00000433, -0.00002011, 0.00000248, -0.00000753, 0.00000374, -0.00000463, -0.00000237, 0.00001546, -0.00000674, 0.00000695, -0.00000557, -0.00000402, -0.00000080, 0.00000123, -0.00000162, 0.00000387, 0.00000194, -0.00000196, 0.00000185, -0.00001072, 0.00000466, -0.00000107, -0.00000025, -0.00000104, -0.00000182, 0.00000133, -0.00000138, 0.00000104, 0.00000065, -0.00000054, -0.00000005, 0.00000855, -0.00000262, -0.00000286, -0.00000205, -0.00000313, -0.00000089, 0.00000074, 0.00000082, 0.00000002, 0.00000272, 0.00000304, -0.00000009, -0.00000562, 0.00000144, -0.00000055, -0.00000064, -0.00000109, -0.00000059, -0.00000147, 0.00000092, 0.00000384, -0.00000137, -0.00000111, -0.00000017, 0.00000399, -0.00000084, -0.00000052, -0.00000136, -0.00000030, -0.00000120, 0.00000039, -0.00000040, -0.00000534, 0.00000173, 0.00000338, 0.00000180, -0.00000038, 0.00000169, -0.00000042, -0.00000088, -0.00000319, -0.00000146, -0.00000470, 0.00000132, 0.00000710, -0.00000146, 0.00000029, 0.00000113, 0.00000074, -0.00000166, 0.00000135, -0.00000377, 0.00000158, -0.00000241, 0.00000293, 0.00000014, -0.00000958, 0.00000326, -0.00000354, 0.00000435, 0.00000202, 0.00000034, -0.00000138, 0.00000153, -0.00000812, 0.00000000, 0.00000281, -0.00000372, 0.00001664, -0.00000846, 0.00000004, 0.00000010, -0.00000240, -0.00000538, -0.00001715, -0.00000083, 0.00000380, -0.00001644, -0.00001014, -0.00001514, -0.00006944, 0.00002810, 0.00006944, 0.00003117, },
		{ -0.00000733, -0.00010602, -0.00000246, 0.00007949, 0.00003643, 0.00008528, -0.00000003, -0.00007305, -0.00000550, -0.00001619, -0.00001284, 0.00000472, -0.00000314, -0.00001986, 0.00000072, -0.00000609, 0.00000254, -0.00000321, -0.00000428, 0.00001635, -0.00000654, 0.00000629, -0.00000435, -0.00000504, -0.00000018, 0.00000045, -0.00000091, 0.00000323, 0.00000271, -0.00000260, 0.00000255, -0.00001083, 0.00000380, -0.00000014, -0.00000081, -0.00000030, -0.00000251, 0.00000151, -0.00000155, 0.00000107, 0.00000070, -0.00000026, -0.00000059, 0.00000842, -0.00000171, -0.00000341, -0.00000178, -0.00000321, -0.00000090, 0.00000043, 0.00000112, -0.00000020, 0.00000258, 0.00000312, 0.00000042, -0.00000567, 0.00000103, -0.00000032, -0.00000076, -0.00000091, -0.00000078, -0.00000122, 0.00000039, 0.00000401, -0.00000112, -0.00000113, -0.00000051, 0.00000408, -0.00000050, -0.00000072, -0.00000118, -0.00000051, -0.00000105, 0.00000016, 0.00000010, -0.00000552, 0.00000132, 0.00000330, 0.00000211, -0.00000039, 0.00000156, -0.00000015, -0.00000089, -0.00000323, -0.00000148, -0.00000436, 0.00000029, 0.00000768, -0.00000130, 0.00000016, 0.00000093, 0.00000093, -0.00000170, 0.00000153, -0.00000351, 0.00000106, -0.00000194, 0.00000192, 0.00000148, -0.00001022, 0.00000324, -0.00000339, 0.00000391, 0.00000231, 0.00000008, -0.00000088, 0.00000102, -0.00000713, -0.00000151, 0.00000418, -0.00000502, 0.00001697, -0.00000737, -0.00000082, 0.00000064, -0.00000340, -0.00000335, -0.00001827, -0.00000147, 0.00000416, -0.00001530, -0.00001224, -0.00001146, -0.00007198, 0.00001995, 0.00007493, 0.00003255, 0.00007132, },
		},
		{ { -0.00000653, -0.00000790, -0.00001290, -0.00001183, -0.00006042, 0.00000954, 0.00008186, 0.00003797, 0.00008599, 0.00001270, -0.00006280, -0.00001551, 0.00000427, -0.00000812, -0.00000693, -0.00000560, -0.00000427, -0.00000059, -0.00001124, 0.00000203, -0.00001079, -0.00000228, 0.00000289, 0.00000131, -0.00001395, 0.00000091, 0.00000224, -0.00000386, 0.00000063, -0.00000751, 0.00002120, -0.00000610, -0.00000302, 0.00000026, 0.00000172, 0.00000427, 0.00000002, 0.00000317, -0.00000776, -0.00000095, 0.00000588, -0.00000043, -0.00000948, 0.00000086, 0.00000414, 0.00000234, 0.00000214, 0.00000285, 0.00000069, -0.00000130, -0.00000840, 0.00000042, -0.00000012, 0.00000056, 0.00000432, -0.00000254, -0.00000386, -0.00000037, 0.00000720, -0.00000132, -0.00000179, -0.00000192, -0.00000271, 0.00000164, 0.00000242, 0.00000164, -0.00000271, -0.00000192, -0.00000179, -0.00000132, 0.00000720, -0.00000037, -0.00000386, -0.00000254, 0.00000432, 0.00000056, -0.00000012, 0.00000042, -0.00000840, -0.00000130, 0.00000069, 0.00000285, 0.00000214, 0.00000234, 0.00000414, 0.00000086, -0.00000948, -0.00000043, 0.00000588, -0.00000095, -0.00000776, 0.00000317, 0.00000002, 0.00000427, 0.00000172, 0.00000026, -0.00000302, -0.00000610, 0.00002120, -0.00000751, 0.00000063, -0.00000386, 0.00000224, 0.00000091, -0.00001395, 0.00000131, 0.00000289, -0.00000228, -0.00001079, 0.00000203, -0.00001124, -0.00000059, -0.00000427, -0.00000560, -0.00000693, -0.00000812, 0.00000427, -0.00001551, -0.00006280, 0.00001270, 0.00008599, 0.00003797, 0.00008186, 0.00000954, -0.00006042, -0.00001183, -0.00001290, -0.00000790, },
		{ -0.00000748, -0.00001272, -0.00001033, -0.00005602, -0.00000107, 0.00007971, 0.00003666, 0.00007809, 0.00002991, -0.00006044, -0.00002233, 0.00000586, -0.00000798, -0.00000737, -0.00000547, -0.00000435, -0.00000039, -0.00001140, 0.00000192, -0.00000965, -0.00000379, 0.00000274, 0.00000231, -0.00001244, -0.00000262, 0.00000462, -0.00000517, 0.00000237, -0.00000955, 0.00001995, -0.00000227, -0.00000565, 0.00000154, 0.00000024, 0.00000527, -0.00000057, 0.00000413, -0.00000728, -0.00000227, 0.00000547, 0.00000114, -0.00000963, 0.00000004, 0.00000395, 0.00000256, 0.00000195, 0.00000291, 0.00000087, -0.00000044, -0.00000803, -0.00000142, 0.00000095, -0.00000032, 0.00000505, -0.00000251, -0.00000328, -0.00000186, 0.00000734, -0.00000016, -0.00000224, -0.00000187, -0.00000278, 0.00000148, 0.00000230, 0.00000203, -0.00000278, -0.00000194, -0.00000140, -0.00000212, 0.00000652, 0.00000135, -0.00000438, -0.00000255, 0.00000355, 0.00000142, -0.00000109, 0.00000181, -0.00000810, -0.00000251, 0.00000070, 0.00000252, 0.00000258, 0.00000191, 0.00000441, 0.00000157, -0.00000906, -0.00000205, 0.00000605, 0.00000034, -0.00000784, 0.00000190, 0.00000077, 0.00000310, 0.00000326, -0.00000100, -0.00000061, -0.00000902, 0.00002098, -0.00000431, -0.00000174, -0.00000200, -0.00000048, 0.00000409, -0.00001467, -0.00000010, 0.00000294, -0.00000073, -0.00001170, 0.00000195, -0.00001102, -0.00000088, -0.00000399, -0.00000596, -0.00000636, -0.00000833, 0.00000295, -0.00001040, -0.00006112, -0.00000657, 0.00009153, 0.00004149, 0.00008238, 0.00002145, -0.00006438, -0.00001390, -0.00001288, -0.00000850, -0.00000648, },
		{ -0.00001237, -0.00000927, -0.00005141, -0.00001024, 0.00007602, 0.00003710, 0.00006895, 0.00004412, -0.00005420, -0.00003007, 0.00000707, -0.00000747, -0.00000797, -0.00000533, -0.00000445, -0.00000007, -0.00001165, 0.00000174, -0.00000841, -0.00000514, 0.00000243, 0.00000297, -0.00001038, -0.00000606, 0.00000627, -0.00000569, 0.00000330, -0.00001038, 0.00001740, 0.00000213, -0.00000816, 0.00000259, -0.00000101, 0.00000597, -0.00000087, 0.00000470, -0.00000641, -0.00000356, 0.00000487, 0.00000257, -0.00000944, -0.00000095, 0.00000391, 0.00000251, 0.00000201, 0.00000274, 0.00000121, 0.00000008, -0.00000711, -0.00000348, 0.00000187, -0.00000104, 0.00000558, -0.00000234, -0.00000281, -0.00000296, 0.00000686, 0.00000134, -0.00000273, -0.00000176, -0.00000302, 0.00000153, 0.00000197, 0.00000255, -0.00000290, -0.00000198, -0.00000106, -0.00000258, 0.00000543, 0.00000308, -0.00000467, -0.00000270, 0.00000289, 0.00000212, -0.00000179, 0.00000260, -0.00000711, -0.00000402, 0.00000086, 0.00000196, 0.00000321, 0.00000137, 0.00000463, 0.00000226, -0.00000846, -0.00000359, 0.00000592, 0.00000159, -0.00000756, 0.00000044, 0.00000154, 0.00000193, 0.00000467, -0.00000202, 0.00000131, -0.00001082, 0.00001931, -0.00000023, -0.00000440, 0.00000002, -0.00000308, 0.00000654, -0.00001442, -0.00000193, 0.00000294, 0.00000070, -0.00001223, 0.00000154, -0.00001057, -0.00000150, -0.00000326, -0.00000678, -0.00000539, -0.00000896, 0.00000238, -0.00000742, -0.00005573, -0.00002655, 0.00009352, 0.00004757, 0.00008127, 0.00003447, -0.00006763, -0.00001663, -0.00001263, -0.00000928, -0.00000636, -0.00000722, },
		{ -0.00000851, -0.00004682, -0.00001790, 0.00007093, 0.00003881, 0.00005950, 0.00005482, -0.00004472, -0.00003775, 0.00000722, -0.00000622, -0.00000895, -0.00000499, -0.00000478, 0.00000052, -0.00001208, 0.00000154, -0.00000717, -0.00000624, 0.00000193, 0.00000336, -0.00000802, -0.00000906, 0.00000693, -0.00000531, 0.00000342, -0.00001009, 0.00001385, 0.00000667, -0.00001022, 0.00000318, -0.00000185, 0.00000625, -0.00000081, 0.00000485, -0.00000520, -0.00000479, 0.00000410, 0.00000377, -0.00000888, -0.00000210, 0.00000402, 0.00000224, 0.00000225, 0.00000242, 0.00000164, 0.00000034, -0.00000581, -0.00000549, 0.00000243, -0.00000143, 0.00000581, -0.00000191, -0.00000258, -0.00000354, 0.00000577, 0.00000308, -0.00000318, -0.00000159, -0.00000342, 0.00000175, 0.00000152, 0.00000311, -0.00000295, -0.00000211, -0.00000074, -0.00000278, 0.00000413, 0.00000459, -0.00000454, -0.00000310, 0.00000244, 0.00000253, -0.00000211, 0.00000275, -0.00000552, -0.00000568, 0.00000105, 0.00000127, 0.00000389, 0.00000089, 0.00000468, 0.00000301, -0.00000777, -0.00000495, 0.00000548, 0.00000278, -0.00000697, -0.00000108, 0.00000218, 0.00000091, 0.00000579, -0.00000265, 0.00000259, -0.00001145, 0.00001638, 0.00000435, -0.00000692, 0.00000178, -0.00000514, 0.00000795, -0.00001312, -0.00000413, 0.00000292, 0.00000191, -0.00001228, 0.00000072, -0.00000974, -0.00000261, -0.00000194, -0.00000816, -0.00000392, -0.00001016, 0.00000276, -0.00000654, -0.00004750, -0.00004556, 0.00009081, 0.00005634, 0.00007869, 0.00004833, -0.00006990, -0.00002009, -0.00001218, -0.00001020, -0.00000622, -0.00000706, -0.00001194, },
		{ -0.00004246, -0.00002400, 0.00006464, 0.00004134, 0.00005044, 0.00006194, -0.00003295, -0.00004431, 0.00000577, -0.00000402, -0.00001041, -0.00000437, -0.00000540, 0.00000141, -0.00001266, 0.00000130, -0.00000596, -0.00000707, 0.00000127, 0.00000351, -0.00000558, -0.00001134, 0.00000650, -0.00000412, 0.00000288, -0.00000894, 0.00000974, 0.00001088, -0.00001149, 0.00000314, -0.00000219, 0.00000604, -0.00000036, 0.00000459, -0.00000373, -0.00000588, 0.00000320, 0.00000469, -0.00000795, -0.00000339, 0.00000426, 0.00000182, 0.00000255, 0.00000206, 0.00000206, 0.00000046, -0.00000434, -0.00000721, 0.00000246, -0.00000142, 0.00000566, -0.00000116, -0.00000265, -0.00000360, 0.00000417, 0.00000490, -0.00000350, -0.00000141, -0.00000393, 0.00000202, 0.00000108, 0.00000357, -0.00000282, -0.00000242, -0.00000041, -0.00000280, 0.00000279, 0.00000571, -0.00000392, -0.00000378, 0.00000223, 0.00000263, -0.00000202, 0.00000232, -0.00000351, -0.00000728, 0.00000112, 0.00000060, 0.00000445, 0.00000063, 0.00000441, 0.00000389, -0.00000706, -0.00000605, 0.00000471, 0.00000389, -0.00000614, -0.00000252, 0.00000258, 0.00000012, 0.00000653, -0.00000281, 0.00000318, -0.00001100, 0.00001256, 0.00000893, -0.00000886, 0.00000289, -0.00000631, 0.00000815, -0.00001086, -0.00000655, 0.00000284, 0.00000285, -0.00001181, -0.00000054, -0.00000851, -0.00000424, -0.00000005, -0.00001005, -0.00000199, -0.00001188, 0.00000395, -0.00000730, -0.00003770, -0.00006186, 0.00008252, 0.00006763, 0.00007499, 0.00006265, -0.00007090, -0.00002436, -0.00001156, -0.00001119, -0.00000611, -0.00000693, -0.00001152, -0.00000789, },
		{ -0.00002855, 0.00005737, 0.00004427, 0.00004228, 0.00006572, -0.00001997, -0.00004886, 0.00000244, -0.00000091, -0.00001223, -0.00000358, -0.00000623, 0.00000250, -0.00001325, 0.00000091, -0.00000472, -0.00000768, 0.00000053, 0.00000342, -0.00000320, -0.00001278, 0.00000505, -0.00000236, 0.00000197, -0.00000730, 0.00000553, 0.00001434, -0.00001175, 0.00000238, -0.00000199, 0.00000535, 0.00000043, 0.00000401, -0.00000212, -0.00000676, 0.00000219, 0.00000531, -0.00000670, -0.00000473, 0.00000453, 0.00000139, 0.00000277, 0.00000180, 0.00000237, 0.00000053, -0.00000287, -0.00000842, 0.00000186, -0.00000099, 0.00000516, -0.00000012, -0.00000298, -0.00000322, 0.00000223, 0.00000659, -0.00000355, -0.00000129, -0.00000445, 0.00000223, 0.00000079, 0.00000379, -0.00000241, -0.00000292, -0.00000004, -0.00000273, 0.00000156, 0.00000630, -0.00000279, -0.00000471, 0.00000223, 0.00000246, -0.00000160, 0.00000145, -0.00000135, -0.00000857, 0.00000091, 0.00000011, 0.00000474, 0.00000074, 0.00000376, 0.00000493, -0.00000639, -0.00000686, 0.00000365, 0.00000487, -0.00000512, -0.00000378, 0.00000267, -0.00000038, 0.00000687, -0.00000249, 0.00000314, -0.00000969, 0.00000825, 0.00001302, -0.00000984, 0.00000309, -0.00000644, 0.00000719, -0.00000788, -0.00000896, 0.00000260, 0.00000353, -0.00001086, -0.00000215, -0.00000693, -0.00000632, 0.00000226, -0.00001224, 0.00000012, -0.00001382, 0.00000553, -0.00000893, -0.00002784, -0.00007381, 0.00006824, 0.00008086, 0.00007076, 0.00007694, -0.00007026, -0.00002950, -0.00001081, -0.00001221, -0.00000610, -0.00000677, -0.00001119, -0.00000731, -0.00003848, },
		{ 0.00004941, 0.00004723, 0.00003531, 0.00006661, -0.00000683, -0.00005081, -0.00000270, 0.00000279, -0.00001410, -0.00000285, -0.00000707, 0.00000357, -0.00001363, 0.00000020, -0.00000335, -0.00000817, -0.00000017, 0.00000306, -0.00000097, -0.00001335, 0.00000281, -0.00000041, 0.00000104, -0.00000553, 0.00000166, 0.00001670, -0.00001087, 0.00000093, -0.00000132, 0.00000429, 0.00000147, 0.00000324, -0.00000051, -0.00000735, 0.00000109, 0.00000563, -0.00000522, -0.00000604, 0.00000472, 0.00000107, 0.00000280, 0.00000171, 0.00000250, 0.00000065, -0.00000157, -0.00000902, 0.00000063, -0.00000022, 0.00000438, 0.00000111, -0.00000349, -0.00000253, 0.00000018, 0.00000794, -0.00000322, -0.00000129, -0.00000487, 0.00000223, 0.00000077, 0.00000370, -0.00000170, -0.00000361, 0.00000036, -0.00000261, 0.00000056, 0.00000633, -0.00000121, -0.00000576, 0.00000232, 0.00000213, -0.00000099, 0.00000039, 0.00000067, -0.00000931, 0.00000025, -0.00000007, 0.00000461, 0.00000131, 0.00000270, 0.00000608, -0.00000574, -0.00000738, 0.00000236, 0.00000568, -0.00000396, -0.00000478, 0.00000242, -0.00000061, 0.00000684, -0.00000177, 0.00000259, -0.00000780, 0.00000394, 0.00001617, -0.00000960, 0.00000227, -0.00000558, 0.00000531, -0.00000455, -0.00001102, 0.00000204, 0.00000405, -0.00000958, -0.00000396, -0.00000516, -0.00000864, 0.00000470, -0.00001434, 0.00000203, -0.00001552, 0.00000690, -0.00001045, -0.00001937, -0.00008019, 0.00004813, 0.00009495, 0.00006680, 0.00009062, -0.00006759, -0.00003560, -0.00000993, -0.00001325, -0.00000619, -0.00000655, -0.00001100, -0.00000667, -0.00003496, -0.00003166, },
		{ 0.00004984, 0.00002973, 0.00006511, 0.00000560, -0.00004994, -0.00000921, 0.00000653, -0.00001556, -0.00000253, -0.00000764, 0.00000433, -0.00001353, -0.00000102, -0.00000175, -0.00000865, -0.00000071, 0.00000238, 0.00000108, -0.00001316, 0.00000010, 0.00000135, 0.00000042, -0.00000396, -0.00000153, 0.00001780, -0.00000890, -0.00000109, -0.00000032, 0.00000297, 0.00000261, 0.00000241, 0.00000096, -0.00000759, -0.00000008, 0.00000566, -0.00000360, -0.00000718, 0.00000472, 0.00000098, 0.00000254, 0.00000184, 0.00000246, 0.00000084, -0.00000052, -0.00000897, -0.00000110, 0.00000072, 0.00000349, 0.00000239, -0.00000402, -0.00000174, -0.00000172, 0.00000874, -0.00000244, -0.00000148, -0.00000511, 0.00000192, 0.00000109, 0.00000327, -0.00000071, -0.00000444, 0.00000073, -0.00000247, -0.00000018, 0.00000585, 0.00000065, -0.00000676, 0.00000236, 0.00000180, -0.00000038, -0.00000064, 0.00000228, -0.00000931, -0.00000093, 0.00000013, 0.00000402, 0.00000235, 0.00000130, 0.00000725, -0.00000505, -0.00000764, 0.00000094, 0.00000626, -0.00000270, -0.00000547, 0.00000187, -0.00000064, 0.00000652, -0.00000076, 0.00000174, -0.00000566, 0.00000001, 0.00001808, -0.00000807, 0.00000054, -0.00000396, 0.00000293, -0.00000137, -0.00001239, 0.00000101, 0.00000452, -0.00000816, -0.00000574, -0.00000345, -0.00001089, 0.00000687, -0.00001594, 0.00000326, -0.00001646, 0.00000739, -0.00001094, -0.00001352, -0.00008038, 0.00002311, 0.00010842, 0.00006408, 0.00010302, -0.00006249, -0.00004275, -0.00000893, -0.00001433, -0.00000635, -0.00000630, -0.00001093, -0.00000597, -0.00003191, -0.00003351, 0.00004108, },
		{ 0.00002568, 0.00006173, 0.00001674, -0.00004645, -0.00001635, 0.00000959, -0.00001609, -0.00000297, -0.00000765, 0.00000451, -0.00001271, -0.00000290, 0.00000017, -0.00000924, -0.00000097, 0.00000137, 0.00000293, -0.00001235, -0.00000271, 0.00000255, 0.00000036, -0.00000281, -0.00000385, 0.00001760, -0.00000601, -0.00000342, 0.00000080, 0.00000160, 0.00000369, 0.00000168, 0.00000217, -0.00000745, -0.00000129, 0.00000543, -0.00000193, -0.00000808, 0.00000446, 0.00000119, 0.00000199, 0.00000215, 0.00000228, 0.00000109, 0.00000029, -0.00000834, -0.00000316, 0.00000161, 0.00000266, 0.00000355, -0.00000441, -0.00000103, -0.00000326, 0.00000886, -0.00000116, -0.00000184, -0.00000512, 0.00000127, 0.00000174, 0.00000256, 0.00000048, -0.00000527, 0.00000100, -0.00000229, -0.00000067, 0.00000497, 0.00000257, -0.00000752, 0.00000216, 0.00000164, 0.00000006, -0.00000139, 0.00000325, -0.00000849, -0.00000263, 0.00000069, 0.00000298, 0.00000376, -0.00000027, 0.00000828, -0.00000424, -0.00000775, -0.00000049, 0.00000653, -0.00000137, -0.00000587, 0.00000108, -0.00000055, 0.00000601, 0.00000040, 0.00000078, -0.00000354, -0.00000320, 0.00001863, -0.00000540, -0.00000179, -0.00000199, 0.00000054, 0.00000119, -0.00001279, -0.00000062, 0.00000507, -0.00000677, -0.00000726, -0.00000204, -0.00001275, 0.00000839, -0.00001660, 0.00000340, -0.00001617, 0.00000645, -0.00000967, -0.00001102, -0.00007449, -0.00000523, 0.00011942, 0.00006360, 0.00011352, -0.00005458, -0.00005099, -0.00000777, -0.00001551, -0.00000651, -0.00000610, -0.00001089, -0.00000529, -0.00002922, -0.00003436, 0.00003275, 0.00005176, },
		{ 0.00005689, 0.00002628, -0.00004086, -0.00002320, 0.00001125, -0.00001525, -0.00000444, -0.00000689, 0.00000393, -0.00001104, -0.00000547, 0.00000241, -0.00000994, -0.00000092, 0.00000007, 0.00000454, -0.00001109, -0.00000531, 0.00000293, 0.00000098, -0.00000216, -0.00000526, 0.00001626, -0.00000252, -0.00000577, 0.00000182, 0.00000034, 0.00000456, 0.00000118, 0.00000303, -0.00000694, -0.00000249, 0.00000496, -0.00000032, -0.00000865, 0.00000388, 0.00000171, 0.00000119, 0.00000255, 0.00000205, 0.00000134, 0.00000087, -0.00000726, -0.00000529, 0.00000223, 0.00000208, 0.00000445, -0.00000453, -0.00000058, -0.00000426, 0.00000824, 0.00000054, -0.00000234, -0.00000489, 0.00000028, 0.00000265, 0.00000168, 0.00000170, -0.00000598, 0.00000107, -0.00000204, -0.00000097, 0.00000387, 0.00000434, -0.00000782, 0.00000161, 0.00000177, 0.00000018, -0.00000173, 0.00000349, -0.00000687, -0.00000473, 0.00000150, 0.00000165, 0.00000536, -0.00000179, 0.00000897, -0.00000319, -0.00000778, -0.00000181, 0.00000645, 0.00000001, -0.00000597, 0.00000015, -0.00000045, 0.00000543, 0.00000157, -0.00000013, -0.00000168, -0.00000551, 0.00001786, -0.00000190, -0.00000429, -0.00000016, -0.00000131, 0.00000270, -0.00001201, -0.00000287, 0.00000574, -0.00000558, -0.00000834, -0.00000112, -0.00001397, 0.00000894, -0.00001603, 0.00000217, -0.00001436, 0.00000377, -0.00000631, -0.00001203, -0.00006343, -0.00003466, 0.00012597, 0.00006629, 0.00012159, -0.00004360, -0.00006030, -0.00000649, -0.00001683, -0.00000659, -0.00000605, -0.00001079, -0.00000473, -0.00002674, -0.00003455, 0.00002482, 0.00005268, 0.00002321, },
		{ 0.00003418, -0.00003392, -0.00002885, 0.00001095, -0.00001279, -0.00000703, -0.00000533, 0.00000255, -0.00000854, -0.00000864, 0.00000482, -0.00001070, -0.00000062, -0.00000143, 0.00000585, -0.00000950, -0.00000743, 0.00000236, 0.00000222, -0.00000192, -0.00000589, 0.00001404, 0.00000120, -0.00000781, 0.00000251, -0.00000063, 0.00000508, 0.00000099, 0.00000353, -0.00000611, -0.00000360, 0.00000427, 0.00000117, -0.00000886, 0.00000300, 0.00000248, 0.00000027, 0.00000291, 0.00000189, 0.00000150, 0.00000131, -0.00000591, -0.00000723, 0.00000237, 0.00000187, 0.00000497, -0.00000427, -0.00000047, -0.00000467, 0.00000692, 0.00000255, -0.00000287, -0.00000449, -0.00000096, 0.00000367, 0.00000080, 0.00000280, -0.00000641, 0.00000087, -0.00000166, -0.00000117, 0.00000271, 0.00000572, -0.00000754, 0.00000061, 0.00000225, -0.00000007, -0.00000158, 0.00000300, -0.00000459, -0.00000703, 0.00000236, 0.00000020, 0.00000692, -0.00000301, 0.00000914, -0.00000183, -0.00000783, -0.00000290, 0.00000597, 0.00000142, -0.00000583, -0.00000082, -0.00000045, 0.00000487, 0.00000264, -0.00000086, -0.00000022, -0.00000690, 0.00001599, 0.00000199, -0.00000646, 0.00000103, -0.00000221, 0.00000292, -0.00001006, -0.00000564, 0.00000651, -0.00000464, -0.00000891, -0.00000081, -0.00001440, 0.00000837, -0.00001412, -0.00000048, -0.00001104, -0.00000062, -0.00000100, -0.00001604, -0.00004881, -0.00006262, 0.00012620, 0.00007272, 0.00012698, -0.00002946, -0.00007054, -0.00000517, -0.00001832, -0.00000651, -0.00000625, -0.00001049, -0.00000445, -0.00002429, -0.00003440, 0.00001767, 0.00005231, 0.00002236, 0.00005094, },
		{ -0.00002644, -0.00003253, 0.00000839, -0.00000875, -0.00001057, -0.00000313, 0.00000052, -0.00000539, -0.00001215, 0.00000718, -0.00001133, -0.00000021, -0.00000294, 0.00000678, -0.00000768, -0.00000895, 0.00000090, 0.00000387, -0.00000189, -0.00000596, 0.00001133, 0.00000474, -0.00000926, 0.00000270, -0.00000120, 0.00000521, 0.00000112, 0.00000367, -0.00000505, -0.00000456, 0.00000337, 0.00000248, -0.00000870, 0.00000187, 0.00000340, -0.00000062, 0.00000306, 0.00000191, 0.00000151, 0.00000168, -0.00000445, -0.00000876, 0.00000188, 0.00000208, 0.00000509, -0.00000363, -0.00000073, -0.00000451, 0.00000505, 0.00000465, -0.00000329, -0.00000401, -0.00000231, 0.00000464, 0.00000010, 0.00000361, -0.00000645, 0.00000034, -0.00000116, -0.00000133, 0.00000166, 0.00000658, -0.00000660, -0.00000082, 0.00000306, -0.00000067, -0.00000100, 0.00000192, -0.00000193, -0.00000922, 0.00000304, -0.00000110, 0.00000816, -0.00000368, 0.00000864, -0.00000010, -0.00000795, -0.00000370, 0.00000511, 0.00000281, -0.00000546, -0.00000172, -0.00000061, 0.00000437, 0.00000357, -0.00000136, 0.00000082, -0.00000745, 0.00001337, 0.00000574, -0.00000781, 0.00000121, -0.00000190, 0.00000180, -0.00000712, -0.00000867, 0.00000721, -0.00000390, -0.00000898, -0.00000108, -0.00001405, 0.00000671, -0.00001099, -0.00000434, -0.00000650, -0.00000630, 0.00000568, -0.00002200, -0.00003269, -0.00008646, 0.00011865, 0.00008302, 0.00012976, -0.00001234, -0.00008135, -0.00000400, -0.00001992, -0.00000624, -0.00000677, -0.00000991, -0.00000455, -0.00002168, -0.00003423, 0.00001160, 0.00005049, 0.00002307, 0.00004416, 0.00004056, },
		{ -0.00003378, 0.00000367, -0.00000350, -0.00001465, -0.00000067, -0.00000182, -0.00000196, -0.00001561, 0.00000913, -0.00001160, 0.00000010, -0.00000424, 0.00000718, -0.00000567, -0.00000984, -0.00000127, 0.00000562, -0.00000179, -0.00000576, 0.00000849, 0.00000774, -0.00000989, 0.00000232, -0.00000135, 0.00000493, 0.00000156, 0.00000355, -0.00000389, -0.00000528, 0.00000229, 0.00000358, -0.00000821, 0.00000058, 0.00000435, -0.00000132, 0.00000288, 0.00000220, 0.00000132, 0.00000203, -0.00000304, -0.00000973, 0.00000074, 0.00000268, 0.00000487, -0.00000266, -0.00000129, -0.00000389, 0.00000284, 0.00000660, -0.00000345, -0.00000356, -0.00000362, 0.00000535, -0.00000025, 0.00000400, -0.00000603, -0.00000051, -0.00000052, -0.00000151, 0.00000083, 0.00000683, -0.00000506, -0.00000258, 0.00000409, -0.00000149, -0.00000015, 0.00000049, 0.00000078, -0.00001099, 0.00000327, -0.00000202, 0.00000885, -0.00000362, 0.00000739, 0.00000197, -0.00000815, -0.00000419, 0.00000392, 0.00000411, -0.00000489, -0.00000250, -0.00000097, 0.00000394, 0.00000435, -0.00000162, 0.00000148, -0.00000734, 0.00001041, 0.00000887, -0.00000799, 0.00000019, -0.00000036, -0.00000043, -0.00000360, -0.00001157, 0.00000761, -0.00000320, -0.00000872, -0.00000175, -0.00001307, 0.00000419, -0.00000699, -0.00000894, -0.00000131, -0.00001257, 0.00001278, -0.00002846, -0.00001736, -0.00010383, 0.00010254, 0.00009663, 0.00013042, 0.00000730, -0.00009220, -0.00000330, -0.00002150, -0.00000585, -0.00000758, -0.00000902, -0.00000511, -0.00001879, -0.00003426, 0.00000683, 0.00004715, 0.00002518, 0.00003685, 0.00004559, -0.00001912, },
		{ -0.00000274, 0.00000229, -0.00001862, 0.00000152, -0.00000401, 0.00000129, -0.00001855, 0.00001030, -0.00001124, 0.00000006, -0.00000510, 0.00000692, -0.00000347, -0.00001016, -0.00000384, 0.00000707, -0.00000131, -0.00000557, 0.00000587, 0.00000991, -0.00000962, 0.00000137, -0.00000112, 0.00000432, 0.00000220, 0.00000326, -0.00000274, -0.00000569, 0.00000105, 0.00000443, -0.00000743, -0.00000078, 0.00000518, -0.00000168, 0.00000229, 0.00000276, 0.00000096, 0.00000238, -0.00000177, -0.00001006, -0.00000099, 0.00000352, 0.00000443, -0.00000150, -0.00000201, -0.00000299, 0.00000057, 0.00000815, -0.00000319, -0.00000325, -0.00000473, 0.00000566, -0.00000014, 0.00000392, -0.00000517, -0.00000163, 0.00000018, -0.00000171, 0.00000028, 0.00000650, -0.00000304, -0.00000448, 0.00000514, -0.00000234, 0.00000077, -0.00000100, 0.00000318, -0.00001202, 0.00000285, -0.00000238, 0.00000881, -0.00000273, 0.00000543, 0.00000429, -0.00000835, -0.00000441, 0.00000249, 0.00000524, -0.00000415, -0.00000310, -0.00000151, 0.00000354, 0.00000501, -0.00000170, 0.00000184, -0.00000681, 0.00000750, 0.00001102, -0.00000682, -0.00000201, 0.00000216, -0.00000331, -0.00000008, -0.00001386, 0.00000739, -0.00000234, -0.00000835, -0.00000258, -0.00001176, 0.00000118, -0.00000261, -0.00001367, 0.00000381, -0.00001854, 0.00001916, -0.00003390, -0.00000493, -0.00011301, 0.00007803, 0.00011230, 0.00012985, 0.00002873, -0.00010235, -0.00000354, -0.00002281, -0.00000549, -0.00000859, -0.00000788, -0.00000607, -0.00001563, -0.00003460, 0.00000339, 0.00004243, 0.00002842, 0.00002935, 0.00004941, -0.00001245, -0.00003254, },
		{ 0.00000774, -0.00002174, 0.00000286, -0.00000556, 0.00000387, -0.00002052, 0.00001034, -0.00001002, -0.00000053, -0.00000531, 0.00000593, -0.00000109, -0.00001004, -0.00000648, 0.00000788, -0.00000024, -0.00000559, 0.00000370, 0.00001113, -0.00000849, -0.00000002, -0.00000062, 0.00000349, 0.00000294, 0.00000293, -0.00000172, -0.00000577, -0.00000030, 0.00000502, -0.00000643, -0.00000210, 0.00000576, -0.00000161, 0.00000128, 0.00000355, 0.00000049, 0.00000270, -0.00000066, -0.00000978, -0.00000316, 0.00000440, 0.00000397, -0.00000032, -0.00000272, -0.00000204, -0.00000151, 0.00000910, -0.00000243, -0.00000314, -0.00000555, 0.00000546, 0.00000047, 0.00000339, -0.00000396, -0.00000289, 0.00000086, -0.00000191, -0.00000001, 0.00000568, -0.00000077, -0.00000629, 0.00000601, -0.00000302, 0.00000151, -0.00000226, 0.00000495, -0.00001209, 0.00000165, -0.00000209, 0.00000800, -0.00000104, 0.00000291, 0.00000667, -0.00000846, -0.00000443, 0.00000095, 0.00000609, -0.00000322, -0.00000352, -0.00000218, 0.00000312, 0.00000560, -0.00000167, 0.00000203, -0.00000610, 0.00000497, 0.00001196, -0.00000437, -0.00000508, 0.00000519, -0.00000622, 0.00000281, -0.00001505, 0.00000628, -0.00000108, -0.00000816, -0.00000327, -0.00001045, -0.00000186, 0.00000155, -0.00001782, 0.00000807, -0.00002327, 0.00002372, -0.00003690, 0.00000296, -0.00011321, 0.00004630, 0.00012810, 0.00012928, 0.00005102, -0.00011092, -0.00000524, -0.00002354, -0.00000539, -0.00000961, -0.00000666, -0.00000732, -0.00001228, -0.00003521, 0.00000116, 0.00003662, 0.00003235, 0.00002205, 0.00005206, -0.00000664, -0.00002915, -0.00000999, },
		{ -0.00002328, 0.00000282, -0.00000602, 0.00000537, -0.00002116, 0.00000902, -0.00000780, -0.00000178, -0.00000477, 0.00000419, 0.00000141, -0.00000958, -0.00000887, 0.00000776, 0.00000155, -0.00000590, 0.00000208, 0.00001141, -0.00000664, -0.00000165, -0.00000002, 0.00000258, 0.00000364, 0.00000267, -0.00000091, -0.00000550, -0.00000169, 0.00000533, -0.00000527, -0.00000328, 0.00000604, -0.00000108, -0.00000010, 0.00000443, 0.00000005, 0.00000290, 0.00000030, -0.00000899, -0.00000551, 0.00000507, 0.00000367, 0.00000070, -0.00000324, -0.00000124, -0.00000317, 0.00000930, -0.00000115, -0.00000325, -0.00000601, 0.00000473, 0.00000152, 0.00000255, -0.00000256, -0.00000412, 0.00000140, -0.00000204, -0.00000011, 0.00000454, 0.00000151, -0.00000775, 0.00000647, -0.00000330, 0.00000186, -0.00000303, 0.00000587, -0.00001111, -0.00000032, -0.00000117, 0.00000649, 0.00000127, 0.00000009, 0.00000886, -0.00000832, -0.00000439, -0.00000053, 0.00000656, -0.00000212, -0.00000376, -0.00000291, 0.00000259, 0.00000619, -0.00000159, 0.00000216, -0.00000539, 0.00000303, 0.00001167, -0.00000091, -0.00000854, 0.00000810, -0.00000846, 0.00000446, -0.00001475, 0.00000406, 0.00000075, -0.00000835, -0.00000355, -0.00000945, -0.00000453, 0.00000494, -0.00002079, 0.00001079, -0.00002602, 0.00002559, -0.00003648, 0.00000539, -0.00010469, 0.00000953, 0.00014160, 0.00013004, 0.00007318, -0.00011700, -0.00000895, -0.00002336, -0.00000584, -0.00001037, -0.00000558, -0.00000864, -0.00000894, -0.00003592, -0.00000015, 0.00003015, 0.00003643, 0.00001544, 0.00005342, -0.00000158, -0.00002431, -0.00001704, 0.00001188, },
		{ 0.00000106, -0.00000514, 0.00000555, -0.00002029, 0.00000627, -0.00000462, -0.00000368, -0.00000350, 0.00000180, 0.00000395, -0.00000889, -0.00001077, 0.00000658, 0.00000400, -0.00000643, 0.00000096, 0.00001089, -0.00000435, -0.00000329, 0.00000050, 0.00000173, 0.00000419, 0.00000259, -0.00000037, -0.00000493, -0.00000303, 0.00000531, -0.00000399, -0.00000428, 0.00000597, -0.00000014, -0.00000167, 0.00000522, -0.00000021, 0.00000289, 0.00000117, -0.00000784, -0.00000780, 0.00000529, 0.00000371, 0.00000142, -0.00000344, -0.00000073, -0.00000425, 0.00000872, 0.00000058, -0.00000352, -0.00000613, 0.00000353, 0.00000286, 0.00000156, -0.00000116, -0.00000515, 0.00000168, -0.00000206, -0.00000010, 0.00000328, 0.00000353, -0.00000863, 0.00000637, -0.00000307, 0.00000168, -0.00000317, 0.00000584, -0.00000911, -0.00000292, 0.00000021, 0.00000447, 0.00000394, -0.00000271, 0.00001058, -0.00000780, -0.00000439, -0.00000181, 0.00000659, -0.00000085, -0.00000384, -0.00000361, 0.00000187, 0.00000682, -0.00000152, 0.00000230, -0.00000480, 0.00000175, 0.00001035, 0.00000309, -0.00001176, 0.00001021, -0.00000939, 0.00000438, -0.00001276, 0.00000071, 0.00000318, -0.00000904, -0.00000327, -0.00000897, -0.00000649, 0.00000719, -0.00002216, 0.00001156, -0.00002633, 0.00002433, -0.00003226, 0.00000232, -0.00008881, -0.00002928, 0.00015012, 0.00013338, 0.00009430, -0.00011975, -0.00001511, -0.00002198, -0.00000708, -0.00001064, -0.00000490, -0.00000976, -0.00000588, -0.00003647, -0.00000094, 0.00002358, 0.00004004, 0.00001004, 0.00005329, 0.00000303, -0.00001892, -0.00002279, 0.00001387, -0.00002271, },
		{ -0.00000292, 0.00000441, -0.00001797, 0.00000228, -0.00000070, -0.00000605, -0.00000166, -0.00000100, 0.00000633, -0.00000799, -0.00001204, 0.00000436, 0.00000690, -0.00000699, 0.00000021, 0.00000982, -0.00000192, -0.00000467, 0.00000079, 0.00000105, 0.00000451, 0.00000271, -0.00000008, -0.00000415, -0.00000420, 0.00000495, -0.00000263, -0.00000507, 0.00000557, 0.00000110, -0.00000326, 0.00000571, -0.00000013, 0.00000258, 0.00000202, -0.00000650, -0.00000976, 0.00000489, 0.00000418, 0.00000176, -0.00000326, -0.00000060, -0.00000472, 0.00000742, 0.00000262, -0.00000383, -0.00000598, 0.00000199, 0.00000430, 0.00000066, 0.00000002, -0.00000580, 0.00000162, -0.00000190, -0.00000009, 0.00000208, 0.00000507, -0.00000879, 0.00000561, -0.00000225, 0.00000094, -0.00000266, 0.00000492, -0.00000629, -0.00000586, 0.00000178, 0.00000223, 0.00000661, -0.00000512, 0.00001157, -0.00000675, -0.00000455, -0.00000277, 0.00000612, 0.00000056, -0.00000379, -0.00000420, 0.00000093, 0.00000750, -0.00000146, 0.00000246, -0.00000435, 0.00000105, 0.00000831, 0.00000705, -0.00001410, 0.00001093, -0.00000854, 0.00000237, -0.00000912, -0.00000361, 0.00000608, -0.00001017, -0.00000244, -0.00000907, -0.00000763, 0.00000811, -0.00002178, 0.00001027, -0.00002415, 0.00002000, -0.00002450, -0.00000542, -0.00006782, -0.00006664, 0.00015117, 0.00014017, 0.00011377, -0.00011857, -0.00002399, -0.00001927, -0.00000927, -0.00001021, -0.00000482, -0.00001046, -0.00000336, -0.00003656, -0.00000163, 0.00001748, 0.00004258, 0.00000633, 0.00005144, 0.00000762, -0.00001390, -0.00002629, 0.00001315, -0.00001978, -0.00000248, },
		{ 0.00000222, -0.00001449, -0.00000259, 0.00000357, -0.00000858, 0.00000046, -0.00000391, 0.00000832, -0.00000685, -0.00001266, 0.00000130, 0.00000993, -0.00000732, -0.00000040, 0.00000849, 0.00000032, -0.00000559, 0.00000074, 0.00000061, 0.00000457, 0.00000303, 0.00000001, -0.00000328, -0.00000509, 0.00000422, -0.00000124, -0.00000563, 0.00000491, 0.00000249, -0.00000463, 0.00000572, 0.00000039, 0.00000194, 0.00000289, -0.00000512, -0.00001119, 0.00000375, 0.00000509, 0.00000174, -0.00000270, -0.00000082, -0.00000462, 0.00000557, 0.00000472, -0.00000404, -0.00000569, 0.00000030, 0.00000560, 0.00000006, 0.00000080, -0.00000597, 0.00000117, -0.00000154, -0.00000014, 0.00000111, 0.00000600, -0.00000818, 0.00000424, -0.00000092, -0.00000030, -0.00000159, 0.00000329, -0.00000299, -0.00000879, 0.00000321, 0.00000012, 0.00000893, -0.00000681, 0.00001163, -0.00000511, -0.00000493, -0.00000331, 0.00000515, 0.00000204, -0.00000360, -0.00000463, -0.00000025, 0.00000820, -0.00000137, 0.00000261, -0.00000399, 0.00000077, 0.00000599, 0.00001039, -0.00001500, 0.00000984, -0.00000571, -0.00000147, -0.00000419, -0.00000848, 0.00000914, -0.00001152, -0.00000125, -0.00000960, -0.00000802, 0.00000779, -0.00001983, 0.00000718, -0.00001986, 0.00001316, -0.00001412, -0.00001626, -0.00004459, -0.00009903, 0.00014277, 0.00015062, 0.00013142, -0.00011317, -0.00003556, -0.00001531, -0.00001243, -0.00000901, -0.00000547, -0.00001055, -0.00000159, -0.00003594, -0.00000263, 0.00001237, 0.00004356, 0.00000468, 0.00004771, 0.00001260, -0.00001004, -0.00002694, 0.00000955, -0.00001465, -0.00000751, 0.00000037, },
		{ -0.00001036, -0.00000778, 0.00000767, -0.00001083, 0.00000246, -0.00000652, 0.00000965, -0.00000538, -0.00001274, -0.00000228, 0.00001266, -0.00000711, -0.00000111, 0.00000719, 0.00000212, -0.00000590, 0.00000031, 0.00000041, 0.00000440, 0.00000349, -0.00000001, -0.00000246, -0.00000562, 0.00000313, 0.00000015, -0.00000599, 0.00000408, 0.00000386, -0.00000559, 0.00000511, 0.00000139, 0.00000099, 0.00000376, -0.00000380, -0.00001198, 0.00000189, 0.00000634, 0.00000146, -0.00000187, -0.00000130, -0.00000410, 0.00000342, 0.00000664, -0.00000395, -0.00000539, -0.00000134, 0.00000654, -0.00000004, 0.00000106, -0.00000562, 0.00000036, -0.00000103, -0.00000028, 0.00000043, 0.00000627, -0.00000687, 0.00000238, 0.00000078, -0.00000185, -0.00000018, 0.00000126, 0.00000039, -0.00001131, 0.00000417, -0.00000155, 0.00001056, -0.00000752, 0.00001068, -0.00000292, -0.00000552, -0.00000345, 0.00000377, 0.00000351, -0.00000329, -0.00000487, -0.00000165, 0.00000885, -0.00000119, 0.00000268, -0.00000362, 0.00000066, 0.00000382, 0.00001258, -0.00001410, 0.00000681, -0.00000105, -0.00000672, 0.00000142, -0.00001330, 0.00001187, -0.00001272, -0.00000003, -0.00001025, -0.00000793, 0.00000657, -0.00001673, 0.00000287, -0.00001417, 0.00000480, -0.00000247, -0.00002818, -0.00002220, -0.00012333, 0.00012391, 0.00016409, 0.00014758, -0.00010365, -0.00004948, -0.00001041, -0.00001638, -0.00000713, -0.00000683, -0.00001000, -0.00000069, -0.00003444, -0.00000422, 0.00000860, 0.00004266, 0.00000525, 0.00004210, 0.00001821, -0.00000784, -0.00002456, 0.00000339, -0.00000789, -0.00001343, 0.00000417, -0.00000056, },
		{ -0.00001264, 0.00001105, -0.00001236, 0.00000395, -0.00000847, 0.00001007, -0.00000351, -0.00001242, -0.00000602, 0.00001468, -0.00000610, -0.00000210, 0.00000612, 0.00000331, -0.00000558, -0.00000044, 0.00000037, 0.00000406, 0.00000401, -0.00000002, -0.00000180, -0.00000570, 0.00000172, 0.00000149, -0.00000614, 0.00000316, 0.00000506, -0.00000599, 0.00000383, 0.00000283, -0.00000019, 0.00000459, -0.00000259, -0.00001211, -0.00000057, 0.00000775, 0.00000108, -0.00000093, -0.00000188, -0.00000336, 0.00000124, 0.00000812, -0.00000344, -0.00000520, -0.00000275, 0.00000693, 0.00000043, 0.00000080, -0.00000481, -0.00000069, -0.00000044, -0.00000051, 0.00000008, 0.00000592, -0.00000502, 0.00000027, 0.00000259, -0.00000346, 0.00000129, -0.00000080, 0.00000344, -0.00001303, 0.00000437, -0.00000252, 0.00001129, -0.00000713, 0.00000874, -0.00000030, -0.00000624, -0.00000322, 0.00000209, 0.00000483, -0.00000281, -0.00000492, -0.00000317, 0.00000933, -0.00000081, 0.00000257, -0.00000311, 0.00000049, 0.00000218, 0.00001330, -0.00001131, 0.00000204, 0.00000495, -0.00001264, 0.00000689, -0.00001734, 0.00001374, -0.00001333, 0.00000080, -0.00001064, -0.00000778, 0.00000491, -0.00001313, -0.00000190, -0.00000803, -0.00000388, 0.00000886, -0.00003904, -0.00000346, -0.00013733, 0.00009479, 0.00017903, 0.00016313, -0.00009048, -0.00006509, -0.00000513, -0.00002075, -0.00000481, -0.00000873, -0.00000890, -0.00000061, -0.00003206, -0.00000655, 0.00000634, 0.00003979, 0.00000796, 0.00003480, 0.00002446, -0.00000741, -0.00001947, -0.00000455, -0.00000039, -0.00001940, 0.00000778, -0.00000328, -0.00000621, },
		{ 0.00001317, -0.00001279, 0.00000459, -0.00000944, 0.00000941, -0.00000118, -0.00001185, -0.00000953, 0.00001562, -0.00000414, -0.00000344, 0.00000540, 0.00000386, -0.00000472, -0.00000138, 0.00000039, 0.00000367, 0.00000447, 0.00000009, -0.00000140, -0.00000534, 0.00000005, 0.00000271, -0.00000611, 0.00000222, 0.00000597, -0.00000574, 0.00000194, 0.00000456, -0.00000142, 0.00000526, -0.00000146, -0.00001163, -0.00000342, 0.00000906, 0.00000082, -0.00000006, -0.00000237, -0.00000263, -0.00000069, 0.00000895, -0.00000241, -0.00000519, -0.00000382, 0.00000668, 0.00000148, 0.00000009, -0.00000369, -0.00000183, 0.00000012, -0.00000076, -0.00000001, 0.00000512, -0.00000290, -0.00000181, 0.00000425, -0.00000486, 0.00000251, -0.00000256, 0.00000577, -0.00001369, 0.00000364, -0.00000265, 0.00001103, -0.00000568, 0.00000600, 0.00000253, -0.00000696, -0.00000274, 0.00000029, 0.00000588, -0.00000214, -0.00000480, -0.00000473, 0.00000953, -0.00000015, 0.00000220, -0.00000236, 0.00000006, 0.00000134, 0.00001243, -0.00000682, -0.00000396, 0.00001153, -0.00001830, 0.00001132, -0.00001990, 0.00001423, -0.00001292, 0.00000083, -0.00001035, -0.00000798, 0.00000335, -0.00000969, -0.00000631, -0.00000241, -0.00001166, 0.00001836, -0.00004696, 0.00000949, -0.00014000, 0.00005694, 0.00019299, 0.00017931, -0.00007438, -0.00008151, -0.00000019, -0.00002504, -0.00000245, -0.00001086, -0.00000748, -0.00000119, -0.00002891, -0.00000956, 0.00000555, 0.00003514, 0.00001245, 0.00002629, 0.00003105, -0.00000844, -0.00001247, -0.00001313, 0.00000670, -0.00002444, 0.00001040, -0.00000529, -0.00000269, -0.00001656, },
		{ -0.00001188, 0.00000418, -0.00000926, 0.00000763, 0.00000155, -0.00001116, -0.00001252, 0.00001529, -0.00000124, -0.00000506, 0.00000499, 0.00000383, -0.00000350, -0.00000231, 0.00000032, 0.00000333, 0.00000479, 0.00000040, -0.00000131, -0.00000455, -0.00000175, 0.00000373, -0.00000590, 0.00000133, 0.00000653, -0.00000488, -0.00000041, 0.00000635, -0.00000248, 0.00000565, -0.00000036, -0.00001069, -0.00000637, 0.00001000, 0.00000089, 0.00000054, -0.00000261, -0.00000210, -0.00000217, 0.00000900, -0.00000086, -0.00000538, -0.00000449, 0.00000581, 0.00000301, -0.00000087, -0.00000248, -0.00000285, 0.00000049, -0.00000095, 0.00000007, 0.00000404, -0.00000077, -0.00000357, 0.00000551, -0.00000581, 0.00000325, -0.00000373, 0.00000713, -0.00001314, 0.00000195, -0.00000198, 0.00000986, -0.00000335, 0.00000276, 0.00000526, -0.00000748, -0.00000217, -0.00000145, 0.00000652, -0.00000125, -0.00000456, -0.00000621, 0.00000935, 0.00000085, 0.00000152, -0.00000130, -0.00000073, 0.00000139, 0.00001009, -0.00000111, -0.00001046, 0.00001776, -0.00002273, 0.00001386, -0.00002039, 0.00001297, -0.00001118, -0.00000026, -0.00000907, -0.00000887, 0.00000235, -0.00000700, -0.00000968, 0.00000189, -0.00001755, 0.00002487, -0.00005064, 0.00001548, -0.00013175, 0.00001322, 0.00020294, 0.00019748, -0.00005614, -0.00009775, 0.00000362, -0.00002868, -0.00000048, -0.00001284, -0.00000605, -0.00000215, -0.00002526, -0.00001305, 0.00000596, 0.00002912, 0.00001815, 0.00001723, 0.00003746, -0.00001031, -0.00000463, -0.00002106, 0.00001225, -0.00002768, 0.00001138, -0.00000601, -0.00000033, -0.00001903, 0.00001367, },
		{ 0.00000270, -0.00000794, 0.00000487, 0.00000454, -0.00001038, -0.00001480, 0.00001365, 0.00000242, -0.00000677, 0.00000477, 0.00000344, -0.00000217, -0.00000303, 0.00000004, 0.00000313, 0.00000488, 0.00000095, -0.00000153, -0.00000344, -0.00000355, 0.00000445, -0.00000548, 0.00000051, 0.00000673, -0.00000349, -0.00000299, 0.00000793, -0.00000316, 0.00000560, 0.00000078, -0.00000945, -0.00000915, 0.00001030, 0.00000144, 0.00000076, -0.00000247, -0.00000189, -0.00000307, 0.00000827, 0.00000111, -0.00000569, -0.00000479, 0.00000440, 0.00000480, -0.00000185, -0.00000141, -0.00000355, 0.00000057, -0.00000102, 0.00000021, 0.00000290, 0.00000109, -0.00000479, 0.00000616, -0.00000613, 0.00000334, -0.00000415, 0.00000742, -0.00001142, -0.00000054, -0.00000067, 0.00000799, -0.00000047, -0.00000058, 0.00000757, -0.00000763, -0.00000167, -0.00000293, 0.00000663, -0.00000012, -0.00000424, -0.00000749, 0.00000868, 0.00000224, 0.00000053, 0.00000006, -0.00000189, 0.00000226, 0.00000663, 0.00000515, -0.00001655, 0.00002268, -0.00002505, 0.00001388, -0.00001847, 0.00000982, -0.00000803, -0.00000257, -0.00000669, -0.00001061, 0.00000217, -0.00000541, -0.00001159, 0.00000437, -0.00002098, 0.00002777, -0.00004953, 0.00001448, -0.00011435, -0.00003254, 0.00020559, 0.00021882, -0.00003643, -0.00011291, 0.00000558, -0.00003117, 0.00000068, -0.00001432, -0.00000495, -0.00000319, -0.00002142, -0.00001671, 0.00000717, 0.00002231, 0.00002430, 0.00000843, 0.00004300, -0.00001218, 0.00000280, -0.00002707, 0.00001529, -0.00002844, 0.00001027, -0.00000512, 0.00000057, -0.00001980, 0.00001243, -0.00000960, },
		{ -0.00000570, 0.00000140, 0.00000753, -0.00000946, -0.00001633, 0.00001084, 0.00000649, -0.00000827, 0.00000451, 0.00000292, -0.00000101, -0.00000333, -0.00000056, 0.00000313, 0.00000473, 0.00000172, -0.00000200, -0.00000215, -0.00000516, 0.00000477, -0.00000485, -0.00000026, 0.00000662, -0.00000174, -0.00000552, 0.00000902, -0.00000326, 0.00000504, 0.00000202, -0.00000807, -0.00001150, 0.00000981, 0.00000256, 0.00000056, -0.00000192, -0.00000206, -0.00000337, 0.00000685, 0.00000332, -0.00000598, -0.00000482, 0.00000265, 0.00000658, -0.00000257, -0.00000074, -0.00000376, 0.00000026, -0.00000090, 0.00000031, 0.00000188, 0.00000247, -0.00000530, 0.00000613, -0.00000577, 0.00000275, -0.00000380, 0.00000669, -0.00000875, -0.00000353, 0.00000096, 0.00000578, 0.00000255, -0.00000359, 0.00000914, -0.00000723, -0.00000137, -0.00000399, 0.00000614, 0.00000124, -0.00000388, -0.00000848, 0.00000750, 0.00000398, -0.00000071, 0.00000163, -0.00000333, 0.00000373, 0.00000256, 0.00001117, -0.00002136, 0.00002547, -0.00002465, 0.00001110, -0.00001419, 0.00000495, -0.00000365, -0.00000594, -0.00000336, -0.00001311, 0.00000280, -0.00000500, -0.00001196, 0.00000496, -0.00002186, 0.00002709, -0.00004398, 0.00000757, -0.00009069, -0.00007592, 0.00019794, 0.00024395, -0.00001550, -0.00012631, 0.00000519, -0.00003219, 0.00000073, -0.00001499, -0.00000443, -0.00000403, -0.00001771, -0.00002019, 0.00000871, 0.00001535, 0.00003012, 0.00000071, 0.00004696, -0.00001318, 0.00000867, -0.00003015, 0.00001520, -0.00002647, 0.00000699, -0.00000260, -0.00000003, -0.00001890, 0.00000957, -0.00000615, 0.00000034, },
		{ -0.00000233, 0.00001016, -0.00000827, -0.00001718, 0.00000717, 0.00001051, -0.00000919, 0.00000395, 0.00000253, -0.00000025, -0.00000310, -0.00000150, 0.00000333, 0.00000437, 0.00000266, -0.00000260, -0.00000085, -0.00000644, 0.00000463, -0.00000399, -0.00000097, 0.00000629, 0.00000016, -0.00000771, 0.00000937, -0.00000264, 0.00000394, 0.00000338, -0.00000669, -0.00001323, 0.00000843, 0.00000420, -0.00000002, -0.00000102, -0.00000254, -0.00000317, 0.00000495, 0.00000553, -0.00000608, -0.00000472, 0.00000078, 0.00000808, -0.00000278, -0.00000063, -0.00000342, -0.00000043, -0.00000060, 0.00000031, 0.00000113, 0.00000324, -0.00000508, 0.00000546, -0.00000480, 0.00000156, -0.00000281, 0.00000517, -0.00000548, -0.00000663, 0.00000255, 0.00000361, 0.00000528, -0.00000587, 0.00000974, -0.00000620, -0.00000137, -0.00000453, 0.00000505, 0.00000276, -0.00000349, -0.00000913, 0.00000580, 0.00000598, -0.00000207, 0.00000328, -0.00000486, 0.00000548, -0.00000154, 0.00001619, -0.00002415, 0.00002557, -0.00002128, 0.00000566, -0.00000794, -0.00000111, 0.00000151, -0.00000998, 0.00000055, -0.00001602, 0.00000397, -0.00000550, -0.00001110, 0.00000402, -0.00002063, 0.00002350, -0.00003508, -0.00000328, -0.00006441, -0.00011251, 0.00017778, 0.00027254, 0.00000694, -0.00013767, 0.00000227, -0.00003167, -0.00000041, -0.00001476, -0.00000464, -0.00000449, -0.00001437, -0.00002320, 0.00001016, 0.00000885, 0.00003490, -0.00000523, 0.00004873, -0.00001259, 0.00001213, -0.00002973, 0.00001189, -0.00002194, 0.00000185, 0.00000124, -0.00000184, -0.00001662, 0.00000550, -0.00000198, -0.00000251, -0.00000292, },
		{ 0.00001203, -0.00000662, -0.00001755, 0.00000306, 0.00001399, -0.00000919, 0.00000290, 0.00000250, -0.00000007, -0.00000228, -0.00000274, 0.00000364, 0.00000387, 0.00000366, -0.00000320, 0.00000030, -0.00000722, 0.00000397, -0.00000292, -0.00000165, 0.00000581, 0.00000202, -0.00000932, 0.00000885, -0.00000129, 0.00000236, 0.00000477, -0.00000536, -0.00001425, 0.00000623, 0.00000623, -0.00000083, 0.00000009, -0.00000319, -0.00000265, 0.00000284, 0.00000744, -0.00000582, -0.00000464, -0.00000099, 0.00000904, -0.00000230, -0.00000117, -0.00000252, -0.00000142, -0.00000018, 0.00000019, 0.00000069, 0.00000340, -0.00000421, 0.00000428, -0.00000339, -0.00000002, -0.00000143, 0.00000320, -0.00000206, -0.00000939, 0.00000369, 0.00000186, 0.00000736, -0.00000714, 0.00000924, -0.00000455, -0.00000169, -0.00000451, 0.00000341, 0.00000434, -0.00000305, -0.00000941, 0.00000365, 0.00000810, -0.00000340, 0.00000484, -0.00000625, 0.00000714, -0.00000510, 0.00001956, -0.00002445, 0.00002282, -0.00001515, -0.00000188, -0.00000053, -0.00000759, 0.00000674, -0.00001404, 0.00000442, -0.00001879, 0.00000516, -0.00000638, -0.00000961, 0.00000223, -0.00001811, 0.00001812, -0.00002446, -0.00001559, -0.00003929, -0.00013857, 0.00014424, 0.00030310, 0.00003187, -0.00014710, -0.00000301, -0.00002983, -0.00000258, -0.00001371, -0.00000552, -0.00000455, -0.00001152, -0.00002556, 0.00001119, 0.00000328, 0.00003810, -0.00000890, 0.00004796, -0.00000999, 0.00001276, -0.00002580, 0.00000578, -0.00001550, -0.00000447, 0.00000578, -0.00000432, -0.00001352, 0.00000082, 0.00000234, -0.00000533, -0.00000008, -0.00000581, },
		{ -0.00000436, -0.00001765, -0.00000106, 0.00001646, -0.00000803, 0.00000124, 0.00000292, -0.00000050, -0.00000095, -0.00000413, 0.00000394, 0.00000336, 0.00000459, -0.00000364, 0.00000112, -0.00000743, 0.00000281, -0.00000168, -0.00000229, 0.00000528, 0.00000367, -0.00001018, 0.00000742, 0.00000069, 0.00000047, 0.00000610, -0.00000408, -0.00001458, 0.00000337, 0.00000839, -0.00000164, 0.00000120, -0.00000382, -0.00000204, 0.00000081, 0.00000881, -0.00000504, -0.00000472, -0.00000244, 0.00000927, -0.00000105, -0.00000232, -0.00000120, -0.00000256, 0.00000025, 0.00000000, 0.00000054, 0.00000304, -0.00000289, 0.00000286, -0.00000181, -0.00000170, 0.00000004, 0.00000116, 0.00000107, -0.00001138, 0.00000405, 0.00000083, 0.00000851, -0.00000724, 0.00000768, -0.00000240, -0.00000227, -0.00000398, 0.00000138, 0.00000585, -0.00000253, -0.00000935, 0.00000118, 0.00001017, -0.00000452, 0.00000610, -0.00000730, 0.00000839, -0.00000764, 0.00002089, -0.00002213, 0.00001744, -0.00000689, -0.00001058, 0.00000702, -0.00001353, 0.00001124, -0.00001741, 0.00000756, -0.00002075, 0.00000577, -0.00000697, -0.00000827, 0.00000048, -0.00001537, 0.00001232, -0.00001398, -0.00002678, -0.00001871, -0.00015156, 0.00009808, 0.00033293, 0.00006088, -0.00015510, -0.00001015, -0.00002718, -0.00000541, -0.00001214, -0.00000685, -0.00000434, -0.00000911, -0.00002721, 0.00001161, -0.00000107, 0.00003944, -0.00001012, 0.00004464, -0.00000533, 0.00001061, -0.00001892, -0.00000219, -0.00000817, -0.00001103, 0.00001021, -0.00000675, -0.00001030, -0.00000376, 0.00000614, -0.00000758, 0.00000229, -0.00000855, 0.00001281, },
		{ -0.00001767, -0.00000478, 0.00001759, -0.00000561, -0.00000102, 0.00000377, -0.00000148, 0.00000071, -0.00000549, 0.00000409, 0.00000297, 0.00000533, -0.00000381, 0.00000151, -0.00000703, 0.00000123, -0.00000036, -0.00000287, 0.00000474, 0.00000498, -0.00001023, 0.00000518, 0.00000309, -0.00000148, 0.00000717, -0.00000280, -0.00001431, 0.00000009, 0.00001040, -0.00000221, 0.00000211, -0.00000424, -0.00000155, -0.00000087, 0.00000943, -0.00000368, -0.00000501, -0.00000347, 0.00000871, 0.00000091, -0.00000391, 0.00000031, -0.00000362, 0.00000054, -0.00000017, 0.00000060, 0.00000233, -0.00000139, 0.00000148, -0.00000034, -0.00000320, 0.00000128, -0.00000058, 0.00000351, -0.00001231, 0.00000344, 0.00000066, 0.00000864, -0.00000620, 0.00000526, 0.00000002, -0.00000296, -0.00000306, -0.00000087, 0.00000712, -0.00000187, -0.00000900, -0.00000149, 0.00001200, -0.00000525, 0.00000691, -0.00000780, 0.00000895, -0.00000888, 0.00002004, -0.00001743, 0.00001008, 0.00000253, -0.00001925, 0.00001355, -0.00001796, 0.00001423, -0.00001938, 0.00000930, -0.00002128, 0.00000516, -0.00000660, -0.00000785, -0.00000037, -0.00001339, 0.00000740, -0.00000534, -0.00003468, -0.00000515, -0.00015058, 0.00004187, 0.00035817, 0.00009592, -0.00016232, -0.00001846, -0.00002439, -0.00000836, -0.00001048, -0.00000827, -0.00000413, -0.00000699, -0.00002827, 0.00001143, -0.00000415, 0.00003891, -0.00000904, 0.00003907, 0.00000102, 0.00000626, -0.00001008, -0.00001075, -0.00000117, -0.00001677, 0.00001366, -0.00000840, -0.00000766, -0.00000755, 0.00000882, -0.00000877, 0.00000375, -0.00001014, 0.00001223, -0.00000139, },
		{ -0.00000784, 0.00001721, -0.00000205, -0.00000367, 0.00000491, -0.00000280, 0.00000244, -0.00000657, 0.00000390, 0.00000282, 0.00000576, -0.00000362, 0.00000143, -0.00000607, -0.00000066, 0.00000092, -0.00000335, 0.00000423, 0.00000592, -0.00000951, 0.00000232, 0.00000559, -0.00000320, 0.00000780, -0.00000142, -0.00001360, -0.00000329, 0.00001195, -0.00000231, 0.00000263, -0.00000426, -0.00000138, -0.00000201, 0.00000921, -0.00000176, -0.00000551, -0.00000402, 0.00000738, 0.00000339, -0.00000568, 0.00000174, -0.00000436, 0.00000052, -0.00000023, 0.00000074, 0.00000147, 0.00000002, 0.00000042, 0.00000076, -0.00000426, 0.00000201, -0.00000174, 0.00000500, -0.00001202, 0.00000183, 0.00000134, 0.00000783, -0.00000424, 0.00000230, 0.00000239, -0.00000357, -0.00000193, -0.00000310, 0.00000798, -0.00000101, -0.00000842, -0.00000417, 0.00001340, -0.00000545, 0.00000716, -0.00000765, 0.00000870, -0.00000870, 0.00001719, -0.00001093, 0.00000166, 0.00001192, -0.00002665, 0.00001797, -0.00002007, 0.00001509, -0.00001943, 0.00000915, -0.00001994, 0.00000293, -0.00000479, -0.00000889, 0.00000031, -0.00001294, 0.00000434, 0.00000026, -0.00003792, 0.00000025, -0.00013658, -0.00002015, 0.00037423, 0.00013899, -0.00016937, -0.00002725, -0.00002214, -0.00001089, -0.00000918, -0.00000939, -0.00000424, -0.00000492, -0.00002892, 0.00001077, -0.00000606, 0.00003673, -0.00000607, 0.00003189, 0.00000837, 0.00000062, -0.00000062, -0.00001849, 0.00000430, -0.00002076, 0.00001541, -0.00000867, -0.00000614, -0.00001004, 0.00000998, -0.00000860, 0.00000404, -0.00001035, 0.00001025, 0.00000224, -0.00001769, },
		{ 0.00001540, 0.00000235, -0.00000639, 0.00000606, -0.00000418, 0.00000394, -0.00000713, 0.00000324, 0.00000299, 0.00000584, -0.00000303, 0.00000089, -0.00000466, -0.00000267, 0.00000202, -0.00000365, 0.00000372, 0.00000649, -0.00000816, -0.00000084, 0.00000785, -0.00000440, 0.00000782, 0.00000014, -0.00001262, -0.00000647, 0.00001276, -0.00000178, 0.00000263, -0.00000380, -0.00000162, -0.00000252, 0.00000817, 0.00000059, -0.00000612, -0.00000414, 0.00000543, 0.00000613, -0.00000731, 0.00000276, -0.00000456, 0.00000009, -0.00000010, 0.00000086, 0.00000069, 0.00000106, -0.00000009, 0.00000132, -0.00000472, 0.00000211, -0.00000215, 0.00000543, -0.00001055, -0.00000063, 0.00000266, 0.00000635, -0.00000170, -0.00000078, 0.00000438, -0.00000391, -0.00000078, -0.00000508, 0.00000827, 0.00000010, -0.00000770, -0.00000670, 0.00001422, -0.00000504, 0.00000680, -0.00000683, 0.00000761, -0.00000722, 0.00001274, -0.00000341, -0.00000674, 0.00002008, -0.00003168, 0.00001946, -0.00001938, 0.00001354, -0.00001733, 0.00000692, -0.00001658, -0.00000106, -0.00000137, -0.00001162, 0.00000274, -0.00001429, 0.00000354, 0.00000234, -0.00003611, -0.00000218, -0.00011227, -0.00008245, 0.00037636, 0.00019160, -0.00017638, -0.00003608, -0.00002092, -0.00001260, -0.00000861, -0.00000987, -0.00000493, -0.00000270, -0.00002935, 0.00000981, -0.00000705, 0.00003329, -0.00000182, 0.00002387, 0.00001586, -0.00000520, 0.00000808, -0.00002409, 0.00000725, -0.00002233, 0.00001502, -0.00000723, -0.00000605, -0.00001094, 0.00000943, -0.00000700, 0.00000309, -0.00000918, 0.00000699, 0.00000633, -0.00001770, -0.00001013, },
		{ 0.00000710, -0.00000874, 0.00000689, -0.00000529, 0.00000489, -0.00000697, 0.00000203, 0.00000351, 0.00000555, -0.00000208, -0.00000001, -0.00000296, -0.00000459, 0.00000280, -0.00000372, 0.00000319, 0.00000676, -0.00000637, -0.00000397, 0.00000953, -0.00000483, 0.00000712, 0.00000190, -0.00001151, -0.00000922, 0.00001268, -0.00000056, 0.00000210, -0.00000284, -0.00000229, -0.00000242, 0.00000645, 0.00000316, -0.00000669, -0.00000398, 0.00000311, 0.00000878, -0.00000844, 0.00000311, -0.00000406, -0.00000082, 0.00000026, 0.00000084, 0.00000015, 0.00000158, 0.00000004, 0.00000127, -0.00000455, 0.00000154, -0.00000184, 0.00000489, -0.00000815, -0.00000360, 0.00000427, 0.00000456, 0.00000096, -0.00000349, 0.00000564, -0.00000380, 0.00000022, -0.00000660, 0.00000789, 0.00000145, -0.00000691, -0.00000893, 0.00001435, -0.00000397, 0.00000584, -0.00000539, 0.00000580, -0.00000470, 0.00000731, 0.00000420, -0.00001405, 0.00002598, -0.00003355, 0.00001764, -0.00001582, 0.00000970, -0.00001326, 0.00000280, -0.00001144, -0.00000656, 0.00000343, -0.00001580, 0.00000666, -0.00001718, 0.00000476, 0.00000125, -0.00003000, -0.00001073, -0.00008178, -0.00013887, 0.00036030, 0.00025422, -0.00018274, -0.00004499, -0.00002089, -0.00001338, -0.00000890, -0.00000959, -0.00000632, -0.00000022, -0.00002971, 0.00000871, -0.00000737, 0.00002904, 0.00000306, 0.00001587, 0.00002257, -0.00001013, 0.00001482, -0.00002662, 0.00000717, -0.00002126, 0.00001242, -0.00000411, -0.00000738, -0.00001031, 0.00000734, -0.00000419, 0.00000113, -0.00000688, 0.00000282, 0.00001050, -0.00001753, -0.00001171, 0.00001242, },
		{ -0.00001025, 0.00000707, -0.00000584, 0.00000506, -0.00000597, 0.00000026, 0.00000435, 0.00000496, -0.00000085, -0.00000115, -0.00000118, -0.00000623, 0.00000314, -0.00000350, 0.00000260, 0.00000682, -0.00000438, -0.00000674, 0.00001036, -0.00000437, 0.00000570, 0.00000385, -0.00001036, -0.00001136, 0.00001165, 0.00000128, 0.00000114, -0.00000149, -0.00000328, -0.00000185, 0.00000429, 0.00000565, -0.00000701, -0.00000370, 0.00000069, 0.00001099, -0.00000879, 0.00000261, -0.00000281, -0.00000216, 0.00000080, 0.00000068, -0.00000007, 0.00000150, 0.00000081, 0.00000071, -0.00000386, 0.00000045, -0.00000098, 0.00000365, -0.00000520, -0.00000664, 0.00000575, 0.00000291, 0.00000327, -0.00000542, 0.00000592, -0.00000312, 0.00000093, -0.00000749, 0.00000678, 0.00000304, -0.00000610, -0.00001075, 0.00001372, -0.00000230, 0.00000438, -0.00000346, 0.00000347, -0.00000153, 0.00000156, 0.00001100, -0.00001939, 0.00002894, -0.00003194, 0.00001263, -0.00000985, 0.00000412, -0.00000777, -0.00000263, -0.00000515, -0.00001291, 0.00000898, -0.00002077, 0.00001134, -0.00002083, 0.00000714, -0.00000190, -0.00002121, -0.00002261, -0.00005006, -0.00018340, 0.00032309, 0.00032585, -0.00018684, -0.00005463, -0.00002172, -0.00001349, -0.00000986, -0.00000867, -0.00000828, 0.00000244, -0.00002998, 0.00000752, -0.00000720, 0.00002436, 0.00000798, 0.00000859, 0.00002776, -0.00001333, 0.00001877, -0.00002564, 0.00000406, -0.00001783, 0.00000803, 0.00000030, -0.00000974, -0.00000853, 0.00000414, -0.00000063, -0.00000141, -0.00000389, -0.00000171, 0.00001429, -0.00001691, -0.00001279, 0.00000872, 0.00001163, },
		{ 0.00000635, -0.00000560, 0.00000429, -0.00000413, -0.00000198, 0.00000540, 0.00000417, 0.00000052, -0.00000236, 0.00000048, -0.00000742, 0.00000297, -0.00000300, 0.00000194, 0.00000676, -0.00000240, -0.00000889, 0.00001019, -0.00000301, 0.00000369, 0.00000585, -0.00000918, -0.00001284, 0.00000974, 0.00000356, -0.00000008, 0.00000007, -0.00000440, -0.00000103, 0.00000201, 0.00000778, -0.00000689, -0.00000347, -0.00000155, 0.00001246, -0.00000817, 0.00000120, -0.00000089, -0.00000380, 0.00000143, 0.00000039, 0.00000003, 0.00000088, 0.00000207, -0.00000018, -0.00000286, -0.00000091, 0.00000017, 0.00000207, -0.00000217, -0.00000926, 0.00000666, 0.00000181, 0.00000483, -0.00000626, 0.00000511, -0.00000187, 0.00000128, -0.00000767, 0.00000497, 0.00000480, -0.00000529, -0.00001206, 0.00001235, -0.00000010, 0.00000254, -0.00000122, 0.00000087, 0.00000190, -0.00000386, 0.00001629, -0.00002220, 0.00002871, -0.00002707, 0.00000506, -0.00000233, -0.00000227, -0.00000174, -0.00000851, 0.00000139, -0.00001919, 0.00001433, -0.00002551, 0.00001566, -0.00002403, 0.00000936, -0.00000548, -0.00001202, -0.00003446, -0.00002218, -0.00021104, 0.00026374, 0.00040362, -0.00018591, -0.00006632, -0.00002257, -0.00001360, -0.00001097, -0.00000751, -0.00001046, 0.00000495, -0.00002995, 0.00000613, -0.00000659, 0.00001951, 0.00001251, 0.00000254, 0.00003099, -0.00001436, 0.00001966, -0.00002134, -0.00000145, -0.00001281, 0.00000261, 0.00000527, -0.00001249, -0.00000623, 0.00000050, 0.00000301, -0.00000395, -0.00000083, -0.00000597, 0.00001715, -0.00001553, -0.00001368, 0.00000481, 0.00001531, -0.00001049, },
		{ -0.00000448, 0.00000258, -0.00000157, -0.00000450, 0.00000649, 0.00000332, 0.00000190, -0.00000347, 0.00000184, -0.00000804, 0.00000227, -0.00000224, 0.00000123, 0.00000665, -0.00000059, -0.00001026, 0.00000899, -0.00000091, 0.00000131, 0.00000771, -0.00000793, -0.00001370, 0.00000715, 0.00000600, -0.00000128, 0.00000160, -0.00000543, -0.00000022, -0.00000010, 0.00000925, -0.00000618, -0.00000343, -0.00000340, 0.00001300, -0.00000651, -0.00000101, 0.00000149, -0.00000548, 0.00000196, 0.00000008, 0.00000038, -0.00000013, 0.00000354, -0.00000108, -0.00000186, -0.00000224, 0.00000123, 0.00000057, 0.00000045, -0.00001101, 0.00000663, 0.00000159, 0.00000535, -0.00000584, 0.00000323, -0.00000015, 0.00000129, -0.00000714, 0.00000256, 0.00000660, -0.00000449, -0.00001283, 0.00001027, 0.00000249, 0.00000050, 0.00000115, -0.00000177, 0.00000518, -0.00000843, 0.00001958, -0.00002228, 0.00002553, -0.00001964, -0.00000398, 0.00000549, -0.00000831, 0.00000376, -0.00001377, 0.00000710, -0.00002433, 0.00001840, -0.00002888, 0.00001838, -0.00002545, 0.00000996, -0.00000765, -0.00000483, -0.00004296, -0.00000250, -0.00021858, 0.00018369, 0.00048267, -0.00017619, -0.00008187, -0.00002225, -0.00001471, -0.00001145, -0.00000680, -0.00001227, 0.00000683, -0.00002925, 0.00000429, -0.00000543, 0.00001454, 0.00001648, -0.00000211, 0.00003218, -0.00001323, 0.00001776, -0.00001449, -0.00000828, -0.00000735, -0.00000279, 0.00000987, -0.00001482, -0.00000418, -0.00000280, 0.00000602, -0.00000583, 0.00000169, -0.00000933, 0.00001858, -0.00001310, -0.00001470, 0.00000124, 0.00001762, -0.00000918, 0.00000460, },
		{ 0.00000006, 0.00000148, -0.00000706, 0.00000743, 0.00000258, 0.00000313, -0.00000435, 0.00000279, -0.00000805, 0.00000111, -0.00000131, 0.00000050, 0.00000651, 0.00000094, -0.00001080, 0.00000691, 0.00000167, -0.00000112, 0.00000920, -0.00000649, -0.00001406, 0.00000417, 0.00000828, -0.00000220, 0.00000285, -0.00000614, 0.00000033, -0.00000174, 0.00000988, -0.00000481, -0.00000367, -0.00000470, 0.00001252, -0.00000390, -0.00000380, 0.00000401, -0.00000692, 0.00000220, -0.00000013, 0.00000084, -0.00000129, 0.00000493, -0.00000170, -0.00000116, -0.00000322, 0.00000189, -0.00000047, 0.00000226, -0.00001156, 0.00000544, 0.00000240, 0.00000474, -0.00000423, 0.00000049, 0.00000183, 0.00000105, -0.00000599, -0.00000027, 0.00000830, -0.00000368, -0.00001306, 0.00000759, 0.00000529, -0.00000157, 0.00000347, -0.00000422, 0.00000803, -0.00001180, 0.00002072, -0.00001989, 0.00002010, -0.00001076, -0.00001312, 0.00001227, -0.00001282, 0.00000769, -0.00001742, 0.00001097, -0.00002730, 0.00002017, -0.00002977, 0.00001832, -0.00002387, 0.00000759, -0.00000681, -0.00000171, -0.00004545, 0.00000597, -0.00020516, 0.00008707, 0.00055642, -0.00015322, -0.00010333, -0.00001940, -0.00001794, -0.00001039, -0.00000731, -0.00001299, 0.00000747, -0.00002741, 0.00000161, -0.00000350, 0.00000936, 0.00001994, -0.00000548, 0.00003162, -0.00001040, 0.00001386, -0.00000634, -0.00001501, -0.00000275, -0.00000705, 0.00001319, -0.00001591, -0.00000313, -0.00000505, 0.00000773, -0.00000651, 0.00000312, -0.00001125, 0.00001823, -0.00000944, -0.00001608, -0.00000156, 0.00001821, -0.00000624, 0.00000186, -0.00000256, },
		{ 0.00000470, -0.00000937, 0.00000802, 0.00000207, 0.00000411, -0.00000490, 0.00000328, -0.00000749, -0.00000038, -0.00000036, -0.00000015, 0.00000632, 0.00000217, -0.00001058, 0.00000420, 0.00000435, -0.00000323, 0.00001005, -0.00000475, -0.00001411, 0.00000112, 0.00001005, -0.00000258, 0.00000362, -0.00000633, 0.00000044, -0.00000272, 0.00000958, -0.00000283, -0.00000415, -0.00000543, 0.00001109, -0.00000059, -0.00000680, 0.00000630, -0.00000781, 0.00000195, -0.00000011, 0.00000123, -0.00000235, 0.00000592, -0.00000172, -0.00000102, -0.00000359, 0.00000189, -0.00000077, 0.00000301, -0.00001078, 0.00000308, 0.00000419, 0.00000314, -0.00000165, -0.00000275, 0.00000376, 0.00000072, -0.00000436, -0.00000330, 0.00000973, -0.00000282, -0.00001279, 0.00000445, 0.00000814, -0.00000353, 0.00000558, -0.00000631, 0.00001024, -0.00001384, 0.00001986, -0.00001559, 0.00001340, -0.00000176, -0.00002095, 0.00001675, -0.00001479, 0.00000920, -0.00001865, 0.00001221, -0.00002736, 0.00001889, -0.00002744, 0.00001470, -0.00001851, 0.00000144, -0.00000192, -0.00000394, -0.00004046, 0.00000214, -0.00017256, -0.00001952, 0.00061697, -0.00011248, -0.00013247, -0.00001287, -0.00002426, -0.00000698, -0.00000975, -0.00001197, 0.00000628, -0.00002397, -0.00000226, -0.00000051, 0.00000377, 0.00002311, -0.00000790, 0.00002988, -0.00000664, 0.00000906, 0.00000165, -0.00002014, -0.00000022, -0.00000923, 0.00001444, -0.00001515, -0.00000364, -0.00000570, 0.00000772, -0.00000562, 0.00000312, -0.00001143, 0.00001595, -0.00000456, -0.00001793, -0.00000333, 0.00001695, -0.00000181, -0.00000160, -0.00000006, -0.00000299, },
		{ -0.00001112, 0.00000807, 0.00000191, 0.00000477, -0.00000508, 0.00000330, -0.00000647, -0.00000201, 0.00000045, -0.00000063, 0.00000605, 0.00000315, -0.00000976, 0.00000121, 0.00000672, -0.00000466, 0.00001005, -0.00000261, -0.00001399, -0.00000173, 0.00001107, -0.00000226, 0.00000378, -0.00000591, 0.00000001, -0.00000299, 0.00000839, -0.00000037, -0.00000478, -0.00000564, 0.00000889, 0.00000307, -0.00000960, 0.00000796, -0.00000788, 0.00000108, 0.00000025, 0.00000141, -0.00000309, 0.00000625, -0.00000096, -0.00000157, -0.00000324, 0.00000111, -0.00000024, 0.00000263, -0.00000874, -0.00000025, 0.00000670, 0.00000084, 0.00000146, -0.00000601, 0.00000529, 0.00000049, -0.00000245, -0.00000627, 0.00001072, -0.00000187, -0.00001208, 0.00000103, 0.00001083, -0.00000522, 0.00000737, -0.00000796, 0.00001173, -0.00001458, 0.00001738, -0.00001020, 0.00000659, 0.00000606, -0.00002625, 0.00001799, -0.00001360, 0.00000786, -0.00001710, 0.00001047, -0.00002421, 0.00001432, -0.00002161, 0.00000730, -0.00000923, -0.00000862, 0.00000718, -0.00001170, -0.00002798, -0.00001301, -0.00012510, -0.00012750, 0.00065603, -0.00005010, -0.00017027, -0.00000208, -0.00003421, -0.00000078, -0.00001455, -0.00000879, 0.00000290, -0.00001862, -0.00000760, 0.00000373, -0.00000241, 0.00002625, -0.00000985, 0.00002767, -0.00000288, 0.00000460, 0.00000806, -0.00002244, -0.00000066, -0.00000872, 0.00001324, -0.00001223, -0.00000597, -0.00000454, 0.00000584, -0.00000312, 0.00000163, -0.00000984, 0.00001190, 0.00000131, -0.00002018, -0.00000403, 0.00001398, 0.00000369, -0.00000537, 0.00000266, -0.00000620, 0.00000770, },
		{ 0.00000747, 0.00000214, 0.00000508, -0.00000491, 0.00000294, -0.00000515, -0.00000359, 0.00000096, -0.00000085, 0.00000564, 0.00000395, -0.00000859, -0.00000169, 0.00000841, -0.00000515, 0.00000911, -0.00000006, -0.00001382, -0.00000413, 0.00001117, -0.00000119, 0.00000333, -0.00000490, -0.00000094, -0.00000259, 0.00000650, 0.00000232, -0.00000538, -0.00000549, 0.00000620, 0.00000669, -0.00001178, 0.00000867, -0.00000695, -0.00000045, 0.00000095, 0.00000131, -0.00000334, 0.00000579, 0.00000064, -0.00000282, -0.00000220, -0.00000039, 0.00000105, 0.00000130, -0.00000573, -0.00000416, 0.00000950, -0.00000168, 0.00000458, -0.00000880, 0.00000608, 0.00000054, -0.00000047, -0.00000891, 0.00001110, -0.00000077, -0.00001103, -0.00000249, 0.00001320, -0.00000653, 0.00000878, -0.00000911, 0.00001251, -0.00001422, 0.00001383, -0.00000462, 0.00000074, 0.00001162, -0.00002818, 0.00001557, -0.00000918, 0.00000377, -0.00001290, 0.00000594, -0.00001811, 0.00000676, -0.00001266, -0.00000340, 0.00000333, -0.00002182, 0.00001968, -0.00002398, -0.00000959, -0.00003652, -0.00006915, -0.00022723, 0.00066585, 0.00003628, -0.00021633, 0.00001254, -0.00004758, 0.00000807, -0.00002166, -0.00000342, -0.00000269, -0.00001138, -0.00001439, 0.00000921, -0.00000922, 0.00002955, -0.00001173, 0.00002565, 0.00000004, 0.00000151, 0.00001177, -0.00002112, -0.00000441, -0.00000548, 0.00000966, -0.00000730, -0.00000998, -0.00000173, 0.00000233, 0.00000074, -0.00000111, -0.00000677, 0.00000648, 0.00000770, -0.00002252, -0.00000386, 0.00000971, 0.00000966, -0.00000888, 0.00000516, -0.00000916, 0.00001010, -0.00001206, },
		{ 0.00000275, 0.00000507, -0.00000444, 0.00000231, -0.00000372, -0.00000491, 0.00000105, -0.00000074, 0.00000505, 0.00000466, -0.00000727, -0.00000416, 0.00000914, -0.00000457, 0.00000724, 0.00000281, -0.00001364, -0.00000597, 0.00001033, 0.00000053, 0.00000238, -0.00000342, -0.00000228, -0.00000170, 0.00000417, 0.00000492, -0.00000574, -0.00000514, 0.00000332, 0.00000986, -0.00001300, 0.00000823, -0.00000500, -0.00000256, 0.00000193, 0.00000092, -0.00000307, 0.00000454, 0.00000297, -0.00000460, -0.00000065, -0.00000239, 0.00000284, -0.00000066, -0.00000222, -0.00000811, 0.00001207, -0.00000389, 0.00000717, -0.00001062, 0.00000585, 0.00000100, 0.00000139, -0.00001098, 0.00001075, 0.00000049, -0.00000973, -0.00000591, 0.00001506, -0.00000735, 0.00000974, -0.00000975, 0.00001264, -0.00001302, 0.00000979, 0.00000033, -0.00000330, 0.00001428, -0.00002648, 0.00000971, -0.00000207, -0.00000242, -0.00000672, -0.00000067, -0.00000986, -0.00000288, -0.00000161, -0.00001622, 0.00001780, -0.00003660, 0.00003381, -0.00003865, 0.00001183, -0.00006387, -0.00001234, -0.00030911, 0.00064040, 0.00014673, -0.00026842, 0.00002934, -0.00006322, 0.00001869, -0.00003041, 0.00000361, -0.00001005, -0.00000267, -0.00002228, 0.00001565, -0.00001648, 0.00003299, -0.00001374, 0.00002425, 0.00000154, 0.00000049, 0.00001222, -0.00001606, -0.00001116, -0.00000004, 0.00000429, -0.00000094, -0.00001510, 0.00000218, -0.00000225, 0.00000537, -0.00000456, -0.00000278, 0.00000036, 0.00001399, -0.00002454, -0.00000320, 0.00000473, 0.00001537, -0.00001154, 0.00000700, -0.00001146, 0.00001158, -0.00001201, 0.00000617, },
		{ 0.00000482, -0.00000376, 0.00000153, -0.00000233, -0.00000583, 0.00000066, -0.00000030, 0.00000428, 0.00000535, -0.00000601, -0.00000596, 0.00000877, -0.00000295, 0.00000461, 0.00000582, -0.00001339, -0.00000725, 0.00000867, 0.00000270, 0.00000116, -0.00000168, -0.00000378, -0.00000056, 0.00000173, 0.00000712, -0.00000566, -0.00000480, 0.00000056, 0.00001223, -0.00001302, 0.00000659, -0.00000217, -0.00000504, 0.00000302, 0.00000034, -0.00000235, 0.00000265, 0.00000577, -0.00000659, 0.00000108, -0.00000453, 0.00000475, -0.00000277, 0.00000122, -0.00001152, 0.00001386, -0.00000527, 0.00000873, -0.00001111, 0.00000445, 0.00000192, 0.00000297, -0.00001230, 0.00000962, 0.00000193, -0.00000830, -0.00000903, 0.00001627, -0.00000762, 0.00001021, -0.00000986, 0.00001218, -0.00001122, 0.00000577, 0.00000406, -0.00000508, 0.00001392, -0.00002154, 0.00000124, 0.00000662, -0.00000954, 0.00000028, -0.00000817, -0.00000075, -0.00001320, 0.00001004, -0.00002945, 0.00003221, -0.00005078, 0.00004715, -0.00005280, 0.00003251, -0.00008972, 0.00003743, -0.00036480, 0.00057638, 0.00027836, -0.00032223, 0.00004531, -0.00007904, 0.00002943, -0.00003950, 0.00001125, -0.00001822, 0.00000665, -0.00003052, 0.00002239, -0.00002366, 0.00003627, -0.00001583, 0.00002358, 0.00000145, 0.00000168, 0.00000951, -0.00000786, -0.00001992, 0.00000653, -0.00000184, 0.00000589, -0.00002044, 0.00000635, -0.00000703, 0.00000996, -0.00000799, 0.00000140, -0.00000564, 0.00001943, -0.00002569, -0.00000255, -0.00000023, 0.00002005, -0.00001280, 0.00000780, -0.00001278, 0.00001190, -0.00001089, 0.00000420, 0.00000368, },
		{ -0.00000295, 0.00000071, -0.00000112, -0.00000626, -0.00000020, 0.00000039, 0.00000337, 0.00000604, -0.00000494, -0.00000697, 0.00000733, -0.00000050, 0.00000155, 0.00000867, -0.00001293, -0.00000806, 0.00000642, 0.00000499, -0.00000004, 0.00000004, -0.00000518, 0.00000053, -0.00000048, 0.00000864, -0.00000499, -0.00000461, -0.00000183, 0.00001358, -0.00001179, 0.00000388, 0.00000128, -0.00000757, 0.00000398, -0.00000028, -0.00000134, 0.00000037, 0.00000868, -0.00000838, 0.00000257, -0.00000641, 0.00000631, -0.00000453, 0.00000402, -0.00001382, 0.00001439, -0.00000541, 0.00000891, -0.00001008, 0.00000189, 0.00000326, 0.00000420, -0.00001275, 0.00000769, 0.00000353, -0.00000686, -0.00001166, 0.00001668, -0.00000725, 0.00001014, -0.00000943, 0.00001119, -0.00000903, 0.00000211, 0.00000627, -0.00000462, 0.00001099, -0.00001430, -0.00000848, 0.00001537, -0.00001613, 0.00000670, -0.00001510, 0.00000766, -0.00002255, 0.00002049, -0.00004108, 0.00004431, -0.00006189, 0.00005700, -0.00006325, 0.00004844, -0.00010871, 0.00007316, -0.00038836, 0.00047408, 0.00042509, -0.00037139, 0.00005627, -0.00009208, 0.00003798, -0.00004710, 0.00001794, -0.00002584, 0.00001534, -0.00003803, 0.00002848, -0.00002997, 0.00003879, -0.00001758, 0.00002334, 0.00000011, 0.00000465, 0.00000441, 0.00000221, -0.00002915, 0.00001274, -0.00000739, 0.00001200, -0.00002494, 0.00000977, -0.00001105, 0.00001361, -0.00001057, 0.00000494, -0.00001068, 0.00002328, -0.00002545, -0.00000244, -0.00000447, 0.00002303, -0.00001224, 0.00000731, -0.00001293, 0.00001096, -0.00000873, 0.00000169, 0.00000482, 0.00000441, },
		{ -0.00000008, -0.00000016, -0.00000623, -0.00000142, 0.00000121, 0.00000244, 0.00000670, -0.00000409, -0.00000720, 0.00000503, 0.00000242, -0.00000153, 0.00001103, -0.00001209, -0.00000860, 0.00000392, 0.00000705, -0.00000094, 0.00000148, -0.00000623, 0.00000130, -0.00000217, 0.00000928, -0.00000367, -0.00000463, -0.00000369, 0.00001382, -0.00000942, 0.00000040, 0.00000491, -0.00000976, 0.00000451, -0.00000072, -0.00000028, -0.00000195, 0.00001126, -0.00000954, 0.00000343, -0.00000760, 0.00000712, -0.00000548, 0.00000569, -0.00001461, 0.00001335, -0.00000409, 0.00000759, -0.00000754, -0.00000164, 0.00000487, 0.00000506, -0.00001229, 0.00000502, 0.00000525, -0.00000551, -0.00001362, 0.00001620, -0.00000620, 0.00000947, -0.00000841, 0.00000968, -0.00000656, -0.00000104, 0.00000702, -0.00000236, 0.00000643, -0.00000617, -0.00001776, 0.00002249, -0.00002063, 0.00001106, -0.00002000, 0.00001383, -0.00002929, 0.00002797, -0.00004914, 0.00005194, -0.00006764, 0.00006086, -0.00006710, 0.00005609, -0.00011641, 0.00008969, -0.00037715, 0.00033778, 0.00057767, -0.00040792, 0.00005725, -0.00009896, 0.00004170, -0.00005108, 0.00002189, -0.00003129, 0.00002194, -0.00004351, 0.00003275, -0.00003441, 0.00003973, -0.00001831, 0.00002291, -0.00000176, 0.00000845, -0.00000175, 0.00001242, -0.00003705, 0.00001695, -0.00001099, 0.00001619, -0.00002751, 0.00001145, -0.00001337, 0.00001548, -0.00001158, 0.00000710, -0.00001400, 0.00002496, -0.00002343, -0.00000331, -0.00000739, 0.00002381, -0.00000967, 0.00000549, -0.00001187, 0.00000881, -0.00000570, -0.00000119, 0.00000602, 0.00000393, -0.00000210, },
		{ 0.00000057, -0.00000581, -0.00000282, 0.00000195, 0.00000164, 0.00000723, -0.00000341, -0.00000680, 0.00000221, 0.00000535, -0.00000418, 0.00001256, -0.00001066, -0.00000907, 0.00000149, 0.00000853, -0.00000125, 0.00000240, -0.00000673, 0.00000157, -0.00000316, 0.00000897, -0.00000175, -0.00000484, -0.00000499, 0.00001301, -0.00000617, -0.00000341, 0.00000825, -0.00001121, 0.00000438, -0.00000079, 0.00000059, -0.00000397, 0.00001310, -0.00000970, 0.00000332, -0.00000780, 0.00000685, -0.00000528, 0.00000594, -0.00001368, 0.00001069, -0.00000130, 0.00000484, -0.00000373, -0.00000581, 0.00000650, 0.00000558, -0.00001094, 0.00000175, 0.00000701, -0.00000435, -0.00001478, 0.00001476, -0.00000443, 0.00000816, -0.00000675, 0.00000759, -0.00000381, -0.00000373, 0.00000663, 0.00000090, 0.00000148, 0.00000127, -0.00002490, 0.00002642, -0.00002169, 0.00001218, -0.00002164, 0.00001648, -0.00003207, 0.00003103, -0.00005205, 0.00005342, -0.00006629, 0.00005691, -0.00006230, 0.00005305, -0.00011004, 0.00008457, -0.00033240, 0.00017575, 0.00072432, -0.00042291, 0.00004322, -0.00009633, 0.00003805, -0.00004941, 0.00002135, -0.00003299, 0.00002500, -0.00004566, 0.00003403, -0.00003594, 0.00003820, -0.00001722, 0.00002142, -0.00000314, 0.00001183, -0.00000738, 0.00002087, -0.00004176, 0.00001768, -0.00001146, 0.00001748, -0.00002731, 0.00001062, -0.00001330, 0.00001497, -0.00001049, 0.00000736, -0.00001511, 0.00002411, -0.00001943, -0.00000543, -0.00000861, 0.00002218, -0.00000516, 0.00000250, -0.00000976, 0.00000565, -0.00000204, -0.00000419, 0.00000715, 0.00000346, -0.00000124, -0.00000080, },
		{ -0.00000517, -0.00000417, 0.00000241, 0.00000111, 0.00000753, -0.00000278, -0.00000602, -0.00000070, 0.00000780, -0.00000594, 0.00001296, -0.00000852, -0.00000965, -0.00000057, 0.00000917, -0.00000082, 0.00000270, -0.00000656, 0.00000123, -0.00000338, 0.00000778, 0.00000060, -0.00000515, -0.00000578, 0.00001135, -0.00000244, -0.00000708, 0.00001084, -0.00001157, 0.00000338, -0.00000034, 0.00000107, -0.00000538, 0.00001388, -0.00000860, 0.00000206, -0.00000683, 0.00000536, -0.00000383, 0.00000469, -0.00001110, 0.00000656, 0.00000268, 0.00000102, 0.00000090, -0.00001014, 0.00000785, 0.00000588, -0.00000882, -0.00000195, 0.00000872, -0.00000342, -0.00001507, 0.00001236, -0.00000197, 0.00000619, -0.00000440, 0.00000491, -0.00000074, -0.00000612, 0.00000567, 0.00000417, -0.00000253, 0.00000649, -0.00002846, 0.00002598, -0.00001848, 0.00000933, -0.00001933, 0.00001488, -0.00003014, 0.00002889, -0.00004895, 0.00004789, -0.00005705, 0.00004444, -0.00004808, 0.00003853, -0.00008902, 0.00005837, -0.00025917, -0.00000035, 0.00085168, -0.00040766, 0.00000989, -0.00008160, 0.00002508, -0.00004054, 0.00001499, -0.00002968, 0.00002338, -0.00004345, 0.00003138, -0.00003370, 0.00003342, -0.00001354, 0.00001804, -0.00000295, 0.00001347, -0.00001084, 0.00002581, -0.00004183, 0.00001389, -0.00000809, 0.00001533, -0.00002390, 0.00000692, -0.00001051, 0.00001184, -0.00000712, 0.00000552, -0.00001381, 0.00002072, -0.00001354, -0.00000881, -0.00000802, 0.00001824, 0.00000093, -0.00000128, -0.00000689, 0.00000177, 0.00000191, -0.00000707, 0.00000806, 0.00000304, -0.00000039, -0.00000148, 0.00000111, },
		{ -0.00000522, 0.00000239, 0.00000102, 0.00000748, -0.00000207, -0.00000514, -0.00000327, 0.00000934, -0.00000651, 0.00001210, -0.00000562, -0.00001046, -0.00000203, 0.00000883, 0.00000039, 0.00000236, -0.00000576, 0.00000031, -0.00000291, 0.00000588, 0.00000312, -0.00000536, -0.00000621, 0.00000912, 0.00000133, -0.00001010, 0.00001227, -0.00001065, 0.00000145, 0.00000066, 0.00000104, -0.00000599, 0.00001343, -0.00000617, -0.00000033, -0.00000476, 0.00000273, -0.00000121, 0.00000212, -0.00000716, 0.00000140, 0.00000739, -0.00000338, 0.00000577, -0.00001410, 0.00000862, 0.00000605, -0.00000609, -0.00000583, 0.00001025, -0.00000277, -0.00001447, 0.00000907, 0.00000111, 0.00000357, -0.00000136, 0.00000161, 0.00000271, -0.00000847, 0.00000472, 0.00000648, -0.00000442, 0.00000835, -0.00002752, 0.00002069, -0.00001082, 0.00000250, -0.00001306, 0.00000901, -0.00002350, 0.00002155, -0.00003989, 0.00003552, -0.00004031, 0.00002403, -0.00002521, 0.00001362, -0.00005511, 0.00001477, -0.00016594, -0.00017646, 0.00094619, -0.00035482, -0.00004535, -0.00005356, 0.00000198, -0.00002385, 0.00000222, -0.00002080, 0.00001649, -0.00003636, 0.00002429, -0.00002723, 0.00002496, -0.00000680, 0.00001212, -0.00000033, 0.00001228, -0.00001087, 0.00002599, -0.00003642, 0.00000524, -0.00000083, 0.00000979, -0.00001738, 0.00000046, -0.00000514, 0.00000628, -0.00000169, 0.00000176, -0.00001030, 0.00001509, -0.00000616, -0.00001320, -0.00000582, 0.00001241, 0.00000800, -0.00000533, -0.00000367, -0.00000242, 0.00000580, -0.00000958, 0.00000866, 0.00000271, 0.00000048, -0.00000215, 0.00000158, -0.00000448, },
		{ 0.00000173, 0.00000144, 0.00000701, -0.00000114, -0.00000443, -0.00000513, 0.00000965, -0.00000572, 0.00000998, -0.00000207, -0.00001149, -0.00000283, 0.00000752, 0.00000226, 0.00000152, -0.00000444, -0.00000103, -0.00000193, 0.00000358, 0.00000551, -0.00000527, -0.00000645, 0.00000664, 0.00000471, -0.00001210, 0.00001230, -0.00000840, -0.00000129, 0.00000214, 0.00000049, -0.00000576, 0.00001176, -0.00000256, -0.00000364, -0.00000184, -0.00000075, 0.00000222, -0.00000136, -0.00000240, -0.00000419, 0.00001221, -0.00000769, 0.00001024, -0.00001714, 0.00000853, 0.00000622, -0.00000294, -0.00000962, 0.00001147, -0.00000235, -0.00001304, 0.00000506, 0.00000467, 0.00000042, 0.00000229, -0.00000224, 0.00000654, -0.00001095, 0.00000431, 0.00000708, -0.00000343, 0.00000628, -0.00002194, 0.00001086, 0.00000065, -0.00000756, -0.00000357, -0.00000036, -0.00001298, 0.00000991, -0.00002588, 0.00001755, -0.00001765, -0.00000239, 0.00000402, -0.00001881, -0.00001228, -0.00004000, -0.00006360, -0.00033764, 0.00099569, -0.00025976, -0.00012253, -0.00001297, -0.00003054, 0.00000000, -0.00001648, -0.00000666, 0.00000459, -0.00002459, 0.00001291, -0.00001664, 0.00001289, 0.00000303, 0.00000346, 0.00000516, 0.00000767, -0.00000679, 0.00002095, -0.00002555, -0.00000778, 0.00000964, 0.00000156, -0.00000841, -0.00000812, 0.00000217, -0.00000104, 0.00000517, -0.00000338, -0.00000513, 0.00000792, 0.00000205, -0.00001809, -0.00000247, 0.00000535, 0.00001525, -0.00000901, -0.00000056, -0.00000652, 0.00000927, -0.00001151, 0.00000887, 0.00000244, 0.00000138, -0.00000287, 0.00000207, -0.00000394, -0.00000579, },
		{ 0.00000236, 0.00000612, 0.00000006, -0.00000408, -0.00000605, 0.00000862, -0.00000362, 0.00000681, 0.00000188, -0.00001263, -0.00000302, 0.00000541, 0.00000455, 0.00000040, -0.00000284, -0.00000259, -0.00000070, 0.00000120, 0.00000744, -0.00000471, -0.00000666, 0.00000421, 0.00000734, -0.00001281, 0.00001088, -0.00000498, -0.00000461, 0.00000387, -0.00000043, -0.00000480, 0.00000906, 0.00000190, -0.00000746, 0.00000151, -0.00000461, 0.00000597, -0.00000516, 0.00000250, -0.00000949, 0.00001644, -0.00001128, 0.00001369, -0.00001881, 0.00000739, 0.00000648, 0.00000041, -0.00001307, 0.00001221, -0.00000212, -0.00001094, 0.00000056, 0.00000847, -0.00000307, 0.00000635, -0.00000646, 0.00001061, -0.00001364, 0.00000474, 0.00000558, 0.00000066, 0.00000043, -0.00001237, -0.00000238, 0.00001451, -0.00001941, 0.00000770, -0.00001177, -0.00000014, -0.00000433, -0.00000880, -0.00000381, 0.00000830, -0.00003176, 0.00003606, -0.00005446, 0.00003388, -0.00009795, 0.00003581, -0.00046967, 0.00099095, -0.00012161, -0.00021856, 0.00003709, -0.00006998, 0.00002900, -0.00003946, 0.00001141, -0.00001120, -0.00000915, -0.00000188, -0.00000272, -0.00000208, 0.00001540, -0.00000759, 0.00001334, -0.00000028, 0.00000128, 0.00001112, -0.00001019, -0.00002383, 0.00002190, -0.00000807, 0.00000183, -0.00001771, 0.00001037, -0.00000912, 0.00001252, -0.00000905, 0.00000085, 0.00000009, 0.00001022, -0.00002281, 0.00000137, -0.00000207, 0.00002182, -0.00001169, 0.00000200, -0.00001014, 0.00001203, -0.00001270, 0.00000867, 0.00000223, 0.00000235, -0.00000368, 0.00000266, -0.00000367, -0.00000575, 0.00000041, },
		{ 0.00000489, 0.00000154, -0.00000418, -0.00000595, 0.00000632, -0.00000049, 0.00000298, 0.00000589, -0.00001366, -0.00000278, 0.00000278, 0.00000691, -0.00000067, -0.00000123, -0.00000406, 0.00000047, -0.00000091, 0.00000864, -0.00000355, -0.00000697, 0.00000207, 0.00000900, -0.00001218, 0.00000818, -0.00000075, -0.00000809, 0.00000555, -0.00000150, -0.00000335, 0.00000571, 0.00000672, -0.00001124, 0.00000474, -0.00000830, 0.00000943, -0.00000863, 0.00000683, -0.00001381, 0.00001945, -0.00001355, 0.00001562, -0.00001881, 0.00000512, 0.00000686, 0.00000375, -0.00001588, 0.00001234, -0.00000196, -0.00000841, -0.00000409, 0.00001219, -0.00000662, 0.00001051, -0.00001074, 0.00001466, -0.00001641, 0.00000603, 0.00000204, 0.00000745, -0.00000832, -0.00000022, -0.00001720, 0.00002878, -0.00003107, 0.00001883, -0.00002325, 0.00001294, -0.00001894, 0.00000887, -0.00002571, 0.00003425, -0.00006031, 0.00006662, -0.00008819, 0.00007694, -0.00015049, 0.00012061, -0.00056073, 0.00092709, 0.00005594, -0.00032687, 0.00009123, -0.00011216, 0.00005974, -0.00006398, 0.00003111, -0.00002886, 0.00000817, -0.00001848, 0.00001305, -0.00001862, 0.00002914, -0.00002007, 0.00002340, -0.00001078, 0.00001240, -0.00000217, 0.00000783, -0.00004086, 0.00003403, -0.00001740, 0.00001182, -0.00002692, 0.00001815, -0.00001670, 0.00001922, -0.00001424, 0.00000664, -0.00000734, 0.00001741, -0.00002660, 0.00000494, -0.00000894, 0.00002687, -0.00001285, 0.00000365, -0.00001296, 0.00001384, -0.00001307, 0.00000805, 0.00000206, 0.00000339, -0.00000459, 0.00000338, -0.00000372, -0.00000509, -0.00000151, 0.00000369, },
		{ 0.00000318, -0.00000469, -0.00000493, 0.00000305, 0.00000325, -0.00000101, 0.00000949, -0.00001429, -0.00000236, 0.00000002, 0.00000894, -0.00000139, 0.00000010, -0.00000521, 0.00000132, -0.00000248, 0.00000896, -0.00000177, -0.00000741, 0.00000033, 0.00000960, -0.00001034, 0.00000454, 0.00000381, -0.00001125, 0.00000680, -0.00000242, -0.00000174, 0.00000214, 0.00001131, -0.00001438, 0.00000727, -0.00001122, 0.00001202, -0.00001115, 0.00000994, -0.00001654, 0.00002075, -0.00001409, 0.00001573, -0.00001702, 0.00000181, 0.00000734, 0.00000687, -0.00001781, 0.00001171, -0.00000172, -0.00000573, -0.00000849, 0.00001544, -0.00000983, 0.00001437, -0.00001462, 0.00001829, -0.00001891, 0.00000788, -0.00000301, 0.00001595, -0.00001845, 0.00001251, -0.00003132, 0.00004115, -0.00004037, 0.00002773, -0.00003266, 0.00002400, -0.00003148, 0.00002442, -0.00004504, 0.00005666, -0.00008410, 0.00009125, -0.00011484, 0.00011060, -0.00018965, 0.00018100, -0.00060292, 0.00080454, 0.00026432, -0.00043766, 0.00014204, -0.00015153, 0.00008781, -0.00008644, 0.00004942, -0.00004575, 0.00002498, -0.00003473, 0.00002871, -0.00003489, 0.00004259, -0.00003248, 0.00003396, -0.00002235, 0.00002489, -0.00001688, 0.00002609, -0.00005640, 0.00004384, -0.00002458, 0.00001993, -0.00003425, 0.00002413, -0.00002251, 0.00002413, -0.00001791, 0.00001125, -0.00001337, 0.00002274, -0.00002877, 0.00000752, -0.00001445, 0.00002972, -0.00001211, 0.00000418, -0.00001478, 0.00001456, -0.00001257, 0.00000705, 0.00000194, 0.00000444, -0.00000552, 0.00000417, -0.00000402, -0.00000393, -0.00000382, 0.00000519, 0.00000350, },
		{ -0.00000545, -0.00000326, -0.00000074, 0.00000700, -0.00000458, 0.00001225, -0.00001423, -0.00000207, -0.00000248, 0.00001027, -0.00000147, 0.00000092, -0.00000581, 0.00000165, -0.00000335, 0.00000834, 0.00000055, -0.00000792, -0.00000099, 0.00000926, -0.00000759, 0.00000045, 0.00000811, -0.00001359, 0.00000727, -0.00000291, -0.00000031, -0.00000115, 0.00001512, -0.00001634, 0.00000860, -0.00001291, 0.00001324, -0.00001222, 0.00001137, -0.00001731, 0.00002004, -0.00001273, 0.00001394, -0.00001355, -0.00000233, 0.00000786, 0.00000960, -0.00001866, 0.00001022, -0.00000125, -0.00000323, -0.00001223, 0.00001784, -0.00001230, 0.00001744, -0.00001760, 0.00002096, -0.00002068, 0.00000972, -0.00000865, 0.00002475, -0.00002808, 0.00002359, -0.00004238, 0.00004938, -0.00004527, 0.00003247, -0.00003804, 0.00003095, -0.00003971, 0.00003537, -0.00005899, 0.00007235, -0.00009963, 0.00010612, -0.00012998, 0.00012969, -0.00020922, 0.00021025, -0.00059319, 0.00062946, 0.00049045, -0.00053860, 0.00018086, -0.00018181, 0.00010831, -0.00010288, 0.00006298, -0.00005890, 0.00003862, -0.00004822, 0.00004198, -0.00004879, 0.00005381, -0.00004300, 0.00004321, -0.00003310, 0.00003662, -0.00003051, 0.00004188, -0.00006792, 0.00004923, -0.00002792, 0.00002470, -0.00003839, 0.00002712, -0.00002547, 0.00002629, -0.00001923, 0.00001383, -0.00001716, 0.00002552, -0.00002879, 0.00000853, -0.00001792, 0.00002993, -0.00000934, 0.00000355, -0.00001553, 0.00001417, -0.00001126, 0.00000569, 0.00000189, 0.00000541, -0.00000635, 0.00000490, -0.00000445, -0.00000248, -0.00000623, 0.00000656, 0.00000220, 0.00000479, },
		{ -0.00000130, -0.00000450, 0.00001018, -0.00000718, 0.00001377, -0.00001326, -0.00000215, -0.00000439, 0.00001062, -0.00000075, 0.00000111, -0.00000576, 0.00000138, -0.00000346, 0.00000688, 0.00000319, -0.00000834, -0.00000197, 0.00000817, -0.00000438, -0.00000352, 0.00001158, -0.00001468, 0.00000668, -0.00000272, 0.00000062, -0.00000373, 0.00001766, -0.00001671, 0.00000840, -0.00001303, 0.00001281, -0.00001158, 0.00001088, -0.00001599, 0.00001735, -0.00000954, 0.00001044, -0.00000871, -0.00000696, 0.00000828, 0.00001181, -0.00001828, 0.00000782, -0.00000042, -0.00000124, -0.00001491, 0.00001899, -0.00001362, 0.00001924, -0.00001915, 0.00002215, -0.00002115, 0.00001084, -0.00001379, 0.00003227, -0.00003524, 0.00003087, -0.00004831, 0.00005166, -0.00004426, 0.00003165, -0.00003797, 0.00003228, -0.00004198, 0.00003986, -0.00006542, 0.00007899, -0.00010441, 0.00010856, -0.00013070, 0.00013092, -0.00020568, 0.00020565, -0.00053393, 0.00041346, 0.00071765, -0.00061594, 0.00019885, -0.00019688, 0.00011658, -0.00010957, 0.00006865, -0.00006553, 0.00004654, -0.00005663, 0.00005070, -0.00005825, 0.00006085, -0.00004980, 0.00004930, -0.00004100, 0.00004533, -0.00004057, 0.00005266, -0.00007324, 0.00004856, -0.00002621, 0.00002513, -0.00003844, 0.00002630, -0.00002485, 0.00002510, -0.00001768, 0.00001384, -0.00001819, 0.00002538, -0.00002640, 0.00000765, -0.00001899, 0.00002736, -0.00000469, 0.00000190, -0.00001526, 0.00001270, -0.00000920, 0.00000401, 0.00000198, 0.00000617, -0.00000691, 0.00000538, -0.00000477, -0.00000105, -0.00000837, 0.00000749, 0.00000125, 0.00000615, -0.00000621, },
		{ -0.00000767, 0.00001223, -0.00000840, 0.00001381, -0.00001126, -0.00000277, -0.00000545, 0.00000985, 0.00000080, 0.00000068, -0.00000509, 0.00000054, -0.00000290, 0.00000480, 0.00000586, -0.00000849, -0.00000278, 0.00000665, -0.00000116, -0.00000681, 0.00001375, -0.00001421, 0.00000490, -0.00000174, 0.00000084, -0.00000532, 0.00001864, -0.00001529, 0.00000654, -0.00001152, 0.00001067, -0.00000921, 0.00000853, -0.00001275, 0.00001293, -0.00000486, 0.00000563, -0.00000298, -0.00001164, 0.00000847, 0.00001339, -0.00001661, 0.00000451, 0.00000088, -0.00000003, -0.00001622, 0.00001862, -0.00001345, 0.00001938, -0.00001883, 0.00002139, -0.00001981, 0.00001052, -0.00001730, 0.00003698, -0.00003821, 0.00003266, -0.00004769, 0.00004699, -0.00003665, 0.00002470, -0.00003184, 0.00002732, -0.00003752, 0.00003700, -0.00006331, 0.00007548, -0.00009740, 0.00009759, -0.00011603, 0.00011349, -0.00017872, 0.00016875, -0.00043269, 0.00017265, 0.00092708, -0.00065610, 0.00018823, -0.00019179, 0.00010898, -0.00010368, 0.00006404, -0.00006351, 0.00004681, -0.00005813, 0.00005312, -0.00006160, 0.00006213, -0.00005131, 0.00005060, -0.00004426, 0.00004906, -0.00004497, 0.00005650, -0.00007092, 0.00004098, -0.00001894, 0.00002083, -0.00003408, 0.00002140, -0.00002045, 0.00002045, -0.00001316, 0.00001117, -0.00001634, 0.00002233, -0.00002168, 0.00000485, -0.00001762, 0.00002224, 0.00000144, -0.00000046, -0.00001413, 0.00001032, -0.00000653, 0.00000206, 0.00000228, 0.00000656, -0.00000702, 0.00000541, -0.00000475, 0.00000005, -0.00000989, 0.00000764, 0.00000092, 0.00000704, -0.00000672, 0.00000052, },
		{ 0.00001275, -0.00000798, 0.00001230, -0.00000827, -0.00000396, -0.00000560, 0.00000796, 0.00000307, -0.00000025, -0.00000393, -0.00000071, -0.00000189, 0.00000241, 0.00000824, -0.00000814, -0.00000359, 0.00000501, 0.00000159, -0.00000900, 0.00001432, -0.00001212, 0.00000198, -0.00000002, 0.00000031, -0.00000577, 0.00001798, -0.00001212, 0.00000317, -0.00000854, 0.00000705, -0.00000538, 0.00000466, -0.00000804, 0.00000730, 0.00000076, 0.00000008, 0.00000304, -0.00001592, 0.00000828, 0.00001430, -0.00001368, 0.00000042, 0.00000266, 0.00000023, -0.00001598, 0.00001657, -0.00001163, 0.00001761, -0.00001637, 0.00001840, -0.00001631, 0.00000822, -0.00001830, 0.00003775, -0.00003586, 0.00002808, -0.00004006, 0.00003535, -0.00002273, 0.00001203, -0.00002004, 0.00001641, -0.00002665, 0.00002705, -0.00005290, 0.00006216, -0.00007917, 0.00007410, -0.00008721, 0.00007920, -0.00013127, 0.00010521, -0.00030139, -0.00007385, 0.00109968, -0.00064733, 0.00014360, -0.00016368, 0.00008366, -0.00008387, 0.00004802, -0.00005179, 0.00003844, -0.00005180, 0.00004829, -0.00005790, 0.00005672, -0.00004657, 0.00004605, -0.00004165, 0.00004648, -0.00004240, 0.00005238, -0.00006052, 0.00002661, -0.00000650, 0.00001221, -0.00002566, 0.00001279, -0.00001265, 0.00001274, -0.00000608, 0.00000611, -0.00001190, 0.00001675, -0.00001504, 0.00000040, -0.00001409, 0.00001506, 0.00000845, -0.00000311, -0.00001239, 0.00000720, -0.00000340, -0.00000012, 0.00000288, 0.00000643, -0.00000649, 0.00000479, -0.00000416, 0.00000053, -0.00001048, 0.00000679, 0.00000138, 0.00000726, -0.00000674, 0.00000182, -0.00000980, },
		{ -0.00000594, 0.00000942, -0.00000450, -0.00000561, -0.00000489, 0.00000514, 0.00000579, -0.00000143, -0.00000250, -0.00000214, -0.00000068, 0.00000004, 0.00001000, -0.00000712, -0.00000456, 0.00000356, 0.00000353, -0.00000981, 0.00001320, -0.00000853, -0.00000181, 0.00000225, -0.00000087, -0.00000517, 0.00001581, -0.00000749, -0.00000133, -0.00000452, 0.00000238, -0.00000057, -0.00000017, -0.00000250, 0.00000115, 0.00000665, -0.00000552, 0.00000870, -0.00001934, 0.00000757, 0.00001452, -0.00000961, -0.00000427, 0.00000486, -0.00000051, -0.00001417, 0.00001291, -0.00000817, 0.00001391, -0.00001177, 0.00001317, -0.00001061, 0.00000371, -0.00001629, 0.00003401, -0.00002783, 0.00001718, -0.00002602, 0.00001781, -0.00000387, -0.00000497, -0.00000393, 0.00000090, -0.00001075, 0.00001142, -0.00003569, 0.00004082, -0.00005191, 0.00004075, -0.00004750, 0.00003227, -0.00006914, 0.00002395, -0.00015494, -0.00030590, 0.00121823, -0.00058148, 0.00006312, -0.00011257, 0.00004108, -0.00005071, 0.00002111, -0.00003076, 0.00002171, -0.00003782, 0.00003629, -0.00004713, 0.00004456, -0.00003549, 0.00003542, -0.00003283, 0.00003719, -0.00003257, 0.00004043, -0.00004280, 0.00000664, 0.00000983, 0.00000040, -0.00001419, 0.00000142, -0.00000237, 0.00000287, 0.00000276, -0.00000063, -0.00000558, 0.00000940, -0.00000717, -0.00000513, -0.00000896, 0.00000658, 0.00001558, -0.00000555, -0.00001033, 0.00000358, 0.00000001, -0.00000248, 0.00000383, 0.00000568, -0.00000522, 0.00000340, -0.00000285, 0.00000020, -0.00000997, 0.00000484, 0.00000269, 0.00000676, -0.00000610, 0.00000234, -0.00001056, 0.00001158, },
		{ 0.00000553, -0.00000032, -0.00000748, -0.00000353, 0.00000173, 0.00000858, -0.00000255, -0.00000109, -0.00000349, 0.00000043, -0.00000197, 0.00001091, -0.00000535, -0.00000576, 0.00000247, 0.00000445, -0.00000919, 0.00001056, -0.00000382, -0.00000606, 0.00000472, -0.00000246, -0.00000375, 0.00001250, -0.00000191, -0.00000638, -0.00000004, -0.00000271, 0.00000455, -0.00000526, 0.00000312, -0.00000476, 0.00001208, -0.00001051, 0.00001344, -0.00002155, 0.00000629, 0.00001407, -0.00000462, -0.00000924, 0.00000730, -0.00000214, -0.00001098, 0.00000788, -0.00000333, 0.00000855, -0.00000530, 0.00000603, -0.00000300, -0.00000286, -0.00001131, 0.00002590, -0.00001466, 0.00000108, -0.00000722, -0.00000354, 0.00001770, -0.00002406, 0.00001432, -0.00001701, 0.00000793, -0.00000752, -0.00001426, 0.00001442, -0.00001913, 0.00000169, -0.00000182, -0.00002127, -0.00000016, -0.00006404, -0.00000948, -0.00050447, 0.00126934, -0.00045535, -0.00005071, -0.00004184, -0.00001568, -0.00000690, -0.00001445, -0.00000233, -0.00000175, -0.00001759, 0.00001830, -0.00003034, 0.00002663, -0.00001891, 0.00001943, -0.00001840, 0.00002183, -0.00001638, 0.00002198, -0.00001962, -0.00001682, 0.00002805, -0.00001286, -0.00000123, -0.00001129, 0.00000904, -0.00000789, 0.00001219, -0.00000803, 0.00000167, 0.00000126, 0.00000100, -0.00001101, -0.00000300, -0.00000228, 0.00002202, -0.00000732, -0.00000820, -0.00000030, 0.00000353, -0.00000493, 0.00000514, 0.00000428, -0.00000319, 0.00000125, -0.00000080, -0.00000099, -0.00000835, 0.00000186, 0.00000473, 0.00000558, -0.00000478, 0.00000194, -0.00000988, 0.00000877, -0.00000251, },
		{ 0.00000380, -0.00000925, -0.00000184, -0.00000186, 0.00001102, -0.00000327, 0.00000004, -0.00000450, 0.00000120, -0.00000337, 0.00001081, -0.00000283, -0.00000717, 0.00000183, 0.00000432, -0.00000730, 0.00000679, 0.00000145, -0.00001020, 0.00000696, -0.00000410, -0.00000189, 0.00000855, 0.00000397, -0.00001130, 0.00000423, -0.00000754, 0.00000929, -0.00000986, 0.00000805, -0.00000972, 0.00001643, -0.00001432, 0.00001677, -0.00002230, 0.00000440, 0.00001304, 0.00000095, -0.00001412, 0.00000972, -0.00000442, -0.00000680, 0.00000197, 0.00000241, 0.00000202, 0.00000245, -0.00000239, 0.00000587, -0.00001087, -0.00000396, 0.00001429, 0.00000225, -0.00001826, 0.00001383, -0.00002592, 0.00003913, -0.00004251, 0.00003203, -0.00003463, 0.00002658, -0.00002681, 0.00000814, -0.00001333, 0.00001486, -0.00003811, 0.00004405, -0.00007442, 0.00006691, -0.00014718, 0.00011954, -0.00065372, 0.00124520, -0.00027169, -0.00019070, 0.00004181, -0.00008107, 0.00004293, -0.00005474, 0.00003020, -0.00002909, 0.00000638, -0.00000346, -0.00000956, 0.00000480, 0.00000142, -0.00000029, 0.00000007, 0.00000210, 0.00000424, -0.00000059, 0.00000629, -0.00004097, 0.00004565, -0.00002545, 0.00001139, -0.00002363, 0.00001999, -0.00001806, 0.00002093, -0.00001494, 0.00000874, -0.00000657, 0.00000853, -0.00001643, 0.00000296, -0.00001060, 0.00002703, -0.00000801, -0.00000626, -0.00000420, 0.00000695, -0.00000738, 0.00000675, 0.00000230, -0.00000051, -0.00000154, 0.00000186, -0.00000292, -0.00000578, -0.00000190, 0.00000727, 0.00000393, -0.00000289, 0.00000069, -0.00000786, 0.00000465, 0.00000181, 0.00000117, },
		{ -0.00001057, -0.00000018, -0.00000520, 0.00001274, -0.00000334, 0.00000067, -0.00000499, 0.00000143, -0.00000402, 0.00000971, 0.00000027, -0.00000867, 0.00000158, 0.00000331, -0.00000450, 0.00000246, 0.00000662, -0.00001364, 0.00000852, -0.00000541, -0.00000003, 0.00000453, 0.00000948, -0.00001541, 0.00000763, -0.00001147, 0.00001297, -0.00001332, 0.00001166, -0.00001315, 0.00001918, -0.00001653, 0.00001839, -0.00002149, 0.00000196, 0.00001153, 0.00000674, -0.00001845, 0.00001176, -0.00000695, -0.00000216, -0.00000420, 0.00000841, -0.00000497, 0.00001064, -0.00001117, 0.00001504, -0.00001938, 0.00000470, 0.00000065, 0.00002086, -0.00003820, 0.00003415, -0.00004619, 0.00005738, -0.00005745, 0.00004643, -0.00004915, 0.00004227, -0.00004333, 0.00002807, -0.00003850, 0.00004556, -0.00007352, 0.00008419, -0.00012022, 0.00012362, -0.00021484, 0.00021913, -0.00074271, 0.00114464, -0.00003946, -0.00034525, 0.00012883, -0.00014753, 0.00009261, -0.00009470, 0.00006251, -0.00005651, 0.00003074, -0.00002600, 0.00001245, -0.00001832, 0.00002301, -0.00002139, 0.00002024, -0.00001954, 0.00002655, -0.00002419, 0.00003161, -0.00006268, 0.00006000, -0.00003522, 0.00002180, -0.00003391, 0.00002892, -0.00002623, 0.00002770, -0.00002026, 0.00001456, -0.00001303, 0.00001449, -0.00002063, 0.00000812, -0.00001752, 0.00003002, -0.00000738, -0.00000465, -0.00000792, 0.00001006, -0.00000965, 0.00000850, -0.00000006, 0.00000257, -0.00000471, 0.00000487, -0.00000532, -0.00000260, -0.00000604, 0.00000990, 0.00000211, -0.00000068, -0.00000121, -0.00000486, -0.00000025, 0.00000640, -0.00000302, 0.00000734, },
		{ 0.00000114, -0.00000788, 0.00001340, -0.00000256, 0.00000069, -0.00000486, 0.00000108, -0.00000391, 0.00000776, 0.00000370, -0.00001004, 0.00000156, 0.00000169, -0.00000130, -0.00000181, 0.00001103, -0.00001585, 0.00000901, -0.00000605, 0.00000143, 0.00000099, 0.00001398, -0.00001811, 0.00000964, -0.00001396, 0.00001511, -0.00001514, 0.00001349, -0.00001470, 0.00002008, -0.00001693, 0.00001820, -0.00001919, -0.00000092, 0.00000966, 0.00001229, -0.00002179, 0.00001306, -0.00000929, 0.00000231, -0.00000989, 0.00001390, -0.00001155, 0.00001832, -0.00001923, 0.00002337, -0.00002718, 0.00001326, -0.00001316, 0.00003871, -0.00005588, 0.00005067, -0.00006137, 0.00006969, -0.00006637, 0.00005512, -0.00005810, 0.00005244, -0.00005432, 0.00004247, -0.00005758, 0.00006897, -0.00010001, 0.00011350, -0.00015274, 0.00016298, -0.00025870, 0.00028027, -0.00076662, 0.00097363, 0.00022655, -0.00049927, 0.00020778, -0.00020643, 0.00013527, -0.00012866, 0.00008982, -0.00007987, 0.00005180, -0.00004599, 0.00003258, -0.00003975, 0.00004303, -0.00004111, 0.00003935, -0.00004017, 0.00004740, -0.00004548, 0.00005302, -0.00007903, 0.00006875, -0.00004035, 0.00002842, -0.00004066, 0.00003449, -0.00003118, 0.00003147, -0.00002305, 0.00001822, -0.00001729, 0.00001819, -0.00002302, 0.00001181, -0.00002240, 0.00003060, -0.00000534, -0.00000346, -0.00001127, 0.00001267, -0.00001158, 0.00001020, -0.00000250, 0.00000570, -0.00000787, 0.00000784, -0.00000777, 0.00000071, -0.00001003, 0.00001216, 0.00000051, 0.00000154, -0.00000341, -0.00000135, -0.00000529, 0.00001057, -0.00000642, 0.00000984, -0.00001111, },
		{ -0.00000961, 0.00001284, -0.00000092, 0.00000010, -0.00000415, 0.00000020, -0.00000317, 0.00000523, 0.00000709, -0.00001101, 0.00000153, -0.00000016, 0.00000177, -0.00000537, 0.00001409, -0.00001641, 0.00000820, -0.00000580, 0.00000217, -0.00000163, 0.00001700, -0.00001901, 0.00000992, -0.00001474, 0.00001541, -0.00001509, 0.00001336, -0.00001425, 0.00001910, -0.00001555, 0.00001631, -0.00001560, -0.00000408, 0.00000760, 0.00001718, -0.00002369, 0.00001326, -0.00001099, 0.00000597, -0.00001439, 0.00001812, -0.00001687, 0.00002447, -0.00002548, 0.00002971, -0.00003303, 0.00002025, -0.00002519, 0.00005339, -0.00006861, 0.00006072, -0.00006905, 0.00007403, -0.00006752, 0.00005642, -0.00005980, 0.00005526, -0.00005780, 0.00004910, -0.00006797, 0.00008213, -0.00011432, 0.00012835, -0.00016790, 0.00018042, -0.00027379, 0.00029882, -0.00072716, 0.00074478, 0.00050693, -0.00063565, 0.00026670, -0.00024912, 0.00016421, -0.00015125, 0.00010759, -0.00009522, 0.00006605, -0.00006023, 0.00004784, -0.00005660, 0.00005868, -0.00005673, 0.00005463, -0.00005688, 0.00006372, -0.00006131, 0.00006760, -0.00008764, 0.00007022, -0.00003960, 0.00003020, -0.00004292, 0.00003583, -0.00003218, 0.00003162, -0.00002278, 0.00001918, -0.00001881, 0.00001922, -0.00002326, 0.00001363, -0.00002482, 0.00002867, -0.00000199, -0.00000268, -0.00001408, 0.00001457, -0.00001294, 0.00001156, -0.00000468, 0.00000847, -0.00001058, 0.00001031, -0.00000983, 0.00000366, -0.00001335, 0.00001359, -0.00000050, 0.00000342, -0.00000551, 0.00000213, -0.00000979, 0.00001365, -0.00000849, 0.00001096, -0.00001067, 0.00000186, },
		{ 0.00001106, 0.00000149, -0.00000097, -0.00000301, -0.00000103, -0.00000202, 0.00000245, 0.00001006, -0.00001131, 0.00000126, -0.00000186, 0.00000419, -0.00000772, 0.00001541, -0.00001514, 0.00000604, -0.00000459, 0.00000200, -0.00000308, 0.00001829, -0.00001794, 0.00000842, -0.00001375, 0.00001388, -0.00001320, 0.00001136, -0.00001198, 0.00001646, -0.00001264, 0.00001300, -0.00001104, -0.00000732, 0.00000550, 0.00002099, -0.00002383, 0.00001209, -0.00001163, 0.00000825, -0.00001708, 0.00002043, -0.00002019, 0.00002825, -0.00002900, 0.00003307, -0.00003586, 0.00002432, -0.00003369, 0.00006283, -0.00007430, 0.00006249, -0.00006785, 0.00006942, -0.00006021, 0.00004975, -0.00005359, 0.00005001, -0.00005295, 0.00004692, -0.00006839, 0.00008359, -0.00011489, 0.00012712, -0.00016403, 0.00017436, -0.00025903, 0.00027581, -0.00063228, 0.00047623, 0.00077946, -0.00073707, 0.00029470, -0.00026826, 0.00017397, -0.00015816, 0.00011225, -0.00009946, 0.00007076, -0.00006623, 0.00005579, -0.00006650, 0.00006765, -0.00006594, 0.00006373, -0.00006720, 0.00007299, -0.00006928, 0.00007332, -0.00008715, 0.00006369, -0.00003259, 0.00002685, -0.00004037, 0.00003265, -0.00002899, 0.00002802, -0.00001933, 0.00001728, -0.00001744, 0.00001755, -0.00002133, 0.00001343, -0.00002470, 0.00002440, 0.00000239, -0.00000222, -0.00001626, 0.00001558, -0.00001351, 0.00001229, -0.00000620, 0.00001044, -0.00001241, 0.00001187, -0.00001104, 0.00000573, -0.00001549, 0.00001377, -0.00000058, 0.00000464, -0.00000713, 0.00000503, -0.00001314, 0.00001514, -0.00000890, 0.00001053, -0.00000920, 0.00000187, -0.00001027, },
		{ 0.00000439, -0.00000225, -0.00000165, -0.00000239, -0.00000076, -0.00000019, 0.00001227, -0.00001072, 0.00000056, -0.00000308, 0.00000557, -0.00000852, 0.00001483, -0.00001211, 0.00000268, -0.00000252, 0.00000094, -0.00000328, 0.00001783, -0.00001503, 0.00000532, -0.00001122, 0.00001077, -0.00000976, 0.00000784, -0.00000830, 0.00001258, -0.00000861, 0.00000868, -0.00000592, -0.00001045, 0.00000352, 0.00002340, -0.00002198, 0.00000943, -0.00001095, 0.00000875, -0.00001755, 0.00002041, -0.00002100, 0.00002906, -0.00002916, 0.00003280, -0.00003492, 0.00002454, -0.00003743, 0.00006566, -0.00007178, 0.00005527, -0.00005759, 0.00005615, -0.00004498, 0.00003568, -0.00004000, 0.00003717, -0.00004015, 0.00003624, -0.00005905, 0.00007357, -0.00010209, 0.00011041, -0.00014203, 0.00014633, -0.00021723, 0.00021713, -0.00049513, 0.00018982, 0.00102129, -0.00078811, 0.00028349, -0.00025894, 0.00016123, -0.00014695, 0.00010184, -0.00009092, 0.00006441, -0.00006257, 0.00005504, -0.00006801, 0.00006850, -0.00006731, 0.00006513, -0.00006953, 0.00007366, -0.00006804, 0.00006935, -0.00007740, 0.00004950, -0.00001989, 0.00001882, -0.00003340, 0.00002532, -0.00002201, 0.00002105, -0.00001306, 0.00001279, -0.00001346, 0.00001348, -0.00001750, 0.00001137, -0.00002224, 0.00001820, 0.00000738, -0.00000194, -0.00001770, 0.00001554, -0.00001312, 0.00001213, -0.00000674, 0.00001125, -0.00001298, 0.00001214, -0.00001104, 0.00000654, -0.00001611, 0.00001246, 0.00000044, 0.00000500, -0.00000796, 0.00000695, -0.00001492, 0.00001474, -0.00000756, 0.00000864, -0.00000682, 0.00000121, -0.00000986, 0.00000822, },
		{ -0.00000346, -0.00000035, -0.00000361, 0.00000033, -0.00000237, 0.00001344, -0.00000912, -0.00000067, -0.00000362, 0.00000569, -0.00000770, 0.00001249, -0.00000764, -0.00000148, 0.00000010, -0.00000082, -0.00000240, 0.00001586, -0.00001065, 0.00000106, -0.00000760, 0.00000652, -0.00000526, 0.00000333, -0.00000377, 0.00000801, -0.00000396, 0.00000385, -0.00000067, -0.00001326, 0.00000179, 0.00002420, -0.00001814, 0.00000531, -0.00000886, 0.00000729, -0.00001565, 0.00001794, -0.00001912, 0.00002669, -0.00002575, 0.00002870, -0.00003001, 0.00002055, -0.00003588, 0.00006144, -0.00006102, 0.00003962, -0.00003942, 0.00003575, -0.00002353, 0.00001591, -0.00002068, 0.00001837, -0.00002106, 0.00001867, -0.00004160, 0.00005393, -0.00007812, 0.00008086, -0.00010520, 0.00010071, -0.00015453, 0.00013249, -0.00033240, -0.00009122, 0.00121139, -0.00077718, 0.00022882, -0.00021969, 0.00012546, -0.00011752, 0.00007639, -0.00006968, 0.00004709, -0.00004925, 0.00004541, -0.00006088, 0.00006091, -0.00006046, 0.00005838, -0.00006340, 0.00006536, -0.00005759, 0.00005620, -0.00005947, 0.00002912, -0.00000292, 0.00000732, -0.00002301, 0.00001477, -0.00001212, 0.00001159, -0.00000474, 0.00000637, -0.00000748, 0.00000765, -0.00001232, 0.00000785, -0.00001788, 0.00001068, 0.00001247, -0.00000168, -0.00001836, 0.00001437, -0.00001163, 0.00001088, -0.00000605, 0.00001067, -0.00001210, 0.00001095, -0.00000963, 0.00000590, -0.00001504, 0.00000962, 0.00000257, 0.00000444, -0.00000787, 0.00000763, -0.00001492, 0.00001243, -0.00000462, 0.00000557, -0.00000382, 0.00000006, -0.00000856, 0.00000463, 0.00000742, },
		{ 0.00000064, -0.00000445, 0.00000101, -0.00000383, 0.00001344, -0.00000655, -0.00000241, -0.00000342, 0.00000455, -0.00000542, 0.00000875, -0.00000225, -0.00000592, 0.00000288, -0.00000298, -0.00000076, 0.00001280, -0.00000534, -0.00000375, -0.00000346, 0.00000175, -0.00000033, -0.00000153, 0.00000097, 0.00000337, 0.00000075, -0.00000100, 0.00000429, -0.00001558, 0.00000040, 0.00002333, -0.00001248, 0.00000001, -0.00000551, 0.00000398, -0.00001157, 0.00001324, -0.00001475, 0.00002135, -0.00001906, 0.00002114, -0.00002148, 0.00001265, -0.00002931, 0.00005069, -0.00004312, 0.00001733, -0.00001565, 0.00001082, 0.00000146, -0.00000696, 0.00000183, -0.00000385, 0.00000175, -0.00000312, -0.00001888, 0.00002782, -0.00004663, 0.00004278, -0.00005872, 0.00004396, -0.00007946, 0.00003409, -0.00016235, -0.00034466, 0.00133281, -0.00069818, 0.00013140, -0.00015295, 0.00006928, -0.00007235, 0.00003811, -0.00003767, 0.00002049, -0.00002772, 0.00002813, -0.00004610, 0.00004578, -0.00004619, 0.00004423, -0.00004954, 0.00004903, -0.00003921, 0.00003566, -0.00003561, 0.00000490, 0.00001617, -0.00000588, -0.00001071, 0.00000238, -0.00000064, 0.00000084, 0.00000457, -0.00000106, -0.00000037, 0.00000090, -0.00000649, 0.00000343, -0.00001225, 0.00000256, 0.00001714, -0.00000129, -0.00001820, 0.00001204, -0.00000903, 0.00000849, -0.00000403, 0.00000862, -0.00000972, 0.00000828, -0.00000684, 0.00000381, -0.00001238, 0.00000541, 0.00000561, 0.00000309, -0.00000686, 0.00000704, -0.00001318, 0.00000847, -0.00000050, 0.00000180, -0.00000061, -0.00000132, -0.00000666, 0.00000068, 0.00001019, -0.00000429, },
		{ -0.00000475, 0.00000112, -0.00000446, 0.00001230, -0.00000318, -0.00000450, -0.00000258, 0.00000233, -0.00000207, 0.00000418, 0.00000336, -0.00001002, 0.00000534, -0.00000512, 0.00000122, 0.00000919, 0.00000023, -0.00000845, 0.00000054, -0.00000289, 0.00000438, -0.00000607, 0.00000527, -0.00000080, 0.00000503, -0.00000541, 0.00000862, -0.00001726, -0.00000062, 0.00002089, -0.00000539, -0.00000605, -0.00000124, -0.00000081, -0.00000581, 0.00000686, -0.00000845, 0.00001365, -0.00000982, 0.00001094, -0.00001024, 0.00000175, -0.00001876, 0.00003489, -0.00002022, -0.00000886, 0.00001057, -0.00001533, 0.00002676, -0.00002982, 0.00002449, -0.00002642, 0.00002511, -0.00002585, 0.00000560, -0.00000078, -0.00001217, 0.00000145, -0.00000882, -0.00001629, -0.00000170, -0.00006509, -0.00000269, -0.00055156, 0.00137455, -0.00055156, -0.00000269, -0.00006509, -0.00000170, -0.00001629, -0.00000882, 0.00000145, -0.00001217, -0.00000078, 0.00000560, -0.00002585, 0.00002511, -0.00002642, 0.00002449, -0.00002982, 0.00002676, -0.00001533, 0.00001057, -0.00000886, -0.00002022, 0.00003489, -0.00001876, 0.00000175, -0.00001024, 0.00001094, -0.00000982, 0.00001365, -0.00000845, 0.00000686, -0.00000581, -0.00000081, -0.00000124, -0.00000605, -0.00000539, 0.00002089, -0.00000062, -0.00001726, 0.00000862, -0.00000541, 0.00000503, -0.00000080, 0.00000527, -0.00000607, 0.00000438, -0.00000289, 0.00000054, -0.00000845, 0.00000023, 0.00000919, 0.00000122, -0.00000512, 0.00000534, -0.00001002, 0.00000336, 0.00000418, -0.00000207, 0.00000233, -0.00000258, -0.00000450, -0.00000318, 0.00001230, -0.00000446, 0.00000112, },
		{ 0.00000064, -0.00000429, 0.00001019, 0.00000068, -0.00000666, -0.00000132, -0.00000061, 0.00000180, -0.00000050, 0.00000847, -0.00001318, 0.00000704, -0.00000686, 0.00000309, 0.00000561, 0.00000541, -0.00001238, 0.00000381, -0.00000684, 0.00000828, -0.00000972, 0.00000862, -0.00000403, 0.00000849, -0.00000903, 0.00001204, -0.00001820, -0.00000129, 0.00001714, 0.00000256, -0.00001225, 0.00000343, -0.00000649, 0.00000090, -0.00000037, -0.00000106, 0.00000457, 0.00000084, -0.00000064, 0.00000238, -0.00001071, -0.00000588, 0.00001617, 0.00000490, -0.00003561, 0.00003566, -0.00003921, 0.00004903, -0.00004954, 0.00004423, -0.00004619, 0.00004578, -0.00004610, 0.00002813, -0.00002772, 0.00002049, -0.00003767, 0.00003811, -0.00007235, 0.00006928, -0.00015295, 0.00013140, -0.00069818, 0.00133281, -0.00034466, -0.00016235, 0.00003409, -0.00007946, 0.00004396, -0.00005872, 0.00004278, -0.00004663, 0.00002782, -0.00001888, -0.00000312, 0.00000175, -0.00000385, 0.00000183, -0.00000696, 0.00000146, 0.00001082, -0.00001565, 0.00001733, -0.00004312, 0.00005069, -0.00002931, 0.00001265, -0.00002148, 0.00002114, -0.00001906, 0.00002135, -0.00001475, 0.00001324, -0.00001157, 0.00000398, -0.00000551, 0.00000001, -0.00001248, 0.00002333, 0.00000040, -0.00001558, 0.00000429, -0.00000100, 0.00000075, 0.00000337, 0.00000097, -0.00000153, -0.00000033, 0.00000175, -0.00000346, -0.00000375, -0.00000534, 0.00001280, -0.00000076, -0.00000298, 0.00000288, -0.00000592, -0.00000225, 0.00000875, -0.00000542, 0.00000455, -0.00000342, -0.00000241, -0.00000655, 0.00001344, -0.00000383, 0.00000101, -0.00000445, },
		{ -0.00000346, 0.00000742, 0.00000463, -0.00000856, 0.00000006, -0.00000382, 0.00000557, -0.00000462, 0.00001243, -0.00001492, 0.00000763, -0.00000787, 0.00000444, 0.00000257, 0.00000962, -0.00001504, 0.00000590, -0.00000963, 0.00001095, -0.00001210, 0.00001067, -0.00000605, 0.00001088, -0.00001163, 0.00001437, -0.00001836, -0.00000168, 0.00001247, 0.00001068, -0.00001788, 0.00000785, -0.00001232, 0.00000765, -0.00000748, 0.00000637, -0.00000474, 0.00001159, -0.00001212, 0.00001477, -0.00002301, 0.00000732, -0.00000292, 0.00002912, -0.00005947, 0.00005620, -0.00005759, 0.00006536, -0.00006340, 0.00005838, -0.00006046, 0.00006091, -0.00006088, 0.00004541, -0.00004925, 0.00004709, -0.00006968, 0.00007639, -0.00011752, 0.00012546, -0.00021969, 0.00022882, -0.00077718, 0.00121139, -0.00009122, -0.00033240, 0.00013249, -0.00015453, 0.00010071, -0.00010520, 0.00008086, -0.00007812, 0.00005393, -0.00004160, 0.00001867, -0.00002106, 0.00001837, -0.00002068, 0.00001591, -0.00002353, 0.00003575, -0.00003942, 0.00003962, -0.00006102, 0.00006144, -0.00003588, 0.00002055, -0.00003001, 0.00002870, -0.00002575, 0.00002669, -0.00001912, 0.00001794, -0.00001565, 0.00000729, -0.00000886, 0.00000531, -0.00001814, 0.00002420, 0.00000179, -0.00001326, -0.00000067, 0.00000385, -0.00000396, 0.00000801, -0.00000377, 0.00000333, -0.00000526, 0.00000652, -0.00000760, 0.00000106, -0.00001065, 0.00001586, -0.00000240, -0.00000082, 0.00000010, -0.00000148, -0.00000764, 0.00001249, -0.00000770, 0.00000569, -0.00000362, -0.00000067, -0.00000912, 0.00001344, -0.00000237, 0.00000033, -0.00000361, -0.00000035, },
		{ 0.00000439, 0.00000822, -0.00000986, 0.00000121, -0.00000682, 0.00000864, -0.00000756, 0.00001474, -0.00001492, 0.00000695, -0.00000796, 0.00000500, 0.00000044, 0.00001246, -0.00001611, 0.00000654, -0.00001104, 0.00001214, -0.00001298, 0.00001125, -0.00000674, 0.00001213, -0.00001312, 0.00001554, -0.00001770, -0.00000194, 0.00000738, 0.00001820, -0.00002224, 0.00001137, -0.00001750, 0.00001348, -0.00001346, 0.00001279, -0.00001306, 0.00002105, -0.00002201, 0.00002532, -0.00003340, 0.00001882, -0.00001989, 0.00004950, -0.00007740, 0.00006935, -0.00006804, 0.00007366, -0.00006953, 0.00006513, -0.00006731, 0.00006850, -0.00006801, 0.00005504, -0.00006257, 0.00006441, -0.00009092, 0.00010184, -0.00014695, 0.00016123, -0.00025894, 0.00028349, -0.00078811, 0.00102129, 0.00018982, -0.00049513, 0.00021713, -0.00021723, 0.00014633, -0.00014203, 0.00011041, -0.00010209, 0.00007357, -0.00005905, 0.00003624, -0.00004015, 0.00003717, -0.00004000, 0.00003568, -0.00004498, 0.00005615, -0.00005759, 0.00005527, -0.00007178, 0.00006566, -0.00003743, 0.00002454, -0.00003492, 0.00003280, -0.00002916, 0.00002906, -0.00002100, 0.00002041, -0.00001755, 0.00000875, -0.00001095, 0.00000943, -0.00002198, 0.00002340, 0.00000352, -0.00001045, -0.00000592, 0.00000868, -0.00000861, 0.00001258, -0.00000830, 0.00000784, -0.00000976, 0.00001077, -0.00001122, 0.00000532, -0.00001503, 0.00001783, -0.00000328, 0.00000094, -0.00000252, 0.00000268, -0.00001211, 0.00001483, -0.00000852, 0.00000557, -0.00000308, 0.00000056, -0.00001072, 0.00001227, -0.00000019, -0.00000076, -0.00000239, -0.00000165, -0.00000225, },
		{ 0.00001106, -0.00001027, 0.00000187, -0.00000920, 0.00001053, -0.00000890, 0.00001514, -0.00001314, 0.00000503, -0.00000713, 0.00000464, -0.00000058, 0.00001377, -0.00001549, 0.00000573, -0.00001104, 0.00001187, -0.00001241, 0.00001044, -0.00000620, 0.00001229, -0.00001351, 0.00001558, -0.00001626, -0.00000222, 0.00000239, 0.00002440, -0.00002470, 0.00001343, -0.00002133, 0.00001755, -0.00001744, 0.00001728, -0.00001933, 0.00002802, -0.00002899, 0.00003265, -0.00004037, 0.00002685, -0.00003259, 0.00006369, -0.00008715, 0.00007332, -0.00006928, 0.00007299, -0.00006720, 0.00006373, -0.00006594, 0.00006765, -0.00006650, 0.00005579, -0.00006623, 0.00007076, -0.00009946, 0.00011225, -0.00015816, 0.00017397, -0.00026826, 0.00029470, -0.00073707, 0.00077946, 0.00047623, -0.00063228, 0.00027581, -0.00025903, 0.00017436, -0.00016403, 0.00012712, -0.00011489, 0.00008359, -0.00006839, 0.00004692, -0.00005295, 0.00005001, -0.00005359, 0.00004975, -0.00006021, 0.00006942, -0.00006785, 0.00006249, -0.00007430, 0.00006283, -0.00003369, 0.00002432, -0.00003586, 0.00003307, -0.00002900, 0.00002825, -0.00002019, 0.00002043, -0.00001708, 0.00000825, -0.00001163, 0.00001209, -0.00002383, 0.00002099, 0.00000550, -0.00000732, -0.00001104, 0.00001300, -0.00001264, 0.00001646, -0.00001198, 0.00001136, -0.00001320, 0.00001388, -0.00001375, 0.00000842, -0.00001794, 0.00001829, -0.00000308, 0.00000200, -0.00000459, 0.00000604, -0.00001514, 0.00001541, -0.00000772, 0.00000419, -0.00000186, 0.00000126, -0.00001131, 0.00001006, 0.00000245, -0.00000202, -0.00000103, -0.00000301, -0.00000097, 0.00000149, },
		{ -0.00000961, 0.00000186, -0.00001067, 0.00001096, -0.00000849, 0.00001365, -0.00000979, 0.00000213, -0.00000551, 0.00000342, -0.00000050, 0.00001359, -0.00001335, 0.00000366, -0.00000983, 0.00001031, -0.00001058, 0.00000847, -0.00000468, 0.00001156, -0.00001294, 0.00001457, -0.00001408, -0.00000268, -0.00000199, 0.00002867, -0.00002482, 0.00001363, -0.00002326, 0.00001922, -0.00001881, 0.00001918, -0.00002278, 0.00003162, -0.00003218, 0.00003583, -0.00004292, 0.00003020, -0.00003960, 0.00007022, -0.00008764, 0.00006760, -0.00006131, 0.00006372, -0.00005688, 0.00005463, -0.00005673, 0.00005868, -0.00005660, 0.00004784, -0.00006023, 0.00006605, -0.00009522, 0.00010759, -0.00015125, 0.00016421, -0.00024912, 0.00026670, -0.00063565, 0.00050693, 0.00074478, -0.00072716, 0.00029882, -0.00027379, 0.00018042, -0.00016790, 0.00012835, -0.00011432, 0.00008213, -0.00006797, 0.00004910, -0.00005780, 0.00005526, -0.00005980, 0.00005642, -0.00006752, 0.00007403, -0.00006905, 0.00006072, -0.00006861, 0.00005339, -0.00002519, 0.00002025, -0.00003303, 0.00002971, -0.00002548, 0.00002447, -0.00001687, 0.00001812, -0.00001439, 0.00000597, -0.00001099, 0.00001326, -0.00002369, 0.00001718, 0.00000760, -0.00000408, -0.00001560, 0.00001631, -0.00001555, 0.00001910, -0.00001425, 0.00001336, -0.00001509, 0.00001541, -0.00001474, 0.00000992, -0.00001901, 0.00001700, -0.00000163, 0.00000217, -0.00000580, 0.00000820, -0.00001641, 0.00001409, -0.00000537, 0.00000177, -0.00000016, 0.00000153, -0.00001101, 0.00000709, 0.00000523, -0.00000317, 0.00000020, -0.00000415, 0.00000010, -0.00000092, 0.00001284, },
		{ 0.00000114, -0.00001111, 0.00000984, -0.00000642, 0.00001057, -0.00000529, -0.00000135, -0.00000341, 0.00000154, 0.00000051, 0.00001216, -0.00001003, 0.00000071, -0.00000777, 0.00000784, -0.00000787, 0.00000570, -0.00000250, 0.00001020, -0.00001158, 0.00001267, -0.00001127, -0.00000346, -0.00000534, 0.00003060, -0.00002240, 0.00001181, -0.00002302, 0.00001819, -0.00001729, 0.00001822, -0.00002305, 0.00003147, -0.00003118, 0.00003449, -0.00004066, 0.00002842, -0.00004035, 0.00006875, -0.00007903, 0.00005302, -0.00004548, 0.00004740, -0.00004017, 0.00003935, -0.00004111, 0.00004303, -0.00003975, 0.00003258, -0.00004599, 0.00005180, -0.00007987, 0.00008982, -0.00012866, 0.00013527, -0.00020643, 0.00020778, -0.00049927, 0.00022655, 0.00097363, -0.00076662, 0.00028027, -0.00025870, 0.00016298, -0.00015274, 0.00011350, -0.00010001, 0.00006897, -0.00005758, 0.00004247, -0.00005432, 0.00005244, -0.00005810, 0.00005512, -0.00006637, 0.00006969, -0.00006137, 0.00005067, -0.00005588, 0.00003871, -0.00001316, 0.00001326, -0.00002718, 0.00002337, -0.00001923, 0.00001832, -0.00001155, 0.00001390, -0.00000989, 0.00000231, -0.00000929, 0.00001306, -0.00002179, 0.00001229, 0.00000966, -0.00000092, -0.00001919, 0.00001820, -0.00001693, 0.00002008, -0.00001470, 0.00001349, -0.00001514, 0.00001511, -0.00001396, 0.00000964, -0.00001811, 0.00001398, 0.00000099, 0.00000143, -0.00000605, 0.00000901, -0.00001585, 0.00001103, -0.00000181, -0.00000130, 0.00000169, 0.00000156, -0.00001004, 0.00000370, 0.00000776, -0.00000391, 0.00000108, -0.00000486, 0.00000069, -0.00000256, 0.00001340, -0.00000788, },
		{ -0.00001057, 0.00000734, -0.00000302, 0.00000640, -0.00000025, -0.00000486, -0.00000121, -0.00000068, 0.00000211, 0.00000990, -0.00000604, -0.00000260, -0.00000532, 0.00000487, -0.00000471, 0.00000257, -0.00000006, 0.00000850, -0.00000965, 0.00001006, -0.00000792, -0.00000465, -0.00000738, 0.00003002, -0.00001752, 0.00000812, -0.00002063, 0.00001449, -0.00001303, 0.00001456, -0.00002026, 0.00002770, -0.00002623, 0.00002892, -0.00003391, 0.00002180, -0.00003522, 0.00006000, -0.00006268, 0.00003161, -0.00002419, 0.00002655, -0.00001954, 0.00002024, -0.00002139, 0.00002301, -0.00001832, 0.00001245, -0.00002600, 0.00003074, -0.00005651, 0.00006251, -0.00009470, 0.00009261, -0.00014753, 0.00012883, -0.00034525, -0.00003946, 0.00114464, -0.00074271, 0.00021913, -0.00021484, 0.00012362, -0.00012022, 0.00008419, -0.00007352, 0.00004556, -0.00003850, 0.00002807, -0.00004333, 0.00004227, -0.00004915, 0.00004643, -0.00005745, 0.00005738, -0.00004619, 0.00003415, -0.00003820, 0.00002086, 0.00000065, 0.00000470, -0.00001938, 0.00001504, -0.00001117, 0.00001064, -0.00000497, 0.00000841, -0.00000420, -0.00000216, -0.00000695, 0.00001176, -0.00001845, 0.00000674, 0.00001153, 0.00000196, -0.00002149, 0.00001839, -0.00001653, 0.00001918, -0.00001315, 0.00001166, -0.00001332, 0.00001297, -0.00001147, 0.00000763, -0.00001541, 0.00000948, 0.00000453, -0.00000003, -0.00000541, 0.00000852, -0.00001364, 0.00000662, 0.00000246, -0.00000450, 0.00000331, 0.00000158, -0.00000867, 0.00000027, 0.00000971, -0.00000402, 0.00000143, -0.00000499, 0.00000067, -0.00000334, 0.00001274, -0.00000520, -0.00000018, },
		{ 0.00000380, 0.00000117, 0.00000181, 0.00000465, -0.00000786, 0.00000069, -0.00000289, 0.00000393, 0.00000727, -0.00000190, -0.00000578, -0.00000292, 0.00000186, -0.00000154, -0.00000051, 0.00000230, 0.00000675, -0.00000738, 0.00000695, -0.00000420, -0.00000626, -0.00000801, 0.00002703, -0.00001060, 0.00000296, -0.00001643, 0.00000853, -0.00000657, 0.00000874, -0.00001494, 0.00002093, -0.00001806, 0.00001999, -0.00002363, 0.00001139, -0.00002545, 0.00004565, -0.00004097, 0.00000629, -0.00000059, 0.00000424, 0.00000210, 0.00000007, -0.00000029, 0.00000142, 0.00000480, -0.00000956, -0.00000346, 0.00000638, -0.00002909, 0.00003020, -0.00005474, 0.00004293, -0.00008107, 0.00004181, -0.00019070, -0.00027169, 0.00124520, -0.00065372, 0.00011954, -0.00014718, 0.00006691, -0.00007442, 0.00004405, -0.00003811, 0.00001486, -0.00001333, 0.00000814, -0.00002681, 0.00002658, -0.00003463, 0.00003203, -0.00004251, 0.00003913, -0.00002592, 0.00001383, -0.00001826, 0.00000225, 0.00001429, -0.00000396, -0.00001087, 0.00000587, -0.00000239, 0.00000245, 0.00000202, 0.00000241, 0.00000197, -0.00000680, -0.00000442, 0.00000972, -0.00001412, 0.00000095, 0.00001304, 0.00000440, -0.00002230, 0.00001677, -0.00001432, 0.00001643, -0.00000972, 0.00000805, -0.00000986, 0.00000929, -0.00000754, 0.00000423, -0.00001130, 0.00000397, 0.00000855, -0.00000189, -0.00000410, 0.00000696, -0.00001020, 0.00000145, 0.00000679, -0.00000730, 0.00000432, 0.00000183, -0.00000717, -0.00000283, 0.00001081, -0.00000337, 0.00000120, -0.00000450, 0.00000004, -0.00000327, 0.00001102, -0.00000186, -0.00000184, -0.00000925, },
		{ 0.00000553, -0.00000251, 0.00000877, -0.00000988, 0.00000194, -0.00000478, 0.00000558, 0.00000473, 0.00000186, -0.00000835, -0.00000099, -0.00000080, 0.00000125, -0.00000319, 0.00000428, 0.00000514, -0.00000493, 0.00000353, -0.00000030, -0.00000820, -0.00000732, 0.00002202, -0.00000228, -0.00000300, -0.00001101, 0.00000100, 0.00000126, 0.00000167, -0.00000803, 0.00001219, -0.00000789, 0.00000904, -0.00001129, -0.00000123, -0.00001286, 0.00002805, -0.00001682, -0.00001962, 0.00002198, -0.00001638, 0.00002183, -0.00001840, 0.00001943, -0.00001891, 0.00002663, -0.00003034, 0.00001830, -0.00001759, -0.00000175, -0.00000233, -0.00001445, -0.00000690, -0.00001568, -0.00004184, -0.00005071, -0.00045535, 0.00126934, -0.00050447, -0.00000948, -0.00006404, -0.00000016, -0.00002127, -0.00000182, 0.00000169, -0.00001913, 0.00001442, -0.00001426, -0.00000752, 0.00000793, -0.00001701, 0.00001432, -0.00002406, 0.00001770, -0.00000354, -0.00000722, 0.00000108, -0.00001466, 0.00002590, -0.00001131, -0.00000286, -0.00000300, 0.00000603, -0.00000530, 0.00000855, -0.00000333, 0.00000788, -0.00001098, -0.00000214, 0.00000730, -0.00000924, -0.00000462, 0.00001407, 0.00000629, -0.00002155, 0.00001344, -0.00001051, 0.00001208, -0.00000476, 0.00000312, -0.00000526, 0.00000455, -0.00000271, -0.00000004, -0.00000638, -0.00000191, 0.00001250, -0.00000375, -0.00000246, 0.00000472, -0.00000606, -0.00000382, 0.00001056, -0.00000919, 0.00000445, 0.00000247, -0.00000576, -0.00000535, 0.00001091, -0.00000197, 0.00000043, -0.00000349, -0.00000109, -0.00000255, 0.00000858, 0.00000173, -0.00000353, -0.00000748, -0.00000032, },
		{ -0.00000594, 0.00001158, -0.00001056, 0.00000234, -0.00000610, 0.00000676, 0.00000269, 0.00000484, -0.00000997, 0.00000020, -0.00000285, 0.00000340, -0.00000522, 0.00000568, 0.00000383, -0.00000248, 0.00000001, 0.00000358, -0.00001033, -0.00000555, 0.00001558, 0.00000658, -0.00000896, -0.00000513, -0.00000717, 0.00000940, -0.00000558, -0.00000063, 0.00000276, 0.00000287, -0.00000237, 0.00000142, -0.00001419, 0.00000040, 0.00000983, 0.00000664, -0.00004280, 0.00004043, -0.00003257, 0.00003719, -0.00003283, 0.00003542, -0.00003549, 0.00004456, -0.00004713, 0.00003629, -0.00003782, 0.00002171, -0.00003076, 0.00002111, -0.00005071, 0.00004108, -0.00011257, 0.00006312, -0.00058148, 0.00121823, -0.00030590, -0.00015494, 0.00002395, -0.00006914, 0.00003227, -0.00004750, 0.00004075, -0.00005191, 0.00004082, -0.00003569, 0.00001142, -0.00001075, 0.00000090, -0.00000393, -0.00000497, -0.00000387, 0.00001781, -0.00002602, 0.00001718, -0.00002783, 0.00003401, -0.00001629, 0.00000371, -0.00001061, 0.00001317, -0.00001177, 0.00001391, -0.00000817, 0.00001291, -0.00001417, -0.00000051, 0.00000486, -0.00000427, -0.00000961, 0.00001452, 0.00000757, -0.00001934, 0.00000870, -0.00000552, 0.00000665, 0.00000115, -0.00000250, -0.00000017, -0.00000057, 0.00000238, -0.00000452, -0.00000133, -0.00000749, 0.00001581, -0.00000517, -0.00000087, 0.00000225, -0.00000181, -0.00000853, 0.00001320, -0.00000981, 0.00000353, 0.00000356, -0.00000456, -0.00000712, 0.00001000, 0.00000004, -0.00000068, -0.00000214, -0.00000250, -0.00000143, 0.00000579, 0.00000514, -0.00000489, -0.00000561, -0.00000450, 0.00000942, },
		{ 0.00001275, -0.00000980, 0.00000182, -0.00000674, 0.00000726, 0.00000138, 0.00000679, -0.00001048, 0.00000053, -0.00000416, 0.00000479, -0.00000649, 0.00000643, 0.00000288, -0.00000012, -0.00000340, 0.00000720, -0.00001239, -0.00000311, 0.00000845, 0.00001506, -0.00001409, 0.00000040, -0.00001504, 0.00001675, -0.00001190, 0.00000611, -0.00000608, 0.00001274, -0.00001265, 0.00001279, -0.00002566, 0.00001221, -0.00000650, 0.00002661, -0.00006052, 0.00005238, -0.00004240, 0.00004648, -0.00004165, 0.00004605, -0.00004657, 0.00005672, -0.00005790, 0.00004829, -0.00005180, 0.00003844, -0.00005179, 0.00004802, -0.00008387, 0.00008366, -0.00016368, 0.00014360, -0.00064733, 0.00109968, -0.00007385, -0.00030139, 0.00010521, -0.00013127, 0.00007920, -0.00008721, 0.00007410, -0.00007917, 0.00006216, -0.00005290, 0.00002705, -0.00002665, 0.00001641, -0.00002004, 0.00001203, -0.00002273, 0.00003535, -0.00004006, 0.00002808, -0.00003586, 0.00003775, -0.00001830, 0.00000822, -0.00001631, 0.00001840, -0.00001637, 0.00001761, -0.00001163, 0.00001657, -0.00001598, 0.00000023, 0.00000266, 0.00000042, -0.00001368, 0.00001430, 0.00000828, -0.00001592, 0.00000304, 0.00000008, 0.00000076, 0.00000730, -0.00000804, 0.00000466, -0.00000538, 0.00000705, -0.00000854, 0.00000317, -0.00001212, 0.00001798, -0.00000577, 0.00000031, -0.00000002, 0.00000198, -0.00001212, 0.00001432, -0.00000900, 0.00000159, 0.00000501, -0.00000359, -0.00000814, 0.00000824, 0.00000241, -0.00000189, -0.00000071, -0.00000393, -0.00000025, 0.00000307, 0.00000796, -0.00000560, -0.00000396, -0.00000827, 0.00001230, -0.00000798, },
		{ -0.00000767, 0.00000052, -0.00000672, 0.00000704, 0.00000092, 0.00000764, -0.00000989, 0.00000005, -0.00000475, 0.00000541, -0.00000702, 0.00000656, 0.00000228, 0.00000206, -0.00000653, 0.00001032, -0.00001413, -0.00000046, 0.00000144, 0.00002224, -0.00001762, 0.00000485, -0.00002168, 0.00002233, -0.00001634, 0.00001117, -0.00001316, 0.00002045, -0.00002045, 0.00002140, -0.00003408, 0.00002083, -0.00001894, 0.00004098, -0.00007092, 0.00005650, -0.00004497, 0.00004906, -0.00004426, 0.00005060, -0.00005131, 0.00006213, -0.00006160, 0.00005312, -0.00005813, 0.00004681, -0.00006351, 0.00006404, -0.00010368, 0.00010898, -0.00019179, 0.00018823, -0.00065610, 0.00092708, 0.00017265, -0.00043269, 0.00016875, -0.00017872, 0.00011349, -0.00011603, 0.00009759, -0.00009740, 0.00007548, -0.00006331, 0.00003700, -0.00003752, 0.00002732, -0.00003184, 0.00002470, -0.00003665, 0.00004699, -0.00004769, 0.00003266, -0.00003821, 0.00003698, -0.00001730, 0.00001052, -0.00001981, 0.00002139, -0.00001883, 0.00001938, -0.00001345, 0.00001862, -0.00001622, -0.00000003, 0.00000088, 0.00000451, -0.00001661, 0.00001339, 0.00000847, -0.00001164, -0.00000298, 0.00000563, -0.00000486, 0.00001293, -0.00001275, 0.00000853, -0.00000921, 0.00001067, -0.00001152, 0.00000654, -0.00001529, 0.00001864, -0.00000532, 0.00000084, -0.00000174, 0.00000490, -0.00001421, 0.00001375, -0.00000681, -0.00000116, 0.00000665, -0.00000278, -0.00000849, 0.00000586, 0.00000480, -0.00000290, 0.00000054, -0.00000509, 0.00000068, 0.00000080, 0.00000985, -0.00000545, -0.00000277, -0.00001126, 0.00001381, -0.00000840, 0.00001223, },
		{ -0.00000130, -0.00000621, 0.00000615, 0.00000125, 0.00000749, -0.00000837, -0.00000105, -0.00000477, 0.00000538, -0.00000691, 0.00000617, 0.00000198, 0.00000401, -0.00000920, 0.00001270, -0.00001526, 0.00000190, -0.00000469, 0.00002736, -0.00001899, 0.00000765, -0.00002640, 0.00002538, -0.00001819, 0.00001384, -0.00001768, 0.00002510, -0.00002485, 0.00002630, -0.00003844, 0.00002513, -0.00002621, 0.00004856, -0.00007324, 0.00005266, -0.00004057, 0.00004533, -0.00004100, 0.00004930, -0.00004980, 0.00006085, -0.00005825, 0.00005070, -0.00005663, 0.00004654, -0.00006553, 0.00006865, -0.00010957, 0.00011658, -0.00019688, 0.00019885, -0.00061594, 0.00071765, 0.00041346, -0.00053393, 0.00020565, -0.00020568, 0.00013092, -0.00013070, 0.00010856, -0.00010441, 0.00007899, -0.00006542, 0.00003986, -0.00004198, 0.00003228, -0.00003797, 0.00003165, -0.00004426, 0.00005166, -0.00004831, 0.00003087, -0.00003524, 0.00003227, -0.00001379, 0.00001084, -0.00002115, 0.00002215, -0.00001915, 0.00001924, -0.00001362, 0.00001899, -0.00001491, -0.00000124, -0.00000042, 0.00000782, -0.00001828, 0.00001181, 0.00000828, -0.00000696, -0.00000871, 0.00001044, -0.00000954, 0.00001735, -0.00001599, 0.00001088, -0.00001158, 0.00001281, -0.00001303, 0.00000840, -0.00001671, 0.00001766, -0.00000373, 0.00000062, -0.00000272, 0.00000668, -0.00001468, 0.00001158, -0.00000352, -0.00000438, 0.00000817, -0.00000197, -0.00000834, 0.00000319, 0.00000688, -0.00000346, 0.00000138, -0.00000576, 0.00000111, -0.00000075, 0.00001062, -0.00000439, -0.00000215, -0.00001326, 0.00001377, -0.00000718, 0.00001018, -0.00000450, },
		{ -0.00000545, 0.00000479, 0.00000220, 0.00000656, -0.00000623, -0.00000248, -0.00000445, 0.00000490, -0.00000635, 0.00000541, 0.00000189, 0.00000569, -0.00001126, 0.00001417, -0.00001553, 0.00000355, -0.00000934, 0.00002993, -0.00001792, 0.00000853, -0.00002879, 0.00002552, -0.00001716, 0.00001383, -0.00001923, 0.00002629, -0.00002547, 0.00002712, -0.00003839, 0.00002470, -0.00002792, 0.00004923, -0.00006792, 0.00004188, -0.00003051, 0.00003662, -0.00003310, 0.00004321, -0.00004300, 0.00005381, -0.00004879, 0.00004198, -0.00004822, 0.00003862, -0.00005890, 0.00006298, -0.00010288, 0.00010831, -0.00018181, 0.00018086, -0.00053860, 0.00049045, 0.00062946, -0.00059319, 0.00021025, -0.00020922, 0.00012969, -0.00012998, 0.00010612, -0.00009963, 0.00007235, -0.00005899, 0.00003537, -0.00003971, 0.00003095, -0.00003804, 0.00003247, -0.00004527, 0.00004938, -0.00004238, 0.00002359, -0.00002808, 0.00002475, -0.00000865, 0.00000972, -0.00002068, 0.00002096, -0.00001760, 0.00001744, -0.00001230, 0.00001784, -0.00001223, -0.00000323, -0.00000125, 0.00001022, -0.00001866, 0.00000960, 0.00000786, -0.00000233, -0.00001355, 0.00001394, -0.00001273, 0.00002004, -0.00001731, 0.00001137, -0.00001222, 0.00001324, -0.00001291, 0.00000860, -0.00001634, 0.00001512, -0.00000115, -0.00000031, -0.00000291, 0.00000727, -0.00001359, 0.00000811, 0.00000045, -0.00000759, 0.00000926, -0.00000099, -0.00000792, 0.00000055, 0.00000834, -0.00000335, 0.00000165, -0.00000581, 0.00000092, -0.00000147, 0.00001027, -0.00000248, -0.00000207, -0.00001423, 0.00001225, -0.00000458, 0.00000700, -0.00000074, -0.00000326, },
		{ 0.00000318, 0.00000350, 0.00000519, -0.00000382, -0.00000393, -0.00000402, 0.00000417, -0.00000552, 0.00000444, 0.00000194, 0.00000705, -0.00001257, 0.00001456, -0.00001478, 0.00000418, -0.00001211, 0.00002972, -0.00001445, 0.00000752, -0.00002877, 0.00002274, -0.00001337, 0.00001125, -0.00001791, 0.00002413, -0.00002251, 0.00002413, -0.00003425, 0.00001993, -0.00002458, 0.00004384, -0.00005640, 0.00002609, -0.00001688, 0.00002489, -0.00002235, 0.00003396, -0.00003248, 0.00004259, -0.00003489, 0.00002871, -0.00003473, 0.00002498, -0.00004575, 0.00004942, -0.00008644, 0.00008781, -0.00015153, 0.00014204, -0.00043766, 0.00026432, 0.00080454, -0.00060292, 0.00018100, -0.00018965, 0.00011060, -0.00011484, 0.00009125, -0.00008410, 0.00005666, -0.00004504, 0.00002442, -0.00003148, 0.00002400, -0.00003266, 0.00002773, -0.00004037, 0.00004115, -0.00003132, 0.00001251, -0.00001845, 0.00001595, -0.00000301, 0.00000788, -0.00001891, 0.00001829, -0.00001462, 0.00001437, -0.00000983, 0.00001544, -0.00000849, -0.00000573, -0.00000172, 0.00001171, -0.00001781, 0.00000687, 0.00000734, 0.00000181, -0.00001702, 0.00001573, -0.00001409, 0.00002075, -0.00001654, 0.00000994, -0.00001115, 0.00001202, -0.00001122, 0.00000727, -0.00001438, 0.00001131, 0.00000214, -0.00000174, -0.00000242, 0.00000680, -0.00001125, 0.00000381, 0.00000454, -0.00001034, 0.00000960, 0.00000033, -0.00000741, -0.00000177, 0.00000896, -0.00000248, 0.00000132, -0.00000521, 0.00000010, -0.00000139, 0.00000894, 0.00000002, -0.00000236, -0.00001429, 0.00000949, -0.00000101, 0.00000325, 0.00000305, -0.00000493, -0.00000469, },
		{ 0.00000489, 0.00000369, -0.00000151, -0.00000509, -0.00000372, 0.00000338, -0.00000459, 0.00000339, 0.00000206, 0.00000805, -0.00001307, 0.00001384, -0.00001296, 0.00000365, -0.00001285, 0.00002687, -0.00000894, 0.00000494, -0.00002660, 0.00001741, -0.00000734, 0.00000664, -0.00001424, 0.00001922, -0.00001670, 0.00001815, -0.00002692, 0.00001182, -0.00001740, 0.00003403, -0.00004086, 0.00000783, -0.00000217, 0.00001240, -0.00001078, 0.00002340, -0.00002007, 0.00002914, -0.00001862, 0.00001305, -0.00001848, 0.00000817, -0.00002886, 0.00003111, -0.00006398, 0.00005974, -0.00011216, 0.00009123, -0.00032687, 0.00005594, 0.00092709, -0.00056073, 0.00012061, -0.00015049, 0.00007694, -0.00008819, 0.00006662, -0.00006031, 0.00003425, -0.00002571, 0.00000887, -0.00001894, 0.00001294, -0.00002325, 0.00001883, -0.00003107, 0.00002878, -0.00001720, -0.00000022, -0.00000832, 0.00000745, 0.00000204, 0.00000603, -0.00001641, 0.00001466, -0.00001074, 0.00001051, -0.00000662, 0.00001219, -0.00000409, -0.00000841, -0.00000196, 0.00001234, -0.00001588, 0.00000375, 0.00000686, 0.00000512, -0.00001881, 0.00001562, -0.00001355, 0.00001945, -0.00001381, 0.00000683, -0.00000863, 0.00000943, -0.00000830, 0.00000474, -0.00001124, 0.00000672, 0.00000571, -0.00000335, -0.00000150, 0.00000555, -0.00000809, -0.00000075, 0.00000818, -0.00001218, 0.00000900, 0.00000207, -0.00000697, -0.00000355, 0.00000864, -0.00000091, 0.00000047, -0.00000406, -0.00000123, -0.00000067, 0.00000691, 0.00000278, -0.00000278, -0.00001366, 0.00000589, 0.00000298, -0.00000049, 0.00000632, -0.00000595, -0.00000418, 0.00000154, },
		{ 0.00000236, 0.00000041, -0.00000575, -0.00000367, 0.00000266, -0.00000368, 0.00000235, 0.00000223, 0.00000867, -0.00001270, 0.00001203, -0.00001014, 0.00000200, -0.00001169, 0.00002182, -0.00000207, 0.00000137, -0.00002281, 0.00001022, 0.00000009, 0.00000085, -0.00000905, 0.00001252, -0.00000912, 0.00001037, -0.00001771, 0.00000183, -0.00000807, 0.00002190, -0.00002383, -0.00001019, 0.00001112, 0.00000128, -0.00000028, 0.00001334, -0.00000759, 0.00001540, -0.00000208, -0.00000272, -0.00000188, -0.00000915, -0.00001120, 0.00001141, -0.00003946, 0.00002900, -0.00006998, 0.00003709, -0.00021856, -0.00012161, 0.00099095, -0.00046967, 0.00003581, -0.00009795, 0.00003388, -0.00005446, 0.00003606, -0.00003176, 0.00000830, -0.00000381, -0.00000880, -0.00000433, -0.00000014, -0.00001177, 0.00000770, -0.00001941, 0.00001451, -0.00000238, -0.00001237, 0.00000043, 0.00000066, 0.00000558, 0.00000474, -0.00001364, 0.00001061, -0.00000646, 0.00000635, -0.00000307, 0.00000847, 0.00000056, -0.00001094, -0.00000212, 0.00001221, -0.00001307, 0.00000041, 0.00000648, 0.00000739, -0.00001881, 0.00001369, -0.00001128, 0.00001644, -0.00000949, 0.00000250, -0.00000516, 0.00000597, -0.00000461, 0.00000151, -0.00000746, 0.00000190, 0.00000906, -0.00000480, -0.00000043, 0.00000387, -0.00000461, -0.00000498, 0.00001088, -0.00001281, 0.00000734, 0.00000421, -0.00000666, -0.00000471, 0.00000744, 0.00000120, -0.00000070, -0.00000259, -0.00000284, 0.00000040, 0.00000455, 0.00000541, -0.00000302, -0.00001263, 0.00000188, 0.00000681, -0.00000362, 0.00000862, -0.00000605, -0.00000408, 0.00000006, 0.00000612, },
		{ 0.00000173, -0.00000579, -0.00000394, 0.00000207, -0.00000287, 0.00000138, 0.00000244, 0.00000887, -0.00001151, 0.00000927, -0.00000652, -0.00000056, -0.00000901, 0.00001525, 0.00000535, -0.00000247, -0.00001809, 0.00000205, 0.00000792, -0.00000513, -0.00000338, 0.00000517, -0.00000104, 0.00000217, -0.00000812, -0.00000841, 0.00000156, 0.00000964, -0.00000778, -0.00002555, 0.00002095, -0.00000679, 0.00000767, 0.00000516, 0.00000346, 0.00000303, 0.00001289, -0.00001664, 0.00001291, -0.00002459, 0.00000459, -0.00000666, -0.00001648, 0.00000000, -0.00003054, -0.00001297, -0.00012253, -0.00025976, 0.00099569, -0.00033764, -0.00006360, -0.00004000, -0.00001228, -0.00001881, 0.00000402, -0.00000239, -0.00001765, 0.00001755, -0.00002588, 0.00000991, -0.00001298, -0.00000036, -0.00000357, -0.00000756, 0.00000065, 0.00001086, -0.00002194, 0.00000628, -0.00000343, 0.00000708, 0.00000431, -0.00001095, 0.00000654, -0.00000224, 0.00000229, 0.00000042, 0.00000467, 0.00000506, -0.00001304, -0.00000235, 0.00001147, -0.00000962, -0.00000294, 0.00000622, 0.00000853, -0.00001714, 0.00001024, -0.00000769, 0.00001221, -0.00000419, -0.00000240, -0.00000136, 0.00000222, -0.00000075, -0.00000184, -0.00000364, -0.00000256, 0.00001176, -0.00000576, 0.00000049, 0.00000214, -0.00000129, -0.00000840, 0.00001230, -0.00001210, 0.00000471, 0.00000664, -0.00000645, -0.00000527, 0.00000551, 0.00000358, -0.00000193, -0.00000103, -0.00000444, 0.00000152, 0.00000226, 0.00000752, -0.00000283, -0.00001149, -0.00000207, 0.00000998, -0.00000572, 0.00000965, -0.00000513, -0.00000443, -0.00000114, 0.00000701, 0.00000144, },
		{ -0.00000522, -0.00000448, 0.00000158, -0.00000215, 0.00000048, 0.00000271, 0.00000866, -0.00000958, 0.00000580, -0.00000242, -0.00000367, -0.00000533, 0.00000800, 0.00001241, -0.00000582, -0.00001320, -0.00000616, 0.00001509, -0.00001030, 0.00000176, -0.00000169, 0.00000628, -0.00000514, 0.00000046, -0.00001738, 0.00000979, -0.00000083, 0.00000524, -0.00003642, 0.00002599, -0.00001087, 0.00001228, -0.00000033, 0.00001212, -0.00000680, 0.00002496, -0.00002723, 0.00002429, -0.00003636, 0.00001649, -0.00002080, 0.00000222, -0.00002385, 0.00000198, -0.00005356, -0.00004535, -0.00035482, 0.00094619, -0.00017646, -0.00016594, 0.00001477, -0.00005511, 0.00001362, -0.00002521, 0.00002403, -0.00004031, 0.00003552, -0.00003989, 0.00002155, -0.00002350, 0.00000901, -0.00001306, 0.00000250, -0.00001082, 0.00002069, -0.00002752, 0.00000835, -0.00000442, 0.00000648, 0.00000472, -0.00000847, 0.00000271, 0.00000161, -0.00000136, 0.00000357, 0.00000111, 0.00000907, -0.00001447, -0.00000277, 0.00001025, -0.00000583, -0.00000609, 0.00000605, 0.00000862, -0.00001410, 0.00000577, -0.00000338, 0.00000739, 0.00000140, -0.00000716, 0.00000212, -0.00000121, 0.00000273, -0.00000476, -0.00000033, -0.00000617, 0.00001343, -0.00000599, 0.00000104, 0.00000066, 0.00000145, -0.00001065, 0.00001227, -0.00001010, 0.00000133, 0.00000912, -0.00000621, -0.00000536, 0.00000312, 0.00000588, -0.00000291, 0.00000031, -0.00000576, 0.00000236, 0.00000039, 0.00000883, -0.00000203, -0.00001046, -0.00000562, 0.00001210, -0.00000651, 0.00000934, -0.00000327, -0.00000514, -0.00000207, 0.00000748, 0.00000102, 0.00000239, },
		{ -0.00000517, 0.00000111, -0.00000148, -0.00000039, 0.00000304, 0.00000806, -0.00000707, 0.00000191, 0.00000177, -0.00000689, -0.00000128, 0.00000093, 0.00001824, -0.00000802, -0.00000881, -0.00001354, 0.00002072, -0.00001381, 0.00000552, -0.00000712, 0.00001184, -0.00001051, 0.00000692, -0.00002390, 0.00001533, -0.00000809, 0.00001389, -0.00004183, 0.00002581, -0.00001084, 0.00001347, -0.00000295, 0.00001804, -0.00001354, 0.00003342, -0.00003370, 0.00003138, -0.00004345, 0.00002338, -0.00002968, 0.00001499, -0.00004054, 0.00002508, -0.00008160, 0.00000989, -0.00040766, 0.00085168, -0.00000035, -0.00025917, 0.00005837, -0.00008902, 0.00003853, -0.00004808, 0.00004444, -0.00005705, 0.00004789, -0.00004895, 0.00002889, -0.00003014, 0.00001488, -0.00001933, 0.00000933, -0.00001848, 0.00002598, -0.00002846, 0.00000649, -0.00000253, 0.00000417, 0.00000567, -0.00000612, -0.00000074, 0.00000491, -0.00000440, 0.00000619, -0.00000197, 0.00001236, -0.00001507, -0.00000342, 0.00000872, -0.00000195, -0.00000882, 0.00000588, 0.00000785, -0.00001014, 0.00000090, 0.00000102, 0.00000268, 0.00000656, -0.00001110, 0.00000469, -0.00000383, 0.00000536, -0.00000683, 0.00000206, -0.00000860, 0.00001388, -0.00000538, 0.00000107, -0.00000034, 0.00000338, -0.00001157, 0.00001084, -0.00000708, -0.00000244, 0.00001135, -0.00000578, -0.00000515, 0.00000060, 0.00000778, -0.00000338, 0.00000123, -0.00000656, 0.00000270, -0.00000082, 0.00000917, -0.00000057, -0.00000965, -0.00000852, 0.00001296, -0.00000594, 0.00000780, -0.00000070, -0.00000602, -0.00000278, 0.00000753, 0.00000111, 0.00000241, -0.00000417, },
		{ 0.00000057, -0.00000080, -0.00000124, 0.00000346, 0.00000715, -0.00000419, -0.00000204, 0.00000565, -0.00000976, 0.00000250, -0.00000516, 0.00002218, -0.00000861, -0.00000543, -0.00001943, 0.00002411, -0.00001511, 0.00000736, -0.00001049, 0.00001497, -0.00001330, 0.00001062, -0.00002731, 0.00001748, -0.00001146, 0.00001768, -0.00004176, 0.00002087, -0.00000738, 0.00001183, -0.00000314, 0.00002142, -0.00001722, 0.00003820, -0.00003594, 0.00003403, -0.00004566, 0.00002500, -0.00003299, 0.00002135, -0.00004941, 0.00003805, -0.00009633, 0.00004322, -0.00042291, 0.00072432, 0.00017575, -0.00033240, 0.00008457, -0.00011004, 0.00005305, -0.00006230, 0.00005691, -0.00006629, 0.00005342, -0.00005205, 0.00003103, -0.00003207, 0.00001648, -0.00002164, 0.00001218, -0.00002169, 0.00002642, -0.00002490, 0.00000127, 0.00000148, 0.00000090, 0.00000663, -0.00000373, -0.00000381, 0.00000759, -0.00000675, 0.00000816, -0.00000443, 0.00001476, -0.00001478, -0.00000435, 0.00000701, 0.00000175, -0.00001094, 0.00000558, 0.00000650, -0.00000581, -0.00000373, 0.00000484, -0.00000130, 0.00001069, -0.00001368, 0.00000594, -0.00000528, 0.00000685, -0.00000780, 0.00000332, -0.00000970, 0.00001310, -0.00000397, 0.00000059, -0.00000079, 0.00000438, -0.00001121, 0.00000825, -0.00000341, -0.00000617, 0.00001301, -0.00000499, -0.00000484, -0.00000175, 0.00000897, -0.00000316, 0.00000157, -0.00000673, 0.00000240, -0.00000125, 0.00000853, 0.00000149, -0.00000907, -0.00001066, 0.00001256, -0.00000418, 0.00000535, 0.00000221, -0.00000680, -0.00000341, 0.00000723, 0.00000164, 0.00000195, -0.00000282, -0.00000581, },
		{ -0.00000008, -0.00000210, 0.00000393, 0.00000602, -0.00000119, -0.00000570, 0.00000881, -0.00001187, 0.00000549, -0.00000967, 0.00002381, -0.00000739, -0.00000331, -0.00002343, 0.00002496, -0.00001400, 0.00000710, -0.00001158, 0.00001548, -0.00001337, 0.00001145, -0.00002751, 0.00001619, -0.00001099, 0.00001695, -0.00003705, 0.00001242, -0.00000175, 0.00000845, -0.00000176, 0.00002291, -0.00001831, 0.00003973, -0.00003441, 0.00003275, -0.00004351, 0.00002194, -0.00003129, 0.00002189, -0.00005108, 0.00004170, -0.00009896, 0.00005725, -0.00040792, 0.00057767, 0.00033778, -0.00037715, 0.00008969, -0.00011641, 0.00005609, -0.00006710, 0.00006086, -0.00006764, 0.00005194, -0.00004914, 0.00002797, -0.00002929, 0.00001383, -0.00002000, 0.00001106, -0.00002063, 0.00002249, -0.00001776, -0.00000617, 0.00000643, -0.00000236, 0.00000702, -0.00000104, -0.00000656, 0.00000968, -0.00000841, 0.00000947, -0.00000620, 0.00001620, -0.00001362, -0.00000551, 0.00000525, 0.00000502, -0.00001229, 0.00000506, 0.00000487, -0.00000164, -0.00000754, 0.00000759, -0.00000409, 0.00001335, -0.00001461, 0.00000569, -0.00000548, 0.00000712, -0.00000760, 0.00000343, -0.00000954, 0.00001126, -0.00000195, -0.00000028, -0.00000072, 0.00000451, -0.00000976, 0.00000491, 0.00000040, -0.00000942, 0.00001382, -0.00000369, -0.00000463, -0.00000367, 0.00000928, -0.00000217, 0.00000130, -0.00000623, 0.00000148, -0.00000094, 0.00000705, 0.00000392, -0.00000860, -0.00001209, 0.00001103, -0.00000153, 0.00000242, 0.00000503, -0.00000720, -0.00000409, 0.00000670, 0.00000244, 0.00000121, -0.00000142, -0.00000623, -0.00000016, },
		{ -0.00000295, 0.00000441, 0.00000482, 0.00000169, -0.00000873, 0.00001096, -0.00001293, 0.00000731, -0.00001224, 0.00002303, -0.00000447, -0.00000244, -0.00002545, 0.00002328, -0.00001068, 0.00000494, -0.00001057, 0.00001361, -0.00001105, 0.00000977, -0.00002494, 0.00001200, -0.00000739, 0.00001274, -0.00002915, 0.00000221, 0.00000441, 0.00000465, 0.00000011, 0.00002334, -0.00001758, 0.00003879, -0.00002997, 0.00002848, -0.00003803, 0.00001534, -0.00002584, 0.00001794, -0.00004710, 0.00003798, -0.00009208, 0.00005627, -0.00037139, 0.00042509, 0.00047408, -0.00038836, 0.00007316, -0.00010871, 0.00004844, -0.00006325, 0.00005700, -0.00006189, 0.00004431, -0.00004108, 0.00002049, -0.00002255, 0.00000766, -0.00001510, 0.00000670, -0.00001613, 0.00001537, -0.00000848, -0.00001430, 0.00001099, -0.00000462, 0.00000627, 0.00000211, -0.00000903, 0.00001119, -0.00000943, 0.00001014, -0.00000725, 0.00001668, -0.00001166, -0.00000686, 0.00000353, 0.00000769, -0.00001275, 0.00000420, 0.00000326, 0.00000189, -0.00001008, 0.00000891, -0.00000541, 0.00001439, -0.00001382, 0.00000402, -0.00000453, 0.00000631, -0.00000641, 0.00000257, -0.00000838, 0.00000868, 0.00000037, -0.00000134, -0.00000028, 0.00000398, -0.00000757, 0.00000128, 0.00000388, -0.00001179, 0.00001358, -0.00000183, -0.00000461, -0.00000499, 0.00000864, -0.00000048, 0.00000053, -0.00000518, 0.00000004, -0.00000004, 0.00000499, 0.00000642, -0.00000806, -0.00001293, 0.00000867, 0.00000155, -0.00000050, 0.00000733, -0.00000697, -0.00000494, 0.00000604, 0.00000337, 0.00000039, -0.00000020, -0.00000626, -0.00000112, 0.00000071, },
		{ 0.00000482, 0.00000368, 0.00000420, -0.00001089, 0.00001190, -0.00001278, 0.00000780, -0.00001280, 0.00002005, -0.00000023, -0.00000255, -0.00002569, 0.00001943, -0.00000564, 0.00000140, -0.00000799, 0.00000996, -0.00000703, 0.00000635, -0.00002044, 0.00000589, -0.00000184, 0.00000653, -0.00001992, -0.00000786, 0.00000951, 0.00000168, 0.00000145, 0.00002358, -0.00001583, 0.00003627, -0.00002366, 0.00002239, -0.00003052, 0.00000665, -0.00001822, 0.00001125, -0.00003950, 0.00002943, -0.00007904, 0.00004531, -0.00032223, 0.00027836, 0.00057638, -0.00036480, 0.00003743, -0.00008972, 0.00003251, -0.00005280, 0.00004715, -0.00005078, 0.00003221, -0.00002945, 0.00001004, -0.00001320, -0.00000075, -0.00000817, 0.00000028, -0.00000954, 0.00000662, 0.00000124, -0.00002154, 0.00001392, -0.00000508, 0.00000406, 0.00000577, -0.00001122, 0.00001218, -0.00000986, 0.00001021, -0.00000762, 0.00001627, -0.00000903, -0.00000830, 0.00000193, 0.00000962, -0.00001230, 0.00000297, 0.00000192, 0.00000445, -0.00001111, 0.00000873, -0.00000527, 0.00001386, -0.00001152, 0.00000122, -0.00000277, 0.00000475, -0.00000453, 0.00000108, -0.00000659, 0.00000577, 0.00000265, -0.00000235, 0.00000034, 0.00000302, -0.00000504, -0.00000217, 0.00000659, -0.00001302, 0.00001223, 0.00000056, -0.00000480, -0.00000566, 0.00000712, 0.00000173, -0.00000056, -0.00000378, -0.00000168, 0.00000116, 0.00000270, 0.00000867, -0.00000725, -0.00001339, 0.00000582, 0.00000461, -0.00000295, 0.00000877, -0.00000596, -0.00000601, 0.00000535, 0.00000428, -0.00000030, 0.00000066, -0.00000583, -0.00000233, 0.00000153, -0.00000376, },
		{ 0.00000275, 0.00000617, -0.00001201, 0.00001158, -0.00001146, 0.00000700, -0.00001154, 0.00001537, 0.00000473, -0.00000320, -0.00002454, 0.00001399, 0.00000036, -0.00000278, -0.00000456, 0.00000537, -0.00000225, 0.00000218, -0.00001510, -0.00000094, 0.00000429, -0.00000004, -0.00001116, -0.00001606, 0.00001222, 0.00000049, 0.00000154, 0.00002425, -0.00001374, 0.00003299, -0.00001648, 0.00001565, -0.00002228, -0.00000267, -0.00001005, 0.00000361, -0.00003041, 0.00001869, -0.00006322, 0.00002934, -0.00026842, 0.00014673, 0.00064040, -0.00030911, -0.00001234, -0.00006387, 0.00001183, -0.00003865, 0.00003381, -0.00003660, 0.00001780, -0.00001622, -0.00000161, -0.00000288, -0.00000986, -0.00000067, -0.00000672, -0.00000242, -0.00000207, 0.00000971, -0.00002648, 0.00001428, -0.00000330, 0.00000033, 0.00000979, -0.00001302, 0.00001264, -0.00000975, 0.00000974, -0.00000735, 0.00001506, -0.00000591, -0.00000973, 0.00000049, 0.00001075, -0.00001098, 0.00000139, 0.00000100, 0.00000585, -0.00001062, 0.00000717, -0.00000389, 0.00001207, -0.00000811, -0.00000222, -0.00000066, 0.00000284, -0.00000239, -0.00000065, -0.00000460, 0.00000297, 0.00000454, -0.00000307, 0.00000092, 0.00000193, -0.00000256, -0.00000500, 0.00000823, -0.00001300, 0.00000986, 0.00000332, -0.00000514, -0.00000574, 0.00000492, 0.00000417, -0.00000170, -0.00000228, -0.00000342, 0.00000238, 0.00000053, 0.00001033, -0.00000597, -0.00001364, 0.00000281, 0.00000724, -0.00000457, 0.00000914, -0.00000416, -0.00000727, 0.00000466, 0.00000505, -0.00000074, 0.00000105, -0.00000491, -0.00000372, 0.00000231, -0.00000444, 0.00000507, },
		{ 0.00000747, -0.00001206, 0.00001010, -0.00000916, 0.00000516, -0.00000888, 0.00000966, 0.00000971, -0.00000386, -0.00002252, 0.00000770, 0.00000648, -0.00000677, -0.00000111, 0.00000074, 0.00000233, -0.00000173, -0.00000998, -0.00000730, 0.00000966, -0.00000548, -0.00000441, -0.00002112, 0.00001177, 0.00000151, 0.00000004, 0.00002565, -0.00001173, 0.00002955, -0.00000922, 0.00000921, -0.00001439, -0.00001138, -0.00000269, -0.00000342, -0.00002166, 0.00000807, -0.00004758, 0.00001254, -0.00021633, 0.00003628, 0.00066585, -0.00022723, -0.00006915, -0.00003652, -0.00000959, -0.00002398, 0.00001968, -0.00002182, 0.00000333, -0.00000340, -0.00001266, 0.00000676, -0.00001811, 0.00000594, -0.00001290, 0.00000377, -0.00000918, 0.00001557, -0.00002818, 0.00001162, 0.00000074, -0.00000462, 0.00001383, -0.00001422, 0.00001251, -0.00000911, 0.00000878, -0.00000653, 0.00001320, -0.00000249, -0.00001103, -0.00000077, 0.00001110, -0.00000891, -0.00000047, 0.00000054, 0.00000608, -0.00000880, 0.00000458, -0.00000168, 0.00000950, -0.00000416, -0.00000573, 0.00000130, 0.00000105, -0.00000039, -0.00000220, -0.00000282, 0.00000064, 0.00000579, -0.00000334, 0.00000131, 0.00000095, -0.00000045, -0.00000695, 0.00000867, -0.00001178, 0.00000669, 0.00000620, -0.00000549, -0.00000538, 0.00000232, 0.00000650, -0.00000259, -0.00000094, -0.00000490, 0.00000333, -0.00000119, 0.00001117, -0.00000413, -0.00001382, -0.00000006, 0.00000911, -0.00000515, 0.00000841, -0.00000169, -0.00000859, 0.00000395, 0.00000564, -0.00000085, 0.00000096, -0.00000359, -0.00000515, 0.00000294, -0.00000491, 0.00000508, 0.00000214, },
		{ -0.00001112, 0.00000770, -0.00000620, 0.00000266, -0.00000537, 0.00000369, 0.00001398, -0.00000403, -0.00002018, 0.00000131, 0.00001190, -0.00000984, 0.00000163, -0.00000312, 0.00000584, -0.00000454, -0.00000597, -0.00001223, 0.00001324, -0.00000872, -0.00000066, -0.00002244, 0.00000806, 0.00000460, -0.00000288, 0.00002767, -0.00000985, 0.00002625, -0.00000241, 0.00000373, -0.00000760, -0.00001862, 0.00000290, -0.00000879, -0.00001455, -0.00000078, -0.00003421, -0.00000208, -0.00017027, -0.00005010, 0.00065603, -0.00012750, -0.00012510, -0.00001301, -0.00002798, -0.00001170, 0.00000718, -0.00000862, -0.00000923, 0.00000730, -0.00002161, 0.00001432, -0.00002421, 0.00001047, -0.00001710, 0.00000786, -0.00001360, 0.00001799, -0.00002625, 0.00000606, 0.00000659, -0.00001020, 0.00001738, -0.00001458, 0.00001173, -0.00000796, 0.00000737, -0.00000522, 0.00001083, 0.00000103, -0.00001208, -0.00000187, 0.00001072, -0.00000627, -0.00000245, 0.00000049, 0.00000529, -0.00000601, 0.00000146, 0.00000084, 0.00000670, -0.00000025, -0.00000874, 0.00000263, -0.00000024, 0.00000111, -0.00000324, -0.00000157, -0.00000096, 0.00000625, -0.00000309, 0.00000141, 0.00000025, 0.00000108, -0.00000788, 0.00000796, -0.00000960, 0.00000307, 0.00000889, -0.00000564, -0.00000478, -0.00000037, 0.00000839, -0.00000299, 0.00000001, -0.00000591, 0.00000378, -0.00000226, 0.00001107, -0.00000173, -0.00001399, -0.00000261, 0.00001005, -0.00000466, 0.00000672, 0.00000121, -0.00000976, 0.00000315, 0.00000605, -0.00000063, 0.00000045, -0.00000201, -0.00000647, 0.00000330, -0.00000508, 0.00000477, 0.00000191, 0.00000807, },
		{ 0.00000470, -0.00000299, -0.00000006, -0.00000160, -0.00000181, 0.00001695, -0.00000333, -0.00001793, -0.00000456, 0.00001595, -0.00001143, 0.00000312, -0.00000562, 0.00000772, -0.00000570, -0.00000364, -0.00001515, 0.00001444, -0.00000923, -0.00000022, -0.00002014, 0.00000165, 0.00000906, -0.00000664, 0.00002988, -0.00000790, 0.00002311, 0.00000377, -0.00000051, -0.00000226, -0.00002397, 0.00000628, -0.00001197, -0.00000975, -0.00000698, -0.00002426, -0.00001287, -0.00013247, -0.00011248, 0.00061697, -0.00001952, -0.00017256, 0.00000214, -0.00004046, -0.00000394, -0.00000192, 0.00000144, -0.00001851, 0.00001470, -0.00002744, 0.00001889, -0.00002736, 0.00001221, -0.00001865, 0.00000920, -0.00001479, 0.00001675, -0.00002095, -0.00000176, 0.00001340, -0.00001559, 0.00001986, -0.00001384, 0.00001024, -0.00000631, 0.00000558, -0.00000353, 0.00000814, 0.00000445, -0.00001279, -0.00000282, 0.00000973, -0.00000330, -0.00000436, 0.00000072, 0.00000376, -0.00000275, -0.00000165, 0.00000314, 0.00000419, 0.00000308, -0.00001078, 0.00000301, -0.00000077, 0.00000189, -0.00000359, -0.00000102, -0.00000172, 0.00000592, -0.00000235, 0.00000123, -0.00000011, 0.00000195, -0.00000781, 0.00000630, -0.00000680, -0.00000059, 0.00001109, -0.00000543, -0.00000415, -0.00000283, 0.00000958, -0.00000272, 0.00000044, -0.00000633, 0.00000362, -0.00000258, 0.00001005, 0.00000112, -0.00001411, -0.00000475, 0.00001005, -0.00000323, 0.00000435, 0.00000420, -0.00001058, 0.00000217, 0.00000632, -0.00000015, -0.00000036, -0.00000038, -0.00000749, 0.00000328, -0.00000490, 0.00000411, 0.00000207, 0.00000802, -0.00000937, },
		{ 0.00000006, -0.00000256, 0.00000186, -0.00000624, 0.00001821, -0.00000156, -0.00001608, -0.00000944, 0.00001823, -0.00001125, 0.00000312, -0.00000651, 0.00000773, -0.00000505, -0.00000313, -0.00001591, 0.00001319, -0.00000705, -0.00000275, -0.00001501, -0.00000634, 0.00001386, -0.00001040, 0.00003162, -0.00000548, 0.00001994, 0.00000936, -0.00000350, 0.00000161, -0.00002741, 0.00000747, -0.00001299, -0.00000731, -0.00001039, -0.00001794, -0.00001940, -0.00010333, -0.00015322, 0.00055642, 0.00008707, -0.00020516, 0.00000597, -0.00004545, -0.00000171, -0.00000681, 0.00000759, -0.00002387, 0.00001832, -0.00002977, 0.00002017, -0.00002730, 0.00001097, -0.00001742, 0.00000769, -0.00001282, 0.00001227, -0.00001312, -0.00001076, 0.00002010, -0.00001989, 0.00002072, -0.00001180, 0.00000803, -0.00000422, 0.00000347, -0.00000157, 0.00000529, 0.00000759, -0.00001306, -0.00000368, 0.00000830, -0.00000027, -0.00000599, 0.00000105, 0.00000183, 0.00000049, -0.00000423, 0.00000474, 0.00000240, 0.00000544, -0.00001156, 0.00000226, -0.00000047, 0.00000189, -0.00000322, -0.00000116, -0.00000170, 0.00000493, -0.00000129, 0.00000084, -0.00000013, 0.00000220, -0.00000692, 0.00000401, -0.00000380, -0.00000390, 0.00001252, -0.00000470, -0.00000367, -0.00000481, 0.00000988, -0.00000174, 0.00000033, -0.00000614, 0.00000285, -0.00000220, 0.00000828, 0.00000417, -0.00001406, -0.00000649, 0.00000920, -0.00000112, 0.00000167, 0.00000691, -0.00001080, 0.00000094, 0.00000651, 0.00000050, -0.00000131, 0.00000111, -0.00000805, 0.00000279, -0.00000435, 0.00000313, 0.00000258, 0.00000743, -0.00000706, 0.00000148, },
		{ -0.00000448, 0.00000460, -0.00000918, 0.00001762, 0.00000124, -0.00001470, -0.00001310, 0.00001858, -0.00000933, 0.00000169, -0.00000583, 0.00000602, -0.00000280, -0.00000418, -0.00001482, 0.00000987, -0.00000279, -0.00000735, -0.00000828, -0.00001449, 0.00001776, -0.00001323, 0.00003218, -0.00000211, 0.00001648, 0.00001454, -0.00000543, 0.00000429, -0.00002925, 0.00000683, -0.00001227, -0.00000680, -0.00001145, -0.00001471, -0.00002225, -0.00008187, -0.00017619, 0.00048267, 0.00018369, -0.00021858, -0.00000250, -0.00004296, -0.00000483, -0.00000765, 0.00000996, -0.00002545, 0.00001838, -0.00002888, 0.00001840, -0.00002433, 0.00000710, -0.00001377, 0.00000376, -0.00000831, 0.00000549, -0.00000398, -0.00001964, 0.00002553, -0.00002228, 0.00001958, -0.00000843, 0.00000518, -0.00000177, 0.00000115, 0.00000050, 0.00000249, 0.00001027, -0.00001283, -0.00000449, 0.00000660, 0.00000256, -0.00000714, 0.00000129, -0.00000015, 0.00000323, -0.00000584, 0.00000535, 0.00000159, 0.00000663, -0.00001101, 0.00000045, 0.00000057, 0.00000123, -0.00000224, -0.00000186, -0.00000108, 0.00000354, -0.00000013, 0.00000038, 0.00000008, 0.00000196, -0.00000548, 0.00000149, -0.00000101, -0.00000651, 0.00001300, -0.00000340, -0.00000343, -0.00000618, 0.00000925, -0.00000010, -0.00000022, -0.00000543, 0.00000160, -0.00000128, 0.00000600, 0.00000715, -0.00001370, -0.00000793, 0.00000771, 0.00000131, -0.00000091, 0.00000899, -0.00001026, -0.00000059, 0.00000665, 0.00000123, -0.00000224, 0.00000227, -0.00000804, 0.00000184, -0.00000347, 0.00000190, 0.00000332, 0.00000649, -0.00000450, -0.00000157, 0.00000258, },
		{ 0.00000635, -0.00001049, 0.00001531, 0.00000481, -0.00001368, -0.00001553, 0.00001715, -0.00000597, -0.00000083, -0.00000395, 0.00000301, 0.00000050, -0.00000623, -0.00001249, 0.00000527, 0.00000261, -0.00001281, -0.00000145, -0.00002134, 0.00001966, -0.00001436, 0.00003099, 0.00000254, 0.00001251, 0.00001951, -0.00000659, 0.00000613, -0.00002995, 0.00000495, -0.00001046, -0.00000751, -0.00001097, -0.00001360, -0.00002257, -0.00006632, -0.00018591, 0.00040362, 0.00026374, -0.00021104, -0.00002218, -0.00003446, -0.00001202, -0.00000548, 0.00000936, -0.00002403, 0.00001566, -0.00002551, 0.00001433, -0.00001919, 0.00000139, -0.00000851, -0.00000174, -0.00000227, -0.00000233, 0.00000506, -0.00002707, 0.00002871, -0.00002220, 0.00001629, -0.00000386, 0.00000190, 0.00000087, -0.00000122, 0.00000254, -0.00000010, 0.00001235, -0.00001206, -0.00000529, 0.00000480, 0.00000497, -0.00000767, 0.00000128, -0.00000187, 0.00000511, -0.00000626, 0.00000483, 0.00000181, 0.00000666, -0.00000926, -0.00000217, 0.00000207, 0.00000017, -0.00000091, -0.00000286, -0.00000018, 0.00000207, 0.00000088, 0.00000003, 0.00000039, 0.00000143, -0.00000380, -0.00000089, 0.00000120, -0.00000817, 0.00001246, -0.00000155, -0.00000347, -0.00000689, 0.00000778, 0.00000201, -0.00000103, -0.00000440, 0.00000007, -0.00000008, 0.00000356, 0.00000974, -0.00001284, -0.00000918, 0.00000585, 0.00000369, -0.00000301, 0.00001019, -0.00000889, -0.00000240, 0.00000676, 0.00000194, -0.00000300, 0.00000297, -0.00000742, 0.00000048, -0.00000236, 0.00000052, 0.00000417, 0.00000540, -0.00000198, -0.00000413, 0.00000429, -0.00000560, },
		{ -0.00001025, 0.00001163, 0.00000872, -0.00001279, -0.00001691, 0.00001429, -0.00000171, -0.00000389, -0.00000141, -0.00000063, 0.00000414, -0.00000853, -0.00000974, 0.00000030, 0.00000803, -0.00001783, 0.00000406, -0.00002564, 0.00001877, -0.00001333, 0.00002776, 0.00000859, 0.00000798, 0.00002436, -0.00000720, 0.00000752, -0.00002998, 0.00000244, -0.00000828, -0.00000867, -0.00000986, -0.00001349, -0.00002172, -0.00005463, -0.00018684, 0.00032585, 0.00032309, -0.00018340, -0.00005006, -0.00002261, -0.00002121, -0.00000190, 0.00000714, -0.00002083, 0.00001134, -0.00002077, 0.00000898, -0.00001291, -0.00000515, -0.00000263, -0.00000777, 0.00000412, -0.00000985, 0.00001263, -0.00003194, 0.00002894, -0.00001939, 0.00001100, 0.00000156, -0.00000153, 0.00000347, -0.00000346, 0.00000438, -0.00000230, 0.00001372, -0.00001075, -0.00000610, 0.00000304, 0.00000678, -0.00000749, 0.00000093, -0.00000312, 0.00000592, -0.00000542, 0.00000327, 0.00000291, 0.00000575, -0.00000664, -0.00000520, 0.00000365, -0.00000098, 0.00000045, -0.00000386, 0.00000071, 0.00000081, 0.00000150, -0.00000007, 0.00000068, 0.00000080, -0.00000216, -0.00000281, 0.00000261, -0.00000879, 0.00001099, 0.00000069, -0.00000370, -0.00000701, 0.00000565, 0.00000429, -0.00000185, -0.00000328, -0.00000149, 0.00000114, 0.00000128, 0.00001165, -0.00001136, -0.00001036, 0.00000385, 0.00000570, -0.00000437, 0.00001036, -0.00000674, -0.00000438, 0.00000682, 0.00000260, -0.00000350, 0.00000314, -0.00000623, -0.00000118, -0.00000115, -0.00000085, 0.00000496, 0.00000435, 0.00000026, -0.00000597, 0.00000506, -0.00000584, 0.00000707, },
		{ 0.00000710, 0.00001242, -0.00001171, -0.00001753, 0.00001050, 0.00000282, -0.00000688, 0.00000113, -0.00000419, 0.00000734, -0.00001031, -0.00000738, -0.00000411, 0.00001242, -0.00002126, 0.00000717, -0.00002662, 0.00001482, -0.00001013, 0.00002257, 0.00001587, 0.00000306, 0.00002904, -0.00000737, 0.00000871, -0.00002971, -0.00000022, -0.00000632, -0.00000959, -0.00000890, -0.00001338, -0.00002089, -0.00004499, -0.00018274, 0.00025422, 0.00036030, -0.00013887, -0.00008178, -0.00001073, -0.00003000, 0.00000125, 0.00000476, -0.00001718, 0.00000666, -0.00001580, 0.00000343, -0.00000656, -0.00001144, 0.00000280, -0.00001326, 0.00000970, -0.00001582, 0.00001764, -0.00003355, 0.00002598, -0.00001405, 0.00000420, 0.00000731, -0.00000470, 0.00000580, -0.00000539, 0.00000584, -0.00000397, 0.00001435, -0.00000893, -0.00000691, 0.00000145, 0.00000789, -0.00000660, 0.00000022, -0.00000380, 0.00000564, -0.00000349, 0.00000096, 0.00000456, 0.00000427, -0.00000360, -0.00000815, 0.00000489, -0.00000184, 0.00000154, -0.00000455, 0.00000127, 0.00000004, 0.00000158, 0.00000015, 0.00000084, 0.00000026, -0.00000082, -0.00000406, 0.00000311, -0.00000844, 0.00000878, 0.00000311, -0.00000398, -0.00000669, 0.00000316, 0.00000645, -0.00000242, -0.00000229, -0.00000284, 0.00000210, -0.00000056, 0.00001268, -0.00000922, -0.00001151, 0.00000190, 0.00000712, -0.00000483, 0.00000953, -0.00000397, -0.00000637, 0.00000676, 0.00000319, -0.00000372, 0.00000280, -0.00000459, -0.00000296, -0.00000001, -0.00000208, 0.00000555, 0.00000351, 0.00000203, -0.00000697, 0.00000489, -0.00000529, 0.00000689, -0.00000874, },
		{ 0.00001540, -0.00001013, -0.00001770, 0.00000633, 0.00000699, -0.00000918, 0.00000309, -0.00000700, 0.00000943, -0.00001094, -0.00000605, -0.00000723, 0.00001502, -0.00002233, 0.00000725, -0.00002409, 0.00000808, -0.00000520, 0.00001586, 0.00002387, -0.00000182, 0.00003329, -0.00000705, 0.00000981, -0.00002935, -0.00000270, -0.00000493, -0.00000987, -0.00000861, -0.00001260, -0.00002092, -0.00003608, -0.00017638, 0.00019160, 0.00037636, -0.00008245, -0.00011227, -0.00000218, -0.00003611, 0.00000234, 0.00000354, -0.00001429, 0.00000274, -0.00001162, -0.00000137, -0.00000106, -0.00001658, 0.00000692, -0.00001733, 0.00001354, -0.00001938, 0.00001946, -0.00003168, 0.00002008, -0.00000674, -0.00000341, 0.00001274, -0.00000722, 0.00000761, -0.00000683, 0.00000680, -0.00000504, 0.00001422, -0.00000670, -0.00000770, 0.00000010, 0.00000827, -0.00000508, -0.00000078, -0.00000391, 0.00000438, -0.00000078, -0.00000170, 0.00000635, 0.00000266, -0.00000063, -0.00001055, 0.00000543, -0.00000215, 0.00000211, -0.00000472, 0.00000132, -0.00000009, 0.00000106, 0.00000069, 0.00000086, -0.00000010, 0.00000009, -0.00000456, 0.00000276, -0.00000731, 0.00000613, 0.00000543, -0.00000414, -0.00000612, 0.00000059, 0.00000817, -0.00000252, -0.00000162, -0.00000380, 0.00000263, -0.00000178, 0.00001276, -0.00000647, -0.00001262, 0.00000014, 0.00000782, -0.00000440, 0.00000785, -0.00000084, -0.00000816, 0.00000649, 0.00000372, -0.00000365, 0.00000202, -0.00000267, -0.00000466, 0.00000089, -0.00000303, 0.00000584, 0.00000299, 0.00000324, -0.00000713, 0.00000394, -0.00000418, 0.00000606, -0.00000639, 0.00000235, },
		{ -0.00000784, -0.00001769, 0.00000224, 0.00001025, -0.00001035, 0.00000404, -0.00000860, 0.00000998, -0.00001004, -0.00000614, -0.00000867, 0.00001541, -0.00002076, 0.00000430, -0.00001849, -0.00000062, 0.00000062, 0.00000837, 0.00003189, -0.00000607, 0.00003673, -0.00000606, 0.00001077, -0.00002892, -0.00000492, -0.00000424, -0.00000939, -0.00000918, -0.00001089, -0.00002214, -0.00002725, -0.00016937, 0.00013899, 0.00037423, -0.00002015, -0.00013658, 0.00000025, -0.00003792, 0.00000026, 0.00000434, -0.00001294, 0.00000031, -0.00000889, -0.00000479, 0.00000293, -0.00001994, 0.00000915, -0.00001943, 0.00001509, -0.00002007, 0.00001797, -0.00002665, 0.00001192, 0.00000166, -0.00001093, 0.00001719, -0.00000870, 0.00000870, -0.00000765, 0.00000716, -0.00000545, 0.00001340, -0.00000417, -0.00000842, -0.00000101, 0.00000798, -0.00000310, -0.00000193, -0.00000357, 0.00000239, 0.00000230, -0.00000424, 0.00000783, 0.00000134, 0.00000183, -0.00001202, 0.00000500, -0.00000174, 0.00000201, -0.00000426, 0.00000076, 0.00000042, 0.00000002, 0.00000147, 0.00000074, -0.00000023, 0.00000052, -0.00000436, 0.00000174, -0.00000568, 0.00000339, 0.00000738, -0.00000402, -0.00000551, -0.00000176, 0.00000921, -0.00000201, -0.00000138, -0.00000426, 0.00000263, -0.00000231, 0.00001195, -0.00000329, -0.00001360, -0.00000142, 0.00000780, -0.00000320, 0.00000559, 0.00000232, -0.00000951, 0.00000592, 0.00000423, -0.00000335, 0.00000092, -0.00000066, -0.00000607, 0.00000143, -0.00000362, 0.00000576, 0.00000282, 0.00000390, -0.00000657, 0.00000244, -0.00000280, 0.00000491, -0.00000367, -0.00000205, 0.00001721, },
		{ -0.00001767, -0.00000139, 0.00001223, -0.00001014, 0.00000375, -0.00000877, 0.00000882, -0.00000755, -0.00000766, -0.00000840, 0.00001366, -0.00001677, -0.00000117, -0.00001075, -0.00001008, 0.00000626, 0.00000102, 0.00003907, -0.00000904, 0.00003891, -0.00000415, 0.00001143, -0.00002827, -0.00000699, -0.00000413, -0.00000827, -0.00001048, -0.00000836, -0.00002439, -0.00001846, -0.00016232, 0.00009592, 0.00035817, 0.00004187, -0.00015058, -0.00000515, -0.00003468, -0.00000534, 0.00000740, -0.00001339, -0.00000037, -0.00000785, -0.00000660, 0.00000516, -0.00002128, 0.00000930, -0.00001938, 0.00001423, -0.00001796, 0.00001355, -0.00001925, 0.00000253, 0.00001008, -0.00001743, 0.00002004, -0.00000888, 0.00000895, -0.00000780, 0.00000691, -0.00000525, 0.00001200, -0.00000149, -0.00000900, -0.00000187, 0.00000712, -0.00000087, -0.00000306, -0.00000296, 0.00000002, 0.00000526, -0.00000620, 0.00000864, 0.00000066, 0.00000344, -0.00001231, 0.00000351, -0.00000058, 0.00000128, -0.00000320, -0.00000034, 0.00000148, -0.00000139, 0.00000233, 0.00000060, -0.00000017, 0.00000054, -0.00000362, 0.00000031, -0.00000391, 0.00000091, 0.00000871, -0.00000347, -0.00000501, -0.00000368, 0.00000943, -0.00000087, -0.00000155, -0.00000424, 0.00000211, -0.00000221, 0.00001040, 0.00000009, -0.00001431, -0.00000280, 0.00000717, -0.00000148, 0.00000309, 0.00000518, -0.00001023, 0.00000498, 0.00000474, -0.00000287, -0.00000036, 0.00000123, -0.00000703, 0.00000151, -0.00000381, 0.00000533, 0.00000297, 0.00000409, -0.00000549, 0.00000071, -0.00000148, 0.00000377, -0.00000102, -0.00000561, 0.00001759, -0.00000478, },
		{ -0.00000436, 0.00001281, -0.00000855, 0.00000229, -0.00000758, 0.00000614, -0.00000376, -0.00001030, -0.00000675, 0.00001021, -0.00001103, -0.00000817, -0.00000219, -0.00001892, 0.00001061, -0.00000533, 0.00004464, -0.00001012, 0.00003944, -0.00000107, 0.00001161, -0.00002721, -0.00000911, -0.00000434, -0.00000685, -0.00001214, -0.00000541, -0.00002718, -0.00001015, -0.00015510, 0.00006088, 0.00033293, 0.00009808, -0.00015156, -0.00001871, -0.00002678, -0.00001398, 0.00001232, -0.00001537, 0.00000048, -0.00000827, -0.00000697, 0.00000577, -0.00002075, 0.00000756, -0.00001741, 0.00001124, -0.00001353, 0.00000702, -0.00001058, -0.00000689, 0.00001744, -0.00002213, 0.00002089, -0.00000764, 0.00000839, -0.00000730, 0.00000610, -0.00000452, 0.00001017, 0.00000118, -0.00000935, -0.00000253, 0.00000585, 0.00000138, -0.00000398, -0.00000227, -0.00000240, 0.00000768, -0.00000724, 0.00000851, 0.00000083, 0.00000405, -0.00001138, 0.00000107, 0.00000116, 0.00000004, -0.00000170, -0.00000181, 0.00000286, -0.00000289, 0.00000304, 0.00000054, 0.00000000, 0.00000025, -0.00000256, -0.00000120, -0.00000232, -0.00000105, 0.00000927, -0.00000244, -0.00000472, -0.00000504, 0.00000881, 0.00000081, -0.00000204, -0.00000382, 0.00000120, -0.00000164, 0.00000839, 0.00000337, -0.00001458, -0.00000408, 0.00000610, 0.00000047, 0.00000069, 0.00000742, -0.00001018, 0.00000367, 0.00000528, -0.00000229, -0.00000168, 0.00000281, -0.00000743, 0.00000112, -0.00000364, 0.00000459, 0.00000336, 0.00000394, -0.00000413, -0.00000095, -0.00000050, 0.00000292, 0.00000124, -0.00000803, 0.00001646, -0.00000106, -0.00001765, },
		{ 0.00001203, -0.00000581, -0.00000008, -0.00000533, 0.00000234, 0.00000082, -0.00001352, -0.00000432, 0.00000578, -0.00000447, -0.00001550, 0.00000578, -0.00002580, 0.00001276, -0.00000999, 0.00004796, -0.00000890, 0.00003810, 0.00000328, 0.00001119, -0.00002556, -0.00001152, -0.00000455, -0.00000552, -0.00001371, -0.00000258, -0.00002983, -0.00000301, -0.00014710, 0.00003187, 0.00030310, 0.00014424, -0.00013857, -0.00003929, -0.00001559, -0.00002446, 0.00001812, -0.00001811, 0.00000223, -0.00000961, -0.00000638, 0.00000516, -0.00001879, 0.00000442, -0.00001404, 0.00000674, -0.00000759, -0.00000053, -0.00000188, -0.00001515, 0.00002282, -0.00002445, 0.00001956, -0.00000510, 0.00000714, -0.00000625, 0.00000484, -0.00000340, 0.00000810, 0.00000365, -0.00000941, -0.00000305, 0.00000434, 0.00000341, -0.00000451, -0.00000169, -0.00000455, 0.00000924, -0.00000714, 0.00000736, 0.00000186, 0.00000369, -0.00000939, -0.00000206, 0.00000320, -0.00000143, -0.00000002, -0.00000339, 0.00000428, -0.00000421, 0.00000340, 0.00000069, 0.00000019, -0.00000018, -0.00000142, -0.00000252, -0.00000117, -0.00000230, 0.00000904, -0.00000099, -0.00000464, -0.00000582, 0.00000744, 0.00000284, -0.00000265, -0.00000319, 0.00000009, -0.00000083, 0.00000623, 0.00000623, -0.00001425, -0.00000536, 0.00000477, 0.00000236, -0.00000129, 0.00000885, -0.00000932, 0.00000202, 0.00000581, -0.00000165, -0.00000292, 0.00000397, -0.00000722, 0.00000030, -0.00000320, 0.00000366, 0.00000387, 0.00000364, -0.00000274, -0.00000228, -0.00000007, 0.00000250, 0.00000290, -0.00000919, 0.00001399, 0.00000306, -0.00001755, -0.00000662, },
		{ -0.00000233, -0.00000292, -0.00000251, -0.00000198, 0.00000550, -0.00001662, -0.00000184, 0.00000124, 0.00000185, -0.00002194, 0.00001189, -0.00002973, 0.00001213, -0.00001259, 0.00004873, -0.00000523, 0.00003490, 0.00000885, 0.00001016, -0.00002320, -0.00001437, -0.00000449, -0.00000464, -0.00001476, -0.00000041, -0.00003167, 0.00000227, -0.00013767, 0.00000694, 0.00027254, 0.00017778, -0.00011251, -0.00006441, -0.00000328, -0.00003508, 0.00002350, -0.00002063, 0.00000402, -0.00001110, -0.00000550, 0.00000397, -0.00001602, 0.00000055, -0.00000998, 0.00000151, -0.00000111, -0.00000794, 0.00000566, -0.00002128, 0.00002557, -0.00002415, 0.00001619, -0.00000154, 0.00000548, -0.00000486, 0.00000328, -0.00000207, 0.00000598, 0.00000580, -0.00000913, -0.00000349, 0.00000276, 0.00000505, -0.00000453, -0.00000137, -0.00000620, 0.00000974, -0.00000587, 0.00000528, 0.00000361, 0.00000255, -0.00000663, -0.00000548, 0.00000517, -0.00000281, 0.00000156, -0.00000480, 0.00000546, -0.00000508, 0.00000324, 0.00000113, 0.00000031, -0.00000060, -0.00000043, -0.00000342, -0.00000063, -0.00000278, 0.00000808, 0.00000078, -0.00000472, -0.00000608, 0.00000553, 0.00000495, -0.00000317, -0.00000254, -0.00000102, -0.00000002, 0.00000420, 0.00000843, -0.00001323, -0.00000669, 0.00000338, 0.00000394, -0.00000264, 0.00000937, -0.00000771, 0.00000016, 0.00000629, -0.00000097, -0.00000399, 0.00000463, -0.00000644, -0.00000085, -0.00000260, 0.00000266, 0.00000437, 0.00000333, -0.00000150, -0.00000310, -0.00000025, 0.00000253, 0.00000395, -0.00000919, 0.00001051, 0.00000717, -0.00001718, -0.00000827, 0.00001016, },
		{ -0.00000570, 0.00000034, -0.00000615, 0.00000957, -0.00001890, -0.00000003, -0.00000260, 0.00000699, -0.00002647, 0.00001520, -0.00003015, 0.00000867, -0.00001318, 0.00004696, 0.00000071, 0.00003012, 0.00001535, 0.00000871, -0.00002019, -0.00001771, -0.00000403, -0.00000443, -0.00001499, 0.00000073, -0.00003219, 0.00000519, -0.00012631, -0.00001550, 0.00024395, 0.00019794, -0.00007592, -0.00009069, 0.00000757, -0.00004398, 0.00002709, -0.00002186, 0.00000496, -0.00001196, -0.00000500, 0.00000280, -0.00001311, -0.00000336, -0.00000594, -0.00000365, 0.00000495, -0.00001419, 0.00001110, -0.00002465, 0.00002547, -0.00002136, 0.00001117, 0.00000256, 0.00000373, -0.00000333, 0.00000163, -0.00000071, 0.00000398, 0.00000750, -0.00000848, -0.00000388, 0.00000124, 0.00000614, -0.00000399, -0.00000137, -0.00000723, 0.00000914, -0.00000359, 0.00000255, 0.00000578, 0.00000096, -0.00000353, -0.00000875, 0.00000669, -0.00000380, 0.00000275, -0.00000577, 0.00000613, -0.00000530, 0.00000247, 0.00000188, 0.00000031, -0.00000090, 0.00000026, -0.00000376, -0.00000074, -0.00000257, 0.00000658, 0.00000265, -0.00000482, -0.00000598, 0.00000332, 0.00000685, -0.00000337, -0.00000206, -0.00000192, 0.00000056, 0.00000256, 0.00000981, -0.00001150, -0.00000807, 0.00000202, 0.00000504, -0.00000326, 0.00000902, -0.00000552, -0.00000174, 0.00000662, -0.00000026, -0.00000485, 0.00000477, -0.00000516, -0.00000215, -0.00000200, 0.00000172, 0.00000473, 0.00000313, -0.00000056, -0.00000333, -0.00000101, 0.00000292, 0.00000451, -0.00000827, 0.00000649, 0.00001084, -0.00001633, -0.00000946, 0.00000753, 0.00000140, },
		{ 0.00000270, -0.00000960, 0.00001243, -0.00001980, 0.00000057, -0.00000512, 0.00001027, -0.00002844, 0.00001529, -0.00002707, 0.00000280, -0.00001218, 0.00004300, 0.00000843, 0.00002430, 0.00002231, 0.00000717, -0.00001671, -0.00002142, -0.00000319, -0.00000495, -0.00001432, 0.00000068, -0.00003117, 0.00000558, -0.00011291, -0.00003643, 0.00021882, 0.00020559, -0.00003254, -0.00011435, 0.00001448, -0.00004953, 0.00002777, -0.00002098, 0.00000437, -0.00001159, -0.00000541, 0.00000217, -0.00001061, -0.00000669, -0.00000257, -0.00000803, 0.00000982, -0.00001847, 0.00001388, -0.00002505, 0.00002268, -0.00001655, 0.00000515, 0.00000663, 0.00000226, -0.00000189, 0.00000006, 0.00000053, 0.00000224, 0.00000868, -0.00000749, -0.00000424, -0.00000012, 0.00000663, -0.00000293, -0.00000167, -0.00000763, 0.00000757, -0.00000058, -0.00000047, 0.00000799, -0.00000067, -0.00000054, -0.00001142, 0.00000742, -0.00000415, 0.00000334, -0.00000613, 0.00000616, -0.00000479, 0.00000109, 0.00000290, 0.00000021, -0.00000102, 0.00000057, -0.00000355, -0.00000141, -0.00000185, 0.00000480, 0.00000440, -0.00000479, -0.00000569, 0.00000111, 0.00000827, -0.00000307, -0.00000189, -0.00000247, 0.00000076, 0.00000144, 0.00001030, -0.00000915, -0.00000945, 0.00000078, 0.00000560, -0.00000316, 0.00000793, -0.00000299, -0.00000349, 0.00000673, 0.00000051, -0.00000548, 0.00000445, -0.00000355, -0.00000344, -0.00000153, 0.00000095, 0.00000488, 0.00000313, 0.00000004, -0.00000303, -0.00000217, 0.00000344, 0.00000477, -0.00000677, 0.00000242, 0.00001365, -0.00001480, -0.00001038, 0.00000454, 0.00000487, -0.00000794, },
		{ -0.00001188, 0.00001367, -0.00001903, -0.00000033, -0.00000601, 0.00001138, -0.00002768, 0.00001225, -0.00002106, -0.00000463, -0.00001031, 0.00003746, 0.00001723, 0.00001815, 0.00002912, 0.00000596, -0.00001305, -0.00002526, -0.00000215, -0.00000605, -0.00001284, -0.00000048, -0.00002868, 0.00000362, -0.00009775, -0.00005614, 0.00019748, 0.00020294, 0.00001322, -0.00013175, 0.00001548, -0.00005064, 0.00002487, -0.00001755, 0.00000189, -0.00000968, -0.00000700, 0.00000235, -0.00000887, -0.00000907, -0.00000026, -0.00001118, 0.00001297, -0.00002039, 0.00001386, -0.00002273, 0.00001776, -0.00001046, -0.00000111, 0.00001009, 0.00000139, -0.00000073, -0.00000130, 0.00000152, 0.00000085, 0.00000935, -0.00000621, -0.00000456, -0.00000125, 0.00000652, -0.00000145, -0.00000217, -0.00000748, 0.00000526, 0.00000276, -0.00000335, 0.00000986, -0.00000198, 0.00000195, -0.00001314, 0.00000713, -0.00000373, 0.00000325, -0.00000581, 0.00000551, -0.00000357, -0.00000077, 0.00000404, 0.00000007, -0.00000095, 0.00000049, -0.00000285, -0.00000248, -0.00000087, 0.00000301, 0.00000581, -0.00000449, -0.00000538, -0.00000086, 0.00000900, -0.00000217, -0.00000210, -0.00000261, 0.00000054, 0.00000089, 0.00001000, -0.00000637, -0.00001069, -0.00000036, 0.00000565, -0.00000248, 0.00000635, -0.00000041, -0.00000488, 0.00000653, 0.00000133, -0.00000590, 0.00000373, -0.00000175, -0.00000455, -0.00000131, 0.00000040, 0.00000479, 0.00000333, 0.00000032, -0.00000231, -0.00000350, 0.00000383, 0.00000499, -0.00000506, -0.00000124, 0.00001529, -0.00001252, -0.00001116, 0.00000155, 0.00000763, -0.00000926, 0.00000418, },
		{ 0.00001317, -0.00001656, -0.00000269, -0.00000529, 0.00001040, -0.00002444, 0.00000670, -0.00001313, -0.00001247, -0.00000844, 0.00003105, 0.00002629, 0.00001245, 0.00003514, 0.00000555, -0.00000956, -0.00002891, -0.00000119, -0.00000748, -0.00001086, -0.00000245, -0.00002504, -0.00000019, -0.00008151, -0.00007438, 0.00017931, 0.00019299, 0.00005694, -0.00014000, 0.00000949, -0.00004696, 0.00001836, -0.00001166, -0.00000241, -0.00000631, -0.00000969, 0.00000335, -0.00000798, -0.00001035, 0.00000083, -0.00001292, 0.00001423, -0.00001990, 0.00001132, -0.00001830, 0.00001153, -0.00000396, -0.00000682, 0.00001243, 0.00000134, 0.00000006, -0.00000236, 0.00000220, -0.00000015, 0.00000953, -0.00000473, -0.00000480, -0.00000214, 0.00000588, 0.00000029, -0.00000274, -0.00000696, 0.00000253, 0.00000600, -0.00000568, 0.00001103, -0.00000265, 0.00000364, -0.00001369, 0.00000577, -0.00000256, 0.00000251, -0.00000486, 0.00000425, -0.00000181, -0.00000290, 0.00000512, -0.00000001, -0.00000076, 0.00000012, -0.00000183, -0.00000369, 0.00000009, 0.00000148, 0.00000668, -0.00000382, -0.00000519, -0.00000241, 0.00000895, -0.00000069, -0.00000263, -0.00000237, -0.00000006, 0.00000082, 0.00000906, -0.00000342, -0.00001163, -0.00000146, 0.00000526, -0.00000142, 0.00000456, 0.00000194, -0.00000574, 0.00000597, 0.00000222, -0.00000611, 0.00000271, 0.00000005, -0.00000534, -0.00000140, 0.00000009, 0.00000447, 0.00000367, 0.00000039, -0.00000138, -0.00000472, 0.00000386, 0.00000540, -0.00000344, -0.00000414, 0.00001562, -0.00000953, -0.00001185, -0.00000118, 0.00000941, -0.00000944, 0.00000459, -0.00001279, },
		{ -0.00001264, -0.00000621, -0.00000328, 0.00000778, -0.00001940, -0.00000039, -0.00000455, -0.00001947, -0.00000741, 0.00002446, 0.00003480, 0.00000796, 0.00003979, 0.00000634, -0.00000655, -0.00003206, -0.00000061, -0.00000890, -0.00000873, -0.00000481, -0.00002075, -0.00000513, -0.00006509, -0.00009048, 0.00016313, 0.00017903, 0.00009479, -0.00013733, -0.00000346, -0.00003904, 0.00000886, -0.00000388, -0.00000803, -0.00000190, -0.00001313, 0.00000491, -0.00000778, -0.00001064, 0.00000080, -0.00001333, 0.00001374, -0.00001734, 0.00000689, -0.00001264, 0.00000495, 0.00000204, -0.00001131, 0.00001330, 0.00000218, 0.00000049, -0.00000311, 0.00000257, -0.00000081, 0.00000933, -0.00000317, -0.00000492, -0.00000281, 0.00000483, 0.00000209, -0.00000322, -0.00000624, -0.00000030, 0.00000874, -0.00000713, 0.00001129, -0.00000252, 0.00000437, -0.00001303, 0.00000344, -0.00000080, 0.00000129, -0.00000346, 0.00000259, 0.00000027, -0.00000502, 0.00000592, 0.00000008, -0.00000051, -0.00000044, -0.00000069, -0.00000481, 0.00000080, 0.00000043, 0.00000693, -0.00000275, -0.00000520, -0.00000344, 0.00000812, 0.00000124, -0.00000336, -0.00000188, -0.00000093, 0.00000108, 0.00000775, -0.00000057, -0.00001211, -0.00000259, 0.00000459, -0.00000019, 0.00000283, 0.00000383, -0.00000599, 0.00000506, 0.00000316, -0.00000614, 0.00000149, 0.00000172, -0.00000570, -0.00000180, -0.00000002, 0.00000401, 0.00000406, 0.00000037, -0.00000044, -0.00000558, 0.00000331, 0.00000612, -0.00000210, -0.00000610, 0.00001468, -0.00000602, -0.00001242, -0.00000351, 0.00001007, -0.00000847, 0.00000395, -0.00001236, 0.00001105, },
		{ -0.00001036, -0.00000056, 0.00000417, -0.00001343, -0.00000789, 0.00000339, -0.00002456, -0.00000784, 0.00001821, 0.00004210, 0.00000525, 0.00004266, 0.00000860, -0.00000422, -0.00003444, -0.00000069, -0.00001000, -0.00000683, -0.00000713, -0.00001638, -0.00001041, -0.00004948, -0.00010365, 0.00014758, 0.00016409, 0.00012391, -0.00012333, -0.00002220, -0.00002818, -0.00000247, 0.00000480, -0.00001417, 0.00000287, -0.00001673, 0.00000657, -0.00000793, -0.00001025, -0.00000003, -0.00001272, 0.00001187, -0.00001330, 0.00000142, -0.00000672, -0.00000105, 0.00000681, -0.00001410, 0.00001258, 0.00000382, 0.00000066, -0.00000362, 0.00000268, -0.00000119, 0.00000885, -0.00000165, -0.00000487, -0.00000329, 0.00000351, 0.00000377, -0.00000345, -0.00000552, -0.00000292, 0.00001068, -0.00000752, 0.00001056, -0.00000155, 0.00000417, -0.00001131, 0.00000039, 0.00000126, -0.00000018, -0.00000185, 0.00000078, 0.00000238, -0.00000687, 0.00000627, 0.00000043, -0.00000028, -0.00000103, 0.00000036, -0.00000562, 0.00000106, -0.00000004, 0.00000654, -0.00000134, -0.00000539, -0.00000395, 0.00000664, 0.00000342, -0.00000410, -0.00000130, -0.00000187, 0.00000146, 0.00000634, 0.00000189, -0.00001198, -0.00000380, 0.00000376, 0.00000099, 0.00000139, 0.00000511, -0.00000559, 0.00000386, 0.00000408, -0.00000599, 0.00000015, 0.00000313, -0.00000562, -0.00000246, -0.00000001, 0.00000349, 0.00000440, 0.00000041, 0.00000031, -0.00000590, 0.00000212, 0.00000719, -0.00000111, -0.00000711, 0.00001266, -0.00000228, -0.00001274, -0.00000538, 0.00000965, -0.00000652, 0.00000246, -0.00001083, 0.00000767, -0.00000778, },
		{ 0.00000222, 0.00000037, -0.00000751, -0.00001465, 0.00000955, -0.00002694, -0.00001004, 0.00001260, 0.00004771, 0.00000468, 0.00004356, 0.00001237, -0.00000263, -0.00003594, -0.00000159, -0.00001055, -0.00000547, -0.00000901, -0.00001243, -0.00001531, -0.00003556, -0.00011317, 0.00013142, 0.00015062, 0.00014277, -0.00009903, -0.00004459, -0.00001626, -0.00001412, 0.00001316, -0.00001986, 0.00000718, -0.00001983, 0.00000779, -0.00000802, -0.00000960, -0.00000125, -0.00001152, 0.00000914, -0.00000848, -0.00000419, -0.00000147, -0.00000571, 0.00000984, -0.00001500, 0.00001039, 0.00000599, 0.00000077, -0.00000399, 0.00000261, -0.00000137, 0.00000820, -0.00000025, -0.00000463, -0.00000360, 0.00000204, 0.00000515, -0.00000331, -0.00000493, -0.00000511, 0.00001163, -0.00000681, 0.00000893, 0.00000012, 0.00000321, -0.00000879, -0.00000299, 0.00000329, -0.00000159, -0.00000030, -0.00000092, 0.00000424, -0.00000818, 0.00000600, 0.00000111, -0.00000014, -0.00000154, 0.00000117, -0.00000597, 0.00000080, 0.00000006, 0.00000560, 0.00000030, -0.00000569, -0.00000404, 0.00000472, 0.00000557, -0.00000462, -0.00000082, -0.00000270, 0.00000174, 0.00000509, 0.00000375, -0.00001119, -0.00000512, 0.00000289, 0.00000194, 0.00000039, 0.00000572, -0.00000463, 0.00000249, 0.00000491, -0.00000563, -0.00000124, 0.00000422, -0.00000509, -0.00000328, 0.00000001, 0.00000303, 0.00000457, 0.00000061, 0.00000074, -0.00000559, 0.00000032, 0.00000849, -0.00000040, -0.00000732, 0.00000993, 0.00000130, -0.00001266, -0.00000685, 0.00000832, -0.00000391, 0.00000046, -0.00000858, 0.00000357, -0.00000259, -0.00001449, },
		{ -0.00000292, -0.00000248, -0.00001978, 0.00001315, -0.00002629, -0.00001390, 0.00000762, 0.00005144, 0.00000633, 0.00004258, 0.00001748, -0.00000163, -0.00003656, -0.00000336, -0.00001046, -0.00000482, -0.00001021, -0.00000927, -0.00001927, -0.00002399, -0.00011857, 0.00011377, 0.00014017, 0.00015117, -0.00006664, -0.00006782, -0.00000542, -0.00002450, 0.00002000, -0.00002415, 0.00001027, -0.00002178, 0.00000811, -0.00000763, -0.00000907, -0.00000244, -0.00001017, 0.00000608, -0.00000361, -0.00000912, 0.00000237, -0.00000854, 0.00001093, -0.00001410, 0.00000705, 0.00000831, 0.00000105, -0.00000435, 0.00000246, -0.00000146, 0.00000750, 0.00000093, -0.00000420, -0.00000379, 0.00000056, 0.00000612, -0.00000277, -0.00000455, -0.00000675, 0.00001157, -0.00000512, 0.00000661, 0.00000223, 0.00000178, -0.00000586, -0.00000629, 0.00000492, -0.00000266, 0.00000094, -0.00000225, 0.00000561, -0.00000879, 0.00000507, 0.00000208, -0.00000009, -0.00000190, 0.00000162, -0.00000580, 0.00000002, 0.00000066, 0.00000430, 0.00000199, -0.00000598, -0.00000383, 0.00000262, 0.00000742, -0.00000472, -0.00000060, -0.00000326, 0.00000176, 0.00000418, 0.00000489, -0.00000976, -0.00000650, 0.00000202, 0.00000258, -0.00000013, 0.00000571, -0.00000326, 0.00000110, 0.00000557, -0.00000507, -0.00000263, 0.00000495, -0.00000420, -0.00000415, -0.00000008, 0.00000271, 0.00000451, 0.00000105, 0.00000079, -0.00000467, -0.00000192, 0.00000982, 0.00000021, -0.00000699, 0.00000690, 0.00000436, -0.00001204, -0.00000799, 0.00000633, -0.00000100, -0.00000166, -0.00000605, -0.00000070, 0.00000228, -0.00001797, 0.00000441, },
		{ 0.00000106, -0.00002271, 0.00001387, -0.00002279, -0.00001892, 0.00000303, 0.00005329, 0.00001004, 0.00004004, 0.00002358, -0.00000094, -0.00003647, -0.00000588, -0.00000976, -0.00000490, -0.00001064, -0.00000708, -0.00002198, -0.00001511, -0.00011975, 0.00009430, 0.00013338, 0.00015012, -0.00002928, -0.00008881, 0.00000232, -0.00003226, 0.00002433, -0.00002633, 0.00001156, -0.00002216, 0.00000719, -0.00000649, -0.00000897, -0.00000327, -0.00000904, 0.00000318, 0.00000071, -0.00001276, 0.00000438, -0.00000939, 0.00001021, -0.00001176, 0.00000309, 0.00001035, 0.00000175, -0.00000480, 0.00000230, -0.00000152, 0.00000682, 0.00000187, -0.00000361, -0.00000384, -0.00000085, 0.00000659, -0.00000181, -0.00000439, -0.00000780, 0.00001058, -0.00000271, 0.00000394, 0.00000447, 0.00000021, -0.00000292, -0.00000911, 0.00000584, -0.00000317, 0.00000168, -0.00000307, 0.00000637, -0.00000863, 0.00000353, 0.00000328, -0.00000010, -0.00000206, 0.00000168, -0.00000515, -0.00000116, 0.00000156, 0.00000286, 0.00000353, -0.00000613, -0.00000352, 0.00000058, 0.00000872, -0.00000425, -0.00000073, -0.00000344, 0.00000142, 0.00000371, 0.00000529, -0.00000780, -0.00000784, 0.00000117, 0.00000289, -0.00000021, 0.00000522, -0.00000167, -0.00000014, 0.00000597, -0.00000428, -0.00000399, 0.00000531, -0.00000303, -0.00000493, -0.00000037, 0.00000259, 0.00000419, 0.00000173, 0.00000050, -0.00000329, -0.00000435, 0.00001089, 0.00000096, -0.00000643, 0.00000400, 0.00000658, -0.00001077, -0.00000889, 0.00000395, 0.00000180, -0.00000350, -0.00000368, -0.00000462, 0.00000627, -0.00002029, 0.00000555, -0.00000514, },
		{ -0.00002328, 0.00001188, -0.00001704, -0.00002431, -0.00000158, 0.00005342, 0.00001544, 0.00003643, 0.00003015, -0.00000015, -0.00003592, -0.00000894, -0.00000864, -0.00000558, -0.00001037, -0.00000584, -0.00002336, -0.00000895, -0.00011700, 0.00007318, 0.00013004, 0.00014160, 0.00000953, -0.00010469, 0.00000539, -0.00003648, 0.00002559, -0.00002602, 0.00001079, -0.00002079, 0.00000494, -0.00000453, -0.00000945, -0.00000355, -0.00000835, 0.00000075, 0.00000406, -0.00001475, 0.00000446, -0.00000846, 0.00000810, -0.00000854, -0.00000091, 0.00001167, 0.00000303, -0.00000539, 0.00000216, -0.00000159, 0.00000619, 0.00000259, -0.00000291, -0.00000376, -0.00000212, 0.00000656, -0.00000053, -0.00000439, -0.00000832, 0.00000886, 0.00000009, 0.00000127, 0.00000649, -0.00000117, -0.00000032, -0.00001111, 0.00000587, -0.00000303, 0.00000186, -0.00000330, 0.00000647, -0.00000775, 0.00000151, 0.00000454, -0.00000011, -0.00000204, 0.00000140, -0.00000412, -0.00000256, 0.00000255, 0.00000152, 0.00000473, -0.00000601, -0.00000325, -0.00000115, 0.00000930, -0.00000317, -0.00000124, -0.00000324, 0.00000070, 0.00000367, 0.00000507, -0.00000551, -0.00000899, 0.00000030, 0.00000290, 0.00000005, 0.00000443, -0.00000010, -0.00000108, 0.00000604, -0.00000328, -0.00000527, 0.00000533, -0.00000169, -0.00000550, -0.00000091, 0.00000267, 0.00000364, 0.00000258, -0.00000002, -0.00000165, -0.00000664, 0.00001141, 0.00000208, -0.00000590, 0.00000155, 0.00000776, -0.00000887, -0.00000958, 0.00000141, 0.00000419, -0.00000477, -0.00000178, -0.00000780, 0.00000902, -0.00002116, 0.00000537, -0.00000602, 0.00000282, },
		{ 0.00000774, -0.00000999, -0.00002915, -0.00000664, 0.00005206, 0.00002205, 0.00003235, 0.00003662, 0.00000116, -0.00003521, -0.00001228, -0.00000732, -0.00000666, -0.00000961, -0.00000539, -0.00002354, -0.00000524, -0.00011092, 0.00005102, 0.00012928, 0.00012810, 0.00004630, -0.00011321, 0.00000296, -0.00003690, 0.00002372, -0.00002327, 0.00000807, -0.00001782, 0.00000155, -0.00000186, -0.00001045, -0.00000327, -0.00000816, -0.00000108, 0.00000628, -0.00001505, 0.00000281, -0.00000622, 0.00000519, -0.00000508, -0.00000437, 0.00001196, 0.00000497, -0.00000610, 0.00000203, -0.00000167, 0.00000560, 0.00000312, -0.00000218, -0.00000352, -0.00000322, 0.00000609, 0.00000095, -0.00000443, -0.00000846, 0.00000667, 0.00000291, -0.00000104, 0.00000800, -0.00000209, 0.00000165, -0.00001209, 0.00000495, -0.00000226, 0.00000151, -0.00000302, 0.00000601, -0.00000629, -0.00000077, 0.00000568, -0.00000001, -0.00000191, 0.00000086, -0.00000289, -0.00000396, 0.00000339, 0.00000047, 0.00000546, -0.00000555, -0.00000314, -0.00000243, 0.00000910, -0.00000151, -0.00000204, -0.00000272, -0.00000032, 0.00000397, 0.00000440, -0.00000316, -0.00000978, -0.00000066, 0.00000270, 0.00000049, 0.00000355, 0.00000128, -0.00000161, 0.00000576, -0.00000210, -0.00000643, 0.00000502, -0.00000030, -0.00000577, -0.00000172, 0.00000293, 0.00000294, 0.00000349, -0.00000062, -0.00000002, -0.00000849, 0.00001113, 0.00000370, -0.00000559, -0.00000024, 0.00000788, -0.00000648, -0.00001004, -0.00000109, 0.00000593, -0.00000531, -0.00000053, -0.00001002, 0.00001034, -0.00002052, 0.00000387, -0.00000556, 0.00000286, -0.00002174, },
		{ -0.00000274, -0.00003254, -0.00001245, 0.00004941, 0.00002935, 0.00002842, 0.00004243, 0.00000339, -0.00003460, -0.00001563, -0.00000607, -0.00000788, -0.00000859, -0.00000549, -0.00002281, -0.00000354, -0.00010235, 0.00002873, 0.00012985, 0.00011230, 0.00007803, -0.00011301, -0.00000493, -0.00003390, 0.00001916, -0.00001854, 0.00000381, -0.00001367, -0.00000261, 0.00000118, -0.00001176, -0.00000258, -0.00000835, -0.00000234, 0.00000739, -0.00001386, -0.00000008, -0.00000331, 0.00000216, -0.00000201, -0.00000682, 0.00001102, 0.00000750, -0.00000681, 0.00000184, -0.00000170, 0.00000501, 0.00000354, -0.00000151, -0.00000310, -0.00000415, 0.00000524, 0.00000249, -0.00000441, -0.00000835, 0.00000429, 0.00000543, -0.00000273, 0.00000881, -0.00000238, 0.00000285, -0.00001202, 0.00000318, -0.00000100, 0.00000077, -0.00000234, 0.00000514, -0.00000448, -0.00000304, 0.00000650, 0.00000028, -0.00000171, 0.00000018, -0.00000163, -0.00000517, 0.00000392, -0.00000014, 0.00000566, -0.00000473, -0.00000325, -0.00000319, 0.00000815, 0.00000057, -0.00000299, -0.00000201, -0.00000150, 0.00000443, 0.00000352, -0.00000099, -0.00001006, -0.00000177, 0.00000238, 0.00000096, 0.00000276, 0.00000229, -0.00000168, 0.00000518, -0.00000078, -0.00000743, 0.00000443, 0.00000105, -0.00000569, -0.00000274, 0.00000326, 0.00000220, 0.00000432, -0.00000112, 0.00000137, -0.00000962, 0.00000991, 0.00000587, -0.00000557, -0.00000131, 0.00000707, -0.00000384, -0.00001016, -0.00000347, 0.00000692, -0.00000510, 0.00000006, -0.00001124, 0.00001030, -0.00001855, 0.00000129, -0.00000401, 0.00000152, -0.00001862, 0.00000229, },
		{ -0.00003378, -0.00001912, 0.00004559, 0.00003685, 0.00002518, 0.00004715, 0.00000683, -0.00003426, -0.00001879, -0.00000511, -0.00000902, -0.00000758, -0.00000585, -0.00002150, -0.00000330, -0.00009220, 0.00000730, 0.00013042, 0.00009663, 0.00010254, -0.00010383, -0.00001736, -0.00002846, 0.00001278, -0.00001257, -0.00000131, -0.00000894, -0.00000699, 0.00000419, -0.00001307, -0.00000175, -0.00000872, -0.00000320, 0.00000761, -0.00001157, -0.00000360, -0.00000043, -0.00000036, 0.00000019, -0.00000799, 0.00000887, 0.00001041, -0.00000734, 0.00000148, -0.00000162, 0.00000435, 0.00000394, -0.00000097, -0.00000250, -0.00000489, 0.00000411, 0.00000392, -0.00000419, -0.00000815, 0.00000197, 0.00000739, -0.00000362, 0.00000885, -0.00000202, 0.00000327, -0.00001099, 0.00000078, 0.00000049, -0.00000015, -0.00000149, 0.00000409, -0.00000258, -0.00000506, 0.00000683, 0.00000083, -0.00000151, -0.00000052, -0.00000051, -0.00000603, 0.00000400, -0.00000025, 0.00000535, -0.00000362, -0.00000356, -0.00000345, 0.00000660, 0.00000284, -0.00000389, -0.00000129, -0.00000266, 0.00000487, 0.00000268, 0.00000074, -0.00000973, -0.00000304, 0.00000203, 0.00000132, 0.00000220, 0.00000288, -0.00000132, 0.00000435, 0.00000058, -0.00000821, 0.00000358, 0.00000229, -0.00000528, -0.00000389, 0.00000355, 0.00000156, 0.00000493, -0.00000135, 0.00000232, -0.00000989, 0.00000774, 0.00000849, -0.00000576, -0.00000179, 0.00000562, -0.00000127, -0.00000984, -0.00000567, 0.00000718, -0.00000424, 0.00000010, -0.00001160, 0.00000913, -0.00001561, -0.00000196, -0.00000182, -0.00000067, -0.00001465, -0.00000350, 0.00000367, },
		{ -0.00002644, 0.00004056, 0.00004416, 0.00002307, 0.00005049, 0.00001160, -0.00003423, -0.00002168, -0.00000455, -0.00000991, -0.00000677, -0.00000624, -0.00001992, -0.00000400, -0.00008135, -0.00001234, 0.00012976, 0.00008302, 0.00011865, -0.00008646, -0.00003269, -0.00002200, 0.00000568, -0.00000630, -0.00000650, -0.00000434, -0.00001099, 0.00000671, -0.00001405, -0.00000108, -0.00000898, -0.00000390, 0.00000721, -0.00000867, -0.00000712, 0.00000180, -0.00000190, 0.00000121, -0.00000781, 0.00000574, 0.00001337, -0.00000745, 0.00000082, -0.00000136, 0.00000357, 0.00000437, -0.00000061, -0.00000172, -0.00000546, 0.00000281, 0.00000511, -0.00000370, -0.00000795, -0.00000010, 0.00000864, -0.00000368, 0.00000816, -0.00000110, 0.00000304, -0.00000922, -0.00000193, 0.00000192, -0.00000100, -0.00000067, 0.00000306, -0.00000082, -0.00000660, 0.00000658, 0.00000166, -0.00000133, -0.00000116, 0.00000034, -0.00000645, 0.00000361, 0.00000010, 0.00000464, -0.00000231, -0.00000401, -0.00000329, 0.00000465, 0.00000505, -0.00000451, -0.00000073, -0.00000363, 0.00000509, 0.00000208, 0.00000188, -0.00000876, -0.00000445, 0.00000168, 0.00000151, 0.00000191, 0.00000306, -0.00000062, 0.00000340, 0.00000187, -0.00000870, 0.00000248, 0.00000337, -0.00000456, -0.00000505, 0.00000367, 0.00000112, 0.00000521, -0.00000120, 0.00000270, -0.00000926, 0.00000474, 0.00001133, -0.00000596, -0.00000189, 0.00000387, 0.00000090, -0.00000895, -0.00000768, 0.00000678, -0.00000294, -0.00000021, -0.00001133, 0.00000718, -0.00001215, -0.00000539, 0.00000052, -0.00000313, -0.00001057, -0.00000875, 0.00000839, -0.00003253, },
		{ 0.00003418, 0.00005094, 0.00002236, 0.00005231, 0.00001767, -0.00003440, -0.00002429, -0.00000445, -0.00001049, -0.00000625, -0.00000651, -0.00001832, -0.00000517, -0.00007054, -0.00002946, 0.00012698, 0.00007272, 0.00012620, -0.00006262, -0.00004881, -0.00001604, -0.00000100, -0.00000062, -0.00001104, -0.00000048, -0.00001412, 0.00000837, -0.00001440, -0.00000081, -0.00000891, -0.00000464, 0.00000651, -0.00000564, -0.00001006, 0.00000292, -0.00000221, 0.00000103, -0.00000646, 0.00000199, 0.00001599, -0.00000690, -0.00000022, -0.00000086, 0.00000264, 0.00000487, -0.00000045, -0.00000082, -0.00000583, 0.00000142, 0.00000597, -0.00000290, -0.00000783, -0.00000183, 0.00000914, -0.00000301, 0.00000692, 0.00000020, 0.00000236, -0.00000703, -0.00000459, 0.00000300, -0.00000158, -0.00000007, 0.00000225, 0.00000061, -0.00000754, 0.00000572, 0.00000271, -0.00000117, -0.00000166, 0.00000087, -0.00000641, 0.00000280, 0.00000080, 0.00000367, -0.00000096, -0.00000449, -0.00000287, 0.00000255, 0.00000692, -0.00000467, -0.00000047, -0.00000427, 0.00000497, 0.00000187, 0.00000237, -0.00000723, -0.00000591, 0.00000131, 0.00000150, 0.00000189, 0.00000291, 0.00000027, 0.00000248, 0.00000300, -0.00000886, 0.00000117, 0.00000427, -0.00000360, -0.00000611, 0.00000353, 0.00000099, 0.00000508, -0.00000063, 0.00000251, -0.00000781, 0.00000120, 0.00001404, -0.00000589, -0.00000192, 0.00000222, 0.00000236, -0.00000743, -0.00000950, 0.00000585, -0.00000143, -0.00000062, -0.00001070, 0.00000482, -0.00000864, -0.00000854, 0.00000255, -0.00000533, -0.00000703, -0.00001279, 0.00001095, -0.00002885, -0.00003392, },
		{ 0.00005689, 0.00002321, 0.00005268, 0.00002482, -0.00003455, -0.00002674, -0.00000473, -0.00001079, -0.00000605, -0.00000659, -0.00001683, -0.00000649, -0.00006030, -0.00004360, 0.00012159, 0.00006629, 0.00012597, -0.00003466, -0.00006343, -0.00001203, -0.00000631, 0.00000377, -0.00001436, 0.00000217, -0.00001603, 0.00000894, -0.00001397, -0.00000112, -0.00000834, -0.00000558, 0.00000574, -0.00000287, -0.00001201, 0.00000270, -0.00000131, -0.00000016, -0.00000429, -0.00000190, 0.00001786, -0.00000551, -0.00000168, -0.00000013, 0.00000157, 0.00000543, -0.00000045, 0.00000015, -0.00000597, 0.00000001, 0.00000645, -0.00000181, -0.00000778, -0.00000319, 0.00000897, -0.00000179, 0.00000536, 0.00000165, 0.00000150, -0.00000473, -0.00000687, 0.00000349, -0.00000173, 0.00000018, 0.00000177, 0.00000161, -0.00000782, 0.00000434, 0.00000387, -0.00000097, -0.00000204, 0.00000107, -0.00000598, 0.00000170, 0.00000168, 0.00000265, 0.00000028, -0.00000489, -0.00000234, 0.00000054, 0.00000824, -0.00000426, -0.00000058, -0.00000453, 0.00000445, 0.00000208, 0.00000223, -0.00000529, -0.00000726, 0.00000087, 0.00000134, 0.00000205, 0.00000255, 0.00000119, 0.00000171, 0.00000388, -0.00000865, -0.00000032, 0.00000496, -0.00000249, -0.00000694, 0.00000303, 0.00000118, 0.00000456, 0.00000034, 0.00000182, -0.00000577, -0.00000252, 0.00001626, -0.00000526, -0.00000216, 0.00000098, 0.00000293, -0.00000531, -0.00001109, 0.00000454, 0.00000007, -0.00000092, -0.00000994, 0.00000241, -0.00000547, -0.00001104, 0.00000393, -0.00000689, -0.00000444, -0.00001525, 0.00001125, -0.00002320, -0.00004086, 0.00002628, },
		{ 0.00002568, 0.00005176, 0.00003275, -0.00003436, -0.00002922, -0.00000529, -0.00001089, -0.00000610, -0.00000651, -0.00001551, -0.00000777, -0.00005099, -0.00005458, 0.00011352, 0.00006360, 0.00011942, -0.00000523, -0.00007449, -0.00001102, -0.00000967, 0.00000645, -0.00001617, 0.00000340, -0.00001660, 0.00000839, -0.00001275, -0.00000204, -0.00000726, -0.00000677, 0.00000507, -0.00000062, -0.00001279, 0.00000119, 0.00000054, -0.00000199, -0.00000179, -0.00000540, 0.00001863, -0.00000320, -0.00000354, 0.00000078, 0.00000040, 0.00000601, -0.00000055, 0.00000108, -0.00000587, -0.00000137, 0.00000653, -0.00000049, -0.00000775, -0.00000424, 0.00000828, -0.00000027, 0.00000376, 0.00000298, 0.00000069, -0.00000263, -0.00000849, 0.00000325, -0.00000139, 0.00000006, 0.00000164, 0.00000216, -0.00000752, 0.00000257, 0.00000497, -0.00000067, -0.00000229, 0.00000100, -0.00000527, 0.00000048, 0.00000256, 0.00000174, 0.00000127, -0.00000512, -0.00000184, -0.00000116, 0.00000886, -0.00000326, -0.00000103, -0.00000441, 0.00000355, 0.00000266, 0.00000161, -0.00000316, -0.00000834, 0.00000029, 0.00000109, 0.00000228, 0.00000215, 0.00000199, 0.00000119, 0.00000446, -0.00000808, -0.00000193, 0.00000543, -0.00000129, -0.00000745, 0.00000217, 0.00000168, 0.00000369, 0.00000160, 0.00000080, -0.00000342, -0.00000601, 0.00001760, -0.00000385, -0.00000281, 0.00000036, 0.00000255, -0.00000271, -0.00001235, 0.00000293, 0.00000137, -0.00000097, -0.00000924, 0.00000017, -0.00000290, -0.00001271, 0.00000451, -0.00000765, -0.00000297, -0.00001609, 0.00000959, -0.00001635, -0.00004645, 0.00001674, 0.00006173, },
		{ 0.00004984, 0.00004108, -0.00003351, -0.00003191, -0.00000597, -0.00001093, -0.00000630, -0.00000635, -0.00001433, -0.00000893, -0.00004275, -0.00006249, 0.00010302, 0.00006408, 0.00010842, 0.00002311, -0.00008038, -0.00001352, -0.00001094, 0.00000739, -0.00001646, 0.00000326, -0.00001594, 0.00000687, -0.00001089, -0.00000345, -0.00000574, -0.00000816, 0.00000452, 0.00000101, -0.00001239, -0.00000137, 0.00000293, -0.00000396, 0.00000054, -0.00000807, 0.00001808, 0.00000001, -0.00000566, 0.00000174, -0.00000076, 0.00000652, -0.00000064, 0.00000187, -0.00000547, -0.00000270, 0.00000626, 0.00000094, -0.00000764, -0.00000505, 0.00000725, 0.00000130, 0.00000235, 0.00000402, 0.00000013, -0.00000093, -0.00000931, 0.00000228, -0.00000064, -0.00000038, 0.00000180, 0.00000236, -0.00000676, 0.00000065, 0.00000585, -0.00000018, -0.00000247, 0.00000073, -0.00000444, -0.00000071, 0.00000327, 0.00000109, 0.00000192, -0.00000511, -0.00000148, -0.00000244, 0.00000874, -0.00000172, -0.00000174, -0.00000402, 0.00000239, 0.00000349, 0.00000072, -0.00000110, -0.00000897, -0.00000052, 0.00000084, 0.00000246, 0.00000184, 0.00000254, 0.00000098, 0.00000472, -0.00000718, -0.00000360, 0.00000566, -0.00000008, -0.00000759, 0.00000096, 0.00000241, 0.00000261, 0.00000297, -0.00000032, -0.00000109, -0.00000890, 0.00001780, -0.00000153, -0.00000396, 0.00000042, 0.00000135, 0.00000010, -0.00001316, 0.00000108, 0.00000238, -0.00000071, -0.00000865, -0.00000175, -0.00000102, -0.00001353, 0.00000433, -0.00000764, -0.00000253, -0.00001556, 0.00000653, -0.00000921, -0.00004994, 0.00000560, 0.00006511, 0.00002973, },
		{ 0.00004941, -0.00003166, -0.00003496, -0.00000667, -0.00001100, -0.00000655, -0.00000619, -0.00001325, -0.00000993, -0.00003560, -0.00006759, 0.00009062, 0.00006680, 0.00009495, 0.00004813, -0.00008019, -0.00001937, -0.00001045, 0.00000690, -0.00001552, 0.00000203, -0.00001434, 0.00000470, -0.00000864, -0.00000516, -0.00000396, -0.00000958, 0.00000405, 0.00000204, -0.00001102, -0.00000455, 0.00000531, -0.00000558, 0.00000227, -0.00000960, 0.00001617, 0.00000394, -0.00000780, 0.00000259, -0.00000177, 0.00000684, -0.00000061, 0.00000242, -0.00000478, -0.00000396, 0.00000568, 0.00000236, -0.00000738, -0.00000574, 0.00000608, 0.00000270, 0.00000131, 0.00000461, -0.00000007, 0.00000025, -0.00000931, 0.00000067, 0.00000039, -0.00000099, 0.00000213, 0.00000232, -0.00000576, -0.00000121, 0.00000633, 0.00000056, -0.00000261, 0.00000036, -0.00000361, -0.00000170, 0.00000370, 0.00000077, 0.00000223, -0.00000487, -0.00000129, -0.00000322, 0.00000794, 0.00000018, -0.00000253, -0.00000349, 0.00000111, 0.00000438, -0.00000022, 0.00000063, -0.00000902, -0.00000157, 0.00000065, 0.00000250, 0.00000171, 0.00000280, 0.00000107, 0.00000472, -0.00000604, -0.00000522, 0.00000563, 0.00000109, -0.00000735, -0.00000051, 0.00000324, 0.00000147, 0.00000429, -0.00000132, 0.00000093, -0.00001087, 0.00001670, 0.00000166, -0.00000553, 0.00000104, -0.00000041, 0.00000281, -0.00001335, -0.00000097, 0.00000306, -0.00000017, -0.00000817, -0.00000335, 0.00000020, -0.00001363, 0.00000357, -0.00000707, -0.00000285, -0.00001410, 0.00000279, -0.00000270, -0.00005081, -0.00000683, 0.00006661, 0.00003531, 0.00004723, },
		{ -0.00002855, -0.00003848, -0.00000731, -0.00001119, -0.00000677, -0.00000610, -0.00001221, -0.00001081, -0.00002950, -0.00007026, 0.00007694, 0.00007076, 0.00008086, 0.00006824, -0.00007381, -0.00002784, -0.00000893, 0.00000553, -0.00001382, 0.00000012, -0.00001224, 0.00000226, -0.00000632, -0.00000693, -0.00000215, -0.00001086, 0.00000353, 0.00000260, -0.00000896, -0.00000788, 0.00000719, -0.00000644, 0.00000309, -0.00000984, 0.00001302, 0.00000825, -0.00000969, 0.00000314, -0.00000249, 0.00000687, -0.00000038, 0.00000267, -0.00000378, -0.00000512, 0.00000487, 0.00000365, -0.00000686, -0.00000639, 0.00000493, 0.00000376, 0.00000074, 0.00000474, 0.00000011, 0.00000091, -0.00000857, -0.00000135, 0.00000145, -0.00000160, 0.00000246, 0.00000223, -0.00000471, -0.00000279, 0.00000630, 0.00000156, -0.00000273, -0.00000004, -0.00000292, -0.00000241, 0.00000379, 0.00000079, 0.00000223, -0.00000445, -0.00000129, -0.00000355, 0.00000659, 0.00000223, -0.00000322, -0.00000298, -0.00000012, 0.00000516, -0.00000099, 0.00000186, -0.00000842, -0.00000287, 0.00000053, 0.00000237, 0.00000180, 0.00000277, 0.00000139, 0.00000453, -0.00000473, -0.00000670, 0.00000531, 0.00000219, -0.00000676, -0.00000212, 0.00000401, 0.00000043, 0.00000535, -0.00000199, 0.00000238, -0.00001175, 0.00001434, 0.00000553, -0.00000730, 0.00000197, -0.00000236, 0.00000505, -0.00001278, -0.00000320, 0.00000342, 0.00000053, -0.00000768, -0.00000472, 0.00000091, -0.00001325, 0.00000250, -0.00000623, -0.00000358, -0.00001223, -0.00000091, 0.00000244, -0.00004886, -0.00001997, 0.00006572, 0.00004228, 0.00004427, 0.00005737, },
		{ -0.00004246, -0.00000789, -0.00001152, -0.00000693, -0.00000611, -0.00001119, -0.00001156, -0.00002436, -0.00007090, 0.00006265, 0.00007499, 0.00006763, 0.00008252, -0.00006186, -0.00003770, -0.00000730, 0.00000395, -0.00001188, -0.00000199, -0.00001005, -0.00000005, -0.00000424, -0.00000851, -0.00000054, -0.00001181, 0.00000285, 0.00000284, -0.00000655, -0.00001086, 0.00000815, -0.00000631, 0.00000289, -0.00000886, 0.00000893, 0.00001256, -0.00001100, 0.00000318, -0.00000281, 0.00000653, 0.00000012, 0.00000258, -0.00000252, -0.00000614, 0.00000389, 0.00000471, -0.00000605, -0.00000706, 0.00000389, 0.00000441, 0.00000063, 0.00000445, 0.00000060, 0.00000112, -0.00000728, -0.00000351, 0.00000232, -0.00000202, 0.00000263, 0.00000223, -0.00000378, -0.00000392, 0.00000571, 0.00000279, -0.00000280, -0.00000041, -0.00000242, -0.00000282, 0.00000357, 0.00000108, 0.00000202, -0.00000393, -0.00000141, -0.00000350, 0.00000490, 0.00000417, -0.00000360, -0.00000265, -0.00000116, 0.00000566, -0.00000142, 0.00000246, -0.00000721, -0.00000434, 0.00000046, 0.00000206, 0.00000206, 0.00000255, 0.00000182, 0.00000426, -0.00000339, -0.00000795, 0.00000469, 0.00000320, -0.00000588, -0.00000373, 0.00000459, -0.00000036, 0.00000604, -0.00000219, 0.00000314, -0.00001149, 0.00001088, 0.00000974, -0.00000894, 0.00000288, -0.00000412, 0.00000650, -0.00001134, -0.00000558, 0.00000351, 0.00000127, -0.00000707, -0.00000596, 0.00000130, -0.00001266, 0.00000141, -0.00000540, -0.00000437, -0.00001041, -0.00000402, 0.00000577, -0.00004431, -0.00003295, 0.00006194, 0.00005044, 0.00004134, 0.00006464, -0.00002400, },
		{ -0.00000851, -0.00001194, -0.00000706, -0.00000622, -0.00001020, -0.00001218, -0.00002009, -0.00006990, 0.00004833, 0.00007869, 0.00005634, 0.00009081, -0.00004556, -0.00004750, -0.00000654, 0.00000276, -0.00001016, -0.00000392, -0.00000816, -0.00000194, -0.00000261, -0.00000974, 0.00000072, -0.00001228, 0.00000191, 0.00000292, -0.00000413, -0.00001312, 0.00000795, -0.00000514, 0.00000178, -0.00000692, 0.00000435, 0.00001638, -0.00001145, 0.00000259, -0.00000265, 0.00000579, 0.00000091, 0.00000218, -0.00000108, -0.00000697, 0.00000278, 0.00000548, -0.00000495, -0.00000777, 0.00000301, 0.00000468, 0.00000089, 0.00000389, 0.00000127, 0.00000105, -0.00000568, -0.00000552, 0.00000275, -0.00000211, 0.00000253, 0.00000244, -0.00000310, -0.00000454, 0.00000459, 0.00000413, -0.00000278, -0.00000074, -0.00000211, -0.00000295, 0.00000311, 0.00000152, 0.00000175, -0.00000342, -0.00000159, -0.00000318, 0.00000308, 0.00000577, -0.00000354, -0.00000258, -0.00000191, 0.00000581, -0.00000143, 0.00000243, -0.00000549, -0.00000581, 0.00000034, 0.00000164, 0.00000242, 0.00000225, 0.00000224, 0.00000402, -0.00000210, -0.00000888, 0.00000377, 0.00000410, -0.00000479, -0.00000520, 0.00000485, -0.00000081, 0.00000625, -0.00000185, 0.00000318, -0.00001022, 0.00000667, 0.00001385, -0.00001009, 0.00000342, -0.00000531, 0.00000693, -0.00000906, -0.00000802, 0.00000336, 0.00000193, -0.00000624, -0.00000717, 0.00000154, -0.00001208, 0.00000052, -0.00000478, -0.00000499, -0.00000895, -0.00000622, 0.00000722, -0.00003775, -0.00004472, 0.00005482, 0.00005950, 0.00003881, 0.00007093, -0.00001790, -0.00004682, },
		{ -0.00001237, -0.00000722, -0.00000636, -0.00000928, -0.00001263, -0.00001663, -0.00006763, 0.00003447, 0.00008127, 0.00004757, 0.00009352, -0.00002655, -0.00005573, -0.00000742, 0.00000238, -0.00000896, -0.00000539, -0.00000678, -0.00000326, -0.00000150, -0.00001057, 0.00000154, -0.00001223, 0.00000070, 0.00000294, -0.00000193, -0.00001442, 0.00000654, -0.00000308, 0.00000002, -0.00000440, -0.00000023, 0.00001931, -0.00001082, 0.00000131, -0.00000202, 0.00000467, 0.00000193, 0.00000154, 0.00000044, -0.00000756, 0.00000159, 0.00000592, -0.00000359, -0.00000846, 0.00000226, 0.00000463, 0.00000137, 0.00000321, 0.00000196, 0.00000086, -0.00000402, -0.00000711, 0.00000260, -0.00000179, 0.00000212, 0.00000289, -0.00000270, -0.00000467, 0.00000308, 0.00000543, -0.00000258, -0.00000106, -0.00000198, -0.00000290, 0.00000255, 0.00000197, 0.00000153, -0.00000302, -0.00000176, -0.00000273, 0.00000134, 0.00000686, -0.00000296, -0.00000281, -0.00000234, 0.00000558, -0.00000104, 0.00000187, -0.00000348, -0.00000711, 0.00000008, 0.00000121, 0.00000274, 0.00000201, 0.00000251, 0.00000391, -0.00000095, -0.00000944, 0.00000257, 0.00000487, -0.00000356, -0.00000641, 0.00000470, -0.00000087, 0.00000597, -0.00000101, 0.00000259, -0.00000816, 0.00000213, 0.00001740, -0.00001038, 0.00000330, -0.00000569, 0.00000627, -0.00000606, -0.00001038, 0.00000297, 0.00000243, -0.00000514, -0.00000841, 0.00000174, -0.00001165, -0.00000007, -0.00000445, -0.00000533, -0.00000797, -0.00000747, 0.00000707, -0.00003007, -0.00005420, 0.00004412, 0.00006895, 0.00003710, 0.00007602, -0.00001024, -0.00005141, -0.00000927, },
		{ -0.00000748, -0.00000648, -0.00000850, -0.00001288, -0.00001390, -0.00006438, 0.00002145, 0.00008238, 0.00004149, 0.00009153, -0.00000657, -0.00006112, -0.00001040, 0.00000295, -0.00000833, -0.00000636, -0.00000596, -0.00000399, -0.00000088, -0.00001102, 0.00000195, -0.00001170, -0.00000073, 0.00000294, -0.00000010, -0.00001467, 0.00000409, -0.00000048, -0.00000200, -0.00000174, -0.00000431, 0.00002098, -0.00000902, -0.00000061, -0.00000100, 0.00000326, 0.00000310, 0.00000077, 0.00000190, -0.00000784, 0.00000034, 0.00000605, -0.00000205, -0.00000906, 0.00000157, 0.00000441, 0.00000191, 0.00000258, 0.00000252, 0.00000070, -0.00000251, -0.00000810, 0.00000181, -0.00000109, 0.00000142, 0.00000355, -0.00000255, -0.00000438, 0.00000135, 0.00000652, -0.00000212, -0.00000140, -0.00000194, -0.00000278, 0.00000203, 0.00000230, 0.00000148, -0.00000278, -0.00000187, -0.00000224, -0.00000016, 0.00000734, -0.00000186, -0.00000328, -0.00000251, 0.00000505, -0.00000032, 0.00000095, -0.00000142, -0.00000803, -0.00000044, 0.00000087, 0.00000291, 0.00000195, 0.00000256, 0.00000395, 0.00000004, -0.00000963, 0.00000114, 0.00000547, -0.00000227, -0.00000728, 0.00000413, -0.00000057, 0.00000527, 0.00000024, 0.00000154, -0.00000565, -0.00000227, 0.00001995, -0.00000955, 0.00000237, -0.00000517, 0.00000462, -0.00000262, -0.00001244, 0.00000231, 0.00000274, -0.00000379, -0.00000965, 0.00000192, -0.00001140, -0.00000039, -0.00000435, -0.00000547, -0.00000737, -0.00000798, 0.00000586, -0.00002233, -0.00006044, 0.00002991, 0.00007809, 0.00003666, 0.00007971, -0.00000107, -0.00005602, -0.00001033, -0.00001272, },
		},
		{ { -0.00000147, -0.00000215, -0.00000515, -0.00000281, -0.00000547, -0.00000545, -0.00000988, -0.00001493, -0.00006230, 0.00001005, 0.00008679, 0.00003825, 0.00008451, 0.00001149, -0.00005947, -0.00000948, -0.00000785, -0.00000521, -0.00001114, -0.00000329, 0.00000312, 0.00000021, -0.00000401, -0.00000320, 0.00001098, -0.00000544, -0.00000144, -0.00000330, 0.00000081, 0.00000231, -0.00000685, 0.00000177, -0.00000223, -0.00000440, -0.00000349, 0.00000010, -0.00001613, 0.00000134, 0.00000499, 0.00000088, -0.00000737, -0.00000272, 0.00001597, -0.00000366, -0.00000727, 0.00000089, 0.00000440, -0.00000164, -0.00000106, 0.00000036, 0.00001051, -0.00000341, -0.00000557, -0.00000241, -0.00000116, 0.00000558, 0.00000687, 0.00000243, -0.00001683, 0.00000172, 0.00001140, -0.00000012, -0.00000188, -0.00000105, -0.00000075, -0.00000105, -0.00000188, -0.00000012, 0.00001140, 0.00000172, -0.00001683, 0.00000243, 0.00000687, 0.00000558, -0.00000116, -0.00000241, -0.00000557, -0.00000341, 0.00001051, 0.00000036, -0.00000106, -0.00000164, 0.00000440, 0.00000089, -0.00000727, -0.00000366, 0.00001597, -0.00000272, -0.00000737, 0.00000088, 0.00000499, 0.00000134, -0.00001613, 0.00000010, -0.00000349, -0.00000440, -0.00000223, 0.00000177, -0.00000685, 0.00000231, 0.00000081, -0.00000330, -0.00000144, -0.00000544, 0.00001098, -0.00000320, -0.00000401, 0.00000021, 0.00000312, -0.00000329, -0.00001114, -0.00000521, -0.00000785, -0.00000948, -0.00005947, 0.00001149, 0.00008451, 0.00003825, 0.00008679, 0.00001005, -0.00006230, -0.00001493, -0.00000988, -0.00000545, -0.00000547, -0.00000281, -0.00000515, -0.00000215, },
		{ -0.00000170, -0.00000485, -0.00000340, -0.00000455, -0.00000546, -0.00000929, -0.00001164, -0.00005588, -0.00000903, 0.00008431, 0.00003928, 0.00007341, 0.00003650, -0.00005717, -0.00001796, -0.00000552, -0.00000661, -0.00000940, -0.00000621, 0.00000379, 0.00000021, -0.00000266, -0.00000539, 0.00001123, -0.00000331, -0.00000331, -0.00000202, -0.00000084, 0.00000369, -0.00000733, 0.00000191, -0.00000176, -0.00000373, -0.00000501, 0.00000150, -0.00001413, -0.00000357, 0.00000726, 0.00000035, -0.00000504, -0.00000663, 0.00001538, 0.00000071, -0.00000852, -0.00000056, 0.00000447, -0.00000050, -0.00000160, -0.00000039, 0.00000980, -0.00000032, -0.00000717, -0.00000163, -0.00000261, 0.00000527, 0.00000616, 0.00000529, -0.00001532, -0.00000345, 0.00001267, 0.00000180, -0.00000266, -0.00000115, -0.00000055, -0.00000135, -0.00000100, -0.00000125, 0.00000914, 0.00000628, -0.00001645, -0.00000146, 0.00000773, 0.00000542, 0.00000077, -0.00000341, -0.00000381, -0.00000606, 0.00001040, 0.00000150, -0.00000073, -0.00000220, 0.00000361, 0.00000239, -0.00000582, -0.00000676, 0.00001418, 0.00000253, -0.00000993, 0.00000175, 0.00000243, 0.00000530, -0.00001607, -0.00000282, -0.00000152, -0.00000508, -0.00000254, 0.00000144, -0.00000627, 0.00000080, 0.00000248, -0.00000442, -0.00000001, -0.00000663, 0.00000974, -0.00000058, -0.00000533, 0.00000033, 0.00000207, -0.00000047, -0.00001219, -0.00000463, -0.00000914, -0.00000460, -0.00005407, -0.00001631, 0.00008994, 0.00004257, 0.00008424, 0.00003185, -0.00006655, -0.00001972, -0.00001027, -0.00000572, -0.00000602, -0.00000284, -0.00000481, -0.00000299, -0.00000140, },
		{ -0.00000406, -0.00000429, -0.00000363, -0.00000541, -0.00000875, -0.00000937, -0.00004850, -0.00002434, 0.00007693, 0.00004418, 0.00005996, 0.00005564, -0.00004713, -0.00002827, -0.00000359, -0.00000780, -0.00000786, -0.00000841, 0.00000353, 0.00000069, -0.00000166, -0.00000674, 0.00001022, -0.00000026, -0.00000548, -0.00000074, -0.00000239, 0.00000486, -0.00000759, 0.00000169, -0.00000101, -0.00000328, -0.00000580, 0.00000141, -0.00001073, -0.00000838, 0.00000836, 0.00000074, -0.00000347, -0.00000880, 0.00001263, 0.00000562, -0.00000905, -0.00000210, 0.00000399, 0.00000090, -0.00000203, -0.00000099, 0.00000862, 0.00000272, -0.00000814, -0.00000144, -0.00000334, 0.00000434, 0.00000580, 0.00000697, -0.00001226, -0.00000851, 0.00001258, 0.00000431, -0.00000306, -0.00000174, -0.00000006, -0.00000181, -0.00000034, -0.00000153, 0.00000634, 0.00000968, -0.00001417, -0.00000591, 0.00000834, 0.00000510, 0.00000274, -0.00000418, -0.00000239, -0.00000781, 0.00000921, 0.00000318, -0.00000081, -0.00000201, 0.00000211, 0.00000395, -0.00000451, -0.00000827, 0.00001035, 0.00000826, -0.00001188, 0.00000216, 0.00000055, 0.00000742, -0.00001369, -0.00000692, 0.00000047, -0.00000552, -0.00000284, 0.00000103, -0.00000560, -0.00000090, 0.00000418, -0.00000536, 0.00000098, -0.00000707, 0.00000790, 0.00000198, -0.00000619, 0.00000016, 0.00000118, 0.00000156, -0.00001186, -0.00000550, -0.00000862, -0.00000375, -0.00004275, -0.00004268, 0.00008658, 0.00005281, 0.00007737, 0.00005476, -0.00006727, -0.00002639, -0.00001028, -0.00000652, -0.00000599, -0.00000359, -0.00000392, -0.00000407, -0.00000125, -0.00000158, },
		{ -0.00000507, -0.00000311, -0.00000498, -0.00000850, -0.00000768, -0.00004112, -0.00003540, 0.00006544, 0.00005107, 0.00004701, 0.00006737, -0.00003096, -0.00003790, -0.00000358, -0.00000791, -0.00000724, -0.00000941, 0.00000213, 0.00000173, -0.00000116, -0.00000710, 0.00000802, 0.00000341, -0.00000757, 0.00000026, -0.00000362, 0.00000563, -0.00000740, 0.00000091, 0.00000015, -0.00000316, -0.00000582, 0.00000015, -0.00000678, -0.00001216, 0.00000777, 0.00000216, -0.00000270, -0.00000937, 0.00000843, 0.00001009, -0.00000836, -0.00000384, 0.00000317, 0.00000218, -0.00000207, -0.00000164, 0.00000728, 0.00000531, -0.00000821, -0.00000199, -0.00000331, 0.00000287, 0.00000576, 0.00000759, -0.00000826, -0.00001269, 0.00001091, 0.00000714, -0.00000285, -0.00000274, 0.00000048, -0.00000211, -0.00000017, -0.00000104, 0.00000348, 0.00001158, -0.00001036, -0.00001024, 0.00000823, 0.00000498, 0.00000435, -0.00000431, -0.00000167, -0.00000840, 0.00000691, 0.00000534, -0.00000129, -0.00000114, 0.00000014, 0.00000540, -0.00000341, -0.00000835, 0.00000534, 0.00001330, -0.00001230, 0.00000136, 0.00000007, 0.00000732, -0.00000929, -0.00001145, 0.00000188, -0.00000546, -0.00000323, 0.00000057, -0.00000480, -0.00000283, 0.00000585, -0.00000605, 0.00000154, -0.00000692, 0.00000579, 0.00000414, -0.00000636, -0.00000047, 0.00000075, 0.00000246, -0.00001000, -0.00000775, -0.00000652, -0.00000595, -0.00002874, -0.00006312, 0.00007243, 0.00006814, 0.00006793, 0.00007666, -0.00006313, -0.00003518, -0.00000987, -0.00000795, -0.00000536, -0.00000491, -0.00000274, -0.00000508, -0.00000124, -0.00000163, -0.00000311, },
		{ -0.00000327, -0.00000404, -0.00000861, -0.00000629, -0.00003434, -0.00004229, 0.00005109, 0.00005807, 0.00003662, 0.00007163, -0.00001132, -0.00004439, -0.00000644, -0.00000666, -0.00000764, -0.00000926, -0.00000015, 0.00000304, -0.00000097, -0.00000670, 0.00000507, 0.00000710, -0.00000901, 0.00000056, -0.00000423, 0.00000574, -0.00000649, -0.00000061, 0.00000177, -0.00000336, -0.00000518, -0.00000181, -0.00000307, -0.00001427, 0.00000547, 0.00000430, -0.00000241, -0.00000884, 0.00000373, 0.00001323, -0.00000615, -0.00000575, 0.00000215, 0.00000307, -0.00000155, -0.00000241, 0.00000594, 0.00000724, -0.00000733, -0.00000316, -0.00000273, 0.00000114, 0.00000581, 0.00000752, -0.00000405, -0.00001537, 0.00000771, 0.00000988, -0.00000193, -0.00000391, 0.00000074, -0.00000199, -0.00000058, -0.00000001, 0.00000100, 0.00001191, -0.00000558, -0.00001372, 0.00000703, 0.00000529, 0.00000531, -0.00000358, -0.00000178, -0.00000784, 0.00000373, 0.00000770, -0.00000195, 0.00000010, -0.00000189, 0.00000642, -0.00000235, -0.00000749, 0.00000021, 0.00001655, -0.00001052, -0.00000103, 0.00000120, 0.00000525, -0.00000375, -0.00001538, 0.00000206, -0.00000468, -0.00000379, 0.00000011, -0.00000387, -0.00000492, 0.00000734, -0.00000630, 0.00000157, -0.00000623, 0.00000356, 0.00000582, -0.00000592, -0.00000142, 0.00000067, 0.00000244, -0.00000714, -0.00001055, -0.00000394, -0.00000923, -0.00001576, -0.00007404, 0.00004743, 0.00008602, 0.00005858, 0.00009519, -0.00005295, -0.00004599, -0.00000920, -0.00000994, -0.00000429, -0.00000650, -0.00000167, -0.00000571, -0.00000154, -0.00000164, -0.00000230, -0.00000536, },
		{ -0.00000268, -0.00000895, -0.00000516, -0.00002840, -0.00004552, 0.00003536, 0.00006352, 0.00002996, 0.00006946, 0.00000893, -0.00004603, -0.00001207, -0.00000460, -0.00000854, -0.00000854, -0.00000262, 0.00000401, -0.00000062, -0.00000603, 0.00000201, 0.00001006, -0.00000918, -0.00000024, -0.00000394, 0.00000497, -0.00000471, -0.00000287, 0.00000370, -0.00000375, -0.00000414, -0.00000393, -0.00000025, -0.00001452, 0.00000187, 0.00000652, -0.00000210, -0.00000784, -0.00000056, 0.00001441, -0.00000252, -0.00000754, 0.00000095, 0.00000346, -0.00000054, -0.00000322, 0.00000463, 0.00000849, -0.00000567, -0.00000464, -0.00000198, -0.00000046, 0.00000567, 0.00000716, -0.00000028, -0.00001620, 0.00000329, 0.00001211, -0.00000029, -0.00000497, 0.00000048, -0.00000134, -0.00000147, 0.00000121, -0.00000075, 0.00001082, -0.00000055, -0.00001579, 0.00000460, 0.00000606, 0.00000557, -0.00000204, -0.00000256, -0.00000644, 0.00000015, 0.00000980, -0.00000239, 0.00000126, -0.00000345, 0.00000667, -0.00000107, -0.00000631, -0.00000400, 0.00001725, -0.00000642, -0.00000479, 0.00000351, 0.00000210, 0.00000167, -0.00001764, 0.00000056, -0.00000311, -0.00000451, -0.00000033, -0.00000287, -0.00000698, 0.00000832, -0.00000584, 0.00000089, -0.00000493, 0.00000115, 0.00000719, -0.00000521, -0.00000230, 0.00000055, 0.00000205, -0.00000419, -0.00001273, -0.00000232, -0.00001145, -0.00000693, -0.00007377, 0.00001399, 0.00010232, 0.00005250, 0.00010807, -0.00003604, -0.00005834, -0.00000863, -0.00001227, -0.00000309, -0.00000801, -0.00000104, -0.00000577, -0.00000225, -0.00000150, -0.00000185, -0.00000500, -0.00000412, },
		{ -0.00000918, -0.00000442, -0.00002322, -0.00004595, 0.00001978, 0.00006614, 0.00002742, 0.00006250, 0.00002731, -0.00004236, -0.00001933, -0.00000290, -0.00000903, -0.00000812, -0.00000445, 0.00000399, 0.00000034, -0.00000556, -0.00000051, 0.00001165, -0.00000770, -0.00000227, -0.00000273, 0.00000330, -0.00000213, -0.00000565, 0.00000563, -0.00000401, -0.00000305, -0.00000571, 0.00000132, -0.00001308, -0.00000234, 0.00000814, -0.00000130, -0.00000688, -0.00000383, 0.00001351, 0.00000204, -0.00000876, -0.00000049, 0.00000339, 0.00000076, -0.00000382, 0.00000325, 0.00000918, -0.00000355, -0.00000605, -0.00000144, -0.00000161, 0.00000508, 0.00000684, 0.00000263, -0.00001519, -0.00000176, 0.00001336, 0.00000202, -0.00000566, -0.00000040, -0.00000029, -0.00000255, 0.00000220, -0.00000152, 0.00000863, 0.00000405, -0.00001608, 0.00000109, 0.00000708, 0.00000529, 0.00000001, -0.00000365, -0.00000469, -0.00000323, 0.00001111, -0.00000222, 0.00000190, -0.00000415, 0.00000590, 0.00000062, -0.00000533, -0.00000664, 0.00001527, -0.00000053, -0.00000911, 0.00000602, -0.00000089, 0.00000563, -0.00001740, -0.00000275, -0.00000094, -0.00000528, -0.00000074, -0.00000189, -0.00000873, 0.00000842, -0.00000436, -0.00000068, -0.00000292, -0.00000152, 0.00000848, -0.00000460, -0.00000269, -0.00000002, 0.00000192, -0.00000205, -0.00001325, -0.00000276, -0.00001113, -0.00000381, -0.00006322, -0.00002314, 0.00011199, 0.00005267, 0.00011373, -0.00001250, -0.00007120, -0.00000877, -0.00001461, -0.00000212, -0.00000909, -0.00000107, -0.00000521, -0.00000327, -0.00000126, -0.00000174, -0.00000408, -0.00000539, -0.00000126, },
		{ -0.00000436, -0.00001851, -0.00004460, 0.00000572, 0.00006513, 0.00002877, 0.00005255, 0.00004218, -0.00003404, -0.00002639, -0.00000290, -0.00000825, -0.00000869, -0.00000508, 0.00000265, 0.00000209, -0.00000556, -0.00000212, 0.00001160, -0.00000459, -0.00000528, -0.00000090, 0.00000108, 0.00000086, -0.00000842, 0.00000702, -0.00000375, -0.00000230, -0.00000676, 0.00000152, -0.00001041, -0.00000640, 0.00000864, 0.00000016, -0.00000618, -0.00000588, 0.00001088, 0.00000670, -0.00000889, -0.00000228, 0.00000297, 0.00000206, -0.00000399, 0.00000171, 0.00000948, -0.00000128, -0.00000703, -0.00000137, -0.00000214, 0.00000402, 0.00000668, 0.00000456, -0.00001273, -0.00000668, 0.00001321, 0.00000484, -0.00000577, -0.00000181, 0.00000086, -0.00000342, 0.00000259, -0.00000125, 0.00000581, 0.00000766, -0.00001457, -0.00000307, 0.00000795, 0.00000481, 0.00000215, -0.00000456, -0.00000312, -0.00000586, 0.00001125, -0.00000116, 0.00000177, -0.00000378, 0.00000412, 0.00000271, -0.00000478, -0.00000755, 0.00001114, 0.00000603, -0.00001280, 0.00000755, -0.00000249, 0.00000716, -0.00001440, -0.00000748, 0.00000134, -0.00000585, -0.00000118, -0.00000104, -0.00000990, 0.00000734, -0.00000173, -0.00000313, -0.00000029, -0.00000443, 0.00000977, -0.00000428, -0.00000242, -0.00000122, 0.00000237, -0.00000121, -0.00001169, -0.00000551, -0.00000803, -0.00000597, -0.00004578, -0.00005772, 0.00011026, 0.00006092, 0.00011184, 0.00001649, -0.00008292, -0.00001048, -0.00001648, -0.00000175, -0.00000951, -0.00000181, -0.00000421, -0.00000434, -0.00000113, -0.00000182, -0.00000292, -0.00000662, -0.00000029, -0.00000887, },
		{ -0.00001399, -0.00004237, -0.00000581, 0.00006026, 0.00003325, 0.00004137, 0.00005266, -0.00002253, -0.00003139, -0.00000545, -0.00000594, -0.00001036, -0.00000451, 0.00000018, 0.00000435, -0.00000585, -0.00000291, 0.00001015, -0.00000038, -0.00000858, 0.00000088, -0.00000109, 0.00000359, -0.00001049, 0.00000730, -0.00000259, -0.00000220, -0.00000684, 0.00000045, -0.00000706, -0.00000969, 0.00000783, 0.00000216, -0.00000566, -0.00000687, 0.00000721, 0.00001056, -0.00000751, -0.00000440, 0.00000230, 0.00000311, -0.00000358, -0.00000000, 0.00000946, 0.00000091, -0.00000742, -0.00000184, -0.00000211, 0.00000264, 0.00000658, 0.00000568, -0.00000944, -0.00001070, 0.00001147, 0.00000791, -0.00000521, -0.00000349, 0.00000174, -0.00000374, 0.00000219, -0.00000011, 0.00000286, 0.00000991, -0.00001152, -0.00000727, 0.00000824, 0.00000447, 0.00000395, -0.00000488, -0.00000214, -0.00000737, 0.00001004, 0.00000080, 0.00000086, -0.00000245, 0.00000160, 0.00000494, -0.00000456, -0.00000707, 0.00000587, 0.00001190, -0.00001465, 0.00000711, -0.00000190, 0.00000597, -0.00000913, -0.00001269, 0.00000300, -0.00000588, -0.00000176, -0.00000035, -0.00001032, 0.00000502, 0.00000189, -0.00000614, 0.00000260, -0.00000722, 0.00001082, -0.00000406, -0.00000172, -0.00000279, 0.00000322, -0.00000147, -0.00000851, -0.00000978, -0.00000332, -0.00001117, -0.00002639, -0.00008343, 0.00009398, 0.00007697, 0.00010377, 0.00004856, -0.00009126, -0.00001480, -0.00001737, -0.00000237, -0.00000911, -0.00000316, -0.00000305, -0.00000517, -0.00000131, -0.00000183, -0.00000191, -0.00000729, -0.00000022, -0.00000769, -0.00000516, },
		{ -0.00003987, -0.00001436, 0.00005196, 0.00003964, 0.00003061, 0.00005851, -0.00000949, -0.00003307, -0.00001049, -0.00000263, -0.00001254, -0.00000339, -0.00000263, 0.00000637, -0.00000587, -0.00000336, 0.00000795, 0.00000401, -0.00001120, 0.00000178, -0.00000249, 0.00000536, -0.00001118, 0.00000601, -0.00000032, -0.00000286, -0.00000594, -0.00000157, -0.00000364, -0.00001180, 0.00000583, 0.00000431, -0.00000504, -0.00000718, 0.00000332, 0.00001286, -0.00000456, -0.00000659, 0.00000136, 0.00000377, -0.00000260, -0.00000176, 0.00000911, 0.00000292, -0.00000722, -0.00000271, -0.00000176, 0.00000122, 0.00000634, 0.00000630, -0.00000599, -0.00001321, 0.00000817, 0.00001087, -0.00000393, -0.00000514, 0.00000200, -0.00000331, 0.00000103, 0.00000154, 0.00000032, 0.00001063, -0.00000742, -0.00001088, 0.00000758, 0.00000452, 0.00000513, -0.00000438, -0.00000192, -0.00000770, 0.00000765, 0.00000339, -0.00000055, -0.00000054, -0.00000114, 0.00000686, -0.00000435, -0.00000585, 0.00000067, 0.00001581, -0.00001382, 0.00000430, 0.00000089, 0.00000261, -0.00000279, -0.00001709, 0.00000323, -0.00000504, -0.00000262, 0.00000021, -0.00001000, 0.00000168, 0.00000598, -0.00000908, 0.00000510, -0.00000931, 0.00001113, -0.00000343, -0.00000118, -0.00000409, 0.00000385, -0.00000209, -0.00000487, -0.00001396, 0.00000099, -0.00001628, -0.00001022, -0.00009557, 0.00006290, 0.00009798, 0.00009263, 0.00008034, -0.00009366, -0.00002276, -0.00001679, -0.00000428, -0.00000783, -0.00000498, -0.00000198, -0.00000554, -0.00000193, -0.00000163, -0.00000133, -0.00000709, -0.00000125, -0.00000557, -0.00000677, -0.00000956, },
		{ -0.00002009, 0.00004124, 0.00004641, 0.00002178, 0.00005978, 0.00000375, -0.00003120, -0.00001691, 0.00000035, -0.00001407, -0.00000281, -0.00000471, 0.00000720, -0.00000490, -0.00000413, 0.00000585, 0.00000756, -0.00001219, 0.00000109, -0.00000253, 0.00000565, -0.00001010, 0.00000303, 0.00000295, -0.00000413, -0.00000431, -0.00000405, -0.00000070, -0.00001256, 0.00000300, 0.00000612, -0.00000404, -0.00000717, -0.00000013, 0.00001324, -0.00000036, -0.00000840, 0.00000005, 0.00000404, -0.00000122, -0.00000332, 0.00000832, 0.00000479, -0.00000656, -0.00000372, -0.00000141, 0.00000011, 0.00000572, 0.00000675, -0.00000294, -0.00001396, 0.00000371, 0.00001322, -0.00000193, -0.00000648, 0.00000150, -0.00000222, -0.00000058, 0.00000320, -0.00000136, 0.00000988, -0.00000285, -0.00001336, 0.00000585, 0.00000501, 0.00000562, -0.00000310, -0.00000238, -0.00000705, 0.00000448, 0.00000614, -0.00000201, 0.00000141, -0.00000350, 0.00000801, -0.00000376, -0.00000460, -0.00000341, 0.00001700, -0.00001019, -0.00000052, 0.00000511, -0.00000162, 0.00000302, -0.00001936, 0.00000141, -0.00000316, -0.00000381, 0.00000069, -0.00000911, -0.00000221, 0.00000980, -0.00001111, 0.00000642, -0.00000999, 0.00001005, -0.00000179, -0.00000148, -0.00000437, 0.00000350, -0.00000208, -0.00000217, -0.00001632, 0.00000283, -0.00001840, -0.00000107, -0.00009242, 0.00002032, 0.00011869, 0.00008268, 0.00010803, -0.00008770, -0.00003503, -0.00001448, -0.00000763, -0.00000571, -0.00000708, -0.00000117, -0.00000543, -0.00000289, -0.00000126, -0.00000125, -0.00000601, -0.00000321, -0.00000281, -0.00000884, -0.00000544, -0.00003728, },
		{ 0.00002956, 0.00005184, 0.00001615, 0.00005671, 0.00001637, -0.00002649, -0.00002290, 0.00000142, -0.00001374, -0.00000382, -0.00000512, 0.00000613, -0.00000245, -0.00000570, 0.00000451, 0.00000951, -0.00001100, -0.00000146, -0.00000108, 0.00000437, -0.00000733, -0.00000130, 0.00000669, -0.00000554, -0.00000244, -0.00000639, 0.00000129, -0.00001201, -0.00000020, 0.00000723, -0.00000256, -0.00000705, -0.00000276, 0.00001178, 0.00000438, -0.00000925, -0.00000176, 0.00000400, 0.00000029, -0.00000441, 0.00000700, 0.00000654, -0.00000562, -0.00000460, -0.00000133, -0.00000050, 0.00000466, 0.00000717, -0.00000057, -0.00001309, -0.00000123, 0.00001445, 0.00000077, -0.00000733, 0.00000030, -0.00000078, -0.00000218, 0.00000435, -0.00000189, 0.00000792, 0.00000157, -0.00001437, 0.00000313, 0.00000576, 0.00000555, -0.00000126, -0.00000321, -0.00000582, 0.00000108, 0.00000849, -0.00000301, 0.00000288, -0.00000499, 0.00000805, -0.00000249, -0.00000384, -0.00000577, 0.00001538, -0.00000440, -0.00000626, 0.00000938, -0.00000510, 0.00000670, -0.00001856, -0.00000261, -0.00000044, -0.00000519, 0.00000109, -0.00000784, -0.00000608, 0.00001253, -0.00001146, 0.00000591, -0.00000872, 0.00000718, 0.00000128, -0.00000307, -0.00000317, 0.00000166, -0.00000074, -0.00000140, -0.00001569, 0.00000098, -0.00001602, -0.00000033, -0.00007584, -0.00002729, 0.00013240, 0.00007834, 0.00012830, -0.00007178, -0.00005158, -0.00001068, -0.00001228, -0.00000298, -0.00000929, -0.00000068, -0.00000498, -0.00000392, -0.00000095, -0.00000146, -0.00000437, -0.00000558, -0.00000005, -0.00001072, -0.00000211, -0.00003436, -0.00002367, },
		{ 0.00005443, 0.00001461, 0.00004970, 0.00002798, -0.00002026, -0.00002656, -0.00000068, -0.00001084, -0.00000688, -0.00000353, 0.00000306, 0.00000141, -0.00000804, 0.00000408, 0.00000968, -0.00000770, -0.00000551, 0.00000140, 0.00000200, -0.00000348, -0.00000612, 0.00001006, -0.00000638, -0.00000096, -0.00000800, 0.00000201, -0.00001032, -0.00000336, 0.00000745, -0.00000069, -0.00000683, -0.00000454, 0.00000899, 0.00000875, -0.00000860, -0.00000411, 0.00000371, 0.00000168, -0.00000482, 0.00000515, 0.00000816, -0.00000452, -0.00000521, -0.00000162, -0.00000061, 0.00000326, 0.00000754, 0.00000115, -0.00001107, -0.00000581, 0.00001412, 0.00000405, -0.00000760, -0.00000134, 0.00000055, -0.00000326, 0.00000460, -0.00000122, 0.00000518, 0.00000528, -0.00001381, -0.00000025, 0.00000649, 0.00000518, 0.00000078, -0.00000402, -0.00000447, -0.00000201, 0.00000998, -0.00000320, 0.00000354, -0.00000534, 0.00000690, -0.00000055, -0.00000377, -0.00000634, 0.00001158, 0.00000229, -0.00001143, 0.00001216, -0.00000637, 0.00000725, -0.00001452, -0.00000830, 0.00000252, -0.00000641, 0.00000129, -0.00000637, -0.00000942, 0.00001358, -0.00000970, 0.00000335, -0.00000546, 0.00000257, 0.00000564, -0.00000586, -0.00000060, -0.00000155, 0.00000196, -0.00000275, -0.00001201, -0.00000438, -0.00000950, -0.00000656, -0.00005091, -0.00007153, 0.00013262, 0.00008273, 0.00013925, -0.00004574, -0.00007122, -0.00000628, -0.00001763, -0.00000011, -0.00001131, -0.00000051, -0.00000445, -0.00000467, -0.00000101, -0.00000166, -0.00000269, -0.00000765, 0.00000196, -0.00001168, -0.00000012, -0.00003071, -0.00002607, 0.00001849, },
		{ 0.00001737, 0.00003953, 0.00003830, -0.00001375, -0.00002666, -0.00000617, -0.00000566, -0.00001157, -0.00000047, -0.00000134, 0.00000591, -0.00001054, 0.00000418, 0.00000850, -0.00000304, -0.00001011, 0.00000397, -0.00000054, 0.00000043, -0.00001033, 0.00001204, -0.00000594, -0.00000044, -0.00000841, 0.00000132, -0.00000773, -0.00000617, 0.00000685, 0.00000124, -0.00000632, -0.00000566, 0.00000556, 0.00001191, -0.00000621, -0.00000680, 0.00000311, 0.00000280, -0.00000452, 0.00000291, 0.00000949, -0.00000323, -0.00000554, -0.00000219, -0.00000042, 0.00000183, 0.00000761, 0.00000250, -0.00000859, -0.00000924, 0.00001200, 0.00000768, -0.00000726, -0.00000308, 0.00000133, -0.00000350, 0.00000382, 0.00000039, 0.00000224, 0.00000783, -0.00001180, -0.00000384, 0.00000686, 0.00000476, 0.00000267, -0.00000447, -0.00000336, -0.00000438, 0.00001035, -0.00000245, 0.00000328, -0.00000459, 0.00000480, 0.00000183, -0.00000422, -0.00000558, 0.00000668, 0.00000838, -0.00001459, 0.00001228, -0.00000465, 0.00000456, -0.00000801, -0.00001445, 0.00000474, -0.00000695, 0.00000104, -0.00000473, -0.00001194, 0.00001269, -0.00000591, -0.00000091, -0.00000074, -0.00000309, 0.00001056, -0.00000908, 0.00000257, -0.00000533, 0.00000527, -0.00000541, -0.00000642, -0.00001161, -0.00000101, -0.00001615, -0.00002453, -0.00010412, 0.00011503, 0.00009638, 0.00014122, -0.00001126, -0.00009152, -0.00000291, -0.00002264, 0.00000217, -0.00001276, -0.00000072, -0.00000404, -0.00000488, -0.00000166, -0.00000154, -0.00000144, -0.00000879, 0.00000259, -0.00001118, 0.00000010, -0.00002595, -0.00002816, 0.00000933, 0.00005316, },
		{ 0.00002752, 0.00004682, -0.00000765, -0.00002311, -0.00001404, 0.00000043, -0.00001653, 0.00000273, -0.00000572, 0.00000977, -0.00001215, 0.00000393, 0.00000692, 0.00000180, -0.00001395, 0.00000544, -0.00000216, 0.00000330, -0.00001284, 0.00001184, -0.00000375, -0.00000121, -0.00000743, -0.00000067, -0.00000464, -0.00000841, 0.00000565, 0.00000286, -0.00000535, -0.00000636, 0.00000216, 0.00001328, -0.00000223, -0.00000937, 0.00000205, 0.00000365, -0.00000366, 0.00000060, 0.00001031, -0.00000167, -0.00000573, -0.00000281, -0.00000026, 0.00000073, 0.00000714, 0.00000378, -0.00000627, -0.00001100, 0.00000822, 0.00001124, -0.00000627, -0.00000461, 0.00000128, -0.00000284, 0.00000223, 0.00000245, -0.00000027, 0.00000895, -0.00000867, -0.00000715, 0.00000662, 0.00000449, 0.00000417, -0.00000436, -0.00000269, -0.00000587, 0.00000958, -0.00000088, 0.00000230, -0.00000304, 0.00000220, 0.00000421, -0.00000478, -0.00000419, 0.00000193, 0.00001258, -0.00001483, 0.00000935, -0.00000016, -0.00000041, -0.00000064, -0.00001943, 0.00000517, -0.00000625, 0.00000009, -0.00000287, -0.00001356, 0.00001005, -0.00000071, -0.00000597, 0.00000436, -0.00000858, 0.00001479, -0.00001151, 0.00000506, -0.00000837, 0.00000789, -0.00000792, -0.00000088, -0.00001821, 0.00000632, -0.00002459, -0.00000348, -0.00011910, 0.00007909, 0.00011650, 0.00013690, 0.00002823, -0.00010901, -0.00000272, -0.00002599, 0.00000293, -0.00001315, -0.00000145, -0.00000377, -0.00000451, -0.00000288, -0.00000103, -0.00000090, -0.00000865, 0.00000160, -0.00000909, -0.00000145, -0.00002009, -0.00003044, 0.00000278, 0.00004789, 0.00002379, },
		{ 0.00005260, -0.00000182, -0.00001704, -0.00002218, 0.00000534, -0.00001991, 0.00000442, -0.00000849, 0.00001161, -0.00001180, 0.00000245, 0.00000593, 0.00000556, -0.00001579, 0.00000482, -0.00000203, 0.00000431, -0.00001295, 0.00000909, 0.00000020, -0.00000316, -0.00000526, -0.00000352, -0.00000151, -0.00000994, 0.00000410, 0.00000391, -0.00000389, -0.00000679, -0.00000074, 0.00001274, 0.00000276, -0.00001117, 0.00000028, 0.00000432, -0.00000250, -0.00000141, 0.00001035, 0.00000030, -0.00000594, -0.00000323, -0.00000043, 0.00000025, 0.00000598, 0.00000520, -0.00000452, -0.00001099, 0.00000331, 0.00001416, -0.00000454, -0.00000576, 0.00000044, -0.00000156, 0.00000033, 0.00000430, -0.00000181, 0.00000857, -0.00000482, -0.00000978, 0.00000565, 0.00000442, 0.00000519, -0.00000368, -0.00000246, -0.00000653, 0.00000789, 0.00000119, 0.00000094, -0.00000113, -0.00000038, 0.00000612, -0.00000504, -0.00000289, -0.00000171, 0.00001414, -0.00001206, 0.00000392, 0.00000592, -0.00000592, 0.00000559, -0.00002163, 0.00000299, -0.00000403, -0.00000168, -0.00000080, -0.00001437, 0.00000617, 0.00000493, -0.00001061, 0.00000852, -0.00001255, 0.00001700, -0.00001186, 0.00000562, -0.00000940, 0.00000850, -0.00000870, 0.00000260, -0.00002173, 0.00000964, -0.00002807, 0.00000751, -0.00011452, 0.00002880, 0.00013714, 0.00013080, 0.00006818, -0.00011974, -0.00000784, -0.00002641, 0.00000132, -0.00001201, -0.00000293, -0.00000350, -0.00000380, -0.00000435, -0.00000036, -0.00000099, -0.00000735, -0.00000074, -0.00000580, -0.00000430, -0.00001356, -0.00003285, -0.00000128, 0.00003938, 0.00003235, 0.00001556, },
		{ 0.00000452, -0.00001037, -0.00002806, 0.00000693, -0.00002006, 0.00000322, -0.00000848, 0.00001050, -0.00000884, -0.00000078, 0.00000621, 0.00000742, -0.00001493, 0.00000170, 0.00000009, 0.00000328, -0.00001064, 0.00000416, 0.00000531, -0.00000570, -0.00000251, -0.00000657, 0.00000109, -0.00001056, 0.00000234, 0.00000432, -0.00000212, -0.00000692, -0.00000299, 0.00001060, 0.00000785, -0.00001147, -0.00000236, 0.00000488, -0.00000133, -0.00000279, 0.00000947, 0.00000268, -0.00000624, -0.00000332, -0.00000106, 0.00000043, 0.00000420, 0.00000672, -0.00000338, -0.00000957, -0.00000187, 0.00001576, -0.00000197, -0.00000652, -0.00000092, -0.00000021, -0.00000124, 0.00000533, -0.00000202, 0.00000684, -0.00000078, -0.00001141, 0.00000402, 0.00000446, 0.00000576, -0.00000254, -0.00000252, -0.00000654, 0.00000563, 0.00000338, -0.00000043, 0.00000070, -0.00000248, 0.00000722, -0.00000469, -0.00000215, -0.00000373, 0.00001310, -0.00000702, -0.00000265, 0.00001180, -0.00000993, 0.00000883, -0.00002002, -0.00000195, -0.00000052, -0.00000406, 0.00000136, -0.00001445, 0.00000166, 0.00000998, -0.00001368, 0.00001060, -0.00001389, 0.00001619, -0.00000932, 0.00000348, -0.00000767, 0.00000635, -0.00000682, 0.00000275, -0.00002073, 0.00000742, -0.00002480, 0.00000709, -0.00009305, -0.00002785, 0.00015049, 0.00012790, 0.00010409, -0.00012030, -0.00001978, -0.00002315, -0.00000311, -0.00000915, -0.00000529, -0.00000299, -0.00000318, -0.00000558, 0.00000006, -0.00000135, -0.00000538, -0.00000373, -0.00000212, -0.00000755, -0.00000722, -0.00003478, -0.00000363, 0.00002917, 0.00004093, 0.00000582, 0.00005447, },
		{ -0.00000508, -0.00002960, 0.00000391, -0.00001625, -0.00000131, -0.00000542, 0.00000639, -0.00000347, -0.00000555, 0.00000773, 0.00000727, -0.00001148, -0.00000351, 0.00000367, 0.00000074, -0.00000665, -0.00000193, 0.00001045, -0.00000787, -0.00000004, -0.00000899, 0.00000257, -0.00001014, 0.00000042, 0.00000428, -0.00000041, -0.00000657, -0.00000466, 0.00000747, 0.00001203, -0.00000979, -0.00000582, 0.00000530, -0.00000033, -0.00000339, 0.00000773, 0.00000531, -0.00000652, -0.00000314, -0.00000204, 0.00000108, 0.00000217, 0.00000803, -0.00000257, -0.00000742, -0.00000632, 0.00001549, 0.00000144, -0.00000695, -0.00000237, 0.00000062, -0.00000193, 0.00000518, -0.00000091, 0.00000421, 0.00000291, -0.00001187, 0.00000188, 0.00000450, 0.00000598, -0.00000112, -0.00000269, -0.00000613, 0.00000313, 0.00000532, -0.00000148, 0.00000213, -0.00000385, 0.00000741, -0.00000368, -0.00000212, -0.00000418, 0.00001014, -0.00000107, -0.00000862, 0.00001564, -0.00001080, 0.00000808, -0.00001456, -0.00000893, 0.00000349, -0.00000648, 0.00000324, -0.00001380, -0.00000297, 0.00001361, -0.00001442, 0.00000999, -0.00001217, 0.00001218, -0.00000388, -0.00000123, -0.00000338, 0.00000165, -0.00000238, -0.00000043, -0.00001535, 0.00000016, -0.00001560, -0.00000251, -0.00006132, -0.00008059, 0.00014897, 0.00013189, 0.00013277, -0.00010869, -0.00003866, -0.00001649, -0.00001010, -0.00000480, -0.00000842, -0.00000209, -0.00000304, -0.00000608, -0.00000021, -0.00000151, -0.00000340, -0.00000648, 0.00000096, -0.00001017, -0.00000208, -0.00003542, -0.00000546, 0.00001916, 0.00004725, 0.00000031, 0.00005143, 0.00001227, },
		{ -0.00002595, -0.00000359, -0.00000910, -0.00000838, -0.00000019, 0.00000031, 0.00000320, -0.00001085, 0.00000975, 0.00000578, -0.00000635, -0.00000964, 0.00000752, -0.00000215, -0.00000221, -0.00000776, 0.00001430, -0.00000863, 0.00000130, -0.00001005, 0.00000251, -0.00000857, -0.00000170, 0.00000410, 0.00000081, -0.00000555, -0.00000594, 0.00000405, 0.00001452, -0.00000600, -0.00000969, 0.00000535, 0.00000053, -0.00000330, 0.00000544, 0.00000782, -0.00000653, -0.00000293, -0.00000306, 0.00000175, 0.00000039, 0.00000870, -0.00000161, -0.00000538, -0.00000917, 0.00001305, 0.00000553, -0.00000713, -0.00000349, 0.00000051, -0.00000150, 0.00000385, 0.00000117, 0.00000130, 0.00000572, -0.00001111, -0.00000054, 0.00000439, 0.00000594, 0.00000040, -0.00000284, -0.00000549, 0.00000065, 0.00000681, -0.00000208, 0.00000304, -0.00000445, 0.00000680, -0.00000218, -0.00000262, -0.00000354, 0.00000634, 0.00000428, -0.00001244, 0.00001614, -0.00000783, 0.00000352, -0.00000641, -0.00001637, 0.00000670, -0.00000808, 0.00000427, -0.00001227, -0.00000740, 0.00001542, -0.00001269, 0.00000685, -0.00000781, 0.00000569, 0.00000346, -0.00000739, 0.00000227, -0.00000433, 0.00000339, -0.00000563, -0.00000733, -0.00000983, -0.00000353, -0.00001640, -0.00002803, -0.00011948, 0.00012765, 0.00014358, 0.00015335, -0.00008502, -0.00006290, -0.00000789, -0.00001856, 0.00000017, -0.00001185, -0.00000090, -0.00000356, -0.00000558, -0.00000141, -0.00000115, -0.00000199, -0.00000821, 0.00000262, -0.00001132, 0.00000104, -0.00003402, -0.00000792, 0.00001107, 0.00004952, 0.00000030, 0.00004321, 0.00002183, -0.00000241, },
		{ -0.00001384, -0.00000057, -0.00001607, 0.00000536, -0.00000591, 0.00000940, -0.00001515, 0.00001104, 0.00000413, -0.00000095, -0.00001509, 0.00001016, -0.00000404, 0.00000125, -0.00001188, 0.00001565, -0.00000719, 0.00000091, -0.00000933, 0.00000080, -0.00000599, -0.00000401, 0.00000404, 0.00000125, -0.00000385, -0.00000694, 0.00000091, 0.00001493, -0.00000053, -0.00001326, 0.00000462, 0.00000139, -0.00000281, 0.00000306, 0.00000975, -0.00000591, -0.00000304, -0.00000371, 0.00000197, -0.00000061, 0.00000834, -0.00000011, -0.00000409, -0.00000993, 0.00000865, 0.00000985, -0.00000697, -0.00000405, -0.00000065, -0.00000012, 0.00000179, 0.00000355, -0.00000115, 0.00000723, -0.00000920, -0.00000301, 0.00000409, 0.00000565, 0.00000194, -0.00000288, -0.00000469, -0.00000164, 0.00000775, -0.00000221, 0.00000345, -0.00000441, 0.00000567, -0.00000050, -0.00000332, -0.00000245, 0.00000283, 0.00000781, -0.00001322, 0.00001302, -0.00000156, -0.00000344, 0.00000236, -0.00002221, 0.00000767, -0.00000800, 0.00000394, -0.00000967, -0.00001151, 0.00001543, -0.00000895, 0.00000199, -0.00000194, -0.00000184, 0.00001102, -0.00001321, 0.00000745, -0.00000974, 0.00000858, -0.00001069, 0.00000065, -0.00001920, 0.00000731, -0.00002881, -0.00000146, -0.00013756, 0.00008612, 0.00016000, 0.00016757, -0.00005155, -0.00008926, 0.00000010, -0.00002666, 0.00000444, -0.00001476, 0.00000017, -0.00000455, -0.00000422, -0.00000338, -0.00000026, -0.00000140, -0.00000855, 0.00000252, -0.00001066, 0.00000182, -0.00003029, -0.00001165, 0.00000593, 0.00004693, 0.00000584, 0.00003065, 0.00003257, -0.00000231, -0.00001791, },
		{ 0.00000662, -0.00002197, 0.00000905, -0.00001026, 0.00001330, -0.00001694, 0.00001035, 0.00000355, 0.00000326, -0.00001838, 0.00001032, -0.00000385, 0.00000269, -0.00001332, 0.00001392, -0.00000336, -0.00000122, -0.00000695, -0.00000226, -0.00000279, -0.00000632, 0.00000417, 0.00000093, -0.00000171, -0.00000758, -0.00000170, 0.00001336, 0.00000574, -0.00001559, 0.00000271, 0.00000247, -0.00000228, 0.00000109, 0.00001067, -0.00000437, -0.00000371, -0.00000371, 0.00000141, -0.00000056, 0.00000679, 0.00000209, -0.00000383, -0.00000867, 0.00000297, 0.00001367, -0.00000626, -0.00000409, -0.00000257, 0.00000164, -0.00000027, 0.00000541, -0.00000247, 0.00000721, -0.00000637, -0.00000530, 0.00000361, 0.00000510, 0.00000345, -0.00000282, -0.00000377, -0.00000368, 0.00000815, -0.00000195, 0.00000349, -0.00000393, 0.00000430, 0.00000108, -0.00000390, -0.00000151, 0.00000037, 0.00000898, -0.00001101, 0.00000711, 0.00000639, -0.00001050, 0.00000925, -0.00002452, 0.00000536, -0.00000573, 0.00000197, -0.00000599, -0.00001522, 0.00001394, -0.00000404, -0.00000335, 0.00000393, -0.00000862, 0.00001689, -0.00001682, 0.00001031, -0.00001271, 0.00001130, -0.00001347, 0.00000596, -0.00002483, 0.00001326, -0.00003500, 0.00001283, -0.00013280, 0.00002927, 0.00017477, 0.00017909, -0.00001203, -0.00011351, 0.00000446, -0.00003232, 0.00000651, -0.00001618, 0.00000051, -0.00000556, -0.00000250, -0.00000562, 0.00000080, -0.00000148, -0.00000762, 0.00000087, -0.00000843, 0.00000053, -0.00002460, -0.00001661, 0.00000381, 0.00003990, 0.00001561, 0.00001580, 0.00004282, -0.00000345, -0.00000770, -0.00002397, },
		{ -0.00002397, 0.00000916, -0.00001121, 0.00001363, -0.00001529, 0.00000702, 0.00000477, 0.00000535, -0.00001863, 0.00000742, -0.00000121, 0.00000177, -0.00001190, 0.00000932, 0.00000230, -0.00000448, -0.00000354, -0.00000592, 0.00000039, -0.00000825, 0.00000430, 0.00000017, 0.00000038, -0.00000761, -0.00000380, 0.00001036, 0.00001166, -0.00001581, -0.00000063, 0.00000387, -0.00000200, -0.00000012, 0.00001037, -0.00000187, -0.00000500, -0.00000297, 0.00000003, 0.00000051, 0.00000429, 0.00000476, -0.00000439, -0.00000601, -0.00000289, 0.00001613, -0.00000467, -0.00000387, -0.00000464, 0.00000295, -0.00000150, 0.00000606, -0.00000226, 0.00000572, -0.00000295, -0.00000721, 0.00000301, 0.00000426, 0.00000489, -0.00000268, -0.00000276, -0.00000540, 0.00000803, -0.00000138, 0.00000323, -0.00000317, 0.00000288, 0.00000241, -0.00000419, -0.00000101, -0.00000077, 0.00000803, -0.00000676, 0.00000008, 0.00001375, -0.00001521, 0.00001216, -0.00002211, -0.00000044, -0.00000149, -0.00000138, -0.00000158, -0.00001831, 0.00001128, 0.00000112, -0.00000796, 0.00000842, -0.00001315, 0.00001960, -0.00001689, 0.00000967, -0.00001215, 0.00001043, -0.00001268, 0.00000699, -0.00002491, 0.00001245, -0.00003287, 0.00001341, -0.00010869, -0.00003349, 0.00017964, 0.00019198, 0.00002954, -0.00013147, 0.00000251, -0.00003384, 0.00000525, -0.00001539, -0.00000043, -0.00000601, -0.00000111, -0.00000741, 0.00000147, -0.00000178, -0.00000593, -0.00000165, -0.00000538, -0.00000200, -0.00001781, -0.00002207, 0.00000391, 0.00002995, 0.00002725, 0.00000153, 0.00005023, -0.00000366, 0.00000172, -0.00003083, 0.00000987, },
		{ 0.00000507, -0.00000839, 0.00001025, -0.00001031, 0.00000128, 0.00000769, 0.00000523, -0.00001588, 0.00000184, 0.00000335, -0.00000095, -0.00000832, 0.00000293, 0.00000857, -0.00000775, -0.00000013, -0.00000918, 0.00000272, -0.00000931, 0.00000406, -0.00000048, 0.00000179, -0.00000672, -0.00000556, 0.00000669, 0.00001612, -0.00001338, -0.00000529, 0.00000545, -0.00000203, -0.00000046, 0.00000893, 0.00000135, -0.00000665, -0.00000176, -0.00000188, 0.00000215, 0.00000141, 0.00000733, -0.00000518, -0.00000296, -0.00000768, 0.00001645, -0.00000196, -0.00000376, -0.00000617, 0.00000307, -0.00000129, 0.00000514, -0.00000055, 0.00000314, 0.00000056, -0.00000855, 0.00000232, 0.00000317, 0.00000617, -0.00000243, -0.00000173, -0.00000672, 0.00000740, -0.00000053, 0.00000272, -0.00000217, 0.00000146, 0.00000351, -0.00000424, -0.00000096, -0.00000082, 0.00000576, -0.00000194, -0.00000611, 0.00001839, -0.00001573, 0.00001006, -0.00001512, -0.00000880, 0.00000370, -0.00000525, 0.00000282, -0.00002035, 0.00000763, 0.00000587, -0.00001106, 0.00001070, -0.00001464, 0.00001856, -0.00001314, 0.00000548, -0.00000814, 0.00000613, -0.00000839, 0.00000374, -0.00001964, 0.00000545, -0.00002344, 0.00000310, -0.00007314, -0.00009048, 0.00016703, 0.00020877, 0.00007027, -0.00014005, -0.00000720, -0.00003059, 0.00000030, -0.00001223, -0.00000286, -0.00000553, -0.00000062, -0.00000814, 0.00000121, -0.00000178, -0.00000419, -0.00000416, -0.00000245, -0.00000475, -0.00001111, -0.00002694, 0.00000484, 0.00001920, 0.00003788, -0.00000914, 0.00005252, -0.00000088, 0.00000773, -0.00003208, 0.00000768, -0.00002116, },
		{ -0.00000274, 0.00000422, -0.00000322, -0.00000569, 0.00001138, 0.00000368, -0.00001108, -0.00000518, 0.00000852, -0.00000424, -0.00000396, -0.00000367, 0.00001385, -0.00000973, 0.00000214, -0.00001107, 0.00000349, -0.00000904, 0.00000303, -0.00000046, 0.00000202, -0.00000476, -0.00000716, 0.00000306, 0.00001833, -0.00000831, -0.00001070, 0.00000678, -0.00000220, -0.00000018, 0.00000678, 0.00000476, -0.00000815, -0.00000057, -0.00000374, 0.00000370, -0.00000107, 0.00000907, -0.00000543, -0.00000059, -0.00001037, 0.00001421, 0.00000188, -0.00000399, -0.00000661, 0.00000155, 0.00000049, 0.00000285, 0.00000218, 0.00000017, 0.00000357, -0.00000909, 0.00000152, 0.00000196, 0.00000712, -0.00000194, -0.00000084, -0.00000746, 0.00000617, 0.00000065, 0.00000189, -0.00000092, -0.00000001, 0.00000452, -0.00000419, -0.00000114, -0.00000032, 0.00000329, 0.00000194, -0.00000983, 0.00001896, -0.00001147, 0.00000345, -0.00000508, -0.00001774, 0.00000816, -0.00000836, 0.00000617, -0.00002071, 0.00000301, 0.00000989, -0.00001238, 0.00001067, -0.00001320, 0.00001423, -0.00000643, -0.00000110, -0.00000198, -0.00000020, -0.00000202, -0.00000222, -0.00001106, -0.00000503, -0.00001030, -0.00001234, -0.00003617, -0.00013086, 0.00013278, 0.00022875, 0.00010940, -0.00013803, -0.00002440, -0.00002330, -0.00000759, -0.00000733, -0.00000642, -0.00000412, -0.00000125, -0.00000752, -0.00000020, -0.00000119, -0.00000293, -0.00000593, -0.00000044, -0.00000680, -0.00000555, -0.00003013, 0.00000518, 0.00000980, 0.00004495, -0.00001392, 0.00004831, 0.00000603, 0.00000893, -0.00002702, 0.00000030, -0.00001421, -0.00000235, },
		{ -0.00000248, 0.00000404, -0.00001209, 0.00001433, 0.00000202, -0.00000574, -0.00001194, 0.00001264, -0.00000657, -0.00000035, -0.00000884, 0.00001672, -0.00000937, 0.00000248, -0.00001094, 0.00000233, -0.00000728, 0.00000101, 0.00000055, 0.00000090, -0.00000188, -0.00000854, -0.00000008, 0.00001803, -0.00000124, -0.00001583, 0.00000721, -0.00000210, 0.00000028, 0.00000454, 0.00000768, -0.00000889, 0.00000001, -0.00000493, 0.00000447, -0.00000243, 0.00000937, -0.00000453, 0.00000033, -0.00001040, 0.00000955, 0.00000645, -0.00000455, -0.00000578, -0.00000151, 0.00000338, -0.00000013, 0.00000506, -0.00000233, 0.00000551, -0.00000866, 0.00000053, 0.00000086, 0.00000749, -0.00000107, -0.00000030, -0.00000744, 0.00000428, 0.00000221, 0.00000069, 0.00000064, -0.00000161, 0.00000555, -0.00000420, -0.00000131, 0.00000016, 0.00000154, 0.00000387, -0.00001033, 0.00001536, -0.00000335, -0.00000572, 0.00000537, -0.00002473, 0.00001004, -0.00000941, 0.00000746, -0.00001884, -0.00000266, 0.00001314, -0.00001217, 0.00000890, -0.00000971, 0.00000792, 0.00000151, -0.00000814, 0.00000426, -0.00000643, 0.00000422, -0.00000839, -0.00000223, -0.00001521, 0.00000186, -0.00002632, -0.00000697, -0.00014766, 0.00007807, 0.00024724, 0.00014842, -0.00012605, -0.00004712, -0.00001401, -0.00001677, -0.00000199, -0.00001026, -0.00000226, -0.00000276, -0.00000578, -0.00000252, -0.00000007, -0.00000234, -0.00000660, 0.00000026, -0.00000767, -0.00000177, -0.00003099, 0.00000393, 0.00000325, 0.00004688, -0.00001190, 0.00003774, 0.00001668, 0.00000571, -0.00001691, -0.00001026, -0.00000526, -0.00001098, 0.00000371, },
		{ 0.00000942, -0.00001614, 0.00001513, 0.00000156, -0.00000135, -0.00001680, 0.00001426, -0.00000671, 0.00000128, -0.00001139, 0.00001633, -0.00000626, 0.00000068, -0.00000876, -0.00000059, -0.00000429, -0.00000181, 0.00000242, -0.00000125, 0.00000135, -0.00000942, -0.00000262, 0.00001557, 0.00000666, -0.00001949, 0.00000607, -0.00000134, 0.00000041, 0.00000283, 0.00000950, -0.00000837, -0.00000047, -0.00000500, 0.00000398, -0.00000227, 0.00000798, -0.00000225, -0.00000047, -0.00000794, 0.00000327, 0.00001095, -0.00000508, -0.00000398, -0.00000546, 0.00000643, -0.00000272, 0.00000708, -0.00000357, 0.00000597, -0.00000718, -0.00000072, 0.00000015, 0.00000707, 0.00000033, -0.00000028, -0.00000656, 0.00000173, 0.00000411, -0.00000083, 0.00000243, -0.00000329, 0.00000660, -0.00000427, -0.00000137, 0.00000027, 0.00000099, 0.00000361, -0.00000792, 0.00000877, 0.00000648, -0.00001459, 0.00001324, -0.00002743, 0.00000796, -0.00000759, 0.00000614, -0.00001455, -0.00000921, 0.00001565, -0.00001091, 0.00000621, -0.00000536, 0.00000127, 0.00000873, -0.00001357, 0.00000853, -0.00001051, 0.00000819, -0.00001239, 0.00000396, -0.00002167, 0.00000909, -0.00003378, 0.00000875, -0.00013975, 0.00001002, 0.00025635, 0.00019010, -0.00010586, -0.00007233, -0.00000544, -0.00002511, 0.00000219, -0.00001331, -0.00000068, -0.00000455, -0.00000352, -0.00000515, 0.00000114, -0.00000224, -0.00000629, -0.00000020, -0.00000738, 0.00000016, -0.00002945, 0.00000076, 0.00000013, 0.00004344, -0.00000379, 0.00002262, 0.00002910, 0.00000011, -0.00000458, -0.00002080, 0.00000272, -0.00001821, 0.00000861, -0.00000765, },
		{ -0.00001672, 0.00001295, 0.00000306, 0.00000112, -0.00001878, 0.00001268, -0.00000418, 0.00000051, -0.00001101, 0.00001277, -0.00000086, -0.00000268, -0.00000514, -0.00000454, -0.00000079, -0.00000488, 0.00000464, -0.00000370, 0.00000412, -0.00000933, -0.00000474, 0.00001168, 0.00001397, -0.00002055, 0.00000292, 0.00000031, -0.00000013, 0.00000208, 0.00000989, -0.00000641, -0.00000212, -0.00000385, 0.00000220, -0.00000064, 0.00000515, 0.00000109, -0.00000261, -0.00000383, -0.00000335, 0.00001434, -0.00000504, -0.00000182, -0.00000927, 0.00000846, -0.00000386, 0.00000740, -0.00000304, 0.00000483, -0.00000478, -0.00000223, -0.00000002, 0.00000575, 0.00000227, -0.00000081, -0.00000487, -0.00000131, 0.00000618, -0.00000249, 0.00000424, -0.00000484, 0.00000746, -0.00000420, -0.00000142, -0.00000000, 0.00000154, 0.00000171, -0.00000383, 0.00000123, 0.00001530, -0.00002025, 0.00001610, -0.00002450, 0.00000161, -0.00000301, 0.00000243, -0.00000830, -0.00001608, 0.00001725, -0.00000897, 0.00000333, -0.00000121, -0.00000433, 0.00001368, -0.00001594, 0.00000953, -0.00001121, 0.00000870, -0.00001284, 0.00000587, -0.00002263, 0.00000955, -0.00003272, 0.00001009, -0.00011221, -0.00005968, 0.00024711, 0.00023663, -0.00007900, -0.00009700, -0.00000015, -0.00003078, 0.00000389, -0.00001468, -0.00000008, -0.00000595, -0.00000151, -0.00000734, 0.00000187, -0.00000220, -0.00000543, -0.00000133, -0.00000639, 0.00000063, -0.00002599, -0.00000393, 0.00000003, 0.00003567, 0.00000828, 0.00000597, 0.00004032, -0.00000492, 0.00000651, -0.00002796, 0.00000704, -0.00002180, 0.00001007, -0.00000966, 0.00001148, },
		{ 0.00000798, 0.00000639, 0.00000159, -0.00001784, 0.00000815, 0.00000053, -0.00000216, -0.00000833, 0.00000703, 0.00000559, -0.00000644, -0.00000121, -0.00000841, 0.00000223, -0.00000737, 0.00000637, -0.00000550, 0.00000553, -0.00000785, -0.00000672, 0.00000728, 0.00001940, -0.00001833, -0.00000217, 0.00000270, -0.00000136, 0.00000232, 0.00000892, -0.00000323, -0.00000464, -0.00000182, -0.00000047, 0.00000188, 0.00000162, 0.00000472, -0.00000521, 0.00000063, -0.00000879, 0.00001560, -0.00000386, -0.00000003, -0.00001187, 0.00000842, -0.00000282, 0.00000566, -0.00000075, 0.00000240, -0.00000181, -0.00000389, 0.00000039, 0.00000364, 0.00000456, -0.00000173, -0.00000261, -0.00000447, 0.00000804, -0.00000391, 0.00000568, -0.00000590, 0.00000778, -0.00000365, -0.00000176, -0.00000038, 0.00000266, -0.00000073, 0.00000023, -0.00000502, 0.00002066, -0.00002067, 0.00001294, -0.00001618, -0.00000789, 0.00000312, -0.00000259, -0.00000126, -0.00002223, 0.00001741, -0.00000636, 0.00000059, 0.00000219, -0.00000818, 0.00001572, -0.00001489, 0.00000718, -0.00000867, 0.00000597, -0.00000993, 0.00000367, -0.00001853, 0.00000413, -0.00002462, 0.00000077, -0.00007466, -0.00011772, 0.00021255, 0.00028746, -0.00004551, -0.00011907, 0.00000032, -0.00003290, 0.00000261, -0.00001409, -0.00000072, -0.00000654, -0.00000031, -0.00000849, 0.00000167, -0.00000184, -0.00000451, -0.00000253, -0.00000528, 0.00000025, -0.00002141, -0.00000933, 0.00000187, 0.00002545, 0.00002139, -0.00000873, 0.00004728, -0.00000659, 0.00001343, -0.00002943, 0.00000620, -0.00002078, 0.00000746, -0.00000806, 0.00000996, -0.00001382, },
		{ 0.00001054, 0.00000086, -0.00001483, 0.00000182, 0.00000615, -0.00000549, -0.00000469, 0.00000071, 0.00001144, -0.00000915, 0.00000174, -0.00001101, 0.00000380, -0.00000848, 0.00000684, -0.00000581, 0.00000499, -0.00000485, -0.00000869, 0.00000313, 0.00002209, -0.00001285, -0.00000856, 0.00000527, -0.00000294, 0.00000323, 0.00000707, 0.00000054, -0.00000738, 0.00000040, -0.00000325, 0.00000442, -0.00000164, 0.00000766, -0.00000722, 0.00000408, -0.00001180, 0.00001409, -0.00000128, 0.00000080, -0.00001245, 0.00000581, 0.00000047, 0.00000217, 0.00000272, -0.00000062, 0.00000115, -0.00000540, 0.00000119, 0.00000109, 0.00000680, -0.00000268, -0.00000026, -0.00000722, 0.00000921, -0.00000464, 0.00000632, -0.00000605, 0.00000716, -0.00000226, -0.00000269, -0.00000051, 0.00000363, -0.00000254, 0.00000275, -0.00000831, 0.00002118, -0.00001538, 0.00000452, -0.00000441, -0.00001820, 0.00000875, -0.00000718, 0.00000493, -0.00002630, 0.00001543, -0.00000283, -0.00000216, 0.00000493, -0.00001035, 0.00001520, -0.00001119, 0.00000263, -0.00000423, 0.00000147, -0.00000521, -0.00000084, -0.00001168, -0.00000412, -0.00001358, -0.00001263, -0.00003833, -0.00015287, 0.00015078, 0.00033771, -0.00000318, -0.00013792, -0.00000415, -0.00003185, -0.00000112, -0.00001206, -0.00000231, -0.00000636, -0.00000008, -0.00000847, 0.00000049, -0.00000108, -0.00000381, -0.00000338, -0.00000445, -0.00000046, -0.00001647, -0.00001454, 0.00000439, 0.00001490, 0.00003267, -0.00001858, 0.00004786, -0.00000327, 0.00001476, -0.00002467, 0.00000053, -0.00001582, 0.00000160, -0.00000374, 0.00000583, -0.00000853, 0.00000140, },
		{ 0.00000017, -0.00001112, -0.00000467, 0.00001103, -0.00000795, -0.00000161, -0.00000453, 0.00001515, -0.00000962, 0.00000275, -0.00001154, 0.00000331, -0.00000776, 0.00000560, -0.00000424, 0.00000242, -0.00000065, -0.00001051, -0.00000034, 0.00002183, -0.00000486, -0.00001507, 0.00000717, -0.00000428, 0.00000419, 0.00000506, 0.00000411, -0.00000954, 0.00000201, -0.00000532, 0.00000609, -0.00000369, 0.00000906, -0.00000772, 0.00000552, -0.00001167, 0.00000983, 0.00000251, 0.00000051, -0.00001077, 0.00000083, 0.00000531, -0.00000210, 0.00000632, -0.00000329, 0.00000345, -0.00000638, 0.00000201, -0.00000137, 0.00000846, -0.00000319, 0.00000166, -0.00000900, 0.00000922, -0.00000427, 0.00000579, -0.00000499, 0.00000536, 0.00000013, -0.00000431, -0.00000014, 0.00000390, -0.00000287, 0.00000286, -0.00000803, 0.00001697, -0.00000567, -0.00000668, 0.00000765, -0.00002629, 0.00001158, -0.00000955, 0.00000865, -0.00002706, 0.00001066, 0.00000196, -0.00000527, 0.00000742, -0.00001145, 0.00001313, -0.00000635, -0.00000231, 0.00000010, -0.00000267, -0.00000094, -0.00000513, -0.00000520, -0.00001133, -0.00000439, -0.00002327, -0.00001255, -0.00015935, 0.00006685, 0.00037794, 0.00005197, -0.00015402, -0.00001253, -0.00002901, -0.00000600, -0.00000971, -0.00000408, -0.00000586, -0.00000050, -0.00000757, -0.00000131, -0.00000017, -0.00000330, -0.00000380, -0.00000400, -0.00000127, -0.00001168, -0.00001888, 0.00000651, 0.00000573, 0.00004004, -0.00002195, 0.00004157, 0.00000491, 0.00001108, -0.00001517, -0.00000792, -0.00000899, -0.00000550, 0.00000138, 0.00000093, -0.00000271, -0.00000500, 0.00001399, },
		{ -0.00000804, -0.00000984, 0.00001375, -0.00000836, -0.00000030, -0.00000748, 0.00001579, -0.00000732, 0.00000148, -0.00000980, 0.00000078, -0.00000532, 0.00000277, -0.00000103, -0.00000158, 0.00000391, -0.00001169, -0.00000313, 0.00001906, 0.00000429, -0.00002026, 0.00000749, -0.00000475, 0.00000454, 0.00000362, 0.00000672, -0.00001041, 0.00000236, -0.00000603, 0.00000623, -0.00000395, 0.00000847, -0.00000634, 0.00000457, -0.00000855, 0.00000358, 0.00000677, -0.00000057, -0.00000727, -0.00000556, 0.00001037, -0.00000579, 0.00000884, -0.00000468, 0.00000447, -0.00000649, 0.00000245, -0.00000318, 0.00000902, -0.00000283, 0.00000265, -0.00000938, 0.00000779, -0.00000260, 0.00000398, -0.00000271, 0.00000245, 0.00000338, -0.00000646, 0.00000065, 0.00000330, -0.00000154, 0.00000064, -0.00000479, 0.00000958, 0.00000581, -0.00001726, 0.00001651, -0.00002943, 0.00000989, -0.00000848, 0.00000887, -0.00002388, 0.00000300, 0.00000800, -0.00000893, 0.00001004, -0.00001211, 0.00001061, -0.00000190, -0.00000588, 0.00000254, -0.00000466, 0.00000104, -0.00000705, -0.00000169, -0.00001447, -0.00000060, -0.00002660, -0.00000206, -0.00013875, -0.00002730, 0.00039575, 0.00012391, -0.00016764, -0.00002354, -0.00002607, -0.00001050, -0.00000823, -0.00000521, -0.00000565, -0.00000107, -0.00000641, -0.00000309, 0.00000040, -0.00000268, -0.00000402, -0.00000368, -0.00000224, -0.00000726, -0.00002204, 0.00000761, -0.00000101, 0.00004257, -0.00001883, 0.00002969, 0.00001607, 0.00000464, -0.00000389, -0.00001605, -0.00000303, -0.00001141, 0.00000516, -0.00000277, 0.00000177, -0.00000952, 0.00001536, 0.00000071, },
		{ -0.00001283, 0.00001365, -0.00000625, -0.00000113, -0.00000780, 0.00001337, -0.00000262, -0.00000155, -0.00000639, -0.00000311, -0.00000186, -0.00000096, 0.00000299, -0.00000591, 0.00000768, -0.00001159, -0.00000551, 0.00001467, 0.00001299, -0.00002280, 0.00000559, -0.00000395, 0.00000385, 0.00000324, 0.00000791, -0.00000963, 0.00000118, -0.00000517, 0.00000472, -0.00000240, 0.00000606, -0.00000331, 0.00000164, -0.00000338, -0.00000329, 0.00001040, -0.00000176, -0.00000286, -0.00001190, 0.00001405, -0.00000752, 0.00000925, -0.00000415, 0.00000393, -0.00000556, 0.00000220, -0.00000391, 0.00000817, -0.00000138, 0.00000248, -0.00000821, 0.00000492, 0.00000024, 0.00000110, 0.00000049, -0.00000118, 0.00000700, -0.00000862, 0.00000150, 0.00000211, 0.00000102, -0.00000297, -0.00000017, 0.00000143, 0.00001590, -0.00002388, 0.00001948, -0.00002614, 0.00000325, -0.00000395, 0.00000565, -0.00001716, -0.00000681, 0.00001466, -0.00001279, 0.00001267, -0.00001255, 0.00000828, 0.00000123, -0.00000715, 0.00000230, -0.00000382, 0.00000006, -0.00000585, -0.00000210, -0.00001258, -0.00000304, -0.00002186, -0.00000579, -0.00009968, -0.00011514, 0.00037903, 0.00021399, -0.00017722, -0.00003676, -0.00002404, -0.00001375, -0.00000828, -0.00000529, -0.00000605, -0.00000137, -0.00000549, -0.00000430, 0.00000022, -0.00000169, -0.00000435, -0.00000318, -0.00000356, -0.00000321, -0.00002396, 0.00000751, -0.00000504, 0.00004047, -0.00001062, 0.00001485, 0.00002724, -0.00000148, 0.00000580, -0.00002090, -0.00000023, -0.00001426, 0.00000604, -0.00000393, 0.00000370, -0.00001117, 0.00001396, 0.00000303, -0.00000629, },
		{ 0.00001098, -0.00000212, -0.00000358, -0.00000610, 0.00000881, 0.00000328, -0.00000517, -0.00000245, -0.00000718, 0.00000148, -0.00000450, 0.00000662, -0.00000922, 0.00000952, -0.00000968, -0.00000783, 0.00000968, 0.00001977, -0.00002181, 0.00000135, -0.00000190, 0.00000203, 0.00000399, 0.00000767, -0.00000733, -0.00000131, -0.00000304, 0.00000195, 0.00000039, 0.00000255, 0.00000056, -0.00000225, 0.00000236, -0.00000913, 0.00001224, -0.00000224, 0.00000131, -0.00001666, 0.00001496, -0.00000635, 0.00000711, -0.00000163, 0.00000196, -0.00000370, 0.00000117, -0.00000337, 0.00000590, 0.00000107, 0.00000124, -0.00000569, 0.00000100, 0.00000377, -0.00000227, 0.00000396, -0.00000480, 0.00001021, -0.00001006, 0.00000179, 0.00000089, 0.00000391, -0.00000659, 0.00000389, -0.00000499, 0.00002187, -0.00002439, 0.00001554, -0.00001680, -0.00000707, 0.00000272, 0.00000023, -0.00000838, -0.00001713, 0.00002056, -0.00001586, 0.00001455, -0.00001228, 0.00000600, 0.00000307, -0.00000638, -0.00000009, -0.00000094, -0.00000294, -0.00000249, -0.00000531, -0.00000718, -0.00000956, -0.00001214, -0.00001774, -0.00005524, -0.00017972, 0.00032014, 0.00031816, -0.00017791, -0.00005359, -0.00002257, -0.00001604, -0.00000947, -0.00000476, -0.00000676, -0.00000149, -0.00000496, -0.00000476, -0.00000078, -0.00000030, -0.00000487, -0.00000234, -0.00000523, 0.00000038, -0.00002469, 0.00000625, -0.00000657, 0.00003476, 0.00000043, 0.00000017, 0.00003541, -0.00000450, 0.00001135, -0.00002070, -0.00000158, -0.00001349, 0.00000373, -0.00000243, 0.00000305, -0.00001005, 0.00001012, 0.00000677, -0.00000576, -0.00001368, },
		{ 0.00000280, -0.00000644, -0.00000366, 0.00000364, 0.00000875, -0.00000790, 0.00000066, -0.00001014, 0.00000355, -0.00000675, 0.00000872, -0.00001040, 0.00000872, -0.00000585, -0.00001021, 0.00000493, 0.00002364, -0.00001719, -0.00000468, 0.00000088, -0.00000054, 0.00000552, 0.00000645, -0.00000406, -0.00000447, -0.00000034, -0.00000128, 0.00000350, -0.00000100, 0.00000417, -0.00000584, 0.00000709, -0.00001250, 0.00001149, -0.00000146, 0.00000427, -0.00001866, 0.00001235, -0.00000217, 0.00000276, 0.00000230, -0.00000083, -0.00000135, -0.00000044, -0.00000176, 0.00000258, 0.00000408, -0.00000067, -0.00000238, -0.00000327, 0.00000726, -0.00000538, 0.00000691, -0.00000762, 0.00001221, -0.00001008, 0.00000100, 0.00000024, 0.00000618, -0.00000885, 0.00000578, -0.00000794, 0.00002237, -0.00001845, 0.00000579, -0.00000371, -0.00001834, 0.00000911, -0.00000525, 0.00000023, -0.00002575, 0.00002390, -0.00001671, 0.00001447, -0.00001033, 0.00000298, 0.00000454, -0.00000484, -0.00000296, 0.00000208, -0.00000587, 0.00000078, -0.00000872, -0.00000153, -0.00001601, -0.00000286, -0.00002951, -0.00001903, -0.00020841, 0.00021953, 0.00042544, -0.00016131, -0.00007714, -0.00001993, -0.00001890, -0.00001041, -0.00000483, -0.00000688, -0.00000203, -0.00000439, -0.00000485, -0.00000213, 0.00000104, -0.00000532, -0.00000137, -0.00000695, 0.00000319, -0.00002411, 0.00000387, -0.00000602, 0.00002672, 0.00001199, -0.00001163, 0.00003847, -0.00000297, 0.00001186, -0.00001555, -0.00000625, -0.00001011, -0.00000066, 0.00000065, 0.00000088, -0.00000733, 0.00000510, 0.00001076, -0.00000555, -0.00001322, 0.00000681, },
		{ -0.00000831, -0.00000188, -0.00000057, 0.00001229, -0.00000853, 0.00000189, -0.00001105, 0.00000355, -0.00000700, 0.00000857, -0.00000892, 0.00000524, -0.00000054, -0.00001241, 0.00000085, 0.00002433, -0.00000962, -0.00001135, 0.00000351, -0.00000319, 0.00000714, 0.00000494, -0.00000060, -0.00000746, 0.00000203, -0.00000402, 0.00000591, -0.00000354, 0.00000649, -0.00000801, 0.00000959, -0.00001255, 0.00000799, 0.00000064, 0.00000556, -0.00001745, 0.00000640, 0.00000423, -0.00000271, 0.00000652, -0.00000351, 0.00000085, -0.00000220, 0.00000046, -0.00000112, 0.00000697, -0.00000258, 0.00000098, -0.00000704, 0.00000987, -0.00000744, 0.00000860, -0.00000892, 0.00001239, -0.00000827, -0.00000115, 0.00000051, 0.00000719, -0.00000894, 0.00000473, -0.00000696, 0.00001773, -0.00000774, -0.00000684, 0.00000945, -0.00002709, 0.00001250, -0.00000849, 0.00000636, -0.00003051, 0.00002302, -0.00001405, 0.00001126, -0.00000571, -0.00000171, 0.00000681, -0.00000415, -0.00000432, 0.00000302, -0.00000638, 0.00000143, -0.00000945, 0.00000085, -0.00001809, 0.00000069, -0.00003361, -0.00000088, -0.00019688, 0.00008772, 0.00051825, -0.00011689, -0.00011104, -0.00001391, -0.00002439, -0.00000923, -0.00000709, -0.00000521, -0.00000388, -0.00000306, -0.00000533, -0.00000302, 0.00000160, -0.00000509, -0.00000083, -0.00000809, 0.00000468, -0.00002196, 0.00000034, -0.00000374, 0.00001754, 0.00002224, -0.00001896, 0.00003592, 0.00000275, 0.00000839, -0.00000739, -0.00001197, -0.00000627, -0.00000511, 0.00000343, -0.00000107, -0.00000469, 0.00000061, 0.00001351, -0.00000447, -0.00001263, 0.00000255, 0.00000699, },
		{ -0.00000175, -0.00000277, 0.00001303, -0.00000651, 0.00000087, -0.00000965, 0.00000137, -0.00000519, 0.00000620, -0.00000508, -0.00000015, 0.00000521, -0.00001384, -0.00000255, 0.00002223, -0.00000044, -0.00001718, 0.00000500, -0.00000517, 0.00000813, 0.00000391, 0.00000227, -0.00000949, 0.00000331, -0.00000551, 0.00000687, -0.00000437, 0.00000693, -0.00000816, 0.00000934, -0.00000934, 0.00000240, 0.00000352, 0.00000529, -0.00001341, -0.00000178, 0.00001127, -0.00000769, 0.00000968, -0.00000508, 0.00000223, -0.00000360, 0.00000264, -0.00000439, 0.00000897, -0.00000383, 0.00000360, -0.00000955, 0.00001094, -0.00000787, 0.00000854, -0.00000835, 0.00001056, -0.00000463, -0.00000447, 0.00000160, 0.00000684, -0.00000679, 0.00000107, -0.00000294, 0.00000983, 0.00000468, -0.00001845, 0.00001880, -0.00003027, 0.00001086, -0.00000790, 0.00000842, -0.00003010, 0.00001711, -0.00000732, 0.00000448, 0.00000191, -0.00000842, 0.00001058, -0.00000545, -0.00000281, 0.00000047, -0.00000306, -0.00000202, -0.00000575, -0.00000218, -0.00001336, -0.00000433, -0.00002641, -0.00000382, -0.00015086, -0.00005555, 0.00057533, -0.00003506, -0.00015711, -0.00000344, -0.00003382, -0.00000468, -0.00001256, -0.00000101, -0.00000765, -0.00000035, -0.00000694, -0.00000262, 0.00000063, -0.00000360, -0.00000134, -0.00000794, 0.00000425, -0.00001796, -0.00000444, -0.00000004, 0.00000810, 0.00003011, -0.00002149, 0.00002879, 0.00001075, 0.00000341, 0.00000075, -0.00001585, -0.00000440, -0.00000753, 0.00000410, -0.00000121, -0.00000365, -0.00000194, 0.00001385, -0.00000163, -0.00001285, -0.00000064, 0.00000930, -0.00000817, },
		{ -0.00000279, 0.00001105, -0.00000224, -0.00000189, -0.00000650, -0.00000235, -0.00000202, 0.00000238, 0.00000008, -0.00000603, 0.00001000, -0.00001377, -0.00000555, 0.00001823, 0.00000871, -0.00002079, 0.00000458, -0.00000593, 0.00000794, 0.00000384, 0.00000405, -0.00001009, 0.00000310, -0.00000543, 0.00000613, -0.00000339, 0.00000551, -0.00000639, 0.00000665, -0.00000379, -0.00000398, 0.00000621, 0.00000411, -0.00000759, -0.00001044, 0.00001700, -0.00001055, 0.00001060, -0.00000482, 0.00000235, -0.00000424, 0.00000419, -0.00000650, 0.00000949, -0.00000388, 0.00000494, -0.00001029, 0.00001012, -0.00000645, 0.00000666, -0.00000599, 0.00000697, 0.00000031, -0.00000834, 0.00000302, 0.00000551, -0.00000308, -0.00000401, 0.00000228, 0.00000140, 0.00001527, -0.00002544, 0.00002146, -0.00002630, 0.00000370, -0.00000328, 0.00000618, -0.00002457, 0.00000670, 0.00000274, -0.00000513, 0.00001171, -0.00001645, 0.00001548, -0.00000867, 0.00000155, -0.00000538, 0.00000375, -0.00000913, 0.00000196, -0.00001020, -0.00000245, -0.00001683, -0.00000959, -0.00002322, -0.00008502, -0.00018510, 0.00057682, 0.00008854, -0.00021266, 0.00000968, -0.00004646, 0.00000285, -0.00002091, 0.00000536, -0.00001308, 0.00000369, -0.00000983, -0.00000065, -0.00000208, -0.00000065, -0.00000319, -0.00000611, 0.00000168, -0.00001213, -0.00001031, 0.00000468, -0.00000092, 0.00003521, -0.00001997, 0.00001917, 0.00001836, -0.00000018, 0.00000595, -0.00001559, -0.00000606, -0.00000673, 0.00000177, 0.00000116, -0.00000482, -0.00000201, 0.00001150, 0.00000306, -0.00001408, -0.00000236, 0.00000935, -0.00000582, -0.00000345, },
		{ 0.00000736, 0.00000304, -0.00000519, -0.00000278, -0.00000638, 0.00000131, -0.00000164, 0.00000508, -0.00001077, 0.00001251, -0.00001161, -0.00000846, 0.00001333, 0.00001633, -0.00002121, 0.00000199, -0.00000535, 0.00000644, 0.00000486, 0.00000465, -0.00000926, 0.00000149, -0.00000397, 0.00000399, -0.00000105, 0.00000285, -0.00000341, 0.00000252, 0.00000259, -0.00000949, 0.00000757, 0.00000289, -0.00000140, -0.00001768, 0.00001970, -0.00001011, 0.00000870, -0.00000254, 0.00000115, -0.00000397, 0.00000474, -0.00000704, 0.00000826, -0.00000258, 0.00000478, -0.00000915, 0.00000748, -0.00000343, 0.00000334, -0.00000234, 0.00000233, 0.00000566, -0.00001186, 0.00000403, 0.00000395, 0.00000111, -0.00000891, 0.00000659, -0.00000491, 0.00002125, -0.00002564, 0.00001655, -0.00001579, -0.00000751, 0.00000389, 0.00000097, -0.00001554, -0.00000621, 0.00001407, -0.00001551, 0.00002161, -0.00002385, 0.00001991, -0.00001236, 0.00000711, -0.00001257, 0.00001178, -0.00001740, 0.00001100, -0.00002013, 0.00001083, -0.00003197, 0.00001048, -0.00004857, -0.00001903, -0.00027626, 0.00051016, 0.00024911, -0.00026842, 0.00001965, -0.00005875, 0.00001069, -0.00002990, 0.00001197, -0.00001867, 0.00000795, -0.00001324, 0.00000228, -0.00000596, 0.00000322, -0.00000602, -0.00000289, -0.00000262, -0.00000507, -0.00001667, 0.00000971, -0.00000880, 0.00003745, -0.00001558, 0.00000934, 0.00002318, -0.00000027, 0.00000657, -0.00001050, -0.00001122, -0.00000306, -0.00000302, 0.00000538, -0.00000758, -0.00000026, 0.00000728, 0.00000868, -0.00001562, -0.00000309, 0.00000770, -0.00000199, -0.00000621, -0.00000140, },
		{ 0.00000771, -0.00000753, 0.00000009, -0.00000936, 0.00000350, -0.00000451, 0.00000848, -0.00001300, 0.00001189, -0.00000724, -0.00001135, 0.00000836, 0.00002135, -0.00001818, -0.00000239, -0.00000377, 0.00000389, 0.00000662, 0.00000441, -0.00000737, -0.00000103, -0.00000171, 0.00000119, 0.00000180, -0.00000009, -0.00000024, -0.00000176, 0.00000814, -0.00001266, 0.00000678, 0.00000233, 0.00000390, -0.00002195, 0.00001833, -0.00000610, 0.00000423, 0.00000128, -0.00000093, -0.00000295, 0.00000421, -0.00000594, 0.00000545, -0.00000013, 0.00000333, -0.00000645, 0.00000353, 0.00000059, -0.00000071, 0.00000178, -0.00000244, 0.00001042, -0.00001410, 0.00000391, 0.00000285, 0.00000471, -0.00001219, 0.00000838, -0.00000744, 0.00002138, -0.00001902, 0.00000558, -0.00000151, -0.00001959, 0.00001082, -0.00000471, -0.00000573, -0.00001863, 0.00002373, -0.00002379, 0.00002876, -0.00002798, 0.00002147, -0.00001410, 0.00001114, -0.00001796, 0.00001754, -0.00002305, 0.00001726, -0.00002732, 0.00002086, -0.00004288, 0.00002512, -0.00006727, 0.00002812, -0.00031150, 0.00037529, 0.00043030, -0.00030833, 0.00001711, -0.00006475, 0.00001426, -0.00003570, 0.00001564, -0.00002185, 0.00001041, -0.00001551, 0.00000469, -0.00000954, 0.00000668, -0.00000870, 0.00000067, -0.00000745, 0.00000192, -0.00002231, 0.00001392, -0.00001462, 0.00003673, -0.00000936, 0.00000094, 0.00002402, 0.00000358, 0.00000296, -0.00000193, -0.00001799, 0.00000156, -0.00000838, 0.00000965, -0.00001023, 0.00000167, 0.00000289, 0.00001358, -0.00001612, -0.00000395, 0.00000557, 0.00000195, -0.00000871, 0.00000009, 0.00000349, },
		{ -0.00000768, 0.00000104, -0.00001028, 0.00000363, -0.00000533, 0.00000935, -0.00001208, 0.00000813, -0.00000122, -0.00001394, 0.00000380, 0.00002341, -0.00001225, -0.00000761, -0.00000192, 0.00000090, 0.00000854, 0.00000392, -0.00000506, -0.00000374, 0.00000054, -0.00000143, 0.00000421, -0.00000233, 0.00000210, -0.00000501, 0.00001155, -0.00001262, 0.00000361, 0.00000272, 0.00000747, -0.00002249, 0.00001287, 0.00000075, -0.00000171, 0.00000558, -0.00000312, -0.00000164, 0.00000288, -0.00000356, 0.00000160, 0.00000289, 0.00000110, -0.00000284, -0.00000096, 0.00000477, -0.00000463, 0.00000548, -0.00000642, 0.00001371, -0.00001442, 0.00000230, 0.00000261, 0.00000704, -0.00001307, 0.00000702, -0.00000595, 0.00001635, -0.00000770, -0.00000796, 0.00001232, -0.00002861, 0.00001436, -0.00000807, 0.00000187, -0.00002742, 0.00002874, -0.00002714, 0.00003041, -0.00002635, 0.00001792, -0.00001155, 0.00001090, -0.00001843, 0.00001767, -0.00002252, 0.00001690, -0.00002736, 0.00002242, -0.00004336, 0.00002681, -0.00006918, 0.00004374, -0.00028563, 0.00018720, 0.00060499, -0.00031181, -0.00000859, -0.00005789, 0.00000848, -0.00003413, 0.00001290, -0.00001986, 0.00000876, -0.00001461, 0.00000463, -0.00001092, 0.00000797, -0.00000966, 0.00000302, -0.00001115, 0.00000714, -0.00002569, 0.00001587, -0.00001729, 0.00003278, -0.00000188, -0.00000543, 0.00002126, 0.00001002, -0.00000267, 0.00000710, -0.00002326, 0.00000435, -0.00001174, 0.00001164, -0.00001070, 0.00000187, 0.00000026, 0.00001599, -0.00001417, -0.00000620, 0.00000432, 0.00000455, -0.00000958, 0.00000031, 0.00000091, 0.00001031, },
		{ -0.00000028, -0.00000883, 0.00000148, -0.00000392, 0.00000762, -0.00000827, 0.00000208, 0.00000531, -0.00001559, -0.00000030, 0.00002277, -0.00000454, -0.00001243, -0.00000063, -0.00000185, 0.00000994, 0.00000376, -0.00000293, -0.00000598, 0.00000210, -0.00000316, 0.00000549, -0.00000323, 0.00000301, -0.00000654, 0.00001223, -0.00000938, -0.00000135, 0.00000372, 0.00000918, -0.00001946, 0.00000439, 0.00000881, -0.00000749, 0.00000904, -0.00000452, -0.00000059, 0.00000117, -0.00000051, -0.00000248, 0.00000575, -0.00000122, 0.00000090, -0.00000513, 0.00000828, -0.00000762, 0.00000804, -0.00000893, 0.00001499, -0.00001261, -0.00000071, 0.00000318, 0.00000800, -0.00001158, 0.00000303, -0.00000161, 0.00000840, 0.00000478, -0.00001973, 0.00002142, -0.00003117, 0.00001214, -0.00000712, 0.00000508, -0.00003037, 0.00002717, -0.00002381, 0.00002502, -0.00001773, 0.00000822, -0.00000353, 0.00000489, -0.00001227, 0.00001041, -0.00001405, 0.00000805, -0.00001808, 0.00001294, -0.00003056, 0.00001239, -0.00005042, 0.00002538, -0.00020798, -0.00002539, 0.00073943, -0.00025864, -0.00006553, -0.00003387, -0.00000999, -0.00002241, 0.00000151, -0.00001091, 0.00000147, -0.00000906, 0.00000058, -0.00000854, 0.00000562, -0.00000753, 0.00000271, -0.00001207, 0.00000897, -0.00002540, 0.00001430, -0.00001588, 0.00002533, 0.00000670, -0.00001016, 0.00001657, 0.00001641, -0.00000706, 0.00001300, -0.00002387, 0.00000277, -0.00001098, 0.00000956, -0.00000750, -0.00000104, 0.00000069, 0.00001480, -0.00000899, -0.00001049, 0.00000478, 0.00000491, -0.00000801, -0.00000150, 0.00000047, 0.00001013, -0.00000519, },
		{ -0.00000560, -0.00000230, -0.00000094, 0.00000405, -0.00000273, -0.00000468, 0.00001083, -0.00001552, -0.00000413, 0.00002011, 0.00000353, -0.00001563, -0.00000057, -0.00000386, 0.00001033, 0.00000433, -0.00000136, -0.00000734, 0.00000259, -0.00000366, 0.00000538, -0.00000265, 0.00000247, -0.00000634, 0.00001040, -0.00000382, -0.00000689, 0.00000455, 0.00000945, -0.00001384, -0.00000530, 0.00001601, -0.00001135, 0.00001045, -0.00000443, -0.00000024, -0.00000044, 0.00000251, -0.00000597, 0.00000771, -0.00000297, 0.00000402, -0.00000822, 0.00001044, -0.00000911, 0.00000900, -0.00000966, 0.00001418, -0.00000895, -0.00000461, 0.00000408, 0.00000799, -0.00000846, -0.00000225, 0.00000353, 0.00000051, 0.00001469, -0.00002595, 0.00002282, -0.00002570, 0.00000368, -0.00000156, 0.00000336, -0.00002701, 0.00001891, -0.00001397, 0.00001299, -0.00000292, -0.00000661, 0.00000909, -0.00000622, -0.00000018, -0.00000335, 0.00000125, -0.00000801, -0.00000082, -0.00000606, -0.00000651, -0.00001529, -0.00001507, -0.00001846, -0.00010063, -0.00022532, 0.00080055, -0.00013537, -0.00015463, 0.00000629, -0.00004049, -0.00000101, -0.00001805, 0.00000452, -0.00001115, 0.00000108, -0.00000767, -0.00000202, -0.00000074, -0.00000189, -0.00000088, -0.00000939, 0.00000658, -0.00002078, 0.00000864, -0.00001013, 0.00001457, 0.00001623, -0.00001404, 0.00001206, 0.00001994, -0.00000723, 0.00001306, -0.00001795, -0.00000424, -0.00000543, 0.00000301, -0.00000046, -0.00000711, 0.00000423, 0.00001013, -0.00000095, -0.00001647, 0.00000680, 0.00000310, -0.00000419, -0.00000513, 0.00000202, 0.00000747, -0.00000058, -0.00000330, },
		{ -0.00000644, 0.00000236, 0.00000003, 0.00000290, -0.00001035, 0.00001391, -0.00001319, -0.00000787, 0.00001629, 0.00001060, -0.00001638, -0.00000194, -0.00000496, 0.00000958, 0.00000566, -0.00000038, -0.00000780, 0.00000205, -0.00000309, 0.00000415, -0.00000100, 0.00000098, -0.00000496, 0.00000695, 0.00000259, -0.00001146, 0.00000428, 0.00000900, -0.00000703, -0.00001416, 0.00002043, -0.00001200, 0.00000916, -0.00000261, -0.00000072, -0.00000164, 0.00000491, -0.00000820, 0.00000828, -0.00000369, 0.00000599, -0.00000974, 0.00001089, -0.00000888, 0.00000826, -0.00000865, 0.00001158, -0.00000406, -0.00000862, 0.00000467, 0.00000763, -0.00000473, -0.00000723, 0.00000738, -0.00000472, 0.00001932, -0.00002465, 0.00001598, -0.00001316, -0.00000923, 0.00000692, -0.00000191, -0.00001888, 0.00000590, 0.00000016, -0.00000310, 0.00001508, -0.00002342, 0.00002327, -0.00001943, 0.00001461, -0.00001994, 0.00001927, -0.00002680, 0.00001953, -0.00002900, 0.00002204, -0.00004784, 0.00002599, -0.00007096, 0.00000720, -0.00037533, 0.00076457, 0.00005830, -0.00026612, 0.00005425, -0.00007703, 0.00002535, -0.00004182, 0.00002305, -0.00002640, 0.00001370, -0.00001846, 0.00000726, -0.00000989, 0.00000624, -0.00000706, -0.00000359, 0.00000049, -0.00001244, -0.00000046, -0.00000083, 0.00000150, 0.00002610, -0.00001757, 0.00000934, 0.00001884, -0.00000180, 0.00000660, -0.00000594, -0.00001555, 0.00000357, -0.00000651, 0.00000885, -0.00001484, 0.00000947, 0.00000349, 0.00000831, -0.00002265, 0.00000917, 0.00000025, 0.00000069, -0.00000944, 0.00000448, 0.00000354, 0.00000476, -0.00000670, -0.00000186, },
		{ 0.00000454, -0.00000298, 0.00000699, -0.00001342, 0.00001366, -0.00000854, -0.00001149, 0.00001201, 0.00001575, -0.00001446, -0.00000446, -0.00000538, 0.00000788, 0.00000743, 0.00000030, -0.00000763, 0.00000082, -0.00000190, 0.00000238, 0.00000099, -0.00000067, -0.00000328, 0.00000307, 0.00000825, -0.00001370, 0.00000222, 0.00000849, -0.00000044, -0.00002041, 0.00002082, -0.00000900, 0.00000531, 0.00000059, -0.00000182, -0.00000233, 0.00000633, -0.00000881, 0.00000726, -0.00000324, 0.00000659, -0.00000951, 0.00000959, -0.00000702, 0.00000604, -0.00000625, 0.00000773, 0.00000127, -0.00001197, 0.00000442, 0.00000743, -0.00000129, -0.00001066, 0.00000855, -0.00000592, 0.00001785, -0.00001637, 0.00000303, 0.00000301, -0.00002282, 0.00001503, -0.00000782, -0.00000908, -0.00000833, 0.00001475, -0.00001908, 0.00003175, -0.00003760, 0.00003454, -0.00003022, 0.00002716, -0.00003392, 0.00003410, -0.00004191, 0.00003599, -0.00004792, 0.00004564, -0.00007391, 0.00005864, -0.00011239, 0.00008697, -0.00044732, 0.00062470, 0.00030551, -0.00037836, 0.00009445, -0.00010865, 0.00004813, -0.00006273, 0.00003879, -0.00003942, 0.00002479, -0.00002839, 0.00001626, -0.00001904, 0.00001438, -0.00001371, 0.00000344, -0.00000741, -0.00000236, -0.00001107, 0.00000998, -0.00001185, 0.00003486, -0.00002039, 0.00000868, 0.00001322, 0.00000833, -0.00000448, 0.00000924, -0.00002785, 0.00001284, -0.00001594, 0.00001749, -0.00002152, 0.00001389, -0.00000264, 0.00001631, -0.00002682, 0.00001005, -0.00000189, 0.00000485, -0.00001271, 0.00000622, 0.00000008, 0.00000908, -0.00000889, 0.00000090, -0.00000948, },
		{ -0.00000389, 0.00000847, -0.00001312, 0.00001005, -0.00000219, -0.00001459, 0.00000764, 0.00001861, -0.00001030, -0.00000738, -0.00000560, 0.00000568, 0.00000916, 0.00000109, -0.00000722, -0.00000060, -0.00000068, 0.00000076, 0.00000257, -0.00000171, -0.00000211, -0.00000015, 0.00001194, -0.00001287, -0.00000171, 0.00000822, 0.00000494, -0.00002305, 0.00001696, -0.00000292, -0.00000016, 0.00000429, -0.00000299, -0.00000265, 0.00000668, -0.00000780, 0.00000483, -0.00000176, 0.00000587, -0.00000769, 0.00000682, -0.00000387, 0.00000275, -0.00000295, 0.00000326, 0.00000634, -0.00001412, 0.00000312, 0.00000755, 0.00000147, -0.00001206, 0.00000680, -0.00000336, 0.00001160, -0.00000392, -0.00001178, 0.00001788, -0.00003249, 0.00001903, -0.00001104, -0.00000120, -0.00001979, 0.00002557, -0.00003043, 0.00004234, -0.00004449, 0.00003840, -0.00003397, 0.00003248, -0.00003984, 0.00003990, -0.00004717, 0.00004184, -0.00005519, 0.00005542, -0.00008311, 0.00007041, -0.00012654, 0.00011872, -0.00042994, 0.00039545, 0.00057213, -0.00046040, 0.00010714, -0.00012208, 0.00005712, -0.00007243, 0.00004484, -0.00004456, 0.00002952, -0.00003319, 0.00002109, -0.00002455, 0.00001923, -0.00001790, 0.00000890, -0.00001423, 0.00000655, -0.00002035, 0.00001942, -0.00002272, 0.00004041, -0.00002123, 0.00000896, 0.00000509, 0.00002005, -0.00001615, 0.00002293, -0.00003664, 0.00001836, -0.00002163, 0.00002215, -0.00002418, 0.00001479, -0.00000564, 0.00002056, -0.00002688, 0.00000766, -0.00000163, 0.00000656, -0.00001333, 0.00000572, -0.00000135, 0.00001087, -0.00000857, 0.00000151, -0.00001030, 0.00000458, },
		{ 0.00000717, -0.00000970, 0.00000398, 0.00000465, -0.00001649, 0.00000328, 0.00001932, -0.00000481, -0.00000979, -0.00000613, 0.00000343, 0.00001037, 0.00000235, -0.00000688, -0.00000182, 0.00000011, -0.00000023, 0.00000325, -0.00000169, -0.00000188, -0.00000213, 0.00001309, -0.00000906, -0.00000689, 0.00000793, 0.00000874, -0.00002199, 0.00000970, 0.00000475, -0.00000573, 0.00000734, -0.00000361, -0.00000284, 0.00000609, -0.00000546, 0.00000145, 0.00000037, 0.00000415, -0.00000467, 0.00000305, 0.00000005, -0.00000107, 0.00000073, -0.00000127, 0.00001064, -0.00001488, 0.00000096, 0.00000773, 0.00000367, -0.00001175, 0.00000303, 0.00000132, 0.00000341, 0.00000852, -0.00002351, 0.00002661, -0.00003437, 0.00001612, -0.00000912, 0.00000196, -0.00002532, 0.00002934, -0.00003374, 0.00004341, -0.00004094, 0.00003201, -0.00002778, 0.00002738, -0.00003430, 0.00003316, -0.00003894, 0.00003321, -0.00004642, 0.00004635, -0.00006986, 0.00005509, -0.00010608, 0.00009673, -0.00033215, 0.00011190, 0.00081167, -0.00047827, 0.00007384, -0.00010597, 0.00004388, -0.00006417, 0.00003575, -0.00003731, 0.00002401, -0.00002930, 0.00001835, -0.00002321, 0.00001785, -0.00001687, 0.00001003, -0.00001708, 0.00001134, -0.00002538, 0.00002454, -0.00002833, 0.00004057, -0.00001839, 0.00000808, -0.00000228, 0.00002915, -0.00002362, 0.00003025, -0.00003773, 0.00001676, -0.00002071, 0.00002035, -0.00002072, 0.00001037, -0.00000379, 0.00001950, -0.00002164, 0.00000112, 0.00000194, 0.00000490, -0.00001048, 0.00000225, 0.00000004, 0.00000945, -0.00000529, -0.00000038, -0.00000858, 0.00000218, -0.00000245, },
		{ -0.00000432, -0.00000294, 0.00001046, -0.00001649, -0.00000114, 0.00001836, 0.00000094, -0.00001086, -0.00000730, 0.00000138, 0.00001078, 0.00000418, -0.00000668, -0.00000272, 0.00000034, -0.00000046, 0.00000297, -0.00000068, -0.00000248, -0.00000289, 0.00001197, -0.00000322, -0.00001210, 0.00000702, 0.00001113, -0.00001797, 0.00000069, 0.00001207, -0.00000979, 0.00000872, -0.00000317, -0.00000313, 0.00000480, -0.00000228, -0.00000226, 0.00000263, 0.00000185, -0.00000097, -0.00000112, 0.00000416, -0.00000487, 0.00000427, -0.00000538, 0.00001384, -0.00001434, -0.00000158, 0.00000746, 0.00000577, -0.00001057, -0.00000124, 0.00000578, -0.00000344, 0.00001690, -0.00002812, 0.00002608, -0.00002680, 0.00000571, -0.00000159, -0.00000049, -0.00002380, 0.00002494, -0.00002792, 0.00003413, -0.00002662, 0.00001543, -0.00001179, 0.00001191, -0.00001743, 0.00001425, -0.00001784, 0.00001084, -0.00002240, 0.00001943, -0.00003578, 0.00001527, -0.00005520, 0.00003100, -0.00018157, -0.00017661, 0.00097451, -0.00040385, -0.00001580, -0.00005591, 0.00000550, -0.00003581, 0.00000994, -0.00001650, 0.00000712, -0.00001550, 0.00000671, -0.00001363, 0.00000893, -0.00000932, 0.00000530, -0.00001417, 0.00001011, -0.00002426, 0.00002335, -0.00002684, 0.00003392, -0.00001059, 0.00000404, -0.00000572, 0.00003181, -0.00002304, 0.00002791, -0.00002893, 0.00000677, -0.00001236, 0.00001158, -0.00001095, 0.00000058, 0.00000294, 0.00001322, -0.00001149, -0.00000903, 0.00000834, 0.00000023, -0.00000460, -0.00000375, 0.00000386, 0.00000529, 0.00000030, -0.00000416, -0.00000489, -0.00000203, 0.00000072, 0.00000385, },
		{ -0.00000888, 0.00001380, -0.00001411, -0.00000563, 0.00001629, 0.00000598, -0.00001017, -0.00000902, -0.00000048, 0.00001045, 0.00000636, -0.00000638, -0.00000347, 0.00000017, -0.00000018, 0.00000209, 0.00000082, -0.00000340, -0.00000294, 0.00000943, 0.00000319, -0.00001588, 0.00000474, 0.00001259, -0.00001223, -0.00000810, 0.00001716, -0.00001110, 0.00000784, -0.00000153, -0.00000354, 0.00000307, 0.00000119, -0.00000567, 0.00000454, -0.00000053, 0.00000281, -0.00000506, 0.00000784, -0.00000812, 0.00000720, -0.00000867, 0.00001577, -0.00001274, -0.00000400, 0.00000631, 0.00000821, -0.00000937, -0.00000454, 0.00000802, -0.00000643, 0.00001869, -0.00002395, 0.00001611, -0.00001112, -0.00001004, 0.00000960, -0.00000711, -0.00001653, 0.00001387, -0.00001471, 0.00001671, -0.00000444, -0.00000794, 0.00001047, -0.00001024, 0.00000672, -0.00001223, 0.00001097, -0.00001958, 0.00001061, -0.00001811, 0.00001029, -0.00003796, 0.00001148, -0.00005580, -0.00001771, -0.00041707, 0.00101961, -0.00022434, -0.00015791, 0.00002181, -0.00005269, 0.00000816, -0.00002859, 0.00001440, -0.00001834, 0.00000601, -0.00001208, 0.00000271, -0.00000626, 0.00000376, -0.00000469, -0.00000573, 0.00000293, -0.00001694, 0.00001568, -0.00001819, 0.00002070, 0.00000209, -0.00000398, -0.00000345, 0.00002617, -0.00001311, 0.00001558, -0.00001119, -0.00000995, 0.00000158, -0.00000221, 0.00000310, -0.00001253, 0.00001262, 0.00000364, 0.00000145, -0.00002069, 0.00001576, -0.00000579, 0.00000268, -0.00001070, 0.00000864, -0.00000008, 0.00000662, -0.00000837, -0.00000060, -0.00000672, 0.00000431, -0.00000007, 0.00000131, },
		{ 0.00001385, -0.00000941, -0.00000996, 0.00001352, 0.00000972, -0.00000777, -0.00001082, -0.00000245, 0.00000963, 0.00000845, -0.00000560, -0.00000437, -0.00000003, 0.00000014, 0.00000121, 0.00000212, -0.00000389, -0.00000303, 0.00000655, 0.00000867, -0.00001701, 0.00000067, 0.00001355, -0.00000604, -0.00001496, 0.00001877, -0.00000919, 0.00000484, 0.00000097, -0.00000393, 0.00000109, 0.00000444, -0.00000822, 0.00000567, -0.00000257, 0.00000609, -0.00000815, 0.00001053, -0.00001028, 0.00000906, -0.00001076, 0.00001626, -0.00001022, -0.00000600, 0.00000416, 0.00001102, -0.00000857, -0.00000607, 0.00000710, -0.00000473, 0.00001375, -0.00001228, -0.00000033, 0.00000840, -0.00002663, 0.00002053, -0.00001453, -0.00000686, -0.00000018, 0.00000177, -0.00000408, 0.00002006, -0.00003211, 0.00003284, -0.00003265, 0.00003116, -0.00003861, 0.00003912, -0.00004898, 0.00004263, -0.00005484, 0.00005480, -0.00008824, 0.00007348, -0.00013519, 0.00011821, -0.00056525, 0.00092562, 0.00005079, -0.00033124, 0.00010900, -0.00011680, 0.00005647, -0.00007035, 0.00004738, -0.00004568, 0.00002958, -0.00003312, 0.00002146, -0.00002376, 0.00001890, -0.00001706, 0.00000575, -0.00000792, -0.00000556, 0.00000348, -0.00000441, 0.00000310, 0.00001778, -0.00001495, 0.00000401, 0.00001321, 0.00000421, -0.00000359, 0.00001133, -0.00002896, 0.00001688, -0.00001696, 0.00001752, -0.00002532, 0.00002190, -0.00000596, 0.00001392, -0.00003078, 0.00002146, -0.00001068, 0.00000895, -0.00001634, 0.00001231, -0.00000460, 0.00001168, -0.00001122, 0.00000266, -0.00001030, 0.00000679, -0.00000303, 0.00000550, -0.00001229, },
		{ -0.00000309, -0.00001361, 0.00001025, 0.00001199, -0.00000423, -0.00001200, -0.00000489, 0.00000869, 0.00000999, -0.00000393, -0.00000573, 0.00000007, 0.00000007, 0.00000085, 0.00000261, -0.00000341, -0.00000374, 0.00000424, 0.00001210, -0.00001493, -0.00000506, 0.00001412, -0.00000034, -0.00001882, 0.00001658, -0.00000453, 0.00000050, 0.00000365, -0.00000397, -0.00000098, 0.00000705, -0.00000953, 0.00000575, -0.00000390, 0.00000835, -0.00000985, 0.00001174, -0.00001091, 0.00000943, -0.00001127, 0.00001509, -0.00000680, -0.00000766, 0.00000141, 0.00001371, -0.00000788, -0.00000604, 0.00000352, 0.00000058, 0.00000435, 0.00000301, -0.00001797, 0.00002584, -0.00003853, 0.00002661, -0.00001872, 0.00000098, -0.00001254, 0.00001634, -0.00002235, 0.00004034, -0.00005023, 0.00004837, -0.00004807, 0.00004808, -0.00005645, 0.00005757, -0.00006765, 0.00006304, -0.00007868, 0.00008373, -0.00011913, 0.00011100, -0.00018149, 0.00019399, -0.00059653, 0.00069836, 0.00038749, -0.00049862, 0.00017836, -0.00016701, 0.00009352, -0.00010245, 0.00007165, -0.00006586, 0.00004743, -0.00004954, 0.00003638, -0.00003788, 0.00003103, -0.00002731, 0.00001616, -0.00001850, 0.00000610, -0.00000957, 0.00001068, -0.00001498, 0.00003299, -0.00002601, 0.00001384, -0.00000334, 0.00002408, -0.00002378, 0.00003226, -0.00004421, 0.00002814, -0.00002777, 0.00002775, -0.00003362, 0.00002700, -0.00001203, 0.00002242, -0.00003612, 0.00002275, -0.00001200, 0.00001189, -0.00001853, 0.00001291, -0.00000638, 0.00001370, -0.00001121, 0.00000352, -0.00001149, 0.00000698, -0.00000383, 0.00000711, -0.00001239, 0.00001060, },
		{ -0.00001586, 0.00000648, 0.00001297, -0.00000035, -0.00001188, -0.00000803, 0.00000784, 0.00001066, -0.00000126, -0.00000756, 0.00000057, -0.00000058, 0.00000125, 0.00000206, -0.00000179, -0.00000522, 0.00000289, 0.00001311, -0.00000997, -0.00001161, 0.00001398, 0.00000444, -0.00001946, 0.00001124, 0.00000166, -0.00000393, 0.00000565, -0.00000336, -0.00000303, 0.00000870, -0.00000939, 0.00000467, -0.00000428, 0.00000920, -0.00000979, 0.00001109, -0.00000965, 0.00000798, -0.00000989, 0.00001207, -0.00000236, -0.00000930, -0.00000117, 0.00001534, -0.00000648, -0.00000550, -0.00000106, 0.00000693, -0.00000567, 0.00001673, -0.00003086, 0.00003546, -0.00004105, 0.00002429, -0.00001651, 0.00000351, -0.00001920, 0.00002448, -0.00003299, 0.00005082, -0.00005670, 0.00005157, -0.00005086, 0.00005144, -0.00005933, 0.00005960, -0.00006852, 0.00006417, -0.00008099, 0.00008721, -0.00011951, 0.00011131, -0.00017957, 0.00019487, -0.00051234, 0.00037208, 0.00073132, -0.00061365, 0.00019988, -0.00018304, 0.00010382, -0.00011236, 0.00007694, -0.00007032, 0.00005211, -0.00005460, 0.00004127, -0.00004293, 0.00003501, -0.00003078, 0.00002106, -0.00002445, 0.00001374, -0.00001921, 0.00002263, -0.00002899, 0.00004350, -0.00003332, 0.00002184, -0.00001817, 0.00004020, -0.00003821, 0.00004500, -0.00005004, 0.00003066, -0.00003051, 0.00003013, -0.00003420, 0.00002511, -0.00001191, 0.00002446, -0.00003460, 0.00001792, -0.00000826, 0.00001011, -0.00001606, 0.00000940, -0.00000441, 0.00001185, -0.00000774, 0.00000149, -0.00000980, 0.00000445, -0.00000211, 0.00000590, -0.00000940, 0.00000494, 0.00000362, },
		{ 0.00000222, 0.00001300, 0.00000308, -0.00001005, -0.00001174, 0.00000700, 0.00001050, 0.00000219, -0.00000956, 0.00000123, -0.00000160, 0.00000223, 0.00000072, 0.00000060, -0.00000711, 0.00000228, 0.00001209, -0.00000321, -0.00001761, 0.00001249, 0.00000835, -0.00001741, 0.00000415, 0.00000769, -0.00000712, 0.00000625, -0.00000197, -0.00000483, 0.00000918, -0.00000777, 0.00000250, -0.00000361, 0.00000842, -0.00000780, 0.00000850, -0.00000644, 0.00000464, -0.00000655, 0.00000719, 0.00000314, -0.00001132, -0.00000275, 0.00001497, -0.00000340, -0.00000577, -0.00000455, 0.00001137, -0.00001233, 0.00002418, -0.00003443, 0.00003366, -0.00003207, 0.00001255, -0.00000702, -0.00000065, -0.00001826, 0.00002391, -0.00003329, 0.00004872, -0.00004907, 0.00004035, -0.00003898, 0.00003914, -0.00004518, 0.00004326, -0.00004977, 0.00004413, -0.00005965, 0.00006302, -0.00008746, 0.00007306, -0.00012932, 0.00012696, -0.00033994, 0.00000359, 0.00101755, -0.00063144, 0.00014988, -0.00015099, 0.00007743, -0.00009241, 0.00005718, -0.00005408, 0.00003922, -0.00004422, 0.00003230, -0.00003536, 0.00002762, -0.00002445, 0.00001740, -0.00002259, 0.00001411, -0.00002203, 0.00002774, -0.00003513, 0.00004577, -0.00003342, 0.00002391, -0.00002615, 0.00004696, -0.00004142, 0.00004489, -0.00004312, 0.00002216, -0.00002343, 0.00002324, -0.00002601, 0.00001546, -0.00000493, 0.00001942, -0.00002593, 0.00000704, 0.00000037, 0.00000376, -0.00000917, 0.00000208, 0.00000099, 0.00000653, -0.00000138, -0.00000291, -0.00000573, -0.00000026, 0.00000155, 0.00000260, -0.00000443, -0.00000157, 0.00000924, -0.00001603, },
		{ 0.00001239, 0.00000551, -0.00000659, -0.00001547, 0.00000578, 0.00000985, 0.00000583, -0.00001108, 0.00000152, -0.00000251, 0.00000324, -0.00000077, 0.00000299, -0.00000864, 0.00000178, 0.00000999, 0.00000376, -0.00002156, 0.00000902, 0.00001166, -0.00001364, -0.00000302, 0.00001200, -0.00000813, 0.00000517, 0.00000006, -0.00000613, 0.00000837, -0.00000484, -0.00000053, -0.00000199, 0.00000608, -0.00000406, 0.00000421, -0.00000154, -0.00000030, -0.00000156, 0.00000082, 0.00000942, -0.00001383, -0.00000284, 0.00001206, 0.00000186, -0.00000777, -0.00000540, 0.00001173, -0.00001297, 0.00002279, -0.00002705, 0.00002036, -0.00001309, -0.00000634, 0.00000772, -0.00001012, -0.00001067, 0.00001534, -0.00002390, 0.00003498, -0.00002896, 0.00001693, -0.00001495, 0.00001394, -0.00001721, 0.00001236, -0.00001574, 0.00000778, -0.00002007, 0.00001761, -0.00003116, 0.00000696, -0.00004550, 0.00001491, -0.00012564, -0.00033966, 0.00118546, -0.00052112, 0.00001992, -0.00006957, 0.00001460, -0.00004336, 0.00001349, -0.00001828, 0.00000962, -0.00001895, 0.00000979, -0.00001539, 0.00000907, -0.00000838, 0.00000489, -0.00001230, 0.00000633, -0.00001686, 0.00002442, -0.00003169, 0.00003825, -0.00002469, 0.00001773, -0.00002422, 0.00004136, -0.00003120, 0.00003102, -0.00002392, 0.00000386, -0.00000797, 0.00000863, -0.00001075, -0.00000016, 0.00000721, 0.00000898, -0.00001199, -0.00000785, 0.00001194, -0.00000536, 0.00000038, -0.00000731, 0.00000821, -0.00000063, 0.00000624, -0.00000820, -0.00000064, -0.00000580, 0.00000585, -0.00000139, 0.00000085, -0.00000718, 0.00001249, -0.00001380, -0.00000241, },
		{ 0.00000679, -0.00000207, -0.00001835, 0.00000361, 0.00000925, 0.00000894, -0.00001141, 0.00000084, -0.00000273, 0.00000363, -0.00000166, 0.00000452, -0.00000901, 0.00000062, 0.00000790, 0.00000943, -0.00002230, 0.00000331, 0.00001455, -0.00000916, -0.00000879, 0.00001354, -0.00000667, 0.00000267, 0.00000227, -0.00000659, 0.00000631, -0.00000095, -0.00000401, 0.00000028, 0.00000255, 0.00000088, -0.00000113, 0.00000431, -0.00000608, 0.00000428, -0.00000620, 0.00001573, -0.00001642, -0.00000161, 0.00000692, 0.00000892, -0.00001146, -0.00000325, 0.00000750, -0.00000723, 0.00001302, -0.00001074, -0.00000072, 0.00001092, -0.00002720, 0.00002320, -0.00002108, 0.00000005, 0.00000232, -0.00000859, 0.00001404, -0.00000175, -0.00001259, 0.00001469, -0.00001712, 0.00001681, -0.00002450, 0.00002408, -0.00003449, 0.00002622, -0.00003568, 0.00003341, -0.00006742, 0.00004685, -0.00010563, 0.00007701, -0.00059482, 0.00119307, -0.00027659, -0.00017695, 0.00004646, -0.00007227, 0.00002420, -0.00004490, 0.00002906, -0.00002992, 0.00001545, -0.00002118, 0.00001238, -0.00001646, 0.00001379, -0.00001343, 0.00000390, -0.00000748, -0.00000536, 0.00001389, -0.00001980, 0.00002220, -0.00000822, 0.00000379, -0.00001261, 0.00002429, -0.00000960, 0.00000678, 0.00000307, -0.00001948, 0.00001139, -0.00000944, 0.00000746, -0.00001779, 0.00002085, -0.00000337, 0.00000366, -0.00002322, 0.00002323, -0.00001428, 0.00000974, -0.00001612, 0.00001480, -0.00000726, 0.00001289, -0.00001242, 0.00000371, -0.00001048, 0.00000916, -0.00000446, 0.00000478, -0.00001044, 0.00001264, -0.00000936, -0.00000698, 0.00001125, },
		{ 0.00000262, -0.00001946, 0.00000000, 0.00000914, 0.00001088, -0.00001001, -0.00000119, -0.00000187, 0.00000293, -0.00000138, 0.00000456, -0.00000769, -0.00000167, 0.00000662, 0.00001274, -0.00001939, -0.00000429, 0.00001689, -0.00000471, -0.00001228, 0.00001210, -0.00000320, -0.00000051, 0.00000403, -0.00000599, 0.00000321, 0.00000336, -0.00000737, 0.00000269, -0.00000150, 0.00000612, -0.00000648, 0.00001000, -0.00001150, 0.00000972, -0.00001255, 0.00002087, -0.00001814, 0.00000004, 0.00000072, 0.00001642, -0.00001563, 0.00000075, 0.00000012, 0.00000276, -0.00000169, 0.00000943, -0.00002311, 0.00003291, -0.00004345, 0.00003380, -0.00002863, 0.00000907, -0.00001007, 0.00000706, -0.00000763, 0.00002508, -0.00004003, 0.00004132, -0.00004484, 0.00004687, -0.00005641, 0.00005789, -0.00006991, 0.00006510, -0.00008071, 0.00008733, -0.00012766, 0.00012026, -0.00019847, 0.00022080, -0.00071828, 0.00102873, 0.00007785, -0.00040520, 0.00016754, -0.00016028, 0.00009163, -0.00010225, 0.00007462, -0.00006809, 0.00004910, -0.00005179, 0.00003994, -0.00004168, 0.00003555, -0.00003185, 0.00002090, -0.00002268, 0.00000829, -0.00000006, -0.00000322, 0.00000157, 0.00001223, -0.00001447, 0.00000509, 0.00000039, 0.00001746, -0.00002088, 0.00003046, -0.00004093, 0.00002842, -0.00002528, 0.00002327, -0.00003245, 0.00003145, -0.00001337, 0.00001683, -0.00003508, 0.00003062, -0.00001979, 0.00001594, -0.00002160, 0.00001829, -0.00001105, 0.00001648, -0.00001381, 0.00000573, -0.00001281, 0.00001013, -0.00000534, 0.00000623, -0.00001064, 0.00000979, -0.00000349, -0.00001083, 0.00000950, 0.00000712, },
		{ -0.00001810, -0.00000515, 0.00000966, 0.00001136, -0.00000677, -0.00000451, 0.00000004, 0.00000111, 0.00000018, 0.00000304, -0.00000470, -0.00000504, 0.00000637, 0.00001347, -0.00001333, -0.00001270, 0.00001815, -0.00000056, -0.00001333, 0.00000828, 0.00000120, -0.00000335, 0.00000475, -0.00000427, -0.00000047, 0.00000740, -0.00000994, 0.00000458, -0.00000522, 0.00001055, -0.00001062, 0.00001423, -0.00001519, 0.00001332, -0.00001673, 0.00002340, -0.00001771, 0.00000073, -0.00000474, 0.00002232, -0.00001828, 0.00000442, -0.00000758, 0.00001317, -0.00001614, 0.00002689, -0.00003958, 0.00004589, -0.00004919, 0.00003484, -0.00002858, 0.00001206, -0.00001703, 0.00001759, -0.00002361, 0.00004421, -0.00005761, 0.00005687, -0.00006071, 0.00006381, -0.00007356, 0.00007527, -0.00008732, 0.00008433, -0.00010359, 0.00011465, -0.00015541, 0.00015317, -0.00023722, 0.00027653, -0.00069583, 0.00071589, 0.00048903, -0.00061230, 0.00025547, -0.00022169, 0.00013706, -0.00014056, 0.00010336, -0.00009213, 0.00007085, -0.00007199, 0.00005812, -0.00005824, 0.00004943, -0.00004365, 0.00003252, -0.00003349, 0.00001874, -0.00001230, 0.00001272, -0.00001809, 0.00003116, -0.00003169, 0.00002297, -0.00002328, 0.00004198, -0.00004352, 0.00005014, -0.00005345, 0.00003717, -0.00003360, 0.00003184, -0.00003974, 0.00003508, -0.00001738, 0.00002395, -0.00004015, 0.00003135, -0.00001948, 0.00001681, -0.00002185, 0.00001703, -0.00001052, 0.00001577, -0.00001143, 0.00000461, -0.00001205, 0.00000814, -0.00000355, 0.00000493, -0.00000799, 0.00000481, 0.00000258, -0.00001320, 0.00000700, 0.00000691, 0.00000653, },
		{ -0.00001143, 0.00001045, 0.00001062, -0.00000214, -0.00000854, 0.00000248, -0.00000139, 0.00000258, 0.00000043, -0.00000068, -0.00000885, 0.00000672, 0.00001218, -0.00000541, -0.00002035, 0.00001755, 0.00000338, -0.00001238, 0.00000326, 0.00000519, -0.00000493, 0.00000410, -0.00000169, -0.00000411, 0.00001042, -0.00001107, 0.00000530, -0.00000770, 0.00001308, -0.00001242, 0.00001581, -0.00001595, 0.00001382, -0.00001744, 0.00002214, -0.00001406, -0.00000086, -0.00000769, 0.00002464, -0.00001737, 0.00000534, -0.00001243, 0.00001988, -0.00002512, 0.00003579, -0.00004446, 0.00004526, -0.00004134, 0.00002435, -0.00001919, 0.00000685, -0.00001582, 0.00001956, -0.00002963, 0.00005080, -0.00006038, 0.00005648, -0.00005968, 0.00006230, -0.00007043, 0.00007055, -0.00008083, 0.00007752, -0.00009713, 0.00010737, -0.00014209, 0.00013646, -0.00021241, 0.00023964, -0.00054562, 0.00030952, 0.00088424, -0.00073991, 0.00027348, -0.00023202, 0.00014208, -0.00014514, 0.00010332, -0.00009203, 0.00007189, -0.00007380, 0.00005961, -0.00005952, 0.00004949, -0.00004345, 0.00003367, -0.00003506, 0.00002132, -0.00001811, 0.00002292, -0.00003136, 0.00004311, -0.00004230, 0.00003477, -0.00003951, 0.00005631, -0.00005384, 0.00005587, -0.00005229, 0.00003398, -0.00003134, 0.00003045, -0.00003729, 0.00002979, -0.00001361, 0.00002331, -0.00003699, 0.00002441, -0.00001264, 0.00001174, -0.00001646, 0.00001077, -0.00000552, 0.00001079, -0.00000550, 0.00000059, -0.00000840, 0.00000351, 0.00000047, 0.00000152, -0.00000353, -0.00000087, 0.00000752, -0.00001350, 0.00000371, 0.00000654, 0.00000892, -0.00001410, },
		{ 0.00001077, 0.00000930, 0.00000299, -0.00001228, 0.00000454, -0.00000374, 0.00000498, -0.00000232, 0.00000328, -0.00001204, 0.00000683, 0.00000996, 0.00000262, -0.00002560, 0.00001439, 0.00000732, -0.00001023, -0.00000162, 0.00000761, -0.00000471, 0.00000220, 0.00000120, -0.00000697, 0.00001176, -0.00001028, 0.00000438, -0.00000830, 0.00001296, -0.00001117, 0.00001409, -0.00001311, 0.00001061, -0.00001411, 0.00001663, -0.00000678, -0.00000539, -0.00000705, 0.00002215, -0.00001171, 0.00000196, -0.00001222, 0.00002000, -0.00002526, 0.00003294, -0.00003565, 0.00003048, -0.00002094, 0.00000411, -0.00000206, -0.00000558, -0.00000682, 0.00001276, -0.00002492, 0.00004396, -0.00004784, 0.00004001, -0.00004183, 0.00004258, -0.00004759, 0.00004472, -0.00005179, 0.00004628, -0.00006315, 0.00006795, -0.00009149, 0.00007594, -0.00013352, 0.00012991, -0.00031297, -0.00011539, 0.00118654, -0.00073848, 0.00019765, -0.00017865, 0.00009835, -0.00010995, 0.00007002, -0.00006426, 0.00004906, -0.00005425, 0.00004169, -0.00004308, 0.00003363, -0.00002929, 0.00002234, -0.00002528, 0.00001382, -0.00001497, 0.00002429, -0.00003472, 0.00004449, -0.00004247, 0.00003602, -0.00004325, 0.00005568, -0.00004806, 0.00004538, -0.00003667, 0.00001891, -0.00001878, 0.00001950, -0.00002567, 0.00001632, -0.00000280, 0.00001557, -0.00002647, 0.00001098, -0.00000056, 0.00000204, -0.00000675, 0.00000090, 0.00000261, 0.00000293, 0.00000256, -0.00000510, -0.00000304, -0.00000255, 0.00000550, -0.00000268, 0.00000122, -0.00000566, 0.00001023, -0.00001150, -0.00000017, 0.00000620, 0.00000954, -0.00000796, -0.00001784, },
		{ 0.00000824, 0.00000752, -0.00001461, 0.00000525, -0.00000502, 0.00000642, -0.00000416, 0.00000605, -0.00001356, 0.00000577, 0.00000801, 0.00000913, -0.00002718, 0.00000843, 0.00001131, -0.00000763, -0.00000523, 0.00000781, -0.00000275, -0.00000041, 0.00000366, -0.00000838, 0.00001103, -0.00000745, 0.00000173, -0.00000679, 0.00001001, -0.00000687, 0.00000916, -0.00000693, 0.00000401, -0.00000717, 0.00000749, 0.00000344, -0.00001242, -0.00000296, 0.00001506, -0.00000158, -0.00000569, -0.00000662, 0.00001303, -0.00001626, 0.00001898, -0.00001536, 0.00000549, 0.00000688, -0.00002052, 0.00001810, -0.00002125, 0.00000652, 0.00000027, -0.00001238, 0.00002693, -0.00002398, 0.00001221, -0.00001240, 0.00001046, -0.00001161, 0.00000522, -0.00000850, -0.00000025, -0.00001192, 0.00000851, -0.00001846, -0.00000970, -0.00002535, -0.00001522, -0.00005846, -0.00048053, 0.00133240, -0.00058140, 0.00002645, -0.00006692, 0.00001183, -0.00004080, 0.00000904, -0.00001383, 0.00000665, -0.00001704, 0.00000767, -0.00001205, 0.00000478, -0.00000377, 0.00000068, -0.00000591, -0.00000240, -0.00000362, 0.00001682, -0.00002778, 0.00003481, -0.00003151, 0.00002561, -0.00003336, 0.00003983, -0.00002730, 0.00002127, -0.00001033, -0.00000406, 0.00000034, 0.00000246, -0.00000827, -0.00000198, 0.00001194, 0.00000364, -0.00001154, -0.00000585, 0.00001373, -0.00000947, 0.00000456, -0.00001002, 0.00001145, -0.00000549, 0.00001059, -0.00001055, 0.00000233, -0.00000840, 0.00000993, -0.00000609, 0.00000472, -0.00000835, 0.00001020, -0.00000758, -0.00000411, 0.00000579, 0.00000859, -0.00000072, -0.00002305, 0.00000968, },
		{ 0.00001053, -0.00001468, 0.00000395, -0.00000460, 0.00000623, -0.00000433, 0.00000680, -0.00001273, 0.00000300, 0.00000718, 0.00001302, -0.00002464, 0.00000010, 0.00001505, -0.00000505, -0.00000702, 0.00000587, 0.00000028, -0.00000286, 0.00000496, -0.00000796, 0.00000824, -0.00000290, -0.00000226, -0.00000353, 0.00000479, -0.00000033, 0.00000204, 0.00000143, -0.00000461, 0.00000186, -0.00000355, 0.00001479, -0.00002034, 0.00000307, 0.00000508, 0.00001105, -0.00001572, 0.00000254, 0.00000110, -0.00000101, -0.00000178, 0.00001037, -0.00002235, 0.00003421, -0.00004214, 0.00003486, -0.00003451, 0.00001889, -0.00001269, 0.00000257, 0.00000590, 0.00000394, -0.00001881, 0.00001983, -0.00002450, 0.00002695, -0.00003634, 0.00003635, -0.00004817, 0.00004107, -0.00005305, 0.00005572, -0.00009480, 0.00007992, -0.00015194, 0.00015745, -0.00072204, 0.00128718, -0.00027488, -0.00021492, 0.00007851, -0.00009703, 0.00004497, -0.00006460, 0.00004623, -0.00004413, 0.00002796, -0.00003350, 0.00002534, -0.00002948, 0.00002630, -0.00002528, 0.00001765, -0.00002256, 0.00001194, 0.00000378, -0.00001349, 0.00001705, -0.00001230, 0.00000627, -0.00001300, 0.00001314, 0.00000265, -0.00000953, 0.00001944, -0.00002812, 0.00001994, -0.00001517, 0.00000975, -0.00002020, 0.00002609, -0.00000830, 0.00000362, -0.00002193, 0.00002637, -0.00001927, 0.00001419, -0.00001895, 0.00001826, -0.00001192, 0.00001630, -0.00001386, 0.00000600, -0.00001240, 0.00001228, -0.00000743, 0.00000594, -0.00000837, 0.00000767, -0.00000259, -0.00000735, 0.00000505, 0.00000667, 0.00000627, -0.00002572, 0.00000639, 0.00000815, },
		{ -0.00001220, 0.00000056, -0.00000242, 0.00000427, -0.00000264, 0.00000536, -0.00000955, -0.00000141, 0.00000767, 0.00001403, -0.00001850, -0.00000944, 0.00001785, -0.00000256, -0.00000703, 0.00000249, 0.00000335, -0.00000429, 0.00000465, -0.00000576, 0.00000390, 0.00000256, -0.00000675, 0.00000056, -0.00000152, 0.00000694, -0.00000556, 0.00001003, -0.00001315, 0.00001067, -0.00001397, 0.00002471, -0.00002672, 0.00000859, -0.00000494, 0.00002302, -0.00002489, 0.00001182, -0.00001173, 0.00001534, -0.00002270, 0.00003361, -0.00004448, 0.00005278, -0.00005359, 0.00004223, -0.00003996, 0.00002497, -0.00002064, 0.00001394, -0.00001215, 0.00002783, -0.00004415, 0.00004533, -0.00005200, 0.00005684, -0.00006775, 0.00006960, -0.00008321, 0.00007997, -0.00009859, 0.00010997, -0.00015461, 0.00015246, -0.00024242, 0.00028917, -0.00080474, 0.00105448, 0.00013989, -0.00047714, 0.00021803, -0.00019776, 0.00012259, -0.00013006, 0.00009825, -0.00008811, 0.00006734, -0.00006967, 0.00005795, -0.00005899, 0.00005177, -0.00004731, 0.00003791, -0.00003990, 0.00002575, -0.00000947, 0.00000285, -0.00000334, 0.00000962, -0.00001631, 0.00001137, -0.00001663, 0.00003301, -0.00003791, 0.00004396, -0.00004572, 0.00003359, -0.00002765, 0.00002305, -0.00003337, 0.00003516, -0.00001612, 0.00001495, -0.00003333, 0.00003381, -0.00002421, 0.00001930, -0.00002334, 0.00002080, -0.00001436, 0.00001804, -0.00001370, 0.00000682, -0.00001350, 0.00001171, -0.00000610, 0.00000460, -0.00000598, 0.00000352, 0.00000230, -0.00000916, 0.00000367, 0.00000446, 0.00001174, -0.00002495, 0.00000068, 0.00000926, 0.00001160, },
		{ -0.00000433, 0.00000095, 0.00000106, 0.00000045, 0.00000225, -0.00000472, -0.00000668, 0.00000895, 0.00001279, -0.00001012, -0.00001850, 0.00001878, 0.00000007, -0.00000578, -0.00000125, 0.00000537, -0.00000409, 0.00000273, -0.00000232, -0.00000104, 0.00000780, -0.00001060, 0.00000425, -0.00000738, 0.00001312, -0.00001164, 0.00001670, -0.00001928, 0.00001678, -0.00002112, 0.00003055, -0.00002904, 0.00001096, -0.00001195, 0.00003094, -0.00002970, 0.00001735, -0.00002083, 0.00002715, -0.00003704, 0.00004700, -0.00005384, 0.00005662, -0.00005037, 0.00003681, -0.00003452, 0.00002150, -0.00001990, 0.00001734, -0.00002177, 0.00004129, -0.00005686, 0.00005678, -0.00006417, 0.00006960, -0.00008001, 0.00008172, -0.00009514, 0.00009350, -0.00011528, 0.00012968, -0.00017268, 0.00017342, -0.00026478, 0.00031615, -0.00072921, 0.00067636, 0.00059319, -0.00069662, 0.00030679, -0.00025820, 0.00016691, -0.00016677, 0.00012510, -0.00011070, 0.00008820, -0.00008915, 0.00007514, -0.00007409, 0.00006399, -0.00005761, 0.00004770, -0.00004786, 0.00003187, -0.00001729, 0.00001540, -0.00002013, 0.00002780, -0.00003528, 0.00003204, -0.00004082, 0.00005477, -0.00005527, 0.00005585, -0.00005105, 0.00003661, -0.00003091, 0.00002788, -0.00003803, 0.00003614, -0.00001707, 0.00001971, -0.00003746, 0.00003387, -0.00002252, 0.00001836, -0.00002192, 0.00001806, -0.00001201, 0.00001526, -0.00000981, 0.00000455, -0.00001150, 0.00000821, -0.00000235, 0.00000125, -0.00000212, -0.00000096, 0.00000593, -0.00000904, 0.00000154, 0.00000253, 0.00001483, -0.00002055, -0.00000697, 0.00001116, 0.00001093, -0.00000759, },
		{ 0.00000442, -0.00000242, 0.00000394, -0.00000143, 0.00000051, -0.00001162, 0.00001005, 0.00001045, -0.00000130, -0.00002532, 0.00001705, 0.00000311, -0.00000399, -0.00000424, 0.00000560, -0.00000217, -0.00000028, 0.00000144, -0.00000543, 0.00001162, -0.00001266, 0.00000633, -0.00001127, 0.00001651, -0.00001445, 0.00001961, -0.00002111, 0.00001827, -0.00002304, 0.00003042, -0.00002556, 0.00000832, -0.00001367, 0.00003231, -0.00002761, 0.00001628, -0.00002273, 0.00003024, -0.00004011, 0.00004614, -0.00004715, 0.00004404, -0.00003233, 0.00001917, -0.00001867, 0.00000849, -0.00000995, 0.00001152, -0.00002080, 0.00004157, -0.00005408, 0.00005132, -0.00005804, 0.00006209, -0.00007001, 0.00006966, -0.00008087, 0.00007831, -0.00009933, 0.00011089, -0.00014536, 0.00013986, -0.00021795, 0.00024603, -0.00052994, 0.00022435, 0.00100107, -0.00081042, 0.00030728, -0.00025443, 0.00016037, -0.00016098, 0.00011579, -0.00010274, 0.00008244, -0.00008460, 0.00007026, -0.00006889, 0.00005778, -0.00005150, 0.00004269, -0.00004243, 0.00002660, -0.00001591, 0.00001985, -0.00002848, 0.00003703, -0.00004497, 0.00004261, -0.00005252, 0.00006143, -0.00005630, 0.00005143, -0.00004195, 0.00002770, -0.00002394, 0.00002327, -0.00003334, 0.00002845, -0.00001064, 0.00001731, -0.00003383, 0.00002641, -0.00001433, 0.00001163, -0.00001511, 0.00001058, -0.00000550, 0.00000874, -0.00000303, -0.00000006, -0.00000712, 0.00000264, 0.00000283, -0.00000296, 0.00000188, -0.00000450, 0.00000751, -0.00000696, -0.00000109, 0.00000111, 0.00001531, -0.00001316, -0.00001534, 0.00001290, 0.00000932, -0.00000188, -0.00000955, },
		{ -0.00000499, 0.00000666, -0.00000444, 0.00000477, -0.00001493, 0.00000989, 0.00000832, 0.00000620, -0.00002854, 0.00001230, 0.00000665, -0.00000229, -0.00000571, 0.00000383, 0.00000095, -0.00000345, 0.00000441, -0.00000817, 0.00001306, -0.00001215, 0.00000598, -0.00001222, 0.00001611, -0.00001306, 0.00001788, -0.00001785, 0.00001440, -0.00001908, 0.00002390, -0.00001605, 0.00000034, -0.00000949, 0.00002650, -0.00001807, 0.00000790, -0.00001638, 0.00002336, -0.00003092, 0.00003108, -0.00002608, 0.00001836, -0.00000394, -0.00000601, 0.00000347, -0.00001055, 0.00000625, -0.00000122, -0.00001096, 0.00003027, -0.00003776, 0.00003131, -0.00003626, 0.00003737, -0.00004137, 0.00003767, -0.00004526, 0.00003978, -0.00005685, 0.00006107, -0.00008230, 0.00006468, -0.00012038, 0.00011019, -0.00026526, -0.00021663, 0.00128397, -0.00077259, 0.00020134, -0.00017925, 0.00009934, -0.00011085, 0.00006955, -0.00006392, 0.00004978, -0.00005573, 0.00004315, -0.00004345, 0.00003339, -0.00002929, 0.00002308, -0.00002381, 0.00001000, -0.00000488, 0.00001493, -0.00002648, 0.00003505, -0.00004272, 0.00003999, -0.00004875, 0.00005106, -0.00004063, 0.00003193, -0.00002078, 0.00000946, -0.00000910, 0.00001131, -0.00002128, 0.00001412, 0.00000127, 0.00000939, -0.00002410, 0.00001332, -0.00000164, 0.00000109, -0.00000485, 0.00000030, 0.00000331, 0.00000030, 0.00000490, -0.00000551, -0.00000174, -0.00000359, 0.00000801, -0.00000664, 0.00000476, -0.00000618, 0.00000682, -0.00000347, -0.00000366, 0.00000005, 0.00001355, -0.00000413, -0.00002276, 0.00001326, 0.00000781, 0.00000364, -0.00001373, 0.00000675, },
		{ 0.00000768, -0.00000577, 0.00000699, -0.00001570, 0.00000778, 0.00000734, 0.00001118, -0.00002756, 0.00000488, 0.00001037, -0.00000097, -0.00000542, 0.00000050, 0.00000433, -0.00000570, 0.00000567, -0.00000859, 0.00001175, -0.00000893, 0.00000306, -0.00001005, 0.00001189, -0.00000769, 0.00001191, -0.00001008, 0.00000598, -0.00001022, 0.00001228, -0.00000208, -0.00001148, -0.00000079, 0.00001503, -0.00000291, -0.00000594, -0.00000363, 0.00000863, -0.00001241, 0.00000628, 0.00000321, -0.00001297, 0.00002685, -0.00003130, 0.00002532, -0.00002973, 0.00002327, -0.00001585, 0.00000297, 0.00001253, -0.00001381, 0.00000338, -0.00000608, 0.00000346, -0.00000308, -0.00000422, 0.00000064, -0.00000990, -0.00000154, -0.00000369, -0.00000300, -0.00002806, -0.00000325, -0.00004659, -0.00000194, -0.00056755, 0.00138513, -0.00056755, -0.00000194, -0.00004659, -0.00000325, -0.00002806, -0.00000300, -0.00000369, -0.00000154, -0.00000990, 0.00000064, -0.00000422, -0.00000308, 0.00000346, -0.00000608, 0.00000338, -0.00001381, 0.00001253, 0.00000297, -0.00001585, 0.00002327, -0.00002973, 0.00002532, -0.00003130, 0.00002685, -0.00001297, 0.00000321, 0.00000628, -0.00001241, 0.00000863, -0.00000363, -0.00000594, -0.00000291, 0.00001503, -0.00000079, -0.00001148, -0.00000208, 0.00001228, -0.00001022, 0.00000598, -0.00001008, 0.00001191, -0.00000769, 0.00001189, -0.00001005, 0.00000306, -0.00000893, 0.00001175, -0.00000859, 0.00000567, -0.00000570, 0.00000433, 0.00000050, -0.00000542, -0.00000097, 0.00001037, 0.00000488, -0.00002756, 0.00001118, 0.00000734, 0.00000778, -0.00001570, 0.00000699, -0.00000577, },
		{ -0.00000499, 0.00000675, -0.00001373, 0.00000364, 0.00000781, 0.00001326, -0.00002276, -0.00000413, 0.00001355, 0.00000005, -0.00000366, -0.00000347, 0.00000682, -0.00000618, 0.00000476, -0.00000664, 0.00000801, -0.00000359, -0.00000174, -0.00000551, 0.00000490, 0.00000030, 0.00000331, 0.00000030, -0.00000485, 0.00000109, -0.00000164, 0.00001332, -0.00002410, 0.00000939, 0.00000127, 0.00001412, -0.00002128, 0.00001131, -0.00000910, 0.00000946, -0.00002078, 0.00003193, -0.00004063, 0.00005106, -0.00004875, 0.00003999, -0.00004272, 0.00003505, -0.00002648, 0.00001493, -0.00000488, 0.00001000, -0.00002381, 0.00002308, -0.00002929, 0.00003339, -0.00004345, 0.00004315, -0.00005573, 0.00004978, -0.00006392, 0.00006955, -0.00011085, 0.00009934, -0.00017925, 0.00020134, -0.00077259, 0.00128397, -0.00021663, -0.00026526, 0.00011019, -0.00012038, 0.00006468, -0.00008230, 0.00006107, -0.00005685, 0.00003978, -0.00004526, 0.00003767, -0.00004137, 0.00003737, -0.00003626, 0.00003131, -0.00003776, 0.00003027, -0.00001096, -0.00000122, 0.00000625, -0.00001055, 0.00000347, -0.00000601, -0.00000394, 0.00001836, -0.00002608, 0.00003108, -0.00003092, 0.00002336, -0.00001638, 0.00000790, -0.00001807, 0.00002650, -0.00000949, 0.00000034, -0.00001605, 0.00002390, -0.00001908, 0.00001440, -0.00001785, 0.00001788, -0.00001306, 0.00001611, -0.00001222, 0.00000598, -0.00001215, 0.00001306, -0.00000817, 0.00000441, -0.00000345, 0.00000095, 0.00000383, -0.00000571, -0.00000229, 0.00000665, 0.00001230, -0.00002854, 0.00000620, 0.00000832, 0.00000989, -0.00001493, 0.00000477, -0.00000444, 0.00000666, },
		{ 0.00000442, -0.00000955, -0.00000188, 0.00000932, 0.00001290, -0.00001534, -0.00001316, 0.00001531, 0.00000111, -0.00000109, -0.00000696, 0.00000751, -0.00000450, 0.00000188, -0.00000296, 0.00000283, 0.00000264, -0.00000712, -0.00000006, -0.00000303, 0.00000874, -0.00000550, 0.00001058, -0.00001511, 0.00001163, -0.00001433, 0.00002641, -0.00003383, 0.00001731, -0.00001064, 0.00002845, -0.00003334, 0.00002327, -0.00002394, 0.00002770, -0.00004195, 0.00005143, -0.00005630, 0.00006143, -0.00005252, 0.00004261, -0.00004497, 0.00003703, -0.00002848, 0.00001985, -0.00001591, 0.00002660, -0.00004243, 0.00004269, -0.00005150, 0.00005778, -0.00006889, 0.00007026, -0.00008460, 0.00008244, -0.00010274, 0.00011579, -0.00016098, 0.00016037, -0.00025443, 0.00030728, -0.00081042, 0.00100107, 0.00022435, -0.00052994, 0.00024603, -0.00021795, 0.00013986, -0.00014536, 0.00011089, -0.00009933, 0.00007831, -0.00008087, 0.00006966, -0.00007001, 0.00006209, -0.00005804, 0.00005132, -0.00005408, 0.00004157, -0.00002080, 0.00001152, -0.00000995, 0.00000849, -0.00001867, 0.00001917, -0.00003233, 0.00004404, -0.00004715, 0.00004614, -0.00004011, 0.00003024, -0.00002273, 0.00001628, -0.00002761, 0.00003231, -0.00001367, 0.00000832, -0.00002556, 0.00003042, -0.00002304, 0.00001827, -0.00002111, 0.00001961, -0.00001445, 0.00001651, -0.00001127, 0.00000633, -0.00001266, 0.00001162, -0.00000543, 0.00000144, -0.00000028, -0.00000217, 0.00000560, -0.00000424, -0.00000399, 0.00000311, 0.00001705, -0.00002532, -0.00000130, 0.00001045, 0.00001005, -0.00001162, 0.00000051, -0.00000143, 0.00000394, -0.00000242, },
		{ -0.00000433, -0.00000759, 0.00001093, 0.00001116, -0.00000697, -0.00002055, 0.00001483, 0.00000253, 0.00000154, -0.00000904, 0.00000593, -0.00000096, -0.00000212, 0.00000125, -0.00000235, 0.00000821, -0.00001150, 0.00000455, -0.00000981, 0.00001526, -0.00001201, 0.00001806, -0.00002192, 0.00001836, -0.00002252, 0.00003387, -0.00003746, 0.00001971, -0.00001707, 0.00003614, -0.00003803, 0.00002788, -0.00003091, 0.00003661, -0.00005105, 0.00005585, -0.00005527, 0.00005477, -0.00004082, 0.00003204, -0.00003528, 0.00002780, -0.00002013, 0.00001540, -0.00001729, 0.00003187, -0.00004786, 0.00004770, -0.00005761, 0.00006399, -0.00007409, 0.00007514, -0.00008915, 0.00008820, -0.00011070, 0.00012510, -0.00016677, 0.00016691, -0.00025820, 0.00030679, -0.00069662, 0.00059319, 0.00067636, -0.00072921, 0.00031615, -0.00026478, 0.00017342, -0.00017268, 0.00012968, -0.00011528, 0.00009350, -0.00009514, 0.00008172, -0.00008001, 0.00006960, -0.00006417, 0.00005678, -0.00005686, 0.00004129, -0.00002177, 0.00001734, -0.00001990, 0.00002150, -0.00003452, 0.00003681, -0.00005037, 0.00005662, -0.00005384, 0.00004700, -0.00003704, 0.00002715, -0.00002083, 0.00001735, -0.00002970, 0.00003094, -0.00001195, 0.00001096, -0.00002904, 0.00003055, -0.00002112, 0.00001678, -0.00001928, 0.00001670, -0.00001164, 0.00001312, -0.00000738, 0.00000425, -0.00001060, 0.00000780, -0.00000104, -0.00000232, 0.00000273, -0.00000409, 0.00000537, -0.00000125, -0.00000578, 0.00000007, 0.00001878, -0.00001850, -0.00001012, 0.00001279, 0.00000895, -0.00000668, -0.00000472, 0.00000225, 0.00000045, 0.00000106, 0.00000095, },
		{ -0.00001220, 0.00001160, 0.00000926, 0.00000068, -0.00002495, 0.00001174, 0.00000446, 0.00000367, -0.00000916, 0.00000230, 0.00000352, -0.00000598, 0.00000460, -0.00000610, 0.00001171, -0.00001350, 0.00000682, -0.00001370, 0.00001804, -0.00001436, 0.00002080, -0.00002334, 0.00001930, -0.00002421, 0.00003381, -0.00003333, 0.00001495, -0.00001612, 0.00003516, -0.00003337, 0.00002305, -0.00002765, 0.00003359, -0.00004572, 0.00004396, -0.00003791, 0.00003301, -0.00001663, 0.00001137, -0.00001631, 0.00000962, -0.00000334, 0.00000285, -0.00000947, 0.00002575, -0.00003990, 0.00003791, -0.00004731, 0.00005177, -0.00005899, 0.00005795, -0.00006967, 0.00006734, -0.00008811, 0.00009825, -0.00013006, 0.00012259, -0.00019776, 0.00021803, -0.00047714, 0.00013989, 0.00105448, -0.00080474, 0.00028917, -0.00024242, 0.00015246, -0.00015461, 0.00010997, -0.00009859, 0.00007997, -0.00008321, 0.00006960, -0.00006775, 0.00005684, -0.00005200, 0.00004533, -0.00004415, 0.00002783, -0.00001215, 0.00001394, -0.00002064, 0.00002497, -0.00003996, 0.00004223, -0.00005359, 0.00005278, -0.00004448, 0.00003361, -0.00002270, 0.00001534, -0.00001173, 0.00001182, -0.00002489, 0.00002302, -0.00000494, 0.00000859, -0.00002672, 0.00002471, -0.00001397, 0.00001067, -0.00001315, 0.00001003, -0.00000556, 0.00000694, -0.00000152, 0.00000056, -0.00000675, 0.00000256, 0.00000390, -0.00000576, 0.00000465, -0.00000429, 0.00000335, 0.00000249, -0.00000703, -0.00000256, 0.00001785, -0.00000944, -0.00001850, 0.00001403, 0.00000767, -0.00000141, -0.00000955, 0.00000536, -0.00000264, 0.00000427, -0.00000242, 0.00000056, },
		{ 0.00001053, 0.00000815, 0.00000639, -0.00002572, 0.00000627, 0.00000667, 0.00000505, -0.00000735, -0.00000259, 0.00000767, -0.00000837, 0.00000594, -0.00000743, 0.00001228, -0.00001240, 0.00000600, -0.00001386, 0.00001630, -0.00001192, 0.00001826, -0.00001895, 0.00001419, -0.00001927, 0.00002637, -0.00002193, 0.00000362, -0.00000830, 0.00002609, -0.00002020, 0.00000975, -0.00001517, 0.00001994, -0.00002812, 0.00001944, -0.00000953, 0.00000265, 0.00001314, -0.00001300, 0.00000627, -0.00001230, 0.00001705, -0.00001349, 0.00000378, 0.00001194, -0.00002256, 0.00001765, -0.00002528, 0.00002630, -0.00002948, 0.00002534, -0.00003350, 0.00002796, -0.00004413, 0.00004623, -0.00006460, 0.00004497, -0.00009703, 0.00007851, -0.00021492, -0.00027488, 0.00128718, -0.00072204, 0.00015745, -0.00015194, 0.00007992, -0.00009480, 0.00005572, -0.00005305, 0.00004107, -0.00004817, 0.00003635, -0.00003634, 0.00002695, -0.00002450, 0.00001983, -0.00001881, 0.00000394, 0.00000590, 0.00000257, -0.00001269, 0.00001889, -0.00003451, 0.00003486, -0.00004214, 0.00003421, -0.00002235, 0.00001037, -0.00000178, -0.00000101, 0.00000110, 0.00000254, -0.00001572, 0.00001105, 0.00000508, 0.00000307, -0.00002034, 0.00001479, -0.00000355, 0.00000186, -0.00000461, 0.00000143, 0.00000204, -0.00000033, 0.00000479, -0.00000353, -0.00000226, -0.00000290, 0.00000824, -0.00000796, 0.00000496, -0.00000286, 0.00000028, 0.00000587, -0.00000702, -0.00000505, 0.00001505, 0.00000010, -0.00002464, 0.00001302, 0.00000718, 0.00000300, -0.00001273, 0.00000680, -0.00000433, 0.00000623, -0.00000460, 0.00000395, -0.00001468, },
		{ 0.00000824, 0.00000968, -0.00002305, -0.00000072, 0.00000859, 0.00000579, -0.00000411, -0.00000758, 0.00001020, -0.00000835, 0.00000472, -0.00000609, 0.00000993, -0.00000840, 0.00000233, -0.00001055, 0.00001059, -0.00000549, 0.00001145, -0.00001002, 0.00000456, -0.00000947, 0.00001373, -0.00000585, -0.00001154, 0.00000364, 0.00001194, -0.00000198, -0.00000827, 0.00000246, 0.00000034, -0.00000406, -0.00001033, 0.00002127, -0.00002730, 0.00003983, -0.00003336, 0.00002561, -0.00003151, 0.00003481, -0.00002778, 0.00001682, -0.00000362, -0.00000240, -0.00000591, 0.00000068, -0.00000377, 0.00000478, -0.00001205, 0.00000767, -0.00001704, 0.00000665, -0.00001383, 0.00000904, -0.00004080, 0.00001183, -0.00006692, 0.00002645, -0.00058140, 0.00133240, -0.00048053, -0.00005846, -0.00001522, -0.00002535, -0.00000970, -0.00001846, 0.00000851, -0.00001192, -0.00000025, -0.00000850, 0.00000522, -0.00001161, 0.00001046, -0.00001240, 0.00001221, -0.00002398, 0.00002693, -0.00001238, 0.00000027, 0.00000652, -0.00002125, 0.00001810, -0.00002052, 0.00000688, 0.00000549, -0.00001536, 0.00001898, -0.00001626, 0.00001303, -0.00000662, -0.00000569, -0.00000158, 0.00001506, -0.00000296, -0.00001242, 0.00000344, 0.00000749, -0.00000717, 0.00000401, -0.00000693, 0.00000916, -0.00000687, 0.00001001, -0.00000679, 0.00000173, -0.00000745, 0.00001103, -0.00000838, 0.00000366, -0.00000041, -0.00000275, 0.00000781, -0.00000523, -0.00000763, 0.00001131, 0.00000843, -0.00002718, 0.00000913, 0.00000801, 0.00000577, -0.00001356, 0.00000605, -0.00000416, 0.00000642, -0.00000502, 0.00000525, -0.00001461, 0.00000752, },
		{ 0.00001077, -0.00001784, -0.00000796, 0.00000954, 0.00000620, -0.00000017, -0.00001150, 0.00001023, -0.00000566, 0.00000122, -0.00000268, 0.00000550, -0.00000255, -0.00000304, -0.00000510, 0.00000256, 0.00000293, 0.00000261, 0.00000090, -0.00000675, 0.00000204, -0.00000056, 0.00001098, -0.00002647, 0.00001557, -0.00000280, 0.00001632, -0.00002567, 0.00001950, -0.00001878, 0.00001891, -0.00003667, 0.00004538, -0.00004806, 0.00005568, -0.00004325, 0.00003602, -0.00004247, 0.00004449, -0.00003472, 0.00002429, -0.00001497, 0.00001382, -0.00002528, 0.00002234, -0.00002929, 0.00003363, -0.00004308, 0.00004169, -0.00005425, 0.00004906, -0.00006426, 0.00007002, -0.00010995, 0.00009835, -0.00017865, 0.00019765, -0.00073848, 0.00118654, -0.00011539, -0.00031297, 0.00012991, -0.00013352, 0.00007594, -0.00009149, 0.00006795, -0.00006315, 0.00004628, -0.00005179, 0.00004472, -0.00004759, 0.00004258, -0.00004183, 0.00004001, -0.00004784, 0.00004396, -0.00002492, 0.00001276, -0.00000682, -0.00000558, -0.00000206, 0.00000411, -0.00002094, 0.00003048, -0.00003565, 0.00003294, -0.00002526, 0.00002000, -0.00001222, 0.00000196, -0.00001171, 0.00002215, -0.00000705, -0.00000539, -0.00000678, 0.00001663, -0.00001411, 0.00001061, -0.00001311, 0.00001409, -0.00001117, 0.00001296, -0.00000830, 0.00000438, -0.00001028, 0.00001176, -0.00000697, 0.00000120, 0.00000220, -0.00000471, 0.00000761, -0.00000162, -0.00001023, 0.00000732, 0.00001439, -0.00002560, 0.00000262, 0.00000996, 0.00000683, -0.00001204, 0.00000328, -0.00000232, 0.00000498, -0.00000374, 0.00000454, -0.00001228, 0.00000299, 0.00000930, },
		{ -0.00001143, -0.00001410, 0.00000892, 0.00000654, 0.00000371, -0.00001350, 0.00000752, -0.00000087, -0.00000353, 0.00000152, 0.00000047, 0.00000351, -0.00000840, 0.00000059, -0.00000550, 0.00001079, -0.00000552, 0.00001077, -0.00001646, 0.00001174, -0.00001264, 0.00002441, -0.00003699, 0.00002331, -0.00001361, 0.00002979, -0.00003729, 0.00003045, -0.00003134, 0.00003398, -0.00005229, 0.00005587, -0.00005384, 0.00005631, -0.00003951, 0.00003477, -0.00004230, 0.00004311, -0.00003136, 0.00002292, -0.00001811, 0.00002132, -0.00003506, 0.00003367, -0.00004345, 0.00004949, -0.00005952, 0.00005961, -0.00007380, 0.00007189, -0.00009203, 0.00010332, -0.00014514, 0.00014208, -0.00023202, 0.00027348, -0.00073991, 0.00088424, 0.00030952, -0.00054562, 0.00023964, -0.00021241, 0.00013646, -0.00014209, 0.00010737, -0.00009713, 0.00007752, -0.00008083, 0.00007055, -0.00007043, 0.00006230, -0.00005968, 0.00005648, -0.00006038, 0.00005080, -0.00002963, 0.00001956, -0.00001582, 0.00000685, -0.00001919, 0.00002435, -0.00004134, 0.00004526, -0.00004446, 0.00003579, -0.00002512, 0.00001988, -0.00001243, 0.00000534, -0.00001737, 0.00002464, -0.00000769, -0.00000086, -0.00001406, 0.00002214, -0.00001744, 0.00001382, -0.00001595, 0.00001581, -0.00001242, 0.00001308, -0.00000770, 0.00000530, -0.00001107, 0.00001042, -0.00000411, -0.00000169, 0.00000410, -0.00000493, 0.00000519, 0.00000326, -0.00001238, 0.00000338, 0.00001755, -0.00002035, -0.00000541, 0.00001218, 0.00000672, -0.00000885, -0.00000068, 0.00000043, 0.00000258, -0.00000139, 0.00000248, -0.00000854, -0.00000214, 0.00001062, 0.00001045, },
		{ -0.00001810, 0.00000653, 0.00000691, 0.00000700, -0.00001320, 0.00000258, 0.00000481, -0.00000799, 0.00000493, -0.00000355, 0.00000814, -0.00001205, 0.00000461, -0.00001143, 0.00001577, -0.00001052, 0.00001703, -0.00002185, 0.00001681, -0.00001948, 0.00003135, -0.00004015, 0.00002395, -0.00001738, 0.00003508, -0.00003974, 0.00003184, -0.00003360, 0.00003717, -0.00005345, 0.00005014, -0.00004352, 0.00004198, -0.00002328, 0.00002297, -0.00003169, 0.00003116, -0.00001809, 0.00001272, -0.00001230, 0.00001874, -0.00003349, 0.00003252, -0.00004365, 0.00004943, -0.00005824, 0.00005812, -0.00007199, 0.00007085, -0.00009213, 0.00010336, -0.00014056, 0.00013706, -0.00022169, 0.00025547, -0.00061230, 0.00048903, 0.00071589, -0.00069583, 0.00027653, -0.00023722, 0.00015317, -0.00015541, 0.00011465, -0.00010359, 0.00008433, -0.00008732, 0.00007527, -0.00007356, 0.00006381, -0.00006071, 0.00005687, -0.00005761, 0.00004421, -0.00002361, 0.00001759, -0.00001703, 0.00001206, -0.00002858, 0.00003484, -0.00004919, 0.00004589, -0.00003958, 0.00002689, -0.00001614, 0.00001317, -0.00000758, 0.00000442, -0.00001828, 0.00002232, -0.00000474, 0.00000073, -0.00001771, 0.00002340, -0.00001673, 0.00001332, -0.00001519, 0.00001423, -0.00001062, 0.00001055, -0.00000522, 0.00000458, -0.00000994, 0.00000740, -0.00000047, -0.00000427, 0.00000475, -0.00000335, 0.00000120, 0.00000828, -0.00001333, -0.00000056, 0.00001815, -0.00001270, -0.00001333, 0.00001347, 0.00000637, -0.00000504, -0.00000470, 0.00000304, 0.00000018, 0.00000111, 0.00000004, -0.00000451, -0.00000677, 0.00001136, 0.00000966, -0.00000515, },
		{ 0.00000262, 0.00000712, 0.00000950, -0.00001083, -0.00000349, 0.00000979, -0.00001064, 0.00000623, -0.00000534, 0.00001013, -0.00001281, 0.00000573, -0.00001381, 0.00001648, -0.00001105, 0.00001829, -0.00002160, 0.00001594, -0.00001979, 0.00003062, -0.00003508, 0.00001683, -0.00001337, 0.00003145, -0.00003245, 0.00002327, -0.00002528, 0.00002842, -0.00004093, 0.00003046, -0.00002088, 0.00001746, 0.00000039, 0.00000509, -0.00001447, 0.00001223, 0.00000157, -0.00000322, -0.00000006, 0.00000829, -0.00002268, 0.00002090, -0.00003185, 0.00003555, -0.00004168, 0.00003994, -0.00005179, 0.00004910, -0.00006809, 0.00007462, -0.00010225, 0.00009163, -0.00016028, 0.00016754, -0.00040520, 0.00007785, 0.00102873, -0.00071828, 0.00022080, -0.00019847, 0.00012026, -0.00012766, 0.00008733, -0.00008071, 0.00006510, -0.00006991, 0.00005789, -0.00005641, 0.00004687, -0.00004484, 0.00004132, -0.00004003, 0.00002508, -0.00000763, 0.00000706, -0.00001007, 0.00000907, -0.00002863, 0.00003380, -0.00004345, 0.00003291, -0.00002311, 0.00000943, -0.00000169, 0.00000276, 0.00000012, 0.00000075, -0.00001563, 0.00001642, 0.00000072, 0.00000004, -0.00001814, 0.00002087, -0.00001255, 0.00000972, -0.00001150, 0.00001000, -0.00000648, 0.00000612, -0.00000150, 0.00000269, -0.00000737, 0.00000336, 0.00000321, -0.00000599, 0.00000403, -0.00000051, -0.00000320, 0.00001210, -0.00001228, -0.00000471, 0.00001689, -0.00000429, -0.00001939, 0.00001274, 0.00000662, -0.00000167, -0.00000769, 0.00000456, -0.00000138, 0.00000293, -0.00000187, -0.00000119, -0.00001001, 0.00001088, 0.00000914, 0.00000000, -0.00001946, },
		{ 0.00000679, 0.00001125, -0.00000698, -0.00000936, 0.00001264, -0.00001044, 0.00000478, -0.00000446, 0.00000916, -0.00001048, 0.00000371, -0.00001242, 0.00001289, -0.00000726, 0.00001480, -0.00001612, 0.00000974, -0.00001428, 0.00002323, -0.00002322, 0.00000366, -0.00000337, 0.00002085, -0.00001779, 0.00000746, -0.00000944, 0.00001139, -0.00001948, 0.00000307, 0.00000678, -0.00000960, 0.00002429, -0.00001261, 0.00000379, -0.00000822, 0.00002220, -0.00001980, 0.00001389, -0.00000536, -0.00000748, 0.00000390, -0.00001343, 0.00001379, -0.00001646, 0.00001238, -0.00002118, 0.00001545, -0.00002992, 0.00002906, -0.00004490, 0.00002420, -0.00007227, 0.00004646, -0.00017695, -0.00027659, 0.00119307, -0.00059482, 0.00007701, -0.00010563, 0.00004685, -0.00006742, 0.00003341, -0.00003568, 0.00002622, -0.00003449, 0.00002408, -0.00002450, 0.00001681, -0.00001712, 0.00001469, -0.00001259, -0.00000175, 0.00001404, -0.00000859, 0.00000232, 0.00000005, -0.00002108, 0.00002320, -0.00002720, 0.00001092, -0.00000072, -0.00001074, 0.00001302, -0.00000723, 0.00000750, -0.00000325, -0.00001146, 0.00000892, 0.00000692, -0.00000161, -0.00001642, 0.00001573, -0.00000620, 0.00000428, -0.00000608, 0.00000431, -0.00000113, 0.00000088, 0.00000255, 0.00000028, -0.00000401, -0.00000095, 0.00000631, -0.00000659, 0.00000227, 0.00000267, -0.00000667, 0.00001354, -0.00000879, -0.00000916, 0.00001455, 0.00000331, -0.00002230, 0.00000943, 0.00000790, 0.00000062, -0.00000901, 0.00000452, -0.00000166, 0.00000363, -0.00000273, 0.00000084, -0.00001141, 0.00000894, 0.00000925, 0.00000361, -0.00001835, -0.00000207, },
		{ 0.00001239, -0.00000241, -0.00001380, 0.00001249, -0.00000718, 0.00000085, -0.00000139, 0.00000585, -0.00000580, -0.00000064, -0.00000820, 0.00000624, -0.00000063, 0.00000821, -0.00000731, 0.00000038, -0.00000536, 0.00001194, -0.00000785, -0.00001199, 0.00000898, 0.00000721, -0.00000016, -0.00001075, 0.00000863, -0.00000797, 0.00000386, -0.00002392, 0.00003102, -0.00003120, 0.00004136, -0.00002422, 0.00001773, -0.00002469, 0.00003825, -0.00003169, 0.00002442, -0.00001686, 0.00000633, -0.00001230, 0.00000489, -0.00000838, 0.00000907, -0.00001539, 0.00000979, -0.00001895, 0.00000962, -0.00001828, 0.00001349, -0.00004336, 0.00001460, -0.00006957, 0.00001992, -0.00052112, 0.00118546, -0.00033966, -0.00012564, 0.00001491, -0.00004550, 0.00000696, -0.00003116, 0.00001761, -0.00002007, 0.00000778, -0.00001574, 0.00001236, -0.00001721, 0.00001394, -0.00001495, 0.00001693, -0.00002896, 0.00003498, -0.00002390, 0.00001534, -0.00001067, -0.00001012, 0.00000772, -0.00000634, -0.00001309, 0.00002036, -0.00002705, 0.00002279, -0.00001297, 0.00001173, -0.00000540, -0.00000777, 0.00000186, 0.00001206, -0.00000284, -0.00001383, 0.00000942, 0.00000082, -0.00000156, -0.00000030, -0.00000154, 0.00000421, -0.00000406, 0.00000608, -0.00000199, -0.00000053, -0.00000484, 0.00000837, -0.00000613, 0.00000006, 0.00000517, -0.00000813, 0.00001200, -0.00000302, -0.00001364, 0.00001166, 0.00000902, -0.00002156, 0.00000376, 0.00000999, 0.00000178, -0.00000864, 0.00000299, -0.00000077, 0.00000324, -0.00000251, 0.00000152, -0.00001108, 0.00000583, 0.00000985, 0.00000578, -0.00001547, -0.00000659, 0.00000551, },
		{ 0.00000222, -0.00001603, 0.00000924, -0.00000157, -0.00000443, 0.00000260, 0.00000155, -0.00000026, -0.00000573, -0.00000291, -0.00000138, 0.00000653, 0.00000099, 0.00000208, -0.00000917, 0.00000376, 0.00000037, 0.00000704, -0.00002593, 0.00001942, -0.00000493, 0.00001546, -0.00002601, 0.00002324, -0.00002343, 0.00002216, -0.00004312, 0.00004489, -0.00004142, 0.00004696, -0.00002615, 0.00002391, -0.00003342, 0.00004577, -0.00003513, 0.00002774, -0.00002203, 0.00001411, -0.00002259, 0.00001740, -0.00002445, 0.00002762, -0.00003536, 0.00003230, -0.00004422, 0.00003922, -0.00005408, 0.00005718, -0.00009241, 0.00007743, -0.00015099, 0.00014988, -0.00063144, 0.00101755, 0.00000359, -0.00033994, 0.00012696, -0.00012932, 0.00007306, -0.00008746, 0.00006302, -0.00005965, 0.00004413, -0.00004977, 0.00004326, -0.00004518, 0.00003914, -0.00003898, 0.00004035, -0.00004907, 0.00004872, -0.00003329, 0.00002391, -0.00001826, -0.00000065, -0.00000702, 0.00001255, -0.00003207, 0.00003366, -0.00003443, 0.00002418, -0.00001233, 0.00001137, -0.00000455, -0.00000577, -0.00000340, 0.00001497, -0.00000275, -0.00001132, 0.00000314, 0.00000719, -0.00000655, 0.00000464, -0.00000644, 0.00000850, -0.00000780, 0.00000842, -0.00000361, 0.00000250, -0.00000777, 0.00000918, -0.00000483, -0.00000197, 0.00000625, -0.00000712, 0.00000769, 0.00000415, -0.00001741, 0.00000835, 0.00001249, -0.00001761, -0.00000321, 0.00001209, 0.00000228, -0.00000711, 0.00000060, 0.00000072, 0.00000223, -0.00000160, 0.00000123, -0.00000956, 0.00000219, 0.00001050, 0.00000700, -0.00001174, -0.00001005, 0.00000308, 0.00001300, },
		{ -0.00001586, 0.00000362, 0.00000494, -0.00000940, 0.00000590, -0.00000211, 0.00000445, -0.00000980, 0.00000149, -0.00000774, 0.00001185, -0.00000441, 0.00000940, -0.00001606, 0.00001011, -0.00000826, 0.00001792, -0.00003460, 0.00002446, -0.00001191, 0.00002511, -0.00003420, 0.00003013, -0.00003051, 0.00003066, -0.00005004, 0.00004500, -0.00003821, 0.00004020, -0.00001817, 0.00002184, -0.00003332, 0.00004350, -0.00002899, 0.00002263, -0.00001921, 0.00001374, -0.00002445, 0.00002106, -0.00003078, 0.00003501, -0.00004293, 0.00004127, -0.00005460, 0.00005211, -0.00007032, 0.00007694, -0.00011236, 0.00010382, -0.00018304, 0.00019988, -0.00061365, 0.00073132, 0.00037208, -0.00051234, 0.00019487, -0.00017957, 0.00011131, -0.00011951, 0.00008721, -0.00008099, 0.00006417, -0.00006852, 0.00005960, -0.00005933, 0.00005144, -0.00005086, 0.00005157, -0.00005670, 0.00005082, -0.00003299, 0.00002448, -0.00001920, 0.00000351, -0.00001651, 0.00002429, -0.00004105, 0.00003546, -0.00003086, 0.00001673, -0.00000567, 0.00000693, -0.00000106, -0.00000550, -0.00000648, 0.00001534, -0.00000117, -0.00000930, -0.00000236, 0.00001207, -0.00000989, 0.00000798, -0.00000965, 0.00001109, -0.00000979, 0.00000920, -0.00000428, 0.00000467, -0.00000939, 0.00000870, -0.00000303, -0.00000336, 0.00000565, -0.00000393, 0.00000166, 0.00001124, -0.00001946, 0.00000444, 0.00001398, -0.00001161, -0.00000997, 0.00001311, 0.00000289, -0.00000522, -0.00000179, 0.00000206, 0.00000125, -0.00000058, 0.00000057, -0.00000756, -0.00000126, 0.00001066, 0.00000784, -0.00000803, -0.00001188, -0.00000035, 0.00001297, 0.00000648, },
		{ -0.00000309, 0.00001060, -0.00001239, 0.00000711, -0.00000383, 0.00000698, -0.00001149, 0.00000352, -0.00001121, 0.00001370, -0.00000638, 0.00001291, -0.00001853, 0.00001189, -0.00001200, 0.00002275, -0.00003612, 0.00002242, -0.00001203, 0.00002700, -0.00003362, 0.00002775, -0.00002777, 0.00002814, -0.00004421, 0.00003226, -0.00002378, 0.00002408, -0.00000334, 0.00001384, -0.00002601, 0.00003299, -0.00001498, 0.00001068, -0.00000957, 0.00000610, -0.00001850, 0.00001616, -0.00002731, 0.00003103, -0.00003788, 0.00003638, -0.00004954, 0.00004743, -0.00006586, 0.00007165, -0.00010245, 0.00009352, -0.00016701, 0.00017836, -0.00049862, 0.00038749, 0.00069836, -0.00059653, 0.00019399, -0.00018149, 0.00011100, -0.00011913, 0.00008373, -0.00007868, 0.00006304, -0.00006765, 0.00005757, -0.00005645, 0.00004808, -0.00004807, 0.00004837, -0.00005023, 0.00004034, -0.00002235, 0.00001634, -0.00001254, 0.00000098, -0.00001872, 0.00002661, -0.00003853, 0.00002584, -0.00001797, 0.00000301, 0.00000435, 0.00000058, 0.00000352, -0.00000604, -0.00000788, 0.00001371, 0.00000141, -0.00000766, -0.00000680, 0.00001509, -0.00001127, 0.00000943, -0.00001091, 0.00001174, -0.00000985, 0.00000835, -0.00000390, 0.00000575, -0.00000953, 0.00000705, -0.00000098, -0.00000397, 0.00000365, 0.00000050, -0.00000453, 0.00001658, -0.00001882, -0.00000034, 0.00001412, -0.00000506, -0.00001493, 0.00001210, 0.00000424, -0.00000374, -0.00000341, 0.00000261, 0.00000085, 0.00000007, 0.00000007, -0.00000573, -0.00000393, 0.00000999, 0.00000869, -0.00000489, -0.00001200, -0.00000423, 0.00001199, 0.00001025, -0.00001361, },
		{ 0.00001385, -0.00001229, 0.00000550, -0.00000303, 0.00000679, -0.00001030, 0.00000266, -0.00001122, 0.00001168, -0.00000460, 0.00001231, -0.00001634, 0.00000895, -0.00001068, 0.00002146, -0.00003078, 0.00001392, -0.00000596, 0.00002190, -0.00002532, 0.00001752, -0.00001696, 0.00001688, -0.00002896, 0.00001133, -0.00000359, 0.00000421, 0.00001321, 0.00000401, -0.00001495, 0.00001778, 0.00000310, -0.00000441, 0.00000348, -0.00000556, -0.00000792, 0.00000575, -0.00001706, 0.00001890, -0.00002376, 0.00002146, -0.00003312, 0.00002958, -0.00004568, 0.00004738, -0.00007035, 0.00005647, -0.00011680, 0.00010900, -0.00033124, 0.00005079, 0.00092562, -0.00056525, 0.00011821, -0.00013519, 0.00007348, -0.00008824, 0.00005480, -0.00005484, 0.00004263, -0.00004898, 0.00003912, -0.00003861, 0.00003116, -0.00003265, 0.00003284, -0.00003211, 0.00002006, -0.00000408, 0.00000177, -0.00000018, -0.00000686, -0.00001453, 0.00002053, -0.00002663, 0.00000840, -0.00000033, -0.00001228, 0.00001375, -0.00000473, 0.00000710, -0.00000607, -0.00000857, 0.00001102, 0.00000416, -0.00000600, -0.00001022, 0.00001626, -0.00001076, 0.00000906, -0.00001028, 0.00001053, -0.00000815, 0.00000609, -0.00000257, 0.00000567, -0.00000822, 0.00000444, 0.00000109, -0.00000393, 0.00000097, 0.00000484, -0.00000919, 0.00001877, -0.00001496, -0.00000604, 0.00001355, 0.00000067, -0.00001701, 0.00000867, 0.00000655, -0.00000303, -0.00000389, 0.00000212, 0.00000121, 0.00000014, -0.00000003, -0.00000437, -0.00000560, 0.00000845, 0.00000963, -0.00000245, -0.00001082, -0.00000777, 0.00000972, 0.00001352, -0.00000996, -0.00000941, },
		{ -0.00000888, 0.00000131, -0.00000007, 0.00000431, -0.00000672, -0.00000060, -0.00000837, 0.00000662, -0.00000008, 0.00000864, -0.00001070, 0.00000268, -0.00000579, 0.00001576, -0.00002069, 0.00000145, 0.00000364, 0.00001262, -0.00001253, 0.00000310, -0.00000221, 0.00000158, -0.00000995, -0.00001119, 0.00001558, -0.00001311, 0.00002617, -0.00000345, -0.00000398, 0.00000209, 0.00002070, -0.00001819, 0.00001568, -0.00001694, 0.00000293, -0.00000573, -0.00000469, 0.00000376, -0.00000626, 0.00000271, -0.00001208, 0.00000601, -0.00001834, 0.00001440, -0.00002859, 0.00000816, -0.00005269, 0.00002181, -0.00015791, -0.00022434, 0.00101961, -0.00041707, -0.00001771, -0.00005580, 0.00001148, -0.00003796, 0.00001029, -0.00001811, 0.00001061, -0.00001958, 0.00001097, -0.00001223, 0.00000672, -0.00001024, 0.00001047, -0.00000794, -0.00000444, 0.00001671, -0.00001471, 0.00001387, -0.00001653, -0.00000711, 0.00000960, -0.00001004, -0.00001112, 0.00001611, -0.00002395, 0.00001869, -0.00000643, 0.00000802, -0.00000454, -0.00000937, 0.00000821, 0.00000631, -0.00000400, -0.00001274, 0.00001577, -0.00000867, 0.00000720, -0.00000812, 0.00000784, -0.00000506, 0.00000281, -0.00000053, 0.00000454, -0.00000567, 0.00000119, 0.00000307, -0.00000354, -0.00000153, 0.00000784, -0.00001110, 0.00001716, -0.00000810, -0.00001223, 0.00001259, 0.00000474, -0.00001588, 0.00000319, 0.00000943, -0.00000294, -0.00000340, 0.00000082, 0.00000209, -0.00000018, 0.00000017, -0.00000347, -0.00000638, 0.00000636, 0.00001045, -0.00000048, -0.00000902, -0.00001017, 0.00000598, 0.00001629, -0.00000563, -0.00001411, 0.00001380, },
		{ -0.00000432, 0.00000385, 0.00000072, -0.00000203, -0.00000489, -0.00000416, 0.00000030, 0.00000529, 0.00000386, -0.00000375, -0.00000460, 0.00000023, 0.00000834, -0.00000903, -0.00001149, 0.00001322, 0.00000294, 0.00000058, -0.00001095, 0.00001158, -0.00001236, 0.00000677, -0.00002893, 0.00002791, -0.00002304, 0.00003181, -0.00000572, 0.00000404, -0.00001059, 0.00003392, -0.00002684, 0.00002335, -0.00002426, 0.00001011, -0.00001417, 0.00000530, -0.00000932, 0.00000893, -0.00001363, 0.00000671, -0.00001550, 0.00000712, -0.00001650, 0.00000994, -0.00003581, 0.00000550, -0.00005591, -0.00001580, -0.00040385, 0.00097451, -0.00017661, -0.00018157, 0.00003100, -0.00005520, 0.00001527, -0.00003578, 0.00001943, -0.00002240, 0.00001084, -0.00001784, 0.00001425, -0.00001743, 0.00001191, -0.00001179, 0.00001543, -0.00002662, 0.00003413, -0.00002792, 0.00002494, -0.00002380, -0.00000049, -0.00000159, 0.00000571, -0.00002680, 0.00002608, -0.00002812, 0.00001690, -0.00000344, 0.00000578, -0.00000124, -0.00001057, 0.00000577, 0.00000746, -0.00000158, -0.00001434, 0.00001384, -0.00000538, 0.00000427, -0.00000487, 0.00000416, -0.00000112, -0.00000097, 0.00000185, 0.00000263, -0.00000226, -0.00000228, 0.00000480, -0.00000313, -0.00000317, 0.00000872, -0.00000979, 0.00001207, 0.00000069, -0.00001797, 0.00001113, 0.00000702, -0.00001210, -0.00000322, 0.00001197, -0.00000289, -0.00000248, -0.00000068, 0.00000297, -0.00000046, 0.00000034, -0.00000272, -0.00000668, 0.00000418, 0.00001078, 0.00000138, -0.00000730, -0.00001086, 0.00000094, 0.00001836, -0.00000114, -0.00001649, 0.00001046, -0.00000294, },
		{ 0.00000717, -0.00000245, 0.00000218, -0.00000858, -0.00000038, -0.00000529, 0.00000945, 0.00000004, 0.00000225, -0.00001048, 0.00000490, 0.00000194, 0.00000112, -0.00002164, 0.00001950, -0.00000379, 0.00001037, -0.00002072, 0.00002035, -0.00002071, 0.00001676, -0.00003773, 0.00003025, -0.00002362, 0.00002915, -0.00000228, 0.00000808, -0.00001839, 0.00004057, -0.00002833, 0.00002454, -0.00002538, 0.00001134, -0.00001708, 0.00001003, -0.00001687, 0.00001785, -0.00002321, 0.00001835, -0.00002930, 0.00002401, -0.00003731, 0.00003575, -0.00006417, 0.00004388, -0.00010597, 0.00007384, -0.00047827, 0.00081167, 0.00011190, -0.00033215, 0.00009673, -0.00010608, 0.00005509, -0.00006986, 0.00004635, -0.00004642, 0.00003321, -0.00003894, 0.00003316, -0.00003430, 0.00002738, -0.00002778, 0.00003201, -0.00004094, 0.00004341, -0.00003374, 0.00002934, -0.00002532, 0.00000196, -0.00000912, 0.00001612, -0.00003437, 0.00002661, -0.00002351, 0.00000852, 0.00000341, 0.00000132, 0.00000303, -0.00001175, 0.00000367, 0.00000773, 0.00000096, -0.00001488, 0.00001064, -0.00000127, 0.00000073, -0.00000107, 0.00000005, 0.00000305, -0.00000467, 0.00000415, 0.00000037, 0.00000145, -0.00000546, 0.00000609, -0.00000284, -0.00000361, 0.00000734, -0.00000573, 0.00000475, 0.00000970, -0.00002199, 0.00000874, 0.00000793, -0.00000689, -0.00000906, 0.00001309, -0.00000213, -0.00000188, -0.00000169, 0.00000325, -0.00000023, 0.00000011, -0.00000182, -0.00000688, 0.00000235, 0.00001037, 0.00000343, -0.00000613, -0.00000979, -0.00000481, 0.00001932, 0.00000328, -0.00001649, 0.00000465, 0.00000398, -0.00000970, },
		{ -0.00000389, 0.00000458, -0.00001030, 0.00000151, -0.00000857, 0.00001087, -0.00000135, 0.00000572, -0.00001333, 0.00000656, -0.00000163, 0.00000766, -0.00002688, 0.00002056, -0.00000564, 0.00001479, -0.00002418, 0.00002215, -0.00002163, 0.00001836, -0.00003664, 0.00002293, -0.00001615, 0.00002005, 0.00000509, 0.00000896, -0.00002123, 0.00004041, -0.00002272, 0.00001942, -0.00002035, 0.00000655, -0.00001423, 0.00000890, -0.00001790, 0.00001923, -0.00002455, 0.00002109, -0.00003319, 0.00002952, -0.00004456, 0.00004484, -0.00007243, 0.00005712, -0.00012208, 0.00010714, -0.00046040, 0.00057213, 0.00039545, -0.00042994, 0.00011872, -0.00012654, 0.00007041, -0.00008311, 0.00005542, -0.00005519, 0.00004184, -0.00004717, 0.00003990, -0.00003984, 0.00003248, -0.00003397, 0.00003840, -0.00004449, 0.00004234, -0.00003043, 0.00002557, -0.00001979, -0.00000120, -0.00001104, 0.00001903, -0.00003249, 0.00001788, -0.00001178, -0.00000392, 0.00001160, -0.00000336, 0.00000680, -0.00001206, 0.00000147, 0.00000755, 0.00000312, -0.00001412, 0.00000634, 0.00000326, -0.00000295, 0.00000275, -0.00000387, 0.00000682, -0.00000769, 0.00000587, -0.00000176, 0.00000483, -0.00000780, 0.00000668, -0.00000265, -0.00000299, 0.00000429, -0.00000016, -0.00000292, 0.00001696, -0.00002305, 0.00000494, 0.00000822, -0.00000171, -0.00001287, 0.00001194, -0.00000015, -0.00000211, -0.00000171, 0.00000257, 0.00000076, -0.00000068, -0.00000060, -0.00000722, 0.00000109, 0.00000916, 0.00000568, -0.00000560, -0.00000738, -0.00001030, 0.00001861, 0.00000764, -0.00001459, -0.00000219, 0.00001005, -0.00001312, 0.00000847, },
		{ 0.00000454, -0.00000948, 0.00000090, -0.00000889, 0.00000908, 0.00000008, 0.00000622, -0.00001271, 0.00000485, -0.00000189, 0.00001005, -0.00002682, 0.00001631, -0.00000264, 0.00001389, -0.00002152, 0.00001749, -0.00001594, 0.00001284, -0.00002785, 0.00000924, -0.00000448, 0.00000833, 0.00001322, 0.00000868, -0.00002039, 0.00003486, -0.00001185, 0.00000998, -0.00001107, -0.00000236, -0.00000741, 0.00000344, -0.00001371, 0.00001438, -0.00001904, 0.00001626, -0.00002839, 0.00002479, -0.00003942, 0.00003879, -0.00006273, 0.00004813, -0.00010865, 0.00009445, -0.00037836, 0.00030551, 0.00062470, -0.00044732, 0.00008697, -0.00011239, 0.00005864, -0.00007391, 0.00004564, -0.00004792, 0.00003599, -0.00004191, 0.00003410, -0.00003392, 0.00002716, -0.00003022, 0.00003454, -0.00003760, 0.00003175, -0.00001908, 0.00001475, -0.00000833, -0.00000908, -0.00000782, 0.00001503, -0.00002282, 0.00000301, 0.00000303, -0.00001637, 0.00001785, -0.00000592, 0.00000855, -0.00001066, -0.00000129, 0.00000743, 0.00000442, -0.00001197, 0.00000127, 0.00000773, -0.00000625, 0.00000604, -0.00000702, 0.00000959, -0.00000951, 0.00000659, -0.00000324, 0.00000726, -0.00000881, 0.00000633, -0.00000233, -0.00000182, 0.00000059, 0.00000531, -0.00000900, 0.00002082, -0.00002041, -0.00000044, 0.00000849, 0.00000222, -0.00001370, 0.00000825, 0.00000307, -0.00000328, -0.00000067, 0.00000099, 0.00000238, -0.00000190, 0.00000082, -0.00000763, 0.00000030, 0.00000743, 0.00000788, -0.00000538, -0.00000446, -0.00001446, 0.00001575, 0.00001201, -0.00001149, -0.00000854, 0.00001366, -0.00001342, 0.00000699, -0.00000298, },
		{ -0.00000644, -0.00000186, -0.00000670, 0.00000476, 0.00000354, 0.00000448, -0.00000944, 0.00000069, 0.00000025, 0.00000917, -0.00002265, 0.00000831, 0.00000349, 0.00000947, -0.00001484, 0.00000885, -0.00000651, 0.00000357, -0.00001555, -0.00000594, 0.00000660, -0.00000180, 0.00001884, 0.00000934, -0.00001757, 0.00002610, 0.00000150, -0.00000083, -0.00000046, -0.00001244, 0.00000049, -0.00000359, -0.00000706, 0.00000624, -0.00000989, 0.00000726, -0.00001846, 0.00001370, -0.00002640, 0.00002305, -0.00004182, 0.00002535, -0.00007703, 0.00005425, -0.00026612, 0.00005830, 0.00076457, -0.00037533, 0.00000720, -0.00007096, 0.00002599, -0.00004784, 0.00002204, -0.00002900, 0.00001953, -0.00002680, 0.00001927, -0.00001994, 0.00001461, -0.00001943, 0.00002327, -0.00002342, 0.00001508, -0.00000310, 0.00000016, 0.00000590, -0.00001888, -0.00000191, 0.00000692, -0.00000923, -0.00001316, 0.00001598, -0.00002465, 0.00001932, -0.00000472, 0.00000738, -0.00000723, -0.00000473, 0.00000763, 0.00000467, -0.00000862, -0.00000406, 0.00001158, -0.00000865, 0.00000826, -0.00000888, 0.00001089, -0.00000974, 0.00000599, -0.00000369, 0.00000828, -0.00000820, 0.00000491, -0.00000164, -0.00000072, -0.00000261, 0.00000916, -0.00001200, 0.00002043, -0.00001416, -0.00000703, 0.00000900, 0.00000428, -0.00001146, 0.00000259, 0.00000695, -0.00000496, 0.00000098, -0.00000100, 0.00000415, -0.00000309, 0.00000205, -0.00000780, -0.00000038, 0.00000566, 0.00000958, -0.00000496, -0.00000194, -0.00001638, 0.00001060, 0.00001629, -0.00000787, -0.00001319, 0.00001391, -0.00001035, 0.00000290, 0.00000003, 0.00000236, },
		{ -0.00000560, -0.00000330, -0.00000058, 0.00000747, 0.00000202, -0.00000513, -0.00000419, 0.00000310, 0.00000680, -0.00001647, -0.00000095, 0.00001013, 0.00000423, -0.00000711, -0.00000046, 0.00000301, -0.00000543, -0.00000424, -0.00001795, 0.00001306, -0.00000723, 0.00001994, 0.00001206, -0.00001404, 0.00001623, 0.00001457, -0.00001013, 0.00000864, -0.00002078, 0.00000658, -0.00000939, -0.00000088, -0.00000189, -0.00000074, -0.00000202, -0.00000767, 0.00000108, -0.00001115, 0.00000452, -0.00001805, -0.00000101, -0.00004049, 0.00000629, -0.00015463, -0.00013537, 0.00080055, -0.00022532, -0.00010063, -0.00001846, -0.00001507, -0.00001529, -0.00000651, -0.00000606, -0.00000082, -0.00000801, 0.00000125, -0.00000335, -0.00000018, -0.00000622, 0.00000909, -0.00000661, -0.00000292, 0.00001299, -0.00001397, 0.00001891, -0.00002701, 0.00000336, -0.00000156, 0.00000368, -0.00002570, 0.00002282, -0.00002595, 0.00001469, 0.00000051, 0.00000353, -0.00000225, -0.00000846, 0.00000799, 0.00000408, -0.00000461, -0.00000895, 0.00001418, -0.00000966, 0.00000900, -0.00000911, 0.00001044, -0.00000822, 0.00000402, -0.00000297, 0.00000771, -0.00000597, 0.00000251, -0.00000044, -0.00000024, -0.00000443, 0.00001045, -0.00001135, 0.00001601, -0.00000530, -0.00001384, 0.00000945, 0.00000455, -0.00000689, -0.00000382, 0.00001040, -0.00000634, 0.00000247, -0.00000265, 0.00000538, -0.00000366, 0.00000259, -0.00000734, -0.00000136, 0.00000433, 0.00001033, -0.00000386, -0.00000057, -0.00001563, 0.00000353, 0.00002011, -0.00000413, -0.00001552, 0.00001083, -0.00000468, -0.00000273, 0.00000405, -0.00000094, -0.00000230, },
		{ -0.00000028, -0.00000519, 0.00001013, 0.00000047, -0.00000150, -0.00000801, 0.00000491, 0.00000478, -0.00001049, -0.00000899, 0.00001480, 0.00000069, -0.00000104, -0.00000750, 0.00000956, -0.00001098, 0.00000277, -0.00002387, 0.00001300, -0.00000706, 0.00001641, 0.00001657, -0.00001016, 0.00000670, 0.00002533, -0.00001588, 0.00001430, -0.00002540, 0.00000897, -0.00001207, 0.00000271, -0.00000753, 0.00000562, -0.00000854, 0.00000058, -0.00000906, 0.00000147, -0.00001091, 0.00000151, -0.00002241, -0.00000999, -0.00003387, -0.00006553, -0.00025864, 0.00073943, -0.00002539, -0.00020798, 0.00002538, -0.00005042, 0.00001239, -0.00003056, 0.00001294, -0.00001808, 0.00000805, -0.00001405, 0.00001041, -0.00001227, 0.00000489, -0.00000353, 0.00000822, -0.00001773, 0.00002502, -0.00002381, 0.00002717, -0.00003037, 0.00000508, -0.00000712, 0.00001214, -0.00003117, 0.00002142, -0.00001973, 0.00000478, 0.00000840, -0.00000161, 0.00000303, -0.00001158, 0.00000800, 0.00000318, -0.00000071, -0.00001261, 0.00001499, -0.00000893, 0.00000804, -0.00000762, 0.00000828, -0.00000513, 0.00000090, -0.00000122, 0.00000575, -0.00000248, -0.00000051, 0.00000117, -0.00000059, -0.00000452, 0.00000904, -0.00000749, 0.00000881, 0.00000439, -0.00001946, 0.00000918, 0.00000372, -0.00000135, -0.00000938, 0.00001223, -0.00000654, 0.00000301, -0.00000323, 0.00000549, -0.00000316, 0.00000210, -0.00000598, -0.00000293, 0.00000376, 0.00000994, -0.00000185, -0.00000063, -0.00001243, -0.00000454, 0.00002277, -0.00000030, -0.00001559, 0.00000531, 0.00000208, -0.00000827, 0.00000762, -0.00000392, 0.00000148, -0.00000883, },
		{ -0.00000768, 0.00001031, 0.00000091, 0.00000031, -0.00000958, 0.00000455, 0.00000432, -0.00000620, -0.00001417, 0.00001599, 0.00000026, 0.00000187, -0.00001070, 0.00001164, -0.00001174, 0.00000435, -0.00002326, 0.00000710, -0.00000267, 0.00001002, 0.00002126, -0.00000543, -0.00000188, 0.00003278, -0.00001729, 0.00001587, -0.00002569, 0.00000714, -0.00001115, 0.00000302, -0.00000966, 0.00000797, -0.00001092, 0.00000463, -0.00001461, 0.00000876, -0.00001986, 0.00001290, -0.00003413, 0.00000848, -0.00005789, -0.00000859, -0.00031181, 0.00060499, 0.00018720, -0.00028563, 0.00004374, -0.00006918, 0.00002681, -0.00004336, 0.00002242, -0.00002736, 0.00001690, -0.00002252, 0.00001767, -0.00001843, 0.00001090, -0.00001155, 0.00001792, -0.00002635, 0.00003041, -0.00002714, 0.00002874, -0.00002742, 0.00000187, -0.00000807, 0.00001436, -0.00002861, 0.00001232, -0.00000796, -0.00000770, 0.00001635, -0.00000595, 0.00000702, -0.00001307, 0.00000704, 0.00000261, 0.00000230, -0.00001442, 0.00001371, -0.00000642, 0.00000548, -0.00000463, 0.00000477, -0.00000096, -0.00000284, 0.00000110, 0.00000289, 0.00000160, -0.00000356, 0.00000288, -0.00000164, -0.00000312, 0.00000558, -0.00000171, 0.00000075, 0.00001287, -0.00002249, 0.00000747, 0.00000272, 0.00000361, -0.00001262, 0.00001155, -0.00000501, 0.00000210, -0.00000233, 0.00000421, -0.00000143, 0.00000054, -0.00000374, -0.00000506, 0.00000392, 0.00000854, 0.00000090, -0.00000192, -0.00000761, -0.00001225, 0.00002341, 0.00000380, -0.00001394, -0.00000122, 0.00000813, -0.00001208, 0.00000935, -0.00000533, 0.00000363, -0.00001028, 0.00000104, },
		{ 0.00000771, 0.00000349, 0.00000009, -0.00000871, 0.00000195, 0.00000557, -0.00000395, -0.00001612, 0.00001358, 0.00000289, 0.00000167, -0.00001023, 0.00000965, -0.00000838, 0.00000156, -0.00001799, -0.00000193, 0.00000296, 0.00000358, 0.00002402, 0.00000094, -0.00000936, 0.00003673, -0.00001462, 0.00001392, -0.00002231, 0.00000192, -0.00000745, 0.00000067, -0.00000870, 0.00000668, -0.00000954, 0.00000469, -0.00001551, 0.00001041, -0.00002185, 0.00001564, -0.00003570, 0.00001426, -0.00006475, 0.00001711, -0.00030833, 0.00043030, 0.00037529, -0.00031150, 0.00002812, -0.00006727, 0.00002512, -0.00004288, 0.00002086, -0.00002732, 0.00001726, -0.00002305, 0.00001754, -0.00001796, 0.00001114, -0.00001410, 0.00002147, -0.00002798, 0.00002876, -0.00002379, 0.00002373, -0.00001863, -0.00000573, -0.00000471, 0.00001082, -0.00001959, -0.00000151, 0.00000558, -0.00001902, 0.00002138, -0.00000744, 0.00000838, -0.00001219, 0.00000471, 0.00000285, 0.00000391, -0.00001410, 0.00001042, -0.00000244, 0.00000178, -0.00000071, 0.00000059, 0.00000353, -0.00000645, 0.00000333, -0.00000013, 0.00000545, -0.00000594, 0.00000421, -0.00000295, -0.00000093, 0.00000128, 0.00000423, -0.00000610, 0.00001833, -0.00002195, 0.00000390, 0.00000233, 0.00000678, -0.00001266, 0.00000814, -0.00000176, -0.00000024, -0.00000009, 0.00000180, 0.00000119, -0.00000171, -0.00000103, -0.00000737, 0.00000441, 0.00000662, 0.00000389, -0.00000377, -0.00000239, -0.00001818, 0.00002135, 0.00000836, -0.00001135, -0.00000724, 0.00001189, -0.00001300, 0.00000848, -0.00000451, 0.00000350, -0.00000936, 0.00000009, -0.00000753, },
		{ 0.00000736, -0.00000140, -0.00000621, -0.00000199, 0.00000770, -0.00000309, -0.00001562, 0.00000868, 0.00000728, -0.00000026, -0.00000758, 0.00000538, -0.00000302, -0.00000306, -0.00001122, -0.00001050, 0.00000657, -0.00000027, 0.00002318, 0.00000934, -0.00001558, 0.00003745, -0.00000880, 0.00000971, -0.00001667, -0.00000507, -0.00000262, -0.00000289, -0.00000602, 0.00000322, -0.00000596, 0.00000228, -0.00001324, 0.00000795, -0.00001867, 0.00001197, -0.00002990, 0.00001069, -0.00005875, 0.00001965, -0.00026842, 0.00024911, 0.00051016, -0.00027626, -0.00001903, -0.00004857, 0.00001048, -0.00003197, 0.00001083, -0.00002013, 0.00001100, -0.00001740, 0.00001178, -0.00001257, 0.00000711, -0.00001236, 0.00001991, -0.00002385, 0.00002161, -0.00001551, 0.00001407, -0.00000621, -0.00001554, 0.00000097, 0.00000389, -0.00000751, -0.00001579, 0.00001655, -0.00002564, 0.00002125, -0.00000491, 0.00000659, -0.00000891, 0.00000111, 0.00000395, 0.00000403, -0.00001186, 0.00000566, 0.00000233, -0.00000234, 0.00000334, -0.00000343, 0.00000748, -0.00000915, 0.00000478, -0.00000258, 0.00000826, -0.00000704, 0.00000474, -0.00000397, 0.00000115, -0.00000254, 0.00000870, -0.00001011, 0.00001970, -0.00001768, -0.00000140, 0.00000289, 0.00000757, -0.00000949, 0.00000259, 0.00000252, -0.00000341, 0.00000285, -0.00000105, 0.00000399, -0.00000397, 0.00000149, -0.00000926, 0.00000465, 0.00000486, 0.00000644, -0.00000535, 0.00000199, -0.00002121, 0.00001633, 0.00001333, -0.00000846, -0.00001161, 0.00001251, -0.00001077, 0.00000508, -0.00000164, 0.00000131, -0.00000638, -0.00000278, -0.00000519, 0.00000304, },
		{ -0.00000279, -0.00000345, -0.00000582, 0.00000935, -0.00000236, -0.00001408, 0.00000306, 0.00001150, -0.00000201, -0.00000482, 0.00000116, 0.00000177, -0.00000673, -0.00000606, -0.00001559, 0.00000595, -0.00000018, 0.00001836, 0.00001917, -0.00001997, 0.00003521, -0.00000092, 0.00000468, -0.00001031, -0.00001213, 0.00000168, -0.00000611, -0.00000319, -0.00000065, -0.00000208, -0.00000065, -0.00000983, 0.00000369, -0.00001308, 0.00000536, -0.00002091, 0.00000285, -0.00004646, 0.00000968, -0.00021266, 0.00008854, 0.00057682, -0.00018510, -0.00008502, -0.00002322, -0.00000959, -0.00001683, -0.00000245, -0.00001020, 0.00000196, -0.00000913, 0.00000375, -0.00000538, 0.00000155, -0.00000867, 0.00001548, -0.00001645, 0.00001171, -0.00000513, 0.00000274, 0.00000670, -0.00002457, 0.00000618, -0.00000328, 0.00000370, -0.00002630, 0.00002146, -0.00002544, 0.00001527, 0.00000140, 0.00000228, -0.00000401, -0.00000308, 0.00000551, 0.00000302, -0.00000834, 0.00000031, 0.00000697, -0.00000599, 0.00000666, -0.00000645, 0.00001012, -0.00001029, 0.00000494, -0.00000388, 0.00000949, -0.00000650, 0.00000419, -0.00000424, 0.00000235, -0.00000482, 0.00001060, -0.00001055, 0.00001700, -0.00001044, -0.00000759, 0.00000411, 0.00000621, -0.00000398, -0.00000379, 0.00000665, -0.00000639, 0.00000551, -0.00000339, 0.00000613, -0.00000543, 0.00000310, -0.00001009, 0.00000405, 0.00000384, 0.00000794, -0.00000593, 0.00000458, -0.00002079, 0.00000871, 0.00001823, -0.00000555, -0.00001377, 0.00001000, -0.00000603, 0.00000008, 0.00000238, -0.00000202, -0.00000235, -0.00000650, -0.00000189, -0.00000224, 0.00001105, },
		{ -0.00000175, -0.00000817, 0.00000930, -0.00000064, -0.00001285, -0.00000163, 0.00001385, -0.00000194, -0.00000365, -0.00000121, 0.00000410, -0.00000753, -0.00000440, -0.00001585, 0.00000075, 0.00000341, 0.00001075, 0.00002879, -0.00002149, 0.00003011, 0.00000810, -0.00000004, -0.00000444, -0.00001796, 0.00000425, -0.00000794, -0.00000134, -0.00000360, 0.00000063, -0.00000262, -0.00000694, -0.00000035, -0.00000765, -0.00000101, -0.00001256, -0.00000468, -0.00003382, -0.00000344, -0.00015711, -0.00003506, 0.00057533, -0.00005555, -0.00015086, -0.00000382, -0.00002641, -0.00000433, -0.00001336, -0.00000218, -0.00000575, -0.00000202, -0.00000306, 0.00000047, -0.00000281, -0.00000545, 0.00001058, -0.00000842, 0.00000191, 0.00000448, -0.00000732, 0.00001711, -0.00003010, 0.00000842, -0.00000790, 0.00001086, -0.00003027, 0.00001880, -0.00001845, 0.00000468, 0.00000983, -0.00000294, 0.00000107, -0.00000679, 0.00000684, 0.00000160, -0.00000447, -0.00000463, 0.00001056, -0.00000835, 0.00000854, -0.00000787, 0.00001094, -0.00000955, 0.00000360, -0.00000383, 0.00000897, -0.00000439, 0.00000264, -0.00000360, 0.00000223, -0.00000508, 0.00000968, -0.00000769, 0.00001127, -0.00000178, -0.00001341, 0.00000529, 0.00000352, 0.00000240, -0.00000934, 0.00000934, -0.00000816, 0.00000693, -0.00000437, 0.00000687, -0.00000551, 0.00000331, -0.00000949, 0.00000227, 0.00000391, 0.00000813, -0.00000517, 0.00000500, -0.00001718, -0.00000044, 0.00002223, -0.00000255, -0.00001384, 0.00000521, -0.00000015, -0.00000508, 0.00000620, -0.00000519, 0.00000137, -0.00000965, 0.00000087, -0.00000651, 0.00001303, -0.00000277, },
		{ -0.00000831, 0.00000699, 0.00000255, -0.00001263, -0.00000447, 0.00001351, 0.00000061, -0.00000469, -0.00000107, 0.00000343, -0.00000511, -0.00000627, -0.00001197, -0.00000739, 0.00000839, 0.00000275, 0.00003592, -0.00001896, 0.00002224, 0.00001754, -0.00000374, 0.00000034, -0.00002196, 0.00000468, -0.00000809, -0.00000083, -0.00000509, 0.00000160, -0.00000302, -0.00000533, -0.00000306, -0.00000388, -0.00000521, -0.00000709, -0.00000923, -0.00002439, -0.00001391, -0.00011104, -0.00011689, 0.00051825, 0.00008772, -0.00019688, -0.00000088, -0.00003361, 0.00000069, -0.00001809, 0.00000085, -0.00000945, 0.00000143, -0.00000638, 0.00000302, -0.00000432, -0.00000415, 0.00000681, -0.00000171, -0.00000571, 0.00001126, -0.00001405, 0.00002302, -0.00003051, 0.00000636, -0.00000849, 0.00001250, -0.00002709, 0.00000945, -0.00000684, -0.00000774, 0.00001773, -0.00000696, 0.00000473, -0.00000894, 0.00000719, 0.00000051, -0.00000115, -0.00000827, 0.00001239, -0.00000892, 0.00000860, -0.00000744, 0.00000987, -0.00000704, 0.00000098, -0.00000258, 0.00000697, -0.00000112, 0.00000046, -0.00000220, 0.00000085, -0.00000351, 0.00000652, -0.00000271, 0.00000423, 0.00000640, -0.00001745, 0.00000556, 0.00000064, 0.00000799, -0.00001255, 0.00000959, -0.00000801, 0.00000649, -0.00000354, 0.00000591, -0.00000402, 0.00000203, -0.00000746, -0.00000060, 0.00000494, 0.00000714, -0.00000319, 0.00000351, -0.00001135, -0.00000962, 0.00002433, 0.00000085, -0.00001241, -0.00000054, 0.00000524, -0.00000892, 0.00000857, -0.00000700, 0.00000355, -0.00001105, 0.00000189, -0.00000853, 0.00001229, -0.00000057, -0.00000188, },
		{ 0.00000280, 0.00000681, -0.00001322, -0.00000555, 0.00001076, 0.00000510, -0.00000733, 0.00000088, 0.00000065, -0.00000066, -0.00001011, -0.00000625, -0.00001555, 0.00001186, -0.00000297, 0.00003847, -0.00001163, 0.00001199, 0.00002672, -0.00000602, 0.00000387, -0.00002411, 0.00000319, -0.00000695, -0.00000137, -0.00000532, 0.00000104, -0.00000213, -0.00000485, -0.00000439, -0.00000203, -0.00000688, -0.00000483, -0.00001041, -0.00001890, -0.00001993, -0.00007714, -0.00016131, 0.00042544, 0.00021953, -0.00020841, -0.00001903, -0.00002951, -0.00000286, -0.00001601, -0.00000153, -0.00000872, 0.00000078, -0.00000587, 0.00000208, -0.00000296, -0.00000484, 0.00000454, 0.00000298, -0.00001033, 0.00001447, -0.00001671, 0.00002390, -0.00002575, 0.00000023, -0.00000525, 0.00000911, -0.00001834, -0.00000371, 0.00000579, -0.00001845, 0.00002237, -0.00000794, 0.00000578, -0.00000885, 0.00000618, 0.00000024, 0.00000100, -0.00001008, 0.00001221, -0.00000762, 0.00000691, -0.00000538, 0.00000726, -0.00000327, -0.00000238, -0.00000067, 0.00000408, 0.00000258, -0.00000176, -0.00000044, -0.00000135, -0.00000083, 0.00000230, 0.00000276, -0.00000217, 0.00001235, -0.00001866, 0.00000427, -0.00000146, 0.00001149, -0.00001250, 0.00000709, -0.00000584, 0.00000417, -0.00000100, 0.00000350, -0.00000128, -0.00000034, -0.00000447, -0.00000406, 0.00000645, 0.00000552, -0.00000054, 0.00000088, -0.00000468, -0.00001719, 0.00002364, 0.00000493, -0.00001021, -0.00000585, 0.00000872, -0.00001040, 0.00000872, -0.00000675, 0.00000355, -0.00001014, 0.00000066, -0.00000790, 0.00000875, 0.00000364, -0.00000366, -0.00000644, },
		{ 0.00001098, -0.00001368, -0.00000576, 0.00000677, 0.00001012, -0.00001005, 0.00000305, -0.00000243, 0.00000373, -0.00001349, -0.00000158, -0.00002070, 0.00001135, -0.00000450, 0.00003541, 0.00000017, 0.00000043, 0.00003476, -0.00000657, 0.00000625, -0.00002469, 0.00000038, -0.00000523, -0.00000234, -0.00000487, -0.00000030, -0.00000078, -0.00000476, -0.00000496, -0.00000149, -0.00000676, -0.00000476, -0.00000947, -0.00001604, -0.00002257, -0.00005359, -0.00017791, 0.00031816, 0.00032014, -0.00017972, -0.00005524, -0.00001774, -0.00001214, -0.00000956, -0.00000718, -0.00000531, -0.00000249, -0.00000294, -0.00000094, -0.00000009, -0.00000638, 0.00000307, 0.00000600, -0.00001228, 0.00001455, -0.00001586, 0.00002056, -0.00001713, -0.00000838, 0.00000023, 0.00000272, -0.00000707, -0.00001680, 0.00001554, -0.00002439, 0.00002187, -0.00000499, 0.00000389, -0.00000659, 0.00000391, 0.00000089, 0.00000179, -0.00001006, 0.00001021, -0.00000480, 0.00000396, -0.00000227, 0.00000377, 0.00000100, -0.00000569, 0.00000124, 0.00000107, 0.00000590, -0.00000337, 0.00000117, -0.00000370, 0.00000196, -0.00000163, 0.00000711, -0.00000635, 0.00001496, -0.00001666, 0.00000131, -0.00000224, 0.00001224, -0.00000913, 0.00000236, -0.00000225, 0.00000056, 0.00000255, 0.00000039, 0.00000195, -0.00000304, -0.00000131, -0.00000733, 0.00000767, 0.00000399, 0.00000203, -0.00000190, 0.00000135, -0.00002181, 0.00001977, 0.00000968, -0.00000783, -0.00000968, 0.00000952, -0.00000922, 0.00000662, -0.00000450, 0.00000148, -0.00000718, -0.00000245, -0.00000517, 0.00000328, 0.00000881, -0.00000610, -0.00000358, -0.00000212, },
		{ -0.00001283, -0.00000629, 0.00000303, 0.00001396, -0.00001117, 0.00000370, -0.00000393, 0.00000604, -0.00001426, -0.00000023, -0.00002090, 0.00000580, -0.00000148, 0.00002724, 0.00001485, -0.00001062, 0.00004047, -0.00000504, 0.00000751, -0.00002396, -0.00000321, -0.00000356, -0.00000318, -0.00000435, -0.00000169, 0.00000022, -0.00000430, -0.00000549, -0.00000137, -0.00000605, -0.00000529, -0.00000828, -0.00001375, -0.00002404, -0.00003676, -0.00017722, 0.00021399, 0.00037903, -0.00011514, -0.00009968, -0.00000579, -0.00002186, -0.00000304, -0.00001258, -0.00000210, -0.00000585, 0.00000006, -0.00000382, 0.00000230, -0.00000715, 0.00000123, 0.00000828, -0.00001255, 0.00001267, -0.00001279, 0.00001466, -0.00000681, -0.00001716, 0.00000565, -0.00000395, 0.00000325, -0.00002614, 0.00001948, -0.00002388, 0.00001590, 0.00000143, -0.00000017, -0.00000297, 0.00000102, 0.00000211, 0.00000150, -0.00000862, 0.00000700, -0.00000118, 0.00000049, 0.00000110, 0.00000024, 0.00000492, -0.00000821, 0.00000248, -0.00000138, 0.00000817, -0.00000391, 0.00000220, -0.00000556, 0.00000393, -0.00000415, 0.00000925, -0.00000752, 0.00001405, -0.00001190, -0.00000286, -0.00000176, 0.00001040, -0.00000329, -0.00000338, 0.00000164, -0.00000331, 0.00000606, -0.00000240, 0.00000472, -0.00000517, 0.00000118, -0.00000963, 0.00000791, 0.00000324, 0.00000385, -0.00000395, 0.00000559, -0.00002280, 0.00001299, 0.00001467, -0.00000551, -0.00001159, 0.00000768, -0.00000591, 0.00000299, -0.00000096, -0.00000186, -0.00000311, -0.00000639, -0.00000155, -0.00000262, 0.00001337, -0.00000780, -0.00000113, -0.00000625, 0.00001365, },
		{ -0.00000804, 0.00000071, 0.00001536, -0.00000952, 0.00000177, -0.00000277, 0.00000516, -0.00001141, -0.00000303, -0.00001605, -0.00000389, 0.00000464, 0.00001607, 0.00002969, -0.00001883, 0.00004257, -0.00000101, 0.00000761, -0.00002204, -0.00000726, -0.00000224, -0.00000368, -0.00000402, -0.00000268, 0.00000040, -0.00000309, -0.00000641, -0.00000107, -0.00000565, -0.00000521, -0.00000823, -0.00001050, -0.00002607, -0.00002354, -0.00016764, 0.00012391, 0.00039575, -0.00002730, -0.00013875, -0.00000206, -0.00002660, -0.00000060, -0.00001447, -0.00000169, -0.00000705, 0.00000104, -0.00000466, 0.00000254, -0.00000588, -0.00000190, 0.00001061, -0.00001211, 0.00001004, -0.00000893, 0.00000800, 0.00000300, -0.00002388, 0.00000887, -0.00000848, 0.00000989, -0.00002943, 0.00001651, -0.00001726, 0.00000581, 0.00000958, -0.00000479, 0.00000064, -0.00000154, 0.00000330, 0.00000065, -0.00000646, 0.00000338, 0.00000245, -0.00000271, 0.00000398, -0.00000260, 0.00000779, -0.00000938, 0.00000265, -0.00000283, 0.00000902, -0.00000318, 0.00000245, -0.00000649, 0.00000447, -0.00000468, 0.00000884, -0.00000579, 0.00001037, -0.00000556, -0.00000727, -0.00000057, 0.00000677, 0.00000358, -0.00000855, 0.00000457, -0.00000634, 0.00000847, -0.00000395, 0.00000623, -0.00000603, 0.00000236, -0.00001041, 0.00000672, 0.00000362, 0.00000454, -0.00000475, 0.00000749, -0.00002026, 0.00000429, 0.00001906, -0.00000313, -0.00001169, 0.00000391, -0.00000158, -0.00000103, 0.00000277, -0.00000532, 0.00000078, -0.00000980, 0.00000148, -0.00000732, 0.00001579, -0.00000748, -0.00000030, -0.00000836, 0.00001375, -0.00000984, },
		{ 0.00000017, 0.00001399, -0.00000500, -0.00000271, 0.00000093, 0.00000138, -0.00000550, -0.00000899, -0.00000792, -0.00001517, 0.00001108, 0.00000491, 0.00004157, -0.00002195, 0.00004004, 0.00000573, 0.00000651, -0.00001888, -0.00001168, -0.00000127, -0.00000400, -0.00000380, -0.00000330, -0.00000017, -0.00000131, -0.00000757, -0.00000050, -0.00000586, -0.00000408, -0.00000971, -0.00000600, -0.00002901, -0.00001253, -0.00015402, 0.00005197, 0.00037794, 0.00006685, -0.00015935, -0.00001255, -0.00002327, -0.00000439, -0.00001133, -0.00000520, -0.00000513, -0.00000094, -0.00000267, 0.00000010, -0.00000231, -0.00000635, 0.00001313, -0.00001145, 0.00000742, -0.00000527, 0.00000196, 0.00001066, -0.00002706, 0.00000865, -0.00000955, 0.00001158, -0.00002629, 0.00000765, -0.00000668, -0.00000567, 0.00001697, -0.00000803, 0.00000286, -0.00000287, 0.00000390, -0.00000014, -0.00000431, 0.00000013, 0.00000536, -0.00000499, 0.00000579, -0.00000427, 0.00000922, -0.00000900, 0.00000166, -0.00000319, 0.00000846, -0.00000137, 0.00000201, -0.00000638, 0.00000345, -0.00000329, 0.00000632, -0.00000210, 0.00000531, 0.00000083, -0.00001077, 0.00000051, 0.00000251, 0.00000983, -0.00001167, 0.00000552, -0.00000772, 0.00000906, -0.00000369, 0.00000609, -0.00000532, 0.00000201, -0.00000954, 0.00000411, 0.00000506, 0.00000419, -0.00000428, 0.00000717, -0.00001507, -0.00000486, 0.00002183, -0.00000034, -0.00001051, -0.00000065, 0.00000242, -0.00000424, 0.00000560, -0.00000776, 0.00000331, -0.00001154, 0.00000275, -0.00000962, 0.00001515, -0.00000453, -0.00000161, -0.00000795, 0.00001103, -0.00000467, -0.00001112, },
		{ 0.00001054, 0.00000140, -0.00000853, 0.00000583, -0.00000374, 0.00000160, -0.00001582, 0.00000053, -0.00002467, 0.00001476, -0.00000327, 0.00004786, -0.00001858, 0.00003267, 0.00001490, 0.00000439, -0.00001454, -0.00001647, -0.00000046, -0.00000445, -0.00000338, -0.00000381, -0.00000108, 0.00000049, -0.00000847, -0.00000008, -0.00000636, -0.00000231, -0.00001206, -0.00000112, -0.00003185, -0.00000415, -0.00013792, -0.00000318, 0.00033771, 0.00015078, -0.00015287, -0.00003833, -0.00001263, -0.00001358, -0.00000412, -0.00001168, -0.00000084, -0.00000521, 0.00000147, -0.00000423, 0.00000263, -0.00001119, 0.00001520, -0.00001035, 0.00000493, -0.00000216, -0.00000283, 0.00001543, -0.00002630, 0.00000493, -0.00000718, 0.00000875, -0.00001820, -0.00000441, 0.00000452, -0.00001538, 0.00002118, -0.00000831, 0.00000275, -0.00000254, 0.00000363, -0.00000051, -0.00000269, -0.00000226, 0.00000716, -0.00000605, 0.00000632, -0.00000464, 0.00000921, -0.00000722, -0.00000026, -0.00000268, 0.00000680, 0.00000109, 0.00000119, -0.00000540, 0.00000115, -0.00000062, 0.00000272, 0.00000217, 0.00000047, 0.00000581, -0.00001245, 0.00000080, -0.00000128, 0.00001409, -0.00001180, 0.00000408, -0.00000722, 0.00000766, -0.00000164, 0.00000442, -0.00000325, 0.00000040, -0.00000738, 0.00000054, 0.00000707, 0.00000323, -0.00000294, 0.00000527, -0.00000856, -0.00001285, 0.00002209, 0.00000313, -0.00000869, -0.00000485, 0.00000499, -0.00000581, 0.00000684, -0.00000848, 0.00000380, -0.00001101, 0.00000174, -0.00000915, 0.00001144, 0.00000071, -0.00000469, -0.00000549, 0.00000615, 0.00000182, -0.00001483, 0.00000086, },
		{ 0.00000798, -0.00001382, 0.00000996, -0.00000806, 0.00000746, -0.00002078, 0.00000620, -0.00002943, 0.00001343, -0.00000659, 0.00004728, -0.00000873, 0.00002139, 0.00002545, 0.00000187, -0.00000933, -0.00002141, 0.00000025, -0.00000528, -0.00000253, -0.00000451, -0.00000184, 0.00000167, -0.00000849, -0.00000031, -0.00000654, -0.00000072, -0.00001409, 0.00000261, -0.00003290, 0.00000032, -0.00011907, -0.00004551, 0.00028746, 0.00021255, -0.00011772, -0.00007466, 0.00000077, -0.00002462, 0.00000413, -0.00001853, 0.00000367, -0.00000993, 0.00000597, -0.00000867, 0.00000718, -0.00001489, 0.00001572, -0.00000818, 0.00000219, 0.00000059, -0.00000636, 0.00001741, -0.00002223, -0.00000126, -0.00000259, 0.00000312, -0.00000789, -0.00001618, 0.00001294, -0.00002067, 0.00002066, -0.00000502, 0.00000023, -0.00000073, 0.00000266, -0.00000038, -0.00000176, -0.00000365, 0.00000778, -0.00000590, 0.00000568, -0.00000391, 0.00000804, -0.00000447, -0.00000261, -0.00000173, 0.00000456, 0.00000364, 0.00000039, -0.00000389, -0.00000181, 0.00000240, -0.00000075, 0.00000566, -0.00000282, 0.00000842, -0.00001187, -0.00000003, -0.00000386, 0.00001560, -0.00000879, 0.00000063, -0.00000521, 0.00000472, 0.00000162, 0.00000188, -0.00000047, -0.00000182, -0.00000464, -0.00000323, 0.00000892, 0.00000232, -0.00000136, 0.00000270, -0.00000217, -0.00001833, 0.00001940, 0.00000728, -0.00000672, -0.00000785, 0.00000553, -0.00000550, 0.00000637, -0.00000737, 0.00000223, -0.00000841, -0.00000121, -0.00000644, 0.00000559, 0.00000703, -0.00000833, -0.00000216, 0.00000053, 0.00000815, -0.00001784, 0.00000159, 0.00000639, },
		{ -0.00001672, 0.00001148, -0.00000966, 0.00001007, -0.00002180, 0.00000704, -0.00002796, 0.00000651, -0.00000492, 0.00004032, 0.00000597, 0.00000828, 0.00003567, 0.00000003, -0.00000393, -0.00002599, 0.00000063, -0.00000639, -0.00000133, -0.00000543, -0.00000220, 0.00000187, -0.00000734, -0.00000151, -0.00000595, -0.00000008, -0.00001468, 0.00000389, -0.00003078, -0.00000015, -0.00009700, -0.00007900, 0.00023663, 0.00024711, -0.00005968, -0.00011221, 0.00001009, -0.00003272, 0.00000955, -0.00002263, 0.00000587, -0.00001284, 0.00000870, -0.00001121, 0.00000953, -0.00001594, 0.00001368, -0.00000433, -0.00000121, 0.00000333, -0.00000897, 0.00001725, -0.00001608, -0.00000830, 0.00000243, -0.00000301, 0.00000161, -0.00002450, 0.00001610, -0.00002025, 0.00001530, 0.00000123, -0.00000383, 0.00000171, 0.00000154, -0.00000000, -0.00000142, -0.00000420, 0.00000746, -0.00000484, 0.00000424, -0.00000249, 0.00000618, -0.00000131, -0.00000487, -0.00000081, 0.00000227, 0.00000575, -0.00000002, -0.00000223, -0.00000478, 0.00000483, -0.00000304, 0.00000740, -0.00000386, 0.00000846, -0.00000927, -0.00000182, -0.00000504, 0.00001434, -0.00000335, -0.00000383, -0.00000261, 0.00000109, 0.00000515, -0.00000064, 0.00000220, -0.00000385, -0.00000212, -0.00000641, 0.00000989, 0.00000208, -0.00000013, 0.00000031, 0.00000292, -0.00002055, 0.00001397, 0.00001168, -0.00000474, -0.00000933, 0.00000412, -0.00000370, 0.00000464, -0.00000488, -0.00000079, -0.00000454, -0.00000514, -0.00000268, -0.00000086, 0.00001277, -0.00001101, 0.00000051, -0.00000418, 0.00001268, -0.00001878, 0.00000112, 0.00000306, 0.00001295, },
		{ 0.00000942, -0.00000765, 0.00000861, -0.00001821, 0.00000272, -0.00002080, -0.00000458, 0.00000011, 0.00002910, 0.00002262, -0.00000379, 0.00004344, 0.00000013, 0.00000076, -0.00002945, 0.00000016, -0.00000738, -0.00000020, -0.00000629, -0.00000224, 0.00000114, -0.00000515, -0.00000352, -0.00000455, -0.00000068, -0.00001331, 0.00000219, -0.00002511, -0.00000544, -0.00007233, -0.00010586, 0.00019010, 0.00025635, 0.00001002, -0.00013975, 0.00000875, -0.00003378, 0.00000909, -0.00002167, 0.00000396, -0.00001239, 0.00000819, -0.00001051, 0.00000853, -0.00001357, 0.00000873, 0.00000127, -0.00000536, 0.00000621, -0.00001091, 0.00001565, -0.00000921, -0.00001455, 0.00000614, -0.00000759, 0.00000796, -0.00002743, 0.00001324, -0.00001459, 0.00000648, 0.00000877, -0.00000792, 0.00000361, 0.00000099, 0.00000027, -0.00000137, -0.00000427, 0.00000660, -0.00000329, 0.00000243, -0.00000083, 0.00000411, 0.00000173, -0.00000656, -0.00000028, 0.00000033, 0.00000707, 0.00000015, -0.00000072, -0.00000718, 0.00000597, -0.00000357, 0.00000708, -0.00000272, 0.00000643, -0.00000546, -0.00000398, -0.00000508, 0.00001095, 0.00000327, -0.00000794, -0.00000047, -0.00000225, 0.00000798, -0.00000227, 0.00000398, -0.00000500, -0.00000047, -0.00000837, 0.00000950, 0.00000283, 0.00000041, -0.00000134, 0.00000607, -0.00001949, 0.00000666, 0.00001557, -0.00000262, -0.00000942, 0.00000135, -0.00000125, 0.00000242, -0.00000181, -0.00000429, -0.00000059, -0.00000876, 0.00000068, -0.00000626, 0.00001633, -0.00001139, 0.00000128, -0.00000671, 0.00001426, -0.00001680, -0.00000135, 0.00000156, 0.00001513, -0.00001614, },
		{ -0.00000248, 0.00000371, -0.00001098, -0.00000526, -0.00001026, -0.00001691, 0.00000571, 0.00001668, 0.00003774, -0.00001190, 0.00004688, 0.00000325, 0.00000393, -0.00003099, -0.00000177, -0.00000767, 0.00000026, -0.00000660, -0.00000234, -0.00000007, -0.00000252, -0.00000578, -0.00000276, -0.00000226, -0.00001026, -0.00000199, -0.00001677, -0.00001401, -0.00004712, -0.00012605, 0.00014842, 0.00024724, 0.00007807, -0.00014766, -0.00000697, -0.00002632, 0.00000186, -0.00001521, -0.00000223, -0.00000839, 0.00000422, -0.00000643, 0.00000426, -0.00000814, 0.00000151, 0.00000792, -0.00000971, 0.00000890, -0.00001217, 0.00001314, -0.00000266, -0.00001884, 0.00000746, -0.00000941, 0.00001004, -0.00002473, 0.00000537, -0.00000572, -0.00000335, 0.00001536, -0.00001033, 0.00000387, 0.00000154, 0.00000016, -0.00000131, -0.00000420, 0.00000555, -0.00000161, 0.00000064, 0.00000069, 0.00000221, 0.00000428, -0.00000744, -0.00000030, -0.00000107, 0.00000749, 0.00000086, 0.00000053, -0.00000866, 0.00000551, -0.00000233, 0.00000506, -0.00000013, 0.00000338, -0.00000151, -0.00000578, -0.00000455, 0.00000645, 0.00000955, -0.00001040, 0.00000033, -0.00000453, 0.00000937, -0.00000243, 0.00000447, -0.00000493, 0.00000001, -0.00000889, 0.00000768, 0.00000454, 0.00000028, -0.00000210, 0.00000721, -0.00001583, -0.00000124, 0.00001803, -0.00000008, -0.00000854, -0.00000188, 0.00000090, 0.00000055, 0.00000101, -0.00000728, 0.00000233, -0.00001094, 0.00000248, -0.00000937, 0.00001672, -0.00000884, -0.00000035, -0.00000657, 0.00001264, -0.00001194, -0.00000574, 0.00000202, 0.00001433, -0.00001209, 0.00000404, },
		{ -0.00000274, -0.00000235, -0.00001421, 0.00000030, -0.00002702, 0.00000893, 0.00000603, 0.00004831, -0.00001392, 0.00004495, 0.00000980, 0.00000518, -0.00003013, -0.00000555, -0.00000680, -0.00000044, -0.00000593, -0.00000293, -0.00000119, -0.00000020, -0.00000752, -0.00000125, -0.00000412, -0.00000642, -0.00000733, -0.00000759, -0.00002330, -0.00002440, -0.00013803, 0.00010940, 0.00022875, 0.00013278, -0.00013086, -0.00003617, -0.00001234, -0.00001030, -0.00000503, -0.00001106, -0.00000222, -0.00000202, -0.00000020, -0.00000198, -0.00000110, -0.00000643, 0.00001423, -0.00001320, 0.00001067, -0.00001238, 0.00000989, 0.00000301, -0.00002071, 0.00000617, -0.00000836, 0.00000816, -0.00001774, -0.00000508, 0.00000345, -0.00001147, 0.00001896, -0.00000983, 0.00000194, 0.00000329, -0.00000032, -0.00000114, -0.00000419, 0.00000452, -0.00000001, -0.00000092, 0.00000189, 0.00000065, 0.00000617, -0.00000746, -0.00000084, -0.00000194, 0.00000712, 0.00000196, 0.00000152, -0.00000909, 0.00000357, 0.00000017, 0.00000218, 0.00000285, 0.00000049, 0.00000155, -0.00000661, -0.00000399, 0.00000188, 0.00001421, -0.00001037, -0.00000059, -0.00000543, 0.00000907, -0.00000107, 0.00000370, -0.00000374, -0.00000057, -0.00000815, 0.00000476, 0.00000678, -0.00000018, -0.00000220, 0.00000678, -0.00001070, -0.00000831, 0.00001833, 0.00000306, -0.00000716, -0.00000476, 0.00000202, -0.00000046, 0.00000303, -0.00000904, 0.00000349, -0.00001107, 0.00000214, -0.00000973, 0.00001385, -0.00000367, -0.00000396, -0.00000424, 0.00000852, -0.00000518, -0.00001108, 0.00000368, 0.00001138, -0.00000569, -0.00000322, 0.00000422, },
		{ 0.00000507, -0.00002116, 0.00000768, -0.00003208, 0.00000773, -0.00000088, 0.00005252, -0.00000914, 0.00003788, 0.00001920, 0.00000484, -0.00002694, -0.00001111, -0.00000475, -0.00000245, -0.00000416, -0.00000419, -0.00000178, 0.00000121, -0.00000814, -0.00000062, -0.00000553, -0.00000286, -0.00001223, 0.00000030, -0.00003059, -0.00000720, -0.00014005, 0.00007027, 0.00020877, 0.00016703, -0.00009048, -0.00007314, 0.00000310, -0.00002344, 0.00000545, -0.00001964, 0.00000374, -0.00000839, 0.00000613, -0.00000814, 0.00000548, -0.00001314, 0.00001856, -0.00001464, 0.00001070, -0.00001106, 0.00000587, 0.00000763, -0.00002035, 0.00000282, -0.00000525, 0.00000370, -0.00000880, -0.00001512, 0.00001006, -0.00001573, 0.00001839, -0.00000611, -0.00000194, 0.00000576, -0.00000082, -0.00000096, -0.00000424, 0.00000351, 0.00000146, -0.00000217, 0.00000272, -0.00000053, 0.00000740, -0.00000672, -0.00000173, -0.00000243, 0.00000617, 0.00000317, 0.00000232, -0.00000855, 0.00000056, 0.00000314, -0.00000055, 0.00000514, -0.00000129, 0.00000307, -0.00000617, -0.00000376, -0.00000196, 0.00001645, -0.00000768, -0.00000296, -0.00000518, 0.00000733, 0.00000141, 0.00000215, -0.00000188, -0.00000176, -0.00000665, 0.00000135, 0.00000893, -0.00000046, -0.00000203, 0.00000545, -0.00000529, -0.00001338, 0.00001612, 0.00000669, -0.00000556, -0.00000672, 0.00000179, -0.00000048, 0.00000406, -0.00000931, 0.00000272, -0.00000918, -0.00000013, -0.00000775, 0.00000857, 0.00000293, -0.00000832, -0.00000095, 0.00000335, 0.00000184, -0.00001588, 0.00000523, 0.00000769, 0.00000128, -0.00001031, 0.00001025, -0.00000839, },
		{ -0.00002397, 0.00000987, -0.00003083, 0.00000172, -0.00000366, 0.00005023, 0.00000153, 0.00002725, 0.00002995, 0.00000391, -0.00002207, -0.00001781, -0.00000200, -0.00000538, -0.00000165, -0.00000593, -0.00000178, 0.00000147, -0.00000741, -0.00000111, -0.00000601, -0.00000043, -0.00001539, 0.00000525, -0.00003384, 0.00000251, -0.00013147, 0.00002954, 0.00019198, 0.00017964, -0.00003349, -0.00010869, 0.00001341, -0.00003287, 0.00001245, -0.00002491, 0.00000699, -0.00001268, 0.00001043, -0.00001215, 0.00000967, -0.00001689, 0.00001960, -0.00001315, 0.00000842, -0.00000796, 0.00000112, 0.00001128, -0.00001831, -0.00000158, -0.00000138, -0.00000149, -0.00000044, -0.00002211, 0.00001216, -0.00001521, 0.00001375, 0.00000008, -0.00000676, 0.00000803, -0.00000077, -0.00000101, -0.00000419, 0.00000241, 0.00000288, -0.00000317, 0.00000323, -0.00000138, 0.00000803, -0.00000540, -0.00000276, -0.00000268, 0.00000489, 0.00000426, 0.00000301, -0.00000721, -0.00000295, 0.00000572, -0.00000226, 0.00000606, -0.00000150, 0.00000295, -0.00000464, -0.00000387, -0.00000467, 0.00001613, -0.00000289, -0.00000601, -0.00000439, 0.00000476, 0.00000429, 0.00000051, 0.00000003, -0.00000297, -0.00000500, -0.00000187, 0.00001037, -0.00000012, -0.00000200, 0.00000387, -0.00000063, -0.00001581, 0.00001166, 0.00001036, -0.00000380, -0.00000761, 0.00000038, 0.00000017, 0.00000430, -0.00000825, 0.00000039, -0.00000592, -0.00000354, -0.00000448, 0.00000230, 0.00000932, -0.00001190, 0.00000177, -0.00000121, 0.00000742, -0.00001863, 0.00000535, 0.00000477, 0.00000702, -0.00001529, 0.00001363, -0.00001121, 0.00000916, },
		{ 0.00000662, -0.00002397, -0.00000770, -0.00000345, 0.00004282, 0.00001580, 0.00001561, 0.00003990, 0.00000381, -0.00001661, -0.00002460, 0.00000053, -0.00000843, 0.00000087, -0.00000762, -0.00000148, 0.00000080, -0.00000562, -0.00000250, -0.00000556, 0.00000051, -0.00001618, 0.00000651, -0.00003232, 0.00000446, -0.00011351, -0.00001203, 0.00017909, 0.00017477, 0.00002927, -0.00013280, 0.00001283, -0.00003500, 0.00001326, -0.00002483, 0.00000596, -0.00001347, 0.00001130, -0.00001271, 0.00001031, -0.00001682, 0.00001689, -0.00000862, 0.00000393, -0.00000335, -0.00000404, 0.00001394, -0.00001522, -0.00000599, 0.00000197, -0.00000573, 0.00000536, -0.00002452, 0.00000925, -0.00001050, 0.00000639, 0.00000711, -0.00001101, 0.00000898, 0.00000037, -0.00000151, -0.00000390, 0.00000108, 0.00000430, -0.00000393, 0.00000349, -0.00000195, 0.00000815, -0.00000368, -0.00000377, -0.00000282, 0.00000345, 0.00000510, 0.00000361, -0.00000530, -0.00000637, 0.00000721, -0.00000247, 0.00000541, -0.00000027, 0.00000164, -0.00000257, -0.00000409, -0.00000626, 0.00001367, 0.00000297, -0.00000867, -0.00000383, 0.00000209, 0.00000679, -0.00000056, 0.00000141, -0.00000371, -0.00000371, -0.00000437, 0.00001067, 0.00000109, -0.00000228, 0.00000247, 0.00000271, -0.00001559, 0.00000574, 0.00001336, -0.00000170, -0.00000758, -0.00000171, 0.00000093, 0.00000417, -0.00000632, -0.00000279, -0.00000226, -0.00000695, -0.00000122, -0.00000336, 0.00001392, -0.00001332, 0.00000269, -0.00000385, 0.00001032, -0.00001838, 0.00000326, 0.00000355, 0.00001035, -0.00001694, 0.00001330, -0.00001026, 0.00000905, -0.00002197, },
		{ -0.00001384, -0.00001791, -0.00000231, 0.00003257, 0.00003065, 0.00000584, 0.00004693, 0.00000593, -0.00001165, -0.00003029, 0.00000182, -0.00001066, 0.00000252, -0.00000855, -0.00000140, -0.00000026, -0.00000338, -0.00000422, -0.00000455, 0.00000017, -0.00001476, 0.00000444, -0.00002666, 0.00000010, -0.00008926, -0.00005155, 0.00016757, 0.00016000, 0.00008612, -0.00013756, -0.00000146, -0.00002881, 0.00000731, -0.00001920, 0.00000065, -0.00001069, 0.00000858, -0.00000974, 0.00000745, -0.00001321, 0.00001102, -0.00000184, -0.00000194, 0.00000199, -0.00000895, 0.00001543, -0.00001151, -0.00000967, 0.00000394, -0.00000800, 0.00000767, -0.00002221, 0.00000236, -0.00000344, -0.00000156, 0.00001302, -0.00001322, 0.00000781, 0.00000283, -0.00000245, -0.00000332, -0.00000050, 0.00000567, -0.00000441, 0.00000345, -0.00000221, 0.00000775, -0.00000164, -0.00000469, -0.00000288, 0.00000194, 0.00000565, 0.00000409, -0.00000301, -0.00000920, 0.00000723, -0.00000115, 0.00000355, 0.00000179, -0.00000012, -0.00000065, -0.00000405, -0.00000697, 0.00000985, 0.00000865, -0.00000993, -0.00000409, -0.00000011, 0.00000834, -0.00000061, 0.00000197, -0.00000371, -0.00000304, -0.00000591, 0.00000975, 0.00000306, -0.00000281, 0.00000139, 0.00000462, -0.00001326, -0.00000053, 0.00001493, 0.00000091, -0.00000694, -0.00000385, 0.00000125, 0.00000404, -0.00000401, -0.00000599, 0.00000080, -0.00000933, 0.00000091, -0.00000719, 0.00001565, -0.00001188, 0.00000125, -0.00000404, 0.00001016, -0.00001509, -0.00000095, 0.00000413, 0.00001104, -0.00001515, 0.00000940, -0.00000591, 0.00000536, -0.00001607, -0.00000057, },
		{ -0.00002595, -0.00000241, 0.00002183, 0.00004321, 0.00000030, 0.00004952, 0.00001107, -0.00000792, -0.00003402, 0.00000104, -0.00001132, 0.00000262, -0.00000821, -0.00000199, -0.00000115, -0.00000141, -0.00000558, -0.00000356, -0.00000090, -0.00001185, 0.00000017, -0.00001856, -0.00000789, -0.00006290, -0.00008502, 0.00015335, 0.00014358, 0.00012765, -0.00011948, -0.00002803, -0.00001640, -0.00000353, -0.00000983, -0.00000733, -0.00000563, 0.00000339, -0.00000433, 0.00000227, -0.00000739, 0.00000346, 0.00000569, -0.00000781, 0.00000685, -0.00001269, 0.00001542, -0.00000740, -0.00001227, 0.00000427, -0.00000808, 0.00000670, -0.00001637, -0.00000641, 0.00000352, -0.00000783, 0.00001614, -0.00001244, 0.00000428, 0.00000634, -0.00000354, -0.00000262, -0.00000218, 0.00000680, -0.00000445, 0.00000304, -0.00000208, 0.00000681, 0.00000065, -0.00000549, -0.00000284, 0.00000040, 0.00000594, 0.00000439, -0.00000054, -0.00001111, 0.00000572, 0.00000130, 0.00000117, 0.00000385, -0.00000150, 0.00000051, -0.00000349, -0.00000713, 0.00000553, 0.00001305, -0.00000917, -0.00000538, -0.00000161, 0.00000870, 0.00000039, 0.00000175, -0.00000306, -0.00000293, -0.00000653, 0.00000782, 0.00000544, -0.00000330, 0.00000053, 0.00000535, -0.00000969, -0.00000600, 0.00001452, 0.00000405, -0.00000594, -0.00000555, 0.00000081, 0.00000410, -0.00000170, -0.00000857, 0.00000251, -0.00001005, 0.00000130, -0.00000863, 0.00001430, -0.00000776, -0.00000221, -0.00000215, 0.00000752, -0.00000964, -0.00000635, 0.00000578, 0.00000975, -0.00001085, 0.00000320, 0.00000031, -0.00000019, -0.00000838, -0.00000910, -0.00000359, },
		{ -0.00000508, 0.00001227, 0.00005143, 0.00000031, 0.00004725, 0.00001916, -0.00000546, -0.00003542, -0.00000208, -0.00001017, 0.00000096, -0.00000648, -0.00000340, -0.00000151, -0.00000021, -0.00000608, -0.00000304, -0.00000209, -0.00000842, -0.00000480, -0.00001010, -0.00001649, -0.00003866, -0.00010869, 0.00013277, 0.00013189, 0.00014897, -0.00008059, -0.00006132, -0.00000251, -0.00001560, 0.00000016, -0.00001535, -0.00000043, -0.00000238, 0.00000165, -0.00000338, -0.00000123, -0.00000388, 0.00001218, -0.00001217, 0.00000999, -0.00001442, 0.00001361, -0.00000297, -0.00001380, 0.00000324, -0.00000648, 0.00000349, -0.00000893, -0.00001456, 0.00000808, -0.00001080, 0.00001564, -0.00000862, -0.00000107, 0.00001014, -0.00000418, -0.00000212, -0.00000368, 0.00000741, -0.00000385, 0.00000213, -0.00000148, 0.00000532, 0.00000313, -0.00000613, -0.00000269, -0.00000112, 0.00000598, 0.00000450, 0.00000188, -0.00001187, 0.00000291, 0.00000421, -0.00000091, 0.00000518, -0.00000193, 0.00000062, -0.00000237, -0.00000695, 0.00000144, 0.00001549, -0.00000632, -0.00000742, -0.00000257, 0.00000803, 0.00000217, 0.00000108, -0.00000204, -0.00000314, -0.00000652, 0.00000531, 0.00000773, -0.00000339, -0.00000033, 0.00000530, -0.00000582, -0.00000979, 0.00001203, 0.00000747, -0.00000466, -0.00000657, -0.00000041, 0.00000428, 0.00000042, -0.00001014, 0.00000257, -0.00000899, -0.00000004, -0.00000787, 0.00001045, -0.00000193, -0.00000665, 0.00000074, 0.00000367, -0.00000351, -0.00001148, 0.00000727, 0.00000773, -0.00000555, -0.00000347, 0.00000639, -0.00000542, -0.00000131, -0.00001625, 0.00000391, -0.00002960, },
		{ 0.00000452, 0.00005447, 0.00000582, 0.00004093, 0.00002917, -0.00000363, -0.00003478, -0.00000722, -0.00000755, -0.00000212, -0.00000373, -0.00000538, -0.00000135, 0.00000006, -0.00000558, -0.00000318, -0.00000299, -0.00000529, -0.00000915, -0.00000311, -0.00002315, -0.00001978, -0.00012030, 0.00010409, 0.00012790, 0.00015049, -0.00002785, -0.00009305, 0.00000709, -0.00002480, 0.00000742, -0.00002073, 0.00000275, -0.00000682, 0.00000635, -0.00000767, 0.00000348, -0.00000932, 0.00001619, -0.00001389, 0.00001060, -0.00001368, 0.00000998, 0.00000166, -0.00001445, 0.00000136, -0.00000406, -0.00000052, -0.00000195, -0.00002002, 0.00000883, -0.00000993, 0.00001180, -0.00000265, -0.00000702, 0.00001310, -0.00000373, -0.00000215, -0.00000469, 0.00000722, -0.00000248, 0.00000070, -0.00000043, 0.00000338, 0.00000563, -0.00000654, -0.00000252, -0.00000254, 0.00000576, 0.00000446, 0.00000402, -0.00001141, -0.00000078, 0.00000684, -0.00000202, 0.00000533, -0.00000124, -0.00000021, -0.00000092, -0.00000652, -0.00000197, 0.00001576, -0.00000187, -0.00000957, -0.00000338, 0.00000672, 0.00000420, 0.00000043, -0.00000106, -0.00000332, -0.00000624, 0.00000268, 0.00000947, -0.00000279, -0.00000133, 0.00000488, -0.00000236, -0.00001147, 0.00000785, 0.00001060, -0.00000299, -0.00000692, -0.00000212, 0.00000432, 0.00000234, -0.00001056, 0.00000109, -0.00000657, -0.00000251, -0.00000570, 0.00000531, 0.00000416, -0.00001064, 0.00000328, 0.00000009, 0.00000170, -0.00001493, 0.00000742, 0.00000621, -0.00000078, -0.00000884, 0.00001050, -0.00000848, 0.00000322, -0.00002006, 0.00000693, -0.00002806, -0.00001037, },
		{ 0.00005260, 0.00001556, 0.00003235, 0.00003938, -0.00000128, -0.00003285, -0.00001356, -0.00000430, -0.00000580, -0.00000074, -0.00000735, -0.00000099, -0.00000036, -0.00000435, -0.00000380, -0.00000350, -0.00000293, -0.00001201, 0.00000132, -0.00002641, -0.00000784, -0.00011974, 0.00006818, 0.00013080, 0.00013714, 0.00002880, -0.00011452, 0.00000751, -0.00002807, 0.00000964, -0.00002173, 0.00000260, -0.00000870, 0.00000850, -0.00000940, 0.00000562, -0.00001186, 0.00001700, -0.00001255, 0.00000852, -0.00001061, 0.00000493, 0.00000617, -0.00001437, -0.00000080, -0.00000168, -0.00000403, 0.00000299, -0.00002163, 0.00000559, -0.00000592, 0.00000592, 0.00000392, -0.00001206, 0.00001414, -0.00000171, -0.00000289, -0.00000504, 0.00000612, -0.00000038, -0.00000113, 0.00000094, 0.00000119, 0.00000789, -0.00000653, -0.00000246, -0.00000368, 0.00000519, 0.00000442, 0.00000565, -0.00000978, -0.00000482, 0.00000857, -0.00000181, 0.00000430, 0.00000033, -0.00000156, 0.00000044, -0.00000576, -0.00000454, 0.00001416, 0.00000331, -0.00001099, -0.00000452, 0.00000520, 0.00000598, 0.00000025, -0.00000043, -0.00000323, -0.00000594, 0.00000030, 0.00001035, -0.00000141, -0.00000250, 0.00000432, 0.00000028, -0.00001117, 0.00000276, 0.00001274, -0.00000074, -0.00000679, -0.00000389, 0.00000391, 0.00000410, -0.00000994, -0.00000151, -0.00000352, -0.00000526, -0.00000316, 0.00000020, 0.00000909, -0.00001295, 0.00000431, -0.00000203, 0.00000482, -0.00001579, 0.00000556, 0.00000593, 0.00000245, -0.00001180, 0.00001161, -0.00000849, 0.00000442, -0.00001991, 0.00000534, -0.00002218, -0.00001704, -0.00000182, },
		{ 0.00002752, 0.00002379, 0.00004789, 0.00000278, -0.00003044, -0.00002009, -0.00000145, -0.00000909, 0.00000160, -0.00000865, -0.00000090, -0.00000103, -0.00000288, -0.00000451, -0.00000377, -0.00000145, -0.00001315, 0.00000293, -0.00002599, -0.00000272, -0.00010901, 0.00002823, 0.00013690, 0.00011650, 0.00007909, -0.00011910, -0.00000348, -0.00002459, 0.00000632, -0.00001821, -0.00000088, -0.00000792, 0.00000789, -0.00000837, 0.00000506, -0.00001151, 0.00001479, -0.00000858, 0.00000436, -0.00000597, -0.00000071, 0.00001005, -0.00001356, -0.00000287, 0.00000009, -0.00000625, 0.00000517, -0.00001943, -0.00000064, -0.00000041, -0.00000016, 0.00000935, -0.00001483, 0.00001258, 0.00000193, -0.00000419, -0.00000478, 0.00000421, 0.00000220, -0.00000304, 0.00000230, -0.00000088, 0.00000958, -0.00000587, -0.00000269, -0.00000436, 0.00000417, 0.00000449, 0.00000662, -0.00000715, -0.00000867, 0.00000895, -0.00000027, 0.00000245, 0.00000223, -0.00000284, 0.00000128, -0.00000461, -0.00000627, 0.00001124, 0.00000822, -0.00001100, -0.00000627, 0.00000378, 0.00000714, 0.00000073, -0.00000026, -0.00000281, -0.00000573, -0.00000167, 0.00001031, 0.00000060, -0.00000366, 0.00000365, 0.00000205, -0.00000937, -0.00000223, 0.00001328, 0.00000216, -0.00000636, -0.00000535, 0.00000286, 0.00000565, -0.00000841, -0.00000464, -0.00000067, -0.00000743, -0.00000121, -0.00000375, 0.00001184, -0.00001284, 0.00000330, -0.00000216, 0.00000544, -0.00001395, 0.00000180, 0.00000692, 0.00000393, -0.00001215, 0.00000977, -0.00000572, 0.00000273, -0.00001653, 0.00000043, -0.00001404, -0.00002311, -0.00000765, 0.00004682, },
		{ 0.00001737, 0.00005316, 0.00000933, -0.00002816, -0.00002595, 0.00000010, -0.00001118, 0.00000259, -0.00000879, -0.00000144, -0.00000154, -0.00000166, -0.00000488, -0.00000404, -0.00000072, -0.00001276, 0.00000217, -0.00002264, -0.00000291, -0.00009152, -0.00001126, 0.00014122, 0.00009638, 0.00011503, -0.00010412, -0.00002453, -0.00001615, -0.00000101, -0.00001161, -0.00000642, -0.00000541, 0.00000527, -0.00000533, 0.00000257, -0.00000908, 0.00001056, -0.00000309, -0.00000074, -0.00000091, -0.00000591, 0.00001269, -0.00001194, -0.00000473, 0.00000104, -0.00000695, 0.00000474, -0.00001445, -0.00000801, 0.00000456, -0.00000465, 0.00001228, -0.00001459, 0.00000838, 0.00000668, -0.00000558, -0.00000422, 0.00000183, 0.00000480, -0.00000459, 0.00000328, -0.00000245, 0.00001035, -0.00000438, -0.00000336, -0.00000447, 0.00000267, 0.00000476, 0.00000686, -0.00000384, -0.00001180, 0.00000783, 0.00000224, 0.00000039, 0.00000382, -0.00000350, 0.00000133, -0.00000308, -0.00000726, 0.00000768, 0.00001200, -0.00000924, -0.00000859, 0.00000250, 0.00000761, 0.00000183, -0.00000042, -0.00000219, -0.00000554, -0.00000323, 0.00000949, 0.00000291, -0.00000452, 0.00000280, 0.00000311, -0.00000680, -0.00000621, 0.00001191, 0.00000556, -0.00000566, -0.00000632, 0.00000124, 0.00000685, -0.00000617, -0.00000773, 0.00000132, -0.00000841, -0.00000044, -0.00000594, 0.00001204, -0.00001033, 0.00000043, -0.00000054, 0.00000397, -0.00001011, -0.00000304, 0.00000850, 0.00000418, -0.00001054, 0.00000591, -0.00000134, -0.00000047, -0.00001157, -0.00000566, -0.00000617, -0.00002666, -0.00001375, 0.00003830, 0.00003953, },
		{ 0.00005443, 0.00001849, -0.00002607, -0.00003071, -0.00000012, -0.00001168, 0.00000196, -0.00000765, -0.00000269, -0.00000166, -0.00000101, -0.00000467, -0.00000445, -0.00000051, -0.00001131, -0.00000011, -0.00001763, -0.00000628, -0.00007122, -0.00004574, 0.00013925, 0.00008273, 0.00013262, -0.00007153, -0.00005091, -0.00000656, -0.00000950, -0.00000438, -0.00001201, -0.00000275, 0.00000196, -0.00000155, -0.00000060, -0.00000586, 0.00000564, 0.00000257, -0.00000546, 0.00000335, -0.00000970, 0.00001358, -0.00000942, -0.00000637, 0.00000129, -0.00000641, 0.00000252, -0.00000830, -0.00001452, 0.00000725, -0.00000637, 0.00001216, -0.00001143, 0.00000229, 0.00001158, -0.00000634, -0.00000377, -0.00000055, 0.00000690, -0.00000534, 0.00000354, -0.00000320, 0.00000998, -0.00000201, -0.00000447, -0.00000402, 0.00000078, 0.00000518, 0.00000649, -0.00000025, -0.00001381, 0.00000528, 0.00000518, -0.00000122, 0.00000460, -0.00000326, 0.00000055, -0.00000134, -0.00000760, 0.00000405, 0.00001412, -0.00000581, -0.00001107, 0.00000115, 0.00000754, 0.00000326, -0.00000061, -0.00000162, -0.00000521, -0.00000452, 0.00000816, 0.00000515, -0.00000482, 0.00000168, 0.00000371, -0.00000411, -0.00000860, 0.00000875, 0.00000899, -0.00000454, -0.00000683, -0.00000069, 0.00000745, -0.00000336, -0.00001032, 0.00000201, -0.00000800, -0.00000096, -0.00000638, 0.00001006, -0.00000612, -0.00000348, 0.00000200, 0.00000140, -0.00000551, -0.00000770, 0.00000968, 0.00000408, -0.00000804, 0.00000141, 0.00000306, -0.00000353, -0.00000688, -0.00001084, -0.00000068, -0.00002656, -0.00002026, 0.00002798, 0.00004970, 0.00001461, },
		{ 0.00002956, -0.00002367, -0.00003436, -0.00000211, -0.00001072, -0.00000005, -0.00000558, -0.00000437, -0.00000146, -0.00000095, -0.00000392, -0.00000498, -0.00000068, -0.00000929, -0.00000298, -0.00001228, -0.00001068, -0.00005158, -0.00007178, 0.00012830, 0.00007834, 0.00013240, -0.00002729, -0.00007584, -0.00000033, -0.00001602, 0.00000098, -0.00001569, -0.00000140, -0.00000074, 0.00000166, -0.00000317, -0.00000307, 0.00000128, 0.00000718, -0.00000872, 0.00000591, -0.00001146, 0.00001253, -0.00000608, -0.00000784, 0.00000109, -0.00000519, -0.00000044, -0.00000261, -0.00001856, 0.00000670, -0.00000510, 0.00000938, -0.00000626, -0.00000440, 0.00001538, -0.00000577, -0.00000384, -0.00000249, 0.00000805, -0.00000499, 0.00000288, -0.00000301, 0.00000849, 0.00000108, -0.00000582, -0.00000321, -0.00000126, 0.00000555, 0.00000576, 0.00000313, -0.00001437, 0.00000157, 0.00000792, -0.00000189, 0.00000435, -0.00000218, -0.00000078, 0.00000030, -0.00000733, 0.00000077, 0.00001445, -0.00000123, -0.00001309, -0.00000057, 0.00000717, 0.00000466, -0.00000050, -0.00000133, -0.00000460, -0.00000562, 0.00000654, 0.00000700, -0.00000441, 0.00000029, 0.00000400, -0.00000176, -0.00000925, 0.00000438, 0.00001178, -0.00000276, -0.00000705, -0.00000256, 0.00000723, -0.00000020, -0.00001201, 0.00000129, -0.00000639, -0.00000244, -0.00000554, 0.00000669, -0.00000130, -0.00000733, 0.00000437, -0.00000108, -0.00000146, -0.00001100, 0.00000951, 0.00000451, -0.00000570, -0.00000245, 0.00000613, -0.00000512, -0.00000382, -0.00001374, 0.00000142, -0.00002290, -0.00002649, 0.00001637, 0.00005671, 0.00001615, 0.00005184, },
		{ -0.00002009, -0.00003728, -0.00000544, -0.00000884, -0.00000281, -0.00000321, -0.00000601, -0.00000125, -0.00000126, -0.00000289, -0.00000543, -0.00000117, -0.00000708, -0.00000571, -0.00000763, -0.00001448, -0.00003503, -0.00008770, 0.00010803, 0.00008268, 0.00011869, 0.00002032, -0.00009242, -0.00000107, -0.00001840, 0.00000283, -0.00001632, -0.00000217, -0.00000208, 0.00000350, -0.00000437, -0.00000148, -0.00000179, 0.00001005, -0.00000999, 0.00000642, -0.00001111, 0.00000980, -0.00000221, -0.00000911, 0.00000069, -0.00000381, -0.00000316, 0.00000141, -0.00001936, 0.00000302, -0.00000162, 0.00000511, -0.00000052, -0.00001019, 0.00001700, -0.00000341, -0.00000460, -0.00000376, 0.00000801, -0.00000350, 0.00000141, -0.00000201, 0.00000614, 0.00000448, -0.00000705, -0.00000238, -0.00000310, 0.00000562, 0.00000501, 0.00000585, -0.00001336, -0.00000285, 0.00000988, -0.00000136, 0.00000320, -0.00000058, -0.00000222, 0.00000150, -0.00000648, -0.00000193, 0.00001322, 0.00000371, -0.00001396, -0.00000294, 0.00000675, 0.00000572, 0.00000011, -0.00000141, -0.00000372, -0.00000656, 0.00000479, 0.00000832, -0.00000332, -0.00000122, 0.00000404, 0.00000005, -0.00000840, -0.00000036, 0.00001324, -0.00000013, -0.00000717, -0.00000404, 0.00000612, 0.00000300, -0.00001256, -0.00000070, -0.00000405, -0.00000431, -0.00000413, 0.00000295, 0.00000303, -0.00001010, 0.00000565, -0.00000253, 0.00000109, -0.00001219, 0.00000756, 0.00000585, -0.00000413, -0.00000490, 0.00000720, -0.00000471, -0.00000281, -0.00001407, 0.00000035, -0.00001691, -0.00003120, 0.00000375, 0.00005978, 0.00002178, 0.00004641, 0.00004124, },
		{ -0.00003987, -0.00000956, -0.00000677, -0.00000557, -0.00000125, -0.00000709, -0.00000133, -0.00000163, -0.00000193, -0.00000554, -0.00000198, -0.00000498, -0.00000783, -0.00000428, -0.00001679, -0.00002276, -0.00009366, 0.00008034, 0.00009263, 0.00009798, 0.00006290, -0.00009557, -0.00001022, -0.00001628, 0.00000099, -0.00001396, -0.00000487, -0.00000209, 0.00000385, -0.00000409, -0.00000118, -0.00000343, 0.00001113, -0.00000931, 0.00000510, -0.00000908, 0.00000598, 0.00000168, -0.00001000, 0.00000021, -0.00000262, -0.00000504, 0.00000323, -0.00001709, -0.00000279, 0.00000261, 0.00000089, 0.00000430, -0.00001382, 0.00001581, 0.00000067, -0.00000585, -0.00000435, 0.00000686, -0.00000114, -0.00000054, -0.00000055, 0.00000339, 0.00000765, -0.00000770, -0.00000192, -0.00000438, 0.00000513, 0.00000452, 0.00000758, -0.00001088, -0.00000742, 0.00001063, 0.00000032, 0.00000154, 0.00000103, -0.00000331, 0.00000200, -0.00000514, -0.00000393, 0.00001087, 0.00000817, -0.00001321, -0.00000599, 0.00000630, 0.00000634, 0.00000122, -0.00000176, -0.00000271, -0.00000722, 0.00000292, 0.00000911, -0.00000176, -0.00000260, 0.00000377, 0.00000136, -0.00000659, -0.00000456, 0.00001286, 0.00000332, -0.00000718, -0.00000504, 0.00000431, 0.00000583, -0.00001180, -0.00000364, -0.00000157, -0.00000594, -0.00000286, -0.00000032, 0.00000601, -0.00001118, 0.00000536, -0.00000249, 0.00000178, -0.00001120, 0.00000401, 0.00000795, -0.00000336, -0.00000587, 0.00000637, -0.00000263, -0.00000339, -0.00001254, -0.00000263, -0.00001049, -0.00003307, -0.00000949, 0.00005851, 0.00003061, 0.00003964, 0.00005196, -0.00001436, },
		{ -0.00001399, -0.00000516, -0.00000769, -0.00000022, -0.00000729, -0.00000191, -0.00000183, -0.00000131, -0.00000517, -0.00000305, -0.00000316, -0.00000911, -0.00000237, -0.00001737, -0.00001480, -0.00009126, 0.00004856, 0.00010377, 0.00007697, 0.00009398, -0.00008343, -0.00002639, -0.00001117, -0.00000332, -0.00000978, -0.00000851, -0.00000147, 0.00000322, -0.00000279, -0.00000172, -0.00000406, 0.00001082, -0.00000722, 0.00000260, -0.00000614, 0.00000189, 0.00000502, -0.00001032, -0.00000035, -0.00000176, -0.00000588, 0.00000300, -0.00001269, -0.00000913, 0.00000597, -0.00000190, 0.00000711, -0.00001465, 0.00001190, 0.00000587, -0.00000707, -0.00000456, 0.00000494, 0.00000160, -0.00000245, 0.00000086, 0.00000080, 0.00001004, -0.00000737, -0.00000214, -0.00000488, 0.00000395, 0.00000447, 0.00000824, -0.00000727, -0.00001152, 0.00000991, 0.00000286, -0.00000011, 0.00000219, -0.00000374, 0.00000174, -0.00000349, -0.00000521, 0.00000791, 0.00001147, -0.00001070, -0.00000944, 0.00000568, 0.00000658, 0.00000264, -0.00000211, -0.00000184, -0.00000742, 0.00000091, 0.00000946, -0.00000000, -0.00000358, 0.00000311, 0.00000230, -0.00000440, -0.00000751, 0.00001056, 0.00000721, -0.00000687, -0.00000566, 0.00000216, 0.00000783, -0.00000969, -0.00000706, 0.00000045, -0.00000684, -0.00000220, -0.00000259, 0.00000730, -0.00001049, 0.00000359, -0.00000109, 0.00000088, -0.00000858, -0.00000038, 0.00001015, -0.00000291, -0.00000585, 0.00000435, 0.00000018, -0.00000451, -0.00001036, -0.00000594, -0.00000545, -0.00003139, -0.00002253, 0.00005266, 0.00004137, 0.00003325, 0.00006026, -0.00000581, -0.00004237, },
		{ -0.00000436, -0.00000887, -0.00000029, -0.00000662, -0.00000292, -0.00000182, -0.00000113, -0.00000434, -0.00000421, -0.00000181, -0.00000951, -0.00000175, -0.00001648, -0.00001048, -0.00008292, 0.00001649, 0.00011184, 0.00006092, 0.00011026, -0.00005772, -0.00004578, -0.00000597, -0.00000803, -0.00000551, -0.00001169, -0.00000121, 0.00000237, -0.00000122, -0.00000242, -0.00000428, 0.00000977, -0.00000443, -0.00000029, -0.00000313, -0.00000173, 0.00000734, -0.00000990, -0.00000104, -0.00000118, -0.00000585, 0.00000134, -0.00000748, -0.00001440, 0.00000716, -0.00000249, 0.00000755, -0.00001280, 0.00000603, 0.00001114, -0.00000755, -0.00000478, 0.00000271, 0.00000412, -0.00000378, 0.00000177, -0.00000116, 0.00001125, -0.00000586, -0.00000312, -0.00000456, 0.00000215, 0.00000481, 0.00000795, -0.00000307, -0.00001457, 0.00000766, 0.00000581, -0.00000125, 0.00000259, -0.00000342, 0.00000086, -0.00000181, -0.00000577, 0.00000484, 0.00001321, -0.00000668, -0.00001273, 0.00000456, 0.00000668, 0.00000402, -0.00000214, -0.00000137, -0.00000703, -0.00000128, 0.00000948, 0.00000171, -0.00000399, 0.00000206, 0.00000297, -0.00000228, -0.00000889, 0.00000670, 0.00001088, -0.00000588, -0.00000618, 0.00000016, 0.00000864, -0.00000640, -0.00001041, 0.00000152, -0.00000676, -0.00000230, -0.00000375, 0.00000702, -0.00000842, 0.00000086, 0.00000108, -0.00000090, -0.00000528, -0.00000459, 0.00001160, -0.00000212, -0.00000556, 0.00000209, 0.00000265, -0.00000508, -0.00000869, -0.00000825, -0.00000290, -0.00002639, -0.00003404, 0.00004218, 0.00005255, 0.00002877, 0.00006513, 0.00000572, -0.00004460, -0.00001851, },
		{ -0.00000918, -0.00000126, -0.00000539, -0.00000408, -0.00000174, -0.00000126, -0.00000327, -0.00000521, -0.00000107, -0.00000909, -0.00000212, -0.00001461, -0.00000877, -0.00007120, -0.00001250, 0.00011373, 0.00005267, 0.00011199, -0.00002314, -0.00006322, -0.00000381, -0.00001113, -0.00000276, -0.00001325, -0.00000205, 0.00000192, -0.00000002, -0.00000269, -0.00000460, 0.00000848, -0.00000152, -0.00000292, -0.00000068, -0.00000436, 0.00000842, -0.00000873, -0.00000189, -0.00000074, -0.00000528, -0.00000094, -0.00000275, -0.00001740, 0.00000563, -0.00000089, 0.00000602, -0.00000911, -0.00000053, 0.00001527, -0.00000664, -0.00000533, 0.00000062, 0.00000590, -0.00000415, 0.00000190, -0.00000222, 0.00001111, -0.00000323, -0.00000469, -0.00000365, 0.00000001, 0.00000529, 0.00000708, 0.00000109, -0.00001608, 0.00000405, 0.00000863, -0.00000152, 0.00000220, -0.00000255, -0.00000029, -0.00000040, -0.00000566, 0.00000202, 0.00001336, -0.00000176, -0.00001519, 0.00000263, 0.00000684, 0.00000508, -0.00000161, -0.00000144, -0.00000605, -0.00000355, 0.00000918, 0.00000325, -0.00000382, 0.00000076, 0.00000339, -0.00000049, -0.00000876, 0.00000204, 0.00001351, -0.00000383, -0.00000688, -0.00000130, 0.00000814, -0.00000234, -0.00001308, 0.00000132, -0.00000571, -0.00000305, -0.00000401, 0.00000563, -0.00000565, -0.00000213, 0.00000330, -0.00000273, -0.00000227, -0.00000770, 0.00001165, -0.00000051, -0.00000556, 0.00000034, 0.00000399, -0.00000445, -0.00000812, -0.00000903, -0.00000290, -0.00001933, -0.00004236, 0.00002731, 0.00006250, 0.00002742, 0.00006614, 0.00001978, -0.00004595, -0.00002322, -0.00000442, },
		{ -0.00000268, -0.00000412, -0.00000500, -0.00000185, -0.00000150, -0.00000225, -0.00000577, -0.00000104, -0.00000801, -0.00000309, -0.00001227, -0.00000863, -0.00005834, -0.00003604, 0.00010807, 0.00005250, 0.00010232, 0.00001399, -0.00007377, -0.00000693, -0.00001145, -0.00000232, -0.00001273, -0.00000419, 0.00000205, 0.00000055, -0.00000230, -0.00000521, 0.00000719, 0.00000115, -0.00000493, 0.00000089, -0.00000584, 0.00000832, -0.00000698, -0.00000287, -0.00000033, -0.00000451, -0.00000311, 0.00000056, -0.00001764, 0.00000167, 0.00000210, 0.00000351, -0.00000479, -0.00000642, 0.00001725, -0.00000400, -0.00000631, -0.00000107, 0.00000667, -0.00000345, 0.00000126, -0.00000239, 0.00000980, 0.00000015, -0.00000644, -0.00000256, -0.00000204, 0.00000557, 0.00000606, 0.00000460, -0.00001579, -0.00000055, 0.00001082, -0.00000075, 0.00000121, -0.00000147, -0.00000134, 0.00000048, -0.00000497, -0.00000029, 0.00001211, 0.00000329, -0.00001620, -0.00000028, 0.00000716, 0.00000567, -0.00000046, -0.00000198, -0.00000464, -0.00000567, 0.00000849, 0.00000463, -0.00000322, -0.00000054, 0.00000346, 0.00000095, -0.00000754, -0.00000252, 0.00001441, -0.00000056, -0.00000784, -0.00000210, 0.00000652, 0.00000187, -0.00001452, -0.00000025, -0.00000393, -0.00000414, -0.00000375, 0.00000370, -0.00000287, -0.00000471, 0.00000497, -0.00000394, -0.00000024, -0.00000918, 0.00001006, 0.00000201, -0.00000603, -0.00000062, 0.00000401, -0.00000262, -0.00000854, -0.00000854, -0.00000460, -0.00001207, -0.00004603, 0.00000893, 0.00006946, 0.00002996, 0.00006352, 0.00003536, -0.00004552, -0.00002840, -0.00000516, -0.00000895, },
		{ -0.00000327, -0.00000536, -0.00000230, -0.00000164, -0.00000154, -0.00000571, -0.00000167, -0.00000650, -0.00000429, -0.00000994, -0.00000920, -0.00004599, -0.00005295, 0.00009519, 0.00005858, 0.00008602, 0.00004743, -0.00007404, -0.00001576, -0.00000923, -0.00000394, -0.00001055, -0.00000714, 0.00000244, 0.00000067, -0.00000142, -0.00000592, 0.00000582, 0.00000356, -0.00000623, 0.00000157, -0.00000630, 0.00000734, -0.00000492, -0.00000387, 0.00000011, -0.00000379, -0.00000468, 0.00000206, -0.00001538, -0.00000375, 0.00000525, 0.00000120, -0.00000103, -0.00001052, 0.00001655, 0.00000021, -0.00000749, -0.00000235, 0.00000642, -0.00000189, 0.00000010, -0.00000195, 0.00000770, 0.00000373, -0.00000784, -0.00000178, -0.00000358, 0.00000531, 0.00000529, 0.00000703, -0.00001372, -0.00000558, 0.00001191, 0.00000100, -0.00000001, -0.00000058, -0.00000199, 0.00000074, -0.00000391, -0.00000193, 0.00000988, 0.00000771, -0.00001537, -0.00000405, 0.00000752, 0.00000581, 0.00000114, -0.00000273, -0.00000316, -0.00000733, 0.00000724, 0.00000594, -0.00000241, -0.00000155, 0.00000307, 0.00000215, -0.00000575, -0.00000615, 0.00001323, 0.00000373, -0.00000884, -0.00000241, 0.00000430, 0.00000547, -0.00001427, -0.00000307, -0.00000181, -0.00000518, -0.00000336, 0.00000177, -0.00000061, -0.00000649, 0.00000574, -0.00000423, 0.00000056, -0.00000901, 0.00000710, 0.00000507, -0.00000670, -0.00000097, 0.00000304, -0.00000015, -0.00000926, -0.00000764, -0.00000666, -0.00000644, -0.00004439, -0.00001132, 0.00007163, 0.00003662, 0.00005807, 0.00005109, -0.00004229, -0.00003434, -0.00000629, -0.00000861, -0.00000404, },
		{ -0.00000507, -0.00000311, -0.00000163, -0.00000124, -0.00000508, -0.00000274, -0.00000491, -0.00000536, -0.00000795, -0.00000987, -0.00003518, -0.00006313, 0.00007666, 0.00006793, 0.00006814, 0.00007243, -0.00006312, -0.00002874, -0.00000595, -0.00000652, -0.00000775, -0.00001000, 0.00000246, 0.00000075, -0.00000047, -0.00000636, 0.00000414, 0.00000579, -0.00000692, 0.00000154, -0.00000605, 0.00000585, -0.00000283, -0.00000480, 0.00000057, -0.00000323, -0.00000546, 0.00000188, -0.00001145, -0.00000929, 0.00000732, 0.00000007, 0.00000136, -0.00001230, 0.00001330, 0.00000534, -0.00000835, -0.00000341, 0.00000540, 0.00000014, -0.00000114, -0.00000129, 0.00000534, 0.00000691, -0.00000840, -0.00000167, -0.00000431, 0.00000435, 0.00000498, 0.00000823, -0.00001024, -0.00001036, 0.00001158, 0.00000348, -0.00000104, -0.00000017, -0.00000211, 0.00000048, -0.00000274, -0.00000285, 0.00000714, 0.00001091, -0.00001269, -0.00000826, 0.00000759, 0.00000576, 0.00000287, -0.00000331, -0.00000199, -0.00000821, 0.00000531, 0.00000728, -0.00000164, -0.00000207, 0.00000218, 0.00000317, -0.00000384, -0.00000836, 0.00001009, 0.00000843, -0.00000937, -0.00000270, 0.00000216, 0.00000777, -0.00001216, -0.00000678, 0.00000015, -0.00000582, -0.00000316, 0.00000015, 0.00000091, -0.00000740, 0.00000563, -0.00000362, 0.00000026, -0.00000757, 0.00000341, 0.00000802, -0.00000710, -0.00000116, 0.00000173, 0.00000213, -0.00000941, -0.00000724, -0.00000791, -0.00000358, -0.00003790, -0.00003096, 0.00006737, 0.00004701, 0.00005107, 0.00006544, -0.00003540, -0.00004112, -0.00000768, -0.00000850, -0.00000498, -0.00000311, },
		{ -0.00000406, -0.00000158, -0.00000125, -0.00000407, -0.00000392, -0.00000359, -0.00000599, -0.00000652, -0.00001028, -0.00002639, -0.00006727, 0.00005476, 0.00007737, 0.00005281, 0.00008658, -0.00004268, -0.00004275, -0.00000375, -0.00000862, -0.00000550, -0.00001186, 0.00000156, 0.00000118, 0.00000016, -0.00000619, 0.00000198, 0.00000790, -0.00000707, 0.00000098, -0.00000536, 0.00000418, -0.00000090, -0.00000560, 0.00000103, -0.00000284, -0.00000552, 0.00000047, -0.00000692, -0.00001369, 0.00000742, 0.00000055, 0.00000216, -0.00001188, 0.00000826, 0.00001035, -0.00000827, -0.00000451, 0.00000395, 0.00000211, -0.00000201, -0.00000081, 0.00000318, 0.00000921, -0.00000781, -0.00000239, -0.00000418, 0.00000274, 0.00000510, 0.00000834, -0.00000591, -0.00001417, 0.00000968, 0.00000634, -0.00000153, -0.00000034, -0.00000181, -0.00000006, -0.00000174, -0.00000306, 0.00000431, 0.00001258, -0.00000851, -0.00001226, 0.00000697, 0.00000580, 0.00000434, -0.00000334, -0.00000144, -0.00000814, 0.00000272, 0.00000862, -0.00000099, -0.00000203, 0.00000090, 0.00000399, -0.00000210, -0.00000905, 0.00000562, 0.00001263, -0.00000880, -0.00000347, 0.00000074, 0.00000836, -0.00000838, -0.00001073, 0.00000141, -0.00000580, -0.00000328, -0.00000101, 0.00000169, -0.00000759, 0.00000486, -0.00000239, -0.00000074, -0.00000548, -0.00000026, 0.00001022, -0.00000674, -0.00000166, 0.00000069, 0.00000353, -0.00000841, -0.00000786, -0.00000780, -0.00000359, -0.00002827, -0.00004713, 0.00005564, 0.00005996, 0.00004418, 0.00007693, -0.00002434, -0.00004850, -0.00000937, -0.00000875, -0.00000541, -0.00000363, -0.00000429, },
		{ -0.00000170, -0.00000140, -0.00000299, -0.00000481, -0.00000284, -0.00000602, -0.00000572, -0.00001027, -0.00001972, -0.00006655, 0.00003185, 0.00008424, 0.00004257, 0.00008994, -0.00001631, -0.00005407, -0.00000460, -0.00000914, -0.00000463, -0.00001219, -0.00000047, 0.00000207, 0.00000033, -0.00000533, -0.00000058, 0.00000974, -0.00000663, -0.00000001, -0.00000442, 0.00000248, 0.00000080, -0.00000627, 0.00000144, -0.00000254, -0.00000508, -0.00000152, -0.00000282, -0.00001607, 0.00000530, 0.00000243, 0.00000175, -0.00000993, 0.00000253, 0.00001418, -0.00000676, -0.00000582, 0.00000239, 0.00000361, -0.00000220, -0.00000073, 0.00000150, 0.00001040, -0.00000606, -0.00000381, -0.00000341, 0.00000077, 0.00000542, 0.00000773, -0.00000146, -0.00001645, 0.00000628, 0.00000914, -0.00000125, -0.00000100, -0.00000135, -0.00000055, -0.00000115, -0.00000266, 0.00000180, 0.00001267, -0.00000345, -0.00001532, 0.00000529, 0.00000616, 0.00000527, -0.00000261, -0.00000163, -0.00000717, -0.00000032, 0.00000980, -0.00000039, -0.00000160, -0.00000050, 0.00000447, -0.00000056, -0.00000852, 0.00000071, 0.00001538, -0.00000663, -0.00000504, 0.00000035, 0.00000726, -0.00000357, -0.00001413, 0.00000150, -0.00000501, -0.00000373, -0.00000176, 0.00000191, -0.00000733, 0.00000369, -0.00000084, -0.00000202, -0.00000331, -0.00000331, 0.00001123, -0.00000539, -0.00000266, 0.00000021, 0.00000379, -0.00000621, -0.00000940, -0.00000661, -0.00000552, -0.00001796, -0.00005717, 0.00003650, 0.00007341, 0.00003928, 0.00008431, -0.00000903, -0.00005588, -0.00001164, -0.00000929, -0.00000546, -0.00000455, -0.00000340, -0.00000485, },
		},
		{ { -0.00000302, -0.00000364, 0.00000832, -0.00000562, -0.00000196, -0.00000290, -0.00000340, -0.00000080, -0.00000266, -0.00000655, -0.00001330, -0.00001194, -0.00005983, 0.00001057, 0.00008605, 0.00003656, 0.00008162, 0.00001219, -0.00005587, -0.00000935, -0.00001192, -0.00000462, -0.00000733, -0.00000408, -0.00000235, 0.00000015, -0.00000262, 0.00000050, -0.00000213, -0.00000084, -0.00000699, -0.00000020, 0.00000092, 0.00000190, -0.00000344, -0.00000091, 0.00001501, -0.00000269, -0.00001040, 0.00000010, 0.00000130, 0.00000284, -0.00001268, 0.00000123, 0.00000341, -0.00000545, -0.00000584, 0.00000136, -0.00000629, 0.00000033, -0.00000391, 0.00000023, 0.00000248, -0.00000030, 0.00001653, -0.00000761, -0.00000494, -0.00000079, 0.00001608, 0.00000041, -0.00000913, 0.00000252, 0.00000016, 0.00000010, -0.00000137, 0.00000010, 0.00000016, 0.00000252, -0.00000913, 0.00000041, 0.00001608, -0.00000079, -0.00000494, -0.00000761, 0.00001653, -0.00000030, 0.00000248, 0.00000023, -0.00000391, 0.00000033, -0.00000629, 0.00000136, -0.00000584, -0.00000545, 0.00000341, 0.00000123, -0.00001268, 0.00000284, 0.00000130, 0.00000010, -0.00001040, -0.00000269, 0.00001501, -0.00000091, -0.00000344, 0.00000190, 0.00000092, -0.00000020, -0.00000699, -0.00000084, -0.00000213, 0.00000050, -0.00000262, 0.00000015, -0.00000235, -0.00000408, -0.00000733, -0.00000462, -0.00001192, -0.00000935, -0.00005587, 0.00001219, 0.00008162, 0.00003656, 0.00008605, 0.00001057, -0.00005983, -0.00001194, -0.00001330, -0.00000655, -0.00000266, -0.00000080, -0.00000340, -0.00000290, -0.00000196, -0.00000562, 0.00000832, -0.00000364, },
		{ -0.00000455, 0.00000654, -0.00000128, -0.00000511, -0.00000068, -0.00000508, 0.00000030, -0.00000327, -0.00000370, -0.00001442, -0.00000750, -0.00005157, -0.00001488, 0.00008204, 0.00004096, 0.00006641, 0.00004431, -0.00005322, -0.00001909, -0.00000944, -0.00000668, -0.00000612, -0.00000535, -0.00000269, -0.00000005, -0.00000231, 0.00000032, -0.00000175, -0.00000065, -0.00000635, -0.00000187, 0.00000129, 0.00000150, -0.00000133, -0.00000437, 0.00001378, 0.00000275, -0.00001123, -0.00000223, 0.00000150, 0.00000396, -0.00001070, -0.00000333, 0.00000625, -0.00000478, -0.00000653, 0.00000021, -0.00000359, -0.00000326, -0.00000012, -0.00000389, 0.00000571, -0.00000438, 0.00001645, -0.00000208, -0.00000733, -0.00000236, 0.00001324, 0.00000695, -0.00001134, 0.00000154, 0.00000076, 0.00000029, -0.00000138, -0.00000001, -0.00000030, 0.00000264, -0.00000608, -0.00000493, 0.00001620, 0.00000226, -0.00000349, -0.00000984, 0.00001203, 0.00000660, -0.00000203, 0.00000490, -0.00000721, 0.00000258, -0.00000716, 0.00000087, -0.00000445, -0.00000575, 0.00000052, 0.00000436, -0.00001254, 0.00000032, 0.00000201, 0.00000047, -0.00000707, -0.00000847, 0.00001473, 0.00000288, -0.00000451, 0.00000139, 0.00000096, 0.00000095, -0.00000682, -0.00000166, -0.00000191, 0.00000009, -0.00000232, -0.00000039, -0.00000152, -0.00000335, -0.00000751, -0.00000403, -0.00001217, -0.00000597, -0.00004609, -0.00002343, 0.00008681, 0.00004180, 0.00008050, 0.00004003, -0.00006308, -0.00001980, -0.00001134, -0.00000963, -0.00000213, -0.00000216, -0.00000155, -0.00000508, 0.00000055, -0.00000806, 0.00000742, -0.00000094, -0.00000351, },
		{ 0.00000309, 0.00000340, -0.00000740, 0.00000052, -0.00000585, 0.00000063, -0.00000344, -0.00000167, -0.00001413, -0.00000611, -0.00004084, -0.00003382, 0.00006916, 0.00005092, 0.00004915, 0.00006542, -0.00003714, -0.00003219, -0.00000723, -0.00000870, -0.00000495, -0.00000643, -0.00000300, -0.00000061, -0.00000177, -0.00000015, -0.00000104, -0.00000091, -0.00000510, -0.00000389, 0.00000198, 0.00000041, 0.00000121, -0.00000686, 0.00001119, 0.00000725, -0.00000974, -0.00000546, 0.00000158, 0.00000447, -0.00000767, -0.00000779, 0.00000765, -0.00000297, -0.00000705, -0.00000163, -0.00000072, -0.00000605, 0.00000188, -0.00000543, 0.00000594, -0.00000449, 0.00001197, 0.00000537, -0.00000953, -0.00000301, 0.00000871, 0.00001308, -0.00001157, -0.00000065, 0.00000155, 0.00000050, -0.00000136, -0.00000007, -0.00000062, 0.00000222, -0.00000309, -0.00000819, 0.00001355, 0.00000641, -0.00000301, -0.00000906, 0.00000471, 0.00001354, -0.00000503, 0.00000750, -0.00000769, 0.00000181, -0.00000534, -0.00000152, -0.00000222, -0.00000627, -0.00000139, 0.00000522, -0.00000998, -0.00000369, 0.00000385, -0.00000105, -0.00000221, -0.00001326, 0.00001247, 0.00000675, -0.00000452, 0.00000035, 0.00000108, 0.00000170, -0.00000594, -0.00000303, -0.00000118, -0.00000086, -0.00000141, -0.00000165, -0.00000021, -0.00000337, -0.00000636, -0.00000526, -0.00000973, -0.00000826, -0.00002952, -0.00005264, 0.00007549, 0.00005743, 0.00006783, 0.00006921, -0.00005857, -0.00003098, -0.00000928, -0.00001239, -0.00000209, -0.00000330, -0.00000023, -0.00000631, 0.00000140, -0.00000789, 0.00000395, 0.00000294, -0.00000453, -0.00000409, },
		{ 0.00000672, -0.00000764, 0.00000005, -0.00000536, -0.00000004, -0.00000293, -0.00000080, -0.00001233, -0.00000698, -0.00002981, -0.00004540, 0.00005005, 0.00006134, 0.00003620, 0.00007268, -0.00001158, -0.00004316, -0.00000826, -0.00000922, -0.00000477, -0.00000671, -0.00000368, -0.00000113, -0.00000136, -0.00000063, -0.00000024, -0.00000135, -0.00000355, -0.00000588, 0.00000257, -0.00000074, 0.00000324, -0.00000765, 0.00000727, 0.00001070, -0.00000681, -0.00000812, 0.00000050, 0.00000511, -0.00000465, -0.00001085, 0.00000671, 0.00000023, -0.00000762, -0.00000335, 0.00000101, -0.00000666, 0.00000114, -0.00000407, 0.00000340, -0.00000156, 0.00000522, 0.00001218, -0.00000988, -0.00000379, 0.00000402, 0.00001711, -0.00000897, -0.00000408, 0.00000244, 0.00000065, -0.00000118, -0.00000026, -0.00000068, 0.00000143, -0.00000060, -0.00000935, 0.00000905, 0.00001051, -0.00000264, -0.00000677, -0.00000262, 0.00001745, -0.00000422, 0.00000641, -0.00000446, -0.00000200, -0.00000146, -0.00000491, 0.00000016, -0.00000686, -0.00000223, 0.00000408, -0.00000578, -0.00000823, 0.00000592, -0.00000321, 0.00000222, -0.00001534, 0.00000769, 0.00001073, -0.00000392, -0.00000053, 0.00000078, 0.00000235, -0.00000466, -0.00000456, -0.00000038, -0.00000185, -0.00000041, -0.00000297, 0.00000092, -0.00000355, -0.00000463, -0.00000722, -0.00000626, -0.00001265, -0.00001388, -0.00006708, 0.00004613, 0.00007913, 0.00005375, 0.00009256, -0.00004389, -0.00004488, -0.00000785, -0.00001458, -0.00000262, -0.00000405, 0.00000025, -0.00000627, 0.00000045, -0.00000557, -0.00000086, 0.00000654, -0.00000499, -0.00000331, -0.00000046, },
		{ -0.00000549, -0.00000191, -0.00000391, -0.00000144, -0.00000187, -0.00000091, -0.00000946, -0.00000893, -0.00002021, -0.00004995, 0.00002810, 0.00006802, 0.00003064, 0.00006798, 0.00001661, -0.00004672, -0.00001402, -0.00000812, -0.00000556, -0.00000628, -0.00000460, -0.00000161, -0.00000111, -0.00000112, 0.00000055, -0.00000181, -0.00000199, -0.00000744, 0.00000256, -0.00000128, 0.00000398, -0.00000640, 0.00000237, 0.00001301, -0.00000333, -0.00000923, -0.00000206, 0.00000593, -0.00000205, -0.00001204, 0.00000357, 0.00000415, -0.00000783, -0.00000475, 0.00000123, -0.00000512, -0.00000163, -0.00000118, -0.00000013, 0.00000214, -0.00000092, 0.00001582, -0.00000714, -0.00000544, 0.00000040, 0.00001797, -0.00000353, -0.00000818, 0.00000307, 0.00000079, -0.00000074, -0.00000069, -0.00000037, 0.00000032, 0.00000131, -0.00000894, 0.00000402, 0.00001325, -0.00000137, -0.00000461, -0.00000756, 0.00001654, 0.00000086, 0.00000210, 0.00000126, -0.00000694, 0.00000236, -0.00000743, 0.00000139, -0.00000675, -0.00000271, 0.00000213, -0.00000157, -0.00001171, 0.00000671, -0.00000417, 0.00000418, -0.00001347, 0.00000041, 0.00001460, -0.00000307, -0.00000074, -0.00000017, 0.00000305, -0.00000330, -0.00000571, -0.00000014, -0.00000222, -0.00000007, -0.00000350, 0.00000100, -0.00000296, -0.00000354, -0.00000819, -0.00000416, -0.00001503, -0.00000531, -0.00006373, 0.00000408, 0.00009786, 0.00004578, 0.00010485, -0.00001797, -0.00005996, -0.00000788, -0.00001617, -0.00000359, -0.00000453, 0.00000003, -0.00000527, -0.00000159, -0.00000245, -0.00000518, 0.00000826, -0.00000385, -0.00000320, -0.00000286, 0.00000763, },
		{ -0.00000442, -0.00000228, -0.00000297, -0.00000073, -0.00000153, -0.00000630, -0.00001078, -0.00001309, -0.00004870, 0.00000658, 0.00006856, 0.00003232, 0.00005612, 0.00004055, -0.00004024, -0.00002328, -0.00000691, -0.00000636, -0.00000588, -0.00000522, -0.00000238, -0.00000079, -0.00000177, 0.00000137, -0.00000221, -0.00000064, -0.00000822, 0.00000162, -0.00000083, 0.00000322, -0.00000348, -0.00000266, 0.00001386, 0.00000020, -0.00000862, -0.00000542, 0.00000625, 0.00000041, -0.00001174, -0.00000071, 0.00000745, -0.00000686, -0.00000613, 0.00000040, -0.00000255, -0.00000467, 0.00000117, -0.00000255, 0.00000438, -0.00000419, 0.00001503, -0.00000143, -0.00000773, -0.00000181, 0.00001572, 0.00000368, -0.00001170, 0.00000280, 0.00000110, -0.00000013, -0.00000132, 0.00000025, -0.00000093, 0.00000261, -0.00000757, -0.00000042, 0.00001387, 0.00000122, -0.00000350, -0.00000914, 0.00001120, 0.00000840, -0.00000298, 0.00000662, -0.00001013, 0.00000367, -0.00000734, 0.00000042, -0.00000526, -0.00000369, 0.00000067, 0.00000118, -0.00001279, 0.00000504, -0.00000270, 0.00000268, -0.00000779, -0.00000822, 0.00001747, -0.00000179, -0.00000039, -0.00000144, 0.00000355, -0.00000195, -0.00000621, -0.00000080, -0.00000160, -0.00000076, -0.00000281, -0.00000035, -0.00000120, -0.00000380, -0.00000717, -0.00000482, -0.00001335, -0.00000552, -0.00004643, -0.00003942, 0.00010272, 0.00005013, 0.00010348, 0.00001756, -0.00007323, -0.00001055, -0.00001702, -0.00000493, -0.00000486, -0.00000061, -0.00000393, -0.00000369, -0.00000001, -0.00000748, 0.00000719, -0.00000086, -0.00000401, -0.00000369, 0.00000614, -0.00000167, },
		{ -0.00000131, -0.00000403, 0.00000002, -0.00000211, -0.00000356, -0.00001173, -0.00000870, -0.00004349, -0.00001185, 0.00006237, 0.00003891, 0.00004251, 0.00005578, -0.00002473, -0.00003251, -0.00000784, -0.00000607, -0.00000619, -0.00000513, -0.00000362, -0.00000026, -0.00000263, 0.00000209, -0.00000236, 0.00000020, -0.00000797, -0.00000035, 0.00000054, 0.00000135, 0.00000012, -0.00000667, 0.00001284, 0.00000371, -0.00000686, -0.00000835, 0.00000515, 0.00000317, -0.00001065, -0.00000483, 0.00000888, -0.00000420, -0.00000769, -0.00000084, -0.00000027, -0.00000634, 0.00000147, -0.00000272, 0.00000421, -0.00000404, 0.00001047, 0.00000563, -0.00000945, -0.00000315, 0.00001145, 0.00001083, -0.00001319, 0.00000103, 0.00000178, 0.00000049, -0.00000195, 0.00000091, -0.00000199, 0.00000318, -0.00000565, -0.00000374, 0.00001245, 0.00000469, -0.00000324, -0.00000804, 0.00000366, 0.00001516, -0.00000564, 0.00000877, -0.00000928, 0.00000114, -0.00000420, -0.00000266, -0.00000257, -0.00000533, 0.00000024, 0.00000192, -0.00001123, 0.00000091, 0.00000106, -0.00000157, -0.00000018, -0.00001583, 0.00001778, 0.00000067, -0.00000022, -0.00000218, 0.00000325, -0.00000038, -0.00000615, -0.00000218, -0.00000032, -0.00000210, -0.00000134, -0.00000256, 0.00000112, -0.00000491, -0.00000467, -0.00000752, -0.00000868, -0.00001139, -0.00002423, -0.00007132, 0.00008574, 0.00006807, 0.00009062, 0.00005755, -0.00008006, -0.00001756, -0.00001662, -0.00000698, -0.00000487, -0.00000159, -0.00000267, -0.00000513, 0.00000091, -0.00000726, 0.00000366, 0.00000313, -0.00000508, -0.00000346, 0.00000324, 0.00000242, -0.00000622, },
		{ -0.00000432, 0.00000015, -0.00000234, -0.00000163, -0.00001152, -0.00000660, -0.00003631, -0.00002550, 0.00005042, 0.00004720, 0.00003144, 0.00006107, -0.00000407, -0.00003760, -0.00001231, -0.00000457, -0.00000710, -0.00000455, -0.00000501, 0.00000014, -0.00000332, 0.00000229, -0.00000189, 0.00000024, -0.00000656, -0.00000316, 0.00000236, -0.00000081, 0.00000322, -0.00000865, 0.00000981, 0.00000709, -0.00000466, -0.00000983, 0.00000220, 0.00000620, -0.00000915, -0.00000796, 0.00000799, -0.00000013, -0.00000909, -0.00000219, 0.00000099, -0.00000601, -0.00000038, -0.00000102, 0.00000230, -0.00000165, 0.00000434, 0.00001154, -0.00000913, -0.00000451, 0.00000667, 0.00001604, -0.00001159, -0.00000236, 0.00000274, 0.00000101, -0.00000233, 0.00000124, -0.00000246, 0.00000280, -0.00000331, -0.00000596, 0.00000972, 0.00000802, -0.00000293, -0.00000589, -0.00000319, 0.00001813, -0.00000378, 0.00000651, -0.00000424, -0.00000434, 0.00000055, -0.00000632, 0.00000009, -0.00000687, 0.00000033, 0.00000130, -0.00000801, -0.00000439, 0.00000551, -0.00000632, 0.00000636, -0.00001965, 0.00001403, 0.00000504, -0.00000106, -0.00000156, 0.00000165, 0.00000165, -0.00000587, -0.00000364, 0.00000077, -0.00000310, -0.00000019, -0.00000430, 0.00000260, -0.00000545, -0.00000244, -0.00000993, -0.00000430, -0.00001720, -0.00000710, -0.00008232, 0.00004638, 0.00009374, 0.00007350, 0.00009400, -0.00007504, -0.00003053, -0.00001434, -0.00001035, -0.00000413, -0.00000307, -0.00000155, -0.00000580, 0.00000038, -0.00000522, -0.00000089, 0.00000652, -0.00000526, -0.00000319, 0.00000031, 0.00000531, -0.00000635, -0.00000148, },
		{ -0.00000026, -0.00000220, -0.00000053, -0.00001034, -0.00000595, -0.00002888, -0.00003382, 0.00003477, 0.00005397, 0.00002539, 0.00005789, 0.00001691, -0.00003594, -0.00001967, -0.00000323, -0.00000754, -0.00000440, -0.00000573, -0.00000024, -0.00000321, 0.00000144, -0.00000045, -0.00000072, -0.00000410, -0.00000635, 0.00000396, -0.00000242, 0.00000491, -0.00000824, 0.00000521, 0.00000997, -0.00000240, -0.00000952, -0.00000207, 0.00000877, -0.00000702, -0.00001005, 0.00000533, 0.00000420, -0.00000947, -0.00000383, 0.00000124, -0.00000421, -0.00000320, 0.00000098, 0.00000029, 0.00000101, -0.00000085, 0.00001423, -0.00000594, -0.00000635, 0.00000255, 0.00001816, -0.00000684, -0.00000674, 0.00000353, 0.00000150, -0.00000229, 0.00000099, -0.00000207, 0.00000145, -0.00000071, -0.00000732, 0.00000659, 0.00001023, -0.00000173, -0.00000420, -0.00000725, 0.00001605, 0.00000239, 0.00000112, 0.00000280, -0.00000998, 0.00000414, -0.00000828, 0.00000106, -0.00000711, -0.00000014, 0.00000061, -0.00000477, -0.00000893, 0.00000850, -0.00000897, 0.00000904, -0.00001783, 0.00000586, 0.00001117, -0.00000303, 0.00000057, -0.00000111, 0.00000398, -0.00000558, -0.00000457, 0.00000082, -0.00000280, -0.00000041, -0.00000439, 0.00000202, -0.00000412, -0.00000218, -0.00000977, -0.00000329, -0.00001822, -0.00000120, -0.00007109, -0.00000674, 0.00011510, 0.00006220, 0.00011850, -0.00005410, -0.00004967, -0.00001025, -0.00001526, -0.00000246, -0.00000517, -0.00000043, -0.00000604, -0.00000084, -0.00000276, -0.00000459, 0.00000774, -0.00000365, -0.00000364, -0.00000163, 0.00000617, -0.00000461, -0.00000264, -0.00000400, },
		{ -0.00000182, -0.00000011, -0.00000859, -0.00000603, -0.00002217, -0.00003742, 0.00001810, 0.00005668, 0.00002505, 0.00004911, 0.00003431, -0.00002750, -0.00002718, -0.00000421, -0.00000625, -0.00000556, -0.00000510, -0.00000190, -0.00000191, -0.00000069, 0.00000193, -0.00000243, -0.00000113, -0.00000917, 0.00000460, -0.00000283, 0.00000484, -0.00000581, 0.00000008, 0.00001168, -0.00000005, -0.00000788, -0.00000644, 0.00000973, -0.00000379, -0.00001149, 0.00000196, 0.00000740, -0.00000803, -0.00000598, 0.00000091, -0.00000201, -0.00000544, 0.00000173, -0.00000045, 0.00000241, -0.00000349, 0.00001301, -0.00000041, -0.00000823, -0.00000058, 0.00001719, -0.00000001, -0.00001076, 0.00000340, 0.00000214, -0.00000182, 0.00000011, -0.00000088, -0.00000060, 0.00000181, -0.00000788, 0.00000360, 0.00001092, 0.00000055, -0.00000355, -0.00000806, 0.00000994, 0.00001037, -0.00000422, 0.00000834, -0.00001248, 0.00000407, -0.00000693, -0.00000060, -0.00000544, -0.00000184, 0.00000082, -0.00000273, -0.00001130, 0.00000846, -0.00000792, 0.00000671, -0.00001061, -0.00000522, 0.00001748, -0.00000507, 0.00000323, -0.00000399, 0.00000583, -0.00000502, -0.00000484, -0.00000041, -0.00000105, -0.00000215, -0.00000265, -0.00000070, -0.00000084, -0.00000423, -0.00000651, -0.00000636, -0.00001344, -0.00000619, -0.00004510, -0.00005826, 0.00011839, 0.00006503, 0.00012595, -0.00001696, -0.00007233, -0.00000610, -0.00002086, -0.00000048, -0.00000749, 0.00000060, -0.00000616, -0.00000195, -0.00000116, -0.00000617, 0.00000615, -0.00000025, -0.00000481, -0.00000235, 0.00000512, -0.00000162, -0.00000408, -0.00000352, -0.00000093, },
		{ -0.00000019, -0.00000654, -0.00000644, -0.00001639, -0.00003774, 0.00000303, 0.00005387, 0.00002960, 0.00003792, 0.00004603, -0.00001453, -0.00003146, -0.00000906, -0.00000294, -0.00000798, -0.00000331, -0.00000453, 0.00000020, -0.00000352, 0.00000448, -0.00000405, 0.00000140, -0.00001068, 0.00000366, -0.00000178, 0.00000323, -0.00000229, -0.00000436, 0.00001161, 0.00000252, -0.00000565, -0.00000958, 0.00000826, 0.00000063, -0.00001252, -0.00000122, 0.00000859, -0.00000468, -0.00000840, 0.00000023, -0.00000028, -0.00000611, 0.00000058, 0.00000036, 0.00000232, -0.00000354, 0.00000887, 0.00000570, -0.00000904, -0.00000309, 0.00001405, 0.00000714, -0.00001292, 0.00000175, 0.00000303, -0.00000103, -0.00000116, 0.00000071, -0.00000271, 0.00000364, -0.00000748, 0.00000083, 0.00001038, 0.00000338, -0.00000352, -0.00000668, 0.00000240, 0.00001664, -0.00000616, 0.00000955, -0.00000995, -0.00000032, -0.00000251, -0.00000425, -0.00000248, -0.00000440, 0.00000183, -0.00000204, -0.00001128, 0.00000532, -0.00000345, 0.00000052, -0.00000054, -0.00001589, 0.00002129, -0.00000526, 0.00000468, -0.00000537, 0.00000604, -0.00000357, -0.00000489, -0.00000230, 0.00000131, -0.00000439, -0.00000021, -0.00000418, 0.00000292, -0.00000708, -0.00000189, -0.00001127, -0.00000595, -0.00001610, -0.00001706, -0.00009211, 0.00009454, 0.00008377, 0.00011759, 0.00003119, -0.00009231, -0.00000555, -0.00002493, 0.00000015, -0.00000890, 0.00000090, -0.00000601, -0.00000271, -0.00000080, -0.00000553, 0.00000244, 0.00000380, -0.00000587, -0.00000247, 0.00000302, 0.00000148, -0.00000494, -0.00000333, -0.00000164, -0.00000133, },
		{ -0.00000432, -0.00000713, -0.00001129, -0.00003631, -0.00000862, 0.00004560, 0.00003700, 0.00002732, 0.00005153, -0.00000002, -0.00003035, -0.00001723, 0.00000090, -0.00001031, -0.00000159, -0.00000693, 0.00000191, -0.00000576, 0.00000588, -0.00000438, 0.00000238, -0.00001009, 0.00000083, 0.00000055, 0.00000070, 0.00000126, -0.00000717, 0.00000963, 0.00000520, -0.00000341, -0.00001078, 0.00000447, 0.00000555, -0.00001277, -0.00000404, 0.00000792, -0.00000029, -0.00001018, -0.00000102, 0.00000080, -0.00000525, -0.00000184, 0.00000178, 0.00000160, -0.00000218, 0.00000386, 0.00001024, -0.00000769, -0.00000545, 0.00000994, 0.00001285, -0.00001222, -0.00000146, 0.00000400, 0.00000005, -0.00000249, 0.00000209, -0.00000413, 0.00000417, -0.00000588, -0.00000187, 0.00000922, 0.00000587, -0.00000325, -0.00000477, -0.00000373, 0.00001846, -0.00000305, 0.00000593, -0.00000315, -0.00000715, 0.00000269, -0.00000775, 0.00000002, -0.00000647, 0.00000256, -0.00000176, -0.00000983, 0.00000042, 0.00000251, -0.00000661, 0.00000846, -0.00002233, 0.00001995, -0.00000190, 0.00000343, -0.00000400, 0.00000383, -0.00000076, -0.00000531, -0.00000382, 0.00000285, -0.00000551, 0.00000110, -0.00000633, 0.00000489, -0.00000830, 0.00000115, -0.00001429, -0.00000074, -0.00002307, 0.00000101, -0.00009852, 0.00004449, 0.00011105, 0.00010154, 0.00008044, -0.00010112, -0.00001311, -0.00002473, -0.00000263, -0.00000797, -0.00000045, -0.00000502, -0.00000355, -0.00000117, -0.00000375, -0.00000168, 0.00000675, -0.00000561, -0.00000285, 0.00000100, 0.00000361, -0.00000464, -0.00000357, -0.00000232, -0.00000078, -0.00000065, },
		{ -0.00000800, -0.00000670, -0.00003408, -0.00001643, 0.00003354, 0.00004435, 0.00001991, 0.00005113, 0.00001370, -0.00002404, -0.00002586, 0.00000244, -0.00001042, -0.00000167, -0.00000754, 0.00000181, -0.00000603, 0.00000492, -0.00000251, 0.00000117, -0.00000716, -0.00000356, 0.00000353, -0.00000189, 0.00000395, -0.00000800, 0.00000625, 0.00000745, -0.00000124, -0.00001016, -0.00000056, 0.00000954, -0.00001142, -0.00000687, 0.00000624, 0.00000369, -0.00001023, -0.00000322, 0.00000146, -0.00000366, -0.00000420, 0.00000242, 0.00000145, -0.00000082, -0.00000012, 0.00001178, -0.00000396, -0.00000768, 0.00000574, 0.00001612, -0.00000864, -0.00000553, 0.00000452, 0.00000141, -0.00000354, 0.00000275, -0.00000431, 0.00000312, -0.00000316, -0.00000453, 0.00000792, 0.00000743, -0.00000209, -0.00000359, -0.00000682, 0.00001520, 0.00000413, -0.00000035, 0.00000474, -0.00001283, 0.00000544, -0.00000863, 0.00000024, -0.00000667, 0.00000184, -0.00000079, -0.00000828, -0.00000440, 0.00000742, -0.00001145, 0.00001288, -0.00002194, 0.00001245, 0.00000524, -0.00000065, 0.00000008, -0.00000046, 0.00000309, -0.00000627, -0.00000423, 0.00000248, -0.00000433, 0.00000001, -0.00000571, 0.00000354, -0.00000621, 0.00000045, -0.00001265, -0.00000137, -0.00002204, 0.00000346, -0.00007852, -0.00001953, 0.00013218, 0.00008964, 0.00011950, -0.00009109, -0.00003170, -0.00001877, -0.00000991, -0.00000407, -0.00000387, -0.00000277, -0.00000502, -0.00000134, -0.00000232, -0.00000436, 0.00000714, -0.00000332, -0.00000404, -0.00000024, 0.00000422, -0.00000316, -0.00000406, -0.00000304, -0.00000021, -0.00000137, -0.00000209, },
		{ -0.00000293, -0.00003105, -0.00002138, 0.00002048, 0.00004850, 0.00001758, 0.00004550, 0.00002565, -0.00001484, -0.00003108, -0.00000097, -0.00000690, -0.00000450, -0.00000560, -0.00000075, -0.00000379, 0.00000136, 0.00000145, -0.00000187, -0.00000265, -0.00000844, 0.00000606, -0.00000364, 0.00000523, -0.00000704, 0.00000236, 0.00000864, 0.00000101, -0.00000840, -0.00000530, 0.00001119, -0.00000786, -0.00001006, 0.00000446, 0.00000609, -0.00000802, -0.00000628, 0.00000187, -0.00000212, -0.00000540, 0.00000146, 0.00000235, -0.00000014, -0.00000223, 0.00001026, 0.00000117, -0.00000907, 0.00000169, 0.00001687, -0.00000314, -0.00000919, 0.00000392, 0.00000307, -0.00000402, 0.00000238, -0.00000314, 0.00000075, 0.00000013, -0.00000676, 0.00000649, 0.00000801, -0.00000010, -0.00000339, -0.00000690, 0.00000850, 0.00001227, -0.00000553, 0.00000981, -0.00001398, 0.00000353, -0.00000570, -0.00000236, -0.00000465, -0.00000063, 0.00000124, -0.00000732, -0.00000785, 0.00000950, -0.00001203, 0.00001123, -0.00001475, 0.00000033, 0.00001422, -0.00000587, 0.00000518, -0.00000515, 0.00000666, -0.00000712, -0.00000366, 0.00000025, -0.00000100, -0.00000325, -0.00000255, -0.00000075, -0.00000120, -0.00000376, -0.00000656, -0.00000757, -0.00001347, -0.00000690, -0.00004349, -0.00007814, 0.00013123, 0.00009166, 0.00014063, -0.00005912, -0.00006005, -0.00000864, -0.00002049, 0.00000171, -0.00000857, 0.00000030, -0.00000714, -0.00000080, -0.00000218, -0.00000468, 0.00000470, 0.00000064, -0.00000571, -0.00000072, 0.00000353, -0.00000103, -0.00000437, -0.00000388, 0.00000026, -0.00000209, -0.00000014, -0.00000875, },
		{ -0.00002657, -0.00002517, 0.00000931, 0.00004716, 0.00002094, 0.00003590, 0.00003557, -0.00000563, -0.00003006, -0.00000985, -0.00000044, -0.00000922, -0.00000194, -0.00000492, 0.00000005, -0.00000363, 0.00000609, -0.00000522, 0.00000178, -0.00001215, 0.00000686, -0.00000377, 0.00000487, -0.00000481, -0.00000116, 0.00000839, 0.00000336, -0.00000618, -0.00000854, 0.00000984, -0.00000235, -0.00001326, 0.00000281, 0.00000667, -0.00000409, -0.00000936, 0.00000167, -0.00000088, -0.00000523, -0.00000083, 0.00000372, 0.00000023, -0.00000282, 0.00000693, 0.00000592, -0.00000866, -0.00000230, 0.00001567, 0.00000284, -0.00001118, 0.00000184, 0.00000483, -0.00000373, 0.00000101, -0.00000099, -0.00000220, 0.00000307, -0.00000792, 0.00000461, 0.00000808, 0.00000212, -0.00000359, -0.00000531, 0.00000121, 0.00001774, -0.00000637, 0.00000962, -0.00000944, -0.00000277, -0.00000012, -0.00000643, -0.00000167, -0.00000379, 0.00000347, -0.00000656, -0.00000979, 0.00000847, -0.00000845, 0.00000469, -0.00000354, -0.00001258, 0.00002146, -0.00000924, 0.00000859, -0.00000783, 0.00000815, -0.00000667, -0.00000308, -0.00000267, 0.00000290, -0.00000686, 0.00000118, -0.00000572, 0.00000428, -0.00000883, 0.00000081, -0.00001532, -0.00000270, -0.00002105, -0.00000977, -0.00011280, 0.00009890, 0.00010940, 0.00014356, -0.00000919, -0.00009143, 0.00000052, -0.00003054, 0.00000633, -0.00001236, 0.00000274, -0.00000900, -0.00000004, -0.00000307, -0.00000316, 0.00000058, 0.00000481, -0.00000679, -0.00000112, 0.00000229, 0.00000094, -0.00000408, -0.00000479, 0.00000039, -0.00000243, 0.00000109, -0.00000878, -0.00000068, },
		{ -0.00002911, 0.00000174, 0.00003991, 0.00002870, 0.00002455, 0.00004296, 0.00000200, -0.00002282, -0.00002163, 0.00000587, -0.00001319, 0.00000117, -0.00000859, 0.00000336, -0.00000778, 0.00000913, -0.00000680, 0.00000418, -0.00001308, 0.00000497, -0.00000191, 0.00000300, -0.00000189, -0.00000382, 0.00000690, 0.00000542, -0.00000376, -0.00000992, 0.00000608, 0.00000376, -0.00001533, 0.00000071, 0.00000625, 0.00000009, -0.00001105, 0.00000021, 0.00000034, -0.00000430, -0.00000326, 0.00000432, 0.00000116, -0.00000288, 0.00000347, 0.00000865, -0.00000594, -0.00000613, 0.00001315, 0.00000804, -0.00001086, -0.00000144, 0.00000617, -0.00000246, -0.00000111, 0.00000149, -0.00000474, 0.00000472, -0.00000747, 0.00000199, 0.00000812, 0.00000387, -0.00000339, -0.00000365, -0.00000408, 0.00001828, -0.00000190, 0.00000455, -0.00000108, -0.00001049, 0.00000500, -0.00000928, 0.00000011, -0.00000583, 0.00000440, -0.00000497, -0.00001091, 0.00000533, -0.00000255, -0.00000366, 0.00000732, -0.00002171, 0.00002336, -0.00000808, 0.00000806, -0.00000669, 0.00000630, -0.00000401, -0.00000356, -0.00000464, 0.00000527, -0.00000848, 0.00000290, -0.00000848, 0.00000706, -0.00001124, 0.00000522, -0.00001938, 0.00000355, -0.00002919, 0.00000904, -0.00011398, 0.00003828, 0.00013417, 0.00013607, 0.00004833, -0.00011514, 0.00000147, -0.00003507, 0.00000603, -0.00001259, 0.00000265, -0.00000916, -0.00000021, -0.00000382, -0.00000137, -0.00000314, 0.00000726, -0.00000608, -0.00000219, 0.00000131, 0.00000207, -0.00000305, -0.00000552, -0.00000012, -0.00000210, 0.00000124, -0.00000760, -0.00000058, -0.00002009, },
		{ -0.00000247, 0.00002866, 0.00003767, 0.00001472, 0.00004646, 0.00000847, -0.00001233, -0.00003146, 0.00000783, -0.00001333, 0.00000125, -0.00000950, 0.00000398, -0.00000892, 0.00000864, -0.00000505, 0.00000331, -0.00001050, 0.00000032, 0.00000161, 0.00000015, 0.00000118, -0.00000551, 0.00000485, 0.00000653, -0.00000104, -0.00000986, 0.00000135, 0.00000857, -0.00001491, -0.00000266, 0.00000596, 0.00000296, -0.00001023, -0.00000288, 0.00000180, -0.00000341, -0.00000463, 0.00000318, 0.00000314, -0.00000304, 0.00000096, 0.00000874, -0.00000145, -0.00000913, 0.00000961, 0.00001184, -0.00000845, -0.00000500, 0.00000639, -0.00000015, -0.00000353, 0.00000352, -0.00000606, 0.00000454, -0.00000532, -0.00000125, 0.00000824, 0.00000491, -0.00000241, -0.00000287, -0.00000623, 0.00001398, 0.00000608, -0.00000229, 0.00000706, -0.00001544, 0.00000622, -0.00000849, -0.00000100, -0.00000540, 0.00000300, -0.00000190, -0.00001184, 0.00000140, 0.00000330, -0.00001038, 0.00001389, -0.00002382, 0.00001816, -0.00000160, 0.00000317, -0.00000168, 0.00000140, 0.00000060, -0.00000536, -0.00000471, 0.00000473, -0.00000666, 0.00000111, -0.00000734, 0.00000533, -0.00000891, 0.00000405, -0.00001661, 0.00000142, -0.00002620, 0.00000818, -0.00008551, -0.00003480, 0.00014962, 0.00013009, 0.00010071, -0.00012034, -0.00001168, -0.00003054, -0.00000155, -0.00000782, -0.00000104, -0.00000668, -0.00000231, -0.00000320, -0.00000094, -0.00000468, 0.00000669, -0.00000307, -0.00000416, 0.00000094, 0.00000218, -0.00000155, -0.00000573, -0.00000140, -0.00000105, 0.00000031, -0.00000515, -0.00000259, -0.00001195, -0.00003321, },
		{ 0.00001693, 0.00004371, 0.00000987, 0.00004443, 0.00001554, -0.00000290, -0.00003454, 0.00000251, -0.00000815, -0.00000273, -0.00000675, 0.00000103, -0.00000627, 0.00000421, -0.00000003, -0.00000050, -0.00000516, -0.00000594, 0.00000558, -0.00000271, 0.00000376, -0.00000630, 0.00000293, 0.00000622, 0.00000205, -0.00000899, -0.00000280, 0.00001055, -0.00001130, -0.00000750, 0.00000635, 0.00000382, -0.00000689, -0.00000701, 0.00000318, -0.00000283, -0.00000456, 0.00000043, 0.00000563, -0.00000305, -0.00000055, 0.00000688, 0.00000329, -0.00001023, 0.00000505, 0.00001426, -0.00000485, -0.00000770, 0.00000496, 0.00000299, -0.00000564, 0.00000445, -0.00000572, 0.00000258, -0.00000203, -0.00000447, 0.00000807, 0.00000548, -0.00000087, -0.00000295, -0.00000576, 0.00000701, 0.00001394, -0.00000680, 0.00001095, -0.00001453, 0.00000198, -0.00000365, -0.00000488, -0.00000279, -0.00000036, 0.00000210, -0.00001239, -0.00000254, 0.00000738, -0.00001332, 0.00001432, -0.00001850, 0.00000703, 0.00000832, -0.00000409, 0.00000508, -0.00000449, 0.00000552, -0.00000757, -0.00000326, 0.00000157, -0.00000188, -0.00000356, -0.00000294, -0.00000015, -0.00000264, -0.00000194, -0.00000793, -0.00000773, -0.00001414, -0.00000694, -0.00004258, -0.00009771, 0.00013927, 0.00013495, 0.00013853, -0.00010091, -0.00003985, -0.00001756, -0.00001542, 0.00000078, -0.00000742, -0.00000205, -0.00000621, -0.00000097, -0.00000245, -0.00000355, 0.00000326, 0.00000147, -0.00000636, 0.00000085, 0.00000166, -0.00000008, -0.00000517, -0.00000328, 0.00000034, -0.00000124, -0.00000201, -0.00000579, -0.00000360, -0.00003605, -0.00000535, },
		{ 0.00004350, 0.00001205, 0.00003628, 0.00002445, 0.00000234, -0.00002880, -0.00000965, 0.00000080, -0.00000921, -0.00000167, -0.00000426, -0.00000113, -0.00000251, 0.00000634, -0.00000525, 0.00000078, -0.00001161, 0.00000825, -0.00000435, 0.00000517, -0.00000616, 0.00000141, 0.00000467, 0.00000509, -0.00000763, -0.00000552, 0.00000941, -0.00000520, -0.00001283, 0.00000681, 0.00000332, -0.00000238, -0.00001062, 0.00000347, -0.00000210, -0.00000366, -0.00000271, 0.00000723, -0.00000200, -0.00000180, 0.00000449, 0.00000653, -0.00000859, -0.00000038, 0.00001545, -0.00000101, -0.00000872, 0.00000195, 0.00000625, -0.00000668, 0.00000386, -0.00000379, -0.00000053, 0.00000141, -0.00000679, 0.00000703, 0.00000610, 0.00000065, -0.00000324, -0.00000415, 0.00000029, 0.00001825, -0.00000613, 0.00000886, -0.00000766, -0.00000625, 0.00000300, -0.00000928, 0.00000006, -0.00000382, 0.00000533, -0.00001155, -0.00000652, 0.00000922, -0.00001221, 0.00000942, -0.00000824, -0.00000623, 0.00001778, -0.00001013, 0.00001029, -0.00000843, 0.00000843, -0.00000848, -0.00000175, -0.00000245, 0.00000358, -0.00000857, 0.00000196, -0.00000633, 0.00000427, -0.00000906, 0.00000212, -0.00001822, -0.00000042, -0.00002461, -0.00000444, -0.00013072, 0.00009531, 0.00015121, 0.00016003, -0.00005845, -0.00007721, -0.00000179, -0.00003082, 0.00000923, -0.00001355, 0.00000258, -0.00001033, 0.00000173, -0.00000496, -0.00000088, -0.00000135, 0.00000574, -0.00000750, 0.00000029, 0.00000116, 0.00000082, -0.00000380, -0.00000528, 0.00000142, -0.00000258, 0.00000079, -0.00000865, 0.00000288, -0.00003557, -0.00000951, 0.00000831, },
		{ 0.00002057, 0.00002377, 0.00003439, 0.00000332, -0.00001632, -0.00002431, 0.00000911, -0.00001457, 0.00000269, -0.00000909, 0.00000369, -0.00000850, 0.00001105, -0.00000814, 0.00000468, -0.00001440, 0.00000796, -0.00000377, 0.00000486, -0.00000491, 0.00000007, 0.00000271, 0.00000718, -0.00000552, -0.00000694, 0.00000617, 0.00000147, -0.00001671, 0.00000591, 0.00000300, 0.00000124, -0.00001185, 0.00000160, -0.00000052, -0.00000301, -0.00000470, 0.00000661, 0.00000081, -0.00000343, 0.00000278, 0.00000721, -0.00000436, -0.00000591, 0.00001531, 0.00000257, -0.00000804, -0.00000183, 0.00000859, -0.00000599, 0.00000164, -0.00000075, -0.00000385, 0.00000396, -0.00000748, 0.00000482, 0.00000704, 0.00000176, -0.00000313, -0.00000278, -0.00000401, 0.00001740, -0.00000022, 0.00000230, 0.00000204, -0.00001437, 0.00000753, -0.00001102, 0.00000061, -0.00000522, 0.00000595, -0.00000829, -0.00001078, 0.00000915, -0.00000829, 0.00000183, 0.00000282, -0.00001696, 0.00002268, -0.00001161, 0.00001109, -0.00000815, 0.00000765, -0.00000676, -0.00000171, -0.00000516, 0.00000699, -0.00001090, 0.00000420, -0.00000970, 0.00000795, -0.00001290, 0.00000821, -0.00002358, 0.00000687, -0.00003364, 0.00001460, -0.00012579, 0.00002413, 0.00016870, 0.00017131, -0.00000171, -0.00011255, 0.00000791, -0.00004105, 0.00001240, -0.00001573, 0.00000444, -0.00001238, 0.00000291, -0.00000660, 0.00000121, -0.00000479, 0.00000769, -0.00000639, -0.00000136, 0.00000119, 0.00000089, -0.00000195, -0.00000674, 0.00000151, -0.00000297, 0.00000226, -0.00000974, 0.00000571, -0.00003034, -0.00001655, 0.00000465, 0.00003619, },
		{ 0.00001103, 0.00004226, 0.00000307, -0.00000243, -0.00003503, 0.00001159, -0.00001506, 0.00000328, -0.00001066, 0.00000538, -0.00001089, 0.00001151, -0.00000708, 0.00000473, -0.00001300, 0.00000413, -0.00000079, 0.00000278, -0.00000252, -0.00000154, 0.00000139, 0.00000737, -0.00000229, -0.00000775, 0.00000253, 0.00000645, -0.00001729, 0.00000236, 0.00000425, 0.00000243, -0.00000973, -0.00000262, 0.00000203, -0.00000323, -0.00000469, 0.00000344, 0.00000500, -0.00000518, 0.00000194, 0.00000564, 0.00000109, -0.00001009, 0.00001333, 0.00000595, -0.00000639, -0.00000508, 0.00000894, -0.00000322, -0.00000184, 0.00000258, -0.00000643, 0.00000491, -0.00000637, 0.00000166, 0.00000808, 0.00000250, -0.00000249, -0.00000227, -0.00000534, 0.00001227, 0.00000827, -0.00000473, 0.00000978, -0.00001782, 0.00000658, -0.00000802, -0.00000242, -0.00000355, 0.00000338, -0.00000274, -0.00001492, 0.00000749, -0.00000317, -0.00000550, 0.00001090, -0.00002173, 0.00002077, -0.00000722, 0.00000678, -0.00000343, 0.00000330, -0.00000245, -0.00000362, -0.00000537, 0.00000676, -0.00000895, 0.00000214, -0.00000846, 0.00000641, -0.00001110, 0.00000740, -0.00002040, 0.00000386, -0.00002928, 0.00001120, -0.00009006, -0.00005489, 0.00017070, 0.00018190, 0.00005774, -0.00013364, 0.00000344, -0.00004080, 0.00000671, -0.00001171, 0.00000179, -0.00001081, 0.00000109, -0.00000579, 0.00000096, -0.00000533, 0.00000625, -0.00000282, -0.00000398, 0.00000170, 0.00000037, -0.00000018, -0.00000710, 0.00000028, -0.00000214, 0.00000200, -0.00000843, 0.00000439, -0.00002062, -0.00002590, 0.00000505, 0.00002407, 0.00003179, },
		{ 0.00004418, 0.00000567, 0.00000703, -0.00003650, 0.00000521, -0.00000917, -0.00000092, -0.00000794, 0.00000283, -0.00000864, 0.00000711, -0.00000191, 0.00000108, -0.00000795, -0.00000230, 0.00000361, -0.00000032, 0.00000062, -0.00000356, 0.00000131, 0.00000536, 0.00000197, -0.00000841, -0.00000013, 0.00000828, -0.00001388, -0.00000385, 0.00000715, 0.00000114, -0.00000496, -0.00000794, 0.00000457, -0.00000387, -0.00000315, -0.00000109, 0.00000897, -0.00000587, 0.00000116, 0.00000324, 0.00000566, -0.00001133, 0.00000905, 0.00000942, -0.00000470, -0.00000667, 0.00000682, 0.00000128, -0.00000571, 0.00000522, -0.00000751, 0.00000404, -0.00000389, -0.00000172, 0.00000862, 0.00000333, -0.00000166, -0.00000236, -0.00000460, 0.00000544, 0.00001541, -0.00000803, 0.00001178, -0.00001419, -0.00000044, -0.00000094, -0.00000792, 0.00000002, -0.00000105, 0.00000340, -0.00001752, 0.00000405, 0.00000204, -0.00001063, 0.00001409, -0.00001972, 0.00001270, 0.00000153, -0.00000076, 0.00000359, -0.00000254, 0.00000275, -0.00000648, -0.00000357, 0.00000331, -0.00000345, -0.00000325, -0.00000361, 0.00000074, -0.00000473, 0.00000081, -0.00001031, -0.00000706, -0.00001512, -0.00000676, -0.00004202, -0.00011719, 0.00014266, 0.00019763, 0.00011121, -0.00013231, -0.00001836, -0.00002936, -0.00000744, -0.00000241, -0.00000465, -0.00000596, -0.00000362, -0.00000245, -0.00000186, -0.00000294, 0.00000203, 0.00000208, -0.00000655, 0.00000204, -0.00000015, 0.00000089, -0.00000609, -0.00000207, -0.00000043, 0.00000039, -0.00000518, -0.00000003, -0.00000862, -0.00003492, 0.00000635, 0.00001162, 0.00004044, 0.00000308, },
		{ 0.00001356, 0.00000893, -0.00002741, -0.00000872, 0.00000090, -0.00000799, -0.00000248, -0.00000263, -0.00000315, -0.00000033, 0.00000525, -0.00000414, -0.00000154, -0.00000899, 0.00000748, -0.00000306, 0.00000355, -0.00000552, 0.00000218, 0.00000193, 0.00000632, -0.00000868, -0.00000153, 0.00000713, -0.00000760, -0.00001110, 0.00001016, -0.00000095, 0.00000020, -0.00001201, 0.00000539, -0.00000373, -0.00000157, -0.00000489, 0.00001062, -0.00000413, -0.00000060, 0.00000162, 0.00000755, -0.00000884, 0.00000272, 0.00001288, -0.00000348, -0.00000628, 0.00000272, 0.00000632, -0.00000868, 0.00000621, -0.00000671, 0.00000165, -0.00000083, -0.00000442, 0.00000805, 0.00000463, -0.00000102, -0.00000250, -0.00000324, -0.00000033, 0.00001816, -0.00000545, 0.00000733, -0.00000474, -0.00001059, 0.00000668, -0.00001262, 0.00000267, -0.00000463, 0.00000745, -0.00001677, -0.00000156, 0.00000681, -0.00001301, 0.00001264, -0.00001275, 0.00000162, 0.00001102, -0.00000792, 0.00000947, -0.00000682, 0.00000639, -0.00000836, -0.00000149, -0.00000122, 0.00000287, -0.00000894, 0.00000164, -0.00000560, 0.00000245, -0.00000721, 0.00000127, -0.00001905, -0.00000007, -0.00002573, -0.00000251, -0.00014396, 0.00008112, 0.00021519, 0.00015660, -0.00010741, -0.00005363, -0.00001172, -0.00002527, 0.00000796, -0.00001171, -0.00000021, -0.00000926, 0.00000177, -0.00000574, 0.00000066, -0.00000284, 0.00000617, -0.00000760, 0.00000135, -0.00000003, 0.00000091, -0.00000394, -0.00000482, 0.00000132, -0.00000160, -0.00000132, -0.00000537, 0.00000235, -0.00004003, 0.00000470, 0.00000338, 0.00004205, 0.00000342, 0.00003797, },
		{ 0.00000460, -0.00001149, -0.00002468, 0.00000998, -0.00001380, 0.00000236, -0.00000795, 0.00000247, -0.00000744, 0.00001102, -0.00000782, 0.00000330, -0.00001329, 0.00000882, -0.00000398, 0.00000515, -0.00000652, 0.00000294, -0.00000135, 0.00000918, -0.00000767, -0.00000240, 0.00000450, -0.00000090, -0.00001671, 0.00001095, -0.00000148, 0.00000313, -0.00001260, 0.00000307, -0.00000174, -0.00000138, -0.00000617, 0.00000862, 0.00000046, -0.00000361, 0.00000152, 0.00000630, -0.00000332, -0.00000432, 0.00001547, -0.00000238, -0.00000463, -0.00000189, 0.00001021, -0.00000942, 0.00000488, -0.00000405, -0.00000166, 0.00000195, -0.00000584, 0.00000621, 0.00000636, -0.00000060, -0.00000238, -0.00000231, -0.00000346, 0.00001579, 0.00000199, -0.00000072, 0.00000604, -0.00001867, 0.00001021, -0.00001295, 0.00000166, -0.00000490, 0.00000737, -0.00001173, -0.00000911, 0.00001067, -0.00001286, 0.00000815, -0.00000394, -0.00000853, 0.00001745, -0.00001133, 0.00001132, -0.00000724, 0.00000669, -0.00000774, -0.00000089, -0.00000443, 0.00000701, -0.00001166, 0.00000394, -0.00000898, 0.00000642, -0.00001193, 0.00000851, -0.00002506, 0.00000734, -0.00003430, 0.00001501, -0.00013069, -0.00000221, 0.00022188, 0.00019768, -0.00006397, -0.00009298, 0.00000336, -0.00003961, 0.00001378, -0.00001534, 0.00000325, -0.00001309, 0.00000437, -0.00000831, 0.00000303, -0.00000587, 0.00000748, -0.00000606, -0.00000078, 0.00000091, 0.00000000, -0.00000134, -0.00000702, 0.00000214, -0.00000289, 0.00000166, -0.00000930, 0.00000926, -0.00003855, -0.00000225, 0.00000173, 0.00003516, 0.00001204, 0.00002490, 0.00002558, },
		{ 0.00000434, -0.00003541, 0.00001255, -0.00001447, 0.00000344, -0.00001016, 0.00000517, -0.00001104, 0.00001253, -0.00000758, 0.00000447, -0.00001353, 0.00000655, -0.00000237, 0.00000472, -0.00000577, 0.00000247, -0.00000289, 0.00000919, -0.00000459, -0.00000375, 0.00000231, 0.00000372, -0.00001837, 0.00000776, 0.00000114, 0.00000236, -0.00000898, -0.00000229, 0.00000195, -0.00000287, -0.00000459, 0.00000344, 0.00000666, -0.00000677, 0.00000229, 0.00000304, 0.00000309, -0.00000983, 0.00001580, -0.00000046, -0.00000310, -0.00000522, 0.00001136, -0.00000710, 0.00000123, -0.00000012, -0.00000507, 0.00000380, -0.00000582, 0.00000351, 0.00000795, 0.00000001, -0.00000216, -0.00000197, -0.00000404, 0.00001001, 0.00001072, -0.00000762, 0.00001284, -0.00002000, 0.00000661, -0.00000739, -0.00000369, -0.00000150, 0.00000323, -0.00000348, -0.00001690, 0.00001254, -0.00001044, 0.00000231, 0.00000392, -0.00001486, 0.00001867, -0.00000961, 0.00000838, -0.00000359, 0.00000372, -0.00000459, -0.00000232, -0.00000509, 0.00000736, -0.00000991, 0.00000187, -0.00000785, 0.00000539, -0.00001112, 0.00000858, -0.00002203, 0.00000387, -0.00002896, 0.00000969, -0.00008882, -0.00008394, 0.00020167, 0.00023878, -0.00000898, -0.00012563, 0.00000745, -0.00004453, 0.00001105, -0.00001299, 0.00000243, -0.00001319, 0.00000355, -0.00000784, 0.00000244, -0.00000559, 0.00000529, -0.00000211, -0.00000383, 0.00000219, -0.00000118, 0.00000079, -0.00000778, 0.00000144, -0.00000283, 0.00000278, -0.00001034, 0.00001061, -0.00003009, -0.00001406, 0.00000561, 0.00002215, 0.00002506, 0.00000974, 0.00003719, -0.00000099, },
		{ -0.00003565, 0.00000589, -0.00000877, -0.00000014, -0.00000824, 0.00000371, -0.00000989, 0.00000900, -0.00000314, 0.00000188, -0.00000990, 0.00000126, 0.00000120, 0.00000253, -0.00000321, 0.00000030, -0.00000197, 0.00000606, 0.00000043, -0.00000592, 0.00000164, 0.00000501, -0.00001547, 0.00000072, 0.00000651, -0.00000137, -0.00000267, -0.00000872, 0.00000567, -0.00000485, -0.00000154, -0.00000273, 0.00001187, -0.00000806, 0.00000236, -0.00000010, 0.00000772, -0.00001170, 0.00001279, 0.00000308, -0.00000288, -0.00000600, 0.00000917, -0.00000199, -0.00000388, 0.00000401, -0.00000766, 0.00000432, -0.00000466, 0.00000068, 0.00000867, 0.00000134, -0.00000223, -0.00000186, -0.00000328, 0.00000367, 0.00001673, -0.00000926, 0.00001238, -0.00001318, -0.00000343, 0.00000206, -0.00001105, 0.00000334, -0.00000246, 0.00000488, -0.00002202, 0.00001083, -0.00000569, -0.00000383, 0.00000942, -0.00001661, 0.00001499, -0.00000388, 0.00000234, 0.00000213, -0.00000060, -0.00000045, -0.00000479, -0.00000375, 0.00000455, -0.00000473, -0.00000322, -0.00000361, 0.00000080, -0.00000613, 0.00000302, -0.00001224, -0.00000702, -0.00001482, -0.00000828, -0.00003968, -0.00013908, 0.00014489, 0.00027821, 0.00005336, -0.00014381, -0.00000361, -0.00003854, -0.00000026, -0.00000540, -0.00000216, -0.00000976, -0.00000060, -0.00000440, -0.00000097, -0.00000249, 0.00000075, 0.00000266, -0.00000641, 0.00000279, -0.00000172, 0.00000164, -0.00000672, -0.00000078, -0.00000149, 0.00000200, -0.00000839, 0.00000697, -0.00001682, -0.00002748, 0.00001115, 0.00000814, 0.00003642, -0.00000127, 0.00004286, -0.00000207, 0.00001346, },
		{ -0.00000799, 0.00000071, -0.00000628, -0.00000372, -0.00000072, -0.00000527, 0.00000214, 0.00000350, -0.00000253, -0.00000443, -0.00000502, 0.00000507, -0.00000026, 0.00000029, -0.00000296, 0.00000077, 0.00000099, 0.00000601, -0.00000811, 0.00000218, 0.00000360, -0.00000947, -0.00000797, 0.00001221, -0.00000538, 0.00000317, -0.00001315, 0.00000704, -0.00000538, 0.00000070, -0.00000701, 0.00001344, -0.00000576, 0.00000042, -0.00000135, 0.00000871, -0.00000903, 0.00000654, 0.00000806, -0.00000413, -0.00000429, 0.00000441, 0.00000446, -0.00000875, 0.00000692, -0.00000858, 0.00000336, -0.00000281, -0.00000160, 0.00000806, 0.00000354, -0.00000268, -0.00000158, -0.00000246, -0.00000079, 0.00001759, -0.00000446, 0.00000528, -0.00000105, -0.00001536, 0.00001052, -0.00001605, 0.00000586, -0.00000605, 0.00000966, -0.00002171, 0.00000446, 0.00000122, -0.00000969, 0.00001239, -0.00001476, 0.00000857, 0.00000308, -0.00000377, 0.00000690, -0.00000362, 0.00000251, -0.00000653, -0.00000213, 0.00000072, 0.00000108, -0.00000819, 0.00000048, -0.00000395, -0.00000065, -0.00000377, -0.00000141, -0.00001800, -0.00000139, -0.00002474, -0.00000381, -0.00015256, 0.00005625, 0.00030478, 0.00012348, -0.00014463, -0.00002823, -0.00002528, -0.00001600, 0.00000360, -0.00000763, -0.00000501, -0.00000611, 0.00000011, -0.00000529, 0.00000130, -0.00000371, 0.00000598, -0.00000692, 0.00000179, -0.00000100, 0.00000091, -0.00000411, -0.00000382, 0.00000041, 0.00000005, -0.00000456, 0.00000056, -0.00000263, -0.00003798, 0.00001356, -0.00000150, 0.00004068, -0.00000349, 0.00003916, 0.00000446, 0.00001309, -0.00002496, },
		{ 0.00000872, -0.00001103, 0.00000031, -0.00000531, -0.00000007, -0.00000481, 0.00000913, -0.00000587, 0.00000018, -0.00000978, 0.00000728, -0.00000213, 0.00000335, -0.00000594, 0.00000366, -0.00000385, 0.00001000, -0.00000881, 0.00000255, 0.00000145, -0.00000321, -0.00001505, 0.00001505, -0.00000650, 0.00000541, -0.00001306, 0.00000440, -0.00000312, 0.00000045, -0.00000753, 0.00001019, 0.00000023, -0.00000345, -0.00000032, 0.00000600, -0.00000287, -0.00000125, 0.00001308, -0.00000580, -0.00000146, -0.00000095, 0.00000999, -0.00001143, 0.00000730, -0.00000727, 0.00000101, -0.00000067, -0.00000303, 0.00000630, 0.00000609, -0.00000306, -0.00000127, -0.00000212, -0.00000255, 0.00001361, 0.00000451, -0.00000424, 0.00001056, -0.00002300, 0.00001275, -0.00001483, 0.00000316, -0.00000495, 0.00000868, -0.00001508, -0.00000593, 0.00000908, -0.00001443, 0.00001302, -0.00001082, 0.00000184, 0.00000859, -0.00000736, 0.00000851, -0.00000365, 0.00000298, -0.00000632, -0.00000172, -0.00000198, 0.00000480, -0.00001016, 0.00000158, -0.00000584, 0.00000196, -0.00000766, 0.00000516, -0.00002269, 0.00000373, -0.00002983, 0.00000849, -0.00012635, -0.00004392, 0.00030075, 0.00020335, -0.00012814, -0.00006087, -0.00001143, -0.00003015, 0.00000917, -0.00001055, -0.00000167, -0.00001039, 0.00000319, -0.00000806, 0.00000331, -0.00000575, 0.00000613, -0.00000462, -0.00000094, 0.00000086, -0.00000094, -0.00000091, -0.00000655, 0.00000183, -0.00000190, -0.00000054, -0.00000577, 0.00000846, -0.00004169, 0.00000954, -0.00000368, 0.00003563, 0.00000383, 0.00002687, 0.00001719, 0.00000557, -0.00000808, -0.00002306, },
		{ -0.00001108, 0.00000123, -0.00000756, 0.00000298, -0.00000892, 0.00001114, -0.00000602, 0.00000201, -0.00001139, 0.00000664, -0.00000211, 0.00000485, -0.00000725, 0.00000496, -0.00000636, 0.00001061, -0.00000676, 0.00000139, 0.00000061, 0.00000070, -0.00001788, 0.00001285, -0.00000296, 0.00000276, -0.00000805, -0.00000185, 0.00000148, -0.00000221, -0.00000449, 0.00000332, 0.00000789, -0.00000750, 0.00000166, 0.00000139, 0.00000411, -0.00000783, 0.00001600, -0.00000608, 0.00000045, -0.00000460, 0.00001251, -0.00001054, 0.00000460, -0.00000381, -0.00000235, 0.00000142, -0.00000372, 0.00000407, 0.00000807, -0.00000257, -0.00000142, -0.00000201, -0.00000238, 0.00000729, 0.00001330, -0.00001072, 0.00001599, -0.00002182, 0.00000630, -0.00000667, -0.00000461, 0.00000045, 0.00000282, -0.00000428, -0.00001729, 0.00001522, -0.00001650, 0.00001107, -0.00000572, -0.00000383, 0.00001147, -0.00000766, 0.00000674, -0.00000099, 0.00000141, -0.00000444, -0.00000270, -0.00000277, 0.00000539, -0.00000832, -0.00000088, -0.00000433, 0.00000096, -0.00000739, 0.00000571, -0.00001938, -0.00000076, -0.00002273, 0.00000056, -0.00007806, -0.00012686, 0.00025086, 0.00029000, -0.00009326, -0.00009568, -0.00000301, -0.00003819, 0.00000835, -0.00000925, -0.00000115, -0.00001192, 0.00000330, -0.00000789, 0.00000231, -0.00000455, 0.00000309, -0.00000030, -0.00000431, 0.00000282, -0.00000276, 0.00000159, -0.00000782, 0.00000189, -0.00000285, 0.00000218, -0.00000970, 0.00001382, -0.00003712, -0.00000118, 0.00000103, 0.00002322, 0.00001708, 0.00001089, 0.00003084, -0.00000301, 0.00000747, -0.00003222, 0.00001026, },
		{ -0.00000133, -0.00000676, 0.00000281, -0.00000911, 0.00000885, -0.00000277, 0.00000094, -0.00000986, 0.00000340, -0.00000030, 0.00000450, -0.00000630, 0.00000378, -0.00000558, 0.00000741, -0.00000198, -0.00000160, 0.00000191, 0.00000126, -0.00001588, 0.00000579, 0.00000442, -0.00000332, -0.00000042, -0.00000907, 0.00000615, -0.00000540, -0.00000012, -0.00000409, 0.00001388, -0.00000910, 0.00000251, -0.00000252, 0.00000886, -0.00001060, 0.00001509, -0.00000358, -0.00000008, -0.00000509, 0.00001112, -0.00000597, -0.00000062, 0.00000093, -0.00000591, 0.00000307, -0.00000393, 0.00000213, 0.00000864, -0.00000066, -0.00000241, -0.00000160, -0.00000172, 0.00000168, 0.00001792, -0.00001045, 0.00001276, -0.00001165, -0.00000667, 0.00000500, -0.00001381, 0.00000668, -0.00000416, 0.00000618, -0.00002513, 0.00001636, -0.00001412, 0.00000598, 0.00000043, -0.00000842, 0.00001227, -0.00000580, 0.00000329, 0.00000237, -0.00000034, -0.00000242, -0.00000396, -0.00000253, 0.00000389, -0.00000426, -0.00000486, -0.00000150, -0.00000161, -0.00000492, 0.00000252, -0.00001142, -0.00000996, -0.00001058, -0.00001473, -0.00003177, -0.00016813, 0.00015303, 0.00036986, -0.00003489, -0.00012912, -0.00000273, -0.00003934, 0.00000174, -0.00000496, -0.00000257, -0.00001113, 0.00000084, -0.00000528, -0.00000096, -0.00000130, -0.00000129, 0.00000387, -0.00000645, 0.00000343, -0.00000325, 0.00000225, -0.00000692, 0.00000022, -0.00000237, 0.00000283, -0.00001020, 0.00001299, -0.00002563, -0.00001575, 0.00000891, 0.00000846, 0.00003011, -0.00000214, 0.00003910, -0.00000632, 0.00001535, -0.00003108, 0.00000354, -0.00000595, },
		{ -0.00000433, 0.00000047, -0.00000641, 0.00000379, 0.00000214, -0.00000146, -0.00000670, -0.00000096, 0.00000218, 0.00000295, -0.00000355, 0.00000055, -0.00000218, 0.00000171, 0.00000397, -0.00000529, 0.00000453, -0.00000051, -0.00001074, -0.00000356, 0.00001250, -0.00000923, 0.00000594, -0.00001362, 0.00000796, -0.00000657, 0.00000268, -0.00000863, 0.00001521, -0.00000638, 0.00000075, -0.00000384, 0.00000945, -0.00000850, 0.00001013, 0.00000172, -0.00000315, -0.00000261, 0.00000661, 0.00000083, -0.00000660, 0.00000536, -0.00000854, 0.00000379, -0.00000372, 0.00000087, 0.00000761, 0.00000250, -0.00000399, -0.00000080, -0.00000158, -0.00000130, 0.00001676, -0.00000338, 0.00000300, 0.00000296, -0.00002004, 0.00001399, -0.00001900, 0.00000908, -0.00000768, 0.00001156, -0.00002560, 0.00001040, -0.00000660, -0.00000222, 0.00000752, -0.00001236, 0.00001215, -0.00000355, 0.00000034, 0.00000418, -0.00000034, -0.00000182, -0.00000410, -0.00000271, 0.00000229, -0.00000067, -0.00000739, -0.00000040, -0.00000268, -0.00000350, -0.00000041, -0.00000427, -0.00001669, -0.00000259, -0.00002381, -0.00000592, -0.00015941, 0.00002452, 0.00041871, 0.00005429, -0.00015918, -0.00001016, -0.00003622, -0.00000723, -0.00000099, -0.00000354, -0.00000987, -0.00000237, -0.00000221, -0.00000431, 0.00000152, -0.00000435, 0.00000551, -0.00000571, 0.00000167, -0.00000171, 0.00000071, -0.00000401, -0.00000274, -0.00000077, 0.00000161, -0.00000765, 0.00000750, -0.00001076, -0.00002956, 0.00001501, -0.00000311, 0.00003716, -0.00000705, 0.00003801, -0.00000125, 0.00001358, -0.00002029, -0.00000870, 0.00000143, -0.00000520, },
		{ -0.00000201, -0.00000308, -0.00000146, 0.00000616, -0.00000304, -0.00000385, -0.00000466, 0.00000394, 0.00000129, -0.00000017, -0.00000333, 0.00000197, -0.00000408, 0.00000871, -0.00000773, 0.00000658, -0.00000230, -0.00000546, -0.00001157, 0.00001743, -0.00001119, 0.00000752, -0.00001286, 0.00000516, -0.00000426, 0.00000212, -0.00000847, 0.00001099, 0.00000041, -0.00000327, -0.00000244, 0.00000607, -0.00000265, 0.00000280, 0.00000822, -0.00000720, 0.00000105, 0.00000120, 0.00000740, -0.00001106, 0.00000772, -0.00000916, 0.00000308, -0.00000295, 0.00000008, 0.00000560, 0.00000592, -0.00000523, -0.00000016, -0.00000195, -0.00000161, 0.00001119, 0.00000700, -0.00000777, 0.00001501, -0.00002681, 0.00001467, -0.00001622, 0.00000474, -0.00000514, 0.00000966, -0.00001787, -0.00000181, 0.00000439, -0.00001185, 0.00001434, -0.00001531, 0.00001153, -0.00000177, -0.00000103, 0.00000357, 0.00000183, -0.00000288, -0.00000282, -0.00000396, 0.00000184, 0.00000072, -0.00000667, -0.00000255, -0.00000077, -0.00000481, -0.00000064, -0.00000148, -0.00001666, -0.00000369, -0.00002056, -0.00000479, -0.00011281, -0.00010132, 0.00040967, 0.00017690, -0.00018125, -0.00002467, -0.00003194, -0.00001509, -0.00000018, -0.00000227, -0.00000961, -0.00000465, -0.00000060, -0.00000573, 0.00000189, -0.00000430, 0.00000353, -0.00000190, -0.00000211, 0.00000133, -0.00000223, -0.00000022, -0.00000585, 0.00000104, -0.00000057, -0.00000350, 0.00000012, 0.00000319, -0.00003819, 0.00001541, -0.00000785, 0.00003541, -0.00000248, 0.00002799, 0.00001055, 0.00000529, -0.00000511, -0.00002052, 0.00000634, -0.00000716, -0.00000263, },
		{ -0.00000090, -0.00000484, 0.00000752, -0.00000252, -0.00000246, -0.00000666, 0.00000422, 0.00000033, 0.00000270, -0.00000628, 0.00000487, -0.00000765, 0.00001023, -0.00000736, 0.00000640, -0.00000207, -0.00000255, -0.00001549, 0.00001674, -0.00000732, 0.00000330, -0.00000679, -0.00000150, 0.00000080, -0.00000135, -0.00000436, 0.00000304, 0.00000871, -0.00000736, 0.00000002, 0.00000080, 0.00000411, -0.00000398, 0.00001334, -0.00000987, 0.00000343, -0.00000259, 0.00001134, -0.00001213, 0.00000686, -0.00000721, 0.00000074, -0.00000146, -0.00000071, 0.00000364, 0.00000826, -0.00000496, -0.00000052, -0.00000215, -0.00000066, 0.00000443, 0.00001566, -0.00001363, 0.00001875, -0.00002290, 0.00000541, -0.00000568, -0.00000525, 0.00000220, 0.00000222, -0.00000505, -0.00001603, 0.00001469, -0.00001943, 0.00001833, -0.00001564, 0.00000942, 0.00000027, -0.00000157, 0.00000169, 0.00000465, -0.00000408, -0.00000135, -0.00000543, 0.00000208, 0.00000029, -0.00000352, -0.00000670, 0.00000264, -0.00000747, 0.00000088, -0.00000233, -0.00001110, -0.00001147, -0.00000904, -0.00001827, -0.00005504, -0.00018759, 0.00032668, 0.00032203, -0.00018355, -0.00004862, -0.00002720, -0.00002108, -0.00000254, 0.00000055, -0.00000987, -0.00000592, -0.00000079, -0.00000488, -0.00000033, -0.00000147, -0.00000095, 0.00000319, -0.00000602, 0.00000414, -0.00000484, 0.00000271, -0.00000763, 0.00000190, -0.00000244, 0.00000043, -0.00000628, 0.00001271, -0.00003902, 0.00000884, -0.00000539, 0.00002588, 0.00000873, 0.00001347, 0.00002379, -0.00000335, 0.00000748, -0.00002604, 0.00000520, -0.00000525, -0.00000304, -0.00000315, },
		{ -0.00000588, 0.00000617, -0.00000024, -0.00000224, -0.00000715, 0.00000320, 0.00000012, 0.00000446, -0.00000735, 0.00000535, -0.00000788, 0.00000796, -0.00000400, 0.00000364, 0.00000078, -0.00000273, -0.00001478, 0.00001061, 0.00000126, -0.00000462, 0.00000165, -0.00000890, 0.00000588, -0.00000531, 0.00000086, -0.00000495, 0.00001474, -0.00000871, 0.00000126, -0.00000360, 0.00000868, -0.00000747, 0.00001491, -0.00000922, 0.00000271, -0.00000314, 0.00001148, -0.00000928, 0.00000288, -0.00000311, -0.00000281, 0.00000061, -0.00000186, 0.00000254, 0.00000865, -0.00000262, -0.00000223, -0.00000156, -0.00000006, -0.00000041, 0.00001883, -0.00001141, 0.00001276, -0.00000960, -0.00001005, 0.00000770, -0.00001591, 0.00000964, -0.00000578, 0.00000700, -0.00002625, 0.00001925, -0.00002111, 0.00001665, -0.00001135, 0.00000417, 0.00000415, -0.00000309, 0.00000078, 0.00000561, -0.00000313, -0.00000168, -0.00000536, 0.00000131, 0.00000007, -0.00000068, -0.00000952, 0.00000407, -0.00000801, 0.00000074, -0.00000291, -0.00000540, -0.00001847, 0.00000050, -0.00002967, -0.00001390, -0.00021028, 0.00017685, 0.00046053, -0.00014686, -0.00008722, -0.00001974, -0.00002769, -0.00000503, 0.00000175, -0.00000829, -0.00000780, -0.00000134, -0.00000335, -0.00000323, 0.00000174, -0.00000501, 0.00000655, -0.00000747, 0.00000459, -0.00000523, 0.00000316, -0.00000695, 0.00000096, -0.00000297, 0.00000261, -0.00000963, 0.00001598, -0.00003203, -0.00000312, 0.00000168, 0.00001251, 0.00002133, 0.00000048, 0.00003254, -0.00000668, 0.00001255, -0.00002269, -0.00000209, -0.00000036, -0.00000488, -0.00000311, -0.00000021, },
		{ 0.00000356, 0.00000221, -0.00000195, -0.00000714, 0.00000178, 0.00000018, 0.00000517, -0.00000650, 0.00000348, -0.00000525, 0.00000305, 0.00000095, -0.00000048, 0.00000507, -0.00000469, -0.00001119, 0.00000171, 0.00001094, -0.00001202, 0.00000805, -0.00001309, 0.00000780, -0.00000690, 0.00000395, -0.00000936, 0.00001555, -0.00000557, -0.00000011, -0.00000529, 0.00000922, -0.00000648, 0.00001230, -0.00000490, -0.00000124, -0.00000065, 0.00000838, -0.00000359, -0.00000283, 0.00000175, -0.00000646, 0.00000263, -0.00000318, 0.00000241, 0.00000715, 0.00000131, -0.00000474, -0.00000035, -0.00000049, -0.00000198, 0.00001579, -0.00000230, 0.00000067, 0.00000692, -0.00002416, 0.00001662, -0.00002095, 0.00001176, -0.00000900, 0.00001269, -0.00002764, 0.00001480, -0.00001489, 0.00000815, -0.00000176, -0.00000475, 0.00001057, -0.00000661, 0.00000232, 0.00000318, 0.00000117, -0.00000477, -0.00000270, -0.00000177, 0.00000195, -0.00000074, -0.00000813, 0.00000068, -0.00000371, -0.00000386, 0.00000035, -0.00000467, -0.00001789, -0.00000058, -0.00002666, -0.00000440, -0.00017017, -0.00000595, 0.00054964, -0.00005138, -0.00014359, -0.00000768, -0.00003796, -0.00000405, -0.00000199, -0.00000256, -0.00001207, -0.00000036, -0.00000338, -0.00000422, 0.00000244, -0.00000569, 0.00000563, -0.00000472, 0.00000154, -0.00000254, 0.00000058, -0.00000367, -0.00000180, -0.00000185, 0.00000242, -0.00000924, 0.00001330, -0.00001948, -0.00001692, 0.00000931, 0.00000004, 0.00003018, -0.00000627, 0.00003333, -0.00000251, 0.00000967, -0.00001263, -0.00001179, 0.00000407, -0.00000576, -0.00000337, 0.00000000, -0.00000562, },
		{ 0.00000327, -0.00000049, -0.00000747, 0.00000066, 0.00000009, 0.00000512, -0.00000427, 0.00000026, -0.00000135, -0.00000234, 0.00000530, -0.00000397, 0.00000870, -0.00000607, -0.00000756, -0.00000636, 0.00001758, -0.00001476, 0.00000888, -0.00001159, 0.00000494, -0.00000470, 0.00000321, -0.00000863, 0.00001065, 0.00000143, -0.00000352, -0.00000430, 0.00000598, -0.00000198, 0.00000686, 0.00000166, -0.00000677, 0.00000303, 0.00000409, 0.00000272, -0.00000809, 0.00000554, -0.00000886, 0.00000367, -0.00000404, 0.00000260, 0.00000478, 0.00000541, -0.00000668, 0.00000054, -0.00000148, -0.00000098, 0.00000890, 0.00000908, -0.00001086, 0.00001877, -0.00002947, 0.00001541, -0.00001654, 0.00000585, -0.00000501, 0.00000990, -0.00001936, 0.00000214, -0.00000245, -0.00000513, 0.00001092, -0.00001533, 0.00001784, -0.00001092, 0.00000548, -0.00000170, 0.00000754, -0.00000935, 0.00000151, -0.00000652, 0.00000579, -0.00000391, -0.00000246, -0.00000712, 0.00000466, -0.00001213, 0.00000725, -0.00000953, -0.00000886, -0.00001284, -0.00000900, -0.00002198, -0.00009345, -0.00017032, 0.00054790, 0.00011048, -0.00021062, 0.00000513, -0.00005112, 0.00000070, -0.00001087, 0.00000686, -0.00001834, 0.00000240, -0.00000577, -0.00000218, -0.00000059, -0.00000204, 0.00000021, 0.00000168, -0.00000408, 0.00000226, -0.00000391, 0.00000098, -0.00000534, 0.00000030, 0.00000031, -0.00000590, 0.00000665, -0.00000504, -0.00002850, 0.00001380, -0.00000795, 0.00003236, -0.00000527, 0.00002657, 0.00000706, 0.00000262, -0.00000133, -0.00001847, 0.00000423, -0.00000350, -0.00000525, 0.00000081, -0.00000537, 0.00000137, },
		{ 0.00000204, -0.00000796, -0.00000027, -0.00000003, 0.00000440, -0.00000131, -0.00000309, 0.00000210, -0.00000621, 0.00000727, -0.00000528, 0.00000995, -0.00000500, -0.00000605, -0.00001099, 0.00001862, -0.00001103, 0.00000347, -0.00000487, -0.00000160, 0.00000026, -0.00000056, -0.00000398, 0.00000238, 0.00000937, -0.00000668, -0.00000230, 0.00000099, 0.00000349, 0.00000118, 0.00000793, -0.00001133, 0.00000542, 0.00000095, 0.00000740, -0.00001092, 0.00000680, -0.00000899, 0.00000298, -0.00000376, 0.00000224, 0.00000288, 0.00000808, -0.00000666, 0.00000007, -0.00000198, 0.00000066, 0.00000191, 0.00001732, -0.00001575, 0.00002050, -0.00002272, 0.00000353, -0.00000402, -0.00000593, 0.00000396, 0.00000126, -0.00000538, -0.00001364, 0.00001094, -0.00001807, 0.00002183, -0.00002313, 0.00002204, -0.00001263, 0.00000717, -0.00000576, 0.00001245, -0.00001205, 0.00000430, -0.00001022, 0.00000904, -0.00000749, 0.00000411, -0.00001518, 0.00001247, -0.00001924, 0.00001313, -0.00001516, 0.00000256, -0.00002799, 0.00001178, -0.00004665, -0.00001967, -0.00026705, 0.00043496, 0.00032138, -0.00026507, 0.00000603, -0.00006011, 0.00000406, -0.00002013, 0.00001539, -0.00002297, 0.00000442, -0.00000870, 0.00000131, -0.00000563, 0.00000378, -0.00000690, 0.00000853, -0.00000926, 0.00000646, -0.00000778, 0.00000466, -0.00000782, 0.00000205, -0.00000231, -0.00000148, -0.00000099, 0.00000743, -0.00003459, 0.00001300, -0.00001012, 0.00002780, 0.00000172, 0.00001580, 0.00001736, -0.00000315, 0.00000567, -0.00001827, -0.00000133, 0.00000188, -0.00000837, 0.00000211, -0.00000549, 0.00000025, 0.00000256, },
		{ -0.00000742, -0.00000189, 0.00000050, 0.00000278, 0.00000198, -0.00000572, 0.00000405, -0.00000760, 0.00000637, -0.00000416, 0.00000845, -0.00000119, -0.00000696, -0.00001179, 0.00001426, -0.00000221, -0.00000555, 0.00000363, -0.00000838, 0.00000503, -0.00000464, 0.00000140, -0.00000532, 0.00001450, -0.00000697, -0.00000135, -0.00000328, 0.00000719, -0.00000230, 0.00001166, -0.00001277, 0.00000468, 0.00000043, 0.00000922, -0.00001049, 0.00000508, -0.00000669, 0.00000051, -0.00000215, 0.00000081, 0.00000232, 0.00000842, -0.00000417, -0.00000205, -0.00000140, 0.00000133, -0.00000221, 0.00001908, -0.00001170, 0.00001196, -0.00000678, -0.00001369, 0.00001028, -0.00001738, 0.00001211, -0.00000721, 0.00000737, -0.00002530, 0.00001867, -0.00002471, 0.00002556, -0.00002337, 0.00001905, -0.00000818, 0.00000404, -0.00000538, 0.00001215, -0.00000946, 0.00000252, -0.00000972, 0.00000834, -0.00000752, 0.00000673, -0.00001790, 0.00001375, -0.00001915, 0.00001185, -0.00001460, 0.00000739, -0.00003426, 0.00002035, -0.00005567, 0.00001741, -0.00027072, 0.00022669, 0.00053464, -0.00026977, -0.00002328, -0.00005399, -0.00000269, -0.00002196, 0.00001604, -0.00002051, 0.00000153, -0.00000851, 0.00000261, -0.00000882, 0.00000750, -0.00001114, 0.00001144, -0.00001026, 0.00000698, -0.00000831, 0.00000506, -0.00000744, 0.00000191, -0.00000375, 0.00000179, -0.00000669, 0.00001491, -0.00003361, 0.00000675, -0.00000720, 0.00001856, 0.00001122, 0.00000546, 0.00002397, -0.00000385, 0.00000599, -0.00001137, -0.00001029, 0.00000749, -0.00001040, 0.00000234, -0.00000500, -0.00000057, 0.00000121, 0.00000422, },
		{ -0.00000502, 0.00000221, 0.00000017, 0.00000525, -0.00000725, 0.00000432, -0.00000679, 0.00000350, -0.00000165, 0.00000527, 0.00000402, -0.00000882, -0.00001043, 0.00000701, 0.00000797, -0.00001364, 0.00000937, -0.00001154, 0.00000651, -0.00000619, 0.00000430, -0.00000905, 0.00001431, -0.00000316, -0.00000242, -0.00000533, 0.00000764, -0.00000247, 0.00001198, -0.00001043, 0.00000064, 0.00000235, 0.00000848, -0.00000732, 0.00000121, -0.00000284, -0.00000300, 0.00000028, -0.00000141, 0.00000300, 0.00000673, 0.00000005, -0.00000499, -0.00000015, 0.00000065, -0.00000269, 0.00001457, -0.00000110, -0.00000175, 0.00001071, -0.00002752, 0.00001824, -0.00002162, 0.00001352, -0.00000966, 0.00001276, -0.00002727, 0.00001634, -0.00002149, 0.00001917, -0.00001386, 0.00000736, 0.00000341, -0.00000485, 0.00000071, 0.00000536, -0.00000063, -0.00000467, -0.00000389, 0.00000207, -0.00000176, 0.00000245, -0.00001202, 0.00000535, -0.00000891, 0.00000049, -0.00000408, 0.00000028, -0.00002466, 0.00000814, -0.00003765, 0.00000621, -0.00019222, -0.00002367, 0.00068423, -0.00018668, -0.00009612, -0.00002570, -0.00002564, -0.00001070, 0.00000404, -0.00000760, -0.00000890, -0.00000251, -0.00000134, -0.00000669, 0.00000543, -0.00000890, 0.00000745, -0.00000500, 0.00000227, -0.00000423, 0.00000121, -0.00000357, -0.00000078, -0.00000292, 0.00000237, -0.00000855, 0.00001608, -0.00002589, -0.00000345, -0.00000120, 0.00000745, 0.00001983, -0.00000132, 0.00002502, 0.00000072, 0.00000150, -0.00000199, -0.00001749, 0.00000903, -0.00000841, -0.00000038, -0.00000255, -0.00000246, 0.00000093, 0.00000430, -0.00000457, },
		{ 0.00000469, -0.00000283, 0.00000789, -0.00000743, 0.00000324, -0.00000477, 0.00000014, 0.00000070, 0.00000201, 0.00000861, -0.00000947, -0.00000925, 0.00000007, 0.00001547, -0.00001674, 0.00000911, -0.00000904, 0.00000347, -0.00000409, 0.00000337, -0.00000774, 0.00000891, 0.00000360, -0.00000467, -0.00000534, 0.00000515, -0.00000008, 0.00000968, -0.00000531, -0.00000524, 0.00000510, 0.00000667, -0.00000293, -0.00000326, 0.00000113, -0.00000632, 0.00000249, -0.00000354, 0.00000400, 0.00000428, 0.00000433, -0.00000716, 0.00000063, -0.00000054, -0.00000086, 0.00000695, 0.00001056, -0.00001319, 0.00002141, -0.00003055, 0.00001460, -0.00001533, 0.00000599, -0.00000410, 0.00000895, -0.00001879, 0.00000450, -0.00000963, 0.00000443, 0.00000294, -0.00000997, 0.00001882, -0.00001640, 0.00000983, -0.00000520, 0.00001140, -0.00001437, 0.00000483, -0.00000771, 0.00000810, -0.00000722, 0.00000058, -0.00001011, 0.00000819, -0.00001736, 0.00001306, -0.00001589, -0.00000253, -0.00002004, 0.00000038, -0.00003775, -0.00007500, -0.00024191, 0.00070807, 0.00000255, -0.00020707, 0.00001881, -0.00006126, 0.00001129, -0.00001837, 0.00001317, -0.00002466, 0.00000804, -0.00001016, 0.00000088, -0.00000273, -0.00000021, -0.00000271, 0.00000513, -0.00000612, 0.00000306, -0.00000550, 0.00000248, -0.00000514, -0.00000012, 0.00000018, -0.00000647, 0.00001169, -0.00001365, -0.00001494, 0.00000538, -0.00000294, 0.00002536, -0.00000363, 0.00002155, 0.00000760, -0.00000316, 0.00000420, -0.00001806, 0.00000362, -0.00000126, -0.00000639, 0.00000205, -0.00000588, 0.00000253, 0.00000165, 0.00000062, -0.00000936, },
		{ -0.00000475, 0.00000884, -0.00000593, 0.00000116, -0.00000241, -0.00000256, 0.00000189, -0.00000023, 0.00001117, -0.00000754, -0.00000969, -0.00000451, 0.00001800, -0.00001335, 0.00000276, -0.00000199, -0.00000251, 0.00000029, -0.00000023, -0.00000306, 0.00000104, 0.00001027, -0.00000596, -0.00000473, 0.00000137, 0.00000300, 0.00000658, 0.00000061, -0.00001068, 0.00000660, 0.00000544, 0.00000111, -0.00000681, 0.00000387, -0.00000829, 0.00000346, -0.00000458, 0.00000421, 0.00000251, 0.00000703, -0.00000717, -0.00000010, -0.00000111, 0.00000115, 0.00000018, 0.00001785, -0.00001659, 0.00002064, -0.00002077, 0.00000022, -0.00000119, -0.00000715, 0.00000616, -0.00000048, -0.00000457, -0.00001121, 0.00000499, -0.00001246, 0.00002033, -0.00002593, 0.00003116, -0.00002431, 0.00001617, -0.00001364, 0.00002060, -0.00002085, 0.00001112, -0.00001581, 0.00001673, -0.00001646, 0.00001311, -0.00002475, 0.00002338, -0.00003234, 0.00002723, -0.00003067, 0.00001949, -0.00004791, 0.00003689, -0.00008138, 0.00002445, -0.00036143, 0.00057625, 0.00027799, -0.00032296, 0.00005625, -0.00009366, 0.00003165, -0.00004050, 0.00003223, -0.00003791, 0.00001710, -0.00001897, 0.00000962, -0.00001276, 0.00001044, -0.00001409, 0.00001516, -0.00001372, 0.00000964, -0.00001154, 0.00000765, -0.00000882, 0.00000289, -0.00000325, -0.00000220, 0.00000434, -0.00000036, -0.00002458, 0.00001015, -0.00001065, 0.00002687, -0.00000224, 0.00001616, 0.00001302, -0.00000379, 0.00000368, -0.00001065, -0.00000768, 0.00000875, -0.00001332, 0.00000690, -0.00000948, 0.00000498, -0.00000239, 0.00000637, -0.00001308, 0.00000630, },
		{ 0.00000708, -0.00000244, -0.00000175, -0.00000007, -0.00000435, 0.00000197, -0.00000142, 0.00001151, -0.00000321, -0.00001148, -0.00000670, 0.00001589, -0.00000510, -0.00000659, 0.00000584, -0.00000790, 0.00000398, -0.00000370, 0.00000172, -0.00000546, 0.00001353, -0.00000431, -0.00000490, -0.00000193, 0.00000488, 0.00000433, 0.00000551, -0.00001378, 0.00000548, 0.00000565, 0.00000399, -0.00000863, 0.00000474, -0.00000838, 0.00000273, -0.00000406, 0.00000304, 0.00000222, 0.00000740, -0.00000474, -0.00000239, -0.00000065, 0.00000191, -0.00000320, 0.00001819, -0.00001079, 0.00000982, -0.00000279, -0.00001791, 0.00001312, -0.00001857, 0.00001435, -0.00000866, 0.00000775, -0.00002284, 0.00001471, -0.00002380, 0.00003039, -0.00003277, 0.00003326, -0.00002231, 0.00001402, -0.00001415, 0.00002113, -0.00001867, 0.00000979, -0.00001676, 0.00001812, -0.00001841, 0.00001758, -0.00002956, 0.00002699, -0.00003441, 0.00002794, -0.00003227, 0.00002686, -0.00005722, 0.00004885, -0.00009295, 0.00006292, -0.00035171, 0.00031057, 0.00057616, -0.00038687, 0.00005221, -0.00010009, 0.00003274, -0.00004746, 0.00003688, -0.00003849, 0.00001648, -0.00002064, 0.00001277, -0.00001792, 0.00001612, -0.00001980, 0.00001876, -0.00001518, 0.00001100, -0.00001294, 0.00000855, -0.00000908, 0.00000375, -0.00000548, 0.00000147, -0.00000246, 0.00001013, -0.00002945, 0.00001115, -0.00001428, 0.00002409, 0.00000172, 0.00001096, 0.00001494, 0.00000062, -0.00000262, 0.00000118, -0.00001958, 0.00001630, -0.00001675, 0.00000858, -0.00001055, 0.00000578, -0.00000506, 0.00000970, -0.00001358, 0.00000497, -0.00000398, },
		{ 0.00000278, -0.00000521, 0.00000218, -0.00000560, 0.00000169, -0.00000230, 0.00001045, 0.00000209, -0.00001298, -0.00000798, 0.00001134, 0.00000439, -0.00001442, 0.00001007, -0.00000923, 0.00000431, -0.00000459, 0.00000380, -0.00000782, 0.00001170, 0.00000061, -0.00000606, -0.00000397, 0.00000480, 0.00000352, 0.00000860, -0.00001382, 0.00000170, 0.00000694, 0.00000590, -0.00000881, 0.00000389, -0.00000683, 0.00000055, -0.00000224, 0.00000076, 0.00000315, 0.00000598, -0.00000085, -0.00000515, 0.00000018, 0.00000143, -0.00000306, 0.00001277, 0.00000059, -0.00000460, 0.00001451, -0.00003027, 0.00001906, -0.00002116, 0.00001440, -0.00000963, 0.00001189, -0.00002450, 0.00001430, -0.00002450, 0.00002816, -0.00002614, 0.00002186, -0.00000818, 0.00000162, -0.00000483, 0.00001115, -0.00000638, -0.00000065, -0.00000853, 0.00000954, -0.00000962, 0.00000981, -0.00001999, 0.00001456, -0.00001933, 0.00001092, -0.00001546, 0.00001261, -0.00003928, 0.00002621, -0.00006024, 0.00003062, -0.00023261, -0.00001600, 0.00080489, -0.00033786, -0.00002185, -0.00006407, 0.00000212, -0.00002904, 0.00001887, -0.00002025, 0.00000126, -0.00001048, 0.00000542, -0.00001302, 0.00001158, -0.00001496, 0.00001196, -0.00000758, 0.00000478, -0.00000761, 0.00000345, -0.00000452, 0.00000096, -0.00000457, 0.00000210, -0.00000578, 0.00001500, -0.00002781, 0.00000732, -0.00001305, 0.00001703, 0.00000754, 0.00000649, 0.00001412, 0.00000737, -0.00000977, 0.00001064, -0.00002512, 0.00001581, -0.00001282, 0.00000436, -0.00000689, 0.00000277, -0.00000389, 0.00000823, -0.00000912, -0.00000040, 0.00000016, 0.00000244, },
		{ -0.00000832, 0.00000408, -0.00000660, 0.00000172, -0.00000348, 0.00000893, 0.00000674, -0.00001250, -0.00000997, 0.00000678, 0.00001160, -0.00001712, 0.00000818, -0.00000531, 0.00000080, -0.00000237, 0.00000247, -0.00000574, 0.00000576, 0.00000700, -0.00000696, -0.00000519, 0.00000324, 0.00000360, 0.00001016, -0.00001130, -0.00000355, 0.00000803, 0.00000762, -0.00000790, 0.00000191, -0.00000430, -0.00000232, 0.00000004, -0.00000175, 0.00000431, 0.00000408, 0.00000280, -0.00000679, 0.00000021, 0.00000070, -0.00000114, 0.00000529, 0.00001153, -0.00001481, 0.00002283, -0.00003000, 0.00001224, -0.00001251, 0.00000493, -0.00000218, 0.00000666, -0.00001574, 0.00000407, -0.00001504, 0.00001452, -0.00000797, 0.00000014, 0.00001392, -0.00001676, 0.00001035, -0.00000528, 0.00001168, -0.00001608, 0.00000537, -0.00000592, 0.00000703, -0.00000728, 0.00000046, -0.00000948, 0.00000752, -0.00001779, 0.00001369, -0.00001710, -0.00000147, -0.00002095, 0.00000224, -0.00004536, -0.00006643, -0.00030303, 0.00087613, -0.00014058, -0.00016755, 0.00001010, -0.00005640, 0.00001149, -0.00001843, 0.00001307, -0.00002531, 0.00000935, -0.00001126, 0.00000138, -0.00000278, -0.00000033, -0.00000375, 0.00000710, -0.00000707, 0.00000276, -0.00000611, 0.00000360, -0.00000482, -0.00000048, -0.00000088, -0.00000463, 0.00001355, -0.00001985, -0.00000094, -0.00000726, 0.00000631, 0.00001472, 0.00000217, 0.00001300, 0.00001205, -0.00001196, 0.00001167, -0.00001975, 0.00000509, -0.00000110, -0.00000543, 0.00000115, -0.00000403, 0.00000137, 0.00000200, -0.00000039, -0.00000867, 0.00000636, -0.00000366, 0.00000839, },
		{ 0.00000480, -0.00000700, 0.00000187, -0.00000470, 0.00000711, 0.00000995, -0.00000945, -0.00001307, 0.00000331, 0.00001492, -0.00001387, 0.00000091, 0.00000197, -0.00000443, 0.00000118, -0.00000065, -0.00000139, -0.00000120, 0.00001177, -0.00000576, -0.00000661, 0.00000132, 0.00000351, 0.00001091, -0.00000729, -0.00000867, 0.00000767, 0.00000967, -0.00000624, -0.00000067, -0.00000144, -0.00000511, 0.00000191, -0.00000354, 0.00000468, 0.00000294, 0.00000486, -0.00000626, -0.00000125, 0.00000063, 0.00000057, -0.00000051, 0.00001705, -0.00001590, 0.00001886, -0.00001681, -0.00000470, 0.00000309, -0.00000933, 0.00000920, -0.00000336, -0.00000192, -0.00001011, -0.00000134, -0.00000398, 0.00001400, -0.00002305, 0.00003465, -0.00003212, 0.00002294, -0.00001968, 0.00002688, -0.00002824, 0.00001710, -0.00002033, 0.00002312, -0.00002444, 0.00002125, -0.00003310, 0.00003276, -0.00004377, 0.00003987, -0.00004532, 0.00003589, -0.00006698, 0.00006115, -0.00011638, 0.00006985, -0.00046146, 0.00074407, 0.00018835, -0.00034573, 0.00009046, -0.00011885, 0.00005474, -0.00005829, 0.00004675, -0.00005096, 0.00002894, -0.00002901, 0.00001797, -0.00001998, 0.00001700, -0.00002109, 0.00002186, -0.00001844, 0.00001292, -0.00001549, 0.00001132, -0.00001057, 0.00000454, -0.00000567, -0.00000060, 0.00000769, -0.00000808, -0.00001140, 0.00000125, -0.00000619, 0.00002201, -0.00000219, 0.00001348, 0.00001175, -0.00000632, 0.00000283, -0.00000468, -0.00001233, 0.00001393, -0.00001655, 0.00000996, -0.00001165, 0.00000804, -0.00000602, 0.00000917, -0.00001639, 0.00001147, -0.00000841, 0.00001212, -0.00000959, },
		{ -0.00000593, 0.00000115, -0.00000489, 0.00000437, 0.00001203, -0.00000470, -0.00001615, 0.00000035, 0.00001501, -0.00000665, -0.00000822, 0.00000856, -0.00000785, 0.00000319, -0.00000292, 0.00000214, -0.00000575, 0.00001236, -0.00000143, -0.00000866, -0.00000024, 0.00000263, 0.00001124, -0.00000264, -0.00001249, 0.00000537, 0.00001181, -0.00000367, -0.00000370, 0.00000136, -0.00000733, 0.00000286, -0.00000416, 0.00000388, 0.00000297, 0.00000507, -0.00000380, -0.00000374, 0.00000113, 0.00000119, -0.00000289, 0.00001575, -0.00000823, 0.00000587, 0.00000273, -0.00002308, 0.00001669, -0.00002002, 0.00001685, -0.00001061, 0.00000888, -0.00002006, 0.00000851, -0.00001856, 0.00002991, -0.00003723, 0.00004382, -0.00003521, 0.00002439, -0.00002351, 0.00003069, -0.00002903, 0.00001867, -0.00002513, 0.00002915, -0.00003100, 0.00003031, -0.00004287, 0.00004193, -0.00005173, 0.00004681, -0.00005373, 0.00005077, -0.00008489, 0.00008314, -0.00013840, 0.00011941, -0.00045056, 0.00043124, 0.00057225, -0.00048081, 0.00012563, -0.00014903, 0.00007238, -0.00007718, 0.00006094, -0.00005948, 0.00003488, -0.00003613, 0.00002591, -0.00002953, 0.00002672, -0.00003019, 0.00002789, -0.00002188, 0.00001645, -0.00001900, 0.00001378, -0.00001226, 0.00000697, -0.00000892, 0.00000296, 0.00000120, 0.00000342, -0.00002047, 0.00000918, -0.00001717, 0.00002685, -0.00000526, 0.00001508, 0.00000720, 0.00000492, -0.00001125, 0.00001297, -0.00002812, 0.00002440, -0.00002243, 0.00001419, -0.00001546, 0.00001174, -0.00001121, 0.00001482, -0.00001927, 0.00001192, -0.00000886, 0.00001172, -0.00000752, 0.00000330, },
		{ -0.00000138, -0.00000298, 0.00000003, 0.00001372, 0.00000015, -0.00001734, -0.00000353, 0.00001375, 0.00000117, -0.00001489, 0.00001055, -0.00000669, 0.00000175, -0.00000252, 0.00000278, -0.00000607, 0.00000828, 0.00000515, -0.00001048, -0.00000166, 0.00000126, 0.00001085, 0.00000235, -0.00001463, 0.00000164, 0.00001305, 0.00000032, -0.00000713, 0.00000387, -0.00000873, 0.00000278, -0.00000371, 0.00000224, 0.00000367, 0.00000429, -0.00000074, -0.00000583, 0.00000119, 0.00000124, -0.00000252, 0.00000990, 0.00000329, -0.00000838, 0.00001872, -0.00003286, 0.00001962, -0.00002011, 0.00001484, -0.00000935, 0.00001070, -0.00002009, 0.00000897, -0.00002297, 0.00003286, -0.00003572, 0.00003578, -0.00002181, 0.00001127, -0.00001333, 0.00001981, -0.00001550, 0.00000701, -0.00001642, 0.00002027, -0.00002125, 0.00002110, -0.00003180, 0.00002795, -0.00003463, 0.00002724, -0.00003381, 0.00003262, -0.00006213, 0.00005412, -0.00009549, 0.00007089, -0.00029559, 0.00002778, 0.00089298, -0.00048650, 0.00006812, -0.00011738, 0.00004256, -0.00005773, 0.00004172, -0.00004002, 0.00001826, -0.00002445, 0.00001722, -0.00002354, 0.00002118, -0.00002412, 0.00001953, -0.00001295, 0.00000960, -0.00001324, 0.00000802, -0.00000731, 0.00000418, -0.00000770, 0.00000282, -0.00000219, 0.00001064, -0.00002451, 0.00001288, -0.00002281, 0.00002603, -0.00000449, 0.00001511, 0.00000246, 0.00001557, -0.00002221, 0.00002357, -0.00003323, 0.00002319, -0.00001792, 0.00000989, -0.00001204, 0.00000903, -0.00000994, 0.00001313, -0.00001465, 0.00000582, -0.00000371, 0.00000643, -0.00000174, -0.00000077, -0.00000278, },
		{ 0.00000096, -0.00000558, 0.00001514, 0.00000398, -0.00001545, -0.00000911, 0.00001260, 0.00000692, -0.00001635, 0.00000645, -0.00000086, -0.00000267, 0.00000020, 0.00000072, -0.00000292, 0.00000156, 0.00001132, -0.00001024, -0.00000388, 0.00000031, 0.00000907, 0.00000767, -0.00001521, -0.00000253, 0.00001236, 0.00000589, -0.00001056, 0.00000565, -0.00000903, 0.00000170, -0.00000259, 0.00000044, 0.00000414, 0.00000372, 0.00000136, -0.00000608, -0.00000023, 0.00000173, -0.00000137, 0.00000352, 0.00001241, -0.00001613, 0.00002339, -0.00002822, 0.00000881, -0.00000848, 0.00000296, 0.00000055, 0.00000333, -0.00001039, 0.00000031, -0.00001691, 0.00002249, -0.00001919, 0.00001292, 0.00000418, -0.00001197, 0.00000653, -0.00000130, 0.00000756, -0.00001342, 0.00000206, -0.00000020, 0.00000156, -0.00000287, -0.00000411, -0.00000389, 0.00000111, -0.00001156, 0.00000683, -0.00001044, -0.00000873, -0.00001198, -0.00000812, -0.00003852, -0.00007683, -0.00033888, 0.00103302, -0.00030501, -0.00009629, -0.00002198, -0.00003417, -0.00000143, -0.00000854, 0.00000432, -0.00001812, 0.00000418, -0.00000713, -0.00000238, 0.00000071, -0.00000372, -0.00000161, 0.00000625, -0.00000576, 0.00000030, -0.00000466, 0.00000331, -0.00000355, -0.00000158, -0.00000201, -0.00000084, 0.00001160, -0.00002166, 0.00001013, -0.00002059, 0.00001748, 0.00000205, 0.00001074, 0.00000206, 0.00001927, -0.00002269, 0.00002034, -0.00002313, 0.00000840, -0.00000286, -0.00000248, -0.00000175, -0.00000003, -0.00000202, 0.00000429, -0.00000345, -0.00000536, 0.00000538, -0.00000204, 0.00000620, -0.00000635, 0.00000192, -0.00000559, },
		{ -0.00001072, 0.00001525, 0.00000692, -0.00001091, -0.00001551, 0.00001141, 0.00001001, -0.00001272, -0.00000196, 0.00000678, -0.00000724, 0.00000294, -0.00000190, 0.00000103, -0.00000432, 0.00001399, -0.00000639, -0.00000751, 0.00000038, 0.00000565, 0.00001294, -0.00001426, -0.00000635, 0.00000946, 0.00001223, -0.00001284, 0.00000580, -0.00000778, -0.00000040, -0.00000109, -0.00000102, 0.00000382, 0.00000405, 0.00000183, -0.00000415, -0.00000318, 0.00000300, -0.00000089, -0.00000033, 0.00001513, -0.00001387, 0.00001532, -0.00001107, -0.00001098, 0.00000875, -0.00001249, 0.00001318, -0.00000746, 0.00000292, -0.00001149, -0.00000595, 0.00000502, 0.00000449, -0.00001505, 0.00003187, -0.00003444, 0.00002559, -0.00002204, 0.00002936, -0.00003217, 0.00002023, -0.00002189, 0.00002617, -0.00002916, 0.00002618, -0.00003773, 0.00003784, -0.00005035, 0.00004738, -0.00005502, 0.00004727, -0.00008003, 0.00007822, -0.00014374, 0.00010631, -0.00055327, 0.00092464, 0.00005726, -0.00032825, 0.00010050, -0.00012824, 0.00006694, -0.00006828, 0.00005457, -0.00005822, 0.00003642, -0.00003583, 0.00002382, -0.00002510, 0.00002151, -0.00002606, 0.00002693, -0.00002201, 0.00001506, -0.00001848, 0.00001471, -0.00001241, 0.00000649, -0.00000924, 0.00000366, 0.00000755, -0.00001309, 0.00000159, -0.00001103, 0.00000221, 0.00001355, 0.00000172, 0.00000771, 0.00001360, -0.00001096, 0.00000373, -0.00000117, -0.00001443, 0.00001654, -0.00001726, 0.00001058, -0.00001131, 0.00000864, -0.00000749, 0.00000964, -0.00001692, 0.00001409, -0.00000982, 0.00001301, -0.00001092, 0.00000625, -0.00000986, 0.00000525, },
		{ 0.00001235, 0.00001006, -0.00000550, -0.00002049, 0.00000866, 0.00001170, -0.00000653, -0.00001063, 0.00001191, -0.00000847, 0.00000302, -0.00000264, 0.00000294, -0.00000650, 0.00001148, 0.00000099, -0.00001188, 0.00000105, 0.00000140, 0.00001701, -0.00001135, -0.00000976, 0.00000512, 0.00001757, -0.00001230, 0.00000329, -0.00000450, -0.00000361, 0.00000071, -0.00000214, 0.00000292, 0.00000491, 0.00000126, -0.00000117, -0.00000629, 0.00000409, -0.00000100, -0.00000112, 0.00001166, -0.00000392, -0.00000001, 0.00000984, -0.00002931, 0.00002129, -0.00002213, 0.00002006, -0.00001352, 0.00001156, -0.00001842, 0.00000202, -0.00001050, 0.00002438, -0.00003574, 0.00004856, -0.00004415, 0.00003254, -0.00003110, 0.00003854, -0.00003822, 0.00002702, -0.00003280, 0.00003923, -0.00004289, 0.00004245, -0.00005526, 0.00005572, -0.00006788, 0.00006490, -0.00007514, 0.00007504, -0.00011269, 0.00011778, -0.00018567, 0.00017941, -0.00055875, 0.00058410, 0.00051664, -0.00053748, 0.00018342, -0.00019047, 0.00010829, -0.00010502, 0.00008332, -0.00007927, 0.00005312, -0.00005201, 0.00003961, -0.00004163, 0.00003752, -0.00004066, 0.00003741, -0.00002924, 0.00002242, -0.00002571, 0.00002020, -0.00001675, 0.00001153, -0.00001431, 0.00000725, 0.00000236, -0.00000270, -0.00000911, 0.00000192, -0.00001519, 0.00002566, -0.00000850, 0.00001652, 0.00000202, 0.00000761, -0.00001819, 0.00002226, -0.00003428, 0.00003025, -0.00002602, 0.00001788, -0.00001847, 0.00001591, -0.00001586, 0.00001825, -0.00002301, 0.00001735, -0.00001259, 0.00001516, -0.00001163, 0.00000770, -0.00001173, 0.00000728, -0.00001281, },
		{ 0.00001432, -0.00000118, -0.00002178, 0.00000280, 0.00001369, -0.00000088, -0.00001560, 0.00001125, -0.00000446, -0.00000059, -0.00000052, 0.00000171, -0.00000434, 0.00000474, 0.00000963, -0.00001488, 0.00000083, -0.00000203, 0.00001830, -0.00000582, -0.00001314, 0.00000077, 0.00001988, -0.00000761, -0.00000227, 0.00000069, -0.00000765, 0.00000267, -0.00000315, 0.00000210, 0.00000530, 0.00000110, 0.00000093, -0.00000756, 0.00000339, -0.00000043, -0.00000060, 0.00000557, 0.00000740, -0.00001352, 0.00002372, -0.00003577, 0.00002060, -0.00001920, 0.00001554, -0.00000950, 0.00001017, -0.00001542, 0.00000170, -0.00001727, 0.00003214, -0.00004033, 0.00004615, -0.00003446, 0.00002168, -0.00002283, 0.00002953, -0.00002630, 0.00001685, -0.00002623, 0.00003279, -0.00003508, 0.00003480, -0.00004589, 0.00004379, -0.00005287, 0.00004741, -0.00005712, 0.00005829, -0.00009106, 0.00008947, -0.00014101, 0.00012493, -0.00038041, 0.00011319, 0.00093268, -0.00061301, 0.00015896, -0.00017609, 0.00008919, -0.00009226, 0.00006923, -0.00006414, 0.00003988, -0.00004277, 0.00003286, -0.00003726, 0.00003339, -0.00003564, 0.00002959, -0.00002083, 0.00001655, -0.00002100, 0.00001501, -0.00001230, 0.00000936, -0.00001302, 0.00000584, 0.00000033, 0.00000482, -0.00001702, 0.00001250, -0.00002815, 0.00003205, -0.00001410, 0.00002243, -0.00000798, 0.00002326, -0.00003365, 0.00003479, -0.00004007, 0.00002958, -0.00002226, 0.00001485, -0.00001667, 0.00001492, -0.00001582, 0.00001768, -0.00001975, 0.00001219, -0.00000806, 0.00001085, -0.00000699, 0.00000474, -0.00000938, 0.00000501, -0.00001004, 0.00000544, },
		{ 0.00000134, -0.00001861, -0.00000620, 0.00001645, 0.00000255, -0.00001513, 0.00000451, 0.00000362, -0.00000627, 0.00000308, -0.00000143, 0.00000022, -0.00000303, 0.00001590, -0.00001397, -0.00000186, -0.00000308, 0.00001568, 0.00000239, -0.00001656, -0.00000247, 0.00001803, 0.00000114, -0.00000987, 0.00000654, -0.00001146, 0.00000423, -0.00000406, 0.00000186, 0.00000440, 0.00000246, 0.00000079, -0.00000577, -0.00000002, 0.00000187, -0.00000089, 0.00000106, 0.00001381, -0.00001778, 0.00002371, -0.00002597, 0.00000518, -0.00000408, 0.00000080, 0.00000340, -0.00000016, -0.00000375, -0.00000629, -0.00001447, 0.00002619, -0.00002789, 0.00002549, -0.00000808, -0.00000343, -0.00000081, 0.00000614, -0.00000053, -0.00000653, -0.00000507, 0.00000930, -0.00000822, 0.00000607, -0.00001323, 0.00000672, -0.00001113, 0.00000163, -0.00000811, 0.00000500, -0.00002548, 0.00000834, -0.00003299, -0.00001348, -0.00011272, -0.00033616, 0.00115802, -0.00047138, -0.00000425, -0.00007261, 0.00000332, -0.00002675, 0.00001113, -0.00001308, -0.00000284, -0.00000802, 0.00000291, -0.00001106, 0.00000836, -0.00001095, 0.00000435, 0.00000186, -0.00000145, -0.00000504, -0.00000030, 0.00000067, -0.00000039, -0.00000447, -0.00000184, 0.00000347, 0.00000664, -0.00001856, 0.00001597, -0.00003124, 0.00002771, -0.00001032, 0.00001998, -0.00000938, 0.00002739, -0.00003360, 0.00002880, -0.00002702, 0.00001254, -0.00000568, 0.00000164, -0.00000580, 0.00000521, -0.00000678, 0.00000775, -0.00000775, -0.00000021, 0.00000238, 0.00000147, 0.00000182, -0.00000195, -0.00000306, -0.00000152, -0.00000277, -0.00000453, 0.00001918, },
		{ -0.00001238, -0.00001619, 0.00001846, 0.00000436, -0.00001050, -0.00000533, 0.00001187, -0.00001046, 0.00000526, -0.00000374, 0.00000390, -0.00000797, 0.00001667, -0.00000781, -0.00000736, -0.00000144, 0.00000952, 0.00001198, -0.00001915, -0.00000457, 0.00001257, 0.00001196, -0.00001690, 0.00001068, -0.00001328, 0.00000429, -0.00000432, 0.00000197, 0.00000238, 0.00000519, -0.00000137, -0.00000151, -0.00000526, 0.00000537, -0.00000264, 0.00000021, 0.00001270, -0.00001113, 0.00001066, -0.00000434, -0.00001784, 0.00001512, -0.00001616, 0.00001770, -0.00001231, 0.00000962, -0.00001579, -0.00000715, 0.00001190, -0.00000559, -0.00000394, 0.00002376, -0.00003113, 0.00002346, -0.00001992, 0.00002705, -0.00003139, 0.00001921, -0.00001926, 0.00002450, -0.00002901, 0.00002616, -0.00003685, 0.00003658, -0.00004963, 0.00004686, -0.00005633, 0.00004971, -0.00008246, 0.00008222, -0.00015571, 0.00012305, -0.00062061, 0.00109848, -0.00010223, -0.00027312, 0.00008373, -0.00011807, 0.00006437, -0.00006697, 0.00005264, -0.00005682, 0.00003680, -0.00003688, 0.00002495, -0.00002619, 0.00002220, -0.00002731, 0.00002872, -0.00002294, 0.00001472, -0.00001914, 0.00001653, -0.00001323, 0.00000768, -0.00001284, 0.00000984, 0.00000358, -0.00001365, 0.00001124, -0.00002324, 0.00001219, 0.00000309, 0.00000821, -0.00000016, 0.00001768, -0.00001703, 0.00000606, -0.00000001, -0.00001401, 0.00001656, -0.00001541, 0.00000878, -0.00000854, 0.00000676, -0.00000669, 0.00000771, -0.00001454, 0.00001382, -0.00000859, 0.00001107, -0.00000933, 0.00000463, -0.00000972, 0.00000605, -0.00001436, 0.00002235, 0.00000317, },
		{ -0.00002370, 0.00001708, 0.00000667, -0.00000493, -0.00001367, 0.00001568, -0.00000968, 0.00000355, -0.00000298, 0.00000420, -0.00000779, 0.00001118, 0.00000247, -0.00001399, 0.00000139, 0.00000192, 0.00002032, -0.00001907, -0.00000668, 0.00000564, 0.00002140, -0.00001994, 0.00001053, -0.00001135, 0.00000172, -0.00000308, 0.00000153, 0.00000043, 0.00000778, -0.00000367, 0.00000280, -0.00000973, 0.00000790, -0.00000438, 0.00000183, 0.00000628, 0.00000176, -0.00000743, 0.00001809, -0.00003621, 0.00002678, -0.00002494, 0.00002407, -0.00001747, 0.00001618, -0.00001911, -0.00000262, -0.00000208, 0.00001552, -0.00002888, 0.00004679, -0.00004745, 0.00003652, -0.00003501, 0.00004263, -0.00004404, 0.00003266, -0.00003766, 0.00004604, -0.00005154, 0.00005136, -0.00006395, 0.00006526, -0.00007930, 0.00007816, -0.00009197, 0.00009464, -0.00013489, 0.00014580, -0.00022621, 0.00023221, -0.00066259, 0.00075674, 0.00041178, -0.00054913, 0.00021532, -0.00021543, 0.00013294, -0.00012481, 0.00009891, -0.00009333, 0.00006712, -0.00006476, 0.00005089, -0.00005162, 0.00004620, -0.00004906, 0.00004535, -0.00003556, 0.00002740, -0.00003163, 0.00002658, -0.00002161, 0.00001667, -0.00002104, 0.00001441, -0.00000025, -0.00000580, 0.00000147, -0.00000801, -0.00000940, 0.00002074, -0.00000796, 0.00001496, -0.00000020, 0.00000805, -0.00002251, 0.00002807, -0.00003708, 0.00003291, -0.00002665, 0.00001885, -0.00001881, 0.00001743, -0.00001807, 0.00001910, -0.00002380, 0.00002011, -0.00001402, 0.00001616, -0.00001338, 0.00000983, -0.00001556, 0.00001208, -0.00001985, 0.00002086, 0.00000638, -0.00000595, },
		{ 0.00001035, 0.00001121, -0.00000136, -0.00001685, 0.00001241, -0.00000291, -0.00000221, 0.00000092, 0.00000100, -0.00000309, 0.00000175, 0.00001323, -0.00001843, 0.00000270, -0.00000406, 0.00002437, -0.00001445, -0.00001022, -0.00000010, 0.00002604, -0.00001637, 0.00000467, -0.00000506, -0.00000372, 0.00000003, -0.00000029, -0.00000002, 0.00000832, -0.00000373, 0.00000435, -0.00001055, 0.00000691, -0.00000374, 0.00000286, -0.00000029, 0.00001300, -0.00002009, 0.00002962, -0.00003927, 0.00002249, -0.00001903, 0.00001715, -0.00001073, 0.00001128, -0.00001215, -0.00000540, -0.00000912, 0.00002644, -0.00003903, 0.00005086, -0.00004343, 0.00003025, -0.00003100, 0.00003801, -0.00003645, 0.00002670, -0.00003585, 0.00004482, -0.00004867, 0.00004842, -0.00005968, 0.00005922, -0.00007083, 0.00006787, -0.00008155, 0.00008558, -0.00012160, 0.00012701, -0.00019079, 0.00018604, -0.00048014, 0.00023820, 0.00091475, -0.00070069, 0.00023543, -0.00022886, 0.00013344, -0.00012609, 0.00009621, -0.00008816, 0.00006230, -0.00006229, 0.00004974, -0.00005202, 0.00004629, -0.00004776, 0.00004062, -0.00002997, 0.00002459, -0.00002994, 0.00002374, -0.00001913, 0.00001634, -0.00002065, 0.00001203, -0.00000006, 0.00000002, -0.00000774, 0.00000729, -0.00002844, 0.00003379, -0.00001990, 0.00002695, -0.00001569, 0.00002876, -0.00004230, 0.00004266, -0.00004410, 0.00003351, -0.00002451, 0.00001800, -0.00001951, 0.00001906, -0.00002017, 0.00002054, -0.00002299, 0.00001715, -0.00001140, 0.00001416, -0.00001134, 0.00000956, -0.00001573, 0.00001184, -0.00001802, 0.00001296, 0.00001237, -0.00000182, -0.00002583, },
		{ 0.00001765, -0.00000059, -0.00001416, 0.00000297, 0.00000743, -0.00000929, 0.00000560, -0.00000339, 0.00000301, -0.00000723, 0.00001990, -0.00001737, 0.00000015, -0.00000600, 0.00002226, -0.00000479, -0.00001562, -0.00000307, 0.00002422, -0.00000596, -0.00000584, 0.00000406, -0.00001068, 0.00000419, -0.00000338, 0.00000156, 0.00000588, -0.00000035, 0.00000176, -0.00000653, 0.00000145, 0.00000049, 0.00000086, -0.00000258, 0.00001628, -0.00002036, 0.00002445, -0.00002412, 0.00000239, -0.00000034, -0.00000054, 0.00000537, -0.00000257, 0.00000259, -0.00001415, -0.00000832, 0.00002467, -0.00003192, 0.00003497, -0.00001995, 0.00000651, -0.00000985, 0.00001535, -0.00001102, 0.00000325, -0.00001479, 0.00002125, -0.00002091, 0.00001825, -0.00002562, 0.00002091, -0.00002760, 0.00002010, -0.00002951, 0.00002775, -0.00005026, 0.00003846, -0.00007109, 0.00002923, -0.00017560, -0.00028658, 0.00123280, -0.00061853, 0.00009324, -0.00013234, 0.00004984, -0.00006027, 0.00003753, -0.00003664, 0.00001857, -0.00002584, 0.00001787, -0.00002389, 0.00001951, -0.00002145, 0.00001375, -0.00000592, 0.00000575, -0.00001320, 0.00000713, -0.00000475, 0.00000520, -0.00000995, 0.00000104, 0.00000627, 0.00000062, -0.00001171, 0.00001602, -0.00003706, 0.00003458, -0.00002018, 0.00002776, -0.00001934, 0.00003439, -0.00004273, 0.00003535, -0.00002991, 0.00001611, -0.00000825, 0.00000571, -0.00000974, 0.00001037, -0.00001163, 0.00001116, -0.00001196, 0.00000541, -0.00000140, 0.00000563, -0.00000344, 0.00000358, -0.00000961, 0.00000472, -0.00000898, -0.00000021, 0.00002029, -0.00000051, -0.00002196, -0.00000126, },
		{ -0.00000078, -0.00000806, -0.00000849, 0.00001645, -0.00001338, 0.00000757, -0.00000568, 0.00000668, -0.00001154, 0.00001926, -0.00000958, -0.00000653, -0.00000366, 0.00001457, 0.00000812, -0.00002141, -0.00000369, 0.00001696, 0.00000849, -0.00001733, 0.00001248, -0.00001626, 0.00000733, -0.00000635, 0.00000416, 0.00000140, 0.00000544, -0.00000376, 0.00000072, -0.00000664, 0.00000688, -0.00000385, 0.00000030, 0.00001060, -0.00000862, 0.00000589, 0.00000224, -0.00002407, 0.00002112, -0.00001935, 0.00002185, -0.00001685, 0.00001706, -0.00002234, -0.00000435, 0.00001464, -0.00001347, 0.00000746, 0.00001255, -0.00002347, 0.00001721, -0.00001375, 0.00002023, -0.00002586, 0.00001387, -0.00001232, 0.00001791, -0.00002357, 0.00002068, -0.00002995, 0.00002840, -0.00004081, 0.00003710, -0.00004758, 0.00004114, -0.00007173, 0.00006956, -0.00014692, 0.00011157, -0.00064824, 0.00124372, -0.00027085, -0.00019031, 0.00004373, -0.00008934, 0.00004680, -0.00005370, 0.00004027, -0.00004589, 0.00002896, -0.00003097, 0.00002021, -0.00002220, 0.00001814, -0.00002385, 0.00002619, -0.00002025, 0.00001097, -0.00001640, 0.00001558, -0.00001184, 0.00000681, -0.00001481, 0.00001619, -0.00000314, -0.00000953, 0.00001555, -0.00003225, 0.00002096, -0.00000714, 0.00001534, -0.00000845, 0.00002240, -0.00002317, 0.00000879, -0.00000029, -0.00001188, 0.00001472, -0.00001170, 0.00000531, -0.00000413, 0.00000313, -0.00000426, 0.00000411, -0.00000995, 0.00001121, -0.00000525, 0.00000679, -0.00000522, 0.00000013, -0.00000632, 0.00000357, -0.00001445, 0.00002689, -0.00000017, -0.00001430, -0.00001454, 0.00002323, },
		{ -0.00000250, -0.00001678, 0.00001931, -0.00001108, 0.00000451, -0.00000379, 0.00000572, -0.00000942, 0.00001136, 0.00000292, -0.00001485, 0.00000070, 0.00000439, 0.00002043, -0.00002456, -0.00000415, 0.00000758, 0.00002210, -0.00002474, 0.00001600, -0.00001723, 0.00000701, -0.00000714, 0.00000570, -0.00000265, 0.00001073, -0.00000881, 0.00000729, -0.00001329, 0.00001177, -0.00000818, 0.00000523, 0.00000046, 0.00000792, -0.00001544, 0.00002646, -0.00004289, 0.00003244, -0.00002788, 0.00002842, -0.00002198, 0.00002243, -0.00002255, -0.00000376, 0.00000408, 0.00000597, -0.00001865, 0.00003943, -0.00004489, 0.00003554, -0.00003429, 0.00004186, -0.00004509, 0.00003409, -0.00003820, 0.00004792, -0.00005500, 0.00005494, -0.00006675, 0.00006811, -0.00008314, 0.00008314, -0.00010014, 0.00010495, -0.00014609, 0.00016044, -0.00025126, 0.00026605, -0.00074535, 0.00093031, 0.00026904, -0.00051621, 0.00021670, -0.00021862, 0.00014119, -0.00013205, 0.00010385, -0.00009810, 0.00007349, -0.00007127, 0.00005698, -0.00005709, 0.00005059, -0.00005332, 0.00004971, -0.00003908, 0.00002978, -0.00003510, 0.00003129, -0.00002540, 0.00002098, -0.00002768, 0.00002335, -0.00000663, -0.00000444, 0.00000842, -0.00001739, -0.00000207, 0.00001353, -0.00000464, 0.00001112, -0.00000004, 0.00000653, -0.00002423, 0.00003036, -0.00003649, 0.00003229, -0.00002424, 0.00001708, -0.00001647, 0.00001619, -0.00001767, 0.00001726, -0.00002149, 0.00001984, -0.00001273, 0.00001428, -0.00001217, 0.00000878, -0.00001612, 0.00001392, -0.00002411, 0.00002798, 0.00000223, -0.00000664, -0.00002489, 0.00002417, 0.00000108, },
		{ -0.00001852, 0.00001390, -0.00000212, -0.00000294, 0.00000158, 0.00000085, -0.00000256, -0.00000029, 0.00001541, -0.00002053, 0.00000337, -0.00000404, 0.00002776, -0.00002197, -0.00000699, -0.00000005, 0.00002998, -0.00002393, 0.00001183, -0.00001183, 0.00000193, -0.00000438, 0.00000430, -0.00000371, 0.00001234, -0.00000973, 0.00000912, -0.00001445, 0.00001137, -0.00000832, 0.00000748, -0.00000722, 0.00001959, -0.00002763, 0.00003598, -0.00004313, 0.00002530, -0.00001976, 0.00001989, -0.00001331, 0.00001462, -0.00001159, -0.00001009, -0.00000111, 0.00001760, -0.00003248, 0.00004927, -0.00004704, 0.00003500, -0.00003589, 0.00004321, -0.00004370, 0.00003427, -0.00004305, 0.00005392, -0.00005933, 0.00005918, -0.00007025, 0.00007100, -0.00008482, 0.00008440, -0.00010234, 0.00010923, -0.00014782, 0.00015965, -0.00023626, 0.00024377, -0.00058209, 0.00039246, 0.00083891, -0.00073921, 0.00028532, -0.00026542, 0.00016682, -0.00015238, 0.00011699, -0.00010709, 0.00008100, -0.00007911, 0.00006455, -0.00006496, 0.00005733, -0.00005813, 0.00005046, -0.00003852, 0.00003202, -0.00003842, 0.00003273, -0.00002661, 0.00002411, -0.00002976, 0.00002122, -0.00000442, -0.00000152, 0.00000036, -0.00000023, -0.00002473, 0.00003139, -0.00002163, 0.00002817, -0.00002006, 0.00003124, -0.00004718, 0.00004626, -0.00004445, 0.00003414, -0.00002383, 0.00001854, -0.00001980, 0.00002058, -0.00002206, 0.00002089, -0.00002341, 0.00001960, -0.00001259, 0.00001516, -0.00001339, 0.00001205, -0.00001992, 0.00001707, -0.00002489, 0.00002087, 0.00000874, -0.00000214, -0.00002865, 0.00001795, 0.00000700, -0.00000046, },
		{ 0.00000220, 0.00001005, -0.00001115, 0.00000722, -0.00000469, 0.00000492, -0.00001033, 0.00002252, -0.00001966, 0.00000137, -0.00000747, 0.00002737, -0.00001244, -0.00001322, -0.00000359, 0.00002957, -0.00001389, 0.00000043, -0.00000118, -0.00000682, 0.00000142, -0.00000027, -0.00000082, 0.00000888, -0.00000490, 0.00000453, -0.00000865, 0.00000441, -0.00000264, 0.00000411, -0.00000756, 0.00002001, -0.00002412, 0.00002597, -0.00002327, 0.00000124, 0.00000188, -0.00000018, 0.00000556, -0.00000266, 0.00000679, -0.00002102, -0.00000039, 0.00001855, -0.00003045, 0.00003935, -0.00002879, 0.00001534, -0.00001831, 0.00002412, -0.00002164, 0.00001377, -0.00002505, 0.00003343, -0.00003417, 0.00003131, -0.00003884, 0.00003598, -0.00004526, 0.00004051, -0.00005378, 0.00005407, -0.00007902, 0.00007373, -0.00011722, 0.00008415, -0.00026053, -0.00018950, 0.00124540, -0.00072775, 0.00017984, -0.00018939, 0.00009664, -0.00009540, 0.00006547, -0.00006192, 0.00004227, -0.00004613, 0.00003520, -0.00003877, 0.00003231, -0.00003352, 0.00002508, -0.00001586, 0.00001481, -0.00002320, 0.00001689, -0.00001254, 0.00001303, -0.00001809, 0.00000747, 0.00000573, -0.00000397, -0.00000353, 0.00001150, -0.00003754, 0.00003682, -0.00002609, 0.00003257, -0.00002632, 0.00003868, -0.00004852, 0.00003862, -0.00003048, 0.00001780, -0.00000935, 0.00000853, -0.00001238, 0.00001412, -0.00001525, 0.00001330, -0.00001474, 0.00001002, -0.00000452, 0.00000894, -0.00000795, 0.00000845, -0.00001586, 0.00001137, -0.00001618, 0.00000639, 0.00001859, -0.00000146, -0.00002508, 0.00000512, 0.00001617, -0.00000201, -0.00001384, },
		{ 0.00001977, -0.00001528, 0.00000921, -0.00000715, 0.00000877, -0.00001429, 0.00002103, -0.00001106, -0.00000564, -0.00000545, 0.00001970, 0.00000218, -0.00002122, -0.00000350, 0.00002189, 0.00000245, -0.00001409, 0.00001064, -0.00001576, 0.00000759, -0.00000610, 0.00000450, 0.00000181, 0.00000401, -0.00000447, 0.00000168, -0.00000653, 0.00000674, -0.00000376, -0.00000078, 0.00000965, -0.00000723, 0.00000205, 0.00000742, -0.00002836, 0.00002548, -0.00002087, 0.00002441, -0.00001962, 0.00002348, -0.00002946, 0.00000166, 0.00001246, -0.00001710, 0.00001637, 0.00000105, -0.00001370, 0.00000856, -0.00000510, 0.00001032, -0.00001677, 0.00000525, -0.00000216, 0.00000749, -0.00001381, 0.00001069, -0.00001810, 0.00001448, -0.00002508, 0.00001917, -0.00002959, 0.00002221, -0.00004843, 0.00004033, -0.00011630, 0.00006829, -0.00062562, 0.00134032, -0.00042672, -0.00009492, -0.00001074, -0.00004756, 0.00001775, -0.00003095, 0.00001945, -0.00002699, 0.00001389, -0.00001871, 0.00001001, -0.00001346, 0.00000961, -0.00001582, 0.00001929, -0.00001386, 0.00000370, -0.00000995, 0.00001125, -0.00000749, 0.00000288, -0.00001356, 0.00002052, -0.00001046, -0.00000210, 0.00001426, -0.00003618, 0.00002626, -0.00001503, 0.00002117, -0.00001539, 0.00002602, -0.00002782, 0.00001057, -0.00000074, -0.00000930, 0.00001221, -0.00000730, 0.00000133, 0.00000067, -0.00000107, -0.00000133, 0.00000008, -0.00000443, 0.00000746, -0.00000102, 0.00000146, 0.00000011, -0.00000601, -0.00000057, -0.00000169, -0.00001088, 0.00002806, -0.00000237, -0.00001675, -0.00001045, 0.00002496, -0.00000438, -0.00000619, -0.00001071, },
		{ -0.00001204, 0.00000542, -0.00000466, 0.00000704, -0.00001090, 0.00001162, 0.00000265, -0.00001471, -0.00000063, 0.00000832, 0.00001733, -0.00002720, -0.00000263, 0.00001095, 0.00001927, -0.00002555, 0.00001814, -0.00002039, 0.00001050, -0.00000994, 0.00000897, -0.00000510, 0.00001265, -0.00001299, 0.00001114, -0.00001595, 0.00001466, -0.00001136, 0.00000805, -0.00000471, 0.00001334, -0.00002274, 0.00003355, -0.00004802, 0.00003712, -0.00002990, 0.00003212, -0.00002591, 0.00002915, -0.00002804, -0.00000092, 0.00000602, -0.00000152, -0.00000783, 0.00002864, -0.00003768, 0.00003015, -0.00002927, 0.00003633, -0.00004116, 0.00003089, -0.00003403, 0.00004434, -0.00005250, 0.00005233, -0.00006277, 0.00006325, -0.00007807, 0.00007810, -0.00009733, 0.00010311, -0.00014287, 0.00015708, -0.00025421, 0.00027100, -0.00079070, 0.00108274, 0.00010680, -0.00044714, 0.00018953, -0.00019968, 0.00013173, -0.00012519, 0.00009668, -0.00009205, 0.00007049, -0.00006978, 0.00005627, -0.00005657, 0.00004936, -0.00005211, 0.00004915, -0.00003852, 0.00002832, -0.00003477, 0.00003284, -0.00002661, 0.00002286, -0.00003227, 0.00003205, -0.00001543, 0.00000131, 0.00001012, -0.00002343, 0.00000424, 0.00000608, -0.00000023, 0.00000638, 0.00000126, 0.00000407, -0.00002416, 0.00002995, -0.00003329, 0.00002912, -0.00001946, 0.00001325, -0.00001219, 0.00001284, -0.00001523, 0.00001341, -0.00001671, 0.00001707, -0.00000927, 0.00001002, -0.00000838, 0.00000474, -0.00001316, 0.00001210, -0.00002440, 0.00003213, -0.00000119, -0.00000808, -0.00002338, 0.00002866, -0.00000375, -0.00000016, -0.00001922, 0.00002206, },
		{ -0.00000294, 0.00000154, 0.00000115, -0.00000264, -0.00000135, 0.00001600, -0.00002100, 0.00000274, -0.00000179, 0.00002768, -0.00002713, -0.00000437, 0.00000156, 0.00003049, -0.00002843, 0.00001714, -0.00001761, 0.00000754, -0.00000911, 0.00000937, -0.00000791, 0.00001639, -0.00001588, 0.00001428, -0.00001841, 0.00001590, -0.00001331, 0.00001240, -0.00001426, 0.00002618, -0.00003512, 0.00004179, -0.00004655, 0.00002844, -0.00002092, 0.00002338, -0.00001682, 0.00001995, -0.00001418, -0.00001075, 0.00000413, 0.00000834, -0.00002273, 0.00004229, -0.00004506, 0.00003509, -0.00003649, 0.00004390, -0.00004652, 0.00003790, -0.00004612, 0.00005820, -0.00006491, 0.00006477, -0.00007519, 0.00007642, -0.00009166, 0.00009321, -0.00011504, 0.00012419, -0.00016391, 0.00018016, -0.00026817, 0.00028604, -0.00066988, 0.00055842, 0.00071443, -0.00072671, 0.00030208, -0.00027907, 0.00018318, -0.00016579, 0.00012704, -0.00011677, 0.00009207, -0.00008965, 0.00007414, -0.00007330, 0.00006405, -0.00006439, 0.00005687, -0.00004442, 0.00003696, -0.00004449, 0.00004005, -0.00003299, 0.00003098, -0.00003863, 0.00003207, -0.00001252, 0.00000137, 0.00000474, -0.00000706, -0.00001917, 0.00002614, -0.00002011, 0.00002659, -0.00002145, 0.00003083, -0.00004832, 0.00004567, -0.00004117, 0.00003144, -0.00002019, 0.00001646, -0.00001751, 0.00001933, -0.00002129, 0.00001859, -0.00002083, 0.00001918, -0.00001126, 0.00001340, -0.00001252, 0.00001131, -0.00002061, 0.00001907, -0.00002875, 0.00002723, 0.00000473, -0.00000280, -0.00002931, 0.00002416, 0.00000224, 0.00000120, -0.00002013, 0.00001531, -0.00000177, },
		{ 0.00000767, -0.00000508, 0.00000569, -0.00001193, 0.00002326, -0.00002030, 0.00000128, -0.00000677, 0.00002974, -0.00001904, -0.00001041, -0.00000315, 0.00003257, -0.00002065, 0.00000709, -0.00000759, -0.00000107, -0.00000328, 0.00000464, -0.00000481, 0.00001296, -0.00001071, 0.00000874, -0.00001184, 0.00000850, -0.00000717, 0.00000853, -0.00001349, 0.00002460, -0.00002873, 0.00002807, -0.00002348, 0.00000195, 0.00000227, 0.00000225, 0.00000357, 0.00000029, 0.00000748, -0.00002461, 0.00000663, 0.00000984, -0.00002426, 0.00003807, -0.00003299, 0.00002108, -0.00002425, 0.00003038, -0.00003010, 0.00002272, -0.00003356, 0.00004340, -0.00004532, 0.00004248, -0.00005001, 0.00004872, -0.00006044, 0.00005864, -0.00007619, 0.00007879, -0.00010599, 0.00010728, -0.00016320, 0.00014157, -0.00035627, -0.00005304, 0.00119300, -0.00078647, 0.00024183, -0.00023263, 0.00013471, -0.00012492, 0.00008901, -0.00008368, 0.00006355, -0.00006485, 0.00005147, -0.00005273, 0.00004415, -0.00004472, 0.00003609, -0.00002598, 0.00002388, -0.00003323, 0.00002733, -0.00002136, 0.00002197, -0.00002796, 0.00001717, 0.00000090, -0.00000501, 0.00000315, 0.00000491, -0.00003382, 0.00003457, -0.00002775, 0.00003387, -0.00002973, 0.00003951, -0.00005022, 0.00003794, -0.00002805, 0.00001688, -0.00000819, 0.00000934, -0.00001295, 0.00001561, -0.00001676, 0.00001334, -0.00001515, 0.00001256, -0.00000588, 0.00001020, -0.00001033, 0.00001105, -0.00001987, 0.00001638, -0.00002232, 0.00001353, 0.00001503, -0.00000193, -0.00002714, 0.00001179, 0.00001268, -0.00000182, -0.00001417, 0.00000222, 0.00001136, -0.00001163, },
		{ -0.00000770, 0.00000968, -0.00001564, 0.00002132, -0.00001157, -0.00000538, -0.00000591, 0.00002356, -0.00000450, -0.00001934, -0.00000332, 0.00002606, -0.00000466, -0.00000825, 0.00000575, -0.00001193, 0.00000491, -0.00000329, 0.00000262, 0.00000396, 0.00000058, -0.00000310, 0.00000093, -0.00000459, 0.00000458, -0.00000196, -0.00000339, 0.00001025, -0.00000749, -0.00000018, 0.00001028, -0.00002972, 0.00002723, -0.00001968, 0.00002434, -0.00001927, 0.00002701, -0.00003482, 0.00000878, 0.00000617, -0.00001581, 0.00002089, -0.00000818, -0.00000427, -0.00000031, 0.00000391, -0.00000048, -0.00000616, -0.00000474, 0.00000917, -0.00000462, -0.00000186, -0.00000162, -0.00000374, -0.00000241, -0.00000548, -0.00000372, -0.00000560, -0.00000363, -0.00001645, -0.00000125, -0.00006797, -0.00000379, -0.00054977, 0.00137421, -0.00054977, -0.00000379, -0.00006797, -0.00000125, -0.00001645, -0.00000363, -0.00000560, -0.00000372, -0.00000548, -0.00000241, -0.00000374, -0.00000162, -0.00000186, -0.00000462, 0.00000917, -0.00000474, -0.00000616, -0.00000048, 0.00000391, -0.00000031, -0.00000427, -0.00000818, 0.00002089, -0.00001581, 0.00000617, 0.00000878, -0.00003482, 0.00002701, -0.00001927, 0.00002434, -0.00001968, 0.00002723, -0.00002972, 0.00001028, -0.00000018, -0.00000749, 0.00001025, -0.00000339, -0.00000196, 0.00000458, -0.00000459, 0.00000093, -0.00000310, 0.00000058, 0.00000396, 0.00000262, -0.00000329, 0.00000491, -0.00001193, 0.00000575, -0.00000825, -0.00000466, 0.00002606, -0.00000332, -0.00001934, -0.00000450, 0.00002356, -0.00000591, -0.00000538, -0.00001157, 0.00002132, -0.00001564, 0.00000968, },
		{ 0.00000767, -0.00001163, 0.00001136, 0.00000222, -0.00001417, -0.00000182, 0.00001268, 0.00001179, -0.00002714, -0.00000193, 0.00001503, 0.00001353, -0.00002232, 0.00001638, -0.00001987, 0.00001105, -0.00001033, 0.00001020, -0.00000588, 0.00001256, -0.00001515, 0.00001334, -0.00001676, 0.00001561, -0.00001295, 0.00000934, -0.00000819, 0.00001688, -0.00002805, 0.00003794, -0.00005022, 0.00003951, -0.00002973, 0.00003387, -0.00002775, 0.00003457, -0.00003382, 0.00000491, 0.00000315, -0.00000501, 0.00000090, 0.00001717, -0.00002796, 0.00002197, -0.00002136, 0.00002733, -0.00003323, 0.00002388, -0.00002598, 0.00003609, -0.00004472, 0.00004415, -0.00005273, 0.00005147, -0.00006485, 0.00006355, -0.00008368, 0.00008901, -0.00012492, 0.00013471, -0.00023263, 0.00024183, -0.00078647, 0.00119300, -0.00005304, -0.00035627, 0.00014157, -0.00016320, 0.00010728, -0.00010599, 0.00007879, -0.00007619, 0.00005864, -0.00006044, 0.00004872, -0.00005001, 0.00004248, -0.00004532, 0.00004340, -0.00003356, 0.00002272, -0.00003010, 0.00003038, -0.00002425, 0.00002108, -0.00003299, 0.00003807, -0.00002426, 0.00000984, 0.00000663, -0.00002461, 0.00000748, 0.00000029, 0.00000357, 0.00000225, 0.00000227, 0.00000195, -0.00002348, 0.00002807, -0.00002873, 0.00002460, -0.00001349, 0.00000853, -0.00000717, 0.00000850, -0.00001184, 0.00000874, -0.00001071, 0.00001296, -0.00000481, 0.00000464, -0.00000328, -0.00000107, -0.00000759, 0.00000709, -0.00002065, 0.00003257, -0.00000315, -0.00001041, -0.00001904, 0.00002974, -0.00000677, 0.00000128, -0.00002030, 0.00002326, -0.00001193, 0.00000569, -0.00000508, },
		{ -0.00000294, -0.00000177, 0.00001531, -0.00002013, 0.00000120, 0.00000224, 0.00002416, -0.00002931, -0.00000280, 0.00000473, 0.00002723, -0.00002875, 0.00001907, -0.00002061, 0.00001131, -0.00001252, 0.00001340, -0.00001126, 0.00001918, -0.00002083, 0.00001859, -0.00002129, 0.00001933, -0.00001751, 0.00001646, -0.00002019, 0.00003144, -0.00004117, 0.00004567, -0.00004832, 0.00003083, -0.00002145, 0.00002659, -0.00002011, 0.00002614, -0.00001917, -0.00000706, 0.00000474, 0.00000137, -0.00001252, 0.00003207, -0.00003863, 0.00003098, -0.00003299, 0.00004005, -0.00004449, 0.00003696, -0.00004442, 0.00005687, -0.00006439, 0.00006405, -0.00007330, 0.00007414, -0.00008965, 0.00009207, -0.00011677, 0.00012704, -0.00016579, 0.00018318, -0.00027907, 0.00030208, -0.00072671, 0.00071443, 0.00055842, -0.00066988, 0.00028604, -0.00026817, 0.00018016, -0.00016391, 0.00012419, -0.00011504, 0.00009321, -0.00009166, 0.00007642, -0.00007519, 0.00006477, -0.00006491, 0.00005820, -0.00004612, 0.00003790, -0.00004652, 0.00004390, -0.00003649, 0.00003509, -0.00004506, 0.00004229, -0.00002273, 0.00000834, 0.00000413, -0.00001075, -0.00001418, 0.00001995, -0.00001682, 0.00002338, -0.00002092, 0.00002844, -0.00004655, 0.00004179, -0.00003512, 0.00002618, -0.00001426, 0.00001240, -0.00001331, 0.00001590, -0.00001841, 0.00001428, -0.00001588, 0.00001639, -0.00000791, 0.00000937, -0.00000911, 0.00000754, -0.00001761, 0.00001714, -0.00002843, 0.00003049, 0.00000156, -0.00000437, -0.00002713, 0.00002768, -0.00000179, 0.00000274, -0.00002100, 0.00001600, -0.00000135, -0.00000264, 0.00000115, 0.00000154, },
		{ -0.00001204, 0.00002206, -0.00001922, -0.00000016, -0.00000375, 0.00002866, -0.00002338, -0.00000808, -0.00000119, 0.00003213, -0.00002440, 0.00001210, -0.00001316, 0.00000474, -0.00000838, 0.00001002, -0.00000927, 0.00001707, -0.00001671, 0.00001341, -0.00001523, 0.00001284, -0.00001219, 0.00001325, -0.00001946, 0.00002912, -0.00003329, 0.00002995, -0.00002416, 0.00000407, 0.00000126, 0.00000638, -0.00000023, 0.00000608, 0.00000424, -0.00002343, 0.00001012, 0.00000131, -0.00001543, 0.00003205, -0.00003227, 0.00002286, -0.00002661, 0.00003284, -0.00003477, 0.00002832, -0.00003852, 0.00004915, -0.00005211, 0.00004936, -0.00005657, 0.00005627, -0.00006978, 0.00007049, -0.00009205, 0.00009668, -0.00012519, 0.00013173, -0.00019968, 0.00018953, -0.00044714, 0.00010680, 0.00108274, -0.00079070, 0.00027100, -0.00025421, 0.00015708, -0.00014287, 0.00010311, -0.00009733, 0.00007810, -0.00007807, 0.00006325, -0.00006277, 0.00005233, -0.00005250, 0.00004434, -0.00003403, 0.00003089, -0.00004116, 0.00003633, -0.00002927, 0.00003015, -0.00003768, 0.00002864, -0.00000783, -0.00000152, 0.00000602, -0.00000092, -0.00002804, 0.00002915, -0.00002591, 0.00003212, -0.00002990, 0.00003712, -0.00004802, 0.00003355, -0.00002274, 0.00001334, -0.00000471, 0.00000805, -0.00001136, 0.00001466, -0.00001595, 0.00001114, -0.00001299, 0.00001265, -0.00000510, 0.00000897, -0.00000994, 0.00001050, -0.00002039, 0.00001814, -0.00002555, 0.00001927, 0.00001095, -0.00000263, -0.00002720, 0.00001733, 0.00000832, -0.00000063, -0.00001471, 0.00000265, 0.00001162, -0.00001090, 0.00000704, -0.00000466, 0.00000542, },
		{ 0.00001977, -0.00001071, -0.00000619, -0.00000438, 0.00002496, -0.00001045, -0.00001675, -0.00000237, 0.00002806, -0.00001088, -0.00000169, -0.00000057, -0.00000601, 0.00000011, 0.00000146, -0.00000102, 0.00000746, -0.00000443, 0.00000008, -0.00000133, -0.00000107, 0.00000067, 0.00000133, -0.00000730, 0.00001221, -0.00000930, -0.00000074, 0.00001057, -0.00002782, 0.00002602, -0.00001539, 0.00002117, -0.00001503, 0.00002626, -0.00003618, 0.00001426, -0.00000210, -0.00001046, 0.00002052, -0.00001356, 0.00000288, -0.00000749, 0.00001125, -0.00000995, 0.00000370, -0.00001386, 0.00001929, -0.00001582, 0.00000961, -0.00001346, 0.00001001, -0.00001871, 0.00001389, -0.00002699, 0.00001945, -0.00003095, 0.00001775, -0.00004756, -0.00001074, -0.00009492, -0.00042672, 0.00134032, -0.00062562, 0.00006829, -0.00011630, 0.00004033, -0.00004843, 0.00002221, -0.00002959, 0.00001917, -0.00002508, 0.00001448, -0.00001810, 0.00001069, -0.00001381, 0.00000749, -0.00000216, 0.00000525, -0.00001677, 0.00001032, -0.00000510, 0.00000856, -0.00001370, 0.00000105, 0.00001637, -0.00001710, 0.00001246, 0.00000166, -0.00002946, 0.00002348, -0.00001962, 0.00002441, -0.00002087, 0.00002548, -0.00002836, 0.00000742, 0.00000205, -0.00000723, 0.00000965, -0.00000078, -0.00000376, 0.00000674, -0.00000653, 0.00000168, -0.00000447, 0.00000401, 0.00000181, 0.00000450, -0.00000610, 0.00000759, -0.00001576, 0.00001064, -0.00001409, 0.00000245, 0.00002189, -0.00000350, -0.00002122, 0.00000218, 0.00001970, -0.00000545, -0.00000564, -0.00001106, 0.00002103, -0.00001429, 0.00000877, -0.00000715, 0.00000921, -0.00001528, },
		{ 0.00000220, -0.00001384, -0.00000201, 0.00001617, 0.00000512, -0.00002508, -0.00000146, 0.00001859, 0.00000639, -0.00001618, 0.00001137, -0.00001586, 0.00000845, -0.00000795, 0.00000894, -0.00000452, 0.00001002, -0.00001474, 0.00001330, -0.00001525, 0.00001412, -0.00001238, 0.00000853, -0.00000935, 0.00001780, -0.00003048, 0.00003862, -0.00004852, 0.00003868, -0.00002632, 0.00003257, -0.00002609, 0.00003682, -0.00003754, 0.00001150, -0.00000353, -0.00000397, 0.00000573, 0.00000747, -0.00001809, 0.00001303, -0.00001254, 0.00001689, -0.00002320, 0.00001481, -0.00001586, 0.00002508, -0.00003352, 0.00003231, -0.00003877, 0.00003520, -0.00004613, 0.00004227, -0.00006192, 0.00006547, -0.00009540, 0.00009664, -0.00018939, 0.00017984, -0.00072775, 0.00124540, -0.00018950, -0.00026053, 0.00008415, -0.00011722, 0.00007373, -0.00007902, 0.00005407, -0.00005378, 0.00004051, -0.00004526, 0.00003598, -0.00003884, 0.00003131, -0.00003417, 0.00003343, -0.00002505, 0.00001377, -0.00002164, 0.00002412, -0.00001831, 0.00001534, -0.00002879, 0.00003935, -0.00003045, 0.00001855, -0.00000039, -0.00002102, 0.00000679, -0.00000266, 0.00000556, -0.00000018, 0.00000188, 0.00000124, -0.00002327, 0.00002597, -0.00002412, 0.00002001, -0.00000756, 0.00000411, -0.00000264, 0.00000441, -0.00000865, 0.00000453, -0.00000490, 0.00000888, -0.00000082, -0.00000027, 0.00000142, -0.00000682, -0.00000118, 0.00000043, -0.00001389, 0.00002957, -0.00000359, -0.00001322, -0.00001244, 0.00002737, -0.00000747, 0.00000137, -0.00001966, 0.00002252, -0.00001033, 0.00000492, -0.00000469, 0.00000722, -0.00001115, 0.00001005, },
		{ -0.00001852, -0.00000046, 0.00000700, 0.00001795, -0.00002865, -0.00000214, 0.00000874, 0.00002087, -0.00002489, 0.00001707, -0.00001992, 0.00001205, -0.00001339, 0.00001516, -0.00001259, 0.00001960, -0.00002341, 0.00002089, -0.00002206, 0.00002058, -0.00001980, 0.00001854, -0.00002383, 0.00003414, -0.00004445, 0.00004626, -0.00004718, 0.00003124, -0.00002006, 0.00002817, -0.00002163, 0.00003139, -0.00002473, -0.00000023, 0.00000036, -0.00000152, -0.00000442, 0.00002122, -0.00002976, 0.00002411, -0.00002661, 0.00003273, -0.00003842, 0.00003202, -0.00003852, 0.00005046, -0.00005813, 0.00005733, -0.00006496, 0.00006455, -0.00007911, 0.00008100, -0.00010709, 0.00011699, -0.00015238, 0.00016682, -0.00026542, 0.00028532, -0.00073921, 0.00083891, 0.00039246, -0.00058209, 0.00024377, -0.00023626, 0.00015965, -0.00014782, 0.00010923, -0.00010234, 0.00008440, -0.00008482, 0.00007100, -0.00007025, 0.00005918, -0.00005933, 0.00005392, -0.00004305, 0.00003427, -0.00004370, 0.00004321, -0.00003589, 0.00003500, -0.00004704, 0.00004927, -0.00003248, 0.00001760, -0.00000111, -0.00001009, -0.00001159, 0.00001462, -0.00001331, 0.00001989, -0.00001976, 0.00002530, -0.00004313, 0.00003598, -0.00002763, 0.00001959, -0.00000722, 0.00000748, -0.00000832, 0.00001137, -0.00001445, 0.00000912, -0.00000973, 0.00001234, -0.00000371, 0.00000430, -0.00000438, 0.00000193, -0.00001183, 0.00001183, -0.00002393, 0.00002998, -0.00000005, -0.00000699, -0.00002197, 0.00002776, -0.00000404, 0.00000337, -0.00002053, 0.00001541, -0.00000029, -0.00000256, 0.00000085, 0.00000158, -0.00000294, -0.00000212, 0.00001390, },
		{ -0.00000250, 0.00000108, 0.00002417, -0.00002489, -0.00000664, 0.00000223, 0.00002798, -0.00002411, 0.00001392, -0.00001612, 0.00000878, -0.00001217, 0.00001428, -0.00001273, 0.00001984, -0.00002149, 0.00001726, -0.00001767, 0.00001619, -0.00001647, 0.00001708, -0.00002424, 0.00003229, -0.00003649, 0.00003036, -0.00002423, 0.00000653, -0.00000004, 0.00001112, -0.00000464, 0.00001353, -0.00000207, -0.00001739, 0.00000842, -0.00000444, -0.00000663, 0.00002335, -0.00002768, 0.00002098, -0.00002540, 0.00003129, -0.00003510, 0.00002978, -0.00003908, 0.00004971, -0.00005332, 0.00005059, -0.00005709, 0.00005698, -0.00007127, 0.00007349, -0.00009810, 0.00010385, -0.00013205, 0.00014119, -0.00021862, 0.00021670, -0.00051621, 0.00026904, 0.00093031, -0.00074535, 0.00026605, -0.00025126, 0.00016044, -0.00014609, 0.00010495, -0.00010014, 0.00008314, -0.00008314, 0.00006811, -0.00006675, 0.00005494, -0.00005500, 0.00004792, -0.00003820, 0.00003409, -0.00004509, 0.00004186, -0.00003429, 0.00003554, -0.00004489, 0.00003943, -0.00001865, 0.00000597, 0.00000408, -0.00000376, -0.00002255, 0.00002243, -0.00002198, 0.00002842, -0.00002788, 0.00003244, -0.00004289, 0.00002646, -0.00001544, 0.00000792, 0.00000046, 0.00000523, -0.00000818, 0.00001177, -0.00001329, 0.00000729, -0.00000881, 0.00001073, -0.00000265, 0.00000570, -0.00000714, 0.00000701, -0.00001723, 0.00001600, -0.00002474, 0.00002210, 0.00000758, -0.00000415, -0.00002456, 0.00002043, 0.00000439, 0.00000070, -0.00001485, 0.00000292, 0.00001136, -0.00000942, 0.00000572, -0.00000379, 0.00000451, -0.00001108, 0.00001931, -0.00001678, },
		{ -0.00000078, 0.00002323, -0.00001454, -0.00001430, -0.00000017, 0.00002689, -0.00001445, 0.00000357, -0.00000632, 0.00000013, -0.00000522, 0.00000679, -0.00000525, 0.00001121, -0.00000995, 0.00000411, -0.00000426, 0.00000313, -0.00000413, 0.00000531, -0.00001170, 0.00001472, -0.00001188, -0.00000029, 0.00000879, -0.00002317, 0.00002240, -0.00000845, 0.00001534, -0.00000714, 0.00002096, -0.00003225, 0.00001555, -0.00000953, -0.00000314, 0.00001619, -0.00001481, 0.00000681, -0.00001184, 0.00001558, -0.00001640, 0.00001097, -0.00002025, 0.00002619, -0.00002385, 0.00001814, -0.00002220, 0.00002021, -0.00003097, 0.00002896, -0.00004589, 0.00004027, -0.00005370, 0.00004680, -0.00008934, 0.00004373, -0.00019031, -0.00027085, 0.00124372, -0.00064824, 0.00011157, -0.00014692, 0.00006956, -0.00007173, 0.00004114, -0.00004758, 0.00003710, -0.00004081, 0.00002840, -0.00002995, 0.00002068, -0.00002357, 0.00001791, -0.00001232, 0.00001387, -0.00002586, 0.00002023, -0.00001375, 0.00001721, -0.00002347, 0.00001255, 0.00000746, -0.00001347, 0.00001464, -0.00000435, -0.00002234, 0.00001706, -0.00001685, 0.00002185, -0.00001935, 0.00002112, -0.00002407, 0.00000224, 0.00000589, -0.00000862, 0.00001060, 0.00000030, -0.00000385, 0.00000688, -0.00000664, 0.00000072, -0.00000376, 0.00000544, 0.00000140, 0.00000416, -0.00000635, 0.00000733, -0.00001626, 0.00001248, -0.00001733, 0.00000849, 0.00001696, -0.00000369, -0.00002141, 0.00000812, 0.00001457, -0.00000366, -0.00000653, -0.00000958, 0.00001926, -0.00001154, 0.00000668, -0.00000568, 0.00000757, -0.00001338, 0.00001645, -0.00000849, -0.00000806, },
		{ 0.00001765, -0.00000126, -0.00002196, -0.00000051, 0.00002029, -0.00000021, -0.00000898, 0.00000472, -0.00000961, 0.00000358, -0.00000344, 0.00000563, -0.00000140, 0.00000541, -0.00001196, 0.00001116, -0.00001163, 0.00001037, -0.00000974, 0.00000571, -0.00000825, 0.00001611, -0.00002991, 0.00003535, -0.00004273, 0.00003439, -0.00001934, 0.00002776, -0.00002018, 0.00003458, -0.00003706, 0.00001602, -0.00001171, 0.00000062, 0.00000627, 0.00000104, -0.00000995, 0.00000520, -0.00000475, 0.00000713, -0.00001320, 0.00000575, -0.00000592, 0.00001375, -0.00002145, 0.00001951, -0.00002389, 0.00001787, -0.00002584, 0.00001857, -0.00003664, 0.00003753, -0.00006027, 0.00004984, -0.00013234, 0.00009324, -0.00061853, 0.00123280, -0.00028658, -0.00017560, 0.00002923, -0.00007109, 0.00003846, -0.00005026, 0.00002775, -0.00002951, 0.00002010, -0.00002760, 0.00002091, -0.00002562, 0.00001825, -0.00002091, 0.00002125, -0.00001479, 0.00000325, -0.00001102, 0.00001535, -0.00000985, 0.00000651, -0.00001995, 0.00003497, -0.00003192, 0.00002467, -0.00000832, -0.00001415, 0.00000259, -0.00000257, 0.00000537, -0.00000054, -0.00000034, 0.00000239, -0.00002412, 0.00002445, -0.00002036, 0.00001628, -0.00000258, 0.00000086, 0.00000049, 0.00000145, -0.00000653, 0.00000176, -0.00000035, 0.00000588, 0.00000156, -0.00000338, 0.00000419, -0.00001068, 0.00000406, -0.00000584, -0.00000596, 0.00002422, -0.00000307, -0.00001562, -0.00000479, 0.00002226, -0.00000600, 0.00000015, -0.00001737, 0.00001990, -0.00000723, 0.00000301, -0.00000339, 0.00000560, -0.00000929, 0.00000743, 0.00000297, -0.00001416, -0.00000059, },
		{ 0.00001035, -0.00002583, -0.00000182, 0.00001237, 0.00001296, -0.00001802, 0.00001184, -0.00001573, 0.00000956, -0.00001134, 0.00001416, -0.00001140, 0.00001715, -0.00002299, 0.00002054, -0.00002017, 0.00001906, -0.00001951, 0.00001800, -0.00002451, 0.00003351, -0.00004410, 0.00004266, -0.00004230, 0.00002876, -0.00001569, 0.00002695, -0.00001990, 0.00003379, -0.00002844, 0.00000729, -0.00000774, 0.00000002, -0.00000006, 0.00001203, -0.00002065, 0.00001634, -0.00001913, 0.00002374, -0.00002994, 0.00002459, -0.00002997, 0.00004062, -0.00004776, 0.00004629, -0.00005202, 0.00004974, -0.00006229, 0.00006230, -0.00008816, 0.00009621, -0.00012609, 0.00013344, -0.00022886, 0.00023543, -0.00070069, 0.00091475, 0.00023820, -0.00048014, 0.00018604, -0.00019079, 0.00012701, -0.00012160, 0.00008558, -0.00008155, 0.00006787, -0.00007083, 0.00005922, -0.00005968, 0.00004842, -0.00004867, 0.00004482, -0.00003585, 0.00002670, -0.00003645, 0.00003801, -0.00003100, 0.00003025, -0.00004343, 0.00005086, -0.00003903, 0.00002644, -0.00000912, -0.00000540, -0.00001215, 0.00001128, -0.00001073, 0.00001715, -0.00001903, 0.00002249, -0.00003927, 0.00002962, -0.00002009, 0.00001300, -0.00000029, 0.00000286, -0.00000374, 0.00000691, -0.00001055, 0.00000435, -0.00000373, 0.00000832, -0.00000002, -0.00000029, 0.00000003, -0.00000372, -0.00000506, 0.00000467, -0.00001637, 0.00002604, -0.00000010, -0.00001022, -0.00001445, 0.00002437, -0.00000406, 0.00000270, -0.00001843, 0.00001323, 0.00000175, -0.00000309, 0.00000100, 0.00000092, -0.00000221, -0.00000291, 0.00001241, -0.00001685, -0.00000136, 0.00001121, },
		{ -0.00002370, -0.00000595, 0.00000638, 0.00002086, -0.00001985, 0.00001208, -0.00001556, 0.00000983, -0.00001338, 0.00001616, -0.00001402, 0.00002011, -0.00002380, 0.00001910, -0.00001807, 0.00001743, -0.00001881, 0.00001885, -0.00002665, 0.00003291, -0.00003708, 0.00002807, -0.00002251, 0.00000805, -0.00000020, 0.00001496, -0.00000796, 0.00002074, -0.00000940, -0.00000801, 0.00000147, -0.00000580, -0.00000025, 0.00001441, -0.00002104, 0.00001667, -0.00002161, 0.00002658, -0.00003163, 0.00002740, -0.00003556, 0.00004535, -0.00004906, 0.00004620, -0.00005162, 0.00005089, -0.00006476, 0.00006712, -0.00009333, 0.00009891, -0.00012481, 0.00013294, -0.00021543, 0.00021532, -0.00054913, 0.00041178, 0.00075674, -0.00066259, 0.00023221, -0.00022621, 0.00014580, -0.00013489, 0.00009464, -0.00009197, 0.00007816, -0.00007930, 0.00006526, -0.00006395, 0.00005136, -0.00005154, 0.00004604, -0.00003766, 0.00003266, -0.00004404, 0.00004263, -0.00003501, 0.00003652, -0.00004745, 0.00004679, -0.00002888, 0.00001552, -0.00000208, -0.00000262, -0.00001911, 0.00001618, -0.00001747, 0.00002407, -0.00002494, 0.00002678, -0.00003621, 0.00001809, -0.00000743, 0.00000176, 0.00000628, 0.00000183, -0.00000438, 0.00000790, -0.00000973, 0.00000280, -0.00000367, 0.00000778, 0.00000043, 0.00000153, -0.00000308, 0.00000172, -0.00001135, 0.00001053, -0.00001994, 0.00002140, 0.00000564, -0.00000668, -0.00001907, 0.00002032, 0.00000192, 0.00000139, -0.00001399, 0.00000247, 0.00001118, -0.00000779, 0.00000420, -0.00000298, 0.00000355, -0.00000968, 0.00001568, -0.00001367, -0.00000493, 0.00000667, 0.00001708, },
		{ -0.00001238, 0.00000317, 0.00002235, -0.00001436, 0.00000605, -0.00000972, 0.00000463, -0.00000933, 0.00001107, -0.00000859, 0.00001382, -0.00001454, 0.00000771, -0.00000669, 0.00000676, -0.00000854, 0.00000878, -0.00001541, 0.00001656, -0.00001401, -0.00000001, 0.00000606, -0.00001703, 0.00001768, -0.00000016, 0.00000821, 0.00000309, 0.00001219, -0.00002324, 0.00001124, -0.00001365, 0.00000358, 0.00000984, -0.00001284, 0.00000768, -0.00001323, 0.00001653, -0.00001914, 0.00001472, -0.00002294, 0.00002872, -0.00002731, 0.00002220, -0.00002619, 0.00002495, -0.00003688, 0.00003680, -0.00005682, 0.00005264, -0.00006697, 0.00006437, -0.00011807, 0.00008373, -0.00027312, -0.00010223, 0.00109848, -0.00062061, 0.00012305, -0.00015571, 0.00008222, -0.00008246, 0.00004971, -0.00005633, 0.00004686, -0.00004963, 0.00003658, -0.00003685, 0.00002616, -0.00002901, 0.00002450, -0.00001926, 0.00001921, -0.00003139, 0.00002705, -0.00001992, 0.00002346, -0.00003113, 0.00002376, -0.00000394, -0.00000559, 0.00001190, -0.00000715, -0.00001579, 0.00000962, -0.00001231, 0.00001770, -0.00001616, 0.00001512, -0.00001784, -0.00000434, 0.00001066, -0.00001113, 0.00001270, 0.00000021, -0.00000264, 0.00000537, -0.00000526, -0.00000151, -0.00000137, 0.00000519, 0.00000238, 0.00000197, -0.00000432, 0.00000429, -0.00001328, 0.00001068, -0.00001690, 0.00001196, 0.00001257, -0.00000457, -0.00001915, 0.00001198, 0.00000952, -0.00000144, -0.00000736, -0.00000781, 0.00001667, -0.00000797, 0.00000390, -0.00000374, 0.00000526, -0.00001046, 0.00001187, -0.00000533, -0.00001050, 0.00000436, 0.00001846, -0.00001619, },
		{ 0.00000134, 0.00001918, -0.00000453, -0.00000277, -0.00000152, -0.00000306, -0.00000195, 0.00000182, 0.00000147, 0.00000238, -0.00000021, -0.00000775, 0.00000775, -0.00000678, 0.00000521, -0.00000580, 0.00000164, -0.00000568, 0.00001254, -0.00002702, 0.00002880, -0.00003360, 0.00002739, -0.00000938, 0.00001998, -0.00001032, 0.00002771, -0.00003124, 0.00001597, -0.00001856, 0.00000664, 0.00000347, -0.00000184, -0.00000447, -0.00000039, 0.00000067, -0.00000030, -0.00000504, -0.00000145, 0.00000186, 0.00000435, -0.00001095, 0.00000836, -0.00001106, 0.00000291, -0.00000802, -0.00000284, -0.00001308, 0.00001113, -0.00002675, 0.00000332, -0.00007261, -0.00000425, -0.00047138, 0.00115802, -0.00033616, -0.00011272, -0.00001348, -0.00003299, 0.00000834, -0.00002548, 0.00000500, -0.00000811, 0.00000163, -0.00001113, 0.00000672, -0.00001323, 0.00000607, -0.00000822, 0.00000930, -0.00000507, -0.00000653, -0.00000053, 0.00000614, -0.00000081, -0.00000343, -0.00000808, 0.00002549, -0.00002789, 0.00002619, -0.00001447, -0.00000629, -0.00000375, -0.00000016, 0.00000340, 0.00000080, -0.00000408, 0.00000518, -0.00002597, 0.00002371, -0.00001778, 0.00001381, 0.00000106, -0.00000089, 0.00000187, -0.00000002, -0.00000577, 0.00000079, 0.00000246, 0.00000440, 0.00000186, -0.00000406, 0.00000423, -0.00001146, 0.00000654, -0.00000987, 0.00000114, 0.00001803, -0.00000247, -0.00001656, 0.00000239, 0.00001568, -0.00000308, -0.00000186, -0.00001397, 0.00001590, -0.00000303, 0.00000022, -0.00000143, 0.00000308, -0.00000627, 0.00000362, 0.00000451, -0.00001513, 0.00000255, 0.00001645, -0.00000620, -0.00001861, },
		{ 0.00001432, 0.00000544, -0.00001004, 0.00000501, -0.00000938, 0.00000474, -0.00000699, 0.00001085, -0.00000806, 0.00001219, -0.00001975, 0.00001768, -0.00001582, 0.00001492, -0.00001667, 0.00001485, -0.00002226, 0.00002958, -0.00004007, 0.00003479, -0.00003365, 0.00002326, -0.00000798, 0.00002243, -0.00001410, 0.00003205, -0.00002815, 0.00001250, -0.00001702, 0.00000482, 0.00000033, 0.00000584, -0.00001302, 0.00000936, -0.00001230, 0.00001501, -0.00002100, 0.00001655, -0.00002083, 0.00002959, -0.00003564, 0.00003339, -0.00003726, 0.00003286, -0.00004277, 0.00003988, -0.00006414, 0.00006923, -0.00009226, 0.00008919, -0.00017609, 0.00015896, -0.00061301, 0.00093268, 0.00011319, -0.00038041, 0.00012493, -0.00014101, 0.00008947, -0.00009106, 0.00005829, -0.00005712, 0.00004741, -0.00005287, 0.00004379, -0.00004589, 0.00003480, -0.00003508, 0.00003279, -0.00002623, 0.00001685, -0.00002630, 0.00002953, -0.00002283, 0.00002168, -0.00003446, 0.00004615, -0.00004033, 0.00003214, -0.00001727, 0.00000170, -0.00001542, 0.00001017, -0.00000950, 0.00001554, -0.00001920, 0.00002060, -0.00003577, 0.00002372, -0.00001352, 0.00000740, 0.00000557, -0.00000060, -0.00000043, 0.00000339, -0.00000756, 0.00000093, 0.00000110, 0.00000530, 0.00000210, -0.00000315, 0.00000267, -0.00000765, 0.00000069, -0.00000227, -0.00000761, 0.00001988, 0.00000077, -0.00001314, -0.00000582, 0.00001830, -0.00000203, 0.00000083, -0.00001488, 0.00000963, 0.00000474, -0.00000434, 0.00000171, -0.00000052, -0.00000059, -0.00000446, 0.00001125, -0.00001560, -0.00000088, 0.00001369, 0.00000280, -0.00002178, -0.00000118, },
		{ 0.00001235, -0.00001281, 0.00000728, -0.00001173, 0.00000770, -0.00001163, 0.00001516, -0.00001259, 0.00001735, -0.00002301, 0.00001825, -0.00001586, 0.00001591, -0.00001847, 0.00001788, -0.00002602, 0.00003025, -0.00003428, 0.00002226, -0.00001819, 0.00000761, 0.00000202, 0.00001652, -0.00000850, 0.00002566, -0.00001519, 0.00000192, -0.00000911, -0.00000270, 0.00000236, 0.00000725, -0.00001431, 0.00001153, -0.00001675, 0.00002020, -0.00002571, 0.00002242, -0.00002924, 0.00003741, -0.00004066, 0.00003752, -0.00004163, 0.00003961, -0.00005201, 0.00005312, -0.00007927, 0.00008332, -0.00010502, 0.00010829, -0.00019047, 0.00018342, -0.00053748, 0.00051664, 0.00058410, -0.00055875, 0.00017941, -0.00018567, 0.00011778, -0.00011269, 0.00007504, -0.00007514, 0.00006490, -0.00006788, 0.00005572, -0.00005526, 0.00004245, -0.00004289, 0.00003923, -0.00003280, 0.00002702, -0.00003822, 0.00003854, -0.00003110, 0.00003254, -0.00004415, 0.00004856, -0.00003574, 0.00002438, -0.00001050, 0.00000202, -0.00001842, 0.00001156, -0.00001352, 0.00002006, -0.00002213, 0.00002129, -0.00002931, 0.00000984, -0.00000001, -0.00000392, 0.00001166, -0.00000112, -0.00000100, 0.00000409, -0.00000629, -0.00000117, 0.00000126, 0.00000491, 0.00000292, -0.00000214, 0.00000071, -0.00000361, -0.00000450, 0.00000329, -0.00001230, 0.00001757, 0.00000512, -0.00000976, -0.00001135, 0.00001701, 0.00000140, 0.00000105, -0.00001188, 0.00000099, 0.00001148, -0.00000650, 0.00000294, -0.00000264, 0.00000302, -0.00000847, 0.00001191, -0.00001063, -0.00000653, 0.00001170, 0.00000866, -0.00002049, -0.00000550, 0.00001006, },
		{ -0.00001072, 0.00000525, -0.00000986, 0.00000625, -0.00001092, 0.00001301, -0.00000982, 0.00001409, -0.00001692, 0.00000964, -0.00000749, 0.00000864, -0.00001131, 0.00001058, -0.00001726, 0.00001654, -0.00001443, -0.00000117, 0.00000373, -0.00001096, 0.00001360, 0.00000771, 0.00000172, 0.00001355, 0.00000221, -0.00001103, 0.00000159, -0.00001309, 0.00000755, 0.00000366, -0.00000924, 0.00000649, -0.00001241, 0.00001471, -0.00001848, 0.00001506, -0.00002201, 0.00002693, -0.00002606, 0.00002151, -0.00002510, 0.00002382, -0.00003583, 0.00003642, -0.00005822, 0.00005457, -0.00006828, 0.00006694, -0.00012824, 0.00010050, -0.00032825, 0.00005726, 0.00092464, -0.00055327, 0.00010631, -0.00014374, 0.00007822, -0.00008003, 0.00004727, -0.00005502, 0.00004738, -0.00005035, 0.00003784, -0.00003773, 0.00002618, -0.00002916, 0.00002617, -0.00002189, 0.00002023, -0.00003217, 0.00002936, -0.00002204, 0.00002559, -0.00003444, 0.00003187, -0.00001505, 0.00000449, 0.00000502, -0.00000595, -0.00001149, 0.00000292, -0.00000746, 0.00001318, -0.00001249, 0.00000875, -0.00001098, -0.00001107, 0.00001532, -0.00001387, 0.00001513, -0.00000033, -0.00000089, 0.00000300, -0.00000318, -0.00000415, 0.00000183, 0.00000405, 0.00000382, -0.00000102, -0.00000109, -0.00000040, -0.00000778, 0.00000580, -0.00001284, 0.00001223, 0.00000946, -0.00000635, -0.00001426, 0.00001294, 0.00000565, 0.00000038, -0.00000751, -0.00000639, 0.00001399, -0.00000432, 0.00000103, -0.00000190, 0.00000294, -0.00000724, 0.00000678, -0.00000196, -0.00001272, 0.00001001, 0.00001141, -0.00001551, -0.00001091, 0.00000692, 0.00001525, },
		{ 0.00000096, -0.00000559, 0.00000192, -0.00000635, 0.00000620, -0.00000204, 0.00000538, -0.00000536, -0.00000345, 0.00000429, -0.00000202, -0.00000003, -0.00000175, -0.00000248, -0.00000286, 0.00000840, -0.00002313, 0.00002034, -0.00002269, 0.00001927, 0.00000206, 0.00001074, 0.00000205, 0.00001748, -0.00002059, 0.00001013, -0.00002166, 0.00001160, -0.00000084, -0.00000201, -0.00000158, -0.00000355, 0.00000331, -0.00000466, 0.00000030, -0.00000576, 0.00000625, -0.00000161, -0.00000372, 0.00000071, -0.00000238, -0.00000713, 0.00000418, -0.00001812, 0.00000432, -0.00000854, -0.00000143, -0.00003417, -0.00002198, -0.00009629, -0.00030501, 0.00103302, -0.00033888, -0.00007683, -0.00003852, -0.00000812, -0.00001198, -0.00000873, -0.00001044, 0.00000683, -0.00001156, 0.00000111, -0.00000389, -0.00000411, -0.00000287, 0.00000156, -0.00000020, 0.00000206, -0.00001342, 0.00000756, -0.00000130, 0.00000653, -0.00001197, 0.00000418, 0.00001292, -0.00001919, 0.00002249, -0.00001691, 0.00000031, -0.00001039, 0.00000333, 0.00000055, 0.00000296, -0.00000848, 0.00000881, -0.00002822, 0.00002339, -0.00001613, 0.00001241, 0.00000352, -0.00000137, 0.00000173, -0.00000023, -0.00000608, 0.00000136, 0.00000372, 0.00000414, 0.00000044, -0.00000259, 0.00000170, -0.00000903, 0.00000565, -0.00001056, 0.00000589, 0.00001236, -0.00000253, -0.00001521, 0.00000767, 0.00000907, 0.00000031, -0.00000388, -0.00001024, 0.00001132, 0.00000156, -0.00000292, 0.00000072, 0.00000020, -0.00000267, -0.00000086, 0.00000645, -0.00001635, 0.00000692, 0.00001260, -0.00000911, -0.00001545, 0.00000398, 0.00001514, -0.00000558, },
		{ -0.00000138, -0.00000278, -0.00000077, -0.00000174, 0.00000643, -0.00000371, 0.00000582, -0.00001465, 0.00001313, -0.00000994, 0.00000903, -0.00001204, 0.00000989, -0.00001792, 0.00002319, -0.00003323, 0.00002357, -0.00002221, 0.00001557, 0.00000246, 0.00001511, -0.00000449, 0.00002603, -0.00002281, 0.00001288, -0.00002451, 0.00001064, -0.00000219, 0.00000282, -0.00000770, 0.00000418, -0.00000731, 0.00000802, -0.00001324, 0.00000960, -0.00001295, 0.00001953, -0.00002412, 0.00002118, -0.00002354, 0.00001722, -0.00002445, 0.00001826, -0.00004002, 0.00004172, -0.00005773, 0.00004256, -0.00011738, 0.00006812, -0.00048650, 0.00089298, 0.00002778, -0.00029559, 0.00007089, -0.00009549, 0.00005412, -0.00006213, 0.00003262, -0.00003381, 0.00002724, -0.00003463, 0.00002795, -0.00003180, 0.00002110, -0.00002125, 0.00002027, -0.00001642, 0.00000701, -0.00001550, 0.00001981, -0.00001333, 0.00001127, -0.00002181, 0.00003578, -0.00003572, 0.00003286, -0.00002297, 0.00000897, -0.00002009, 0.00001070, -0.00000935, 0.00001484, -0.00002011, 0.00001962, -0.00003286, 0.00001872, -0.00000838, 0.00000329, 0.00000990, -0.00000252, 0.00000124, 0.00000119, -0.00000583, -0.00000074, 0.00000429, 0.00000367, 0.00000224, -0.00000371, 0.00000278, -0.00000873, 0.00000387, -0.00000713, 0.00000032, 0.00001305, 0.00000164, -0.00001463, 0.00000235, 0.00001085, 0.00000126, -0.00000166, -0.00001048, 0.00000515, 0.00000828, -0.00000607, 0.00000278, -0.00000252, 0.00000175, -0.00000669, 0.00001055, -0.00001489, 0.00000117, 0.00001375, -0.00000353, -0.00001734, 0.00000015, 0.00001372, 0.00000003, -0.00000298, },
		{ -0.00000593, 0.00000330, -0.00000752, 0.00001172, -0.00000886, 0.00001192, -0.00001927, 0.00001482, -0.00001121, 0.00001174, -0.00001546, 0.00001419, -0.00002243, 0.00002440, -0.00002812, 0.00001297, -0.00001125, 0.00000492, 0.00000720, 0.00001508, -0.00000526, 0.00002685, -0.00001717, 0.00000918, -0.00002047, 0.00000342, 0.00000120, 0.00000296, -0.00000892, 0.00000697, -0.00001226, 0.00001378, -0.00001900, 0.00001645, -0.00002188, 0.00002789, -0.00003019, 0.00002672, -0.00002953, 0.00002591, -0.00003613, 0.00003488, -0.00005948, 0.00006094, -0.00007718, 0.00007238, -0.00014903, 0.00012563, -0.00048081, 0.00057225, 0.00043124, -0.00045056, 0.00011941, -0.00013840, 0.00008314, -0.00008489, 0.00005077, -0.00005373, 0.00004681, -0.00005173, 0.00004193, -0.00004287, 0.00003031, -0.00003100, 0.00002915, -0.00002513, 0.00001867, -0.00002903, 0.00003069, -0.00002351, 0.00002439, -0.00003521, 0.00004382, -0.00003723, 0.00002991, -0.00001856, 0.00000851, -0.00002006, 0.00000888, -0.00001061, 0.00001685, -0.00002002, 0.00001669, -0.00002308, 0.00000273, 0.00000587, -0.00000823, 0.00001575, -0.00000289, 0.00000119, 0.00000113, -0.00000374, -0.00000380, 0.00000507, 0.00000297, 0.00000388, -0.00000416, 0.00000286, -0.00000733, 0.00000136, -0.00000370, -0.00000367, 0.00001181, 0.00000537, -0.00001249, -0.00000264, 0.00001124, 0.00000263, -0.00000024, -0.00000866, -0.00000143, 0.00001236, -0.00000575, 0.00000214, -0.00000292, 0.00000319, -0.00000785, 0.00000856, -0.00000822, -0.00000665, 0.00001501, 0.00000035, -0.00001615, -0.00000470, 0.00001203, 0.00000437, -0.00000489, 0.00000115, },
		{ 0.00000480, -0.00000959, 0.00001212, -0.00000841, 0.00001147, -0.00001639, 0.00000917, -0.00000602, 0.00000804, -0.00001165, 0.00000996, -0.00001655, 0.00001393, -0.00001233, -0.00000468, 0.00000283, -0.00000632, 0.00001175, 0.00001348, -0.00000219, 0.00002201, -0.00000619, 0.00000125, -0.00001140, -0.00000808, 0.00000769, -0.00000060, -0.00000567, 0.00000454, -0.00001057, 0.00001132, -0.00001549, 0.00001292, -0.00001844, 0.00002186, -0.00002109, 0.00001700, -0.00001998, 0.00001797, -0.00002901, 0.00002894, -0.00005096, 0.00004675, -0.00005829, 0.00005474, -0.00011885, 0.00009046, -0.00034573, 0.00018835, 0.00074407, -0.00046146, 0.00006985, -0.00011638, 0.00006115, -0.00006698, 0.00003589, -0.00004532, 0.00003987, -0.00004377, 0.00003276, -0.00003310, 0.00002125, -0.00002444, 0.00002312, -0.00002033, 0.00001710, -0.00002824, 0.00002688, -0.00001968, 0.00002294, -0.00003212, 0.00003465, -0.00002305, 0.00001400, -0.00000398, -0.00000134, -0.00001011, -0.00000192, -0.00000336, 0.00000920, -0.00000933, 0.00000309, -0.00000470, -0.00001681, 0.00001886, -0.00001590, 0.00001705, -0.00000051, 0.00000057, 0.00000063, -0.00000125, -0.00000626, 0.00000486, 0.00000294, 0.00000468, -0.00000354, 0.00000191, -0.00000511, -0.00000144, -0.00000067, -0.00000624, 0.00000967, 0.00000767, -0.00000867, -0.00000729, 0.00001091, 0.00000351, 0.00000132, -0.00000661, -0.00000576, 0.00001177, -0.00000120, -0.00000139, -0.00000065, 0.00000118, -0.00000443, 0.00000197, 0.00000091, -0.00001387, 0.00001492, 0.00000331, -0.00001307, -0.00000945, 0.00000995, 0.00000711, -0.00000470, 0.00000187, -0.00000700, },
		{ -0.00000832, 0.00000839, -0.00000366, 0.00000636, -0.00000867, -0.00000039, 0.00000200, 0.00000137, -0.00000403, 0.00000115, -0.00000543, -0.00000110, 0.00000509, -0.00001975, 0.00001167, -0.00001196, 0.00001205, 0.00001300, 0.00000217, 0.00001472, 0.00000631, -0.00000726, -0.00000094, -0.00001985, 0.00001355, -0.00000463, -0.00000088, -0.00000048, -0.00000482, 0.00000360, -0.00000611, 0.00000276, -0.00000707, 0.00000710, -0.00000375, -0.00000033, -0.00000278, 0.00000138, -0.00001126, 0.00000935, -0.00002531, 0.00001307, -0.00001843, 0.00001149, -0.00005640, 0.00001010, -0.00016755, -0.00014058, 0.00087613, -0.00030303, -0.00006643, -0.00004536, 0.00000224, -0.00002095, -0.00000147, -0.00001710, 0.00001369, -0.00001779, 0.00000752, -0.00000948, 0.00000046, -0.00000728, 0.00000703, -0.00000592, 0.00000537, -0.00001608, 0.00001168, -0.00000528, 0.00001035, -0.00001676, 0.00001392, 0.00000014, -0.00000797, 0.00001452, -0.00001504, 0.00000407, -0.00001574, 0.00000666, -0.00000218, 0.00000493, -0.00001251, 0.00001224, -0.00003000, 0.00002283, -0.00001481, 0.00001153, 0.00000529, -0.00000114, 0.00000070, 0.00000021, -0.00000679, 0.00000280, 0.00000408, 0.00000431, -0.00000175, 0.00000004, -0.00000232, -0.00000430, 0.00000191, -0.00000790, 0.00000762, 0.00000803, -0.00000355, -0.00001130, 0.00001016, 0.00000360, 0.00000324, -0.00000519, -0.00000696, 0.00000700, 0.00000576, -0.00000574, 0.00000247, -0.00000237, 0.00000080, -0.00000531, 0.00000818, -0.00001712, 0.00001160, 0.00000678, -0.00000997, -0.00001250, 0.00000674, 0.00000893, -0.00000348, 0.00000172, -0.00000660, 0.00000408, },
		{ 0.00000278, 0.00000244, 0.00000016, -0.00000040, -0.00000912, 0.00000823, -0.00000389, 0.00000277, -0.00000689, 0.00000436, -0.00001282, 0.00001581, -0.00002512, 0.00001064, -0.00000977, 0.00000737, 0.00001412, 0.00000649, 0.00000754, 0.00001703, -0.00001305, 0.00000732, -0.00002781, 0.00001500, -0.00000578, 0.00000210, -0.00000457, 0.00000096, -0.00000452, 0.00000345, -0.00000761, 0.00000478, -0.00000758, 0.00001196, -0.00001496, 0.00001158, -0.00001302, 0.00000542, -0.00001048, 0.00000126, -0.00002025, 0.00001887, -0.00002904, 0.00000212, -0.00006407, -0.00002185, -0.00033786, 0.00080489, -0.00001600, -0.00023261, 0.00003062, -0.00006024, 0.00002621, -0.00003928, 0.00001261, -0.00001546, 0.00001092, -0.00001933, 0.00001456, -0.00001999, 0.00000981, -0.00000962, 0.00000954, -0.00000853, -0.00000065, -0.00000638, 0.00001115, -0.00000483, 0.00000162, -0.00000818, 0.00002186, -0.00002614, 0.00002816, -0.00002450, 0.00001430, -0.00002450, 0.00001189, -0.00000963, 0.00001440, -0.00002116, 0.00001906, -0.00003027, 0.00001451, -0.00000460, 0.00000059, 0.00001277, -0.00000306, 0.00000143, 0.00000018, -0.00000515, -0.00000085, 0.00000598, 0.00000315, 0.00000076, -0.00000224, 0.00000055, -0.00000683, 0.00000389, -0.00000881, 0.00000590, 0.00000694, 0.00000170, -0.00001382, 0.00000860, 0.00000352, 0.00000480, -0.00000397, -0.00000606, 0.00000061, 0.00001170, -0.00000782, 0.00000380, -0.00000459, 0.00000431, -0.00000923, 0.00001007, -0.00001442, 0.00000439, 0.00001134, -0.00000798, -0.00001298, 0.00000209, 0.00001045, -0.00000230, 0.00000169, -0.00000560, 0.00000218, -0.00000521, },
		{ 0.00000708, -0.00000398, 0.00000497, -0.00001358, 0.00000970, -0.00000506, 0.00000578, -0.00001055, 0.00000858, -0.00001675, 0.00001630, -0.00001958, 0.00000118, -0.00000262, 0.00000062, 0.00001494, 0.00001096, 0.00000172, 0.00002409, -0.00001428, 0.00001115, -0.00002945, 0.00001013, -0.00000246, 0.00000147, -0.00000548, 0.00000375, -0.00000908, 0.00000855, -0.00001294, 0.00001100, -0.00001518, 0.00001876, -0.00001980, 0.00001612, -0.00001792, 0.00001277, -0.00002064, 0.00001648, -0.00003849, 0.00003688, -0.00004746, 0.00003274, -0.00010009, 0.00005221, -0.00038687, 0.00057616, 0.00031057, -0.00035171, 0.00006292, -0.00009295, 0.00004885, -0.00005722, 0.00002686, -0.00003227, 0.00002794, -0.00003441, 0.00002699, -0.00002956, 0.00001758, -0.00001841, 0.00001812, -0.00001676, 0.00000979, -0.00001867, 0.00002113, -0.00001415, 0.00001402, -0.00002231, 0.00003326, -0.00003277, 0.00003039, -0.00002380, 0.00001471, -0.00002284, 0.00000775, -0.00000866, 0.00001435, -0.00001857, 0.00001312, -0.00001791, -0.00000279, 0.00000982, -0.00001079, 0.00001819, -0.00000320, 0.00000191, -0.00000065, -0.00000239, -0.00000474, 0.00000740, 0.00000222, 0.00000304, -0.00000406, 0.00000273, -0.00000838, 0.00000474, -0.00000863, 0.00000399, 0.00000565, 0.00000548, -0.00001378, 0.00000551, 0.00000433, 0.00000488, -0.00000193, -0.00000490, -0.00000431, 0.00001353, -0.00000546, 0.00000172, -0.00000370, 0.00000398, -0.00000790, 0.00000584, -0.00000659, -0.00000510, 0.00001589, -0.00000670, -0.00001148, -0.00000321, 0.00001151, -0.00000142, 0.00000197, -0.00000435, -0.00000007, -0.00000175, -0.00000244, },
		{ -0.00000475, 0.00000630, -0.00001308, 0.00000637, -0.00000239, 0.00000498, -0.00000948, 0.00000690, -0.00001332, 0.00000875, -0.00000768, -0.00001065, 0.00000368, -0.00000379, 0.00001302, 0.00001616, -0.00000224, 0.00002687, -0.00001065, 0.00001015, -0.00002458, -0.00000036, 0.00000434, -0.00000220, -0.00000325, 0.00000289, -0.00000882, 0.00000765, -0.00001154, 0.00000964, -0.00001372, 0.00001516, -0.00001409, 0.00001044, -0.00001276, 0.00000962, -0.00001897, 0.00001710, -0.00003791, 0.00003223, -0.00004050, 0.00003165, -0.00009366, 0.00005625, -0.00032296, 0.00027799, 0.00057625, -0.00036143, 0.00002445, -0.00008138, 0.00003689, -0.00004791, 0.00001949, -0.00003067, 0.00002723, -0.00003234, 0.00002338, -0.00002475, 0.00001311, -0.00001646, 0.00001673, -0.00001581, 0.00001112, -0.00002085, 0.00002060, -0.00001364, 0.00001617, -0.00002431, 0.00003116, -0.00002593, 0.00002033, -0.00001246, 0.00000499, -0.00001121, -0.00000457, -0.00000048, 0.00000616, -0.00000715, -0.00000119, 0.00000022, -0.00002077, 0.00002064, -0.00001659, 0.00001785, 0.00000018, 0.00000115, -0.00000111, -0.00000010, -0.00000717, 0.00000703, 0.00000251, 0.00000421, -0.00000458, 0.00000346, -0.00000829, 0.00000387, -0.00000681, 0.00000111, 0.00000544, 0.00000660, -0.00001068, 0.00000061, 0.00000658, 0.00000300, 0.00000137, -0.00000473, -0.00000596, 0.00001027, 0.00000104, -0.00000306, -0.00000023, 0.00000029, -0.00000251, -0.00000199, 0.00000276, -0.00001335, 0.00001800, -0.00000451, -0.00000969, -0.00000754, 0.00001117, -0.00000023, 0.00000189, -0.00000256, -0.00000241, 0.00000116, -0.00000593, 0.00000884, },
		{ 0.00000469, -0.00000936, 0.00000062, 0.00000165, 0.00000253, -0.00000588, 0.00000205, -0.00000639, -0.00000126, 0.00000362, -0.00001806, 0.00000420, -0.00000316, 0.00000760, 0.00002155, -0.00000363, 0.00002536, -0.00000294, 0.00000538, -0.00001494, -0.00001365, 0.00001169, -0.00000647, 0.00000018, -0.00000012, -0.00000514, 0.00000248, -0.00000550, 0.00000306, -0.00000612, 0.00000513, -0.00000271, -0.00000021, -0.00000273, 0.00000088, -0.00001016, 0.00000804, -0.00002466, 0.00001317, -0.00001837, 0.00001129, -0.00006126, 0.00001881, -0.00020707, 0.00000255, 0.00070807, -0.00024191, -0.00007500, -0.00003775, 0.00000038, -0.00002004, -0.00000253, -0.00001589, 0.00001306, -0.00001736, 0.00000819, -0.00001011, 0.00000058, -0.00000722, 0.00000810, -0.00000771, 0.00000483, -0.00001437, 0.00001140, -0.00000520, 0.00000983, -0.00001640, 0.00001882, -0.00000997, 0.00000294, 0.00000443, -0.00000963, 0.00000450, -0.00001879, 0.00000895, -0.00000410, 0.00000599, -0.00001533, 0.00001460, -0.00003055, 0.00002141, -0.00001319, 0.00001056, 0.00000695, -0.00000086, -0.00000054, 0.00000063, -0.00000716, 0.00000433, 0.00000428, 0.00000400, -0.00000354, 0.00000249, -0.00000632, 0.00000113, -0.00000326, -0.00000293, 0.00000667, 0.00000510, -0.00000524, -0.00000531, 0.00000968, -0.00000008, 0.00000515, -0.00000534, -0.00000467, 0.00000360, 0.00000891, -0.00000774, 0.00000337, -0.00000409, 0.00000347, -0.00000904, 0.00000911, -0.00001674, 0.00001547, 0.00000007, -0.00000925, -0.00000947, 0.00000861, 0.00000201, 0.00000070, 0.00000014, -0.00000477, 0.00000324, -0.00000743, 0.00000789, -0.00000283, },
		{ -0.00000502, -0.00000457, 0.00000430, 0.00000093, -0.00000246, -0.00000255, -0.00000038, -0.00000841, 0.00000903, -0.00001749, -0.00000199, 0.00000150, 0.00000072, 0.00002502, -0.00000132, 0.00001983, 0.00000745, -0.00000120, -0.00000345, -0.00002589, 0.00001608, -0.00000855, 0.00000237, -0.00000292, -0.00000078, -0.00000357, 0.00000121, -0.00000423, 0.00000227, -0.00000500, 0.00000745, -0.00000890, 0.00000543, -0.00000669, -0.00000134, -0.00000251, -0.00000890, -0.00000760, 0.00000404, -0.00001070, -0.00002564, -0.00002570, -0.00009612, -0.00018668, 0.00068423, -0.00002367, -0.00019222, 0.00000621, -0.00003765, 0.00000814, -0.00002466, 0.00000028, -0.00000408, 0.00000049, -0.00000891, 0.00000535, -0.00001202, 0.00000245, -0.00000176, 0.00000207, -0.00000389, -0.00000467, -0.00000063, 0.00000536, 0.00000071, -0.00000485, 0.00000341, 0.00000736, -0.00001386, 0.00001917, -0.00002149, 0.00001634, -0.00002727, 0.00001276, -0.00000966, 0.00001352, -0.00002162, 0.00001824, -0.00002752, 0.00001071, -0.00000175, -0.00000110, 0.00001457, -0.00000269, 0.00000065, -0.00000015, -0.00000499, 0.00000005, 0.00000673, 0.00000300, -0.00000141, 0.00000028, -0.00000300, -0.00000284, 0.00000121, -0.00000732, 0.00000848, 0.00000235, 0.00000064, -0.00001043, 0.00001198, -0.00000247, 0.00000764, -0.00000533, -0.00000242, -0.00000316, 0.00001431, -0.00000905, 0.00000430, -0.00000619, 0.00000651, -0.00001154, 0.00000937, -0.00001364, 0.00000797, 0.00000701, -0.00001043, -0.00000882, 0.00000402, 0.00000527, -0.00000165, 0.00000350, -0.00000679, 0.00000432, -0.00000725, 0.00000525, 0.00000017, 0.00000221, },
		{ -0.00000742, 0.00000422, 0.00000121, -0.00000057, -0.00000500, 0.00000234, -0.00001040, 0.00000749, -0.00001029, -0.00001137, 0.00000599, -0.00000385, 0.00002397, 0.00000546, 0.00001122, 0.00001856, -0.00000720, 0.00000675, -0.00003361, 0.00001491, -0.00000669, 0.00000179, -0.00000375, 0.00000191, -0.00000744, 0.00000506, -0.00000831, 0.00000698, -0.00001026, 0.00001144, -0.00001114, 0.00000750, -0.00000882, 0.00000261, -0.00000851, 0.00000153, -0.00002051, 0.00001604, -0.00002196, -0.00000269, -0.00005399, -0.00002328, -0.00026977, 0.00053464, 0.00022669, -0.00027072, 0.00001741, -0.00005567, 0.00002035, -0.00003426, 0.00000739, -0.00001460, 0.00001185, -0.00001915, 0.00001375, -0.00001790, 0.00000673, -0.00000752, 0.00000834, -0.00000972, 0.00000252, -0.00000946, 0.00001215, -0.00000538, 0.00000404, -0.00000818, 0.00001905, -0.00002337, 0.00002556, -0.00002471, 0.00001867, -0.00002530, 0.00000737, -0.00000721, 0.00001211, -0.00001738, 0.00001028, -0.00001369, -0.00000678, 0.00001196, -0.00001170, 0.00001908, -0.00000221, 0.00000133, -0.00000140, -0.00000205, -0.00000417, 0.00000842, 0.00000232, 0.00000081, -0.00000215, 0.00000051, -0.00000669, 0.00000508, -0.00001049, 0.00000922, 0.00000043, 0.00000468, -0.00001277, 0.00001166, -0.00000230, 0.00000719, -0.00000328, -0.00000135, -0.00000697, 0.00001450, -0.00000532, 0.00000140, -0.00000464, 0.00000503, -0.00000838, 0.00000363, -0.00000555, -0.00000221, 0.00001426, -0.00001179, -0.00000696, -0.00000119, 0.00000845, -0.00000416, 0.00000637, -0.00000760, 0.00000405, -0.00000572, 0.00000198, 0.00000278, 0.00000050, -0.00000189, },
		{ 0.00000204, 0.00000256, 0.00000025, -0.00000549, 0.00000211, -0.00000837, 0.00000188, -0.00000133, -0.00001827, 0.00000567, -0.00000315, 0.00001736, 0.00001580, 0.00000172, 0.00002780, -0.00001012, 0.00001300, -0.00003459, 0.00000743, -0.00000099, -0.00000148, -0.00000231, 0.00000205, -0.00000782, 0.00000466, -0.00000778, 0.00000646, -0.00000926, 0.00000853, -0.00000690, 0.00000378, -0.00000563, 0.00000131, -0.00000870, 0.00000442, -0.00002297, 0.00001539, -0.00002013, 0.00000406, -0.00006011, 0.00000603, -0.00026507, 0.00032138, 0.00043496, -0.00026705, -0.00001967, -0.00004665, 0.00001178, -0.00002799, 0.00000256, -0.00001516, 0.00001313, -0.00001924, 0.00001247, -0.00001518, 0.00000411, -0.00000749, 0.00000904, -0.00001022, 0.00000430, -0.00001205, 0.00001245, -0.00000576, 0.00000717, -0.00001263, 0.00002204, -0.00002313, 0.00002183, -0.00001807, 0.00001094, -0.00001364, -0.00000538, 0.00000126, 0.00000396, -0.00000593, -0.00000402, 0.00000353, -0.00002272, 0.00002050, -0.00001575, 0.00001732, 0.00000191, 0.00000066, -0.00000198, 0.00000007, -0.00000666, 0.00000808, 0.00000288, 0.00000224, -0.00000376, 0.00000298, -0.00000899, 0.00000680, -0.00001092, 0.00000740, 0.00000095, 0.00000542, -0.00001133, 0.00000793, 0.00000118, 0.00000349, 0.00000099, -0.00000230, -0.00000668, 0.00000937, 0.00000238, -0.00000398, -0.00000056, 0.00000026, -0.00000160, -0.00000487, 0.00000347, -0.00001103, 0.00001862, -0.00001099, -0.00000605, -0.00000500, 0.00000995, -0.00000528, 0.00000727, -0.00000621, 0.00000210, -0.00000309, -0.00000131, 0.00000440, -0.00000003, -0.00000027, -0.00000796, },
		{ 0.00000327, 0.00000137, -0.00000537, 0.00000081, -0.00000525, -0.00000350, 0.00000423, -0.00001847, -0.00000133, 0.00000262, 0.00000706, 0.00002657, -0.00000527, 0.00003236, -0.00000795, 0.00001380, -0.00002850, -0.00000504, 0.00000665, -0.00000590, 0.00000031, 0.00000030, -0.00000534, 0.00000098, -0.00000391, 0.00000226, -0.00000408, 0.00000168, 0.00000021, -0.00000204, -0.00000059, -0.00000218, -0.00000577, 0.00000240, -0.00001834, 0.00000686, -0.00001087, 0.00000070, -0.00005112, 0.00000513, -0.00021062, 0.00011048, 0.00054790, -0.00017032, -0.00009345, -0.00002198, -0.00000900, -0.00001284, -0.00000886, -0.00000953, 0.00000725, -0.00001213, 0.00000466, -0.00000712, -0.00000246, -0.00000391, 0.00000579, -0.00000652, 0.00000151, -0.00000935, 0.00000754, -0.00000170, 0.00000548, -0.00001092, 0.00001784, -0.00001533, 0.00001092, -0.00000513, -0.00000245, 0.00000214, -0.00001936, 0.00000990, -0.00000501, 0.00000585, -0.00001654, 0.00001541, -0.00002947, 0.00001877, -0.00001086, 0.00000908, 0.00000890, -0.00000098, -0.00000148, 0.00000054, -0.00000668, 0.00000541, 0.00000478, 0.00000260, -0.00000404, 0.00000367, -0.00000886, 0.00000554, -0.00000809, 0.00000272, 0.00000409, 0.00000303, -0.00000677, 0.00000166, 0.00000686, -0.00000198, 0.00000598, -0.00000430, -0.00000352, 0.00000143, 0.00001065, -0.00000863, 0.00000321, -0.00000470, 0.00000494, -0.00001159, 0.00000888, -0.00001476, 0.00001758, -0.00000636, -0.00000756, -0.00000607, 0.00000870, -0.00000397, 0.00000530, -0.00000234, -0.00000135, 0.00000026, -0.00000427, 0.00000512, 0.00000009, 0.00000066, -0.00000747, -0.00000049, },
		{ 0.00000356, -0.00000562, 0.00000000, -0.00000337, -0.00000576, 0.00000407, -0.00001179, -0.00001263, 0.00000967, -0.00000251, 0.00003333, -0.00000627, 0.00003018, 0.00000004, 0.00000931, -0.00001692, -0.00001948, 0.00001330, -0.00000924, 0.00000242, -0.00000185, -0.00000180, -0.00000367, 0.00000058, -0.00000254, 0.00000154, -0.00000472, 0.00000563, -0.00000569, 0.00000244, -0.00000422, -0.00000338, -0.00000036, -0.00001207, -0.00000256, -0.00000199, -0.00000405, -0.00003796, -0.00000768, -0.00014359, -0.00005138, 0.00054964, -0.00000595, -0.00017017, -0.00000440, -0.00002666, -0.00000058, -0.00001789, -0.00000467, 0.00000035, -0.00000386, -0.00000371, 0.00000068, -0.00000813, -0.00000074, 0.00000195, -0.00000177, -0.00000270, -0.00000477, 0.00000117, 0.00000318, 0.00000232, -0.00000661, 0.00001057, -0.00000475, -0.00000176, 0.00000815, -0.00001489, 0.00001480, -0.00002764, 0.00001269, -0.00000900, 0.00001176, -0.00002095, 0.00001662, -0.00002416, 0.00000692, 0.00000067, -0.00000230, 0.00001579, -0.00000198, -0.00000049, -0.00000035, -0.00000474, 0.00000131, 0.00000715, 0.00000241, -0.00000318, 0.00000263, -0.00000646, 0.00000175, -0.00000283, -0.00000359, 0.00000838, -0.00000065, -0.00000124, -0.00000490, 0.00001230, -0.00000648, 0.00000922, -0.00000529, -0.00000011, -0.00000557, 0.00001555, -0.00000936, 0.00000395, -0.00000690, 0.00000780, -0.00001309, 0.00000805, -0.00001202, 0.00001094, 0.00000171, -0.00001119, -0.00000469, 0.00000507, -0.00000048, 0.00000095, 0.00000305, -0.00000525, 0.00000348, -0.00000650, 0.00000517, 0.00000018, 0.00000178, -0.00000714, -0.00000195, 0.00000221, },
		{ -0.00000588, -0.00000021, -0.00000311, -0.00000488, -0.00000036, -0.00000209, -0.00002269, 0.00001255, -0.00000668, 0.00003254, 0.00000048, 0.00002133, 0.00001251, 0.00000168, -0.00000312, -0.00003203, 0.00001598, -0.00000963, 0.00000261, -0.00000297, 0.00000096, -0.00000695, 0.00000316, -0.00000523, 0.00000459, -0.00000747, 0.00000655, -0.00000501, 0.00000174, -0.00000323, -0.00000335, -0.00000134, -0.00000780, -0.00000829, 0.00000175, -0.00000503, -0.00002769, -0.00001974, -0.00008722, -0.00014686, 0.00046053, 0.00017685, -0.00021028, -0.00001390, -0.00002967, 0.00000050, -0.00001847, -0.00000540, -0.00000291, 0.00000074, -0.00000801, 0.00000407, -0.00000952, -0.00000068, 0.00000007, 0.00000131, -0.00000536, -0.00000168, -0.00000313, 0.00000561, 0.00000078, -0.00000309, 0.00000415, 0.00000417, -0.00001135, 0.00001665, -0.00002111, 0.00001925, -0.00002625, 0.00000700, -0.00000578, 0.00000964, -0.00001591, 0.00000770, -0.00001005, -0.00000960, 0.00001276, -0.00001141, 0.00001883, -0.00000041, -0.00000006, -0.00000156, -0.00000223, -0.00000262, 0.00000865, 0.00000254, -0.00000186, 0.00000061, -0.00000281, -0.00000311, 0.00000288, -0.00000928, 0.00001148, -0.00000314, 0.00000271, -0.00000922, 0.00001491, -0.00000747, 0.00000868, -0.00000360, 0.00000126, -0.00000871, 0.00001474, -0.00000495, 0.00000086, -0.00000531, 0.00000588, -0.00000890, 0.00000165, -0.00000462, 0.00000126, 0.00001061, -0.00001478, -0.00000273, 0.00000078, 0.00000364, -0.00000400, 0.00000796, -0.00000788, 0.00000535, -0.00000735, 0.00000446, 0.00000012, 0.00000320, -0.00000715, -0.00000224, -0.00000024, 0.00000617, },
		{ -0.00000090, -0.00000315, -0.00000304, -0.00000525, 0.00000520, -0.00002604, 0.00000748, -0.00000335, 0.00002379, 0.00001347, 0.00000873, 0.00002588, -0.00000539, 0.00000884, -0.00003902, 0.00001271, -0.00000628, 0.00000043, -0.00000244, 0.00000190, -0.00000763, 0.00000271, -0.00000484, 0.00000414, -0.00000602, 0.00000319, -0.00000095, -0.00000147, -0.00000033, -0.00000488, -0.00000079, -0.00000592, -0.00000987, 0.00000055, -0.00000254, -0.00002108, -0.00002720, -0.00004862, -0.00018355, 0.00032203, 0.00032668, -0.00018759, -0.00005504, -0.00001827, -0.00000904, -0.00001147, -0.00001110, -0.00000233, 0.00000088, -0.00000747, 0.00000264, -0.00000670, -0.00000352, 0.00000029, 0.00000208, -0.00000543, -0.00000135, -0.00000408, 0.00000465, 0.00000169, -0.00000157, 0.00000027, 0.00000942, -0.00001564, 0.00001833, -0.00001943, 0.00001469, -0.00001603, -0.00000505, 0.00000222, 0.00000220, -0.00000525, -0.00000568, 0.00000541, -0.00002290, 0.00001875, -0.00001363, 0.00001566, 0.00000443, -0.00000066, -0.00000215, -0.00000052, -0.00000496, 0.00000826, 0.00000364, -0.00000071, -0.00000146, 0.00000074, -0.00000721, 0.00000686, -0.00001213, 0.00001134, -0.00000259, 0.00000343, -0.00000987, 0.00001334, -0.00000398, 0.00000411, 0.00000080, 0.00000002, -0.00000736, 0.00000871, 0.00000304, -0.00000436, -0.00000135, 0.00000080, -0.00000150, -0.00000679, 0.00000330, -0.00000732, 0.00001674, -0.00001549, -0.00000255, -0.00000207, 0.00000640, -0.00000736, 0.00001023, -0.00000765, 0.00000487, -0.00000628, 0.00000270, 0.00000033, 0.00000422, -0.00000666, -0.00000246, -0.00000252, 0.00000752, -0.00000484, },
		{ -0.00000201, -0.00000263, -0.00000716, 0.00000634, -0.00002052, -0.00000511, 0.00000529, 0.00001055, 0.00002799, -0.00000248, 0.00003541, -0.00000785, 0.00001541, -0.00003819, 0.00000319, 0.00000012, -0.00000350, -0.00000057, 0.00000104, -0.00000585, -0.00000022, -0.00000223, 0.00000133, -0.00000211, -0.00000190, 0.00000353, -0.00000430, 0.00000189, -0.00000573, -0.00000060, -0.00000465, -0.00000961, -0.00000227, -0.00000018, -0.00001509, -0.00003194, -0.00002467, -0.00018125, 0.00017690, 0.00040967, -0.00010132, -0.00011281, -0.00000479, -0.00002056, -0.00000369, -0.00001666, -0.00000148, -0.00000064, -0.00000481, -0.00000077, -0.00000255, -0.00000667, 0.00000072, 0.00000184, -0.00000396, -0.00000282, -0.00000288, 0.00000183, 0.00000357, -0.00000103, -0.00000177, 0.00001153, -0.00001531, 0.00001434, -0.00001185, 0.00000439, -0.00000181, -0.00001787, 0.00000966, -0.00000514, 0.00000474, -0.00001622, 0.00001467, -0.00002681, 0.00001501, -0.00000777, 0.00000700, 0.00001119, -0.00000161, -0.00000195, -0.00000016, -0.00000523, 0.00000592, 0.00000560, 0.00000008, -0.00000295, 0.00000308, -0.00000916, 0.00000772, -0.00001106, 0.00000740, 0.00000120, 0.00000105, -0.00000720, 0.00000822, 0.00000280, -0.00000265, 0.00000607, -0.00000244, -0.00000327, 0.00000041, 0.00001099, -0.00000847, 0.00000212, -0.00000426, 0.00000516, -0.00001286, 0.00000752, -0.00001119, 0.00001743, -0.00001157, -0.00000546, -0.00000230, 0.00000658, -0.00000773, 0.00000871, -0.00000408, 0.00000197, -0.00000333, -0.00000017, 0.00000129, 0.00000394, -0.00000466, -0.00000385, -0.00000304, 0.00000616, -0.00000146, -0.00000308, },
		{ -0.00000433, -0.00000520, 0.00000143, -0.00000870, -0.00002029, 0.00001358, -0.00000125, 0.00003801, -0.00000705, 0.00003716, -0.00000311, 0.00001501, -0.00002956, -0.00001076, 0.00000750, -0.00000765, 0.00000161, -0.00000077, -0.00000274, -0.00000401, 0.00000071, -0.00000171, 0.00000167, -0.00000571, 0.00000551, -0.00000435, 0.00000152, -0.00000431, -0.00000221, -0.00000237, -0.00000987, -0.00000354, -0.00000099, -0.00000723, -0.00003622, -0.00001016, -0.00015918, 0.00005429, 0.00041871, 0.00002452, -0.00015941, -0.00000592, -0.00002381, -0.00000259, -0.00001669, -0.00000427, -0.00000041, -0.00000350, -0.00000268, -0.00000040, -0.00000739, -0.00000067, 0.00000229, -0.00000271, -0.00000410, -0.00000182, -0.00000034, 0.00000418, 0.00000034, -0.00000355, 0.00001215, -0.00001236, 0.00000752, -0.00000222, -0.00000660, 0.00001040, -0.00002560, 0.00001156, -0.00000768, 0.00000908, -0.00001900, 0.00001399, -0.00002004, 0.00000296, 0.00000300, -0.00000338, 0.00001676, -0.00000130, -0.00000158, -0.00000080, -0.00000399, 0.00000250, 0.00000761, 0.00000087, -0.00000372, 0.00000379, -0.00000854, 0.00000536, -0.00000660, 0.00000083, 0.00000661, -0.00000261, -0.00000315, 0.00000172, 0.00001013, -0.00000850, 0.00000945, -0.00000384, 0.00000075, -0.00000638, 0.00001521, -0.00000863, 0.00000268, -0.00000657, 0.00000796, -0.00001362, 0.00000594, -0.00000923, 0.00001250, -0.00000356, -0.00001074, -0.00000051, 0.00000453, -0.00000529, 0.00000397, 0.00000171, -0.00000218, 0.00000055, -0.00000355, 0.00000295, 0.00000218, -0.00000096, -0.00000670, -0.00000146, 0.00000214, 0.00000379, -0.00000641, 0.00000047, },
		{ -0.00000133, -0.00000595, 0.00000354, -0.00003108, 0.00001535, -0.00000632, 0.00003910, -0.00000214, 0.00003011, 0.00000846, 0.00000891, -0.00001575, -0.00002563, 0.00001299, -0.00001020, 0.00000283, -0.00000237, 0.00000022, -0.00000692, 0.00000225, -0.00000325, 0.00000343, -0.00000645, 0.00000387, -0.00000129, -0.00000130, -0.00000096, -0.00000528, 0.00000084, -0.00001113, -0.00000257, -0.00000496, 0.00000174, -0.00003934, -0.00000273, -0.00012912, -0.00003489, 0.00036986, 0.00015303, -0.00016813, -0.00003177, -0.00001473, -0.00001058, -0.00000996, -0.00001142, 0.00000252, -0.00000492, -0.00000161, -0.00000150, -0.00000486, -0.00000426, 0.00000389, -0.00000253, -0.00000396, -0.00000242, -0.00000034, 0.00000237, 0.00000329, -0.00000580, 0.00001227, -0.00000842, 0.00000043, 0.00000598, -0.00001412, 0.00001636, -0.00002513, 0.00000618, -0.00000416, 0.00000668, -0.00001381, 0.00000500, -0.00000667, -0.00001165, 0.00001276, -0.00001045, 0.00001792, 0.00000168, -0.00000172, -0.00000160, -0.00000241, -0.00000066, 0.00000864, 0.00000213, -0.00000393, 0.00000307, -0.00000591, 0.00000093, -0.00000062, -0.00000597, 0.00001112, -0.00000509, -0.00000008, -0.00000358, 0.00001509, -0.00001060, 0.00000886, -0.00000252, 0.00000251, -0.00000910, 0.00001388, -0.00000409, -0.00000012, -0.00000540, 0.00000615, -0.00000907, -0.00000042, -0.00000332, 0.00000442, 0.00000579, -0.00001588, 0.00000126, 0.00000191, -0.00000160, -0.00000198, 0.00000741, -0.00000558, 0.00000378, -0.00000630, 0.00000450, -0.00000030, 0.00000340, -0.00000986, 0.00000094, -0.00000277, 0.00000885, -0.00000911, 0.00000281, -0.00000676, },
		{ -0.00001108, 0.00001026, -0.00003222, 0.00000747, -0.00000301, 0.00003084, 0.00001089, 0.00001708, 0.00002322, 0.00000103, -0.00000118, -0.00003712, 0.00001382, -0.00000970, 0.00000218, -0.00000285, 0.00000189, -0.00000782, 0.00000159, -0.00000276, 0.00000282, -0.00000431, -0.00000030, 0.00000309, -0.00000455, 0.00000231, -0.00000789, 0.00000330, -0.00001192, -0.00000115, -0.00000925, 0.00000835, -0.00003819, -0.00000301, -0.00009568, -0.00009326, 0.00029000, 0.00025086, -0.00012686, -0.00007806, 0.00000056, -0.00002273, -0.00000076, -0.00001938, 0.00000571, -0.00000739, 0.00000096, -0.00000433, -0.00000088, -0.00000832, 0.00000539, -0.00000277, -0.00000270, -0.00000444, 0.00000141, -0.00000099, 0.00000674, -0.00000766, 0.00001147, -0.00000383, -0.00000572, 0.00001107, -0.00001650, 0.00001522, -0.00001729, -0.00000428, 0.00000282, 0.00000045, -0.00000461, -0.00000667, 0.00000630, -0.00002182, 0.00001599, -0.00001072, 0.00001330, 0.00000729, -0.00000238, -0.00000201, -0.00000142, -0.00000257, 0.00000807, 0.00000407, -0.00000372, 0.00000142, -0.00000235, -0.00000381, 0.00000460, -0.00001054, 0.00001251, -0.00000460, 0.00000045, -0.00000608, 0.00001600, -0.00000783, 0.00000411, 0.00000139, 0.00000166, -0.00000750, 0.00000789, 0.00000332, -0.00000449, -0.00000221, 0.00000148, -0.00000185, -0.00000805, 0.00000276, -0.00000296, 0.00001285, -0.00001788, 0.00000070, 0.00000061, 0.00000139, -0.00000676, 0.00001061, -0.00000636, 0.00000496, -0.00000725, 0.00000485, -0.00000211, 0.00000664, -0.00001139, 0.00000201, -0.00000602, 0.00001114, -0.00000892, 0.00000298, -0.00000756, 0.00000123, },
		{ 0.00000872, -0.00002306, -0.00000808, 0.00000557, 0.00001719, 0.00002687, 0.00000383, 0.00003563, -0.00000368, 0.00000954, -0.00004169, 0.00000846, -0.00000577, -0.00000054, -0.00000190, 0.00000183, -0.00000655, -0.00000091, -0.00000094, 0.00000086, -0.00000094, -0.00000462, 0.00000613, -0.00000575, 0.00000331, -0.00000806, 0.00000319, -0.00001039, -0.00000167, -0.00001055, 0.00000917, -0.00003015, -0.00001143, -0.00006087, -0.00012814, 0.00020335, 0.00030075, -0.00004392, -0.00012635, 0.00000849, -0.00002983, 0.00000373, -0.00002269, 0.00000516, -0.00000766, 0.00000196, -0.00000584, 0.00000158, -0.00001016, 0.00000480, -0.00000198, -0.00000172, -0.00000632, 0.00000298, -0.00000365, 0.00000851, -0.00000736, 0.00000859, 0.00000184, -0.00001082, 0.00001302, -0.00001443, 0.00000908, -0.00000593, -0.00001508, 0.00000868, -0.00000495, 0.00000316, -0.00001483, 0.00001275, -0.00002300, 0.00001056, -0.00000424, 0.00000451, 0.00001361, -0.00000255, -0.00000212, -0.00000127, -0.00000306, 0.00000609, 0.00000630, -0.00000303, -0.00000067, 0.00000101, -0.00000727, 0.00000730, -0.00001143, 0.00000999, -0.00000095, -0.00000146, -0.00000580, 0.00001308, -0.00000125, -0.00000287, 0.00000600, -0.00000032, -0.00000345, 0.00000023, 0.00001019, -0.00000753, 0.00000045, -0.00000312, 0.00000440, -0.00001306, 0.00000541, -0.00000650, 0.00001505, -0.00001505, -0.00000321, 0.00000145, 0.00000255, -0.00000881, 0.00001000, -0.00000385, 0.00000366, -0.00000594, 0.00000335, -0.00000213, 0.00000728, -0.00000978, 0.00000018, -0.00000587, 0.00000913, -0.00000481, -0.00000007, -0.00000531, 0.00000031, -0.00001103, },
		{ -0.00000799, -0.00002496, 0.00001309, 0.00000446, 0.00003916, -0.00000349, 0.00004068, -0.00000150, 0.00001356, -0.00003798, -0.00000263, 0.00000056, -0.00000456, 0.00000005, 0.00000041, -0.00000382, -0.00000411, 0.00000091, -0.00000100, 0.00000179, -0.00000692, 0.00000598, -0.00000371, 0.00000130, -0.00000529, 0.00000011, -0.00000611, -0.00000501, -0.00000763, 0.00000360, -0.00001600, -0.00002528, -0.00002823, -0.00014463, 0.00012348, 0.00030478, 0.00005625, -0.00015256, -0.00000381, -0.00002474, -0.00000139, -0.00001800, -0.00000141, -0.00000377, -0.00000065, -0.00000395, 0.00000048, -0.00000819, 0.00000108, 0.00000072, -0.00000213, -0.00000653, 0.00000251, -0.00000362, 0.00000690, -0.00000377, 0.00000308, 0.00000857, -0.00001476, 0.00001239, -0.00000969, 0.00000122, 0.00000446, -0.00002171, 0.00000966, -0.00000605, 0.00000586, -0.00001605, 0.00001052, -0.00001536, -0.00000105, 0.00000528, -0.00000446, 0.00001759, -0.00000079, -0.00000246, -0.00000158, -0.00000268, 0.00000354, 0.00000806, -0.00000160, -0.00000281, 0.00000336, -0.00000858, 0.00000692, -0.00000875, 0.00000446, 0.00000441, -0.00000429, -0.00000413, 0.00000806, 0.00000654, -0.00000903, 0.00000871, -0.00000135, 0.00000042, -0.00000576, 0.00001344, -0.00000701, 0.00000070, -0.00000538, 0.00000704, -0.00001315, 0.00000317, -0.00000538, 0.00001221, -0.00000797, -0.00000947, 0.00000360, 0.00000218, -0.00000811, 0.00000601, 0.00000099, 0.00000077, -0.00000296, 0.00000029, -0.00000026, 0.00000507, -0.00000502, -0.00000443, -0.00000253, 0.00000350, 0.00000214, -0.00000527, -0.00000072, -0.00000372, -0.00000628, 0.00000071, },
		{ -0.00003565, 0.00001346, -0.00000207, 0.00004286, -0.00000127, 0.00003642, 0.00000814, 0.00001115, -0.00002748, -0.00001682, 0.00000697, -0.00000839, 0.00000200, -0.00000149, -0.00000078, -0.00000672, 0.00000164, -0.00000172, 0.00000279, -0.00000641, 0.00000266, 0.00000075, -0.00000249, -0.00000097, -0.00000440, -0.00000060, -0.00000976, -0.00000216, -0.00000540, -0.00000026, -0.00003854, -0.00000361, -0.00014381, 0.00005336, 0.00027821, 0.00014489, -0.00013908, -0.00003968, -0.00000828, -0.00001482, -0.00000702, -0.00001224, 0.00000302, -0.00000613, 0.00000080, -0.00000361, -0.00000322, -0.00000473, 0.00000455, -0.00000375, -0.00000479, -0.00000045, -0.00000060, 0.00000213, 0.00000234, -0.00000388, 0.00001499, -0.00001661, 0.00000942, -0.00000383, -0.00000569, 0.00001083, -0.00002202, 0.00000488, -0.00000246, 0.00000334, -0.00001105, 0.00000206, -0.00000343, -0.00001318, 0.00001238, -0.00000926, 0.00001673, 0.00000367, -0.00000328, -0.00000186, -0.00000223, 0.00000134, 0.00000867, 0.00000068, -0.00000466, 0.00000432, -0.00000766, 0.00000401, -0.00000388, -0.00000199, 0.00000917, -0.00000600, -0.00000288, 0.00000308, 0.00001279, -0.00001170, 0.00000772, -0.00000010, 0.00000236, -0.00000806, 0.00001187, -0.00000273, -0.00000154, -0.00000485, 0.00000567, -0.00000872, -0.00000267, -0.00000137, 0.00000651, 0.00000072, -0.00001547, 0.00000501, 0.00000164, -0.00000592, 0.00000043, 0.00000606, -0.00000197, 0.00000030, -0.00000321, 0.00000253, 0.00000120, 0.00000126, -0.00000990, 0.00000188, -0.00000314, 0.00000900, -0.00000989, 0.00000371, -0.00000824, -0.00000014, -0.00000877, 0.00000589, },
		{ 0.00000434, -0.00000099, 0.00003719, 0.00000974, 0.00002506, 0.00002215, 0.00000561, -0.00001406, -0.00003009, 0.00001061, -0.00001034, 0.00000278, -0.00000283, 0.00000144, -0.00000778, 0.00000079, -0.00000118, 0.00000219, -0.00000383, -0.00000211, 0.00000529, -0.00000559, 0.00000244, -0.00000784, 0.00000355, -0.00001319, 0.00000243, -0.00001299, 0.00001105, -0.00004453, 0.00000745, -0.00012563, -0.00000898, 0.00023878, 0.00020167, -0.00008394, -0.00008882, 0.00000969, -0.00002896, 0.00000387, -0.00002203, 0.00000858, -0.00001112, 0.00000539, -0.00000785, 0.00000187, -0.00000991, 0.00000736, -0.00000509, -0.00000232, -0.00000459, 0.00000372, -0.00000359, 0.00000838, -0.00000961, 0.00001867, -0.00001486, 0.00000392, 0.00000231, -0.00001044, 0.00001254, -0.00001690, -0.00000348, 0.00000323, -0.00000150, -0.00000369, -0.00000739, 0.00000661, -0.00002000, 0.00001284, -0.00000762, 0.00001072, 0.00001001, -0.00000404, -0.00000197, -0.00000216, 0.00000001, 0.00000795, 0.00000351, -0.00000582, 0.00000380, -0.00000507, -0.00000012, 0.00000123, -0.00000710, 0.00001136, -0.00000522, -0.00000310, -0.00000046, 0.00001580, -0.00000983, 0.00000309, 0.00000304, 0.00000229, -0.00000677, 0.00000666, 0.00000344, -0.00000459, -0.00000287, 0.00000195, -0.00000229, -0.00000898, 0.00000236, 0.00000114, 0.00000776, -0.00001837, 0.00000372, 0.00000231, -0.00000375, -0.00000459, 0.00000919, -0.00000289, 0.00000247, -0.00000577, 0.00000472, -0.00000237, 0.00000655, -0.00001353, 0.00000447, -0.00000758, 0.00001253, -0.00001104, 0.00000517, -0.00001016, 0.00000344, -0.00001447, 0.00001255, -0.00003541, },
		{ 0.00000460, 0.00002558, 0.00002490, 0.00001204, 0.00003516, 0.00000173, -0.00000225, -0.00003855, 0.00000926, -0.00000930, 0.00000166, -0.00000289, 0.00000214, -0.00000702, -0.00000134, 0.00000000, 0.00000091, -0.00000078, -0.00000606, 0.00000748, -0.00000587, 0.00000303, -0.00000831, 0.00000437, -0.00001309, 0.00000325, -0.00001534, 0.00001378, -0.00003961, 0.00000336, -0.00009298, -0.00006397, 0.00019768, 0.00022188, -0.00000221, -0.00013069, 0.00001501, -0.00003430, 0.00000734, -0.00002506, 0.00000851, -0.00001193, 0.00000642, -0.00000898, 0.00000394, -0.00001166, 0.00000701, -0.00000443, -0.00000089, -0.00000774, 0.00000669, -0.00000724, 0.00001132, -0.00001133, 0.00001745, -0.00000853, -0.00000394, 0.00000815, -0.00001286, 0.00001067, -0.00000911, -0.00001173, 0.00000737, -0.00000490, 0.00000166, -0.00001295, 0.00001021, -0.00001867, 0.00000604, -0.00000072, 0.00000199, 0.00001579, -0.00000346, -0.00000231, -0.00000238, -0.00000060, 0.00000636, 0.00000621, -0.00000584, 0.00000195, -0.00000166, -0.00000405, 0.00000488, -0.00000942, 0.00001021, -0.00000189, -0.00000463, -0.00000238, 0.00001547, -0.00000432, -0.00000332, 0.00000630, 0.00000152, -0.00000361, 0.00000046, 0.00000862, -0.00000617, -0.00000138, -0.00000174, 0.00000307, -0.00001260, 0.00000313, -0.00000148, 0.00001095, -0.00001671, -0.00000090, 0.00000450, -0.00000240, -0.00000767, 0.00000918, -0.00000135, 0.00000294, -0.00000652, 0.00000515, -0.00000398, 0.00000882, -0.00001329, 0.00000330, -0.00000782, 0.00001102, -0.00000744, 0.00000247, -0.00000795, 0.00000236, -0.00001380, 0.00000998, -0.00002468, -0.00001149, },
		{ 0.00001356, 0.00003797, 0.00000342, 0.00004205, 0.00000338, 0.00000470, -0.00004003, 0.00000235, -0.00000537, -0.00000132, -0.00000160, 0.00000132, -0.00000482, -0.00000394, 0.00000091, -0.00000003, 0.00000135, -0.00000760, 0.00000617, -0.00000284, 0.00000066, -0.00000574, 0.00000177, -0.00000926, -0.00000021, -0.00001171, 0.00000796, -0.00002527, -0.00001172, -0.00005363, -0.00010741, 0.00015660, 0.00021519, 0.00008112, -0.00014396, -0.00000251, -0.00002573, -0.00000007, -0.00001905, 0.00000127, -0.00000721, 0.00000245, -0.00000560, 0.00000164, -0.00000894, 0.00000287, -0.00000122, -0.00000149, -0.00000836, 0.00000639, -0.00000682, 0.00000947, -0.00000792, 0.00001102, 0.00000162, -0.00001275, 0.00001264, -0.00001301, 0.00000681, -0.00000156, -0.00001677, 0.00000745, -0.00000463, 0.00000267, -0.00001262, 0.00000668, -0.00001059, -0.00000474, 0.00000733, -0.00000545, 0.00001816, -0.00000033, -0.00000324, -0.00000250, -0.00000102, 0.00000463, 0.00000805, -0.00000442, -0.00000083, 0.00000165, -0.00000671, 0.00000621, -0.00000868, 0.00000632, 0.00000272, -0.00000628, -0.00000348, 0.00001288, 0.00000272, -0.00000884, 0.00000755, 0.00000162, -0.00000060, -0.00000413, 0.00001062, -0.00000489, -0.00000157, -0.00000373, 0.00000539, -0.00001201, 0.00000020, -0.00000095, 0.00001016, -0.00001110, -0.00000760, 0.00000713, -0.00000153, -0.00000868, 0.00000632, 0.00000193, 0.00000218, -0.00000552, 0.00000355, -0.00000306, 0.00000748, -0.00000899, -0.00000154, -0.00000414, 0.00000525, -0.00000033, -0.00000315, -0.00000263, -0.00000248, -0.00000799, 0.00000090, -0.00000872, -0.00002741, 0.00000893, },
		{ 0.00004418, 0.00000308, 0.00004044, 0.00001162, 0.00000635, -0.00003492, -0.00000862, -0.00000003, -0.00000518, 0.00000039, -0.00000043, -0.00000207, -0.00000609, 0.00000089, -0.00000015, 0.00000204, -0.00000655, 0.00000208, 0.00000203, -0.00000294, -0.00000186, -0.00000245, -0.00000362, -0.00000596, -0.00000465, -0.00000241, -0.00000744, -0.00002936, -0.00001836, -0.00013231, 0.00011121, 0.00019763, 0.00014266, -0.00011719, -0.00004202, -0.00000676, -0.00001512, -0.00000706, -0.00001031, 0.00000081, -0.00000473, 0.00000074, -0.00000361, -0.00000325, -0.00000345, 0.00000331, -0.00000357, -0.00000648, 0.00000275, -0.00000254, 0.00000359, -0.00000076, 0.00000153, 0.00001270, -0.00001972, 0.00001409, -0.00001063, 0.00000204, 0.00000405, -0.00001752, 0.00000340, -0.00000105, 0.00000002, -0.00000792, -0.00000094, -0.00000044, -0.00001419, 0.00001178, -0.00000803, 0.00001541, 0.00000544, -0.00000460, -0.00000236, -0.00000166, 0.00000333, 0.00000862, -0.00000172, -0.00000389, 0.00000404, -0.00000751, 0.00000522, -0.00000571, 0.00000128, 0.00000682, -0.00000667, -0.00000470, 0.00000942, 0.00000905, -0.00001133, 0.00000566, 0.00000324, 0.00000116, -0.00000587, 0.00000897, -0.00000109, -0.00000315, -0.00000387, 0.00000457, -0.00000794, -0.00000496, 0.00000114, 0.00000715, -0.00000385, -0.00001388, 0.00000828, -0.00000013, -0.00000841, 0.00000197, 0.00000536, 0.00000131, -0.00000356, 0.00000062, -0.00000032, 0.00000361, -0.00000230, -0.00000795, 0.00000108, -0.00000191, 0.00000711, -0.00000864, 0.00000283, -0.00000794, -0.00000092, -0.00000917, 0.00000521, -0.00003650, 0.00000703, 0.00000567, },
		{ 0.00001103, 0.00003179, 0.00002407, 0.00000505, -0.00002590, -0.00002062, 0.00000439, -0.00000843, 0.00000200, -0.00000214, 0.00000028, -0.00000710, -0.00000018, 0.00000037, 0.00000170, -0.00000398, -0.00000282, 0.00000625, -0.00000533, 0.00000096, -0.00000579, 0.00000109, -0.00001081, 0.00000179, -0.00001171, 0.00000671, -0.00004080, 0.00000344, -0.00013364, 0.00005774, 0.00018190, 0.00017070, -0.00005489, -0.00009006, 0.00001120, -0.00002928, 0.00000386, -0.00002040, 0.00000740, -0.00001110, 0.00000641, -0.00000846, 0.00000214, -0.00000895, 0.00000676, -0.00000537, -0.00000362, -0.00000245, 0.00000330, -0.00000343, 0.00000678, -0.00000722, 0.00002077, -0.00002173, 0.00001090, -0.00000550, -0.00000317, 0.00000749, -0.00001492, -0.00000274, 0.00000338, -0.00000355, -0.00000242, -0.00000802, 0.00000658, -0.00001782, 0.00000978, -0.00000473, 0.00000827, 0.00001227, -0.00000534, -0.00000227, -0.00000249, 0.00000250, 0.00000808, 0.00000166, -0.00000637, 0.00000491, -0.00000643, 0.00000258, -0.00000184, -0.00000322, 0.00000894, -0.00000508, -0.00000639, 0.00000595, 0.00001333, -0.00001009, 0.00000109, 0.00000564, 0.00000194, -0.00000518, 0.00000500, 0.00000344, -0.00000469, -0.00000323, 0.00000203, -0.00000262, -0.00000973, 0.00000243, 0.00000425, 0.00000236, -0.00001729, 0.00000645, 0.00000253, -0.00000775, -0.00000229, 0.00000737, 0.00000139, -0.00000154, -0.00000252, 0.00000278, -0.00000079, 0.00000413, -0.00001300, 0.00000473, -0.00000708, 0.00001151, -0.00001089, 0.00000538, -0.00001066, 0.00000328, -0.00001506, 0.00001159, -0.00003503, -0.00000243, 0.00000307, 0.00004226, },
		{ 0.00002057, 0.00003619, 0.00000465, -0.00001655, -0.00003034, 0.00000571, -0.00000974, 0.00000226, -0.00000297, 0.00000151, -0.00000674, -0.00000195, 0.00000089, 0.00000119, -0.00000136, -0.00000639, 0.00000769, -0.00000479, 0.00000121, -0.00000660, 0.00000291, -0.00001238, 0.00000444, -0.00001573, 0.00001240, -0.00004105, 0.00000791, -0.00011255, -0.00000171, 0.00017131, 0.00016870, 0.00002413, -0.00012579, 0.00001460, -0.00003364, 0.00000687, -0.00002358, 0.00000821, -0.00001290, 0.00000795, -0.00000970, 0.00000420, -0.00001090, 0.00000699, -0.00000516, -0.00000171, -0.00000676, 0.00000765, -0.00000815, 0.00001109, -0.00001161, 0.00002268, -0.00001696, 0.00000282, 0.00000183, -0.00000829, 0.00000915, -0.00001078, -0.00000829, 0.00000595, -0.00000522, 0.00000061, -0.00001102, 0.00000753, -0.00001437, 0.00000204, 0.00000230, -0.00000022, 0.00001740, -0.00000401, -0.00000278, -0.00000313, 0.00000176, 0.00000704, 0.00000482, -0.00000748, 0.00000396, -0.00000385, -0.00000075, 0.00000164, -0.00000599, 0.00000859, -0.00000183, -0.00000804, 0.00000257, 0.00001531, -0.00000591, -0.00000436, 0.00000721, 0.00000278, -0.00000343, 0.00000081, 0.00000661, -0.00000470, -0.00000301, -0.00000052, 0.00000160, -0.00001185, 0.00000124, 0.00000300, 0.00000591, -0.00001671, 0.00000147, 0.00000617, -0.00000694, -0.00000552, 0.00000718, 0.00000271, 0.00000007, -0.00000491, 0.00000486, -0.00000377, 0.00000796, -0.00001440, 0.00000468, -0.00000814, 0.00001105, -0.00000850, 0.00000369, -0.00000909, 0.00000269, -0.00001457, 0.00000911, -0.00002431, -0.00001632, 0.00000332, 0.00003439, 0.00002377, },
		{ 0.00004350, 0.00000831, -0.00000951, -0.00003557, 0.00000288, -0.00000865, 0.00000079, -0.00000258, 0.00000142, -0.00000528, -0.00000380, 0.00000082, 0.00000116, 0.00000029, -0.00000750, 0.00000574, -0.00000135, -0.00000088, -0.00000496, 0.00000173, -0.00001033, 0.00000258, -0.00001355, 0.00000923, -0.00003082, -0.00000179, -0.00007721, -0.00005845, 0.00016003, 0.00015121, 0.00009531, -0.00013072, -0.00000444, -0.00002461, -0.00000042, -0.00001822, 0.00000212, -0.00000906, 0.00000427, -0.00000633, 0.00000196, -0.00000857, 0.00000358, -0.00000245, -0.00000175, -0.00000848, 0.00000843, -0.00000843, 0.00001029, -0.00001013, 0.00001778, -0.00000623, -0.00000824, 0.00000942, -0.00001221, 0.00000922, -0.00000652, -0.00001155, 0.00000533, -0.00000382, 0.00000006, -0.00000928, 0.00000300, -0.00000625, -0.00000766, 0.00000886, -0.00000613, 0.00001825, 0.00000029, -0.00000415, -0.00000324, 0.00000065, 0.00000610, 0.00000703, -0.00000679, 0.00000141, -0.00000053, -0.00000379, 0.00000386, -0.00000668, 0.00000625, 0.00000195, -0.00000872, -0.00000101, 0.00001545, -0.00000038, -0.00000859, 0.00000653, 0.00000449, -0.00000180, -0.00000200, 0.00000723, -0.00000271, -0.00000366, -0.00000210, 0.00000347, -0.00001062, -0.00000238, 0.00000332, 0.00000681, -0.00001283, -0.00000520, 0.00000941, -0.00000552, -0.00000763, 0.00000509, 0.00000467, 0.00000141, -0.00000616, 0.00000517, -0.00000435, 0.00000825, -0.00001161, 0.00000078, -0.00000525, 0.00000634, -0.00000251, -0.00000113, -0.00000426, -0.00000167, -0.00000921, 0.00000080, -0.00000965, -0.00002880, 0.00000234, 0.00002445, 0.00003628, 0.00001205, },
		{ 0.00001693, -0.00000535, -0.00003605, -0.00000360, -0.00000579, -0.00000201, -0.00000124, 0.00000034, -0.00000328, -0.00000517, -0.00000008, 0.00000166, 0.00000085, -0.00000636, 0.00000147, 0.00000326, -0.00000355, -0.00000245, -0.00000097, -0.00000621, -0.00000205, -0.00000742, 0.00000078, -0.00001542, -0.00001756, -0.00003985, -0.00010091, 0.00013853, 0.00013495, 0.00013927, -0.00009771, -0.00004258, -0.00000694, -0.00001414, -0.00000773, -0.00000793, -0.00000194, -0.00000264, -0.00000015, -0.00000294, -0.00000356, -0.00000188, 0.00000157, -0.00000326, -0.00000757, 0.00000552, -0.00000449, 0.00000508, -0.00000409, 0.00000832, 0.00000703, -0.00001850, 0.00001432, -0.00001332, 0.00000738, -0.00000254, -0.00001239, 0.00000210, -0.00000036, -0.00000279, -0.00000488, -0.00000365, 0.00000198, -0.00001453, 0.00001095, -0.00000680, 0.00001394, 0.00000701, -0.00000576, -0.00000295, -0.00000087, 0.00000548, 0.00000807, -0.00000447, -0.00000203, 0.00000258, -0.00000572, 0.00000445, -0.00000564, 0.00000299, 0.00000496, -0.00000770, -0.00000485, 0.00001426, 0.00000505, -0.00001023, 0.00000329, 0.00000688, -0.00000055, -0.00000305, 0.00000563, 0.00000043, -0.00000456, -0.00000283, 0.00000318, -0.00000701, -0.00000689, 0.00000382, 0.00000635, -0.00000750, -0.00001130, 0.00001055, -0.00000280, -0.00000899, 0.00000205, 0.00000622, 0.00000293, -0.00000630, 0.00000376, -0.00000271, 0.00000558, -0.00000594, -0.00000516, -0.00000050, -0.00000003, 0.00000421, -0.00000627, 0.00000103, -0.00000675, -0.00000273, -0.00000815, 0.00000251, -0.00003454, -0.00000290, 0.00001554, 0.00004443, 0.00000987, 0.00004371, },
		{ -0.00000247, -0.00003321, -0.00001195, -0.00000259, -0.00000515, 0.00000031, -0.00000105, -0.00000140, -0.00000573, -0.00000155, 0.00000218, 0.00000094, -0.00000416, -0.00000307, 0.00000669, -0.00000468, -0.00000094, -0.00000320, -0.00000231, -0.00000668, -0.00000104, -0.00000782, -0.00000155, -0.00003054, -0.00001168, -0.00012034, 0.00010071, 0.00013009, 0.00014962, -0.00003480, -0.00008551, 0.00000818, -0.00002620, 0.00000142, -0.00001661, 0.00000405, -0.00000891, 0.00000533, -0.00000734, 0.00000111, -0.00000666, 0.00000473, -0.00000471, -0.00000536, 0.00000060, 0.00000140, -0.00000168, 0.00000317, -0.00000160, 0.00001816, -0.00002382, 0.00001389, -0.00001038, 0.00000330, 0.00000140, -0.00001184, -0.00000190, 0.00000300, -0.00000540, -0.00000100, -0.00000849, 0.00000622, -0.00001544, 0.00000706, -0.00000229, 0.00000608, 0.00001398, -0.00000623, -0.00000287, -0.00000241, 0.00000491, 0.00000824, -0.00000125, -0.00000532, 0.00000454, -0.00000606, 0.00000352, -0.00000353, -0.00000015, 0.00000639, -0.00000500, -0.00000845, 0.00001184, 0.00000961, -0.00000913, -0.00000145, 0.00000874, 0.00000096, -0.00000304, 0.00000314, 0.00000318, -0.00000463, -0.00000341, 0.00000180, -0.00000288, -0.00001023, 0.00000296, 0.00000596, -0.00000266, -0.00001491, 0.00000857, 0.00000135, -0.00000986, -0.00000104, 0.00000653, 0.00000485, -0.00000551, 0.00000118, 0.00000015, 0.00000161, 0.00000032, -0.00001050, 0.00000331, -0.00000505, 0.00000864, -0.00000892, 0.00000398, -0.00000950, 0.00000125, -0.00001333, 0.00000783, -0.00003146, -0.00001233, 0.00000847, 0.00004646, 0.00001472, 0.00003767, 0.00002866, },
		{ -0.00002911, -0.00002009, -0.00000058, -0.00000760, 0.00000124, -0.00000210, -0.00000012, -0.00000552, -0.00000305, 0.00000207, 0.00000131, -0.00000219, -0.00000608, 0.00000726, -0.00000314, -0.00000137, -0.00000382, -0.00000021, -0.00000916, 0.00000265, -0.00001259, 0.00000603, -0.00003507, 0.00000147, -0.00011514, 0.00004833, 0.00013607, 0.00013417, 0.00003828, -0.00011398, 0.00000904, -0.00002919, 0.00000355, -0.00001938, 0.00000522, -0.00001124, 0.00000706, -0.00000848, 0.00000290, -0.00000848, 0.00000527, -0.00000464, -0.00000356, -0.00000401, 0.00000630, -0.00000669, 0.00000806, -0.00000808, 0.00002336, -0.00002171, 0.00000732, -0.00000366, -0.00000255, 0.00000533, -0.00001091, -0.00000497, 0.00000440, -0.00000583, 0.00000011, -0.00000928, 0.00000500, -0.00001049, -0.00000108, 0.00000455, -0.00000190, 0.00001828, -0.00000408, -0.00000365, -0.00000339, 0.00000387, 0.00000812, 0.00000199, -0.00000747, 0.00000472, -0.00000474, 0.00000149, -0.00000111, -0.00000246, 0.00000617, -0.00000144, -0.00001086, 0.00000804, 0.00001315, -0.00000613, -0.00000594, 0.00000865, 0.00000347, -0.00000288, 0.00000116, 0.00000432, -0.00000326, -0.00000430, 0.00000034, 0.00000021, -0.00001105, 0.00000009, 0.00000625, 0.00000071, -0.00001533, 0.00000376, 0.00000608, -0.00000992, -0.00000376, 0.00000542, 0.00000690, -0.00000382, -0.00000189, 0.00000300, -0.00000191, 0.00000497, -0.00001308, 0.00000418, -0.00000680, 0.00000913, -0.00000778, 0.00000336, -0.00000859, 0.00000117, -0.00001319, 0.00000587, -0.00002163, -0.00002282, 0.00000200, 0.00004296, 0.00002455, 0.00002870, 0.00003991, 0.00000174, },
		{ -0.00002657, -0.00000068, -0.00000878, 0.00000109, -0.00000243, 0.00000039, -0.00000479, -0.00000408, 0.00000094, 0.00000229, -0.00000112, -0.00000679, 0.00000481, 0.00000058, -0.00000316, -0.00000307, -0.00000004, -0.00000900, 0.00000274, -0.00001236, 0.00000633, -0.00003054, 0.00000052, -0.00009143, -0.00000919, 0.00014356, 0.00010940, 0.00009890, -0.00011280, -0.00000977, -0.00002105, -0.00000270, -0.00001532, 0.00000081, -0.00000883, 0.00000428, -0.00000572, 0.00000118, -0.00000686, 0.00000290, -0.00000267, -0.00000308, -0.00000667, 0.00000815, -0.00000783, 0.00000859, -0.00000924, 0.00002146, -0.00001258, -0.00000354, 0.00000469, -0.00000845, 0.00000847, -0.00000979, -0.00000656, 0.00000347, -0.00000379, -0.00000167, -0.00000643, -0.00000012, -0.00000277, -0.00000944, 0.00000962, -0.00000637, 0.00001774, 0.00000121, -0.00000531, -0.00000359, 0.00000212, 0.00000808, 0.00000461, -0.00000792, 0.00000307, -0.00000220, -0.00000099, 0.00000101, -0.00000373, 0.00000483, 0.00000184, -0.00001118, 0.00000284, 0.00001567, -0.00000230, -0.00000866, 0.00000592, 0.00000693, -0.00000282, 0.00000023, 0.00000372, -0.00000083, -0.00000523, -0.00000088, 0.00000167, -0.00000936, -0.00000409, 0.00000667, 0.00000281, -0.00001326, -0.00000235, 0.00000984, -0.00000854, -0.00000618, 0.00000336, 0.00000839, -0.00000116, -0.00000481, 0.00000487, -0.00000377, 0.00000686, -0.00001215, 0.00000178, -0.00000522, 0.00000609, -0.00000363, 0.00000005, -0.00000492, -0.00000194, -0.00000922, -0.00000044, -0.00000985, -0.00003006, -0.00000563, 0.00003557, 0.00003590, 0.00002094, 0.00004716, 0.00000931, -0.00002517, },
		{ -0.00000293, -0.00000875, -0.00000014, -0.00000209, 0.00000026, -0.00000388, -0.00000437, -0.00000103, 0.00000353, -0.00000072, -0.00000571, 0.00000064, 0.00000470, -0.00000468, -0.00000218, -0.00000080, -0.00000714, 0.00000030, -0.00000857, 0.00000171, -0.00002049, -0.00000864, -0.00006005, -0.00005912, 0.00014063, 0.00009166, 0.00013123, -0.00007814, -0.00004349, -0.00000690, -0.00001347, -0.00000757, -0.00000656, -0.00000376, -0.00000120, -0.00000075, -0.00000255, -0.00000325, -0.00000100, 0.00000025, -0.00000366, -0.00000712, 0.00000666, -0.00000515, 0.00000518, -0.00000587, 0.00001422, 0.00000033, -0.00001475, 0.00001123, -0.00001203, 0.00000950, -0.00000785, -0.00000732, 0.00000124, -0.00000063, -0.00000465, -0.00000236, -0.00000570, 0.00000353, -0.00001398, 0.00000981, -0.00000553, 0.00001227, 0.00000850, -0.00000690, -0.00000339, -0.00000010, 0.00000801, 0.00000649, -0.00000676, 0.00000013, 0.00000075, -0.00000314, 0.00000238, -0.00000402, 0.00000307, 0.00000392, -0.00000919, -0.00000314, 0.00001687, 0.00000169, -0.00000907, 0.00000117, 0.00001026, -0.00000223, -0.00000014, 0.00000235, 0.00000146, -0.00000540, -0.00000212, 0.00000187, -0.00000628, -0.00000802, 0.00000609, 0.00000446, -0.00001006, -0.00000786, 0.00001119, -0.00000530, -0.00000840, 0.00000101, 0.00000864, 0.00000236, -0.00000704, 0.00000523, -0.00000364, 0.00000606, -0.00000844, -0.00000265, -0.00000187, 0.00000145, 0.00000136, -0.00000379, -0.00000075, -0.00000560, -0.00000450, -0.00000690, -0.00000097, -0.00003108, -0.00001484, 0.00002565, 0.00004550, 0.00001758, 0.00004850, 0.00002048, -0.00002138, -0.00003105, },
		{ -0.00000800, -0.00000209, -0.00000137, -0.00000021, -0.00000304, -0.00000406, -0.00000316, 0.00000422, -0.00000024, -0.00000404, -0.00000332, 0.00000714, -0.00000436, -0.00000232, -0.00000134, -0.00000502, -0.00000277, -0.00000387, -0.00000407, -0.00000991, -0.00001877, -0.00003170, -0.00009109, 0.00011950, 0.00008964, 0.00013218, -0.00001953, -0.00007852, 0.00000346, -0.00002204, -0.00000137, -0.00001265, 0.00000045, -0.00000621, 0.00000354, -0.00000571, 0.00000001, -0.00000433, 0.00000248, -0.00000423, -0.00000627, 0.00000309, -0.00000046, 0.00000008, -0.00000065, 0.00000524, 0.00001245, -0.00002194, 0.00001288, -0.00001145, 0.00000742, -0.00000440, -0.00000828, -0.00000079, 0.00000184, -0.00000667, 0.00000024, -0.00000863, 0.00000544, -0.00001283, 0.00000474, -0.00000035, 0.00000413, 0.00001520, -0.00000682, -0.00000359, -0.00000209, 0.00000743, 0.00000792, -0.00000453, -0.00000316, 0.00000312, -0.00000431, 0.00000275, -0.00000354, 0.00000141, 0.00000452, -0.00000553, -0.00000864, 0.00001612, 0.00000574, -0.00000768, -0.00000396, 0.00001178, -0.00000012, -0.00000082, 0.00000145, 0.00000242, -0.00000420, -0.00000366, 0.00000146, -0.00000322, -0.00001023, 0.00000369, 0.00000624, -0.00000687, -0.00001142, 0.00000954, -0.00000056, -0.00001016, -0.00000124, 0.00000745, 0.00000625, -0.00000800, 0.00000395, -0.00000189, 0.00000353, -0.00000356, -0.00000716, 0.00000117, -0.00000251, 0.00000492, -0.00000603, 0.00000181, -0.00000754, -0.00000167, -0.00001042, 0.00000244, -0.00002586, -0.00002404, 0.00001370, 0.00005113, 0.00001991, 0.00004435, 0.00003354, -0.00001643, -0.00003408, -0.00000670, },
		{ -0.00000432, -0.00000065, -0.00000078, -0.00000232, -0.00000357, -0.00000464, 0.00000361, 0.00000100, -0.00000285, -0.00000561, 0.00000675, -0.00000168, -0.00000375, -0.00000117, -0.00000355, -0.00000502, -0.00000045, -0.00000797, -0.00000263, -0.00002473, -0.00001311, -0.00010112, 0.00008044, 0.00010154, 0.00011105, 0.00004449, -0.00009852, 0.00000101, -0.00002307, -0.00000074, -0.00001429, 0.00000115, -0.00000830, 0.00000489, -0.00000633, 0.00000110, -0.00000551, 0.00000285, -0.00000382, -0.00000531, -0.00000076, 0.00000383, -0.00000400, 0.00000343, -0.00000190, 0.00001995, -0.00002233, 0.00000846, -0.00000661, 0.00000251, 0.00000042, -0.00000983, -0.00000176, 0.00000256, -0.00000647, 0.00000002, -0.00000775, 0.00000269, -0.00000715, -0.00000315, 0.00000593, -0.00000305, 0.00001846, -0.00000373, -0.00000477, -0.00000325, 0.00000587, 0.00000922, -0.00000187, -0.00000588, 0.00000417, -0.00000413, 0.00000209, -0.00000249, 0.00000005, 0.00000400, -0.00000146, -0.00001222, 0.00001285, 0.00000994, -0.00000545, -0.00000769, 0.00001024, 0.00000386, -0.00000218, 0.00000160, 0.00000178, -0.00000184, -0.00000525, 0.00000080, -0.00000102, -0.00001018, -0.00000029, 0.00000792, -0.00000404, -0.00001277, 0.00000555, 0.00000447, -0.00001078, -0.00000341, 0.00000520, 0.00000963, -0.00000717, 0.00000126, 0.00000070, 0.00000055, 0.00000083, -0.00001009, 0.00000238, -0.00000438, 0.00000588, -0.00000576, 0.00000191, -0.00000693, -0.00000159, -0.00001031, 0.00000090, -0.00001723, -0.00003035, -0.00000002, 0.00005153, 0.00002732, 0.00003700, 0.00004560, -0.00000862, -0.00003631, -0.00001129, -0.00000713, },
		{ -0.00000019, -0.00000133, -0.00000164, -0.00000333, -0.00000494, 0.00000148, 0.00000302, -0.00000247, -0.00000587, 0.00000380, 0.00000244, -0.00000553, -0.00000080, -0.00000271, -0.00000601, 0.00000090, -0.00000890, 0.00000015, -0.00002493, -0.00000555, -0.00009231, 0.00003119, 0.00011759, 0.00008377, 0.00009454, -0.00009211, -0.00001706, -0.00001610, -0.00000595, -0.00001127, -0.00000189, -0.00000708, 0.00000292, -0.00000418, -0.00000021, -0.00000439, 0.00000131, -0.00000230, -0.00000489, -0.00000357, 0.00000604, -0.00000537, 0.00000468, -0.00000526, 0.00002129, -0.00001589, -0.00000054, 0.00000052, -0.00000345, 0.00000532, -0.00001128, -0.00000204, 0.00000183, -0.00000440, -0.00000248, -0.00000425, -0.00000251, -0.00000032, -0.00000995, 0.00000955, -0.00000616, 0.00001664, 0.00000240, -0.00000668, -0.00000352, 0.00000338, 0.00001038, 0.00000083, -0.00000748, 0.00000364, -0.00000271, 0.00000071, -0.00000116, -0.00000103, 0.00000303, 0.00000175, -0.00001292, 0.00000714, 0.00001405, -0.00000309, -0.00000904, 0.00000570, 0.00000887, -0.00000354, 0.00000232, 0.00000036, 0.00000058, -0.00000611, -0.00000028, 0.00000023, -0.00000840, -0.00000468, 0.00000859, -0.00000122, -0.00001252, 0.00000063, 0.00000826, -0.00000958, -0.00000565, 0.00000252, 0.00001161, -0.00000436, -0.00000229, 0.00000323, -0.00000178, 0.00000366, -0.00001068, 0.00000140, -0.00000405, 0.00000448, -0.00000352, 0.00000020, -0.00000453, -0.00000331, -0.00000798, -0.00000294, -0.00000906, -0.00003146, -0.00001453, 0.00004603, 0.00003792, 0.00002960, 0.00005387, 0.00000303, -0.00003774, -0.00001639, -0.00000644, -0.00000654, },
		{ -0.00000182, -0.00000093, -0.00000352, -0.00000408, -0.00000162, 0.00000512, -0.00000235, -0.00000481, -0.00000025, 0.00000615, -0.00000617, -0.00000116, -0.00000195, -0.00000616, 0.00000060, -0.00000749, -0.00000048, -0.00002086, -0.00000610, -0.00007233, -0.00001696, 0.00012595, 0.00006503, 0.00011839, -0.00005826, -0.00004510, -0.00000619, -0.00001344, -0.00000636, -0.00000651, -0.00000423, -0.00000084, -0.00000070, -0.00000265, -0.00000215, -0.00000105, -0.00000041, -0.00000484, -0.00000502, 0.00000583, -0.00000399, 0.00000323, -0.00000507, 0.00001748, -0.00000522, -0.00001061, 0.00000671, -0.00000792, 0.00000846, -0.00001130, -0.00000273, 0.00000082, -0.00000184, -0.00000544, -0.00000060, -0.00000693, 0.00000407, -0.00001248, 0.00000834, -0.00000422, 0.00001037, 0.00000994, -0.00000806, -0.00000355, 0.00000055, 0.00001092, 0.00000360, -0.00000788, 0.00000181, -0.00000060, -0.00000088, 0.00000011, -0.00000182, 0.00000214, 0.00000340, -0.00001076, -0.00000001, 0.00001719, -0.00000058, -0.00000823, -0.00000041, 0.00001301, -0.00000349, 0.00000241, -0.00000045, 0.00000173, -0.00000544, -0.00000201, 0.00000091, -0.00000598, -0.00000803, 0.00000740, 0.00000196, -0.00001149, -0.00000379, 0.00000973, -0.00000644, -0.00000788, -0.00000005, 0.00001168, 0.00000008, -0.00000581, 0.00000484, -0.00000283, 0.00000460, -0.00000917, -0.00000113, -0.00000243, 0.00000193, -0.00000069, -0.00000191, -0.00000190, -0.00000510, -0.00000556, -0.00000625, -0.00000421, -0.00002718, -0.00002750, 0.00003431, 0.00004911, 0.00002505, 0.00005668, 0.00001810, -0.00003742, -0.00002217, -0.00000603, -0.00000859, -0.00000011, },
		{ -0.00000026, -0.00000400, -0.00000264, -0.00000461, 0.00000617, -0.00000163, -0.00000364, -0.00000365, 0.00000774, -0.00000459, -0.00000276, -0.00000084, -0.00000604, -0.00000043, -0.00000517, -0.00000246, -0.00001526, -0.00001025, -0.00004967, -0.00005410, 0.00011850, 0.00006220, 0.00011510, -0.00000674, -0.00007109, -0.00000120, -0.00001822, -0.00000329, -0.00000977, -0.00000218, -0.00000412, 0.00000202, -0.00000439, -0.00000041, -0.00000280, 0.00000082, -0.00000457, -0.00000558, 0.00000398, -0.00000111, 0.00000057, -0.00000303, 0.00001117, 0.00000586, -0.00001783, 0.00000904, -0.00000897, 0.00000850, -0.00000893, -0.00000477, 0.00000061, -0.00000014, -0.00000711, 0.00000106, -0.00000828, 0.00000414, -0.00000998, 0.00000280, 0.00000112, 0.00000239, 0.00001605, -0.00000725, -0.00000420, -0.00000173, 0.00001023, 0.00000659, -0.00000732, -0.00000071, 0.00000145, -0.00000207, 0.00000099, -0.00000229, 0.00000150, 0.00000353, -0.00000674, -0.00000684, 0.00001816, 0.00000255, -0.00000635, -0.00000594, 0.00001423, -0.00000085, 0.00000101, 0.00000029, 0.00000098, -0.00000320, -0.00000421, 0.00000124, -0.00000383, -0.00000947, 0.00000420, 0.00000533, -0.00001005, -0.00000702, 0.00000877, -0.00000207, -0.00000952, -0.00000240, 0.00000997, 0.00000521, -0.00000824, 0.00000491, -0.00000242, 0.00000396, -0.00000635, -0.00000410, -0.00000072, -0.00000045, 0.00000144, -0.00000321, -0.00000024, -0.00000573, -0.00000440, -0.00000754, -0.00000323, -0.00001967, -0.00003594, 0.00001691, 0.00005789, 0.00002539, 0.00005397, 0.00003477, -0.00003382, -0.00002888, -0.00000595, -0.00001034, -0.00000053, -0.00000220, },
		{ -0.00000432, -0.00000148, -0.00000635, 0.00000531, 0.00000031, -0.00000319, -0.00000526, 0.00000652, -0.00000089, -0.00000522, 0.00000038, -0.00000580, -0.00000155, -0.00000307, -0.00000413, -0.00001035, -0.00001434, -0.00003053, -0.00007504, 0.00009400, 0.00007350, 0.00009374, 0.00004638, -0.00008232, -0.00000710, -0.00001720, -0.00000430, -0.00000993, -0.00000244, -0.00000545, 0.00000260, -0.00000430, -0.00000019, -0.00000310, 0.00000077, -0.00000364, -0.00000587, 0.00000165, 0.00000165, -0.00000156, -0.00000106, 0.00000504, 0.00001403, -0.00001965, 0.00000636, -0.00000632, 0.00000551, -0.00000439, -0.00000801, 0.00000130, 0.00000033, -0.00000687, 0.00000009, -0.00000632, 0.00000055, -0.00000434, -0.00000424, 0.00000651, -0.00000378, 0.00001813, -0.00000319, -0.00000589, -0.00000293, 0.00000802, 0.00000972, -0.00000596, -0.00000331, 0.00000280, -0.00000246, 0.00000124, -0.00000233, 0.00000101, 0.00000274, -0.00000236, -0.00001159, 0.00001604, 0.00000667, -0.00000451, -0.00000913, 0.00001154, 0.00000434, -0.00000165, 0.00000230, -0.00000102, -0.00000038, -0.00000601, 0.00000099, -0.00000219, -0.00000909, -0.00000013, 0.00000799, -0.00000796, -0.00000915, 0.00000620, 0.00000220, -0.00000983, -0.00000466, 0.00000709, 0.00000981, -0.00000865, 0.00000322, -0.00000081, 0.00000236, -0.00000316, -0.00000656, 0.00000024, -0.00000189, 0.00000229, -0.00000332, 0.00000014, -0.00000501, -0.00000455, -0.00000710, -0.00000457, -0.00001231, -0.00003760, -0.00000407, 0.00006107, 0.00003144, 0.00004720, 0.00005042, -0.00002550, -0.00003631, -0.00000660, -0.00001152, -0.00000163, -0.00000234, 0.00000015, },
		{ -0.00000131, -0.00000622, 0.00000242, 0.00000324, -0.00000346, -0.00000508, 0.00000313, 0.00000366, -0.00000726, 0.00000091, -0.00000513, -0.00000267, -0.00000159, -0.00000487, -0.00000698, -0.00001662, -0.00001756, -0.00008006, 0.00005755, 0.00009062, 0.00006807, 0.00008574, -0.00007132, -0.00002423, -0.00001139, -0.00000868, -0.00000752, -0.00000467, -0.00000491, 0.00000112, -0.00000256, -0.00000134, -0.00000210, -0.00000032, -0.00000218, -0.00000615, -0.00000038, 0.00000325, -0.00000218, -0.00000022, 0.00000067, 0.00001778, -0.00001583, -0.00000018, -0.00000157, 0.00000106, 0.00000091, -0.00001123, 0.00000192, 0.00000024, -0.00000533, -0.00000257, -0.00000266, -0.00000420, 0.00000114, -0.00000928, 0.00000877, -0.00000564, 0.00001516, 0.00000366, -0.00000804, -0.00000324, 0.00000469, 0.00001245, -0.00000374, -0.00000565, 0.00000318, -0.00000199, 0.00000091, -0.00000195, 0.00000049, 0.00000178, 0.00000103, -0.00001319, 0.00001083, 0.00001145, -0.00000315, -0.00000945, 0.00000563, 0.00001047, -0.00000404, 0.00000421, -0.00000272, 0.00000147, -0.00000634, -0.00000027, -0.00000084, -0.00000769, -0.00000420, 0.00000888, -0.00000483, -0.00001065, 0.00000317, 0.00000515, -0.00000835, -0.00000686, 0.00000371, 0.00001284, -0.00000667, 0.00000012, 0.00000135, 0.00000054, -0.00000035, -0.00000797, 0.00000020, -0.00000236, 0.00000209, -0.00000263, -0.00000026, -0.00000362, -0.00000513, -0.00000619, -0.00000607, -0.00000784, -0.00003251, -0.00002473, 0.00005578, 0.00004251, 0.00003891, 0.00006237, -0.00001185, -0.00004349, -0.00000870, -0.00001173, -0.00000356, -0.00000211, 0.00000002, -0.00000403, },
		{ -0.00000442, -0.00000167, 0.00000614, -0.00000369, -0.00000401, -0.00000086, 0.00000719, -0.00000748, -0.00000001, -0.00000369, -0.00000393, -0.00000061, -0.00000486, -0.00000493, -0.00001702, -0.00001055, -0.00007323, 0.00001756, 0.00010348, 0.00005013, 0.00010272, -0.00003942, -0.00004643, -0.00000552, -0.00001335, -0.00000482, -0.00000717, -0.00000380, -0.00000120, -0.00000035, -0.00000281, -0.00000076, -0.00000160, -0.00000080, -0.00000621, -0.00000195, 0.00000355, -0.00000144, -0.00000039, -0.00000179, 0.00001747, -0.00000822, -0.00000779, 0.00000268, -0.00000270, 0.00000504, -0.00001279, 0.00000118, 0.00000067, -0.00000369, -0.00000526, 0.00000042, -0.00000734, 0.00000367, -0.00001013, 0.00000662, -0.00000298, 0.00000840, 0.00001120, -0.00000914, -0.00000350, 0.00000122, 0.00001387, -0.00000042, -0.00000757, 0.00000261, -0.00000093, 0.00000025, -0.00000132, -0.00000013, 0.00000110, 0.00000280, -0.00001170, 0.00000368, 0.00001572, -0.00000181, -0.00000773, -0.00000143, 0.00001503, -0.00000419, 0.00000438, -0.00000255, 0.00000117, -0.00000467, -0.00000255, 0.00000040, -0.00000613, -0.00000686, 0.00000745, -0.00000071, -0.00001174, 0.00000041, 0.00000625, -0.00000542, -0.00000862, 0.00000020, 0.00001386, -0.00000266, -0.00000348, 0.00000322, -0.00000083, 0.00000162, -0.00000822, -0.00000064, -0.00000221, 0.00000137, -0.00000177, -0.00000079, -0.00000238, -0.00000522, -0.00000588, -0.00000636, -0.00000691, -0.00002328, -0.00004024, 0.00004055, 0.00005612, 0.00003232, 0.00006856, 0.00000658, -0.00004870, -0.00001309, -0.00001078, -0.00000630, -0.00000153, -0.00000073, -0.00000297, -0.00000228, },
		{ -0.00000549, 0.00000763, -0.00000286, -0.00000320, -0.00000385, 0.00000826, -0.00000518, -0.00000245, -0.00000159, -0.00000527, 0.00000003, -0.00000453, -0.00000359, -0.00001617, -0.00000788, -0.00005996, -0.00001797, 0.00010485, 0.00004578, 0.00009786, 0.00000408, -0.00006373, -0.00000531, -0.00001503, -0.00000416, -0.00000819, -0.00000354, -0.00000296, 0.00000100, -0.00000350, -0.00000007, -0.00000222, -0.00000014, -0.00000571, -0.00000330, 0.00000305, -0.00000017, -0.00000074, -0.00000307, 0.00001460, 0.00000041, -0.00001347, 0.00000418, -0.00000417, 0.00000671, -0.00001171, -0.00000157, 0.00000213, -0.00000271, -0.00000675, 0.00000139, -0.00000743, 0.00000236, -0.00000694, 0.00000126, 0.00000210, 0.00000086, 0.00001654, -0.00000756, -0.00000461, -0.00000137, 0.00001325, 0.00000402, -0.00000894, 0.00000131, 0.00000032, -0.00000037, -0.00000069, -0.00000074, 0.00000079, 0.00000307, -0.00000818, -0.00000353, 0.00001797, 0.00000040, -0.00000544, -0.00000714, 0.00001582, -0.00000092, 0.00000214, -0.00000013, -0.00000118, -0.00000163, -0.00000512, 0.00000123, -0.00000475, -0.00000783, 0.00000415, 0.00000357, -0.00001204, -0.00000205, 0.00000593, -0.00000206, -0.00000923, -0.00000333, 0.00001301, 0.00000237, -0.00000640, 0.00000398, -0.00000128, 0.00000256, -0.00000744, -0.00000199, -0.00000181, 0.00000055, -0.00000112, -0.00000111, -0.00000161, -0.00000460, -0.00000628, -0.00000556, -0.00000812, -0.00001402, -0.00004672, 0.00001661, 0.00006798, 0.00003064, 0.00006802, 0.00002810, -0.00004995, -0.00002021, -0.00000893, -0.00000946, -0.00000091, -0.00000187, -0.00000144, -0.00000391, -0.00000191, },
		{ 0.00000672, -0.00000046, -0.00000331, -0.00000499, 0.00000654, -0.00000086, -0.00000557, 0.00000045, -0.00000627, 0.00000025, -0.00000405, -0.00000262, -0.00001458, -0.00000785, -0.00004488, -0.00004389, 0.00009256, 0.00005375, 0.00007913, 0.00004613, -0.00006708, -0.00001388, -0.00001265, -0.00000626, -0.00000722, -0.00000463, -0.00000355, 0.00000092, -0.00000297, -0.00000041, -0.00000185, -0.00000038, -0.00000456, -0.00000466, 0.00000235, 0.00000078, -0.00000053, -0.00000392, 0.00001073, 0.00000769, -0.00001534, 0.00000222, -0.00000321, 0.00000592, -0.00000823, -0.00000578, 0.00000408, -0.00000223, -0.00000686, 0.00000016, -0.00000491, -0.00000146, -0.00000200, -0.00000446, 0.00000641, -0.00000422, 0.00001745, -0.00000262, -0.00000677, -0.00000264, 0.00001051, 0.00000905, -0.00000935, -0.00000060, 0.00000143, -0.00000068, -0.00000026, -0.00000118, 0.00000065, 0.00000244, -0.00000408, -0.00000897, 0.00001711, 0.00000402, -0.00000379, -0.00000988, 0.00001218, 0.00000522, -0.00000156, 0.00000340, -0.00000407, 0.00000114, -0.00000666, 0.00000101, -0.00000335, -0.00000762, 0.00000023, 0.00000671, -0.00001085, -0.00000465, 0.00000511, 0.00000050, -0.00000812, -0.00000681, 0.00001070, 0.00000727, -0.00000765, 0.00000324, -0.00000074, 0.00000257, -0.00000588, -0.00000355, -0.00000135, -0.00000024, -0.00000063, -0.00000136, -0.00000113, -0.00000368, -0.00000671, -0.00000477, -0.00000922, -0.00000826, -0.00004316, -0.00001158, 0.00007268, 0.00003620, 0.00006134, 0.00005005, -0.00004540, -0.00002981, -0.00000698, -0.00001233, -0.00000080, -0.00000293, -0.00000004, -0.00000536, 0.00000005, -0.00000764, },
		{ 0.00000309, -0.00000409, -0.00000453, 0.00000294, 0.00000395, -0.00000789, 0.00000140, -0.00000631, -0.00000023, -0.00000330, -0.00000209, -0.00001239, -0.00000928, -0.00003098, -0.00005857, 0.00006921, 0.00006783, 0.00005743, 0.00007549, -0.00005264, -0.00002952, -0.00000826, -0.00000973, -0.00000526, -0.00000636, -0.00000337, -0.00000021, -0.00000165, -0.00000141, -0.00000086, -0.00000118, -0.00000303, -0.00000594, 0.00000170, 0.00000108, 0.00000035, -0.00000452, 0.00000675, 0.00001247, -0.00001326, -0.00000221, -0.00000105, 0.00000385, -0.00000369, -0.00000998, 0.00000522, -0.00000139, -0.00000627, -0.00000222, -0.00000152, -0.00000534, 0.00000181, -0.00000769, 0.00000750, -0.00000503, 0.00001354, 0.00000471, -0.00000906, -0.00000301, 0.00000641, 0.00001355, -0.00000819, -0.00000309, 0.00000222, -0.00000062, -0.00000007, -0.00000136, 0.00000050, 0.00000155, -0.00000065, -0.00001157, 0.00001308, 0.00000871, -0.00000301, -0.00000953, 0.00000537, 0.00001197, -0.00000449, 0.00000594, -0.00000543, 0.00000188, -0.00000605, -0.00000072, -0.00000163, -0.00000705, -0.00000297, 0.00000765, -0.00000779, -0.00000767, 0.00000447, 0.00000158, -0.00000546, -0.00000974, 0.00000725, 0.00001119, -0.00000686, 0.00000121, 0.00000041, 0.00000198, -0.00000389, -0.00000510, -0.00000091, -0.00000104, -0.00000015, -0.00000177, -0.00000061, -0.00000300, -0.00000643, -0.00000495, -0.00000870, -0.00000723, -0.00003219, -0.00003714, 0.00006542, 0.00004915, 0.00005092, 0.00006916, -0.00003382, -0.00004084, -0.00000611, -0.00001413, -0.00000167, -0.00000344, 0.00000063, -0.00000585, 0.00000052, -0.00000740, 0.00000340, },
		{ -0.00000455, -0.00000351, -0.00000094, 0.00000742, -0.00000806, 0.00000055, -0.00000508, -0.00000155, -0.00000216, -0.00000213, -0.00000963, -0.00001134, -0.00001980, -0.00006308, 0.00004003, 0.00008050, 0.00004180, 0.00008681, -0.00002343, -0.00004609, -0.00000597, -0.00001217, -0.00000403, -0.00000751, -0.00000335, -0.00000152, -0.00000039, -0.00000232, 0.00000009, -0.00000191, -0.00000166, -0.00000682, 0.00000095, 0.00000096, 0.00000139, -0.00000451, 0.00000288, 0.00001473, -0.00000847, -0.00000707, 0.00000047, 0.00000201, 0.00000032, -0.00001254, 0.00000436, 0.00000052, -0.00000575, -0.00000445, 0.00000087, -0.00000716, 0.00000258, -0.00000721, 0.00000490, -0.00000203, 0.00000660, 0.00001203, -0.00000984, -0.00000349, 0.00000226, 0.00001620, -0.00000493, -0.00000608, 0.00000264, -0.00000030, -0.00000001, -0.00000138, 0.00000029, 0.00000076, 0.00000154, -0.00001134, 0.00000695, 0.00001324, -0.00000236, -0.00000733, -0.00000208, 0.00001645, -0.00000438, 0.00000571, -0.00000389, -0.00000012, -0.00000326, -0.00000359, 0.00000021, -0.00000653, -0.00000478, 0.00000625, -0.00000333, -0.00001070, 0.00000396, 0.00000150, -0.00000223, -0.00001123, 0.00000275, 0.00001378, -0.00000437, -0.00000133, 0.00000150, 0.00000129, -0.00000187, -0.00000635, -0.00000065, -0.00000175, 0.00000032, -0.00000231, -0.00000005, -0.00000269, -0.00000535, -0.00000612, -0.00000668, -0.00000944, -0.00001909, -0.00005322, 0.00004431, 0.00006641, 0.00004096, 0.00008204, -0.00001488, -0.00005157, -0.00000750, -0.00001442, -0.00000370, -0.00000327, 0.00000030, -0.00000508, -0.00000068, -0.00000511, -0.00000128, 0.00000654, },
		},
		{ { -0.00000230, -0.00000123, -0.00000726, 0.00000108, 0.00000102, -0.00000281, -0.00000070, -0.00000249, -0.00000057, 0.00000063, -0.00000188, -0.00000176, -0.00001095, -0.00000732, -0.00000864, -0.00001035, -0.00005534, 0.00001274, 0.00007504, 0.00003905, 0.00008531, 0.00001145, -0.00005635, -0.00001140, -0.00001019, -0.00000785, 0.00000000, -0.00000612, 0.00000487, -0.00000494, -0.00000940, -0.00000212, 0.00000143, 0.00000277, 0.00000819, -0.00000256, -0.00000583, -0.00000107, 0.00001001, -0.00000018, -0.00000083, -0.00000067, -0.00000194, 0.00000071, 0.00000130, 0.00000224, 0.00000079, -0.00000331, 0.00000399, -0.00000307, 0.00000031, -0.00000038, 0.00000320, 0.00000031, -0.00001189, 0.00000324, -0.00000067, -0.00000309, -0.00001033, 0.00000445, -0.00001351, -0.00000047, 0.00001540, -0.00000281, -0.00000148, -0.00000281, 0.00001540, -0.00000047, -0.00001351, 0.00000445, -0.00001033, -0.00000309, -0.00000067, 0.00000324, -0.00001189, 0.00000031, 0.00000320, -0.00000038, 0.00000031, -0.00000307, 0.00000399, -0.00000331, 0.00000079, 0.00000224, 0.00000130, 0.00000071, -0.00000194, -0.00000067, -0.00000083, -0.00000018, 0.00001001, -0.00000107, -0.00000583, -0.00000256, 0.00000819, 0.00000277, 0.00000143, -0.00000212, -0.00000940, -0.00000494, 0.00000487, -0.00000612, 0.00000000, -0.00000785, -0.00001019, -0.00001140, -0.00005635, 0.00001145, 0.00008531, 0.00003905, 0.00007504, 0.00001274, -0.00005534, -0.00001035, -0.00000864, -0.00000732, -0.00001095, -0.00000176, -0.00000188, 0.00000063, -0.00000057, -0.00000249, -0.00000070, -0.00000281, 0.00000102, 0.00000108, -0.00000726, -0.00000123, },
		{ -0.00000115, -0.00000610, -0.00000211, 0.00000296, -0.00000285, -0.00000100, -0.00000189, -0.00000181, 0.00000106, -0.00000158, -0.00000054, -0.00000869, -0.00000888, -0.00000880, -0.00000554, -0.00004471, -0.00001816, 0.00007331, 0.00003897, 0.00007015, 0.00004852, -0.00004881, -0.00002719, -0.00000422, -0.00001277, 0.00000090, -0.00000640, 0.00000242, -0.00000019, -0.00001009, -0.00000464, 0.00000153, 0.00000128, 0.00000805, 0.00000110, -0.00000600, -0.00000375, 0.00000841, 0.00000356, -0.00000241, 0.00000062, -0.00000271, 0.00000079, 0.00000061, 0.00000216, 0.00000169, -0.00000293, 0.00000215, -0.00000004, -0.00000254, 0.00000129, 0.00000015, 0.00000519, -0.00001285, 0.00000107, 0.00000042, -0.00000122, -0.00001076, 0.00000134, -0.00000571, -0.00001126, 0.00001693, 0.00000197, -0.00000279, -0.00000297, 0.00000837, 0.00001055, -0.00001629, 0.00000205, -0.00000611, -0.00000664, -0.00000061, 0.00000308, -0.00000811, -0.00000573, 0.00000593, -0.00000202, 0.00000298, -0.00000493, 0.00000417, -0.00000223, -0.00000072, 0.00000222, 0.00000191, 0.00000077, -0.00000110, -0.00000194, 0.00000055, -0.00000266, 0.00000930, 0.00000241, -0.00000513, -0.00000503, 0.00000611, 0.00000522, 0.00000091, -0.00000011, -0.00000753, -0.00000847, 0.00000393, -0.00000268, -0.00000360, -0.00000172, -0.00001452, -0.00000444, -0.00004589, -0.00002907, 0.00008404, 0.00005232, 0.00006569, 0.00004645, -0.00005677, -0.00002038, -0.00000780, -0.00000694, -0.00001087, -0.00000478, -0.00000122, -0.00000083, 0.00000091, -0.00000269, -0.00000095, -0.00000200, -0.00000123, 0.00000324, -0.00000672, -0.00000255, -0.00000184, },
		{ -0.00000412, -0.00000511, 0.00000365, -0.00000188, -0.00000165, -0.00000138, -0.00000229, 0.00000033, -0.00000047, -0.00000066, -0.00000537, -0.00001016, -0.00000896, -0.00000444, -0.00003065, -0.00003958, 0.00005842, 0.00004843, 0.00004996, 0.00007076, -0.00002398, -0.00004345, -0.00000369, -0.00001227, -0.00000286, -0.00000328, -0.00000225, 0.00000406, -0.00000901, -0.00000709, 0.00000050, 0.00000117, 0.00000606, 0.00000498, -0.00000532, -0.00000534, 0.00000498, 0.00000742, -0.00000327, 0.00000133, -0.00000295, 0.00000061, 0.00000021, 0.00000183, 0.00000208, -0.00000163, -0.00000022, 0.00000266, -0.00000418, 0.00000185, -0.00000183, 0.00000734, -0.00001022, -0.00000340, 0.00000260, -0.00000117, -0.00000805, -0.00000455, 0.00000191, -0.00001633, 0.00001077, 0.00001022, -0.00000482, -0.00000128, 0.00000111, 0.00001596, -0.00001121, -0.00000533, -0.00000010, -0.00001001, -0.00000058, 0.00000171, -0.00000330, -0.00001083, 0.00000674, -0.00000240, 0.00000409, -0.00000451, 0.00000223, 0.00000017, -0.00000249, 0.00000206, 0.00000222, 0.00000118, -0.00000042, -0.00000274, 0.00000111, -0.00000340, 0.00000659, 0.00000599, -0.00000394, -0.00000599, 0.00000245, 0.00000747, 0.00000095, 0.00000091, -0.00000504, -0.00001004, 0.00000038, 0.00000153, -0.00000640, 0.00000124, -0.00001299, -0.00000729, -0.00002610, -0.00005782, 0.00005943, 0.00007524, 0.00005268, 0.00007422, -0.00004432, -0.00003546, -0.00000652, -0.00000831, -0.00000855, -0.00000874, -0.00000050, -0.00000238, 0.00000173, -0.00000207, -0.00000172, -0.00000097, -0.00000286, 0.00000359, -0.00000436, -0.00000479, -0.00000086, -0.00000183, },
		{ -0.00000680, 0.00000258, -0.00000004, -0.00000232, -0.00000118, -0.00000214, -0.00000090, 0.00000050, -0.00000086, -0.00000269, -0.00000974, -0.00000978, -0.00000514, -0.00001797, -0.00004855, 0.00003341, 0.00006000, 0.00003468, 0.00007487, 0.00000928, -0.00004991, -0.00001265, -0.00000656, -0.00000914, 0.00000041, -0.00000608, 0.00000551, -0.00000567, -0.00000912, -0.00000150, 0.00000172, 0.00000358, 0.00000761, -0.00000331, -0.00000611, 0.00000090, 0.00000988, -0.00000238, 0.00000082, -0.00000218, -0.00000028, 0.00000049, 0.00000103, 0.00000234, -0.00000028, -0.00000189, 0.00000379, -0.00000377, 0.00000078, -0.00000204, 0.00000638, -0.00000466, -0.00000875, 0.00000457, -0.00000164, -0.00000441, -0.00000956, 0.00000521, -0.00001393, -0.00000044, 0.00001716, -0.00000368, -0.00000140, -0.00000210, 0.00001356, -0.00000024, -0.00001337, 0.00000372, -0.00001058, -0.00000211, 0.00000056, 0.00000057, -0.00001311, 0.00000456, -0.00000077, 0.00000293, -0.00000160, -0.00000130, 0.00000305, -0.00000390, 0.00000161, 0.00000214, 0.00000193, -0.00000001, -0.00000284, 0.00000058, -0.00000250, 0.00000271, 0.00000886, -0.00000228, -0.00000561, -0.00000164, 0.00000839, 0.00000203, 0.00000110, -0.00000270, -0.00000955, -0.00000422, 0.00000418, -0.00000633, 0.00000001, -0.00000698, -0.00001386, -0.00000972, -0.00006414, 0.00001496, 0.00009611, 0.00004653, 0.00008829, -0.00001659, -0.00005269, -0.00000643, -0.00001068, -0.00000545, -0.00001193, -0.00000096, -0.00000312, 0.00000131, -0.00000067, -0.00000254, -0.00000036, -0.00000326, 0.00000213, -0.00000095, -0.00000690, -0.00000029, -0.00000237, -0.00000243, },
		{ 0.00000001, 0.00000204, -0.00000264, -0.00000124, -0.00000185, -0.00000172, 0.00000043, -0.00000026, -0.00000155, -0.00000745, -0.00001099, -0.00000649, -0.00000924, -0.00004625, 0.00000478, 0.00006597, 0.00002966, 0.00006466, 0.00003977, -0.00004115, -0.00002804, -0.00000113, -0.00001281, 0.00000057, -0.00000611, 0.00000286, -0.00000042, -0.00001041, -0.00000379, 0.00000171, 0.00000208, 0.00000797, 0.00000013, -0.00000640, -0.00000251, 0.00000976, 0.00000065, -0.00000096, -0.00000041, -0.00000189, 0.00000144, -0.00000018, 0.00000266, 0.00000055, -0.00000221, 0.00000301, -0.00000151, -0.00000147, -0.00000087, 0.00000322, 0.00000182, -0.00001271, 0.00000459, -0.00000109, -0.00000194, -0.00001133, 0.00000319, -0.00000655, -0.00001106, 0.00001791, 0.00000263, -0.00000420, -0.00000117, 0.00000634, 0.00001057, -0.00001634, 0.00000195, -0.00000703, -0.00000557, 0.00000032, 0.00000236, -0.00001184, -0.00000051, 0.00000252, -0.00000008, 0.00000274, -0.00000495, 0.00000508, -0.00000416, 0.00000068, 0.00000177, 0.00000280, 0.00000026, -0.00000227, -0.00000083, -0.00000060, -0.00000120, 0.00001020, -0.00000008, -0.00000439, -0.00000500, 0.00000738, 0.00000390, 0.00000109, -0.00000115, -0.00000741, -0.00000831, 0.00000420, -0.00000374, -0.00000318, -0.00000117, -0.00001670, -0.00000500, -0.00004892, -0.00003435, 0.00009909, 0.00005563, 0.00008581, 0.00002264, -0.00006616, -0.00001040, -0.00001234, -0.00000366, -0.00001270, -0.00000328, -0.00000280, -0.00000015, 0.00000086, -0.00000276, -0.00000063, -0.00000242, -0.00000038, 0.00000220, -0.00000771, -0.00000091, -0.00000208, -0.00000174, -0.00000677, },
		{ 0.00000347, -0.00000224, -0.00000148, -0.00000175, -0.00000179, -0.00000064, 0.00000087, -0.00000141, -0.00000452, -0.00001136, -0.00000840, -0.00000457, -0.00003695, -0.00001997, 0.00006112, 0.00003502, 0.00004757, 0.00005950, -0.00001951, -0.00004128, -0.00000283, -0.00001011, -0.00000434, -0.00000230, -0.00000252, 0.00000472, -0.00001005, -0.00000599, 0.00000053, 0.00000199, 0.00000626, 0.00000415, -0.00000603, -0.00000467, 0.00000705, 0.00000499, -0.00000308, 0.00000153, -0.00000343, 0.00000230, -0.00000114, 0.00000269, 0.00000100, -0.00000151, 0.00000112, 0.00000135, -0.00000358, 0.00000040, -0.00000021, 0.00000667, -0.00001317, 0.00000150, 0.00000123, -0.00000146, -0.00000968, -0.00000226, 0.00000124, -0.00001619, 0.00001107, 0.00001189, -0.00000669, 0.00000033, -0.00000000, 0.00001515, -0.00001108, -0.00000529, -0.00000101, -0.00000936, 0.00000022, 0.00000234, -0.00000787, -0.00000687, 0.00000587, -0.00000322, 0.00000639, -0.00000666, 0.00000486, -0.00000266, -0.00000090, 0.00000142, 0.00000330, 0.00000078, -0.00000145, -0.00000239, 0.00000114, -0.00000378, 0.00000936, 0.00000280, -0.00000305, -0.00000663, 0.00000439, 0.00000599, 0.00000136, -0.00000040, -0.00000468, -0.00001057, 0.00000162, -0.00000025, -0.00000551, 0.00000095, -0.00001298, -0.00001063, -0.00002460, -0.00006851, 0.00007315, 0.00007964, 0.00007205, 0.00006384, -0.00006783, -0.00002142, -0.00001160, -0.00000487, -0.00001040, -0.00000716, -0.00000196, -0.00000196, 0.00000176, -0.00000204, -0.00000162, -0.00000100, -0.00000271, 0.00000382, -0.00000650, -0.00000290, -0.00000102, -0.00000198, -0.00000543, -0.00000311, },
		{ -0.00000095, -0.00000187, -0.00000177, -0.00000150, -0.00000182, 0.00000142, -0.00000094, -0.00000263, -0.00000972, -0.00001096, -0.00000281, -0.00002569, -0.00003536, 0.00004472, 0.00004638, 0.00003150, 0.00006568, 0.00000751, -0.00004412, -0.00001383, -0.00000276, -0.00001109, 0.00000165, -0.00000677, 0.00000666, -0.00000689, -0.00000824, -0.00000131, 0.00000238, 0.00000402, 0.00000709, -0.00000431, -0.00000589, 0.00000303, 0.00000875, -0.00000382, 0.00000222, -0.00000368, 0.00000197, -0.00000090, 0.00000176, 0.00000168, -0.00000076, -0.00000050, 0.00000318, -0.00000406, 0.00000038, -0.00000208, 0.00000802, -0.00000932, -0.00000429, 0.00000456, -0.00000220, -0.00000617, -0.00000799, 0.00000562, -0.00001439, -0.00000020, 0.00001879, -0.00000469, -0.00000117, -0.00000162, 0.00001176, 0.00000017, -0.00001356, 0.00000328, -0.00001061, -0.00000125, 0.00000175, -0.00000320, -0.00001200, 0.00000698, -0.00000430, 0.00000710, -0.00000483, 0.00000172, 0.00000061, -0.00000306, 0.00000142, 0.00000290, 0.00000199, -0.00000098, -0.00000320, 0.00000155, -0.00000404, 0.00000609, 0.00000630, -0.00000220, -0.00000606, -0.00000005, 0.00000769, 0.00000191, 0.00000020, -0.00000269, -0.00001014, -0.00000289, 0.00000271, -0.00000587, -0.00000038, -0.00000591, -0.00001807, -0.00000704, -0.00007416, 0.00002105, 0.00010609, 0.00005973, 0.00009443, -0.00005048, -0.00004053, -0.00000822, -0.00000933, -0.00000581, -0.00001136, -0.00000147, -0.00000347, 0.00000172, -0.00000070, -0.00000258, -0.00000006, -0.00000372, 0.00000322, -0.00000342, -0.00000560, 0.00000015, -0.00000248, -0.00000372, -0.00000563, 0.00000346, },
		{ -0.00000223, -0.00000175, -0.00000142, -0.00000221, 0.00000058, 0.00000057, -0.00000242, -0.00000626, -0.00001341, -0.00000312, -0.00001592, -0.00004005, 0.00002096, 0.00005656, 0.00002281, 0.00005968, 0.00003217, -0.00003416, -0.00002909, 0.00000236, -0.00001365, 0.00000100, -0.00000631, 0.00000355, -0.00000090, -0.00001042, -0.00000301, 0.00000190, 0.00000288, 0.00000763, -0.00000088, -0.00000669, -0.00000065, 0.00001003, -0.00000191, 0.00000077, -0.00000194, -0.00000007, 0.00000087, -0.00000028, 0.00000289, -0.00000061, -0.00000093, 0.00000309, -0.00000237, -0.00000133, -0.00000180, 0.00000579, -0.00000248, -0.00001058, 0.00000686, -0.00000243, -0.00000300, -0.00001137, 0.00000511, -0.00000776, -0.00001052, 0.00001877, 0.00000314, -0.00000552, 0.00000054, 0.00000418, 0.00001083, -0.00001662, 0.00000190, -0.00000751, -0.00000478, 0.00000154, 0.00000028, -0.00001378, 0.00000431, -0.00000208, 0.00000402, 0.00000047, -0.00000340, 0.00000452, -0.00000504, 0.00000161, 0.00000160, 0.00000377, -0.00000098, -0.00000283, 0.00000022, -0.00000203, 0.00000119, 0.00000967, -0.00000172, -0.00000382, -0.00000470, 0.00000824, 0.00000263, 0.00000124, -0.00000212, -0.00000728, -0.00000798, 0.00000421, -0.00000477, -0.00000256, -0.00000081, -0.00001917, -0.00000480, -0.00005348, -0.00003863, 0.00011467, 0.00006245, 0.00010500, -0.00001241, -0.00006404, -0.00000493, -0.00001505, -0.00000126, -0.00001408, -0.00000215, -0.00000437, 0.00000093, 0.00000047, -0.00000261, -0.00000036, -0.00000295, 0.00000070, 0.00000046, -0.00000779, 0.00000047, -0.00000249, -0.00000252, -0.00000675, 0.00000182, 0.00000095, },
		{ -0.00000166, -0.00000175, -0.00000166, -0.00000132, 0.00000246, -0.00000288, -0.00000273, -0.00001406, -0.00000544, -0.00000869, -0.00003674, -0.00000293, 0.00005848, 0.00002455, 0.00004545, 0.00004963, -0.00001536, -0.00003958, -0.00000150, -0.00000858, -0.00000549, -0.00000132, -0.00000300, 0.00000553, -0.00001123, -0.00000453, 0.00000022, 0.00000314, 0.00000601, 0.00000344, -0.00000691, -0.00000317, 0.00000834, 0.00000233, -0.00000214, 0.00000099, -0.00000293, 0.00000318, -0.00000245, 0.00000393, -0.00000071, -0.00000043, 0.00000160, 0.00000054, -0.00000364, -0.00000047, 0.00000188, 0.00000433, -0.00001422, 0.00000588, -0.00000059, -0.00000171, -0.00001127, 0.00000064, -0.00000009, -0.00001565, 0.00001129, 0.00001346, -0.00000859, 0.00000201, -0.00000137, 0.00001460, -0.00001109, -0.00000538, -0.00000143, -0.00000892, 0.00000133, 0.00000194, -0.00001180, -0.00000171, 0.00000254, -0.00000141, 0.00000689, -0.00000786, 0.00000685, -0.00000544, 0.00000118, 0.00000025, 0.00000517, -0.00000073, -0.00000187, -0.00000197, 0.00000079, -0.00000336, 0.00001131, -0.00000052, -0.00000148, -0.00000769, 0.00000669, 0.00000379, 0.00000241, -0.00000228, -0.00000370, -0.00001153, 0.00000332, -0.00000279, -0.00000372, -0.00000010, -0.00001267, -0.00001407, -0.00002315, -0.00008034, 0.00008921, 0.00008464, 0.00009614, 0.00003883, -0.00008224, -0.00000753, -0.00001816, 0.00000002, -0.00001351, -0.00000459, -0.00000452, -0.00000037, 0.00000101, -0.00000155, -0.00000170, -0.00000108, -0.00000237, 0.00000350, -0.00000820, -0.00000071, -0.00000164, -0.00000217, -0.00000638, -0.00000094, 0.00000275, -0.00000222, },
		{ -0.00000222, -0.00000087, -0.00000299, 0.00000322, -0.00000230, -0.00000116, -0.00001155, -0.00000967, -0.00000371, -0.00002979, -0.00002037, 0.00004875, 0.00003500, 0.00002891, 0.00005779, 0.00000586, -0.00003886, -0.00001484, 0.00000090, -0.00001335, 0.00000351, -0.00000817, 0.00000838, -0.00000866, -0.00000666, -0.00000169, 0.00000351, 0.00000397, 0.00000671, -0.00000563, -0.00000487, 0.00000485, 0.00000689, -0.00000439, 0.00000305, -0.00000463, 0.00000415, -0.00000305, 0.00000348, 0.00000001, -0.00000018, 0.00000024, 0.00000278, -0.00000466, -0.00000001, -0.00000108, 0.00000797, -0.00001290, 0.00000075, 0.00000343, -0.00000236, -0.00000835, -0.00000546, 0.00000532, -0.00001458, 0.00000006, 0.00002033, -0.00000583, -0.00000077, -0.00000141, 0.00001014, 0.00000057, -0.00001395, 0.00000323, -0.00001066, -0.00000030, 0.00000251, -0.00000752, -0.00000860, 0.00000672, -0.00000592, 0.00001076, -0.00000846, 0.00000543, -0.00000299, -0.00000081, 0.00000003, 0.00000491, 0.00000080, -0.00000149, -0.00000339, 0.00000229, -0.00000528, 0.00000960, 0.00000259, -0.00000074, -0.00000747, 0.00000249, 0.00000575, 0.00000280, -0.00000160, -0.00000172, -0.00001154, -0.00000061, 0.00000001, -0.00000414, -0.00000184, -0.00000409, -0.00002315, -0.00000311, -0.00008684, 0.00003044, 0.00011482, 0.00008055, 0.00008742, -0.00008205, -0.00002219, -0.00001514, -0.00000436, -0.00000885, -0.00000877, -0.00000384, -0.00000223, 0.00000115, -0.00000019, -0.00000288, 0.00000036, -0.00000421, 0.00000418, -0.00000612, -0.00000340, -0.00000016, -0.00000247, -0.00000503, -0.00000386, 0.00000362, -0.00000151, -0.00000165, },
		{ -0.00000065, -0.00000332, 0.00000189, 0.00000010, -0.00000210, -0.00000640, -0.00001438, -0.00000105, -0.00002235, -0.00002878, 0.00002980, 0.00004765, 0.00001706, 0.00005582, 0.00002496, -0.00002738, -0.00003053, 0.00000635, -0.00001543, 0.00000251, -0.00000748, 0.00000498, -0.00000208, -0.00000970, -0.00000273, 0.00000248, 0.00000333, 0.00000736, -0.00000221, -0.00000639, 0.00000135, 0.00000938, -0.00000389, 0.00000247, -0.00000363, 0.00000248, -0.00000109, 0.00000090, 0.00000210, -0.00000099, 0.00000015, 0.00000302, -0.00000324, -0.00000154, -0.00000158, 0.00000722, -0.00000683, -0.00000690, 0.00000772, -0.00000335, -0.00000460, -0.00001049, 0.00000665, -0.00000907, -0.00000973, 0.00001950, 0.00000351, -0.00000671, 0.00000209, 0.00000205, 0.00001124, -0.00001716, 0.00000209, -0.00000779, -0.00000412, 0.00000287, -0.00000309, -0.00001333, 0.00000739, -0.00000645, 0.00000938, -0.00000366, -0.00000006, 0.00000202, -0.00000416, 0.00000126, 0.00000259, 0.00000382, -0.00000217, -0.00000309, 0.00000124, -0.00000367, 0.00000446, 0.00000744, -0.00000186, -0.00000416, -0.00000346, 0.00000799, 0.00000212, 0.00000071, -0.00000235, -0.00000779, -0.00000678, 0.00000315, -0.00000480, -0.00000269, 0.00000014, -0.00002261, -0.00000306, -0.00006071, -0.00004039, 0.00012917, 0.00007627, 0.00011686, -0.00005413, -0.00005010, -0.00000644, -0.00001363, -0.00000163, -0.00001347, -0.00000275, -0.00000467, 0.00000140, 0.00000036, -0.00000263, 0.00000007, -0.00000370, 0.00000205, -0.00000201, -0.00000671, 0.00000119, -0.00000283, -0.00000347, -0.00000602, 0.00000308, -0.00000009, -0.00000177, -0.00000242, },
		{ -0.00000225, -0.00000098, 0.00000327, -0.00000398, -0.00000118, -0.00001685, -0.00000168, -0.00001524, -0.00003005, 0.00000854, 0.00005407, 0.00001555, 0.00004439, 0.00004023, -0.00001095, -0.00003874, 0.00000068, -0.00000821, -0.00000557, -0.00000111, -0.00000299, 0.00000590, -0.00001200, -0.00000320, -0.00000003, 0.00000421, 0.00000565, 0.00000248, -0.00000750, -0.00000119, 0.00000881, -0.00000030, -0.00000062, -0.00000031, -0.00000120, 0.00000254, -0.00000274, 0.00000464, -0.00000226, 0.00000101, 0.00000160, -0.00000006, -0.00000423, -0.00000036, 0.00000358, 0.00000099, -0.00001336, 0.00000920, -0.00000242, -0.00000225, -0.00001239, 0.00000376, -0.00000198, -0.00001464, 0.00001131, 0.00001496, -0.00001052, 0.00000372, -0.00000290, 0.00001431, -0.00001133, -0.00000536, -0.00000153, -0.00000872, 0.00000277, 0.00000019, -0.00001416, 0.00000342, -0.00000245, 0.00000311, 0.00000478, -0.00000705, 0.00000710, -0.00000701, 0.00000265, -0.00000048, 0.00000694, -0.00000288, -0.00000174, -0.00000151, 0.00000002, -0.00000187, 0.00001181, -0.00000304, -0.00000008, -0.00000849, 0.00000867, 0.00000151, 0.00000353, -0.00000413, -0.00000270, -0.00001229, 0.00000473, -0.00000526, -0.00000174, -0.00000136, -0.00001261, -0.00001681, -0.00002291, -0.00009186, 0.00010601, 0.00009347, 0.00012069, -0.00000060, -0.00008290, 0.00000114, -0.00002265, 0.00000388, -0.00001602, -0.00000254, -0.00000692, 0.00000155, -0.00000009, -0.00000086, -0.00000179, -0.00000135, -0.00000168, 0.00000240, -0.00000900, 0.00000132, -0.00000255, -0.00000241, -0.00000686, 0.00000116, 0.00000164, -0.00000185, -0.00000223, -0.00000124, },
		{ -0.00000367, 0.00000509, -0.00000429, 0.00000120, -0.00001479, -0.00000615, -0.00000824, -0.00002805, -0.00000791, 0.00004884, 0.00002508, 0.00002735, 0.00005045, 0.00000472, -0.00003445, -0.00001524, 0.00000380, -0.00001515, 0.00000528, -0.00000971, 0.00001024, -0.00001060, -0.00000469, -0.00000240, 0.00000489, 0.00000364, 0.00000623, -0.00000697, -0.00000316, 0.00000611, 0.00000453, -0.00000403, 0.00000299, -0.00000446, 0.00000534, -0.00000500, 0.00000553, -0.00000228, 0.00000129, 0.00000026, 0.00000272, -0.00000587, 0.00000030, 0.00000016, 0.00000652, -0.00001494, 0.00000566, 0.00000156, -0.00000233, -0.00001067, -0.00000208, 0.00000412, -0.00001422, 0.00000014, 0.00002187, -0.00000712, -0.00000021, -0.00000147, 0.00000879, 0.00000081, -0.00001436, 0.00000354, -0.00001084, 0.00000090, 0.00000239, -0.00001156, -0.00000356, 0.00000356, -0.00000450, 0.00001232, -0.00001107, 0.00000861, -0.00000671, 0.00000206, -0.00000199, 0.00000782, -0.00000172, -0.00000119, -0.00000365, 0.00000276, -0.00000589, 0.00001259, -0.00000156, 0.00000167, -0.00000958, 0.00000574, 0.00000275, 0.00000461, -0.00000418, 0.00000005, -0.00001356, 0.00000237, -0.00000373, -0.00000118, -0.00000445, -0.00000153, -0.00002886, 0.00000157, -0.00010161, 0.00004253, 0.00012378, 0.00010838, 0.00006274, -0.00010353, -0.00000339, -0.00002418, 0.00000252, -0.00001348, -0.00000493, -0.00000755, 0.00000044, -0.00000047, 0.00000098, -0.00000351, 0.00000083, -0.00000461, 0.00000475, -0.00000863, -0.00000059, -0.00000121, -0.00000224, -0.00000628, -0.00000163, 0.00000304, -0.00000157, -0.00000186, -0.00000220, -0.00000075, },
		{ 0.00000409, -0.00000174, -0.00000026, -0.00000850, -0.00001285, -0.00000216, -0.00002522, -0.00001651, 0.00003317, 0.00003976, 0.00001223, 0.00005285, 0.00001808, -0.00002085, -0.00003210, 0.00001041, -0.00001769, 0.00000481, -0.00000954, 0.00000716, -0.00000398, -0.00000815, -0.00000307, 0.00000349, 0.00000338, 0.00000715, -0.00000386, -0.00000532, 0.00000312, 0.00000794, -0.00000504, 0.00000370, -0.00000485, 0.00000502, -0.00000390, 0.00000327, 0.00000010, -0.00000020, 0.00000049, 0.00000331, -0.00000469, -0.00000132, -0.00000076, 0.00000738, -0.00001055, -0.00000226, 0.00000726, -0.00000375, -0.00000678, -0.00000845, 0.00000735, -0.00001007, -0.00000893, 0.00002016, 0.00000371, -0.00000775, 0.00000342, 0.00000009, 0.00001161, -0.00001780, 0.00000259, -0.00000804, -0.00000344, 0.00000396, -0.00000733, -0.00001031, 0.00000763, -0.00000887, 0.00001436, -0.00000859, 0.00000443, -0.00000206, -0.00000153, -0.00000058, 0.00000517, 0.00000225, -0.00000256, -0.00000346, 0.00000231, -0.00000535, 0.00000829, 0.00000365, -0.00000028, -0.00000581, -0.00000088, 0.00000619, 0.00000287, -0.00000094, -0.00000141, -0.00000936, -0.00000431, 0.00000046, -0.00000314, -0.00000435, 0.00000235, -0.00002750, 0.00000066, -0.00007125, -0.00003877, 0.00014188, 0.00009993, 0.00011387, -0.00009451, -0.00002794, -0.00001447, -0.00000770, -0.00000532, -0.00001031, -0.00000568, -0.00000290, 0.00000051, 0.00000111, -0.00000316, 0.00000082, -0.00000473, 0.00000357, -0.00000506, -0.00000437, 0.00000090, -0.00000281, -0.00000466, -0.00000453, 0.00000356, -0.00000079, -0.00000154, -0.00000291, 0.00000005, -0.00000463, },
		{ 0.00000257, -0.00000369, -0.00000113, -0.00001804, 0.00000062, -0.00002125, -0.00001934, 0.00001410, 0.00004984, 0.00000716, 0.00004476, 0.00003071, -0.00000593, -0.00003889, 0.00000373, -0.00000910, -0.00000426, -0.00000217, -0.00000197, 0.00000538, -0.00001190, -0.00000244, 0.00000011, 0.00000494, 0.00000541, 0.00000106, -0.00000738, 0.00000086, 0.00000847, -0.00000260, 0.00000114, -0.00000198, 0.00000149, 0.00000024, -0.00000141, 0.00000404, -0.00000282, 0.00000203, 0.00000176, -0.00000108, -0.00000472, 0.00000051, 0.00000445, -0.00000268, -0.00001080, 0.00001115, -0.00000388, -0.00000336, -0.00001256, 0.00000656, -0.00000410, -0.00001327, 0.00001110, 0.00001640, -0.00001244, 0.00000535, -0.00000444, 0.00001414, -0.00001177, -0.00000511, -0.00000144, -0.00000866, 0.00000444, -0.00000297, -0.00001414, 0.00000701, -0.00000756, 0.00000941, 0.00000014, -0.00000387, 0.00000498, -0.00000650, 0.00000257, 0.00000027, 0.00000749, -0.00000474, -0.00000149, -0.00000086, -0.00000125, 0.00000080, 0.00001038, -0.00000391, 0.00000026, -0.00000818, 0.00000947, 0.00000000, 0.00000397, -0.00000519, -0.00000245, -0.00001200, 0.00000487, -0.00000655, -0.00000071, -0.00000182, -0.00001374, -0.00001779, -0.00002535, -0.00010124, 0.00012135, 0.00011072, 0.00013817, -0.00004958, -0.00006893, 0.00000142, -0.00002225, 0.00000443, -0.00001612, -0.00000259, -0.00000770, 0.00000259, -0.00000071, -0.00000049, -0.00000155, -0.00000206, -0.00000047, 0.00000030, -0.00000844, 0.00000258, -0.00000333, -0.00000286, -0.00000668, 0.00000284, 0.00000043, -0.00000132, -0.00000303, -0.00000033, -0.00000356, 0.00000068, },
		{ -0.00000576, 0.00000323, -0.00001804, -0.00000192, -0.00001491, -0.00002062, -0.00000045, 0.00004793, 0.00001543, 0.00002762, 0.00004271, 0.00000472, -0.00003134, -0.00001455, 0.00000532, -0.00001569, 0.00000610, -0.00001063, 0.00001159, -0.00001212, -0.00000285, -0.00000307, 0.00000616, 0.00000332, 0.00000533, -0.00000785, -0.00000110, 0.00000675, 0.00000190, -0.00000281, 0.00000200, -0.00000286, 0.00000475, -0.00000566, 0.00000687, -0.00000431, 0.00000297, 0.00000004, 0.00000260, -0.00000737, 0.00000150, 0.00000100, 0.00000413, -0.00001525, 0.00000982, -0.00000054, -0.00000245, -0.00001267, 0.00000177, 0.00000211, -0.00001322, -0.00000012, 0.00002347, -0.00000856, 0.00000045, -0.00000170, 0.00000768, 0.00000081, -0.00001460, 0.00000409, -0.00001122, 0.00000246, 0.00000101, -0.00001431, 0.00000189, -0.00000179, 0.00000023, 0.00001069, -0.00001132, 0.00000989, -0.00000916, 0.00000429, -0.00000349, 0.00001060, -0.00000490, -0.00000026, -0.00000389, 0.00000277, -0.00000543, 0.00001424, -0.00000508, 0.00000405, -0.00001158, 0.00000891, -0.00000059, 0.00000673, -0.00000691, 0.00000197, -0.00001547, 0.00000517, -0.00000757, 0.00000216, -0.00000748, 0.00000104, -0.00003421, 0.00000558, -0.00011677, 0.00005524, 0.00013693, 0.00013866, 0.00001984, -0.00010892, 0.00000971, -0.00003116, 0.00000839, -0.00001754, -0.00000161, -0.00001112, 0.00000351, -0.00000252, 0.00000242, -0.00000416, 0.00000107, -0.00000460, 0.00000453, -0.00001032, 0.00000222, -0.00000272, -0.00000188, -0.00000729, 0.00000078, 0.00000186, -0.00000115, -0.00000258, -0.00000155, -0.00000150, -0.00000314, 0.00000598, },
		{ 0.00000256, -0.00001207, -0.00000903, -0.00000682, -0.00002254, -0.00000706, 0.00003404, 0.00003201, 0.00000904, 0.00004986, 0.00001210, -0.00001510, -0.00003322, 0.00001378, -0.00001955, 0.00000709, -0.00001184, 0.00000962, -0.00000620, -0.00000611, -0.00000380, 0.00000474, 0.00000325, 0.00000672, -0.00000545, -0.00000361, 0.00000449, 0.00000581, -0.00000516, 0.00000402, -0.00000490, 0.00000647, -0.00000647, 0.00000604, -0.00000263, 0.00000153, 0.00000013, 0.00000394, -0.00000685, -0.00000014, -0.00000003, 0.00000647, -0.00001313, 0.00000268, 0.00000579, -0.00000380, -0.00000932, -0.00000528, 0.00000694, -0.00001041, -0.00000840, 0.00002089, 0.00000371, -0.00000864, 0.00000452, -0.00000164, 0.00001178, -0.00001837, 0.00000333, -0.00000838, -0.00000255, 0.00000439, -0.00001167, -0.00000519, 0.00000456, -0.00000791, 0.00001708, -0.00001260, 0.00000867, -0.00000655, 0.00000197, -0.00000328, 0.00000890, -0.00000096, -0.00000185, -0.00000411, 0.00000332, -0.00000668, 0.00001197, -0.00000098, 0.00000258, -0.00000852, 0.00000281, 0.00000293, 0.00000486, -0.00000363, 0.00000058, -0.00001180, -0.00000083, -0.00000364, 0.00000018, -0.00000761, 0.00000582, -0.00003361, 0.00000584, -0.00008455, -0.00003444, 0.00015429, 0.00013351, 0.00008992, -0.00012418, -0.00000419, -0.00002567, 0.00000088, -0.00001118, -0.00000546, -0.00001035, 0.00000070, -0.00000181, 0.00000279, -0.00000420, 0.00000180, -0.00000583, 0.00000495, -0.00000825, -0.00000108, -0.00000045, -0.00000230, -0.00000608, -0.00000234, 0.00000315, -0.00000093, -0.00000180, -0.00000288, 0.00000015, -0.00000533, 0.00000629, -0.00000425, },
		{ -0.00000319, -0.00001661, -0.00000031, -0.00002363, -0.00000840, 0.00001562, 0.00004592, -0.00000041, 0.00004612, 0.00002118, -0.00000054, -0.00003966, 0.00000717, -0.00001074, -0.00000190, -0.00000437, 0.00000005, 0.00000391, -0.00001084, -0.00000236, 0.00000068, 0.00000532, 0.00000521, -0.00000072, -0.00000643, 0.00000270, 0.00000735, -0.00000419, 0.00000259, -0.00000332, 0.00000427, -0.00000311, 0.00000140, 0.00000195, -0.00000202, 0.00000214, 0.00000251, -0.00000303, -0.00000432, 0.00000156, 0.00000443, -0.00000609, -0.00000698, 0.00001166, -0.00000478, -0.00000516, -0.00001144, 0.00000849, -0.00000596, -0.00001182, 0.00001079, 0.00001775, -0.00001431, 0.00000686, -0.00000585, 0.00001388, -0.00001223, -0.00000463, -0.00000129, -0.00000864, 0.00000609, -0.00000731, -0.00001131, 0.00000768, -0.00001080, 0.00001564, -0.00000582, 0.00000105, 0.00000077, -0.00000384, 0.00000064, 0.00000298, 0.00000601, -0.00000549, -0.00000160, 0.00000010, -0.00000292, 0.00000445, 0.00000701, -0.00000275, -0.00000103, -0.00000621, 0.00000843, -0.00000005, 0.00000313, -0.00000492, -0.00000348, -0.00001016, 0.00000305, -0.00000578, -0.00000159, -0.00000062, -0.00001674, -0.00001632, -0.00003140, -0.00010730, 0.00013396, 0.00014054, 0.00013926, -0.00009897, -0.00004407, -0.00000654, -0.00001627, 0.00000087, -0.00001304, -0.00000554, -0.00000591, 0.00000185, -0.00000016, -0.00000089, -0.00000074, -0.00000333, 0.00000127, -0.00000275, -0.00000629, 0.00000260, -0.00000357, -0.00000370, -0.00000563, 0.00000375, -0.00000049, -0.00000098, -0.00000366, 0.00000050, -0.00000507, 0.00000332, 0.00000030, -0.00000167, },
		{ -0.00001964, 0.00000081, -0.00002090, -0.00000988, 0.00000193, 0.00004744, 0.00000549, 0.00002999, 0.00003404, 0.00000615, -0.00002969, -0.00001271, 0.00000534, -0.00001460, 0.00000542, -0.00001034, 0.00001190, -0.00001271, -0.00000158, -0.00000336, 0.00000710, 0.00000318, 0.00000383, -0.00000791, 0.00000095, 0.00000676, -0.00000066, -0.00000108, 0.00000039, -0.00000007, 0.00000220, -0.00000436, 0.00000664, -0.00000517, 0.00000404, 0.00000024, 0.00000172, -0.00000840, 0.00000322, 0.00000117, 0.00000129, -0.00001386, 0.00001275, -0.00000243, -0.00000308, -0.00001380, 0.00000547, -0.00000034, -0.00001167, -0.00000074, 0.00002514, -0.00001013, 0.00000112, -0.00000194, 0.00000663, 0.00000069, -0.00001463, 0.00000480, -0.00001177, 0.00000437, -0.00000188, -0.00001476, 0.00000610, -0.00000770, 0.00000736, 0.00000584, -0.00000860, 0.00000841, -0.00000926, 0.00000473, -0.00000328, 0.00001191, -0.00000767, 0.00000069, -0.00000380, 0.00000208, -0.00000357, 0.00001382, -0.00000691, 0.00000530, -0.00001243, 0.00001093, -0.00000325, 0.00000824, -0.00000889, 0.00000310, -0.00001626, 0.00000660, -0.00001014, 0.00000450, -0.00000967, 0.00000241, -0.00003779, 0.00000708, -0.00013002, 0.00006561, 0.00016014, 0.00016273, -0.00003640, -0.00009639, 0.00001292, -0.00003252, 0.00001047, -0.00001881, -0.00000074, -0.00001275, 0.00000550, -0.00000398, 0.00000342, -0.00000432, 0.00000067, -0.00000384, 0.00000310, -0.00001052, 0.00000423, -0.00000412, -0.00000168, -0.00000772, 0.00000293, 0.00000043, -0.00000047, -0.00000358, -0.00000046, -0.00000306, -0.00000111, 0.00000514, -0.00000567, 0.00000365, },
		{ -0.00000444, -0.00001359, -0.00001454, -0.00000256, 0.00003496, 0.00002327, 0.00000844, 0.00004574, 0.00000776, -0.00001072, -0.00003327, 0.00001576, -0.00002010, 0.00000837, -0.00001351, 0.00001161, -0.00000808, -0.00000411, -0.00000454, 0.00000588, 0.00000323, 0.00000571, -0.00000647, -0.00000165, 0.00000544, 0.00000315, -0.00000420, 0.00000323, -0.00000335, 0.00000594, -0.00000757, 0.00000803, -0.00000508, 0.00000342, -0.00000035, 0.00000436, -0.00000925, 0.00000203, 0.00000013, 0.00000482, -0.00001429, 0.00000724, 0.00000379, -0.00000382, -0.00001175, -0.00000138, 0.00000546, -0.00000992, -0.00000831, 0.00002179, 0.00000348, -0.00000943, 0.00000550, -0.00000318, 0.00001174, -0.00001871, 0.00000425, -0.00000886, -0.00000132, 0.00000368, -0.00001506, 0.00000085, -0.00000128, -0.00000305, 0.00001610, -0.00001399, 0.00001098, -0.00000985, 0.00000491, -0.00000552, 0.00001257, -0.00000500, -0.00000036, -0.00000483, 0.00000391, -0.00000704, 0.00001452, -0.00000525, 0.00000566, -0.00001138, 0.00000676, -0.00000102, 0.00000745, -0.00000668, 0.00000287, -0.00001428, 0.00000266, -0.00000805, 0.00000413, -0.00001160, 0.00000964, -0.00003972, 0.00001085, -0.00009858, -0.00003001, 0.00017104, 0.00017249, 0.00004313, -0.00013532, 0.00001365, -0.00003505, 0.00000867, -0.00001663, -0.00000100, -0.00001500, 0.00000486, -0.00000474, 0.00000487, -0.00000534, 0.00000257, -0.00000654, 0.00000559, -0.00001077, 0.00000242, -0.00000247, -0.00000144, -0.00000748, 0.00000027, 0.00000197, -0.00000048, -0.00000262, -0.00000214, -0.00000073, -0.00000467, 0.00000739, -0.00000621, 0.00000509, -0.00001573, },
		{ -0.00000486, -0.00002052, -0.00000073, 0.00001596, 0.00004118, -0.00000607, 0.00004730, 0.00001240, 0.00000456, -0.00004033, 0.00001014, -0.00001215, 0.00000060, -0.00000701, 0.00000256, 0.00000193, -0.00000915, -0.00000275, 0.00000147, 0.00000560, 0.00000469, -0.00000241, -0.00000488, 0.00000426, 0.00000542, -0.00000472, 0.00000316, -0.00000350, 0.00000598, -0.00000629, 0.00000480, -0.00000106, -0.00000015, 0.00000148, 0.00000372, -0.00000593, -0.00000259, 0.00000213, 0.00000370, -0.00000879, -0.00000247, 0.00001091, -0.00000519, -0.00000751, -0.00000898, 0.00000917, -0.00000714, -0.00001062, 0.00001052, 0.00001892, -0.00001612, 0.00000823, -0.00000705, 0.00001340, -0.00001250, -0.00000401, -0.00000116, -0.00000851, 0.00000732, -0.00001215, -0.00000594, 0.00000467, -0.00001039, 0.00001956, -0.00001113, 0.00000612, -0.00000426, 0.00000005, -0.00000246, 0.00000718, 0.00000251, -0.00000478, -0.00000225, 0.00000126, -0.00000456, 0.00000839, 0.00000237, 0.00000006, -0.00000377, -0.00000269, 0.00000556, 0.00000145, 0.00000102, -0.00000338, -0.00000563, -0.00000698, -0.00000055, -0.00000296, -0.00000451, 0.00000231, -0.00002144, -0.00001285, -0.00004055, -0.00011071, 0.00014510, 0.00018421, 0.00011571, -0.00013746, -0.00001590, -0.00001910, -0.00000666, -0.00000559, -0.00000768, -0.00001076, -0.00000177, -0.00000077, 0.00000166, -0.00000207, 0.00000053, -0.00000491, 0.00000315, -0.00000630, -0.00000284, 0.00000124, -0.00000305, -0.00000503, -0.00000369, 0.00000366, -0.00000081, -0.00000116, -0.00000373, 0.00000067, -0.00000592, 0.00000599, -0.00000275, 0.00000150, -0.00000692, -0.00001272, },
		{ -0.00002321, 0.00000053, 0.00000126, 0.00004708, -0.00000414, 0.00003387, 0.00002466, 0.00000875, -0.00002913, -0.00001020, 0.00000435, -0.00001221, 0.00000335, -0.00000882, 0.00001109, -0.00001225, -0.00000103, -0.00000325, 0.00000774, 0.00000308, 0.00000188, -0.00000707, 0.00000279, 0.00000604, -0.00000269, 0.00000056, -0.00000106, 0.00000301, -0.00000168, -0.00000123, 0.00000461, -0.00000445, 0.00000401, 0.00000130, -0.00000047, -0.00000813, 0.00000475, 0.00000073, -0.00000158, -0.00001103, 0.00001423, -0.00000381, -0.00000445, -0.00001360, 0.00000837, -0.00000269, -0.00000990, -0.00000159, 0.00002684, -0.00001179, 0.00000174, -0.00000204, 0.00000542, 0.00000065, -0.00001451, 0.00000561, -0.00001244, 0.00000648, -0.00000622, -0.00001225, 0.00000742, -0.00001196, 0.00001488, -0.00000101, -0.00000349, 0.00000429, -0.00000679, 0.00000294, -0.00000071, 0.00001083, -0.00000900, 0.00000104, -0.00000314, 0.00000063, -0.00000033, 0.00001112, -0.00000645, 0.00000467, -0.00001134, 0.00001092, -0.00000433, 0.00000840, -0.00000941, 0.00000278, -0.00001523, 0.00000578, -0.00001036, 0.00000461, -0.00000988, 0.00000158, -0.00003857, 0.00000474, -0.00013969, 0.00007141, 0.00019935, 0.00016922, -0.00009577, -0.00006954, 0.00000581, -0.00002702, 0.00000750, -0.00001616, -0.00000341, -0.00001124, 0.00000523, -0.00000392, 0.00000337, -0.00000364, -0.00000059, -0.00000218, 0.00000032, -0.00000883, 0.00000477, -0.00000481, -0.00000202, -0.00000725, 0.00000434, -0.00000078, 0.00000005, -0.00000443, 0.00000057, -0.00000485, 0.00000202, 0.00000253, -0.00000366, 0.00000181, -0.00001848, 0.00000025, },
		{ -0.00000359, -0.00000280, 0.00003698, 0.00001317, 0.00001077, 0.00003988, 0.00000550, -0.00000802, -0.00003207, 0.00001610, -0.00001886, 0.00000801, -0.00001388, 0.00001253, -0.00000907, -0.00000264, -0.00000497, 0.00000669, 0.00000347, 0.00000393, -0.00000653, 0.00000021, 0.00000592, 0.00000027, -0.00000247, 0.00000162, -0.00000039, 0.00000314, -0.00000638, 0.00000823, -0.00000625, 0.00000456, -0.00000019, 0.00000376, -0.00001095, 0.00000461, -0.00000036, 0.00000273, -0.00001388, 0.00001082, 0.00000178, -0.00000423, -0.00001345, 0.00000263, 0.00000328, -0.00000871, -0.00000872, 0.00002289, 0.00000302, -0.00001020, 0.00000653, -0.00000475, 0.00001164, -0.00001884, 0.00000528, -0.00000953, 0.00000035, 0.00000144, -0.00001634, 0.00000606, -0.00000825, 0.00000494, 0.00001114, -0.00001186, 0.00001017, -0.00001057, 0.00000587, -0.00000584, 0.00001465, -0.00000857, 0.00000114, -0.00000515, 0.00000365, -0.00000585, 0.00001496, -0.00000786, 0.00000763, -0.00001312, 0.00000966, -0.00000442, 0.00000956, -0.00000903, 0.00000436, -0.00001558, 0.00000472, -0.00001117, 0.00000704, -0.00001473, 0.00001226, -0.00004404, 0.00001335, -0.00011043, -0.00002949, 0.00019945, 0.00020722, -0.00002207, -0.00012465, 0.00002001, -0.00003808, 0.00001216, -0.00001891, 0.00000067, -0.00001744, 0.00000777, -0.00000697, 0.00000641, -0.00000586, 0.00000251, -0.00000627, 0.00000484, -0.00001172, 0.00000515, -0.00000443, -0.00000068, -0.00000844, 0.00000278, 0.00000040, 0.00000030, -0.00000377, -0.00000091, -0.00000251, -0.00000237, 0.00000659, -0.00000653, 0.00000601, -0.00001777, -0.00000126, -0.00001970, },
		{ 0.00000175, 0.00001745, 0.00003435, -0.00000861, 0.00004696, 0.00000535, 0.00000852, -0.00004013, 0.00001179, -0.00001229, 0.00000215, -0.00000911, 0.00000473, 0.00000015, -0.00000739, -0.00000321, 0.00000214, 0.00000611, 0.00000344, -0.00000344, -0.00000315, 0.00000556, 0.00000273, -0.00000401, 0.00000253, -0.00000198, 0.00000555, -0.00000789, 0.00000743, -0.00000388, 0.00000192, 0.00000072, 0.00000471, -0.00000915, 0.00000034, 0.00000187, 0.00000250, -0.00001042, 0.00000206, 0.00000935, -0.00000539, -0.00000996, -0.00000551, 0.00000856, -0.00000739, -0.00000994, 0.00001046, 0.00001988, -0.00001791, 0.00000958, -0.00000813, 0.00001269, -0.00001247, -0.00000332, -0.00000111, -0.00000809, 0.00000762, -0.00001647, 0.00000093, -0.00000171, -0.00000550, 0.00001936, -0.00001370, 0.00000937, -0.00000826, 0.00000353, -0.00000525, 0.00001153, -0.00000207, -0.00000306, -0.00000313, 0.00000212, -0.00000542, 0.00001148, -0.00000224, 0.00000338, -0.00000697, 0.00000144, 0.00000165, 0.00000383, -0.00000166, -0.00000137, -0.00000798, -0.00000361, -0.00000471, 0.00000077, -0.00000844, 0.00000591, -0.00002644, -0.00000926, -0.00005047, -0.00011463, 0.00015982, 0.00023782, 0.00006366, -0.00015492, 0.00000650, -0.00003038, 0.00000257, -0.00001198, -0.00000245, -0.00001621, 0.00000322, -0.00000426, 0.00000407, -0.00000344, 0.00000164, -0.00000615, 0.00000444, -0.00000941, 0.00000112, -0.00000103, -0.00000195, -0.00000658, -0.00000111, 0.00000266, -0.00000049, -0.00000194, -0.00000304, -0.00000021, -0.00000531, 0.00000751, -0.00000535, 0.00000455, -0.00001092, -0.00000785, -0.00001166, -0.00001218, },
		{ 0.00000039, 0.00004537, -0.00001202, 0.00003784, 0.00001551, 0.00001168, -0.00002884, -0.00000794, 0.00000342, -0.00000947, 0.00000060, -0.00000658, 0.00000958, -0.00001104, -0.00000099, -0.00000294, 0.00000836, 0.00000259, 0.00000001, -0.00000561, 0.00000441, 0.00000443, -0.00000370, 0.00000140, -0.00000145, 0.00000507, -0.00000554, 0.00000277, 0.00000137, -0.00000247, 0.00000305, 0.00000304, -0.00000394, -0.00000611, 0.00000543, -0.00000007, -0.00000411, -0.00000718, 0.00001429, -0.00000461, -0.00000648, -0.00001193, 0.00000997, -0.00000442, -0.00000831, -0.00000247, 0.00002848, -0.00001351, 0.00000230, -0.00000193, 0.00000391, 0.00000094, -0.00001436, 0.00000650, -0.00001312, 0.00000845, -0.00001148, -0.00000676, 0.00000476, -0.00001247, 0.00002024, -0.00000766, 0.00000234, -0.00000117, -0.00000263, -0.00000042, 0.00000379, 0.00000728, -0.00000853, 0.00000056, -0.00000199, -0.00000115, 0.00000365, 0.00000671, -0.00000394, 0.00000218, -0.00000829, 0.00000870, -0.00000358, 0.00000705, -0.00000843, 0.00000105, -0.00001252, 0.00000279, -0.00000808, 0.00000220, -0.00000786, -0.00000143, -0.00003682, -0.00000113, -0.00014620, 0.00007320, 0.00025772, 0.00014716, -0.00014483, -0.00003684, -0.00000782, -0.00001660, 0.00000066, -0.00001043, -0.00000906, -0.00000673, 0.00000255, -0.00000217, 0.00000220, -0.00000220, -0.00000248, 0.00000001, -0.00000337, -0.00000541, 0.00000362, -0.00000448, -0.00000308, -0.00000569, 0.00000468, -0.00000137, 0.00000004, -0.00000467, 0.00000089, -0.00000596, 0.00000525, -0.00000107, -0.00000027, -0.00000191, -0.00001467, -0.00000406, -0.00002002, 0.00000722, },
		{ 0.00003940, 0.00000252, 0.00001536, 0.00003248, 0.00000506, -0.00000667, -0.00003003, 0.00001524, -0.00001610, 0.00000602, -0.00001282, 0.00001220, -0.00000894, -0.00000188, -0.00000501, 0.00000723, 0.00000379, 0.00000159, -0.00000558, 0.00000181, 0.00000577, -0.00000230, -0.00000058, -0.00000006, 0.00000310, -0.00000131, -0.00000297, 0.00000631, -0.00000562, 0.00000440, 0.00000110, 0.00000149, -0.00001099, 0.00000675, -0.00000128, 0.00000043, -0.00001195, 0.00001303, 0.00000017, -0.00000534, -0.00001388, 0.00000598, 0.00000097, -0.00000710, -0.00000952, 0.00002415, 0.00000238, -0.00001101, 0.00000778, -0.00000659, 0.00001174, -0.00001889, 0.00000645, -0.00001040, 0.00000240, -0.00000242, -0.00001460, 0.00000855, -0.00001399, 0.00001397, 0.00000335, -0.00000656, 0.00000613, -0.00000824, 0.00000420, -0.00000341, 0.00001394, -0.00001043, 0.00000183, -0.00000463, 0.00000229, -0.00000290, 0.00001280, -0.00000793, 0.00000746, -0.00001269, 0.00001036, -0.00000614, 0.00001022, -0.00000978, 0.00000416, -0.00001475, 0.00000418, -0.00001154, 0.00000733, -0.00001545, 0.00001223, -0.00004502, 0.00001140, -0.00011759, -0.00003650, 0.00024772, 0.00022416, -0.00009466, -0.00009519, 0.00001346, -0.00003271, 0.00000948, -0.00001636, -0.00000199, -0.00001605, 0.00000788, -0.00000727, 0.00000653, -0.00000515, 0.00000119, -0.00000469, 0.00000233, -0.00001045, 0.00000618, -0.00000553, -0.00000059, -0.00000843, 0.00000457, -0.00000098, 0.00000092, -0.00000475, 0.00000023, -0.00000453, 0.00000116, 0.00000383, -0.00000477, 0.00000458, -0.00001703, -0.00000133, -0.00002221, 0.00000673, -0.00000561, },
		{ 0.00002505, -0.00000747, 0.00004426, 0.00000067, 0.00001088, -0.00003871, 0.00001173, -0.00001053, 0.00000197, -0.00000986, 0.00000584, -0.00000080, -0.00000609, -0.00000339, 0.00000247, 0.00000699, 0.00000124, -0.00000336, -0.00000162, 0.00000659, -0.00000046, -0.00000226, 0.00000085, 0.00000120, 0.00000254, -0.00000691, 0.00000811, -0.00000536, 0.00000315, 0.00000076, 0.00000450, -0.00001157, 0.00000373, 0.00000085, 0.00000097, -0.00001069, 0.00000593, 0.00000753, -0.00000583, -0.00001187, -0.00000168, 0.00000699, -0.00000673, -0.00000987, 0.00001069, 0.00002059, -0.00001974, 0.00001108, -0.00000931, 0.00001195, -0.00001217, -0.00000260, -0.00000120, -0.00000725, 0.00000651, -0.00001901, 0.00000752, -0.00000995, 0.00000332, 0.00001441, -0.00001220, 0.00000920, -0.00000952, 0.00000492, -0.00000600, 0.00001421, -0.00000620, -0.00000130, -0.00000356, 0.00000202, -0.00000462, 0.00001246, -0.00000526, 0.00000566, -0.00000913, 0.00000467, -0.00000187, 0.00000584, -0.00000367, -0.00000022, -0.00000904, -0.00000177, -0.00000746, 0.00000337, -0.00001142, 0.00000818, -0.00002942, -0.00000850, -0.00005739, -0.00012429, 0.00018696, 0.00029115, -0.00001366, -0.00014592, 0.00001565, -0.00003453, 0.00000700, -0.00001478, -0.00000035, -0.00001918, 0.00000677, -0.00000694, 0.00000581, -0.00000404, 0.00000173, -0.00000621, 0.00000417, -0.00001091, 0.00000433, -0.00000332, -0.00000089, -0.00000780, 0.00000151, 0.00000119, 0.00000019, -0.00000304, -0.00000186, -0.00000204, -0.00000294, 0.00000703, -0.00000631, 0.00000610, -0.00001348, -0.00000418, -0.00001802, -0.00000094, -0.00000057, 0.00002093, },
		{ -0.00001656, 0.00004027, 0.00000781, 0.00001393, -0.00002790, -0.00000693, 0.00000370, -0.00000759, -0.00000173, -0.00000453, 0.00000812, -0.00000966, -0.00000106, -0.00000280, 0.00000934, 0.00000121, -0.00000110, -0.00000403, 0.00000594, 0.00000187, -0.00000334, 0.00000098, -0.00000003, 0.00000490, -0.00000780, 0.00000612, -0.00000185, -0.00000018, 0.00000193, 0.00000465, -0.00000793, -0.00000257, 0.00000495, -0.00000105, -0.00000593, -0.00000294, 0.00001330, -0.00000506, -0.00000881, -0.00000901, 0.00001014, -0.00000517, -0.00000722, -0.00000317, 0.00002998, -0.00001534, 0.00000292, -0.00000170, 0.00000209, 0.00000168, -0.00001431, 0.00000744, -0.00001366, 0.00000978, -0.00001670, 0.00000084, -0.00000189, -0.00000797, 0.00002117, -0.00001160, 0.00000660, -0.00000590, 0.00000141, -0.00000374, 0.00000874, 0.00000228, -0.00000671, -0.00000037, -0.00000095, -0.00000240, 0.00000715, 0.00000192, -0.00000057, -0.00000113, -0.00000423, 0.00000509, -0.00000165, 0.00000492, -0.00000682, -0.00000106, -0.00000938, -0.00000106, -0.00000456, -0.00000158, -0.00000480, -0.00000510, -0.00003461, -0.00000794, -0.00015306, 0.00007600, 0.00033277, 0.00008979, -0.00017051, -0.00000924, -0.00002112, -0.00000585, -0.00000667, -0.00000438, -0.00001537, -0.00000095, -0.00000140, 0.00000045, 0.00000063, -0.00000075, -0.00000421, 0.00000183, -0.00000694, -0.00000116, 0.00000125, -0.00000332, -0.00000463, -0.00000328, 0.00000397, -0.00000126, -0.00000059, -0.00000413, 0.00000009, -0.00000554, 0.00000734, -0.00000436, 0.00000324, -0.00000617, -0.00000961, -0.00001087, -0.00001150, 0.00000837, 0.00000170, 0.00004079, },
		{ 0.00002063, 0.00002460, 0.00000558, -0.00000587, -0.00002811, 0.00001426, -0.00001279, 0.00000313, -0.00001083, 0.00001101, -0.00000798, -0.00000167, -0.00000491, 0.00000784, 0.00000367, -0.00000071, -0.00000397, 0.00000323, 0.00000474, -0.00000393, 0.00000068, -0.00000081, 0.00000572, -0.00000600, 0.00000167, 0.00000283, -0.00000348, 0.00000309, 0.00000335, -0.00000243, -0.00000887, 0.00000770, -0.00000229, -0.00000182, -0.00000878, 0.00001378, -0.00000086, -0.00000718, -0.00001276, 0.00000808, -0.00000083, -0.00000553, -0.00001047, 0.00002549, 0.00000155, -0.00001186, 0.00000932, -0.00000888, 0.00001233, -0.00001904, 0.00000780, -0.00001141, 0.00000458, -0.00000759, -0.00000956, 0.00000690, -0.00001603, 0.00002120, -0.00000487, 0.00000018, 0.00000009, -0.00000367, 0.00000043, 0.00000150, 0.00001024, -0.00001004, 0.00000133, -0.00000326, 0.00000014, 0.00000134, 0.00000840, -0.00000550, 0.00000499, -0.00000982, 0.00000845, -0.00000570, 0.00000907, -0.00000870, 0.00000215, -0.00001175, 0.00000089, -0.00000880, 0.00000443, -0.00001321, 0.00000921, -0.00004253, 0.00000481, -0.00011981, -0.00005193, 0.00032173, 0.00020900, -0.00015895, -0.00005605, -0.00000236, -0.00002055, 0.00000150, -0.00000958, -0.00000863, -0.00001077, 0.00000482, -0.00000528, 0.00000504, -0.00000323, -0.00000125, -0.00000203, -0.00000165, -0.00000695, 0.00000513, -0.00000529, -0.00000150, -0.00000713, 0.00000518, -0.00000169, 0.00000095, -0.00000503, 0.00000058, -0.00000583, 0.00000486, -0.00000010, -0.00000142, 0.00000109, -0.00001349, -0.00000526, -0.00001920, 0.00001301, -0.00000796, 0.00004039, -0.00000689, },
		{ 0.00003929, -0.00000175, 0.00001179, -0.00003635, 0.00001030, -0.00000703, -0.00000006, -0.00000901, 0.00000560, -0.00000059, -0.00000552, -0.00000321, 0.00000252, 0.00000803, -0.00000170, -0.00000212, -0.00000045, 0.00000717, -0.00000363, -0.00000005, -0.00000117, 0.00000520, -0.00000249, -0.00000329, 0.00000635, -0.00000484, 0.00000289, 0.00000219, 0.00000232, -0.00001207, 0.00000655, -0.00000055, -0.00000082, -0.00000949, 0.00000860, 0.00000593, -0.00000688, -0.00001263, 0.00000172, 0.00000506, -0.00000550, -0.00001032, 0.00001120, 0.00002106, -0.00002168, 0.00001291, -0.00001088, 0.00001149, -0.00001179, -0.00000177, -0.00000149, -0.00000593, 0.00000371, -0.00001864, 0.00001177, -0.00001757, 0.00001402, 0.00000565, -0.00000670, 0.00000517, -0.00000724, 0.00000323, -0.00000356, 0.00001372, -0.00000835, -0.00000059, -0.00000286, 0.00000044, -0.00000164, 0.00001050, -0.00000548, 0.00000552, -0.00000883, 0.00000549, -0.00000355, 0.00000621, -0.00000380, -0.00000114, -0.00000750, -0.00000305, -0.00000693, 0.00000274, -0.00001133, 0.00000705, -0.00002812, -0.00001344, -0.00005760, -0.00014525, 0.00023730, 0.00032855, -0.00010503, -0.00011213, 0.00000857, -0.00002827, 0.00000375, -0.00001158, -0.00000360, -0.00001747, 0.00000685, -0.00000716, 0.00000565, -0.00000292, 0.00000005, -0.00000442, 0.00000166, -0.00000979, 0.00000560, -0.00000456, -0.00000065, -0.00000797, 0.00000339, -0.00000009, 0.00000065, -0.00000389, -0.00000083, -0.00000410, 0.00000073, 0.00000447, -0.00000514, 0.00000531, -0.00001332, -0.00000363, -0.00002097, 0.00000962, -0.00000537, 0.00002546, 0.00001417, -0.00000324, },
		{ 0.00000244, 0.00001479, -0.00002578, -0.00000772, 0.00000598, -0.00000754, -0.00000270, -0.00000352, 0.00000746, -0.00000872, -0.00000086, -0.00000308, 0.00001086, -0.00000134, -0.00000095, -0.00000278, 0.00000741, -0.00000150, -0.00000167, -0.00000069, 0.00000327, 0.00000187, -0.00000723, 0.00000744, -0.00000373, 0.00000126, 0.00000168, 0.00000499, -0.00001116, 0.00000160, 0.00000350, -0.00000217, -0.00000668, 0.00000099, 0.00001179, -0.00000560, -0.00001080, -0.00000548, 0.00000912, -0.00000490, -0.00000682, -0.00000356, 0.00003129, -0.00001731, 0.00000377, -0.00000160, 0.00000017, 0.00000283, -0.00001438, 0.00000836, -0.00001388, 0.00000993, -0.00002060, 0.00000884, -0.00001120, 0.00000132, 0.00001659, -0.00001102, 0.00000725, -0.00000783, 0.00000331, -0.00000500, 0.00001206, -0.00000238, -0.00000477, -0.00000085, -0.00000092, -0.00000196, 0.00000861, -0.00000141, 0.00000188, -0.00000352, -0.00000089, 0.00000168, 0.00000001, 0.00000346, -0.00000614, -0.00000178, -0.00000786, -0.00000343, -0.00000225, -0.00000431, -0.00000319, -0.00000652, -0.00003561, -0.00001098, -0.00016678, 0.00008977, 0.00041453, -0.00000193, -0.00016432, 0.00000336, -0.00002664, -0.00000035, -0.00001009, -0.00000173, -0.00001901, 0.00000332, -0.00000449, 0.00000233, -0.00000002, -0.00000049, -0.00000458, 0.00000200, -0.00000891, 0.00000246, -0.00000125, -0.00000210, -0.00000596, -0.00000071, 0.00000272, -0.00000081, -0.00000153, -0.00000308, -0.00000169, -0.00000328, 0.00000741, -0.00000611, 0.00000545, -0.00000924, -0.00000543, -0.00001757, 0.00000009, 0.00000439, 0.00000598, 0.00003271, -0.00001687, 0.00004002, },
		{ 0.00000587, -0.00000457, -0.00002738, 0.00001444, -0.00001023, 0.00000049, -0.00000885, 0.00000972, -0.00000676, -0.00000158, -0.00000503, 0.00000893, 0.00000253, -0.00000222, -0.00000225, 0.00000467, 0.00000260, -0.00000415, 0.00000067, 0.00000027, 0.00000608, -0.00000915, 0.00000589, -0.00000089, -0.00000084, 0.00000148, 0.00000568, -0.00000717, -0.00000484, 0.00000714, -0.00000318, -0.00000371, -0.00000489, 0.00001329, -0.00000147, -0.00000946, -0.00001025, 0.00000866, -0.00000169, -0.00000441, -0.00001133, 0.00002681, 0.00000052, -0.00001264, 0.00001104, -0.00001162, 0.00001358, -0.00001946, 0.00000933, -0.00001245, 0.00000645, -0.00001321, -0.00000183, 0.00000075, -0.00001270, 0.00002390, -0.00001067, 0.00000581, -0.00000571, 0.00000123, -0.00000374, 0.00000734, 0.00000459, -0.00000789, 0.00000003, -0.00000166, -0.00000187, 0.00000559, 0.00000310, -0.00000171, 0.00000119, -0.00000541, 0.00000466, -0.00000367, 0.00000679, -0.00000665, -0.00000062, -0.00000781, -0.00000383, -0.00000418, -0.00000054, -0.00000918, 0.00000475, -0.00003865, -0.00000380, -0.00012048, -0.00007161, 0.00042156, 0.00015109, -0.00019842, -0.00002016, -0.00001965, -0.00000672, -0.00000804, -0.00000157, -0.00001676, -0.00000346, -0.00000016, -0.00000190, 0.00000273, -0.00000093, -0.00000388, 0.00000068, -0.00000599, -0.00000213, 0.00000244, -0.00000386, -0.00000323, -0.00000471, 0.00000457, -0.00000161, 0.00000027, -0.00000443, -0.00000029, -0.00000551, 0.00000744, -0.00000389, 0.00000229, -0.00000319, -0.00000841, -0.00001195, -0.00001072, 0.00001339, -0.00000733, 0.00003805, -0.00001307, 0.00002467, 0.00001765, },
		{ 0.00001200, -0.00003397, 0.00000853, -0.00000271, -0.00000328, -0.00000699, 0.00000431, 0.00000056, -0.00000554, -0.00000289, 0.00000265, 0.00000869, -0.00000475, -0.00000007, 0.00000047, 0.00000693, -0.00000606, 0.00000176, -0.00000244, 0.00000857, -0.00000810, 0.00000200, 0.00000265, -0.00000252, 0.00000122, 0.00000490, -0.00000190, -0.00001003, 0.00000790, -0.00000183, -0.00000278, -0.00000691, 0.00000984, 0.00000484, -0.00000867, -0.00001183, 0.00000396, 0.00000346, -0.00000414, -0.00001108, 0.00001192, 0.00002129, -0.00002370, 0.00001514, -0.00001305, 0.00001166, -0.00001158, -0.00000073, -0.00000199, -0.00000430, -0.00000064, -0.00001482, 0.00001195, -0.00002184, 0.00002357, -0.00000445, 0.00000118, -0.00000169, -0.00000200, -0.00000123, 0.00000202, 0.00000961, -0.00000772, -0.00000153, -0.00000081, -0.00000253, 0.00000329, 0.00000564, -0.00000261, 0.00000248, -0.00000551, 0.00000316, -0.00000257, 0.00000427, -0.00000153, -0.00000457, -0.00000296, -0.00000802, -0.00000226, -0.00000219, -0.00000704, 0.00000154, -0.00002165, -0.00002519, -0.00004949, -0.00018068, 0.00032030, 0.00033206, -0.00019271, -0.00006282, -0.00001179, -0.00001249, -0.00000690, -0.00000241, -0.00001234, -0.00001055, 0.00000268, -0.00000419, 0.00000310, 0.00000014, -0.00000347, -0.00000081, -0.00000306, -0.00000580, 0.00000426, -0.00000402, -0.00000180, -0.00000655, 0.00000388, -0.00000055, 0.00000036, -0.00000387, -0.00000075, -0.00000533, 0.00000459, 0.00000061, -0.00000224, 0.00000238, -0.00001030, -0.00000693, -0.00001850, 0.00001604, -0.00000943, 0.00002887, 0.00000374, 0.00000244, 0.00003307, -0.00000273, },
		{ -0.00002259, -0.00001013, 0.00001030, -0.00000960, -0.00000187, -0.00000397, 0.00000800, -0.00000855, -0.00000025, -0.00000372, 0.00001273, -0.00000488, 0.00000060, -0.00000205, 0.00000865, -0.00000517, 0.00000079, -0.00000300, 0.00000772, -0.00000360, -0.00000360, 0.00000601, -0.00000344, 0.00000103, 0.00000307, 0.00000312, -0.00001231, 0.00000517, 0.00000163, -0.00000352, -0.00000607, 0.00000396, 0.00001030, -0.00000664, -0.00001176, -0.00000216, 0.00000752, -0.00000389, -0.00000706, -0.00000359, 0.00003236, -0.00001948, 0.00000504, -0.00000195, -0.00000147, 0.00000417, -0.00001447, 0.00000917, -0.00001367, 0.00000849, -0.00002191, 0.00001510, -0.00002073, 0.00001352, 0.00000714, -0.00000556, 0.00000339, -0.00000570, 0.00000167, -0.00000264, 0.00001185, -0.00000479, -0.00000412, 0.00000012, -0.00000277, 0.00000107, 0.00000677, -0.00000166, 0.00000162, -0.00000318, -0.00000019, 0.00000034, -0.00000025, 0.00000426, -0.00000804, 0.00000072, -0.00001013, -0.00000180, -0.00000393, -0.00000314, -0.00000595, -0.00000241, -0.00004393, -0.00000490, -0.00019521, 0.00012800, 0.00048584, -0.00011755, -0.00012560, -0.00000439, -0.00001927, -0.00000432, -0.00000606, -0.00000564, -0.00001696, 0.00000330, -0.00000446, 0.00000171, 0.00000163, -0.00000261, -0.00000251, -0.00000061, -0.00000786, 0.00000385, -0.00000248, -0.00000188, -0.00000612, 0.00000106, 0.00000175, -0.00000070, -0.00000206, -0.00000232, -0.00000364, 0.00000031, 0.00000533, -0.00000576, 0.00000545, -0.00000978, -0.00000411, -0.00002121, 0.00001119, -0.00000231, 0.00001217, 0.00002200, -0.00001329, 0.00003692, -0.00000042, 0.00001418, },
		{ -0.00002858, 0.00001676, -0.00000956, -0.00000079, -0.00000787, 0.00000918, -0.00000597, -0.00000120, -0.00000568, 0.00001077, 0.00000000, -0.00000230, -0.00000099, 0.00000626, -0.00000063, -0.00000283, -0.00000078, 0.00000344, 0.00000335, -0.00000931, 0.00000807, -0.00000336, 0.00000099, 0.00000074, 0.00000677, -0.00001131, 0.00000009, 0.00000533, -0.00000397, -0.00000483, -0.00000097, 0.00001205, -0.00000202, -0.00001157, -0.00000689, 0.00000790, -0.00000146, -0.00000399, -0.00001189, 0.00002803, -0.00000076, -0.00001315, 0.00001268, -0.00001461, 0.00001546, -0.00002022, 0.00001103, -0.00001333, 0.00000743, -0.00001804, 0.00000704, -0.00000889, -0.00000379, 0.00002048, -0.00001168, 0.00000784, -0.00000880, 0.00000412, -0.00000599, 0.00001176, -0.00000098, -0.00000537, -0.00000095, -0.00000100, -0.00000233, 0.00000804, -0.00000106, 0.00000143, -0.00000199, -0.00000139, 0.00000075, -0.00000166, 0.00000504, -0.00000543, -0.00000211, -0.00000537, -0.00000726, -0.00000053, -0.00000475, -0.00000630, 0.00000228, -0.00003772, -0.00000892, -0.00012704, -0.00008520, 0.00053876, 0.00004778, -0.00020051, -0.00000041, -0.00002910, 0.00000192, -0.00001376, 0.00000313, -0.00002237, 0.00000256, -0.00000449, 0.00000086, 0.00000128, 0.00000019, -0.00000517, 0.00000182, -0.00000892, 0.00000230, -0.00000062, -0.00000217, -0.00000496, -0.00000192, 0.00000331, -0.00000115, -0.00000073, -0.00000330, -0.00000222, -0.00000325, 0.00000796, -0.00000627, 0.00000492, -0.00000657, -0.00000386, -0.00001895, 0.00000113, 0.00000820, -0.00000280, 0.00003142, -0.00001472, 0.00002596, 0.00001283, 0.00000500, -0.00000206, },
		{ 0.00000781, 0.00000106, -0.00000650, -0.00000476, 0.00000274, 0.00000207, -0.00000575, -0.00000281, 0.00000335, 0.00000827, -0.00000704, 0.00000213, 0.00000147, 0.00000552, -0.00000710, 0.00000239, -0.00000197, 0.00000974, -0.00001237, 0.00000720, -0.00000165, 0.00000056, -0.00000098, 0.00000797, -0.00000734, -0.00000566, 0.00000739, -0.00000273, -0.00000468, -0.00000335, 0.00000973, 0.00000422, -0.00001100, -0.00000953, 0.00000468, 0.00000272, -0.00000314, -0.00001185, 0.00001275, 0.00002123, -0.00002568, 0.00001767, -0.00001586, 0.00001267, -0.00001178, 0.00000059, -0.00000263, -0.00000274, -0.00000584, -0.00000786, 0.00000731, -0.00002064, 0.00002883, -0.00001272, 0.00000871, -0.00000905, 0.00000428, -0.00000678, 0.00000926, 0.00000283, -0.00000469, -0.00000381, 0.00000202, -0.00000600, 0.00000898, -0.00000089, 0.00000237, -0.00000264, 0.00000015, -0.00000182, 0.00000070, 0.00000055, 0.00000242, -0.00000959, 0.00000347, -0.00001550, 0.00000551, -0.00001057, 0.00000056, -0.00000712, -0.00001170, -0.00004168, -0.00003554, -0.00022841, 0.00043986, 0.00028617, -0.00025640, -0.00001272, -0.00003703, 0.00000750, -0.00002115, 0.00000971, -0.00002408, -0.00000024, -0.00000460, 0.00000114, -0.00000110, 0.00000433, -0.00000793, 0.00000362, -0.00000894, 0.00000026, 0.00000063, -0.00000175, -0.00000427, -0.00000357, 0.00000286, 0.00000002, -0.00000088, -0.00000273, -0.00000213, -0.00000475, 0.00000728, -0.00000322, 0.00000119, -0.00000153, -0.00000553, -0.00001323, -0.00001043, 0.00001630, -0.00000999, 0.00002884, -0.00000387, 0.00000743, 0.00002718, -0.00000352, 0.00001262, -0.00003272, },
		{ 0.00001582, -0.00001324, 0.00000044, -0.00000572, 0.00000967, -0.00000910, 0.00000060, -0.00000435, 0.00001436, -0.00000878, 0.00000318, -0.00000171, 0.00000927, -0.00000840, 0.00000317, -0.00000487, 0.00001187, -0.00001010, 0.00000221, 0.00000221, -0.00000104, -0.00000090, 0.00000609, -0.00000118, -0.00001058, 0.00000708, -0.00000001, -0.00000516, -0.00000404, 0.00000560, 0.00000921, -0.00000840, -0.00001123, 0.00000017, 0.00000603, -0.00000258, -0.00000776, -0.00000328, 0.00003316, -0.00002181, 0.00000682, -0.00000298, -0.00000246, 0.00000539, -0.00001446, 0.00000984, -0.00001312, 0.00000546, -0.00001983, 0.00001762, -0.00002755, 0.00002549, -0.00000483, 0.00000342, -0.00000427, 0.00000023, -0.00000353, 0.00000366, 0.00000729, -0.00000380, -0.00000569, 0.00000306, -0.00000674, 0.00000687, 0.00000126, 0.00000191, -0.00000227, 0.00000095, -0.00000336, 0.00000236, -0.00000353, 0.00000831, -0.00001346, 0.00000738, -0.00001737, 0.00000538, -0.00001147, 0.00000397, -0.00001507, 0.00000931, -0.00006219, 0.00001397, -0.00024461, 0.00020424, 0.00052511, -0.00023808, -0.00006275, -0.00003106, 0.00000146, -0.00001867, 0.00000637, -0.00001711, -0.00000793, -0.00000245, -0.00000000, -0.00000241, 0.00000628, -0.00000759, 0.00000240, -0.00000639, -0.00000309, 0.00000194, -0.00000136, -0.00000356, -0.00000427, 0.00000113, 0.00000187, -0.00000163, -0.00000144, -0.00000277, -0.00000457, 0.00000404, 0.00000191, -0.00000366, 0.00000334, -0.00000751, -0.00000656, -0.00001964, 0.00001828, -0.00000834, 0.00001785, 0.00001075, -0.00000742, 0.00003186, -0.00000146, 0.00001276, -0.00001915, -0.00001319, },
		{ -0.00001130, -0.00000007, -0.00000850, 0.00000995, -0.00000604, -0.00000030, -0.00000683, 0.00001320, -0.00000384, -0.00000073, -0.00000046, 0.00000790, -0.00000453, -0.00000037, -0.00000317, 0.00000811, -0.00000226, -0.00000600, 0.00000725, -0.00000352, 0.00000102, 0.00000183, 0.00000548, -0.00001332, 0.00000451, 0.00000298, -0.00000488, -0.00000477, 0.00000227, 0.00001063, -0.00000296, -0.00001282, -0.00000351, 0.00000633, -0.00000035, -0.00000432, -0.00001204, 0.00002905, -0.00000233, -0.00001320, 0.00001392, -0.00001745, 0.00001774, -0.00002122, 0.00001281, -0.00001394, 0.00000707, -0.00002073, 0.00001491, -0.00001972, 0.00000911, 0.00001112, -0.00000698, 0.00000480, -0.00000744, 0.00000312, -0.00000430, 0.00001242, -0.00000418, -0.00000428, -0.00000026, -0.00000251, 0.00000012, 0.00000696, -0.00000196, 0.00000170, -0.00000226, -0.00000013, -0.00000097, -0.00000175, 0.00000587, -0.00000724, 0.00000012, -0.00000724, -0.00000616, -0.00000144, -0.00000444, -0.00000850, 0.00000626, -0.00004526, -0.00000340, -0.00014964, -0.00007707, 0.00065582, -0.00009285, -0.00016085, -0.00000543, -0.00002317, -0.00000070, -0.00001064, 0.00000014, -0.00002138, 0.00000359, -0.00000513, 0.00000060, 0.00000268, -0.00000161, -0.00000349, -0.00000027, -0.00000850, 0.00000431, -0.00000227, -0.00000160, -0.00000544, 0.00000001, 0.00000239, -0.00000114, -0.00000120, -0.00000254, -0.00000428, 0.00000038, 0.00000632, -0.00000663, 0.00000555, -0.00000768, -0.00000185, -0.00002332, 0.00001288, -0.00000017, 0.00000462, 0.00002111, -0.00001181, 0.00002398, 0.00001061, 0.00000265, 0.00000177, -0.00003174, 0.00002145, },
		{ -0.00000845, -0.00000341, 0.00000183, 0.00000322, -0.00000566, -0.00000332, 0.00000496, 0.00000629, -0.00000780, 0.00000380, 0.00000277, 0.00000282, -0.00000645, 0.00000147, 0.00000074, 0.00000767, -0.00001360, 0.00001049, -0.00000488, 0.00000296, -0.00000239, 0.00000996, -0.00001244, -0.00000003, 0.00000533, -0.00000327, -0.00000610, 0.00000055, 0.00000869, 0.00000374, -0.00001331, -0.00000621, 0.00000395, 0.00000308, -0.00000282, -0.00001240, 0.00001360, 0.00002084, -0.00002742, 0.00002024, -0.00001909, 0.00001453, -0.00001249, 0.00000221, -0.00000327, -0.00000177, -0.00001073, 0.00000090, -0.00000152, -0.00001328, 0.00002762, -0.00001624, 0.00001295, -0.00001410, 0.00000895, -0.00001084, 0.00001559, -0.00000443, -0.00000079, -0.00000619, 0.00000430, -0.00000836, 0.00001343, -0.00000690, 0.00000732, -0.00000777, 0.00000611, -0.00000760, 0.00000458, -0.00000322, 0.00000609, -0.00001394, 0.00000915, -0.00002257, 0.00001331, -0.00001931, 0.00000828, -0.00001521, -0.00000294, -0.00005707, -0.00002334, -0.00027903, 0.00059075, 0.00018305, -0.00027857, 0.00002194, -0.00005585, 0.00002356, -0.00003268, 0.00001949, -0.00003422, 0.00000965, -0.00001201, 0.00000650, -0.00000495, 0.00000779, -0.00001148, 0.00000697, -0.00001395, 0.00000652, -0.00000390, 0.00000126, -0.00000718, 0.00000016, 0.00000087, 0.00000119, -0.00000262, -0.00000086, -0.00000478, -0.00000202, 0.00000781, -0.00000571, 0.00000372, -0.00000476, -0.00000100, -0.00002022, 0.00000133, 0.00001054, -0.00000588, 0.00002398, -0.00000690, 0.00000984, 0.00002309, -0.00000527, 0.00001454, -0.00003353, 0.00000928, 0.00000294, },
		{ 0.00000331, -0.00000803, 0.00001192, -0.00000998, 0.00000127, -0.00000445, 0.00001501, -0.00001209, 0.00000607, -0.00000138, 0.00000878, -0.00001041, 0.00000456, -0.00000515, 0.00001403, -0.00001558, 0.00000840, -0.00000264, 0.00000246, -0.00000365, 0.00000983, -0.00000714, -0.00000609, 0.00000682, -0.00000102, -0.00000698, -0.00000084, 0.00000585, 0.00000858, -0.00001075, -0.00000915, 0.00000103, 0.00000528, -0.00000149, -0.00000863, -0.00000271, 0.00003362, -0.00002415, 0.00000899, -0.00000477, -0.00000253, 0.00000622, -0.00001420, 0.00001040, -0.00001252, 0.00000137, -0.00001441, 0.00001528, -0.00002915, 0.00003378, -0.00001591, 0.00001313, -0.00001347, 0.00000818, -0.00001079, 0.00001262, -0.00000093, 0.00000048, -0.00000941, 0.00000762, -0.00001216, 0.00001447, -0.00000702, 0.00000865, -0.00000934, 0.00000855, -0.00001031, 0.00000775, -0.00000972, 0.00001526, -0.00002186, 0.00001752, -0.00002891, 0.00001763, -0.00002466, 0.00001684, -0.00003016, 0.00002796, -0.00008963, 0.00004509, -0.00031614, 0.00032746, 0.00051141, -0.00033962, 0.00000831, -0.00006834, 0.00003067, -0.00004004, 0.00002459, -0.00003408, 0.00000669, -0.00001319, 0.00000842, -0.00000957, 0.00001333, -0.00001473, 0.00000935, -0.00001453, 0.00000486, -0.00000322, 0.00000234, -0.00000732, -0.00000022, -0.00000085, 0.00000347, -0.00000394, 0.00000074, -0.00000509, -0.00000336, 0.00000643, -0.00000147, -0.00000100, 0.00000021, -0.00000343, -0.00001208, -0.00001247, 0.00001913, -0.00001067, 0.00002036, 0.00000163, -0.00000151, 0.00002646, -0.00000194, 0.00001160, -0.00001664, -0.00001552, 0.00002115, -0.00001726, },
		{ -0.00001072, 0.00001209, -0.00000704, 0.00000099, -0.00000817, 0.00001567, -0.00000838, 0.00000218, -0.00000053, 0.00000917, -0.00000836, 0.00000241, -0.00000549, 0.00001289, -0.00000941, 0.00000005, 0.00000361, -0.00000126, -0.00000094, 0.00000489, 0.00000140, -0.00001219, 0.00000716, 0.00000089, -0.00000614, -0.00000332, 0.00000432, 0.00000947, -0.00000457, -0.00001266, -0.00000093, 0.00000467, 0.00000125, -0.00000526, -0.00001175, 0.00002980, -0.00000411, -0.00001272, 0.00001449, -0.00001976, 0.00002010, -0.00002234, 0.00001463, -0.00001431, 0.00000523, -0.00002030, 0.00001961, -0.00002872, 0.00002296, -0.00000208, 0.00000246, -0.00000306, -0.00000137, -0.00000229, 0.00000215, 0.00000797, -0.00000333, -0.00000602, 0.00000308, -0.00000692, 0.00000618, 0.00000140, 0.00000171, -0.00000244, 0.00000206, -0.00000351, 0.00000142, -0.00000570, 0.00001090, -0.00001371, 0.00000782, -0.00001550, 0.00000204, -0.00000988, 0.00000364, -0.00001915, 0.00002040, -0.00006594, 0.00001909, -0.00019815, -0.00002952, 0.00074836, -0.00025192, -0.00008553, -0.00003652, 0.00000088, -0.00001731, 0.00000378, -0.00001287, -0.00001133, -0.00000285, 0.00000011, -0.00000447, 0.00000830, -0.00000747, 0.00000219, -0.00000661, -0.00000345, 0.00000228, -0.00000093, -0.00000350, -0.00000344, -0.00000012, 0.00000292, -0.00000256, -0.00000014, -0.00000334, -0.00000508, 0.00000405, 0.00000335, -0.00000532, 0.00000420, -0.00000613, -0.00000338, -0.00002284, 0.00002095, -0.00000826, 0.00001241, 0.00000917, -0.00000571, 0.00001939, 0.00001055, -0.00000075, 0.00000625, -0.00003606, 0.00002786, -0.00001509, 0.00000251, },
		{ 0.00000230, 0.00000343, -0.00000497, -0.00000451, 0.00000742, 0.00000269, -0.00000667, 0.00000458, 0.00000438, -0.00000091, -0.00000433, -0.00000068, 0.00000528, 0.00000233, -0.00001101, 0.00001061, -0.00000576, 0.00000350, -0.00000187, 0.00000951, -0.00001549, 0.00000529, 0.00000251, -0.00000376, -0.00000658, 0.00000404, 0.00000728, 0.00000299, -0.00001492, -0.00000264, 0.00000221, 0.00000445, -0.00000332, -0.00001253, 0.00001432, 0.00002010, -0.00002873, 0.00002250, -0.00002239, 0.00001705, -0.00001369, 0.00000410, -0.00000378, -0.00000190, -0.00001395, 0.00000941, -0.00001248, -0.00000092, 0.00001958, -0.00001347, 0.00001183, -0.00001455, 0.00000964, -0.00001092, 0.00001821, -0.00000947, 0.00000181, -0.00000694, 0.00000439, -0.00000781, 0.00001445, -0.00000984, 0.00000957, -0.00001019, 0.00000956, -0.00001144, 0.00000654, -0.00000450, 0.00000675, -0.00001459, 0.00001058, -0.00002523, 0.00001675, -0.00002380, 0.00001123, -0.00001718, -0.00000219, -0.00006265, -0.00002474, -0.00031596, 0.00075694, 0.00002649, -0.00024965, 0.00002831, -0.00005779, 0.00002745, -0.00003476, 0.00002114, -0.00003748, 0.00001443, -0.00001567, 0.00000874, -0.00000581, 0.00000816, -0.00001187, 0.00000697, -0.00001564, 0.00001047, -0.00000715, 0.00000327, -0.00000899, 0.00000319, -0.00000086, 0.00000192, -0.00000383, 0.00000066, -0.00000766, 0.00000223, 0.00000607, -0.00000623, 0.00000440, -0.00000594, 0.00000133, -0.00002505, 0.00001347, 0.00000102, 0.00000209, 0.00001454, -0.00000487, 0.00000875, 0.00002161, -0.00000860, 0.00001809, -0.00003666, 0.00001346, 0.00000218, -0.00000826, -0.00000376, },
		{ 0.00001396, -0.00001068, 0.00000137, -0.00000359, 0.00001403, -0.00001390, 0.00000849, -0.00000068, 0.00000689, -0.00001068, 0.00000442, -0.00000317, 0.00001291, -0.00001811, 0.00001293, -0.00000673, 0.00000553, -0.00000582, 0.00001276, -0.00001314, 0.00000007, 0.00000465, -0.00000143, -0.00000861, 0.00000297, 0.00000503, 0.00000815, -0.00001321, -0.00000589, 0.00000037, 0.00000560, -0.00000101, -0.00000940, -0.00000198, 0.00003367, -0.00002628, 0.00001130, -0.00000712, -0.00000163, 0.00000648, -0.00001365, 0.00001096, -0.00001240, -0.00000274, -0.00000672, 0.00000834, -0.00002437, 0.00003568, -0.00002273, 0.00002028, -0.00002117, 0.00001533, -0.00001740, 0.00002157, -0.00001058, 0.00000651, -0.00001404, 0.00001244, -0.00001736, 0.00002188, -0.00001587, 0.00001646, -0.00001756, 0.00001767, -0.00001927, 0.00001496, -0.00001713, 0.00002318, -0.00003101, 0.00002853, -0.00004189, 0.00003201, -0.00004057, 0.00003250, -0.00004773, 0.00004919, -0.00012097, 0.00008198, -0.00040298, 0.00049749, 0.00043035, -0.00039909, 0.00006811, -0.00010332, 0.00005928, -0.00006148, 0.00004290, -0.00005161, 0.00002285, -0.00002579, 0.00001835, -0.00001767, 0.00002082, -0.00002207, 0.00001632, -0.00002293, 0.00001410, -0.00001026, 0.00000765, -0.00001235, 0.00000533, -0.00000440, 0.00000617, -0.00000723, 0.00000411, -0.00000915, 0.00000043, 0.00000637, -0.00000342, 0.00000079, -0.00000233, 0.00000060, -0.00001859, -0.00000135, 0.00001364, -0.00000779, 0.00001783, -0.00000316, 0.00000224, 0.00002240, -0.00000316, 0.00001176, -0.00001615, -0.00001577, 0.00002481, -0.00002027, 0.00000555, -0.00000990, },
		{ -0.00000864, 0.00000231, -0.00000914, 0.00001738, -0.00001264, 0.00000570, -0.00000080, 0.00000954, -0.00001126, 0.00000452, -0.00000651, 0.00001600, -0.00001600, 0.00000705, -0.00000158, 0.00000250, -0.00000409, 0.00000911, -0.00000481, -0.00000789, 0.00000737, -0.00000040, -0.00000776, -0.00000058, 0.00000500, 0.00000872, -0.00000679, -0.00001095, 0.00000027, 0.00000356, 0.00000281, -0.00000658, -0.00001102, 0.00003014, -0.00000592, -0.00001184, 0.00001432, -0.00002124, 0.00002225, -0.00002344, 0.00001653, -0.00001472, 0.00000232, -0.00001656, 0.00001972, -0.00003312, 0.00003408, -0.00001568, 0.00001400, -0.00001375, 0.00000793, -0.00001095, 0.00001246, -0.00000130, 0.00000186, -0.00001091, 0.00000910, -0.00001396, 0.00001534, -0.00000825, 0.00000985, -0.00001109, 0.00001123, -0.00001205, 0.00000858, -0.00001398, 0.00002043, -0.00002498, 0.00002105, -0.00003035, 0.00001782, -0.00002675, 0.00002059, -0.00003931, 0.00004575, -0.00010118, 0.00006102, -0.00027815, 0.00007225, 0.00079007, -0.00040290, 0.00000963, -0.00008662, 0.00003952, -0.00004577, 0.00002799, -0.00003480, 0.00000728, -0.00001682, 0.00001150, -0.00001451, 0.00001812, -0.00001718, 0.00001154, -0.00001687, 0.00000622, -0.00000427, 0.00000407, -0.00000848, 0.00000164, -0.00000305, 0.00000558, -0.00000599, 0.00000310, -0.00000659, -0.00000326, 0.00000606, 0.00000058, -0.00000356, 0.00000194, -0.00000284, -0.00000796, -0.00001691, 0.00002294, -0.00001278, 0.00001760, -0.00000158, 0.00000132, 0.00001370, 0.00001149, -0.00000418, 0.00001024, -0.00004017, 0.00003466, -0.00001983, 0.00000614, -0.00001392, 0.00001514, },
		{ -0.00000368, -0.00000611, 0.00001023, -0.00000200, -0.00000387, 0.00000449, 0.00000594, -0.00000499, -0.00000148, -0.00000314, 0.00001038, -0.00000510, -0.00000513, 0.00000750, -0.00000394, 0.00000179, 0.00000094, 0.00000598, -0.00001522, 0.00000882, -0.00000009, -0.00000459, -0.00000575, 0.00000653, 0.00000598, 0.00000163, -0.00001523, 0.00000032, 0.00000016, 0.00000648, -0.00000458, -0.00001212, 0.00001478, 0.00001913, -0.00002956, 0.00002421, -0.00002536, 0.00001993, -0.00001527, 0.00000627, -0.00000420, -0.00000333, -0.00001442, 0.00001544, -0.00002261, 0.00001359, 0.00000639, -0.00000481, 0.00000498, -0.00000953, 0.00000522, -0.00000558, 0.00001517, -0.00001002, 0.00000117, -0.00000454, 0.00000098, -0.00000302, 0.00001040, -0.00000768, 0.00000688, -0.00000749, 0.00000787, -0.00001069, 0.00000412, -0.00000089, 0.00000192, -0.00000883, 0.00000462, -0.00001975, 0.00001156, -0.00001936, 0.00000445, -0.00000743, -0.00001638, -0.00004924, -0.00005312, -0.00031815, 0.00091290, -0.00016684, -0.00017151, 0.00000104, -0.00003688, 0.00001391, -0.00002285, 0.00001055, -0.00002986, 0.00001025, -0.00001220, 0.00000506, -0.00000140, 0.00000345, -0.00000729, 0.00000184, -0.00001195, 0.00000981, -0.00000696, 0.00000240, -0.00000799, 0.00000389, -0.00000088, 0.00000094, -0.00000321, 0.00000038, -0.00000912, 0.00000648, 0.00000300, -0.00000516, 0.00000326, -0.00000466, 0.00000034, -0.00002545, 0.00002243, -0.00000879, 0.00001145, 0.00000234, 0.00000122, 0.00000443, 0.00002262, -0.00001330, 0.00002285, -0.00004153, 0.00001992, -0.00000112, -0.00000582, -0.00000603, 0.00000439, 0.00000250, },
		{ 0.00000072, -0.00000171, 0.00001123, -0.00001370, 0.00000993, 0.00000051, 0.00000370, -0.00000926, 0.00000285, 0.00000089, 0.00000823, -0.00001660, 0.00001427, -0.00000853, 0.00000676, -0.00000605, 0.00001331, -0.00001728, 0.00000622, 0.00000148, -0.00000159, -0.00000955, 0.00000665, 0.00000367, 0.00000753, -0.00001508, -0.00000221, -0.00000144, 0.00000699, -0.00000135, -0.00000981, -0.00000125, 0.00003330, -0.00002801, 0.00001339, -0.00000970, 0.00000009, 0.00000616, -0.00001277, 0.00001162, -0.00001325, -0.00000553, 0.00000133, -0.00000148, -0.00001387, 0.00003021, -0.00002308, 0.00002221, -0.00002454, 0.00001885, -0.00002044, 0.00002735, -0.00001862, 0.00001184, -0.00001755, 0.00001554, -0.00002017, 0.00002650, -0.00002241, 0.00002235, -0.00002389, 0.00002518, -0.00002725, 0.00002116, -0.00002293, 0.00002896, -0.00003743, 0.00003644, -0.00005178, 0.00004356, -0.00005401, 0.00004545, -0.00006157, 0.00006544, -0.00014671, 0.00011230, -0.00048947, 0.00070208, 0.00027933, -0.00040040, 0.00009899, -0.00012238, 0.00007685, -0.00007460, 0.00005412, -0.00006330, 0.00003491, -0.00003559, 0.00002596, -0.00002346, 0.00002579, -0.00002679, 0.00002052, -0.00002866, 0.00002173, -0.00001672, 0.00001257, -0.00001688, 0.00001074, -0.00000815, 0.00000881, -0.00001033, 0.00000749, -0.00001383, 0.00000615, 0.00000374, -0.00000327, 0.00000074, -0.00000289, 0.00000262, -0.00002334, 0.00001060, 0.00000383, -0.00000024, 0.00001004, -0.00000261, 0.00000244, 0.00002083, -0.00000600, 0.00001381, -0.00001819, -0.00001316, 0.00002577, -0.00002119, 0.00000615, -0.00001046, 0.00001512, -0.00001078, },
		{ -0.00000924, 0.00001756, -0.00001561, 0.00000895, -0.00000079, 0.00000854, -0.00001254, 0.00000525, -0.00000537, 0.00001596, -0.00001989, 0.00001279, -0.00000646, 0.00000616, -0.00000696, 0.00001288, -0.00001153, -0.00000145, 0.00000534, -0.00000083, -0.00000947, 0.00000300, 0.00000453, 0.00000819, -0.00000919, -0.00000798, -0.00000003, 0.00000341, 0.00000392, -0.00000799, -0.00000993, 0.00002996, -0.00000749, -0.00001083, 0.00001359, -0.00002183, 0.00002401, -0.00002443, 0.00001864, -0.00001567, -0.00000071, -0.00001035, 0.00001513, -0.00003130, 0.00003934, -0.00002595, 0.00002415, -0.00002415, 0.00001764, -0.00002021, 0.00002410, -0.00001335, 0.00001008, -0.00001791, 0.00001661, -0.00002213, 0.00002579, -0.00002007, 0.00002068, -0.00002256, 0.00002367, -0.00002441, 0.00001936, -0.00002536, 0.00003292, -0.00003923, 0.00003768, -0.00004946, 0.00003888, -0.00004987, 0.00004429, -0.00006652, 0.00007924, -0.00014741, 0.00011852, -0.00038722, 0.00023472, 0.00075924, -0.00051736, 0.00010272, -0.00014211, 0.00008359, -0.00007925, 0.00005646, -0.00006097, 0.00003068, -0.00003550, 0.00002687, -0.00002768, 0.00003035, -0.00002895, 0.00002274, -0.00002917, 0.00001886, -0.00001421, 0.00001205, -0.00001597, 0.00000933, -0.00000850, 0.00001019, -0.00001120, 0.00000833, -0.00001234, 0.00000182, 0.00000515, -0.00000049, -0.00000284, 0.00000038, 0.00000040, -0.00001366, -0.00000693, 0.00001847, -0.00001191, 0.00001792, -0.00000856, 0.00000685, 0.00000873, 0.00001204, -0.00000650, 0.00001255, -0.00004262, 0.00004028, -0.00002407, 0.00000959, -0.00001710, 0.00001834, -0.00001031, 0.00000322, },
		{ 0.00001261, -0.00000681, -0.00000010, 0.00000393, 0.00000687, -0.00000851, 0.00000111, -0.00000468, 0.00001424, -0.00001253, 0.00000235, 0.00000226, -0.00000020, -0.00000151, 0.00000536, -0.00000016, -0.00001142, 0.00000971, -0.00000177, -0.00000592, -0.00000356, 0.00000773, 0.00000500, -0.00000033, -0.00001402, 0.00000206, -0.00000155, 0.00000870, -0.00000645, -0.00001110, 0.00001475, 0.00001816, -0.00003006, 0.00002532, -0.00002774, 0.00002290, -0.00001713, 0.00000879, -0.00000484, -0.00000573, -0.00001176, 0.00001736, -0.00002898, 0.00002656, -0.00000866, 0.00000743, -0.00000607, -0.00000003, -0.00000367, 0.00000484, 0.00000614, -0.00000517, -0.00000362, 0.00000160, -0.00000624, 0.00000617, 0.00000096, 0.00000022, -0.00000161, 0.00000142, -0.00000032, -0.00000382, -0.00000408, 0.00000882, -0.00000954, 0.00000451, -0.00001017, -0.00000424, -0.00000472, -0.00000312, -0.00001514, 0.00001744, -0.00004988, -0.00001060, -0.00011909, -0.00026502, 0.00102805, -0.00036962, -0.00005791, -0.00005572, 0.00000595, -0.00001714, 0.00000344, -0.00001285, -0.00001035, -0.00000428, -0.00000017, -0.00000575, 0.00000921, -0.00000700, 0.00000277, -0.00000892, -0.00000212, 0.00000332, -0.00000195, -0.00000259, -0.00000300, 0.00000103, 0.00000193, -0.00000275, 0.00000025, -0.00000297, -0.00000745, 0.00000877, 0.00000031, -0.00000379, 0.00000138, -0.00000177, -0.00000355, -0.00002068, 0.00002565, -0.00001548, 0.00001908, -0.00000979, 0.00000925, -0.00000181, 0.00002517, -0.00001854, 0.00002779, -0.00004692, 0.00002750, -0.00000605, -0.00000177, -0.00000976, 0.00000781, 0.00000063, -0.00000211, -0.00000760, },
		{ 0.00000701, -0.00001155, 0.00001029, 0.00000192, -0.00000022, -0.00000671, 0.00000060, 0.00000594, 0.00000094, -0.00001131, 0.00001203, -0.00000743, 0.00000554, -0.00000375, 0.00001057, -0.00001806, 0.00001067, -0.00000160, -0.00000202, -0.00000930, 0.00000953, 0.00000228, 0.00000637, -0.00001578, 0.00000105, -0.00000376, 0.00000918, -0.00000256, -0.00000966, -0.00000073, 0.00003266, -0.00002931, 0.00001503, -0.00001213, 0.00000238, 0.00000529, -0.00001149, 0.00001231, -0.00001531, -0.00000585, 0.00000752, -0.00001139, -0.00000016, 0.00001852, -0.00001670, 0.00001785, -0.00002205, 0.00001694, -0.00001781, 0.00002734, -0.00002217, 0.00001395, -0.00001787, 0.00001502, -0.00001862, 0.00002601, -0.00002390, 0.00002340, -0.00002521, 0.00002773, -0.00003084, 0.00002316, -0.00002392, 0.00002928, -0.00003746, 0.00003700, -0.00005368, 0.00004673, -0.00005890, 0.00004917, -0.00006432, 0.00006770, -0.00015527, 0.00012007, -0.00055287, 0.00091697, 0.00007023, -0.00033918, 0.00009022, -0.00011553, 0.00007506, -0.00007237, 0.00005209, -0.00006336, 0.00003752, -0.00003790, 0.00002732, -0.00002362, 0.00002532, -0.00002617, 0.00001929, -0.00002884, 0.00002472, -0.00001975, 0.00001467, -0.00001867, 0.00001392, -0.00001028, 0.00000974, -0.00001160, 0.00000909, -0.00001720, 0.00001208, -0.00000043, -0.00000144, -0.00000113, -0.00000106, 0.00000151, -0.00002405, 0.00001988, -0.00000693, 0.00000972, -0.00000154, 0.00000273, -0.00000105, 0.00002202, -0.00001058, 0.00001761, -0.00002250, -0.00000781, 0.00002384, -0.00001960, 0.00000463, -0.00000924, 0.00001502, -0.00001017, -0.00000051, 0.00000082, },
		{ -0.00001659, 0.00001127, -0.00000024, 0.00000612, -0.00001202, 0.00000446, -0.00000201, 0.00001220, -0.00001970, 0.00001548, -0.00000925, 0.00000814, -0.00000800, 0.00001446, -0.00001670, 0.00000535, 0.00000198, -0.00000080, -0.00001071, 0.00000670, 0.00000339, 0.00000751, -0.00001111, -0.00000445, -0.00000154, 0.00000432, 0.00000430, -0.00000922, -0.00000862, 0.00002922, -0.00000861, -0.00001006, 0.00001263, -0.00002165, 0.00002531, -0.00002528, 0.00002105, -0.00001769, -0.00000254, -0.00000343, 0.00000724, -0.00002347, 0.00003717, -0.00003010, 0.00002974, -0.00003104, 0.00002459, -0.00002685, 0.00003365, -0.00002495, 0.00001875, -0.00002488, 0.00002348, -0.00002908, 0.00003479, -0.00003101, 0.00003108, -0.00003372, 0.00003621, -0.00003755, 0.00003094, -0.00003692, 0.00004521, -0.00005288, 0.00005360, -0.00006819, 0.00006021, -0.00007403, 0.00006929, -0.00009465, 0.00011349, -0.00019561, 0.00018046, -0.00051282, 0.00045181, 0.00064508, -0.00057190, 0.00017130, -0.00018659, 0.00012088, -0.00010808, 0.00008108, -0.00008421, 0.00005266, -0.00005376, 0.00004200, -0.00004036, 0.00004172, -0.00003964, 0.00003268, -0.00004032, 0.00003139, -0.00002496, 0.00002092, -0.00002416, 0.00001800, -0.00001513, 0.00001557, -0.00001701, 0.00001434, -0.00001949, 0.00000959, 0.00000105, 0.00000093, -0.00000422, 0.00000096, 0.00000162, -0.00001781, 0.00000421, 0.00000929, -0.00000579, 0.00001264, -0.00001027, 0.00000912, 0.00000593, 0.00001110, -0.00000692, 0.00001240, -0.00004238, 0.00004345, -0.00002653, 0.00001170, -0.00001924, 0.00002089, -0.00001157, 0.00000347, -0.00000830, 0.00001590, },
		{ 0.00000374, 0.00000345, 0.00000662, -0.00001067, 0.00000259, -0.00000430, 0.00001522, -0.00001767, 0.00000912, -0.00000318, 0.00000387, -0.00000491, 0.00000979, -0.00000734, -0.00000500, 0.00000803, -0.00000237, -0.00000755, -0.00000036, 0.00000774, 0.00000424, -0.00000250, -0.00001153, 0.00000238, -0.00000249, 0.00001070, -0.00000869, -0.00000946, 0.00001404, 0.00001754, -0.00003057, 0.00002602, -0.00002948, 0.00002583, -0.00001926, 0.00001187, -0.00000623, -0.00000817, -0.00000665, 0.00001476, -0.00002965, 0.00003456, -0.00002166, 0.00001977, -0.00001832, 0.00001137, -0.00001469, 0.00001818, -0.00000725, 0.00000420, -0.00001197, 0.00001071, -0.00001621, 0.00001842, -0.00001250, 0.00001268, -0.00001488, 0.00001565, -0.00001444, 0.00000880, -0.00001757, 0.00002390, -0.00002665, 0.00002423, -0.00003253, 0.00002021, -0.00003131, 0.00002438, -0.00004693, 0.00005660, -0.00010196, 0.00005336, -0.00022601, -0.00014255, 0.00107349, -0.00054931, 0.00006841, -0.00012917, 0.00006273, -0.00006003, 0.00003968, -0.00004539, 0.00001821, -0.00002722, 0.00001906, -0.00002254, 0.00002484, -0.00002195, 0.00001702, -0.00002397, 0.00001278, -0.00000846, 0.00000772, -0.00001165, 0.00000600, -0.00000559, 0.00000779, -0.00000928, 0.00000673, -0.00000980, -0.00000167, 0.00000749, -0.00000026, -0.00000380, 0.00000048, 0.00000092, -0.00000854, -0.00001191, 0.00002253, -0.00001691, 0.00002235, -0.00001904, 0.00001671, -0.00000814, 0.00002789, -0.00002318, 0.00003171, -0.00005134, 0.00003463, -0.00001118, 0.00000274, -0.00001402, 0.00001186, -0.00000171, -0.00000065, -0.00000845, 0.00001373, -0.00001066, },
		{ 0.00000991, 0.00000301, -0.00000400, -0.00000399, -0.00000114, 0.00001024, -0.00000696, -0.00000376, 0.00000712, -0.00000400, 0.00000235, 0.00000061, 0.00000479, -0.00001500, 0.00001234, -0.00000370, -0.00000302, -0.00000768, 0.00001121, 0.00000112, 0.00000463, -0.00001501, 0.00000327, -0.00000596, 0.00001180, -0.00000459, -0.00000876, -0.00000072, 0.00003206, -0.00003037, 0.00001622, -0.00001423, 0.00000505, 0.00000387, -0.00000964, 0.00001269, -0.00001828, -0.00000323, 0.00001011, -0.00001841, 0.00001318, 0.00000359, -0.00000541, 0.00000813, -0.00001404, 0.00000956, -0.00000913, 0.00002048, -0.00001959, 0.00001122, -0.00001369, 0.00000985, -0.00001176, 0.00001923, -0.00001881, 0.00001778, -0.00001944, 0.00002293, -0.00002751, 0.00001853, -0.00001782, 0.00002179, -0.00002859, 0.00002734, -0.00004406, 0.00003727, -0.00005039, 0.00003834, -0.00004994, 0.00004848, -0.00013650, 0.00008991, -0.00056797, 0.00110948, -0.00017154, -0.00022470, 0.00004124, -0.00007979, 0.00005063, -0.00005174, 0.00003392, -0.00004879, 0.00002750, -0.00002973, 0.00001985, -0.00001601, 0.00001762, -0.00001864, 0.00001109, -0.00002172, 0.00002091, -0.00001716, 0.00001191, -0.00001585, 0.00001303, -0.00000910, 0.00000749, -0.00000954, 0.00000720, -0.00001714, 0.00001592, -0.00000420, 0.00000062, -0.00000364, 0.00000231, -0.00000243, -0.00001987, 0.00002389, -0.00001516, 0.00001894, -0.00001425, 0.00001103, -0.00000725, 0.00002532, -0.00001631, 0.00002237, -0.00002808, -0.00000063, 0.00001966, -0.00001590, 0.00000121, -0.00000633, 0.00001368, -0.00000903, -0.00000189, 0.00000324, 0.00000225, -0.00000805, },
		{ 0.00000072, 0.00000266, -0.00001008, 0.00000273, 0.00000268, 0.00000540, -0.00001534, 0.00001438, -0.00000905, 0.00000757, -0.00000636, 0.00001267, -0.00001858, 0.00001063, -0.00000145, -0.00000092, -0.00001093, 0.00000980, 0.00000210, 0.00000633, -0.00001193, -0.00000119, -0.00000371, 0.00000609, 0.00000380, -0.00000998, -0.00000735, 0.00002808, -0.00000924, -0.00000976, 0.00001179, -0.00002089, 0.00002613, -0.00002585, 0.00002368, -0.00002105, -0.00000199, 0.00000205, -0.00000139, -0.00001170, 0.00002815, -0.00002718, 0.00002902, -0.00003222, 0.00002638, -0.00002818, 0.00003792, -0.00003272, 0.00002487, -0.00002927, 0.00002732, -0.00003230, 0.00003943, -0.00003780, 0.00003759, -0.00004089, 0.00004494, -0.00004755, 0.00003956, -0.00004492, 0.00005330, -0.00006154, 0.00006378, -0.00008079, 0.00007533, -0.00009220, 0.00008804, -0.00011521, 0.00013827, -0.00023293, 0.00022995, -0.00063297, 0.00070344, 0.00045204, -0.00055424, 0.00019844, -0.00020576, 0.00013989, -0.00012278, 0.00009364, -0.00009704, 0.00006640, -0.00006570, 0.00005193, -0.00004831, 0.00004844, -0.00004572, 0.00003790, -0.00004672, 0.00004012, -0.00003314, 0.00002775, -0.00003040, 0.00002520, -0.00002080, 0.00001983, -0.00002151, 0.00001910, -0.00002593, 0.00001822, -0.00000516, 0.00000444, -0.00000769, 0.00000412, -0.00000033, -0.00001815, 0.00001309, -0.00000139, 0.00000358, 0.00000279, -0.00000674, 0.00000748, 0.00000595, 0.00000821, -0.00000526, 0.00000975, -0.00003922, 0.00004366, -0.00002650, 0.00001173, -0.00001966, 0.00002226, -0.00001218, 0.00000309, -0.00000656, 0.00001260, -0.00001547, 0.00001242, },
		{ 0.00000493, -0.00001106, 0.00000261, -0.00000170, 0.00001246, -0.00001885, 0.00001312, -0.00000688, 0.00000654, -0.00000673, 0.00001234, -0.00001339, 0.00000222, 0.00000469, -0.00000228, -0.00000896, 0.00000318, 0.00000700, 0.00000332, -0.00000425, -0.00000842, 0.00000150, -0.00000249, 0.00001218, -0.00001103, -0.00000728, 0.00001254, 0.00001756, -0.00003152, 0.00002671, -0.00003074, 0.00002866, -0.00002163, 0.00001565, -0.00000899, -0.00000935, -0.00000070, 0.00000874, -0.00002440, 0.00003553, -0.00002937, 0.00002870, -0.00002833, 0.00002137, -0.00002452, 0.00003100, -0.00002184, 0.00001563, -0.00002186, 0.00002080, -0.00002668, 0.00003111, -0.00002712, 0.00002685, -0.00003005, 0.00003234, -0.00003179, 0.00002473, -0.00003384, 0.00004154, -0.00004625, 0.00004667, -0.00005836, 0.00004924, -0.00006378, 0.00005864, -0.00008592, 0.00010413, -0.00016563, 0.00013465, -0.00036664, 0.00005124, 0.00102934, -0.00067546, 0.00018123, -0.00020126, 0.00012054, -0.00010477, 0.00007760, -0.00007988, 0.00004970, -0.00005360, 0.00004144, -0.00004186, 0.00004235, -0.00003838, 0.00003245, -0.00004024, 0.00002984, -0.00002314, 0.00002020, -0.00002318, 0.00001762, -0.00001485, 0.00001574, -0.00001763, 0.00001519, -0.00001921, 0.00000790, 0.00000212, 0.00000233, -0.00000641, 0.00000210, 0.00000138, -0.00001201, -0.00000187, 0.00001455, -0.00001286, 0.00002008, -0.00002347, 0.00002153, -0.00001292, 0.00002952, -0.00002625, 0.00003365, -0.00005362, 0.00003991, -0.00001512, 0.00000643, -0.00001776, 0.00001572, -0.00000400, 0.00000039, -0.00000838, 0.00001313, -0.00001273, 0.00000684, 0.00000343, },
		{ -0.00000674, -0.00000203, -0.00000127, 0.00001201, -0.00001309, 0.00000372, 0.00000141, 0.00000025, -0.00000139, 0.00000550, -0.00000257, -0.00000886, 0.00001115, -0.00000453, -0.00000451, -0.00000492, 0.00001172, 0.00000012, 0.00000266, -0.00001299, 0.00000418, -0.00000761, 0.00001447, -0.00000729, -0.00000697, -0.00000157, 0.00003195, -0.00003166, 0.00001726, -0.00001602, 0.00000805, 0.00000183, -0.00000691, 0.00001211, -0.00002122, 0.00000179, 0.00000847, -0.00002041, 0.00002258, -0.00001069, 0.00000754, -0.00000433, -0.00000264, -0.00000144, 0.00000400, 0.00000776, -0.00001110, 0.00000358, -0.00000514, 0.00000041, -0.00000018, 0.00000672, -0.00000743, 0.00000561, -0.00000650, 0.00001034, -0.00001654, 0.00000664, -0.00000419, 0.00000623, -0.00001066, 0.00000715, -0.00002227, 0.00001399, -0.00002676, 0.00001101, -0.00001626, 0.00000475, -0.00008548, 0.00001211, -0.00051344, 0.00124496, -0.00041161, -0.00007808, -0.00003799, -0.00002035, 0.00000680, -0.00001500, 0.00000138, -0.00002065, 0.00000519, -0.00001102, 0.00000343, -0.00000065, 0.00000288, -0.00000450, -0.00000369, -0.00000742, 0.00000995, -0.00000825, 0.00000357, -0.00000766, 0.00000720, -0.00000379, 0.00000137, -0.00000346, 0.00000089, -0.00001216, 0.00001561, -0.00000545, 0.00000099, -0.00000491, 0.00000532, -0.00000740, -0.00001182, 0.00002181, -0.00001851, 0.00002460, -0.00002518, 0.00001987, -0.00001451, 0.00002952, -0.00002217, 0.00002695, -0.00003355, 0.00000694, 0.00001450, -0.00001107, -0.00000336, -0.00000226, 0.00001145, -0.00000768, -0.00000291, 0.00000474, -0.00000196, -0.00000412, 0.00000935, 0.00000316, },
		{ -0.00000759, 0.00000112, 0.00000708, -0.00000260, -0.00000811, 0.00001012, -0.00000616, 0.00000467, -0.00000227, 0.00000747, -0.00001639, 0.00001310, -0.00000390, -0.00000164, -0.00000979, 0.00001181, 0.00000096, 0.00000458, -0.00001133, 0.00000115, -0.00000594, 0.00000845, 0.00000238, -0.00000998, -0.00000653, 0.00002693, -0.00000965, -0.00000988, 0.00001120, -0.00001967, 0.00002638, -0.00002588, 0.00002613, -0.00002548, 0.00000151, 0.00000427, -0.00000785, 0.00000063, 0.00001482, -0.00001838, 0.00002221, -0.00002732, 0.00002221, -0.00002305, 0.00003507, -0.00003425, 0.00002609, -0.00002908, 0.00002634, -0.00003001, 0.00003767, -0.00003798, 0.00003746, -0.00004103, 0.00004647, -0.00005084, 0.00004178, -0.00004599, 0.00005367, -0.00006140, 0.00006388, -0.00008214, 0.00007825, -0.00009761, 0.00009311, -0.00011977, 0.00014328, -0.00024596, 0.00024808, -0.00071999, 0.00095784, 0.00020052, -0.00046664, 0.00017727, -0.00019170, 0.00013352, -0.00011714, 0.00008859, -0.00009413, 0.00006671, -0.00006657, 0.00005251, -0.00004801, 0.00004744, -0.00004440, 0.00003573, -0.00004551, 0.00004193, -0.00003565, 0.00002972, -0.00003207, 0.00002841, -0.00002326, 0.00002094, -0.00002270, 0.00002042, -0.00002910, 0.00002504, -0.00001130, 0.00000844, -0.00001200, 0.00000898, -0.00000521, -0.00001375, 0.00001708, -0.00001014, 0.00001316, -0.00000927, 0.00000064, 0.00000249, 0.00000853, 0.00000365, -0.00000202, 0.00000529, -0.00003375, 0.00004124, -0.00002405, 0.00000960, -0.00001821, 0.00002227, -0.00001214, 0.00000242, -0.00000467, 0.00000844, -0.00001273, 0.00001262, 0.00000162, -0.00000103, },
		{ 0.00000154, 0.00000248, 0.00000637, -0.00001562, 0.00001328, -0.00000764, 0.00000665, -0.00000582, 0.00001156, -0.00001633, 0.00000820, 0.00000104, -0.00000220, -0.00000950, 0.00000630, 0.00000602, 0.00000190, -0.00000493, -0.00000550, -0.00000007, -0.00000168, 0.00001294, -0.00001318, -0.00000482, 0.00001036, 0.00001830, -0.00003317, 0.00002773, -0.00003172, 0.00003138, -0.00002413, 0.00002013, -0.00001348, -0.00000798, 0.00000397, 0.00000164, -0.00001492, 0.00002951, -0.00003022, 0.00003187, -0.00003342, 0.00002707, -0.00003005, 0.00003973, -0.00003393, 0.00002587, -0.00003049, 0.00002918, -0.00003483, 0.00004104, -0.00003935, 0.00003897, -0.00004320, 0.00004736, -0.00004829, 0.00004006, -0.00004897, 0.00005753, -0.00006365, 0.00006652, -0.00008161, 0.00007611, -0.00009489, 0.00009196, -0.00012354, 0.00014982, -0.00022839, 0.00021760, -0.00052248, 0.00030331, 0.00089055, -0.00072694, 0.00025748, -0.00025385, 0.00016529, -0.00013996, 0.00010750, -0.00010762, 0.00007632, -0.00007671, 0.00006131, -0.00005886, 0.00005739, -0.00005220, 0.00004510, -0.00005369, 0.00004502, -0.00003707, 0.00003231, -0.00003435, 0.00002927, -0.00002451, 0.00002377, -0.00002581, 0.00002352, -0.00002904, 0.00001946, -0.00000636, 0.00000779, -0.00001174, 0.00000681, -0.00000166, -0.00001167, 0.00000613, 0.00000472, -0.00000503, 0.00001283, -0.00002260, 0.00002262, -0.00001514, 0.00002933, -0.00002732, 0.00003326, -0.00005320, 0.00004252, -0.00001692, 0.00000838, -0.00002013, 0.00001871, -0.00000589, 0.00000095, -0.00000755, 0.00001086, -0.00001272, 0.00000882, 0.00000388, 0.00000203, -0.00000988, },
		{ 0.00000069, 0.00001019, -0.00001553, 0.00000887, -0.00000301, 0.00000345, -0.00000389, 0.00000892, -0.00000932, -0.00000143, 0.00000796, -0.00000444, -0.00000601, -0.00000163, 0.00001142, -0.00000104, 0.00000110, -0.00001032, 0.00000399, -0.00000854, 0.00001692, -0.00001048, -0.00000427, -0.00000351, 0.00003277, -0.00003370, 0.00001861, -0.00001775, 0.00001144, -0.00000093, -0.00000298, 0.00000982, -0.00002275, 0.00000767, 0.00000348, -0.00001685, 0.00002568, -0.00002082, 0.00001853, -0.00001611, 0.00000897, -0.00001295, 0.00001842, -0.00000803, 0.00000128, -0.00000736, 0.00000619, -0.00001149, 0.00001398, -0.00000924, 0.00000799, -0.00001092, 0.00001150, -0.00000818, 0.00000049, -0.00001088, 0.00001506, -0.00001519, 0.00001374, -0.00002067, 0.00000870, -0.00002030, 0.00000932, -0.00003001, 0.00003348, -0.00005992, -0.00000546, -0.00011299, -0.00037827, 0.00129475, -0.00061399, 0.00007267, -0.00012972, 0.00005074, -0.00004746, 0.00003058, -0.00003930, 0.00001585, -0.00002530, 0.00001502, -0.00001924, 0.00001996, -0.00001647, 0.00001380, -0.00002255, 0.00001175, -0.00000639, 0.00000571, -0.00000932, 0.00000506, -0.00000298, 0.00000515, -0.00000805, 0.00000602, -0.00000938, -0.00000214, 0.00001019, -0.00000277, -0.00000187, -0.00000316, 0.00000578, -0.00001078, -0.00000253, 0.00001494, -0.00001647, 0.00002513, -0.00003209, 0.00002699, -0.00002109, 0.00003333, -0.00002716, 0.00003036, -0.00003766, 0.00001347, 0.00000975, -0.00000635, -0.00000806, 0.00000219, 0.00000881, -0.00000643, -0.00000326, 0.00000476, -0.00000469, -0.00000061, 0.00000906, 0.00000199, -0.00000789, -0.00000134, },
		{ 0.00000940, -0.00000942, -0.00000026, 0.00000446, -0.00000199, 0.00000077, 0.00000288, 0.00000013, -0.00001071, 0.00001244, -0.00000493, -0.00000303, -0.00000741, 0.00001266, -0.00000005, 0.00000257, -0.00000948, 0.00000231, -0.00000781, 0.00001110, 0.00000005, -0.00000892, -0.00000663, 0.00002636, -0.00001041, -0.00001002, 0.00001069, -0.00001793, 0.00002586, -0.00002490, 0.00002763, -0.00003001, 0.00000749, 0.00000248, -0.00000989, 0.00000997, 0.00000088, -0.00000659, 0.00001140, -0.00001784, 0.00001325, -0.00001231, 0.00002528, -0.00002893, 0.00002155, -0.00002359, 0.00002007, -0.00002190, 0.00002905, -0.00003080, 0.00002969, -0.00003288, 0.00003912, -0.00004546, 0.00003568, -0.00003833, 0.00004460, -0.00005067, 0.00005182, -0.00006958, 0.00006554, -0.00008604, 0.00007972, -0.00010286, 0.00012156, -0.00022493, 0.00021927, -0.00074673, 0.00117726, -0.00007653, -0.00032571, 0.00011266, -0.00014524, 0.00010131, -0.00009034, 0.00006500, -0.00007416, 0.00005183, -0.00005443, 0.00004197, -0.00003797, 0.00003754, -0.00003469, 0.00002530, -0.00003562, 0.00003529, -0.00003074, 0.00002512, -0.00002749, 0.00002596, -0.00002093, 0.00001753, -0.00001923, 0.00001674, -0.00002690, 0.00002748, -0.00001484, 0.00001069, -0.00001502, 0.00001352, -0.00001120, -0.00000554, 0.00001527, -0.00001448, 0.00002006, -0.00002069, 0.00000963, -0.00000446, 0.00001272, -0.00000180, 0.00000189, 0.00000015, -0.00002714, 0.00003717, -0.00001993, 0.00000583, -0.00001519, 0.00002108, -0.00001168, 0.00000189, -0.00000342, 0.00000447, -0.00000921, 0.00001238, 0.00000183, -0.00000404, -0.00000547, 0.00000077, },
		{ -0.00000144, -0.00000900, 0.00000992, -0.00000546, 0.00000419, -0.00000215, 0.00000710, -0.00001510, 0.00001138, -0.00000168, -0.00000272, -0.00000873, 0.00000843, 0.00000516, -0.00000015, -0.00000417, -0.00000343, -0.00000177, -0.00000027, 0.00001284, -0.00001474, -0.00000255, 0.00000795, 0.00001946, -0.00003545, 0.00002918, -0.00003249, 0.00003386, -0.00002645, 0.00002488, -0.00001949, -0.00000339, 0.00000548, -0.00000370, -0.00000432, 0.00001858, -0.00002478, 0.00002883, -0.00003247, 0.00002700, -0.00002941, 0.00004182, -0.00004045, 0.00003195, -0.00003525, 0.00003343, -0.00003821, 0.00004542, -0.00004598, 0.00004555, -0.00005053, 0.00005655, -0.00005959, 0.00005057, -0.00005876, 0.00006747, -0.00007407, 0.00007834, -0.00009597, 0.00009355, -0.00011649, 0.00011545, -0.00014976, 0.00018164, -0.00027501, 0.00028191, -0.00066629, 0.00058617, 0.00066942, -0.00069707, 0.00028299, -0.00027374, 0.00018590, -0.00015624, 0.00012122, -0.00012107, 0.00009099, -0.00009016, 0.00007315, -0.00006875, 0.00006573, -0.00005953, 0.00005125, -0.00006047, 0.00005428, -0.00004636, 0.00004056, -0.00004193, 0.00003791, -0.00003183, 0.00002942, -0.00003140, 0.00002914, -0.00003641, 0.00003010, -0.00001563, 0.00001446, -0.00001852, 0.00001378, -0.00000803, -0.00000648, 0.00000944, -0.00000361, 0.00000375, 0.00000260, -0.00001735, 0.00002007, -0.00001467, 0.00002724, -0.00002654, 0.00003090, -0.00005036, 0.00004247, -0.00001634, 0.00000822, -0.00002070, 0.00002048, -0.00000725, 0.00000124, -0.00000647, 0.00000757, -0.00001095, 0.00000977, 0.00000443, -0.00000138, -0.00000783, 0.00000029, 0.00000676, },
		{ -0.00001350, 0.00001028, -0.00000468, 0.00000424, -0.00000374, 0.00000916, -0.00001327, 0.00000511, 0.00000419, -0.00000419, -0.00000682, 0.00000141, 0.00001084, -0.00000277, 0.00000064, -0.00000784, 0.00000316, -0.00000883, 0.00001895, -0.00001388, -0.00000083, -0.00000652, 0.00003470, -0.00003686, 0.00002073, -0.00001973, 0.00001530, -0.00000442, 0.00000227, 0.00000526, -0.00002146, 0.00001227, -0.00000268, -0.00000911, 0.00002205, -0.00002475, 0.00002480, -0.00002418, 0.00001763, -0.00002166, 0.00003045, -0.00002315, 0.00001430, -0.00001886, 0.00001768, -0.00002305, 0.00002756, -0.00002515, 0.00002385, -0.00002804, 0.00003064, -0.00002881, 0.00001999, -0.00003044, 0.00003605, -0.00003814, 0.00003968, -0.00005052, 0.00004272, -0.00005907, 0.00005109, -0.00007727, 0.00009056, -0.00013517, 0.00009132, -0.00027280, -0.00016643, 0.00124332, -0.00074911, 0.00019955, -0.00021332, 0.00011810, -0.00009996, 0.00007480, -0.00007912, 0.00005273, -0.00005721, 0.00004271, -0.00004329, 0.00004147, -0.00003633, 0.00003227, -0.00004146, 0.00003180, -0.00002458, 0.00002171, -0.00002410, 0.00001990, -0.00001546, 0.00001589, -0.00001891, 0.00001696, -0.00002164, 0.00001139, 0.00000032, 0.00000387, -0.00000836, 0.00000246, 0.00000220, -0.00001012, 0.00000475, 0.00000610, -0.00001044, 0.00002067, -0.00003406, 0.00003096, -0.00002574, 0.00003580, -0.00003061, 0.00003202, -0.00003968, 0.00001798, 0.00000652, -0.00000280, -0.00001194, 0.00000631, 0.00000615, -0.00000537, -0.00000299, 0.00000324, -0.00000551, 0.00000195, 0.00000919, -0.00000043, -0.00000742, -0.00000179, 0.00000434, 0.00000488, },
		{ 0.00000578, -0.00000046, 0.00000159, -0.00000229, 0.00000706, -0.00000716, -0.00000323, 0.00000943, -0.00000481, -0.00000465, -0.00000431, 0.00001267, -0.00000126, 0.00000096, -0.00000698, 0.00000247, -0.00000920, 0.00001385, -0.00000315, -0.00000662, -0.00000808, 0.00002697, -0.00001220, -0.00000953, 0.00000987, -0.00001541, 0.00002426, -0.00002238, 0.00002722, -0.00003313, 0.00001445, -0.00000260, -0.00000678, 0.00001377, -0.00001008, 0.00000469, -0.00000023, -0.00000663, 0.00000219, 0.00000142, 0.00001073, -0.00001812, 0.00001214, -0.00001373, 0.00000963, -0.00000933, 0.00001502, -0.00001756, 0.00001549, -0.00001749, 0.00002361, -0.00003177, 0.00002162, -0.00002254, 0.00002689, -0.00003040, 0.00002874, -0.00004411, 0.00003777, -0.00005758, 0.00004773, -0.00006425, 0.00007235, -0.00016749, 0.00013647, -0.00069362, 0.00132604, -0.00034033, -0.00015837, 0.00001970, -0.00007562, 0.00004974, -0.00004742, 0.00002705, -0.00004048, 0.00002412, -0.00003094, 0.00002161, -0.00001944, 0.00002003, -0.00001797, 0.00000806, -0.00001832, 0.00002098, -0.00001872, 0.00001408, -0.00001668, 0.00001764, -0.00001360, 0.00000951, -0.00001109, 0.00000787, -0.00001856, 0.00002395, -0.00001382, 0.00000921, -0.00001463, 0.00001530, -0.00001559, 0.00000391, 0.00000878, -0.00001369, 0.00002246, -0.00002908, 0.00001799, -0.00001173, 0.00001735, -0.00000721, 0.00000552, -0.00000447, -0.00002075, 0.00003276, -0.00001523, 0.00000129, -0.00001123, 0.00001906, -0.00001100, 0.00000180, -0.00000337, 0.00000163, -0.00000579, 0.00001222, 0.00000088, -0.00000568, -0.00000436, 0.00000230, 0.00000836, -0.00001293, },
		{ 0.00000457, -0.00000153, 0.00000027, 0.00000310, -0.00000001, -0.00001000, 0.00001125, -0.00000284, -0.00000406, -0.00000660, 0.00000941, 0.00000443, -0.00000252, -0.00000210, -0.00000250, -0.00000324, 0.00000156, 0.00001175, -0.00001528, -0.00000107, 0.00000600, 0.00002037, -0.00003786, 0.00003081, -0.00003287, 0.00003576, -0.00002798, 0.00002905, -0.00002606, 0.00000405, 0.00000295, -0.00000501, 0.00000397, 0.00000615, -0.00001543, 0.00002103, -0.00002633, 0.00002158, -0.00002265, 0.00003660, -0.00003992, 0.00003216, -0.00003460, 0.00003222, -0.00003556, 0.00004279, -0.00004520, 0.00004450, -0.00004962, 0.00005703, -0.00006249, 0.00005308, -0.00006011, 0.00006820, -0.00007415, 0.00007830, -0.00009686, 0.00009602, -0.00012208, 0.00012176, -0.00015619, 0.00018924, -0.00029180, 0.00030768, -0.00076771, 0.00086250, 0.00039393, -0.00059509, 0.00025573, -0.00025621, 0.00017752, -0.00014917, 0.00011470, -0.00011620, 0.00008953, -0.00008993, 0.00007331, -0.00006841, 0.00006468, -0.00005799, 0.00004867, -0.00005822, 0.00005491, -0.00004815, 0.00004220, -0.00004330, 0.00004098, -0.00003450, 0.00003064, -0.00003238, 0.00002982, -0.00003860, 0.00003671, -0.00002274, 0.00001980, -0.00002442, 0.00002085, -0.00001594, 0.00000271, 0.00000705, -0.00000789, 0.00001061, -0.00000797, -0.00000962, 0.00001490, -0.00001208, 0.00002373, -0.00002454, 0.00002741, -0.00004595, 0.00004044, -0.00001381, 0.00000617, -0.00001955, 0.00002097, -0.00000814, 0.00000159, -0.00000584, 0.00000420, -0.00000817, 0.00001012, 0.00000449, -0.00000442, -0.00000581, -0.00000006, 0.00000939, -0.00000867, -0.00000112, },
		{ -0.00000327, 0.00000228, -0.00000061, 0.00000558, -0.00001306, 0.00000898, 0.00000124, -0.00000452, -0.00000635, 0.00000355, 0.00001038, -0.00000522, 0.00000169, -0.00000624, 0.00000225, -0.00000866, 0.00002038, -0.00001709, 0.00000288, -0.00001021, 0.00003750, -0.00004113, 0.00002376, -0.00002207, 0.00001952, -0.00000835, 0.00000852, -0.00000151, -0.00001646, 0.00001360, -0.00000725, -0.00000002, 0.00001336, -0.00002250, 0.00002528, -0.00002689, 0.00002134, -0.00002515, 0.00003702, -0.00003408, 0.00002462, -0.00002794, 0.00002652, -0.00003140, 0.00003732, -0.00003740, 0.00003625, -0.00004157, 0.00004639, -0.00004688, 0.00003743, -0.00004750, 0.00005401, -0.00005739, 0.00006124, -0.00007561, 0.00007202, -0.00009377, 0.00008926, -0.00012044, 0.00014273, -0.00020593, 0.00018564, -0.00044362, 0.00010204, 0.00109267, -0.00080042, 0.00028119, -0.00027088, 0.00016738, -0.00013888, 0.00010754, -0.00010891, 0.00008158, -0.00008305, 0.00006557, -0.00006314, 0.00005890, -0.00005210, 0.00004651, -0.00005595, 0.00004815, -0.00004026, 0.00003584, -0.00003719, 0.00003356, -0.00002731, 0.00002580, -0.00002862, 0.00002658, -0.00003289, 0.00002548, -0.00001170, 0.00001289, -0.00001732, 0.00001124, -0.00000544, -0.00000425, 0.00000729, -0.00000141, -0.00000303, 0.00001284, -0.00003155, 0.00003145, -0.00002797, 0.00003656, -0.00003241, 0.00003200, -0.00003956, 0.00002015, 0.00000531, -0.00000102, -0.00001441, 0.00000957, 0.00000371, -0.00000439, -0.00000253, 0.00000067, -0.00000458, 0.00000351, 0.00000948, -0.00000357, -0.00000588, -0.00000262, 0.00000842, -0.00000258, -0.00000770, 0.00000793, },
		{ 0.00000299, -0.00000292, 0.00000835, -0.00001206, 0.00000379, 0.00000553, -0.00000433, -0.00000577, -0.00000130, 0.00001237, -0.00000309, 0.00000047, -0.00000469, 0.00000211, -0.00001021, 0.00001656, -0.00000702, -0.00000309, -0.00001104, 0.00002914, -0.00001561, -0.00000778, 0.00000823, -0.00001186, 0.00002138, -0.00001795, 0.00002408, -0.00003320, 0.00002016, -0.00000892, 0.00000033, 0.00001130, -0.00001570, 0.00001248, -0.00000958, 0.00000317, -0.00000773, 0.00001460, -0.00000511, -0.00000471, 0.00000026, -0.00000174, -0.00000257, 0.00000494, -0.00000144, -0.00000131, -0.00000203, 0.00000194, 0.00000297, -0.00001249, 0.00000231, -0.00000157, 0.00000389, -0.00000443, -0.00000099, -0.00001044, -0.00000022, -0.00001706, 0.00000226, -0.00000977, 0.00000230, -0.00008069, 0.00000473, -0.00055466, 0.00137871, -0.00055466, 0.00000473, -0.00008069, 0.00000230, -0.00000977, 0.00000226, -0.00001706, -0.00000022, -0.00001044, -0.00000099, -0.00000443, 0.00000389, -0.00000157, 0.00000231, -0.00001249, 0.00000297, 0.00000194, -0.00000203, -0.00000131, -0.00000144, 0.00000494, -0.00000257, -0.00000174, 0.00000026, -0.00000471, -0.00000511, 0.00001460, -0.00000773, 0.00000317, -0.00000958, 0.00001248, -0.00001570, 0.00001130, 0.00000033, -0.00000892, 0.00002016, -0.00003320, 0.00002408, -0.00001795, 0.00002138, -0.00001186, 0.00000823, -0.00000778, -0.00001561, 0.00002914, -0.00001104, -0.00000309, -0.00000702, 0.00001656, -0.00001021, 0.00000211, -0.00000469, 0.00000047, -0.00000309, 0.00001237, -0.00000130, -0.00000577, -0.00000433, 0.00000553, 0.00000379, -0.00001206, 0.00000835, -0.00000292, },
		{ -0.00000327, 0.00000793, -0.00000770, -0.00000258, 0.00000842, -0.00000262, -0.00000588, -0.00000357, 0.00000948, 0.00000351, -0.00000458, 0.00000067, -0.00000253, -0.00000439, 0.00000371, 0.00000957, -0.00001441, -0.00000102, 0.00000531, 0.00002015, -0.00003956, 0.00003200, -0.00003241, 0.00003656, -0.00002797, 0.00003145, -0.00003155, 0.00001284, -0.00000303, -0.00000141, 0.00000729, -0.00000425, -0.00000544, 0.00001124, -0.00001732, 0.00001289, -0.00001170, 0.00002548, -0.00003289, 0.00002658, -0.00002862, 0.00002580, -0.00002731, 0.00003356, -0.00003719, 0.00003584, -0.00004026, 0.00004815, -0.00005595, 0.00004651, -0.00005210, 0.00005890, -0.00006314, 0.00006557, -0.00008305, 0.00008158, -0.00010891, 0.00010754, -0.00013888, 0.00016738, -0.00027088, 0.00028119, -0.00080042, 0.00109267, 0.00010204, -0.00044362, 0.00018564, -0.00020593, 0.00014273, -0.00012044, 0.00008926, -0.00009377, 0.00007202, -0.00007561, 0.00006124, -0.00005739, 0.00005401, -0.00004750, 0.00003743, -0.00004688, 0.00004639, -0.00004157, 0.00003625, -0.00003740, 0.00003732, -0.00003140, 0.00002652, -0.00002794, 0.00002462, -0.00003408, 0.00003702, -0.00002515, 0.00002134, -0.00002689, 0.00002528, -0.00002250, 0.00001336, -0.00000002, -0.00000725, 0.00001360, -0.00001646, -0.00000151, 0.00000852, -0.00000835, 0.00001952, -0.00002207, 0.00002376, -0.00004113, 0.00003750, -0.00001021, 0.00000288, -0.00001709, 0.00002038, -0.00000866, 0.00000225, -0.00000624, 0.00000169, -0.00000522, 0.00001038, 0.00000355, -0.00000635, -0.00000452, 0.00000124, 0.00000898, -0.00001306, 0.00000558, -0.00000061, 0.00000228, },
		{ 0.00000457, -0.00000112, -0.00000867, 0.00000939, -0.00000006, -0.00000581, -0.00000442, 0.00000449, 0.00001012, -0.00000817, 0.00000420, -0.00000584, 0.00000159, -0.00000814, 0.00002097, -0.00001955, 0.00000617, -0.00001381, 0.00004044, -0.00004595, 0.00002741, -0.00002454, 0.00002373, -0.00001208, 0.00001490, -0.00000962, -0.00000797, 0.00001061, -0.00000789, 0.00000705, 0.00000271, -0.00001594, 0.00002085, -0.00002442, 0.00001980, -0.00002274, 0.00003671, -0.00003860, 0.00002982, -0.00003238, 0.00003064, -0.00003450, 0.00004098, -0.00004330, 0.00004220, -0.00004815, 0.00005491, -0.00005822, 0.00004867, -0.00005799, 0.00006468, -0.00006841, 0.00007331, -0.00008993, 0.00008953, -0.00011620, 0.00011470, -0.00014917, 0.00017752, -0.00025621, 0.00025573, -0.00059509, 0.00039393, 0.00086250, -0.00076771, 0.00030768, -0.00029180, 0.00018924, -0.00015619, 0.00012176, -0.00012208, 0.00009602, -0.00009686, 0.00007830, -0.00007415, 0.00006820, -0.00006011, 0.00005308, -0.00006249, 0.00005703, -0.00004962, 0.00004450, -0.00004520, 0.00004279, -0.00003556, 0.00003222, -0.00003460, 0.00003216, -0.00003992, 0.00003660, -0.00002265, 0.00002158, -0.00002633, 0.00002103, -0.00001543, 0.00000615, 0.00000397, -0.00000501, 0.00000295, 0.00000405, -0.00002606, 0.00002905, -0.00002798, 0.00003576, -0.00003287, 0.00003081, -0.00003786, 0.00002037, 0.00000600, -0.00000107, -0.00001528, 0.00001175, 0.00000156, -0.00000324, -0.00000250, -0.00000210, -0.00000252, 0.00000443, 0.00000941, -0.00000660, -0.00000406, -0.00000284, 0.00001125, -0.00001000, -0.00000001, 0.00000310, 0.00000027, -0.00000153, },
		{ 0.00000578, -0.00001293, 0.00000836, 0.00000230, -0.00000436, -0.00000568, 0.00000088, 0.00001222, -0.00000579, 0.00000163, -0.00000337, 0.00000180, -0.00001100, 0.00001906, -0.00001123, 0.00000129, -0.00001523, 0.00003276, -0.00002075, -0.00000447, 0.00000552, -0.00000721, 0.00001735, -0.00001173, 0.00001799, -0.00002908, 0.00002246, -0.00001369, 0.00000878, 0.00000391, -0.00001559, 0.00001530, -0.00001463, 0.00000921, -0.00001382, 0.00002395, -0.00001856, 0.00000787, -0.00001109, 0.00000951, -0.00001360, 0.00001764, -0.00001668, 0.00001408, -0.00001872, 0.00002098, -0.00001832, 0.00000806, -0.00001797, 0.00002003, -0.00001944, 0.00002161, -0.00003094, 0.00002412, -0.00004048, 0.00002705, -0.00004742, 0.00004974, -0.00007562, 0.00001970, -0.00015837, -0.00034033, 0.00132604, -0.00069362, 0.00013647, -0.00016749, 0.00007235, -0.00006425, 0.00004773, -0.00005758, 0.00003777, -0.00004411, 0.00002874, -0.00003040, 0.00002689, -0.00002254, 0.00002162, -0.00003177, 0.00002361, -0.00001749, 0.00001549, -0.00001756, 0.00001502, -0.00000933, 0.00000963, -0.00001373, 0.00001214, -0.00001812, 0.00001073, 0.00000142, 0.00000219, -0.00000663, -0.00000023, 0.00000469, -0.00001008, 0.00001377, -0.00000678, -0.00000260, 0.00001445, -0.00003313, 0.00002722, -0.00002238, 0.00002426, -0.00001541, 0.00000987, -0.00000953, -0.00001220, 0.00002697, -0.00000808, -0.00000662, -0.00000315, 0.00001385, -0.00000920, 0.00000247, -0.00000698, 0.00000096, -0.00000126, 0.00001267, -0.00000431, -0.00000465, -0.00000481, 0.00000943, -0.00000323, -0.00000716, 0.00000706, -0.00000229, 0.00000159, -0.00000046, },
		{ -0.00001350, 0.00000488, 0.00000434, -0.00000179, -0.00000742, -0.00000043, 0.00000919, 0.00000195, -0.00000551, 0.00000324, -0.00000299, -0.00000537, 0.00000615, 0.00000631, -0.00001194, -0.00000280, 0.00000652, 0.00001798, -0.00003968, 0.00003202, -0.00003061, 0.00003580, -0.00002574, 0.00003096, -0.00003406, 0.00002067, -0.00001044, 0.00000610, 0.00000475, -0.00001012, 0.00000220, 0.00000246, -0.00000836, 0.00000387, 0.00000032, 0.00001139, -0.00002164, 0.00001696, -0.00001891, 0.00001589, -0.00001546, 0.00001990, -0.00002410, 0.00002171, -0.00002458, 0.00003180, -0.00004146, 0.00003227, -0.00003633, 0.00004147, -0.00004329, 0.00004271, -0.00005721, 0.00005273, -0.00007912, 0.00007480, -0.00009996, 0.00011810, -0.00021332, 0.00019955, -0.00074911, 0.00124332, -0.00016643, -0.00027280, 0.00009132, -0.00013517, 0.00009056, -0.00007727, 0.00005109, -0.00005907, 0.00004272, -0.00005052, 0.00003968, -0.00003814, 0.00003605, -0.00003044, 0.00001999, -0.00002881, 0.00003064, -0.00002804, 0.00002385, -0.00002515, 0.00002756, -0.00002305, 0.00001768, -0.00001886, 0.00001430, -0.00002315, 0.00003045, -0.00002166, 0.00001763, -0.00002418, 0.00002480, -0.00002475, 0.00002205, -0.00000911, -0.00000268, 0.00001227, -0.00002146, 0.00000526, 0.00000227, -0.00000442, 0.00001530, -0.00001973, 0.00002073, -0.00003686, 0.00003470, -0.00000652, -0.00000083, -0.00001388, 0.00001895, -0.00000883, 0.00000316, -0.00000784, 0.00000064, -0.00000277, 0.00001084, 0.00000141, -0.00000682, -0.00000419, 0.00000419, 0.00000511, -0.00001327, 0.00000916, -0.00000374, 0.00000424, -0.00000468, 0.00001028, },
		{ -0.00000144, 0.00000676, 0.00000029, -0.00000783, -0.00000138, 0.00000443, 0.00000977, -0.00001095, 0.00000757, -0.00000647, 0.00000124, -0.00000725, 0.00002048, -0.00002070, 0.00000822, -0.00001634, 0.00004247, -0.00005036, 0.00003090, -0.00002654, 0.00002724, -0.00001467, 0.00002007, -0.00001735, 0.00000260, 0.00000375, -0.00000361, 0.00000944, -0.00000648, -0.00000803, 0.00001378, -0.00001852, 0.00001446, -0.00001563, 0.00003010, -0.00003641, 0.00002914, -0.00003140, 0.00002942, -0.00003183, 0.00003791, -0.00004193, 0.00004056, -0.00004636, 0.00005428, -0.00006047, 0.00005125, -0.00005953, 0.00006573, -0.00006875, 0.00007315, -0.00009016, 0.00009099, -0.00012107, 0.00012122, -0.00015624, 0.00018590, -0.00027374, 0.00028299, -0.00069707, 0.00066942, 0.00058617, -0.00066629, 0.00028191, -0.00027501, 0.00018164, -0.00014976, 0.00011545, -0.00011649, 0.00009355, -0.00009597, 0.00007834, -0.00007407, 0.00006747, -0.00005876, 0.00005057, -0.00005959, 0.00005655, -0.00005053, 0.00004555, -0.00004598, 0.00004542, -0.00003821, 0.00003343, -0.00003525, 0.00003195, -0.00004045, 0.00004182, -0.00002941, 0.00002700, -0.00003247, 0.00002883, -0.00002478, 0.00001858, -0.00000432, -0.00000370, 0.00000548, -0.00000339, -0.00001949, 0.00002488, -0.00002645, 0.00003386, -0.00003249, 0.00002918, -0.00003545, 0.00001946, 0.00000795, -0.00000255, -0.00001474, 0.00001284, -0.00000027, -0.00000177, -0.00000343, -0.00000417, -0.00000015, 0.00000516, 0.00000843, -0.00000873, -0.00000272, -0.00000168, 0.00001138, -0.00001510, 0.00000710, -0.00000215, 0.00000419, -0.00000546, 0.00000992, -0.00000900, },
		{ 0.00000940, 0.00000077, -0.00000547, -0.00000404, 0.00000183, 0.00001238, -0.00000921, 0.00000447, -0.00000342, 0.00000189, -0.00001168, 0.00002108, -0.00001519, 0.00000583, -0.00001993, 0.00003717, -0.00002714, 0.00000015, 0.00000189, -0.00000180, 0.00001272, -0.00000446, 0.00000963, -0.00002069, 0.00002006, -0.00001448, 0.00001527, -0.00000554, -0.00001120, 0.00001352, -0.00001502, 0.00001069, -0.00001484, 0.00002748, -0.00002690, 0.00001674, -0.00001923, 0.00001753, -0.00002093, 0.00002596, -0.00002749, 0.00002512, -0.00003074, 0.00003529, -0.00003562, 0.00002530, -0.00003469, 0.00003754, -0.00003797, 0.00004197, -0.00005443, 0.00005183, -0.00007416, 0.00006500, -0.00009034, 0.00010131, -0.00014524, 0.00011266, -0.00032571, -0.00007653, 0.00117726, -0.00074673, 0.00021927, -0.00022493, 0.00012156, -0.00010286, 0.00007972, -0.00008604, 0.00006554, -0.00006958, 0.00005182, -0.00005067, 0.00004460, -0.00003833, 0.00003568, -0.00004546, 0.00003912, -0.00003288, 0.00002969, -0.00003080, 0.00002905, -0.00002190, 0.00002007, -0.00002359, 0.00002155, -0.00002893, 0.00002528, -0.00001231, 0.00001325, -0.00001784, 0.00001140, -0.00000659, 0.00000088, 0.00000997, -0.00000989, 0.00000248, 0.00000749, -0.00003001, 0.00002763, -0.00002490, 0.00002586, -0.00001793, 0.00001069, -0.00001002, -0.00001041, 0.00002636, -0.00000663, -0.00000892, 0.00000005, 0.00001110, -0.00000781, 0.00000231, -0.00000948, 0.00000257, -0.00000005, 0.00001266, -0.00000741, -0.00000303, -0.00000493, 0.00001244, -0.00001071, 0.00000013, 0.00000288, 0.00000077, -0.00000199, 0.00000446, -0.00000026, -0.00000942, },
		{ 0.00000069, -0.00000134, -0.00000789, 0.00000199, 0.00000906, -0.00000061, -0.00000469, 0.00000476, -0.00000326, -0.00000643, 0.00000881, 0.00000219, -0.00000806, -0.00000635, 0.00000975, 0.00001347, -0.00003766, 0.00003036, -0.00002716, 0.00003333, -0.00002109, 0.00002699, -0.00003209, 0.00002513, -0.00001647, 0.00001494, -0.00000253, -0.00001078, 0.00000578, -0.00000316, -0.00000187, -0.00000277, 0.00001019, -0.00000214, -0.00000938, 0.00000602, -0.00000805, 0.00000515, -0.00000298, 0.00000506, -0.00000932, 0.00000571, -0.00000639, 0.00001175, -0.00002255, 0.00001380, -0.00001647, 0.00001996, -0.00001924, 0.00001502, -0.00002530, 0.00001585, -0.00003930, 0.00003058, -0.00004746, 0.00005074, -0.00012972, 0.00007267, -0.00061399, 0.00129475, -0.00037827, -0.00011299, -0.00000546, -0.00005992, 0.00003348, -0.00003001, 0.00000932, -0.00002030, 0.00000870, -0.00002067, 0.00001374, -0.00001519, 0.00001506, -0.00001088, 0.00000049, -0.00000818, 0.00001150, -0.00001092, 0.00000799, -0.00000924, 0.00001398, -0.00001149, 0.00000619, -0.00000736, 0.00000128, -0.00000803, 0.00001842, -0.00001295, 0.00000897, -0.00001611, 0.00001853, -0.00002082, 0.00002568, -0.00001685, 0.00000348, 0.00000767, -0.00002275, 0.00000982, -0.00000298, -0.00000093, 0.00001144, -0.00001775, 0.00001861, -0.00003370, 0.00003277, -0.00000351, -0.00000427, -0.00001048, 0.00001692, -0.00000854, 0.00000399, -0.00001032, 0.00000110, -0.00000104, 0.00001142, -0.00000163, -0.00000601, -0.00000444, 0.00000796, -0.00000143, -0.00000932, 0.00000892, -0.00000389, 0.00000345, -0.00000301, 0.00000887, -0.00001553, 0.00001019, },
		{ 0.00000154, -0.00000988, 0.00000203, 0.00000388, 0.00000882, -0.00001272, 0.00001086, -0.00000755, 0.00000095, -0.00000589, 0.00001871, -0.00002013, 0.00000838, -0.00001692, 0.00004252, -0.00005320, 0.00003326, -0.00002732, 0.00002933, -0.00001514, 0.00002262, -0.00002260, 0.00001283, -0.00000503, 0.00000472, 0.00000613, -0.00001167, -0.00000166, 0.00000681, -0.00001174, 0.00000779, -0.00000636, 0.00001946, -0.00002904, 0.00002352, -0.00002581, 0.00002377, -0.00002451, 0.00002927, -0.00003435, 0.00003231, -0.00003707, 0.00004502, -0.00005369, 0.00004510, -0.00005220, 0.00005739, -0.00005886, 0.00006131, -0.00007671, 0.00007632, -0.00010762, 0.00010750, -0.00013996, 0.00016529, -0.00025385, 0.00025748, -0.00072694, 0.00089055, 0.00030331, -0.00052248, 0.00021760, -0.00022839, 0.00014982, -0.00012354, 0.00009196, -0.00009489, 0.00007611, -0.00008161, 0.00006652, -0.00006365, 0.00005753, -0.00004897, 0.00004006, -0.00004829, 0.00004736, -0.00004320, 0.00003897, -0.00003935, 0.00004104, -0.00003483, 0.00002918, -0.00003049, 0.00002587, -0.00003393, 0.00003973, -0.00003005, 0.00002707, -0.00003342, 0.00003187, -0.00003022, 0.00002951, -0.00001492, 0.00000164, 0.00000397, -0.00000798, -0.00001348, 0.00002013, -0.00002413, 0.00003138, -0.00003172, 0.00002773, -0.00003317, 0.00001830, 0.00001036, -0.00000482, -0.00001318, 0.00001294, -0.00000168, -0.00000007, -0.00000550, -0.00000493, 0.00000190, 0.00000602, 0.00000630, -0.00000950, -0.00000220, 0.00000104, 0.00000820, -0.00001633, 0.00001156, -0.00000582, 0.00000665, -0.00000764, 0.00001328, -0.00001562, 0.00000637, 0.00000248, },
		{ -0.00000759, -0.00000103, 0.00000162, 0.00001262, -0.00001273, 0.00000844, -0.00000467, 0.00000242, -0.00001214, 0.00002227, -0.00001821, 0.00000960, -0.00002405, 0.00004124, -0.00003375, 0.00000529, -0.00000202, 0.00000365, 0.00000853, 0.00000249, 0.00000064, -0.00000927, 0.00001316, -0.00001014, 0.00001708, -0.00001375, -0.00000521, 0.00000898, -0.00001200, 0.00000844, -0.00001130, 0.00002504, -0.00002910, 0.00002042, -0.00002270, 0.00002094, -0.00002326, 0.00002841, -0.00003207, 0.00002972, -0.00003565, 0.00004193, -0.00004551, 0.00003573, -0.00004440, 0.00004744, -0.00004801, 0.00005251, -0.00006657, 0.00006671, -0.00009413, 0.00008859, -0.00011714, 0.00013352, -0.00019170, 0.00017727, -0.00046664, 0.00020052, 0.00095784, -0.00071999, 0.00024808, -0.00024596, 0.00014328, -0.00011977, 0.00009311, -0.00009761, 0.00007825, -0.00008214, 0.00006388, -0.00006140, 0.00005367, -0.00004599, 0.00004178, -0.00005084, 0.00004647, -0.00004103, 0.00003746, -0.00003798, 0.00003767, -0.00003001, 0.00002634, -0.00002908, 0.00002609, -0.00003425, 0.00003507, -0.00002305, 0.00002221, -0.00002732, 0.00002221, -0.00001838, 0.00001482, 0.00000063, -0.00000785, 0.00000427, 0.00000151, -0.00002548, 0.00002613, -0.00002588, 0.00002638, -0.00001967, 0.00001120, -0.00000988, -0.00000965, 0.00002693, -0.00000653, -0.00000998, 0.00000238, 0.00000845, -0.00000594, 0.00000115, -0.00001133, 0.00000458, 0.00000096, 0.00001181, -0.00000979, -0.00000164, -0.00000390, 0.00001310, -0.00001639, 0.00000747, -0.00000227, 0.00000467, -0.00000616, 0.00001012, -0.00000811, -0.00000260, 0.00000708, 0.00000112, },
		{ -0.00000674, 0.00000316, 0.00000935, -0.00000412, -0.00000196, 0.00000474, -0.00000291, -0.00000768, 0.00001145, -0.00000226, -0.00000336, -0.00001107, 0.00001450, 0.00000694, -0.00003355, 0.00002695, -0.00002217, 0.00002952, -0.00001451, 0.00001987, -0.00002518, 0.00002460, -0.00001851, 0.00002181, -0.00001182, -0.00000740, 0.00000532, -0.00000491, 0.00000099, -0.00000545, 0.00001561, -0.00001216, 0.00000089, -0.00000346, 0.00000137, -0.00000379, 0.00000720, -0.00000766, 0.00000357, -0.00000825, 0.00000995, -0.00000742, -0.00000369, -0.00000450, 0.00000288, -0.00000065, 0.00000343, -0.00001102, 0.00000519, -0.00002065, 0.00000138, -0.00001500, 0.00000680, -0.00002035, -0.00003799, -0.00007808, -0.00041161, 0.00124496, -0.00051344, 0.00001211, -0.00008548, 0.00000475, -0.00001626, 0.00001101, -0.00002676, 0.00001399, -0.00002227, 0.00000715, -0.00001066, 0.00000623, -0.00000419, 0.00000664, -0.00001654, 0.00001034, -0.00000650, 0.00000561, -0.00000743, 0.00000672, -0.00000018, 0.00000041, -0.00000514, 0.00000358, -0.00001110, 0.00000776, 0.00000400, -0.00000144, -0.00000264, -0.00000433, 0.00000754, -0.00001069, 0.00002258, -0.00002041, 0.00000847, 0.00000179, -0.00002122, 0.00001211, -0.00000691, 0.00000183, 0.00000805, -0.00001602, 0.00001726, -0.00003166, 0.00003195, -0.00000157, -0.00000697, -0.00000729, 0.00001447, -0.00000761, 0.00000418, -0.00001299, 0.00000266, 0.00000012, 0.00001172, -0.00000492, -0.00000451, -0.00000453, 0.00001115, -0.00000886, -0.00000257, 0.00000550, -0.00000139, 0.00000025, 0.00000141, 0.00000372, -0.00001309, 0.00001201, -0.00000127, -0.00000203, },
		{ 0.00000493, 0.00000343, 0.00000684, -0.00001273, 0.00001313, -0.00000838, 0.00000039, -0.00000400, 0.00001572, -0.00001776, 0.00000643, -0.00001512, 0.00003991, -0.00005362, 0.00003365, -0.00002625, 0.00002952, -0.00001292, 0.00002153, -0.00002347, 0.00002008, -0.00001286, 0.00001455, -0.00000187, -0.00001201, 0.00000138, 0.00000210, -0.00000641, 0.00000233, 0.00000212, 0.00000790, -0.00001921, 0.00001519, -0.00001763, 0.00001574, -0.00001485, 0.00001762, -0.00002318, 0.00002020, -0.00002314, 0.00002984, -0.00004024, 0.00003245, -0.00003838, 0.00004235, -0.00004186, 0.00004144, -0.00005360, 0.00004970, -0.00007988, 0.00007760, -0.00010477, 0.00012054, -0.00020126, 0.00018123, -0.00067546, 0.00102934, 0.00005124, -0.00036664, 0.00013465, -0.00016563, 0.00010413, -0.00008592, 0.00005864, -0.00006378, 0.00004924, -0.00005836, 0.00004667, -0.00004625, 0.00004154, -0.00003384, 0.00002473, -0.00003179, 0.00003234, -0.00003005, 0.00002685, -0.00002712, 0.00003111, -0.00002668, 0.00002080, -0.00002186, 0.00001563, -0.00002184, 0.00003100, -0.00002452, 0.00002137, -0.00002833, 0.00002870, -0.00002937, 0.00003553, -0.00002440, 0.00000874, -0.00000070, -0.00000935, -0.00000899, 0.00001565, -0.00002163, 0.00002866, -0.00003074, 0.00002671, -0.00003152, 0.00001756, 0.00001254, -0.00000728, -0.00001103, 0.00001218, -0.00000249, 0.00000150, -0.00000842, -0.00000425, 0.00000332, 0.00000700, 0.00000318, -0.00000896, -0.00000228, 0.00000469, 0.00000222, -0.00001339, 0.00001234, -0.00000673, 0.00000654, -0.00000688, 0.00001312, -0.00001885, 0.00001246, -0.00000170, 0.00000261, -0.00001106, },
		{ 0.00000072, 0.00001242, -0.00001547, 0.00001260, -0.00000656, 0.00000309, -0.00001218, 0.00002226, -0.00001966, 0.00001173, -0.00002650, 0.00004366, -0.00003922, 0.00000975, -0.00000526, 0.00000821, 0.00000595, 0.00000748, -0.00000674, 0.00000279, 0.00000358, -0.00000139, 0.00001309, -0.00001815, -0.00000033, 0.00000412, -0.00000769, 0.00000444, -0.00000516, 0.00001822, -0.00002593, 0.00001910, -0.00002151, 0.00001983, -0.00002080, 0.00002520, -0.00003040, 0.00002775, -0.00003314, 0.00004012, -0.00004672, 0.00003790, -0.00004572, 0.00004844, -0.00004831, 0.00005193, -0.00006570, 0.00006640, -0.00009704, 0.00009364, -0.00012278, 0.00013989, -0.00020576, 0.00019844, -0.00055424, 0.00045204, 0.00070344, -0.00063297, 0.00022995, -0.00023293, 0.00013827, -0.00011521, 0.00008804, -0.00009220, 0.00007533, -0.00008079, 0.00006378, -0.00006154, 0.00005330, -0.00004492, 0.00003956, -0.00004755, 0.00004494, -0.00004089, 0.00003759, -0.00003780, 0.00003943, -0.00003230, 0.00002732, -0.00002927, 0.00002487, -0.00003272, 0.00003792, -0.00002818, 0.00002638, -0.00003222, 0.00002902, -0.00002718, 0.00002815, -0.00001170, -0.00000139, 0.00000205, -0.00000199, -0.00002105, 0.00002368, -0.00002585, 0.00002613, -0.00002089, 0.00001179, -0.00000976, -0.00000924, 0.00002808, -0.00000735, -0.00000998, 0.00000380, 0.00000609, -0.00000371, -0.00000119, -0.00001193, 0.00000633, 0.00000210, 0.00000980, -0.00001093, -0.00000092, -0.00000145, 0.00001063, -0.00001858, 0.00001267, -0.00000636, 0.00000757, -0.00000905, 0.00001438, -0.00001534, 0.00000540, 0.00000268, 0.00000273, -0.00001008, 0.00000266, },
		{ 0.00000991, -0.00000805, 0.00000225, 0.00000324, -0.00000189, -0.00000903, 0.00001368, -0.00000633, 0.00000121, -0.00001590, 0.00001966, -0.00000063, -0.00002808, 0.00002237, -0.00001631, 0.00002532, -0.00000725, 0.00001103, -0.00001425, 0.00001894, -0.00001516, 0.00002389, -0.00001987, -0.00000243, 0.00000231, -0.00000364, 0.00000062, -0.00000420, 0.00001592, -0.00001714, 0.00000720, -0.00000954, 0.00000749, -0.00000910, 0.00001303, -0.00001585, 0.00001191, -0.00001716, 0.00002091, -0.00002172, 0.00001109, -0.00001864, 0.00001762, -0.00001601, 0.00001985, -0.00002973, 0.00002750, -0.00004879, 0.00003392, -0.00005174, 0.00005063, -0.00007979, 0.00004124, -0.00022470, -0.00017154, 0.00110948, -0.00056797, 0.00008991, -0.00013650, 0.00004848, -0.00004994, 0.00003834, -0.00005039, 0.00003727, -0.00004406, 0.00002734, -0.00002859, 0.00002179, -0.00001782, 0.00001853, -0.00002751, 0.00002293, -0.00001944, 0.00001778, -0.00001881, 0.00001923, -0.00001176, 0.00000985, -0.00001369, 0.00001122, -0.00001959, 0.00002048, -0.00000913, 0.00000956, -0.00001404, 0.00000813, -0.00000541, 0.00000359, 0.00001318, -0.00001841, 0.00001011, -0.00000323, -0.00001828, 0.00001269, -0.00000964, 0.00000387, 0.00000505, -0.00001423, 0.00001622, -0.00003037, 0.00003206, -0.00000072, -0.00000876, -0.00000459, 0.00001180, -0.00000596, 0.00000327, -0.00001501, 0.00000463, 0.00000112, 0.00001121, -0.00000768, -0.00000302, -0.00000370, 0.00001234, -0.00001500, 0.00000479, 0.00000061, 0.00000235, -0.00000400, 0.00000712, -0.00000376, -0.00000696, 0.00001024, -0.00000114, -0.00000399, -0.00000400, 0.00000301, },
		{ 0.00000374, -0.00001066, 0.00001373, -0.00000845, -0.00000065, -0.00000171, 0.00001186, -0.00001402, 0.00000274, -0.00001118, 0.00003463, -0.00005134, 0.00003171, -0.00002318, 0.00002789, -0.00000814, 0.00001671, -0.00001904, 0.00002235, -0.00001691, 0.00002253, -0.00001191, -0.00000854, 0.00000092, 0.00000048, -0.00000380, -0.00000026, 0.00000749, -0.00000167, -0.00000980, 0.00000673, -0.00000928, 0.00000779, -0.00000559, 0.00000600, -0.00001165, 0.00000772, -0.00000846, 0.00001278, -0.00002397, 0.00001702, -0.00002195, 0.00002484, -0.00002254, 0.00001906, -0.00002722, 0.00001821, -0.00004539, 0.00003968, -0.00006003, 0.00006273, -0.00012917, 0.00006841, -0.00054931, 0.00107349, -0.00014255, -0.00022601, 0.00005336, -0.00010196, 0.00005660, -0.00004693, 0.00002438, -0.00003131, 0.00002021, -0.00003253, 0.00002423, -0.00002665, 0.00002390, -0.00001757, 0.00000880, -0.00001444, 0.00001565, -0.00001488, 0.00001268, -0.00001250, 0.00001842, -0.00001621, 0.00001071, -0.00001197, 0.00000420, -0.00000725, 0.00001818, -0.00001469, 0.00001137, -0.00001832, 0.00001977, -0.00002166, 0.00003456, -0.00002965, 0.00001476, -0.00000665, -0.00000817, -0.00000623, 0.00001187, -0.00001926, 0.00002583, -0.00002948, 0.00002602, -0.00003057, 0.00001754, 0.00001404, -0.00000946, -0.00000869, 0.00001070, -0.00000249, 0.00000238, -0.00001153, -0.00000250, 0.00000424, 0.00000774, -0.00000036, -0.00000755, -0.00000237, 0.00000803, -0.00000500, -0.00000734, 0.00000979, -0.00000491, 0.00000387, -0.00000318, 0.00000912, -0.00001767, 0.00001522, -0.00000430, 0.00000259, -0.00001067, 0.00000662, 0.00000345, },
		{ -0.00001659, 0.00001590, -0.00000830, 0.00000347, -0.00001157, 0.00002089, -0.00001924, 0.00001170, -0.00002653, 0.00004345, -0.00004238, 0.00001240, -0.00000692, 0.00001110, 0.00000593, 0.00000912, -0.00001027, 0.00001264, -0.00000579, 0.00000929, 0.00000421, -0.00001781, 0.00000162, 0.00000096, -0.00000422, 0.00000093, 0.00000105, 0.00000959, -0.00001949, 0.00001434, -0.00001701, 0.00001557, -0.00001513, 0.00001800, -0.00002416, 0.00002092, -0.00002496, 0.00003139, -0.00004032, 0.00003268, -0.00003964, 0.00004172, -0.00004036, 0.00004200, -0.00005376, 0.00005266, -0.00008421, 0.00008108, -0.00010808, 0.00012088, -0.00018659, 0.00017130, -0.00057190, 0.00064508, 0.00045181, -0.00051282, 0.00018046, -0.00019561, 0.00011349, -0.00009465, 0.00006929, -0.00007403, 0.00006021, -0.00006819, 0.00005360, -0.00005288, 0.00004521, -0.00003692, 0.00003094, -0.00003755, 0.00003621, -0.00003372, 0.00003108, -0.00003101, 0.00003479, -0.00002908, 0.00002348, -0.00002488, 0.00001875, -0.00002495, 0.00003365, -0.00002685, 0.00002459, -0.00003104, 0.00002974, -0.00003010, 0.00003717, -0.00002347, 0.00000724, -0.00000343, -0.00000254, -0.00001769, 0.00002105, -0.00002528, 0.00002531, -0.00002165, 0.00001263, -0.00001006, -0.00000861, 0.00002922, -0.00000862, -0.00000922, 0.00000430, 0.00000432, -0.00000154, -0.00000445, -0.00001111, 0.00000751, 0.00000339, 0.00000670, -0.00001071, -0.00000080, 0.00000198, 0.00000535, -0.00001670, 0.00001446, -0.00000800, 0.00000814, -0.00000925, 0.00001548, -0.00001970, 0.00001220, -0.00000201, 0.00000446, -0.00001202, 0.00000612, -0.00000024, 0.00001127, },
		{ 0.00000701, 0.00000082, -0.00000051, -0.00001017, 0.00001502, -0.00000924, 0.00000463, -0.00001960, 0.00002384, -0.00000781, -0.00002250, 0.00001761, -0.00001058, 0.00002202, -0.00000105, 0.00000273, -0.00000154, 0.00000972, -0.00000693, 0.00001988, -0.00002405, 0.00000151, -0.00000106, -0.00000113, -0.00000144, -0.00000043, 0.00001208, -0.00001720, 0.00000909, -0.00001160, 0.00000974, -0.00001028, 0.00001392, -0.00001867, 0.00001467, -0.00001975, 0.00002472, -0.00002884, 0.00001929, -0.00002617, 0.00002532, -0.00002362, 0.00002732, -0.00003790, 0.00003752, -0.00006336, 0.00005209, -0.00007237, 0.00007506, -0.00011553, 0.00009022, -0.00033918, 0.00007023, 0.00091697, -0.00055287, 0.00012007, -0.00015527, 0.00006770, -0.00006432, 0.00004917, -0.00005890, 0.00004673, -0.00005368, 0.00003700, -0.00003746, 0.00002928, -0.00002392, 0.00002316, -0.00003084, 0.00002773, -0.00002521, 0.00002340, -0.00002390, 0.00002601, -0.00001862, 0.00001502, -0.00001787, 0.00001395, -0.00002217, 0.00002734, -0.00001781, 0.00001694, -0.00002205, 0.00001785, -0.00001670, 0.00001852, -0.00000016, -0.00001139, 0.00000752, -0.00000585, -0.00001531, 0.00001231, -0.00001149, 0.00000529, 0.00000238, -0.00001213, 0.00001503, -0.00002931, 0.00003266, -0.00000073, -0.00000966, -0.00000256, 0.00000918, -0.00000376, 0.00000105, -0.00001578, 0.00000637, 0.00000228, 0.00000953, -0.00000930, -0.00000202, -0.00000160, 0.00001067, -0.00001806, 0.00001057, -0.00000375, 0.00000554, -0.00000743, 0.00001203, -0.00001131, 0.00000094, 0.00000594, 0.00000060, -0.00000671, -0.00000022, 0.00000192, 0.00001029, -0.00001155, },
		{ 0.00001261, -0.00000760, -0.00000211, 0.00000063, 0.00000781, -0.00000976, -0.00000177, -0.00000605, 0.00002750, -0.00004692, 0.00002779, -0.00001854, 0.00002517, -0.00000181, 0.00000925, -0.00000979, 0.00001908, -0.00001548, 0.00002565, -0.00002068, -0.00000355, -0.00000177, 0.00000138, -0.00000379, 0.00000031, 0.00000877, -0.00000745, -0.00000297, 0.00000025, -0.00000275, 0.00000193, 0.00000103, -0.00000300, -0.00000259, -0.00000195, 0.00000332, -0.00000212, -0.00000892, 0.00000277, -0.00000700, 0.00000921, -0.00000575, -0.00000017, -0.00000428, -0.00001035, -0.00001285, 0.00000344, -0.00001714, 0.00000595, -0.00005572, -0.00005791, -0.00036962, 0.00102805, -0.00026502, -0.00011909, -0.00001060, -0.00004988, 0.00001744, -0.00001514, -0.00000312, -0.00000472, -0.00000424, -0.00001017, 0.00000451, -0.00000954, 0.00000882, -0.00000408, -0.00000382, -0.00000032, 0.00000142, -0.00000161, 0.00000022, 0.00000096, 0.00000617, -0.00000624, 0.00000160, -0.00000362, -0.00000517, 0.00000614, 0.00000484, -0.00000367, -0.00000003, -0.00000607, 0.00000743, -0.00000866, 0.00002656, -0.00002898, 0.00001736, -0.00001176, -0.00000573, -0.00000484, 0.00000879, -0.00001713, 0.00002290, -0.00002774, 0.00002532, -0.00003006, 0.00001816, 0.00001475, -0.00001110, -0.00000645, 0.00000870, -0.00000155, 0.00000206, -0.00001402, -0.00000033, 0.00000500, 0.00000773, -0.00000356, -0.00000592, -0.00000177, 0.00000971, -0.00001142, -0.00000016, 0.00000536, -0.00000151, -0.00000020, 0.00000226, 0.00000235, -0.00001253, 0.00001424, -0.00000468, 0.00000111, -0.00000851, 0.00000687, 0.00000393, -0.00000010, -0.00000681, },
		{ -0.00000924, 0.00000322, -0.00001031, 0.00001834, -0.00001710, 0.00000959, -0.00002407, 0.00004028, -0.00004262, 0.00001255, -0.00000650, 0.00001204, 0.00000873, 0.00000685, -0.00000856, 0.00001792, -0.00001191, 0.00001847, -0.00000693, -0.00001366, 0.00000040, 0.00000038, -0.00000284, -0.00000049, 0.00000515, 0.00000182, -0.00001234, 0.00000833, -0.00001120, 0.00001019, -0.00000850, 0.00000933, -0.00001597, 0.00001205, -0.00001421, 0.00001886, -0.00002917, 0.00002274, -0.00002895, 0.00003035, -0.00002768, 0.00002687, -0.00003550, 0.00003068, -0.00006097, 0.00005646, -0.00007925, 0.00008359, -0.00014211, 0.00010272, -0.00051736, 0.00075924, 0.00023472, -0.00038722, 0.00011852, -0.00014741, 0.00007924, -0.00006652, 0.00004429, -0.00004987, 0.00003888, -0.00004946, 0.00003768, -0.00003923, 0.00003292, -0.00002536, 0.00001936, -0.00002441, 0.00002367, -0.00002256, 0.00002068, -0.00002007, 0.00002579, -0.00002213, 0.00001661, -0.00001791, 0.00001008, -0.00001335, 0.00002410, -0.00002021, 0.00001764, -0.00002415, 0.00002415, -0.00002595, 0.00003934, -0.00003130, 0.00001513, -0.00001035, -0.00000071, -0.00001567, 0.00001864, -0.00002443, 0.00002401, -0.00002183, 0.00001359, -0.00001083, -0.00000749, 0.00002996, -0.00000993, -0.00000799, 0.00000392, 0.00000341, -0.00000003, -0.00000798, -0.00000919, 0.00000819, 0.00000453, 0.00000300, -0.00000947, -0.00000083, 0.00000534, -0.00000145, -0.00001153, 0.00001288, -0.00000696, 0.00000616, -0.00000646, 0.00001279, -0.00001989, 0.00001596, -0.00000537, 0.00000525, -0.00001254, 0.00000854, -0.00000079, 0.00000895, -0.00001561, 0.00001756, },
		{ 0.00000072, -0.00001078, 0.00001512, -0.00001046, 0.00000615, -0.00002119, 0.00002577, -0.00001316, -0.00001819, 0.00001381, -0.00000600, 0.00002083, 0.00000244, -0.00000261, 0.00001004, -0.00000024, 0.00000383, 0.00001060, -0.00002334, 0.00000262, -0.00000289, 0.00000074, -0.00000327, 0.00000374, 0.00000615, -0.00001383, 0.00000749, -0.00001033, 0.00000881, -0.00000815, 0.00001074, -0.00001688, 0.00001257, -0.00001672, 0.00002173, -0.00002866, 0.00002052, -0.00002679, 0.00002579, -0.00002346, 0.00002596, -0.00003559, 0.00003491, -0.00006330, 0.00005412, -0.00007460, 0.00007685, -0.00012238, 0.00009899, -0.00040040, 0.00027933, 0.00070208, -0.00048947, 0.00011230, -0.00014671, 0.00006544, -0.00006157, 0.00004545, -0.00005401, 0.00004356, -0.00005178, 0.00003644, -0.00003743, 0.00002896, -0.00002293, 0.00002116, -0.00002725, 0.00002518, -0.00002389, 0.00002235, -0.00002241, 0.00002650, -0.00002017, 0.00001554, -0.00001755, 0.00001184, -0.00001862, 0.00002735, -0.00002044, 0.00001885, -0.00002454, 0.00002221, -0.00002308, 0.00003021, -0.00001387, -0.00000148, 0.00000133, -0.00000553, -0.00001325, 0.00001162, -0.00001277, 0.00000616, 0.00000009, -0.00000970, 0.00001339, -0.00002801, 0.00003330, -0.00000125, -0.00000981, -0.00000135, 0.00000699, -0.00000144, -0.00000221, -0.00001508, 0.00000753, 0.00000367, 0.00000665, -0.00000955, -0.00000159, 0.00000148, 0.00000622, -0.00001728, 0.00001331, -0.00000605, 0.00000676, -0.00000853, 0.00001427, -0.00001660, 0.00000823, 0.00000089, 0.00000285, -0.00000926, 0.00000370, 0.00000051, 0.00000993, -0.00001370, 0.00001123, -0.00000171, },
		{ -0.00000368, 0.00000250, 0.00000439, -0.00000603, -0.00000582, -0.00000112, 0.00001992, -0.00004153, 0.00002285, -0.00001330, 0.00002262, 0.00000443, 0.00000122, 0.00000234, 0.00001145, -0.00000879, 0.00002243, -0.00002545, 0.00000034, -0.00000466, 0.00000326, -0.00000516, 0.00000300, 0.00000648, -0.00000912, 0.00000038, -0.00000321, 0.00000094, -0.00000088, 0.00000389, -0.00000799, 0.00000240, -0.00000696, 0.00000981, -0.00001195, 0.00000184, -0.00000729, 0.00000345, -0.00000140, 0.00000506, -0.00001220, 0.00001025, -0.00002986, 0.00001055, -0.00002285, 0.00001391, -0.00003688, 0.00000104, -0.00017151, -0.00016684, 0.00091290, -0.00031815, -0.00005312, -0.00004924, -0.00001638, -0.00000743, 0.00000445, -0.00001936, 0.00001156, -0.00001975, 0.00000462, -0.00000883, 0.00000192, -0.00000089, 0.00000412, -0.00001069, 0.00000787, -0.00000749, 0.00000688, -0.00000768, 0.00001040, -0.00000302, 0.00000098, -0.00000454, 0.00000117, -0.00001002, 0.00001517, -0.00000558, 0.00000522, -0.00000953, 0.00000498, -0.00000481, 0.00000639, 0.00001359, -0.00002261, 0.00001544, -0.00001442, -0.00000333, -0.00000420, 0.00000627, -0.00001527, 0.00001993, -0.00002536, 0.00002421, -0.00002956, 0.00001913, 0.00001478, -0.00001212, -0.00000458, 0.00000648, 0.00000016, 0.00000032, -0.00001523, 0.00000163, 0.00000598, 0.00000653, -0.00000575, -0.00000459, -0.00000009, 0.00000882, -0.00001522, 0.00000598, 0.00000094, 0.00000179, -0.00000394, 0.00000750, -0.00000513, -0.00000510, 0.00001038, -0.00000314, -0.00000148, -0.00000499, 0.00000594, 0.00000449, -0.00000387, -0.00000200, 0.00001023, -0.00000611, },
		{ -0.00000864, 0.00001514, -0.00001392, 0.00000614, -0.00001983, 0.00003466, -0.00004017, 0.00001024, -0.00000418, 0.00001149, 0.00001370, 0.00000132, -0.00000158, 0.00001760, -0.00001278, 0.00002294, -0.00001691, -0.00000796, -0.00000284, 0.00000194, -0.00000356, 0.00000058, 0.00000606, -0.00000326, -0.00000659, 0.00000310, -0.00000599, 0.00000558, -0.00000305, 0.00000164, -0.00000848, 0.00000407, -0.00000427, 0.00000622, -0.00001687, 0.00001154, -0.00001718, 0.00001812, -0.00001451, 0.00001150, -0.00001682, 0.00000728, -0.00003480, 0.00002799, -0.00004577, 0.00003952, -0.00008662, 0.00000963, -0.00040290, 0.00079007, 0.00007225, -0.00027815, 0.00006102, -0.00010118, 0.00004575, -0.00003931, 0.00002059, -0.00002675, 0.00001782, -0.00003035, 0.00002105, -0.00002498, 0.00002043, -0.00001398, 0.00000858, -0.00001205, 0.00001123, -0.00001109, 0.00000985, -0.00000825, 0.00001534, -0.00001396, 0.00000910, -0.00001091, 0.00000186, -0.00000130, 0.00001246, -0.00001095, 0.00000793, -0.00001375, 0.00001400, -0.00001568, 0.00003408, -0.00003312, 0.00001972, -0.00001656, 0.00000232, -0.00001472, 0.00001653, -0.00002344, 0.00002225, -0.00002124, 0.00001432, -0.00001184, -0.00000592, 0.00003014, -0.00001102, -0.00000658, 0.00000281, 0.00000356, 0.00000027, -0.00001095, -0.00000679, 0.00000872, 0.00000500, -0.00000058, -0.00000776, -0.00000040, 0.00000737, -0.00000789, -0.00000481, 0.00000911, -0.00000409, 0.00000250, -0.00000158, 0.00000705, -0.00001600, 0.00001600, -0.00000651, 0.00000452, -0.00001126, 0.00000954, -0.00000080, 0.00000570, -0.00001264, 0.00001738, -0.00000914, 0.00000231, },
		{ 0.00001396, -0.00000990, 0.00000555, -0.00002027, 0.00002481, -0.00001577, -0.00001615, 0.00001176, -0.00000316, 0.00002240, 0.00000224, -0.00000316, 0.00001783, -0.00000779, 0.00001364, -0.00000135, -0.00001859, 0.00000060, -0.00000233, 0.00000079, -0.00000342, 0.00000637, 0.00000043, -0.00000915, 0.00000411, -0.00000723, 0.00000617, -0.00000440, 0.00000533, -0.00001235, 0.00000765, -0.00001026, 0.00001410, -0.00002293, 0.00001632, -0.00002207, 0.00002082, -0.00001767, 0.00001835, -0.00002579, 0.00002285, -0.00005161, 0.00004290, -0.00006148, 0.00005928, -0.00010332, 0.00006811, -0.00039909, 0.00043035, 0.00049749, -0.00040298, 0.00008198, -0.00012097, 0.00004919, -0.00004773, 0.00003250, -0.00004057, 0.00003201, -0.00004189, 0.00002853, -0.00003101, 0.00002318, -0.00001713, 0.00001496, -0.00001927, 0.00001767, -0.00001756, 0.00001646, -0.00001587, 0.00002188, -0.00001736, 0.00001244, -0.00001404, 0.00000651, -0.00001058, 0.00002157, -0.00001740, 0.00001533, -0.00002117, 0.00002028, -0.00002273, 0.00003568, -0.00002437, 0.00000834, -0.00000672, -0.00000274, -0.00001240, 0.00001096, -0.00001365, 0.00000648, -0.00000163, -0.00000712, 0.00001130, -0.00002628, 0.00003367, -0.00000198, -0.00000940, -0.00000101, 0.00000560, 0.00000037, -0.00000589, -0.00001321, 0.00000815, 0.00000503, 0.00000297, -0.00000861, -0.00000143, 0.00000465, 0.00000007, -0.00001314, 0.00001276, -0.00000582, 0.00000553, -0.00000673, 0.00001293, -0.00001811, 0.00001291, -0.00000317, 0.00000442, -0.00001068, 0.00000689, -0.00000068, 0.00000849, -0.00001390, 0.00001403, -0.00000359, 0.00000137, -0.00001068, },
		{ 0.00000230, -0.00000376, -0.00000826, 0.00000218, 0.00001346, -0.00003666, 0.00001809, -0.00000860, 0.00002161, 0.00000875, -0.00000487, 0.00001454, 0.00000209, 0.00000102, 0.00001347, -0.00002505, 0.00000133, -0.00000594, 0.00000440, -0.00000623, 0.00000607, 0.00000223, -0.00000766, 0.00000066, -0.00000383, 0.00000192, -0.00000086, 0.00000319, -0.00000899, 0.00000327, -0.00000715, 0.00001047, -0.00001564, 0.00000697, -0.00001187, 0.00000816, -0.00000581, 0.00000874, -0.00001567, 0.00001443, -0.00003748, 0.00002114, -0.00003476, 0.00002745, -0.00005779, 0.00002831, -0.00024965, 0.00002649, 0.00075694, -0.00031596, -0.00002474, -0.00006265, -0.00000219, -0.00001718, 0.00001123, -0.00002380, 0.00001675, -0.00002523, 0.00001058, -0.00001459, 0.00000675, -0.00000450, 0.00000654, -0.00001144, 0.00000956, -0.00001019, 0.00000957, -0.00000984, 0.00001445, -0.00000781, 0.00000439, -0.00000694, 0.00000181, -0.00000947, 0.00001821, -0.00001092, 0.00000964, -0.00001455, 0.00001183, -0.00001347, 0.00001958, -0.00000092, -0.00001248, 0.00000941, -0.00001395, -0.00000190, -0.00000378, 0.00000410, -0.00001369, 0.00001705, -0.00002239, 0.00002250, -0.00002873, 0.00002010, 0.00001432, -0.00001253, -0.00000332, 0.00000445, 0.00000221, -0.00000264, -0.00001492, 0.00000299, 0.00000728, 0.00000404, -0.00000658, -0.00000376, 0.00000251, 0.00000529, -0.00001549, 0.00000951, -0.00000187, 0.00000350, -0.00000576, 0.00001061, -0.00001101, 0.00000233, 0.00000528, -0.00000068, -0.00000433, -0.00000091, 0.00000438, 0.00000458, -0.00000667, 0.00000269, 0.00000742, -0.00000451, -0.00000497, 0.00000343, },
		{ -0.00001072, 0.00000251, -0.00001509, 0.00002786, -0.00003606, 0.00000625, -0.00000075, 0.00001055, 0.00001939, -0.00000571, 0.00000917, 0.00001241, -0.00000826, 0.00002095, -0.00002284, -0.00000338, -0.00000613, 0.00000420, -0.00000532, 0.00000335, 0.00000405, -0.00000508, -0.00000334, -0.00000014, -0.00000256, 0.00000292, -0.00000012, -0.00000344, -0.00000350, -0.00000093, 0.00000228, -0.00000345, -0.00000661, 0.00000219, -0.00000747, 0.00000830, -0.00000447, 0.00000011, -0.00000285, -0.00001133, -0.00001287, 0.00000378, -0.00001731, 0.00000088, -0.00003652, -0.00008553, -0.00025192, 0.00074836, -0.00002952, -0.00019815, 0.00001909, -0.00006594, 0.00002040, -0.00001915, 0.00000364, -0.00000988, 0.00000204, -0.00001550, 0.00000782, -0.00001371, 0.00001090, -0.00000570, 0.00000142, -0.00000351, 0.00000206, -0.00000244, 0.00000171, 0.00000140, 0.00000618, -0.00000692, 0.00000308, -0.00000602, -0.00000333, 0.00000797, 0.00000215, -0.00000229, -0.00000137, -0.00000306, 0.00000246, -0.00000208, 0.00002296, -0.00002872, 0.00001961, -0.00002030, 0.00000523, -0.00001431, 0.00001463, -0.00002234, 0.00002010, -0.00001976, 0.00001449, -0.00001272, -0.00000411, 0.00002980, -0.00001175, -0.00000526, 0.00000125, 0.00000467, -0.00000093, -0.00001266, -0.00000457, 0.00000947, 0.00000432, -0.00000332, -0.00000614, 0.00000089, 0.00000716, -0.00001219, 0.00000140, 0.00000489, -0.00000094, -0.00000126, 0.00000361, 0.00000005, -0.00000941, 0.00001289, -0.00000549, 0.00000241, -0.00000836, 0.00000917, -0.00000053, 0.00000218, -0.00000838, 0.00001567, -0.00000817, 0.00000099, -0.00000704, 0.00001209, },
		{ 0.00000331, -0.00001726, 0.00002115, -0.00001552, -0.00001664, 0.00001160, -0.00000194, 0.00002646, -0.00000151, 0.00000163, 0.00002036, -0.00001067, 0.00001913, -0.00001247, -0.00001208, -0.00000343, 0.00000021, -0.00000100, -0.00000147, 0.00000643, -0.00000336, -0.00000509, 0.00000074, -0.00000394, 0.00000347, -0.00000085, -0.00000022, -0.00000732, 0.00000234, -0.00000322, 0.00000486, -0.00001453, 0.00000935, -0.00001473, 0.00001333, -0.00000957, 0.00000842, -0.00001319, 0.00000669, -0.00003408, 0.00002459, -0.00004004, 0.00003067, -0.00006834, 0.00000831, -0.00033962, 0.00051141, 0.00032746, -0.00031614, 0.00004509, -0.00008963, 0.00002796, -0.00003016, 0.00001684, -0.00002466, 0.00001763, -0.00002891, 0.00001752, -0.00002186, 0.00001526, -0.00000972, 0.00000775, -0.00001031, 0.00000855, -0.00000934, 0.00000865, -0.00000702, 0.00001447, -0.00001216, 0.00000762, -0.00000941, 0.00000048, -0.00000093, 0.00001262, -0.00001079, 0.00000818, -0.00001347, 0.00001313, -0.00001591, 0.00003378, -0.00002915, 0.00001528, -0.00001441, 0.00000137, -0.00001252, 0.00001040, -0.00001420, 0.00000622, -0.00000253, -0.00000477, 0.00000899, -0.00002415, 0.00003362, -0.00000271, -0.00000863, -0.00000149, 0.00000528, 0.00000103, -0.00000915, -0.00001075, 0.00000858, 0.00000585, -0.00000084, -0.00000698, -0.00000102, 0.00000682, -0.00000609, -0.00000714, 0.00000983, -0.00000365, 0.00000246, -0.00000264, 0.00000840, -0.00001558, 0.00001403, -0.00000515, 0.00000456, -0.00001041, 0.00000878, -0.00000138, 0.00000607, -0.00001209, 0.00001501, -0.00000445, 0.00000127, -0.00000998, 0.00001192, -0.00000803, },
		{ -0.00000845, 0.00000294, 0.00000928, -0.00003353, 0.00001454, -0.00000527, 0.00002309, 0.00000984, -0.00000690, 0.00002398, -0.00000588, 0.00001054, 0.00000133, -0.00002022, -0.00000100, -0.00000476, 0.00000372, -0.00000571, 0.00000781, -0.00000202, -0.00000478, -0.00000086, -0.00000262, 0.00000119, 0.00000087, 0.00000016, -0.00000718, 0.00000126, -0.00000390, 0.00000652, -0.00001395, 0.00000697, -0.00001148, 0.00000779, -0.00000495, 0.00000650, -0.00001201, 0.00000965, -0.00003422, 0.00001949, -0.00003268, 0.00002356, -0.00005585, 0.00002194, -0.00027857, 0.00018305, 0.00059075, -0.00027903, -0.00002334, -0.00005707, -0.00000294, -0.00001521, 0.00000828, -0.00001931, 0.00001331, -0.00002257, 0.00000915, -0.00001394, 0.00000609, -0.00000322, 0.00000458, -0.00000760, 0.00000611, -0.00000777, 0.00000732, -0.00000690, 0.00001343, -0.00000836, 0.00000430, -0.00000619, -0.00000079, -0.00000443, 0.00001559, -0.00001084, 0.00000895, -0.00001410, 0.00001295, -0.00001624, 0.00002762, -0.00001328, -0.00000152, 0.00000090, -0.00001073, -0.00000177, -0.00000327, 0.00000221, -0.00001249, 0.00001453, -0.00001909, 0.00002024, -0.00002742, 0.00002084, 0.00001360, -0.00001240, -0.00000282, 0.00000308, 0.00000395, -0.00000621, -0.00001331, 0.00000374, 0.00000869, 0.00000055, -0.00000610, -0.00000327, 0.00000533, -0.00000003, -0.00001244, 0.00000996, -0.00000239, 0.00000296, -0.00000488, 0.00001049, -0.00001360, 0.00000767, 0.00000074, 0.00000147, -0.00000645, 0.00000282, 0.00000277, 0.00000380, -0.00000780, 0.00000629, 0.00000496, -0.00000332, -0.00000566, 0.00000322, 0.00000183, -0.00000341, },
		{ -0.00001130, 0.00002145, -0.00003174, 0.00000177, 0.00000265, 0.00001061, 0.00002398, -0.00001181, 0.00002111, 0.00000462, -0.00000017, 0.00001288, -0.00002332, -0.00000185, -0.00000768, 0.00000555, -0.00000663, 0.00000632, 0.00000038, -0.00000428, -0.00000254, -0.00000120, -0.00000114, 0.00000239, 0.00000001, -0.00000544, -0.00000160, -0.00000227, 0.00000431, -0.00000850, -0.00000027, -0.00000349, -0.00000161, 0.00000268, 0.00000060, -0.00000513, 0.00000359, -0.00002138, 0.00000014, -0.00001064, -0.00000070, -0.00002317, -0.00000543, -0.00016085, -0.00009285, 0.00065582, -0.00007707, -0.00014964, -0.00000340, -0.00004526, 0.00000626, -0.00000850, -0.00000444, -0.00000144, -0.00000616, -0.00000724, 0.00000012, -0.00000724, 0.00000587, -0.00000175, -0.00000097, -0.00000013, -0.00000226, 0.00000170, -0.00000196, 0.00000696, 0.00000012, -0.00000251, -0.00000026, -0.00000428, -0.00000418, 0.00001242, -0.00000430, 0.00000312, -0.00000744, 0.00000480, -0.00000698, 0.00001112, 0.00000911, -0.00001972, 0.00001491, -0.00002073, 0.00000707, -0.00001394, 0.00001281, -0.00002122, 0.00001774, -0.00001745, 0.00001392, -0.00001320, -0.00000233, 0.00002905, -0.00001204, -0.00000432, -0.00000035, 0.00000633, -0.00000351, -0.00001282, -0.00000296, 0.00001063, 0.00000227, -0.00000477, -0.00000488, 0.00000298, 0.00000451, -0.00001332, 0.00000548, 0.00000183, 0.00000102, -0.00000352, 0.00000725, -0.00000600, -0.00000226, 0.00000811, -0.00000317, -0.00000037, -0.00000453, 0.00000790, -0.00000046, -0.00000073, -0.00000384, 0.00001320, -0.00000683, -0.00000030, -0.00000604, 0.00000995, -0.00000850, -0.00000007, },
		{ 0.00001582, -0.00001319, -0.00001915, 0.00001276, -0.00000146, 0.00003186, -0.00000742, 0.00001075, 0.00001785, -0.00000834, 0.00001828, -0.00001964, -0.00000656, -0.00000751, 0.00000334, -0.00000366, 0.00000191, 0.00000404, -0.00000457, -0.00000277, -0.00000144, -0.00000163, 0.00000187, 0.00000113, -0.00000427, -0.00000356, -0.00000136, 0.00000194, -0.00000309, -0.00000639, 0.00000240, -0.00000759, 0.00000628, -0.00000241, -0.00000000, -0.00000245, -0.00000793, -0.00001711, 0.00000637, -0.00001867, 0.00000146, -0.00003106, -0.00006275, -0.00023808, 0.00052511, 0.00020424, -0.00024461, 0.00001397, -0.00006219, 0.00000931, -0.00001507, 0.00000397, -0.00001147, 0.00000538, -0.00001737, 0.00000738, -0.00001346, 0.00000831, -0.00000353, 0.00000236, -0.00000336, 0.00000095, -0.00000227, 0.00000191, 0.00000126, 0.00000687, -0.00000674, 0.00000306, -0.00000569, -0.00000380, 0.00000729, 0.00000366, -0.00000353, 0.00000023, -0.00000427, 0.00000342, -0.00000483, 0.00002549, -0.00002755, 0.00001762, -0.00001983, 0.00000546, -0.00001312, 0.00000984, -0.00001446, 0.00000539, -0.00000246, -0.00000298, 0.00000682, -0.00002181, 0.00003316, -0.00000328, -0.00000776, -0.00000258, 0.00000603, 0.00000017, -0.00001123, -0.00000840, 0.00000921, 0.00000560, -0.00000404, -0.00000516, -0.00000001, 0.00000708, -0.00001058, -0.00000118, 0.00000609, -0.00000090, -0.00000104, 0.00000221, 0.00000221, -0.00001010, 0.00001187, -0.00000487, 0.00000317, -0.00000840, 0.00000927, -0.00000171, 0.00000318, -0.00000878, 0.00001436, -0.00000435, 0.00000060, -0.00000910, 0.00000967, -0.00000572, 0.00000044, -0.00001324, },
		{ 0.00000781, -0.00003272, 0.00001262, -0.00000352, 0.00002718, 0.00000743, -0.00000387, 0.00002884, -0.00000999, 0.00001630, -0.00001043, -0.00001323, -0.00000553, -0.00000153, 0.00000119, -0.00000322, 0.00000728, -0.00000475, -0.00000213, -0.00000273, -0.00000088, 0.00000002, 0.00000286, -0.00000357, -0.00000427, -0.00000175, 0.00000063, 0.00000026, -0.00000894, 0.00000362, -0.00000793, 0.00000433, -0.00000110, 0.00000114, -0.00000460, -0.00000024, -0.00002408, 0.00000971, -0.00002115, 0.00000750, -0.00003703, -0.00001272, -0.00025640, 0.00028617, 0.00043986, -0.00022841, -0.00003554, -0.00004168, -0.00001170, -0.00000712, 0.00000056, -0.00001057, 0.00000551, -0.00001550, 0.00000347, -0.00000959, 0.00000242, 0.00000055, 0.00000070, -0.00000182, 0.00000015, -0.00000264, 0.00000237, -0.00000089, 0.00000898, -0.00000600, 0.00000202, -0.00000381, -0.00000469, 0.00000283, 0.00000926, -0.00000678, 0.00000428, -0.00000905, 0.00000871, -0.00001272, 0.00002883, -0.00002064, 0.00000731, -0.00000786, -0.00000584, -0.00000274, -0.00000263, 0.00000059, -0.00001178, 0.00001267, -0.00001586, 0.00001767, -0.00002568, 0.00002123, 0.00001275, -0.00001185, -0.00000314, 0.00000272, 0.00000468, -0.00000953, -0.00001100, 0.00000422, 0.00000973, -0.00000335, -0.00000468, -0.00000273, 0.00000739, -0.00000566, -0.00000734, 0.00000797, -0.00000098, 0.00000056, -0.00000165, 0.00000720, -0.00001237, 0.00000974, -0.00000197, 0.00000239, -0.00000710, 0.00000552, 0.00000147, 0.00000213, -0.00000704, 0.00000827, 0.00000335, -0.00000281, -0.00000575, 0.00000207, 0.00000274, -0.00000476, -0.00000650, 0.00000106, },
		{ -0.00002858, -0.00000206, 0.00000500, 0.00001283, 0.00002596, -0.00001472, 0.00003142, -0.00000280, 0.00000820, 0.00000113, -0.00001895, -0.00000386, -0.00000657, 0.00000492, -0.00000627, 0.00000796, -0.00000325, -0.00000222, -0.00000330, -0.00000073, -0.00000115, 0.00000331, -0.00000192, -0.00000496, -0.00000217, -0.00000062, 0.00000230, -0.00000892, 0.00000182, -0.00000517, 0.00000019, 0.00000128, 0.00000086, -0.00000449, 0.00000256, -0.00002237, 0.00000313, -0.00001376, 0.00000192, -0.00002910, -0.00000041, -0.00020051, 0.00004778, 0.00053876, -0.00008520, -0.00012704, -0.00000892, -0.00003772, 0.00000228, -0.00000630, -0.00000475, -0.00000053, -0.00000726, -0.00000537, -0.00000211, -0.00000543, 0.00000504, -0.00000166, 0.00000075, -0.00000139, -0.00000199, 0.00000143, -0.00000106, 0.00000804, -0.00000233, -0.00000100, -0.00000095, -0.00000537, -0.00000098, 0.00001176, -0.00000599, 0.00000412, -0.00000880, 0.00000784, -0.00001168, 0.00002048, -0.00000379, -0.00000889, 0.00000704, -0.00001804, 0.00000743, -0.00001333, 0.00001103, -0.00002022, 0.00001546, -0.00001461, 0.00001268, -0.00001315, -0.00000076, 0.00002803, -0.00001189, -0.00000399, -0.00000146, 0.00000790, -0.00000689, -0.00001157, -0.00000202, 0.00001205, -0.00000097, -0.00000483, -0.00000397, 0.00000533, 0.00000009, -0.00001131, 0.00000677, 0.00000074, 0.00000099, -0.00000336, 0.00000807, -0.00000931, 0.00000335, 0.00000344, -0.00000078, -0.00000283, -0.00000063, 0.00000626, -0.00000099, -0.00000230, 0.00000000, 0.00001077, -0.00000568, -0.00000120, -0.00000597, 0.00000918, -0.00000787, -0.00000079, -0.00000956, 0.00001676, },
		{ -0.00002259, 0.00001418, -0.00000042, 0.00003692, -0.00001329, 0.00002200, 0.00001217, -0.00000231, 0.00001119, -0.00002121, -0.00000411, -0.00000978, 0.00000545, -0.00000576, 0.00000533, 0.00000031, -0.00000364, -0.00000232, -0.00000206, -0.00000070, 0.00000175, 0.00000106, -0.00000612, -0.00000188, -0.00000248, 0.00000385, -0.00000786, -0.00000061, -0.00000251, -0.00000261, 0.00000163, 0.00000171, -0.00000446, 0.00000330, -0.00001696, -0.00000564, -0.00000606, -0.00000432, -0.00001927, -0.00000439, -0.00012560, -0.00011755, 0.00048584, 0.00012800, -0.00019521, -0.00000490, -0.00004393, -0.00000241, -0.00000595, -0.00000314, -0.00000393, -0.00000180, -0.00001013, 0.00000072, -0.00000804, 0.00000426, -0.00000025, 0.00000034, -0.00000019, -0.00000318, 0.00000162, -0.00000166, 0.00000677, 0.00000107, -0.00000277, 0.00000012, -0.00000412, -0.00000479, 0.00001185, -0.00000264, 0.00000167, -0.00000570, 0.00000339, -0.00000556, 0.00000714, 0.00001352, -0.00002073, 0.00001510, -0.00002191, 0.00000849, -0.00001367, 0.00000917, -0.00001447, 0.00000417, -0.00000147, -0.00000195, 0.00000504, -0.00001948, 0.00003236, -0.00000359, -0.00000706, -0.00000389, 0.00000752, -0.00000216, -0.00001176, -0.00000664, 0.00001030, 0.00000396, -0.00000607, -0.00000352, 0.00000163, 0.00000517, -0.00001231, 0.00000312, 0.00000307, 0.00000103, -0.00000344, 0.00000601, -0.00000360, -0.00000360, 0.00000772, -0.00000300, 0.00000079, -0.00000517, 0.00000865, -0.00000205, 0.00000060, -0.00000488, 0.00001273, -0.00000372, -0.00000025, -0.00000855, 0.00000800, -0.00000397, -0.00000187, -0.00000960, 0.00001030, -0.00001013, },
		{ 0.00001200, -0.00000273, 0.00003307, 0.00000244, 0.00000374, 0.00002887, -0.00000943, 0.00001604, -0.00001850, -0.00000693, -0.00001030, 0.00000238, -0.00000224, 0.00000061, 0.00000459, -0.00000533, -0.00000075, -0.00000387, 0.00000036, -0.00000055, 0.00000388, -0.00000655, -0.00000180, -0.00000402, 0.00000426, -0.00000580, -0.00000306, -0.00000081, -0.00000347, 0.00000014, 0.00000310, -0.00000419, 0.00000268, -0.00001055, -0.00001234, -0.00000241, -0.00000690, -0.00001249, -0.00001179, -0.00006282, -0.00019271, 0.00033206, 0.00032030, -0.00018068, -0.00004949, -0.00002519, -0.00002165, 0.00000154, -0.00000704, -0.00000219, -0.00000226, -0.00000802, -0.00000296, -0.00000457, -0.00000153, 0.00000427, -0.00000257, 0.00000316, -0.00000551, 0.00000248, -0.00000261, 0.00000564, 0.00000329, -0.00000253, -0.00000081, -0.00000153, -0.00000772, 0.00000961, 0.00000202, -0.00000123, -0.00000200, -0.00000169, 0.00000118, -0.00000445, 0.00002357, -0.00002184, 0.00001195, -0.00001482, -0.00000064, -0.00000430, -0.00000199, -0.00000073, -0.00001158, 0.00001166, -0.00001305, 0.00001514, -0.00002370, 0.00002129, 0.00001192, -0.00001108, -0.00000414, 0.00000346, 0.00000396, -0.00001183, -0.00000867, 0.00000484, 0.00000984, -0.00000691, -0.00000278, -0.00000183, 0.00000790, -0.00001003, -0.00000190, 0.00000490, 0.00000122, -0.00000252, 0.00000265, 0.00000200, -0.00000810, 0.00000857, -0.00000244, 0.00000176, -0.00000606, 0.00000693, 0.00000047, -0.00000007, -0.00000475, 0.00000869, 0.00000265, -0.00000289, -0.00000554, 0.00000056, 0.00000431, -0.00000699, -0.00000328, -0.00000271, 0.00000853, -0.00003397, },
		{ 0.00000587, 0.00001765, 0.00002467, -0.00001307, 0.00003805, -0.00000733, 0.00001339, -0.00001072, -0.00001195, -0.00000841, -0.00000319, 0.00000229, -0.00000389, 0.00000744, -0.00000551, -0.00000029, -0.00000443, 0.00000027, -0.00000161, 0.00000457, -0.00000471, -0.00000323, -0.00000386, 0.00000244, -0.00000213, -0.00000599, 0.00000068, -0.00000388, -0.00000093, 0.00000273, -0.00000190, -0.00000016, -0.00000346, -0.00001676, -0.00000157, -0.00000804, -0.00000672, -0.00001965, -0.00002016, -0.00019842, 0.00015109, 0.00042156, -0.00007161, -0.00012048, -0.00000380, -0.00003865, 0.00000475, -0.00000918, -0.00000054, -0.00000418, -0.00000383, -0.00000781, -0.00000062, -0.00000665, 0.00000679, -0.00000367, 0.00000466, -0.00000541, 0.00000119, -0.00000171, 0.00000310, 0.00000559, -0.00000187, -0.00000166, 0.00000003, -0.00000789, 0.00000459, 0.00000734, -0.00000374, 0.00000123, -0.00000571, 0.00000581, -0.00001067, 0.00002390, -0.00001270, 0.00000075, -0.00000183, -0.00001321, 0.00000645, -0.00001245, 0.00000933, -0.00001946, 0.00001358, -0.00001162, 0.00001104, -0.00001264, 0.00000052, 0.00002681, -0.00001133, -0.00000441, -0.00000169, 0.00000866, -0.00001025, -0.00000946, -0.00000147, 0.00001329, -0.00000489, -0.00000371, -0.00000318, 0.00000714, -0.00000484, -0.00000717, 0.00000568, 0.00000148, -0.00000084, -0.00000089, 0.00000589, -0.00000915, 0.00000608, 0.00000027, 0.00000067, -0.00000415, 0.00000260, 0.00000467, -0.00000225, -0.00000222, 0.00000253, 0.00000893, -0.00000503, -0.00000158, -0.00000676, 0.00000972, -0.00000885, 0.00000049, -0.00001023, 0.00001444, -0.00002738, -0.00000457, },
		{ 0.00000244, 0.00004002, -0.00001687, 0.00003271, 0.00000598, 0.00000439, 0.00000009, -0.00001757, -0.00000543, -0.00000924, 0.00000545, -0.00000611, 0.00000741, -0.00000328, -0.00000169, -0.00000308, -0.00000153, -0.00000081, 0.00000272, -0.00000071, -0.00000596, -0.00000210, -0.00000125, 0.00000246, -0.00000891, 0.00000200, -0.00000458, -0.00000049, -0.00000002, 0.00000233, -0.00000449, 0.00000332, -0.00001901, -0.00000173, -0.00001009, -0.00000035, -0.00002664, 0.00000336, -0.00016432, -0.00000193, 0.00041453, 0.00008977, -0.00016678, -0.00001098, -0.00003561, -0.00000652, -0.00000319, -0.00000431, -0.00000225, -0.00000343, -0.00000786, -0.00000178, -0.00000614, 0.00000346, 0.00000001, 0.00000168, -0.00000089, -0.00000352, 0.00000188, -0.00000141, 0.00000861, -0.00000196, -0.00000092, -0.00000085, -0.00000477, -0.00000238, 0.00001206, -0.00000500, 0.00000331, -0.00000783, 0.00000725, -0.00001102, 0.00001659, 0.00000132, -0.00001120, 0.00000884, -0.00002060, 0.00000993, -0.00001388, 0.00000836, -0.00001438, 0.00000283, 0.00000017, -0.00000160, 0.00000377, -0.00001731, 0.00003129, -0.00000356, -0.00000682, -0.00000490, 0.00000912, -0.00000548, -0.00001080, -0.00000560, 0.00001179, 0.00000099, -0.00000668, -0.00000217, 0.00000350, 0.00000160, -0.00001116, 0.00000499, 0.00000168, 0.00000126, -0.00000373, 0.00000744, -0.00000723, 0.00000187, 0.00000327, -0.00000069, -0.00000167, -0.00000150, 0.00000741, -0.00000278, -0.00000095, -0.00000134, 0.00001086, -0.00000308, -0.00000086, -0.00000872, 0.00000746, -0.00000352, -0.00000270, -0.00000754, 0.00000598, -0.00000772, -0.00002578, 0.00001479, },
		{ 0.00003929, -0.00000324, 0.00001417, 0.00002546, -0.00000537, 0.00000962, -0.00002097, -0.00000363, -0.00001332, 0.00000531, -0.00000514, 0.00000447, 0.00000073, -0.00000410, -0.00000083, -0.00000389, 0.00000065, -0.00000009, 0.00000339, -0.00000797, -0.00000065, -0.00000456, 0.00000560, -0.00000979, 0.00000166, -0.00000442, 0.00000005, -0.00000292, 0.00000565, -0.00000716, 0.00000685, -0.00001747, -0.00000360, -0.00001158, 0.00000375, -0.00002827, 0.00000857, -0.00011213, -0.00010503, 0.00032855, 0.00023730, -0.00014525, -0.00005760, -0.00001344, -0.00002812, 0.00000705, -0.00001133, 0.00000274, -0.00000693, -0.00000305, -0.00000750, -0.00000114, -0.00000380, 0.00000621, -0.00000355, 0.00000549, -0.00000883, 0.00000552, -0.00000548, 0.00001050, -0.00000164, 0.00000044, -0.00000286, -0.00000059, -0.00000835, 0.00001372, -0.00000356, 0.00000323, -0.00000724, 0.00000517, -0.00000670, 0.00000565, 0.00001402, -0.00001757, 0.00001177, -0.00001864, 0.00000371, -0.00000593, -0.00000149, -0.00000177, -0.00001179, 0.00001149, -0.00001088, 0.00001291, -0.00002168, 0.00002106, 0.00001120, -0.00001032, -0.00000550, 0.00000506, 0.00000172, -0.00001263, -0.00000688, 0.00000593, 0.00000860, -0.00000949, -0.00000082, -0.00000055, 0.00000655, -0.00001207, 0.00000232, 0.00000219, 0.00000289, -0.00000484, 0.00000635, -0.00000329, -0.00000249, 0.00000520, -0.00000117, -0.00000005, -0.00000363, 0.00000717, -0.00000045, -0.00000212, -0.00000170, 0.00000803, 0.00000252, -0.00000321, -0.00000552, -0.00000059, 0.00000560, -0.00000901, -0.00000006, -0.00000703, 0.00001030, -0.00003635, 0.00001179, -0.00000175, },
		{ 0.00002063, -0.00000689, 0.00004039, -0.00000796, 0.00001301, -0.00001920, -0.00000526, -0.00001349, 0.00000109, -0.00000142, -0.00000010, 0.00000486, -0.00000583, 0.00000058, -0.00000503, 0.00000095, -0.00000169, 0.00000518, -0.00000713, -0.00000150, -0.00000529, 0.00000513, -0.00000695, -0.00000165, -0.00000203, -0.00000125, -0.00000323, 0.00000504, -0.00000528, 0.00000482, -0.00001077, -0.00000863, -0.00000958, 0.00000150, -0.00002055, -0.00000236, -0.00005605, -0.00015895, 0.00020900, 0.00032173, -0.00005193, -0.00011981, 0.00000481, -0.00004253, 0.00000921, -0.00001321, 0.00000443, -0.00000880, 0.00000089, -0.00001175, 0.00000215, -0.00000870, 0.00000907, -0.00000570, 0.00000845, -0.00000982, 0.00000499, -0.00000550, 0.00000840, 0.00000134, 0.00000014, -0.00000326, 0.00000133, -0.00001004, 0.00001024, 0.00000150, 0.00000043, -0.00000367, 0.00000009, 0.00000018, -0.00000487, 0.00002120, -0.00001603, 0.00000690, -0.00000956, -0.00000759, 0.00000458, -0.00001141, 0.00000780, -0.00001904, 0.00001233, -0.00000888, 0.00000932, -0.00001186, 0.00000155, 0.00002549, -0.00001047, -0.00000553, -0.00000083, 0.00000808, -0.00001276, -0.00000718, -0.00000086, 0.00001378, -0.00000878, -0.00000182, -0.00000229, 0.00000770, -0.00000887, -0.00000243, 0.00000335, 0.00000309, -0.00000348, 0.00000283, 0.00000167, -0.00000600, 0.00000572, -0.00000081, 0.00000068, -0.00000393, 0.00000474, 0.00000323, -0.00000397, -0.00000071, 0.00000367, 0.00000784, -0.00000491, -0.00000167, -0.00000798, 0.00001101, -0.00001083, 0.00000313, -0.00001279, 0.00001426, -0.00002811, -0.00000587, 0.00000558, 0.00002460, },
		{ -0.00001656, 0.00004079, 0.00000170, 0.00000837, -0.00001150, -0.00001087, -0.00000961, -0.00000617, 0.00000324, -0.00000436, 0.00000734, -0.00000554, 0.00000009, -0.00000413, -0.00000059, -0.00000126, 0.00000397, -0.00000328, -0.00000463, -0.00000332, 0.00000125, -0.00000116, -0.00000694, 0.00000183, -0.00000421, -0.00000075, 0.00000063, 0.00000045, -0.00000140, -0.00000095, -0.00001537, -0.00000438, -0.00000667, -0.00000585, -0.00002112, -0.00000924, -0.00017051, 0.00008979, 0.00033277, 0.00007600, -0.00015306, -0.00000794, -0.00003461, -0.00000510, -0.00000480, -0.00000158, -0.00000456, -0.00000106, -0.00000938, -0.00000106, -0.00000682, 0.00000492, -0.00000165, 0.00000509, -0.00000423, -0.00000113, -0.00000057, 0.00000192, 0.00000715, -0.00000240, -0.00000095, -0.00000037, -0.00000671, 0.00000228, 0.00000874, -0.00000374, 0.00000141, -0.00000590, 0.00000660, -0.00001160, 0.00002117, -0.00000797, -0.00000189, 0.00000084, -0.00001670, 0.00000978, -0.00001366, 0.00000744, -0.00001431, 0.00000168, 0.00000209, -0.00000170, 0.00000292, -0.00001534, 0.00002998, -0.00000317, -0.00000722, -0.00000517, 0.00001014, -0.00000901, -0.00000881, -0.00000506, 0.00001330, -0.00000294, -0.00000593, -0.00000105, 0.00000495, -0.00000257, -0.00000793, 0.00000465, 0.00000193, -0.00000018, -0.00000185, 0.00000612, -0.00000780, 0.00000490, -0.00000003, 0.00000098, -0.00000334, 0.00000187, 0.00000594, -0.00000403, -0.00000110, 0.00000121, 0.00000934, -0.00000280, -0.00000106, -0.00000966, 0.00000812, -0.00000453, -0.00000173, -0.00000759, 0.00000370, -0.00000693, -0.00002790, 0.00001393, 0.00000781, 0.00004027, },
		{ 0.00002505, 0.00002093, -0.00000057, -0.00000094, -0.00001802, -0.00000418, -0.00001348, 0.00000610, -0.00000631, 0.00000703, -0.00000294, -0.00000204, -0.00000186, -0.00000304, 0.00000019, 0.00000119, 0.00000151, -0.00000780, -0.00000089, -0.00000332, 0.00000433, -0.00001091, 0.00000417, -0.00000621, 0.00000173, -0.00000404, 0.00000581, -0.00000694, 0.00000677, -0.00001918, -0.00000035, -0.00001478, 0.00000700, -0.00003453, 0.00001565, -0.00014592, -0.00001366, 0.00029115, 0.00018696, -0.00012429, -0.00005739, -0.00000850, -0.00002942, 0.00000818, -0.00001142, 0.00000337, -0.00000746, -0.00000177, -0.00000904, -0.00000022, -0.00000367, 0.00000584, -0.00000187, 0.00000467, -0.00000913, 0.00000566, -0.00000526, 0.00001246, -0.00000462, 0.00000202, -0.00000356, -0.00000130, -0.00000620, 0.00001421, -0.00000600, 0.00000492, -0.00000952, 0.00000920, -0.00001220, 0.00001441, 0.00000332, -0.00000995, 0.00000752, -0.00001901, 0.00000651, -0.00000725, -0.00000120, -0.00000260, -0.00001217, 0.00001195, -0.00000931, 0.00001108, -0.00001974, 0.00002059, 0.00001069, -0.00000987, -0.00000673, 0.00000699, -0.00000168, -0.00001187, -0.00000583, 0.00000753, 0.00000593, -0.00001069, 0.00000097, 0.00000085, 0.00000373, -0.00001157, 0.00000450, 0.00000076, 0.00000315, -0.00000536, 0.00000811, -0.00000691, 0.00000254, 0.00000120, 0.00000085, -0.00000226, -0.00000046, 0.00000659, -0.00000162, -0.00000336, 0.00000124, 0.00000699, 0.00000247, -0.00000339, -0.00000609, -0.00000080, 0.00000584, -0.00000986, 0.00000197, -0.00001053, 0.00001173, -0.00003871, 0.00001088, 0.00000067, 0.00004426, -0.00000747, },
		{ 0.00003940, -0.00000561, 0.00000673, -0.00002221, -0.00000133, -0.00001703, 0.00000458, -0.00000477, 0.00000383, 0.00000116, -0.00000453, 0.00000023, -0.00000475, 0.00000092, -0.00000098, 0.00000457, -0.00000843, -0.00000059, -0.00000553, 0.00000618, -0.00001045, 0.00000233, -0.00000469, 0.00000119, -0.00000515, 0.00000653, -0.00000727, 0.00000788, -0.00001605, -0.00000199, -0.00001636, 0.00000948, -0.00003271, 0.00001346, -0.00009519, -0.00009466, 0.00022416, 0.00024772, -0.00003650, -0.00011759, 0.00001140, -0.00004502, 0.00001223, -0.00001545, 0.00000733, -0.00001154, 0.00000418, -0.00001475, 0.00000416, -0.00000978, 0.00001022, -0.00000614, 0.00001036, -0.00001269, 0.00000746, -0.00000793, 0.00001280, -0.00000290, 0.00000229, -0.00000463, 0.00000183, -0.00001043, 0.00001394, -0.00000341, 0.00000420, -0.00000824, 0.00000613, -0.00000656, 0.00000335, 0.00001397, -0.00001399, 0.00000855, -0.00001460, -0.00000242, 0.00000240, -0.00001040, 0.00000645, -0.00001889, 0.00001174, -0.00000659, 0.00000778, -0.00001101, 0.00000238, 0.00002415, -0.00000952, -0.00000710, 0.00000097, 0.00000598, -0.00001388, -0.00000534, 0.00000017, 0.00001303, -0.00001195, 0.00000043, -0.00000128, 0.00000675, -0.00001099, 0.00000149, 0.00000110, 0.00000440, -0.00000562, 0.00000631, -0.00000297, -0.00000131, 0.00000310, -0.00000006, -0.00000058, -0.00000230, 0.00000577, 0.00000181, -0.00000558, 0.00000159, 0.00000379, 0.00000723, -0.00000501, -0.00000188, -0.00000894, 0.00001220, -0.00001282, 0.00000602, -0.00001610, 0.00001524, -0.00003003, -0.00000667, 0.00000506, 0.00003248, 0.00001536, 0.00000252, },
		{ 0.00000039, 0.00000722, -0.00002002, -0.00000406, -0.00001467, -0.00000191, -0.00000027, -0.00000107, 0.00000525, -0.00000596, 0.00000089, -0.00000467, 0.00000004, -0.00000137, 0.00000468, -0.00000569, -0.00000308, -0.00000448, 0.00000362, -0.00000541, -0.00000337, 0.00000001, -0.00000248, -0.00000220, 0.00000220, -0.00000217, 0.00000255, -0.00000673, -0.00000906, -0.00001043, 0.00000066, -0.00001660, -0.00000782, -0.00003684, -0.00014483, 0.00014716, 0.00025772, 0.00007320, -0.00014620, -0.00000113, -0.00003682, -0.00000143, -0.00000786, 0.00000220, -0.00000808, 0.00000279, -0.00001252, 0.00000105, -0.00000843, 0.00000705, -0.00000358, 0.00000870, -0.00000829, 0.00000218, -0.00000394, 0.00000671, 0.00000365, -0.00000115, -0.00000199, 0.00000056, -0.00000853, 0.00000728, 0.00000379, -0.00000042, -0.00000263, -0.00000117, 0.00000234, -0.00000766, 0.00002024, -0.00001247, 0.00000476, -0.00000676, -0.00001148, 0.00000845, -0.00001312, 0.00000650, -0.00001436, 0.00000094, 0.00000391, -0.00000193, 0.00000230, -0.00001351, 0.00002848, -0.00000247, -0.00000831, -0.00000442, 0.00000997, -0.00001193, -0.00000648, -0.00000461, 0.00001429, -0.00000718, -0.00000411, -0.00000007, 0.00000543, -0.00000611, -0.00000394, 0.00000304, 0.00000305, -0.00000247, 0.00000137, 0.00000277, -0.00000554, 0.00000507, -0.00000145, 0.00000140, -0.00000370, 0.00000443, 0.00000441, -0.00000561, 0.00000001, 0.00000259, 0.00000836, -0.00000294, -0.00000099, -0.00001104, 0.00000958, -0.00000658, 0.00000060, -0.00000947, 0.00000342, -0.00000794, -0.00002884, 0.00001168, 0.00001551, 0.00003784, -0.00001202, 0.00004537, },
		{ 0.00000175, -0.00001218, -0.00001166, -0.00000785, -0.00001092, 0.00000455, -0.00000535, 0.00000751, -0.00000531, -0.00000021, -0.00000304, -0.00000194, -0.00000049, 0.00000266, -0.00000111, -0.00000658, -0.00000195, -0.00000103, 0.00000112, -0.00000941, 0.00000444, -0.00000615, 0.00000164, -0.00000344, 0.00000407, -0.00000426, 0.00000322, -0.00001621, -0.00000245, -0.00001198, 0.00000257, -0.00003038, 0.00000650, -0.00015492, 0.00006366, 0.00023782, 0.00015982, -0.00011463, -0.00005047, -0.00000926, -0.00002644, 0.00000591, -0.00000844, 0.00000077, -0.00000471, -0.00000361, -0.00000798, -0.00000137, -0.00000166, 0.00000383, 0.00000165, 0.00000144, -0.00000697, 0.00000338, -0.00000224, 0.00001148, -0.00000542, 0.00000212, -0.00000313, -0.00000306, -0.00000207, 0.00001153, -0.00000525, 0.00000353, -0.00000826, 0.00000937, -0.00001370, 0.00001936, -0.00000550, -0.00000171, 0.00000093, -0.00001647, 0.00000762, -0.00000809, -0.00000111, -0.00000332, -0.00001247, 0.00001269, -0.00000813, 0.00000958, -0.00001791, 0.00001988, 0.00001046, -0.00000994, -0.00000739, 0.00000856, -0.00000551, -0.00000996, -0.00000539, 0.00000935, 0.00000206, -0.00001042, 0.00000250, 0.00000187, 0.00000034, -0.00000915, 0.00000471, 0.00000072, 0.00000192, -0.00000388, 0.00000743, -0.00000789, 0.00000555, -0.00000198, 0.00000253, -0.00000401, 0.00000273, 0.00000556, -0.00000315, -0.00000344, 0.00000344, 0.00000611, 0.00000214, -0.00000321, -0.00000739, 0.00000015, 0.00000473, -0.00000911, 0.00000215, -0.00001229, 0.00001179, -0.00004013, 0.00000852, 0.00000535, 0.00004696, -0.00000861, 0.00003435, 0.00001745, },
		{ -0.00000359, -0.00001970, -0.00000126, -0.00001777, 0.00000601, -0.00000653, 0.00000659, -0.00000237, -0.00000251, -0.00000091, -0.00000377, 0.00000030, 0.00000040, 0.00000278, -0.00000844, -0.00000068, -0.00000443, 0.00000515, -0.00001172, 0.00000484, -0.00000627, 0.00000251, -0.00000586, 0.00000641, -0.00000697, 0.00000777, -0.00001744, 0.00000067, -0.00001891, 0.00001216, -0.00003808, 0.00002001, -0.00012465, -0.00002207, 0.00020722, 0.00019945, -0.00002949, -0.00011043, 0.00001335, -0.00004404, 0.00001226, -0.00001473, 0.00000704, -0.00001117, 0.00000472, -0.00001558, 0.00000436, -0.00000903, 0.00000956, -0.00000442, 0.00000966, -0.00001312, 0.00000763, -0.00000786, 0.00001496, -0.00000585, 0.00000365, -0.00000515, 0.00000114, -0.00000857, 0.00001465, -0.00000584, 0.00000587, -0.00001057, 0.00001017, -0.00001186, 0.00001114, 0.00000494, -0.00000825, 0.00000606, -0.00001634, 0.00000144, 0.00000035, -0.00000953, 0.00000528, -0.00001884, 0.00001164, -0.00000475, 0.00000653, -0.00001020, 0.00000302, 0.00002289, -0.00000872, -0.00000871, 0.00000328, 0.00000263, -0.00001345, -0.00000423, 0.00000178, 0.00001082, -0.00001388, 0.00000273, -0.00000036, 0.00000461, -0.00001095, 0.00000376, -0.00000019, 0.00000456, -0.00000625, 0.00000823, -0.00000638, 0.00000314, -0.00000039, 0.00000162, -0.00000247, 0.00000027, 0.00000592, 0.00000021, -0.00000653, 0.00000393, 0.00000347, 0.00000669, -0.00000497, -0.00000264, -0.00000907, 0.00001253, -0.00001388, 0.00000801, -0.00001886, 0.00001610, -0.00003207, -0.00000802, 0.00000550, 0.00003988, 0.00001077, 0.00001317, 0.00003698, -0.00000280, },
		{ -0.00002321, 0.00000025, -0.00001848, 0.00000181, -0.00000366, 0.00000253, 0.00000202, -0.00000485, 0.00000057, -0.00000443, 0.00000005, -0.00000078, 0.00000434, -0.00000725, -0.00000202, -0.00000481, 0.00000477, -0.00000883, 0.00000032, -0.00000218, -0.00000059, -0.00000364, 0.00000337, -0.00000392, 0.00000523, -0.00001124, -0.00000341, -0.00001616, 0.00000750, -0.00002702, 0.00000581, -0.00006954, -0.00009577, 0.00016922, 0.00019935, 0.00007141, -0.00013969, 0.00000474, -0.00003857, 0.00000158, -0.00000988, 0.00000461, -0.00001036, 0.00000578, -0.00001523, 0.00000278, -0.00000941, 0.00000840, -0.00000433, 0.00001092, -0.00001134, 0.00000467, -0.00000645, 0.00001112, -0.00000033, 0.00000063, -0.00000314, 0.00000104, -0.00000900, 0.00001083, -0.00000071, 0.00000294, -0.00000679, 0.00000429, -0.00000349, -0.00000101, 0.00001488, -0.00001196, 0.00000742, -0.00001225, -0.00000622, 0.00000648, -0.00001244, 0.00000561, -0.00001451, 0.00000065, 0.00000542, -0.00000204, 0.00000174, -0.00001179, 0.00002684, -0.00000159, -0.00000990, -0.00000269, 0.00000837, -0.00001360, -0.00000445, -0.00000381, 0.00001423, -0.00001103, -0.00000158, 0.00000073, 0.00000475, -0.00000813, -0.00000047, 0.00000130, 0.00000401, -0.00000445, 0.00000461, -0.00000123, -0.00000168, 0.00000301, -0.00000106, 0.00000056, -0.00000269, 0.00000604, 0.00000279, -0.00000707, 0.00000188, 0.00000308, 0.00000774, -0.00000325, -0.00000103, -0.00001225, 0.00001109, -0.00000882, 0.00000335, -0.00001221, 0.00000435, -0.00001020, -0.00002913, 0.00000875, 0.00002466, 0.00003387, -0.00000414, 0.00004708, 0.00000126, 0.00000053, },
		{ -0.00000486, -0.00001272, -0.00000692, 0.00000150, -0.00000275, 0.00000599, -0.00000592, 0.00000067, -0.00000373, -0.00000116, -0.00000081, 0.00000366, -0.00000369, -0.00000503, -0.00000305, 0.00000124, -0.00000284, -0.00000630, 0.00000315, -0.00000491, 0.00000053, -0.00000207, 0.00000166, -0.00000077, -0.00000177, -0.00001076, -0.00000768, -0.00000559, -0.00000666, -0.00001910, -0.00001590, -0.00013746, 0.00011571, 0.00018421, 0.00014510, -0.00011071, -0.00004055, -0.00001285, -0.00002144, 0.00000231, -0.00000451, -0.00000296, -0.00000055, -0.00000698, -0.00000563, -0.00000338, 0.00000102, 0.00000145, 0.00000556, -0.00000269, -0.00000377, 0.00000006, 0.00000237, 0.00000839, -0.00000456, 0.00000126, -0.00000225, -0.00000478, 0.00000251, 0.00000718, -0.00000246, 0.00000005, -0.00000426, 0.00000612, -0.00001113, 0.00001956, -0.00001039, 0.00000467, -0.00000594, -0.00001215, 0.00000732, -0.00000851, -0.00000116, -0.00000401, -0.00001250, 0.00001340, -0.00000705, 0.00000823, -0.00001612, 0.00001892, 0.00001052, -0.00001062, -0.00000714, 0.00000917, -0.00000898, -0.00000751, -0.00000519, 0.00001091, -0.00000247, -0.00000879, 0.00000370, 0.00000213, -0.00000259, -0.00000593, 0.00000372, 0.00000148, -0.00000015, -0.00000106, 0.00000480, -0.00000629, 0.00000598, -0.00000350, 0.00000316, -0.00000472, 0.00000542, 0.00000426, -0.00000488, -0.00000241, 0.00000469, 0.00000560, 0.00000147, -0.00000275, -0.00000915, 0.00000193, 0.00000256, -0.00000701, 0.00000060, -0.00001215, 0.00001014, -0.00004033, 0.00000456, 0.00001240, 0.00004730, -0.00000607, 0.00004118, 0.00001596, -0.00000073, -0.00002052, },
		{ -0.00000444, -0.00001573, 0.00000509, -0.00000621, 0.00000739, -0.00000467, -0.00000073, -0.00000214, -0.00000262, -0.00000048, 0.00000197, 0.00000027, -0.00000748, -0.00000144, -0.00000247, 0.00000242, -0.00001077, 0.00000559, -0.00000654, 0.00000257, -0.00000534, 0.00000487, -0.00000474, 0.00000486, -0.00001500, -0.00000100, -0.00001663, 0.00000867, -0.00003505, 0.00001365, -0.00013532, 0.00004313, 0.00017249, 0.00017104, -0.00003001, -0.00009858, 0.00001085, -0.00003972, 0.00000964, -0.00001160, 0.00000413, -0.00000805, 0.00000266, -0.00001428, 0.00000287, -0.00000668, 0.00000745, -0.00000102, 0.00000676, -0.00001138, 0.00000566, -0.00000525, 0.00001452, -0.00000704, 0.00000391, -0.00000483, -0.00000036, -0.00000500, 0.00001257, -0.00000552, 0.00000491, -0.00000985, 0.00001098, -0.00001399, 0.00001610, -0.00000305, -0.00000128, 0.00000085, -0.00001506, 0.00000368, -0.00000132, -0.00000886, 0.00000425, -0.00001871, 0.00001174, -0.00000318, 0.00000550, -0.00000943, 0.00000348, 0.00002179, -0.00000831, -0.00000992, 0.00000546, -0.00000138, -0.00001175, -0.00000382, 0.00000379, 0.00000724, -0.00001429, 0.00000482, 0.00000013, 0.00000203, -0.00000925, 0.00000436, -0.00000035, 0.00000342, -0.00000508, 0.00000803, -0.00000757, 0.00000594, -0.00000335, 0.00000323, -0.00000420, 0.00000315, 0.00000544, -0.00000165, -0.00000647, 0.00000571, 0.00000323, 0.00000588, -0.00000454, -0.00000411, -0.00000808, 0.00001161, -0.00001351, 0.00000837, -0.00002010, 0.00001576, -0.00003327, -0.00001072, 0.00000776, 0.00004574, 0.00000844, 0.00002327, 0.00003496, -0.00000256, -0.00001454, -0.00001359, },
		{ -0.00001964, 0.00000365, -0.00000567, 0.00000514, -0.00000111, -0.00000306, -0.00000046, -0.00000358, -0.00000047, 0.00000043, 0.00000293, -0.00000772, -0.00000168, -0.00000412, 0.00000423, -0.00001052, 0.00000310, -0.00000384, 0.00000067, -0.00000432, 0.00000342, -0.00000398, 0.00000550, -0.00001275, -0.00000074, -0.00001881, 0.00001047, -0.00003252, 0.00001292, -0.00009639, -0.00003640, 0.00016273, 0.00016014, 0.00006561, -0.00013002, 0.00000708, -0.00003779, 0.00000241, -0.00000967, 0.00000450, -0.00001014, 0.00000660, -0.00001626, 0.00000310, -0.00000889, 0.00000824, -0.00000325, 0.00001093, -0.00001243, 0.00000530, -0.00000691, 0.00001382, -0.00000357, 0.00000208, -0.00000380, 0.00000069, -0.00000767, 0.00001191, -0.00000328, 0.00000473, -0.00000926, 0.00000841, -0.00000860, 0.00000584, 0.00000736, -0.00000770, 0.00000610, -0.00001476, -0.00000188, 0.00000437, -0.00001177, 0.00000480, -0.00001463, 0.00000069, 0.00000663, -0.00000194, 0.00000112, -0.00001013, 0.00002514, -0.00000074, -0.00001167, -0.00000034, 0.00000547, -0.00001380, -0.00000308, -0.00000243, 0.00001275, -0.00001386, 0.00000129, 0.00000117, 0.00000322, -0.00000840, 0.00000172, 0.00000024, 0.00000404, -0.00000517, 0.00000664, -0.00000436, 0.00000220, -0.00000007, 0.00000039, -0.00000108, -0.00000066, 0.00000676, 0.00000095, -0.00000791, 0.00000383, 0.00000318, 0.00000710, -0.00000336, -0.00000158, -0.00001271, 0.00001190, -0.00001034, 0.00000542, -0.00001460, 0.00000534, -0.00001271, -0.00002969, 0.00000615, 0.00003404, 0.00002999, 0.00000549, 0.00004744, 0.00000193, -0.00000988, -0.00002090, 0.00000081, },
		{ -0.00000319, -0.00000167, 0.00000030, 0.00000332, -0.00000507, 0.00000050, -0.00000366, -0.00000098, -0.00000049, 0.00000375, -0.00000563, -0.00000370, -0.00000357, 0.00000260, -0.00000629, -0.00000275, 0.00000127, -0.00000333, -0.00000074, -0.00000089, -0.00000016, 0.00000185, -0.00000591, -0.00000554, -0.00001304, 0.00000087, -0.00001627, -0.00000654, -0.00004407, -0.00009897, 0.00013926, 0.00014054, 0.00013396, -0.00010730, -0.00003140, -0.00001632, -0.00001674, -0.00000062, -0.00000159, -0.00000578, 0.00000305, -0.00001016, -0.00000348, -0.00000492, 0.00000313, -0.00000005, 0.00000843, -0.00000621, -0.00000103, -0.00000275, 0.00000701, 0.00000445, -0.00000292, 0.00000010, -0.00000160, -0.00000549, 0.00000601, 0.00000298, 0.00000064, -0.00000384, 0.00000077, 0.00000105, -0.00000582, 0.00001564, -0.00001080, 0.00000768, -0.00001131, -0.00000731, 0.00000609, -0.00000864, -0.00000129, -0.00000463, -0.00001223, 0.00001388, -0.00000585, 0.00000686, -0.00001431, 0.00001775, 0.00001079, -0.00001182, -0.00000596, 0.00000849, -0.00001144, -0.00000516, -0.00000478, 0.00001166, -0.00000698, -0.00000609, 0.00000443, 0.00000156, -0.00000432, -0.00000303, 0.00000251, 0.00000214, -0.00000202, 0.00000195, 0.00000140, -0.00000311, 0.00000427, -0.00000332, 0.00000259, -0.00000419, 0.00000735, 0.00000270, -0.00000643, -0.00000072, 0.00000521, 0.00000532, 0.00000068, -0.00000236, -0.00001084, 0.00000391, 0.00000005, -0.00000437, -0.00000190, -0.00001074, 0.00000717, -0.00003966, -0.00000054, 0.00002118, 0.00004612, -0.00000041, 0.00004592, 0.00001562, -0.00000840, -0.00002363, -0.00000031, -0.00001661, },
		{ 0.00000256, -0.00000425, 0.00000629, -0.00000533, 0.00000015, -0.00000288, -0.00000180, -0.00000093, 0.00000315, -0.00000234, -0.00000608, -0.00000230, -0.00000045, -0.00000108, -0.00000825, 0.00000495, -0.00000583, 0.00000180, -0.00000420, 0.00000279, -0.00000181, 0.00000070, -0.00001035, -0.00000546, -0.00001118, 0.00000088, -0.00002567, -0.00000419, -0.00012418, 0.00008992, 0.00013351, 0.00015429, -0.00003444, -0.00008455, 0.00000584, -0.00003361, 0.00000582, -0.00000761, 0.00000018, -0.00000364, -0.00000083, -0.00001180, 0.00000058, -0.00000363, 0.00000486, 0.00000293, 0.00000281, -0.00000852, 0.00000258, -0.00000098, 0.00001197, -0.00000668, 0.00000332, -0.00000411, -0.00000185, -0.00000096, 0.00000890, -0.00000328, 0.00000197, -0.00000655, 0.00000867, -0.00001260, 0.00001708, -0.00000791, 0.00000456, -0.00000519, -0.00001167, 0.00000439, -0.00000255, -0.00000838, 0.00000333, -0.00001837, 0.00001178, -0.00000164, 0.00000452, -0.00000864, 0.00000371, 0.00002089, -0.00000840, -0.00001041, 0.00000694, -0.00000528, -0.00000932, -0.00000380, 0.00000579, 0.00000268, -0.00001313, 0.00000647, -0.00000003, -0.00000014, -0.00000685, 0.00000394, 0.00000013, 0.00000153, -0.00000263, 0.00000604, -0.00000647, 0.00000647, -0.00000490, 0.00000402, -0.00000516, 0.00000581, 0.00000449, -0.00000361, -0.00000545, 0.00000672, 0.00000325, 0.00000474, -0.00000380, -0.00000611, -0.00000620, 0.00000962, -0.00001184, 0.00000709, -0.00001955, 0.00001378, -0.00003322, -0.00001510, 0.00001210, 0.00004986, 0.00000904, 0.00003201, 0.00003404, -0.00000706, -0.00002254, -0.00000682, -0.00000903, -0.00001207, },
		{ -0.00000576, 0.00000598, -0.00000314, -0.00000150, -0.00000155, -0.00000258, -0.00000115, 0.00000186, 0.00000078, -0.00000729, -0.00000188, -0.00000272, 0.00000222, -0.00001032, 0.00000453, -0.00000460, 0.00000107, -0.00000416, 0.00000242, -0.00000252, 0.00000351, -0.00001112, -0.00000161, -0.00001754, 0.00000839, -0.00003116, 0.00000971, -0.00010892, 0.00001984, 0.00013866, 0.00013693, 0.00005524, -0.00011677, 0.00000558, -0.00003421, 0.00000104, -0.00000748, 0.00000216, -0.00000757, 0.00000517, -0.00001547, 0.00000197, -0.00000691, 0.00000673, -0.00000059, 0.00000891, -0.00001158, 0.00000405, -0.00000508, 0.00001424, -0.00000543, 0.00000277, -0.00000389, -0.00000026, -0.00000490, 0.00001060, -0.00000349, 0.00000429, -0.00000916, 0.00000989, -0.00001132, 0.00001069, 0.00000023, -0.00000179, 0.00000189, -0.00001431, 0.00000101, 0.00000246, -0.00001122, 0.00000409, -0.00001460, 0.00000081, 0.00000768, -0.00000170, 0.00000045, -0.00000856, 0.00002347, -0.00000012, -0.00001322, 0.00000211, 0.00000177, -0.00001267, -0.00000245, -0.00000054, 0.00000982, -0.00001525, 0.00000413, 0.00000100, 0.00000150, -0.00000737, 0.00000260, 0.00000004, 0.00000297, -0.00000431, 0.00000687, -0.00000566, 0.00000475, -0.00000286, 0.00000200, -0.00000281, 0.00000190, 0.00000675, -0.00000110, -0.00000785, 0.00000533, 0.00000332, 0.00000616, -0.00000307, -0.00000285, -0.00001212, 0.00001159, -0.00001063, 0.00000610, -0.00001569, 0.00000532, -0.00001455, -0.00003134, 0.00000472, 0.00004271, 0.00002762, 0.00001543, 0.00004793, -0.00000045, -0.00002062, -0.00001491, -0.00000192, -0.00001804, 0.00000323, },
		{ 0.00000257, 0.00000068, -0.00000356, -0.00000033, -0.00000303, -0.00000132, 0.00000043, 0.00000284, -0.00000668, -0.00000286, -0.00000333, 0.00000258, -0.00000844, 0.00000030, -0.00000047, -0.00000206, -0.00000155, -0.00000049, -0.00000071, 0.00000259, -0.00000770, -0.00000259, -0.00001612, 0.00000443, -0.00002225, 0.00000142, -0.00006893, -0.00004958, 0.00013817, 0.00011072, 0.00012135, -0.00010124, -0.00002535, -0.00001779, -0.00001374, -0.00000182, -0.00000071, -0.00000655, 0.00000487, -0.00001200, -0.00000245, -0.00000519, 0.00000397, 0.00000000, 0.00000947, -0.00000818, 0.00000026, -0.00000391, 0.00001038, 0.00000080, -0.00000125, -0.00000086, -0.00000149, -0.00000474, 0.00000749, 0.00000027, 0.00000257, -0.00000650, 0.00000498, -0.00000387, 0.00000014, 0.00000941, -0.00000756, 0.00000701, -0.00001414, -0.00000297, 0.00000444, -0.00000866, -0.00000144, -0.00000511, -0.00001177, 0.00001414, -0.00000444, 0.00000535, -0.00001244, 0.00001640, 0.00001110, -0.00001327, -0.00000410, 0.00000656, -0.00001256, -0.00000336, -0.00000388, 0.00001115, -0.00001080, -0.00000268, 0.00000445, 0.00000051, -0.00000472, -0.00000108, 0.00000176, 0.00000203, -0.00000282, 0.00000404, -0.00000141, 0.00000024, 0.00000149, -0.00000198, 0.00000114, -0.00000260, 0.00000847, 0.00000086, -0.00000738, 0.00000106, 0.00000541, 0.00000494, 0.00000011, -0.00000244, -0.00001190, 0.00000538, -0.00000197, -0.00000217, -0.00000426, -0.00000910, 0.00000373, -0.00003889, -0.00000593, 0.00003071, 0.00004476, 0.00000716, 0.00004984, 0.00001410, -0.00001934, -0.00002125, 0.00000062, -0.00001804, -0.00000113, -0.00000369, },
		{ 0.00000409, -0.00000463, 0.00000005, -0.00000291, -0.00000154, -0.00000079, 0.00000356, -0.00000453, -0.00000466, -0.00000281, 0.00000090, -0.00000437, -0.00000506, 0.00000357, -0.00000473, 0.00000082, -0.00000316, 0.00000111, 0.00000051, -0.00000290, -0.00000568, -0.00001031, -0.00000532, -0.00000770, -0.00001447, -0.00002794, -0.00009451, 0.00011387, 0.00009993, 0.00014188, -0.00003877, -0.00007125, 0.00000066, -0.00002750, 0.00000235, -0.00000435, -0.00000314, 0.00000046, -0.00000431, -0.00000936, -0.00000141, -0.00000094, 0.00000287, 0.00000619, -0.00000088, -0.00000581, -0.00000028, 0.00000365, 0.00000829, -0.00000535, 0.00000231, -0.00000346, -0.00000256, 0.00000225, 0.00000517, -0.00000058, -0.00000153, -0.00000206, 0.00000443, -0.00000859, 0.00001436, -0.00000887, 0.00000763, -0.00001031, -0.00000733, 0.00000396, -0.00000344, -0.00000804, 0.00000259, -0.00001780, 0.00001161, 0.00000009, 0.00000342, -0.00000775, 0.00000371, 0.00002016, -0.00000893, -0.00001007, 0.00000735, -0.00000845, -0.00000678, -0.00000375, 0.00000726, -0.00000226, -0.00001055, 0.00000738, -0.00000076, -0.00000132, -0.00000469, 0.00000331, 0.00000049, -0.00000020, 0.00000010, 0.00000327, -0.00000390, 0.00000502, -0.00000485, 0.00000370, -0.00000504, 0.00000794, 0.00000312, -0.00000532, -0.00000386, 0.00000715, 0.00000338, 0.00000349, -0.00000307, -0.00000815, -0.00000398, 0.00000716, -0.00000954, 0.00000481, -0.00001769, 0.00001041, -0.00003210, -0.00002085, 0.00001808, 0.00005285, 0.00001223, 0.00003976, 0.00003317, -0.00001651, -0.00002522, -0.00000216, -0.00001285, -0.00000850, -0.00000026, -0.00000174, },
		{ -0.00000367, -0.00000075, -0.00000220, -0.00000186, -0.00000157, 0.00000304, -0.00000163, -0.00000628, -0.00000224, -0.00000121, -0.00000059, -0.00000863, 0.00000475, -0.00000461, 0.00000083, -0.00000351, 0.00000098, -0.00000047, 0.00000044, -0.00000755, -0.00000493, -0.00001348, 0.00000252, -0.00002418, -0.00000339, -0.00010353, 0.00006274, 0.00010838, 0.00012378, 0.00004253, -0.00010161, 0.00000157, -0.00002886, -0.00000153, -0.00000445, -0.00000118, -0.00000373, 0.00000237, -0.00001356, 0.00000005, -0.00000418, 0.00000461, 0.00000275, 0.00000574, -0.00000958, 0.00000167, -0.00000156, 0.00001259, -0.00000589, 0.00000276, -0.00000365, -0.00000119, -0.00000172, 0.00000782, -0.00000199, 0.00000206, -0.00000671, 0.00000861, -0.00001107, 0.00001232, -0.00000450, 0.00000356, -0.00000356, -0.00001156, 0.00000239, 0.00000090, -0.00001084, 0.00000354, -0.00001436, 0.00000081, 0.00000879, -0.00000147, -0.00000021, -0.00000712, 0.00002187, 0.00000014, -0.00001422, 0.00000412, -0.00000208, -0.00001067, -0.00000233, 0.00000156, 0.00000566, -0.00001494, 0.00000652, 0.00000016, 0.00000030, -0.00000587, 0.00000272, 0.00000026, 0.00000129, -0.00000228, 0.00000553, -0.00000500, 0.00000534, -0.00000446, 0.00000299, -0.00000403, 0.00000453, 0.00000611, -0.00000316, -0.00000697, 0.00000623, 0.00000364, 0.00000489, -0.00000240, -0.00000469, -0.00001060, 0.00001024, -0.00000971, 0.00000528, -0.00001515, 0.00000380, -0.00001524, -0.00003445, 0.00000472, 0.00005045, 0.00002735, 0.00002508, 0.00004884, -0.00000791, -0.00002805, -0.00000824, -0.00000615, -0.00001479, 0.00000120, -0.00000429, 0.00000509, },
		{ -0.00000225, -0.00000124, -0.00000223, -0.00000185, 0.00000164, 0.00000116, -0.00000686, -0.00000241, -0.00000255, 0.00000132, -0.00000900, 0.00000240, -0.00000168, -0.00000135, -0.00000179, -0.00000086, -0.00000009, 0.00000155, -0.00000692, -0.00000254, -0.00001602, 0.00000388, -0.00002265, 0.00000114, -0.00008290, -0.00000060, 0.00012069, 0.00009347, 0.00010601, -0.00009186, -0.00002291, -0.00001681, -0.00001261, -0.00000136, -0.00000174, -0.00000526, 0.00000473, -0.00001229, -0.00000270, -0.00000413, 0.00000353, 0.00000151, 0.00000867, -0.00000849, -0.00000008, -0.00000304, 0.00001181, -0.00000187, 0.00000002, -0.00000151, -0.00000174, -0.00000288, 0.00000694, -0.00000048, 0.00000265, -0.00000701, 0.00000710, -0.00000705, 0.00000478, 0.00000311, -0.00000245, 0.00000342, -0.00001416, 0.00000019, 0.00000277, -0.00000872, -0.00000153, -0.00000536, -0.00001133, 0.00001431, -0.00000290, 0.00000372, -0.00001052, 0.00001496, 0.00001131, -0.00001464, -0.00000198, 0.00000376, -0.00001239, -0.00000225, -0.00000242, 0.00000920, -0.00001336, 0.00000099, 0.00000358, -0.00000036, -0.00000423, -0.00000006, 0.00000160, 0.00000101, -0.00000226, 0.00000464, -0.00000274, 0.00000254, -0.00000120, -0.00000031, -0.00000062, -0.00000030, 0.00000881, -0.00000119, -0.00000750, 0.00000248, 0.00000565, 0.00000421, -0.00000003, -0.00000320, -0.00001200, 0.00000590, -0.00000299, -0.00000111, -0.00000557, -0.00000821, 0.00000068, -0.00003874, -0.00001095, 0.00004023, 0.00004439, 0.00001555, 0.00005407, 0.00000854, -0.00003005, -0.00001524, -0.00000168, -0.00001685, -0.00000118, -0.00000398, 0.00000327, -0.00000098, },
		{ -0.00000065, -0.00000242, -0.00000177, -0.00000009, 0.00000308, -0.00000602, -0.00000347, -0.00000283, 0.00000119, -0.00000671, -0.00000201, 0.00000205, -0.00000370, 0.00000007, -0.00000263, 0.00000036, 0.00000140, -0.00000467, -0.00000275, -0.00001347, -0.00000163, -0.00001363, -0.00000644, -0.00005010, -0.00005413, 0.00011686, 0.00007627, 0.00012917, -0.00004039, -0.00006071, -0.00000306, -0.00002261, 0.00000014, -0.00000269, -0.00000480, 0.00000315, -0.00000678, -0.00000779, -0.00000235, 0.00000071, 0.00000212, 0.00000799, -0.00000346, -0.00000416, -0.00000186, 0.00000744, 0.00000446, -0.00000367, 0.00000124, -0.00000309, -0.00000217, 0.00000382, 0.00000259, 0.00000126, -0.00000416, 0.00000202, -0.00000006, -0.00000366, 0.00000938, -0.00000645, 0.00000739, -0.00001333, -0.00000309, 0.00000287, -0.00000412, -0.00000779, 0.00000209, -0.00001716, 0.00001124, 0.00000205, 0.00000209, -0.00000671, 0.00000351, 0.00001950, -0.00000973, -0.00000907, 0.00000665, -0.00001049, -0.00000460, -0.00000335, 0.00000772, -0.00000690, -0.00000683, 0.00000722, -0.00000158, -0.00000154, -0.00000324, 0.00000302, 0.00000015, -0.00000099, 0.00000210, 0.00000090, -0.00000109, 0.00000248, -0.00000363, 0.00000247, -0.00000389, 0.00000938, 0.00000135, -0.00000639, -0.00000221, 0.00000736, 0.00000333, 0.00000248, -0.00000273, -0.00000970, -0.00000208, 0.00000498, -0.00000748, 0.00000251, -0.00001543, 0.00000635, -0.00003053, -0.00002738, 0.00002496, 0.00005582, 0.00001706, 0.00004765, 0.00002980, -0.00002878, -0.00002235, -0.00000105, -0.00001438, -0.00000640, -0.00000210, 0.00000010, 0.00000189, -0.00000332, },
		{ -0.00000222, -0.00000165, -0.00000151, 0.00000362, -0.00000386, -0.00000503, -0.00000247, -0.00000016, -0.00000340, -0.00000612, 0.00000418, -0.00000421, 0.00000036, -0.00000288, -0.00000019, 0.00000115, -0.00000223, -0.00000384, -0.00000877, -0.00000885, -0.00000436, -0.00001514, -0.00002219, -0.00008205, 0.00008742, 0.00008055, 0.00011482, 0.00003044, -0.00008684, -0.00000311, -0.00002315, -0.00000409, -0.00000184, -0.00000414, 0.00000001, -0.00000061, -0.00001154, -0.00000172, -0.00000160, 0.00000280, 0.00000575, 0.00000249, -0.00000747, -0.00000074, 0.00000259, 0.00000960, -0.00000528, 0.00000229, -0.00000339, -0.00000149, 0.00000080, 0.00000491, 0.00000003, -0.00000081, -0.00000299, 0.00000543, -0.00000846, 0.00001076, -0.00000592, 0.00000672, -0.00000860, -0.00000752, 0.00000251, -0.00000030, -0.00001066, 0.00000323, -0.00001395, 0.00000057, 0.00001014, -0.00000141, -0.00000077, -0.00000583, 0.00002033, 0.00000006, -0.00001458, 0.00000532, -0.00000546, -0.00000835, -0.00000236, 0.00000343, 0.00000075, -0.00001290, 0.00000797, -0.00000108, -0.00000001, -0.00000466, 0.00000278, 0.00000024, -0.00000018, 0.00000001, 0.00000348, -0.00000305, 0.00000415, -0.00000463, 0.00000305, -0.00000439, 0.00000689, 0.00000485, -0.00000487, -0.00000563, 0.00000671, 0.00000397, 0.00000351, -0.00000169, -0.00000666, -0.00000866, 0.00000838, -0.00000817, 0.00000351, -0.00001335, 0.00000090, -0.00001484, -0.00003886, 0.00000586, 0.00005779, 0.00002891, 0.00003500, 0.00004875, -0.00002037, -0.00002979, -0.00000371, -0.00000967, -0.00001155, -0.00000116, -0.00000230, 0.00000322, -0.00000299, -0.00000087, },
		{ -0.00000166, -0.00000222, 0.00000275, -0.00000094, -0.00000638, -0.00000217, -0.00000164, -0.00000071, -0.00000820, 0.00000350, -0.00000237, -0.00000108, -0.00000170, -0.00000155, 0.00000101, -0.00000037, -0.00000452, -0.00000459, -0.00001351, 0.00000002, -0.00001816, -0.00000753, -0.00008224, 0.00003883, 0.00009614, 0.00008464, 0.00008921, -0.00008034, -0.00002315, -0.00001407, -0.00001267, -0.00000010, -0.00000372, -0.00000279, 0.00000332, -0.00001153, -0.00000370, -0.00000228, 0.00000241, 0.00000379, 0.00000669, -0.00000769, -0.00000148, -0.00000052, 0.00001131, -0.00000336, 0.00000079, -0.00000197, -0.00000187, -0.00000073, 0.00000517, 0.00000025, 0.00000118, -0.00000544, 0.00000685, -0.00000786, 0.00000689, -0.00000141, 0.00000254, -0.00000171, -0.00001180, 0.00000194, 0.00000133, -0.00000892, -0.00000143, -0.00000538, -0.00001109, 0.00001460, -0.00000137, 0.00000201, -0.00000859, 0.00001346, 0.00001129, -0.00001565, -0.00000009, 0.00000064, -0.00001127, -0.00000171, -0.00000059, 0.00000588, -0.00001422, 0.00000433, 0.00000188, -0.00000047, -0.00000364, 0.00000054, 0.00000160, -0.00000043, -0.00000071, 0.00000393, -0.00000245, 0.00000318, -0.00000293, 0.00000099, -0.00000214, 0.00000233, 0.00000834, -0.00000317, -0.00000691, 0.00000344, 0.00000601, 0.00000314, 0.00000022, -0.00000453, -0.00001123, 0.00000553, -0.00000300, -0.00000132, -0.00000549, -0.00000858, -0.00000150, -0.00003958, -0.00001536, 0.00004963, 0.00004545, 0.00002455, 0.00005848, -0.00000293, -0.00003674, -0.00000869, -0.00000544, -0.00001406, -0.00000273, -0.00000288, 0.00000246, -0.00000132, -0.00000166, -0.00000175, },
		{ -0.00000223, 0.00000095, 0.00000182, -0.00000675, -0.00000252, -0.00000249, 0.00000047, -0.00000779, 0.00000046, 0.00000070, -0.00000295, -0.00000036, -0.00000261, 0.00000047, 0.00000093, -0.00000437, -0.00000215, -0.00001408, -0.00000126, -0.00001505, -0.00000493, -0.00006404, -0.00001241, 0.00010500, 0.00006245, 0.00011467, -0.00003863, -0.00005348, -0.00000480, -0.00001917, -0.00000081, -0.00000256, -0.00000477, 0.00000421, -0.00000798, -0.00000728, -0.00000212, 0.00000124, 0.00000263, 0.00000824, -0.00000470, -0.00000382, -0.00000172, 0.00000967, 0.00000119, -0.00000203, 0.00000022, -0.00000283, -0.00000098, 0.00000377, 0.00000160, 0.00000161, -0.00000504, 0.00000452, -0.00000340, 0.00000047, 0.00000402, -0.00000208, 0.00000431, -0.00001378, 0.00000028, 0.00000154, -0.00000478, -0.00000751, 0.00000190, -0.00001662, 0.00001083, 0.00000418, 0.00000054, -0.00000552, 0.00000314, 0.00001877, -0.00001052, -0.00000776, 0.00000511, -0.00001137, -0.00000300, -0.00000243, 0.00000686, -0.00001058, -0.00000248, 0.00000579, -0.00000180, -0.00000133, -0.00000237, 0.00000309, -0.00000093, -0.00000061, 0.00000289, -0.00000028, 0.00000087, -0.00000007, -0.00000194, 0.00000077, -0.00000191, 0.00001003, -0.00000065, -0.00000669, -0.00000088, 0.00000763, 0.00000288, 0.00000190, -0.00000301, -0.00001042, -0.00000090, 0.00000355, -0.00000631, 0.00000100, -0.00001365, 0.00000236, -0.00002909, -0.00003416, 0.00003217, 0.00005968, 0.00002281, 0.00005656, 0.00002096, -0.00004005, -0.00001592, -0.00000312, -0.00001341, -0.00000626, -0.00000242, 0.00000057, 0.00000058, -0.00000221, -0.00000142, -0.00000175, },
		{ -0.00000095, 0.00000346, -0.00000563, -0.00000372, -0.00000248, 0.00000015, -0.00000560, -0.00000342, 0.00000322, -0.00000372, -0.00000006, -0.00000258, -0.00000070, 0.00000172, -0.00000347, -0.00000147, -0.00001136, -0.00000581, -0.00000933, -0.00000822, -0.00004053, -0.00005048, 0.00009443, 0.00005973, 0.00010609, 0.00002105, -0.00007416, -0.00000704, -0.00001807, -0.00000591, -0.00000038, -0.00000587, 0.00000271, -0.00000289, -0.00001014, -0.00000269, 0.00000020, 0.00000191, 0.00000769, -0.00000005, -0.00000606, -0.00000220, 0.00000630, 0.00000609, -0.00000404, 0.00000155, -0.00000320, -0.00000098, 0.00000199, 0.00000290, 0.00000142, -0.00000306, 0.00000061, 0.00000172, -0.00000483, 0.00000710, -0.00000430, 0.00000698, -0.00001200, -0.00000320, 0.00000175, -0.00000125, -0.00001061, 0.00000328, -0.00001356, 0.00000017, 0.00001176, -0.00000162, -0.00000117, -0.00000469, 0.00001879, -0.00000020, -0.00001439, 0.00000562, -0.00000799, -0.00000617, -0.00000220, 0.00000456, -0.00000429, -0.00000932, 0.00000802, -0.00000208, 0.00000038, -0.00000406, 0.00000318, -0.00000050, -0.00000076, 0.00000168, 0.00000176, -0.00000090, 0.00000197, -0.00000368, 0.00000222, -0.00000382, 0.00000875, 0.00000303, -0.00000589, -0.00000431, 0.00000709, 0.00000402, 0.00000238, -0.00000131, -0.00000824, -0.00000689, 0.00000666, -0.00000677, 0.00000165, -0.00001109, -0.00000276, -0.00001383, -0.00004412, 0.00000751, 0.00006568, 0.00003150, 0.00004638, 0.00004472, -0.00003536, -0.00002569, -0.00000281, -0.00001096, -0.00000972, -0.00000263, -0.00000094, 0.00000142, -0.00000182, -0.00000150, -0.00000177, -0.00000187, },
		{ 0.00000347, -0.00000311, -0.00000543, -0.00000198, -0.00000102, -0.00000290, -0.00000650, 0.00000382, -0.00000271, -0.00000100, -0.00000162, -0.00000204, 0.00000176, -0.00000196, -0.00000196, -0.00000716, -0.00001040, -0.00000487, -0.00001160, -0.00002142, -0.00006783, 0.00006384, 0.00007205, 0.00007964, 0.00007315, -0.00006851, -0.00002460, -0.00001063, -0.00001298, 0.00000095, -0.00000551, -0.00000025, 0.00000162, -0.00001057, -0.00000468, -0.00000040, 0.00000136, 0.00000599, 0.00000439, -0.00000663, -0.00000305, 0.00000280, 0.00000936, -0.00000378, 0.00000114, -0.00000239, -0.00000145, 0.00000078, 0.00000330, 0.00000142, -0.00000090, -0.00000266, 0.00000486, -0.00000666, 0.00000639, -0.00000322, 0.00000587, -0.00000687, -0.00000787, 0.00000234, 0.00000022, -0.00000936, -0.00000101, -0.00000529, -0.00001108, 0.00001515, -0.00000000, 0.00000033, -0.00000669, 0.00001189, 0.00001107, -0.00001619, 0.00000124, -0.00000226, -0.00000968, -0.00000146, 0.00000123, 0.00000150, -0.00001317, 0.00000667, -0.00000021, 0.00000040, -0.00000358, 0.00000135, 0.00000112, -0.00000151, 0.00000100, 0.00000269, -0.00000114, 0.00000230, -0.00000343, 0.00000153, -0.00000308, 0.00000499, 0.00000705, -0.00000467, -0.00000603, 0.00000415, 0.00000626, 0.00000199, 0.00000053, -0.00000599, -0.00001005, 0.00000472, -0.00000252, -0.00000230, -0.00000434, -0.00001011, -0.00000283, -0.00004128, -0.00001951, 0.00005950, 0.00004757, 0.00003502, 0.00006112, -0.00001997, -0.00003695, -0.00000457, -0.00000840, -0.00001136, -0.00000452, -0.00000141, 0.00000087, -0.00000064, -0.00000179, -0.00000175, -0.00000148, -0.00000224, },
		{ 0.00000001, -0.00000677, -0.00000174, -0.00000208, -0.00000091, -0.00000771, 0.00000220, -0.00000038, -0.00000242, -0.00000063, -0.00000276, 0.00000086, -0.00000015, -0.00000280, -0.00000328, -0.00001270, -0.00000366, -0.00001234, -0.00001040, -0.00006616, 0.00002264, 0.00008581, 0.00005563, 0.00009909, -0.00003435, -0.00004892, -0.00000500, -0.00001670, -0.00000117, -0.00000318, -0.00000374, 0.00000420, -0.00000831, -0.00000741, -0.00000115, 0.00000109, 0.00000390, 0.00000738, -0.00000500, -0.00000439, -0.00000008, 0.00001020, -0.00000120, -0.00000060, -0.00000083, -0.00000227, 0.00000026, 0.00000280, 0.00000177, 0.00000068, -0.00000416, 0.00000508, -0.00000495, 0.00000274, -0.00000008, 0.00000252, -0.00000051, -0.00001184, 0.00000236, 0.00000032, -0.00000557, -0.00000703, 0.00000195, -0.00001634, 0.00001057, 0.00000634, -0.00000117, -0.00000420, 0.00000263, 0.00001791, -0.00001106, -0.00000655, 0.00000319, -0.00001133, -0.00000194, -0.00000109, 0.00000459, -0.00001271, 0.00000182, 0.00000322, -0.00000087, -0.00000147, -0.00000151, 0.00000301, -0.00000221, 0.00000055, 0.00000266, -0.00000018, 0.00000144, -0.00000189, -0.00000041, -0.00000096, 0.00000065, 0.00000976, -0.00000251, -0.00000640, 0.00000013, 0.00000797, 0.00000208, 0.00000171, -0.00000379, -0.00001041, -0.00000042, 0.00000286, -0.00000611, 0.00000057, -0.00001281, -0.00000113, -0.00002804, -0.00004115, 0.00003977, 0.00006466, 0.00002966, 0.00006597, 0.00000478, -0.00004625, -0.00000924, -0.00000649, -0.00001099, -0.00000745, -0.00000155, -0.00000026, 0.00000043, -0.00000172, -0.00000185, -0.00000124, -0.00000264, 0.00000204, },
		{ -0.00000680, -0.00000243, -0.00000237, -0.00000029, -0.00000690, -0.00000095, 0.00000213, -0.00000326, -0.00000036, -0.00000254, -0.00000067, 0.00000131, -0.00000312, -0.00000096, -0.00001193, -0.00000545, -0.00001068, -0.00000643, -0.00005269, -0.00001659, 0.00008829, 0.00004653, 0.00009611, 0.00001496, -0.00006414, -0.00000972, -0.00001386, -0.00000698, 0.00000001, -0.00000633, 0.00000418, -0.00000422, -0.00000955, -0.00000270, 0.00000110, 0.00000203, 0.00000839, -0.00000164, -0.00000561, -0.00000228, 0.00000886, 0.00000271, -0.00000250, 0.00000058, -0.00000284, -0.00000001, 0.00000193, 0.00000214, 0.00000161, -0.00000390, 0.00000305, -0.00000130, -0.00000160, 0.00000293, -0.00000077, 0.00000456, -0.00001311, 0.00000057, 0.00000056, -0.00000211, -0.00001058, 0.00000372, -0.00001337, -0.00000024, 0.00001356, -0.00000210, -0.00000140, -0.00000368, 0.00001716, -0.00000044, -0.00001393, 0.00000521, -0.00000956, -0.00000441, -0.00000164, 0.00000457, -0.00000875, -0.00000466, 0.00000638, -0.00000204, 0.00000078, -0.00000377, 0.00000379, -0.00000189, -0.00000028, 0.00000234, 0.00000103, 0.00000049, -0.00000028, -0.00000218, 0.00000082, -0.00000238, 0.00000988, 0.00000090, -0.00000611, -0.00000331, 0.00000761, 0.00000358, 0.00000172, -0.00000150, -0.00000912, -0.00000567, 0.00000551, -0.00000608, 0.00000041, -0.00000914, -0.00000656, -0.00001265, -0.00004991, 0.00000928, 0.00007487, 0.00003468, 0.00006000, 0.00003341, -0.00004855, -0.00001797, -0.00000514, -0.00000978, -0.00000974, -0.00000269, -0.00000086, 0.00000050, -0.00000090, -0.00000214, -0.00000118, -0.00000232, -0.00000004, 0.00000258, },
		{ -0.00000412, -0.00000183, -0.00000086, -0.00000479, -0.00000436, 0.00000359, -0.00000286, -0.00000097, -0.00000172, -0.00000207, 0.00000173, -0.00000238, -0.00000050, -0.00000874, -0.00000855, -0.00000831, -0.00000652, -0.00003546, -0.00004432, 0.00007422, 0.00005268, 0.00007524, 0.00005943, -0.00005782, -0.00002610, -0.00000729, -0.00001299, 0.00000124, -0.00000640, 0.00000153, 0.00000038, -0.00001004, -0.00000504, 0.00000091, 0.00000095, 0.00000747, 0.00000245, -0.00000599, -0.00000394, 0.00000599, 0.00000659, -0.00000340, 0.00000111, -0.00000274, -0.00000042, 0.00000118, 0.00000222, 0.00000206, -0.00000249, 0.00000017, 0.00000223, -0.00000451, 0.00000409, -0.00000240, 0.00000674, -0.00001083, -0.00000330, 0.00000171, -0.00000058, -0.00001001, -0.00000010, -0.00000533, -0.00001121, 0.00001596, 0.00000111, -0.00000128, -0.00000482, 0.00001022, 0.00001077, -0.00001633, 0.00000191, -0.00000455, -0.00000805, -0.00000117, 0.00000260, -0.00000340, -0.00001022, 0.00000734, -0.00000183, 0.00000185, -0.00000418, 0.00000266, -0.00000022, -0.00000163, 0.00000208, 0.00000183, 0.00000021, 0.00000061, -0.00000295, 0.00000133, -0.00000327, 0.00000742, 0.00000498, -0.00000534, -0.00000532, 0.00000498, 0.00000606, 0.00000117, 0.00000050, -0.00000709, -0.00000901, 0.00000406, -0.00000225, -0.00000328, -0.00000286, -0.00001227, -0.00000369, -0.00004345, -0.00002398, 0.00007076, 0.00004996, 0.00004843, 0.00005842, -0.00003958, -0.00003065, -0.00000444, -0.00000896, -0.00001016, -0.00000537, -0.00000066, -0.00000047, 0.00000033, -0.00000229, -0.00000138, -0.00000165, -0.00000188, 0.00000365, -0.00000511, },
		{ -0.00000115, -0.00000184, -0.00000255, -0.00000672, 0.00000324, -0.00000123, -0.00000200, -0.00000095, -0.00000269, 0.00000091, -0.00000083, -0.00000122, -0.00000478, -0.00001087, -0.00000694, -0.00000780, -0.00002038, -0.00005677, 0.00004645, 0.00006569, 0.00005232, 0.00008404, -0.00002907, -0.00004589, -0.00000444, -0.00001452, -0.00000172, -0.00000360, -0.00000268, 0.00000393, -0.00000847, -0.00000753, -0.00000011, 0.00000091, 0.00000522, 0.00000611, -0.00000503, -0.00000513, 0.00000241, 0.00000930, -0.00000266, 0.00000055, -0.00000194, -0.00000110, 0.00000077, 0.00000191, 0.00000222, -0.00000072, -0.00000223, 0.00000417, -0.00000493, 0.00000298, -0.00000202, 0.00000593, -0.00000573, -0.00000811, 0.00000308, -0.00000061, -0.00000664, -0.00000611, 0.00000205, -0.00001629, 0.00001055, 0.00000837, -0.00000297, -0.00000279, 0.00000197, 0.00001693, -0.00001126, -0.00000571, 0.00000134, -0.00001076, -0.00000122, 0.00000042, 0.00000107, -0.00001285, 0.00000519, 0.00000015, 0.00000129, -0.00000254, -0.00000004, 0.00000215, -0.00000293, 0.00000169, 0.00000216, 0.00000061, 0.00000079, -0.00000271, 0.00000062, -0.00000241, 0.00000356, 0.00000841, -0.00000375, -0.00000600, 0.00000110, 0.00000805, 0.00000128, 0.00000153, -0.00000464, -0.00001009, -0.00000019, 0.00000242, -0.00000640, 0.00000090, -0.00001277, -0.00000422, -0.00002719, -0.00004881, 0.00004852, 0.00007015, 0.00003897, 0.00007331, -0.00001816, -0.00004471, -0.00000554, -0.00000880, -0.00000888, -0.00000869, -0.00000054, -0.00000158, 0.00000106, -0.00000181, -0.00000189, -0.00000100, -0.00000285, 0.00000296, -0.00000211, -0.00000610, },
		},
		{ { 0.00000476, 0.00000319, -0.00000988, -0.00000039, -0.00000139, -0.00000013, -0.00000022, -0.00000310, 0.00000443, -0.00000237, -0.00000117, -0.00000231, 0.00000587, 0.00000017, -0.00000643, -0.00000449, -0.00000654, -0.00000841, 0.00000108, -0.00001326, -0.00006090, 0.00001164, 0.00008027, 0.00003891, 0.00008439, 0.00001224, -0.00006519, -0.00000818, -0.00001583, -0.00000122, 0.00000864, -0.00000180, -0.00000558, -0.00000714, -0.00000881, -0.00000199, 0.00000232, -0.00000127, -0.00000103, -0.00000163, -0.00000235, 0.00000100, 0.00000256, 0.00000094, 0.00000048, -0.00000050, -0.00000384, 0.00000187, 0.00000150, -0.00000101, 0.00001654, -0.00000336, -0.00000697, -0.00000271, 0.00000191, 0.00000002, -0.00001241, 0.00000351, 0.00000134, -0.00000682, 0.00001237, 0.00000022, -0.00001443, 0.00000106, -0.00000082, 0.00000106, -0.00001443, 0.00000022, 0.00001237, -0.00000682, 0.00000134, 0.00000351, -0.00001241, 0.00000002, 0.00000191, -0.00000271, -0.00000697, -0.00000336, 0.00001654, -0.00000101, 0.00000150, 0.00000187, -0.00000384, -0.00000050, 0.00000048, 0.00000094, 0.00000256, 0.00000100, -0.00000235, -0.00000163, -0.00000103, -0.00000127, 0.00000232, -0.00000199, -0.00000881, -0.00000714, -0.00000558, -0.00000180, 0.00000864, -0.00000122, -0.00001583, -0.00000818, -0.00006519, 0.00001224, 0.00008439, 0.00003891, 0.00008027, 0.00001164, -0.00006090, -0.00001326, 0.00000108, -0.00000841, -0.00000654, -0.00000449, -0.00000643, 0.00000017, 0.00000587, -0.00000231, -0.00000117, -0.00000237, 0.00000443, -0.00000310, -0.00000022, -0.00000013, -0.00000139, -0.00000039, -0.00000988, 0.00000319, },
		{ 0.00000500, -0.00000555, -0.00000655, 0.00000224, -0.00000304, 0.00000248, -0.00000486, 0.00000342, 0.00000002, -0.00000175, -0.00000264, 0.00000332, 0.00000389, -0.00000577, -0.00000502, -0.00000501, -0.00000853, -0.00000295, -0.00000106, -0.00005001, -0.00002550, 0.00007292, 0.00004641, 0.00006173, 0.00005923, -0.00005554, -0.00002831, -0.00000817, -0.00001150, 0.00001020, 0.00000095, -0.00000429, -0.00000693, -0.00000827, -0.00000557, 0.00000264, -0.00000052, -0.00000114, -0.00000132, -0.00000214, -0.00000090, 0.00000290, 0.00000124, 0.00000091, -0.00000020, -0.00000241, -0.00000178, 0.00000462, -0.00000395, 0.00001214, 0.00000701, -0.00001027, -0.00000159, -0.00000277, 0.00000509, -0.00001077, -0.00000348, 0.00000596, -0.00000785, 0.00000626, 0.00000925, -0.00001513, -0.00000289, 0.00000059, 0.00000083, -0.00000883, -0.00000934, 0.00001417, -0.00000203, -0.00000357, 0.00000637, -0.00000784, -0.00000755, 0.00000470, -0.00000203, -0.00000431, -0.00000891, 0.00001234, 0.00000736, -0.00000308, 0.00000486, -0.00000371, -0.00000147, 0.00000013, 0.00000084, 0.00000168, 0.00000249, -0.00000165, -0.00000209, -0.00000117, -0.00000130, 0.00000089, 0.00000113, -0.00000789, -0.00000774, -0.00000643, -0.00000343, 0.00000425, 0.00000754, -0.00001728, -0.00000360, -0.00004649, -0.00004102, 0.00008441, 0.00005140, 0.00006858, 0.00005321, -0.00005712, -0.00003145, 0.00000219, -0.00000622, -0.00000817, -0.00000484, -0.00000582, -0.00000360, 0.00000635, -0.00000027, -0.00000166, -0.00000244, 0.00000192, 0.00000133, -0.00000397, 0.00000279, -0.00000393, 0.00000221, -0.00000856, -0.00000208, 0.00000521, },
		{ 0.00000062, -0.00001061, 0.00000185, -0.00000213, 0.00000135, -0.00000253, -0.00000068, 0.00000327, -0.00000232, -0.00000206, 0.00000011, 0.00000621, -0.00000367, -0.00000562, -0.00000457, -0.00000668, -0.00000738, 0.00000403, -0.00003224, -0.00004828, 0.00004685, 0.00006245, 0.00003949, 0.00007872, -0.00001773, -0.00005080, -0.00000566, -0.00001570, 0.00000544, 0.00000557, -0.00000331, -0.00000611, -0.00000776, -0.00000771, 0.00000071, 0.00000123, -0.00000161, -0.00000092, -0.00000185, -0.00000193, 0.00000179, 0.00000227, 0.00000081, 0.00000044, -0.00000145, -0.00000316, 0.00000350, -0.00000085, 0.00000300, 0.00001564, -0.00000892, -0.00000308, -0.00000390, 0.00000334, -0.00000292, -0.00001127, 0.00000726, -0.00000488, -0.00000109, 0.00001382, -0.00000949, -0.00000990, 0.00000298, -0.00000104, -0.00000271, -0.00001468, 0.00000979, 0.00000526, -0.00000685, 0.00000537, -0.00000133, -0.00001153, 0.00000170, 0.00000184, -0.00000376, -0.00000889, 0.00000293, 0.00001478, -0.00000349, 0.00000370, -0.00000050, -0.00000360, 0.00000028, 0.00000033, 0.00000128, 0.00000259, 0.00000018, -0.00000266, -0.00000129, -0.00000133, -0.00000022, 0.00000226, -0.00000491, -0.00000896, -0.00000648, -0.00000502, 0.00000055, 0.00000993, -0.00000942, -0.00001154, -0.00001853, -0.00007186, 0.00004839, 0.00007915, 0.00005070, 0.00008383, -0.00003284, -0.00005206, -0.00000153, -0.00000324, -0.00000849, -0.00000645, -0.00000465, -0.00000621, 0.00000435, 0.00000292, -0.00000259, -0.00000126, -0.00000150, 0.00000445, -0.00000467, 0.00000203, -0.00000223, -0.00000029, -0.00000295, -0.00000786, 0.00000488, 0.00000473, },
		{ -0.00000868, -0.00000374, 0.00000177, -0.00000181, 0.00000108, -0.00000430, 0.00000439, -0.00000127, -0.00000175, -0.00000222, 0.00000624, -0.00000046, -0.00000571, -0.00000508, -0.00000440, -0.00000968, 0.00000273, -0.00001435, -0.00005434, 0.00001186, 0.00007139, 0.00003299, 0.00006968, 0.00002923, -0.00005661, -0.00001628, -0.00001210, -0.00000383, 0.00000954, -0.00000159, -0.00000501, -0.00000735, -0.00000834, -0.00000256, 0.00000264, -0.00000149, -0.00000088, -0.00000153, -0.00000205, -0.00000009, 0.00000307, 0.00000085, 0.00000105, -0.00000093, -0.00000269, 0.00000025, 0.00000333, -0.00000320, 0.00001584, -0.00000009, -0.00000827, -0.00000089, -0.00000294, 0.00000525, -0.00001395, 0.00000274, 0.00000103, -0.00000687, 0.00001269, 0.00000004, -0.00001567, 0.00000256, -0.00000095, -0.00000005, -0.00001345, 0.00000038, 0.00001221, -0.00000713, 0.00000219, 0.00000328, -0.00000970, -0.00000523, 0.00000527, -0.00000260, -0.00000700, -0.00000512, 0.00001523, 0.00000255, -0.00000098, 0.00000381, -0.00000496, -0.00000023, 0.00000004, 0.00000100, 0.00000201, 0.00000191, -0.00000235, -0.00000182, -0.00000127, -0.00000089, 0.00000191, -0.00000144, -0.00000917, -0.00000700, -0.00000587, -0.00000220, 0.00000758, 0.00000128, -0.00001856, -0.00000306, -0.00006619, -0.00001241, 0.00009891, 0.00004655, 0.00009083, 0.00001064, -0.00006782, -0.00001165, -0.00000101, -0.00000700, -0.00000865, -0.00000400, -0.00000704, 0.00000064, 0.00000560, -0.00000241, -0.00000076, -0.00000299, 0.00000373, -0.00000130, -0.00000184, 0.00000169, -0.00000402, 0.00000148, -0.00000940, 0.00000122, 0.00000502, 0.00000417, },
		{ -0.00000992, 0.00000357, -0.00000251, 0.00000192, -0.00000416, 0.00000179, 0.00000158, -0.00000198, -0.00000300, 0.00000401, 0.00000317, -0.00000514, -0.00000557, -0.00000345, -0.00000863, -0.00000260, -0.00000102, -0.00004700, -0.00002008, 0.00006394, 0.00004258, 0.00004724, 0.00006250, -0.00003681, -0.00003522, -0.00000791, -0.00001097, 0.00000856, 0.00000251, -0.00000463, -0.00000624, -0.00000855, -0.00000526, 0.00000207, 0.00000001, -0.00000160, -0.00000088, -0.00000206, -0.00000130, 0.00000237, 0.00000190, 0.00000081, 0.00000011, -0.00000227, -0.00000173, 0.00000403, -0.00000259, 0.00000787, 0.00001166, -0.00001196, 0.00000098, -0.00000668, 0.00000655, -0.00000807, -0.00000646, 0.00000667, -0.00000878, 0.00000703, 0.00000902, -0.00001599, -0.00000275, 0.00000213, -0.00000121, -0.00000762, -0.00000957, 0.00001488, -0.00000333, -0.00000194, 0.00000512, -0.00000463, -0.00001060, 0.00000383, 0.00000131, -0.00000592, -0.00000843, 0.00000864, 0.00001103, -0.00000406, 0.00000498, -0.00000297, -0.00000259, 0.00000069, 0.00000013, 0.00000182, 0.00000228, -0.00000054, -0.00000290, -0.00000090, -0.00000162, 0.00000137, 0.00000058, -0.00000700, -0.00000880, -0.00000545, -0.00000480, 0.00000420, 0.00000755, -0.00001569, -0.00000611, -0.00003623, -0.00006565, 0.00008390, 0.00006691, 0.00007532, 0.00006073, -0.00006779, -0.00002940, -0.00000040, -0.00000477, -0.00000982, -0.00000493, -0.00000622, -0.00000335, 0.00000622, -0.00000032, -0.00000167, -0.00000209, 0.00000023, 0.00000315, -0.00000481, 0.00000337, -0.00000432, 0.00000091, -0.00000550, -0.00000483, 0.00000563, 0.00000462, -0.00000261, },
		{ -0.00000043, 0.00000088, -0.00000060, -0.00000086, -0.00000253, 0.00000406, -0.00000170, -0.00000273, 0.00000062, 0.00000604, -0.00000380, -0.00000539, -0.00000434, -0.00000523, -0.00000832, 0.00000524, -0.00003209, -0.00004066, 0.00004090, 0.00005750, 0.00002975, 0.00007037, 0.00000081, -0.00004806, -0.00001187, -0.00001173, 0.00000206, 0.00000757, -0.00000400, -0.00000495, -0.00000839, -0.00000693, -0.00000023, 0.00000194, -0.00000214, -0.00000047, -0.00000197, -0.00000160, 0.00000050, 0.00000322, 0.00000037, 0.00000127, -0.00000208, -0.00000196, 0.00000167, 0.00000166, -0.00000091, 0.00001748, -0.00000759, -0.00000313, -0.00000364, -0.00000024, 0.00000270, -0.00001427, 0.00000741, -0.00000561, -0.00000064, 0.00001381, -0.00000986, -0.00001099, 0.00000536, -0.00000287, -0.00000212, -0.00001466, 0.00001053, 0.00000420, -0.00000597, 0.00000507, -0.00000011, -0.00001068, -0.00000265, 0.00000586, -0.00000453, -0.00000849, 0.00000006, 0.00001536, -0.00000116, 0.00000144, 0.00000182, -0.00000540, 0.00000112, -0.00000057, 0.00000169, 0.00000179, 0.00000156, -0.00000330, -0.00000102, -0.00000196, 0.00000067, 0.00000128, -0.00000368, -0.00001014, -0.00000545, -0.00000619, 0.00000070, 0.00000851, -0.00000597, -0.00001599, -0.00000920, -0.00008185, 0.00002807, 0.00009885, 0.00005618, 0.00009655, -0.00004190, -0.00005350, -0.00000212, -0.00000359, -0.00000853, -0.00000767, -0.00000455, -0.00000639, 0.00000447, 0.00000282, -0.00000276, -0.00000067, -0.00000278, 0.00000479, -0.00000368, 0.00000087, -0.00000070, -0.00000305, -0.00000014, -0.00000881, 0.00000381, 0.00000465, 0.00000256, -0.00001100, },
		{ 0.00000452, -0.00000261, 0.00000157, -0.00000461, 0.00000351, 0.00000018, -0.00000230, -0.00000228, 0.00000681, -0.00000129, -0.00000485, -0.00000579, -0.00000222, -0.00001085, 0.00000412, -0.00001534, -0.00004753, 0.00001074, 0.00006461, 0.00002761, 0.00005662, 0.00003864, -0.00004302, -0.00002446, -0.00000956, -0.00000499, 0.00000918, -0.00000068, -0.00000506, -0.00000683, -0.00000848, -0.00000248, 0.00000229, -0.00000108, -0.00000121, -0.00000117, -0.00000191, -0.00000083, 0.00000301, 0.00000122, 0.00000118, -0.00000088, -0.00000226, -0.00000034, 0.00000386, -0.00000350, 0.00001294, 0.00000457, -0.00001068, 0.00000224, -0.00000816, 0.00000932, -0.00001364, 0.00000075, 0.00000128, -0.00000719, 0.00001301, -0.00000000, -0.00001722, 0.00000441, -0.00000131, -0.00000084, -0.00001258, 0.00000035, 0.00001234, -0.00000779, 0.00000345, 0.00000238, -0.00000651, -0.00000958, 0.00000654, -0.00000064, -0.00000793, -0.00000584, 0.00001261, 0.00000650, -0.00000310, 0.00000513, -0.00000521, -0.00000082, 0.00000028, 0.00000050, 0.00000195, 0.00000211, -0.00000163, -0.00000249, -0.00000115, -0.00000091, 0.00000205, -0.00000161, -0.00000867, -0.00000774, -0.00000507, -0.00000365, 0.00000742, 0.00000221, -0.00001849, -0.00000282, -0.00005887, -0.00004262, 0.00010983, 0.00005780, 0.00010227, 0.00000970, -0.00007585, -0.00000901, -0.00000381, -0.00000530, -0.00001085, -0.00000345, -0.00000775, 0.00000116, 0.00000518, -0.00000227, -0.00000071, -0.00000306, 0.00000242, 0.00000075, -0.00000331, 0.00000304, -0.00000549, 0.00000157, -0.00000742, -0.00000141, 0.00000560, 0.00000427, -0.00000616, -0.00000778, },
		{ -0.00000040, 0.00000036, -0.00000270, -0.00000020, 0.00000295, -0.00000185, -0.00000362, 0.00000484, 0.00000240, -0.00000456, -0.00000609, -0.00000197, -0.00000867, -0.00000219, -0.00000145, -0.00004309, -0.00001626, 0.00005636, 0.00003938, 0.00003507, 0.00006110, -0.00001954, -0.00003686, -0.00001172, -0.00000757, 0.00000502, 0.00000522, -0.00000610, -0.00000439, -0.00000975, -0.00000406, 0.00000069, 0.00000122, -0.00000253, -0.00000006, -0.00000244, -0.00000095, 0.00000099, 0.00000320, 0.00000004, 0.00000120, -0.00000310, -0.00000059, 0.00000236, -0.00000007, 0.00000293, 0.00001584, -0.00001296, 0.00000277, -0.00000873, 0.00000548, -0.00000352, -0.00000990, 0.00000717, -0.00000938, 0.00000738, 0.00000915, -0.00001718, -0.00000244, 0.00000368, -0.00000324, -0.00000613, -0.00001024, 0.00001592, -0.00000485, -0.00000022, 0.00000384, -0.00000198, -0.00001196, 0.00000129, 0.00000499, -0.00000684, -0.00000837, 0.00000547, 0.00001311, -0.00000318, 0.00000346, -0.00000075, -0.00000483, 0.00000204, -0.00000125, 0.00000263, 0.00000122, 0.00000128, -0.00000419, -0.00000009, -0.00000267, 0.00000276, -0.00000099, -0.00000504, -0.00001089, -0.00000330, -0.00000724, 0.00000532, 0.00000557, -0.00001135, -0.00001198, -0.00002289, -0.00008704, 0.00007373, 0.00008671, 0.00008277, 0.00006972, -0.00008050, -0.00002611, -0.00000363, -0.00000304, -0.00001174, -0.00000466, -0.00000704, -0.00000272, 0.00000564, 0.00000010, -0.00000213, -0.00000135, -0.00000157, 0.00000457, -0.00000492, 0.00000308, -0.00000357, -0.00000135, -0.00000241, -0.00000683, 0.00000539, 0.00000430, 0.00000013, -0.00001207, 0.00000297, },
		{ -0.00000239, 0.00000051, -0.00000380, 0.00000408, -0.00000059, -0.00000352, 0.00000105, 0.00000604, -0.00000414, -0.00000500, -0.00000425, -0.00000373, -0.00000913, 0.00000604, -0.00003146, -0.00003352, 0.00003458, 0.00005433, 0.00002084, 0.00006197, 0.00001301, -0.00003911, -0.00002040, -0.00000713, -0.00000117, 0.00000918, -0.00000462, -0.00000380, -0.00000883, -0.00000628, -0.00000097, 0.00000238, -0.00000235, -0.00000023, -0.00000207, -0.00000107, -0.00000086, 0.00000401, 0.00000004, 0.00000205, -0.00000267, -0.00000095, 0.00000029, 0.00000346, -0.00000338, 0.00001713, -0.00000443, -0.00000487, -0.00000127, -0.00000534, 0.00000830, -0.00001603, 0.00000639, -0.00000558, -0.00000071, 0.00001403, -0.00001030, -0.00001220, 0.00000797, -0.00000487, -0.00000114, -0.00001496, 0.00001129, 0.00000328, -0.00000533, 0.00000508, 0.00000031, -0.00000877, -0.00000693, 0.00000831, -0.00000344, -0.00000918, -0.00000157, 0.00001406, 0.00000252, -0.00000145, 0.00000438, -0.00000703, 0.00000161, -0.00000119, 0.00000193, 0.00000107, 0.00000266, -0.00000355, -0.00000093, -0.00000258, 0.00000159, 0.00000035, -0.00000263, -0.00001100, -0.00000474, -0.00000683, 0.00000041, 0.00000714, -0.00000288, -0.00001911, -0.00000352, -0.00008268, -0.00000164, 0.00011833, 0.00006467, 0.00011029, -0.00005192, -0.00005506, -0.00000230, -0.00000430, -0.00000850, -0.00000873, -0.00000464, -0.00000642, 0.00000439, 0.00000285, -0.00000295, -0.00000022, -0.00000365, 0.00000445, -0.00000214, -0.00000058, 0.00000088, -0.00000517, 0.00000116, -0.00000810, 0.00000172, 0.00000513, 0.00000347, -0.00000960, -0.00000427, 0.00000441, },
		{ 0.00000093, -0.00000387, 0.00000185, 0.00000181, -0.00000269, -0.00000255, 0.00000753, -0.00000223, -0.00000399, -0.00000647, -0.00000016, -0.00001183, 0.00000534, -0.00001649, -0.00004021, 0.00000823, 0.00005974, 0.00002250, 0.00004554, 0.00004223, -0.00002789, -0.00002980, -0.00001005, -0.00000357, 0.00000688, 0.00000146, -0.00000625, -0.00000511, -0.00000960, -0.00000145, 0.00000102, 0.00000014, -0.00000214, -0.00000037, -0.00000218, -0.00000087, 0.00000208, 0.00000233, 0.00000059, -0.00000001, -0.00000285, 0.00000033, 0.00000297, -0.00000199, 0.00000835, 0.00000991, -0.00001338, 0.00000552, -0.00001234, 0.00001115, -0.00001108, -0.00000242, 0.00000194, -0.00000758, 0.00001316, 0.00000019, -0.00001904, 0.00000648, -0.00000174, -0.00000145, -0.00001165, -0.00000004, 0.00001280, -0.00000873, 0.00000493, 0.00000111, -0.00000340, -0.00001251, 0.00000578, 0.00000246, -0.00000886, -0.00000634, 0.00000967, 0.00000973, -0.00000389, 0.00000503, -0.00000395, -0.00000270, 0.00000152, -0.00000083, 0.00000268, 0.00000134, 0.00000001, -0.00000384, -0.00000037, -0.00000173, 0.00000319, -0.00000295, -0.00000688, -0.00000978, -0.00000277, -0.00000651, 0.00000872, 0.00000081, -0.00001488, -0.00000776, -0.00004468, -0.00007423, 0.00011277, 0.00007455, 0.00011395, 0.00000956, -0.00008561, -0.00000503, -0.00000746, -0.00000320, -0.00001331, -0.00000258, -0.00000881, 0.00000200, 0.00000436, -0.00000166, -0.00000118, -0.00000258, 0.00000067, 0.00000276, -0.00000431, 0.00000360, -0.00000567, 0.00000020, -0.00000461, -0.00000399, 0.00000601, 0.00000397, -0.00000308, -0.00001144, 0.00000585, -0.00000190, },
		{ -0.00000088, -0.00000218, 0.00000385, -0.00000131, -0.00000439, 0.00000562, 0.00000179, -0.00000422, -0.00000635, -0.00000076, -0.00000852, -0.00000179, -0.00000222, -0.00003864, -0.00001334, 0.00004933, 0.00003741, 0.00002403, 0.00005791, -0.00000653, -0.00003242, -0.00001909, -0.00000216, 0.00000046, 0.00000841, -0.00000816, -0.00000195, -0.00001131, -0.00000249, -0.00000099, 0.00000258, -0.00000351, 0.00000082, -0.00000305, -0.00000009, -0.00000088, 0.00000477, -0.00000101, 0.00000264, -0.00000436, 0.00000096, 0.00000042, 0.00000262, -0.00000148, 0.00001841, -0.00001228, 0.00000279, -0.00000825, 0.00000194, 0.00000224, -0.00001306, 0.00000697, -0.00000934, 0.00000718, 0.00000961, -0.00001855, -0.00000210, 0.00000535, -0.00000532, -0.00000432, -0.00001133, 0.00001713, -0.00000637, 0.00000141, 0.00000267, -0.00000003, -0.00001191, -0.00000201, 0.00000787, -0.00000621, -0.00000923, 0.00000335, 0.00001320, -0.00000039, 0.00000051, 0.00000255, -0.00000766, 0.00000367, -0.00000287, 0.00000374, -0.00000026, 0.00000335, -0.00000558, 0.00000095, -0.00000415, 0.00000471, -0.00000314, -0.00000257, -0.00001339, -0.00000071, -0.00000996, 0.00000685, 0.00000254, -0.00000570, -0.00001888, -0.00001029, -0.00010062, 0.00005165, 0.00010989, 0.00009213, 0.00007970, -0.00009481, -0.00002209, -0.00000711, -0.00000124, -0.00001386, -0.00000400, -0.00000829, -0.00000169, 0.00000458, 0.00000094, -0.00000291, -0.00000041, -0.00000327, 0.00000546, -0.00000438, 0.00000218, -0.00000225, -0.00000363, -0.00000025, -0.00000739, 0.00000409, 0.00000451, 0.00000200, -0.00001234, 0.00000025, 0.00000303, -0.00000180, },
		{ -0.00000414, 0.00000320, 0.00000093, -0.00000429, 0.00000129, 0.00000626, -0.00000468, -0.00000446, -0.00000428, -0.00000225, -0.00000979, 0.00000655, -0.00003073, -0.00002635, 0.00002743, 0.00005313, 0.00001210, 0.00005480, 0.00001919, -0.00002636, -0.00002855, -0.00000398, -0.00000273, 0.00000932, -0.00000430, -0.00000351, -0.00000830, -0.00000643, -0.00000093, 0.00000201, -0.00000181, -0.00000057, -0.00000183, -0.00000077, -0.00000178, 0.00000409, 0.00000032, 0.00000223, -0.00000260, -0.00000079, 0.00000001, 0.00000392, -0.00000386, 0.00001439, 0.00000042, -0.00000796, 0.00000244, -0.00001068, 0.00001259, -0.00001575, 0.00000388, -0.00000470, -0.00000124, 0.00001435, -0.00001069, -0.00001355, 0.00001074, -0.00000693, 0.00000009, -0.00001538, 0.00001183, 0.00000265, -0.00000497, 0.00000535, 0.00000018, -0.00000640, -0.00001035, 0.00000874, -0.00000069, -0.00001043, -0.00000245, 0.00001160, 0.00000653, -0.00000393, 0.00000618, -0.00000756, 0.00000102, -0.00000094, 0.00000145, 0.00000097, 0.00000295, -0.00000298, -0.00000145, -0.00000268, 0.00000191, 0.00000018, -0.00000254, -0.00001072, -0.00000521, -0.00000607, -0.00000120, 0.00000696, -0.00000170, -0.00001896, -0.00000356, -0.00007360, -0.00003814, 0.00013331, 0.00007880, 0.00012388, -0.00006187, -0.00005765, -0.00000146, -0.00000571, -0.00000820, -0.00000987, -0.00000465, -0.00000657, 0.00000440, 0.00000270, -0.00000286, -0.00000014, -0.00000400, 0.00000353, -0.00000033, -0.00000190, 0.00000197, -0.00000611, 0.00000077, -0.00000597, -0.00000100, 0.00000592, 0.00000351, -0.00000683, -0.00000879, 0.00000741, -0.00000320, 0.00000104, },
		{ -0.00000035, 0.00000339, -0.00000281, -0.00000297, 0.00000825, -0.00000303, -0.00000335, -0.00000689, 0.00000155, -0.00001251, 0.00000636, -0.00001779, -0.00003253, 0.00000479, 0.00005606, 0.00001812, 0.00003572, 0.00004258, -0.00001459, -0.00003021, -0.00001434, 0.00000056, 0.00000276, 0.00000475, -0.00000855, -0.00000231, -0.00001155, 0.00000036, -0.00000096, 0.00000193, -0.00000345, 0.00000073, -0.00000286, -0.00000020, 0.00000032, 0.00000408, -0.00000065, 0.00000159, -0.00000432, 0.00000192, 0.00000112, 0.00000063, 0.00000307, 0.00001476, -0.00001519, 0.00000761, -0.00001424, 0.00001010, -0.00000633, -0.00000635, 0.00000262, -0.00000774, 0.00001294, 0.00000064, -0.00002102, 0.00000858, -0.00000210, -0.00000202, -0.00001051, -0.00000085, 0.00001357, -0.00000981, 0.00000644, -0.00000026, -0.00000069, -0.00001396, 0.00000356, 0.00000562, -0.00000880, -0.00000738, 0.00000724, 0.00001138, -0.00000275, 0.00000313, -0.00000105, -0.00000580, 0.00000360, -0.00000287, 0.00000411, -0.00000030, 0.00000239, -0.00000574, 0.00000103, -0.00000338, 0.00000535, -0.00000542, -0.00000391, -0.00001295, 0.00000074, -0.00001047, 0.00001122, -0.00000268, -0.00000833, -0.00001642, -0.00002704, -0.00010175, 0.00010341, 0.00009780, 0.00012607, 0.00001032, -0.00009716, 0.00000018, -0.00001180, -0.00000075, -0.00001612, -0.00000124, -0.00001042, 0.00000335, 0.00000295, -0.00000045, -0.00000218, -0.00000162, -0.00000131, 0.00000450, -0.00000472, 0.00000340, -0.00000484, -0.00000192, -0.00000194, -0.00000571, 0.00000564, 0.00000385, -0.00000031, -0.00001366, 0.00000500, 0.00000057, -0.00000084, -0.00000223, },
		{ 0.00000401, -0.00000030, -0.00000520, 0.00000619, 0.00000150, -0.00000424, -0.00000628, 0.00000011, -0.00000815, -0.00000145, -0.00000317, -0.00003408, -0.00001062, 0.00004205, 0.00003722, 0.00001316, 0.00005496, 0.00000108, -0.00002287, -0.00002819, 0.00000352, -0.00000368, 0.00001097, -0.00000987, 0.00000018, -0.00001239, -0.00000133, -0.00000224, 0.00000347, -0.00000398, 0.00000132, -0.00000351, 0.00000084, -0.00000270, 0.00000602, -0.00000177, 0.00000383, -0.00000534, 0.00000209, -0.00000091, 0.00000450, -0.00000431, 0.00001854, -0.00000937, 0.00000067, -0.00000535, -0.00000329, 0.00000807, -0.00001503, 0.00000554, -0.00000839, 0.00000634, 0.00001034, -0.00001997, -0.00000183, 0.00000712, -0.00000740, -0.00000228, -0.00001268, 0.00001830, -0.00000770, 0.00000281, 0.00000169, 0.00000129, -0.00001092, -0.00000518, 0.00000909, -0.00000373, -0.00001099, 0.00000225, 0.00001157, 0.00000370, -0.00000307, 0.00000603, -0.00001015, 0.00000478, -0.00000404, 0.00000452, -0.00000153, 0.00000503, -0.00000649, 0.00000172, -0.00000552, 0.00000654, -0.00000507, -0.00000046, -0.00001532, 0.00000127, -0.00001183, 0.00000764, -0.00000015, -0.00000074, -0.00002403, -0.00000214, -0.00010305, 0.00001773, 0.00013340, 0.00010597, 0.00008938, -0.00010968, -0.00001829, -0.00001012, 0.00000019, -0.00001593, -0.00000314, -0.00000978, -0.00000046, 0.00000329, 0.00000195, -0.00000374, 0.00000044, -0.00000465, 0.00000577, -0.00000334, 0.00000105, -0.00000098, -0.00000514, 0.00000034, -0.00000628, 0.00000171, 0.00000535, 0.00000274, -0.00001065, -0.00000417, 0.00000702, -0.00000381, 0.00000071, -0.00000366, },
		{ 0.00000272, -0.00000492, 0.00000134, 0.00000660, -0.00000527, -0.00000397, -0.00000421, -0.00000099, -0.00001019, 0.00000683, -0.00003010, -0.00001895, 0.00001951, 0.00005349, 0.00000365, 0.00004880, 0.00002114, -0.00001305, -0.00003362, -0.00000398, -0.00000159, 0.00000737, -0.00000253, -0.00000458, -0.00000643, -0.00000769, 0.00000016, 0.00000065, -0.00000041, -0.00000154, -0.00000110, -0.00000095, -0.00000192, 0.00000321, 0.00000146, 0.00000156, -0.00000158, -0.00000175, 0.00000098, 0.00000297, -0.00000248, 0.00000978, 0.00000615, -0.00001144, 0.00000616, -0.00001474, 0.00001438, -0.00001294, -0.00000011, -0.00000312, -0.00000203, 0.00001459, -0.00001095, -0.00001497, 0.00001348, -0.00000886, 0.00000137, -0.00001570, 0.00001199, 0.00000240, -0.00000484, 0.00000575, -0.00000030, -0.00000401, -0.00001250, 0.00000728, 0.00000294, -0.00001134, -0.00000334, 0.00000890, 0.00000978, -0.00000500, 0.00000637, -0.00000634, -0.00000107, 0.00000047, -0.00000000, 0.00000179, 0.00000214, -0.00000137, -0.00000278, -0.00000196, 0.00000123, 0.00000124, -0.00000385, -0.00000886, -0.00000727, -0.00000354, -0.00000448, 0.00000848, -0.00000319, -0.00001478, -0.00000960, -0.00005639, -0.00007641, 0.00013827, 0.00010125, 0.00013638, -0.00007094, -0.00006192, 0.00000075, -0.00000801, -0.00000742, -0.00001132, -0.00000429, -0.00000716, 0.00000481, 0.00000206, -0.00000222, -0.00000063, -0.00000375, 0.00000212, 0.00000152, -0.00000282, 0.00000231, -0.00000581, -0.00000098, -0.00000314, -0.00000363, 0.00000647, 0.00000318, -0.00000351, -0.00001283, 0.00000880, -0.00000237, 0.00000035, -0.00000334, 0.00000141, },
		{ -0.00000256, -0.00000346, 0.00000877, -0.00000350, -0.00000313, -0.00000686, 0.00000275, -0.00001275, 0.00000711, -0.00001913, -0.00002485, 0.00000113, 0.00005259, 0.00001524, 0.00002607, 0.00004213, -0.00000544, -0.00002513, -0.00002184, 0.00000650, -0.00000225, 0.00000843, -0.00001132, 0.00000092, -0.00001368, 0.00000235, -0.00000307, 0.00000372, -0.00000467, 0.00000178, -0.00000368, 0.00000091, -0.00000188, 0.00000602, -0.00000209, 0.00000344, -0.00000605, 0.00000372, -0.00000083, 0.00000333, -0.00000161, 0.00001787, -0.00001501, 0.00000745, -0.00001315, 0.00000622, -0.00000007, -0.00001027, 0.00000276, -0.00000731, 0.00001218, 0.00000137, -0.00002302, 0.00001057, -0.00000230, -0.00000261, -0.00000915, -0.00000204, 0.00001448, -0.00001082, 0.00000782, -0.00000160, 0.00000150, -0.00001415, 0.00000070, 0.00000777, -0.00000703, -0.00000934, 0.00000576, 0.00001114, 0.00000031, -0.00000029, 0.00000302, -0.00000947, 0.00000589, -0.00000505, 0.00000576, -0.00000229, 0.00000498, -0.00000771, 0.00000266, -0.00000547, 0.00000805, -0.00000843, -0.00000046, -0.00001644, 0.00000457, -0.00001452, 0.00001396, -0.00000709, -0.00000059, -0.00002601, -0.00001049, -0.00011963, 0.00007883, 0.00012665, 0.00014027, 0.00001113, -0.00010982, 0.00000592, -0.00001627, 0.00000174, -0.00001910, 0.00000052, -0.00001252, 0.00000516, 0.00000101, 0.00000123, -0.00000356, -0.00000042, -0.00000327, 0.00000582, -0.00000455, 0.00000265, -0.00000355, -0.00000391, -0.00000032, -0.00000593, 0.00000412, 0.00000436, 0.00000148, -0.00001381, 0.00000180, 0.00000449, -0.00000328, -0.00000010, -0.00000266, 0.00000458, },
		{ -0.00000595, 0.00000647, 0.00000151, -0.00000455, -0.00000596, 0.00000075, -0.00000771, -0.00000108, -0.00000421, -0.00002977, -0.00000760, 0.00003411, 0.00003889, 0.00000213, 0.00005299, 0.00000392, -0.00001060, -0.00003642, 0.00000741, -0.00000587, 0.00001179, -0.00001033, 0.00000112, -0.00001220, -0.00000124, -0.00000247, 0.00000335, -0.00000353, 0.00000112, -0.00000349, 0.00000141, -0.00000393, 0.00000641, -0.00000173, 0.00000421, -0.00000542, 0.00000214, -0.00000101, 0.00000494, -0.00000501, 0.00001602, -0.00000439, -0.00000318, -0.00000092, -0.00000879, 0.00001257, -0.00001488, 0.00000248, -0.00000639, 0.00000490, 0.00001122, -0.00002134, -0.00000162, 0.00000886, -0.00000932, -0.00000019, -0.00001403, 0.00001917, -0.00000864, 0.00000391, 0.00000088, 0.00000212, -0.00000945, -0.00000755, 0.00000833, 0.00000025, -0.00001302, 0.00000162, 0.00000893, 0.00000806, -0.00000617, 0.00000860, -0.00001130, 0.00000454, -0.00000409, 0.00000438, -0.00000199, 0.00000572, -0.00000640, 0.00000170, -0.00000616, 0.00000754, -0.00000596, 0.00000039, -0.00001576, 0.00000166, -0.00001187, 0.00000669, -0.00000122, 0.00000180, -0.00002522, -0.00000088, -0.00009342, -0.00002501, 0.00015192, 0.00012787, 0.00009701, -0.00012373, -0.00001584, -0.00001192, 0.00000077, -0.00001757, -0.00000244, -0.00001113, 0.00000058, 0.00000214, 0.00000272, -0.00000424, 0.00000088, -0.00000550, 0.00000549, -0.00000202, 0.00000003, -0.00000024, -0.00000543, -0.00000074, -0.00000378, -0.00000131, 0.00000651, 0.00000257, -0.00000750, -0.00000932, 0.00001042, -0.00000478, 0.00000145, -0.00000418, 0.00000322, 0.00000115, },
		{ 0.00000123, 0.00000688, -0.00000566, -0.00000377, -0.00000381, -0.00000019, -0.00001015, 0.00000683, -0.00002956, -0.00001149, 0.00001132, 0.00005453, -0.00000375, 0.00004304, 0.00002126, -0.00000217, -0.00003377, -0.00000781, 0.00000235, 0.00000347, 0.00000055, -0.00000689, -0.00000338, -0.00000983, 0.00000205, -0.00000143, 0.00000155, -0.00000290, -0.00000006, -0.00000157, -0.00000130, 0.00000143, 0.00000330, 0.00000018, 0.00000022, -0.00000357, 0.00000282, 0.00000112, 0.00000004, 0.00000441, 0.00001148, -0.00001396, 0.00000839, -0.00001612, 0.00001294, -0.00000760, -0.00000519, -0.00000125, -0.00000275, 0.00001455, -0.00001103, -0.00001634, 0.00001600, -0.00001049, 0.00000253, -0.00001577, 0.00001166, 0.00000249, -0.00000485, 0.00000615, -0.00000093, -0.00000187, -0.00001332, 0.00000452, 0.00000636, -0.00001097, -0.00000495, 0.00000677, 0.00001145, -0.00000404, 0.00000456, -0.00000325, -0.00000454, 0.00000284, -0.00000223, 0.00000337, 0.00000041, 0.00000107, -0.00000477, -0.00000048, -0.00000045, 0.00000348, -0.00000647, -0.00000564, -0.00001065, 0.00000036, -0.00000896, 0.00001134, -0.00000697, -0.00000735, -0.00001986, -0.00003510, -0.00010980, 0.00012771, 0.00013364, 0.00014793, -0.00007909, -0.00006788, 0.00000417, -0.00001102, -0.00000623, -0.00001314, -0.00000340, -0.00000837, 0.00000581, 0.00000076, -0.00000092, -0.00000172, -0.00000299, 0.00000042, 0.00000319, -0.00000322, 0.00000192, -0.00000457, -0.00000333, -0.00000055, -0.00000532, 0.00000618, 0.00000311, -0.00000056, -0.00001528, 0.00000781, 0.00000037, -0.00000155, -0.00000153, -0.00000105, 0.00000454, -0.00000531, },
		{ 0.00000898, -0.00000354, -0.00000339, -0.00000634, 0.00000342, -0.00001260, 0.00000763, -0.00002038, -0.00001761, -0.00000202, 0.00004849, 0.00001442, 0.00001580, 0.00004243, -0.00000116, -0.00001566, -0.00003070, 0.00001247, -0.00000667, 0.00001132, -0.00001354, 0.00000360, -0.00001509, 0.00000371, -0.00000454, 0.00000484, -0.00000521, 0.00000231, -0.00000423, 0.00000199, -0.00000392, 0.00000752, -0.00000309, 0.00000487, -0.00000728, 0.00000486, -0.00000200, 0.00000510, -0.00000460, 0.00001832, -0.00001220, 0.00000465, -0.00000923, 0.00000040, 0.00000646, -0.00001312, 0.00000172, -0.00000594, 0.00001074, 0.00000233, -0.00002494, 0.00001236, -0.00000237, -0.00000313, -0.00000768, -0.00000339, 0.00001527, -0.00001155, 0.00000892, -0.00000280, 0.00000317, -0.00001347, -0.00000198, 0.00000819, -0.00000341, -0.00001204, 0.00000507, 0.00000931, 0.00000465, -0.00000436, 0.00000722, -0.00001266, 0.00000747, -0.00000660, 0.00000690, -0.00000391, 0.00000703, -0.00000907, 0.00000389, -0.00000729, 0.00001046, -0.00001102, 0.00000239, -0.00001908, 0.00000742, -0.00001732, 0.00001554, -0.00001074, 0.00000600, -0.00003324, 0.00000053, -0.00012378, 0.00003880, 0.00015755, 0.00015989, 0.00001008, -0.00012213, 0.00001094, -0.00001997, 0.00000365, -0.00002185, 0.00000235, -0.00001478, 0.00000710, -0.00000110, 0.00000302, -0.00000494, 0.00000067, -0.00000491, 0.00000660, -0.00000391, 0.00000169, -0.00000238, -0.00000502, -0.00000030, -0.00000446, 0.00000147, 0.00000555, 0.00000207, -0.00001185, -0.00000326, 0.00000900, -0.00000559, 0.00000184, -0.00000452, 0.00000503, -0.00000183, -0.00000396, },
		{ 0.00000169, -0.00000494, -0.00000562, 0.00000140, -0.00000742, -0.00000051, -0.00000533, -0.00002586, -0.00000413, 0.00002561, 0.00004186, -0.00000858, 0.00005156, 0.00000385, 0.00000133, -0.00004120, 0.00000788, -0.00000511, 0.00001024, -0.00000902, 0.00000039, -0.00001040, -0.00000249, -0.00000148, 0.00000208, -0.00000209, 0.00000015, -0.00000288, 0.00000139, -0.00000425, 0.00000566, -0.00000066, 0.00000355, -0.00000436, 0.00000091, 0.00000024, 0.00000389, -0.00000371, 0.00001141, 0.00000179, -0.00000764, 0.00000352, -0.00001285, 0.00001439, -0.00001198, -0.00000231, -0.00000348, 0.00000305, 0.00001209, -0.00002259, -0.00000139, 0.00001038, -0.00001085, 0.00000170, -0.00001512, 0.00001951, -0.00000908, 0.00000470, 0.00000021, 0.00000259, -0.00000785, -0.00000878, 0.00000583, 0.00000485, -0.00001436, 0.00000068, 0.00000622, 0.00001158, -0.00000773, 0.00000930, -0.00001038, 0.00000251, -0.00000274, 0.00000303, -0.00000132, 0.00000513, -0.00000508, 0.00000064, -0.00000571, 0.00000722, -0.00000529, -0.00000055, -0.00001419, -0.00000002, -0.00000968, 0.00000361, -0.00000009, 0.00000106, -0.00002158, -0.00000684, -0.00007371, -0.00007064, 0.00015862, 0.00016160, 0.00010109, -0.00013587, -0.00001558, -0.00001198, 0.00000018, -0.00001849, -0.00000225, -0.00001194, 0.00000100, 0.00000156, 0.00000286, -0.00000407, 0.00000067, -0.00000569, 0.00000467, -0.00000058, -0.00000066, -0.00000023, -0.00000450, -0.00000307, -0.00000068, -0.00000419, 0.00000739, 0.00000206, -0.00000379, -0.00001392, 0.00001205, -0.00000397, 0.00000089, -0.00000351, 0.00000142, 0.00000295, -0.00000657, 0.00000652, },
		{ -0.00000566, -0.00000404, -0.00000289, -0.00000000, -0.00000960, 0.00000650, -0.00002900, -0.00000436, 0.00000365, 0.00005516, -0.00000916, 0.00003642, 0.00002170, 0.00000432, -0.00002861, -0.00001482, 0.00000811, -0.00000139, 0.00000412, -0.00000972, 0.00000009, -0.00001210, 0.00000406, -0.00000356, 0.00000345, -0.00000409, 0.00000088, -0.00000230, -0.00000023, -0.00000081, 0.00000535, -0.00000140, 0.00000224, -0.00000559, 0.00000474, -0.00000075, 0.00000261, -0.00000039, 0.00001502, -0.00001425, 0.00000795, -0.00001410, 0.00000832, -0.00000045, -0.00001051, 0.00000026, -0.00000296, 0.00001401, -0.00001090, -0.00001756, 0.00001815, -0.00001173, 0.00000354, -0.00001557, 0.00001093, 0.00000275, -0.00000479, 0.00000640, -0.00000160, -0.00000010, -0.00001306, 0.00000129, 0.00000857, -0.00000870, -0.00000753, 0.00000552, 0.00001128, -0.00000109, 0.00000109, 0.00000114, -0.00000865, 0.00000541, -0.00000459, 0.00000513, -0.00000165, 0.00000370, -0.00000684, 0.00000127, -0.00000261, 0.00000630, -0.00000967, -0.00000189, -0.00001436, 0.00000451, -0.00001342, 0.00001432, -0.00001160, 0.00000121, -0.00003101, -0.00001520, -0.00013152, 0.00009762, 0.00017539, 0.00016041, -0.00008752, -0.00007458, 0.00000789, -0.00001404, -0.00000507, -0.00001509, -0.00000212, -0.00001008, 0.00000727, -0.00000107, 0.00000086, -0.00000317, -0.00000198, -0.00000130, 0.00000451, -0.00000310, 0.00000103, -0.00000294, -0.00000542, 0.00000089, -0.00000548, 0.00000466, 0.00000371, 0.00000134, -0.00001549, 0.00000425, 0.00000470, -0.00000419, 0.00000084, -0.00000371, 0.00000601, -0.00000531, 0.00000105, 0.00000693, },
		{ -0.00000399, -0.00000550, 0.00000374, -0.00001224, 0.00000802, -0.00002149, -0.00001110, -0.00000418, 0.00004336, 0.00001567, 0.00000488, 0.00004380, -0.00000087, -0.00000424, -0.00003834, 0.00001640, -0.00000892, 0.00001227, -0.00001424, 0.00000480, -0.00001495, 0.00000374, -0.00000477, 0.00000476, -0.00000465, 0.00000198, -0.00000417, 0.00000258, -0.00000523, 0.00000801, -0.00000312, 0.00000528, -0.00000739, 0.00000469, -0.00000177, 0.00000531, -0.00000529, 0.00001585, -0.00000689, -0.00000033, -0.00000351, -0.00000583, 0.00001170, -0.00001384, -0.00000104, -0.00000340, 0.00000863, 0.00000345, -0.00002671, 0.00001401, -0.00000246, -0.00000339, -0.00000634, -0.00000461, 0.00001564, -0.00001177, 0.00000963, -0.00000383, 0.00000439, -0.00001229, -0.00000391, 0.00000668, 0.00000156, -0.00001472, 0.00000456, 0.00000668, 0.00000917, -0.00000786, 0.00001033, -0.00001423, 0.00000740, -0.00000674, 0.00000683, -0.00000445, 0.00000783, -0.00000918, 0.00000407, -0.00000812, 0.00001170, -0.00001220, 0.00000358, -0.00001973, 0.00000811, -0.00001767, 0.00001474, -0.00001208, 0.00000933, -0.00003544, 0.00000297, -0.00011278, -0.00001344, 0.00018422, 0.00018964, 0.00000464, -0.00013221, 0.00001374, -0.00002184, 0.00000426, -0.00002378, 0.00000371, -0.00001665, 0.00000860, -0.00000285, 0.00000438, -0.00000581, 0.00000123, -0.00000594, 0.00000671, -0.00000289, 0.00000077, -0.00000171, -0.00000491, -0.00000191, -0.00000168, -0.00000180, 0.00000703, 0.00000174, -0.00000834, -0.00000907, 0.00001286, -0.00000677, 0.00000287, -0.00000539, 0.00000444, -0.00000057, -0.00000443, 0.00000888, -0.00000322, },
		{ -0.00000556, 0.00000235, -0.00000756, 0.00000043, -0.00000661, -0.00002232, -0.00000041, 0.00001713, 0.00004511, -0.00001795, 0.00004953, 0.00000324, 0.00001012, -0.00004086, 0.00000428, -0.00000130, 0.00000649, -0.00000609, -0.00000186, -0.00000721, -0.00000478, 0.00000045, -0.00000002, 0.00000001, -0.00000127, -0.00000187, 0.00000083, -0.00000369, 0.00000387, 0.00000126, 0.00000203, -0.00000237, -0.00000128, 0.00000240, 0.00000191, -0.00000117, 0.00000584, 0.00000775, -0.00001120, 0.00000629, -0.00001397, 0.00001270, -0.00000628, -0.00000845, -0.00000005, 0.00000114, 0.00001272, -0.00002366, -0.00000103, 0.00001148, -0.00001181, 0.00000321, -0.00001578, 0.00001923, -0.00000904, 0.00000527, -0.00000042, 0.00000281, -0.00000631, -0.00000891, 0.00000225, 0.00000896, -0.00001406, -0.00000122, 0.00000420, 0.00001343, -0.00000710, 0.00000775, -0.00000728, -0.00000118, -0.00000021, 0.00000072, 0.00000029, 0.00000343, -0.00000273, -0.00000129, -0.00000426, 0.00000561, -0.00000311, -0.00000313, -0.00001087, -0.00000343, -0.00000570, -0.00000105, 0.00000279, -0.00000251, -0.00001400, -0.00001801, -0.00004868, -0.00011119, 0.00014639, 0.00020985, 0.00010136, -0.00014598, -0.00001758, -0.00001044, -0.00000143, -0.00001869, -0.00000267, -0.00001200, 0.00000060, 0.00000176, 0.00000216, -0.00000310, -0.00000024, -0.00000528, 0.00000345, 0.00000081, -0.00000093, -0.00000087, -0.00000272, -0.00000586, 0.00000205, -0.00000606, 0.00000733, 0.00000187, -0.00000049, -0.00001680, 0.00001107, -0.00000108, -0.00000099, -0.00000170, -0.00000118, 0.00000486, -0.00000695, 0.00000639, 0.00000182, -0.00000512, },
		{ -0.00000148, -0.00000038, -0.00000860, 0.00000590, -0.00002832, 0.00000204, -0.00000280, 0.00005449, -0.00001194, 0.00002826, 0.00002366, 0.00000602, -0.00001939, -0.00002310, 0.00001387, -0.00000566, 0.00000693, -0.00001197, 0.00000290, -0.00001353, 0.00000530, -0.00000495, 0.00000457, -0.00000452, 0.00000124, -0.00000272, 0.00000077, -0.00000286, 0.00000692, -0.00000249, 0.00000375, -0.00000696, 0.00000582, -0.00000172, 0.00000420, -0.00000341, 0.00001573, -0.00001156, 0.00000438, -0.00000883, 0.00000145, 0.00000720, -0.00001491, 0.00000063, -0.00000220, 0.00001277, -0.00001055, -0.00001857, 0.00001992, -0.00001264, 0.00000447, -0.00001524, 0.00001005, 0.00000290, -0.00000441, 0.00000636, -0.00000220, 0.00000129, -0.00001202, -0.00000170, 0.00000893, -0.00000452, -0.00001075, 0.00000491, 0.00000962, 0.00000312, -0.00000308, 0.00000567, -0.00001217, 0.00000711, -0.00000613, 0.00000621, -0.00000317, 0.00000564, -0.00000816, 0.00000249, -0.00000438, 0.00000868, -0.00001226, 0.00000105, -0.00001696, 0.00000733, -0.00001621, 0.00001570, -0.00001499, 0.00000806, -0.00003909, -0.00000215, -0.00013624, 0.00004704, 0.00022278, 0.00017787, -0.00009880, -0.00008008, 0.00001031, -0.00001587, -0.00000477, -0.00001656, -0.00000097, -0.00001179, 0.00000873, -0.00000295, 0.00000261, -0.00000450, -0.00000116, -0.00000265, 0.00000524, -0.00000245, -0.00000014, -0.00000144, -0.00000658, 0.00000071, -0.00000396, 0.00000200, 0.00000501, 0.00000201, -0.00001346, -0.00000129, 0.00000964, -0.00000668, 0.00000304, -0.00000593, 0.00000670, -0.00000475, 0.00000079, 0.00000669, -0.00000520, -0.00000479, },
		{ 0.00000402, -0.00001198, 0.00000853, -0.00002252, -0.00000545, -0.00000526, 0.00003734, 0.00001838, -0.00000596, 0.00004544, -0.00000254, 0.00000615, -0.00004225, 0.00001663, -0.00000797, 0.00001059, -0.00001284, 0.00000398, -0.00001290, 0.00000216, -0.00000354, 0.00000332, -0.00000293, 0.00000076, -0.00000339, 0.00000244, -0.00000549, 0.00000718, -0.00000192, 0.00000444, -0.00000611, 0.00000303, -0.00000006, 0.00000392, -0.00000382, 0.00001099, -0.00000002, -0.00000623, 0.00000231, -0.00001056, 0.00001415, -0.00001163, -0.00000573, 0.00000024, 0.00000599, 0.00000459, -0.00002830, 0.00001561, -0.00000279, -0.00000314, -0.00000541, -0.00000538, 0.00001534, -0.00001134, 0.00000992, -0.00000468, 0.00000518, -0.00001083, -0.00000490, 0.00000361, 0.00000689, -0.00001633, 0.00000339, 0.00000422, 0.00001268, -0.00000963, 0.00001127, -0.00001334, 0.00000513, -0.00000509, 0.00000521, -0.00000352, 0.00000702, -0.00000770, 0.00000285, -0.00000745, 0.00001116, -0.00001131, 0.00000242, -0.00001776, 0.00000601, -0.00001500, 0.00001101, -0.00001034, 0.00000830, -0.00003144, -0.00000381, -0.00008865, -0.00007114, 0.00019829, 0.00023474, -0.00000759, -0.00013847, 0.00001313, -0.00002110, 0.00000303, -0.00002439, 0.00000407, -0.00001754, 0.00000912, -0.00000367, 0.00000478, -0.00000574, 0.00000090, -0.00000612, 0.00000611, -0.00000157, 0.00000001, -0.00000166, -0.00000367, -0.00000462, 0.00000156, -0.00000482, 0.00000813, 0.00000114, -0.00000428, -0.00001425, 0.00001475, -0.00000598, 0.00000248, -0.00000492, 0.00000269, 0.00000118, -0.00000486, 0.00000856, -0.00000273, -0.00000465, -0.00000463, },
		{ -0.00000827, 0.00000183, -0.00000809, -0.00001903, 0.00000314, 0.00000945, 0.00004748, -0.00002484, 0.00004568, 0.00000407, 0.00001409, -0.00003513, -0.00000267, 0.00000454, 0.00000154, -0.00000242, -0.00000485, -0.00000346, -0.00000731, 0.00000256, -0.00000224, 0.00000208, -0.00000257, -0.00000091, 0.00000008, -0.00000259, 0.00000154, 0.00000350, 0.00000022, -0.00000009, -0.00000368, 0.00000458, -0.00000008, 0.00000147, 0.00000076, 0.00001197, -0.00001238, 0.00000603, -0.00001130, 0.00000750, 0.00000155, -0.00001506, 0.00000315, -0.00000029, 0.00001285, -0.00002450, -0.00000047, 0.00001204, -0.00001214, 0.00000429, -0.00001599, 0.00001843, -0.00000871, 0.00000581, -0.00000111, 0.00000286, -0.00000493, -0.00000818, -0.00000160, 0.00001156, -0.00001158, -0.00000427, 0.00000312, 0.00001339, -0.00000435, 0.00000429, -0.00000263, -0.00000568, 0.00000265, -0.00000182, 0.00000217, 0.00000131, -0.00000008, -0.00000344, -0.00000238, 0.00000333, -0.00000016, -0.00000648, -0.00000681, -0.00000739, -0.00000130, -0.00000584, 0.00000595, -0.00000714, -0.00000500, -0.00003046, -0.00002487, -0.00013822, 0.00010966, 0.00027285, 0.00009964, -0.00015545, -0.00002069, -0.00000837, -0.00000323, -0.00001871, -0.00000340, -0.00001152, -0.00000045, 0.00000258, 0.00000081, -0.00000156, -0.00000159, -0.00000454, 0.00000213, 0.00000194, -0.00000075, -0.00000197, -0.00000064, -0.00000825, 0.00000351, -0.00000632, 0.00000596, 0.00000243, 0.00000169, -0.00001726, 0.00000731, 0.00000355, -0.00000370, 0.00000076, -0.00000409, 0.00000648, -0.00000694, 0.00000615, 0.00000175, -0.00000491, -0.00000594, 0.00000377, },
		{ -0.00000741, 0.00000521, -0.00002752, 0.00000742, -0.00000757, 0.00005212, -0.00001208, 0.00001867, 0.00002710, 0.00000409, -0.00000858, -0.00003007, 0.00001753, -0.00000771, 0.00000773, -0.00001255, 0.00000403, -0.00001323, 0.00000502, -0.00000493, 0.00000431, -0.00000372, 0.00000065, -0.00000246, 0.00000121, -0.00000411, 0.00000738, -0.00000247, 0.00000409, -0.00000700, 0.00000540, -0.00000113, 0.00000412, -0.00000402, 0.00001329, -0.00000597, -0.00000181, -0.00000147, -0.00000597, 0.00001363, -0.00001713, -0.00000081, -0.00000016, 0.00001075, -0.00000998, -0.00001938, 0.00002138, -0.00001342, 0.00000558, -0.00001509, 0.00000937, 0.00000256, -0.00000346, 0.00000588, -0.00000263, 0.00000226, -0.00001048, -0.00000399, 0.00000739, 0.00000093, -0.00001373, 0.00000418, 0.00000737, 0.00000741, -0.00000661, 0.00000893, -0.00001379, 0.00000684, -0.00000594, 0.00000577, -0.00000330, 0.00000605, -0.00000794, 0.00000235, -0.00000482, 0.00000951, -0.00001301, 0.00000184, -0.00001700, 0.00000732, -0.00001578, 0.00001388, -0.00001518, 0.00001054, -0.00004068, 0.00000006, -0.00012166, -0.00002096, 0.00026871, 0.00020625, -0.00011647, -0.00008184, 0.00000942, -0.00001505, -0.00000637, -0.00001666, -0.00000075, -0.00001272, 0.00000941, -0.00000413, 0.00000362, -0.00000506, -0.00000111, -0.00000320, 0.00000516, -0.00000127, -0.00000141, -0.00000035, -0.00000657, -0.00000100, -0.00000124, -0.00000121, 0.00000653, 0.00000180, -0.00000983, -0.00000761, 0.00001383, -0.00000792, 0.00000423, -0.00000707, 0.00000621, -0.00000350, 0.00000044, 0.00000621, -0.00000443, -0.00000582, 0.00000021, -0.00000109, },
		{ 0.00000936, -0.00002357, -0.00000056, -0.00000553, 0.00003106, 0.00002139, -0.00001545, 0.00004592, -0.00000375, 0.00001283, -0.00004081, 0.00001254, -0.00000370, 0.00000642, -0.00000948, 0.00000130, -0.00000919, -0.00000069, -0.00000119, 0.00000088, -0.00000043, -0.00000103, -0.00000210, 0.00000165, -0.00000472, 0.00000516, 0.00000032, 0.00000254, -0.00000371, 0.00000027, 0.00000264, 0.00000152, -0.00000096, 0.00000497, 0.00000689, -0.00001133, 0.00000633, -0.00001213, 0.00001284, -0.00000632, -0.00001202, 0.00000458, 0.00000320, 0.00000554, -0.00002966, 0.00001729, -0.00000357, -0.00000219, -0.00000511, -0.00000552, 0.00001425, -0.00001027, 0.00000985, -0.00000536, 0.00000558, -0.00000923, -0.00000506, -0.00000027, 0.00001137, -0.00001590, 0.00000090, 0.00000268, 0.00001434, -0.00000895, 0.00000958, -0.00000982, 0.00000080, -0.00000190, 0.00000227, -0.00000131, 0.00000476, -0.00000485, 0.00000036, -0.00000533, 0.00000884, -0.00000837, -0.00000097, -0.00001336, 0.00000145, -0.00000978, 0.00000489, -0.00000594, 0.00000336, -0.00002212, -0.00001767, -0.00005675, -0.00012487, 0.00019073, 0.00029943, -0.00002771, -0.00014038, 0.00000880, -0.00001772, -0.00000005, -0.00002355, 0.00000317, -0.00001714, 0.00000830, -0.00000325, 0.00000392, -0.00000451, -0.00000043, -0.00000543, 0.00000485, -0.00000004, -0.00000055, -0.00000210, -0.00000176, -0.00000756, 0.00000420, -0.00000667, 0.00000812, 0.00000101, -0.00000075, -0.00001751, 0.00001376, -0.00000286, 0.00000057, -0.00000314, -0.00000008, 0.00000322, -0.00000515, 0.00000813, -0.00000234, -0.00000506, -0.00000407, 0.00000460, -0.00001211, },
		{ -0.00000968, -0.00001589, 0.00000611, 0.00000332, 0.00004801, -0.00002853, 0.00003933, 0.00000730, 0.00001310, -0.00002536, -0.00001107, 0.00001054, -0.00000297, 0.00000068, -0.00000738, -0.00000033, -0.00000900, 0.00000389, -0.00000368, 0.00000333, -0.00000306, -0.00000053, -0.00000037, -0.00000152, -0.00000066, 0.00000529, -0.00000111, 0.00000167, -0.00000537, 0.00000583, -0.00000109, 0.00000311, -0.00000255, 0.00001328, -0.00001028, 0.00000220, -0.00000501, -0.00000024, 0.00001009, -0.00002086, 0.00000523, -0.00000071, 0.00001221, -0.00002503, 0.00000030, 0.00001212, -0.00001195, 0.00000508, -0.00001596, 0.00001740, -0.00000844, 0.00000662, -0.00000204, 0.00000287, -0.00000376, -0.00000686, -0.00000504, 0.00001214, -0.00000703, -0.00000799, 0.00000261, 0.00001189, -0.00000025, -0.00000002, 0.00000227, -0.00000961, 0.00000459, -0.00000347, 0.00000329, -0.00000020, 0.00000183, -0.00000480, -0.00000105, 0.00000145, 0.00000233, -0.00000920, -0.00000360, -0.00001013, 0.00000161, -0.00000871, 0.00000725, -0.00001028, 0.00000196, -0.00003933, -0.00000907, -0.00014469, 0.00004612, 0.00034718, 0.00010046, -0.00016744, -0.00002248, -0.00000776, -0.00000373, -0.00001962, -0.00000360, -0.00001122, -0.00000148, 0.00000337, -0.00000056, -0.00000010, -0.00000276, -0.00000404, 0.00000118, 0.00000250, -0.00000004, -0.00000335, 0.00000129, -0.00000964, 0.00000331, -0.00000488, 0.00000340, 0.00000373, 0.00000258, -0.00001532, 0.00000139, 0.00000887, -0.00000628, 0.00000307, -0.00000659, 0.00000730, -0.00000630, 0.00000576, 0.00000143, -0.00000428, -0.00000676, 0.00000561, -0.00000947, 0.00000358, },
		{ -0.00002670, 0.00001175, -0.00001065, 0.00004825, -0.00001030, 0.00000864, 0.00003090, 0.00000076, 0.00000078, -0.00003320, 0.00001736, -0.00000640, 0.00000575, -0.00001081, 0.00000288, -0.00001076, 0.00000292, -0.00000325, 0.00000253, -0.00000161, -0.00000094, -0.00000137, 0.00000083, -0.00000419, 0.00000639, -0.00000107, 0.00000300, -0.00000546, 0.00000328, 0.00000108, 0.00000233, -0.00000231, 0.00000821, 0.00000156, -0.00000924, 0.00000610, -0.00001187, 0.00001716, -0.00001622, -0.00000442, 0.00000322, 0.00000805, -0.00000927, -0.00001999, 0.00002272, -0.00001435, 0.00000717, -0.00001545, 0.00000926, 0.00000145, -0.00000177, 0.00000489, -0.00000282, 0.00000275, -0.00000856, -0.00000549, 0.00000447, 0.00000646, -0.00001527, 0.00000243, 0.00000558, 0.00001049, -0.00000816, 0.00000969, -0.00001251, 0.00000390, -0.00000352, 0.00000333, -0.00000153, 0.00000440, -0.00000567, 0.00000032, -0.00000325, 0.00000797, -0.00001104, -0.00000044, -0.00001360, 0.00000358, -0.00001129, 0.00000799, -0.00001100, 0.00000713, -0.00003407, -0.00000985, -0.00008943, -0.00009906, 0.00030326, 0.00025245, -0.00014415, -0.00007743, 0.00000344, -0.00001036, -0.00001080, -0.00001458, -0.00000230, -0.00001201, 0.00000850, -0.00000384, 0.00000315, -0.00000420, -0.00000233, -0.00000257, 0.00000403, 0.00000053, -0.00000279, 0.00000033, -0.00000562, -0.00000359, 0.00000172, -0.00000401, 0.00000749, 0.00000146, -0.00000574, -0.00001316, 0.00001583, -0.00000694, 0.00000378, -0.00000667, 0.00000432, -0.00000153, -0.00000006, 0.00000565, -0.00000362, -0.00000679, 0.00000179, -0.00000174, -0.00000640, 0.00000471, },
		{ 0.00000371, -0.00000545, 0.00002539, 0.00002343, -0.00002225, 0.00004388, -0.00000257, 0.00001426, -0.00003383, 0.00000481, 0.00000288, 0.00000077, -0.00000505, -0.00000240, -0.00000469, -0.00000395, 0.00000148, -0.00000179, 0.00000214, -0.00000275, -0.00000077, 0.00000057, -0.00000330, 0.00000245, 0.00000301, 0.00000023, -0.00000090, -0.00000278, 0.00000538, -0.00000091, 0.00000208, -0.00000069, 0.00001215, -0.00001395, 0.00000701, -0.00000954, 0.00000765, 0.00000144, -0.00001905, 0.00000885, 0.00000083, 0.00000599, -0.00003072, 0.00001910, -0.00000488, -0.00000049, -0.00000544, -0.00000501, 0.00001249, -0.00000878, 0.00000962, -0.00000598, 0.00000564, -0.00000752, -0.00000463, -0.00000413, 0.00001400, -0.00001290, -0.00000304, 0.00000229, 0.00001391, -0.00000587, 0.00000559, -0.00000435, -0.00000471, 0.00000193, -0.00000118, 0.00000145, 0.00000180, -0.00000138, -0.00000268, -0.00000242, 0.00000543, -0.00000419, -0.00000564, -0.00000767, -0.00000424, -0.00000353, -0.00000195, -0.00000060, -0.00000348, -0.00001035, -0.00003409, -0.00002479, -0.00016423, 0.00015378, 0.00038538, -0.00005443, -0.00013922, 0.00000188, -0.00001275, -0.00000413, -0.00002179, 0.00000133, -0.00001566, 0.00000634, -0.00000175, 0.00000200, -0.00000237, -0.00000248, -0.00000415, 0.00000321, 0.00000148, -0.00000089, -0.00000281, 0.00000025, -0.00000985, 0.00000536, -0.00000670, 0.00000659, 0.00000181, 0.00000149, -0.00001809, 0.00000965, 0.00000226, -0.00000237, -0.00000051, -0.00000335, 0.00000513, -0.00000515, 0.00000764, -0.00000222, -0.00000499, -0.00000403, 0.00000566, -0.00001281, 0.00001062, -0.00002471, },
		{ 0.00000826, -0.00000085, 0.00004631, -0.00002899, 0.00003073, 0.00001261, 0.00000854, -0.00001412, -0.00001825, 0.00001452, -0.00000534, 0.00000184, -0.00000824, 0.00000104, -0.00000886, 0.00000360, -0.00000361, 0.00000311, -0.00000225, -0.00000114, -0.00000011, -0.00000103, -0.00000205, 0.00000594, -0.00000127, 0.00000217, -0.00000560, 0.00000540, -0.00000043, 0.00000300, -0.00000334, 0.00001122, -0.00000490, -0.00000469, 0.00000368, -0.00000873, 0.00001751, -0.00002443, 0.00000534, 0.00000036, 0.00001062, -0.00002517, 0.00000118, 0.00001187, -0.00001152, 0.00000591, -0.00001605, 0.00001658, -0.00000864, 0.00000799, -0.00000338, 0.00000298, -0.00000291, -0.00000510, -0.00000775, 0.00001077, -0.00000123, -0.00001133, 0.00000181, 0.00000996, 0.00000389, -0.00000366, 0.00000581, -0.00001144, 0.00000430, -0.00000313, 0.00000263, -0.00000006, 0.00000195, -0.00000435, -0.00000135, 0.00000118, 0.00000294, -0.00000970, -0.00000297, -0.00000984, 0.00000109, -0.00000765, 0.00000456, -0.00000933, 0.00000345, -0.00004020, -0.00000664, -0.00012685, -0.00004195, 0.00042534, 0.00011085, -0.00018658, -0.00001961, -0.00001123, -0.00000097, -0.00002290, -0.00000204, -0.00001221, -0.00000142, 0.00000311, -0.00000096, 0.00000035, -0.00000290, -0.00000451, 0.00000119, 0.00000207, 0.00000137, -0.00000499, 0.00000289, -0.00000989, 0.00000164, -0.00000230, 0.00000032, 0.00000519, 0.00000262, -0.00001173, -0.00000536, 0.00001339, -0.00000750, 0.00000430, -0.00000793, 0.00000682, -0.00000482, 0.00000512, 0.00000099, -0.00000343, -0.00000776, 0.00000759, -0.00001087, 0.00000538, -0.00001117, -0.00001294, },
		{ -0.00001237, 0.00004359, -0.00000785, -0.00000033, 0.00003334, -0.00000143, 0.00000605, -0.00003085, 0.00001264, -0.00000155, 0.00000104, -0.00000681, -0.00000045, -0.00000635, -0.00000068, -0.00000025, -0.00000043, 0.00000143, -0.00000321, 0.00000033, -0.00000031, -0.00000312, 0.00000406, 0.00000155, 0.00000068, -0.00000261, -0.00000010, 0.00000436, -0.00000056, 0.00000092, 0.00000173, 0.00000938, -0.00001600, 0.00001175, -0.00001437, 0.00001662, -0.00001184, -0.00001000, 0.00000757, 0.00000502, -0.00000863, -0.00002043, 0.00002410, -0.00001567, 0.00000946, -0.00001655, 0.00000995, -0.00000060, 0.00000069, 0.00000342, -0.00000274, 0.00000272, -0.00000628, -0.00000643, 0.00000102, 0.00001078, -0.00001431, -0.00000109, 0.00000506, 0.00001142, -0.00000693, 0.00000734, -0.00000804, -0.00000166, 0.00000100, -0.00000095, 0.00000205, 0.00000077, -0.00000146, -0.00000358, 0.00000042, 0.00000394, -0.00000622, -0.00000583, -0.00000679, -0.00000371, -0.00000305, -0.00000158, -0.00000271, -0.00000198, -0.00001988, -0.00002993, -0.00004526, -0.00017645, 0.00031521, 0.00032281, -0.00018429, -0.00006554, -0.00000849, -0.00000135, -0.00001840, -0.00000989, -0.00000612, -0.00000912, 0.00000547, -0.00000156, 0.00000074, -0.00000157, -0.00000507, -0.00000060, 0.00000180, 0.00000296, -0.00000435, 0.00000084, -0.00000429, -0.00000608, 0.00000377, -0.00000533, 0.00000705, 0.00000184, -0.00000238, -0.00001650, 0.00001455, -0.00000326, 0.00000152, -0.00000466, 0.00000110, 0.00000103, -0.00000067, 0.00000517, -0.00000308, -0.00000731, 0.00000287, -0.00000193, -0.00000593, 0.00000468, -0.00002601, 0.00001516, },
		{ 0.00002109, 0.00002348, -0.00002550, 0.00003857, 0.00000182, 0.00001048, -0.00002271, -0.00000461, 0.00000987, -0.00000468, -0.00000094, -0.00000586, -0.00000068, -0.00000645, 0.00000343, -0.00000372, 0.00000389, -0.00000369, 0.00000001, -0.00000027, -0.00000185, -0.00000021, 0.00000534, -0.00000165, 0.00000141, -0.00000510, 0.00000712, -0.00000230, 0.00000410, -0.00000458, 0.00001442, -0.00001302, 0.00000366, -0.00000291, -0.00000044, 0.00001024, -0.00002544, 0.00001203, -0.00000047, 0.00000560, -0.00003134, 0.00002099, -0.00000662, 0.00000176, -0.00000622, -0.00000413, 0.00001040, -0.00000725, 0.00000955, -0.00000671, 0.00000549, -0.00000580, -0.00000386, -0.00000734, 0.00001430, -0.00000753, -0.00000784, 0.00000257, 0.00001194, -0.00000124, 0.00000046, 0.00000165, -0.00000984, 0.00000499, -0.00000386, 0.00000356, -0.00000069, 0.00000152, -0.00000511, 0.00000013, 0.00000219, -0.00000024, -0.00000991, -0.00000259, -0.00000899, 0.00000146, -0.00000709, 0.00000310, -0.00000910, -0.00000032, -0.00004720, -0.00000117, -0.00017997, 0.00008306, 0.00049022, -0.00008326, -0.00013852, -0.00000484, -0.00000852, -0.00000745, -0.00002042, -0.00000041, -0.00001396, 0.00000405, 0.00000005, -0.00000022, -0.00000006, -0.00000453, -0.00000292, 0.00000175, 0.00000261, -0.00000085, -0.00000366, 0.00000194, -0.00001093, 0.00000465, -0.00000486, 0.00000367, 0.00000349, 0.00000229, -0.00001599, 0.00000308, 0.00000830, -0.00000533, 0.00000213, -0.00000637, 0.00000634, -0.00000458, 0.00000703, -0.00000238, -0.00000445, -0.00000446, 0.00000715, -0.00001398, 0.00001219, -0.00002586, 0.00000748, -0.00000546, },
		{ 0.00004261, -0.00002698, 0.00002103, 0.00001857, 0.00000287, -0.00000451, -0.00002162, 0.00001463, -0.00000429, 0.00000016, -0.00000663, -0.00000005, -0.00000635, 0.00000132, -0.00000174, 0.00000122, -0.00000001, -0.00000284, 0.00000103, -0.00000142, -0.00000220, 0.00000505, 0.00000007, 0.00000109, -0.00000406, 0.00000311, 0.00000198, 0.00000108, -0.00000166, 0.00000627, 0.00000279, -0.00001317, 0.00001272, -0.00001570, 0.00002191, -0.00002462, 0.00000295, 0.00000307, 0.00000810, -0.00002493, 0.00000207, 0.00001158, -0.00001124, 0.00000716, -0.00001668, 0.00001642, -0.00000972, 0.00001017, -0.00000528, 0.00000335, -0.00000254, -0.00000290, -0.00000978, 0.00000811, 0.00000447, -0.00001293, -0.00000038, 0.00000876, 0.00000661, -0.00000512, 0.00000652, -0.00000993, 0.00000088, -0.00000007, -0.00000051, 0.00000246, -0.00000047, -0.00000135, -0.00000408, 0.00000350, 0.00000056, -0.00000677, -0.00000619, -0.00000520, -0.00000420, -0.00000136, -0.00000350, -0.00000253, -0.00000279, -0.00003041, -0.00002000, -0.00008538, -0.00014719, 0.00049614, 0.00013954, -0.00021800, -0.00000856, -0.00002140, 0.00000694, -0.00003000, 0.00000258, -0.00001573, 0.00000096, 0.00000064, 0.00000069, -0.00000124, -0.00000111, -0.00000669, 0.00000277, 0.00000025, 0.00000374, -0.00000704, 0.00000431, -0.00000939, -0.00000070, 0.00000029, -0.00000214, 0.00000590, 0.00000269, -0.00000774, -0.00001121, 0.00001548, -0.00000626, 0.00000367, -0.00000751, 0.00000467, -0.00000234, 0.00000412, 0.00000066, -0.00000271, -0.00000850, 0.00000925, -0.00001198, 0.00000680, -0.00001223, -0.00001031, 0.00000957, -0.00000307, },
		{ -0.00000613, -0.00000670, 0.00003282, -0.00000051, 0.00000573, -0.00002281, 0.00000401, 0.00000589, -0.00000542, -0.00000143, -0.00000511, -0.00000091, -0.00000490, 0.00000324, -0.00000375, 0.00000462, -0.00000549, 0.00000215, -0.00000183, -0.00000130, 0.00000091, 0.00000475, -0.00000220, 0.00000078, -0.00000386, 0.00000772, -0.00000350, 0.00000440, -0.00000450, 0.00001566, -0.00002019, 0.00001372, -0.00001232, 0.00001177, -0.00000454, -0.00001671, 0.00001213, 0.00000224, -0.00000834, -0.00002062, 0.00002560, -0.00001749, 0.00001249, -0.00001842, 0.00001145, -0.00000346, 0.00000371, 0.00000167, -0.00000244, 0.00000215, -0.00000363, -0.00000710, -0.00000208, 0.00001281, -0.00001029, -0.00000648, 0.00000602, 0.00000995, -0.00000287, 0.00000215, -0.00000103, -0.00000894, 0.00000666, -0.00000624, 0.00000667, -0.00000410, 0.00000394, -0.00000864, 0.00000554, -0.00000193, 0.00000064, -0.00001336, 0.00000224, -0.00001318, 0.00000736, -0.00001310, 0.00000790, -0.00001465, -0.00000117, -0.00005522, 0.00000207, -0.00024061, 0.00029412, 0.00042114, -0.00023680, -0.00004687, -0.00002555, 0.00001115, -0.00002850, -0.00000298, -0.00001201, -0.00000414, 0.00000037, 0.00000265, -0.00000352, 0.00000266, -0.00000914, 0.00000252, -0.00000134, 0.00000588, -0.00000612, 0.00000146, -0.00000324, -0.00000750, 0.00000394, -0.00000447, 0.00000472, 0.00000348, -0.00000060, -0.00001677, 0.00000971, 0.00000282, -0.00000213, -0.00000145, -0.00000297, 0.00000378, -0.00000123, 0.00000485, -0.00000302, -0.00000713, 0.00000320, -0.00000142, -0.00000619, 0.00000523, -0.00002557, 0.00001785, -0.00001326, 0.00003909, },
		{ -0.00002511, 0.00003028, 0.00000891, 0.00000307, -0.00001010, -0.00001302, 0.00001499, -0.00000811, 0.00000136, -0.00000771, 0.00000155, -0.00000708, 0.00000371, -0.00000410, 0.00000412, -0.00000328, -0.00000024, -0.00000039, -0.00000095, -0.00000206, 0.00000651, -0.00000235, 0.00000242, -0.00000586, 0.00000704, -0.00000191, 0.00000429, -0.00000586, 0.00001311, -0.00000839, -0.00000325, 0.00000650, -0.00000956, 0.00001810, -0.00002962, 0.00001313, -0.00000011, 0.00000411, -0.00003138, 0.00002280, -0.00000854, 0.00000420, -0.00000702, -0.00000331, 0.00000853, -0.00000621, 0.00001004, -0.00000780, 0.00000533, -0.00000425, -0.00000280, -0.00000969, 0.00001248, -0.00000079, -0.00001227, 0.00000249, 0.00000954, 0.00000349, -0.00000411, 0.00000634, -0.00001282, 0.00000571, -0.00000443, 0.00000377, -0.00000140, 0.00000251, -0.00000562, 0.00000095, 0.00000066, 0.00000172, -0.00001183, -0.00000026, -0.00001047, 0.00000271, -0.00000791, 0.00000237, -0.00001014, 0.00000361, -0.00005126, 0.00000692, -0.00016613, -0.00002077, 0.00060677, -0.00010646, -0.00014382, -0.00000722, -0.00000828, -0.00000752, -0.00002136, -0.00000046, -0.00001351, 0.00000282, 0.00000084, -0.00000151, 0.00000121, -0.00000553, -0.00000267, 0.00000123, 0.00000278, -0.00000009, -0.00000478, 0.00000325, -0.00001077, 0.00000240, -0.00000182, 0.00000016, 0.00000542, 0.00000213, -0.00001207, -0.00000448, 0.00001360, -0.00000696, 0.00000372, -0.00000826, 0.00000623, -0.00000311, 0.00000614, -0.00000264, -0.00000368, -0.00000506, 0.00000874, -0.00001526, 0.00001372, -0.00002677, 0.00001073, -0.00000584, 0.00001862, 0.00002109, },
		{ 0.00001196, 0.00002317, -0.00000120, 0.00000075, -0.00001944, 0.00001003, 0.00000049, -0.00000444, -0.00000253, -0.00000360, -0.00000166, -0.00000270, 0.00000165, -0.00000203, 0.00000329, -0.00000532, 0.00000289, -0.00000266, -0.00000109, 0.00000269, 0.00000276, -0.00000137, -0.00000104, -0.00000062, 0.00000559, -0.00000203, 0.00000166, -0.00000029, 0.00001107, -0.00002114, 0.00001975, -0.00001905, 0.00002193, -0.00002090, -0.00000188, 0.00000718, 0.00000494, -0.00002444, 0.00000288, 0.00001150, -0.00001143, 0.00000914, -0.00001816, 0.00001722, -0.00001188, 0.00001325, -0.00000776, 0.00000410, -0.00000284, -0.00000016, -0.00001147, 0.00000519, 0.00000860, -0.00001156, -0.00000480, 0.00000924, 0.00000683, -0.00000337, 0.00000358, -0.00000459, -0.00000583, 0.00000573, -0.00000614, 0.00000744, -0.00000554, 0.00000432, -0.00000945, 0.00000874, -0.00000522, 0.00000000, -0.00001364, 0.00000407, -0.00001441, 0.00001022, -0.00001696, 0.00001034, -0.00001711, -0.00000996, -0.00004786, -0.00002578, -0.00025781, 0.00054616, 0.00019523, -0.00026597, 0.00001321, -0.00003999, 0.00002115, -0.00004184, 0.00001116, -0.00002272, 0.00000659, -0.00000491, 0.00000523, -0.00000561, 0.00000321, -0.00001106, 0.00000630, -0.00000325, 0.00000729, -0.00000975, 0.00000597, -0.00000889, -0.00000254, 0.00000160, -0.00000279, 0.00000485, 0.00000377, -0.00000474, -0.00001454, 0.00001389, -0.00000192, 0.00000086, -0.00000512, 0.00000084, 0.00000106, 0.00000269, 0.00000068, -0.00000250, -0.00000854, 0.00001013, -0.00001235, 0.00000739, -0.00001253, -0.00000820, 0.00001021, -0.00000376, 0.00003771, -0.00002387, },
		{ 0.00002847, 0.00000434, -0.00000007, -0.00001048, -0.00000660, 0.00001403, -0.00001193, 0.00000391, -0.00000976, 0.00000420, -0.00000848, 0.00000609, -0.00000643, 0.00000705, -0.00000702, 0.00000346, -0.00000317, 0.00000064, -0.00000226, 0.00000766, -0.00000471, 0.00000373, -0.00000690, 0.00001002, -0.00000534, 0.00000683, -0.00000895, 0.00001886, -0.00002051, 0.00001116, -0.00000572, 0.00000352, 0.00000428, -0.00002314, 0.00001579, 0.00000046, -0.00000880, -0.00002040, 0.00002714, -0.00001968, 0.00001605, -0.00002081, 0.00001346, -0.00000675, 0.00000685, -0.00000003, -0.00000212, 0.00000116, -0.00000068, -0.00000773, -0.00000419, 0.00001211, -0.00000348, -0.00001310, 0.00000795, 0.00000665, 0.00000312, -0.00000465, 0.00000700, -0.00001629, 0.00001194, -0.00001115, 0.00001104, -0.00000889, 0.00000921, -0.00001357, 0.00001082, -0.00000818, 0.00000789, -0.00002117, 0.00001140, -0.00002248, 0.00001733, -0.00002377, 0.00001784, -0.00002727, 0.00001721, -0.00007882, 0.00004238, -0.00027962, 0.00023275, 0.00054701, -0.00029797, -0.00002481, -0.00004494, 0.00002477, -0.00003927, 0.00000477, -0.00001892, 0.00000203, -0.00000593, 0.00000797, -0.00000881, 0.00000768, -0.00001377, 0.00000610, -0.00000480, 0.00000888, -0.00000793, 0.00000230, -0.00000289, -0.00000730, 0.00000184, -0.00000137, 0.00000062, 0.00000635, -0.00000062, -0.00001393, 0.00000194, 0.00001021, -0.00000613, 0.00000208, -0.00000711, 0.00000609, -0.00000145, 0.00000463, -0.00000345, -0.00000625, 0.00000282, -0.00000027, -0.00000710, 0.00000627, -0.00002529, 0.00001998, -0.00001386, 0.00003563, -0.00000629, -0.00000939, },
		{ 0.00001701, -0.00000531, 0.00000079, -0.00001770, 0.00001626, -0.00000810, 0.00000079, -0.00000702, 0.00000118, -0.00000521, 0.00000185, -0.00000254, 0.00000253, -0.00000132, -0.00000165, 0.00000044, -0.00000100, -0.00000262, 0.00000608, -0.00000143, 0.00000172, -0.00000468, 0.00000491, 0.00000039, 0.00000252, -0.00000450, 0.00000864, -0.00000101, -0.00001218, 0.00001649, -0.00001727, 0.00002295, -0.00003025, 0.00001141, 0.00000224, 0.00000144, -0.00003077, 0.00002432, -0.00001024, 0.00000631, -0.00000733, -0.00000310, 0.00000743, -0.00000616, 0.00001142, -0.00000948, 0.00000542, -0.00000317, -0.00000131, -0.00001137, 0.00000941, 0.00000577, -0.00001477, 0.00000077, 0.00000810, 0.00000668, -0.00000636, 0.00000798, -0.00001213, 0.00000290, -0.00000186, 0.00000109, 0.00000066, 0.00000056, -0.00000316, -0.00000112, 0.00000221, 0.00000016, -0.00000971, -0.00000248, -0.00000682, -0.00000169, -0.00000244, -0.00000492, -0.00000395, -0.00000194, -0.00004227, -0.00000470, -0.00012166, -0.00015120, 0.00072319, -0.00011376, -0.00016183, -0.00000051, -0.00001562, -0.00000167, -0.00002672, 0.00000307, -0.00001607, 0.00000431, -0.00000094, -0.00000038, 0.00000008, -0.00000425, -0.00000442, 0.00000254, 0.00000131, 0.00000185, -0.00000651, 0.00000453, -0.00000995, -0.00000042, 0.00000115, -0.00000266, 0.00000649, 0.00000207, -0.00000773, -0.00001113, 0.00001635, -0.00000598, 0.00000332, -0.00000825, 0.00000431, -0.00000048, 0.00000473, -0.00000268, -0.00000313, -0.00000533, 0.00000990, -0.00001612, 0.00001471, -0.00002707, 0.00001331, -0.00000661, 0.00001795, 0.00001652, -0.00002187, 0.00002027, },
		{ -0.00000155, 0.00000009, -0.00001140, 0.00000122, 0.00000817, -0.00001107, 0.00000329, -0.00000881, 0.00000432, -0.00000755, 0.00000572, -0.00000582, 0.00000688, -0.00000791, 0.00000495, -0.00000438, 0.00000087, -0.00000061, 0.00000619, -0.00000453, 0.00000267, -0.00000484, 0.00000933, -0.00000524, 0.00000531, -0.00000675, 0.00001793, -0.00002647, 0.00002276, -0.00001743, 0.00001715, -0.00001366, -0.00000841, 0.00001193, 0.00000172, -0.00002399, 0.00000364, 0.00001176, -0.00001226, 0.00001195, -0.00002056, 0.00001905, -0.00001506, 0.00001706, -0.00001068, 0.00000523, -0.00000389, 0.00000325, -0.00001325, 0.00000301, 0.00000997, -0.00000657, -0.00001165, 0.00001169, 0.00000418, 0.00000174, -0.00000287, 0.00000403, -0.00001500, 0.00001341, -0.00001351, 0.00001421, -0.00001262, 0.00001199, -0.00001687, 0.00001638, -0.00001388, 0.00000996, -0.00002449, 0.00001695, -0.00002833, 0.00002566, -0.00003423, 0.00002767, -0.00003756, 0.00001826, -0.00008522, 0.00004243, -0.00035929, 0.00056201, 0.00028455, -0.00033231, 0.00004616, -0.00006700, 0.00004140, -0.00005823, 0.00002371, -0.00003335, 0.00001570, -0.00001378, 0.00001285, -0.00001290, 0.00001011, -0.00001763, 0.00001176, -0.00000841, 0.00001203, -0.00001331, 0.00000828, -0.00000918, -0.00000278, 0.00000050, -0.00000076, 0.00000141, 0.00000657, -0.00000373, -0.00001430, 0.00000816, 0.00000535, -0.00000379, -0.00000109, -0.00000430, 0.00000506, 0.00000094, 0.00000118, -0.00000309, -0.00000757, 0.00000991, -0.00001167, 0.00000693, -0.00001190, -0.00000673, 0.00001044, -0.00000353, 0.00003268, -0.00002123, 0.00000531, 0.00002454, },
		{ -0.00000966, 0.00000346, -0.00001647, 0.00002049, -0.00001659, 0.00000762, -0.00001293, 0.00000760, -0.00001020, 0.00000727, -0.00000753, 0.00000791, -0.00000715, 0.00000373, -0.00000378, 0.00000202, -0.00000465, 0.00000944, -0.00000602, 0.00000532, -0.00000827, 0.00001036, -0.00000527, 0.00000731, -0.00001065, 0.00001825, -0.00001666, 0.00000442, 0.00000417, -0.00000620, 0.00001256, -0.00002760, 0.00001742, 0.00000038, -0.00001036, -0.00001959, 0.00002851, -0.00002193, 0.00001967, -0.00002323, 0.00001544, -0.00000984, 0.00000949, -0.00000119, -0.00000207, 0.00000002, 0.00000229, -0.00000831, -0.00000520, 0.00000901, 0.00000500, -0.00001956, 0.00000965, 0.00000276, 0.00000947, -0.00001122, 0.00001400, -0.00002168, 0.00001503, -0.00001406, 0.00001364, -0.00001207, 0.00001277, -0.00001680, 0.00001459, -0.00001297, 0.00001343, -0.00002691, 0.00001808, -0.00002879, 0.00002383, -0.00003038, 0.00002361, -0.00003561, 0.00002981, -0.00009353, 0.00006628, -0.00028464, 0.00012914, 0.00069458, -0.00036011, -0.00000538, -0.00006197, 0.00003574, -0.00004781, 0.00001110, -0.00002489, 0.00000761, -0.00001177, 0.00001283, -0.00001366, 0.00001212, -0.00001770, 0.00000905, -0.00000767, 0.00001125, -0.00000931, 0.00000314, -0.00000321, -0.00000553, -0.00000212, 0.00000325, -0.00000440, 0.00000973, -0.00000188, -0.00000887, -0.00000722, 0.00001715, -0.00000902, 0.00000479, -0.00001032, 0.00000725, -0.00000091, 0.00000424, -0.00000413, -0.00000498, 0.00000207, 0.00000115, -0.00000826, 0.00000739, -0.00002493, 0.00002155, -0.00001450, 0.00003372, -0.00000889, -0.00000822, 0.00002050, 0.00001250, },
		{ 0.00000713, -0.00001677, 0.00001263, -0.00000415, -0.00000288, -0.00000359, -0.00000191, -0.00000088, -0.00000198, 0.00000075, -0.00000061, 0.00000187, -0.00000396, 0.00000209, -0.00000200, -0.00000182, 0.00000404, 0.00000099, -0.00000055, -0.00000182, 0.00000113, 0.00000405, -0.00000059, -0.00000132, 0.00000227, 0.00000734, -0.00002088, 0.00002449, -0.00002127, 0.00002319, -0.00002656, 0.00000676, 0.00000643, -0.00000222, -0.00002955, 0.00002539, -0.00001138, 0.00000765, -0.00000673, -0.00000393, 0.00000753, -0.00000744, 0.00001389, -0.00001189, 0.00000602, -0.00000287, 0.00000091, -0.00001293, 0.00000627, 0.00001045, -0.00001388, -0.00000362, 0.00000873, 0.00000702, -0.00000501, 0.00000543, -0.00000698, -0.00000387, 0.00000411, -0.00000476, 0.00000588, -0.00000475, 0.00000272, -0.00000664, 0.00000755, -0.00000576, -0.00000270, -0.00001009, 0.00000277, -0.00001250, 0.00000999, -0.00001950, 0.00001049, -0.00001824, -0.00001904, -0.00003605, -0.00005103, -0.00029622, 0.00082433, -0.00009405, -0.00019891, 0.00001940, -0.00003347, 0.00001221, -0.00003822, 0.00001176, -0.00002316, 0.00001002, -0.00000668, 0.00000444, -0.00000459, 0.00000027, -0.00000901, 0.00000635, -0.00000238, 0.00000549, -0.00000936, 0.00000640, -0.00000945, -0.00000241, 0.00000251, -0.00000337, 0.00000549, 0.00000328, -0.00000456, -0.00001501, 0.00001508, -0.00000154, 0.00000043, -0.00000597, 0.00000037, 0.00000337, 0.00000265, -0.00000215, -0.00000330, -0.00000471, 0.00001006, -0.00001598, 0.00001465, -0.00002638, 0.00001498, -0.00000757, 0.00001864, 0.00001061, -0.00001723, 0.00001048, 0.00002385, -0.00001175, },
		{ 0.00000119, -0.00000996, 0.00001690, -0.00001807, 0.00000937, -0.00001430, 0.00001023, -0.00001194, 0.00000929, -0.00000908, 0.00000980, -0.00000981, 0.00000656, -0.00000598, 0.00000303, -0.00000404, 0.00000943, -0.00000742, 0.00000600, -0.00000836, 0.00001198, -0.00000732, 0.00000789, -0.00001142, 0.00002166, -0.00002762, 0.00002069, -0.00001069, 0.00000836, -0.00000427, -0.00001523, 0.00001620, -0.00000077, -0.00002399, 0.00000451, 0.00001231, -0.00001366, 0.00001539, -0.00002365, 0.00002160, -0.00001886, 0.00002115, -0.00001370, 0.00000660, -0.00000564, 0.00000728, -0.00001537, 0.00000222, 0.00000809, 0.00000179, -0.00002027, 0.00001558, -0.00000079, 0.00000939, -0.00001164, 0.00001442, -0.00002496, 0.00002139, -0.00002117, 0.00002142, -0.00002036, 0.00002034, -0.00002503, 0.00002511, -0.00002393, 0.00002142, -0.00003681, 0.00003124, -0.00004349, 0.00004222, -0.00005236, 0.00004626, -0.00006026, 0.00004898, -0.00012444, 0.00010746, -0.00043687, 0.00053299, 0.00040992, -0.00041496, 0.00008776, -0.00010009, 0.00006565, -0.00007758, 0.00003906, -0.00004669, 0.00002751, -0.00002521, 0.00002284, -0.00002239, 0.00001886, -0.00002569, 0.00001854, -0.00001472, 0.00001763, -0.00001759, 0.00001144, -0.00001078, -0.00000077, -0.00000346, 0.00000411, -0.00000439, 0.00001113, -0.00000502, -0.00001035, -0.00000113, 0.00001449, -0.00000929, 0.00000375, -0.00000996, 0.00000903, -0.00000081, 0.00000212, -0.00000449, -0.00000556, 0.00000861, -0.00000996, 0.00000547, -0.00001046, -0.00000587, 0.00001047, -0.00000300, 0.00002851, -0.00002037, 0.00000237, 0.00002158, 0.00000272, -0.00000640, },
		{ -0.00002276, 0.00002316, -0.00001781, 0.00000846, -0.00001352, 0.00000830, -0.00000929, 0.00000618, -0.00000656, 0.00000683, -0.00000562, 0.00000274, -0.00000338, 0.00000240, -0.00000569, 0.00000953, -0.00000560, 0.00000507, -0.00000752, 0.00000844, -0.00000308, 0.00000565, -0.00000950, 0.00001413, -0.00000952, -0.00000491, 0.00001505, -0.00001483, 0.00001803, -0.00002852, 0.00001609, 0.00000250, -0.00001322, -0.00001800, 0.00002943, -0.00002375, 0.00002275, -0.00002504, 0.00001673, -0.00001205, 0.00001102, -0.00000136, -0.00000263, -0.00000091, 0.00000486, -0.00000851, -0.00000547, 0.00000452, 0.00001338, -0.00002406, 0.00000963, -0.00000012, 0.00001432, -0.00001552, 0.00001791, -0.00002327, 0.00001442, -0.00001366, 0.00001317, -0.00001229, 0.00001323, -0.00001689, 0.00001529, -0.00001450, 0.00001523, -0.00002836, 0.00001991, -0.00002959, 0.00002422, -0.00003014, 0.00002216, -0.00003595, 0.00003190, -0.00009354, 0.00006724, -0.00025194, -0.00001215, 0.00085240, -0.00041206, 0.00000346, -0.00007080, 0.00003956, -0.00005064, 0.00001319, -0.00002747, 0.00001036, -0.00001502, 0.00001525, -0.00001621, 0.00001422, -0.00001941, 0.00001005, -0.00000880, 0.00001204, -0.00000953, 0.00000341, -0.00000367, -0.00000295, -0.00000679, 0.00000793, -0.00000887, 0.00001237, -0.00000318, -0.00000318, -0.00001565, 0.00002163, -0.00000934, 0.00000557, -0.00001168, 0.00000658, 0.00000078, 0.00000337, -0.00000464, -0.00000385, 0.00000155, 0.00000223, -0.00000905, 0.00000803, -0.00002404, 0.00002241, -0.00001524, 0.00003341, -0.00001370, -0.00000392, 0.00001025, 0.00002212, -0.00002024, 0.00001572, },
		{ 0.00000444, 0.00000306, -0.00000893, 0.00000190, -0.00000702, 0.00000506, -0.00000693, 0.00000496, -0.00000451, 0.00000554, -0.00000653, 0.00000408, -0.00000361, -0.00000002, 0.00000090, 0.00000433, -0.00000368, 0.00000191, -0.00000329, 0.00000797, -0.00000389, 0.00000233, -0.00000419, 0.00001449, -0.00002698, 0.00002823, -0.00001994, 0.00001819, -0.00001878, -0.00000021, 0.00001179, -0.00000632, -0.00002798, 0.00002596, -0.00001173, 0.00000794, -0.00000500, -0.00000599, 0.00000899, -0.00001007, 0.00001739, -0.00001498, 0.00000725, -0.00000362, 0.00000412, -0.00001493, 0.00000427, 0.00001188, -0.00000876, -0.00001108, 0.00001192, 0.00000393, 0.00000035, -0.00000140, 0.00000233, -0.00001398, 0.00001285, -0.00001322, 0.00001379, -0.00001301, 0.00001163, -0.00001532, 0.00001648, -0.00001583, 0.00000891, -0.00002266, 0.00001769, -0.00002893, 0.00002841, -0.00004029, 0.00003214, -0.00004400, 0.00001623, -0.00008279, 0.00003625, -0.00043988, 0.00089397, -0.00003757, -0.00025933, 0.00005467, -0.00006306, 0.00003486, -0.00005650, 0.00002632, -0.00003556, 0.00002070, -0.00001708, 0.00001359, -0.00001334, 0.00000845, -0.00001673, 0.00001297, -0.00000854, 0.00001111, -0.00001376, 0.00000954, -0.00001030, -0.00000225, 0.00000092, -0.00000087, 0.00000157, 0.00000667, -0.00000380, -0.00001482, 0.00000910, 0.00000636, -0.00000479, -0.00000158, -0.00000533, 0.00000822, -0.00000010, -0.00000082, -0.00000454, -0.00000281, 0.00000883, -0.00001447, 0.00001322, -0.00002448, 0.00001559, -0.00000844, 0.00002001, 0.00000453, -0.00001292, 0.00000293, 0.00002725, -0.00001392, 0.00000720, -0.00000974, },
		{ 0.00002427, -0.00002347, 0.00001407, -0.00001851, 0.00001455, -0.00001457, 0.00001125, -0.00001081, 0.00001116, -0.00001030, 0.00000715, -0.00000688, 0.00000467, -0.00000672, 0.00001157, -0.00000909, 0.00000796, -0.00001014, 0.00001256, -0.00000734, 0.00000839, -0.00001321, 0.00002133, -0.00002411, 0.00001378, -0.00000004, -0.00000249, 0.00000516, -0.00002052, 0.00001869, -0.00000169, -0.00002489, 0.00000573, 0.00001291, -0.00001525, 0.00001896, -0.00002687, 0.00002428, -0.00002263, 0.00002487, -0.00001631, 0.00000787, -0.00000779, 0.00001166, -0.00001777, 0.00000292, 0.00000333, 0.00001232, -0.00002919, 0.00001964, -0.00000675, 0.00001791, -0.00002078, 0.00002437, -0.00003349, 0.00002767, -0.00002733, 0.00002738, -0.00002705, 0.00002760, -0.00003214, 0.00003304, -0.00003328, 0.00003200, -0.00004797, 0.00004403, -0.00005669, 0.00005640, -0.00006761, 0.00006199, -0.00008023, 0.00007575, -0.00015682, 0.00015756, -0.00047833, 0.00045357, 0.00056795, -0.00050730, 0.00013220, -0.00013456, 0.00009005, -0.00009686, 0.00005482, -0.00006072, 0.00004017, -0.00003747, 0.00003354, -0.00003251, 0.00002798, -0.00003391, 0.00002549, -0.00002119, 0.00002330, -0.00002207, 0.00001517, -0.00001364, 0.00000342, -0.00000989, 0.00001107, -0.00001168, 0.00001672, -0.00000806, -0.00000359, -0.00001240, 0.00002368, -0.00001412, 0.00000818, -0.00001509, 0.00001217, -0.00000209, 0.00000318, -0.00000644, -0.00000285, 0.00000660, -0.00000763, 0.00000344, -0.00000862, -0.00000531, 0.00001031, -0.00000258, 0.00002585, -0.00002198, 0.00000353, 0.00001440, 0.00001100, -0.00001701, 0.00001567, -0.00002078, },
		{ -0.00001490, 0.00000600, -0.00001115, 0.00000602, -0.00000569, 0.00000288, -0.00000366, 0.00000398, -0.00000267, 0.00000070, -0.00000203, 0.00000170, -0.00000525, 0.00000787, -0.00000350, 0.00000306, -0.00000487, 0.00000464, 0.00000069, 0.00000242, -0.00000627, 0.00000775, -0.00000092, -0.00001444, 0.00002412, -0.00001982, 0.00001887, -0.00002491, 0.00001147, 0.00000681, -0.00001729, -0.00001560, 0.00002964, -0.00002469, 0.00002470, -0.00002566, 0.00001675, -0.00001279, 0.00001096, -0.00000020, -0.00000408, -0.00000122, 0.00000653, -0.00000787, -0.00000574, 0.00000010, 0.00001970, -0.00002488, 0.00000659, -0.00000063, 0.00001606, -0.00001593, 0.00001724, -0.00001990, 0.00000930, -0.00000930, 0.00000897, -0.00000880, 0.00000975, -0.00001294, 0.00001188, -0.00001152, 0.00001189, -0.00002405, 0.00001535, -0.00002335, 0.00001693, -0.00002149, 0.00001177, -0.00002610, 0.00002083, -0.00007598, 0.00004312, -0.00018394, -0.00018005, 0.00100448, -0.00044075, -0.00000667, -0.00006564, 0.00003199, -0.00004459, 0.00000842, -0.00002426, 0.00000800, -0.00001356, 0.00001325, -0.00001464, 0.00001236, -0.00001746, 0.00000786, -0.00000710, 0.00001032, -0.00000776, 0.00000230, -0.00000336, -0.00000080, -0.00001051, 0.00001090, -0.00001112, 0.00001281, -0.00000312, 0.00000132, -0.00002123, 0.00002190, -0.00000602, 0.00000369, -0.00001060, 0.00000371, 0.00000388, 0.00000171, -0.00000449, -0.00000348, 0.00000193, 0.00000231, -0.00000880, 0.00000759, -0.00002220, 0.00002230, -0.00001588, 0.00003424, -0.00001989, 0.00000201, -0.00000023, 0.00003068, -0.00002868, 0.00002329, -0.00002342, 0.00002080, },
		{ -0.00001576, 0.00000807, -0.00001278, 0.00001124, -0.00001169, 0.00000889, -0.00000808, 0.00000867, -0.00000853, 0.00000573, -0.00000522, 0.00000210, -0.00000250, 0.00000762, -0.00000668, 0.00000538, -0.00000710, 0.00001083, -0.00000610, 0.00000494, -0.00000892, 0.00001853, -0.00002867, 0.00002641, -0.00001296, 0.00000859, -0.00000816, -0.00000811, 0.00001720, -0.00001005, -0.00002647, 0.00002619, -0.00001130, 0.00000719, -0.00000227, -0.00000907, 0.00001156, -0.00001371, 0.00002151, -0.00001846, 0.00000902, -0.00000545, 0.00000841, -0.00001776, 0.00000419, 0.00000943, 0.00000045, -0.00002104, 0.00001723, -0.00000216, 0.00000905, -0.00001148, 0.00001438, -0.00002585, 0.00002284, -0.00002296, 0.00002315, -0.00002299, 0.00002236, -0.00002596, 0.00002781, -0.00002873, 0.00002359, -0.00003844, 0.00003590, -0.00004866, 0.00005022, -0.00006448, 0.00005796, -0.00007548, 0.00005837, -0.00013706, 0.00012732, -0.00056486, 0.00091773, 0.00006158, -0.00034351, 0.00010435, -0.00010315, 0.00006503, -0.00008061, 0.00004614, -0.00005283, 0.00003601, -0.00003178, 0.00002668, -0.00002575, 0.00001981, -0.00002717, 0.00002202, -0.00001691, 0.00001856, -0.00001976, 0.00001431, -0.00001321, 0.00000093, -0.00000426, 0.00000518, -0.00000542, 0.00001247, -0.00000595, -0.00001024, -0.00000117, 0.00001679, -0.00001143, 0.00000418, -0.00001211, 0.00001351, -0.00000325, 0.00000131, -0.00000695, 0.00000046, 0.00000616, -0.00001156, 0.00001045, -0.00002145, 0.00001519, -0.00000899, 0.00002136, -0.00000059, -0.00001041, -0.00000083, 0.00002587, -0.00001077, 0.00000098, 0.00000224, -0.00000663, 0.00001179, },
		{ 0.00001600, -0.00002020, 0.00001617, -0.00001453, 0.00001087, -0.00001041, 0.00001049, -0.00000906, 0.00000642, -0.00000672, 0.00000527, -0.00000801, 0.00001196, -0.00000896, 0.00000795, -0.00000965, 0.00001067, -0.00000504, 0.00000652, -0.00001188, 0.00001712, -0.00001672, 0.00000359, 0.00001214, -0.00001275, 0.00001223, -0.00002254, 0.00001832, -0.00000042, -0.00002699, 0.00000757, 0.00001320, -0.00001651, 0.00002198, -0.00002949, 0.00002632, -0.00002555, 0.00002747, -0.00001792, 0.00000861, -0.00000988, 0.00001583, -0.00002002, 0.00000461, -0.00000312, 0.00002308, -0.00003639, 0.00002217, -0.00001194, 0.00002525, -0.00002801, 0.00003152, -0.00003844, 0.00003043, -0.00003035, 0.00003046, -0.00003100, 0.00003202, -0.00003641, 0.00003819, -0.00003969, 0.00003919, -0.00005521, 0.00005232, -0.00006474, 0.00006482, -0.00007636, 0.00007087, -0.00009259, 0.00009241, -0.00017460, 0.00018341, -0.00047644, 0.00032515, 0.00074888, -0.00059831, 0.00017099, -0.00016386, 0.00010945, -0.00011207, 0.00006768, -0.00007251, 0.00005102, -0.00004805, 0.00004263, -0.00004108, 0.00003544, -0.00004050, 0.00003104, -0.00002646, 0.00002789, -0.00002584, 0.00001875, -0.00001708, 0.00000890, -0.00001748, 0.00001856, -0.00001886, 0.00002195, -0.00001156, 0.00000429, -0.00002340, 0.00003075, -0.00001666, 0.00001096, -0.00001862, 0.00001368, -0.00000239, 0.00000396, -0.00000842, -0.00000005, 0.00000454, -0.00000534, 0.00000153, -0.00000701, -0.00000460, 0.00000979, -0.00000238, 0.00002480, -0.00002597, 0.00000815, 0.00000427, 0.00002135, -0.00002885, 0.00002871, -0.00002835, 0.00002805, -0.00002556, },
		{ -0.00000635, 0.00000135, -0.00000011, -0.00000185, 0.00000043, 0.00000013, 0.00000096, -0.00000178, -0.00000018, 0.00000024, -0.00000364, 0.00000493, -0.00000028, -0.00000004, -0.00000111, -0.00000002, 0.00000491, -0.00000120, -0.00000237, 0.00000101, 0.00000683, -0.00002156, 0.00002887, -0.00001931, 0.00001418, -0.00001677, 0.00000402, 0.00001277, -0.00002211, -0.00001257, 0.00002902, -0.00002443, 0.00002511, -0.00002473, 0.00001517, -0.00001179, 0.00000912, 0.00000236, -0.00000652, -0.00000063, 0.00000686, -0.00000589, -0.00000681, -0.00000283, 0.00002233, -0.00002092, -0.00000011, 0.00000199, 0.00001384, -0.00001171, 0.00001154, -0.00001153, -0.00000004, -0.00000128, 0.00000126, -0.00000171, 0.00000237, -0.00000494, 0.00000421, -0.00000377, 0.00000313, -0.00001373, 0.00000425, -0.00001001, 0.00000206, -0.00000465, -0.00000735, -0.00000609, -0.00000331, -0.00004164, -0.00000325, -0.00008907, -0.00035825, 0.00113247, -0.00043352, -0.00004267, -0.00004231, 0.00001021, -0.00002762, -0.00000493, -0.00001363, -0.00000103, -0.00000588, 0.00000548, -0.00000771, 0.00000545, -0.00001097, 0.00000172, -0.00000188, 0.00000542, -0.00000337, -0.00000093, -0.00000129, -0.00000046, -0.00001164, 0.00001050, -0.00000975, 0.00000990, -0.00000051, 0.00000310, -0.00002237, 0.00001700, 0.00000122, -0.00000092, -0.00000700, -0.00000132, 0.00000830, -0.00000087, -0.00000332, -0.00000437, 0.00000373, 0.00000086, -0.00000701, 0.00000565, -0.00001910, 0.00002101, -0.00001612, 0.00003551, -0.00002620, 0.00000774, -0.00000869, 0.00003576, -0.00003244, 0.00002432, -0.00001773, 0.00001350, -0.00000835, 0.00000082, },
		{ -0.00001760, 0.00001611, -0.00001489, 0.00001137, -0.00001025, 0.00001035, -0.00000920, 0.00000643, -0.00000620, 0.00000377, -0.00000524, 0.00000989, -0.00000852, 0.00000751, -0.00000915, 0.00001157, -0.00000616, 0.00000537, -0.00001066, 0.00001833, -0.00002527, 0.00001909, -0.00000144, -0.00000374, 0.00000319, -0.00001510, 0.00002127, -0.00001249, -0.00002553, 0.00002635, -0.00001035, 0.00000577, 0.00000097, -0.00001264, 0.00001464, -0.00001769, 0.00002559, -0.00002183, 0.00001105, -0.00000810, 0.00001352, -0.00002138, 0.00000616, 0.00000344, 0.00001257, -0.00003200, 0.00002338, -0.00000995, 0.00001944, -0.00002284, 0.00002692, -0.00003720, 0.00003204, -0.00003210, 0.00003220, -0.00003293, 0.00003309, -0.00003674, 0.00003959, -0.00004228, 0.00003886, -0.00005467, 0.00005435, -0.00006831, 0.00007173, -0.00008809, 0.00008354, -0.00010730, 0.00010027, -0.00018906, 0.00020833, -0.00065548, 0.00088591, 0.00020353, -0.00044697, 0.00016377, -0.00014967, 0.00009932, -0.00010786, 0.00006905, -0.00007313, 0.00005426, -0.00004919, 0.00004222, -0.00004038, 0.00003298, -0.00003907, 0.00003243, -0.00002657, 0.00002714, -0.00002693, 0.00002053, -0.00001824, 0.00000715, -0.00001275, 0.00001415, -0.00001470, 0.00002002, -0.00001056, -0.00000207, -0.00001422, 0.00002798, -0.00001803, 0.00001013, -0.00001890, 0.00001844, -0.00000634, 0.00000396, -0.00001027, 0.00000479, 0.00000239, -0.00000761, 0.00000675, -0.00001770, 0.00001404, -0.00000919, 0.00002223, -0.00000399, -0.00001059, -0.00000019, 0.00001961, -0.00000280, -0.00000989, 0.00001661, -0.00001792, 0.00001966, -0.00002140, 0.00001323, },
		{ 0.00001465, -0.00001161, 0.00000811, -0.00000790, 0.00000789, -0.00000625, 0.00000451, -0.00000550, 0.00000468, -0.00000771, 0.00001047, -0.00000698, 0.00000599, -0.00000700, 0.00000664, -0.00000091, 0.00000282, -0.00000818, 0.00001025, -0.00000728, -0.00000736, 0.00002289, -0.00001964, 0.00001484, -0.00002004, 0.00001451, 0.00000324, -0.00003035, 0.00001017, 0.00001283, -0.00001687, 0.00002373, -0.00003079, 0.00002699, -0.00002689, 0.00002834, -0.00001807, 0.00000841, -0.00001137, 0.00001914, -0.00002143, 0.00000633, -0.00000959, 0.00003184, -0.00003989, 0.00002159, -0.00001469, 0.00002951, -0.00003133, 0.00003401, -0.00003826, 0.00002847, -0.00002918, 0.00002960, -0.00003105, 0.00003232, -0.00003645, 0.00003903, -0.00004137, 0.00004100, -0.00005640, 0.00005385, -0.00006527, 0.00006501, -0.00007605, 0.00007000, -0.00009376, 0.00009456, -0.00017271, 0.00018008, -0.00043050, 0.00015653, 0.00093718, -0.00067393, 0.00019425, -0.00018087, 0.00011845, -0.00011901, 0.00007415, -0.00007892, 0.00005711, -0.00005421, 0.00004756, -0.00004575, 0.00003912, -0.00004358, 0.00003355, -0.00002906, 0.00003011, -0.00002774, 0.00002108, -0.00001993, 0.00001416, -0.00002435, 0.00002455, -0.00002403, 0.00002513, -0.00001387, 0.00001123, -0.00003176, 0.00003370, -0.00001559, 0.00001112, -0.00001973, 0.00001297, -0.00000128, 0.00000397, -0.00000982, 0.00000207, 0.00000324, -0.00000390, 0.00000053, -0.00000633, -0.00000320, 0.00000860, -0.00000220, 0.00002498, -0.00003151, 0.00001474, -0.00000666, 0.00003110, -0.00003866, 0.00003718, -0.00003042, 0.00002680, -0.00002345, 0.00001456, -0.00001880, },
		{ 0.00000609, -0.00000674, 0.00000455, -0.00000365, 0.00000429, -0.00000386, 0.00000151, -0.00000139, -0.00000155, 0.00000155, 0.00000308, -0.00000318, 0.00000256, -0.00000423, 0.00000819, -0.00000382, 0.00000060, -0.00000415, 0.00001160, -0.00002420, 0.00002770, -0.00001274, 0.00000439, -0.00000517, -0.00000500, 0.00001931, -0.00002689, -0.00000931, 0.00002767, -0.00002291, 0.00002393, -0.00002229, 0.00001208, -0.00000918, 0.00000575, 0.00000603, -0.00000976, 0.00000093, 0.00000565, -0.00000231, -0.00000925, -0.00000328, 0.00002040, -0.00001208, -0.00001013, 0.00000750, 0.00000783, -0.00000326, 0.00000157, 0.00000067, -0.00001226, 0.00000912, -0.00000884, 0.00000801, -0.00000796, 0.00000619, -0.00000686, 0.00000792, -0.00001013, 0.00000148, -0.00001209, 0.00000887, -0.00001857, 0.00001833, -0.00003300, 0.00002168, -0.00003750, 0.00000508, -0.00006465, 0.00001967, -0.00052765, 0.00121866, -0.00038071, -0.00010799, 0.00000039, -0.00002612, 0.00000026, -0.00002696, 0.00000468, -0.00001707, 0.00000835, -0.00000834, 0.00000476, -0.00000659, 0.00000008, -0.00000836, 0.00000691, -0.00000274, 0.00000386, -0.00000667, 0.00000321, -0.00000292, -0.00000901, 0.00000578, -0.00000410, 0.00000321, 0.00000514, 0.00000138, -0.00001845, 0.00000709, 0.00001166, -0.00000756, -0.00000146, -0.00000797, 0.00001365, -0.00000426, -0.00000101, -0.00000675, 0.00000716, -0.00000229, -0.00000354, 0.00000212, -0.00001474, 0.00001854, -0.00001570, 0.00003647, -0.00003141, 0.00001158, -0.00001332, 0.00003571, -0.00003026, 0.00001857, -0.00000663, 0.00000275, 0.00000026, -0.00000564, -0.00000040, -0.00000441, },
		{ -0.00001550, 0.00001156, -0.00001032, 0.00001001, -0.00000812, 0.00000581, -0.00000613, 0.00000442, -0.00000661, 0.00001041, -0.00000852, 0.00000760, -0.00000880, 0.00000966, -0.00000371, 0.00000323, -0.00000906, 0.00001390, -0.00001739, 0.00000778, 0.00001223, -0.00001607, 0.00001273, -0.00001923, 0.00002273, -0.00001286, -0.00002559, 0.00002680, -0.00000929, 0.00000423, 0.00000399, -0.00001590, 0.00001739, -0.00002115, 0.00002885, -0.00002446, 0.00001286, -0.00001109, 0.00001888, -0.00002531, 0.00000959, -0.00000483, 0.00002554, -0.00004185, 0.00002853, -0.00001753, 0.00002934, -0.00003302, 0.00003740, -0.00004565, 0.00003836, -0.00003876, 0.00003908, -0.00004089, 0.00004181, -0.00004557, 0.00004957, -0.00005393, 0.00005188, -0.00006821, 0.00006961, -0.00008420, 0.00008896, -0.00010685, 0.00010408, -0.00013359, 0.00013448, -0.00022919, 0.00026710, -0.00070052, 0.00079575, 0.00038144, -0.00056003, 0.00022480, -0.00019611, 0.00013248, -0.00013410, 0.00009163, -0.00009343, 0.00007267, -0.00006669, 0.00005772, -0.00005489, 0.00004582, -0.00005052, 0.00004254, -0.00003608, 0.00003562, -0.00003426, 0.00002743, -0.00002467, 0.00001548, -0.00002322, 0.00002445, -0.00002465, 0.00002790, -0.00001630, 0.00000798, -0.00002773, 0.00003765, -0.00002286, 0.00001493, -0.00002454, 0.00002211, -0.00000876, 0.00000664, -0.00001396, 0.00000951, -0.00000175, -0.00000335, 0.00000283, -0.00001391, 0.00001267, -0.00000922, 0.00002246, -0.00000546, -0.00001350, 0.00000437, 0.00000963, 0.00000806, -0.00002253, 0.00003004, -0.00002649, 0.00002434, -0.00002403, 0.00001589, -0.00002009, 0.00001841, },
		{ 0.00000362, -0.00000394, 0.00000404, -0.00000254, 0.00000199, -0.00000362, 0.00000318, -0.00000610, 0.00000752, -0.00000367, 0.00000270, -0.00000301, 0.00000150, 0.00000393, -0.00000152, -0.00000353, 0.00000267, 0.00000175, -0.00001624, 0.00002940, -0.00002104, 0.00001182, -0.00001274, 0.00000752, 0.00000892, -0.00003464, 0.00001346, 0.00001157, -0.00001591, 0.00002369, -0.00003025, 0.00002581, -0.00002623, 0.00002714, -0.00001653, 0.00000700, -0.00001179, 0.00002093, -0.00002130, 0.00000703, -0.00001439, 0.00003665, -0.00003824, 0.00001692, -0.00001395, 0.00002946, -0.00002963, 0.00003095, -0.00003247, 0.00002163, -0.00002371, 0.00002459, -0.00002684, 0.00002805, -0.00003173, 0.00003482, -0.00003744, 0.00003646, -0.00005056, 0.00004766, -0.00005737, 0.00005613, -0.00006582, 0.00005840, -0.00008241, 0.00008067, -0.00014994, 0.00014794, -0.00034664, -0.00003700, 0.00111358, -0.00071926, 0.00019270, -0.00017938, 0.00011260, -0.00011431, 0.00007135, -0.00007729, 0.00005598, -0.00005364, 0.00004623, -0.00004461, 0.00003734, -0.00004170, 0.00003174, -0.00002782, 0.00002885, -0.00002672, 0.00002104, -0.00002084, 0.00001749, -0.00002849, 0.00002705, -0.00002543, 0.00002478, -0.00001351, 0.00001538, -0.00003552, 0.00003123, -0.00001035, 0.00000835, -0.00001810, 0.00000985, 0.00000138, 0.00000292, -0.00001010, 0.00000286, 0.00000339, -0.00000397, 0.00000106, -0.00000716, -0.00000068, 0.00000647, -0.00000170, 0.00002564, -0.00003733, 0.00002138, -0.00001596, 0.00003757, -0.00004367, 0.00003901, -0.00002603, 0.00002035, -0.00001740, 0.00001010, -0.00001466, 0.00001043, -0.00000641, },
		{ 0.00000744, -0.00000625, 0.00000632, -0.00000474, 0.00000236, -0.00000251, 0.00000022, -0.00000132, 0.00000556, -0.00000529, 0.00000498, -0.00000673, 0.00000935, -0.00000428, 0.00000136, -0.00000628, 0.00001203, -0.00002139, 0.00002048, -0.00000105, -0.00000875, 0.00000783, -0.00001371, 0.00002499, -0.00003067, -0.00000634, 0.00002591, -0.00002039, 0.00002145, -0.00001875, 0.00000797, -0.00000551, 0.00000147, 0.00001022, -0.00001339, 0.00000325, 0.00000302, 0.00000274, -0.00001317, -0.00000098, 0.00001413, 0.00000057, -0.00002210, 0.00001473, -0.00000078, 0.00000792, -0.00001083, 0.00001456, -0.00002516, 0.00001991, -0.00001958, 0.00001869, -0.00001959, 0.00001878, -0.00001963, 0.00002175, -0.00002586, 0.00001928, -0.00003107, 0.00003037, -0.00004175, 0.00004393, -0.00006134, 0.00005297, -0.00007647, 0.00005706, -0.00013092, 0.00012706, -0.00066956, 0.00124910, -0.00027814, -0.00020108, 0.00005973, -0.00007429, 0.00003667, -0.00005579, 0.00002922, -0.00003888, 0.00002798, -0.00002709, 0.00002168, -0.00002267, 0.00001463, -0.00002145, 0.00001848, -0.00001353, 0.00001346, -0.00001472, 0.00001022, -0.00000849, -0.00000237, -0.00000320, 0.00000541, -0.00000671, 0.00001338, -0.00000358, -0.00001007, -0.00000650, 0.00002366, -0.00001487, 0.00000496, -0.00001529, 0.00001919, -0.00000804, 0.00000225, -0.00001044, 0.00001200, -0.00000686, 0.00000132, -0.00000265, -0.00000947, 0.00001512, -0.00001456, 0.00003659, -0.00003464, 0.00001250, -0.00001327, 0.00003017, -0.00002248, 0.00000754, 0.00000748, -0.00000890, 0.00000862, -0.00001143, 0.00000501, -0.00000966, 0.00001135, -0.00001037, },
		{ -0.00000820, 0.00000764, -0.00000540, 0.00000395, -0.00000495, 0.00000383, -0.00000633, 0.00000898, -0.00000653, 0.00000557, -0.00000610, 0.00000538, 0.00000083, -0.00000104, -0.00000476, 0.00000639, -0.00000688, -0.00000492, 0.00002494, -0.00002544, 0.00001815, -0.00001902, 0.00002079, -0.00001075, -0.00002687, 0.00002782, -0.00000860, 0.00000326, 0.00000597, -0.00001801, 0.00001895, -0.00002323, 0.00003051, -0.00002576, 0.00001390, -0.00001375, 0.00002367, -0.00002870, 0.00001338, -0.00001352, 0.00003693, -0.00004834, 0.00003085, -0.00002301, 0.00003656, -0.00003974, 0.00004357, -0.00004927, 0.00004024, -0.00004152, 0.00004234, -0.00004532, 0.00004684, -0.00005064, 0.00005571, -0.00006137, 0.00006007, -0.00007630, 0.00007870, -0.00009317, 0.00009860, -0.00011719, 0.00011555, -0.00014939, 0.00015488, -0.00025004, 0.00029544, -0.00069530, 0.00065247, 0.00058174, -0.00066890, 0.00027697, -0.00023455, 0.00015838, -0.00015448, 0.00010978, -0.00011002, 0.00008781, -0.00008107, 0.00007024, -0.00006656, 0.00005583, -0.00005932, 0.00005042, -0.00004369, 0.00004243, -0.00004038, 0.00003369, -0.00003113, 0.00002424, -0.00003359, 0.00003387, -0.00003316, 0.00003424, -0.00002136, 0.00001766, -0.00003912, 0.00004358, -0.00002442, 0.00001745, -0.00002805, 0.00002375, -0.00000994, 0.00000876, -0.00001728, 0.00001377, -0.00000536, 0.00000035, -0.00000041, -0.00001087, 0.00001170, -0.00000945, 0.00002225, -0.00000534, -0.00001838, 0.00001143, -0.00000190, 0.00001903, -0.00003356, 0.00003931, -0.00002993, 0.00002419, -0.00002259, 0.00001524, -0.00001954, 0.00001755, -0.00001319, 0.00000928, },
		{ 0.00000000, 0.00000107, -0.00000031, -0.00000174, 0.00000136, -0.00000383, 0.00000394, 0.00000000, -0.00000084, 0.00000111, -0.00000339, 0.00000800, -0.00000502, 0.00000034, -0.00000353, 0.00000801, -0.00002068, 0.00002979, -0.00001610, 0.00000332, -0.00000149, -0.00000157, 0.00001563, -0.00003915, 0.00001715, 0.00000946, -0.00001350, 0.00002168, -0.00002778, 0.00002271, -0.00002355, 0.00002395, -0.00001342, 0.00000438, -0.00001088, 0.00002077, -0.00001911, 0.00000589, -0.00001623, 0.00003637, -0.00003105, 0.00000821, -0.00000962, 0.00002494, -0.00002292, 0.00002271, -0.00002186, 0.00001092, -0.00001485, 0.00001615, -0.00001894, 0.00001972, -0.00002268, 0.00002594, -0.00002818, 0.00002592, -0.00003817, 0.00003438, -0.00004187, 0.00003916, -0.00004685, 0.00003728, -0.00005974, 0.00005240, -0.00010912, 0.00009259, -0.00023673, -0.00023587, 0.00125794, -0.00072130, 0.00015986, -0.00015581, 0.00008969, -0.00009644, 0.00005790, -0.00006625, 0.00004630, -0.00004512, 0.00003757, -0.00003675, 0.00002932, -0.00003425, 0.00002506, -0.00002218, 0.00002353, -0.00002211, 0.00001780, -0.00001873, 0.00001748, -0.00002836, 0.00002471, -0.00002203, 0.00002011, -0.00000964, 0.00001562, -0.00003375, 0.00002319, -0.00000137, 0.00000306, -0.00001405, 0.00000462, 0.00000538, 0.00000075, -0.00000896, 0.00000191, 0.00000536, -0.00000588, 0.00000345, -0.00000970, 0.00000307, 0.00000329, -0.00000058, 0.00002599, -0.00004212, 0.00002627, -0.00002161, 0.00003883, -0.00004238, 0.00003374, -0.00001587, 0.00000997, -0.00000882, 0.00000392, -0.00000895, 0.00000470, -0.00000021, -0.00000136, 0.00000033, },
		{ 0.00000641, -0.00000395, 0.00000197, -0.00000265, 0.00000104, -0.00000288, 0.00000635, -0.00000556, 0.00000533, -0.00000675, 0.00000774, -0.00000207, -0.00000066, -0.00000485, 0.00000792, -0.00001358, 0.00000857, 0.00001342, -0.00002253, 0.00001964, -0.00002007, 0.00002841, -0.00003254, -0.00000421, 0.00002415, -0.00001734, 0.00001830, -0.00001487, 0.00000365, -0.00000165, -0.00000284, 0.00001410, -0.00001677, 0.00000586, -0.00000055, 0.00000872, -0.00001808, 0.00000351, 0.00000475, 0.00001501, -0.00003385, 0.00002184, -0.00001010, 0.00001964, -0.00002316, 0.00002751, -0.00003626, 0.00002893, -0.00002897, 0.00002839, -0.00003052, 0.00003073, -0.00003194, 0.00003537, -0.00004142, 0.00003670, -0.00004942, 0.00005092, -0.00006357, 0.00006790, -0.00008779, 0.00008260, -0.00011383, 0.00010609, -0.00019112, 0.00021848, -0.00076883, 0.00121634, -0.00012868, -0.00031479, 0.00012895, -0.00012858, 0.00007690, -0.00008767, 0.00005686, -0.00006367, 0.00005043, -0.00004831, 0.00004075, -0.00004063, 0.00003066, -0.00003574, 0.00003127, -0.00002559, 0.00002430, -0.00002417, 0.00001903, -0.00001657, 0.00000749, -0.00001524, 0.00001732, -0.00001831, 0.00002286, -0.00001056, 0.00000114, -0.00002149, 0.00003497, -0.00002114, 0.00001093, -0.00002213, 0.00002401, -0.00001160, 0.00000597, -0.00001492, 0.00001761, -0.00001217, 0.00000687, -0.00000795, -0.00000394, 0.00001128, -0.00001287, 0.00003566, -0.00003555, 0.00001027, -0.00000880, 0.00002016, -0.00001093, -0.00000596, 0.00002135, -0.00001854, 0.00001434, -0.00001467, 0.00000833, -0.00001291, 0.00001431, -0.00001163, 0.00000821, -0.00000686, },
		{ -0.00000163, 0.00000137, -0.00000300, 0.00000222, -0.00000462, 0.00000594, -0.00000304, 0.00000202, -0.00000180, -0.00000029, 0.00000633, -0.00000623, 0.00000079, -0.00000221, 0.00000373, -0.00001598, 0.00003364, -0.00002948, 0.00001797, -0.00001389, 0.00001543, -0.00000637, -0.00002919, 0.00002949, -0.00000863, 0.00000339, 0.00000628, -0.00001832, 0.00001869, -0.00002331, 0.00003009, -0.00002532, 0.00001377, -0.00001542, 0.00002703, -0.00003062, 0.00001623, -0.00002065, 0.00004449, -0.00004970, 0.00002904, -0.00002499, 0.00003954, -0.00004145, 0.00004413, -0.00004716, 0.00003708, -0.00003982, 0.00004132, -0.00004537, 0.00004720, -0.00005084, 0.00005670, -0.00006306, 0.00006177, -0.00007718, 0.00007983, -0.00009337, 0.00009873, -0.00011710, 0.00011563, -0.00015178, 0.00015796, -0.00024789, 0.00029060, -0.00064251, 0.00046886, 0.00078571, -0.00075757, 0.00030938, -0.00025725, 0.00017123, -0.00016449, 0.00011963, -0.00011938, 0.00009642, -0.00008930, 0.00007699, -0.00007286, 0.00006077, -0.00006351, 0.00005429, -0.00004775, 0.00004605, -0.00004383, 0.00003784, -0.00003596, 0.00003140, -0.00004157, 0.00004011, -0.00003816, 0.00003724, -0.00002398, 0.00002484, -0.00004613, 0.00004421, -0.00002185, 0.00001708, -0.00002886, 0.00002296, -0.00000955, 0.00000984, -0.00001953, 0.00001673, -0.00000762, 0.00000265, -0.00000220, -0.00000927, 0.00001166, -0.00001025, 0.00002198, -0.00000437, -0.00002398, 0.00001904, -0.00001245, 0.00002723, -0.00004004, 0.00004216, -0.00002705, 0.00001882, -0.00001712, 0.00001147, -0.00001609, 0.00001383, -0.00000841, 0.00000509, -0.00000446, 0.00000386, },
		{ -0.00000158, -0.00000055, -0.00000011, -0.00000174, 0.00000074, 0.00000294, -0.00000348, 0.00000411, -0.00000665, 0.00000998, -0.00000639, 0.00000200, -0.00000670, 0.00000990, -0.00001942, 0.00002364, -0.00000553, -0.00000913, 0.00001181, -0.00001095, 0.00002200, -0.00004294, 0.00002071, 0.00000675, -0.00000982, 0.00001794, -0.00002370, 0.00001808, -0.00001929, 0.00001929, -0.00000923, 0.00000088, -0.00000871, 0.00001862, -0.00001480, 0.00000261, -0.00001464, 0.00003095, -0.00001912, -0.00000339, -0.00000265, 0.00001689, -0.00001242, 0.00001087, -0.00000834, -0.00000174, -0.00000432, 0.00000579, -0.00000875, 0.00000868, -0.00001066, 0.00001371, -0.00001501, 0.00001098, -0.00002109, 0.00001612, -0.00002114, 0.00001678, -0.00002207, 0.00000981, -0.00002930, 0.00001423, -0.00005654, 0.00002354, -0.00011625, -0.00041939, 0.00135263, -0.00067177, 0.00009387, -0.00011041, 0.00005060, -0.00006634, 0.00003448, -0.00004624, 0.00002842, -0.00002900, 0.00002198, -0.00002261, 0.00001558, -0.00002175, 0.00001396, -0.00001243, 0.00001436, -0.00001399, 0.00001122, -0.00001317, 0.00001346, -0.00002333, 0.00001724, -0.00001387, 0.00001135, -0.00000240, 0.00001189, -0.00002671, 0.00001062, 0.00000995, -0.00000359, -0.00000843, -0.00000195, 0.00001014, -0.00000227, -0.00000645, -0.00000076, 0.00000907, -0.00000952, 0.00000753, -0.00001376, 0.00000785, -0.00000080, 0.00000132, 0.00002540, -0.00004485, 0.00002812, -0.00002246, 0.00003424, -0.00003489, 0.00002266, -0.00000210, -0.00000199, 0.00000010, -0.00000211, -0.00000329, -0.00000094, 0.00000544, -0.00000543, 0.00000366, -0.00000308, 0.00000358, },
		{ 0.00000033, -0.00000166, 0.00000061, -0.00000277, 0.00000515, -0.00000378, 0.00000345, -0.00000424, 0.00000354, 0.00000248, -0.00000508, -0.00000039, 0.00000033, -0.00000251, -0.00000539, 0.00002748, -0.00003388, 0.00002776, -0.00002237, 0.00002859, -0.00003195, -0.00000323, 0.00002279, -0.00001434, 0.00001520, -0.00001153, 0.00000005, 0.00000147, -0.00000623, 0.00001685, -0.00001923, 0.00000817, -0.00000434, 0.00001473, -0.00002306, 0.00000900, -0.00000578, 0.00002865, -0.00004298, 0.00002680, -0.00001805, 0.00002954, -0.00003293, 0.00003703, -0.00004337, 0.00003434, -0.00003531, 0.00003539, -0.00003891, 0.00004008, -0.00004165, 0.00004640, -0.00005410, 0.00005076, -0.00006391, 0.00006701, -0.00008031, 0.00008626, -0.00010804, 0.00010564, -0.00014353, 0.00014455, -0.00023570, 0.00028252, -0.00081639, 0.00112106, 0.00005738, -0.00043684, 0.00019806, -0.00018113, 0.00011472, -0.00011762, 0.00008335, -0.00008761, 0.00007214, -0.00006868, 0.00005885, -0.00005757, 0.00004553, -0.00004892, 0.00004324, -0.00003708, 0.00003471, -0.00003355, 0.00002828, -0.00002574, 0.00001883, -0.00002822, 0.00002935, -0.00002942, 0.00003165, -0.00001770, 0.00001287, -0.00003520, 0.00004326, -0.00002476, 0.00001523, -0.00002738, 0.00002725, -0.00001431, 0.00000952, -0.00001941, 0.00002309, -0.00001727, 0.00001219, -0.00001288, 0.00000100, 0.00000765, -0.00001103, 0.00003387, -0.00003443, 0.00000552, -0.00000121, 0.00000777, 0.00000158, -0.00001851, 0.00003172, -0.00002363, 0.00001564, -0.00001414, 0.00000859, -0.00001329, 0.00001425, -0.00001003, 0.00000654, -0.00000530, 0.00000447, -0.00000149, },
		{ -0.00000091, 0.00000016, -0.00000212, 0.00000211, 0.00000100, -0.00000201, 0.00000286, -0.00000591, 0.00001128, -0.00001078, 0.00000577, -0.00000967, 0.00001187, -0.00002274, 0.00003613, -0.00002702, 0.00001202, -0.00000439, 0.00000755, -0.00000052, -0.00003200, 0.00003165, -0.00000949, 0.00000488, 0.00000458, -0.00001654, 0.00001635, -0.00002123, 0.00002750, -0.00002310, 0.00001226, -0.00001566, 0.00002832, -0.00003028, 0.00001704, -0.00002466, 0.00004672, -0.00004516, 0.00002277, -0.00002302, 0.00003771, -0.00003773, 0.00003900, -0.00003964, 0.00002942, -0.00003411, 0.00003626, -0.00004113, 0.00004286, -0.00004604, 0.00005228, -0.00005863, 0.00005660, -0.00007055, 0.00007275, -0.00008468, 0.00008931, -0.00010658, 0.00010420, -0.00014042, 0.00014356, -0.00022338, 0.00025554, -0.00055158, 0.00026341, 0.00097222, -0.00081069, 0.00031294, -0.00025841, 0.00016701, -0.00016108, 0.00011850, -0.00011898, 0.00009614, -0.00008926, 0.00007608, -0.00007210, 0.00005915, -0.00006182, 0.00005301, -0.00004716, 0.00004538, -0.00004349, 0.00003862, -0.00003765, 0.00003514, -0.00004518, 0.00004140, -0.00003821, 0.00003573, -0.00002295, 0.00002803, -0.00004746, 0.00003899, -0.00001527, 0.00001398, -0.00002699, 0.00001977, -0.00000755, 0.00000962, -0.00002022, 0.00001782, -0.00000794, 0.00000304, -0.00000205, -0.00000950, 0.00001285, -0.00001188, 0.00002208, -0.00000342, -0.00002889, 0.00002525, -0.00001975, 0.00003046, -0.00004020, 0.00003788, -0.00001824, 0.00000920, -0.00000880, 0.00000568, -0.00001076, 0.00000830, -0.00000232, 0.00000012, -0.00000016, -0.00000033, 0.00000223, -0.00000116, },
		{ -0.00000066, -0.00000052, -0.00000125, 0.00000427, -0.00000434, 0.00000506, -0.00000739, 0.00000912, -0.00000497, 0.00000072, -0.00000610, 0.00000697, -0.00001265, 0.00001216, 0.00000845, -0.00002289, 0.00002458, -0.00001854, 0.00002655, -0.00004507, 0.00002360, 0.00000388, -0.00000538, 0.00001306, -0.00001872, 0.00001271, -0.00001429, 0.00001400, -0.00000472, -0.00000296, -0.00000564, 0.00001484, -0.00000879, -0.00000247, -0.00001003, 0.00002149, -0.00000436, -0.00001585, 0.00000521, 0.00000707, -0.00000020, -0.00000218, 0.00000556, -0.00001392, 0.00000576, -0.00000458, 0.00000191, -0.00000321, 0.00000238, 0.00000017, -0.00000013, -0.00000584, -0.00000214, -0.00000401, 0.00000137, -0.00000726, 0.00000445, -0.00001959, 0.00000390, -0.00002757, -0.00000047, -0.00004780, -0.00000137, -0.00056910, 0.00138562, -0.00056910, -0.00000137, -0.00004780, -0.00000047, -0.00002757, 0.00000390, -0.00001959, 0.00000445, -0.00000726, 0.00000137, -0.00000401, -0.00000214, -0.00000584, -0.00000013, 0.00000017, 0.00000238, -0.00000321, 0.00000191, -0.00000458, 0.00000576, -0.00001392, 0.00000556, -0.00000218, -0.00000020, 0.00000707, 0.00000521, -0.00001585, -0.00000436, 0.00002149, -0.00001003, -0.00000247, -0.00000879, 0.00001484, -0.00000564, -0.00000296, -0.00000472, 0.00001400, -0.00001429, 0.00001271, -0.00001872, 0.00001306, -0.00000538, 0.00000388, 0.00002360, -0.00004507, 0.00002655, -0.00001854, 0.00002458, -0.00002289, 0.00000845, 0.00001216, -0.00001265, 0.00000697, -0.00000610, 0.00000072, -0.00000497, 0.00000912, -0.00000739, 0.00000506, -0.00000434, 0.00000427, -0.00000125, -0.00000052, },
		{ -0.00000091, -0.00000116, 0.00000223, -0.00000033, -0.00000016, 0.00000012, -0.00000232, 0.00000830, -0.00001076, 0.00000568, -0.00000880, 0.00000920, -0.00001824, 0.00003788, -0.00004020, 0.00003046, -0.00001975, 0.00002525, -0.00002889, -0.00000342, 0.00002208, -0.00001188, 0.00001285, -0.00000950, -0.00000205, 0.00000304, -0.00000794, 0.00001782, -0.00002022, 0.00000962, -0.00000755, 0.00001977, -0.00002699, 0.00001398, -0.00001527, 0.00003899, -0.00004746, 0.00002803, -0.00002295, 0.00003573, -0.00003821, 0.00004140, -0.00004518, 0.00003514, -0.00003765, 0.00003862, -0.00004349, 0.00004538, -0.00004716, 0.00005301, -0.00006182, 0.00005915, -0.00007210, 0.00007608, -0.00008926, 0.00009614, -0.00011898, 0.00011850, -0.00016108, 0.00016701, -0.00025841, 0.00031294, -0.00081069, 0.00097222, 0.00026341, -0.00055158, 0.00025554, -0.00022338, 0.00014356, -0.00014042, 0.00010420, -0.00010658, 0.00008931, -0.00008468, 0.00007275, -0.00007055, 0.00005660, -0.00005863, 0.00005228, -0.00004604, 0.00004286, -0.00004113, 0.00003626, -0.00003411, 0.00002942, -0.00003964, 0.00003900, -0.00003773, 0.00003771, -0.00002302, 0.00002277, -0.00004516, 0.00004672, -0.00002466, 0.00001704, -0.00003028, 0.00002832, -0.00001566, 0.00001226, -0.00002310, 0.00002750, -0.00002123, 0.00001635, -0.00001654, 0.00000458, 0.00000488, -0.00000949, 0.00003165, -0.00003200, -0.00000052, 0.00000755, -0.00000439, 0.00001202, -0.00002702, 0.00003613, -0.00002274, 0.00001187, -0.00000967, 0.00000577, -0.00001078, 0.00001128, -0.00000591, 0.00000286, -0.00000201, 0.00000100, 0.00000211, -0.00000212, 0.00000016, },
		{ 0.00000033, -0.00000149, 0.00000447, -0.00000530, 0.00000654, -0.00001003, 0.00001425, -0.00001329, 0.00000859, -0.00001414, 0.00001564, -0.00002363, 0.00003172, -0.00001851, 0.00000158, 0.00000777, -0.00000121, 0.00000552, -0.00003443, 0.00003387, -0.00001103, 0.00000765, 0.00000100, -0.00001288, 0.00001219, -0.00001727, 0.00002309, -0.00001941, 0.00000952, -0.00001431, 0.00002725, -0.00002738, 0.00001523, -0.00002476, 0.00004326, -0.00003520, 0.00001287, -0.00001770, 0.00003165, -0.00002942, 0.00002935, -0.00002822, 0.00001883, -0.00002574, 0.00002828, -0.00003355, 0.00003471, -0.00003708, 0.00004324, -0.00004892, 0.00004553, -0.00005757, 0.00005885, -0.00006868, 0.00007214, -0.00008761, 0.00008335, -0.00011762, 0.00011472, -0.00018113, 0.00019806, -0.00043684, 0.00005738, 0.00112106, -0.00081639, 0.00028252, -0.00023570, 0.00014455, -0.00014353, 0.00010564, -0.00010804, 0.00008626, -0.00008031, 0.00006701, -0.00006391, 0.00005076, -0.00005410, 0.00004640, -0.00004165, 0.00004008, -0.00003891, 0.00003539, -0.00003531, 0.00003434, -0.00004337, 0.00003703, -0.00003293, 0.00002954, -0.00001805, 0.00002680, -0.00004298, 0.00002865, -0.00000578, 0.00000900, -0.00002306, 0.00001473, -0.00000434, 0.00000817, -0.00001923, 0.00001685, -0.00000623, 0.00000147, 0.00000005, -0.00001153, 0.00001520, -0.00001434, 0.00002279, -0.00000323, -0.00003195, 0.00002859, -0.00002237, 0.00002776, -0.00003388, 0.00002748, -0.00000539, -0.00000251, 0.00000033, -0.00000039, -0.00000508, 0.00000248, 0.00000354, -0.00000424, 0.00000345, -0.00000378, 0.00000515, -0.00000277, 0.00000061, -0.00000166, },
		{ -0.00000158, 0.00000358, -0.00000308, 0.00000366, -0.00000543, 0.00000544, -0.00000094, -0.00000329, -0.00000211, 0.00000010, -0.00000199, -0.00000210, 0.00002266, -0.00003489, 0.00003424, -0.00002246, 0.00002812, -0.00004485, 0.00002540, 0.00000132, -0.00000080, 0.00000785, -0.00001376, 0.00000753, -0.00000952, 0.00000907, -0.00000076, -0.00000645, -0.00000227, 0.00001014, -0.00000195, -0.00000843, -0.00000359, 0.00000995, 0.00001062, -0.00002671, 0.00001189, -0.00000240, 0.00001135, -0.00001387, 0.00001724, -0.00002333, 0.00001346, -0.00001317, 0.00001122, -0.00001399, 0.00001436, -0.00001243, 0.00001396, -0.00002175, 0.00001558, -0.00002261, 0.00002198, -0.00002900, 0.00002842, -0.00004624, 0.00003448, -0.00006634, 0.00005060, -0.00011041, 0.00009387, -0.00067177, 0.00135263, -0.00041939, -0.00011625, 0.00002354, -0.00005654, 0.00001423, -0.00002930, 0.00000981, -0.00002207, 0.00001678, -0.00002114, 0.00001612, -0.00002109, 0.00001098, -0.00001501, 0.00001371, -0.00001066, 0.00000868, -0.00000875, 0.00000579, -0.00000432, -0.00000174, -0.00000834, 0.00001087, -0.00001242, 0.00001689, -0.00000265, -0.00000339, -0.00001912, 0.00003095, -0.00001464, 0.00000261, -0.00001480, 0.00001862, -0.00000871, 0.00000088, -0.00000923, 0.00001929, -0.00001929, 0.00001808, -0.00002370, 0.00001794, -0.00000982, 0.00000675, 0.00002071, -0.00004294, 0.00002200, -0.00001095, 0.00001181, -0.00000913, -0.00000553, 0.00002364, -0.00001942, 0.00000990, -0.00000670, 0.00000200, -0.00000639, 0.00000998, -0.00000665, 0.00000411, -0.00000348, 0.00000294, 0.00000074, -0.00000174, -0.00000011, -0.00000055, },
		{ -0.00000163, 0.00000386, -0.00000446, 0.00000509, -0.00000841, 0.00001383, -0.00001609, 0.00001147, -0.00001712, 0.00001882, -0.00002705, 0.00004216, -0.00004004, 0.00002723, -0.00001245, 0.00001904, -0.00002398, -0.00000437, 0.00002198, -0.00001025, 0.00001166, -0.00000927, -0.00000220, 0.00000265, -0.00000762, 0.00001673, -0.00001953, 0.00000984, -0.00000955, 0.00002296, -0.00002886, 0.00001708, -0.00002185, 0.00004421, -0.00004613, 0.00002484, -0.00002398, 0.00003724, -0.00003816, 0.00004011, -0.00004157, 0.00003140, -0.00003596, 0.00003784, -0.00004383, 0.00004605, -0.00004775, 0.00005429, -0.00006351, 0.00006077, -0.00007286, 0.00007699, -0.00008930, 0.00009642, -0.00011938, 0.00011963, -0.00016449, 0.00017123, -0.00025725, 0.00030938, -0.00075757, 0.00078571, 0.00046886, -0.00064251, 0.00029060, -0.00024789, 0.00015796, -0.00015178, 0.00011563, -0.00011710, 0.00009873, -0.00009337, 0.00007983, -0.00007718, 0.00006177, -0.00006306, 0.00005670, -0.00005084, 0.00004720, -0.00004537, 0.00004132, -0.00003982, 0.00003708, -0.00004716, 0.00004413, -0.00004145, 0.00003954, -0.00002499, 0.00002904, -0.00004970, 0.00004449, -0.00002065, 0.00001623, -0.00003062, 0.00002703, -0.00001542, 0.00001377, -0.00002532, 0.00003009, -0.00002331, 0.00001869, -0.00001832, 0.00000628, 0.00000339, -0.00000863, 0.00002949, -0.00002919, -0.00000637, 0.00001543, -0.00001389, 0.00001797, -0.00002948, 0.00003364, -0.00001598, 0.00000373, -0.00000221, 0.00000079, -0.00000623, 0.00000633, -0.00000029, -0.00000180, 0.00000202, -0.00000304, 0.00000594, -0.00000462, 0.00000222, -0.00000300, 0.00000137, },
		{ 0.00000641, -0.00000686, 0.00000821, -0.00001163, 0.00001431, -0.00001291, 0.00000833, -0.00001467, 0.00001434, -0.00001854, 0.00002135, -0.00000596, -0.00001093, 0.00002016, -0.00000880, 0.00001027, -0.00003555, 0.00003566, -0.00001287, 0.00001128, -0.00000394, -0.00000795, 0.00000687, -0.00001217, 0.00001761, -0.00001492, 0.00000597, -0.00001160, 0.00002401, -0.00002213, 0.00001093, -0.00002114, 0.00003497, -0.00002149, 0.00000114, -0.00001056, 0.00002286, -0.00001831, 0.00001732, -0.00001524, 0.00000749, -0.00001657, 0.00001903, -0.00002417, 0.00002430, -0.00002559, 0.00003127, -0.00003574, 0.00003066, -0.00004063, 0.00004075, -0.00004831, 0.00005043, -0.00006367, 0.00005686, -0.00008767, 0.00007690, -0.00012858, 0.00012895, -0.00031479, -0.00012868, 0.00121634, -0.00076883, 0.00021848, -0.00019112, 0.00010609, -0.00011383, 0.00008260, -0.00008779, 0.00006790, -0.00006357, 0.00005092, -0.00004942, 0.00003670, -0.00004142, 0.00003537, -0.00003194, 0.00003073, -0.00003052, 0.00002839, -0.00002897, 0.00002893, -0.00003626, 0.00002751, -0.00002316, 0.00001964, -0.00001010, 0.00002184, -0.00003385, 0.00001501, 0.00000475, 0.00000351, -0.00001808, 0.00000872, -0.00000055, 0.00000586, -0.00001677, 0.00001410, -0.00000284, -0.00000165, 0.00000365, -0.00001487, 0.00001830, -0.00001734, 0.00002415, -0.00000421, -0.00003254, 0.00002841, -0.00002007, 0.00001964, -0.00002253, 0.00001342, 0.00000857, -0.00001358, 0.00000792, -0.00000485, -0.00000066, -0.00000207, 0.00000774, -0.00000675, 0.00000533, -0.00000556, 0.00000635, -0.00000288, 0.00000104, -0.00000265, 0.00000197, -0.00000395, },
		{ 0.00000000, 0.00000033, -0.00000136, -0.00000021, 0.00000470, -0.00000895, 0.00000392, -0.00000882, 0.00000997, -0.00001587, 0.00003374, -0.00004238, 0.00003883, -0.00002161, 0.00002627, -0.00004212, 0.00002599, -0.00000058, 0.00000329, 0.00000307, -0.00000970, 0.00000345, -0.00000588, 0.00000536, 0.00000191, -0.00000896, 0.00000075, 0.00000538, 0.00000462, -0.00001405, 0.00000306, -0.00000137, 0.00002319, -0.00003375, 0.00001562, -0.00000964, 0.00002011, -0.00002203, 0.00002471, -0.00002836, 0.00001748, -0.00001873, 0.00001780, -0.00002211, 0.00002353, -0.00002218, 0.00002506, -0.00003425, 0.00002932, -0.00003675, 0.00003757, -0.00004512, 0.00004630, -0.00006625, 0.00005790, -0.00009644, 0.00008969, -0.00015581, 0.00015986, -0.00072130, 0.00125794, -0.00023587, -0.00023673, 0.00009259, -0.00010912, 0.00005240, -0.00005974, 0.00003728, -0.00004685, 0.00003916, -0.00004187, 0.00003438, -0.00003817, 0.00002592, -0.00002818, 0.00002594, -0.00002268, 0.00001972, -0.00001894, 0.00001615, -0.00001485, 0.00001092, -0.00002186, 0.00002271, -0.00002292, 0.00002494, -0.00000962, 0.00000821, -0.00003105, 0.00003637, -0.00001623, 0.00000589, -0.00001911, 0.00002077, -0.00001088, 0.00000438, -0.00001342, 0.00002395, -0.00002355, 0.00002271, -0.00002778, 0.00002168, -0.00001350, 0.00000946, 0.00001715, -0.00003915, 0.00001563, -0.00000157, -0.00000149, 0.00000332, -0.00001610, 0.00002979, -0.00002068, 0.00000801, -0.00000353, 0.00000034, -0.00000502, 0.00000800, -0.00000339, 0.00000111, -0.00000084, 0.00000000, 0.00000394, -0.00000383, 0.00000136, -0.00000174, -0.00000031, 0.00000107, },
		{ -0.00000820, 0.00000928, -0.00001319, 0.00001755, -0.00001954, 0.00001524, -0.00002259, 0.00002419, -0.00002993, 0.00003931, -0.00003356, 0.00001903, -0.00000190, 0.00001143, -0.00001838, -0.00000534, 0.00002225, -0.00000945, 0.00001170, -0.00001087, -0.00000041, 0.00000035, -0.00000536, 0.00001377, -0.00001728, 0.00000876, -0.00000994, 0.00002375, -0.00002805, 0.00001745, -0.00002442, 0.00004358, -0.00003912, 0.00001766, -0.00002136, 0.00003424, -0.00003316, 0.00003387, -0.00003359, 0.00002424, -0.00003113, 0.00003369, -0.00004038, 0.00004243, -0.00004369, 0.00005042, -0.00005932, 0.00005583, -0.00006656, 0.00007024, -0.00008107, 0.00008781, -0.00011002, 0.00010978, -0.00015448, 0.00015838, -0.00023455, 0.00027697, -0.00066890, 0.00058174, 0.00065247, -0.00069530, 0.00029544, -0.00025004, 0.00015488, -0.00014939, 0.00011555, -0.00011719, 0.00009860, -0.00009317, 0.00007870, -0.00007630, 0.00006007, -0.00006137, 0.00005571, -0.00005064, 0.00004684, -0.00004532, 0.00004234, -0.00004152, 0.00004024, -0.00004927, 0.00004357, -0.00003974, 0.00003656, -0.00002301, 0.00003085, -0.00004834, 0.00003693, -0.00001352, 0.00001338, -0.00002870, 0.00002367, -0.00001375, 0.00001390, -0.00002576, 0.00003051, -0.00002323, 0.00001895, -0.00001801, 0.00000597, 0.00000326, -0.00000860, 0.00002782, -0.00002687, -0.00001075, 0.00002079, -0.00001902, 0.00001815, -0.00002544, 0.00002494, -0.00000492, -0.00000688, 0.00000639, -0.00000476, -0.00000104, 0.00000083, 0.00000538, -0.00000610, 0.00000557, -0.00000653, 0.00000898, -0.00000633, 0.00000383, -0.00000495, 0.00000395, -0.00000540, 0.00000764, },
		{ 0.00000744, -0.00001037, 0.00001135, -0.00000966, 0.00000501, -0.00001143, 0.00000862, -0.00000890, 0.00000748, 0.00000754, -0.00002248, 0.00003017, -0.00001327, 0.00001250, -0.00003464, 0.00003659, -0.00001456, 0.00001512, -0.00000947, -0.00000265, 0.00000132, -0.00000686, 0.00001200, -0.00001044, 0.00000225, -0.00000804, 0.00001919, -0.00001529, 0.00000496, -0.00001487, 0.00002366, -0.00000650, -0.00001007, -0.00000358, 0.00001338, -0.00000671, 0.00000541, -0.00000320, -0.00000237, -0.00000849, 0.00001022, -0.00001472, 0.00001346, -0.00001353, 0.00001848, -0.00002145, 0.00001463, -0.00002267, 0.00002168, -0.00002709, 0.00002798, -0.00003888, 0.00002922, -0.00005579, 0.00003667, -0.00007429, 0.00005973, -0.00020108, -0.00027814, 0.00124910, -0.00066956, 0.00012706, -0.00013092, 0.00005706, -0.00007647, 0.00005297, -0.00006134, 0.00004393, -0.00004175, 0.00003037, -0.00003107, 0.00001928, -0.00002586, 0.00002175, -0.00001963, 0.00001878, -0.00001959, 0.00001869, -0.00001958, 0.00001991, -0.00002516, 0.00001456, -0.00001083, 0.00000792, -0.00000078, 0.00001473, -0.00002210, 0.00000057, 0.00001413, -0.00000098, -0.00001317, 0.00000274, 0.00000302, 0.00000325, -0.00001339, 0.00001022, 0.00000147, -0.00000551, 0.00000797, -0.00001875, 0.00002145, -0.00002039, 0.00002591, -0.00000634, -0.00003067, 0.00002499, -0.00001371, 0.00000783, -0.00000875, -0.00000105, 0.00002048, -0.00002139, 0.00001203, -0.00000628, 0.00000136, -0.00000428, 0.00000935, -0.00000673, 0.00000498, -0.00000529, 0.00000556, -0.00000132, 0.00000022, -0.00000251, 0.00000236, -0.00000474, 0.00000632, -0.00000625, },
		{ 0.00000362, -0.00000641, 0.00001043, -0.00001466, 0.00001010, -0.00001740, 0.00002035, -0.00002603, 0.00003901, -0.00004367, 0.00003757, -0.00001596, 0.00002138, -0.00003733, 0.00002564, -0.00000170, 0.00000647, -0.00000068, -0.00000716, 0.00000106, -0.00000397, 0.00000339, 0.00000286, -0.00001010, 0.00000292, 0.00000138, 0.00000985, -0.00001810, 0.00000835, -0.00001035, 0.00003123, -0.00003552, 0.00001538, -0.00001351, 0.00002478, -0.00002543, 0.00002705, -0.00002849, 0.00001749, -0.00002084, 0.00002104, -0.00002672, 0.00002885, -0.00002782, 0.00003174, -0.00004170, 0.00003734, -0.00004461, 0.00004623, -0.00005364, 0.00005598, -0.00007729, 0.00007135, -0.00011431, 0.00011260, -0.00017938, 0.00019270, -0.00071926, 0.00111358, -0.00003700, -0.00034664, 0.00014794, -0.00014994, 0.00008067, -0.00008241, 0.00005840, -0.00006582, 0.00005613, -0.00005737, 0.00004766, -0.00005056, 0.00003646, -0.00003744, 0.00003482, -0.00003173, 0.00002805, -0.00002684, 0.00002459, -0.00002371, 0.00002163, -0.00003247, 0.00003095, -0.00002963, 0.00002946, -0.00001395, 0.00001692, -0.00003824, 0.00003665, -0.00001439, 0.00000703, -0.00002130, 0.00002093, -0.00001179, 0.00000700, -0.00001653, 0.00002714, -0.00002623, 0.00002581, -0.00003025, 0.00002369, -0.00001591, 0.00001157, 0.00001346, -0.00003464, 0.00000892, 0.00000752, -0.00001274, 0.00001182, -0.00002104, 0.00002940, -0.00001624, 0.00000175, 0.00000267, -0.00000353, -0.00000152, 0.00000393, 0.00000150, -0.00000301, 0.00000270, -0.00000367, 0.00000752, -0.00000610, 0.00000318, -0.00000362, 0.00000199, -0.00000254, 0.00000404, -0.00000394, },
		{ -0.00001550, 0.00001841, -0.00002009, 0.00001589, -0.00002403, 0.00002434, -0.00002649, 0.00003004, -0.00002253, 0.00000806, 0.00000963, 0.00000437, -0.00001350, -0.00000546, 0.00002246, -0.00000922, 0.00001267, -0.00001391, 0.00000283, -0.00000335, -0.00000175, 0.00000951, -0.00001396, 0.00000664, -0.00000876, 0.00002211, -0.00002454, 0.00001493, -0.00002286, 0.00003765, -0.00002773, 0.00000798, -0.00001630, 0.00002790, -0.00002465, 0.00002445, -0.00002322, 0.00001548, -0.00002467, 0.00002743, -0.00003426, 0.00003562, -0.00003608, 0.00004254, -0.00005052, 0.00004582, -0.00005489, 0.00005772, -0.00006669, 0.00007267, -0.00009343, 0.00009163, -0.00013410, 0.00013248, -0.00019611, 0.00022480, -0.00056003, 0.00038144, 0.00079575, -0.00070052, 0.00026710, -0.00022919, 0.00013448, -0.00013359, 0.00010408, -0.00010685, 0.00008896, -0.00008420, 0.00006961, -0.00006821, 0.00005188, -0.00005393, 0.00004957, -0.00004557, 0.00004181, -0.00004089, 0.00003908, -0.00003876, 0.00003836, -0.00004565, 0.00003740, -0.00003302, 0.00002934, -0.00001753, 0.00002853, -0.00004185, 0.00002554, -0.00000483, 0.00000959, -0.00002531, 0.00001888, -0.00001109, 0.00001286, -0.00002446, 0.00002885, -0.00002115, 0.00001739, -0.00001590, 0.00000399, 0.00000423, -0.00000929, 0.00002680, -0.00002559, -0.00001286, 0.00002273, -0.00001923, 0.00001273, -0.00001607, 0.00001223, 0.00000778, -0.00001739, 0.00001390, -0.00000906, 0.00000323, -0.00000371, 0.00000966, -0.00000880, 0.00000760, -0.00000852, 0.00001041, -0.00000661, 0.00000442, -0.00000613, 0.00000581, -0.00000812, 0.00001001, -0.00001032, 0.00001156, },
		{ 0.00000609, -0.00000441, -0.00000040, -0.00000564, 0.00000026, 0.00000275, -0.00000663, 0.00001857, -0.00003026, 0.00003571, -0.00001332, 0.00001158, -0.00003141, 0.00003647, -0.00001570, 0.00001854, -0.00001474, 0.00000212, -0.00000354, -0.00000229, 0.00000716, -0.00000675, -0.00000101, -0.00000426, 0.00001365, -0.00000797, -0.00000146, -0.00000756, 0.00001166, 0.00000709, -0.00001845, 0.00000138, 0.00000514, 0.00000321, -0.00000410, 0.00000578, -0.00000901, -0.00000292, 0.00000321, -0.00000667, 0.00000386, -0.00000274, 0.00000691, -0.00000836, 0.00000008, -0.00000659, 0.00000476, -0.00000834, 0.00000835, -0.00001707, 0.00000468, -0.00002696, 0.00000026, -0.00002612, 0.00000039, -0.00010799, -0.00038071, 0.00121866, -0.00052765, 0.00001967, -0.00006465, 0.00000508, -0.00003750, 0.00002168, -0.00003300, 0.00001833, -0.00001857, 0.00000887, -0.00001209, 0.00000148, -0.00001013, 0.00000792, -0.00000686, 0.00000619, -0.00000796, 0.00000801, -0.00000884, 0.00000912, -0.00001226, 0.00000067, 0.00000157, -0.00000326, 0.00000783, 0.00000750, -0.00001013, -0.00001208, 0.00002040, -0.00000328, -0.00000925, -0.00000231, 0.00000565, 0.00000093, -0.00000976, 0.00000603, 0.00000575, -0.00000918, 0.00001208, -0.00002229, 0.00002393, -0.00002291, 0.00002767, -0.00000931, -0.00002689, 0.00001931, -0.00000500, -0.00000517, 0.00000439, -0.00001274, 0.00002770, -0.00002420, 0.00001160, -0.00000415, 0.00000060, -0.00000382, 0.00000819, -0.00000423, 0.00000256, -0.00000318, 0.00000308, 0.00000155, -0.00000155, -0.00000139, 0.00000151, -0.00000386, 0.00000429, -0.00000365, 0.00000455, -0.00000674, },
		{ 0.00001465, -0.00001880, 0.00001456, -0.00002345, 0.00002680, -0.00003042, 0.00003718, -0.00003866, 0.00003110, -0.00000666, 0.00001474, -0.00003151, 0.00002498, -0.00000220, 0.00000860, -0.00000320, -0.00000633, 0.00000053, -0.00000390, 0.00000324, 0.00000207, -0.00000982, 0.00000397, -0.00000128, 0.00001297, -0.00001973, 0.00001112, -0.00001559, 0.00003370, -0.00003176, 0.00001123, -0.00001387, 0.00002513, -0.00002403, 0.00002455, -0.00002435, 0.00001416, -0.00001993, 0.00002108, -0.00002774, 0.00003011, -0.00002906, 0.00003355, -0.00004358, 0.00003912, -0.00004575, 0.00004756, -0.00005421, 0.00005711, -0.00007892, 0.00007415, -0.00011901, 0.00011845, -0.00018087, 0.00019425, -0.00067393, 0.00093718, 0.00015653, -0.00043050, 0.00018008, -0.00017271, 0.00009456, -0.00009376, 0.00007000, -0.00007605, 0.00006501, -0.00006527, 0.00005385, -0.00005640, 0.00004100, -0.00004137, 0.00003903, -0.00003645, 0.00003232, -0.00003105, 0.00002960, -0.00002918, 0.00002847, -0.00003826, 0.00003401, -0.00003133, 0.00002951, -0.00001469, 0.00002159, -0.00003989, 0.00003184, -0.00000959, 0.00000633, -0.00002143, 0.00001914, -0.00001137, 0.00000841, -0.00001807, 0.00002834, -0.00002689, 0.00002699, -0.00003079, 0.00002373, -0.00001687, 0.00001283, 0.00001017, -0.00003035, 0.00000324, 0.00001451, -0.00002004, 0.00001484, -0.00001964, 0.00002289, -0.00000736, -0.00000728, 0.00001025, -0.00000818, 0.00000282, -0.00000091, 0.00000664, -0.00000700, 0.00000599, -0.00000698, 0.00001047, -0.00000771, 0.00000468, -0.00000550, 0.00000451, -0.00000625, 0.00000789, -0.00000790, 0.00000811, -0.00001161, },
		{ -0.00001760, 0.00001323, -0.00002140, 0.00001966, -0.00001792, 0.00001661, -0.00000989, -0.00000280, 0.00001961, -0.00000019, -0.00001059, -0.00000399, 0.00002223, -0.00000919, 0.00001404, -0.00001770, 0.00000675, -0.00000761, 0.00000239, 0.00000479, -0.00001027, 0.00000396, -0.00000634, 0.00001844, -0.00001890, 0.00001013, -0.00001803, 0.00002798, -0.00001422, -0.00000207, -0.00001056, 0.00002002, -0.00001470, 0.00001415, -0.00001275, 0.00000715, -0.00001824, 0.00002053, -0.00002693, 0.00002714, -0.00002657, 0.00003243, -0.00003907, 0.00003298, -0.00004038, 0.00004222, -0.00004919, 0.00005426, -0.00007313, 0.00006905, -0.00010786, 0.00009932, -0.00014967, 0.00016377, -0.00044697, 0.00020353, 0.00088591, -0.00065548, 0.00020833, -0.00018906, 0.00010027, -0.00010730, 0.00008354, -0.00008809, 0.00007173, -0.00006831, 0.00005435, -0.00005467, 0.00003886, -0.00004228, 0.00003959, -0.00003674, 0.00003309, -0.00003293, 0.00003220, -0.00003210, 0.00003204, -0.00003720, 0.00002692, -0.00002284, 0.00001944, -0.00000995, 0.00002338, -0.00003200, 0.00001257, 0.00000344, 0.00000616, -0.00002138, 0.00001352, -0.00000810, 0.00001105, -0.00002183, 0.00002559, -0.00001769, 0.00001464, -0.00001264, 0.00000097, 0.00000577, -0.00001035, 0.00002635, -0.00002553, -0.00001249, 0.00002127, -0.00001510, 0.00000319, -0.00000374, -0.00000144, 0.00001909, -0.00002527, 0.00001833, -0.00001066, 0.00000537, -0.00000616, 0.00001157, -0.00000915, 0.00000751, -0.00000852, 0.00000989, -0.00000524, 0.00000377, -0.00000620, 0.00000643, -0.00000920, 0.00001035, -0.00001025, 0.00001137, -0.00001489, 0.00001611, },
		{ -0.00000635, 0.00000082, -0.00000835, 0.00001350, -0.00001773, 0.00002432, -0.00003244, 0.00003576, -0.00000869, 0.00000774, -0.00002620, 0.00003551, -0.00001612, 0.00002101, -0.00001910, 0.00000565, -0.00000701, 0.00000086, 0.00000373, -0.00000437, -0.00000332, -0.00000087, 0.00000830, -0.00000132, -0.00000700, -0.00000092, 0.00000122, 0.00001700, -0.00002237, 0.00000310, -0.00000051, 0.00000990, -0.00000975, 0.00001050, -0.00001164, -0.00000046, -0.00000129, -0.00000093, -0.00000337, 0.00000542, -0.00000188, 0.00000172, -0.00001097, 0.00000545, -0.00000771, 0.00000548, -0.00000588, -0.00000103, -0.00001363, -0.00000493, -0.00002762, 0.00001021, -0.00004231, -0.00004267, -0.00043352, 0.00113247, -0.00035825, -0.00008907, -0.00000325, -0.00004164, -0.00000331, -0.00000609, -0.00000735, -0.00000465, 0.00000206, -0.00001001, 0.00000425, -0.00001373, 0.00000313, -0.00000377, 0.00000421, -0.00000494, 0.00000237, -0.00000171, 0.00000126, -0.00000128, -0.00000004, -0.00001153, 0.00001154, -0.00001171, 0.00001384, 0.00000199, -0.00000011, -0.00002092, 0.00002233, -0.00000283, -0.00000681, -0.00000589, 0.00000686, -0.00000063, -0.00000652, 0.00000236, 0.00000912, -0.00001179, 0.00001517, -0.00002473, 0.00002511, -0.00002443, 0.00002902, -0.00001257, -0.00002211, 0.00001277, 0.00000402, -0.00001677, 0.00001418, -0.00001931, 0.00002887, -0.00002156, 0.00000683, 0.00000101, -0.00000237, -0.00000120, 0.00000491, -0.00000002, -0.00000111, -0.00000004, -0.00000028, 0.00000493, -0.00000364, 0.00000024, -0.00000018, -0.00000178, 0.00000096, 0.00000013, 0.00000043, -0.00000185, -0.00000011, 0.00000135, },
		{ 0.00001600, -0.00002556, 0.00002805, -0.00002835, 0.00002871, -0.00002885, 0.00002135, 0.00000427, 0.00000815, -0.00002597, 0.00002480, -0.00000238, 0.00000979, -0.00000460, -0.00000701, 0.00000153, -0.00000534, 0.00000454, -0.00000005, -0.00000842, 0.00000396, -0.00000239, 0.00001368, -0.00001862, 0.00001096, -0.00001666, 0.00003075, -0.00002340, 0.00000429, -0.00001156, 0.00002195, -0.00001886, 0.00001856, -0.00001748, 0.00000890, -0.00001708, 0.00001875, -0.00002584, 0.00002789, -0.00002646, 0.00003104, -0.00004050, 0.00003544, -0.00004108, 0.00004263, -0.00004805, 0.00005102, -0.00007251, 0.00006768, -0.00011207, 0.00010945, -0.00016386, 0.00017099, -0.00059831, 0.00074888, 0.00032515, -0.00047644, 0.00018341, -0.00017460, 0.00009241, -0.00009259, 0.00007087, -0.00007636, 0.00006482, -0.00006474, 0.00005232, -0.00005521, 0.00003919, -0.00003969, 0.00003819, -0.00003641, 0.00003202, -0.00003100, 0.00003046, -0.00003035, 0.00003043, -0.00003844, 0.00003152, -0.00002801, 0.00002525, -0.00001194, 0.00002217, -0.00003639, 0.00002308, -0.00000312, 0.00000461, -0.00002002, 0.00001583, -0.00000988, 0.00000861, -0.00001792, 0.00002747, -0.00002555, 0.00002632, -0.00002949, 0.00002198, -0.00001651, 0.00001320, 0.00000757, -0.00002699, -0.00000042, 0.00001832, -0.00002254, 0.00001223, -0.00001275, 0.00001214, 0.00000359, -0.00001672, 0.00001712, -0.00001188, 0.00000652, -0.00000504, 0.00001067, -0.00000965, 0.00000795, -0.00000896, 0.00001196, -0.00000801, 0.00000527, -0.00000672, 0.00000642, -0.00000906, 0.00001049, -0.00001041, 0.00001087, -0.00001453, 0.00001617, -0.00002020, },
		{ -0.00001576, 0.00001179, -0.00000663, 0.00000224, 0.00000098, -0.00001077, 0.00002587, -0.00000083, -0.00001041, -0.00000059, 0.00002136, -0.00000899, 0.00001519, -0.00002145, 0.00001045, -0.00001156, 0.00000616, 0.00000046, -0.00000695, 0.00000131, -0.00000325, 0.00001351, -0.00001211, 0.00000418, -0.00001143, 0.00001679, -0.00000117, -0.00001024, -0.00000595, 0.00001247, -0.00000542, 0.00000518, -0.00000426, 0.00000093, -0.00001321, 0.00001431, -0.00001976, 0.00001856, -0.00001691, 0.00002202, -0.00002717, 0.00001981, -0.00002575, 0.00002668, -0.00003178, 0.00003601, -0.00005283, 0.00004614, -0.00008061, 0.00006503, -0.00010315, 0.00010435, -0.00034351, 0.00006158, 0.00091773, -0.00056486, 0.00012732, -0.00013706, 0.00005837, -0.00007548, 0.00005796, -0.00006448, 0.00005022, -0.00004866, 0.00003590, -0.00003844, 0.00002359, -0.00002873, 0.00002781, -0.00002596, 0.00002236, -0.00002299, 0.00002315, -0.00002296, 0.00002284, -0.00002585, 0.00001438, -0.00001148, 0.00000905, -0.00000216, 0.00001723, -0.00002104, 0.00000045, 0.00000943, 0.00000419, -0.00001776, 0.00000841, -0.00000545, 0.00000902, -0.00001846, 0.00002151, -0.00001371, 0.00001156, -0.00000907, -0.00000227, 0.00000719, -0.00001130, 0.00002619, -0.00002647, -0.00001005, 0.00001720, -0.00000811, -0.00000816, 0.00000859, -0.00001296, 0.00002641, -0.00002867, 0.00001853, -0.00000892, 0.00000494, -0.00000610, 0.00001083, -0.00000710, 0.00000538, -0.00000668, 0.00000762, -0.00000250, 0.00000210, -0.00000522, 0.00000573, -0.00000853, 0.00000867, -0.00000808, 0.00000889, -0.00001169, 0.00001124, -0.00001278, 0.00000807, },
		{ -0.00001490, 0.00002080, -0.00002342, 0.00002329, -0.00002868, 0.00003068, -0.00000023, 0.00000201, -0.00001989, 0.00003424, -0.00001588, 0.00002230, -0.00002220, 0.00000759, -0.00000880, 0.00000231, 0.00000193, -0.00000348, -0.00000449, 0.00000171, 0.00000388, 0.00000371, -0.00001060, 0.00000369, -0.00000602, 0.00002190, -0.00002123, 0.00000132, -0.00000312, 0.00001281, -0.00001112, 0.00001090, -0.00001051, -0.00000080, -0.00000336, 0.00000230, -0.00000776, 0.00001032, -0.00000710, 0.00000786, -0.00001746, 0.00001236, -0.00001464, 0.00001325, -0.00001356, 0.00000800, -0.00002426, 0.00000842, -0.00004459, 0.00003199, -0.00006564, -0.00000667, -0.00044075, 0.00100448, -0.00018005, -0.00018394, 0.00004312, -0.00007598, 0.00002083, -0.00002610, 0.00001177, -0.00002149, 0.00001693, -0.00002335, 0.00001535, -0.00002405, 0.00001189, -0.00001152, 0.00001188, -0.00001294, 0.00000975, -0.00000880, 0.00000897, -0.00000930, 0.00000930, -0.00001990, 0.00001724, -0.00001593, 0.00001606, -0.00000063, 0.00000659, -0.00002488, 0.00001970, 0.00000010, -0.00000574, -0.00000787, 0.00000653, -0.00000122, -0.00000408, -0.00000020, 0.00001096, -0.00001279, 0.00001675, -0.00002566, 0.00002470, -0.00002469, 0.00002964, -0.00001560, -0.00001729, 0.00000681, 0.00001147, -0.00002491, 0.00001887, -0.00001982, 0.00002412, -0.00001444, -0.00000092, 0.00000775, -0.00000627, 0.00000242, 0.00000069, 0.00000464, -0.00000487, 0.00000306, -0.00000350, 0.00000787, -0.00000525, 0.00000170, -0.00000203, 0.00000070, -0.00000267, 0.00000398, -0.00000366, 0.00000288, -0.00000569, 0.00000602, -0.00001115, 0.00000600, },
		{ 0.00002427, -0.00002078, 0.00001567, -0.00001701, 0.00001100, 0.00001440, 0.00000353, -0.00002198, 0.00002585, -0.00000258, 0.00001031, -0.00000531, -0.00000862, 0.00000344, -0.00000763, 0.00000660, -0.00000285, -0.00000644, 0.00000318, -0.00000209, 0.00001217, -0.00001509, 0.00000818, -0.00001412, 0.00002368, -0.00001240, -0.00000359, -0.00000806, 0.00001672, -0.00001168, 0.00001107, -0.00000989, 0.00000342, -0.00001364, 0.00001517, -0.00002207, 0.00002330, -0.00002119, 0.00002549, -0.00003391, 0.00002798, -0.00003251, 0.00003354, -0.00003747, 0.00004017, -0.00006072, 0.00005482, -0.00009686, 0.00009005, -0.00013456, 0.00013220, -0.00050730, 0.00056795, 0.00045357, -0.00047833, 0.00015756, -0.00015682, 0.00007575, -0.00008023, 0.00006199, -0.00006761, 0.00005640, -0.00005669, 0.00004403, -0.00004797, 0.00003200, -0.00003328, 0.00003304, -0.00003214, 0.00002760, -0.00002705, 0.00002738, -0.00002733, 0.00002767, -0.00003349, 0.00002437, -0.00002078, 0.00001791, -0.00000675, 0.00001964, -0.00002919, 0.00001232, 0.00000333, 0.00000292, -0.00001777, 0.00001166, -0.00000779, 0.00000787, -0.00001631, 0.00002487, -0.00002263, 0.00002428, -0.00002687, 0.00001896, -0.00001525, 0.00001291, 0.00000573, -0.00002489, -0.00000169, 0.00001869, -0.00002052, 0.00000516, -0.00000249, -0.00000004, 0.00001378, -0.00002411, 0.00002133, -0.00001321, 0.00000839, -0.00000734, 0.00001256, -0.00001014, 0.00000796, -0.00000909, 0.00001157, -0.00000672, 0.00000467, -0.00000688, 0.00000715, -0.00001030, 0.00001116, -0.00001081, 0.00001125, -0.00001457, 0.00001455, -0.00001851, 0.00001407, -0.00002347, },
		{ 0.00000444, -0.00000974, 0.00000720, -0.00001392, 0.00002725, 0.00000293, -0.00001292, 0.00000453, 0.00002001, -0.00000844, 0.00001559, -0.00002448, 0.00001322, -0.00001447, 0.00000883, -0.00000281, -0.00000454, -0.00000082, -0.00000010, 0.00000822, -0.00000533, -0.00000158, -0.00000479, 0.00000636, 0.00000910, -0.00001482, -0.00000380, 0.00000667, 0.00000157, -0.00000087, 0.00000092, -0.00000225, -0.00001030, 0.00000954, -0.00001376, 0.00001111, -0.00000854, 0.00001297, -0.00001673, 0.00000845, -0.00001334, 0.00001359, -0.00001708, 0.00002070, -0.00003556, 0.00002632, -0.00005650, 0.00003486, -0.00006306, 0.00005467, -0.00025933, -0.00003757, 0.00089397, -0.00043988, 0.00003625, -0.00008279, 0.00001623, -0.00004400, 0.00003214, -0.00004029, 0.00002841, -0.00002893, 0.00001769, -0.00002266, 0.00000891, -0.00001583, 0.00001648, -0.00001532, 0.00001163, -0.00001301, 0.00001379, -0.00001322, 0.00001285, -0.00001398, 0.00000233, -0.00000140, 0.00000035, 0.00000393, 0.00001192, -0.00001108, -0.00000876, 0.00001188, 0.00000427, -0.00001493, 0.00000412, -0.00000362, 0.00000725, -0.00001498, 0.00001739, -0.00001007, 0.00000899, -0.00000599, -0.00000500, 0.00000794, -0.00001173, 0.00002596, -0.00002798, -0.00000632, 0.00001179, -0.00000021, -0.00001878, 0.00001819, -0.00001994, 0.00002823, -0.00002698, 0.00001449, -0.00000419, 0.00000233, -0.00000389, 0.00000797, -0.00000329, 0.00000191, -0.00000368, 0.00000433, 0.00000090, -0.00000002, -0.00000361, 0.00000408, -0.00000653, 0.00000554, -0.00000451, 0.00000496, -0.00000693, 0.00000506, -0.00000702, 0.00000190, -0.00000893, 0.00000306, },
		{ -0.00002276, 0.00001572, -0.00002024, 0.00002212, 0.00001025, -0.00000392, -0.00001370, 0.00003341, -0.00001524, 0.00002241, -0.00002404, 0.00000803, -0.00000905, 0.00000223, 0.00000155, -0.00000385, -0.00000464, 0.00000337, 0.00000078, 0.00000658, -0.00001168, 0.00000557, -0.00000934, 0.00002163, -0.00001565, -0.00000318, -0.00000318, 0.00001237, -0.00000887, 0.00000793, -0.00000679, -0.00000295, -0.00000367, 0.00000341, -0.00000953, 0.00001204, -0.00000880, 0.00001005, -0.00001941, 0.00001422, -0.00001621, 0.00001525, -0.00001502, 0.00001036, -0.00002747, 0.00001319, -0.00005064, 0.00003956, -0.00007080, 0.00000346, -0.00041206, 0.00085240, -0.00001215, -0.00025194, 0.00006724, -0.00009354, 0.00003190, -0.00003595, 0.00002216, -0.00003014, 0.00002422, -0.00002959, 0.00001991, -0.00002836, 0.00001523, -0.00001450, 0.00001529, -0.00001689, 0.00001323, -0.00001229, 0.00001317, -0.00001366, 0.00001442, -0.00002327, 0.00001791, -0.00001552, 0.00001432, -0.00000012, 0.00000963, -0.00002406, 0.00001338, 0.00000452, -0.00000547, -0.00000851, 0.00000486, -0.00000091, -0.00000263, -0.00000136, 0.00001102, -0.00001205, 0.00001673, -0.00002504, 0.00002275, -0.00002375, 0.00002943, -0.00001800, -0.00001322, 0.00000250, 0.00001609, -0.00002852, 0.00001803, -0.00001483, 0.00001505, -0.00000491, -0.00000952, 0.00001413, -0.00000950, 0.00000565, -0.00000308, 0.00000844, -0.00000752, 0.00000507, -0.00000560, 0.00000953, -0.00000569, 0.00000240, -0.00000338, 0.00000274, -0.00000562, 0.00000683, -0.00000656, 0.00000618, -0.00000929, 0.00000830, -0.00001352, 0.00000846, -0.00001781, 0.00002316, },
		{ 0.00000119, -0.00000640, 0.00000272, 0.00002158, 0.00000237, -0.00002037, 0.00002851, -0.00000300, 0.00001047, -0.00000587, -0.00001046, 0.00000547, -0.00000996, 0.00000861, -0.00000556, -0.00000449, 0.00000212, -0.00000081, 0.00000903, -0.00000996, 0.00000375, -0.00000929, 0.00001449, -0.00000113, -0.00001035, -0.00000502, 0.00001113, -0.00000439, 0.00000411, -0.00000346, -0.00000077, -0.00001078, 0.00001144, -0.00001759, 0.00001763, -0.00001472, 0.00001854, -0.00002569, 0.00001886, -0.00002239, 0.00002284, -0.00002521, 0.00002751, -0.00004669, 0.00003906, -0.00007758, 0.00006565, -0.00010009, 0.00008776, -0.00041496, 0.00040992, 0.00053299, -0.00043687, 0.00010746, -0.00012444, 0.00004898, -0.00006026, 0.00004626, -0.00005236, 0.00004222, -0.00004349, 0.00003124, -0.00003681, 0.00002142, -0.00002393, 0.00002511, -0.00002503, 0.00002034, -0.00002036, 0.00002142, -0.00002117, 0.00002139, -0.00002496, 0.00001442, -0.00001164, 0.00000939, -0.00000079, 0.00001558, -0.00002027, 0.00000179, 0.00000809, 0.00000222, -0.00001537, 0.00000728, -0.00000564, 0.00000660, -0.00001370, 0.00002115, -0.00001886, 0.00002160, -0.00002365, 0.00001539, -0.00001366, 0.00001231, 0.00000451, -0.00002399, -0.00000077, 0.00001620, -0.00001523, -0.00000427, 0.00000836, -0.00001069, 0.00002069, -0.00002762, 0.00002166, -0.00001142, 0.00000789, -0.00000732, 0.00001198, -0.00000836, 0.00000600, -0.00000742, 0.00000943, -0.00000404, 0.00000303, -0.00000598, 0.00000656, -0.00000981, 0.00000980, -0.00000908, 0.00000929, -0.00001194, 0.00001023, -0.00001430, 0.00000937, -0.00001807, 0.00001690, -0.00000996, },
		{ 0.00000713, -0.00001175, 0.00002385, 0.00001048, -0.00001723, 0.00001061, 0.00001864, -0.00000757, 0.00001498, -0.00002638, 0.00001465, -0.00001598, 0.00001006, -0.00000471, -0.00000330, -0.00000215, 0.00000265, 0.00000337, 0.00000037, -0.00000597, 0.00000043, -0.00000154, 0.00001508, -0.00001501, -0.00000456, 0.00000328, 0.00000549, -0.00000337, 0.00000251, -0.00000241, -0.00000945, 0.00000640, -0.00000936, 0.00000549, -0.00000238, 0.00000635, -0.00000901, 0.00000027, -0.00000459, 0.00000444, -0.00000668, 0.00001002, -0.00002316, 0.00001176, -0.00003822, 0.00001221, -0.00003347, 0.00001940, -0.00019891, -0.00009405, 0.00082433, -0.00029622, -0.00005103, -0.00003605, -0.00001904, -0.00001824, 0.00001049, -0.00001950, 0.00000999, -0.00001250, 0.00000277, -0.00001009, -0.00000270, -0.00000576, 0.00000755, -0.00000664, 0.00000272, -0.00000475, 0.00000588, -0.00000476, 0.00000411, -0.00000387, -0.00000698, 0.00000543, -0.00000501, 0.00000702, 0.00000873, -0.00000362, -0.00001388, 0.00001045, 0.00000627, -0.00001293, 0.00000091, -0.00000287, 0.00000602, -0.00001189, 0.00001389, -0.00000744, 0.00000753, -0.00000393, -0.00000673, 0.00000765, -0.00001138, 0.00002539, -0.00002955, -0.00000222, 0.00000643, 0.00000676, -0.00002656, 0.00002319, -0.00002127, 0.00002449, -0.00002088, 0.00000734, 0.00000227, -0.00000132, -0.00000059, 0.00000405, 0.00000113, -0.00000182, -0.00000055, 0.00000099, 0.00000404, -0.00000182, -0.00000200, 0.00000209, -0.00000396, 0.00000187, -0.00000061, 0.00000075, -0.00000198, -0.00000088, -0.00000191, -0.00000359, -0.00000288, -0.00000415, 0.00001263, -0.00001677, },
		{ -0.00000966, 0.00001250, 0.00002050, -0.00000822, -0.00000889, 0.00003372, -0.00001450, 0.00002155, -0.00002493, 0.00000739, -0.00000826, 0.00000115, 0.00000207, -0.00000498, -0.00000413, 0.00000424, -0.00000091, 0.00000725, -0.00001032, 0.00000479, -0.00000902, 0.00001715, -0.00000722, -0.00000887, -0.00000188, 0.00000973, -0.00000440, 0.00000325, -0.00000212, -0.00000553, -0.00000321, 0.00000314, -0.00000931, 0.00001125, -0.00000767, 0.00000905, -0.00001770, 0.00001212, -0.00001366, 0.00001283, -0.00001177, 0.00000761, -0.00002489, 0.00001110, -0.00004781, 0.00003574, -0.00006197, -0.00000538, -0.00036011, 0.00069458, 0.00012914, -0.00028464, 0.00006628, -0.00009353, 0.00002981, -0.00003561, 0.00002361, -0.00003038, 0.00002383, -0.00002879, 0.00001808, -0.00002691, 0.00001343, -0.00001297, 0.00001459, -0.00001680, 0.00001277, -0.00001207, 0.00001364, -0.00001406, 0.00001503, -0.00002168, 0.00001400, -0.00001122, 0.00000947, 0.00000276, 0.00000965, -0.00001956, 0.00000500, 0.00000901, -0.00000520, -0.00000831, 0.00000229, 0.00000002, -0.00000207, -0.00000119, 0.00000949, -0.00000984, 0.00001544, -0.00002323, 0.00001967, -0.00002193, 0.00002851, -0.00001959, -0.00001036, 0.00000038, 0.00001742, -0.00002760, 0.00001256, -0.00000620, 0.00000417, 0.00000442, -0.00001666, 0.00001825, -0.00001065, 0.00000731, -0.00000527, 0.00001036, -0.00000827, 0.00000532, -0.00000602, 0.00000944, -0.00000465, 0.00000202, -0.00000378, 0.00000373, -0.00000715, 0.00000791, -0.00000753, 0.00000727, -0.00001020, 0.00000760, -0.00001293, 0.00000762, -0.00001659, 0.00002049, -0.00001647, 0.00000346, },
		{ -0.00000155, 0.00002454, 0.00000531, -0.00002123, 0.00003268, -0.00000353, 0.00001044, -0.00000673, -0.00001190, 0.00000693, -0.00001167, 0.00000991, -0.00000757, -0.00000309, 0.00000118, 0.00000094, 0.00000506, -0.00000430, -0.00000109, -0.00000379, 0.00000535, 0.00000816, -0.00001430, -0.00000373, 0.00000657, 0.00000141, -0.00000076, 0.00000050, -0.00000278, -0.00000918, 0.00000828, -0.00001331, 0.00001203, -0.00000841, 0.00001176, -0.00001763, 0.00001011, -0.00001290, 0.00001285, -0.00001378, 0.00001570, -0.00003335, 0.00002371, -0.00005823, 0.00004140, -0.00006700, 0.00004616, -0.00033231, 0.00028455, 0.00056201, -0.00035929, 0.00004243, -0.00008522, 0.00001826, -0.00003756, 0.00002767, -0.00003423, 0.00002566, -0.00002833, 0.00001695, -0.00002449, 0.00000996, -0.00001388, 0.00001638, -0.00001687, 0.00001199, -0.00001262, 0.00001421, -0.00001351, 0.00001341, -0.00001500, 0.00000403, -0.00000287, 0.00000174, 0.00000418, 0.00001169, -0.00001165, -0.00000657, 0.00000997, 0.00000301, -0.00001325, 0.00000325, -0.00000389, 0.00000523, -0.00001068, 0.00001706, -0.00001506, 0.00001905, -0.00002056, 0.00001195, -0.00001226, 0.00001176, 0.00000364, -0.00002399, 0.00000172, 0.00001193, -0.00000841, -0.00001366, 0.00001715, -0.00001743, 0.00002276, -0.00002647, 0.00001793, -0.00000675, 0.00000531, -0.00000524, 0.00000933, -0.00000484, 0.00000267, -0.00000453, 0.00000619, -0.00000061, 0.00000087, -0.00000438, 0.00000495, -0.00000791, 0.00000688, -0.00000582, 0.00000572, -0.00000755, 0.00000432, -0.00000881, 0.00000329, -0.00001107, 0.00000817, 0.00000122, -0.00001140, 0.00000009, },
		{ 0.00001701, 0.00002027, -0.00002187, 0.00001652, 0.00001795, -0.00000661, 0.00001331, -0.00002707, 0.00001471, -0.00001612, 0.00000990, -0.00000533, -0.00000313, -0.00000268, 0.00000473, -0.00000048, 0.00000431, -0.00000825, 0.00000332, -0.00000598, 0.00001635, -0.00001113, -0.00000773, 0.00000207, 0.00000649, -0.00000266, 0.00000115, -0.00000042, -0.00000995, 0.00000453, -0.00000651, 0.00000185, 0.00000131, 0.00000254, -0.00000442, -0.00000425, 0.00000008, -0.00000038, -0.00000094, 0.00000431, -0.00001607, 0.00000307, -0.00002672, -0.00000167, -0.00001562, -0.00000051, -0.00016183, -0.00011376, 0.00072319, -0.00015120, -0.00012166, -0.00000470, -0.00004227, -0.00000194, -0.00000395, -0.00000492, -0.00000244, -0.00000169, -0.00000682, -0.00000248, -0.00000971, 0.00000016, 0.00000221, -0.00000112, -0.00000316, 0.00000056, 0.00000066, 0.00000109, -0.00000186, 0.00000290, -0.00001213, 0.00000798, -0.00000636, 0.00000668, 0.00000810, 0.00000077, -0.00001477, 0.00000577, 0.00000941, -0.00001137, -0.00000131, -0.00000317, 0.00000542, -0.00000948, 0.00001142, -0.00000616, 0.00000743, -0.00000310, -0.00000733, 0.00000631, -0.00001024, 0.00002432, -0.00003077, 0.00000144, 0.00000224, 0.00001141, -0.00003025, 0.00002295, -0.00001727, 0.00001649, -0.00001218, -0.00000101, 0.00000864, -0.00000450, 0.00000252, 0.00000039, 0.00000491, -0.00000468, 0.00000172, -0.00000143, 0.00000608, -0.00000262, -0.00000100, 0.00000044, -0.00000165, -0.00000132, 0.00000253, -0.00000254, 0.00000185, -0.00000521, 0.00000118, -0.00000702, 0.00000079, -0.00000810, 0.00001626, -0.00001770, 0.00000079, -0.00000531, },
		{ 0.00002847, -0.00000939, -0.00000629, 0.00003563, -0.00001386, 0.00001998, -0.00002529, 0.00000627, -0.00000710, -0.00000027, 0.00000282, -0.00000625, -0.00000345, 0.00000463, -0.00000145, 0.00000609, -0.00000711, 0.00000208, -0.00000613, 0.00001021, 0.00000194, -0.00001393, -0.00000062, 0.00000635, 0.00000062, -0.00000137, 0.00000184, -0.00000730, -0.00000289, 0.00000230, -0.00000793, 0.00000888, -0.00000480, 0.00000610, -0.00001377, 0.00000768, -0.00000881, 0.00000797, -0.00000593, 0.00000203, -0.00001892, 0.00000477, -0.00003927, 0.00002477, -0.00004494, -0.00002481, -0.00029797, 0.00054701, 0.00023275, -0.00027962, 0.00004238, -0.00007882, 0.00001721, -0.00002727, 0.00001784, -0.00002377, 0.00001733, -0.00002248, 0.00001140, -0.00002117, 0.00000789, -0.00000818, 0.00001082, -0.00001357, 0.00000921, -0.00000889, 0.00001104, -0.00001115, 0.00001194, -0.00001629, 0.00000700, -0.00000465, 0.00000312, 0.00000665, 0.00000795, -0.00001310, -0.00000348, 0.00001211, -0.00000419, -0.00000773, -0.00000068, 0.00000116, -0.00000212, -0.00000003, 0.00000685, -0.00000675, 0.00001346, -0.00002081, 0.00001605, -0.00001968, 0.00002714, -0.00002040, -0.00000880, 0.00000046, 0.00001579, -0.00002314, 0.00000428, 0.00000352, -0.00000572, 0.00001116, -0.00002051, 0.00001886, -0.00000895, 0.00000683, -0.00000534, 0.00001002, -0.00000690, 0.00000373, -0.00000471, 0.00000766, -0.00000226, 0.00000064, -0.00000317, 0.00000346, -0.00000702, 0.00000705, -0.00000643, 0.00000609, -0.00000848, 0.00000420, -0.00000976, 0.00000391, -0.00001193, 0.00001403, -0.00000660, -0.00001048, -0.00000007, 0.00000434, },
		{ 0.00001196, -0.00002387, 0.00003771, -0.00000376, 0.00001021, -0.00000820, -0.00001253, 0.00000739, -0.00001235, 0.00001013, -0.00000854, -0.00000250, 0.00000068, 0.00000269, 0.00000106, 0.00000084, -0.00000512, 0.00000086, -0.00000192, 0.00001389, -0.00001454, -0.00000474, 0.00000377, 0.00000485, -0.00000279, 0.00000160, -0.00000254, -0.00000889, 0.00000597, -0.00000975, 0.00000729, -0.00000325, 0.00000630, -0.00001106, 0.00000321, -0.00000561, 0.00000523, -0.00000491, 0.00000659, -0.00002272, 0.00001116, -0.00004184, 0.00002115, -0.00003999, 0.00001321, -0.00026597, 0.00019523, 0.00054616, -0.00025781, -0.00002578, -0.00004786, -0.00000996, -0.00001711, 0.00001034, -0.00001696, 0.00001022, -0.00001441, 0.00000407, -0.00001364, 0.00000000, -0.00000522, 0.00000874, -0.00000945, 0.00000432, -0.00000554, 0.00000744, -0.00000614, 0.00000573, -0.00000583, -0.00000459, 0.00000358, -0.00000337, 0.00000683, 0.00000924, -0.00000480, -0.00001156, 0.00000860, 0.00000519, -0.00001147, -0.00000016, -0.00000284, 0.00000410, -0.00000776, 0.00001325, -0.00001188, 0.00001722, -0.00001816, 0.00000914, -0.00001143, 0.00001150, 0.00000288, -0.00002444, 0.00000494, 0.00000718, -0.00000188, -0.00002090, 0.00002193, -0.00001905, 0.00001975, -0.00002114, 0.00001107, -0.00000029, 0.00000166, -0.00000203, 0.00000559, -0.00000062, -0.00000104, -0.00000137, 0.00000276, 0.00000269, -0.00000109, -0.00000266, 0.00000289, -0.00000532, 0.00000329, -0.00000203, 0.00000165, -0.00000270, -0.00000166, -0.00000360, -0.00000253, -0.00000444, 0.00000049, 0.00001003, -0.00001944, 0.00000075, -0.00000120, 0.00002317, },
		{ -0.00002511, 0.00002109, 0.00001862, -0.00000584, 0.00001073, -0.00002677, 0.00001372, -0.00001526, 0.00000874, -0.00000506, -0.00000368, -0.00000264, 0.00000614, -0.00000311, 0.00000623, -0.00000826, 0.00000372, -0.00000696, 0.00001360, -0.00000448, -0.00001207, 0.00000213, 0.00000542, 0.00000016, -0.00000182, 0.00000240, -0.00001077, 0.00000325, -0.00000478, -0.00000009, 0.00000278, 0.00000123, -0.00000267, -0.00000553, 0.00000121, -0.00000151, 0.00000084, 0.00000282, -0.00001351, -0.00000046, -0.00002136, -0.00000752, -0.00000828, -0.00000722, -0.00014382, -0.00010646, 0.00060677, -0.00002077, -0.00016613, 0.00000692, -0.00005126, 0.00000361, -0.00001014, 0.00000237, -0.00000791, 0.00000271, -0.00001047, -0.00000026, -0.00001183, 0.00000172, 0.00000066, 0.00000095, -0.00000562, 0.00000251, -0.00000140, 0.00000377, -0.00000443, 0.00000571, -0.00001282, 0.00000634, -0.00000411, 0.00000349, 0.00000954, 0.00000249, -0.00001227, -0.00000079, 0.00001248, -0.00000969, -0.00000280, -0.00000425, 0.00000533, -0.00000780, 0.00001004, -0.00000621, 0.00000853, -0.00000331, -0.00000702, 0.00000420, -0.00000854, 0.00002280, -0.00003138, 0.00000411, -0.00000011, 0.00001313, -0.00002962, 0.00001810, -0.00000956, 0.00000650, -0.00000325, -0.00000839, 0.00001311, -0.00000586, 0.00000429, -0.00000191, 0.00000704, -0.00000586, 0.00000242, -0.00000235, 0.00000651, -0.00000206, -0.00000095, -0.00000039, -0.00000024, -0.00000328, 0.00000412, -0.00000410, 0.00000371, -0.00000708, 0.00000155, -0.00000771, 0.00000136, -0.00000811, 0.00001499, -0.00001302, -0.00001010, 0.00000307, 0.00000891, 0.00003028, },
		{ -0.00000613, 0.00003909, -0.00001326, 0.00001785, -0.00002557, 0.00000523, -0.00000619, -0.00000142, 0.00000320, -0.00000713, -0.00000302, 0.00000485, -0.00000123, 0.00000378, -0.00000297, -0.00000145, -0.00000213, 0.00000282, 0.00000971, -0.00001677, -0.00000060, 0.00000348, 0.00000472, -0.00000447, 0.00000394, -0.00000750, -0.00000324, 0.00000146, -0.00000612, 0.00000588, -0.00000134, 0.00000252, -0.00000914, 0.00000266, -0.00000352, 0.00000265, 0.00000037, -0.00000414, -0.00001201, -0.00000298, -0.00002850, 0.00001115, -0.00002555, -0.00004687, -0.00023680, 0.00042114, 0.00029412, -0.00024061, 0.00000207, -0.00005522, -0.00000117, -0.00001465, 0.00000790, -0.00001310, 0.00000736, -0.00001318, 0.00000224, -0.00001336, 0.00000064, -0.00000193, 0.00000554, -0.00000864, 0.00000394, -0.00000410, 0.00000667, -0.00000624, 0.00000666, -0.00000894, -0.00000103, 0.00000215, -0.00000287, 0.00000995, 0.00000602, -0.00000648, -0.00001029, 0.00001281, -0.00000208, -0.00000710, -0.00000363, 0.00000215, -0.00000244, 0.00000167, 0.00000371, -0.00000346, 0.00001145, -0.00001842, 0.00001249, -0.00001749, 0.00002560, -0.00002062, -0.00000834, 0.00000224, 0.00001213, -0.00001671, -0.00000454, 0.00001177, -0.00001232, 0.00001372, -0.00002019, 0.00001566, -0.00000450, 0.00000440, -0.00000350, 0.00000772, -0.00000386, 0.00000078, -0.00000220, 0.00000475, 0.00000091, -0.00000130, -0.00000183, 0.00000215, -0.00000549, 0.00000462, -0.00000375, 0.00000324, -0.00000490, -0.00000091, -0.00000511, -0.00000143, -0.00000542, 0.00000589, 0.00000401, -0.00002281, 0.00000573, -0.00000051, 0.00003282, -0.00000670, },
		{ 0.00004261, -0.00000307, 0.00000957, -0.00001031, -0.00001223, 0.00000680, -0.00001198, 0.00000925, -0.00000850, -0.00000271, 0.00000066, 0.00000412, -0.00000234, 0.00000467, -0.00000751, 0.00000367, -0.00000626, 0.00001548, -0.00001121, -0.00000774, 0.00000269, 0.00000590, -0.00000214, 0.00000029, -0.00000070, -0.00000939, 0.00000431, -0.00000704, 0.00000374, 0.00000025, 0.00000277, -0.00000669, -0.00000111, -0.00000124, 0.00000069, 0.00000064, 0.00000096, -0.00001573, 0.00000258, -0.00003000, 0.00000694, -0.00002140, -0.00000856, -0.00021800, 0.00013954, 0.00049614, -0.00014719, -0.00008538, -0.00002000, -0.00003041, -0.00000279, -0.00000253, -0.00000350, -0.00000136, -0.00000420, -0.00000520, -0.00000619, -0.00000677, 0.00000056, 0.00000350, -0.00000408, -0.00000135, -0.00000047, 0.00000246, -0.00000051, -0.00000007, 0.00000088, -0.00000993, 0.00000652, -0.00000512, 0.00000661, 0.00000876, -0.00000038, -0.00001293, 0.00000447, 0.00000811, -0.00000978, -0.00000290, -0.00000254, 0.00000335, -0.00000528, 0.00001017, -0.00000972, 0.00001642, -0.00001668, 0.00000716, -0.00001124, 0.00001158, 0.00000207, -0.00002493, 0.00000810, 0.00000307, 0.00000295, -0.00002462, 0.00002191, -0.00001570, 0.00001272, -0.00001317, 0.00000279, 0.00000627, -0.00000166, 0.00000108, 0.00000198, 0.00000311, -0.00000406, 0.00000109, 0.00000007, 0.00000505, -0.00000220, -0.00000142, 0.00000103, -0.00000284, -0.00000001, 0.00000122, -0.00000174, 0.00000132, -0.00000635, -0.00000005, -0.00000663, 0.00000016, -0.00000429, 0.00001463, -0.00002162, -0.00000451, 0.00000287, 0.00001857, 0.00002103, -0.00002698, },
		{ 0.00002109, -0.00000546, 0.00000748, -0.00002586, 0.00001219, -0.00001398, 0.00000715, -0.00000446, -0.00000445, -0.00000238, 0.00000703, -0.00000458, 0.00000634, -0.00000637, 0.00000213, -0.00000533, 0.00000830, 0.00000308, -0.00001599, 0.00000229, 0.00000349, 0.00000367, -0.00000486, 0.00000465, -0.00001093, 0.00000194, -0.00000366, -0.00000085, 0.00000261, 0.00000175, -0.00000292, -0.00000453, -0.00000006, -0.00000022, 0.00000005, 0.00000405, -0.00001396, -0.00000041, -0.00002042, -0.00000745, -0.00000852, -0.00000484, -0.00013852, -0.00008326, 0.00049022, 0.00008306, -0.00017997, -0.00000117, -0.00004720, -0.00000032, -0.00000910, 0.00000310, -0.00000709, 0.00000146, -0.00000899, -0.00000259, -0.00000991, -0.00000024, 0.00000219, 0.00000013, -0.00000511, 0.00000152, -0.00000069, 0.00000356, -0.00000386, 0.00000499, -0.00000984, 0.00000165, 0.00000046, -0.00000124, 0.00001194, 0.00000257, -0.00000784, -0.00000753, 0.00001430, -0.00000734, -0.00000386, -0.00000580, 0.00000549, -0.00000671, 0.00000955, -0.00000725, 0.00001040, -0.00000413, -0.00000622, 0.00000176, -0.00000662, 0.00002099, -0.00003134, 0.00000560, -0.00000047, 0.00001203, -0.00002544, 0.00001024, -0.00000044, -0.00000291, 0.00000366, -0.00001302, 0.00001442, -0.00000458, 0.00000410, -0.00000230, 0.00000712, -0.00000510, 0.00000141, -0.00000165, 0.00000534, -0.00000021, -0.00000185, -0.00000027, 0.00000001, -0.00000369, 0.00000389, -0.00000372, 0.00000343, -0.00000645, -0.00000068, -0.00000586, -0.00000094, -0.00000468, 0.00000987, -0.00000461, -0.00002271, 0.00001048, 0.00000182, 0.00003857, -0.00002550, 0.00002348, },
		{ -0.00001237, 0.00001516, -0.00002601, 0.00000468, -0.00000593, -0.00000193, 0.00000287, -0.00000731, -0.00000308, 0.00000517, -0.00000067, 0.00000103, 0.00000110, -0.00000466, 0.00000152, -0.00000326, 0.00001455, -0.00001650, -0.00000238, 0.00000184, 0.00000705, -0.00000533, 0.00000377, -0.00000608, -0.00000429, 0.00000084, -0.00000435, 0.00000296, 0.00000180, -0.00000060, -0.00000507, -0.00000157, 0.00000074, -0.00000156, 0.00000547, -0.00000912, -0.00000612, -0.00000989, -0.00001840, -0.00000135, -0.00000849, -0.00006554, -0.00018429, 0.00032281, 0.00031521, -0.00017645, -0.00004526, -0.00002993, -0.00001988, -0.00000198, -0.00000271, -0.00000158, -0.00000305, -0.00000371, -0.00000679, -0.00000583, -0.00000622, 0.00000394, 0.00000042, -0.00000358, -0.00000146, 0.00000077, 0.00000205, -0.00000095, 0.00000100, -0.00000166, -0.00000804, 0.00000734, -0.00000693, 0.00001142, 0.00000506, -0.00000109, -0.00001431, 0.00001078, 0.00000102, -0.00000643, -0.00000628, 0.00000272, -0.00000274, 0.00000342, 0.00000069, -0.00000060, 0.00000995, -0.00001655, 0.00000946, -0.00001567, 0.00002410, -0.00002043, -0.00000863, 0.00000502, 0.00000757, -0.00001000, -0.00001184, 0.00001662, -0.00001437, 0.00001175, -0.00001600, 0.00000938, 0.00000173, 0.00000092, -0.00000056, 0.00000436, -0.00000010, -0.00000261, 0.00000068, 0.00000155, 0.00000406, -0.00000312, -0.00000031, 0.00000033, -0.00000321, 0.00000143, -0.00000043, -0.00000025, -0.00000068, -0.00000635, -0.00000045, -0.00000681, 0.00000104, -0.00000155, 0.00001264, -0.00003085, 0.00000605, -0.00000143, 0.00003334, -0.00000033, -0.00000785, 0.00004359, },
		{ 0.00000826, -0.00001294, -0.00001117, 0.00000538, -0.00001087, 0.00000759, -0.00000776, -0.00000343, 0.00000099, 0.00000512, -0.00000482, 0.00000682, -0.00000793, 0.00000430, -0.00000750, 0.00001339, -0.00000536, -0.00001173, 0.00000262, 0.00000519, 0.00000032, -0.00000230, 0.00000164, -0.00000989, 0.00000289, -0.00000499, 0.00000137, 0.00000207, 0.00000119, -0.00000451, -0.00000290, 0.00000035, -0.00000096, 0.00000311, -0.00000142, -0.00001221, -0.00000204, -0.00002290, -0.00000097, -0.00001123, -0.00001961, -0.00018658, 0.00011085, 0.00042534, -0.00004195, -0.00012685, -0.00000664, -0.00004020, 0.00000345, -0.00000933, 0.00000456, -0.00000765, 0.00000109, -0.00000984, -0.00000297, -0.00000970, 0.00000294, 0.00000118, -0.00000135, -0.00000435, 0.00000195, -0.00000006, 0.00000263, -0.00000313, 0.00000430, -0.00001144, 0.00000581, -0.00000366, 0.00000389, 0.00000996, 0.00000181, -0.00001133, -0.00000123, 0.00001077, -0.00000775, -0.00000510, -0.00000291, 0.00000298, -0.00000338, 0.00000799, -0.00000864, 0.00001658, -0.00001605, 0.00000591, -0.00001152, 0.00001187, 0.00000118, -0.00002517, 0.00001062, 0.00000036, 0.00000534, -0.00002443, 0.00001751, -0.00000873, 0.00000368, -0.00000469, -0.00000490, 0.00001122, -0.00000334, 0.00000300, -0.00000043, 0.00000540, -0.00000560, 0.00000217, -0.00000127, 0.00000594, -0.00000205, -0.00000103, -0.00000011, -0.00000114, -0.00000225, 0.00000311, -0.00000361, 0.00000360, -0.00000886, 0.00000104, -0.00000824, 0.00000184, -0.00000534, 0.00001452, -0.00001825, -0.00001412, 0.00000854, 0.00001261, 0.00003073, -0.00002899, 0.00004631, -0.00000085, },
		{ 0.00000371, -0.00002471, 0.00001062, -0.00001281, 0.00000566, -0.00000403, -0.00000499, -0.00000222, 0.00000764, -0.00000515, 0.00000513, -0.00000335, -0.00000051, -0.00000237, 0.00000226, 0.00000965, -0.00001809, 0.00000149, 0.00000181, 0.00000659, -0.00000670, 0.00000536, -0.00000985, 0.00000025, -0.00000281, -0.00000089, 0.00000148, 0.00000321, -0.00000415, -0.00000248, -0.00000237, 0.00000200, -0.00000175, 0.00000634, -0.00001566, 0.00000133, -0.00002179, -0.00000413, -0.00001275, 0.00000188, -0.00013922, -0.00005443, 0.00038538, 0.00015378, -0.00016423, -0.00002479, -0.00003409, -0.00001035, -0.00000348, -0.00000060, -0.00000195, -0.00000353, -0.00000424, -0.00000767, -0.00000564, -0.00000419, 0.00000543, -0.00000242, -0.00000268, -0.00000138, 0.00000180, 0.00000145, -0.00000118, 0.00000193, -0.00000471, -0.00000435, 0.00000559, -0.00000587, 0.00001391, 0.00000229, -0.00000304, -0.00001290, 0.00001400, -0.00000413, -0.00000463, -0.00000752, 0.00000564, -0.00000598, 0.00000962, -0.00000878, 0.00001249, -0.00000501, -0.00000544, -0.00000049, -0.00000488, 0.00001910, -0.00003072, 0.00000599, 0.00000083, 0.00000885, -0.00001905, 0.00000144, 0.00000765, -0.00000954, 0.00000701, -0.00001395, 0.00001215, -0.00000069, 0.00000208, -0.00000091, 0.00000538, -0.00000278, -0.00000090, 0.00000023, 0.00000301, 0.00000245, -0.00000330, 0.00000057, -0.00000077, -0.00000275, 0.00000214, -0.00000179, 0.00000148, -0.00000395, -0.00000469, -0.00000240, -0.00000505, 0.00000077, 0.00000288, 0.00000481, -0.00003383, 0.00001426, -0.00000257, 0.00004388, -0.00002225, 0.00002343, 0.00002539, -0.00000545, },
		{ -0.00002670, 0.00000471, -0.00000640, -0.00000174, 0.00000179, -0.00000679, -0.00000362, 0.00000565, -0.00000006, -0.00000153, 0.00000432, -0.00000667, 0.00000378, -0.00000694, 0.00001583, -0.00001316, -0.00000574, 0.00000146, 0.00000749, -0.00000401, 0.00000172, -0.00000359, -0.00000562, 0.00000033, -0.00000279, 0.00000053, 0.00000403, -0.00000257, -0.00000233, -0.00000420, 0.00000315, -0.00000384, 0.00000850, -0.00001201, -0.00000230, -0.00001458, -0.00001080, -0.00001036, 0.00000344, -0.00007743, -0.00014415, 0.00025245, 0.00030326, -0.00009906, -0.00008943, -0.00000985, -0.00003407, 0.00000713, -0.00001100, 0.00000799, -0.00001129, 0.00000358, -0.00001360, -0.00000044, -0.00001104, 0.00000797, -0.00000325, 0.00000032, -0.00000567, 0.00000440, -0.00000153, 0.00000333, -0.00000352, 0.00000390, -0.00001251, 0.00000969, -0.00000816, 0.00001049, 0.00000558, 0.00000243, -0.00001527, 0.00000646, 0.00000447, -0.00000549, -0.00000856, 0.00000275, -0.00000282, 0.00000489, -0.00000177, 0.00000145, 0.00000926, -0.00001545, 0.00000717, -0.00001435, 0.00002272, -0.00001999, -0.00000927, 0.00000805, 0.00000322, -0.00000442, -0.00001622, 0.00001716, -0.00001187, 0.00000610, -0.00000924, 0.00000156, 0.00000821, -0.00000231, 0.00000233, 0.00000108, 0.00000328, -0.00000546, 0.00000300, -0.00000107, 0.00000639, -0.00000419, 0.00000083, -0.00000137, -0.00000094, -0.00000161, 0.00000253, -0.00000325, 0.00000292, -0.00001076, 0.00000288, -0.00001081, 0.00000575, -0.00000640, 0.00001736, -0.00003320, 0.00000078, 0.00000076, 0.00003090, 0.00000864, -0.00001030, 0.00004825, -0.00001065, 0.00001175, },
		{ -0.00000968, 0.00000358, -0.00000947, 0.00000561, -0.00000676, -0.00000428, 0.00000143, 0.00000576, -0.00000630, 0.00000730, -0.00000659, 0.00000307, -0.00000628, 0.00000887, 0.00000139, -0.00001532, 0.00000258, 0.00000373, 0.00000340, -0.00000488, 0.00000331, -0.00000964, 0.00000129, -0.00000335, -0.00000004, 0.00000250, 0.00000118, -0.00000404, -0.00000276, -0.00000010, -0.00000056, 0.00000337, -0.00000148, -0.00001122, -0.00000360, -0.00001962, -0.00000373, -0.00000776, -0.00002248, -0.00016744, 0.00010046, 0.00034718, 0.00004612, -0.00014469, -0.00000907, -0.00003933, 0.00000196, -0.00001028, 0.00000725, -0.00000871, 0.00000161, -0.00001013, -0.00000360, -0.00000920, 0.00000233, 0.00000145, -0.00000105, -0.00000480, 0.00000183, -0.00000020, 0.00000329, -0.00000347, 0.00000459, -0.00000961, 0.00000227, -0.00000002, -0.00000025, 0.00001189, 0.00000261, -0.00000799, -0.00000703, 0.00001214, -0.00000504, -0.00000686, -0.00000376, 0.00000287, -0.00000204, 0.00000662, -0.00000844, 0.00001740, -0.00001596, 0.00000508, -0.00001195, 0.00001212, 0.00000030, -0.00002503, 0.00001221, -0.00000071, 0.00000523, -0.00002086, 0.00001009, -0.00000024, -0.00000501, 0.00000220, -0.00001028, 0.00001328, -0.00000255, 0.00000311, -0.00000109, 0.00000583, -0.00000537, 0.00000167, -0.00000111, 0.00000529, -0.00000066, -0.00000152, -0.00000037, -0.00000053, -0.00000306, 0.00000333, -0.00000368, 0.00000389, -0.00000900, -0.00000033, -0.00000738, 0.00000068, -0.00000297, 0.00001054, -0.00001107, -0.00002536, 0.00001310, 0.00000730, 0.00003933, -0.00002853, 0.00004801, 0.00000332, 0.00000611, -0.00001589, },
		{ 0.00000936, -0.00001211, 0.00000460, -0.00000407, -0.00000506, -0.00000234, 0.00000813, -0.00000515, 0.00000322, -0.00000008, -0.00000314, 0.00000057, -0.00000286, 0.00001376, -0.00001751, -0.00000075, 0.00000101, 0.00000812, -0.00000667, 0.00000420, -0.00000756, -0.00000176, -0.00000210, -0.00000055, -0.00000004, 0.00000485, -0.00000543, -0.00000043, -0.00000451, 0.00000392, -0.00000325, 0.00000830, -0.00001714, 0.00000317, -0.00002355, -0.00000005, -0.00001772, 0.00000880, -0.00014038, -0.00002771, 0.00029943, 0.00019073, -0.00012487, -0.00005675, -0.00001767, -0.00002212, 0.00000336, -0.00000594, 0.00000489, -0.00000978, 0.00000145, -0.00001336, -0.00000097, -0.00000837, 0.00000884, -0.00000533, 0.00000036, -0.00000485, 0.00000476, -0.00000131, 0.00000227, -0.00000190, 0.00000080, -0.00000982, 0.00000958, -0.00000895, 0.00001434, 0.00000268, 0.00000090, -0.00001590, 0.00001137, -0.00000027, -0.00000506, -0.00000923, 0.00000558, -0.00000536, 0.00000985, -0.00001027, 0.00001425, -0.00000552, -0.00000511, -0.00000219, -0.00000357, 0.00001729, -0.00002966, 0.00000554, 0.00000320, 0.00000458, -0.00001202, -0.00000632, 0.00001284, -0.00001213, 0.00000633, -0.00001133, 0.00000689, 0.00000497, -0.00000096, 0.00000152, 0.00000264, 0.00000027, -0.00000371, 0.00000254, 0.00000032, 0.00000516, -0.00000472, 0.00000165, -0.00000210, -0.00000103, -0.00000043, 0.00000088, -0.00000119, -0.00000069, -0.00000919, 0.00000130, -0.00000948, 0.00000642, -0.00000370, 0.00001254, -0.00004081, 0.00001283, -0.00000375, 0.00004592, -0.00001545, 0.00002139, 0.00003106, -0.00000553, -0.00000056, -0.00002357, },
		{ -0.00000741, -0.00000109, 0.00000021, -0.00000582, -0.00000443, 0.00000621, 0.00000044, -0.00000350, 0.00000621, -0.00000707, 0.00000423, -0.00000792, 0.00001383, -0.00000761, -0.00000983, 0.00000180, 0.00000653, -0.00000121, -0.00000124, -0.00000100, -0.00000657, -0.00000035, -0.00000141, -0.00000127, 0.00000516, -0.00000320, -0.00000111, -0.00000506, 0.00000362, -0.00000413, 0.00000941, -0.00001272, -0.00000075, -0.00001666, -0.00000637, -0.00001505, 0.00000942, -0.00008184, -0.00011647, 0.00020625, 0.00026871, -0.00002096, -0.00012166, 0.00000006, -0.00004068, 0.00001054, -0.00001518, 0.00001388, -0.00001578, 0.00000732, -0.00001700, 0.00000184, -0.00001301, 0.00000951, -0.00000482, 0.00000235, -0.00000794, 0.00000605, -0.00000330, 0.00000577, -0.00000594, 0.00000684, -0.00001379, 0.00000893, -0.00000661, 0.00000741, 0.00000737, 0.00000418, -0.00001373, 0.00000093, 0.00000739, -0.00000399, -0.00001048, 0.00000226, -0.00000263, 0.00000588, -0.00000346, 0.00000256, 0.00000937, -0.00001509, 0.00000558, -0.00001342, 0.00002138, -0.00001938, -0.00000998, 0.00001075, -0.00000016, -0.00000081, -0.00001713, 0.00001363, -0.00000597, -0.00000147, -0.00000181, -0.00000597, 0.00001329, -0.00000402, 0.00000412, -0.00000113, 0.00000540, -0.00000700, 0.00000409, -0.00000247, 0.00000738, -0.00000411, 0.00000121, -0.00000246, 0.00000065, -0.00000372, 0.00000431, -0.00000493, 0.00000502, -0.00001323, 0.00000403, -0.00001255, 0.00000773, -0.00000771, 0.00001753, -0.00003007, -0.00000858, 0.00000409, 0.00002710, 0.00001867, -0.00001208, 0.00005212, -0.00000757, 0.00000742, -0.00002752, 0.00000521, },
		{ -0.00000827, 0.00000377, -0.00000594, -0.00000491, 0.00000175, 0.00000615, -0.00000694, 0.00000648, -0.00000409, 0.00000076, -0.00000370, 0.00000355, 0.00000731, -0.00001726, 0.00000169, 0.00000243, 0.00000596, -0.00000632, 0.00000351, -0.00000825, -0.00000064, -0.00000197, -0.00000075, 0.00000194, 0.00000213, -0.00000454, -0.00000159, -0.00000156, 0.00000081, 0.00000258, -0.00000045, -0.00001152, -0.00000340, -0.00001871, -0.00000323, -0.00000837, -0.00002069, -0.00015545, 0.00009964, 0.00027285, 0.00010966, -0.00013822, -0.00002487, -0.00003046, -0.00000500, -0.00000714, 0.00000595, -0.00000584, -0.00000130, -0.00000739, -0.00000681, -0.00000648, -0.00000016, 0.00000333, -0.00000238, -0.00000344, -0.00000008, 0.00000131, 0.00000217, -0.00000182, 0.00000265, -0.00000568, -0.00000263, 0.00000429, -0.00000435, 0.00001339, 0.00000312, -0.00000427, -0.00001158, 0.00001156, -0.00000160, -0.00000818, -0.00000493, 0.00000286, -0.00000111, 0.00000581, -0.00000871, 0.00001843, -0.00001599, 0.00000429, -0.00001214, 0.00001204, -0.00000047, -0.00002450, 0.00001285, -0.00000029, 0.00000315, -0.00001506, 0.00000155, 0.00000750, -0.00001130, 0.00000603, -0.00001238, 0.00001197, 0.00000076, 0.00000147, -0.00000008, 0.00000458, -0.00000368, -0.00000009, 0.00000022, 0.00000350, 0.00000154, -0.00000259, 0.00000008, -0.00000091, -0.00000257, 0.00000208, -0.00000224, 0.00000256, -0.00000731, -0.00000346, -0.00000485, -0.00000242, 0.00000154, 0.00000454, -0.00000267, -0.00003513, 0.00001409, 0.00000407, 0.00004568, -0.00002484, 0.00004748, 0.00000945, 0.00000314, -0.00001903, -0.00000809, 0.00000183, },
		{ 0.00000402, -0.00000463, -0.00000465, -0.00000273, 0.00000856, -0.00000486, 0.00000118, 0.00000269, -0.00000492, 0.00000248, -0.00000598, 0.00001475, -0.00001425, -0.00000428, 0.00000114, 0.00000813, -0.00000482, 0.00000156, -0.00000462, -0.00000367, -0.00000166, 0.00000001, -0.00000157, 0.00000611, -0.00000612, 0.00000090, -0.00000574, 0.00000478, -0.00000367, 0.00000912, -0.00001754, 0.00000407, -0.00002439, 0.00000303, -0.00002110, 0.00001313, -0.00013847, -0.00000759, 0.00023474, 0.00019829, -0.00007114, -0.00008865, -0.00000381, -0.00003144, 0.00000830, -0.00001034, 0.00001101, -0.00001500, 0.00000601, -0.00001776, 0.00000242, -0.00001131, 0.00001116, -0.00000745, 0.00000285, -0.00000770, 0.00000702, -0.00000352, 0.00000521, -0.00000509, 0.00000513, -0.00001334, 0.00001127, -0.00000963, 0.00001268, 0.00000422, 0.00000339, -0.00001633, 0.00000689, 0.00000361, -0.00000490, -0.00001083, 0.00000518, -0.00000468, 0.00000992, -0.00001134, 0.00001534, -0.00000538, -0.00000541, -0.00000314, -0.00000279, 0.00001561, -0.00002830, 0.00000459, 0.00000599, 0.00000024, -0.00000573, -0.00001163, 0.00001415, -0.00001056, 0.00000231, -0.00000623, -0.00000002, 0.00001099, -0.00000382, 0.00000392, -0.00000006, 0.00000303, -0.00000611, 0.00000444, -0.00000192, 0.00000718, -0.00000549, 0.00000244, -0.00000339, 0.00000076, -0.00000293, 0.00000332, -0.00000354, 0.00000216, -0.00001290, 0.00000398, -0.00001284, 0.00001059, -0.00000797, 0.00001663, -0.00004225, 0.00000615, -0.00000254, 0.00004544, -0.00000596, 0.00001838, 0.00003734, -0.00000526, -0.00000545, -0.00002252, 0.00000853, -0.00001198, },
		{ -0.00000148, -0.00000479, -0.00000520, 0.00000669, 0.00000079, -0.00000475, 0.00000670, -0.00000593, 0.00000304, -0.00000668, 0.00000964, -0.00000129, -0.00001346, 0.00000201, 0.00000501, 0.00000200, -0.00000396, 0.00000071, -0.00000658, -0.00000144, -0.00000014, -0.00000245, 0.00000524, -0.00000265, -0.00000116, -0.00000450, 0.00000261, -0.00000295, 0.00000873, -0.00001179, -0.00000097, -0.00001656, -0.00000477, -0.00001587, 0.00001031, -0.00008008, -0.00009880, 0.00017787, 0.00022278, 0.00004704, -0.00013624, -0.00000215, -0.00003909, 0.00000806, -0.00001499, 0.00001570, -0.00001621, 0.00000733, -0.00001696, 0.00000105, -0.00001226, 0.00000868, -0.00000438, 0.00000249, -0.00000816, 0.00000564, -0.00000317, 0.00000621, -0.00000613, 0.00000711, -0.00001217, 0.00000567, -0.00000308, 0.00000312, 0.00000962, 0.00000491, -0.00001075, -0.00000452, 0.00000893, -0.00000170, -0.00001202, 0.00000129, -0.00000220, 0.00000636, -0.00000441, 0.00000290, 0.00001005, -0.00001524, 0.00000447, -0.00001264, 0.00001992, -0.00001857, -0.00001055, 0.00001277, -0.00000220, 0.00000063, -0.00001491, 0.00000720, 0.00000145, -0.00000883, 0.00000438, -0.00001156, 0.00001573, -0.00000341, 0.00000420, -0.00000172, 0.00000582, -0.00000696, 0.00000375, -0.00000249, 0.00000692, -0.00000286, 0.00000077, -0.00000272, 0.00000124, -0.00000452, 0.00000457, -0.00000495, 0.00000530, -0.00001353, 0.00000290, -0.00001197, 0.00000693, -0.00000566, 0.00001387, -0.00002310, -0.00001939, 0.00000602, 0.00002366, 0.00002826, -0.00001194, 0.00005449, -0.00000280, 0.00000204, -0.00002832, 0.00000590, -0.00000860, -0.00000038, },
		{ -0.00000556, -0.00000512, 0.00000182, 0.00000639, -0.00000695, 0.00000486, -0.00000118, -0.00000170, -0.00000099, -0.00000108, 0.00001107, -0.00001680, -0.00000049, 0.00000187, 0.00000733, -0.00000606, 0.00000205, -0.00000586, -0.00000272, -0.00000087, -0.00000093, 0.00000081, 0.00000345, -0.00000528, -0.00000024, -0.00000310, 0.00000216, 0.00000176, 0.00000060, -0.00001200, -0.00000267, -0.00001869, -0.00000143, -0.00001044, -0.00001758, -0.00014598, 0.00010136, 0.00020985, 0.00014639, -0.00011119, -0.00004868, -0.00001801, -0.00001400, -0.00000251, 0.00000279, -0.00000105, -0.00000570, -0.00000343, -0.00001087, -0.00000313, -0.00000311, 0.00000561, -0.00000426, -0.00000129, -0.00000273, 0.00000343, 0.00000029, 0.00000072, -0.00000021, -0.00000118, -0.00000728, 0.00000775, -0.00000710, 0.00001343, 0.00000420, -0.00000122, -0.00001406, 0.00000896, 0.00000225, -0.00000891, -0.00000631, 0.00000281, -0.00000042, 0.00000527, -0.00000904, 0.00001923, -0.00001578, 0.00000321, -0.00001181, 0.00001148, -0.00000103, -0.00002366, 0.00001272, 0.00000114, -0.00000005, -0.00000845, -0.00000628, 0.00001270, -0.00001397, 0.00000629, -0.00001120, 0.00000775, 0.00000584, -0.00000117, 0.00000191, 0.00000240, -0.00000128, -0.00000237, 0.00000203, 0.00000126, 0.00000387, -0.00000369, 0.00000083, -0.00000187, -0.00000127, 0.00000001, -0.00000002, 0.00000045, -0.00000478, -0.00000721, -0.00000186, -0.00000609, 0.00000649, -0.00000130, 0.00000428, -0.00004086, 0.00001012, 0.00000324, 0.00004953, -0.00001795, 0.00004511, 0.00001713, -0.00000041, -0.00002232, -0.00000661, 0.00000043, -0.00000756, 0.00000235, },
		{ -0.00000399, -0.00000322, 0.00000888, -0.00000443, -0.00000057, 0.00000444, -0.00000539, 0.00000287, -0.00000677, 0.00001286, -0.00000907, -0.00000834, 0.00000174, 0.00000703, -0.00000180, -0.00000168, -0.00000191, -0.00000491, -0.00000171, 0.00000077, -0.00000289, 0.00000671, -0.00000594, 0.00000123, -0.00000581, 0.00000438, -0.00000285, 0.00000860, -0.00001665, 0.00000371, -0.00002378, 0.00000426, -0.00002184, 0.00001374, -0.00013221, 0.00000464, 0.00018964, 0.00018422, -0.00001344, -0.00011278, 0.00000297, -0.00003544, 0.00000933, -0.00001208, 0.00001474, -0.00001767, 0.00000811, -0.00001973, 0.00000358, -0.00001220, 0.00001170, -0.00000812, 0.00000407, -0.00000918, 0.00000783, -0.00000445, 0.00000683, -0.00000674, 0.00000740, -0.00001423, 0.00001033, -0.00000786, 0.00000917, 0.00000668, 0.00000456, -0.00001472, 0.00000156, 0.00000668, -0.00000391, -0.00001229, 0.00000439, -0.00000383, 0.00000963, -0.00001177, 0.00001564, -0.00000461, -0.00000634, -0.00000339, -0.00000246, 0.00001401, -0.00002671, 0.00000345, 0.00000863, -0.00000340, -0.00000104, -0.00001384, 0.00001170, -0.00000583, -0.00000351, -0.00000033, -0.00000689, 0.00001585, -0.00000529, 0.00000531, -0.00000177, 0.00000469, -0.00000739, 0.00000528, -0.00000312, 0.00000801, -0.00000523, 0.00000258, -0.00000417, 0.00000198, -0.00000465, 0.00000476, -0.00000477, 0.00000374, -0.00001495, 0.00000480, -0.00001424, 0.00001227, -0.00000892, 0.00001640, -0.00003834, -0.00000424, -0.00000087, 0.00004380, 0.00000488, 0.00001567, 0.00004336, -0.00000418, -0.00001110, -0.00002149, 0.00000802, -0.00001224, 0.00000374, -0.00000550, },
		{ -0.00000566, 0.00000693, 0.00000105, -0.00000531, 0.00000601, -0.00000371, 0.00000084, -0.00000419, 0.00000470, 0.00000425, -0.00001549, 0.00000134, 0.00000371, 0.00000466, -0.00000548, 0.00000089, -0.00000542, -0.00000294, 0.00000103, -0.00000310, 0.00000451, -0.00000130, -0.00000198, -0.00000317, 0.00000086, -0.00000107, 0.00000727, -0.00001008, -0.00000212, -0.00001509, -0.00000507, -0.00001404, 0.00000789, -0.00007458, -0.00008752, 0.00016041, 0.00017539, 0.00009762, -0.00013152, -0.00001520, -0.00003101, 0.00000121, -0.00001160, 0.00001432, -0.00001342, 0.00000451, -0.00001436, -0.00000189, -0.00000967, 0.00000630, -0.00000261, 0.00000127, -0.00000684, 0.00000370, -0.00000165, 0.00000513, -0.00000459, 0.00000541, -0.00000865, 0.00000114, 0.00000109, -0.00000109, 0.00001128, 0.00000552, -0.00000753, -0.00000870, 0.00000857, 0.00000129, -0.00001306, -0.00000010, -0.00000160, 0.00000640, -0.00000479, 0.00000275, 0.00001093, -0.00001557, 0.00000354, -0.00001173, 0.00001815, -0.00001756, -0.00001090, 0.00001401, -0.00000296, 0.00000026, -0.00001051, -0.00000045, 0.00000832, -0.00001410, 0.00000795, -0.00001425, 0.00001502, -0.00000039, 0.00000261, -0.00000075, 0.00000474, -0.00000559, 0.00000224, -0.00000140, 0.00000535, -0.00000081, -0.00000023, -0.00000230, 0.00000088, -0.00000409, 0.00000345, -0.00000356, 0.00000406, -0.00001210, 0.00000009, -0.00000972, 0.00000412, -0.00000139, 0.00000811, -0.00001482, -0.00002861, 0.00000432, 0.00002170, 0.00003642, -0.00000916, 0.00005516, 0.00000365, -0.00000436, -0.00002900, 0.00000650, -0.00000960, -0.00000000, -0.00000289, -0.00000404, },
		{ 0.00000169, 0.00000652, -0.00000657, 0.00000295, 0.00000142, -0.00000351, 0.00000089, -0.00000397, 0.00001205, -0.00001392, -0.00000379, 0.00000206, 0.00000739, -0.00000419, -0.00000068, -0.00000307, -0.00000450, -0.00000023, -0.00000066, -0.00000058, 0.00000467, -0.00000569, 0.00000067, -0.00000407, 0.00000286, 0.00000156, 0.00000100, -0.00001194, -0.00000225, -0.00001849, 0.00000018, -0.00001198, -0.00001558, -0.00013587, 0.00010109, 0.00016160, 0.00015862, -0.00007064, -0.00007371, -0.00000684, -0.00002158, 0.00000106, -0.00000009, 0.00000361, -0.00000968, -0.00000002, -0.00001419, -0.00000055, -0.00000529, 0.00000722, -0.00000571, 0.00000064, -0.00000508, 0.00000513, -0.00000132, 0.00000303, -0.00000274, 0.00000251, -0.00001038, 0.00000930, -0.00000773, 0.00001158, 0.00000622, 0.00000068, -0.00001436, 0.00000485, 0.00000583, -0.00000878, -0.00000785, 0.00000259, 0.00000021, 0.00000470, -0.00000908, 0.00001951, -0.00001512, 0.00000170, -0.00001085, 0.00001038, -0.00000139, -0.00002259, 0.00001209, 0.00000305, -0.00000348, -0.00000231, -0.00001198, 0.00001439, -0.00001285, 0.00000352, -0.00000764, 0.00000179, 0.00001141, -0.00000371, 0.00000389, 0.00000024, 0.00000091, -0.00000436, 0.00000355, -0.00000066, 0.00000566, -0.00000425, 0.00000139, -0.00000288, 0.00000015, -0.00000209, 0.00000208, -0.00000148, -0.00000249, -0.00001040, 0.00000039, -0.00000902, 0.00001024, -0.00000511, 0.00000788, -0.00004120, 0.00000133, 0.00000385, 0.00005156, -0.00000858, 0.00004186, 0.00002561, -0.00000413, -0.00002586, -0.00000533, -0.00000051, -0.00000742, 0.00000140, -0.00000562, -0.00000494, },
		{ 0.00000898, -0.00000396, -0.00000183, 0.00000503, -0.00000452, 0.00000184, -0.00000559, 0.00000900, -0.00000326, -0.00001185, 0.00000207, 0.00000555, 0.00000147, -0.00000446, -0.00000030, -0.00000502, -0.00000238, 0.00000169, -0.00000391, 0.00000660, -0.00000491, 0.00000067, -0.00000494, 0.00000302, -0.00000110, 0.00000710, -0.00001478, 0.00000235, -0.00002185, 0.00000365, -0.00001997, 0.00001094, -0.00012213, 0.00001008, 0.00015989, 0.00015755, 0.00003880, -0.00012378, 0.00000053, -0.00003324, 0.00000600, -0.00001074, 0.00001554, -0.00001732, 0.00000742, -0.00001908, 0.00000239, -0.00001102, 0.00001046, -0.00000729, 0.00000389, -0.00000907, 0.00000703, -0.00000391, 0.00000690, -0.00000660, 0.00000747, -0.00001266, 0.00000722, -0.00000436, 0.00000465, 0.00000931, 0.00000507, -0.00001204, -0.00000341, 0.00000819, -0.00000198, -0.00001347, 0.00000317, -0.00000280, 0.00000892, -0.00001155, 0.00001527, -0.00000339, -0.00000768, -0.00000313, -0.00000237, 0.00001236, -0.00002494, 0.00000233, 0.00001074, -0.00000594, 0.00000172, -0.00001312, 0.00000646, 0.00000040, -0.00000923, 0.00000465, -0.00001220, 0.00001832, -0.00000460, 0.00000510, -0.00000200, 0.00000486, -0.00000728, 0.00000487, -0.00000309, 0.00000752, -0.00000392, 0.00000199, -0.00000423, 0.00000231, -0.00000521, 0.00000484, -0.00000454, 0.00000371, -0.00001509, 0.00000360, -0.00001354, 0.00001132, -0.00000667, 0.00001247, -0.00003070, -0.00001566, -0.00000116, 0.00004243, 0.00001580, 0.00001442, 0.00004849, -0.00000202, -0.00001761, -0.00002038, 0.00000763, -0.00001260, 0.00000342, -0.00000634, -0.00000339, -0.00000354, },
		{ 0.00000123, -0.00000531, 0.00000454, -0.00000105, -0.00000153, -0.00000155, 0.00000037, 0.00000781, -0.00001528, -0.00000056, 0.00000311, 0.00000618, -0.00000532, -0.00000055, -0.00000333, -0.00000457, 0.00000192, -0.00000322, 0.00000319, 0.00000042, -0.00000299, -0.00000172, -0.00000092, 0.00000076, 0.00000581, -0.00000837, -0.00000340, -0.00001314, -0.00000623, -0.00001102, 0.00000417, -0.00006788, -0.00007909, 0.00014793, 0.00013364, 0.00012771, -0.00010980, -0.00003510, -0.00001986, -0.00000735, -0.00000697, 0.00001134, -0.00000896, 0.00000036, -0.00001065, -0.00000564, -0.00000647, 0.00000348, -0.00000045, -0.00000048, -0.00000477, 0.00000107, 0.00000041, 0.00000337, -0.00000223, 0.00000284, -0.00000454, -0.00000325, 0.00000456, -0.00000404, 0.00001145, 0.00000677, -0.00000495, -0.00001097, 0.00000636, 0.00000452, -0.00001332, -0.00000187, -0.00000093, 0.00000615, -0.00000485, 0.00000249, 0.00001166, -0.00001577, 0.00000253, -0.00001049, 0.00001600, -0.00001634, -0.00001103, 0.00001455, -0.00000275, -0.00000125, -0.00000519, -0.00000760, 0.00001294, -0.00001612, 0.00000839, -0.00001396, 0.00001148, 0.00000441, 0.00000004, 0.00000112, 0.00000282, -0.00000357, 0.00000022, 0.00000018, 0.00000330, 0.00000143, -0.00000130, -0.00000157, -0.00000006, -0.00000290, 0.00000155, -0.00000143, 0.00000205, -0.00000983, -0.00000338, -0.00000689, 0.00000055, 0.00000347, 0.00000235, -0.00000781, -0.00003377, -0.00000217, 0.00002126, 0.00004304, -0.00000375, 0.00005453, 0.00001132, -0.00001149, -0.00002956, 0.00000683, -0.00001015, -0.00000019, -0.00000381, -0.00000377, -0.00000566, 0.00000688, },
		{ -0.00000595, 0.00000115, 0.00000322, -0.00000418, 0.00000145, -0.00000478, 0.00001042, -0.00000932, -0.00000750, 0.00000257, 0.00000651, -0.00000131, -0.00000378, -0.00000074, -0.00000543, -0.00000024, 0.00000003, -0.00000202, 0.00000549, -0.00000550, 0.00000088, -0.00000424, 0.00000272, 0.00000214, 0.00000058, -0.00001113, -0.00000244, -0.00001757, 0.00000077, -0.00001192, -0.00001584, -0.00012373, 0.00009701, 0.00012787, 0.00015192, -0.00002501, -0.00009342, -0.00000088, -0.00002522, 0.00000180, -0.00000122, 0.00000669, -0.00001187, 0.00000166, -0.00001576, 0.00000039, -0.00000596, 0.00000754, -0.00000616, 0.00000170, -0.00000640, 0.00000572, -0.00000199, 0.00000438, -0.00000409, 0.00000454, -0.00001130, 0.00000860, -0.00000617, 0.00000806, 0.00000893, 0.00000162, -0.00001302, 0.00000025, 0.00000833, -0.00000755, -0.00000945, 0.00000212, 0.00000088, 0.00000391, -0.00000864, 0.00001917, -0.00001403, -0.00000019, -0.00000932, 0.00000886, -0.00000162, -0.00002134, 0.00001122, 0.00000490, -0.00000639, 0.00000248, -0.00001488, 0.00001257, -0.00000879, -0.00000092, -0.00000318, -0.00000439, 0.00001602, -0.00000501, 0.00000494, -0.00000101, 0.00000214, -0.00000542, 0.00000421, -0.00000173, 0.00000641, -0.00000393, 0.00000141, -0.00000349, 0.00000112, -0.00000353, 0.00000335, -0.00000247, -0.00000124, -0.00001220, 0.00000112, -0.00001033, 0.00001179, -0.00000587, 0.00000741, -0.00003642, -0.00001060, 0.00000392, 0.00005299, 0.00000213, 0.00003889, 0.00003411, -0.00000760, -0.00002977, -0.00000421, -0.00000108, -0.00000771, 0.00000075, -0.00000596, -0.00000455, 0.00000151, 0.00000647, },
		{ -0.00000256, 0.00000458, -0.00000266, -0.00000010, -0.00000328, 0.00000449, 0.00000180, -0.00001381, 0.00000148, 0.00000436, 0.00000412, -0.00000593, -0.00000032, -0.00000391, -0.00000355, 0.00000265, -0.00000455, 0.00000582, -0.00000327, -0.00000042, -0.00000356, 0.00000123, 0.00000101, 0.00000516, -0.00001252, 0.00000052, -0.00001910, 0.00000174, -0.00001627, 0.00000592, -0.00010982, 0.00001113, 0.00014027, 0.00012665, 0.00007883, -0.00011963, -0.00001049, -0.00002601, -0.00000059, -0.00000709, 0.00001396, -0.00001452, 0.00000457, -0.00001644, -0.00000046, -0.00000843, 0.00000805, -0.00000547, 0.00000266, -0.00000771, 0.00000498, -0.00000229, 0.00000576, -0.00000505, 0.00000589, -0.00000947, 0.00000302, -0.00000029, 0.00000031, 0.00001114, 0.00000576, -0.00000934, -0.00000703, 0.00000777, 0.00000070, -0.00001415, 0.00000150, -0.00000160, 0.00000782, -0.00001082, 0.00001448, -0.00000204, -0.00000915, -0.00000261, -0.00000230, 0.00001057, -0.00002302, 0.00000137, 0.00001218, -0.00000731, 0.00000276, -0.00001027, -0.00000007, 0.00000622, -0.00001315, 0.00000745, -0.00001501, 0.00001787, -0.00000161, 0.00000333, -0.00000083, 0.00000372, -0.00000605, 0.00000344, -0.00000209, 0.00000602, -0.00000188, 0.00000091, -0.00000368, 0.00000178, -0.00000467, 0.00000372, -0.00000307, 0.00000235, -0.00001368, 0.00000092, -0.00001132, 0.00000843, -0.00000225, 0.00000650, -0.00002184, -0.00002513, -0.00000544, 0.00004213, 0.00002607, 0.00001524, 0.00005259, 0.00000113, -0.00002485, -0.00001913, 0.00000711, -0.00001275, 0.00000275, -0.00000686, -0.00000313, -0.00000350, 0.00000877, -0.00000346, },
		{ 0.00000272, 0.00000141, -0.00000334, 0.00000035, -0.00000237, 0.00000880, -0.00001283, -0.00000351, 0.00000318, 0.00000647, -0.00000363, -0.00000314, -0.00000098, -0.00000581, 0.00000231, -0.00000282, 0.00000152, 0.00000212, -0.00000375, -0.00000063, -0.00000222, 0.00000206, 0.00000481, -0.00000716, -0.00000429, -0.00001132, -0.00000742, -0.00000801, 0.00000075, -0.00006192, -0.00007094, 0.00013638, 0.00010125, 0.00013827, -0.00007641, -0.00005639, -0.00000960, -0.00001478, -0.00000319, 0.00000848, -0.00000448, -0.00000354, -0.00000727, -0.00000886, -0.00000385, 0.00000124, 0.00000123, -0.00000196, -0.00000278, -0.00000137, 0.00000214, 0.00000179, -0.00000000, 0.00000047, -0.00000107, -0.00000634, 0.00000637, -0.00000500, 0.00000978, 0.00000890, -0.00000334, -0.00001134, 0.00000294, 0.00000728, -0.00001250, -0.00000401, -0.00000030, 0.00000575, -0.00000484, 0.00000240, 0.00001199, -0.00001570, 0.00000137, -0.00000886, 0.00001348, -0.00001497, -0.00001095, 0.00001459, -0.00000203, -0.00000312, -0.00000011, -0.00001294, 0.00001438, -0.00001474, 0.00000616, -0.00001144, 0.00000615, 0.00000978, -0.00000248, 0.00000297, 0.00000098, -0.00000175, -0.00000158, 0.00000156, 0.00000146, 0.00000321, -0.00000192, -0.00000095, -0.00000110, -0.00000154, -0.00000041, 0.00000065, 0.00000016, -0.00000769, -0.00000643, -0.00000458, -0.00000253, 0.00000737, -0.00000159, -0.00000398, -0.00003362, -0.00001305, 0.00002114, 0.00004880, 0.00000365, 0.00005349, 0.00001951, -0.00001895, -0.00003010, 0.00000683, -0.00001019, -0.00000099, -0.00000421, -0.00000397, -0.00000527, 0.00000660, 0.00000134, -0.00000492, },
		{ 0.00000401, -0.00000366, 0.00000071, -0.00000381, 0.00000702, -0.00000417, -0.00001065, 0.00000274, 0.00000535, 0.00000171, -0.00000628, 0.00000034, -0.00000514, -0.00000098, 0.00000105, -0.00000334, 0.00000577, -0.00000465, 0.00000044, -0.00000374, 0.00000195, 0.00000329, -0.00000046, -0.00000978, -0.00000314, -0.00001593, 0.00000019, -0.00001012, -0.00001829, -0.00010968, 0.00008938, 0.00010597, 0.00013340, 0.00001773, -0.00010305, -0.00000214, -0.00002403, -0.00000074, -0.00000015, 0.00000764, -0.00001183, 0.00000127, -0.00001532, -0.00000046, -0.00000507, 0.00000654, -0.00000552, 0.00000172, -0.00000649, 0.00000503, -0.00000153, 0.00000452, -0.00000404, 0.00000478, -0.00001015, 0.00000603, -0.00000307, 0.00000370, 0.00001157, 0.00000225, -0.00001099, -0.00000373, 0.00000909, -0.00000518, -0.00001092, 0.00000129, 0.00000169, 0.00000281, -0.00000770, 0.00001830, -0.00001268, -0.00000228, -0.00000740, 0.00000712, -0.00000183, -0.00001997, 0.00001034, 0.00000634, -0.00000839, 0.00000554, -0.00001503, 0.00000807, -0.00000329, -0.00000535, 0.00000067, -0.00000937, 0.00001854, -0.00000431, 0.00000450, -0.00000091, 0.00000209, -0.00000534, 0.00000383, -0.00000177, 0.00000602, -0.00000270, 0.00000084, -0.00000351, 0.00000132, -0.00000398, 0.00000347, -0.00000224, -0.00000133, -0.00001239, 0.00000018, -0.00000987, 0.00001097, -0.00000368, 0.00000352, -0.00002819, -0.00002287, 0.00000108, 0.00005496, 0.00001316, 0.00003722, 0.00004205, -0.00001062, -0.00003408, -0.00000317, -0.00000145, -0.00000815, 0.00000011, -0.00000628, -0.00000424, 0.00000150, 0.00000619, -0.00000520, -0.00000030, },
		{ -0.00000035, -0.00000223, -0.00000084, 0.00000057, 0.00000500, -0.00001366, -0.00000031, 0.00000385, 0.00000564, -0.00000571, -0.00000194, -0.00000192, -0.00000484, 0.00000340, -0.00000472, 0.00000450, -0.00000131, -0.00000162, -0.00000218, -0.00000045, 0.00000295, 0.00000335, -0.00001042, -0.00000124, -0.00001612, -0.00000075, -0.00001180, 0.00000018, -0.00009716, 0.00001032, 0.00012607, 0.00009780, 0.00010341, -0.00010175, -0.00002704, -0.00001642, -0.00000833, -0.00000268, 0.00001122, -0.00001047, 0.00000074, -0.00001295, -0.00000391, -0.00000542, 0.00000535, -0.00000338, 0.00000103, -0.00000574, 0.00000239, -0.00000030, 0.00000411, -0.00000287, 0.00000360, -0.00000580, -0.00000105, 0.00000313, -0.00000275, 0.00001138, 0.00000724, -0.00000738, -0.00000880, 0.00000562, 0.00000356, -0.00001396, -0.00000069, -0.00000026, 0.00000644, -0.00000981, 0.00001357, -0.00000085, -0.00001051, -0.00000202, -0.00000210, 0.00000858, -0.00002102, 0.00000064, 0.00001294, -0.00000774, 0.00000262, -0.00000635, -0.00000633, 0.00001010, -0.00001424, 0.00000761, -0.00001519, 0.00001476, 0.00000307, 0.00000063, 0.00000112, 0.00000192, -0.00000432, 0.00000159, -0.00000065, 0.00000408, 0.00000032, -0.00000020, -0.00000286, 0.00000073, -0.00000345, 0.00000193, -0.00000096, 0.00000036, -0.00001155, -0.00000231, -0.00000855, 0.00000475, 0.00000276, 0.00000056, -0.00001434, -0.00003021, -0.00001459, 0.00004258, 0.00003572, 0.00001812, 0.00005606, 0.00000479, -0.00003253, -0.00001779, 0.00000636, -0.00001251, 0.00000155, -0.00000689, -0.00000335, -0.00000303, 0.00000825, -0.00000297, -0.00000281, 0.00000339, },
		{ -0.00000414, 0.00000104, -0.00000320, 0.00000741, -0.00000879, -0.00000683, 0.00000351, 0.00000592, -0.00000100, -0.00000597, 0.00000077, -0.00000611, 0.00000197, -0.00000190, -0.00000033, 0.00000353, -0.00000400, -0.00000014, -0.00000286, 0.00000270, 0.00000440, -0.00000657, -0.00000465, -0.00000987, -0.00000820, -0.00000571, -0.00000146, -0.00005765, -0.00006187, 0.00012388, 0.00007880, 0.00013331, -0.00003814, -0.00007360, -0.00000356, -0.00001896, -0.00000170, 0.00000696, -0.00000120, -0.00000607, -0.00000521, -0.00001072, -0.00000254, 0.00000018, 0.00000191, -0.00000268, -0.00000145, -0.00000298, 0.00000295, 0.00000097, 0.00000145, -0.00000094, 0.00000102, -0.00000756, 0.00000618, -0.00000393, 0.00000653, 0.00001160, -0.00000245, -0.00001043, -0.00000069, 0.00000874, -0.00001035, -0.00000640, 0.00000018, 0.00000535, -0.00000497, 0.00000265, 0.00001183, -0.00001538, 0.00000009, -0.00000693, 0.00001074, -0.00001355, -0.00001069, 0.00001435, -0.00000124, -0.00000470, 0.00000388, -0.00001575, 0.00001259, -0.00001068, 0.00000244, -0.00000796, 0.00000042, 0.00001439, -0.00000386, 0.00000392, 0.00000001, -0.00000079, -0.00000260, 0.00000223, 0.00000032, 0.00000409, -0.00000178, -0.00000077, -0.00000183, -0.00000057, -0.00000181, 0.00000201, -0.00000093, -0.00000643, -0.00000830, -0.00000351, -0.00000430, 0.00000932, -0.00000273, -0.00000398, -0.00002855, -0.00002636, 0.00001919, 0.00005480, 0.00001210, 0.00005313, 0.00002743, -0.00002635, -0.00003073, 0.00000655, -0.00000979, -0.00000225, -0.00000428, -0.00000446, -0.00000468, 0.00000626, 0.00000129, -0.00000429, 0.00000093, 0.00000320, },
		{ -0.00000088, -0.00000180, 0.00000303, 0.00000025, -0.00001234, 0.00000200, 0.00000451, 0.00000409, -0.00000739, -0.00000025, -0.00000363, -0.00000225, 0.00000218, -0.00000438, 0.00000546, -0.00000327, -0.00000041, -0.00000291, 0.00000094, 0.00000458, -0.00000169, -0.00000829, -0.00000400, -0.00001386, -0.00000124, -0.00000711, -0.00002209, -0.00009481, 0.00007970, 0.00009213, 0.00010989, 0.00005165, -0.00010062, -0.00001029, -0.00001888, -0.00000570, 0.00000254, 0.00000685, -0.00000996, -0.00000071, -0.00001339, -0.00000257, -0.00000314, 0.00000471, -0.00000415, 0.00000095, -0.00000558, 0.00000335, -0.00000026, 0.00000374, -0.00000287, 0.00000367, -0.00000766, 0.00000255, 0.00000051, -0.00000039, 0.00001320, 0.00000335, -0.00000923, -0.00000621, 0.00000787, -0.00000201, -0.00001191, -0.00000003, 0.00000267, 0.00000141, -0.00000637, 0.00001713, -0.00001133, -0.00000432, -0.00000532, 0.00000535, -0.00000210, -0.00001855, 0.00000961, 0.00000718, -0.00000934, 0.00000697, -0.00001306, 0.00000224, 0.00000194, -0.00000825, 0.00000279, -0.00001228, 0.00001841, -0.00000148, 0.00000262, 0.00000042, 0.00000096, -0.00000436, 0.00000264, -0.00000101, 0.00000477, -0.00000088, -0.00000009, -0.00000305, 0.00000082, -0.00000351, 0.00000258, -0.00000099, -0.00000249, -0.00001131, -0.00000195, -0.00000816, 0.00000841, 0.00000046, -0.00000216, -0.00001909, -0.00003242, -0.00000653, 0.00005791, 0.00002403, 0.00003741, 0.00004933, -0.00001334, -0.00003864, -0.00000222, -0.00000179, -0.00000852, -0.00000076, -0.00000635, -0.00000422, 0.00000179, 0.00000562, -0.00000439, -0.00000131, 0.00000385, -0.00000218, },
		{ 0.00000093, -0.00000190, 0.00000585, -0.00001144, -0.00000308, 0.00000397, 0.00000601, -0.00000399, -0.00000461, 0.00000020, -0.00000567, 0.00000360, -0.00000431, 0.00000276, 0.00000067, -0.00000258, -0.00000118, -0.00000166, 0.00000436, 0.00000200, -0.00000881, -0.00000258, -0.00001331, -0.00000320, -0.00000746, -0.00000503, -0.00008561, 0.00000956, 0.00011395, 0.00007455, 0.00011277, -0.00007423, -0.00004468, -0.00000776, -0.00001488, 0.00000081, 0.00000872, -0.00000651, -0.00000277, -0.00000978, -0.00000688, -0.00000295, 0.00000319, -0.00000173, -0.00000037, -0.00000384, 0.00000001, 0.00000134, 0.00000268, -0.00000083, 0.00000152, -0.00000270, -0.00000395, 0.00000503, -0.00000389, 0.00000973, 0.00000967, -0.00000634, -0.00000886, 0.00000246, 0.00000578, -0.00001251, -0.00000340, 0.00000111, 0.00000493, -0.00000873, 0.00001280, -0.00000004, -0.00001165, -0.00000145, -0.00000174, 0.00000648, -0.00001904, 0.00000019, 0.00001316, -0.00000758, 0.00000194, -0.00000242, -0.00001108, 0.00001115, -0.00001234, 0.00000552, -0.00001338, 0.00000991, 0.00000835, -0.00000199, 0.00000297, 0.00000033, -0.00000285, -0.00000001, 0.00000059, 0.00000233, 0.00000208, -0.00000087, -0.00000218, -0.00000037, -0.00000214, 0.00000014, 0.00000102, -0.00000145, -0.00000960, -0.00000511, -0.00000625, 0.00000146, 0.00000688, -0.00000357, -0.00001005, -0.00002980, -0.00002789, 0.00004223, 0.00004554, 0.00002250, 0.00005974, 0.00000823, -0.00004021, -0.00001649, 0.00000534, -0.00001183, -0.00000016, -0.00000647, -0.00000399, -0.00000223, 0.00000753, -0.00000255, -0.00000269, 0.00000181, 0.00000185, -0.00000387, },
		{ -0.00000239, 0.00000441, -0.00000427, -0.00000960, 0.00000347, 0.00000513, 0.00000172, -0.00000810, 0.00000116, -0.00000517, 0.00000088, -0.00000058, -0.00000214, 0.00000445, -0.00000365, -0.00000022, -0.00000295, 0.00000285, 0.00000439, -0.00000642, -0.00000464, -0.00000873, -0.00000850, -0.00000430, -0.00000230, -0.00005506, -0.00005192, 0.00011029, 0.00006467, 0.00011833, -0.00000164, -0.00008268, -0.00000352, -0.00001911, -0.00000288, 0.00000714, 0.00000041, -0.00000683, -0.00000474, -0.00001100, -0.00000263, 0.00000035, 0.00000159, -0.00000258, -0.00000093, -0.00000355, 0.00000266, 0.00000107, 0.00000193, -0.00000119, 0.00000161, -0.00000703, 0.00000438, -0.00000145, 0.00000252, 0.00001406, -0.00000157, -0.00000918, -0.00000344, 0.00000831, -0.00000693, -0.00000877, 0.00000031, 0.00000508, -0.00000533, 0.00000328, 0.00001129, -0.00001496, -0.00000114, -0.00000487, 0.00000797, -0.00001220, -0.00001030, 0.00001403, -0.00000071, -0.00000558, 0.00000639, -0.00001603, 0.00000830, -0.00000534, -0.00000127, -0.00000487, -0.00000443, 0.00001713, -0.00000338, 0.00000346, 0.00000029, -0.00000095, -0.00000267, 0.00000205, 0.00000004, 0.00000401, -0.00000086, -0.00000107, -0.00000207, -0.00000023, -0.00000235, 0.00000238, -0.00000097, -0.00000628, -0.00000883, -0.00000380, -0.00000462, 0.00000918, -0.00000117, -0.00000713, -0.00002040, -0.00003911, 0.00001301, 0.00006197, 0.00002084, 0.00005433, 0.00003458, -0.00003352, -0.00003146, 0.00000604, -0.00000913, -0.00000373, -0.00000425, -0.00000500, -0.00000414, 0.00000604, 0.00000105, -0.00000352, -0.00000059, 0.00000408, -0.00000380, 0.00000051, },
		{ -0.00000040, 0.00000297, -0.00001207, 0.00000013, 0.00000430, 0.00000539, -0.00000683, -0.00000241, -0.00000135, -0.00000357, 0.00000308, -0.00000492, 0.00000457, -0.00000157, -0.00000135, -0.00000213, 0.00000010, 0.00000564, -0.00000272, -0.00000704, -0.00000466, -0.00001174, -0.00000304, -0.00000363, -0.00002611, -0.00008050, 0.00006972, 0.00008277, 0.00008671, 0.00007373, -0.00008704, -0.00002289, -0.00001198, -0.00001135, 0.00000557, 0.00000532, -0.00000724, -0.00000330, -0.00001089, -0.00000504, -0.00000099, 0.00000276, -0.00000267, -0.00000009, -0.00000419, 0.00000128, 0.00000122, 0.00000263, -0.00000125, 0.00000204, -0.00000483, -0.00000075, 0.00000346, -0.00000318, 0.00001311, 0.00000547, -0.00000837, -0.00000684, 0.00000499, 0.00000129, -0.00001196, -0.00000198, 0.00000384, -0.00000022, -0.00000485, 0.00001592, -0.00001024, -0.00000613, -0.00000324, 0.00000368, -0.00000244, -0.00001718, 0.00000915, 0.00000738, -0.00000938, 0.00000717, -0.00000990, -0.00000352, 0.00000548, -0.00000873, 0.00000277, -0.00001296, 0.00001584, 0.00000293, -0.00000007, 0.00000236, -0.00000059, -0.00000310, 0.00000120, 0.00000004, 0.00000320, 0.00000099, -0.00000095, -0.00000244, -0.00000006, -0.00000253, 0.00000122, 0.00000069, -0.00000406, -0.00000975, -0.00000439, -0.00000610, 0.00000522, 0.00000502, -0.00000757, -0.00001172, -0.00003686, -0.00001954, 0.00006110, 0.00003507, 0.00003938, 0.00005636, -0.00001626, -0.00004309, -0.00000145, -0.00000219, -0.00000867, -0.00000197, -0.00000609, -0.00000456, 0.00000240, 0.00000484, -0.00000362, -0.00000185, 0.00000295, -0.00000020, -0.00000270, 0.00000036, },
		{ 0.00000452, -0.00000778, -0.00000616, 0.00000427, 0.00000560, -0.00000141, -0.00000742, 0.00000157, -0.00000549, 0.00000304, -0.00000331, 0.00000075, 0.00000242, -0.00000306, -0.00000071, -0.00000227, 0.00000518, 0.00000116, -0.00000775, -0.00000345, -0.00001085, -0.00000530, -0.00000381, -0.00000901, -0.00007585, 0.00000970, 0.00010227, 0.00005780, 0.00010983, -0.00004262, -0.00005887, -0.00000282, -0.00001849, 0.00000221, 0.00000742, -0.00000365, -0.00000507, -0.00000774, -0.00000867, -0.00000161, 0.00000205, -0.00000091, -0.00000115, -0.00000249, -0.00000163, 0.00000211, 0.00000195, 0.00000050, 0.00000028, -0.00000082, -0.00000521, 0.00000513, -0.00000310, 0.00000650, 0.00001261, -0.00000584, -0.00000793, -0.00000064, 0.00000654, -0.00000958, -0.00000651, 0.00000238, 0.00000345, -0.00000779, 0.00001234, 0.00000035, -0.00001258, -0.00000084, -0.00000131, 0.00000441, -0.00001722, -0.00000000, 0.00001301, -0.00000719, 0.00000128, 0.00000075, -0.00001364, 0.00000932, -0.00000816, 0.00000224, -0.00001068, 0.00000457, 0.00001294, -0.00000350, 0.00000386, -0.00000034, -0.00000226, -0.00000088, 0.00000118, 0.00000122, 0.00000301, -0.00000083, -0.00000191, -0.00000117, -0.00000121, -0.00000108, 0.00000229, -0.00000248, -0.00000848, -0.00000683, -0.00000506, -0.00000068, 0.00000918, -0.00000499, -0.00000956, -0.00002446, -0.00004302, 0.00003864, 0.00005662, 0.00002761, 0.00006461, 0.00001074, -0.00004753, -0.00001534, 0.00000412, -0.00001085, -0.00000222, -0.00000579, -0.00000485, -0.00000129, 0.00000681, -0.00000228, -0.00000230, 0.00000018, 0.00000351, -0.00000461, 0.00000157, -0.00000261, },
		{ -0.00000043, -0.00001100, 0.00000256, 0.00000465, 0.00000381, -0.00000881, -0.00000014, -0.00000305, -0.00000070, 0.00000087, -0.00000368, 0.00000479, -0.00000278, -0.00000067, -0.00000276, 0.00000282, 0.00000447, -0.00000639, -0.00000455, -0.00000767, -0.00000853, -0.00000359, -0.00000212, -0.00005350, -0.00004190, 0.00009655, 0.00005618, 0.00009885, 0.00002807, -0.00008185, -0.00000920, -0.00001599, -0.00000597, 0.00000851, 0.00000070, -0.00000619, -0.00000545, -0.00001014, -0.00000368, 0.00000128, 0.00000067, -0.00000196, -0.00000102, -0.00000330, 0.00000156, 0.00000179, 0.00000169, -0.00000057, 0.00000112, -0.00000540, 0.00000182, 0.00000144, -0.00000116, 0.00001536, 0.00000006, -0.00000849, -0.00000453, 0.00000586, -0.00000265, -0.00001068, -0.00000011, 0.00000507, -0.00000597, 0.00000420, 0.00001053, -0.00001466, -0.00000212, -0.00000287, 0.00000536, -0.00001099, -0.00000986, 0.00001381, -0.00000064, -0.00000561, 0.00000741, -0.00001427, 0.00000270, -0.00000024, -0.00000364, -0.00000313, -0.00000759, 0.00001748, -0.00000091, 0.00000166, 0.00000167, -0.00000196, -0.00000208, 0.00000127, 0.00000037, 0.00000322, 0.00000050, -0.00000160, -0.00000197, -0.00000047, -0.00000214, 0.00000194, -0.00000023, -0.00000693, -0.00000839, -0.00000495, -0.00000400, 0.00000757, 0.00000206, -0.00001173, -0.00001187, -0.00004806, 0.00000081, 0.00007037, 0.00002975, 0.00005750, 0.00004090, -0.00004066, -0.00003209, 0.00000524, -0.00000832, -0.00000523, -0.00000434, -0.00000539, -0.00000380, 0.00000604, 0.00000062, -0.00000273, -0.00000170, 0.00000406, -0.00000253, -0.00000086, -0.00000060, 0.00000088, },
		{ -0.00000992, -0.00000261, 0.00000462, 0.00000563, -0.00000483, -0.00000550, 0.00000091, -0.00000432, 0.00000337, -0.00000481, 0.00000315, 0.00000023, -0.00000209, -0.00000167, -0.00000032, 0.00000622, -0.00000335, -0.00000622, -0.00000493, -0.00000982, -0.00000477, -0.00000040, -0.00002940, -0.00006779, 0.00006073, 0.00007532, 0.00006691, 0.00008390, -0.00006565, -0.00003623, -0.00000611, -0.00001569, 0.00000755, 0.00000420, -0.00000480, -0.00000545, -0.00000880, -0.00000700, 0.00000058, 0.00000137, -0.00000162, -0.00000090, -0.00000290, -0.00000054, 0.00000228, 0.00000182, 0.00000013, 0.00000069, -0.00000259, -0.00000297, 0.00000498, -0.00000406, 0.00001103, 0.00000864, -0.00000843, -0.00000592, 0.00000131, 0.00000383, -0.00001060, -0.00000463, 0.00000512, -0.00000194, -0.00000333, 0.00001488, -0.00000957, -0.00000762, -0.00000121, 0.00000213, -0.00000275, -0.00001599, 0.00000902, 0.00000703, -0.00000878, 0.00000667, -0.00000646, -0.00000807, 0.00000655, -0.00000668, 0.00000098, -0.00001196, 0.00001166, 0.00000787, -0.00000259, 0.00000403, -0.00000173, -0.00000227, 0.00000011, 0.00000081, 0.00000190, 0.00000237, -0.00000130, -0.00000206, -0.00000088, -0.00000160, 0.00000001, 0.00000207, -0.00000526, -0.00000855, -0.00000624, -0.00000463, 0.00000251, 0.00000856, -0.00001097, -0.00000791, -0.00003522, -0.00003681, 0.00006250, 0.00004724, 0.00004258, 0.00006394, -0.00002008, -0.00004700, -0.00000102, -0.00000260, -0.00000863, -0.00000345, -0.00000557, -0.00000514, 0.00000317, 0.00000401, -0.00000300, -0.00000198, 0.00000158, 0.00000179, -0.00000416, 0.00000192, -0.00000251, 0.00000357, },
		{ -0.00000868, 0.00000417, 0.00000502, 0.00000122, -0.00000940, 0.00000148, -0.00000402, 0.00000169, -0.00000184, -0.00000130, 0.00000373, -0.00000299, -0.00000076, -0.00000241, 0.00000560, 0.00000064, -0.00000704, -0.00000400, -0.00000865, -0.00000700, -0.00000101, -0.00001165, -0.00006782, 0.00001064, 0.00009083, 0.00004655, 0.00009891, -0.00001241, -0.00006619, -0.00000306, -0.00001856, 0.00000128, 0.00000758, -0.00000220, -0.00000587, -0.00000700, -0.00000917, -0.00000144, 0.00000191, -0.00000089, -0.00000127, -0.00000182, -0.00000235, 0.00000191, 0.00000201, 0.00000100, 0.00000004, -0.00000023, -0.00000496, 0.00000381, -0.00000098, 0.00000255, 0.00001523, -0.00000512, -0.00000700, -0.00000260, 0.00000527, -0.00000523, -0.00000970, 0.00000328, 0.00000219, -0.00000713, 0.00001221, 0.00000038, -0.00001345, -0.00000005, -0.00000095, 0.00000256, -0.00001567, 0.00000004, 0.00001269, -0.00000687, 0.00000103, 0.00000274, -0.00001395, 0.00000525, -0.00000294, -0.00000089, -0.00000827, -0.00000009, 0.00001584, -0.00000320, 0.00000333, 0.00000025, -0.00000269, -0.00000093, 0.00000105, 0.00000085, 0.00000307, -0.00000009, -0.00000205, -0.00000153, -0.00000088, -0.00000149, 0.00000264, -0.00000256, -0.00000834, -0.00000735, -0.00000501, -0.00000159, 0.00000954, -0.00000383, -0.00001210, -0.00001628, -0.00005661, 0.00002923, 0.00006968, 0.00003299, 0.00007139, 0.00001186, -0.00005434, -0.00001435, 0.00000273, -0.00000968, -0.00000440, -0.00000508, -0.00000571, -0.00000046, 0.00000624, -0.00000222, -0.00000175, -0.00000127, 0.00000439, -0.00000430, 0.00000108, -0.00000181, 0.00000177, -0.00000374, },
		{ 0.00000062, 0.00000473, 0.00000488, -0.00000786, -0.00000295, -0.00000029, -0.00000223, 0.00000203, -0.00000467, 0.00000445, -0.00000150, -0.00000126, -0.00000259, 0.00000292, 0.00000435, -0.00000621, -0.00000465, -0.00000645, -0.00000849, -0.00000324, -0.00000153, -0.00005206, -0.00003284, 0.00008383, 0.00005070, 0.00007915, 0.00004839, -0.00007186, -0.00001853, -0.00001154, -0.00000942, 0.00000993, 0.00000055, -0.00000502, -0.00000648, -0.00000896, -0.00000491, 0.00000226, -0.00000022, -0.00000133, -0.00000129, -0.00000266, 0.00000018, 0.00000259, 0.00000128, 0.00000033, 0.00000028, -0.00000360, -0.00000050, 0.00000370, -0.00000349, 0.00001478, 0.00000293, -0.00000889, -0.00000376, 0.00000184, 0.00000170, -0.00001153, -0.00000133, 0.00000537, -0.00000685, 0.00000526, 0.00000979, -0.00001468, -0.00000271, -0.00000104, 0.00000298, -0.00000990, -0.00000949, 0.00001382, -0.00000109, -0.00000488, 0.00000726, -0.00001127, -0.00000292, 0.00000334, -0.00000390, -0.00000308, -0.00000892, 0.00001564, 0.00000300, -0.00000085, 0.00000350, -0.00000316, -0.00000145, 0.00000044, 0.00000081, 0.00000227, 0.00000179, -0.00000193, -0.00000185, -0.00000092, -0.00000161, 0.00000123, 0.00000071, -0.00000771, -0.00000776, -0.00000611, -0.00000331, 0.00000557, 0.00000544, -0.00001570, -0.00000566, -0.00005080, -0.00001773, 0.00007872, 0.00003949, 0.00006245, 0.00004685, -0.00004828, -0.00003224, 0.00000403, -0.00000738, -0.00000668, -0.00000457, -0.00000562, -0.00000367, 0.00000621, 0.00000011, -0.00000206, -0.00000232, 0.00000327, -0.00000068, -0.00000253, 0.00000135, -0.00000213, 0.00000185, -0.00001061, },
		{ 0.00000500, 0.00000521, -0.00000208, -0.00000856, 0.00000221, -0.00000393, 0.00000279, -0.00000397, 0.00000133, 0.00000192, -0.00000244, -0.00000166, -0.00000027, 0.00000635, -0.00000360, -0.00000582, -0.00000484, -0.00000817, -0.00000622, 0.00000219, -0.00003145, -0.00005712, 0.00005321, 0.00006858, 0.00005140, 0.00008441, -0.00004102, -0.00004649, -0.00000360, -0.00001728, 0.00000754, 0.00000425, -0.00000343, -0.00000643, -0.00000774, -0.00000789, 0.00000113, 0.00000089, -0.00000130, -0.00000117, -0.00000209, -0.00000165, 0.00000249, 0.00000168, 0.00000084, 0.00000013, -0.00000147, -0.00000371, 0.00000486, -0.00000308, 0.00000736, 0.00001234, -0.00000891, -0.00000431, -0.00000203, 0.00000470, -0.00000755, -0.00000784, 0.00000637, -0.00000357, -0.00000203, 0.00001417, -0.00000934, -0.00000883, 0.00000083, 0.00000059, -0.00000289, -0.00001513, 0.00000925, 0.00000626, -0.00000785, 0.00000596, -0.00000348, -0.00001077, 0.00000509, -0.00000277, -0.00000159, -0.00001027, 0.00000701, 0.00001214, -0.00000395, 0.00000462, -0.00000178, -0.00000241, -0.00000020, 0.00000091, 0.00000124, 0.00000290, -0.00000090, -0.00000214, -0.00000132, -0.00000114, -0.00000052, 0.00000264, -0.00000557, -0.00000827, -0.00000693, -0.00000429, 0.00000095, 0.00001020, -0.00001150, -0.00000817, -0.00002831, -0.00005554, 0.00005923, 0.00006173, 0.00004641, 0.00007292, -0.00002550, -0.00005001, -0.00000106, -0.00000295, -0.00000853, -0.00000501, -0.00000502, -0.00000577, 0.00000389, 0.00000332, -0.00000264, -0.00000175, 0.00000002, 0.00000342, -0.00000486, 0.00000248, -0.00000304, 0.00000224, -0.00000655, -0.00000555, },
		},
		{ { -0.00000364, -0.00000285, 0.00000683, -0.00000101, 0.00000351, 0.00000039, -0.00000384, 0.00000383, -0.00000527, -0.00000137, -0.00000372, -0.00000220, 0.00000955, 0.00000015, -0.00000516, 0.00000035, 0.00000033, -0.00000094, -0.00000594, -0.00000413, -0.00000274, -0.00000587, -0.00001431, -0.00001027, -0.00005711, 0.00001352, 0.00008514, 0.00003715, 0.00008323, 0.00001059, -0.00006457, -0.00001129, -0.00000844, -0.00000271, 0.00000196, -0.00000119, -0.00000500, -0.00000642, 0.00001202, -0.00000477, -0.00000207, -0.00000388, -0.00000269, 0.00000116, -0.00001453, 0.00000177, 0.00000897, -0.00000480, -0.00000367, -0.00000044, -0.00001524, 0.00000170, -0.00000166, 0.00000123, -0.00000245, -0.00000155, 0.00001263, -0.00000288, 0.00000428, 0.00000128, -0.00000695, -0.00000272, 0.00000008, 0.00000088, 0.00000270, 0.00000088, 0.00000008, -0.00000272, -0.00000695, 0.00000128, 0.00000428, -0.00000288, 0.00001263, -0.00000155, -0.00000245, 0.00000123, -0.00000166, 0.00000170, -0.00001524, -0.00000044, -0.00000367, -0.00000480, 0.00000897, 0.00000177, -0.00001453, 0.00000116, -0.00000269, -0.00000388, -0.00000207, -0.00000477, 0.00001202, -0.00000642, -0.00000500, -0.00000119, 0.00000196, -0.00000271, -0.00000844, -0.00001129, -0.00006457, 0.00001059, 0.00008323, 0.00003715, 0.00008514, 0.00001352, -0.00005711, -0.00001027, -0.00001431, -0.00000587, -0.00000274, -0.00000413, -0.00000594, -0.00000094, 0.00000033, 0.00000035, -0.00000516, 0.00000015, 0.00000955, -0.00000220, -0.00000372, -0.00000137, -0.00000527, 0.00000383, -0.00000384, 0.00000039, 0.00000351, -0.00000101, 0.00000683, -0.00000285, },
		{ -0.00000414, 0.00000391, 0.00000348, -0.00000134, 0.00000517, -0.00000533, 0.00000268, -0.00000179, -0.00000389, -0.00000100, -0.00000551, 0.00000552, 0.00000674, -0.00000550, -0.00000125, 0.00000013, 0.00000030, -0.00000442, -0.00000504, -0.00000358, -0.00000221, -0.00001317, -0.00000780, -0.00003857, -0.00003247, 0.00007628, 0.00004889, 0.00005754, 0.00006214, -0.00005007, -0.00003608, -0.00000280, -0.00000916, 0.00000343, -0.00000163, -0.00000084, -0.00001005, 0.00000618, 0.00000527, -0.00000733, 0.00000090, -0.00000795, 0.00000422, -0.00000978, -0.00000803, 0.00000946, 0.00000160, -0.00000748, 0.00000207, -0.00001207, -0.00000608, 0.00000303, -0.00000331, 0.00000260, -0.00000687, 0.00001135, 0.00000269, -0.00000070, 0.00000581, -0.00000527, -0.00000535, -0.00000058, 0.00000056, 0.00000214, 0.00000195, -0.00000005, -0.00000079, -0.00000556, -0.00000402, 0.00000604, -0.00000255, 0.00000702, 0.00000733, -0.00000717, 0.00000369, -0.00000331, 0.00000389, -0.00001095, -0.00000780, 0.00000125, -0.00000794, 0.00000397, 0.00000865, -0.00001021, -0.00000797, 0.00000315, -0.00000672, 0.00000039, -0.00000778, 0.00000694, 0.00000392, -0.00000973, -0.00000068, -0.00000096, 0.00000274, -0.00001022, -0.00000330, -0.00004415, -0.00004669, 0.00007756, 0.00005416, 0.00006535, 0.00006586, -0.00005425, -0.00002688, -0.00000916, -0.00001261, -0.00000181, -0.00000415, -0.00000517, -0.00000400, 0.00000106, 0.00000031, -0.00000211, -0.00000522, 0.00000794, 0.00000408, -0.00000544, -0.00000077, -0.00000529, 0.00000037, 0.00000128, -0.00000505, 0.00000584, -0.00000148, 0.00000538, 0.00000100, -0.00000403, },
		{ -0.00000084, 0.00000734, -0.00000303, 0.00000461, -0.00000150, -0.00000171, 0.00000221, -0.00000522, -0.00000039, -0.00000459, -0.00000091, 0.00000967, -0.00000123, -0.00000420, 0.00000042, -0.00000003, -0.00000193, -0.00000561, -0.00000419, -0.00000174, -0.00000808, -0.00001091, -0.00001964, -0.00005120, 0.00003768, 0.00007148, 0.00003473, 0.00007793, -0.00000313, -0.00005719, -0.00000826, -0.00000801, -0.00000156, 0.00000168, -0.00000155, -0.00000546, -0.00000515, 0.00001277, -0.00000672, 0.00000008, -0.00000589, -0.00000143, -0.00000009, -0.00001421, 0.00000398, 0.00000757, -0.00000639, -0.00000123, -0.00000438, -0.00001375, 0.00000467, -0.00000411, 0.00000265, -0.00000519, 0.00000362, 0.00001024, -0.00000391, 0.00000616, -0.00000062, -0.00000696, -0.00000262, 0.00000067, 0.00000126, 0.00000263, 0.00000043, -0.00000040, -0.00000276, -0.00000676, 0.00000300, 0.00000181, -0.00000006, 0.00001263, -0.00000556, 0.00000028, 0.00000009, 0.00000067, -0.00000272, -0.00001427, 0.00000212, -0.00000607, -0.00000248, 0.00000967, -0.00000058, -0.00001433, 0.00000192, -0.00000333, -0.00000236, -0.00000416, -0.00000258, 0.00001077, -0.00000725, -0.00000469, -0.00000058, 0.00000187, -0.00000385, -0.00000880, -0.00001546, -0.00007113, 0.00002718, 0.00008575, 0.00004423, 0.00009402, -0.00001682, -0.00005319, -0.00000482, -0.00001697, -0.00000371, -0.00000394, -0.00000387, -0.00000620, 0.00000010, 0.00000061, 0.00000006, -0.00000592, 0.00000162, 0.00000914, -0.00000330, -0.00000284, -0.00000277, -0.00000432, 0.00000450, -0.00000556, 0.00000232, 0.00000246, 0.00000125, 0.00000516, -0.00000402, -0.00000372, },
		{ 0.00000667, 0.00000064, -0.00000032, 0.00000382, -0.00000461, 0.00000305, -0.00000343, -0.00000228, -0.00000164, -0.00000531, 0.00000691, 0.00000525, -0.00000533, -0.00000071, 0.00000004, -0.00000077, -0.00000437, -0.00000524, -0.00000232, -0.00000392, -0.00001142, -0.00001091, -0.00004472, -0.00000824, 0.00007688, 0.00003664, 0.00006015, 0.00004676, -0.00005064, -0.00002909, -0.00000270, -0.00000784, 0.00000367, -0.00000213, -0.00000111, -0.00001011, 0.00000841, 0.00000325, -0.00000633, 0.00000072, -0.00000875, 0.00000520, -0.00001164, -0.00000535, 0.00000943, -0.00000063, -0.00000618, 0.00000135, -0.00001514, -0.00000043, 0.00000019, -0.00000178, 0.00000028, -0.00000448, 0.00001332, -0.00000129, 0.00000197, 0.00000462, -0.00000609, -0.00000523, -0.00000015, 0.00000115, 0.00000219, 0.00000169, -0.00000068, -0.00000078, -0.00000585, -0.00000240, 0.00000515, -0.00000285, 0.00001049, 0.00000239, -0.00000594, 0.00000396, -0.00000221, 0.00000257, -0.00001408, -0.00000333, -0.00000077, -0.00000732, 0.00000606, 0.00000723, -0.00001180, -0.00000641, 0.00000226, -0.00000532, -0.00000067, -0.00000764, 0.00000813, 0.00000180, -0.00000930, -0.00000049, -0.00000029, 0.00000168, -0.00001095, -0.00000449, -0.00005293, -0.00004020, 0.00009223, 0.00005137, 0.00008457, 0.00004353, -0.00006809, -0.00001392, -0.00001336, -0.00001016, -0.00000272, -0.00000390, -0.00000569, -0.00000314, 0.00000153, 0.00000049, -0.00000316, -0.00000455, 0.00000883, 0.00000262, -0.00000514, -0.00000105, -0.00000607, 0.00000249, -0.00000075, -0.00000396, 0.00000594, -0.00000083, 0.00000614, -0.00000162, -0.00000372, -0.00000378, },
		{ 0.00000638, -0.00000425, 0.00000534, -0.00000292, 0.00000030, 0.00000006, -0.00000427, 0.00000004, -0.00000539, 0.00000061, 0.00000936, -0.00000229, -0.00000333, 0.00000048, -0.00000071, -0.00000251, -0.00000556, -0.00000352, -0.00000187, -0.00000883, -0.00000928, -0.00002946, -0.00003738, 0.00005326, 0.00005693, 0.00003517, 0.00007048, -0.00001404, -0.00004930, -0.00000654, -0.00000703, -0.00000090, 0.00000149, -0.00000200, -0.00000579, -0.00000370, 0.00001314, -0.00000844, 0.00000223, -0.00000835, 0.00000056, -0.00000207, -0.00001300, 0.00000556, 0.00000589, -0.00000724, 0.00000078, -0.00000918, -0.00000967, 0.00000514, -0.00000541, 0.00000338, -0.00000699, 0.00000888, 0.00000609, -0.00000300, 0.00000694, -0.00000232, -0.00000696, -0.00000238, 0.00000134, 0.00000153, 0.00000246, -0.00000008, -0.00000069, -0.00000284, -0.00000624, 0.00000409, -0.00000045, 0.00000381, 0.00001034, -0.00000788, 0.00000250, -0.00000023, 0.00000181, -0.00000732, -0.00001147, 0.00000306, -0.00000795, 0.00000037, 0.00000944, -0.00000263, -0.00001399, 0.00000254, -0.00000359, -0.00000127, -0.00000618, -0.00000012, 0.00000894, -0.00000748, -0.00000478, 0.00000057, 0.00000101, -0.00000444, -0.00000972, -0.00002031, -0.00007682, 0.00004670, 0.00008510, 0.00005762, 0.00009317, -0.00004802, -0.00004070, -0.00000485, -0.00001717, -0.00000287, -0.00000444, -0.00000398, -0.00000601, 0.00000088, 0.00000100, -0.00000062, -0.00000630, 0.00000299, 0.00000857, -0.00000424, -0.00000210, -0.00000432, -0.00000255, 0.00000408, -0.00000652, 0.00000398, 0.00000174, 0.00000327, 0.00000278, -0.00000443, -0.00000392, 0.00000195, },
		{ -0.00000261, 0.00000142, 0.00000191, -0.00000299, 0.00000221, -0.00000412, -0.00000084, -0.00000253, -0.00000479, 0.00000812, 0.00000362, -0.00000482, -0.00000041, -0.00000012, -0.00000189, -0.00000413, -0.00000532, -0.00000109, -0.00000600, -0.00000847, -0.00001802, -0.00004313, 0.00001342, 0.00007115, 0.00002834, 0.00006153, 0.00003179, -0.00004847, -0.00002382, -0.00000225, -0.00000681, 0.00000387, -0.00000273, -0.00000123, -0.00001008, 0.00001070, 0.00000078, -0.00000463, -0.00000040, -0.00000868, 0.00000560, -0.00001307, -0.00000262, 0.00000877, -0.00000256, -0.00000426, -0.00000115, -0.00001597, 0.00000452, -0.00000362, 0.00000090, -0.00000310, -0.00000001, 0.00001261, -0.00000378, 0.00000454, 0.00000289, -0.00000659, -0.00000512, 0.00000045, 0.00000169, 0.00000206, 0.00000143, -0.00000131, -0.00000058, -0.00000617, -0.00000063, 0.00000337, -0.00000136, 0.00001216, -0.00000236, -0.00000392, 0.00000410, -0.00000093, -0.00000035, -0.00001514, 0.00000057, -0.00000357, -0.00000545, 0.00000732, 0.00000569, -0.00001311, -0.00000490, 0.00000147, -0.00000378, -0.00000236, -0.00000670, 0.00000854, 0.00000023, -0.00000921, 0.00000023, -0.00000023, 0.00000090, -0.00001196, -0.00000601, -0.00006262, -0.00002976, 0.00010472, 0.00005169, 0.00010271, 0.00001025, -0.00007016, -0.00000427, -0.00001763, -0.00000731, -0.00000403, -0.00000335, -0.00000623, -0.00000216, 0.00000195, 0.00000045, -0.00000414, -0.00000372, 0.00000956, 0.00000109, -0.00000457, -0.00000189, -0.00000597, 0.00000404, -0.00000289, -0.00000233, 0.00000569, 0.00000035, 0.00000576, -0.00000364, -0.00000343, -0.00000267, 0.00000867, },
		{ -0.00000424, 0.00000528, -0.00000345, 0.00000159, -0.00000202, -0.00000273, -0.00000008, -0.00000602, 0.00000234, 0.00000854, -0.00000286, -0.00000272, 0.00000066, -0.00000182, -0.00000257, -0.00000588, -0.00000205, -0.00000342, -0.00000733, -0.00001138, -0.00003580, -0.00001968, 0.00006071, 0.00004358, 0.00003730, 0.00006114, -0.00002179, -0.00004170, -0.00000567, -0.00000564, -0.00000075, 0.00000147, -0.00000258, -0.00000592, -0.00000209, 0.00001309, -0.00000978, 0.00000412, -0.00001092, 0.00000302, -0.00000465, -0.00001089, 0.00000625, 0.00000427, -0.00000738, 0.00000184, -0.00001388, -0.00000360, 0.00000276, -0.00000460, 0.00000248, -0.00000678, 0.00001282, 0.00000137, -0.00000063, 0.00000653, -0.00000359, -0.00000701, -0.00000195, 0.00000207, 0.00000159, 0.00000228, -0.00000069, -0.00000071, -0.00000306, -0.00000533, 0.00000426, -0.00000171, 0.00000766, 0.00000643, -0.00000861, 0.00000429, 0.00000015, 0.00000133, -0.00001108, -0.00000758, 0.00000238, -0.00000878, 0.00000319, 0.00000849, -0.00000428, -0.00001367, 0.00000317, -0.00000358, -0.00000068, -0.00000793, 0.00000237, 0.00000677, -0.00000726, -0.00000511, 0.00000206, -0.00000056, -0.00000441, -0.00001147, -0.00002566, -0.00008118, 0.00006862, 0.00008190, 0.00007746, 0.00007825, -0.00007313, -0.00002388, -0.00000939, -0.00001477, -0.00000367, -0.00000390, -0.00000466, -0.00000527, 0.00000133, 0.00000149, -0.00000161, -0.00000632, 0.00000424, 0.00000783, -0.00000496, -0.00000165, -0.00000564, -0.00000037, 0.00000273, -0.00000662, 0.00000524, 0.00000147, 0.00000470, 0.00000010, -0.00000426, -0.00000402, 0.00000523, 0.00000386, },
		{ 0.00000341, -0.00000018, -0.00000087, 0.00000036, -0.00000372, 0.00000013, -0.00000355, -0.00000390, 0.00000894, 0.00000211, -0.00000422, -0.00000014, -0.00000060, -0.00000272, -0.00000407, -0.00000473, -0.00000085, -0.00000696, -0.00000649, -0.00002658, -0.00003438, 0.00002954, 0.00006228, 0.00002325, 0.00006149, 0.00001765, -0.00004389, -0.00002064, -0.00000083, -0.00000666, 0.00000453, -0.00000381, -0.00000089, -0.00001022, 0.00001320, -0.00000224, -0.00000215, -0.00000262, -0.00000740, 0.00000497, -0.00001352, -0.00000044, 0.00000790, -0.00000420, -0.00000198, -0.00000534, -0.00001378, 0.00000719, -0.00000682, 0.00000363, -0.00000661, 0.00000576, 0.00000933, -0.00000415, 0.00000622, 0.00000113, -0.00000697, -0.00000492, 0.00000119, 0.00000212, 0.00000179, 0.00000116, -0.00000188, -0.00000028, -0.00000639, 0.00000091, 0.00000126, 0.00000158, 0.00001167, -0.00000616, -0.00000164, 0.00000458, -0.00000038, -0.00000394, -0.00001430, 0.00000338, -0.00000659, -0.00000244, 0.00000739, 0.00000456, -0.00001456, -0.00000311, 0.00000056, -0.00000201, -0.00000476, -0.00000476, 0.00000785, -0.00000036, -0.00000993, 0.00000202, -0.00000148, 0.00000123, -0.00001417, -0.00000709, -0.00007375, -0.00001430, 0.00011340, 0.00005824, 0.00011377, -0.00002924, -0.00005995, -0.00000101, -0.00001941, -0.00000577, -0.00000459, -0.00000325, -0.00000627, -0.00000150, 0.00000262, -0.00000008, -0.00000477, -0.00000294, 0.00001027, -0.00000060, -0.00000379, -0.00000319, -0.00000491, 0.00000466, -0.00000467, -0.00000051, 0.00000530, 0.00000173, 0.00000441, -0.00000491, -0.00000331, -0.00000062, 0.00000932, -0.00000556, },
		{ 0.00000421, -0.00000282, 0.00000167, -0.00000340, -0.00000105, -0.00000065, -0.00000640, 0.00000415, 0.00000730, -0.00000297, -0.00000238, 0.00000088, -0.00000320, -0.00000225, -0.00000624, -0.00000036, -0.00000573, -0.00000412, -0.00001756, -0.00003594, -0.00000252, 0.00006224, 0.00003198, 0.00004057, 0.00004991, -0.00002584, -0.00003556, -0.00000461, -0.00000456, -0.00000056, 0.00000122, -0.00000290, -0.00000616, -0.00000009, 0.00001230, -0.00001037, 0.00000524, -0.00001298, 0.00000537, -0.00000731, -0.00000832, 0.00000625, 0.00000276, -0.00000675, 0.00000134, -0.00001718, 0.00000306, -0.00000182, -0.00000157, -0.00000043, -0.00000394, 0.00001422, -0.00000250, 0.00000226, 0.00000526, -0.00000448, -0.00000710, -0.00000134, 0.00000283, 0.00000138, 0.00000219, -0.00000143, -0.00000042, -0.00000344, -0.00000410, 0.00000347, -0.00000147, 0.00001045, 0.00000186, -0.00000817, 0.00000601, 0.00000059, -0.00000053, -0.00001345, -0.00000331, 0.00000021, -0.00000807, 0.00000520, 0.00000744, -0.00000590, -0.00001311, 0.00000365, -0.00000314, -0.00000090, -0.00000893, 0.00000435, 0.00000486, -0.00000724, -0.00000500, 0.00000313, -0.00000205, -0.00000450, -0.00001344, -0.00003215, -0.00008294, 0.00009134, 0.00007846, 0.00010112, 0.00004769, -0.00008585, -0.00000844, -0.00001550, -0.00001120, -0.00000531, -0.00000279, -0.00000557, -0.00000431, 0.00000173, 0.00000181, -0.00000264, -0.00000619, 0.00000554, 0.00000679, -0.00000534, -0.00000169, -0.00000635, 0.00000168, 0.00000085, -0.00000595, 0.00000609, 0.00000161, 0.00000534, -0.00000248, -0.00000376, -0.00000375, 0.00000845, 0.00000006, -0.00000282, },
		{ -0.00000187, 0.00000112, -0.00000198, -0.00000237, 0.00000049, -0.00000458, -0.00000260, 0.00000921, 0.00000096, -0.00000378, 0.00000029, -0.00000159, -0.00000301, -0.00000440, -0.00000322, -0.00000218, -0.00000565, -0.00000772, -0.00003302, -0.00002139, 0.00003970, 0.00005221, 0.00002092, 0.00005953, 0.00000524, -0.00003787, -0.00001918, 0.00000153, -0.00000749, 0.00000578, -0.00000544, 0.00000001, -0.00001057, 0.00001585, -0.00000569, 0.00000095, -0.00000579, -0.00000494, 0.00000318, -0.00001277, 0.00000081, 0.00000720, -0.00000558, 0.00000020, -0.00001066, -0.00000853, 0.00000649, -0.00000784, 0.00000497, -0.00000886, 0.00001138, 0.00000441, -0.00000250, 0.00000660, -0.00000029, -0.00000738, -0.00000454, 0.00000204, 0.00000233, 0.00000149, 0.00000083, -0.00000227, 0.00000001, -0.00000639, 0.00000188, -0.00000046, 0.00000519, 0.00000925, -0.00000869, 0.00000077, 0.00000539, -0.00000108, -0.00000728, -0.00001196, 0.00000470, -0.00000910, 0.00000115, 0.00000635, 0.00000401, -0.00001632, -0.00000089, -0.00000056, -0.00000009, -0.00000773, -0.00000191, 0.00000608, 0.00000005, -0.00001153, 0.00000496, -0.00000430, 0.00000309, -0.00001816, -0.00000720, -0.00008653, 0.00000675, 0.00011740, 0.00007327, 0.00011112, -0.00006713, -0.00004085, -0.00000440, -0.00001769, -0.00000640, -0.00000372, -0.00000406, -0.00000554, -0.00000134, 0.00000363, -0.00000115, -0.00000498, -0.00000228, 0.00001096, -0.00000239, -0.00000290, -0.00000470, -0.00000308, 0.00000426, -0.00000579, 0.00000127, 0.00000493, 0.00000305, 0.00000233, -0.00000538, -0.00000337, 0.00000239, 0.00000810, -0.00000738, 0.00000500, },
		{ -0.00000111, 0.00000039, -0.00000363, 0.00000032, -0.00000149, -0.00000643, 0.00000583, 0.00000587, -0.00000280, -0.00000214, 0.00000091, -0.00000451, -0.00000192, -0.00000607, 0.00000059, -0.00000739, -0.00000105, -0.00002609, -0.00002951, 0.00001102, 0.00006026, 0.00002208, 0.00004457, 0.00003701, -0.00002593, -0.00003186, -0.00000233, -0.00000471, 0.00000042, 0.00000012, -0.00000243, -0.00000699, 0.00000271, 0.00001038, -0.00000980, 0.00000511, -0.00001388, 0.00000688, -0.00000931, -0.00000599, 0.00000604, 0.00000120, -0.00000531, -0.00000117, -0.00001776, 0.00000839, -0.00000698, 0.00000277, -0.00000484, 0.00000136, 0.00001252, -0.00000438, 0.00000464, 0.00000367, -0.00000516, -0.00000713, -0.00000057, 0.00000357, 0.00000089, 0.00000224, -0.00000231, 0.00000018, -0.00000398, -0.00000275, 0.00000201, 0.00000034, 0.00001150, -0.00000253, -0.00000695, 0.00000804, 0.00000030, -0.00000309, -0.00001432, 0.00000073, -0.00000310, -0.00000558, 0.00000572, 0.00000702, -0.00000807, -0.00001179, 0.00000354, -0.00000194, -0.00000236, -0.00000860, 0.00000513, 0.00000397, -0.00000824, -0.00000351, 0.00000263, -0.00000216, -0.00000610, -0.00001433, -0.00004109, -0.00008010, 0.00011239, 0.00007865, 0.00012269, 0.00000431, -0.00008291, 0.00000062, -0.00001972, -0.00000860, -0.00000633, -0.00000209, -0.00000600, -0.00000370, 0.00000250, 0.00000155, -0.00000331, -0.00000622, 0.00000711, 0.00000525, -0.00000523, -0.00000234, -0.00000615, 0.00000310, -0.00000110, -0.00000475, 0.00000660, 0.00000207, 0.00000514, -0.00000463, -0.00000320, -0.00000274, 0.00001089, -0.00000434, -0.00000027, 0.00000222, },
		{ 0.00000225, -0.00000410, -0.00000048, 0.00000028, -0.00000549, -0.00000095, 0.00000889, 0.00000029, -0.00000363, 0.00000092, -0.00000309, -0.00000274, -0.00000496, -0.00000114, -0.00000478, -0.00000203, -0.00001324, -0.00003424, -0.00000809, 0.00004535, 0.00004153, 0.00002142, 0.00005486, -0.00000410, -0.00003206, -0.00001825, 0.00000405, -0.00000872, 0.00000718, -0.00000722, 0.00000111, -0.00001082, 0.00001824, -0.00000912, 0.00000415, -0.00000932, -0.00000179, 0.00000062, -0.00001110, 0.00000120, 0.00000671, -0.00000653, 0.00000160, -0.00001589, -0.00000115, 0.00000237, -0.00000581, 0.00000380, -0.00000853, 0.00001517, -0.00000064, 0.00000040, 0.00000581, -0.00000126, -0.00000786, -0.00000397, 0.00000301, 0.00000221, 0.00000131, 0.00000032, -0.00000234, 0.00000015, -0.00000612, 0.00000209, -0.00000113, 0.00000850, 0.00000557, -0.00001008, 0.00000354, 0.00000601, -0.00000293, -0.00000982, -0.00000858, 0.00000424, -0.00001025, 0.00000438, 0.00000487, 0.00000366, -0.00001809, 0.00000153, -0.00000163, 0.00000154, -0.00001072, 0.00000131, 0.00000381, 0.00000090, -0.00001338, 0.00000837, -0.00000802, 0.00000593, -0.00002356, -0.00000677, -0.00010013, 0.00003268, 0.00011773, 0.00009644, 0.00008996, -0.00009464, -0.00001920, -0.00001189, -0.00001347, -0.00000870, -0.00000167, -0.00000558, -0.00000426, -0.00000149, 0.00000473, -0.00000249, -0.00000497, -0.00000153, 0.00001143, -0.00000408, -0.00000216, -0.00000602, -0.00000092, 0.00000302, -0.00000613, 0.00000283, 0.00000466, 0.00000407, -0.00000019, -0.00000517, -0.00000346, 0.00000606, 0.00000489, -0.00000746, 0.00000553, -0.00000257, },
		{ -0.00000201, -0.00000259, 0.00000108, -0.00000246, -0.00000604, 0.00000714, 0.00000453, -0.00000264, -0.00000177, 0.00000047, -0.00000540, -0.00000188, -0.00000496, -0.00000005, -0.00000693, -0.00000053, -0.00003375, -0.00001861, 0.00002008, 0.00005626, 0.00001411, 0.00004846, 0.00002332, -0.00002263, -0.00003078, 0.00000166, -0.00000657, 0.00000268, -0.00000221, -0.00000077, -0.00000872, 0.00000650, 0.00000714, -0.00000787, 0.00000349, -0.00001327, 0.00000708, -0.00001014, -0.00000452, 0.00000614, -0.00000061, -0.00000315, -0.00000570, -0.00001477, 0.00001058, -0.00001069, 0.00000673, -0.00000933, 0.00000799, 0.00000812, -0.00000383, 0.00000576, 0.00000232, -0.00000588, -0.00000697, 0.00000032, 0.00000418, 0.00000022, 0.00000240, -0.00000324, 0.00000104, -0.00000461, -0.00000156, 0.00000042, 0.00000328, 0.00001064, -0.00000621, -0.00000512, 0.00001033, -0.00000123, -0.00000557, -0.00001381, 0.00000393, -0.00000685, -0.00000160, 0.00000452, 0.00000770, -0.00001122, -0.00000931, 0.00000251, 0.00000023, -0.00000530, -0.00000656, 0.00000426, 0.00000463, -0.00001084, 0.00000004, -0.00000036, 0.00000022, -0.00001055, -0.00001273, -0.00005388, -0.00007063, 0.00012933, 0.00008664, 0.00013403, -0.00004450, -0.00006562, 0.00000098, -0.00001973, -0.00000864, -0.00000547, -0.00000269, -0.00000535, -0.00000392, 0.00000398, 0.00000046, -0.00000338, -0.00000660, 0.00000909, 0.00000310, -0.00000458, -0.00000355, -0.00000498, 0.00000358, -0.00000267, -0.00000327, 0.00000685, 0.00000272, 0.00000409, -0.00000609, -0.00000279, -0.00000068, 0.00001171, -0.00000833, 0.00000274, -0.00000019, 0.00000123, },
		{ -0.00000526, 0.00000140, -0.00000036, -0.00000620, 0.00000089, 0.00000810, 0.00000003, -0.00000369, 0.00000159, -0.00000485, -0.00000218, -0.00000524, 0.00000062, -0.00000744, 0.00000251, -0.00002218, -0.00002903, 0.00000234, 0.00004849, 0.00003023, 0.00002486, 0.00004682, -0.00000915, -0.00002819, -0.00001631, 0.00000546, -0.00000929, 0.00000784, -0.00000833, 0.00000170, -0.00001031, 0.00001972, -0.00001185, 0.00000670, -0.00001240, 0.00000123, -0.00000192, -0.00000922, 0.00000124, 0.00000613, -0.00000670, 0.00000136, -0.00001944, 0.00000651, -0.00000398, -0.00000102, -0.00000010, -0.00000498, 0.00001585, -0.00000425, 0.00000340, 0.00000437, -0.00000192, -0.00000834, -0.00000323, 0.00000408, 0.00000166, 0.00000140, -0.00000048, -0.00000197, 0.00000002, -0.00000561, 0.00000157, -0.00000043, 0.00001067, 0.00000139, -0.00001054, 0.00000692, 0.00000566, -0.00000533, -0.00001138, -0.00000461, 0.00000198, -0.00000933, 0.00000614, 0.00000397, 0.00000270, -0.00001913, 0.00000349, -0.00000204, 0.00000214, -0.00001282, 0.00000391, 0.00000205, 0.00000108, -0.00001425, 0.00001080, -0.00001111, 0.00000813, -0.00002879, -0.00000742, -0.00011240, 0.00006119, 0.00011780, 0.00012342, 0.00004978, -0.00010502, -0.00000200, -0.00001926, -0.00000918, -0.00001101, 0.00000042, -0.00000696, -0.00000314, -0.00000136, 0.00000539, -0.00000359, -0.00000517, -0.00000033, 0.00001129, -0.00000535, -0.00000187, -0.00000672, 0.00000104, 0.00000135, -0.00000578, 0.00000409, 0.00000454, 0.00000462, -0.00000279, -0.00000449, -0.00000325, 0.00000978, 0.00000005, -0.00000558, 0.00000466, -0.00000190, 0.00000195, },
		{ -0.00000051, 0.00000111, -0.00000342, -0.00000519, 0.00000793, 0.00000352, -0.00000270, -0.00000111, -0.00000059, -0.00000572, -0.00000216, -0.00000297, -0.00000243, -0.00000367, -0.00000436, -0.00003710, -0.00000681, 0.00002580, 0.00005043, 0.00000891, 0.00005074, 0.00001053, -0.00001751, -0.00003144, 0.00000683, -0.00000984, 0.00000601, -0.00000558, 0.00000192, -0.00001120, 0.00001109, 0.00000279, -0.00000482, 0.00000065, -0.00001129, 0.00000604, -0.00000976, -0.00000407, 0.00000676, -0.00000259, -0.00000079, -0.00001154, -0.00000833, 0.00000861, -0.00001120, 0.00000851, -0.00001212, 0.00001411, 0.00000231, -0.00000123, 0.00000540, 0.00000151, -0.00000677, -0.00000656, 0.00000135, 0.00000450, -0.00000044, 0.00000252, -0.00000406, 0.00000199, -0.00000524, -0.00000076, -0.00000068, 0.00000657, 0.00000820, -0.00000906, -0.00000249, 0.00001239, -0.00000398, -0.00000747, -0.00001212, 0.00000573, -0.00001002, 0.00000293, 0.00000209, 0.00000929, -0.00001525, -0.00000572, 0.00000065, 0.00000316, -0.00000945, -0.00000302, 0.00000188, 0.00000667, -0.00001486, 0.00000549, -0.00000582, 0.00000531, -0.00001831, -0.00000811, -0.00007101, -0.00005356, 0.00014124, 0.00010485, 0.00012713, -0.00008835, -0.00003978, -0.00000610, -0.00001548, -0.00001157, -0.00000248, -0.00000478, -0.00000353, -0.00000499, 0.00000610, -0.00000138, -0.00000291, -0.00000722, 0.00001134, 0.00000046, -0.00000354, -0.00000510, -0.00000303, 0.00000303, -0.00000358, -0.00000175, 0.00000691, 0.00000346, 0.00000229, -0.00000672, -0.00000262, 0.00000248, 0.00001029, -0.00001082, 0.00000528, -0.00000227, 0.00000343, -0.00000489, },
		{ 0.00000275, -0.00000120, -0.00000662, 0.00000270, 0.00000709, -0.00000005, -0.00000374, 0.00000202, -0.00000654, -0.00000165, -0.00000474, 0.00000103, -0.00000849, 0.00000560, -0.00003183, -0.00001863, 0.00000864, 0.00005039, 0.00001861, 0.00003079, 0.00003550, -0.00000946, -0.00002738, -0.00001216, 0.00000465, -0.00000818, 0.00000690, -0.00000794, 0.00000104, -0.00000841, 0.00001965, -0.00001327, 0.00000801, -0.00001431, 0.00000333, -0.00000366, -0.00000795, 0.00000165, 0.00000503, -0.00000577, -0.00000116, -0.00001977, 0.00001214, -0.00001040, 0.00000505, -0.00000581, 0.00000137, 0.00001302, -0.00000533, 0.00000537, 0.00000296, -0.00000258, -0.00000866, -0.00000237, 0.00000518, 0.00000071, 0.00000183, -0.00000160, -0.00000109, -0.00000044, -0.00000499, 0.00000063, 0.00000155, 0.00001124, -0.00000270, -0.00001015, 0.00001091, 0.00000375, -0.00000756, -0.00001198, -0.00000056, -0.00000175, -0.00000612, 0.00000568, 0.00000461, 0.00000025, -0.00001860, 0.00000421, -0.00000109, 0.00000091, -0.00001308, 0.00000488, 0.00000188, -0.00000056, -0.00001277, 0.00001062, -0.00001165, 0.00000755, -0.00003162, -0.00001155, -0.00012036, 0.00008903, 0.00012272, 0.00014601, -0.00000425, -0.00009608, 0.00000566, -0.00002261, -0.00000750, -0.00001137, 0.00000112, -0.00000715, -0.00000302, -0.00000029, 0.00000503, -0.00000395, -0.00000599, 0.00000171, 0.00001017, -0.00000587, -0.00000227, -0.00000646, 0.00000227, -0.00000027, -0.00000494, 0.00000502, 0.00000460, 0.00000453, -0.00000511, -0.00000365, -0.00000230, 0.00001263, -0.00000553, -0.00000214, 0.00000248, 0.00000009, 0.00000006, -0.00000499, },
		{ 0.00000052, -0.00000427, -0.00000397, 0.00000817, 0.00000291, -0.00000304, -0.00000017, -0.00000222, -0.00000552, -0.00000241, -0.00000081, -0.00000571, 0.00000170, -0.00001267, -0.00003402, 0.00000238, 0.00003025, 0.00004221, 0.00000756, 0.00004973, 0.00000074, -0.00001280, -0.00003207, 0.00001181, -0.00001341, 0.00000951, -0.00000918, 0.00000494, -0.00001380, 0.00001582, -0.00000201, -0.00000132, -0.00000271, -0.00000865, 0.00000438, -0.00000874, -0.00000423, 0.00000760, -0.00000434, 0.00000088, -0.00001724, 0.00000025, 0.00000282, -0.00000782, 0.00000689, -0.00001167, 0.00001783, -0.00000310, 0.00000231, 0.00000394, 0.00000120, -0.00000780, -0.00000591, 0.00000254, 0.00000435, -0.00000085, 0.00000239, -0.00000454, 0.00000283, -0.00000582, -0.00000045, -0.00000082, 0.00000940, 0.00000475, -0.00001111, 0.00000121, 0.00001337, -0.00000737, -0.00000862, -0.00000942, 0.00000561, -0.00001149, 0.00000665, -0.00000042, 0.00001092, -0.00001937, -0.00000170, -0.00000140, 0.00000603, -0.00001391, 0.00000115, -0.00000108, 0.00000913, -0.00001918, 0.00001158, -0.00001249, 0.00001189, -0.00002835, -0.00000144, -0.00009135, -0.00002987, 0.00014991, 0.00013176, 0.00009724, -0.00011693, -0.00001378, -0.00001636, -0.00000920, -0.00001589, 0.00000167, -0.00000762, -0.00000119, -0.00000636, 0.00000834, -0.00000345, -0.00000237, -0.00000761, 0.00001339, -0.00000227, -0.00000245, -0.00000654, -0.00000077, 0.00000171, -0.00000375, -0.00000040, 0.00000688, 0.00000412, -0.00000008, -0.00000656, -0.00000257, 0.00000647, 0.00000645, -0.00001104, 0.00000648, -0.00000324, 0.00000460, -0.00000737, 0.00000206, },
		{ -0.00000204, -0.00000669, 0.00000426, 0.00000612, -0.00000020, -0.00000356, 0.00000195, -0.00000785, -0.00000133, -0.00000324, -0.00000052, -0.00000665, 0.00000483, -0.00003858, -0.00000629, 0.00001172, 0.00005097, 0.00000787, 0.00003773, 0.00002219, -0.00000585, -0.00002956, -0.00000560, 0.00000127, -0.00000500, 0.00000399, -0.00000570, -0.00000121, -0.00000485, 0.00001777, -0.00001313, 0.00000785, -0.00001476, 0.00000419, -0.00000421, -0.00000777, 0.00000289, 0.00000314, -0.00000379, -0.00000598, -0.00001603, 0.00001381, -0.00001445, 0.00001020, -0.00001144, 0.00000898, 0.00000741, -0.00000367, 0.00000568, 0.00000215, -0.00000351, -0.00000867, -0.00000141, 0.00000616, -0.00000052, 0.00000252, -0.00000296, 0.00000025, -0.00000127, -0.00000439, -0.00000027, 0.00000435, 0.00001018, -0.00000637, -0.00000874, 0.00001502, 0.00000019, -0.00000905, -0.00001169, 0.00000289, -0.00000607, -0.00000117, 0.00000298, 0.00000709, -0.00000409, -0.00001605, 0.00000328, 0.00000154, -0.00000250, -0.00001102, 0.00000370, 0.00000384, -0.00000463, -0.00000822, 0.00000679, -0.00000832, 0.00000248, -0.00003011, -0.00002127, -0.00012138, 0.00011351, 0.00013731, 0.00015393, -0.00006172, -0.00007138, 0.00000259, -0.00002011, -0.00000986, -0.00000860, -0.00000049, -0.00000549, -0.00000439, 0.00000208, 0.00000339, -0.00000334, -0.00000759, 0.00000470, 0.00000792, -0.00000551, -0.00000340, -0.00000514, 0.00000245, -0.00000141, -0.00000388, 0.00000563, 0.00000487, 0.00000366, -0.00000680, -0.00000294, -0.00000024, 0.00001361, -0.00001057, 0.00000197, -0.00000039, 0.00000286, -0.00000311, -0.00000319, 0.00000325, },
		{ -0.00000496, -0.00000249, 0.00000797, 0.00000261, -0.00000356, 0.00000087, -0.00000417, -0.00000507, -0.00000211, 0.00000045, -0.00000826, 0.00000711, -0.00002358, -0.00002471, 0.00000706, 0.00003501, 0.00003130, 0.00001069, 0.00004426, -0.00000438, -0.00001058, -0.00003082, 0.00001496, -0.00001581, 0.00001192, -0.00001182, 0.00000724, -0.00001558, 0.00001979, -0.00000645, 0.00000184, -0.00000570, -0.00000623, 0.00000300, -0.00000800, -0.00000415, 0.00000801, -0.00000519, 0.00000077, -0.00002090, 0.00000867, -0.00000507, -0.00000132, 0.00000192, -0.00000743, 0.00001784, -0.00000643, 0.00000542, 0.00000212, 0.00000109, -0.00000881, -0.00000510, 0.00000390, 0.00000362, -0.00000081, 0.00000183, -0.00000446, 0.00000335, -0.00000630, -0.00000055, 0.00000019, 0.00001111, 0.00000085, -0.00001237, 0.00000608, 0.00001248, -0.00001056, -0.00000915, -0.00000595, 0.00000340, -0.00001039, 0.00000822, -0.00000160, 0.00001131, -0.00002238, 0.00000164, -0.00000259, 0.00000769, -0.00001738, 0.00000462, -0.00000324, 0.00001046, -0.00002205, 0.00001627, -0.00001813, 0.00001753, -0.00003819, 0.00000478, -0.00011206, -0.00000245, 0.00016010, 0.00016076, 0.00004564, -0.00012352, 0.00000435, -0.00002445, -0.00000435, -0.00001909, 0.00000510, -0.00000981, 0.00000052, -0.00000706, 0.00000985, -0.00000499, -0.00000242, -0.00000715, 0.00001460, -0.00000454, -0.00000176, -0.00000735, 0.00000119, 0.00000007, -0.00000328, 0.00000066, 0.00000688, 0.00000451, -0.00000269, -0.00000582, -0.00000227, 0.00001053, 0.00000067, -0.00000876, 0.00000586, -0.00000260, 0.00000411, -0.00000860, 0.00000443, -0.00000045, },
		{ -0.00000641, 0.00000542, 0.00000539, -0.00000062, -0.00000301, 0.00000124, -0.00000869, -0.00000106, -0.00000116, -0.00000369, -0.00000178, -0.00000101, -0.00003939, 0.00000414, 0.00001390, 0.00004899, -0.00000010, 0.00004340, 0.00000919, -0.00000033, -0.00003338, 0.00000242, -0.00000397, -0.00000027, -0.00000043, -0.00000192, -0.00000476, 0.00000008, 0.00001433, -0.00001174, 0.00000660, -0.00001412, 0.00000411, -0.00000385, -0.00000852, 0.00000490, 0.00000071, -0.00000138, -0.00001221, -0.00000859, 0.00001068, -0.00001436, 0.00001230, -0.00001481, 0.00001566, 0.00000070, -0.00000003, 0.00000433, 0.00000214, -0.00000482, -0.00000831, -0.00000032, 0.00000681, -0.00000175, 0.00000326, -0.00000435, 0.00000189, -0.00000240, -0.00000393, -0.00000067, 0.00000727, 0.00000781, -0.00000956, -0.00000591, 0.00001832, -0.00000450, -0.00000966, -0.00001049, 0.00000495, -0.00000967, 0.00000421, -0.00000102, 0.00001085, -0.00000988, -0.00001178, 0.00000098, 0.00000544, -0.00000768, -0.00000697, 0.00000069, 0.00000754, -0.00001067, -0.00000107, -0.00000036, -0.00000120, -0.00000730, -0.00002375, -0.00003721, -0.00011451, 0.00013435, 0.00016337, 0.00013819, -0.00010972, -0.00003929, -0.00000813, -0.00001294, -0.00001561, -0.00000303, -0.00000420, -0.00000221, -0.00000705, 0.00000548, 0.00000073, -0.00000202, -0.00000966, 0.00000834, 0.00000477, -0.00000445, -0.00000502, -0.00000299, 0.00000157, -0.00000178, -0.00000286, 0.00000598, 0.00000535, 0.00000199, -0.00000762, -0.00000250, 0.00000306, 0.00001199, -0.00001374, 0.00000553, -0.00000300, 0.00000549, -0.00000676, -0.00000022, 0.00000283, -0.00000277, },
		{ -0.00000095, 0.00000755, 0.00000240, -0.00000405, 0.00000173, -0.00000614, -0.00000458, -0.00000088, -0.00000013, -0.00000836, 0.00000986, -0.00003363, -0.00001173, 0.00000758, 0.00004022, 0.00001843, 0.00001768, 0.00003441, -0.00000433, -0.00001192, -0.00002649, 0.00001508, -0.00001590, 0.00001221, -0.00001255, 0.00000793, -0.00001577, 0.00002226, -0.00000988, 0.00000409, -0.00000771, -0.00000471, 0.00000261, -0.00000835, -0.00000305, 0.00000737, -0.00000465, -0.00000190, -0.00002084, 0.00001435, -0.00001237, 0.00000624, -0.00000501, -0.00000014, 0.00001398, -0.00000671, 0.00000696, 0.00000076, 0.00000075, -0.00000957, -0.00000420, 0.00000537, 0.00000233, -0.00000023, 0.00000074, -0.00000368, 0.00000335, -0.00000669, -0.00000080, 0.00000217, 0.00001141, -0.00000310, -0.00001262, 0.00001173, 0.00000935, -0.00001283, -0.00000918, -0.00000221, -0.00000047, -0.00000661, 0.00000688, -0.00000039, 0.00000938, -0.00002318, 0.00000326, -0.00000194, 0.00000705, -0.00001865, 0.00000614, -0.00000334, 0.00000927, -0.00002183, 0.00001756, -0.00002035, 0.00001949, -0.00004487, 0.00000739, -0.00012965, 0.00002512, 0.00017808, 0.00018089, -0.00001943, -0.00010765, 0.00000992, -0.00002631, -0.00000390, -0.00001883, 0.00000599, -0.00000996, 0.00000048, -0.00000620, 0.00000987, -0.00000537, -0.00000360, -0.00000531, 0.00001442, -0.00000585, -0.00000186, -0.00000712, 0.00000230, -0.00000136, -0.00000241, 0.00000136, 0.00000707, 0.00000434, -0.00000511, -0.00000484, -0.00000121, 0.00001360, -0.00000593, -0.00000451, 0.00000354, -0.00000039, 0.00000176, -0.00000807, 0.00000601, -0.00000149, -0.00000547, },
		{ 0.00000616, 0.00000494, -0.00000133, -0.00000211, -0.00000003, -0.00000915, -0.00000039, 0.00000053, -0.00000717, 0.00000472, -0.00001117, -0.00003316, 0.00000982, 0.00001747, 0.00004312, -0.00000340, 0.00004545, -0.00000093, 0.00000450, -0.00003670, 0.00001010, -0.00000955, 0.00000477, -0.00000523, 0.00000240, -0.00000878, 0.00000560, 0.00001004, -0.00000975, 0.00000503, -0.00001320, 0.00000393, -0.00000346, -0.00000939, 0.00000700, -0.00000154, 0.00000031, -0.00001811, 0.00000083, 0.00000354, -0.00000974, 0.00001022, -0.00001429, 0.00001934, -0.00000506, 0.00000420, 0.00000198, 0.00000271, -0.00000639, -0.00000766, 0.00000097, 0.00000691, -0.00000266, 0.00000374, -0.00000542, 0.00000352, -0.00000377, -0.00000362, -0.00000026, 0.00000963, 0.00000459, -0.00001221, -0.00000141, 0.00001976, -0.00000937, -0.00000960, -0.00000829, 0.00000495, -0.00001119, 0.00000829, -0.00000470, 0.00001453, -0.00001591, -0.00000686, -0.00000167, 0.00000947, -0.00001342, -0.00000212, -0.00000290, 0.00001159, -0.00001707, 0.00000690, -0.00000897, 0.00000783, -0.00002010, -0.00001409, -0.00005782, -0.00010134, 0.00015490, 0.00019728, 0.00009492, -0.00013676, -0.00001018, -0.00002054, -0.00000468, -0.00002220, 0.00000348, -0.00000862, 0.00000154, -0.00001002, 0.00000906, -0.00000214, -0.00000075, -0.00001143, 0.00001189, 0.00000131, -0.00000316, -0.00000660, -0.00000054, -0.00000008, -0.00000134, -0.00000210, 0.00000621, 0.00000589, -0.00000036, -0.00000755, -0.00000221, 0.00000726, 0.00000759, -0.00001413, 0.00000746, -0.00000439, 0.00000697, -0.00000986, 0.00000322, 0.00000170, -0.00000332, -0.00000580, },
		{ 0.00000712, 0.00000205, -0.00000426, 0.00000218, -0.00000794, -0.00000403, 0.00000118, -0.00000279, -0.00000505, 0.00000779, -0.00003915, 0.00000094, 0.00000627, 0.00004442, 0.00000565, 0.00002655, 0.00002180, -0.00000017, -0.00001646, -0.00001917, 0.00001204, -0.00001343, 0.00001013, -0.00001103, 0.00000667, -0.00001406, 0.00002296, -0.00001211, 0.00000539, -0.00000873, -0.00000413, 0.00000331, -0.00000999, -0.00000059, 0.00000545, -0.00000277, -0.00000706, -0.00001632, 0.00001533, -0.00001640, 0.00001220, -0.00001154, 0.00000823, 0.00000736, -0.00000402, 0.00000643, 0.00000040, -0.00000011, -0.00000990, -0.00000323, 0.00000677, 0.00000065, 0.00000079, -0.00000078, -0.00000220, 0.00000272, -0.00000698, -0.00000084, 0.00000464, 0.00001037, -0.00000694, -0.00001140, 0.00001725, 0.00000433, -0.00001389, -0.00000864, 0.00000102, -0.00000491, -0.00000098, 0.00000293, 0.00000333, 0.00000482, -0.00002134, 0.00000273, 0.00000088, 0.00000373, -0.00001722, 0.00000522, -0.00000090, 0.00000505, -0.00001784, 0.00001447, -0.00001783, 0.00001597, -0.00004622, 0.00000401, -0.00014155, 0.00005066, 0.00020883, 0.00017971, -0.00008419, -0.00007564, 0.00000329, -0.00002102, -0.00000874, -0.00001420, 0.00000353, -0.00000747, -0.00000176, -0.00000345, 0.00000819, -0.00000443, -0.00000599, -0.00000199, 0.00001267, -0.00000601, -0.00000284, -0.00000573, 0.00000222, -0.00000212, -0.00000144, 0.00000168, 0.00000757, 0.00000338, -0.00000689, -0.00000397, 0.00000104, 0.00001453, -0.00001180, 0.00000054, 0.00000028, 0.00000274, -0.00000203, -0.00000575, 0.00000646, -0.00000236, -0.00000577, 0.00000049, },
		{ 0.00000469, -0.00000222, -0.00000101, -0.00000169, -0.00000941, 0.00000113, 0.00000069, -0.00000907, 0.00001028, -0.00002288, -0.00002136, 0.00001009, 0.00002343, 0.00003297, -0.00000118, 0.00004245, -0.00000630, 0.00000650, -0.00003754, 0.00001557, -0.00001381, 0.00000865, -0.00000909, 0.00000608, -0.00001222, 0.00001077, 0.00000566, -0.00000788, 0.00000392, -0.00001287, 0.00000460, -0.00000409, -0.00000929, 0.00000823, -0.00000271, -0.00000001, -0.00002153, 0.00000952, -0.00000532, -0.00000190, 0.00000436, -0.00000952, 0.00001877, -0.00000811, 0.00000746, -0.00000041, 0.00000339, -0.00000796, -0.00000684, 0.00000250, 0.00000629, -0.00000297, 0.00000369, -0.00000588, 0.00000482, -0.00000523, -0.00000335, 0.00000101, 0.00001098, 0.00000090, -0.00001409, 0.00000461, 0.00001861, -0.00001343, -0.00000917, -0.00000524, 0.00000269, -0.00000972, 0.00000956, -0.00000635, 0.00001645, -0.00002058, -0.00000276, -0.00000327, 0.00001214, -0.00001812, 0.00000196, -0.00000543, 0.00001413, -0.00002168, 0.00001317, -0.00001625, 0.00001572, -0.00003278, -0.00000427, -0.00007986, -0.00008538, 0.00018183, 0.00022913, 0.00002814, -0.00013689, 0.00000730, -0.00002837, 0.00000034, -0.00002628, 0.00000833, -0.00001176, 0.00000414, -0.00001195, 0.00001166, -0.00000423, -0.00000041, -0.00001203, 0.00001446, -0.00000170, -0.00000226, -0.00000752, 0.00000152, -0.00000195, -0.00000026, -0.00000178, 0.00000654, 0.00000622, -0.00000305, -0.00000681, -0.00000169, 0.00001149, 0.00000104, -0.00001155, 0.00000718, -0.00000396, 0.00000655, -0.00001147, 0.00000629, 0.00000023, -0.00000373, -0.00000496, 0.00000657, },
		{ 0.00000140, -0.00000407, 0.00000211, -0.00000953, -0.00000309, 0.00000332, -0.00000672, 0.00000119, 0.00000034, -0.00003770, 0.00000953, 0.00000615, 0.00004536, -0.00000432, 0.00003444, 0.00000910, 0.00000578, -0.00002249, -0.00001020, 0.00000687, -0.00000917, 0.00000632, -0.00000777, 0.00000386, -0.00001080, 0.00002217, -0.00001351, 0.00000627, -0.00000937, -0.00000384, 0.00000439, -0.00001235, 0.00000277, 0.00000273, -0.00000038, -0.00001359, -0.00000805, 0.00001108, -0.00001548, 0.00001428, -0.00001522, 0.00001518, 0.00000001, 0.00000054, 0.00000410, 0.00000110, -0.00000158, -0.00000978, -0.00000213, 0.00000783, -0.00000108, 0.00000196, -0.00000245, -0.00000025, 0.00000149, -0.00000716, -0.00000037, 0.00000701, 0.00000828, -0.00001061, -0.00000821, 0.00002138, -0.00000164, -0.00001392, -0.00000730, 0.00000276, -0.00000833, 0.00000476, -0.00000225, 0.00000857, -0.00000161, -0.00001742, 0.00000058, 0.00000526, -0.00000166, -0.00001362, 0.00000247, 0.00000332, -0.00000133, -0.00001100, 0.00000777, -0.00001114, 0.00000720, -0.00004211, -0.00000560, -0.00014769, 0.00007563, 0.00025258, 0.00014777, -0.00013315, -0.00003870, -0.00001036, -0.00001120, -0.00001714, -0.00000636, -0.00000148, -0.00000298, -0.00000567, 0.00000067, 0.00000530, -0.00000267, -0.00000904, 0.00000230, 0.00000971, -0.00000528, -0.00000438, -0.00000346, 0.00000100, -0.00000194, -0.00000070, 0.00000173, 0.00000843, 0.00000148, -0.00000773, -0.00000340, 0.00000462, 0.00001251, -0.00001541, 0.00000494, -0.00000277, 0.00000574, -0.00000631, -0.00000213, 0.00000579, -0.00000295, -0.00000587, 0.00000170, 0.00000685, },
		{ -0.00000307, 0.00000004, -0.00000355, -0.00000949, 0.00000366, -0.00000140, -0.00000776, 0.00001210, -0.00003226, -0.00000754, 0.00000673, 0.00003080, 0.00001981, 0.00000573, 0.00003459, -0.00000654, 0.00000478, -0.00003489, 0.00001775, -0.00001564, 0.00001037, -0.00001101, 0.00000820, -0.00001430, 0.00001492, 0.00000169, -0.00000642, 0.00000358, -0.00001353, 0.00000661, -0.00000643, -0.00000741, 0.00000782, -0.00000217, -0.00000316, -0.00002074, 0.00001471, -0.00001279, 0.00000653, -0.00000338, -0.00000164, 0.00001410, -0.00000767, 0.00000863, -0.00000192, 0.00000365, -0.00000923, -0.00000594, 0.00000421, 0.00000497, -0.00000259, 0.00000298, -0.00000549, 0.00000549, -0.00000666, -0.00000292, 0.00000289, 0.00001117, -0.00000305, -0.00001469, 0.00001140, 0.00001483, -0.00001609, -0.00000848, -0.00000188, -0.00000125, -0.00000534, 0.00000738, -0.00000491, 0.00001545, -0.00002265, -0.00000067, -0.00000273, 0.00001222, -0.00002046, 0.00000397, -0.00000560, 0.00001372, -0.00002279, 0.00001563, -0.00001964, 0.00001947, -0.00004200, 0.00000228, -0.00009982, -0.00007024, 0.00022277, 0.00024449, -0.00004992, -0.00011216, 0.00000943, -0.00002783, -0.00000085, -0.00002534, 0.00000947, -0.00001201, 0.00000427, -0.00001179, 0.00001242, -0.00000481, -0.00000163, -0.00001082, 0.00001538, -0.00000363, -0.00000221, -0.00000731, 0.00000259, -0.00000344, 0.00000117, -0.00000197, 0.00000724, 0.00000595, -0.00000556, -0.00000579, -0.00000035, 0.00001453, -0.00000630, -0.00000670, 0.00000489, -0.00000178, 0.00000408, -0.00001103, 0.00000835, -0.00000119, -0.00000403, -0.00000401, 0.00000683, 0.00000442, },
		{ -0.00000348, 0.00000153, -0.00001097, -0.00000128, 0.00000437, -0.00001019, 0.00000835, -0.00001084, -0.00002924, 0.00001204, 0.00000941, 0.00004128, -0.00000915, 0.00003867, -0.00000094, 0.00001086, -0.00002771, -0.00000161, 0.00000128, -0.00000454, 0.00000207, -0.00000386, 0.00000040, -0.00000678, 0.00002054, -0.00001469, 0.00000755, -0.00001060, -0.00000274, 0.00000467, -0.00001420, 0.00000597, 0.00000027, 0.00000113, -0.00001943, 0.00000181, 0.00000293, -0.00000968, 0.00001155, -0.00001441, 0.00001852, -0.00000578, 0.00000526, 0.00000091, 0.00000247, -0.00000343, -0.00000932, -0.00000079, 0.00000826, -0.00000245, 0.00000289, -0.00000387, 0.00000180, -0.00000022, -0.00000717, 0.00000076, 0.00000876, 0.00000547, -0.00001392, -0.00000288, 0.00002297, -0.00000733, -0.00001342, -0.00000490, 0.00000220, -0.00000920, 0.00000858, -0.00000660, 0.00001350, -0.00000827, -0.00001289, -0.00000182, 0.00000970, -0.00000760, -0.00000933, -0.00000057, 0.00000754, -0.00000780, -0.00000363, -0.00000005, -0.00000282, -0.00000441, -0.00003479, -0.00001917, -0.00015099, 0.00010613, 0.00030229, 0.00008298, -0.00015413, -0.00000904, -0.00002331, -0.00000189, -0.00002533, 0.00000188, -0.00000694, 0.00000177, -0.00000980, 0.00000493, 0.00000233, -0.00000115, -0.00001169, 0.00000655, 0.00000635, -0.00000430, -0.00000585, -0.00000096, -0.00000096, -0.00000084, -0.00000044, 0.00000174, 0.00000947, -0.00000126, -0.00000755, -0.00000303, 0.00000913, 0.00000742, -0.00001576, 0.00000741, -0.00000448, 0.00000745, -0.00000989, 0.00000193, 0.00000434, -0.00000328, -0.00000577, 0.00000266, 0.00000677, 0.00000042, },
		{ 0.00000085, -0.00000551, -0.00000916, 0.00000675, -0.00000556, -0.00000274, 0.00000843, -0.00003584, 0.00000395, 0.00000303, 0.00003707, 0.00000637, 0.00001494, 0.00002372, -0.00000296, 0.00000007, -0.00002918, 0.00001677, -0.00001501, 0.00000986, -0.00001083, 0.00000856, -0.00001484, 0.00001792, -0.00000192, -0.00000508, 0.00000362, -0.00001479, 0.00000965, -0.00001033, -0.00000370, 0.00000572, -0.00000007, -0.00000893, -0.00001521, 0.00001455, -0.00001609, 0.00001252, -0.00001021, 0.00000699, 0.00000685, -0.00000416, 0.00000739, -0.00000207, 0.00000317, -0.00001003, -0.00000498, 0.00000588, 0.00000319, -0.00000167, 0.00000172, -0.00000430, 0.00000537, -0.00000794, -0.00000209, 0.00000492, 0.00001033, -0.00000723, -0.00001333, 0.00001765, 0.00000916, -0.00001735, -0.00000730, 0.00000080, -0.00000550, 0.00000070, 0.00000246, -0.00000057, 0.00001142, -0.00002185, -0.00000087, 0.00000017, 0.00000948, -0.00002011, 0.00000365, -0.00000326, 0.00001019, -0.00002007, 0.00001367, -0.00001819, 0.00001763, -0.00004596, 0.00000368, -0.00011615, -0.00005670, 0.00028250, 0.00022850, -0.00012139, -0.00007238, -0.00000123, -0.00001934, -0.00000824, -0.00001906, 0.00000644, -0.00000899, 0.00000166, -0.00000934, 0.00001126, -0.00000390, -0.00000432, -0.00000782, 0.00001452, -0.00000432, -0.00000308, -0.00000588, 0.00000236, -0.00000407, 0.00000256, -0.00000268, 0.00000853, 0.00000468, -0.00000733, -0.00000494, 0.00000231, 0.00001510, -0.00001262, -0.00000102, 0.00000152, 0.00000139, 0.00000010, -0.00000863, 0.00000907, -0.00000231, -0.00000426, -0.00000308, 0.00000711, 0.00000392, -0.00000366, },
		{ 0.00000053, -0.00001231, 0.00000175, 0.00000322, -0.00001118, 0.00001349, -0.00002233, -0.00001635, 0.00000906, 0.00001626, 0.00003201, -0.00000800, 0.00003780, -0.00000653, 0.00001310, -0.00003028, 0.00000480, -0.00000306, -0.00000100, -0.00000128, -0.00000049, -0.00000273, -0.00000279, 0.00001865, -0.00001614, 0.00000980, -0.00001318, 0.00000010, 0.00000295, -0.00001423, 0.00000776, -0.00000080, 0.00000030, -0.00002224, 0.00001021, -0.00000627, -0.00000091, 0.00000488, -0.00000905, 0.00001721, -0.00000831, 0.00000847, -0.00000199, 0.00000388, -0.00000532, -0.00000869, 0.00000086, 0.00000789, -0.00000315, 0.00000324, -0.00000463, 0.00000351, -0.00000219, -0.00000691, 0.00000248, 0.00000962, 0.00000214, -0.00001640, 0.00000415, 0.00002148, -0.00001176, -0.00001276, -0.00000157, -0.00000079, -0.00000667, 0.00000891, -0.00000822, 0.00001621, -0.00001330, -0.00000944, -0.00000284, 0.00001249, -0.00001231, -0.00000609, -0.00000207, 0.00000969, -0.00001195, 0.00000146, -0.00000580, 0.00000356, -0.00001511, -0.00002798, -0.00003289, -0.00015615, 0.00015182, 0.00034343, -0.00000669, -0.00014276, 0.00000457, -0.00002868, 0.00000189, -0.00002926, 0.00000726, -0.00001029, 0.00000474, -0.00001248, 0.00000792, 0.00000046, -0.00000096, -0.00001284, 0.00000967, 0.00000353, -0.00000381, -0.00000652, 0.00000100, -0.00000302, 0.00000095, -0.00000085, 0.00000203, 0.00001031, -0.00000443, -0.00000659, -0.00000241, 0.00001357, 0.00000007, -0.00001275, 0.00000736, -0.00000423, 0.00000711, -0.00001177, 0.00000548, 0.00000257, -0.00000346, -0.00000551, 0.00000344, 0.00000676, -0.00000075, -0.00000260, },
		{ -0.00000757, -0.00000797, 0.00000934, -0.00001049, 0.00000487, -0.00000053, -0.00003195, 0.00000976, 0.00000233, 0.00003935, -0.00000412, 0.00002326, 0.00001269, 0.00000200, -0.00000569, -0.00002205, 0.00001391, -0.00001291, 0.00000798, -0.00000924, 0.00000766, -0.00001426, 0.00002015, -0.00000571, -0.00000304, 0.00000304, -0.00001555, 0.00001253, -0.00001470, 0.00000101, 0.00000270, 0.00000250, -0.00001587, -0.00000606, 0.00000898, -0.00001380, 0.00001382, -0.00001348, 0.00001356, -0.00000059, 0.00000094, 0.00000434, -0.00000097, 0.00000195, -0.00001036, -0.00000384, 0.00000718, 0.00000136, -0.00000057, 0.00000029, -0.00000257, 0.00000448, -0.00000898, -0.00000067, 0.00000660, 0.00000874, -0.00001155, -0.00000947, 0.00002188, 0.00000293, -0.00001773, -0.00000512, 0.00000160, -0.00000815, 0.00000625, -0.00000329, 0.00000520, 0.00000554, -0.00001914, -0.00000245, 0.00000444, 0.00000486, -0.00001798, 0.00000204, 0.00000033, 0.00000500, -0.00001510, 0.00000884, -0.00001335, 0.00001137, -0.00004555, 0.00000092, -0.00013079, -0.00004031, 0.00035876, 0.00017150, -0.00016803, -0.00003199, -0.00001697, -0.00000743, -0.00001866, -0.00000969, 0.00000081, -0.00000395, -0.00000268, -0.00000551, 0.00000903, -0.00000235, -0.00000761, -0.00000382, 0.00001246, -0.00000420, -0.00000441, -0.00000366, 0.00000099, -0.00000365, 0.00000360, -0.00000374, 0.00001047, 0.00000217, -0.00000794, -0.00000450, 0.00000644, 0.00001233, -0.00001621, 0.00000378, -0.00000159, 0.00000434, -0.00000427, -0.00000492, 0.00000857, -0.00000304, -0.00000439, -0.00000225, 0.00000753, 0.00000304, -0.00000383, 0.00000130, },
		{ -0.00001338, 0.00000586, -0.00000055, -0.00000829, 0.00001394, -0.00003007, -0.00000327, 0.00000352, 0.00002474, 0.00001937, -0.00000200, 0.00003227, -0.00000762, 0.00001203, -0.00002958, 0.00000824, -0.00000533, 0.00000071, -0.00000298, 0.00000164, -0.00000496, 0.00000076, 0.00001662, -0.00001774, 0.00001293, -0.00001718, 0.00000491, -0.00000124, -0.00001169, 0.00000735, 0.00000019, -0.00000368, -0.00002034, 0.00001428, -0.00001312, 0.00000769, -0.00000327, -0.00000076, 0.00001176, -0.00000705, 0.00000915, -0.00000364, 0.00000472, -0.00000690, -0.00000800, 0.00000274, 0.00000675, -0.00000309, 0.00000290, -0.00000452, 0.00000453, -0.00000420, -0.00000622, 0.00000449, 0.00000955, -0.00000169, -0.00001728, 0.00001176, 0.00001721, -0.00001454, -0.00001193, 0.00000199, -0.00000537, -0.00000113, 0.00000541, -0.00000619, 0.00001556, -0.00001551, -0.00000821, -0.00000141, 0.00001248, -0.00001456, -0.00000502, -0.00000095, 0.00000852, -0.00001224, 0.00000232, -0.00000712, 0.00000513, -0.00002170, -0.00002491, -0.00004365, -0.00016684, 0.00022244, 0.00035684, -0.00010371, -0.00010463, 0.00000013, -0.00002347, -0.00000243, -0.00002655, 0.00000769, -0.00000986, 0.00000455, -0.00001259, 0.00000880, 0.00000038, -0.00000268, -0.00001186, 0.00001095, 0.00000190, -0.00000431, -0.00000591, 0.00000178, -0.00000451, 0.00000304, -0.00000203, 0.00000302, 0.00001039, -0.00000740, -0.00000535, -0.00000089, 0.00001655, -0.00000791, -0.00000733, 0.00000510, -0.00000214, 0.00000462, -0.00001145, 0.00000786, 0.00000094, -0.00000358, -0.00000516, 0.00000419, 0.00000662, -0.00000192, -0.00000161, -0.00000082, },
		{ -0.00000541, 0.00001013, -0.00001414, 0.00001255, -0.00001235, -0.00002165, 0.00000890, 0.00000635, 0.00003585, -0.00000922, 0.00002800, 0.00000415, 0.00000583, -0.00001028, -0.00001554, 0.00001097, -0.00001085, 0.00000607, -0.00000737, 0.00000643, -0.00001332, 0.00002220, -0.00001028, 0.00000060, 0.00000066, -0.00001443, 0.00001369, -0.00001792, 0.00000522, 0.00000013, 0.00000388, -0.00002162, 0.00000406, -0.00000009, -0.00000656, 0.00000986, -0.00001171, 0.00001591, -0.00000582, 0.00000569, 0.00000066, 0.00000081, 0.00000030, -0.00001040, -0.00000235, 0.00000776, -0.00000005, 0.00000023, -0.00000083, -0.00000080, 0.00000305, -0.00000972, 0.00000139, 0.00000752, 0.00000661, -0.00001570, -0.00000306, 0.00002292, -0.00000246, -0.00001794, -0.00000157, -0.00000036, -0.00000754, 0.00000899, -0.00000743, 0.00001012, -0.00000015, -0.00001635, -0.00000364, 0.00000821, 0.00000026, -0.00001591, 0.00000114, 0.00000285, 0.00000083, -0.00001089, 0.00000441, -0.00000861, 0.00000423, -0.00004427, -0.00000212, -0.00014932, -0.00001066, 0.00043972, 0.00007401, -0.00017725, -0.00000506, -0.00002823, 0.00000133, -0.00002699, -0.00000116, -0.00000440, 0.00000068, -0.00000675, -0.00000197, 0.00000723, -0.00000156, -0.00001010, -0.00000014, 0.00001028, -0.00000416, -0.00000538, -0.00000146, -0.00000095, -0.00000228, 0.00000399, -0.00000486, 0.00001287, -0.00000150, -0.00000727, -0.00000439, 0.00001155, 0.00000621, -0.00001604, 0.00000631, -0.00000318, 0.00000580, -0.00000773, -0.00000091, 0.00000733, -0.00000351, -0.00000440, -0.00000157, 0.00000813, 0.00000174, -0.00000354, 0.00000134, -0.00000976, },
		{ 0.00001028, -0.00000631, -0.00000155, 0.00000845, -0.00003101, 0.00000563, -0.00000065, 0.00003152, 0.00000665, 0.00000608, 0.00002418, -0.00000580, 0.00000880, -0.00002649, 0.00000927, -0.00000588, 0.00000085, -0.00000319, 0.00000260, -0.00000637, 0.00000406, 0.00001401, -0.00001869, 0.00001604, -0.00002168, 0.00001089, -0.00000734, -0.00000683, 0.00000487, 0.00000290, -0.00001035, -0.00001351, 0.00001245, -0.00001485, 0.00001284, -0.00000974, 0.00000767, 0.00000413, -0.00000281, 0.00000726, -0.00000368, 0.00000471, -0.00000804, -0.00000719, 0.00000457, 0.00000515, -0.00000252, 0.00000204, -0.00000361, 0.00000472, -0.00000607, -0.00000492, 0.00000632, 0.00000878, -0.00000606, -0.00001569, 0.00001829, 0.00001136, -0.00001606, -0.00001038, 0.00000450, -0.00000977, 0.00000566, -0.00000072, -0.00000117, 0.00001186, -0.00001500, -0.00000911, 0.00000230, 0.00000979, -0.00001446, -0.00000586, 0.00000237, 0.00000451, -0.00000908, -0.00000082, -0.00000394, 0.00000143, -0.00002341, -0.00002617, -0.00005169, -0.00018198, 0.00032274, 0.00032441, -0.00018510, -0.00005402, -0.00001684, -0.00001010, -0.00001346, -0.00001790, 0.00000343, -0.00000581, 0.00000134, -0.00001032, 0.00000783, 0.00000173, -0.00000593, -0.00000904, 0.00001050, 0.00000143, -0.00000569, -0.00000408, 0.00000119, -0.00000501, 0.00000507, -0.00000399, 0.00000506, 0.00000913, -0.00000943, -0.00000437, 0.00000213, 0.00001667, -0.00001445, -0.00000122, 0.00000177, 0.00000086, 0.00000073, -0.00000916, 0.00000884, -0.00000033, -0.00000367, -0.00000482, 0.00000507, 0.00000612, -0.00000283, -0.00000071, -0.00000250, -0.00001375, },
		{ 0.00000814, -0.00001449, 0.00001717, -0.00002312, -0.00000841, 0.00000318, 0.00001442, 0.00002683, -0.00000837, 0.00002801, -0.00000052, 0.00000704, -0.00001223, -0.00001117, 0.00000933, -0.00001005, 0.00000529, -0.00000625, 0.00000566, -0.00001258, 0.00002438, -0.00001584, 0.00000623, -0.00000421, -0.00001047, 0.00001188, -0.00001849, 0.00000747, -0.00000064, 0.00000250, -0.00002371, 0.00001184, -0.00000931, 0.00000308, 0.00000207, -0.00000527, 0.00001326, -0.00000724, 0.00000836, -0.00000233, 0.00000250, -0.00000135, -0.00001037, -0.00000042, 0.00000743, -0.00000070, 0.00000036, -0.00000117, 0.00000050, 0.00000139, -0.00001007, 0.00000395, 0.00000752, 0.00000400, -0.00001900, 0.00000518, 0.00002046, -0.00000609, -0.00001836, 0.00000321, -0.00000513, -0.00000292, 0.00000738, -0.00000792, 0.00001212, -0.00000362, -0.00001535, -0.00000264, 0.00000958, -0.00000239, -0.00001576, 0.00000290, 0.00000203, 0.00000033, -0.00001057, 0.00000401, -0.00000808, 0.00000066, -0.00004669, -0.00000047, -0.00017877, 0.00004656, 0.00050451, -0.00005076, -0.00014703, -0.00000006, -0.00002771, 0.00000131, -0.00002855, 0.00000265, -0.00000613, 0.00000243, -0.00000856, -0.00000035, 0.00000723, -0.00000284, -0.00001048, 0.00000196, 0.00000910, -0.00000510, -0.00000510, -0.00000021, -0.00000267, -0.00000033, 0.00000356, -0.00000558, 0.00001521, -0.00000587, -0.00000559, -0.00000411, 0.00001655, -0.00000226, -0.00001215, 0.00000597, -0.00000262, 0.00000501, -0.00000922, 0.00000237, 0.00000591, -0.00000389, -0.00000426, -0.00000099, 0.00000877, 0.00000014, -0.00000284, 0.00000094, -0.00001196, -0.00000130, },
		{ -0.00001237, 0.00000737, -0.00000201, -0.00002439, 0.00000768, -0.00000023, 0.00003344, -0.00000267, 0.00001296, 0.00001639, -0.00000355, 0.00000553, -0.00002290, 0.00000945, -0.00000600, 0.00000059, -0.00000284, 0.00000313, -0.00000763, 0.00000775, 0.00000997, -0.00001779, 0.00001762, -0.00002510, 0.00001639, -0.00001380, -0.00000092, 0.00000144, 0.00000597, -0.00001791, -0.00000334, 0.00000518, -0.00001047, 0.00001244, -0.00001179, 0.00001324, -0.00000296, 0.00000254, 0.00000371, -0.00000243, 0.00000398, -0.00000885, -0.00000605, 0.00000592, 0.00000362, -0.00000191, 0.00000118, -0.00000233, 0.00000421, -0.00000774, -0.00000286, 0.00000748, 0.00000759, -0.00001092, -0.00001106, 0.00002203, 0.00000562, -0.00001722, -0.00000732, 0.00000445, -0.00001173, 0.00001104, -0.00000694, 0.00000475, 0.00000684, -0.00001328, -0.00001064, 0.00000669, 0.00000598, -0.00001357, -0.00000682, 0.00000574, 0.00000011, -0.00000515, -0.00000517, 0.00000086, -0.00000475, -0.00002303, -0.00002841, -0.00006228, -0.00019303, 0.00044752, 0.00023599, -0.00022916, -0.00000936, -0.00003544, 0.00000449, -0.00002607, -0.00000704, -0.00000278, -0.00000030, -0.00000313, -0.00000719, 0.00000643, 0.00000311, -0.00000930, -0.00000561, 0.00000930, 0.00000131, -0.00000720, -0.00000171, -0.00000037, -0.00000449, 0.00000676, -0.00000661, 0.00000828, 0.00000611, -0.00000993, -0.00000404, 0.00000682, 0.00001299, -0.00001767, 0.00000361, -0.00000105, 0.00000333, -0.00000325, -0.00000577, 0.00000872, -0.00000129, -0.00000365, -0.00000463, 0.00000624, 0.00000506, -0.00000325, -0.00000006, -0.00000453, -0.00001287, 0.00001372, },
		{ -0.00001046, 0.00001629, -0.00002882, 0.00000307, -0.00000359, 0.00002363, 0.00001476, -0.00000316, 0.00002415, -0.00000179, 0.00000579, -0.00001155, -0.00000916, 0.00000932, -0.00001081, 0.00000599, -0.00000621, 0.00000556, -0.00001204, 0.00002637, -0.00002184, 0.00001337, -0.00001127, -0.00000371, 0.00000679, -0.00001578, 0.00000701, 0.00000097, -0.00000235, -0.00002058, 0.00001443, -0.00001503, 0.00001148, -0.00000652, 0.00000372, 0.00000660, -0.00000469, 0.00000812, -0.00000373, 0.00000347, -0.00000266, -0.00001034, 0.00000183, 0.00000627, -0.00000056, -0.00000024, -0.00000056, 0.00000097, -0.00000022, -0.00000992, 0.00000665, 0.00000670, 0.00000082, -0.00002045, 0.00001375, 0.00001521, -0.00000792, -0.00001873, 0.00000825, -0.00001152, 0.00000489, 0.00000145, -0.00000417, 0.00001026, -0.00000387, -0.00001708, 0.00000142, 0.00000760, -0.00000213, -0.00001834, 0.00000806, -0.00000300, 0.00000465, -0.00001567, 0.00000960, -0.00001430, 0.00000356, -0.00005579, 0.00000881, -0.00022363, 0.00014463, 0.00052755, -0.00017850, -0.00008757, -0.00001656, -0.00001375, -0.00000926, -0.00002142, -0.00000008, -0.00000286, 0.00000006, -0.00000712, -0.00000135, 0.00000954, -0.00000660, -0.00000824, 0.00000186, 0.00000948, -0.00000744, -0.00000312, -0.00000051, -0.00000350, 0.00000180, 0.00000217, -0.00000534, 0.00001672, -0.00001012, -0.00000349, -0.00000294, 0.00001986, -0.00001117, -0.00000577, 0.00000329, -0.00000019, 0.00000206, -0.00000840, 0.00000431, 0.00000475, -0.00000431, -0.00000404, -0.00000035, 0.00000924, -0.00000153, -0.00000188, 0.00000007, -0.00001383, 0.00000395, 0.00000325, },
		{ 0.00001554, -0.00001432, -0.00001227, 0.00000299, 0.00000584, 0.00002897, -0.00000634, 0.00001618, 0.00001122, -0.00000305, 0.00000425, -0.00002079, 0.00001055, -0.00000721, 0.00000132, -0.00000317, 0.00000420, -0.00000948, 0.00001244, 0.00000378, -0.00001393, 0.00001625, -0.00002575, 0.00001950, -0.00001863, 0.00000419, -0.00000129, 0.00000748, -0.00002371, 0.00000708, -0.00000498, -0.00000129, 0.00000642, -0.00000821, 0.00001384, -0.00000702, 0.00000684, -0.00000004, -0.00000068, 0.00000297, -0.00000962, -0.00000432, 0.00000635, 0.00000273, -0.00000186, 0.00000094, -0.00000133, 0.00000339, -0.00000924, 0.00000003, 0.00000759, 0.00000618, -0.00001588, -0.00000343, 0.00002186, 0.00000154, -0.00001895, -0.00000216, 0.00000080, -0.00000935, 0.00001236, -0.00001034, 0.00000882, 0.00000305, -0.00001270, -0.00001048, 0.00000936, 0.00000349, -0.00001432, -0.00000523, 0.00000606, -0.00000113, -0.00000448, -0.00000620, 0.00000226, -0.00000804, -0.00002624, -0.00002490, -0.00008518, -0.00018367, 0.00057910, 0.00009511, -0.00022273, 0.00001306, -0.00004385, 0.00001177, -0.00003351, 0.00000067, -0.00000678, 0.00000332, -0.00000609, -0.00000549, 0.00000665, 0.00000257, -0.00001091, -0.00000335, 0.00000886, 0.00000028, -0.00000760, 0.00000004, -0.00000202, -0.00000329, 0.00000791, -0.00000956, 0.00001253, 0.00000132, -0.00000863, -0.00000436, 0.00001271, 0.00000558, -0.00001651, 0.00000564, -0.00000199, 0.00000393, -0.00000586, -0.00000249, 0.00000813, -0.00000218, -0.00000340, -0.00000468, 0.00000775, 0.00000335, -0.00000306, 0.00000020, -0.00000684, -0.00001031, 0.00001486, -0.00001649, },
		{ 0.00000925, -0.00002697, 0.00000868, -0.00000713, 0.00003007, 0.00000339, 0.00000329, 0.00001883, -0.00000163, 0.00000377, -0.00000972, -0.00000836, 0.00001003, -0.00001233, 0.00000752, -0.00000674, 0.00000560, -0.00001105, 0.00002722, -0.00002694, 0.00002055, -0.00001917, 0.00000462, -0.00000064, -0.00001034, 0.00000420, 0.00000443, -0.00000992, -0.00001240, 0.00001063, -0.00001466, 0.00001526, -0.00001239, 0.00001201, -0.00000162, 0.00000056, 0.00000527, -0.00000338, 0.00000357, -0.00000362, -0.00001013, 0.00000401, 0.00000473, -0.00000002, -0.00000122, 0.00000076, 0.00000058, -0.00000166, -0.00000909, 0.00000899, 0.00000536, -0.00000311, -0.00001906, 0.00002067, 0.00000885, -0.00000878, -0.00001811, 0.00001188, -0.00001727, 0.00001354, -0.00000692, 0.00000258, 0.00000543, -0.00000160, -0.00002082, 0.00000776, 0.00000303, 0.00000023, -0.00002260, 0.00001531, -0.00001079, 0.00001231, -0.00002476, 0.00001991, -0.00002628, 0.00001214, -0.00007033, 0.00002336, -0.00028125, 0.00028929, 0.00048596, -0.00027960, -0.00001881, -0.00004516, 0.00000856, -0.00002700, -0.00000778, -0.00000801, 0.00000438, -0.00000557, -0.00000326, -0.00000409, 0.00001322, -0.00001191, -0.00000416, 0.00000010, 0.00001099, -0.00001076, 0.00000021, -0.00000232, -0.00000316, 0.00000382, -0.00000028, -0.00000359, 0.00001656, -0.00001328, -0.00000171, -0.00000017, 0.00001993, -0.00001817, 0.00000104, -0.00000033, 0.00000292, -0.00000207, -0.00000573, 0.00000488, 0.00000397, -0.00000471, -0.00000388, 0.00000058, 0.00000921, -0.00000295, -0.00000087, -0.00000128, -0.00001483, 0.00000931, -0.00000353, -0.00000256, },
		{ -0.00002429, 0.00000114, -0.00000554, 0.00001589, 0.00001903, -0.00000433, 0.00001512, 0.00000942, -0.00000510, 0.00000583, -0.00002120, 0.00001357, -0.00001043, 0.00000397, -0.00000499, 0.00000642, -0.00001230, 0.00001823, -0.00000459, -0.00000681, 0.00001124, -0.00002262, 0.00001889, -0.00002023, 0.00000691, -0.00000185, 0.00000581, -0.00002523, 0.00001421, -0.00001421, 0.00000950, -0.00000315, 0.00000013, 0.00000911, -0.00000668, 0.00000839, -0.00000258, 0.00000068, 0.00000219, -0.00001059, -0.00000187, 0.00000564, 0.00000282, -0.00000276, 0.00000184, -0.00000125, 0.00000271, -0.00001061, 0.00000363, 0.00000650, 0.00000460, -0.00002015, 0.00000625, 0.00001765, -0.00000007, -0.00002164, 0.00000494, -0.00000643, -0.00000200, 0.00000809, -0.00000879, 0.00000875, 0.00000270, -0.00001537, -0.00000657, 0.00000810, 0.00000454, -0.00001885, 0.00000118, 0.00000068, 0.00000392, -0.00001078, 0.00000044, -0.00000481, -0.00000274, -0.00003927, -0.00000833, -0.00013143, -0.00013305, 0.00068867, -0.00007890, -0.00016648, 0.00000511, -0.00003431, 0.00000547, -0.00003035, 0.00000064, -0.00000489, 0.00000204, -0.00000512, -0.00000721, 0.00001016, -0.00000144, -0.00000917, -0.00000380, 0.00001055, -0.00000284, -0.00000579, 0.00000000, -0.00000276, -0.00000196, 0.00000844, -0.00001235, 0.00001722, -0.00000477, -0.00000578, -0.00000487, 0.00001861, -0.00000439, -0.00001117, 0.00000434, -0.00000043, 0.00000193, -0.00000606, -0.00000040, 0.00000778, -0.00000328, -0.00000282, -0.00000493, 0.00000949, 0.00000108, -0.00000223, -0.00000009, -0.00000915, -0.00000609, 0.00001289, -0.00001676, 0.00001981, },
		{ -0.00001774, 0.00000658, -0.00000456, 0.00003056, -0.00000374, 0.00000771, 0.00001486, -0.00000256, 0.00000331, -0.00000897, -0.00000684, 0.00000974, -0.00001302, 0.00000855, -0.00000674, 0.00000479, -0.00000861, 0.00002566, -0.00002947, 0.00002584, -0.00002587, 0.00001248, -0.00000846, -0.00000382, 0.00000049, 0.00000808, -0.00001805, -0.00000128, 0.00000151, -0.00000776, 0.00001259, -0.00001282, 0.00001643, -0.00000839, 0.00000627, 0.00000110, -0.00000187, 0.00000312, -0.00000449, -0.00000938, 0.00000552, 0.00000349, 0.00000024, -0.00000186, 0.00000212, -0.00000030, -0.00000302, -0.00000731, 0.00001039, 0.00000390, -0.00000779, -0.00001413, 0.00002402, 0.00000345, -0.00001004, -0.00001527, 0.00001217, -0.00001964, 0.00001976, -0.00001448, 0.00000948, 0.00000013, 0.00000089, -0.00002423, 0.00001393, -0.00000167, 0.00000218, -0.00002564, 0.00002127, -0.00001756, 0.00001912, -0.00003330, 0.00003003, -0.00003886, 0.00002089, -0.00008367, 0.00003377, -0.00033863, 0.00047289, 0.00036800, -0.00032761, 0.00003566, -0.00007085, 0.00002909, -0.00004424, 0.00000659, -0.00001672, 0.00001206, -0.00001154, 0.00000053, -0.00000625, 0.00001602, -0.00001662, -0.00000018, -0.00000167, 0.00001221, -0.00001373, 0.00000372, -0.00000483, -0.00000196, 0.00000563, -0.00000381, 0.00000003, 0.00001402, -0.00001446, -0.00000085, 0.00000451, 0.00001574, -0.00002117, 0.00000602, -0.00000301, 0.00000504, -0.00000576, -0.00000242, 0.00000465, 0.00000331, -0.00000487, -0.00000404, 0.00000205, 0.00000836, -0.00000375, -0.00000010, -0.00000306, -0.00001441, 0.00001342, -0.00001022, 0.00000707, -0.00000227, },
		{ 0.00001094, -0.00001337, 0.00002616, 0.00000679, 0.00000118, 0.00001124, 0.00000984, -0.00000876, 0.00000948, -0.00002364, 0.00001822, -0.00001543, 0.00000842, -0.00000823, 0.00000959, -0.00001567, 0.00002442, -0.00001422, 0.00000276, 0.00000310, -0.00001596, 0.00001443, -0.00001809, 0.00000660, 0.00000022, 0.00000044, -0.00002112, 0.00001534, -0.00001876, 0.00001784, -0.00001267, 0.00001051, 0.00000053, -0.00000218, 0.00000661, -0.00000310, 0.00000107, 0.00000192, -0.00001179, 0.00000114, 0.00000393, 0.00000385, -0.00000459, 0.00000400, -0.00000243, 0.00000252, -0.00001180, 0.00000761, 0.00000435, 0.00000267, -0.00002264, 0.00001619, 0.00001055, 0.00000052, -0.00002475, 0.00001278, -0.00001574, 0.00000911, -0.00000134, -0.00000201, 0.00000391, 0.00000653, -0.00002194, 0.00000175, 0.00000220, 0.00000982, -0.00002768, 0.00001287, -0.00001099, 0.00001609, -0.00002527, 0.00001646, -0.00002272, 0.00001410, -0.00006535, 0.00002504, -0.00020795, -0.00002234, 0.00074229, -0.00025359, -0.00007601, -0.00002988, -0.00000678, -0.00001520, -0.00001530, -0.00000858, 0.00000417, -0.00000515, 0.00000053, -0.00001279, 0.00001724, -0.00000917, -0.00000374, -0.00000739, 0.00001480, -0.00000841, -0.00000133, -0.00000242, -0.00000186, -0.00000096, 0.00000819, -0.00001430, 0.00002140, -0.00001121, -0.00000206, -0.00000478, 0.00002284, -0.00001475, -0.00000316, 0.00000043, 0.00000315, -0.00000239, -0.00000368, -0.00000002, 0.00000804, -0.00000471, -0.00000197, -0.00000522, 0.00001119, -0.00000150, -0.00000089, -0.00000102, -0.00001099, -0.00000083, 0.00000802, -0.00001247, 0.00001808, -0.00002814, },
		{ -0.00000204, 0.00000433, 0.00002406, -0.00000477, 0.00000803, 0.00001414, -0.00000638, 0.00000622, -0.00001120, -0.00000280, 0.00000684, -0.00001133, 0.00000768, -0.00000508, 0.00000223, -0.00000394, 0.00002079, -0.00002809, 0.00002752, -0.00002945, 0.00001773, -0.00001444, 0.00000172, -0.00000227, 0.00000984, -0.00002388, 0.00000933, -0.00000979, 0.00000368, 0.00000395, -0.00000696, 0.00001504, -0.00001117, 0.00001005, -0.00000265, -0.00000024, 0.00000275, -0.00000574, -0.00000766, 0.00000578, 0.00000328, -0.00000054, -0.00000130, 0.00000258, -0.00000098, -0.00000458, -0.00000437, 0.00001032, 0.00000266, -0.00001292, -0.00000574, 0.00002262, 0.00000078, -0.00001297, -0.00000928, 0.00000776, -0.00001641, 0.00002043, -0.00001772, 0.00001309, -0.00000241, 0.00000048, -0.00002417, 0.00001664, -0.00000317, 0.00000029, -0.00002371, 0.00002162, -0.00001837, 0.00001943, -0.00003486, 0.00003271, -0.00004403, 0.00002088, -0.00008499, 0.00002470, -0.00037255, 0.00067168, 0.00017956, -0.00030789, 0.00005646, -0.00007875, 0.00003672, -0.00005200, 0.00001439, -0.00002045, 0.00001549, -0.00001395, 0.00000092, -0.00000487, 0.00001515, -0.00001801, 0.00000119, -0.00000123, 0.00001118, -0.00001451, 0.00000568, -0.00000665, -0.00000067, 0.00000738, -0.00000830, 0.00000554, 0.00000885, -0.00001314, -0.00000114, 0.00001073, 0.00000734, -0.00001905, 0.00000747, -0.00000321, 0.00000464, -0.00000735, 0.00000007, 0.00000454, 0.00000227, -0.00000446, -0.00000475, 0.00000424, 0.00000648, -0.00000364, 0.00000011, -0.00000502, -0.00001232, 0.00001506, -0.00001459, 0.00001529, -0.00001469, -0.00000403, },
		{ -0.00001614, 0.00003235, -0.00000362, 0.00000674, 0.00000730, 0.00001001, -0.00001168, 0.00001301, -0.00002626, 0.00002289, -0.00002076, 0.00001346, -0.00001189, 0.00001274, -0.00001849, 0.00002958, -0.00002335, 0.00001298, -0.00000647, -0.00000727, 0.00000739, -0.00001307, 0.00000389, 0.00000413, -0.00000756, -0.00001201, 0.00000976, -0.00001635, 0.00002033, -0.00001834, 0.00001929, -0.00000895, 0.00000471, 0.00000220, -0.00000172, 0.00000052, 0.00000202, -0.00001291, 0.00000416, 0.00000184, 0.00000524, -0.00000679, 0.00000694, -0.00000470, 0.00000284, -0.00001257, 0.00001139, 0.00000154, 0.00000014, -0.00002221, 0.00002413, 0.00000265, 0.00000194, -0.00002692, 0.00001923, -0.00002436, 0.00002104, -0.00001334, 0.00000806, -0.00000416, 0.00001314, -0.00003099, 0.00001297, -0.00000685, 0.00001771, -0.00003885, 0.00002753, -0.00002644, 0.00003275, -0.00004523, 0.00003917, -0.00004897, 0.00004007, -0.00010162, 0.00007130, -0.00031176, 0.00015718, 0.00071027, -0.00039107, 0.00002193, -0.00007786, 0.00003040, -0.00004440, 0.00000759, -0.00002416, 0.00001817, -0.00001641, 0.00000919, -0.00002055, 0.00002618, -0.00001898, 0.00000396, -0.00001301, 0.00002065, -0.00001554, 0.00000504, -0.00000688, 0.00000068, -0.00000045, 0.00000692, -0.00001460, 0.00002391, -0.00001672, 0.00000159, -0.00000327, 0.00002370, -0.00002288, 0.00000509, -0.00000431, 0.00000726, -0.00000770, 0.00000047, -0.00000109, 0.00000883, -0.00000627, -0.00000115, -0.00000519, 0.00001240, -0.00000393, 0.00000067, -0.00000257, -0.00001181, 0.00000431, 0.00000154, -0.00000456, 0.00001031, -0.00002404, 0.00001347, },
		{ 0.00001690, 0.00001224, -0.00000032, 0.00000427, 0.00001671, -0.00001318, 0.00001278, -0.00001696, 0.00000443, 0.00000063, -0.00000650, 0.00000419, -0.00000125, -0.00000235, 0.00000300, 0.00001257, -0.00002247, 0.00002489, -0.00002889, 0.00001907, -0.00001702, 0.00000467, -0.00000263, 0.00000811, -0.00002493, 0.00001579, -0.00001909, 0.00001586, -0.00000790, 0.00000380, 0.00000786, -0.00000884, 0.00001026, -0.00000446, 0.00000048, 0.00000308, -0.00000772, -0.00000473, 0.00000447, 0.00000453, -0.00000286, 0.00000112, 0.00000131, -0.00000072, -0.00000666, -0.00000015, 0.00000849, 0.00000180, -0.00001779, 0.00000509, 0.00001649, 0.00000165, -0.00001805, -0.00000016, -0.00000140, -0.00000687, 0.00001393, -0.00001430, 0.00001085, 0.00000037, -0.00000532, -0.00001808, 0.00001317, 0.00000130, -0.00000819, -0.00001383, 0.00001291, -0.00000915, 0.00000846, -0.00002378, 0.00002120, -0.00003391, 0.00000309, -0.00006312, -0.00002066, -0.00035402, 0.00084805, -0.00005383, -0.00022332, 0.00003520, -0.00006014, 0.00002405, -0.00004372, 0.00000998, -0.00001453, 0.00001081, -0.00000966, -0.00000464, 0.00000228, 0.00000850, -0.00001402, -0.00000206, 0.00000326, 0.00000627, -0.00001150, 0.00000449, -0.00000634, -0.00000022, 0.00000921, -0.00001339, 0.00001246, 0.00000136, -0.00000936, -0.00000231, 0.00001737, -0.00000404, -0.00001210, 0.00000488, -0.00000030, 0.00000100, -0.00000577, 0.00000056, 0.00000541, 0.00000040, -0.00000326, -0.00000608, 0.00000713, 0.00000353, -0.00000247, -0.00000053, -0.00000672, -0.00000881, 0.00001370, -0.00001498, 0.00001912, -0.00002372, 0.00000939, -0.00001335, },
		{ 0.00003142, -0.00000871, 0.00000908, 0.00000614, 0.00000730, -0.00001122, 0.00001376, -0.00002662, 0.00002538, -0.00002435, 0.00001728, -0.00001442, 0.00001449, -0.00001939, 0.00003208, -0.00002995, 0.00002157, -0.00001519, 0.00000118, -0.00000008, -0.00000707, 0.00000046, 0.00000796, -0.00001572, -0.00000046, -0.00000086, -0.00000713, 0.00001552, -0.00001751, 0.00002315, -0.00001604, 0.00001134, -0.00000312, 0.00000064, -0.00000039, 0.00000198, -0.00001333, 0.00000639, 0.00000028, 0.00000603, -0.00000831, 0.00000961, -0.00000732, 0.00000326, -0.00001246, 0.00001422, -0.00000135, -0.00000314, -0.00001807, 0.00002794, -0.00000352, 0.00000230, -0.00002635, 0.00002188, -0.00002901, 0.00002979, -0.00002387, 0.00001769, -0.00001205, 0.00001925, -0.00003913, 0.00002355, -0.00001547, 0.00002447, -0.00004813, 0.00004033, -0.00004026, 0.00004787, -0.00006402, 0.00006129, -0.00007571, 0.00006660, -0.00013785, 0.00011694, -0.00042620, 0.00039700, 0.00057724, -0.00045835, 0.00009729, -0.00011871, 0.00006318, -0.00007132, 0.00002990, -0.00003955, 0.00003180, -0.00002726, 0.00001696, -0.00002691, 0.00003356, -0.00002764, 0.00001101, -0.00001807, 0.00002579, -0.00002210, 0.00001146, -0.00001192, 0.00000406, -0.00000008, 0.00000428, -0.00001255, 0.00002365, -0.00002002, 0.00000422, 0.00000019, 0.00002004, -0.00002644, 0.00001097, -0.00000770, 0.00000988, -0.00001200, 0.00000477, -0.00000262, 0.00000956, -0.00000747, -0.00000081, -0.00000439, 0.00001259, -0.00000564, 0.00000195, -0.00000451, -0.00001120, 0.00000802, -0.00000449, 0.00000452, -0.00000127, -0.00001299, 0.00000775, -0.00001150, },
		{ -0.00000108, 0.00000683, -0.00000148, 0.00002070, -0.00002119, 0.00002140, -0.00002510, 0.00001397, -0.00000816, 0.00000096, -0.00000153, 0.00000430, -0.00000829, 0.00001126, 0.00000213, -0.00001355, 0.00001858, -0.00002456, 0.00001658, -0.00001590, 0.00000460, -0.00000033, 0.00000266, -0.00002020, 0.00001569, -0.00002268, 0.00002444, -0.00001890, 0.00001618, -0.00000310, -0.00000205, 0.00000659, -0.00000367, -0.00000023, 0.00000437, -0.00001046, -0.00000076, 0.00000177, 0.00000713, -0.00000666, 0.00000546, -0.00000207, 0.00000094, -0.00000937, 0.00000514, 0.00000492, 0.00000128, -0.00002135, 0.00001647, 0.00000699, 0.00000553, -0.00002462, 0.00001078, -0.00001369, 0.00000754, 0.00000092, -0.00000412, 0.00000224, 0.00000910, -0.00001712, -0.00000534, 0.00000279, 0.00001242, -0.00002377, 0.00000450, -0.00000554, 0.00001102, -0.00001518, 0.00000198, -0.00000741, -0.00000467, -0.00003713, -0.00001197, -0.00011366, -0.00025650, 0.00095819, -0.00029097, -0.00009493, -0.00002210, -0.00001634, -0.00000891, -0.00001855, -0.00000790, 0.00000221, -0.00000295, 0.00000205, -0.00001657, 0.00001541, -0.00000410, -0.00000435, -0.00001035, 0.00001223, -0.00000298, -0.00000417, -0.00000059, -0.00000305, -0.00000127, 0.00001113, -0.00001840, 0.00001980, -0.00000747, -0.00000377, -0.00000364, 0.00002278, -0.00001606, -0.00000204, -0.00000085, 0.00000509, -0.00000541, -0.00000102, -0.00000139, 0.00000764, -0.00000242, -0.00000131, -0.00000787, 0.00001044, -0.00000021, -0.00000033, -0.00000219, -0.00000755, -0.00000474, 0.00000989, -0.00001114, 0.00001695, -0.00002601, 0.00001772, -0.00002233, 0.00002870, },
		{ -0.00000705, 0.00000653, 0.00000926, 0.00000022, -0.00000572, 0.00000981, -0.00002279, 0.00002383, -0.00002439, 0.00001823, -0.00001447, 0.00001372, -0.00001736, 0.00003081, -0.00003251, 0.00002672, -0.00002110, 0.00000726, -0.00000573, -0.00000223, -0.00000174, 0.00000958, -0.00002107, 0.00000980, -0.00001286, 0.00000625, 0.00000449, -0.00000971, 0.00002034, -0.00001816, 0.00001513, -0.00000731, 0.00000263, -0.00000074, 0.00000108, -0.00001237, 0.00000699, 0.00000023, 0.00000517, -0.00000795, 0.00001069, -0.00000915, 0.00000309, -0.00001087, 0.00001525, -0.00000366, -0.00000711, -0.00001016, 0.00002628, -0.00000589, -0.00000013, -0.00002162, 0.00001887, -0.00002697, 0.00003171, -0.00002871, 0.00002265, -0.00001562, 0.00002075, -0.00004213, 0.00002907, -0.00001909, 0.00002536, -0.00005029, 0.00004531, -0.00004571, 0.00005376, -0.00007285, 0.00007278, -0.00009159, 0.00008079, -0.00015836, 0.00014053, -0.00052131, 0.00066845, 0.00034970, -0.00043752, 0.00012612, -0.00013347, 0.00007694, -0.00008398, 0.00004179, -0.00004676, 0.00003843, -0.00003216, 0.00001908, -0.00002758, 0.00003536, -0.00003130, 0.00001381, -0.00001930, 0.00002728, -0.00002535, 0.00001535, -0.00001540, 0.00000681, 0.00000082, -0.00000005, -0.00000770, 0.00001991, -0.00002014, 0.00000521, 0.00000549, 0.00001174, -0.00002417, 0.00001255, -0.00000791, 0.00000923, -0.00001335, 0.00000738, -0.00000332, 0.00000938, -0.00000769, -0.00000143, -0.00000240, 0.00001129, -0.00000608, 0.00000239, -0.00000632, -0.00000922, 0.00000940, -0.00000804, 0.00001167, -0.00001282, 0.00000142, -0.00000405, -0.00000007, 0.00002290, },
		{ 0.00001285, -0.00000601, 0.00002310, -0.00002740, 0.00002918, -0.00003308, 0.00002358, -0.00001755, 0.00000937, -0.00000806, 0.00001022, -0.00001418, 0.00001916, -0.00000859, -0.00000331, 0.00001040, -0.00001814, 0.00001172, -0.00001224, 0.00000245, 0.00000352, -0.00000507, -0.00001078, 0.00000882, -0.00001870, 0.00002614, -0.00002504, 0.00002620, -0.00001448, 0.00000696, 0.00000016, -0.00000067, -0.00000217, 0.00000632, -0.00001347, 0.00000356, -0.00000156, 0.00001033, -0.00001112, 0.00001102, -0.00000719, 0.00000391, -0.00001245, 0.00001093, 0.00000004, 0.00000085, -0.00002247, 0.00002598, -0.00000339, 0.00001065, -0.00003092, 0.00002110, -0.00002598, 0.00002339, -0.00001545, 0.00001029, -0.00001057, 0.00002175, -0.00003283, 0.00001195, -0.00001239, 0.00002793, -0.00004379, 0.00002823, -0.00003043, 0.00003866, -0.00004785, 0.00003875, -0.00004962, 0.00004037, -0.00009622, 0.00006450, -0.00025316, -0.00006588, 0.00096347, -0.00048439, 0.00004314, -0.00009650, 0.00004097, -0.00005386, 0.00001752, -0.00003488, 0.00002632, -0.00002289, 0.00001857, -0.00003234, 0.00003207, -0.00002037, 0.00000898, -0.00002196, 0.00002419, -0.00001522, 0.00000641, -0.00000887, 0.00000302, -0.00000377, 0.00001273, -0.00002234, 0.00002617, -0.00001614, 0.00000252, -0.00000424, 0.00002517, -0.00002592, 0.00000838, -0.00000763, 0.00001115, -0.00001292, 0.00000573, -0.00000520, 0.00001090, -0.00000584, 0.00000097, -0.00000961, 0.00001356, -0.00000416, 0.00000236, -0.00000476, -0.00000703, -0.00000128, 0.00000514, -0.00000439, 0.00000934, -0.00002046, 0.00001788, -0.00002477, 0.00003520, -0.00001144, },
		{ -0.00000026, 0.00001607, -0.00001074, 0.00000464, 0.00000091, -0.00001428, 0.00001764, -0.00002015, 0.00001558, -0.00001150, 0.00001011, -0.00001233, 0.00002567, -0.00003066, 0.00002773, -0.00002328, 0.00000987, -0.00000820, 0.00000006, -0.00000145, 0.00000757, -0.00002132, 0.00001521, -0.00002189, 0.00001952, -0.00000945, 0.00000315, 0.00001122, -0.00001441, 0.00001447, -0.00000874, 0.00000303, 0.00000026, -0.00000126, -0.00000958, 0.00000546, 0.00000230, 0.00000196, -0.00000480, 0.00000899, -0.00000903, 0.00000154, -0.00000730, 0.00001382, -0.00000489, -0.00001135, 0.00000066, 0.00001908, -0.00000347, -0.00000615, -0.00001238, 0.00000979, -0.00001725, 0.00002481, -0.00002505, 0.00001981, -0.00001169, 0.00001441, -0.00003660, 0.00002590, -0.00001400, 0.00001654, -0.00004112, 0.00003768, -0.00003725, 0.00004393, -0.00006401, 0.00006441, -0.00008568, 0.00006974, -0.00014690, 0.00011831, -0.00055935, 0.00092558, 0.00005804, -0.00033213, 0.00009845, -0.00011129, 0.00006230, -0.00007400, 0.00003590, -0.00003958, 0.00003286, -0.00002672, 0.00001188, -0.00001928, 0.00002851, -0.00002703, 0.00000954, -0.00001406, 0.00002265, -0.00002288, 0.00001440, -0.00001519, 0.00000732, 0.00000297, -0.00000614, -0.00000014, 0.00001261, -0.00001670, 0.00000444, 0.00001182, -0.00000006, -0.00001629, 0.00000920, -0.00000416, 0.00000447, -0.00001051, 0.00000689, -0.00000203, 0.00000755, -0.00000644, -0.00000332, 0.00000098, 0.00000826, -0.00000486, 0.00000144, -0.00000728, -0.00000649, 0.00000835, -0.00000793, 0.00001430, -0.00002027, 0.00001423, -0.00001743, 0.00001469, 0.00000917, 0.00000010, },
		{ -0.00000646, 0.00002109, -0.00002883, 0.00003303, -0.00003795, 0.00003058, -0.00002503, 0.00001654, -0.00001356, 0.00001487, -0.00001841, 0.00002489, -0.00001743, 0.00000597, 0.00000260, -0.00001185, 0.00000672, -0.00000807, 0.00000009, 0.00000682, -0.00001236, 0.00000033, -0.00000250, -0.00000794, 0.00001995, -0.00002386, 0.00003050, -0.00002272, 0.00001517, -0.00000691, 0.00000326, -0.00000444, 0.00000811, -0.00001589, 0.00000720, -0.00000433, 0.00001283, -0.00001486, 0.00001636, -0.00001290, 0.00000743, -0.00001518, 0.00001623, -0.00000535, 0.00000018, -0.00002027, 0.00003132, -0.00001179, 0.00001458, -0.00003465, 0.00002789, -0.00003448, 0.00003612, -0.00003037, 0.00002436, -0.00002328, 0.00003404, -0.00004806, 0.00002918, -0.00002766, 0.00004287, -0.00006272, 0.00005114, -0.00005486, 0.00006605, -0.00008100, 0.00007709, -0.00009518, 0.00008999, -0.00016109, 0.00014990, -0.00042083, 0.00021115, 0.00084236, -0.00059276, 0.00015329, -0.00016214, 0.00009334, -0.00009642, 0.00005314, -0.00006197, 0.00005041, -0.00004276, 0.00003444, -0.00004691, 0.00004750, -0.00003585, 0.00002190, -0.00003321, 0.00003584, -0.00002745, 0.00001754, -0.00001822, 0.00001040, -0.00000690, 0.00001325, -0.00002407, 0.00003000, -0.00002294, 0.00000820, -0.00000334, 0.00002324, -0.00003106, 0.00001621, -0.00001290, 0.00001551, -0.00001914, 0.00001239, -0.00000943, 0.00001421, -0.00000906, 0.00000285, -0.00001059, 0.00001571, -0.00000748, 0.00000487, -0.00000772, -0.00000505, 0.00000050, 0.00000140, 0.00000269, -0.00000109, -0.00000880, 0.00000974, -0.00001890, 0.00003363, -0.00001547, 0.00001450, },
		{ 0.00002410, -0.00002314, 0.00001768, -0.00001128, -0.00000236, 0.00000781, -0.00001234, 0.00000986, -0.00000608, 0.00000442, -0.00000527, 0.00001780, -0.00002537, 0.00002533, -0.00002231, 0.00000939, -0.00000759, -0.00000026, 0.00000125, 0.00000211, -0.00001593, 0.00001379, -0.00002447, 0.00002824, -0.00002187, 0.00001739, -0.00000174, -0.00000582, 0.00000929, -0.00000683, 0.00000130, 0.00000292, -0.00000511, -0.00000506, 0.00000191, 0.00000642, -0.00000359, 0.00000131, 0.00000398, -0.00000623, -0.00000184, -0.00000159, 0.00000964, -0.00000474, -0.00001513, 0.00001262, 0.00000776, 0.00000318, -0.00001522, 0.00000026, -0.00000388, -0.00000117, 0.00000974, -0.00001274, 0.00000855, 0.00000053, -0.00000062, -0.00002156, 0.00001291, 0.00000096, -0.00000301, -0.00001955, 0.00001607, -0.00001313, 0.00001604, -0.00003430, 0.00003178, -0.00005213, 0.00002599, -0.00009330, 0.00003262, -0.00050469, 0.00111433, -0.00024843, -0.00016754, 0.00002190, -0.00005393, 0.00001926, -0.00004030, 0.00001060, -0.00001639, 0.00001365, -0.00000983, -0.00000543, -0.00000147, 0.00001249, -0.00001417, -0.00000253, -0.00000158, 0.00001108, -0.00001378, 0.00000748, -0.00001002, 0.00000456, 0.00000674, -0.00001360, 0.00000935, 0.00000249, -0.00001008, 0.00000230, 0.00001774, -0.00001311, -0.00000459, 0.00000188, 0.00000284, -0.00000387, -0.00000356, 0.00000288, 0.00000171, 0.00000381, -0.00000364, -0.00000640, 0.00000553, 0.00000370, -0.00000199, -0.00000115, -0.00000667, -0.00000408, 0.00000582, -0.00000442, 0.00001145, -0.00002087, 0.00002094, -0.00002720, 0.00002786, -0.00000543, 0.00000943, -0.00000865, },
		{ 0.00001339, -0.00002385, 0.00003091, -0.00003760, 0.00003287, -0.00002856, 0.00002059, -0.00001646, 0.00001698, -0.00001984, 0.00002727, -0.00002291, 0.00001258, -0.00000306, -0.00000758, 0.00000361, -0.00000539, -0.00000061, 0.00000747, -0.00001626, 0.00000927, -0.00001427, 0.00000636, 0.00000753, -0.00001525, 0.00002747, -0.00002527, 0.00001982, -0.00001229, 0.00000642, -0.00000580, 0.00000871, -0.00001672, 0.00000901, -0.00000527, 0.00001324, -0.00001623, 0.00001964, -0.00001749, 0.00001027, -0.00001648, 0.00001979, -0.00001020, -0.00000102, -0.00001444, 0.00003092, -0.00001571, 0.00001504, -0.00003376, 0.00002873, -0.00003593, 0.00004142, -0.00003888, 0.00003297, -0.00003078, 0.00004078, -0.00005735, 0.00004064, -0.00003711, 0.00005101, -0.00007369, 0.00006551, -0.00007014, 0.00008335, -0.00010336, 0.00010410, -0.00012939, 0.00012733, -0.00021140, 0.00021742, -0.00058162, 0.00054353, 0.00059920, -0.00059263, 0.00020544, -0.00019484, 0.00012190, -0.00012106, 0.00007537, -0.00007846, 0.00006550, -0.00005493, 0.00004301, -0.00005428, 0.00005611, -0.00004527, 0.00002949, -0.00003957, 0.00004298, -0.00003575, 0.00002565, -0.00002550, 0.00001673, -0.00000919, 0.00001170, -0.00002254, 0.00002996, -0.00002628, 0.00001211, -0.00000063, 0.00001662, -0.00002998, 0.00001921, -0.00001449, 0.00001604, -0.00002171, 0.00001666, -0.00001228, 0.00001627, -0.00001104, 0.00000348, -0.00001006, 0.00001607, -0.00000928, 0.00000628, -0.00001021, -0.00000210, 0.00000015, 0.00000029, 0.00000725, -0.00001048, 0.00000496, -0.00000376, -0.00000613, 0.00002408, -0.00001216, 0.00001046, -0.00000161, },
		{ -0.00003348, 0.00002998, -0.00002375, 0.00001035, -0.00000336, -0.00000287, 0.00000265, 0.00000036, -0.00000188, 0.00000212, 0.00000913, -0.00001856, 0.00002129, -0.00001992, 0.00000749, -0.00000537, -0.00000189, 0.00000486, -0.00000495, -0.00000657, 0.00000601, -0.00001932, 0.00002946, -0.00002878, 0.00002883, -0.00001491, 0.00000505, 0.00000101, -0.00000218, -0.00000219, 0.00000684, -0.00000995, 0.00000049, -0.00000282, 0.00001173, -0.00001054, 0.00000954, -0.00000389, -0.00000081, -0.00000689, 0.00000581, 0.00000299, -0.00000328, -0.00001754, 0.00002343, -0.00000514, 0.00001213, -0.00002544, 0.00001381, -0.00001897, 0.00001766, -0.00001011, 0.00000541, -0.00000869, 0.00001873, -0.00002200, 0.00000064, -0.00000760, 0.00002329, -0.00003046, 0.00001126, -0.00001611, 0.00002300, -0.00002614, 0.00001255, -0.00002170, 0.00000612, -0.00004781, 0.00000043, -0.00011961, -0.00033553, 0.00118579, -0.00051250, 0.00001531, -0.00008047, 0.00002424, -0.00004195, 0.00000968, -0.00002861, 0.00001848, -0.00001559, 0.00001532, -0.00002981, 0.00002297, -0.00001005, 0.00000487, -0.00002033, 0.00001636, -0.00000589, 0.00000072, -0.00000462, -0.00000020, -0.00000147, 0.00001178, -0.00002145, 0.00001934, -0.00000891, -0.00000138, -0.00000039, 0.00002160, -0.00002457, 0.00000804, -0.00000717, 0.00001114, -0.00001388, 0.00000607, -0.00000386, 0.00000742, -0.00000140, 0.00000022, -0.00001012, 0.00001059, -0.00000175, 0.00000204, -0.00000519, -0.00000408, -0.00000318, 0.00000350, 0.00000074, 0.00000427, -0.00001430, 0.00001923, -0.00002951, 0.00003489, -0.00001627, 0.00001684, -0.00001511, 0.00002991, },
		{ -0.00001290, 0.00002269, -0.00003157, 0.00002983, -0.00002735, 0.00002074, -0.00001617, 0.00001618, -0.00001837, 0.00002622, -0.00002479, 0.00001607, -0.00000604, -0.00000596, 0.00000327, -0.00000520, 0.00000127, 0.00000443, -0.00001484, 0.00001276, -0.00002214, 0.00001966, -0.00000737, -0.00000154, 0.00001782, -0.00002139, 0.00001931, -0.00001418, 0.00000736, -0.00000518, 0.00000724, -0.00001520, 0.00000822, -0.00000348, 0.00001054, -0.00001395, 0.00001921, -0.00001924, 0.00001102, -0.00001527, 0.00002038, -0.00001343, -0.00000285, -0.00000548, 0.00002445, -0.00001382, 0.00001069, -0.00002730, 0.00002262, -0.00002880, 0.00003678, -0.00003755, 0.00003227, -0.00002902, 0.00003776, -0.00005621, 0.00004152, -0.00003572, 0.00004707, -0.00007093, 0.00006481, -0.00006884, 0.00008190, -0.00010473, 0.00010761, -0.00013774, 0.00013510, -0.00022540, 0.00023637, -0.00069000, 0.00087911, 0.00026651, -0.00048572, 0.00018616, -0.00018030, 0.00011421, -0.00011665, 0.00007438, -0.00007591, 0.00006436, -0.00005320, 0.00003897, -0.00004968, 0.00005344, -0.00004440, 0.00002772, -0.00003721, 0.00004200, -0.00003664, 0.00002737, -0.00002760, 0.00001945, -0.00000902, 0.00000727, -0.00001715, 0.00002534, -0.00002516, 0.00001349, 0.00000358, 0.00000614, -0.00002267, 0.00001649, -0.00001136, 0.00001163, -0.00001914, 0.00001679, -0.00001219, 0.00001592, -0.00001091, 0.00000221, -0.00000750, 0.00001410, -0.00000886, 0.00000573, -0.00001113, 0.00000075, -0.00000182, 0.00000232, 0.00000734, -0.00001522, 0.00001597, -0.00001774, 0.00000953, 0.00000947, -0.00000327, 0.00000181, 0.00000755, 0.00000085, },
		{ 0.00003815, -0.00003327, 0.00002091, -0.00001318, 0.00000590, -0.00000409, 0.00000610, -0.00000718, 0.00000813, 0.00000159, -0.00001228, 0.00001762, -0.00001816, 0.00000634, -0.00000366, -0.00000280, 0.00000711, -0.00001065, 0.00000339, -0.00000527, -0.00000791, 0.00002272, -0.00002801, 0.00003419, -0.00002463, 0.00001493, -0.00000792, 0.00000360, -0.00000630, 0.00001096, -0.00001469, 0.00000583, -0.00000737, 0.00001670, -0.00001720, 0.00001810, -0.00001311, 0.00000617, -0.00001270, 0.00001378, -0.00000520, -0.00000091, -0.00001771, 0.00003075, -0.00001650, 0.00002059, -0.00003418, 0.00002507, -0.00003144, 0.00003435, -0.00002948, 0.00002425, -0.00002694, 0.00003791, -0.00004456, 0.00002468, -0.00003013, 0.00004721, -0.00005939, 0.00004418, -0.00005099, 0.00006239, -0.00007288, 0.00006581, -0.00008437, 0.00007638, -0.00013724, 0.00011720, -0.00032124, -0.00005374, 0.00110997, -0.00068337, 0.00017098, -0.00017720, 0.00010066, -0.00010370, 0.00006188, -0.00007047, 0.00005572, -0.00004696, 0.00004181, -0.00005489, 0.00004812, -0.00003359, 0.00002505, -0.00003925, 0.00003556, -0.00002445, 0.00001727, -0.00001912, 0.00001223, -0.00000943, 0.00001694, -0.00002818, 0.00002794, -0.00001955, 0.00000782, -0.00000270, 0.00002204, -0.00003180, 0.00001843, -0.00001509, 0.00001806, -0.00002283, 0.00001589, -0.00001148, 0.00001379, -0.00000699, 0.00000417, -0.00001351, 0.00001512, -0.00000700, 0.00000624, -0.00000989, 0.00000026, -0.00000453, 0.00000314, 0.00000496, -0.00000428, -0.00000295, 0.00000980, -0.00002327, 0.00003341, -0.00002018, 0.00001918, -0.00001676, 0.00003059, -0.00003855, },
		{ 0.00001026, -0.00002124, 0.00002252, -0.00002215, 0.00001752, -0.00001323, 0.00001319, -0.00001492, 0.00002286, -0.00002408, 0.00001736, -0.00000725, -0.00000614, 0.00000503, -0.00000699, 0.00000512, -0.00000160, -0.00000820, 0.00000945, -0.00002312, 0.00002770, -0.00002022, 0.00001337, 0.00000424, -0.00001233, 0.00001372, -0.00001201, 0.00000544, -0.00000212, 0.00000343, -0.00001119, 0.00000473, 0.00000117, 0.00000454, -0.00000761, 0.00001421, -0.00001696, 0.00000869, -0.00001084, 0.00001718, -0.00001419, -0.00000515, 0.00000529, 0.00001304, -0.00000639, 0.00000169, -0.00001587, 0.00001053, -0.00001401, 0.00002249, -0.00002582, 0.00002117, -0.00001665, 0.00002345, -0.00004287, 0.00002981, -0.00002139, 0.00002896, -0.00005211, 0.00004634, -0.00004773, 0.00005770, -0.00007994, 0.00008086, -0.00011142, 0.00010200, -0.00018773, 0.00018188, -0.00070123, 0.00115488, -0.00010058, -0.00029929, 0.00010257, -0.00011911, 0.00006898, -0.00008087, 0.00004730, -0.00005150, 0.00004443, -0.00003540, 0.00002054, -0.00003157, 0.00003804, -0.00003178, 0.00001508, -0.00002460, 0.00003129, -0.00002845, 0.00002085, -0.00002256, 0.00001679, -0.00000533, -0.00000032, -0.00000810, 0.00001632, -0.00001941, 0.00001217, 0.00000838, -0.00000631, -0.00001074, 0.00000883, -0.00000402, 0.00000270, -0.00001136, 0.00001217, -0.00000845, 0.00001261, -0.00000834, -0.00000113, -0.00000288, 0.00000976, -0.00000602, 0.00000277, -0.00000956, 0.00000208, -0.00000394, 0.00000655, 0.00000280, -0.00001338, 0.00002034, -0.00002719, 0.00002295, -0.00000555, 0.00000754, -0.00000844, 0.00001816, -0.00001373, 0.00000156, },
		{ -0.00003760, 0.00002714, -0.00001954, 0.00001208, -0.00000880, 0.00000993, -0.00001047, 0.00001176, -0.00000364, -0.00000778, 0.00001559, -0.00001840, 0.00000758, -0.00000419, -0.00000129, 0.00000605, -0.00001217, 0.00001009, -0.00001576, 0.00000603, 0.00001021, -0.00001987, 0.00003211, -0.00002844, 0.00002096, -0.00001492, 0.00000852, -0.00000947, 0.00001394, -0.00001801, 0.00000954, -0.00001015, 0.00001958, -0.00002152, 0.00002468, -0.00002144, 0.00001292, -0.00001772, 0.00002065, -0.00001347, 0.00000173, -0.00001511, 0.00003282, -0.00002351, 0.00002576, -0.00003883, 0.00003115, -0.00003759, 0.00004414, -0.00004278, 0.00003794, -0.00004025, 0.00005193, -0.00006184, 0.00004373, -0.00004760, 0.00006520, -0.00008157, 0.00007004, -0.00007833, 0.00009346, -0.00011095, 0.00011034, -0.00013891, 0.00013883, -0.00021865, 0.00022681, -0.00053502, 0.00031070, 0.00088591, -0.00072996, 0.00026225, -0.00023839, 0.00015195, -0.00014670, 0.00010012, -0.00010145, 0.00008390, -0.00007064, 0.00006102, -0.00007291, 0.00006676, -0.00005140, 0.00004009, -0.00005343, 0.00005059, -0.00003955, 0.00003120, -0.00003187, 0.00002394, -0.00001697, 0.00002049, -0.00003208, 0.00003316, -0.00002725, 0.00001576, -0.00000387, 0.00001840, -0.00003305, 0.00002400, -0.00001936, 0.00002111, -0.00002797, 0.00002314, -0.00001768, 0.00001904, -0.00001152, 0.00000697, -0.00001547, 0.00001798, -0.00001081, 0.00000935, -0.00001396, 0.00000537, -0.00000800, 0.00000580, 0.00000582, -0.00001054, 0.00000898, -0.00000384, -0.00001053, 0.00002402, -0.00001655, 0.00001538, -0.00001257, 0.00002493, -0.00003671, 0.00004009, },
		{ -0.00000930, 0.00001324, -0.00001484, 0.00001245, -0.00000900, 0.00000936, -0.00001104, 0.00001891, -0.00002258, 0.00001817, -0.00000853, -0.00000612, 0.00000702, -0.00000900, 0.00000899, -0.00000826, 0.00000138, 0.00000056, -0.00001669, 0.00002807, -0.00002728, 0.00002535, -0.00000954, -0.00000084, 0.00000464, -0.00000651, 0.00000106, 0.00000300, -0.00000227, -0.00000532, -0.00000073, 0.00000785, -0.00000390, 0.00000201, 0.00000500, -0.00001054, 0.00000314, -0.00000327, 0.00001010, -0.00001211, -0.00000749, 0.00001598, -0.00000100, 0.00000480, -0.00001034, -0.00000160, -0.00000475, 0.00000516, 0.00000174, -0.00000637, 0.00000194, 0.00000419, -0.00000005, -0.00001937, 0.00000746, 0.00000383, -0.00000102, -0.00001979, 0.00001279, -0.00000967, 0.00001363, -0.00003167, 0.00002593, -0.00005161, 0.00002814, -0.00009655, 0.00004480, -0.00058465, 0.00131186, -0.00043662, -0.00007914, -0.00001991, -0.00002676, -0.00000289, -0.00002156, -0.00000016, -0.00000959, 0.00000926, -0.00000462, -0.00000936, -0.00000270, 0.00001242, -0.00000964, -0.00000647, -0.00000336, 0.00001221, -0.00001216, 0.00000668, -0.00001045, 0.00000848, 0.00000188, -0.00001040, 0.00000347, 0.00000415, -0.00000983, 0.00000860, 0.00001253, -0.00001816, 0.00000305, -0.00000158, 0.00000560, -0.00000883, 0.00000009, 0.00000365, -0.00000151, 0.00000671, -0.00000371, -0.00000607, 0.00000321, 0.00000366, -0.00000122, -0.00000240, -0.00000512, 0.00000069, -0.00000435, 0.00001080, -0.00000455, -0.00000553, 0.00001663, -0.00002886, 0.00002987, -0.00001643, 0.00001614, -0.00001666, 0.00002666, -0.00002667, 0.00001592, -0.00000325, },
		{ 0.00002841, -0.00002169, 0.00001491, -0.00001094, 0.00001157, -0.00001175, 0.00001317, -0.00000667, -0.00000504, 0.00001510, -0.00002060, 0.00001141, -0.00000738, 0.00000307, 0.00000111, -0.00000811, 0.00001069, -0.00002139, 0.00001790, -0.00000404, -0.00000698, 0.00002354, -0.00002574, 0.00002158, -0.00001809, 0.00001089, -0.00001035, 0.00001456, -0.00001884, 0.00001051, -0.00000993, 0.00001897, -0.00002180, 0.00002712, -0.00002653, 0.00001738, -0.00002021, 0.00002458, -0.00002009, 0.00000389, -0.00000974, 0.00002894, -0.00002438, 0.00002570, -0.00003772, 0.00003037, -0.00003531, 0.00004393, -0.00004595, 0.00004184, -0.00004367, 0.00005553, -0.00006817, 0.00005173, -0.00005363, 0.00007050, -0.00008958, 0.00008056, -0.00008881, 0.00010546, -0.00012778, 0.00013123, -0.00016759, 0.00017227, -0.00026560, 0.00029311, -0.00071020, 0.00070301, 0.00054459, -0.00064931, 0.00027026, -0.00024505, 0.00016174, -0.00015648, 0.00011159, -0.00011042, 0.00009335, -0.00007826, 0.00006565, -0.00007724, 0.00007276, -0.00005773, 0.00004450, -0.00005766, 0.00005648, -0.00004645, 0.00003798, -0.00003863, 0.00003126, -0.00002142, 0.00002061, -0.00003168, 0.00003349, -0.00003021, 0.00002093, -0.00000356, 0.00001103, -0.00002797, 0.00002353, -0.00001856, 0.00001886, -0.00002745, 0.00002562, -0.00002041, 0.00002151, -0.00001366, 0.00000754, -0.00001505, 0.00001822, -0.00001212, 0.00001013, -0.00001591, 0.00000960, -0.00001236, 0.00001119, 0.00000216, -0.00001157, 0.00001700, -0.00001674, 0.00000434, 0.00001003, -0.00000745, 0.00000680, -0.00000373, 0.00001397, -0.00002851, 0.00003571, -0.00003631, },
		{ 0.00000455, -0.00000765, 0.00000735, -0.00000497, 0.00000604, -0.00000810, 0.00001593, -0.00002190, 0.00002016, -0.00001171, -0.00000380, 0.00000698, -0.00000900, 0.00001039, -0.00001236, 0.00001015, -0.00001045, -0.00000506, 0.00002100, -0.00002688, 0.00003140, -0.00002001, 0.00000980, -0.00000536, 0.00000058, -0.00000452, 0.00000902, -0.00000870, 0.00000116, -0.00000673, 0.00001497, -0.00001294, 0.00001298, -0.00000676, -0.00000113, -0.00000470, 0.00000637, 0.00000001, -0.00000748, -0.00000927, 0.00002451, -0.00001465, 0.00001694, -0.00002274, 0.00001243, -0.00001934, 0.00002391, -0.00002019, 0.00001556, -0.00002027, 0.00002830, -0.00002739, 0.00000880, -0.00001990, 0.00003397, -0.00003633, 0.00001885, -0.00002794, 0.00003654, -0.00004055, 0.00002935, -0.00004555, 0.00002926, -0.00007281, 0.00003281, -0.00016172, -0.00033490, 0.00131045, -0.00068162, 0.00012269, -0.00014560, 0.00007128, -0.00008159, 0.00004542, -0.00005526, 0.00003927, -0.00003221, 0.00003132, -0.00004364, 0.00003040, -0.00001737, 0.00001644, -0.00003182, 0.00002185, -0.00001106, 0.00000849, -0.00001202, 0.00000636, -0.00000390, 0.00001131, -0.00002138, 0.00001560, -0.00000904, 0.00000194, 0.00000368, 0.00001479, -0.00002688, 0.00001551, -0.00001181, 0.00001474, -0.00002010, 0.00001255, -0.00000671, 0.00000714, -0.00000055, 0.00000187, -0.00001153, 0.00000962, -0.00000301, 0.00000441, -0.00000878, 0.00000168, -0.00000388, -0.00000146, 0.00001247, -0.00001149, 0.00000537, 0.00000638, -0.00002242, 0.00002844, -0.00002033, 0.00001955, -0.00002010, 0.00003019, -0.00003477, 0.00002682, -0.00001466, 0.00000141, },
		{ -0.00002035, 0.00001493, -0.00001104, 0.00001172, -0.00001196, 0.00001353, -0.00000873, -0.00000277, 0.00001473, -0.00002323, 0.00001647, -0.00001205, 0.00000905, -0.00000635, 0.00000072, 0.00000464, -0.00001985, 0.00002389, -0.00001564, 0.00000674, 0.00001125, -0.00001790, 0.00001690, -0.00001677, 0.00000991, -0.00000821, 0.00001225, -0.00001669, 0.00000833, -0.00000628, 0.00001434, -0.00001722, 0.00002409, -0.00002665, 0.00001791, -0.00001878, 0.00002405, -0.00002348, 0.00000496, -0.00000224, 0.00001968, -0.00001884, 0.00001988, -0.00003069, 0.00002292, -0.00002480, 0.00003339, -0.00003777, 0.00003414, -0.00003503, 0.00004625, -0.00006074, 0.00004552, -0.00004486, 0.00005956, -0.00007953, 0.00007134, -0.00007737, 0.00009233, -0.00011588, 0.00011899, -0.00015832, 0.00016131, -0.00025753, 0.00028466, -0.00079473, 0.00105439, 0.00014320, -0.00046723, 0.00019921, -0.00019479, 0.00012611, -0.00012837, 0.00009129, -0.00009256, 0.00007967, -0.00006590, 0.00005230, -0.00006483, 0.00006327, -0.00004985, 0.00003564, -0.00004926, 0.00005053, -0.00004241, 0.00003476, -0.00003646, 0.00003139, -0.00002068, 0.00001611, -0.00002631, 0.00002832, -0.00002750, 0.00002238, -0.00000189, 0.00000125, -0.00001774, 0.00001740, -0.00001284, 0.00001138, -0.00002094, 0.00002244, -0.00001859, 0.00002026, -0.00001266, 0.00000535, -0.00001189, 0.00001547, -0.00001043, 0.00000786, -0.00001451, 0.00001117, -0.00001559, 0.00001761, -0.00000528, -0.00000642, 0.00001808, -0.00002438, 0.00001637, -0.00000391, 0.00000333, -0.00000347, 0.00000683, 0.00000056, -0.00001642, 0.00002685, -0.00003071, 0.00002574, },
		{ -0.00000221, 0.00000349, -0.00000209, 0.00000390, -0.00000664, 0.00001453, -0.00002270, 0.00002397, -0.00001759, 0.00000200, 0.00000350, -0.00000548, 0.00000757, -0.00001125, 0.00001432, -0.00001915, 0.00000763, 0.00000924, -0.00001982, 0.00003056, -0.00002494, 0.00001683, -0.00001362, 0.00000710, -0.00000953, 0.00001439, -0.00001438, 0.00000667, -0.00001151, 0.00002056, -0.00002028, 0.00002266, -0.00001844, 0.00000907, -0.00001296, 0.00001613, -0.00001136, -0.00000117, -0.00000986, 0.00002910, -0.00002496, 0.00002696, -0.00003262, 0.00002308, -0.00002942, 0.00003736, -0.00003751, 0.00003376, -0.00003904, 0.00004902, -0.00005153, 0.00003423, -0.00004452, 0.00006082, -0.00006801, 0.00005393, -0.00006488, 0.00007852, -0.00009072, 0.00008697, -0.00011515, 0.00011015, -0.00017605, 0.00016961, -0.00040170, 0.00002271, 0.00114184, -0.00079586, 0.00026182, -0.00023944, 0.00014778, -0.00014470, 0.00010124, -0.00010213, 0.00008162, -0.00006838, 0.00006223, -0.00007308, 0.00005923, -0.00004350, 0.00003912, -0.00005408, 0.00004459, -0.00003256, 0.00002803, -0.00003029, 0.00002373, -0.00001725, 0.00002060, -0.00003103, 0.00002586, -0.00002064, 0.00001371, -0.00000150, 0.00001434, -0.00003062, 0.00002390, -0.00001912, 0.00002071, -0.00002812, 0.00002293, -0.00001623, 0.00001535, -0.00000740, 0.00000688, -0.00001610, 0.00001492, -0.00000874, 0.00000930, -0.00001473, 0.00000937, -0.00001090, 0.00000519, 0.00000962, -0.00001462, 0.00001506, -0.00000649, -0.00001052, 0.00001979, -0.00001706, 0.00001700, -0.00001790, 0.00002772, -0.00003656, 0.00003229, -0.00002191, 0.00000891, -0.00000168, },
		{ 0.00001351, -0.00001032, 0.00001157, -0.00001247, 0.00001444, -0.00001155, 0.00000087, 0.00001235, -0.00002386, 0.00002028, -0.00001584, 0.00001414, -0.00001342, 0.00001133, -0.00000599, -0.00001155, 0.00002237, -0.00002141, 0.00001749, -0.00000121, -0.00000761, 0.00000854, -0.00001167, 0.00000586, -0.00000329, 0.00000724, -0.00001192, 0.00000348, 0.00000023, 0.00000632, -0.00000832, 0.00001569, -0.00002131, 0.00001386, -0.00001296, 0.00001841, -0.00002265, 0.00000462, 0.00000623, 0.00000681, -0.00000816, 0.00000933, -0.00001914, 0.00001081, -0.00000865, 0.00001510, -0.00002039, 0.00001655, -0.00001581, 0.00002543, -0.00004073, 0.00002614, -0.00002231, 0.00003354, -0.00005265, 0.00004364, -0.00004531, 0.00005526, -0.00007597, 0.00007348, -0.00010949, 0.00010246, -0.00018772, 0.00018557, -0.00074962, 0.00129788, -0.00024774, -0.00023109, 0.00007408, -0.00010208, 0.00005489, -0.00006923, 0.00004394, -0.00005126, 0.00004544, -0.00003576, 0.00002299, -0.00003754, 0.00003995, -0.00002923, 0.00001475, -0.00002920, 0.00003341, -0.00002779, 0.00002155, -0.00002487, 0.00002347, -0.00001408, 0.00000694, -0.00001643, 0.00001827, -0.00001933, 0.00001992, 0.00000056, -0.00000897, -0.00000471, 0.00000748, -0.00000382, 0.00000034, -0.00000975, 0.00001424, -0.00001242, 0.00001541, -0.00000869, 0.00000062, -0.00000633, 0.00001014, -0.00000602, 0.00000263, -0.00000933, 0.00000875, -0.00001553, 0.00002237, -0.00001395, 0.00000343, 0.00001178, -0.00002433, 0.00002183, -0.00001357, 0.00001189, -0.00001197, 0.00001572, -0.00001181, -0.00000384, 0.00001645, -0.00002326, 0.00002122, -0.00001725, },
		{ 0.00000115, -0.00000034, 0.00000260, -0.00000607, 0.00001403, -0.00002424, 0.00002875, -0.00002534, 0.00001077, -0.00000326, 0.00000148, 0.00000047, -0.00000412, 0.00001162, -0.00002188, 0.00001690, -0.00000302, -0.00000897, 0.00002402, -0.00002396, 0.00001878, -0.00001819, 0.00001121, -0.00001240, 0.00001763, -0.00001792, 0.00000981, -0.00001357, 0.00002288, -0.00002385, 0.00002847, -0.00002718, 0.00001745, -0.00001933, 0.00002361, -0.00002170, 0.00000548, -0.00000878, 0.00002869, -0.00002975, 0.00003238, -0.00003772, 0.00002815, -0.00003243, 0.00004203, -0.00004567, 0.00004304, -0.00004877, 0.00006030, -0.00006594, 0.00004993, -0.00005901, 0.00007651, -0.00008747, 0.00007594, -0.00008741, 0.00010412, -0.00012272, 0.00012449, -0.00016303, 0.00016738, -0.00025238, 0.00027497, -0.00062292, 0.00043499, 0.00083173, -0.00076930, 0.00031290, -0.00027729, 0.00018231, -0.00017441, 0.00013014, -0.00012703, 0.00010544, -0.00008876, 0.00007886, -0.00008929, 0.00007608, -0.00005876, 0.00005161, -0.00006677, 0.00005871, -0.00004642, 0.00004084, -0.00004283, 0.00003691, -0.00002783, 0.00002698, -0.00003704, 0.00003195, -0.00002813, 0.00002323, -0.00000589, 0.00001104, -0.00002869, 0.00002663, -0.00002174, 0.00002173, -0.00003071, 0.00002866, -0.00002244, 0.00002106, -0.00001213, 0.00000984, -0.00001849, 0.00001784, -0.00001215, 0.00001191, -0.00001837, 0.00001579, -0.00001846, 0.00001451, 0.00000192, -0.00001178, 0.00001959, -0.00001718, 0.00000236, 0.00000745, -0.00000890, 0.00000999, -0.00001138, 0.00002036, -0.00003264, 0.00003236, -0.00002468, 0.00001277, -0.00000494, 0.00000099, },
		{ -0.00000992, 0.00001211, -0.00001424, 0.00001701, -0.00001639, 0.00000726, 0.00000626, -0.00002032, 0.00002044, -0.00001627, 0.00001563, -0.00001664, 0.00001959, -0.00001715, 0.00000049, 0.00001443, -0.00002036, 0.00002284, -0.00001073, 0.00000209, -0.00000104, -0.00000450, -0.00000003, 0.00000335, 0.00000054, -0.00000563, -0.00000277, 0.00000811, -0.00000337, 0.00000306, 0.00000350, -0.00001154, 0.00000597, -0.00000354, 0.00000823, -0.00001751, 0.00000299, 0.00001413, -0.00000712, 0.00000520, -0.00000371, -0.00000563, -0.00000269, 0.00000895, -0.00000616, 0.00000139, -0.00000619, 0.00000919, -0.00000203, -0.00001314, -0.00000132, 0.00000866, -0.00000175, -0.00001528, 0.00000440, -0.00000038, 0.00000284, -0.00001749, 0.00000479, -0.00003166, 0.00000681, -0.00006770, 0.00000312, -0.00056125, 0.00138495, -0.00056125, 0.00000312, -0.00006770, 0.00000681, -0.00003166, 0.00000479, -0.00001749, 0.00000284, -0.00000038, 0.00000440, -0.00001528, -0.00000175, 0.00000866, -0.00000132, -0.00001314, -0.00000203, 0.00000919, -0.00000619, 0.00000139, -0.00000616, 0.00000895, -0.00000269, -0.00000563, -0.00000371, 0.00000520, -0.00000712, 0.00001413, 0.00000299, -0.00001751, 0.00000823, -0.00000354, 0.00000597, -0.00001154, 0.00000350, 0.00000306, -0.00000337, 0.00000811, -0.00000277, -0.00000563, 0.00000054, 0.00000335, -0.00000003, -0.00000450, -0.00000104, 0.00000209, -0.00001073, 0.00002284, -0.00002036, 0.00001443, 0.00000049, -0.00001715, 0.00001959, -0.00001664, 0.00001563, -0.00001627, 0.00002044, -0.00002032, 0.00000626, 0.00000726, -0.00001639, 0.00001701, -0.00001424, 0.00001211, },
		{ 0.00000115, 0.00000099, -0.00000494, 0.00001277, -0.00002468, 0.00003236, -0.00003264, 0.00002036, -0.00001138, 0.00000999, -0.00000890, 0.00000745, 0.00000236, -0.00001718, 0.00001959, -0.00001178, 0.00000192, 0.00001451, -0.00001846, 0.00001579, -0.00001837, 0.00001191, -0.00001215, 0.00001784, -0.00001849, 0.00000984, -0.00001213, 0.00002106, -0.00002244, 0.00002866, -0.00003071, 0.00002173, -0.00002174, 0.00002663, -0.00002869, 0.00001104, -0.00000589, 0.00002323, -0.00002813, 0.00003195, -0.00003704, 0.00002698, -0.00002783, 0.00003691, -0.00004283, 0.00004084, -0.00004642, 0.00005871, -0.00006677, 0.00005161, -0.00005876, 0.00007608, -0.00008929, 0.00007886, -0.00008876, 0.00010544, -0.00012703, 0.00013014, -0.00017441, 0.00018231, -0.00027729, 0.00031290, -0.00076930, 0.00083173, 0.00043499, -0.00062292, 0.00027497, -0.00025238, 0.00016738, -0.00016303, 0.00012449, -0.00012272, 0.00010412, -0.00008741, 0.00007594, -0.00008747, 0.00007651, -0.00005901, 0.00004993, -0.00006594, 0.00006030, -0.00004877, 0.00004304, -0.00004567, 0.00004203, -0.00003243, 0.00002815, -0.00003772, 0.00003238, -0.00002975, 0.00002869, -0.00000878, 0.00000548, -0.00002170, 0.00002361, -0.00001933, 0.00001745, -0.00002718, 0.00002847, -0.00002385, 0.00002288, -0.00001357, 0.00000981, -0.00001792, 0.00001763, -0.00001240, 0.00001121, -0.00001819, 0.00001878, -0.00002396, 0.00002402, -0.00000897, -0.00000302, 0.00001690, -0.00002188, 0.00001162, -0.00000412, 0.00000047, 0.00000148, -0.00000326, 0.00001077, -0.00002534, 0.00002875, -0.00002424, 0.00001403, -0.00000607, 0.00000260, -0.00000034, },
		{ 0.00001351, -0.00001725, 0.00002122, -0.00002326, 0.00001645, -0.00000384, -0.00001181, 0.00001572, -0.00001197, 0.00001189, -0.00001357, 0.00002183, -0.00002433, 0.00001178, 0.00000343, -0.00001395, 0.00002237, -0.00001553, 0.00000875, -0.00000933, 0.00000263, -0.00000602, 0.00001014, -0.00000633, 0.00000062, -0.00000869, 0.00001541, -0.00001242, 0.00001424, -0.00000975, 0.00000034, -0.00000382, 0.00000748, -0.00000471, -0.00000897, 0.00000056, 0.00001992, -0.00001933, 0.00001827, -0.00001643, 0.00000694, -0.00001408, 0.00002347, -0.00002487, 0.00002155, -0.00002779, 0.00003341, -0.00002920, 0.00001475, -0.00002923, 0.00003995, -0.00003754, 0.00002299, -0.00003576, 0.00004544, -0.00005126, 0.00004394, -0.00006923, 0.00005489, -0.00010208, 0.00007408, -0.00023109, -0.00024774, 0.00129788, -0.00074962, 0.00018557, -0.00018772, 0.00010246, -0.00010949, 0.00007348, -0.00007597, 0.00005526, -0.00004531, 0.00004364, -0.00005265, 0.00003354, -0.00002231, 0.00002614, -0.00004073, 0.00002543, -0.00001581, 0.00001655, -0.00002039, 0.00001510, -0.00000865, 0.00001081, -0.00001914, 0.00000933, -0.00000816, 0.00000681, 0.00000623, 0.00000462, -0.00002265, 0.00001841, -0.00001296, 0.00001386, -0.00002131, 0.00001569, -0.00000832, 0.00000632, 0.00000023, 0.00000348, -0.00001192, 0.00000724, -0.00000329, 0.00000586, -0.00001167, 0.00000854, -0.00000761, -0.00000121, 0.00001749, -0.00002141, 0.00002237, -0.00001155, -0.00000599, 0.00001133, -0.00001342, 0.00001414, -0.00001584, 0.00002028, -0.00002386, 0.00001235, 0.00000087, -0.00001155, 0.00001444, -0.00001247, 0.00001157, -0.00001032, },
		{ -0.00000221, -0.00000168, 0.00000891, -0.00002191, 0.00003229, -0.00003656, 0.00002772, -0.00001790, 0.00001700, -0.00001706, 0.00001979, -0.00001052, -0.00000649, 0.00001506, -0.00001462, 0.00000962, 0.00000519, -0.00001090, 0.00000937, -0.00001473, 0.00000930, -0.00000874, 0.00001492, -0.00001610, 0.00000688, -0.00000740, 0.00001535, -0.00001623, 0.00002293, -0.00002812, 0.00002071, -0.00001912, 0.00002390, -0.00003062, 0.00001434, -0.00000150, 0.00001371, -0.00002064, 0.00002586, -0.00003103, 0.00002060, -0.00001725, 0.00002373, -0.00003029, 0.00002803, -0.00003256, 0.00004459, -0.00005408, 0.00003912, -0.00004350, 0.00005923, -0.00007308, 0.00006223, -0.00006838, 0.00008162, -0.00010213, 0.00010124, -0.00014470, 0.00014778, -0.00023944, 0.00026182, -0.00079586, 0.00114184, 0.00002271, -0.00040170, 0.00016961, -0.00017605, 0.00011015, -0.00011515, 0.00008697, -0.00009072, 0.00007852, -0.00006488, 0.00005393, -0.00006801, 0.00006082, -0.00004452, 0.00003423, -0.00005153, 0.00004902, -0.00003904, 0.00003376, -0.00003751, 0.00003736, -0.00002942, 0.00002308, -0.00003262, 0.00002696, -0.00002496, 0.00002910, -0.00000986, -0.00000117, -0.00001136, 0.00001613, -0.00001296, 0.00000907, -0.00001844, 0.00002266, -0.00002028, 0.00002056, -0.00001151, 0.00000667, -0.00001438, 0.00001439, -0.00000953, 0.00000710, -0.00001362, 0.00001683, -0.00002494, 0.00003056, -0.00001982, 0.00000924, 0.00000763, -0.00001915, 0.00001432, -0.00001125, 0.00000757, -0.00000548, 0.00000350, 0.00000200, -0.00001759, 0.00002397, -0.00002270, 0.00001453, -0.00000664, 0.00000390, -0.00000209, 0.00000349, },
		{ -0.00002035, 0.00002574, -0.00003071, 0.00002685, -0.00001642, 0.00000056, 0.00000683, -0.00000347, 0.00000333, -0.00000391, 0.00001637, -0.00002438, 0.00001808, -0.00000642, -0.00000528, 0.00001761, -0.00001559, 0.00001117, -0.00001451, 0.00000786, -0.00001043, 0.00001547, -0.00001189, 0.00000535, -0.00001266, 0.00002026, -0.00001859, 0.00002244, -0.00002094, 0.00001138, -0.00001284, 0.00001740, -0.00001774, 0.00000125, -0.00000189, 0.00002238, -0.00002750, 0.00002832, -0.00002631, 0.00001611, -0.00002068, 0.00003139, -0.00003646, 0.00003476, -0.00004241, 0.00005053, -0.00004926, 0.00003564, -0.00004985, 0.00006327, -0.00006483, 0.00005230, -0.00006590, 0.00007967, -0.00009256, 0.00009129, -0.00012837, 0.00012611, -0.00019479, 0.00019921, -0.00046723, 0.00014320, 0.00105439, -0.00079473, 0.00028466, -0.00025753, 0.00016131, -0.00015832, 0.00011899, -0.00011588, 0.00009233, -0.00007737, 0.00007134, -0.00007953, 0.00005956, -0.00004486, 0.00004552, -0.00006074, 0.00004625, -0.00003503, 0.00003414, -0.00003777, 0.00003339, -0.00002480, 0.00002292, -0.00003069, 0.00001988, -0.00001884, 0.00001968, -0.00000224, 0.00000496, -0.00002348, 0.00002405, -0.00001878, 0.00001791, -0.00002665, 0.00002409, -0.00001722, 0.00001434, -0.00000628, 0.00000833, -0.00001669, 0.00001225, -0.00000821, 0.00000991, -0.00001677, 0.00001690, -0.00001790, 0.00001125, 0.00000674, -0.00001564, 0.00002389, -0.00001985, 0.00000464, 0.00000072, -0.00000635, 0.00000905, -0.00001205, 0.00001647, -0.00002323, 0.00001473, -0.00000277, -0.00000873, 0.00001353, -0.00001196, 0.00001172, -0.00001104, 0.00001493, },
		{ 0.00000455, 0.00000141, -0.00001466, 0.00002682, -0.00003477, 0.00003019, -0.00002010, 0.00001955, -0.00002033, 0.00002844, -0.00002242, 0.00000638, 0.00000537, -0.00001149, 0.00001247, -0.00000146, -0.00000388, 0.00000168, -0.00000878, 0.00000441, -0.00000301, 0.00000962, -0.00001153, 0.00000187, -0.00000055, 0.00000714, -0.00000671, 0.00001255, -0.00002010, 0.00001474, -0.00001181, 0.00001551, -0.00002688, 0.00001479, 0.00000368, 0.00000194, -0.00000904, 0.00001560, -0.00002138, 0.00001131, -0.00000390, 0.00000636, -0.00001202, 0.00000849, -0.00001106, 0.00002185, -0.00003182, 0.00001644, -0.00001737, 0.00003040, -0.00004364, 0.00003132, -0.00003221, 0.00003927, -0.00005526, 0.00004542, -0.00008159, 0.00007128, -0.00014560, 0.00012269, -0.00068162, 0.00131045, -0.00033490, -0.00016172, 0.00003281, -0.00007281, 0.00002926, -0.00004555, 0.00002935, -0.00004055, 0.00003654, -0.00002794, 0.00001885, -0.00003633, 0.00003397, -0.00001990, 0.00000880, -0.00002739, 0.00002830, -0.00002027, 0.00001556, -0.00002019, 0.00002391, -0.00001934, 0.00001243, -0.00002274, 0.00001694, -0.00001465, 0.00002451, -0.00000927, -0.00000748, 0.00000001, 0.00000637, -0.00000470, -0.00000113, -0.00000676, 0.00001298, -0.00001294, 0.00001497, -0.00000673, 0.00000116, -0.00000870, 0.00000902, -0.00000452, 0.00000058, -0.00000536, 0.00000980, -0.00002001, 0.00003140, -0.00002688, 0.00002100, -0.00000506, -0.00001045, 0.00001015, -0.00001236, 0.00001039, -0.00000900, 0.00000698, -0.00000380, -0.00001171, 0.00002016, -0.00002190, 0.00001593, -0.00000810, 0.00000604, -0.00000497, 0.00000735, -0.00000765, },
		{ 0.00002841, -0.00003631, 0.00003571, -0.00002851, 0.00001397, -0.00000373, 0.00000680, -0.00000745, 0.00001003, 0.00000434, -0.00001674, 0.00001700, -0.00001157, 0.00000216, 0.00001119, -0.00001236, 0.00000960, -0.00001591, 0.00001013, -0.00001212, 0.00001822, -0.00001505, 0.00000754, -0.00001366, 0.00002151, -0.00002041, 0.00002562, -0.00002745, 0.00001886, -0.00001856, 0.00002353, -0.00002797, 0.00001103, -0.00000356, 0.00002093, -0.00003021, 0.00003349, -0.00003168, 0.00002061, -0.00002142, 0.00003126, -0.00003863, 0.00003798, -0.00004645, 0.00005648, -0.00005766, 0.00004450, -0.00005773, 0.00007276, -0.00007724, 0.00006565, -0.00007826, 0.00009335, -0.00011042, 0.00011159, -0.00015648, 0.00016174, -0.00024505, 0.00027026, -0.00064931, 0.00054459, 0.00070301, -0.00071020, 0.00029311, -0.00026560, 0.00017227, -0.00016759, 0.00013123, -0.00012778, 0.00010546, -0.00008881, 0.00008056, -0.00008958, 0.00007050, -0.00005363, 0.00005173, -0.00006817, 0.00005553, -0.00004367, 0.00004184, -0.00004595, 0.00004393, -0.00003531, 0.00003037, -0.00003772, 0.00002570, -0.00002438, 0.00002894, -0.00000974, 0.00000389, -0.00002009, 0.00002458, -0.00002021, 0.00001738, -0.00002653, 0.00002712, -0.00002180, 0.00001897, -0.00000993, 0.00001051, -0.00001884, 0.00001456, -0.00001035, 0.00001089, -0.00001809, 0.00002158, -0.00002574, 0.00002354, -0.00000698, -0.00000404, 0.00001790, -0.00002139, 0.00001069, -0.00000811, 0.00000111, 0.00000307, -0.00000738, 0.00001141, -0.00002060, 0.00001510, -0.00000504, -0.00000667, 0.00001317, -0.00001175, 0.00001157, -0.00001094, 0.00001491, -0.00002169, },
		{ -0.00000930, -0.00000325, 0.00001592, -0.00002667, 0.00002666, -0.00001666, 0.00001614, -0.00001643, 0.00002987, -0.00002886, 0.00001663, -0.00000553, -0.00000455, 0.00001080, -0.00000435, 0.00000069, -0.00000512, -0.00000240, -0.00000122, 0.00000366, 0.00000321, -0.00000607, -0.00000371, 0.00000671, -0.00000151, 0.00000365, 0.00000009, -0.00000883, 0.00000560, -0.00000158, 0.00000305, -0.00001816, 0.00001253, 0.00000860, -0.00000983, 0.00000415, 0.00000347, -0.00001040, 0.00000188, 0.00000848, -0.00001045, 0.00000668, -0.00001216, 0.00001221, -0.00000336, -0.00000647, -0.00000964, 0.00001242, -0.00000270, -0.00000936, -0.00000462, 0.00000926, -0.00000959, -0.00000016, -0.00002156, -0.00000289, -0.00002676, -0.00001991, -0.00007914, -0.00043662, 0.00131186, -0.00058465, 0.00004480, -0.00009655, 0.00002814, -0.00005161, 0.00002593, -0.00003167, 0.00001363, -0.00000967, 0.00001279, -0.00001979, -0.00000102, 0.00000383, 0.00000746, -0.00001937, -0.00000005, 0.00000419, 0.00000194, -0.00000637, 0.00000174, 0.00000516, -0.00000475, -0.00000160, -0.00001034, 0.00000480, -0.00000100, 0.00001598, -0.00000749, -0.00001211, 0.00001010, -0.00000327, 0.00000314, -0.00001054, 0.00000500, 0.00000201, -0.00000390, 0.00000785, -0.00000073, -0.00000532, -0.00000227, 0.00000300, 0.00000106, -0.00000651, 0.00000464, -0.00000084, -0.00000954, 0.00002535, -0.00002728, 0.00002807, -0.00001669, 0.00000056, 0.00000138, -0.00000826, 0.00000899, -0.00000900, 0.00000702, -0.00000612, -0.00000853, 0.00001817, -0.00002258, 0.00001891, -0.00001104, 0.00000936, -0.00000900, 0.00001245, -0.00001484, 0.00001324, },
		{ -0.00003760, 0.00004009, -0.00003671, 0.00002493, -0.00001257, 0.00001538, -0.00001655, 0.00002402, -0.00001053, -0.00000384, 0.00000898, -0.00001054, 0.00000582, 0.00000580, -0.00000800, 0.00000537, -0.00001396, 0.00000935, -0.00001081, 0.00001798, -0.00001547, 0.00000697, -0.00001152, 0.00001904, -0.00001768, 0.00002314, -0.00002797, 0.00002111, -0.00001936, 0.00002400, -0.00003305, 0.00001840, -0.00000387, 0.00001576, -0.00002725, 0.00003316, -0.00003208, 0.00002049, -0.00001697, 0.00002394, -0.00003187, 0.00003120, -0.00003955, 0.00005059, -0.00005343, 0.00004009, -0.00005140, 0.00006676, -0.00007291, 0.00006102, -0.00007064, 0.00008390, -0.00010145, 0.00010012, -0.00014670, 0.00015195, -0.00023839, 0.00026225, -0.00072996, 0.00088591, 0.00031070, -0.00053502, 0.00022681, -0.00021865, 0.00013883, -0.00013891, 0.00011034, -0.00011095, 0.00009346, -0.00007833, 0.00007004, -0.00008157, 0.00006520, -0.00004760, 0.00004373, -0.00006184, 0.00005193, -0.00004025, 0.00003794, -0.00004278, 0.00004414, -0.00003759, 0.00003115, -0.00003883, 0.00002576, -0.00002351, 0.00003282, -0.00001511, 0.00000173, -0.00001347, 0.00002065, -0.00001772, 0.00001292, -0.00002144, 0.00002468, -0.00002152, 0.00001958, -0.00001015, 0.00000954, -0.00001801, 0.00001394, -0.00000947, 0.00000852, -0.00001492, 0.00002096, -0.00002844, 0.00003211, -0.00001987, 0.00001021, 0.00000603, -0.00001576, 0.00001009, -0.00001217, 0.00000605, -0.00000129, -0.00000419, 0.00000758, -0.00001840, 0.00001559, -0.00000778, -0.00000364, 0.00001176, -0.00001047, 0.00000993, -0.00000880, 0.00001208, -0.00001954, 0.00002714, },
		{ 0.00001026, 0.00000156, -0.00001373, 0.00001816, -0.00000844, 0.00000754, -0.00000555, 0.00002295, -0.00002719, 0.00002034, -0.00001338, 0.00000280, 0.00000655, -0.00000394, 0.00000208, -0.00000956, 0.00000277, -0.00000602, 0.00000976, -0.00000288, -0.00000113, -0.00000834, 0.00001261, -0.00000845, 0.00001217, -0.00001136, 0.00000270, -0.00000402, 0.00000883, -0.00001074, -0.00000631, 0.00000838, 0.00001217, -0.00001941, 0.00001632, -0.00000810, -0.00000032, -0.00000533, 0.00001679, -0.00002256, 0.00002085, -0.00002845, 0.00003129, -0.00002460, 0.00001508, -0.00003178, 0.00003804, -0.00003157, 0.00002054, -0.00003540, 0.00004443, -0.00005150, 0.00004730, -0.00008087, 0.00006898, -0.00011911, 0.00010257, -0.00029929, -0.00010058, 0.00115488, -0.00070123, 0.00018188, -0.00018773, 0.00010200, -0.00011142, 0.00008086, -0.00007994, 0.00005770, -0.00004773, 0.00004634, -0.00005211, 0.00002896, -0.00002139, 0.00002981, -0.00004287, 0.00002345, -0.00001665, 0.00002117, -0.00002582, 0.00002249, -0.00001401, 0.00001053, -0.00001587, 0.00000169, -0.00000639, 0.00001304, 0.00000529, -0.00000515, -0.00001419, 0.00001718, -0.00001084, 0.00000869, -0.00001696, 0.00001421, -0.00000761, 0.00000454, 0.00000117, 0.00000473, -0.00001119, 0.00000343, -0.00000212, 0.00000544, -0.00001201, 0.00001372, -0.00001233, 0.00000424, 0.00001337, -0.00002022, 0.00002770, -0.00002312, 0.00000945, -0.00000820, -0.00000160, 0.00000512, -0.00000699, 0.00000503, -0.00000614, -0.00000725, 0.00001736, -0.00002408, 0.00002286, -0.00001492, 0.00001319, -0.00001323, 0.00001752, -0.00002215, 0.00002252, -0.00002124, },
		{ 0.00003815, -0.00003855, 0.00003059, -0.00001676, 0.00001918, -0.00002018, 0.00003341, -0.00002327, 0.00000980, -0.00000295, -0.00000428, 0.00000496, 0.00000314, -0.00000453, 0.00000026, -0.00000989, 0.00000624, -0.00000700, 0.00001512, -0.00001351, 0.00000417, -0.00000699, 0.00001379, -0.00001148, 0.00001589, -0.00002283, 0.00001806, -0.00001509, 0.00001843, -0.00003180, 0.00002204, -0.00000270, 0.00000782, -0.00001955, 0.00002794, -0.00002818, 0.00001694, -0.00000943, 0.00001223, -0.00001912, 0.00001727, -0.00002445, 0.00003556, -0.00003925, 0.00002505, -0.00003359, 0.00004812, -0.00005489, 0.00004181, -0.00004696, 0.00005572, -0.00007047, 0.00006188, -0.00010370, 0.00010066, -0.00017720, 0.00017098, -0.00068337, 0.00110997, -0.00005374, -0.00032124, 0.00011720, -0.00013724, 0.00007638, -0.00008437, 0.00006581, -0.00007288, 0.00006239, -0.00005099, 0.00004418, -0.00005939, 0.00004721, -0.00003013, 0.00002468, -0.00004456, 0.00003791, -0.00002694, 0.00002425, -0.00002948, 0.00003435, -0.00003144, 0.00002507, -0.00003418, 0.00002059, -0.00001650, 0.00003075, -0.00001771, -0.00000091, -0.00000520, 0.00001378, -0.00001270, 0.00000617, -0.00001311, 0.00001810, -0.00001720, 0.00001670, -0.00000737, 0.00000583, -0.00001469, 0.00001096, -0.00000630, 0.00000360, -0.00000792, 0.00001493, -0.00002463, 0.00003419, -0.00002801, 0.00002272, -0.00000791, -0.00000527, 0.00000339, -0.00001065, 0.00000711, -0.00000280, -0.00000366, 0.00000634, -0.00001816, 0.00001762, -0.00001228, 0.00000159, 0.00000813, -0.00000718, 0.00000610, -0.00000409, 0.00000590, -0.00001318, 0.00002091, -0.00003327, },
		{ -0.00001290, 0.00000085, 0.00000755, 0.00000181, -0.00000327, 0.00000947, 0.00000953, -0.00001774, 0.00001597, -0.00001522, 0.00000734, 0.00000232, -0.00000182, 0.00000075, -0.00001113, 0.00000573, -0.00000886, 0.00001410, -0.00000750, 0.00000221, -0.00001091, 0.00001592, -0.00001219, 0.00001679, -0.00001914, 0.00001163, -0.00001136, 0.00001649, -0.00002267, 0.00000614, 0.00000358, 0.00001349, -0.00002516, 0.00002534, -0.00001715, 0.00000727, -0.00000902, 0.00001945, -0.00002760, 0.00002737, -0.00003664, 0.00004200, -0.00003721, 0.00002772, -0.00004440, 0.00005344, -0.00004968, 0.00003897, -0.00005320, 0.00006436, -0.00007591, 0.00007438, -0.00011665, 0.00011421, -0.00018030, 0.00018616, -0.00048572, 0.00026651, 0.00087911, -0.00069000, 0.00023637, -0.00022540, 0.00013510, -0.00013774, 0.00010761, -0.00010473, 0.00008190, -0.00006884, 0.00006481, -0.00007093, 0.00004707, -0.00003572, 0.00004152, -0.00005621, 0.00003776, -0.00002902, 0.00003227, -0.00003755, 0.00003678, -0.00002880, 0.00002262, -0.00002730, 0.00001069, -0.00001382, 0.00002445, -0.00000548, -0.00000285, -0.00001343, 0.00002038, -0.00001527, 0.00001102, -0.00001924, 0.00001921, -0.00001395, 0.00001054, -0.00000348, 0.00000822, -0.00001520, 0.00000724, -0.00000518, 0.00000736, -0.00001418, 0.00001931, -0.00002139, 0.00001782, -0.00000154, -0.00000737, 0.00001966, -0.00002214, 0.00001276, -0.00001484, 0.00000443, 0.00000127, -0.00000520, 0.00000327, -0.00000596, -0.00000604, 0.00001607, -0.00002479, 0.00002622, -0.00001837, 0.00001618, -0.00001617, 0.00002074, -0.00002735, 0.00002983, -0.00003157, 0.00002269, },
		{ -0.00003348, 0.00002991, -0.00001511, 0.00001684, -0.00001627, 0.00003489, -0.00002951, 0.00001923, -0.00001430, 0.00000427, 0.00000074, 0.00000350, -0.00000318, -0.00000408, -0.00000519, 0.00000204, -0.00000175, 0.00001059, -0.00001012, 0.00000022, -0.00000140, 0.00000742, -0.00000386, 0.00000607, -0.00001388, 0.00001114, -0.00000717, 0.00000804, -0.00002457, 0.00002160, -0.00000039, -0.00000138, -0.00000891, 0.00001934, -0.00002145, 0.00001178, -0.00000147, -0.00000020, -0.00000462, 0.00000072, -0.00000589, 0.00001636, -0.00002033, 0.00000487, -0.00001005, 0.00002297, -0.00002981, 0.00001532, -0.00001559, 0.00001848, -0.00002861, 0.00000968, -0.00004195, 0.00002424, -0.00008047, 0.00001531, -0.00051250, 0.00118579, -0.00033553, -0.00011961, 0.00000043, -0.00004781, 0.00000612, -0.00002170, 0.00001255, -0.00002614, 0.00002300, -0.00001611, 0.00001126, -0.00003046, 0.00002329, -0.00000760, 0.00000064, -0.00002200, 0.00001873, -0.00000869, 0.00000541, -0.00001011, 0.00001766, -0.00001897, 0.00001381, -0.00002544, 0.00001213, -0.00000514, 0.00002343, -0.00001754, -0.00000328, 0.00000299, 0.00000581, -0.00000689, -0.00000081, -0.00000389, 0.00000954, -0.00001054, 0.00001173, -0.00000282, 0.00000049, -0.00000995, 0.00000684, -0.00000219, -0.00000218, 0.00000101, 0.00000505, -0.00001491, 0.00002883, -0.00002878, 0.00002946, -0.00001932, 0.00000601, -0.00000657, -0.00000495, 0.00000486, -0.00000189, -0.00000537, 0.00000749, -0.00001992, 0.00002129, -0.00001856, 0.00000913, 0.00000212, -0.00000188, 0.00000036, 0.00000265, -0.00000287, -0.00000336, 0.00001035, -0.00002375, 0.00002998, },
		{ 0.00001339, -0.00000161, 0.00001046, -0.00001216, 0.00002408, -0.00000613, -0.00000376, 0.00000496, -0.00001048, 0.00000725, 0.00000029, 0.00000015, -0.00000210, -0.00001021, 0.00000628, -0.00000928, 0.00001607, -0.00001006, 0.00000348, -0.00001104, 0.00001627, -0.00001228, 0.00001666, -0.00002171, 0.00001604, -0.00001449, 0.00001921, -0.00002998, 0.00001662, -0.00000063, 0.00001211, -0.00002628, 0.00002996, -0.00002254, 0.00001170, -0.00000919, 0.00001673, -0.00002550, 0.00002565, -0.00003575, 0.00004298, -0.00003957, 0.00002949, -0.00004527, 0.00005611, -0.00005428, 0.00004301, -0.00005493, 0.00006550, -0.00007846, 0.00007537, -0.00012106, 0.00012190, -0.00019484, 0.00020544, -0.00059263, 0.00059920, 0.00054353, -0.00058162, 0.00021742, -0.00021140, 0.00012733, -0.00012939, 0.00010410, -0.00010336, 0.00008335, -0.00007014, 0.00006551, -0.00007369, 0.00005101, -0.00003711, 0.00004064, -0.00005735, 0.00004078, -0.00003078, 0.00003297, -0.00003888, 0.00004142, -0.00003593, 0.00002873, -0.00003376, 0.00001504, -0.00001571, 0.00003092, -0.00001444, -0.00000102, -0.00001020, 0.00001979, -0.00001648, 0.00001027, -0.00001749, 0.00001964, -0.00001623, 0.00001324, -0.00000527, 0.00000901, -0.00001672, 0.00000871, -0.00000580, 0.00000642, -0.00001229, 0.00001982, -0.00002527, 0.00002747, -0.00001525, 0.00000753, 0.00000636, -0.00001427, 0.00000927, -0.00001626, 0.00000747, -0.00000061, -0.00000539, 0.00000361, -0.00000758, -0.00000306, 0.00001258, -0.00002291, 0.00002727, -0.00001984, 0.00001698, -0.00001646, 0.00002059, -0.00002856, 0.00003287, -0.00003760, 0.00003091, -0.00002385, },
		{ 0.00002410, -0.00000865, 0.00000943, -0.00000543, 0.00002786, -0.00002720, 0.00002094, -0.00002087, 0.00001145, -0.00000442, 0.00000582, -0.00000408, -0.00000667, -0.00000115, -0.00000199, 0.00000370, 0.00000553, -0.00000640, -0.00000364, 0.00000381, 0.00000171, 0.00000288, -0.00000356, -0.00000387, 0.00000284, 0.00000188, -0.00000459, -0.00001311, 0.00001774, 0.00000230, -0.00001008, 0.00000249, 0.00000935, -0.00001360, 0.00000674, 0.00000456, -0.00001002, 0.00000748, -0.00001378, 0.00001108, -0.00000158, -0.00000253, -0.00001417, 0.00001249, -0.00000147, -0.00000543, -0.00000983, 0.00001365, -0.00001639, 0.00001060, -0.00004030, 0.00001926, -0.00005393, 0.00002190, -0.00016754, -0.00024843, 0.00111433, -0.00050469, 0.00003262, -0.00009330, 0.00002599, -0.00005213, 0.00003178, -0.00003430, 0.00001604, -0.00001313, 0.00001607, -0.00001955, -0.00000301, 0.00000096, 0.00001291, -0.00002156, -0.00000062, 0.00000053, 0.00000855, -0.00001274, 0.00000974, -0.00000117, -0.00000388, 0.00000026, -0.00001522, 0.00000318, 0.00000776, 0.00001262, -0.00001513, -0.00000474, 0.00000964, -0.00000159, -0.00000184, -0.00000623, 0.00000398, 0.00000131, -0.00000359, 0.00000642, 0.00000191, -0.00000506, -0.00000511, 0.00000292, 0.00000130, -0.00000683, 0.00000929, -0.00000582, -0.00000174, 0.00001739, -0.00002187, 0.00002824, -0.00002447, 0.00001379, -0.00001593, 0.00000211, 0.00000125, -0.00000026, -0.00000759, 0.00000939, -0.00002231, 0.00002533, -0.00002537, 0.00001780, -0.00000527, 0.00000442, -0.00000608, 0.00000986, -0.00001234, 0.00000781, -0.00000236, -0.00001128, 0.00001768, -0.00002314, },
		{ -0.00000646, 0.00001450, -0.00001547, 0.00003363, -0.00001890, 0.00000974, -0.00000880, -0.00000109, 0.00000269, 0.00000140, 0.00000050, -0.00000505, -0.00000772, 0.00000487, -0.00000748, 0.00001571, -0.00001059, 0.00000285, -0.00000906, 0.00001421, -0.00000943, 0.00001239, -0.00001914, 0.00001551, -0.00001290, 0.00001621, -0.00003106, 0.00002324, -0.00000334, 0.00000820, -0.00002294, 0.00003000, -0.00002407, 0.00001325, -0.00000690, 0.00001040, -0.00001822, 0.00001754, -0.00002745, 0.00003584, -0.00003321, 0.00002190, -0.00003585, 0.00004750, -0.00004691, 0.00003444, -0.00004276, 0.00005041, -0.00006197, 0.00005314, -0.00009642, 0.00009334, -0.00016214, 0.00015329, -0.00059276, 0.00084236, 0.00021115, -0.00042083, 0.00014990, -0.00016109, 0.00008999, -0.00009518, 0.00007709, -0.00008100, 0.00006605, -0.00005486, 0.00005114, -0.00006272, 0.00004287, -0.00002766, 0.00002918, -0.00004806, 0.00003404, -0.00002328, 0.00002436, -0.00003037, 0.00003612, -0.00003448, 0.00002789, -0.00003465, 0.00001458, -0.00001179, 0.00003132, -0.00002027, 0.00000018, -0.00000535, 0.00001623, -0.00001518, 0.00000743, -0.00001290, 0.00001636, -0.00001486, 0.00001283, -0.00000433, 0.00000720, -0.00001589, 0.00000811, -0.00000444, 0.00000326, -0.00000691, 0.00001517, -0.00002272, 0.00003050, -0.00002386, 0.00001995, -0.00000794, -0.00000250, 0.00000033, -0.00001236, 0.00000682, 0.00000009, -0.00000807, 0.00000672, -0.00001185, 0.00000260, 0.00000597, -0.00001743, 0.00002489, -0.00001841, 0.00001487, -0.00001356, 0.00001654, -0.00002503, 0.00003058, -0.00003795, 0.00003303, -0.00002883, 0.00002109, },
		{ -0.00000026, 0.00000010, 0.00000917, 0.00001469, -0.00001743, 0.00001423, -0.00002027, 0.00001430, -0.00000793, 0.00000835, -0.00000649, -0.00000728, 0.00000144, -0.00000486, 0.00000826, 0.00000098, -0.00000332, -0.00000644, 0.00000755, -0.00000203, 0.00000689, -0.00001051, 0.00000447, -0.00000416, 0.00000920, -0.00001629, -0.00000006, 0.00001182, 0.00000444, -0.00001670, 0.00001261, -0.00000014, -0.00000614, 0.00000297, 0.00000732, -0.00001519, 0.00001440, -0.00002288, 0.00002265, -0.00001406, 0.00000954, -0.00002703, 0.00002851, -0.00001928, 0.00001188, -0.00002672, 0.00003286, -0.00003958, 0.00003590, -0.00007400, 0.00006230, -0.00011129, 0.00009845, -0.00033213, 0.00005804, 0.00092558, -0.00055935, 0.00011831, -0.00014690, 0.00006974, -0.00008568, 0.00006441, -0.00006401, 0.00004393, -0.00003725, 0.00003768, -0.00004112, 0.00001654, -0.00001400, 0.00002590, -0.00003660, 0.00001441, -0.00001169, 0.00001981, -0.00002505, 0.00002481, -0.00001725, 0.00000979, -0.00001238, -0.00000615, -0.00000347, 0.00001908, 0.00000066, -0.00001135, -0.00000489, 0.00001382, -0.00000730, 0.00000154, -0.00000903, 0.00000899, -0.00000480, 0.00000196, 0.00000230, 0.00000546, -0.00000958, -0.00000126, 0.00000026, 0.00000303, -0.00000874, 0.00001447, -0.00001441, 0.00001122, 0.00000315, -0.00000945, 0.00001952, -0.00002189, 0.00001521, -0.00002132, 0.00000757, -0.00000145, 0.00000006, -0.00000820, 0.00000987, -0.00002328, 0.00002773, -0.00003066, 0.00002567, -0.00001233, 0.00001011, -0.00001150, 0.00001558, -0.00002015, 0.00001764, -0.00001428, 0.00000091, 0.00000464, -0.00001074, 0.00001607, },
		{ 0.00001285, -0.00001144, 0.00003520, -0.00002477, 0.00001788, -0.00002046, 0.00000934, -0.00000439, 0.00000514, -0.00000128, -0.00000703, -0.00000476, 0.00000236, -0.00000416, 0.00001356, -0.00000961, 0.00000097, -0.00000584, 0.00001090, -0.00000520, 0.00000573, -0.00001292, 0.00001115, -0.00000763, 0.00000838, -0.00002592, 0.00002517, -0.00000424, 0.00000252, -0.00001614, 0.00002617, -0.00002234, 0.00001273, -0.00000377, 0.00000302, -0.00000887, 0.00000641, -0.00001522, 0.00002419, -0.00002196, 0.00000898, -0.00002037, 0.00003207, -0.00003234, 0.00001857, -0.00002289, 0.00002632, -0.00003488, 0.00001752, -0.00005386, 0.00004097, -0.00009650, 0.00004314, -0.00048439, 0.00096347, -0.00006588, -0.00025316, 0.00006450, -0.00009622, 0.00004037, -0.00004962, 0.00003875, -0.00004785, 0.00003866, -0.00003043, 0.00002823, -0.00004379, 0.00002793, -0.00001239, 0.00001195, -0.00003283, 0.00002175, -0.00001057, 0.00001029, -0.00001545, 0.00002339, -0.00002598, 0.00002110, -0.00003092, 0.00001065, -0.00000339, 0.00002598, -0.00002247, 0.00000085, 0.00000004, 0.00001093, -0.00001245, 0.00000391, -0.00000719, 0.00001102, -0.00001112, 0.00001033, -0.00000156, 0.00000356, -0.00001347, 0.00000632, -0.00000217, -0.00000067, 0.00000016, 0.00000696, -0.00001448, 0.00002620, -0.00002504, 0.00002614, -0.00001870, 0.00000882, -0.00001078, -0.00000507, 0.00000352, 0.00000245, -0.00001224, 0.00001172, -0.00001814, 0.00001040, -0.00000331, -0.00000859, 0.00001916, -0.00001418, 0.00001022, -0.00000806, 0.00000937, -0.00001755, 0.00002358, -0.00003308, 0.00002918, -0.00002740, 0.00002310, -0.00000601, },
		{ -0.00000705, 0.00002290, -0.00000007, -0.00000405, 0.00000142, -0.00001282, 0.00001167, -0.00000804, 0.00000940, -0.00000922, -0.00000632, 0.00000239, -0.00000608, 0.00001129, -0.00000240, -0.00000143, -0.00000769, 0.00000938, -0.00000332, 0.00000738, -0.00001335, 0.00000923, -0.00000791, 0.00001255, -0.00002417, 0.00001174, 0.00000549, 0.00000521, -0.00002014, 0.00001991, -0.00000770, -0.00000005, 0.00000082, 0.00000681, -0.00001540, 0.00001535, -0.00002535, 0.00002728, -0.00001930, 0.00001381, -0.00003130, 0.00003536, -0.00002758, 0.00001908, -0.00003216, 0.00003843, -0.00004676, 0.00004179, -0.00008398, 0.00007694, -0.00013347, 0.00012612, -0.00043752, 0.00034970, 0.00066845, -0.00052131, 0.00014053, -0.00015836, 0.00008079, -0.00009159, 0.00007278, -0.00007285, 0.00005376, -0.00004571, 0.00004531, -0.00005029, 0.00002536, -0.00001909, 0.00002907, -0.00004213, 0.00002075, -0.00001562, 0.00002265, -0.00002871, 0.00003171, -0.00002697, 0.00001887, -0.00002162, -0.00000013, -0.00000589, 0.00002628, -0.00001016, -0.00000711, -0.00000366, 0.00001525, -0.00001087, 0.00000309, -0.00000915, 0.00001069, -0.00000795, 0.00000517, 0.00000023, 0.00000699, -0.00001237, 0.00000108, -0.00000074, 0.00000263, -0.00000731, 0.00001513, -0.00001816, 0.00002034, -0.00000971, 0.00000449, 0.00000625, -0.00001286, 0.00000980, -0.00002107, 0.00000958, -0.00000174, -0.00000223, -0.00000573, 0.00000726, -0.00002110, 0.00002672, -0.00003251, 0.00003081, -0.00001736, 0.00001372, -0.00001447, 0.00001823, -0.00002439, 0.00002383, -0.00002279, 0.00000981, -0.00000572, 0.00000022, 0.00000926, 0.00000653, },
		{ -0.00000108, 0.00002870, -0.00002233, 0.00001772, -0.00002601, 0.00001695, -0.00001114, 0.00000989, -0.00000474, -0.00000755, -0.00000219, -0.00000033, -0.00000021, 0.00001044, -0.00000787, -0.00000131, -0.00000242, 0.00000764, -0.00000139, -0.00000102, -0.00000541, 0.00000509, -0.00000085, -0.00000204, -0.00001606, 0.00002278, -0.00000364, -0.00000377, -0.00000747, 0.00001980, -0.00001840, 0.00001113, -0.00000127, -0.00000305, -0.00000059, -0.00000417, -0.00000298, 0.00001223, -0.00001035, -0.00000435, -0.00000410, 0.00001541, -0.00001657, 0.00000205, -0.00000295, 0.00000221, -0.00000790, -0.00001855, -0.00000891, -0.00001634, -0.00002210, -0.00009493, -0.00029097, 0.00095819, -0.00025650, -0.00011366, -0.00001197, -0.00003713, -0.00000467, -0.00000741, 0.00000198, -0.00001518, 0.00001102, -0.00000554, 0.00000450, -0.00002377, 0.00001242, 0.00000279, -0.00000534, -0.00001712, 0.00000910, 0.00000224, -0.00000412, 0.00000092, 0.00000754, -0.00001369, 0.00001078, -0.00002462, 0.00000553, 0.00000699, 0.00001647, -0.00002135, 0.00000128, 0.00000492, 0.00000514, -0.00000937, 0.00000094, -0.00000207, 0.00000546, -0.00000666, 0.00000713, 0.00000177, -0.00000076, -0.00001046, 0.00000437, -0.00000023, -0.00000367, 0.00000659, -0.00000205, -0.00000310, 0.00001618, -0.00001890, 0.00002444, -0.00002268, 0.00001569, -0.00002020, 0.00000266, -0.00000033, 0.00000460, -0.00001590, 0.00001658, -0.00002456, 0.00001858, -0.00001355, 0.00000213, 0.00001126, -0.00000829, 0.00000430, -0.00000153, 0.00000096, -0.00000816, 0.00001397, -0.00002510, 0.00002140, -0.00002119, 0.00002070, -0.00000148, 0.00000683, },
		{ 0.00003142, -0.00001150, 0.00000775, -0.00001299, -0.00000127, 0.00000452, -0.00000449, 0.00000802, -0.00001120, -0.00000451, 0.00000195, -0.00000564, 0.00001259, -0.00000439, -0.00000081, -0.00000747, 0.00000956, -0.00000262, 0.00000477, -0.00001200, 0.00000988, -0.00000770, 0.00001097, -0.00002644, 0.00002004, 0.00000019, 0.00000422, -0.00002002, 0.00002365, -0.00001255, 0.00000428, -0.00000008, 0.00000406, -0.00001192, 0.00001146, -0.00002210, 0.00002579, -0.00001807, 0.00001101, -0.00002764, 0.00003356, -0.00002691, 0.00001696, -0.00002726, 0.00003180, -0.00003955, 0.00002990, -0.00007132, 0.00006318, -0.00011871, 0.00009729, -0.00045835, 0.00057724, 0.00039700, -0.00042620, 0.00011694, -0.00013785, 0.00006660, -0.00007571, 0.00006129, -0.00006402, 0.00004787, -0.00004026, 0.00004033, -0.00004813, 0.00002447, -0.00001547, 0.00002355, -0.00003913, 0.00001925, -0.00001205, 0.00001769, -0.00002387, 0.00002979, -0.00002901, 0.00002188, -0.00002635, 0.00000230, -0.00000352, 0.00002794, -0.00001807, -0.00000314, -0.00000135, 0.00001422, -0.00001246, 0.00000326, -0.00000732, 0.00000961, -0.00000831, 0.00000603, 0.00000028, 0.00000639, -0.00001333, 0.00000198, -0.00000039, 0.00000064, -0.00000312, 0.00001134, -0.00001604, 0.00002315, -0.00001751, 0.00001552, -0.00000713, -0.00000086, -0.00000046, -0.00001572, 0.00000796, 0.00000046, -0.00000707, -0.00000008, 0.00000118, -0.00001519, 0.00002157, -0.00002995, 0.00003208, -0.00001939, 0.00001449, -0.00001442, 0.00001728, -0.00002435, 0.00002538, -0.00002662, 0.00001376, -0.00001122, 0.00000730, 0.00000614, 0.00000908, -0.00000871, },
		{ 0.00001690, -0.00001335, 0.00000939, -0.00002372, 0.00001912, -0.00001498, 0.00001370, -0.00000881, -0.00000672, -0.00000053, -0.00000247, 0.00000353, 0.00000713, -0.00000608, -0.00000326, 0.00000040, 0.00000541, 0.00000056, -0.00000577, 0.00000100, -0.00000030, 0.00000488, -0.00001210, -0.00000404, 0.00001737, -0.00000231, -0.00000936, 0.00000136, 0.00001246, -0.00001339, 0.00000921, -0.00000022, -0.00000634, 0.00000449, -0.00001150, 0.00000627, 0.00000326, -0.00000206, -0.00001402, 0.00000850, 0.00000228, -0.00000464, -0.00000966, 0.00001081, -0.00001453, 0.00000998, -0.00004372, 0.00002405, -0.00006014, 0.00003520, -0.00022332, -0.00005383, 0.00084805, -0.00035402, -0.00002066, -0.00006312, 0.00000309, -0.00003391, 0.00002120, -0.00002378, 0.00000846, -0.00000915, 0.00001291, -0.00001383, -0.00000819, 0.00000130, 0.00001317, -0.00001808, -0.00000532, 0.00000037, 0.00001085, -0.00001430, 0.00001393, -0.00000687, -0.00000140, -0.00000016, -0.00001805, 0.00000165, 0.00001649, 0.00000509, -0.00001779, 0.00000180, 0.00000849, -0.00000015, -0.00000666, -0.00000072, 0.00000131, 0.00000112, -0.00000286, 0.00000453, 0.00000447, -0.00000473, -0.00000772, 0.00000308, 0.00000048, -0.00000446, 0.00001026, -0.00000884, 0.00000786, 0.00000380, -0.00000790, 0.00001586, -0.00001909, 0.00001579, -0.00002493, 0.00000811, -0.00000263, 0.00000467, -0.00001702, 0.00001907, -0.00002889, 0.00002489, -0.00002247, 0.00001257, 0.00000300, -0.00000235, -0.00000125, 0.00000419, -0.00000650, 0.00000063, 0.00000443, -0.00001696, 0.00001278, -0.00001318, 0.00001671, 0.00000427, -0.00000032, 0.00001224, },
		{ -0.00001614, 0.00001347, -0.00002404, 0.00001031, -0.00000456, 0.00000154, 0.00000431, -0.00001181, -0.00000257, 0.00000067, -0.00000393, 0.00001240, -0.00000519, -0.00000115, -0.00000627, 0.00000883, -0.00000109, 0.00000047, -0.00000770, 0.00000726, -0.00000431, 0.00000509, -0.00002288, 0.00002370, -0.00000327, 0.00000159, -0.00001672, 0.00002391, -0.00001460, 0.00000692, -0.00000045, 0.00000068, -0.00000688, 0.00000504, -0.00001554, 0.00002065, -0.00001301, 0.00000396, -0.00001898, 0.00002618, -0.00002055, 0.00000919, -0.00001641, 0.00001817, -0.00002416, 0.00000759, -0.00004440, 0.00003040, -0.00007786, 0.00002193, -0.00039107, 0.00071027, 0.00015718, -0.00031176, 0.00007130, -0.00010162, 0.00004007, -0.00004897, 0.00003917, -0.00004523, 0.00003275, -0.00002644, 0.00002753, -0.00003885, 0.00001771, -0.00000685, 0.00001297, -0.00003099, 0.00001314, -0.00000416, 0.00000806, -0.00001334, 0.00002104, -0.00002436, 0.00001923, -0.00002692, 0.00000194, 0.00000265, 0.00002413, -0.00002221, 0.00000014, 0.00000154, 0.00001139, -0.00001257, 0.00000284, -0.00000470, 0.00000694, -0.00000679, 0.00000524, 0.00000184, 0.00000416, -0.00001291, 0.00000202, 0.00000052, -0.00000172, 0.00000220, 0.00000471, -0.00000895, 0.00001929, -0.00001834, 0.00002033, -0.00001635, 0.00000976, -0.00001201, -0.00000756, 0.00000413, 0.00000389, -0.00001307, 0.00000739, -0.00000727, -0.00000647, 0.00001298, -0.00002335, 0.00002958, -0.00001849, 0.00001274, -0.00001189, 0.00001346, -0.00002076, 0.00002289, -0.00002626, 0.00001301, -0.00001168, 0.00001001, 0.00000730, 0.00000674, -0.00000362, 0.00003235, },
		{ -0.00000204, -0.00000403, -0.00001469, 0.00001529, -0.00001459, 0.00001506, -0.00001232, -0.00000502, 0.00000011, -0.00000364, 0.00000648, 0.00000424, -0.00000475, -0.00000446, 0.00000227, 0.00000454, 0.00000007, -0.00000735, 0.00000464, -0.00000321, 0.00000747, -0.00001905, 0.00000734, 0.00001073, -0.00000114, -0.00001314, 0.00000885, 0.00000554, -0.00000830, 0.00000738, -0.00000067, -0.00000665, 0.00000568, -0.00001451, 0.00001118, -0.00000123, 0.00000119, -0.00001801, 0.00001515, -0.00000487, 0.00000092, -0.00001395, 0.00001549, -0.00002045, 0.00001439, -0.00005200, 0.00003672, -0.00007875, 0.00005646, -0.00030789, 0.00017956, 0.00067168, -0.00037255, 0.00002470, -0.00008499, 0.00002088, -0.00004403, 0.00003271, -0.00003486, 0.00001943, -0.00001837, 0.00002162, -0.00002371, 0.00000029, -0.00000317, 0.00001664, -0.00002417, 0.00000048, -0.00000241, 0.00001309, -0.00001772, 0.00002043, -0.00001641, 0.00000776, -0.00000928, -0.00001297, 0.00000078, 0.00002262, -0.00000574, -0.00001292, 0.00000266, 0.00001032, -0.00000437, -0.00000458, -0.00000098, 0.00000258, -0.00000130, -0.00000054, 0.00000328, 0.00000578, -0.00000766, -0.00000574, 0.00000275, -0.00000024, -0.00000265, 0.00001005, -0.00001117, 0.00001504, -0.00000696, 0.00000395, 0.00000368, -0.00000979, 0.00000933, -0.00002388, 0.00000984, -0.00000227, 0.00000172, -0.00001444, 0.00001773, -0.00002945, 0.00002752, -0.00002809, 0.00002079, -0.00000394, 0.00000223, -0.00000508, 0.00000768, -0.00001133, 0.00000684, -0.00000280, -0.00001120, 0.00000622, -0.00000638, 0.00001414, 0.00000803, -0.00000477, 0.00002406, 0.00000433, },
		{ 0.00001094, -0.00002814, 0.00001808, -0.00001247, 0.00000802, -0.00000083, -0.00001099, -0.00000102, -0.00000089, -0.00000150, 0.00001119, -0.00000522, -0.00000197, -0.00000471, 0.00000804, -0.00000002, -0.00000368, -0.00000239, 0.00000315, 0.00000043, -0.00000316, -0.00001475, 0.00002284, -0.00000478, -0.00000206, -0.00001121, 0.00002140, -0.00001430, 0.00000819, -0.00000096, -0.00000186, -0.00000242, -0.00000133, -0.00000841, 0.00001480, -0.00000739, -0.00000374, -0.00000917, 0.00001724, -0.00001279, 0.00000053, -0.00000515, 0.00000417, -0.00000858, -0.00001530, -0.00001520, -0.00000678, -0.00002988, -0.00007601, -0.00025359, 0.00074229, -0.00002234, -0.00020795, 0.00002504, -0.00006535, 0.00001410, -0.00002272, 0.00001646, -0.00002527, 0.00001609, -0.00001099, 0.00001287, -0.00002768, 0.00000982, 0.00000220, 0.00000175, -0.00002194, 0.00000653, 0.00000391, -0.00000201, -0.00000134, 0.00000911, -0.00001574, 0.00001278, -0.00002475, 0.00000052, 0.00001055, 0.00001619, -0.00002264, 0.00000267, 0.00000435, 0.00000761, -0.00001180, 0.00000252, -0.00000243, 0.00000400, -0.00000459, 0.00000385, 0.00000393, 0.00000114, -0.00001179, 0.00000192, 0.00000107, -0.00000310, 0.00000661, -0.00000218, 0.00000053, 0.00001051, -0.00001267, 0.00001784, -0.00001876, 0.00001534, -0.00002112, 0.00000044, 0.00000022, 0.00000660, -0.00001809, 0.00001443, -0.00001596, 0.00000310, 0.00000276, -0.00001422, 0.00002442, -0.00001567, 0.00000959, -0.00000823, 0.00000842, -0.00001543, 0.00001822, -0.00002364, 0.00000948, -0.00000876, 0.00000984, 0.00001124, 0.00000118, 0.00000679, 0.00002616, -0.00001337, },
		{ -0.00001774, -0.00000227, 0.00000707, -0.00001022, 0.00001342, -0.00001441, -0.00000306, -0.00000010, -0.00000375, 0.00000836, 0.00000205, -0.00000404, -0.00000487, 0.00000331, 0.00000465, -0.00000242, -0.00000576, 0.00000504, -0.00000301, 0.00000602, -0.00002117, 0.00001574, 0.00000451, -0.00000085, -0.00001446, 0.00001402, 0.00000003, -0.00000381, 0.00000563, -0.00000196, -0.00000483, 0.00000372, -0.00001373, 0.00001221, -0.00000167, -0.00000018, -0.00001662, 0.00001602, -0.00000625, 0.00000053, -0.00001154, 0.00001206, -0.00001672, 0.00000659, -0.00004424, 0.00002909, -0.00007085, 0.00003566, -0.00032761, 0.00036800, 0.00047289, -0.00033863, 0.00003377, -0.00008367, 0.00002089, -0.00003886, 0.00003003, -0.00003330, 0.00001912, -0.00001756, 0.00002127, -0.00002564, 0.00000218, -0.00000167, 0.00001393, -0.00002423, 0.00000089, 0.00000013, 0.00000948, -0.00001448, 0.00001976, -0.00001964, 0.00001217, -0.00001527, -0.00001004, 0.00000345, 0.00002402, -0.00001413, -0.00000779, 0.00000390, 0.00001039, -0.00000731, -0.00000302, -0.00000030, 0.00000212, -0.00000186, 0.00000024, 0.00000349, 0.00000552, -0.00000938, -0.00000449, 0.00000312, -0.00000187, 0.00000110, 0.00000627, -0.00000839, 0.00001643, -0.00001282, 0.00001259, -0.00000776, 0.00000151, -0.00000128, -0.00001805, 0.00000808, 0.00000049, -0.00000382, -0.00000846, 0.00001248, -0.00002587, 0.00002584, -0.00002947, 0.00002566, -0.00000861, 0.00000479, -0.00000674, 0.00000855, -0.00001302, 0.00000974, -0.00000684, -0.00000897, 0.00000331, -0.00000256, 0.00001486, 0.00000771, -0.00000374, 0.00003056, -0.00000456, 0.00000658, },
		{ -0.00002429, 0.00001981, -0.00001676, 0.00001289, -0.00000609, -0.00000915, -0.00000009, -0.00000223, 0.00000108, 0.00000949, -0.00000493, -0.00000282, -0.00000328, 0.00000778, -0.00000040, -0.00000606, 0.00000193, -0.00000043, 0.00000434, -0.00001117, -0.00000439, 0.00001861, -0.00000487, -0.00000578, -0.00000477, 0.00001722, -0.00001235, 0.00000844, -0.00000196, -0.00000276, 0.00000000, -0.00000579, -0.00000284, 0.00001055, -0.00000380, -0.00000917, -0.00000144, 0.00001016, -0.00000721, -0.00000512, 0.00000204, -0.00000489, 0.00000064, -0.00003035, 0.00000547, -0.00003431, 0.00000511, -0.00016648, -0.00007890, 0.00068867, -0.00013305, -0.00013143, -0.00000833, -0.00003927, -0.00000274, -0.00000481, 0.00000044, -0.00001078, 0.00000392, 0.00000068, 0.00000118, -0.00001885, 0.00000454, 0.00000810, -0.00000657, -0.00001537, 0.00000270, 0.00000875, -0.00000879, 0.00000809, -0.00000200, -0.00000643, 0.00000494, -0.00002164, -0.00000007, 0.00001765, 0.00000625, -0.00002015, 0.00000460, 0.00000650, 0.00000363, -0.00001061, 0.00000271, -0.00000125, 0.00000184, -0.00000276, 0.00000282, 0.00000564, -0.00000187, -0.00001059, 0.00000219, 0.00000068, -0.00000258, 0.00000839, -0.00000668, 0.00000911, 0.00000013, -0.00000315, 0.00000950, -0.00001421, 0.00001421, -0.00002523, 0.00000581, -0.00000185, 0.00000691, -0.00002023, 0.00001889, -0.00002262, 0.00001124, -0.00000681, -0.00000459, 0.00001823, -0.00001230, 0.00000642, -0.00000499, 0.00000397, -0.00001043, 0.00001357, -0.00002120, 0.00000583, -0.00000510, 0.00000942, 0.00001512, -0.00000433, 0.00001903, 0.00001589, -0.00000554, 0.00000114, },
		{ 0.00000925, -0.00000256, -0.00000353, 0.00000931, -0.00001483, -0.00000128, -0.00000087, -0.00000295, 0.00000921, 0.00000058, -0.00000388, -0.00000471, 0.00000397, 0.00000488, -0.00000573, -0.00000207, 0.00000292, -0.00000033, 0.00000104, -0.00001817, 0.00001993, -0.00000017, -0.00000171, -0.00001328, 0.00001656, -0.00000359, -0.00000028, 0.00000382, -0.00000316, -0.00000232, 0.00000021, -0.00001076, 0.00001099, 0.00000010, -0.00000416, -0.00001191, 0.00001322, -0.00000409, -0.00000326, -0.00000557, 0.00000438, -0.00000801, -0.00000778, -0.00002700, 0.00000856, -0.00004516, -0.00001881, -0.00027960, 0.00048596, 0.00028929, -0.00028125, 0.00002336, -0.00007033, 0.00001214, -0.00002628, 0.00001991, -0.00002476, 0.00001231, -0.00001079, 0.00001531, -0.00002260, 0.00000023, 0.00000303, 0.00000776, -0.00002082, -0.00000160, 0.00000543, 0.00000258, -0.00000692, 0.00001354, -0.00001727, 0.00001188, -0.00001811, -0.00000878, 0.00000885, 0.00002067, -0.00001906, -0.00000311, 0.00000536, 0.00000899, -0.00000909, -0.00000166, 0.00000058, 0.00000076, -0.00000122, -0.00000002, 0.00000473, 0.00000401, -0.00001013, -0.00000362, 0.00000357, -0.00000338, 0.00000527, 0.00000056, -0.00000162, 0.00001201, -0.00001239, 0.00001526, -0.00001466, 0.00001063, -0.00001240, -0.00000992, 0.00000443, 0.00000420, -0.00001034, -0.00000064, 0.00000462, -0.00001917, 0.00002055, -0.00002694, 0.00002722, -0.00001105, 0.00000560, -0.00000674, 0.00000752, -0.00001233, 0.00001003, -0.00000836, -0.00000972, 0.00000377, -0.00000163, 0.00001883, 0.00000329, 0.00000339, 0.00003007, -0.00000713, 0.00000868, -0.00002697, },
		{ 0.00001554, -0.00001649, 0.00001486, -0.00001031, -0.00000684, 0.00000020, -0.00000306, 0.00000335, 0.00000775, -0.00000468, -0.00000340, -0.00000218, 0.00000813, -0.00000249, -0.00000586, 0.00000393, -0.00000199, 0.00000564, -0.00001651, 0.00000558, 0.00001271, -0.00000436, -0.00000863, 0.00000132, 0.00001253, -0.00000956, 0.00000791, -0.00000329, -0.00000202, 0.00000004, -0.00000760, 0.00000028, 0.00000886, -0.00000335, -0.00001091, 0.00000257, 0.00000665, -0.00000549, -0.00000609, 0.00000332, -0.00000678, 0.00000067, -0.00003351, 0.00001177, -0.00004385, 0.00001306, -0.00022273, 0.00009511, 0.00057910, -0.00018367, -0.00008518, -0.00002490, -0.00002624, -0.00000804, 0.00000226, -0.00000620, -0.00000448, -0.00000113, 0.00000606, -0.00000523, -0.00001432, 0.00000349, 0.00000936, -0.00001048, -0.00001270, 0.00000305, 0.00000882, -0.00001034, 0.00001236, -0.00000935, 0.00000080, -0.00000216, -0.00001895, 0.00000154, 0.00002186, -0.00000343, -0.00001588, 0.00000618, 0.00000759, 0.00000003, -0.00000924, 0.00000339, -0.00000133, 0.00000094, -0.00000186, 0.00000273, 0.00000635, -0.00000432, -0.00000962, 0.00000297, -0.00000068, -0.00000004, 0.00000684, -0.00000702, 0.00001384, -0.00000821, 0.00000642, -0.00000129, -0.00000498, 0.00000708, -0.00002371, 0.00000748, -0.00000129, 0.00000419, -0.00001863, 0.00001950, -0.00002575, 0.00001625, -0.00001393, 0.00000378, 0.00001244, -0.00000948, 0.00000420, -0.00000317, 0.00000132, -0.00000721, 0.00001055, -0.00002079, 0.00000425, -0.00000305, 0.00001122, 0.00001618, -0.00000634, 0.00002897, 0.00000584, 0.00000299, -0.00001227, -0.00001432, },
		{ -0.00001046, 0.00000325, 0.00000395, -0.00001383, 0.00000007, -0.00000188, -0.00000153, 0.00000924, -0.00000035, -0.00000404, -0.00000431, 0.00000475, 0.00000431, -0.00000840, 0.00000206, -0.00000019, 0.00000329, -0.00000577, -0.00001117, 0.00001986, -0.00000294, -0.00000349, -0.00001012, 0.00001672, -0.00000534, 0.00000217, 0.00000180, -0.00000350, -0.00000051, -0.00000312, -0.00000744, 0.00000948, 0.00000186, -0.00000824, -0.00000660, 0.00000954, -0.00000135, -0.00000712, 0.00000006, -0.00000286, -0.00000008, -0.00002142, -0.00000926, -0.00001375, -0.00001656, -0.00008757, -0.00017850, 0.00052755, 0.00014463, -0.00022363, 0.00000881, -0.00005579, 0.00000356, -0.00001430, 0.00000960, -0.00001567, 0.00000465, -0.00000300, 0.00000806, -0.00001834, -0.00000213, 0.00000760, 0.00000142, -0.00001708, -0.00000387, 0.00001026, -0.00000417, 0.00000145, 0.00000489, -0.00001152, 0.00000825, -0.00001873, -0.00000792, 0.00001521, 0.00001375, -0.00002045, 0.00000082, 0.00000670, 0.00000665, -0.00000992, -0.00000022, 0.00000097, -0.00000056, -0.00000024, -0.00000056, 0.00000627, 0.00000183, -0.00001034, -0.00000266, 0.00000347, -0.00000373, 0.00000812, -0.00000469, 0.00000660, 0.00000372, -0.00000652, 0.00001148, -0.00001503, 0.00001443, -0.00002058, -0.00000235, 0.00000097, 0.00000701, -0.00001578, 0.00000679, -0.00000371, -0.00001127, 0.00001337, -0.00002184, 0.00002637, -0.00001204, 0.00000556, -0.00000621, 0.00000599, -0.00001081, 0.00000932, -0.00000916, -0.00001155, 0.00000579, -0.00000179, 0.00002415, -0.00000316, 0.00001476, 0.00002363, -0.00000359, 0.00000307, -0.00002882, 0.00001629, },
		{ -0.00001237, 0.00001372, -0.00001287, -0.00000453, -0.00000006, -0.00000325, 0.00000506, 0.00000624, -0.00000463, -0.00000365, -0.00000129, 0.00000872, -0.00000577, -0.00000325, 0.00000333, -0.00000105, 0.00000361, -0.00001767, 0.00001299, 0.00000682, -0.00000404, -0.00000993, 0.00000611, 0.00000828, -0.00000661, 0.00000676, -0.00000449, -0.00000037, -0.00000171, -0.00000720, 0.00000131, 0.00000930, -0.00000561, -0.00000930, 0.00000311, 0.00000643, -0.00000719, -0.00000313, -0.00000030, -0.00000278, -0.00000704, -0.00002607, 0.00000449, -0.00003544, -0.00000936, -0.00022916, 0.00023599, 0.00044752, -0.00019303, -0.00006228, -0.00002841, -0.00002303, -0.00000475, 0.00000086, -0.00000517, -0.00000515, 0.00000011, 0.00000574, -0.00000682, -0.00001357, 0.00000598, 0.00000669, -0.00001064, -0.00001328, 0.00000684, 0.00000475, -0.00000694, 0.00001104, -0.00001173, 0.00000445, -0.00000732, -0.00001722, 0.00000562, 0.00002203, -0.00001106, -0.00001092, 0.00000759, 0.00000748, -0.00000286, -0.00000774, 0.00000421, -0.00000233, 0.00000118, -0.00000191, 0.00000362, 0.00000592, -0.00000605, -0.00000885, 0.00000398, -0.00000243, 0.00000371, 0.00000254, -0.00000296, 0.00001324, -0.00001179, 0.00001244, -0.00001047, 0.00000518, -0.00000334, -0.00001791, 0.00000597, 0.00000144, -0.00000092, -0.00001380, 0.00001639, -0.00002510, 0.00001762, -0.00001779, 0.00000997, 0.00000775, -0.00000763, 0.00000313, -0.00000284, 0.00000059, -0.00000600, 0.00000945, -0.00002290, 0.00000553, -0.00000355, 0.00001639, 0.00001296, -0.00000267, 0.00003344, -0.00000023, 0.00000768, -0.00002439, -0.00000201, 0.00000737, },
		{ 0.00000814, -0.00000130, -0.00001196, 0.00000094, -0.00000284, 0.00000014, 0.00000877, -0.00000099, -0.00000426, -0.00000389, 0.00000591, 0.00000237, -0.00000922, 0.00000501, -0.00000262, 0.00000597, -0.00001215, -0.00000226, 0.00001655, -0.00000411, -0.00000559, -0.00000587, 0.00001521, -0.00000558, 0.00000356, -0.00000033, -0.00000267, -0.00000021, -0.00000510, -0.00000510, 0.00000910, 0.00000196, -0.00001048, -0.00000284, 0.00000723, -0.00000035, -0.00000856, 0.00000243, -0.00000613, 0.00000265, -0.00002855, 0.00000131, -0.00002771, -0.00000006, -0.00014703, -0.00005076, 0.00050451, 0.00004656, -0.00017877, -0.00000047, -0.00004669, 0.00000066, -0.00000808, 0.00000401, -0.00001057, 0.00000033, 0.00000203, 0.00000290, -0.00001576, -0.00000239, 0.00000958, -0.00000264, -0.00001535, -0.00000362, 0.00001212, -0.00000792, 0.00000738, -0.00000292, -0.00000513, 0.00000321, -0.00001836, -0.00000609, 0.00002046, 0.00000518, -0.00001900, 0.00000400, 0.00000752, 0.00000395, -0.00001007, 0.00000139, 0.00000050, -0.00000117, 0.00000036, -0.00000070, 0.00000743, -0.00000042, -0.00001037, -0.00000135, 0.00000250, -0.00000233, 0.00000836, -0.00000724, 0.00001326, -0.00000527, 0.00000207, 0.00000308, -0.00000931, 0.00001184, -0.00002371, 0.00000250, -0.00000064, 0.00000747, -0.00001849, 0.00001188, -0.00001047, -0.00000421, 0.00000623, -0.00001584, 0.00002438, -0.00001258, 0.00000566, -0.00000625, 0.00000529, -0.00001005, 0.00000933, -0.00001117, -0.00001223, 0.00000704, -0.00000052, 0.00002801, -0.00000837, 0.00002683, 0.00001442, 0.00000318, -0.00000841, -0.00002312, 0.00001717, -0.00001449, },
		{ 0.00001028, -0.00001375, -0.00000250, -0.00000071, -0.00000283, 0.00000612, 0.00000507, -0.00000482, -0.00000367, -0.00000033, 0.00000884, -0.00000916, 0.00000073, 0.00000086, 0.00000177, -0.00000122, -0.00001445, 0.00001667, 0.00000213, -0.00000437, -0.00000943, 0.00000913, 0.00000506, -0.00000399, 0.00000507, -0.00000501, 0.00000119, -0.00000408, -0.00000569, 0.00000143, 0.00001050, -0.00000904, -0.00000593, 0.00000173, 0.00000783, -0.00001032, 0.00000134, -0.00000581, 0.00000343, -0.00001790, -0.00001346, -0.00001010, -0.00001684, -0.00005402, -0.00018510, 0.00032441, 0.00032274, -0.00018198, -0.00005169, -0.00002617, -0.00002341, 0.00000143, -0.00000394, -0.00000082, -0.00000908, 0.00000451, 0.00000237, -0.00000586, -0.00001446, 0.00000979, 0.00000230, -0.00000911, -0.00001500, 0.00001186, -0.00000117, -0.00000072, 0.00000566, -0.00000977, 0.00000450, -0.00001038, -0.00001606, 0.00001136, 0.00001829, -0.00001569, -0.00000606, 0.00000878, 0.00000632, -0.00000492, -0.00000607, 0.00000472, -0.00000361, 0.00000204, -0.00000252, 0.00000515, 0.00000457, -0.00000719, -0.00000804, 0.00000471, -0.00000368, 0.00000726, -0.00000281, 0.00000413, 0.00000767, -0.00000974, 0.00001284, -0.00001485, 0.00001245, -0.00001351, -0.00001035, 0.00000290, 0.00000487, -0.00000683, -0.00000734, 0.00001089, -0.00002168, 0.00001604, -0.00001869, 0.00001401, 0.00000406, -0.00000637, 0.00000260, -0.00000319, 0.00000085, -0.00000588, 0.00000927, -0.00002649, 0.00000880, -0.00000580, 0.00002418, 0.00000608, 0.00000665, 0.00003152, -0.00000065, 0.00000563, -0.00003101, 0.00000845, -0.00000155, -0.00000631, },
		{ -0.00000541, -0.00000976, 0.00000134, -0.00000354, 0.00000174, 0.00000813, -0.00000157, -0.00000440, -0.00000351, 0.00000733, -0.00000091, -0.00000773, 0.00000580, -0.00000318, 0.00000631, -0.00001604, 0.00000621, 0.00001155, -0.00000439, -0.00000727, -0.00000150, 0.00001287, -0.00000486, 0.00000399, -0.00000228, -0.00000095, -0.00000146, -0.00000538, -0.00000416, 0.00001028, -0.00000014, -0.00001010, -0.00000156, 0.00000723, -0.00000197, -0.00000675, 0.00000068, -0.00000440, -0.00000116, -0.00002699, 0.00000133, -0.00002823, -0.00000506, -0.00017725, 0.00007401, 0.00043972, -0.00001066, -0.00014932, -0.00000212, -0.00004427, 0.00000423, -0.00000861, 0.00000441, -0.00001089, 0.00000083, 0.00000285, 0.00000114, -0.00001591, 0.00000026, 0.00000821, -0.00000364, -0.00001635, -0.00000015, 0.00001012, -0.00000743, 0.00000899, -0.00000754, -0.00000036, -0.00000157, -0.00001794, -0.00000246, 0.00002292, -0.00000306, -0.00001570, 0.00000661, 0.00000752, 0.00000139, -0.00000972, 0.00000305, -0.00000080, -0.00000083, 0.00000023, -0.00000005, 0.00000776, -0.00000235, -0.00001040, 0.00000030, 0.00000081, 0.00000066, 0.00000569, -0.00000582, 0.00001591, -0.00001171, 0.00000986, -0.00000656, -0.00000009, 0.00000406, -0.00002162, 0.00000388, 0.00000013, 0.00000522, -0.00001792, 0.00001369, -0.00001443, 0.00000066, 0.00000060, -0.00001028, 0.00002220, -0.00001332, 0.00000643, -0.00000737, 0.00000607, -0.00001085, 0.00001097, -0.00001554, -0.00001028, 0.00000583, 0.00000415, 0.00002800, -0.00000922, 0.00003585, 0.00000635, 0.00000890, -0.00002165, -0.00001235, 0.00001255, -0.00001414, 0.00001013, },
		{ -0.00001338, -0.00000082, -0.00000161, -0.00000192, 0.00000662, 0.00000419, -0.00000516, -0.00000358, 0.00000094, 0.00000786, -0.00001145, 0.00000462, -0.00000214, 0.00000510, -0.00000733, -0.00000791, 0.00001655, -0.00000089, -0.00000535, -0.00000740, 0.00001039, 0.00000302, -0.00000203, 0.00000304, -0.00000451, 0.00000178, -0.00000591, -0.00000431, 0.00000190, 0.00001095, -0.00001186, -0.00000268, 0.00000038, 0.00000880, -0.00001259, 0.00000455, -0.00000986, 0.00000769, -0.00002655, -0.00000243, -0.00002347, 0.00000013, -0.00010463, -0.00010371, 0.00035684, 0.00022244, -0.00016684, -0.00004365, -0.00002491, -0.00002170, 0.00000513, -0.00000712, 0.00000232, -0.00001224, 0.00000852, -0.00000095, -0.00000502, -0.00001456, 0.00001248, -0.00000141, -0.00000821, -0.00001551, 0.00001556, -0.00000619, 0.00000541, -0.00000113, -0.00000537, 0.00000199, -0.00001193, -0.00001454, 0.00001721, 0.00001176, -0.00001728, -0.00000169, 0.00000955, 0.00000449, -0.00000622, -0.00000420, 0.00000453, -0.00000452, 0.00000290, -0.00000309, 0.00000675, 0.00000274, -0.00000800, -0.00000690, 0.00000472, -0.00000364, 0.00000915, -0.00000705, 0.00001176, -0.00000076, -0.00000327, 0.00000769, -0.00001312, 0.00001428, -0.00002034, -0.00000368, 0.00000019, 0.00000735, -0.00001169, -0.00000124, 0.00000491, -0.00001718, 0.00001293, -0.00001774, 0.00001662, 0.00000076, -0.00000496, 0.00000164, -0.00000298, 0.00000071, -0.00000533, 0.00000824, -0.00002958, 0.00001203, -0.00000762, 0.00003227, -0.00000200, 0.00001937, 0.00002474, 0.00000352, -0.00000327, -0.00003007, 0.00001394, -0.00000829, -0.00000055, 0.00000586, },
		{ -0.00000757, 0.00000130, -0.00000383, 0.00000304, 0.00000753, -0.00000225, -0.00000439, -0.00000304, 0.00000857, -0.00000492, -0.00000427, 0.00000434, -0.00000159, 0.00000378, -0.00001621, 0.00001233, 0.00000644, -0.00000450, -0.00000794, 0.00000217, 0.00001047, -0.00000374, 0.00000360, -0.00000365, 0.00000099, -0.00000366, -0.00000441, -0.00000420, 0.00001246, -0.00000382, -0.00000761, -0.00000235, 0.00000903, -0.00000551, -0.00000268, -0.00000395, 0.00000081, -0.00000969, -0.00001866, -0.00000743, -0.00001697, -0.00003199, -0.00016803, 0.00017150, 0.00035876, -0.00004031, -0.00013079, 0.00000092, -0.00004555, 0.00001137, -0.00001335, 0.00000884, -0.00001510, 0.00000500, 0.00000033, 0.00000204, -0.00001798, 0.00000486, 0.00000444, -0.00000245, -0.00001914, 0.00000554, 0.00000520, -0.00000329, 0.00000625, -0.00000815, 0.00000160, -0.00000512, -0.00001773, 0.00000293, 0.00002188, -0.00000947, -0.00001155, 0.00000874, 0.00000660, -0.00000067, -0.00000898, 0.00000448, -0.00000257, 0.00000029, -0.00000057, 0.00000136, 0.00000718, -0.00000384, -0.00001036, 0.00000195, -0.00000097, 0.00000434, 0.00000094, -0.00000059, 0.00001356, -0.00001348, 0.00001382, -0.00001380, 0.00000898, -0.00000606, -0.00001587, 0.00000250, 0.00000270, 0.00000101, -0.00001470, 0.00001253, -0.00001555, 0.00000304, -0.00000304, -0.00000571, 0.00002015, -0.00001426, 0.00000766, -0.00000924, 0.00000798, -0.00001291, 0.00001391, -0.00002205, -0.00000569, 0.00000200, 0.00001269, 0.00002326, -0.00000412, 0.00003935, 0.00000233, 0.00000976, -0.00003195, -0.00000053, 0.00000487, -0.00001049, 0.00000934, -0.00000797, },
		{ 0.00000053, -0.00000260, -0.00000075, 0.00000676, 0.00000344, -0.00000551, -0.00000346, 0.00000257, 0.00000548, -0.00001177, 0.00000711, -0.00000423, 0.00000736, -0.00001275, 0.00000007, 0.00001357, -0.00000241, -0.00000659, -0.00000443, 0.00001031, 0.00000203, -0.00000085, 0.00000095, -0.00000302, 0.00000100, -0.00000652, -0.00000381, 0.00000353, 0.00000967, -0.00001284, -0.00000096, 0.00000046, 0.00000792, -0.00001248, 0.00000474, -0.00001029, 0.00000726, -0.00002926, 0.00000189, -0.00002868, 0.00000457, -0.00014276, -0.00000669, 0.00034343, 0.00015182, -0.00015615, -0.00003289, -0.00002798, -0.00001511, 0.00000356, -0.00000580, 0.00000146, -0.00001195, 0.00000969, -0.00000207, -0.00000609, -0.00001231, 0.00001249, -0.00000284, -0.00000944, -0.00001330, 0.00001621, -0.00000822, 0.00000891, -0.00000667, -0.00000079, -0.00000157, -0.00001276, -0.00001176, 0.00002148, 0.00000415, -0.00001640, 0.00000214, 0.00000962, 0.00000248, -0.00000691, -0.00000219, 0.00000351, -0.00000463, 0.00000324, -0.00000315, 0.00000789, 0.00000086, -0.00000869, -0.00000532, 0.00000388, -0.00000199, 0.00000847, -0.00000831, 0.00001721, -0.00000905, 0.00000488, -0.00000091, -0.00000627, 0.00001021, -0.00002224, 0.00000030, -0.00000080, 0.00000776, -0.00001423, 0.00000295, 0.00000010, -0.00001318, 0.00000980, -0.00001614, 0.00001865, -0.00000279, -0.00000273, -0.00000049, -0.00000128, -0.00000100, -0.00000306, 0.00000480, -0.00003028, 0.00001310, -0.00000653, 0.00003780, -0.00000800, 0.00003201, 0.00001626, 0.00000906, -0.00001635, -0.00002233, 0.00001349, -0.00001118, 0.00000322, 0.00000175, -0.00001231, },
		{ 0.00000085, -0.00000366, 0.00000392, 0.00000711, -0.00000308, -0.00000426, -0.00000231, 0.00000907, -0.00000863, 0.00000010, 0.00000139, 0.00000152, -0.00000102, -0.00001262, 0.00001510, 0.00000231, -0.00000494, -0.00000733, 0.00000468, 0.00000853, -0.00000268, 0.00000256, -0.00000407, 0.00000236, -0.00000588, -0.00000308, -0.00000432, 0.00001452, -0.00000782, -0.00000432, -0.00000390, 0.00001126, -0.00000934, 0.00000166, -0.00000899, 0.00000644, -0.00001906, -0.00000824, -0.00001934, -0.00000123, -0.00007238, -0.00012139, 0.00022850, 0.00028250, -0.00005670, -0.00011615, 0.00000368, -0.00004596, 0.00001763, -0.00001819, 0.00001367, -0.00002007, 0.00001019, -0.00000326, 0.00000365, -0.00002011, 0.00000948, 0.00000017, -0.00000087, -0.00002185, 0.00001142, -0.00000057, 0.00000246, 0.00000070, -0.00000550, 0.00000080, -0.00000730, -0.00001735, 0.00000916, 0.00001765, -0.00001333, -0.00000723, 0.00001033, 0.00000492, -0.00000209, -0.00000794, 0.00000537, -0.00000430, 0.00000172, -0.00000167, 0.00000319, 0.00000588, -0.00000498, -0.00001003, 0.00000317, -0.00000207, 0.00000739, -0.00000416, 0.00000685, 0.00000699, -0.00001021, 0.00001252, -0.00001609, 0.00001455, -0.00001521, -0.00000893, -0.00000007, 0.00000572, -0.00000370, -0.00001033, 0.00000965, -0.00001479, 0.00000362, -0.00000508, -0.00000192, 0.00001792, -0.00001484, 0.00000856, -0.00001083, 0.00000986, -0.00001501, 0.00001677, -0.00002918, 0.00000007, -0.00000296, 0.00002372, 0.00001494, 0.00000637, 0.00003707, 0.00000303, 0.00000395, -0.00003584, 0.00000843, -0.00000274, -0.00000556, 0.00000675, -0.00000916, -0.00000551, },
		{ -0.00000348, 0.00000042, 0.00000677, 0.00000266, -0.00000577, -0.00000328, 0.00000434, 0.00000193, -0.00000989, 0.00000745, -0.00000448, 0.00000741, -0.00001576, 0.00000742, 0.00000913, -0.00000303, -0.00000755, -0.00000126, 0.00000947, 0.00000174, -0.00000044, -0.00000084, -0.00000096, -0.00000096, -0.00000585, -0.00000430, 0.00000635, 0.00000655, -0.00001169, -0.00000115, 0.00000233, 0.00000493, -0.00000980, 0.00000177, -0.00000694, 0.00000188, -0.00002533, -0.00000189, -0.00002331, -0.00000904, -0.00015413, 0.00008298, 0.00030229, 0.00010613, -0.00015099, -0.00001917, -0.00003479, -0.00000441, -0.00000282, -0.00000005, -0.00000363, -0.00000780, 0.00000754, -0.00000057, -0.00000933, -0.00000760, 0.00000970, -0.00000182, -0.00001289, -0.00000827, 0.00001350, -0.00000660, 0.00000858, -0.00000920, 0.00000220, -0.00000490, -0.00001342, -0.00000733, 0.00002297, -0.00000288, -0.00001392, 0.00000547, 0.00000876, 0.00000076, -0.00000717, -0.00000022, 0.00000180, -0.00000387, 0.00000289, -0.00000245, 0.00000826, -0.00000079, -0.00000932, -0.00000343, 0.00000247, 0.00000091, 0.00000526, -0.00000578, 0.00001852, -0.00001441, 0.00001155, -0.00000968, 0.00000293, 0.00000181, -0.00001943, 0.00000113, 0.00000027, 0.00000597, -0.00001420, 0.00000467, -0.00000274, -0.00001060, 0.00000755, -0.00001469, 0.00002054, -0.00000678, 0.00000040, -0.00000386, 0.00000207, -0.00000454, 0.00000128, -0.00000161, -0.00002771, 0.00001086, -0.00000094, 0.00003867, -0.00000915, 0.00004128, 0.00000941, 0.00001204, -0.00002924, -0.00001084, 0.00000835, -0.00001019, 0.00000437, -0.00000128, -0.00001097, 0.00000153, },
		{ -0.00000307, 0.00000442, 0.00000683, -0.00000401, -0.00000403, -0.00000119, 0.00000835, -0.00001103, 0.00000408, -0.00000178, 0.00000489, -0.00000670, -0.00000630, 0.00001453, -0.00000035, -0.00000579, -0.00000556, 0.00000595, 0.00000724, -0.00000197, 0.00000117, -0.00000344, 0.00000259, -0.00000731, -0.00000221, -0.00000363, 0.00001538, -0.00001082, -0.00000163, -0.00000481, 0.00001242, -0.00001179, 0.00000427, -0.00001201, 0.00000947, -0.00002534, -0.00000085, -0.00002783, 0.00000943, -0.00011216, -0.00004992, 0.00024449, 0.00022277, -0.00007024, -0.00009982, 0.00000228, -0.00004200, 0.00001947, -0.00001964, 0.00001563, -0.00002279, 0.00001372, -0.00000560, 0.00000397, -0.00002046, 0.00001222, -0.00000273, -0.00000067, -0.00002265, 0.00001545, -0.00000491, 0.00000738, -0.00000534, -0.00000125, -0.00000188, -0.00000848, -0.00001609, 0.00001483, 0.00001140, -0.00001469, -0.00000305, 0.00001117, 0.00000289, -0.00000292, -0.00000666, 0.00000549, -0.00000549, 0.00000298, -0.00000259, 0.00000497, 0.00000421, -0.00000594, -0.00000923, 0.00000365, -0.00000192, 0.00000863, -0.00000767, 0.00001410, -0.00000164, -0.00000338, 0.00000653, -0.00001279, 0.00001471, -0.00002074, -0.00000316, -0.00000217, 0.00000782, -0.00000741, -0.00000643, 0.00000661, -0.00001353, 0.00000358, -0.00000642, 0.00000169, 0.00001492, -0.00001430, 0.00000820, -0.00001101, 0.00001037, -0.00001564, 0.00001775, -0.00003489, 0.00000478, -0.00000654, 0.00003459, 0.00000573, 0.00001981, 0.00003080, 0.00000673, -0.00000754, -0.00003226, 0.00001210, -0.00000776, -0.00000140, 0.00000366, -0.00000949, -0.00000355, 0.00000004, },
		{ 0.00000140, 0.00000685, 0.00000170, -0.00000587, -0.00000295, 0.00000579, -0.00000213, -0.00000631, 0.00000574, -0.00000277, 0.00000494, -0.00001541, 0.00001251, 0.00000462, -0.00000340, -0.00000773, 0.00000148, 0.00000843, 0.00000173, -0.00000070, -0.00000194, 0.00000100, -0.00000346, -0.00000438, -0.00000528, 0.00000971, 0.00000230, -0.00000904, -0.00000267, 0.00000530, 0.00000067, -0.00000567, -0.00000298, -0.00000148, -0.00000636, -0.00001714, -0.00001120, -0.00001036, -0.00003870, -0.00013315, 0.00014777, 0.00025258, 0.00007563, -0.00014769, -0.00000560, -0.00004211, 0.00000720, -0.00001114, 0.00000777, -0.00001100, -0.00000133, 0.00000332, 0.00000247, -0.00001362, -0.00000166, 0.00000526, 0.00000058, -0.00001742, -0.00000161, 0.00000857, -0.00000225, 0.00000476, -0.00000833, 0.00000276, -0.00000730, -0.00001392, -0.00000164, 0.00002138, -0.00000821, -0.00001061, 0.00000828, 0.00000701, -0.00000037, -0.00000716, 0.00000149, -0.00000025, -0.00000245, 0.00000196, -0.00000108, 0.00000783, -0.00000213, -0.00000978, -0.00000158, 0.00000110, 0.00000410, 0.00000054, 0.00000001, 0.00001518, -0.00001522, 0.00001428, -0.00001548, 0.00001108, -0.00000805, -0.00001359, -0.00000038, 0.00000273, 0.00000277, -0.00001235, 0.00000439, -0.00000384, -0.00000937, 0.00000627, -0.00001351, 0.00002217, -0.00001080, 0.00000386, -0.00000777, 0.00000632, -0.00000917, 0.00000687, -0.00001020, -0.00002249, 0.00000578, 0.00000910, 0.00003444, -0.00000432, 0.00004536, 0.00000615, 0.00000953, -0.00003770, 0.00000034, 0.00000119, -0.00000672, 0.00000332, -0.00000309, -0.00000953, 0.00000211, -0.00000407, },
		{ 0.00000469, 0.00000657, -0.00000496, -0.00000373, 0.00000023, 0.00000629, -0.00001147, 0.00000655, -0.00000396, 0.00000718, -0.00001155, 0.00000104, 0.00001149, -0.00000169, -0.00000681, -0.00000305, 0.00000622, 0.00000654, -0.00000178, -0.00000026, -0.00000195, 0.00000152, -0.00000752, -0.00000226, -0.00000170, 0.00001446, -0.00001203, -0.00000041, -0.00000423, 0.00001166, -0.00001195, 0.00000414, -0.00001176, 0.00000833, -0.00002628, 0.00000034, -0.00002837, 0.00000730, -0.00013689, 0.00002814, 0.00022913, 0.00018183, -0.00008538, -0.00007986, -0.00000427, -0.00003278, 0.00001572, -0.00001625, 0.00001317, -0.00002168, 0.00001413, -0.00000543, 0.00000196, -0.00001812, 0.00001214, -0.00000327, -0.00000276, -0.00002058, 0.00001645, -0.00000635, 0.00000956, -0.00000972, 0.00000269, -0.00000524, -0.00000917, -0.00001343, 0.00001861, 0.00000461, -0.00001409, 0.00000090, 0.00001098, 0.00000101, -0.00000335, -0.00000523, 0.00000482, -0.00000588, 0.00000369, -0.00000297, 0.00000629, 0.00000250, -0.00000684, -0.00000796, 0.00000339, -0.00000041, 0.00000746, -0.00000811, 0.00001877, -0.00000952, 0.00000436, -0.00000190, -0.00000532, 0.00000952, -0.00002153, -0.00000001, -0.00000271, 0.00000823, -0.00000929, -0.00000409, 0.00000460, -0.00001287, 0.00000392, -0.00000788, 0.00000566, 0.00001077, -0.00001222, 0.00000608, -0.00000909, 0.00000865, -0.00001381, 0.00001557, -0.00003754, 0.00000650, -0.00000630, 0.00004245, -0.00000118, 0.00003297, 0.00002343, 0.00001009, -0.00002136, -0.00002288, 0.00001028, -0.00000907, 0.00000069, 0.00000113, -0.00000941, -0.00000169, -0.00000101, -0.00000222, },
		{ 0.00000712, 0.00000049, -0.00000577, -0.00000236, 0.00000646, -0.00000575, -0.00000203, 0.00000274, 0.00000028, 0.00000054, -0.00001180, 0.00001453, 0.00000104, -0.00000397, -0.00000689, 0.00000338, 0.00000757, 0.00000168, -0.00000144, -0.00000212, 0.00000222, -0.00000573, -0.00000284, -0.00000601, 0.00001267, -0.00000199, -0.00000599, -0.00000443, 0.00000819, -0.00000345, -0.00000176, -0.00000747, 0.00000353, -0.00001420, -0.00000874, -0.00002102, 0.00000329, -0.00007564, -0.00008419, 0.00017971, 0.00020883, 0.00005066, -0.00014155, 0.00000401, -0.00004622, 0.00001597, -0.00001783, 0.00001447, -0.00001784, 0.00000505, -0.00000090, 0.00000522, -0.00001722, 0.00000373, 0.00000088, 0.00000273, -0.00002134, 0.00000482, 0.00000333, 0.00000293, -0.00000098, -0.00000491, 0.00000102, -0.00000864, -0.00001389, 0.00000433, 0.00001725, -0.00001140, -0.00000694, 0.00001037, 0.00000464, -0.00000084, -0.00000698, 0.00000272, -0.00000220, -0.00000078, 0.00000079, 0.00000065, 0.00000677, -0.00000323, -0.00000990, -0.00000011, 0.00000040, 0.00000643, -0.00000402, 0.00000736, 0.00000823, -0.00001154, 0.00001220, -0.00001640, 0.00001533, -0.00001632, -0.00000706, -0.00000277, 0.00000545, -0.00000059, -0.00000999, 0.00000331, -0.00000413, -0.00000873, 0.00000539, -0.00001211, 0.00002296, -0.00001406, 0.00000667, -0.00001103, 0.00001013, -0.00001343, 0.00001204, -0.00001917, -0.00001646, -0.00000017, 0.00002180, 0.00002655, 0.00000565, 0.00004442, 0.00000627, 0.00000094, -0.00003915, 0.00000779, -0.00000505, -0.00000279, 0.00000118, -0.00000403, -0.00000794, 0.00000218, -0.00000426, 0.00000205, },
		{ 0.00000616, -0.00000580, -0.00000332, 0.00000170, 0.00000322, -0.00000986, 0.00000697, -0.00000439, 0.00000746, -0.00001413, 0.00000759, 0.00000726, -0.00000221, -0.00000755, -0.00000036, 0.00000589, 0.00000621, -0.00000210, -0.00000134, -0.00000008, -0.00000054, -0.00000660, -0.00000316, 0.00000131, 0.00001189, -0.00001143, -0.00000075, -0.00000214, 0.00000906, -0.00001002, 0.00000154, -0.00000862, 0.00000348, -0.00002220, -0.00000468, -0.00002054, -0.00001018, -0.00013676, 0.00009492, 0.00019728, 0.00015490, -0.00010134, -0.00005782, -0.00001409, -0.00002010, 0.00000783, -0.00000897, 0.00000690, -0.00001707, 0.00001159, -0.00000290, -0.00000212, -0.00001342, 0.00000947, -0.00000167, -0.00000686, -0.00001591, 0.00001453, -0.00000470, 0.00000829, -0.00001119, 0.00000495, -0.00000829, -0.00000960, -0.00000937, 0.00001976, -0.00000141, -0.00001221, 0.00000459, 0.00000963, -0.00000026, -0.00000362, -0.00000377, 0.00000352, -0.00000542, 0.00000374, -0.00000266, 0.00000691, 0.00000097, -0.00000766, -0.00000639, 0.00000271, 0.00000198, 0.00000420, -0.00000506, 0.00001934, -0.00001429, 0.00001022, -0.00000974, 0.00000354, 0.00000083, -0.00001811, 0.00000031, -0.00000154, 0.00000700, -0.00000939, -0.00000346, 0.00000393, -0.00001320, 0.00000503, -0.00000975, 0.00001004, 0.00000560, -0.00000878, 0.00000240, -0.00000523, 0.00000477, -0.00000955, 0.00001010, -0.00003670, 0.00000450, -0.00000093, 0.00004545, -0.00000340, 0.00004312, 0.00001747, 0.00000982, -0.00003316, -0.00001117, 0.00000472, -0.00000717, 0.00000053, -0.00000039, -0.00000915, -0.00000003, -0.00000211, -0.00000133, 0.00000494, },
		{ -0.00000095, -0.00000547, -0.00000149, 0.00000601, -0.00000807, 0.00000176, -0.00000039, 0.00000354, -0.00000451, -0.00000593, 0.00001360, -0.00000121, -0.00000484, -0.00000511, 0.00000434, 0.00000707, 0.00000136, -0.00000241, -0.00000136, 0.00000230, -0.00000712, -0.00000186, -0.00000585, 0.00001442, -0.00000531, -0.00000360, -0.00000537, 0.00000987, -0.00000620, 0.00000048, -0.00000996, 0.00000599, -0.00001883, -0.00000390, -0.00002631, 0.00000992, -0.00010765, -0.00001943, 0.00018089, 0.00017808, 0.00002512, -0.00012965, 0.00000739, -0.00004487, 0.00001949, -0.00002035, 0.00001756, -0.00002183, 0.00000927, -0.00000334, 0.00000614, -0.00001865, 0.00000705, -0.00000194, 0.00000326, -0.00002318, 0.00000938, -0.00000039, 0.00000688, -0.00000661, -0.00000047, -0.00000221, -0.00000918, -0.00001283, 0.00000935, 0.00001173, -0.00001262, -0.00000310, 0.00001141, 0.00000217, -0.00000080, -0.00000669, 0.00000335, -0.00000368, 0.00000074, -0.00000023, 0.00000233, 0.00000537, -0.00000420, -0.00000957, 0.00000075, 0.00000076, 0.00000696, -0.00000671, 0.00001398, -0.00000014, -0.00000501, 0.00000624, -0.00001237, 0.00001435, -0.00002084, -0.00000190, -0.00000465, 0.00000737, -0.00000305, -0.00000835, 0.00000261, -0.00000471, -0.00000771, 0.00000409, -0.00000988, 0.00002226, -0.00001577, 0.00000793, -0.00001255, 0.00001221, -0.00001590, 0.00001508, -0.00002649, -0.00001192, -0.00000433, 0.00003441, 0.00001768, 0.00001843, 0.00004022, 0.00000758, -0.00001173, -0.00003363, 0.00000986, -0.00000836, -0.00000013, -0.00000088, -0.00000458, -0.00000614, 0.00000173, -0.00000405, 0.00000240, 0.00000755, },
		{ -0.00000641, -0.00000277, 0.00000283, -0.00000022, -0.00000676, 0.00000549, -0.00000300, 0.00000553, -0.00001374, 0.00001199, 0.00000306, -0.00000250, -0.00000762, 0.00000199, 0.00000535, 0.00000598, -0.00000286, -0.00000178, 0.00000157, -0.00000299, -0.00000502, -0.00000445, 0.00000477, 0.00000834, -0.00000966, -0.00000202, 0.00000073, 0.00000548, -0.00000705, -0.00000221, -0.00000420, -0.00000303, -0.00001561, -0.00001294, -0.00000813, -0.00003929, -0.00010972, 0.00013819, 0.00016337, 0.00013435, -0.00011451, -0.00003721, -0.00002375, -0.00000730, -0.00000120, -0.00000036, -0.00000107, -0.00001067, 0.00000754, 0.00000069, -0.00000697, -0.00000768, 0.00000544, 0.00000098, -0.00001178, -0.00000988, 0.00001085, -0.00000102, 0.00000421, -0.00000967, 0.00000495, -0.00001049, -0.00000966, -0.00000450, 0.00001832, -0.00000591, -0.00000956, 0.00000781, 0.00000727, -0.00000067, -0.00000393, -0.00000240, 0.00000189, -0.00000435, 0.00000326, -0.00000175, 0.00000681, -0.00000032, -0.00000831, -0.00000482, 0.00000214, 0.00000433, -0.00000003, 0.00000070, 0.00001566, -0.00001481, 0.00001230, -0.00001436, 0.00001068, -0.00000859, -0.00001221, -0.00000138, 0.00000071, 0.00000490, -0.00000852, -0.00000385, 0.00000411, -0.00001412, 0.00000660, -0.00001174, 0.00001433, 0.00000008, -0.00000476, -0.00000192, -0.00000043, -0.00000027, -0.00000397, 0.00000242, -0.00003338, -0.00000033, 0.00000919, 0.00004340, -0.00000010, 0.00004899, 0.00001390, 0.00000414, -0.00003939, -0.00000101, -0.00000178, -0.00000369, -0.00000116, -0.00000106, -0.00000869, 0.00000124, -0.00000301, -0.00000062, 0.00000539, 0.00000542, },
		{ -0.00000496, -0.00000045, 0.00000443, -0.00000860, 0.00000411, -0.00000260, 0.00000586, -0.00000876, 0.00000067, 0.00001053, -0.00000227, -0.00000582, -0.00000269, 0.00000451, 0.00000688, 0.00000066, -0.00000328, 0.00000007, 0.00000119, -0.00000735, -0.00000176, -0.00000454, 0.00001460, -0.00000715, -0.00000242, -0.00000499, 0.00000985, -0.00000706, 0.00000052, -0.00000981, 0.00000510, -0.00001909, -0.00000435, -0.00002445, 0.00000435, -0.00012352, 0.00004564, 0.00016076, 0.00016010, -0.00000245, -0.00011206, 0.00000478, -0.00003819, 0.00001753, -0.00001813, 0.00001627, -0.00002205, 0.00001046, -0.00000324, 0.00000462, -0.00001738, 0.00000769, -0.00000259, 0.00000164, -0.00002238, 0.00001131, -0.00000160, 0.00000822, -0.00001039, 0.00000340, -0.00000595, -0.00000915, -0.00001056, 0.00001248, 0.00000608, -0.00001237, 0.00000085, 0.00001111, 0.00000019, -0.00000055, -0.00000630, 0.00000335, -0.00000446, 0.00000183, -0.00000081, 0.00000362, 0.00000390, -0.00000510, -0.00000881, 0.00000109, 0.00000212, 0.00000542, -0.00000643, 0.00001784, -0.00000743, 0.00000192, -0.00000132, -0.00000507, 0.00000867, -0.00002090, 0.00000077, -0.00000519, 0.00000801, -0.00000415, -0.00000800, 0.00000300, -0.00000623, -0.00000570, 0.00000184, -0.00000645, 0.00001979, -0.00001558, 0.00000724, -0.00001182, 0.00001192, -0.00001581, 0.00001496, -0.00003082, -0.00001058, -0.00000438, 0.00004426, 0.00001069, 0.00003130, 0.00003501, 0.00000706, -0.00002471, -0.00002358, 0.00000711, -0.00000826, 0.00000045, -0.00000211, -0.00000507, -0.00000417, 0.00000087, -0.00000356, 0.00000261, 0.00000797, -0.00000249, },
		{ -0.00000204, 0.00000325, -0.00000319, -0.00000311, 0.00000286, -0.00000039, 0.00000197, -0.00001057, 0.00001361, -0.00000024, -0.00000294, -0.00000680, 0.00000366, 0.00000487, 0.00000563, -0.00000388, -0.00000141, 0.00000245, -0.00000514, -0.00000340, -0.00000551, 0.00000792, 0.00000470, -0.00000759, -0.00000334, 0.00000339, 0.00000208, -0.00000439, -0.00000549, -0.00000049, -0.00000860, -0.00000986, -0.00002011, 0.00000259, -0.00007138, -0.00006172, 0.00015393, 0.00013731, 0.00011351, -0.00012138, -0.00002127, -0.00003011, 0.00000248, -0.00000832, 0.00000679, -0.00000822, -0.00000463, 0.00000384, 0.00000370, -0.00001102, -0.00000250, 0.00000154, 0.00000328, -0.00001605, -0.00000409, 0.00000709, 0.00000298, -0.00000117, -0.00000607, 0.00000289, -0.00001169, -0.00000905, 0.00000019, 0.00001502, -0.00000874, -0.00000637, 0.00001018, 0.00000435, -0.00000027, -0.00000439, -0.00000127, 0.00000025, -0.00000296, 0.00000252, -0.00000052, 0.00000616, -0.00000141, -0.00000867, -0.00000351, 0.00000215, 0.00000568, -0.00000367, 0.00000741, 0.00000898, -0.00001144, 0.00001020, -0.00001445, 0.00001381, -0.00001603, -0.00000598, -0.00000379, 0.00000314, 0.00000289, -0.00000777, -0.00000421, 0.00000419, -0.00001476, 0.00000785, -0.00001313, 0.00001777, -0.00000485, -0.00000121, -0.00000570, 0.00000399, -0.00000500, 0.00000127, -0.00000560, -0.00002956, -0.00000585, 0.00002219, 0.00003773, 0.00000787, 0.00005097, 0.00001172, -0.00000629, -0.00003858, 0.00000483, -0.00000665, -0.00000052, -0.00000324, -0.00000133, -0.00000785, 0.00000195, -0.00000356, -0.00000020, 0.00000612, 0.00000426, -0.00000669, },
		{ 0.00000052, 0.00000206, -0.00000737, 0.00000460, -0.00000324, 0.00000648, -0.00001104, 0.00000645, 0.00000647, -0.00000257, -0.00000656, -0.00000008, 0.00000412, 0.00000688, -0.00000040, -0.00000375, 0.00000171, -0.00000077, -0.00000654, -0.00000245, -0.00000227, 0.00001339, -0.00000761, -0.00000237, -0.00000345, 0.00000834, -0.00000636, -0.00000119, -0.00000762, 0.00000167, -0.00001589, -0.00000920, -0.00001636, -0.00001378, -0.00011693, 0.00009724, 0.00013176, 0.00014991, -0.00002987, -0.00009135, -0.00000144, -0.00002835, 0.00001189, -0.00001249, 0.00001158, -0.00001918, 0.00000913, -0.00000108, 0.00000115, -0.00001391, 0.00000603, -0.00000140, -0.00000170, -0.00001937, 0.00001092, -0.00000042, 0.00000665, -0.00001149, 0.00000561, -0.00000942, -0.00000862, -0.00000737, 0.00001337, 0.00000121, -0.00001111, 0.00000475, 0.00000940, -0.00000082, -0.00000045, -0.00000582, 0.00000283, -0.00000454, 0.00000239, -0.00000085, 0.00000435, 0.00000254, -0.00000591, -0.00000780, 0.00000120, 0.00000394, 0.00000231, -0.00000310, 0.00001783, -0.00001167, 0.00000689, -0.00000782, 0.00000282, 0.00000025, -0.00001724, 0.00000088, -0.00000434, 0.00000760, -0.00000423, -0.00000874, 0.00000438, -0.00000865, -0.00000271, -0.00000132, -0.00000201, 0.00001582, -0.00001380, 0.00000494, -0.00000918, 0.00000951, -0.00001341, 0.00001181, -0.00003207, -0.00001280, 0.00000074, 0.00004973, 0.00000756, 0.00004221, 0.00003025, 0.00000238, -0.00003402, -0.00001267, 0.00000170, -0.00000571, -0.00000081, -0.00000241, -0.00000552, -0.00000222, -0.00000017, -0.00000304, 0.00000291, 0.00000817, -0.00000397, -0.00000427, },
		{ 0.00000275, -0.00000499, 0.00000006, 0.00000009, 0.00000248, -0.00000214, -0.00000553, 0.00001263, -0.00000230, -0.00000365, -0.00000511, 0.00000453, 0.00000460, 0.00000502, -0.00000494, -0.00000027, 0.00000227, -0.00000646, -0.00000227, -0.00000587, 0.00001017, 0.00000171, -0.00000599, -0.00000395, 0.00000503, -0.00000029, -0.00000302, -0.00000715, 0.00000112, -0.00001137, -0.00000750, -0.00002261, 0.00000566, -0.00009608, -0.00000425, 0.00014601, 0.00012272, 0.00008903, -0.00012036, -0.00001155, -0.00003162, 0.00000755, -0.00001165, 0.00001062, -0.00001277, -0.00000056, 0.00000188, 0.00000488, -0.00001308, 0.00000091, -0.00000109, 0.00000421, -0.00001860, 0.00000025, 0.00000461, 0.00000568, -0.00000612, -0.00000175, -0.00000056, -0.00001198, -0.00000756, 0.00000375, 0.00001091, -0.00001015, -0.00000270, 0.00001124, 0.00000155, 0.00000063, -0.00000499, -0.00000044, -0.00000109, -0.00000160, 0.00000183, 0.00000071, 0.00000518, -0.00000237, -0.00000866, -0.00000258, 0.00000296, 0.00000537, -0.00000533, 0.00001302, 0.00000137, -0.00000581, 0.00000505, -0.00001040, 0.00001214, -0.00001977, -0.00000116, -0.00000577, 0.00000503, 0.00000165, -0.00000795, -0.00000366, 0.00000333, -0.00001431, 0.00000801, -0.00001327, 0.00001965, -0.00000841, 0.00000104, -0.00000794, 0.00000690, -0.00000818, 0.00000465, -0.00001216, -0.00002738, -0.00000946, 0.00003550, 0.00003079, 0.00001861, 0.00005039, 0.00000864, -0.00001863, -0.00003183, 0.00000560, -0.00000849, 0.00000103, -0.00000474, -0.00000165, -0.00000654, 0.00000202, -0.00000374, -0.00000005, 0.00000709, 0.00000270, -0.00000662, -0.00000120, },
		{ -0.00000051, -0.00000489, 0.00000343, -0.00000227, 0.00000528, -0.00001082, 0.00001029, 0.00000248, -0.00000262, -0.00000672, 0.00000229, 0.00000346, 0.00000691, -0.00000175, -0.00000358, 0.00000303, -0.00000303, -0.00000510, -0.00000354, 0.00000046, 0.00001134, -0.00000722, -0.00000291, -0.00000138, 0.00000610, -0.00000499, -0.00000353, -0.00000478, -0.00000248, -0.00001157, -0.00001548, -0.00000610, -0.00003978, -0.00008835, 0.00012713, 0.00010485, 0.00014124, -0.00005356, -0.00007101, -0.00000811, -0.00001831, 0.00000531, -0.00000582, 0.00000549, -0.00001486, 0.00000667, 0.00000188, -0.00000302, -0.00000945, 0.00000316, 0.00000065, -0.00000572, -0.00001525, 0.00000929, 0.00000209, 0.00000293, -0.00001002, 0.00000573, -0.00001212, -0.00000747, -0.00000398, 0.00001239, -0.00000249, -0.00000906, 0.00000820, 0.00000657, -0.00000068, -0.00000076, -0.00000524, 0.00000199, -0.00000406, 0.00000252, -0.00000044, 0.00000450, 0.00000135, -0.00000656, -0.00000677, 0.00000151, 0.00000540, -0.00000123, 0.00000231, 0.00001411, -0.00001212, 0.00000851, -0.00001120, 0.00000861, -0.00000833, -0.00001154, -0.00000079, -0.00000259, 0.00000676, -0.00000407, -0.00000976, 0.00000604, -0.00001129, 0.00000065, -0.00000482, 0.00000279, 0.00001109, -0.00001120, 0.00000192, -0.00000558, 0.00000601, -0.00000984, 0.00000683, -0.00003144, -0.00001751, 0.00001053, 0.00005074, 0.00000891, 0.00005043, 0.00002580, -0.00000681, -0.00003710, -0.00000436, -0.00000367, -0.00000243, -0.00000297, -0.00000216, -0.00000572, -0.00000059, -0.00000111, -0.00000270, 0.00000352, 0.00000793, -0.00000519, -0.00000342, 0.00000111, },
		{ -0.00000526, 0.00000195, -0.00000190, 0.00000466, -0.00000558, 0.00000005, 0.00000978, -0.00000325, -0.00000449, -0.00000279, 0.00000462, 0.00000454, 0.00000409, -0.00000578, 0.00000135, 0.00000104, -0.00000672, -0.00000187, -0.00000535, 0.00001129, -0.00000033, -0.00000517, -0.00000359, 0.00000539, -0.00000136, -0.00000314, -0.00000696, 0.00000042, -0.00001101, -0.00000918, -0.00001926, -0.00000200, -0.00010502, 0.00004978, 0.00012342, 0.00011780, 0.00006119, -0.00011240, -0.00000742, -0.00002879, 0.00000813, -0.00001111, 0.00001080, -0.00001425, 0.00000108, 0.00000205, 0.00000391, -0.00001282, 0.00000214, -0.00000204, 0.00000349, -0.00001913, 0.00000270, 0.00000397, 0.00000614, -0.00000933, 0.00000198, -0.00000461, -0.00001138, -0.00000533, 0.00000566, 0.00000692, -0.00001054, 0.00000139, 0.00001067, -0.00000043, 0.00000157, -0.00000561, 0.00000002, -0.00000197, -0.00000048, 0.00000140, 0.00000166, 0.00000408, -0.00000323, -0.00000834, -0.00000192, 0.00000437, 0.00000340, -0.00000425, 0.00001585, -0.00000498, -0.00000010, -0.00000102, -0.00000398, 0.00000651, -0.00001944, 0.00000136, -0.00000670, 0.00000613, 0.00000124, -0.00000922, -0.00000192, 0.00000123, -0.00001240, 0.00000670, -0.00001185, 0.00001972, -0.00001031, 0.00000170, -0.00000833, 0.00000784, -0.00000929, 0.00000546, -0.00001631, -0.00002819, -0.00000915, 0.00004682, 0.00002486, 0.00003023, 0.00004849, 0.00000234, -0.00002903, -0.00002218, 0.00000251, -0.00000744, 0.00000062, -0.00000524, -0.00000218, -0.00000485, 0.00000159, -0.00000369, 0.00000003, 0.00000810, 0.00000089, -0.00000620, -0.00000036, 0.00000140, },
		{ -0.00000201, 0.00000123, -0.00000019, 0.00000274, -0.00000833, 0.00001171, -0.00000068, -0.00000279, -0.00000609, 0.00000409, 0.00000272, 0.00000685, -0.00000327, -0.00000267, 0.00000358, -0.00000498, -0.00000355, -0.00000458, 0.00000310, 0.00000909, -0.00000660, -0.00000338, 0.00000046, 0.00000398, -0.00000392, -0.00000535, -0.00000269, -0.00000547, -0.00000864, -0.00001973, 0.00000098, -0.00006562, -0.00004450, 0.00013403, 0.00008664, 0.00012933, -0.00007063, -0.00005388, -0.00001273, -0.00001055, 0.00000022, -0.00000036, 0.00000004, -0.00001084, 0.00000463, 0.00000426, -0.00000656, -0.00000530, 0.00000023, 0.00000251, -0.00000931, -0.00001122, 0.00000770, 0.00000452, -0.00000160, -0.00000685, 0.00000393, -0.00001381, -0.00000557, -0.00000123, 0.00001033, -0.00000512, -0.00000621, 0.00001064, 0.00000328, 0.00000042, -0.00000156, -0.00000461, 0.00000104, -0.00000324, 0.00000240, 0.00000022, 0.00000418, 0.00000032, -0.00000697, -0.00000588, 0.00000232, 0.00000576, -0.00000383, 0.00000812, 0.00000799, -0.00000933, 0.00000673, -0.00001069, 0.00001058, -0.00001477, -0.00000570, -0.00000315, -0.00000061, 0.00000614, -0.00000452, -0.00001014, 0.00000708, -0.00001327, 0.00000349, -0.00000787, 0.00000714, 0.00000650, -0.00000872, -0.00000077, -0.00000221, 0.00000268, -0.00000657, 0.00000166, -0.00003078, -0.00002263, 0.00002332, 0.00004846, 0.00001411, 0.00005626, 0.00002008, -0.00001861, -0.00003375, -0.00000053, -0.00000693, -0.00000005, -0.00000496, -0.00000188, -0.00000540, 0.00000047, -0.00000177, -0.00000264, 0.00000453, 0.00000714, -0.00000604, -0.00000246, 0.00000108, -0.00000259, },
		{ 0.00000225, -0.00000257, 0.00000553, -0.00000746, 0.00000489, 0.00000606, -0.00000346, -0.00000517, -0.00000019, 0.00000407, 0.00000466, 0.00000283, -0.00000613, 0.00000302, -0.00000092, -0.00000602, -0.00000216, -0.00000408, 0.00001143, -0.00000153, -0.00000497, -0.00000249, 0.00000473, -0.00000149, -0.00000426, -0.00000558, -0.00000167, -0.00000870, -0.00001347, -0.00001189, -0.00001920, -0.00009464, 0.00008996, 0.00009644, 0.00011773, 0.00003268, -0.00010013, -0.00000677, -0.00002356, 0.00000593, -0.00000802, 0.00000837, -0.00001338, 0.00000090, 0.00000381, 0.00000131, -0.00001072, 0.00000154, -0.00000163, 0.00000153, -0.00001809, 0.00000366, 0.00000487, 0.00000438, -0.00001025, 0.00000424, -0.00000858, -0.00000982, -0.00000293, 0.00000601, 0.00000354, -0.00001008, 0.00000557, 0.00000850, -0.00000113, 0.00000209, -0.00000612, 0.00000015, -0.00000234, 0.00000032, 0.00000131, 0.00000221, 0.00000301, -0.00000397, -0.00000786, -0.00000126, 0.00000581, 0.00000040, -0.00000064, 0.00001517, -0.00000853, 0.00000380, -0.00000581, 0.00000237, -0.00000115, -0.00001589, 0.00000160, -0.00000653, 0.00000671, 0.00000120, -0.00001110, 0.00000062, -0.00000179, -0.00000932, 0.00000415, -0.00000912, 0.00001824, -0.00001082, 0.00000111, -0.00000722, 0.00000718, -0.00000872, 0.00000405, -0.00001825, -0.00003206, -0.00000410, 0.00005486, 0.00002142, 0.00004153, 0.00004535, -0.00000809, -0.00003424, -0.00001324, -0.00000203, -0.00000478, -0.00000114, -0.00000496, -0.00000274, -0.00000309, 0.00000092, -0.00000363, 0.00000029, 0.00000889, -0.00000095, -0.00000549, 0.00000028, -0.00000048, -0.00000410, },
		{ -0.00000111, 0.00000222, -0.00000027, -0.00000434, 0.00001089, -0.00000274, -0.00000320, -0.00000463, 0.00000514, 0.00000207, 0.00000660, -0.00000475, -0.00000110, 0.00000310, -0.00000615, -0.00000234, -0.00000523, 0.00000525, 0.00000711, -0.00000622, -0.00000331, 0.00000155, 0.00000250, -0.00000370, -0.00000600, -0.00000209, -0.00000633, -0.00000860, -0.00001972, 0.00000062, -0.00008291, 0.00000431, 0.00012269, 0.00007865, 0.00011239, -0.00008010, -0.00004109, -0.00001433, -0.00000610, -0.00000216, 0.00000263, -0.00000351, -0.00000824, 0.00000397, 0.00000513, -0.00000860, -0.00000236, -0.00000194, 0.00000354, -0.00001179, -0.00000807, 0.00000702, 0.00000572, -0.00000558, -0.00000310, 0.00000073, -0.00001432, -0.00000309, 0.00000030, 0.00000804, -0.00000695, -0.00000253, 0.00001150, 0.00000034, 0.00000201, -0.00000275, -0.00000398, 0.00000018, -0.00000231, 0.00000224, 0.00000089, 0.00000357, -0.00000057, -0.00000713, -0.00000516, 0.00000367, 0.00000464, -0.00000438, 0.00001252, 0.00000136, -0.00000484, 0.00000277, -0.00000698, 0.00000839, -0.00001776, -0.00000117, -0.00000531, 0.00000120, 0.00000604, -0.00000599, -0.00000931, 0.00000688, -0.00001388, 0.00000511, -0.00000980, 0.00001038, 0.00000271, -0.00000699, -0.00000243, 0.00000012, 0.00000042, -0.00000471, -0.00000233, -0.00003186, -0.00002593, 0.00003701, 0.00004457, 0.00002208, 0.00006026, 0.00001102, -0.00002951, -0.00002609, -0.00000105, -0.00000739, 0.00000059, -0.00000607, -0.00000192, -0.00000451, 0.00000091, -0.00000214, -0.00000280, 0.00000587, 0.00000583, -0.00000643, -0.00000149, 0.00000032, -0.00000363, 0.00000039, },
		{ -0.00000187, 0.00000500, -0.00000738, 0.00000810, 0.00000239, -0.00000337, -0.00000538, 0.00000233, 0.00000305, 0.00000493, 0.00000127, -0.00000579, 0.00000426, -0.00000308, -0.00000470, -0.00000290, -0.00000239, 0.00001096, -0.00000228, -0.00000498, -0.00000115, 0.00000363, -0.00000134, -0.00000554, -0.00000406, -0.00000372, -0.00000640, -0.00001769, -0.00000440, -0.00004085, -0.00006713, 0.00011112, 0.00007327, 0.00011740, 0.00000675, -0.00008653, -0.00000720, -0.00001816, 0.00000309, -0.00000430, 0.00000496, -0.00001153, 0.00000005, 0.00000608, -0.00000191, -0.00000773, -0.00000009, -0.00000056, -0.00000089, -0.00001632, 0.00000401, 0.00000635, 0.00000115, -0.00000910, 0.00000470, -0.00001196, -0.00000728, -0.00000108, 0.00000539, 0.00000077, -0.00000869, 0.00000925, 0.00000519, -0.00000046, 0.00000188, -0.00000639, 0.00000001, -0.00000227, 0.00000083, 0.00000149, 0.00000233, 0.00000204, -0.00000454, -0.00000738, -0.00000029, 0.00000660, -0.00000250, 0.00000441, 0.00001138, -0.00000886, 0.00000497, -0.00000784, 0.00000649, -0.00000853, -0.00001066, 0.00000020, -0.00000558, 0.00000720, 0.00000081, -0.00001277, 0.00000318, -0.00000494, -0.00000579, 0.00000095, -0.00000569, 0.00001585, -0.00001057, 0.00000001, -0.00000544, 0.00000578, -0.00000749, 0.00000153, -0.00001918, -0.00003787, 0.00000524, 0.00005953, 0.00002092, 0.00005221, 0.00003970, -0.00002139, -0.00003302, -0.00000772, -0.00000565, -0.00000218, -0.00000322, -0.00000440, -0.00000301, -0.00000159, 0.00000029, -0.00000378, 0.00000096, 0.00000921, -0.00000260, -0.00000458, 0.00000049, -0.00000237, -0.00000198, 0.00000112, },
		{ 0.00000421, -0.00000282, 0.00000006, 0.00000845, -0.00000375, -0.00000376, -0.00000248, 0.00000534, 0.00000161, 0.00000609, -0.00000595, 0.00000085, 0.00000168, -0.00000635, -0.00000169, -0.00000534, 0.00000679, 0.00000554, -0.00000619, -0.00000264, 0.00000181, 0.00000173, -0.00000431, -0.00000557, -0.00000279, -0.00000531, -0.00001120, -0.00001550, -0.00000844, -0.00008585, 0.00004769, 0.00010112, 0.00007846, 0.00009134, -0.00008294, -0.00003215, -0.00001344, -0.00000450, -0.00000205, 0.00000313, -0.00000500, -0.00000724, 0.00000486, 0.00000435, -0.00000893, -0.00000090, -0.00000314, 0.00000365, -0.00001311, -0.00000590, 0.00000744, 0.00000520, -0.00000807, 0.00000021, -0.00000331, -0.00001345, -0.00000053, 0.00000059, 0.00000601, -0.00000817, 0.00000186, 0.00001045, -0.00000147, 0.00000347, -0.00000410, -0.00000344, -0.00000042, -0.00000143, 0.00000219, 0.00000138, 0.00000283, -0.00000134, -0.00000710, -0.00000448, 0.00000526, 0.00000226, -0.00000250, 0.00001422, -0.00000394, -0.00000043, -0.00000157, -0.00000182, 0.00000306, -0.00001718, 0.00000134, -0.00000675, 0.00000276, 0.00000625, -0.00000832, -0.00000731, 0.00000537, -0.00001298, 0.00000524, -0.00001037, 0.00001230, -0.00000009, -0.00000616, -0.00000290, 0.00000122, -0.00000056, -0.00000456, -0.00000461, -0.00003556, -0.00002584, 0.00004991, 0.00004057, 0.00003198, 0.00006224, -0.00000252, -0.00003594, -0.00001756, -0.00000412, -0.00000573, -0.00000036, -0.00000624, -0.00000225, -0.00000320, 0.00000088, -0.00000238, -0.00000297, 0.00000730, 0.00000415, -0.00000640, -0.00000065, -0.00000105, -0.00000340, 0.00000167, -0.00000282, },
		{ 0.00000341, -0.00000556, 0.00000932, -0.00000062, -0.00000331, -0.00000491, 0.00000441, 0.00000173, 0.00000530, -0.00000051, -0.00000467, 0.00000466, -0.00000491, -0.00000319, -0.00000379, -0.00000060, 0.00001027, -0.00000294, -0.00000477, -0.00000008, 0.00000262, -0.00000150, -0.00000627, -0.00000325, -0.00000459, -0.00000577, -0.00001941, -0.00000101, -0.00005995, -0.00002924, 0.00011377, 0.00005824, 0.00011340, -0.00001430, -0.00007375, -0.00000709, -0.00001417, 0.00000123, -0.00000148, 0.00000202, -0.00000993, -0.00000036, 0.00000785, -0.00000476, -0.00000476, -0.00000201, 0.00000056, -0.00000311, -0.00001456, 0.00000456, 0.00000739, -0.00000244, -0.00000659, 0.00000338, -0.00001430, -0.00000394, -0.00000038, 0.00000458, -0.00000164, -0.00000616, 0.00001167, 0.00000158, 0.00000126, 0.00000091, -0.00000639, -0.00000028, -0.00000188, 0.00000116, 0.00000179, 0.00000212, 0.00000119, -0.00000492, -0.00000697, 0.00000113, 0.00000622, -0.00000415, 0.00000933, 0.00000576, -0.00000661, 0.00000363, -0.00000682, 0.00000719, -0.00001378, -0.00000534, -0.00000198, -0.00000420, 0.00000790, -0.00000044, -0.00001352, 0.00000497, -0.00000740, -0.00000262, -0.00000215, -0.00000224, 0.00001320, -0.00001022, -0.00000089, -0.00000381, 0.00000453, -0.00000666, -0.00000083, -0.00002064, -0.00004389, 0.00001765, 0.00006149, 0.00002325, 0.00006228, 0.00002954, -0.00003438, -0.00002658, -0.00000649, -0.00000696, -0.00000085, -0.00000473, -0.00000407, -0.00000272, -0.00000060, -0.00000014, -0.00000422, 0.00000211, 0.00000894, -0.00000390, -0.00000355, 0.00000013, -0.00000372, 0.00000036, -0.00000087, -0.00000018, },
		{ -0.00000424, 0.00000386, 0.00000523, -0.00000402, -0.00000426, 0.00000010, 0.00000470, 0.00000147, 0.00000524, -0.00000662, 0.00000273, -0.00000037, -0.00000564, -0.00000165, -0.00000496, 0.00000783, 0.00000424, -0.00000632, -0.00000161, 0.00000149, 0.00000133, -0.00000527, -0.00000466, -0.00000390, -0.00000367, -0.00001477, -0.00000939, -0.00002388, -0.00007313, 0.00007825, 0.00007746, 0.00008190, 0.00006862, -0.00008118, -0.00002566, -0.00001147, -0.00000441, -0.00000056, 0.00000206, -0.00000511, -0.00000726, 0.00000677, 0.00000237, -0.00000793, -0.00000068, -0.00000358, 0.00000317, -0.00001367, -0.00000428, 0.00000849, 0.00000319, -0.00000878, 0.00000238, -0.00000758, -0.00001108, 0.00000133, 0.00000015, 0.00000429, -0.00000861, 0.00000643, 0.00000766, -0.00000171, 0.00000426, -0.00000533, -0.00000306, -0.00000071, -0.00000069, 0.00000228, 0.00000159, 0.00000207, -0.00000195, -0.00000701, -0.00000359, 0.00000653, -0.00000063, 0.00000137, 0.00001282, -0.00000678, 0.00000248, -0.00000460, 0.00000276, -0.00000360, -0.00001388, 0.00000184, -0.00000738, 0.00000427, 0.00000625, -0.00001089, -0.00000465, 0.00000302, -0.00001092, 0.00000412, -0.00000978, 0.00001309, -0.00000209, -0.00000592, -0.00000258, 0.00000147, -0.00000075, -0.00000564, -0.00000567, -0.00004170, -0.00002179, 0.00006114, 0.00003730, 0.00004358, 0.00006071, -0.00001968, -0.00003580, -0.00001138, -0.00000733, -0.00000342, -0.00000205, -0.00000588, -0.00000257, -0.00000182, 0.00000066, -0.00000272, -0.00000286, 0.00000854, 0.00000234, -0.00000602, -0.00000008, -0.00000273, -0.00000202, 0.00000159, -0.00000345, 0.00000528, },
		{ -0.00000261, 0.00000867, -0.00000267, -0.00000343, -0.00000364, 0.00000576, 0.00000035, 0.00000569, -0.00000233, -0.00000289, 0.00000404, -0.00000597, -0.00000189, -0.00000457, 0.00000109, 0.00000956, -0.00000372, -0.00000414, 0.00000045, 0.00000195, -0.00000216, -0.00000623, -0.00000335, -0.00000403, -0.00000731, -0.00001763, -0.00000427, -0.00007016, 0.00001025, 0.00010271, 0.00005169, 0.00010472, -0.00002976, -0.00006262, -0.00000601, -0.00001196, 0.00000090, -0.00000023, 0.00000023, -0.00000921, 0.00000023, 0.00000854, -0.00000670, -0.00000236, -0.00000378, 0.00000147, -0.00000490, -0.00001311, 0.00000569, 0.00000732, -0.00000545, -0.00000357, 0.00000057, -0.00001514, -0.00000035, -0.00000093, 0.00000410, -0.00000392, -0.00000236, 0.00001216, -0.00000136, 0.00000337, -0.00000063, -0.00000617, -0.00000058, -0.00000131, 0.00000143, 0.00000206, 0.00000169, 0.00000045, -0.00000512, -0.00000659, 0.00000289, 0.00000454, -0.00000378, 0.00001261, -0.00000001, -0.00000310, 0.00000090, -0.00000362, 0.00000452, -0.00001597, -0.00000115, -0.00000426, -0.00000256, 0.00000877, -0.00000262, -0.00001307, 0.00000560, -0.00000868, -0.00000040, -0.00000463, 0.00000078, 0.00001070, -0.00001008, -0.00000123, -0.00000273, 0.00000387, -0.00000681, -0.00000225, -0.00002382, -0.00004847, 0.00003179, 0.00006153, 0.00002834, 0.00007115, 0.00001342, -0.00004313, -0.00001802, -0.00000847, -0.00000600, -0.00000109, -0.00000532, -0.00000413, -0.00000189, -0.00000012, -0.00000041, -0.00000482, 0.00000362, 0.00000812, -0.00000479, -0.00000253, -0.00000084, -0.00000412, 0.00000221, -0.00000299, 0.00000191, 0.00000142, },
		{ 0.00000638, 0.00000195, -0.00000392, -0.00000443, 0.00000278, 0.00000327, 0.00000174, 0.00000398, -0.00000652, 0.00000408, -0.00000255, -0.00000432, -0.00000210, -0.00000424, 0.00000857, 0.00000299, -0.00000630, -0.00000062, 0.00000100, 0.00000088, -0.00000601, -0.00000398, -0.00000444, -0.00000287, -0.00001717, -0.00000485, -0.00004070, -0.00004802, 0.00009317, 0.00005762, 0.00008510, 0.00004670, -0.00007682, -0.00002031, -0.00000972, -0.00000444, 0.00000101, 0.00000057, -0.00000478, -0.00000748, 0.00000894, -0.00000012, -0.00000618, -0.00000127, -0.00000359, 0.00000254, -0.00001399, -0.00000263, 0.00000944, 0.00000037, -0.00000795, 0.00000306, -0.00001147, -0.00000732, 0.00000181, -0.00000023, 0.00000250, -0.00000788, 0.00001034, 0.00000381, -0.00000045, 0.00000409, -0.00000624, -0.00000284, -0.00000069, -0.00000008, 0.00000246, 0.00000153, 0.00000134, -0.00000238, -0.00000696, -0.00000232, 0.00000694, -0.00000300, 0.00000609, 0.00000888, -0.00000699, 0.00000338, -0.00000541, 0.00000514, -0.00000967, -0.00000918, 0.00000078, -0.00000724, 0.00000589, 0.00000556, -0.00001300, -0.00000207, 0.00000056, -0.00000835, 0.00000223, -0.00000844, 0.00001314, -0.00000370, -0.00000579, -0.00000200, 0.00000149, -0.00000090, -0.00000703, -0.00000654, -0.00004930, -0.00001404, 0.00007048, 0.00003517, 0.00005693, 0.00005326, -0.00003738, -0.00002946, -0.00000928, -0.00000883, -0.00000187, -0.00000352, -0.00000556, -0.00000251, -0.00000071, 0.00000048, -0.00000333, -0.00000229, 0.00000936, 0.00000061, -0.00000539, 0.00000004, -0.00000427, 0.00000006, 0.00000030, -0.00000292, 0.00000534, -0.00000425, },
		{ 0.00000667, -0.00000378, -0.00000372, -0.00000162, 0.00000614, -0.00000083, 0.00000594, -0.00000396, -0.00000075, 0.00000249, -0.00000607, -0.00000105, -0.00000514, 0.00000262, 0.00000883, -0.00000455, -0.00000316, 0.00000049, 0.00000153, -0.00000314, -0.00000569, -0.00000390, -0.00000272, -0.00001016, -0.00001336, -0.00001392, -0.00006809, 0.00004353, 0.00008457, 0.00005137, 0.00009223, -0.00004020, -0.00005293, -0.00000449, -0.00001095, 0.00000168, -0.00000029, -0.00000049, -0.00000930, 0.00000180, 0.00000813, -0.00000764, -0.00000067, -0.00000532, 0.00000226, -0.00000641, -0.00001180, 0.00000723, 0.00000606, -0.00000732, -0.00000077, -0.00000333, -0.00001408, 0.00000257, -0.00000221, 0.00000396, -0.00000594, 0.00000239, 0.00001049, -0.00000285, 0.00000515, -0.00000240, -0.00000585, -0.00000078, -0.00000068, 0.00000169, 0.00000219, 0.00000115, -0.00000015, -0.00000523, -0.00000609, 0.00000462, 0.00000197, -0.00000129, 0.00001332, -0.00000448, 0.00000028, -0.00000178, 0.00000019, -0.00000043, -0.00001514, 0.00000135, -0.00000618, -0.00000063, 0.00000943, -0.00000535, -0.00001164, 0.00000520, -0.00000875, 0.00000072, -0.00000633, 0.00000325, 0.00000841, -0.00001011, -0.00000111, -0.00000213, 0.00000367, -0.00000784, -0.00000270, -0.00002909, -0.00005064, 0.00004676, 0.00006015, 0.00003664, 0.00007688, -0.00000824, -0.00004472, -0.00001091, -0.00001142, -0.00000392, -0.00000232, -0.00000524, -0.00000437, -0.00000077, 0.00000004, -0.00000071, -0.00000533, 0.00000525, 0.00000691, -0.00000531, -0.00000164, -0.00000228, -0.00000343, 0.00000305, -0.00000461, 0.00000382, -0.00000032, 0.00000064, },
		{ -0.00000084, -0.00000372, -0.00000402, 0.00000516, 0.00000125, 0.00000246, 0.00000232, -0.00000556, 0.00000450, -0.00000432, -0.00000277, -0.00000284, -0.00000330, 0.00000914, 0.00000162, -0.00000592, 0.00000006, 0.00000061, 0.00000010, -0.00000620, -0.00000387, -0.00000394, -0.00000371, -0.00001697, -0.00000482, -0.00005319, -0.00001682, 0.00009402, 0.00004423, 0.00008575, 0.00002718, -0.00007113, -0.00001546, -0.00000880, -0.00000385, 0.00000187, -0.00000058, -0.00000469, -0.00000725, 0.00001077, -0.00000258, -0.00000416, -0.00000236, -0.00000333, 0.00000192, -0.00001433, -0.00000058, 0.00000967, -0.00000248, -0.00000607, 0.00000212, -0.00001427, -0.00000272, 0.00000067, 0.00000009, 0.00000028, -0.00000556, 0.00001263, -0.00000006, 0.00000181, 0.00000300, -0.00000676, -0.00000276, -0.00000040, 0.00000043, 0.00000263, 0.00000126, 0.00000067, -0.00000262, -0.00000696, -0.00000062, 0.00000616, -0.00000391, 0.00001024, 0.00000362, -0.00000519, 0.00000265, -0.00000411, 0.00000467, -0.00001375, -0.00000438, -0.00000123, -0.00000639, 0.00000757, 0.00000398, -0.00001421, -0.00000009, -0.00000143, -0.00000589, 0.00000008, -0.00000672, 0.00001277, -0.00000515, -0.00000546, -0.00000155, 0.00000168, -0.00000156, -0.00000801, -0.00000826, -0.00005719, -0.00000313, 0.00007793, 0.00003473, 0.00007148, 0.00003768, -0.00005120, -0.00001964, -0.00001091, -0.00000808, -0.00000174, -0.00000419, -0.00000561, -0.00000193, -0.00000003, 0.00000042, -0.00000420, -0.00000123, 0.00000967, -0.00000091, -0.00000459, -0.00000039, -0.00000522, 0.00000221, -0.00000171, -0.00000150, 0.00000461, -0.00000303, 0.00000734, },
		{ -0.00000414, -0.00000403, 0.00000100, 0.00000538, -0.00000148, 0.00000584, -0.00000505, 0.00000128, 0.00000037, -0.00000529, -0.00000077, -0.00000544, 0.00000408, 0.00000794, -0.00000522, -0.00000211, 0.00000031, 0.00000106, -0.00000400, -0.00000517, -0.00000415, -0.00000181, -0.00001261, -0.00000916, -0.00002688, -0.00005425, 0.00006586, 0.00006535, 0.00005416, 0.00007756, -0.00004669, -0.00004415, -0.00000330, -0.00001022, 0.00000274, -0.00000096, -0.00000068, -0.00000973, 0.00000392, 0.00000694, -0.00000778, 0.00000039, -0.00000672, 0.00000315, -0.00000797, -0.00001021, 0.00000865, 0.00000397, -0.00000794, 0.00000125, -0.00000780, -0.00001095, 0.00000389, -0.00000331, 0.00000369, -0.00000717, 0.00000733, 0.00000702, -0.00000255, 0.00000604, -0.00000402, -0.00000556, -0.00000079, -0.00000005, 0.00000195, 0.00000214, 0.00000056, -0.00000058, -0.00000535, -0.00000527, 0.00000581, -0.00000070, 0.00000269, 0.00001135, -0.00000687, 0.00000260, -0.00000331, 0.00000303, -0.00000608, -0.00001207, 0.00000207, -0.00000748, 0.00000160, 0.00000946, -0.00000803, -0.00000978, 0.00000422, -0.00000795, 0.00000090, -0.00000733, 0.00000527, 0.00000618, -0.00001005, -0.00000084, -0.00000163, 0.00000343, -0.00000916, -0.00000280, -0.00003608, -0.00005007, 0.00006214, 0.00005754, 0.00004889, 0.00007628, -0.00003247, -0.00003857, -0.00000780, -0.00001317, -0.00000221, -0.00000358, -0.00000504, -0.00000442, 0.00000030, 0.00000013, -0.00000125, -0.00000550, 0.00000674, 0.00000552, -0.00000551, -0.00000100, -0.00000389, -0.00000179, 0.00000268, -0.00000533, 0.00000517, -0.00000134, 0.00000348, 0.00000391, },
		},
		{ { -0.00000114, -0.00000101, 0.00000451, -0.00000094, -0.00000504, -0.00000012, 0.00000493, -0.00000251, -0.00000338, -0.00000169, 0.00000702, 0.00000151, -0.00001009, -0.00000135, 0.00000529, -0.00000216, -0.00000296, -0.00000268, 0.00000712, -0.00000126, -0.00000457, -0.00000393, 0.00000863, -0.00000230, -0.00000557, -0.00001122, 0.00000390, -0.00001283, -0.00005945, 0.00001014, 0.00008332, 0.00003948, 0.00007644, 0.00001187, -0.00006385, -0.00001486, -0.00000780, -0.00000144, -0.00001916, -0.00000176, -0.00000044, 0.00000147, -0.00001114, -0.00000283, 0.00001362, -0.00000286, -0.00000948, 0.00000038, 0.00000169, 0.00000261, -0.00000488, 0.00000062, 0.00000283, 0.00000057, -0.00000486, 0.00000119, 0.00000073, 0.00000011, -0.00000645, -0.00000125, 0.00000644, 0.00000168, -0.00000355, 0.00000010, -0.00000107, 0.00000010, -0.00000355, 0.00000168, 0.00000644, -0.00000125, -0.00000645, 0.00000011, 0.00000073, 0.00000119, -0.00000486, 0.00000057, 0.00000283, 0.00000062, -0.00000488, 0.00000261, 0.00000169, 0.00000038, -0.00000948, -0.00000286, 0.00001362, -0.00000283, -0.00001114, 0.00000147, -0.00000044, -0.00000176, -0.00001916, -0.00000144, -0.00000780, -0.00001486, -0.00006385, 0.00001187, 0.00007644, 0.00003948, 0.00008332, 0.00001014, -0.00005945, -0.00001283, 0.00000390, -0.00001122, -0.00000557, -0.00000230, 0.00000863, -0.00000393, -0.00000457, -0.00000126, 0.00000712, -0.00000268, -0.00000296, -0.00000216, 0.00000529, -0.00000135, -0.00001009, 0.00000151, 0.00000702, -0.00000169, -0.00000338, -0.00000251, 0.00000493, -0.00000012, -0.00000504, -0.00000094, 0.00000451, -0.00000101, },
		{ -0.00000147, 0.00000194, 0.00000321, -0.00000450, -0.00000312, 0.00000369, 0.00000163, -0.00000404, -0.00000284, 0.00000241, 0.00000722, -0.00000671, -0.00000733, 0.00000479, 0.00000090, -0.00000244, -0.00000392, 0.00000269, 0.00000461, -0.00000413, -0.00000542, 0.00000330, 0.00000504, -0.00000587, -0.00000741, -0.00000763, 0.00000669, -0.00004531, -0.00003440, 0.00006319, 0.00005883, 0.00004748, 0.00006615, -0.00004514, -0.00003930, -0.00000875, -0.00000176, -0.00001261, -0.00001169, 0.00000153, -0.00000050, -0.00000248, -0.00001253, 0.00001003, 0.00000687, -0.00001019, -0.00000316, 0.00000093, 0.00000335, -0.00000173, -0.00000321, 0.00000281, 0.00000215, -0.00000312, -0.00000203, 0.00000173, 0.00000053, -0.00000298, -0.00000601, 0.00000444, 0.00000503, -0.00000241, -0.00000177, -0.00000018, -0.00000050, -0.00000151, -0.00000232, 0.00000583, 0.00000339, -0.00000582, -0.00000308, 0.00000079, 0.00000175, -0.00000246, -0.00000329, 0.00000313, 0.00000212, -0.00000292, -0.00000201, 0.00000370, 0.00000088, -0.00000405, -0.00001045, 0.00000917, 0.00000781, -0.00001152, -0.00000419, 0.00000112, 0.00000131, -0.00001356, -0.00001139, -0.00000141, -0.00000991, -0.00003843, -0.00005275, 0.00007186, 0.00005170, 0.00006687, 0.00006423, -0.00004777, -0.00004030, 0.00000474, -0.00000616, -0.00000827, -0.00000605, 0.00000613, 0.00000264, -0.00000580, -0.00000372, 0.00000404, 0.00000370, -0.00000484, -0.00000256, 0.00000143, 0.00000469, -0.00000880, -0.00000518, 0.00000700, 0.00000248, -0.00000315, -0.00000385, 0.00000171, 0.00000404, -0.00000408, -0.00000396, 0.00000318, 0.00000194, -0.00000162, },
		{ -0.00000097, 0.00000468, -0.00000148, -0.00000461, 0.00000022, 0.00000458, -0.00000265, -0.00000322, -0.00000168, 0.00000739, 0.00000044, -0.00001013, -0.00000001, 0.00000457, -0.00000236, -0.00000212, -0.00000287, 0.00000670, -0.00000082, -0.00000503, -0.00000291, 0.00000833, -0.00000347, -0.00000385, -0.00001283, 0.00000565, -0.00001614, -0.00005540, 0.00002080, 0.00007293, 0.00003726, 0.00006812, 0.00001470, -0.00005981, -0.00001585, -0.00000649, -0.00000302, -0.00001890, 0.00000054, -0.00000280, 0.00000274, -0.00001155, -0.00000167, 0.00001357, -0.00000467, -0.00000808, 0.00000064, 0.00000143, 0.00000264, -0.00000499, 0.00000101, 0.00000263, -0.00000007, -0.00000426, 0.00000136, 0.00000043, 0.00000022, -0.00000678, -0.00000047, 0.00000649, 0.00000105, -0.00000364, 0.00000019, -0.00000112, 0.00000010, -0.00000347, 0.00000234, 0.00000620, -0.00000190, -0.00000609, -0.00000002, 0.00000106, 0.00000093, -0.00000542, 0.00000140, 0.00000280, 0.00000033, -0.00000474, 0.00000251, 0.00000195, 0.00000005, -0.00001080, -0.00000076, 0.00001314, -0.00000366, -0.00001100, 0.00000059, 0.00000165, -0.00000415, -0.00001900, -0.00000021, -0.00000901, -0.00001398, -0.00006753, 0.00000816, 0.00008548, 0.00004233, 0.00009393, -0.00000317, -0.00006125, -0.00001029, 0.00000231, -0.00000958, -0.00000732, -0.00000101, 0.00000867, -0.00000481, -0.00000407, -0.00000177, 0.00000768, -0.00000270, -0.00000369, -0.00000187, 0.00000592, -0.00000285, -0.00000971, 0.00000240, 0.00000661, -0.00000165, -0.00000351, -0.00000240, 0.00000534, -0.00000066, -0.00000530, -0.00000042, 0.00000429, -0.00000103, -0.00000114, },
		{ 0.00000198, 0.00000312, -0.00000483, -0.00000225, 0.00000330, 0.00000156, -0.00000420, -0.00000252, 0.00000246, 0.00000715, -0.00000804, -0.00000580, 0.00000475, 0.00000034, -0.00000209, -0.00000321, 0.00000176, 0.00000515, -0.00000452, -0.00000500, 0.00000404, 0.00000361, -0.00000523, -0.00000712, -0.00000824, 0.00000733, -0.00004858, -0.00002158, 0.00006041, 0.00005188, 0.00004368, 0.00006062, -0.00003828, -0.00003950, -0.00000803, -0.00000186, -0.00001423, -0.00000907, 0.00000090, -0.00000147, -0.00000117, -0.00001315, 0.00001193, 0.00000451, -0.00000970, -0.00000230, 0.00000090, 0.00000305, -0.00000160, -0.00000325, 0.00000329, 0.00000121, -0.00000283, -0.00000166, 0.00000163, 0.00000033, -0.00000299, -0.00000601, 0.00000538, 0.00000419, -0.00000249, -0.00000194, 0.00000010, -0.00000082, -0.00000119, -0.00000222, 0.00000654, 0.00000226, -0.00000545, -0.00000327, 0.00000111, 0.00000169, -0.00000292, -0.00000329, 0.00000409, 0.00000128, -0.00000243, -0.00000243, 0.00000411, 0.00000074, -0.00000497, -0.00001043, 0.00001129, 0.00000539, -0.00001025, -0.00000614, 0.00000325, 0.00000028, -0.00001460, -0.00001064, -0.00000077, -0.00001155, -0.00003686, -0.00006115, 0.00007773, 0.00005635, 0.00007619, 0.00006286, -0.00006089, -0.00003398, 0.00000162, -0.00000387, -0.00000969, -0.00000573, 0.00000682, 0.00000209, -0.00000614, -0.00000333, 0.00000349, 0.00000470, -0.00000592, -0.00000246, 0.00000192, 0.00000442, -0.00001013, -0.00000351, 0.00000654, 0.00000265, -0.00000344, -0.00000367, 0.00000182, 0.00000429, -0.00000507, -0.00000325, 0.00000302, 0.00000197, -0.00000177, -0.00000131, },
		{ 0.00000477, -0.00000199, -0.00000405, 0.00000038, 0.00000429, -0.00000284, -0.00000301, -0.00000159, 0.00000767, -0.00000076, -0.00000987, 0.00000116, 0.00000379, -0.00000242, -0.00000127, -0.00000323, 0.00000641, -0.00000047, -0.00000545, -0.00000174, 0.00000773, -0.00000439, -0.00000233, -0.00001418, 0.00000723, -0.00001977, -0.00004949, 0.00002878, 0.00006328, 0.00003524, 0.00006074, 0.00001650, -0.00005528, -0.00001706, -0.00000496, -0.00000507, -0.00001802, 0.00000250, -0.00000518, 0.00000424, -0.00001215, -0.00000023, 0.00001304, -0.00000618, -0.00000665, 0.00000082, 0.00000122, 0.00000256, -0.00000501, 0.00000145, 0.00000224, -0.00000052, -0.00000368, 0.00000144, 0.00000015, 0.00000027, -0.00000705, 0.00000040, 0.00000636, 0.00000046, -0.00000372, 0.00000035, -0.00000127, 0.00000020, -0.00000340, 0.00000304, 0.00000577, -0.00000239, -0.00000575, -0.00000015, 0.00000137, 0.00000057, -0.00000592, 0.00000240, 0.00000253, 0.00000021, -0.00000463, 0.00000239, 0.00000220, -0.00000037, -0.00001201, 0.00000155, 0.00001215, -0.00000409, -0.00001118, 0.00000022, 0.00000324, -0.00000637, -0.00001860, 0.00000083, -0.00001025, -0.00001306, -0.00007100, 0.00000375, 0.00009496, 0.00004632, 0.00010412, -0.00001883, -0.00006057, -0.00000885, 0.00000114, -0.00000813, -0.00000894, 0.00000029, 0.00000851, -0.00000555, -0.00000355, -0.00000234, 0.00000835, -0.00000297, -0.00000426, -0.00000152, 0.00000646, -0.00000448, -0.00000899, 0.00000305, 0.00000621, -0.00000157, -0.00000360, -0.00000232, 0.00000580, -0.00000140, -0.00000535, 0.00000004, 0.00000404, -0.00000102, -0.00000114, -0.00000088, },
		{ 0.00000290, -0.00000496, -0.00000154, 0.00000293, 0.00000145, -0.00000432, -0.00000220, 0.00000261, 0.00000679, -0.00000909, -0.00000428, 0.00000457, -0.00000023, -0.00000153, -0.00000278, 0.00000096, 0.00000560, -0.00000488, -0.00000452, 0.00000479, 0.00000192, -0.00000419, -0.00000733, -0.00000801, 0.00000660, -0.00004983, -0.00000997, 0.00005660, 0.00004572, 0.00004036, 0.00005522, -0.00003210, -0.00003913, -0.00000765, -0.00000182, -0.00001609, -0.00000580, -0.00000055, -0.00000170, -0.00000036, -0.00001330, 0.00001339, 0.00000219, -0.00000903, -0.00000148, 0.00000081, 0.00000280, -0.00000160, -0.00000305, 0.00000350, 0.00000040, -0.00000246, -0.00000134, 0.00000147, 0.00000019, -0.00000309, -0.00000579, 0.00000616, 0.00000332, -0.00000258, -0.00000202, 0.00000029, -0.00000110, -0.00000086, -0.00000207, 0.00000713, 0.00000111, -0.00000494, -0.00000354, 0.00000148, 0.00000153, -0.00000343, -0.00000309, 0.00000493, 0.00000036, -0.00000176, -0.00000299, 0.00000457, 0.00000051, -0.00000589, -0.00001009, 0.00001312, 0.00000291, -0.00000885, -0.00000814, 0.00000568, -0.00000141, -0.00001485, -0.00001037, 0.00000016, -0.00001361, -0.00003462, -0.00007024, 0.00008362, 0.00006162, 0.00008685, 0.00005847, -0.00007295, -0.00002688, -0.00000239, -0.00000093, -0.00001157, -0.00000495, 0.00000710, 0.00000169, -0.00000645, -0.00000296, 0.00000302, 0.00000559, -0.00000707, -0.00000216, 0.00000236, 0.00000394, -0.00001126, -0.00000178, 0.00000587, 0.00000293, -0.00000371, -0.00000350, 0.00000199, 0.00000438, -0.00000603, -0.00000242, 0.00000275, 0.00000204, -0.00000191, -0.00000115, 0.00000204, },
		{ -0.00000242, -0.00000342, 0.00000040, 0.00000406, -0.00000306, -0.00000276, -0.00000142, 0.00000782, -0.00000202, -0.00000934, 0.00000214, 0.00000295, -0.00000231, -0.00000049, -0.00000369, 0.00000623, -0.00000022, -0.00000582, -0.00000044, 0.00000679, -0.00000497, -0.00000117, -0.00001506, 0.00000834, -0.00002325, -0.00004227, 0.00003421, 0.00005463, 0.00003312, 0.00005447, 0.00001721, -0.00005022, -0.00001859, -0.00000314, -0.00000765, -0.00001639, 0.00000387, -0.00000733, 0.00000578, -0.00001281, 0.00000137, 0.00001209, -0.00000736, -0.00000522, 0.00000092, 0.00000106, 0.00000235, -0.00000488, 0.00000184, 0.00000172, -0.00000078, -0.00000312, 0.00000143, -0.00000008, 0.00000024, -0.00000722, 0.00000132, 0.00000608, -0.00000010, -0.00000376, 0.00000054, -0.00000150, 0.00000040, -0.00000336, 0.00000376, 0.00000516, -0.00000270, -0.00000545, -0.00000026, 0.00000166, 0.00000011, -0.00000631, 0.00000350, 0.00000200, 0.00000028, -0.00000459, 0.00000227, 0.00000241, -0.00000085, -0.00001306, 0.00000401, 0.00001068, -0.00000415, -0.00001167, 0.00000041, 0.00000418, -0.00000819, -0.00001817, 0.00000179, -0.00001165, -0.00001195, -0.00007441, -0.00000116, 0.00010459, 0.00005203, 0.00011303, -0.00003632, -0.00005739, -0.00000877, 0.00000062, -0.00000704, -0.00001026, 0.00000146, 0.00000823, -0.00000617, -0.00000302, -0.00000294, 0.00000909, -0.00000349, -0.00000463, -0.00000115, 0.00000687, -0.00000617, -0.00000796, 0.00000344, 0.00000586, -0.00000145, -0.00000367, -0.00000225, 0.00000625, -0.00000232, -0.00000516, 0.00000040, 0.00000379, -0.00000101, -0.00000115, -0.00000076, 0.00000476, },
		{ -0.00000486, -0.00000101, 0.00000260, 0.00000128, -0.00000438, -0.00000187, 0.00000284, 0.00000615, -0.00000983, -0.00000283, 0.00000430, -0.00000076, -0.00000078, -0.00000263, 0.00000037, 0.00000592, -0.00000517, -0.00000398, 0.00000549, 0.00000007, -0.00000286, -0.00000796, -0.00000706, 0.00000460, -0.00004893, -0.00000012, 0.00005243, 0.00003999, 0.00003764, 0.00004985, -0.00002647, -0.00003836, -0.00000746, -0.00000180, -0.00001799, -0.00000211, -0.00000271, -0.00000119, -0.00000012, -0.00001292, 0.00001435, 0.00000003, -0.00000823, -0.00000070, 0.00000068, 0.00000256, -0.00000169, -0.00000262, 0.00000345, -0.00000024, -0.00000206, -0.00000107, 0.00000127, 0.00000009, -0.00000327, -0.00000536, 0.00000676, 0.00000246, -0.00000268, -0.00000197, 0.00000037, -0.00000129, -0.00000055, -0.00000187, 0.00000756, -0.00000001, -0.00000433, -0.00000386, 0.00000190, 0.00000126, -0.00000394, -0.00000267, 0.00000559, -0.00000056, -0.00000098, -0.00000364, 0.00000506, 0.00000020, -0.00000681, -0.00000942, 0.00001456, 0.00000050, -0.00000744, -0.00001001, 0.00000817, -0.00000356, -0.00001442, -0.00001053, 0.00000127, -0.00001601, -0.00003185, -0.00007985, 0.00008931, 0.00006783, 0.00009871, 0.00005063, -0.00008316, -0.00001963, -0.00000695, 0.00000246, -0.00001379, -0.00000378, 0.00000699, 0.00000143, -0.00000671, -0.00000264, 0.00000266, 0.00000629, -0.00000819, -0.00000172, 0.00000276, 0.00000326, -0.00001210, -0.00000009, 0.00000504, 0.00000329, -0.00000395, -0.00000337, 0.00000224, 0.00000427, -0.00000688, -0.00000153, 0.00000239, 0.00000213, -0.00000205, -0.00000098, 0.00000210, 0.00000259, },
		{ -0.00000280, 0.00000032, 0.00000386, -0.00000331, -0.00000248, -0.00000115, 0.00000779, -0.00000326, -0.00000859, 0.00000295, 0.00000207, -0.00000200, 0.00000013, -0.00000416, 0.00000612, -0.00000004, -0.00000612, 0.00000098, 0.00000551, -0.00000515, -0.00000048, -0.00001531, 0.00000872, -0.00002612, -0.00003433, 0.00003745, 0.00004710, 0.00003074, 0.00004935, 0.00001681, -0.00004459, -0.00002046, -0.00000100, -0.00001078, -0.00001391, 0.00000446, -0.00000899, 0.00000715, -0.00001340, 0.00000302, 0.00001079, -0.00000821, -0.00000381, 0.00000093, 0.00000097, 0.00000198, -0.00000457, 0.00000213, 0.00000116, -0.00000087, -0.00000260, 0.00000135, -0.00000025, 0.00000012, -0.00000724, 0.00000224, 0.00000568, -0.00000065, -0.00000372, 0.00000072, -0.00000177, 0.00000068, -0.00000334, 0.00000448, 0.00000439, -0.00000282, -0.00000521, -0.00000033, 0.00000189, -0.00000045, -0.00000654, 0.00000465, 0.00000125, 0.00000057, -0.00000465, 0.00000219, 0.00000256, -0.00000140, -0.00001388, 0.00000650, 0.00000880, -0.00000387, -0.00001242, 0.00000113, 0.00000438, -0.00000947, -0.00001785, 0.00000279, -0.00001331, -0.00001054, -0.00007786, -0.00000644, 0.00011410, 0.00006007, 0.00011973, -0.00005492, -0.00005187, -0.00001011, 0.00000089, -0.00000645, -0.00001115, 0.00000240, 0.00000788, -0.00000668, -0.00000251, -0.00000350, 0.00000985, -0.00000421, -0.00000475, -0.00000078, 0.00000713, -0.00000788, -0.00000663, 0.00000354, 0.00000558, -0.00000132, -0.00000374, -0.00000214, 0.00000666, -0.00000337, -0.00000475, 0.00000064, 0.00000356, -0.00000098, -0.00000116, -0.00000058, 0.00000464, -0.00000272, },
		{ -0.00000069, 0.00000237, 0.00000102, -0.00000435, -0.00000154, 0.00000312, 0.00000524, -0.00001024, -0.00000149, 0.00000393, -0.00000123, 0.00000010, -0.00000277, 0.00000001, 0.00000607, -0.00000539, -0.00000336, 0.00000607, -0.00000181, -0.00000140, -0.00000882, -0.00000558, 0.00000155, -0.00004595, 0.00000762, 0.00004846, 0.00003436, 0.00003568, 0.00004435, -0.00002121, -0.00003735, -0.00000731, -0.00000197, -0.00001969, 0.00000174, -0.00000539, -0.00000001, -0.00000045, -0.00001200, 0.00001476, -0.00000190, -0.00000734, 0.00000002, 0.00000053, 0.00000230, -0.00000183, -0.00000203, 0.00000313, -0.00000067, -0.00000168, -0.00000084, 0.00000103, 0.00000001, -0.00000350, -0.00000473, 0.00000716, 0.00000161, -0.00000279, -0.00000182, 0.00000031, -0.00000136, -0.00000029, -0.00000158, 0.00000781, -0.00000105, -0.00000367, -0.00000421, 0.00000233, 0.00000089, -0.00000445, -0.00000202, 0.00000600, -0.00000137, -0.00000017, -0.00000433, 0.00000556, -0.00000020, -0.00000771, -0.00000839, 0.00001551, -0.00000173, -0.00000615, -0.00001156, 0.00001047, -0.00000594, -0.00001349, -0.00001098, 0.00000245, -0.00001856, -0.00002874, -0.00008974, 0.00009443, 0.00007545, 0.00011138, 0.00003908, -0.00009083, -0.00001282, -0.00001163, 0.00000602, -0.00001615, -0.00000234, 0.00000654, 0.00000129, -0.00000692, -0.00000239, 0.00000245, 0.00000671, -0.00000918, -0.00000119, 0.00000311, 0.00000235, -0.00001260, 0.00000148, 0.00000413, 0.00000371, -0.00000416, -0.00000327, 0.00000259, 0.00000391, -0.00000754, -0.00000062, 0.00000197, 0.00000223, -0.00000217, -0.00000081, 0.00000216, 0.00000220, -0.00000457, },
		{ 0.00000019, 0.00000367, -0.00000356, -0.00000217, -0.00000079, 0.00000754, -0.00000440, -0.00000769, 0.00000357, 0.00000116, -0.00000146, 0.00000053, -0.00000457, 0.00000602, 0.00000008, -0.00000634, 0.00000247, 0.00000391, -0.00000491, -0.00000031, -0.00001482, 0.00000815, -0.00002795, -0.00002633, 0.00003893, 0.00004062, 0.00002806, 0.00004533, 0.00001539, -0.00003849, -0.00002263, 0.00000142, -0.00001441, -0.00001058, 0.00000414, -0.00000995, 0.00000816, -0.00001378, 0.00000460, 0.00000924, -0.00000877, -0.00000245, 0.00000088, 0.00000094, 0.00000145, -0.00000402, 0.00000223, 0.00000062, -0.00000084, -0.00000214, 0.00000120, -0.00000035, -0.00000011, -0.00000710, 0.00000310, 0.00000517, -0.00000119, -0.00000359, 0.00000085, -0.00000204, 0.00000100, -0.00000331, 0.00000517, 0.00000350, -0.00000277, -0.00000504, -0.00000035, 0.00000206, -0.00000110, -0.00000655, 0.00000576, 0.00000034, 0.00000106, -0.00000482, 0.00000216, 0.00000262, -0.00000202, -0.00001441, 0.00000892, 0.00000661, -0.00000331, -0.00001333, 0.00000231, 0.00000383, -0.00001010, -0.00001775, 0.00000389, -0.00001528, -0.00000879, -0.00008143, -0.00001200, 0.00012330, 0.00007095, 0.00012320, -0.00007372, -0.00004441, -0.00001284, 0.00000196, -0.00000643, -0.00001155, 0.00000303, 0.00000752, -0.00000709, -0.00000203, -0.00000399, 0.00001054, -0.00000509, -0.00000464, -0.00000043, 0.00000721, -0.00000951, -0.00000509, 0.00000338, 0.00000539, -0.00000116, -0.00000381, -0.00000196, 0.00000696, -0.00000450, -0.00000412, 0.00000076, 0.00000334, -0.00000095, -0.00000117, -0.00000037, 0.00000440, -0.00000286, -0.00000225, },
		{ 0.00000224, 0.00000066, -0.00000423, -0.00000120, 0.00000342, 0.00000410, -0.00001032, -0.00000030, 0.00000350, -0.00000159, 0.00000104, -0.00000314, -0.00000010, 0.00000605, -0.00000551, -0.00000266, 0.00000646, -0.00000360, 0.00000005, -0.00000973, -0.00000382, -0.00000223, -0.00004111, 0.00001307, 0.00004512, 0.00002855, 0.00003461, 0.00003861, -0.00001618, -0.00003629, -0.00000704, -0.00000250, -0.00002094, 0.00000546, -0.00000834, 0.00000171, -0.00000130, -0.00001057, 0.00001459, -0.00000354, -0.00000641, 0.00000069, 0.00000039, 0.00000199, -0.00000196, -0.00000132, 0.00000258, -0.00000086, -0.00000135, -0.00000066, 0.00000079, -0.00000006, -0.00000376, -0.00000392, 0.00000735, 0.00000080, -0.00000290, -0.00000156, 0.00000011, -0.00000129, -0.00000011, -0.00000119, 0.00000785, -0.00000195, -0.00000300, -0.00000455, 0.00000276, 0.00000042, -0.00000492, -0.00000115, 0.00000610, -0.00000202, 0.00000060, -0.00000501, 0.00000601, -0.00000066, -0.00000858, -0.00000702, 0.00001593, -0.00000367, -0.00000509, -0.00001263, 0.00001232, -0.00000828, -0.00001225, -0.00001156, 0.00000349, -0.00002106, -0.00002553, -0.00009957, 0.00009854, 0.00008513, 0.00012426, 0.00002384, -0.00009539, -0.00000705, -0.00001603, 0.00000944, -0.00001844, -0.00000077, 0.00000582, 0.00000126, -0.00000707, -0.00000220, 0.00000239, 0.00000681, -0.00000995, -0.00000062, 0.00000342, 0.00000121, -0.00001272, 0.00000285, 0.00000321, 0.00000416, -0.00000433, -0.00000320, 0.00000302, 0.00000328, -0.00000796, 0.00000022, 0.00000152, 0.00000233, -0.00000227, -0.00000062, 0.00000220, 0.00000178, -0.00000412, -0.00000056, },
		{ 0.00000343, -0.00000377, -0.00000184, -0.00000033, 0.00000705, -0.00000538, -0.00000670, 0.00000404, 0.00000025, -0.00000071, 0.00000065, -0.00000483, 0.00000590, 0.00000015, -0.00000648, 0.00000399, 0.00000204, -0.00000427, -0.00000068, -0.00001354, 0.00000652, -0.00002841, -0.00001886, 0.00003920, 0.00003502, 0.00002515, 0.00004229, 0.00001310, -0.00003204, -0.00002499, 0.00000399, -0.00001839, -0.00000647, 0.00000287, -0.00001006, 0.00000865, -0.00001382, 0.00000596, 0.00000756, -0.00000906, -0.00000115, 0.00000077, 0.00000097, 0.00000077, -0.00000325, 0.00000211, 0.00000019, -0.00000073, -0.00000173, 0.00000098, -0.00000038, -0.00000046, -0.00000675, 0.00000387, 0.00000458, -0.00000174, -0.00000334, 0.00000089, -0.00000226, 0.00000133, -0.00000325, 0.00000580, 0.00000252, -0.00000257, -0.00000493, -0.00000031, 0.00000214, -0.00000183, -0.00000631, 0.00000676, -0.00000067, 0.00000170, -0.00000510, 0.00000218, 0.00000261, -0.00000271, -0.00001460, 0.00001115, 0.00000422, -0.00000260, -0.00001425, 0.00000379, 0.00000263, -0.00001010, -0.00001788, 0.00000508, -0.00001753, -0.00000671, -0.00008506, -0.00001785, 0.00013216, 0.00008508, 0.00012248, -0.00009166, -0.00003557, -0.00001671, 0.00000376, -0.00000694, -0.00001143, 0.00000332, 0.00000717, -0.00000738, -0.00000163, -0.00000433, 0.00001106, -0.00000603, -0.00000432, -0.00000009, 0.00000707, -0.00001099, -0.00000340, 0.00000299, 0.00000528, -0.00000097, -0.00000392, -0.00000169, 0.00000709, -0.00000562, -0.00000333, 0.00000075, 0.00000315, -0.00000092, -0.00000117, -0.00000011, 0.00000403, -0.00000281, -0.00000182, 0.00000008, },
		{ 0.00000019, -0.00000400, -0.00000085, 0.00000369, 0.00000280, -0.00001008, 0.00000071, 0.00000300, -0.00000181, 0.00000195, -0.00000371, 0.00000001, 0.00000583, -0.00000553, -0.00000187, 0.00000661, -0.00000519, 0.00000133, -0.00001049, -0.00000207, -0.00000637, -0.00003479, 0.00001626, 0.00004270, 0.00002235, 0.00003451, 0.00003253, -0.00001123, -0.00003533, -0.00000649, -0.00000355, -0.00002151, 0.00000873, -0.00001126, 0.00000378, -0.00000256, -0.00000869, 0.00001388, -0.00000484, -0.00000546, 0.00000128, 0.00000028, 0.00000160, -0.00000203, -0.00000058, 0.00000185, -0.00000082, -0.00000109, -0.00000051, 0.00000055, -0.00000016, -0.00000401, -0.00000297, 0.00000734, 0.00000001, -0.00000298, -0.00000122, -0.00000021, -0.00000107, -0.00000002, -0.00000069, 0.00000765, -0.00000267, -0.00000237, -0.00000484, 0.00000315, -0.00000016, -0.00000530, -0.00000008, 0.00000588, -0.00000242, 0.00000125, -0.00000561, 0.00000637, -0.00000116, -0.00000937, -0.00000533, 0.00001576, -0.00000525, -0.00000431, -0.00001310, 0.00001353, -0.00001033, -0.00001093, -0.00001210, 0.00000422, -0.00002327, -0.00002252, -0.00010899, 0.00010115, 0.00009762, 0.00013650, 0.00000522, -0.00009650, -0.00000279, -0.00001977, 0.00001244, -0.00002043, 0.00000077, 0.00000494, 0.00000130, -0.00000714, -0.00000209, 0.00000249, 0.00000655, -0.00001042, -0.00000006, 0.00000369, -0.00000013, -0.00001241, 0.00000394, 0.00000232, 0.00000462, -0.00000445, -0.00000314, 0.00000351, 0.00000239, -0.00000808, 0.00000096, 0.00000109, 0.00000242, -0.00000234, -0.00000042, 0.00000220, 0.00000137, -0.00000355, -0.00000061, 0.00000221, },
		{ -0.00000392, -0.00000152, 0.00000023, 0.00000629, -0.00000615, -0.00000568, 0.00000436, -0.00000063, 0.00000025, 0.00000044, -0.00000488, 0.00000569, 0.00000022, -0.00000652, 0.00000548, -0.00000002, -0.00000329, -0.00000153, -0.00001154, 0.00000382, -0.00002730, -0.00001244, 0.00003880, 0.00003002, 0.00002220, 0.00004001, 0.00001017, -0.00002544, -0.00002738, 0.00000655, -0.00002254, -0.00000177, 0.00000068, -0.00000925, 0.00000849, -0.00001341, 0.00000701, 0.00000584, -0.00000913, 0.00000009, 0.00000061, 0.00000104, -0.00000003, -0.00000225, 0.00000172, -0.00000009, -0.00000059, -0.00000136, 0.00000072, -0.00000036, -0.00000092, -0.00000619, 0.00000450, 0.00000396, -0.00000229, -0.00000294, 0.00000079, -0.00000237, 0.00000162, -0.00000312, 0.00000631, 0.00000152, -0.00000225, -0.00000487, -0.00000023, 0.00000213, -0.00000263, -0.00000578, 0.00000754, -0.00000167, 0.00000245, -0.00000544, 0.00000223, 0.00000252, -0.00000349, -0.00001438, 0.00001304, 0.00000180, -0.00000185, -0.00001501, 0.00000535, 0.00000092, -0.00000953, -0.00001820, 0.00000628, -0.00001995, -0.00000443, -0.00008862, -0.00002414, 0.00014077, 0.00010268, 0.00011666, -0.00010765, -0.00002608, -0.00002139, 0.00000610, -0.00000789, -0.00001084, 0.00000327, 0.00000683, -0.00000754, -0.00000133, -0.00000448, 0.00001131, -0.00000693, -0.00000385, 0.00000025, 0.00000665, -0.00001223, -0.00000166, 0.00000241, 0.00000526, -0.00000077, -0.00000406, -0.00000128, 0.00000698, -0.00000665, -0.00000242, 0.00000064, 0.00000298, -0.00000087, -0.00000118, 0.00000019, 0.00000355, -0.00000256, -0.00000155, 0.00000003, 0.00000312, },
		{ -0.00000367, -0.00000048, 0.00000390, 0.00000138, -0.00000955, 0.00000154, 0.00000246, -0.00000184, 0.00000276, -0.00000438, 0.00000032, 0.00000543, -0.00000543, -0.00000099, 0.00000649, -0.00000647, 0.00000230, -0.00001092, -0.00000060, -0.00001044, -0.00002745, 0.00001740, 0.00004130, 0.00001567, 0.00003540, 0.00002608, -0.00000630, -0.00003456, -0.00000559, -0.00000522, -0.00002122, 0.00001127, -0.00001385, 0.00000596, -0.00000411, -0.00000648, 0.00001267, -0.00000580, -0.00000454, 0.00000181, 0.00000021, 0.00000109, -0.00000197, 0.00000013, 0.00000100, -0.00000057, -0.00000092, -0.00000039, 0.00000032, -0.00000030, -0.00000420, -0.00000192, 0.00000714, -0.00000074, -0.00000300, -0.00000085, -0.00000063, -0.00000071, -0.00000004, -0.00000009, 0.00000723, -0.00000319, -0.00000183, -0.00000506, 0.00000347, -0.00000083, -0.00000557, 0.00000113, 0.00000534, -0.00000254, 0.00000173, -0.00000608, 0.00000661, -0.00000168, -0.00001007, -0.00000339, 0.00001502, -0.00000643, -0.00000385, -0.00001292, 0.00001392, -0.00001187, -0.00000973, -0.00001242, 0.00000443, -0.00002496, -0.00001997, -0.00011763, 0.00010175, 0.00011377, 0.00014703, -0.00001619, -0.00009407, -0.00000040, -0.00002252, 0.00001477, -0.00002192, 0.00000210, 0.00000398, 0.00000135, -0.00000712, -0.00000202, 0.00000272, 0.00000592, -0.00001053, 0.00000044, 0.00000390, -0.00000164, -0.00001169, 0.00000471, 0.00000154, 0.00000505, -0.00000454, -0.00000308, 0.00000404, 0.00000125, -0.00000789, 0.00000155, 0.00000070, 0.00000246, -0.00000238, -0.00000021, 0.00000214, 0.00000100, -0.00000291, -0.00000080, 0.00000227, -0.00000037, },
		{ -0.00000121, 0.00000086, 0.00000527, -0.00000665, -0.00000468, 0.00000455, -0.00000146, 0.00000137, -0.00000009, -0.00000467, 0.00000534, 0.00000030, -0.00000645, 0.00000687, -0.00000216, -0.00000207, -0.00000276, -0.00000891, 0.00000016, -0.00002453, -0.00000750, 0.00003824, 0.00002532, 0.00001939, 0.00003825, 0.00000685, -0.00001892, -0.00002961, 0.00000889, -0.00002658, 0.00000329, -0.00000227, -0.00000751, 0.00000760, -0.00001247, 0.00000763, 0.00000419, -0.00000902, 0.00000124, 0.00000043, 0.00000111, -0.00000091, -0.00000106, 0.00000106, -0.00000015, -0.00000047, -0.00000103, 0.00000041, -0.00000028, -0.00000148, -0.00000541, 0.00000496, 0.00000332, -0.00000286, -0.00000240, 0.00000052, -0.00000234, 0.00000183, -0.00000289, 0.00000666, 0.00000056, -0.00000186, -0.00000483, -0.00000014, 0.00000203, -0.00000350, -0.00000493, 0.00000802, -0.00000257, 0.00000321, -0.00000580, 0.00000227, 0.00000238, -0.00000435, -0.00001370, 0.00001446, -0.00000053, -0.00000120, -0.00001542, 0.00000676, -0.00000106, -0.00000855, -0.00001857, 0.00000732, -0.00002235, -0.00000217, -0.00009187, -0.00003116, 0.00014946, 0.00012374, 0.00010503, -0.00012061, -0.00001674, -0.00002639, 0.00000869, -0.00000909, -0.00000989, 0.00000294, 0.00000647, -0.00000755, -0.00000118, -0.00000436, 0.00001119, -0.00000768, -0.00000329, 0.00000062, 0.00000592, -0.00001312, 0.00000001, 0.00000172, 0.00000528, -0.00000052, -0.00000426, -0.00000072, 0.00000657, -0.00000751, -0.00000147, 0.00000046, 0.00000281, -0.00000080, -0.00000120, 0.00000051, 0.00000295, -0.00000211, -0.00000149, 0.00000009, 0.00000268, -0.00000397, },
		{ -0.00000008, 0.00000400, -0.00000008, -0.00000878, 0.00000217, 0.00000189, -0.00000166, 0.00000338, -0.00000508, 0.00000078, 0.00000487, -0.00000523, -0.00000005, 0.00000606, -0.00000737, 0.00000286, -0.00001085, 0.00000034, -0.00001400, -0.00001966, 0.00001684, 0.00004087, 0.00000855, 0.00003719, 0.00001931, -0.00000135, -0.00003404, -0.00000429, -0.00000753, -0.00001994, 0.00001286, -0.00001583, 0.00000801, -0.00000575, -0.00000407, 0.00001105, -0.00000641, -0.00000365, 0.00000226, 0.00000018, 0.00000046, -0.00000174, 0.00000071, 0.00000013, -0.00000015, -0.00000085, -0.00000029, 0.00000012, -0.00000050, -0.00000431, -0.00000081, 0.00000677, -0.00000146, -0.00000296, -0.00000048, -0.00000112, -0.00000023, -0.00000015, 0.00000060, 0.00000658, -0.00000349, -0.00000139, -0.00000517, 0.00000369, -0.00000159, -0.00000568, 0.00000242, 0.00000449, -0.00000235, 0.00000200, -0.00000639, 0.00000669, -0.00000223, -0.00001062, -0.00000127, 0.00001374, -0.00000721, -0.00000368, -0.00001210, 0.00001343, -0.00001274, -0.00000882, -0.00001238, 0.00000400, -0.00002596, -0.00001811, -0.00012517, 0.00009985, 0.00013443, 0.00015463, -0.00003952, -0.00008826, -0.00000005, -0.00002408, 0.00001622, -0.00002274, 0.00000311, 0.00000303, 0.00000139, -0.00000701, -0.00000197, 0.00000301, 0.00000495, -0.00001027, 0.00000084, 0.00000404, -0.00000328, -0.00001058, 0.00000513, 0.00000089, 0.00000544, -0.00000460, -0.00000300, 0.00000453, -0.00000009, -0.00000739, 0.00000195, 0.00000037, 0.00000246, -0.00000237, 0.00000001, 0.00000201, 0.00000072, -0.00000228, -0.00000108, 0.00000238, -0.00000100, -0.00000324, },
		{ 0.00000156, 0.00000400, -0.00000687, -0.00000375, 0.00000463, -0.00000219, 0.00000260, -0.00000094, -0.00000417, 0.00000483, 0.00000043, -0.00000627, 0.00000810, -0.00000429, -0.00000074, -0.00000422, -0.00000586, -0.00000428, -0.00002019, -0.00000427, 0.00003795, 0.00002063, 0.00001695, 0.00003674, 0.00000343, -0.00001274, -0.00003145, 0.00001078, -0.00003023, 0.00000841, -0.00000579, -0.00000494, 0.00000599, -0.00001098, 0.00000778, 0.00000270, -0.00000876, 0.00000231, 0.00000025, 0.00000116, -0.00000181, 0.00000027, 0.00000016, 0.00000005, -0.00000043, -0.00000071, 0.00000006, -0.00000016, -0.00000213, -0.00000441, 0.00000522, 0.00000268, -0.00000343, -0.00000171, 0.00000007, -0.00000212, 0.00000190, -0.00000253, 0.00000680, -0.00000030, -0.00000147, -0.00000476, -0.00000005, 0.00000185, -0.00000440, -0.00000374, 0.00000812, -0.00000326, 0.00000388, -0.00000609, 0.00000226, 0.00000222, -0.00000531, -0.00001252, 0.00001531, -0.00000260, -0.00000080, -0.00001530, 0.00000774, -0.00000303, -0.00000740, -0.00001880, 0.00000800, -0.00002447, -0.00000022, -0.00009445, -0.00003933, 0.00015877, 0.00014794, 0.00008712, -0.00012955, -0.00000840, -0.00003117, 0.00001117, -0.00001030, -0.00000876, 0.00000242, 0.00000605, -0.00000735, -0.00000120, -0.00000394, 0.00001060, -0.00000814, -0.00000272, 0.00000104, 0.00000484, -0.00001358, 0.00000151, 0.00000100, 0.00000531, -0.00000022, -0.00000453, 0.00000001, 0.00000581, -0.00000809, -0.00000056, 0.00000025, 0.00000261, -0.00000070, -0.00000122, 0.00000086, 0.00000228, -0.00000146, -0.00000164, 0.00000028, 0.00000210, -0.00000389, -0.00000092, },
		{ 0.00000394, -0.00000150, -0.00000782, 0.00000262, 0.00000130, -0.00000124, 0.00000373, -0.00000570, 0.00000131, 0.00000419, -0.00000490, 0.00000094, 0.00000535, -0.00000787, 0.00000293, -0.00001019, 0.00000055, -0.00001668, -0.00001198, 0.00001504, 0.00004121, 0.00000112, 0.00003971, 0.00001234, 0.00000354, -0.00003373, -0.00000265, -0.00001045, -0.00001765, 0.00001332, -0.00001695, 0.00000969, -0.00000732, -0.00000161, 0.00000911, -0.00000671, -0.00000281, 0.00000265, 0.00000020, -0.00000030, -0.00000131, 0.00000110, -0.00000068, 0.00000039, -0.00000087, -0.00000021, -0.00000005, -0.00000077, -0.00000430, 0.00000030, 0.00000625, -0.00000215, -0.00000281, -0.00000016, -0.00000161, 0.00000033, -0.00000032, 0.00000135, 0.00000574, -0.00000356, -0.00000107, -0.00000516, 0.00000379, -0.00000242, -0.00000557, 0.00000370, 0.00000341, -0.00000188, 0.00000203, -0.00000651, 0.00000659, -0.00000279, -0.00001097, 0.00000094, 0.00001199, -0.00000764, -0.00000372, -0.00001070, 0.00001205, -0.00001285, -0.00000830, -0.00001190, 0.00000283, -0.00002614, -0.00001712, -0.00013138, 0.00009509, 0.00016043, 0.00015798, -0.00006366, -0.00007950, -0.00000171, -0.00002436, 0.00001670, -0.00002279, 0.00000369, 0.00000217, 0.00000140, -0.00000681, -0.00000191, 0.00000330, 0.00000371, -0.00000964, 0.00000114, 0.00000406, -0.00000496, -0.00000912, 0.00000519, 0.00000042, 0.00000576, -0.00000463, -0.00000284, 0.00000493, -0.00000156, -0.00000658, 0.00000214, 0.00000012, 0.00000241, -0.00000232, 0.00000024, 0.00000180, 0.00000056, -0.00000171, -0.00000140, 0.00000251, -0.00000167, -0.00000272, 0.00000034, },
		{ 0.00000252, -0.00000678, -0.00000295, 0.00000461, -0.00000278, 0.00000386, -0.00000206, -0.00000338, 0.00000412, 0.00000062, -0.00000596, 0.00000910, -0.00000629, 0.00000057, -0.00000574, -0.00000261, -0.00000919, -0.00001450, -0.00000284, 0.00003822, 0.00001573, 0.00001508, 0.00003524, 0.00000018, -0.00000714, -0.00003270, 0.00001198, -0.00003318, 0.00001324, -0.00000957, -0.00000171, 0.00000372, -0.00000896, 0.00000742, 0.00000142, -0.00000841, 0.00000328, 0.00000007, 0.00000115, -0.00000267, 0.00000167, -0.00000097, 0.00000050, -0.00000049, -0.00000039, -0.00000032, -0.00000001, -0.00000283, -0.00000321, 0.00000527, 0.00000207, -0.00000399, -0.00000090, -0.00000058, -0.00000167, 0.00000180, -0.00000198, 0.00000669, -0.00000100, -0.00000112, -0.00000462, -0.00000000, 0.00000159, -0.00000532, -0.00000222, 0.00000778, -0.00000363, 0.00000437, -0.00000625, 0.00000212, 0.00000208, -0.00000638, -0.00001081, 0.00001547, -0.00000428, -0.00000075, -0.00001448, 0.00000805, -0.00000472, -0.00000632, -0.00001867, 0.00000804, -0.00002599, 0.00000103, -0.00009592, -0.00004922, 0.00016942, 0.00017466, 0.00006278, -0.00013364, -0.00000187, -0.00003517, 0.00001314, -0.00001123, -0.00000764, 0.00000184, 0.00000549, -0.00000690, -0.00000145, -0.00000318, 0.00000948, -0.00000820, -0.00000224, 0.00000156, 0.00000338, -0.00001352, 0.00000273, 0.00000034, 0.00000532, 0.00000015, -0.00000487, 0.00000092, 0.00000466, -0.00000833, 0.00000024, 0.00000007, 0.00000237, -0.00000054, -0.00000127, 0.00000122, 0.00000154, -0.00000066, -0.00000201, 0.00000064, 0.00000136, -0.00000365, -0.00000066, 0.00000229, },
		{ -0.00000283, -0.00000673, 0.00000290, 0.00000071, -0.00000060, 0.00000376, -0.00000616, 0.00000184, 0.00000342, -0.00000447, 0.00000194, 0.00000438, -0.00000796, 0.00000252, -0.00000889, -0.00000010, -0.00001817, -0.00000497, 0.00001252, 0.00004205, -0.00000637, 0.00004271, 0.00000535, 0.00000825, -0.00003354, -0.00000077, -0.00001384, -0.00001440, 0.00001259, -0.00001703, 0.00001078, -0.00000864, 0.00000076, 0.00000698, -0.00000672, -0.00000203, 0.00000297, 0.00000027, -0.00000116, -0.00000067, 0.00000126, -0.00000136, 0.00000099, -0.00000096, -0.00000016, -0.00000019, -0.00000113, -0.00000413, 0.00000137, 0.00000561, -0.00000282, -0.00000254, 0.00000006, -0.00000206, 0.00000092, -0.00000053, 0.00000212, 0.00000475, -0.00000343, -0.00000087, -0.00000503, 0.00000375, -0.00000330, -0.00000521, 0.00000489, 0.00000218, -0.00000117, 0.00000185, -0.00000643, 0.00000630, -0.00000336, -0.00001106, 0.00000312, 0.00000988, -0.00000779, -0.00000386, -0.00000888, 0.00000986, -0.00001222, -0.00000819, -0.00001100, 0.00000095, -0.00002550, -0.00001703, -0.00013617, 0.00008723, 0.00019244, 0.00015574, -0.00008729, -0.00006846, -0.00000516, -0.00002341, 0.00001621, -0.00002203, 0.00000378, 0.00000143, 0.00000135, -0.00000655, -0.00000178, 0.00000349, 0.00000230, -0.00000867, 0.00000135, 0.00000393, -0.00000659, -0.00000740, 0.00000492, 0.00000013, 0.00000602, -0.00000465, -0.00000259, 0.00000517, -0.00000308, -0.00000553, 0.00000215, -0.00000005, 0.00000231, -0.00000222, 0.00000047, 0.00000152, 0.00000054, -0.00000126, -0.00000168, 0.00000259, -0.00000234, -0.00000214, 0.00000081, 0.00000371, },
		{ -0.00000639, -0.00000228, 0.00000451, -0.00000321, 0.00000508, -0.00000338, -0.00000231, 0.00000322, 0.00000089, -0.00000554, 0.00000983, -0.00000805, 0.00000173, -0.00000713, 0.00000057, -0.00001422, -0.00000782, -0.00000313, 0.00003922, 0.00001046, 0.00001390, 0.00003354, -0.00000270, -0.00000231, -0.00003317, 0.00001231, -0.00003516, 0.00001744, -0.00001329, 0.00000196, 0.00000091, -0.00000646, 0.00000656, 0.00000042, -0.00000800, 0.00000417, -0.00000008, 0.00000105, -0.00000343, 0.00000307, -0.00000224, 0.00000121, -0.00000068, -0.00000006, -0.00000073, 0.00000014, -0.00000355, -0.00000183, 0.00000510, 0.00000149, -0.00000454, 0.00000003, -0.00000141, -0.00000100, 0.00000150, -0.00000124, 0.00000629, -0.00000147, -0.00000087, -0.00000437, -0.00000005, 0.00000128, -0.00000623, -0.00000039, 0.00000696, -0.00000361, 0.00000458, -0.00000618, 0.00000180, 0.00000200, -0.00000756, -0.00000856, 0.00001487, -0.00000546, -0.00000116, -0.00001283, 0.00000743, -0.00000581, -0.00000561, -0.00001791, 0.00000716, -0.00002656, 0.00000118, -0.00009577, -0.00006146, 0.00018236, 0.00020297, 0.00003221, -0.00013230, 0.00000214, -0.00003783, 0.00001421, -0.00001157, -0.00000676, 0.00000135, 0.00000469, -0.00000614, -0.00000194, -0.00000206, 0.00000776, -0.00000776, -0.00000193, 0.00000219, 0.00000153, -0.00001290, 0.00000356, -0.00000018, 0.00000523, 0.00000060, -0.00000528, 0.00000198, 0.00000313, -0.00000816, 0.00000084, -0.00000004, 0.00000205, -0.00000030, -0.00000134, 0.00000158, 0.00000078, 0.00000027, -0.00000257, 0.00000115, 0.00000043, -0.00000324, -0.00000045, 0.00000303, 0.00000085, },
		{ -0.00000558, 0.00000302, 0.00000014, 0.00000026, 0.00000342, -0.00000638, 0.00000230, 0.00000262, -0.00000393, 0.00000290, 0.00000319, -0.00000769, 0.00000164, -0.00000698, -0.00000164, -0.00001826, 0.00000091, 0.00000980, 0.00004302, -0.00001363, 0.00004590, -0.00000141, 0.00001260, -0.00003331, 0.00000116, -0.00001749, -0.00001035, 0.00001070, -0.00001598, 0.00001113, -0.00000955, 0.00000288, 0.00000478, -0.00000650, -0.00000130, 0.00000323, 0.00000035, -0.00000211, 0.00000017, 0.00000113, -0.00000182, 0.00000158, -0.00000110, -0.00000013, -0.00000029, -0.00000158, -0.00000378, 0.00000235, 0.00000488, -0.00000348, -0.00000213, 0.00000013, -0.00000241, 0.00000149, -0.00000073, 0.00000286, 0.00000366, -0.00000314, -0.00000078, -0.00000478, 0.00000356, -0.00000424, -0.00000457, 0.00000588, 0.00000089, -0.00000028, 0.00000148, -0.00000618, 0.00000583, -0.00000395, -0.00001083, 0.00000513, 0.00000755, -0.00000776, -0.00000395, -0.00000684, 0.00000705, -0.00001094, -0.00000842, -0.00000975, -0.00000155, -0.00002414, -0.00001775, -0.00013961, 0.00007629, 0.00023093, 0.00014665, -0.00010899, -0.00005603, -0.00001000, -0.00002145, 0.00001486, -0.00002053, 0.00000341, 0.00000081, 0.00000128, -0.00000626, -0.00000153, 0.00000347, 0.00000083, -0.00000744, 0.00000148, 0.00000360, -0.00000808, -0.00000552, 0.00000438, 0.00000001, 0.00000622, -0.00000467, -0.00000220, 0.00000518, -0.00000455, -0.00000429, 0.00000198, -0.00000014, 0.00000216, -0.00000207, 0.00000070, 0.00000115, 0.00000068, -0.00000096, -0.00000188, 0.00000259, -0.00000295, -0.00000153, 0.00000131, 0.00000328, -0.00000399, },
		{ -0.00000178, 0.00000434, -0.00000342, 0.00000617, -0.00000482, -0.00000103, 0.00000213, 0.00000124, -0.00000501, 0.00001023, -0.00000948, 0.00000259, -0.00000821, 0.00000340, -0.00001897, -0.00000058, -0.00000490, 0.00004096, 0.00000479, 0.00001348, 0.00003152, -0.00000503, 0.00000159, -0.00003276, 0.00001163, -0.00003592, 0.00002067, -0.00001660, 0.00000580, -0.00000227, -0.00000359, 0.00000524, -0.00000030, -0.00000756, 0.00000496, -0.00000019, 0.00000085, -0.00000404, 0.00000436, -0.00000359, 0.00000213, -0.00000101, 0.00000029, -0.00000116, 0.00000029, -0.00000427, -0.00000031, 0.00000471, 0.00000096, -0.00000504, 0.00000102, -0.00000240, -0.00000011, 0.00000099, -0.00000028, 0.00000558, -0.00000167, -0.00000079, -0.00000395, -0.00000023, 0.00000093, -0.00000710, 0.00000171, 0.00000565, -0.00000313, 0.00000443, -0.00000580, 0.00000121, 0.00000202, -0.00000885, -0.00000578, 0.00001346, -0.00000605, -0.00000209, -0.00001026, 0.00000573, -0.00000605, -0.00000553, -0.00001627, 0.00000509, -0.00002583, -0.00000019, -0.00009348, -0.00007677, 0.00019866, 0.00023161, -0.00000403, -0.00012525, 0.00000307, -0.00003869, 0.00001401, -0.00001103, -0.00000637, 0.00000112, 0.00000357, -0.00000502, -0.00000271, -0.00000058, 0.00000542, -0.00000673, -0.00000186, 0.00000296, -0.00000071, -0.00001167, 0.00000392, -0.00000046, 0.00000502, 0.00000116, -0.00000575, 0.00000317, 0.00000121, -0.00000753, 0.00000118, -0.00000003, 0.00000162, 0.00000004, -0.00000146, 0.00000195, 0.00000002, 0.00000128, -0.00000330, 0.00000180, -0.00000066, -0.00000264, -0.00000029, 0.00000375, -0.00000094, -0.00000573, },
		{ 0.00000301, -0.00000038, 0.00000131, 0.00000272, -0.00000631, 0.00000262, 0.00000184, -0.00000329, 0.00000379, 0.00000188, -0.00000712, 0.00000039, -0.00000455, -0.00000399, -0.00001688, 0.00000530, 0.00000737, 0.00004375, -0.00002032, 0.00004892, -0.00000765, 0.00001637, -0.00003283, 0.00000292, -0.00002114, -0.00000575, 0.00000777, -0.00001378, 0.00001063, -0.00000994, 0.00000463, 0.00000261, -0.00000609, -0.00000063, 0.00000345, 0.00000043, -0.00000310, 0.00000118, 0.00000071, -0.00000202, 0.00000210, -0.00000127, -0.00000013, -0.00000035, -0.00000210, -0.00000324, 0.00000319, 0.00000409, -0.00000411, -0.00000158, 0.00000003, -0.00000259, 0.00000196, -0.00000085, 0.00000350, 0.00000254, -0.00000273, -0.00000075, -0.00000445, 0.00000323, -0.00000520, -0.00000360, 0.00000658, -0.00000033, 0.00000068, 0.00000100, -0.00000580, 0.00000522, -0.00000460, -0.00001022, 0.00000682, 0.00000515, -0.00000770, -0.00000381, -0.00000483, 0.00000387, -0.00000921, -0.00000884, -0.00000835, -0.00000443, -0.00002229, -0.00001905, -0.00014195, 0.00006250, 0.00027611, 0.00012962, -0.00012729, -0.00004325, -0.00001564, -0.00001881, 0.00001287, -0.00001842, 0.00000265, 0.00000027, 0.00000123, -0.00000598, -0.00000109, 0.00000313, -0.00000054, -0.00000605, 0.00000160, 0.00000299, -0.00000930, -0.00000359, 0.00000365, 0.00000001, 0.00000637, -0.00000471, -0.00000165, 0.00000486, -0.00000585, -0.00000296, 0.00000169, -0.00000019, 0.00000200, -0.00000190, 0.00000093, 0.00000072, 0.00000100, -0.00000087, -0.00000195, 0.00000244, -0.00000345, -0.00000090, 0.00000185, 0.00000263, -0.00000494, -0.00000442, },
		{ 0.00000412, -0.00000341, 0.00000705, -0.00000628, 0.00000041, 0.00000089, 0.00000169, -0.00000437, 0.00001029, -0.00001052, 0.00000308, -0.00000883, 0.00000566, -0.00002306, 0.00000671, -0.00000777, 0.00004330, -0.00000121, 0.00001378, 0.00002913, -0.00000674, 0.00000451, -0.00003141, 0.00000987, -0.00003532, 0.00002266, -0.00001918, 0.00000949, -0.00000563, -0.00000049, 0.00000354, -0.00000073, -0.00000710, 0.00000565, -0.00000027, 0.00000053, -0.00000445, 0.00000549, -0.00000493, 0.00000324, -0.00000149, 0.00000068, -0.00000161, 0.00000043, -0.00000495, 0.00000131, 0.00000410, 0.00000048, -0.00000548, 0.00000205, -0.00000351, 0.00000100, 0.00000027, 0.00000088, 0.00000455, -0.00000156, -0.00000091, -0.00000333, -0.00000060, 0.00000058, -0.00000790, 0.00000403, 0.00000387, -0.00000215, 0.00000384, -0.00000506, 0.00000031, 0.00000219, -0.00001025, -0.00000252, 0.00001124, -0.00000601, -0.00000356, -0.00000675, 0.00000282, -0.00000524, -0.00000629, -0.00001356, 0.00000157, -0.00002351, -0.00000347, -0.00008859, -0.00009583, 0.00021949, 0.00025909, -0.00004499, -0.00011254, 0.00000058, -0.00003738, 0.00001223, -0.00000936, -0.00000667, 0.00000129, 0.00000204, -0.00000348, -0.00000375, 0.00000122, 0.00000248, -0.00000508, -0.00000211, 0.00000387, -0.00000331, -0.00000983, 0.00000376, -0.00000044, 0.00000463, 0.00000183, -0.00000626, 0.00000445, -0.00000106, -0.00000643, 0.00000121, 0.00000014, 0.00000106, 0.00000049, -0.00000164, 0.00000231, -0.00000071, 0.00000231, -0.00000413, 0.00000255, -0.00000190, -0.00000186, -0.00000017, 0.00000440, -0.00000280, -0.00000481, -0.00000145, },
		{ -0.00000083, 0.00000248, 0.00000166, -0.00000590, 0.00000275, 0.00000112, -0.00000256, 0.00000456, 0.00000051, -0.00000636, -0.00000111, -0.00000176, -0.00000701, -0.00001407, 0.00000798, 0.00000563, 0.00004390, -0.00002612, 0.00005144, -0.00001308, 0.00001932, -0.00003189, 0.00000425, -0.00002447, -0.00000092, 0.00000405, -0.00001052, 0.00000925, -0.00000972, 0.00000591, 0.00000059, -0.00000557, -0.00000000, 0.00000362, 0.00000049, -0.00000408, 0.00000231, 0.00000002, -0.00000192, 0.00000249, -0.00000142, -0.00000017, -0.00000038, -0.00000270, -0.00000250, 0.00000387, 0.00000327, -0.00000473, -0.00000087, -0.00000028, -0.00000256, 0.00000228, -0.00000085, 0.00000399, 0.00000147, -0.00000228, -0.00000075, -0.00000407, 0.00000277, -0.00000619, -0.00000229, 0.00000689, -0.00000135, 0.00000160, 0.00000051, -0.00000538, 0.00000452, -0.00000533, -0.00000914, 0.00000806, 0.00000284, -0.00000773, -0.00000324, -0.00000315, 0.00000063, -0.00000729, -0.00000921, -0.00000708, -0.00000738, -0.00002029, -0.00002055, -0.00014366, 0.00004640, 0.00032782, 0.00010382, -0.00014082, -0.00003123, -0.00002140, -0.00001595, 0.00001055, -0.00001593, 0.00000163, -0.00000027, 0.00000128, -0.00000579, -0.00000039, 0.00000236, -0.00000167, -0.00000460, 0.00000176, 0.00000204, -0.00001015, -0.00000177, 0.00000281, 0.00000011, 0.00000650, -0.00000479, -0.00000091, 0.00000418, -0.00000688, -0.00000164, 0.00000134, -0.00000022, 0.00000184, -0.00000171, 0.00000117, 0.00000021, 0.00000147, -0.00000099, -0.00000184, 0.00000211, -0.00000380, -0.00000029, 0.00000243, 0.00000175, -0.00000562, -0.00000332, 0.00000289, },
		{ -0.00000316, 0.00000765, -0.00000765, 0.00000193, -0.00000047, 0.00000223, -0.00000365, 0.00001001, -0.00001112, 0.00000311, -0.00000888, 0.00000712, -0.00002613, 0.00001352, -0.00001128, 0.00004600, -0.00000736, 0.00001466, 0.00002643, -0.00000784, 0.00000647, -0.00002916, 0.00000705, -0.00003328, 0.00002322, -0.00002072, 0.00001275, -0.00000892, 0.00000268, 0.00000155, -0.00000090, -0.00000663, 0.00000625, -0.00000031, 0.00000009, -0.00000462, 0.00000637, -0.00000617, 0.00000445, -0.00000209, 0.00000110, -0.00000208, 0.00000053, -0.00000555, 0.00000299, 0.00000329, 0.00000007, -0.00000585, 0.00000306, -0.00000469, 0.00000226, -0.00000065, 0.00000225, 0.00000322, -0.00000114, -0.00000126, -0.00000248, -0.00000118, 0.00000024, -0.00000859, 0.00000648, 0.00000166, -0.00000068, 0.00000280, -0.00000389, -0.00000095, 0.00000254, -0.00001171, 0.00000116, 0.00000823, -0.00000531, -0.00000553, -0.00000237, -0.00000132, -0.00000326, -0.00000806, -0.00000962, -0.00000353, -0.00001936, -0.00000895, -0.00008069, -0.00011925, 0.00024601, 0.00028371, -0.00008936, -0.00009455, -0.00000543, -0.00003371, 0.00000870, -0.00000638, -0.00000782, 0.00000199, 0.00000003, -0.00000149, -0.00000508, 0.00000330, -0.00000100, -0.00000280, -0.00000269, 0.00000492, -0.00000621, -0.00000740, 0.00000305, -0.00000007, 0.00000403, 0.00000262, -0.00000679, 0.00000575, -0.00000360, -0.00000488, 0.00000091, 0.00000051, 0.00000035, 0.00000108, -0.00000188, 0.00000267, -0.00000139, 0.00000331, -0.00000502, 0.00000336, -0.00000324, -0.00000091, -0.00000010, 0.00000495, -0.00000467, -0.00000370, -0.00000130, 0.00000386, },
		{ 0.00000373, 0.00000029, -0.00000515, 0.00000265, 0.00000050, -0.00000176, 0.00000515, -0.00000082, -0.00000551, -0.00000270, 0.00000118, -0.00001044, -0.00001004, 0.00000887, 0.00000484, 0.00004319, -0.00003073, 0.00005315, -0.00001743, 0.00002124, -0.00003028, 0.00000489, -0.00002715, 0.00000378, -0.00000016, -0.00000639, 0.00000703, -0.00000887, 0.00000666, -0.00000120, -0.00000497, 0.00000058, 0.00000375, 0.00000049, -0.00000500, 0.00000350, -0.00000091, -0.00000149, 0.00000270, -0.00000152, -0.00000026, -0.00000038, -0.00000336, -0.00000155, 0.00000434, 0.00000245, -0.00000532, -0.00000002, -0.00000081, -0.00000227, 0.00000239, -0.00000067, 0.00000425, 0.00000054, -0.00000184, -0.00000071, -0.00000369, 0.00000220, -0.00000719, -0.00000063, 0.00000674, -0.00000206, 0.00000234, 0.00000011, -0.00000500, 0.00000379, -0.00000619, -0.00000754, 0.00000870, 0.00000080, -0.00000801, -0.00000208, -0.00000207, -0.00000232, -0.00000551, -0.00000922, -0.00000626, -0.00000999, -0.00001861, -0.00002171, -0.00014540, 0.00002882, 0.00038554, 0.00006875, -0.00014833, -0.00002112, -0.00002650, -0.00001337, 0.00000829, -0.00001335, 0.00000056, -0.00000095, 0.00000151, -0.00000575, 0.00000063, 0.00000104, -0.00000239, -0.00000325, 0.00000202, 0.00000069, -0.00001049, -0.00000019, 0.00000200, 0.00000021, 0.00000664, -0.00000492, 0.00000004, 0.00000306, -0.00000754, -0.00000043, 0.00000098, -0.00000029, 0.00000172, -0.00000153, 0.00000142, -0.00000034, 0.00000209, -0.00000134, -0.00000152, 0.00000155, -0.00000395, 0.00000026, 0.00000303, 0.00000067, -0.00000601, -0.00000232, 0.00000269, -0.00000117, },
		{ 0.00000790, -0.00000884, 0.00000342, -0.00000188, 0.00000285, -0.00000287, 0.00000941, -0.00001128, 0.00000267, -0.00000831, 0.00000766, -0.00002788, 0.00001935, -0.00001493, 0.00004871, -0.00001339, 0.00001591, 0.00002353, -0.00000840, 0.00000757, -0.00002615, 0.00000331, -0.00002985, 0.00002224, -0.00002102, 0.00001531, -0.00001192, 0.00000574, -0.00000060, -0.00000084, -0.00000616, 0.00000677, -0.00000033, -0.00000046, -0.00000454, 0.00000694, -0.00000722, 0.00000570, -0.00000280, 0.00000156, -0.00000257, 0.00000059, -0.00000604, 0.00000469, 0.00000231, -0.00000028, -0.00000612, 0.00000401, -0.00000588, 0.00000364, -0.00000173, 0.00000378, 0.00000161, -0.00000040, -0.00000186, -0.00000138, -0.00000201, -0.00000004, -0.00000915, 0.00000899, -0.00000089, 0.00000126, 0.00000128, -0.00000229, -0.00000260, 0.00000308, -0.00001321, 0.00000517, 0.00000451, -0.00000401, -0.00000792, 0.00000278, -0.00000662, -0.00000009, -0.00001089, -0.00000442, -0.00001029, -0.00001328, -0.00001680, -0.00006953, -0.00014749, 0.00027933, 0.00030369, -0.00013555, -0.00007201, -0.00001482, -0.00002766, 0.00000334, -0.00000201, -0.00000993, 0.00000331, -0.00000249, 0.00000095, -0.00000666, 0.00000556, -0.00000494, 0.00000010, -0.00000361, 0.00000607, -0.00000935, -0.00000446, 0.00000180, 0.00000067, 0.00000319, 0.00000353, -0.00000731, 0.00000700, -0.00000634, -0.00000292, 0.00000027, 0.00000108, -0.00000052, 0.00000181, -0.00000220, 0.00000302, -0.00000198, 0.00000423, -0.00000590, 0.00000418, -0.00000465, 0.00000019, -0.00000008, 0.00000538, -0.00000647, -0.00000244, -0.00000131, 0.00000359, -0.00000266, },
		{ -0.00000133, -0.00000408, 0.00000229, 0.00000003, -0.00000090, 0.00000554, -0.00000203, -0.00000469, -0.00000422, 0.00000405, -0.00001399, -0.00000508, 0.00000809, 0.00000514, 0.00004145, -0.00003396, 0.00005381, -0.00002049, 0.00002194, -0.00002780, 0.00000462, -0.00002885, 0.00000797, -0.00000450, -0.00000164, 0.00000410, -0.00000744, 0.00000685, -0.00000269, -0.00000435, 0.00000114, 0.00000385, 0.00000041, -0.00000579, 0.00000467, -0.00000202, -0.00000075, 0.00000270, -0.00000155, -0.00000040, -0.00000036, -0.00000405, -0.00000042, 0.00000459, 0.00000166, -0.00000588, 0.00000095, -0.00000156, -0.00000171, 0.00000224, -0.00000025, 0.00000423, -0.00000019, -0.00000151, -0.00000056, -0.00000339, 0.00000158, -0.00000819, 0.00000138, 0.00000606, -0.00000233, 0.00000278, -0.00000008, -0.00000477, 0.00000313, -0.00000721, -0.00000536, 0.00000861, -0.00000081, -0.00000867, -0.00000014, -0.00000188, -0.00000465, -0.00000420, -0.00000854, -0.00000630, -0.00001180, -0.00001779, -0.00002189, -0.00014800, 0.00001086, 0.00044831, 0.00002435, -0.00014884, -0.00001396, -0.00003016, -0.00001166, 0.00000654, -0.00001101, -0.00000035, -0.00000192, 0.00000205, -0.00000595, 0.00000202, -0.00000091, -0.00000256, -0.00000211, 0.00000246, -0.00000111, -0.00001024, 0.00000102, 0.00000133, 0.00000026, 0.00000683, -0.00000514, 0.00000121, 0.00000148, -0.00000772, 0.00000056, 0.00000071, -0.00000045, 0.00000168, -0.00000139, 0.00000169, -0.00000094, 0.00000284, -0.00000190, -0.00000099, 0.00000075, -0.00000385, 0.00000074, 0.00000365, -0.00000060, -0.00000610, -0.00000146, 0.00000241, -0.00000137, 0.00000496, },
		{ -0.00000976, 0.00000481, -0.00000329, 0.00000354, -0.00000207, 0.00000852, -0.00001103, 0.00000179, -0.00000711, 0.00000723, -0.00002813, 0.00002378, -0.00001819, 0.00005104, -0.00001896, 0.00001724, 0.00002064, -0.00000861, 0.00000798, -0.00002258, -0.00000115, -0.00002518, 0.00001975, -0.00001995, 0.00001695, -0.00001443, 0.00000851, -0.00000279, -0.00000061, -0.00000568, 0.00000719, -0.00000034, -0.00000108, -0.00000423, 0.00000717, -0.00000800, 0.00000691, -0.00000358, 0.00000204, -0.00000307, 0.00000062, -0.00000640, 0.00000635, 0.00000118, -0.00000057, -0.00000628, 0.00000484, -0.00000701, 0.00000507, -0.00000290, 0.00000542, -0.00000023, 0.00000062, -0.00000269, -0.00000003, -0.00000311, -0.00000026, -0.00000954, 0.00001144, -0.00000367, 0.00000358, -0.00000066, -0.00000026, -0.00000464, 0.00000381, -0.00001469, 0.00000938, 0.00000019, -0.00000216, -0.00001063, 0.00000851, -0.00001291, 0.00000420, -0.00001474, 0.00000199, -0.00001864, -0.00000530, -0.00002703, -0.00005505, -0.00018081, 0.00032034, 0.00031722, -0.00018174, -0.00004596, -0.00002717, -0.00001941, -0.00000375, 0.00000373, -0.00001303, 0.00000526, -0.00000554, 0.00000383, -0.00000846, 0.00000792, -0.00000921, 0.00000352, -0.00000487, 0.00000726, -0.00001261, -0.00000109, 0.00000004, 0.00000179, 0.00000212, 0.00000456, -0.00000780, 0.00000813, -0.00000916, -0.00000061, -0.00000069, 0.00000184, -0.00000155, 0.00000267, -0.00000261, 0.00000338, -0.00000249, 0.00000503, -0.00000671, 0.00000494, -0.00000606, 0.00000141, -0.00000009, 0.00000566, -0.00000815, -0.00000108, -0.00000147, 0.00000331, -0.00000193, 0.00000777, },
		{ -0.00000272, 0.00000167, -0.00000028, -0.00000000, 0.00000570, -0.00000305, -0.00000402, -0.00000551, 0.00000659, -0.00001730, 0.00000041, 0.00000587, 0.00000653, 0.00003861, -0.00003569, 0.00005325, -0.00002212, 0.00002130, -0.00002434, 0.00000326, -0.00002931, 0.00001127, -0.00000858, 0.00000341, 0.00000063, -0.00000549, 0.00000651, -0.00000387, -0.00000374, 0.00000167, 0.00000392, 0.00000024, -0.00000640, 0.00000575, -0.00000324, 0.00000027, 0.00000246, -0.00000147, -0.00000062, -0.00000031, -0.00000477, 0.00000089, 0.00000459, 0.00000092, -0.00000641, 0.00000201, -0.00000252, -0.00000085, 0.00000180, 0.00000044, 0.00000388, -0.00000064, -0.00000136, -0.00000025, -0.00000322, 0.00000093, -0.00000918, 0.00000371, 0.00000481, -0.00000206, 0.00000279, 0.00000006, -0.00000479, 0.00000261, -0.00000844, -0.00000258, 0.00000770, -0.00000185, -0.00000982, 0.00000269, -0.00000282, -0.00000599, -0.00000373, -0.00000683, -0.00000758, -0.00001231, -0.00001840, -0.00002040, -0.00015241, -0.00000610, 0.00051478, -0.00002902, -0.00014169, -0.00001066, -0.00003168, -0.00001134, 0.00000571, -0.00000927, -0.00000083, -0.00000334, 0.00000302, -0.00000648, 0.00000384, -0.00000356, -0.00000204, -0.00000132, 0.00000315, -0.00000341, -0.00000931, 0.00000171, 0.00000092, 0.00000016, 0.00000712, -0.00000544, 0.00000257, -0.00000057, -0.00000737, 0.00000124, 0.00000059, -0.00000074, 0.00000177, -0.00000132, 0.00000200, -0.00000156, 0.00000368, -0.00000265, -0.00000026, -0.00000030, -0.00000350, 0.00000112, 0.00000427, -0.00000203, -0.00000588, -0.00000078, 0.00000210, -0.00000142, 0.00000611, -0.00000311, },
		{ 0.00000601, -0.00000463, 0.00000427, -0.00000127, 0.00000740, -0.00001043, 0.00000052, -0.00000537, 0.00000586, -0.00002678, 0.00002648, -0.00002062, 0.00005257, -0.00002372, 0.00001833, 0.00001800, -0.00000868, 0.00000795, -0.00001871, -0.00000602, -0.00001952, 0.00001587, -0.00001747, 0.00001752, -0.00001627, 0.00001085, -0.00000489, -0.00000026, -0.00000518, 0.00000754, -0.00000036, -0.00000174, -0.00000370, 0.00000705, -0.00000845, 0.00000800, -0.00000438, 0.00000253, -0.00000358, 0.00000060, -0.00000660, 0.00000793, -0.00000008, -0.00000078, -0.00000632, 0.00000554, -0.00000803, 0.00000648, -0.00000412, 0.00000710, -0.00000222, 0.00000188, -0.00000372, 0.00000153, -0.00000447, -0.00000039, -0.00000974, 0.00001375, -0.00000657, 0.00000619, -0.00000296, 0.00000213, -0.00000704, 0.00000472, -0.00001608, 0.00001365, -0.00000460, 0.00000011, -0.00001348, 0.00001455, -0.00001996, 0.00000943, -0.00001947, 0.00000942, -0.00002837, 0.00000441, -0.00003948, -0.00003742, -0.00021917, 0.00036971, 0.00032261, -0.00022600, -0.00001767, -0.00004182, -0.00000932, -0.00001235, 0.00001069, -0.00001703, 0.00000781, -0.00000906, 0.00000711, -0.00001040, 0.00001026, -0.00001365, 0.00000737, -0.00000639, 0.00000843, -0.00001587, 0.00000257, -0.00000214, 0.00000325, 0.00000082, 0.00000569, -0.00000822, 0.00000906, -0.00001195, 0.00000193, -0.00000192, 0.00000277, -0.00000270, 0.00000366, -0.00000308, 0.00000373, -0.00000290, 0.00000567, -0.00000741, 0.00000558, -0.00000740, 0.00000270, -0.00000012, 0.00000576, -0.00000964, 0.00000030, -0.00000176, 0.00000307, -0.00000102, 0.00000725, -0.00001033, },
		{ 0.00000082, -0.00000041, 0.00000092, 0.00000560, -0.00000380, -0.00000359, -0.00000642, 0.00000859, -0.00002003, 0.00000597, 0.00000257, 0.00000884, 0.00003473, -0.00003592, 0.00005141, -0.00002226, 0.00001931, -0.00001985, 0.00000073, -0.00002833, 0.00001338, -0.00001201, 0.00000842, -0.00000315, -0.00000317, 0.00000569, -0.00000470, -0.00000316, 0.00000219, 0.00000396, -0.00000002, -0.00000680, 0.00000665, -0.00000447, 0.00000152, 0.00000199, -0.00000127, -0.00000091, -0.00000025, -0.00000547, 0.00000235, 0.00000433, 0.00000027, -0.00000690, 0.00000315, -0.00000366, 0.00000028, 0.00000105, 0.00000143, 0.00000316, -0.00000074, -0.00000145, 0.00000031, -0.00000327, 0.00000033, -0.00001015, 0.00000634, 0.00000299, -0.00000120, 0.00000227, 0.00000064, -0.00000518, 0.00000232, -0.00000990, 0.00000083, 0.00000588, -0.00000219, -0.00001155, 0.00000651, -0.00000506, -0.00000609, -0.00000440, -0.00000377, -0.00001048, -0.00001106, -0.00002102, -0.00001650, -0.00015966, -0.00002053, 0.00058323, -0.00009049, -0.00012659, -0.00001190, -0.00003043, -0.00001293, 0.00000624, -0.00000846, -0.00000063, -0.00000539, 0.00000455, -0.00000742, 0.00000610, -0.00000694, -0.00000075, -0.00000099, 0.00000414, -0.00000622, -0.00000765, 0.00000180, 0.00000088, -0.00000016, 0.00000756, -0.00000585, 0.00000411, -0.00000308, -0.00000643, 0.00000153, 0.00000070, -0.00000122, 0.00000204, -0.00000136, 0.00000235, -0.00000220, 0.00000458, -0.00000355, 0.00000065, -0.00000160, -0.00000286, 0.00000137, 0.00000488, -0.00000357, -0.00000538, -0.00000030, 0.00000178, -0.00000130, 0.00000710, -0.00000495, -0.00000116, },
		{ -0.00000583, 0.00000502, -0.00000051, 0.00000613, -0.00000955, -0.00000101, -0.00000320, 0.00000368, -0.00002388, 0.00002726, -0.00002183, 0.00005296, -0.00002731, 0.00001886, 0.00001588, -0.00000885, 0.00000776, -0.00001486, -0.00001095, -0.00001322, 0.00001086, -0.00001366, 0.00001694, -0.00001733, 0.00001263, -0.00000679, 0.00000013, -0.00000465, 0.00000780, -0.00000042, -0.00000240, -0.00000300, 0.00000660, -0.00000853, 0.00000891, -0.00000516, 0.00000302, -0.00000410, 0.00000054, -0.00000663, 0.00000938, -0.00000142, -0.00000092, -0.00000625, 0.00000605, -0.00000888, 0.00000778, -0.00000531, 0.00000876, -0.00000430, 0.00000331, -0.00000489, 0.00000327, -0.00000608, -0.00000042, -0.00000971, 0.00001580, -0.00000943, 0.00000896, -0.00000551, 0.00000481, -0.00000974, 0.00000576, -0.00001730, 0.00001783, -0.00000966, 0.00000267, -0.00001629, 0.00002064, -0.00002743, 0.00001530, -0.00002482, 0.00001758, -0.00003915, 0.00001549, -0.00005376, -0.00001706, -0.00026220, 0.00042772, 0.00031839, -0.00026636, 0.00001141, -0.00005792, 0.00000208, -0.00002208, 0.00001863, -0.00002179, 0.00001085, -0.00001295, 0.00001070, -0.00001242, 0.00001243, -0.00001807, 0.00001149, -0.00000810, 0.00000950, -0.00001899, 0.00000638, -0.00000463, 0.00000500, -0.00000067, 0.00000690, -0.00000854, 0.00000970, -0.00001458, 0.00000461, -0.00000335, 0.00000383, -0.00000394, 0.00000476, -0.00000363, 0.00000409, -0.00000321, 0.00000613, -0.00000793, 0.00000604, -0.00000862, 0.00000403, -0.00000018, 0.00000567, -0.00001090, 0.00000166, -0.00000215, 0.00000288, 0.00000005, 0.00000634, -0.00001051, 0.00000694, },
		{ -0.00000035, 0.00000183, 0.00000526, -0.00000426, -0.00000345, -0.00000683, 0.00000983, -0.00002186, 0.00001112, -0.00000135, 0.00001179, 0.00003002, -0.00003478, 0.00004834, -0.00002096, 0.00001601, -0.00001438, -0.00000298, -0.00002579, 0.00001403, -0.00001446, 0.00001303, -0.00000698, -0.00000064, 0.00000447, -0.00000521, -0.00000263, 0.00000269, 0.00000395, -0.00000038, -0.00000694, 0.00000731, -0.00000563, 0.00000294, 0.00000129, -0.00000094, -0.00000130, -0.00000016, -0.00000616, 0.00000393, 0.00000381, -0.00000029, -0.00000733, 0.00000431, -0.00000493, 0.00000167, -0.00000001, 0.00000273, 0.00000205, -0.00000045, -0.00000185, 0.00000117, -0.00000359, -0.00000019, -0.00001110, 0.00000920, 0.00000062, 0.00000030, 0.00000115, 0.00000176, -0.00000603, 0.00000234, -0.00001164, 0.00000486, 0.00000312, -0.00000174, -0.00001389, 0.00001132, -0.00000871, -0.00000471, -0.00000648, 0.00000089, -0.00001531, -0.00000762, -0.00002618, -0.00000954, -0.00017074, -0.00003076, 0.00065162, -0.00015877, -0.00010368, -0.00001809, -0.00002597, -0.00001679, 0.00000846, -0.00000889, 0.00000047, -0.00000823, 0.00000676, -0.00000882, 0.00000881, -0.00001104, 0.00000139, -0.00000122, 0.00000547, -0.00000953, -0.00000523, 0.00000119, 0.00000131, -0.00000078, 0.00000819, -0.00000636, 0.00000579, -0.00000600, -0.00000490, 0.00000136, 0.00000108, -0.00000193, 0.00000251, -0.00000153, 0.00000276, -0.00000286, 0.00000550, -0.00000456, 0.00000170, -0.00000311, -0.00000194, 0.00000150, 0.00000546, -0.00000519, -0.00000461, -0.00000004, 0.00000146, -0.00000100, 0.00000785, -0.00000674, 0.00000054, -0.00000023, },
		{ 0.00000577, 0.00000017, 0.00000477, -0.00000850, -0.00000268, -0.00000079, 0.00000090, -0.00001960, 0.00002607, -0.00002158, 0.00005190, -0.00002944, 0.00001852, 0.00001454, -0.00000938, 0.00000768, -0.00001135, -0.00001556, -0.00000670, 0.00000505, -0.00000873, 0.00001525, -0.00001755, 0.00001376, -0.00000839, 0.00000053, -0.00000408, 0.00000798, -0.00000054, -0.00000299, -0.00000220, 0.00000584, -0.00000823, 0.00000958, -0.00000587, 0.00000349, -0.00000463, 0.00000044, -0.00000647, 0.00001066, -0.00000282, -0.00000099, -0.00000606, 0.00000636, -0.00000949, 0.00000891, -0.00000638, 0.00001030, -0.00000636, 0.00000483, -0.00000615, 0.00000512, -0.00000789, -0.00000035, -0.00000944, 0.00001749, -0.00001212, 0.00001172, -0.00000817, 0.00000766, -0.00001265, 0.00000687, -0.00001826, 0.00002173, -0.00001481, 0.00000536, -0.00001886, 0.00002643, -0.00003494, 0.00002148, -0.00003043, 0.00002607, -0.00005051, 0.00002743, -0.00006931, 0.00000535, -0.00030920, 0.00049426, 0.00030341, -0.00030096, 0.00003972, -0.00007448, 0.00001411, -0.00003245, 0.00002717, -0.00002707, 0.00001422, -0.00001708, 0.00001449, -0.00001442, 0.00001429, -0.00002227, 0.00001571, -0.00000990, 0.00001036, -0.00002182, 0.00001015, -0.00000730, 0.00000694, -0.00000230, 0.00000815, -0.00000873, 0.00000999, -0.00001690, 0.00000729, -0.00000490, 0.00000496, -0.00000523, 0.00000592, -0.00000424, 0.00000445, -0.00000343, 0.00000640, -0.00000825, 0.00000627, -0.00000965, 0.00000535, -0.00000024, 0.00000538, -0.00001189, 0.00000293, -0.00000262, 0.00000277, 0.00000120, 0.00000509, -0.00001028, 0.00000754, -0.00000684, },
		{ 0.00000273, 0.00000468, -0.00000442, -0.00000364, -0.00000668, 0.00001020, -0.00002251, 0.00001541, -0.00000539, 0.00001502, 0.00002474, -0.00003246, 0.00004421, -0.00001837, 0.00001156, -0.00000806, -0.00000776, -0.00002173, 0.00001311, -0.00001564, 0.00001691, -0.00001059, 0.00000192, 0.00000298, -0.00000544, -0.00000215, 0.00000319, 0.00000390, -0.00000082, -0.00000684, 0.00000766, -0.00000663, 0.00000445, 0.00000040, -0.00000047, -0.00000178, -0.00000006, -0.00000679, 0.00000560, 0.00000304, -0.00000072, -0.00000772, 0.00000547, -0.00000630, 0.00000327, -0.00000135, 0.00000432, 0.00000055, 0.00000024, -0.00000258, 0.00000236, -0.00000425, -0.00000056, -0.00001202, 0.00001224, -0.00000225, 0.00000244, -0.00000061, 0.00000348, -0.00000742, 0.00000275, -0.00001365, 0.00000946, -0.00000058, -0.00000047, -0.00001686, 0.00001708, -0.00001378, -0.00000174, -0.00001012, 0.00000734, -0.00002229, -0.00000169, -0.00003427, 0.00000107, -0.00018661, -0.00003511, 0.00071767, -0.00023212, -0.00007354, -0.00002933, -0.00001806, -0.00002318, 0.00001262, -0.00001079, 0.00000267, -0.00001198, 0.00000974, -0.00001075, 0.00001194, -0.00001582, 0.00000440, -0.00000207, 0.00000714, -0.00001331, -0.00000207, -0.00000014, 0.00000228, -0.00000177, 0.00000904, -0.00000698, 0.00000756, -0.00000927, -0.00000278, 0.00000070, 0.00000179, -0.00000291, 0.00000322, -0.00000187, 0.00000326, -0.00000352, 0.00000641, -0.00000564, 0.00000285, -0.00000479, -0.00000075, 0.00000148, 0.00000598, -0.00000684, -0.00000361, 0.00000001, 0.00000119, -0.00000052, 0.00000831, -0.00000838, 0.00000227, -0.00000142, -0.00000012, },
		{ 0.00000074, 0.00000340, -0.00000739, -0.00000434, 0.00000166, -0.00000222, -0.00001422, 0.00002303, -0.00001975, 0.00004923, -0.00002989, 0.00001708, 0.00001419, -0.00001049, 0.00000797, -0.00000850, -0.00001945, -0.00000037, -0.00000116, -0.00000294, 0.00001255, -0.00001692, 0.00001419, -0.00000963, 0.00000087, -0.00000346, 0.00000809, -0.00000073, -0.00000346, -0.00000138, 0.00000483, -0.00000755, 0.00000995, -0.00000647, 0.00000390, -0.00000516, 0.00000032, -0.00000613, 0.00001173, -0.00000423, -0.00000099, -0.00000577, 0.00000646, -0.00000984, 0.00000980, -0.00000727, 0.00001165, -0.00000831, 0.00000634, -0.00000741, 0.00000700, -0.00000987, -0.00000019, -0.00000891, 0.00001875, -0.00001447, 0.00001431, -0.00001080, 0.00001055, -0.00001565, 0.00000798, -0.00001887, 0.00002519, -0.00001983, 0.00000799, -0.00002097, 0.00003162, -0.00004210, 0.00002755, -0.00003591, 0.00003440, -0.00006188, 0.00003960, -0.00008539, 0.00002887, -0.00035909, 0.00056873, 0.00027693, -0.00032814, 0.00006573, -0.00009041, 0.00002602, -0.00004286, 0.00003587, -0.00003254, 0.00001767, -0.00002124, 0.00001832, -0.00001628, 0.00001572, -0.00002605, 0.00001986, -0.00001168, 0.00001092, -0.00002421, 0.00001372, -0.00000999, 0.00000897, -0.00000401, 0.00000941, -0.00000877, 0.00000986, -0.00001882, 0.00000982, -0.00000645, 0.00000608, -0.00000651, 0.00000710, -0.00000487, 0.00000481, -0.00000356, 0.00000647, -0.00000835, 0.00000624, -0.00001044, 0.00000661, -0.00000031, 0.00000490, -0.00001257, 0.00000407, -0.00000313, 0.00000275, 0.00000237, 0.00000356, -0.00000965, 0.00000778, -0.00000762, 0.00000650, },
		{ 0.00000389, -0.00000429, -0.00000416, -0.00000596, 0.00000962, -0.00002179, 0.00001841, -0.00000904, 0.00001809, 0.00001926, -0.00002930, 0.00003925, -0.00001473, 0.00000622, -0.00000112, -0.00001339, -0.00001626, 0.00001059, -0.00001538, 0.00001978, -0.00001374, 0.00000432, 0.00000134, -0.00000543, -0.00000172, 0.00000367, 0.00000377, -0.00000131, -0.00000649, 0.00000769, -0.00000740, 0.00000596, -0.00000065, 0.00000014, -0.00000236, 0.00000007, -0.00000735, 0.00000731, 0.00000203, -0.00000102, -0.00000805, 0.00000659, -0.00000770, 0.00000503, -0.00000293, 0.00000618, -0.00000131, 0.00000135, -0.00000367, 0.00000391, -0.00000529, -0.00000073, -0.00001291, 0.00001537, -0.00000554, 0.00000518, -0.00000303, 0.00000582, -0.00000939, 0.00000358, -0.00001593, 0.00001458, -0.00000518, 0.00000163, -0.00002040, 0.00002367, -0.00002020, 0.00000284, -0.00001539, 0.00001560, -0.00003151, 0.00000690, -0.00004557, 0.00001573, -0.00020800, -0.00003199, 0.00077899, -0.00030846, -0.00003712, -0.00004539, -0.00000671, -0.00003217, 0.00001882, -0.00001431, 0.00000608, -0.00001671, 0.00001354, -0.00001322, 0.00001546, -0.00002120, 0.00000824, -0.00000355, 0.00000915, -0.00001748, 0.00000177, -0.00000220, 0.00000381, -0.00000316, 0.00001015, -0.00000770, 0.00000935, -0.00001279, -0.00000014, -0.00000044, 0.00000283, -0.00000416, 0.00000418, -0.00000240, 0.00000386, -0.00000419, 0.00000729, -0.00000671, 0.00000403, -0.00000659, 0.00000068, 0.00000132, 0.00000644, -0.00000845, -0.00000243, -0.00000015, 0.00000097, 0.00000011, 0.00000845, -0.00000978, 0.00000394, -0.00000268, 0.00000028, 0.00000357, },
		{ 0.00000209, -0.00000630, -0.00000583, 0.00000393, -0.00000536, -0.00000811, 0.00001839, -0.00001638, 0.00004487, -0.00002855, 0.00001438, 0.00001498, -0.00001234, 0.00000886, -0.00000655, -0.00002227, 0.00000530, -0.00000731, 0.00000334, 0.00000903, -0.00001551, 0.00001394, -0.00001046, 0.00000111, -0.00000277, 0.00000812, -0.00000101, -0.00000377, -0.00000060, 0.00000367, -0.00000652, 0.00001002, -0.00000691, 0.00000424, -0.00000567, 0.00000019, -0.00000561, 0.00001256, -0.00000561, -0.00000093, -0.00000539, 0.00000635, -0.00000990, 0.00001039, -0.00000790, 0.00001272, -0.00001004, 0.00000775, -0.00000857, 0.00000882, -0.00001193, 0.00000006, -0.00000812, 0.00001949, -0.00001637, 0.00001658, -0.00001321, 0.00001331, -0.00001863, 0.00000900, -0.00001903, 0.00002803, -0.00002452, 0.00001040, -0.00002245, 0.00003588, -0.00004847, 0.00003307, -0.00004080, 0.00004205, -0.00007264, 0.00005126, -0.00010111, 0.00005237, -0.00041045, 0.00065008, 0.00023872, -0.00034654, 0.00008801, -0.00010465, 0.00003698, -0.00005269, 0.00004423, -0.00003784, 0.00002095, -0.00002523, 0.00002205, -0.00001791, 0.00001658, -0.00002922, 0.00002372, -0.00001331, 0.00001109, -0.00002602, 0.00001689, -0.00001253, 0.00001097, -0.00000570, 0.00001064, -0.00000864, 0.00000928, -0.00002021, 0.00001209, -0.00000792, 0.00000711, -0.00000770, 0.00000826, -0.00000550, 0.00000516, -0.00000361, 0.00000635, -0.00000822, 0.00000592, -0.00001093, 0.00000776, -0.00000037, 0.00000424, -0.00001294, 0.00000503, -0.00000367, 0.00000285, 0.00000348, 0.00000183, -0.00000865, 0.00000764, -0.00000812, 0.00000719, 0.00000119, },
		{ -0.00000391, -0.00000495, -0.00000472, 0.00000810, -0.00001962, 0.00001984, -0.00001183, 0.00002057, 0.00001400, -0.00002565, 0.00003378, -0.00001036, 0.00000030, 0.00000612, -0.00001957, -0.00000963, 0.00000658, -0.00001360, 0.00002143, -0.00001619, 0.00000638, -0.00000031, -0.00000526, -0.00000131, 0.00000415, 0.00000356, -0.00000181, -0.00000593, 0.00000738, -0.00000786, 0.00000740, -0.00000182, 0.00000087, -0.00000305, 0.00000022, -0.00000783, 0.00000904, 0.00000080, -0.00000116, -0.00000834, 0.00000763, -0.00000908, 0.00000688, -0.00000470, 0.00000825, -0.00000349, 0.00000283, -0.00000511, 0.00000583, -0.00000674, -0.00000066, -0.00001375, 0.00001852, -0.00000914, 0.00000843, -0.00000604, 0.00000877, -0.00001196, 0.00000486, -0.00001844, 0.00002010, -0.00001058, 0.00000451, -0.00002442, 0.00003093, -0.00002780, 0.00000892, -0.00002221, 0.00002562, -0.00004290, 0.00001815, -0.00006014, 0.00003459, -0.00023543, -0.00002000, 0.00083316, -0.00038545, 0.00000423, -0.00006571, 0.00000785, -0.00004364, 0.00002706, -0.00001948, 0.00001074, -0.00002244, 0.00001818, -0.00001624, 0.00001928, -0.00002704, 0.00001284, -0.00000564, 0.00001146, -0.00002196, 0.00000620, -0.00000494, 0.00000592, -0.00000496, 0.00001152, -0.00000852, 0.00001111, -0.00001645, 0.00000294, -0.00000205, 0.00000419, -0.00000569, 0.00000542, -0.00000312, 0.00000457, -0.00000488, 0.00000809, -0.00000775, 0.00000518, -0.00000844, 0.00000232, 0.00000103, 0.00000680, -0.00000998, -0.00000112, -0.00000050, 0.00000084, 0.00000086, 0.00000824, -0.00001084, 0.00000544, -0.00000395, 0.00000081, 0.00000433, 0.00000294, },
		{ -0.00000535, -0.00000700, 0.00000578, -0.00000820, -0.00000172, 0.00001252, -0.00001167, 0.00003893, -0.00002542, 0.00001038, 0.00001697, -0.00001501, 0.00001046, -0.00000571, -0.00002375, 0.00000994, -0.00001295, 0.00000972, 0.00000493, -0.00001346, 0.00001304, -0.00001087, 0.00000125, -0.00000202, 0.00000806, -0.00000139, -0.00000387, 0.00000004, 0.00000243, -0.00000519, 0.00000977, -0.00000717, 0.00000448, -0.00000617, 0.00000004, -0.00000491, 0.00001313, -0.00000693, -0.00000080, -0.00000494, 0.00000604, -0.00000964, 0.00001064, -0.00000822, 0.00001344, -0.00001147, 0.00000895, -0.00000954, 0.00001047, -0.00001399, 0.00000038, -0.00000708, 0.00001967, -0.00001768, 0.00001836, -0.00001526, 0.00001577, -0.00002143, 0.00000985, -0.00001866, 0.00003011, -0.00002866, 0.00001243, -0.00002313, 0.00003895, -0.00005369, 0.00003762, -0.00004464, 0.00004848, -0.00008210, 0.00006162, -0.00011550, 0.00007454, -0.00046155, 0.00073675, 0.00018907, -0.00035519, 0.00010530, -0.00011620, 0.00004624, -0.00006130, 0.00005174, -0.00004258, 0.00002376, -0.00002879, 0.00002547, -0.00001919, 0.00001676, -0.00003159, 0.00002711, -0.00001468, 0.00001077, -0.00002713, 0.00001950, -0.00001477, 0.00001281, -0.00000731, 0.00001178, -0.00000833, 0.00000823, -0.00002098, 0.00001396, -0.00000918, 0.00000797, -0.00000873, 0.00000932, -0.00000609, 0.00000548, -0.00000361, 0.00000607, -0.00000786, 0.00000531, -0.00001109, 0.00000877, -0.00000043, 0.00000342, -0.00001298, 0.00000577, -0.00000419, 0.00000307, 0.00000446, -0.00000000, -0.00000735, 0.00000715, -0.00000832, 0.00000780, 0.00000150, 0.00000090, },
		{ -0.00000592, -0.00000305, 0.00000575, -0.00001602, 0.00001949, -0.00001339, 0.00002206, 0.00000934, -0.00002191, 0.00002815, -0.00000560, -0.00000584, 0.00001331, -0.00002593, -0.00000220, 0.00000130, -0.00001037, 0.00002176, -0.00001779, 0.00000795, -0.00000186, -0.00000498, -0.00000091, 0.00000462, 0.00000324, -0.00000228, -0.00000520, 0.00000675, -0.00000797, 0.00000868, -0.00000303, 0.00000171, -0.00000384, 0.00000041, -0.00000820, 0.00001073, -0.00000063, -0.00000113, -0.00000858, 0.00000858, -0.00001038, 0.00000874, -0.00000658, 0.00001046, -0.00000591, 0.00000465, -0.00000686, 0.00000808, -0.00000860, -0.00000031, -0.00001453, 0.00002158, -0.00001292, 0.00001208, -0.00000957, 0.00001228, -0.00001511, 0.00000658, -0.00002113, 0.00002589, -0.00001665, 0.00000808, -0.00002879, 0.00003862, -0.00003632, 0.00001629, -0.00003041, 0.00003718, -0.00005624, 0.00003188, -0.00007782, 0.00005757, -0.00026909, 0.00000194, 0.00087791, -0.00046062, 0.00004889, -0.00008945, 0.00002515, -0.00005729, 0.00003714, -0.00002620, 0.00001658, -0.00002912, 0.00002362, -0.00001976, 0.00002330, -0.00003319, 0.00001807, -0.00000830, 0.00001398, -0.00002660, 0.00001107, -0.00000828, 0.00000856, -0.00000718, 0.00001315, -0.00000942, 0.00001275, -0.00002012, 0.00000636, -0.00000405, 0.00000586, -0.00000747, 0.00000690, -0.00000405, 0.00000540, -0.00000558, 0.00000882, -0.00000870, 0.00000625, -0.00001027, 0.00000411, 0.00000061, 0.00000707, -0.00001139, 0.00000027, -0.00000103, 0.00000082, 0.00000169, 0.00000769, -0.00001153, 0.00000670, -0.00000515, 0.00000145, 0.00000500, 0.00000189, -0.00000335, },
		{ -0.00000775, 0.00000704, -0.00001041, 0.00000448, 0.00000591, -0.00000593, 0.00003161, -0.00002065, 0.00000513, 0.00002011, -0.00001850, 0.00001284, -0.00000605, -0.00002369, 0.00001320, -0.00001766, 0.00001579, 0.00000055, -0.00001093, 0.00001159, -0.00001088, 0.00000127, -0.00000120, 0.00000792, -0.00000185, -0.00000375, 0.00000049, 0.00000120, -0.00000364, 0.00000920, -0.00000722, 0.00000460, -0.00000663, -0.00000010, -0.00000405, 0.00001340, -0.00000814, -0.00000063, -0.00000444, 0.00000555, -0.00000909, 0.00001053, -0.00000819, 0.00001375, -0.00001252, 0.00000985, -0.00001022, 0.00001185, -0.00001597, 0.00000072, -0.00000579, 0.00001927, -0.00001832, 0.00001952, -0.00001677, 0.00001779, -0.00002391, 0.00001042, -0.00001769, 0.00003127, -0.00003208, 0.00001392, -0.00002288, 0.00004061, -0.00005741, 0.00004080, -0.00004702, 0.00005316, -0.00008963, 0.00006990, -0.00012755, 0.00009399, -0.00051046, 0.00082675, 0.00012881, -0.00035360, 0.00011665, -0.00012417, 0.00005306, -0.00006808, 0.00005789, -0.00004638, 0.00002579, -0.00003167, 0.00002842, -0.00002001, 0.00001618, -0.00003302, 0.00002987, -0.00001567, 0.00000992, -0.00002744, 0.00002140, -0.00001653, 0.00001435, -0.00000871, 0.00001280, -0.00000785, 0.00000671, -0.00002110, 0.00001534, -0.00001014, 0.00000858, -0.00000953, 0.00001023, -0.00000660, 0.00000576, -0.00000356, 0.00000565, -0.00000731, 0.00000441, -0.00001090, 0.00000959, -0.00000047, 0.00000246, -0.00001271, 0.00000629, -0.00000466, 0.00000342, 0.00000525, -0.00000182, -0.00000582, 0.00000632, -0.00000822, 0.00000832, 0.00000165, -0.00000013, -0.00000460, },
		{ -0.00000110, 0.00000273, -0.00001114, 0.00001731, -0.00001345, 0.00002221, 0.00000566, -0.00001844, 0.00002272, -0.00000083, -0.00001181, 0.00002006, -0.00003203, 0.00000560, -0.00000490, -0.00000583, 0.00002073, -0.00001843, 0.00000892, -0.00000318, -0.00000467, -0.00000051, 0.00000507, 0.00000280, -0.00000265, -0.00000438, 0.00000584, -0.00000773, 0.00000975, -0.00000424, 0.00000262, -0.00000474, 0.00000064, -0.00000847, 0.00001235, -0.00000223, -0.00000093, -0.00000879, 0.00000941, -0.00001155, 0.00001054, -0.00000850, 0.00001274, -0.00000850, 0.00000674, -0.00000887, 0.00001061, -0.00001085, 0.00000032, -0.00001525, 0.00002447, -0.00001675, 0.00001599, -0.00001349, 0.00001624, -0.00001875, 0.00000869, -0.00002393, 0.00003181, -0.00002322, 0.00001222, -0.00003334, 0.00004645, -0.00004542, 0.00002465, -0.00003968, 0.00004993, -0.00007117, 0.00004770, -0.00009823, 0.00008429, -0.00030881, 0.00003455, 0.00091119, -0.00053145, 0.00009502, -0.00011554, 0.00004450, -0.00007264, 0.00004874, -0.00003427, 0.00002346, -0.00003661, 0.00002976, -0.00002372, 0.00002739, -0.00003946, 0.00002377, -0.00001142, 0.00001664, -0.00003126, 0.00001622, -0.00001209, 0.00001165, -0.00000976, 0.00001503, -0.00001037, 0.00001422, -0.00002367, 0.00000997, -0.00000637, 0.00000777, -0.00000945, 0.00000861, -0.00000517, 0.00000635, -0.00000631, 0.00000946, -0.00000952, 0.00000717, -0.00001201, 0.00000600, 0.00000008, 0.00000722, -0.00001262, 0.00000167, -0.00000171, 0.00000094, 0.00000254, 0.00000684, -0.00001180, 0.00000763, -0.00000621, 0.00000216, 0.00000555, 0.00000078, -0.00000268, -0.00000699, },
		{ 0.00000755, -0.00001172, 0.00001004, -0.00000094, 0.00000041, 0.00002325, -0.00001449, -0.00000118, 0.00002427, -0.00002270, 0.00001592, -0.00000756, -0.00002201, 0.00001487, -0.00002109, 0.00002113, -0.00000380, -0.00000813, 0.00000972, -0.00001054, 0.00000118, -0.00000032, 0.00000768, -0.00000237, -0.00000339, 0.00000070, 0.00000009, -0.00000194, 0.00000835, -0.00000704, 0.00000458, -0.00000704, -0.00000023, -0.00000305, 0.00001338, -0.00000920, -0.00000042, -0.00000391, 0.00000492, -0.00000826, 0.00001006, -0.00000779, 0.00001363, -0.00001312, 0.00001039, -0.00001053, 0.00001285, -0.00001776, 0.00000107, -0.00000428, 0.00001828, -0.00001824, 0.00001997, -0.00001763, 0.00001919, -0.00002594, 0.00001064, -0.00001608, 0.00003144, -0.00003463, 0.00001477, -0.00002162, 0.00004071, -0.00005937, 0.00004231, -0.00004758, 0.00005564, -0.00009465, 0.00007535, -0.00013630, 0.00010932, -0.00055509, 0.00091774, 0.00005934, -0.00034175, 0.00012140, -0.00012793, 0.00005688, -0.00007251, 0.00006223, -0.00004887, 0.00002677, -0.00003365, 0.00003071, -0.00002030, 0.00001478, -0.00003339, 0.00003184, -0.00001619, 0.00000848, -0.00002689, 0.00002247, -0.00001769, 0.00001546, -0.00000984, 0.00001363, -0.00000720, 0.00000476, -0.00002052, 0.00001613, -0.00001073, 0.00000887, -0.00001003, 0.00001093, -0.00000698, 0.00000598, -0.00000346, 0.00000513, -0.00000659, 0.00000325, -0.00001036, 0.00001021, -0.00000051, 0.00000140, -0.00001215, 0.00000655, -0.00000505, 0.00000389, 0.00000579, -0.00000354, -0.00000417, 0.00000523, -0.00000783, 0.00000873, 0.00000167, -0.00000097, -0.00000408, -0.00000799, },
		{ -0.00000070, -0.00000521, 0.00001341, -0.00001189, 0.00002081, 0.00000324, -0.00001560, 0.00001783, 0.00000362, -0.00001724, 0.00002598, -0.00003743, 0.00001331, -0.00001160, -0.00000029, 0.00001845, -0.00001809, 0.00000922, -0.00000421, -0.00000437, -0.00000008, 0.00000549, 0.00000224, -0.00000289, -0.00000354, 0.00000471, -0.00000714, 0.00001054, -0.00000540, 0.00000357, -0.00000572, 0.00000091, -0.00000862, 0.00001386, -0.00000395, -0.00000055, -0.00000898, 0.00001011, -0.00001255, 0.00001220, -0.00001036, 0.00001498, -0.00001116, 0.00000899, -0.00001104, 0.00001336, -0.00001344, 0.00000125, -0.00001590, 0.00002710, -0.00002046, 0.00001998, -0.00001765, 0.00002053, -0.00002281, 0.00001114, -0.00002672, 0.00003767, -0.00003009, 0.00001677, -0.00003790, 0.00005413, -0.00005472, 0.00003362, -0.00004963, 0.00006341, -0.00008715, 0.00006507, -0.00012075, 0.00011405, -0.00035399, 0.00007804, 0.00093135, -0.00059556, 0.00014067, -0.00014271, 0.00006504, -0.00008907, 0.00006141, -0.00004337, 0.00003114, -0.00004471, 0.00003645, -0.00002800, 0.00003141, -0.00004562, 0.00002974, -0.00001489, 0.00001931, -0.00003577, 0.00002146, -0.00001621, 0.00001510, -0.00001265, 0.00001712, -0.00001136, 0.00001543, -0.00002697, 0.00001364, -0.00000890, 0.00000984, -0.00001157, 0.00001049, -0.00000646, 0.00000741, -0.00000706, 0.00001000, -0.00001020, 0.00000791, -0.00001360, 0.00000794, -0.00000055, 0.00000726, -0.00001365, 0.00000305, -0.00000251, 0.00000121, 0.00000335, 0.00000573, -0.00001165, 0.00000820, -0.00000708, 0.00000290, 0.00000596, -0.00000031, -0.00000199, -0.00000800, 0.00000095, },
		{ -0.00001192, 0.00001454, -0.00000748, 0.00000689, 0.00001427, -0.00000730, -0.00000828, 0.00002920, -0.00002742, 0.00001957, -0.00001010, -0.00001878, 0.00001485, -0.00002298, 0.00002541, -0.00000780, -0.00000527, 0.00000756, -0.00000990, 0.00000099, 0.00000062, 0.00000733, -0.00000294, -0.00000280, 0.00000061, -0.00000082, -0.00000019, 0.00000726, -0.00000664, 0.00000440, -0.00000739, -0.00000035, -0.00000192, 0.00001304, -0.00001008, -0.00000020, -0.00000337, 0.00000418, -0.00000719, 0.00000924, -0.00000701, 0.00001305, -0.00001323, 0.00001049, -0.00001040, 0.00001340, -0.00001927, 0.00000138, -0.00000259, 0.00001674, -0.00001741, 0.00001962, -0.00001772, 0.00001988, -0.00002739, 0.00001045, -0.00001380, 0.00003053, -0.00003617, 0.00001490, -0.00001936, 0.00003919, -0.00005942, 0.00004192, -0.00004606, 0.00005560, -0.00009667, 0.00007737, -0.00014091, 0.00011922, -0.00059331, 0.00100709, -0.00001748, -0.00032012, 0.00011930, -0.00012704, 0.00005729, -0.00007421, 0.00006441, -0.00004976, 0.00002646, -0.00003450, 0.00003218, -0.00001998, 0.00001255, -0.00003263, 0.00003293, -0.00001619, 0.00000647, -0.00002546, 0.00002265, -0.00001814, 0.00001605, -0.00001060, 0.00001425, -0.00000641, 0.00000242, -0.00001925, 0.00001631, -0.00001088, 0.00000878, -0.00001018, 0.00001136, -0.00000719, 0.00000611, -0.00000333, 0.00000452, -0.00000575, 0.00000187, -0.00000946, 0.00001060, -0.00000053, 0.00000027, -0.00001132, 0.00000655, -0.00000534, 0.00000445, 0.00000603, -0.00000504, -0.00000250, 0.00000393, -0.00000717, 0.00000902, 0.00000155, -0.00000161, -0.00000383, -0.00000770, 0.00000722, },
		{ 0.00000141, 0.00000800, -0.00000871, 0.00001775, 0.00000229, -0.00001364, 0.00001374, 0.00000742, -0.00002182, 0.00003074, -0.00004173, 0.00002044, -0.00001832, 0.00000591, 0.00001508, -0.00001681, 0.00000883, -0.00000489, -0.00000414, 0.00000040, 0.00000588, 0.00000153, -0.00000294, -0.00000276, 0.00000344, -0.00000622, 0.00001104, -0.00000645, 0.00000453, -0.00000677, 0.00000124, -0.00000865, 0.00001522, -0.00000576, -0.00000001, -0.00000917, 0.00001068, -0.00001334, 0.00001367, -0.00001209, 0.00001711, -0.00001377, 0.00001130, -0.00001330, 0.00001621, -0.00001630, 0.00000245, -0.00001646, 0.00002939, -0.00002391, 0.00002389, -0.00002187, 0.00002496, -0.00002713, 0.00001383, -0.00002942, 0.00004327, -0.00003703, 0.00002154, -0.00004226, 0.00006134, -0.00006380, 0.00004273, -0.00005978, 0.00007707, -0.00010356, 0.00008327, -0.00014456, 0.00014588, -0.00040365, 0.00013214, 0.00093718, -0.00065079, 0.00018392, -0.00016963, 0.00008582, -0.00010585, 0.00007457, -0.00005309, 0.00003929, -0.00005315, 0.00004350, -0.00003246, 0.00003521, -0.00005146, 0.00003577, -0.00001855, 0.00002188, -0.00003996, 0.00002658, -0.00002045, 0.00001875, -0.00001574, 0.00001937, -0.00001235, 0.00001634, -0.00002988, 0.00001720, -0.00001151, 0.00001198, -0.00001376, 0.00001249, -0.00000787, 0.00000857, -0.00000785, 0.00001045, -0.00001072, 0.00000842, -0.00001497, 0.00000985, -0.00000125, 0.00000718, -0.00001444, 0.00000435, -0.00000340, 0.00000164, 0.00000406, 0.00000445, -0.00001112, 0.00000838, -0.00000770, 0.00000365, 0.00000622, -0.00000134, -0.00000137, -0.00000884, 0.00000290, -0.00000428, },
		{ 0.00001764, -0.00001322, 0.00001299, 0.00000515, 0.00000050, -0.00001582, 0.00003459, -0.00003238, 0.00002353, -0.00001342, -0.00001417, 0.00001316, -0.00002318, 0.00002834, -0.00001119, -0.00000257, 0.00000527, -0.00000906, 0.00000074, 0.00000159, 0.00000687, -0.00000351, -0.00000203, 0.00000024, -0.00000148, 0.00000152, 0.00000597, -0.00000603, 0.00000404, -0.00000765, -0.00000044, -0.00000069, 0.00001239, -0.00001074, 0.00000001, -0.00000283, 0.00000337, -0.00000593, 0.00000810, -0.00000587, 0.00001201, -0.00001284, 0.00001013, -0.00000980, 0.00001341, -0.00002042, 0.00000162, -0.00000074, 0.00001471, -0.00001586, 0.00001847, -0.00001699, 0.00001974, -0.00002817, 0.00000979, -0.00001087, 0.00002855, -0.00003665, 0.00001427, -0.00001611, 0.00003607, -0.00005750, 0.00003953, -0.00004233, 0.00005283, -0.00009539, 0.00007550, -0.00014070, 0.00012253, -0.00062308, 0.00109204, -0.00009939, -0.00028969, 0.00011046, -0.00012139, 0.00005409, -0.00007294, 0.00006417, -0.00004884, 0.00002469, -0.00003408, 0.00003271, -0.00001903, 0.00000952, -0.00003075, 0.00003305, -0.00001563, 0.00000389, -0.00002317, 0.00002190, -0.00001781, 0.00001604, -0.00001091, 0.00001460, -0.00000550, -0.00000024, -0.00001735, 0.00001584, -0.00001055, 0.00000829, -0.00000994, 0.00001148, -0.00000719, 0.00000611, -0.00000317, 0.00000388, -0.00000483, 0.00000032, -0.00000825, 0.00001075, -0.00000055, -0.00000090, -0.00001027, 0.00000630, -0.00000548, 0.00000508, 0.00000597, -0.00000626, -0.00000091, 0.00000251, -0.00000628, 0.00000917, 0.00000134, -0.00000209, -0.00000380, -0.00000689, 0.00000605, -0.00001089, },
		{ 0.00000145, -0.00000411, 0.00001307, 0.00000286, -0.00001272, 0.00001065, 0.00001037, -0.00002528, 0.00003407, -0.00004456, 0.00002651, -0.00002458, 0.00001232, 0.00001089, -0.00001471, 0.00000780, -0.00000519, -0.00000399, 0.00000092, 0.00000622, 0.00000069, -0.00000276, -0.00000211, 0.00000213, -0.00000503, 0.00001122, -0.00000734, 0.00000547, -0.00000788, 0.00000161, -0.00000856, 0.00001639, -0.00000760, 0.00000068, -0.00000936, 0.00001111, -0.00001391, 0.00001487, -0.00001361, 0.00001901, -0.00001623, 0.00001357, -0.00001551, 0.00001906, -0.00001934, 0.00000389, -0.00001692, 0.00003125, -0.00002695, 0.00002752, -0.00002594, 0.00002936, -0.00003155, 0.00001663, -0.00003188, 0.00004842, -0.00004380, 0.00002634, -0.00004623, 0.00006777, -0.00007225, 0.00005153, -0.00006961, 0.00009030, -0.00011966, 0.00010147, -0.00016867, 0.00017857, -0.00045641, 0.00019599, 0.00092805, -0.00069534, 0.00022295, -0.00019492, 0.00010581, -0.00012216, 0.00008761, -0.00006293, 0.00004753, -0.00006161, 0.00005066, -0.00003695, 0.00003862, -0.00005674, 0.00004161, -0.00002226, 0.00002421, -0.00004367, 0.00003138, -0.00002461, 0.00002245, -0.00001893, 0.00002170, -0.00001332, 0.00001690, -0.00003229, 0.00002051, -0.00001408, 0.00001409, -0.00001592, 0.00001452, -0.00000936, 0.00000980, -0.00000867, 0.00001081, -0.00001107, 0.00000868, -0.00001606, 0.00001169, -0.00000200, 0.00000698, -0.00001499, 0.00000552, -0.00000433, 0.00000223, 0.00000459, 0.00000308, -0.00001025, 0.00000817, -0.00000805, 0.00000436, 0.00000632, -0.00000226, -0.00000088, -0.00000938, 0.00000455, -0.00000769, 0.00000832, },
		{ -0.00001773, 0.00001823, -0.00000361, 0.00000843, -0.00002338, 0.00004008, -0.00003725, 0.00002749, -0.00001720, -0.00000847, 0.00000998, -0.00002169, 0.00002974, -0.00001374, -0.00000024, 0.00000300, -0.00000809, 0.00000045, 0.00000259, 0.00000628, -0.00000404, -0.00000112, -0.00000042, -0.00000184, 0.00000313, 0.00000455, -0.00000521, 0.00000350, -0.00000782, -0.00000051, 0.00000061, 0.00001144, -0.00001115, 0.00000020, -0.00000230, 0.00000252, -0.00000452, 0.00000670, -0.00000441, 0.00001056, -0.00001195, 0.00000929, -0.00000869, 0.00001283, -0.00002112, 0.00000176, 0.00000120, 0.00001225, -0.00001365, 0.00001653, -0.00001542, 0.00001873, -0.00002821, 0.00000866, -0.00000732, 0.00002551, -0.00003603, 0.00001289, -0.00001199, 0.00003148, -0.00005367, 0.00003514, -0.00003640, 0.00004727, -0.00009067, 0.00006950, -0.00013525, 0.00011840, -0.00064259, 0.00116978, -0.00018381, -0.00025185, 0.00009539, -0.00011113, 0.00004731, -0.00006865, 0.00006143, -0.00004600, 0.00002138, -0.00003229, 0.00003222, -0.00001744, 0.00000576, -0.00002777, 0.00003220, -0.00001451, 0.00000082, -0.00002010, 0.00002024, -0.00001668, 0.00001538, -0.00001074, 0.00001467, -0.00000452, -0.00000311, -0.00001488, 0.00001476, -0.00000974, 0.00000737, -0.00000928, 0.00001126, -0.00000695, 0.00000598, -0.00000298, 0.00000322, -0.00000389, -0.00000133, -0.00000675, 0.00001067, -0.00000056, -0.00000206, -0.00000903, 0.00000582, -0.00000547, 0.00000573, 0.00000561, -0.00000712, 0.00000051, 0.00000105, -0.00000519, 0.00000918, 0.00000105, -0.00000243, -0.00000395, -0.00000563, 0.00000408, -0.00000861, 0.00001912, },
		{ 0.00000163, 0.00000696, 0.00000491, -0.00001289, 0.00000866, 0.00001233, -0.00002748, 0.00003579, -0.00004566, 0.00003115, -0.00002989, 0.00001851, 0.00000620, -0.00001197, 0.00000620, -0.00000514, -0.00000394, 0.00000150, 0.00000651, -0.00000027, -0.00000235, -0.00000168, 0.00000086, -0.00000364, 0.00001110, -0.00000804, 0.00000634, -0.00000901, 0.00000204, -0.00000836, 0.00001735, -0.00000943, 0.00000151, -0.00000957, 0.00001143, -0.00001423, 0.00001577, -0.00001484, 0.00002062, -0.00001843, 0.00001567, -0.00001758, 0.00002177, -0.00002244, 0.00000552, -0.00001726, 0.00003265, -0.00002946, 0.00003071, -0.00002967, 0.00003351, -0.00003591, 0.00001943, -0.00003400, 0.00005294, -0.00005015, 0.00003095, -0.00004961, 0.00007315, -0.00007966, 0.00005955, -0.00007861, 0.00010246, -0.00013469, 0.00011878, -0.00019199, 0.00021071, -0.00051055, 0.00026822, 0.00090391, -0.00072783, 0.00025616, -0.00021732, 0.00012401, -0.00013720, 0.00009986, -0.00007239, 0.00005543, -0.00006972, 0.00005765, -0.00004129, 0.00004151, -0.00006127, 0.00004705, -0.00002583, 0.00002619, -0.00004673, 0.00003564, -0.00002850, 0.00002603, -0.00002209, 0.00002406, -0.00001424, 0.00001708, -0.00003411, 0.00002343, -0.00001647, 0.00001606, -0.00001795, 0.00001652, -0.00001087, 0.00001106, -0.00000950, 0.00001110, -0.00001126, 0.00000869, -0.00001683, 0.00001340, -0.00000279, 0.00000668, -0.00001528, 0.00000654, -0.00000526, 0.00000298, 0.00000491, 0.00000172, -0.00000912, 0.00000761, -0.00000811, 0.00000502, 0.00000627, -0.00000302, -0.00000060, -0.00000952, 0.00000570, -0.00001062, 0.00001507, -0.00000580, },
		{ 0.00002221, -0.00001156, 0.00001601, -0.00003055, 0.00004526, -0.00004166, 0.00003111, -0.00002103, -0.00000208, 0.00000559, -0.00001862, 0.00002956, -0.00001529, 0.00000159, 0.00000089, -0.00000709, 0.00000018, 0.00000358, 0.00000556, -0.00000447, -0.00000013, -0.00000131, -0.00000186, 0.00000454, 0.00000307, -0.00000422, 0.00000277, -0.00000787, -0.00000057, 0.00000194, 0.00001021, -0.00001128, 0.00000032, -0.00000178, 0.00000168, -0.00000303, 0.00000510, -0.00000270, 0.00000874, -0.00001061, 0.00000802, -0.00000711, 0.00001166, -0.00002133, 0.00000177, 0.00000320, 0.00000948, -0.00001087, 0.00001388, -0.00001305, 0.00001686, -0.00002748, 0.00000705, -0.00000324, 0.00002151, -0.00003435, 0.00001082, -0.00000715, 0.00002561, -0.00004809, 0.00002891, -0.00002842, 0.00003907, -0.00008258, 0.00005937, -0.00012442, 0.00010628, -0.00065029, 0.00123764, -0.00026801, -0.00020832, 0.00007494, -0.00009670, 0.00003720, -0.00006145, 0.00005622, -0.00004128, 0.00001653, -0.00002910, 0.00003066, -0.00001524, 0.00000138, -0.00002378, 0.00003040, -0.00001287, -0.00000264, -0.00001634, 0.00001774, -0.00001477, 0.00001404, -0.00001004, 0.00001444, -0.00000349, -0.00000610, -0.00001195, 0.00001311, -0.00000848, 0.00000604, -0.00000820, 0.00001068, -0.00000644, 0.00000568, -0.00000274, 0.00000256, -0.00000298, -0.00000303, -0.00000501, 0.00001035, -0.00000057, -0.00000318, -0.00000765, 0.00000511, -0.00000527, 0.00000636, 0.00000497, -0.00000759, 0.00000167, -0.00000037, -0.00000396, 0.00000905, 0.00000072, -0.00000268, -0.00000420, -0.00000404, 0.00000145, -0.00000516, 0.00001886, -0.00002068, },
		{ -0.00000029, 0.00000823, -0.00001407, 0.00000775, 0.00001327, -0.00002836, 0.00003585, -0.00004488, 0.00003404, -0.00003384, 0.00002401, 0.00000134, -0.00000881, 0.00000416, -0.00000478, -0.00000398, 0.00000212, 0.00000674, -0.00000133, -0.00000168, -0.00000152, -0.00000026, -0.00000214, 0.00001069, -0.00000854, 0.00000710, -0.00001015, 0.00000252, -0.00000805, 0.00001805, -0.00001119, 0.00000245, -0.00000980, 0.00001165, -0.00001432, 0.00001635, -0.00001575, 0.00002186, -0.00002028, 0.00001751, -0.00001938, 0.00002423, -0.00002549, 0.00000728, -0.00001748, 0.00003353, -0.00003135, 0.00003330, -0.00003289, 0.00003722, -0.00004000, 0.00002208, -0.00003565, 0.00005663, -0.00005585, 0.00003520, -0.00005226, 0.00007724, -0.00008570, 0.00006636, -0.00008627, 0.00011295, -0.00014792, 0.00013428, -0.00021337, 0.00024077, -0.00056409, 0.00034698, 0.00086535, -0.00074743, 0.00028225, -0.00023570, 0.00013950, -0.00015019, 0.00011067, -0.00008093, 0.00006259, -0.00007713, 0.00006418, -0.00004530, 0.00004372, -0.00006484, 0.00005186, -0.00002912, 0.00002770, -0.00004901, 0.00003920, -0.00003192, 0.00002931, -0.00002507, 0.00002634, -0.00001508, 0.00001687, -0.00003526, 0.00002583, -0.00001857, 0.00001778, -0.00001975, 0.00001838, -0.00001232, 0.00001231, -0.00001035, 0.00001134, -0.00001131, 0.00000845, -0.00001726, 0.00001492, -0.00000359, 0.00000629, -0.00001531, 0.00000736, -0.00000616, 0.00000387, 0.00000496, 0.00000046, -0.00000782, 0.00000673, -0.00000788, 0.00000560, 0.00000606, -0.00000360, -0.00000056, -0.00000917, 0.00000621, -0.00001281, 0.00002123, -0.00001325, 0.00000809, },
		{ -0.00001828, 0.00002281, -0.00003692, 0.00004976, -0.00004524, 0.00003400, -0.00002449, 0.00000454, 0.00000039, -0.00001421, 0.00002783, -0.00001576, 0.00000279, -0.00000094, -0.00000612, -0.00000005, 0.00000456, 0.00000471, -0.00000475, 0.00000086, -0.00000238, -0.00000156, 0.00000571, 0.00000157, -0.00000310, 0.00000187, -0.00000779, -0.00000060, 0.00000329, 0.00000873, -0.00001113, 0.00000036, -0.00000128, 0.00000087, -0.00000153, 0.00000337, -0.00000080, 0.00000663, -0.00000887, 0.00000634, -0.00000507, 0.00000990, -0.00002101, 0.00000162, 0.00000518, 0.00000651, -0.00000765, 0.00001062, -0.00000997, 0.00001415, -0.00002600, 0.00000502, 0.00000125, 0.00001666, -0.00003167, 0.00000815, -0.00000179, 0.00001873, -0.00004103, 0.00002110, -0.00001867, 0.00002852, -0.00007141, 0.00004537, -0.00010837, 0.00008602, -0.00064511, 0.00129320, -0.00034918, -0.00016110, 0.00005022, -0.00007880, 0.00002421, -0.00005167, 0.00004873, -0.00003479, 0.00001024, -0.00002456, 0.00002807, -0.00001250, -0.00000348, -0.00001893, 0.00002772, -0.00001078, -0.00000638, -0.00001205, 0.00001452, -0.00001214, 0.00001207, -0.00000882, 0.00001390, -0.00000246, -0.00000908, -0.00000868, 0.00001098, -0.00000681, 0.00000435, -0.00000673, 0.00000975, -0.00000566, 0.00000519, -0.00000245, 0.00000193, -0.00000212, -0.00000471, -0.00000309, 0.00000981, -0.00000058, -0.00000422, -0.00000619, 0.00000421, -0.00000486, 0.00000691, 0.00000411, -0.00000767, 0.00000253, -0.00000167, -0.00000264, 0.00000877, 0.00000039, -0.00000290, -0.00000446, -0.00000226, -0.00000165, -0.00000073, 0.00001691, -0.00002188, 0.00002459, },
		{ 0.00001255, -0.00001610, 0.00000780, 0.00001328, -0.00002800, 0.00003432, -0.00004219, 0.00003502, -0.00003612, 0.00002842, -0.00000331, -0.00000546, 0.00000183, -0.00000418, -0.00000407, 0.00000277, 0.00000688, -0.00000245, -0.00000077, -0.00000165, -0.00000115, -0.00000061, 0.00001003, -0.00000880, 0.00000773, -0.00001125, 0.00000304, -0.00000763, 0.00001849, -0.00001282, 0.00000346, -0.00001006, 0.00001177, -0.00001419, 0.00001660, -0.00001629, 0.00002268, -0.00002171, 0.00001899, -0.00002082, 0.00002629, -0.00002834, 0.00000910, -0.00001757, 0.00003388, -0.00003253, 0.00003517, -0.00003541, 0.00004031, -0.00004365, 0.00002445, -0.00003674, 0.00005935, -0.00006069, 0.00003889, -0.00005405, 0.00007986, -0.00009006, 0.00007160, -0.00009218, 0.00012124, -0.00015864, 0.00014711, -0.00023169, 0.00026719, -0.00061485, 0.00042998, 0.00081354, -0.00075384, 0.00030031, -0.00024914, 0.00015151, -0.00016046, 0.00011947, -0.00008808, 0.00006857, -0.00008347, 0.00006997, -0.00004880, 0.00004515, -0.00006731, 0.00005584, -0.00003198, 0.00002867, -0.00005042, 0.00004189, -0.00003469, 0.00003212, -0.00002774, 0.00002847, -0.00001583, 0.00001628, -0.00003572, 0.00002761, -0.00002026, 0.00001916, -0.00002124, 0.00002003, -0.00001365, 0.00001350, -0.00001119, 0.00001152, -0.00001124, 0.00000797, -0.00001733, 0.00001622, -0.00000438, 0.00000583, -0.00001510, 0.00000797, -0.00000699, 0.00000488, 0.00000471, -0.00000060, -0.00000644, 0.00000561, -0.00000739, 0.00000609, 0.00000572, -0.00000400, -0.00000077, -0.00000832, 0.00000598, -0.00001401, 0.00002637, -0.00002036, 0.00001478, -0.00000826, },
		{ 0.00002846, -0.00004214, 0.00005325, -0.00004768, 0.00003585, -0.00002716, 0.00001090, -0.00000517, -0.00000883, 0.00002470, -0.00001517, 0.00000331, -0.00000243, -0.00000524, -0.00000022, 0.00000549, 0.00000371, -0.00000485, 0.00000176, -0.00000355, -0.00000097, 0.00000660, 0.00000013, -0.00000187, 0.00000079, -0.00000756, -0.00000061, 0.00000462, 0.00000702, -0.00001069, 0.00000029, -0.00000078, 0.00000011, -0.00000006, 0.00000159, 0.00000120, 0.00000432, -0.00000683, 0.00000434, -0.00000267, 0.00000760, -0.00002014, 0.00000131, 0.00000710, 0.00000346, -0.00000414, 0.00000690, -0.00000629, 0.00001070, -0.00002382, 0.00000263, 0.00000601, 0.00001113, -0.00002812, 0.00000499, 0.00000387, 0.00001116, -0.00003285, 0.00001207, -0.00000757, 0.00001608, -0.00005762, 0.00002798, -0.00008760, 0.00005792, -0.00062644, 0.00133445, -0.00042466, -0.00011231, 0.00002260, -0.00005831, 0.00000899, -0.00003976, 0.00003931, -0.00002680, 0.00000272, -0.00001882, 0.00002451, -0.00000931, -0.00000864, -0.00001339, 0.00002427, -0.00000832, -0.00001025, -0.00000741, 0.00001072, -0.00000890, 0.00000952, -0.00000710, 0.00001305, -0.00000148, -0.00001194, -0.00000523, 0.00000847, -0.00000481, 0.00000235, -0.00000491, 0.00000849, -0.00000462, 0.00000450, -0.00000210, 0.00000133, -0.00000137, -0.00000629, -0.00000105, 0.00000907, -0.00000059, -0.00000516, -0.00000470, 0.00000315, -0.00000425, 0.00000731, 0.00000308, -0.00000736, 0.00000303, -0.00000279, -0.00000128, 0.00000835, 0.00000011, -0.00000314, -0.00000461, -0.00000048, -0.00000498, 0.00000441, 0.00001343, -0.00002131, 0.00002520, -0.00002347, },
		{ -0.00001871, 0.00000860, 0.00001254, -0.00002656, 0.00003138, -0.00003770, 0.00003406, -0.00003653, 0.00003141, -0.00000742, -0.00000220, -0.00000060, -0.00000343, -0.00000419, 0.00000344, 0.00000693, -0.00000360, 0.00000035, -0.00000210, -0.00000174, 0.00000086, 0.00000917, -0.00000884, 0.00000820, -0.00001229, 0.00000360, -0.00000712, 0.00001863, -0.00001427, 0.00000450, -0.00001033, 0.00001182, -0.00001386, 0.00001651, -0.00001645, 0.00002304, -0.00002264, 0.00002002, -0.00002179, 0.00002786, -0.00003088, 0.00001089, -0.00001751, 0.00003371, -0.00003296, 0.00003624, -0.00003712, 0.00004260, -0.00004669, 0.00002641, -0.00003718, 0.00006099, -0.00006450, 0.00004189, -0.00005489, 0.00008091, -0.00009256, 0.00007499, -0.00009600, 0.00012688, -0.00016629, 0.00015652, -0.00024594, 0.00028850, -0.00066062, 0.00051461, 0.00075020, -0.00074732, 0.00030982, -0.00025704, 0.00015945, -0.00016746, 0.00012576, -0.00009342, 0.00007304, -0.00008841, 0.00007475, -0.00005163, 0.00004571, -0.00006855, 0.00005884, -0.00003429, 0.00002904, -0.00005089, 0.00004362, -0.00003668, 0.00003432, -0.00002997, 0.00003034, -0.00001645, 0.00001534, -0.00003547, 0.00002872, -0.00002147, 0.00002012, -0.00002233, 0.00002138, -0.00001479, 0.00001456, -0.00001199, 0.00001165, -0.00001108, 0.00000730, -0.00001705, 0.00001726, -0.00000514, 0.00000531, -0.00001467, 0.00000836, -0.00000770, 0.00000596, 0.00000415, -0.00000140, -0.00000510, 0.00000433, -0.00000667, 0.00000647, 0.00000526, -0.00000423, -0.00000122, -0.00000695, 0.00000497, -0.00001409, 0.00003016, -0.00002663, 0.00002120, -0.00001648, 0.00001746, },
		{ -0.00004593, 0.00005546, -0.00004870, 0.00003635, -0.00002865, 0.00001654, -0.00001058, -0.00000290, 0.00002045, -0.00001360, 0.00000316, -0.00000352, -0.00000450, -0.00000028, 0.00000635, 0.00000260, -0.00000472, 0.00000249, -0.00000473, -0.00000014, 0.00000720, -0.00000121, -0.00000059, -0.00000043, -0.00000719, -0.00000061, 0.00000589, 0.00000514, -0.00000996, 0.00000009, -0.00000028, -0.00000056, 0.00000131, -0.00000016, 0.00000321, 0.00000191, -0.00000457, 0.00000211, 0.00000001, 0.00000482, -0.00001875, 0.00000083, 0.00000890, 0.00000045, -0.00000050, 0.00000289, -0.00000218, 0.00000665, -0.00002103, -0.00000002, 0.00001088, 0.00000513, -0.00002384, 0.00000151, 0.00000957, 0.00000326, -0.00002395, 0.00000227, 0.00000437, 0.00000233, -0.00004187, 0.00000795, -0.00006294, 0.00002270, -0.00059425, 0.00135983, -0.00049198, -0.00006409, -0.00000647, -0.00003628, -0.00000768, -0.00002632, 0.00002841, -0.00001766, -0.00000573, -0.00001207, 0.00002012, -0.00000582, -0.00001389, -0.00000737, 0.00002020, -0.00000563, -0.00001409, -0.00000260, 0.00000653, -0.00000520, 0.00000648, -0.00000493, 0.00001191, -0.00000058, -0.00001457, -0.00000172, 0.00000572, -0.00000258, 0.00000013, -0.00000280, 0.00000693, -0.00000332, 0.00000362, -0.00000167, 0.00000077, -0.00000074, -0.00000773, 0.00000104, 0.00000815, -0.00000060, -0.00000598, -0.00000321, 0.00000197, -0.00000343, 0.00000753, 0.00000198, -0.00000672, 0.00000318, -0.00000369, 0.00000007, 0.00000780, -0.00000012, -0.00000347, -0.00000456, 0.00000113, -0.00000826, 0.00000991, 0.00000869, -0.00001905, 0.00002397, -0.00002695, 0.00003269, },
		{ 0.00000987, 0.00001130, -0.00002431, 0.00002732, -0.00003167, 0.00003125, -0.00003500, 0.00003275, -0.00001072, 0.00000076, -0.00000298, -0.00000262, -0.00000428, 0.00000409, 0.00000689, -0.00000472, 0.00000161, -0.00000286, -0.00000199, 0.00000218, 0.00000816, -0.00000866, 0.00000847, -0.00001323, 0.00000417, -0.00000653, 0.00001848, -0.00001548, 0.00000551, -0.00001060, 0.00001181, -0.00001336, 0.00001612, -0.00001623, 0.00002293, -0.00002306, 0.00002057, -0.00002225, 0.00002882, -0.00003298, 0.00001258, -0.00001730, 0.00003303, -0.00003266, 0.00003644, -0.00003792, 0.00004396, -0.00004898, 0.00002788, -0.00003693, 0.00006146, -0.00006713, 0.00004407, -0.00005475, 0.00008037, -0.00009310, 0.00007637, -0.00009750, 0.00012958, -0.00017044, 0.00016193, -0.00025527, 0.00030337, -0.00069923, 0.00059808, 0.00067753, -0.00072869, 0.00031072, -0.00025908, 0.00016296, -0.00017085, 0.00012920, -0.00009663, 0.00007570, -0.00009168, 0.00007827, -0.00005364, 0.00004535, -0.00006849, 0.00006074, -0.00003596, 0.00002877, -0.00005040, 0.00004431, -0.00003777, 0.00003579, -0.00003164, 0.00003189, -0.00001695, 0.00001409, -0.00003454, 0.00002913, -0.00002214, 0.00002060, -0.00002295, 0.00002237, -0.00001567, 0.00001545, -0.00001272, 0.00001175, -0.00001086, 0.00000647, -0.00001642, 0.00001802, -0.00000586, 0.00000475, -0.00001403, 0.00000853, -0.00000827, 0.00000707, 0.00000330, -0.00000187, -0.00000388, 0.00000297, -0.00000576, 0.00000673, 0.00000471, -0.00000431, -0.00000186, -0.00000514, 0.00000319, -0.00001299, 0.00003233, -0.00003163, 0.00002684, -0.00002445, 0.00002254, -0.00002157, },
		{ 0.00005621, -0.00004814, 0.00003533, -0.00002864, 0.00002101, -0.00001535, 0.00000310, 0.00001539, -0.00001122, 0.00000240, -0.00000422, -0.00000391, -0.00000025, 0.00000713, 0.00000138, -0.00000434, 0.00000298, -0.00000582, 0.00000087, 0.00000750, -0.00000241, 0.00000071, -0.00000177, -0.00000666, -0.00000061, 0.00000708, 0.00000312, -0.00000896, -0.00000025, 0.00000023, -0.00000116, 0.00000255, -0.00000182, 0.00000515, -0.00000050, -0.00000222, -0.00000024, 0.00000285, 0.00000167, -0.00001687, 0.00000020, 0.00001052, -0.00000238, 0.00000309, -0.00000121, 0.00000216, 0.00000215, -0.00001774, -0.00000281, 0.00001567, -0.00000112, -0.00001903, -0.00000215, 0.00001507, -0.00000461, -0.00001476, -0.00000781, 0.00001656, -0.00001205, -0.00002493, -0.00001382, -0.00003546, -0.00001847, -0.00054905, 0.00136840, -0.00054905, -0.00001847, -0.00003546, -0.00001382, -0.00002493, -0.00001205, 0.00001656, -0.00000781, -0.00001476, -0.00000461, 0.00001507, -0.00000215, -0.00001903, -0.00000112, 0.00001567, -0.00000281, -0.00001774, 0.00000215, 0.00000216, -0.00000121, 0.00000309, -0.00000238, 0.00001052, 0.00000020, -0.00001687, 0.00000167, 0.00000285, -0.00000024, -0.00000222, -0.00000050, 0.00000515, -0.00000182, 0.00000255, -0.00000116, 0.00000023, -0.00000025, -0.00000896, 0.00000312, 0.00000708, -0.00000061, -0.00000666, -0.00000177, 0.00000071, -0.00000241, 0.00000750, 0.00000087, -0.00000582, 0.00000298, -0.00000434, 0.00000138, 0.00000713, -0.00000025, -0.00000391, -0.00000422, 0.00000240, -0.00001122, 0.00001539, 0.00000310, -0.00001535, 0.00002101, -0.00002864, 0.00003533, -0.00004814, },
		{ 0.00000987, -0.00002157, 0.00002254, -0.00002445, 0.00002684, -0.00003163, 0.00003233, -0.00001299, 0.00000319, -0.00000514, -0.00000186, -0.00000431, 0.00000471, 0.00000673, -0.00000576, 0.00000297, -0.00000388, -0.00000187, 0.00000330, 0.00000707, -0.00000827, 0.00000853, -0.00001403, 0.00000475, -0.00000586, 0.00001802, -0.00001642, 0.00000647, -0.00001086, 0.00001175, -0.00001272, 0.00001545, -0.00001567, 0.00002237, -0.00002295, 0.00002060, -0.00002214, 0.00002913, -0.00003454, 0.00001409, -0.00001695, 0.00003189, -0.00003164, 0.00003579, -0.00003777, 0.00004431, -0.00005040, 0.00002877, -0.00003596, 0.00006074, -0.00006849, 0.00004535, -0.00005364, 0.00007827, -0.00009168, 0.00007570, -0.00009663, 0.00012920, -0.00017085, 0.00016296, -0.00025908, 0.00031072, -0.00072869, 0.00067753, 0.00059808, -0.00069923, 0.00030337, -0.00025527, 0.00016193, -0.00017044, 0.00012958, -0.00009750, 0.00007637, -0.00009310, 0.00008037, -0.00005475, 0.00004407, -0.00006713, 0.00006146, -0.00003693, 0.00002788, -0.00004898, 0.00004396, -0.00003792, 0.00003644, -0.00003266, 0.00003303, -0.00001730, 0.00001258, -0.00003298, 0.00002882, -0.00002225, 0.00002057, -0.00002306, 0.00002293, -0.00001623, 0.00001612, -0.00001336, 0.00001181, -0.00001060, 0.00000551, -0.00001548, 0.00001848, -0.00000653, 0.00000417, -0.00001323, 0.00000847, -0.00000866, 0.00000816, 0.00000218, -0.00000199, -0.00000286, 0.00000161, -0.00000472, 0.00000689, 0.00000409, -0.00000428, -0.00000262, -0.00000298, 0.00000076, -0.00001072, 0.00003275, -0.00003500, 0.00003125, -0.00003167, 0.00002732, -0.00002431, 0.00001130, },
		{ -0.00004593, 0.00003269, -0.00002695, 0.00002397, -0.00001905, 0.00000869, 0.00000991, -0.00000826, 0.00000113, -0.00000456, -0.00000347, -0.00000012, 0.00000780, 0.00000007, -0.00000369, 0.00000318, -0.00000672, 0.00000198, 0.00000753, -0.00000343, 0.00000197, -0.00000321, -0.00000598, -0.00000060, 0.00000815, 0.00000104, -0.00000773, -0.00000074, 0.00000077, -0.00000167, 0.00000362, -0.00000332, 0.00000693, -0.00000280, 0.00000013, -0.00000258, 0.00000572, -0.00000172, -0.00001457, -0.00000058, 0.00001191, -0.00000493, 0.00000648, -0.00000520, 0.00000653, -0.00000260, -0.00001409, -0.00000563, 0.00002020, -0.00000737, -0.00001389, -0.00000582, 0.00002012, -0.00001207, -0.00000573, -0.00001766, 0.00002841, -0.00002632, -0.00000768, -0.00003628, -0.00000647, -0.00006409, -0.00049198, 0.00135983, -0.00059425, 0.00002270, -0.00006294, 0.00000795, -0.00004187, 0.00000233, 0.00000437, 0.00000227, -0.00002395, 0.00000326, 0.00000957, 0.00000151, -0.00002384, 0.00000513, 0.00001088, -0.00000002, -0.00002103, 0.00000665, -0.00000218, 0.00000289, -0.00000050, 0.00000045, 0.00000890, 0.00000083, -0.00001875, 0.00000482, 0.00000001, 0.00000211, -0.00000457, 0.00000191, 0.00000321, -0.00000016, 0.00000131, -0.00000056, -0.00000028, 0.00000009, -0.00000996, 0.00000514, 0.00000589, -0.00000061, -0.00000719, -0.00000043, -0.00000059, -0.00000121, 0.00000720, -0.00000014, -0.00000473, 0.00000249, -0.00000472, 0.00000260, 0.00000635, -0.00000028, -0.00000450, -0.00000352, 0.00000316, -0.00001360, 0.00002045, -0.00000290, -0.00001058, 0.00001654, -0.00002865, 0.00003635, -0.00004870, 0.00005546, },
		{ -0.00001871, 0.00001746, -0.00001648, 0.00002120, -0.00002663, 0.00003016, -0.00001409, 0.00000497, -0.00000695, -0.00000122, -0.00000423, 0.00000526, 0.00000647, -0.00000667, 0.00000433, -0.00000510, -0.00000140, 0.00000415, 0.00000596, -0.00000770, 0.00000836, -0.00001467, 0.00000531, -0.00000514, 0.00001726, -0.00001705, 0.00000730, -0.00001108, 0.00001165, -0.00001199, 0.00001456, -0.00001479, 0.00002138, -0.00002233, 0.00002012, -0.00002147, 0.00002872, -0.00003547, 0.00001534, -0.00001645, 0.00003034, -0.00002997, 0.00003432, -0.00003668, 0.00004362, -0.00005089, 0.00002904, -0.00003429, 0.00005884, -0.00006855, 0.00004571, -0.00005163, 0.00007475, -0.00008841, 0.00007304, -0.00009342, 0.00012576, -0.00016746, 0.00015945, -0.00025704, 0.00030982, -0.00074732, 0.00075020, 0.00051461, -0.00066062, 0.00028850, -0.00024594, 0.00015652, -0.00016629, 0.00012688, -0.00009600, 0.00007499, -0.00009256, 0.00008091, -0.00005489, 0.00004189, -0.00006450, 0.00006099, -0.00003718, 0.00002641, -0.00004669, 0.00004260, -0.00003712, 0.00003624, -0.00003296, 0.00003371, -0.00001751, 0.00001089, -0.00003088, 0.00002786, -0.00002179, 0.00002002, -0.00002264, 0.00002304, -0.00001645, 0.00001651, -0.00001386, 0.00001182, -0.00001033, 0.00000450, -0.00001427, 0.00001863, -0.00000712, 0.00000360, -0.00001229, 0.00000820, -0.00000884, 0.00000917, 0.00000086, -0.00000174, -0.00000210, 0.00000035, -0.00000360, 0.00000693, 0.00000344, -0.00000419, -0.00000343, -0.00000060, -0.00000220, -0.00000742, 0.00003141, -0.00003653, 0.00003406, -0.00003770, 0.00003138, -0.00002656, 0.00001254, 0.00000860, },
		{ 0.00002846, -0.00002347, 0.00002520, -0.00002131, 0.00001343, 0.00000441, -0.00000498, -0.00000048, -0.00000461, -0.00000314, 0.00000011, 0.00000835, -0.00000128, -0.00000279, 0.00000303, -0.00000736, 0.00000308, 0.00000731, -0.00000425, 0.00000315, -0.00000470, -0.00000516, -0.00000059, 0.00000907, -0.00000105, -0.00000629, -0.00000137, 0.00000133, -0.00000210, 0.00000450, -0.00000462, 0.00000849, -0.00000491, 0.00000235, -0.00000481, 0.00000847, -0.00000523, -0.00001194, -0.00000148, 0.00001305, -0.00000710, 0.00000952, -0.00000890, 0.00001072, -0.00000741, -0.00001025, -0.00000832, 0.00002427, -0.00001339, -0.00000864, -0.00000931, 0.00002451, -0.00001882, 0.00000272, -0.00002680, 0.00003931, -0.00003976, 0.00000899, -0.00005831, 0.00002260, -0.00011231, -0.00042466, 0.00133445, -0.00062644, 0.00005792, -0.00008760, 0.00002798, -0.00005762, 0.00001608, -0.00000757, 0.00001207, -0.00003285, 0.00001116, 0.00000387, 0.00000499, -0.00002812, 0.00001113, 0.00000601, 0.00000263, -0.00002382, 0.00001070, -0.00000629, 0.00000690, -0.00000414, 0.00000346, 0.00000710, 0.00000131, -0.00002014, 0.00000760, -0.00000267, 0.00000434, -0.00000683, 0.00000432, 0.00000120, 0.00000159, -0.00000006, 0.00000011, -0.00000078, 0.00000029, -0.00001069, 0.00000702, 0.00000462, -0.00000061, -0.00000756, 0.00000079, -0.00000187, 0.00000013, 0.00000660, -0.00000097, -0.00000355, 0.00000176, -0.00000485, 0.00000371, 0.00000549, -0.00000022, -0.00000524, -0.00000243, 0.00000331, -0.00001517, 0.00002470, -0.00000883, -0.00000517, 0.00001090, -0.00002716, 0.00003585, -0.00004768, 0.00005325, -0.00004214, },
		{ 0.00001255, -0.00000826, 0.00001478, -0.00002036, 0.00002637, -0.00001401, 0.00000598, -0.00000832, -0.00000077, -0.00000400, 0.00000572, 0.00000609, -0.00000739, 0.00000561, -0.00000644, -0.00000060, 0.00000471, 0.00000488, -0.00000699, 0.00000797, -0.00001510, 0.00000583, -0.00000438, 0.00001622, -0.00001733, 0.00000797, -0.00001124, 0.00001152, -0.00001119, 0.00001350, -0.00001365, 0.00002003, -0.00002124, 0.00001916, -0.00002026, 0.00002761, -0.00003572, 0.00001628, -0.00001583, 0.00002847, -0.00002774, 0.00003212, -0.00003469, 0.00004189, -0.00005042, 0.00002867, -0.00003198, 0.00005584, -0.00006731, 0.00004515, -0.00004880, 0.00006997, -0.00008347, 0.00006857, -0.00008808, 0.00011947, -0.00016046, 0.00015151, -0.00024914, 0.00030031, -0.00075384, 0.00081354, 0.00042998, -0.00061485, 0.00026719, -0.00023169, 0.00014711, -0.00015864, 0.00012124, -0.00009218, 0.00007160, -0.00009006, 0.00007986, -0.00005405, 0.00003889, -0.00006069, 0.00005935, -0.00003674, 0.00002445, -0.00004365, 0.00004031, -0.00003541, 0.00003517, -0.00003253, 0.00003388, -0.00001757, 0.00000910, -0.00002834, 0.00002629, -0.00002082, 0.00001899, -0.00002171, 0.00002268, -0.00001629, 0.00001660, -0.00001419, 0.00001177, -0.00001006, 0.00000346, -0.00001282, 0.00001849, -0.00000763, 0.00000304, -0.00001125, 0.00000773, -0.00000880, 0.00001003, -0.00000061, -0.00000115, -0.00000165, -0.00000077, -0.00000245, 0.00000688, 0.00000277, -0.00000407, -0.00000418, 0.00000183, -0.00000546, -0.00000331, 0.00002842, -0.00003612, 0.00003502, -0.00004219, 0.00003432, -0.00002800, 0.00001328, 0.00000780, -0.00001610, },
		{ -0.00001828, 0.00002459, -0.00002188, 0.00001691, -0.00000073, -0.00000165, -0.00000226, -0.00000446, -0.00000290, 0.00000039, 0.00000877, -0.00000264, -0.00000167, 0.00000253, -0.00000767, 0.00000411, 0.00000691, -0.00000486, 0.00000421, -0.00000619, -0.00000422, -0.00000058, 0.00000981, -0.00000309, -0.00000471, -0.00000212, 0.00000193, -0.00000245, 0.00000519, -0.00000566, 0.00000975, -0.00000673, 0.00000435, -0.00000681, 0.00001098, -0.00000868, -0.00000908, -0.00000246, 0.00001390, -0.00000882, 0.00001207, -0.00001214, 0.00001452, -0.00001205, -0.00000638, -0.00001078, 0.00002772, -0.00001893, -0.00000348, -0.00001250, 0.00002807, -0.00002456, 0.00001024, -0.00003479, 0.00004873, -0.00005167, 0.00002421, -0.00007880, 0.00005022, -0.00016110, -0.00034918, 0.00129320, -0.00064511, 0.00008602, -0.00010837, 0.00004537, -0.00007141, 0.00002852, -0.00001867, 0.00002110, -0.00004103, 0.00001873, -0.00000179, 0.00000815, -0.00003167, 0.00001666, 0.00000125, 0.00000502, -0.00002600, 0.00001415, -0.00000997, 0.00001062, -0.00000765, 0.00000651, 0.00000518, 0.00000162, -0.00002101, 0.00000990, -0.00000507, 0.00000634, -0.00000887, 0.00000663, -0.00000080, 0.00000337, -0.00000153, 0.00000087, -0.00000128, 0.00000036, -0.00001113, 0.00000873, 0.00000329, -0.00000060, -0.00000779, 0.00000187, -0.00000310, 0.00000157, 0.00000571, -0.00000156, -0.00000238, 0.00000086, -0.00000475, 0.00000471, 0.00000456, -0.00000005, -0.00000612, -0.00000094, 0.00000279, -0.00001576, 0.00002783, -0.00001421, 0.00000039, 0.00000454, -0.00002449, 0.00003400, -0.00004524, 0.00004976, -0.00003692, 0.00002281, },
		{ -0.00000029, 0.00000809, -0.00001325, 0.00002123, -0.00001281, 0.00000621, -0.00000917, -0.00000056, -0.00000360, 0.00000606, 0.00000560, -0.00000788, 0.00000673, -0.00000782, 0.00000046, 0.00000496, 0.00000387, -0.00000616, 0.00000736, -0.00001531, 0.00000629, -0.00000359, 0.00001492, -0.00001726, 0.00000845, -0.00001131, 0.00001134, -0.00001035, 0.00001231, -0.00001232, 0.00001838, -0.00001975, 0.00001778, -0.00001857, 0.00002583, -0.00003526, 0.00001687, -0.00001508, 0.00002634, -0.00002507, 0.00002931, -0.00003192, 0.00003920, -0.00004901, 0.00002770, -0.00002912, 0.00005186, -0.00006484, 0.00004372, -0.00004530, 0.00006418, -0.00007713, 0.00006259, -0.00008093, 0.00011067, -0.00015019, 0.00013950, -0.00023570, 0.00028225, -0.00074743, 0.00086535, 0.00034698, -0.00056409, 0.00024077, -0.00021337, 0.00013428, -0.00014792, 0.00011295, -0.00008627, 0.00006636, -0.00008570, 0.00007724, -0.00005226, 0.00003520, -0.00005585, 0.00005663, -0.00003565, 0.00002208, -0.00004000, 0.00003722, -0.00003289, 0.00003330, -0.00003135, 0.00003353, -0.00001748, 0.00000728, -0.00002549, 0.00002423, -0.00001938, 0.00001751, -0.00002028, 0.00002186, -0.00001575, 0.00001635, -0.00001432, 0.00001165, -0.00000980, 0.00000245, -0.00001119, 0.00001805, -0.00000805, 0.00000252, -0.00001015, 0.00000710, -0.00000854, 0.00001069, -0.00000214, -0.00000026, -0.00000152, -0.00000168, -0.00000133, 0.00000674, 0.00000212, -0.00000398, -0.00000478, 0.00000416, -0.00000881, 0.00000134, 0.00002401, -0.00003384, 0.00003404, -0.00004488, 0.00003585, -0.00002836, 0.00001327, 0.00000775, -0.00001407, 0.00000823, },
		{ 0.00002221, -0.00002068, 0.00001886, -0.00000516, 0.00000145, -0.00000404, -0.00000420, -0.00000268, 0.00000072, 0.00000905, -0.00000396, -0.00000037, 0.00000167, -0.00000759, 0.00000497, 0.00000636, -0.00000527, 0.00000511, -0.00000765, -0.00000318, -0.00000057, 0.00001035, -0.00000501, -0.00000303, -0.00000298, 0.00000256, -0.00000274, 0.00000568, -0.00000644, 0.00001068, -0.00000820, 0.00000604, -0.00000848, 0.00001311, -0.00001195, -0.00000610, -0.00000349, 0.00001444, -0.00001004, 0.00001404, -0.00001477, 0.00001774, -0.00001634, -0.00000264, -0.00001287, 0.00003040, -0.00002378, 0.00000138, -0.00001524, 0.00003066, -0.00002910, 0.00001653, -0.00004128, 0.00005622, -0.00006145, 0.00003720, -0.00009670, 0.00007494, -0.00020832, -0.00026801, 0.00123764, -0.00065029, 0.00010628, -0.00012442, 0.00005937, -0.00008258, 0.00003907, -0.00002842, 0.00002891, -0.00004809, 0.00002561, -0.00000715, 0.00001082, -0.00003435, 0.00002151, -0.00000324, 0.00000705, -0.00002748, 0.00001686, -0.00001305, 0.00001388, -0.00001087, 0.00000948, 0.00000320, 0.00000177, -0.00002133, 0.00001166, -0.00000711, 0.00000802, -0.00001061, 0.00000874, -0.00000270, 0.00000510, -0.00000303, 0.00000168, -0.00000178, 0.00000032, -0.00001128, 0.00001021, 0.00000194, -0.00000057, -0.00000787, 0.00000277, -0.00000422, 0.00000307, 0.00000454, -0.00000186, -0.00000131, -0.00000013, -0.00000447, 0.00000556, 0.00000358, 0.00000018, -0.00000709, 0.00000089, 0.00000159, -0.00001529, 0.00002956, -0.00001862, 0.00000559, -0.00000208, -0.00002103, 0.00003111, -0.00004166, 0.00004526, -0.00003055, 0.00001601, -0.00001156, },
		{ 0.00000163, -0.00000580, 0.00001507, -0.00001062, 0.00000570, -0.00000952, -0.00000060, -0.00000302, 0.00000627, 0.00000502, -0.00000811, 0.00000761, -0.00000912, 0.00000172, 0.00000491, 0.00000298, -0.00000526, 0.00000654, -0.00001528, 0.00000668, -0.00000279, 0.00001340, -0.00001683, 0.00000869, -0.00001126, 0.00001110, -0.00000950, 0.00001106, -0.00001087, 0.00001652, -0.00001795, 0.00001606, -0.00001647, 0.00002343, -0.00003411, 0.00001708, -0.00001424, 0.00002406, -0.00002209, 0.00002603, -0.00002850, 0.00003564, -0.00004673, 0.00002619, -0.00002583, 0.00004705, -0.00006127, 0.00004151, -0.00004129, 0.00005765, -0.00006972, 0.00005543, -0.00007239, 0.00009986, -0.00013720, 0.00012401, -0.00021732, 0.00025616, -0.00072783, 0.00090391, 0.00026822, -0.00051055, 0.00021071, -0.00019199, 0.00011878, -0.00013469, 0.00010246, -0.00007861, 0.00005955, -0.00007966, 0.00007315, -0.00004961, 0.00003095, -0.00005015, 0.00005294, -0.00003400, 0.00001943, -0.00003591, 0.00003351, -0.00002967, 0.00003071, -0.00002946, 0.00003265, -0.00001726, 0.00000552, -0.00002244, 0.00002177, -0.00001758, 0.00001567, -0.00001843, 0.00002062, -0.00001484, 0.00001577, -0.00001423, 0.00001143, -0.00000957, 0.00000151, -0.00000943, 0.00001735, -0.00000836, 0.00000204, -0.00000901, 0.00000634, -0.00000804, 0.00001110, -0.00000364, 0.00000086, -0.00000168, -0.00000235, -0.00000027, 0.00000651, 0.00000150, -0.00000394, -0.00000514, 0.00000620, -0.00001197, 0.00000620, 0.00001851, -0.00002989, 0.00003115, -0.00004566, 0.00003579, -0.00002748, 0.00001233, 0.00000866, -0.00001289, 0.00000491, 0.00000696, },
		{ -0.00001773, 0.00001912, -0.00000861, 0.00000408, -0.00000563, -0.00000395, -0.00000243, 0.00000105, 0.00000918, -0.00000519, 0.00000105, 0.00000051, -0.00000712, 0.00000561, 0.00000573, -0.00000547, 0.00000582, -0.00000903, -0.00000206, -0.00000056, 0.00001067, -0.00000675, -0.00000133, -0.00000389, 0.00000322, -0.00000298, 0.00000598, -0.00000695, 0.00001126, -0.00000928, 0.00000737, -0.00000974, 0.00001476, -0.00001488, -0.00000311, -0.00000452, 0.00001467, -0.00001074, 0.00001538, -0.00001668, 0.00002024, -0.00002010, 0.00000082, -0.00001451, 0.00003220, -0.00002777, 0.00000576, -0.00001744, 0.00003222, -0.00003229, 0.00002138, -0.00004600, 0.00006143, -0.00006865, 0.00004731, -0.00011113, 0.00009539, -0.00025185, -0.00018381, 0.00116978, -0.00064259, 0.00011840, -0.00013525, 0.00006950, -0.00009067, 0.00004727, -0.00003640, 0.00003514, -0.00005367, 0.00003148, -0.00001199, 0.00001289, -0.00003603, 0.00002551, -0.00000732, 0.00000866, -0.00002821, 0.00001873, -0.00001542, 0.00001653, -0.00001365, 0.00001225, 0.00000120, 0.00000176, -0.00002112, 0.00001283, -0.00000869, 0.00000929, -0.00001195, 0.00001056, -0.00000441, 0.00000670, -0.00000452, 0.00000252, -0.00000230, 0.00000020, -0.00001115, 0.00001144, 0.00000061, -0.00000051, -0.00000782, 0.00000350, -0.00000521, 0.00000455, 0.00000313, -0.00000184, -0.00000042, -0.00000112, -0.00000404, 0.00000628, 0.00000259, 0.00000045, -0.00000809, 0.00000300, -0.00000024, -0.00001374, 0.00002974, -0.00002169, 0.00000998, -0.00000847, -0.00001720, 0.00002749, -0.00003725, 0.00004008, -0.00002338, 0.00000843, -0.00000361, 0.00001823, },
		{ 0.00000145, 0.00000832, -0.00000769, 0.00000455, -0.00000938, -0.00000088, -0.00000226, 0.00000632, 0.00000436, -0.00000805, 0.00000817, -0.00001025, 0.00000308, 0.00000459, 0.00000223, -0.00000433, 0.00000552, -0.00001499, 0.00000698, -0.00000200, 0.00001169, -0.00001606, 0.00000868, -0.00001107, 0.00001081, -0.00000867, 0.00000980, -0.00000936, 0.00001452, -0.00001592, 0.00001409, -0.00001408, 0.00002051, -0.00003229, 0.00001690, -0.00001332, 0.00002170, -0.00001893, 0.00002245, -0.00002461, 0.00003138, -0.00004367, 0.00002421, -0.00002226, 0.00004161, -0.00005674, 0.00003862, -0.00003695, 0.00005066, -0.00006161, 0.00004753, -0.00006293, 0.00008761, -0.00012216, 0.00010581, -0.00019492, 0.00022295, -0.00069534, 0.00092805, 0.00019599, -0.00045641, 0.00017857, -0.00016867, 0.00010147, -0.00011966, 0.00009030, -0.00006961, 0.00005153, -0.00007225, 0.00006777, -0.00004623, 0.00002634, -0.00004380, 0.00004842, -0.00003188, 0.00001663, -0.00003155, 0.00002936, -0.00002594, 0.00002752, -0.00002695, 0.00003125, -0.00001692, 0.00000389, -0.00001934, 0.00001906, -0.00001551, 0.00001357, -0.00001623, 0.00001901, -0.00001361, 0.00001487, -0.00001391, 0.00001111, -0.00000936, 0.00000068, -0.00000760, 0.00001639, -0.00000856, 0.00000161, -0.00000788, 0.00000547, -0.00000734, 0.00001122, -0.00000503, 0.00000213, -0.00000211, -0.00000276, 0.00000069, 0.00000622, 0.00000092, -0.00000399, -0.00000519, 0.00000780, -0.00001471, 0.00001089, 0.00001232, -0.00002458, 0.00002651, -0.00004456, 0.00003407, -0.00002528, 0.00001037, 0.00001065, -0.00001272, 0.00000286, 0.00001307, -0.00000411, },
		{ 0.00001764, -0.00001089, 0.00000605, -0.00000689, -0.00000380, -0.00000209, 0.00000134, 0.00000917, -0.00000628, 0.00000251, -0.00000091, -0.00000626, 0.00000597, 0.00000508, -0.00000548, 0.00000630, -0.00001027, -0.00000090, -0.00000055, 0.00001075, -0.00000825, 0.00000032, -0.00000483, 0.00000388, -0.00000317, 0.00000611, -0.00000719, 0.00001148, -0.00000994, 0.00000829, -0.00001055, 0.00001584, -0.00001735, -0.00000024, -0.00000550, 0.00001460, -0.00001091, 0.00001604, -0.00001781, 0.00002190, -0.00002317, 0.00000389, -0.00001563, 0.00003305, -0.00003075, 0.00000952, -0.00001903, 0.00003271, -0.00003408, 0.00002469, -0.00004884, 0.00006417, -0.00007294, 0.00005409, -0.00012139, 0.00011046, -0.00028969, -0.00009939, 0.00109204, -0.00062308, 0.00012253, -0.00014070, 0.00007550, -0.00009539, 0.00005283, -0.00004233, 0.00003953, -0.00005750, 0.00003607, -0.00001611, 0.00001427, -0.00003665, 0.00002855, -0.00001087, 0.00000979, -0.00002817, 0.00001974, -0.00001699, 0.00001847, -0.00001586, 0.00001471, -0.00000074, 0.00000162, -0.00002042, 0.00001341, -0.00000980, 0.00001013, -0.00001284, 0.00001201, -0.00000587, 0.00000810, -0.00000593, 0.00000337, -0.00000283, 0.00000001, -0.00001074, 0.00001239, -0.00000069, -0.00000044, -0.00000765, 0.00000404, -0.00000603, 0.00000597, 0.00000152, -0.00000148, 0.00000024, -0.00000203, -0.00000351, 0.00000687, 0.00000159, 0.00000074, -0.00000906, 0.00000527, -0.00000257, -0.00001119, 0.00002834, -0.00002318, 0.00001316, -0.00001417, -0.00001342, 0.00002353, -0.00003238, 0.00003459, -0.00001582, 0.00000050, 0.00000515, 0.00001299, -0.00001322, },
		{ 0.00000141, -0.00000428, 0.00000290, -0.00000884, -0.00000137, -0.00000134, 0.00000622, 0.00000365, -0.00000770, 0.00000838, -0.00001112, 0.00000445, 0.00000406, 0.00000164, -0.00000340, 0.00000435, -0.00001444, 0.00000718, -0.00000125, 0.00000985, -0.00001497, 0.00000842, -0.00001072, 0.00001045, -0.00000785, 0.00000857, -0.00000787, 0.00001249, -0.00001376, 0.00001198, -0.00001151, 0.00001720, -0.00002988, 0.00001634, -0.00001235, 0.00001937, -0.00001574, 0.00001875, -0.00002045, 0.00002658, -0.00003996, 0.00002188, -0.00001855, 0.00003577, -0.00005146, 0.00003521, -0.00003246, 0.00004350, -0.00005315, 0.00003929, -0.00005309, 0.00007457, -0.00010585, 0.00008582, -0.00016963, 0.00018392, -0.00065079, 0.00093718, 0.00013214, -0.00040365, 0.00014588, -0.00014456, 0.00008327, -0.00010356, 0.00007707, -0.00005978, 0.00004273, -0.00006380, 0.00006134, -0.00004226, 0.00002154, -0.00003703, 0.00004327, -0.00002942, 0.00001383, -0.00002713, 0.00002496, -0.00002187, 0.00002389, -0.00002391, 0.00002939, -0.00001646, 0.00000245, -0.00001630, 0.00001621, -0.00001330, 0.00001130, -0.00001377, 0.00001711, -0.00001209, 0.00001367, -0.00001334, 0.00001068, -0.00000917, -0.00000001, -0.00000576, 0.00001522, -0.00000865, 0.00000124, -0.00000677, 0.00000453, -0.00000645, 0.00001104, -0.00000622, 0.00000344, -0.00000276, -0.00000294, 0.00000153, 0.00000588, 0.00000040, -0.00000414, -0.00000489, 0.00000883, -0.00001681, 0.00001508, 0.00000591, -0.00001832, 0.00002044, -0.00004173, 0.00003074, -0.00002182, 0.00000742, 0.00001374, -0.00001364, 0.00000229, 0.00001775, -0.00000871, 0.00000800, },
		{ -0.00001192, 0.00000722, -0.00000770, -0.00000383, -0.00000161, 0.00000155, 0.00000902, -0.00000717, 0.00000393, -0.00000250, -0.00000504, 0.00000603, 0.00000445, -0.00000534, 0.00000655, -0.00001132, 0.00000027, -0.00000053, 0.00001060, -0.00000946, 0.00000187, -0.00000575, 0.00000452, -0.00000333, 0.00000611, -0.00000719, 0.00001136, -0.00001018, 0.00000878, -0.00001088, 0.00001631, -0.00001925, 0.00000242, -0.00000641, 0.00001425, -0.00001060, 0.00001605, -0.00001814, 0.00002265, -0.00002546, 0.00000647, -0.00001619, 0.00003293, -0.00003263, 0.00001255, -0.00001998, 0.00003218, -0.00003450, 0.00002646, -0.00004976, 0.00006441, -0.00007421, 0.00005729, -0.00012704, 0.00011930, -0.00032012, -0.00001748, 0.00100709, -0.00059331, 0.00011922, -0.00014091, 0.00007737, -0.00009667, 0.00005560, -0.00004606, 0.00004192, -0.00005942, 0.00003919, -0.00001936, 0.00001490, -0.00003617, 0.00003053, -0.00001380, 0.00001045, -0.00002739, 0.00001988, -0.00001772, 0.00001962, -0.00001741, 0.00001674, -0.00000259, 0.00000138, -0.00001927, 0.00001340, -0.00001040, 0.00001049, -0.00001323, 0.00001305, -0.00000701, 0.00000924, -0.00000719, 0.00000418, -0.00000337, -0.00000020, -0.00001008, 0.00001304, -0.00000192, -0.00000035, -0.00000739, 0.00000440, -0.00000664, 0.00000726, -0.00000019, -0.00000082, 0.00000061, -0.00000280, -0.00000294, 0.00000733, 0.00000062, 0.00000099, -0.00000990, 0.00000756, -0.00000527, -0.00000780, 0.00002541, -0.00002298, 0.00001485, -0.00001878, -0.00001010, 0.00001957, -0.00002742, 0.00002920, -0.00000828, -0.00000730, 0.00001427, 0.00000689, -0.00000748, 0.00001454, },
		{ -0.00000070, 0.00000095, -0.00000800, -0.00000199, -0.00000031, 0.00000596, 0.00000290, -0.00000708, 0.00000820, -0.00001165, 0.00000573, 0.00000335, 0.00000121, -0.00000251, 0.00000305, -0.00001365, 0.00000726, -0.00000055, 0.00000794, -0.00001360, 0.00000791, -0.00001020, 0.00001000, -0.00000706, 0.00000741, -0.00000646, 0.00001049, -0.00001157, 0.00000984, -0.00000890, 0.00001364, -0.00002697, 0.00001543, -0.00001136, 0.00001712, -0.00001265, 0.00001510, -0.00001621, 0.00002146, -0.00003577, 0.00001931, -0.00001489, 0.00002974, -0.00004562, 0.00003141, -0.00002800, 0.00003645, -0.00004471, 0.00003114, -0.00004337, 0.00006141, -0.00008907, 0.00006504, -0.00014271, 0.00014067, -0.00059556, 0.00093135, 0.00007804, -0.00035399, 0.00011405, -0.00012075, 0.00006507, -0.00008715, 0.00006341, -0.00004963, 0.00003362, -0.00005472, 0.00005413, -0.00003790, 0.00001677, -0.00003009, 0.00003767, -0.00002672, 0.00001114, -0.00002281, 0.00002053, -0.00001765, 0.00001998, -0.00002046, 0.00002710, -0.00001590, 0.00000125, -0.00001344, 0.00001336, -0.00001104, 0.00000899, -0.00001116, 0.00001498, -0.00001036, 0.00001220, -0.00001255, 0.00001011, -0.00000898, -0.00000055, -0.00000395, 0.00001386, -0.00000862, 0.00000091, -0.00000572, 0.00000357, -0.00000540, 0.00001054, -0.00000714, 0.00000471, -0.00000354, -0.00000289, 0.00000224, 0.00000549, -0.00000008, -0.00000437, -0.00000421, 0.00000922, -0.00001809, 0.00001845, -0.00000029, -0.00001160, 0.00001331, -0.00003743, 0.00002598, -0.00001724, 0.00000362, 0.00001783, -0.00001560, 0.00000324, 0.00002081, -0.00001189, 0.00001341, -0.00000521, },
		{ 0.00000755, -0.00000799, -0.00000408, -0.00000097, 0.00000167, 0.00000873, -0.00000783, 0.00000523, -0.00000417, -0.00000354, 0.00000579, 0.00000389, -0.00000505, 0.00000655, -0.00001215, 0.00000140, -0.00000051, 0.00001021, -0.00001036, 0.00000325, -0.00000659, 0.00000513, -0.00000346, 0.00000598, -0.00000698, 0.00001093, -0.00001003, 0.00000887, -0.00001073, 0.00001613, -0.00002052, 0.00000476, -0.00000720, 0.00001363, -0.00000984, 0.00001546, -0.00001769, 0.00002247, -0.00002689, 0.00000848, -0.00001619, 0.00003184, -0.00003339, 0.00001478, -0.00002030, 0.00003071, -0.00003365, 0.00002677, -0.00004887, 0.00006223, -0.00007251, 0.00005688, -0.00012793, 0.00012140, -0.00034175, 0.00005934, 0.00091774, -0.00055509, 0.00010932, -0.00013630, 0.00007535, -0.00009465, 0.00005564, -0.00004758, 0.00004231, -0.00005937, 0.00004071, -0.00002162, 0.00001477, -0.00003463, 0.00003144, -0.00001608, 0.00001064, -0.00002594, 0.00001919, -0.00001763, 0.00001997, -0.00001824, 0.00001828, -0.00000428, 0.00000107, -0.00001776, 0.00001285, -0.00001053, 0.00001039, -0.00001312, 0.00001363, -0.00000779, 0.00001006, -0.00000826, 0.00000492, -0.00000391, -0.00000042, -0.00000920, 0.00001338, -0.00000305, -0.00000023, -0.00000704, 0.00000458, -0.00000704, 0.00000835, -0.00000194, 0.00000009, 0.00000070, -0.00000339, -0.00000237, 0.00000768, -0.00000032, 0.00000118, -0.00001054, 0.00000972, -0.00000813, -0.00000380, 0.00002113, -0.00002109, 0.00001487, -0.00002201, -0.00000756, 0.00001592, -0.00002270, 0.00002427, -0.00000118, -0.00001449, 0.00002325, 0.00000041, -0.00000094, 0.00001004, -0.00001172, },
		{ -0.00000110, -0.00000699, -0.00000268, 0.00000078, 0.00000555, 0.00000216, -0.00000621, 0.00000763, -0.00001180, 0.00000684, 0.00000254, 0.00000094, -0.00000171, 0.00000167, -0.00001262, 0.00000722, 0.00000008, 0.00000600, -0.00001201, 0.00000717, -0.00000952, 0.00000946, -0.00000631, 0.00000635, -0.00000517, 0.00000861, -0.00000945, 0.00000777, -0.00000637, 0.00000997, -0.00002367, 0.00001422, -0.00001037, 0.00001503, -0.00000976, 0.00001165, -0.00001209, 0.00001622, -0.00003126, 0.00001664, -0.00001142, 0.00002377, -0.00003946, 0.00002739, -0.00002372, 0.00002976, -0.00003661, 0.00002346, -0.00003427, 0.00004874, -0.00007264, 0.00004450, -0.00011554, 0.00009502, -0.00053145, 0.00091119, 0.00003455, -0.00030881, 0.00008429, -0.00009823, 0.00004770, -0.00007117, 0.00004993, -0.00003968, 0.00002465, -0.00004542, 0.00004645, -0.00003334, 0.00001222, -0.00002322, 0.00003181, -0.00002393, 0.00000869, -0.00001875, 0.00001624, -0.00001349, 0.00001599, -0.00001675, 0.00002447, -0.00001525, 0.00000032, -0.00001085, 0.00001061, -0.00000887, 0.00000674, -0.00000850, 0.00001274, -0.00000850, 0.00001054, -0.00001155, 0.00000941, -0.00000879, -0.00000093, -0.00000223, 0.00001235, -0.00000847, 0.00000064, -0.00000474, 0.00000262, -0.00000424, 0.00000975, -0.00000773, 0.00000584, -0.00000438, -0.00000265, 0.00000280, 0.00000507, -0.00000051, -0.00000467, -0.00000318, 0.00000892, -0.00001843, 0.00002073, -0.00000583, -0.00000490, 0.00000560, -0.00003203, 0.00002006, -0.00001181, -0.00000083, 0.00002272, -0.00001844, 0.00000566, 0.00002221, -0.00001345, 0.00001731, -0.00001114, 0.00000273, },
		{ -0.00000775, -0.00000460, -0.00000013, 0.00000165, 0.00000832, -0.00000822, 0.00000632, -0.00000582, -0.00000182, 0.00000525, 0.00000342, -0.00000466, 0.00000629, -0.00001271, 0.00000246, -0.00000047, 0.00000959, -0.00001090, 0.00000441, -0.00000731, 0.00000565, -0.00000356, 0.00000576, -0.00000660, 0.00001023, -0.00000953, 0.00000858, -0.00001014, 0.00001534, -0.00002110, 0.00000671, -0.00000785, 0.00001280, -0.00000871, 0.00001435, -0.00001653, 0.00002140, -0.00002744, 0.00000992, -0.00001567, 0.00002987, -0.00003302, 0.00001618, -0.00002001, 0.00002842, -0.00003167, 0.00002579, -0.00004638, 0.00005789, -0.00006808, 0.00005306, -0.00012417, 0.00011665, -0.00035360, 0.00012881, 0.00082675, -0.00051046, 0.00009399, -0.00012755, 0.00006990, -0.00008963, 0.00005316, -0.00004702, 0.00004080, -0.00005741, 0.00004061, -0.00002288, 0.00001392, -0.00003208, 0.00003127, -0.00001769, 0.00001042, -0.00002391, 0.00001779, -0.00001677, 0.00001952, -0.00001832, 0.00001927, -0.00000579, 0.00000072, -0.00001597, 0.00001185, -0.00001022, 0.00000985, -0.00001252, 0.00001375, -0.00000819, 0.00001053, -0.00000909, 0.00000555, -0.00000444, -0.00000063, -0.00000814, 0.00001340, -0.00000405, -0.00000010, -0.00000663, 0.00000460, -0.00000722, 0.00000920, -0.00000364, 0.00000120, 0.00000049, -0.00000375, -0.00000185, 0.00000792, -0.00000120, 0.00000127, -0.00001088, 0.00001159, -0.00001093, 0.00000055, 0.00001579, -0.00001766, 0.00001320, -0.00002369, -0.00000605, 0.00001284, -0.00001850, 0.00002011, 0.00000513, -0.00002065, 0.00003161, -0.00000593, 0.00000591, 0.00000448, -0.00001041, 0.00000704, },
		{ -0.00000592, -0.00000335, 0.00000189, 0.00000500, 0.00000145, -0.00000515, 0.00000670, -0.00001153, 0.00000769, 0.00000169, 0.00000082, -0.00000103, 0.00000027, -0.00001139, 0.00000707, 0.00000061, 0.00000411, -0.00001027, 0.00000625, -0.00000870, 0.00000882, -0.00000558, 0.00000540, -0.00000405, 0.00000690, -0.00000747, 0.00000586, -0.00000405, 0.00000636, -0.00002012, 0.00001275, -0.00000942, 0.00001315, -0.00000718, 0.00000856, -0.00000828, 0.00001107, -0.00002660, 0.00001398, -0.00000830, 0.00001807, -0.00003319, 0.00002330, -0.00001976, 0.00002362, -0.00002912, 0.00001658, -0.00002620, 0.00003714, -0.00005729, 0.00002515, -0.00008945, 0.00004889, -0.00046062, 0.00087791, 0.00000194, -0.00026909, 0.00005757, -0.00007782, 0.00003188, -0.00005624, 0.00003718, -0.00003041, 0.00001629, -0.00003632, 0.00003862, -0.00002879, 0.00000808, -0.00001665, 0.00002589, -0.00002113, 0.00000658, -0.00001511, 0.00001228, -0.00000957, 0.00001208, -0.00001292, 0.00002158, -0.00001453, -0.00000031, -0.00000860, 0.00000808, -0.00000686, 0.00000465, -0.00000591, 0.00001046, -0.00000658, 0.00000874, -0.00001038, 0.00000858, -0.00000858, -0.00000113, -0.00000063, 0.00001073, -0.00000820, 0.00000041, -0.00000384, 0.00000171, -0.00000303, 0.00000868, -0.00000797, 0.00000675, -0.00000520, -0.00000228, 0.00000324, 0.00000462, -0.00000091, -0.00000498, -0.00000186, 0.00000795, -0.00001779, 0.00002176, -0.00001037, 0.00000130, -0.00000220, -0.00002593, 0.00001331, -0.00000584, -0.00000560, 0.00002815, -0.00002191, 0.00000934, 0.00002206, -0.00001339, 0.00001949, -0.00001602, 0.00000575, -0.00000305, },
		{ -0.00000535, 0.00000090, 0.00000150, 0.00000780, -0.00000832, 0.00000715, -0.00000735, -0.00000000, 0.00000446, 0.00000307, -0.00000419, 0.00000577, -0.00001298, 0.00000342, -0.00000043, 0.00000877, -0.00001109, 0.00000531, -0.00000786, 0.00000607, -0.00000361, 0.00000548, -0.00000609, 0.00000932, -0.00000873, 0.00000797, -0.00000918, 0.00001396, -0.00002098, 0.00000823, -0.00000833, 0.00001178, -0.00000731, 0.00001281, -0.00001477, 0.00001950, -0.00002713, 0.00001077, -0.00001468, 0.00002711, -0.00003159, 0.00001676, -0.00001919, 0.00002547, -0.00002879, 0.00002376, -0.00004258, 0.00005174, -0.00006130, 0.00004624, -0.00011620, 0.00010530, -0.00035519, 0.00018907, 0.00073675, -0.00046155, 0.00007454, -0.00011550, 0.00006162, -0.00008210, 0.00004848, -0.00004464, 0.00003762, -0.00005369, 0.00003895, -0.00002313, 0.00001243, -0.00002866, 0.00003011, -0.00001866, 0.00000985, -0.00002143, 0.00001577, -0.00001526, 0.00001836, -0.00001768, 0.00001967, -0.00000708, 0.00000038, -0.00001399, 0.00001047, -0.00000954, 0.00000895, -0.00001147, 0.00001344, -0.00000822, 0.00001064, -0.00000964, 0.00000604, -0.00000494, -0.00000080, -0.00000693, 0.00001313, -0.00000491, 0.00000004, -0.00000617, 0.00000448, -0.00000717, 0.00000977, -0.00000519, 0.00000243, 0.00000004, -0.00000387, -0.00000139, 0.00000806, -0.00000202, 0.00000125, -0.00001087, 0.00001304, -0.00001346, 0.00000493, 0.00000972, -0.00001295, 0.00000994, -0.00002375, -0.00000571, 0.00001046, -0.00001501, 0.00001697, 0.00001038, -0.00002542, 0.00003893, -0.00001167, 0.00001252, -0.00000172, -0.00000820, 0.00000578, -0.00000700, },
		{ -0.00000391, 0.00000294, 0.00000433, 0.00000081, -0.00000395, 0.00000544, -0.00001084, 0.00000824, 0.00000086, 0.00000084, -0.00000050, -0.00000112, -0.00000998, 0.00000680, 0.00000103, 0.00000232, -0.00000844, 0.00000518, -0.00000775, 0.00000809, -0.00000488, 0.00000457, -0.00000312, 0.00000542, -0.00000569, 0.00000419, -0.00000205, 0.00000294, -0.00001645, 0.00001111, -0.00000852, 0.00001152, -0.00000496, 0.00000592, -0.00000494, 0.00000620, -0.00002196, 0.00001146, -0.00000564, 0.00001284, -0.00002704, 0.00001928, -0.00001624, 0.00001818, -0.00002244, 0.00001074, -0.00001948, 0.00002706, -0.00004364, 0.00000785, -0.00006571, 0.00000423, -0.00038545, 0.00083316, -0.00002000, -0.00023543, 0.00003459, -0.00006014, 0.00001815, -0.00004290, 0.00002562, -0.00002221, 0.00000892, -0.00002780, 0.00003093, -0.00002442, 0.00000451, -0.00001058, 0.00002010, -0.00001844, 0.00000486, -0.00001196, 0.00000877, -0.00000604, 0.00000843, -0.00000914, 0.00001852, -0.00001375, -0.00000066, -0.00000674, 0.00000583, -0.00000511, 0.00000283, -0.00000349, 0.00000825, -0.00000470, 0.00000688, -0.00000908, 0.00000763, -0.00000834, -0.00000116, 0.00000080, 0.00000904, -0.00000783, 0.00000022, -0.00000305, 0.00000087, -0.00000182, 0.00000740, -0.00000786, 0.00000738, -0.00000593, -0.00000181, 0.00000356, 0.00000415, -0.00000131, -0.00000526, -0.00000031, 0.00000638, -0.00001619, 0.00002143, -0.00001360, 0.00000658, -0.00000963, -0.00001957, 0.00000612, 0.00000030, -0.00001036, 0.00003378, -0.00002565, 0.00001400, 0.00002057, -0.00001183, 0.00001984, -0.00001962, 0.00000810, -0.00000472, -0.00000495, },
		{ 0.00000209, 0.00000119, 0.00000719, -0.00000812, 0.00000764, -0.00000865, 0.00000183, 0.00000348, 0.00000285, -0.00000367, 0.00000503, -0.00001294, 0.00000424, -0.00000037, 0.00000776, -0.00001093, 0.00000592, -0.00000822, 0.00000635, -0.00000361, 0.00000516, -0.00000550, 0.00000826, -0.00000770, 0.00000711, -0.00000792, 0.00001209, -0.00002021, 0.00000928, -0.00000864, 0.00001064, -0.00000570, 0.00001097, -0.00001253, 0.00001689, -0.00002602, 0.00001109, -0.00001331, 0.00002372, -0.00002922, 0.00001658, -0.00001791, 0.00002205, -0.00002523, 0.00002095, -0.00003784, 0.00004423, -0.00005269, 0.00003698, -0.00010465, 0.00008801, -0.00034654, 0.00023872, 0.00065008, -0.00041045, 0.00005237, -0.00010111, 0.00005126, -0.00007264, 0.00004205, -0.00004080, 0.00003307, -0.00004847, 0.00003588, -0.00002245, 0.00001040, -0.00002452, 0.00002803, -0.00001903, 0.00000900, -0.00001863, 0.00001331, -0.00001321, 0.00001658, -0.00001637, 0.00001949, -0.00000812, 0.00000006, -0.00001193, 0.00000882, -0.00000857, 0.00000775, -0.00001004, 0.00001272, -0.00000790, 0.00001039, -0.00000990, 0.00000635, -0.00000539, -0.00000093, -0.00000561, 0.00001256, -0.00000561, 0.00000019, -0.00000567, 0.00000424, -0.00000691, 0.00001002, -0.00000652, 0.00000367, -0.00000060, -0.00000377, -0.00000101, 0.00000812, -0.00000277, 0.00000111, -0.00001046, 0.00001394, -0.00001551, 0.00000903, 0.00000334, -0.00000731, 0.00000530, -0.00002227, -0.00000655, 0.00000886, -0.00001234, 0.00001498, 0.00001438, -0.00002855, 0.00004487, -0.00001638, 0.00001839, -0.00000811, -0.00000536, 0.00000393, -0.00000583, -0.00000630, },
		{ 0.00000389, 0.00000357, 0.00000028, -0.00000268, 0.00000394, -0.00000978, 0.00000845, 0.00000011, 0.00000097, -0.00000015, -0.00000243, -0.00000845, 0.00000644, 0.00000132, 0.00000068, -0.00000659, 0.00000403, -0.00000671, 0.00000729, -0.00000419, 0.00000386, -0.00000240, 0.00000418, -0.00000416, 0.00000283, -0.00000044, -0.00000014, -0.00001279, 0.00000935, -0.00000770, 0.00001015, -0.00000316, 0.00000381, -0.00000220, 0.00000177, -0.00001748, 0.00000915, -0.00000355, 0.00000824, -0.00002120, 0.00001546, -0.00001322, 0.00001354, -0.00001671, 0.00000608, -0.00001431, 0.00001882, -0.00003217, -0.00000671, -0.00004539, -0.00003712, -0.00030846, 0.00077899, -0.00003199, -0.00020800, 0.00001573, -0.00004557, 0.00000690, -0.00003151, 0.00001560, -0.00001539, 0.00000284, -0.00002020, 0.00002367, -0.00002040, 0.00000163, -0.00000518, 0.00001458, -0.00001593, 0.00000358, -0.00000939, 0.00000582, -0.00000303, 0.00000518, -0.00000554, 0.00001537, -0.00001291, -0.00000073, -0.00000529, 0.00000391, -0.00000367, 0.00000135, -0.00000131, 0.00000618, -0.00000293, 0.00000503, -0.00000770, 0.00000659, -0.00000805, -0.00000102, 0.00000203, 0.00000731, -0.00000735, 0.00000007, -0.00000236, 0.00000014, -0.00000065, 0.00000596, -0.00000740, 0.00000769, -0.00000649, -0.00000131, 0.00000377, 0.00000367, -0.00000172, -0.00000543, 0.00000134, 0.00000432, -0.00001374, 0.00001978, -0.00001538, 0.00001059, -0.00001626, -0.00001339, -0.00000112, 0.00000622, -0.00001473, 0.00003925, -0.00002930, 0.00001926, 0.00001809, -0.00000904, 0.00001841, -0.00002179, 0.00000962, -0.00000596, -0.00000416, -0.00000429, },
		{ 0.00000074, 0.00000650, -0.00000762, 0.00000778, -0.00000965, 0.00000356, 0.00000237, 0.00000275, -0.00000313, 0.00000407, -0.00001257, 0.00000490, -0.00000031, 0.00000661, -0.00001044, 0.00000624, -0.00000835, 0.00000647, -0.00000356, 0.00000481, -0.00000487, 0.00000710, -0.00000651, 0.00000608, -0.00000645, 0.00000982, -0.00001882, 0.00000986, -0.00000877, 0.00000941, -0.00000401, 0.00000897, -0.00000999, 0.00001372, -0.00002421, 0.00001092, -0.00001168, 0.00001986, -0.00002605, 0.00001572, -0.00001628, 0.00001832, -0.00002124, 0.00001767, -0.00003254, 0.00003587, -0.00004286, 0.00002602, -0.00009041, 0.00006573, -0.00032814, 0.00027693, 0.00056873, -0.00035909, 0.00002887, -0.00008539, 0.00003960, -0.00006188, 0.00003440, -0.00003591, 0.00002755, -0.00004210, 0.00003162, -0.00002097, 0.00000799, -0.00001983, 0.00002519, -0.00001887, 0.00000798, -0.00001565, 0.00001055, -0.00001080, 0.00001431, -0.00001447, 0.00001875, -0.00000891, -0.00000019, -0.00000987, 0.00000700, -0.00000741, 0.00000634, -0.00000831, 0.00001165, -0.00000727, 0.00000980, -0.00000984, 0.00000646, -0.00000577, -0.00000099, -0.00000423, 0.00001173, -0.00000613, 0.00000032, -0.00000516, 0.00000390, -0.00000647, 0.00000995, -0.00000755, 0.00000483, -0.00000138, -0.00000346, -0.00000073, 0.00000809, -0.00000346, 0.00000087, -0.00000963, 0.00001419, -0.00001692, 0.00001255, -0.00000294, -0.00000116, -0.00000037, -0.00001945, -0.00000850, 0.00000797, -0.00001049, 0.00001419, 0.00001708, -0.00002989, 0.00004923, -0.00001975, 0.00002303, -0.00001422, -0.00000222, 0.00000166, -0.00000434, -0.00000739, 0.00000340, },
		{ 0.00000273, -0.00000012, -0.00000142, 0.00000227, -0.00000838, 0.00000831, -0.00000052, 0.00000119, 0.00000001, -0.00000361, -0.00000684, 0.00000598, 0.00000148, -0.00000075, -0.00000479, 0.00000285, -0.00000564, 0.00000641, -0.00000352, 0.00000326, -0.00000187, 0.00000322, -0.00000291, 0.00000179, 0.00000070, -0.00000278, -0.00000927, 0.00000756, -0.00000698, 0.00000904, -0.00000177, 0.00000228, -0.00000014, -0.00000207, -0.00001331, 0.00000714, -0.00000207, 0.00000440, -0.00001582, 0.00001194, -0.00001075, 0.00000974, -0.00001198, 0.00000267, -0.00001079, 0.00001262, -0.00002318, -0.00001806, -0.00002933, -0.00007354, -0.00023212, 0.00071767, -0.00003511, -0.00018661, 0.00000107, -0.00003427, -0.00000169, -0.00002229, 0.00000734, -0.00001012, -0.00000174, -0.00001378, 0.00001708, -0.00001686, -0.00000047, -0.00000058, 0.00000946, -0.00001365, 0.00000275, -0.00000742, 0.00000348, -0.00000061, 0.00000244, -0.00000225, 0.00001224, -0.00001202, -0.00000056, -0.00000425, 0.00000236, -0.00000258, 0.00000024, 0.00000055, 0.00000432, -0.00000135, 0.00000327, -0.00000630, 0.00000547, -0.00000772, -0.00000072, 0.00000304, 0.00000560, -0.00000679, -0.00000006, -0.00000178, -0.00000047, 0.00000040, 0.00000445, -0.00000663, 0.00000766, -0.00000684, -0.00000082, 0.00000390, 0.00000319, -0.00000215, -0.00000544, 0.00000298, 0.00000192, -0.00001059, 0.00001691, -0.00001564, 0.00001311, -0.00002173, -0.00000776, -0.00000806, 0.00001156, -0.00001837, 0.00004421, -0.00003246, 0.00002474, 0.00001502, -0.00000539, 0.00001541, -0.00002251, 0.00001020, -0.00000668, -0.00000364, -0.00000442, 0.00000468, },
		{ 0.00000577, -0.00000684, 0.00000754, -0.00001028, 0.00000509, 0.00000120, 0.00000277, -0.00000262, 0.00000293, -0.00001189, 0.00000538, -0.00000024, 0.00000535, -0.00000965, 0.00000627, -0.00000825, 0.00000640, -0.00000343, 0.00000445, -0.00000424, 0.00000592, -0.00000523, 0.00000496, -0.00000490, 0.00000729, -0.00001690, 0.00000999, -0.00000873, 0.00000815, -0.00000230, 0.00000694, -0.00000730, 0.00001015, -0.00002182, 0.00001036, -0.00000990, 0.00001571, -0.00002227, 0.00001429, -0.00001442, 0.00001449, -0.00001708, 0.00001422, -0.00002707, 0.00002717, -0.00003245, 0.00001411, -0.00007448, 0.00003972, -0.00030096, 0.00030341, 0.00049426, -0.00030920, 0.00000535, -0.00006931, 0.00002743, -0.00005051, 0.00002607, -0.00003043, 0.00002148, -0.00003494, 0.00002643, -0.00001886, 0.00000536, -0.00001481, 0.00002173, -0.00001826, 0.00000687, -0.00001265, 0.00000766, -0.00000817, 0.00001172, -0.00001212, 0.00001749, -0.00000944, -0.00000035, -0.00000789, 0.00000512, -0.00000615, 0.00000483, -0.00000636, 0.00001030, -0.00000638, 0.00000891, -0.00000949, 0.00000636, -0.00000606, -0.00000099, -0.00000282, 0.00001066, -0.00000647, 0.00000044, -0.00000463, 0.00000349, -0.00000587, 0.00000958, -0.00000823, 0.00000584, -0.00000220, -0.00000299, -0.00000054, 0.00000798, -0.00000408, 0.00000053, -0.00000839, 0.00001376, -0.00001755, 0.00001525, -0.00000873, 0.00000505, -0.00000670, -0.00001556, -0.00001135, 0.00000768, -0.00000938, 0.00001454, 0.00001852, -0.00002944, 0.00005190, -0.00002158, 0.00002607, -0.00001960, 0.00000090, -0.00000079, -0.00000268, -0.00000850, 0.00000477, 0.00000017, },
		{ -0.00000035, -0.00000023, 0.00000054, -0.00000674, 0.00000785, -0.00000100, 0.00000146, -0.00000004, -0.00000461, -0.00000519, 0.00000546, 0.00000150, -0.00000194, -0.00000311, 0.00000170, -0.00000456, 0.00000550, -0.00000286, 0.00000276, -0.00000153, 0.00000251, -0.00000193, 0.00000108, 0.00000136, -0.00000490, -0.00000600, 0.00000579, -0.00000636, 0.00000819, -0.00000078, 0.00000131, 0.00000119, -0.00000523, -0.00000953, 0.00000547, -0.00000122, 0.00000139, -0.00001104, 0.00000881, -0.00000882, 0.00000676, -0.00000823, 0.00000047, -0.00000889, 0.00000846, -0.00001679, -0.00002597, -0.00001809, -0.00010368, -0.00015877, 0.00065162, -0.00003076, -0.00017074, -0.00000954, -0.00002618, -0.00000762, -0.00001531, 0.00000089, -0.00000648, -0.00000471, -0.00000871, 0.00001132, -0.00001389, -0.00000174, 0.00000312, 0.00000486, -0.00001164, 0.00000234, -0.00000603, 0.00000176, 0.00000115, 0.00000030, 0.00000062, 0.00000920, -0.00001110, -0.00000019, -0.00000359, 0.00000117, -0.00000185, -0.00000045, 0.00000205, 0.00000273, -0.00000001, 0.00000167, -0.00000493, 0.00000431, -0.00000733, -0.00000029, 0.00000381, 0.00000393, -0.00000616, -0.00000016, -0.00000130, -0.00000094, 0.00000129, 0.00000294, -0.00000563, 0.00000731, -0.00000694, -0.00000038, 0.00000395, 0.00000269, -0.00000263, -0.00000521, 0.00000447, -0.00000064, -0.00000698, 0.00001303, -0.00001446, 0.00001403, -0.00002579, -0.00000298, -0.00001438, 0.00001601, -0.00002096, 0.00004834, -0.00003478, 0.00003002, 0.00001179, -0.00000135, 0.00001112, -0.00002186, 0.00000983, -0.00000683, -0.00000345, -0.00000426, 0.00000526, 0.00000183, },
		{ -0.00000583, 0.00000694, -0.00001051, 0.00000634, 0.00000005, 0.00000288, -0.00000215, 0.00000166, -0.00001090, 0.00000567, -0.00000018, 0.00000403, -0.00000862, 0.00000604, -0.00000793, 0.00000613, -0.00000321, 0.00000409, -0.00000363, 0.00000476, -0.00000394, 0.00000383, -0.00000335, 0.00000461, -0.00001458, 0.00000970, -0.00000854, 0.00000690, -0.00000067, 0.00000500, -0.00000463, 0.00000638, -0.00001899, 0.00000950, -0.00000810, 0.00001149, -0.00001807, 0.00001243, -0.00001242, 0.00001070, -0.00001295, 0.00001085, -0.00002179, 0.00001863, -0.00002208, 0.00000208, -0.00005792, 0.00001141, -0.00026636, 0.00031839, 0.00042772, -0.00026220, -0.00001706, -0.00005376, 0.00001549, -0.00003915, 0.00001758, -0.00002482, 0.00001530, -0.00002743, 0.00002064, -0.00001629, 0.00000267, -0.00000966, 0.00001783, -0.00001730, 0.00000576, -0.00000974, 0.00000481, -0.00000551, 0.00000896, -0.00000943, 0.00001580, -0.00000971, -0.00000042, -0.00000608, 0.00000327, -0.00000489, 0.00000331, -0.00000430, 0.00000876, -0.00000531, 0.00000778, -0.00000888, 0.00000605, -0.00000625, -0.00000092, -0.00000142, 0.00000938, -0.00000663, 0.00000054, -0.00000410, 0.00000302, -0.00000516, 0.00000891, -0.00000853, 0.00000660, -0.00000300, -0.00000240, -0.00000042, 0.00000780, -0.00000465, 0.00000013, -0.00000679, 0.00001263, -0.00001733, 0.00001694, -0.00001366, 0.00001086, -0.00001322, -0.00001095, -0.00001486, 0.00000776, -0.00000885, 0.00001588, 0.00001886, -0.00002731, 0.00005296, -0.00002183, 0.00002726, -0.00002388, 0.00000368, -0.00000320, -0.00000101, -0.00000955, 0.00000613, -0.00000051, 0.00000502, },
		{ 0.00000082, -0.00000116, -0.00000495, 0.00000710, -0.00000130, 0.00000178, -0.00000030, -0.00000538, -0.00000357, 0.00000488, 0.00000137, -0.00000286, -0.00000160, 0.00000065, -0.00000355, 0.00000458, -0.00000220, 0.00000235, -0.00000136, 0.00000204, -0.00000122, 0.00000070, 0.00000153, -0.00000643, -0.00000308, 0.00000411, -0.00000585, 0.00000756, -0.00000016, 0.00000088, 0.00000180, -0.00000765, -0.00000622, 0.00000414, -0.00000099, -0.00000075, -0.00000694, 0.00000610, -0.00000742, 0.00000455, -0.00000539, -0.00000063, -0.00000846, 0.00000624, -0.00001293, -0.00003043, -0.00001190, -0.00012659, -0.00009049, 0.00058323, -0.00002053, -0.00015966, -0.00001650, -0.00002102, -0.00001106, -0.00001048, -0.00000377, -0.00000440, -0.00000609, -0.00000506, 0.00000651, -0.00001155, -0.00000219, 0.00000588, 0.00000083, -0.00000990, 0.00000232, -0.00000518, 0.00000064, 0.00000227, -0.00000120, 0.00000299, 0.00000634, -0.00001015, 0.00000033, -0.00000327, 0.00000031, -0.00000145, -0.00000074, 0.00000316, 0.00000143, 0.00000105, 0.00000028, -0.00000366, 0.00000315, -0.00000690, 0.00000027, 0.00000433, 0.00000235, -0.00000547, -0.00000025, -0.00000091, -0.00000127, 0.00000199, 0.00000152, -0.00000447, 0.00000665, -0.00000680, -0.00000002, 0.00000396, 0.00000219, -0.00000316, -0.00000470, 0.00000569, -0.00000317, -0.00000315, 0.00000842, -0.00001201, 0.00001338, -0.00002833, 0.00000073, -0.00001985, 0.00001931, -0.00002226, 0.00005141, -0.00003592, 0.00003473, 0.00000884, 0.00000257, 0.00000597, -0.00002003, 0.00000859, -0.00000642, -0.00000359, -0.00000380, 0.00000560, 0.00000092, -0.00000041, },
		{ 0.00000601, -0.00001033, 0.00000725, -0.00000102, 0.00000307, -0.00000176, 0.00000030, -0.00000964, 0.00000576, -0.00000012, 0.00000270, -0.00000740, 0.00000558, -0.00000741, 0.00000567, -0.00000290, 0.00000373, -0.00000308, 0.00000366, -0.00000270, 0.00000277, -0.00000192, 0.00000193, -0.00001195, 0.00000906, -0.00000822, 0.00000569, 0.00000082, 0.00000325, -0.00000214, 0.00000257, -0.00001587, 0.00000843, -0.00000639, 0.00000737, -0.00001365, 0.00001026, -0.00001040, 0.00000711, -0.00000906, 0.00000781, -0.00001703, 0.00001069, -0.00001235, -0.00000932, -0.00004182, -0.00001767, -0.00022600, 0.00032261, 0.00036971, -0.00021917, -0.00003742, -0.00003948, 0.00000441, -0.00002837, 0.00000942, -0.00001947, 0.00000943, -0.00001996, 0.00001455, -0.00001348, 0.00000011, -0.00000460, 0.00001365, -0.00001608, 0.00000472, -0.00000704, 0.00000213, -0.00000296, 0.00000619, -0.00000657, 0.00001375, -0.00000974, -0.00000039, -0.00000447, 0.00000153, -0.00000372, 0.00000188, -0.00000222, 0.00000710, -0.00000412, 0.00000648, -0.00000803, 0.00000554, -0.00000632, -0.00000078, -0.00000008, 0.00000793, -0.00000660, 0.00000060, -0.00000358, 0.00000253, -0.00000438, 0.00000800, -0.00000845, 0.00000705, -0.00000370, -0.00000174, -0.00000036, 0.00000754, -0.00000518, -0.00000026, -0.00000489, 0.00001085, -0.00001627, 0.00001752, -0.00001747, 0.00001587, -0.00001952, -0.00000602, -0.00001871, 0.00000795, -0.00000868, 0.00001800, 0.00001833, -0.00002372, 0.00005257, -0.00002062, 0.00002648, -0.00002678, 0.00000586, -0.00000537, 0.00000052, -0.00001043, 0.00000740, -0.00000127, 0.00000427, -0.00000463, },
		{ -0.00000272, -0.00000311, 0.00000611, -0.00000142, 0.00000210, -0.00000078, -0.00000588, -0.00000203, 0.00000427, 0.00000112, -0.00000350, -0.00000030, -0.00000026, -0.00000265, 0.00000368, -0.00000156, 0.00000200, -0.00000132, 0.00000177, -0.00000074, 0.00000059, 0.00000124, -0.00000737, -0.00000057, 0.00000257, -0.00000544, 0.00000712, 0.00000016, 0.00000092, 0.00000171, -0.00000931, -0.00000341, 0.00000315, -0.00000132, -0.00000204, -0.00000356, 0.00000384, -0.00000648, 0.00000302, -0.00000334, -0.00000083, -0.00000927, 0.00000571, -0.00001134, -0.00003168, -0.00001066, -0.00014169, -0.00002902, 0.00051478, -0.00000610, -0.00015241, -0.00002040, -0.00001840, -0.00001231, -0.00000758, -0.00000683, -0.00000373, -0.00000599, -0.00000282, 0.00000269, -0.00000982, -0.00000185, 0.00000770, -0.00000258, -0.00000844, 0.00000261, -0.00000479, 0.00000006, 0.00000279, -0.00000206, 0.00000481, 0.00000371, -0.00000918, 0.00000093, -0.00000322, -0.00000025, -0.00000136, -0.00000064, 0.00000388, 0.00000044, 0.00000180, -0.00000085, -0.00000252, 0.00000201, -0.00000641, 0.00000092, 0.00000459, 0.00000089, -0.00000477, -0.00000031, -0.00000062, -0.00000147, 0.00000246, 0.00000027, -0.00000324, 0.00000575, -0.00000640, 0.00000024, 0.00000392, 0.00000167, -0.00000374, -0.00000387, 0.00000651, -0.00000549, 0.00000063, 0.00000341, -0.00000858, 0.00001127, -0.00002931, 0.00000326, -0.00002434, 0.00002130, -0.00002212, 0.00005325, -0.00003569, 0.00003861, 0.00000653, 0.00000587, 0.00000041, -0.00001730, 0.00000659, -0.00000551, -0.00000402, -0.00000305, 0.00000570, -0.00000000, -0.00000028, 0.00000167, },
		{ -0.00000976, 0.00000777, -0.00000193, 0.00000331, -0.00000147, -0.00000108, -0.00000815, 0.00000566, -0.00000009, 0.00000141, -0.00000606, 0.00000494, -0.00000671, 0.00000503, -0.00000249, 0.00000338, -0.00000261, 0.00000267, -0.00000155, 0.00000184, -0.00000069, -0.00000061, -0.00000916, 0.00000813, -0.00000780, 0.00000456, 0.00000212, 0.00000179, 0.00000004, -0.00000109, -0.00001261, 0.00000726, -0.00000487, 0.00000352, -0.00000921, 0.00000792, -0.00000846, 0.00000383, -0.00000554, 0.00000526, -0.00001303, 0.00000373, -0.00000375, -0.00001941, -0.00002717, -0.00004596, -0.00018174, 0.00031722, 0.00032034, -0.00018081, -0.00005505, -0.00002703, -0.00000530, -0.00001864, 0.00000199, -0.00001474, 0.00000420, -0.00001291, 0.00000851, -0.00001063, -0.00000216, 0.00000019, 0.00000938, -0.00001469, 0.00000381, -0.00000464, -0.00000026, -0.00000066, 0.00000358, -0.00000367, 0.00001144, -0.00000954, -0.00000026, -0.00000311, -0.00000003, -0.00000269, 0.00000062, -0.00000023, 0.00000542, -0.00000290, 0.00000507, -0.00000701, 0.00000484, -0.00000628, -0.00000057, 0.00000118, 0.00000635, -0.00000640, 0.00000062, -0.00000307, 0.00000204, -0.00000358, 0.00000691, -0.00000800, 0.00000717, -0.00000423, -0.00000108, -0.00000034, 0.00000719, -0.00000568, -0.00000061, -0.00000279, 0.00000851, -0.00001443, 0.00001695, -0.00001995, 0.00001975, -0.00002518, -0.00000115, -0.00002258, 0.00000798, -0.00000861, 0.00002064, 0.00001724, -0.00001896, 0.00005104, -0.00001819, 0.00002378, -0.00002813, 0.00000723, -0.00000711, 0.00000179, -0.00001103, 0.00000852, -0.00000207, 0.00000354, -0.00000329, 0.00000481, },
		{ -0.00000133, 0.00000496, -0.00000137, 0.00000241, -0.00000146, -0.00000610, -0.00000060, 0.00000365, 0.00000074, -0.00000385, 0.00000075, -0.00000099, -0.00000190, 0.00000284, -0.00000094, 0.00000169, -0.00000139, 0.00000168, -0.00000045, 0.00000071, 0.00000056, -0.00000772, 0.00000148, 0.00000121, -0.00000514, 0.00000683, 0.00000026, 0.00000133, 0.00000102, -0.00001024, -0.00000111, 0.00000246, -0.00000211, -0.00000256, -0.00000091, 0.00000202, -0.00000595, 0.00000205, -0.00000192, -0.00000035, -0.00001101, 0.00000654, -0.00001166, -0.00003016, -0.00001396, -0.00014884, 0.00002435, 0.00044831, 0.00001086, -0.00014800, -0.00002189, -0.00001779, -0.00001180, -0.00000630, -0.00000854, -0.00000420, -0.00000465, -0.00000188, -0.00000014, -0.00000867, -0.00000081, 0.00000861, -0.00000536, -0.00000721, 0.00000313, -0.00000477, -0.00000008, 0.00000278, -0.00000233, 0.00000606, 0.00000138, -0.00000819, 0.00000158, -0.00000339, -0.00000056, -0.00000151, -0.00000019, 0.00000423, -0.00000025, 0.00000224, -0.00000171, -0.00000156, 0.00000095, -0.00000588, 0.00000166, 0.00000459, -0.00000042, -0.00000405, -0.00000036, -0.00000040, -0.00000155, 0.00000270, -0.00000075, -0.00000202, 0.00000467, -0.00000579, 0.00000041, 0.00000385, 0.00000114, -0.00000435, -0.00000269, 0.00000685, -0.00000744, 0.00000410, -0.00000164, -0.00000450, 0.00000797, -0.00002885, 0.00000462, -0.00002780, 0.00002194, -0.00002049, 0.00005381, -0.00003396, 0.00004145, 0.00000514, 0.00000809, -0.00000508, -0.00001399, 0.00000405, -0.00000422, -0.00000469, -0.00000203, 0.00000554, -0.00000090, 0.00000003, 0.00000229, -0.00000408, },
		{ 0.00000790, -0.00000266, 0.00000359, -0.00000131, -0.00000244, -0.00000647, 0.00000538, -0.00000008, 0.00000019, -0.00000465, 0.00000418, -0.00000590, 0.00000423, -0.00000198, 0.00000302, -0.00000220, 0.00000181, -0.00000052, 0.00000108, 0.00000027, -0.00000292, -0.00000634, 0.00000700, -0.00000731, 0.00000353, 0.00000319, 0.00000067, 0.00000180, -0.00000446, -0.00000935, 0.00000607, -0.00000361, 0.00000010, -0.00000494, 0.00000556, -0.00000666, 0.00000095, -0.00000249, 0.00000331, -0.00000993, -0.00000201, 0.00000334, -0.00002766, -0.00001482, -0.00007201, -0.00013555, 0.00030369, 0.00027933, -0.00014749, -0.00006953, -0.00001680, -0.00001328, -0.00001029, -0.00000442, -0.00001089, -0.00000009, -0.00000662, 0.00000278, -0.00000792, -0.00000401, 0.00000451, 0.00000517, -0.00001321, 0.00000308, -0.00000260, -0.00000229, 0.00000128, 0.00000126, -0.00000089, 0.00000899, -0.00000915, -0.00000004, -0.00000201, -0.00000138, -0.00000186, -0.00000040, 0.00000161, 0.00000378, -0.00000173, 0.00000364, -0.00000588, 0.00000401, -0.00000612, -0.00000028, 0.00000231, 0.00000469, -0.00000604, 0.00000059, -0.00000257, 0.00000156, -0.00000280, 0.00000570, -0.00000722, 0.00000694, -0.00000454, -0.00000046, -0.00000033, 0.00000677, -0.00000616, -0.00000084, -0.00000060, 0.00000574, -0.00001192, 0.00001531, -0.00002102, 0.00002224, -0.00002985, 0.00000331, -0.00002615, 0.00000757, -0.00000840, 0.00002353, 0.00001591, -0.00001339, 0.00004871, -0.00001493, 0.00001935, -0.00002788, 0.00000766, -0.00000831, 0.00000267, -0.00001128, 0.00000941, -0.00000287, 0.00000285, -0.00000188, 0.00000342, -0.00000884, },
		{ 0.00000373, -0.00000117, 0.00000269, -0.00000232, -0.00000601, 0.00000067, 0.00000303, 0.00000026, -0.00000395, 0.00000155, -0.00000152, -0.00000134, 0.00000209, -0.00000034, 0.00000142, -0.00000153, 0.00000172, -0.00000029, 0.00000098, -0.00000043, -0.00000754, 0.00000306, 0.00000004, -0.00000492, 0.00000664, 0.00000021, 0.00000200, -0.00000019, -0.00001049, 0.00000069, 0.00000202, -0.00000325, -0.00000239, 0.00000104, 0.00000063, -0.00000575, 0.00000151, -0.00000095, 0.00000056, -0.00001335, 0.00000829, -0.00001337, -0.00002650, -0.00002112, -0.00014833, 0.00006875, 0.00038554, 0.00002882, -0.00014540, -0.00002171, -0.00001861, -0.00000999, -0.00000626, -0.00000922, -0.00000551, -0.00000232, -0.00000207, -0.00000208, -0.00000801, 0.00000080, 0.00000870, -0.00000754, -0.00000619, 0.00000379, -0.00000500, 0.00000011, 0.00000234, -0.00000206, 0.00000674, -0.00000063, -0.00000719, 0.00000220, -0.00000369, -0.00000071, -0.00000184, 0.00000054, 0.00000425, -0.00000067, 0.00000239, -0.00000227, -0.00000081, -0.00000002, -0.00000532, 0.00000245, 0.00000434, -0.00000155, -0.00000336, -0.00000038, -0.00000026, -0.00000152, 0.00000270, -0.00000149, -0.00000091, 0.00000350, -0.00000500, 0.00000049, 0.00000375, 0.00000058, -0.00000497, -0.00000120, 0.00000666, -0.00000887, 0.00000703, -0.00000639, -0.00000016, 0.00000378, -0.00002715, 0.00000489, -0.00003028, 0.00002124, -0.00001743, 0.00005315, -0.00003073, 0.00004319, 0.00000484, 0.00000887, -0.00001004, -0.00001044, 0.00000118, -0.00000270, -0.00000551, -0.00000082, 0.00000515, -0.00000176, 0.00000050, 0.00000265, -0.00000515, 0.00000029, },
		{ -0.00000316, 0.00000386, -0.00000130, -0.00000370, -0.00000467, 0.00000495, -0.00000010, -0.00000091, -0.00000324, 0.00000336, -0.00000502, 0.00000331, -0.00000139, 0.00000267, -0.00000188, 0.00000108, 0.00000035, 0.00000051, 0.00000091, -0.00000488, -0.00000360, 0.00000575, -0.00000679, 0.00000262, 0.00000403, -0.00000007, 0.00000305, -0.00000740, -0.00000621, 0.00000492, -0.00000269, -0.00000280, -0.00000100, 0.00000330, -0.00000508, -0.00000149, 0.00000003, 0.00000199, -0.00000782, -0.00000638, 0.00000870, -0.00003371, -0.00000543, -0.00009455, -0.00008936, 0.00028371, 0.00024601, -0.00011925, -0.00008069, -0.00000895, -0.00001936, -0.00000353, -0.00000962, -0.00000806, -0.00000326, -0.00000132, -0.00000237, -0.00000553, -0.00000531, 0.00000823, 0.00000116, -0.00001171, 0.00000254, -0.00000095, -0.00000389, 0.00000280, -0.00000068, 0.00000166, 0.00000648, -0.00000859, 0.00000024, -0.00000118, -0.00000248, -0.00000126, -0.00000114, 0.00000322, 0.00000225, -0.00000065, 0.00000226, -0.00000469, 0.00000306, -0.00000585, 0.00000007, 0.00000329, 0.00000299, -0.00000555, 0.00000053, -0.00000208, 0.00000110, -0.00000209, 0.00000445, -0.00000617, 0.00000637, -0.00000462, 0.00000009, -0.00000031, 0.00000625, -0.00000663, -0.00000090, 0.00000155, 0.00000268, -0.00000892, 0.00001275, -0.00002072, 0.00002322, -0.00003328, 0.00000705, -0.00002916, 0.00000647, -0.00000784, 0.00002643, 0.00001466, -0.00000736, 0.00004600, -0.00001128, 0.00001352, -0.00002613, 0.00000712, -0.00000888, 0.00000311, -0.00001112, 0.00001001, -0.00000365, 0.00000223, -0.00000047, 0.00000193, -0.00000765, 0.00000765, },
		{ -0.00000083, 0.00000289, -0.00000332, -0.00000562, 0.00000175, 0.00000243, -0.00000029, -0.00000380, 0.00000211, -0.00000184, -0.00000099, 0.00000147, 0.00000021, 0.00000117, -0.00000171, 0.00000184, -0.00000022, 0.00000134, -0.00000164, -0.00000688, 0.00000418, -0.00000091, -0.00000479, 0.00000650, 0.00000011, 0.00000281, -0.00000177, -0.00001015, 0.00000204, 0.00000176, -0.00000460, -0.00000167, 0.00000236, -0.00000039, -0.00000579, 0.00000128, -0.00000027, 0.00000163, -0.00001593, 0.00001055, -0.00001595, -0.00002140, -0.00003123, -0.00014082, 0.00010382, 0.00032782, 0.00004640, -0.00014366, -0.00002055, -0.00002029, -0.00000738, -0.00000708, -0.00000921, -0.00000729, 0.00000063, -0.00000315, -0.00000324, -0.00000773, 0.00000284, 0.00000806, -0.00000914, -0.00000533, 0.00000452, -0.00000538, 0.00000051, 0.00000160, -0.00000135, 0.00000689, -0.00000229, -0.00000619, 0.00000277, -0.00000407, -0.00000075, -0.00000228, 0.00000147, 0.00000399, -0.00000085, 0.00000228, -0.00000256, -0.00000028, -0.00000087, -0.00000473, 0.00000327, 0.00000387, -0.00000250, -0.00000270, -0.00000038, -0.00000017, -0.00000142, 0.00000249, -0.00000192, 0.00000002, 0.00000231, -0.00000408, 0.00000049, 0.00000362, -0.00000000, -0.00000557, 0.00000059, 0.00000591, -0.00000972, 0.00000925, -0.00001052, 0.00000405, -0.00000092, -0.00002447, 0.00000425, -0.00003189, 0.00001932, -0.00001308, 0.00005144, -0.00002612, 0.00004390, 0.00000563, 0.00000798, -0.00001407, -0.00000701, -0.00000176, -0.00000111, -0.00000636, 0.00000051, 0.00000456, -0.00000256, 0.00000112, 0.00000275, -0.00000590, 0.00000166, 0.00000248, },
		{ 0.00000412, -0.00000145, -0.00000481, -0.00000280, 0.00000440, -0.00000017, -0.00000186, -0.00000190, 0.00000255, -0.00000413, 0.00000231, -0.00000071, 0.00000231, -0.00000164, 0.00000049, 0.00000106, 0.00000014, 0.00000121, -0.00000643, -0.00000106, 0.00000445, -0.00000626, 0.00000183, 0.00000463, -0.00000044, 0.00000376, -0.00000983, -0.00000331, 0.00000387, -0.00000211, -0.00000508, 0.00000248, 0.00000122, -0.00000375, -0.00000348, 0.00000204, 0.00000129, -0.00000667, -0.00000936, 0.00001223, -0.00003738, 0.00000058, -0.00011254, -0.00004499, 0.00025909, 0.00021949, -0.00009583, -0.00008859, -0.00000347, -0.00002351, 0.00000157, -0.00001356, -0.00000629, -0.00000524, 0.00000282, -0.00000675, -0.00000356, -0.00000601, 0.00001124, -0.00000252, -0.00001025, 0.00000219, 0.00000031, -0.00000506, 0.00000384, -0.00000215, 0.00000387, 0.00000403, -0.00000790, 0.00000058, -0.00000060, -0.00000333, -0.00000091, -0.00000156, 0.00000455, 0.00000088, 0.00000027, 0.00000100, -0.00000351, 0.00000205, -0.00000548, 0.00000048, 0.00000410, 0.00000131, -0.00000495, 0.00000043, -0.00000161, 0.00000068, -0.00000149, 0.00000324, -0.00000493, 0.00000549, -0.00000445, 0.00000053, -0.00000027, 0.00000565, -0.00000710, -0.00000073, 0.00000354, -0.00000049, -0.00000563, 0.00000949, -0.00001918, 0.00002266, -0.00003532, 0.00000987, -0.00003141, 0.00000451, -0.00000674, 0.00002913, 0.00001378, -0.00000121, 0.00004330, -0.00000777, 0.00000671, -0.00002306, 0.00000566, -0.00000883, 0.00000308, -0.00001052, 0.00001029, -0.00000437, 0.00000169, 0.00000089, 0.00000041, -0.00000628, 0.00000705, -0.00000341, },
		{ 0.00000301, -0.00000442, -0.00000494, 0.00000263, 0.00000185, -0.00000090, -0.00000345, 0.00000244, -0.00000195, -0.00000087, 0.00000100, 0.00000072, 0.00000093, -0.00000190, 0.00000200, -0.00000019, 0.00000169, -0.00000296, -0.00000585, 0.00000486, -0.00000165, -0.00000471, 0.00000637, 0.00000001, 0.00000365, -0.00000359, -0.00000930, 0.00000299, 0.00000160, -0.00000605, -0.00000054, 0.00000313, -0.00000109, -0.00000598, 0.00000123, 0.00000027, 0.00000265, -0.00001842, 0.00001287, -0.00001881, -0.00001564, -0.00004325, -0.00012729, 0.00012962, 0.00027611, 0.00006250, -0.00014195, -0.00001905, -0.00002229, -0.00000443, -0.00000835, -0.00000884, -0.00000921, 0.00000387, -0.00000483, -0.00000381, -0.00000770, 0.00000515, 0.00000682, -0.00001022, -0.00000460, 0.00000522, -0.00000580, 0.00000100, 0.00000068, -0.00000033, 0.00000658, -0.00000360, -0.00000520, 0.00000323, -0.00000445, -0.00000075, -0.00000273, 0.00000254, 0.00000350, -0.00000085, 0.00000196, -0.00000259, 0.00000003, -0.00000158, -0.00000411, 0.00000409, 0.00000319, -0.00000324, -0.00000210, -0.00000035, -0.00000013, -0.00000127, 0.00000210, -0.00000202, 0.00000071, 0.00000118, -0.00000310, 0.00000043, 0.00000345, -0.00000063, -0.00000609, 0.00000261, 0.00000463, -0.00000994, 0.00001063, -0.00001378, 0.00000777, -0.00000575, -0.00002114, 0.00000292, -0.00003283, 0.00001637, -0.00000765, 0.00004892, -0.00002032, 0.00004375, 0.00000737, 0.00000530, -0.00001688, -0.00000399, -0.00000455, 0.00000039, -0.00000712, 0.00000188, 0.00000379, -0.00000329, 0.00000184, 0.00000262, -0.00000631, 0.00000272, 0.00000131, -0.00000038, },
		{ -0.00000178, -0.00000573, -0.00000094, 0.00000375, -0.00000029, -0.00000264, -0.00000066, 0.00000180, -0.00000330, 0.00000128, 0.00000002, 0.00000195, -0.00000146, 0.00000004, 0.00000162, -0.00000003, 0.00000118, -0.00000753, 0.00000121, 0.00000317, -0.00000575, 0.00000116, 0.00000502, -0.00000046, 0.00000392, -0.00001167, -0.00000071, 0.00000296, -0.00000186, -0.00000673, 0.00000542, -0.00000058, -0.00000271, -0.00000502, 0.00000357, 0.00000112, -0.00000637, -0.00001103, 0.00001401, -0.00003869, 0.00000307, -0.00012525, -0.00000403, 0.00023161, 0.00019866, -0.00007677, -0.00009348, -0.00000019, -0.00002583, 0.00000509, -0.00001627, -0.00000553, -0.00000605, 0.00000573, -0.00001026, -0.00000209, -0.00000605, 0.00001346, -0.00000578, -0.00000885, 0.00000202, 0.00000121, -0.00000580, 0.00000443, -0.00000313, 0.00000565, 0.00000171, -0.00000710, 0.00000093, -0.00000023, -0.00000395, -0.00000079, -0.00000167, 0.00000558, -0.00000028, 0.00000099, -0.00000011, -0.00000240, 0.00000102, -0.00000504, 0.00000096, 0.00000471, -0.00000031, -0.00000427, 0.00000029, -0.00000116, 0.00000029, -0.00000101, 0.00000213, -0.00000359, 0.00000436, -0.00000404, 0.00000085, -0.00000019, 0.00000496, -0.00000756, -0.00000030, 0.00000524, -0.00000359, -0.00000227, 0.00000580, -0.00001660, 0.00002067, -0.00003592, 0.00001163, -0.00003276, 0.00000159, -0.00000503, 0.00003152, 0.00001348, 0.00000479, 0.00004096, -0.00000490, -0.00000058, -0.00001897, 0.00000340, -0.00000821, 0.00000259, -0.00000948, 0.00001023, -0.00000501, 0.00000124, 0.00000213, -0.00000103, -0.00000482, 0.00000617, -0.00000342, 0.00000434, },
		{ -0.00000558, -0.00000399, 0.00000328, 0.00000131, -0.00000153, -0.00000295, 0.00000259, -0.00000188, -0.00000096, 0.00000068, 0.00000115, 0.00000070, -0.00000207, 0.00000216, -0.00000014, 0.00000198, -0.00000429, -0.00000455, 0.00000518, -0.00000220, -0.00000467, 0.00000622, 0.00000001, 0.00000438, -0.00000552, -0.00000808, 0.00000360, 0.00000148, -0.00000744, 0.00000083, 0.00000347, -0.00000153, -0.00000626, 0.00000128, 0.00000081, 0.00000341, -0.00002053, 0.00001486, -0.00002145, -0.00001000, -0.00005603, -0.00010899, 0.00014665, 0.00023093, 0.00007629, -0.00013961, -0.00001775, -0.00002414, -0.00000155, -0.00000975, -0.00000842, -0.00001094, 0.00000705, -0.00000684, -0.00000395, -0.00000776, 0.00000755, 0.00000513, -0.00001083, -0.00000395, 0.00000583, -0.00000618, 0.00000148, -0.00000028, 0.00000089, 0.00000588, -0.00000457, -0.00000424, 0.00000356, -0.00000478, -0.00000078, -0.00000314, 0.00000366, 0.00000286, -0.00000073, 0.00000149, -0.00000241, 0.00000013, -0.00000213, -0.00000348, 0.00000488, 0.00000235, -0.00000378, -0.00000158, -0.00000029, -0.00000013, -0.00000110, 0.00000158, -0.00000182, 0.00000113, 0.00000017, -0.00000211, 0.00000035, 0.00000323, -0.00000130, -0.00000650, 0.00000478, 0.00000288, -0.00000955, 0.00001113, -0.00001598, 0.00001070, -0.00001035, -0.00001749, 0.00000116, -0.00003331, 0.00001260, -0.00000141, 0.00004590, -0.00001363, 0.00004302, 0.00000980, 0.00000091, -0.00001826, -0.00000164, -0.00000698, 0.00000164, -0.00000769, 0.00000319, 0.00000290, -0.00000393, 0.00000262, 0.00000230, -0.00000638, 0.00000342, 0.00000026, 0.00000014, 0.00000302, },
		{ -0.00000639, 0.00000085, 0.00000303, -0.00000045, -0.00000324, 0.00000043, 0.00000115, -0.00000257, 0.00000027, 0.00000078, 0.00000158, -0.00000134, -0.00000030, 0.00000205, -0.00000004, 0.00000084, -0.00000816, 0.00000313, 0.00000198, -0.00000528, 0.00000060, 0.00000523, -0.00000018, 0.00000356, -0.00001290, 0.00000153, 0.00000219, -0.00000193, -0.00000776, 0.00000776, -0.00000206, -0.00000194, -0.00000614, 0.00000469, 0.00000135, -0.00000676, -0.00001157, 0.00001421, -0.00003783, 0.00000214, -0.00013230, 0.00003221, 0.00020297, 0.00018236, -0.00006146, -0.00009577, 0.00000118, -0.00002656, 0.00000716, -0.00001791, -0.00000561, -0.00000581, 0.00000743, -0.00001283, -0.00000116, -0.00000546, 0.00001487, -0.00000856, -0.00000756, 0.00000200, 0.00000180, -0.00000618, 0.00000458, -0.00000361, 0.00000696, -0.00000039, -0.00000623, 0.00000128, -0.00000005, -0.00000437, -0.00000087, -0.00000147, 0.00000629, -0.00000124, 0.00000150, -0.00000100, -0.00000141, 0.00000003, -0.00000454, 0.00000149, 0.00000510, -0.00000183, -0.00000355, 0.00000014, -0.00000073, -0.00000006, -0.00000068, 0.00000121, -0.00000224, 0.00000307, -0.00000343, 0.00000105, -0.00000008, 0.00000417, -0.00000800, 0.00000042, 0.00000656, -0.00000646, 0.00000091, 0.00000196, -0.00001329, 0.00001744, -0.00003516, 0.00001231, -0.00003317, -0.00000231, -0.00000270, 0.00003354, 0.00001390, 0.00001046, 0.00003922, -0.00000313, -0.00000782, -0.00001422, 0.00000057, -0.00000713, 0.00000173, -0.00000805, 0.00000983, -0.00000554, 0.00000089, 0.00000322, -0.00000231, -0.00000338, 0.00000508, -0.00000321, 0.00000451, -0.00000228, },
		{ -0.00000283, 0.00000371, 0.00000081, -0.00000214, -0.00000234, 0.00000259, -0.00000168, -0.00000126, 0.00000054, 0.00000152, 0.00000047, -0.00000222, 0.00000231, -0.00000005, 0.00000215, -0.00000553, -0.00000308, 0.00000517, -0.00000259, -0.00000465, 0.00000602, 0.00000013, 0.00000492, -0.00000740, -0.00000659, 0.00000393, 0.00000135, -0.00000867, 0.00000230, 0.00000349, -0.00000178, -0.00000655, 0.00000135, 0.00000143, 0.00000378, -0.00002203, 0.00001621, -0.00002341, -0.00000516, -0.00006846, -0.00008729, 0.00015574, 0.00019244, 0.00008723, -0.00013617, -0.00001703, -0.00002550, 0.00000095, -0.00001100, -0.00000819, -0.00001222, 0.00000986, -0.00000888, -0.00000386, -0.00000779, 0.00000988, 0.00000312, -0.00001106, -0.00000336, 0.00000630, -0.00000643, 0.00000185, -0.00000117, 0.00000218, 0.00000489, -0.00000521, -0.00000330, 0.00000375, -0.00000503, -0.00000087, -0.00000343, 0.00000475, 0.00000212, -0.00000053, 0.00000092, -0.00000206, 0.00000006, -0.00000254, -0.00000282, 0.00000561, 0.00000137, -0.00000413, -0.00000113, -0.00000019, -0.00000016, -0.00000096, 0.00000099, -0.00000136, 0.00000126, -0.00000067, -0.00000116, 0.00000027, 0.00000297, -0.00000203, -0.00000672, 0.00000698, 0.00000076, -0.00000864, 0.00001078, -0.00001703, 0.00001259, -0.00001440, -0.00001384, -0.00000077, -0.00003354, 0.00000825, 0.00000535, 0.00004271, -0.00000637, 0.00004205, 0.00001252, -0.00000497, -0.00001817, -0.00000010, -0.00000889, 0.00000252, -0.00000796, 0.00000438, 0.00000194, -0.00000447, 0.00000342, 0.00000184, -0.00000616, 0.00000376, -0.00000060, 0.00000071, 0.00000290, -0.00000673, },
		{ 0.00000252, 0.00000229, -0.00000066, -0.00000365, 0.00000136, 0.00000064, -0.00000201, -0.00000066, 0.00000154, 0.00000122, -0.00000127, -0.00000054, 0.00000237, 0.00000007, 0.00000024, -0.00000833, 0.00000466, 0.00000092, -0.00000487, 0.00000015, 0.00000532, 0.00000034, 0.00000273, -0.00001352, 0.00000338, 0.00000156, -0.00000224, -0.00000820, 0.00000948, -0.00000318, -0.00000145, -0.00000690, 0.00000549, 0.00000184, -0.00000764, -0.00001123, 0.00001314, -0.00003517, -0.00000187, -0.00013364, 0.00006278, 0.00017466, 0.00016942, -0.00004922, -0.00009592, 0.00000103, -0.00002599, 0.00000804, -0.00001867, -0.00000632, -0.00000472, 0.00000805, -0.00001448, -0.00000075, -0.00000428, 0.00001547, -0.00001081, -0.00000638, 0.00000208, 0.00000212, -0.00000625, 0.00000437, -0.00000363, 0.00000778, -0.00000222, -0.00000532, 0.00000159, -0.00000000, -0.00000462, -0.00000112, -0.00000100, 0.00000669, -0.00000198, 0.00000180, -0.00000167, -0.00000058, -0.00000090, -0.00000399, 0.00000207, 0.00000527, -0.00000321, -0.00000283, -0.00000001, -0.00000032, -0.00000039, -0.00000049, 0.00000050, -0.00000097, 0.00000167, -0.00000267, 0.00000115, 0.00000007, 0.00000328, -0.00000841, 0.00000142, 0.00000742, -0.00000896, 0.00000372, -0.00000171, -0.00000957, 0.00001324, -0.00003318, 0.00001198, -0.00003270, -0.00000714, 0.00000018, 0.00003524, 0.00001508, 0.00001573, 0.00003822, -0.00000284, -0.00001450, -0.00000919, -0.00000261, -0.00000574, 0.00000057, -0.00000629, 0.00000910, -0.00000596, 0.00000062, 0.00000412, -0.00000338, -0.00000206, 0.00000386, -0.00000278, 0.00000461, -0.00000295, -0.00000678, },
		{ 0.00000394, 0.00000034, -0.00000272, -0.00000167, 0.00000251, -0.00000140, -0.00000171, 0.00000056, 0.00000180, 0.00000024, -0.00000232, 0.00000241, 0.00000012, 0.00000214, -0.00000658, -0.00000156, 0.00000493, -0.00000284, -0.00000463, 0.00000576, 0.00000042, 0.00000519, -0.00000912, -0.00000496, 0.00000406, 0.00000114, -0.00000964, 0.00000371, 0.00000330, -0.00000191, -0.00000681, 0.00000140, 0.00000217, 0.00000369, -0.00002279, 0.00001670, -0.00002436, -0.00000171, -0.00007950, -0.00006366, 0.00015798, 0.00016043, 0.00009509, -0.00013138, -0.00001712, -0.00002614, 0.00000283, -0.00001190, -0.00000830, -0.00001285, 0.00001205, -0.00001070, -0.00000372, -0.00000764, 0.00001199, 0.00000094, -0.00001097, -0.00000279, 0.00000659, -0.00000651, 0.00000203, -0.00000188, 0.00000341, 0.00000370, -0.00000557, -0.00000242, 0.00000379, -0.00000516, -0.00000107, -0.00000356, 0.00000574, 0.00000135, -0.00000032, 0.00000033, -0.00000161, -0.00000016, -0.00000281, -0.00000215, 0.00000625, 0.00000030, -0.00000430, -0.00000077, -0.00000005, -0.00000021, -0.00000087, 0.00000039, -0.00000068, 0.00000110, -0.00000131, -0.00000030, 0.00000020, 0.00000265, -0.00000281, -0.00000671, 0.00000911, -0.00000161, -0.00000732, 0.00000969, -0.00001695, 0.00001332, -0.00001765, -0.00001045, -0.00000265, -0.00003373, 0.00000354, 0.00001234, 0.00003971, 0.00000112, 0.00004121, 0.00001504, -0.00001198, -0.00001668, 0.00000055, -0.00001019, 0.00000293, -0.00000787, 0.00000535, 0.00000094, -0.00000490, 0.00000419, 0.00000131, -0.00000570, 0.00000373, -0.00000124, 0.00000130, 0.00000262, -0.00000782, -0.00000150, },
		{ 0.00000156, -0.00000092, -0.00000389, 0.00000210, 0.00000028, -0.00000164, -0.00000146, 0.00000228, 0.00000086, -0.00000122, -0.00000070, 0.00000261, 0.00000025, -0.00000056, -0.00000809, 0.00000581, 0.00000001, -0.00000453, -0.00000022, 0.00000531, 0.00000100, 0.00000151, -0.00001358, 0.00000484, 0.00000104, -0.00000272, -0.00000814, 0.00001060, -0.00000394, -0.00000120, -0.00000735, 0.00000605, 0.00000242, -0.00000876, -0.00001030, 0.00001117, -0.00003117, -0.00000840, -0.00012955, 0.00008712, 0.00014794, 0.00015877, -0.00003933, -0.00009445, -0.00000022, -0.00002447, 0.00000800, -0.00001880, -0.00000740, -0.00000303, 0.00000774, -0.00001530, -0.00000080, -0.00000260, 0.00001531, -0.00001252, -0.00000531, 0.00000222, 0.00000226, -0.00000609, 0.00000388, -0.00000326, 0.00000812, -0.00000374, -0.00000440, 0.00000185, -0.00000005, -0.00000476, -0.00000147, -0.00000030, 0.00000680, -0.00000253, 0.00000190, -0.00000212, 0.00000007, -0.00000171, -0.00000343, 0.00000268, 0.00000522, -0.00000441, -0.00000213, -0.00000016, 0.00000006, -0.00000071, -0.00000043, 0.00000005, 0.00000016, 0.00000027, -0.00000181, 0.00000116, 0.00000025, 0.00000231, -0.00000876, 0.00000270, 0.00000778, -0.00001098, 0.00000599, -0.00000494, -0.00000579, 0.00000841, -0.00003023, 0.00001078, -0.00003145, -0.00001274, 0.00000343, 0.00003674, 0.00001695, 0.00002063, 0.00003795, -0.00000427, -0.00002019, -0.00000428, -0.00000586, -0.00000422, -0.00000074, -0.00000429, 0.00000810, -0.00000627, 0.00000043, 0.00000483, -0.00000417, -0.00000094, 0.00000260, -0.00000219, 0.00000463, -0.00000375, -0.00000687, 0.00000400, },
		{ -0.00000008, -0.00000324, -0.00000100, 0.00000238, -0.00000108, -0.00000228, 0.00000072, 0.00000201, 0.00000001, -0.00000237, 0.00000246, 0.00000037, 0.00000195, -0.00000739, -0.00000009, 0.00000453, -0.00000300, -0.00000460, 0.00000544, 0.00000089, 0.00000513, -0.00001058, -0.00000328, 0.00000404, 0.00000084, -0.00001027, 0.00000495, 0.00000301, -0.00000197, -0.00000701, 0.00000139, 0.00000303, 0.00000311, -0.00002274, 0.00001622, -0.00002408, -0.00000005, -0.00008826, -0.00003952, 0.00015463, 0.00013443, 0.00009985, -0.00012517, -0.00001811, -0.00002596, 0.00000400, -0.00001238, -0.00000882, -0.00001274, 0.00001343, -0.00001210, -0.00000368, -0.00000721, 0.00001374, -0.00000127, -0.00001062, -0.00000223, 0.00000669, -0.00000639, 0.00000200, -0.00000235, 0.00000449, 0.00000242, -0.00000568, -0.00000159, 0.00000369, -0.00000517, -0.00000139, -0.00000349, 0.00000658, 0.00000060, -0.00000015, -0.00000023, -0.00000112, -0.00000048, -0.00000296, -0.00000146, 0.00000677, -0.00000081, -0.00000431, -0.00000050, 0.00000012, -0.00000029, -0.00000085, -0.00000015, 0.00000013, 0.00000071, -0.00000174, 0.00000046, 0.00000018, 0.00000226, -0.00000365, -0.00000641, 0.00001105, -0.00000407, -0.00000575, 0.00000801, -0.00001583, 0.00001286, -0.00001994, -0.00000753, -0.00000429, -0.00003404, -0.00000135, 0.00001931, 0.00003719, 0.00000855, 0.00004087, 0.00001684, -0.00001966, -0.00001400, 0.00000034, -0.00001085, 0.00000286, -0.00000737, 0.00000606, -0.00000005, -0.00000523, 0.00000487, 0.00000078, -0.00000508, 0.00000338, -0.00000166, 0.00000189, 0.00000217, -0.00000878, -0.00000008, 0.00000400, },
		{ -0.00000121, -0.00000397, 0.00000268, 0.00000009, -0.00000149, -0.00000211, 0.00000295, 0.00000051, -0.00000120, -0.00000080, 0.00000281, 0.00000046, -0.00000147, -0.00000751, 0.00000657, -0.00000072, -0.00000426, -0.00000052, 0.00000528, 0.00000172, 0.00000001, -0.00001312, 0.00000592, 0.00000062, -0.00000329, -0.00000768, 0.00001119, -0.00000436, -0.00000118, -0.00000755, 0.00000647, 0.00000294, -0.00000989, -0.00000909, 0.00000869, -0.00002639, -0.00001674, -0.00012061, 0.00010503, 0.00012374, 0.00014946, -0.00003116, -0.00009187, -0.00000217, -0.00002235, 0.00000732, -0.00001857, -0.00000855, -0.00000106, 0.00000676, -0.00001542, -0.00000120, -0.00000053, 0.00001446, -0.00001370, -0.00000435, 0.00000238, 0.00000227, -0.00000580, 0.00000321, -0.00000257, 0.00000802, -0.00000493, -0.00000350, 0.00000203, -0.00000014, -0.00000483, -0.00000186, 0.00000056, 0.00000666, -0.00000289, 0.00000183, -0.00000234, 0.00000052, -0.00000240, -0.00000286, 0.00000332, 0.00000496, -0.00000541, -0.00000148, -0.00000028, 0.00000041, -0.00000103, -0.00000047, -0.00000015, 0.00000106, -0.00000106, -0.00000091, 0.00000111, 0.00000043, 0.00000124, -0.00000902, 0.00000419, 0.00000763, -0.00001247, 0.00000760, -0.00000751, -0.00000227, 0.00000329, -0.00002658, 0.00000889, -0.00002961, -0.00001892, 0.00000685, 0.00003825, 0.00001939, 0.00002532, 0.00003824, -0.00000750, -0.00002453, 0.00000016, -0.00000891, -0.00000276, -0.00000207, -0.00000216, 0.00000687, -0.00000645, 0.00000030, 0.00000534, -0.00000467, -0.00000009, 0.00000137, -0.00000146, 0.00000455, -0.00000468, -0.00000665, 0.00000527, 0.00000086, },
		{ -0.00000367, -0.00000037, 0.00000227, -0.00000080, -0.00000291, 0.00000100, 0.00000214, -0.00000021, -0.00000238, 0.00000246, 0.00000070, 0.00000155, -0.00000789, 0.00000125, 0.00000404, -0.00000308, -0.00000454, 0.00000505, 0.00000154, 0.00000471, -0.00001169, -0.00000164, 0.00000390, 0.00000044, -0.00001053, 0.00000592, 0.00000272, -0.00000202, -0.00000712, 0.00000135, 0.00000398, 0.00000210, -0.00002192, 0.00001477, -0.00002252, -0.00000040, -0.00009407, -0.00001619, 0.00014703, 0.00011377, 0.00010175, -0.00011763, -0.00001997, -0.00002496, 0.00000443, -0.00001242, -0.00000973, -0.00001187, 0.00001392, -0.00001292, -0.00000385, -0.00000643, 0.00001502, -0.00000339, -0.00001007, -0.00000168, 0.00000661, -0.00000608, 0.00000173, -0.00000254, 0.00000534, 0.00000113, -0.00000557, -0.00000083, 0.00000347, -0.00000506, -0.00000183, -0.00000319, 0.00000723, -0.00000009, -0.00000004, -0.00000071, -0.00000063, -0.00000085, -0.00000300, -0.00000074, 0.00000714, -0.00000192, -0.00000420, -0.00000030, 0.00000032, -0.00000039, -0.00000092, -0.00000057, 0.00000100, 0.00000013, -0.00000197, 0.00000109, 0.00000021, 0.00000181, -0.00000454, -0.00000580, 0.00001267, -0.00000648, -0.00000411, 0.00000596, -0.00001385, 0.00001127, -0.00002122, -0.00000522, -0.00000559, -0.00003456, -0.00000630, 0.00002608, 0.00003540, 0.00001567, 0.00004130, 0.00001740, -0.00002745, -0.00001044, -0.00000060, -0.00001092, 0.00000230, -0.00000647, 0.00000649, -0.00000099, -0.00000543, 0.00000543, 0.00000032, -0.00000438, 0.00000276, -0.00000184, 0.00000246, 0.00000154, -0.00000955, 0.00000138, 0.00000390, -0.00000048, },
		{ -0.00000392, 0.00000312, 0.00000003, -0.00000155, -0.00000256, 0.00000355, 0.00000019, -0.00000118, -0.00000087, 0.00000298, 0.00000064, -0.00000242, -0.00000665, 0.00000698, -0.00000128, -0.00000406, -0.00000077, 0.00000526, 0.00000241, -0.00000166, -0.00001223, 0.00000665, 0.00000025, -0.00000385, -0.00000693, 0.00001131, -0.00000448, -0.00000133, -0.00000754, 0.00000683, 0.00000327, -0.00001084, -0.00000789, 0.00000610, -0.00002139, -0.00002608, -0.00010765, 0.00011666, 0.00010268, 0.00014077, -0.00002414, -0.00008862, -0.00000443, -0.00001995, 0.00000628, -0.00001820, -0.00000953, 0.00000092, 0.00000535, -0.00001501, -0.00000185, 0.00000180, 0.00001304, -0.00001438, -0.00000349, 0.00000252, 0.00000223, -0.00000544, 0.00000245, -0.00000167, 0.00000754, -0.00000578, -0.00000263, 0.00000213, -0.00000023, -0.00000487, -0.00000225, 0.00000152, 0.00000631, -0.00000312, 0.00000162, -0.00000237, 0.00000079, -0.00000294, -0.00000229, 0.00000396, 0.00000450, -0.00000619, -0.00000092, -0.00000036, 0.00000072, -0.00000136, -0.00000059, -0.00000009, 0.00000172, -0.00000225, -0.00000003, 0.00000104, 0.00000061, 0.00000009, -0.00000913, 0.00000584, 0.00000701, -0.00001341, 0.00000849, -0.00000925, 0.00000068, -0.00000177, -0.00002254, 0.00000655, -0.00002738, -0.00002544, 0.00001017, 0.00004001, 0.00002220, 0.00003002, 0.00003880, -0.00001244, -0.00002730, 0.00000382, -0.00001154, -0.00000153, -0.00000329, -0.00000002, 0.00000548, -0.00000652, 0.00000022, 0.00000569, -0.00000488, 0.00000044, 0.00000025, -0.00000063, 0.00000436, -0.00000568, -0.00000615, 0.00000629, 0.00000023, -0.00000152, },
		{ 0.00000019, 0.00000221, -0.00000061, -0.00000355, 0.00000137, 0.00000220, -0.00000042, -0.00000234, 0.00000242, 0.00000109, 0.00000096, -0.00000808, 0.00000239, 0.00000351, -0.00000314, -0.00000445, 0.00000462, 0.00000232, 0.00000394, -0.00001241, -0.00000013, 0.00000369, -0.00000006, -0.00001042, 0.00000655, 0.00000249, -0.00000209, -0.00000714, 0.00000130, 0.00000494, 0.00000077, -0.00002043, 0.00001244, -0.00001977, -0.00000279, -0.00009650, 0.00000522, 0.00013650, 0.00009762, 0.00010115, -0.00010899, -0.00002252, -0.00002327, 0.00000422, -0.00001210, -0.00001093, -0.00001033, 0.00001353, -0.00001310, -0.00000431, -0.00000525, 0.00001576, -0.00000533, -0.00000937, -0.00000116, 0.00000637, -0.00000561, 0.00000125, -0.00000242, 0.00000588, -0.00000008, -0.00000530, -0.00000016, 0.00000315, -0.00000484, -0.00000237, -0.00000267, 0.00000765, -0.00000069, -0.00000002, -0.00000107, -0.00000021, -0.00000122, -0.00000298, 0.00000001, 0.00000734, -0.00000297, -0.00000401, -0.00000016, 0.00000055, -0.00000051, -0.00000109, -0.00000082, 0.00000185, -0.00000058, -0.00000203, 0.00000160, 0.00000028, 0.00000128, -0.00000546, -0.00000484, 0.00001388, -0.00000869, -0.00000256, 0.00000378, -0.00001126, 0.00000873, -0.00002151, -0.00000355, -0.00000649, -0.00003533, -0.00001123, 0.00003253, 0.00003451, 0.00002235, 0.00004270, 0.00001626, -0.00003479, -0.00000637, -0.00000207, -0.00001049, 0.00000133, -0.00000519, 0.00000661, -0.00000187, -0.00000553, 0.00000583, 0.00000001, -0.00000371, 0.00000195, -0.00000181, 0.00000300, 0.00000071, -0.00001008, 0.00000280, 0.00000369, -0.00000085, -0.00000400, },
		{ 0.00000343, 0.00000008, -0.00000182, -0.00000281, 0.00000403, -0.00000011, -0.00000117, -0.00000092, 0.00000315, 0.00000075, -0.00000333, -0.00000562, 0.00000709, -0.00000169, -0.00000392, -0.00000097, 0.00000528, 0.00000299, -0.00000340, -0.00001099, 0.00000707, -0.00000009, -0.00000432, -0.00000603, 0.00001106, -0.00000433, -0.00000163, -0.00000738, 0.00000717, 0.00000332, -0.00001143, -0.00000694, 0.00000376, -0.00001671, -0.00003557, -0.00009166, 0.00012248, 0.00008508, 0.00013216, -0.00001785, -0.00008506, -0.00000671, -0.00001753, 0.00000508, -0.00001788, -0.00001010, 0.00000263, 0.00000379, -0.00001425, -0.00000260, 0.00000422, 0.00001115, -0.00001460, -0.00000271, 0.00000261, 0.00000218, -0.00000510, 0.00000170, -0.00000067, 0.00000676, -0.00000631, -0.00000183, 0.00000214, -0.00000031, -0.00000493, -0.00000257, 0.00000252, 0.00000580, -0.00000325, 0.00000133, -0.00000226, 0.00000089, -0.00000334, -0.00000174, 0.00000458, 0.00000387, -0.00000675, -0.00000046, -0.00000038, 0.00000098, -0.00000173, -0.00000073, 0.00000019, 0.00000211, -0.00000325, 0.00000077, 0.00000097, 0.00000077, -0.00000115, -0.00000906, 0.00000756, 0.00000596, -0.00001382, 0.00000865, -0.00001006, 0.00000287, -0.00000647, -0.00001839, 0.00000399, -0.00002499, -0.00003204, 0.00001310, 0.00004229, 0.00002515, 0.00003502, 0.00003920, -0.00001886, -0.00002841, 0.00000652, -0.00001354, -0.00000068, -0.00000427, 0.00000204, 0.00000399, -0.00000648, 0.00000015, 0.00000590, -0.00000483, 0.00000065, -0.00000071, 0.00000025, 0.00000404, -0.00000670, -0.00000538, 0.00000705, -0.00000033, -0.00000184, -0.00000377, },
		{ 0.00000224, -0.00000056, -0.00000412, 0.00000178, 0.00000220, -0.00000062, -0.00000227, 0.00000233, 0.00000152, 0.00000022, -0.00000796, 0.00000328, 0.00000302, -0.00000320, -0.00000433, 0.00000416, 0.00000321, 0.00000285, -0.00001272, 0.00000121, 0.00000342, -0.00000062, -0.00000995, 0.00000681, 0.00000239, -0.00000220, -0.00000707, 0.00000126, 0.00000582, -0.00000077, -0.00001844, 0.00000944, -0.00001603, -0.00000705, -0.00009539, 0.00002384, 0.00012426, 0.00008513, 0.00009854, -0.00009957, -0.00002553, -0.00002106, 0.00000349, -0.00001156, -0.00001225, -0.00000828, 0.00001232, -0.00001263, -0.00000509, -0.00000367, 0.00001593, -0.00000702, -0.00000858, -0.00000066, 0.00000601, -0.00000501, 0.00000060, -0.00000202, 0.00000610, -0.00000115, -0.00000492, 0.00000042, 0.00000276, -0.00000455, -0.00000300, -0.00000195, 0.00000785, -0.00000119, -0.00000011, -0.00000129, 0.00000011, -0.00000156, -0.00000290, 0.00000080, 0.00000735, -0.00000392, -0.00000376, -0.00000006, 0.00000079, -0.00000066, -0.00000135, -0.00000086, 0.00000258, -0.00000132, -0.00000196, 0.00000199, 0.00000039, 0.00000069, -0.00000641, -0.00000354, 0.00001459, -0.00001057, -0.00000130, 0.00000171, -0.00000834, 0.00000546, -0.00002094, -0.00000250, -0.00000704, -0.00003629, -0.00001618, 0.00003861, 0.00003461, 0.00002855, 0.00004512, 0.00001307, -0.00004111, -0.00000223, -0.00000382, -0.00000973, 0.00000005, -0.00000360, 0.00000646, -0.00000266, -0.00000551, 0.00000605, -0.00000010, -0.00000314, 0.00000104, -0.00000159, 0.00000350, -0.00000030, -0.00001032, 0.00000410, 0.00000342, -0.00000120, -0.00000423, 0.00000066, },
		{ 0.00000019, -0.00000225, -0.00000286, 0.00000440, -0.00000037, -0.00000117, -0.00000095, 0.00000334, 0.00000076, -0.00000412, -0.00000450, 0.00000696, -0.00000196, -0.00000381, -0.00000116, 0.00000539, 0.00000338, -0.00000509, -0.00000951, 0.00000721, -0.00000043, -0.00000464, -0.00000509, 0.00001054, -0.00000399, -0.00000203, -0.00000709, 0.00000752, 0.00000303, -0.00001155, -0.00000643, 0.00000196, -0.00001284, -0.00004441, -0.00007372, 0.00012320, 0.00007095, 0.00012330, -0.00001200, -0.00008143, -0.00000879, -0.00001528, 0.00000389, -0.00001775, -0.00001010, 0.00000383, 0.00000231, -0.00001333, -0.00000331, 0.00000661, 0.00000892, -0.00001441, -0.00000202, 0.00000262, 0.00000216, -0.00000482, 0.00000106, 0.00000034, 0.00000576, -0.00000655, -0.00000110, 0.00000206, -0.00000035, -0.00000504, -0.00000277, 0.00000350, 0.00000517, -0.00000331, 0.00000100, -0.00000204, 0.00000085, -0.00000359, -0.00000119, 0.00000517, 0.00000310, -0.00000710, -0.00000011, -0.00000035, 0.00000120, -0.00000214, -0.00000084, 0.00000062, 0.00000223, -0.00000402, 0.00000145, 0.00000094, 0.00000088, -0.00000245, -0.00000877, 0.00000924, 0.00000460, -0.00001378, 0.00000816, -0.00000995, 0.00000414, -0.00001058, -0.00001441, 0.00000142, -0.00002263, -0.00003849, 0.00001539, 0.00004533, 0.00002806, 0.00004062, 0.00003893, -0.00002633, -0.00002795, 0.00000815, -0.00001482, -0.00000031, -0.00000491, 0.00000391, 0.00000247, -0.00000634, 0.00000008, 0.00000602, -0.00000457, 0.00000053, -0.00000146, 0.00000116, 0.00000357, -0.00000769, -0.00000440, 0.00000754, -0.00000079, -0.00000217, -0.00000356, 0.00000367, },
		{ -0.00000069, -0.00000457, 0.00000220, 0.00000216, -0.00000081, -0.00000217, 0.00000223, 0.00000197, -0.00000062, -0.00000754, 0.00000391, 0.00000259, -0.00000327, -0.00000416, 0.00000371, 0.00000413, 0.00000148, -0.00001260, 0.00000235, 0.00000311, -0.00000119, -0.00000918, 0.00000671, 0.00000245, -0.00000239, -0.00000692, 0.00000129, 0.00000654, -0.00000234, -0.00001615, 0.00000602, -0.00001163, -0.00001282, -0.00009083, 0.00003908, 0.00011138, 0.00007545, 0.00009443, -0.00008974, -0.00002874, -0.00001856, 0.00000245, -0.00001098, -0.00001349, -0.00000594, 0.00001047, -0.00001156, -0.00000615, -0.00000173, 0.00001551, -0.00000839, -0.00000771, -0.00000020, 0.00000556, -0.00000433, -0.00000017, -0.00000137, 0.00000600, -0.00000202, -0.00000445, 0.00000089, 0.00000233, -0.00000421, -0.00000367, -0.00000105, 0.00000781, -0.00000158, -0.00000029, -0.00000136, 0.00000031, -0.00000182, -0.00000279, 0.00000161, 0.00000716, -0.00000473, -0.00000350, 0.00000001, 0.00000103, -0.00000084, -0.00000168, -0.00000067, 0.00000313, -0.00000203, -0.00000183, 0.00000230, 0.00000053, 0.00000002, -0.00000734, -0.00000190, 0.00001476, -0.00001200, -0.00000045, -0.00000001, -0.00000539, 0.00000174, -0.00001969, -0.00000197, -0.00000731, -0.00003735, -0.00002121, 0.00004435, 0.00003568, 0.00003436, 0.00004846, 0.00000762, -0.00004595, 0.00000155, -0.00000558, -0.00000882, -0.00000140, -0.00000181, 0.00000607, -0.00000336, -0.00000539, 0.00000607, 0.00000001, -0.00000277, 0.00000010, -0.00000123, 0.00000393, -0.00000149, -0.00001024, 0.00000524, 0.00000312, -0.00000154, -0.00000435, 0.00000102, 0.00000237, },
		{ -0.00000280, -0.00000272, 0.00000464, -0.00000058, -0.00000116, -0.00000098, 0.00000356, 0.00000064, -0.00000475, -0.00000337, 0.00000666, -0.00000214, -0.00000374, -0.00000132, 0.00000558, 0.00000354, -0.00000663, -0.00000788, 0.00000713, -0.00000078, -0.00000475, -0.00000421, 0.00000985, -0.00000350, -0.00000251, -0.00000668, 0.00000788, 0.00000240, -0.00001115, -0.00000645, 0.00000089, -0.00001011, -0.00005187, -0.00005492, 0.00011973, 0.00006007, 0.00011410, -0.00000644, -0.00007786, -0.00001054, -0.00001331, 0.00000279, -0.00001785, -0.00000947, 0.00000438, 0.00000113, -0.00001242, -0.00000387, 0.00000880, 0.00000650, -0.00001388, -0.00000140, 0.00000256, 0.00000219, -0.00000465, 0.00000057, 0.00000125, 0.00000465, -0.00000654, -0.00000045, 0.00000189, -0.00000033, -0.00000521, -0.00000282, 0.00000439, 0.00000448, -0.00000334, 0.00000068, -0.00000177, 0.00000072, -0.00000372, -0.00000065, 0.00000568, 0.00000224, -0.00000724, 0.00000012, -0.00000025, 0.00000135, -0.00000260, -0.00000087, 0.00000116, 0.00000213, -0.00000457, 0.00000198, 0.00000097, 0.00000093, -0.00000381, -0.00000821, 0.00001079, 0.00000302, -0.00001340, 0.00000715, -0.00000899, 0.00000446, -0.00001391, -0.00001078, -0.00000100, -0.00002046, -0.00004459, 0.00001681, 0.00004935, 0.00003074, 0.00004710, 0.00003745, -0.00003433, -0.00002612, 0.00000872, -0.00001531, -0.00000048, -0.00000515, 0.00000551, 0.00000098, -0.00000612, -0.00000004, 0.00000612, -0.00000416, 0.00000013, -0.00000200, 0.00000207, 0.00000295, -0.00000859, -0.00000326, 0.00000779, -0.00000115, -0.00000248, -0.00000331, 0.00000386, 0.00000032, },
		{ -0.00000486, 0.00000259, 0.00000210, -0.00000098, -0.00000205, 0.00000213, 0.00000239, -0.00000153, -0.00000688, 0.00000427, 0.00000224, -0.00000337, -0.00000395, 0.00000329, 0.00000504, -0.00000009, -0.00001210, 0.00000326, 0.00000276, -0.00000172, -0.00000819, 0.00000629, 0.00000266, -0.00000264, -0.00000671, 0.00000143, 0.00000699, -0.00000378, -0.00001379, 0.00000246, -0.00000695, -0.00001963, -0.00008316, 0.00005063, 0.00009871, 0.00006783, 0.00008931, -0.00007985, -0.00003185, -0.00001601, 0.00000127, -0.00001053, -0.00001442, -0.00000356, 0.00000817, -0.00001001, -0.00000744, 0.00000050, 0.00001456, -0.00000942, -0.00000681, 0.00000020, 0.00000506, -0.00000364, -0.00000098, -0.00000056, 0.00000559, -0.00000267, -0.00000394, 0.00000126, 0.00000190, -0.00000386, -0.00000433, -0.00000001, 0.00000756, -0.00000187, -0.00000055, -0.00000129, 0.00000037, -0.00000197, -0.00000268, 0.00000246, 0.00000676, -0.00000536, -0.00000327, 0.00000009, 0.00000127, -0.00000107, -0.00000206, -0.00000024, 0.00000345, -0.00000262, -0.00000169, 0.00000256, 0.00000068, -0.00000070, -0.00000823, 0.00000003, 0.00001435, -0.00001292, -0.00000012, -0.00000119, -0.00000271, -0.00000211, -0.00001799, -0.00000180, -0.00000746, -0.00003836, -0.00002647, 0.00004985, 0.00003764, 0.00003999, 0.00005243, -0.00000012, -0.00004893, 0.00000460, -0.00000706, -0.00000796, -0.00000286, 0.00000007, 0.00000549, -0.00000398, -0.00000517, 0.00000592, 0.00000037, -0.00000263, -0.00000078, -0.00000076, 0.00000430, -0.00000283, -0.00000983, 0.00000615, 0.00000284, -0.00000187, -0.00000438, 0.00000128, 0.00000260, -0.00000101, },
		{ -0.00000242, 0.00000476, -0.00000076, -0.00000115, -0.00000101, 0.00000379, 0.00000040, -0.00000516, -0.00000232, 0.00000625, -0.00000225, -0.00000367, -0.00000145, 0.00000586, 0.00000344, -0.00000796, -0.00000617, 0.00000687, -0.00000115, -0.00000463, -0.00000349, 0.00000909, -0.00000294, -0.00000302, -0.00000617, 0.00000823, 0.00000146, -0.00001026, -0.00000704, 0.00000062, -0.00000877, -0.00005739, -0.00003632, 0.00011303, 0.00005203, 0.00010459, -0.00000116, -0.00007441, -0.00001195, -0.00001165, 0.00000179, -0.00001817, -0.00000819, 0.00000418, 0.00000041, -0.00001167, -0.00000415, 0.00001068, 0.00000401, -0.00001306, -0.00000085, 0.00000241, 0.00000227, -0.00000459, 0.00000028, 0.00000200, 0.00000350, -0.00000631, 0.00000011, 0.00000166, -0.00000026, -0.00000545, -0.00000270, 0.00000516, 0.00000376, -0.00000336, 0.00000040, -0.00000150, 0.00000054, -0.00000376, -0.00000010, 0.00000608, 0.00000132, -0.00000722, 0.00000024, -0.00000008, 0.00000143, -0.00000312, -0.00000078, 0.00000172, 0.00000184, -0.00000488, 0.00000235, 0.00000106, 0.00000092, -0.00000522, -0.00000736, 0.00001209, 0.00000137, -0.00001281, 0.00000578, -0.00000733, 0.00000387, -0.00001639, -0.00000765, -0.00000314, -0.00001859, -0.00005022, 0.00001721, 0.00005447, 0.00003312, 0.00005463, 0.00003421, -0.00004227, -0.00002325, 0.00000834, -0.00001506, -0.00000117, -0.00000497, 0.00000679, -0.00000044, -0.00000582, -0.00000022, 0.00000623, -0.00000369, -0.00000049, -0.00000231, 0.00000295, 0.00000214, -0.00000934, -0.00000202, 0.00000782, -0.00000142, -0.00000276, -0.00000306, 0.00000406, 0.00000040, -0.00000342, },
		{ 0.00000290, 0.00000204, -0.00000115, -0.00000191, 0.00000204, 0.00000275, -0.00000242, -0.00000603, 0.00000438, 0.00000199, -0.00000350, -0.00000371, 0.00000293, 0.00000587, -0.00000178, -0.00001126, 0.00000394, 0.00000236, -0.00000216, -0.00000707, 0.00000559, 0.00000302, -0.00000296, -0.00000645, 0.00000169, 0.00000710, -0.00000495, -0.00001157, -0.00000093, -0.00000239, -0.00002688, -0.00007295, 0.00005847, 0.00008685, 0.00006162, 0.00008362, -0.00007024, -0.00003462, -0.00001361, 0.00000016, -0.00001037, -0.00001485, -0.00000141, 0.00000568, -0.00000814, -0.00000885, 0.00000291, 0.00001312, -0.00001009, -0.00000589, 0.00000051, 0.00000457, -0.00000299, -0.00000176, 0.00000036, 0.00000493, -0.00000309, -0.00000343, 0.00000153, 0.00000148, -0.00000354, -0.00000494, 0.00000111, 0.00000713, -0.00000207, -0.00000086, -0.00000110, 0.00000029, -0.00000202, -0.00000258, 0.00000332, 0.00000616, -0.00000579, -0.00000309, 0.00000019, 0.00000147, -0.00000134, -0.00000246, 0.00000040, 0.00000350, -0.00000305, -0.00000160, 0.00000280, 0.00000081, -0.00000148, -0.00000903, 0.00000219, 0.00001339, -0.00001330, -0.00000036, -0.00000170, -0.00000055, -0.00000580, -0.00001609, -0.00000182, -0.00000765, -0.00003913, -0.00003210, 0.00005522, 0.00004036, 0.00004572, 0.00005660, -0.00000997, -0.00004983, 0.00000660, -0.00000801, -0.00000733, -0.00000419, 0.00000192, 0.00000479, -0.00000452, -0.00000488, 0.00000560, 0.00000096, -0.00000278, -0.00000153, -0.00000023, 0.00000457, -0.00000428, -0.00000909, 0.00000679, 0.00000261, -0.00000220, -0.00000432, 0.00000145, 0.00000293, -0.00000154, -0.00000496, },
		{ 0.00000477, -0.00000088, -0.00000114, -0.00000102, 0.00000404, 0.00000004, -0.00000535, -0.00000140, 0.00000580, -0.00000232, -0.00000360, -0.00000157, 0.00000621, 0.00000305, -0.00000899, -0.00000448, 0.00000646, -0.00000152, -0.00000426, -0.00000297, 0.00000835, -0.00000234, -0.00000355, -0.00000555, 0.00000851, 0.00000029, -0.00000894, -0.00000813, 0.00000114, -0.00000885, -0.00006057, -0.00001883, 0.00010412, 0.00004632, 0.00009496, 0.00000375, -0.00007100, -0.00001306, -0.00001025, 0.00000083, -0.00001860, -0.00000637, 0.00000324, 0.00000022, -0.00001118, -0.00000409, 0.00001215, 0.00000155, -0.00001201, -0.00000037, 0.00000220, 0.00000239, -0.00000463, 0.00000021, 0.00000253, 0.00000240, -0.00000592, 0.00000057, 0.00000137, -0.00000015, -0.00000575, -0.00000239, 0.00000577, 0.00000304, -0.00000340, 0.00000020, -0.00000127, 0.00000035, -0.00000372, 0.00000046, 0.00000636, 0.00000040, -0.00000705, 0.00000027, 0.00000015, 0.00000144, -0.00000368, -0.00000052, 0.00000224, 0.00000145, -0.00000501, 0.00000256, 0.00000122, 0.00000082, -0.00000665, -0.00000618, 0.00001304, -0.00000023, -0.00001215, 0.00000424, -0.00000518, 0.00000250, -0.00001802, -0.00000507, -0.00000496, -0.00001706, -0.00005528, 0.00001650, 0.00006074, 0.00003524, 0.00006328, 0.00002878, -0.00004949, -0.00001977, 0.00000723, -0.00001418, -0.00000233, -0.00000439, 0.00000773, -0.00000174, -0.00000545, -0.00000047, 0.00000641, -0.00000323, -0.00000127, -0.00000242, 0.00000379, 0.00000116, -0.00000987, -0.00000076, 0.00000767, -0.00000159, -0.00000301, -0.00000284, 0.00000429, 0.00000038, -0.00000405, -0.00000199, },
		{ 0.00000198, -0.00000131, -0.00000177, 0.00000197, 0.00000302, -0.00000325, -0.00000507, 0.00000429, 0.00000182, -0.00000367, -0.00000344, 0.00000265, 0.00000654, -0.00000351, -0.00001013, 0.00000442, 0.00000192, -0.00000246, -0.00000592, 0.00000470, 0.00000349, -0.00000333, -0.00000614, 0.00000209, 0.00000682, -0.00000573, -0.00000969, -0.00000387, 0.00000162, -0.00003398, -0.00006089, 0.00006286, 0.00007619, 0.00005635, 0.00007773, -0.00006115, -0.00003686, -0.00001155, -0.00000077, -0.00001064, -0.00001460, 0.00000028, 0.00000325, -0.00000614, -0.00001025, 0.00000539, 0.00001129, -0.00001043, -0.00000497, 0.00000074, 0.00000411, -0.00000243, -0.00000243, 0.00000128, 0.00000409, -0.00000329, -0.00000292, 0.00000169, 0.00000111, -0.00000327, -0.00000545, 0.00000226, 0.00000654, -0.00000222, -0.00000119, -0.00000082, 0.00000010, -0.00000194, -0.00000249, 0.00000419, 0.00000538, -0.00000601, -0.00000299, 0.00000033, 0.00000163, -0.00000166, -0.00000283, 0.00000121, 0.00000329, -0.00000325, -0.00000160, 0.00000305, 0.00000090, -0.00000230, -0.00000970, 0.00000451, 0.00001193, -0.00001315, -0.00000117, -0.00000147, 0.00000090, -0.00000907, -0.00001423, -0.00000186, -0.00000803, -0.00003950, -0.00003828, 0.00006062, 0.00004368, 0.00005188, 0.00006041, -0.00002158, -0.00004858, 0.00000733, -0.00000824, -0.00000712, -0.00000523, 0.00000361, 0.00000404, -0.00000500, -0.00000452, 0.00000515, 0.00000176, -0.00000321, -0.00000209, 0.00000034, 0.00000475, -0.00000580, -0.00000804, 0.00000715, 0.00000246, -0.00000252, -0.00000420, 0.00000156, 0.00000330, -0.00000225, -0.00000483, 0.00000312, },
		{ -0.00000097, -0.00000114, -0.00000103, 0.00000429, -0.00000042, -0.00000530, -0.00000066, 0.00000534, -0.00000240, -0.00000351, -0.00000165, 0.00000661, 0.00000240, -0.00000971, -0.00000285, 0.00000592, -0.00000187, -0.00000369, -0.00000270, 0.00000768, -0.00000177, -0.00000407, -0.00000481, 0.00000867, -0.00000101, -0.00000732, -0.00000958, 0.00000231, -0.00001029, -0.00006125, -0.00000317, 0.00009393, 0.00004233, 0.00008548, 0.00000816, -0.00006753, -0.00001398, -0.00000901, -0.00000021, -0.00001900, -0.00000415, 0.00000165, 0.00000059, -0.00001100, -0.00000366, 0.00001314, -0.00000076, -0.00001080, 0.00000005, 0.00000195, 0.00000251, -0.00000474, 0.00000033, 0.00000280, 0.00000140, -0.00000542, 0.00000093, 0.00000106, -0.00000002, -0.00000609, -0.00000190, 0.00000620, 0.00000234, -0.00000347, 0.00000010, -0.00000112, 0.00000019, -0.00000364, 0.00000105, 0.00000649, -0.00000047, -0.00000678, 0.00000022, 0.00000043, 0.00000136, -0.00000426, -0.00000007, 0.00000263, 0.00000101, -0.00000499, 0.00000264, 0.00000143, 0.00000064, -0.00000808, -0.00000467, 0.00001357, -0.00000167, -0.00001155, 0.00000274, -0.00000280, 0.00000054, -0.00001890, -0.00000302, -0.00000649, -0.00001585, -0.00005981, 0.00001470, 0.00006812, 0.00003726, 0.00007293, 0.00002080, -0.00005540, -0.00001614, 0.00000565, -0.00001283, -0.00000385, -0.00000347, 0.00000833, -0.00000291, -0.00000503, -0.00000082, 0.00000670, -0.00000287, -0.00000212, -0.00000236, 0.00000457, -0.00000001, -0.00001013, 0.00000044, 0.00000739, -0.00000168, -0.00000322, -0.00000265, 0.00000458, 0.00000022, -0.00000461, -0.00000148, 0.00000468, },
		{ -0.00000147, -0.00000162, 0.00000194, 0.00000318, -0.00000396, -0.00000408, 0.00000404, 0.00000171, -0.00000385, -0.00000315, 0.00000248, 0.00000700, -0.00000518, -0.00000880, 0.00000469, 0.00000143, -0.00000256, -0.00000484, 0.00000370, 0.00000404, -0.00000372, -0.00000580, 0.00000264, 0.00000613, -0.00000605, -0.00000827, -0.00000616, 0.00000474, -0.00004030, -0.00004777, 0.00006423, 0.00006687, 0.00005170, 0.00007186, -0.00005275, -0.00003843, -0.00000991, -0.00000141, -0.00001139, -0.00001356, 0.00000131, 0.00000112, -0.00000419, -0.00001152, 0.00000781, 0.00000917, -0.00001045, -0.00000405, 0.00000088, 0.00000370, -0.00000201, -0.00000292, 0.00000212, 0.00000313, -0.00000329, -0.00000246, 0.00000175, 0.00000079, -0.00000308, -0.00000582, 0.00000339, 0.00000583, -0.00000232, -0.00000151, -0.00000050, -0.00000018, -0.00000177, -0.00000241, 0.00000503, 0.00000444, -0.00000601, -0.00000298, 0.00000053, 0.00000173, -0.00000203, -0.00000312, 0.00000215, 0.00000281, -0.00000321, -0.00000173, 0.00000335, 0.00000093, -0.00000316, -0.00001019, 0.00000687, 0.00001003, -0.00001253, -0.00000248, -0.00000050, 0.00000153, -0.00001169, -0.00001261, -0.00000176, -0.00000875, -0.00003930, -0.00004514, 0.00006615, 0.00004748, 0.00005883, 0.00006319, -0.00003440, -0.00004531, 0.00000669, -0.00000763, -0.00000741, -0.00000587, 0.00000504, 0.00000330, -0.00000542, -0.00000413, 0.00000461, 0.00000269, -0.00000392, -0.00000244, 0.00000090, 0.00000479, -0.00000733, -0.00000671, 0.00000722, 0.00000241, -0.00000284, -0.00000404, 0.00000163, 0.00000369, -0.00000312, -0.00000450, 0.00000321, 0.00000194, },
		},
		{ { -0.00000025, 0.00000013, -0.00000118, 0.00000196, -0.00000488, 0.00000058, -0.00000426, 0.00000122, 0.00000230, 0.00000191, 0.00000349, -0.00000107, -0.00000562, -0.00000454, 0.00001215, -0.00000217, 0.00000114, 0.00000211, -0.00000974, 0.00000227, -0.00001354, 0.00000301, -0.00000185, -0.00000653, -0.00000440, 0.00000073, -0.00001576, -0.00000035, -0.00000338, -0.00000447, -0.00001041, -0.00001359, -0.00005288, 0.00000800, 0.00008903, 0.00003839, 0.00008321, 0.00000964, -0.00005913, -0.00000959, -0.00000878, -0.00000673, -0.00000028, -0.00000459, -0.00000191, -0.00000139, 0.00000389, 0.00000105, -0.00000160, -0.00000438, 0.00000063, -0.00000322, 0.00000762, -0.00000289, -0.00000236, -0.00000226, -0.00000136, -0.00000056, 0.00000707, 0.00000084, 0.00000391, -0.00000229, -0.00000995, 0.00000147, 0.00000330, 0.00000147, -0.00000995, -0.00000229, 0.00000391, 0.00000084, 0.00000707, -0.00000056, -0.00000136, -0.00000226, -0.00000236, -0.00000289, 0.00000762, -0.00000322, 0.00000063, -0.00000438, -0.00000160, 0.00000105, 0.00000389, -0.00000139, -0.00000191, -0.00000459, -0.00000028, -0.00000673, -0.00000878, -0.00000959, -0.00005913, 0.00000964, 0.00008321, 0.00003839, 0.00008903, 0.00000800, -0.00005288, -0.00001359, -0.00001041, -0.00000447, -0.00000338, -0.00000035, -0.00001576, 0.00000073, -0.00000440, -0.00000653, -0.00000185, 0.00000301, -0.00001354, 0.00000227, -0.00000974, 0.00000211, 0.00000114, -0.00000217, 0.00001215, -0.00000454, -0.00000562, -0.00000107, 0.00000349, 0.00000191, 0.00000230, 0.00000122, -0.00000426, 0.00000058, -0.00000488, 0.00000196, -0.00000118, 0.00000013, },
		{ 0.00000007, -0.00000126, 0.00000138, -0.00000112, -0.00000362, -0.00000016, -0.00000413, 0.00000425, 0.00000034, 0.00000371, 0.00000130, -0.00000209, -0.00000794, 0.00000341, 0.00000984, -0.00000638, 0.00000768, -0.00000835, -0.00000184, -0.00000448, -0.00000954, 0.00000517, -0.00000608, -0.00000599, -0.00000065, -0.00000682, -0.00001233, 0.00000359, -0.00000781, -0.00000386, -0.00001241, -0.00002796, -0.00004489, 0.00006616, 0.00006014, 0.00005077, 0.00007025, -0.00003721, -0.00004005, -0.00000178, -0.00001236, -0.00000002, -0.00000414, -0.00000252, -0.00000280, 0.00000208, 0.00000240, 0.00000105, -0.00000495, -0.00000030, -0.00000294, 0.00000213, 0.00000496, -0.00000484, -0.00000130, -0.00000228, -0.00000170, 0.00000320, 0.00000575, 0.00000000, 0.00000452, -0.00000851, -0.00000524, 0.00000324, 0.00000334, -0.00000299, -0.00001039, 0.00000376, 0.00000070, 0.00000498, 0.00000372, -0.00000123, -0.00000209, -0.00000135, -0.00000464, 0.00000293, 0.00000378, -0.00000300, 0.00000007, -0.00000537, 0.00000086, 0.00000175, 0.00000301, -0.00000327, -0.00000215, -0.00000381, -0.00000039, -0.00001206, -0.00000316, -0.00003138, -0.00005371, 0.00006431, 0.00006345, 0.00005958, 0.00007760, -0.00004830, -0.00003082, -0.00001239, -0.00000570, -0.00000672, 0.00000285, -0.00001159, -0.00000783, 0.00000030, -0.00000583, -0.00000697, 0.00000426, -0.00000599, -0.00000792, -0.00000045, -0.00000868, 0.00000721, -0.00000535, 0.00000649, 0.00000686, -0.00000818, -0.00000289, 0.00000006, 0.00000479, 0.00000032, 0.00000407, -0.00000346, -0.00000084, -0.00000227, -0.00000211, 0.00000115, -0.00000125, 0.00000036, },
		{ -0.00000003, -0.00000132, 0.00000292, -0.00000494, -0.00000033, -0.00000346, 0.00000004, 0.00000337, 0.00000061, 0.00000381, -0.00000026, -0.00000432, -0.00000663, 0.00001112, 0.00000104, -0.00000144, 0.00000481, -0.00001228, 0.00000437, -0.00001314, 0.00000025, 0.00000019, -0.00000648, -0.00000493, 0.00000053, -0.00001510, -0.00000104, -0.00000267, -0.00000552, -0.00000752, -0.00001512, -0.00004576, 0.00000185, 0.00008316, 0.00003557, 0.00007237, 0.00002699, -0.00005509, -0.00001644, -0.00000559, -0.00000847, 0.00000120, -0.00000617, -0.00000092, -0.00000223, 0.00000400, 0.00000127, -0.00000092, -0.00000476, 0.00000102, -0.00000450, 0.00000755, -0.00000117, -0.00000311, -0.00000184, -0.00000190, -0.00000106, 0.00000698, 0.00000153, 0.00000281, 0.00000011, -0.00001029, -0.00000013, 0.00000335, 0.00000256, -0.00000870, -0.00000508, 0.00000486, 0.00000034, 0.00000714, -0.00000031, -0.00000063, -0.00000265, -0.00000165, -0.00000425, 0.00000712, -0.00000179, 0.00000031, -0.00000397, -0.00000228, 0.00000093, 0.00000351, -0.00000031, -0.00000303, -0.00000311, -0.00000135, -0.00000548, -0.00001117, -0.00000513, -0.00005751, -0.00001222, 0.00009240, 0.00004353, 0.00009409, 0.00001561, -0.00006046, -0.00001233, -0.00001305, -0.00000374, -0.00000388, 0.00000023, -0.00001637, 0.00000087, -0.00000389, -0.00000631, -0.00000394, 0.00000500, -0.00001264, -0.00000065, -0.00000722, -0.00000003, 0.00000302, -0.00000427, 0.00001165, -0.00000149, -0.00000704, -0.00000170, 0.00000263, 0.00000354, 0.00000120, 0.00000227, -0.00000486, 0.00000117, -0.00000450, 0.00000108, -0.00000122, 0.00000029, -0.00000021, },
		{ -0.00000134, 0.00000141, 0.00000017, -0.00000492, 0.00000024, -0.00000446, 0.00000397, 0.00000085, 0.00000232, 0.00000248, -0.00000146, -0.00000707, -0.00000016, 0.00001231, -0.00000623, 0.00000694, -0.00000654, -0.00000466, -0.00000032, -0.00001279, 0.00000547, -0.00000507, -0.00000583, -0.00000186, -0.00000557, -0.00001333, 0.00000459, -0.00000907, -0.00000209, -0.00001226, -0.00002575, -0.00004099, 0.00005555, 0.00006020, 0.00004089, 0.00007070, -0.00001975, -0.00004567, -0.00000293, -0.00001114, -0.00000079, -0.00000340, -0.00000376, -0.00000173, 0.00000079, 0.00000316, 0.00000112, -0.00000434, -0.00000085, -0.00000252, 0.00000014, 0.00000677, -0.00000462, -0.00000149, -0.00000218, -0.00000225, 0.00000254, 0.00000649, -0.00000038, 0.00000462, -0.00000619, -0.00000722, 0.00000282, 0.00000327, -0.00000075, -0.00001152, 0.00000225, 0.00000170, 0.00000417, 0.00000420, -0.00000101, -0.00000155, -0.00000163, -0.00000412, 0.00000083, 0.00000501, -0.00000270, 0.00000028, -0.00000561, 0.00000051, 0.00000132, 0.00000347, -0.00000307, -0.00000264, -0.00000252, -0.00000164, -0.00001066, -0.00000622, -0.00002152, -0.00006666, 0.00005143, 0.00007872, 0.00005892, 0.00008970, -0.00005111, -0.00003427, -0.00001236, -0.00000743, -0.00000593, 0.00000244, -0.00001114, -0.00000863, 0.00000099, -0.00000544, -0.00000766, 0.00000285, -0.00000251, -0.00001026, -0.00000057, -0.00000780, 0.00000596, -0.00000364, 0.00000289, 0.00000962, -0.00000749, -0.00000396, -0.00000110, 0.00000539, 0.00000087, 0.00000348, -0.00000253, -0.00000169, -0.00000097, -0.00000280, 0.00000085, -0.00000137, 0.00000069, -0.00000015, },
		{ -0.00000164, 0.00000383, -0.00000459, -0.00000156, -0.00000254, -0.00000114, 0.00000423, -0.00000021, 0.00000354, 0.00000079, -0.00000331, -0.00000764, 0.00000875, 0.00000493, -0.00000419, 0.00000738, -0.00001405, 0.00000502, -0.00001121, -0.00000326, 0.00000218, -0.00000633, -0.00000521, 0.00000000, -0.00001412, -0.00000213, -0.00000146, -0.00000714, -0.00000419, -0.00001704, -0.00003899, -0.00000313, 0.00007693, 0.00003404, 0.00006163, 0.00003899, -0.00004631, -0.00002423, -0.00000270, -0.00000992, 0.00000242, -0.00000729, -0.00000048, -0.00000255, 0.00000367, 0.00000169, -0.00000033, -0.00000505, 0.00000135, -0.00000546, 0.00000684, 0.00000083, -0.00000378, -0.00000154, -0.00000215, -0.00000181, 0.00000678, 0.00000235, 0.00000174, 0.00000193, -0.00000972, -0.00000212, 0.00000347, 0.00000311, -0.00000669, -0.00000798, 0.00000545, 0.00000009, 0.00000717, -0.00000026, 0.00000013, -0.00000288, -0.00000107, -0.00000520, 0.00000616, -0.00000036, 0.00000015, -0.00000363, -0.00000289, 0.00000079, 0.00000308, 0.00000067, -0.00000384, -0.00000226, -0.00000144, -0.00000537, -0.00001198, -0.00000386, -0.00005038, -0.00003662, 0.00009759, 0.00005210, 0.00009790, 0.00002502, -0.00006870, -0.00001110, -0.00001576, -0.00000299, -0.00000451, 0.00000106, -0.00001725, 0.00000120, -0.00000363, -0.00000569, -0.00000608, 0.00000628, -0.00001082, -0.00000373, -0.00000541, -0.00000116, 0.00000391, -0.00000510, 0.00000974, 0.00000218, -0.00000829, -0.00000226, 0.00000136, 0.00000529, 0.00000027, 0.00000305, -0.00000519, 0.00000142, -0.00000389, 0.00000033, -0.00000139, 0.00000038, -0.00000009, -0.00000017, },
		{ 0.00000119, 0.00000174, -0.00000604, 0.00000029, -0.00000444, 0.00000330, 0.00000170, 0.00000086, 0.00000338, -0.00000082, -0.00000592, -0.00000330, 0.00001343, -0.00000464, 0.00000487, -0.00000337, -0.00000847, 0.00000392, -0.00001521, 0.00000495, -0.00000390, -0.00000544, -0.00000320, -0.00000429, -0.00001437, 0.00000557, -0.00001018, -0.00000069, -0.00001161, -0.00002441, -0.00003653, 0.00004567, 0.00005976, 0.00003337, 0.00006753, -0.00000370, -0.00004703, -0.00000677, -0.00000851, -0.00000255, -0.00000177, -0.00000559, -0.00000034, -0.00000062, 0.00000389, 0.00000115, -0.00000358, -0.00000156, -0.00000176, -0.00000203, 0.00000821, -0.00000397, -0.00000189, -0.00000193, -0.00000270, 0.00000162, 0.00000719, -0.00000047, 0.00000424, -0.00000374, -0.00000864, 0.00000197, 0.00000320, 0.00000117, -0.00001170, -0.00000002, 0.00000293, 0.00000331, 0.00000477, -0.00000114, -0.00000067, -0.00000208, -0.00000336, -0.00000120, 0.00000577, -0.00000213, 0.00000043, -0.00000571, 0.00000004, 0.00000109, 0.00000349, -0.00000232, -0.00000373, -0.00000071, -0.00000318, -0.00000893, -0.00000972, -0.00001255, -0.00007381, 0.00003105, 0.00009551, 0.00005908, 0.00010198, -0.00005297, -0.00003843, -0.00001223, -0.00000917, -0.00000530, 0.00000221, -0.00001078, -0.00000942, 0.00000164, -0.00000504, -0.00000798, 0.00000101, 0.00000067, -0.00001140, -0.00000194, -0.00000626, 0.00000451, -0.00000184, -0.00000035, 0.00001122, -0.00000573, -0.00000532, -0.00000207, 0.00000539, 0.00000201, 0.00000253, -0.00000141, -0.00000264, 0.00000022, -0.00000324, 0.00000058, -0.00000167, 0.00000105, -0.00000031, -0.00000143, },
		{ 0.00000457, -0.00000376, -0.00000303, -0.00000164, -0.00000218, 0.00000475, -0.00000047, 0.00000272, 0.00000202, -0.00000261, -0.00000765, 0.00000543, 0.00000888, -0.00000642, 0.00000910, -0.00001439, 0.00000384, -0.00000775, -0.00000730, 0.00000402, -0.00000619, -0.00000506, -0.00000100, -0.00001265, -0.00000377, 0.00000034, -0.00000936, -0.00000043, -0.00001928, -0.00003263, -0.00000709, 0.00007059, 0.00003314, 0.00005201, 0.00004584, -0.00003450, -0.00003121, -0.00000123, -0.00001035, 0.00000282, -0.00000753, -0.00000085, -0.00000221, 0.00000286, 0.00000232, 0.00000013, -0.00000515, 0.00000151, -0.00000597, 0.00000550, 0.00000300, -0.00000425, -0.00000142, -0.00000209, -0.00000268, 0.00000640, 0.00000330, 0.00000082, 0.00000310, -0.00000835, -0.00000430, 0.00000355, 0.00000322, -0.00000417, -0.00001065, 0.00000548, 0.00000015, 0.00000713, -0.00000027, 0.00000072, -0.00000281, -0.00000070, -0.00000572, 0.00000484, 0.00000094, 0.00000023, -0.00000338, -0.00000334, 0.00000051, 0.00000279, 0.00000128, -0.00000400, -0.00000239, -0.00000026, -0.00000663, -0.00001098, -0.00000575, -0.00003898, -0.00006073, 0.00009624, 0.00006493, 0.00010029, 0.00003641, -0.00007772, -0.00000970, -0.00001881, -0.00000195, -0.00000557, 0.00000236, -0.00001856, 0.00000188, -0.00000372, -0.00000465, -0.00000821, 0.00000693, -0.00000850, -0.00000637, -0.00000475, -0.00000111, 0.00000383, -0.00000479, 0.00000677, 0.00000598, -0.00000906, -0.00000288, -0.00000015, 0.00000691, -0.00000030, 0.00000347, -0.00000522, 0.00000138, -0.00000314, -0.00000026, -0.00000162, 0.00000034, 0.00000010, -0.00000025, -0.00000209, },
		{ 0.00000347, -0.00000684, -0.00000008, -0.00000407, 0.00000233, 0.00000271, -0.00000045, 0.00000385, -0.00000002, -0.00000482, -0.00000558, 0.00001296, -0.00000162, 0.00000167, 0.00000071, -0.00001253, 0.00000745, -0.00001629, 0.00000343, -0.00000250, -0.00000500, -0.00000436, -0.00000327, -0.00001511, 0.00000615, -0.00001076, -0.00000004, -0.00001013, -0.00002417, -0.00003138, 0.00003634, 0.00005906, 0.00002752, 0.00006258, 0.00000922, -0.00004384, -0.00001274, -0.00000511, -0.00000478, 0.00000024, -0.00000755, 0.00000098, -0.00000185, 0.00000439, 0.00000123, -0.00000273, -0.00000239, -0.00000077, -0.00000420, 0.00000910, -0.00000286, -0.00000239, -0.00000165, -0.00000291, 0.00000042, 0.00000781, -0.00000027, 0.00000354, -0.00000147, -0.00000930, 0.00000061, 0.00000323, 0.00000257, -0.00001082, -0.00000295, 0.00000426, 0.00000244, 0.00000547, -0.00000164, 0.00000046, -0.00000259, -0.00000247, -0.00000308, 0.00000608, -0.00000139, 0.00000060, -0.00000576, -0.00000047, 0.00000096, 0.00000326, -0.00000132, -0.00000497, 0.00000098, -0.00000422, -0.00000783, -0.00001231, -0.00000636, -0.00007374, 0.00000387, 0.00011174, 0.00006149, 0.00011373, -0.00005335, -0.00004366, -0.00001172, -0.00001124, -0.00000450, 0.00000178, -0.00001014, -0.00001054, 0.00000256, -0.00000494, -0.00000772, -0.00000128, 0.00000347, -0.00001147, -0.00000406, -0.00000473, 0.00000347, -0.00000045, -0.00000274, 0.00001138, -0.00000292, -0.00000688, -0.00000283, 0.00000477, 0.00000364, 0.00000135, -0.00000022, -0.00000359, 0.00000126, -0.00000349, 0.00000040, -0.00000211, 0.00000139, -0.00000042, -0.00000147, 0.00000066, },
		{ -0.00000241, -0.00000464, -0.00000089, -0.00000294, 0.00000483, -0.00000018, 0.00000147, 0.00000331, -0.00000214, -0.00000690, 0.00000175, 0.00001220, -0.00000752, 0.00000940, -0.00001289, 0.00000079, -0.00000308, -0.00001147, 0.00000553, -0.00000607, -0.00000447, -0.00000246, -0.00001076, -0.00000585, 0.00000259, -0.00001198, 0.00000350, -0.00002152, -0.00002694, -0.00000996, 0.00006416, 0.00003260, 0.00004376, 0.00004855, -0.00002175, -0.00003575, -0.00000199, -0.00000933, 0.00000212, -0.00000672, -0.00000209, -0.00000122, 0.00000164, 0.00000311, 0.00000044, -0.00000503, 0.00000141, -0.00000594, 0.00000356, 0.00000518, -0.00000441, -0.00000153, -0.00000178, -0.00000354, 0.00000571, 0.00000434, 0.00000012, 0.00000363, -0.00000641, -0.00000638, 0.00000344, 0.00000302, -0.00000147, -0.00001271, 0.00000473, 0.00000058, 0.00000695, -0.00000018, 0.00000096, -0.00000235, -0.00000057, -0.00000584, 0.00000326, 0.00000201, 0.00000053, -0.00000324, -0.00000361, 0.00000005, 0.00000276, 0.00000136, -0.00000338, -0.00000358, 0.00000213, -0.00000905, -0.00000852, -0.00000995, -0.00002542, -0.00008129, 0.00008616, 0.00008217, 0.00010159, 0.00004964, -0.00008739, -0.00000813, -0.00002227, -0.00000054, -0.00000709, 0.00000415, -0.00002028, 0.00000283, -0.00000409, -0.00000334, -0.00001016, 0.00000697, -0.00000597, -0.00000820, -0.00000536, -0.00000005, 0.00000308, -0.00000371, 0.00000327, 0.00000936, -0.00000902, -0.00000370, -0.00000171, 0.00000820, -0.00000036, 0.00000348, -0.00000494, 0.00000106, -0.00000229, -0.00000073, -0.00000180, 0.00000011, 0.00000036, -0.00000029, -0.00000258, 0.00000500, },
		{ -0.00000717, -0.00000091, -0.00000342, 0.00000119, 0.00000367, -0.00000139, 0.00000376, 0.00000105, -0.00000400, -0.00000676, 0.00001093, 0.00000249, -0.00000212, 0.00000494, -0.00001594, 0.00000950, -0.00001562, 0.00000080, -0.00000079, -0.00000472, -0.00000507, -0.00000281, -0.00001524, 0.00000601, -0.00001048, -0.00000042, -0.00000761, -0.00002513, -0.00002555, 0.00002743, 0.00005832, 0.00002268, 0.00005724, 0.00001818, -0.00003673, -0.00001966, -0.00000188, -0.00000677, 0.00000204, -0.00000913, 0.00000183, -0.00000262, 0.00000450, 0.00000145, -0.00000187, -0.00000324, 0.00000031, -0.00000616, 0.00000933, -0.00000133, -0.00000292, -0.00000147, -0.00000279, -0.00000102, 0.00000827, 0.00000018, 0.00000271, 0.00000040, -0.00000912, -0.00000121, 0.00000340, 0.00000335, -0.00000894, -0.00000629, 0.00000547, 0.00000163, 0.00000629, -0.00000241, 0.00000166, -0.00000302, -0.00000156, -0.00000470, 0.00000598, -0.00000061, 0.00000088, -0.00000584, -0.00000090, 0.00000078, 0.00000302, -0.00000046, -0.00000583, 0.00000189, -0.00000404, -0.00000818, -0.00001295, -0.00000414, -0.00006624, -0.00002805, 0.00012448, 0.00006779, 0.00012418, -0.00005164, -0.00005040, -0.00001042, -0.00001407, -0.00000311, 0.00000073, -0.00000880, -0.00001235, 0.00000405, -0.00000541, -0.00000673, -0.00000400, 0.00000588, -0.00001078, -0.00000633, -0.00000384, 0.00000330, 0.00000017, -0.00000402, 0.00001006, 0.00000072, -0.00000843, -0.00000343, 0.00000357, 0.00000561, 0.00000013, 0.00000092, -0.00000446, 0.00000213, -0.00000360, 0.00000035, -0.00000264, 0.00000165, -0.00000047, -0.00000141, -0.00000015, 0.00000521, },
		{ -0.00000623, -0.00000041, -0.00000335, 0.00000445, 0.00000056, -0.00000002, 0.00000447, -0.00000173, -0.00000578, -0.00000171, 0.00001427, -0.00000701, 0.00000796, -0.00000952, -0.00000375, 0.00000214, -0.00001518, 0.00000639, -0.00000585, -0.00000359, -0.00000414, -0.00000870, -0.00000809, 0.00000493, -0.00001458, 0.00000716, -0.00002329, -0.00002231, -0.00001153, 0.00005748, 0.00003247, 0.00003661, 0.00004859, -0.00001008, -0.00003670, -0.00000525, -0.00000689, 0.00000039, -0.00000499, -0.00000400, 0.00000018, 0.00000020, 0.00000393, 0.00000066, -0.00000467, 0.00000101, -0.00000536, 0.00000115, 0.00000719, -0.00000420, -0.00000182, -0.00000135, -0.00000418, 0.00000464, 0.00000544, -0.00000033, 0.00000363, -0.00000418, -0.00000810, 0.00000299, 0.00000270, 0.00000108, -0.00001384, 0.00000311, 0.00000141, 0.00000656, 0.00000018, 0.00000069, -0.00000148, -0.00000067, -0.00000560, 0.00000154, 0.00000282, 0.00000103, -0.00000315, -0.00000375, -0.00000058, 0.00000298, 0.00000094, -0.00000210, -0.00000554, 0.00000526, -0.00001197, -0.00000551, -0.00001502, -0.00001229, -0.00009525, 0.00006612, 0.00010292, 0.00010286, 0.00006414, -0.00009729, -0.00000668, -0.00002593, 0.00000101, -0.00000886, 0.00000617, -0.00002211, 0.00000374, -0.00000442, -0.00000207, -0.00001166, 0.00000637, -0.00000343, -0.00000909, -0.00000707, 0.00000155, 0.00000215, -0.00000236, -0.00000017, 0.00001180, -0.00000796, -0.00000478, -0.00000319, 0.00000896, 0.00000017, 0.00000307, -0.00000437, 0.00000050, -0.00000136, -0.00000113, -0.00000184, -0.00000031, 0.00000067, -0.00000028, -0.00000302, 0.00000501, -0.00000060, },
		{ -0.00000214, -0.00000261, 0.00000005, 0.00000437, -0.00000181, 0.00000305, 0.00000241, -0.00000357, -0.00000683, 0.00000767, 0.00000708, -0.00000575, 0.00000847, -0.00001784, 0.00000949, -0.00001303, -0.00000283, 0.00000117, -0.00000474, -0.00000511, -0.00000311, -0.00001457, 0.00000499, -0.00000923, -0.00000189, -0.00000406, -0.00002716, -0.00001923, 0.00001898, 0.00005759, 0.00001846, 0.00005228, 0.00002328, -0.00002710, -0.00002600, 0.00000011, -0.00000781, 0.00000309, -0.00000989, 0.00000191, -0.00000275, 0.00000413, 0.00000184, -0.00000107, -0.00000401, 0.00000132, -0.00000768, 0.00000879, 0.00000054, -0.00000334, -0.00000148, -0.00000232, -0.00000258, 0.00000844, 0.00000088, 0.00000188, 0.00000173, -0.00000811, -0.00000335, 0.00000365, 0.00000351, -0.00000626, -0.00000971, 0.00000632, 0.00000100, 0.00000715, -0.00000329, 0.00000270, -0.00000321, -0.00000074, -0.00000602, 0.00000553, 0.00000011, 0.00000134, -0.00000603, -0.00000115, 0.00000040, 0.00000302, -0.00000010, -0.00000588, 0.00000155, -0.00000217, -0.00001041, -0.00001115, -0.00000617, -0.00005243, -0.00006144, 0.00013040, 0.00007947, 0.00013280, -0.00004738, -0.00005898, -0.00000796, -0.00001803, -0.00000076, -0.00000129, -0.00000647, -0.00001509, 0.00000629, -0.00000658, -0.00000497, -0.00000707, 0.00000791, -0.00000962, -0.00000821, -0.00000405, 0.00000418, -0.00000002, -0.00000419, 0.00000750, 0.00000477, -0.00000967, -0.00000396, 0.00000193, 0.00000773, -0.00000097, 0.00000188, -0.00000519, 0.00000280, -0.00000359, 0.00000041, -0.00000317, 0.00000176, -0.00000047, -0.00000123, -0.00000117, 0.00000679, -0.00000689, },
		{ -0.00000034, -0.00000336, 0.00000366, 0.00000161, -0.00000150, 0.00000528, -0.00000118, -0.00000464, -0.00000439, 0.00001468, -0.00000475, 0.00000487, -0.00000468, -0.00000902, 0.00000705, -0.00001777, 0.00000626, -0.00000533, -0.00000267, -0.00000568, -0.00000687, -0.00001005, 0.00000689, -0.00001667, 0.00001003, -0.00002411, -0.00001911, -0.00001157, 0.00005030, 0.00003299, 0.00003006, 0.00004742, -0.00000100, -0.00003368, -0.00001062, -0.00000353, -0.00000192, -0.00000275, -0.00000617, 0.00000164, -0.00000119, 0.00000461, 0.00000085, -0.00000410, 0.00000032, -0.00000429, -0.00000152, 0.00000882, -0.00000356, -0.00000226, -0.00000094, -0.00000446, 0.00000315, 0.00000654, -0.00000050, 0.00000323, -0.00000193, -0.00000921, 0.00000208, 0.00000245, 0.00000315, -0.00001384, 0.00000061, 0.00000256, 0.00000594, 0.00000089, -0.00000011, -0.00000025, -0.00000092, -0.00000507, -0.00000025, 0.00000335, 0.00000163, -0.00000303, -0.00000381, -0.00000128, 0.00000333, 0.00000021, -0.00000048, -0.00000779, 0.00000839, -0.00001445, -0.00000310, -0.00001928, -0.00000208, -0.00010036, 0.00003629, 0.00012499, 0.00010584, 0.00007898, -0.00010666, -0.00000591, -0.00002929, 0.00000222, -0.00001039, 0.00000794, -0.00002355, 0.00000415, -0.00000430, -0.00000127, -0.00001240, 0.00000502, -0.00000093, -0.00000916, -0.00000941, 0.00000308, 0.00000160, -0.00000123, -0.00000302, 0.00001293, -0.00000580, -0.00000609, -0.00000449, 0.00000908, 0.00000131, 0.00000228, -0.00000356, -0.00000025, -0.00000037, -0.00000153, -0.00000167, -0.00000092, 0.00000098, -0.00000026, -0.00000331, 0.00000455, 0.00000158, -0.00000761, },
		{ -0.00000179, -0.00000092, 0.00000469, -0.00000165, 0.00000180, 0.00000397, -0.00000347, -0.00000601, 0.00000372, 0.00001139, -0.00000845, 0.00001050, -0.00001761, 0.00000719, -0.00000872, -0.00000712, 0.00000320, -0.00000506, -0.00000447, -0.00000412, -0.00001316, 0.00000319, -0.00000716, -0.00000420, 0.00000021, -0.00002985, -0.00001282, 0.00001126, 0.00005669, 0.00001484, 0.00004784, 0.00002538, -0.00001676, -0.00003023, 0.00000004, -0.00000742, 0.00000303, -0.00000961, 0.00000112, -0.00000222, 0.00000333, 0.00000240, -0.00000036, -0.00000461, 0.00000210, -0.00000860, 0.00000747, 0.00000262, -0.00000356, -0.00000172, -0.00000159, -0.00000407, 0.00000820, 0.00000180, 0.00000116, 0.00000247, -0.00000643, -0.00000559, 0.00000388, 0.00000319, -0.00000309, -0.00001280, 0.00000654, 0.00000067, 0.00000791, -0.00000406, 0.00000335, -0.00000303, -0.00000007, -0.00000699, 0.00000477, 0.00000071, 0.00000200, -0.00000634, -0.00000117, -0.00000028, 0.00000341, -0.00000044, -0.00000489, -0.00000021, 0.00000141, -0.00001440, -0.00000719, -0.00001168, -0.00003464, -0.00009228, 0.00012628, 0.00009732, 0.00013969, -0.00004054, -0.00006947, -0.00000424, -0.00002328, 0.00000266, -0.00000437, -0.00000310, -0.00001873, 0.00000921, -0.00000834, -0.00000264, -0.00001025, 0.00000949, -0.00000824, -0.00000935, -0.00000550, 0.00000601, -0.00000083, -0.00000350, 0.00000413, 0.00000872, -0.00001030, -0.00000454, 0.00000001, 0.00000975, -0.00000175, 0.00000257, -0.00000570, 0.00000323, -0.00000346, 0.00000051, -0.00000357, 0.00000168, -0.00000042, -0.00000095, -0.00000228, 0.00000800, -0.00000595, -0.00000371, },
		{ -0.00000299, 0.00000257, 0.00000274, -0.00000274, 0.00000556, -0.00000032, -0.00000380, -0.00000595, 0.00001332, -0.00000094, 0.00000059, 0.00000084, -0.00001402, 0.00001070, -0.00001864, 0.00000489, -0.00000432, -0.00000205, -0.00000669, -0.00000566, -0.00001131, 0.00000803, -0.00001781, 0.00001173, -0.00002363, -0.00001760, -0.00000999, 0.00004249, 0.00003431, 0.00002363, 0.00004615, 0.00000474, -0.00002717, -0.00001709, -0.00000012, -0.00000415, -0.00000058, -0.00000811, 0.00000279, -0.00000224, 0.00000500, 0.00000107, -0.00000337, -0.00000061, -0.00000288, -0.00000422, 0.00000988, -0.00000247, -0.00000273, -0.00000068, -0.00000429, 0.00000130, 0.00000752, -0.00000039, 0.00000258, 0.00000006, -0.00000956, 0.00000068, 0.00000236, 0.00000455, -0.00001262, -0.00000263, 0.00000389, 0.00000514, 0.00000197, -0.00000138, 0.00000115, -0.00000120, -0.00000432, -0.00000203, 0.00000361, 0.00000225, -0.00000279, -0.00000393, -0.00000190, 0.00000362, -0.00000051, 0.00000100, -0.00000966, 0.00001070, -0.00001553, -0.00000242, -0.00002125, 0.00000336, -0.00009569, -0.00000146, 0.00014500, 0.00011279, 0.00009299, -0.00011462, -0.00000653, -0.00003171, 0.00000244, -0.00001107, 0.00000884, -0.00002402, 0.00000355, -0.00000329, -0.00000133, -0.00001210, 0.00000286, 0.00000155, -0.00000865, -0.00001182, 0.00000391, 0.00000188, -0.00000070, -0.00000493, 0.00001258, -0.00000266, -0.00000751, -0.00000558, 0.00000851, 0.00000299, 0.00000122, -0.00000259, -0.00000116, 0.00000065, -0.00000196, -0.00000130, -0.00000165, 0.00000123, -0.00000023, -0.00000340, 0.00000362, 0.00000393, -0.00000856, -0.00000077, },
		{ -0.00000160, 0.00000453, -0.00000093, 0.00000015, 0.00000556, -0.00000353, -0.00000465, -0.00000026, 0.00001469, -0.00000955, 0.00001050, -0.00001502, 0.00000283, -0.00000324, -0.00001149, 0.00000495, -0.00000550, -0.00000331, -0.00000560, -0.00001131, 0.00000094, -0.00000469, -0.00000688, 0.00000466, -0.00003261, -0.00000690, 0.00000468, 0.00005523, 0.00001209, 0.00004357, 0.00002576, -0.00000753, -0.00003123, -0.00000248, -0.00000553, 0.00000189, -0.00000835, -0.00000041, -0.00000119, 0.00000223, 0.00000302, 0.00000026, -0.00000498, 0.00000252, -0.00000882, 0.00000543, 0.00000474, -0.00000349, -0.00000216, -0.00000073, -0.00000529, 0.00000745, 0.00000291, 0.00000061, 0.00000267, -0.00000430, -0.00000768, 0.00000392, 0.00000256, 0.00000017, -0.00001514, 0.00000590, 0.00000076, 0.00000843, -0.00000447, 0.00000340, -0.00000241, 0.00000040, -0.00000758, 0.00000375, 0.00000116, 0.00000282, -0.00000673, -0.00000098, -0.00000125, 0.00000423, -0.00000149, -0.00000294, -0.00000319, 0.00000627, -0.00001947, -0.00000199, -0.00001907, -0.00001598, -0.00011645, 0.00010981, 0.00012093, 0.00014581, -0.00003162, -0.00008148, 0.00000048, -0.00002959, 0.00000692, -0.00000826, 0.00000101, -0.00002293, 0.00001243, -0.00001032, -0.00000013, -0.00001318, 0.00001048, -0.00000675, -0.00000964, -0.00000802, 0.00000836, -0.00000181, -0.00000238, 0.00000050, 0.00001202, -0.00001003, -0.00000528, -0.00000202, 0.00001143, -0.00000204, 0.00000288, -0.00000594, 0.00000339, -0.00000316, 0.00000054, -0.00000373, 0.00000135, -0.00000033, -0.00000061, -0.00000335, 0.00000869, -0.00000434, -0.00000544, -0.00000113, },
		{ 0.00000134, 0.00000374, -0.00000351, 0.00000520, 0.00000093, -0.00000344, -0.00000628, 0.00001040, 0.00000389, -0.00000409, 0.00000606, -0.00001771, 0.00001228, -0.00001739, 0.00000218, -0.00000273, -0.00000195, -0.00000687, -0.00000535, -0.00001160, 0.00000810, -0.00001778, 0.00001208, -0.00002174, -0.00001773, -0.00000695, 0.00003412, 0.00003641, 0.00001715, 0.00004532, 0.00000726, -0.00001837, -0.00002326, 0.00000231, -0.00000559, 0.00000096, -0.00000938, 0.00000329, -0.00000275, 0.00000501, 0.00000138, -0.00000254, -0.00000166, -0.00000130, -0.00000666, 0.00001020, -0.00000096, -0.00000315, -0.00000066, -0.00000365, -0.00000076, 0.00000822, 0.00000002, 0.00000184, 0.00000162, -0.00000907, -0.00000116, 0.00000248, 0.00000518, -0.00001030, -0.00000632, 0.00000515, 0.00000427, 0.00000332, -0.00000293, 0.00000250, -0.00000136, -0.00000347, -0.00000370, 0.00000364, 0.00000279, -0.00000236, -0.00000419, -0.00000229, 0.00000364, -0.00000089, 0.00000188, -0.00001056, 0.00001148, -0.00001449, -0.00000421, -0.00002005, 0.00000323, -0.00008199, -0.00004358, 0.00015871, 0.00012599, 0.00010521, -0.00012040, -0.00000920, -0.00003254, 0.00000104, -0.00001032, 0.00000834, -0.00002305, 0.00000154, -0.00000107, -0.00000247, -0.00001065, -0.00000007, 0.00000400, -0.00000785, -0.00001378, 0.00000357, 0.00000323, -0.00000090, -0.00000577, 0.00001085, 0.00000116, -0.00000880, -0.00000650, 0.00000731, 0.00000508, 0.00000005, -0.00000155, -0.00000214, 0.00000167, -0.00000241, -0.00000076, -0.00000237, 0.00000137, -0.00000020, -0.00000327, 0.00000231, 0.00000623, -0.00000890, -0.00000171, -0.00000231, },
		{ 0.00000389, 0.00000021, -0.00000166, 0.00000694, -0.00000353, -0.00000318, -0.00000363, 0.00001637, -0.00000864, 0.00000835, -0.00001033, -0.00000291, 0.00000250, -0.00001518, 0.00000596, -0.00000577, -0.00000201, -0.00000710, -0.00000949, -0.00000124, -0.00000239, -0.00000934, 0.00000866, -0.00003479, -0.00000208, -0.00000030, 0.00005278, 0.00001059, 0.00003893, 0.00002572, -0.00000081, -0.00002850, -0.00000721, -0.00000251, -0.00000001, -0.00000645, -0.00000235, 0.00000004, 0.00000107, 0.00000359, 0.00000080, -0.00000506, 0.00000252, -0.00000832, 0.00000283, 0.00000671, -0.00000305, -0.00000274, 0.00000009, -0.00000604, 0.00000613, 0.00000413, 0.00000028, 0.00000241, -0.00000200, -0.00000933, 0.00000361, 0.00000187, 0.00000313, -0.00001638, 0.00000427, 0.00000131, 0.00000859, -0.00000436, 0.00000277, -0.00000138, 0.00000069, -0.00000779, 0.00000250, 0.00000147, 0.00000371, -0.00000713, -0.00000066, -0.00000240, 0.00000532, -0.00000304, -0.00000040, -0.00000680, 0.00001156, -0.00002453, 0.00000302, -0.00002624, 0.00000027, -0.00013051, 0.00008021, 0.00014832, 0.00015314, -0.00002175, -0.00009420, 0.00000554, -0.00003635, 0.00001141, -0.00001237, 0.00000526, -0.00002706, 0.00001536, -0.00001198, 0.00000201, -0.00001540, 0.00001065, -0.00000513, -0.00000920, -0.00001117, 0.00001063, -0.00000245, -0.00000130, -0.00000283, 0.00001423, -0.00000871, -0.00000621, -0.00000400, 0.00001257, -0.00000172, 0.00000279, -0.00000587, 0.00000325, -0.00000266, 0.00000042, -0.00000357, 0.00000081, -0.00000023, -0.00000024, -0.00000423, 0.00000875, -0.00000214, -0.00000713, -0.00000079, -0.00000188, },
		{ 0.00000443, -0.00000370, 0.00000418, 0.00000256, -0.00000356, -0.00000551, 0.00000640, 0.00000896, -0.00000827, 0.00000996, -0.00001920, 0.00001131, -0.00001399, -0.00000168, -0.00000067, -0.00000243, -0.00000612, -0.00000600, -0.00001091, 0.00000711, -0.00001667, 0.00001123, -0.00001869, -0.00001914, -0.00000288, 0.00002554, 0.00003895, 0.00001081, 0.00004489, 0.00000739, -0.00000896, -0.00002765, 0.00000287, -0.00000567, 0.00000149, -0.00000968, 0.00000298, -0.00000266, 0.00000461, 0.00000177, -0.00000166, -0.00000272, 0.00000022, -0.00000859, 0.00000968, 0.00000089, -0.00000342, -0.00000093, -0.00000264, -0.00000281, 0.00000849, 0.00000074, 0.00000110, 0.00000264, -0.00000784, -0.00000329, 0.00000275, 0.00000511, -0.00000713, -0.00001006, 0.00000604, 0.00000350, 0.00000476, -0.00000448, 0.00000352, -0.00000124, -0.00000261, -0.00000518, 0.00000342, 0.00000321, -0.00000170, -0.00000464, -0.00000236, 0.00000323, -0.00000071, 0.00000184, -0.00001014, 0.00001039, -0.00001106, -0.00000860, -0.00001566, -0.00000202, -0.00006158, -0.00008532, 0.00016168, 0.00014719, 0.00011527, -0.00012363, -0.00001428, -0.00003141, -0.00000232, -0.00000786, 0.00000619, -0.00002046, -0.00000196, 0.00000236, -0.00000467, -0.00000820, -0.00000354, 0.00000632, -0.00000695, -0.00001494, 0.00000190, 0.00000557, -0.00000170, -0.00000568, 0.00000806, 0.00000518, -0.00000966, -0.00000733, 0.00000559, 0.00000735, -0.00000106, -0.00000058, -0.00000309, 0.00000261, -0.00000283, -0.00000017, -0.00000295, 0.00000131, -0.00000018, -0.00000293, 0.00000078, 0.00000822, -0.00000849, -0.00000313, -0.00000148, 0.00000015, },
		{ 0.00000159, -0.00000337, 0.00000788, -0.00000323, -0.00000198, -0.00000591, 0.00001613, -0.00000575, 0.00000434, -0.00000427, -0.00000898, 0.00000745, -0.00001740, 0.00000579, -0.00000554, -0.00000099, -0.00000811, -0.00000822, -0.00000280, -0.00000080, -0.00001099, 0.00001164, -0.00003582, 0.00000118, -0.00000337, 0.00004901, 0.00001065, 0.00003345, 0.00002625, 0.00000269, -0.00002237, -0.00001331, 0.00000084, -0.00000203, -0.00000441, -0.00000425, 0.00000112, 0.00000009, 0.00000397, 0.00000129, -0.00000485, 0.00000210, -0.00000719, -0.00000011, 0.00000829, -0.00000220, -0.00000337, 0.00000071, -0.00000619, 0.00000429, 0.00000533, 0.00000022, 0.00000185, 0.00000022, -0.00001035, 0.00000284, 0.00000130, 0.00000546, -0.00001631, 0.00000167, 0.00000227, 0.00000835, -0.00000365, 0.00000145, -0.00000003, 0.00000086, -0.00000765, 0.00000106, 0.00000168, 0.00000458, -0.00000742, -0.00000036, -0.00000353, 0.00000644, -0.00000469, 0.00000219, -0.00001027, 0.00001625, -0.00002831, 0.00000633, -0.00003115, 0.00001133, -0.00013245, 0.00003873, 0.00017587, 0.00016451, -0.00001245, -0.00010649, 0.00000998, -0.00004267, 0.00001527, -0.00001586, 0.00000885, -0.00003036, 0.00001730, -0.00001267, 0.00000322, -0.00001648, 0.00000980, -0.00000329, -0.00000831, -0.00001439, 0.00001217, -0.00000226, -0.00000068, -0.00000542, 0.00001504, -0.00000635, -0.00000728, -0.00000582, 0.00001303, -0.00000077, 0.00000229, -0.00000548, 0.00000278, -0.00000194, 0.00000008, -0.00000307, 0.00000010, -0.00000016, 0.00000010, -0.00000484, 0.00000816, 0.00000045, -0.00000853, -0.00000091, -0.00000172, 0.00000284, },
		{ -0.00000326, 0.00000259, 0.00000441, -0.00000402, -0.00000397, 0.00000199, 0.00001341, -0.00001106, 0.00001175, -0.00001798, 0.00000778, -0.00000884, -0.00000617, 0.00000154, -0.00000335, -0.00000460, -0.00000736, -0.00000949, 0.00000540, -0.00001486, 0.00000966, -0.00001501, -0.00002122, 0.00000155, 0.00001733, 0.00004136, 0.00000511, 0.00004431, 0.00000639, -0.00000068, -0.00002904, 0.00000102, -0.00000420, 0.00000089, -0.00000899, 0.00000191, -0.00000205, 0.00000392, 0.00000221, -0.00000079, -0.00000365, 0.00000148, -0.00000980, 0.00000827, 0.00000296, -0.00000344, -0.00000146, -0.00000143, -0.00000460, 0.00000818, 0.00000173, 0.00000046, 0.00000313, -0.00000602, -0.00000548, 0.00000301, 0.00000450, -0.00000352, -0.00001336, 0.00000624, 0.00000304, 0.00000607, -0.00000572, 0.00000394, -0.00000073, -0.00000184, -0.00000639, 0.00000293, 0.00000353, -0.00000085, -0.00000527, -0.00000210, 0.00000235, 0.00000012, 0.00000081, -0.00000836, 0.00000751, -0.00000559, -0.00001499, -0.00000896, -0.00001074, -0.00003804, -0.00012133, 0.00015010, 0.00017689, 0.00012382, -0.00012469, -0.00002158, -0.00002842, -0.00000757, -0.00000379, 0.00000255, -0.00001646, -0.00000666, 0.00000668, -0.00000757, -0.00000515, -0.00000713, 0.00000828, -0.00000602, -0.00001519, -0.00000097, 0.00000852, -0.00000272, -0.00000502, 0.00000471, 0.00000889, -0.00000982, -0.00000817, 0.00000352, 0.00000955, -0.00000188, 0.00000018, -0.00000391, 0.00000336, -0.00000312, 0.00000031, -0.00000323, 0.00000102, -0.00000015, -0.00000244, -0.00000081, 0.00000968, -0.00000727, -0.00000490, -0.00000064, -0.00000084, 0.00000466, },
		{ -0.00000473, 0.00000818, -0.00000246, -0.00000133, -0.00000686, 0.00001400, -0.00000126, -0.00000077, 0.00000212, -0.00001418, 0.00001057, -0.00001752, 0.00000413, -0.00000455, -0.00000061, -0.00000821, -0.00000792, -0.00000332, -0.00000035, -0.00001145, 0.00001327, -0.00003544, 0.00000271, -0.00000450, 0.00004388, 0.00001232, 0.00002699, 0.00002778, 0.00000305, -0.00001392, -0.00001949, 0.00000352, -0.00000345, -0.00000283, -0.00000565, 0.00000171, -0.00000047, 0.00000408, 0.00000176, -0.00000438, 0.00000131, -0.00000560, -0.00000307, 0.00000927, -0.00000092, -0.00000396, 0.00000100, -0.00000573, 0.00000207, 0.00000636, 0.00000045, 0.00000110, 0.00000211, -0.00001058, 0.00000157, 0.00000098, 0.00000696, -0.00001489, -0.00000175, 0.00000347, 0.00000777, -0.00000237, -0.00000038, 0.00000140, 0.00000103, -0.00000721, -0.00000052, 0.00000179, 0.00000534, -0.00000749, -0.00000024, -0.00000443, 0.00000729, -0.00000602, 0.00000419, -0.00001279, 0.00001934, -0.00002971, 0.00000674, -0.00003228, 0.00001554, -0.00012215, -0.00001120, 0.00019852, 0.00018319, -0.00000532, -0.00011718, 0.00001280, -0.00004761, 0.00001757, -0.00001790, 0.00001098, -0.00003211, 0.00001762, -0.00001189, 0.00000308, -0.00001613, 0.00000784, -0.00000116, -0.00000728, -0.00001710, 0.00001247, -0.00000096, -0.00000072, -0.00000703, 0.00001442, -0.00000313, -0.00000834, -0.00000745, 0.00001275, 0.00000077, 0.00000148, -0.00000484, 0.00000203, -0.00000100, -0.00000048, -0.00000230, -0.00000065, -0.00000018, 0.00000039, -0.00000510, 0.00000699, 0.00000319, -0.00000940, -0.00000159, -0.00000116, 0.00000151, 0.00000299, },
		{ 0.00000065, 0.00000626, -0.00000457, -0.00000213, -0.00000209, 0.00001645, -0.00001179, 0.00001101, -0.00001408, 0.00000219, -0.00000274, -0.00001055, 0.00000341, -0.00000437, -0.00000270, -0.00000898, -0.00000786, 0.00000353, -0.00001299, 0.00000803, -0.00001142, -0.00002318, 0.00000558, 0.00001014, 0.00004298, 0.00000070, 0.00004288, 0.00000559, 0.00000502, -0.00002683, -0.00000320, -0.00000141, -0.00000058, -0.00000755, 0.00000032, -0.00000115, 0.00000309, 0.00000261, 0.00000003, -0.00000433, 0.00000230, -0.00001017, 0.00000606, 0.00000507, -0.00000317, -0.00000217, -0.00000021, -0.00000588, 0.00000718, 0.00000297, -0.00000003, 0.00000314, -0.00000386, -0.00000746, 0.00000309, 0.00000362, 0.00000008, -0.00001575, 0.00000549, 0.00000304, 0.00000700, -0.00000636, 0.00000357, 0.00000024, -0.00000122, -0.00000721, 0.00000213, 0.00000380, 0.00000008, -0.00000597, -0.00000162, 0.00000114, 0.00000145, -0.00000101, -0.00000559, 0.00000344, 0.00000103, -0.00002216, -0.00000156, -0.00002045, -0.00001555, -0.00014664, 0.00012170, 0.00021380, 0.00013287, -0.00012478, -0.00003027, -0.00002423, -0.00001409, 0.00000124, -0.00000195, -0.00001170, -0.00001190, 0.00001119, -0.00001050, -0.00000216, -0.00001028, 0.00000955, -0.00000494, -0.00001468, -0.00000458, 0.00001146, -0.00000345, -0.00000425, 0.00000136, 0.00001175, -0.00000903, -0.00000911, 0.00000131, 0.00001142, -0.00000223, 0.00000062, -0.00000446, 0.00000381, -0.00000315, 0.00000053, -0.00000309, 0.00000046, -0.00000013, -0.00000187, -0.00000225, 0.00001043, -0.00000526, -0.00000683, -0.00000001, -0.00000147, 0.00000436, -0.00000226, },
		{ 0.00000776, -0.00000116, -0.00000131, -0.00000648, 0.00001036, 0.00000409, -0.00000601, 0.00000764, -0.00001737, 0.00001109, -0.00001524, 0.00000097, -0.00000276, -0.00000109, -0.00000715, -0.00000877, -0.00000265, -0.00000115, -0.00001068, 0.00001355, -0.00003373, 0.00000272, -0.00000398, 0.00003764, 0.00001529, 0.00001981, 0.00003012, 0.00000114, -0.00000472, -0.00002426, 0.00000456, -0.00000365, -0.00000215, -0.00000621, 0.00000160, -0.00000053, 0.00000386, 0.00000221, -0.00000370, 0.00000029, -0.00000379, -0.00000574, 0.00000945, 0.00000075, -0.00000441, 0.00000092, -0.00000472, -0.00000029, 0.00000701, 0.00000101, 0.00000028, 0.00000355, -0.00001002, -0.00000014, 0.00000095, 0.00000756, -0.00001228, -0.00000561, 0.00000463, 0.00000703, -0.00000072, -0.00000243, 0.00000263, 0.00000134, -0.00000656, -0.00000212, 0.00000178, 0.00000594, -0.00000728, -0.00000038, -0.00000493, 0.00000760, -0.00000666, 0.00000512, -0.00001377, 0.00002019, -0.00002812, 0.00000368, -0.00002911, 0.00001273, -0.00010156, -0.00006424, 0.00021042, 0.00021223, -0.00000146, -0.00012551, 0.00001328, -0.00005043, 0.00001764, -0.00001790, 0.00001113, -0.00003185, 0.00001600, -0.00000942, 0.00000144, -0.00001436, 0.00000490, 0.00000119, -0.00000633, -0.00001890, 0.00001127, 0.00000145, -0.00000137, -0.00000770, 0.00001256, 0.00000058, -0.00000914, -0.00000889, 0.00001178, 0.00000277, 0.00000048, -0.00000402, 0.00000107, 0.00000007, -0.00000118, -0.00000140, -0.00000128, -0.00000036, 0.00000062, -0.00000503, 0.00000538, 0.00000578, -0.00000953, -0.00000285, -0.00000030, 0.00000007, 0.00000417, -0.00000554, },
		{ 0.00000782, -0.00000490, -0.00000046, -0.00000521, 0.00001754, -0.00001021, 0.00000780, -0.00000805, -0.00000447, 0.00000319, -0.00001390, 0.00000434, -0.00000504, -0.00000095, -0.00001023, -0.00000667, 0.00000216, -0.00001173, 0.00000705, -0.00000864, -0.00002432, 0.00000853, 0.00000455, 0.00004322, -0.00000185, 0.00003993, 0.00000604, 0.00000737, -0.00002112, -0.00000911, 0.00000202, -0.00000239, -0.00000581, -0.00000138, -0.00000026, 0.00000235, 0.00000289, 0.00000077, -0.00000467, 0.00000258, -0.00000967, 0.00000322, 0.00000701, -0.00000254, -0.00000295, 0.00000078, -0.00000646, 0.00000549, 0.00000433, -0.00000032, 0.00000280, -0.00000163, -0.00000895, 0.00000280, 0.00000274, 0.00000321, -0.00001685, 0.00000362, 0.00000361, 0.00000741, -0.00000620, 0.00000232, 0.00000163, -0.00000077, -0.00000759, 0.00000096, 0.00000411, 0.00000095, -0.00000659, -0.00000110, -0.00000017, 0.00000296, -0.00000316, -0.00000248, -0.00000092, 0.00000755, -0.00002852, 0.00000457, -0.00002841, 0.00000194, -0.00015759, 0.00007645, 0.00025445, 0.00014573, -0.00012586, -0.00003897, -0.00001996, -0.00002081, 0.00000621, -0.00000633, -0.00000716, -0.00001670, 0.00001500, -0.00001264, 0.00000000, -0.00001234, 0.00000973, -0.00000350, -0.00001377, -0.00000831, 0.00001369, -0.00000335, -0.00000382, -0.00000147, 0.00001337, -0.00000718, -0.00001018, -0.00000088, 0.00001272, -0.00000195, 0.00000061, -0.00000465, 0.00000384, -0.00000282, 0.00000036, -0.00000249, -0.00000032, -0.00000012, -0.00000132, -0.00000337, 0.00001035, -0.00000261, -0.00000867, 0.00000025, -0.00000166, 0.00000356, -0.00000087, -0.00000139, },
		{ 0.00000057, -0.00000183, -0.00000506, 0.00000586, 0.00000936, -0.00001028, 0.00001120, -0.00001772, 0.00000868, -0.00001069, -0.00000336, -0.00000036, -0.00000237, -0.00000500, -0.00001065, -0.00000098, -0.00000293, -0.00000900, 0.00001289, -0.00003115, 0.00000176, -0.00000245, 0.00003089, 0.00001892, 0.00001258, 0.00003257, -0.00000173, 0.00000344, -0.00002632, 0.00000326, -0.00000227, -0.00000258, -0.00000580, 0.00000074, -0.00000009, 0.00000339, 0.00000263, -0.00000289, -0.00000080, -0.00000203, -0.00000780, 0.00000870, 0.00000275, -0.00000464, 0.00000049, -0.00000336, -0.00000249, 0.00000707, 0.00000192, -0.00000052, 0.00000448, -0.00000878, -0.00000210, 0.00000110, 0.00000739, -0.00000883, -0.00000944, 0.00000537, 0.00000636, 0.00000102, -0.00000432, 0.00000334, 0.00000196, -0.00000583, -0.00000362, 0.00000159, 0.00000642, -0.00000683, -0.00000081, -0.00000498, 0.00000729, -0.00000643, 0.00000477, -0.00001301, 0.00001867, -0.00002362, -0.00000249, -0.00002223, 0.00000432, -0.00007452, -0.00011381, 0.00020587, 0.00025364, -0.00000089, -0.00013147, 0.00001128, -0.00005094, 0.00001530, -0.00001573, 0.00000925, -0.00002962, 0.00001255, -0.00000546, -0.00000143, -0.00001151, 0.00000137, 0.00000354, -0.00000550, -0.00001967, 0.00000868, 0.00000463, -0.00000233, -0.00000768, 0.00000986, 0.00000428, -0.00000940, -0.00001024, 0.00001023, 0.00000498, -0.00000048, -0.00000320, 0.00000004, 0.00000114, -0.00000187, -0.00000056, -0.00000161, -0.00000078, 0.00000081, -0.00000469, 0.00000356, 0.00000791, -0.00000876, -0.00000462, 0.00000072, -0.00000128, 0.00000497, -0.00000570, 0.00000664, },
		{ -0.00000476, 0.00000067, -0.00000696, 0.00001651, -0.00000649, 0.00000273, -0.00000093, -0.00001091, 0.00000774, -0.00001539, 0.00000383, -0.00000492, 0.00000013, -0.00001053, -0.00000648, 0.00000188, -0.00001166, 0.00000726, -0.00000717, -0.00002415, 0.00001000, 0.00000086, 0.00004175, -0.00000226, 0.00003518, 0.00000821, 0.00000638, -0.00001284, -0.00001556, 0.00000514, -0.00000383, -0.00000434, -0.00000274, 0.00000027, 0.00000190, 0.00000298, 0.00000141, -0.00000465, 0.00000231, -0.00000842, 0.00000002, 0.00000855, -0.00000153, -0.00000368, 0.00000139, -0.00000625, 0.00000321, 0.00000567, -0.00000038, 0.00000223, 0.00000039, -0.00000974, 0.00000201, 0.00000209, 0.00000551, -0.00001645, 0.00000065, 0.00000469, 0.00000722, -0.00000518, 0.00000027, 0.00000329, -0.00000045, -0.00000748, -0.00000058, 0.00000452, 0.00000163, -0.00000694, -0.00000078, -0.00000125, 0.00000422, -0.00000504, 0.00000017, -0.00000448, 0.00001261, -0.00003249, 0.00000758, -0.00003225, 0.00001147, -0.00015260, 0.00001706, 0.00029327, 0.00016675, -0.00013034, -0.00004602, -0.00001704, -0.00002643, 0.00000986, -0.00000942, -0.00000393, -0.00002005, 0.00001719, -0.00001320, 0.00000066, -0.00001281, 0.00000854, -0.00000149, -0.00001286, -0.00001149, 0.00001459, -0.00000204, -0.00000402, -0.00000342, 0.00001355, -0.00000432, -0.00001131, -0.00000289, 0.00001326, -0.00000095, 0.00000014, -0.00000442, 0.00000341, -0.00000206, -0.00000025, -0.00000145, -0.00000122, -0.00000015, -0.00000085, -0.00000406, 0.00000948, 0.00000045, -0.00001013, -0.00000001, -0.00000140, 0.00000237, 0.00000069, -0.00000322, 0.00000883, },
		{ -0.00000265, -0.00000308, 0.00000130, 0.00001360, -0.00001267, 0.00001204, -0.00001492, 0.00000358, -0.00000452, -0.00000820, 0.00000218, -0.00000413, -0.00000211, -0.00001310, 0.00000121, -0.00000515, -0.00000702, 0.00001195, -0.00002843, 0.00000060, -0.00000073, 0.00002441, 0.00002236, 0.00000619, 0.00003417, -0.00000410, 0.00000904, -0.00002483, -0.00000056, 0.00000061, -0.00000401, -0.00000456, -0.00000069, 0.00000068, 0.00000277, 0.00000299, -0.00000204, -0.00000177, -0.00000060, -0.00000901, 0.00000699, 0.00000497, -0.00000462, -0.00000019, -0.00000190, -0.00000420, 0.00000638, 0.00000317, -0.00000125, 0.00000493, -0.00000706, -0.00000405, 0.00000124, 0.00000670, -0.00000503, -0.00001267, 0.00000530, 0.00000606, 0.00000251, -0.00000566, 0.00000322, 0.00000303, -0.00000516, -0.00000483, 0.00000108, 0.00000689, -0.00000627, -0.00000140, -0.00000467, 0.00000645, -0.00000544, 0.00000331, -0.00001079, 0.00001529, -0.00001706, -0.00001055, -0.00001333, -0.00000696, -0.00004608, -0.00015306, 0.00018038, 0.00030749, -0.00000205, -0.00013619, 0.00000753, -0.00004967, 0.00001104, -0.00001193, 0.00000587, -0.00002601, 0.00000793, -0.00000068, -0.00000486, -0.00000827, -0.00000211, 0.00000544, -0.00000462, -0.00001957, 0.00000514, 0.00000796, -0.00000306, -0.00000742, 0.00000690, 0.00000741, -0.00000883, -0.00001159, 0.00000833, 0.00000712, -0.00000117, -0.00000255, -0.00000086, 0.00000201, -0.00000235, -0.00000003, -0.00000145, -0.00000148, 0.00000097, -0.00000418, 0.00000178, 0.00000931, -0.00000704, -0.00000676, 0.00000171, -0.00000238, 0.00000525, -0.00000521, 0.00000497, 0.00000252, },
		{ 0.00000104, -0.00000721, 0.00001356, -0.00000125, -0.00000323, 0.00000598, -0.00001576, 0.00000987, -0.00001440, 0.00000162, -0.00000376, 0.00000014, -0.00000947, -0.00000766, 0.00000302, -0.00001307, 0.00000890, -0.00000719, -0.00002258, 0.00000996, -0.00000101, 0.00003863, -0.00000065, 0.00002880, 0.00001189, 0.00000284, -0.00000347, -0.00002108, 0.00000690, -0.00000419, -0.00000364, -0.00000340, 0.00000020, 0.00000186, 0.00000284, 0.00000194, -0.00000425, 0.00000156, -0.00000661, -0.00000320, 0.00000945, -0.00000015, -0.00000425, 0.00000150, -0.00000528, 0.00000057, 0.00000677, -0.00000016, 0.00000156, 0.00000202, -0.00000973, 0.00000067, 0.00000182, 0.00000680, -0.00001455, -0.00000320, 0.00000609, 0.00000653, -0.00000341, -0.00000234, 0.00000499, -0.00000018, -0.00000690, -0.00000248, 0.00000505, 0.00000202, -0.00000691, -0.00000086, -0.00000180, 0.00000484, -0.00000609, 0.00000160, -0.00000633, 0.00001513, -0.00003293, 0.00000628, -0.00003056, 0.00001175, -0.00013263, -0.00005108, 0.00032306, 0.00020065, -0.00014047, -0.00004993, -0.00001672, -0.00002975, 0.00001108, -0.00001022, -0.00000291, -0.00002114, 0.00001710, -0.00001164, -0.00000061, -0.00001141, 0.00000590, 0.00000116, -0.00001228, -0.00001361, 0.00001377, 0.00000060, -0.00000494, -0.00000437, 0.00001230, -0.00000071, -0.00001236, -0.00000467, 0.00001297, 0.00000072, -0.00000075, -0.00000378, 0.00000251, -0.00000090, -0.00000125, -0.00000010, -0.00000210, -0.00000030, -0.00000051, -0.00000428, 0.00000792, 0.00000361, -0.00001096, -0.00000088, -0.00000074, 0.00000095, 0.00000218, -0.00000460, 0.00000911, -0.00000399, },
		{ -0.00000106, -0.00000256, 0.00001605, -0.00001259, 0.00000993, -0.00000928, -0.00000339, 0.00000225, -0.00001263, 0.00000422, -0.00000587, 0.00000093, -0.00001544, 0.00000319, -0.00000706, -0.00000553, 0.00001155, -0.00002640, 0.00000011, 0.00000036, 0.00001898, 0.00002476, 0.00000150, 0.00003399, -0.00000475, 0.00001109, -0.00001967, -0.00000646, 0.00000448, -0.00000600, -0.00000286, -0.00000238, 0.00000152, 0.00000217, 0.00000325, -0.00000123, -0.00000244, 0.00000027, -0.00000922, 0.00000438, 0.00000724, -0.00000432, -0.00000098, -0.00000064, -0.00000515, 0.00000485, 0.00000471, -0.00000189, 0.00000503, -0.00000515, -0.00000569, 0.00000115, 0.00000584, -0.00000145, -0.00001477, 0.00000411, 0.00000635, 0.00000343, -0.00000608, 0.00000206, 0.00000462, -0.00000467, -0.00000555, 0.00000007, 0.00000754, -0.00000581, -0.00000194, -0.00000426, 0.00000541, -0.00000409, 0.00000125, -0.00000787, 0.00001110, -0.00000989, -0.00001862, -0.00000485, -0.00001761, -0.00002166, -0.00017606, 0.00013173, 0.00037132, -0.00000155, -0.00014196, 0.00000363, -0.00004790, 0.00000603, -0.00000764, 0.00000211, -0.00002209, 0.00000321, 0.00000386, -0.00000785, -0.00000559, -0.00000470, 0.00000632, -0.00000334, -0.00001906, 0.00000138, 0.00001066, -0.00000297, -0.00000741, 0.00000425, 0.00000946, -0.00000721, -0.00001304, 0.00000630, 0.00000884, -0.00000131, -0.00000228, -0.00000143, 0.00000246, -0.00000239, -0.00000002, -0.00000068, -0.00000249, 0.00000112, -0.00000365, 0.00000029, 0.00000979, -0.00000444, -0.00000904, 0.00000247, -0.00000310, 0.00000501, -0.00000421, 0.00000301, 0.00000438, -0.00000342, },
		{ -0.00000614, 0.00000925, 0.00000459, -0.00000888, 0.00001134, -0.00001788, 0.00000890, -0.00001078, -0.00000224, -0.00000158, -0.00000102, -0.00000692, -0.00001026, 0.00000557, -0.00001587, 0.00001179, -0.00000844, -0.00001993, 0.00000883, -0.00000153, 0.00003435, 0.00000239, 0.00002150, 0.00001630, -0.00000186, 0.00000519, -0.00002424, 0.00000645, -0.00000299, -0.00000404, -0.00000313, -0.00000056, 0.00000225, 0.00000252, 0.00000234, -0.00000355, 0.00000049, -0.00000456, -0.00000606, 0.00000949, 0.00000160, -0.00000459, 0.00000112, -0.00000375, -0.00000210, 0.00000739, 0.00000040, 0.00000087, 0.00000315, -0.00000894, -0.00000111, 0.00000192, 0.00000706, -0.00001137, -0.00000749, 0.00000746, 0.00000557, -0.00000118, -0.00000511, 0.00000638, 0.00000019, -0.00000595, -0.00000459, 0.00000563, 0.00000213, -0.00000647, -0.00000145, -0.00000167, 0.00000456, -0.00000596, 0.00000137, -0.00000597, 0.00001461, -0.00002945, 0.00000045, -0.00002333, 0.00000349, -0.00010126, -0.00012035, 0.00033590, 0.00025159, -0.00015761, -0.00004995, -0.00001969, -0.00003007, 0.00000923, -0.00000820, -0.00000454, -0.00001966, 0.00001452, -0.00000789, -0.00000378, -0.00000830, 0.00000204, 0.00000433, -0.00001215, -0.00001443, 0.00001122, 0.00000431, -0.00000636, -0.00000447, 0.00000994, 0.00000325, -0.00001309, -0.00000621, 0.00001188, 0.00000293, -0.00000190, -0.00000285, 0.00000126, 0.00000054, -0.00000249, 0.00000134, -0.00000273, -0.00000062, -0.00000030, -0.00000406, 0.00000591, 0.00000650, -0.00001092, -0.00000234, 0.00000021, -0.00000052, 0.00000336, -0.00000535, 0.00000859, -0.00000261, 0.00000067, },
		{ -0.00000511, 0.00001624, -0.00000991, 0.00000519, -0.00000166, -0.00001095, 0.00000831, -0.00001562, 0.00000508, -0.00000700, 0.00000344, -0.00001694, 0.00000423, -0.00000788, -0.00000531, 0.00001243, -0.00002577, 0.00000094, 0.00000017, 0.00001518, 0.00002550, -0.00000085, 0.00003144, -0.00000300, 0.00000943, -0.00001151, -0.00001346, 0.00000846, -0.00000788, -0.00000123, -0.00000390, 0.00000214, 0.00000175, 0.00000339, -0.00000057, -0.00000267, 0.00000046, -0.00000839, 0.00000107, 0.00000935, -0.00000373, -0.00000170, 0.00000017, -0.00000515, 0.00000250, 0.00000644, -0.00000242, 0.00000492, -0.00000335, -0.00000672, 0.00000057, 0.00000514, 0.00000140, -0.00001530, 0.00000160, 0.00000740, 0.00000353, -0.00000534, -0.00000021, 0.00000672, -0.00000449, -0.00000558, -0.00000161, 0.00000855, -0.00000571, -0.00000212, -0.00000414, 0.00000462, -0.00000297, -0.00000060, -0.00000528, 0.00000746, -0.00000388, -0.00002453, 0.00000060, -0.00002411, -0.00000599, -0.00017890, 0.00006085, 0.00043982, 0.00000580, -0.00015208, 0.00000186, -0.00004749, 0.00000200, -0.00000447, -0.00000053, -0.00001930, -0.00000026, 0.00000693, -0.00000927, -0.00000447, -0.00000559, 0.00000563, -0.00000123, -0.00001869, -0.00000181, 0.00001192, -0.00000154, -0.00000810, 0.00000243, 0.00001005, -0.00000444, -0.00001466, 0.00000439, 0.00000986, -0.00000068, -0.00000257, -0.00000148, 0.00000229, -0.00000182, -0.00000068, 0.00000073, -0.00000374, 0.00000130, -0.00000323, -0.00000069, 0.00000924, -0.00000112, -0.00001119, 0.00000284, -0.00000340, 0.00000431, -0.00000290, 0.00000108, 0.00000583, -0.00000378, 0.00000048, },
		{ 0.00000437, 0.00000996, -0.00001303, 0.00001405, -0.00001659, 0.00000472, -0.00000486, -0.00000724, 0.00000129, -0.00000317, -0.00000315, -0.00001391, 0.00000910, -0.00001956, 0.00001536, -0.00001027, -0.00001689, 0.00000736, -0.00000151, 0.00002969, 0.00000597, 0.00001430, 0.00002027, -0.00000612, 0.00001144, -0.00002398, 0.00000336, -0.00000010, -0.00000554, -0.00000197, -0.00000194, 0.00000298, 0.00000208, 0.00000260, -0.00000266, -0.00000067, -0.00000260, -0.00000819, 0.00000853, 0.00000366, -0.00000467, 0.00000036, -0.00000193, -0.00000445, 0.00000730, 0.00000133, 0.00000021, 0.00000380, -0.00000753, -0.00000313, 0.00000225, 0.00000649, -0.00000740, -0.00001166, 0.00000837, 0.00000468, 0.00000111, -0.00000756, 0.00000710, 0.00000084, -0.00000481, -0.00000669, 0.00000608, 0.00000211, -0.00000573, -0.00000244, -0.00000089, 0.00000341, -0.00000465, -0.00000050, -0.00000347, 0.00001131, -0.00002263, -0.00000892, -0.00001200, -0.00001064, -0.00006414, -0.00018194, 0.00032436, 0.00032206, -0.00018147, -0.00004660, -0.00002567, -0.00002754, 0.00000446, -0.00000359, -0.00000854, -0.00001595, 0.00000989, -0.00000246, -0.00000830, -0.00000410, -0.00000241, 0.00000756, -0.00001229, -0.00001409, 0.00000735, 0.00000853, -0.00000779, -0.00000409, 0.00000695, 0.00000701, -0.00001320, -0.00000760, 0.00001016, 0.00000540, -0.00000307, -0.00000183, -0.00000013, 0.00000202, -0.00000371, 0.00000256, -0.00000291, -0.00000120, -0.00000020, -0.00000352, 0.00000373, 0.00000877, -0.00000988, -0.00000432, 0.00000127, -0.00000185, 0.00000409, -0.00000540, 0.00000736, -0.00000077, -0.00000025, -0.00000416, },
		{ 0.00001415, -0.00000503, -0.00000132, 0.00000661, -0.00001763, 0.00001238, -0.00001633, 0.00000423, -0.00000706, 0.00000484, -0.00001697, 0.00000374, -0.00000705, -0.00000686, 0.00001505, -0.00002693, 0.00000342, -0.00000154, 0.00001319, 0.00002435, -0.00000071, 0.00002648, 0.00000108, 0.00000470, -0.00000171, -0.00002015, 0.00001150, -0.00000893, -0.00000019, -0.00000487, 0.00000229, 0.00000164, 0.00000341, -0.00000013, -0.00000239, -0.00000006, -0.00000666, -0.00000265, 0.00001108, -0.00000286, -0.00000219, 0.00000035, -0.00000414, -0.00000052, 0.00000817, -0.00000282, 0.00000476, -0.00000193, -0.00000692, -0.00000065, 0.00000491, 0.00000313, -0.00001404, -0.00000223, 0.00000918, 0.00000274, -0.00000341, -0.00000350, 0.00000921, -0.00000465, -0.00000478, -0.00000408, 0.00001009, -0.00000619, -0.00000166, -0.00000467, 0.00000460, -0.00000274, -0.00000139, -0.00000414, 0.00000577, -0.00000073, -0.00002634, 0.00000082, -0.00002375, -0.00000219, -0.00016056, -0.00002784, 0.00050500, 0.00002643, -0.00017029, 0.00000475, -0.00005051, 0.00000079, -0.00000414, -0.00000051, -0.00001905, -0.00000119, 0.00000737, -0.00000817, -0.00000574, -0.00000415, 0.00000294, 0.00000206, -0.00001900, -0.00000367, 0.00001116, 0.00000156, -0.00000972, 0.00000174, 0.00000901, -0.00000059, -0.00001642, 0.00000282, 0.00000991, 0.00000085, -0.00000352, -0.00000088, 0.00000139, -0.00000054, -0.00000207, 0.00000273, -0.00000515, 0.00000149, -0.00000305, -0.00000103, 0.00000772, 0.00000263, -0.00001294, 0.00000270, -0.00000328, 0.00000330, -0.00000155, -0.00000050, 0.00000659, -0.00000347, 0.00000120, -0.00000603, },
		{ 0.00001385, -0.00001478, 0.00001348, -0.00001182, -0.00000213, 0.00000247, -0.00001254, 0.00000424, -0.00000579, 0.00000129, -0.00001795, 0.00001290, -0.00002334, 0.00001874, -0.00001181, -0.00001440, 0.00000647, -0.00000185, 0.00002555, 0.00000905, 0.00000835, 0.00002257, -0.00000845, 0.00001405, -0.00001988, -0.00000221, 0.00000414, -0.00000784, -0.00000019, -0.00000370, 0.00000385, 0.00000164, 0.00000273, -0.00000173, -0.00000166, -0.00000107, -0.00000931, 0.00000651, 0.00000594, -0.00000450, -0.00000060, -0.00000017, -0.00000607, 0.00000631, 0.00000266, -0.00000042, 0.00000406, -0.00000576, -0.00000506, 0.00000255, 0.00000547, -0.00000323, -0.00001504, 0.00000834, 0.00000424, 0.00000300, -0.00000919, 0.00000678, 0.00000197, -0.00000371, -0.00000845, 0.00000610, 0.00000223, -0.00000498, -0.00000357, 0.00000023, 0.00000173, -0.00000256, -0.00000349, 0.00000041, 0.00000628, -0.00001399, -0.00001978, 0.00000070, -0.00002654, -0.00002821, -0.00022707, 0.00028294, 0.00041190, -0.00020951, -0.00004196, -0.00003316, -0.00002337, -0.00000212, 0.00000253, -0.00001385, -0.00001107, 0.00000430, 0.00000358, -0.00001312, 0.00000018, -0.00000654, 0.00001018, -0.00001226, -0.00001308, 0.00000291, 0.00001239, -0.00000857, -0.00000377, 0.00000398, 0.00000997, -0.00001239, -0.00000899, 0.00000808, 0.00000775, -0.00000393, -0.00000099, -0.00000138, 0.00000324, -0.00000458, 0.00000326, -0.00000243, -0.00000210, -0.00000013, -0.00000285, 0.00000171, 0.00001011, -0.00000777, -0.00000667, 0.00000228, -0.00000290, 0.00000431, -0.00000484, 0.00000565, 0.00000122, -0.00000134, -0.00000187, -0.00000019, },
		{ 0.00000122, -0.00000836, 0.00001401, -0.00002199, 0.00001345, -0.00001422, 0.00000146, -0.00000580, 0.00000479, -0.00001518, 0.00000144, -0.00000437, -0.00001031, 0.00001943, -0.00002981, 0.00000741, -0.00000457, 0.00001276, 0.00002156, 0.00000155, 0.00001966, 0.00000675, -0.00000175, 0.00000797, -0.00002504, 0.00001260, -0.00000852, -0.00000016, -0.00000501, 0.00000184, 0.00000185, 0.00000336, 0.00000004, -0.00000159, -0.00000122, -0.00000428, -0.00000639, 0.00001219, -0.00000171, -0.00000235, -0.00000019, -0.00000223, -0.00000392, 0.00000970, -0.00000304, 0.00000467, -0.00000108, -0.00000618, -0.00000257, 0.00000529, 0.00000353, -0.00001101, -0.00000712, 0.00001149, 0.00000116, -0.00000042, -0.00000751, 0.00001185, -0.00000512, -0.00000312, -0.00000733, 0.00001221, -0.00000737, -0.00000035, -0.00000617, 0.00000577, -0.00000393, -0.00000041, -0.00000533, 0.00000707, -0.00000160, -0.00002288, -0.00000536, -0.00001527, -0.00001106, -0.00012330, -0.00012660, 0.00055698, 0.00006709, -0.00019998, 0.00001439, -0.00005865, 0.00000395, -0.00000803, 0.00000338, -0.00002239, 0.00000131, 0.00000448, -0.00000396, -0.00000981, -0.00000013, -0.00000181, 0.00000663, -0.00002035, -0.00000376, 0.00000813, 0.00000629, -0.00001224, 0.00000224, 0.00000642, 0.00000409, -0.00001821, 0.00000171, 0.00000888, 0.00000328, -0.00000512, 0.00000038, -0.00000025, 0.00000143, -0.00000409, 0.00000512, -0.00000653, 0.00000166, -0.00000315, -0.00000070, 0.00000539, 0.00000646, -0.00001402, 0.00000200, -0.00000285, 0.00000218, -0.00000043, -0.00000148, 0.00000648, -0.00000238, 0.00000097, -0.00000534, 0.00001016, },
		{ -0.00001363, 0.00000962, -0.00000418, -0.00001054, 0.00000995, -0.00001707, 0.00000649, -0.00000824, 0.00000563, -0.00002154, 0.00001608, -0.00002631, 0.00002102, -0.00001211, -0.00001335, 0.00000703, -0.00000340, 0.00002276, 0.00001070, 0.00000463, 0.00002224, -0.00000789, 0.00001255, -0.00001229, -0.00000951, 0.00000901, -0.00001038, 0.00000177, -0.00000548, 0.00000462, 0.00000132, 0.00000277, -0.00000092, -0.00000224, -0.00000027, -0.00000923, 0.00000350, 0.00000831, -0.00000412, -0.00000152, 0.00000115, -0.00000666, 0.00000432, 0.00000436, -0.00000107, 0.00000413, -0.00000400, -0.00000652, 0.00000249, 0.00000444, 0.00000046, -0.00001698, 0.00000695, 0.00000462, 0.00000402, -0.00000953, 0.00000515, 0.00000374, -0.00000294, -0.00000948, 0.00000532, 0.00000287, -0.00000464, -0.00000439, 0.00000116, 0.00000015, -0.00000047, -0.00000661, 0.00000438, 0.00000121, -0.00000573, -0.00002938, 0.00001130, -0.00003939, -0.00000041, -0.00024852, 0.00020925, 0.00051753, -0.00023670, -0.00003979, -0.00003942, -0.00001974, -0.00000852, 0.00000829, -0.00001871, -0.00000669, -0.00000067, 0.00000872, -0.00001683, 0.00000324, -0.00000921, 0.00001133, -0.00001141, -0.00001214, -0.00000112, 0.00001494, -0.00000801, -0.00000409, 0.00000166, 0.00001160, -0.00001040, -0.00001055, 0.00000598, 0.00000955, -0.00000412, -0.00000066, -0.00000215, 0.00000385, -0.00000478, 0.00000314, -0.00000114, -0.00000336, -0.00000001, -0.00000225, 0.00000018, 0.00001030, -0.00000469, -0.00000916, 0.00000306, -0.00000359, 0.00000404, -0.00000387, 0.00000378, 0.00000299, -0.00000217, 0.00000012, -0.00000364, 0.00001554, },
		{ -0.00001459, 0.00001915, -0.00002295, 0.00001100, -0.00000932, -0.00000301, -0.00000335, 0.00000328, -0.00001163, -0.00000250, -0.00000006, -0.00001534, 0.00002516, -0.00003391, 0.00001231, -0.00000826, 0.00001321, 0.00001781, 0.00000511, 0.00001203, 0.00001274, -0.00000820, 0.00001568, -0.00002683, 0.00001111, -0.00000635, -0.00000126, -0.00000427, 0.00000078, 0.00000234, 0.00000331, -0.00000010, -0.00000035, -0.00000280, -0.00000160, -0.00000974, 0.00001244, -0.00000031, -0.00000212, -0.00000137, 0.00000033, -0.00000734, 0.00001075, -0.00000304, 0.00000472, -0.00000086, -0.00000456, -0.00000508, 0.00000627, 0.00000266, -0.00000655, -0.00001259, 0.00001394, -0.00000094, 0.00000325, -0.00001179, 0.00001430, -0.00000578, -0.00000068, -0.00001122, 0.00001477, -0.00000920, 0.00000180, -0.00000871, 0.00000827, -0.00000678, 0.00000265, -0.00000920, 0.00001174, -0.00000679, -0.00001406, -0.00001772, 0.00000071, -0.00003083, -0.00007255, -0.00022510, 0.00058514, 0.00013364, -0.00024322, 0.00003174, -0.00007268, 0.00001218, -0.00001673, 0.00001158, -0.00002963, 0.00000744, -0.00000182, 0.00000326, -0.00001649, 0.00000616, -0.00000828, 0.00001224, -0.00002273, -0.00000202, 0.00000304, 0.00001221, -0.00001534, 0.00000370, 0.00000264, 0.00000917, -0.00001981, 0.00000106, 0.00000681, 0.00000648, -0.00000720, 0.00000219, -0.00000249, 0.00000390, -0.00000651, 0.00000763, -0.00000766, 0.00000174, -0.00000356, 0.00000023, 0.00000252, 0.00000997, -0.00001420, 0.00000076, -0.00000225, 0.00000115, 0.00000025, -0.00000170, 0.00000549, -0.00000057, -0.00000009, -0.00000335, 0.00000500, 0.00000776, },
		{ 0.00000306, 0.00000513, -0.00001900, 0.00001617, -0.00001977, 0.00000732, -0.00000984, 0.00000910, -0.00002383, 0.00001784, -0.00002765, 0.00002141, -0.00001046, -0.00001440, 0.00000961, -0.00000666, 0.00002178, 0.00001037, 0.00000367, 0.00001888, -0.00000419, 0.00000731, -0.00000230, -0.00001728, 0.00001349, -0.00001242, 0.00000338, -0.00000691, 0.00000506, 0.00000118, 0.00000279, -0.00000042, -0.00000219, -0.00000038, -0.00000792, -0.00000032, 0.00001059, -0.00000358, -0.00000216, 0.00000171, -0.00000600, 0.00000134, 0.00000636, -0.00000178, 0.00000423, -0.00000260, -0.00000716, 0.00000173, 0.00000388, 0.00000303, -0.00001698, 0.00000391, 0.00000605, 0.00000384, -0.00000827, 0.00000208, 0.00000624, -0.00000276, -0.00000940, 0.00000336, 0.00000441, -0.00000515, -0.00000434, 0.00000123, -0.00000050, 0.00000062, -0.00000858, 0.00000686, -0.00000192, -0.00000030, -0.00003480, 0.00001632, -0.00004469, 0.00001347, -0.00024186, 0.00010497, 0.00063170, -0.00025569, -0.00004512, -0.00004089, -0.00001957, -0.00001217, 0.00001132, -0.00002096, -0.00000480, -0.00000317, 0.00001128, -0.00001795, 0.00000377, -0.00000935, 0.00001020, -0.00000905, -0.00001212, -0.00000371, 0.00001527, -0.00000553, -0.00000553, 0.00000055, 0.00001147, -0.00000708, -0.00001243, 0.00000423, 0.00001038, -0.00000331, -0.00000110, -0.00000214, 0.00000356, -0.00000402, 0.00000199, 0.00000102, -0.00000496, 0.00000023, -0.00000195, -0.00000059, 0.00000925, -0.00000080, -0.00001154, 0.00000350, -0.00000393, 0.00000344, -0.00000275, 0.00000213, 0.00000415, -0.00000236, 0.00000129, -0.00000543, 0.00001469, -0.00000966, },
		{ 0.00002098, -0.00002004, 0.00000521, -0.00000222, -0.00000852, -0.00000020, 0.00000069, -0.00000676, -0.00000753, 0.00000523, -0.00002121, 0.00003140, -0.00003832, 0.00001716, -0.00001165, 0.00001360, 0.00001403, 0.00000885, 0.00000492, 0.00001755, -0.00001291, 0.00001991, -0.00002477, 0.00000686, -0.00000254, -0.00000337, -0.00000280, -0.00000071, 0.00000296, 0.00000334, -0.00000050, 0.00000115, -0.00000450, 0.00000095, -0.00001229, 0.00001168, 0.00000130, -0.00000152, -0.00000301, 0.00000314, -0.00001031, 0.00001107, -0.00000274, 0.00000490, -0.00000121, -0.00000227, -0.00000791, 0.00000763, 0.00000083, -0.00000121, -0.00001794, 0.00001600, -0.00000310, 0.00000709, -0.00001575, 0.00001613, -0.00000642, 0.00000228, -0.00001540, 0.00001748, -0.00001141, 0.00000455, -0.00001209, 0.00001191, -0.00001108, 0.00000757, -0.00001546, 0.00001929, -0.00001556, -0.00000108, -0.00003452, 0.00002173, -0.00005740, -0.00001618, -0.00031178, 0.00057977, 0.00022969, -0.00029978, 0.00005600, -0.00009194, 0.00002498, -0.00002971, 0.00002349, -0.00004012, 0.00001649, -0.00001074, 0.00001268, -0.00002494, 0.00001385, -0.00001563, 0.00001826, -0.00002575, 0.00000115, -0.00000345, 0.00001854, -0.00001839, 0.00000565, -0.00000176, 0.00001408, -0.00002093, 0.00000079, 0.00000389, 0.00001010, -0.00000948, 0.00000428, -0.00000506, 0.00000655, -0.00000896, 0.00000986, -0.00000829, 0.00000166, -0.00000420, 0.00000159, -0.00000051, 0.00001278, -0.00001335, -0.00000092, -0.00000163, 0.00000038, 0.00000038, -0.00000119, 0.00000378, 0.00000170, -0.00000165, -0.00000064, -0.00000039, 0.00001340, -0.00001876, },
		{ 0.00001456, -0.00002589, 0.00001981, -0.00001980, 0.00000622, -0.00001011, 0.00001110, -0.00002422, 0.00001762, -0.00002686, 0.00001947, -0.00000652, -0.00001779, 0.00001434, -0.00001168, 0.00002267, 0.00000797, 0.00000547, 0.00001274, 0.00000212, -0.00000049, 0.00000840, -0.00002404, 0.00001654, -0.00001330, 0.00000421, -0.00000770, 0.00000504, 0.00000124, 0.00000289, -0.00000037, -0.00000139, -0.00000145, -0.00000550, -0.00000466, 0.00001259, -0.00000296, -0.00000230, 0.00000131, -0.00000407, -0.00000246, 0.00000852, -0.00000258, 0.00000458, -0.00000190, -0.00000670, 0.00000002, 0.00000416, 0.00000402, -0.00001476, -0.00000079, 0.00000859, 0.00000230, -0.00000531, -0.00000237, 0.00000945, -0.00000340, -0.00000788, -0.00000006, 0.00000718, -0.00000693, -0.00000293, -0.00000020, 0.00000059, -0.00000030, -0.00000814, 0.00000631, -0.00000125, 0.00000014, -0.00003361, 0.00001303, -0.00003917, 0.00000992, -0.00020642, -0.00002381, 0.00074382, -0.00025753, -0.00006356, -0.00003374, -0.00002589, -0.00001039, 0.00000924, -0.00001851, -0.00000726, -0.00000155, 0.00000983, -0.00001524, 0.00000073, -0.00000615, 0.00000619, -0.00000463, -0.00001372, -0.00000404, 0.00001277, -0.00000084, -0.00000834, 0.00000100, 0.00000936, -0.00000242, -0.00001472, 0.00000316, 0.00000988, -0.00000124, -0.00000255, -0.00000112, 0.00000214, -0.00000216, -0.00000024, 0.00000398, -0.00000682, 0.00000066, -0.00000214, -0.00000043, 0.00000698, 0.00000361, -0.00001356, 0.00000353, -0.00000401, 0.00000273, -0.00000183, 0.00000106, 0.00000441, -0.00000164, 0.00000136, -0.00000535, 0.00001143, -0.00000347, -0.00000506, },
		{ -0.00001347, -0.00000310, 0.00000600, -0.00001405, 0.00000290, -0.00000233, -0.00000132, -0.00001277, 0.00001059, -0.00002693, 0.00003711, -0.00004197, 0.00002090, -0.00001371, 0.00001293, 0.00001123, 0.00001160, -0.00000038, 0.00001987, -0.00001454, 0.00001980, -0.00001883, 0.00000032, 0.00000239, -0.00000601, -0.00000094, -0.00000236, 0.00000351, 0.00000352, -0.00000111, 0.00000269, -0.00000599, 0.00000297, -0.00001369, 0.00000984, 0.00000309, -0.00000064, -0.00000482, 0.00000576, -0.00001241, 0.00001044, -0.00000208, 0.00000513, -0.00000191, 0.00000032, -0.00001065, 0.00000900, -0.00000148, 0.00000424, -0.00002241, 0.00001709, -0.00000483, 0.00001047, -0.00001877, 0.00001692, -0.00000674, 0.00000538, -0.00001936, 0.00001981, -0.00001353, 0.00000743, -0.00001580, 0.00001612, -0.00001616, 0.00001353, -0.00002308, 0.00002840, -0.00002615, 0.00001371, -0.00005270, 0.00004381, -0.00008497, 0.00003665, -0.00037543, 0.00053364, 0.00035539, -0.00036651, 0.00008414, -0.00011410, 0.00004039, -0.00004520, 0.00003739, -0.00005222, 0.00002686, -0.00002072, 0.00002279, -0.00003371, 0.00002158, -0.00002263, 0.00002372, -0.00002867, 0.00000496, -0.00001028, 0.00002422, -0.00002061, 0.00000746, -0.00000607, 0.00001818, -0.00002122, 0.00000067, 0.00000050, 0.00001368, -0.00001152, 0.00000627, -0.00000754, 0.00000895, -0.00001101, 0.00001142, -0.00000819, 0.00000132, -0.00000496, 0.00000311, -0.00000330, 0.00001454, -0.00001144, -0.00000288, -0.00000116, -0.00000005, -0.00000004, -0.00000009, 0.00000166, 0.00000404, -0.00000325, 0.00000210, -0.00000508, 0.00001714, -0.00002003, 0.00001907, },
		{ -0.00002979, 0.00001999, -0.00001681, 0.00000309, -0.00000890, 0.00001136, -0.00002248, 0.00001530, -0.00002389, 0.00001526, -0.00000045, -0.00002322, 0.00002082, -0.00001801, 0.00002495, 0.00000396, 0.00000939, 0.00000473, 0.00000979, -0.00000911, 0.00001794, -0.00002834, 0.00001729, -0.00001254, 0.00000400, -0.00000772, 0.00000453, 0.00000142, 0.00000318, -0.00000087, 0.00000018, -0.00000337, -0.00000225, -0.00000911, 0.00001408, -0.00000229, -0.00000180, -0.00000011, -0.00000099, -0.00000678, 0.00001065, -0.00000348, 0.00000534, -0.00000210, -0.00000500, -0.00000272, 0.00000549, 0.00000322, -0.00001035, -0.00000693, 0.00001208, -0.00000054, -0.00000080, -0.00000796, 0.00001323, -0.00000495, -0.00000475, -0.00000506, 0.00001134, -0.00001023, 0.00000022, -0.00000363, 0.00000405, -0.00000401, -0.00000437, 0.00000162, 0.00000446, -0.00000578, -0.00002447, 0.00000015, -0.00002157, -0.00001153, -0.00014565, -0.00016695, 0.00084098, -0.00023287, -0.00010022, -0.00001477, -0.00004121, -0.00000105, 0.00000022, -0.00000981, -0.00001539, 0.00000529, 0.00000347, -0.00000803, -0.00000638, 0.00000071, -0.00000087, 0.00000204, -0.00001736, -0.00000165, 0.00000727, 0.00000598, -0.00001250, 0.00000306, 0.00000535, 0.00000339, -0.00001741, 0.00000300, 0.00000785, 0.00000215, -0.00000504, 0.00000099, -0.00000046, 0.00000080, -0.00000346, 0.00000753, -0.00000880, 0.00000130, -0.00000291, 0.00000071, 0.00000370, 0.00000817, -0.00001499, 0.00000316, -0.00000399, 0.00000214, -0.00000137, 0.00000081, 0.00000361, 0.00000007, 0.00000028, -0.00000355, 0.00000632, 0.00000391, -0.00001332, 0.00002245, },
		{ -0.00001254, 0.00001389, -0.00001846, 0.00000509, -0.00000500, 0.00000378, -0.00001725, 0.00001503, -0.00003148, 0.00004127, -0.00004387, 0.00002258, -0.00001359, 0.00001042, 0.00001020, 0.00001245, -0.00000289, 0.00001882, -0.00001242, 0.00001536, -0.00000977, -0.00000751, 0.00000756, -0.00000856, 0.00000083, -0.00000383, 0.00000383, 0.00000388, -0.00000176, 0.00000398, -0.00000692, 0.00000408, -0.00001373, 0.00000693, 0.00000497, 0.00000039, -0.00000646, 0.00000773, -0.00001326, 0.00000873, -0.00000102, 0.00000529, -0.00000269, 0.00000276, -0.00001284, 0.00000992, -0.00000365, 0.00000898, -0.00002522, 0.00001667, -0.00000562, 0.00001280, -0.00002025, 0.00001629, -0.00000643, 0.00000811, -0.00002245, 0.00002114, -0.00001491, 0.00000973, -0.00001906, 0.00001999, -0.00002093, 0.00001923, -0.00003045, 0.00003708, -0.00003605, 0.00002719, -0.00006843, 0.00006220, -0.00010711, 0.00007699, -0.00040691, 0.00044357, 0.00050656, -0.00043698, 0.00011100, -0.00013520, 0.00005517, -0.00006026, 0.00005058, -0.00006343, 0.00003623, -0.00002959, 0.00003155, -0.00004094, 0.00002766, -0.00002784, 0.00002747, -0.00003050, 0.00000831, -0.00001620, 0.00002811, -0.00002120, 0.00000843, -0.00000955, 0.00002083, -0.00002034, 0.00000043, -0.00000288, 0.00001665, -0.00001283, 0.00000772, -0.00000946, 0.00001061, -0.00001220, 0.00001194, -0.00000717, 0.00000064, -0.00000568, 0.00000447, -0.00000544, 0.00001499, -0.00000851, -0.00000493, -0.00000093, -0.00000013, -0.00000088, 0.00000132, -0.00000047, 0.00000600, -0.00000437, 0.00000421, -0.00000829, 0.00001831, -0.00001806, 0.00001365, -0.00000419, },
		{ 0.00001643, -0.00001100, -0.00000169, -0.00000647, 0.00001004, -0.00001884, 0.00001123, -0.00001917, 0.00000933, 0.00000707, -0.00002994, 0.00002820, -0.00002475, 0.00002775, -0.00000079, 0.00001438, -0.00000381, 0.00001723, -0.00001665, 0.00002454, -0.00002911, 0.00001531, -0.00001005, 0.00000277, -0.00000703, 0.00000365, 0.00000160, 0.00000377, -0.00000194, 0.00000240, -0.00000590, 0.00000142, -0.00001323, 0.00001486, -0.00000163, -0.00000063, -0.00000245, 0.00000291, -0.00001122, 0.00001250, -0.00000446, 0.00000657, -0.00000325, -0.00000212, -0.00000640, 0.00000784, 0.00000069, -0.00000412, -0.00001399, 0.00001613, -0.00000440, 0.00000489, -0.00001421, 0.00001729, -0.00000738, -0.00000004, -0.00001152, 0.00001678, -0.00001503, 0.00000514, -0.00000920, 0.00001009, -0.00001078, 0.00000309, -0.00000756, 0.00001552, -0.00001822, -0.00000752, -0.00002177, 0.00000699, -0.00004818, -0.00006683, -0.00031125, 0.00090944, -0.00017347, -0.00015854, 0.00001771, -0.00006671, 0.00001685, -0.00001651, 0.00000566, -0.00002952, 0.00001751, -0.00000779, 0.00000357, -0.00001730, 0.00001088, -0.00001058, 0.00001071, -0.00002299, 0.00000343, -0.00000090, 0.00001441, -0.00001763, 0.00000653, -0.00000023, 0.00000996, -0.00002036, 0.00000383, 0.00000427, 0.00000675, -0.00000847, 0.00000411, -0.00000409, 0.00000465, -0.00000739, 0.00001136, -0.00001070, 0.00000213, -0.00000431, 0.00000274, -0.00000030, 0.00001249, -0.00001566, 0.00000248, -0.00000403, 0.00000189, -0.00000160, 0.00000151, 0.00000179, 0.00000264, -0.00000172, -0.00000050, 0.00000023, 0.00001125, -0.00002026, 0.00002742, -0.00002977, },
		{ 0.00001995, -0.00002064, 0.00000563, -0.00000660, 0.00000772, -0.00002012, 0.00001772, -0.00003406, 0.00004313, -0.00004333, 0.00002160, -0.00001080, 0.00000569, 0.00001131, 0.00001093, -0.00000221, 0.00001424, -0.00000675, 0.00000744, 0.00000099, -0.00001525, 0.00001192, -0.00001032, 0.00000209, -0.00000486, 0.00000379, 0.00000437, -0.00000228, 0.00000477, -0.00000700, 0.00000405, -0.00001233, 0.00000313, 0.00000684, 0.00000143, -0.00000759, 0.00000865, -0.00001263, 0.00000591, 0.00000045, 0.00000522, -0.00000319, 0.00000458, -0.00001400, 0.00000993, -0.00000505, 0.00001224, -0.00002573, 0.00001434, -0.00000507, 0.00001356, -0.00001976, 0.00001395, -0.00000521, 0.00000994, -0.00002398, 0.00002080, -0.00001486, 0.00001067, -0.00002096, 0.00002241, -0.00002409, 0.00002311, -0.00003567, 0.00004303, -0.00004251, 0.00003603, -0.00007781, 0.00007226, -0.00011795, 0.00009752, -0.00040069, 0.00031134, 0.00067450, -0.00050179, 0.00012966, -0.00015016, 0.00006516, -0.00007126, 0.00005980, -0.00007083, 0.00004195, -0.00003496, 0.00003683, -0.00004473, 0.00003041, -0.00002985, 0.00002836, -0.00003025, 0.00001009, -0.00001997, 0.00002918, -0.00001948, 0.00000800, -0.00001154, 0.00002152, -0.00001802, -0.00000026, -0.00000571, 0.00001841, -0.00001293, 0.00000817, -0.00001035, 0.00001108, -0.00001215, 0.00001115, -0.00000514, -0.00000043, -0.00000617, 0.00000535, -0.00000659, 0.00001401, -0.00000474, -0.00000688, -0.00000100, 0.00000004, -0.00000190, 0.00000265, -0.00000216, 0.00000714, -0.00000459, 0.00000521, -0.00000955, 0.00001673, -0.00001314, 0.00000562, 0.00000634, -0.00002143, },
		{ -0.00000321, -0.00000728, -0.00000344, 0.00000769, -0.00001394, 0.00000617, -0.00001353, 0.00000261, 0.00001504, -0.00003686, 0.00003533, -0.00003080, 0.00003001, -0.00000522, 0.00001915, -0.00001134, 0.00002277, -0.00002139, 0.00002698, -0.00002594, 0.00001071, -0.00000614, 0.00000082, -0.00000591, 0.00000263, 0.00000159, 0.00000474, -0.00000351, 0.00000504, -0.00000868, 0.00000504, -0.00001659, 0.00001477, -0.00000098, 0.00000115, -0.00000546, 0.00000719, -0.00001531, 0.00001381, -0.00000543, 0.00000824, -0.00000524, 0.00000168, -0.00001072, 0.00001095, -0.00000317, 0.00000329, -0.00002123, 0.00002017, -0.00000877, 0.00001116, -0.00002050, 0.00002123, -0.00001047, 0.00000595, -0.00001902, 0.00002309, -0.00002096, 0.00001154, -0.00001662, 0.00001840, -0.00002025, 0.00001379, -0.00002065, 0.00003110, -0.00003597, 0.00001546, -0.00005025, 0.00004306, -0.00009444, 0.00001988, -0.00044208, 0.00093653, -0.00007383, -0.00023912, 0.00006308, -0.00010173, 0.00004266, -0.00004022, 0.00002710, -0.00004876, 0.00003417, -0.00002297, 0.00001851, -0.00003100, 0.00002331, -0.00002199, 0.00002064, -0.00003007, 0.00001059, -0.00001089, 0.00002355, -0.00002308, 0.00001089, -0.00000682, 0.00001676, -0.00002332, 0.00000558, -0.00000064, 0.00001220, -0.00001251, 0.00000793, -0.00000843, 0.00000901, -0.00001160, 0.00001506, -0.00001228, 0.00000310, -0.00000626, 0.00000545, -0.00000461, 0.00001618, -0.00001545, 0.00000159, -0.00000431, 0.00000213, -0.00000257, 0.00000308, -0.00000083, 0.00000571, -0.00000420, 0.00000311, -0.00000581, 0.00001734, -0.00002466, 0.00002858, -0.00002564, 0.00000957, },
		{ -0.00001989, 0.00000409, -0.00000671, 0.00000996, -0.00002083, 0.00001822, -0.00003427, 0.00004236, -0.00004016, 0.00001789, -0.00000543, -0.00000110, 0.00001441, 0.00000715, 0.00000142, 0.00000669, 0.00000149, -0.00000237, 0.00001162, -0.00002140, 0.00001450, -0.00001074, 0.00000251, -0.00000529, 0.00000338, 0.00000486, -0.00000245, 0.00000481, -0.00000605, 0.00000279, -0.00000961, -0.00000131, 0.00000856, 0.00000231, -0.00000791, 0.00000827, -0.00001044, 0.00000210, 0.00000228, 0.00000476, -0.00000310, 0.00000536, -0.00001372, 0.00000866, -0.00000518, 0.00001343, -0.00002359, 0.00000995, -0.00000298, 0.00001245, -0.00001704, 0.00000982, -0.00000286, 0.00001038, -0.00002336, 0.00001821, -0.00001276, 0.00000953, -0.00002060, 0.00002232, -0.00002434, 0.00002364, -0.00003692, 0.00004408, -0.00004301, 0.00003736, -0.00007766, 0.00007043, -0.00011326, 0.00009385, -0.00035585, 0.00014411, 0.00084637, -0.00054949, 0.00013241, -0.00015356, 0.00006603, -0.00007447, 0.00006179, -0.00007158, 0.00004156, -0.00003467, 0.00003674, -0.00004347, 0.00002849, -0.00002756, 0.00002553, -0.00002708, 0.00000931, -0.00002057, 0.00002672, -0.00001507, 0.00000577, -0.00001157, 0.00001990, -0.00001408, -0.00000167, -0.00000746, 0.00001845, -0.00001141, 0.00000728, -0.00000985, 0.00001004, -0.00001058, 0.00000892, -0.00000211, -0.00000190, -0.00000626, 0.00000547, -0.00000652, 0.00001160, -0.00000036, -0.00000854, -0.00000133, 0.00000028, -0.00000280, 0.00000354, -0.00000303, 0.00000715, -0.00000365, 0.00000484, -0.00000876, 0.00001269, -0.00000606, -0.00000368, 0.00001638, -0.00002811, 0.00002297, },
		{ -0.00001254, -0.00000069, 0.00000509, -0.00000869, 0.00000113, -0.00000802, -0.00000379, 0.00002229, -0.00004279, 0.00004104, -0.00003504, 0.00003071, -0.00000827, 0.00002247, -0.00001652, 0.00002511, -0.00002225, 0.00002480, -0.00001915, 0.00000420, -0.00000151, -0.00000129, -0.00000473, 0.00000175, 0.00000125, 0.00000609, -0.00000541, 0.00000780, -0.00001128, 0.00000811, -0.00001882, 0.00001372, -0.00000037, 0.00000339, -0.00000875, 0.00001131, -0.00001858, 0.00001436, -0.00000629, 0.00001021, -0.00000779, 0.00000597, -0.00001520, 0.00001436, -0.00000777, 0.00001099, -0.00002777, 0.00002353, -0.00001305, 0.00001727, -0.00002613, 0.00002454, -0.00001385, 0.00001267, -0.00002684, 0.00002958, -0.00002737, 0.00001872, -0.00002514, 0.00002813, -0.00003142, 0.00002652, -0.00003612, 0.00004928, -0.00005657, 0.00004130, -0.00008124, 0.00008139, -0.00014278, 0.00010288, -0.00054532, 0.00091262, 0.00006742, -0.00033891, 0.00011797, -0.00014341, 0.00007391, -0.00006862, 0.00005226, -0.00007100, 0.00005324, -0.00004012, 0.00003494, -0.00004572, 0.00003639, -0.00003366, 0.00003063, -0.00003762, 0.00001880, -0.00002144, 0.00003222, -0.00002801, 0.00001547, -0.00001365, 0.00002311, -0.00002591, 0.00000796, -0.00000641, 0.00001796, -0.00001664, 0.00001198, -0.00001297, 0.00001334, -0.00001555, 0.00001817, -0.00001329, 0.00000408, -0.00000859, 0.00000849, -0.00000875, 0.00001888, -0.00001434, 0.00000066, -0.00000495, 0.00000291, -0.00000422, 0.00000529, -0.00000386, 0.00000883, -0.00000662, 0.00000655, -0.00001086, 0.00002122, -0.00002580, 0.00002574, -0.00001794, 0.00000055, 0.00000526, },
		{ 0.00000053, -0.00000528, 0.00001037, -0.00001933, 0.00001657, -0.00003224, 0.00003921, -0.00003471, 0.00001194, 0.00000192, -0.00000929, 0.00001886, 0.00000173, 0.00000717, -0.00000263, 0.00001074, -0.00001214, 0.00002024, -0.00002469, 0.00001465, -0.00000955, 0.00000203, -0.00000515, 0.00000274, 0.00000517, -0.00000207, 0.00000396, -0.00000402, 0.00000040, -0.00000584, -0.00000601, 0.00000998, 0.00000291, -0.00000722, 0.00000650, -0.00000683, -0.00000246, 0.00000437, 0.00000381, -0.00000216, 0.00000480, -0.00001181, 0.00000589, -0.00000373, 0.00001224, -0.00001874, 0.00000365, 0.00000059, 0.00000939, -0.00001214, 0.00000402, 0.00000067, 0.00000906, -0.00002015, 0.00001303, -0.00000824, 0.00000579, -0.00001734, 0.00001887, -0.00002069, 0.00001964, -0.00003283, 0.00003872, -0.00003587, 0.00002940, -0.00006620, 0.00005497, -0.00009137, 0.00006531, -0.00027643, -0.00004608, 0.00100644, -0.00056794, 0.00011200, -0.00014063, 0.00005415, -0.00006686, 0.00005402, -0.00006358, 0.00003328, -0.00002724, 0.00003010, -0.00003621, 0.00002118, -0.00002047, 0.00001859, -0.00002059, 0.00000542, -0.00001746, 0.00002051, -0.00000799, 0.00000166, -0.00000944, 0.00001591, -0.00000857, -0.00000398, -0.00000772, 0.00001638, -0.00000806, 0.00000483, -0.00000775, 0.00000732, -0.00000746, 0.00000533, 0.00000177, -0.00000371, -0.00000582, 0.00000463, -0.00000511, 0.00000793, 0.00000430, -0.00000978, -0.00000180, 0.00000034, -0.00000327, 0.00000365, -0.00000280, 0.00000588, -0.00000152, 0.00000315, -0.00000619, 0.00000688, 0.00000198, -0.00001268, 0.00002425, -0.00003127, 0.00002222, -0.00001601, },
		{ 0.00000086, 0.00000311, -0.00000406, -0.00000285, -0.00000368, -0.00000883, 0.00002777, -0.00004673, 0.00004439, -0.00003665, 0.00002911, -0.00000922, 0.00002349, -0.00001845, 0.00002356, -0.00001891, 0.00001844, -0.00000984, -0.00000304, 0.00000289, -0.00000293, -0.00000391, 0.00000125, 0.00000050, 0.00000772, -0.00000739, 0.00001029, -0.00001327, 0.00001022, -0.00001965, 0.00001170, 0.00000020, 0.00000586, -0.00001190, 0.00001474, -0.00002061, 0.00001395, -0.00000691, 0.00001222, -0.00001049, 0.00001019, -0.00001924, 0.00001747, -0.00001231, 0.00001801, -0.00003273, 0.00002555, -0.00001659, 0.00002247, -0.00003036, 0.00002674, -0.00001702, 0.00001939, -0.00003406, 0.00003532, -0.00003331, 0.00002567, -0.00003362, 0.00003793, -0.00004270, 0.00003940, -0.00005170, 0.00006729, -0.00007665, 0.00006587, -0.00010976, 0.00011581, -0.00018495, 0.00017085, -0.00060934, 0.00083277, 0.00024650, -0.00045086, 0.00017622, -0.00018688, 0.00010647, -0.00009796, 0.00007775, -0.00009309, 0.00007183, -0.00005657, 0.00005041, -0.00005923, 0.00004810, -0.00004385, 0.00003926, -0.00004439, 0.00002669, -0.00003109, 0.00003915, -0.00003152, 0.00001948, -0.00001989, 0.00002832, -0.00002770, 0.00001056, -0.00001243, 0.00002330, -0.00002025, 0.00001569, -0.00001709, 0.00001703, -0.00001868, 0.00002027, -0.00001351, 0.00000495, -0.00001104, 0.00001146, -0.00001225, 0.00002030, -0.00001237, -0.00000017, -0.00000599, 0.00000413, -0.00000633, 0.00000776, -0.00000682, 0.00001147, -0.00000843, 0.00000913, -0.00001419, 0.00002232, -0.00002355, 0.00001942, -0.00000790, -0.00000904, 0.00001284, -0.00001624, },
		{ -0.00000275, 0.00000929, -0.00001602, 0.00001331, -0.00002859, 0.00003440, -0.00002782, 0.00000468, 0.00001020, -0.00001781, 0.00002365, -0.00000431, 0.00001377, -0.00001215, 0.00001918, -0.00001992, 0.00002531, -0.00002434, 0.00001219, -0.00000688, 0.00000085, -0.00000465, 0.00000207, 0.00000508, -0.00000099, 0.00000218, -0.00000103, -0.00000287, -0.00000144, -0.00001056, 0.00001096, 0.00000313, -0.00000547, 0.00000343, -0.00000210, -0.00000740, 0.00000655, 0.00000232, -0.00000024, 0.00000280, -0.00000824, 0.00000165, -0.00000065, 0.00000872, -0.00001151, -0.00000411, 0.00000537, 0.00000460, -0.00000534, -0.00000313, 0.00000529, 0.00000585, -0.00001425, 0.00000524, -0.00000125, -0.00000068, -0.00001092, 0.00001172, -0.00001270, 0.00001064, -0.00002288, 0.00002641, -0.00002064, 0.00001187, -0.00004351, 0.00002643, -0.00005351, 0.00001508, -0.00017104, -0.00024355, 0.00113788, -0.00054617, 0.00006304, -0.00010843, 0.00002741, -0.00004680, 0.00003524, -0.00004595, 0.00001652, -0.00001233, 0.00001673, -0.00002296, 0.00000859, -0.00000878, 0.00000781, -0.00001090, -0.00000163, -0.00001065, 0.00001086, 0.00000132, -0.00000406, -0.00000524, 0.00000975, -0.00000170, -0.00000723, -0.00000627, 0.00001211, -0.00000290, 0.00000088, -0.00000406, 0.00000301, -0.00000295, 0.00000061, 0.00000621, -0.00000574, -0.00000476, 0.00000273, -0.00000243, 0.00000328, 0.00000890, -0.00001051, -0.00000227, -0.00000001, -0.00000306, 0.00000277, -0.00000138, 0.00000341, 0.00000162, 0.00000045, -0.00000242, 0.00000029, 0.00000970, -0.00001991, 0.00002868, -0.00003023, 0.00001772, -0.00000948, -0.00000450, },
		{ 0.00000243, -0.00000085, -0.00000497, -0.00000126, -0.00001178, 0.00003083, -0.00004810, 0.00004489, -0.00003530, 0.00002500, -0.00000782, 0.00002186, -0.00001690, 0.00001821, -0.00001195, 0.00000915, 0.00000039, -0.00000959, 0.00000606, -0.00000351, -0.00000377, 0.00000128, -0.00000066, 0.00000945, -0.00000913, 0.00001213, -0.00001429, 0.00001106, -0.00001898, 0.00000883, 0.00000070, 0.00000829, -0.00001443, 0.00001700, -0.00002111, 0.00001250, -0.00000718, 0.00001398, -0.00001285, 0.00001371, -0.00002224, 0.00001966, -0.00001602, 0.00002342, -0.00003533, 0.00002569, -0.00001881, 0.00002605, -0.00003258, 0.00002734, -0.00001942, 0.00002522, -0.00003966, 0.00003929, -0.00003773, 0.00003122, -0.00004069, 0.00004619, -0.00005219, 0.00005020, -0.00006476, 0.00008203, -0.00009253, 0.00008484, -0.00013077, 0.00014034, -0.00021339, 0.00021447, -0.00062671, 0.00069789, 0.00045422, -0.00056429, 0.00022947, -0.00022572, 0.00013504, -0.00012363, 0.00009942, -0.00011138, 0.00008666, -0.00006939, 0.00006230, -0.00006923, 0.00005642, -0.00005084, 0.00004508, -0.00004904, 0.00003284, -0.00003837, 0.00004319, -0.00003283, 0.00002220, -0.00002474, 0.00003172, -0.00002826, 0.00001288, -0.00001795, 0.00002745, -0.00002269, 0.00001843, -0.00002018, 0.00001948, -0.00002049, 0.00002100, -0.00001280, 0.00000556, -0.00001332, 0.00001392, -0.00001469, 0.00002029, -0.00000968, -0.00000077, -0.00000737, 0.00000560, -0.00000855, 0.00001005, -0.00000921, 0.00001318, -0.00000918, 0.00001041, -0.00001542, 0.00002063, -0.00001840, 0.00001077, 0.00000277, -0.00001742, 0.00001806, -0.00001735, 0.00000052, },
		{ 0.00000740, -0.00001173, 0.00000937, -0.00002430, 0.00002900, -0.00002063, -0.00000269, 0.00001817, -0.00002546, 0.00002764, -0.00000981, 0.00001985, -0.00002024, 0.00002514, -0.00002419, 0.00002594, -0.00002028, 0.00000752, -0.00000324, -0.00000058, -0.00000413, 0.00000167, 0.00000441, 0.00000085, -0.00000043, 0.00000265, -0.00000663, 0.00000309, -0.00001454, 0.00001139, 0.00000294, -0.00000271, -0.00000065, 0.00000328, -0.00001229, 0.00000862, 0.00000032, 0.00000263, -0.00000053, -0.00000327, -0.00000377, 0.00000381, 0.00000325, -0.00000257, -0.00001262, 0.00001081, -0.00000148, 0.00000277, -0.00001106, 0.00001069, 0.00000084, -0.00000589, -0.00000475, 0.00000783, -0.00000960, -0.00000158, 0.00000111, -0.00000067, -0.00000303, -0.00000754, 0.00000784, 0.00000169, -0.00001378, -0.00001167, -0.00001233, -0.00000371, -0.00005025, -0.00005171, -0.00043083, 0.00122487, -0.00047613, -0.00001662, -0.00005667, -0.00001402, -0.00001460, 0.00000599, -0.00001941, -0.00000786, 0.00000914, -0.00000237, -0.00000473, -0.00000824, 0.00000648, -0.00000593, 0.00000128, -0.00001132, -0.00000077, -0.00000135, 0.00001200, -0.00001085, 0.00000061, 0.00000189, 0.00000610, -0.00001128, -0.00000310, 0.00000581, 0.00000375, -0.00000430, 0.00000095, -0.00000257, 0.00000256, -0.00000478, 0.00001084, -0.00000783, -0.00000310, -0.00000019, 0.00000130, -0.00000194, 0.00001310, -0.00001068, -0.00000256, -0.00000096, -0.00000200, 0.00000085, 0.00000112, -0.00000000, 0.00000537, -0.00000276, 0.00000181, -0.00000605, 0.00001590, -0.00002425, 0.00002902, -0.00002511, 0.00001022, -0.00000136, -0.00000997, 0.00000011, },
		{ 0.00000050, -0.00000482, -0.00000106, -0.00001240, 0.00003131, -0.00004685, 0.00004262, -0.00003122, 0.00001870, -0.00000437, 0.00001784, -0.00001232, 0.00000991, -0.00000270, -0.00000125, 0.00000967, -0.00001408, 0.00000721, -0.00000265, -0.00000443, 0.00000183, -0.00000206, 0.00001101, -0.00001029, 0.00001299, -0.00001406, 0.00001050, -0.00001688, 0.00000531, 0.00000112, 0.00001042, -0.00001596, 0.00001777, -0.00001995, 0.00001003, -0.00000699, 0.00001520, -0.00001440, 0.00001599, -0.00002366, 0.00002037, -0.00001820, 0.00002648, -0.00003506, 0.00002363, -0.00001930, 0.00002750, -0.00003234, 0.00002602, -0.00002054, 0.00002932, -0.00004265, 0.00004055, -0.00003965, 0.00003420, -0.00004498, 0.00005130, -0.00005802, 0.00005675, -0.00007278, 0.00009058, -0.00010089, 0.00009446, -0.00014008, 0.00015024, -0.00022249, 0.00022783, -0.00059536, 0.00051509, 0.00067654, -0.00066586, 0.00026819, -0.00025297, 0.00015403, -0.00014084, 0.00011316, -0.00012227, 0.00009463, -0.00007589, 0.00006828, -0.00007370, 0.00005965, -0.00005322, 0.00004690, -0.00005044, 0.00003602, -0.00004209, 0.00004349, -0.00003140, 0.00002307, -0.00002754, 0.00003279, -0.00002722, 0.00001440, -0.00002221, 0.00002970, -0.00002336, 0.00001968, -0.00002169, 0.00002021, -0.00002062, 0.00002017, -0.00001109, 0.00000580, -0.00001512, 0.00001548, -0.00001575, 0.00001878, -0.00000643, -0.00000108, -0.00000894, 0.00000702, -0.00001049, 0.00001171, -0.00001060, 0.00001361, -0.00000862, 0.00001019, -0.00001457, 0.00001657, -0.00001133, 0.00000133, 0.00001226, -0.00002297, 0.00001983, -0.00001530, -0.00000202, 0.00000339, },
		{ -0.00000749, 0.00000587, -0.00002049, 0.00002416, -0.00001434, -0.00000896, 0.00002465, -0.00003114, 0.00002986, -0.00001376, 0.00002422, -0.00002561, 0.00002742, -0.00002415, 0.00002209, -0.00001319, 0.00000154, 0.00000055, -0.00000167, -0.00000399, 0.00000179, 0.00000304, 0.00000339, -0.00000363, 0.00000664, -0.00001039, 0.00000724, -0.00001757, 0.00001117, 0.00000232, 0.00000086, -0.00000533, 0.00000877, -0.00001667, 0.00001038, -0.00000206, 0.00000628, -0.00000487, 0.00000264, -0.00000988, 0.00000913, -0.00000348, 0.00000717, -0.00002098, 0.00001626, -0.00000821, 0.00001146, -0.00001907, 0.00001643, -0.00000555, 0.00000430, -0.00001615, 0.00001821, -0.00002019, 0.00000988, -0.00001209, 0.00001438, -0.00002011, 0.00001169, -0.00001513, 0.00002878, -0.00004459, 0.00002556, -0.00005653, 0.00005182, -0.00012173, 0.00006777, -0.00059087, 0.00125488, -0.00035443, -0.00012499, 0.00001167, -0.00006744, 0.00002722, -0.00003129, 0.00001368, -0.00003761, 0.00003500, -0.00002514, 0.00001651, -0.00002746, 0.00002364, -0.00002113, 0.00001473, -0.00002257, 0.00001104, -0.00001483, 0.00002291, -0.00001792, 0.00000748, -0.00000696, 0.00001422, -0.00001579, 0.00000151, -0.00000206, 0.00001132, -0.00001020, 0.00000678, -0.00000887, 0.00000849, -0.00001030, 0.00001522, -0.00000978, -0.00000092, -0.00000393, 0.00000573, -0.00000725, 0.00001657, -0.00001032, -0.00000248, -0.00000264, -0.00000011, -0.00000199, 0.00000442, -0.00000392, 0.00000923, -0.00000589, 0.00000574, -0.00001123, 0.00001973, -0.00002518, 0.00002538, -0.00001681, 0.00000112, 0.00000686, -0.00001462, 0.00000231, 0.00000564, },
		{ -0.00000254, -0.00000282, -0.00001102, 0.00002964, -0.00004352, 0.00003823, -0.00002518, 0.00001102, 0.00000035, 0.00001221, -0.00000571, 0.00000007, 0.00000705, -0.00001076, 0.00001629, -0.00001548, 0.00000593, -0.00000034, -0.00000576, 0.00000269, -0.00000343, 0.00001209, -0.00001059, 0.00001266, -0.00001251, 0.00000859, -0.00001359, 0.00000144, 0.00000141, 0.00001199, -0.00001619, 0.00001687, -0.00001717, 0.00000666, -0.00000632, 0.00001560, -0.00001475, 0.00001660, -0.00002312, 0.00001925, -0.00001838, 0.00002674, -0.00003174, 0.00001934, -0.00001792, 0.00002652, -0.00002946, 0.00002262, -0.00001996, 0.00003097, -0.00004228, 0.00003843, -0.00003830, 0.00003372, -0.00004539, 0.00005194, -0.00005867, 0.00005732, -0.00007379, 0.00009080, -0.00009940, 0.00009224, -0.00013515, 0.00014291, -0.00020960, 0.00020929, -0.00051892, 0.00029716, 0.00089584, -0.00074117, 0.00028307, -0.00026226, 0.00015850, -0.00014550, 0.00011556, -0.00012298, 0.00009340, -0.00007413, 0.00006676, -0.00007133, 0.00005674, -0.00005014, 0.00004401, -0.00004787, 0.00003540, -0.00004150, 0.00003966, -0.00002706, 0.00002179, -0.00002787, 0.00003124, -0.00002436, 0.00001468, -0.00002457, 0.00002947, -0.00002188, 0.00001907, -0.00002125, 0.00001894, -0.00001888, 0.00001774, -0.00000842, 0.00000556, -0.00001615, 0.00001579, -0.00001524, 0.00001588, -0.00000284, -0.00000108, -0.00001046, 0.00000805, -0.00001176, 0.00001235, -0.00001071, 0.00001263, -0.00000676, 0.00000860, -0.00001202, 0.00001098, -0.00000362, -0.00000729, 0.00001897, -0.00002461, 0.00001769, -0.00001015, -0.00000651, 0.00000585, 0.00000006, },
		{ 0.00000382, -0.00001815, 0.00002088, -0.00000990, -0.00001322, 0.00002877, -0.00003413, 0.00002967, -0.00001552, 0.00002608, -0.00002749, 0.00002559, -0.00001988, 0.00001459, -0.00000437, -0.00000447, 0.00000352, -0.00000183, -0.00000457, 0.00000258, 0.00000098, 0.00000643, -0.00000705, 0.00001049, -0.00001365, 0.00001056, -0.00001940, 0.00001031, 0.00000132, 0.00000494, -0.00001008, 0.00001379, -0.00002011, 0.00001161, -0.00000464, 0.00001038, -0.00000972, 0.00000885, -0.00001600, 0.00001464, -0.00001062, 0.00001664, -0.00002825, 0.00002100, -0.00001487, 0.00001986, -0.00002636, 0.00002190, -0.00001269, 0.00001538, -0.00002782, 0.00002876, -0.00003130, 0.00002220, -0.00002643, 0.00003076, -0.00003864, 0.00003249, -0.00003970, 0.00005727, -0.00007652, 0.00006325, -0.00010026, 0.00010573, -0.00018947, 0.00017379, -0.00070927, 0.00122048, -0.00018335, -0.00025551, 0.00009036, -0.00012762, 0.00007414, -0.00007250, 0.00004952, -0.00006924, 0.00006206, -0.00004866, 0.00003813, -0.00004671, 0.00004056, -0.00003596, 0.00002788, -0.00003400, 0.00002331, -0.00002804, 0.00003276, -0.00002439, 0.00001458, -0.00001595, 0.00002196, -0.00002031, 0.00000709, -0.00001078, 0.00001901, -0.00001610, 0.00001277, -0.00001521, 0.00001417, -0.00001534, 0.00001891, -0.00001138, 0.00000163, -0.00000820, 0.00001039, -0.00001216, 0.00001903, -0.00000948, -0.00000193, -0.00000502, 0.00000249, -0.00000545, 0.00000808, -0.00000782, 0.00001264, -0.00000839, 0.00000875, -0.00001464, 0.00002080, -0.00002282, 0.00001860, -0.00000684, -0.00000782, 0.00001357, -0.00001723, 0.00000295, 0.00000488, -0.00000428, },
		{ -0.00000580, -0.00000844, 0.00002669, -0.00003907, 0.00003276, -0.00001832, 0.00000311, 0.00000528, 0.00000606, 0.00000156, -0.00000960, 0.00001543, -0.00001756, 0.00001900, -0.00001335, 0.00000233, 0.00000308, -0.00000738, 0.00000351, -0.00000445, 0.00001239, -0.00000981, 0.00001107, -0.00000972, 0.00000556, -0.00000949, -0.00000245, 0.00000154, 0.00001283, -0.00001499, 0.00001436, -0.00001303, 0.00000266, -0.00000516, 0.00001500, -0.00001364, 0.00001532, -0.00002050, 0.00001618, -0.00001638, 0.00002412, -0.00002557, 0.00001311, -0.00001474, 0.00002313, -0.00002402, 0.00001725, -0.00001750, 0.00002970, -0.00003815, 0.00003263, -0.00003337, 0.00002929, -0.00004129, 0.00004737, -0.00005331, 0.00005102, -0.00006686, 0.00008173, -0.00008717, 0.00007748, -0.00011561, 0.00011838, -0.00017541, 0.00016166, -0.00040628, 0.00006121, 0.00109294, -0.00077672, 0.00026668, -0.00024909, 0.00014517, -0.00013508, 0.00010469, -0.00011204, 0.00008192, -0.00006335, 0.00005720, -0.00006180, 0.00004751, -0.00004155, 0.00003638, -0.00004119, 0.00003072, -0.00003648, 0.00003187, -0.00002004, 0.00001835, -0.00002559, 0.00002704, -0.00001965, 0.00001344, -0.00002458, 0.00002648, -0.00001810, 0.00001646, -0.00001874, 0.00001563, -0.00001537, 0.00001389, -0.00000496, 0.00000482, -0.00001619, 0.00001467, -0.00001315, 0.00001180, 0.00000085, -0.00000080, -0.00001168, 0.00000836, -0.00001204, 0.00001174, -0.00000944, 0.00001031, -0.00000382, 0.00000600, -0.00000841, 0.00000484, 0.00000343, -0.00001369, 0.00002185, -0.00002198, 0.00001197, -0.00000263, -0.00001218, 0.00000921, -0.00000167, 0.00000125, },
		{ -0.00001791, 0.00001972, -0.00000782, -0.00001504, 0.00003020, -0.00003424, 0.00002701, -0.00001499, 0.00002525, -0.00002584, 0.00002000, -0.00001231, 0.00000496, 0.00000446, -0.00000912, 0.00000479, -0.00000059, -0.00000603, 0.00000405, -0.00000160, 0.00000965, -0.00001027, 0.00001371, -0.00001600, 0.00001269, -0.00001990, 0.00000887, -0.00000001, 0.00000916, -0.00001438, 0.00001781, -0.00002229, 0.00001216, -0.00000719, 0.00001450, -0.00001448, 0.00001464, -0.00002140, 0.00001958, -0.00001723, 0.00002479, -0.00003357, 0.00002437, -0.00002072, 0.00002712, -0.00003217, 0.00002647, -0.00001979, 0.00002618, -0.00003842, 0.00003815, -0.00004158, 0.00003386, -0.00004015, 0.00004641, -0.00005619, 0.00005203, -0.00006260, 0.00008333, -0.00010508, 0.00009620, -0.00013743, 0.00015076, -0.00024423, 0.00025484, -0.00077615, 0.00112066, 0.00002876, -0.00039735, 0.00017044, -0.00018751, 0.00012023, -0.00011242, 0.00008348, -0.00009864, 0.00008666, -0.00006970, 0.00005723, -0.00006342, 0.00005502, -0.00004850, 0.00003911, -0.00004407, 0.00003443, -0.00003940, 0.00004037, -0.00002940, 0.00002108, -0.00002417, 0.00002856, -0.00002428, 0.00001296, -0.00001944, 0.00002593, -0.00002124, 0.00001821, -0.00002084, 0.00001889, -0.00001933, 0.00002153, -0.00001244, 0.00000434, -0.00001260, 0.00001475, -0.00001620, 0.00002030, -0.00000826, -0.00000084, -0.00000797, 0.00000550, -0.00000913, 0.00001160, -0.00001118, 0.00001515, -0.00000984, 0.00001045, -0.00001604, 0.00001928, -0.00001788, 0.00001006, 0.00000295, -0.00001481, 0.00001736, -0.00001687, 0.00000139, 0.00000575, -0.00000282, 0.00000393, },
		{ -0.00000575, 0.00002362, -0.00003469, 0.00002744, -0.00001190, -0.00000381, 0.00000929, 0.00000052, 0.00000811, -0.00001753, 0.00002088, -0.00002043, 0.00001733, -0.00000799, -0.00000292, 0.00000689, -0.00000870, 0.00000388, -0.00000479, 0.00001169, -0.00000788, 0.00000833, -0.00000597, 0.00000180, -0.00000503, -0.00000601, 0.00000147, 0.00001280, -0.00001237, 0.00001046, -0.00000791, -0.00000168, -0.00000361, 0.00001333, -0.00001103, 0.00001219, -0.00001596, 0.00001134, -0.00001237, 0.00001893, -0.00001711, 0.00000552, -0.00001013, 0.00001760, -0.00001644, 0.00001024, -0.00001319, 0.00002542, -0.00003031, 0.00002338, -0.00002503, 0.00002100, -0.00003271, 0.00003759, -0.00004200, 0.00003797, -0.00005225, 0.00006386, -0.00006501, 0.00005145, -0.00008335, 0.00007934, -0.00012384, 0.00009170, -0.00027033, -0.00017341, 0.00124945, -0.00076196, 0.00021488, -0.00021179, 0.00011322, -0.00010919, 0.00008053, -0.00008975, 0.00006066, -0.00004417, 0.00004032, -0.00004587, 0.00003274, -0.00002819, 0.00002467, -0.00003089, 0.00002234, -0.00002748, 0.00002081, -0.00001099, 0.00001308, -0.00002089, 0.00002045, -0.00001330, 0.00001061, -0.00002209, 0.00002078, -0.00001222, 0.00001203, -0.00001432, 0.00001052, -0.00001040, 0.00000897, -0.00000098, 0.00000361, -0.00001513, 0.00001209, -0.00000963, 0.00000689, 0.00000438, -0.00000033, -0.00001231, 0.00000770, -0.00001116, 0.00000982, -0.00000691, 0.00000694, -0.00000018, 0.00000290, -0.00000447, -0.00000084, 0.00000874, -0.00001697, 0.00002056, -0.00001558, 0.00000375, 0.00000597, -0.00001787, 0.00001254, -0.00000381, 0.00000557, -0.00000892, },
		{ 0.00002071, -0.00000805, -0.00001458, 0.00002919, -0.00003186, 0.00002234, -0.00001258, 0.00002211, -0.00002129, 0.00001174, -0.00000300, -0.00000487, 0.00001153, -0.00001126, 0.00000379, 0.00000218, -0.00000829, 0.00000602, -0.00000440, 0.00001266, -0.00001285, 0.00001591, -0.00001711, 0.00001346, -0.00001909, 0.00000698, -0.00000158, 0.00001315, -0.00001774, 0.00002042, -0.00002300, 0.00001193, -0.00000948, 0.00001819, -0.00001853, 0.00001932, -0.00002541, 0.00002323, -0.00002245, 0.00003075, -0.00003627, 0.00002588, -0.00002516, 0.00003250, -0.00003584, 0.00002954, -0.00002597, 0.00003547, -0.00004662, 0.00004509, -0.00004964, 0.00004328, -0.00005140, 0.00005917, -0.00007032, 0.00006752, -0.00008061, 0.00010332, -0.00012614, 0.00011975, -0.00016285, 0.00018092, -0.00027885, 0.00030304, -0.00078743, 0.00096120, 0.00026806, -0.00053652, 0.00024139, -0.00023916, 0.00015902, -0.00014550, 0.00011079, -0.00012170, 0.00010526, -0.00008517, 0.00007115, -0.00007531, 0.00006506, -0.00005709, 0.00004696, -0.00005140, 0.00004297, -0.00004759, 0.00004476, -0.00003229, 0.00002626, -0.00003081, 0.00003332, -0.00002714, 0.00001840, -0.00002708, 0.00003117, -0.00002489, 0.00002239, -0.00002508, 0.00002207, -0.00002180, 0.00002281, -0.00001281, 0.00000694, -0.00001668, 0.00001831, -0.00001896, 0.00002027, -0.00000674, 0.00000074, -0.00001121, 0.00000852, -0.00001255, 0.00001448, -0.00001357, 0.00001644, -0.00001004, 0.00001076, -0.00001556, 0.00001574, -0.00001150, 0.00000143, 0.00001076, -0.00001839, 0.00001736, -0.00001317, -0.00000250, 0.00000844, -0.00000340, 0.00000641, -0.00001984, },
		{ 0.00002156, -0.00003153, 0.00002339, -0.00000702, -0.00000876, 0.00001151, -0.00000347, 0.00001281, -0.00002253, 0.00002244, -0.00001892, 0.00001163, -0.00000035, -0.00000875, 0.00001018, -0.00000911, 0.00000340, -0.00000420, 0.00000991, -0.00000491, 0.00000472, -0.00000164, -0.00000222, -0.00000070, -0.00000894, 0.00000121, 0.00001188, -0.00000852, 0.00000558, -0.00000231, -0.00000596, -0.00000178, 0.00001067, -0.00000706, 0.00000748, -0.00000991, 0.00000517, -0.00000679, 0.00001179, -0.00000721, -0.00000266, -0.00000464, 0.00001048, -0.00000734, 0.00000214, -0.00000735, 0.00001841, -0.00001934, 0.00001142, -0.00001400, 0.00000950, -0.00002034, 0.00002341, -0.00002571, 0.00001939, -0.00003146, 0.00003908, -0.00003532, 0.00001718, -0.00004224, 0.00003071, -0.00006136, 0.00000893, -0.00012616, -0.00038742, 0.00135009, -0.00069105, 0.00012796, -0.00015223, 0.00006463, -0.00006982, 0.00004516, -0.00005819, 0.00003168, -0.00001857, 0.00001800, -0.00002530, 0.00001408, -0.00001156, 0.00001014, -0.00001800, 0.00001117, -0.00001554, 0.00000764, -0.00000082, 0.00000654, -0.00001423, 0.00001200, -0.00000573, 0.00000636, -0.00001725, 0.00001274, -0.00000474, 0.00000621, -0.00000841, 0.00000409, -0.00000449, 0.00000346, 0.00000320, 0.00000201, -0.00001297, 0.00000817, -0.00000502, 0.00000155, 0.00000753, 0.00000023, -0.00001213, 0.00000592, -0.00000908, 0.00000672, -0.00000340, 0.00000295, 0.00000363, -0.00000013, -0.00000089, -0.00000530, 0.00001165, -0.00001688, 0.00001559, -0.00000664, -0.00000536, 0.00001403, -0.00002222, 0.00001479, -0.00000535, 0.00000930, -0.00001104, -0.00000410, },
		{ -0.00000996, -0.00001253, 0.00002652, -0.00002788, 0.00001659, -0.00000914, 0.00001754, -0.00001497, 0.00000237, 0.00000618, -0.00001297, 0.00001538, -0.00001015, 0.00000039, 0.00000626, -0.00001103, 0.00000811, -0.00000702, 0.00001502, -0.00001441, 0.00001681, -0.00001684, 0.00001288, -0.00001716, 0.00000485, -0.00000329, 0.00001652, -0.00001975, 0.00002137, -0.00002220, 0.00001093, -0.00001128, 0.00002101, -0.00002134, 0.00002232, -0.00002750, 0.00002505, -0.00002564, 0.00003390, -0.00003603, 0.00002531, -0.00002778, 0.00003546, -0.00003691, 0.00003065, -0.00003045, 0.00004215, -0.00005128, 0.00004853, -0.00005435, 0.00004910, -0.00005857, 0.00006721, -0.00007894, 0.00007662, -0.00009115, 0.00011439, -0.00013663, 0.00013061, -0.00017305, 0.00019250, -0.00028927, 0.00031511, -0.00074525, 0.00075418, 0.00051656, -0.00065749, 0.00029262, -0.00027502, 0.00018455, -0.00016684, 0.00012739, -0.00013505, 0.00011503, -0.00009272, 0.00007793, -0.00008072, 0.00006930, -0.00006059, 0.00005042, -0.00005498, 0.00004785, -0.00005168, 0.00004538, -0.00003266, 0.00002955, -0.00003519, 0.00003573, -0.00002841, 0.00002270, -0.00003284, 0.00003399, -0.00002646, 0.00002477, -0.00002739, 0.00002327, -0.00002249, 0.00002263, -0.00001242, 0.00000921, -0.00002001, 0.00002062, -0.00002018, 0.00001898, -0.00000505, 0.00000269, -0.00001441, 0.00001112, -0.00001525, 0.00001631, -0.00001468, 0.00001637, -0.00000902, 0.00000985, -0.00001368, 0.00001107, -0.00000500, -0.00000571, 0.00001515, -0.00001776, 0.00001341, -0.00000643, -0.00000827, 0.00001263, -0.00000574, 0.00001091, -0.00002348, 0.00002330, },
		{ -0.00003037, 0.00002136, -0.00000438, -0.00001117, 0.00001149, -0.00000541, 0.00001506, -0.00002404, 0.00001994, -0.00001349, 0.00000307, 0.00000808, -0.00001383, 0.00001201, -0.00000808, 0.00000182, -0.00000261, 0.00000714, -0.00000117, 0.00000064, 0.00000275, -0.00000600, 0.00000306, -0.00001101, 0.00000077, 0.00001008, -0.00000374, 0.00000023, 0.00000322, -0.00000983, 0.00000015, 0.00000720, -0.00000207, 0.00000171, -0.00000298, -0.00000166, -0.00000033, 0.00000357, 0.00000310, -0.00001054, 0.00000108, 0.00000244, 0.00000244, -0.00000628, -0.00000053, 0.00000934, -0.00000624, -0.00000205, -0.00000145, -0.00000400, -0.00000548, 0.00000636, -0.00000624, -0.00000260, -0.00000700, 0.00001039, -0.00000168, -0.00002100, 0.00000249, -0.00002123, 0.00000414, -0.00007584, 0.00001103, -0.00056406, 0.00138482, -0.00056406, 0.00001103, -0.00007584, 0.00000414, -0.00002123, 0.00000249, -0.00002100, -0.00000168, 0.00001039, -0.00000700, -0.00000260, -0.00000624, 0.00000636, -0.00000548, -0.00000400, -0.00000145, -0.00000205, -0.00000624, 0.00000934, -0.00000053, -0.00000628, 0.00000244, 0.00000244, 0.00000108, -0.00001054, 0.00000310, 0.00000357, -0.00000033, -0.00000166, -0.00000298, 0.00000171, -0.00000207, 0.00000720, 0.00000015, -0.00000983, 0.00000322, 0.00000023, -0.00000374, 0.00001008, 0.00000077, -0.00001101, 0.00000306, -0.00000600, 0.00000275, 0.00000064, -0.00000117, 0.00000714, -0.00000261, 0.00000182, -0.00000808, 0.00001201, -0.00001383, 0.00000808, 0.00000307, -0.00001349, 0.00001994, -0.00002404, 0.00001506, -0.00000541, 0.00001149, -0.00001117, -0.00000438, 0.00002136, },
		{ -0.00000996, 0.00002330, -0.00002348, 0.00001091, -0.00000574, 0.00001263, -0.00000827, -0.00000643, 0.00001341, -0.00001776, 0.00001515, -0.00000571, -0.00000500, 0.00001107, -0.00001368, 0.00000985, -0.00000902, 0.00001637, -0.00001468, 0.00001631, -0.00001525, 0.00001112, -0.00001441, 0.00000269, -0.00000505, 0.00001898, -0.00002018, 0.00002062, -0.00002001, 0.00000921, -0.00001242, 0.00002263, -0.00002249, 0.00002327, -0.00002739, 0.00002477, -0.00002646, 0.00003399, -0.00003284, 0.00002270, -0.00002841, 0.00003573, -0.00003519, 0.00002955, -0.00003266, 0.00004538, -0.00005168, 0.00004785, -0.00005498, 0.00005042, -0.00006059, 0.00006930, -0.00008072, 0.00007793, -0.00009272, 0.00011503, -0.00013505, 0.00012739, -0.00016684, 0.00018455, -0.00027502, 0.00029262, -0.00065749, 0.00051656, 0.00075418, -0.00074525, 0.00031511, -0.00028927, 0.00019250, -0.00017305, 0.00013061, -0.00013663, 0.00011439, -0.00009115, 0.00007662, -0.00007894, 0.00006721, -0.00005857, 0.00004910, -0.00005435, 0.00004853, -0.00005128, 0.00004215, -0.00003045, 0.00003065, -0.00003691, 0.00003546, -0.00002778, 0.00002531, -0.00003603, 0.00003390, -0.00002564, 0.00002505, -0.00002750, 0.00002232, -0.00002134, 0.00002101, -0.00001128, 0.00001093, -0.00002220, 0.00002137, -0.00001975, 0.00001652, -0.00000329, 0.00000485, -0.00001716, 0.00001288, -0.00001684, 0.00001681, -0.00001441, 0.00001502, -0.00000702, 0.00000811, -0.00001103, 0.00000626, 0.00000039, -0.00001015, 0.00001538, -0.00001297, 0.00000618, 0.00000237, -0.00001497, 0.00001754, -0.00000914, 0.00001659, -0.00002788, 0.00002652, -0.00001253, },
		{ 0.00002156, -0.00000410, -0.00001104, 0.00000930, -0.00000535, 0.00001479, -0.00002222, 0.00001403, -0.00000536, -0.00000664, 0.00001559, -0.00001688, 0.00001165, -0.00000530, -0.00000089, -0.00000013, 0.00000363, 0.00000295, -0.00000340, 0.00000672, -0.00000908, 0.00000592, -0.00001213, 0.00000023, 0.00000753, 0.00000155, -0.00000502, 0.00000817, -0.00001297, 0.00000201, 0.00000320, 0.00000346, -0.00000449, 0.00000409, -0.00000841, 0.00000621, -0.00000474, 0.00001274, -0.00001725, 0.00000636, -0.00000573, 0.00001200, -0.00001423, 0.00000654, -0.00000082, 0.00000764, -0.00001554, 0.00001117, -0.00001800, 0.00001014, -0.00001156, 0.00001408, -0.00002530, 0.00001800, -0.00001857, 0.00003168, -0.00005819, 0.00004516, -0.00006982, 0.00006463, -0.00015223, 0.00012796, -0.00069105, 0.00135009, -0.00038742, -0.00012616, 0.00000893, -0.00006136, 0.00003071, -0.00004224, 0.00001718, -0.00003532, 0.00003908, -0.00003146, 0.00001939, -0.00002571, 0.00002341, -0.00002034, 0.00000950, -0.00001400, 0.00001142, -0.00001934, 0.00001841, -0.00000735, 0.00000214, -0.00000734, 0.00001048, -0.00000464, -0.00000266, -0.00000721, 0.00001179, -0.00000679, 0.00000517, -0.00000991, 0.00000748, -0.00000706, 0.00001067, -0.00000178, -0.00000596, -0.00000231, 0.00000558, -0.00000852, 0.00001188, 0.00000121, -0.00000894, -0.00000070, -0.00000222, -0.00000164, 0.00000472, -0.00000491, 0.00000991, -0.00000420, 0.00000340, -0.00000911, 0.00001018, -0.00000875, -0.00000035, 0.00001163, -0.00001892, 0.00002244, -0.00002253, 0.00001281, -0.00000347, 0.00001151, -0.00000876, -0.00000702, 0.00002339, -0.00003153, },
		{ 0.00002071, -0.00001984, 0.00000641, -0.00000340, 0.00000844, -0.00000250, -0.00001317, 0.00001736, -0.00001839, 0.00001076, 0.00000143, -0.00001150, 0.00001574, -0.00001556, 0.00001076, -0.00001004, 0.00001644, -0.00001357, 0.00001448, -0.00001255, 0.00000852, -0.00001121, 0.00000074, -0.00000674, 0.00002027, -0.00001896, 0.00001831, -0.00001668, 0.00000694, -0.00001281, 0.00002281, -0.00002180, 0.00002207, -0.00002508, 0.00002239, -0.00002489, 0.00003117, -0.00002708, 0.00001840, -0.00002714, 0.00003332, -0.00003081, 0.00002626, -0.00003229, 0.00004476, -0.00004759, 0.00004297, -0.00005140, 0.00004696, -0.00005709, 0.00006506, -0.00007531, 0.00007115, -0.00008517, 0.00010526, -0.00012170, 0.00011079, -0.00014550, 0.00015902, -0.00023916, 0.00024139, -0.00053652, 0.00026806, 0.00096120, -0.00078743, 0.00030304, -0.00027885, 0.00018092, -0.00016285, 0.00011975, -0.00012614, 0.00010332, -0.00008061, 0.00006752, -0.00007032, 0.00005917, -0.00005140, 0.00004328, -0.00004964, 0.00004509, -0.00004662, 0.00003547, -0.00002597, 0.00002954, -0.00003584, 0.00003250, -0.00002516, 0.00002588, -0.00003627, 0.00003075, -0.00002245, 0.00002323, -0.00002541, 0.00001932, -0.00001853, 0.00001819, -0.00000948, 0.00001193, -0.00002300, 0.00002042, -0.00001774, 0.00001315, -0.00000158, 0.00000698, -0.00001909, 0.00001346, -0.00001711, 0.00001591, -0.00001285, 0.00001266, -0.00000440, 0.00000602, -0.00000829, 0.00000218, 0.00000379, -0.00001126, 0.00001153, -0.00000487, -0.00000300, 0.00001174, -0.00002129, 0.00002211, -0.00001258, 0.00002234, -0.00003186, 0.00002919, -0.00001458, -0.00000805, },
		{ -0.00000575, -0.00000892, 0.00000557, -0.00000381, 0.00001254, -0.00001787, 0.00000597, 0.00000375, -0.00001558, 0.00002056, -0.00001697, 0.00000874, -0.00000084, -0.00000447, 0.00000290, -0.00000018, 0.00000694, -0.00000691, 0.00000982, -0.00001116, 0.00000770, -0.00001231, -0.00000033, 0.00000438, 0.00000689, -0.00000963, 0.00001209, -0.00001513, 0.00000361, -0.00000098, 0.00000897, -0.00001040, 0.00001052, -0.00001432, 0.00001203, -0.00001222, 0.00002078, -0.00002209, 0.00001061, -0.00001330, 0.00002045, -0.00002089, 0.00001308, -0.00001099, 0.00002081, -0.00002748, 0.00002234, -0.00003089, 0.00002467, -0.00002819, 0.00003274, -0.00004587, 0.00004032, -0.00004417, 0.00006066, -0.00008975, 0.00008053, -0.00010919, 0.00011322, -0.00021179, 0.00021488, -0.00076196, 0.00124945, -0.00017341, -0.00027033, 0.00009170, -0.00012384, 0.00007934, -0.00008335, 0.00005145, -0.00006501, 0.00006386, -0.00005225, 0.00003797, -0.00004200, 0.00003759, -0.00003271, 0.00002100, -0.00002503, 0.00002338, -0.00003031, 0.00002542, -0.00001319, 0.00001024, -0.00001644, 0.00001760, -0.00001013, 0.00000552, -0.00001711, 0.00001893, -0.00001237, 0.00001134, -0.00001596, 0.00001219, -0.00001103, 0.00001333, -0.00000361, -0.00000168, -0.00000791, 0.00001046, -0.00001237, 0.00001280, 0.00000147, -0.00000601, -0.00000503, 0.00000180, -0.00000597, 0.00000833, -0.00000788, 0.00001169, -0.00000479, 0.00000388, -0.00000870, 0.00000689, -0.00000292, -0.00000799, 0.00001733, -0.00002043, 0.00002088, -0.00001753, 0.00000811, 0.00000052, 0.00000929, -0.00000381, -0.00001190, 0.00002744, -0.00003469, 0.00002362, },
		{ -0.00001791, 0.00000393, -0.00000282, 0.00000575, 0.00000139, -0.00001687, 0.00001736, -0.00001481, 0.00000295, 0.00001006, -0.00001788, 0.00001928, -0.00001604, 0.00001045, -0.00000984, 0.00001515, -0.00001118, 0.00001160, -0.00000913, 0.00000550, -0.00000797, -0.00000084, -0.00000826, 0.00002030, -0.00001620, 0.00001475, -0.00001260, 0.00000434, -0.00001244, 0.00002153, -0.00001933, 0.00001889, -0.00002084, 0.00001821, -0.00002124, 0.00002593, -0.00001944, 0.00001296, -0.00002428, 0.00002856, -0.00002417, 0.00002108, -0.00002940, 0.00004037, -0.00003940, 0.00003443, -0.00004407, 0.00003911, -0.00004850, 0.00005502, -0.00006342, 0.00005723, -0.00006970, 0.00008666, -0.00009864, 0.00008348, -0.00011242, 0.00012023, -0.00018751, 0.00017044, -0.00039735, 0.00002876, 0.00112066, -0.00077615, 0.00025484, -0.00024423, 0.00015076, -0.00013743, 0.00009620, -0.00010508, 0.00008333, -0.00006260, 0.00005203, -0.00005619, 0.00004641, -0.00004015, 0.00003386, -0.00004158, 0.00003815, -0.00003842, 0.00002618, -0.00001979, 0.00002647, -0.00003217, 0.00002712, -0.00002072, 0.00002437, -0.00003357, 0.00002479, -0.00001723, 0.00001958, -0.00002140, 0.00001464, -0.00001448, 0.00001450, -0.00000719, 0.00001216, -0.00002229, 0.00001781, -0.00001438, 0.00000916, -0.00000001, 0.00000887, -0.00001990, 0.00001269, -0.00001600, 0.00001371, -0.00001027, 0.00000965, -0.00000160, 0.00000405, -0.00000603, -0.00000059, 0.00000479, -0.00000912, 0.00000446, 0.00000496, -0.00001231, 0.00002000, -0.00002584, 0.00002525, -0.00001499, 0.00002701, -0.00003424, 0.00003020, -0.00001504, -0.00000782, 0.00001972, },
		{ -0.00000580, 0.00000125, -0.00000167, 0.00000921, -0.00001218, -0.00000263, 0.00001197, -0.00002198, 0.00002185, -0.00001369, 0.00000343, 0.00000484, -0.00000841, 0.00000600, -0.00000382, 0.00001031, -0.00000944, 0.00001174, -0.00001204, 0.00000836, -0.00001168, -0.00000080, 0.00000085, 0.00001180, -0.00001315, 0.00001467, -0.00001619, 0.00000482, -0.00000496, 0.00001389, -0.00001537, 0.00001563, -0.00001874, 0.00001646, -0.00001810, 0.00002648, -0.00002458, 0.00001344, -0.00001965, 0.00002704, -0.00002559, 0.00001835, -0.00002004, 0.00003187, -0.00003648, 0.00003072, -0.00004119, 0.00003638, -0.00004155, 0.00004751, -0.00006180, 0.00005720, -0.00006335, 0.00008192, -0.00011204, 0.00010469, -0.00013508, 0.00014517, -0.00024909, 0.00026668, -0.00077672, 0.00109294, 0.00006121, -0.00040628, 0.00016166, -0.00017541, 0.00011838, -0.00011561, 0.00007748, -0.00008717, 0.00008173, -0.00006686, 0.00005102, -0.00005331, 0.00004737, -0.00004129, 0.00002929, -0.00003337, 0.00003263, -0.00003815, 0.00002970, -0.00001750, 0.00001725, -0.00002402, 0.00002313, -0.00001474, 0.00001311, -0.00002557, 0.00002412, -0.00001638, 0.00001618, -0.00002050, 0.00001532, -0.00001364, 0.00001500, -0.00000516, 0.00000266, -0.00001303, 0.00001436, -0.00001499, 0.00001283, 0.00000154, -0.00000245, -0.00000949, 0.00000556, -0.00000972, 0.00001107, -0.00000981, 0.00001239, -0.00000445, 0.00000351, -0.00000738, 0.00000308, 0.00000233, -0.00001335, 0.00001900, -0.00001756, 0.00001543, -0.00000960, 0.00000156, 0.00000606, 0.00000528, 0.00000311, -0.00001832, 0.00003276, -0.00003907, 0.00002669, -0.00000844, },
		{ 0.00000382, -0.00000428, 0.00000488, 0.00000295, -0.00001723, 0.00001357, -0.00000782, -0.00000684, 0.00001860, -0.00002282, 0.00002080, -0.00001464, 0.00000875, -0.00000839, 0.00001264, -0.00000782, 0.00000808, -0.00000545, 0.00000249, -0.00000502, -0.00000193, -0.00000948, 0.00001903, -0.00001216, 0.00001039, -0.00000820, 0.00000163, -0.00001138, 0.00001891, -0.00001534, 0.00001417, -0.00001521, 0.00001277, -0.00001610, 0.00001901, -0.00001078, 0.00000709, -0.00002031, 0.00002196, -0.00001595, 0.00001458, -0.00002439, 0.00003276, -0.00002804, 0.00002331, -0.00003400, 0.00002788, -0.00003596, 0.00004056, -0.00004671, 0.00003813, -0.00004866, 0.00006206, -0.00006924, 0.00004952, -0.00007250, 0.00007414, -0.00012762, 0.00009036, -0.00025551, -0.00018335, 0.00122048, -0.00070927, 0.00017379, -0.00018947, 0.00010573, -0.00010026, 0.00006325, -0.00007652, 0.00005727, -0.00003970, 0.00003249, -0.00003864, 0.00003076, -0.00002643, 0.00002220, -0.00003130, 0.00002876, -0.00002782, 0.00001538, -0.00001269, 0.00002190, -0.00002636, 0.00001986, -0.00001487, 0.00002100, -0.00002825, 0.00001664, -0.00001062, 0.00001464, -0.00001600, 0.00000885, -0.00000972, 0.00001038, -0.00000464, 0.00001161, -0.00002011, 0.00001379, -0.00001008, 0.00000494, 0.00000132, 0.00001031, -0.00001940, 0.00001056, -0.00001365, 0.00001049, -0.00000705, 0.00000643, 0.00000098, 0.00000258, -0.00000457, -0.00000183, 0.00000352, -0.00000447, -0.00000437, 0.00001459, -0.00001988, 0.00002559, -0.00002749, 0.00002608, -0.00001552, 0.00002967, -0.00003413, 0.00002877, -0.00001322, -0.00000990, 0.00002088, -0.00001815, },
		{ -0.00000254, 0.00000006, 0.00000585, -0.00000651, -0.00001015, 0.00001769, -0.00002461, 0.00001897, -0.00000729, -0.00000362, 0.00001098, -0.00001202, 0.00000860, -0.00000676, 0.00001263, -0.00001071, 0.00001235, -0.00001176, 0.00000805, -0.00001046, -0.00000108, -0.00000284, 0.00001588, -0.00001524, 0.00001579, -0.00001615, 0.00000556, -0.00000842, 0.00001774, -0.00001888, 0.00001894, -0.00002125, 0.00001907, -0.00002188, 0.00002947, -0.00002457, 0.00001468, -0.00002436, 0.00003124, -0.00002787, 0.00002179, -0.00002706, 0.00003966, -0.00004150, 0.00003540, -0.00004787, 0.00004401, -0.00005014, 0.00005674, -0.00007133, 0.00006676, -0.00007413, 0.00009340, -0.00012298, 0.00011556, -0.00014550, 0.00015850, -0.00026226, 0.00028307, -0.00074117, 0.00089584, 0.00029716, -0.00051892, 0.00020929, -0.00020960, 0.00014291, -0.00013515, 0.00009224, -0.00009940, 0.00009080, -0.00007379, 0.00005732, -0.00005867, 0.00005194, -0.00004539, 0.00003372, -0.00003830, 0.00003843, -0.00004228, 0.00003097, -0.00001996, 0.00002262, -0.00002946, 0.00002652, -0.00001792, 0.00001934, -0.00003174, 0.00002674, -0.00001838, 0.00001925, -0.00002312, 0.00001660, -0.00001475, 0.00001560, -0.00000632, 0.00000666, -0.00001717, 0.00001687, -0.00001619, 0.00001199, 0.00000141, 0.00000144, -0.00001359, 0.00000859, -0.00001251, 0.00001266, -0.00001059, 0.00001209, -0.00000343, 0.00000269, -0.00000576, -0.00000034, 0.00000593, -0.00001548, 0.00001629, -0.00001076, 0.00000705, 0.00000007, -0.00000571, 0.00001221, 0.00000035, 0.00001102, -0.00002518, 0.00003823, -0.00004352, 0.00002964, -0.00001102, -0.00000282, },
		{ -0.00000749, 0.00000564, 0.00000231, -0.00001462, 0.00000686, 0.00000112, -0.00001681, 0.00002538, -0.00002518, 0.00001973, -0.00001123, 0.00000574, -0.00000589, 0.00000923, -0.00000392, 0.00000442, -0.00000199, -0.00000011, -0.00000264, -0.00000248, -0.00001032, 0.00001657, -0.00000725, 0.00000573, -0.00000393, -0.00000092, -0.00000978, 0.00001522, -0.00001030, 0.00000849, -0.00000887, 0.00000678, -0.00001020, 0.00001132, -0.00000206, 0.00000151, -0.00001579, 0.00001422, -0.00000696, 0.00000748, -0.00001792, 0.00002291, -0.00001483, 0.00001104, -0.00002257, 0.00001473, -0.00002113, 0.00002364, -0.00002746, 0.00001651, -0.00002514, 0.00003500, -0.00003761, 0.00001368, -0.00003129, 0.00002722, -0.00006744, 0.00001167, -0.00012499, -0.00035443, 0.00125488, -0.00059087, 0.00006777, -0.00012173, 0.00005182, -0.00005653, 0.00002556, -0.00004459, 0.00002878, -0.00001513, 0.00001169, -0.00002011, 0.00001438, -0.00001209, 0.00000988, -0.00002019, 0.00001821, -0.00001615, 0.00000430, -0.00000555, 0.00001643, -0.00001907, 0.00001146, -0.00000821, 0.00001626, -0.00002098, 0.00000717, -0.00000348, 0.00000913, -0.00000988, 0.00000264, -0.00000487, 0.00000628, -0.00000206, 0.00001038, -0.00001667, 0.00000877, -0.00000533, 0.00000086, 0.00000232, 0.00001117, -0.00001757, 0.00000724, -0.00001039, 0.00000664, -0.00000363, 0.00000339, 0.00000304, 0.00000179, -0.00000399, -0.00000167, 0.00000055, 0.00000154, -0.00001319, 0.00002209, -0.00002415, 0.00002742, -0.00002561, 0.00002422, -0.00001376, 0.00002986, -0.00003114, 0.00002465, -0.00000896, -0.00001434, 0.00002416, -0.00002049, 0.00000587, },
		{ 0.00000050, 0.00000339, -0.00000202, -0.00001530, 0.00001983, -0.00002297, 0.00001226, 0.00000133, -0.00001133, 0.00001657, -0.00001457, 0.00001019, -0.00000862, 0.00001361, -0.00001060, 0.00001171, -0.00001049, 0.00000702, -0.00000894, -0.00000108, -0.00000643, 0.00001878, -0.00001575, 0.00001548, -0.00001512, 0.00000580, -0.00001109, 0.00002017, -0.00002062, 0.00002021, -0.00002169, 0.00001968, -0.00002336, 0.00002970, -0.00002221, 0.00001440, -0.00002722, 0.00003279, -0.00002754, 0.00002307, -0.00003140, 0.00004349, -0.00004209, 0.00003602, -0.00005044, 0.00004690, -0.00005322, 0.00005965, -0.00007370, 0.00006828, -0.00007589, 0.00009463, -0.00012227, 0.00011316, -0.00014084, 0.00015403, -0.00025297, 0.00026819, -0.00066586, 0.00067654, 0.00051509, -0.00059536, 0.00022783, -0.00022249, 0.00015024, -0.00014008, 0.00009446, -0.00010089, 0.00009058, -0.00007278, 0.00005675, -0.00005802, 0.00005130, -0.00004498, 0.00003420, -0.00003965, 0.00004055, -0.00004265, 0.00002932, -0.00002054, 0.00002602, -0.00003234, 0.00002750, -0.00001930, 0.00002363, -0.00003506, 0.00002648, -0.00001820, 0.00002037, -0.00002366, 0.00001599, -0.00001440, 0.00001520, -0.00000699, 0.00001003, -0.00001995, 0.00001777, -0.00001596, 0.00001042, 0.00000112, 0.00000531, -0.00001688, 0.00001050, -0.00001406, 0.00001299, -0.00001029, 0.00001101, -0.00000206, 0.00000183, -0.00000443, -0.00000265, 0.00000721, -0.00001408, 0.00000967, -0.00000125, -0.00000270, 0.00000991, -0.00001232, 0.00001784, -0.00000437, 0.00001870, -0.00003122, 0.00004262, -0.00004685, 0.00003131, -0.00001240, -0.00000106, -0.00000482, },
		{ 0.00000740, 0.00000011, -0.00000997, -0.00000136, 0.00001022, -0.00002511, 0.00002902, -0.00002425, 0.00001590, -0.00000605, 0.00000181, -0.00000276, 0.00000537, -0.00000000, 0.00000112, 0.00000085, -0.00000200, -0.00000096, -0.00000256, -0.00001068, 0.00001310, -0.00000194, 0.00000130, -0.00000019, -0.00000310, -0.00000783, 0.00001084, -0.00000478, 0.00000256, -0.00000257, 0.00000095, -0.00000430, 0.00000375, 0.00000581, -0.00000310, -0.00001128, 0.00000610, 0.00000189, 0.00000061, -0.00001085, 0.00001200, -0.00000135, -0.00000077, -0.00001132, 0.00000128, -0.00000593, 0.00000648, -0.00000824, -0.00000473, -0.00000237, 0.00000914, -0.00000786, -0.00001941, 0.00000599, -0.00001460, -0.00001402, -0.00005667, -0.00001662, -0.00047613, 0.00122487, -0.00043083, -0.00005171, -0.00005025, -0.00000371, -0.00001233, -0.00001167, -0.00001378, 0.00000169, 0.00000784, -0.00000754, -0.00000303, -0.00000067, 0.00000111, -0.00000158, -0.00000960, 0.00000783, -0.00000475, -0.00000589, 0.00000084, 0.00001069, -0.00001106, 0.00000277, -0.00000148, 0.00001081, -0.00001262, -0.00000257, 0.00000325, 0.00000381, -0.00000377, -0.00000327, -0.00000053, 0.00000263, 0.00000032, 0.00000862, -0.00001229, 0.00000328, -0.00000065, -0.00000271, 0.00000294, 0.00001139, -0.00001454, 0.00000309, -0.00000663, 0.00000265, -0.00000043, 0.00000085, 0.00000441, 0.00000167, -0.00000413, -0.00000058, -0.00000324, 0.00000752, -0.00002028, 0.00002594, -0.00002419, 0.00002514, -0.00002024, 0.00001985, -0.00000981, 0.00002764, -0.00002546, 0.00001817, -0.00000269, -0.00002063, 0.00002900, -0.00002430, 0.00000937, -0.00001173, },
		{ 0.00000243, 0.00000052, -0.00001735, 0.00001806, -0.00001742, 0.00000277, 0.00001077, -0.00001840, 0.00002063, -0.00001542, 0.00001041, -0.00000918, 0.00001318, -0.00000921, 0.00001005, -0.00000855, 0.00000560, -0.00000737, -0.00000077, -0.00000968, 0.00002029, -0.00001469, 0.00001392, -0.00001332, 0.00000556, -0.00001280, 0.00002100, -0.00002049, 0.00001948, -0.00002018, 0.00001843, -0.00002269, 0.00002745, -0.00001795, 0.00001288, -0.00002826, 0.00003172, -0.00002474, 0.00002220, -0.00003283, 0.00004319, -0.00003837, 0.00003284, -0.00004904, 0.00004508, -0.00005084, 0.00005642, -0.00006923, 0.00006230, -0.00006939, 0.00008666, -0.00011138, 0.00009942, -0.00012363, 0.00013504, -0.00022572, 0.00022947, -0.00056429, 0.00045422, 0.00069789, -0.00062671, 0.00021447, -0.00021339, 0.00014034, -0.00013077, 0.00008484, -0.00009253, 0.00008203, -0.00006476, 0.00005020, -0.00005219, 0.00004619, -0.00004069, 0.00003122, -0.00003773, 0.00003929, -0.00003966, 0.00002522, -0.00001942, 0.00002734, -0.00003258, 0.00002605, -0.00001881, 0.00002569, -0.00003533, 0.00002342, -0.00001602, 0.00001966, -0.00002224, 0.00001371, -0.00001285, 0.00001398, -0.00000718, 0.00001250, -0.00002111, 0.00001700, -0.00001443, 0.00000829, 0.00000070, 0.00000883, -0.00001898, 0.00001106, -0.00001429, 0.00001213, -0.00000913, 0.00000945, -0.00000066, 0.00000128, -0.00000377, -0.00000351, 0.00000606, -0.00000959, 0.00000039, 0.00000915, -0.00001195, 0.00001821, -0.00001690, 0.00002186, -0.00000782, 0.00002500, -0.00003530, 0.00004489, -0.00004810, 0.00003083, -0.00001178, -0.00000126, -0.00000497, -0.00000085, },
		{ -0.00000275, -0.00000450, -0.00000948, 0.00001772, -0.00003023, 0.00002868, -0.00001991, 0.00000970, 0.00000029, -0.00000242, 0.00000045, 0.00000162, 0.00000341, -0.00000138, 0.00000277, -0.00000306, -0.00000001, -0.00000227, -0.00001051, 0.00000890, 0.00000328, -0.00000243, 0.00000273, -0.00000476, -0.00000574, 0.00000621, 0.00000061, -0.00000295, 0.00000301, -0.00000406, 0.00000088, -0.00000290, 0.00001211, -0.00000627, -0.00000723, -0.00000170, 0.00000975, -0.00000524, -0.00000406, 0.00000132, 0.00001086, -0.00001065, -0.00000163, -0.00001090, 0.00000781, -0.00000878, 0.00000859, -0.00002296, 0.00001673, -0.00001233, 0.00001652, -0.00004595, 0.00003524, -0.00004680, 0.00002741, -0.00010843, 0.00006304, -0.00054617, 0.00113788, -0.00024355, -0.00017104, 0.00001508, -0.00005351, 0.00002643, -0.00004351, 0.00001187, -0.00002064, 0.00002641, -0.00002288, 0.00001064, -0.00001270, 0.00001172, -0.00001092, -0.00000068, -0.00000125, 0.00000524, -0.00001425, 0.00000585, 0.00000529, -0.00000313, -0.00000534, 0.00000460, 0.00000537, -0.00000411, -0.00001151, 0.00000872, -0.00000065, 0.00000165, -0.00000824, 0.00000280, -0.00000024, 0.00000232, 0.00000655, -0.00000740, -0.00000210, 0.00000343, -0.00000547, 0.00000313, 0.00001096, -0.00001056, -0.00000144, -0.00000287, -0.00000103, 0.00000218, -0.00000099, 0.00000508, 0.00000207, -0.00000465, 0.00000085, -0.00000688, 0.00001219, -0.00002434, 0.00002531, -0.00001992, 0.00001918, -0.00001215, 0.00001377, -0.00000431, 0.00002365, -0.00001781, 0.00001020, 0.00000468, -0.00002782, 0.00003440, -0.00002859, 0.00001331, -0.00001602, 0.00000929, },
		{ 0.00000086, -0.00001624, 0.00001284, -0.00000904, -0.00000790, 0.00001942, -0.00002355, 0.00002232, -0.00001419, 0.00000913, -0.00000843, 0.00001147, -0.00000682, 0.00000776, -0.00000633, 0.00000413, -0.00000599, -0.00000017, -0.00001237, 0.00002030, -0.00001225, 0.00001146, -0.00001104, 0.00000495, -0.00001351, 0.00002027, -0.00001868, 0.00001703, -0.00001709, 0.00001569, -0.00002025, 0.00002330, -0.00001243, 0.00001056, -0.00002770, 0.00002832, -0.00001989, 0.00001948, -0.00003152, 0.00003915, -0.00003109, 0.00002669, -0.00004439, 0.00003926, -0.00004385, 0.00004810, -0.00005923, 0.00005041, -0.00005657, 0.00007183, -0.00009309, 0.00007775, -0.00009796, 0.00010647, -0.00018688, 0.00017622, -0.00045086, 0.00024650, 0.00083277, -0.00060934, 0.00017085, -0.00018495, 0.00011581, -0.00010976, 0.00006587, -0.00007665, 0.00006729, -0.00005170, 0.00003940, -0.00004270, 0.00003793, -0.00003362, 0.00002567, -0.00003331, 0.00003532, -0.00003406, 0.00001939, -0.00001702, 0.00002674, -0.00003036, 0.00002247, -0.00001659, 0.00002555, -0.00003273, 0.00001801, -0.00001231, 0.00001747, -0.00001924, 0.00001019, -0.00001049, 0.00001222, -0.00000691, 0.00001395, -0.00002061, 0.00001474, -0.00001190, 0.00000586, 0.00000020, 0.00001170, -0.00001965, 0.00001022, -0.00001327, 0.00001029, -0.00000739, 0.00000772, 0.00000050, 0.00000125, -0.00000391, -0.00000293, 0.00000289, -0.00000304, -0.00000984, 0.00001844, -0.00001891, 0.00002356, -0.00001845, 0.00002349, -0.00000922, 0.00002911, -0.00003665, 0.00004439, -0.00004673, 0.00002777, -0.00000883, -0.00000368, -0.00000285, -0.00000406, 0.00000311, },
		{ 0.00000053, -0.00001601, 0.00002222, -0.00003127, 0.00002425, -0.00001268, 0.00000198, 0.00000688, -0.00000619, 0.00000315, -0.00000152, 0.00000588, -0.00000280, 0.00000365, -0.00000327, 0.00000034, -0.00000180, -0.00000978, 0.00000430, 0.00000793, -0.00000511, 0.00000463, -0.00000582, -0.00000371, 0.00000177, 0.00000533, -0.00000746, 0.00000732, -0.00000775, 0.00000483, -0.00000806, 0.00001638, -0.00000772, -0.00000398, -0.00000857, 0.00001591, -0.00000944, 0.00000166, -0.00000799, 0.00002051, -0.00001746, 0.00000542, -0.00002059, 0.00001859, -0.00002047, 0.00002118, -0.00003621, 0.00003010, -0.00002724, 0.00003328, -0.00006358, 0.00005402, -0.00006686, 0.00005415, -0.00014063, 0.00011200, -0.00056794, 0.00100644, -0.00004608, -0.00027643, 0.00006531, -0.00009137, 0.00005497, -0.00006620, 0.00002940, -0.00003587, 0.00003872, -0.00003283, 0.00001964, -0.00002069, 0.00001887, -0.00001734, 0.00000579, -0.00000824, 0.00001303, -0.00002015, 0.00000906, 0.00000067, 0.00000402, -0.00001214, 0.00000939, 0.00000059, 0.00000365, -0.00001874, 0.00001224, -0.00000373, 0.00000589, -0.00001181, 0.00000480, -0.00000216, 0.00000381, 0.00000437, -0.00000246, -0.00000683, 0.00000650, -0.00000722, 0.00000291, 0.00000998, -0.00000601, -0.00000584, 0.00000040, -0.00000402, 0.00000396, -0.00000207, 0.00000517, 0.00000274, -0.00000515, 0.00000203, -0.00000955, 0.00001465, -0.00002469, 0.00002024, -0.00001214, 0.00001074, -0.00000263, 0.00000717, 0.00000173, 0.00001886, -0.00000929, 0.00000192, 0.00001194, -0.00003471, 0.00003921, -0.00003224, 0.00001657, -0.00001933, 0.00001037, -0.00000528, },
		{ -0.00001254, 0.00000526, 0.00000055, -0.00001794, 0.00002574, -0.00002580, 0.00002122, -0.00001086, 0.00000655, -0.00000662, 0.00000883, -0.00000386, 0.00000529, -0.00000422, 0.00000291, -0.00000495, 0.00000066, -0.00001434, 0.00001888, -0.00000875, 0.00000849, -0.00000859, 0.00000408, -0.00001329, 0.00001817, -0.00001555, 0.00001334, -0.00001297, 0.00001198, -0.00001664, 0.00001796, -0.00000641, 0.00000796, -0.00002591, 0.00002311, -0.00001365, 0.00001547, -0.00002801, 0.00003222, -0.00002144, 0.00001880, -0.00003762, 0.00003063, -0.00003366, 0.00003639, -0.00004572, 0.00003494, -0.00004012, 0.00005324, -0.00007100, 0.00005226, -0.00006862, 0.00007391, -0.00014341, 0.00011797, -0.00033891, 0.00006742, 0.00091262, -0.00054532, 0.00010288, -0.00014278, 0.00008139, -0.00008124, 0.00004130, -0.00005657, 0.00004928, -0.00003612, 0.00002652, -0.00003142, 0.00002813, -0.00002514, 0.00001872, -0.00002737, 0.00002958, -0.00002684, 0.00001267, -0.00001385, 0.00002454, -0.00002613, 0.00001727, -0.00001305, 0.00002353, -0.00002777, 0.00001099, -0.00000777, 0.00001436, -0.00001520, 0.00000597, -0.00000779, 0.00001021, -0.00000629, 0.00001436, -0.00001858, 0.00001131, -0.00000875, 0.00000339, -0.00000037, 0.00001372, -0.00001882, 0.00000811, -0.00001128, 0.00000780, -0.00000541, 0.00000609, 0.00000125, 0.00000175, -0.00000473, -0.00000129, -0.00000151, 0.00000420, -0.00001915, 0.00002480, -0.00002225, 0.00002511, -0.00001652, 0.00002247, -0.00000827, 0.00003071, -0.00003504, 0.00004104, -0.00004279, 0.00002229, -0.00000379, -0.00000802, 0.00000113, -0.00000869, 0.00000509, -0.00000069, },
		{ -0.00001989, 0.00002297, -0.00002811, 0.00001638, -0.00000368, -0.00000606, 0.00001269, -0.00000876, 0.00000484, -0.00000365, 0.00000715, -0.00000303, 0.00000354, -0.00000280, 0.00000028, -0.00000133, -0.00000854, -0.00000036, 0.00001160, -0.00000652, 0.00000547, -0.00000626, -0.00000190, -0.00000211, 0.00000892, -0.00001058, 0.00001004, -0.00000985, 0.00000728, -0.00001141, 0.00001845, -0.00000746, -0.00000167, -0.00001408, 0.00001990, -0.00001157, 0.00000577, -0.00001507, 0.00002672, -0.00002057, 0.00000931, -0.00002708, 0.00002553, -0.00002756, 0.00002849, -0.00004347, 0.00003674, -0.00003467, 0.00004156, -0.00007158, 0.00006179, -0.00007447, 0.00006603, -0.00015356, 0.00013241, -0.00054949, 0.00084637, 0.00014411, -0.00035585, 0.00009385, -0.00011326, 0.00007043, -0.00007766, 0.00003736, -0.00004301, 0.00004408, -0.00003692, 0.00002364, -0.00002434, 0.00002232, -0.00002060, 0.00000953, -0.00001276, 0.00001821, -0.00002336, 0.00001038, -0.00000286, 0.00000982, -0.00001704, 0.00001245, -0.00000298, 0.00000995, -0.00002359, 0.00001343, -0.00000518, 0.00000866, -0.00001372, 0.00000536, -0.00000310, 0.00000476, 0.00000228, 0.00000210, -0.00001044, 0.00000827, -0.00000791, 0.00000231, 0.00000856, -0.00000131, -0.00000961, 0.00000279, -0.00000605, 0.00000481, -0.00000245, 0.00000486, 0.00000338, -0.00000529, 0.00000251, -0.00001074, 0.00001450, -0.00002140, 0.00001162, -0.00000237, 0.00000149, 0.00000669, 0.00000142, 0.00000715, 0.00001441, -0.00000110, -0.00000543, 0.00001789, -0.00004016, 0.00004236, -0.00003427, 0.00001822, -0.00002083, 0.00000996, -0.00000671, 0.00000409, },
		{ -0.00000321, 0.00000957, -0.00002564, 0.00002858, -0.00002466, 0.00001734, -0.00000581, 0.00000311, -0.00000420, 0.00000571, -0.00000083, 0.00000308, -0.00000257, 0.00000213, -0.00000431, 0.00000159, -0.00001545, 0.00001618, -0.00000461, 0.00000545, -0.00000626, 0.00000310, -0.00001228, 0.00001506, -0.00001160, 0.00000901, -0.00000843, 0.00000793, -0.00001251, 0.00001220, -0.00000064, 0.00000558, -0.00002332, 0.00001676, -0.00000682, 0.00001089, -0.00002308, 0.00002355, -0.00001089, 0.00001059, -0.00003007, 0.00002064, -0.00002199, 0.00002331, -0.00003100, 0.00001851, -0.00002297, 0.00003417, -0.00004876, 0.00002710, -0.00004022, 0.00004266, -0.00010173, 0.00006308, -0.00023912, -0.00007383, 0.00093653, -0.00044208, 0.00001988, -0.00009444, 0.00004306, -0.00005025, 0.00001546, -0.00003597, 0.00003110, -0.00002065, 0.00001379, -0.00002025, 0.00001840, -0.00001662, 0.00001154, -0.00002096, 0.00002309, -0.00001902, 0.00000595, -0.00001047, 0.00002123, -0.00002050, 0.00001116, -0.00000877, 0.00002017, -0.00002123, 0.00000329, -0.00000317, 0.00001095, -0.00001072, 0.00000168, -0.00000524, 0.00000824, -0.00000543, 0.00001381, -0.00001531, 0.00000719, -0.00000546, 0.00000115, -0.00000098, 0.00001477, -0.00001659, 0.00000504, -0.00000868, 0.00000504, -0.00000351, 0.00000474, 0.00000159, 0.00000263, -0.00000591, 0.00000082, -0.00000614, 0.00001071, -0.00002594, 0.00002698, -0.00002139, 0.00002277, -0.00001134, 0.00001915, -0.00000522, 0.00003001, -0.00003080, 0.00003533, -0.00003686, 0.00001504, 0.00000261, -0.00001353, 0.00000617, -0.00001394, 0.00000769, -0.00000344, -0.00000728, },
		{ 0.00001995, -0.00002143, 0.00000634, 0.00000562, -0.00001314, 0.00001673, -0.00000955, 0.00000521, -0.00000459, 0.00000714, -0.00000216, 0.00000265, -0.00000190, 0.00000004, -0.00000100, -0.00000688, -0.00000474, 0.00001401, -0.00000659, 0.00000535, -0.00000617, -0.00000043, -0.00000514, 0.00001115, -0.00001215, 0.00001108, -0.00001035, 0.00000817, -0.00001293, 0.00001841, -0.00000571, -0.00000026, -0.00001802, 0.00002152, -0.00001154, 0.00000800, -0.00001948, 0.00002918, -0.00001997, 0.00001009, -0.00003025, 0.00002836, -0.00002985, 0.00003041, -0.00004473, 0.00003683, -0.00003496, 0.00004195, -0.00007083, 0.00005980, -0.00007126, 0.00006516, -0.00015016, 0.00012966, -0.00050179, 0.00067450, 0.00031134, -0.00040069, 0.00009752, -0.00011795, 0.00007226, -0.00007781, 0.00003603, -0.00004251, 0.00004303, -0.00003567, 0.00002311, -0.00002409, 0.00002241, -0.00002096, 0.00001067, -0.00001486, 0.00002080, -0.00002398, 0.00000994, -0.00000521, 0.00001395, -0.00001976, 0.00001356, -0.00000507, 0.00001434, -0.00002573, 0.00001224, -0.00000505, 0.00000993, -0.00001400, 0.00000458, -0.00000319, 0.00000522, 0.00000045, 0.00000591, -0.00001263, 0.00000865, -0.00000759, 0.00000143, 0.00000684, 0.00000313, -0.00001233, 0.00000405, -0.00000700, 0.00000477, -0.00000228, 0.00000437, 0.00000379, -0.00000486, 0.00000209, -0.00001032, 0.00001192, -0.00001525, 0.00000099, 0.00000744, -0.00000675, 0.00001424, -0.00000221, 0.00001093, 0.00001131, 0.00000569, -0.00001080, 0.00002160, -0.00004333, 0.00004313, -0.00003406, 0.00001772, -0.00002012, 0.00000772, -0.00000660, 0.00000563, -0.00002064, },
		{ 0.00001643, -0.00002977, 0.00002742, -0.00002026, 0.00001125, 0.00000023, -0.00000050, -0.00000172, 0.00000264, 0.00000179, 0.00000151, -0.00000160, 0.00000189, -0.00000403, 0.00000248, -0.00001566, 0.00001249, -0.00000030, 0.00000274, -0.00000431, 0.00000213, -0.00001070, 0.00001136, -0.00000739, 0.00000465, -0.00000409, 0.00000411, -0.00000847, 0.00000675, 0.00000427, 0.00000383, -0.00002036, 0.00000996, -0.00000023, 0.00000653, -0.00001763, 0.00001441, -0.00000090, 0.00000343, -0.00002299, 0.00001071, -0.00001058, 0.00001088, -0.00001730, 0.00000357, -0.00000779, 0.00001751, -0.00002952, 0.00000566, -0.00001651, 0.00001685, -0.00006671, 0.00001771, -0.00015854, -0.00017347, 0.00090944, -0.00031125, -0.00006683, -0.00004818, 0.00000699, -0.00002177, -0.00000752, -0.00001822, 0.00001552, -0.00000756, 0.00000309, -0.00001078, 0.00001009, -0.00000920, 0.00000514, -0.00001503, 0.00001678, -0.00001152, -0.00000004, -0.00000738, 0.00001729, -0.00001421, 0.00000489, -0.00000440, 0.00001613, -0.00001399, -0.00000412, 0.00000069, 0.00000784, -0.00000640, -0.00000212, -0.00000325, 0.00000657, -0.00000446, 0.00001250, -0.00001122, 0.00000291, -0.00000245, -0.00000063, -0.00000163, 0.00001486, -0.00001323, 0.00000142, -0.00000590, 0.00000240, -0.00000194, 0.00000377, 0.00000160, 0.00000365, -0.00000703, 0.00000277, -0.00001005, 0.00001531, -0.00002911, 0.00002454, -0.00001665, 0.00001723, -0.00000381, 0.00001438, -0.00000079, 0.00002775, -0.00002475, 0.00002820, -0.00002994, 0.00000707, 0.00000933, -0.00001917, 0.00001123, -0.00001884, 0.00001004, -0.00000647, -0.00000169, -0.00001100, },
		{ -0.00001254, -0.00000419, 0.00001365, -0.00001806, 0.00001831, -0.00000829, 0.00000421, -0.00000437, 0.00000600, -0.00000047, 0.00000132, -0.00000088, -0.00000013, -0.00000093, -0.00000493, -0.00000851, 0.00001499, -0.00000544, 0.00000447, -0.00000568, 0.00000064, -0.00000717, 0.00001194, -0.00001220, 0.00001061, -0.00000946, 0.00000772, -0.00001283, 0.00001665, -0.00000288, 0.00000043, -0.00002034, 0.00002083, -0.00000955, 0.00000843, -0.00002120, 0.00002811, -0.00001620, 0.00000831, -0.00003050, 0.00002747, -0.00002784, 0.00002766, -0.00004094, 0.00003155, -0.00002959, 0.00003623, -0.00006343, 0.00005058, -0.00006026, 0.00005517, -0.00013520, 0.00011100, -0.00043698, 0.00050656, 0.00044357, -0.00040691, 0.00007699, -0.00010711, 0.00006220, -0.00006843, 0.00002719, -0.00003605, 0.00003708, -0.00003045, 0.00001923, -0.00002093, 0.00001999, -0.00001906, 0.00000973, -0.00001491, 0.00002114, -0.00002245, 0.00000811, -0.00000643, 0.00001629, -0.00002025, 0.00001280, -0.00000562, 0.00001667, -0.00002522, 0.00000898, -0.00000365, 0.00000992, -0.00001284, 0.00000276, -0.00000269, 0.00000529, -0.00000102, 0.00000873, -0.00001326, 0.00000773, -0.00000646, 0.00000039, 0.00000497, 0.00000693, -0.00001373, 0.00000408, -0.00000692, 0.00000398, -0.00000176, 0.00000388, 0.00000383, -0.00000383, 0.00000083, -0.00000856, 0.00000756, -0.00000751, -0.00000977, 0.00001536, -0.00001242, 0.00001882, -0.00000289, 0.00001245, 0.00001020, 0.00001042, -0.00001359, 0.00002258, -0.00004387, 0.00004127, -0.00003148, 0.00001503, -0.00001725, 0.00000378, -0.00000500, 0.00000509, -0.00001846, 0.00001389, },
		{ -0.00002979, 0.00002245, -0.00001332, 0.00000391, 0.00000632, -0.00000355, 0.00000028, 0.00000007, 0.00000361, 0.00000081, -0.00000137, 0.00000214, -0.00000399, 0.00000316, -0.00001499, 0.00000817, 0.00000370, 0.00000071, -0.00000291, 0.00000130, -0.00000880, 0.00000753, -0.00000346, 0.00000080, -0.00000046, 0.00000099, -0.00000504, 0.00000215, 0.00000785, 0.00000300, -0.00001741, 0.00000339, 0.00000535, 0.00000306, -0.00001250, 0.00000598, 0.00000727, -0.00000165, -0.00001736, 0.00000204, -0.00000087, 0.00000071, -0.00000638, -0.00000803, 0.00000347, 0.00000529, -0.00001539, -0.00000981, 0.00000022, -0.00000105, -0.00004121, -0.00001477, -0.00010022, -0.00023287, 0.00084098, -0.00016695, -0.00014565, -0.00001153, -0.00002157, 0.00000015, -0.00002447, -0.00000578, 0.00000446, 0.00000162, -0.00000437, -0.00000401, 0.00000405, -0.00000363, 0.00000022, -0.00001023, 0.00001134, -0.00000506, -0.00000475, -0.00000495, 0.00001323, -0.00000796, -0.00000080, -0.00000054, 0.00001208, -0.00000693, -0.00001035, 0.00000322, 0.00000549, -0.00000272, -0.00000500, -0.00000210, 0.00000534, -0.00000348, 0.00001065, -0.00000678, -0.00000099, -0.00000011, -0.00000180, -0.00000229, 0.00001408, -0.00000911, -0.00000225, -0.00000337, 0.00000018, -0.00000087, 0.00000318, 0.00000142, 0.00000453, -0.00000772, 0.00000400, -0.00001254, 0.00001729, -0.00002834, 0.00001794, -0.00000911, 0.00000979, 0.00000473, 0.00000939, 0.00000396, 0.00002495, -0.00001801, 0.00002082, -0.00002322, -0.00000045, 0.00001526, -0.00002389, 0.00001530, -0.00002248, 0.00001136, -0.00000890, 0.00000309, -0.00001681, 0.00001999, },
		{ -0.00001347, 0.00001907, -0.00002003, 0.00001714, -0.00000508, 0.00000210, -0.00000325, 0.00000404, 0.00000166, -0.00000009, -0.00000004, -0.00000005, -0.00000116, -0.00000288, -0.00001144, 0.00001454, -0.00000330, 0.00000311, -0.00000496, 0.00000132, -0.00000819, 0.00001142, -0.00001101, 0.00000895, -0.00000754, 0.00000627, -0.00001152, 0.00001368, 0.00000050, 0.00000067, -0.00002122, 0.00001818, -0.00000607, 0.00000746, -0.00002061, 0.00002422, -0.00001028, 0.00000496, -0.00002867, 0.00002372, -0.00002263, 0.00002158, -0.00003371, 0.00002279, -0.00002072, 0.00002686, -0.00005222, 0.00003739, -0.00004520, 0.00004039, -0.00011410, 0.00008414, -0.00036651, 0.00035539, 0.00053364, -0.00037543, 0.00003665, -0.00008497, 0.00004381, -0.00005270, 0.00001371, -0.00002615, 0.00002840, -0.00002308, 0.00001353, -0.00001616, 0.00001612, -0.00001580, 0.00000743, -0.00001353, 0.00001981, -0.00001936, 0.00000538, -0.00000674, 0.00001692, -0.00001877, 0.00001047, -0.00000483, 0.00001709, -0.00002241, 0.00000424, -0.00000148, 0.00000900, -0.00001065, 0.00000032, -0.00000191, 0.00000513, -0.00000208, 0.00001044, -0.00001241, 0.00000576, -0.00000482, -0.00000064, 0.00000309, 0.00000984, -0.00001369, 0.00000297, -0.00000599, 0.00000269, -0.00000111, 0.00000352, 0.00000351, -0.00000236, -0.00000094, -0.00000601, 0.00000239, 0.00000032, -0.00001883, 0.00001980, -0.00001454, 0.00001987, -0.00000038, 0.00001160, 0.00001123, 0.00001293, -0.00001371, 0.00002090, -0.00004197, 0.00003711, -0.00002693, 0.00001059, -0.00001277, -0.00000132, -0.00000233, 0.00000290, -0.00001405, 0.00000600, -0.00000310, },
		{ 0.00001456, -0.00000506, -0.00000347, 0.00001143, -0.00000535, 0.00000136, -0.00000164, 0.00000441, 0.00000106, -0.00000183, 0.00000273, -0.00000401, 0.00000353, -0.00001356, 0.00000361, 0.00000698, -0.00000043, -0.00000214, 0.00000066, -0.00000682, 0.00000398, -0.00000024, -0.00000216, 0.00000214, -0.00000112, -0.00000255, -0.00000124, 0.00000988, 0.00000316, -0.00001472, -0.00000242, 0.00000936, 0.00000100, -0.00000834, -0.00000084, 0.00001277, -0.00000404, -0.00001372, -0.00000463, 0.00000619, -0.00000615, 0.00000073, -0.00001524, 0.00000983, -0.00000155, -0.00000726, -0.00001851, 0.00000924, -0.00001039, -0.00002589, -0.00003374, -0.00006356, -0.00025753, 0.00074382, -0.00002381, -0.00020642, 0.00000992, -0.00003917, 0.00001303, -0.00003361, 0.00000014, -0.00000125, 0.00000631, -0.00000814, -0.00000030, 0.00000059, -0.00000020, -0.00000293, -0.00000693, 0.00000718, -0.00000006, -0.00000788, -0.00000340, 0.00000945, -0.00000237, -0.00000531, 0.00000230, 0.00000859, -0.00000079, -0.00001476, 0.00000402, 0.00000416, 0.00000002, -0.00000670, -0.00000190, 0.00000458, -0.00000258, 0.00000852, -0.00000246, -0.00000407, 0.00000131, -0.00000230, -0.00000296, 0.00001259, -0.00000466, -0.00000550, -0.00000145, -0.00000139, -0.00000037, 0.00000289, 0.00000124, 0.00000504, -0.00000770, 0.00000421, -0.00001330, 0.00001654, -0.00002404, 0.00000840, -0.00000049, 0.00000212, 0.00001274, 0.00000547, 0.00000797, 0.00002267, -0.00001168, 0.00001434, -0.00001779, -0.00000652, 0.00001947, -0.00002686, 0.00001762, -0.00002422, 0.00001110, -0.00001011, 0.00000622, -0.00001980, 0.00001981, -0.00002589, },
		{ 0.00002098, -0.00001876, 0.00001340, -0.00000039, -0.00000064, -0.00000165, 0.00000170, 0.00000378, -0.00000119, 0.00000038, 0.00000038, -0.00000163, -0.00000092, -0.00001335, 0.00001278, -0.00000051, 0.00000159, -0.00000420, 0.00000166, -0.00000829, 0.00000986, -0.00000896, 0.00000655, -0.00000506, 0.00000428, -0.00000948, 0.00001010, 0.00000389, 0.00000079, -0.00002093, 0.00001408, -0.00000176, 0.00000565, -0.00001839, 0.00001854, -0.00000345, 0.00000115, -0.00002575, 0.00001826, -0.00001563, 0.00001385, -0.00002494, 0.00001268, -0.00001074, 0.00001649, -0.00004012, 0.00002349, -0.00002971, 0.00002498, -0.00009194, 0.00005600, -0.00029978, 0.00022969, 0.00057977, -0.00031178, -0.00001618, -0.00005740, 0.00002173, -0.00003452, -0.00000108, -0.00001556, 0.00001929, -0.00001546, 0.00000757, -0.00001108, 0.00001191, -0.00001209, 0.00000455, -0.00001141, 0.00001748, -0.00001540, 0.00000228, -0.00000642, 0.00001613, -0.00001575, 0.00000709, -0.00000310, 0.00001600, -0.00001794, -0.00000121, 0.00000083, 0.00000763, -0.00000791, -0.00000227, -0.00000121, 0.00000490, -0.00000274, 0.00001107, -0.00001031, 0.00000314, -0.00000301, -0.00000152, 0.00000130, 0.00001168, -0.00001229, 0.00000095, -0.00000450, 0.00000115, -0.00000050, 0.00000334, 0.00000296, -0.00000071, -0.00000280, -0.00000337, -0.00000254, 0.00000686, -0.00002477, 0.00001991, -0.00001291, 0.00001755, 0.00000492, 0.00000885, 0.00001403, 0.00001360, -0.00001165, 0.00001716, -0.00003832, 0.00003140, -0.00002121, 0.00000523, -0.00000753, -0.00000676, 0.00000069, -0.00000020, -0.00000852, -0.00000222, 0.00000521, -0.00002004, },
		{ 0.00000306, -0.00000966, 0.00001469, -0.00000543, 0.00000129, -0.00000236, 0.00000415, 0.00000213, -0.00000275, 0.00000344, -0.00000393, 0.00000350, -0.00001154, -0.00000080, 0.00000925, -0.00000059, -0.00000195, 0.00000023, -0.00000496, 0.00000102, 0.00000199, -0.00000402, 0.00000356, -0.00000214, -0.00000110, -0.00000331, 0.00001038, 0.00000423, -0.00001243, -0.00000708, 0.00001147, 0.00000055, -0.00000553, -0.00000553, 0.00001527, -0.00000371, -0.00001212, -0.00000905, 0.00001020, -0.00000935, 0.00000377, -0.00001795, 0.00001128, -0.00000317, -0.00000480, -0.00002096, 0.00001132, -0.00001217, -0.00001957, -0.00004089, -0.00004512, -0.00025569, 0.00063170, 0.00010497, -0.00024186, 0.00001347, -0.00004469, 0.00001632, -0.00003480, -0.00000030, -0.00000192, 0.00000686, -0.00000858, 0.00000062, -0.00000050, 0.00000123, -0.00000434, -0.00000515, 0.00000441, 0.00000336, -0.00000940, -0.00000276, 0.00000624, 0.00000208, -0.00000827, 0.00000384, 0.00000605, 0.00000391, -0.00001698, 0.00000303, 0.00000388, 0.00000173, -0.00000716, -0.00000260, 0.00000423, -0.00000178, 0.00000636, 0.00000134, -0.00000600, 0.00000171, -0.00000216, -0.00000358, 0.00001059, -0.00000032, -0.00000792, -0.00000038, -0.00000219, -0.00000042, 0.00000279, 0.00000118, 0.00000506, -0.00000691, 0.00000338, -0.00001242, 0.00001349, -0.00001728, -0.00000230, 0.00000731, -0.00000419, 0.00001888, 0.00000367, 0.00001037, 0.00002178, -0.00000666, 0.00000961, -0.00001440, -0.00001046, 0.00002141, -0.00002765, 0.00001784, -0.00002383, 0.00000910, -0.00000984, 0.00000732, -0.00001977, 0.00001617, -0.00001900, 0.00000513, },
		{ -0.00001459, 0.00000776, 0.00000500, -0.00000335, -0.00000009, -0.00000057, 0.00000549, -0.00000170, 0.00000025, 0.00000115, -0.00000225, 0.00000076, -0.00001420, 0.00000997, 0.00000252, 0.00000023, -0.00000356, 0.00000174, -0.00000766, 0.00000763, -0.00000651, 0.00000390, -0.00000249, 0.00000219, -0.00000720, 0.00000648, 0.00000681, 0.00000106, -0.00001981, 0.00000917, 0.00000264, 0.00000370, -0.00001534, 0.00001221, 0.00000304, -0.00000202, -0.00002273, 0.00001224, -0.00000828, 0.00000616, -0.00001649, 0.00000326, -0.00000182, 0.00000744, -0.00002963, 0.00001158, -0.00001673, 0.00001218, -0.00007268, 0.00003174, -0.00024322, 0.00013364, 0.00058514, -0.00022510, -0.00007255, -0.00003083, 0.00000071, -0.00001772, -0.00001406, -0.00000679, 0.00001174, -0.00000920, 0.00000265, -0.00000678, 0.00000827, -0.00000871, 0.00000180, -0.00000920, 0.00001477, -0.00001122, -0.00000068, -0.00000578, 0.00001430, -0.00001179, 0.00000325, -0.00000094, 0.00001394, -0.00001259, -0.00000655, 0.00000266, 0.00000627, -0.00000508, -0.00000456, -0.00000086, 0.00000472, -0.00000304, 0.00001075, -0.00000734, 0.00000033, -0.00000137, -0.00000212, -0.00000031, 0.00001244, -0.00000974, -0.00000160, -0.00000280, -0.00000035, -0.00000010, 0.00000331, 0.00000234, 0.00000078, -0.00000427, -0.00000126, -0.00000635, 0.00001111, -0.00002683, 0.00001568, -0.00000820, 0.00001274, 0.00001203, 0.00000511, 0.00001781, 0.00001321, -0.00000826, 0.00001231, -0.00003391, 0.00002516, -0.00001534, -0.00000006, -0.00000250, -0.00001163, 0.00000328, -0.00000335, -0.00000301, -0.00000932, 0.00001100, -0.00002295, 0.00001915, },
		{ -0.00001363, 0.00001554, -0.00000364, 0.00000012, -0.00000217, 0.00000299, 0.00000378, -0.00000387, 0.00000404, -0.00000359, 0.00000306, -0.00000916, -0.00000469, 0.00001030, 0.00000018, -0.00000225, -0.00000001, -0.00000336, -0.00000114, 0.00000314, -0.00000478, 0.00000385, -0.00000215, -0.00000066, -0.00000412, 0.00000955, 0.00000598, -0.00001055, -0.00001040, 0.00001160, 0.00000166, -0.00000409, -0.00000801, 0.00001494, -0.00000112, -0.00001214, -0.00001141, 0.00001133, -0.00000921, 0.00000324, -0.00001683, 0.00000872, -0.00000067, -0.00000669, -0.00001871, 0.00000829, -0.00000852, -0.00001974, -0.00003942, -0.00003979, -0.00023670, 0.00051753, 0.00020925, -0.00024852, -0.00000041, -0.00003939, 0.00001130, -0.00002938, -0.00000573, 0.00000121, 0.00000438, -0.00000661, -0.00000047, 0.00000015, 0.00000116, -0.00000439, -0.00000464, 0.00000287, 0.00000532, -0.00000948, -0.00000294, 0.00000374, 0.00000515, -0.00000953, 0.00000402, 0.00000462, 0.00000695, -0.00001698, 0.00000046, 0.00000444, 0.00000249, -0.00000652, -0.00000400, 0.00000413, -0.00000107, 0.00000436, 0.00000432, -0.00000666, 0.00000115, -0.00000152, -0.00000412, 0.00000831, 0.00000350, -0.00000923, -0.00000027, -0.00000224, -0.00000092, 0.00000277, 0.00000132, 0.00000462, -0.00000548, 0.00000177, -0.00001038, 0.00000901, -0.00000951, -0.00001229, 0.00001255, -0.00000789, 0.00002224, 0.00000463, 0.00001070, 0.00002276, -0.00000340, 0.00000703, -0.00001335, -0.00001211, 0.00002102, -0.00002631, 0.00001608, -0.00002154, 0.00000563, -0.00000824, 0.00000649, -0.00001707, 0.00000995, -0.00001054, -0.00000418, 0.00000962, },
		{ 0.00000122, 0.00001016, -0.00000534, 0.00000097, -0.00000238, 0.00000648, -0.00000148, -0.00000043, 0.00000218, -0.00000285, 0.00000200, -0.00001402, 0.00000646, 0.00000539, -0.00000070, -0.00000315, 0.00000166, -0.00000653, 0.00000512, -0.00000409, 0.00000143, -0.00000025, 0.00000038, -0.00000512, 0.00000328, 0.00000888, 0.00000171, -0.00001821, 0.00000409, 0.00000642, 0.00000224, -0.00001224, 0.00000629, 0.00000813, -0.00000376, -0.00002035, 0.00000663, -0.00000181, -0.00000013, -0.00000981, -0.00000396, 0.00000448, 0.00000131, -0.00002239, 0.00000338, -0.00000803, 0.00000395, -0.00005865, 0.00001439, -0.00019998, 0.00006709, 0.00055698, -0.00012660, -0.00012330, -0.00001106, -0.00001527, -0.00000536, -0.00002288, -0.00000160, 0.00000707, -0.00000533, -0.00000041, -0.00000393, 0.00000577, -0.00000617, -0.00000035, -0.00000737, 0.00001221, -0.00000733, -0.00000312, -0.00000512, 0.00001185, -0.00000751, -0.00000042, 0.00000116, 0.00001149, -0.00000712, -0.00001101, 0.00000353, 0.00000529, -0.00000257, -0.00000618, -0.00000108, 0.00000467, -0.00000304, 0.00000970, -0.00000392, -0.00000223, -0.00000019, -0.00000235, -0.00000171, 0.00001219, -0.00000639, -0.00000428, -0.00000122, -0.00000159, 0.00000004, 0.00000336, 0.00000185, 0.00000184, -0.00000501, -0.00000016, -0.00000852, 0.00001260, -0.00002504, 0.00000797, -0.00000175, 0.00000675, 0.00001966, 0.00000155, 0.00002156, 0.00001276, -0.00000457, 0.00000741, -0.00002981, 0.00001943, -0.00001031, -0.00000437, 0.00000144, -0.00001518, 0.00000479, -0.00000580, 0.00000146, -0.00001422, 0.00001345, -0.00002199, 0.00001401, -0.00000836, },
		{ 0.00001385, -0.00000019, -0.00000187, -0.00000134, 0.00000122, 0.00000565, -0.00000484, 0.00000431, -0.00000290, 0.00000228, -0.00000667, -0.00000777, 0.00001011, 0.00000171, -0.00000285, -0.00000013, -0.00000210, -0.00000243, 0.00000326, -0.00000458, 0.00000324, -0.00000138, -0.00000099, -0.00000393, 0.00000775, 0.00000808, -0.00000899, -0.00001239, 0.00000997, 0.00000398, -0.00000377, -0.00000857, 0.00001239, 0.00000291, -0.00001308, -0.00001226, 0.00001018, -0.00000654, 0.00000018, -0.00001312, 0.00000358, 0.00000430, -0.00001107, -0.00001385, 0.00000253, -0.00000212, -0.00002337, -0.00003316, -0.00004196, -0.00020951, 0.00041190, 0.00028294, -0.00022707, -0.00002821, -0.00002654, 0.00000070, -0.00001978, -0.00001399, 0.00000628, 0.00000041, -0.00000349, -0.00000256, 0.00000173, 0.00000023, -0.00000357, -0.00000498, 0.00000223, 0.00000610, -0.00000845, -0.00000371, 0.00000197, 0.00000678, -0.00000919, 0.00000300, 0.00000424, 0.00000834, -0.00001504, -0.00000323, 0.00000547, 0.00000255, -0.00000506, -0.00000576, 0.00000406, -0.00000042, 0.00000266, 0.00000631, -0.00000607, -0.00000017, -0.00000060, -0.00000450, 0.00000594, 0.00000651, -0.00000931, -0.00000107, -0.00000166, -0.00000173, 0.00000273, 0.00000164, 0.00000385, -0.00000370, -0.00000019, -0.00000784, 0.00000414, -0.00000221, -0.00001988, 0.00001405, -0.00000845, 0.00002257, 0.00000835, 0.00000905, 0.00002555, -0.00000185, 0.00000647, -0.00001440, -0.00001181, 0.00001874, -0.00002334, 0.00001290, -0.00001795, 0.00000129, -0.00000579, 0.00000424, -0.00001254, 0.00000247, -0.00000213, -0.00001182, 0.00001348, -0.00001478, },
		{ 0.00001415, -0.00000603, 0.00000120, -0.00000347, 0.00000659, -0.00000050, -0.00000155, 0.00000330, -0.00000328, 0.00000270, -0.00001294, 0.00000263, 0.00000772, -0.00000103, -0.00000305, 0.00000149, -0.00000515, 0.00000273, -0.00000207, -0.00000054, 0.00000139, -0.00000088, -0.00000352, 0.00000085, 0.00000991, 0.00000282, -0.00001642, -0.00000059, 0.00000901, 0.00000174, -0.00000972, 0.00000156, 0.00001116, -0.00000367, -0.00001900, 0.00000206, 0.00000294, -0.00000415, -0.00000574, -0.00000817, 0.00000737, -0.00000119, -0.00001905, -0.00000051, -0.00000414, 0.00000079, -0.00005051, 0.00000475, -0.00017029, 0.00002643, 0.00050500, -0.00002784, -0.00016056, -0.00000219, -0.00002375, 0.00000082, -0.00002634, -0.00000073, 0.00000577, -0.00000414, -0.00000139, -0.00000274, 0.00000460, -0.00000467, -0.00000166, -0.00000619, 0.00001009, -0.00000408, -0.00000478, -0.00000465, 0.00000921, -0.00000350, -0.00000341, 0.00000274, 0.00000918, -0.00000223, -0.00001404, 0.00000313, 0.00000491, -0.00000065, -0.00000692, -0.00000193, 0.00000476, -0.00000282, 0.00000817, -0.00000052, -0.00000414, 0.00000035, -0.00000219, -0.00000286, 0.00001108, -0.00000265, -0.00000666, -0.00000006, -0.00000239, -0.00000013, 0.00000341, 0.00000164, 0.00000229, -0.00000487, -0.00000019, -0.00000893, 0.00001150, -0.00002015, -0.00000171, 0.00000470, 0.00000108, 0.00002648, -0.00000071, 0.00002435, 0.00001319, -0.00000154, 0.00000342, -0.00002693, 0.00001505, -0.00000686, -0.00000705, 0.00000374, -0.00001697, 0.00000484, -0.00000706, 0.00000423, -0.00001633, 0.00001238, -0.00001763, 0.00000661, -0.00000132, -0.00000503, },
		{ 0.00000437, -0.00000416, -0.00000025, -0.00000077, 0.00000736, -0.00000540, 0.00000409, -0.00000185, 0.00000127, -0.00000432, -0.00000988, 0.00000877, 0.00000373, -0.00000352, -0.00000020, -0.00000120, -0.00000291, 0.00000256, -0.00000371, 0.00000202, -0.00000013, -0.00000183, -0.00000307, 0.00000540, 0.00001016, -0.00000760, -0.00001320, 0.00000701, 0.00000695, -0.00000409, -0.00000779, 0.00000853, 0.00000735, -0.00001409, -0.00001229, 0.00000756, -0.00000241, -0.00000410, -0.00000830, -0.00000246, 0.00000989, -0.00001595, -0.00000854, -0.00000359, 0.00000446, -0.00002754, -0.00002567, -0.00004660, -0.00018147, 0.00032206, 0.00032436, -0.00018194, -0.00006414, -0.00001064, -0.00001200, -0.00000892, -0.00002263, 0.00001131, -0.00000347, -0.00000050, -0.00000465, 0.00000341, -0.00000089, -0.00000244, -0.00000573, 0.00000211, 0.00000608, -0.00000669, -0.00000481, 0.00000084, 0.00000710, -0.00000756, 0.00000111, 0.00000468, 0.00000837, -0.00001166, -0.00000740, 0.00000649, 0.00000225, -0.00000313, -0.00000753, 0.00000380, 0.00000021, 0.00000133, 0.00000730, -0.00000445, -0.00000193, 0.00000036, -0.00000467, 0.00000366, 0.00000853, -0.00000819, -0.00000260, -0.00000067, -0.00000266, 0.00000260, 0.00000208, 0.00000298, -0.00000194, -0.00000197, -0.00000554, -0.00000010, 0.00000336, -0.00002398, 0.00001144, -0.00000612, 0.00002027, 0.00001430, 0.00000597, 0.00002969, -0.00000151, 0.00000736, -0.00001689, -0.00001027, 0.00001536, -0.00001956, 0.00000910, -0.00001391, -0.00000315, -0.00000317, 0.00000129, -0.00000724, -0.00000486, 0.00000472, -0.00001659, 0.00001405, -0.00001303, 0.00000996, },
		{ -0.00000511, 0.00000048, -0.00000378, 0.00000583, 0.00000108, -0.00000290, 0.00000431, -0.00000340, 0.00000284, -0.00001119, -0.00000112, 0.00000924, -0.00000069, -0.00000323, 0.00000130, -0.00000374, 0.00000073, -0.00000068, -0.00000182, 0.00000229, -0.00000148, -0.00000257, -0.00000068, 0.00000986, 0.00000439, -0.00001466, -0.00000444, 0.00001005, 0.00000243, -0.00000810, -0.00000154, 0.00001192, -0.00000181, -0.00001869, -0.00000123, 0.00000563, -0.00000559, -0.00000447, -0.00000927, 0.00000693, -0.00000026, -0.00001930, -0.00000053, -0.00000447, 0.00000200, -0.00004749, 0.00000186, -0.00015208, 0.00000580, 0.00043982, 0.00006085, -0.00017890, -0.00000599, -0.00002411, 0.00000060, -0.00002453, -0.00000388, 0.00000746, -0.00000528, -0.00000060, -0.00000297, 0.00000462, -0.00000414, -0.00000212, -0.00000571, 0.00000855, -0.00000161, -0.00000558, -0.00000449, 0.00000672, -0.00000021, -0.00000534, 0.00000353, 0.00000740, 0.00000160, -0.00001530, 0.00000140, 0.00000514, 0.00000057, -0.00000672, -0.00000335, 0.00000492, -0.00000242, 0.00000644, 0.00000250, -0.00000515, 0.00000017, -0.00000170, -0.00000373, 0.00000935, 0.00000107, -0.00000839, 0.00000046, -0.00000267, -0.00000057, 0.00000339, 0.00000175, 0.00000214, -0.00000390, -0.00000123, -0.00000788, 0.00000846, -0.00001346, -0.00001151, 0.00000943, -0.00000300, 0.00003144, -0.00000085, 0.00002550, 0.00001518, 0.00000017, 0.00000094, -0.00002577, 0.00001243, -0.00000531, -0.00000788, 0.00000423, -0.00001694, 0.00000344, -0.00000700, 0.00000508, -0.00001562, 0.00000831, -0.00001095, -0.00000166, 0.00000519, -0.00000991, 0.00001624, },
		{ -0.00000614, 0.00000067, -0.00000261, 0.00000859, -0.00000535, 0.00000336, -0.00000052, 0.00000021, -0.00000234, -0.00001092, 0.00000650, 0.00000591, -0.00000406, -0.00000030, -0.00000062, -0.00000273, 0.00000134, -0.00000249, 0.00000054, 0.00000126, -0.00000285, -0.00000190, 0.00000293, 0.00001188, -0.00000621, -0.00001309, 0.00000325, 0.00000994, -0.00000447, -0.00000636, 0.00000431, 0.00001122, -0.00001443, -0.00001215, 0.00000433, 0.00000204, -0.00000830, -0.00000378, -0.00000789, 0.00001452, -0.00001966, -0.00000454, -0.00000820, 0.00000923, -0.00003007, -0.00001969, -0.00004995, -0.00015761, 0.00025159, 0.00033590, -0.00012035, -0.00010126, 0.00000349, -0.00002333, 0.00000045, -0.00002945, 0.00001461, -0.00000597, 0.00000137, -0.00000596, 0.00000456, -0.00000167, -0.00000145, -0.00000647, 0.00000213, 0.00000563, -0.00000459, -0.00000595, 0.00000019, 0.00000638, -0.00000511, -0.00000118, 0.00000557, 0.00000746, -0.00000749, -0.00001137, 0.00000706, 0.00000192, -0.00000111, -0.00000894, 0.00000315, 0.00000087, 0.00000040, 0.00000739, -0.00000210, -0.00000375, 0.00000112, -0.00000459, 0.00000160, 0.00000949, -0.00000606, -0.00000456, 0.00000049, -0.00000355, 0.00000234, 0.00000252, 0.00000225, -0.00000056, -0.00000313, -0.00000404, -0.00000299, 0.00000645, -0.00002424, 0.00000519, -0.00000186, 0.00001630, 0.00002150, 0.00000239, 0.00003435, -0.00000153, 0.00000883, -0.00001993, -0.00000844, 0.00001179, -0.00001587, 0.00000557, -0.00001026, -0.00000692, -0.00000102, -0.00000158, -0.00000224, -0.00001078, 0.00000890, -0.00001788, 0.00001134, -0.00000888, 0.00000459, 0.00000925, },
		{ -0.00000106, -0.00000342, 0.00000438, 0.00000301, -0.00000421, 0.00000501, -0.00000310, 0.00000247, -0.00000904, -0.00000444, 0.00000979, 0.00000029, -0.00000365, 0.00000112, -0.00000249, -0.00000068, -0.00000002, -0.00000239, 0.00000246, -0.00000143, -0.00000228, -0.00000131, 0.00000884, 0.00000630, -0.00001304, -0.00000721, 0.00000946, 0.00000425, -0.00000741, -0.00000297, 0.00001066, 0.00000138, -0.00001906, -0.00000334, 0.00000632, -0.00000470, -0.00000559, -0.00000785, 0.00000386, 0.00000321, -0.00002209, 0.00000211, -0.00000764, 0.00000603, -0.00004790, 0.00000363, -0.00014196, -0.00000155, 0.00037132, 0.00013173, -0.00017606, -0.00002166, -0.00001761, -0.00000485, -0.00001862, -0.00000989, 0.00001110, -0.00000787, 0.00000125, -0.00000409, 0.00000541, -0.00000426, -0.00000194, -0.00000581, 0.00000754, 0.00000007, -0.00000555, -0.00000467, 0.00000462, 0.00000206, -0.00000608, 0.00000343, 0.00000635, 0.00000411, -0.00001477, -0.00000145, 0.00000584, 0.00000115, -0.00000569, -0.00000515, 0.00000503, -0.00000189, 0.00000471, 0.00000485, -0.00000515, -0.00000064, -0.00000098, -0.00000432, 0.00000724, 0.00000438, -0.00000922, 0.00000027, -0.00000244, -0.00000123, 0.00000325, 0.00000217, 0.00000152, -0.00000238, -0.00000286, -0.00000600, 0.00000448, -0.00000646, -0.00001967, 0.00001109, -0.00000475, 0.00003399, 0.00000150, 0.00002476, 0.00001898, 0.00000036, 0.00000011, -0.00002640, 0.00001155, -0.00000553, -0.00000706, 0.00000319, -0.00001544, 0.00000093, -0.00000587, 0.00000422, -0.00001263, 0.00000225, -0.00000339, -0.00000928, 0.00000993, -0.00001259, 0.00001605, -0.00000256, },
		{ 0.00000104, -0.00000399, 0.00000911, -0.00000460, 0.00000218, 0.00000095, -0.00000074, -0.00000088, -0.00001096, 0.00000361, 0.00000792, -0.00000428, -0.00000051, -0.00000030, -0.00000210, -0.00000010, -0.00000125, -0.00000090, 0.00000251, -0.00000378, -0.00000075, 0.00000072, 0.00001297, -0.00000467, -0.00001236, -0.00000071, 0.00001230, -0.00000437, -0.00000494, 0.00000060, 0.00001377, -0.00001361, -0.00001228, 0.00000116, 0.00000590, -0.00001141, -0.00000061, -0.00001164, 0.00001710, -0.00002114, -0.00000291, -0.00001022, 0.00001108, -0.00002975, -0.00001672, -0.00004993, -0.00014047, 0.00020065, 0.00032306, -0.00005108, -0.00013263, 0.00001175, -0.00003056, 0.00000628, -0.00003293, 0.00001513, -0.00000633, 0.00000160, -0.00000609, 0.00000484, -0.00000180, -0.00000086, -0.00000691, 0.00000202, 0.00000505, -0.00000248, -0.00000690, -0.00000018, 0.00000499, -0.00000234, -0.00000341, 0.00000653, 0.00000609, -0.00000320, -0.00001455, 0.00000680, 0.00000182, 0.00000067, -0.00000973, 0.00000202, 0.00000156, -0.00000016, 0.00000677, 0.00000057, -0.00000528, 0.00000150, -0.00000425, -0.00000015, 0.00000945, -0.00000320, -0.00000661, 0.00000156, -0.00000425, 0.00000194, 0.00000284, 0.00000186, 0.00000020, -0.00000340, -0.00000364, -0.00000419, 0.00000690, -0.00002108, -0.00000347, 0.00000284, 0.00001189, 0.00002880, -0.00000065, 0.00003863, -0.00000101, 0.00000996, -0.00002258, -0.00000719, 0.00000890, -0.00001307, 0.00000302, -0.00000766, -0.00000947, 0.00000014, -0.00000376, 0.00000162, -0.00001440, 0.00000987, -0.00001576, 0.00000598, -0.00000323, -0.00000125, 0.00001356, -0.00000721, },
		{ -0.00000265, 0.00000252, 0.00000497, -0.00000521, 0.00000525, -0.00000238, 0.00000171, -0.00000676, -0.00000704, 0.00000931, 0.00000178, -0.00000418, 0.00000097, -0.00000148, -0.00000145, -0.00000003, -0.00000235, 0.00000201, -0.00000086, -0.00000255, -0.00000117, 0.00000712, 0.00000833, -0.00001159, -0.00000883, 0.00000741, 0.00000690, -0.00000742, -0.00000306, 0.00000796, 0.00000514, -0.00001957, -0.00000462, 0.00000544, -0.00000211, -0.00000827, -0.00000486, -0.00000068, 0.00000793, -0.00002601, 0.00000587, -0.00001193, 0.00001104, -0.00004967, 0.00000753, -0.00013619, -0.00000205, 0.00030749, 0.00018038, -0.00015306, -0.00004608, -0.00000696, -0.00001333, -0.00001055, -0.00001706, 0.00001529, -0.00001079, 0.00000331, -0.00000544, 0.00000645, -0.00000467, -0.00000140, -0.00000627, 0.00000689, 0.00000108, -0.00000483, -0.00000516, 0.00000303, 0.00000322, -0.00000566, 0.00000251, 0.00000606, 0.00000530, -0.00001267, -0.00000503, 0.00000670, 0.00000124, -0.00000405, -0.00000706, 0.00000493, -0.00000125, 0.00000317, 0.00000638, -0.00000420, -0.00000190, -0.00000019, -0.00000462, 0.00000497, 0.00000699, -0.00000901, -0.00000060, -0.00000177, -0.00000204, 0.00000299, 0.00000277, 0.00000068, -0.00000069, -0.00000456, -0.00000401, 0.00000061, -0.00000056, -0.00002483, 0.00000904, -0.00000410, 0.00003417, 0.00000619, 0.00002236, 0.00002441, -0.00000073, 0.00000060, -0.00002843, 0.00001195, -0.00000702, -0.00000515, 0.00000121, -0.00001310, -0.00000211, -0.00000413, 0.00000218, -0.00000820, -0.00000452, 0.00000358, -0.00001492, 0.00001204, -0.00001267, 0.00001360, 0.00000130, -0.00000308, },
		{ -0.00000476, 0.00000883, -0.00000322, 0.00000069, 0.00000237, -0.00000140, -0.00000001, -0.00001013, 0.00000045, 0.00000948, -0.00000406, -0.00000085, -0.00000015, -0.00000122, -0.00000145, -0.00000025, -0.00000206, 0.00000341, -0.00000442, 0.00000014, -0.00000095, 0.00001326, -0.00000289, -0.00001131, -0.00000432, 0.00001355, -0.00000342, -0.00000402, -0.00000204, 0.00001459, -0.00001149, -0.00001286, -0.00000149, 0.00000854, -0.00001281, 0.00000066, -0.00001320, 0.00001719, -0.00002005, -0.00000393, -0.00000942, 0.00000986, -0.00002643, -0.00001704, -0.00004602, -0.00013034, 0.00016675, 0.00029327, 0.00001706, -0.00015260, 0.00001147, -0.00003225, 0.00000758, -0.00003249, 0.00001261, -0.00000448, 0.00000017, -0.00000504, 0.00000422, -0.00000125, -0.00000078, -0.00000694, 0.00000163, 0.00000452, -0.00000058, -0.00000748, -0.00000045, 0.00000329, 0.00000027, -0.00000518, 0.00000722, 0.00000469, 0.00000065, -0.00001645, 0.00000551, 0.00000209, 0.00000201, -0.00000974, 0.00000039, 0.00000223, -0.00000038, 0.00000567, 0.00000321, -0.00000625, 0.00000139, -0.00000368, -0.00000153, 0.00000855, 0.00000002, -0.00000842, 0.00000231, -0.00000465, 0.00000141, 0.00000298, 0.00000190, 0.00000027, -0.00000274, -0.00000434, -0.00000383, 0.00000514, -0.00001556, -0.00001284, 0.00000638, 0.00000821, 0.00003518, -0.00000226, 0.00004175, 0.00000086, 0.00001000, -0.00002415, -0.00000717, 0.00000726, -0.00001166, 0.00000188, -0.00000648, -0.00001053, 0.00000013, -0.00000492, 0.00000383, -0.00001539, 0.00000774, -0.00001091, -0.00000093, 0.00000273, -0.00000649, 0.00001651, -0.00000696, 0.00000067, },
		{ 0.00000057, 0.00000664, -0.00000570, 0.00000497, -0.00000128, 0.00000072, -0.00000462, -0.00000876, 0.00000791, 0.00000356, -0.00000469, 0.00000081, -0.00000078, -0.00000161, -0.00000056, -0.00000187, 0.00000114, 0.00000004, -0.00000320, -0.00000048, 0.00000498, 0.00001023, -0.00001024, -0.00000940, 0.00000428, 0.00000986, -0.00000768, -0.00000233, 0.00000463, 0.00000868, -0.00001967, -0.00000550, 0.00000354, 0.00000137, -0.00001151, -0.00000143, -0.00000546, 0.00001255, -0.00002962, 0.00000925, -0.00001573, 0.00001530, -0.00005094, 0.00001128, -0.00013147, -0.00000089, 0.00025364, 0.00020587, -0.00011381, -0.00007452, 0.00000432, -0.00002223, -0.00000249, -0.00002362, 0.00001867, -0.00001301, 0.00000477, -0.00000643, 0.00000729, -0.00000498, -0.00000081, -0.00000683, 0.00000642, 0.00000159, -0.00000362, -0.00000583, 0.00000196, 0.00000334, -0.00000432, 0.00000102, 0.00000636, 0.00000537, -0.00000944, -0.00000883, 0.00000739, 0.00000110, -0.00000210, -0.00000878, 0.00000448, -0.00000052, 0.00000192, 0.00000707, -0.00000249, -0.00000336, 0.00000049, -0.00000464, 0.00000275, 0.00000870, -0.00000780, -0.00000203, -0.00000080, -0.00000289, 0.00000263, 0.00000339, -0.00000009, 0.00000074, -0.00000580, -0.00000258, -0.00000227, 0.00000326, -0.00002632, 0.00000344, -0.00000173, 0.00003257, 0.00001258, 0.00001892, 0.00003089, -0.00000245, 0.00000176, -0.00003115, 0.00001289, -0.00000900, -0.00000293, -0.00000098, -0.00001065, -0.00000500, -0.00000237, -0.00000036, -0.00000336, -0.00001069, 0.00000868, -0.00001772, 0.00001120, -0.00001028, 0.00000936, 0.00000586, -0.00000506, -0.00000183, },
		{ 0.00000782, -0.00000139, -0.00000087, 0.00000356, -0.00000166, 0.00000025, -0.00000867, -0.00000261, 0.00001035, -0.00000337, -0.00000132, -0.00000012, -0.00000032, -0.00000249, 0.00000036, -0.00000282, 0.00000384, -0.00000465, 0.00000061, -0.00000195, 0.00001272, -0.00000088, -0.00001018, -0.00000718, 0.00001337, -0.00000147, -0.00000382, -0.00000335, 0.00001369, -0.00000831, -0.00001377, -0.00000350, 0.00000973, -0.00001234, 0.00000000, -0.00001264, 0.00001500, -0.00001670, -0.00000716, -0.00000633, 0.00000621, -0.00002081, -0.00001996, -0.00003897, -0.00012586, 0.00014573, 0.00025445, 0.00007645, -0.00015759, 0.00000194, -0.00002841, 0.00000457, -0.00002852, 0.00000755, -0.00000092, -0.00000248, -0.00000316, 0.00000296, -0.00000017, -0.00000110, -0.00000659, 0.00000095, 0.00000411, 0.00000096, -0.00000759, -0.00000077, 0.00000163, 0.00000232, -0.00000620, 0.00000741, 0.00000361, 0.00000362, -0.00001685, 0.00000321, 0.00000274, 0.00000280, -0.00000895, -0.00000163, 0.00000280, -0.00000032, 0.00000433, 0.00000549, -0.00000646, 0.00000078, -0.00000295, -0.00000254, 0.00000701, 0.00000322, -0.00000967, 0.00000258, -0.00000467, 0.00000077, 0.00000289, 0.00000235, -0.00000026, -0.00000138, -0.00000581, -0.00000239, 0.00000202, -0.00000911, -0.00002112, 0.00000737, 0.00000604, 0.00003993, -0.00000185, 0.00004322, 0.00000455, 0.00000853, -0.00002432, -0.00000864, 0.00000705, -0.00001173, 0.00000216, -0.00000667, -0.00001023, -0.00000095, -0.00000504, 0.00000434, -0.00001390, 0.00000319, -0.00000447, -0.00000805, 0.00000780, -0.00001021, 0.00001754, -0.00000521, -0.00000046, -0.00000490, },
		{ 0.00000776, -0.00000554, 0.00000417, 0.00000007, -0.00000030, -0.00000285, -0.00000953, 0.00000578, 0.00000538, -0.00000503, 0.00000062, -0.00000036, -0.00000128, -0.00000140, -0.00000118, 0.00000007, 0.00000107, -0.00000402, 0.00000048, 0.00000277, 0.00001178, -0.00000889, -0.00000914, 0.00000058, 0.00001256, -0.00000770, -0.00000137, 0.00000145, 0.00001127, -0.00001890, -0.00000633, 0.00000119, 0.00000490, -0.00001436, 0.00000144, -0.00000942, 0.00001600, -0.00003185, 0.00001113, -0.00001790, 0.00001764, -0.00005043, 0.00001328, -0.00012551, -0.00000146, 0.00021223, 0.00021042, -0.00006424, -0.00010156, 0.00001273, -0.00002911, 0.00000368, -0.00002812, 0.00002019, -0.00001377, 0.00000512, -0.00000666, 0.00000760, -0.00000493, -0.00000038, -0.00000728, 0.00000594, 0.00000178, -0.00000212, -0.00000656, 0.00000134, 0.00000263, -0.00000243, -0.00000072, 0.00000703, 0.00000463, -0.00000561, -0.00001228, 0.00000756, 0.00000095, -0.00000014, -0.00001002, 0.00000355, 0.00000028, 0.00000101, 0.00000701, -0.00000029, -0.00000472, 0.00000092, -0.00000441, 0.00000075, 0.00000945, -0.00000574, -0.00000379, 0.00000029, -0.00000370, 0.00000221, 0.00000386, -0.00000053, 0.00000160, -0.00000621, -0.00000215, -0.00000365, 0.00000456, -0.00002426, -0.00000472, 0.00000114, 0.00003012, 0.00001981, 0.00001529, 0.00003764, -0.00000398, 0.00000272, -0.00003373, 0.00001355, -0.00001068, -0.00000115, -0.00000265, -0.00000877, -0.00000715, -0.00000109, -0.00000276, 0.00000097, -0.00001524, 0.00001109, -0.00001737, 0.00000764, -0.00000601, 0.00000409, 0.00001036, -0.00000648, -0.00000131, -0.00000116, },
		{ 0.00000065, -0.00000226, 0.00000436, -0.00000147, -0.00000001, -0.00000683, -0.00000526, 0.00001043, -0.00000225, -0.00000187, -0.00000013, 0.00000046, -0.00000309, 0.00000053, -0.00000315, 0.00000381, -0.00000446, 0.00000062, -0.00000223, 0.00001142, 0.00000131, -0.00000911, -0.00000903, 0.00001175, 0.00000136, -0.00000425, -0.00000345, 0.00001146, -0.00000458, -0.00001468, -0.00000494, 0.00000955, -0.00001028, -0.00000216, -0.00001050, 0.00001119, -0.00001190, -0.00001170, -0.00000195, 0.00000124, -0.00001409, -0.00002423, -0.00003027, -0.00012478, 0.00013287, 0.00021380, 0.00012170, -0.00014664, -0.00001555, -0.00002045, -0.00000156, -0.00002216, 0.00000103, 0.00000344, -0.00000559, -0.00000101, 0.00000145, 0.00000114, -0.00000162, -0.00000597, 0.00000008, 0.00000380, 0.00000213, -0.00000721, -0.00000122, 0.00000024, 0.00000357, -0.00000636, 0.00000700, 0.00000304, 0.00000549, -0.00001575, 0.00000008, 0.00000362, 0.00000309, -0.00000746, -0.00000386, 0.00000314, -0.00000003, 0.00000297, 0.00000718, -0.00000588, -0.00000021, -0.00000217, -0.00000317, 0.00000507, 0.00000606, -0.00001017, 0.00000230, -0.00000433, 0.00000003, 0.00000261, 0.00000309, -0.00000115, 0.00000032, -0.00000755, -0.00000058, -0.00000141, -0.00000320, -0.00002683, 0.00000502, 0.00000559, 0.00004288, 0.00000070, 0.00004298, 0.00001014, 0.00000558, -0.00002318, -0.00001142, 0.00000803, -0.00001299, 0.00000353, -0.00000786, -0.00000898, -0.00000270, -0.00000437, 0.00000341, -0.00001055, -0.00000274, 0.00000219, -0.00001408, 0.00001101, -0.00001179, 0.00001645, -0.00000209, -0.00000213, -0.00000457, 0.00000626, },
		{ -0.00000473, 0.00000299, 0.00000151, -0.00000116, -0.00000159, -0.00000940, 0.00000319, 0.00000699, -0.00000510, 0.00000039, -0.00000018, -0.00000065, -0.00000230, -0.00000048, -0.00000100, 0.00000203, -0.00000484, 0.00000148, 0.00000077, 0.00001275, -0.00000745, -0.00000834, -0.00000313, 0.00001442, -0.00000703, -0.00000072, -0.00000096, 0.00001247, -0.00001710, -0.00000728, -0.00000116, 0.00000784, -0.00001613, 0.00000308, -0.00001189, 0.00001762, -0.00003211, 0.00001098, -0.00001790, 0.00001757, -0.00004761, 0.00001280, -0.00011718, -0.00000532, 0.00018319, 0.00019852, -0.00001120, -0.00012215, 0.00001554, -0.00003228, 0.00000674, -0.00002971, 0.00001934, -0.00001279, 0.00000419, -0.00000602, 0.00000729, -0.00000443, -0.00000024, -0.00000749, 0.00000534, 0.00000179, -0.00000052, -0.00000721, 0.00000103, 0.00000140, -0.00000038, -0.00000237, 0.00000777, 0.00000347, -0.00000175, -0.00001489, 0.00000696, 0.00000098, 0.00000157, -0.00001058, 0.00000211, 0.00000110, 0.00000045, 0.00000636, 0.00000207, -0.00000573, 0.00000100, -0.00000396, -0.00000092, 0.00000927, -0.00000307, -0.00000560, 0.00000131, -0.00000438, 0.00000176, 0.00000408, -0.00000047, 0.00000171, -0.00000565, -0.00000283, -0.00000345, 0.00000352, -0.00001949, -0.00001392, 0.00000305, 0.00002778, 0.00002699, 0.00001232, 0.00004388, -0.00000450, 0.00000271, -0.00003544, 0.00001327, -0.00001145, -0.00000035, -0.00000332, -0.00000792, -0.00000821, -0.00000061, -0.00000455, 0.00000413, -0.00001752, 0.00001057, -0.00001418, 0.00000212, -0.00000077, -0.00000126, 0.00001400, -0.00000686, -0.00000133, -0.00000246, 0.00000818, },
		{ -0.00000326, 0.00000466, -0.00000084, -0.00000064, -0.00000490, -0.00000727, 0.00000968, -0.00000081, -0.00000244, -0.00000015, 0.00000102, -0.00000323, 0.00000031, -0.00000312, 0.00000336, -0.00000391, 0.00000018, -0.00000188, 0.00000955, 0.00000352, -0.00000817, -0.00000982, 0.00000889, 0.00000471, -0.00000502, -0.00000272, 0.00000852, -0.00000097, -0.00001519, -0.00000602, 0.00000828, -0.00000713, -0.00000515, -0.00000757, 0.00000668, -0.00000666, -0.00001646, 0.00000255, -0.00000379, -0.00000757, -0.00002842, -0.00002158, -0.00012469, 0.00012382, 0.00017689, 0.00015010, -0.00012133, -0.00003804, -0.00001074, -0.00000896, -0.00001499, -0.00000559, 0.00000751, -0.00000836, 0.00000081, 0.00000012, 0.00000235, -0.00000210, -0.00000527, -0.00000085, 0.00000353, 0.00000293, -0.00000639, -0.00000184, -0.00000073, 0.00000394, -0.00000572, 0.00000607, 0.00000304, 0.00000624, -0.00001336, -0.00000352, 0.00000450, 0.00000301, -0.00000548, -0.00000602, 0.00000313, 0.00000046, 0.00000173, 0.00000818, -0.00000460, -0.00000143, -0.00000146, -0.00000344, 0.00000296, 0.00000827, -0.00000980, 0.00000148, -0.00000365, -0.00000079, 0.00000221, 0.00000392, -0.00000205, 0.00000191, -0.00000899, 0.00000089, -0.00000420, 0.00000102, -0.00002904, -0.00000068, 0.00000639, 0.00004431, 0.00000511, 0.00004136, 0.00001733, 0.00000155, -0.00002122, -0.00001501, 0.00000966, -0.00001486, 0.00000540, -0.00000949, -0.00000736, -0.00000460, -0.00000335, 0.00000154, -0.00000617, -0.00000884, 0.00000778, -0.00001798, 0.00001175, -0.00001106, 0.00001341, 0.00000199, -0.00000397, -0.00000402, 0.00000441, 0.00000259, },
		{ 0.00000159, 0.00000284, -0.00000172, -0.00000091, -0.00000853, 0.00000045, 0.00000816, -0.00000484, 0.00000010, -0.00000016, 0.00000010, -0.00000307, 0.00000008, -0.00000194, 0.00000278, -0.00000548, 0.00000229, -0.00000077, 0.00001303, -0.00000582, -0.00000728, -0.00000635, 0.00001504, -0.00000542, -0.00000068, -0.00000226, 0.00001217, -0.00001439, -0.00000831, -0.00000329, 0.00000980, -0.00001648, 0.00000322, -0.00001267, 0.00001730, -0.00003036, 0.00000885, -0.00001586, 0.00001527, -0.00004267, 0.00000998, -0.00010649, -0.00001245, 0.00016451, 0.00017587, 0.00003873, -0.00013245, 0.00001133, -0.00003115, 0.00000633, -0.00002831, 0.00001625, -0.00001027, 0.00000219, -0.00000469, 0.00000644, -0.00000353, -0.00000036, -0.00000742, 0.00000458, 0.00000168, 0.00000106, -0.00000765, 0.00000086, -0.00000003, 0.00000145, -0.00000365, 0.00000835, 0.00000227, 0.00000167, -0.00001631, 0.00000546, 0.00000130, 0.00000284, -0.00001035, 0.00000022, 0.00000185, 0.00000022, 0.00000533, 0.00000429, -0.00000619, 0.00000071, -0.00000337, -0.00000220, 0.00000829, -0.00000011, -0.00000719, 0.00000210, -0.00000485, 0.00000129, 0.00000397, 0.00000009, 0.00000112, -0.00000425, -0.00000441, -0.00000203, 0.00000084, -0.00001331, -0.00002237, 0.00000269, 0.00002625, 0.00003345, 0.00001065, 0.00004901, -0.00000337, 0.00000118, -0.00003582, 0.00001164, -0.00001099, -0.00000080, -0.00000280, -0.00000822, -0.00000811, -0.00000099, -0.00000554, 0.00000579, -0.00001740, 0.00000745, -0.00000898, -0.00000427, 0.00000434, -0.00000575, 0.00001613, -0.00000591, -0.00000198, -0.00000323, 0.00000788, -0.00000337, },
		{ 0.00000443, 0.00000015, -0.00000148, -0.00000313, -0.00000849, 0.00000822, 0.00000078, -0.00000293, -0.00000018, 0.00000131, -0.00000295, -0.00000017, -0.00000283, 0.00000261, -0.00000309, -0.00000058, -0.00000106, 0.00000735, 0.00000559, -0.00000733, -0.00000966, 0.00000518, 0.00000806, -0.00000568, -0.00000170, 0.00000557, 0.00000190, -0.00001494, -0.00000695, 0.00000632, -0.00000354, -0.00000820, -0.00000467, 0.00000236, -0.00000196, -0.00002046, 0.00000619, -0.00000786, -0.00000232, -0.00003141, -0.00001428, -0.00012363, 0.00011527, 0.00014719, 0.00016168, -0.00008532, -0.00006158, -0.00000202, -0.00001566, -0.00000860, -0.00001106, 0.00001039, -0.00001014, 0.00000184, -0.00000071, 0.00000323, -0.00000236, -0.00000464, -0.00000170, 0.00000321, 0.00000342, -0.00000518, -0.00000261, -0.00000124, 0.00000352, -0.00000448, 0.00000476, 0.00000350, 0.00000604, -0.00001006, -0.00000713, 0.00000511, 0.00000275, -0.00000329, -0.00000784, 0.00000264, 0.00000110, 0.00000074, 0.00000849, -0.00000281, -0.00000264, -0.00000093, -0.00000342, 0.00000089, 0.00000968, -0.00000859, 0.00000022, -0.00000272, -0.00000166, 0.00000177, 0.00000461, -0.00000266, 0.00000298, -0.00000968, 0.00000149, -0.00000567, 0.00000287, -0.00002765, -0.00000896, 0.00000739, 0.00004489, 0.00001081, 0.00003895, 0.00002554, -0.00000288, -0.00001914, -0.00001869, 0.00001123, -0.00001667, 0.00000711, -0.00001091, -0.00000600, -0.00000612, -0.00000243, -0.00000067, -0.00000168, -0.00001399, 0.00001131, -0.00001920, 0.00000996, -0.00000827, 0.00000896, 0.00000640, -0.00000551, -0.00000356, 0.00000256, 0.00000418, -0.00000370, },
		{ 0.00000389, -0.00000188, -0.00000079, -0.00000713, -0.00000214, 0.00000875, -0.00000423, -0.00000024, -0.00000023, 0.00000081, -0.00000357, 0.00000042, -0.00000266, 0.00000325, -0.00000587, 0.00000279, -0.00000172, 0.00001257, -0.00000400, -0.00000621, -0.00000871, 0.00001423, -0.00000283, -0.00000130, -0.00000245, 0.00001063, -0.00001117, -0.00000920, -0.00000513, 0.00001065, -0.00001540, 0.00000201, -0.00001198, 0.00001536, -0.00002706, 0.00000526, -0.00001237, 0.00001141, -0.00003635, 0.00000554, -0.00009420, -0.00002175, 0.00015314, 0.00014832, 0.00008021, -0.00013051, 0.00000027, -0.00002624, 0.00000302, -0.00002453, 0.00001156, -0.00000680, -0.00000040, -0.00000304, 0.00000532, -0.00000240, -0.00000066, -0.00000713, 0.00000371, 0.00000147, 0.00000250, -0.00000779, 0.00000069, -0.00000138, 0.00000277, -0.00000436, 0.00000859, 0.00000131, 0.00000427, -0.00001638, 0.00000313, 0.00000187, 0.00000361, -0.00000933, -0.00000200, 0.00000241, 0.00000028, 0.00000413, 0.00000613, -0.00000604, 0.00000009, -0.00000274, -0.00000305, 0.00000671, 0.00000283, -0.00000832, 0.00000252, -0.00000506, 0.00000080, 0.00000359, 0.00000107, 0.00000004, -0.00000235, -0.00000645, -0.00000001, -0.00000251, -0.00000721, -0.00002850, -0.00000081, 0.00002572, 0.00003893, 0.00001059, 0.00005278, -0.00000030, -0.00000208, -0.00003479, 0.00000866, -0.00000934, -0.00000239, -0.00000124, -0.00000949, -0.00000710, -0.00000201, -0.00000577, 0.00000596, -0.00001518, 0.00000250, -0.00000291, -0.00001033, 0.00000835, -0.00000864, 0.00001637, -0.00000363, -0.00000318, -0.00000353, 0.00000694, -0.00000166, 0.00000021, },
		{ 0.00000134, -0.00000231, -0.00000171, -0.00000890, 0.00000623, 0.00000231, -0.00000327, -0.00000020, 0.00000137, -0.00000237, -0.00000076, -0.00000241, 0.00000167, -0.00000214, -0.00000155, 0.00000005, 0.00000508, 0.00000731, -0.00000650, -0.00000880, 0.00000116, 0.00001085, -0.00000577, -0.00000090, 0.00000323, 0.00000357, -0.00001378, -0.00000785, 0.00000400, -0.00000007, -0.00001065, -0.00000247, -0.00000107, 0.00000154, -0.00002305, 0.00000834, -0.00001032, 0.00000104, -0.00003254, -0.00000920, -0.00012040, 0.00010521, 0.00012599, 0.00015871, -0.00004358, -0.00008199, 0.00000323, -0.00002005, -0.00000421, -0.00001449, 0.00001148, -0.00001056, 0.00000188, -0.00000089, 0.00000364, -0.00000229, -0.00000419, -0.00000236, 0.00000279, 0.00000364, -0.00000370, -0.00000347, -0.00000136, 0.00000250, -0.00000293, 0.00000332, 0.00000427, 0.00000515, -0.00000632, -0.00001030, 0.00000518, 0.00000248, -0.00000116, -0.00000907, 0.00000162, 0.00000184, 0.00000002, 0.00000822, -0.00000076, -0.00000365, -0.00000066, -0.00000315, -0.00000096, 0.00001020, -0.00000666, -0.00000130, -0.00000166, -0.00000254, 0.00000138, 0.00000501, -0.00000275, 0.00000329, -0.00000938, 0.00000096, -0.00000559, 0.00000231, -0.00002326, -0.00001837, 0.00000726, 0.00004532, 0.00001715, 0.00003641, 0.00003412, -0.00000695, -0.00001773, -0.00002174, 0.00001208, -0.00001778, 0.00000810, -0.00001160, -0.00000535, -0.00000687, -0.00000195, -0.00000273, 0.00000218, -0.00001739, 0.00001228, -0.00001771, 0.00000606, -0.00000409, 0.00000389, 0.00001040, -0.00000628, -0.00000344, 0.00000093, 0.00000520, -0.00000351, 0.00000374, },
		{ -0.00000160, -0.00000113, -0.00000544, -0.00000434, 0.00000869, -0.00000335, -0.00000061, -0.00000033, 0.00000135, -0.00000373, 0.00000054, -0.00000316, 0.00000339, -0.00000594, 0.00000288, -0.00000204, 0.00001143, -0.00000202, -0.00000528, -0.00001003, 0.00001202, 0.00000050, -0.00000238, -0.00000181, 0.00000836, -0.00000802, -0.00000964, -0.00000675, 0.00001048, -0.00001318, -0.00000013, -0.00001032, 0.00001243, -0.00002293, 0.00000101, -0.00000826, 0.00000692, -0.00002959, 0.00000048, -0.00008148, -0.00003162, 0.00014581, 0.00012093, 0.00010981, -0.00011645, -0.00001598, -0.00001907, -0.00000199, -0.00001947, 0.00000627, -0.00000319, -0.00000294, -0.00000149, 0.00000423, -0.00000125, -0.00000098, -0.00000673, 0.00000282, 0.00000116, 0.00000375, -0.00000758, 0.00000040, -0.00000241, 0.00000340, -0.00000447, 0.00000843, 0.00000076, 0.00000590, -0.00001514, 0.00000017, 0.00000256, 0.00000392, -0.00000768, -0.00000430, 0.00000267, 0.00000061, 0.00000291, 0.00000745, -0.00000529, -0.00000073, -0.00000216, -0.00000349, 0.00000474, 0.00000543, -0.00000882, 0.00000252, -0.00000498, 0.00000026, 0.00000302, 0.00000223, -0.00000119, -0.00000041, -0.00000835, 0.00000189, -0.00000553, -0.00000248, -0.00003123, -0.00000753, 0.00002576, 0.00004357, 0.00001209, 0.00005523, 0.00000468, -0.00000690, -0.00003261, 0.00000466, -0.00000688, -0.00000469, 0.00000094, -0.00001131, -0.00000560, -0.00000331, -0.00000550, 0.00000495, -0.00001149, -0.00000324, 0.00000283, -0.00001502, 0.00001050, -0.00000955, 0.00001469, -0.00000026, -0.00000465, -0.00000353, 0.00000556, 0.00000015, -0.00000093, 0.00000453, },
		{ -0.00000299, -0.00000077, -0.00000856, 0.00000393, 0.00000362, -0.00000340, -0.00000023, 0.00000123, -0.00000165, -0.00000130, -0.00000196, 0.00000065, -0.00000116, -0.00000259, 0.00000122, 0.00000299, 0.00000851, -0.00000558, -0.00000751, -0.00000266, 0.00001258, -0.00000493, -0.00000070, 0.00000188, 0.00000391, -0.00001182, -0.00000865, 0.00000155, 0.00000286, -0.00001210, -0.00000133, -0.00000329, 0.00000355, -0.00002402, 0.00000884, -0.00001107, 0.00000244, -0.00003171, -0.00000653, -0.00011462, 0.00009299, 0.00011279, 0.00014500, -0.00000146, -0.00009569, 0.00000336, -0.00002125, -0.00000242, -0.00001553, 0.00001070, -0.00000966, 0.00000100, -0.00000051, 0.00000362, -0.00000190, -0.00000393, -0.00000279, 0.00000225, 0.00000361, -0.00000203, -0.00000432, -0.00000120, 0.00000115, -0.00000138, 0.00000197, 0.00000514, 0.00000389, -0.00000263, -0.00001262, 0.00000455, 0.00000236, 0.00000068, -0.00000956, 0.00000006, 0.00000258, -0.00000039, 0.00000752, 0.00000130, -0.00000429, -0.00000068, -0.00000273, -0.00000247, 0.00000988, -0.00000422, -0.00000288, -0.00000061, -0.00000337, 0.00000107, 0.00000500, -0.00000224, 0.00000279, -0.00000811, -0.00000058, -0.00000415, -0.00000012, -0.00001709, -0.00002717, 0.00000474, 0.00004615, 0.00002363, 0.00003431, 0.00004249, -0.00000999, -0.00001760, -0.00002363, 0.00001173, -0.00001781, 0.00000803, -0.00001131, -0.00000566, -0.00000669, -0.00000205, -0.00000432, 0.00000489, -0.00001864, 0.00001070, -0.00001402, 0.00000084, 0.00000059, -0.00000094, 0.00001332, -0.00000595, -0.00000380, -0.00000032, 0.00000556, -0.00000274, 0.00000274, 0.00000257, },
		{ -0.00000179, -0.00000371, -0.00000595, 0.00000800, -0.00000228, -0.00000095, -0.00000042, 0.00000168, -0.00000357, 0.00000051, -0.00000346, 0.00000323, -0.00000570, 0.00000257, -0.00000175, 0.00000975, 0.00000001, -0.00000454, -0.00001030, 0.00000872, 0.00000413, -0.00000350, -0.00000083, 0.00000601, -0.00000550, -0.00000935, -0.00000824, 0.00000949, -0.00001025, -0.00000264, -0.00000834, 0.00000921, -0.00001873, -0.00000310, -0.00000437, 0.00000266, -0.00002328, -0.00000424, -0.00006947, -0.00004054, 0.00013969, 0.00009732, 0.00012628, -0.00009228, -0.00003464, -0.00001168, -0.00000719, -0.00001440, 0.00000141, -0.00000021, -0.00000489, -0.00000044, 0.00000341, -0.00000028, -0.00000117, -0.00000634, 0.00000200, 0.00000071, 0.00000477, -0.00000699, -0.00000007, -0.00000303, 0.00000335, -0.00000406, 0.00000791, 0.00000067, 0.00000654, -0.00001280, -0.00000309, 0.00000319, 0.00000388, -0.00000559, -0.00000643, 0.00000247, 0.00000116, 0.00000180, 0.00000820, -0.00000407, -0.00000159, -0.00000172, -0.00000356, 0.00000262, 0.00000747, -0.00000860, 0.00000210, -0.00000461, -0.00000036, 0.00000240, 0.00000333, -0.00000222, 0.00000112, -0.00000961, 0.00000303, -0.00000742, 0.00000004, -0.00003023, -0.00001676, 0.00002538, 0.00004784, 0.00001484, 0.00005669, 0.00001126, -0.00001282, -0.00002985, 0.00000021, -0.00000420, -0.00000716, 0.00000319, -0.00001316, -0.00000412, -0.00000447, -0.00000506, 0.00000320, -0.00000712, -0.00000872, 0.00000719, -0.00001761, 0.00001050, -0.00000845, 0.00001139, 0.00000372, -0.00000601, -0.00000347, 0.00000397, 0.00000180, -0.00000165, 0.00000469, -0.00000092, },
		{ -0.00000034, -0.00000761, 0.00000158, 0.00000455, -0.00000331, -0.00000026, 0.00000098, -0.00000092, -0.00000167, -0.00000153, -0.00000037, -0.00000025, -0.00000356, 0.00000228, 0.00000131, 0.00000908, -0.00000449, -0.00000609, -0.00000580, 0.00001293, -0.00000302, -0.00000123, 0.00000160, 0.00000308, -0.00000941, -0.00000916, -0.00000093, 0.00000502, -0.00001240, -0.00000127, -0.00000430, 0.00000415, -0.00002355, 0.00000794, -0.00001039, 0.00000222, -0.00002929, -0.00000591, -0.00010666, 0.00007898, 0.00010584, 0.00012499, 0.00003629, -0.00010036, -0.00000208, -0.00001928, -0.00000310, -0.00001445, 0.00000839, -0.00000779, -0.00000048, 0.00000021, 0.00000333, -0.00000128, -0.00000381, -0.00000303, 0.00000163, 0.00000335, -0.00000025, -0.00000507, -0.00000092, -0.00000025, -0.00000011, 0.00000089, 0.00000594, 0.00000256, 0.00000061, -0.00001384, 0.00000315, 0.00000245, 0.00000208, -0.00000921, -0.00000193, 0.00000323, -0.00000050, 0.00000654, 0.00000315, -0.00000446, -0.00000094, -0.00000226, -0.00000356, 0.00000882, -0.00000152, -0.00000429, 0.00000032, -0.00000410, 0.00000085, 0.00000461, -0.00000119, 0.00000164, -0.00000617, -0.00000275, -0.00000192, -0.00000353, -0.00001062, -0.00003368, -0.00000100, 0.00004742, 0.00003006, 0.00003299, 0.00005030, -0.00001157, -0.00001911, -0.00002411, 0.00001003, -0.00001667, 0.00000689, -0.00001005, -0.00000687, -0.00000568, -0.00000267, -0.00000533, 0.00000626, -0.00001777, 0.00000705, -0.00000902, -0.00000468, 0.00000487, -0.00000475, 0.00001468, -0.00000439, -0.00000464, -0.00000118, 0.00000528, -0.00000150, 0.00000161, 0.00000366, -0.00000336, },
		{ -0.00000214, -0.00000689, 0.00000679, -0.00000117, -0.00000123, -0.00000047, 0.00000176, -0.00000317, 0.00000041, -0.00000359, 0.00000280, -0.00000519, 0.00000188, -0.00000097, 0.00000773, 0.00000193, -0.00000396, -0.00000967, 0.00000477, 0.00000750, -0.00000419, -0.00000002, 0.00000418, -0.00000405, -0.00000821, -0.00000962, 0.00000791, -0.00000707, -0.00000497, -0.00000658, 0.00000629, -0.00001509, -0.00000647, -0.00000129, -0.00000076, -0.00001803, -0.00000796, -0.00005898, -0.00004738, 0.00013280, 0.00007947, 0.00013040, -0.00006144, -0.00005243, -0.00000617, -0.00001115, -0.00001041, -0.00000217, 0.00000155, -0.00000588, -0.00000010, 0.00000302, 0.00000040, -0.00000115, -0.00000603, 0.00000134, 0.00000011, 0.00000553, -0.00000602, -0.00000074, -0.00000321, 0.00000270, -0.00000329, 0.00000715, 0.00000100, 0.00000632, -0.00000971, -0.00000626, 0.00000351, 0.00000365, -0.00000335, -0.00000811, 0.00000173, 0.00000188, 0.00000088, 0.00000844, -0.00000258, -0.00000232, -0.00000148, -0.00000334, 0.00000054, 0.00000879, -0.00000768, 0.00000132, -0.00000401, -0.00000107, 0.00000184, 0.00000413, -0.00000275, 0.00000191, -0.00000989, 0.00000309, -0.00000781, 0.00000011, -0.00002600, -0.00002710, 0.00002328, 0.00005228, 0.00001846, 0.00005759, 0.00001898, -0.00001923, -0.00002716, -0.00000406, -0.00000189, -0.00000923, 0.00000499, -0.00001457, -0.00000311, -0.00000511, -0.00000474, 0.00000117, -0.00000283, -0.00001303, 0.00000949, -0.00001784, 0.00000847, -0.00000575, 0.00000708, 0.00000767, -0.00000683, -0.00000357, 0.00000241, 0.00000305, -0.00000181, 0.00000437, 0.00000005, -0.00000261, },
		{ -0.00000623, -0.00000060, 0.00000501, -0.00000302, -0.00000028, 0.00000067, -0.00000031, -0.00000184, -0.00000113, -0.00000136, 0.00000050, -0.00000437, 0.00000307, 0.00000017, 0.00000896, -0.00000319, -0.00000478, -0.00000796, 0.00001180, -0.00000017, -0.00000236, 0.00000215, 0.00000155, -0.00000707, -0.00000909, -0.00000343, 0.00000637, -0.00001166, -0.00000207, -0.00000442, 0.00000374, -0.00002211, 0.00000617, -0.00000886, 0.00000101, -0.00002593, -0.00000668, -0.00009729, 0.00006414, 0.00010286, 0.00010292, 0.00006612, -0.00009525, -0.00001229, -0.00001502, -0.00000551, -0.00001197, 0.00000526, -0.00000554, -0.00000210, 0.00000094, 0.00000298, -0.00000058, -0.00000375, -0.00000315, 0.00000103, 0.00000282, 0.00000154, -0.00000560, -0.00000067, -0.00000148, 0.00000069, 0.00000018, 0.00000656, 0.00000141, 0.00000311, -0.00001384, 0.00000108, 0.00000270, 0.00000299, -0.00000810, -0.00000418, 0.00000363, -0.00000033, 0.00000544, 0.00000464, -0.00000418, -0.00000135, -0.00000182, -0.00000420, 0.00000719, 0.00000115, -0.00000536, 0.00000101, -0.00000467, 0.00000066, 0.00000393, 0.00000020, 0.00000018, -0.00000400, -0.00000499, 0.00000039, -0.00000689, -0.00000525, -0.00003670, -0.00001008, 0.00004859, 0.00003661, 0.00003247, 0.00005748, -0.00001153, -0.00002231, -0.00002329, 0.00000716, -0.00001458, 0.00000493, -0.00000809, -0.00000870, -0.00000414, -0.00000359, -0.00000585, 0.00000639, -0.00001518, 0.00000214, -0.00000375, -0.00000952, 0.00000796, -0.00000701, 0.00001427, -0.00000171, -0.00000578, -0.00000173, 0.00000447, -0.00000002, 0.00000056, 0.00000445, -0.00000335, -0.00000041, },
		{ -0.00000717, 0.00000521, -0.00000015, -0.00000141, -0.00000047, 0.00000165, -0.00000264, 0.00000035, -0.00000360, 0.00000213, -0.00000446, 0.00000092, 0.00000013, 0.00000561, 0.00000357, -0.00000343, -0.00000843, 0.00000072, 0.00001006, -0.00000402, 0.00000017, 0.00000330, -0.00000384, -0.00000633, -0.00001078, 0.00000588, -0.00000400, -0.00000673, -0.00000541, 0.00000405, -0.00001235, -0.00000880, 0.00000073, -0.00000311, -0.00001407, -0.00001042, -0.00005040, -0.00005164, 0.00012418, 0.00006779, 0.00012448, -0.00002805, -0.00006624, -0.00000414, -0.00001295, -0.00000818, -0.00000404, 0.00000189, -0.00000583, -0.00000046, 0.00000302, 0.00000078, -0.00000090, -0.00000584, 0.00000088, -0.00000061, 0.00000598, -0.00000470, -0.00000156, -0.00000302, 0.00000166, -0.00000241, 0.00000629, 0.00000163, 0.00000547, -0.00000629, -0.00000894, 0.00000335, 0.00000340, -0.00000121, -0.00000912, 0.00000040, 0.00000271, 0.00000018, 0.00000827, -0.00000102, -0.00000279, -0.00000147, -0.00000292, -0.00000133, 0.00000933, -0.00000616, 0.00000031, -0.00000324, -0.00000187, 0.00000145, 0.00000450, -0.00000262, 0.00000183, -0.00000913, 0.00000204, -0.00000677, -0.00000188, -0.00001966, -0.00003673, 0.00001818, 0.00005724, 0.00002268, 0.00005832, 0.00002743, -0.00002555, -0.00002513, -0.00000761, -0.00000042, -0.00001048, 0.00000601, -0.00001524, -0.00000281, -0.00000507, -0.00000472, -0.00000079, 0.00000080, -0.00001562, 0.00000950, -0.00001594, 0.00000494, -0.00000212, 0.00000249, 0.00001093, -0.00000676, -0.00000400, 0.00000105, 0.00000376, -0.00000139, 0.00000367, 0.00000119, -0.00000342, -0.00000091, },
		{ -0.00000241, 0.00000500, -0.00000258, -0.00000029, 0.00000036, 0.00000011, -0.00000180, -0.00000073, -0.00000229, 0.00000106, -0.00000494, 0.00000348, -0.00000036, 0.00000820, -0.00000171, -0.00000370, -0.00000902, 0.00000936, 0.00000327, -0.00000371, 0.00000308, -0.00000005, -0.00000536, -0.00000820, -0.00000597, 0.00000697, -0.00001016, -0.00000334, -0.00000409, 0.00000283, -0.00002028, 0.00000415, -0.00000709, -0.00000054, -0.00002227, -0.00000813, -0.00008739, 0.00004964, 0.00010159, 0.00008217, 0.00008616, -0.00008129, -0.00002542, -0.00000995, -0.00000852, -0.00000905, 0.00000213, -0.00000358, -0.00000338, 0.00000136, 0.00000276, 0.00000005, -0.00000361, -0.00000324, 0.00000053, 0.00000201, 0.00000326, -0.00000584, -0.00000057, -0.00000235, 0.00000096, -0.00000018, 0.00000695, 0.00000058, 0.00000473, -0.00001271, -0.00000147, 0.00000302, 0.00000344, -0.00000638, -0.00000641, 0.00000363, 0.00000012, 0.00000434, 0.00000571, -0.00000354, -0.00000178, -0.00000153, -0.00000441, 0.00000518, 0.00000356, -0.00000594, 0.00000141, -0.00000503, 0.00000044, 0.00000311, 0.00000164, -0.00000122, -0.00000209, -0.00000672, 0.00000212, -0.00000933, -0.00000199, -0.00003575, -0.00002175, 0.00004855, 0.00004376, 0.00003260, 0.00006416, -0.00000996, -0.00002694, -0.00002152, 0.00000350, -0.00001198, 0.00000259, -0.00000585, -0.00001076, -0.00000246, -0.00000447, -0.00000607, 0.00000553, -0.00001147, -0.00000308, 0.00000079, -0.00001289, 0.00000940, -0.00000752, 0.00001220, 0.00000175, -0.00000690, -0.00000214, 0.00000331, 0.00000147, -0.00000018, 0.00000483, -0.00000294, -0.00000089, -0.00000464, },
		{ 0.00000347, 0.00000066, -0.00000147, -0.00000042, 0.00000139, -0.00000211, 0.00000040, -0.00000349, 0.00000126, -0.00000359, -0.00000022, 0.00000135, 0.00000364, 0.00000477, -0.00000283, -0.00000688, -0.00000292, 0.00001138, -0.00000274, -0.00000045, 0.00000347, -0.00000473, -0.00000406, -0.00001147, 0.00000347, -0.00000128, -0.00000772, -0.00000494, 0.00000256, -0.00001054, -0.00001014, 0.00000178, -0.00000450, -0.00001124, -0.00001172, -0.00004366, -0.00005335, 0.00011373, 0.00006149, 0.00011174, 0.00000387, -0.00007374, -0.00000636, -0.00001231, -0.00000783, -0.00000422, 0.00000098, -0.00000497, -0.00000132, 0.00000326, 0.00000096, -0.00000047, -0.00000576, 0.00000060, -0.00000139, 0.00000608, -0.00000308, -0.00000247, -0.00000259, 0.00000046, -0.00000164, 0.00000547, 0.00000244, 0.00000426, -0.00000295, -0.00001082, 0.00000257, 0.00000323, 0.00000061, -0.00000930, -0.00000147, 0.00000354, -0.00000027, 0.00000781, 0.00000042, -0.00000291, -0.00000165, -0.00000239, -0.00000286, 0.00000910, -0.00000420, -0.00000077, -0.00000239, -0.00000273, 0.00000123, 0.00000439, -0.00000185, 0.00000098, -0.00000755, 0.00000024, -0.00000478, -0.00000511, -0.00001274, -0.00004384, 0.00000922, 0.00006258, 0.00002752, 0.00005906, 0.00003634, -0.00003138, -0.00002417, -0.00001013, -0.00000004, -0.00001076, 0.00000615, -0.00001511, -0.00000327, -0.00000436, -0.00000500, -0.00000250, 0.00000343, -0.00001629, 0.00000745, -0.00001253, 0.00000071, 0.00000167, -0.00000162, 0.00001296, -0.00000558, -0.00000482, -0.00000002, 0.00000385, -0.00000045, 0.00000271, 0.00000233, -0.00000407, -0.00000008, -0.00000684, },
		{ 0.00000457, -0.00000209, -0.00000025, 0.00000010, 0.00000034, -0.00000162, -0.00000026, -0.00000314, 0.00000138, -0.00000522, 0.00000347, -0.00000030, 0.00000691, -0.00000015, -0.00000288, -0.00000906, 0.00000598, 0.00000677, -0.00000479, 0.00000383, -0.00000111, -0.00000475, -0.00000637, -0.00000850, 0.00000693, -0.00000821, -0.00000465, -0.00000372, 0.00000188, -0.00001856, 0.00000236, -0.00000557, -0.00000195, -0.00001881, -0.00000970, -0.00007772, 0.00003641, 0.00010029, 0.00006493, 0.00009624, -0.00006073, -0.00003898, -0.00000575, -0.00001098, -0.00000663, -0.00000026, -0.00000239, -0.00000400, 0.00000128, 0.00000279, 0.00000051, -0.00000334, -0.00000338, 0.00000023, 0.00000094, 0.00000484, -0.00000572, -0.00000070, -0.00000281, 0.00000072, -0.00000027, 0.00000713, 0.00000015, 0.00000548, -0.00001065, -0.00000417, 0.00000322, 0.00000355, -0.00000430, -0.00000835, 0.00000310, 0.00000082, 0.00000330, 0.00000640, -0.00000268, -0.00000209, -0.00000142, -0.00000425, 0.00000300, 0.00000550, -0.00000597, 0.00000151, -0.00000515, 0.00000013, 0.00000232, 0.00000286, -0.00000221, -0.00000085, -0.00000753, 0.00000282, -0.00001035, -0.00000123, -0.00003121, -0.00003450, 0.00004584, 0.00005201, 0.00003314, 0.00007059, -0.00000709, -0.00003263, -0.00001928, -0.00000043, -0.00000936, 0.00000034, -0.00000377, -0.00001265, -0.00000100, -0.00000506, -0.00000619, 0.00000402, -0.00000730, -0.00000775, 0.00000384, -0.00001439, 0.00000910, -0.00000642, 0.00000888, 0.00000543, -0.00000765, -0.00000261, 0.00000202, 0.00000272, -0.00000047, 0.00000475, -0.00000218, -0.00000164, -0.00000303, -0.00000376, },
		{ 0.00000119, -0.00000143, -0.00000031, 0.00000105, -0.00000167, 0.00000058, -0.00000324, 0.00000022, -0.00000264, -0.00000141, 0.00000253, 0.00000201, 0.00000539, -0.00000207, -0.00000532, -0.00000573, 0.00001122, -0.00000035, -0.00000184, 0.00000451, -0.00000626, -0.00000194, -0.00001140, 0.00000067, 0.00000101, -0.00000798, -0.00000504, 0.00000164, -0.00000942, -0.00001078, 0.00000221, -0.00000530, -0.00000917, -0.00001223, -0.00003843, -0.00005297, 0.00010198, 0.00005908, 0.00009551, 0.00003105, -0.00007381, -0.00001255, -0.00000972, -0.00000893, -0.00000318, -0.00000071, -0.00000373, -0.00000232, 0.00000349, 0.00000109, 0.00000004, -0.00000571, 0.00000043, -0.00000213, 0.00000577, -0.00000120, -0.00000336, -0.00000208, -0.00000067, -0.00000114, 0.00000477, 0.00000331, 0.00000293, -0.00000002, -0.00001170, 0.00000117, 0.00000320, 0.00000197, -0.00000864, -0.00000374, 0.00000424, -0.00000047, 0.00000719, 0.00000162, -0.00000270, -0.00000193, -0.00000189, -0.00000397, 0.00000821, -0.00000203, -0.00000176, -0.00000156, -0.00000358, 0.00000115, 0.00000389, -0.00000062, -0.00000034, -0.00000559, -0.00000177, -0.00000255, -0.00000851, -0.00000677, -0.00004703, -0.00000370, 0.00006753, 0.00003337, 0.00005976, 0.00004567, -0.00003653, -0.00002441, -0.00001161, -0.00000069, -0.00001018, 0.00000557, -0.00001437, -0.00000429, -0.00000320, -0.00000544, -0.00000390, 0.00000495, -0.00001521, 0.00000392, -0.00000847, -0.00000337, 0.00000487, -0.00000464, 0.00001343, -0.00000330, -0.00000592, -0.00000082, 0.00000338, 0.00000086, 0.00000170, 0.00000330, -0.00000444, 0.00000029, -0.00000604, 0.00000174, },
		{ -0.00000164, -0.00000017, -0.00000009, 0.00000038, -0.00000139, 0.00000033, -0.00000389, 0.00000142, -0.00000519, 0.00000305, 0.00000027, 0.00000529, 0.00000136, -0.00000226, -0.00000829, 0.00000218, 0.00000974, -0.00000510, 0.00000391, -0.00000116, -0.00000541, -0.00000373, -0.00001082, 0.00000628, -0.00000608, -0.00000569, -0.00000363, 0.00000120, -0.00001725, 0.00000106, -0.00000451, -0.00000299, -0.00001576, -0.00001110, -0.00006870, 0.00002502, 0.00009790, 0.00005210, 0.00009759, -0.00003662, -0.00005038, -0.00000386, -0.00001198, -0.00000537, -0.00000144, -0.00000226, -0.00000384, 0.00000067, 0.00000308, 0.00000079, -0.00000289, -0.00000363, 0.00000015, -0.00000036, 0.00000616, -0.00000520, -0.00000107, -0.00000288, 0.00000013, -0.00000026, 0.00000717, 0.00000009, 0.00000545, -0.00000798, -0.00000669, 0.00000311, 0.00000347, -0.00000212, -0.00000972, 0.00000193, 0.00000174, 0.00000235, 0.00000678, -0.00000181, -0.00000215, -0.00000154, -0.00000378, 0.00000083, 0.00000684, -0.00000546, 0.00000135, -0.00000505, -0.00000033, 0.00000169, 0.00000367, -0.00000255, -0.00000048, -0.00000729, 0.00000242, -0.00000992, -0.00000270, -0.00002423, -0.00004631, 0.00003899, 0.00006163, 0.00003404, 0.00007693, -0.00000313, -0.00003899, -0.00001704, -0.00000419, -0.00000714, -0.00000146, -0.00000213, -0.00001412, 0.00000000, -0.00000521, -0.00000633, 0.00000218, -0.00000326, -0.00001121, 0.00000502, -0.00001405, 0.00000738, -0.00000419, 0.00000493, 0.00000875, -0.00000764, -0.00000331, 0.00000079, 0.00000354, -0.00000021, 0.00000423, -0.00000114, -0.00000254, -0.00000156, -0.00000459, 0.00000383, },
		{ -0.00000134, -0.00000015, 0.00000069, -0.00000137, 0.00000085, -0.00000280, -0.00000097, -0.00000169, -0.00000253, 0.00000348, 0.00000087, 0.00000539, -0.00000110, -0.00000396, -0.00000749, 0.00000962, 0.00000289, -0.00000364, 0.00000596, -0.00000780, -0.00000057, -0.00001026, -0.00000251, 0.00000285, -0.00000766, -0.00000544, 0.00000099, -0.00000863, -0.00001114, 0.00000244, -0.00000593, -0.00000743, -0.00001236, -0.00003427, -0.00005111, 0.00008970, 0.00005892, 0.00007872, 0.00005143, -0.00006666, -0.00002152, -0.00000622, -0.00001066, -0.00000164, -0.00000252, -0.00000264, -0.00000307, 0.00000347, 0.00000132, 0.00000051, -0.00000561, 0.00000028, -0.00000270, 0.00000501, 0.00000083, -0.00000412, -0.00000163, -0.00000155, -0.00000101, 0.00000420, 0.00000417, 0.00000170, 0.00000225, -0.00001152, -0.00000075, 0.00000327, 0.00000282, -0.00000722, -0.00000619, 0.00000462, -0.00000038, 0.00000649, 0.00000254, -0.00000225, -0.00000218, -0.00000149, -0.00000462, 0.00000677, 0.00000014, -0.00000252, -0.00000085, -0.00000434, 0.00000112, 0.00000316, 0.00000079, -0.00000173, -0.00000376, -0.00000340, -0.00000079, -0.00001114, -0.00000293, -0.00004567, -0.00001975, 0.00007070, 0.00004089, 0.00006020, 0.00005555, -0.00004099, -0.00002575, -0.00001226, -0.00000209, -0.00000907, 0.00000459, -0.00001333, -0.00000557, -0.00000186, -0.00000583, -0.00000507, 0.00000547, -0.00001279, -0.00000032, -0.00000466, -0.00000654, 0.00000694, -0.00000623, 0.00001231, -0.00000016, -0.00000707, -0.00000146, 0.00000248, 0.00000232, 0.00000085, 0.00000397, -0.00000446, 0.00000024, -0.00000492, 0.00000017, 0.00000141, },
		{ -0.00000003, -0.00000021, 0.00000029, -0.00000122, 0.00000108, -0.00000450, 0.00000117, -0.00000486, 0.00000227, 0.00000120, 0.00000354, 0.00000263, -0.00000170, -0.00000704, -0.00000149, 0.00001165, -0.00000427, 0.00000302, -0.00000003, -0.00000722, -0.00000065, -0.00001264, 0.00000500, -0.00000394, -0.00000631, -0.00000389, 0.00000087, -0.00001637, 0.00000023, -0.00000388, -0.00000374, -0.00001305, -0.00001233, -0.00006046, 0.00001561, 0.00009409, 0.00004353, 0.00009240, -0.00001222, -0.00005751, -0.00000513, -0.00001117, -0.00000548, -0.00000135, -0.00000311, -0.00000303, -0.00000031, 0.00000351, 0.00000093, -0.00000228, -0.00000397, 0.00000031, -0.00000179, 0.00000712, -0.00000425, -0.00000165, -0.00000265, -0.00000063, -0.00000031, 0.00000714, 0.00000034, 0.00000486, -0.00000508, -0.00000870, 0.00000256, 0.00000335, -0.00000013, -0.00001029, 0.00000011, 0.00000281, 0.00000153, 0.00000698, -0.00000106, -0.00000190, -0.00000184, -0.00000311, -0.00000117, 0.00000755, -0.00000450, 0.00000102, -0.00000476, -0.00000092, 0.00000127, 0.00000400, -0.00000223, -0.00000092, -0.00000617, 0.00000120, -0.00000847, -0.00000559, -0.00001644, -0.00005509, 0.00002699, 0.00007237, 0.00003557, 0.00008316, 0.00000185, -0.00004576, -0.00001512, -0.00000752, -0.00000552, -0.00000267, -0.00000104, -0.00001510, 0.00000053, -0.00000493, -0.00000648, 0.00000019, 0.00000025, -0.00001314, 0.00000437, -0.00001228, 0.00000481, -0.00000144, 0.00000104, 0.00001112, -0.00000663, -0.00000432, -0.00000026, 0.00000381, 0.00000061, 0.00000337, 0.00000004, -0.00000346, -0.00000033, -0.00000494, 0.00000292, -0.00000132, },
		{ 0.00000007, 0.00000036, -0.00000125, 0.00000115, -0.00000211, -0.00000227, -0.00000084, -0.00000346, 0.00000407, 0.00000032, 0.00000479, 0.00000006, -0.00000289, -0.00000818, 0.00000686, 0.00000649, -0.00000535, 0.00000721, -0.00000868, -0.00000045, -0.00000792, -0.00000599, 0.00000426, -0.00000697, -0.00000583, 0.00000030, -0.00000783, -0.00001159, 0.00000285, -0.00000672, -0.00000570, -0.00001239, -0.00003082, -0.00004830, 0.00007760, 0.00005958, 0.00006345, 0.00006431, -0.00005371, -0.00003138, -0.00000316, -0.00001206, -0.00000039, -0.00000381, -0.00000215, -0.00000327, 0.00000301, 0.00000175, 0.00000086, -0.00000537, 0.00000007, -0.00000300, 0.00000378, 0.00000293, -0.00000464, -0.00000135, -0.00000209, -0.00000123, 0.00000372, 0.00000498, 0.00000070, 0.00000376, -0.00001039, -0.00000299, 0.00000334, 0.00000324, -0.00000524, -0.00000851, 0.00000452, 0.00000000, 0.00000575, 0.00000320, -0.00000170, -0.00000228, -0.00000130, -0.00000484, 0.00000496, 0.00000213, -0.00000294, -0.00000030, -0.00000495, 0.00000105, 0.00000240, 0.00000208, -0.00000280, -0.00000252, -0.00000414, -0.00000002, -0.00001236, -0.00000178, -0.00004005, -0.00003721, 0.00007025, 0.00005077, 0.00006014, 0.00006616, -0.00004489, -0.00002796, -0.00001241, -0.00000386, -0.00000781, 0.00000359, -0.00001233, -0.00000682, -0.00000065, -0.00000599, -0.00000608, 0.00000517, -0.00000954, -0.00000448, -0.00000184, -0.00000835, 0.00000768, -0.00000638, 0.00000984, 0.00000341, -0.00000794, -0.00000209, 0.00000130, 0.00000371, 0.00000034, 0.00000425, -0.00000413, -0.00000016, -0.00000362, -0.00000112, 0.00000138, -0.00000126, },
		},
		{ { -0.00000080, -0.00000304, 0.00001097, -0.00000624, 0.00000439, 0.00000144, -0.00000066, 0.00000303, -0.00001249, 0.00000257, -0.00000545, -0.00000322, 0.00000174, 0.00000317, -0.00001534, 0.00000134, -0.00000226, -0.00000299, 0.00000212, -0.00000272, 0.00001362, -0.00000374, -0.00000202, 0.00000171, -0.00000047, -0.00000027, -0.00000304, -0.00000255, -0.00000231, -0.00000332, 0.00000018, -0.00000270, -0.00000204, -0.00000353, -0.00001577, -0.00001041, -0.00005989, 0.00001079, 0.00009212, 0.00003788, 0.00008260, 0.00000844, -0.00005648, -0.00001123, -0.00000948, -0.00000501, -0.00001748, -0.00000393, -0.00000135, 0.00000089, 0.00000621, 0.00000118, -0.00000840, 0.00000178, -0.00000115, -0.00000170, 0.00000443, -0.00000307, -0.00000484, -0.00000211, -0.00000373, 0.00000302, 0.00001198, -0.00000277, -0.00000522, -0.00000277, 0.00001198, 0.00000302, -0.00000373, -0.00000211, -0.00000484, -0.00000307, 0.00000443, -0.00000170, -0.00000115, 0.00000178, -0.00000840, 0.00000118, 0.00000621, 0.00000089, -0.00000135, -0.00000393, -0.00001748, -0.00000501, -0.00000948, -0.00001123, -0.00005648, 0.00000844, 0.00008260, 0.00003788, 0.00009212, 0.00001079, -0.00005989, -0.00001041, -0.00001577, -0.00000353, -0.00000204, -0.00000270, 0.00000018, -0.00000332, -0.00000231, -0.00000255, -0.00000304, -0.00000027, -0.00000047, 0.00000171, -0.00000202, -0.00000374, 0.00001362, -0.00000272, 0.00000212, -0.00000299, -0.00000226, 0.00000134, -0.00001534, 0.00000317, 0.00000174, -0.00000322, -0.00000545, 0.00000257, -0.00001249, 0.00000303, -0.00000066, 0.00000144, 0.00000439, -0.00000624, 0.00001097, -0.00000304, },
		{ -0.00000205, 0.00000181, 0.00000767, -0.00000740, 0.00000946, -0.00000502, 0.00000567, -0.00000586, -0.00000697, 0.00000081, -0.00000580, -0.00000139, 0.00000446, -0.00000500, -0.00001268, 0.00000603, -0.00000823, 0.00000377, -0.00000370, 0.00000605, 0.00000856, -0.00000698, 0.00000189, 0.00000010, 0.00000029, -0.00000181, -0.00000292, -0.00000231, -0.00000259, -0.00000272, 0.00000106, -0.00000486, 0.00000066, -0.00001129, -0.00001014, -0.00002846, -0.00005118, 0.00006347, 0.00006670, 0.00004561, 0.00007409, -0.00003270, -0.00003948, -0.00000782, -0.00000682, -0.00000961, -0.00001505, 0.00000001, -0.00000205, 0.00000367, 0.00000590, -0.00000406, -0.00000577, 0.00000384, -0.00000449, 0.00000235, 0.00000220, -0.00000491, -0.00000396, -0.00000186, -0.00000365, 0.00000982, 0.00000628, -0.00000512, -0.00000543, 0.00000334, 0.00001196, -0.00000286, -0.00000196, -0.00000373, -0.00000440, -0.00000064, 0.00000431, -0.00000430, 0.00000277, -0.00000382, -0.00000592, 0.00000452, 0.00000529, -0.00000173, -0.00000005, -0.00001083, -0.00001417, -0.00000489, -0.00000943, -0.00002401, -0.00005809, 0.00005784, 0.00006587, 0.00005651, 0.00008689, -0.00004450, -0.00003947, -0.00000825, -0.00001466, 0.00000075, -0.00000425, -0.00000048, -0.00000160, -0.00000303, -0.00000240, -0.00000285, -0.00000237, 0.00000061, -0.00000098, 0.00000266, -0.00000665, 0.00000550, 0.00000925, -0.00000481, 0.00000403, -0.00000732, 0.00000420, -0.00000870, -0.00000894, 0.00000568, -0.00000046, -0.00000614, -0.00000101, -0.00000173, -0.00001134, 0.00000756, -0.00000538, 0.00000833, -0.00000461, 0.00000183, 0.00000733, -0.00000359, },
		{ -0.00000379, 0.00000787, -0.00000017, -0.00000105, 0.00000648, -0.00000494, 0.00000657, -0.00001173, -0.00000083, -0.00000333, -0.00000336, -0.00000011, 0.00000474, -0.00001400, -0.00000236, 0.00000107, -0.00000623, 0.00000476, -0.00000505, 0.00001339, -0.00000131, -0.00000344, 0.00000200, 0.00000024, -0.00000058, -0.00000227, -0.00000317, -0.00000144, -0.00000391, -0.00000008, -0.00000120, -0.00000353, -0.00000149, -0.00001457, -0.00000910, -0.00004913, -0.00001086, 0.00008520, 0.00004035, 0.00006495, 0.00004056, -0.00005093, -0.00002225, -0.00000756, -0.00000526, -0.00001450, -0.00000908, 0.00000078, -0.00000137, 0.00000623, 0.00000286, -0.00000765, -0.00000032, 0.00000134, -0.00000434, 0.00000510, -0.00000129, -0.00000556, -0.00000230, -0.00000337, 0.00000070, 0.00001227, -0.00000050, -0.00000520, -0.00000442, 0.00001046, 0.00000594, -0.00000420, -0.00000174, -0.00000454, -0.00000358, 0.00000218, 0.00000147, -0.00000317, 0.00000310, -0.00000824, -0.00000070, 0.00000541, 0.00000357, -0.00000315, -0.00000049, -0.00001729, -0.00000740, -0.00000915, -0.00000716, -0.00004671, -0.00003089, 0.00009047, 0.00004519, 0.00009040, 0.00003900, -0.00006890, -0.00001445, -0.00001555, -0.00000653, -0.00000084, -0.00000315, -0.00000067, -0.00000212, -0.00000333, -0.00000196, -0.00000374, -0.00000017, -0.00000066, 0.00000084, -0.00000039, -0.00000573, 0.00001282, 0.00000038, -0.00000061, -0.00000030, -0.00000441, 0.00000323, -0.00001469, 0.00000022, 0.00000385, -0.00000258, -0.00000738, 0.00000373, -0.00000956, -0.00000317, 0.00000429, -0.00000294, 0.00000823, -0.00000912, 0.00000982, 0.00000064, -0.00000162, },
		{ -0.00000246, 0.00000991, -0.00000588, 0.00000664, -0.00000089, 0.00000067, 0.00000067, -0.00001077, 0.00000085, -0.00000507, -0.00000143, 0.00000224, -0.00000084, -0.00001579, 0.00000627, -0.00000746, 0.00000195, -0.00000135, 0.00000233, 0.00001127, -0.00000681, 0.00000092, 0.00000087, 0.00000054, -0.00000173, -0.00000252, -0.00000278, -0.00000156, -0.00000409, 0.00000206, -0.00000435, -0.00000050, -0.00000792, -0.00001131, -0.00002092, -0.00005097, 0.00003911, 0.00007339, 0.00003398, 0.00007319, -0.00000096, -0.00004788, -0.00001183, -0.00000664, -0.00000713, -0.00001587, -0.00000303, -0.00000027, 0.00000089, 0.00000734, -0.00000254, -0.00000661, 0.00000338, -0.00000281, -0.00000096, 0.00000501, -0.00000483, -0.00000436, -0.00000186, -0.00000398, 0.00000736, 0.00000863, -0.00000429, -0.00000543, 0.00000037, 0.00001317, -0.00000121, -0.00000242, -0.00000335, -0.00000392, -0.00000261, 0.00000426, -0.00000226, 0.00000062, -0.00000130, -0.00000767, 0.00000323, 0.00000548, 0.00000059, -0.00000230, -0.00000509, -0.00001849, -0.00000315, -0.00001257, -0.00001020, -0.00006738, 0.00002265, 0.00008962, 0.00004814, 0.00010503, -0.00002879, -0.00005349, -0.00000660, -0.00001738, -0.00000034, -0.00000311, -0.00000166, -0.00000140, -0.00000257, -0.00000314, -0.00000229, -0.00000338, 0.00000128, -0.00000199, 0.00000291, -0.00000567, 0.00000217, 0.00001173, -0.00000464, 0.00000302, -0.00000541, 0.00000187, -0.00000502, -0.00001167, 0.00000540, 0.00000126, -0.00000592, -0.00000390, 0.00000297, -0.00001361, 0.00000531, -0.00000229, 0.00000445, 0.00000044, -0.00000470, 0.00001124, -0.00000347, -0.00000089, },
		{ 0.00000245, 0.00000605, -0.00000490, 0.00000902, -0.00000567, 0.00000564, -0.00000716, -0.00000558, -0.00000154, -0.00000344, -0.00000086, 0.00000434, -0.00001054, -0.00000745, 0.00000482, -0.00000923, 0.00000668, -0.00000628, 0.00001210, 0.00000145, -0.00000461, 0.00000179, 0.00000118, -0.00000079, -0.00000161, -0.00000361, -0.00000099, -0.00000375, -0.00000116, 0.00000051, -0.00000428, -0.00000101, -0.00001203, -0.00000944, -0.00003877, -0.00002512, 0.00007155, 0.00004758, 0.00004676, 0.00005859, -0.00003185, -0.00003472, -0.00000714, -0.00000574, -0.00001095, -0.00001313, 0.00000096, -0.00000170, 0.00000466, 0.00000493, -0.00000687, -0.00000207, 0.00000299, -0.00000518, 0.00000352, 0.00000179, -0.00000662, -0.00000225, -0.00000327, -0.00000093, 0.00001152, 0.00000206, -0.00000506, -0.00000539, 0.00000795, 0.00000894, -0.00000431, -0.00000152, -0.00000435, -0.00000337, -0.00000050, 0.00000371, -0.00000358, 0.00000288, -0.00000671, -0.00000308, 0.00000456, 0.00000524, -0.00000284, -0.00000039, -0.00001304, -0.00001283, -0.00000571, -0.00001019, -0.00002744, -0.00006481, 0.00007870, 0.00006483, 0.00008077, 0.00007116, -0.00007339, -0.00002200, -0.00001431, -0.00000964, -0.00000073, -0.00000233, -0.00000228, -0.00000085, -0.00000408, -0.00000166, -0.00000412, -0.00000053, -0.00000024, -0.00000044, 0.00000112, -0.00000697, 0.00001088, 0.00000401, -0.00000300, 0.00000145, -0.00000526, 0.00000369, -0.00001288, -0.00000303, 0.00000531, -0.00000113, -0.00000879, 0.00000263, -0.00000439, -0.00000923, 0.00000700, -0.00000437, 0.00000875, -0.00000787, 0.00000478, 0.00000603, -0.00000323, -0.00000244, },
		{ 0.00000758, -0.00000045, 0.00000127, 0.00000478, -0.00000435, 0.00000507, -0.00001115, -0.00000142, -0.00000390, -0.00000116, 0.00000020, 0.00000212, -0.00001667, 0.00000405, -0.00000464, -0.00000144, 0.00000208, -0.00000174, 0.00001364, -0.00000631, 0.00000004, 0.00000104, 0.00000145, -0.00000212, -0.00000164, -0.00000375, -0.00000019, -0.00000527, 0.00000229, -0.00000309, -0.00000168, -0.00000575, -0.00001076, -0.00001694, -0.00004605, 0.00001695, 0.00007428, 0.00003083, 0.00006201, 0.00002674, -0.00004405, -0.00002144, -0.00000501, -0.00000662, -0.00001380, -0.00000784, 0.00000202, -0.00000165, 0.00000803, -0.00000095, -0.00000680, 0.00000209, -0.00000033, -0.00000410, 0.00000640, -0.00000334, -0.00000542, -0.00000153, -0.00000434, 0.00000513, 0.00001007, -0.00000289, -0.00000528, -0.00000222, 0.00001313, 0.00000128, -0.00000311, -0.00000288, -0.00000360, -0.00000354, 0.00000260, 0.00000056, -0.00000136, 0.00000061, -0.00000835, 0.00000148, 0.00000492, 0.00000345, -0.00000439, -0.00000097, -0.00001939, -0.00000476, -0.00001255, -0.00000499, -0.00005734, -0.00002458, 0.00010588, 0.00004898, 0.00011339, -0.00000256, -0.00006962, -0.00000578, -0.00001965, -0.00000177, -0.00000274, -0.00000137, -0.00000270, -0.00000105, -0.00000453, -0.00000123, -0.00000478, 0.00000208, -0.00000267, 0.00000257, -0.00000421, -0.00000105, 0.00001306, -0.00000304, 0.00000101, -0.00000307, -0.00000028, -0.00000212, -0.00001316, 0.00000403, 0.00000305, -0.00000473, -0.00000722, 0.00000587, -0.00001223, 0.00000010, 0.00000199, 0.00000045, 0.00000485, -0.00000887, 0.00001132, -0.00000052, -0.00000134, -0.00000381, },
		{ 0.00000908, -0.00000445, 0.00000720, -0.00000196, 0.00000085, -0.00000070, -0.00000969, -0.00000095, -0.00000345, -0.00000050, 0.00000246, -0.00000601, -0.00001249, 0.00000755, -0.00001084, 0.00000706, -0.00000585, 0.00000944, 0.00000477, -0.00000579, 0.00000148, 0.00000183, -0.00000048, -0.00000140, -0.00000350, -0.00000128, -0.00000273, -0.00000275, 0.00000187, -0.00000402, -0.00000159, -0.00000930, -0.00000969, -0.00003074, -0.00003210, 0.00005374, 0.00005523, 0.00003324, 0.00006276, -0.00000678, -0.00004138, -0.00001166, -0.00000453, -0.00000893, -0.00001419, -0.00000187, 0.00000022, 0.00000160, 0.00000744, -0.00000647, -0.00000292, 0.00000322, -0.00000396, 0.00000015, 0.00000529, -0.00000731, -0.00000242, -0.00000308, -0.00000222, 0.00001023, 0.00000444, -0.00000461, -0.00000582, 0.00000491, 0.00001140, -0.00000365, -0.00000168, -0.00000411, -0.00000287, -0.00000276, 0.00000430, -0.00000223, 0.00000145, -0.00000438, -0.00000538, 0.00000362, 0.00000554, -0.00000037, -0.00000288, -0.00000663, -0.00001872, -0.00000193, -0.00001546, -0.00000921, -0.00008057, 0.00004281, 0.00009291, 0.00006839, 0.00010207, -0.00006952, -0.00003414, -0.00001203, -0.00001287, -0.00000136, -0.00000124, -0.00000338, -0.00000058, -0.00000385, -0.00000216, -0.00000376, -0.00000167, 0.00000086, -0.00000192, 0.00000226, -0.00000731, 0.00000795, 0.00000769, -0.00000453, 0.00000200, -0.00000497, 0.00000326, -0.00001064, -0.00000595, 0.00000580, 0.00000082, -0.00000914, -0.00000036, 0.00000124, -0.00001296, 0.00000617, -0.00000274, 0.00000656, -0.00000390, -0.00000170, 0.00001052, -0.00000364, -0.00000103, -0.00000227, },
		{ 0.00000599, -0.00000327, 0.00000839, -0.00000593, 0.00000490, -0.00000736, -0.00000558, -0.00000252, -0.00000118, -0.00000061, 0.00000292, -0.00001483, -0.00000056, -0.00000028, -0.00000577, 0.00000594, -0.00000560, 0.00001529, -0.00000532, -0.00000080, 0.00000070, 0.00000281, -0.00000274, -0.00000047, -0.00000497, 0.00000113, -0.00000597, 0.00000188, -0.00000183, -0.00000191, -0.00000540, -0.00000850, -0.00001550, -0.00003902, -0.00000070, 0.00006905, 0.00003356, 0.00004764, 0.00004436, -0.00002920, -0.00003197, -0.00000521, -0.00000602, -0.00001112, -0.00001182, 0.00000253, -0.00000230, 0.00000694, 0.00000148, -0.00000728, 0.00000109, 0.00000152, -0.00000557, 0.00000537, -0.00000012, -0.00000722, -0.00000079, -0.00000482, 0.00000324, 0.00001067, -0.00000114, -0.00000497, -0.00000424, 0.00001192, 0.00000417, -0.00000359, -0.00000260, -0.00000331, -0.00000370, 0.00000021, 0.00000284, -0.00000201, 0.00000105, -0.00000750, -0.00000094, 0.00000429, 0.00000535, -0.00000427, -0.00000053, -0.00001543, -0.00001064, -0.00000791, -0.00000922, -0.00003393, -0.00006912, 0.00010154, 0.00006462, 0.00011008, 0.00003342, -0.00008565, -0.00000675, -0.00002164, -0.00000280, -0.00000391, 0.00000064, -0.00000522, 0.00000105, -0.00000617, 0.00000001, -0.00000623, 0.00000261, -0.00000279, 0.00000174, -0.00000263, -0.00000360, 0.00001282, -0.00000002, -0.00000166, -0.00000079, -0.00000198, 0.00000010, -0.00001386, 0.00000229, 0.00000421, -0.00000250, -0.00001025, 0.00000639, -0.00000805, -0.00000577, 0.00000485, -0.00000151, 0.00000676, -0.00000931, 0.00000741, 0.00000454, -0.00000291, -0.00000268, 0.00000226, },
		{ 0.00000054, 0.00000190, 0.00000419, -0.00000472, 0.00000426, -0.00001078, -0.00000253, -0.00000303, 0.00000019, 0.00000047, -0.00000213, -0.00001572, 0.00000784, -0.00001010, 0.00000524, -0.00000335, 0.00000520, 0.00000882, -0.00000720, 0.00000145, 0.00000173, 0.00000073, -0.00000197, -0.00000248, -0.00000261, -0.00000081, -0.00000472, 0.00000291, -0.00000344, -0.00000184, -0.00000803, -0.00000817, -0.00002623, -0.00003273, 0.00003435, 0.00006057, 0.00002565, 0.00005766, 0.00001589, -0.00003785, -0.00002096, -0.00000236, -0.00000836, -0.00001256, -0.00000665, 0.00000296, -0.00000161, 0.00000946, -0.00000590, -0.00000340, 0.00000266, -0.00000166, -0.00000350, 0.00000768, -0.00000654, -0.00000352, -0.00000229, -0.00000360, 0.00000886, 0.00000633, -0.00000382, -0.00000581, 0.00000177, 0.00001295, -0.00000222, -0.00000199, -0.00000400, -0.00000218, -0.00000435, 0.00000342, -0.00000003, 0.00000000, -0.00000253, -0.00000657, 0.00000205, 0.00000510, 0.00000280, -0.00000543, -0.00000158, -0.00002118, -0.00000193, -0.00001689, -0.00000118, -0.00007239, -0.00001138, 0.00011766, 0.00006233, 0.00012487, -0.00005312, -0.00005200, -0.00000815, -0.00001713, -0.00000131, -0.00000167, -0.00000236, -0.00000237, -0.00000198, -0.00000392, -0.00000226, -0.00000383, 0.00000260, -0.00000346, 0.00000300, -0.00000698, 0.00000465, 0.00001053, -0.00000450, 0.00000108, -0.00000366, 0.00000222, -0.00000814, -0.00000862, 0.00000579, 0.00000242, -0.00000787, -0.00000472, 0.00000594, -0.00001350, 0.00000246, 0.00000022, 0.00000376, 0.00000027, -0.00000667, 0.00001170, -0.00000138, -0.00000113, -0.00000413, 0.00000730, },
		{ -0.00000363, 0.00000730, -0.00000247, 0.00000017, -0.00000094, -0.00000982, -0.00000183, -0.00000148, -0.00000011, 0.00000188, -0.00001108, -0.00000622, 0.00000414, -0.00000970, 0.00000907, -0.00000821, 0.00001540, -0.00000321, -0.00000207, 0.00000033, 0.00000397, -0.00000299, 0.00000043, -0.00000576, 0.00000172, -0.00000573, 0.00000079, -0.00000078, -0.00000123, -0.00000606, -0.00000600, -0.00001443, -0.00003284, -0.00001206, 0.00005820, 0.00003954, 0.00003438, 0.00005163, -0.00000950, -0.00003707, -0.00001027, -0.00000371, -0.00000974, -0.00001301, -0.00000002, -0.00000045, 0.00000387, 0.00000498, -0.00000849, 0.00000094, 0.00000208, -0.00000484, 0.00000209, 0.00000411, -0.00000900, -0.00000023, -0.00000501, 0.00000136, 0.00001081, 0.00000061, -0.00000443, -0.00000564, 0.00000981, 0.00000698, -0.00000354, -0.00000258, -0.00000313, -0.00000324, -0.00000226, 0.00000389, -0.00000114, 0.00000030, -0.00000572, -0.00000337, 0.00000351, 0.00000583, -0.00000169, -0.00000325, -0.00000837, -0.00001817, -0.00000171, -0.00001738, -0.00000959, -0.00009464, 0.00006840, 0.00009367, 0.00009870, 0.00007442, -0.00009697, -0.00001178, -0.00002242, -0.00000407, -0.00000580, 0.00000302, -0.00000735, 0.00000233, -0.00000699, 0.00000056, -0.00000692, 0.00000220, -0.00000200, 0.00000042, -0.00000113, -0.00000526, 0.00001105, 0.00000389, -0.00000422, 0.00000083, -0.00000296, 0.00000170, -0.00001399, 0.00000039, 0.00000474, 0.00000015, -0.00001200, 0.00000434, -0.00000239, -0.00001029, 0.00000498, -0.00000106, 0.00000633, -0.00000687, 0.00000147, 0.00000932, -0.00000375, -0.00000102, -0.00000293, 0.00000971, },
		{ -0.00000340, 0.00000882, -0.00000704, 0.00000463, -0.00000750, -0.00000634, -0.00000207, 0.00000041, -0.00000036, -0.00000019, -0.00001614, 0.00000522, -0.00000700, 0.00000137, 0.00000096, -0.00000026, 0.00001319, -0.00000854, 0.00000163, 0.00000084, 0.00000279, -0.00000329, -0.00000055, -0.00000486, 0.00000174, -0.00000680, 0.00000381, -0.00000332, -0.00000067, -0.00000902, -0.00000458, -0.00002474, -0.00002927, 0.00001626, 0.00006181, 0.00002344, 0.00004820, 0.00003136, -0.00002530, -0.00003071, -0.00000235, -0.00000736, -0.00001032, -0.00001086, 0.00000420, -0.00000313, 0.00000967, -0.00000407, -0.00000461, 0.00000246, 0.00000025, -0.00000571, 0.00000760, -0.00000365, -0.00000580, -0.00000080, -0.00000513, 0.00000748, 0.00000779, -0.00000288, -0.00000535, -0.00000125, 0.00001355, -0.00000037, -0.00000205, -0.00000425, -0.00000127, -0.00000532, 0.00000171, 0.00000185, -0.00000029, -0.00000206, -0.00000608, -0.00000038, 0.00000462, 0.00000501, -0.00000565, -0.00000057, -0.00001793, -0.00000761, -0.00001184, -0.00000564, -0.00004547, -0.00006774, 0.00012294, 0.00007135, 0.00013446, -0.00002204, -0.00007514, -0.00000284, -0.00002306, 0.00000051, -0.00000476, 0.00000136, -0.00000625, 0.00000134, -0.00000678, 0.00000024, -0.00000678, 0.00000464, -0.00000480, 0.00000341, -0.00000639, 0.00000177, 0.00001170, -0.00000245, -0.00000133, -0.00000152, 0.00000069, -0.00000523, -0.00001141, 0.00000586, 0.00000304, -0.00000492, -0.00000952, 0.00000867, -0.00001113, -0.00000235, 0.00000235, 0.00000222, 0.00000278, -0.00000847, 0.00000893, 0.00000326, -0.00000265, -0.00000311, 0.00000184, 0.00000746, },
		{ 0.00000180, 0.00000439, -0.00000613, 0.00000463, -0.00001147, -0.00000307, -0.00000145, 0.00000083, 0.00000038, -0.00000724, -0.00001095, 0.00000687, -0.00001181, 0.00001021, -0.00000847, 0.00001312, 0.00000062, -0.00000424, 0.00000051, 0.00000422, -0.00000216, 0.00000032, -0.00000531, 0.00000081, -0.00000402, -0.00000124, 0.00000046, -0.00000072, -0.00000589, -0.00000545, -0.00001136, -0.00002983, -0.00001647, 0.00004295, 0.00004665, 0.00002388, 0.00005171, 0.00000814, -0.00003298, -0.00001996, -0.00000044, -0.00000964, -0.00001152, -0.00000484, 0.00000275, -0.00000003, 0.00000866, -0.00000978, 0.00000106, 0.00000184, -0.00000272, -0.00000203, 0.00000772, -0.00000949, -0.00000074, -0.00000427, -0.00000091, 0.00001080, 0.00000220, -0.00000376, -0.00000635, 0.00000702, 0.00000946, -0.00000308, -0.00000242, -0.00000343, -0.00000206, -0.00000462, 0.00000384, 0.00000043, -0.00000041, -0.00000440, -0.00000459, 0.00000192, 0.00000556, 0.00000182, -0.00000641, -0.00000215, -0.00002266, 0.00000080, -0.00002164, 0.00000359, -0.00009169, 0.00000982, 0.00012489, 0.00008876, 0.00011227, -0.00009711, -0.00002419, -0.00001988, -0.00000761, -0.00000614, 0.00000320, -0.00000683, 0.00000110, -0.00000573, -0.00000065, -0.00000590, 0.00000012, 0.00000006, -0.00000149, 0.00000040, -0.00000631, 0.00000849, 0.00000751, -0.00000549, 0.00000096, -0.00000277, 0.00000242, -0.00001318, -0.00000215, 0.00000542, 0.00000204, -0.00001146, -0.00000013, 0.00000348, -0.00001246, 0.00000271, 0.00000046, 0.00000518, -0.00000368, -0.00000368, 0.00001131, -0.00000213, -0.00000072, -0.00000537, 0.00000838, 0.00000123, },
		{ 0.00000848, -0.00000376, -0.00000023, -0.00000085, -0.00001086, -0.00000135, 0.00000008, 0.00000020, -0.00000017, -0.00001420, 0.00000069, -0.00000268, -0.00000346, 0.00000597, -0.00000570, 0.00001672, -0.00000890, 0.00000137, -0.00000029, 0.00000505, -0.00000472, 0.00000166, -0.00000723, 0.00000403, -0.00000828, 0.00000428, -0.00000369, 0.00000170, -0.00001128, -0.00000059, -0.00002389, -0.00002523, 0.00000262, 0.00005748, 0.00002598, 0.00003717, 0.00003965, -0.00000934, -0.00003495, -0.00000751, -0.00000431, -0.00000938, -0.00001242, 0.00000235, -0.00000194, 0.00000740, -0.00000038, -0.00000721, 0.00000330, 0.00000089, -0.00000561, 0.00000484, 0.00000090, -0.00000863, 0.00000079, -0.00000627, 0.00000568, 0.00000913, -0.00000204, -0.00000448, -0.00000394, 0.00001326, 0.00000162, -0.00000171, -0.00000473, -0.00000049, -0.00000548, -0.00000047, 0.00000286, 0.00000074, -0.00000273, -0.00000458, -0.00000293, 0.00000399, 0.00000568, -0.00000301, -0.00000348, -0.00001033, -0.00001662, -0.00000321, -0.00001690, -0.00001387, -0.00010620, 0.00009620, 0.00009747, 0.00013162, 0.00002114, -0.00009966, 0.00000143, -0.00002938, 0.00000322, -0.00000951, 0.00000628, -0.00001053, 0.00000460, -0.00000954, 0.00000271, -0.00000949, 0.00000603, -0.00000538, 0.00000335, -0.00000571, -0.00000033, 0.00001100, 0.00000128, -0.00000452, 0.00000073, -0.00000092, -0.00000207, -0.00001418, 0.00000587, 0.00000298, -0.00000124, -0.00001329, 0.00000860, -0.00000658, -0.00000662, 0.00000248, 0.00000235, 0.00000354, -0.00000748, 0.00000374, 0.00000819, -0.00000391, -0.00000105, -0.00000418, 0.00001221, -0.00000470, },
		{ 0.00001043, -0.00000952, 0.00000567, -0.00000856, -0.00000683, -0.00000067, 0.00000121, -0.00000014, -0.00000484, -0.00001336, 0.00000694, -0.00001148, 0.00000880, -0.00000594, 0.00000827, 0.00000606, -0.00000717, 0.00000135, 0.00000325, 0.00000005, -0.00000116, -0.00000318, -0.00000189, -0.00000076, -0.00000429, 0.00000239, -0.00000154, -0.00000337, -0.00000809, -0.00000543, -0.00003032, -0.00001528, 0.00002585, 0.00005253, 0.00001691, 0.00004771, 0.00002015, -0.00002112, -0.00003001, 0.00000058, -0.00000892, -0.00000929, -0.00000954, 0.00000497, -0.00000279, 0.00001090, -0.00000972, 0.00000010, 0.00000206, -0.00000072, -0.00000509, 0.00000904, -0.00000765, -0.00000286, -0.00000239, -0.00000355, 0.00001055, 0.00000379, -0.00000325, -0.00000620, 0.00000366, 0.00001163, -0.00000258, -0.00000165, -0.00000458, -0.00000011, -0.00000683, 0.00000317, 0.00000161, 0.00000012, -0.00000460, -0.00000390, -0.00000081, 0.00000536, 0.00000434, -0.00000714, -0.00000028, -0.00002071, -0.00000370, -0.00001756, 0.00000091, -0.00006347, -0.00005781, 0.00013989, 0.00009157, 0.00013893, -0.00008057, -0.00004627, -0.00001247, -0.00001522, -0.00000288, -0.00000074, -0.00000231, -0.00000337, -0.00000192, -0.00000404, -0.00000278, -0.00000381, 0.00000331, -0.00000396, 0.00000216, -0.00000728, 0.00000618, 0.00000954, -0.00000444, -0.00000098, -0.00000114, 0.00000188, -0.00001088, -0.00000598, 0.00000698, 0.00000232, -0.00000836, -0.00000617, 0.00000831, -0.00001214, -0.00000061, 0.00000127, 0.00000490, -0.00000148, -0.00000620, 0.00000947, 0.00000211, -0.00000216, -0.00000447, 0.00000254, 0.00000909, -0.00000474, },
		{ 0.00000407, -0.00000765, 0.00000570, -0.00001327, -0.00000236, -0.00000003, 0.00000124, -0.00000079, -0.00001157, -0.00000389, 0.00000108, -0.00000769, 0.00001009, -0.00000951, 0.00001794, -0.00000717, -0.00000009, -0.00000086, 0.00000656, -0.00000526, 0.00000302, -0.00000843, 0.00000476, -0.00000811, 0.00000356, -0.00000365, 0.00000364, -0.00001236, 0.00000098, -0.00002063, -0.00002402, -0.00000420, 0.00004699, 0.00003254, 0.00002564, 0.00004342, 0.00000396, -0.00003020, -0.00001769, 0.00000009, -0.00000983, -0.00001112, -0.00000221, 0.00000123, 0.00000345, 0.00000451, -0.00001057, 0.00000462, 0.00000064, -0.00000377, 0.00000058, 0.00000556, -0.00001059, 0.00000137, -0.00000622, 0.00000305, 0.00001056, -0.00000131, -0.00000346, -0.00000595, 0.00001197, 0.00000378, -0.00000135, -0.00000478, -0.00000059, -0.00000434, -0.00000319, 0.00000324, 0.00000219, -0.00000316, -0.00000366, -0.00000412, 0.00000237, 0.00000555, 0.00000093, -0.00000746, -0.00000271, -0.00002356, 0.00000261, -0.00002534, 0.00000702, -0.00011267, 0.00003703, 0.00013132, 0.00012476, 0.00006857, -0.00011809, -0.00000025, -0.00003263, 0.00000374, -0.00001275, 0.00000912, -0.00001225, 0.00000548, -0.00001032, 0.00000346, -0.00001039, 0.00000549, -0.00000437, 0.00000233, -0.00000461, -0.00000199, 0.00000919, 0.00000536, -0.00000698, 0.00000187, -0.00000172, 0.00000052, -0.00001599, 0.00000477, 0.00000345, 0.00000152, -0.00001442, 0.00000527, -0.00000071, -0.00000950, 0.00000091, 0.00000301, 0.00000382, -0.00000549, -0.00000123, 0.00001081, -0.00000297, -0.00000014, -0.00000760, 0.00001169, -0.00000007, 0.00000264, },
		{ -0.00000675, 0.00000073, -0.00000151, -0.00001195, 0.00000027, 0.00000082, 0.00000070, -0.00000411, -0.00001353, 0.00000485, -0.00000932, 0.00000538, -0.00000124, 0.00000179, 0.00001202, -0.00000988, 0.00000224, 0.00000147, 0.00000321, -0.00000361, 0.00000017, -0.00000572, 0.00000322, -0.00000766, 0.00000475, -0.00000377, 0.00000129, -0.00001296, 0.00000168, -0.00003201, -0.00001205, 0.00001081, 0.00005424, 0.00001472, 0.00004071, 0.00002736, -0.00000706, -0.00003445, -0.00000387, -0.00000580, -0.00000835, -0.00001180, 0.00000428, -0.00000292, 0.00001051, -0.00000718, -0.00000294, 0.00000363, 0.00000004, -0.00000588, 0.00000741, -0.00000356, -0.00000616, 0.00000004, -0.00000590, 0.00000954, 0.00000577, -0.00000319, -0.00000518, -0.00000004, 0.00001344, -0.00000224, -0.00000021, -0.00000633, 0.00000208, -0.00000841, 0.00000198, 0.00000196, 0.00000206, -0.00000607, -0.00000196, -0.00000387, 0.00000513, 0.00000509, -0.00000447, -0.00000325, -0.00001292, -0.00001362, -0.00000723, -0.00001260, -0.00002484, -0.00011106, 0.00012173, 0.00011281, 0.00015175, -0.00004655, -0.00007638, -0.00000154, -0.00002629, 0.00000359, -0.00000824, 0.00000506, -0.00000969, 0.00000321, -0.00000864, 0.00000156, -0.00000863, 0.00000683, -0.00000636, 0.00000396, -0.00000838, 0.00000469, 0.00000941, -0.00000097, -0.00000461, 0.00000149, 0.00000030, -0.00000713, -0.00001093, 0.00000917, 0.00000139, -0.00000375, -0.00001201, 0.00001074, -0.00000950, -0.00000377, 0.00000048, 0.00000577, -0.00000050, -0.00000613, 0.00000491, 0.00000719, -0.00000387, -0.00000186, -0.00000475, 0.00001558, -0.00000766, 0.00001136, },
		{ -0.00001274, 0.00000791, -0.00001098, -0.00000617, 0.00000083, 0.00000157, -0.00000050, -0.00000987, -0.00000697, 0.00000309, -0.00001029, 0.00001220, -0.00001051, 0.00001597, -0.00000288, -0.00000303, -0.00000045, 0.00000662, -0.00000411, 0.00000261, -0.00000742, 0.00000301, -0.00000561, 0.00000101, -0.00000212, 0.00000336, -0.00001005, -0.00000187, -0.00001330, -0.00002701, -0.00000421, 0.00003180, 0.00004115, 0.00001484, 0.00004467, 0.00001194, -0.00001801, -0.00002870, 0.00000260, -0.00000979, -0.00000875, -0.00000729, 0.00000420, -0.00000028, 0.00000886, -0.00001300, 0.00000484, 0.00000081, -0.00000170, -0.00000320, 0.00000840, -0.00001034, 0.00000011, -0.00000459, -0.00000038, 0.00001179, -0.00000031, -0.00000281, -0.00000682, 0.00000947, 0.00000640, -0.00000160, -0.00000358, -0.00000226, -0.00000158, -0.00000655, 0.00000353, 0.00000287, -0.00000195, -0.00000467, -0.00000294, -0.00000080, 0.00000566, 0.00000387, -0.00000900, 0.00000051, -0.00002368, 0.00000055, -0.00002409, 0.00000904, -0.00008694, -0.00003933, 0.00015351, 0.00012653, 0.00011046, -0.00012216, -0.00001300, -0.00002909, -0.00000140, -0.00001111, 0.00000674, -0.00000903, 0.00000229, -0.00000777, 0.00000122, -0.00000832, 0.00000217, -0.00000128, -0.00000002, -0.00000256, -0.00000399, 0.00000757, 0.00000807, -0.00000710, 0.00000067, -0.00000080, 0.00000153, -0.00001557, 0.00000133, 0.00000566, 0.00000204, -0.00001211, -0.00000089, 0.00000535, -0.00001069, -0.00000126, 0.00000265, 0.00000501, -0.00000415, -0.00000398, 0.00000972, 0.00000080, -0.00000125, -0.00000719, 0.00000570, 0.00000890, -0.00000558, 0.00001211, },
		{ -0.00000799, 0.00000641, -0.00001554, -0.00000028, 0.00000057, 0.00000196, -0.00000385, -0.00001291, 0.00000219, -0.00000676, 0.00000120, 0.00000418, -0.00000459, 0.00001671, -0.00001096, 0.00000218, -0.00000023, 0.00000627, -0.00000589, 0.00000344, -0.00000912, 0.00000630, -0.00000993, 0.00000642, -0.00000611, 0.00000607, -0.00001707, 0.00000654, -0.00003121, -0.00001108, 0.00000170, 0.00004926, 0.00001841, 0.00003048, 0.00003273, 0.00000330, -0.00002961, -0.00001407, -0.00000069, -0.00000905, -0.00001105, 0.00000054, -0.00000059, 0.00000765, -0.00000233, -0.00000770, 0.00000614, -0.00000029, -0.00000461, 0.00000365, 0.00000144, -0.00000925, 0.00000177, -0.00000697, 0.00000720, 0.00000832, -0.00000349, -0.00000366, -0.00000354, 0.00001448, -0.00000169, 0.00000122, -0.00000761, 0.00000328, -0.00000837, -0.00000022, 0.00000200, 0.00000429, -0.00000724, -0.00000065, -0.00000551, 0.00000384, 0.00000487, 0.00000005, -0.00000820, -0.00000376, -0.00002311, 0.00000217, -0.00002581, 0.00000551, -0.00013064, 0.00006535, 0.00014590, 0.00015681, -0.00000081, -0.00010783, 0.00000774, -0.00003684, 0.00000971, -0.00001570, 0.00001157, -0.00001455, 0.00000699, -0.00001223, 0.00000505, -0.00001236, 0.00000891, -0.00000751, 0.00000504, -0.00000909, 0.00000366, 0.00000780, 0.00000363, -0.00000836, 0.00000372, -0.00000123, -0.00000302, -0.00001561, 0.00001043, 0.00000091, 0.00000024, -0.00001542, 0.00000960, -0.00000482, -0.00000623, -0.00000148, 0.00000663, 0.00000040, -0.00000500, 0.00000010, 0.00001047, -0.00000375, 0.00000000, -0.00000984, 0.00001679, -0.00000398, 0.00000562, 0.00000191, },
		{ 0.00000340, -0.00000367, -0.00001209, 0.00000233, 0.00000088, 0.00000116, -0.00000945, -0.00000841, 0.00000348, -0.00001137, 0.00001213, -0.00000858, 0.00001108, 0.00000329, -0.00000671, 0.00000062, 0.00000526, -0.00000127, 0.00000035, -0.00000416, -0.00000101, -0.00000116, -0.00000310, 0.00000109, 0.00000030, -0.00000405, -0.00000862, -0.00000316, -0.00003240, -0.00000045, 0.00001580, 0.00004813, 0.00000735, 0.00004294, 0.00001634, -0.00000421, -0.00003432, -0.00000040, -0.00000720, -0.00000748, -0.00001038, 0.00000476, -0.00000193, 0.00001096, -0.00001282, 0.00000254, 0.00000256, -0.00000073, -0.00000496, 0.00000831, -0.00000748, -0.00000287, -0.00000189, -0.00000383, 0.00001209, 0.00000156, -0.00000303, -0.00000629, 0.00000574, 0.00000961, -0.00000279, -0.00000095, -0.00000536, 0.00000228, -0.00000997, 0.00000370, 0.00000236, 0.00000116, -0.00000751, 0.00000005, -0.00000480, 0.00000609, 0.00000468, -0.00000653, -0.00000209, -0.00001642, -0.00000912, -0.00001367, -0.00000447, -0.00004347, -0.00010699, 0.00014324, 0.00014566, 0.00014116, -0.00010741, -0.00003842, -0.00001788, -0.00001319, -0.00000369, -0.00000134, -0.00000111, -0.00000438, -0.00000236, -0.00000371, -0.00000356, -0.00000346, 0.00000332, -0.00000341, 0.00000048, -0.00000676, 0.00000706, 0.00000834, -0.00000419, -0.00000305, 0.00000193, 0.00000064, -0.00001247, -0.00000464, 0.00000957, 0.00000051, -0.00000722, -0.00000810, 0.00000997, -0.00000999, -0.00000312, 0.00000059, 0.00000722, -0.00000372, -0.00000426, 0.00000559, 0.00000599, -0.00000327, -0.00000421, -0.00000299, 0.00001777, -0.00001101, 0.00001521, -0.00001127, },
		{ 0.00001046, -0.00001443, -0.00000389, 0.00000157, 0.00000222, -0.00000246, -0.00001308, 0.00000041, -0.00000492, -0.00000270, 0.00000897, -0.00000921, 0.00001862, -0.00000935, 0.00000051, -0.00000110, 0.00000819, -0.00000681, 0.00000520, -0.00001045, 0.00000690, -0.00000973, 0.00000605, -0.00000677, 0.00000831, -0.00001727, 0.00000612, -0.00002522, -0.00001498, 0.00000019, 0.00003734, 0.00002727, 0.00001777, 0.00003797, 0.00000765, -0.00001693, -0.00002604, 0.00000324, -0.00000964, -0.00000876, -0.00000440, 0.00000237, 0.00000395, 0.00000325, -0.00001247, 0.00000793, -0.00000041, -0.00000267, -0.00000031, 0.00000533, -0.00001053, 0.00000170, -0.00000616, 0.00000352, 0.00001104, -0.00000357, -0.00000244, -0.00000600, 0.00001407, -0.00000016, 0.00000151, -0.00000712, 0.00000224, -0.00000584, -0.00000394, 0.00000258, 0.00000529, -0.00000627, -0.00000181, -0.00000422, 0.00000051, 0.00000505, 0.00000371, -0.00001100, 0.00000134, -0.00002603, 0.00000368, -0.00002910, 0.00001524, -0.00011176, -0.00001647, 0.00017159, 0.00016670, 0.00004690, -0.00013115, 0.00000828, -0.00004133, 0.00001045, -0.00001839, 0.00001295, -0.00001454, 0.00000681, -0.00001258, 0.00000559, -0.00001304, 0.00000796, -0.00000629, 0.00000445, -0.00000846, 0.00000199, 0.00000624, 0.00000729, -0.00001011, 0.00000376, -0.00000121, -0.00000022, -0.00001800, 0.00000881, 0.00000272, 0.00000167, -0.00001486, 0.00000464, 0.00000120, -0.00000788, -0.00000344, 0.00000594, 0.00000249, -0.00000440, -0.00000280, 0.00001016, -0.00000069, -0.00000030, -0.00001073, 0.00001151, 0.00000559, -0.00000423, 0.00001226, -0.00001537, },
		{ 0.00000575, -0.00001730, 0.00000271, 0.00000017, 0.00000315, -0.00000908, -0.00000955, 0.00000358, -0.00001192, 0.00001057, -0.00000460, 0.00000463, 0.00000968, -0.00000972, 0.00000140, 0.00000321, 0.00000246, -0.00000288, 0.00000026, -0.00000590, 0.00000368, -0.00000736, 0.00000484, -0.00000424, 0.00000354, -0.00001612, 0.00000600, -0.00003614, 0.00000227, 0.00000393, 0.00004931, 0.00000617, 0.00003636, 0.00002100, 0.00000495, -0.00003037, -0.00000983, -0.00000203, -0.00000803, -0.00001047, 0.00000224, -0.00000104, 0.00001027, -0.00000951, -0.00000244, 0.00000579, -0.00000103, -0.00000450, 0.00000569, -0.00000304, -0.00000633, 0.00000061, -0.00000613, 0.00001065, 0.00000462, -0.00000410, -0.00000466, 0.00000133, 0.00001289, -0.00000438, 0.00000220, -0.00000856, 0.00000564, -0.00001200, 0.00000286, 0.00000143, 0.00000487, -0.00001036, 0.00000268, -0.00000757, 0.00000558, 0.00000413, -0.00000142, -0.00000790, -0.00000604, -0.00002063, -0.00000140, -0.00002158, -0.00000378, -0.00014146, 0.00009046, 0.00017856, 0.00016414, -0.00007645, -0.00007219, -0.00000295, -0.00002833, 0.00000638, -0.00001182, 0.00000804, -0.00001141, 0.00000340, -0.00000925, 0.00000191, -0.00000945, 0.00000771, -0.00000660, 0.00000372, -0.00000979, 0.00000741, 0.00000654, 0.00000081, -0.00000796, 0.00000519, -0.00000118, -0.00000771, -0.00001164, 0.00001341, -0.00000119, -0.00000214, -0.00001368, 0.00001136, -0.00000710, -0.00000452, -0.00000248, 0.00000918, -0.00000312, -0.00000339, 0.00000071, 0.00001001, -0.00000401, -0.00000126, -0.00001031, 0.00002176, -0.00000949, 0.00001059, -0.00000249, -0.00000634, },
		{ -0.00000742, -0.00001051, 0.00000383, 0.00000096, 0.00000061, -0.00001438, -0.00000040, -0.00000368, -0.00000624, 0.00001253, -0.00001130, 0.00001731, -0.00000515, -0.00000247, -0.00000109, 0.00000855, -0.00000576, 0.00000468, -0.00000894, 0.00000448, -0.00000672, 0.00000319, -0.00000475, 0.00000658, -0.00001228, -0.00000004, -0.00001439, -0.00002295, 0.00000434, 0.00002153, 0.00003783, 0.00000569, 0.00004162, 0.00000844, -0.00000244, -0.00003335, 0.00000202, -0.00000782, -0.00000720, -0.00000801, 0.00000368, 0.00000139, 0.00000753, -0.00001508, 0.00000715, 0.00000109, -0.00000159, -0.00000270, 0.00000665, -0.00000921, -0.00000040, -0.00000378, -0.00000068, 0.00001291, -0.00000251, -0.00000232, -0.00000676, 0.00001173, 0.00000292, -0.00000009, -0.00000424, -0.00000131, -0.00000102, -0.00000879, 0.00000379, 0.00000443, -0.00000256, -0.00000570, -0.00000024, -0.00000439, 0.00000602, 0.00000484, -0.00000925, -0.00000018, -0.00002029, -0.00000430, -0.00002064, 0.00000486, -0.00006711, -0.00009645, 0.00016595, 0.00019220, 0.00008929, -0.00013916, -0.00000459, -0.00003633, 0.00000280, -0.00001373, 0.00000726, -0.00000863, 0.00000213, -0.00000912, 0.00000250, -0.00001003, 0.00000364, -0.00000256, 0.00000192, -0.00000592, -0.00000117, 0.00000612, 0.00000836, -0.00000847, 0.00000069, 0.00000114, 0.00000019, -0.00001685, 0.00000339, 0.00000736, 0.00000018, -0.00001067, -0.00000272, 0.00000699, -0.00000837, -0.00000462, 0.00000311, 0.00000587, -0.00000469, -0.00000322, 0.00000648, 0.00000440, -0.00000218, -0.00000808, 0.00000201, 0.00001682, -0.00001270, 0.00001839, -0.00001618, 0.00000724, },
		{ -0.00001807, -0.00000014, 0.00000098, 0.00000374, -0.00000695, -0.00001201, 0.00000469, -0.00001274, 0.00000819, 0.00000033, -0.00000173, 0.00001454, -0.00001071, 0.00000105, 0.00000130, 0.00000596, -0.00000571, 0.00000414, -0.00000965, 0.00000689, -0.00000993, 0.00000733, -0.00000791, 0.00000937, -0.00002029, 0.00001011, -0.00003442, 0.00000008, -0.00000045, 0.00004245, 0.00001244, 0.00002441, 0.00002809, 0.00000693, -0.00001777, -0.00002221, 0.00000286, -0.00000891, -0.00000865, -0.00000200, 0.00000107, 0.00000787, -0.00000420, -0.00000858, 0.00000893, -0.00000141, -0.00000301, 0.00000225, 0.00000104, -0.00000855, 0.00000156, -0.00000639, 0.00000731, 0.00000845, -0.00000530, -0.00000287, -0.00000268, 0.00001515, -0.00000511, 0.00000423, -0.00000995, 0.00000656, -0.00001107, -0.00000009, 0.00000140, 0.00000719, -0.00001085, 0.00000251, -0.00000702, 0.00000270, 0.00000399, 0.00000323, -0.00001224, 0.00000118, -0.00002653, 0.00000395, -0.00002994, 0.00001532, -0.00013251, 0.00000466, 0.00020657, 0.00019118, -0.00003747, -0.00010549, 0.00000820, -0.00004047, 0.00001326, -0.00001918, 0.00001338, -0.00001472, 0.00000632, -0.00001263, 0.00000545, -0.00001325, 0.00000974, -0.00000796, 0.00000578, -0.00001176, 0.00000731, 0.00000438, 0.00000563, -0.00001165, 0.00000687, -0.00000209, -0.00000340, -0.00001707, 0.00001451, -0.00000055, 0.00000051, -0.00001533, 0.00000848, -0.00000214, -0.00000566, -0.00000520, 0.00000920, -0.00000103, -0.00000299, -0.00000257, 0.00001068, -0.00000190, -0.00000034, -0.00001342, 0.00001854, -0.00000054, -0.00000010, 0.00000988, -0.00001622, 0.00001219, },
		{ -0.00001753, 0.00000564, -0.00000072, 0.00000417, -0.00001537, -0.00000172, -0.00000170, -0.00001016, 0.00001503, -0.00001103, 0.00001347, 0.00000046, -0.00000563, -0.00000083, 0.00000764, -0.00000304, 0.00000223, -0.00000513, -0.00000008, -0.00000197, -0.00000124, -0.00000067, 0.00000163, -0.00000372, -0.00000913, -0.00000231, -0.00003106, 0.00000944, 0.00000715, 0.00004485, -0.00000126, 0.00004025, 0.00001055, 0.00000701, -0.00003111, -0.00000602, -0.00000303, -0.00000745, -0.00000887, 0.00000222, 0.00000097, 0.00000932, -0.00001431, 0.00000306, 0.00000451, -0.00000184, -0.00000294, 0.00000533, -0.00000591, -0.00000351, -0.00000105, -0.00000405, 0.00001287, 0.00000028, -0.00000353, -0.00000583, 0.00000771, 0.00000725, -0.00000318, 0.00000024, -0.00000608, 0.00000440, -0.00001309, 0.00000453, 0.00000258, 0.00000255, -0.00001046, 0.00000413, -0.00000864, 0.00000649, 0.00000406, -0.00000390, -0.00000631, -0.00000952, -0.00001658, -0.00000722, -0.00001379, -0.00002023, -0.00014523, 0.00011407, 0.00023211, 0.00012805, -0.00013151, -0.00002958, -0.00002360, -0.00001153, -0.00000363, -0.00000316, 0.00000046, -0.00000460, -0.00000373, -0.00000275, -0.00000475, -0.00000256, 0.00000233, -0.00000167, -0.00000196, -0.00000558, 0.00000752, 0.00000674, -0.00000382, -0.00000467, 0.00000500, -0.00000128, -0.00001271, -0.00000466, 0.00001318, -0.00000243, -0.00000521, -0.00000964, 0.00001037, -0.00000702, -0.00000511, -0.00000103, 0.00000914, -0.00000482, -0.00000227, 0.00000147, 0.00000903, -0.00000350, -0.00000444, -0.00000748, 0.00002424, -0.00001449, 0.00001616, -0.00000840, -0.00000277, 0.00000321, },
		{ -0.00000691, 0.00000385, 0.00000203, -0.00000234, -0.00001609, 0.00000669, -0.00001339, 0.00000481, 0.00000562, -0.00000691, 0.00001693, -0.00000929, -0.00000050, -0.00000015, 0.00000844, -0.00000708, 0.00000617, -0.00001084, 0.00000720, -0.00000970, 0.00000726, -0.00000879, 0.00001092, -0.00001865, 0.00000742, -0.00002632, -0.00000757, 0.00000235, 0.00002903, 0.00002374, 0.00000982, 0.00003592, 0.00000443, -0.00000250, -0.00003110, 0.00000328, -0.00000772, -0.00000711, -0.00000554, 0.00000227, 0.00000562, 0.00000104, -0.00001348, 0.00000982, -0.00000029, -0.00000209, -0.00000012, 0.00000305, -0.00000842, 0.00000043, -0.00000470, 0.00000282, 0.00001186, -0.00000546, -0.00000206, -0.00000517, 0.00001535, -0.00000380, 0.00000375, -0.00000823, 0.00000407, -0.00000672, -0.00000525, 0.00000279, 0.00000700, -0.00000778, -0.00000137, -0.00000279, -0.00000256, 0.00000510, 0.00000511, -0.00001175, 0.00000134, -0.00002304, -0.00000135, -0.00002489, 0.00001062, -0.00008999, -0.00008627, 0.00020297, 0.00023443, 0.00000220, -0.00013017, 0.00000894, -0.00004425, 0.00001220, -0.00001931, 0.00001167, -0.00001218, 0.00000487, -0.00001235, 0.00000547, -0.00001338, 0.00000825, -0.00000666, 0.00000572, -0.00001150, 0.00000538, 0.00000367, 0.00000810, -0.00001208, 0.00000532, -0.00000062, -0.00000136, -0.00001876, 0.00001106, 0.00000373, -0.00000035, -0.00001265, 0.00000213, 0.00000367, -0.00000636, -0.00000664, 0.00000655, 0.00000273, -0.00000356, -0.00000334, 0.00000770, 0.00000271, -0.00000147, -0.00001216, 0.00000951, 0.00001217, -0.00001139, 0.00001939, -0.00002008, 0.00001119, -0.00001222, },
		{ 0.00000435, -0.00000080, 0.00000600, -0.00001382, -0.00000545, 0.00000235, -0.00001501, 0.00001662, -0.00000895, 0.00000836, 0.00000588, -0.00000767, -0.00000106, 0.00000607, 0.00000051, -0.00000104, -0.00000054, -0.00000488, 0.00000254, -0.00000538, 0.00000360, -0.00000407, 0.00000482, -0.00001664, 0.00000625, -0.00003473, 0.00001068, -0.00000106, 0.00004425, 0.00000008, 0.00003162, 0.00001719, 0.00000818, -0.00001942, -0.00001814, 0.00000232, -0.00000839, -0.00000763, -0.00000117, 0.00000194, 0.00000903, -0.00001066, -0.00000324, 0.00000851, -0.00000247, -0.00000193, 0.00000280, -0.00000239, -0.00000595, 0.00000050, -0.00000542, 0.00001045, 0.00000451, -0.00000543, -0.00000409, 0.00000310, 0.00001160, -0.00000625, 0.00000429, -0.00000971, 0.00000791, -0.00001461, 0.00000310, 0.00000150, 0.00000663, -0.00001315, 0.00000581, -0.00000951, 0.00000451, 0.00000343, 0.00000165, -0.00001195, -0.00000062, -0.00002476, 0.00000099, -0.00002598, 0.00000759, -0.00014666, 0.00002174, 0.00026801, 0.00017477, -0.00011455, -0.00005973, -0.00000972, -0.00002657, 0.00000617, -0.00001273, 0.00000760, -0.00000922, 0.00000037, -0.00000735, 0.00000010, -0.00000799, 0.00000611, -0.00000460, 0.00000197, -0.00000988, 0.00000913, 0.00000404, 0.00000168, -0.00000993, 0.00000819, -0.00000274, -0.00000777, -0.00001243, 0.00001708, -0.00000316, -0.00000160, -0.00001318, 0.00000961, -0.00000336, -0.00000549, -0.00000483, 0.00001040, -0.00000337, -0.00000158, -0.00000242, 0.00001076, -0.00000232, -0.00000230, -0.00001328, 0.00002430, -0.00000767, 0.00000595, 0.00000510, -0.00001480, 0.00001222, -0.00002082, },
		{ 0.00000729, -0.00000095, 0.00000347, -0.00002009, 0.00000768, -0.00001215, -0.00000052, 0.00001134, -0.00001068, 0.00001700, -0.00000617, -0.00000243, -0.00000142, 0.00000977, -0.00000672, 0.00000599, -0.00000929, 0.00000477, -0.00000701, 0.00000480, -0.00000666, 0.00000799, -0.00001183, -0.00000021, -0.00001464, -0.00001750, 0.00000836, 0.00001393, 0.00003471, -0.00000209, 0.00003989, 0.00000323, 0.00000799, -0.00003087, -0.00000318, -0.00000340, -0.00000724, -0.00000670, 0.00000115, 0.00000482, 0.00000458, -0.00001531, 0.00000718, 0.00000302, -0.00000251, -0.00000051, 0.00000254, -0.00000618, -0.00000208, -0.00000215, -0.00000134, 0.00001347, -0.00000364, -0.00000282, -0.00000572, 0.00001321, -0.00000022, 0.00000065, -0.00000376, -0.00000093, -0.00000039, -0.00001107, 0.00000455, 0.00000500, -0.00000222, -0.00000731, 0.00000294, -0.00000805, 0.00000631, 0.00000446, -0.00000676, -0.00000435, -0.00001281, -0.00001315, -0.00001207, -0.00000693, -0.00003869, -0.00014824, 0.00014778, 0.00029448, 0.00004402, -0.00014441, -0.00000180, -0.00003925, 0.00000311, -0.00001264, 0.00000404, -0.00000549, 0.00000062, -0.00000946, 0.00000265, -0.00001047, 0.00000407, -0.00000341, 0.00000384, -0.00000895, 0.00000135, 0.00000502, 0.00000751, -0.00000893, 0.00000069, 0.00000307, -0.00000180, -0.00001650, 0.00000358, 0.00001050, -0.00000328, -0.00000765, -0.00000499, 0.00000805, -0.00000573, -0.00000701, 0.00000211, 0.00000676, -0.00000417, -0.00000244, 0.00000277, 0.00000762, -0.00000279, -0.00000879, -0.00000122, 0.00002280, -0.00001702, 0.00002055, -0.00001447, 0.00000196, -0.00000112, -0.00001552, },
		{ 0.00000200, 0.00000460, -0.00000850, -0.00001216, 0.00000829, -0.00001998, 0.00001660, -0.00000516, 0.00000282, 0.00001012, -0.00000807, -0.00000191, 0.00000415, 0.00000408, -0.00000395, 0.00000327, -0.00000815, 0.00000512, -0.00000765, 0.00000627, -0.00000798, 0.00000996, -0.00001902, 0.00000831, -0.00003173, 0.00000609, -0.00000168, 0.00003561, 0.00000900, 0.00001733, 0.00002713, 0.00000349, -0.00000388, -0.00002811, 0.00000401, -0.00000763, -0.00000632, -0.00000431, 0.00000243, 0.00000819, -0.00000591, -0.00000940, 0.00001083, -0.00000178, -0.00000128, 0.00000097, -0.00000050, -0.00000636, 0.00000005, -0.00000457, 0.00000623, 0.00000899, -0.00000669, -0.00000286, -0.00000068, 0.00001445, -0.00000750, 0.00000585, -0.00001011, 0.00000767, -0.00001206, -0.00000130, 0.00000233, 0.00000790, -0.00001181, 0.00000308, -0.00000589, -0.00000056, 0.00000441, 0.00000438, -0.00001278, 0.00000119, -0.00002332, -0.00000193, -0.00002402, 0.00000893, -0.00010700, -0.00008278, 0.00026973, 0.00024345, -0.00009494, -0.00008742, -0.00000177, -0.00003603, 0.00000991, -0.00001636, 0.00000866, -0.00000884, 0.00000100, -0.00000924, 0.00000238, -0.00001061, 0.00000710, -0.00000555, 0.00000445, -0.00001242, 0.00000919, 0.00000238, 0.00000551, -0.00001256, 0.00000856, -0.00000223, -0.00000443, -0.00001699, 0.00001635, 0.00000012, -0.00000169, -0.00001213, 0.00000476, 0.00000187, -0.00000586, -0.00000712, 0.00000868, -0.00000000, -0.00000187, -0.00000389, 0.00000873, 0.00000160, -0.00000235, -0.00001432, 0.00001724, 0.00000505, -0.00000723, 0.00001769, -0.00002202, 0.00001419, -0.00001712, -0.00000164, },
		{ -0.00000271, 0.00000772, -0.00002109, 0.00000510, -0.00000717, -0.00000850, 0.00001739, -0.00001308, 0.00001545, -0.00000256, -0.00000371, -0.00000298, 0.00001014, -0.00000498, 0.00000422, -0.00000600, 0.00000102, -0.00000338, 0.00000136, -0.00000288, 0.00000258, -0.00000306, -0.00000845, -0.00000420, -0.00002493, 0.00001246, 0.00000262, 0.00003993, -0.00000645, 0.00003605, 0.00000779, 0.00000949, -0.00002065, -0.00001466, 0.00000220, -0.00000834, -0.00000570, -0.00000190, 0.00000523, 0.00000619, -0.00001387, 0.00000153, 0.00000755, -0.00000358, 0.00000048, 0.00000061, -0.00000336, -0.00000436, -0.00000034, -0.00000375, 0.00001247, 0.00000017, -0.00000471, -0.00000496, 0.00000957, 0.00000458, -0.00000359, 0.00000140, -0.00000591, 0.00000509, -0.00001486, 0.00000456, 0.00000329, 0.00000307, -0.00001195, 0.00000653, -0.00001045, 0.00000522, 0.00000363, -0.00000085, -0.00001061, -0.00000312, -0.00002238, -0.00000269, -0.00002057, -0.00000438, -0.00015860, 0.00004262, 0.00035241, 0.00010107, -0.00015389, -0.00001873, -0.00003074, -0.00000892, -0.00000433, -0.00000428, 0.00000042, -0.00000247, -0.00000696, -0.00000049, -0.00000695, -0.00000043, -0.00000028, 0.00000162, -0.00000532, -0.00000359, 0.00000734, 0.00000515, -0.00000398, -0.00000495, 0.00000703, -0.00000308, -0.00001224, -0.00000512, 0.00001648, -0.00000504, -0.00000377, -0.00000951, 0.00000870, -0.00000286, -0.00000716, -0.00000227, 0.00000896, -0.00000332, -0.00000146, -0.00000162, 0.00001018, -0.00000211, -0.00000609, -0.00000935, 0.00002661, -0.00001350, 0.00001218, -0.00000105, -0.00001143, 0.00001021, -0.00002163, 0.00000827, },
		{ 0.00000063, -0.00000054, -0.00001999, 0.00001377, -0.00002266, 0.00001336, 0.00000077, -0.00000294, 0.00001314, -0.00000731, -0.00000269, 0.00000170, 0.00000733, -0.00000589, 0.00000552, -0.00000915, 0.00000531, -0.00000772, 0.00000675, -0.00000896, 0.00001046, -0.00001570, 0.00000449, -0.00002369, -0.00000228, 0.00000260, 0.00002262, 0.00002082, 0.00000270, 0.00003524, -0.00000073, 0.00000764, -0.00002972, -0.00000101, -0.00000366, -0.00000658, -0.00000531, 0.00000090, 0.00000820, -0.00000197, -0.00001312, 0.00000965, 0.00000127, -0.00000205, 0.00000105, -0.00000102, -0.00000462, -0.00000201, -0.00000240, 0.00000168, 0.00001212, -0.00000607, -0.00000306, -0.00000281, 0.00001499, -0.00000608, 0.00000417, -0.00000699, 0.00000395, -0.00000634, -0.00000749, 0.00000433, 0.00000667, -0.00000703, -0.00000290, 0.00000041, -0.00000681, 0.00000610, 0.00000406, -0.00000841, -0.00000386, -0.00001376, -0.00001314, -0.00001200, -0.00000670, -0.00005192, -0.00016013, 0.00021089, 0.00033708, -0.00007044, -0.00011113, -0.00000180, -0.00003823, 0.00000627, -0.00001338, 0.00000386, -0.00000449, -0.00000081, -0.00000894, 0.00000220, -0.00001048, 0.00000557, -0.00000474, 0.00000527, -0.00001261, 0.00000691, 0.00000285, 0.00000641, -0.00001163, 0.00000557, 0.00000075, -0.00000362, -0.00001729, 0.00001067, 0.00000691, -0.00000460, -0.00000804, -0.00000182, 0.00000649, -0.00000532, -0.00000807, 0.00000475, 0.00000398, -0.00000236, -0.00000344, 0.00000423, 0.00000643, -0.00000314, -0.00001230, 0.00000673, 0.00001782, -0.00001642, 0.00002266, -0.00001950, 0.00000686, -0.00000664, -0.00001107, 0.00000671, },
		{ 0.00000802, -0.00001701, -0.00000234, 0.00000174, -0.00001826, 0.00002264, -0.00001363, 0.00001265, 0.00000081, -0.00000395, -0.00000475, 0.00000959, -0.00000233, 0.00000175, -0.00000235, -0.00000237, -0.00000049, -0.00000148, 0.00000070, -0.00000259, 0.00000394, -0.00001313, 0.00000120, -0.00002668, 0.00001161, -0.00000204, 0.00003715, -0.00000206, 0.00002443, 0.00001786, 0.00000378, -0.00000538, -0.00002528, 0.00000496, -0.00000807, -0.00000439, -0.00000492, 0.00000525, 0.00000700, -0.00001055, -0.00000497, 0.00001102, -0.00000361, 0.00000126, -0.00000079, -0.00000190, -0.00000487, -0.00000037, -0.00000385, 0.00000916, 0.00000489, -0.00000642, -0.00000420, 0.00000588, 0.00000889, -0.00000683, 0.00000469, -0.00000818, 0.00000715, -0.00001452, 0.00000127, 0.00000360, 0.00000566, -0.00001254, 0.00000536, -0.00000776, 0.00000041, 0.00000472, 0.00000222, -0.00001210, -0.00000054, -0.00002175, -0.00000495, -0.00001940, 0.00000079, -0.00011922, -0.00008391, 0.00037300, 0.00019144, -0.00016582, -0.00003497, -0.00002554, -0.00001742, -0.00000057, -0.00000766, 0.00000091, -0.00000113, -0.00000723, -0.00000171, -0.00000512, -0.00000301, 0.00000092, 0.00000049, -0.00000212, -0.00000768, 0.00000872, 0.00000327, 0.00000011, -0.00000886, 0.00000883, -0.00000296, -0.00000867, -0.00001161, 0.00001828, -0.00000280, -0.00000355, -0.00000948, 0.00000497, 0.00000205, -0.00000751, -0.00000514, 0.00000809, -0.00000047, -0.00000131, -0.00000385, 0.00000907, 0.00000134, -0.00000537, -0.00001286, 0.00002258, -0.00000225, -0.00000159, 0.00001382, -0.00002183, 0.00001552, -0.00002098, 0.00000421, -0.00000110, },
		{ 0.00000686, -0.00002533, 0.00001528, -0.00002026, 0.00000545, 0.00000914, -0.00000893, 0.00001538, -0.00000636, -0.00000250, -0.00000160, 0.00001020, -0.00000680, 0.00000628, -0.00000836, 0.00000399, -0.00000658, 0.00000583, -0.00000768, 0.00000753, -0.00000917, -0.00000150, -0.00001495, -0.00001003, 0.00000699, 0.00001073, 0.00002956, -0.00000626, 0.00003614, 0.00000121, 0.00000994, -0.00002106, -0.00001178, 0.00000223, -0.00000809, -0.00000403, -0.00000261, 0.00000919, 0.00000045, -0.00001355, 0.00000491, 0.00000616, -0.00000388, 0.00000276, -0.00000316, -0.00000201, -0.00000422, -0.00000047, -0.00000161, 0.00001278, -0.00000326, -0.00000448, -0.00000361, 0.00001365, -0.00000263, 0.00000036, -0.00000219, -0.00000083, -0.00000030, -0.00001258, 0.00000511, 0.00000518, -0.00000166, -0.00000854, 0.00000534, -0.00001042, 0.00000567, 0.00000332, -0.00000255, -0.00001031, -0.00000371, -0.00002291, -0.00000221, -0.00002046, -0.00001217, -0.00017972, 0.00008733, 0.00043632, -0.00002465, -0.00013814, -0.00000468, -0.00003785, -0.00000017, -0.00000860, -0.00000172, -0.00000105, -0.00000116, -0.00000931, 0.00000192, -0.00000982, 0.00000368, -0.00000399, 0.00000561, -0.00001130, 0.00000315, 0.00000468, 0.00000532, -0.00000847, 0.00000089, 0.00000459, -0.00000409, -0.00001480, 0.00000250, 0.00001404, -0.00000705, -0.00000442, -0.00000648, 0.00000778, -0.00000256, -0.00000875, 0.00000065, 0.00000633, -0.00000152, -0.00000259, -0.00000029, 0.00000947, -0.00000238, -0.00001011, -0.00000259, 0.00002495, -0.00001666, 0.00001714, -0.00000736, -0.00000682, 0.00000623, -0.00001973, 0.00001019, -0.00000320, },
		{ -0.00000831, -0.00001310, 0.00001210, -0.00002684, 0.00002468, -0.00001110, 0.00000829, 0.00000414, -0.00000376, -0.00000587, 0.00000765, 0.00000107, -0.00000086, 0.00000080, -0.00000450, 0.00000099, -0.00000312, 0.00000312, -0.00000587, 0.00000715, -0.00001279, 0.00000112, -0.00002330, 0.00000672, -0.00000135, 0.00002857, 0.00000766, 0.00000967, 0.00002836, -0.00000267, 0.00000688, -0.00002847, 0.00000130, -0.00000455, -0.00000463, -0.00000587, 0.00000325, 0.00000843, -0.00000731, -0.00000976, 0.00001122, -0.00000113, 0.00000044, -0.00000017, -0.00000290, -0.00000314, -0.00000218, -0.00000224, 0.00000486, 0.00000888, -0.00000657, -0.00000451, 0.00000318, 0.00001151, -0.00000767, 0.00000470, -0.00000673, 0.00000531, -0.00001023, -0.00000468, 0.00000565, 0.00000535, -0.00000901, -0.00000007, -0.00000125, -0.00000658, 0.00000722, 0.00000168, -0.00000765, -0.00000590, -0.00001152, -0.00001740, -0.00000581, -0.00001546, -0.00005657, -0.00018595, 0.00032069, 0.00032322, -0.00017802, -0.00004994, -0.00002558, -0.00002028, -0.00000328, -0.00000481, -0.00000427, 0.00000394, -0.00000949, -0.00000136, -0.00000504, -0.00000354, 0.00000023, 0.00000041, 0.00000020, -0.00000998, 0.00000801, 0.00000326, 0.00000177, -0.00000965, 0.00000743, -0.00000049, -0.00000726, -0.00001406, 0.00001472, 0.00000370, -0.00000671, -0.00000588, -0.00000121, 0.00000699, -0.00000721, -0.00000659, 0.00000483, 0.00000315, -0.00000135, -0.00000408, 0.00000513, 0.00000613, -0.00000566, -0.00001276, 0.00001359, 0.00001127, -0.00001351, 0.00002247, -0.00002290, 0.00001105, -0.00001237, -0.00000475, 0.00000382, 0.00000423, },
		{ -0.00002489, 0.00001033, -0.00001157, -0.00000685, 0.00001887, -0.00001439, 0.00001687, -0.00000570, -0.00000109, -0.00000548, 0.00001235, -0.00000672, 0.00000604, -0.00000675, 0.00000249, -0.00000555, 0.00000475, -0.00000562, 0.00000359, -0.00000301, -0.00000546, -0.00000961, -0.00001360, 0.00000797, 0.00000366, 0.00003160, -0.00000664, 0.00002820, 0.00001027, 0.00000384, -0.00000623, -0.00002300, 0.00000617, -0.00000869, -0.00000210, -0.00000636, 0.00000982, 0.00000209, -0.00001163, -0.00000157, 0.00001077, -0.00000512, 0.00000453, -0.00000502, -0.00000051, -0.00000494, -0.00000008, -0.00000279, 0.00001097, 0.00000084, -0.00000590, -0.00000429, 0.00001175, 0.00000118, -0.00000329, 0.00000146, -0.00000357, 0.00000289, -0.00001376, 0.00000244, 0.00000579, 0.00000121, -0.00001026, 0.00000530, -0.00000836, 0.00000077, 0.00000508, 0.00000011, -0.00001165, -0.00000130, -0.00002190, -0.00000547, -0.00001765, -0.00000580, -0.00013732, -0.00007298, 0.00049953, 0.00006781, -0.00017914, -0.00000259, -0.00004220, -0.00000242, -0.00000884, -0.00000171, -0.00000387, 0.00000380, -0.00001316, 0.00000401, -0.00001110, 0.00000381, -0.00000522, 0.00000691, -0.00000998, -0.00000030, 0.00000588, 0.00000449, -0.00000573, -0.00000270, 0.00000675, -0.00000352, -0.00001232, -0.00000443, 0.00001759, -0.00000555, -0.00000447, -0.00000658, 0.00000438, 0.00000276, -0.00000992, -0.00000191, 0.00000553, 0.00000139, -0.00000239, -0.00000296, 0.00000895, 0.00000114, -0.00000955, -0.00000781, 0.00002409, -0.00000782, 0.00000396, 0.00000877, -0.00001980, 0.00001479, -0.00002261, 0.00000899, -0.00000384, 0.00001031, },
		{ -0.00002315, 0.00001957, -0.00003031, 0.00002088, -0.00000432, 0.00000182, 0.00000812, -0.00000430, -0.00000520, 0.00000382, 0.00000524, -0.00000348, 0.00000336, -0.00000560, 0.00000164, -0.00000421, 0.00000475, -0.00000733, 0.00000711, -0.00000914, -0.00000147, -0.00001850, 0.00000156, 0.00000058, 0.00001906, 0.00001683, -0.00000168, 0.00003289, -0.00000311, 0.00000995, -0.00002123, -0.00000879, 0.00000158, -0.00000656, -0.00000418, -0.00000109, 0.00001089, -0.00000528, -0.00001135, 0.00000743, 0.00000383, -0.00000205, 0.00000258, -0.00000597, -0.00000013, -0.00000459, -0.00000032, 0.00000112, 0.00001087, -0.00000467, -0.00000579, 0.00000132, 0.00001257, -0.00000636, 0.00000220, -0.00000309, 0.00000175, -0.00000477, -0.00001030, 0.00000690, 0.00000445, -0.00000425, -0.00000602, 0.00000470, -0.00001146, 0.00000780, 0.00000050, -0.00000122, -0.00001343, 0.00000019, -0.00002947, 0.00000662, -0.00003176, -0.00000760, -0.00022221, 0.00018175, 0.00047911, -0.00017066, -0.00007424, -0.00002420, -0.00002288, -0.00000750, -0.00000090, -0.00000962, 0.00000730, -0.00000926, -0.00000276, -0.00000406, -0.00000445, 0.00000002, -0.00000058, 0.00000282, -0.00001123, 0.00000592, 0.00000448, 0.00000171, -0.00000838, 0.00000419, 0.00000303, -0.00000712, -0.00001349, 0.00000782, 0.00001152, -0.00001003, -0.00000228, -0.00000598, 0.00000887, -0.00000455, -0.00000794, 0.00000137, 0.00000514, -0.00000001, -0.00000375, 0.00000072, 0.00000955, -0.00000464, -0.00001187, 0.00000443, 0.00002071, -0.00001730, 0.00002031, -0.00001304, -0.00000173, 0.00000084, -0.00001502, 0.00000922, -0.00000113, 0.00000178, },
		{ -0.00000055, 0.00000151, -0.00002060, 0.00002712, -0.00001746, 0.00001667, -0.00000473, 0.00000059, -0.00000869, 0.00001289, -0.00000537, 0.00000500, -0.00000494, 0.00000152, -0.00000516, 0.00000399, -0.00000367, 0.00000040, 0.00000047, -0.00000544, -0.00000870, -0.00001286, 0.00000564, 0.00000119, 0.00002774, -0.00000043, 0.00001501, 0.00002184, -0.00000429, 0.00000672, -0.00002778, 0.00000409, -0.00000611, -0.00000163, -0.00000813, 0.00000830, 0.00000438, -0.00000929, -0.00000706, 0.00001253, -0.00000396, 0.00000466, -0.00000471, -0.00000161, -0.00000333, -0.00000154, -0.00000210, 0.00000772, 0.00000465, -0.00000591, -0.00000591, 0.00001035, 0.00000399, -0.00000503, 0.00000206, -0.00000267, 0.00000207, -0.00001049, -0.00000365, 0.00000873, 0.00000078, -0.00000723, -0.00000003, -0.00000114, -0.00000771, 0.00000945, -0.00000198, -0.00000561, -0.00000876, -0.00000868, -0.00002227, 0.00000176, -0.00002782, -0.00005949, -0.00021667, 0.00047829, 0.00022492, -0.00023556, 0.00000565, -0.00005360, 0.00000224, -0.00001678, 0.00000593, -0.00001384, 0.00001376, -0.00001988, 0.00000825, -0.00001447, 0.00000619, -0.00000840, 0.00000910, -0.00000909, -0.00000238, 0.00000512, 0.00000552, -0.00000519, -0.00000350, 0.00000583, -0.00000058, -0.00001172, -0.00000769, 0.00001557, 0.00000093, -0.00000861, -0.00000246, -0.00000227, 0.00000887, -0.00001081, -0.00000288, 0.00000212, 0.00000522, -0.00000235, -0.00000362, 0.00000530, 0.00000643, -0.00001012, -0.00000925, 0.00001719, 0.00000538, -0.00000977, 0.00002066, -0.00002461, 0.00001386, -0.00001704, 0.00000196, -0.00000019, 0.00000849, -0.00001796, },
		{ 0.00002021, -0.00002595, 0.00001029, 0.00000646, -0.00000636, 0.00001289, -0.00000615, -0.00000234, -0.00000174, 0.00000980, -0.00000598, 0.00000558, -0.00000642, 0.00000259, -0.00000584, 0.00000652, -0.00000808, 0.00000584, -0.00000549, -0.00000251, -0.00001634, -0.00000025, 0.00000019, 0.00001266, 0.00002100, -0.00000536, 0.00002826, 0.00000473, 0.00000358, -0.00000669, -0.00002070, 0.00000682, -0.00000828, -0.00000120, -0.00000632, 0.00001332, -0.00000411, -0.00001017, 0.00000099, 0.00000950, -0.00000479, 0.00000612, -0.00000930, 0.00000223, -0.00000595, 0.00000058, -0.00000104, 0.00001066, -0.00000156, -0.00000679, -0.00000095, 0.00001334, -0.00000451, -0.00000049, -0.00000031, -0.00000007, -0.00000183, -0.00001209, 0.00000436, 0.00000603, -0.00000209, -0.00000800, 0.00000533, -0.00000991, 0.00000286, 0.00000284, 0.00000117, -0.00001494, 0.00000304, -0.00002892, 0.00000333, -0.00002806, 0.00000131, -0.00017873, -0.00001985, 0.00061185, -0.00010876, -0.00012546, -0.00000983, -0.00003491, -0.00000431, -0.00000448, -0.00000706, 0.00000200, -0.00000132, -0.00000982, 0.00000125, -0.00000901, 0.00000332, -0.00000491, 0.00000752, -0.00001312, 0.00000444, 0.00000472, 0.00000241, -0.00000783, 0.00000203, 0.00000485, -0.00000577, -0.00001278, 0.00000154, 0.00001623, -0.00000936, -0.00000261, -0.00000590, 0.00000548, 0.00000146, -0.00001041, -0.00000018, 0.00000360, 0.00000371, -0.00000410, -0.00000200, 0.00000940, -0.00000065, -0.00001245, -0.00000130, 0.00002243, -0.00001116, 0.00000856, 0.00000333, -0.00001637, 0.00001200, -0.00002120, 0.00001112, -0.00000436, 0.00000903, -0.00002805, },
		{ 0.00001436, -0.00003105, 0.00003015, -0.00001585, 0.00001338, -0.00000213, 0.00000090, -0.00000957, 0.00001075, -0.00000226, 0.00000293, -0.00000277, 0.00000070, -0.00000470, 0.00000279, -0.00000145, -0.00000199, 0.00000150, -0.00000278, -0.00000964, -0.00001114, 0.00000341, -0.00000017, 0.00002216, 0.00000693, 0.00000324, 0.00002884, -0.00000684, 0.00001066, -0.00002207, -0.00000499, -0.00000026, -0.00000332, -0.00000679, 0.00000382, 0.00000812, -0.00000808, -0.00000949, 0.00001001, 0.00000027, 0.00000245, -0.00000156, -0.00000543, 0.00000011, -0.00000404, -0.00000061, 0.00000443, 0.00000694, -0.00000403, -0.00000832, 0.00000921, 0.00000589, -0.00000493, 0.00000013, 0.00000059, -0.00000080, -0.00000541, -0.00001017, 0.00001132, -0.00000055, -0.00000237, -0.00000688, 0.00000662, -0.00001503, 0.00001263, -0.00000550, 0.00000360, -0.00002004, 0.00000811, -0.00004111, 0.00002268, -0.00005388, 0.00000924, -0.00028804, 0.00034333, 0.00043724, -0.00028725, 0.00000679, -0.00006347, 0.00000727, -0.00002667, 0.00001539, -0.00002470, 0.00002242, -0.00002411, 0.00001049, -0.00001653, 0.00000769, -0.00001057, 0.00000968, -0.00000707, -0.00000402, 0.00000317, 0.00000770, -0.00000609, -0.00000252, 0.00000304, 0.00000353, -0.00001229, -0.00000789, 0.00000954, 0.00000976, -0.00001371, 0.00000226, -0.00000814, 0.00001229, -0.00000931, -0.00000393, -0.00000144, 0.00000730, -0.00000073, -0.00000368, 0.00000077, 0.00001073, -0.00000948, -0.00000957, 0.00000892, 0.00001620, -0.00001667, 0.00002207, -0.00001780, 0.00000315, -0.00000494, -0.00000839, 0.00000576, 0.00000320, -0.00000718, -0.00001358, },
		{ -0.00001429, -0.00000505, 0.00001871, -0.00001432, 0.00001731, -0.00000853, 0.00000166, -0.00000762, 0.00001358, -0.00000776, 0.00000735, -0.00000732, 0.00000433, -0.00000832, 0.00000866, -0.00000877, 0.00000495, -0.00000416, 0.00000012, -0.00001812, 0.00000169, -0.00000273, 0.00000975, 0.00001992, -0.00000207, 0.00001690, 0.00001697, -0.00000626, 0.00000729, -0.00002740, 0.00000677, -0.00000727, 0.00000082, -0.00001004, 0.00001378, -0.00000238, -0.00000804, -0.00000538, 0.00001316, -0.00000574, 0.00000846, -0.00001074, 0.00000213, -0.00000511, -0.00000013, -0.00000161, 0.00000913, 0.00000121, -0.00000596, -0.00000477, 0.00001492, -0.00000362, -0.00000130, -0.00000078, 0.00000208, -0.00000332, -0.00000864, -0.00000280, 0.00001108, -0.00000418, -0.00000411, -0.00000091, -0.00000117, -0.00000791, 0.00001002, -0.00000352, -0.00000629, -0.00000757, -0.00001146, -0.00001959, -0.00000007, -0.00003009, -0.00007939, -0.00022436, 0.00065758, 0.00004103, -0.00021414, 0.00002320, -0.00006160, 0.00001138, -0.00002040, 0.00000745, -0.00001458, 0.00001541, -0.00002303, 0.00001149, -0.00001803, 0.00001093, -0.00001301, 0.00001452, -0.00001632, 0.00000492, 0.00000226, 0.00000587, -0.00001016, 0.00000306, 0.00000309, -0.00000144, -0.00001427, -0.00000104, 0.00001495, -0.00000268, -0.00000818, -0.00000048, -0.00000244, 0.00000956, -0.00001353, 0.00000051, -0.00000116, 0.00000881, -0.00000478, -0.00000255, 0.00000560, 0.00000578, -0.00001445, -0.00000311, 0.00001727, 0.00000124, -0.00000625, 0.00001788, -0.00002471, 0.00001477, -0.00001931, 0.00000708, -0.00000317, 0.00001077, -0.00002555, 0.00001295, },
		{ -0.00003394, 0.00002540, -0.00000843, 0.00000637, 0.00000295, -0.00000135, -0.00000711, 0.00000554, 0.00000267, -0.00000032, 0.00000021, -0.00000094, -0.00000283, -0.00000011, 0.00000204, -0.00000464, 0.00000203, -0.00000078, -0.00000829, -0.00001267, 0.00000501, -0.00000401, 0.00001896, 0.00001050, -0.00000180, 0.00002665, -0.00000012, 0.00000408, -0.00000781, -0.00001740, 0.00000596, -0.00000578, -0.00000316, -0.00000264, 0.00001264, -0.00000816, -0.00000857, 0.00000384, 0.00000640, -0.00000122, 0.00000357, -0.00001056, 0.00000373, -0.00000616, 0.00000053, 0.00000195, 0.00000763, -0.00000127, -0.00000989, 0.00000673, 0.00000841, -0.00000482, -0.00000175, 0.00000274, -0.00000172, -0.00000291, -0.00001267, 0.00000956, 0.00000127, -0.00000051, -0.00000951, 0.00000867, -0.00001518, 0.00000924, -0.00000449, 0.00000794, -0.00002453, 0.00001502, -0.00004568, 0.00002522, -0.00005637, 0.00003043, -0.00025751, 0.00010764, 0.00065673, -0.00029061, -0.00002533, -0.00005432, -0.00000129, -0.00002619, 0.00001510, -0.00002556, 0.00002016, -0.00001821, 0.00000467, -0.00001203, 0.00000346, -0.00000723, 0.00000492, -0.00000114, -0.00000766, 0.00000267, 0.00000819, -0.00000536, -0.00000305, 0.00000179, 0.00000562, -0.00001118, -0.00000821, 0.00000379, 0.00001592, -0.00001484, 0.00000279, -0.00000881, 0.00000981, -0.00000349, -0.00000700, -0.00000238, 0.00000514, 0.00000389, -0.00000467, -0.00000210, 0.00001133, -0.00000544, -0.00001148, 0.00000357, 0.00001977, -0.00001324, 0.00001239, -0.00000222, -0.00001200, 0.00000759, -0.00001679, 0.00000996, -0.00000179, 0.00000292, -0.00002356, 0.00002182, },
		{ -0.00002040, 0.00002822, -0.00001902, 0.00001922, -0.00000951, 0.00000462, -0.00001158, 0.00001488, -0.00000773, 0.00000796, -0.00000773, 0.00000610, -0.00001055, 0.00001004, -0.00000868, 0.00000434, -0.00000492, 0.00000508, -0.00002124, 0.00000398, -0.00000473, 0.00000705, 0.00001729, 0.00000324, 0.00000538, 0.00002607, -0.00001119, 0.00001268, -0.00002379, -0.00000058, -0.00000264, 0.00000053, -0.00001050, 0.00001087, 0.00000119, -0.00000696, -0.00000901, 0.00001264, -0.00000350, 0.00000820, -0.00000888, -0.00000120, -0.00000205, -0.00000207, -0.00000130, 0.00000739, 0.00000249, -0.00000286, -0.00000985, 0.00001688, -0.00000351, -0.00000021, -0.00000419, 0.00000713, -0.00000753, -0.00000236, -0.00001168, 0.00001693, -0.00000817, 0.00000290, -0.00001043, 0.00001022, -0.00001973, 0.00001816, -0.00001203, 0.00000859, -0.00002595, 0.00001453, -0.00005054, 0.00003656, -0.00007543, 0.00002372, -0.00035874, 0.00056467, 0.00028555, -0.00032657, 0.00006301, -0.00009471, 0.00003334, -0.00004326, 0.00002842, -0.00003632, 0.00003448, -0.00003666, 0.00002208, -0.00002770, 0.00001931, -0.00002144, 0.00002083, -0.00001887, 0.00000627, -0.00000212, 0.00001137, -0.00001461, 0.00000637, -0.00000117, 0.00000486, -0.00001747, -0.00000011, 0.00000865, 0.00000786, -0.00001617, 0.00000683, -0.00001085, 0.00001602, -0.00001484, 0.00000146, -0.00000643, 0.00001235, -0.00000389, -0.00000229, 0.00000032, 0.00001191, -0.00001556, -0.00000350, 0.00000967, 0.00001307, -0.00001597, 0.00002297, -0.00002152, 0.00000707, -0.00000968, -0.00000170, 0.00000163, 0.00000758, -0.00001701, 0.00000053, 0.00000063, },
		{ 0.00001332, 0.00000342, -0.00000312, 0.00000980, -0.00000576, -0.00000195, -0.00000167, 0.00000845, -0.00000426, 0.00000402, -0.00000389, 0.00000110, -0.00000518, 0.00000712, -0.00000821, 0.00000375, -0.00000203, -0.00000219, -0.00001911, 0.00001114, -0.00001071, 0.00001889, 0.00000926, 0.00000037, 0.00001641, 0.00001295, -0.00000773, 0.00000758, -0.00002626, 0.00000810, -0.00000649, 0.00000062, -0.00000879, 0.00001610, -0.00000829, -0.00000584, -0.00000353, 0.00001187, -0.00000439, 0.00000871, -0.00001478, 0.00000561, -0.00000669, 0.00000075, 0.00000027, 0.00000767, 0.00000050, -0.00000860, 0.00000124, 0.00001308, -0.00000662, -0.00000089, 0.00000060, 0.00000251, -0.00000604, -0.00000849, 0.00000128, 0.00000857, -0.00000472, -0.00000437, 0.00000153, -0.00000529, -0.00000319, 0.00000467, 0.00000173, -0.00001478, 0.00000334, -0.00002666, -0.00000062, -0.00002318, -0.00000625, -0.00014012, -0.00016790, 0.00080530, -0.00019362, -0.00011551, -0.00001131, -0.00003515, -0.00000502, -0.00000431, -0.00000828, 0.00000067, 0.00000209, -0.00001234, 0.00000194, -0.00000935, 0.00000420, -0.00000709, 0.00001015, -0.00001497, 0.00000595, 0.00000422, 0.00000002, -0.00000818, 0.00000506, 0.00000288, -0.00000572, -0.00001179, 0.00000234, 0.00001534, -0.00000864, -0.00000349, -0.00000250, 0.00000097, 0.00000631, -0.00001232, 0.00000022, -0.00000151, 0.00001095, -0.00000663, -0.00000229, 0.00000747, 0.00000215, -0.00001559, 0.00000250, 0.00001570, -0.00000175, -0.00000292, 0.00001426, -0.00002323, 0.00001358, -0.00001837, 0.00000904, -0.00000322, 0.00000854, -0.00002677, 0.00002103, -0.00002795, },
		{ 0.00003104, -0.00001793, 0.00001680, -0.00000727, 0.00000449, -0.00001174, 0.00001245, -0.00000505, 0.00000657, -0.00000659, 0.00000634, -0.00001058, 0.00000873, -0.00000625, 0.00000248, -0.00000541, 0.00000858, -0.00002096, 0.00000169, -0.00000146, 0.00000058, 0.00001749, 0.00000515, -0.00000003, 0.00002598, -0.00000599, 0.00000648, -0.00001037, -0.00001274, 0.00000359, -0.00000147, -0.00000768, 0.00000475, 0.00000663, -0.00000781, -0.00000874, 0.00000764, 0.00000167, 0.00000532, -0.00000375, -0.00000709, 0.00000216, -0.00000430, -0.00000081, 0.00000593, 0.00000251, 0.00000109, -0.00001400, 0.00001692, -0.00000200, 0.00000042, -0.00000736, 0.00001095, -0.00000967, 0.00000144, -0.00001673, 0.00001835, -0.00000879, 0.00000696, -0.00001609, 0.00001631, -0.00002461, 0.00001982, -0.00001629, 0.00001935, -0.00003868, 0.00003197, -0.00006843, 0.00005551, -0.00009761, 0.00007632, -0.00037061, 0.00032501, 0.00058559, -0.00041640, 0.00007713, -0.00011133, 0.00004335, -0.00005723, 0.00004113, -0.00004928, 0.00004331, -0.00004028, 0.00002445, -0.00003048, 0.00002151, -0.00002375, 0.00002096, -0.00001651, 0.00000487, -0.00000483, 0.00001517, -0.00001721, 0.00000785, -0.00000385, 0.00000924, -0.00001892, 0.00000072, 0.00000202, 0.00001669, -0.00002086, 0.00001031, -0.00001424, 0.00001644, -0.00001129, -0.00000020, -0.00000857, 0.00001103, 0.00000063, -0.00000319, -0.00000347, 0.00001442, -0.00001292, -0.00000573, 0.00000457, 0.00001831, -0.00001541, 0.00001600, -0.00000778, -0.00000728, 0.00000270, -0.00001065, 0.00000652, 0.00000276, -0.00000660, -0.00001190, 0.00001315, -0.00003045, },
		{ 0.00001587, -0.00001206, 0.00001613, -0.00001026, 0.00000360, -0.00000847, 0.00001311, -0.00000754, 0.00000772, -0.00000744, 0.00000621, -0.00001113, 0.00001240, -0.00001179, 0.00000647, -0.00000630, 0.00000737, -0.00002781, 0.00001815, -0.00001655, 0.00001857, 0.00000666, 0.00000523, 0.00000445, 0.00002463, -0.00001583, 0.00001533, -0.00002549, 0.00000321, -0.00000385, 0.00000265, -0.00001232, 0.00001654, -0.00000680, -0.00000361, -0.00000904, 0.00001402, -0.00000514, 0.00001192, -0.00001607, 0.00000440, -0.00000517, 0.00000013, -0.00000115, 0.00000816, -0.00000004, -0.00000370, -0.00000720, 0.00001961, -0.00001035, 0.00000311, -0.00000605, 0.00001140, -0.00001371, 0.00000066, -0.00001128, 0.00001928, -0.00001335, 0.00000669, -0.00001225, 0.00001111, -0.00002086, 0.00001919, -0.00001319, 0.00000711, -0.00002345, 0.00001004, -0.00004583, 0.00003267, -0.00007622, 0.00000797, -0.00039352, 0.00080436, 0.00003543, -0.00026532, 0.00006183, -0.00009131, 0.00003462, -0.00004086, 0.00002460, -0.00003233, 0.00003221, -0.00003628, 0.00002199, -0.00002792, 0.00002100, -0.00002355, 0.00002424, -0.00002439, 0.00001240, -0.00000407, 0.00001001, -0.00001776, 0.00001254, -0.00000425, 0.00000367, -0.00001879, 0.00000580, 0.00000801, 0.00000372, -0.00001463, 0.00000802, -0.00001067, 0.00001660, -0.00001749, 0.00000447, -0.00000995, 0.00001742, -0.00000764, -0.00000097, 0.00000096, 0.00001078, -0.00001976, 0.00000356, 0.00000789, 0.00001122, -0.00001536, 0.00002302, -0.00002388, 0.00000931, -0.00001191, 0.00000287, -0.00000066, 0.00000889, -0.00002299, 0.00001309, -0.00001568, -0.00000232, },
		{ -0.00001085, 0.00000998, -0.00000149, 0.00000076, -0.00000784, 0.00000650, 0.00000004, 0.00000303, -0.00000324, 0.00000385, -0.00000698, 0.00000354, -0.00000051, -0.00000188, -0.00000325, 0.00000721, -0.00001382, -0.00000783, 0.00000863, -0.00001067, 0.00002184, 0.00000176, 0.00000245, 0.00001633, 0.00000780, -0.00000713, 0.00000629, -0.00002334, 0.00000729, -0.00000304, -0.00000319, -0.00000280, 0.00001257, -0.00000991, -0.00000541, -0.00000012, 0.00000786, 0.00000116, 0.00000326, -0.00001373, 0.00000591, -0.00000593, -0.00000032, 0.00000426, 0.00000279, 0.00000325, -0.00001414, 0.00001224, 0.00000352, -0.00000245, -0.00000639, 0.00000890, -0.00000495, -0.00000207, -0.00001330, 0.00001104, -0.00000154, 0.00000271, -0.00001177, 0.00001069, -0.00001650, 0.00000913, -0.00000904, 0.00001612, -0.00003326, 0.00002593, -0.00005624, 0.00003721, -0.00007199, 0.00005106, -0.00025650, -0.00000992, 0.00085618, -0.00041487, 0.00002466, -0.00008452, 0.00002046, -0.00004442, 0.00002960, -0.00003932, 0.00003012, -0.00002477, 0.00001117, -0.00001966, 0.00001123, -0.00001425, 0.00001020, -0.00000542, -0.00000295, -0.00000168, 0.00001266, -0.00001316, 0.00000275, -0.00000038, 0.00000740, -0.00001455, -0.00000298, 0.00000041, 0.00001814, -0.00001714, 0.00000553, -0.00000908, 0.00000865, -0.00000173, -0.00000597, -0.00000549, 0.00000352, 0.00000906, -0.00000602, -0.00000393, 0.00001154, -0.00000540, -0.00001131, 0.00000440, 0.00001510, -0.00000512, 0.00000093, 0.00000966, -0.00002034, 0.00001075, -0.00001454, 0.00000770, -0.00000006, 0.00000137, -0.00001969, 0.00001903, -0.00003213, 0.00002581, },
		{ -0.00001712, 0.00001923, -0.00001226, 0.00000679, -0.00001222, 0.00001462, -0.00000860, 0.00000984, -0.00000992, 0.00001031, -0.00001530, 0.00001523, -0.00001307, 0.00000806, -0.00001066, 0.00001589, -0.00003307, 0.00002001, -0.00001599, 0.00001307, 0.00000777, 0.00000666, -0.00000192, 0.00002738, -0.00001331, 0.00001075, -0.00001401, -0.00000754, 0.00000111, 0.00000265, -0.00001225, 0.00001313, -0.00000273, -0.00000363, -0.00001056, 0.00001149, -0.00000270, 0.00001211, -0.00001342, -0.00000013, -0.00000196, -0.00000092, -0.00000255, 0.00000922, -0.00000239, 0.00000307, -0.00001589, 0.00002506, -0.00001320, 0.00000732, -0.00001351, 0.00002020, -0.00002044, 0.00000871, -0.00002165, 0.00002711, -0.00002008, 0.00001659, -0.00002440, 0.00002478, -0.00003427, 0.00003002, -0.00002719, 0.00002913, -0.00004992, 0.00004462, -0.00008545, 0.00007926, -0.00013324, 0.00011516, -0.00048826, 0.00061383, 0.00037902, -0.00043626, 0.00013327, -0.00014576, 0.00007330, -0.00007757, 0.00005722, -0.00006361, 0.00005804, -0.00005514, 0.00003799, -0.00004326, 0.00003482, -0.00003662, 0.00003428, -0.00003047, 0.00001756, -0.00001243, 0.00002026, -0.00002721, 0.00001959, -0.00001121, 0.00001269, -0.00002535, 0.00001037, -0.00000133, 0.00001649, -0.00002446, 0.00001641, -0.00001881, 0.00002199, -0.00001848, 0.00000660, -0.00001568, 0.00001914, -0.00000496, -0.00000055, -0.00000482, 0.00001659, -0.00002049, 0.00000295, 0.00000189, 0.00001868, -0.00001818, 0.00001949, -0.00001294, -0.00000311, -0.00000104, -0.00000503, 0.00000324, 0.00000632, -0.00001534, 0.00000226, -0.00000234, -0.00001611, 0.00002444, },
		{ 0.00000096, 0.00000613, -0.00000503, -0.00000168, -0.00000099, 0.00000594, -0.00000166, 0.00000177, -0.00000112, -0.00000016, -0.00000469, 0.00000756, -0.00000819, 0.00000182, 0.00000052, -0.00000043, -0.00002245, 0.00002217, -0.00002296, 0.00002699, -0.00000378, 0.00000887, 0.00000290, 0.00002260, -0.00001901, 0.00001701, -0.00002573, 0.00000495, -0.00000230, 0.00000098, -0.00000940, 0.00001693, -0.00001154, -0.00000127, -0.00000761, 0.00001242, -0.00000233, 0.00001012, -0.00001893, 0.00000767, -0.00000646, 0.00000044, 0.00000152, 0.00000496, 0.00000149, -0.00000851, 0.00000182, 0.00001375, -0.00000981, 0.00000028, -0.00000060, 0.00000766, -0.00001337, -0.00000140, -0.00000451, 0.00001330, -0.00001014, 0.00000291, -0.00000654, 0.00000375, -0.00001288, 0.00001002, -0.00000292, -0.00000734, -0.00000543, -0.00001360, -0.00001661, -0.00000307, -0.00003674, -0.00006801, -0.00034040, 0.00099291, -0.00026055, -0.00011877, -0.00000276, -0.00004335, 0.00000177, -0.00001164, -0.00000245, -0.00000707, 0.00001017, -0.00001766, 0.00000502, -0.00001197, 0.00000729, -0.00001119, 0.00001399, -0.00001751, 0.00000855, 0.00000208, -0.00000059, -0.00001054, 0.00001056, -0.00000183, -0.00000321, -0.00001285, 0.00000601, 0.00001046, -0.00000434, -0.00000726, 0.00000344, -0.00000512, 0.00001138, -0.00001427, 0.00000166, -0.00000810, 0.00001881, -0.00000927, -0.00000161, 0.00000430, 0.00000523, -0.00001871, 0.00000773, 0.00000647, 0.00000905, -0.00001397, 0.00002171, -0.00002450, 0.00000956, -0.00001095, 0.00000399, 0.00000050, 0.00000501, -0.00002119, 0.00001763, -0.00002626, 0.00001473, -0.00000034, },
		{ 0.00001773, -0.00001025, 0.00000600, -0.00001159, 0.00001216, -0.00000667, 0.00000927, -0.00000971, 0.00001114, -0.00001513, 0.00001321, -0.00000995, 0.00000615, -0.00001155, 0.00001850, -0.00002973, 0.00001225, -0.00000578, 0.00000000, 0.00001507, 0.00000150, 0.00000070, 0.00001891, 0.00000012, -0.00000357, 0.00000296, -0.00001875, 0.00000500, 0.00000192, -0.00000918, 0.00000662, 0.00000348, -0.00000620, -0.00000784, 0.00000490, 0.00000215, 0.00000951, -0.00000712, -0.00000713, 0.00000209, -0.00000221, -0.00000336, 0.00000955, -0.00000414, 0.00000796, -0.00002034, 0.00002497, -0.00001090, 0.00000710, -0.00001612, 0.00002229, -0.00001926, 0.00000850, -0.00002265, 0.00002506, -0.00001761, 0.00001699, -0.00002557, 0.00002566, -0.00003337, 0.00002707, -0.00002844, 0.00003627, -0.00005738, 0.00005453, -0.00009285, 0.00008472, -0.00013599, 0.00013008, -0.00041871, 0.00026097, 0.00075905, -0.00055163, 0.00014989, -0.00016080, 0.00008118, -0.00008895, 0.00006683, -0.00007276, 0.00006200, -0.00005445, 0.00003772, -0.00004427, 0.00003530, -0.00003682, 0.00003211, -0.00002638, 0.00001514, -0.00001440, 0.00002393, -0.00002950, 0.00001932, -0.00001149, 0.00001517, -0.00002566, 0.00001001, -0.00000575, 0.00002227, -0.00002596, 0.00001621, -0.00001782, 0.00001827, -0.00001226, 0.00000342, -0.00001502, 0.00001332, 0.00000290, -0.00000277, -0.00000698, 0.00001656, -0.00001537, -0.00000204, 0.00000143, 0.00001707, -0.00000996, 0.00000576, 0.00000418, -0.00001669, 0.00000758, -0.00000956, 0.00000480, 0.00000424, -0.00000798, -0.00000668, 0.00000726, -0.00002392, 0.00002639, -0.00001665, },
		{ 0.00001276, -0.00001012, 0.00000381, -0.00000719, 0.00001038, -0.00000569, 0.00000681, -0.00000677, 0.00000754, -0.00001306, 0.00001494, -0.00001382, 0.00000746, -0.00000840, 0.00001436, -0.00003582, 0.00003209, -0.00002954, 0.00002688, -0.00000554, 0.00001238, -0.00000605, 0.00002953, -0.00002068, 0.00001546, -0.00001724, -0.00000350, 0.00000061, 0.00000379, -0.00001325, 0.00001803, -0.00001100, 0.00000124, -0.00001198, 0.00001322, -0.00000365, 0.00001497, -0.00002085, 0.00000648, -0.00000576, 0.00000159, -0.00000244, 0.00000919, -0.00000392, 0.00000148, -0.00001180, 0.00002574, -0.00001886, 0.00000988, -0.00001406, 0.00002355, -0.00002710, 0.00001299, -0.00002189, 0.00002942, -0.00002529, 0.00002099, -0.00002734, 0.00002700, -0.00003679, 0.00003200, -0.00002857, 0.00002776, -0.00004744, 0.00004020, -0.00008085, 0.00007591, -0.00013638, 0.00010957, -0.00055626, 0.00091623, 0.00006320, -0.00033530, 0.00011296, -0.00013063, 0.00006661, -0.00006936, 0.00004852, -0.00005539, 0.00005217, -0.00005132, 0.00003449, -0.00003996, 0.00003316, -0.00003582, 0.00003514, -0.00003361, 0.00002182, -0.00001292, 0.00001675, -0.00002796, 0.00002477, -0.00001438, 0.00001125, -0.00002550, 0.00001559, -0.00000232, 0.00001240, -0.00002232, 0.00001717, -0.00001854, 0.00002240, -0.00002072, 0.00000877, -0.00001892, 0.00002525, -0.00000982, 0.00000098, -0.00000390, 0.00001492, -0.00002413, 0.00001052, -0.00000179, 0.00001948, -0.00002069, 0.00002216, -0.00001695, -0.00000042, -0.00000217, -0.00000212, 0.00000270, 0.00000560, -0.00001826, 0.00001187, -0.00001683, 0.00000236, 0.00000946, -0.00000700, },
		{ -0.00000485, 0.00000177, -0.00000739, 0.00000673, -0.00000246, 0.00000614, -0.00000638, 0.00000788, -0.00000986, 0.00000601, -0.00000228, 0.00000017, -0.00000742, 0.00001319, -0.00001672, -0.00000461, 0.00001198, -0.00001781, 0.00002596, -0.00000798, 0.00000945, 0.00000390, 0.00001771, -0.00001895, 0.00001637, -0.00002371, 0.00000429, 0.00000210, -0.00000460, -0.00000118, 0.00001030, -0.00001019, -0.00000243, -0.00000345, 0.00000744, 0.00000529, 0.00000150, -0.00001497, 0.00000595, -0.00000384, -0.00000253, 0.00000731, -0.00000244, 0.00000729, -0.00001689, 0.00001629, -0.00000069, -0.00000067, -0.00001000, 0.00001353, -0.00000642, -0.00000336, -0.00001119, 0.00001036, -0.00000285, 0.00000407, -0.00001146, 0.00000953, -0.00001437, 0.00000638, -0.00001119, 0.00002034, -0.00003596, 0.00002885, -0.00005681, 0.00003772, -0.00007280, 0.00004829, -0.00021832, -0.00015862, 0.00105478, -0.00052272, 0.00006204, -0.00010556, 0.00003520, -0.00005607, 0.00003730, -0.00004631, 0.00003363, -0.00002647, 0.00001469, -0.00002479, 0.00001662, -0.00001901, 0.00001354, -0.00000861, 0.00000140, -0.00000548, 0.00001605, -0.00001960, 0.00000715, -0.00000097, 0.00000727, -0.00001601, 0.00000068, -0.00000067, 0.00001676, -0.00001581, 0.00000510, -0.00000645, 0.00000543, 0.00000066, -0.00000517, -0.00000754, 0.00000062, 0.00001448, -0.00000715, -0.00000494, 0.00001039, -0.00000485, -0.00001105, 0.00000632, 0.00000803, 0.00000477, -0.00001075, 0.00001862, -0.00002344, 0.00000840, -0.00000757, 0.00000224, 0.00000447, -0.00000329, -0.00001129, 0.00001125, -0.00002504, 0.00002229, -0.00001171, 0.00001251, },
		{ -0.00001219, 0.00000626, -0.00000997, 0.00001172, -0.00000755, 0.00001006, -0.00001073, 0.00001301, -0.00001799, 0.00001814, -0.00001554, 0.00001023, -0.00001479, 0.00002416, -0.00004083, 0.00003167, -0.00002484, 0.00001705, 0.00000077, 0.00000805, -0.00000515, 0.00002404, -0.00000959, 0.00000223, -0.00000145, -0.00001394, 0.00000326, 0.00000561, -0.00001381, 0.00001536, -0.00000748, 0.00000060, -0.00001170, 0.00000960, -0.00000231, 0.00001667, -0.00001842, 0.00000202, -0.00000375, 0.00000258, -0.00000633, 0.00001348, -0.00000984, 0.00001116, -0.00002292, 0.00003359, -0.00002378, 0.00001620, -0.00002415, 0.00003421, -0.00003421, 0.00002049, -0.00003149, 0.00003725, -0.00003280, 0.00003141, -0.00003934, 0.00003987, -0.00004892, 0.00004297, -0.00004495, 0.00005244, -0.00007592, 0.00007567, -0.00011986, 0.00012096, -0.00018897, 0.00019708, -0.00058470, 0.00061362, 0.00050286, -0.00055493, 0.00020722, -0.00019938, 0.00011595, -0.00011408, 0.00008704, -0.00009065, 0.00007990, -0.00007178, 0.00005323, -0.00005858, 0.00005011, -0.00005138, 0.00004719, -0.00004205, 0.00002984, -0.00002416, 0.00003039, -0.00004077, 0.00003372, -0.00002219, 0.00002119, -0.00003372, 0.00002186, -0.00001213, 0.00002445, -0.00003115, 0.00002401, -0.00002421, 0.00002530, -0.00002056, 0.00001105, -0.00002370, 0.00002405, -0.00000496, 0.00000112, -0.00000923, 0.00001950, -0.00002387, 0.00000879, -0.00000459, 0.00002090, -0.00001582, 0.00001102, -0.00000137, -0.00001339, 0.00000588, -0.00000593, 0.00000323, 0.00000606, -0.00001439, 0.00000592, -0.00000836, -0.00000705, 0.00001532, -0.00001147, 0.00001621, },
		{ -0.00000361, -0.00000209, 0.00000074, 0.00000217, 0.00000177, -0.00000102, 0.00000167, -0.00000116, -0.00000408, 0.00000746, -0.00000778, 0.00000008, 0.00000225, 0.00000186, -0.00002461, 0.00003015, -0.00003309, 0.00003374, -0.00001533, 0.00001712, -0.00000891, 0.00002968, -0.00002572, 0.00001863, -0.00001855, -0.00000196, 0.00000344, 0.00000030, -0.00000829, 0.00001582, -0.00001372, 0.00000309, -0.00001057, 0.00001098, 0.00000100, 0.00001054, -0.00002164, 0.00000853, -0.00000581, 0.00000052, 0.00000163, 0.00000384, 0.00000009, -0.00000550, 0.00000008, 0.00001585, -0.00001454, 0.00000360, -0.00000431, 0.00001521, -0.00002322, 0.00000850, -0.00001239, 0.00001976, -0.00001796, 0.00001327, -0.00001808, 0.00001626, -0.00002547, 0.00001893, -0.00001332, 0.00000770, -0.00002311, 0.00000944, -0.00004319, 0.00002990, -0.00008493, 0.00002460, -0.00051315, 0.00114755, -0.00029163, -0.00014346, 0.00001977, -0.00006210, 0.00001804, -0.00002784, 0.00001122, -0.00002127, 0.00002246, -0.00002557, 0.00001073, -0.00001726, 0.00001286, -0.00001728, 0.00001907, -0.00002102, 0.00001276, -0.00000231, 0.00000129, -0.00001521, 0.00001827, -0.00000896, 0.00000168, -0.00001608, 0.00001255, 0.00000201, 0.00000279, -0.00001292, 0.00001063, -0.00001140, 0.00001555, -0.00001555, 0.00000332, -0.00001457, 0.00002546, -0.00001089, -0.00000065, 0.00000077, 0.00000758, -0.00002069, 0.00001285, -0.00000294, 0.00001831, -0.00002136, 0.00002296, -0.00001915, 0.00000051, -0.00000033, -0.00000267, 0.00000584, -0.00000064, -0.00001248, 0.00001135, -0.00002211, 0.00001545, -0.00000541, 0.00000609, 0.00000172, },
		{ 0.00000515, -0.00000909, 0.00000993, -0.00000693, 0.00001057, -0.00001141, 0.00001407, -0.00001723, 0.00001523, -0.00001162, 0.00000793, -0.00001526, 0.00002476, -0.00003384, 0.00001877, -0.00000850, -0.00000176, 0.00001429, -0.00000354, 0.00000462, 0.00000891, 0.00000911, -0.00001519, 0.00001342, -0.00002010, 0.00000257, 0.00000729, -0.00001147, 0.00000955, -0.00000134, -0.00000327, -0.00000706, 0.00000258, 0.00000105, 0.00001502, -0.00001154, -0.00000536, -0.00000035, 0.00000217, -0.00000783, 0.00001451, -0.00001172, 0.00001483, -0.00002544, 0.00003148, -0.00001895, 0.00001320, -0.00002386, 0.00003202, -0.00002720, 0.00001380, -0.00002614, 0.00002969, -0.00002503, 0.00002549, -0.00003302, 0.00003244, -0.00003931, 0.00003237, -0.00003906, 0.00005028, -0.00007120, 0.00007013, -0.00010802, 0.00010289, -0.00015976, 0.00015964, -0.00042506, 0.00015700, 0.00093906, -0.00067304, 0.00021028, -0.00020114, 0.00011213, -0.00011438, 0.00008596, -0.00008942, 0.00007385, -0.00006315, 0.00004740, -0.00005496, 0.00004609, -0.00004695, 0.00004064, -0.00003454, 0.00002464, -0.00002326, 0.00003154, -0.00004016, 0.00002918, -0.00001772, 0.00001954, -0.00003071, 0.00001789, -0.00001155, 0.00002453, -0.00002737, 0.00001860, -0.00001794, 0.00001729, -0.00001178, 0.00000610, -0.00001979, 0.00001387, 0.00000558, -0.00000217, -0.00000956, 0.00001694, -0.00001661, 0.00000121, -0.00000040, 0.00001311, -0.00000206, -0.00000562, 0.00001409, -0.00002152, 0.00000738, -0.00000400, 0.00000010, 0.00000819, -0.00001187, 0.00000228, -0.00000307, -0.00001204, 0.00001679, -0.00001239, 0.00001620, -0.00001077, },
		{ 0.00000176, -0.00000348, 0.00000538, -0.00000217, 0.00000447, -0.00000501, 0.00000766, -0.00001315, 0.00001528, -0.00001398, 0.00000707, -0.00000910, 0.00001881, -0.00003928, 0.00004014, -0.00003812, 0.00003179, -0.00001432, 0.00001690, -0.00001210, 0.00002921, -0.00001901, 0.00000822, -0.00000511, -0.00001079, 0.00000430, 0.00000518, -0.00001332, 0.00001854, -0.00001499, 0.00000624, -0.00001397, 0.00001118, -0.00000202, 0.00001792, -0.00002508, 0.00000867, -0.00000747, 0.00000490, -0.00000604, 0.00001266, -0.00001056, 0.00000919, -0.00001778, 0.00003238, -0.00002817, 0.00001799, -0.00002336, 0.00003643, -0.00004113, 0.00002607, -0.00003253, 0.00003952, -0.00003795, 0.00003640, -0.00004355, 0.00004373, -0.00005323, 0.00004641, -0.00004733, 0.00005238, -0.00007514, 0.00007342, -0.00011781, 0.00012092, -0.00019803, 0.00020607, -0.00068774, 0.00097398, 0.00012950, -0.00042029, 0.00017120, -0.00017482, 0.00010312, -0.00010184, 0.00007524, -0.00007978, 0.00007168, -0.00006541, 0.00004700, -0.00005232, 0.00004551, -0.00004790, 0.00004551, -0.00004264, 0.00003200, -0.00002312, 0.00002484, -0.00003901, 0.00003760, -0.00002534, 0.00001967, -0.00003293, 0.00002657, -0.00001391, 0.00002110, -0.00002902, 0.00002476, -0.00002409, 0.00002551, -0.00002209, 0.00001196, -0.00002609, 0.00003055, -0.00001047, 0.00000268, -0.00000803, 0.00001715, -0.00002643, 0.00001619, -0.00000975, 0.00002403, -0.00002088, 0.00001534, -0.00000584, -0.00001146, 0.00000696, -0.00000560, 0.00000529, 0.00000247, -0.00001308, 0.00001091, -0.00001888, 0.00000954, -0.00000066, 0.00000079, 0.00000714, -0.00000769, },
		{ -0.00000621, 0.00000661, -0.00000484, 0.00000875, -0.00000878, 0.00001060, -0.00001107, 0.00000711, -0.00000319, 0.00000125, -0.00000983, 0.00001649, -0.00001693, -0.00000246, 0.00001371, -0.00002308, 0.00002880, -0.00001639, 0.00001649, -0.00000745, 0.00002559, -0.00002669, 0.00001912, -0.00001753, -0.00000266, 0.00000884, -0.00000683, 0.00000186, 0.00000616, -0.00000916, 0.00000008, -0.00000558, 0.00000505, 0.00001058, -0.00000120, -0.00001447, 0.00000417, -0.00000064, -0.00000503, 0.00000977, -0.00000645, 0.00000913, -0.00001660, 0.00001758, -0.00000303, -0.00000069, -0.00001117, 0.00001523, -0.00000527, -0.00000716, -0.00000618, 0.00000728, -0.00000217, 0.00000294, -0.00000809, 0.00000489, -0.00000871, 0.00000066, -0.00000973, 0.00001870, -0.00003067, 0.00002198, -0.00004589, 0.00002545, -0.00005777, 0.00002314, -0.00015151, -0.00031713, 0.00122489, -0.00059274, 0.00007425, -0.00010965, 0.00003716, -0.00005651, 0.00003451, -0.00004372, 0.00002883, -0.00002217, 0.00001405, -0.00002587, 0.00001799, -0.00001999, 0.00001373, -0.00000960, 0.00000408, -0.00000724, 0.00001676, -0.00002292, 0.00000870, 0.00000080, 0.00000472, -0.00001491, 0.00000190, 0.00000127, 0.00001150, -0.00001099, 0.00000192, -0.00000162, 0.00000102, 0.00000306, -0.00000441, -0.00000880, -0.00000253, 0.00001874, -0.00000724, -0.00000552, 0.00000843, -0.00000439, -0.00001005, 0.00000823, 0.00000056, 0.00001355, -0.00001918, 0.00002148, -0.00001967, 0.00000054, 0.00000312, -0.00000525, 0.00001089, -0.00001011, 0.00000028, 0.00000018, -0.00001440, 0.00001618, -0.00001183, 0.00001427, -0.00000721, 0.00000194, },
		{ -0.00000506, 0.00000652, -0.00000473, 0.00000846, -0.00000975, 0.00001343, -0.00001769, 0.00001790, -0.00001528, 0.00000990, -0.00001580, 0.00002774, -0.00004229, 0.00003673, -0.00002936, 0.00001797, -0.00000324, 0.00000696, -0.00000419, 0.00001654, -0.00000172, -0.00000914, 0.00000961, -0.00001678, 0.00000227, 0.00000998, -0.00001541, 0.00001764, -0.00001290, 0.00000546, -0.00001246, 0.00000763, -0.00000294, 0.00002182, -0.00002354, 0.00000521, -0.00000748, 0.00000824, -0.00001218, 0.00001938, -0.00001860, 0.00001969, -0.00002901, 0.00004055, -0.00003325, 0.00002432, -0.00003356, 0.00004618, -0.00004587, 0.00003032, -0.00003881, 0.00004503, -0.00004379, 0.00004447, -0.00005216, 0.00005244, -0.00006086, 0.00005422, -0.00006168, 0.00007365, -0.00009804, 0.00010079, -0.00014591, 0.00015224, -0.00023048, 0.00025411, -0.00063410, 0.00056194, 0.00064504, -0.00066631, 0.00027144, -0.00024521, 0.00015271, -0.00014527, 0.00011119, -0.00011150, 0.00009516, -0.00008292, 0.00006474, -0.00007063, 0.00006212, -0.00006279, 0.00005692, -0.00005131, 0.00004082, -0.00003490, 0.00003923, -0.00005234, 0.00004570, -0.00003146, 0.00002813, -0.00004041, 0.00003192, -0.00002115, 0.00002937, -0.00003421, 0.00002791, -0.00002577, 0.00002515, -0.00002038, 0.00001373, -0.00002911, 0.00002631, -0.00000389, 0.00000263, -0.00001245, 0.00002017, -0.00002516, 0.00001321, -0.00000900, 0.00001966, -0.00000997, 0.00000015, 0.00000936, -0.00002005, 0.00000839, -0.00000284, 0.00000069, 0.00000775, -0.00001505, 0.00001203, -0.00001726, 0.00000595, 0.00000189, -0.00000291, 0.00001060, -0.00000936, 0.00000302, },
		{ 0.00000372, -0.00000272, 0.00000574, -0.00000413, 0.00000439, -0.00000213, -0.00000288, 0.00000628, -0.00000657, -0.00000176, 0.00000385, 0.00000323, -0.00002359, 0.00003257, -0.00003804, 0.00003630, -0.00002288, 0.00002233, -0.00001609, 0.00003118, -0.00002543, 0.00001235, -0.00000667, -0.00001026, 0.00000886, -0.00000026, -0.00000621, 0.00001334, -0.00001506, 0.00000703, -0.00001221, 0.00000803, 0.00000465, 0.00001060, -0.00002318, 0.00000866, -0.00000555, 0.00000189, -0.00000027, 0.00000514, -0.00000427, 0.00000055, -0.00000380, 0.00001909, -0.00002050, 0.00000873, -0.00000989, 0.00002386, -0.00003382, 0.00001942, -0.00002076, 0.00002658, -0.00002675, 0.00002499, -0.00003106, 0.00003006, -0.00003907, 0.00003008, -0.00002802, 0.00002809, -0.00004651, 0.00003818, -0.00007500, 0.00006862, -0.00013987, 0.00011900, -0.00065939, 0.00124485, -0.00027486, -0.00019191, 0.00005552, -0.00009015, 0.00004234, -0.00005115, 0.00003068, -0.00003961, 0.00003676, -0.00003460, 0.00001827, -0.00002457, 0.00002019, -0.00002454, 0.00002468, -0.00002513, 0.00001834, -0.00000873, 0.00000539, -0.00002144, 0.00002692, -0.00001715, 0.00000787, -0.00002055, 0.00002073, -0.00000832, 0.00001085, -0.00001856, 0.00001716, -0.00001619, 0.00001762, -0.00001535, 0.00000412, -0.00001926, 0.00002930, -0.00001071, 0.00000005, -0.00000229, 0.00000837, -0.00002058, 0.00001637, -0.00001039, 0.00002366, -0.00002317, 0.00001741, -0.00000852, -0.00001096, 0.00001063, -0.00000852, 0.00001094, -0.00000649, -0.00000281, 0.00000438, -0.00001730, 0.00001649, -0.00001207, 0.00001220, -0.00000338, -0.00000125, -0.00000356, },
		{ 0.00000643, -0.00000611, 0.00001041, -0.00001141, 0.00001464, -0.00001643, 0.00001452, -0.00001108, 0.00000745, -0.00001574, 0.00002642, -0.00003275, 0.00002106, -0.00000988, -0.00000353, 0.00001354, -0.00000836, 0.00000963, -0.00000125, 0.00001722, -0.00002372, 0.00001745, -0.00001539, -0.00000363, 0.00001397, -0.00001398, 0.00001304, -0.00000727, 0.00000052, -0.00000644, 0.00000114, -0.00000162, 0.00002110, -0.00001619, -0.00000231, -0.00000453, 0.00000784, -0.00001290, 0.00001920, -0.00001832, 0.00001961, -0.00002735, 0.00003469, -0.00002440, 0.00001686, -0.00002843, 0.00003785, -0.00003153, 0.00001590, -0.00002607, 0.00003107, -0.00002958, 0.00003072, -0.00003668, 0.00003509, -0.00004107, 0.00003414, -0.00004559, 0.00005795, -0.00007635, 0.00007483, -0.00010997, 0.00010568, -0.00016403, 0.00016091, -0.00039210, 0.00002897, 0.00110064, -0.00075868, 0.00024361, -0.00022222, 0.00012833, -0.00012698, 0.00009314, -0.00009495, 0.00007567, -0.00006412, 0.00005134, -0.00006015, 0.00005139, -0.00005180, 0.00004450, -0.00003911, 0.00003128, -0.00002918, 0.00003565, -0.00004667, 0.00003506, -0.00002077, 0.00002097, -0.00003266, 0.00002275, -0.00001392, 0.00002251, -0.00002460, 0.00001731, -0.00001482, 0.00001392, -0.00001009, 0.00000761, -0.00002261, 0.00001311, 0.00000778, -0.00000092, -0.00001128, 0.00001578, -0.00001687, 0.00000415, -0.00000143, 0.00000796, 0.00000567, -0.00001464, 0.00001844, -0.00001961, 0.00000152, 0.00000552, -0.00000671, 0.00001394, -0.00001746, 0.00001368, -0.00001589, 0.00000290, 0.00000428, -0.00000659, 0.00001299, -0.00000929, 0.00000247, -0.00000496, },
		{ -0.00000126, 0.00000251, 0.00000092, -0.00000211, 0.00000617, -0.00001081, 0.00001287, -0.00001177, 0.00000471, -0.00000732, 0.00001905, -0.00003639, 0.00004040, -0.00004026, 0.00003178, -0.00001860, 0.00001766, -0.00001287, 0.00002342, -0.00001185, -0.00000315, 0.00000685, -0.00001554, 0.00000536, 0.00000768, -0.00001303, 0.00001837, -0.00001876, 0.00001110, -0.00001502, 0.00000862, -0.00000105, 0.00002098, -0.00002863, 0.00001100, -0.00001050, 0.00001004, -0.00001167, 0.00001800, -0.00001891, 0.00001794, -0.00002351, 0.00003786, -0.00003657, 0.00002558, -0.00003170, 0.00004723, -0.00005276, 0.00003708, -0.00004069, 0.00004722, -0.00004873, 0.00004991, -0.00005759, 0.00005798, -0.00006709, 0.00005945, -0.00006573, 0.00007595, -0.00010041, 0.00010244, -0.00014832, 0.00015746, -0.00024751, 0.00027899, -0.00076842, 0.00096787, 0.00022976, -0.00050933, 0.00022637, -0.00021558, 0.00013701, -0.00013181, 0.00009884, -0.00010021, 0.00008638, -0.00007537, 0.00005699, -0.00006253, 0.00005556, -0.00005736, 0.00005318, -0.00004964, 0.00004111, -0.00003271, 0.00003218, -0.00004840, 0.00004823, -0.00003455, 0.00002674, -0.00003895, 0.00003632, -0.00002407, 0.00002731, -0.00003257, 0.00002895, -0.00002595, 0.00002506, -0.00002093, 0.00001304, -0.00003010, 0.00003241, -0.00000942, 0.00000392, -0.00001087, 0.00001697, -0.00002606, 0.00001959, -0.00001482, 0.00002422, -0.00001640, 0.00000469, 0.00000579, -0.00001998, 0.00001241, -0.00000551, 0.00000568, 0.00000100, -0.00000897, 0.00001141, -0.00002229, 0.00001871, -0.00001340, 0.00001024, -0.00000005, -0.00000328, -0.00000234, 0.00000223, },
		{ -0.00000713, 0.00001069, -0.00001024, 0.00001186, -0.00001069, 0.00000718, -0.00000360, 0.00000159, -0.00001056, 0.00001743, -0.00001540, -0.00000007, 0.00001220, -0.00002426, 0.00002801, -0.00002122, 0.00002089, -0.00001467, 0.00002817, -0.00002749, 0.00001389, -0.00000625, -0.00001143, 0.00001530, -0.00000858, 0.00000531, 0.00000105, -0.00000739, 0.00000219, -0.00000645, 0.00000131, 0.00001581, -0.00000380, -0.00001295, 0.00000146, 0.00000248, -0.00000650, 0.00001033, -0.00000792, 0.00000765, -0.00001258, 0.00001567, -0.00000305, -0.00000298, -0.00000935, 0.00001366, -0.00000170, -0.00001273, 0.00000086, 0.00000250, 0.00000008, -0.00000031, -0.00000223, -0.00000233, -0.00000059, -0.00000727, -0.00000473, 0.00001116, -0.00001793, 0.00000663, -0.00002566, 0.00000332, -0.00003112, -0.00001674, -0.00007184, -0.00045969, 0.00133989, -0.00060826, 0.00005398, -0.00009342, 0.00002426, -0.00004438, 0.00002064, -0.00003162, 0.00001635, -0.00001263, 0.00000945, -0.00002254, 0.00001486, -0.00001680, 0.00001048, -0.00000803, 0.00000444, -0.00000617, 0.00001401, -0.00002237, 0.00000705, 0.00000474, 0.00000016, -0.00001153, 0.00000086, 0.00000574, 0.00000344, -0.00000406, -0.00000256, 0.00000386, -0.00000322, 0.00000455, -0.00000321, -0.00000987, -0.00000475, 0.00002061, -0.00000574, -0.00000601, 0.00000624, -0.00000457, -0.00000773, 0.00000899, -0.00000532, 0.00001875, -0.00002202, 0.00001706, -0.00000983, -0.00001062, 0.00001487, -0.00001233, 0.00001721, -0.00001694, 0.00001246, -0.00001131, -0.00000296, 0.00000943, -0.00001223, 0.00001568, -0.00000890, 0.00000177, -0.00000495, 0.00000657, },
		{ -0.00000083, 0.00000540, -0.00000706, 0.00001140, -0.00001422, 0.00001447, -0.00001231, 0.00000693, -0.00001313, 0.00002588, -0.00003703, 0.00003499, -0.00002941, 0.00001629, -0.00000513, 0.00000431, -0.00000056, 0.00000720, 0.00000681, -0.00001871, 0.00001536, -0.00001420, -0.00000228, 0.00001546, -0.00001673, 0.00001962, -0.00001843, 0.00001036, -0.00001276, 0.00000604, -0.00000472, 0.00002694, -0.00002798, 0.00000883, -0.00001246, 0.00001501, -0.00001866, 0.00002521, -0.00002661, 0.00002638, -0.00003206, 0.00004396, -0.00003962, 0.00002954, -0.00003926, 0.00005333, -0.00005279, 0.00003599, -0.00004172, 0.00004861, -0.00005085, 0.00005321, -0.00006020, 0.00005991, -0.00006771, 0.00006116, -0.00007374, 0.00008803, -0.00011096, 0.00011425, -0.00015750, 0.00016613, -0.00024937, 0.00027722, -0.00063240, 0.00046785, 0.00078550, -0.00075047, 0.00031139, -0.00027264, 0.00017499, -0.00016381, 0.00012353, -0.00012115, 0.00010006, -0.00008582, 0.00007005, -0.00007677, 0.00006817, -0.00006828, 0.00006119, -0.00005613, 0.00004821, -0.00004217, 0.00004418, -0.00005909, 0.00005275, -0.00003670, 0.00003159, -0.00004354, 0.00003844, -0.00002635, 0.00002975, -0.00003268, 0.00002753, -0.00002337, 0.00002172, -0.00001804, 0.00001430, -0.00003143, 0.00002600, -0.00000236, 0.00000422, -0.00001437, 0.00001865, -0.00002440, 0.00001606, -0.00001143, 0.00001600, -0.00000291, -0.00000966, 0.00001544, -0.00002043, 0.00000529, 0.00000438, -0.00000407, 0.00001123, -0.00001704, 0.00001977, -0.00002740, 0.00002048, -0.00001335, 0.00000660, 0.00000368, -0.00000465, -0.00000202, 0.00000142, 0.00000010, },
		{ 0.00000974, -0.00000725, 0.00000693, -0.00000331, -0.00000066, 0.00000364, -0.00000427, -0.00000404, 0.00000610, 0.00000238, -0.00001810, 0.00002815, -0.00003621, 0.00003322, -0.00002518, 0.00002307, -0.00001736, 0.00002634, -0.00001865, 0.00000093, 0.00000610, -0.00001676, 0.00001179, 0.00000052, -0.00000406, 0.00001023, -0.00001591, 0.00001052, -0.00001269, 0.00000450, 0.00000746, 0.00001095, -0.00002389, 0.00000853, -0.00000625, 0.00000495, -0.00000432, 0.00000872, -0.00001102, 0.00000858, -0.00000863, 0.00002234, -0.00002653, 0.00001445, -0.00001575, 0.00003147, -0.00004272, 0.00002894, -0.00002750, 0.00003183, -0.00003445, 0.00003562, -0.00004270, 0.00004223, -0.00005104, 0.00004101, -0.00004391, 0.00004953, -0.00007013, 0.00006600, -0.00010427, 0.00010344, -0.00018813, 0.00019569, -0.00075619, 0.00126921, -0.00020811, -0.00025871, 0.00009824, -0.00012206, 0.00006976, -0.00007681, 0.00005141, -0.00005793, 0.00004965, -0.00004243, 0.00002596, -0.00003230, 0.00002765, -0.00003138, 0.00002949, -0.00002881, 0.00002426, -0.00001600, 0.00001040, -0.00002754, 0.00003441, -0.00002437, 0.00001357, -0.00002454, 0.00002847, -0.00001806, 0.00001747, -0.00002211, 0.00002124, -0.00001814, 0.00001683, -0.00001317, 0.00000335, -0.00002106, 0.00002955, -0.00000861, 0.00000035, -0.00000445, 0.00000720, -0.00001799, 0.00001744, -0.00001460, 0.00002413, -0.00001939, 0.00000677, 0.00000391, -0.00002098, 0.00001856, -0.00001101, 0.00001380, -0.00001042, 0.00000552, -0.00000107, -0.00001342, 0.00001830, -0.00001988, 0.00001859, -0.00000869, 0.00000188, -0.00000604, 0.00000764, -0.00000813, },
		{ 0.00000906, -0.00000986, 0.00001299, -0.00001308, 0.00001163, -0.00000888, 0.00000512, -0.00001322, 0.00002390, -0.00002745, 0.00002032, -0.00001111, -0.00000369, 0.00001102, -0.00001107, 0.00001326, -0.00000854, 0.00002095, -0.00002588, 0.00001378, -0.00000547, -0.00001189, 0.00002031, -0.00001549, 0.00001606, -0.00001323, 0.00000433, -0.00000572, 0.00000053, -0.00000521, 0.00002645, -0.00001979, 0.00000097, -0.00000925, 0.00001324, -0.00001683, 0.00002186, -0.00002211, 0.00002067, -0.00002488, 0.00003360, -0.00002622, 0.00001697, -0.00002858, 0.00003856, -0.00003145, 0.00001447, -0.00002235, 0.00002899, -0.00003069, 0.00003192, -0.00003594, 0.00003320, -0.00003841, 0.00003179, -0.00004665, 0.00005756, -0.00007143, 0.00006776, -0.00009841, 0.00009318, -0.00014954, 0.00013733, -0.00033111, -0.00010061, 0.00121767, -0.00078982, 0.00023955, -0.00021786, 0.00012524, -0.00012330, 0.00008599, -0.00008788, 0.00006684, -0.00005708, 0.00004888, -0.00005869, 0.00004998, -0.00005024, 0.00004269, -0.00003902, 0.00003374, -0.00003067, 0.00003480, -0.00004750, 0.00003573, -0.00001992, 0.00001904, -0.00003106, 0.00002404, -0.00001260, 0.00001665, -0.00001852, 0.00001340, -0.00000970, 0.00000933, -0.00000795, 0.00000826, -0.00002380, 0.00001188, 0.00000857, 0.00000132, -0.00001232, 0.00001351, -0.00001653, 0.00000710, -0.00000246, 0.00000336, 0.00001076, -0.00001863, 0.00001550, -0.00001120, -0.00000832, 0.00001675, -0.00001347, 0.00001968, -0.00002273, 0.00002496, -0.00002805, 0.00001746, -0.00000825, -0.00000112, 0.00000928, -0.00000673, -0.00000102, -0.00000034, 0.00000261, -0.00000473, },
		{ -0.00000321, 0.00000159, 0.00000313, -0.00000612, 0.00000791, -0.00000745, 0.00000051, -0.00000281, 0.00001463, -0.00002714, 0.00003311, -0.00003604, 0.00002716, -0.00001892, 0.00001519, -0.00000942, 0.00001411, -0.00000245, -0.00001407, 0.00001457, -0.00001538, 0.00000281, 0.00001152, -0.00001252, 0.00001763, -0.00002206, 0.00001542, -0.00001517, 0.00000619, -0.00000120, 0.00002391, -0.00003101, 0.00001313, -0.00001436, 0.00001615, -0.00001807, 0.00002364, -0.00002702, 0.00002551, -0.00002693, 0.00004034, -0.00004216, 0.00003061, -0.00003669, 0.00005323, -0.00005927, 0.00004355, -0.00004425, 0.00005046, -0.00005526, 0.00005876, -0.00006646, 0.00006674, -0.00007553, 0.00006807, -0.00007963, 0.00009317, -0.00011713, 0.00012052, -0.00016499, 0.00017693, -0.00027404, 0.00031550, -0.00078865, 0.00090037, 0.00034958, -0.00058600, 0.00026566, -0.00024317, 0.00016019, -0.00015217, 0.00011328, -0.00011165, 0.00009248, -0.00007855, 0.00006214, -0.00006819, 0.00006089, -0.00006193, 0.00005615, -0.00005278, 0.00004715, -0.00003949, 0.00003646, -0.00005356, 0.00005401, -0.00003964, 0.00003040, -0.00004156, 0.00004256, -0.00003051, 0.00002924, -0.00003173, 0.00002893, -0.00002384, 0.00002118, -0.00001741, 0.00001180, -0.00003054, 0.00003089, -0.00000718, 0.00000490, -0.00001236, 0.00001449, -0.00002315, 0.00002056, -0.00001690, 0.00002081, -0.00000929, -0.00000631, 0.00001386, -0.00002281, 0.00001231, -0.00000096, 0.00000342, 0.00000181, -0.00000653, 0.00001354, -0.00002637, 0.00002808, -0.00002650, 0.00001939, -0.00000739, 0.00000208, -0.00000752, 0.00000874, -0.00000830, 0.00000742, },
		{ -0.00001073, 0.00001181, -0.00000921, 0.00000682, -0.00000406, 0.00000165, -0.00001008, 0.00001687, -0.00001349, 0.00000376, 0.00000652, -0.00002017, 0.00002240, -0.00002126, 0.00002109, -0.00001617, 0.00002421, -0.00002128, 0.00000285, 0.00000668, -0.00001888, 0.00001919, -0.00000840, 0.00000797, -0.00000389, -0.00000565, 0.00000416, -0.00000643, -0.00000266, 0.00001952, -0.00000520, -0.00001124, -0.00000122, 0.00000441, -0.00000596, 0.00000826, -0.00000639, 0.00000284, -0.00000534, 0.00001118, -0.00000131, -0.00000703, -0.00000510, 0.00000962, 0.00000303, -0.00001846, 0.00000823, -0.00000252, 0.00000275, -0.00000449, 0.00000460, -0.00001030, 0.00000804, -0.00001572, 0.00000242, -0.00000037, -0.00000075, -0.00001310, -0.00000131, -0.00002242, -0.00000080, -0.00005960, 0.00000207, -0.00056223, 0.00138056, -0.00056223, 0.00000207, -0.00005960, -0.00000080, -0.00002242, -0.00000131, -0.00001310, -0.00000075, -0.00000037, 0.00000242, -0.00001572, 0.00000804, -0.00001030, 0.00000460, -0.00000449, 0.00000275, -0.00000252, 0.00000823, -0.00001846, 0.00000303, 0.00000962, -0.00000510, -0.00000703, -0.00000131, 0.00001118, -0.00000534, 0.00000284, -0.00000639, 0.00000826, -0.00000596, 0.00000441, -0.00000122, -0.00001124, -0.00000520, 0.00001952, -0.00000266, -0.00000643, 0.00000416, -0.00000565, -0.00000389, 0.00000797, -0.00000840, 0.00001919, -0.00001888, 0.00000668, 0.00000285, -0.00002128, 0.00002421, -0.00001617, 0.00002109, -0.00002126, 0.00002240, -0.00002017, 0.00000652, 0.00000376, -0.00001349, 0.00001687, -0.00001008, 0.00000165, -0.00000406, 0.00000682, -0.00000921, 0.00001181, },
		{ -0.00000321, 0.00000742, -0.00000830, 0.00000874, -0.00000752, 0.00000208, -0.00000739, 0.00001939, -0.00002650, 0.00002808, -0.00002637, 0.00001354, -0.00000653, 0.00000181, 0.00000342, -0.00000096, 0.00001231, -0.00002281, 0.00001386, -0.00000631, -0.00000929, 0.00002081, -0.00001690, 0.00002056, -0.00002315, 0.00001449, -0.00001236, 0.00000490, -0.00000718, 0.00003089, -0.00003054, 0.00001180, -0.00001741, 0.00002118, -0.00002384, 0.00002893, -0.00003173, 0.00002924, -0.00003051, 0.00004256, -0.00004156, 0.00003040, -0.00003964, 0.00005401, -0.00005356, 0.00003646, -0.00003949, 0.00004715, -0.00005278, 0.00005615, -0.00006193, 0.00006089, -0.00006819, 0.00006214, -0.00007855, 0.00009248, -0.00011165, 0.00011328, -0.00015217, 0.00016019, -0.00024317, 0.00026566, -0.00058600, 0.00034958, 0.00090037, -0.00078865, 0.00031550, -0.00027404, 0.00017693, -0.00016499, 0.00012052, -0.00011713, 0.00009317, -0.00007963, 0.00006807, -0.00007553, 0.00006674, -0.00006646, 0.00005876, -0.00005526, 0.00005046, -0.00004425, 0.00004355, -0.00005927, 0.00005323, -0.00003669, 0.00003061, -0.00004216, 0.00004034, -0.00002693, 0.00002551, -0.00002702, 0.00002364, -0.00001807, 0.00001615, -0.00001436, 0.00001313, -0.00003101, 0.00002391, -0.00000120, 0.00000619, -0.00001517, 0.00001542, -0.00002206, 0.00001763, -0.00001252, 0.00001152, 0.00000281, -0.00001538, 0.00001457, -0.00001407, -0.00000245, 0.00001411, -0.00000942, 0.00001519, -0.00001892, 0.00002716, -0.00003604, 0.00003311, -0.00002714, 0.00001463, -0.00000281, 0.00000051, -0.00000745, 0.00000791, -0.00000612, 0.00000313, 0.00000159, },
		{ 0.00000906, -0.00000473, 0.00000261, -0.00000034, -0.00000102, -0.00000673, 0.00000928, -0.00000112, -0.00000825, 0.00001746, -0.00002805, 0.00002496, -0.00002273, 0.00001968, -0.00001347, 0.00001675, -0.00000832, -0.00001120, 0.00001550, -0.00001863, 0.00001076, 0.00000336, -0.00000246, 0.00000710, -0.00001653, 0.00001351, -0.00001232, 0.00000132, 0.00000857, 0.00001188, -0.00002380, 0.00000826, -0.00000795, 0.00000933, -0.00000970, 0.00001340, -0.00001852, 0.00001665, -0.00001260, 0.00002404, -0.00003106, 0.00001904, -0.00001992, 0.00003573, -0.00004750, 0.00003480, -0.00003067, 0.00003374, -0.00003902, 0.00004269, -0.00005024, 0.00004998, -0.00005869, 0.00004888, -0.00005708, 0.00006684, -0.00008788, 0.00008599, -0.00012330, 0.00012524, -0.00021786, 0.00023955, -0.00078982, 0.00121767, -0.00010061, -0.00033111, 0.00013733, -0.00014954, 0.00009318, -0.00009841, 0.00006776, -0.00007143, 0.00005756, -0.00004665, 0.00003179, -0.00003841, 0.00003320, -0.00003594, 0.00003192, -0.00003069, 0.00002899, -0.00002235, 0.00001447, -0.00003145, 0.00003856, -0.00002858, 0.00001697, -0.00002622, 0.00003360, -0.00002488, 0.00002067, -0.00002211, 0.00002186, -0.00001683, 0.00001324, -0.00000925, 0.00000097, -0.00001979, 0.00002645, -0.00000521, 0.00000053, -0.00000572, 0.00000433, -0.00001323, 0.00001606, -0.00001549, 0.00002031, -0.00001189, -0.00000547, 0.00001378, -0.00002588, 0.00002095, -0.00000854, 0.00001326, -0.00001107, 0.00001102, -0.00000369, -0.00001111, 0.00002032, -0.00002745, 0.00002390, -0.00001322, 0.00000512, -0.00000888, 0.00001163, -0.00001308, 0.00001299, -0.00000986, },
		{ 0.00000974, -0.00000813, 0.00000764, -0.00000604, 0.00000188, -0.00000869, 0.00001859, -0.00001988, 0.00001830, -0.00001342, -0.00000107, 0.00000552, -0.00001042, 0.00001380, -0.00001101, 0.00001856, -0.00002098, 0.00000391, 0.00000677, -0.00001939, 0.00002413, -0.00001460, 0.00001744, -0.00001799, 0.00000720, -0.00000445, 0.00000035, -0.00000861, 0.00002955, -0.00002106, 0.00000335, -0.00001317, 0.00001683, -0.00001814, 0.00002124, -0.00002211, 0.00001747, -0.00001806, 0.00002847, -0.00002454, 0.00001357, -0.00002437, 0.00003441, -0.00002754, 0.00001040, -0.00001600, 0.00002426, -0.00002881, 0.00002949, -0.00003138, 0.00002765, -0.00003230, 0.00002596, -0.00004243, 0.00004965, -0.00005793, 0.00005141, -0.00007681, 0.00006976, -0.00012206, 0.00009824, -0.00025871, -0.00020811, 0.00126921, -0.00075619, 0.00019569, -0.00018813, 0.00010344, -0.00010427, 0.00006600, -0.00007013, 0.00004953, -0.00004391, 0.00004101, -0.00005104, 0.00004223, -0.00004270, 0.00003562, -0.00003445, 0.00003183, -0.00002750, 0.00002894, -0.00004272, 0.00003147, -0.00001575, 0.00001445, -0.00002653, 0.00002234, -0.00000863, 0.00000858, -0.00001102, 0.00000872, -0.00000432, 0.00000495, -0.00000625, 0.00000853, -0.00002389, 0.00001095, 0.00000746, 0.00000450, -0.00001269, 0.00001052, -0.00001591, 0.00001023, -0.00000406, 0.00000052, 0.00001179, -0.00001676, 0.00000610, 0.00000093, -0.00001865, 0.00002634, -0.00001736, 0.00002307, -0.00002518, 0.00003322, -0.00003621, 0.00002815, -0.00001810, 0.00000238, 0.00000610, -0.00000404, -0.00000427, 0.00000364, -0.00000066, -0.00000331, 0.00000693, -0.00000725, },
		{ -0.00000083, 0.00000010, 0.00000142, -0.00000202, -0.00000465, 0.00000368, 0.00000660, -0.00001335, 0.00002048, -0.00002740, 0.00001977, -0.00001704, 0.00001123, -0.00000407, 0.00000438, 0.00000529, -0.00002043, 0.00001544, -0.00000966, -0.00000291, 0.00001600, -0.00001143, 0.00001606, -0.00002440, 0.00001865, -0.00001437, 0.00000422, -0.00000236, 0.00002600, -0.00003143, 0.00001430, -0.00001804, 0.00002172, -0.00002337, 0.00002753, -0.00003268, 0.00002975, -0.00002635, 0.00003844, -0.00004354, 0.00003159, -0.00003670, 0.00005275, -0.00005909, 0.00004418, -0.00004217, 0.00004821, -0.00005613, 0.00006119, -0.00006828, 0.00006817, -0.00007677, 0.00007005, -0.00008582, 0.00010006, -0.00012115, 0.00012353, -0.00016381, 0.00017499, -0.00027264, 0.00031139, -0.00075047, 0.00078550, 0.00046785, -0.00063240, 0.00027722, -0.00024937, 0.00016613, -0.00015750, 0.00011425, -0.00011096, 0.00008803, -0.00007374, 0.00006116, -0.00006771, 0.00005991, -0.00006020, 0.00005321, -0.00005085, 0.00004861, -0.00004172, 0.00003599, -0.00005279, 0.00005333, -0.00003926, 0.00002954, -0.00003962, 0.00004396, -0.00003206, 0.00002638, -0.00002661, 0.00002521, -0.00001866, 0.00001501, -0.00001246, 0.00000883, -0.00002798, 0.00002694, -0.00000472, 0.00000604, -0.00001276, 0.00001036, -0.00001843, 0.00001962, -0.00001673, 0.00001546, -0.00000228, -0.00001420, 0.00001536, -0.00001871, 0.00000681, 0.00000720, -0.00000056, 0.00000431, -0.00000513, 0.00001629, -0.00002941, 0.00003499, -0.00003703, 0.00002588, -0.00001313, 0.00000693, -0.00001231, 0.00001447, -0.00001422, 0.00001140, -0.00000706, 0.00000540, },
		{ -0.00000713, 0.00000657, -0.00000495, 0.00000177, -0.00000890, 0.00001568, -0.00001223, 0.00000943, -0.00000296, -0.00001131, 0.00001246, -0.00001694, 0.00001721, -0.00001233, 0.00001487, -0.00001062, -0.00000983, 0.00001706, -0.00002202, 0.00001875, -0.00000532, 0.00000899, -0.00000773, -0.00000457, 0.00000624, -0.00000601, -0.00000574, 0.00002061, -0.00000475, -0.00000987, -0.00000321, 0.00000455, -0.00000322, 0.00000386, -0.00000256, -0.00000406, 0.00000344, 0.00000574, 0.00000086, -0.00001153, 0.00000016, 0.00000474, 0.00000705, -0.00002237, 0.00001401, -0.00000617, 0.00000444, -0.00000803, 0.00001048, -0.00001680, 0.00001486, -0.00002254, 0.00000945, -0.00001263, 0.00001635, -0.00003162, 0.00002064, -0.00004438, 0.00002426, -0.00009342, 0.00005398, -0.00060826, 0.00133989, -0.00045969, -0.00007184, -0.00001674, -0.00003112, 0.00000332, -0.00002566, 0.00000663, -0.00001793, 0.00001116, -0.00000473, -0.00000727, -0.00000059, -0.00000233, -0.00000223, -0.00000031, 0.00000008, 0.00000250, 0.00000086, -0.00001273, -0.00000170, 0.00001366, -0.00000935, -0.00000298, -0.00000305, 0.00001567, -0.00001258, 0.00000765, -0.00000792, 0.00001033, -0.00000650, 0.00000248, 0.00000146, -0.00001295, -0.00000380, 0.00001581, 0.00000131, -0.00000645, 0.00000219, -0.00000739, 0.00000105, 0.00000531, -0.00000858, 0.00001530, -0.00001143, -0.00000625, 0.00001389, -0.00002749, 0.00002817, -0.00001467, 0.00002089, -0.00002122, 0.00002801, -0.00002426, 0.00001220, -0.00000007, -0.00001540, 0.00001743, -0.00001056, 0.00000159, -0.00000360, 0.00000718, -0.00001069, 0.00001186, -0.00001024, 0.00001069, },
		{ -0.00000126, 0.00000223, -0.00000234, -0.00000328, -0.00000005, 0.00001024, -0.00001340, 0.00001871, -0.00002229, 0.00001141, -0.00000897, 0.00000100, 0.00000568, -0.00000551, 0.00001241, -0.00001998, 0.00000579, 0.00000469, -0.00001640, 0.00002422, -0.00001482, 0.00001959, -0.00002606, 0.00001697, -0.00001087, 0.00000392, -0.00000942, 0.00003241, -0.00003010, 0.00001304, -0.00002093, 0.00002506, -0.00002595, 0.00002895, -0.00003257, 0.00002731, -0.00002407, 0.00003632, -0.00003895, 0.00002674, -0.00003455, 0.00004823, -0.00004840, 0.00003218, -0.00003271, 0.00004111, -0.00004964, 0.00005318, -0.00005736, 0.00005556, -0.00006253, 0.00005699, -0.00007537, 0.00008638, -0.00010021, 0.00009884, -0.00013181, 0.00013701, -0.00021558, 0.00022637, -0.00050933, 0.00022976, 0.00096787, -0.00076842, 0.00027899, -0.00024751, 0.00015746, -0.00014832, 0.00010244, -0.00010041, 0.00007595, -0.00006573, 0.00005945, -0.00006709, 0.00005798, -0.00005759, 0.00004991, -0.00004873, 0.00004722, -0.00004069, 0.00003708, -0.00005276, 0.00004723, -0.00003170, 0.00002558, -0.00003657, 0.00003786, -0.00002351, 0.00001794, -0.00001891, 0.00001800, -0.00001167, 0.00001004, -0.00001050, 0.00001100, -0.00002863, 0.00002098, -0.00000105, 0.00000862, -0.00001502, 0.00001110, -0.00001876, 0.00001837, -0.00001303, 0.00000768, 0.00000536, -0.00001554, 0.00000685, -0.00000315, -0.00001185, 0.00002342, -0.00001287, 0.00001766, -0.00001860, 0.00003178, -0.00004026, 0.00004040, -0.00003639, 0.00001905, -0.00000732, 0.00000471, -0.00001177, 0.00001287, -0.00001081, 0.00000617, -0.00000211, 0.00000092, 0.00000251, },
		{ 0.00000643, -0.00000496, 0.00000247, -0.00000929, 0.00001299, -0.00000659, 0.00000428, 0.00000290, -0.00001589, 0.00001368, -0.00001746, 0.00001394, -0.00000671, 0.00000552, 0.00000152, -0.00001961, 0.00001844, -0.00001464, 0.00000567, 0.00000796, -0.00000143, 0.00000415, -0.00001687, 0.00001578, -0.00001128, -0.00000092, 0.00000778, 0.00001311, -0.00002261, 0.00000761, -0.00001009, 0.00001392, -0.00001482, 0.00001731, -0.00002460, 0.00002251, -0.00001392, 0.00002275, -0.00003266, 0.00002097, -0.00002077, 0.00003506, -0.00004667, 0.00003565, -0.00002918, 0.00003128, -0.00003911, 0.00004450, -0.00005180, 0.00005139, -0.00006015, 0.00005134, -0.00006412, 0.00007567, -0.00009495, 0.00009314, -0.00012698, 0.00012833, -0.00022222, 0.00024361, -0.00075868, 0.00110064, 0.00002897, -0.00039210, 0.00016091, -0.00016403, 0.00010568, -0.00010997, 0.00007483, -0.00007635, 0.00005795, -0.00004559, 0.00003414, -0.00004107, 0.00003509, -0.00003668, 0.00003072, -0.00002958, 0.00003107, -0.00002607, 0.00001590, -0.00003153, 0.00003785, -0.00002843, 0.00001686, -0.00002440, 0.00003469, -0.00002735, 0.00001961, -0.00001832, 0.00001920, -0.00001290, 0.00000784, -0.00000453, -0.00000231, -0.00001619, 0.00002110, -0.00000162, 0.00000114, -0.00000644, 0.00000052, -0.00000727, 0.00001304, -0.00001398, 0.00001397, -0.00000363, -0.00001539, 0.00001745, -0.00002372, 0.00001722, -0.00000125, 0.00000963, -0.00000836, 0.00001354, -0.00000353, -0.00000988, 0.00002106, -0.00003275, 0.00002642, -0.00001574, 0.00000745, -0.00001108, 0.00001452, -0.00001643, 0.00001464, -0.00001141, 0.00001041, -0.00000611, },
		{ 0.00000372, -0.00000356, -0.00000125, -0.00000338, 0.00001220, -0.00001207, 0.00001649, -0.00001730, 0.00000438, -0.00000281, -0.00000649, 0.00001094, -0.00000852, 0.00001063, -0.00001096, -0.00000852, 0.00001741, -0.00002317, 0.00002366, -0.00001039, 0.00001637, -0.00002058, 0.00000837, -0.00000229, 0.00000005, -0.00001071, 0.00002930, -0.00001926, 0.00000412, -0.00001535, 0.00001762, -0.00001619, 0.00001716, -0.00001856, 0.00001085, -0.00000832, 0.00002073, -0.00002055, 0.00000787, -0.00001715, 0.00002692, -0.00002144, 0.00000539, -0.00000873, 0.00001834, -0.00002513, 0.00002468, -0.00002454, 0.00002019, -0.00002457, 0.00001827, -0.00003460, 0.00003676, -0.00003961, 0.00003068, -0.00005115, 0.00004234, -0.00009015, 0.00005552, -0.00019191, -0.00027486, 0.00124485, -0.00065939, 0.00011900, -0.00013987, 0.00006862, -0.00007500, 0.00003818, -0.00004651, 0.00002809, -0.00002802, 0.00003008, -0.00003907, 0.00003006, -0.00003106, 0.00002499, -0.00002675, 0.00002658, -0.00002076, 0.00001942, -0.00003382, 0.00002386, -0.00000989, 0.00000873, -0.00002050, 0.00001909, -0.00000380, 0.00000055, -0.00000427, 0.00000514, -0.00000027, 0.00000189, -0.00000555, 0.00000866, -0.00002318, 0.00001060, 0.00000465, 0.00000803, -0.00001221, 0.00000703, -0.00001506, 0.00001334, -0.00000621, -0.00000026, 0.00000886, -0.00001026, -0.00000667, 0.00001235, -0.00002543, 0.00003118, -0.00001609, 0.00002233, -0.00002288, 0.00003630, -0.00003804, 0.00003257, -0.00002359, 0.00000323, 0.00000385, -0.00000176, -0.00000657, 0.00000628, -0.00000288, -0.00000213, 0.00000439, -0.00000413, 0.00000574, -0.00000272, },
		{ -0.00000506, 0.00000302, -0.00000936, 0.00001060, -0.00000291, 0.00000189, 0.00000595, -0.00001726, 0.00001203, -0.00001505, 0.00000775, 0.00000069, -0.00000284, 0.00000839, -0.00002005, 0.00000936, 0.00000015, -0.00000997, 0.00001966, -0.00000900, 0.00001321, -0.00002516, 0.00002017, -0.00001245, 0.00000263, -0.00000389, 0.00002631, -0.00002911, 0.00001373, -0.00002038, 0.00002515, -0.00002577, 0.00002791, -0.00003421, 0.00002937, -0.00002115, 0.00003192, -0.00004041, 0.00002813, -0.00003146, 0.00004570, -0.00005234, 0.00003923, -0.00003490, 0.00004082, -0.00005131, 0.00005692, -0.00006279, 0.00006212, -0.00007063, 0.00006474, -0.00008292, 0.00009516, -0.00011150, 0.00011119, -0.00014527, 0.00015271, -0.00024521, 0.00027144, -0.00066631, 0.00064504, 0.00056194, -0.00063410, 0.00025411, -0.00023048, 0.00015224, -0.00014591, 0.00010079, -0.00009804, 0.00007365, -0.00006168, 0.00005422, -0.00006086, 0.00005244, -0.00005216, 0.00004447, -0.00004379, 0.00004503, -0.00003881, 0.00003032, -0.00004587, 0.00004618, -0.00003356, 0.00002432, -0.00003325, 0.00004055, -0.00002901, 0.00001969, -0.00001860, 0.00001938, -0.00001218, 0.00000824, -0.00000748, 0.00000521, -0.00002354, 0.00002182, -0.00000294, 0.00000763, -0.00001246, 0.00000546, -0.00001290, 0.00001764, -0.00001541, 0.00000998, 0.00000227, -0.00001678, 0.00000961, -0.00000914, -0.00000172, 0.00001654, -0.00000419, 0.00000696, -0.00000324, 0.00001797, -0.00002936, 0.00003673, -0.00004229, 0.00002774, -0.00001580, 0.00000990, -0.00001528, 0.00001790, -0.00001769, 0.00001343, -0.00000975, 0.00000846, -0.00000473, 0.00000652, },
		{ -0.00000621, 0.00000194, -0.00000721, 0.00001427, -0.00001183, 0.00001618, -0.00001440, 0.00000018, 0.00000028, -0.00001011, 0.00001089, -0.00000525, 0.00000312, 0.00000054, -0.00001967, 0.00002148, -0.00001918, 0.00001355, 0.00000056, 0.00000823, -0.00001005, -0.00000439, 0.00000843, -0.00000552, -0.00000724, 0.00001874, -0.00000253, -0.00000880, -0.00000441, 0.00000306, 0.00000102, -0.00000162, 0.00000192, -0.00001099, 0.00001150, 0.00000127, 0.00000190, -0.00001491, 0.00000472, 0.00000080, 0.00000870, -0.00002292, 0.00001676, -0.00000724, 0.00000408, -0.00000960, 0.00001373, -0.00001999, 0.00001799, -0.00002587, 0.00001405, -0.00002217, 0.00002883, -0.00004372, 0.00003451, -0.00005651, 0.00003716, -0.00010965, 0.00007425, -0.00059274, 0.00122489, -0.00031713, -0.00015151, 0.00002314, -0.00005777, 0.00002545, -0.00004589, 0.00002198, -0.00003067, 0.00001870, -0.00000973, 0.00000066, -0.00000871, 0.00000489, -0.00000809, 0.00000294, -0.00000217, 0.00000728, -0.00000618, -0.00000716, -0.00000527, 0.00001523, -0.00001117, -0.00000069, -0.00000303, 0.00001758, -0.00001660, 0.00000913, -0.00000645, 0.00000977, -0.00000503, -0.00000064, 0.00000417, -0.00001447, -0.00000120, 0.00001058, 0.00000505, -0.00000558, 0.00000008, -0.00000916, 0.00000616, 0.00000186, -0.00000683, 0.00000884, -0.00000266, -0.00001753, 0.00001912, -0.00002669, 0.00002559, -0.00000745, 0.00001649, -0.00001639, 0.00002880, -0.00002308, 0.00001371, -0.00000246, -0.00001693, 0.00001649, -0.00000983, 0.00000125, -0.00000319, 0.00000711, -0.00001107, 0.00001060, -0.00000878, 0.00000875, -0.00000484, 0.00000661, },
		{ 0.00000176, -0.00000769, 0.00000714, 0.00000079, -0.00000066, 0.00000954, -0.00001888, 0.00001091, -0.00001308, 0.00000247, 0.00000529, -0.00000560, 0.00000696, -0.00001146, -0.00000584, 0.00001534, -0.00002088, 0.00002403, -0.00000975, 0.00001619, -0.00002643, 0.00001715, -0.00000803, 0.00000268, -0.00001047, 0.00003055, -0.00002609, 0.00001196, -0.00002209, 0.00002551, -0.00002409, 0.00002476, -0.00002902, 0.00002110, -0.00001391, 0.00002657, -0.00003293, 0.00001967, -0.00002534, 0.00003760, -0.00003901, 0.00002484, -0.00002312, 0.00003200, -0.00004264, 0.00004551, -0.00004790, 0.00004551, -0.00005232, 0.00004700, -0.00006541, 0.00007168, -0.00007978, 0.00007524, -0.00010184, 0.00010312, -0.00017482, 0.00017120, -0.00042029, 0.00012950, 0.00097398, -0.00068774, 0.00020607, -0.00019803, 0.00012092, -0.00011781, 0.00007342, -0.00007514, 0.00005238, -0.00004733, 0.00004641, -0.00005323, 0.00004373, -0.00004355, 0.00003640, -0.00003795, 0.00003952, -0.00003253, 0.00002607, -0.00004113, 0.00003643, -0.00002336, 0.00001799, -0.00002817, 0.00003238, -0.00001778, 0.00000919, -0.00001056, 0.00001266, -0.00000604, 0.00000490, -0.00000747, 0.00000867, -0.00002508, 0.00001792, -0.00000202, 0.00001118, -0.00001397, 0.00000624, -0.00001499, 0.00001854, -0.00001332, 0.00000518, 0.00000430, -0.00001079, -0.00000511, 0.00000822, -0.00001901, 0.00002921, -0.00001210, 0.00001690, -0.00001432, 0.00003179, -0.00003812, 0.00004014, -0.00003928, 0.00001881, -0.00000910, 0.00000707, -0.00001398, 0.00001528, -0.00001315, 0.00000766, -0.00000501, 0.00000447, -0.00000217, 0.00000538, -0.00000348, },
		{ 0.00000515, -0.00001077, 0.00001620, -0.00001239, 0.00001679, -0.00001204, -0.00000307, 0.00000228, -0.00001187, 0.00000819, 0.00000010, -0.00000400, 0.00000738, -0.00002152, 0.00001409, -0.00000562, -0.00000206, 0.00001311, -0.00000040, 0.00000121, -0.00001661, 0.00001694, -0.00000956, -0.00000217, 0.00000558, 0.00001387, -0.00001979, 0.00000610, -0.00001178, 0.00001729, -0.00001794, 0.00001860, -0.00002737, 0.00002453, -0.00001155, 0.00001789, -0.00003071, 0.00001954, -0.00001772, 0.00002918, -0.00004016, 0.00003154, -0.00002326, 0.00002464, -0.00003454, 0.00004064, -0.00004695, 0.00004609, -0.00005496, 0.00004740, -0.00006315, 0.00007385, -0.00008942, 0.00008596, -0.00011438, 0.00011213, -0.00020114, 0.00021028, -0.00067304, 0.00093906, 0.00015700, -0.00042506, 0.00015964, -0.00015976, 0.00010289, -0.00010802, 0.00007013, -0.00007120, 0.00005028, -0.00003906, 0.00003237, -0.00003931, 0.00003244, -0.00003302, 0.00002549, -0.00002503, 0.00002969, -0.00002614, 0.00001380, -0.00002720, 0.00003202, -0.00002386, 0.00001320, -0.00001895, 0.00003148, -0.00002544, 0.00001483, -0.00001172, 0.00001451, -0.00000783, 0.00000217, -0.00000035, -0.00000536, -0.00001154, 0.00001502, 0.00000105, 0.00000258, -0.00000706, -0.00000327, -0.00000134, 0.00000955, -0.00001147, 0.00000729, 0.00000257, -0.00002010, 0.00001342, -0.00001519, 0.00000911, 0.00000891, 0.00000462, -0.00000354, 0.00001429, -0.00000176, -0.00000850, 0.00001877, -0.00003384, 0.00002476, -0.00001526, 0.00000793, -0.00001162, 0.00001523, -0.00001723, 0.00001407, -0.00001141, 0.00001057, -0.00000693, 0.00000993, -0.00000909, },
		{ -0.00000361, 0.00000172, 0.00000609, -0.00000541, 0.00001545, -0.00002211, 0.00001135, -0.00001248, -0.00000064, 0.00000584, -0.00000267, -0.00000033, 0.00000051, -0.00001915, 0.00002296, -0.00002136, 0.00001831, -0.00000294, 0.00001285, -0.00002069, 0.00000758, 0.00000077, -0.00000065, -0.00001089, 0.00002546, -0.00001457, 0.00000332, -0.00001555, 0.00001555, -0.00001140, 0.00001063, -0.00001292, 0.00000279, 0.00000201, 0.00001255, -0.00001608, 0.00000168, -0.00000896, 0.00001827, -0.00001521, 0.00000129, -0.00000231, 0.00001276, -0.00002102, 0.00001907, -0.00001728, 0.00001286, -0.00001726, 0.00001073, -0.00002557, 0.00002246, -0.00002127, 0.00001122, -0.00002784, 0.00001804, -0.00006210, 0.00001977, -0.00014346, -0.00029163, 0.00114755, -0.00051315, 0.00002460, -0.00008493, 0.00002990, -0.00004319, 0.00000944, -0.00002311, 0.00000770, -0.00001332, 0.00001893, -0.00002547, 0.00001626, -0.00001808, 0.00001327, -0.00001796, 0.00001976, -0.00001239, 0.00000850, -0.00002322, 0.00001521, -0.00000431, 0.00000360, -0.00001454, 0.00001585, 0.00000008, -0.00000550, 0.00000009, 0.00000384, 0.00000163, 0.00000052, -0.00000581, 0.00000853, -0.00002164, 0.00001054, 0.00000100, 0.00001098, -0.00001057, 0.00000309, -0.00001372, 0.00001582, -0.00000829, 0.00000030, 0.00000344, -0.00000196, -0.00001855, 0.00001863, -0.00002572, 0.00002968, -0.00000891, 0.00001712, -0.00001533, 0.00003374, -0.00003309, 0.00003015, -0.00002461, 0.00000186, 0.00000225, 0.00000008, -0.00000778, 0.00000746, -0.00000408, -0.00000116, 0.00000167, -0.00000102, 0.00000177, 0.00000217, 0.00000074, -0.00000209, },
		{ -0.00001219, 0.00001621, -0.00001147, 0.00001532, -0.00000705, -0.00000836, 0.00000592, -0.00001439, 0.00000606, 0.00000323, -0.00000593, 0.00000588, -0.00001339, -0.00000137, 0.00001102, -0.00001582, 0.00002090, -0.00000459, 0.00000879, -0.00002387, 0.00001950, -0.00000923, 0.00000112, -0.00000496, 0.00002405, -0.00002370, 0.00001105, -0.00002056, 0.00002530, -0.00002421, 0.00002401, -0.00003115, 0.00002445, -0.00001213, 0.00002186, -0.00003372, 0.00002119, -0.00002219, 0.00003372, -0.00004077, 0.00003039, -0.00002416, 0.00002984, -0.00004205, 0.00004719, -0.00005138, 0.00005011, -0.00005858, 0.00005323, -0.00007178, 0.00007990, -0.00009065, 0.00008704, -0.00011408, 0.00011595, -0.00019938, 0.00020722, -0.00055493, 0.00050286, 0.00061362, -0.00058470, 0.00019708, -0.00018897, 0.00012096, -0.00011986, 0.00007567, -0.00007592, 0.00005244, -0.00004495, 0.00004297, -0.00004892, 0.00003987, -0.00003934, 0.00003141, -0.00003280, 0.00003725, -0.00003149, 0.00002049, -0.00003421, 0.00003421, -0.00002415, 0.00001620, -0.00002378, 0.00003359, -0.00002292, 0.00001116, -0.00000984, 0.00001348, -0.00000633, 0.00000258, -0.00000375, 0.00000202, -0.00001842, 0.00001667, -0.00000231, 0.00000960, -0.00001170, 0.00000060, -0.00000748, 0.00001536, -0.00001381, 0.00000561, 0.00000326, -0.00001394, -0.00000145, 0.00000223, -0.00000959, 0.00002404, -0.00000515, 0.00000805, 0.00000077, 0.00001705, -0.00002484, 0.00003167, -0.00004083, 0.00002416, -0.00001479, 0.00001023, -0.00001554, 0.00001814, -0.00001799, 0.00001301, -0.00001073, 0.00001006, -0.00000755, 0.00001172, -0.00000997, 0.00000626, },
		{ -0.00000485, 0.00001251, -0.00001171, 0.00002229, -0.00002504, 0.00001125, -0.00001129, -0.00000329, 0.00000447, 0.00000224, -0.00000757, 0.00000840, -0.00002344, 0.00001862, -0.00001075, 0.00000477, 0.00000803, 0.00000632, -0.00001105, -0.00000485, 0.00001039, -0.00000494, -0.00000715, 0.00001448, 0.00000062, -0.00000754, -0.00000517, 0.00000066, 0.00000543, -0.00000645, 0.00000510, -0.00001581, 0.00001676, -0.00000067, 0.00000068, -0.00001601, 0.00000727, -0.00000097, 0.00000715, -0.00001960, 0.00001605, -0.00000548, 0.00000140, -0.00000861, 0.00001354, -0.00001901, 0.00001662, -0.00002479, 0.00001469, -0.00002647, 0.00003363, -0.00004631, 0.00003730, -0.00005607, 0.00003520, -0.00010556, 0.00006204, -0.00052272, 0.00105478, -0.00015862, -0.00021832, 0.00004829, -0.00007280, 0.00003772, -0.00005681, 0.00002885, -0.00003596, 0.00002034, -0.00001119, 0.00000638, -0.00001437, 0.00000953, -0.00001146, 0.00000407, -0.00000285, 0.00001036, -0.00001119, -0.00000336, -0.00000642, 0.00001353, -0.00001000, -0.00000067, -0.00000069, 0.00001629, -0.00001689, 0.00000729, -0.00000244, 0.00000731, -0.00000253, -0.00000384, 0.00000595, -0.00001497, 0.00000150, 0.00000529, 0.00000744, -0.00000345, -0.00000243, -0.00001019, 0.00001030, -0.00000118, -0.00000460, 0.00000210, 0.00000429, -0.00002371, 0.00001637, -0.00001895, 0.00001771, 0.00000390, 0.00000945, -0.00000798, 0.00002596, -0.00001781, 0.00001198, -0.00000461, -0.00001672, 0.00001319, -0.00000742, 0.00000017, -0.00000228, 0.00000601, -0.00000986, 0.00000788, -0.00000638, 0.00000614, -0.00000246, 0.00000673, -0.00000739, 0.00000177, },
		{ 0.00001276, -0.00000700, 0.00000946, 0.00000236, -0.00001683, 0.00001187, -0.00001826, 0.00000560, 0.00000270, -0.00000212, -0.00000217, -0.00000042, -0.00001695, 0.00002216, -0.00002069, 0.00001948, -0.00000179, 0.00001052, -0.00002413, 0.00001492, -0.00000390, 0.00000098, -0.00000982, 0.00002525, -0.00001892, 0.00000877, -0.00002072, 0.00002240, -0.00001854, 0.00001717, -0.00002232, 0.00001240, -0.00000232, 0.00001559, -0.00002550, 0.00001125, -0.00001438, 0.00002477, -0.00002796, 0.00001675, -0.00001292, 0.00002182, -0.00003361, 0.00003514, -0.00003582, 0.00003316, -0.00003996, 0.00003449, -0.00005132, 0.00005217, -0.00005539, 0.00004852, -0.00006936, 0.00006661, -0.00013063, 0.00011296, -0.00033530, 0.00006320, 0.00091623, -0.00055626, 0.00010957, -0.00013638, 0.00007591, -0.00008085, 0.00004020, -0.00004744, 0.00002776, -0.00002857, 0.00003200, -0.00003679, 0.00002700, -0.00002734, 0.00002099, -0.00002529, 0.00002942, -0.00002189, 0.00001299, -0.00002710, 0.00002355, -0.00001406, 0.00000988, -0.00001886, 0.00002574, -0.00001180, 0.00000148, -0.00000392, 0.00000919, -0.00000244, 0.00000159, -0.00000576, 0.00000648, -0.00002085, 0.00001497, -0.00000365, 0.00001322, -0.00001198, 0.00000124, -0.00001100, 0.00001803, -0.00001325, 0.00000379, 0.00000061, -0.00000350, -0.00001724, 0.00001546, -0.00002068, 0.00002953, -0.00000605, 0.00001238, -0.00000554, 0.00002688, -0.00002954, 0.00003209, -0.00003582, 0.00001436, -0.00000840, 0.00000746, -0.00001382, 0.00001494, -0.00001306, 0.00000754, -0.00000677, 0.00000681, -0.00000569, 0.00001038, -0.00000719, 0.00000381, -0.00001012, },
		{ 0.00001773, -0.00001665, 0.00002639, -0.00002392, 0.00000726, -0.00000668, -0.00000798, 0.00000424, 0.00000480, -0.00000956, 0.00000758, -0.00001669, 0.00000418, 0.00000576, -0.00000996, 0.00001707, 0.00000143, -0.00000204, -0.00001537, 0.00001656, -0.00000698, -0.00000277, 0.00000290, 0.00001332, -0.00001502, 0.00000342, -0.00001226, 0.00001827, -0.00001782, 0.00001621, -0.00002596, 0.00002227, -0.00000575, 0.00001001, -0.00002566, 0.00001517, -0.00001149, 0.00001932, -0.00002950, 0.00002393, -0.00001440, 0.00001514, -0.00002638, 0.00003211, -0.00003682, 0.00003530, -0.00004427, 0.00003772, -0.00005445, 0.00006200, -0.00007276, 0.00006683, -0.00008895, 0.00008118, -0.00016080, 0.00014989, -0.00055163, 0.00075905, 0.00026097, -0.00041871, 0.00013008, -0.00013599, 0.00008472, -0.00009285, 0.00005453, -0.00005738, 0.00003627, -0.00002844, 0.00002707, -0.00003337, 0.00002566, -0.00002557, 0.00001699, -0.00001761, 0.00002506, -0.00002265, 0.00000850, -0.00001926, 0.00002229, -0.00001612, 0.00000710, -0.00001090, 0.00002497, -0.00002034, 0.00000796, -0.00000414, 0.00000955, -0.00000336, -0.00000221, 0.00000209, -0.00000713, -0.00000712, 0.00000951, 0.00000215, 0.00000490, -0.00000784, -0.00000620, 0.00000348, 0.00000662, -0.00000918, 0.00000192, 0.00000500, -0.00001875, 0.00000296, -0.00000357, 0.00000012, 0.00001891, 0.00000070, 0.00000150, 0.00001507, 0.00000000, -0.00000578, 0.00001225, -0.00002973, 0.00001850, -0.00001155, 0.00000615, -0.00000995, 0.00001321, -0.00001513, 0.00001114, -0.00000971, 0.00000927, -0.00000667, 0.00001216, -0.00001159, 0.00000600, -0.00001025, },
		{ 0.00000096, -0.00000034, 0.00001473, -0.00002626, 0.00001763, -0.00002119, 0.00000501, 0.00000050, 0.00000399, -0.00001095, 0.00000956, -0.00002450, 0.00002171, -0.00001397, 0.00000905, 0.00000647, 0.00000773, -0.00001871, 0.00000523, 0.00000430, -0.00000161, -0.00000927, 0.00001881, -0.00000810, 0.00000166, -0.00001427, 0.00001138, -0.00000512, 0.00000344, -0.00000726, -0.00000434, 0.00001046, 0.00000601, -0.00001285, -0.00000321, -0.00000183, 0.00001056, -0.00001054, -0.00000059, 0.00000208, 0.00000855, -0.00001751, 0.00001399, -0.00001119, 0.00000729, -0.00001197, 0.00000502, -0.00001766, 0.00001017, -0.00000707, -0.00000245, -0.00001164, 0.00000177, -0.00004335, -0.00000276, -0.00011877, -0.00026055, 0.00099291, -0.00034040, -0.00006801, -0.00003674, -0.00000307, -0.00001661, -0.00001360, -0.00000543, -0.00000734, -0.00000292, 0.00001002, -0.00001288, 0.00000375, -0.00000654, 0.00000291, -0.00001014, 0.00001330, -0.00000451, -0.00000140, -0.00001337, 0.00000766, -0.00000060, 0.00000028, -0.00000981, 0.00001375, 0.00000182, -0.00000851, 0.00000149, 0.00000496, 0.00000152, 0.00000044, -0.00000646, 0.00000767, -0.00001893, 0.00001012, -0.00000233, 0.00001242, -0.00000761, -0.00000127, -0.00001154, 0.00001693, -0.00000940, 0.00000098, -0.00000230, 0.00000495, -0.00002573, 0.00001701, -0.00001901, 0.00002260, 0.00000290, 0.00000887, -0.00000378, 0.00002699, -0.00002296, 0.00002217, -0.00002245, -0.00000043, 0.00000052, 0.00000182, -0.00000819, 0.00000756, -0.00000469, -0.00000016, -0.00000112, 0.00000177, -0.00000166, 0.00000594, -0.00000099, -0.00000168, -0.00000503, 0.00000613, },
		{ -0.00001712, 0.00002444, -0.00001611, -0.00000234, 0.00000226, -0.00001534, 0.00000632, 0.00000324, -0.00000503, -0.00000104, -0.00000311, -0.00001294, 0.00001949, -0.00001818, 0.00001868, 0.00000189, 0.00000295, -0.00002049, 0.00001659, -0.00000482, -0.00000055, -0.00000496, 0.00001914, -0.00001568, 0.00000660, -0.00001848, 0.00002199, -0.00001881, 0.00001641, -0.00002446, 0.00001649, -0.00000133, 0.00001037, -0.00002535, 0.00001269, -0.00001121, 0.00001959, -0.00002721, 0.00002026, -0.00001243, 0.00001756, -0.00003047, 0.00003428, -0.00003662, 0.00003482, -0.00004326, 0.00003799, -0.00005514, 0.00005804, -0.00006361, 0.00005722, -0.00007757, 0.00007330, -0.00014576, 0.00013327, -0.00043626, 0.00037902, 0.00061383, -0.00048826, 0.00011516, -0.00013324, 0.00007926, -0.00008545, 0.00004462, -0.00004992, 0.00002913, -0.00002719, 0.00003002, -0.00003427, 0.00002478, -0.00002440, 0.00001659, -0.00002008, 0.00002711, -0.00002165, 0.00000871, -0.00002044, 0.00002020, -0.00001351, 0.00000732, -0.00001320, 0.00002506, -0.00001589, 0.00000307, -0.00000239, 0.00000922, -0.00000255, -0.00000092, -0.00000196, -0.00000013, -0.00001342, 0.00001211, -0.00000270, 0.00001149, -0.00001056, -0.00000363, -0.00000273, 0.00001313, -0.00001225, 0.00000265, 0.00000111, -0.00000754, -0.00001401, 0.00001075, -0.00001331, 0.00002738, -0.00000192, 0.00000666, 0.00000777, 0.00001307, -0.00001599, 0.00002001, -0.00003307, 0.00001589, -0.00001066, 0.00000806, -0.00001307, 0.00001523, -0.00001530, 0.00001031, -0.00000992, 0.00000984, -0.00000860, 0.00001462, -0.00001222, 0.00000679, -0.00001226, 0.00001923, },
		{ -0.00001085, 0.00002581, -0.00003213, 0.00001903, -0.00001969, 0.00000137, -0.00000006, 0.00000770, -0.00001454, 0.00001075, -0.00002034, 0.00000966, 0.00000093, -0.00000512, 0.00001510, 0.00000440, -0.00001131, -0.00000540, 0.00001154, -0.00000393, -0.00000602, 0.00000906, 0.00000352, -0.00000549, -0.00000597, -0.00000173, 0.00000865, -0.00000908, 0.00000553, -0.00001714, 0.00001814, 0.00000041, -0.00000298, -0.00001455, 0.00000740, -0.00000038, 0.00000275, -0.00001316, 0.00001266, -0.00000168, -0.00000295, -0.00000542, 0.00001020, -0.00001425, 0.00001123, -0.00001966, 0.00001117, -0.00002477, 0.00003012, -0.00003932, 0.00002960, -0.00004442, 0.00002046, -0.00008452, 0.00002466, -0.00041487, 0.00085618, -0.00000992, -0.00025650, 0.00005106, -0.00007199, 0.00003721, -0.00005624, 0.00002593, -0.00003326, 0.00001612, -0.00000904, 0.00000913, -0.00001650, 0.00001069, -0.00001177, 0.00000271, -0.00000154, 0.00001104, -0.00001330, -0.00000207, -0.00000495, 0.00000890, -0.00000639, -0.00000245, 0.00000352, 0.00001224, -0.00001414, 0.00000325, 0.00000279, 0.00000426, -0.00000032, -0.00000593, 0.00000591, -0.00001373, 0.00000326, 0.00000116, 0.00000786, -0.00000012, -0.00000541, -0.00000991, 0.00001257, -0.00000280, -0.00000319, -0.00000304, 0.00000729, -0.00002334, 0.00000629, -0.00000713, 0.00000780, 0.00001633, 0.00000245, 0.00000176, 0.00002184, -0.00001067, 0.00000863, -0.00000783, -0.00001382, 0.00000721, -0.00000325, -0.00000188, -0.00000051, 0.00000354, -0.00000698, 0.00000385, -0.00000324, 0.00000303, 0.00000004, 0.00000650, -0.00000784, 0.00000076, -0.00000149, 0.00000998, },
		{ 0.00001587, -0.00000232, -0.00001568, 0.00001309, -0.00002299, 0.00000889, -0.00000066, 0.00000287, -0.00001191, 0.00000931, -0.00002388, 0.00002302, -0.00001536, 0.00001122, 0.00000789, 0.00000356, -0.00001976, 0.00001078, 0.00000096, -0.00000097, -0.00000764, 0.00001742, -0.00000995, 0.00000447, -0.00001749, 0.00001660, -0.00001067, 0.00000802, -0.00001463, 0.00000372, 0.00000801, 0.00000580, -0.00001879, 0.00000367, -0.00000425, 0.00001254, -0.00001776, 0.00001001, -0.00000407, 0.00001240, -0.00002439, 0.00002424, -0.00002355, 0.00002100, -0.00002792, 0.00002199, -0.00003628, 0.00003221, -0.00003233, 0.00002460, -0.00004086, 0.00003462, -0.00009131, 0.00006183, -0.00026532, 0.00003543, 0.00080436, -0.00039352, 0.00000797, -0.00007622, 0.00003267, -0.00004583, 0.00001004, -0.00002345, 0.00000711, -0.00001319, 0.00001919, -0.00002086, 0.00001111, -0.00001225, 0.00000669, -0.00001335, 0.00001928, -0.00001128, 0.00000066, -0.00001371, 0.00001140, -0.00000605, 0.00000311, -0.00001035, 0.00001961, -0.00000720, -0.00000370, -0.00000004, 0.00000816, -0.00000115, 0.00000013, -0.00000517, 0.00000440, -0.00001607, 0.00001192, -0.00000514, 0.00001402, -0.00000904, -0.00000361, -0.00000680, 0.00001654, -0.00001232, 0.00000265, -0.00000385, 0.00000321, -0.00002549, 0.00001533, -0.00001583, 0.00002463, 0.00000445, 0.00000523, 0.00000666, 0.00001857, -0.00001655, 0.00001815, -0.00002781, 0.00000737, -0.00000630, 0.00000647, -0.00001179, 0.00001240, -0.00001113, 0.00000621, -0.00000744, 0.00000772, -0.00000754, 0.00001311, -0.00000847, 0.00000360, -0.00001026, 0.00001613, -0.00001206, },
		{ 0.00003104, -0.00003045, 0.00001315, -0.00001190, -0.00000660, 0.00000276, 0.00000652, -0.00001065, 0.00000270, -0.00000728, -0.00000778, 0.00001600, -0.00001541, 0.00001831, 0.00000457, -0.00000573, -0.00001292, 0.00001442, -0.00000347, -0.00000319, 0.00000063, 0.00001103, -0.00000857, -0.00000020, -0.00001129, 0.00001644, -0.00001424, 0.00001031, -0.00002086, 0.00001669, 0.00000202, 0.00000072, -0.00001892, 0.00000924, -0.00000385, 0.00000785, -0.00001721, 0.00001517, -0.00000483, 0.00000487, -0.00001651, 0.00002096, -0.00002375, 0.00002151, -0.00003048, 0.00002445, -0.00004028, 0.00004331, -0.00004928, 0.00004113, -0.00005723, 0.00004335, -0.00011133, 0.00007713, -0.00041640, 0.00058559, 0.00032501, -0.00037061, 0.00007632, -0.00009761, 0.00005551, -0.00006843, 0.00003197, -0.00003868, 0.00001935, -0.00001629, 0.00001982, -0.00002461, 0.00001631, -0.00001609, 0.00000696, -0.00000879, 0.00001835, -0.00001673, 0.00000144, -0.00000967, 0.00001095, -0.00000736, 0.00000042, -0.00000200, 0.00001692, -0.00001400, 0.00000109, 0.00000251, 0.00000593, -0.00000081, -0.00000430, 0.00000216, -0.00000709, -0.00000375, 0.00000532, 0.00000167, 0.00000764, -0.00000874, -0.00000781, 0.00000663, 0.00000475, -0.00000768, -0.00000147, 0.00000359, -0.00001274, -0.00001037, 0.00000648, -0.00000599, 0.00002598, -0.00000003, 0.00000515, 0.00001749, 0.00000058, -0.00000146, 0.00000169, -0.00002096, 0.00000858, -0.00000541, 0.00000248, -0.00000625, 0.00000873, -0.00001058, 0.00000634, -0.00000659, 0.00000657, -0.00000505, 0.00001245, -0.00001174, 0.00000449, -0.00000727, 0.00001680, -0.00001793, },
		{ 0.00001332, -0.00002795, 0.00002103, -0.00002677, 0.00000854, -0.00000322, 0.00000904, -0.00001837, 0.00001358, -0.00002323, 0.00001426, -0.00000292, -0.00000175, 0.00001570, 0.00000250, -0.00001559, 0.00000215, 0.00000747, -0.00000229, -0.00000663, 0.00001095, -0.00000151, 0.00000022, -0.00001232, 0.00000631, 0.00000097, -0.00000250, -0.00000349, -0.00000864, 0.00001534, 0.00000234, -0.00001179, -0.00000572, 0.00000288, 0.00000506, -0.00000818, 0.00000002, 0.00000422, 0.00000595, -0.00001497, 0.00001015, -0.00000709, 0.00000420, -0.00000935, 0.00000194, -0.00001234, 0.00000209, 0.00000067, -0.00000828, -0.00000431, -0.00000502, -0.00003515, -0.00001131, -0.00011551, -0.00019362, 0.00080530, -0.00016790, -0.00014012, -0.00000625, -0.00002318, -0.00000062, -0.00002666, 0.00000334, -0.00001478, 0.00000173, 0.00000467, -0.00000319, -0.00000529, 0.00000153, -0.00000437, -0.00000472, 0.00000857, 0.00000128, -0.00000849, -0.00000604, 0.00000251, 0.00000060, -0.00000089, -0.00000662, 0.00001308, 0.00000124, -0.00000860, 0.00000050, 0.00000767, 0.00000027, 0.00000075, -0.00000669, 0.00000561, -0.00001478, 0.00000871, -0.00000439, 0.00001187, -0.00000353, -0.00000584, -0.00000829, 0.00001610, -0.00000879, 0.00000062, -0.00000649, 0.00000810, -0.00002626, 0.00000758, -0.00000773, 0.00001295, 0.00001641, 0.00000037, 0.00000926, 0.00001889, -0.00001071, 0.00001114, -0.00001911, -0.00000219, -0.00000203, 0.00000375, -0.00000821, 0.00000712, -0.00000518, 0.00000110, -0.00000389, 0.00000402, -0.00000426, 0.00000845, -0.00000167, -0.00000195, -0.00000576, 0.00000980, -0.00000312, 0.00000342, },
		{ -0.00002040, 0.00000063, 0.00000053, -0.00001701, 0.00000758, 0.00000163, -0.00000170, -0.00000968, 0.00000707, -0.00002152, 0.00002297, -0.00001597, 0.00001307, 0.00000967, -0.00000350, -0.00001556, 0.00001191, 0.00000032, -0.00000229, -0.00000389, 0.00001235, -0.00000643, 0.00000146, -0.00001484, 0.00001602, -0.00001085, 0.00000683, -0.00001617, 0.00000786, 0.00000865, -0.00000011, -0.00001747, 0.00000486, -0.00000117, 0.00000637, -0.00001461, 0.00001137, -0.00000212, 0.00000627, -0.00001887, 0.00002083, -0.00002144, 0.00001931, -0.00002770, 0.00002208, -0.00003666, 0.00003448, -0.00003632, 0.00002842, -0.00004326, 0.00003334, -0.00009471, 0.00006301, -0.00032657, 0.00028555, 0.00056467, -0.00035874, 0.00002372, -0.00007543, 0.00003656, -0.00005054, 0.00001453, -0.00002595, 0.00000859, -0.00001203, 0.00001816, -0.00001973, 0.00001022, -0.00001043, 0.00000290, -0.00000817, 0.00001693, -0.00001168, -0.00000236, -0.00000753, 0.00000713, -0.00000419, -0.00000021, -0.00000351, 0.00001688, -0.00000985, -0.00000286, 0.00000249, 0.00000739, -0.00000130, -0.00000207, -0.00000205, -0.00000120, -0.00000888, 0.00000820, -0.00000350, 0.00001264, -0.00000901, -0.00000696, 0.00000119, 0.00001087, -0.00001050, 0.00000053, -0.00000264, -0.00000058, -0.00002379, 0.00001268, -0.00001119, 0.00002607, 0.00000538, 0.00000324, 0.00001729, 0.00000705, -0.00000473, 0.00000398, -0.00002124, 0.00000508, -0.00000492, 0.00000434, -0.00000868, 0.00001004, -0.00001055, 0.00000610, -0.00000773, 0.00000796, -0.00000773, 0.00001488, -0.00001158, 0.00000462, -0.00000951, 0.00001922, -0.00001902, 0.00002822, },
		{ -0.00003394, 0.00002182, -0.00002356, 0.00000292, -0.00000179, 0.00000996, -0.00001679, 0.00000759, -0.00001200, -0.00000222, 0.00001239, -0.00001324, 0.00001977, 0.00000357, -0.00001148, -0.00000544, 0.00001133, -0.00000210, -0.00000467, 0.00000389, 0.00000514, -0.00000238, -0.00000700, -0.00000349, 0.00000981, -0.00000881, 0.00000279, -0.00001484, 0.00001592, 0.00000379, -0.00000821, -0.00001118, 0.00000562, 0.00000179, -0.00000305, -0.00000536, 0.00000819, 0.00000267, -0.00000766, -0.00000114, 0.00000492, -0.00000723, 0.00000346, -0.00001203, 0.00000467, -0.00001821, 0.00002016, -0.00002556, 0.00001510, -0.00002619, -0.00000129, -0.00005432, -0.00002533, -0.00029061, 0.00065673, 0.00010764, -0.00025751, 0.00003043, -0.00005637, 0.00002522, -0.00004568, 0.00001502, -0.00002453, 0.00000794, -0.00000449, 0.00000924, -0.00001518, 0.00000867, -0.00000951, -0.00000051, 0.00000127, 0.00000956, -0.00001267, -0.00000291, -0.00000172, 0.00000274, -0.00000175, -0.00000482, 0.00000841, 0.00000673, -0.00000989, -0.00000127, 0.00000763, 0.00000195, 0.00000053, -0.00000616, 0.00000373, -0.00001056, 0.00000357, -0.00000122, 0.00000640, 0.00000384, -0.00000857, -0.00000816, 0.00001264, -0.00000264, -0.00000316, -0.00000578, 0.00000596, -0.00001740, -0.00000781, 0.00000408, -0.00000012, 0.00002665, -0.00000180, 0.00001050, 0.00001896, -0.00000401, 0.00000501, -0.00001267, -0.00000829, -0.00000078, 0.00000203, -0.00000464, 0.00000204, -0.00000011, -0.00000283, -0.00000094, 0.00000021, -0.00000032, 0.00000267, 0.00000554, -0.00000711, -0.00000135, 0.00000295, 0.00000637, -0.00000843, 0.00002540, },
		{ -0.00001429, 0.00001295, -0.00002555, 0.00001077, -0.00000317, 0.00000708, -0.00001931, 0.00001477, -0.00002471, 0.00001788, -0.00000625, 0.00000124, 0.00001727, -0.00000311, -0.00001445, 0.00000578, 0.00000560, -0.00000255, -0.00000478, 0.00000881, -0.00000116, 0.00000051, -0.00001353, 0.00000956, -0.00000244, -0.00000048, -0.00000818, -0.00000268, 0.00001495, -0.00000104, -0.00001427, -0.00000144, 0.00000309, 0.00000306, -0.00001016, 0.00000587, 0.00000226, 0.00000492, -0.00001632, 0.00001452, -0.00001301, 0.00001093, -0.00001803, 0.00001149, -0.00002303, 0.00001541, -0.00001458, 0.00000745, -0.00002040, 0.00001138, -0.00006160, 0.00002320, -0.00021414, 0.00004103, 0.00065758, -0.00022436, -0.00007939, -0.00003009, -0.00000007, -0.00001959, -0.00001146, -0.00000757, -0.00000629, -0.00000352, 0.00001002, -0.00000791, -0.00000117, -0.00000091, -0.00000411, -0.00000418, 0.00001108, -0.00000280, -0.00000864, -0.00000332, 0.00000208, -0.00000078, -0.00000130, -0.00000362, 0.00001492, -0.00000477, -0.00000596, 0.00000121, 0.00000913, -0.00000161, -0.00000013, -0.00000511, 0.00000213, -0.00001074, 0.00000846, -0.00000574, 0.00001316, -0.00000538, -0.00000804, -0.00000238, 0.00001378, -0.00001004, 0.00000082, -0.00000727, 0.00000677, -0.00002740, 0.00000729, -0.00000626, 0.00001697, 0.00001690, -0.00000207, 0.00001992, 0.00000975, -0.00000273, 0.00000169, -0.00001812, 0.00000012, -0.00000416, 0.00000495, -0.00000877, 0.00000866, -0.00000832, 0.00000433, -0.00000732, 0.00000735, -0.00000776, 0.00001358, -0.00000762, 0.00000166, -0.00000853, 0.00001731, -0.00001432, 0.00001871, -0.00000505, },
		{ 0.00001436, -0.00001358, -0.00000718, 0.00000320, 0.00000576, -0.00000839, -0.00000494, 0.00000315, -0.00001780, 0.00002207, -0.00001667, 0.00001620, 0.00000892, -0.00000957, -0.00000948, 0.00001073, 0.00000077, -0.00000368, -0.00000073, 0.00000730, -0.00000144, -0.00000393, -0.00000931, 0.00001229, -0.00000814, 0.00000226, -0.00001371, 0.00000976, 0.00000954, -0.00000789, -0.00001229, 0.00000353, 0.00000304, -0.00000252, -0.00000609, 0.00000770, 0.00000317, -0.00000402, -0.00000707, 0.00000968, -0.00001057, 0.00000769, -0.00001653, 0.00001049, -0.00002411, 0.00002242, -0.00002470, 0.00001539, -0.00002667, 0.00000727, -0.00006347, 0.00000679, -0.00028725, 0.00043724, 0.00034333, -0.00028804, 0.00000924, -0.00005388, 0.00002268, -0.00004111, 0.00000811, -0.00002004, 0.00000360, -0.00000550, 0.00001263, -0.00001503, 0.00000662, -0.00000688, -0.00000237, -0.00000055, 0.00001132, -0.00001017, -0.00000541, -0.00000080, 0.00000059, 0.00000013, -0.00000493, 0.00000589, 0.00000921, -0.00000832, -0.00000403, 0.00000694, 0.00000443, -0.00000061, -0.00000404, 0.00000011, -0.00000543, -0.00000156, 0.00000245, 0.00000027, 0.00001001, -0.00000949, -0.00000808, 0.00000812, 0.00000382, -0.00000679, -0.00000332, -0.00000026, -0.00000499, -0.00002207, 0.00001066, -0.00000684, 0.00002884, 0.00000324, 0.00000693, 0.00002216, -0.00000017, 0.00000341, -0.00001114, -0.00000964, -0.00000278, 0.00000150, -0.00000199, -0.00000145, 0.00000279, -0.00000470, 0.00000070, -0.00000277, 0.00000293, -0.00000226, 0.00001075, -0.00000957, 0.00000090, -0.00000213, 0.00001338, -0.00001585, 0.00003015, -0.00003105, },
		{ 0.00002021, -0.00002805, 0.00000903, -0.00000436, 0.00001112, -0.00002120, 0.00001200, -0.00001637, 0.00000333, 0.00000856, -0.00001116, 0.00002243, -0.00000130, -0.00001245, -0.00000065, 0.00000940, -0.00000200, -0.00000410, 0.00000371, 0.00000360, -0.00000018, -0.00001041, 0.00000146, 0.00000548, -0.00000590, -0.00000261, -0.00000936, 0.00001623, 0.00000154, -0.00001278, -0.00000577, 0.00000485, 0.00000203, -0.00000783, 0.00000241, 0.00000472, 0.00000444, -0.00001312, 0.00000752, -0.00000491, 0.00000332, -0.00000901, 0.00000125, -0.00000982, -0.00000132, 0.00000200, -0.00000706, -0.00000448, -0.00000431, -0.00003491, -0.00000983, -0.00012546, -0.00010876, 0.00061185, -0.00001985, -0.00017873, 0.00000131, -0.00002806, 0.00000333, -0.00002892, 0.00000304, -0.00001494, 0.00000117, 0.00000284, 0.00000286, -0.00000991, 0.00000533, -0.00000800, -0.00000209, 0.00000603, 0.00000436, -0.00001209, -0.00000183, -0.00000007, -0.00000031, -0.00000049, -0.00000451, 0.00001334, -0.00000095, -0.00000679, -0.00000156, 0.00001066, -0.00000104, 0.00000058, -0.00000595, 0.00000223, -0.00000930, 0.00000612, -0.00000479, 0.00000950, 0.00000099, -0.00001017, -0.00000411, 0.00001332, -0.00000632, -0.00000120, -0.00000828, 0.00000682, -0.00002070, -0.00000669, 0.00000358, 0.00000473, 0.00002826, -0.00000536, 0.00002100, 0.00001266, 0.00000019, -0.00000025, -0.00001634, -0.00000251, -0.00000549, 0.00000584, -0.00000808, 0.00000652, -0.00000584, 0.00000259, -0.00000642, 0.00000558, -0.00000598, 0.00000980, -0.00000174, -0.00000234, -0.00000615, 0.00001289, -0.00000636, 0.00000646, 0.00001029, -0.00002595, },
		{ -0.00000055, -0.00001796, 0.00000849, -0.00000019, 0.00000196, -0.00001704, 0.00001386, -0.00002461, 0.00002066, -0.00000977, 0.00000538, 0.00001719, -0.00000925, -0.00001012, 0.00000643, 0.00000530, -0.00000362, -0.00000235, 0.00000522, 0.00000212, -0.00000288, -0.00001081, 0.00000887, -0.00000227, -0.00000246, -0.00000861, 0.00000093, 0.00001557, -0.00000769, -0.00001172, -0.00000058, 0.00000583, -0.00000350, -0.00000519, 0.00000552, 0.00000512, -0.00000238, -0.00000909, 0.00000910, -0.00000840, 0.00000619, -0.00001447, 0.00000825, -0.00001988, 0.00001376, -0.00001384, 0.00000593, -0.00001678, 0.00000224, -0.00005360, 0.00000565, -0.00023556, 0.00022492, 0.00047829, -0.00021667, -0.00005949, -0.00002782, 0.00000176, -0.00002227, -0.00000868, -0.00000876, -0.00000561, -0.00000198, 0.00000945, -0.00000771, -0.00000114, -0.00000003, -0.00000723, 0.00000078, 0.00000873, -0.00000365, -0.00001049, 0.00000207, -0.00000267, 0.00000206, -0.00000503, 0.00000399, 0.00001035, -0.00000591, -0.00000591, 0.00000465, 0.00000772, -0.00000210, -0.00000154, -0.00000333, -0.00000161, -0.00000471, 0.00000466, -0.00000396, 0.00001253, -0.00000706, -0.00000929, 0.00000438, 0.00000830, -0.00000813, -0.00000163, -0.00000611, 0.00000409, -0.00002778, 0.00000672, -0.00000429, 0.00002184, 0.00001501, -0.00000043, 0.00002774, 0.00000119, 0.00000564, -0.00001286, -0.00000870, -0.00000544, 0.00000047, 0.00000040, -0.00000367, 0.00000399, -0.00000516, 0.00000152, -0.00000494, 0.00000500, -0.00000537, 0.00001289, -0.00000869, 0.00000059, -0.00000473, 0.00001667, -0.00001746, 0.00002712, -0.00002060, 0.00000151, },
		{ -0.00002315, 0.00000178, -0.00000113, 0.00000922, -0.00001502, 0.00000084, -0.00000173, -0.00001304, 0.00002031, -0.00001730, 0.00002071, 0.00000443, -0.00001187, -0.00000464, 0.00000955, 0.00000072, -0.00000375, -0.00000001, 0.00000514, 0.00000137, -0.00000794, -0.00000455, 0.00000887, -0.00000598, -0.00000228, -0.00001003, 0.00001152, 0.00000782, -0.00001349, -0.00000712, 0.00000303, 0.00000419, -0.00000838, 0.00000171, 0.00000448, 0.00000592, -0.00001123, 0.00000282, -0.00000058, 0.00000002, -0.00000445, -0.00000406, -0.00000276, -0.00000926, 0.00000730, -0.00000962, -0.00000090, -0.00000750, -0.00002288, -0.00002420, -0.00007424, -0.00017066, 0.00047911, 0.00018175, -0.00022221, -0.00000760, -0.00003176, 0.00000662, -0.00002947, 0.00000019, -0.00001343, -0.00000122, 0.00000050, 0.00000780, -0.00001146, 0.00000470, -0.00000602, -0.00000425, 0.00000445, 0.00000690, -0.00001030, -0.00000477, 0.00000175, -0.00000309, 0.00000220, -0.00000636, 0.00001257, 0.00000132, -0.00000579, -0.00000467, 0.00001087, 0.00000112, -0.00000032, -0.00000459, -0.00000013, -0.00000597, 0.00000258, -0.00000205, 0.00000383, 0.00000743, -0.00001135, -0.00000528, 0.00001089, -0.00000109, -0.00000418, -0.00000656, 0.00000158, -0.00000879, -0.00002123, 0.00000995, -0.00000311, 0.00003289, -0.00000168, 0.00001683, 0.00001906, 0.00000058, 0.00000156, -0.00001850, -0.00000147, -0.00000914, 0.00000711, -0.00000733, 0.00000475, -0.00000421, 0.00000164, -0.00000560, 0.00000336, -0.00000348, 0.00000524, 0.00000382, -0.00000520, -0.00000430, 0.00000812, 0.00000182, -0.00000432, 0.00002088, -0.00003031, 0.00001957, },
		{ -0.00002489, 0.00001031, -0.00000384, 0.00000899, -0.00002261, 0.00001479, -0.00001980, 0.00000877, 0.00000396, -0.00000782, 0.00002409, -0.00000781, -0.00000955, 0.00000114, 0.00000895, -0.00000296, -0.00000239, 0.00000139, 0.00000553, -0.00000191, -0.00000992, 0.00000276, 0.00000438, -0.00000658, -0.00000447, -0.00000555, 0.00001759, -0.00000443, -0.00001232, -0.00000352, 0.00000675, -0.00000270, -0.00000573, 0.00000449, 0.00000588, -0.00000030, -0.00000998, 0.00000691, -0.00000522, 0.00000381, -0.00001110, 0.00000401, -0.00001316, 0.00000380, -0.00000387, -0.00000171, -0.00000884, -0.00000242, -0.00004220, -0.00000259, -0.00017914, 0.00006781, 0.00049953, -0.00007298, -0.00013732, -0.00000580, -0.00001765, -0.00000547, -0.00002190, -0.00000130, -0.00001165, 0.00000011, 0.00000508, 0.00000077, -0.00000836, 0.00000530, -0.00001026, 0.00000121, 0.00000579, 0.00000244, -0.00001376, 0.00000289, -0.00000357, 0.00000146, -0.00000329, 0.00000118, 0.00001175, -0.00000429, -0.00000590, 0.00000084, 0.00001097, -0.00000279, -0.00000008, -0.00000494, -0.00000051, -0.00000502, 0.00000453, -0.00000512, 0.00001077, -0.00000157, -0.00001163, 0.00000209, 0.00000982, -0.00000636, -0.00000210, -0.00000869, 0.00000617, -0.00002300, -0.00000623, 0.00000384, 0.00001027, 0.00002820, -0.00000664, 0.00003160, 0.00000366, 0.00000797, -0.00001360, -0.00000961, -0.00000546, -0.00000301, 0.00000359, -0.00000562, 0.00000475, -0.00000555, 0.00000249, -0.00000675, 0.00000604, -0.00000672, 0.00001235, -0.00000548, -0.00000109, -0.00000570, 0.00001687, -0.00001439, 0.00001887, -0.00000685, -0.00001157, 0.00001033, },
		{ -0.00000831, 0.00000423, 0.00000382, -0.00000475, -0.00001237, 0.00001105, -0.00002290, 0.00002247, -0.00001351, 0.00001127, 0.00001359, -0.00001276, -0.00000566, 0.00000613, 0.00000513, -0.00000408, -0.00000135, 0.00000315, 0.00000483, -0.00000659, -0.00000721, 0.00000699, -0.00000121, -0.00000588, -0.00000671, 0.00000370, 0.00001472, -0.00001406, -0.00000726, -0.00000049, 0.00000743, -0.00000965, 0.00000177, 0.00000326, 0.00000801, -0.00000998, 0.00000020, 0.00000041, 0.00000023, -0.00000354, -0.00000504, -0.00000136, -0.00000949, 0.00000394, -0.00000427, -0.00000481, -0.00000328, -0.00002028, -0.00002558, -0.00004994, -0.00017802, 0.00032322, 0.00032069, -0.00018595, -0.00005657, -0.00001546, -0.00000581, -0.00001740, -0.00001152, -0.00000590, -0.00000765, 0.00000168, 0.00000722, -0.00000658, -0.00000125, -0.00000007, -0.00000901, 0.00000535, 0.00000565, -0.00000468, -0.00001023, 0.00000531, -0.00000673, 0.00000470, -0.00000767, 0.00001151, 0.00000318, -0.00000451, -0.00000657, 0.00000888, 0.00000486, -0.00000224, -0.00000218, -0.00000314, -0.00000290, -0.00000017, 0.00000044, -0.00000113, 0.00001122, -0.00000976, -0.00000731, 0.00000843, 0.00000325, -0.00000587, -0.00000463, -0.00000455, 0.00000130, -0.00002847, 0.00000688, -0.00000267, 0.00002836, 0.00000967, 0.00000766, 0.00002857, -0.00000135, 0.00000672, -0.00002330, 0.00000112, -0.00001279, 0.00000715, -0.00000587, 0.00000312, -0.00000312, 0.00000099, -0.00000450, 0.00000080, -0.00000086, 0.00000107, 0.00000765, -0.00000587, -0.00000376, 0.00000414, 0.00000829, -0.00001110, 0.00002468, -0.00002684, 0.00001210, -0.00001310, },
		{ 0.00000686, -0.00000320, 0.00001019, -0.00001973, 0.00000623, -0.00000682, -0.00000736, 0.00001714, -0.00001666, 0.00002495, -0.00000259, -0.00001011, -0.00000238, 0.00000947, -0.00000029, -0.00000259, -0.00000152, 0.00000633, 0.00000065, -0.00000875, -0.00000256, 0.00000778, -0.00000648, -0.00000442, -0.00000705, 0.00001404, 0.00000250, -0.00001480, -0.00000409, 0.00000459, 0.00000089, -0.00000847, 0.00000532, 0.00000468, 0.00000315, -0.00001130, 0.00000561, -0.00000399, 0.00000368, -0.00000982, 0.00000192, -0.00000931, -0.00000116, -0.00000105, -0.00000172, -0.00000860, -0.00000017, -0.00003785, -0.00000468, -0.00013814, -0.00002465, 0.00043632, 0.00008733, -0.00017972, -0.00001217, -0.00002046, -0.00000221, -0.00002291, -0.00000371, -0.00001031, -0.00000255, 0.00000332, 0.00000567, -0.00001042, 0.00000534, -0.00000854, -0.00000166, 0.00000518, 0.00000511, -0.00001258, -0.00000030, -0.00000083, -0.00000219, 0.00000036, -0.00000263, 0.00001365, -0.00000361, -0.00000448, -0.00000326, 0.00001278, -0.00000161, -0.00000047, -0.00000422, -0.00000201, -0.00000316, 0.00000276, -0.00000388, 0.00000616, 0.00000491, -0.00001355, 0.00000045, 0.00000919, -0.00000261, -0.00000403, -0.00000809, 0.00000223, -0.00001178, -0.00002106, 0.00000994, 0.00000121, 0.00003614, -0.00000626, 0.00002956, 0.00001073, 0.00000699, -0.00001003, -0.00001495, -0.00000150, -0.00000917, 0.00000753, -0.00000768, 0.00000583, -0.00000658, 0.00000399, -0.00000836, 0.00000628, -0.00000680, 0.00001020, -0.00000160, -0.00000250, -0.00000636, 0.00001538, -0.00000893, 0.00000914, 0.00000545, -0.00002026, 0.00001528, -0.00002533, },
		{ 0.00000802, -0.00000110, 0.00000421, -0.00002098, 0.00001552, -0.00002183, 0.00001382, -0.00000159, -0.00000225, 0.00002258, -0.00001286, -0.00000537, 0.00000134, 0.00000907, -0.00000385, -0.00000131, -0.00000047, 0.00000809, -0.00000514, -0.00000751, 0.00000205, 0.00000497, -0.00000948, -0.00000355, -0.00000280, 0.00001828, -0.00001161, -0.00000867, -0.00000296, 0.00000883, -0.00000886, 0.00000011, 0.00000327, 0.00000872, -0.00000768, -0.00000212, 0.00000049, 0.00000092, -0.00000301, -0.00000512, -0.00000171, -0.00000723, -0.00000113, 0.00000091, -0.00000766, -0.00000057, -0.00001742, -0.00002554, -0.00003497, -0.00016582, 0.00019144, 0.00037300, -0.00008391, -0.00011922, 0.00000079, -0.00001940, -0.00000495, -0.00002175, -0.00000054, -0.00001210, 0.00000222, 0.00000472, 0.00000041, -0.00000776, 0.00000536, -0.00001254, 0.00000566, 0.00000360, 0.00000127, -0.00001452, 0.00000715, -0.00000818, 0.00000469, -0.00000683, 0.00000889, 0.00000588, -0.00000420, -0.00000642, 0.00000489, 0.00000916, -0.00000385, -0.00000037, -0.00000487, -0.00000190, -0.00000079, 0.00000126, -0.00000361, 0.00001102, -0.00000497, -0.00001055, 0.00000700, 0.00000525, -0.00000492, -0.00000439, -0.00000807, 0.00000496, -0.00002528, -0.00000538, 0.00000378, 0.00001786, 0.00002443, -0.00000206, 0.00003715, -0.00000204, 0.00001161, -0.00002668, 0.00000120, -0.00001313, 0.00000394, -0.00000259, 0.00000070, -0.00000148, -0.00000049, -0.00000237, -0.00000235, 0.00000175, -0.00000233, 0.00000959, -0.00000475, -0.00000395, 0.00000081, 0.00001265, -0.00001363, 0.00002264, -0.00001826, 0.00000174, -0.00000234, -0.00001701, },
		{ 0.00000063, 0.00000671, -0.00001107, -0.00000664, 0.00000686, -0.00001950, 0.00002266, -0.00001642, 0.00001782, 0.00000673, -0.00001230, -0.00000314, 0.00000643, 0.00000423, -0.00000344, -0.00000236, 0.00000398, 0.00000475, -0.00000807, -0.00000532, 0.00000649, -0.00000182, -0.00000804, -0.00000460, 0.00000691, 0.00001067, -0.00001729, -0.00000362, 0.00000075, 0.00000557, -0.00001163, 0.00000641, 0.00000285, 0.00000691, -0.00001261, 0.00000527, -0.00000474, 0.00000557, -0.00001048, 0.00000220, -0.00000894, -0.00000081, -0.00000449, 0.00000386, -0.00001338, 0.00000627, -0.00003823, -0.00000180, -0.00011113, -0.00007044, 0.00033708, 0.00021089, -0.00016013, -0.00005192, -0.00000670, -0.00001200, -0.00001314, -0.00001376, -0.00000386, -0.00000841, 0.00000406, 0.00000610, -0.00000681, 0.00000041, -0.00000290, -0.00000703, 0.00000667, 0.00000433, -0.00000749, -0.00000634, 0.00000395, -0.00000699, 0.00000417, -0.00000608, 0.00001499, -0.00000281, -0.00000306, -0.00000607, 0.00001212, 0.00000168, -0.00000240, -0.00000201, -0.00000462, -0.00000102, 0.00000105, -0.00000205, 0.00000127, 0.00000965, -0.00001312, -0.00000197, 0.00000820, 0.00000090, -0.00000531, -0.00000658, -0.00000366, -0.00000101, -0.00002972, 0.00000764, -0.00000073, 0.00003524, 0.00000270, 0.00002082, 0.00002262, 0.00000260, -0.00000228, -0.00002369, 0.00000449, -0.00001570, 0.00001046, -0.00000896, 0.00000675, -0.00000772, 0.00000531, -0.00000915, 0.00000552, -0.00000589, 0.00000733, 0.00000170, -0.00000269, -0.00000731, 0.00001314, -0.00000294, 0.00000077, 0.00001336, -0.00002266, 0.00001377, -0.00001999, -0.00000054, },
		{ -0.00000271, 0.00000827, -0.00002163, 0.00001021, -0.00001143, -0.00000105, 0.00001218, -0.00001350, 0.00002661, -0.00000935, -0.00000609, -0.00000211, 0.00001018, -0.00000162, -0.00000146, -0.00000332, 0.00000896, -0.00000227, -0.00000716, -0.00000286, 0.00000870, -0.00000951, -0.00000377, -0.00000504, 0.00001648, -0.00000512, -0.00001224, -0.00000308, 0.00000703, -0.00000495, -0.00000398, 0.00000515, 0.00000734, -0.00000359, -0.00000532, 0.00000162, -0.00000028, -0.00000043, -0.00000695, -0.00000049, -0.00000696, -0.00000247, 0.00000042, -0.00000428, -0.00000433, -0.00000892, -0.00003074, -0.00001873, -0.00015389, 0.00010107, 0.00035241, 0.00004262, -0.00015860, -0.00000438, -0.00002057, -0.00000269, -0.00002238, -0.00000312, -0.00001061, -0.00000085, 0.00000363, 0.00000522, -0.00001045, 0.00000653, -0.00001195, 0.00000307, 0.00000329, 0.00000456, -0.00001486, 0.00000509, -0.00000591, 0.00000140, -0.00000359, 0.00000458, 0.00000957, -0.00000496, -0.00000471, 0.00000017, 0.00001247, -0.00000375, -0.00000034, -0.00000436, -0.00000336, 0.00000061, 0.00000048, -0.00000358, 0.00000755, 0.00000153, -0.00001387, 0.00000619, 0.00000523, -0.00000190, -0.00000570, -0.00000834, 0.00000220, -0.00001466, -0.00002065, 0.00000949, 0.00000779, 0.00003605, -0.00000645, 0.00003993, 0.00000262, 0.00001246, -0.00002493, -0.00000420, -0.00000845, -0.00000306, 0.00000258, -0.00000288, 0.00000136, -0.00000338, 0.00000102, -0.00000600, 0.00000422, -0.00000498, 0.00001014, -0.00000298, -0.00000371, -0.00000256, 0.00001545, -0.00001308, 0.00001739, -0.00000850, -0.00000717, 0.00000510, -0.00002109, 0.00000772, },
		{ 0.00000200, -0.00000164, -0.00001712, 0.00001419, -0.00002202, 0.00001769, -0.00000723, 0.00000505, 0.00001724, -0.00001432, -0.00000235, 0.00000160, 0.00000873, -0.00000389, -0.00000187, -0.00000000, 0.00000868, -0.00000712, -0.00000586, 0.00000187, 0.00000476, -0.00001213, -0.00000169, 0.00000012, 0.00001635, -0.00001699, -0.00000443, -0.00000223, 0.00000856, -0.00001256, 0.00000551, 0.00000238, 0.00000919, -0.00001242, 0.00000445, -0.00000555, 0.00000710, -0.00001061, 0.00000238, -0.00000924, 0.00000100, -0.00000884, 0.00000866, -0.00001636, 0.00000991, -0.00003603, -0.00000177, -0.00008742, -0.00009494, 0.00024345, 0.00026973, -0.00008278, -0.00010700, 0.00000893, -0.00002402, -0.00000193, -0.00002332, 0.00000119, -0.00001278, 0.00000438, 0.00000441, -0.00000056, -0.00000589, 0.00000308, -0.00001181, 0.00000790, 0.00000233, -0.00000130, -0.00001206, 0.00000767, -0.00001011, 0.00000585, -0.00000750, 0.00001445, -0.00000068, -0.00000286, -0.00000669, 0.00000899, 0.00000623, -0.00000457, 0.00000005, -0.00000636, -0.00000050, 0.00000097, -0.00000128, -0.00000178, 0.00001083, -0.00000940, -0.00000591, 0.00000819, 0.00000243, -0.00000431, -0.00000632, -0.00000763, 0.00000401, -0.00002811, -0.00000388, 0.00000349, 0.00002713, 0.00001733, 0.00000900, 0.00003561, -0.00000168, 0.00000609, -0.00003173, 0.00000831, -0.00001902, 0.00000996, -0.00000798, 0.00000627, -0.00000765, 0.00000512, -0.00000815, 0.00000327, -0.00000395, 0.00000408, 0.00000415, -0.00000191, -0.00000807, 0.00001012, 0.00000282, -0.00000516, 0.00001660, -0.00001998, 0.00000829, -0.00001216, -0.00000850, 0.00000460, },
		{ 0.00000729, -0.00001552, -0.00000112, 0.00000196, -0.00001447, 0.00002055, -0.00001702, 0.00002280, -0.00000122, -0.00000879, -0.00000279, 0.00000762, 0.00000277, -0.00000244, -0.00000417, 0.00000676, 0.00000211, -0.00000701, -0.00000573, 0.00000805, -0.00000499, -0.00000765, -0.00000328, 0.00001050, 0.00000358, -0.00001650, -0.00000180, 0.00000307, 0.00000069, -0.00000893, 0.00000751, 0.00000502, 0.00000135, -0.00000895, 0.00000384, -0.00000341, 0.00000407, -0.00001047, 0.00000265, -0.00000946, 0.00000062, -0.00000549, 0.00000404, -0.00001264, 0.00000311, -0.00003925, -0.00000180, -0.00014441, 0.00004402, 0.00029448, 0.00014778, -0.00014824, -0.00003869, -0.00000693, -0.00001207, -0.00001315, -0.00001281, -0.00000435, -0.00000676, 0.00000446, 0.00000631, -0.00000805, 0.00000294, -0.00000731, -0.00000222, 0.00000500, 0.00000455, -0.00001107, -0.00000039, -0.00000093, -0.00000376, 0.00000065, -0.00000022, 0.00001321, -0.00000572, -0.00000282, -0.00000364, 0.00001347, -0.00000134, -0.00000215, -0.00000208, -0.00000618, 0.00000254, -0.00000051, -0.00000251, 0.00000302, 0.00000718, -0.00001531, 0.00000458, 0.00000482, 0.00000115, -0.00000670, -0.00000724, -0.00000340, -0.00000318, -0.00003087, 0.00000799, 0.00000323, 0.00003989, -0.00000209, 0.00003471, 0.00001393, 0.00000836, -0.00001750, -0.00001464, -0.00000021, -0.00001183, 0.00000799, -0.00000666, 0.00000480, -0.00000701, 0.00000477, -0.00000929, 0.00000599, -0.00000672, 0.00000977, -0.00000142, -0.00000243, -0.00000617, 0.00001700, -0.00001068, 0.00001134, -0.00000052, -0.00001215, 0.00000768, -0.00002009, 0.00000347, -0.00000095, },
		{ 0.00000435, -0.00002082, 0.00001222, -0.00001480, 0.00000510, 0.00000595, -0.00000767, 0.00002430, -0.00001328, -0.00000230, -0.00000232, 0.00001076, -0.00000242, -0.00000158, -0.00000337, 0.00001040, -0.00000483, -0.00000549, -0.00000336, 0.00000961, -0.00001318, -0.00000160, -0.00000316, 0.00001708, -0.00001243, -0.00000777, -0.00000274, 0.00000819, -0.00000993, 0.00000168, 0.00000404, 0.00000913, -0.00000988, 0.00000197, -0.00000460, 0.00000611, -0.00000799, 0.00000010, -0.00000735, 0.00000037, -0.00000922, 0.00000760, -0.00001273, 0.00000617, -0.00002657, -0.00000972, -0.00005973, -0.00011455, 0.00017477, 0.00026801, 0.00002174, -0.00014666, 0.00000759, -0.00002598, 0.00000099, -0.00002476, -0.00000062, -0.00001195, 0.00000165, 0.00000343, 0.00000451, -0.00000951, 0.00000581, -0.00001315, 0.00000663, 0.00000150, 0.00000310, -0.00001461, 0.00000791, -0.00000971, 0.00000429, -0.00000625, 0.00001160, 0.00000310, -0.00000409, -0.00000543, 0.00000451, 0.00001045, -0.00000542, 0.00000050, -0.00000595, -0.00000239, 0.00000280, -0.00000193, -0.00000247, 0.00000851, -0.00000324, -0.00001066, 0.00000903, 0.00000194, -0.00000117, -0.00000763, -0.00000839, 0.00000232, -0.00001814, -0.00001942, 0.00000818, 0.00001719, 0.00003162, 0.00000008, 0.00004425, -0.00000106, 0.00001068, -0.00003473, 0.00000625, -0.00001664, 0.00000482, -0.00000407, 0.00000360, -0.00000538, 0.00000254, -0.00000488, -0.00000054, -0.00000104, 0.00000051, 0.00000607, -0.00000106, -0.00000767, 0.00000588, 0.00000836, -0.00000895, 0.00001662, -0.00001501, 0.00000235, -0.00000545, -0.00001382, 0.00000600, -0.00000080, },
		{ -0.00000691, -0.00001222, 0.00001119, -0.00002008, 0.00001939, -0.00001139, 0.00001217, 0.00000951, -0.00001216, -0.00000147, 0.00000271, 0.00000770, -0.00000334, -0.00000356, 0.00000273, 0.00000655, -0.00000664, -0.00000636, 0.00000367, 0.00000213, -0.00001265, -0.00000035, 0.00000373, 0.00001106, -0.00001876, -0.00000136, -0.00000062, 0.00000532, -0.00001208, 0.00000810, 0.00000367, 0.00000538, -0.00001150, 0.00000572, -0.00000666, 0.00000825, -0.00001338, 0.00000547, -0.00001235, 0.00000487, -0.00001218, 0.00001167, -0.00001931, 0.00001220, -0.00004425, 0.00000894, -0.00013017, 0.00000220, 0.00023443, 0.00020297, -0.00008627, -0.00008999, 0.00001062, -0.00002489, -0.00000135, -0.00002304, 0.00000134, -0.00001175, 0.00000511, 0.00000510, -0.00000256, -0.00000279, -0.00000137, -0.00000778, 0.00000700, 0.00000279, -0.00000525, -0.00000672, 0.00000407, -0.00000823, 0.00000375, -0.00000380, 0.00001535, -0.00000517, -0.00000206, -0.00000546, 0.00001186, 0.00000282, -0.00000470, 0.00000043, -0.00000842, 0.00000305, -0.00000012, -0.00000209, -0.00000029, 0.00000982, -0.00001348, 0.00000104, 0.00000562, 0.00000227, -0.00000554, -0.00000711, -0.00000772, 0.00000328, -0.00003110, -0.00000250, 0.00000443, 0.00003592, 0.00000982, 0.00002374, 0.00002903, 0.00000235, -0.00000757, -0.00002632, 0.00000742, -0.00001865, 0.00001092, -0.00000879, 0.00000726, -0.00000970, 0.00000720, -0.00001084, 0.00000617, -0.00000708, 0.00000844, -0.00000015, -0.00000050, -0.00000929, 0.00001693, -0.00000691, 0.00000562, 0.00000481, -0.00001339, 0.00000669, -0.00001609, -0.00000234, 0.00000203, 0.00000385, },
		{ -0.00001753, 0.00000321, -0.00000277, -0.00000840, 0.00001616, -0.00001449, 0.00002424, -0.00000748, -0.00000444, -0.00000350, 0.00000903, 0.00000147, -0.00000227, -0.00000482, 0.00000914, -0.00000103, -0.00000511, -0.00000702, 0.00001037, -0.00000964, -0.00000521, -0.00000243, 0.00001318, -0.00000466, -0.00001271, -0.00000128, 0.00000500, -0.00000467, -0.00000382, 0.00000674, 0.00000752, -0.00000558, -0.00000196, -0.00000167, 0.00000233, -0.00000256, -0.00000475, -0.00000275, -0.00000373, -0.00000460, 0.00000046, -0.00000316, -0.00000363, -0.00001153, -0.00002360, -0.00002958, -0.00013151, 0.00012805, 0.00023211, 0.00011407, -0.00014523, -0.00002023, -0.00001379, -0.00000722, -0.00001658, -0.00000952, -0.00000631, -0.00000390, 0.00000406, 0.00000649, -0.00000864, 0.00000413, -0.00001046, 0.00000255, 0.00000258, 0.00000453, -0.00001309, 0.00000440, -0.00000608, 0.00000024, -0.00000318, 0.00000725, 0.00000771, -0.00000583, -0.00000353, 0.00000028, 0.00001287, -0.00000405, -0.00000105, -0.00000351, -0.00000591, 0.00000533, -0.00000294, -0.00000184, 0.00000451, 0.00000306, -0.00001431, 0.00000932, 0.00000097, 0.00000222, -0.00000887, -0.00000745, -0.00000303, -0.00000602, -0.00003111, 0.00000701, 0.00001055, 0.00004025, -0.00000126, 0.00004485, 0.00000715, 0.00000944, -0.00003106, -0.00000231, -0.00000913, -0.00000372, 0.00000163, -0.00000067, -0.00000124, -0.00000197, -0.00000008, -0.00000513, 0.00000223, -0.00000304, 0.00000764, -0.00000083, -0.00000563, 0.00000046, 0.00001347, -0.00001103, 0.00001503, -0.00001016, -0.00000170, -0.00000172, -0.00001537, 0.00000417, -0.00000072, 0.00000564, },
		{ -0.00001807, 0.00001219, -0.00001622, 0.00000988, -0.00000010, -0.00000054, 0.00001854, -0.00001342, -0.00000034, -0.00000190, 0.00001068, -0.00000257, -0.00000299, -0.00000103, 0.00000920, -0.00000520, -0.00000566, -0.00000214, 0.00000848, -0.00001533, 0.00000051, -0.00000055, 0.00001451, -0.00001707, -0.00000340, -0.00000209, 0.00000687, -0.00001165, 0.00000563, 0.00000438, 0.00000731, -0.00001176, 0.00000578, -0.00000796, 0.00000974, -0.00001325, 0.00000545, -0.00001263, 0.00000632, -0.00001472, 0.00001338, -0.00001918, 0.00001326, -0.00004047, 0.00000820, -0.00010549, -0.00003747, 0.00019118, 0.00020657, 0.00000466, -0.00013251, 0.00001532, -0.00002994, 0.00000395, -0.00002653, 0.00000118, -0.00001224, 0.00000323, 0.00000399, 0.00000270, -0.00000702, 0.00000251, -0.00001085, 0.00000719, 0.00000140, -0.00000009, -0.00001107, 0.00000656, -0.00000995, 0.00000423, -0.00000511, 0.00001515, -0.00000268, -0.00000287, -0.00000530, 0.00000845, 0.00000731, -0.00000639, 0.00000156, -0.00000855, 0.00000104, 0.00000225, -0.00000301, -0.00000141, 0.00000893, -0.00000858, -0.00000420, 0.00000787, 0.00000107, -0.00000200, -0.00000865, -0.00000891, 0.00000286, -0.00002221, -0.00001777, 0.00000693, 0.00002809, 0.00002441, 0.00001244, 0.00004245, -0.00000045, 0.00000008, -0.00003442, 0.00001011, -0.00002029, 0.00000937, -0.00000791, 0.00000733, -0.00000993, 0.00000689, -0.00000965, 0.00000414, -0.00000571, 0.00000596, 0.00000130, 0.00000105, -0.00001071, 0.00001454, -0.00000173, 0.00000033, 0.00000819, -0.00001274, 0.00000469, -0.00001201, -0.00000695, 0.00000374, 0.00000098, -0.00000014, },
		{ -0.00000742, 0.00000724, -0.00001618, 0.00001839, -0.00001270, 0.00001682, 0.00000201, -0.00000808, -0.00000218, 0.00000440, 0.00000648, -0.00000322, -0.00000469, 0.00000587, 0.00000311, -0.00000462, -0.00000837, 0.00000699, -0.00000272, -0.00001067, 0.00000018, 0.00000736, 0.00000339, -0.00001685, 0.00000019, 0.00000114, 0.00000069, -0.00000847, 0.00000836, 0.00000612, -0.00000117, -0.00000592, 0.00000192, -0.00000256, 0.00000364, -0.00001003, 0.00000250, -0.00000912, 0.00000213, -0.00000863, 0.00000726, -0.00001373, 0.00000280, -0.00003633, -0.00000459, -0.00013916, 0.00008929, 0.00019220, 0.00016595, -0.00009645, -0.00006711, 0.00000486, -0.00002064, -0.00000430, -0.00002029, -0.00000018, -0.00000925, 0.00000484, 0.00000602, -0.00000439, -0.00000024, -0.00000570, -0.00000256, 0.00000443, 0.00000379, -0.00000879, -0.00000102, -0.00000131, -0.00000424, -0.00000009, 0.00000292, 0.00001173, -0.00000676, -0.00000232, -0.00000251, 0.00001291, -0.00000068, -0.00000378, -0.00000040, -0.00000921, 0.00000665, -0.00000270, -0.00000159, 0.00000109, 0.00000715, -0.00001508, 0.00000753, 0.00000139, 0.00000368, -0.00000801, -0.00000720, -0.00000782, 0.00000202, -0.00003335, -0.00000244, 0.00000844, 0.00004162, 0.00000569, 0.00003783, 0.00002153, 0.00000434, -0.00002295, -0.00001439, -0.00000004, -0.00001228, 0.00000658, -0.00000475, 0.00000319, -0.00000672, 0.00000448, -0.00000894, 0.00000468, -0.00000576, 0.00000855, -0.00000109, -0.00000247, -0.00000515, 0.00001731, -0.00001130, 0.00001253, -0.00000624, -0.00000368, -0.00000040, -0.00001438, 0.00000061, 0.00000096, 0.00000383, -0.00001051, },
		{ 0.00000575, -0.00000634, -0.00000249, 0.00001059, -0.00000949, 0.00002176, -0.00001031, -0.00000126, -0.00000401, 0.00001001, 0.00000071, -0.00000339, -0.00000312, 0.00000918, -0.00000248, -0.00000452, -0.00000710, 0.00001136, -0.00001368, -0.00000214, -0.00000119, 0.00001341, -0.00001164, -0.00000771, -0.00000118, 0.00000519, -0.00000796, 0.00000081, 0.00000654, 0.00000741, -0.00000979, 0.00000372, -0.00000660, 0.00000771, -0.00000945, 0.00000191, -0.00000925, 0.00000340, -0.00001141, 0.00000804, -0.00001182, 0.00000638, -0.00002833, -0.00000295, -0.00007219, -0.00007645, 0.00016414, 0.00017856, 0.00009046, -0.00014146, -0.00000378, -0.00002158, -0.00000140, -0.00002063, -0.00000604, -0.00000790, -0.00000142, 0.00000413, 0.00000558, -0.00000757, 0.00000268, -0.00001036, 0.00000487, 0.00000143, 0.00000286, -0.00001200, 0.00000564, -0.00000856, 0.00000220, -0.00000438, 0.00001289, 0.00000133, -0.00000466, -0.00000410, 0.00000462, 0.00001065, -0.00000613, 0.00000061, -0.00000633, -0.00000304, 0.00000569, -0.00000450, -0.00000103, 0.00000579, -0.00000244, -0.00000951, 0.00001027, -0.00000104, 0.00000224, -0.00001047, -0.00000803, -0.00000203, -0.00000983, -0.00003037, 0.00000495, 0.00002100, 0.00003636, 0.00000617, 0.00004931, 0.00000393, 0.00000227, -0.00003614, 0.00000600, -0.00001612, 0.00000354, -0.00000424, 0.00000484, -0.00000736, 0.00000368, -0.00000590, 0.00000026, -0.00000288, 0.00000246, 0.00000321, 0.00000140, -0.00000972, 0.00000968, 0.00000463, -0.00000460, 0.00001057, -0.00001192, 0.00000358, -0.00000955, -0.00000908, 0.00000315, 0.00000017, 0.00000271, -0.00001730, },
		{ 0.00001046, -0.00001537, 0.00001226, -0.00000423, 0.00000559, 0.00001151, -0.00001073, -0.00000030, -0.00000069, 0.00001016, -0.00000280, -0.00000440, 0.00000249, 0.00000594, -0.00000344, -0.00000788, 0.00000120, 0.00000464, -0.00001486, 0.00000167, 0.00000272, 0.00000881, -0.00001800, -0.00000022, -0.00000121, 0.00000376, -0.00001011, 0.00000729, 0.00000624, 0.00000199, -0.00000846, 0.00000445, -0.00000629, 0.00000796, -0.00001304, 0.00000559, -0.00001258, 0.00000681, -0.00001454, 0.00001295, -0.00001839, 0.00001045, -0.00004133, 0.00000828, -0.00013115, 0.00004690, 0.00016670, 0.00017159, -0.00001647, -0.00011176, 0.00001524, -0.00002910, 0.00000368, -0.00002603, 0.00000134, -0.00001100, 0.00000371, 0.00000505, 0.00000051, -0.00000422, -0.00000181, -0.00000627, 0.00000529, 0.00000258, -0.00000394, -0.00000584, 0.00000224, -0.00000712, 0.00000151, -0.00000016, 0.00001407, -0.00000600, -0.00000244, -0.00000357, 0.00001104, 0.00000352, -0.00000616, 0.00000170, -0.00001053, 0.00000533, -0.00000031, -0.00000267, -0.00000041, 0.00000793, -0.00001247, 0.00000325, 0.00000395, 0.00000237, -0.00000440, -0.00000876, -0.00000964, 0.00000324, -0.00002604, -0.00001693, 0.00000765, 0.00003797, 0.00001777, 0.00002727, 0.00003734, 0.00000019, -0.00001498, -0.00002522, 0.00000612, -0.00001727, 0.00000831, -0.00000677, 0.00000605, -0.00000973, 0.00000690, -0.00001045, 0.00000520, -0.00000681, 0.00000819, -0.00000110, 0.00000051, -0.00000935, 0.00001862, -0.00000921, 0.00000897, -0.00000270, -0.00000492, 0.00000041, -0.00001308, -0.00000246, 0.00000222, 0.00000157, -0.00000389, -0.00001443, },
		{ 0.00000340, -0.00001127, 0.00001521, -0.00001101, 0.00001777, -0.00000299, -0.00000421, -0.00000327, 0.00000599, 0.00000559, -0.00000426, -0.00000372, 0.00000722, 0.00000059, -0.00000312, -0.00000999, 0.00000997, -0.00000810, -0.00000722, 0.00000051, 0.00000957, -0.00000464, -0.00001247, 0.00000064, 0.00000193, -0.00000305, -0.00000419, 0.00000834, 0.00000706, -0.00000676, 0.00000048, -0.00000341, 0.00000332, -0.00000346, -0.00000356, -0.00000371, -0.00000236, -0.00000438, -0.00000111, -0.00000134, -0.00000369, -0.00001319, -0.00001788, -0.00003842, -0.00010741, 0.00014116, 0.00014566, 0.00014324, -0.00010699, -0.00004347, -0.00000447, -0.00001367, -0.00000912, -0.00001642, -0.00000209, -0.00000653, 0.00000468, 0.00000609, -0.00000480, 0.00000005, -0.00000751, 0.00000116, 0.00000236, 0.00000370, -0.00000997, 0.00000228, -0.00000536, -0.00000095, -0.00000279, 0.00000961, 0.00000574, -0.00000629, -0.00000303, 0.00000156, 0.00001209, -0.00000383, -0.00000189, -0.00000287, -0.00000748, 0.00000831, -0.00000496, -0.00000073, 0.00000256, 0.00000254, -0.00001282, 0.00001096, -0.00000193, 0.00000476, -0.00001038, -0.00000748, -0.00000720, -0.00000040, -0.00003432, -0.00000421, 0.00001634, 0.00004294, 0.00000735, 0.00004813, 0.00001580, -0.00000045, -0.00003240, -0.00000316, -0.00000862, -0.00000405, 0.00000030, 0.00000109, -0.00000310, -0.00000116, -0.00000101, -0.00000416, 0.00000035, -0.00000127, 0.00000526, 0.00000062, -0.00000671, 0.00000329, 0.00001108, -0.00000858, 0.00001213, -0.00001137, 0.00000348, -0.00000841, -0.00000945, 0.00000116, 0.00000088, 0.00000233, -0.00001209, -0.00000367, },
		{ -0.00000799, 0.00000191, 0.00000562, -0.00000398, 0.00001679, -0.00000984, 0.00000000, -0.00000375, 0.00001047, 0.00000010, -0.00000500, 0.00000040, 0.00000663, -0.00000148, -0.00000623, -0.00000482, 0.00000960, -0.00001542, 0.00000024, 0.00000091, 0.00001043, -0.00001561, -0.00000302, -0.00000123, 0.00000372, -0.00000836, 0.00000363, 0.00000780, 0.00000366, -0.00000909, 0.00000504, -0.00000751, 0.00000891, -0.00001236, 0.00000505, -0.00001223, 0.00000699, -0.00001455, 0.00001157, -0.00001570, 0.00000971, -0.00003684, 0.00000774, -0.00010783, -0.00000081, 0.00015681, 0.00014590, 0.00006535, -0.00013064, 0.00000551, -0.00002581, 0.00000217, -0.00002311, -0.00000376, -0.00000820, 0.00000005, 0.00000487, 0.00000384, -0.00000551, -0.00000065, -0.00000724, 0.00000429, 0.00000200, -0.00000022, -0.00000837, 0.00000328, -0.00000761, 0.00000122, -0.00000169, 0.00001448, -0.00000354, -0.00000366, -0.00000349, 0.00000832, 0.00000720, -0.00000697, 0.00000177, -0.00000925, 0.00000144, 0.00000365, -0.00000461, -0.00000029, 0.00000614, -0.00000770, -0.00000233, 0.00000765, -0.00000059, 0.00000054, -0.00001105, -0.00000905, -0.00000069, -0.00001407, -0.00002961, 0.00000330, 0.00003273, 0.00003048, 0.00001841, 0.00004926, 0.00000170, -0.00001108, -0.00003121, 0.00000654, -0.00001707, 0.00000607, -0.00000611, 0.00000642, -0.00000993, 0.00000630, -0.00000912, 0.00000344, -0.00000589, 0.00000627, -0.00000023, 0.00000218, -0.00001096, 0.00001671, -0.00000459, 0.00000418, 0.00000120, -0.00000676, 0.00000219, -0.00001291, -0.00000385, 0.00000196, 0.00000057, -0.00000028, -0.00001554, 0.00000641, },
		{ -0.00001274, 0.00001211, -0.00000558, 0.00000890, 0.00000570, -0.00000719, -0.00000125, 0.00000080, 0.00000972, -0.00000398, -0.00000415, 0.00000501, 0.00000265, -0.00000126, -0.00001069, 0.00000535, -0.00000089, -0.00001211, 0.00000204, 0.00000566, 0.00000133, -0.00001557, 0.00000153, -0.00000080, 0.00000067, -0.00000710, 0.00000807, 0.00000757, -0.00000399, -0.00000256, -0.00000002, -0.00000128, 0.00000217, -0.00000832, 0.00000122, -0.00000777, 0.00000229, -0.00000903, 0.00000674, -0.00001111, -0.00000140, -0.00002909, -0.00001300, -0.00012216, 0.00011046, 0.00012653, 0.00015351, -0.00003933, -0.00008694, 0.00000904, -0.00002409, 0.00000055, -0.00002368, 0.00000051, -0.00000900, 0.00000387, 0.00000566, -0.00000080, -0.00000294, -0.00000467, -0.00000195, 0.00000287, 0.00000353, -0.00000655, -0.00000158, -0.00000226, -0.00000358, -0.00000160, 0.00000640, 0.00000947, -0.00000682, -0.00000281, -0.00000031, 0.00001179, -0.00000038, -0.00000459, 0.00000011, -0.00001034, 0.00000840, -0.00000320, -0.00000170, 0.00000081, 0.00000484, -0.00001300, 0.00000886, -0.00000028, 0.00000420, -0.00000729, -0.00000875, -0.00000979, 0.00000260, -0.00002870, -0.00001801, 0.00001194, 0.00004467, 0.00001484, 0.00004115, 0.00003180, -0.00000421, -0.00002701, -0.00001330, -0.00000187, -0.00001005, 0.00000336, -0.00000212, 0.00000101, -0.00000561, 0.00000301, -0.00000742, 0.00000261, -0.00000411, 0.00000662, -0.00000045, -0.00000303, -0.00000288, 0.00001597, -0.00001051, 0.00001220, -0.00001029, 0.00000309, -0.00000697, -0.00000987, -0.00000050, 0.00000157, 0.00000083, -0.00000617, -0.00001098, 0.00000791, },
		{ -0.00000675, 0.00001136, -0.00000766, 0.00001558, -0.00000475, -0.00000186, -0.00000387, 0.00000719, 0.00000491, -0.00000613, -0.00000050, 0.00000577, 0.00000048, -0.00000377, -0.00000950, 0.00001074, -0.00001201, -0.00000375, 0.00000139, 0.00000917, -0.00001093, -0.00000713, 0.00000030, 0.00000149, -0.00000461, -0.00000097, 0.00000941, 0.00000469, -0.00000838, 0.00000396, -0.00000636, 0.00000683, -0.00000863, 0.00000156, -0.00000864, 0.00000321, -0.00000969, 0.00000506, -0.00000824, 0.00000359, -0.00002629, -0.00000154, -0.00007638, -0.00004655, 0.00015175, 0.00011281, 0.00012173, -0.00011106, -0.00002484, -0.00001260, -0.00000723, -0.00001362, -0.00001292, -0.00000325, -0.00000447, 0.00000509, 0.00000513, -0.00000387, -0.00000196, -0.00000607, 0.00000206, 0.00000196, 0.00000198, -0.00000841, 0.00000208, -0.00000633, -0.00000021, -0.00000224, 0.00001344, -0.00000004, -0.00000518, -0.00000319, 0.00000577, 0.00000954, -0.00000590, 0.00000004, -0.00000616, -0.00000356, 0.00000741, -0.00000588, 0.00000004, 0.00000363, -0.00000294, -0.00000718, 0.00001051, -0.00000292, 0.00000428, -0.00001180, -0.00000835, -0.00000580, -0.00000387, -0.00003445, -0.00000706, 0.00002736, 0.00004071, 0.00001472, 0.00005424, 0.00001081, -0.00001205, -0.00003201, 0.00000168, -0.00001296, 0.00000129, -0.00000377, 0.00000475, -0.00000766, 0.00000322, -0.00000572, 0.00000017, -0.00000361, 0.00000321, 0.00000147, 0.00000224, -0.00000988, 0.00001202, 0.00000179, -0.00000124, 0.00000538, -0.00000932, 0.00000485, -0.00001353, -0.00000411, 0.00000070, 0.00000082, 0.00000027, -0.00001195, -0.00000151, 0.00000073, },
		{ 0.00000407, 0.00000264, -0.00000007, 0.00001169, -0.00000760, -0.00000014, -0.00000297, 0.00001081, -0.00000123, -0.00000549, 0.00000382, 0.00000301, 0.00000091, -0.00000950, -0.00000071, 0.00000527, -0.00001442, 0.00000152, 0.00000345, 0.00000477, -0.00001599, 0.00000052, -0.00000172, 0.00000187, -0.00000698, 0.00000536, 0.00000919, -0.00000199, -0.00000461, 0.00000233, -0.00000437, 0.00000549, -0.00001039, 0.00000346, -0.00001032, 0.00000548, -0.00001225, 0.00000912, -0.00001275, 0.00000374, -0.00003263, -0.00000025, -0.00011809, 0.00006857, 0.00012476, 0.00013132, 0.00003703, -0.00011267, 0.00000702, -0.00002534, 0.00000261, -0.00002356, -0.00000271, -0.00000746, 0.00000093, 0.00000555, 0.00000237, -0.00000412, -0.00000366, -0.00000316, 0.00000219, 0.00000324, -0.00000319, -0.00000434, -0.00000059, -0.00000478, -0.00000135, 0.00000378, 0.00001197, -0.00000595, -0.00000346, -0.00000131, 0.00001056, 0.00000305, -0.00000622, 0.00000137, -0.00001059, 0.00000556, 0.00000058, -0.00000377, 0.00000064, 0.00000462, -0.00001057, 0.00000451, 0.00000345, 0.00000123, -0.00000221, -0.00001112, -0.00000983, 0.00000009, -0.00001769, -0.00003020, 0.00000396, 0.00004342, 0.00002564, 0.00003254, 0.00004699, -0.00000420, -0.00002402, -0.00002063, 0.00000098, -0.00001236, 0.00000364, -0.00000365, 0.00000356, -0.00000811, 0.00000476, -0.00000843, 0.00000302, -0.00000526, 0.00000656, -0.00000086, -0.00000009, -0.00000717, 0.00001794, -0.00000951, 0.00001009, -0.00000769, 0.00000108, -0.00000389, -0.00001157, -0.00000079, 0.00000124, -0.00000003, -0.00000236, -0.00001327, 0.00000570, -0.00000765, },
		{ 0.00001043, -0.00000474, 0.00000909, 0.00000254, -0.00000447, -0.00000216, 0.00000211, 0.00000947, -0.00000620, -0.00000148, 0.00000490, 0.00000127, -0.00000061, -0.00001214, 0.00000831, -0.00000617, -0.00000836, 0.00000232, 0.00000698, -0.00000598, -0.00001088, 0.00000188, -0.00000114, -0.00000098, -0.00000444, 0.00000954, 0.00000618, -0.00000728, 0.00000216, -0.00000396, 0.00000331, -0.00000381, -0.00000278, -0.00000404, -0.00000192, -0.00000337, -0.00000231, -0.00000074, -0.00000288, -0.00001522, -0.00001247, -0.00004627, -0.00008057, 0.00013893, 0.00009157, 0.00013989, -0.00005781, -0.00006347, 0.00000091, -0.00001756, -0.00000370, -0.00002071, -0.00000028, -0.00000714, 0.00000434, 0.00000536, -0.00000081, -0.00000390, -0.00000460, 0.00000012, 0.00000161, 0.00000317, -0.00000683, -0.00000011, -0.00000458, -0.00000165, -0.00000258, 0.00001163, 0.00000366, -0.00000620, -0.00000325, 0.00000379, 0.00001055, -0.00000355, -0.00000239, -0.00000286, -0.00000765, 0.00000904, -0.00000509, -0.00000072, 0.00000206, 0.00000010, -0.00000972, 0.00001090, -0.00000279, 0.00000497, -0.00000954, -0.00000929, -0.00000892, 0.00000058, -0.00003001, -0.00002112, 0.00002015, 0.00004771, 0.00001691, 0.00005253, 0.00002585, -0.00001528, -0.00003032, -0.00000543, -0.00000809, -0.00000337, -0.00000154, 0.00000239, -0.00000429, -0.00000076, -0.00000189, -0.00000318, -0.00000116, 0.00000005, 0.00000325, 0.00000135, -0.00000717, 0.00000606, 0.00000827, -0.00000594, 0.00000880, -0.00001148, 0.00000694, -0.00001336, -0.00000484, -0.00000014, 0.00000121, -0.00000067, -0.00000683, -0.00000856, 0.00000567, -0.00000952, },
		{ 0.00000848, -0.00000470, 0.00001221, -0.00000418, -0.00000105, -0.00000391, 0.00000819, 0.00000374, -0.00000748, 0.00000354, 0.00000235, 0.00000248, -0.00000662, -0.00000658, 0.00000860, -0.00001329, -0.00000124, 0.00000298, 0.00000587, -0.00001418, -0.00000207, -0.00000092, 0.00000073, -0.00000452, 0.00000128, 0.00001100, -0.00000033, -0.00000571, 0.00000335, -0.00000538, 0.00000603, -0.00000949, 0.00000271, -0.00000954, 0.00000460, -0.00001053, 0.00000628, -0.00000951, 0.00000322, -0.00002938, 0.00000143, -0.00009966, 0.00002114, 0.00013162, 0.00009747, 0.00009620, -0.00010620, -0.00001387, -0.00001690, -0.00000321, -0.00001662, -0.00001033, -0.00000348, -0.00000301, 0.00000568, 0.00000399, -0.00000293, -0.00000458, -0.00000273, 0.00000074, 0.00000286, -0.00000047, -0.00000548, -0.00000049, -0.00000473, -0.00000171, 0.00000162, 0.00001326, -0.00000394, -0.00000448, -0.00000204, 0.00000913, 0.00000568, -0.00000627, 0.00000079, -0.00000863, 0.00000090, 0.00000484, -0.00000561, 0.00000089, 0.00000330, -0.00000721, -0.00000038, 0.00000740, -0.00000194, 0.00000235, -0.00001242, -0.00000938, -0.00000431, -0.00000751, -0.00003495, -0.00000934, 0.00003965, 0.00003717, 0.00002598, 0.00005748, 0.00000262, -0.00002523, -0.00002389, -0.00000059, -0.00001128, 0.00000170, -0.00000369, 0.00000428, -0.00000828, 0.00000403, -0.00000723, 0.00000166, -0.00000472, 0.00000505, -0.00000029, 0.00000137, -0.00000890, 0.00001672, -0.00000570, 0.00000597, -0.00000346, -0.00000268, 0.00000069, -0.00001420, -0.00000017, 0.00000020, 0.00000008, -0.00000135, -0.00001086, -0.00000085, -0.00000023, -0.00000376, },
		{ 0.00000180, 0.00000123, 0.00000838, -0.00000537, -0.00000072, -0.00000213, 0.00001131, -0.00000368, -0.00000368, 0.00000518, 0.00000046, 0.00000271, -0.00001246, 0.00000348, -0.00000013, -0.00001146, 0.00000204, 0.00000542, -0.00000215, -0.00001318, 0.00000242, -0.00000277, 0.00000096, -0.00000549, 0.00000751, 0.00000849, -0.00000631, 0.00000040, -0.00000149, 0.00000006, 0.00000012, -0.00000590, -0.00000065, -0.00000573, 0.00000110, -0.00000683, 0.00000320, -0.00000614, -0.00000761, -0.00001988, -0.00002419, -0.00009711, 0.00011227, 0.00008876, 0.00012489, 0.00000982, -0.00009169, 0.00000359, -0.00002164, 0.00000080, -0.00002266, -0.00000215, -0.00000641, 0.00000182, 0.00000556, 0.00000192, -0.00000459, -0.00000440, -0.00000041, 0.00000043, 0.00000384, -0.00000462, -0.00000206, -0.00000343, -0.00000242, -0.00000308, 0.00000946, 0.00000702, -0.00000635, -0.00000376, 0.00000220, 0.00001080, -0.00000091, -0.00000427, -0.00000074, -0.00000949, 0.00000772, -0.00000203, -0.00000272, 0.00000184, 0.00000106, -0.00000978, 0.00000866, -0.00000003, 0.00000275, -0.00000484, -0.00001152, -0.00000964, -0.00000044, -0.00001996, -0.00003298, 0.00000814, 0.00005171, 0.00002388, 0.00004665, 0.00004295, -0.00001647, -0.00002983, -0.00001136, -0.00000545, -0.00000589, -0.00000072, 0.00000046, -0.00000124, -0.00000402, 0.00000081, -0.00000531, 0.00000032, -0.00000216, 0.00000422, 0.00000051, -0.00000424, 0.00000062, 0.00001312, -0.00000847, 0.00001021, -0.00001181, 0.00000687, -0.00001095, -0.00000724, 0.00000038, 0.00000083, -0.00000145, -0.00000307, -0.00001147, 0.00000463, -0.00000613, 0.00000439, },
		{ -0.00000340, 0.00000746, 0.00000184, -0.00000311, -0.00000265, 0.00000326, 0.00000893, -0.00000847, 0.00000278, 0.00000222, 0.00000235, -0.00000235, -0.00001113, 0.00000867, -0.00000952, -0.00000492, 0.00000304, 0.00000586, -0.00001141, -0.00000523, 0.00000069, -0.00000152, -0.00000133, -0.00000245, 0.00001170, 0.00000177, -0.00000639, 0.00000341, -0.00000480, 0.00000464, -0.00000678, 0.00000024, -0.00000678, 0.00000134, -0.00000625, 0.00000136, -0.00000476, 0.00000051, -0.00002306, -0.00000284, -0.00007514, -0.00002204, 0.00013446, 0.00007135, 0.00012294, -0.00006774, -0.00004547, -0.00000564, -0.00001184, -0.00000761, -0.00001793, -0.00000057, -0.00000565, 0.00000501, 0.00000462, -0.00000038, -0.00000608, -0.00000206, -0.00000029, 0.00000185, 0.00000171, -0.00000532, -0.00000127, -0.00000425, -0.00000205, -0.00000037, 0.00001355, -0.00000125, -0.00000535, -0.00000288, 0.00000779, 0.00000748, -0.00000513, -0.00000080, -0.00000580, -0.00000365, 0.00000760, -0.00000571, 0.00000025, 0.00000246, -0.00000461, -0.00000407, 0.00000967, -0.00000313, 0.00000420, -0.00001086, -0.00001032, -0.00000736, -0.00000235, -0.00003071, -0.00002530, 0.00003136, 0.00004820, 0.00002344, 0.00006181, 0.00001626, -0.00002927, -0.00002474, -0.00000458, -0.00000902, -0.00000067, -0.00000332, 0.00000381, -0.00000680, 0.00000174, -0.00000486, -0.00000055, -0.00000329, 0.00000279, 0.00000084, 0.00000163, -0.00000854, 0.00001319, -0.00000026, 0.00000096, 0.00000137, -0.00000700, 0.00000522, -0.00001614, -0.00000019, -0.00000036, 0.00000041, -0.00000207, -0.00000634, -0.00000750, 0.00000463, -0.00000704, 0.00000882, },
		{ -0.00000363, 0.00000971, -0.00000293, -0.00000102, -0.00000375, 0.00000932, 0.00000147, -0.00000687, 0.00000633, -0.00000106, 0.00000498, -0.00001029, -0.00000239, 0.00000434, -0.00001200, 0.00000015, 0.00000474, 0.00000039, -0.00001399, 0.00000170, -0.00000296, 0.00000083, -0.00000422, 0.00000389, 0.00001105, -0.00000526, -0.00000113, 0.00000042, -0.00000200, 0.00000220, -0.00000692, 0.00000056, -0.00000699, 0.00000233, -0.00000735, 0.00000302, -0.00000580, -0.00000407, -0.00002242, -0.00001178, -0.00009697, 0.00007442, 0.00009870, 0.00009367, 0.00006840, -0.00009464, -0.00000959, -0.00001738, -0.00000171, -0.00001817, -0.00000837, -0.00000325, -0.00000169, 0.00000583, 0.00000351, -0.00000337, -0.00000572, 0.00000030, -0.00000114, 0.00000389, -0.00000226, -0.00000324, -0.00000313, -0.00000258, -0.00000354, 0.00000698, 0.00000981, -0.00000564, -0.00000443, 0.00000061, 0.00001081, 0.00000136, -0.00000501, -0.00000023, -0.00000900, 0.00000411, 0.00000209, -0.00000484, 0.00000208, 0.00000094, -0.00000849, 0.00000498, 0.00000387, -0.00000045, -0.00000002, -0.00001301, -0.00000974, -0.00000371, -0.00001027, -0.00003707, -0.00000950, 0.00005163, 0.00003438, 0.00003954, 0.00005820, -0.00001206, -0.00003284, -0.00001443, -0.00000600, -0.00000606, -0.00000123, -0.00000078, 0.00000079, -0.00000573, 0.00000172, -0.00000576, 0.00000043, -0.00000299, 0.00000397, 0.00000033, -0.00000207, -0.00000321, 0.00001540, -0.00000821, 0.00000907, -0.00000970, 0.00000414, -0.00000622, -0.00001108, 0.00000188, -0.00000011, -0.00000148, -0.00000183, -0.00000982, -0.00000094, 0.00000017, -0.00000247, 0.00000730, },
		{ 0.00000054, 0.00000730, -0.00000413, -0.00000113, -0.00000138, 0.00001170, -0.00000667, 0.00000027, 0.00000376, 0.00000022, 0.00000246, -0.00001350, 0.00000594, -0.00000472, -0.00000787, 0.00000242, 0.00000579, -0.00000862, -0.00000814, 0.00000222, -0.00000366, 0.00000108, -0.00000450, 0.00001053, 0.00000465, -0.00000698, 0.00000300, -0.00000346, 0.00000260, -0.00000383, -0.00000226, -0.00000392, -0.00000198, -0.00000237, -0.00000236, -0.00000167, -0.00000131, -0.00001713, -0.00000815, -0.00005200, -0.00005312, 0.00012487, 0.00006233, 0.00011766, -0.00001138, -0.00007239, -0.00000118, -0.00001689, -0.00000193, -0.00002118, -0.00000158, -0.00000543, 0.00000280, 0.00000510, 0.00000205, -0.00000657, -0.00000253, 0.00000000, -0.00000003, 0.00000342, -0.00000435, -0.00000218, -0.00000400, -0.00000199, -0.00000222, 0.00001295, 0.00000177, -0.00000581, -0.00000382, 0.00000633, 0.00000886, -0.00000360, -0.00000229, -0.00000352, -0.00000654, 0.00000768, -0.00000350, -0.00000166, 0.00000266, -0.00000340, -0.00000590, 0.00000946, -0.00000161, 0.00000296, -0.00000665, -0.00001256, -0.00000836, -0.00000236, -0.00002096, -0.00003785, 0.00001589, 0.00005766, 0.00002565, 0.00006057, 0.00003435, -0.00003273, -0.00002623, -0.00000817, -0.00000803, -0.00000184, -0.00000344, 0.00000291, -0.00000472, -0.00000081, -0.00000261, -0.00000248, -0.00000197, 0.00000073, 0.00000173, 0.00000145, -0.00000720, 0.00000882, 0.00000520, -0.00000335, 0.00000524, -0.00001010, 0.00000784, -0.00001572, -0.00000213, 0.00000047, 0.00000019, -0.00000303, -0.00000253, -0.00001078, 0.00000426, -0.00000472, 0.00000419, 0.00000190, },
		{ 0.00000599, 0.00000226, -0.00000268, -0.00000291, 0.00000454, 0.00000741, -0.00000931, 0.00000676, -0.00000151, 0.00000485, -0.00000577, -0.00000805, 0.00000639, -0.00001025, -0.00000250, 0.00000421, 0.00000229, -0.00001386, 0.00000010, -0.00000198, -0.00000079, -0.00000166, -0.00000002, 0.00001282, -0.00000360, -0.00000263, 0.00000174, -0.00000279, 0.00000261, -0.00000623, 0.00000001, -0.00000617, 0.00000105, -0.00000522, 0.00000064, -0.00000391, -0.00000280, -0.00002164, -0.00000675, -0.00008565, 0.00003342, 0.00011008, 0.00006462, 0.00010154, -0.00006912, -0.00003393, -0.00000922, -0.00000791, -0.00001064, -0.00001543, -0.00000053, -0.00000427, 0.00000535, 0.00000429, -0.00000094, -0.00000750, 0.00000105, -0.00000201, 0.00000284, 0.00000021, -0.00000370, -0.00000331, -0.00000260, -0.00000359, 0.00000417, 0.00001192, -0.00000424, -0.00000497, -0.00000114, 0.00001067, 0.00000324, -0.00000482, -0.00000079, -0.00000722, -0.00000012, 0.00000537, -0.00000557, 0.00000152, 0.00000109, -0.00000728, 0.00000148, 0.00000694, -0.00000230, 0.00000253, -0.00001182, -0.00001112, -0.00000602, -0.00000521, -0.00003197, -0.00002920, 0.00004436, 0.00004764, 0.00003356, 0.00006905, -0.00000070, -0.00003902, -0.00001550, -0.00000850, -0.00000540, -0.00000191, -0.00000183, 0.00000188, -0.00000597, 0.00000113, -0.00000497, -0.00000047, -0.00000274, 0.00000281, 0.00000070, -0.00000080, -0.00000532, 0.00001529, -0.00000560, 0.00000594, -0.00000577, -0.00000028, -0.00000056, -0.00001483, 0.00000292, -0.00000061, -0.00000118, -0.00000252, -0.00000558, -0.00000736, 0.00000490, -0.00000593, 0.00000839, -0.00000327, },
		{ 0.00000908, -0.00000227, -0.00000103, -0.00000364, 0.00001052, -0.00000170, -0.00000390, 0.00000656, -0.00000274, 0.00000617, -0.00001296, 0.00000124, -0.00000036, -0.00000914, 0.00000082, 0.00000580, -0.00000595, -0.00001064, 0.00000326, -0.00000497, 0.00000200, -0.00000453, 0.00000769, 0.00000795, -0.00000731, 0.00000226, -0.00000192, 0.00000086, -0.00000167, -0.00000376, -0.00000216, -0.00000385, -0.00000058, -0.00000338, -0.00000124, -0.00000136, -0.00001287, -0.00001203, -0.00003414, -0.00006952, 0.00010207, 0.00006839, 0.00009291, 0.00004281, -0.00008057, -0.00000921, -0.00001546, -0.00000193, -0.00001872, -0.00000663, -0.00000288, -0.00000037, 0.00000554, 0.00000362, -0.00000538, -0.00000438, 0.00000145, -0.00000223, 0.00000430, -0.00000276, -0.00000287, -0.00000411, -0.00000168, -0.00000365, 0.00001140, 0.00000491, -0.00000582, -0.00000461, 0.00000444, 0.00001023, -0.00000222, -0.00000308, -0.00000242, -0.00000731, 0.00000529, 0.00000015, -0.00000396, 0.00000322, -0.00000292, -0.00000647, 0.00000744, 0.00000160, 0.00000022, -0.00000187, -0.00001419, -0.00000893, -0.00000453, -0.00001166, -0.00004138, -0.00000678, 0.00006276, 0.00003324, 0.00005523, 0.00005374, -0.00003210, -0.00003074, -0.00000969, -0.00000930, -0.00000159, -0.00000402, 0.00000187, -0.00000275, -0.00000273, -0.00000128, -0.00000350, -0.00000140, -0.00000048, 0.00000183, 0.00000148, -0.00000579, 0.00000477, 0.00000944, -0.00000585, 0.00000706, -0.00001084, 0.00000755, -0.00001249, -0.00000601, 0.00000246, -0.00000050, -0.00000345, -0.00000095, -0.00000969, -0.00000070, 0.00000085, -0.00000196, 0.00000720, -0.00000445, },
		{ 0.00000758, -0.00000381, -0.00000134, -0.00000052, 0.00001132, -0.00000887, 0.00000485, 0.00000045, 0.00000199, 0.00000010, -0.00001223, 0.00000587, -0.00000722, -0.00000473, 0.00000305, 0.00000403, -0.00001316, -0.00000212, -0.00000028, -0.00000307, 0.00000101, -0.00000304, 0.00001306, -0.00000105, -0.00000421, 0.00000257, -0.00000267, 0.00000208, -0.00000478, -0.00000123, -0.00000453, -0.00000105, -0.00000270, -0.00000137, -0.00000274, -0.00000177, -0.00001965, -0.00000578, -0.00006962, -0.00000256, 0.00011339, 0.00004898, 0.00010588, -0.00002458, -0.00005734, -0.00000499, -0.00001255, -0.00000476, -0.00001939, -0.00000097, -0.00000439, 0.00000345, 0.00000492, 0.00000148, -0.00000835, 0.00000061, -0.00000136, 0.00000056, 0.00000260, -0.00000354, -0.00000360, -0.00000288, -0.00000311, 0.00000128, 0.00001313, -0.00000222, -0.00000528, -0.00000289, 0.00001007, 0.00000513, -0.00000434, -0.00000153, -0.00000542, -0.00000334, 0.00000640, -0.00000410, -0.00000033, 0.00000209, -0.00000680, -0.00000095, 0.00000803, -0.00000165, 0.00000202, -0.00000784, -0.00001380, -0.00000662, -0.00000501, -0.00002144, -0.00004405, 0.00002674, 0.00006201, 0.00003083, 0.00007428, 0.00001695, -0.00004605, -0.00001694, -0.00001076, -0.00000575, -0.00000168, -0.00000309, 0.00000229, -0.00000527, -0.00000019, -0.00000375, -0.00000164, -0.00000212, 0.00000145, 0.00000104, 0.00000004, -0.00000631, 0.00001364, -0.00000174, 0.00000208, -0.00000144, -0.00000464, 0.00000405, -0.00001667, 0.00000212, 0.00000020, -0.00000116, -0.00000390, -0.00000142, -0.00001115, 0.00000507, -0.00000435, 0.00000478, 0.00000127, -0.00000045, },
		{ 0.00000245, -0.00000244, -0.00000323, 0.00000603, 0.00000478, -0.00000787, 0.00000875, -0.00000437, 0.00000700, -0.00000923, -0.00000439, 0.00000263, -0.00000879, -0.00000113, 0.00000531, -0.00000303, -0.00001288, 0.00000369, -0.00000526, 0.00000145, -0.00000300, 0.00000401, 0.00001088, -0.00000697, 0.00000112, -0.00000044, -0.00000024, -0.00000053, -0.00000412, -0.00000166, -0.00000408, -0.00000085, -0.00000228, -0.00000233, -0.00000073, -0.00000964, -0.00001431, -0.00002200, -0.00007339, 0.00007116, 0.00008077, 0.00006483, 0.00007870, -0.00006481, -0.00002744, -0.00001019, -0.00000571, -0.00001283, -0.00001304, -0.00000039, -0.00000284, 0.00000524, 0.00000456, -0.00000308, -0.00000671, 0.00000288, -0.00000358, 0.00000371, -0.00000050, -0.00000337, -0.00000435, -0.00000152, -0.00000431, 0.00000894, 0.00000795, -0.00000539, -0.00000506, 0.00000206, 0.00001152, -0.00000093, -0.00000327, -0.00000225, -0.00000662, 0.00000179, 0.00000352, -0.00000518, 0.00000299, -0.00000207, -0.00000687, 0.00000493, 0.00000466, -0.00000170, 0.00000096, -0.00001313, -0.00001095, -0.00000574, -0.00000714, -0.00003472, -0.00003185, 0.00005859, 0.00004676, 0.00004758, 0.00007155, -0.00002512, -0.00003877, -0.00000944, -0.00001203, -0.00000101, -0.00000428, 0.00000051, -0.00000116, -0.00000375, -0.00000099, -0.00000361, -0.00000161, -0.00000079, 0.00000118, 0.00000179, -0.00000461, 0.00000145, 0.00001210, -0.00000628, 0.00000668, -0.00000923, 0.00000482, -0.00000745, -0.00001054, 0.00000434, -0.00000086, -0.00000344, -0.00000154, -0.00000558, -0.00000716, 0.00000564, -0.00000567, 0.00000902, -0.00000490, 0.00000605, },
		{ -0.00000246, -0.00000089, -0.00000347, 0.00001124, -0.00000470, 0.00000044, 0.00000445, -0.00000229, 0.00000531, -0.00001361, 0.00000297, -0.00000390, -0.00000592, 0.00000126, 0.00000540, -0.00001167, -0.00000502, 0.00000187, -0.00000541, 0.00000302, -0.00000464, 0.00001173, 0.00000217, -0.00000567, 0.00000291, -0.00000199, 0.00000128, -0.00000338, -0.00000229, -0.00000314, -0.00000257, -0.00000140, -0.00000166, -0.00000311, -0.00000034, -0.00001738, -0.00000660, -0.00005349, -0.00002879, 0.00010503, 0.00004814, 0.00008962, 0.00002265, -0.00006738, -0.00001020, -0.00001257, -0.00000315, -0.00001849, -0.00000509, -0.00000230, 0.00000059, 0.00000548, 0.00000323, -0.00000767, -0.00000130, 0.00000062, -0.00000226, 0.00000426, -0.00000261, -0.00000392, -0.00000335, -0.00000242, -0.00000121, 0.00001317, 0.00000037, -0.00000543, -0.00000429, 0.00000863, 0.00000736, -0.00000398, -0.00000186, -0.00000436, -0.00000483, 0.00000501, -0.00000096, -0.00000281, 0.00000338, -0.00000661, -0.00000254, 0.00000734, 0.00000089, -0.00000027, -0.00000303, -0.00001587, -0.00000713, -0.00000664, -0.00001183, -0.00004788, -0.00000096, 0.00007319, 0.00003398, 0.00007339, 0.00003911, -0.00005097, -0.00002092, -0.00001131, -0.00000792, -0.00000050, -0.00000435, 0.00000206, -0.00000409, -0.00000156, -0.00000278, -0.00000252, -0.00000173, 0.00000054, 0.00000087, 0.00000092, -0.00000681, 0.00001127, 0.00000233, -0.00000135, 0.00000195, -0.00000746, 0.00000627, -0.00001579, -0.00000084, 0.00000224, -0.00000143, -0.00000507, 0.00000085, -0.00001077, 0.00000067, 0.00000067, -0.00000089, 0.00000664, -0.00000588, 0.00000991, },
		{ -0.00000379, -0.00000162, 0.00000064, 0.00000982, -0.00000912, 0.00000823, -0.00000294, 0.00000429, -0.00000317, -0.00000956, 0.00000373, -0.00000738, -0.00000258, 0.00000385, 0.00000022, -0.00001469, 0.00000323, -0.00000441, -0.00000030, -0.00000061, 0.00000038, 0.00001282, -0.00000573, -0.00000039, 0.00000084, -0.00000066, -0.00000017, -0.00000374, -0.00000196, -0.00000333, -0.00000212, -0.00000067, -0.00000315, -0.00000084, -0.00000653, -0.00001555, -0.00001445, -0.00006890, 0.00003900, 0.00009040, 0.00004519, 0.00009047, -0.00003089, -0.00004671, -0.00000716, -0.00000915, -0.00000740, -0.00001729, -0.00000049, -0.00000315, 0.00000357, 0.00000541, -0.00000070, -0.00000824, 0.00000310, -0.00000317, 0.00000147, 0.00000218, -0.00000358, -0.00000454, -0.00000174, -0.00000420, 0.00000594, 0.00001046, -0.00000442, -0.00000520, -0.00000050, 0.00001227, 0.00000070, -0.00000337, -0.00000230, -0.00000556, -0.00000129, 0.00000510, -0.00000434, 0.00000134, -0.00000032, -0.00000765, 0.00000286, 0.00000623, -0.00000137, 0.00000078, -0.00000908, -0.00001450, -0.00000526, -0.00000756, -0.00002225, -0.00005093, 0.00004056, 0.00006495, 0.00004035, 0.00008520, -0.00001086, -0.00004913, -0.00000910, -0.00001457, -0.00000149, -0.00000353, -0.00000120, -0.00000008, -0.00000391, -0.00000144, -0.00000317, -0.00000227, -0.00000058, 0.00000024, 0.00000200, -0.00000344, -0.00000131, 0.00001339, -0.00000505, 0.00000476, -0.00000623, 0.00000107, -0.00000236, -0.00001400, 0.00000474, -0.00000011, -0.00000336, -0.00000333, -0.00000083, -0.00001173, 0.00000657, -0.00000494, 0.00000648, -0.00000105, -0.00000017, 0.00000787, },
		{ -0.00000205, -0.00000359, 0.00000733, 0.00000183, -0.00000461, 0.00000833, -0.00000538, 0.00000756, -0.00001134, -0.00000173, -0.00000101, -0.00000614, -0.00000046, 0.00000568, -0.00000894, -0.00000870, 0.00000420, -0.00000732, 0.00000403, -0.00000481, 0.00000925, 0.00000550, -0.00000665, 0.00000266, -0.00000098, 0.00000061, -0.00000237, -0.00000285, -0.00000240, -0.00000303, -0.00000160, -0.00000048, -0.00000425, 0.00000075, -0.00001466, -0.00000825, -0.00003947, -0.00004450, 0.00008689, 0.00005651, 0.00006587, 0.00005784, -0.00005809, -0.00002401, -0.00000943, -0.00000489, -0.00001417, -0.00001083, -0.00000005, -0.00000173, 0.00000529, 0.00000452, -0.00000592, -0.00000382, 0.00000277, -0.00000430, 0.00000431, -0.00000064, -0.00000440, -0.00000373, -0.00000196, -0.00000286, 0.00001196, 0.00000334, -0.00000543, -0.00000512, 0.00000628, 0.00000982, -0.00000365, -0.00000186, -0.00000396, -0.00000491, 0.00000220, 0.00000235, -0.00000449, 0.00000384, -0.00000577, -0.00000406, 0.00000590, 0.00000367, -0.00000205, 0.00000001, -0.00001505, -0.00000961, -0.00000682, -0.00000782, -0.00003948, -0.00003270, 0.00007409, 0.00004561, 0.00006670, 0.00006347, -0.00005118, -0.00002846, -0.00001014, -0.00001129, 0.00000066, -0.00000486, 0.00000106, -0.00000272, -0.00000259, -0.00000231, -0.00000292, -0.00000181, 0.00000029, 0.00000010, 0.00000189, -0.00000698, 0.00000856, 0.00000605, -0.00000370, 0.00000377, -0.00000823, 0.00000603, -0.00001268, -0.00000500, 0.00000446, -0.00000139, -0.00000580, 0.00000081, -0.00000697, -0.00000586, 0.00000567, -0.00000502, 0.00000946, -0.00000740, 0.00000767, 0.00000181, },
		},
		{ { -0.00000402, -0.00000064, -0.00001365, 0.00000755, -0.00001370, -0.00000075, 0.00000643, -0.00000586, 0.00000802, -0.00000884, 0.00000903, -0.00000183, 0.00000511, 0.00000042, -0.00000126, -0.00000002, 0.00000110, 0.00000162, 0.00000068, 0.00000023, -0.00000159, -0.00000198, 0.00000027, 0.00000001, 0.00000363, 0.00000192, -0.00000779, 0.00000242, -0.00000064, 0.00000056, -0.00000914, -0.00000185, -0.00000111, -0.00000135, 0.00000211, -0.00000350, 0.00000657, -0.00000616, -0.00001892, -0.00001443, -0.00005906, 0.00001561, 0.00008537, 0.00003782, 0.00008053, 0.00000784, -0.00004964, -0.00001074, -0.00001002, -0.00000682, -0.00000548, -0.00000282, -0.00000925, -0.00000161, 0.00000293, 0.00000291, -0.00000441, 0.00000130, -0.00000288, -0.00000120, -0.00000162, -0.00000513, 0.00000331, -0.00000198, -0.00000043, -0.00000198, 0.00000331, -0.00000513, -0.00000162, -0.00000120, -0.00000288, 0.00000130, -0.00000441, 0.00000291, 0.00000293, -0.00000161, -0.00000925, -0.00000282, -0.00000548, -0.00000682, -0.00001002, -0.00001074, -0.00004964, 0.00000784, 0.00008053, 0.00003782, 0.00008537, 0.00001561, -0.00005906, -0.00001443, -0.00001892, -0.00000616, 0.00000657, -0.00000350, 0.00000211, -0.00000135, -0.00000111, -0.00000185, -0.00000914, 0.00000056, -0.00000064, 0.00000242, -0.00000779, 0.00000192, 0.00000363, 0.00000001, 0.00000027, -0.00000198, -0.00000159, 0.00000023, 0.00000068, 0.00000162, 0.00000110, -0.00000002, -0.00000126, 0.00000042, 0.00000511, -0.00000183, 0.00000903, -0.00000884, 0.00000802, -0.00000586, 0.00000643, -0.00000075, -0.00001370, 0.00000755, -0.00001365, -0.00000064, },
		{ -0.00000261, -0.00000571, -0.00000902, 0.00000775, -0.00001973, 0.00001055, -0.00000289, 0.00000233, 0.00000094, -0.00000508, 0.00000918, -0.00000288, 0.00000667, -0.00000253, 0.00000068, -0.00000093, 0.00000252, 0.00000029, 0.00000126, -0.00000078, -0.00000144, -0.00000197, 0.00000112, -0.00000020, 0.00000540, -0.00000262, -0.00000519, 0.00000260, -0.00000045, -0.00000165, -0.00000989, 0.00000132, -0.00000290, 0.00000127, -0.00000011, -0.00000172, 0.00000728, -0.00001521, -0.00001314, -0.00002941, -0.00005022, 0.00005801, 0.00006418, 0.00004434, 0.00007273, -0.00002292, -0.00003783, -0.00000790, -0.00000918, -0.00000713, -0.00000291, -0.00000516, -0.00000818, 0.00000069, 0.00000378, 0.00000006, -0.00000334, 0.00000180, -0.00000437, 0.00000077, -0.00000498, -0.00000173, 0.00000246, -0.00000198, -0.00000088, -0.00000080, 0.00000211, -0.00000652, 0.00000125, -0.00000389, 0.00000000, -0.00000151, -0.00000264, 0.00000416, 0.00000190, -0.00000460, -0.00000806, -0.00000277, -0.00000673, -0.00000735, -0.00000980, -0.00001857, -0.00005387, 0.00004661, 0.00007067, 0.00004914, 0.00008699, -0.00003182, -0.00004583, -0.00001093, -0.00001961, 0.00000176, 0.00000330, -0.00000298, 0.00000269, -0.00000322, 0.00000027, -0.00000548, -0.00000617, 0.00000084, 0.00000026, 0.00000031, -0.00000794, 0.00000564, 0.00000107, 0.00000084, -0.00000099, -0.00000167, -0.00000160, 0.00000104, 0.00000041, 0.00000255, -0.00000030, 0.00000080, -0.00000313, 0.00000443, 0.00000128, 0.00000222, 0.00000419, -0.00000597, 0.00000847, -0.00000900, 0.00001230, -0.00001206, -0.00000292, -0.00000029, -0.00001075, -0.00000022, },
		{ -0.00000042, -0.00001247, 0.00000142, -0.00000153, -0.00001542, 0.00001435, -0.00000828, 0.00000898, -0.00000776, 0.00000294, 0.00000487, -0.00000032, 0.00000488, -0.00000279, 0.00000108, -0.00000064, 0.00000281, -0.00000043, 0.00000147, -0.00000149, -0.00000149, -0.00000137, 0.00000129, 0.00000064, 0.00000527, -0.00000612, -0.00000175, 0.00000166, 0.00000011, -0.00000519, -0.00000797, 0.00000288, -0.00000409, 0.00000387, -0.00000358, 0.00000306, 0.00000224, -0.00001934, -0.00001041, -0.00004698, -0.00001850, 0.00007827, 0.00004206, 0.00005893, 0.00004907, -0.00003928, -0.00002595, -0.00000678, -0.00000910, -0.00000607, -0.00000146, -0.00000785, -0.00000561, 0.00000249, 0.00000345, -0.00000257, -0.00000070, 0.00000001, -0.00000357, 0.00000041, -0.00000662, 0.00000178, 0.00000053, -0.00000122, -0.00000183, 0.00000107, -0.00000091, -0.00000502, 0.00000168, -0.00000501, 0.00000195, -0.00000455, 0.00000094, 0.00000350, 0.00000091, -0.00000755, -0.00000542, -0.00000507, -0.00000569, -0.00001001, -0.00000754, -0.00003370, -0.00004143, 0.00007992, 0.00005286, 0.00007540, 0.00005781, -0.00006211, -0.00002470, -0.00001642, -0.00001329, 0.00000493, 0.00000062, -0.00000179, 0.00000216, -0.00000420, 0.00000069, -0.00000828, -0.00000225, -0.00000035, 0.00000202, -0.00000359, -0.00000487, 0.00000681, -0.00000074, 0.00000139, -0.00000207, -0.00000136, -0.00000124, 0.00000136, 0.00000086, 0.00000238, -0.00000071, 0.00000032, -0.00000300, 0.00000687, -0.00000198, 0.00000621, -0.00000193, 0.00000091, 0.00000260, -0.00000494, 0.00001025, -0.00001649, 0.00000456, -0.00000889, -0.00000436, -0.00000263, },
		{ -0.00000147, -0.00001478, 0.00001034, -0.00001463, -0.00000258, 0.00000864, -0.00000544, 0.00000821, -0.00001145, 0.00000965, -0.00000025, 0.00000344, 0.00000136, -0.00000112, 0.00000011, 0.00000080, 0.00000180, -0.00000021, 0.00000105, -0.00000179, -0.00000168, -0.00000039, 0.00000102, 0.00000211, 0.00000325, -0.00000760, 0.00000130, 0.00000038, 0.00000022, -0.00000899, -0.00000360, 0.00000179, -0.00000310, 0.00000434, -0.00000566, 0.00000810, -0.00000738, -0.00001638, -0.00001749, -0.00005302, 0.00002296, 0.00007373, 0.00003244, 0.00006842, 0.00001894, -0.00004263, -0.00001696, -0.00000621, -0.00000981, -0.00000310, -0.00000268, -0.00000889, -0.00000278, 0.00000400, 0.00000135, -0.00000309, 0.00000115, -0.00000223, -0.00000164, -0.00000199, -0.00000559, 0.00000374, -0.00000119, -0.00000060, -0.00000238, 0.00000236, -0.00000419, -0.00000163, -0.00000046, -0.00000356, 0.00000121, -0.00000533, 0.00000407, 0.00000204, -0.00000060, -0.00000914, -0.00000344, -0.00000761, -0.00000404, -0.00001334, -0.00000570, -0.00005329, -0.00000803, 0.00009400, 0.00004425, 0.00009871, 0.00000478, -0.00006316, -0.00001197, -0.00002116, -0.00000486, 0.00000469, -0.00000109, -0.00000025, 0.00000042, -0.00000383, -0.00000055, -0.00000881, 0.00000063, -0.00000134, 0.00000306, -0.00000748, 0.00000035, 0.00000517, -0.00000099, 0.00000106, -0.00000261, -0.00000114, -0.00000064, 0.00000146, 0.00000161, 0.00000141, -0.00000034, -0.00000130, -0.00000044, 0.00000632, -0.00000257, 0.00000735, -0.00000519, 0.00000663, -0.00000484, 0.00000293, 0.00000177, -0.00001265, 0.00000431, -0.00001211, -0.00000003, -0.00000393, },
		{ -0.00000723, -0.00000867, 0.00001004, -0.00002193, 0.00001040, -0.00000119, 0.00000294, 0.00000001, -0.00000722, 0.00001088, -0.00000246, 0.00000532, -0.00000120, 0.00000043, -0.00000076, 0.00000223, 0.00000034, 0.00000050, 0.00000028, -0.00000192, -0.00000166, 0.00000057, 0.00000080, 0.00000344, 0.00000006, -0.00000696, 0.00000322, -0.00000071, -0.00000082, -0.00001148, 0.00000184, -0.00000169, 0.00000073, 0.00000103, -0.00000326, 0.00000843, -0.00001579, -0.00001120, -0.00003250, -0.00003906, 0.00005641, 0.00005464, 0.00003737, 0.00006476, -0.00000755, -0.00003842, -0.00001036, -0.00000703, -0.00000935, -0.00000018, -0.00000565, -0.00000794, -0.00000009, 0.00000450, -0.00000147, -0.00000140, 0.00000090, -0.00000312, -0.00000046, -0.00000476, -0.00000251, 0.00000363, -0.00000188, -0.00000076, -0.00000182, 0.00000194, -0.00000577, 0.00000136, -0.00000336, -0.00000075, -0.00000179, -0.00000304, 0.00000504, 0.00000092, -0.00000302, -0.00000862, -0.00000379, -0.00000792, -0.00000474, -0.00001388, -0.00001016, -0.00006730, 0.00003982, 0.00008478, 0.00005750, 0.00009576, -0.00004793, -0.00004192, -0.00001331, -0.00001836, 0.00000020, 0.00000409, -0.00000279, 0.00000185, -0.00000264, -0.00000194, -0.00000362, -0.00000640, 0.00000098, -0.00000056, 0.00000159, -0.00000882, 0.00000519, 0.00000206, 0.00000000, -0.00000015, -0.00000252, -0.00000097, -0.00000004, 0.00000173, 0.00000198, 0.00000047, -0.00000017, -0.00000265, 0.00000315, 0.00000349, -0.00000066, 0.00000534, -0.00000399, 0.00000748, -0.00000870, 0.00000893, -0.00000752, -0.00000487, -0.00000191, -0.00000891, -0.00000013, -0.00000249, },
		{ -0.00001408, 0.00000329, -0.00000092, -0.00001733, 0.00001495, -0.00000666, 0.00000919, -0.00000949, 0.00000178, 0.00000702, -0.00000109, 0.00000438, -0.00000158, 0.00000062, -0.00000049, 0.00000263, -0.00000067, 0.00000116, -0.00000058, -0.00000201, -0.00000130, 0.00000131, 0.00000076, 0.00000418, -0.00000353, -0.00000435, 0.00000341, -0.00000104, -0.00000369, -0.00001073, 0.00000547, -0.00000498, 0.00000522, -0.00000448, 0.00000335, 0.00000172, -0.00001763, -0.00001165, -0.00004534, -0.00000837, 0.00007033, 0.00003564, 0.00004950, 0.00004825, -0.00002488, -0.00003142, -0.00000528, -0.00000972, -0.00000605, 0.00000010, -0.00000798, -0.00000603, 0.00000268, 0.00000295, -0.00000284, 0.00000060, -0.00000066, -0.00000252, -0.00000099, -0.00000619, 0.00000106, 0.00000200, -0.00000161, -0.00000162, -0.00000035, -0.00000023, -0.00000472, 0.00000208, -0.00000482, 0.00000114, -0.00000483, 0.00000104, 0.00000379, 0.00000043, -0.00000593, -0.00000663, -0.00000636, -0.00000533, -0.00000928, -0.00000939, -0.00002637, -0.00006188, 0.00008477, 0.00006385, 0.00008820, 0.00005538, -0.00007426, -0.00001864, -0.00002092, -0.00001010, 0.00000147, 0.00000363, -0.00000390, 0.00000332, -0.00000562, 0.00000019, -0.00000699, -0.00000238, -0.00000062, 0.00000172, -0.00000244, -0.00000636, 0.00000740, -0.00000065, 0.00000125, -0.00000183, -0.00000181, -0.00000100, 0.00000073, 0.00000205, 0.00000182, -0.00000011, -0.00000067, -0.00000257, 0.00000591, 0.00000015, 0.00000219, 0.00000201, 0.00000011, 0.00000345, -0.00000699, 0.00000938, -0.00001235, 0.00000067, -0.00000838, -0.00000313, -0.00000258, -0.00000069, },
		{ -0.00001546, 0.00001262, -0.00001552, -0.00000348, 0.00000930, -0.00000351, 0.00000727, -0.00001307, 0.00000932, 0.00000196, 0.00000155, 0.00000222, -0.00000076, -0.00000003, 0.00000063, 0.00000199, -0.00000101, 0.00000160, -0.00000157, -0.00000184, -0.00000075, 0.00000185, 0.00000084, 0.00000414, -0.00000682, -0.00000032, 0.00000168, -0.00000043, -0.00000835, -0.00000563, 0.00000458, -0.00000453, 0.00000621, -0.00000741, 0.00000916, -0.00000842, -0.00001369, -0.00002091, -0.00004534, 0.00002683, 0.00006434, 0.00002714, 0.00005846, 0.00002543, -0.00003319, -0.00002344, -0.00000261, -0.00001243, -0.00000088, -0.00000278, -0.00000824, -0.00000392, 0.00000506, -0.00000043, -0.00000160, 0.00000093, -0.00000175, -0.00000168, -0.00000275, -0.00000553, 0.00000363, -0.00000001, -0.00000107, -0.00000242, 0.00000101, -0.00000288, -0.00000193, 0.00000046, -0.00000408, 0.00000068, -0.00000568, 0.00000463, 0.00000150, 0.00000006, -0.00000839, -0.00000476, -0.00000910, -0.00000197, -0.00001558, -0.00000255, -0.00005303, -0.00002828, 0.00010751, 0.00005308, 0.00011286, -0.00000926, -0.00006520, -0.00001028, -0.00002299, -0.00000356, 0.00000258, 0.00000138, -0.00000252, 0.00000196, -0.00000613, 0.00000023, -0.00000810, 0.00000049, -0.00000180, 0.00000328, -0.00000681, -0.00000123, 0.00000646, -0.00000176, 0.00000181, -0.00000337, -0.00000065, -0.00000135, 0.00000194, 0.00000190, 0.00000161, -0.00000074, -0.00000134, -0.00000105, 0.00000690, -0.00000238, 0.00000465, -0.00000061, 0.00000414, -0.00000261, -0.00000141, 0.00000459, -0.00001134, 0.00000066, -0.00001022, 0.00000036, -0.00000368, -0.00000252, },
		{ -0.00000775, 0.00001130, -0.00002304, 0.00000994, -0.00000009, 0.00000458, -0.00000190, -0.00000831, 0.00001130, -0.00000081, 0.00000298, 0.00000082, -0.00000031, -0.00000029, 0.00000160, 0.00000090, -0.00000076, 0.00000170, -0.00000259, -0.00000134, -0.00000006, 0.00000200, 0.00000127, 0.00000282, -0.00000855, 0.00000363, -0.00000093, -0.00000002, -0.00001276, 0.00000204, -0.00000055, 0.00000058, 0.00000158, -0.00000397, 0.00000840, -0.00001507, -0.00001059, -0.00003387, -0.00002899, 0.00005335, 0.00004689, 0.00003116, 0.00005721, 0.00000396, -0.00003559, -0.00001448, -0.00000403, -0.00001188, 0.00000269, -0.00000655, -0.00000709, -0.00000120, 0.00000544, -0.00000339, 0.00000099, -0.00000047, -0.00000157, -0.00000174, -0.00000454, -0.00000310, 0.00000439, -0.00000140, -0.00000099, -0.00000243, 0.00000122, -0.00000436, 0.00000070, -0.00000196, -0.00000218, -0.00000164, -0.00000365, 0.00000599, -0.00000034, -0.00000099, -0.00000948, -0.00000460, -0.00000942, -0.00000168, -0.00001837, -0.00000179, -0.00007820, 0.00002793, 0.00010088, 0.00006816, 0.00010347, -0.00006606, -0.00003560, -0.00001737, -0.00001569, -0.00000234, 0.00000540, -0.00000302, 0.00000146, -0.00000258, -0.00000343, -0.00000265, -0.00000578, 0.00000045, -0.00000072, 0.00000198, -0.00000869, 0.00000396, 0.00000349, -0.00000112, 0.00000103, -0.00000394, 0.00000026, -0.00000155, 0.00000330, 0.00000126, 0.00000149, -0.00000165, -0.00000152, 0.00000124, 0.00000603, -0.00000348, 0.00000618, -0.00000167, 0.00000585, -0.00000736, 0.00000453, -0.00000221, -0.00000711, -0.00000355, -0.00000692, -0.00000018, -0.00000224, -0.00000893, },
		{ 0.00000532, -0.00000120, -0.00001763, 0.00001411, -0.00000436, 0.00000932, -0.00001127, 0.00000085, 0.00000851, -0.00000096, 0.00000296, 0.00000043, -0.00000055, 0.00000017, 0.00000200, -0.00000029, 0.00000014, 0.00000101, -0.00000309, -0.00000086, 0.00000099, 0.00000133, 0.00000257, -0.00000049, -0.00000717, 0.00000522, -0.00000234, -0.00000199, -0.00001331, 0.00000761, -0.00000557, 0.00000653, -0.00000554, 0.00000401, 0.00000036, -0.00001474, -0.00001438, -0.00004117, -0.00000150, 0.00006356, 0.00002958, 0.00004207, 0.00004533, -0.00001191, -0.00003434, -0.00000554, -0.00000917, -0.00000660, 0.00000179, -0.00000832, -0.00000596, 0.00000253, 0.00000260, -0.00000336, 0.00000226, -0.00000181, -0.00000099, -0.00000272, -0.00000542, 0.00000014, 0.00000340, -0.00000185, -0.00000154, -0.00000159, 0.00000015, -0.00000397, 0.00000181, -0.00000367, -0.00000070, -0.00000415, 0.00000034, 0.00000464, -0.00000074, -0.00000333, -0.00000883, -0.00000664, -0.00000612, -0.00000707, -0.00001317, -0.00001664, -0.00008266, 0.00008555, 0.00007821, 0.00010252, 0.00005021, -0.00008600, -0.00001169, -0.00002632, -0.00000590, -0.00000286, 0.00000709, -0.00000627, 0.00000465, -0.00000719, 0.00000000, -0.00000626, -0.00000187, -0.00000137, 0.00000185, -0.00000195, -0.00000688, 0.00000703, 0.00000016, 0.00000065, -0.00000121, -0.00000284, -0.00000004, -0.00000058, 0.00000375, 0.00000094, 0.00000079, -0.00000216, -0.00000135, 0.00000394, 0.00000324, -0.00000251, 0.00000639, -0.00000113, 0.00000448, -0.00000867, 0.00000784, -0.00000750, -0.00000367, -0.00000762, -0.00000198, -0.00000250, -0.00000101, -0.00001556, },
		{ 0.00001433, -0.00001638, -0.00000332, 0.00000835, -0.00000023, 0.00000543, -0.00001385, 0.00000820, 0.00000454, -0.00000032, 0.00000284, -0.00000011, -0.00000052, 0.00000067, 0.00000205, -0.00000152, 0.00000172, -0.00000082, -0.00000253, -0.00000073, 0.00000235, -0.00000004, 0.00000450, -0.00000546, -0.00000234, 0.00000318, -0.00000132, -0.00000731, -0.00000772, 0.00000690, -0.00000535, 0.00000752, -0.00000859, 0.00000966, -0.00000914, -0.00001105, -0.00002438, -0.00003646, 0.00002750, 0.00005739, 0.00002141, 0.00005092, 0.00002796, -0.00002248, -0.00002922, 0.00000011, -0.00001420, 0.00000081, -0.00000288, -0.00000747, -0.00000487, 0.00000595, -0.00000221, -0.00000013, 0.00000080, -0.00000153, -0.00000134, -0.00000383, -0.00000502, 0.00000303, 0.00000146, -0.00000172, -0.00000226, -0.00000057, -0.00000135, -0.00000238, 0.00000139, -0.00000425, -0.00000044, -0.00000532, 0.00000449, 0.00000143, 0.00000027, -0.00000690, -0.00000694, -0.00000965, -0.00000100, -0.00001628, -0.00000173, -0.00004882, -0.00005193, 0.00011924, 0.00006599, 0.00012671, -0.00002606, -0.00006519, -0.00000942, -0.00002431, -0.00000236, 0.00000034, 0.00000375, -0.00000450, 0.00000306, -0.00000785, 0.00000052, -0.00000716, 0.00000030, -0.00000209, 0.00000321, -0.00000600, -0.00000256, 0.00000727, -0.00000209, 0.00000232, -0.00000406, -0.00000034, -0.00000169, 0.00000197, 0.00000257, 0.00000163, -0.00000112, -0.00000145, -0.00000133, 0.00000679, -0.00000128, 0.00000110, 0.00000465, 0.00000081, 0.00000039, -0.00000604, 0.00000727, -0.00000965, -0.00000334, -0.00000804, 0.00000056, -0.00000326, -0.00000379, -0.00001562, },
		{ 0.00001146, -0.00002303, 0.00000926, 0.00000026, 0.00000723, -0.00000464, -0.00000846, 0.00001057, 0.00000183, -0.00000002, 0.00000326, -0.00000138, 0.00000034, 0.00000078, 0.00000178, -0.00000225, 0.00000320, -0.00000318, -0.00000120, -0.00000058, 0.00000314, -0.00000078, 0.00000528, -0.00000968, 0.00000362, -0.00000091, 0.00000051, -0.00001344, 0.00000176, 0.00000053, 0.00000085, 0.00000154, -0.00000391, 0.00000732, -0.00001322, -0.00001116, -0.00003356, -0.00002023, 0.00004917, 0.00004095, 0.00002509, 0.00005093, 0.00001144, -0.00002991, -0.00001954, -0.00000079, -0.00001417, 0.00000519, -0.00000748, -0.00000594, -0.00000237, 0.00000634, -0.00000541, 0.00000351, -0.00000196, -0.00000002, -0.00000286, -0.00000444, -0.00000337, 0.00000463, -0.00000047, -0.00000159, -0.00000261, 0.00000001, -0.00000236, -0.00000064, 0.00000016, -0.00000412, -0.00000122, -0.00000431, 0.00000686, -0.00000171, 0.00000131, -0.00001045, -0.00000537, -0.00001090, 0.00000127, -0.00002250, 0.00000547, -0.00008538, 0.00001056, 0.00011815, 0.00008248, 0.00010873, -0.00008512, -0.00002750, -0.00002271, -0.00001192, -0.00000562, 0.00000705, -0.00000362, 0.00000154, -0.00000315, -0.00000401, -0.00000267, -0.00000435, -0.00000066, -0.00000025, 0.00000153, -0.00000764, 0.00000213, 0.00000514, -0.00000231, 0.00000235, -0.00000569, 0.00000181, -0.00000322, 0.00000485, 0.00000063, 0.00000255, -0.00000340, 0.00000001, -0.00000098, 0.00000855, -0.00000593, 0.00000653, 0.00000107, 0.00000353, -0.00000511, -0.00000051, 0.00000336, -0.00000927, -0.00000532, -0.00000476, -0.00000038, -0.00000185, -0.00001080, -0.00000626, },
		{ -0.00000235, -0.00001636, 0.00001196, -0.00000164, 0.00000955, -0.00001313, 0.00000019, 0.00000930, 0.00000003, 0.00000079, 0.00000304, -0.00000226, 0.00000121, 0.00000100, 0.00000062, -0.00000142, 0.00000306, -0.00000452, -0.00000023, 0.00000046, 0.00000216, 0.00000072, 0.00000262, -0.00000980, 0.00000673, -0.00000348, -0.00000043, -0.00001526, 0.00000887, -0.00000560, 0.00000761, -0.00000655, 0.00000484, -0.00000153, -0.00001104, -0.00001813, -0.00003494, 0.00000236, 0.00005806, 0.00002382, 0.00003610, 0.00004144, -0.00000131, -0.00003442, -0.00000748, -0.00000758, -0.00000753, 0.00000336, -0.00000867, -0.00000553, 0.00000216, 0.00000230, -0.00000401, 0.00000407, -0.00000319, 0.00000072, -0.00000449, -0.00000451, -0.00000079, 0.00000454, -0.00000178, -0.00000173, -0.00000252, 0.00000015, -0.00000274, 0.00000084, -0.00000158, -0.00000348, -0.00000263, -0.00000104, 0.00000595, -0.00000249, 0.00000008, -0.00001187, -0.00000601, -0.00000790, -0.00000370, -0.00001827, -0.00000571, -0.00010187, 0.00008051, 0.00009668, 0.00011845, 0.00004168, -0.00009653, -0.00000450, -0.00003213, -0.00000108, -0.00000771, 0.00001068, -0.00000864, 0.00000598, -0.00000886, 0.00000026, -0.00000624, -0.00000067, -0.00000263, 0.00000247, -0.00000220, -0.00000637, 0.00000567, 0.00000163, -0.00000037, -0.00000023, -0.00000437, 0.00000151, -0.00000238, 0.00000575, -0.00000006, 0.00000179, -0.00000396, 0.00000042, 0.00000126, 0.00000687, -0.00000742, 0.00001075, -0.00000241, 0.00000526, -0.00000963, 0.00000561, -0.00000220, -0.00000816, -0.00000680, -0.00000084, -0.00000242, -0.00000138, -0.00001686, 0.00000747, },
		{ -0.00001718, -0.00000215, 0.00000591, 0.00000410, 0.00000302, -0.00001403, 0.00000656, 0.00000721, -0.00000193, 0.00000311, 0.00000084, -0.00000136, 0.00000098, 0.00000190, -0.00000164, 0.00000134, 0.00000044, -0.00000372, -0.00000035, 0.00000249, -0.00000044, 0.00000432, -0.00000363, -0.00000458, 0.00000470, -0.00000233, -0.00000603, -0.00000958, 0.00000842, -0.00000530, 0.00000807, -0.00000907, 0.00000949, -0.00000941, -0.00000870, -0.00002753, -0.00002696, 0.00002548, 0.00005276, 0.00001510, 0.00004560, 0.00002751, -0.00001175, -0.00003340, 0.00000141, -0.00001473, 0.00000170, -0.00000282, -0.00000671, -0.00000550, 0.00000653, -0.00000385, 0.00000117, 0.00000085, -0.00000158, -0.00000072, -0.00000508, -0.00000418, 0.00000206, 0.00000306, -0.00000240, -0.00000204, -0.00000217, 0.00000019, -0.00000283, 0.00000214, -0.00000392, -0.00000222, -0.00000421, 0.00000364, 0.00000183, -0.00000000, -0.00000471, -0.00000997, -0.00000914, -0.00000134, -0.00001529, -0.00000331, -0.00004111, -0.00007739, 0.00012697, 0.00008483, 0.00013900, -0.00004497, -0.00006325, -0.00000945, -0.00002507, -0.00000133, -0.00000189, 0.00000586, -0.00000602, 0.00000356, -0.00000888, 0.00000039, -0.00000616, 0.00000019, -0.00000235, 0.00000300, -0.00000522, -0.00000343, 0.00000738, -0.00000182, 0.00000245, -0.00000451, -0.00000038, -0.00000153, 0.00000148, 0.00000364, 0.00000145, -0.00000145, -0.00000168, -0.00000126, 0.00000600, 0.00000057, -0.00000298, 0.00001022, -0.00000292, 0.00000361, -0.00001030, 0.00000937, -0.00000744, -0.00000760, -0.00000567, 0.00000058, -0.00000268, -0.00000528, -0.00001524, 0.00001540, },
		{ -0.00002192, 0.00000844, -0.00000017, 0.00001071, -0.00000792, -0.00000792, 0.00000897, 0.00000509, -0.00000331, 0.00000579, -0.00000253, 0.00000094, -0.00000006, 0.00000276, -0.00000369, 0.00000450, -0.00000348, -0.00000142, -0.00000082, 0.00000398, -0.00000239, 0.00000712, -0.00001010, 0.00000303, -0.00000053, 0.00000061, -0.00001334, 0.00000092, 0.00000150, 0.00000155, 0.00000096, -0.00000318, 0.00000539, -0.00001050, -0.00001268, -0.00003172, -0.00001292, 0.00004421, 0.00003673, 0.00001883, 0.00004631, 0.00001517, -0.00002225, -0.00002469, 0.00000202, -0.00001567, 0.00000686, -0.00000809, -0.00000477, -0.00000333, 0.00000693, -0.00000722, 0.00000581, -0.00000326, 0.00000125, -0.00000365, -0.00000454, -0.00000329, 0.00000433, 0.00000089, -0.00000252, -0.00000246, -0.00000153, 0.00000001, -0.00000246, 0.00000275, -0.00000628, -0.00000078, -0.00000481, 0.00000744, -0.00000298, 0.00000361, -0.00001133, -0.00000633, -0.00001198, 0.00000358, -0.00002554, 0.00001063, -0.00008798, -0.00001214, 0.00013527, 0.00010211, 0.00010989, -0.00010375, -0.00001853, -0.00002872, -0.00000750, -0.00000926, 0.00000876, -0.00000440, 0.00000200, -0.00000435, -0.00000364, -0.00000363, -0.00000230, -0.00000218, 0.00000070, 0.00000040, -0.00000590, -0.00000003, 0.00000675, -0.00000332, 0.00000357, -0.00000748, 0.00000336, -0.00000470, 0.00000607, 0.00000032, 0.00000341, -0.00000515, 0.00000165, -0.00000320, 0.00001070, -0.00000774, 0.00000625, 0.00000425, 0.00000058, -0.00000221, -0.00000564, 0.00000860, -0.00001098, -0.00000734, -0.00000242, -0.00000076, -0.00000133, -0.00001279, -0.00000423, 0.00001054, },
		{ -0.00001365, 0.00000874, 0.00000122, 0.00001000, -0.00001505, -0.00000023, 0.00000943, 0.00000171, -0.00000185, 0.00000593, -0.00000427, 0.00000245, -0.00000022, 0.00000189, -0.00000327, 0.00000527, -0.00000601, 0.00000035, -0.00000006, 0.00000301, -0.00000104, 0.00000541, -0.00001185, 0.00000763, -0.00000418, 0.00000065, -0.00001621, 0.00000892, -0.00000485, 0.00000828, -0.00000735, 0.00000561, -0.00000363, -0.00000700, -0.00002232, -0.00002732, 0.00000367, 0.00005369, 0.00001850, 0.00003103, 0.00003756, 0.00000630, -0.00003173, -0.00001080, -0.00000527, -0.00000855, 0.00000453, -0.00000879, -0.00000495, 0.00000175, 0.00000192, -0.00000461, 0.00000579, -0.00000450, 0.00000227, -0.00000601, -0.00000368, -0.00000158, 0.00000526, -0.00000129, -0.00000228, -0.00000305, -0.00000023, -0.00000108, -0.00000073, 0.00000127, -0.00000694, -0.00000056, -0.00000282, 0.00000745, -0.00000452, 0.00000396, -0.00001544, -0.00000476, -0.00001028, 0.00000021, -0.00002377, 0.00000493, -0.00011750, 0.00006809, 0.00011970, 0.00013628, 0.00002902, -0.00010488, 0.00000210, -0.00003767, 0.00000377, -0.00001256, 0.00001397, -0.00001070, 0.00000709, -0.00001051, 0.00000098, -0.00000695, 0.00000114, -0.00000428, 0.00000350, -0.00000312, -0.00000492, 0.00000349, 0.00000360, -0.00000163, 0.00000093, -0.00000620, 0.00000338, -0.00000437, 0.00000774, -0.00000094, 0.00000267, -0.00000577, 0.00000241, -0.00000171, 0.00001054, -0.00001197, 0.00001459, -0.00000330, 0.00000536, -0.00000961, 0.00000276, 0.00000322, -0.00001245, -0.00000609, 0.00000033, -0.00000236, -0.00000180, -0.00001793, 0.00000962, -0.00000426, },
		{ -0.00000012, 0.00000229, 0.00000902, 0.00000045, -0.00001388, 0.00000464, 0.00000969, -0.00000310, 0.00000299, 0.00000205, -0.00000251, 0.00000153, 0.00000150, -0.00000133, 0.00000049, 0.00000212, -0.00000527, 0.00000030, 0.00000229, -0.00000039, 0.00000370, -0.00000153, -0.00000680, 0.00000602, -0.00000327, -0.00000471, -0.00001091, 0.00000882, -0.00000420, 0.00000775, -0.00000878, 0.00000859, -0.00000911, -0.00000687, -0.00002995, -0.00001750, 0.00002150, 0.00005002, 0.00000843, 0.00004200, 0.00002520, -0.00000220, -0.00003532, 0.00000097, -0.00001383, 0.00000164, -0.00000250, -0.00000603, -0.00000574, 0.00000671, -0.00000521, 0.00000219, 0.00000110, -0.00000183, 0.00000004, -0.00000630, -0.00000318, 0.00000090, 0.00000461, -0.00000294, -0.00000196, -0.00000360, 0.00000157, -0.00000313, 0.00000258, -0.00000303, -0.00000464, -0.00000243, 0.00000217, 0.00000264, -0.00000069, -0.00000194, -0.00001370, -0.00000762, -0.00000293, -0.00001274, -0.00000693, -0.00003089, -0.00010263, 0.00012827, 0.00011144, 0.00014847, -0.00006533, -0.00005953, -0.00001039, -0.00002518, -0.00000060, -0.00000396, 0.00000752, -0.00000694, 0.00000336, -0.00000920, -0.00000006, -0.00000529, 0.00000029, -0.00000265, 0.00000277, -0.00000463, -0.00000369, 0.00000669, -0.00000089, 0.00000212, -0.00000462, -0.00000083, -0.00000085, 0.00000050, 0.00000502, 0.00000111, -0.00000173, -0.00000197, -0.00000090, 0.00000470, 0.00000288, -0.00000717, 0.00001560, -0.00000656, 0.00000644, -0.00001357, 0.00001048, -0.00000464, -0.00001200, -0.00000320, 0.00000048, -0.00000198, -0.00000697, -0.00001429, 0.00001574, -0.00001779, },
		{ 0.00000748, -0.00000126, 0.00001468, -0.00001132, -0.00000705, 0.00000686, 0.00000854, -0.00000643, 0.00000807, -0.00000353, 0.00000136, -0.00000074, 0.00000361, -0.00000482, 0.00000535, -0.00000332, -0.00000210, -0.00000066, 0.00000436, -0.00000334, 0.00000812, -0.00000967, 0.00000180, 0.00000026, 0.00000019, -0.00001239, -0.00000045, 0.00000229, 0.00000271, -0.00000010, -0.00000193, 0.00000284, -0.00000718, -0.00001486, -0.00002856, -0.00000716, 0.00003881, 0.00003399, 0.00001228, 0.00004336, 0.00001583, -0.00001370, -0.00002902, 0.00000378, -0.00001590, 0.00000733, -0.00000806, -0.00000385, -0.00000386, 0.00000701, -0.00000857, 0.00000760, -0.00000411, 0.00000206, -0.00000403, -0.00000482, -0.00000289, 0.00000355, 0.00000255, -0.00000363, -0.00000212, -0.00000318, 0.00000249, -0.00000448, 0.00000547, -0.00000831, -0.00000061, -0.00000491, 0.00000754, -0.00000393, 0.00000565, -0.00001187, -0.00000769, -0.00001232, 0.00000473, -0.00002684, 0.00001294, -0.00008556, -0.00003944, 0.00015048, 0.00012895, 0.00010518, -0.00012046, -0.00000973, -0.00003466, -0.00000300, -0.00001282, 0.00001021, -0.00000515, 0.00000268, -0.00000606, -0.00000244, -0.00000530, 0.00000006, -0.00000382, 0.00000188, -0.00000115, -0.00000377, -0.00000221, 0.00000800, -0.00000388, 0.00000440, -0.00000899, 0.00000457, -0.00000569, 0.00000671, 0.00000053, 0.00000385, -0.00000664, 0.00000311, -0.00000510, 0.00001220, -0.00000871, 0.00000532, 0.00000777, -0.00000283, 0.00000099, -0.00001029, 0.00001295, -0.00001192, -0.00000971, 0.00000011, -0.00000130, -0.00000071, -0.00001484, -0.00000173, 0.00000858, -0.00001976, },
		{ 0.00000474, 0.00000398, 0.00001071, -0.00001688, -0.00000054, 0.00000908, 0.00000377, -0.00000458, 0.00000873, -0.00000628, 0.00000366, -0.00000144, 0.00000325, -0.00000509, 0.00000729, -0.00000724, 0.00000067, -0.00000040, 0.00000366, -0.00000238, 0.00000749, -0.00001299, 0.00000765, -0.00000421, 0.00000101, -0.00001586, 0.00000755, -0.00000323, 0.00000845, -0.00000776, 0.00000609, -0.00000558, -0.00000310, -0.00002634, -0.00001905, 0.00000305, 0.00005018, 0.00001383, 0.00002631, 0.00003443, 0.00001071, -0.00002667, -0.00001496, -0.00000267, -0.00000933, 0.00000502, -0.00000845, -0.00000441, 0.00000143, 0.00000132, -0.00000498, 0.00000717, -0.00000544, 0.00000340, -0.00000706, -0.00000305, -0.00000210, 0.00000549, -0.00000036, -0.00000317, -0.00000324, -0.00000088, 0.00000082, -0.00000266, 0.00000456, -0.00001065, 0.00000167, -0.00000461, 0.00000879, -0.00000645, 0.00000783, -0.00001903, -0.00000335, -0.00001269, 0.00000387, -0.00002856, 0.00001370, -0.00012762, 0.00004707, 0.00014729, 0.00015654, 0.00001137, -0.00010995, 0.00000715, -0.00004216, 0.00000800, -0.00001683, 0.00001647, -0.00001206, 0.00000770, -0.00001196, 0.00000204, -0.00000825, 0.00000331, -0.00000608, 0.00000473, -0.00000451, -0.00000276, 0.00000076, 0.00000576, -0.00000289, 0.00000203, -0.00000803, 0.00000524, -0.00000617, 0.00000936, -0.00000141, 0.00000314, -0.00000727, 0.00000425, -0.00000451, 0.00001370, -0.00001563, 0.00001744, -0.00000342, 0.00000446, -0.00000846, -0.00000055, 0.00000838, -0.00001621, -0.00000565, 0.00000159, -0.00000233, -0.00000230, -0.00001867, 0.00001162, -0.00000674, -0.00000977, },
		{ -0.00000210, 0.00001403, -0.00000188, -0.00001363, 0.00000267, 0.00001177, -0.00000373, 0.00000250, 0.00000339, -0.00000383, 0.00000224, 0.00000090, -0.00000065, -0.00000072, 0.00000401, -0.00000695, 0.00000107, 0.00000185, 0.00000001, 0.00000280, 0.00000058, -0.00000870, 0.00000689, -0.00000393, -0.00000357, -0.00001144, 0.00000790, -0.00000194, 0.00000654, -0.00000769, 0.00000694, -0.00000816, -0.00000578, -0.00003126, -0.00000876, 0.00001636, 0.00004857, 0.00000183, 0.00003941, 0.00002221, 0.00000522, -0.00003459, -0.00000128, -0.00001151, 0.00000066, -0.00000196, -0.00000541, -0.00000560, 0.00000648, -0.00000623, 0.00000289, 0.00000153, -0.00000218, 0.00000074, -0.00000728, -0.00000221, -0.00000028, 0.00000591, -0.00000317, -0.00000214, -0.00000470, 0.00000266, -0.00000320, 0.00000266, -0.00000160, -0.00000753, -0.00000021, 0.00000030, 0.00000365, -0.00000158, 0.00000112, -0.00001782, -0.00000537, -0.00000547, -0.00000907, -0.00001186, -0.00001956, -0.00012529, 0.00012070, 0.00014747, 0.00015403, -0.00008648, -0.00005420, -0.00001231, -0.00002452, -0.00000030, -0.00000569, 0.00000857, -0.00000712, 0.00000238, -0.00000880, -0.00000071, -0.00000468, 0.00000064, -0.00000301, 0.00000256, -0.00000429, -0.00000330, 0.00000523, 0.00000064, 0.00000135, -0.00000437, -0.00000169, 0.00000027, -0.00000082, 0.00000656, 0.00000073, -0.00000204, -0.00000224, -0.00000041, 0.00000311, 0.00000527, -0.00001098, 0.00002026, -0.00000955, 0.00000826, -0.00001529, 0.00001030, -0.00000128, -0.00001637, -0.00000078, 0.00000031, -0.00000119, -0.00000883, -0.00001275, 0.00001527, -0.00001806, 0.00000248, },
		{ -0.00000280, 0.00001873, -0.00001436, -0.00000623, 0.00000463, 0.00001172, -0.00000898, 0.00000979, -0.00000419, 0.00000145, -0.00000112, 0.00000415, -0.00000541, 0.00000558, -0.00000262, -0.00000324, -0.00000009, 0.00000420, -0.00000350, 0.00000820, -0.00000840, 0.00000000, 0.00000139, -0.00000072, -0.00001061, -0.00000224, 0.00000278, 0.00000432, -0.00000153, -0.00000035, -0.00000005, -0.00000358, -0.00001738, -0.00002434, -0.00000296, 0.00003334, 0.00003236, 0.00000562, 0.00004174, 0.00001435, -0.00000540, -0.00003175, 0.00000403, -0.00001460, 0.00000638, -0.00000723, -0.00000333, -0.00000383, 0.00000644, -0.00000928, 0.00000869, -0.00000435, 0.00000230, -0.00000399, -0.00000519, -0.00000228, 0.00000244, 0.00000432, -0.00000471, -0.00000183, -0.00000466, 0.00000475, -0.00000634, 0.00000794, -0.00000986, -0.00000095, -0.00000444, 0.00000700, -0.00000435, 0.00000718, -0.00001186, -0.00000964, -0.00001163, 0.00000435, -0.00002597, 0.00001200, -0.00007817, -0.00007007, 0.00016171, 0.00016487, 0.00009291, -0.00013372, -0.00000220, -0.00003972, 0.00000094, -0.00001580, 0.00001103, -0.00000560, 0.00000336, -0.00000805, -0.00000064, -0.00000736, 0.00000234, -0.00000524, 0.00000298, -0.00000278, -0.00000159, -0.00000405, 0.00000859, -0.00000376, 0.00000462, -0.00000992, 0.00000515, -0.00000593, 0.00000658, 0.00000139, 0.00000374, -0.00000765, 0.00000415, -0.00000644, 0.00001285, -0.00000876, 0.00000379, 0.00001145, -0.00000642, 0.00000402, -0.00001388, 0.00001590, -0.00001183, -0.00001244, 0.00000275, -0.00000195, -0.00000004, -0.00001684, 0.00000112, 0.00000571, -0.00001663, 0.00000634, },
		{ 0.00000643, 0.00001163, -0.00001844, -0.00000093, 0.00000848, 0.00000590, -0.00000700, 0.00001105, -0.00000795, 0.00000460, -0.00000244, 0.00000445, -0.00000655, 0.00000877, -0.00000794, 0.00000052, -0.00000038, 0.00000391, -0.00000305, 0.00000858, -0.00001299, 0.00000664, -0.00000346, 0.00000051, -0.00001408, 0.00000475, -0.00000078, 0.00000809, -0.00000769, 0.00000609, -0.00000706, 0.00000019, -0.00002960, -0.00001089, 0.00000117, 0.00004717, 0.00001001, 0.00002154, 0.00003247, 0.00001210, -0.00001992, -0.00001929, -0.00000026, -0.00000953, 0.00000461, -0.00000748, -0.00000405, 0.00000133, 0.00000042, -0.00000499, 0.00000800, -0.00000581, 0.00000388, -0.00000749, -0.00000271, -0.00000234, 0.00000524, 0.00000093, -0.00000429, -0.00000320, -0.00000162, 0.00000269, -0.00000462, 0.00000782, -0.00001411, 0.00000356, -0.00000596, 0.00000953, -0.00000783, 0.00001114, -0.00002207, -0.00000237, -0.00001441, 0.00000637, -0.00003148, 0.00001904, -0.00013063, 0.00001679, 0.00017900, 0.00017998, -0.00001221, -0.00011063, 0.00000960, -0.00004473, 0.00001083, -0.00001985, 0.00001764, -0.00001233, 0.00000749, -0.00001294, 0.00000322, -0.00000984, 0.00000549, -0.00000769, 0.00000584, -0.00000603, -0.00000030, -0.00000211, 0.00000774, -0.00000382, 0.00000278, -0.00000952, 0.00000668, -0.00000740, 0.00001028, -0.00000119, 0.00000295, -0.00000814, 0.00000553, -0.00000667, 0.00001584, -0.00001790, 0.00001892, -0.00000247, 0.00000237, -0.00000618, -0.00000406, 0.00001289, -0.00001911, -0.00000562, 0.00000295, -0.00000232, -0.00000292, -0.00001896, 0.00001330, -0.00000951, -0.00000504, 0.00000028, },
		{ 0.00001861, -0.00000362, -0.00001347, 0.00000081, 0.00001332, -0.00000380, 0.00000177, 0.00000467, -0.00000513, 0.00000299, 0.00000019, 0.00000028, -0.00000208, 0.00000584, -0.00000849, 0.00000174, 0.00000133, 0.00000059, 0.00000185, 0.00000242, -0.00001001, 0.00000710, -0.00000414, -0.00000280, -0.00001098, 0.00000554, 0.00000143, 0.00000450, -0.00000585, 0.00000457, -0.00000651, -0.00000557, -0.00003114, -0.00000139, 0.00001093, 0.00004770, -0.00000415, 0.00003707, 0.00001955, 0.00000984, -0.00003115, -0.00000513, -0.00000799, -0.00000108, -0.00000128, -0.00000481, -0.00000513, 0.00000586, -0.00000689, 0.00000327, 0.00000205, -0.00000248, 0.00000119, -0.00000781, -0.00000144, -0.00000132, 0.00000683, -0.00000300, -0.00000267, -0.00000537, 0.00000338, -0.00000302, 0.00000242, 0.00000019, -0.00001061, 0.00000212, -0.00000162, 0.00000457, -0.00000238, 0.00000405, -0.00002185, -0.00000286, -0.00000843, -0.00000501, -0.00001707, -0.00000884, -0.00014293, 0.00010210, 0.00019410, 0.00015491, -0.00010793, -0.00004734, -0.00001532, -0.00002293, -0.00000065, -0.00000685, 0.00000879, -0.00000644, 0.00000057, -0.00000768, -0.00000151, -0.00000436, 0.00000122, -0.00000340, 0.00000235, -0.00000416, -0.00000236, 0.00000313, 0.00000262, 0.00000027, -0.00000385, -0.00000282, 0.00000164, -0.00000224, 0.00000801, 0.00000046, -0.00000249, -0.00000234, 0.00000001, 0.00000154, 0.00000732, -0.00001389, 0.00002368, -0.00001139, 0.00000859, -0.00001506, 0.00000869, 0.00000254, -0.00002049, 0.00000145, 0.00000013, -0.00000034, -0.00001080, -0.00001062, 0.00001393, -0.00001778, 0.00000529, -0.00000678, },
		{ 0.00002236, -0.00001658, -0.00000582, 0.00000266, 0.00001424, -0.00001061, 0.00001071, -0.00000440, 0.00000116, -0.00000112, 0.00000426, -0.00000536, 0.00000512, -0.00000140, -0.00000475, 0.00000082, 0.00000351, -0.00000286, 0.00000738, -0.00000640, -0.00000219, 0.00000272, -0.00000199, -0.00000816, -0.00000427, 0.00000284, 0.00000635, -0.00000320, 0.00000138, -0.00000302, -0.00000001, -0.00001995, -0.00001934, -0.00000032, 0.00002817, 0.00003140, -0.00000080, 0.00004092, 0.00001178, 0.00000159, -0.00003227, 0.00000252, -0.00001167, 0.00000398, -0.00000556, -0.00000325, -0.00000321, 0.00000519, -0.00000926, 0.00000897, -0.00000391, 0.00000201, -0.00000365, -0.00000551, -0.00000165, 0.00000121, 0.00000596, -0.00000553, -0.00000181, -0.00000569, 0.00000649, -0.00000772, 0.00000980, -0.00001063, -0.00000201, -0.00000328, 0.00000573, -0.00000411, 0.00000801, -0.00001116, -0.00001227, -0.00000979, 0.00000225, -0.00002278, 0.00000777, -0.00006642, -0.00010226, 0.00016668, 0.00021164, 0.00007160, -0.00014215, 0.00000299, -0.00004315, 0.00000373, -0.00001773, 0.00001088, -0.00000549, 0.00000377, -0.00001004, 0.00000143, -0.00000941, 0.00000415, -0.00000610, 0.00000371, -0.00000419, 0.00000029, -0.00000528, 0.00000833, -0.00000280, 0.00000405, -0.00001007, 0.00000488, -0.00000526, 0.00000560, 0.00000291, 0.00000300, -0.00000803, 0.00000460, -0.00000707, 0.00001254, -0.00000792, 0.00000185, 0.00001501, -0.00000985, 0.00000642, -0.00001590, 0.00001706, -0.00001056, -0.00001547, 0.00000542, -0.00000263, 0.00000060, -0.00001867, 0.00000415, 0.00000210, -0.00001264, 0.00000496, -0.00000448, },
		{ 0.00001264, -0.00001947, -0.00000161, 0.00000788, 0.00000771, -0.00000869, 0.00001251, -0.00000902, 0.00000505, -0.00000301, 0.00000525, -0.00000737, 0.00000943, -0.00000790, -0.00000022, 0.00000008, 0.00000362, -0.00000289, 0.00000853, -0.00001175, 0.00000459, -0.00000193, -0.00000086, -0.00001093, 0.00000065, 0.00000233, 0.00000727, -0.00000710, 0.00000549, -0.00000783, 0.00000253, -0.00003160, -0.00000353, -0.00000129, 0.00004433, 0.00000720, 0.00001651, 0.00003177, 0.00001096, -0.00001234, -0.00002311, 0.00000152, -0.00000891, 0.00000315, -0.00000579, -0.00000396, 0.00000150, -0.00000081, -0.00000456, 0.00000818, -0.00000548, 0.00000363, -0.00000728, -0.00000261, -0.00000237, 0.00000464, 0.00000239, -0.00000543, -0.00000317, -0.00000215, 0.00000418, -0.00000620, 0.00001054, -0.00001675, 0.00000457, -0.00000640, 0.00000925, -0.00000817, 0.00001329, -0.00002393, -0.00000242, -0.00001474, 0.00000684, -0.00003145, 0.00001963, -0.00012535, -0.00002278, 0.00021383, 0.00020757, -0.00004261, -0.00010583, 0.00000845, -0.00004451, 0.00001152, -0.00002099, 0.00001697, -0.00001111, 0.00000612, -0.00001312, 0.00000417, -0.00001131, 0.00000721, -0.00000867, 0.00000645, -0.00000727, 0.00000204, -0.00000466, 0.00000912, -0.00000409, 0.00000287, -0.00001030, 0.00000732, -0.00000769, 0.00001018, -0.00000009, 0.00000189, -0.00000810, 0.00000592, -0.00000781, 0.00001656, -0.00001843, 0.00001878, -0.00000034, -0.00000095, -0.00000293, -0.00000742, 0.00001636, -0.00002089, -0.00000608, 0.00000439, -0.00000229, -0.00000369, -0.00001871, 0.00001445, -0.00001227, 0.00000015, -0.00000432, 0.00000845, },
		{ -0.00000451, -0.00001350, -0.00000085, 0.00001430, -0.00000338, 0.00000096, 0.00000567, -0.00000621, 0.00000360, -0.00000048, 0.00000129, -0.00000333, 0.00000731, -0.00000960, 0.00000209, 0.00000091, 0.00000118, 0.00000107, 0.00000372, -0.00001049, 0.00000647, -0.00000379, -0.00000253, -0.00000945, 0.00000182, 0.00000574, 0.00000181, -0.00000341, 0.00000160, -0.00000417, -0.00000633, -0.00002938, 0.00000410, 0.00000599, 0.00004670, -0.00000895, 0.00003431, 0.00001798, 0.00001139, -0.00002530, -0.00001013, -0.00000366, -0.00000331, -0.00000063, -0.00000412, -0.00000439, 0.00000490, -0.00000722, 0.00000337, 0.00000256, -0.00000259, 0.00000123, -0.00000777, -0.00000097, -0.00000213, 0.00000729, -0.00000241, -0.00000352, -0.00000564, 0.00000378, -0.00000267, 0.00000198, 0.00000208, -0.00001348, 0.00000414, -0.00000318, 0.00000503, -0.00000270, 0.00000636, -0.00002519, -0.00000069, -0.00001110, -0.00000146, -0.00002131, -0.00000057, -0.00015322, 0.00007077, 0.00025191, 0.00015079, -0.00012939, -0.00003893, -0.00001961, -0.00002016, -0.00000190, -0.00000718, 0.00000798, -0.00000478, -0.00000211, -0.00000585, -0.00000244, -0.00000428, 0.00000187, -0.00000366, 0.00000202, -0.00000410, -0.00000108, 0.00000069, 0.00000475, -0.00000089, -0.00000325, -0.00000401, 0.00000298, -0.00000347, 0.00000909, 0.00000051, -0.00000321, -0.00000212, 0.00000014, 0.00000033, 0.00000860, -0.00001545, 0.00002543, -0.00001168, 0.00000706, -0.00001270, 0.00000569, 0.00000658, -0.00002411, 0.00000333, 0.00000002, 0.00000051, -0.00001284, -0.00000795, 0.00001169, -0.00001678, 0.00000793, -0.00001124, 0.00002231, },
		{ -0.00001757, -0.00000610, 0.00000124, 0.00001580, -0.00001110, 0.00001073, -0.00000413, 0.00000049, -0.00000074, 0.00000390, -0.00000464, 0.00000404, 0.00000019, -0.00000645, 0.00000191, 0.00000241, -0.00000156, 0.00000582, -0.00000389, -0.00000454, 0.00000402, -0.00000344, -0.00000525, -0.00000631, 0.00000238, 0.00000873, -0.00000495, 0.00000306, -0.00000580, 0.00000326, -0.00002227, -0.00001387, 0.00000083, 0.00002364, 0.00003063, -0.00000655, 0.00004023, 0.00000913, 0.00000646, -0.00003030, -0.00000073, -0.00000728, 0.00000033, -0.00000318, -0.00000353, -0.00000207, 0.00000334, -0.00000853, 0.00000846, -0.00000287, 0.00000130, -0.00000316, -0.00000557, -0.00000119, 0.00000008, 0.00000722, -0.00000586, -0.00000228, -0.00000603, 0.00000743, -0.00000834, 0.00001076, -0.00001041, -0.00000385, -0.00000142, 0.00000375, -0.00000318, 0.00000805, -0.00000975, -0.00001553, -0.00000683, -0.00000150, -0.00001741, 0.00000062, -0.00005137, -0.00013391, 0.00016318, 0.00027061, 0.00004020, -0.00014468, 0.00000501, -0.00004433, 0.00000490, -0.00001824, 0.00000949, -0.00000460, 0.00000367, -0.00001171, 0.00000341, -0.00001106, 0.00000511, -0.00000612, 0.00000380, -0.00000508, 0.00000162, -0.00000567, 0.00000708, -0.00000095, 0.00000263, -0.00000932, 0.00000369, -0.00000368, 0.00000383, 0.00000502, 0.00000166, -0.00000774, 0.00000439, -0.00000692, 0.00001132, -0.00000635, -0.00000024, 0.00001815, -0.00001273, 0.00000777, -0.00001596, 0.00001623, -0.00000812, -0.00001867, 0.00000796, -0.00000324, 0.00000108, -0.00002017, 0.00000716, -0.00000203, -0.00000791, 0.00000325, -0.00000600, 0.00002512, },
		{ -0.00001974, -0.00000280, 0.00000752, 0.00000891, -0.00000935, 0.00001286, -0.00000927, 0.00000487, -0.00000302, 0.00000545, -0.00000735, 0.00000912, -0.00000701, -0.00000155, 0.00000098, 0.00000279, -0.00000186, 0.00000733, -0.00000934, 0.00000163, 0.00000024, -0.00000294, -0.00000661, -0.00000446, 0.00000583, 0.00000612, -0.00000599, 0.00000423, -0.00000771, 0.00000363, -0.00003198, 0.00000247, -0.00000375, 0.00004136, 0.00000543, 0.00001118, 0.00003217, 0.00000801, -0.00000485, -0.00002575, 0.00000232, -0.00000737, 0.00000068, -0.00000344, -0.00000411, 0.00000189, -0.00000231, -0.00000372, 0.00000769, -0.00000446, 0.00000270, -0.00000653, -0.00000261, -0.00000235, 0.00000389, 0.00000376, -0.00000629, -0.00000341, -0.00000218, 0.00000494, -0.00000696, 0.00001218, -0.00001801, 0.00000424, -0.00000553, 0.00000758, -0.00000706, 0.00001376, -0.00002404, -0.00000407, -0.00001305, 0.00000454, -0.00002762, 0.00001453, -0.00011127, -0.00007092, 0.00025027, 0.00024038, -0.00008063, -0.00009464, 0.00000280, -0.00004072, 0.00000941, -0.00001966, 0.00001402, -0.00000806, 0.00000329, -0.00001215, 0.00000449, -0.00001217, 0.00000796, -0.00000860, 0.00000615, -0.00000779, 0.00000375, -0.00000642, 0.00000952, -0.00000342, 0.00000203, -0.00001007, 0.00000684, -0.00000675, 0.00000887, 0.00000202, -0.00000013, -0.00000697, 0.00000519, -0.00000765, 0.00001559, -0.00001707, 0.00001697, 0.00000295, -0.00000533, 0.00000099, -0.00001027, 0.00001845, -0.00002137, -0.00000704, 0.00000587, -0.00000219, -0.00000469, -0.00001781, 0.00001489, -0.00001470, 0.00000541, -0.00000876, 0.00000997, 0.00001356, },
		{ -0.00001373, -0.00000229, 0.00001472, -0.00000262, 0.00000030, 0.00000616, -0.00000685, 0.00000391, -0.00000095, 0.00000217, -0.00000423, 0.00000814, -0.00001004, 0.00000194, 0.00000073, 0.00000161, 0.00000065, 0.00000429, -0.00001000, 0.00000494, -0.00000282, -0.00000278, -0.00000689, -0.00000307, 0.00001068, -0.00000128, -0.00000056, -0.00000183, -0.00000125, -0.00000804, -0.00002590, 0.00000732, 0.00000220, 0.00004497, -0.00001213, 0.00003062, 0.00001796, 0.00000997, -0.00001763, -0.00001567, 0.00000098, -0.00000570, -0.00000016, -0.00000326, -0.00000350, 0.00000367, -0.00000722, 0.00000323, 0.00000301, -0.00000242, 0.00000078, -0.00000711, -0.00000084, -0.00000268, 0.00000731, -0.00000148, -0.00000458, -0.00000561, 0.00000399, -0.00000231, 0.00000159, 0.00000371, -0.00001568, 0.00000536, -0.00000395, 0.00000462, -0.00000212, 0.00000751, -0.00002722, 0.00000044, -0.00001266, 0.00000058, -0.00002330, 0.00000351, -0.00015422, 0.00002574, 0.00032068, 0.00014189, -0.00015083, -0.00002884, -0.00002544, -0.00001594, -0.00000434, -0.00000639, 0.00000592, -0.00000205, -0.00000569, -0.00000327, -0.00000358, -0.00000427, 0.00000235, -0.00000358, 0.00000137, -0.00000387, 0.00000022, -0.00000173, 0.00000668, -0.00000185, -0.00000280, -0.00000498, 0.00000397, -0.00000416, 0.00000951, 0.00000106, -0.00000434, -0.00000141, -0.00000023, -0.00000024, 0.00000876, -0.00001530, 0.00002520, -0.00001021, 0.00000357, -0.00000827, 0.00000154, 0.00001055, -0.00002695, 0.00000471, 0.00000006, 0.00000129, -0.00001485, -0.00000482, 0.00000855, -0.00001491, 0.00001002, -0.00001504, 0.00002476, -0.00000440, },
		{ -0.00000723, 0.00000052, 0.00001623, -0.00001037, 0.00000986, -0.00000345, -0.00000045, -0.00000005, 0.00000313, -0.00000333, 0.00000247, 0.00000194, -0.00000810, 0.00000296, 0.00000111, 0.00000021, 0.00000378, -0.00000118, -0.00000674, 0.00000505, -0.00000484, -0.00000213, -0.00000813, 0.00000133, 0.00001134, -0.00000660, 0.00000448, -0.00000818, 0.00000598, -0.00002409, -0.00000822, 0.00000061, 0.00002001, 0.00002956, -0.00001115, 0.00003903, 0.00000725, 0.00000868, -0.00002586, -0.00000539, -0.00000184, -0.00000421, -0.00000037, -0.00000400, -0.00000060, 0.00000104, -0.00000721, 0.00000727, -0.00000138, 0.00000034, -0.00000274, -0.00000521, -0.00000109, -0.00000072, 0.00000788, -0.00000552, -0.00000338, -0.00000552, 0.00000741, -0.00000801, 0.00001063, -0.00000912, -0.00000644, 0.00000103, 0.00000117, -0.00000162, 0.00000732, -0.00000769, -0.00001926, -0.00000296, -0.00000662, -0.00001030, -0.00000869, -0.00003445, -0.00016272, 0.00014924, 0.00034255, -0.00000172, -0.00014067, 0.00000330, -0.00004285, 0.00000413, -0.00001712, 0.00000673, -0.00000280, 0.00000285, -0.00001278, 0.00000497, -0.00001196, 0.00000496, -0.00000514, 0.00000310, -0.00000528, 0.00000223, -0.00000514, 0.00000488, 0.00000169, 0.00000044, -0.00000769, 0.00000161, -0.00000131, 0.00000144, 0.00000751, -0.00000013, -0.00000682, 0.00000355, -0.00000605, 0.00000932, -0.00000429, -0.00000217, 0.00002051, -0.00001473, 0.00000774, -0.00001389, 0.00001341, -0.00000468, -0.00002179, 0.00001018, -0.00000362, 0.00000130, -0.00002118, 0.00000994, -0.00000643, -0.00000263, 0.00000114, -0.00000704, 0.00002663, -0.00001707, },
		{ -0.00000462, 0.00000758, 0.00000927, -0.00000877, 0.00001195, -0.00000863, 0.00000401, -0.00000242, 0.00000498, -0.00000639, 0.00000777, -0.00000532, -0.00000340, 0.00000223, 0.00000148, -0.00000006, 0.00000510, -0.00000599, -0.00000196, 0.00000277, -0.00000546, -0.00000148, -0.00001012, 0.00000936, 0.00000482, -0.00000445, 0.00000232, -0.00000667, 0.00000340, -0.00003053, 0.00000670, -0.00000574, 0.00003806, 0.00000466, 0.00000569, 0.00003328, 0.00000410, 0.00000168, -0.00002673, 0.00000198, -0.00000494, -0.00000262, -0.00000058, -0.00000437, 0.00000238, -0.00000395, -0.00000256, 0.00000663, -0.00000288, 0.00000126, -0.00000545, -0.00000251, -0.00000250, 0.00000327, 0.00000472, -0.00000658, -0.00000422, -0.00000140, 0.00000463, -0.00000650, 0.00001229, -0.00001744, 0.00000219, -0.00000306, 0.00000426, -0.00000418, 0.00001212, -0.00002195, -0.00000777, -0.00000886, -0.00000103, -0.00001946, 0.00000330, -0.00008854, -0.00012619, 0.00028642, 0.00027945, -0.00012678, -0.00007642, -0.00000802, -0.00003279, 0.00000396, -0.00001543, 0.00000846, -0.00000293, -0.00000127, -0.00000972, 0.00000377, -0.00001195, 0.00000728, -0.00000707, 0.00000457, -0.00000718, 0.00000442, -0.00000698, 0.00000863, -0.00000158, 0.00000009, -0.00000861, 0.00000502, -0.00000446, 0.00000630, 0.00000512, -0.00000311, -0.00000472, 0.00000324, -0.00000606, 0.00001286, -0.00001381, 0.00001362, 0.00000716, -0.00001048, 0.00000517, -0.00001223, 0.00001895, -0.00002049, -0.00000844, 0.00000728, -0.00000194, -0.00000595, -0.00001617, 0.00001443, -0.00001648, 0.00001034, -0.00001278, 0.00001099, 0.00001419, -0.00001903, },
		{ -0.00000357, 0.00001466, -0.00000167, -0.00000002, 0.00000593, -0.00000688, 0.00000378, -0.00000110, 0.00000275, -0.00000455, 0.00000810, -0.00000962, 0.00000117, 0.00000089, 0.00000178, 0.00000071, 0.00000402, -0.00000851, 0.00000257, -0.00000132, -0.00000349, -0.00000348, -0.00000875, 0.00001583, -0.00000441, 0.00000245, -0.00000549, 0.00000207, -0.00001057, -0.00002078, 0.00000813, 0.00000001, 0.00004203, -0.00001340, 0.00002573, 0.00001959, 0.00000605, -0.00000896, -0.00002101, 0.00000538, -0.00000792, -0.00000003, -0.00000218, -0.00000249, 0.00000221, -0.00000691, 0.00000287, 0.00000333, -0.00000196, -0.00000013, -0.00000590, -0.00000097, -0.00000305, 0.00000701, -0.00000039, -0.00000564, -0.00000549, 0.00000420, -0.00000218, 0.00000155, 0.00000465, -0.00001671, 0.00000531, -0.00000349, 0.00000298, -0.00000026, 0.00000699, -0.00002730, -0.00000014, -0.00001234, 0.00000016, -0.00002188, 0.00000191, -0.00014461, -0.00003304, 0.00039933, 0.00012897, -0.00017248, -0.00001685, -0.00003312, -0.00000996, -0.00000828, -0.00000420, 0.00000243, 0.00000185, -0.00001020, 0.00000010, -0.00000506, -0.00000407, 0.00000236, -0.00000288, 0.00000015, -0.00000318, 0.00000119, -0.00000370, 0.00000800, -0.00000229, -0.00000278, -0.00000542, 0.00000426, -0.00000402, 0.00000904, 0.00000225, -0.00000596, -0.00000010, -0.00000128, 0.00000006, 0.00000757, -0.00001327, 0.00002291, -0.00000695, -0.00000178, -0.00000207, -0.00000337, 0.00001406, -0.00002876, 0.00000546, 0.00000030, 0.00000195, -0.00001676, -0.00000131, 0.00000459, -0.00001206, 0.00001126, -0.00001779, 0.00002574, -0.00000330, -0.00001408, },
		{ 0.00000054, 0.00001553, -0.00000849, 0.00000821, -0.00000250, -0.00000150, 0.00000081, 0.00000207, -0.00000160, 0.00000064, 0.00000357, -0.00000940, 0.00000373, -0.00000019, 0.00000220, 0.00000155, 0.00000142, -0.00000850, 0.00000558, -0.00000598, 0.00000093, -0.00000955, -0.00000031, 0.00001401, -0.00000795, 0.00000548, -0.00000995, 0.00000796, -0.00002519, -0.00000267, -0.00000082, 0.00001748, 0.00002780, -0.00001420, 0.00003679, 0.00000674, 0.00000811, -0.00001931, -0.00001094, 0.00000409, -0.00000914, 0.00000252, -0.00000440, 0.00000098, -0.00000149, -0.00000549, 0.00000560, 0.00000036, -0.00000064, -0.00000257, -0.00000426, -0.00000148, -0.00000105, 0.00000781, -0.00000444, -0.00000515, -0.00000414, 0.00000637, -0.00000668, 0.00000940, -0.00000683, -0.00000958, 0.00000383, -0.00000179, 0.00000042, 0.00000595, -0.00000519, -0.00002315, 0.00000144, -0.00001261, -0.00000215, -0.00001915, -0.00001736, -0.00018646, 0.00012341, 0.00042745, -0.00005383, -0.00013004, -0.00000231, -0.00003861, 0.00000136, -0.00001433, 0.00000262, -0.00000008, 0.00000121, -0.00001304, 0.00000582, -0.00001186, 0.00000357, -0.00000312, 0.00000159, -0.00000474, 0.00000206, -0.00000375, 0.00000188, 0.00000490, -0.00000235, -0.00000535, -0.00000116, 0.00000162, -0.00000129, 0.00001011, -0.00000218, -0.00000541, 0.00000219, -0.00000462, 0.00000676, -0.00000204, -0.00000363, 0.00002180, -0.00001556, 0.00000614, -0.00000973, 0.00000884, -0.00000056, -0.00002453, 0.00001186, -0.00000364, 0.00000113, -0.00002154, 0.00001227, -0.00001083, 0.00000299, -0.00000138, -0.00000738, 0.00002662, -0.00001496, -0.00000922, },
		{ 0.00000813, 0.00000868, -0.00000690, 0.00000978, -0.00000712, 0.00000254, -0.00000126, 0.00000387, -0.00000453, 0.00000549, -0.00000298, -0.00000555, 0.00000365, -0.00000011, 0.00000232, 0.00000213, -0.00000203, -0.00000581, 0.00000532, -0.00000809, 0.00000402, -0.00001584, 0.00001254, 0.00000356, -0.00000257, -0.00000015, -0.00000474, 0.00000183, -0.00002722, 0.00000894, -0.00000694, 0.00003433, 0.00000469, 0.00000031, 0.00003459, 0.00000010, 0.00000653, -0.00002579, 0.00000052, -0.00000187, -0.00000645, 0.00000247, -0.00000455, 0.00000277, -0.00000550, -0.00000130, 0.00000521, -0.00000098, -0.00000041, -0.00000432, -0.00000205, -0.00000307, 0.00000304, 0.00000497, -0.00000601, -0.00000583, 0.00000042, 0.00000298, -0.00000454, 0.00001053, -0.00001475, -0.00000177, 0.00000116, -0.00000081, 0.00000061, 0.00000815, -0.00001739, -0.00001374, -0.00000196, -0.00001007, -0.00000689, -0.00001394, -0.00005804, -0.00018652, 0.00032009, 0.00032566, -0.00018117, -0.00005093, -0.00002432, -0.00002042, -0.00000508, -0.00000808, 0.00000016, 0.00000436, -0.00000767, -0.00000557, 0.00000164, -0.00001022, 0.00000481, -0.00000382, 0.00000147, -0.00000516, 0.00000370, -0.00000607, 0.00000627, 0.00000150, -0.00000303, -0.00000583, 0.00000181, -0.00000082, 0.00000255, 0.00000905, -0.00000690, -0.00000141, 0.00000012, -0.00000310, 0.00000848, -0.00000891, 0.00000903, 0.00001196, -0.00001596, 0.00000918, -0.00001300, 0.00001775, -0.00001832, -0.00001010, 0.00000849, -0.00000146, -0.00000754, -0.00001375, 0.00001297, -0.00001735, 0.00001459, -0.00001618, 0.00001153, 0.00001436, -0.00001723, -0.00000711, },
		{ 0.00001425, -0.00000071, 0.00000018, 0.00000487, -0.00000620, 0.00000313, -0.00000085, 0.00000290, -0.00000416, 0.00000709, -0.00000828, -0.00000023, 0.00000140, 0.00000163, 0.00000130, 0.00000294, -0.00000612, -0.00000048, 0.00000055, -0.00000448, 0.00000047, -0.00001477, 0.00002069, -0.00000723, 0.00000532, -0.00000911, 0.00000558, -0.00001370, -0.00001426, 0.00000656, -0.00000033, 0.00003767, -0.00001268, 0.00001964, 0.00002266, 0.00000036, -0.00000020, -0.00002543, 0.00000904, -0.00000969, -0.00000031, -0.00000087, -0.00000141, 0.00000056, -0.00000628, 0.00000229, 0.00000354, -0.00000124, -0.00000139, -0.00000429, -0.00000120, -0.00000339, 0.00000658, 0.00000060, -0.00000642, -0.00000553, 0.00000467, -0.00000255, 0.00000216, 0.00000450, -0.00001613, 0.00000360, -0.00000146, -0.00000018, 0.00000319, 0.00000439, -0.00002494, -0.00000301, -0.00000946, -0.00000349, -0.00001613, -0.00000637, -0.00012380, -0.00010464, 0.00048587, 0.00011337, -0.00019476, -0.00000266, -0.00004296, -0.00000193, -0.00001399, -0.00000036, -0.00000265, 0.00000696, -0.00001565, 0.00000433, -0.00000708, -0.00000336, 0.00000155, -0.00000126, -0.00000189, -0.00000173, 0.00000142, -0.00000481, 0.00000834, -0.00000192, -0.00000342, -0.00000507, 0.00000360, -0.00000280, 0.00000751, 0.00000415, -0.00000809, 0.00000186, -0.00000307, 0.00000135, 0.00000492, -0.00000935, 0.00001867, -0.00000211, -0.00000862, 0.00000536, -0.00000851, 0.00001672, -0.00002930, 0.00000550, 0.00000079, 0.00000244, -0.00001845, 0.00000242, -0.00000008, -0.00000823, 0.00001142, -0.00001925, 0.00002516, -0.00000131, -0.00001441, -0.00000474, },
		{ 0.00001380, -0.00000566, 0.00000602, -0.00000151, -0.00000248, 0.00000166, 0.00000090, 0.00000033, -0.00000116, 0.00000480, -0.00001011, 0.00000401, -0.00000124, 0.00000413, -0.00000057, 0.00000359, -0.00000958, 0.00000543, -0.00000667, 0.00000372, -0.00001041, -0.00000246, 0.00001651, -0.00000879, 0.00000590, -0.00001098, 0.00000904, -0.00002540, 0.00000247, -0.00000322, 0.00001613, 0.00002508, -0.00001544, 0.00003320, 0.00000787, 0.00000495, -0.00001130, -0.00001666, 0.00000983, -0.00001387, 0.00000511, -0.00000448, 0.00000242, -0.00000399, -0.00000357, 0.00000365, 0.00000210, -0.00000144, -0.00000279, -0.00000265, -0.00000244, -0.00000084, 0.00000695, -0.00000262, -0.00000750, -0.00000197, 0.00000442, -0.00000444, 0.00000717, -0.00000375, -0.00001297, 0.00000666, -0.00000482, 0.00000267, 0.00000415, -0.00000254, -0.00002682, 0.00000587, -0.00001881, 0.00000616, -0.00002957, -0.00000181, -0.00020318, 0.00008494, 0.00052434, -0.00011499, -0.00011333, -0.00001155, -0.00003183, -0.00000324, -0.00001006, -0.00000266, 0.00000343, -0.00000123, -0.00001237, 0.00000584, -0.00001069, 0.00000098, -0.00000020, -0.00000062, -0.00000352, 0.00000123, -0.00000168, -0.00000163, 0.00000838, -0.00000543, -0.00000254, -0.00000433, 0.00000476, -0.00000401, 0.00001251, -0.00000424, -0.00000370, 0.00000050, -0.00000282, 0.00000392, 0.00000008, -0.00000433, 0.00002181, -0.00001508, 0.00000300, -0.00000374, 0.00000295, 0.00000380, -0.00002653, 0.00001277, -0.00000315, 0.00000046, -0.00002112, 0.00001394, -0.00001491, 0.00000870, -0.00000431, -0.00000688, 0.00002501, -0.00001134, -0.00001191, 0.00000120, },
		{ 0.00000718, -0.00000386, 0.00000651, -0.00000493, 0.00000062, 0.00000032, 0.00000223, -0.00000195, 0.00000250, -0.00000027, -0.00000774, 0.00000498, -0.00000176, 0.00000497, -0.00000122, 0.00000211, -0.00000951, 0.00000752, -0.00001042, 0.00000938, -0.00002110, 0.00001501, 0.00000253, -0.00000047, -0.00000303, -0.00000206, -0.00000091, -0.00002219, 0.00000915, -0.00000717, 0.00003016, 0.00000531, -0.00000461, 0.00003556, -0.00000327, 0.00000924, -0.00002292, -0.00000182, 0.00000148, -0.00001033, 0.00000535, -0.00000436, 0.00000280, -0.00000671, -0.00000016, 0.00000368, 0.00000093, -0.00000198, -0.00000346, -0.00000099, -0.00000427, 0.00000347, 0.00000425, -0.00000435, -0.00000842, 0.00000341, -0.00000013, -0.00000092, 0.00000673, -0.00000985, -0.00000762, 0.00000704, -0.00000754, 0.00000727, 0.00000182, -0.00001035, -0.00002196, 0.00000759, -0.00002240, 0.00000974, -0.00003649, -0.00002129, -0.00024931, 0.00034902, 0.00037954, -0.00024341, -0.00001845, -0.00004602, -0.00000366, -0.00001771, 0.00000237, -0.00001080, 0.00001371, -0.00001590, 0.00000040, -0.00000215, -0.00000669, 0.00000032, 0.00000127, -0.00000327, -0.00000156, 0.00000143, -0.00000355, 0.00000240, 0.00000576, -0.00000722, -0.00000181, -0.00000267, 0.00000398, -0.00000214, 0.00001355, -0.00001125, 0.00000273, -0.00000397, 0.00000101, 0.00000276, -0.00000274, 0.00000366, 0.00001687, -0.00002127, 0.00001255, -0.00001239, 0.00001492, -0.00001508, -0.00001180, 0.00000931, -0.00000064, -0.00000947, -0.00001051, 0.00001042, -0.00001711, 0.00001787, -0.00001878, 0.00001163, 0.00001396, -0.00001431, -0.00001019, 0.00000912, },
		{ 0.00000012, 0.00000097, 0.00000295, -0.00000480, 0.00000198, -0.00000018, 0.00000258, -0.00000300, 0.00000510, -0.00000609, -0.00000217, 0.00000222, 0.00000120, 0.00000234, 0.00000116, -0.00000304, -0.00000393, 0.00000254, -0.00000549, 0.00000458, -0.00002057, 0.00002475, -0.00000936, 0.00000780, -0.00001244, 0.00000899, -0.00001712, -0.00000671, 0.00000287, 0.00000118, 0.00003184, -0.00001011, 0.00001259, 0.00002670, -0.00000622, 0.00000772, -0.00002831, 0.00001159, -0.00001083, -0.00000098, 0.00000059, -0.00000022, -0.00000132, -0.00000531, 0.00000146, 0.00000371, -0.00000043, -0.00000279, -0.00000255, -0.00000130, -0.00000390, 0.00000627, 0.00000120, -0.00000662, -0.00000604, 0.00000567, -0.00000366, 0.00000370, 0.00000293, -0.00001363, -0.00000003, 0.00000233, -0.00000502, 0.00000840, -0.00000054, -0.00001979, -0.00000857, -0.00000358, -0.00001087, -0.00000556, -0.00002175, -0.00009212, -0.00018710, 0.00057747, 0.00009684, -0.00021827, 0.00001399, -0.00005519, 0.00000834, -0.00002167, 0.00000532, -0.00000942, 0.00001328, -0.00002201, 0.00000947, -0.00000983, -0.00000183, -0.00000040, 0.00000150, -0.00000496, 0.00000074, 0.00000059, -0.00000469, 0.00000738, -0.00000051, -0.00000493, -0.00000371, 0.00000178, -0.00000040, 0.00000489, 0.00000674, -0.00001068, 0.00000441, -0.00000559, 0.00000359, 0.00000091, -0.00000377, 0.00001280, 0.00000388, -0.00001638, 0.00001334, -0.00001330, 0.00001819, -0.00002845, 0.00000481, 0.00000151, 0.00000273, -0.00001984, 0.00000623, -0.00000527, -0.00000349, 0.00001039, -0.00001930, 0.00002309, 0.00000136, -0.00001455, -0.00000590, 0.00001358, },
		{ -0.00000217, 0.00000358, -0.00000071, -0.00000317, 0.00000233, -0.00000021, 0.00000223, -0.00000268, 0.00000539, -0.00001003, 0.00000362, -0.00000188, 0.00000576, -0.00000232, 0.00000511, -0.00000980, 0.00000453, -0.00000682, 0.00000610, -0.00001069, -0.00000494, 0.00001858, -0.00000892, 0.00000563, -0.00001117, 0.00000916, -0.00002459, 0.00000696, -0.00000630, 0.00001589, 0.00002130, -0.00001480, 0.00002817, 0.00001060, -0.00000025, -0.00000266, -0.00002175, 0.00001466, -0.00001784, 0.00000700, -0.00000400, 0.00000350, -0.00000625, -0.00000167, 0.00000168, 0.00000364, -0.00000189, -0.00000348, -0.00000042, -0.00000390, -0.00000010, 0.00000538, -0.00000021, -0.00001025, 0.00000077, 0.00000176, -0.00000149, 0.00000419, -0.00000023, -0.00001622, 0.00000914, -0.00000758, 0.00000484, 0.00000221, -0.00000006, -0.00002981, 0.00000977, -0.00002451, 0.00001373, -0.00003874, 0.00001059, -0.00021139, 0.00003396, 0.00063128, -0.00018320, -0.00009168, -0.00002373, -0.00002303, -0.00000924, -0.00000469, -0.00000874, 0.00000745, -0.00000435, -0.00001080, 0.00000498, -0.00000850, -0.00000260, 0.00000335, -0.00000330, -0.00000181, -0.00000006, 0.00000077, -0.00000528, 0.00001171, -0.00000845, 0.00000040, -0.00000754, 0.00000771, -0.00000634, 0.00001436, -0.00000604, -0.00000190, -0.00000131, -0.00000090, 0.00000108, 0.00000181, -0.00000409, 0.00002047, -0.00001332, -0.00000148, 0.00000356, -0.00000365, 0.00000786, -0.00002743, 0.00001273, -0.00000205, -0.00000080, -0.00001980, 0.00001476, -0.00001841, 0.00001423, -0.00000755, -0.00000551, 0.00002188, -0.00000645, -0.00001501, 0.00000227, 0.00001128, },
		{ 0.00000010, 0.00000247, -0.00000233, -0.00000150, 0.00000208, 0.00000028, 0.00000110, -0.00000086, 0.00000248, -0.00000964, 0.00000595, -0.00000318, 0.00000755, -0.00000454, 0.00000599, -0.00001261, 0.00000900, -0.00001210, 0.00001415, -0.00002542, 0.00001646, 0.00000184, 0.00000172, -0.00000617, 0.00000116, -0.00000460, -0.00001571, 0.00000746, -0.00000637, 0.00002564, 0.00000623, -0.00000873, 0.00003572, -0.00000544, 0.00000964, -0.00001837, -0.00000463, 0.00000461, -0.00001377, 0.00000763, -0.00000351, 0.00000223, -0.00000732, 0.00000057, 0.00000236, 0.00000250, -0.00000309, -0.00000317, 0.00000088, -0.00000627, 0.00000474, 0.00000240, -0.00000151, -0.00001203, 0.00000757, -0.00000466, 0.00000431, 0.00000097, -0.00000289, -0.00001513, 0.00001432, -0.00001565, 0.00001553, -0.00000667, -0.00000102, -0.00003218, 0.00001942, -0.00003750, 0.00002966, -0.00006314, 0.00001964, -0.00031167, 0.00037109, 0.00044112, -0.00031248, 0.00002021, -0.00007255, 0.00001704, -0.00003356, 0.00001559, -0.00002408, 0.00002484, -0.00002577, 0.00000815, -0.00000767, -0.00000126, -0.00000618, 0.00000811, -0.00000955, 0.00000357, -0.00000239, 0.00000051, -0.00000280, 0.00001097, -0.00001226, 0.00000322, -0.00000817, 0.00000961, -0.00000738, 0.00001822, -0.00001580, 0.00000737, -0.00000871, 0.00000593, -0.00000386, 0.00000415, -0.00000197, 0.00002138, -0.00002592, 0.00001492, -0.00001035, 0.00001071, -0.00001110, -0.00001323, 0.00000958, 0.00000060, -0.00001175, -0.00000651, 0.00000680, -0.00001563, 0.00001994, -0.00002047, 0.00001136, 0.00001292, -0.00001042, -0.00001366, 0.00001037, 0.00000497, },
		{ 0.00000234, 0.00000026, -0.00000280, 0.00000042, 0.00000083, 0.00000182, -0.00000115, 0.00000229, -0.00000322, -0.00000447, 0.00000319, 0.00000060, 0.00000368, -0.00000108, 0.00000039, -0.00000744, 0.00000434, -0.00000624, 0.00000842, -0.00002560, 0.00002753, -0.00001052, 0.00000965, -0.00001520, 0.00001201, -0.00002046, 0.00000138, -0.00000252, 0.00000434, 0.00002473, -0.00000602, 0.00000500, 0.00003107, -0.00001276, 0.00001400, -0.00002920, 0.00001278, -0.00001128, -0.00000194, 0.00000205, 0.00000115, -0.00000345, -0.00000392, 0.00000031, 0.00000396, 0.00000028, -0.00000403, -0.00000100, -0.00000100, -0.00000483, 0.00000635, 0.00000111, -0.00000594, -0.00000725, 0.00000740, -0.00000571, 0.00000638, -0.00000027, -0.00000904, -0.00000565, 0.00000789, -0.00001152, 0.00001536, -0.00000786, -0.00001177, -0.00001693, 0.00000546, -0.00002211, 0.00000985, -0.00004397, -0.00005077, -0.00027752, 0.00067064, 0.00008143, -0.00024360, 0.00003329, -0.00006994, 0.00002095, -0.00003139, 0.00001287, -0.00001784, 0.00002074, -0.00002920, 0.00001553, -0.00001348, 0.00000080, -0.00000374, 0.00000558, -0.00000920, 0.00000440, -0.00000154, -0.00000311, 0.00000495, 0.00000205, -0.00000736, -0.00000128, -0.00000124, 0.00000316, 0.00000125, 0.00000987, -0.00001358, 0.00000741, -0.00000870, 0.00000664, -0.00000420, 0.00000308, 0.00000580, 0.00001045, -0.00002437, 0.00002109, -0.00001716, 0.00001820, -0.00002616, 0.00000347, 0.00000241, 0.00000284, -0.00002085, 0.00000995, -0.00001076, 0.00000201, 0.00000816, -0.00001795, 0.00001971, 0.00000443, -0.00001433, -0.00000711, 0.00001273, 0.00000075, },
		{ 0.00000118, -0.00000030, -0.00000342, 0.00000266, -0.00000107, 0.00000388, -0.00000370, 0.00000516, -0.00000904, 0.00000251, -0.00000198, 0.00000692, -0.00000350, 0.00000582, -0.00000912, 0.00000287, -0.00000643, 0.00000801, -0.00001040, -0.00000753, 0.00001994, -0.00000821, 0.00000461, -0.00001048, 0.00000828, -0.00002272, 0.00001054, -0.00000973, 0.00001660, 0.00001650, -0.00001238, 0.00002185, 0.00001459, -0.00000673, 0.00000565, -0.00002545, 0.00001798, -0.00002056, 0.00000793, -0.00000280, 0.00000404, -0.00000806, 0.00000001, -0.00000015, 0.00000480, -0.00000192, -0.00000460, 0.00000231, -0.00000573, 0.00000102, 0.00000328, 0.00000254, -0.00001310, 0.00000375, -0.00000128, 0.00000182, 0.00000082, 0.00000333, -0.00001891, 0.00001091, -0.00000973, 0.00000664, 0.00000042, 0.00000188, -0.00003170, 0.00001260, -0.00002899, 0.00001966, -0.00004558, 0.00001857, -0.00021026, -0.00002845, 0.00074531, -0.00025566, -0.00006678, -0.00003779, -0.00001303, -0.00001600, 0.00000125, -0.00001513, 0.00001162, -0.00000789, -0.00000847, 0.00000338, -0.00000551, -0.00000681, 0.00000714, -0.00000611, 0.00000010, -0.00000151, 0.00000325, -0.00000864, 0.00001448, -0.00001102, 0.00000312, -0.00001041, 0.00001010, -0.00000795, 0.00001539, -0.00000735, -0.00000021, -0.00000304, 0.00000092, -0.00000152, 0.00000293, -0.00000283, 0.00001782, -0.00001045, -0.00000689, 0.00001149, -0.00001022, 0.00001110, -0.00002694, 0.00001159, -0.00000027, -0.00000265, -0.00001754, 0.00001459, -0.00002105, 0.00001928, -0.00001093, -0.00000338, 0.00001746, -0.00000070, -0.00001814, 0.00000347, 0.00000830, 0.00000163, },
		{ -0.00000195, 0.00000032, -0.00000349, 0.00000376, -0.00000169, 0.00000429, -0.00000423, 0.00000491, -0.00001095, 0.00000631, -0.00000412, 0.00000974, -0.00000744, 0.00000918, -0.00001476, 0.00000949, -0.00001284, 0.00001793, -0.00002841, 0.00001671, 0.00000156, 0.00000391, -0.00000935, 0.00000467, -0.00000889, -0.00000817, 0.00000413, -0.00000463, 0.00002087, 0.00000717, -0.00001177, 0.00003472, -0.00000612, 0.00000785, -0.00001263, -0.00000739, 0.00000699, -0.00001623, 0.00000893, -0.00000177, 0.00000084, -0.00000710, 0.00000067, 0.00000153, 0.00000341, -0.00000344, -0.00000368, 0.00000367, -0.00000913, 0.00000691, -0.00000062, 0.00000249, -0.00001654, 0.00001272, -0.00001039, 0.00001092, -0.00000649, 0.00000575, -0.00002386, 0.00002252, -0.00002469, 0.00002498, -0.00001689, 0.00001012, -0.00004386, 0.00003291, -0.00005451, 0.00005171, -0.00009228, 0.00006234, -0.00037061, 0.00038453, 0.00050981, -0.00038670, 0.00006361, -0.00010285, 0.00004085, -0.00005194, 0.00003098, -0.00003911, 0.00003723, -0.00003688, 0.00001745, -0.00001482, 0.00000596, -0.00001447, 0.00001638, -0.00001708, 0.00001000, -0.00000754, 0.00000587, -0.00000902, 0.00001675, -0.00001776, 0.00000891, -0.00001427, 0.00001561, -0.00001273, 0.00002263, -0.00002014, 0.00001213, -0.00001369, 0.00001120, -0.00001086, 0.00001119, -0.00000729, 0.00002503, -0.00002949, 0.00001606, -0.00000703, 0.00000551, -0.00000684, -0.00001406, 0.00000911, 0.00000231, -0.00001436, -0.00000181, 0.00000217, -0.00001287, 0.00002066, -0.00002117, 0.00001074, 0.00001131, -0.00000579, -0.00001719, 0.00001162, 0.00000235, 0.00000457, },
		{ -0.00000300, -0.00000040, -0.00000137, 0.00000204, 0.00000076, 0.00000120, -0.00000110, 0.00000004, -0.00000686, 0.00000412, 0.00000001, 0.00000512, -0.00000347, 0.00000382, -0.00001062, 0.00000564, -0.00000643, 0.00001158, -0.00002936, 0.00002868, -0.00001048, 0.00001070, -0.00001719, 0.00001437, -0.00002337, 0.00000949, -0.00000904, 0.00000876, 0.00001668, -0.00000090, -0.00000258, 0.00003509, -0.00001840, 0.00001804, -0.00002790, 0.00001259, -0.00001114, -0.00000298, 0.00000331, 0.00000281, -0.00000591, -0.00000206, -0.00000124, 0.00000445, 0.00000065, -0.00000480, 0.00000004, -0.00000005, -0.00000637, 0.00000706, 0.00000008, -0.00000417, -0.00000935, 0.00001000, -0.00000878, 0.00001027, -0.00000516, -0.00000240, -0.00001313, 0.00001505, -0.00001946, 0.00002385, -0.00001737, -0.00000104, -0.00002794, 0.00001748, -0.00003690, 0.00002961, -0.00007210, -0.00000181, -0.00037219, 0.00076145, 0.00006931, -0.00027124, 0.00005521, -0.00008711, 0.00003577, -0.00004306, 0.00002221, -0.00002776, 0.00002913, -0.00003707, 0.00002246, -0.00001814, 0.00000472, -0.00000860, 0.00001099, -0.00001460, 0.00000929, -0.00000504, 0.00000002, 0.00000102, 0.00000572, -0.00001068, 0.00000218, -0.00000536, 0.00000772, -0.00000318, 0.00001331, -0.00001656, 0.00001065, -0.00001215, 0.00001021, -0.00001002, 0.00001065, -0.00000168, 0.00001693, -0.00003181, 0.00002784, -0.00001960, 0.00001663, -0.00002255, 0.00000163, 0.00000340, 0.00000278, -0.00002142, 0.00001339, -0.00001627, 0.00000804, 0.00000484, -0.00001535, 0.00001532, 0.00000760, -0.00001364, -0.00000836, 0.00001171, 0.00000118, 0.00000420, },
		{ -0.00000045, -0.00000311, 0.00000257, -0.00000159, 0.00000512, -0.00000408, 0.00000406, -0.00000716, 0.00000069, -0.00000153, 0.00000754, -0.00000404, 0.00000570, -0.00000760, 0.00000060, -0.00000556, 0.00000945, -0.00000968, -0.00000994, 0.00002034, -0.00000654, 0.00000285, -0.00000896, 0.00000643, -0.00001977, 0.00001302, -0.00001313, 0.00001799, 0.00001088, -0.00000849, 0.00001461, 0.00001928, -0.00001356, 0.00001275, -0.00002712, 0.00001930, -0.00002169, 0.00000773, -0.00000083, 0.00000397, -0.00000932, 0.00000137, -0.00000168, 0.00000550, -0.00000152, -0.00000603, 0.00000531, -0.00000770, 0.00000232, 0.00000091, 0.00000531, -0.00001568, 0.00000661, -0.00000431, 0.00000513, -0.00000254, 0.00000647, -0.00002064, 0.00001162, -0.00001098, 0.00000780, -0.00000095, 0.00000298, -0.00003212, 0.00001389, -0.00003168, 0.00002325, -0.00004927, 0.00002129, -0.00019973, -0.00010023, 0.00086258, -0.00032894, -0.00004073, -0.00005237, -0.00000281, -0.00002276, 0.00000708, -0.00002126, 0.00001546, -0.00001153, -0.00000563, 0.00000127, -0.00000207, -0.00001118, 0.00001071, -0.00000868, 0.00000189, -0.00000282, 0.00000538, -0.00001130, 0.00001630, -0.00001280, 0.00000528, -0.00001259, 0.00001160, -0.00000857, 0.00001541, -0.00000802, 0.00000119, -0.00000451, 0.00000246, -0.00000368, 0.00000334, -0.00000061, 0.00001410, -0.00000681, -0.00001267, 0.00001925, -0.00001601, 0.00001300, -0.00002482, 0.00000934, 0.00000216, -0.00000507, -0.00001435, 0.00001338, -0.00002263, 0.00002356, -0.00001426, -0.00000068, 0.00001213, 0.00000540, -0.00002086, 0.00000447, 0.00000517, 0.00000538, -0.00000089, },
		{ 0.00000264, -0.00000507, 0.00000508, -0.00000341, 0.00000725, -0.00000722, 0.00000668, -0.00001142, 0.00000589, -0.00000437, 0.00001125, -0.00000958, 0.00001137, -0.00001570, 0.00000882, -0.00001245, 0.00002044, -0.00002985, 0.00001570, 0.00000167, 0.00000604, -0.00001240, 0.00000817, -0.00001342, -0.00000002, -0.00000049, -0.00000209, 0.00001601, 0.00000793, -0.00001353, 0.00003239, -0.00000528, 0.00000429, -0.00000635, -0.00000952, 0.00000811, -0.00001726, 0.00000893, 0.00000104, -0.00000151, -0.00000589, -0.00000005, 0.00000143, 0.00000341, -0.00000282, -0.00000510, 0.00000736, -0.00001279, 0.00000991, -0.00000470, 0.00000745, -0.00002170, 0.00001855, -0.00001697, 0.00001849, -0.00001516, 0.00001551, -0.00003318, 0.00003107, -0.00003406, 0.00003501, -0.00002823, 0.00002244, -0.00005629, 0.00004718, -0.00007236, 0.00007450, -0.00012204, 0.00010425, -0.00042332, 0.00038809, 0.00058430, -0.00046372, 0.00010980, -0.00013544, 0.00006655, -0.00007186, 0.00004768, -0.00005509, 0.00005018, -0.00004869, 0.00002789, -0.00002329, 0.00001464, -0.00002409, 0.00002558, -0.00002539, 0.00001732, -0.00001365, 0.00001211, -0.00001580, 0.00002265, -0.00002329, 0.00001479, -0.00002048, 0.00002146, -0.00001769, 0.00002633, -0.00002386, 0.00001659, -0.00001848, 0.00001635, -0.00001766, 0.00001778, -0.00001181, 0.00002741, -0.00003166, 0.00001586, -0.00000273, -0.00000012, -0.00000280, -0.00001397, 0.00000777, 0.00000453, -0.00001721, 0.00000343, -0.00000329, -0.00000889, 0.00001995, -0.00002086, 0.00000978, 0.00000926, -0.00000078, -0.00002037, 0.00001253, -0.00000032, 0.00000910, -0.00000628, },
		{ 0.00000211, -0.00000314, 0.00000324, -0.00000041, 0.00000380, -0.00000471, 0.00000330, -0.00000903, 0.00000479, -0.00000039, 0.00000639, -0.00000569, 0.00000684, -0.00001310, 0.00000617, -0.00000585, 0.00001371, -0.00003148, 0.00002794, -0.00000918, 0.00001088, -0.00001826, 0.00001583, -0.00002549, 0.00001706, -0.00001608, 0.00001394, 0.00000815, 0.00000472, -0.00000958, 0.00003809, -0.00002247, 0.00001953, -0.00002449, 0.00001119, -0.00001062, -0.00000382, 0.00000414, 0.00000487, -0.00000876, 0.00000034, -0.00000328, 0.00000531, 0.00000044, -0.00000481, 0.00000028, 0.00000175, -0.00000869, 0.00000855, -0.00000205, -0.00000118, -0.00001241, 0.00001348, -0.00001285, 0.00001528, -0.00001160, 0.00000602, -0.00002209, 0.00002340, -0.00002843, 0.00003348, -0.00002865, 0.00001193, -0.00004111, 0.00003192, -0.00005455, 0.00005271, -0.00010457, 0.00005202, -0.00046688, 0.00084575, 0.00006249, -0.00030145, 0.00007949, -0.00010636, 0.00005246, -0.00005637, 0.00003306, -0.00003889, 0.00003818, -0.00004538, 0.00003012, -0.00002378, 0.00000996, -0.00001491, 0.00001759, -0.00002101, 0.00001528, -0.00000983, 0.00000459, -0.00000426, 0.00001033, -0.00001470, 0.00000649, -0.00001034, 0.00001296, -0.00000806, 0.00001675, -0.00001933, 0.00001384, -0.00001563, 0.00001392, -0.00001604, 0.00001829, -0.00000895, 0.00002261, -0.00003795, 0.00003289, -0.00002030, 0.00001353, -0.00001786, -0.00000046, 0.00000434, 0.00000262, -0.00002149, 0.00001640, -0.00002153, 0.00001429, 0.00000065, -0.00001173, 0.00001023, 0.00001059, -0.00001240, -0.00000959, 0.00001049, 0.00000148, 0.00000631, -0.00000653, },
		{ -0.00000225, 0.00000203, -0.00000170, 0.00000586, -0.00000376, 0.00000215, -0.00000453, -0.00000168, -0.00000060, 0.00000761, -0.00000394, 0.00000485, -0.00000542, -0.00000208, -0.00000439, 0.00001052, -0.00000869, -0.00001191, 0.00001958, -0.00000390, 0.00000045, -0.00000670, 0.00000373, -0.00001581, 0.00001426, -0.00001616, 0.00001970, 0.00000478, -0.00000356, 0.00000698, 0.00002399, -0.00001984, 0.00001785, -0.00002635, 0.00001839, -0.00002109, 0.00000640, 0.00000184, 0.00000331, -0.00001000, 0.00000234, -0.00000286, 0.00000571, -0.00000078, -0.00000756, 0.00000827, -0.00000953, 0.00000353, -0.00000142, 0.00000774, -0.00001763, 0.00000894, -0.00000695, 0.00000803, -0.00000548, 0.00000881, -0.00002107, 0.00001105, -0.00001114, 0.00000813, -0.00000170, 0.00000303, -0.00003084, 0.00001335, -0.00003216, 0.00002402, -0.00004933, 0.00001846, -0.00018053, -0.00017842, 0.00097858, -0.00039921, -0.00001589, -0.00006600, 0.00000655, -0.00002868, 0.00001213, -0.00002653, 0.00001852, -0.00001489, -0.00000258, -0.00000103, 0.00000141, -0.00001521, 0.00001357, -0.00001063, 0.00000325, -0.00000365, 0.00000682, -0.00001294, 0.00001686, -0.00001352, 0.00000663, -0.00001382, 0.00001198, -0.00000807, 0.00001433, -0.00000796, 0.00000223, -0.00000561, 0.00000359, -0.00000532, 0.00000305, 0.00000237, 0.00000964, -0.00000285, -0.00001818, 0.00002602, -0.00002033, 0.00001321, -0.00002102, 0.00000604, 0.00000513, -0.00000793, -0.00001033, 0.00001112, -0.00002300, 0.00002681, -0.00001728, 0.00000230, 0.00000634, 0.00001131, -0.00002277, 0.00000498, 0.00000222, 0.00000876, -0.00000243, -0.00000121, },
		{ -0.00000596, 0.00000579, -0.00000462, 0.00000968, -0.00000950, 0.00000753, -0.00001090, 0.00000459, -0.00000383, 0.00001190, -0.00001074, 0.00001234, -0.00001533, 0.00000696, -0.00001092, 0.00002157, -0.00002966, 0.00001352, 0.00000205, 0.00000806, -0.00001511, 0.00001139, -0.00001781, 0.00000824, -0.00000596, 0.00000103, 0.00001119, 0.00000833, -0.00001396, 0.00002877, -0.00000313, -0.00000043, -0.00000025, -0.00001048, 0.00000757, -0.00001655, 0.00000746, 0.00000494, -0.00000483, -0.00000365, -0.00000166, 0.00000217, 0.00000236, -0.00000112, -0.00000742, 0.00001180, -0.00001706, 0.00001357, -0.00000960, 0.00001306, -0.00002714, 0.00002462, -0.00002388, 0.00002649, -0.00002441, 0.00002565, -0.00004237, 0.00003926, -0.00004310, 0.00004497, -0.00003996, 0.00003514, -0.00006863, 0.00006128, -0.00008983, 0.00009650, -0.00015046, 0.00014287, -0.00046738, 0.00038123, 0.00066257, -0.00054063, 0.00015637, -0.00016847, 0.00009266, -0.00009210, 0.00006462, -0.00007106, 0.00006289, -0.00006049, 0.00003888, -0.00003259, 0.00002425, -0.00003441, 0.00003505, -0.00003389, 0.00002500, -0.00002021, 0.00001871, -0.00002259, 0.00002813, -0.00002833, 0.00002038, -0.00002627, 0.00002663, -0.00002178, 0.00002894, -0.00002660, 0.00002038, -0.00002265, 0.00002089, -0.00002375, 0.00002340, -0.00001510, 0.00002827, -0.00003230, 0.00001443, 0.00000210, -0.00000558, 0.00000052, -0.00001269, 0.00000550, 0.00000722, -0.00002019, 0.00000901, -0.00000935, -0.00000384, 0.00001785, -0.00001953, 0.00000847, 0.00000699, 0.00000417, -0.00002279, 0.00001274, -0.00000264, 0.00001319, -0.00001005, 0.00000429, },
		{ -0.00000461, 0.00000419, -0.00000145, 0.00000633, -0.00000816, 0.00000620, -0.00001067, 0.00000497, -0.00000041, 0.00000726, -0.00000743, 0.00000913, -0.00001460, 0.00000575, -0.00000433, 0.00001460, -0.00003173, 0.00002528, -0.00000669, 0.00001023, -0.00001834, 0.00001625, -0.00002653, 0.00002357, -0.00002297, 0.00001934, -0.00000035, 0.00001025, -0.00001546, 0.00003956, -0.00002454, 0.00001847, -0.00001927, 0.00000890, -0.00001001, -0.00000417, 0.00000431, 0.00000742, -0.00001201, 0.00000330, -0.00000583, 0.00000668, -0.00000054, -0.00000383, -0.00000049, 0.00000449, -0.00001181, 0.00001089, -0.00000530, 0.00000299, -0.00001635, 0.00001772, -0.00001771, 0.00002118, -0.00001923, 0.00001574, -0.00003198, 0.00003239, -0.00003789, 0.00004366, -0.00004105, 0.00002646, -0.00005571, 0.00004796, -0.00007400, 0.00007774, -0.00013933, 0.00010750, -0.00055713, 0.00091955, 0.00006266, -0.00033409, 0.00010551, -0.00012709, 0.00007045, -0.00007081, 0.00004497, -0.00005076, 0.00004745, -0.00005377, 0.00003823, -0.00003027, 0.00001640, -0.00002246, 0.00002506, -0.00002810, 0.00002207, -0.00001565, 0.00001034, -0.00001057, 0.00001556, -0.00001911, 0.00001132, -0.00001581, 0.00001847, -0.00001299, 0.00001984, -0.00002160, 0.00001667, -0.00001879, 0.00001737, -0.00002172, 0.00002534, -0.00001533, 0.00002686, -0.00004217, 0.00003576, -0.00001913, 0.00000911, -0.00001243, -0.00000250, 0.00000504, 0.00000245, -0.00002107, 0.00001883, -0.00002625, 0.00002043, -0.00000414, -0.00000738, 0.00000481, 0.00001318, -0.00001064, -0.00001069, 0.00000898, 0.00000179, 0.00000840, -0.00000994, 0.00000439, },
		{ 0.00000111, -0.00000143, 0.00000611, -0.00000283, -0.00000039, -0.00000141, -0.00000435, 0.00000064, 0.00000723, -0.00000333, 0.00000344, -0.00000285, -0.00000487, -0.00000310, 0.00001133, -0.00000761, -0.00001317, 0.00001756, -0.00000037, -0.00000244, -0.00000388, 0.00000033, -0.00001095, 0.00001417, -0.00001849, 0.00002134, -0.00000140, 0.00000186, -0.00000043, 0.00002798, -0.00002474, 0.00002044, -0.00002299, 0.00001525, -0.00001885, 0.00000413, 0.00000500, 0.00000218, -0.00001015, 0.00000294, -0.00000369, 0.00000552, 0.00000013, -0.00000892, 0.00001086, -0.00001090, 0.00000437, -0.00000340, 0.00000948, -0.00001861, 0.00001040, -0.00000884, 0.00001020, -0.00000767, 0.00001003, -0.00002004, 0.00000913, -0.00001017, 0.00000758, -0.00000173, 0.00000193, -0.00002779, 0.00001085, -0.00003028, 0.00002183, -0.00004571, 0.00001037, -0.00015414, -0.00025935, 0.00108834, -0.00046246, 0.00000539, -0.00007723, 0.00001403, -0.00003300, 0.00001576, -0.00003038, 0.00002034, -0.00001759, 0.00000032, -0.00000318, 0.00000449, -0.00001841, 0.00001531, -0.00001165, 0.00000392, -0.00000379, 0.00000733, -0.00001332, 0.00001600, -0.00001305, 0.00000703, -0.00001396, 0.00001116, -0.00000645, 0.00001220, -0.00000721, 0.00000285, -0.00000628, 0.00000429, -0.00000642, 0.00000218, 0.00000581, 0.00000489, 0.00000092, -0.00002277, 0.00003105, -0.00002263, 0.00001153, -0.00001560, 0.00000191, 0.00000845, -0.00001104, -0.00000563, 0.00000792, -0.00002209, 0.00002882, -0.00001977, 0.00000523, 0.00000060, 0.00001649, -0.00002352, 0.00000479, -0.00000032, 0.00001152, -0.00000333, -0.00000257, -0.00000090, },
		{ 0.00000575, -0.00000513, 0.00001133, -0.00001083, 0.00000732, -0.00000935, 0.00000245, -0.00000249, 0.00001161, -0.00001081, 0.00001204, -0.00001368, 0.00000406, -0.00000835, 0.00002135, -0.00002792, 0.00001038, 0.00000255, 0.00000995, -0.00001733, 0.00001408, -0.00002170, 0.00001614, -0.00001186, 0.00000452, 0.00000654, 0.00000831, -0.00001314, 0.00002408, -0.00000011, -0.00000559, 0.00000494, -0.00000986, 0.00000514, -0.00001396, 0.00000452, 0.00000983, -0.00000900, -0.00000046, -0.00000412, 0.00000379, 0.00000024, 0.00000159, -0.00001050, 0.00001672, -0.00002161, 0.00001757, -0.00001493, 0.00001888, -0.00003237, 0.00003039, -0.00003055, 0.00003426, -0.00003354, 0.00003540, -0.00005067, 0.00004644, -0.00005116, 0.00005414, -0.00005131, 0.00004739, -0.00007999, 0.00007418, -0.00010571, 0.00011622, -0.00017571, 0.00017598, -0.00050094, 0.00036419, 0.00074192, -0.00061403, 0.00020058, -0.00019988, 0.00011751, -0.00011130, 0.00008065, -0.00008598, 0.00007444, -0.00007151, 0.00004973, -0.00004208, 0.00003411, -0.00004466, 0.00004403, -0.00004189, 0.00003242, -0.00002665, 0.00002510, -0.00002884, 0.00003268, -0.00003243, 0.00002519, -0.00003115, 0.00003067, -0.00002464, 0.00003017, -0.00002810, 0.00002317, -0.00002585, 0.00002444, -0.00002868, 0.00002764, -0.00001694, 0.00002752, -0.00003145, 0.00001201, 0.00000688, -0.00001025, 0.00000268, -0.00001003, 0.00000231, 0.00001027, -0.00002311, 0.00001466, -0.00001570, 0.00000201, 0.00001446, -0.00001726, 0.00000678, 0.00000475, 0.00000862, -0.00002403, 0.00001194, -0.00000424, 0.00001642, -0.00001288, 0.00000500, -0.00000596, },
		{ 0.00000468, -0.00000213, 0.00000848, -0.00001106, 0.00000838, -0.00001151, 0.00000452, 0.00000009, 0.00000755, -0.00000843, 0.00001042, -0.00001494, 0.00000430, -0.00000188, 0.00001417, -0.00003007, 0.00002083, -0.00000323, 0.00000889, -0.00001745, 0.00001554, -0.00002631, 0.00002861, -0.00002913, 0.00002441, -0.00000832, 0.00001519, -0.00001978, 0.00003916, -0.00002447, 0.00001515, -0.00001279, 0.00000617, -0.00000965, -0.00000375, 0.00000366, 0.00001050, -0.00001565, 0.00000676, -0.00000889, 0.00000859, -0.00000240, -0.00000172, -0.00000236, 0.00000817, -0.00001568, 0.00001400, -0.00000956, 0.00000816, -0.00002097, 0.00002245, -0.00002305, 0.00002756, -0.00002755, 0.00002610, -0.00004208, 0.00004131, -0.00004714, 0.00005368, -0.00005378, 0.00004163, -0.00007078, 0.00006451, -0.00009392, 0.00010299, -0.00017404, 0.00016117, -0.00063852, 0.00097929, 0.00007097, -0.00036862, 0.00013234, -0.00014839, 0.00008890, -0.00008568, 0.00005730, -0.00006279, 0.00005646, -0.00006181, 0.00004645, -0.00003731, 0.00002373, -0.00003082, 0.00003294, -0.00003541, 0.00002924, -0.00002207, 0.00001681, -0.00001747, 0.00002096, -0.00002352, 0.00001628, -0.00002132, 0.00002377, -0.00001754, 0.00002224, -0.00002308, 0.00001885, -0.00002131, 0.00002014, -0.00002655, 0.00003120, -0.00002023, 0.00002919, -0.00004404, 0.00003616, -0.00001620, 0.00000379, -0.00000674, -0.00000415, 0.00000534, 0.00000235, -0.00002015, 0.00002058, -0.00003017, 0.00002609, -0.00000920, -0.00000265, -0.00000058, 0.00001519, -0.00000847, -0.00001148, 0.00000705, 0.00000230, 0.00001016, -0.00001287, 0.00000611, -0.00000552, },
		{ -0.00000090, 0.00000596, -0.00000144, -0.00000329, 0.00000188, -0.00000703, 0.00000197, 0.00000656, -0.00000237, 0.00000172, -0.00000021, -0.00000746, -0.00000193, 0.00001203, -0.00000662, -0.00001356, 0.00001428, 0.00000384, -0.00000556, -0.00000071, -0.00000354, -0.00000540, 0.00001277, -0.00001986, 0.00002256, -0.00000723, 0.00000723, -0.00000702, 0.00003062, -0.00002765, 0.00002027, -0.00001721, 0.00001016, -0.00001524, 0.00000121, 0.00000838, 0.00000078, -0.00000990, 0.00000326, -0.00000424, 0.00000504, 0.00000101, -0.00000983, 0.00001276, -0.00001154, 0.00000460, -0.00000474, 0.00001027, -0.00001835, 0.00001074, -0.00000974, 0.00001139, -0.00000886, 0.00000998, -0.00001751, 0.00000592, -0.00000815, 0.00000617, -0.00000107, -0.00000025, -0.00002308, 0.00000653, -0.00002615, 0.00001686, -0.00003877, -0.00000213, -0.00012269, -0.00033888, 0.00118681, -0.00051487, 0.00002093, -0.00008478, 0.00001874, -0.00003509, 0.00001745, -0.00003239, 0.00002058, -0.00001932, 0.00000277, -0.00000485, 0.00000679, -0.00002036, 0.00001562, -0.00001153, 0.00000375, -0.00000309, 0.00000680, -0.00001237, 0.00001370, -0.00001138, 0.00000644, -0.00001299, 0.00000921, -0.00000385, 0.00000919, -0.00000589, 0.00000313, -0.00000655, 0.00000458, -0.00000706, 0.00000092, 0.00000934, 0.00000033, 0.00000398, -0.00002585, 0.00003372, -0.00002258, 0.00000798, -0.00000885, -0.00000273, 0.00001184, -0.00001415, -0.00000052, 0.00000396, -0.00001995, 0.00002944, -0.00002149, 0.00000780, -0.00000462, 0.00002051, -0.00002287, 0.00000378, -0.00000228, 0.00001354, -0.00000358, -0.00000438, 0.00000077, -0.00000005, },
		{ -0.00000485, 0.00001205, -0.00001109, 0.00000604, -0.00000688, -0.00000038, -0.00000047, 0.00001045, -0.00000984, 0.00001059, -0.00001097, 0.00000038, -0.00000500, 0.00001995, -0.00002486, 0.00000659, 0.00000298, 0.00001171, -0.00001890, 0.00001601, -0.00002479, 0.00002325, -0.00001773, 0.00000813, 0.00000215, 0.00000787, -0.00001128, 0.00001867, 0.00000323, -0.00001041, 0.00000864, -0.00000743, 0.00000081, -0.00000958, 0.00000032, 0.00001543, -0.00001378, 0.00000346, -0.00000728, 0.00000616, -0.00000281, 0.00000514, -0.00001403, 0.00002175, -0.00002608, 0.00002151, -0.00002026, 0.00002441, -0.00003690, 0.00003533, -0.00003639, 0.00004116, -0.00004180, 0.00004400, -0.00005740, 0.00005198, -0.00005765, 0.00006189, -0.00006153, 0.00005837, -0.00008953, 0.00008495, -0.00011887, 0.00013233, -0.00019620, 0.00020183, -0.00052288, 0.00033796, 0.00081910, -0.00068031, 0.00023963, -0.00022753, 0.00013941, -0.00012808, 0.00009460, -0.00009883, 0.00008396, -0.00008095, 0.00005969, -0.00005105, 0.00004346, -0.00005402, 0.00005174, -0.00004871, 0.00003895, -0.00003237, 0.00003071, -0.00003401, 0.00003585, -0.00003518, 0.00002884, -0.00003471, 0.00003320, -0.00002601, 0.00002985, -0.00002822, 0.00002477, -0.00002783, 0.00002671, -0.00003214, 0.00003027, -0.00001722, 0.00002526, -0.00002929, 0.00000900, 0.00001102, -0.00001355, 0.00000335, -0.00000595, -0.00000165, 0.00001351, -0.00002574, 0.00002010, -0.00002198, 0.00000834, 0.00000997, -0.00001414, 0.00000468, 0.00000282, 0.00001214, -0.00002377, 0.00000992, -0.00000485, 0.00001845, -0.00001447, 0.00000461, -0.00000503, 0.00000489, },
		{ -0.00000224, 0.00000995, -0.00001308, 0.00000955, -0.00001140, 0.00000335, 0.00000115, 0.00000715, -0.00000855, 0.00001059, -0.00001406, 0.00000192, 0.00000138, 0.00001252, -0.00002665, 0.00001491, 0.00000086, 0.00000708, -0.00001569, 0.00001372, -0.00002473, 0.00003187, -0.00003407, 0.00002867, -0.00001533, 0.00001914, -0.00002230, 0.00003679, -0.00002240, 0.00001012, -0.00000574, 0.00000349, -0.00000984, -0.00000236, 0.00000210, 0.00001404, -0.00001955, 0.00001062, -0.00001236, 0.00001104, -0.00000516, 0.00000149, -0.00000524, 0.00001260, -0.00002008, 0.00001767, -0.00001456, 0.00001402, -0.00002591, 0.00002729, -0.00002842, 0.00003392, -0.00003592, 0.00003634, -0.00005159, 0.00004942, -0.00005548, 0.00006280, -0.00006597, 0.00005645, -0.00008522, 0.00008035, -0.00011285, 0.00012662, -0.00020624, 0.00020965, -0.00070707, 0.00102211, 0.00008787, -0.00040399, 0.00015870, -0.00016911, 0.00010680, -0.00010010, 0.00006929, -0.00007430, 0.00006460, -0.00006902, 0.00005431, -0.00004449, 0.00003149, -0.00003945, 0.00004063, -0.00004238, 0.00003623, -0.00002857, 0.00002348, -0.00002440, 0.00002606, -0.00002749, 0.00002092, -0.00002639, 0.00002836, -0.00002128, 0.00002364, -0.00002355, 0.00002013, -0.00002288, 0.00002189, -0.00003007, 0.00003537, -0.00002324, 0.00002931, -0.00004336, 0.00003409, -0.00001184, -0.00000192, -0.00000128, -0.00000508, 0.00000508, 0.00000241, -0.00001876, 0.00002157, -0.00003304, 0.00003092, -0.00001419, 0.00000212, -0.00000566, 0.00001654, -0.00000605, -0.00001176, 0.00000459, 0.00000317, 0.00001127, -0.00001495, 0.00000698, -0.00000566, 0.00000451, },
		{ 0.00000555, 0.00000019, -0.00000622, 0.00000496, -0.00000938, 0.00000316, 0.00000579, -0.00000129, -0.00000001, 0.00000219, -0.00000957, -0.00000105, 0.00001270, -0.00000585, -0.00001299, 0.00000987, 0.00000847, -0.00000862, 0.00000253, -0.00000764, 0.00000059, 0.00001016, -0.00002007, 0.00002301, -0.00001231, 0.00001201, -0.00001227, 0.00003142, -0.00002822, 0.00001744, -0.00000947, 0.00000362, -0.00001072, -0.00000193, 0.00001160, -0.00000065, -0.00000940, 0.00000339, -0.00000460, 0.00000442, 0.00000165, -0.00001004, 0.00001370, -0.00001126, 0.00000405, -0.00000525, 0.00000993, -0.00001674, 0.00000985, -0.00000952, 0.00001148, -0.00000896, 0.00000863, -0.00001363, 0.00000167, -0.00000529, 0.00000410, 0.00000017, -0.00000332, -0.00001701, 0.00000070, -0.00002014, 0.00000960, -0.00002925, -0.00001776, -0.00008872, -0.00041273, 0.00126921, -0.00055307, 0.00002901, -0.00008773, 0.00002011, -0.00003455, 0.00001692, -0.00003229, 0.00001903, -0.00001987, 0.00000451, -0.00000577, 0.00000801, -0.00002080, 0.00001436, -0.00001022, 0.00000271, -0.00000156, 0.00000526, -0.00001017, 0.00001012, -0.00000867, 0.00000500, -0.00001106, 0.00000631, -0.00000053, 0.00000559, -0.00000419, 0.00000315, -0.00000650, 0.00000456, -0.00000737, -0.00000045, 0.00001254, -0.00000357, 0.00000590, -0.00002698, 0.00003368, -0.00002011, 0.00000282, -0.00000119, -0.00000744, 0.00001497, -0.00001693, 0.00000473, -0.00000050, -0.00001671, 0.00002861, -0.00002228, 0.00000972, -0.00000893, 0.00002303, -0.00002075, 0.00000197, -0.00000364, 0.00001478, -0.00000329, -0.00000643, 0.00000251, -0.00000124, -0.00000024, },
		{ 0.00001183, -0.00001030, 0.00000383, -0.00000374, -0.00000363, 0.00000199, 0.00000858, -0.00000800, 0.00000822, -0.00000751, -0.00000370, -0.00000122, 0.00001767, -0.00002081, 0.00000251, 0.00000313, 0.00001332, -0.00001971, 0.00001701, -0.00002683, 0.00002917, -0.00002319, 0.00001163, -0.00000193, 0.00000710, -0.00000868, 0.00001300, 0.00000632, -0.00001421, 0.00001044, -0.00000320, -0.00000521, -0.00000372, -0.00000479, 0.00002134, -0.00001879, 0.00000781, -0.00001087, 0.00000906, -0.00000655, 0.00000921, -0.00001766, 0.00002643, -0.00003001, 0.00002499, -0.00002511, 0.00002916, -0.00004024, 0.00003895, -0.00004086, 0.00004660, -0.00004853, 0.00005076, -0.00006198, 0.00005542, -0.00006211, 0.00006768, -0.00006996, 0.00006737, -0.00009653, 0.00009281, -0.00012842, 0.00014380, -0.00021075, 0.00021926, -0.00053285, 0.00030429, 0.00089056, -0.00073585, 0.00027080, -0.00024944, 0.00015680, -0.00014118, 0.00010541, -0.00010868, 0.00009065, -0.00008809, 0.00006804, -0.00005877, 0.00005155, -0.00006171, 0.00005748, -0.00005375, 0.00004404, -0.00003686, 0.00003505, -0.00003770, 0.00003734, -0.00003632, 0.00003103, -0.00003669, 0.00003405, -0.00002580, 0.00002800, -0.00002695, 0.00002513, -0.00002847, 0.00002756, -0.00003399, 0.00003120, -0.00001607, 0.00002174, -0.00002617, 0.00000585, 0.00001394, -0.00001506, 0.00000237, -0.00000056, -0.00000616, 0.00001671, -0.00002782, 0.00002500, -0.00002780, 0.00001474, 0.00000467, -0.00001033, 0.00000219, 0.00000144, 0.00001439, -0.00002182, 0.00000658, -0.00000434, 0.00001908, -0.00001470, 0.00000312, -0.00000321, 0.00000328, -0.00000380, },
		{ 0.00001057, -0.00001400, 0.00000957, -0.00001028, 0.00000151, 0.00000276, 0.00000604, -0.00000776, 0.00000965, -0.00001209, -0.00000120, 0.00000521, 0.00000987, -0.00002180, 0.00000800, 0.00000513, 0.00000509, -0.00001323, 0.00001090, -0.00002182, 0.00003318, -0.00003740, 0.00003172, -0.00002107, 0.00002181, -0.00002294, 0.00003259, -0.00001872, 0.00000411, 0.00000116, 0.00000133, -0.00001082, 0.00000008, -0.00000032, 0.00001790, -0.00002354, 0.00001468, -0.00001608, 0.00001390, -0.00000867, 0.00000566, -0.00000895, 0.00001749, -0.00002469, 0.00002160, -0.00001992, 0.00002011, -0.00003072, 0.00003178, -0.00003334, 0.00003971, -0.00004366, 0.00004564, -0.00005973, 0.00005601, -0.00006222, 0.00007029, -0.00007672, 0.00006988, -0.00009794, 0.00009426, -0.00012935, 0.00014684, -0.00023362, 0.00024990, -0.00075949, 0.00104608, 0.00011307, -0.00043871, 0.00018305, -0.00018794, 0.00012302, -0.00011312, 0.00008014, -0.00008455, 0.00007130, -0.00007485, 0.00006129, -0.00005129, 0.00003913, -0.00004767, 0.00004747, -0.00004842, 0.00004247, -0.00003455, 0.00002974, -0.00003080, 0.00003035, -0.00003059, 0.00002481, -0.00003057, 0.00003184, -0.00002389, 0.00002386, -0.00002288, 0.00002038, -0.00002333, 0.00002239, -0.00003199, 0.00003754, -0.00002414, 0.00002718, -0.00004021, 0.00002981, -0.00000659, -0.00000740, 0.00000344, -0.00000499, 0.00000415, 0.00000271, -0.00001692, 0.00002175, -0.00003470, 0.00003460, -0.00001876, 0.00000662, -0.00001016, 0.00001723, -0.00000359, -0.00001135, 0.00000150, 0.00000455, 0.00001151, -0.00001592, 0.00000680, -0.00000491, 0.00000361, -0.00000168, },
		{ 0.00000180, -0.00000881, 0.00000746, -0.00001111, 0.00000400, 0.00000511, -0.00000030, -0.00000150, 0.00000408, -0.00001097, -0.00000060, 0.00001342, -0.00000535, -0.00001150, 0.00000459, 0.00001312, -0.00001129, 0.00000556, -0.00001169, 0.00000674, 0.00000650, -0.00001907, 0.00002248, -0.00001631, 0.00001580, -0.00001580, 0.00003012, -0.00002642, 0.00001236, -0.00000047, -0.00000369, -0.00000585, -0.00000483, 0.00001431, -0.00000187, -0.00000881, 0.00000346, -0.00000487, 0.00000381, 0.00000188, -0.00000937, 0.00001354, -0.00000996, 0.00000266, -0.00000486, 0.00000842, -0.00001381, 0.00000776, -0.00000821, 0.00001050, -0.00000803, 0.00000616, -0.00000872, -0.00000324, -0.00000196, 0.00000162, 0.00000173, -0.00000696, -0.00001002, -0.00000610, -0.00001285, 0.00000080, -0.00001819, -0.00003490, -0.00005501, -0.00047674, 0.00133135, -0.00057446, 0.00002850, -0.00008559, 0.00001790, -0.00003129, 0.00001409, -0.00003004, 0.00001569, -0.00001915, 0.00000539, -0.00000579, 0.00000800, -0.00001965, 0.00001156, -0.00000782, 0.00000091, 0.00000067, 0.00000288, -0.00000698, 0.00000554, -0.00000519, 0.00000295, -0.00000840, 0.00000279, 0.00000313, 0.00000175, -0.00000237, 0.00000306, -0.00000623, 0.00000433, -0.00000752, -0.00000169, 0.00001506, -0.00000644, 0.00000636, -0.00002592, 0.00003083, -0.00001546, -0.00000348, 0.00000682, -0.00001173, 0.00001748, -0.00001909, 0.00000979, -0.00000516, -0.00001257, 0.00002636, -0.00002204, 0.00001077, -0.00001203, 0.00002390, -0.00001723, -0.00000051, -0.00000444, 0.00001530, -0.00000262, -0.00000843, 0.00000406, -0.00000226, 0.00000036, 0.00000506, },
		{ -0.00000860, 0.00000095, -0.00000025, -0.00000695, 0.00000462, 0.00000625, -0.00000557, 0.00000526, -0.00000371, -0.00000775, 0.00000261, 0.00001485, -0.00001616, -0.00000150, 0.00000284, 0.00001475, -0.00001966, 0.00001697, -0.00002766, 0.00003364, -0.00002788, 0.00001484, -0.00000566, 0.00000611, -0.00000566, 0.00000754, 0.00000858, -0.00001647, 0.00001015, 0.00000260, -0.00001246, 0.00000313, -0.00001027, 0.00002704, -0.00002358, 0.00001219, -0.00001455, 0.00001220, -0.00001063, 0.00001341, -0.00002096, 0.00003032, -0.00003300, 0.00002760, -0.00002905, 0.00003269, -0.00004199, 0.00004086, -0.00004354, 0.00005012, -0.00005323, 0.00005518, -0.00006403, 0.00005649, -0.00006426, 0.00007114, -0.00007610, 0.00007384, -0.00010050, 0.00009725, -0.00013377, 0.00014997, -0.00021868, 0.00022775, -0.00053128, 0.00026548, 0.00095262, -0.00077731, 0.00029176, -0.00026394, 0.00016839, -0.00014960, 0.00011222, -0.00011480, 0.00009392, -0.00009235, 0.00007415, -0.00006460, 0.00005769, -0.00006706, 0.00006072, -0.00005655, 0.00004726, -0.00003972, 0.00003778, -0.00003963, 0.00003697, -0.00003574, 0.00003166, -0.00003698, 0.00003317, -0.00002411, 0.00002478, -0.00002446, 0.00002430, -0.00002780, 0.00002699, -0.00003423, 0.00003055, -0.00001372, 0.00001735, -0.00002253, 0.00000303, 0.00001520, -0.00001453, -0.00000020, 0.00000588, -0.00001086, 0.00001957, -0.00002906, 0.00002903, -0.00003274, 0.00002079, -0.00000112, -0.00000604, -0.00000066, 0.00000076, 0.00001519, -0.00001816, 0.00000202, -0.00000270, 0.00001827, -0.00001359, 0.00000068, -0.00000071, 0.00000108, -0.00000212, 0.00001076, },
		{ -0.00001374, 0.00000841, -0.00000823, -0.00000088, 0.00000476, 0.00000434, -0.00000616, 0.00000774, -0.00000925, -0.00000472, 0.00000922, 0.00000660, -0.00001597, 0.00000066, 0.00000912, 0.00000321, -0.00001028, 0.00000726, -0.00001772, 0.00003254, -0.00003892, 0.00003333, -0.00002531, 0.00002310, -0.00002179, 0.00002688, -0.00001400, -0.00000209, 0.00000719, 0.00000007, -0.00001269, 0.00000354, -0.00000344, 0.00002184, -0.00002733, 0.00001867, -0.00001978, 0.00001695, -0.00001272, 0.00001049, -0.00001313, 0.00002245, -0.00002911, 0.00002537, -0.00002519, 0.00002594, -0.00003493, 0.00003546, -0.00003731, 0.00004434, -0.00005009, 0.00005326, -0.00006577, 0.00006045, -0.00006679, 0.00007551, -0.00008525, 0.00008100, -0.00010794, 0.00010514, -0.00014210, 0.00016207, -0.00025421, 0.00027947, -0.00079342, 0.00105034, 0.00014548, -0.00047091, 0.00020372, -0.00020351, 0.00013642, -0.00012377, 0.00008900, -0.00009283, 0.00007596, -0.00007883, 0.00006689, -0.00005716, 0.00004600, -0.00005481, 0.00005283, -0.00005292, 0.00004738, -0.00003945, 0.00003502, -0.00003611, 0.00003341, -0.00003246, 0.00002760, -0.00003348, 0.00003386, -0.00002515, 0.00002281, -0.00002107, 0.00001956, -0.00002260, 0.00002155, -0.00003218, 0.00003761, -0.00002297, 0.00002300, -0.00003493, 0.00002383, -0.00000109, -0.00001205, 0.00000700, -0.00000368, 0.00000250, 0.00000327, -0.00001467, 0.00002111, -0.00003501, 0.00003687, -0.00002261, 0.00001055, -0.00001390, 0.00001730, -0.00000131, -0.00001012, -0.00000224, 0.00000650, 0.00001071, -0.00001561, 0.00000551, -0.00000326, 0.00000198, -0.00000041, 0.00001025, },
		{ -0.00001078, 0.00000910, -0.00001199, 0.00000430, 0.00000465, 0.00000042, -0.00000255, 0.00000524, -0.00001152, -0.00000063, 0.00001415, -0.00000510, -0.00000923, -0.00000120, 0.00001739, -0.00001322, 0.00000811, -0.00001538, 0.00001271, 0.00000207, -0.00001689, 0.00002085, -0.00001901, 0.00001829, -0.00001741, 0.00002670, -0.00002247, 0.00000568, 0.00000895, -0.00001100, -0.00000121, -0.00000704, 0.00001618, -0.00000267, -0.00000828, 0.00000356, -0.00000515, 0.00000334, 0.00000157, -0.00000776, 0.00001223, -0.00000768, 0.00000050, -0.00000363, 0.00000586, -0.00000975, 0.00000466, -0.00000598, 0.00000862, -0.00000625, 0.00000286, -0.00000322, -0.00000832, 0.00000145, -0.00000093, 0.00000332, -0.00001078, -0.00000268, -0.00001325, -0.00000501, -0.00000862, -0.00000678, -0.00005181, -0.00002429, -0.00052727, 0.00137001, -0.00057741, 0.00001903, -0.00007842, 0.00001226, -0.00002550, 0.00000916, -0.00002583, 0.00001070, -0.00001724, 0.00000540, -0.00000490, 0.00000677, -0.00001698, 0.00000743, -0.00000458, -0.00000141, 0.00000334, -0.00000005, -0.00000314, 0.00000039, -0.00000133, 0.00000059, -0.00000535, -0.00000098, 0.00000674, -0.00000197, -0.00000067, 0.00000300, -0.00000586, 0.00000403, -0.00000765, -0.00000257, 0.00001659, -0.00000800, 0.00000525, -0.00002269, 0.00002539, -0.00000911, -0.00001029, 0.00001456, -0.00001511, 0.00001903, -0.00002030, 0.00001434, -0.00000966, -0.00000783, 0.00002281, -0.00002075, 0.00001085, -0.00001376, 0.00002312, -0.00001256, -0.00000341, -0.00000482, 0.00001525, -0.00000180, -0.00001011, 0.00000516, -0.00000291, 0.00000074, 0.00000467, 0.00000314, },
		{ -0.00000228, 0.00000320, -0.00000997, 0.00000707, 0.00000378, -0.00000288, 0.00000212, -0.00000002, -0.00001132, 0.00000608, 0.00001185, -0.00001129, -0.00000509, 0.00000200, 0.00001595, -0.00001869, 0.00001585, -0.00002721, 0.00003645, -0.00003153, 0.00001757, -0.00000903, 0.00000506, -0.00000260, 0.00000273, 0.00000960, -0.00001686, 0.00000788, 0.00000948, -0.00002033, 0.00001034, -0.00001554, 0.00003199, -0.00002769, 0.00001619, -0.00001796, 0.00001522, -0.00001466, 0.00001732, -0.00002352, 0.00003303, -0.00003470, 0.00002901, -0.00003171, 0.00003466, -0.00004191, 0.00004083, -0.00004420, 0.00005144, -0.00005555, 0.00005699, -0.00006341, 0.00005513, -0.00006403, 0.00007211, -0.00007966, 0.00007746, -0.00010116, 0.00009801, -0.00013466, 0.00015061, -0.00021982, 0.00022743, -0.00051928, 0.00022421, 0.00100181, -0.00080189, 0.00030072, -0.00026984, 0.00017330, -0.00015264, 0.00011451, -0.00011675, 0.00009340, -0.00009334, 0.00007756, -0.00006804, 0.00006137, -0.00006962, 0.00006110, -0.00005685, 0.00004838, -0.00004074, 0.00003872, -0.00003971, 0.00003479, -0.00003351, 0.00003077, -0.00003564, 0.00003073, -0.00002117, 0.00002051, -0.00002101, 0.00002247, -0.00002596, 0.00002515, -0.00003302, 0.00002854, -0.00001055, 0.00001255, -0.00001880, 0.00000096, 0.00001449, -0.00001199, -0.00000410, 0.00001296, -0.00001534, 0.00002178, -0.00002920, 0.00003191, -0.00003643, 0.00002604, -0.00000704, -0.00000149, -0.00000379, 0.00000087, 0.00001447, -0.00001294, -0.00000354, -0.00000011, 0.00001616, -0.00001134, -0.00000246, 0.00000220, -0.00000144, -0.00000001, 0.00000904, -0.00000625, },
		{ 0.00000620, -0.00000546, -0.00000357, 0.00000695, 0.00000224, -0.00000396, 0.00000515, -0.00000590, -0.00000825, 0.00001302, 0.00000312, -0.00000970, -0.00000651, 0.00001236, 0.00000173, -0.00000706, 0.00000304, -0.00001269, 0.00003008, -0.00003857, 0.00003339, -0.00002800, 0.00002304, -0.00001913, 0.00002015, -0.00000890, -0.00000768, 0.00001177, -0.00000008, -0.00001541, 0.00000780, -0.00000698, 0.00002551, -0.00003061, 0.00002227, -0.00002318, 0.00001992, -0.00001696, 0.00001559, -0.00001736, 0.00002702, -0.00003290, 0.00002856, -0.00002987, 0.00003099, -0.00003806, 0.00003789, -0.00003989, 0.00004735, -0.00005463, 0.00005855, -0.00006918, 0.00006231, -0.00006879, 0.00007802, -0.00009092, 0.00008901, -0.00011439, 0.00011208, -0.00015008, 0.00017113, -0.00026655, 0.00029667, -0.00080759, 0.00103514, 0.00018331, -0.00049848, 0.00021902, -0.00021450, 0.00014590, -0.00013118, 0.00009515, -0.00009852, 0.00007812, -0.00008053, 0.00007062, -0.00006156, 0.00005151, -0.00006023, 0.00005614, -0.00005541, 0.00005048, -0.00004278, 0.00003883, -0.00003992, 0.00003490, -0.00003287, 0.00002903, -0.00003489, 0.00003426, -0.00002499, 0.00002056, -0.00001825, 0.00001777, -0.00002074, 0.00001941, -0.00003067, 0.00003566, -0.00001996, 0.00001719, -0.00002807, 0.00001681, 0.00000395, -0.00001533, 0.00000909, -0.00000108, 0.00000020, 0.00000407, -0.00001204, 0.00001970, -0.00003395, 0.00003752, -0.00002549, 0.00001368, -0.00001677, 0.00001683, 0.00000060, -0.00000802, -0.00000656, 0.00000898, 0.00000886, -0.00001403, 0.00000316, -0.00000084, -0.00000026, 0.00000147, 0.00000903, -0.00001233, },
		{ 0.00000967, -0.00001188, 0.00000401, 0.00000451, 0.00000078, -0.00000303, 0.00000559, -0.00001121, -0.00000107, 0.00001481, -0.00000498, -0.00000643, -0.00000707, 0.00002085, -0.00001412, 0.00000993, -0.00001848, 0.00001817, -0.00000283, -0.00001367, 0.00001809, -0.00002033, 0.00001936, -0.00001712, 0.00002137, -0.00001687, -0.00000178, 0.00001786, -0.00001755, 0.00000264, -0.00000817, 0.00001698, -0.00000291, -0.00000788, 0.00000375, -0.00000548, 0.00000307, 0.00000070, -0.00000523, 0.00000988, -0.00000459, -0.00000227, -0.00000172, 0.00000249, -0.00000487, 0.00000089, -0.00000314, 0.00000611, -0.00000392, -0.00000088, 0.00000234, -0.00001307, 0.00000449, -0.00000321, 0.00000464, -0.00001434, 0.00000441, -0.00002005, 0.00000259, -0.00001769, 0.00000374, -0.00006680, 0.00000103, -0.00056144, 0.00138313, -0.00056144, 0.00000103, -0.00006680, 0.00000374, -0.00001769, 0.00000259, -0.00002005, 0.00000441, -0.00001434, 0.00000464, -0.00000321, 0.00000449, -0.00001307, 0.00000234, -0.00000088, -0.00000392, 0.00000611, -0.00000314, 0.00000089, -0.00000487, 0.00000249, -0.00000172, -0.00000227, -0.00000459, 0.00000988, -0.00000523, 0.00000070, 0.00000307, -0.00000548, 0.00000375, -0.00000788, -0.00000291, 0.00001698, -0.00000817, 0.00000264, -0.00001755, 0.00001786, -0.00000178, -0.00001687, 0.00002137, -0.00001712, 0.00001936, -0.00002033, 0.00001809, -0.00001367, -0.00000283, 0.00001817, -0.00001848, 0.00000993, -0.00001412, 0.00002085, -0.00000707, -0.00000643, -0.00000498, 0.00001481, -0.00000107, -0.00001121, 0.00000559, -0.00000303, 0.00000078, 0.00000451, 0.00000401, -0.00001188, },
		{ 0.00000620, -0.00001233, 0.00000903, 0.00000147, -0.00000026, -0.00000084, 0.00000316, -0.00001403, 0.00000886, 0.00000898, -0.00000656, -0.00000802, 0.00000060, 0.00001683, -0.00001677, 0.00001368, -0.00002549, 0.00003752, -0.00003395, 0.00001970, -0.00001204, 0.00000407, 0.00000020, -0.00000108, 0.00000909, -0.00001533, 0.00000395, 0.00001681, -0.00002807, 0.00001719, -0.00001996, 0.00003566, -0.00003067, 0.00001941, -0.00002074, 0.00001777, -0.00001825, 0.00002056, -0.00002499, 0.00003426, -0.00003489, 0.00002903, -0.00003287, 0.00003490, -0.00003992, 0.00003883, -0.00004278, 0.00005048, -0.00005541, 0.00005614, -0.00006023, 0.00005151, -0.00006156, 0.00007062, -0.00008053, 0.00007812, -0.00009852, 0.00009515, -0.00013118, 0.00014590, -0.00021450, 0.00021902, -0.00049848, 0.00018331, 0.00103514, -0.00080759, 0.00029667, -0.00026655, 0.00017113, -0.00015008, 0.00011208, -0.00011439, 0.00008901, -0.00009092, 0.00007802, -0.00006879, 0.00006231, -0.00006918, 0.00005855, -0.00005463, 0.00004735, -0.00003989, 0.00003789, -0.00003806, 0.00003099, -0.00002987, 0.00002856, -0.00003290, 0.00002702, -0.00001736, 0.00001559, -0.00001696, 0.00001992, -0.00002318, 0.00002227, -0.00003061, 0.00002551, -0.00000698, 0.00000780, -0.00001541, -0.00000008, 0.00001177, -0.00000768, -0.00000890, 0.00002015, -0.00001913, 0.00002304, -0.00002800, 0.00003339, -0.00003857, 0.00003008, -0.00001269, 0.00000304, -0.00000706, 0.00000173, 0.00001236, -0.00000651, -0.00000970, 0.00000312, 0.00001302, -0.00000825, -0.00000590, 0.00000515, -0.00000396, 0.00000224, 0.00000695, -0.00000357, -0.00000546, },
		{ -0.00000228, -0.00000625, 0.00000904, -0.00000001, -0.00000144, 0.00000220, -0.00000246, -0.00001134, 0.00001616, -0.00000011, -0.00000354, -0.00001294, 0.00001447, 0.00000087, -0.00000379, -0.00000149, -0.00000704, 0.00002604, -0.00003643, 0.00003191, -0.00002920, 0.00002178, -0.00001534, 0.00001296, -0.00000410, -0.00001199, 0.00001449, 0.00000096, -0.00001880, 0.00001255, -0.00001055, 0.00002854, -0.00003302, 0.00002515, -0.00002596, 0.00002247, -0.00002101, 0.00002051, -0.00002117, 0.00003073, -0.00003564, 0.00003077, -0.00003351, 0.00003479, -0.00003971, 0.00003872, -0.00004074, 0.00004838, -0.00005685, 0.00006110, -0.00006962, 0.00006137, -0.00006804, 0.00007756, -0.00009334, 0.00009340, -0.00011675, 0.00011451, -0.00015264, 0.00017330, -0.00026984, 0.00030072, -0.00080189, 0.00100181, 0.00022421, -0.00051928, 0.00022743, -0.00021982, 0.00015061, -0.00013466, 0.00009801, -0.00010116, 0.00007746, -0.00007966, 0.00007211, -0.00006403, 0.00005513, -0.00006341, 0.00005699, -0.00005555, 0.00005144, -0.00004420, 0.00004083, -0.00004191, 0.00003466, -0.00003171, 0.00002901, -0.00003470, 0.00003303, -0.00002352, 0.00001732, -0.00001466, 0.00001522, -0.00001796, 0.00001619, -0.00002769, 0.00003199, -0.00001554, 0.00001034, -0.00002033, 0.00000948, 0.00000788, -0.00001686, 0.00000960, 0.00000273, -0.00000260, 0.00000506, -0.00000903, 0.00001757, -0.00003153, 0.00003645, -0.00002721, 0.00001585, -0.00001869, 0.00001595, 0.00000200, -0.00000509, -0.00001129, 0.00001185, 0.00000608, -0.00001132, -0.00000002, 0.00000212, -0.00000288, 0.00000378, 0.00000707, -0.00000997, 0.00000320, },
		{ -0.00001078, 0.00000314, 0.00000467, 0.00000074, -0.00000291, 0.00000516, -0.00001011, -0.00000180, 0.00001525, -0.00000482, -0.00000341, -0.00001256, 0.00002312, -0.00001376, 0.00001085, -0.00002075, 0.00002281, -0.00000783, -0.00000966, 0.00001434, -0.00002030, 0.00001903, -0.00001511, 0.00001456, -0.00001029, -0.00000911, 0.00002539, -0.00002269, 0.00000525, -0.00000800, 0.00001659, -0.00000257, -0.00000765, 0.00000403, -0.00000586, 0.00000300, -0.00000067, -0.00000197, 0.00000674, -0.00000098, -0.00000535, 0.00000059, -0.00000133, 0.00000039, -0.00000314, -0.00000005, 0.00000334, -0.00000141, -0.00000458, 0.00000743, -0.00001698, 0.00000677, -0.00000490, 0.00000540, -0.00001724, 0.00001070, -0.00002583, 0.00000916, -0.00002550, 0.00001226, -0.00007842, 0.00001903, -0.00057741, 0.00137001, -0.00052727, -0.00002429, -0.00005181, -0.00000678, -0.00000862, -0.00000501, -0.00001325, -0.00000268, -0.00001078, 0.00000332, -0.00000093, 0.00000145, -0.00000832, -0.00000322, 0.00000286, -0.00000625, 0.00000862, -0.00000598, 0.00000466, -0.00000975, 0.00000586, -0.00000363, 0.00000050, -0.00000768, 0.00001223, -0.00000776, 0.00000157, 0.00000334, -0.00000515, 0.00000356, -0.00000828, -0.00000267, 0.00001618, -0.00000704, -0.00000121, -0.00001100, 0.00000895, 0.00000568, -0.00002247, 0.00002670, -0.00001741, 0.00001829, -0.00001901, 0.00002085, -0.00001689, 0.00000207, 0.00001271, -0.00001538, 0.00000811, -0.00001322, 0.00001739, -0.00000120, -0.00000923, -0.00000510, 0.00001415, -0.00000063, -0.00001152, 0.00000524, -0.00000255, 0.00000042, 0.00000465, 0.00000430, -0.00001199, 0.00000910, },
		{ -0.00001374, 0.00001025, -0.00000041, 0.00000198, -0.00000326, 0.00000551, -0.00001561, 0.00001071, 0.00000650, -0.00000224, -0.00001012, -0.00000131, 0.00001730, -0.00001390, 0.00001055, -0.00002261, 0.00003687, -0.00003501, 0.00002111, -0.00001467, 0.00000327, 0.00000250, -0.00000368, 0.00000700, -0.00001205, -0.00000109, 0.00002383, -0.00003493, 0.00002300, -0.00002297, 0.00003761, -0.00003218, 0.00002155, -0.00002260, 0.00001956, -0.00002107, 0.00002281, -0.00002515, 0.00003386, -0.00003348, 0.00002760, -0.00003246, 0.00003341, -0.00003611, 0.00003502, -0.00003945, 0.00004738, -0.00005292, 0.00005283, -0.00005481, 0.00004600, -0.00005716, 0.00006689, -0.00007883, 0.00007596, -0.00009283, 0.00008900, -0.00012377, 0.00013642, -0.00020351, 0.00020372, -0.00047091, 0.00014548, 0.00105034, -0.00079342, 0.00027947, -0.00025421, 0.00016207, -0.00014210, 0.00010514, -0.00010794, 0.00008100, -0.00008525, 0.00007551, -0.00006679, 0.00006045, -0.00006577, 0.00005326, -0.00005009, 0.00004434, -0.00003731, 0.00003546, -0.00003493, 0.00002594, -0.00002519, 0.00002537, -0.00002911, 0.00002245, -0.00001313, 0.00001049, -0.00001272, 0.00001695, -0.00001978, 0.00001867, -0.00002733, 0.00002184, -0.00000344, 0.00000354, -0.00001269, 0.00000007, 0.00000719, -0.00000209, -0.00001400, 0.00002688, -0.00002179, 0.00002310, -0.00002531, 0.00003333, -0.00003892, 0.00003254, -0.00001772, 0.00000726, -0.00001028, 0.00000321, 0.00000912, 0.00000066, -0.00001597, 0.00000660, 0.00000922, -0.00000472, -0.00000925, 0.00000774, -0.00000616, 0.00000434, 0.00000476, -0.00000088, -0.00000823, 0.00000841, },
		{ -0.00000860, 0.00001076, -0.00000212, 0.00000108, -0.00000071, 0.00000068, -0.00001359, 0.00001827, -0.00000270, 0.00000202, -0.00001816, 0.00001519, 0.00000076, -0.00000066, -0.00000604, -0.00000112, 0.00002079, -0.00003274, 0.00002903, -0.00002906, 0.00001957, -0.00001086, 0.00000588, -0.00000020, -0.00001453, 0.00001520, 0.00000303, -0.00002253, 0.00001735, -0.00001372, 0.00003055, -0.00003423, 0.00002699, -0.00002780, 0.00002430, -0.00002446, 0.00002478, -0.00002411, 0.00003317, -0.00003698, 0.00003166, -0.00003574, 0.00003697, -0.00003963, 0.00003778, -0.00003972, 0.00004726, -0.00005655, 0.00006072, -0.00006706, 0.00005769, -0.00006460, 0.00007415, -0.00009235, 0.00009392, -0.00011480, 0.00011222, -0.00014960, 0.00016839, -0.00026394, 0.00029176, -0.00077731, 0.00095262, 0.00026548, -0.00053128, 0.00022775, -0.00021868, 0.00014997, -0.00013377, 0.00009725, -0.00010050, 0.00007384, -0.00007610, 0.00007114, -0.00006426, 0.00005649, -0.00006403, 0.00005518, -0.00005323, 0.00005012, -0.00004354, 0.00004086, -0.00004199, 0.00003269, -0.00002905, 0.00002760, -0.00003300, 0.00003032, -0.00002096, 0.00001341, -0.00001063, 0.00001220, -0.00001455, 0.00001219, -0.00002358, 0.00002704, -0.00001027, 0.00000313, -0.00001246, 0.00000260, 0.00001015, -0.00001647, 0.00000858, 0.00000754, -0.00000566, 0.00000611, -0.00000566, 0.00001484, -0.00002788, 0.00003364, -0.00002766, 0.00001697, -0.00001966, 0.00001475, 0.00000284, -0.00000150, -0.00001616, 0.00001485, 0.00000261, -0.00000775, -0.00000371, 0.00000526, -0.00000557, 0.00000625, 0.00000462, -0.00000695, -0.00000025, 0.00000095, },
		{ 0.00000180, 0.00000506, 0.00000036, -0.00000226, 0.00000406, -0.00000843, -0.00000262, 0.00001530, -0.00000444, -0.00000051, -0.00001723, 0.00002390, -0.00001203, 0.00001077, -0.00002204, 0.00002636, -0.00001257, -0.00000516, 0.00000979, -0.00001909, 0.00001748, -0.00001173, 0.00000682, -0.00000348, -0.00001546, 0.00003083, -0.00002592, 0.00000636, -0.00000644, 0.00001506, -0.00000169, -0.00000752, 0.00000433, -0.00000623, 0.00000306, -0.00000237, 0.00000175, 0.00000313, 0.00000279, -0.00000840, 0.00000295, -0.00000519, 0.00000554, -0.00000698, 0.00000288, 0.00000067, 0.00000091, -0.00000782, 0.00001156, -0.00001965, 0.00000800, -0.00000579, 0.00000539, -0.00001915, 0.00001569, -0.00003004, 0.00001409, -0.00003129, 0.00001790, -0.00008559, 0.00002850, -0.00057446, 0.00133135, -0.00047674, -0.00005501, -0.00003490, -0.00001819, 0.00000080, -0.00001285, -0.00000610, -0.00001002, -0.00000696, 0.00000173, 0.00000162, -0.00000196, -0.00000324, -0.00000872, 0.00000616, -0.00000803, 0.00001050, -0.00000821, 0.00000776, -0.00001381, 0.00000842, -0.00000486, 0.00000266, -0.00000996, 0.00001354, -0.00000937, 0.00000188, 0.00000381, -0.00000487, 0.00000346, -0.00000881, -0.00000187, 0.00001431, -0.00000483, -0.00000585, -0.00000369, -0.00000047, 0.00001236, -0.00002642, 0.00003012, -0.00001580, 0.00001580, -0.00001631, 0.00002248, -0.00001907, 0.00000650, 0.00000674, -0.00001169, 0.00000556, -0.00001129, 0.00001312, 0.00000459, -0.00001150, -0.00000535, 0.00001342, -0.00000060, -0.00001097, 0.00000408, -0.00000150, -0.00000030, 0.00000511, 0.00000400, -0.00001111, 0.00000746, -0.00000881, },
		{ 0.00001057, -0.00000168, 0.00000361, -0.00000491, 0.00000680, -0.00001592, 0.00001151, 0.00000455, 0.00000150, -0.00001135, -0.00000359, 0.00001723, -0.00001016, 0.00000662, -0.00001876, 0.00003460, -0.00003470, 0.00002175, -0.00001692, 0.00000271, 0.00000415, -0.00000499, 0.00000344, -0.00000740, -0.00000659, 0.00002981, -0.00004021, 0.00002718, -0.00002414, 0.00003754, -0.00003199, 0.00002239, -0.00002333, 0.00002038, -0.00002288, 0.00002386, -0.00002389, 0.00003184, -0.00003057, 0.00002481, -0.00003059, 0.00003035, -0.00003080, 0.00002974, -0.00003455, 0.00004247, -0.00004842, 0.00004747, -0.00004767, 0.00003913, -0.00005129, 0.00006129, -0.00007485, 0.00007130, -0.00008455, 0.00008014, -0.00011312, 0.00012302, -0.00018794, 0.00018305, -0.00043871, 0.00011307, 0.00104608, -0.00075949, 0.00024990, -0.00023362, 0.00014684, -0.00012935, 0.00009426, -0.00009794, 0.00006988, -0.00007672, 0.00007029, -0.00006222, 0.00005601, -0.00005973, 0.00004564, -0.00004366, 0.00003971, -0.00003334, 0.00003178, -0.00003072, 0.00002011, -0.00001992, 0.00002160, -0.00002469, 0.00001749, -0.00000895, 0.00000566, -0.00000867, 0.00001390, -0.00001608, 0.00001468, -0.00002354, 0.00001790, -0.00000032, 0.00000008, -0.00001082, 0.00000133, 0.00000116, 0.00000411, -0.00001872, 0.00003259, -0.00002294, 0.00002181, -0.00002107, 0.00003172, -0.00003740, 0.00003318, -0.00002182, 0.00001090, -0.00001323, 0.00000509, 0.00000513, 0.00000800, -0.00002180, 0.00000987, 0.00000521, -0.00000120, -0.00001209, 0.00000965, -0.00000776, 0.00000604, 0.00000276, 0.00000151, -0.00001028, 0.00000957, -0.00001400, },
		{ 0.00001183, -0.00000380, 0.00000328, -0.00000321, 0.00000312, -0.00001470, 0.00001908, -0.00000434, 0.00000658, -0.00002182, 0.00001439, 0.00000144, 0.00000219, -0.00001033, 0.00000467, 0.00001474, -0.00002780, 0.00002500, -0.00002782, 0.00001671, -0.00000616, -0.00000056, 0.00000237, -0.00001506, 0.00001394, 0.00000585, -0.00002617, 0.00002174, -0.00001607, 0.00003120, -0.00003399, 0.00002756, -0.00002847, 0.00002513, -0.00002695, 0.00002800, -0.00002580, 0.00003405, -0.00003669, 0.00003103, -0.00003632, 0.00003734, -0.00003770, 0.00003505, -0.00003686, 0.00004404, -0.00005375, 0.00005748, -0.00006171, 0.00005155, -0.00005877, 0.00006804, -0.00008809, 0.00009065, -0.00010868, 0.00010541, -0.00014118, 0.00015680, -0.00024944, 0.00027080, -0.00073585, 0.00089056, 0.00030429, -0.00053285, 0.00021926, -0.00021075, 0.00014380, -0.00012842, 0.00009281, -0.00009653, 0.00006737, -0.00006996, 0.00006768, -0.00006211, 0.00005542, -0.00006198, 0.00005076, -0.00004853, 0.00004660, -0.00004086, 0.00003895, -0.00004024, 0.00002916, -0.00002511, 0.00002499, -0.00003001, 0.00002643, -0.00001766, 0.00000921, -0.00000655, 0.00000906, -0.00001087, 0.00000781, -0.00001879, 0.00002134, -0.00000479, -0.00000372, -0.00000521, -0.00000320, 0.00001044, -0.00001421, 0.00000632, 0.00001300, -0.00000868, 0.00000710, -0.00000193, 0.00001163, -0.00002319, 0.00002917, -0.00002683, 0.00001701, -0.00001971, 0.00001332, 0.00000313, 0.00000251, -0.00002081, 0.00001767, -0.00000122, -0.00000370, -0.00000751, 0.00000822, -0.00000800, 0.00000858, 0.00000199, -0.00000363, -0.00000374, 0.00000383, -0.00001030, },
		{ 0.00000555, -0.00000024, -0.00000124, 0.00000251, -0.00000643, -0.00000329, 0.00001478, -0.00000364, 0.00000197, -0.00002075, 0.00002303, -0.00000893, 0.00000972, -0.00002228, 0.00002861, -0.00001671, -0.00000050, 0.00000473, -0.00001693, 0.00001497, -0.00000744, -0.00000119, 0.00000282, -0.00002011, 0.00003368, -0.00002698, 0.00000590, -0.00000357, 0.00001254, -0.00000045, -0.00000737, 0.00000456, -0.00000650, 0.00000315, -0.00000419, 0.00000559, -0.00000053, 0.00000631, -0.00001106, 0.00000500, -0.00000867, 0.00001012, -0.00001017, 0.00000526, -0.00000156, 0.00000271, -0.00001022, 0.00001436, -0.00002080, 0.00000801, -0.00000577, 0.00000451, -0.00001987, 0.00001903, -0.00003229, 0.00001692, -0.00003455, 0.00002011, -0.00008773, 0.00002901, -0.00055307, 0.00126921, -0.00041273, -0.00008872, -0.00001776, -0.00002925, 0.00000960, -0.00002014, 0.00000070, -0.00001701, -0.00000332, 0.00000017, 0.00000410, -0.00000529, 0.00000167, -0.00001363, 0.00000863, -0.00000896, 0.00001148, -0.00000952, 0.00000985, -0.00001674, 0.00000993, -0.00000525, 0.00000405, -0.00001126, 0.00001370, -0.00001004, 0.00000165, 0.00000442, -0.00000460, 0.00000339, -0.00000940, -0.00000065, 0.00001160, -0.00000193, -0.00001072, 0.00000362, -0.00000947, 0.00001744, -0.00002822, 0.00003142, -0.00001227, 0.00001201, -0.00001231, 0.00002301, -0.00002007, 0.00001016, 0.00000059, -0.00000764, 0.00000253, -0.00000862, 0.00000847, 0.00000987, -0.00001299, -0.00000585, 0.00001270, -0.00000105, -0.00000957, 0.00000219, -0.00000001, -0.00000129, 0.00000579, 0.00000316, -0.00000938, 0.00000496, -0.00000622, 0.00000019, },
		{ -0.00000224, 0.00000451, -0.00000566, 0.00000698, -0.00001495, 0.00001127, 0.00000317, 0.00000459, -0.00001176, -0.00000605, 0.00001654, -0.00000566, 0.00000212, -0.00001419, 0.00003092, -0.00003304, 0.00002157, -0.00001876, 0.00000241, 0.00000508, -0.00000508, -0.00000128, -0.00000192, -0.00001184, 0.00003409, -0.00004336, 0.00002931, -0.00002324, 0.00003537, -0.00003007, 0.00002189, -0.00002288, 0.00002013, -0.00002355, 0.00002364, -0.00002128, 0.00002836, -0.00002639, 0.00002092, -0.00002749, 0.00002606, -0.00002440, 0.00002348, -0.00002857, 0.00003623, -0.00004238, 0.00004063, -0.00003945, 0.00003149, -0.00004449, 0.00005431, -0.00006902, 0.00006460, -0.00007430, 0.00006929, -0.00010010, 0.00010680, -0.00016911, 0.00015870, -0.00040399, 0.00008787, 0.00102211, -0.00070707, 0.00020965, -0.00020624, 0.00012662, -0.00011285, 0.00008035, -0.00008522, 0.00005645, -0.00006597, 0.00006280, -0.00005548, 0.00004942, -0.00005159, 0.00003634, -0.00003592, 0.00003392, -0.00002842, 0.00002729, -0.00002591, 0.00001402, -0.00001456, 0.00001767, -0.00002008, 0.00001260, -0.00000524, 0.00000149, -0.00000516, 0.00001104, -0.00001236, 0.00001062, -0.00001955, 0.00001404, 0.00000210, -0.00000236, -0.00000984, 0.00000349, -0.00000574, 0.00001012, -0.00002240, 0.00003679, -0.00002230, 0.00001914, -0.00001533, 0.00002867, -0.00003407, 0.00003187, -0.00002473, 0.00001372, -0.00001569, 0.00000708, 0.00000086, 0.00001491, -0.00002665, 0.00001252, 0.00000138, 0.00000192, -0.00001406, 0.00001059, -0.00000855, 0.00000715, 0.00000115, 0.00000335, -0.00001140, 0.00000955, -0.00001308, 0.00000995, },
		{ -0.00000485, 0.00000489, -0.00000503, 0.00000461, -0.00001447, 0.00001845, -0.00000485, 0.00000992, -0.00002377, 0.00001214, 0.00000282, 0.00000468, -0.00001414, 0.00000997, 0.00000834, -0.00002198, 0.00002010, -0.00002574, 0.00001351, -0.00000165, -0.00000595, 0.00000335, -0.00001355, 0.00001102, 0.00000900, -0.00002929, 0.00002526, -0.00001722, 0.00003027, -0.00003214, 0.00002671, -0.00002783, 0.00002477, -0.00002822, 0.00002985, -0.00002601, 0.00003320, -0.00003471, 0.00002884, -0.00003518, 0.00003585, -0.00003401, 0.00003071, -0.00003237, 0.00003895, -0.00004871, 0.00005174, -0.00005402, 0.00004346, -0.00005105, 0.00005969, -0.00008095, 0.00008396, -0.00009883, 0.00009460, -0.00012808, 0.00013941, -0.00022753, 0.00023963, -0.00068031, 0.00081910, 0.00033796, -0.00052288, 0.00020183, -0.00019620, 0.00013233, -0.00011887, 0.00008495, -0.00008953, 0.00005837, -0.00006153, 0.00006189, -0.00005765, 0.00005198, -0.00005740, 0.00004400, -0.00004180, 0.00004116, -0.00003639, 0.00003533, -0.00003690, 0.00002441, -0.00002026, 0.00002151, -0.00002608, 0.00002175, -0.00001403, 0.00000514, -0.00000281, 0.00000616, -0.00000728, 0.00000346, -0.00001378, 0.00001543, 0.00000032, -0.00000958, 0.00000081, -0.00000743, 0.00000864, -0.00001041, 0.00000323, 0.00001867, -0.00001128, 0.00000787, 0.00000215, 0.00000813, -0.00001773, 0.00002325, -0.00002479, 0.00001601, -0.00001890, 0.00001171, 0.00000298, 0.00000659, -0.00002486, 0.00001995, -0.00000500, 0.00000038, -0.00001097, 0.00001059, -0.00000984, 0.00001045, -0.00000047, -0.00000038, -0.00000688, 0.00000604, -0.00001109, 0.00001205, },
		{ -0.00000090, -0.00000005, 0.00000077, -0.00000438, -0.00000358, 0.00001354, -0.00000228, 0.00000378, -0.00002287, 0.00002051, -0.00000462, 0.00000780, -0.00002149, 0.00002944, -0.00001995, 0.00000396, -0.00000052, -0.00001415, 0.00001184, -0.00000273, -0.00000885, 0.00000798, -0.00002258, 0.00003372, -0.00002585, 0.00000398, 0.00000033, 0.00000934, 0.00000092, -0.00000706, 0.00000458, -0.00000655, 0.00000313, -0.00000589, 0.00000919, -0.00000385, 0.00000921, -0.00001299, 0.00000644, -0.00001138, 0.00001370, -0.00001237, 0.00000680, -0.00000309, 0.00000375, -0.00001153, 0.00001562, -0.00002036, 0.00000679, -0.00000485, 0.00000277, -0.00001932, 0.00002058, -0.00003239, 0.00001745, -0.00003509, 0.00001874, -0.00008478, 0.00002093, -0.00051487, 0.00118681, -0.00033888, -0.00012269, -0.00000213, -0.00003877, 0.00001686, -0.00002615, 0.00000653, -0.00002308, -0.00000025, -0.00000107, 0.00000617, -0.00000815, 0.00000592, -0.00001751, 0.00000998, -0.00000886, 0.00001139, -0.00000974, 0.00001074, -0.00001835, 0.00001027, -0.00000474, 0.00000460, -0.00001154, 0.00001276, -0.00000983, 0.00000101, 0.00000504, -0.00000424, 0.00000326, -0.00000990, 0.00000078, 0.00000838, 0.00000121, -0.00001524, 0.00001016, -0.00001721, 0.00002027, -0.00002765, 0.00003062, -0.00000702, 0.00000723, -0.00000723, 0.00002256, -0.00001986, 0.00001277, -0.00000540, -0.00000354, -0.00000071, -0.00000556, 0.00000384, 0.00001428, -0.00001356, -0.00000662, 0.00001203, -0.00000193, -0.00000746, -0.00000021, 0.00000172, -0.00000237, 0.00000656, 0.00000197, -0.00000703, 0.00000188, -0.00000329, -0.00000144, 0.00000596, },
		{ 0.00000468, -0.00000552, 0.00000611, -0.00001287, 0.00001016, 0.00000230, 0.00000705, -0.00001148, -0.00000847, 0.00001519, -0.00000058, -0.00000265, -0.00000920, 0.00002609, -0.00003017, 0.00002058, -0.00002015, 0.00000235, 0.00000534, -0.00000415, -0.00000674, 0.00000379, -0.00001620, 0.00003616, -0.00004404, 0.00002919, -0.00002023, 0.00003120, -0.00002655, 0.00002014, -0.00002131, 0.00001885, -0.00002308, 0.00002224, -0.00001754, 0.00002377, -0.00002132, 0.00001628, -0.00002352, 0.00002096, -0.00001747, 0.00001681, -0.00002207, 0.00002924, -0.00003541, 0.00003294, -0.00003082, 0.00002373, -0.00003731, 0.00004645, -0.00006181, 0.00005646, -0.00006279, 0.00005730, -0.00008568, 0.00008890, -0.00014839, 0.00013234, -0.00036862, 0.00007097, 0.00097929, -0.00063852, 0.00016117, -0.00017404, 0.00010299, -0.00009392, 0.00006451, -0.00007078, 0.00004163, -0.00005378, 0.00005368, -0.00004714, 0.00004131, -0.00004208, 0.00002610, -0.00002755, 0.00002756, -0.00002305, 0.00002245, -0.00002097, 0.00000816, -0.00000956, 0.00001400, -0.00001568, 0.00000817, -0.00000236, -0.00000172, -0.00000240, 0.00000859, -0.00000889, 0.00000676, -0.00001565, 0.00001050, 0.00000366, -0.00000375, -0.00000965, 0.00000617, -0.00001279, 0.00001515, -0.00002447, 0.00003916, -0.00001978, 0.00001519, -0.00000832, 0.00002441, -0.00002913, 0.00002861, -0.00002631, 0.00001554, -0.00001745, 0.00000889, -0.00000323, 0.00002083, -0.00003007, 0.00001417, -0.00000188, 0.00000430, -0.00001494, 0.00001042, -0.00000843, 0.00000755, 0.00000009, 0.00000452, -0.00001151, 0.00000838, -0.00001106, 0.00000848, -0.00000213, },
		{ 0.00000575, -0.00000596, 0.00000500, -0.00001288, 0.00001642, -0.00000424, 0.00001194, -0.00002403, 0.00000862, 0.00000475, 0.00000678, -0.00001726, 0.00001446, 0.00000201, -0.00001570, 0.00001466, -0.00002311, 0.00001027, 0.00000231, -0.00001003, 0.00000268, -0.00001025, 0.00000688, 0.00001201, -0.00003145, 0.00002752, -0.00001694, 0.00002764, -0.00002868, 0.00002444, -0.00002585, 0.00002317, -0.00002810, 0.00003017, -0.00002464, 0.00003067, -0.00003115, 0.00002519, -0.00003243, 0.00003268, -0.00002884, 0.00002510, -0.00002665, 0.00003242, -0.00004189, 0.00004403, -0.00004466, 0.00003411, -0.00004208, 0.00004973, -0.00007151, 0.00007444, -0.00008598, 0.00008065, -0.00011130, 0.00011751, -0.00019988, 0.00020058, -0.00061403, 0.00074192, 0.00036419, -0.00050094, 0.00017598, -0.00017571, 0.00011622, -0.00010571, 0.00007418, -0.00007999, 0.00004739, -0.00005131, 0.00005414, -0.00005116, 0.00004644, -0.00005067, 0.00003540, -0.00003354, 0.00003426, -0.00003055, 0.00003039, -0.00003237, 0.00001888, -0.00001493, 0.00001757, -0.00002161, 0.00001672, -0.00001050, 0.00000159, 0.00000024, 0.00000379, -0.00000412, -0.00000046, -0.00000900, 0.00000983, 0.00000452, -0.00001396, 0.00000514, -0.00000986, 0.00000494, -0.00000559, -0.00000011, 0.00002408, -0.00001314, 0.00000831, 0.00000654, 0.00000452, -0.00001186, 0.00001614, -0.00002170, 0.00001408, -0.00001733, 0.00000995, 0.00000255, 0.00001038, -0.00002792, 0.00002135, -0.00000835, 0.00000406, -0.00001368, 0.00001204, -0.00001081, 0.00001161, -0.00000249, 0.00000245, -0.00000935, 0.00000732, -0.00001083, 0.00001133, -0.00000513, },
		{ 0.00000111, -0.00000090, -0.00000257, -0.00000333, 0.00001152, -0.00000032, 0.00000479, -0.00002352, 0.00001649, 0.00000060, 0.00000523, -0.00001977, 0.00002882, -0.00002209, 0.00000792, -0.00000563, -0.00001104, 0.00000845, 0.00000191, -0.00001560, 0.00001153, -0.00002263, 0.00003105, -0.00002277, 0.00000092, 0.00000489, 0.00000581, 0.00000218, -0.00000642, 0.00000429, -0.00000628, 0.00000285, -0.00000721, 0.00001220, -0.00000645, 0.00001116, -0.00001396, 0.00000703, -0.00001305, 0.00001600, -0.00001332, 0.00000733, -0.00000379, 0.00000392, -0.00001165, 0.00001531, -0.00001841, 0.00000449, -0.00000318, 0.00000032, -0.00001759, 0.00002034, -0.00003038, 0.00001576, -0.00003300, 0.00001403, -0.00007723, 0.00000539, -0.00046246, 0.00108834, -0.00025935, -0.00015414, 0.00001037, -0.00004571, 0.00002183, -0.00003028, 0.00001085, -0.00002779, 0.00000193, -0.00000173, 0.00000758, -0.00001017, 0.00000913, -0.00002004, 0.00001003, -0.00000767, 0.00001020, -0.00000884, 0.00001040, -0.00001861, 0.00000948, -0.00000340, 0.00000437, -0.00001090, 0.00001086, -0.00000892, 0.00000013, 0.00000552, -0.00000369, 0.00000294, -0.00001015, 0.00000218, 0.00000500, 0.00000413, -0.00001885, 0.00001525, -0.00002299, 0.00002044, -0.00002474, 0.00002798, -0.00000043, 0.00000186, -0.00000140, 0.00002134, -0.00001849, 0.00001417, -0.00001095, 0.00000033, -0.00000388, -0.00000244, -0.00000037, 0.00001756, -0.00001317, -0.00000761, 0.00001133, -0.00000310, -0.00000487, -0.00000285, 0.00000344, -0.00000333, 0.00000723, 0.00000064, -0.00000435, -0.00000141, -0.00000039, -0.00000283, 0.00000611, -0.00000143, },
		{ -0.00000461, 0.00000439, -0.00000994, 0.00000840, 0.00000179, 0.00000898, -0.00001069, -0.00001064, 0.00001318, 0.00000481, -0.00000738, -0.00000414, 0.00002043, -0.00002625, 0.00001883, -0.00002107, 0.00000245, 0.00000504, -0.00000250, -0.00001243, 0.00000911, -0.00001913, 0.00003576, -0.00004217, 0.00002686, -0.00001533, 0.00002534, -0.00002172, 0.00001737, -0.00001879, 0.00001667, -0.00002160, 0.00001984, -0.00001299, 0.00001847, -0.00001581, 0.00001132, -0.00001911, 0.00001556, -0.00001057, 0.00001034, -0.00001565, 0.00002207, -0.00002810, 0.00002506, -0.00002246, 0.00001640, -0.00003027, 0.00003823, -0.00005377, 0.00004745, -0.00005076, 0.00004497, -0.00007081, 0.00007045, -0.00012709, 0.00010551, -0.00033409, 0.00006266, 0.00091955, -0.00055713, 0.00010750, -0.00013933, 0.00007774, -0.00007400, 0.00004796, -0.00005571, 0.00002646, -0.00004105, 0.00004366, -0.00003789, 0.00003239, -0.00003198, 0.00001574, -0.00001923, 0.00002118, -0.00001771, 0.00001772, -0.00001635, 0.00000299, -0.00000530, 0.00001089, -0.00001181, 0.00000449, -0.00000049, -0.00000383, -0.00000054, 0.00000668, -0.00000583, 0.00000330, -0.00001201, 0.00000742, 0.00000431, -0.00000417, -0.00001001, 0.00000890, -0.00001927, 0.00001847, -0.00002454, 0.00003956, -0.00001546, 0.00001025, -0.00000035, 0.00001934, -0.00002297, 0.00002357, -0.00002653, 0.00001625, -0.00001834, 0.00001023, -0.00000669, 0.00002528, -0.00003173, 0.00001460, -0.00000433, 0.00000575, -0.00001460, 0.00000913, -0.00000743, 0.00000726, -0.00000041, 0.00000497, -0.00001067, 0.00000620, -0.00000816, 0.00000633, -0.00000145, 0.00000419, },
		{ -0.00000596, 0.00000429, -0.00001005, 0.00001319, -0.00000264, 0.00001274, -0.00002279, 0.00000417, 0.00000699, 0.00000847, -0.00001953, 0.00001785, -0.00000384, -0.00000935, 0.00000901, -0.00002019, 0.00000722, 0.00000550, -0.00001269, 0.00000052, -0.00000558, 0.00000210, 0.00001443, -0.00003230, 0.00002827, -0.00001510, 0.00002340, -0.00002375, 0.00002089, -0.00002265, 0.00002038, -0.00002660, 0.00002894, -0.00002178, 0.00002663, -0.00002627, 0.00002038, -0.00002833, 0.00002813, -0.00002259, 0.00001871, -0.00002021, 0.00002500, -0.00003389, 0.00003505, -0.00003441, 0.00002425, -0.00003259, 0.00003888, -0.00006049, 0.00006289, -0.00007106, 0.00006462, -0.00009210, 0.00009266, -0.00016847, 0.00015637, -0.00054063, 0.00066257, 0.00038123, -0.00046738, 0.00014287, -0.00015046, 0.00009650, -0.00008983, 0.00006128, -0.00006863, 0.00003514, -0.00003996, 0.00004497, -0.00004310, 0.00003926, -0.00004237, 0.00002565, -0.00002441, 0.00002649, -0.00002388, 0.00002462, -0.00002714, 0.00001306, -0.00000960, 0.00001357, -0.00001706, 0.00001180, -0.00000742, -0.00000112, 0.00000236, 0.00000217, -0.00000166, -0.00000365, -0.00000483, 0.00000494, 0.00000746, -0.00001655, 0.00000757, -0.00001048, -0.00000025, -0.00000043, -0.00000313, 0.00002877, -0.00001396, 0.00000833, 0.00001119, 0.00000103, -0.00000596, 0.00000824, -0.00001781, 0.00001139, -0.00001511, 0.00000806, 0.00000205, 0.00001352, -0.00002966, 0.00002157, -0.00001092, 0.00000696, -0.00001533, 0.00001234, -0.00001074, 0.00001190, -0.00000383, 0.00000459, -0.00001090, 0.00000753, -0.00000950, 0.00000968, -0.00000462, 0.00000579, },
		{ -0.00000225, -0.00000121, -0.00000243, 0.00000876, 0.00000222, 0.00000498, -0.00002277, 0.00001131, 0.00000634, 0.00000230, -0.00001728, 0.00002681, -0.00002300, 0.00001112, -0.00001033, -0.00000793, 0.00000513, 0.00000604, -0.00002102, 0.00001321, -0.00002033, 0.00002602, -0.00001818, -0.00000285, 0.00000964, 0.00000237, 0.00000305, -0.00000532, 0.00000359, -0.00000561, 0.00000223, -0.00000796, 0.00001433, -0.00000807, 0.00001198, -0.00001382, 0.00000663, -0.00001352, 0.00001686, -0.00001294, 0.00000682, -0.00000365, 0.00000325, -0.00001063, 0.00001357, -0.00001521, 0.00000141, -0.00000103, -0.00000258, -0.00001489, 0.00001852, -0.00002653, 0.00001213, -0.00002868, 0.00000655, -0.00006600, -0.00001589, -0.00039921, 0.00097858, -0.00017842, -0.00018053, 0.00001846, -0.00004933, 0.00002402, -0.00003216, 0.00001335, -0.00003084, 0.00000303, -0.00000170, 0.00000813, -0.00001114, 0.00001105, -0.00002107, 0.00000881, -0.00000548, 0.00000803, -0.00000695, 0.00000894, -0.00001763, 0.00000774, -0.00000142, 0.00000353, -0.00000953, 0.00000827, -0.00000756, -0.00000078, 0.00000571, -0.00000286, 0.00000234, -0.00001000, 0.00000331, 0.00000184, 0.00000640, -0.00002109, 0.00001839, -0.00002635, 0.00001785, -0.00001984, 0.00002399, 0.00000698, -0.00000356, 0.00000478, 0.00001970, -0.00001616, 0.00001426, -0.00001581, 0.00000373, -0.00000670, 0.00000045, -0.00000390, 0.00001958, -0.00001191, -0.00000869, 0.00001052, -0.00000439, -0.00000208, -0.00000542, 0.00000485, -0.00000394, 0.00000761, -0.00000060, -0.00000168, -0.00000453, 0.00000215, -0.00000376, 0.00000586, -0.00000170, 0.00000203, },
		{ 0.00000211, -0.00000653, 0.00000631, 0.00000148, 0.00001049, -0.00000959, -0.00001240, 0.00001059, 0.00001023, -0.00001173, 0.00000065, 0.00001429, -0.00002153, 0.00001640, -0.00002149, 0.00000262, 0.00000434, -0.00000046, -0.00001786, 0.00001353, -0.00002030, 0.00003289, -0.00003795, 0.00002261, -0.00000895, 0.00001829, -0.00001604, 0.00001392, -0.00001563, 0.00001384, -0.00001933, 0.00001675, -0.00000806, 0.00001296, -0.00001034, 0.00000649, -0.00001470, 0.00001033, -0.00000426, 0.00000459, -0.00000983, 0.00001528, -0.00002101, 0.00001759, -0.00001491, 0.00000996, -0.00002378, 0.00003012, -0.00004538, 0.00003818, -0.00003889, 0.00003306, -0.00005637, 0.00005246, -0.00010636, 0.00007949, -0.00030145, 0.00006249, 0.00084575, -0.00046688, 0.00005202, -0.00010457, 0.00005271, -0.00005455, 0.00003192, -0.00004111, 0.00001193, -0.00002865, 0.00003348, -0.00002843, 0.00002340, -0.00002209, 0.00000602, -0.00001160, 0.00001528, -0.00001285, 0.00001348, -0.00001241, -0.00000118, -0.00000205, 0.00000855, -0.00000869, 0.00000175, 0.00000028, -0.00000481, 0.00000044, 0.00000531, -0.00000328, 0.00000034, -0.00000876, 0.00000487, 0.00000414, -0.00000382, -0.00001062, 0.00001119, -0.00002449, 0.00001953, -0.00002247, 0.00003809, -0.00000958, 0.00000472, 0.00000815, 0.00001394, -0.00001608, 0.00001706, -0.00002549, 0.00001583, -0.00001826, 0.00001088, -0.00000918, 0.00002794, -0.00003148, 0.00001371, -0.00000585, 0.00000617, -0.00001310, 0.00000684, -0.00000569, 0.00000639, -0.00000039, 0.00000479, -0.00000903, 0.00000330, -0.00000471, 0.00000380, -0.00000041, 0.00000324, -0.00000314, },
		{ 0.00000264, -0.00000628, 0.00000910, -0.00000032, 0.00001253, -0.00002037, -0.00000078, 0.00000926, 0.00000978, -0.00002086, 0.00001995, -0.00000889, -0.00000329, 0.00000343, -0.00001721, 0.00000453, 0.00000777, -0.00001397, -0.00000280, -0.00000012, -0.00000273, 0.00001586, -0.00003166, 0.00002741, -0.00001181, 0.00001778, -0.00001766, 0.00001635, -0.00001848, 0.00001659, -0.00002386, 0.00002633, -0.00001769, 0.00002146, -0.00002048, 0.00001479, -0.00002329, 0.00002265, -0.00001580, 0.00001211, -0.00001365, 0.00001732, -0.00002539, 0.00002558, -0.00002409, 0.00001464, -0.00002329, 0.00002789, -0.00004869, 0.00005018, -0.00005509, 0.00004768, -0.00007186, 0.00006655, -0.00013544, 0.00010980, -0.00046372, 0.00058430, 0.00038809, -0.00042332, 0.00010425, -0.00012204, 0.00007450, -0.00007236, 0.00004718, -0.00005629, 0.00002244, -0.00002823, 0.00003501, -0.00003406, 0.00003107, -0.00003318, 0.00001551, -0.00001516, 0.00001849, -0.00001697, 0.00001855, -0.00002170, 0.00000745, -0.00000470, 0.00000991, -0.00001279, 0.00000736, -0.00000510, -0.00000282, 0.00000341, 0.00000143, -0.00000005, -0.00000589, -0.00000151, 0.00000104, 0.00000893, -0.00001726, 0.00000811, -0.00000952, -0.00000635, 0.00000429, -0.00000528, 0.00003239, -0.00001353, 0.00000793, 0.00001601, -0.00000209, -0.00000049, -0.00000002, -0.00001342, 0.00000817, -0.00001240, 0.00000604, 0.00000167, 0.00001570, -0.00002985, 0.00002044, -0.00001245, 0.00000882, -0.00001570, 0.00001137, -0.00000958, 0.00001125, -0.00000437, 0.00000589, -0.00001142, 0.00000668, -0.00000722, 0.00000725, -0.00000341, 0.00000508, -0.00000507, },
		{ -0.00000045, -0.00000089, 0.00000538, 0.00000517, 0.00000447, -0.00002086, 0.00000540, 0.00001213, -0.00000068, -0.00001426, 0.00002356, -0.00002263, 0.00001338, -0.00001435, -0.00000507, 0.00000216, 0.00000934, -0.00002482, 0.00001300, -0.00001601, 0.00001925, -0.00001267, -0.00000681, 0.00001410, -0.00000061, 0.00000334, -0.00000368, 0.00000246, -0.00000451, 0.00000119, -0.00000802, 0.00001541, -0.00000857, 0.00001160, -0.00001259, 0.00000528, -0.00001280, 0.00001630, -0.00001130, 0.00000538, -0.00000282, 0.00000189, -0.00000868, 0.00001071, -0.00001118, -0.00000207, 0.00000127, -0.00000563, -0.00001153, 0.00001546, -0.00002126, 0.00000708, -0.00002276, -0.00000281, -0.00005237, -0.00004073, -0.00032894, 0.00086258, -0.00010023, -0.00019973, 0.00002129, -0.00004927, 0.00002325, -0.00003168, 0.00001389, -0.00003212, 0.00000298, -0.00000095, 0.00000780, -0.00001098, 0.00001162, -0.00002064, 0.00000647, -0.00000254, 0.00000513, -0.00000431, 0.00000661, -0.00001568, 0.00000531, 0.00000091, 0.00000232, -0.00000770, 0.00000531, -0.00000603, -0.00000152, 0.00000550, -0.00000168, 0.00000137, -0.00000932, 0.00000397, -0.00000083, 0.00000773, -0.00002169, 0.00001930, -0.00002712, 0.00001275, -0.00001356, 0.00001928, 0.00001461, -0.00000849, 0.00001088, 0.00001799, -0.00001313, 0.00001302, -0.00001977, 0.00000643, -0.00000896, 0.00000285, -0.00000654, 0.00002034, -0.00000994, -0.00000968, 0.00000945, -0.00000556, 0.00000060, -0.00000760, 0.00000570, -0.00000404, 0.00000754, -0.00000153, 0.00000069, -0.00000716, 0.00000406, -0.00000408, 0.00000512, -0.00000159, 0.00000257, -0.00000311, },
		{ -0.00000300, 0.00000420, 0.00000118, 0.00001171, -0.00000836, -0.00001364, 0.00000760, 0.00001532, -0.00001535, 0.00000484, 0.00000804, -0.00001627, 0.00001339, -0.00002142, 0.00000278, 0.00000340, 0.00000163, -0.00002255, 0.00001663, -0.00001960, 0.00002784, -0.00003181, 0.00001693, -0.00000168, 0.00001065, -0.00001002, 0.00001021, -0.00001215, 0.00001065, -0.00001656, 0.00001331, -0.00000318, 0.00000772, -0.00000536, 0.00000218, -0.00001068, 0.00000572, 0.00000102, 0.00000002, -0.00000504, 0.00000929, -0.00001460, 0.00001099, -0.00000860, 0.00000472, -0.00001814, 0.00002246, -0.00003707, 0.00002913, -0.00002776, 0.00002221, -0.00004306, 0.00003577, -0.00008711, 0.00005521, -0.00027124, 0.00006931, 0.00076145, -0.00037219, -0.00000181, -0.00007210, 0.00002961, -0.00003690, 0.00001748, -0.00002794, -0.00000104, -0.00001737, 0.00002385, -0.00001946, 0.00001505, -0.00001313, -0.00000240, -0.00000516, 0.00001027, -0.00000878, 0.00001000, -0.00000935, -0.00000417, 0.00000008, 0.00000706, -0.00000637, -0.00000005, 0.00000004, -0.00000480, 0.00000065, 0.00000445, -0.00000124, -0.00000206, -0.00000591, 0.00000281, 0.00000331, -0.00000298, -0.00001114, 0.00001259, -0.00002790, 0.00001804, -0.00001840, 0.00003509, -0.00000258, -0.00000090, 0.00001668, 0.00000876, -0.00000904, 0.00000949, -0.00002337, 0.00001437, -0.00001719, 0.00001070, -0.00001048, 0.00002868, -0.00002936, 0.00001158, -0.00000643, 0.00000564, -0.00001062, 0.00000382, -0.00000347, 0.00000512, 0.00000001, 0.00000412, -0.00000686, 0.00000004, -0.00000110, 0.00000120, 0.00000076, 0.00000204, -0.00000137, -0.00000040, },
		{ -0.00000195, 0.00000457, 0.00000235, 0.00001162, -0.00001719, -0.00000579, 0.00001131, 0.00001074, -0.00002117, 0.00002066, -0.00001287, 0.00000217, -0.00000181, -0.00001436, 0.00000231, 0.00000911, -0.00001406, -0.00000684, 0.00000551, -0.00000703, 0.00001606, -0.00002949, 0.00002503, -0.00000729, 0.00001119, -0.00001086, 0.00001120, -0.00001369, 0.00001213, -0.00002014, 0.00002263, -0.00001273, 0.00001561, -0.00001427, 0.00000891, -0.00001776, 0.00001675, -0.00000902, 0.00000587, -0.00000754, 0.00001000, -0.00001708, 0.00001638, -0.00001447, 0.00000596, -0.00001482, 0.00001745, -0.00003688, 0.00003723, -0.00003911, 0.00003098, -0.00005194, 0.00004085, -0.00010285, 0.00006361, -0.00038670, 0.00050981, 0.00038453, -0.00037061, 0.00006234, -0.00009228, 0.00005171, -0.00005451, 0.00003291, -0.00004386, 0.00001012, -0.00001689, 0.00002498, -0.00002469, 0.00002252, -0.00002386, 0.00000575, -0.00000649, 0.00001092, -0.00001039, 0.00001272, -0.00001654, 0.00000249, -0.00000062, 0.00000691, -0.00000913, 0.00000367, -0.00000368, -0.00000344, 0.00000341, 0.00000153, 0.00000067, -0.00000710, 0.00000084, -0.00000177, 0.00000893, -0.00001623, 0.00000699, -0.00000739, -0.00001263, 0.00000785, -0.00000612, 0.00003472, -0.00001177, 0.00000717, 0.00002087, -0.00000463, 0.00000413, -0.00000817, -0.00000889, 0.00000467, -0.00000935, 0.00000391, 0.00000156, 0.00001671, -0.00002841, 0.00001793, -0.00001284, 0.00000949, -0.00001476, 0.00000918, -0.00000744, 0.00000974, -0.00000412, 0.00000631, -0.00001095, 0.00000491, -0.00000423, 0.00000429, -0.00000169, 0.00000376, -0.00000349, 0.00000032, },
		{ 0.00000118, 0.00000163, 0.00000830, 0.00000347, -0.00001814, -0.00000070, 0.00001746, -0.00000338, -0.00001093, 0.00001928, -0.00002105, 0.00001459, -0.00001754, -0.00000265, -0.00000027, 0.00001159, -0.00002694, 0.00001110, -0.00001022, 0.00001149, -0.00000689, -0.00001045, 0.00001782, -0.00000283, 0.00000293, -0.00000152, 0.00000092, -0.00000304, -0.00000021, -0.00000735, 0.00001539, -0.00000795, 0.00001010, -0.00001041, 0.00000312, -0.00001102, 0.00001448, -0.00000864, 0.00000325, -0.00000151, 0.00000010, -0.00000611, 0.00000714, -0.00000681, -0.00000551, 0.00000338, -0.00000847, -0.00000789, 0.00001162, -0.00001513, 0.00000125, -0.00001600, -0.00001303, -0.00003779, -0.00006678, -0.00025566, 0.00074531, -0.00002845, -0.00021026, 0.00001857, -0.00004558, 0.00001966, -0.00002899, 0.00001260, -0.00003170, 0.00000188, 0.00000042, 0.00000664, -0.00000973, 0.00001091, -0.00001891, 0.00000333, 0.00000082, 0.00000182, -0.00000128, 0.00000375, -0.00001310, 0.00000254, 0.00000328, 0.00000102, -0.00000573, 0.00000231, -0.00000460, -0.00000192, 0.00000480, -0.00000015, 0.00000001, -0.00000806, 0.00000404, -0.00000280, 0.00000793, -0.00002056, 0.00001798, -0.00002545, 0.00000565, -0.00000673, 0.00001459, 0.00002185, -0.00001238, 0.00001650, 0.00001660, -0.00000973, 0.00001054, -0.00002272, 0.00000828, -0.00001048, 0.00000461, -0.00000821, 0.00001994, -0.00000753, -0.00001040, 0.00000801, -0.00000643, 0.00000287, -0.00000912, 0.00000582, -0.00000350, 0.00000692, -0.00000198, 0.00000251, -0.00000904, 0.00000516, -0.00000370, 0.00000388, -0.00000107, 0.00000266, -0.00000342, -0.00000030, },
		{ 0.00000234, 0.00000075, 0.00001273, -0.00000711, -0.00001433, 0.00000443, 0.00001971, -0.00001795, 0.00000816, 0.00000201, -0.00001076, 0.00000995, -0.00002085, 0.00000284, 0.00000241, 0.00000347, -0.00002616, 0.00001820, -0.00001716, 0.00002109, -0.00002437, 0.00001045, 0.00000580, 0.00000308, -0.00000420, 0.00000664, -0.00000870, 0.00000741, -0.00001358, 0.00000987, 0.00000125, 0.00000316, -0.00000124, -0.00000128, -0.00000736, 0.00000205, 0.00000495, -0.00000311, -0.00000154, 0.00000440, -0.00000920, 0.00000558, -0.00000374, 0.00000080, -0.00001348, 0.00001553, -0.00002920, 0.00002074, -0.00001784, 0.00001287, -0.00003139, 0.00002095, -0.00006994, 0.00003329, -0.00024360, 0.00008143, 0.00067064, -0.00027752, -0.00005077, -0.00004397, 0.00000985, -0.00002211, 0.00000546, -0.00001693, -0.00001177, -0.00000786, 0.00001536, -0.00001152, 0.00000789, -0.00000565, -0.00000904, -0.00000027, 0.00000638, -0.00000571, 0.00000740, -0.00000725, -0.00000594, 0.00000111, 0.00000635, -0.00000483, -0.00000100, -0.00000100, -0.00000403, 0.00000028, 0.00000396, 0.00000031, -0.00000392, -0.00000345, 0.00000115, 0.00000205, -0.00000194, -0.00001128, 0.00001278, -0.00002920, 0.00001400, -0.00001276, 0.00003107, 0.00000500, -0.00000602, 0.00002473, 0.00000434, -0.00000252, 0.00000138, -0.00002046, 0.00001201, -0.00001520, 0.00000965, -0.00001052, 0.00002753, -0.00002560, 0.00000842, -0.00000624, 0.00000434, -0.00000744, 0.00000039, -0.00000108, 0.00000368, 0.00000060, 0.00000319, -0.00000447, -0.00000322, 0.00000229, -0.00000115, 0.00000182, 0.00000083, 0.00000042, -0.00000280, 0.00000026, },
		{ 0.00000010, 0.00000497, 0.00001037, -0.00001366, -0.00001042, 0.00001292, 0.00001136, -0.00002047, 0.00001994, -0.00001563, 0.00000680, -0.00000651, -0.00001175, 0.00000060, 0.00000958, -0.00001323, -0.00001110, 0.00001071, -0.00001035, 0.00001492, -0.00002592, 0.00002138, -0.00000197, 0.00000415, -0.00000386, 0.00000593, -0.00000871, 0.00000737, -0.00001580, 0.00001822, -0.00000738, 0.00000961, -0.00000817, 0.00000322, -0.00001226, 0.00001097, -0.00000280, 0.00000051, -0.00000239, 0.00000357, -0.00000955, 0.00000811, -0.00000618, -0.00000126, -0.00000767, 0.00000815, -0.00002577, 0.00002484, -0.00002408, 0.00001559, -0.00003356, 0.00001704, -0.00007255, 0.00002021, -0.00031248, 0.00044112, 0.00037109, -0.00031167, 0.00001964, -0.00006314, 0.00002966, -0.00003750, 0.00001942, -0.00003218, -0.00000102, -0.00000667, 0.00001553, -0.00001565, 0.00001432, -0.00001513, -0.00000289, 0.00000097, 0.00000431, -0.00000466, 0.00000757, -0.00001203, -0.00000151, 0.00000240, 0.00000474, -0.00000627, 0.00000088, -0.00000317, -0.00000309, 0.00000250, 0.00000236, 0.00000057, -0.00000732, 0.00000223, -0.00000351, 0.00000763, -0.00001377, 0.00000461, -0.00000463, -0.00001837, 0.00000964, -0.00000544, 0.00003572, -0.00000873, 0.00000623, 0.00002564, -0.00000637, 0.00000746, -0.00001571, -0.00000460, 0.00000116, -0.00000617, 0.00000172, 0.00000184, 0.00001646, -0.00002542, 0.00001415, -0.00001210, 0.00000900, -0.00001261, 0.00000599, -0.00000454, 0.00000755, -0.00000318, 0.00000595, -0.00000964, 0.00000248, -0.00000086, 0.00000110, 0.00000028, 0.00000208, -0.00000150, -0.00000233, 0.00000247, },
		{ -0.00000217, 0.00001128, 0.00000227, -0.00001501, -0.00000645, 0.00002188, -0.00000551, -0.00000755, 0.00001423, -0.00001841, 0.00001476, -0.00001980, -0.00000080, -0.00000205, 0.00001273, -0.00002743, 0.00000786, -0.00000365, 0.00000356, -0.00000148, -0.00001332, 0.00002047, -0.00000409, 0.00000181, 0.00000108, -0.00000090, -0.00000131, -0.00000190, -0.00000604, 0.00001436, -0.00000634, 0.00000771, -0.00000754, 0.00000040, -0.00000845, 0.00001171, -0.00000528, 0.00000077, -0.00000006, -0.00000181, -0.00000330, 0.00000335, -0.00000260, -0.00000850, 0.00000498, -0.00001080, -0.00000435, 0.00000745, -0.00000874, -0.00000469, -0.00000924, -0.00002303, -0.00002373, -0.00009168, -0.00018320, 0.00063128, 0.00003396, -0.00021139, 0.00001059, -0.00003874, 0.00001373, -0.00002451, 0.00000977, -0.00002981, -0.00000006, 0.00000221, 0.00000484, -0.00000758, 0.00000914, -0.00001622, -0.00000023, 0.00000419, -0.00000149, 0.00000176, 0.00000077, -0.00001025, -0.00000021, 0.00000538, -0.00000010, -0.00000390, -0.00000042, -0.00000348, -0.00000189, 0.00000364, 0.00000168, -0.00000167, -0.00000625, 0.00000350, -0.00000400, 0.00000700, -0.00001784, 0.00001466, -0.00002175, -0.00000266, -0.00000025, 0.00001060, 0.00002817, -0.00001480, 0.00002130, 0.00001589, -0.00000630, 0.00000696, -0.00002459, 0.00000916, -0.00001117, 0.00000563, -0.00000892, 0.00001858, -0.00000494, -0.00001069, 0.00000610, -0.00000682, 0.00000453, -0.00000980, 0.00000511, -0.00000232, 0.00000576, -0.00000188, 0.00000362, -0.00001003, 0.00000539, -0.00000268, 0.00000223, -0.00000021, 0.00000233, -0.00000317, -0.00000071, 0.00000358, },
		{ 0.00000012, 0.00001358, -0.00000590, -0.00001455, 0.00000136, 0.00002309, -0.00001930, 0.00001039, -0.00000349, -0.00000527, 0.00000623, -0.00001984, 0.00000273, 0.00000151, 0.00000481, -0.00002845, 0.00001819, -0.00001330, 0.00001334, -0.00001638, 0.00000388, 0.00001280, -0.00000377, 0.00000091, 0.00000359, -0.00000559, 0.00000441, -0.00001068, 0.00000674, 0.00000489, -0.00000040, 0.00000178, -0.00000371, -0.00000493, -0.00000051, 0.00000738, -0.00000469, 0.00000059, 0.00000074, -0.00000496, 0.00000150, -0.00000040, -0.00000183, -0.00000983, 0.00000947, -0.00002201, 0.00001328, -0.00000942, 0.00000532, -0.00002167, 0.00000834, -0.00005519, 0.00001399, -0.00021827, 0.00009684, 0.00057747, -0.00018710, -0.00009212, -0.00002175, -0.00000556, -0.00001087, -0.00000358, -0.00000857, -0.00001979, -0.00000054, 0.00000840, -0.00000502, 0.00000233, -0.00000003, -0.00001363, 0.00000293, 0.00000370, -0.00000366, 0.00000567, -0.00000604, -0.00000662, 0.00000120, 0.00000627, -0.00000390, -0.00000130, -0.00000255, -0.00000279, -0.00000043, 0.00000371, 0.00000146, -0.00000531, -0.00000132, -0.00000022, 0.00000059, -0.00000098, -0.00001083, 0.00001159, -0.00002831, 0.00000772, -0.00000622, 0.00002670, 0.00001259, -0.00001011, 0.00003184, 0.00000118, 0.00000287, -0.00000671, -0.00001712, 0.00000899, -0.00001244, 0.00000780, -0.00000936, 0.00002475, -0.00002057, 0.00000458, -0.00000549, 0.00000254, -0.00000393, -0.00000304, 0.00000116, 0.00000234, 0.00000120, 0.00000222, -0.00000217, -0.00000609, 0.00000510, -0.00000300, 0.00000258, -0.00000018, 0.00000198, -0.00000480, 0.00000295, 0.00000097, },
		{ 0.00000718, 0.00000912, -0.00001019, -0.00001431, 0.00001396, 0.00001163, -0.00001878, 0.00001787, -0.00001711, 0.00001042, -0.00001051, -0.00000947, -0.00000064, 0.00000931, -0.00001180, -0.00001508, 0.00001492, -0.00001239, 0.00001255, -0.00002127, 0.00001687, 0.00000366, -0.00000274, 0.00000276, 0.00000101, -0.00000397, 0.00000273, -0.00001125, 0.00001355, -0.00000214, 0.00000398, -0.00000267, -0.00000181, -0.00000722, 0.00000576, 0.00000240, -0.00000355, 0.00000143, -0.00000156, -0.00000327, 0.00000127, 0.00000032, -0.00000669, -0.00000215, 0.00000040, -0.00001590, 0.00001371, -0.00001080, 0.00000237, -0.00001771, -0.00000366, -0.00004602, -0.00001845, -0.00024341, 0.00037954, 0.00034902, -0.00024931, -0.00002129, -0.00003649, 0.00000974, -0.00002240, 0.00000759, -0.00002196, -0.00001035, 0.00000182, 0.00000727, -0.00000754, 0.00000704, -0.00000762, -0.00000985, 0.00000673, -0.00000092, -0.00000013, 0.00000341, -0.00000842, -0.00000435, 0.00000425, 0.00000347, -0.00000427, -0.00000099, -0.00000346, -0.00000198, 0.00000093, 0.00000368, -0.00000016, -0.00000671, 0.00000280, -0.00000436, 0.00000535, -0.00001033, 0.00000148, -0.00000182, -0.00002292, 0.00000924, -0.00000327, 0.00003556, -0.00000461, 0.00000531, 0.00003016, -0.00000717, 0.00000915, -0.00002219, -0.00000091, -0.00000206, -0.00000303, -0.00000047, 0.00000253, 0.00001501, -0.00002110, 0.00000938, -0.00001042, 0.00000752, -0.00000951, 0.00000211, -0.00000122, 0.00000497, -0.00000176, 0.00000498, -0.00000774, -0.00000027, 0.00000250, -0.00000195, 0.00000223, 0.00000032, 0.00000062, -0.00000493, 0.00000651, -0.00000386, },
		{ 0.00001380, 0.00000120, -0.00001191, -0.00001134, 0.00002501, -0.00000688, -0.00000431, 0.00000870, -0.00001491, 0.00001394, -0.00002112, 0.00000046, -0.00000315, 0.00001277, -0.00002653, 0.00000380, 0.00000295, -0.00000374, 0.00000300, -0.00001508, 0.00002181, -0.00000433, 0.00000008, 0.00000392, -0.00000282, 0.00000050, -0.00000370, -0.00000424, 0.00001251, -0.00000401, 0.00000476, -0.00000433, -0.00000254, -0.00000543, 0.00000838, -0.00000163, -0.00000168, 0.00000123, -0.00000352, -0.00000062, -0.00000020, 0.00000098, -0.00001069, 0.00000584, -0.00001237, -0.00000123, 0.00000343, -0.00000266, -0.00001006, -0.00000324, -0.00003183, -0.00001155, -0.00011333, -0.00011499, 0.00052434, 0.00008494, -0.00020318, -0.00000181, -0.00002957, 0.00000616, -0.00001881, 0.00000587, -0.00002682, -0.00000254, 0.00000415, 0.00000267, -0.00000482, 0.00000666, -0.00001297, -0.00000375, 0.00000717, -0.00000444, 0.00000442, -0.00000197, -0.00000750, -0.00000262, 0.00000695, -0.00000084, -0.00000244, -0.00000265, -0.00000279, -0.00000144, 0.00000210, 0.00000365, -0.00000357, -0.00000399, 0.00000242, -0.00000448, 0.00000511, -0.00001387, 0.00000983, -0.00001666, -0.00001130, 0.00000495, 0.00000787, 0.00003320, -0.00001544, 0.00002508, 0.00001613, -0.00000322, 0.00000247, -0.00002540, 0.00000904, -0.00001098, 0.00000590, -0.00000879, 0.00001651, -0.00000246, -0.00001041, 0.00000372, -0.00000667, 0.00000543, -0.00000958, 0.00000359, -0.00000057, 0.00000413, -0.00000124, 0.00000401, -0.00001011, 0.00000480, -0.00000116, 0.00000033, 0.00000090, 0.00000166, -0.00000248, -0.00000151, 0.00000602, -0.00000566, },
		{ 0.00001425, -0.00000474, -0.00001441, -0.00000131, 0.00002516, -0.00001925, 0.00001142, -0.00000823, -0.00000008, 0.00000242, -0.00001845, 0.00000244, 0.00000079, 0.00000550, -0.00002930, 0.00001672, -0.00000851, 0.00000536, -0.00000862, -0.00000211, 0.00001867, -0.00000935, 0.00000492, 0.00000135, -0.00000307, 0.00000186, -0.00000809, 0.00000415, 0.00000751, -0.00000280, 0.00000360, -0.00000507, -0.00000342, -0.00000192, 0.00000834, -0.00000481, 0.00000142, -0.00000173, -0.00000189, -0.00000126, 0.00000155, -0.00000336, -0.00000708, 0.00000433, -0.00001565, 0.00000696, -0.00000265, -0.00000036, -0.00001399, -0.00000193, -0.00004296, -0.00000266, -0.00019476, 0.00011337, 0.00048587, -0.00010464, -0.00012380, -0.00000637, -0.00001613, -0.00000349, -0.00000946, -0.00000301, -0.00002494, 0.00000439, 0.00000319, -0.00000018, -0.00000146, 0.00000360, -0.00001613, 0.00000450, 0.00000216, -0.00000255, 0.00000467, -0.00000553, -0.00000642, 0.00000060, 0.00000658, -0.00000339, -0.00000120, -0.00000429, -0.00000139, -0.00000124, 0.00000354, 0.00000229, -0.00000628, 0.00000056, -0.00000141, -0.00000087, -0.00000031, -0.00000969, 0.00000904, -0.00002543, -0.00000020, 0.00000036, 0.00002266, 0.00001964, -0.00001268, 0.00003767, -0.00000033, 0.00000656, -0.00001426, -0.00001370, 0.00000558, -0.00000911, 0.00000532, -0.00000723, 0.00002069, -0.00001477, 0.00000047, -0.00000448, 0.00000055, -0.00000048, -0.00000612, 0.00000294, 0.00000130, 0.00000163, 0.00000140, -0.00000023, -0.00000828, 0.00000709, -0.00000416, 0.00000290, -0.00000085, 0.00000313, -0.00000620, 0.00000487, 0.00000018, -0.00000071, },
		{ 0.00000813, -0.00000711, -0.00001723, 0.00001436, 0.00001153, -0.00001618, 0.00001459, -0.00001735, 0.00001297, -0.00001375, -0.00000754, -0.00000146, 0.00000849, -0.00001010, -0.00001832, 0.00001775, -0.00001300, 0.00000918, -0.00001596, 0.00001196, 0.00000903, -0.00000891, 0.00000848, -0.00000310, 0.00000012, -0.00000141, -0.00000690, 0.00000905, 0.00000255, -0.00000082, 0.00000181, -0.00000583, -0.00000303, 0.00000150, 0.00000627, -0.00000607, 0.00000370, -0.00000516, 0.00000147, -0.00000382, 0.00000481, -0.00001022, 0.00000164, -0.00000557, -0.00000767, 0.00000436, 0.00000016, -0.00000808, -0.00000508, -0.00002042, -0.00002432, -0.00005093, -0.00018117, 0.00032566, 0.00032009, -0.00018652, -0.00005804, -0.00001394, -0.00000689, -0.00001007, -0.00000196, -0.00001374, -0.00001739, 0.00000815, 0.00000061, -0.00000081, 0.00000116, -0.00000177, -0.00001475, 0.00001053, -0.00000454, 0.00000298, 0.00000042, -0.00000583, -0.00000601, 0.00000497, 0.00000304, -0.00000307, -0.00000205, -0.00000432, -0.00000041, -0.00000098, 0.00000521, -0.00000130, -0.00000550, 0.00000277, -0.00000455, 0.00000247, -0.00000645, -0.00000187, 0.00000052, -0.00002579, 0.00000653, 0.00000010, 0.00003459, 0.00000031, 0.00000469, 0.00003433, -0.00000694, 0.00000894, -0.00002722, 0.00000183, -0.00000474, -0.00000015, -0.00000257, 0.00000356, 0.00001254, -0.00001584, 0.00000402, -0.00000809, 0.00000532, -0.00000581, -0.00000203, 0.00000213, 0.00000232, -0.00000011, 0.00000365, -0.00000555, -0.00000298, 0.00000549, -0.00000453, 0.00000387, -0.00000126, 0.00000254, -0.00000712, 0.00000978, -0.00000690, 0.00000868, },
		{ 0.00000054, -0.00000922, -0.00001496, 0.00002662, -0.00000738, -0.00000138, 0.00000299, -0.00001083, 0.00001227, -0.00002154, 0.00000113, -0.00000364, 0.00001186, -0.00002453, -0.00000056, 0.00000884, -0.00000973, 0.00000614, -0.00001556, 0.00002180, -0.00000363, -0.00000204, 0.00000676, -0.00000462, 0.00000219, -0.00000541, -0.00000218, 0.00001011, -0.00000129, 0.00000162, -0.00000116, -0.00000535, -0.00000235, 0.00000490, 0.00000188, -0.00000375, 0.00000206, -0.00000474, 0.00000159, -0.00000312, 0.00000357, -0.00001186, 0.00000582, -0.00001304, 0.00000121, -0.00000008, 0.00000262, -0.00001433, 0.00000136, -0.00003861, -0.00000231, -0.00013004, -0.00005383, 0.00042745, 0.00012341, -0.00018646, -0.00001736, -0.00001915, -0.00000215, -0.00001261, 0.00000144, -0.00002315, -0.00000519, 0.00000595, 0.00000042, -0.00000179, 0.00000383, -0.00000958, -0.00000683, 0.00000940, -0.00000668, 0.00000637, -0.00000414, -0.00000515, -0.00000444, 0.00000781, -0.00000105, -0.00000148, -0.00000426, -0.00000257, -0.00000064, 0.00000036, 0.00000560, -0.00000549, -0.00000149, 0.00000098, -0.00000440, 0.00000252, -0.00000914, 0.00000409, -0.00001094, -0.00001931, 0.00000811, 0.00000674, 0.00003679, -0.00001420, 0.00002780, 0.00001748, -0.00000082, -0.00000267, -0.00002519, 0.00000796, -0.00000995, 0.00000548, -0.00000795, 0.00001401, -0.00000031, -0.00000955, 0.00000093, -0.00000598, 0.00000558, -0.00000850, 0.00000142, 0.00000155, 0.00000220, -0.00000019, 0.00000373, -0.00000940, 0.00000357, 0.00000064, -0.00000160, 0.00000207, 0.00000081, -0.00000150, -0.00000250, 0.00000821, -0.00000849, 0.00001553, },
		{ -0.00000357, -0.00001408, -0.00000330, 0.00002574, -0.00001779, 0.00001126, -0.00001206, 0.00000459, -0.00000131, -0.00001676, 0.00000195, 0.00000030, 0.00000546, -0.00002876, 0.00001406, -0.00000337, -0.00000207, -0.00000178, -0.00000695, 0.00002291, -0.00001327, 0.00000757, 0.00000006, -0.00000128, -0.00000010, -0.00000596, 0.00000225, 0.00000904, -0.00000402, 0.00000426, -0.00000542, -0.00000278, -0.00000229, 0.00000800, -0.00000370, 0.00000119, -0.00000318, 0.00000015, -0.00000288, 0.00000236, -0.00000407, -0.00000506, 0.00000010, -0.00001020, 0.00000185, 0.00000243, -0.00000420, -0.00000828, -0.00000996, -0.00003312, -0.00001685, -0.00017248, 0.00012897, 0.00039933, -0.00003304, -0.00014461, 0.00000191, -0.00002188, 0.00000016, -0.00001234, -0.00000014, -0.00002730, 0.00000699, -0.00000026, 0.00000298, -0.00000349, 0.00000531, -0.00001671, 0.00000465, 0.00000155, -0.00000218, 0.00000420, -0.00000549, -0.00000564, -0.00000039, 0.00000701, -0.00000305, -0.00000097, -0.00000590, -0.00000013, -0.00000196, 0.00000333, 0.00000287, -0.00000691, 0.00000221, -0.00000249, -0.00000218, -0.00000003, -0.00000792, 0.00000538, -0.00002101, -0.00000896, 0.00000605, 0.00001959, 0.00002573, -0.00001340, 0.00004203, 0.00000001, 0.00000813, -0.00002078, -0.00001057, 0.00000207, -0.00000549, 0.00000245, -0.00000441, 0.00001583, -0.00000875, -0.00000348, -0.00000349, -0.00000132, 0.00000257, -0.00000851, 0.00000402, 0.00000071, 0.00000178, 0.00000089, 0.00000117, -0.00000962, 0.00000810, -0.00000455, 0.00000275, -0.00000110, 0.00000378, -0.00000688, 0.00000593, -0.00000002, -0.00000167, 0.00001466, },
		{ -0.00000462, -0.00001903, 0.00001419, 0.00001099, -0.00001278, 0.00001034, -0.00001648, 0.00001443, -0.00001617, -0.00000595, -0.00000194, 0.00000728, -0.00000844, -0.00002049, 0.00001895, -0.00001223, 0.00000517, -0.00001048, 0.00000716, 0.00001362, -0.00001381, 0.00001286, -0.00000606, 0.00000324, -0.00000472, -0.00000311, 0.00000512, 0.00000630, -0.00000446, 0.00000502, -0.00000861, 0.00000009, -0.00000158, 0.00000863, -0.00000698, 0.00000442, -0.00000718, 0.00000457, -0.00000707, 0.00000728, -0.00001195, 0.00000377, -0.00000972, -0.00000127, -0.00000293, 0.00000846, -0.00001543, 0.00000396, -0.00003279, -0.00000802, -0.00007642, -0.00012678, 0.00027945, 0.00028642, -0.00012619, -0.00008854, 0.00000330, -0.00001946, -0.00000103, -0.00000886, -0.00000777, -0.00002195, 0.00001212, -0.00000418, 0.00000426, -0.00000306, 0.00000219, -0.00001744, 0.00001229, -0.00000650, 0.00000463, -0.00000140, -0.00000422, -0.00000658, 0.00000472, 0.00000327, -0.00000250, -0.00000251, -0.00000545, 0.00000126, -0.00000288, 0.00000663, -0.00000256, -0.00000395, 0.00000238, -0.00000437, -0.00000058, -0.00000262, -0.00000494, 0.00000198, -0.00002673, 0.00000168, 0.00000410, 0.00003328, 0.00000569, 0.00000466, 0.00003806, -0.00000574, 0.00000670, -0.00003053, 0.00000340, -0.00000667, 0.00000232, -0.00000445, 0.00000482, 0.00000936, -0.00001012, -0.00000148, -0.00000546, 0.00000277, -0.00000196, -0.00000599, 0.00000510, -0.00000006, 0.00000148, 0.00000223, -0.00000340, -0.00000532, 0.00000777, -0.00000639, 0.00000498, -0.00000242, 0.00000401, -0.00000863, 0.00001195, -0.00000877, 0.00000927, 0.00000758, },
		{ -0.00000723, -0.00001707, 0.00002663, -0.00000704, 0.00000114, -0.00000263, -0.00000643, 0.00000994, -0.00002118, 0.00000130, -0.00000362, 0.00001018, -0.00002179, -0.00000468, 0.00001341, -0.00001389, 0.00000774, -0.00001473, 0.00002051, -0.00000217, -0.00000429, 0.00000932, -0.00000605, 0.00000355, -0.00000682, -0.00000013, 0.00000751, 0.00000144, -0.00000131, 0.00000161, -0.00000769, 0.00000044, 0.00000169, 0.00000488, -0.00000514, 0.00000223, -0.00000528, 0.00000310, -0.00000514, 0.00000496, -0.00001196, 0.00000497, -0.00001278, 0.00000285, -0.00000280, 0.00000673, -0.00001712, 0.00000413, -0.00004285, 0.00000330, -0.00014067, -0.00000172, 0.00034255, 0.00014924, -0.00016272, -0.00003445, -0.00000869, -0.00001030, -0.00000662, -0.00000296, -0.00001926, -0.00000769, 0.00000732, -0.00000162, 0.00000117, 0.00000103, -0.00000644, -0.00000912, 0.00001063, -0.00000801, 0.00000741, -0.00000552, -0.00000338, -0.00000552, 0.00000788, -0.00000072, -0.00000109, -0.00000521, -0.00000274, 0.00000034, -0.00000138, 0.00000727, -0.00000721, 0.00000104, -0.00000060, -0.00000400, -0.00000037, -0.00000421, -0.00000184, -0.00000539, -0.00002586, 0.00000868, 0.00000725, 0.00003903, -0.00001115, 0.00002956, 0.00002001, 0.00000061, -0.00000822, -0.00002409, 0.00000598, -0.00000818, 0.00000448, -0.00000660, 0.00001134, 0.00000133, -0.00000813, -0.00000213, -0.00000484, 0.00000505, -0.00000674, -0.00000118, 0.00000378, 0.00000021, 0.00000111, 0.00000296, -0.00000810, 0.00000194, 0.00000247, -0.00000333, 0.00000313, -0.00000005, -0.00000045, -0.00000345, 0.00000986, -0.00001037, 0.00001623, 0.00000052, },
		{ -0.00001373, -0.00000440, 0.00002476, -0.00001504, 0.00001002, -0.00001491, 0.00000855, -0.00000482, -0.00001485, 0.00000129, 0.00000006, 0.00000471, -0.00002695, 0.00001055, 0.00000154, -0.00000827, 0.00000357, -0.00001021, 0.00002520, -0.00001530, 0.00000876, -0.00000024, -0.00000023, -0.00000141, -0.00000434, 0.00000106, 0.00000951, -0.00000416, 0.00000397, -0.00000498, -0.00000280, -0.00000185, 0.00000668, -0.00000173, 0.00000022, -0.00000387, 0.00000137, -0.00000358, 0.00000235, -0.00000427, -0.00000358, -0.00000327, -0.00000569, -0.00000205, 0.00000592, -0.00000639, -0.00000434, -0.00001594, -0.00002544, -0.00002884, -0.00015083, 0.00014189, 0.00032068, 0.00002574, -0.00015422, 0.00000351, -0.00002330, 0.00000058, -0.00001266, 0.00000044, -0.00002722, 0.00000751, -0.00000212, 0.00000462, -0.00000395, 0.00000536, -0.00001568, 0.00000371, 0.00000159, -0.00000231, 0.00000399, -0.00000561, -0.00000458, -0.00000148, 0.00000731, -0.00000268, -0.00000084, -0.00000711, 0.00000078, -0.00000242, 0.00000301, 0.00000323, -0.00000722, 0.00000367, -0.00000350, -0.00000326, -0.00000016, -0.00000570, 0.00000098, -0.00001567, -0.00001763, 0.00000997, 0.00001796, 0.00003062, -0.00001213, 0.00004497, 0.00000220, 0.00000732, -0.00002590, -0.00000804, -0.00000125, -0.00000183, -0.00000056, -0.00000128, 0.00001068, -0.00000307, -0.00000689, -0.00000278, -0.00000282, 0.00000494, -0.00001000, 0.00000429, 0.00000065, 0.00000161, 0.00000073, 0.00000194, -0.00001004, 0.00000814, -0.00000423, 0.00000217, -0.00000095, 0.00000391, -0.00000685, 0.00000616, 0.00000030, -0.00000262, 0.00001472, -0.00000229, },
		{ -0.00001974, 0.00001356, 0.00000997, -0.00000876, 0.00000541, -0.00001470, 0.00001489, -0.00001781, -0.00000469, -0.00000219, 0.00000587, -0.00000704, -0.00002137, 0.00001845, -0.00001027, 0.00000099, -0.00000533, 0.00000295, 0.00001697, -0.00001707, 0.00001559, -0.00000765, 0.00000519, -0.00000697, -0.00000013, 0.00000202, 0.00000887, -0.00000675, 0.00000684, -0.00001007, 0.00000203, -0.00000342, 0.00000952, -0.00000642, 0.00000375, -0.00000779, 0.00000615, -0.00000860, 0.00000796, -0.00001217, 0.00000449, -0.00001215, 0.00000329, -0.00000806, 0.00001402, -0.00001966, 0.00000941, -0.00004072, 0.00000280, -0.00009464, -0.00008063, 0.00024038, 0.00025027, -0.00007092, -0.00011127, 0.00001453, -0.00002762, 0.00000454, -0.00001305, -0.00000407, -0.00002404, 0.00001376, -0.00000706, 0.00000758, -0.00000553, 0.00000424, -0.00001801, 0.00001218, -0.00000696, 0.00000494, -0.00000218, -0.00000341, -0.00000629, 0.00000376, 0.00000389, -0.00000235, -0.00000261, -0.00000653, 0.00000270, -0.00000446, 0.00000769, -0.00000372, -0.00000231, 0.00000189, -0.00000411, -0.00000344, 0.00000068, -0.00000737, 0.00000232, -0.00002575, -0.00000485, 0.00000801, 0.00003217, 0.00001118, 0.00000543, 0.00004136, -0.00000375, 0.00000247, -0.00003198, 0.00000363, -0.00000771, 0.00000423, -0.00000599, 0.00000612, 0.00000583, -0.00000446, -0.00000661, -0.00000294, 0.00000024, 0.00000163, -0.00000934, 0.00000733, -0.00000186, 0.00000279, 0.00000098, -0.00000155, -0.00000701, 0.00000912, -0.00000735, 0.00000545, -0.00000302, 0.00000487, -0.00000927, 0.00001286, -0.00000935, 0.00000891, 0.00000752, -0.00000280, },
		{ -0.00001757, 0.00002512, -0.00000600, 0.00000325, -0.00000791, -0.00000203, 0.00000716, -0.00002017, 0.00000108, -0.00000324, 0.00000796, -0.00001867, -0.00000812, 0.00001623, -0.00001596, 0.00000777, -0.00001273, 0.00001815, -0.00000024, -0.00000635, 0.00001132, -0.00000692, 0.00000439, -0.00000774, 0.00000166, 0.00000502, 0.00000383, -0.00000368, 0.00000369, -0.00000932, 0.00000263, -0.00000095, 0.00000708, -0.00000567, 0.00000162, -0.00000508, 0.00000380, -0.00000612, 0.00000511, -0.00001106, 0.00000341, -0.00001171, 0.00000367, -0.00000460, 0.00000949, -0.00001824, 0.00000490, -0.00004433, 0.00000501, -0.00014468, 0.00004020, 0.00027061, 0.00016318, -0.00013391, -0.00005137, 0.00000062, -0.00001741, -0.00000150, -0.00000683, -0.00001553, -0.00000975, 0.00000805, -0.00000318, 0.00000375, -0.00000142, -0.00000385, -0.00001041, 0.00001076, -0.00000834, 0.00000743, -0.00000603, -0.00000228, -0.00000586, 0.00000722, 0.00000008, -0.00000119, -0.00000557, -0.00000316, 0.00000130, -0.00000287, 0.00000846, -0.00000853, 0.00000334, -0.00000207, -0.00000353, -0.00000318, 0.00000033, -0.00000728, -0.00000073, -0.00003030, 0.00000646, 0.00000913, 0.00004023, -0.00000655, 0.00003063, 0.00002364, 0.00000083, -0.00001387, -0.00002227, 0.00000326, -0.00000580, 0.00000306, -0.00000495, 0.00000873, 0.00000238, -0.00000631, -0.00000525, -0.00000344, 0.00000402, -0.00000454, -0.00000389, 0.00000582, -0.00000156, 0.00000241, 0.00000191, -0.00000645, 0.00000019, 0.00000404, -0.00000464, 0.00000390, -0.00000074, 0.00000049, -0.00000413, 0.00001073, -0.00001110, 0.00001580, 0.00000124, -0.00000610, },
		{ -0.00000451, 0.00002231, -0.00001124, 0.00000793, -0.00001678, 0.00001169, -0.00000795, -0.00001284, 0.00000051, 0.00000002, 0.00000333, -0.00002411, 0.00000658, 0.00000569, -0.00001270, 0.00000706, -0.00001168, 0.00002543, -0.00001545, 0.00000860, 0.00000033, 0.00000014, -0.00000212, -0.00000321, 0.00000051, 0.00000909, -0.00000347, 0.00000298, -0.00000401, -0.00000325, -0.00000089, 0.00000475, 0.00000069, -0.00000108, -0.00000410, 0.00000202, -0.00000366, 0.00000187, -0.00000428, -0.00000244, -0.00000585, -0.00000211, -0.00000478, 0.00000798, -0.00000718, -0.00000190, -0.00002016, -0.00001961, -0.00003893, -0.00012939, 0.00015079, 0.00025191, 0.00007077, -0.00015322, -0.00000057, -0.00002131, -0.00000146, -0.00001110, -0.00000069, -0.00002519, 0.00000636, -0.00000270, 0.00000503, -0.00000318, 0.00000414, -0.00001348, 0.00000208, 0.00000198, -0.00000267, 0.00000378, -0.00000564, -0.00000352, -0.00000241, 0.00000729, -0.00000213, -0.00000097, -0.00000777, 0.00000123, -0.00000259, 0.00000256, 0.00000337, -0.00000722, 0.00000490, -0.00000439, -0.00000412, -0.00000063, -0.00000331, -0.00000366, -0.00001013, -0.00002530, 0.00001139, 0.00001798, 0.00003431, -0.00000895, 0.00004670, 0.00000599, 0.00000410, -0.00002938, -0.00000633, -0.00000417, 0.00000160, -0.00000341, 0.00000181, 0.00000574, 0.00000182, -0.00000945, -0.00000253, -0.00000379, 0.00000647, -0.00001049, 0.00000372, 0.00000107, 0.00000118, 0.00000091, 0.00000209, -0.00000960, 0.00000731, -0.00000333, 0.00000129, -0.00000048, 0.00000360, -0.00000621, 0.00000567, 0.00000096, -0.00000338, 0.00001430, -0.00000085, -0.00001350, },
		{ 0.00001264, 0.00000845, -0.00000432, 0.00000015, -0.00001227, 0.00001445, -0.00001871, -0.00000369, -0.00000229, 0.00000439, -0.00000608, -0.00002089, 0.00001636, -0.00000742, -0.00000293, -0.00000095, -0.00000034, 0.00001878, -0.00001843, 0.00001656, -0.00000781, 0.00000592, -0.00000810, 0.00000189, -0.00000009, 0.00001018, -0.00000769, 0.00000732, -0.00001030, 0.00000287, -0.00000409, 0.00000912, -0.00000466, 0.00000204, -0.00000727, 0.00000645, -0.00000867, 0.00000721, -0.00001131, 0.00000417, -0.00001312, 0.00000612, -0.00001111, 0.00001697, -0.00002099, 0.00001152, -0.00004451, 0.00000845, -0.00010583, -0.00004261, 0.00020757, 0.00021383, -0.00002278, -0.00012535, 0.00001963, -0.00003145, 0.00000684, -0.00001474, -0.00000242, -0.00002393, 0.00001329, -0.00000817, 0.00000925, -0.00000640, 0.00000457, -0.00001675, 0.00001054, -0.00000620, 0.00000418, -0.00000215, -0.00000317, -0.00000543, 0.00000239, 0.00000464, -0.00000237, -0.00000261, -0.00000728, 0.00000363, -0.00000548, 0.00000818, -0.00000456, -0.00000081, 0.00000150, -0.00000396, -0.00000579, 0.00000315, -0.00000891, 0.00000152, -0.00002311, -0.00001234, 0.00001096, 0.00003177, 0.00001651, 0.00000720, 0.00004433, -0.00000129, -0.00000353, -0.00003160, 0.00000253, -0.00000783, 0.00000549, -0.00000710, 0.00000727, 0.00000233, 0.00000065, -0.00001093, -0.00000086, -0.00000193, 0.00000459, -0.00001175, 0.00000853, -0.00000289, 0.00000362, 0.00000008, -0.00000022, -0.00000790, 0.00000943, -0.00000737, 0.00000525, -0.00000301, 0.00000505, -0.00000902, 0.00001251, -0.00000869, 0.00000771, 0.00000788, -0.00000161, -0.00001947, },
		{ 0.00002236, -0.00000448, 0.00000496, -0.00001264, 0.00000210, 0.00000415, -0.00001867, 0.00000060, -0.00000263, 0.00000542, -0.00001547, -0.00001056, 0.00001706, -0.00001590, 0.00000642, -0.00000985, 0.00001501, 0.00000185, -0.00000792, 0.00001254, -0.00000707, 0.00000460, -0.00000803, 0.00000300, 0.00000291, 0.00000560, -0.00000526, 0.00000488, -0.00001007, 0.00000405, -0.00000280, 0.00000833, -0.00000528, 0.00000029, -0.00000419, 0.00000371, -0.00000610, 0.00000415, -0.00000941, 0.00000143, -0.00001004, 0.00000377, -0.00000549, 0.00001088, -0.00001773, 0.00000373, -0.00004315, 0.00000299, -0.00014215, 0.00007160, 0.00021164, 0.00016668, -0.00010226, -0.00006642, 0.00000777, -0.00002278, 0.00000225, -0.00000979, -0.00001227, -0.00001116, 0.00000801, -0.00000411, 0.00000573, -0.00000328, -0.00000201, -0.00001063, 0.00000980, -0.00000772, 0.00000649, -0.00000569, -0.00000181, -0.00000553, 0.00000596, 0.00000121, -0.00000165, -0.00000551, -0.00000365, 0.00000201, -0.00000391, 0.00000897, -0.00000926, 0.00000519, -0.00000321, -0.00000325, -0.00000556, 0.00000398, -0.00001167, 0.00000252, -0.00003227, 0.00000159, 0.00001178, 0.00004092, -0.00000080, 0.00003140, 0.00002817, -0.00000032, -0.00001934, -0.00001995, -0.00000001, -0.00000302, 0.00000138, -0.00000320, 0.00000635, 0.00000284, -0.00000427, -0.00000816, -0.00000199, 0.00000272, -0.00000219, -0.00000640, 0.00000738, -0.00000286, 0.00000351, 0.00000082, -0.00000475, -0.00000140, 0.00000512, -0.00000536, 0.00000426, -0.00000112, 0.00000116, -0.00000440, 0.00001071, -0.00001061, 0.00001424, 0.00000266, -0.00000582, -0.00001658, },
		{ 0.00001861, -0.00000678, 0.00000529, -0.00001778, 0.00001393, -0.00001062, -0.00001080, -0.00000034, 0.00000013, 0.00000145, -0.00002049, 0.00000254, 0.00000869, -0.00001506, 0.00000859, -0.00001139, 0.00002368, -0.00001389, 0.00000732, 0.00000154, 0.00000001, -0.00000234, -0.00000249, 0.00000046, 0.00000801, -0.00000224, 0.00000164, -0.00000282, -0.00000385, 0.00000027, 0.00000262, 0.00000313, -0.00000236, -0.00000416, 0.00000235, -0.00000340, 0.00000122, -0.00000436, -0.00000151, -0.00000768, 0.00000057, -0.00000644, 0.00000879, -0.00000685, -0.00000065, -0.00002293, -0.00001532, -0.00004734, -0.00010793, 0.00015491, 0.00019410, 0.00010210, -0.00014293, -0.00000884, -0.00001707, -0.00000501, -0.00000843, -0.00000286, -0.00002185, 0.00000405, -0.00000238, 0.00000457, -0.00000162, 0.00000212, -0.00001061, 0.00000019, 0.00000242, -0.00000302, 0.00000338, -0.00000537, -0.00000267, -0.00000300, 0.00000683, -0.00000132, -0.00000144, -0.00000781, 0.00000119, -0.00000248, 0.00000205, 0.00000327, -0.00000689, 0.00000586, -0.00000513, -0.00000481, -0.00000128, -0.00000108, -0.00000799, -0.00000513, -0.00003115, 0.00000984, 0.00001955, 0.00003707, -0.00000415, 0.00004770, 0.00001093, -0.00000139, -0.00003114, -0.00000557, -0.00000651, 0.00000457, -0.00000585, 0.00000450, 0.00000143, 0.00000554, -0.00001098, -0.00000280, -0.00000414, 0.00000710, -0.00001001, 0.00000242, 0.00000185, 0.00000059, 0.00000133, 0.00000174, -0.00000849, 0.00000584, -0.00000208, 0.00000028, 0.00000019, 0.00000299, -0.00000513, 0.00000467, 0.00000177, -0.00000380, 0.00001332, 0.00000081, -0.00001347, -0.00000362, },
		{ 0.00000643, 0.00000028, -0.00000504, -0.00000951, 0.00001330, -0.00001896, -0.00000292, -0.00000232, 0.00000295, -0.00000562, -0.00001911, 0.00001289, -0.00000406, -0.00000618, 0.00000237, -0.00000247, 0.00001892, -0.00001790, 0.00001584, -0.00000667, 0.00000553, -0.00000814, 0.00000295, -0.00000119, 0.00001028, -0.00000740, 0.00000668, -0.00000952, 0.00000278, -0.00000382, 0.00000774, -0.00000211, -0.00000030, -0.00000603, 0.00000584, -0.00000769, 0.00000549, -0.00000984, 0.00000322, -0.00001294, 0.00000749, -0.00001233, 0.00001764, -0.00001985, 0.00001083, -0.00004473, 0.00000960, -0.00011063, -0.00001221, 0.00017998, 0.00017900, 0.00001679, -0.00013063, 0.00001904, -0.00003148, 0.00000637, -0.00001441, -0.00000237, -0.00002207, 0.00001114, -0.00000783, 0.00000953, -0.00000596, 0.00000356, -0.00001411, 0.00000782, -0.00000462, 0.00000269, -0.00000162, -0.00000320, -0.00000429, 0.00000093, 0.00000524, -0.00000234, -0.00000271, -0.00000749, 0.00000388, -0.00000581, 0.00000800, -0.00000499, 0.00000042, 0.00000133, -0.00000405, -0.00000748, 0.00000461, -0.00000953, -0.00000026, -0.00001929, -0.00001992, 0.00001210, 0.00003247, 0.00002154, 0.00001001, 0.00004717, 0.00000117, -0.00001089, -0.00002960, 0.00000019, -0.00000706, 0.00000609, -0.00000769, 0.00000809, -0.00000078, 0.00000475, -0.00001408, 0.00000051, -0.00000346, 0.00000664, -0.00001299, 0.00000858, -0.00000305, 0.00000391, -0.00000038, 0.00000052, -0.00000794, 0.00000877, -0.00000655, 0.00000445, -0.00000244, 0.00000460, -0.00000795, 0.00001105, -0.00000700, 0.00000590, 0.00000848, -0.00000093, -0.00001844, 0.00001163, },
		{ -0.00000280, 0.00000634, -0.00001663, 0.00000571, 0.00000112, -0.00001684, -0.00000004, -0.00000195, 0.00000275, -0.00001244, -0.00001183, 0.00001590, -0.00001388, 0.00000402, -0.00000642, 0.00001145, 0.00000379, -0.00000876, 0.00001285, -0.00000644, 0.00000415, -0.00000765, 0.00000374, 0.00000139, 0.00000658, -0.00000593, 0.00000515, -0.00000992, 0.00000462, -0.00000376, 0.00000859, -0.00000405, -0.00000159, -0.00000278, 0.00000298, -0.00000524, 0.00000234, -0.00000736, -0.00000064, -0.00000805, 0.00000336, -0.00000560, 0.00001103, -0.00001580, 0.00000094, -0.00003972, -0.00000220, -0.00013372, 0.00009291, 0.00016487, 0.00016171, -0.00007007, -0.00007817, 0.00001200, -0.00002597, 0.00000435, -0.00001163, -0.00000964, -0.00001186, 0.00000718, -0.00000435, 0.00000700, -0.00000444, -0.00000095, -0.00000986, 0.00000794, -0.00000634, 0.00000475, -0.00000466, -0.00000183, -0.00000471, 0.00000432, 0.00000244, -0.00000228, -0.00000519, -0.00000399, 0.00000230, -0.00000435, 0.00000869, -0.00000928, 0.00000644, -0.00000383, -0.00000333, -0.00000723, 0.00000638, -0.00001460, 0.00000403, -0.00003175, -0.00000540, 0.00001435, 0.00004174, 0.00000562, 0.00003236, 0.00003334, -0.00000296, -0.00002434, -0.00001738, -0.00000358, -0.00000005, -0.00000035, -0.00000153, 0.00000432, 0.00000278, -0.00000224, -0.00001061, -0.00000072, 0.00000139, 0.00000000, -0.00000840, 0.00000820, -0.00000350, 0.00000420, -0.00000009, -0.00000324, -0.00000262, 0.00000558, -0.00000541, 0.00000415, -0.00000112, 0.00000145, -0.00000419, 0.00000979, -0.00000898, 0.00001172, 0.00000463, -0.00000623, -0.00001436, 0.00001873, },
		{ -0.00000210, 0.00000248, -0.00001806, 0.00001527, -0.00001275, -0.00000883, -0.00000119, 0.00000031, -0.00000078, -0.00001637, -0.00000128, 0.00001030, -0.00001529, 0.00000826, -0.00000955, 0.00002026, -0.00001098, 0.00000527, 0.00000311, -0.00000041, -0.00000224, -0.00000204, 0.00000073, 0.00000656, -0.00000082, 0.00000027, -0.00000169, -0.00000437, 0.00000135, 0.00000064, 0.00000523, -0.00000330, -0.00000429, 0.00000256, -0.00000301, 0.00000064, -0.00000468, -0.00000071, -0.00000880, 0.00000238, -0.00000712, 0.00000857, -0.00000569, -0.00000030, -0.00002452, -0.00001231, -0.00005420, -0.00008648, 0.00015403, 0.00014747, 0.00012070, -0.00012529, -0.00001956, -0.00001186, -0.00000907, -0.00000547, -0.00000537, -0.00001782, 0.00000112, -0.00000158, 0.00000365, 0.00000030, -0.00000021, -0.00000753, -0.00000160, 0.00000266, -0.00000320, 0.00000266, -0.00000470, -0.00000214, -0.00000317, 0.00000591, -0.00000028, -0.00000221, -0.00000728, 0.00000074, -0.00000218, 0.00000153, 0.00000289, -0.00000623, 0.00000648, -0.00000560, -0.00000541, -0.00000196, 0.00000066, -0.00001151, -0.00000128, -0.00003459, 0.00000522, 0.00002221, 0.00003941, 0.00000183, 0.00004857, 0.00001636, -0.00000876, -0.00003126, -0.00000578, -0.00000816, 0.00000694, -0.00000769, 0.00000654, -0.00000194, 0.00000790, -0.00001144, -0.00000357, -0.00000393, 0.00000689, -0.00000870, 0.00000058, 0.00000280, 0.00000001, 0.00000185, 0.00000107, -0.00000695, 0.00000401, -0.00000072, -0.00000065, 0.00000090, 0.00000224, -0.00000383, 0.00000339, 0.00000250, -0.00000373, 0.00001177, 0.00000267, -0.00001363, -0.00000188, 0.00001403, },
		{ 0.00000474, -0.00000977, -0.00000674, 0.00001162, -0.00001867, -0.00000230, -0.00000233, 0.00000159, -0.00000565, -0.00001621, 0.00000838, -0.00000055, -0.00000846, 0.00000446, -0.00000342, 0.00001744, -0.00001563, 0.00001370, -0.00000451, 0.00000425, -0.00000727, 0.00000314, -0.00000141, 0.00000936, -0.00000617, 0.00000524, -0.00000803, 0.00000203, -0.00000289, 0.00000576, 0.00000076, -0.00000276, -0.00000451, 0.00000473, -0.00000608, 0.00000331, -0.00000825, 0.00000204, -0.00001196, 0.00000770, -0.00001206, 0.00001647, -0.00001683, 0.00000800, -0.00004216, 0.00000715, -0.00010995, 0.00001137, 0.00015654, 0.00014729, 0.00004707, -0.00012762, 0.00001370, -0.00002856, 0.00000387, -0.00001269, -0.00000335, -0.00001903, 0.00000783, -0.00000645, 0.00000879, -0.00000461, 0.00000167, -0.00001065, 0.00000456, -0.00000266, 0.00000082, -0.00000088, -0.00000324, -0.00000317, -0.00000036, 0.00000549, -0.00000210, -0.00000305, -0.00000706, 0.00000340, -0.00000544, 0.00000717, -0.00000498, 0.00000132, 0.00000143, -0.00000441, -0.00000845, 0.00000502, -0.00000933, -0.00000267, -0.00001496, -0.00002667, 0.00001071, 0.00003443, 0.00002631, 0.00001383, 0.00005018, 0.00000305, -0.00001905, -0.00002634, -0.00000310, -0.00000558, 0.00000609, -0.00000776, 0.00000845, -0.00000323, 0.00000755, -0.00001586, 0.00000101, -0.00000421, 0.00000765, -0.00001299, 0.00000749, -0.00000238, 0.00000366, -0.00000040, 0.00000067, -0.00000724, 0.00000729, -0.00000509, 0.00000325, -0.00000144, 0.00000366, -0.00000628, 0.00000873, -0.00000458, 0.00000377, 0.00000908, -0.00000054, -0.00001688, 0.00001071, 0.00000398, },
		{ 0.00000748, -0.00001976, 0.00000858, -0.00000173, -0.00001484, -0.00000071, -0.00000130, 0.00000011, -0.00000971, -0.00001192, 0.00001295, -0.00001029, 0.00000099, -0.00000283, 0.00000777, 0.00000532, -0.00000871, 0.00001220, -0.00000510, 0.00000311, -0.00000664, 0.00000385, 0.00000053, 0.00000671, -0.00000569, 0.00000457, -0.00000899, 0.00000440, -0.00000388, 0.00000800, -0.00000221, -0.00000377, -0.00000115, 0.00000188, -0.00000382, 0.00000006, -0.00000530, -0.00000244, -0.00000606, 0.00000268, -0.00000515, 0.00001021, -0.00001282, -0.00000300, -0.00003466, -0.00000973, -0.00012046, 0.00010518, 0.00012895, 0.00015048, -0.00003944, -0.00008556, 0.00001294, -0.00002684, 0.00000473, -0.00001232, -0.00000769, -0.00001187, 0.00000565, -0.00000393, 0.00000754, -0.00000491, -0.00000061, -0.00000831, 0.00000547, -0.00000448, 0.00000249, -0.00000318, -0.00000212, -0.00000363, 0.00000255, 0.00000355, -0.00000289, -0.00000482, -0.00000403, 0.00000206, -0.00000411, 0.00000760, -0.00000857, 0.00000701, -0.00000386, -0.00000385, -0.00000806, 0.00000733, -0.00001590, 0.00000378, -0.00002902, -0.00001370, 0.00001583, 0.00004336, 0.00001228, 0.00003399, 0.00003881, -0.00000716, -0.00002856, -0.00001486, -0.00000718, 0.00000284, -0.00000193, -0.00000010, 0.00000271, 0.00000229, -0.00000045, -0.00001239, 0.00000019, 0.00000026, 0.00000180, -0.00000967, 0.00000812, -0.00000334, 0.00000436, -0.00000066, -0.00000210, -0.00000332, 0.00000535, -0.00000482, 0.00000361, -0.00000074, 0.00000136, -0.00000353, 0.00000807, -0.00000643, 0.00000854, 0.00000686, -0.00000705, -0.00001132, 0.00001468, -0.00000126, },
		{ -0.00000012, -0.00001779, 0.00001574, -0.00001429, -0.00000697, -0.00000198, 0.00000048, -0.00000320, -0.00001200, -0.00000464, 0.00001048, -0.00001357, 0.00000644, -0.00000656, 0.00001560, -0.00000717, 0.00000288, 0.00000470, -0.00000090, -0.00000197, -0.00000173, 0.00000111, 0.00000502, 0.00000050, -0.00000085, -0.00000083, -0.00000462, 0.00000212, -0.00000089, 0.00000669, -0.00000369, -0.00000463, 0.00000277, -0.00000265, 0.00000029, -0.00000529, -0.00000006, -0.00000920, 0.00000336, -0.00000694, 0.00000752, -0.00000396, -0.00000060, -0.00002518, -0.00001039, -0.00005953, -0.00006533, 0.00014847, 0.00011144, 0.00012827, -0.00010263, -0.00003089, -0.00000693, -0.00001274, -0.00000293, -0.00000762, -0.00001370, -0.00000194, -0.00000069, 0.00000264, 0.00000217, -0.00000243, -0.00000464, -0.00000303, 0.00000258, -0.00000313, 0.00000157, -0.00000360, -0.00000196, -0.00000294, 0.00000461, 0.00000090, -0.00000318, -0.00000630, 0.00000004, -0.00000183, 0.00000110, 0.00000219, -0.00000521, 0.00000671, -0.00000574, -0.00000603, -0.00000250, 0.00000164, -0.00001383, 0.00000097, -0.00003532, -0.00000220, 0.00002520, 0.00004200, 0.00000843, 0.00005002, 0.00002150, -0.00001750, -0.00002995, -0.00000687, -0.00000911, 0.00000859, -0.00000878, 0.00000775, -0.00000420, 0.00000882, -0.00001091, -0.00000471, -0.00000327, 0.00000602, -0.00000680, -0.00000153, 0.00000370, -0.00000039, 0.00000229, 0.00000030, -0.00000527, 0.00000212, 0.00000049, -0.00000133, 0.00000150, 0.00000153, -0.00000251, 0.00000205, 0.00000299, -0.00000310, 0.00000969, 0.00000464, -0.00001388, 0.00000045, 0.00000902, 0.00000229, },
		{ -0.00001365, -0.00000426, 0.00000962, -0.00001793, -0.00000180, -0.00000236, 0.00000033, -0.00000609, -0.00001245, 0.00000322, 0.00000276, -0.00000961, 0.00000536, -0.00000330, 0.00001459, -0.00001197, 0.00001054, -0.00000171, 0.00000241, -0.00000577, 0.00000267, -0.00000094, 0.00000774, -0.00000437, 0.00000338, -0.00000620, 0.00000093, -0.00000163, 0.00000360, 0.00000349, -0.00000492, -0.00000312, 0.00000350, -0.00000428, 0.00000114, -0.00000695, 0.00000098, -0.00001051, 0.00000709, -0.00001070, 0.00001397, -0.00001256, 0.00000377, -0.00003767, 0.00000210, -0.00010488, 0.00002902, 0.00013628, 0.00011970, 0.00006809, -0.00011750, 0.00000493, -0.00002377, 0.00000021, -0.00001028, -0.00000476, -0.00001544, 0.00000396, -0.00000452, 0.00000745, -0.00000282, -0.00000056, -0.00000694, 0.00000127, -0.00000073, -0.00000108, -0.00000023, -0.00000305, -0.00000228, -0.00000129, 0.00000526, -0.00000158, -0.00000368, -0.00000601, 0.00000227, -0.00000450, 0.00000579, -0.00000461, 0.00000192, 0.00000175, -0.00000495, -0.00000879, 0.00000453, -0.00000855, -0.00000527, -0.00001080, -0.00003173, 0.00000630, 0.00003756, 0.00003103, 0.00001850, 0.00005369, 0.00000367, -0.00002732, -0.00002232, -0.00000700, -0.00000363, 0.00000561, -0.00000735, 0.00000828, -0.00000485, 0.00000892, -0.00001621, 0.00000065, -0.00000418, 0.00000763, -0.00001185, 0.00000541, -0.00000104, 0.00000301, -0.00000006, 0.00000035, -0.00000601, 0.00000527, -0.00000327, 0.00000189, -0.00000022, 0.00000245, -0.00000427, 0.00000593, -0.00000185, 0.00000171, 0.00000943, -0.00000023, -0.00001505, 0.00001000, 0.00000122, 0.00000874, },
		{ -0.00002192, 0.00001054, -0.00000423, -0.00001279, -0.00000133, -0.00000076, -0.00000242, -0.00000734, -0.00001098, 0.00000860, -0.00000564, -0.00000221, 0.00000058, 0.00000425, 0.00000625, -0.00000774, 0.00001070, -0.00000320, 0.00000165, -0.00000515, 0.00000341, 0.00000032, 0.00000607, -0.00000470, 0.00000336, -0.00000748, 0.00000357, -0.00000332, 0.00000675, -0.00000003, -0.00000590, 0.00000040, 0.00000070, -0.00000218, -0.00000230, -0.00000363, -0.00000364, -0.00000435, 0.00000200, -0.00000440, 0.00000876, -0.00000926, -0.00000750, -0.00002872, -0.00001853, -0.00010375, 0.00010989, 0.00010211, 0.00013527, -0.00001214, -0.00008798, 0.00001063, -0.00002554, 0.00000358, -0.00001198, -0.00000633, -0.00001133, 0.00000361, -0.00000298, 0.00000744, -0.00000481, -0.00000078, -0.00000628, 0.00000275, -0.00000246, 0.00000001, -0.00000153, -0.00000246, -0.00000252, 0.00000089, 0.00000433, -0.00000329, -0.00000454, -0.00000365, 0.00000125, -0.00000326, 0.00000581, -0.00000722, 0.00000693, -0.00000333, -0.00000477, -0.00000809, 0.00000686, -0.00001567, 0.00000202, -0.00002469, -0.00002225, 0.00001517, 0.00004631, 0.00001883, 0.00003673, 0.00004421, -0.00001292, -0.00003172, -0.00001268, -0.00001050, 0.00000539, -0.00000318, 0.00000096, 0.00000155, 0.00000150, 0.00000092, -0.00001334, 0.00000061, -0.00000053, 0.00000303, -0.00001010, 0.00000712, -0.00000239, 0.00000398, -0.00000082, -0.00000142, -0.00000348, 0.00000450, -0.00000369, 0.00000276, -0.00000006, 0.00000094, -0.00000253, 0.00000579, -0.00000331, 0.00000509, 0.00000897, -0.00000792, -0.00000792, 0.00001071, -0.00000017, 0.00000844, },
		{ -0.00001718, 0.00001540, -0.00001524, -0.00000528, -0.00000268, 0.00000058, -0.00000567, -0.00000760, -0.00000744, 0.00000937, -0.00001030, 0.00000361, -0.00000292, 0.00001022, -0.00000298, 0.00000057, 0.00000600, -0.00000126, -0.00000168, -0.00000145, 0.00000145, 0.00000364, 0.00000148, -0.00000153, -0.00000038, -0.00000451, 0.00000245, -0.00000182, 0.00000738, -0.00000343, -0.00000522, 0.00000300, -0.00000235, 0.00000019, -0.00000616, 0.00000039, -0.00000888, 0.00000356, -0.00000602, 0.00000586, -0.00000189, -0.00000133, -0.00002507, -0.00000945, -0.00006325, -0.00004497, 0.00013900, 0.00008483, 0.00012697, -0.00007739, -0.00004111, -0.00000331, -0.00001529, -0.00000134, -0.00000914, -0.00000997, -0.00000471, -0.00000000, 0.00000183, 0.00000364, -0.00000421, -0.00000222, -0.00000392, 0.00000214, -0.00000283, 0.00000019, -0.00000217, -0.00000204, -0.00000240, 0.00000306, 0.00000206, -0.00000418, -0.00000508, -0.00000072, -0.00000158, 0.00000085, 0.00000117, -0.00000385, 0.00000653, -0.00000550, -0.00000671, -0.00000282, 0.00000170, -0.00001473, 0.00000141, -0.00003340, -0.00001175, 0.00002751, 0.00004560, 0.00001510, 0.00005276, 0.00002548, -0.00002696, -0.00002753, -0.00000870, -0.00000941, 0.00000949, -0.00000907, 0.00000807, -0.00000530, 0.00000842, -0.00000958, -0.00000603, -0.00000233, 0.00000470, -0.00000458, -0.00000363, 0.00000432, -0.00000044, 0.00000249, -0.00000035, -0.00000372, 0.00000044, 0.00000134, -0.00000164, 0.00000190, 0.00000098, -0.00000136, 0.00000084, 0.00000311, -0.00000193, 0.00000721, 0.00000656, -0.00001403, 0.00000302, 0.00000410, 0.00000591, -0.00000215, },
		{ -0.00000235, 0.00000747, -0.00001686, -0.00000138, -0.00000242, -0.00000084, -0.00000680, -0.00000816, -0.00000220, 0.00000561, -0.00000963, 0.00000526, -0.00000241, 0.00001075, -0.00000742, 0.00000687, 0.00000126, 0.00000042, -0.00000396, 0.00000179, -0.00000006, 0.00000575, -0.00000238, 0.00000151, -0.00000437, -0.00000023, -0.00000037, 0.00000163, 0.00000567, -0.00000637, -0.00000220, 0.00000247, -0.00000263, -0.00000067, -0.00000624, 0.00000026, -0.00000886, 0.00000598, -0.00000864, 0.00001068, -0.00000771, -0.00000108, -0.00003213, -0.00000450, -0.00009653, 0.00004168, 0.00011845, 0.00009668, 0.00008051, -0.00010187, -0.00000571, -0.00001827, -0.00000370, -0.00000790, -0.00000601, -0.00001187, 0.00000008, -0.00000249, 0.00000595, -0.00000104, -0.00000263, -0.00000348, -0.00000158, 0.00000084, -0.00000274, 0.00000015, -0.00000252, -0.00000173, -0.00000178, 0.00000454, -0.00000079, -0.00000451, -0.00000449, 0.00000072, -0.00000319, 0.00000407, -0.00000401, 0.00000230, 0.00000216, -0.00000553, -0.00000867, 0.00000336, -0.00000753, -0.00000758, -0.00000748, -0.00003442, -0.00000131, 0.00004144, 0.00003610, 0.00002382, 0.00005806, 0.00000236, -0.00003494, -0.00001813, -0.00001104, -0.00000153, 0.00000484, -0.00000655, 0.00000761, -0.00000560, 0.00000887, -0.00001526, -0.00000043, -0.00000348, 0.00000673, -0.00000980, 0.00000262, 0.00000072, 0.00000216, 0.00000046, -0.00000023, -0.00000452, 0.00000306, -0.00000142, 0.00000062, 0.00000100, 0.00000121, -0.00000226, 0.00000304, 0.00000079, 0.00000003, 0.00000930, 0.00000019, -0.00001313, 0.00000955, -0.00000164, 0.00001196, -0.00001636, },
		{ 0.00001146, -0.00000626, -0.00001080, -0.00000185, -0.00000038, -0.00000476, -0.00000532, -0.00000927, 0.00000336, -0.00000051, -0.00000511, 0.00000353, 0.00000107, 0.00000653, -0.00000593, 0.00000855, -0.00000098, 0.00000001, -0.00000340, 0.00000255, 0.00000063, 0.00000485, -0.00000322, 0.00000181, -0.00000569, 0.00000235, -0.00000231, 0.00000514, 0.00000213, -0.00000764, 0.00000153, -0.00000025, -0.00000066, -0.00000435, -0.00000267, -0.00000401, -0.00000315, 0.00000154, -0.00000362, 0.00000705, -0.00000562, -0.00001192, -0.00002271, -0.00002750, -0.00008512, 0.00010873, 0.00008248, 0.00011815, 0.00001056, -0.00008538, 0.00000547, -0.00002250, 0.00000127, -0.00001090, -0.00000537, -0.00001045, 0.00000131, -0.00000171, 0.00000686, -0.00000431, -0.00000122, -0.00000412, 0.00000016, -0.00000064, -0.00000236, 0.00000001, -0.00000261, -0.00000159, -0.00000047, 0.00000463, -0.00000337, -0.00000444, -0.00000286, -0.00000002, -0.00000196, 0.00000351, -0.00000541, 0.00000634, -0.00000237, -0.00000594, -0.00000748, 0.00000519, -0.00001417, -0.00000079, -0.00001954, -0.00002991, 0.00001144, 0.00005093, 0.00002509, 0.00004095, 0.00004917, -0.00002023, -0.00003356, -0.00001116, -0.00001322, 0.00000732, -0.00000391, 0.00000154, 0.00000085, 0.00000053, 0.00000176, -0.00001344, 0.00000051, -0.00000091, 0.00000362, -0.00000968, 0.00000528, -0.00000078, 0.00000314, -0.00000058, -0.00000120, -0.00000318, 0.00000320, -0.00000225, 0.00000178, 0.00000078, 0.00000034, -0.00000138, 0.00000326, -0.00000002, 0.00000183, 0.00001057, -0.00000846, -0.00000464, 0.00000723, 0.00000026, 0.00000926, -0.00002303, },
		{ 0.00001433, -0.00001562, -0.00000379, -0.00000326, 0.00000056, -0.00000804, -0.00000334, -0.00000965, 0.00000727, -0.00000604, 0.00000039, 0.00000081, 0.00000465, 0.00000110, -0.00000128, 0.00000679, -0.00000133, -0.00000145, -0.00000112, 0.00000163, 0.00000257, 0.00000197, -0.00000169, -0.00000034, -0.00000406, 0.00000232, -0.00000209, 0.00000727, -0.00000256, -0.00000600, 0.00000321, -0.00000209, 0.00000030, -0.00000716, 0.00000052, -0.00000785, 0.00000306, -0.00000450, 0.00000375, 0.00000034, -0.00000236, -0.00002431, -0.00000942, -0.00006519, -0.00002606, 0.00012671, 0.00006599, 0.00011924, -0.00005193, -0.00004882, -0.00000173, -0.00001628, -0.00000100, -0.00000965, -0.00000694, -0.00000690, 0.00000027, 0.00000143, 0.00000449, -0.00000532, -0.00000044, -0.00000425, 0.00000139, -0.00000238, -0.00000135, -0.00000057, -0.00000226, -0.00000172, 0.00000146, 0.00000303, -0.00000502, -0.00000383, -0.00000134, -0.00000153, 0.00000080, -0.00000013, -0.00000221, 0.00000595, -0.00000487, -0.00000747, -0.00000288, 0.00000081, -0.00001420, 0.00000011, -0.00002922, -0.00002248, 0.00002796, 0.00005092, 0.00002141, 0.00005739, 0.00002750, -0.00003646, -0.00002438, -0.00001105, -0.00000914, 0.00000966, -0.00000859, 0.00000752, -0.00000535, 0.00000690, -0.00000772, -0.00000731, -0.00000132, 0.00000318, -0.00000234, -0.00000546, 0.00000450, -0.00000004, 0.00000235, -0.00000073, -0.00000253, -0.00000082, 0.00000172, -0.00000152, 0.00000205, 0.00000067, -0.00000052, -0.00000011, 0.00000284, -0.00000032, 0.00000454, 0.00000820, -0.00001385, 0.00000543, -0.00000023, 0.00000835, -0.00000332, -0.00001638, },
		{ 0.00000532, -0.00001556, -0.00000101, -0.00000250, -0.00000198, -0.00000762, -0.00000367, -0.00000750, 0.00000784, -0.00000867, 0.00000448, -0.00000113, 0.00000639, -0.00000251, 0.00000324, 0.00000394, -0.00000135, -0.00000216, 0.00000079, 0.00000094, 0.00000375, -0.00000058, -0.00000004, -0.00000284, -0.00000121, 0.00000065, 0.00000016, 0.00000703, -0.00000688, -0.00000195, 0.00000185, -0.00000137, -0.00000187, -0.00000626, 0.00000000, -0.00000719, 0.00000465, -0.00000627, 0.00000709, -0.00000286, -0.00000590, -0.00002632, -0.00001169, -0.00008600, 0.00005021, 0.00010252, 0.00007821, 0.00008555, -0.00008266, -0.00001664, -0.00001317, -0.00000707, -0.00000612, -0.00000664, -0.00000883, -0.00000333, -0.00000074, 0.00000464, 0.00000034, -0.00000415, -0.00000070, -0.00000367, 0.00000181, -0.00000397, 0.00000015, -0.00000159, -0.00000154, -0.00000185, 0.00000340, 0.00000014, -0.00000542, -0.00000272, -0.00000099, -0.00000181, 0.00000226, -0.00000336, 0.00000260, 0.00000253, -0.00000596, -0.00000832, 0.00000179, -0.00000660, -0.00000917, -0.00000554, -0.00003434, -0.00001191, 0.00004533, 0.00004207, 0.00002958, 0.00006356, -0.00000150, -0.00004117, -0.00001438, -0.00001474, 0.00000036, 0.00000401, -0.00000554, 0.00000653, -0.00000557, 0.00000761, -0.00001331, -0.00000199, -0.00000234, 0.00000522, -0.00000717, -0.00000049, 0.00000257, 0.00000133, 0.00000099, -0.00000086, -0.00000309, 0.00000101, 0.00000014, -0.00000029, 0.00000200, 0.00000017, -0.00000055, 0.00000043, 0.00000296, -0.00000096, 0.00000851, 0.00000085, -0.00001127, 0.00000932, -0.00000436, 0.00001411, -0.00001763, -0.00000120, },
		{ -0.00000775, -0.00000893, -0.00000224, -0.00000018, -0.00000692, -0.00000355, -0.00000711, -0.00000221, 0.00000453, -0.00000736, 0.00000585, -0.00000167, 0.00000618, -0.00000348, 0.00000603, 0.00000124, -0.00000152, -0.00000165, 0.00000149, 0.00000126, 0.00000330, -0.00000155, 0.00000026, -0.00000394, 0.00000103, -0.00000112, 0.00000349, 0.00000396, -0.00000869, 0.00000198, -0.00000072, 0.00000045, -0.00000578, -0.00000265, -0.00000343, -0.00000258, 0.00000146, -0.00000302, 0.00000540, -0.00000234, -0.00001569, -0.00001737, -0.00003560, -0.00006606, 0.00010347, 0.00006816, 0.00010088, 0.00002793, -0.00007820, -0.00000179, -0.00001837, -0.00000168, -0.00000942, -0.00000460, -0.00000948, -0.00000099, -0.00000034, 0.00000599, -0.00000365, -0.00000164, -0.00000218, -0.00000196, 0.00000070, -0.00000436, 0.00000122, -0.00000243, -0.00000099, -0.00000140, 0.00000439, -0.00000310, -0.00000454, -0.00000174, -0.00000157, -0.00000047, 0.00000099, -0.00000339, 0.00000544, -0.00000120, -0.00000709, -0.00000655, 0.00000269, -0.00001188, -0.00000403, -0.00001448, -0.00003559, 0.00000396, 0.00005721, 0.00003116, 0.00004689, 0.00005335, -0.00002899, -0.00003387, -0.00001059, -0.00001507, 0.00000840, -0.00000397, 0.00000158, 0.00000058, -0.00000055, 0.00000204, -0.00001276, -0.00000002, -0.00000093, 0.00000363, -0.00000855, 0.00000282, 0.00000127, 0.00000200, -0.00000006, -0.00000134, -0.00000259, 0.00000170, -0.00000076, 0.00000090, 0.00000160, -0.00000029, -0.00000031, 0.00000082, 0.00000298, -0.00000081, 0.00001130, -0.00000831, -0.00000190, 0.00000458, -0.00000009, 0.00000994, -0.00002304, 0.00001130, },
		{ -0.00001546, -0.00000252, -0.00000368, 0.00000036, -0.00001022, 0.00000066, -0.00001134, 0.00000459, -0.00000141, -0.00000261, 0.00000414, -0.00000061, 0.00000465, -0.00000238, 0.00000690, -0.00000105, -0.00000134, -0.00000074, 0.00000161, 0.00000190, 0.00000194, -0.00000135, -0.00000065, -0.00000337, 0.00000181, -0.00000176, 0.00000646, -0.00000123, -0.00000681, 0.00000328, -0.00000180, 0.00000049, -0.00000810, 0.00000023, -0.00000613, 0.00000196, -0.00000252, 0.00000138, 0.00000258, -0.00000356, -0.00002299, -0.00001028, -0.00006520, -0.00000926, 0.00011286, 0.00005308, 0.00010751, -0.00002828, -0.00005303, -0.00000255, -0.00001558, -0.00000197, -0.00000910, -0.00000476, -0.00000839, 0.00000006, 0.00000150, 0.00000463, -0.00000568, 0.00000068, -0.00000408, 0.00000046, -0.00000193, -0.00000288, 0.00000101, -0.00000242, -0.00000107, -0.00000001, 0.00000363, -0.00000553, -0.00000275, -0.00000168, -0.00000175, 0.00000093, -0.00000160, -0.00000043, 0.00000506, -0.00000392, -0.00000824, -0.00000278, -0.00000088, -0.00001243, -0.00000261, -0.00002344, -0.00003319, 0.00002543, 0.00005846, 0.00002714, 0.00006434, 0.00002683, -0.00004534, -0.00002091, -0.00001369, -0.00000842, 0.00000916, -0.00000741, 0.00000621, -0.00000453, 0.00000458, -0.00000563, -0.00000835, -0.00000043, 0.00000168, -0.00000032, -0.00000682, 0.00000414, 0.00000084, 0.00000185, -0.00000075, -0.00000184, -0.00000157, 0.00000160, -0.00000101, 0.00000199, 0.00000063, -0.00000003, -0.00000076, 0.00000222, 0.00000155, 0.00000196, 0.00000932, -0.00001307, 0.00000727, -0.00000351, 0.00000930, -0.00000348, -0.00001552, 0.00001262, },
		{ -0.00001408, -0.00000069, -0.00000258, -0.00000313, -0.00000838, 0.00000067, -0.00001235, 0.00000938, -0.00000699, 0.00000345, 0.00000011, 0.00000201, 0.00000219, 0.00000015, 0.00000591, -0.00000257, -0.00000067, -0.00000011, 0.00000182, 0.00000205, 0.00000073, -0.00000100, -0.00000181, -0.00000183, 0.00000125, -0.00000065, 0.00000740, -0.00000636, -0.00000244, 0.00000172, -0.00000062, -0.00000238, -0.00000699, 0.00000019, -0.00000562, 0.00000332, -0.00000390, 0.00000363, 0.00000147, -0.00001010, -0.00002092, -0.00001864, -0.00007426, 0.00005538, 0.00008820, 0.00006385, 0.00008477, -0.00006188, -0.00002637, -0.00000939, -0.00000928, -0.00000533, -0.00000636, -0.00000663, -0.00000593, 0.00000043, 0.00000379, 0.00000104, -0.00000483, 0.00000114, -0.00000482, 0.00000208, -0.00000472, -0.00000023, -0.00000035, -0.00000162, -0.00000161, 0.00000200, 0.00000106, -0.00000619, -0.00000099, -0.00000252, -0.00000066, 0.00000060, -0.00000284, 0.00000295, 0.00000268, -0.00000603, -0.00000798, 0.00000010, -0.00000605, -0.00000972, -0.00000528, -0.00003142, -0.00002488, 0.00004825, 0.00004950, 0.00003564, 0.00007033, -0.00000837, -0.00004534, -0.00001165, -0.00001763, 0.00000172, 0.00000335, -0.00000448, 0.00000522, -0.00000498, 0.00000547, -0.00001073, -0.00000369, -0.00000104, 0.00000341, -0.00000435, -0.00000353, 0.00000418, 0.00000076, 0.00000131, -0.00000130, -0.00000201, -0.00000058, 0.00000116, -0.00000067, 0.00000263, -0.00000049, 0.00000062, -0.00000158, 0.00000438, -0.00000109, 0.00000702, 0.00000178, -0.00000949, 0.00000919, -0.00000666, 0.00001495, -0.00001733, -0.00000092, 0.00000329, },
		{ -0.00000723, -0.00000249, -0.00000013, -0.00000891, -0.00000191, -0.00000487, -0.00000752, 0.00000893, -0.00000870, 0.00000748, -0.00000399, 0.00000534, -0.00000066, 0.00000349, 0.00000315, -0.00000265, -0.00000017, 0.00000047, 0.00000198, 0.00000173, -0.00000004, -0.00000097, -0.00000252, -0.00000015, 0.00000000, 0.00000206, 0.00000519, -0.00000882, 0.00000159, -0.00000056, 0.00000098, -0.00000640, -0.00000362, -0.00000194, -0.00000264, 0.00000185, -0.00000279, 0.00000409, 0.00000020, -0.00001836, -0.00001331, -0.00004192, -0.00004793, 0.00009576, 0.00005750, 0.00008478, 0.00003982, -0.00006730, -0.00001016, -0.00001388, -0.00000474, -0.00000792, -0.00000379, -0.00000862, -0.00000302, 0.00000092, 0.00000504, -0.00000304, -0.00000179, -0.00000075, -0.00000336, 0.00000136, -0.00000577, 0.00000194, -0.00000182, -0.00000076, -0.00000188, 0.00000363, -0.00000251, -0.00000476, -0.00000046, -0.00000312, 0.00000090, -0.00000140, -0.00000147, 0.00000450, -0.00000009, -0.00000794, -0.00000565, -0.00000018, -0.00000935, -0.00000703, -0.00001036, -0.00003842, -0.00000755, 0.00006476, 0.00003737, 0.00005464, 0.00005641, -0.00003906, -0.00003250, -0.00001120, -0.00001579, 0.00000843, -0.00000326, 0.00000103, 0.00000073, -0.00000169, 0.00000184, -0.00001148, -0.00000082, -0.00000071, 0.00000322, -0.00000696, 0.00000006, 0.00000344, 0.00000080, 0.00000057, -0.00000166, -0.00000192, 0.00000028, 0.00000050, 0.00000034, 0.00000223, -0.00000076, 0.00000043, -0.00000120, 0.00000532, -0.00000246, 0.00001088, -0.00000722, 0.00000001, 0.00000294, -0.00000119, 0.00001040, -0.00002193, 0.00001004, -0.00000867, },
		{ -0.00000147, -0.00000393, -0.00000003, -0.00001211, 0.00000431, -0.00001265, 0.00000177, 0.00000293, -0.00000484, 0.00000663, -0.00000519, 0.00000735, -0.00000257, 0.00000632, -0.00000044, -0.00000130, -0.00000034, 0.00000141, 0.00000161, 0.00000146, -0.00000064, -0.00000114, -0.00000261, 0.00000106, -0.00000099, 0.00000517, 0.00000035, -0.00000748, 0.00000306, -0.00000134, 0.00000063, -0.00000881, -0.00000055, -0.00000383, 0.00000042, -0.00000025, -0.00000109, 0.00000469, -0.00000486, -0.00002116, -0.00001197, -0.00006316, 0.00000478, 0.00009871, 0.00004425, 0.00009400, -0.00000803, -0.00005329, -0.00000570, -0.00001334, -0.00000404, -0.00000761, -0.00000344, -0.00000914, -0.00000060, 0.00000204, 0.00000407, -0.00000533, 0.00000121, -0.00000356, -0.00000046, -0.00000163, -0.00000419, 0.00000236, -0.00000238, -0.00000060, -0.00000119, 0.00000374, -0.00000559, -0.00000199, -0.00000164, -0.00000223, 0.00000115, -0.00000309, 0.00000135, 0.00000400, -0.00000278, -0.00000889, -0.00000268, -0.00000310, -0.00000981, -0.00000621, -0.00001696, -0.00004263, 0.00001894, 0.00006842, 0.00003244, 0.00007373, 0.00002296, -0.00005302, -0.00001749, -0.00001638, -0.00000738, 0.00000810, -0.00000566, 0.00000434, -0.00000310, 0.00000179, -0.00000360, -0.00000899, 0.00000022, 0.00000038, 0.00000130, -0.00000760, 0.00000325, 0.00000211, 0.00000102, -0.00000039, -0.00000168, -0.00000179, 0.00000105, -0.00000021, 0.00000180, 0.00000080, 0.00000011, -0.00000112, 0.00000136, 0.00000344, -0.00000025, 0.00000965, -0.00001145, 0.00000821, -0.00000544, 0.00000864, -0.00000258, -0.00001463, 0.00001034, -0.00001478, },
		{ -0.00000042, -0.00000263, -0.00000436, -0.00000889, 0.00000456, -0.00001649, 0.00001025, -0.00000494, 0.00000260, 0.00000091, -0.00000193, 0.00000621, -0.00000198, 0.00000687, -0.00000300, 0.00000032, -0.00000071, 0.00000238, 0.00000086, 0.00000136, -0.00000124, -0.00000136, -0.00000207, 0.00000139, -0.00000074, 0.00000681, -0.00000487, -0.00000359, 0.00000202, -0.00000035, -0.00000225, -0.00000828, 0.00000069, -0.00000420, 0.00000216, -0.00000179, 0.00000062, 0.00000493, -0.00001329, -0.00001642, -0.00002470, -0.00006211, 0.00005781, 0.00007540, 0.00005286, 0.00007992, -0.00004143, -0.00003370, -0.00000754, -0.00001001, -0.00000569, -0.00000507, -0.00000542, -0.00000755, 0.00000091, 0.00000350, 0.00000094, -0.00000455, 0.00000195, -0.00000501, 0.00000168, -0.00000502, -0.00000091, 0.00000107, -0.00000183, -0.00000122, 0.00000053, 0.00000178, -0.00000662, 0.00000041, -0.00000357, 0.00000001, -0.00000070, -0.00000257, 0.00000345, 0.00000249, -0.00000561, -0.00000785, -0.00000146, -0.00000607, -0.00000910, -0.00000678, -0.00002595, -0.00003928, 0.00004907, 0.00005893, 0.00004206, 0.00007827, -0.00001850, -0.00004698, -0.00001041, -0.00001934, 0.00000224, 0.00000306, -0.00000358, 0.00000387, -0.00000409, 0.00000288, -0.00000797, -0.00000519, 0.00000011, 0.00000166, -0.00000175, -0.00000612, 0.00000527, 0.00000064, 0.00000129, -0.00000137, -0.00000149, -0.00000149, 0.00000147, -0.00000043, 0.00000281, -0.00000064, 0.00000108, -0.00000279, 0.00000488, -0.00000032, 0.00000487, 0.00000294, -0.00000776, 0.00000898, -0.00000828, 0.00001435, -0.00001542, -0.00000153, 0.00000142, -0.00001247, },
		{ -0.00000261, -0.00000022, -0.00001075, -0.00000029, -0.00000292, -0.00001206, 0.00001230, -0.00000900, 0.00000847, -0.00000597, 0.00000419, 0.00000222, 0.00000128, 0.00000443, -0.00000313, 0.00000080, -0.00000030, 0.00000255, 0.00000041, 0.00000104, -0.00000160, -0.00000167, -0.00000099, 0.00000084, 0.00000107, 0.00000564, -0.00000794, 0.00000031, 0.00000026, 0.00000084, -0.00000617, -0.00000548, 0.00000027, -0.00000322, 0.00000269, -0.00000298, 0.00000330, 0.00000176, -0.00001961, -0.00001093, -0.00004583, -0.00003182, 0.00008699, 0.00004914, 0.00007067, 0.00004661, -0.00005387, -0.00001857, -0.00000980, -0.00000735, -0.00000673, -0.00000277, -0.00000806, -0.00000460, 0.00000190, 0.00000416, -0.00000264, -0.00000151, 0.00000000, -0.00000389, 0.00000125, -0.00000652, 0.00000211, -0.00000080, -0.00000088, -0.00000198, 0.00000246, -0.00000173, -0.00000498, 0.00000077, -0.00000437, 0.00000180, -0.00000334, 0.00000006, 0.00000378, 0.00000069, -0.00000818, -0.00000516, -0.00000291, -0.00000713, -0.00000918, -0.00000790, -0.00003783, -0.00002292, 0.00007273, 0.00004434, 0.00006418, 0.00005801, -0.00005022, -0.00002941, -0.00001314, -0.00001521, 0.00000728, -0.00000172, -0.00000011, 0.00000127, -0.00000290, 0.00000132, -0.00000989, -0.00000165, -0.00000045, 0.00000260, -0.00000519, -0.00000262, 0.00000540, -0.00000020, 0.00000112, -0.00000197, -0.00000144, -0.00000078, 0.00000126, 0.00000029, 0.00000252, -0.00000093, 0.00000068, -0.00000253, 0.00000667, -0.00000288, 0.00000918, -0.00000508, 0.00000094, 0.00000233, -0.00000289, 0.00001055, -0.00001973, 0.00000775, -0.00000902, -0.00000571, },
		},
		{ { 0.00000265, 0.00000174, 0.00000158, -0.00000333, 0.00001341, -0.00000290, -0.00001532, 0.00000231, 0.00000380, 0.00000508, -0.00001619, 0.00000383, -0.00000455, -0.00000296, 0.00000067, 0.00000132, -0.00000666, -0.00000045, -0.00000047, -0.00000281, 0.00000022, -0.00000001, 0.00001334, -0.00000530, -0.00000438, -0.00000160, 0.00000879, -0.00000112, 0.00000147, -0.00000001, 0.00000596, 0.00000024, -0.00000219, -0.00000378, 0.00000253, -0.00000309, -0.00001255, -0.00000136, 0.00000389, 0.00000002, -0.00000322, -0.00000885, -0.00001010, -0.00001224, -0.00005756, 0.00001361, 0.00008261, 0.00003733, 0.00008200, 0.00001223, -0.00006238, -0.00001183, -0.00000396, -0.00000723, -0.00000382, -0.00000631, -0.00000485, -0.00000206, 0.00000229, 0.00000070, -0.00000053, 0.00000157, -0.00001202, -0.00000006, -0.00000082, -0.00000006, -0.00001202, 0.00000157, -0.00000053, 0.00000070, 0.00000229, -0.00000206, -0.00000485, -0.00000631, -0.00000382, -0.00000723, -0.00000396, -0.00001183, -0.00006238, 0.00001223, 0.00008200, 0.00003733, 0.00008261, 0.00001361, -0.00005756, -0.00001224, -0.00001010, -0.00000885, -0.00000322, 0.00000002, 0.00000389, -0.00000136, -0.00001255, -0.00000309, 0.00000253, -0.00000378, -0.00000219, 0.00000024, 0.00000596, -0.00000001, 0.00000147, -0.00000112, 0.00000879, -0.00000160, -0.00000438, -0.00000530, 0.00001334, -0.00000001, 0.00000022, -0.00000281, -0.00000047, -0.00000045, -0.00000666, 0.00000132, 0.00000067, -0.00000296, -0.00000455, 0.00000383, -0.00001619, 0.00000508, 0.00000380, 0.00000231, -0.00001532, -0.00000290, 0.00001341, -0.00000333, 0.00000158, 0.00000174, },
		{ 0.00000240, 0.00000205, -0.00000078, 0.00000039, 0.00001396, -0.00001122, -0.00001011, 0.00000259, 0.00000710, -0.00000213, -0.00001363, 0.00000600, -0.00000759, 0.00000014, -0.00000018, 0.00000105, -0.00000814, 0.00000250, -0.00000310, -0.00000008, -0.00000241, 0.00000495, 0.00000981, -0.00000667, -0.00000437, 0.00000203, 0.00000732, -0.00000216, 0.00000191, 0.00000070, 0.00000667, -0.00000236, -0.00000162, -0.00000404, 0.00000447, -0.00000812, -0.00001004, 0.00000090, 0.00000376, -0.00000153, -0.00000338, -0.00001164, -0.00000614, -0.00002550, -0.00005135, 0.00004777, 0.00006693, 0.00004044, 0.00007921, -0.00001756, -0.00005319, -0.00000397, -0.00000591, -0.00000666, -0.00000345, -0.00000707, -0.00000344, -0.00000174, 0.00000376, -0.00000178, 0.00000251, -0.00000321, -0.00001017, 0.00000112, -0.00000046, -0.00000297, -0.00001089, 0.00000422, -0.00000289, 0.00000329, 0.00000009, -0.00000201, -0.00000631, -0.00000521, -0.00000474, -0.00000698, -0.00000292, -0.00002353, -0.00006373, 0.00004535, 0.00007373, 0.00004603, 0.00008549, -0.00002453, -0.00004868, -0.00000597, -0.00001263, -0.00000616, -0.00000306, 0.00000148, 0.00000372, -0.00000491, -0.00001240, 0.00000079, 0.00000009, -0.00000296, -0.00000269, 0.00000325, 0.00000370, 0.00000072, -0.00000006, 0.00000189, 0.00000747, -0.00000367, -0.00000491, -0.00000206, 0.00001433, -0.00000383, 0.00000262, -0.00000556, 0.00000238, -0.00000402, -0.00000399, 0.00000079, 0.00000149, -0.00000643, 0.00000023, -0.00000211, -0.00001348, 0.00000887, 0.00000104, 0.00000092, -0.00001857, 0.00000683, 0.00000836, -0.00000371, 0.00000287, 0.00000180, },
		{ 0.00000189, 0.00000206, -0.00000284, 0.00000608, 0.00000958, -0.00001573, -0.00000531, 0.00000360, 0.00000828, -0.00000957, -0.00000759, 0.00000436, -0.00000791, 0.00000183, -0.00000044, -0.00000038, -0.00000786, 0.00000403, -0.00000462, 0.00000163, -0.00000398, 0.00000950, 0.00000500, -0.00000681, -0.00000405, 0.00000596, 0.00000382, -0.00000115, 0.00000100, 0.00000318, 0.00000488, -0.00000322, -0.00000212, -0.00000260, 0.00000416, -0.00001216, -0.00000623, 0.00000214, 0.00000310, -0.00000251, -0.00000465, -0.00001244, -0.00000470, -0.00004062, -0.00002983, 0.00006861, 0.00004839, 0.00005092, 0.00006495, -0.00003947, -0.00003985, 0.00000001, -0.00000802, -0.00000523, -0.00000402, -0.00000697, -0.00000251, -0.00000067, 0.00000375, -0.00000280, 0.00000438, -0.00000827, -0.00000637, 0.00000079, 0.00000021, -0.00000681, -0.00000706, 0.00000392, -0.00000319, 0.00000458, -0.00000180, -0.00000224, -0.00000719, -0.00000446, -0.00000546, -0.00000645, -0.00000332, -0.00003840, -0.00005382, 0.00007515, 0.00005925, 0.00006569, 0.00006840, -0.00005426, -0.00003149, -0.00000680, -0.00001202, -0.00000481, -0.00000225, 0.00000252, 0.00000296, -0.00000901, -0.00000952, 0.00000232, -0.00000144, -0.00000266, -0.00000203, 0.00000509, 0.00000148, 0.00000167, -0.00000144, 0.00000557, 0.00000394, -0.00000393, -0.00000608, 0.00000303, 0.00001213, -0.00000525, 0.00000351, -0.00000713, 0.00000435, -0.00000711, -0.00000106, 0.00000010, 0.00000139, -0.00000863, 0.00000454, -0.00000970, -0.00000610, 0.00000782, 0.00000084, -0.00000331, -0.00001716, 0.00001422, 0.00000146, -0.00000136, 0.00000283, 0.00000225, },
		{ 0.00000161, 0.00000108, -0.00000287, 0.00001104, 0.00000210, -0.00001621, -0.00000176, 0.00000580, 0.00000589, -0.00001425, -0.00000096, 0.00000044, -0.00000587, 0.00000191, -0.00000000, -0.00000282, -0.00000576, 0.00000376, -0.00000458, 0.00000173, -0.00000358, 0.00001225, 0.00000030, -0.00000655, -0.00000245, 0.00000840, 0.00000013, 0.00000064, -0.00000019, 0.00000634, 0.00000117, -0.00000228, -0.00000377, 0.00000059, 0.00000083, -0.00001350, -0.00000268, 0.00000290, 0.00000180, -0.00000277, -0.00000725, -0.00001027, -0.00000926, -0.00005000, 0.00000076, 0.00007319, 0.00003508, 0.00006223, 0.00004225, -0.00005145, -0.00002569, 0.00000057, -0.00000945, -0.00000346, -0.00000537, -0.00000593, -0.00000217, 0.00000106, 0.00000222, -0.00000179, 0.00000372, -0.00001160, -0.00000232, -0.00000014, 0.00000027, -0.00001006, -0.00000200, 0.00000129, -0.00000117, 0.00000382, -0.00000259, -0.00000325, -0.00000708, -0.00000459, -0.00000535, -0.00000640, -0.00000559, -0.00005478, -0.00003069, 0.00009444, 0.00004715, 0.00008856, 0.00003168, -0.00006598, -0.00001556, -0.00001066, -0.00000947, -0.00000438, -0.00000111, 0.00000348, 0.00000081, -0.00001196, -0.00000540, 0.00000191, -0.00000188, -0.00000295, -0.00000012, 0.00000512, 0.00000037, 0.00000182, -0.00000146, 0.00000819, -0.00000027, -0.00000313, -0.00000721, 0.00000905, 0.00000707, -0.00000378, 0.00000201, -0.00000657, 0.00000442, -0.00000864, 0.00000116, -0.00000012, -0.00000031, -0.00000811, 0.00000577, -0.00001537, 0.00000280, 0.00000322, 0.00000304, -0.00001030, -0.00001003, 0.00001618, -0.00000389, 0.00000177, 0.00000220, 0.00000263, },
		{ 0.00000169, -0.00000092, -0.00000000, 0.00001273, -0.00000556, -0.00001403, 0.00000111, 0.00000803, 0.00000046, -0.00001471, 0.00000375, -0.00000370, -0.00000268, 0.00000091, 0.00000071, -0.00000563, -0.00000237, 0.00000193, -0.00000307, 0.00000024, -0.00000101, 0.00001247, -0.00000333, -0.00000634, 0.00000080, 0.00000806, -0.00000185, 0.00000155, 0.00000002, 0.00000807, -0.00000251, -0.00000088, -0.00000531, 0.00000410, -0.00000456, -0.00001190, -0.00000018, 0.00000331, 0.00000029, -0.00000304, -0.00000992, -0.00000667, -0.00001993, -0.00004779, 0.00003113, 0.00006471, 0.00003084, 0.00006852, 0.00001595, -0.00005377, -0.00001325, -0.00000142, -0.00000956, -0.00000214, -0.00000685, -0.00000434, -0.00000205, 0.00000282, -0.00000004, 0.00000068, 0.00000038, -0.00001203, 0.00000053, -0.00000077, -0.00000102, -0.00001120, 0.00000222, -0.00000195, 0.00000215, 0.00000130, -0.00000218, -0.00000506, -0.00000605, -0.00000553, -0.00000435, -0.00000718, -0.00001046, -0.00006939, 0.00000460, 0.00009837, 0.00004652, 0.00010174, -0.00001503, -0.00005805, -0.00000779, -0.00001282, -0.00000739, -0.00000362, -0.00000036, 0.00000465, -0.00000326, -0.00001217, -0.00000211, 0.00000098, -0.00000207, -0.00000300, 0.00000209, 0.00000386, 0.00000038, 0.00000106, 0.00000029, 0.00000851, -0.00000353, -0.00000248, -0.00000710, 0.00001429, 0.00000059, -0.00000005, -0.00000177, -0.00000363, 0.00000216, -0.00000798, 0.00000205, 0.00000030, -0.00000348, -0.00000458, 0.00000242, -0.00001585, 0.00000896, -0.00000145, 0.00000511, -0.00001741, 0.00000099, 0.00001210, -0.00000544, 0.00000383, 0.00000186, 0.00000253, },
		{ 0.00000167, -0.00000286, 0.00000492, 0.00001025, -0.00001114, -0.00001077, 0.00000417, 0.00000857, -0.00000594, -0.00001143, 0.00000532, -0.00000633, 0.00000022, -0.00000025, 0.00000094, -0.00000778, 0.00000122, -0.00000058, -0.00000087, -0.00000208, 0.00000310, 0.00001028, -0.00000566, -0.00000587, 0.00000489, 0.00000510, -0.00000145, 0.00000079, 0.00000254, 0.00000678, -0.00000419, -0.00000070, -0.00000508, 0.00000577, -0.00000982, -0.00000846, 0.00000134, 0.00000317, -0.00000079, -0.00000429, -0.00001075, -0.00000509, -0.00003276, -0.00003292, 0.00005325, 0.00004974, 0.00003519, 0.00006630, -0.00000885, -0.00004843, -0.00000411, -0.00000484, -0.00000806, -0.00000204, -0.00000761, -0.00000291, -0.00000158, 0.00000378, -0.00000180, 0.00000315, -0.00000448, -0.00000969, 0.00000159, -0.00000080, -0.00000368, -0.00000957, 0.00000398, -0.00000383, 0.00000499, -0.00000175, -0.00000131, -0.00000703, -0.00000479, -0.00000647, -0.00000333, -0.00000836, -0.00001933, -0.00007704, 0.00004653, 0.00008763, 0.00006178, 0.00009288, -0.00005596, -0.00003787, -0.00000880, -0.00001132, -0.00000697, -0.00000203, -0.00000006, 0.00000549, -0.00000845, -0.00000945, -0.00000083, 0.00000070, -0.00000275, -0.00000204, 0.00000348, 0.00000239, 0.00000082, 0.00000006, 0.00000324, 0.00000628, -0.00000479, -0.00000300, -0.00000454, 0.00001674, -0.00000501, 0.00000409, -0.00000652, 0.00000081, -0.00000190, -0.00000534, 0.00000166, 0.00000085, -0.00000693, 0.00000051, -0.00000464, -0.00001021, 0.00000942, -0.00000274, 0.00000378, -0.00002067, 0.00001167, 0.00000432, -0.00000324, 0.00000414, 0.00000211, 0.00000203, },
		{ 0.00000083, -0.00000316, 0.00000968, 0.00000468, -0.00001389, -0.00000716, 0.00000742, 0.00000637, -0.00001097, -0.00000615, 0.00000396, -0.00000668, 0.00000181, -0.00000076, -0.00000001, -0.00000823, 0.00000373, -0.00000257, 0.00000083, -0.00000392, 0.00000742, 0.00000660, -0.00000709, -0.00000425, 0.00000805, 0.00000121, 0.00000031, -0.00000077, 0.00000648, 0.00000265, -0.00000327, -0.00000237, -0.00000240, 0.00000421, -0.00001292, -0.00000467, 0.00000223, 0.00000245, -0.00000137, -0.00000662, -0.00000898, -0.00000846, -0.00004172, -0.00000932, 0.00006314, 0.00003517, 0.00004455, 0.00005496, -0.00002804, -0.00003839, 0.00000115, -0.00000847, -0.00000532, -0.00000342, -0.00000707, -0.00000222, -0.00000036, 0.00000339, -0.00000206, 0.00000403, -0.00000896, -0.00000583, 0.00000122, -0.00000062, -0.00000687, -0.00000569, 0.00000283, -0.00000308, 0.00000568, -0.00000376, -0.00000116, -0.00000811, -0.00000436, -0.00000621, -0.00000377, -0.00000891, -0.00003373, -0.00007140, 0.00008549, 0.00006996, 0.00008836, 0.00005741, -0.00007667, -0.00001767, -0.00001356, -0.00000795, -0.00000722, -0.00000048, 0.00000073, 0.00000450, -0.00001258, -0.00000545, -0.00000104, 0.00000097, -0.00000375, -0.00000018, 0.00000365, 0.00000139, 0.00000107, -0.00000037, 0.00000620, 0.00000227, -0.00000397, -0.00000497, 0.00000091, 0.00001491, -0.00000740, 0.00000624, -0.00001005, 0.00000488, -0.00000624, -0.00000184, 0.00000074, 0.00000054, -0.00000875, 0.00000431, -0.00001210, -0.00000081, 0.00000450, 0.00000038, -0.00000245, -0.00001714, 0.00001748, -0.00000317, 0.00000076, 0.00000329, 0.00000255, 0.00000158, },
		{ -0.00000101, -0.00000080, 0.00001192, -0.00000177, -0.00001436, -0.00000304, 0.00000985, 0.00000182, -0.00001316, -0.00000092, 0.00000087, -0.00000501, 0.00000181, -0.00000036, -0.00000227, -0.00000652, 0.00000427, -0.00000303, 0.00000097, -0.00000402, 0.00001048, 0.00000259, -0.00000809, -0.00000086, 0.00000856, -0.00000136, 0.00000147, -0.00000111, 0.00000944, -0.00000224, -0.00000104, -0.00000482, 0.00000167, -0.00000022, -0.00001310, -0.00000149, 0.00000259, 0.00000164, -0.00000223, -0.00000870, -0.00000613, -0.00001684, -0.00004208, 0.00001646, 0.00006115, 0.00002608, 0.00005397, 0.00003667, -0.00003947, -0.00002658, 0.00000262, -0.00001102, -0.00000234, -0.00000575, -0.00000532, -0.00000230, 0.00000149, 0.00000179, -0.00000068, 0.00000248, -0.00001144, -0.00000199, 0.00000024, -0.00000091, -0.00000927, -0.00000114, -0.00000026, 0.00000015, 0.00000356, -0.00000371, -0.00000262, -0.00000756, -0.00000542, -0.00000402, -0.00000674, -0.00000836, -0.00005370, -0.00004763, 0.00011082, 0.00005782, 0.00011245, 0.00000306, -0.00007176, -0.00000741, -0.00001573, -0.00000609, -0.00000614, -0.00000032, 0.00000295, 0.00000040, -0.00001337, -0.00000261, -0.00000110, 0.00000078, -0.00000409, 0.00000161, 0.00000314, 0.00000084, 0.00000103, 0.00000022, 0.00000796, -0.00000210, -0.00000199, -0.00000753, 0.00000837, 0.00000881, -0.00000543, 0.00000464, -0.00001032, 0.00000654, -0.00000902, 0.00000108, 0.00000015, -0.00000137, -0.00000751, 0.00000410, -0.00001573, 0.00000761, -0.00000223, 0.00000555, -0.00001179, -0.00000712, 0.00001616, -0.00000696, 0.00000413, 0.00000240, 0.00000264, 0.00000148, },
		{ -0.00000308, 0.00000378, 0.00001050, -0.00000713, -0.00001340, 0.00000180, 0.00001017, -0.00000358, -0.00001232, 0.00000269, -0.00000239, -0.00000235, 0.00000077, 0.00000048, -0.00000518, -0.00000314, 0.00000293, -0.00000189, -0.00000066, -0.00000187, 0.00001142, -0.00000103, -0.00000852, 0.00000383, 0.00000596, -0.00000133, 0.00000068, 0.00000131, 0.00000906, -0.00000521, 0.00000032, -0.00000606, 0.00000490, -0.00000565, -0.00001097, 0.00000085, 0.00000230, 0.00000127, -0.00000426, -0.00000873, -0.00000536, -0.00002690, -0.00003273, 0.00003827, 0.00005067, 0.00002480, 0.00005867, 0.00001551, -0.00004320, -0.00001527, 0.00000083, -0.00001144, -0.00000046, -0.00000781, -0.00000322, -0.00000255, 0.00000326, -0.00000016, 0.00000148, -0.00000112, -0.00001126, 0.00000067, -0.00000067, -0.00000214, -0.00000975, 0.00000224, -0.00000334, 0.00000432, -0.00000051, -0.00000186, -0.00000552, -0.00000563, -0.00000750, -0.00000066, -0.00001163, -0.00000831, -0.00007569, -0.00000598, 0.00011608, 0.00006214, 0.00011656, -0.00005192, -0.00004800, -0.00000922, -0.00001244, -0.00000736, -0.00000300, -0.00000170, 0.00000617, -0.00000634, -0.00001017, -0.00000237, 0.00000019, -0.00000071, -0.00000296, 0.00000238, 0.00000272, 0.00000034, 0.00000102, 0.00000175, 0.00000767, -0.00000520, -0.00000047, -0.00000886, 0.00001546, 0.00000041, 0.00000003, -0.00000071, -0.00000668, 0.00000474, -0.00000906, 0.00000246, 0.00000015, -0.00000454, -0.00000344, -0.00000087, -0.00001288, 0.00001057, -0.00000599, 0.00000837, -0.00001966, 0.00000556, 0.00000895, -0.00000602, 0.00000547, 0.00000217, 0.00000222, 0.00000146, },
		{ -0.00000387, 0.00000869, 0.00000606, -0.00001061, -0.00001126, 0.00000684, 0.00000775, -0.00000812, -0.00000926, 0.00000399, -0.00000447, -0.00000002, -0.00000029, 0.00000076, -0.00000740, 0.00000052, 0.00000084, -0.00000013, -0.00000315, 0.00000190, 0.00001031, -0.00000419, -0.00000752, 0.00000807, 0.00000173, 0.00000061, -0.00000152, 0.00000607, 0.00000483, -0.00000489, -0.00000057, -0.00000472, 0.00000539, -0.00000997, -0.00000770, 0.00000237, 0.00000159, 0.00000118, -0.00000724, -0.00000623, -0.00000910, -0.00003383, -0.00001616, 0.00005219, 0.00003655, 0.00003073, 0.00005561, -0.00000401, -0.00004095, -0.00000587, -0.00000326, -0.00000933, -0.00000075, -0.00000834, -0.00000188, -0.00000212, 0.00000408, -0.00000134, 0.00000302, -0.00000542, -0.00000887, 0.00000177, -0.00000126, -0.00000418, -0.00000795, 0.00000309, -0.00000437, 0.00000709, -0.00000448, 0.00000020, -0.00000849, -0.00000374, -0.00000880, 0.00000146, -0.00001580, -0.00001280, -0.00009169, 0.00004596, 0.00010367, 0.00008488, 0.00008898, -0.00008714, -0.00002056, -0.00001709, -0.00000614, -0.00001011, 0.00000064, -0.00000292, 0.00000810, -0.00001300, -0.00000489, -0.00000400, 0.00000250, -0.00000310, -0.00000067, 0.00000204, 0.00000262, -0.00000021, 0.00000126, 0.00000383, 0.00000512, -0.00000586, -0.00000091, -0.00000695, 0.00001913, -0.00000685, 0.00000606, -0.00000764, -0.00000051, 0.00000013, -0.00000653, 0.00000236, 0.00000013, -0.00000739, 0.00000119, -0.00000833, -0.00000444, 0.00000672, -0.00000399, 0.00000534, -0.00002133, 0.00001542, -0.00000014, -0.00000184, 0.00000494, 0.00000245, 0.00000165, 0.00000084, },
		{ -0.00000217, 0.00001161, 0.00000042, -0.00001245, -0.00000765, 0.00001071, 0.00000321, -0.00001069, -0.00000526, 0.00000319, -0.00000476, 0.00000109, -0.00000056, -0.00000031, -0.00000771, 0.00000292, -0.00000052, 0.00000074, -0.00000491, 0.00000590, 0.00000792, -0.00000717, -0.00000418, 0.00000964, -0.00000155, 0.00000228, -0.00000287, 0.00001060, -0.00000128, -0.00000217, -0.00000320, -0.00000118, 0.00000280, -0.00001201, -0.00000419, 0.00000288, 0.00000120, 0.00000023, -0.00000942, -0.00000314, -0.00001689, -0.00003428, 0.00000348, 0.00005670, 0.00002392, 0.00004058, 0.00004461, -0.00001862, -0.00003495, 0.00000071, -0.00000816, -0.00000535, -0.00000321, -0.00000686, -0.00000189, -0.00000055, 0.00000348, -0.00000111, 0.00000282, -0.00000885, -0.00000543, 0.00000159, -0.00000169, -0.00000640, -0.00000450, 0.00000130, -0.00000229, 0.00000658, -0.00000612, 0.00000047, -0.00000965, -0.00000372, -0.00000724, -0.00000054, -0.00001601, -0.00002655, -0.00009132, 0.00009457, 0.00008538, 0.00011526, 0.00003208, -0.00009073, -0.00000418, -0.00002194, -0.00000206, -0.00001076, 0.00000200, -0.00000159, 0.00000598, -0.00001600, -0.00000090, -0.00000513, 0.00000407, -0.00000486, 0.00000146, 0.00000150, 0.00000233, -0.00000040, 0.00000171, 0.00000586, 0.00000085, -0.00000394, -0.00000380, -0.00000090, 0.00001714, -0.00000962, 0.00000910, -0.00001278, 0.00000535, -0.00000517, -0.00000283, 0.00000161, -0.00000092, -0.00000795, 0.00000317, -0.00001392, 0.00000522, -0.00000113, 0.00000260, -0.00000352, -0.00001490, 0.00001833, -0.00000669, 0.00000283, 0.00000371, 0.00000266, 0.00000134, -0.00000079, },
		{ 0.00000200, 0.00001115, -0.00000462, -0.00001314, -0.00000241, 0.00001197, -0.00000200, -0.00001104, -0.00000155, 0.00000119, -0.00000358, 0.00000092, 0.00000000, -0.00000266, -0.00000584, 0.00000334, -0.00000032, -0.00000026, -0.00000464, 0.00000877, 0.00000499, -0.00000975, 0.00000136, 0.00000744, -0.00000190, 0.00000165, -0.00000105, 0.00001177, -0.00000600, 0.00000049, -0.00000554, 0.00000259, -0.00000155, -0.00001188, -0.00000082, 0.00000216, 0.00000186, -0.00000269, -0.00000863, -0.00000296, -0.00002516, -0.00002804, 0.00002251, 0.00005213, 0.00001718, 0.00004928, 0.00002845, -0.00002730, -0.00002691, 0.00000374, -0.00001189, -0.00000125, -0.00000658, -0.00000410, -0.00000287, 0.00000178, 0.00000186, 0.00000025, 0.00000068, -0.00001038, -0.00000208, 0.00000063, -0.00000226, -0.00000788, -0.00000079, -0.00000192, 0.00000223, 0.00000254, -0.00000450, -0.00000213, -0.00000805, -0.00000627, -0.00000226, -0.00000770, -0.00001120, -0.00005124, -0.00006685, 0.00012600, 0.00007687, 0.00013317, -0.00003632, -0.00006601, -0.00000473, -0.00001833, -0.00000369, -0.00000715, -0.00000025, 0.00000300, -0.00000104, -0.00001338, -0.00000072, -0.00000364, 0.00000327, -0.00000455, 0.00000208, 0.00000177, 0.00000123, 0.00000026, 0.00000220, 0.00000708, -0.00000371, -0.00000070, -0.00000787, 0.00000808, 0.00000962, -0.00000655, 0.00000688, -0.00001339, 0.00000812, -0.00000880, 0.00000041, 0.00000097, -0.00000336, -0.00000542, 0.00000047, -0.00001372, 0.00001065, -0.00000755, 0.00000914, -0.00001418, -0.00000273, 0.00001373, -0.00000825, 0.00000578, 0.00000285, 0.00000240, 0.00000125, -0.00000294, },
		{ 0.00000711, 0.00000753, -0.00000818, -0.00001266, 0.00000375, 0.00001002, -0.00000632, -0.00000962, 0.00000097, -0.00000083, -0.00000191, 0.00000022, 0.00000060, -0.00000523, -0.00000276, 0.00000239, 0.00000096, -0.00000282, -0.00000221, 0.00001002, 0.00000161, -0.00001083, 0.00000734, 0.00000255, 0.00000055, -0.00000133, 0.00000428, 0.00000805, -0.00000695, 0.00000089, -0.00000562, 0.00000442, -0.00000575, -0.00001030, 0.00000207, 0.00000062, 0.00000323, -0.00000730, -0.00000436, -0.00000813, -0.00002946, -0.00001743, 0.00003834, 0.00004049, 0.00001869, 0.00005197, 0.00001166, -0.00003112, -0.00001783, 0.00000280, -0.00001271, 0.00000088, -0.00000889, -0.00000162, -0.00000364, 0.00000384, 0.00000022, 0.00000167, -0.00000259, -0.00000989, 0.00000040, -0.00000063, -0.00000315, -0.00000793, 0.00000161, -0.00000441, 0.00000695, -0.00000321, -0.00000077, -0.00000670, -0.00000466, -0.00000991, 0.00000390, -0.00001752, -0.00000486, -0.00008196, -0.00001867, 0.00013375, 0.00008821, 0.00011954, -0.00009059, -0.00002970, -0.00001668, -0.00000815, -0.00000963, -0.00000072, -0.00000437, 0.00000856, -0.00001052, -0.00000651, -0.00000411, 0.00000051, 0.00000010, -0.00000217, 0.00000108, 0.00000293, -0.00000057, 0.00000172, 0.00000281, 0.00000660, -0.00000668, 0.00000162, -0.00001045, 0.00001664, -0.00000048, 0.00000066, -0.00000006, -0.00000898, 0.00000666, -0.00000947, 0.00000229, 0.00000042, -0.00000619, -0.00000115, -0.00000599, -0.00000697, 0.00000875, -0.00000802, 0.00001048, -0.00002087, 0.00000971, 0.00000474, -0.00000524, 0.00000640, 0.00000265, 0.00000182, 0.00000082, -0.00000431, },
		{ 0.00001094, 0.00000226, -0.00001040, -0.00001046, 0.00000920, 0.00000553, -0.00000883, -0.00000720, 0.00000198, -0.00000194, -0.00000076, -0.00000007, 0.00000020, -0.00000656, -0.00000009, 0.00000159, 0.00000177, -0.00000543, 0.00000134, 0.00001000, -0.00000255, -0.00000891, 0.00001094, -0.00000211, 0.00000343, -0.00000434, 0.00001070, 0.00000091, -0.00000425, -0.00000125, -0.00000313, 0.00000349, -0.00000859, -0.00000774, 0.00000374, -0.00000046, 0.00000360, -0.00001118, 0.00000086, -0.00001765, -0.00002749, -0.00000509, 0.00004892, 0.00002566, 0.00002755, 0.00004622, -0.00000178, -0.00003187, -0.00000859, -0.00000156, -0.00001003, -0.00000014, -0.00000865, -0.00000086, -0.00000304, 0.00000463, -0.00000050, 0.00000203, -0.00000574, -0.00000795, 0.00000172, -0.00000186, -0.00000430, -0.00000635, 0.00000166, -0.00000420, 0.00000902, -0.00000762, 0.00000228, -0.00001061, -0.00000215, -0.00001139, 0.00000671, -0.00002444, -0.00000469, -0.00010686, 0.00004253, 0.00012334, 0.00011581, 0.00006828, -0.00011041, -0.00000220, -0.00002784, 0.00000124, -0.00001459, 0.00000419, -0.00000642, 0.00001077, -0.00001760, 0.00000036, -0.00000803, 0.00000505, -0.00000359, 0.00000070, -0.00000025, 0.00000391, -0.00000205, 0.00000329, 0.00000377, 0.00000389, -0.00000661, 0.00000099, -0.00000879, 0.00002079, -0.00000859, 0.00000793, -0.00000853, -0.00000166, 0.00000198, -0.00000753, 0.00000296, -0.00000083, -0.00000746, 0.00000174, -0.00001234, 0.00000295, 0.00000095, -0.00000194, 0.00000448, -0.00001953, 0.00001691, -0.00000386, -0.00000015, 0.00000550, 0.00000271, 0.00000132, -0.00000047, -0.00000352, },
		{ 0.00001178, -0.00000287, -0.00001168, -0.00000616, 0.00001218, 0.00000006, -0.00000942, -0.00000461, 0.00000181, -0.00000199, -0.00000054, 0.00000040, -0.00000155, -0.00000588, 0.00000101, 0.00000211, 0.00000069, -0.00000638, 0.00000464, 0.00000923, -0.00000738, -0.00000340, 0.00001003, -0.00000359, 0.00000385, -0.00000418, 0.00001428, -0.00000597, -0.00000039, -0.00000413, 0.00000037, 0.00000065, -0.00001002, -0.00000427, 0.00000342, 0.00000029, 0.00000101, -0.00001112, 0.00000246, -0.00002690, -0.00002062, 0.00000788, 0.00005211, 0.00001307, 0.00003927, 0.00003346, -0.00001030, -0.00003054, -0.00000062, -0.00000744, -0.00000506, -0.00000373, -0.00000590, -0.00000205, -0.00000079, 0.00000389, -0.00000008, 0.00000086, -0.00000782, -0.00000531, 0.00000184, -0.00000286, -0.00000544, -0.00000367, -0.00000050, -0.00000063, 0.00000667, -0.00000810, 0.00000194, -0.00001124, -0.00000304, -0.00000790, 0.00000216, -0.00002315, -0.00001859, -0.00011153, 0.00009957, 0.00011005, 0.00014172, -0.00000758, -0.00009189, 0.00000408, -0.00002824, 0.00000302, -0.00001363, 0.00000385, -0.00000336, 0.00000641, -0.00001809, 0.00000287, -0.00000886, 0.00000705, -0.00000537, 0.00000203, -0.00000027, 0.00000336, -0.00000197, 0.00000410, 0.00000505, -0.00000060, -0.00000350, -0.00000297, -0.00000193, 0.00001807, -0.00001093, 0.00001104, -0.00001439, 0.00000510, -0.00000350, -0.00000432, 0.00000300, -0.00000331, -0.00000576, 0.00000044, -0.00001396, 0.00001026, -0.00000748, 0.00000678, -0.00000638, -0.00001039, 0.00001631, -0.00000833, 0.00000425, 0.00000433, 0.00000253, 0.00000106, -0.00000247, -0.00000017, },
		{ 0.00000933, -0.00000681, -0.00001199, -0.00000027, 0.00001174, -0.00000471, -0.00000850, -0.00000255, 0.00000124, -0.00000156, -0.00000074, 0.00000104, -0.00000388, -0.00000376, 0.00000081, 0.00000381, -0.00000242, -0.00000500, 0.00000691, 0.00000759, -0.00001157, 0.00000426, 0.00000487, -0.00000107, 0.00000072, 0.00000071, 0.00001219, -0.00000902, 0.00000164, -0.00000539, 0.00000268, -0.00000243, -0.00001048, -0.00000021, 0.00000120, 0.00000297, -0.00000487, -0.00000570, -0.00000312, -0.00003041, -0.00001256, 0.00002159, 0.00004637, 0.00000809, 0.00004750, 0.00001827, -0.00001509, -0.00002692, 0.00000413, -0.00001203, -0.00000040, -0.00000770, -0.00000228, -0.00000405, 0.00000222, 0.00000231, 0.00000080, -0.00000136, -0.00000856, -0.00000257, 0.00000093, -0.00000352, -0.00000614, -0.00000101, -0.00000338, 0.00000491, 0.00000054, -0.00000441, -0.00000242, -0.00000799, -0.00000749, 0.00000027, -0.00000992, -0.00001303, -0.00004880, -0.00008655, 0.00013737, 0.00010967, 0.00014247, -0.00008052, -0.00004974, -0.00000888, -0.00001675, -0.00000379, -0.00000631, -0.00000162, 0.00000408, -0.00000383, -0.00001154, -0.00000041, -0.00000502, 0.00000514, -0.00000420, 0.00000103, 0.00000159, 0.00000093, 0.00000004, 0.00000396, 0.00000593, -0.00000523, 0.00000081, -0.00000829, 0.00000816, 0.00000933, -0.00000668, 0.00000802, -0.00001497, 0.00000848, -0.00000753, -0.00000113, 0.00000252, -0.00000637, -0.00000174, -0.00000528, -0.00000869, 0.00001044, -0.00001076, 0.00001214, -0.00001632, 0.00000215, 0.00000944, -0.00000771, 0.00000653, 0.00000358, 0.00000201, 0.00000068, -0.00000416, 0.00000473, },
		{ 0.00000464, -0.00000938, -0.00001082, 0.00000572, 0.00000823, -0.00000769, -0.00000680, -0.00000136, 0.00000104, -0.00000153, -0.00000046, 0.00000075, -0.00000530, -0.00000177, 0.00000078, 0.00000507, -0.00000610, -0.00000202, 0.00000842, 0.00000428, -0.00001287, 0.00001077, -0.00000162, 0.00000337, -0.00000405, 0.00000864, 0.00000482, -0.00000721, 0.00000058, -0.00000397, 0.00000259, -0.00000468, -0.00001003, 0.00000336, -0.00000138, 0.00000562, -0.00001144, 0.00000264, -0.00001519, -0.00002568, -0.00000618, 0.00003544, 0.00003261, 0.00001325, 0.00004710, 0.00000591, -0.00001864, -0.00002041, 0.00000419, -0.00001299, 0.00000136, -0.00000952, -0.00000002, -0.00000506, 0.00000458, 0.00000098, 0.00000111, -0.00000368, -0.00000818, -0.00000020, -0.00000069, -0.00000387, -0.00000600, 0.00000037, -0.00000477, 0.00000945, -0.00000637, 0.00000095, -0.00000858, -0.00000316, -0.00001247, 0.00000875, -0.00002428, -0.00000052, -0.00008792, -0.00003375, 0.00015152, 0.00012723, 0.00010312, -0.00012226, -0.00000862, -0.00002755, -0.00000122, -0.00001354, 0.00000270, -0.00000783, 0.00001116, -0.00001501, -0.00000189, -0.00000692, 0.00000173, 0.00000061, -0.00000121, -0.00000114, 0.00000414, -0.00000218, 0.00000306, 0.00000348, 0.00000524, -0.00000772, 0.00000343, -0.00001145, 0.00001722, -0.00000145, 0.00000142, 0.00000028, -0.00001037, 0.00000770, -0.00000913, 0.00000156, 0.00000091, -0.00000806, 0.00000174, -0.00001219, 0.00000127, 0.00000332, -0.00000640, 0.00001002, -0.00001978, 0.00001211, 0.00000059, -0.00000366, 0.00000681, 0.00000320, 0.00000143, -0.00000033, -0.00000422, 0.00000921, },
		{ -0.00000068, -0.00001082, -0.00000770, 0.00000998, 0.00000304, -0.00000860, -0.00000506, -0.00000083, 0.00000138, -0.00000222, 0.00000070, -0.00000105, -0.00000478, -0.00000129, 0.00000240, 0.00000402, -0.00000829, 0.00000107, 0.00000974, -0.00000122, -0.00000944, 0.00001269, -0.00000545, 0.00000601, -0.00000658, 0.00001523, -0.00000419, -0.00000257, -0.00000235, -0.00000093, 0.00000068, -0.00000619, -0.00000812, 0.00000482, -0.00000205, 0.00000520, -0.00001427, 0.00000813, -0.00002820, -0.00001522, -0.00000107, 0.00004643, 0.00001548, 0.00002618, 0.00003719, -0.00000078, -0.00002245, -0.00001150, -0.00000012, -0.00000986, -0.00000063, -0.00000800, -0.00000041, -0.00000389, 0.00000531, 0.00000056, 0.00000025, -0.00000530, -0.00000708, 0.00000141, -0.00000249, -0.00000405, -0.00000491, -0.00000025, -0.00000298, 0.00000999, -0.00001026, 0.00000410, -0.00001272, -0.00000063, -0.00001337, 0.00001110, -0.00003257, 0.00000307, -0.00012037, 0.00003348, 0.00015066, 0.00015145, 0.00002841, -0.00011908, 0.00001106, -0.00003673, 0.00000784, -0.00001837, 0.00000710, -0.00000926, 0.00001225, -0.00002085, 0.00000491, -0.00001179, 0.00000756, -0.00000357, 0.00000114, -0.00000226, 0.00000528, -0.00000392, 0.00000546, 0.00000353, 0.00000227, -0.00000660, 0.00000218, -0.00000947, 0.00002088, -0.00000921, 0.00000869, -0.00000837, -0.00000316, 0.00000394, -0.00000861, 0.00000377, -0.00000243, -0.00000655, 0.00000127, -0.00001529, 0.00001022, -0.00000648, 0.00000297, 0.00000099, -0.00001487, 0.00001537, -0.00000589, 0.00000115, 0.00000610, 0.00000279, 0.00000101, -0.00000203, -0.00000193, 0.00001142, },
		{ -0.00000539, -0.00001115, -0.00000293, 0.00001130, -0.00000217, -0.00000774, -0.00000398, -0.00000024, 0.00000161, -0.00000304, 0.00000211, -0.00000367, -0.00000270, -0.00000221, 0.00000575, 0.00000005, -0.00000774, 0.00000330, 0.00001074, -0.00000796, -0.00000164, 0.00000877, -0.00000425, 0.00000426, -0.00000373, 0.00001616, -0.00001018, 0.00000132, -0.00000455, 0.00000149, -0.00000142, -0.00000766, -0.00000437, 0.00000330, 0.00000042, -0.00000005, -0.00001007, 0.00000514, -0.00003463, -0.00000501, 0.00000597, 0.00004986, 0.00000235, 0.00004009, 0.00002179, -0.00000275, -0.00002581, -0.00000242, -0.00000640, -0.00000455, -0.00000492, -0.00000408, -0.00000294, -0.00000076, 0.00000447, 0.00000080, -0.00000152, -0.00000603, -0.00000544, 0.00000188, -0.00000387, -0.00000419, -0.00000324, -0.00000235, 0.00000191, 0.00000548, -0.00000893, 0.00000237, -0.00001208, -0.00000297, -0.00000743, 0.00000313, -0.00002861, -0.00001202, -0.00012936, 0.00009689, 0.00015026, 0.00015953, -0.00005697, -0.00007925, 0.00000415, -0.00002963, 0.00000502, -0.00001417, 0.00000385, -0.00000364, 0.00000498, -0.00001793, 0.00000481, -0.00001125, 0.00000924, -0.00000495, 0.00000106, -0.00000085, 0.00000365, -0.00000292, 0.00000614, 0.00000422, -0.00000234, -0.00000240, -0.00000272, -0.00000197, 0.00001730, -0.00001063, 0.00001120, -0.00001401, 0.00000345, -0.00000081, -0.00000662, 0.00000514, -0.00000682, -0.00000191, -0.00000436, -0.00001119, 0.00001247, -0.00001240, 0.00001122, -0.00001000, -0.00000437, 0.00001172, -0.00000784, 0.00000470, 0.00000525, 0.00000219, 0.00000056, -0.00000372, 0.00000220, 0.00001059, },
		{ -0.00000891, -0.00001001, 0.00000231, 0.00000962, -0.00000611, -0.00000582, -0.00000388, 0.00000116, 0.00000077, -0.00000283, 0.00000237, -0.00000542, -0.00000083, -0.00000283, 0.00000901, -0.00000545, -0.00000494, 0.00000495, 0.00001016, -0.00001328, 0.00000742, 0.00000116, 0.00000095, -0.00000119, 0.00000425, 0.00001012, -0.00001049, 0.00000196, -0.00000416, 0.00000167, -0.00000249, -0.00000921, 0.00000042, -0.00000017, 0.00000474, -0.00000846, 0.00000009, -0.00000754, -0.00002996, -0.00000019, 0.00001768, 0.00004233, -0.00000022, 0.00004708, 0.00000759, -0.00000369, -0.00002637, 0.00000369, -0.00001118, -0.00000026, -0.00000859, -0.00000029, -0.00000562, 0.00000288, 0.00000303, 0.00000082, -0.00000329, -0.00000623, -0.00000339, 0.00000107, -0.00000447, -0.00000427, -0.00000183, -0.00000423, 0.00000765, -0.00000219, -0.00000341, -0.00000371, -0.00000720, -0.00000906, 0.00000336, -0.00001332, -0.00001357, -0.00004699, -0.00010584, 0.00014350, 0.00016077, 0.00013100, -0.00012023, -0.00002782, -0.00001813, -0.00001147, -0.00000634, -0.00000368, -0.00000428, 0.00000587, -0.00000753, -0.00000816, -0.00000163, -0.00000514, 0.00000639, -0.00000341, -0.00000111, 0.00000249, -0.00000012, 0.00000047, 0.00000531, 0.00000458, -0.00000648, 0.00000230, -0.00000846, 0.00000818, 0.00000834, -0.00000597, 0.00000796, -0.00001476, 0.00000733, -0.00000511, -0.00000350, 0.00000461, -0.00001002, 0.00000304, -0.00001258, -0.00000086, 0.00000627, -0.00001025, 0.00001280, -0.00001673, 0.00000613, 0.00000439, -0.00000582, 0.00000647, 0.00000450, 0.00000156, -0.00000036, -0.00000425, 0.00000664, 0.00000717, },
		{ -0.00001105, -0.00000719, 0.00000647, 0.00000585, -0.00000809, -0.00000379, -0.00000441, 0.00000341, -0.00000153, -0.00000090, 0.00000048, -0.00000478, -0.00000101, -0.00000100, 0.00000953, -0.00000982, -0.00000165, 0.00000701, 0.00000645, -0.00001399, 0.00001324, -0.00000560, 0.00000621, -0.00000646, 0.00001337, -0.00000009, -0.00000607, -0.00000039, -0.00000154, -0.00000016, -0.00000278, -0.00000985, 0.00000420, -0.00000291, 0.00000732, -0.00001507, 0.00001026, -0.00002468, -0.00001612, -0.00000110, 0.00003284, 0.00002523, 0.00000947, 0.00004258, -0.00000004, -0.00000708, -0.00002229, 0.00000468, -0.00001199, 0.00000055, -0.00000916, 0.00000102, -0.00000637, 0.00000540, 0.00000198, -0.00000013, -0.00000423, -0.00000629, -0.00000117, -0.00000074, -0.00000429, -0.00000409, -0.00000155, -0.00000390, 0.00001089, -0.00000903, 0.00000247, -0.00001050, -0.00000180, -0.00001426, 0.00001253, -0.00003026, 0.00000300, -0.00009178, -0.00005334, 0.00017226, 0.00017822, 0.00006219, -0.00013817, 0.00000808, -0.00003720, 0.00000535, -0.00001706, 0.00000570, -0.00001071, 0.00001264, -0.00001838, 0.00000235, -0.00000975, 0.00000313, 0.00000142, -0.00000100, -0.00000317, 0.00000544, -0.00000376, 0.00000442, 0.00000417, 0.00000327, -0.00000781, 0.00000437, -0.00001121, 0.00001632, -0.00000152, 0.00000142, 0.00000100, -0.00001127, 0.00000813, -0.00000835, 0.00000068, 0.00000109, -0.00000937, 0.00000417, -0.00001798, 0.00001021, -0.00000468, -0.00000114, 0.00000637, -0.00001562, 0.00001170, -0.00000236, -0.00000199, 0.00000698, 0.00000367, 0.00000106, -0.00000186, -0.00000282, 0.00000970, 0.00000231, },
		{ -0.00001154, -0.00000311, 0.00000853, 0.00000121, -0.00000792, -0.00000273, -0.00000445, 0.00000547, -0.00000442, 0.00000208, -0.00000306, -0.00000186, -0.00000352, 0.00000396, 0.00000575, -0.00001093, 0.00000060, 0.00000990, -0.00000062, -0.00000863, 0.00001259, -0.00000738, 0.00000741, -0.00000703, 0.00001817, -0.00000921, -0.00000057, -0.00000326, 0.00000112, -0.00000221, -0.00000350, -0.00000827, 0.00000494, -0.00000235, 0.00000479, -0.00001462, 0.00001274, -0.00003703, -0.00000069, -0.00000275, 0.00004558, 0.00000553, 0.00002651, 0.00002817, -0.00000012, -0.00001364, -0.00001405, 0.00000090, -0.00000889, -0.00000220, -0.00000625, -0.00000083, -0.00000432, 0.00000602, 0.00000160, -0.00000199, -0.00000420, -0.00000622, 0.00000068, -0.00000284, -0.00000366, -0.00000364, -0.00000257, -0.00000043, 0.00000921, -0.00001133, 0.00000455, -0.00001374, -0.00000020, -0.00001353, 0.00001289, -0.00003791, 0.00000777, -0.00012903, 0.00001471, 0.00019212, 0.00018441, -0.00002817, -0.00010983, 0.00001441, -0.00003973, 0.00001061, -0.00001920, 0.00000763, -0.00001003, 0.00001128, -0.00002141, 0.00000736, -0.00001404, 0.00000912, -0.00000248, -0.00000007, -0.00000295, 0.00000573, -0.00000493, 0.00000698, 0.00000366, -0.00000011, -0.00000542, 0.00000223, -0.00000857, 0.00001880, -0.00000783, 0.00000735, -0.00000626, -0.00000568, 0.00000650, -0.00001015, 0.00000515, -0.00000506, -0.00000407, -0.00000107, -0.00001574, 0.00001522, -0.00001333, 0.00000913, -0.00000427, -0.00000786, 0.00001080, -0.00000575, 0.00000163, 0.00000688, 0.00000269, 0.00000055, -0.00000342, 0.00000036, 0.00001039, -0.00000288, },
		{ -0.00000999, 0.00000105, 0.00000838, -0.00000320, -0.00000590, -0.00000343, -0.00000271, 0.00000568, -0.00000604, 0.00000416, -0.00000618, 0.00000131, -0.00000642, 0.00001028, -0.00000140, -0.00000861, 0.00000177, 0.00001243, -0.00000883, 0.00000094, 0.00000587, -0.00000334, 0.00000330, -0.00000107, 0.00001528, -0.00001285, 0.00000231, -0.00000402, 0.00000168, -0.00000280, -0.00000546, -0.00000422, 0.00000238, 0.00000169, -0.00000306, -0.00000569, 0.00000329, -0.00003690, 0.00000792, 0.00000144, 0.00004853, -0.00000724, 0.00004160, 0.00001068, 0.00000368, -0.00002085, -0.00000461, -0.00000483, -0.00000428, -0.00000628, -0.00000181, -0.00000440, -0.00000035, 0.00000516, 0.00000130, -0.00000391, -0.00000374, -0.00000573, 0.00000155, -0.00000446, -0.00000294, -0.00000323, -0.00000391, 0.00000501, 0.00000295, -0.00000819, 0.00000120, -0.00001162, -0.00000393, -0.00000546, 0.00000169, -0.00003121, -0.00000847, -0.00014273, 0.00008357, 0.00021232, 0.00015814, -0.00010706, -0.00005636, -0.00000371, -0.00002542, 0.00000309, -0.00001171, 0.00000158, -0.00000224, 0.00000164, -0.00001531, 0.00000447, -0.00001179, 0.00001034, -0.00000370, -0.00000132, -0.00000006, 0.00000289, -0.00000290, 0.00000745, 0.00000357, -0.00000430, -0.00000076, -0.00000292, -0.00000127, 0.00001503, -0.00000867, 0.00000927, -0.00001123, 0.00000006, 0.00000305, -0.00000974, 0.00000793, -0.00001120, 0.00000334, -0.00001104, -0.00000528, 0.00001060, -0.00001384, 0.00001389, -0.00001273, 0.00000170, 0.00000557, -0.00000555, 0.00000421, 0.00000640, 0.00000177, -0.00000036, -0.00000408, 0.00000418, 0.00000871, -0.00000764, },
		{ -0.00000643, 0.00000402, 0.00000659, -0.00000652, -0.00000292, -0.00000573, 0.00000127, 0.00000295, -0.00000481, 0.00000340, -0.00000648, 0.00000197, -0.00000660, 0.00001440, -0.00000881, -0.00000474, 0.00000321, 0.00001223, -0.00001425, 0.00001006, -0.00000286, 0.00000344, -0.00000331, 0.00000882, 0.00000574, -0.00001017, 0.00000141, -0.00000205, -0.00000026, -0.00000181, -0.00000798, 0.00000071, -0.00000135, 0.00000623, -0.00001215, 0.00000726, -0.00001516, -0.00002343, 0.00000627, 0.00001431, 0.00003801, -0.00000635, 0.00004622, -0.00000187, 0.00000580, -0.00002479, 0.00000222, -0.00000906, -0.00000128, -0.00000864, 0.00000128, -0.00000718, 0.00000369, 0.00000387, 0.00000028, -0.00000484, -0.00000361, -0.00000452, 0.00000111, -0.00000510, -0.00000236, -0.00000338, -0.00000388, 0.00000952, -0.00000477, -0.00000214, -0.00000548, -0.00000618, -0.00001021, 0.00000593, -0.00001672, -0.00001391, -0.00004484, -0.00012580, 0.00014565, 0.00023167, 0.00009092, -0.00014493, -0.00000791, -0.00002804, -0.00000520, -0.00000958, -0.00000061, -0.00000703, 0.00000719, -0.00001085, -0.00000439, -0.00000358, -0.00000450, 0.00000750, -0.00000302, -0.00000331, 0.00000367, -0.00000133, 0.00000110, 0.00000656, 0.00000273, -0.00000697, 0.00000319, -0.00000778, 0.00000729, 0.00000760, -0.00000524, 0.00000726, -0.00001312, 0.00000498, -0.00000191, -0.00000620, 0.00000660, -0.00001341, 0.00000780, -0.00002006, 0.00000845, -0.00000126, -0.00000555, 0.00001004, -0.00001423, 0.00000784, -0.00000014, -0.00000333, 0.00000593, 0.00000543, 0.00000112, -0.00000183, -0.00000307, 0.00000726, 0.00000522, -0.00001128, },
		{ -0.00000174, 0.00000521, 0.00000380, -0.00000798, -0.00000050, -0.00000820, 0.00000635, -0.00000212, -0.00000081, -0.00000049, -0.00000318, -0.00000122, -0.00000199, 0.00001324, -0.00001308, -0.00000167, 0.00000628, 0.00000755, -0.00001353, 0.00001386, -0.00000830, 0.00000800, -0.00000732, 0.00001698, -0.00000552, -0.00000412, -0.00000150, 0.00000093, -0.00000305, -0.00000069, -0.00000924, 0.00000393, -0.00000306, 0.00000704, -0.00001619, 0.00001574, -0.00003328, -0.00000368, -0.00000159, 0.00003193, 0.00001750, 0.00000778, 0.00003748, -0.00000492, 0.00000254, -0.00002301, 0.00000418, -0.00000980, -0.00000156, -0.00000757, 0.00000112, -0.00000722, 0.00000622, 0.00000292, -0.00000176, -0.00000432, -0.00000419, -0.00000267, -0.00000046, -0.00000469, -0.00000210, -0.00000426, -0.00000127, 0.00001021, -0.00000986, 0.00000249, -0.00001121, -0.00000181, -0.00001377, 0.00001322, -0.00003295, 0.00000283, -0.00009030, -0.00008150, 0.00020199, 0.00023541, -0.00000355, -0.00013220, 0.00001377, -0.00004057, 0.00000782, -0.00001744, 0.00000612, -0.00001124, 0.00001140, -0.00001893, 0.00000455, -0.00001120, 0.00000372, 0.00000324, -0.00000244, -0.00000380, 0.00000567, -0.00000433, 0.00000497, 0.00000546, 0.00000024, -0.00000642, 0.00000388, -0.00000917, 0.00001319, 0.00000035, -0.00000042, 0.00000305, -0.00001244, 0.00000857, -0.00000768, 0.00000024, 0.00000029, -0.00000918, 0.00000488, -0.00002157, 0.00001753, -0.00001303, 0.00000635, 0.00000009, -0.00000856, 0.00000815, -0.00000338, -0.00000083, 0.00000714, 0.00000401, 0.00000059, -0.00000341, -0.00000052, 0.00000873, 0.00000054, -0.00001293, },
		{ 0.00000241, 0.00000502, 0.00000040, -0.00000703, -0.00000034, -0.00000857, 0.00000997, -0.00000706, 0.00000379, -0.00000548, 0.00000190, -0.00000673, 0.00000649, 0.00000640, -0.00001274, -0.00000038, 0.00001064, -0.00000088, -0.00000637, 0.00001036, -0.00000738, 0.00000705, -0.00000493, 0.00001828, -0.00001286, 0.00000094, -0.00000339, 0.00000232, -0.00000441, -0.00000115, -0.00000778, 0.00000380, -0.00000099, 0.00000188, -0.00001117, 0.00001259, -0.00004066, 0.00001158, -0.00000680, 0.00004528, -0.00000337, 0.00002745, 0.00001969, 0.00000032, -0.00000578, -0.00001613, 0.00000177, -0.00000760, -0.00000433, -0.00000377, -0.00000202, -0.00000426, 0.00000675, 0.00000228, -0.00000422, -0.00000278, -0.00000525, -0.00000065, -0.00000258, -0.00000346, -0.00000245, -0.00000516, 0.00000343, 0.00000626, -0.00001001, 0.00000263, -0.00001268, -0.00000177, -0.00001087, 0.00001066, -0.00003845, 0.00000684, -0.00012975, -0.00001798, 0.00025538, 0.00020372, -0.00009360, -0.00008412, 0.00000623, -0.00003476, 0.00000765, -0.00001563, 0.00000467, -0.00000792, 0.00000720, -0.00001849, 0.00000676, -0.00001388, 0.00000915, -0.00000016, -0.00000310, -0.00000181, 0.00000463, -0.00000450, 0.00000733, 0.00000447, -0.00000332, -0.00000300, 0.00000108, -0.00000617, 0.00001451, -0.00000415, 0.00000344, -0.00000169, -0.00000969, 0.00000994, -0.00001236, 0.00000722, -0.00000877, 0.00000011, -0.00000565, -0.00001272, 0.00001612, -0.00001713, 0.00001425, -0.00000955, 0.00000009, 0.00000407, -0.00000362, 0.00000123, 0.00000781, 0.00000251, -0.00000026, -0.00000415, 0.00000264, 0.00000847, -0.00000482, -0.00001169, },
		{ 0.00000444, 0.00000435, -0.00000328, -0.00000380, -0.00000327, -0.00000511, 0.00000959, -0.00000890, 0.00000589, -0.00000822, 0.00000517, -0.00001070, 0.00001498, -0.00000334, -0.00000906, 0.00000024, 0.00001392, -0.00000949, 0.00000370, 0.00000195, -0.00000131, 0.00000142, 0.00000341, 0.00001146, -0.00001338, 0.00000228, -0.00000251, 0.00000084, -0.00000333, -0.00000352, -0.00000397, 0.00000109, 0.00000340, -0.00000707, 0.00000112, -0.00000282, -0.00003248, 0.00001481, -0.00000196, 0.00004601, -0.00001409, 0.00004195, 0.00000161, 0.00000821, -0.00001541, -0.00000736, -0.00000245, -0.00000479, -0.00000709, 0.00000027, -0.00000602, 0.00000032, 0.00000589, 0.00000129, -0.00000596, -0.00000124, -0.00000612, 0.00000088, -0.00000454, -0.00000180, -0.00000374, -0.00000466, 0.00000788, -0.00000029, -0.00000621, -0.00000141, -0.00000997, -0.00000559, -0.00000251, -0.00000167, -0.00003116, -0.00000806, -0.00015141, 0.00005909, 0.00030032, 0.00012718, -0.00014616, -0.00003064, -0.00001578, -0.00001757, -0.00000160, -0.00000716, -0.00000216, -0.00000000, -0.00000270, -0.00001099, 0.00000230, -0.00001071, 0.00001062, -0.00000227, -0.00000428, 0.00000153, 0.00000144, -0.00000219, 0.00000821, 0.00000286, -0.00000602, 0.00000092, -0.00000302, -0.00000060, 0.00001209, -0.00000571, 0.00000574, -0.00000639, -0.00000477, 0.00000772, -0.00001321, 0.00001073, -0.00001560, 0.00000902, -0.00001855, 0.00000296, 0.00000459, -0.00001068, 0.00001313, -0.00001294, 0.00000617, -0.00000074, -0.00000226, 0.00000312, 0.00000757, 0.00000137, -0.00000171, -0.00000330, 0.00000522, 0.00000672, -0.00001007, -0.00000735, },
		{ 0.00000393, 0.00000365, -0.00000640, 0.00000033, -0.00000828, 0.00000189, 0.00000450, -0.00000618, 0.00000361, -0.00000636, 0.00000362, -0.00000932, 0.00001878, -0.00001159, -0.00000492, 0.00000224, 0.00001316, -0.00001375, 0.00001119, -0.00000607, 0.00000533, -0.00000443, 0.00001304, 0.00000005, -0.00000844, 0.00000035, 0.00000035, -0.00000250, -0.00000085, -0.00000638, 0.00000008, -0.00000145, 0.00000619, -0.00001400, 0.00001325, -0.00002321, -0.00001288, 0.00000659, 0.00001340, 0.00003212, -0.00000941, 0.00004359, -0.00000875, 0.00001248, -0.00002184, -0.00000029, -0.00000577, -0.00000356, -0.00000752, 0.00000197, -0.00000834, 0.00000458, 0.00000459, -0.00000059, -0.00000603, -0.00000072, -0.00000616, 0.00000139, -0.00000568, -0.00000029, -0.00000589, -0.00000163, 0.00000931, -0.00000580, -0.00000195, -0.00000646, -0.00000618, -0.00000943, 0.00000597, -0.00001770, -0.00001668, -0.00003948, -0.00014995, 0.00014882, 0.00031889, 0.00001864, -0.00014570, 0.00000175, -0.00003278, -0.00000204, -0.00001053, 0.00000055, -0.00000790, 0.00000623, -0.00001195, -0.00000196, -0.00000483, -0.00000413, 0.00000922, -0.00000404, -0.00000426, 0.00000393, -0.00000169, 0.00000106, 0.00000833, -0.00000017, -0.00000601, 0.00000275, -0.00000554, 0.00000457, 0.00000826, -0.00000563, 0.00000698, -0.00001100, 0.00000221, 0.00000129, -0.00000843, 0.00000756, -0.00001533, 0.00001098, -0.00002574, 0.00001696, -0.00001020, 0.00000235, 0.00000393, -0.00000850, 0.00000651, -0.00000312, -0.00000098, 0.00000523, 0.00000625, 0.00000062, -0.00000344, -0.00000093, 0.00000659, 0.00000356, -0.00001379, -0.00000107, },
		{ 0.00000196, 0.00000257, -0.00000760, 0.00000292, -0.00001249, 0.00000967, -0.00000321, -0.00000025, -0.00000206, -0.00000052, -0.00000270, -0.00000157, 0.00001534, -0.00001507, -0.00000253, 0.00000657, 0.00000725, -0.00001113, 0.00001206, -0.00000895, 0.00000792, -0.00000558, 0.00001800, -0.00001003, -0.00000223, -0.00000208, 0.00000266, -0.00000527, 0.00000093, -0.00000772, 0.00000212, -0.00000129, 0.00000400, -0.00001336, 0.00001659, -0.00003736, 0.00000740, -0.00000488, 0.00003201, 0.00000988, 0.00000756, 0.00003182, -0.00000834, 0.00000985, -0.00002255, 0.00000307, -0.00000699, -0.00000445, -0.00000510, 0.00000037, -0.00000753, 0.00000709, 0.00000342, -0.00000327, -0.00000427, -0.00000172, -0.00000499, 0.00000058, -0.00000550, 0.00000019, -0.00000782, 0.00000342, 0.00000659, -0.00000765, -0.00000040, -0.00000926, -0.00000458, -0.00000945, 0.00000876, -0.00002965, -0.00000428, -0.00007965, -0.00012332, 0.00024918, 0.00028822, -0.00008837, -0.00010312, 0.00000456, -0.00003398, 0.00000318, -0.00001237, 0.00000216, -0.00000796, 0.00000618, -0.00001535, 0.00000329, -0.00001003, 0.00000260, 0.00000653, -0.00000604, -0.00000218, 0.00000393, -0.00000306, 0.00000398, 0.00000783, -0.00000413, -0.00000324, 0.00000165, -0.00000515, 0.00000753, 0.00000473, -0.00000481, 0.00000720, -0.00001462, 0.00000962, -0.00000766, 0.00000071, -0.00000194, -0.00000686, 0.00000290, -0.00002139, 0.00002089, -0.00001897, 0.00001367, -0.00000712, 0.00000021, 0.00000203, -0.00000240, -0.00000038, 0.00000737, 0.00000428, -0.00000018, -0.00000446, 0.00000199, 0.00000685, -0.00000113, -0.00001422, 0.00000471, },
		{ 0.00000037, 0.00000037, -0.00000576, 0.00000166, -0.00001256, 0.00001422, -0.00000952, 0.00000522, -0.00000753, 0.00000566, -0.00001024, 0.00000960, 0.00000596, -0.00001356, -0.00000184, 0.00001175, -0.00000170, -0.00000307, 0.00000633, -0.00000561, 0.00000510, -0.00000030, 0.00001496, -0.00001422, 0.00000141, -0.00000245, 0.00000233, -0.00000552, 0.00000058, -0.00000666, 0.00000163, 0.00000158, -0.00000295, -0.00000426, 0.00000716, -0.00003711, 0.00001782, -0.00000945, 0.00004342, -0.00000963, 0.00002735, 0.00001291, 0.00000003, 0.00000132, -0.00001799, 0.00000294, -0.00000675, -0.00000616, -0.00000130, -0.00000348, -0.00000386, 0.00000755, 0.00000231, -0.00000594, -0.00000138, -0.00000402, -0.00000270, -0.00000149, -0.00000371, -0.00000131, -0.00000764, 0.00000806, 0.00000137, -0.00000617, -0.00000198, -0.00000912, -0.00000562, -0.00000522, 0.00000403, -0.00003337, -0.00000096, -0.00012094, -0.00006701, 0.00034704, 0.00019692, -0.00015580, -0.00004836, -0.00001114, -0.00002260, -0.00000081, -0.00000779, -0.00000159, -0.00000322, 0.00000043, -0.00001233, 0.00000309, -0.00001114, 0.00000761, 0.00000300, -0.00000739, 0.00000088, 0.00000208, -0.00000264, 0.00000647, 0.00000585, -0.00000701, 0.00000026, -0.00000083, -0.00000290, 0.00000873, 0.00000132, -0.00000271, 0.00000517, -0.00001506, 0.00001413, -0.00001498, 0.00000960, -0.00001300, 0.00000540, -0.00001195, -0.00000634, 0.00001212, -0.00001607, 0.00001611, -0.00001281, 0.00000710, -0.00000338, -0.00000023, 0.00000024, 0.00000872, 0.00000237, -0.00000153, -0.00000386, 0.00000422, 0.00000632, -0.00000703, -0.00001030, 0.00000745, },
		{ 0.00000034, -0.00000295, -0.00000118, -0.00000389, -0.00000685, 0.00001264, -0.00001091, 0.00000659, -0.00000890, 0.00000773, -0.00001384, 0.00001873, -0.00000501, -0.00000953, -0.00000082, 0.00001458, -0.00000919, 0.00000568, -0.00000194, 0.00000085, -0.00000028, 0.00000854, 0.00000542, -0.00001189, 0.00000137, -0.00000039, -0.00000064, -0.00000326, -0.00000159, -0.00000413, 0.00000024, 0.00000434, -0.00001026, 0.00000811, -0.00001103, -0.00002258, 0.00001465, -0.00000189, 0.00004067, -0.00001697, 0.00003964, -0.00000414, 0.00001012, -0.00000925, -0.00001068, 0.00000083, -0.00000639, -0.00000683, 0.00000154, -0.00000729, 0.00000113, 0.00000650, 0.00000084, -0.00000751, 0.00000138, -0.00000674, 0.00000014, -0.00000432, -0.00000065, -0.00000504, -0.00000383, 0.00000929, -0.00000291, -0.00000422, -0.00000435, -0.00000820, -0.00000665, -0.00000030, -0.00000497, -0.00003050, -0.00000862, -0.00015804, 0.00002686, 0.00041366, 0.00005951, -0.00016269, -0.00001152, -0.00002593, -0.00001025, -0.00000611, -0.00000284, -0.00000541, 0.00000129, -0.00000620, -0.00000664, -0.00000040, -0.00000893, 0.00001081, -0.00000172, -0.00000648, 0.00000273, 0.00000026, -0.00000155, 0.00000901, 0.00000150, -0.00000672, 0.00000181, -0.00000219, -0.00000103, 0.00000978, -0.00000302, 0.00000178, -0.00000055, -0.00001013, 0.00001235, -0.00001612, 0.00001254, -0.00001872, 0.00001354, -0.00002501, 0.00001158, -0.00000411, -0.00000330, 0.00000839, -0.00000971, 0.00000778, -0.00000585, 0.00000114, 0.00000184, 0.00000857, 0.00000098, -0.00000334, -0.00000152, 0.00000515, 0.00000487, -0.00001270, -0.00000278, 0.00000607, },
		{ 0.00000146, -0.00000606, 0.00000387, -0.00001144, 0.00000308, 0.00000523, -0.00000669, 0.00000281, -0.00000473, 0.00000349, -0.00001010, 0.00002109, -0.00001289, -0.00000604, 0.00000245, 0.00001246, -0.00001121, 0.00001002, -0.00000746, 0.00000561, -0.00000332, 0.00001531, -0.00000550, -0.00000616, -0.00000054, 0.00000225, -0.00000428, -0.00000023, -0.00000396, -0.00000198, 0.00000023, 0.00000356, -0.00001226, 0.00001536, -0.00002809, -0.00000213, 0.00000295, 0.00001469, 0.00002477, -0.00000959, 0.00003890, -0.00001256, 0.00001609, -0.00001767, -0.00000339, -0.00000186, -0.00000664, -0.00000545, 0.00000176, -0.00000895, 0.00000560, 0.00000477, -0.00000130, -0.00000712, 0.00000261, -0.00000862, 0.00000242, -0.00000674, 0.00000232, -0.00000965, 0.00000311, 0.00000583, -0.00000371, -0.00000456, -0.00000491, -0.00000895, -0.00000477, 0.00000099, -0.00001317, -0.00002552, -0.00002669, -0.00018390, 0.00016162, 0.00041318, -0.00008307, -0.00011801, -0.00000511, -0.00002710, -0.00000608, -0.00000602, -0.00000274, -0.00000481, 0.00000118, -0.00000899, -0.00000270, -0.00000380, -0.00000518, 0.00001233, -0.00000729, -0.00000281, 0.00000209, -0.00000017, -0.00000050, 0.00001123, -0.00000458, -0.00000306, 0.00000045, -0.00000129, -0.00000059, 0.00001124, -0.00000821, 0.00000823, -0.00000949, 0.00000001, 0.00000362, -0.00000936, 0.00000663, -0.00001467, 0.00001114, -0.00002760, 0.00002202, -0.00001769, 0.00001117, -0.00000410, -0.00000035, 0.00000229, -0.00000412, 0.00000077, 0.00000457, 0.00000695, -0.00000011, -0.00000472, 0.00000156, 0.00000509, 0.00000184, -0.00001572, 0.00000548, 0.00000181, },
		{ 0.00000202, -0.00000678, 0.00000599, -0.00001667, 0.00001281, -0.00000434, 0.00000034, -0.00000365, 0.00000262, -0.00000512, 0.00000021, 0.00001553, -0.00001542, -0.00000437, 0.00000780, 0.00000568, -0.00000699, 0.00000787, -0.00000736, 0.00000578, -0.00000089, 0.00001546, -0.00001242, -0.00000118, -0.00000162, 0.00000310, -0.00000624, 0.00000168, -0.00000521, -0.00000116, 0.00000222, -0.00000162, -0.00000636, 0.00001163, -0.00003439, 0.00001312, -0.00000728, 0.00003107, 0.00000390, 0.00000735, 0.00002652, -0.00001070, 0.00001516, -0.00002137, 0.00000204, -0.00000449, -0.00000711, -0.00000255, -0.00000072, -0.00000750, 0.00000819, 0.00000304, -0.00000405, -0.00000450, 0.00000134, -0.00000836, 0.00000278, -0.00000714, 0.00000300, -0.00001213, 0.00001000, -0.00000021, -0.00000174, -0.00000706, -0.00000368, -0.00001104, -0.00000035, -0.00000212, -0.00001854, -0.00002072, -0.00005688, -0.00018314, 0.00032273, 0.00032280, -0.00018072, -0.00005544, -0.00001914, -0.00001668, -0.00000953, -0.00000101, -0.00000686, -0.00000044, -0.00000330, -0.00000722, -0.00000202, -0.00000558, -0.00000065, 0.00001125, -0.00001161, 0.00000178, -0.00000005, 0.00000031, 0.00000122, 0.00001132, -0.00000962, 0.00000153, -0.00000207, 0.00000041, -0.00000020, 0.00001138, -0.00001173, 0.00001356, -0.00001807, 0.00001152, -0.00000842, 0.00000209, -0.00000549, -0.00000245, -0.00000190, -0.00001679, 0.00001875, -0.00002004, 0.00001814, -0.00001298, 0.00000873, -0.00000530, 0.00000002, -0.00000049, 0.00000752, 0.00000460, -0.00000144, -0.00000458, 0.00000395, 0.00000476, -0.00000329, -0.00001385, 0.00001078, -0.00000249, },
		{ 0.00000056, -0.00000375, 0.00000272, -0.00001568, 0.00001739, -0.00001104, 0.00000562, -0.00000829, 0.00000829, -0.00001294, 0.00001235, 0.00000501, -0.00001369, -0.00000336, 0.00001265, -0.00000235, 0.00000035, 0.00000142, -0.00000285, 0.00000242, 0.00000596, 0.00000860, -0.00001307, 0.00000072, -0.00000058, 0.00000127, -0.00000558, 0.00000180, -0.00000533, -0.00000082, 0.00000425, -0.00000810, 0.00000430, -0.00000203, -0.00002679, 0.00001637, -0.00000799, 0.00003808, -0.00001174, 0.00002459, 0.00000903, -0.00000166, 0.00000801, -0.00001990, 0.00000478, -0.00000693, -0.00000690, 0.00000035, -0.00000460, -0.00000336, 0.00000843, 0.00000160, -0.00000686, -0.00000021, -0.00000258, -0.00000527, 0.00000032, -0.00000438, -0.00000043, -0.00000936, 0.00001240, -0.00000438, -0.00000071, -0.00000855, -0.00000373, -0.00001091, 0.00000228, -0.00000573, -0.00002388, -0.00001451, -0.00010394, -0.00013112, 0.00046990, 0.00015324, -0.00020095, -0.00001253, -0.00003180, -0.00000699, -0.00001218, 0.00000233, -0.00000950, 0.00000253, -0.00000745, -0.00000435, -0.00000264, -0.00000653, 0.00000511, 0.00000607, -0.00001175, 0.00000409, -0.00000114, -0.00000002, 0.00000495, 0.00000718, -0.00001034, 0.00000350, -0.00000263, 0.00000010, 0.00000279, 0.00000729, -0.00000993, 0.00001326, -0.00002098, 0.00001831, -0.00001724, 0.00001136, -0.00001655, 0.00001044, -0.00001844, 0.00000197, 0.00000397, -0.00000977, 0.00001344, -0.00001251, 0.00001144, -0.00000993, 0.00000340, -0.00000088, 0.00000935, 0.00000236, -0.00000321, -0.00000249, 0.00000476, 0.00000432, -0.00000972, -0.00000670, 0.00001065, -0.00000422, },
		{ -0.00000265, 0.00000205, -0.00000546, -0.00000759, 0.00001435, -0.00001161, 0.00000588, -0.00000767, 0.00000812, -0.00001475, 0.00002039, -0.00000550, -0.00001061, -0.00000087, 0.00001381, -0.00000721, 0.00000587, -0.00000463, 0.00000202, -0.00000046, 0.00001264, -0.00000146, -0.00000911, 0.00000016, 0.00000163, -0.00000209, -0.00000313, 0.00000078, -0.00000530, 0.00000069, 0.00000326, -0.00001066, 0.00001246, -0.00001794, -0.00001052, 0.00000905, 0.00000184, 0.00003228, -0.00001570, 0.00003415, -0.00000607, 0.00000924, -0.00000251, -0.00001424, 0.00000459, -0.00000882, -0.00000547, 0.00000179, -0.00000798, 0.00000208, 0.00000666, 0.00000035, -0.00000878, 0.00000425, -0.00000796, -0.00000009, -0.00000439, 0.00000094, -0.00000756, -0.00000063, 0.00000782, -0.00000313, -0.00000408, -0.00000576, -0.00000819, -0.00000501, -0.00000148, -0.00000501, -0.00003296, -0.00000587, -0.00016824, -0.00000518, 0.00054547, -0.00004599, -0.00014842, -0.00000775, -0.00002745, -0.00000847, -0.00000657, -0.00000183, -0.00000563, -0.00000061, -0.00000663, -0.00000441, -0.00000181, -0.00000781, 0.00001191, -0.00000312, -0.00000655, 0.00000218, 0.00000053, -0.00000203, 0.00001065, -0.00000123, -0.00000558, 0.00000108, 0.00000034, -0.00000352, 0.00000939, -0.00000205, -0.00000113, 0.00000482, -0.00001473, 0.00001579, -0.00001739, 0.00001218, -0.00001912, 0.00001514, -0.00002816, 0.00001785, -0.00001280, 0.00000635, 0.00000059, -0.00000328, 0.00000612, -0.00000892, 0.00000395, 0.00000070, 0.00000931, 0.00000048, -0.00000481, 0.00000076, 0.00000426, 0.00000294, -0.00001500, 0.00000334, 0.00000544, -0.00000280, },
		{ -0.00000543, 0.00000729, -0.00001476, 0.00000450, 0.00000534, -0.00000656, 0.00000134, -0.00000190, 0.00000162, -0.00000866, 0.00002079, -0.00001219, -0.00000836, 0.00000397, 0.00001008, -0.00000658, 0.00000625, -0.00000654, 0.00000377, 0.00000073, 0.00001430, -0.00000943, -0.00000428, -0.00000062, 0.00000295, -0.00000490, -0.00000044, -0.00000050, -0.00000560, 0.00000381, -0.00000188, -0.00000613, 0.00001177, -0.00002676, 0.00000466, -0.00000103, 0.00001624, 0.00001743, -0.00000820, 0.00003292, -0.00001378, 0.00001715, -0.00001293, -0.00000622, 0.00000163, -0.00000949, -0.00000320, 0.00000112, -0.00000925, 0.00000699, 0.00000391, -0.00000121, -0.00000859, 0.00000670, -0.00001231, 0.00000479, -0.00000889, 0.00000588, -0.00001482, 0.00001059, -0.00000163, 0.00000269, -0.00001142, 0.00000077, -0.00001605, 0.00000545, -0.00001114, -0.00000033, -0.00004391, -0.00000215, -0.00023381, 0.00019481, 0.00050025, -0.00020446, -0.00006349, -0.00003073, -0.00000830, -0.00001981, 0.00000597, -0.00001212, 0.00000353, -0.00000902, -0.00000084, -0.00000787, 0.00000068, -0.00000854, 0.00001721, -0.00001304, 0.00000156, -0.00000249, 0.00000385, -0.00000412, 0.00001557, -0.00001058, 0.00000199, -0.00000380, 0.00000489, -0.00000814, 0.00001677, -0.00001343, 0.00001168, -0.00000941, -0.00000085, 0.00000448, -0.00000851, 0.00000342, -0.00001090, 0.00000749, -0.00002430, 0.00002144, -0.00002079, 0.00001793, -0.00001170, 0.00000842, -0.00000374, -0.00000343, 0.00000191, 0.00000392, 0.00000763, -0.00000129, -0.00000515, 0.00000362, 0.00000345, -0.00000056, -0.00001601, 0.00001189, -0.00000173, 0.00000002, },
		{ -0.00000511, 0.00000804, -0.00001991, 0.00001496, -0.00000480, 0.00000029, -0.00000441, 0.00000521, -0.00000751, 0.00000264, 0.00001418, -0.00001431, -0.00000685, 0.00000949, 0.00000349, -0.00000188, 0.00000202, -0.00000409, 0.00000220, 0.00000617, 0.00000934, -0.00001209, -0.00000140, -0.00000002, 0.00000214, -0.00000583, 0.00000156, -0.00000203, -0.00000527, 0.00000654, -0.00000852, 0.00000351, 0.00000163, -0.00002390, 0.00001112, -0.00000565, 0.00002689, 0.00000133, 0.00000540, 0.00002286, -0.00001287, 0.00001914, -0.00002012, 0.00000181, -0.00000325, -0.00000849, -0.00000092, -0.00000141, -0.00000754, 0.00000968, 0.00000149, -0.00000365, -0.00000534, 0.00000506, -0.00001277, 0.00000623, -0.00000979, 0.00000629, -0.00001672, 0.00001770, -0.00000952, 0.00000746, -0.00001735, 0.00000551, -0.00002104, 0.00001315, -0.00001911, 0.00000030, -0.00004671, -0.00002155, -0.00026237, 0.00042920, 0.00032492, -0.00026533, 0.00000123, -0.00005246, 0.00000862, -0.00002862, 0.00001539, -0.00001990, 0.00001034, -0.00001594, 0.00000454, -0.00001084, 0.00000185, -0.00000573, 0.00001673, -0.00001822, 0.00000731, -0.00000571, 0.00000535, -0.00000291, 0.00001540, -0.00001552, 0.00000711, -0.00000649, 0.00000646, -0.00000879, 0.00001920, -0.00002023, 0.00002141, -0.00002233, 0.00001392, -0.00000952, 0.00000378, -0.00000951, 0.00000310, -0.00000853, -0.00000851, 0.00001105, -0.00001504, 0.00001775, -0.00001534, 0.00001492, -0.00001209, 0.00000285, -0.00000070, 0.00000735, 0.00000511, -0.00000324, -0.00000352, 0.00000485, 0.00000302, -0.00000632, -0.00001094, 0.00001495, -0.00000689, 0.00000157, },
		{ -0.00000075, 0.00000241, -0.00001732, 0.00001883, -0.00001101, 0.00000446, -0.00000705, 0.00000876, -0.00001366, 0.00001359, 0.00000445, -0.00001375, -0.00000435, 0.00001261, -0.00000195, 0.00000276, -0.00000311, -0.00000030, 0.00000039, 0.00001221, 0.00000043, -0.00000994, -0.00000074, 0.00000165, -0.00000030, -0.00000504, 0.00000294, -0.00000446, -0.00000256, 0.00000561, -0.00001146, 0.00001170, -0.00001188, -0.00001242, 0.00000817, -0.00000173, 0.00002865, -0.00000915, 0.00001839, 0.00000868, -0.00000510, 0.00001441, -0.00002181, 0.00000722, -0.00000825, -0.00000617, 0.00000069, -0.00000494, -0.00000292, 0.00000918, 0.00000039, -0.00000704, 0.00000081, -0.00000130, -0.00000783, 0.00000233, -0.00000507, -0.00000026, -0.00000940, 0.00001494, -0.00000919, 0.00000454, -0.00001528, 0.00000172, -0.00001563, 0.00000911, -0.00001560, -0.00001338, -0.00002991, -0.00008367, -0.00020402, 0.00062065, 0.00006684, -0.00021694, 0.00001240, -0.00004796, 0.00000642, -0.00002222, 0.00001134, -0.00001626, 0.00000684, -0.00001400, 0.00000317, -0.00000851, -0.00000148, 0.00000274, 0.00000781, -0.00001465, 0.00000632, -0.00000370, 0.00000224, 0.00000372, 0.00000756, -0.00001227, 0.00000565, -0.00000329, 0.00000155, -0.00000171, 0.00001204, -0.00001647, 0.00002096, -0.00002610, 0.00002135, -0.00001802, 0.00001129, -0.00001791, 0.00001339, -0.00002292, 0.00000970, -0.00000602, 0.00000041, 0.00000646, -0.00000819, 0.00001213, -0.00001433, 0.00000638, -0.00000169, 0.00000955, 0.00000239, -0.00000498, -0.00000039, 0.00000437, 0.00000245, -0.00001255, -0.00000096, 0.00001120, -0.00000772, 0.00000040, },
		{ 0.00000562, -0.00000774, -0.00000739, 0.00001470, -0.00001100, 0.00000386, -0.00000457, 0.00000610, -0.00001291, 0.00001925, -0.00000414, -0.00001269, 0.00000039, 0.00001131, -0.00000317, 0.00000365, -0.00000529, 0.00000135, 0.00000192, 0.00001408, -0.00000748, -0.00000611, -0.00000062, 0.00000276, -0.00000275, -0.00000347, 0.00000392, -0.00000782, 0.00000291, -0.00000053, -0.00000691, 0.00001182, -0.00002006, -0.00000058, 0.00000150, 0.00000742, 0.00002220, -0.00001147, 0.00002618, -0.00000471, 0.00000612, 0.00000414, -0.00001723, 0.00000789, -0.00001122, -0.00000364, 0.00000134, -0.00000823, 0.00000338, 0.00000585, 0.00000046, -0.00001022, 0.00000777, -0.00001032, 0.00000093, -0.00000553, 0.00000358, -0.00001173, 0.00000557, 0.00000228, 0.00000073, -0.00000779, -0.00000347, -0.00001208, 0.00000167, -0.00000895, 0.00000189, -0.00004297, 0.00000568, -0.00018957, -0.00002528, 0.00068243, -0.00018281, -0.00010106, -0.00002471, -0.00001537, -0.00001640, 0.00000039, -0.00000689, -0.00000057, -0.00000764, -0.00000205, -0.00000626, -0.00000016, -0.00000873, 0.00001478, -0.00000721, -0.00000348, -0.00000119, 0.00000327, -0.00000448, 0.00001377, -0.00000577, -0.00000208, -0.00000178, 0.00000491, -0.00000853, 0.00001170, -0.00000381, -0.00000173, 0.00000834, -0.00001737, 0.00001707, -0.00001620, 0.00000889, -0.00001580, 0.00001249, -0.00002614, 0.00001914, -0.00001824, 0.00001526, -0.00000810, 0.00000490, 0.00000186, -0.00000989, 0.00000591, -0.00000021, 0.00000985, -0.00000035, -0.00000561, 0.00000281, 0.00000317, 0.00000037, -0.00001594, 0.00000989, 0.00000289, -0.00000485, -0.00000243, },
		{ 0.00000990, -0.00001746, 0.00000548, 0.00000541, -0.00000636, -0.00000003, 0.00000123, -0.00000125, -0.00000550, 0.00001809, -0.00000955, -0.00001164, 0.00000639, 0.00000649, -0.00000041, 0.00000032, -0.00000352, 0.00000005, 0.00000757, 0.00000951, -0.00001073, -0.00000352, 0.00000048, 0.00000211, -0.00000404, -0.00000162, 0.00000380, -0.00001034, 0.00000865, -0.00000929, 0.00000372, 0.00000244, -0.00001789, 0.00000441, -0.00000176, 0.00001550, 0.00001218, -0.00000721, 0.00002717, -0.00001359, 0.00001667, -0.00000857, -0.00000778, 0.00000350, -0.00001088, -0.00000189, 0.00000087, -0.00000973, 0.00000907, 0.00000155, 0.00000030, -0.00001090, 0.00001177, -0.00001745, 0.00000886, -0.00001259, 0.00001064, -0.00002125, 0.00002044, -0.00001296, 0.00001367, -0.00002310, 0.00001134, -0.00002820, 0.00002185, -0.00003126, 0.00002213, -0.00007374, 0.00003674, -0.00030415, 0.00025864, 0.00056322, -0.00033005, 0.00000984, -0.00007222, 0.00002265, -0.00004294, 0.00002541, -0.00002760, 0.00001699, -0.00002408, 0.00001239, -0.00001766, 0.00000894, -0.00001441, 0.00002362, -0.00002078, 0.00000851, -0.00000966, 0.00001025, -0.00000967, 0.00002107, -0.00001770, 0.00000865, -0.00000949, 0.00001222, -0.00001718, 0.00002413, -0.00002082, 0.00001713, -0.00001089, -0.00000022, 0.00000385, -0.00000610, -0.00000163, -0.00000446, 0.00000038, -0.00001583, 0.00001437, -0.00001751, 0.00001995, -0.00001628, 0.00001551, -0.00000985, -0.00000195, 0.00000277, 0.00000307, 0.00000844, -0.00000308, -0.00000436, 0.00000461, 0.00000230, -0.00000412, -0.00001362, 0.00001661, -0.00000550, -0.00000115, -0.00000409, },
		{ 0.00000846, -0.00002146, 0.00001538, -0.00000396, -0.00000103, -0.00000352, 0.00000614, -0.00000869, 0.00000458, 0.00001218, -0.00001241, -0.00000929, 0.00001080, 0.00000164, 0.00000299, -0.00000424, -0.00000025, -0.00000155, 0.00001414, 0.00000058, -0.00000922, -0.00000274, 0.00000238, 0.00000013, -0.00000434, 0.00000093, 0.00000114, -0.00000918, 0.00001023, -0.00001463, 0.00001368, -0.00001078, -0.00000789, 0.00000159, 0.00000132, 0.00001827, 0.00000327, 0.00000050, 0.00002183, -0.00001561, 0.00002237, -0.00001919, 0.00000276, -0.00000378, -0.00000785, -0.00000098, -0.00000104, -0.00000795, 0.00001154, -0.00000117, -0.00000197, -0.00000682, 0.00000918, -0.00001780, 0.00001056, -0.00001319, 0.00000973, -0.00002074, 0.00002508, -0.00001970, 0.00001840, -0.00002981, 0.00001696, -0.00003301, 0.00002904, -0.00003979, 0.00002428, -0.00007941, 0.00002297, -0.00035761, 0.00056989, 0.00028363, -0.00032649, 0.00005435, -0.00008720, 0.00003623, -0.00004992, 0.00003359, -0.00003431, 0.00002195, -0.00002938, 0.00001779, -0.00002146, 0.00001098, -0.00001166, 0.00002178, -0.00002438, 0.00001298, -0.00001179, 0.00001093, -0.00000744, 0.00001910, -0.00002065, 0.00001233, -0.00001045, 0.00001156, -0.00001651, 0.00002641, -0.00002858, 0.00002924, -0.00002628, 0.00001588, -0.00001001, 0.00000466, -0.00001254, 0.00000806, -0.00001501, 0.00000132, -0.00000049, -0.00000459, 0.00001189, -0.00001302, 0.00001721, -0.00001678, 0.00000505, -0.00000052, 0.00000667, 0.00000578, -0.00000537, -0.00000146, 0.00000459, 0.00000172, -0.00001008, -0.00000529, 0.00001615, -0.00001007, 0.00000058, -0.00000218, },
		{ 0.00000073, -0.00001728, 0.00001828, -0.00000926, 0.00000167, -0.00000364, 0.00000658, -0.00001166, 0.00001243, 0.00000510, -0.00001434, -0.00000428, 0.00001107, 0.00000018, 0.00000318, -0.00000611, 0.00000093, 0.00000036, 0.00001658, -0.00000776, -0.00000588, -0.00000224, 0.00000346, -0.00000163, -0.00000449, 0.00000447, -0.00000463, -0.00000280, 0.00000453, -0.00001117, 0.00001513, -0.00001865, 0.00000176, -0.00000356, 0.00000780, 0.00001618, -0.00000273, 0.00000923, 0.00001152, -0.00000993, 0.00002013, -0.00002321, 0.00000960, -0.00001014, -0.00000432, -0.00000017, -0.00000450, -0.00000240, 0.00000934, -0.00000076, -0.00000692, 0.00000209, -0.00000081, -0.00000952, 0.00000365, -0.00000505, -0.00000145, -0.00000683, 0.00001409, -0.00001093, 0.00000719, -0.00001966, 0.00000473, -0.00001703, 0.00001186, -0.00002136, -0.00000692, -0.00004104, -0.00006828, -0.00027442, 0.00078869, -0.00006053, -0.00019664, 0.00001752, -0.00005229, 0.00001180, -0.00002626, 0.00001546, -0.00001878, 0.00000701, -0.00001654, 0.00000764, -0.00001227, 0.00000223, 0.00000173, 0.00000708, -0.00001469, 0.00000612, -0.00000427, 0.00000294, 0.00000375, 0.00000613, -0.00001186, 0.00000578, -0.00000198, 0.00000047, -0.00000343, 0.00001401, -0.00002060, 0.00002649, -0.00002893, 0.00002204, -0.00001626, 0.00000830, -0.00001569, 0.00001252, -0.00002310, 0.00001391, -0.00001456, 0.00001161, -0.00000306, -0.00000077, 0.00000929, -0.00001611, 0.00000818, -0.00000198, 0.00000933, 0.00000225, -0.00000633, 0.00000178, 0.00000354, 0.00000023, -0.00001462, 0.00000602, 0.00000917, -0.00000967, -0.00000029, 0.00000315, },
		{ -0.00001007, -0.00000675, 0.00001410, -0.00000953, 0.00000119, -0.00000037, 0.00000222, -0.00000861, 0.00001532, -0.00000084, -0.00001575, 0.00000277, 0.00000731, 0.00000268, -0.00000094, -0.00000372, -0.00000131, 0.00000721, 0.00001183, -0.00001121, -0.00000389, -0.00000036, 0.00000243, -0.00000211, -0.00000462, 0.00000786, -0.00001155, 0.00000665, -0.00000676, 0.00000107, 0.00000520, -0.00001515, 0.00000346, -0.00000397, 0.00001194, 0.00001279, -0.00000642, 0.00001742, -0.00000146, 0.00000198, 0.00000966, -0.00001855, 0.00000949, -0.00001229, -0.00000243, 0.00000104, -0.00000856, 0.00000542, 0.00000350, 0.00000191, -0.00001245, 0.00001234, -0.00001431, 0.00000389, -0.00000850, 0.00000783, -0.00001778, 0.00001491, -0.00000787, 0.00000973, -0.00001674, 0.00000414, -0.00002153, 0.00001508, -0.00002482, 0.00001855, -0.00006427, 0.00003092, -0.00022948, -0.00002021, 0.00080648, -0.00033706, -0.00002546, -0.00006262, 0.00001183, -0.00003577, 0.00001637, -0.00001940, 0.00001081, -0.00002057, 0.00000844, -0.00001329, 0.00000566, -0.00001255, 0.00001974, -0.00001404, 0.00000290, -0.00000775, 0.00000883, -0.00000922, 0.00001848, -0.00001207, 0.00000371, -0.00000667, 0.00001122, -0.00001572, 0.00001662, -0.00000849, 0.00000048, 0.00000925, -0.00001736, 0.00001575, -0.00001235, 0.00000263, -0.00000866, 0.00000531, -0.00001818, 0.00001377, -0.00001772, 0.00002022, -0.00001482, 0.00001254, -0.00000345, -0.00000941, 0.00000721, -0.00000103, 0.00001037, -0.00000176, -0.00000524, 0.00000388, 0.00000256, -0.00000341, -0.00001442, 0.00001528, -0.00000050, -0.00000620, -0.00000190, 0.00000887, },
		{ -0.00001869, 0.00000524, 0.00000612, -0.00000677, -0.00000041, 0.00000349, -0.00000389, -0.00000176, 0.00001386, -0.00000564, -0.00001523, 0.00000893, 0.00000267, 0.00000618, -0.00000668, 0.00000067, -0.00000448, 0.00001607, 0.00000145, -0.00000916, -0.00000408, 0.00000270, -0.00000015, -0.00000183, -0.00000349, 0.00000834, -0.00001527, 0.00001345, -0.00001685, 0.00001517, -0.00001074, -0.00000252, -0.00000367, 0.00000289, 0.00001052, 0.00001081, -0.00000841, 0.00002317, -0.00001331, 0.00001574, -0.00000543, -0.00000733, 0.00000290, -0.00000976, -0.00000253, 0.00000187, -0.00001087, 0.00001200, -0.00000247, 0.00000353, -0.00001425, 0.00001775, -0.00002386, 0.00001457, -0.00001785, 0.00001647, -0.00002834, 0.00003152, -0.00002694, 0.00002821, -0.00003882, 0.00002619, -0.00004461, 0.00004332, -0.00005808, 0.00005302, -0.00011398, 0.00008914, -0.00039536, 0.00035965, 0.00058625, -0.00044153, 0.00008916, -0.00012206, 0.00006095, -0.00007216, 0.00004981, -0.00004718, 0.00003375, -0.00004217, 0.00002908, -0.00003089, 0.00002021, -0.00002220, 0.00003068, -0.00002927, 0.00001683, -0.00001837, 0.00001813, -0.00001636, 0.00002687, -0.00002488, 0.00001584, -0.00001552, 0.00001934, -0.00002615, 0.00003178, -0.00002897, 0.00002348, -0.00001324, 0.00000134, 0.00000242, -0.00000304, -0.00000727, 0.00000318, -0.00000857, -0.00000376, 0.00000176, -0.00000758, 0.00001578, -0.00001595, 0.00001891, -0.00001424, -0.00000080, 0.00000388, 0.00000176, 0.00000946, -0.00000542, -0.00000231, 0.00000424, 0.00000191, -0.00000890, -0.00000809, 0.00001831, -0.00000822, -0.00000278, -0.00000151, 0.00001094, },
		{ -0.00002091, 0.00001366, -0.00000147, -0.00000385, -0.00000057, 0.00000467, -0.00000767, 0.00000491, 0.00001057, -0.00001054, -0.00001095, 0.00001094, 0.00000112, 0.00000638, -0.00000955, 0.00000283, -0.00000394, 0.00002106, -0.00000911, -0.00000475, -0.00000460, 0.00000486, -0.00000221, -0.00000236, 0.00000039, 0.00000352, -0.00001185, 0.00001184, -0.00001778, 0.00002109, -0.00002186, 0.00000971, -0.00001280, 0.00001267, 0.00000573, 0.00000942, -0.00000741, 0.00002358, -0.00001911, 0.00002497, -0.00001848, 0.00000458, -0.00000588, -0.00000519, -0.00000290, 0.00000059, -0.00000876, 0.00001345, -0.00000449, 0.00000065, -0.00000857, 0.00001305, -0.00002259, 0.00001492, -0.00001661, 0.00001258, -0.00002310, 0.00003035, -0.00002843, 0.00002855, -0.00004187, 0.00002808, -0.00004410, 0.00004380, -0.00005993, 0.00004841, -0.00011294, 0.00006927, -0.00045971, 0.00073876, 0.00019464, -0.00035183, 0.00009167, -0.00011406, 0.00005908, -0.00006791, 0.00004914, -0.00004643, 0.00003121, -0.00004051, 0.00002959, -0.00003137, 0.00001987, -0.00001701, 0.00002502, -0.00002842, 0.00001704, -0.00001667, 0.00001562, -0.00001114, 0.00002125, -0.00002377, 0.00001593, -0.00001277, 0.00001435, -0.00002163, 0.00003109, -0.00003477, 0.00003514, -0.00002842, 0.00001619, -0.00000875, 0.00000345, -0.00001295, 0.00001043, -0.00001884, 0.00000969, -0.00001271, 0.00000888, 0.00000175, -0.00000626, 0.00001517, -0.00001846, 0.00000589, 0.00000036, 0.00000542, 0.00000646, -0.00000741, 0.00000108, 0.00000354, 0.00000053, -0.00001380, 0.00000244, 0.00001416, -0.00001136, -0.00000129, 0.00000209, 0.00000701, },
		{ -0.00001593, 0.00001607, -0.00000616, -0.00000221, 0.00000139, 0.00000208, -0.00000696, 0.00000860, 0.00000735, -0.00001573, -0.00000301, 0.00000791, 0.00000421, 0.00000120, -0.00000693, 0.00000021, 0.00000316, 0.00001756, -0.00001420, -0.00000212, -0.00000299, 0.00000409, -0.00000203, -0.00000443, 0.00000669, -0.00000607, -0.00000130, 0.00000071, -0.00000641, 0.00001276, -0.00001926, 0.00001139, -0.00001447, 0.00001750, 0.00000314, 0.00000541, -0.00000129, 0.00001622, -0.00001510, 0.00002431, -0.00002312, 0.00001076, -0.00001140, -0.00000236, -0.00000148, -0.00000377, -0.00000139, 0.00000829, -0.00000099, -0.00000721, 0.00000424, -0.00000186, -0.00000935, 0.00000325, -0.00000345, -0.00000459, -0.00000100, 0.00000863, -0.00000779, 0.00000500, -0.00001920, 0.00000273, -0.00001236, 0.00000717, -0.00001876, -0.00000987, -0.00004103, -0.00006741, -0.00032756, 0.00095688, -0.00021869, -0.00014017, -0.00000145, -0.00004020, 0.00000506, -0.00002082, 0.00001177, -0.00001467, 0.00000102, -0.00001305, 0.00000696, -0.00001196, 0.00000305, 0.00000305, 0.00000323, -0.00001106, 0.00000255, -0.00000191, 0.00000124, 0.00000567, 0.00000245, -0.00000865, 0.00000345, 0.00000155, -0.00000342, -0.00000180, 0.00001234, -0.00002115, 0.00002854, -0.00002841, 0.00001962, -0.00001142, 0.00000194, -0.00000926, 0.00000690, -0.00001751, 0.00001221, -0.00001831, 0.00002031, -0.00001222, 0.00000766, 0.00000417, -0.00001566, 0.00000886, -0.00000186, 0.00000889, 0.00000165, -0.00000672, 0.00000327, 0.00000296, -0.00000297, -0.00001474, 0.00001267, 0.00000548, -0.00001033, -0.00000122, 0.00000738, -0.00000181, },
		{ -0.00000648, 0.00001329, -0.00000797, -0.00000112, 0.00000372, -0.00000230, -0.00000292, 0.00000938, 0.00000389, -0.00001926, 0.00000562, 0.00000261, 0.00000945, -0.00000714, -0.00000049, -0.00000532, 0.00001477, 0.00000584, -0.00001191, -0.00000305, 0.00000116, 0.00000051, -0.00000005, -0.00000649, 0.00001219, -0.00001582, 0.00001095, -0.00001406, 0.00001184, -0.00000640, -0.00000314, -0.00000030, -0.00000428, 0.00001263, 0.00000655, -0.00000292, 0.00000988, 0.00000193, -0.00000175, 0.00001297, -0.00001723, 0.00000824, -0.00001078, -0.00000308, 0.00000188, -0.00000963, 0.00000859, -0.00000080, 0.00000524, -0.00001589, 0.00001817, -0.00002011, 0.00000911, -0.00001371, 0.00001395, -0.00002548, 0.00002677, -0.00002208, 0.00002369, -0.00003113, 0.00001751, -0.00003688, 0.00003541, -0.00004939, 0.00004572, -0.00009822, 0.00007218, -0.00029272, 0.00002155, 0.00089807, -0.00048964, 0.00006703, -0.00011612, 0.00005135, -0.00006500, 0.00004038, -0.00003861, 0.00002776, -0.00003858, 0.00002419, -0.00002525, 0.00001567, -0.00001933, 0.00002645, -0.00002284, 0.00001187, -0.00001690, 0.00001675, -0.00001583, 0.00002428, -0.00001939, 0.00001105, -0.00001284, 0.00001827, -0.00002389, 0.00002308, -0.00001524, 0.00000500, 0.00000767, -0.00001484, 0.00001219, -0.00000648, -0.00000563, 0.00000124, -0.00000531, -0.00000514, 0.00000183, -0.00001002, 0.00001892, -0.00001706, 0.00001725, -0.00000788, -0.00000862, 0.00000838, -0.00000208, 0.00001104, -0.00000385, -0.00000349, 0.00000354, 0.00000286, -0.00000865, -0.00000969, 0.00001799, -0.00000334, -0.00000761, -0.00000045, 0.00001096, -0.00001156, },
		{ 0.00000321, 0.00000798, -0.00000826, 0.00000090, 0.00000384, -0.00000505, 0.00000122, 0.00000935, -0.00000145, -0.00001821, 0.00001060, -0.00000025, 0.00001179, -0.00001332, 0.00000471, -0.00000831, 0.00002437, -0.00000844, -0.00000531, -0.00000562, 0.00000547, -0.00000324, 0.00000122, -0.00000550, 0.00001230, -0.00001926, 0.00001667, -0.00002232, 0.00002523, -0.00002435, 0.00001609, -0.00001777, 0.00001284, 0.00000093, 0.00001389, -0.00001263, 0.00002178, -0.00001378, 0.00001506, -0.00000383, -0.00000478, -0.00000035, -0.00000577, -0.00000561, 0.00000459, -0.00001292, 0.00001565, -0.00000788, 0.00000831, -0.00001840, 0.00002412, -0.00003085, 0.00002124, -0.00002411, 0.00002278, -0.00003500, 0.00004208, -0.00004138, 0.00004405, -0.00005637, 0.00004316, -0.00006288, 0.00006685, -0.00008805, 0.00008830, -0.00016012, 0.00014978, -0.00050212, 0.00049771, 0.00055851, -0.00052174, 0.00015975, -0.00016998, 0.00009915, -0.00010183, 0.00007475, -0.00006723, 0.00005058, -0.00006018, 0.00004639, -0.00004520, 0.00003270, -0.00003058, 0.00003701, -0.00003681, 0.00002475, -0.00002698, 0.00002603, -0.00002289, 0.00003171, -0.00003075, 0.00002217, -0.00002057, 0.00002474, -0.00003317, 0.00003774, -0.00003589, 0.00002899, -0.00001517, 0.00000274, 0.00000132, -0.00000068, -0.00001173, 0.00000990, -0.00001682, 0.00000902, -0.00001363, 0.00000720, 0.00000583, -0.00001021, 0.00001754, -0.00001562, -0.00000088, 0.00000564, -0.00000011, 0.00001061, -0.00000798, 0.00000061, 0.00000272, 0.00000207, -0.00001429, 0.00000021, 0.00001644, -0.00000898, -0.00000536, 0.00000258, 0.00000995, -0.00001786, },
		{ 0.00000996, 0.00000245, -0.00000785, 0.00000415, 0.00000077, -0.00000409, 0.00000298, 0.00001014, -0.00000933, -0.00001132, 0.00000941, 0.00000251, 0.00000742, -0.00001284, 0.00000413, -0.00000370, 0.00002505, -0.00001773, 0.00000005, -0.00000613, 0.00000658, -0.00000406, -0.00000057, 0.00000043, 0.00000455, -0.00001263, 0.00001053, -0.00001648, 0.00002320, -0.00002794, 0.00002424, -0.00002772, 0.00002547, -0.00000858, 0.00001808, -0.00001712, 0.00002721, -0.00002276, 0.00002648, -0.00001732, 0.00000633, -0.00000853, -0.00000134, -0.00000611, 0.00000313, -0.00000961, 0.00001477, -0.00000761, 0.00000342, -0.00000985, 0.00001575, -0.00002599, 0.00001809, -0.00001898, 0.00001394, -0.00002269, 0.00003172, -0.00003329, 0.00003507, -0.00005043, 0.00003573, -0.00005087, 0.00005326, -0.00007430, 0.00006631, -0.00013940, 0.00010693, -0.00055447, 0.00092197, 0.00006238, -0.00033554, 0.00010436, -0.00012529, 0.00007076, -0.00007730, 0.00005771, -0.00005270, 0.00003501, -0.00004630, 0.00003699, -0.00003792, 0.00002641, -0.00002031, 0.00002529, -0.00002899, 0.00001803, -0.00001892, 0.00001811, -0.00001292, 0.00002095, -0.00002392, 0.00001696, -0.00001266, 0.00001396, -0.00002299, 0.00003182, -0.00003713, 0.00003744, -0.00002744, 0.00001384, -0.00000486, -0.00000070, -0.00000957, 0.00000867, -0.00001786, 0.00001357, -0.00002187, 0.00002170, -0.00000996, 0.00000319, 0.00000961, -0.00001719, 0.00000524, 0.00000189, 0.00000379, 0.00000683, -0.00000879, 0.00000333, 0.00000236, -0.00000122, -0.00001626, 0.00001080, 0.00000969, -0.00001083, -0.00000385, 0.00000726, 0.00000403, -0.00001844, },
		{ 0.00001293, -0.00000251, -0.00000615, 0.00000683, -0.00000366, -0.00000043, 0.00000263, 0.00001096, -0.00001768, -0.00000091, 0.00000363, 0.00000965, -0.00000287, -0.00000566, -0.00000230, 0.00000872, 0.00001447, -0.00001761, 0.00000038, -0.00000248, 0.00000313, -0.00000110, -0.00000507, 0.00000949, -0.00000848, 0.00000143, -0.00000513, 0.00000220, 0.00000453, -0.00001246, 0.00001346, -0.00002062, 0.00002395, -0.00000748, 0.00001256, -0.00001090, 0.00002087, -0.00001917, 0.00002588, -0.00002051, 0.00000943, -0.00001062, -0.00000164, -0.00000212, -0.00000353, 0.00000067, 0.00000539, 0.00000048, -0.00000865, 0.00000781, -0.00000505, -0.00000657, 0.00000026, 0.00000041, -0.00000997, 0.00000814, -0.00000174, 0.00000106, -0.00000333, -0.00001228, -0.00000590, 0.00000003, -0.00000702, -0.00000491, -0.00002619, -0.00002441, -0.00008969, -0.00034820, 0.00110409, -0.00039019, -0.00005533, -0.00004280, -0.00001143, -0.00001473, -0.00000480, -0.00000074, -0.00000320, -0.00001163, -0.00000290, 0.00000024, -0.00000657, 0.00000010, 0.00000702, -0.00000362, -0.00000388, -0.00000444, 0.00000351, -0.00000300, 0.00000947, -0.00000325, -0.00000288, -0.00000109, 0.00000682, -0.00000949, 0.00000273, 0.00000720, -0.00001795, 0.00002668, -0.00002425, 0.00001412, -0.00000380, -0.00000724, 0.00000084, -0.00000307, -0.00000613, 0.00000368, -0.00001509, 0.00002337, -0.00001798, 0.00001445, -0.00000127, -0.00001404, 0.00000892, -0.00000168, 0.00000852, 0.00000033, -0.00000574, 0.00000345, 0.00000321, -0.00000759, -0.00001189, 0.00001728, 0.00000158, -0.00001049, -0.00000169, 0.00001102, -0.00000429, -0.00001406, },
		{ 0.00001297, -0.00000706, -0.00000212, 0.00000639, -0.00000599, 0.00000266, 0.00000275, 0.00000902, -0.00002232, 0.00000795, -0.00000165, 0.00001583, -0.00001364, 0.00000326, -0.00000936, 0.00002304, -0.00000289, -0.00000971, -0.00000358, 0.00000368, -0.00000265, 0.00000309, -0.00000867, 0.00001622, -0.00001961, 0.00001440, -0.00002038, 0.00002278, -0.00002022, 0.00001371, -0.00001122, 0.00000161, 0.00000807, 0.00000503, -0.00000263, 0.00000515, 0.00000399, -0.00000399, 0.00001346, -0.00001288, 0.00000364, -0.00000596, -0.00000639, 0.00000465, -0.00001195, 0.00001298, -0.00000703, 0.00001056, -0.00002060, 0.00002506, -0.00002738, 0.00001630, -0.00002096, 0.00002163, -0.00003407, 0.00003972, -0.00003869, 0.00004101, -0.00004956, 0.00003541, -0.00005674, 0.00006082, -0.00008054, 0.00008120, -0.00014272, 0.00012755, -0.00037887, 0.00010628, 0.00094039, -0.00061999, 0.00016096, -0.00017547, 0.00009664, -0.00009937, 0.00006885, -0.00006167, 0.00004771, -0.00005915, 0.00004299, -0.00004041, 0.00002867, -0.00002819, 0.00003386, -0.00003215, 0.00002187, -0.00002724, 0.00002579, -0.00002323, 0.00003007, -0.00002650, 0.00001866, -0.00001906, 0.00002460, -0.00003133, 0.00002927, -0.00002235, 0.00001045, 0.00000459, -0.00001072, 0.00000746, 0.00000006, -0.00001412, 0.00001183, -0.00001704, 0.00001050, -0.00001461, 0.00000396, 0.00001082, -0.00001347, 0.00001734, -0.00000976, -0.00000863, 0.00000995, -0.00000360, 0.00001193, -0.00000644, -0.00000053, 0.00000179, 0.00000400, -0.00001496, -0.00000177, 0.00001711, -0.00000449, -0.00000977, 0.00000242, 0.00001117, -0.00001140, -0.00000733, },
		{ 0.00001101, -0.00001081, 0.00000399, 0.00000183, -0.00000393, 0.00000236, 0.00000572, 0.00000207, -0.00001980, 0.00001060, -0.00000132, 0.00001522, -0.00001825, 0.00000731, -0.00000992, 0.00003041, -0.00001814, -0.00000039, -0.00000736, 0.00000799, -0.00000633, 0.00000437, -0.00000700, 0.00001493, -0.00002137, 0.00001710, -0.00002375, 0.00003095, -0.00003446, 0.00003339, -0.00003402, 0.00002577, -0.00001180, 0.00002044, -0.00001940, 0.00002278, -0.00001504, 0.00001460, -0.00000336, -0.00000085, -0.00000550, 0.00000052, -0.00001082, 0.00000886, -0.00001564, 0.00001946, -0.00001390, 0.00001391, -0.00002265, 0.00002995, -0.00003726, 0.00002767, -0.00003029, 0.00002854, -0.00003991, 0.00005001, -0.00005356, 0.00005810, -0.00007249, 0.00005897, -0.00007941, 0.00008808, -0.00011582, 0.00012164, -0.00020458, 0.00020926, -0.00061300, 0.00066445, 0.00047731, -0.00055854, 0.00020860, -0.00020565, 0.00012914, -0.00012549, 0.00009498, -0.00008345, 0.00006372, -0.00007450, 0.00006090, -0.00005764, 0.00004405, -0.00003786, 0.00004115, -0.00004168, 0.00003041, -0.00003369, 0.00003232, -0.00002788, 0.00003437, -0.00003399, 0.00002633, -0.00002346, 0.00002713, -0.00003667, 0.00004019, -0.00003963, 0.00003182, -0.00001523, 0.00000277, 0.00000173, -0.00000033, -0.00001333, 0.00001360, -0.00002166, 0.00001907, -0.00002791, 0.00002337, -0.00000765, -0.00000023, 0.00001170, -0.00001365, -0.00000258, 0.00000814, -0.00000244, 0.00001161, -0.00001024, 0.00000372, 0.00000064, 0.00000212, -0.00001912, 0.00000999, 0.00001147, -0.00000763, -0.00000889, 0.00000795, 0.00000700, -0.00001491, -0.00000066, },
		{ 0.00000705, -0.00001211, 0.00000964, -0.00000475, 0.00000154, -0.00000108, 0.00001105, -0.00000894, -0.00001043, 0.00000630, 0.00000577, 0.00000593, -0.00001370, 0.00000332, -0.00000036, 0.00002487, -0.00002357, 0.00000392, -0.00000658, 0.00000686, -0.00000477, 0.00000055, 0.00000129, 0.00000417, -0.00001155, 0.00000642, -0.00001053, 0.00001897, -0.00002669, 0.00003182, -0.00003865, 0.00003595, -0.00002149, 0.00002653, -0.00002648, 0.00003101, -0.00002537, 0.00002600, -0.00001466, 0.00000667, -0.00001025, 0.00000228, -0.00000942, 0.00000570, -0.00000981, 0.00001466, -0.00000948, 0.00000537, -0.00000981, 0.00001631, -0.00002684, 0.00001882, -0.00001920, 0.00001303, -0.00001879, 0.00002800, -0.00003241, 0.00003555, -0.00005276, 0.00003707, -0.00005028, 0.00005371, -0.00007816, 0.00007190, -0.00015089, 0.00012472, -0.00062498, 0.00109950, -0.00009987, -0.00027999, 0.00008936, -0.00011693, 0.00006744, -0.00007470, 0.00005639, -0.00005075, 0.00003136, -0.00004475, 0.00003788, -0.00003909, 0.00002888, -0.00002054, 0.00002202, -0.00002551, 0.00001518, -0.00001772, 0.00001762, -0.00001221, 0.00001786, -0.00002077, 0.00001510, -0.00001002, 0.00001040, -0.00002038, 0.00002811, -0.00003490, 0.00003523, -0.00002267, 0.00000848, 0.00000179, -0.00000782, -0.00000224, 0.00000233, -0.00001107, 0.00001102, -0.00002484, 0.00003009, -0.00001983, 0.00001263, 0.00000241, -0.00001394, 0.00000354, 0.00000371, 0.00000215, 0.00000651, -0.00000895, 0.00000456, 0.00000176, -0.00000418, -0.00001619, 0.00001791, 0.00000419, -0.00000926, -0.00000640, 0.00001242, 0.00000037, -0.00001479, 0.00000500, },
		{ 0.00000073, -0.00000901, 0.00001124, -0.00000908, 0.00000645, -0.00000427, 0.00001509, -0.00001949, 0.00000120, -0.00000080, 0.00001539, -0.00000805, -0.00000311, -0.00000564, 0.00001596, 0.00000776, -0.00001759, 0.00000121, -0.00000065, 0.00000061, 0.00000111, -0.00000624, 0.00001240, -0.00001110, 0.00000427, -0.00001130, 0.00001265, -0.00000770, 0.00000069, 0.00000732, -0.00001979, 0.00002468, -0.00001323, 0.00001634, -0.00001748, 0.00002364, -0.00002083, 0.00002406, -0.00001472, 0.00000476, -0.00000670, -0.00000334, -0.00000097, -0.00000460, 0.00000419, 0.00000029, 0.00000415, -0.00001167, 0.00001299, -0.00001049, -0.00000096, -0.00000561, 0.00000674, -0.00001734, 0.00001980, -0.00001618, 0.00001509, -0.00001761, 0.00000118, -0.00002119, 0.00001994, -0.00003057, 0.00002043, -0.00005678, 0.00001082, -0.00013982, -0.00032428, 0.00120925, -0.00055338, 0.00004391, -0.00009911, 0.00002970, -0.00004485, 0.00001993, -0.00002065, 0.00001431, -0.00002961, 0.00001281, -0.00001188, 0.00000363, -0.00000651, 0.00001319, -0.00001250, 0.00000579, -0.00001394, 0.00001126, -0.00000914, 0.00001446, -0.00001008, 0.00000451, -0.00000694, 0.00001270, -0.00001637, 0.00000877, -0.00000015, -0.00001193, 0.00002153, -0.00001713, 0.00000644, 0.00000543, -0.00001771, 0.00001290, -0.00001555, 0.00000928, -0.00001059, -0.00000467, 0.00001918, -0.00001810, 0.00001730, -0.00000500, -0.00001258, 0.00000905, -0.00000185, 0.00000845, -0.00000173, -0.00000334, 0.00000211, 0.00000443, -0.00001355, -0.00000571, 0.00001855, -0.00000109, -0.00001107, -0.00000095, 0.00001311, -0.00000571, -0.00001261, 0.00000981, },
		{ -0.00000717, -0.00000127, 0.00000674, -0.00000763, 0.00000668, -0.00000313, 0.00001340, -0.00002399, 0.00000880, -0.00000431, 0.00002043, -0.00001899, 0.00000616, -0.00001184, 0.00003000, -0.00001255, -0.00000553, -0.00000493, 0.00000650, -0.00000632, 0.00000665, -0.00001055, 0.00001915, -0.00002191, 0.00001563, -0.00002360, 0.00003083, -0.00003249, 0.00003136, -0.00002626, 0.00001234, -0.00000126, 0.00000821, -0.00000596, 0.00000379, 0.00000408, -0.00000436, 0.00001128, -0.00000595, -0.00000387, 0.00000197, -0.00001242, 0.00000949, -0.00001554, 0.00001827, -0.00001463, 0.00001738, -0.00002607, 0.00003224, -0.00003517, 0.00002449, -0.00002936, 0.00002999, -0.00004226, 0.00005173, -0.00005507, 0.00005889, -0.00006920, 0.00005502, -0.00007799, 0.00008754, -0.00011375, 0.00011984, -0.00019193, 0.00019030, -0.00048107, 0.00023240, 0.00092338, -0.00071044, 0.00024031, -0.00022902, 0.00013901, -0.00013219, 0.00009647, -0.00008421, 0.00006684, -0.00007867, 0.00006149, -0.00005597, 0.00004247, -0.00003755, 0.00004049, -0.00004019, 0.00003094, -0.00003689, 0.00003427, -0.00002996, 0.00003452, -0.00003193, 0.00002508, -0.00002398, 0.00002877, -0.00003625, 0.00003324, -0.00002779, 0.00001496, 0.00000150, -0.00000637, 0.00000300, 0.00000564, -0.00002079, 0.00002064, -0.00002687, 0.00002522, -0.00003189, 0.00002128, -0.00000257, -0.00000443, 0.00001241, -0.00000826, -0.00001012, 0.00001222, -0.00000563, 0.00001288, -0.00000913, 0.00000306, -0.00000090, 0.00000542, -0.00002133, 0.00000835, 0.00001270, -0.00000345, -0.00001296, 0.00000679, 0.00000925, -0.00000921, -0.00000967, 0.00001355, },
		{ -0.00001382, 0.00000822, -0.00000218, -0.00000052, 0.00000125, 0.00000362, 0.00000435, -0.00001966, 0.00000864, -0.00000013, 0.00001583, -0.00002063, 0.00000772, -0.00000837, 0.00003252, -0.00002568, 0.00000423, -0.00000843, 0.00000941, -0.00000858, 0.00000696, -0.00000754, 0.00001583, -0.00002126, 0.00001433, -0.00002027, 0.00003044, -0.00003796, 0.00004494, -0.00004769, 0.00003798, -0.00002477, 0.00002817, -0.00002703, 0.00002494, -0.00001632, 0.00001370, -0.00000302, 0.00000308, -0.00001102, 0.00000771, -0.00001699, 0.00001385, -0.00001835, 0.00002252, -0.00001934, 0.00001922, -0.00002596, 0.00003404, -0.00004169, 0.00003236, -0.00003499, 0.00003256, -0.00004180, 0.00005343, -0.00006083, 0.00006719, -0.00008371, 0.00006999, -0.00009025, 0.00010225, -0.00013550, 0.00014577, -0.00023832, 0.00025563, -0.00071189, 0.00084359, 0.00034961, -0.00054773, 0.00022746, -0.00022142, 0.00014439, -0.00013765, 0.00010593, -0.00009206, 0.00006990, -0.00008196, 0.00006951, -0.00006537, 0.00005193, -0.00004244, 0.00004198, -0.00004269, 0.00003242, -0.00003709, 0.00003568, -0.00003025, 0.00003400, -0.00003377, 0.00002747, -0.00002357, 0.00002595, -0.00003587, 0.00003808, -0.00003889, 0.00003063, -0.00001239, 0.00000061, 0.00000435, -0.00000265, -0.00001111, 0.00001293, -0.00002112, 0.00002349, -0.00003721, 0.00003669, -0.00002115, 0.00001145, 0.00000299, -0.00000903, -0.00000559, 0.00001106, -0.00000487, 0.00001206, -0.00001155, 0.00000618, -0.00000119, 0.00000125, -0.00002195, 0.00001935, 0.00000473, -0.00000481, -0.00001274, 0.00001380, 0.00000281, -0.00001039, -0.00000587, 0.00001472, },
		{ -0.00001544, 0.00001448, -0.00001046, 0.00000798, -0.00000642, 0.00001285, -0.00000900, -0.00000880, 0.00000230, 0.00001019, 0.00000246, -0.00001252, 0.00000090, 0.00000555, 0.00002034, -0.00002570, 0.00000605, -0.00000557, 0.00000542, -0.00000408, 0.00000089, 0.00000279, 0.00000278, -0.00000934, 0.00000068, -0.00000123, 0.00000925, -0.00001794, 0.00003053, -0.00004232, 0.00004081, -0.00002996, 0.00003201, -0.00003317, 0.00003307, -0.00002566, 0.00002266, -0.00000958, 0.00000436, -0.00000953, 0.00000411, -0.00001140, 0.00000720, -0.00000863, 0.00001241, -0.00000920, 0.00000563, -0.00000779, 0.00001408, -0.00002436, 0.00001625, -0.00001648, 0.00000944, -0.00001138, 0.00001903, -0.00002514, 0.00002887, -0.00004736, 0.00003054, -0.00004075, 0.00004318, -0.00006875, 0.00006127, -0.00014176, 0.00011364, -0.00065515, 0.00124851, -0.00027214, -0.00019541, 0.00005015, -0.00008991, 0.00004908, -0.00005966, 0.00004466, -0.00004023, 0.00002007, -0.00003555, 0.00003163, -0.00003405, 0.00002652, -0.00001741, 0.00001550, -0.00001835, 0.00000870, -0.00001313, 0.00001419, -0.00000918, 0.00001239, -0.00001478, 0.00001078, -0.00000551, 0.00000455, -0.00001469, 0.00002067, -0.00002847, 0.00002871, -0.00001445, 0.00000064, 0.00001043, -0.00001688, 0.00000796, -0.00000756, 0.00000081, 0.00000190, -0.00002017, 0.00003134, -0.00002481, 0.00001921, -0.00000413, -0.00001016, 0.00000163, 0.00000524, 0.00000090, 0.00000529, -0.00000760, 0.00000429, 0.00000215, -0.00000862, -0.00001280, 0.00002208, -0.00000064, -0.00000775, -0.00000797, 0.00001641, -0.00000282, -0.00001103, -0.00000018, 0.00001130, },
		{ -0.00000987, 0.00001335, -0.00001265, 0.00001203, -0.00001059, 0.00001863, -0.00002026, 0.00000232, -0.00000421, 0.00002009, -0.00001295, -0.00000053, -0.00000836, 0.00002309, -0.00000115, -0.00001466, 0.00000060, 0.00000205, -0.00000293, 0.00000413, -0.00000749, 0.00001475, -0.00001298, 0.00000588, -0.00001574, 0.00002206, -0.00002062, 0.00001700, -0.00000511, -0.00001218, 0.00001905, -0.00001312, 0.00001552, -0.00001980, 0.00002343, -0.00001945, 0.00001878, -0.00000588, -0.00000326, 0.00000066, -0.00000801, 0.00000255, -0.00000738, 0.00000919, -0.00000682, 0.00000994, -0.00001617, 0.00001944, -0.00001769, 0.00000697, -0.00001392, 0.00001505, -0.00002580, 0.00003238, -0.00003268, 0.00003224, -0.00003594, 0.00001941, -0.00004118, 0.00004496, -0.00006066, 0.00005427, -0.00009860, 0.00006178, -0.00021639, -0.00025046, 0.00125580, -0.00068677, 0.00014051, -0.00015914, 0.00007536, -0.00007951, 0.00004892, -0.00004438, 0.00003472, -0.00004992, 0.00003141, -0.00002724, 0.00001706, -0.00001565, 0.00002041, -0.00002186, 0.00001620, -0.00002432, 0.00001991, -0.00001596, 0.00001944, -0.00001667, 0.00001212, -0.00001281, 0.00001775, -0.00002231, 0.00001439, -0.00000775, -0.00000485, 0.00001454, -0.00000849, -0.00000189, 0.00001450, -0.00002728, 0.00002440, -0.00002764, 0.00002556, -0.00002758, 0.00001091, 0.00000825, -0.00001200, 0.00001502, -0.00000562, -0.00001233, 0.00000981, -0.00000264, 0.00000870, -0.00000430, 0.00000013, -0.00000049, 0.00000629, -0.00002012, 0.00000327, 0.00001604, -0.00000165, -0.00001267, 0.00000143, 0.00001303, -0.00000533, -0.00001232, 0.00000747, 0.00000282, },
		{ 0.00000123, 0.00000463, -0.00000668, 0.00000823, -0.00000719, 0.00001618, -0.00002369, 0.00000757, -0.00000463, 0.00002227, -0.00002206, 0.00000718, -0.00001153, 0.00003372, -0.00002105, -0.00000075, -0.00000625, 0.00000875, -0.00000961, 0.00000991, -0.00001153, 0.00002040, -0.00002209, 0.00001385, -0.00002231, 0.00003352, -0.00003961, 0.00004524, -0.00004092, 0.00002467, -0.00001306, 0.00001450, -0.00001187, 0.00000514, 0.00000255, -0.00000335, 0.00000740, 0.00000223, -0.00001292, 0.00001181, -0.00002025, 0.00001579, -0.00001986, 0.00002363, -0.00002250, 0.00002466, -0.00003132, 0.00003855, -0.00004208, 0.00003217, -0.00003748, 0.00003773, -0.00004854, 0.00006057, -0.00006826, 0.00007389, -0.00008635, 0.00007255, -0.00009651, 0.00011063, -0.00014303, 0.00015447, -0.00023724, 0.00024974, -0.00058625, 0.00038942, 0.00084644, -0.00074999, 0.00029222, -0.00026619, 0.00016999, -0.00015657, 0.00011759, -0.00010159, 0.00008110, -0.00009328, 0.00007604, -0.00006870, 0.00005446, -0.00004555, 0.00004488, -0.00004533, 0.00003725, -0.00004402, 0.00004055, -0.00003467, 0.00003646, -0.00003449, 0.00002907, -0.00002658, 0.00002976, -0.00003740, 0.00003347, -0.00002982, 0.00001682, -0.00000016, -0.00000310, 0.00000009, 0.00000892, -0.00002395, 0.00002554, -0.00003206, 0.00003546, -0.00004574, 0.00003768, -0.00001800, 0.00000795, 0.00000357, -0.00000366, -0.00001308, 0.00001502, -0.00000792, 0.00001353, -0.00001128, 0.00000648, -0.00000373, 0.00000627, -0.00002629, 0.00001888, 0.00000585, -0.00000060, -0.00001680, 0.00001212, 0.00000561, -0.00000525, -0.00001333, 0.00001490, -0.00000808, },
		{ 0.00001238, -0.00000706, 0.00000433, -0.00000169, 0.00000305, 0.00000539, -0.00001802, 0.00000508, 0.00000297, 0.00001374, -0.00002040, 0.00000598, -0.00000366, 0.00002996, -0.00002964, 0.00000740, -0.00000818, 0.00000918, -0.00000944, 0.00000857, -0.00000700, 0.00001511, -0.00001927, 0.00000900, -0.00001241, 0.00002356, -0.00003352, 0.00004778, -0.00005475, 0.00004565, -0.00003464, 0.00003422, -0.00003320, 0.00002647, -0.00001639, 0.00001137, -0.00000159, 0.00000543, -0.00001509, 0.00001398, -0.00002251, 0.00001827, -0.00002006, 0.00002382, -0.00002298, 0.00002303, -0.00002731, 0.00003531, -0.00004289, 0.00003395, -0.00003690, 0.00003386, -0.00003988, 0.00005121, -0.00006141, 0.00006898, -0.00008726, 0.00007331, -0.00009227, 0.00010560, -0.00014228, 0.00015446, -0.00025310, 0.00027716, -0.00078097, 0.00101337, 0.00019113, -0.00049391, 0.00021479, -0.00021427, 0.00014175, -0.00013535, 0.00010501, -0.00009092, 0.00006735, -0.00008080, 0.00007030, -0.00006647, 0.00005468, -0.00004329, 0.00003904, -0.00003946, 0.00003023, -0.00003652, 0.00003551, -0.00002962, 0.00003048, -0.00003000, 0.00002549, -0.00002102, 0.00002155, -0.00003105, 0.00003150, -0.00003349, 0.00002508, -0.00000643, -0.00000378, 0.00000907, -0.00000736, -0.00000529, 0.00000790, -0.00001476, 0.00002091, -0.00003888, 0.00004351, -0.00003096, 0.00002164, -0.00000618, -0.00000325, -0.00000906, 0.00001376, -0.00000690, 0.00001156, -0.00001144, 0.00000730, -0.00000211, -0.00000114, -0.00002158, 0.00002631, -0.00000197, -0.00000169, -0.00001590, 0.00001893, -0.00000148, -0.00000537, -0.00001141, 0.00001817, -0.00001621, },
		{ 0.00001723, -0.00001474, 0.00001361, -0.00001166, 0.00001460, -0.00000877, -0.00000734, -0.00000150, 0.00001454, -0.00000182, -0.00001029, -0.00000195, 0.00001258, 0.00001243, -0.00002416, 0.00000630, -0.00000326, 0.00000252, -0.00000216, 0.00000065, 0.00000444, 0.00000119, -0.00000719, -0.00000482, 0.00000895, -0.00000338, -0.00000383, 0.00002107, -0.00003773, 0.00003835, -0.00003208, 0.00003260, -0.00003543, 0.00003197, -0.00002276, 0.00001610, -0.00000174, -0.00000118, -0.00000548, 0.00000305, -0.00001093, 0.00000677, -0.00000560, 0.00000772, -0.00000633, 0.00000378, -0.00000362, 0.00000902, -0.00001847, 0.00001020, -0.00001070, 0.00000344, -0.00000132, 0.00000590, -0.00001237, 0.00001561, -0.00003457, 0.00001651, -0.00002289, 0.00002230, -0.00004641, 0.00003418, -0.00011075, 0.00006974, -0.00063363, 0.00134792, -0.00043155, -0.00009755, -0.00000430, -0.00004985, 0.00001946, -0.00003489, 0.00002462, -0.00002299, 0.00000297, -0.00002026, 0.00001933, -0.00002343, 0.00001968, -0.00001148, 0.00000679, -0.00000878, -0.00000025, -0.00000609, 0.00000862, -0.00000467, 0.00000559, -0.00000706, 0.00000504, -0.00000035, -0.00000203, -0.00000760, 0.00001119, -0.00001932, 0.00001910, -0.00000403, -0.00000833, 0.00001953, -0.00002593, 0.00001887, -0.00001862, 0.00001535, -0.00001178, -0.00000870, 0.00002471, -0.00002321, 0.00002090, -0.00000800, -0.00000730, 0.00000034, 0.00000597, 0.00000028, 0.00000316, -0.00000484, 0.00000252, 0.00000347, -0.00001418, -0.00000609, 0.00002234, -0.00000348, -0.00000726, -0.00000779, 0.00001826, -0.00000447, -0.00000820, -0.00000455, 0.00001422, -0.00001673, },
		{ 0.00001257, -0.00001342, 0.00001506, -0.00001503, 0.00002056, -0.00001924, 0.00000167, -0.00000561, 0.00002255, -0.00001630, 0.00000089, -0.00000907, 0.00002809, -0.00001024, -0.00001002, -0.00000072, 0.00000482, -0.00000667, 0.00000730, -0.00000824, 0.00001590, -0.00001335, 0.00000521, -0.00001668, 0.00002763, -0.00003051, 0.00003211, -0.00001991, -0.00000027, 0.00000852, -0.00000793, 0.00001040, -0.00001792, 0.00002018, -0.00001532, 0.00001085, 0.00000495, -0.00001361, 0.00001069, -0.00001519, 0.00000825, -0.00001167, 0.00001514, -0.00001513, 0.00001712, -0.00002145, 0.00002623, -0.00002548, 0.00001597, -0.00002349, 0.00002424, -0.00003396, 0.00004375, -0.00004842, 0.00004940, -0.00005506, 0.00003913, -0.00006230, 0.00007081, -0.00009221, 0.00009070, -0.00014489, 0.00012059, -0.00031102, -0.00013032, 0.00123554, -0.00077348, 0.00021829, -0.00021074, 0.00011632, -0.00011144, 0.00007634, -0.00006713, 0.00005385, -0.00006864, 0.00004923, -0.00004270, 0.00003123, -0.00002552, 0.00002715, -0.00002994, 0.00002537, -0.00003362, 0.00002772, -0.00002200, 0.00002311, -0.00002162, 0.00001850, -0.00001742, 0.00002070, -0.00002573, 0.00001773, -0.00001358, 0.00000131, 0.00000748, -0.00000002, -0.00000918, 0.00002164, -0.00003380, 0.00003275, -0.00003619, 0.00003893, -0.00004311, 0.00002780, -0.00000685, -0.00000099, 0.00000795, -0.00000279, -0.00001371, 0.00001135, -0.00000403, 0.00000909, -0.00000687, 0.00000394, -0.00000364, 0.00000799, -0.00002597, 0.00001358, 0.00001040, -0.00000003, -0.00001530, 0.00000529, 0.00001078, -0.00000299, -0.00001345, 0.00000621, 0.00000356, -0.00000849, },
		{ 0.00000077, -0.00000350, 0.00000726, -0.00000904, 0.00001706, -0.00002135, 0.00000429, -0.00000256, 0.00002106, -0.00002240, 0.00000597, -0.00000795, 0.00003300, -0.00002665, 0.00000330, -0.00000671, 0.00000967, -0.00001174, 0.00001226, -0.00001133, 0.00001984, -0.00002016, 0.00000919, -0.00001633, 0.00002981, -0.00003931, 0.00005162, -0.00005073, 0.00003461, -0.00002405, 0.00002121, -0.00001829, 0.00000764, 0.00000051, -0.00000204, 0.00000330, 0.00000983, -0.00002151, 0.00002164, -0.00002824, 0.00002227, -0.00002388, 0.00002792, -0.00002925, 0.00003105, -0.00003514, 0.00004265, -0.00004656, 0.00003767, -0.00004367, 0.00004343, -0.00005156, 0.00006440, -0.00007567, 0.00008283, -0.00009745, 0.00008423, -0.00010818, 0.00012519, -0.00016235, 0.00017762, -0.00026921, 0.00029332, -0.00067732, 0.00055920, 0.00071908, -0.00073660, 0.00030979, -0.00028011, 0.00018354, -0.00016733, 0.00012782, -0.00011013, 0.00008735, -0.00009999, 0.00008366, -0.00007581, 0.00006227, -0.00005054, 0.00004597, -0.00004654, 0.00003958, -0.00004738, 0.00004347, -0.00003645, 0.00003526, -0.00003354, 0.00002998, -0.00002643, 0.00002737, -0.00003448, 0.00002940, -0.00002761, 0.00001507, 0.00000046, -0.00000168, -0.00000065, 0.00000939, -0.00002291, 0.00002555, -0.00003105, 0.00003877, -0.00005259, 0.00004882, -0.00003144, 0.00002048, -0.00000692, 0.00000277, -0.00001679, 0.00001782, -0.00001000, 0.00001344, -0.00001229, 0.00000887, -0.00000585, 0.00000571, -0.00002837, 0.00002771, -0.00000168, 0.00000300, -0.00002033, 0.00001731, 0.00000124, -0.00000035, -0.00001757, 0.00001636, -0.00000904, 0.00000450, },
		{ -0.00001128, 0.00000920, -0.00000539, 0.00000335, 0.00000574, -0.00001562, 0.00000082, 0.00000704, 0.00000980, -0.00001824, 0.00000295, 0.00000318, 0.00002316, -0.00002956, 0.00000860, -0.00000647, 0.00000722, -0.00000878, 0.00000919, -0.00000566, 0.00001336, -0.00001635, 0.00000265, -0.00000209, 0.00001204, -0.00002215, 0.00004139, -0.00005308, 0.00004616, -0.00003880, 0.00003615, -0.00003584, 0.00002569, -0.00001417, 0.00000684, 0.00000185, 0.00000497, -0.00001618, 0.00001764, -0.00002577, 0.00002084, -0.00001990, 0.00002256, -0.00002386, 0.00002444, -0.00002601, 0.00003316, -0.00004016, 0.00003161, -0.00003523, 0.00003197, -0.00003415, 0.00004338, -0.00005496, 0.00006265, -0.00008197, 0.00006766, -0.00008420, 0.00009655, -0.00013387, 0.00014437, -0.00024390, 0.00026550, -0.00080466, 0.00115063, 0.00002348, -0.00040938, 0.00017574, -0.00018651, 0.00012225, -0.00011900, 0.00009237, -0.00008019, 0.00005633, -0.00007115, 0.00006307, -0.00006046, 0.00005174, -0.00004023, 0.00003275, -0.00003260, 0.00002431, -0.00003227, 0.00003204, -0.00002636, 0.00002445, -0.00002336, 0.00002100, -0.00001663, 0.00001511, -0.00002351, 0.00002170, -0.00002443, 0.00001596, 0.00000183, -0.00000957, 0.00001486, -0.00001311, 0.00000266, 0.00000006, -0.00000392, 0.00001208, -0.00003246, 0.00004185, -0.00003437, 0.00002755, -0.00001332, 0.00000197, -0.00001193, 0.00001554, -0.00000809, 0.00000993, -0.00000973, 0.00000678, -0.00000187, -0.00000511, -0.00001752, 0.00002938, -0.00000691, 0.00000048, -0.00001730, 0.00002212, -0.00000459, -0.00000103, -0.00001617, 0.00002036, -0.00001688, 0.00001504, },
		{ -0.00001633, 0.00001697, -0.00001541, 0.00001529, -0.00000744, -0.00000696, -0.00000400, 0.00001755, -0.00000559, -0.00000822, -0.00000379, 0.00001879, 0.00000291, -0.00001984, 0.00000515, -0.00000027, -0.00000120, 0.00000048, 0.00000029, 0.00000566, 0.00000020, -0.00000628, -0.00000816, 0.00001711, -0.00001534, 0.00001183, 0.00000652, -0.00002640, 0.00002905, -0.00002775, 0.00002785, -0.00003274, 0.00002723, -0.00001667, 0.00000677, 0.00000829, -0.00000954, 0.00000182, -0.00000114, -0.00000763, 0.00000411, -0.00000082, 0.00000091, -0.00000110, 0.00000006, 0.00000221, 0.00000185, -0.00000993, 0.00000137, -0.00000250, -0.00000401, 0.00000973, -0.00000920, 0.00000354, -0.00000194, -0.00001659, -0.00000263, 0.00000042, -0.00000553, -0.00001486, -0.00000537, -0.00006193, -0.00000401, -0.00055705, 0.00138284, -0.00055705, -0.00000401, -0.00006193, -0.00000537, -0.00001486, -0.00000553, 0.00000042, -0.00000263, -0.00001659, -0.00000194, 0.00000354, -0.00000920, 0.00000973, -0.00000401, -0.00000250, 0.00000137, -0.00000993, 0.00000185, 0.00000221, 0.00000006, -0.00000110, 0.00000091, -0.00000082, 0.00000411, -0.00000763, -0.00000114, 0.00000182, -0.00000954, 0.00000829, 0.00000677, -0.00001667, 0.00002723, -0.00003274, 0.00002785, -0.00002775, 0.00002905, -0.00002640, 0.00000652, 0.00001183, -0.00001534, 0.00001711, -0.00000816, -0.00000628, 0.00000020, 0.00000566, 0.00000029, 0.00000048, -0.00000120, -0.00000027, 0.00000515, -0.00001984, 0.00000291, 0.00001879, -0.00000379, -0.00000822, -0.00000559, 0.00001755, -0.00000400, -0.00000696, -0.00000744, 0.00001529, -0.00001541, 0.00001697, },
		{ -0.00001128, 0.00001504, -0.00001688, 0.00002036, -0.00001617, -0.00000103, -0.00000459, 0.00002212, -0.00001730, 0.00000048, -0.00000691, 0.00002938, -0.00001752, -0.00000511, -0.00000187, 0.00000678, -0.00000973, 0.00000993, -0.00000809, 0.00001554, -0.00001193, 0.00000197, -0.00001332, 0.00002755, -0.00003437, 0.00004185, -0.00003246, 0.00001208, -0.00000392, 0.00000006, 0.00000266, -0.00001311, 0.00001486, -0.00000957, 0.00000183, 0.00001596, -0.00002443, 0.00002170, -0.00002351, 0.00001511, -0.00001663, 0.00002100, -0.00002336, 0.00002445, -0.00002636, 0.00003204, -0.00003227, 0.00002431, -0.00003260, 0.00003275, -0.00004023, 0.00005174, -0.00006046, 0.00006307, -0.00007115, 0.00005633, -0.00008019, 0.00009237, -0.00011900, 0.00012225, -0.00018651, 0.00017574, -0.00040938, 0.00002348, 0.00115063, -0.00080466, 0.00026550, -0.00024390, 0.00014437, -0.00013387, 0.00009655, -0.00008420, 0.00006766, -0.00008197, 0.00006265, -0.00005496, 0.00004338, -0.00003415, 0.00003197, -0.00003523, 0.00003161, -0.00004016, 0.00003316, -0.00002601, 0.00002444, -0.00002386, 0.00002256, -0.00001990, 0.00002084, -0.00002577, 0.00001764, -0.00001618, 0.00000497, 0.00000185, 0.00000684, -0.00001417, 0.00002569, -0.00003584, 0.00003615, -0.00003880, 0.00004616, -0.00005308, 0.00004139, -0.00002215, 0.00001204, -0.00000209, 0.00000265, -0.00001635, 0.00001336, -0.00000566, 0.00000919, -0.00000878, 0.00000722, -0.00000647, 0.00000860, -0.00002956, 0.00002316, 0.00000318, 0.00000295, -0.00001824, 0.00000980, 0.00000704, 0.00000082, -0.00001562, 0.00000574, 0.00000335, -0.00000539, 0.00000920, },
		{ 0.00000077, 0.00000450, -0.00000904, 0.00001636, -0.00001757, -0.00000035, 0.00000124, 0.00001731, -0.00002033, 0.00000300, -0.00000168, 0.00002771, -0.00002837, 0.00000571, -0.00000585, 0.00000887, -0.00001229, 0.00001344, -0.00001000, 0.00001782, -0.00001679, 0.00000277, -0.00000692, 0.00002048, -0.00003144, 0.00004882, -0.00005259, 0.00003877, -0.00003105, 0.00002555, -0.00002291, 0.00000939, -0.00000065, -0.00000168, 0.00000046, 0.00001507, -0.00002761, 0.00002940, -0.00003448, 0.00002737, -0.00002643, 0.00002998, -0.00003354, 0.00003526, -0.00003645, 0.00004347, -0.00004738, 0.00003958, -0.00004654, 0.00004597, -0.00005054, 0.00006227, -0.00007581, 0.00008366, -0.00009999, 0.00008735, -0.00011013, 0.00012782, -0.00016733, 0.00018354, -0.00028011, 0.00030979, -0.00073660, 0.00071908, 0.00055920, -0.00067732, 0.00029332, -0.00026921, 0.00017762, -0.00016235, 0.00012519, -0.00010818, 0.00008423, -0.00009745, 0.00008283, -0.00007567, 0.00006440, -0.00005156, 0.00004343, -0.00004367, 0.00003767, -0.00004656, 0.00004265, -0.00003514, 0.00003105, -0.00002925, 0.00002792, -0.00002388, 0.00002227, -0.00002824, 0.00002164, -0.00002151, 0.00000983, 0.00000330, -0.00000204, 0.00000051, 0.00000764, -0.00001829, 0.00002121, -0.00002405, 0.00003461, -0.00005073, 0.00005162, -0.00003931, 0.00002981, -0.00001633, 0.00000919, -0.00002016, 0.00001984, -0.00001133, 0.00001226, -0.00001174, 0.00000967, -0.00000671, 0.00000330, -0.00002665, 0.00003300, -0.00000795, 0.00000597, -0.00002240, 0.00002106, -0.00000256, 0.00000429, -0.00002135, 0.00001706, -0.00000904, 0.00000726, -0.00000350, },
		{ 0.00001257, -0.00000849, 0.00000356, 0.00000621, -0.00001345, -0.00000299, 0.00001078, 0.00000529, -0.00001530, -0.00000003, 0.00001040, 0.00001358, -0.00002597, 0.00000799, -0.00000364, 0.00000394, -0.00000687, 0.00000909, -0.00000403, 0.00001135, -0.00001371, -0.00000279, 0.00000795, -0.00000099, -0.00000685, 0.00002780, -0.00004311, 0.00003893, -0.00003619, 0.00003275, -0.00003380, 0.00002164, -0.00000918, -0.00000002, 0.00000748, 0.00000131, -0.00001358, 0.00001773, -0.00002573, 0.00002070, -0.00001742, 0.00001850, -0.00002162, 0.00002311, -0.00002200, 0.00002772, -0.00003362, 0.00002537, -0.00002994, 0.00002715, -0.00002552, 0.00003123, -0.00004270, 0.00004923, -0.00006864, 0.00005385, -0.00006713, 0.00007634, -0.00011144, 0.00011632, -0.00021074, 0.00021829, -0.00077348, 0.00123554, -0.00013032, -0.00031102, 0.00012059, -0.00014489, 0.00009070, -0.00009221, 0.00007081, -0.00006230, 0.00003913, -0.00005506, 0.00004940, -0.00004842, 0.00004375, -0.00003396, 0.00002424, -0.00002349, 0.00001597, -0.00002548, 0.00002623, -0.00002145, 0.00001712, -0.00001513, 0.00001514, -0.00001167, 0.00000825, -0.00001519, 0.00001069, -0.00001361, 0.00000495, 0.00001085, -0.00001532, 0.00002018, -0.00001792, 0.00001040, -0.00000793, 0.00000852, -0.00000027, -0.00001991, 0.00003211, -0.00003051, 0.00002763, -0.00001668, 0.00000521, -0.00001335, 0.00001590, -0.00000824, 0.00000730, -0.00000667, 0.00000482, -0.00000072, -0.00001002, -0.00001024, 0.00002809, -0.00000907, 0.00000089, -0.00001630, 0.00002255, -0.00000561, 0.00000167, -0.00001924, 0.00002056, -0.00001503, 0.00001506, -0.00001342, },
		{ 0.00001723, -0.00001673, 0.00001422, -0.00000455, -0.00000820, -0.00000447, 0.00001826, -0.00000779, -0.00000726, -0.00000348, 0.00002234, -0.00000609, -0.00001418, 0.00000347, 0.00000252, -0.00000484, 0.00000316, 0.00000028, 0.00000597, 0.00000034, -0.00000730, -0.00000800, 0.00002090, -0.00002321, 0.00002471, -0.00000870, -0.00001178, 0.00001535, -0.00001862, 0.00001887, -0.00002593, 0.00001953, -0.00000833, -0.00000403, 0.00001910, -0.00001932, 0.00001119, -0.00000760, -0.00000203, -0.00000035, 0.00000504, -0.00000706, 0.00000559, -0.00000467, 0.00000862, -0.00000609, -0.00000025, -0.00000878, 0.00000679, -0.00001148, 0.00001968, -0.00002343, 0.00001933, -0.00002026, 0.00000297, -0.00002299, 0.00002462, -0.00003489, 0.00001946, -0.00004985, -0.00000430, -0.00009755, -0.00043155, 0.00134792, -0.00063363, 0.00006974, -0.00011075, 0.00003418, -0.00004641, 0.00002230, -0.00002289, 0.00001651, -0.00003457, 0.00001561, -0.00001237, 0.00000590, -0.00000132, 0.00000344, -0.00001070, 0.00001020, -0.00001847, 0.00000902, -0.00000362, 0.00000378, -0.00000633, 0.00000772, -0.00000560, 0.00000677, -0.00001093, 0.00000305, -0.00000548, -0.00000118, -0.00000174, 0.00001610, -0.00002276, 0.00003197, -0.00003543, 0.00003260, -0.00003208, 0.00003835, -0.00003773, 0.00002107, -0.00000383, -0.00000338, 0.00000895, -0.00000482, -0.00000719, 0.00000119, 0.00000444, 0.00000065, -0.00000216, 0.00000252, -0.00000326, 0.00000630, -0.00002416, 0.00001243, 0.00001258, -0.00000195, -0.00001029, -0.00000182, 0.00001454, -0.00000150, -0.00000734, -0.00000877, 0.00001460, -0.00001166, 0.00001361, -0.00001474, },
		{ 0.00001238, -0.00001621, 0.00001817, -0.00001141, -0.00000537, -0.00000148, 0.00001893, -0.00001590, -0.00000169, -0.00000197, 0.00002631, -0.00002158, -0.00000114, -0.00000211, 0.00000730, -0.00001144, 0.00001156, -0.00000690, 0.00001376, -0.00000906, -0.00000325, -0.00000618, 0.00002164, -0.00003096, 0.00004351, -0.00003888, 0.00002091, -0.00001476, 0.00000790, -0.00000529, -0.00000736, 0.00000907, -0.00000378, -0.00000643, 0.00002508, -0.00003349, 0.00003150, -0.00003105, 0.00002155, -0.00002102, 0.00002549, -0.00003000, 0.00003048, -0.00002962, 0.00003551, -0.00003652, 0.00003023, -0.00003946, 0.00003904, -0.00004329, 0.00005468, -0.00006647, 0.00007030, -0.00008080, 0.00006735, -0.00009092, 0.00010501, -0.00013535, 0.00014175, -0.00021427, 0.00021479, -0.00049391, 0.00019113, 0.00101337, -0.00078097, 0.00027716, -0.00025310, 0.00015446, -0.00014228, 0.00010560, -0.00009227, 0.00007331, -0.00008726, 0.00006898, -0.00006141, 0.00005121, -0.00003988, 0.00003386, -0.00003690, 0.00003395, -0.00004289, 0.00003531, -0.00002731, 0.00002303, -0.00002298, 0.00002382, -0.00002006, 0.00001827, -0.00002251, 0.00001398, -0.00001509, 0.00000543, -0.00000159, 0.00001137, -0.00001639, 0.00002647, -0.00003320, 0.00003422, -0.00003464, 0.00004565, -0.00005475, 0.00004778, -0.00003352, 0.00002356, -0.00001241, 0.00000900, -0.00001927, 0.00001511, -0.00000700, 0.00000857, -0.00000944, 0.00000918, -0.00000818, 0.00000740, -0.00002964, 0.00002996, -0.00000366, 0.00000598, -0.00002040, 0.00001374, 0.00000297, 0.00000508, -0.00001802, 0.00000539, 0.00000305, -0.00000169, 0.00000433, -0.00000706, },
		{ 0.00000123, -0.00000808, 0.00001490, -0.00001333, -0.00000525, 0.00000561, 0.00001212, -0.00001680, -0.00000060, 0.00000585, 0.00001888, -0.00002629, 0.00000627, -0.00000373, 0.00000648, -0.00001128, 0.00001353, -0.00000792, 0.00001502, -0.00001308, -0.00000366, 0.00000357, 0.00000795, -0.00001800, 0.00003768, -0.00004574, 0.00003546, -0.00003206, 0.00002554, -0.00002395, 0.00000892, 0.00000009, -0.00000310, -0.00000016, 0.00001682, -0.00002982, 0.00003347, -0.00003740, 0.00002976, -0.00002658, 0.00002907, -0.00003449, 0.00003646, -0.00003467, 0.00004055, -0.00004402, 0.00003725, -0.00004533, 0.00004488, -0.00004555, 0.00005446, -0.00006870, 0.00007604, -0.00009328, 0.00008110, -0.00010159, 0.00011759, -0.00015657, 0.00016999, -0.00026619, 0.00029222, -0.00074999, 0.00084644, 0.00038942, -0.00058625, 0.00024974, -0.00023724, 0.00015447, -0.00014303, 0.00011063, -0.00009651, 0.00007255, -0.00008635, 0.00007389, -0.00006826, 0.00006057, -0.00004854, 0.00003773, -0.00003748, 0.00003217, -0.00004208, 0.00003855, -0.00003132, 0.00002466, -0.00002250, 0.00002363, -0.00001986, 0.00001579, -0.00002025, 0.00001181, -0.00001292, 0.00000223, 0.00000740, -0.00000335, 0.00000255, 0.00000514, -0.00001187, 0.00001450, -0.00001306, 0.00002467, -0.00004092, 0.00004524, -0.00003961, 0.00003352, -0.00002231, 0.00001385, -0.00002209, 0.00002040, -0.00001153, 0.00000991, -0.00000961, 0.00000875, -0.00000625, -0.00000075, -0.00002105, 0.00003372, -0.00001153, 0.00000718, -0.00002206, 0.00002227, -0.00000463, 0.00000757, -0.00002369, 0.00001618, -0.00000719, 0.00000823, -0.00000668, 0.00000463, },
		{ -0.00000987, 0.00000282, 0.00000747, -0.00001232, -0.00000533, 0.00001303, 0.00000143, -0.00001267, -0.00000165, 0.00001604, 0.00000327, -0.00002012, 0.00000629, -0.00000049, 0.00000013, -0.00000430, 0.00000870, -0.00000264, 0.00000981, -0.00001233, -0.00000562, 0.00001502, -0.00001200, 0.00000825, 0.00001091, -0.00002758, 0.00002556, -0.00002764, 0.00002440, -0.00002728, 0.00001450, -0.00000189, -0.00000849, 0.00001454, -0.00000485, -0.00000775, 0.00001439, -0.00002231, 0.00001775, -0.00001281, 0.00001212, -0.00001667, 0.00001944, -0.00001596, 0.00001991, -0.00002432, 0.00001620, -0.00002186, 0.00002041, -0.00001565, 0.00001706, -0.00002724, 0.00003141, -0.00004992, 0.00003472, -0.00004438, 0.00004892, -0.00007951, 0.00007536, -0.00015914, 0.00014051, -0.00068677, 0.00125580, -0.00025046, -0.00021639, 0.00006178, -0.00009860, 0.00005427, -0.00006066, 0.00004496, -0.00004118, 0.00001941, -0.00003594, 0.00003224, -0.00003268, 0.00003238, -0.00002580, 0.00001505, -0.00001392, 0.00000697, -0.00001769, 0.00001944, -0.00001617, 0.00000994, -0.00000682, 0.00000919, -0.00000738, 0.00000255, -0.00000801, 0.00000066, -0.00000326, -0.00000588, 0.00001878, -0.00001945, 0.00002343, -0.00001980, 0.00001552, -0.00001312, 0.00001905, -0.00001218, -0.00000511, 0.00001700, -0.00002062, 0.00002206, -0.00001574, 0.00000588, -0.00001298, 0.00001475, -0.00000749, 0.00000413, -0.00000293, 0.00000205, 0.00000060, -0.00001466, -0.00000115, 0.00002309, -0.00000836, -0.00000053, -0.00001295, 0.00002009, -0.00000421, 0.00000232, -0.00002026, 0.00001863, -0.00001059, 0.00001203, -0.00001265, 0.00001335, },
		{ -0.00001544, 0.00001130, -0.00000018, -0.00001103, -0.00000282, 0.00001641, -0.00000797, -0.00000775, -0.00000064, 0.00002208, -0.00001280, -0.00000862, 0.00000215, 0.00000429, -0.00000760, 0.00000529, 0.00000090, 0.00000524, 0.00000163, -0.00001016, -0.00000413, 0.00001921, -0.00002481, 0.00003134, -0.00002017, 0.00000190, 0.00000081, -0.00000756, 0.00000796, -0.00001688, 0.00001043, 0.00000064, -0.00001445, 0.00002871, -0.00002847, 0.00002067, -0.00001469, 0.00000455, -0.00000551, 0.00001078, -0.00001478, 0.00001239, -0.00000918, 0.00001419, -0.00001313, 0.00000870, -0.00001835, 0.00001550, -0.00001741, 0.00002652, -0.00003405, 0.00003163, -0.00003555, 0.00002007, -0.00004023, 0.00004466, -0.00005966, 0.00004908, -0.00008991, 0.00005015, -0.00019541, -0.00027214, 0.00124851, -0.00065515, 0.00011364, -0.00014176, 0.00006127, -0.00006875, 0.00004318, -0.00004075, 0.00003054, -0.00004736, 0.00002887, -0.00002514, 0.00001903, -0.00001138, 0.00000944, -0.00001648, 0.00001625, -0.00002436, 0.00001408, -0.00000779, 0.00000563, -0.00000920, 0.00001241, -0.00000863, 0.00000720, -0.00001140, 0.00000411, -0.00000953, 0.00000436, -0.00000958, 0.00002266, -0.00002566, 0.00003307, -0.00003317, 0.00003201, -0.00002996, 0.00004081, -0.00004232, 0.00003053, -0.00001794, 0.00000925, -0.00000123, 0.00000068, -0.00000934, 0.00000278, 0.00000279, 0.00000089, -0.00000408, 0.00000542, -0.00000557, 0.00000605, -0.00002570, 0.00002034, 0.00000555, 0.00000090, -0.00001252, 0.00000246, 0.00001019, 0.00000230, -0.00000880, -0.00000900, 0.00001285, -0.00000642, 0.00000798, -0.00001046, 0.00001448, },
		{ -0.00001382, 0.00001472, -0.00000587, -0.00001039, 0.00000281, 0.00001380, -0.00001274, -0.00000481, 0.00000473, 0.00001935, -0.00002195, 0.00000125, -0.00000119, 0.00000618, -0.00001155, 0.00001206, -0.00000487, 0.00001106, -0.00000559, -0.00000903, 0.00000299, 0.00001145, -0.00002115, 0.00003669, -0.00003721, 0.00002349, -0.00002112, 0.00001293, -0.00001111, -0.00000265, 0.00000435, 0.00000061, -0.00001239, 0.00003063, -0.00003889, 0.00003808, -0.00003587, 0.00002595, -0.00002357, 0.00002747, -0.00003377, 0.00003400, -0.00003025, 0.00003568, -0.00003709, 0.00003242, -0.00004269, 0.00004198, -0.00004244, 0.00005193, -0.00006537, 0.00006951, -0.00008196, 0.00006990, -0.00009206, 0.00010593, -0.00013765, 0.00014439, -0.00022142, 0.00022746, -0.00054773, 0.00034961, 0.00084359, -0.00071189, 0.00025563, -0.00023832, 0.00014577, -0.00013550, 0.00010225, -0.00009025, 0.00006999, -0.00008371, 0.00006719, -0.00006083, 0.00005343, -0.00004180, 0.00003256, -0.00003499, 0.00003236, -0.00004169, 0.00003404, -0.00002596, 0.00001922, -0.00001934, 0.00002252, -0.00001835, 0.00001385, -0.00001699, 0.00000771, -0.00001102, 0.00000308, -0.00000302, 0.00001370, -0.00001632, 0.00002494, -0.00002703, 0.00002817, -0.00002477, 0.00003798, -0.00004769, 0.00004494, -0.00003796, 0.00003044, -0.00002027, 0.00001433, -0.00002126, 0.00001583, -0.00000754, 0.00000696, -0.00000858, 0.00000941, -0.00000843, 0.00000423, -0.00002568, 0.00003252, -0.00000837, 0.00000772, -0.00002063, 0.00001583, -0.00000013, 0.00000864, -0.00001966, 0.00000435, 0.00000362, 0.00000125, -0.00000052, -0.00000218, 0.00000822, },
		{ -0.00000717, 0.00001355, -0.00000967, -0.00000921, 0.00000925, 0.00000679, -0.00001296, -0.00000345, 0.00001270, 0.00000835, -0.00002133, 0.00000542, -0.00000090, 0.00000306, -0.00000913, 0.00001288, -0.00000563, 0.00001222, -0.00001012, -0.00000826, 0.00001241, -0.00000443, -0.00000257, 0.00002128, -0.00003189, 0.00002522, -0.00002687, 0.00002064, -0.00002079, 0.00000564, 0.00000300, -0.00000637, 0.00000150, 0.00001496, -0.00002779, 0.00003324, -0.00003625, 0.00002877, -0.00002398, 0.00002508, -0.00003193, 0.00003452, -0.00002996, 0.00003427, -0.00003689, 0.00003094, -0.00004019, 0.00004049, -0.00003755, 0.00004247, -0.00005597, 0.00006149, -0.00007867, 0.00006684, -0.00008421, 0.00009647, -0.00013219, 0.00013901, -0.00022902, 0.00024031, -0.00071044, 0.00092338, 0.00023240, -0.00048107, 0.00019030, -0.00019193, 0.00011984, -0.00011375, 0.00008754, -0.00007799, 0.00005502, -0.00006920, 0.00005889, -0.00005507, 0.00005173, -0.00004226, 0.00002999, -0.00002936, 0.00002449, -0.00003517, 0.00003224, -0.00002607, 0.00001738, -0.00001463, 0.00001827, -0.00001554, 0.00000949, -0.00001242, 0.00000197, -0.00000387, -0.00000595, 0.00001128, -0.00000436, 0.00000408, 0.00000379, -0.00000596, 0.00000821, -0.00000126, 0.00001234, -0.00002626, 0.00003136, -0.00003249, 0.00003083, -0.00002360, 0.00001563, -0.00002191, 0.00001915, -0.00001055, 0.00000665, -0.00000632, 0.00000650, -0.00000493, -0.00000553, -0.00001255, 0.00003000, -0.00001184, 0.00000616, -0.00001899, 0.00002043, -0.00000431, 0.00000880, -0.00002399, 0.00001340, -0.00000313, 0.00000668, -0.00000763, 0.00000674, -0.00000127, },
		{ 0.00000073, 0.00000981, -0.00001261, -0.00000571, 0.00001311, -0.00000095, -0.00001107, -0.00000109, 0.00001855, -0.00000571, -0.00001355, 0.00000443, 0.00000211, -0.00000334, -0.00000173, 0.00000845, -0.00000185, 0.00000905, -0.00001258, -0.00000500, 0.00001730, -0.00001810, 0.00001918, -0.00000467, -0.00001059, 0.00000928, -0.00001555, 0.00001290, -0.00001771, 0.00000543, 0.00000644, -0.00001713, 0.00002153, -0.00001193, -0.00000015, 0.00000877, -0.00001637, 0.00001270, -0.00000694, 0.00000451, -0.00001008, 0.00001446, -0.00000914, 0.00001126, -0.00001394, 0.00000579, -0.00001250, 0.00001319, -0.00000651, 0.00000363, -0.00001188, 0.00001281, -0.00002961, 0.00001431, -0.00002065, 0.00001993, -0.00004485, 0.00002970, -0.00009911, 0.00004391, -0.00055338, 0.00120925, -0.00032428, -0.00013982, 0.00001082, -0.00005678, 0.00002043, -0.00003057, 0.00001994, -0.00002119, 0.00000118, -0.00001761, 0.00001509, -0.00001618, 0.00001980, -0.00001734, 0.00000674, -0.00000561, -0.00000096, -0.00001049, 0.00001299, -0.00001167, 0.00000415, 0.00000029, 0.00000419, -0.00000460, -0.00000097, -0.00000334, -0.00000670, 0.00000476, -0.00001472, 0.00002406, -0.00002083, 0.00002364, -0.00001748, 0.00001634, -0.00001323, 0.00002468, -0.00001979, 0.00000732, 0.00000069, -0.00000770, 0.00001265, -0.00001130, 0.00000427, -0.00001110, 0.00001240, -0.00000624, 0.00000111, 0.00000061, -0.00000065, 0.00000121, -0.00001759, 0.00000776, 0.00001596, -0.00000564, -0.00000311, -0.00000805, 0.00001539, -0.00000080, 0.00000120, -0.00001949, 0.00001509, -0.00000427, 0.00000645, -0.00000908, 0.00001124, -0.00000901, },
		{ 0.00000705, 0.00000500, -0.00001479, 0.00000037, 0.00001242, -0.00000640, -0.00000926, 0.00000419, 0.00001791, -0.00001619, -0.00000418, 0.00000176, 0.00000456, -0.00000895, 0.00000651, 0.00000215, 0.00000371, 0.00000354, -0.00001394, 0.00000241, 0.00001263, -0.00001983, 0.00003009, -0.00002484, 0.00001102, -0.00001107, 0.00000233, -0.00000224, -0.00000782, 0.00000179, 0.00000848, -0.00002267, 0.00003523, -0.00003490, 0.00002811, -0.00002038, 0.00001040, -0.00001002, 0.00001510, -0.00002077, 0.00001786, -0.00001221, 0.00001762, -0.00001772, 0.00001518, -0.00002551, 0.00002202, -0.00002054, 0.00002888, -0.00003909, 0.00003788, -0.00004475, 0.00003136, -0.00005075, 0.00005639, -0.00007470, 0.00006744, -0.00011693, 0.00008936, -0.00027999, -0.00009987, 0.00109950, -0.00062498, 0.00012472, -0.00015089, 0.00007190, -0.00007816, 0.00005371, -0.00005028, 0.00003707, -0.00005276, 0.00003555, -0.00003241, 0.00002800, -0.00001879, 0.00001303, -0.00001920, 0.00001882, -0.00002684, 0.00001631, -0.00000981, 0.00000537, -0.00000948, 0.00001466, -0.00000981, 0.00000570, -0.00000942, 0.00000228, -0.00001025, 0.00000667, -0.00001466, 0.00002600, -0.00002537, 0.00003101, -0.00002648, 0.00002653, -0.00002149, 0.00003595, -0.00003865, 0.00003182, -0.00002669, 0.00001897, -0.00001053, 0.00000642, -0.00001155, 0.00000417, 0.00000129, 0.00000055, -0.00000477, 0.00000686, -0.00000658, 0.00000392, -0.00002357, 0.00002487, -0.00000036, 0.00000332, -0.00001370, 0.00000593, 0.00000577, 0.00000630, -0.00001043, -0.00000894, 0.00001105, -0.00000108, 0.00000154, -0.00000475, 0.00000964, -0.00001211, },
		{ 0.00001101, -0.00000066, -0.00001491, 0.00000700, 0.00000795, -0.00000889, -0.00000763, 0.00001147, 0.00000999, -0.00001912, 0.00000212, 0.00000064, 0.00000372, -0.00001024, 0.00001161, -0.00000244, 0.00000814, -0.00000258, -0.00001365, 0.00001170, -0.00000023, -0.00000765, 0.00002337, -0.00002791, 0.00001907, -0.00002166, 0.00001360, -0.00001333, -0.00000033, 0.00000173, 0.00000277, -0.00001523, 0.00003182, -0.00003963, 0.00004019, -0.00003667, 0.00002713, -0.00002346, 0.00002633, -0.00003399, 0.00003437, -0.00002788, 0.00003232, -0.00003369, 0.00003041, -0.00004168, 0.00004115, -0.00003786, 0.00004405, -0.00005764, 0.00006090, -0.00007450, 0.00006372, -0.00008345, 0.00009498, -0.00012549, 0.00012914, -0.00020565, 0.00020860, -0.00055854, 0.00047731, 0.00066445, -0.00061300, 0.00020926, -0.00020458, 0.00012164, -0.00011582, 0.00008808, -0.00007941, 0.00005897, -0.00007249, 0.00005810, -0.00005356, 0.00005001, -0.00003991, 0.00002854, -0.00003029, 0.00002767, -0.00003726, 0.00002995, -0.00002265, 0.00001391, -0.00001390, 0.00001946, -0.00001564, 0.00000886, -0.00001082, 0.00000052, -0.00000550, -0.00000085, -0.00000336, 0.00001460, -0.00001504, 0.00002278, -0.00001940, 0.00002044, -0.00001180, 0.00002577, -0.00003402, 0.00003339, -0.00003446, 0.00003095, -0.00002375, 0.00001710, -0.00002137, 0.00001493, -0.00000700, 0.00000437, -0.00000633, 0.00000799, -0.00000736, -0.00000039, -0.00001814, 0.00003041, -0.00000992, 0.00000731, -0.00001825, 0.00001522, -0.00000132, 0.00001060, -0.00001980, 0.00000207, 0.00000572, 0.00000236, -0.00000393, 0.00000183, 0.00000399, -0.00001081, },
		{ 0.00001297, -0.00000733, -0.00001140, 0.00001117, 0.00000242, -0.00000977, -0.00000449, 0.00001711, -0.00000177, -0.00001496, 0.00000400, 0.00000179, -0.00000053, -0.00000644, 0.00001193, -0.00000360, 0.00000995, -0.00000863, -0.00000976, 0.00001734, -0.00001347, 0.00001082, 0.00000396, -0.00001461, 0.00001050, -0.00001704, 0.00001183, -0.00001412, 0.00000006, 0.00000746, -0.00001072, 0.00000459, 0.00001045, -0.00002235, 0.00002927, -0.00003133, 0.00002460, -0.00001906, 0.00001866, -0.00002650, 0.00003007, -0.00002323, 0.00002579, -0.00002724, 0.00002187, -0.00003215, 0.00003386, -0.00002819, 0.00002867, -0.00004041, 0.00004299, -0.00005915, 0.00004771, -0.00006167, 0.00006885, -0.00009937, 0.00009664, -0.00017547, 0.00016096, -0.00061999, 0.00094039, 0.00010628, -0.00037887, 0.00012755, -0.00014272, 0.00008120, -0.00008054, 0.00006082, -0.00005674, 0.00003541, -0.00004956, 0.00004101, -0.00003869, 0.00003972, -0.00003407, 0.00002163, -0.00002096, 0.00001630, -0.00002738, 0.00002506, -0.00002060, 0.00001056, -0.00000703, 0.00001298, -0.00001195, 0.00000465, -0.00000639, -0.00000596, 0.00000364, -0.00001288, 0.00001346, -0.00000399, 0.00000399, 0.00000515, -0.00000263, 0.00000503, 0.00000807, 0.00000161, -0.00001122, 0.00001371, -0.00002022, 0.00002278, -0.00002038, 0.00001440, -0.00001961, 0.00001622, -0.00000867, 0.00000309, -0.00000265, 0.00000368, -0.00000358, -0.00000971, -0.00000289, 0.00002304, -0.00000936, 0.00000326, -0.00001364, 0.00001583, -0.00000165, 0.00000795, -0.00002232, 0.00000902, 0.00000275, 0.00000266, -0.00000599, 0.00000639, -0.00000212, -0.00000706, },
		{ 0.00001293, -0.00001406, -0.00000429, 0.00001102, -0.00000169, -0.00001049, 0.00000158, 0.00001728, -0.00001189, -0.00000759, 0.00000321, 0.00000345, -0.00000574, 0.00000033, 0.00000852, -0.00000168, 0.00000892, -0.00001404, -0.00000127, 0.00001445, -0.00001798, 0.00002337, -0.00001509, 0.00000368, -0.00000613, -0.00000307, 0.00000084, -0.00000724, -0.00000380, 0.00001412, -0.00002425, 0.00002668, -0.00001795, 0.00000720, 0.00000273, -0.00000949, 0.00000682, -0.00000109, -0.00000288, -0.00000325, 0.00000947, -0.00000300, 0.00000351, -0.00000444, -0.00000388, -0.00000362, 0.00000702, 0.00000010, -0.00000657, 0.00000024, -0.00000290, -0.00001163, -0.00000320, -0.00000074, -0.00000480, -0.00001473, -0.00001143, -0.00004280, -0.00005533, -0.00039019, 0.00110409, -0.00034820, -0.00008969, -0.00002441, -0.00002619, -0.00000491, -0.00000702, 0.00000003, -0.00000590, -0.00001228, -0.00000333, 0.00000106, -0.00000174, 0.00000814, -0.00000997, 0.00000041, 0.00000026, -0.00000657, -0.00000505, 0.00000781, -0.00000865, 0.00000048, 0.00000539, 0.00000067, -0.00000353, -0.00000212, -0.00000164, -0.00001062, 0.00000943, -0.00002051, 0.00002588, -0.00001917, 0.00002087, -0.00001090, 0.00001256, -0.00000748, 0.00002395, -0.00002062, 0.00001346, -0.00001246, 0.00000453, 0.00000220, -0.00000513, 0.00000143, -0.00000848, 0.00000949, -0.00000507, -0.00000110, 0.00000313, -0.00000248, 0.00000038, -0.00001761, 0.00001447, 0.00000872, -0.00000230, -0.00000566, -0.00000287, 0.00000965, 0.00000363, -0.00000091, -0.00001768, 0.00001096, 0.00000263, -0.00000043, -0.00000366, 0.00000683, -0.00000615, -0.00000251, },
		{ 0.00000996, -0.00001844, 0.00000403, 0.00000726, -0.00000385, -0.00001083, 0.00000969, 0.00001080, -0.00001626, -0.00000122, 0.00000236, 0.00000333, -0.00000879, 0.00000683, 0.00000379, 0.00000189, 0.00000524, -0.00001719, 0.00000961, 0.00000319, -0.00000996, 0.00002170, -0.00002187, 0.00001357, -0.00001786, 0.00000867, -0.00000957, -0.00000070, -0.00000486, 0.00001384, -0.00002744, 0.00003744, -0.00003713, 0.00003182, -0.00002299, 0.00001396, -0.00001266, 0.00001696, -0.00002392, 0.00002095, -0.00001292, 0.00001811, -0.00001892, 0.00001803, -0.00002899, 0.00002529, -0.00002031, 0.00002641, -0.00003792, 0.00003699, -0.00004630, 0.00003501, -0.00005270, 0.00005771, -0.00007730, 0.00007076, -0.00012529, 0.00010436, -0.00033554, 0.00006238, 0.00092197, -0.00055447, 0.00010693, -0.00013940, 0.00006631, -0.00007430, 0.00005326, -0.00005087, 0.00003573, -0.00005043, 0.00003507, -0.00003329, 0.00003172, -0.00002269, 0.00001394, -0.00001898, 0.00001809, -0.00002599, 0.00001575, -0.00000985, 0.00000342, -0.00000761, 0.00001477, -0.00000961, 0.00000313, -0.00000611, -0.00000134, -0.00000853, 0.00000633, -0.00001732, 0.00002648, -0.00002276, 0.00002721, -0.00001712, 0.00001808, -0.00000858, 0.00002547, -0.00002772, 0.00002424, -0.00002794, 0.00002320, -0.00001648, 0.00001053, -0.00001263, 0.00000455, 0.00000043, -0.00000057, -0.00000406, 0.00000658, -0.00000613, 0.00000005, -0.00001773, 0.00002505, -0.00000370, 0.00000413, -0.00001284, 0.00000742, 0.00000251, 0.00000941, -0.00001132, -0.00000933, 0.00001014, 0.00000298, -0.00000409, 0.00000077, 0.00000415, -0.00000785, 0.00000245, },
		{ 0.00000321, -0.00001786, 0.00000995, 0.00000258, -0.00000536, -0.00000898, 0.00001644, 0.00000021, -0.00001429, 0.00000207, 0.00000272, 0.00000061, -0.00000798, 0.00001061, -0.00000011, 0.00000564, -0.00000088, -0.00001562, 0.00001754, -0.00001021, 0.00000583, 0.00000720, -0.00001363, 0.00000902, -0.00001682, 0.00000990, -0.00001173, -0.00000068, 0.00000132, 0.00000274, -0.00001517, 0.00002899, -0.00003589, 0.00003774, -0.00003317, 0.00002474, -0.00002057, 0.00002217, -0.00003075, 0.00003171, -0.00002289, 0.00002603, -0.00002698, 0.00002475, -0.00003681, 0.00003701, -0.00003058, 0.00003270, -0.00004520, 0.00004639, -0.00006018, 0.00005058, -0.00006723, 0.00007475, -0.00010183, 0.00009915, -0.00016998, 0.00015975, -0.00052174, 0.00055851, 0.00049771, -0.00050212, 0.00014978, -0.00016012, 0.00008830, -0.00008805, 0.00006685, -0.00006288, 0.00004316, -0.00005637, 0.00004405, -0.00004138, 0.00004208, -0.00003500, 0.00002278, -0.00002411, 0.00002124, -0.00003085, 0.00002412, -0.00001840, 0.00000831, -0.00000788, 0.00001565, -0.00001292, 0.00000459, -0.00000561, -0.00000577, -0.00000035, -0.00000478, -0.00000383, 0.00001506, -0.00001378, 0.00002178, -0.00001263, 0.00001389, 0.00000093, 0.00001284, -0.00001777, 0.00001609, -0.00002435, 0.00002523, -0.00002232, 0.00001667, -0.00001926, 0.00001230, -0.00000550, 0.00000122, -0.00000324, 0.00000547, -0.00000562, -0.00000531, -0.00000844, 0.00002437, -0.00000831, 0.00000471, -0.00001332, 0.00001179, -0.00000025, 0.00001060, -0.00001821, -0.00000145, 0.00000935, 0.00000122, -0.00000505, 0.00000384, 0.00000090, -0.00000826, 0.00000798, },
		{ -0.00000648, -0.00001156, 0.00001096, -0.00000045, -0.00000761, -0.00000334, 0.00001799, -0.00000969, -0.00000865, 0.00000286, 0.00000354, -0.00000349, -0.00000385, 0.00001104, -0.00000208, 0.00000838, -0.00000862, -0.00000788, 0.00001725, -0.00001706, 0.00001892, -0.00001002, 0.00000183, -0.00000514, -0.00000531, 0.00000124, -0.00000563, -0.00000648, 0.00001219, -0.00001484, 0.00000767, 0.00000500, -0.00001524, 0.00002308, -0.00002389, 0.00001827, -0.00001284, 0.00001105, -0.00001939, 0.00002428, -0.00001583, 0.00001675, -0.00001690, 0.00001187, -0.00002284, 0.00002645, -0.00001933, 0.00001567, -0.00002525, 0.00002419, -0.00003858, 0.00002776, -0.00003861, 0.00004038, -0.00006500, 0.00005135, -0.00011612, 0.00006703, -0.00048964, 0.00089807, 0.00002155, -0.00029272, 0.00007218, -0.00009822, 0.00004572, -0.00004939, 0.00003541, -0.00003688, 0.00001751, -0.00003113, 0.00002369, -0.00002208, 0.00002677, -0.00002548, 0.00001395, -0.00001371, 0.00000911, -0.00002011, 0.00001817, -0.00001589, 0.00000524, -0.00000080, 0.00000859, -0.00000963, 0.00000188, -0.00000308, -0.00001078, 0.00000824, -0.00001723, 0.00001297, -0.00000175, 0.00000193, 0.00000988, -0.00000292, 0.00000655, 0.00001263, -0.00000428, -0.00000030, -0.00000314, -0.00000640, 0.00001184, -0.00001406, 0.00001095, -0.00001582, 0.00001219, -0.00000649, -0.00000005, 0.00000051, 0.00000116, -0.00000305, -0.00001191, 0.00000584, 0.00001477, -0.00000532, -0.00000049, -0.00000714, 0.00000945, 0.00000261, 0.00000562, -0.00001926, 0.00000389, 0.00000938, -0.00000292, -0.00000230, 0.00000372, -0.00000112, -0.00000797, 0.00001329, },
		{ -0.00001593, -0.00000181, 0.00000738, -0.00000122, -0.00001033, 0.00000548, 0.00001267, -0.00001474, -0.00000297, 0.00000296, 0.00000327, -0.00000672, 0.00000165, 0.00000889, -0.00000186, 0.00000886, -0.00001566, 0.00000417, 0.00000766, -0.00001222, 0.00002031, -0.00001831, 0.00001221, -0.00001751, 0.00000690, -0.00000926, 0.00000194, -0.00001142, 0.00001962, -0.00002841, 0.00002854, -0.00002115, 0.00001234, -0.00000180, -0.00000342, 0.00000155, 0.00000345, -0.00000865, 0.00000245, 0.00000567, 0.00000124, -0.00000191, 0.00000255, -0.00001106, 0.00000323, 0.00000305, 0.00000305, -0.00001196, 0.00000696, -0.00001305, 0.00000102, -0.00001467, 0.00001177, -0.00002082, 0.00000506, -0.00004020, -0.00000145, -0.00014017, -0.00021869, 0.00095688, -0.00032756, -0.00006741, -0.00004103, -0.00000987, -0.00001876, 0.00000717, -0.00001236, 0.00000273, -0.00001920, 0.00000500, -0.00000779, 0.00000863, -0.00000100, -0.00000459, -0.00000345, 0.00000325, -0.00000935, -0.00000186, 0.00000424, -0.00000721, -0.00000099, 0.00000829, -0.00000139, -0.00000377, -0.00000148, -0.00000236, -0.00001140, 0.00001076, -0.00002312, 0.00002431, -0.00001510, 0.00001622, -0.00000129, 0.00000541, 0.00000314, 0.00001750, -0.00001447, 0.00001139, -0.00001926, 0.00001276, -0.00000641, 0.00000071, -0.00000130, -0.00000607, 0.00000669, -0.00000443, -0.00000203, 0.00000409, -0.00000299, -0.00000212, -0.00001420, 0.00001756, 0.00000316, 0.00000021, -0.00000693, 0.00000120, 0.00000421, 0.00000791, -0.00000301, -0.00001573, 0.00000735, 0.00000860, -0.00000696, 0.00000208, 0.00000139, -0.00000221, -0.00000616, 0.00001607, },
		{ -0.00002091, 0.00000701, 0.00000209, -0.00000129, -0.00001136, 0.00001416, 0.00000244, -0.00001380, 0.00000053, 0.00000354, 0.00000108, -0.00000741, 0.00000646, 0.00000542, 0.00000036, 0.00000589, -0.00001846, 0.00001517, -0.00000626, 0.00000175, 0.00000888, -0.00001271, 0.00000969, -0.00001884, 0.00001043, -0.00001295, 0.00000345, -0.00000875, 0.00001619, -0.00002842, 0.00003514, -0.00003477, 0.00003109, -0.00002163, 0.00001435, -0.00001277, 0.00001593, -0.00002377, 0.00002125, -0.00001114, 0.00001562, -0.00001667, 0.00001704, -0.00002842, 0.00002502, -0.00001701, 0.00001987, -0.00003137, 0.00002959, -0.00004051, 0.00003121, -0.00004643, 0.00004914, -0.00006791, 0.00005908, -0.00011406, 0.00009167, -0.00035183, 0.00019464, 0.00073876, -0.00045971, 0.00006927, -0.00011294, 0.00004841, -0.00005993, 0.00004380, -0.00004410, 0.00002808, -0.00004187, 0.00002855, -0.00002843, 0.00003035, -0.00002310, 0.00001258, -0.00001661, 0.00001492, -0.00002259, 0.00001305, -0.00000857, 0.00000065, -0.00000449, 0.00001345, -0.00000876, 0.00000059, -0.00000290, -0.00000519, -0.00000588, 0.00000458, -0.00001848, 0.00002497, -0.00001911, 0.00002358, -0.00000741, 0.00000942, 0.00000573, 0.00001267, -0.00001280, 0.00000971, -0.00002186, 0.00002109, -0.00001778, 0.00001184, -0.00001185, 0.00000352, 0.00000039, -0.00000236, -0.00000221, 0.00000486, -0.00000460, -0.00000475, -0.00000911, 0.00002106, -0.00000394, 0.00000283, -0.00000955, 0.00000638, 0.00000112, 0.00001094, -0.00001095, -0.00001054, 0.00001057, 0.00000491, -0.00000767, 0.00000467, -0.00000057, -0.00000385, -0.00000147, 0.00001366, },
		{ -0.00001869, 0.00001094, -0.00000151, -0.00000278, -0.00000822, 0.00001831, -0.00000809, -0.00000890, 0.00000191, 0.00000424, -0.00000231, -0.00000542, 0.00000946, 0.00000176, 0.00000388, -0.00000080, -0.00001424, 0.00001891, -0.00001595, 0.00001578, -0.00000758, 0.00000176, -0.00000376, -0.00000857, 0.00000318, -0.00000727, -0.00000304, 0.00000242, 0.00000134, -0.00001324, 0.00002348, -0.00002897, 0.00003178, -0.00002615, 0.00001934, -0.00001552, 0.00001584, -0.00002488, 0.00002687, -0.00001636, 0.00001813, -0.00001837, 0.00001683, -0.00002927, 0.00003068, -0.00002220, 0.00002021, -0.00003089, 0.00002908, -0.00004217, 0.00003375, -0.00004718, 0.00004981, -0.00007216, 0.00006095, -0.00012206, 0.00008916, -0.00044153, 0.00058625, 0.00035965, -0.00039536, 0.00008914, -0.00011398, 0.00005302, -0.00005808, 0.00004332, -0.00004461, 0.00002619, -0.00003882, 0.00002821, -0.00002694, 0.00003152, -0.00002834, 0.00001647, -0.00001785, 0.00001457, -0.00002386, 0.00001775, -0.00001425, 0.00000353, -0.00000247, 0.00001200, -0.00001087, 0.00000187, -0.00000253, -0.00000976, 0.00000290, -0.00000733, -0.00000543, 0.00001574, -0.00001331, 0.00002317, -0.00000841, 0.00001081, 0.00001052, 0.00000289, -0.00000367, -0.00000252, -0.00001074, 0.00001517, -0.00001685, 0.00001345, -0.00001527, 0.00000834, -0.00000349, -0.00000183, -0.00000015, 0.00000270, -0.00000408, -0.00000916, 0.00000145, 0.00001607, -0.00000448, 0.00000067, -0.00000668, 0.00000618, 0.00000267, 0.00000893, -0.00001523, -0.00000564, 0.00001386, -0.00000176, -0.00000389, 0.00000349, -0.00000041, -0.00000677, 0.00000612, 0.00000524, },
		{ -0.00001007, 0.00000887, -0.00000190, -0.00000620, -0.00000050, 0.00001528, -0.00001442, -0.00000341, 0.00000256, 0.00000388, -0.00000524, -0.00000176, 0.00001037, -0.00000103, 0.00000721, -0.00000941, -0.00000345, 0.00001254, -0.00001482, 0.00002022, -0.00001772, 0.00001377, -0.00001818, 0.00000531, -0.00000866, 0.00000263, -0.00001235, 0.00001575, -0.00001736, 0.00000925, 0.00000048, -0.00000849, 0.00001662, -0.00001572, 0.00001122, -0.00000667, 0.00000371, -0.00001207, 0.00001848, -0.00000922, 0.00000883, -0.00000775, 0.00000290, -0.00001404, 0.00001974, -0.00001255, 0.00000566, -0.00001329, 0.00000844, -0.00002057, 0.00001081, -0.00001940, 0.00001637, -0.00003577, 0.00001183, -0.00006262, -0.00002546, -0.00033706, 0.00080648, -0.00002021, -0.00022948, 0.00003092, -0.00006427, 0.00001855, -0.00002482, 0.00001508, -0.00002153, 0.00000414, -0.00001674, 0.00000973, -0.00000787, 0.00001491, -0.00001778, 0.00000783, -0.00000850, 0.00000389, -0.00001431, 0.00001234, -0.00001245, 0.00000191, 0.00000350, 0.00000542, -0.00000856, 0.00000104, -0.00000243, -0.00001229, 0.00000949, -0.00001855, 0.00000966, 0.00000198, -0.00000146, 0.00001742, -0.00000642, 0.00001279, 0.00001194, -0.00000397, 0.00000346, -0.00001515, 0.00000520, 0.00000107, -0.00000676, 0.00000665, -0.00001155, 0.00000786, -0.00000462, -0.00000211, 0.00000243, -0.00000036, -0.00000389, -0.00001121, 0.00001183, 0.00000721, -0.00000131, -0.00000372, -0.00000094, 0.00000268, 0.00000731, 0.00000277, -0.00001575, -0.00000084, 0.00001532, -0.00000861, 0.00000222, -0.00000037, 0.00000119, -0.00000953, 0.00001410, -0.00000675, },
		{ 0.00000073, 0.00000315, -0.00000029, -0.00000967, 0.00000917, 0.00000602, -0.00001462, 0.00000023, 0.00000354, 0.00000178, -0.00000633, 0.00000225, 0.00000933, -0.00000198, 0.00000818, -0.00001611, 0.00000929, -0.00000077, -0.00000306, 0.00001161, -0.00001456, 0.00001391, -0.00002310, 0.00001252, -0.00001569, 0.00000830, -0.00001626, 0.00002204, -0.00002893, 0.00002649, -0.00002060, 0.00001401, -0.00000343, 0.00000047, -0.00000198, 0.00000578, -0.00001186, 0.00000613, 0.00000375, 0.00000294, -0.00000427, 0.00000612, -0.00001469, 0.00000708, 0.00000173, 0.00000223, -0.00001227, 0.00000764, -0.00001654, 0.00000701, -0.00001878, 0.00001546, -0.00002626, 0.00001180, -0.00005229, 0.00001752, -0.00019664, -0.00006053, 0.00078869, -0.00027442, -0.00006828, -0.00004104, -0.00000692, -0.00002136, 0.00001186, -0.00001703, 0.00000473, -0.00001966, 0.00000719, -0.00001093, 0.00001409, -0.00000683, -0.00000145, -0.00000505, 0.00000365, -0.00000952, -0.00000081, 0.00000209, -0.00000692, -0.00000076, 0.00000934, -0.00000240, -0.00000450, -0.00000017, -0.00000432, -0.00001014, 0.00000960, -0.00002321, 0.00002013, -0.00000993, 0.00001152, 0.00000923, -0.00000273, 0.00001618, 0.00000780, -0.00000356, 0.00000176, -0.00001865, 0.00001513, -0.00001117, 0.00000453, -0.00000280, -0.00000463, 0.00000447, -0.00000449, -0.00000163, 0.00000346, -0.00000224, -0.00000588, -0.00000776, 0.00001658, 0.00000036, 0.00000093, -0.00000611, 0.00000318, 0.00000018, 0.00001107, -0.00000428, -0.00001434, 0.00000510, 0.00001243, -0.00001166, 0.00000658, -0.00000364, 0.00000167, -0.00000926, 0.00001828, -0.00001728, },
		{ 0.00000846, -0.00000218, 0.00000058, -0.00001007, 0.00001615, -0.00000529, -0.00001008, 0.00000172, 0.00000459, -0.00000146, -0.00000537, 0.00000578, 0.00000667, -0.00000052, 0.00000505, -0.00001678, 0.00001721, -0.00001302, 0.00001189, -0.00000459, -0.00000049, 0.00000132, -0.00001501, 0.00000806, -0.00001254, 0.00000466, -0.00001001, 0.00001588, -0.00002628, 0.00002924, -0.00002858, 0.00002641, -0.00001651, 0.00001156, -0.00001045, 0.00001233, -0.00002065, 0.00001910, -0.00000744, 0.00001093, -0.00001179, 0.00001298, -0.00002438, 0.00002178, -0.00001166, 0.00001098, -0.00002146, 0.00001779, -0.00002938, 0.00002195, -0.00003431, 0.00003359, -0.00004992, 0.00003623, -0.00008720, 0.00005435, -0.00032649, 0.00028363, 0.00056989, -0.00035761, 0.00002297, -0.00007941, 0.00002428, -0.00003979, 0.00002904, -0.00003301, 0.00001696, -0.00002981, 0.00001840, -0.00001970, 0.00002508, -0.00002074, 0.00000973, -0.00001319, 0.00001056, -0.00001780, 0.00000918, -0.00000682, -0.00000197, -0.00000117, 0.00001154, -0.00000795, -0.00000104, -0.00000098, -0.00000785, -0.00000378, 0.00000276, -0.00001919, 0.00002237, -0.00001561, 0.00002183, 0.00000050, 0.00000327, 0.00001827, 0.00000132, 0.00000159, -0.00000789, -0.00001078, 0.00001368, -0.00001463, 0.00001023, -0.00000918, 0.00000114, 0.00000093, -0.00000434, 0.00000013, 0.00000238, -0.00000274, -0.00000922, 0.00000058, 0.00001414, -0.00000155, -0.00000025, -0.00000424, 0.00000299, 0.00000164, 0.00001080, -0.00000929, -0.00001241, 0.00001218, 0.00000458, -0.00000869, 0.00000614, -0.00000352, -0.00000103, -0.00000396, 0.00001538, -0.00002146, },
		{ 0.00000990, -0.00000409, -0.00000115, -0.00000550, 0.00001661, -0.00001362, -0.00000412, 0.00000230, 0.00000461, -0.00000436, -0.00000308, 0.00000844, 0.00000307, 0.00000277, -0.00000195, -0.00000985, 0.00001551, -0.00001628, 0.00001995, -0.00001751, 0.00001437, -0.00001583, 0.00000038, -0.00000446, -0.00000163, -0.00000610, 0.00000385, -0.00000022, -0.00001089, 0.00001713, -0.00002082, 0.00002413, -0.00001718, 0.00001222, -0.00000949, 0.00000865, -0.00001770, 0.00002107, -0.00000967, 0.00001025, -0.00000966, 0.00000851, -0.00002078, 0.00002362, -0.00001441, 0.00000894, -0.00001766, 0.00001239, -0.00002408, 0.00001699, -0.00002760, 0.00002541, -0.00004294, 0.00002265, -0.00007222, 0.00000984, -0.00033005, 0.00056322, 0.00025864, -0.00030415, 0.00003674, -0.00007374, 0.00002213, -0.00003126, 0.00002185, -0.00002820, 0.00001134, -0.00002310, 0.00001367, -0.00001296, 0.00002044, -0.00002125, 0.00001064, -0.00001259, 0.00000886, -0.00001745, 0.00001177, -0.00001090, 0.00000030, 0.00000155, 0.00000907, -0.00000973, 0.00000087, -0.00000189, -0.00001088, 0.00000350, -0.00000778, -0.00000857, 0.00001667, -0.00001359, 0.00002717, -0.00000721, 0.00001218, 0.00001550, -0.00000176, 0.00000441, -0.00001789, 0.00000244, 0.00000372, -0.00000929, 0.00000865, -0.00001034, 0.00000380, -0.00000162, -0.00000404, 0.00000211, 0.00000048, -0.00000352, -0.00001073, 0.00000951, 0.00000757, 0.00000005, -0.00000352, 0.00000032, -0.00000041, 0.00000649, 0.00000639, -0.00001164, -0.00000955, 0.00001809, -0.00000550, -0.00000125, 0.00000123, -0.00000003, -0.00000636, 0.00000541, 0.00000548, -0.00001746, },
		{ 0.00000562, -0.00000243, -0.00000485, 0.00000289, 0.00000989, -0.00001594, 0.00000037, 0.00000317, 0.00000281, -0.00000561, -0.00000035, 0.00000985, -0.00000021, 0.00000591, -0.00000989, 0.00000186, 0.00000490, -0.00000810, 0.00001526, -0.00001824, 0.00001914, -0.00002614, 0.00001249, -0.00001580, 0.00000889, -0.00001620, 0.00001707, -0.00001737, 0.00000834, -0.00000173, -0.00000381, 0.00001170, -0.00000853, 0.00000491, -0.00000178, -0.00000208, -0.00000577, 0.00001377, -0.00000448, 0.00000327, -0.00000119, -0.00000348, -0.00000721, 0.00001478, -0.00000873, -0.00000016, -0.00000626, -0.00000205, -0.00000764, -0.00000057, -0.00000689, 0.00000039, -0.00001640, -0.00001537, -0.00002471, -0.00010106, -0.00018281, 0.00068243, -0.00002528, -0.00018957, 0.00000568, -0.00004297, 0.00000189, -0.00000895, 0.00000167, -0.00001208, -0.00000347, -0.00000779, 0.00000073, 0.00000228, 0.00000557, -0.00001173, 0.00000358, -0.00000553, 0.00000093, -0.00001032, 0.00000777, -0.00001022, 0.00000046, 0.00000585, 0.00000338, -0.00000823, 0.00000134, -0.00000364, -0.00001122, 0.00000789, -0.00001723, 0.00000414, 0.00000612, -0.00000471, 0.00002618, -0.00001147, 0.00002220, 0.00000742, 0.00000150, -0.00000058, -0.00002006, 0.00001182, -0.00000691, -0.00000053, 0.00000291, -0.00000782, 0.00000392, -0.00000347, -0.00000275, 0.00000276, -0.00000062, -0.00000611, -0.00000748, 0.00001408, 0.00000192, 0.00000135, -0.00000529, 0.00000365, -0.00000317, 0.00001131, 0.00000039, -0.00001269, -0.00000414, 0.00001925, -0.00001291, 0.00000610, -0.00000457, 0.00000386, -0.00001100, 0.00001470, -0.00000739, -0.00000774, },
		{ -0.00000075, 0.00000040, -0.00000772, 0.00001120, -0.00000096, -0.00001255, 0.00000245, 0.00000437, -0.00000039, -0.00000498, 0.00000239, 0.00000955, -0.00000169, 0.00000638, -0.00001433, 0.00001213, -0.00000819, 0.00000646, 0.00000041, -0.00000602, 0.00000970, -0.00002292, 0.00001339, -0.00001791, 0.00001129, -0.00001802, 0.00002135, -0.00002610, 0.00002096, -0.00001647, 0.00001204, -0.00000171, 0.00000155, -0.00000329, 0.00000565, -0.00001227, 0.00000756, 0.00000372, 0.00000224, -0.00000370, 0.00000632, -0.00001465, 0.00000781, 0.00000274, -0.00000148, -0.00000851, 0.00000317, -0.00001400, 0.00000684, -0.00001626, 0.00001134, -0.00002222, 0.00000642, -0.00004796, 0.00001240, -0.00021694, 0.00006684, 0.00062065, -0.00020402, -0.00008367, -0.00002991, -0.00001338, -0.00001560, 0.00000911, -0.00001563, 0.00000172, -0.00001528, 0.00000454, -0.00000919, 0.00001494, -0.00000940, -0.00000026, -0.00000507, 0.00000233, -0.00000783, -0.00000130, 0.00000081, -0.00000704, 0.00000039, 0.00000918, -0.00000292, -0.00000494, 0.00000069, -0.00000617, -0.00000825, 0.00000722, -0.00002181, 0.00001441, -0.00000510, 0.00000868, 0.00001839, -0.00000915, 0.00002865, -0.00000173, 0.00000817, -0.00001242, -0.00001188, 0.00001170, -0.00001146, 0.00000561, -0.00000256, -0.00000446, 0.00000294, -0.00000504, -0.00000030, 0.00000165, -0.00000074, -0.00000994, 0.00000043, 0.00001221, 0.00000039, -0.00000030, -0.00000311, 0.00000276, -0.00000195, 0.00001261, -0.00000435, -0.00001375, 0.00000445, 0.00001359, -0.00001366, 0.00000876, -0.00000705, 0.00000446, -0.00001101, 0.00001883, -0.00001732, 0.00000241, },
		{ -0.00000511, 0.00000157, -0.00000689, 0.00001495, -0.00001094, -0.00000632, 0.00000302, 0.00000485, -0.00000352, -0.00000324, 0.00000511, 0.00000735, -0.00000070, 0.00000285, -0.00001209, 0.00001492, -0.00001534, 0.00001775, -0.00001504, 0.00001105, -0.00000851, -0.00000853, 0.00000310, -0.00000951, 0.00000378, -0.00000952, 0.00001392, -0.00002233, 0.00002141, -0.00002023, 0.00001920, -0.00000879, 0.00000646, -0.00000649, 0.00000711, -0.00001552, 0.00001540, -0.00000291, 0.00000535, -0.00000571, 0.00000731, -0.00001822, 0.00001673, -0.00000573, 0.00000185, -0.00001084, 0.00000454, -0.00001594, 0.00001034, -0.00001990, 0.00001539, -0.00002862, 0.00000862, -0.00005246, 0.00000123, -0.00026533, 0.00032492, 0.00042920, -0.00026237, -0.00002155, -0.00004671, 0.00000030, -0.00001911, 0.00001315, -0.00002104, 0.00000551, -0.00001735, 0.00000746, -0.00000952, 0.00001770, -0.00001672, 0.00000629, -0.00000979, 0.00000623, -0.00001277, 0.00000506, -0.00000534, -0.00000365, 0.00000149, 0.00000968, -0.00000754, -0.00000141, -0.00000092, -0.00000849, -0.00000325, 0.00000181, -0.00002012, 0.00001914, -0.00001287, 0.00002286, 0.00000540, 0.00000133, 0.00002689, -0.00000565, 0.00001112, -0.00002390, 0.00000163, 0.00000351, -0.00000852, 0.00000654, -0.00000527, -0.00000203, 0.00000156, -0.00000583, 0.00000214, -0.00000002, -0.00000140, -0.00001209, 0.00000934, 0.00000617, 0.00000220, -0.00000409, 0.00000202, -0.00000188, 0.00000349, 0.00000949, -0.00000685, -0.00001431, 0.00001418, 0.00000264, -0.00000751, 0.00000521, -0.00000441, 0.00000029, -0.00000480, 0.00001496, -0.00001991, 0.00000804, },
		{ -0.00000543, 0.00000002, -0.00000173, 0.00001189, -0.00001601, -0.00000056, 0.00000345, 0.00000362, -0.00000515, -0.00000129, 0.00000763, 0.00000392, 0.00000191, -0.00000343, -0.00000374, 0.00000842, -0.00001170, 0.00001793, -0.00002079, 0.00002144, -0.00002430, 0.00000749, -0.00001090, 0.00000342, -0.00000851, 0.00000448, -0.00000085, -0.00000941, 0.00001168, -0.00001343, 0.00001677, -0.00000814, 0.00000489, -0.00000380, 0.00000199, -0.00001058, 0.00001557, -0.00000412, 0.00000385, -0.00000249, 0.00000156, -0.00001304, 0.00001721, -0.00000854, 0.00000068, -0.00000787, -0.00000084, -0.00000902, 0.00000353, -0.00001212, 0.00000597, -0.00001981, -0.00000830, -0.00003073, -0.00006349, -0.00020446, 0.00050025, 0.00019481, -0.00023381, -0.00000215, -0.00004391, -0.00000033, -0.00001114, 0.00000545, -0.00001605, 0.00000077, -0.00001142, 0.00000269, -0.00000163, 0.00001059, -0.00001482, 0.00000588, -0.00000889, 0.00000479, -0.00001231, 0.00000670, -0.00000859, -0.00000121, 0.00000391, 0.00000699, -0.00000925, 0.00000112, -0.00000320, -0.00000949, 0.00000163, -0.00000622, -0.00001293, 0.00001715, -0.00001378, 0.00003292, -0.00000820, 0.00001743, 0.00001624, -0.00000103, 0.00000466, -0.00002676, 0.00001177, -0.00000613, -0.00000188, 0.00000381, -0.00000560, -0.00000050, -0.00000044, -0.00000490, 0.00000295, -0.00000062, -0.00000428, -0.00000943, 0.00001430, 0.00000073, 0.00000377, -0.00000654, 0.00000625, -0.00000658, 0.00001008, 0.00000397, -0.00000836, -0.00001219, 0.00002079, -0.00000866, 0.00000162, -0.00000190, 0.00000134, -0.00000656, 0.00000534, 0.00000450, -0.00001476, 0.00000729, },
		{ -0.00000265, -0.00000280, 0.00000544, 0.00000334, -0.00001500, 0.00000294, 0.00000426, 0.00000076, -0.00000481, 0.00000048, 0.00000931, 0.00000070, 0.00000395, -0.00000892, 0.00000612, -0.00000328, 0.00000059, 0.00000635, -0.00001280, 0.00001785, -0.00002816, 0.00001514, -0.00001912, 0.00001218, -0.00001739, 0.00001579, -0.00001473, 0.00000482, -0.00000113, -0.00000205, 0.00000939, -0.00000352, 0.00000034, 0.00000108, -0.00000558, -0.00000123, 0.00001065, -0.00000203, 0.00000053, 0.00000218, -0.00000655, -0.00000312, 0.00001191, -0.00000781, -0.00000181, -0.00000441, -0.00000663, -0.00000061, -0.00000563, -0.00000183, -0.00000657, -0.00000847, -0.00002745, -0.00000775, -0.00014842, -0.00004599, 0.00054547, -0.00000518, -0.00016824, -0.00000587, -0.00003296, -0.00000501, -0.00000148, -0.00000501, -0.00000819, -0.00000576, -0.00000408, -0.00000313, 0.00000782, -0.00000063, -0.00000756, 0.00000094, -0.00000439, -0.00000009, -0.00000796, 0.00000425, -0.00000878, 0.00000035, 0.00000666, 0.00000208, -0.00000798, 0.00000179, -0.00000547, -0.00000882, 0.00000459, -0.00001424, -0.00000251, 0.00000924, -0.00000607, 0.00003415, -0.00001570, 0.00003228, 0.00000184, 0.00000905, -0.00001052, -0.00001794, 0.00001246, -0.00001066, 0.00000326, 0.00000069, -0.00000530, 0.00000078, -0.00000313, -0.00000209, 0.00000163, 0.00000016, -0.00000911, -0.00000146, 0.00001264, -0.00000046, 0.00000202, -0.00000463, 0.00000587, -0.00000721, 0.00001381, -0.00000087, -0.00001061, -0.00000550, 0.00002039, -0.00001475, 0.00000812, -0.00000767, 0.00000588, -0.00001161, 0.00001435, -0.00000759, -0.00000546, 0.00000205, },
		{ 0.00000056, -0.00000422, 0.00001065, -0.00000670, -0.00000972, 0.00000432, 0.00000476, -0.00000249, -0.00000321, 0.00000236, 0.00000935, -0.00000088, 0.00000340, -0.00000993, 0.00001144, -0.00001251, 0.00001344, -0.00000977, 0.00000397, 0.00000197, -0.00001844, 0.00001044, -0.00001655, 0.00001136, -0.00001724, 0.00001831, -0.00002098, 0.00001326, -0.00000993, 0.00000729, 0.00000279, 0.00000010, -0.00000263, 0.00000350, -0.00001034, 0.00000718, 0.00000495, -0.00000002, -0.00000114, 0.00000409, -0.00001175, 0.00000607, 0.00000511, -0.00000653, -0.00000264, -0.00000435, -0.00000745, 0.00000253, -0.00000950, 0.00000233, -0.00001218, -0.00000699, -0.00003180, -0.00001253, -0.00020095, 0.00015324, 0.00046990, -0.00013112, -0.00010394, -0.00001451, -0.00002388, -0.00000573, 0.00000228, -0.00001091, -0.00000373, -0.00000855, -0.00000071, -0.00000438, 0.00001240, -0.00000936, -0.00000043, -0.00000438, 0.00000032, -0.00000527, -0.00000258, -0.00000021, -0.00000686, 0.00000160, 0.00000843, -0.00000336, -0.00000460, 0.00000035, -0.00000690, -0.00000693, 0.00000478, -0.00001990, 0.00000801, -0.00000166, 0.00000903, 0.00002459, -0.00001174, 0.00003808, -0.00000799, 0.00001637, -0.00002679, -0.00000203, 0.00000430, -0.00000810, 0.00000425, -0.00000082, -0.00000533, 0.00000180, -0.00000558, 0.00000127, -0.00000058, 0.00000072, -0.00001307, 0.00000860, 0.00000596, 0.00000242, -0.00000285, 0.00000142, 0.00000035, -0.00000235, 0.00001265, -0.00000336, -0.00001369, 0.00000501, 0.00001235, -0.00001294, 0.00000829, -0.00000829, 0.00000562, -0.00001104, 0.00001739, -0.00001568, 0.00000272, -0.00000375, },
		{ 0.00000202, -0.00000249, 0.00001078, -0.00001385, -0.00000329, 0.00000476, 0.00000395, -0.00000458, -0.00000144, 0.00000460, 0.00000752, -0.00000049, 0.00000002, -0.00000530, 0.00000873, -0.00001298, 0.00001814, -0.00002004, 0.00001875, -0.00001679, -0.00000190, -0.00000245, -0.00000549, 0.00000209, -0.00000842, 0.00001152, -0.00001807, 0.00001356, -0.00001173, 0.00001138, -0.00000020, 0.00000041, -0.00000207, 0.00000153, -0.00000962, 0.00001132, 0.00000122, 0.00000031, -0.00000005, 0.00000178, -0.00001161, 0.00001125, -0.00000065, -0.00000558, -0.00000202, -0.00000722, -0.00000330, -0.00000044, -0.00000686, -0.00000101, -0.00000953, -0.00001668, -0.00001914, -0.00005544, -0.00018072, 0.00032280, 0.00032273, -0.00018314, -0.00005688, -0.00002072, -0.00001854, -0.00000212, -0.00000035, -0.00001104, -0.00000368, -0.00000706, -0.00000174, -0.00000021, 0.00001000, -0.00001213, 0.00000300, -0.00000714, 0.00000278, -0.00000836, 0.00000134, -0.00000450, -0.00000405, 0.00000304, 0.00000819, -0.00000750, -0.00000072, -0.00000255, -0.00000711, -0.00000449, 0.00000204, -0.00002137, 0.00001516, -0.00001070, 0.00002652, 0.00000735, 0.00000390, 0.00003107, -0.00000728, 0.00001312, -0.00003439, 0.00001163, -0.00000636, -0.00000162, 0.00000222, -0.00000116, -0.00000521, 0.00000168, -0.00000624, 0.00000310, -0.00000162, -0.00000118, -0.00001242, 0.00001546, -0.00000089, 0.00000578, -0.00000736, 0.00000787, -0.00000699, 0.00000568, 0.00000780, -0.00000437, -0.00001542, 0.00001553, 0.00000021, -0.00000512, 0.00000262, -0.00000365, 0.00000034, -0.00000434, 0.00001281, -0.00001667, 0.00000599, -0.00000678, },
		{ 0.00000146, 0.00000181, 0.00000548, -0.00001572, 0.00000184, 0.00000509, 0.00000156, -0.00000472, -0.00000011, 0.00000695, 0.00000457, 0.00000077, -0.00000412, 0.00000229, -0.00000035, -0.00000410, 0.00001117, -0.00001769, 0.00002202, -0.00002760, 0.00001114, -0.00001467, 0.00000663, -0.00000936, 0.00000362, 0.00000001, -0.00000949, 0.00000823, -0.00000821, 0.00001124, -0.00000059, -0.00000129, 0.00000045, -0.00000306, -0.00000458, 0.00001123, -0.00000050, -0.00000017, 0.00000209, -0.00000281, -0.00000729, 0.00001233, -0.00000518, -0.00000380, -0.00000270, -0.00000899, 0.00000118, -0.00000481, -0.00000274, -0.00000602, -0.00000608, -0.00002710, -0.00000511, -0.00011801, -0.00008307, 0.00041318, 0.00016162, -0.00018390, -0.00002669, -0.00002552, -0.00001317, 0.00000099, -0.00000477, -0.00000895, -0.00000491, -0.00000456, -0.00000371, 0.00000583, 0.00000311, -0.00000965, 0.00000232, -0.00000674, 0.00000242, -0.00000862, 0.00000261, -0.00000712, -0.00000130, 0.00000477, 0.00000560, -0.00000895, 0.00000176, -0.00000545, -0.00000664, -0.00000186, -0.00000339, -0.00001767, 0.00001609, -0.00001256, 0.00003890, -0.00000959, 0.00002477, 0.00001469, 0.00000295, -0.00000213, -0.00002809, 0.00001536, -0.00001226, 0.00000356, 0.00000023, -0.00000198, -0.00000396, -0.00000023, -0.00000428, 0.00000225, -0.00000054, -0.00000616, -0.00000550, 0.00001531, -0.00000332, 0.00000561, -0.00000746, 0.00001002, -0.00001121, 0.00001246, 0.00000245, -0.00000604, -0.00001289, 0.00002109, -0.00001010, 0.00000349, -0.00000473, 0.00000281, -0.00000669, 0.00000523, 0.00000308, -0.00001144, 0.00000387, -0.00000606, },
		{ 0.00000034, 0.00000607, -0.00000278, -0.00001270, 0.00000487, 0.00000515, -0.00000152, -0.00000334, 0.00000098, 0.00000857, 0.00000184, 0.00000114, -0.00000585, 0.00000778, -0.00000971, 0.00000839, -0.00000330, -0.00000411, 0.00001158, -0.00002501, 0.00001354, -0.00001872, 0.00001254, -0.00001612, 0.00001235, -0.00001013, -0.00000055, 0.00000178, -0.00000302, 0.00000978, -0.00000103, -0.00000219, 0.00000181, -0.00000672, 0.00000150, 0.00000901, -0.00000155, 0.00000026, 0.00000273, -0.00000648, -0.00000172, 0.00001081, -0.00000893, -0.00000040, -0.00000664, -0.00000620, 0.00000129, -0.00000541, -0.00000284, -0.00000611, -0.00001025, -0.00002593, -0.00001152, -0.00016269, 0.00005951, 0.00041366, 0.00002686, -0.00015804, -0.00000862, -0.00003050, -0.00000497, -0.00000030, -0.00000665, -0.00000820, -0.00000435, -0.00000422, -0.00000291, 0.00000929, -0.00000383, -0.00000504, -0.00000065, -0.00000432, 0.00000014, -0.00000674, 0.00000138, -0.00000751, 0.00000084, 0.00000650, 0.00000113, -0.00000729, 0.00000154, -0.00000683, -0.00000639, 0.00000083, -0.00001068, -0.00000925, 0.00001012, -0.00000414, 0.00003964, -0.00001697, 0.00004067, -0.00000189, 0.00001465, -0.00002258, -0.00001103, 0.00000811, -0.00001026, 0.00000434, 0.00000024, -0.00000413, -0.00000159, -0.00000326, -0.00000064, -0.00000039, 0.00000137, -0.00001189, 0.00000542, 0.00000854, -0.00000028, 0.00000085, -0.00000194, 0.00000568, -0.00000919, 0.00001458, -0.00000082, -0.00000953, -0.00000501, 0.00001873, -0.00001384, 0.00000773, -0.00000890, 0.00000659, -0.00001091, 0.00001264, -0.00000685, -0.00000389, -0.00000118, -0.00000295, },
		{ 0.00000037, 0.00000745, -0.00001030, -0.00000703, 0.00000632, 0.00000422, -0.00000386, -0.00000153, 0.00000237, 0.00000872, 0.00000024, -0.00000023, -0.00000338, 0.00000710, -0.00001281, 0.00001611, -0.00001607, 0.00001212, -0.00000634, -0.00001195, 0.00000540, -0.00001300, 0.00000960, -0.00001498, 0.00001413, -0.00001506, 0.00000517, -0.00000271, 0.00000132, 0.00000873, -0.00000290, -0.00000083, 0.00000026, -0.00000701, 0.00000585, 0.00000647, -0.00000264, 0.00000208, 0.00000088, -0.00000739, 0.00000300, 0.00000761, -0.00001114, 0.00000309, -0.00001233, 0.00000043, -0.00000322, -0.00000159, -0.00000779, -0.00000081, -0.00002260, -0.00001114, -0.00004836, -0.00015580, 0.00019692, 0.00034704, -0.00006701, -0.00012094, -0.00000096, -0.00003337, 0.00000403, -0.00000522, -0.00000562, -0.00000912, -0.00000198, -0.00000617, 0.00000137, 0.00000806, -0.00000764, -0.00000131, -0.00000371, -0.00000149, -0.00000270, -0.00000402, -0.00000138, -0.00000594, 0.00000231, 0.00000755, -0.00000386, -0.00000348, -0.00000130, -0.00000616, -0.00000675, 0.00000294, -0.00001799, 0.00000132, 0.00000003, 0.00001291, 0.00002735, -0.00000963, 0.00004342, -0.00000945, 0.00001782, -0.00003711, 0.00000716, -0.00000426, -0.00000295, 0.00000158, 0.00000163, -0.00000666, 0.00000058, -0.00000552, 0.00000233, -0.00000245, 0.00000141, -0.00001422, 0.00001496, -0.00000030, 0.00000510, -0.00000561, 0.00000633, -0.00000307, -0.00000170, 0.00001175, -0.00000184, -0.00001356, 0.00000596, 0.00000960, -0.00001024, 0.00000566, -0.00000753, 0.00000522, -0.00000952, 0.00001422, -0.00001256, 0.00000166, -0.00000576, 0.00000037, },
		{ 0.00000196, 0.00000471, -0.00001422, -0.00000113, 0.00000685, 0.00000199, -0.00000446, -0.00000018, 0.00000428, 0.00000737, -0.00000038, -0.00000240, 0.00000203, 0.00000021, -0.00000712, 0.00001367, -0.00001897, 0.00002089, -0.00002139, 0.00000290, -0.00000686, -0.00000194, 0.00000071, -0.00000766, 0.00000962, -0.00001462, 0.00000720, -0.00000481, 0.00000473, 0.00000753, -0.00000515, 0.00000165, -0.00000324, -0.00000413, 0.00000783, 0.00000398, -0.00000306, 0.00000393, -0.00000218, -0.00000604, 0.00000653, 0.00000260, -0.00001003, 0.00000329, -0.00001535, 0.00000618, -0.00000796, 0.00000216, -0.00001237, 0.00000318, -0.00003398, 0.00000456, -0.00010312, -0.00008837, 0.00028822, 0.00024918, -0.00012332, -0.00007965, -0.00000428, -0.00002965, 0.00000876, -0.00000945, -0.00000458, -0.00000926, -0.00000040, -0.00000765, 0.00000659, 0.00000342, -0.00000782, 0.00000019, -0.00000550, 0.00000058, -0.00000499, -0.00000172, -0.00000427, -0.00000327, 0.00000342, 0.00000709, -0.00000753, 0.00000037, -0.00000510, -0.00000445, -0.00000699, 0.00000307, -0.00002255, 0.00000985, -0.00000834, 0.00003182, 0.00000756, 0.00000988, 0.00003201, -0.00000488, 0.00000740, -0.00003736, 0.00001659, -0.00001336, 0.00000400, -0.00000129, 0.00000212, -0.00000772, 0.00000093, -0.00000527, 0.00000266, -0.00000208, -0.00000223, -0.00001003, 0.00001800, -0.00000558, 0.00000792, -0.00000895, 0.00001206, -0.00001113, 0.00000725, 0.00000657, -0.00000253, -0.00001507, 0.00001534, -0.00000157, -0.00000270, -0.00000052, -0.00000206, -0.00000025, -0.00000321, 0.00000967, -0.00001249, 0.00000292, -0.00000760, 0.00000257, },
		{ 0.00000393, -0.00000107, -0.00001379, 0.00000356, 0.00000659, -0.00000093, -0.00000344, 0.00000062, 0.00000625, 0.00000523, -0.00000098, -0.00000312, 0.00000651, -0.00000850, 0.00000393, 0.00000235, -0.00001020, 0.00001696, -0.00002574, 0.00001098, -0.00001533, 0.00000756, -0.00000843, 0.00000129, 0.00000221, -0.00001100, 0.00000698, -0.00000563, 0.00000826, 0.00000457, -0.00000554, 0.00000275, -0.00000601, -0.00000017, 0.00000833, 0.00000106, -0.00000169, 0.00000393, -0.00000426, -0.00000404, 0.00000922, -0.00000413, -0.00000483, -0.00000196, -0.00001195, 0.00000623, -0.00000790, 0.00000055, -0.00001053, -0.00000204, -0.00003278, 0.00000175, -0.00014570, 0.00001864, 0.00031889, 0.00014882, -0.00014995, -0.00003948, -0.00001668, -0.00001770, 0.00000597, -0.00000943, -0.00000618, -0.00000646, -0.00000195, -0.00000580, 0.00000931, -0.00000163, -0.00000589, -0.00000029, -0.00000568, 0.00000139, -0.00000616, -0.00000072, -0.00000603, -0.00000059, 0.00000459, 0.00000458, -0.00000834, 0.00000197, -0.00000752, -0.00000356, -0.00000577, -0.00000029, -0.00002184, 0.00001248, -0.00000875, 0.00004359, -0.00000941, 0.00003212, 0.00001340, 0.00000659, -0.00001288, -0.00002321, 0.00001325, -0.00001400, 0.00000619, -0.00000145, 0.00000008, -0.00000638, -0.00000085, -0.00000250, 0.00000035, 0.00000035, -0.00000844, 0.00000005, 0.00001304, -0.00000443, 0.00000533, -0.00000607, 0.00001119, -0.00001375, 0.00001316, 0.00000224, -0.00000492, -0.00001159, 0.00001878, -0.00000932, 0.00000362, -0.00000636, 0.00000361, -0.00000618, 0.00000450, 0.00000189, -0.00000828, 0.00000033, -0.00000640, 0.00000365, },
		{ 0.00000444, -0.00000735, -0.00001007, 0.00000672, 0.00000522, -0.00000330, -0.00000171, 0.00000137, 0.00000757, 0.00000312, -0.00000226, -0.00000074, 0.00000617, -0.00001294, 0.00001313, -0.00001068, 0.00000459, 0.00000296, -0.00001855, 0.00000902, -0.00001560, 0.00001073, -0.00001321, 0.00000772, -0.00000477, -0.00000639, 0.00000574, -0.00000571, 0.00001209, -0.00000060, -0.00000302, 0.00000092, -0.00000602, 0.00000286, 0.00000821, -0.00000219, 0.00000144, 0.00000153, -0.00000428, -0.00000227, 0.00001062, -0.00001071, 0.00000230, -0.00001099, -0.00000270, -0.00000000, -0.00000216, -0.00000716, -0.00000160, -0.00001757, -0.00001578, -0.00003064, -0.00014616, 0.00012718, 0.00030032, 0.00005909, -0.00015141, -0.00000806, -0.00003116, -0.00000167, -0.00000251, -0.00000559, -0.00000997, -0.00000141, -0.00000621, -0.00000029, 0.00000788, -0.00000466, -0.00000374, -0.00000180, -0.00000454, 0.00000088, -0.00000612, -0.00000124, -0.00000596, 0.00000129, 0.00000589, 0.00000032, -0.00000602, 0.00000027, -0.00000709, -0.00000479, -0.00000245, -0.00000736, -0.00001541, 0.00000821, 0.00000161, 0.00004195, -0.00001409, 0.00004601, -0.00000196, 0.00001481, -0.00003248, -0.00000282, 0.00000112, -0.00000707, 0.00000340, 0.00000109, -0.00000397, -0.00000352, -0.00000333, 0.00000084, -0.00000251, 0.00000228, -0.00001338, 0.00001146, 0.00000341, 0.00000142, -0.00000131, 0.00000195, 0.00000370, -0.00000949, 0.00001392, 0.00000024, -0.00000906, -0.00000334, 0.00001498, -0.00001070, 0.00000517, -0.00000822, 0.00000589, -0.00000890, 0.00000959, -0.00000511, -0.00000327, -0.00000380, -0.00000328, 0.00000435, },
		{ 0.00000241, -0.00001169, -0.00000482, 0.00000847, 0.00000264, -0.00000415, -0.00000026, 0.00000251, 0.00000781, 0.00000123, -0.00000362, 0.00000407, 0.00000009, -0.00000955, 0.00001425, -0.00001713, 0.00001612, -0.00001272, -0.00000565, 0.00000011, -0.00000877, 0.00000722, -0.00001236, 0.00000994, -0.00000969, -0.00000169, 0.00000344, -0.00000415, 0.00001451, -0.00000617, 0.00000108, -0.00000300, -0.00000332, 0.00000447, 0.00000733, -0.00000450, 0.00000463, -0.00000181, -0.00000310, -0.00000016, 0.00000915, -0.00001388, 0.00000676, -0.00001849, 0.00000720, -0.00000792, 0.00000467, -0.00001563, 0.00000765, -0.00003476, 0.00000623, -0.00008412, -0.00009360, 0.00020372, 0.00025538, -0.00001798, -0.00012975, 0.00000684, -0.00003845, 0.00001066, -0.00001087, -0.00000177, -0.00001268, 0.00000263, -0.00001001, 0.00000626, 0.00000343, -0.00000516, -0.00000245, -0.00000346, -0.00000258, -0.00000065, -0.00000525, -0.00000278, -0.00000422, 0.00000228, 0.00000675, -0.00000426, -0.00000202, -0.00000377, -0.00000433, -0.00000760, 0.00000177, -0.00001613, -0.00000578, 0.00000032, 0.00001969, 0.00002745, -0.00000337, 0.00004528, -0.00000680, 0.00001158, -0.00004066, 0.00001259, -0.00001117, 0.00000188, -0.00000099, 0.00000380, -0.00000778, -0.00000115, -0.00000441, 0.00000232, -0.00000339, 0.00000094, -0.00001286, 0.00001828, -0.00000493, 0.00000705, -0.00000738, 0.00001036, -0.00000637, -0.00000088, 0.00001064, -0.00000038, -0.00001274, 0.00000640, 0.00000649, -0.00000673, 0.00000190, -0.00000548, 0.00000379, -0.00000706, 0.00000997, -0.00000857, -0.00000034, -0.00000703, 0.00000040, 0.00000502, },
		{ -0.00000174, -0.00001293, 0.00000054, 0.00000873, -0.00000052, -0.00000341, 0.00000059, 0.00000401, 0.00000714, -0.00000083, -0.00000338, 0.00000815, -0.00000856, 0.00000009, 0.00000635, -0.00001303, 0.00001753, -0.00002157, 0.00000488, -0.00000918, 0.00000029, 0.00000024, -0.00000768, 0.00000857, -0.00001244, 0.00000305, -0.00000042, 0.00000035, 0.00001319, -0.00000917, 0.00000388, -0.00000642, 0.00000024, 0.00000546, 0.00000497, -0.00000433, 0.00000567, -0.00000380, -0.00000244, 0.00000324, 0.00000372, -0.00001120, 0.00000455, -0.00001893, 0.00001140, -0.00001124, 0.00000612, -0.00001744, 0.00000782, -0.00004057, 0.00001377, -0.00013220, -0.00000355, 0.00023541, 0.00020199, -0.00008150, -0.00009030, 0.00000283, -0.00003295, 0.00001322, -0.00001377, -0.00000181, -0.00001121, 0.00000249, -0.00000986, 0.00001021, -0.00000127, -0.00000426, -0.00000210, -0.00000469, -0.00000046, -0.00000267, -0.00000419, -0.00000432, -0.00000176, 0.00000292, 0.00000622, -0.00000722, 0.00000112, -0.00000757, -0.00000156, -0.00000980, 0.00000418, -0.00002301, 0.00000254, -0.00000492, 0.00003748, 0.00000778, 0.00001750, 0.00003193, -0.00000159, -0.00000368, -0.00003328, 0.00001574, -0.00001619, 0.00000704, -0.00000306, 0.00000393, -0.00000924, -0.00000069, -0.00000305, 0.00000093, -0.00000150, -0.00000412, -0.00000552, 0.00001698, -0.00000732, 0.00000800, -0.00000830, 0.00001386, -0.00001353, 0.00000755, 0.00000628, -0.00000167, -0.00001308, 0.00001324, -0.00000199, -0.00000122, -0.00000318, -0.00000049, -0.00000081, -0.00000212, 0.00000635, -0.00000820, -0.00000050, -0.00000798, 0.00000380, 0.00000521, },
		{ -0.00000643, -0.00001128, 0.00000522, 0.00000726, -0.00000307, -0.00000183, 0.00000112, 0.00000543, 0.00000593, -0.00000333, -0.00000014, 0.00000784, -0.00001423, 0.00001004, -0.00000555, -0.00000126, 0.00000845, -0.00002006, 0.00000780, -0.00001341, 0.00000660, -0.00000620, -0.00000191, 0.00000498, -0.00001312, 0.00000726, -0.00000524, 0.00000760, 0.00000729, -0.00000778, 0.00000319, -0.00000697, 0.00000273, 0.00000656, 0.00000110, -0.00000133, 0.00000367, -0.00000331, -0.00000302, 0.00000750, -0.00000450, -0.00000358, -0.00000439, -0.00001085, 0.00000719, -0.00000703, -0.00000061, -0.00000958, -0.00000520, -0.00002804, -0.00000791, -0.00014493, 0.00009092, 0.00023167, 0.00014565, -0.00012580, -0.00004484, -0.00001391, -0.00001672, 0.00000593, -0.00001021, -0.00000618, -0.00000548, -0.00000214, -0.00000477, 0.00000952, -0.00000388, -0.00000338, -0.00000236, -0.00000510, 0.00000111, -0.00000452, -0.00000361, -0.00000484, 0.00000028, 0.00000387, 0.00000369, -0.00000718, 0.00000128, -0.00000864, -0.00000128, -0.00000906, 0.00000222, -0.00002479, 0.00000580, -0.00000187, 0.00004622, -0.00000635, 0.00003801, 0.00001431, 0.00000627, -0.00002343, -0.00001516, 0.00000726, -0.00001215, 0.00000623, -0.00000135, 0.00000071, -0.00000798, -0.00000181, -0.00000026, -0.00000205, 0.00000141, -0.00001017, 0.00000574, 0.00000882, -0.00000331, 0.00000344, -0.00000286, 0.00001006, -0.00001425, 0.00001223, 0.00000321, -0.00000474, -0.00000881, 0.00001440, -0.00000660, 0.00000197, -0.00000648, 0.00000340, -0.00000481, 0.00000295, 0.00000127, -0.00000573, -0.00000292, -0.00000652, 0.00000659, 0.00000402, },
		{ -0.00000999, -0.00000764, 0.00000871, 0.00000418, -0.00000408, -0.00000036, 0.00000177, 0.00000640, 0.00000421, -0.00000555, 0.00000557, 0.00000170, -0.00001273, 0.00001389, -0.00001384, 0.00001060, -0.00000528, -0.00001104, 0.00000334, -0.00001120, 0.00000793, -0.00000974, 0.00000305, 0.00000006, -0.00001123, 0.00000927, -0.00000867, 0.00001503, -0.00000127, -0.00000292, -0.00000076, -0.00000430, 0.00000357, 0.00000745, -0.00000290, 0.00000289, -0.00000006, -0.00000132, -0.00000370, 0.00001034, -0.00001179, 0.00000447, -0.00001531, 0.00000164, -0.00000224, 0.00000158, -0.00001171, 0.00000309, -0.00002542, -0.00000371, -0.00005636, -0.00010706, 0.00015814, 0.00021232, 0.00008357, -0.00014273, -0.00000847, -0.00003121, 0.00000169, -0.00000546, -0.00000393, -0.00001162, 0.00000120, -0.00000819, 0.00000295, 0.00000501, -0.00000391, -0.00000323, -0.00000294, -0.00000446, 0.00000155, -0.00000573, -0.00000374, -0.00000391, 0.00000130, 0.00000516, -0.00000035, -0.00000440, -0.00000181, -0.00000628, -0.00000428, -0.00000483, -0.00000461, -0.00002085, 0.00000368, 0.00001068, 0.00004160, -0.00000724, 0.00004853, 0.00000144, 0.00000792, -0.00003690, 0.00000329, -0.00000569, -0.00000306, 0.00000169, 0.00000238, -0.00000422, -0.00000546, -0.00000280, 0.00000168, -0.00000402, 0.00000231, -0.00001285, 0.00001528, -0.00000107, 0.00000330, -0.00000334, 0.00000587, 0.00000094, -0.00000883, 0.00001243, 0.00000177, -0.00000861, -0.00000140, 0.00001028, -0.00000642, 0.00000131, -0.00000618, 0.00000416, -0.00000604, 0.00000568, -0.00000271, -0.00000343, -0.00000590, -0.00000320, 0.00000838, 0.00000105, },
		{ -0.00001154, -0.00000288, 0.00001039, 0.00000036, -0.00000342, 0.00000055, 0.00000269, 0.00000688, 0.00000163, -0.00000575, 0.00001080, -0.00000786, -0.00000427, 0.00000913, -0.00001333, 0.00001522, -0.00001574, -0.00000107, -0.00000407, -0.00000506, 0.00000515, -0.00001015, 0.00000650, -0.00000568, -0.00000626, 0.00000735, -0.00000783, 0.00001880, -0.00000857, 0.00000223, -0.00000542, -0.00000011, 0.00000366, 0.00000698, -0.00000493, 0.00000573, -0.00000295, -0.00000007, -0.00000248, 0.00000912, -0.00001404, 0.00000736, -0.00002141, 0.00001128, -0.00001003, 0.00000763, -0.00001920, 0.00001061, -0.00003973, 0.00001441, -0.00010983, -0.00002817, 0.00018441, 0.00019212, 0.00001471, -0.00012903, 0.00000777, -0.00003791, 0.00001289, -0.00001353, -0.00000020, -0.00001374, 0.00000455, -0.00001133, 0.00000921, -0.00000043, -0.00000257, -0.00000364, -0.00000366, -0.00000284, 0.00000068, -0.00000622, -0.00000420, -0.00000199, 0.00000160, 0.00000602, -0.00000432, -0.00000083, -0.00000625, -0.00000220, -0.00000889, 0.00000090, -0.00001405, -0.00001364, -0.00000012, 0.00002817, 0.00002651, 0.00000553, 0.00004558, -0.00000275, -0.00000069, -0.00003703, 0.00001274, -0.00001462, 0.00000479, -0.00000235, 0.00000494, -0.00000827, -0.00000350, -0.00000221, 0.00000112, -0.00000326, -0.00000057, -0.00000921, 0.00001817, -0.00000703, 0.00000741, -0.00000738, 0.00001259, -0.00000863, -0.00000062, 0.00000990, 0.00000060, -0.00001093, 0.00000575, 0.00000396, -0.00000352, -0.00000186, -0.00000306, 0.00000208, -0.00000442, 0.00000547, -0.00000445, -0.00000273, -0.00000792, 0.00000121, 0.00000853, -0.00000311, },
		{ -0.00001105, 0.00000231, 0.00000970, -0.00000282, -0.00000186, 0.00000106, 0.00000367, 0.00000698, -0.00000199, -0.00000236, 0.00001170, -0.00001562, 0.00000637, -0.00000114, -0.00000468, 0.00001021, -0.00001798, 0.00000417, -0.00000937, 0.00000109, 0.00000068, -0.00000835, 0.00000813, -0.00001127, 0.00000100, 0.00000142, -0.00000152, 0.00001632, -0.00001121, 0.00000437, -0.00000781, 0.00000327, 0.00000417, 0.00000442, -0.00000376, 0.00000544, -0.00000317, -0.00000100, 0.00000142, 0.00000313, -0.00000975, 0.00000235, -0.00001838, 0.00001264, -0.00001071, 0.00000570, -0.00001706, 0.00000535, -0.00003720, 0.00000808, -0.00013817, 0.00006219, 0.00017822, 0.00017226, -0.00005334, -0.00009178, 0.00000300, -0.00003026, 0.00001253, -0.00001426, -0.00000180, -0.00001050, 0.00000247, -0.00000903, 0.00001089, -0.00000390, -0.00000155, -0.00000409, -0.00000429, -0.00000074, -0.00000117, -0.00000629, -0.00000423, -0.00000013, 0.00000198, 0.00000540, -0.00000637, 0.00000102, -0.00000916, 0.00000055, -0.00001199, 0.00000468, -0.00002229, -0.00000708, -0.00000004, 0.00004258, 0.00000947, 0.00002523, 0.00003284, -0.00000110, -0.00001612, -0.00002468, 0.00001026, -0.00001507, 0.00000732, -0.00000291, 0.00000420, -0.00000985, -0.00000278, -0.00000016, -0.00000154, -0.00000039, -0.00000607, -0.00000009, 0.00001337, -0.00000646, 0.00000621, -0.00000560, 0.00001324, -0.00001399, 0.00000645, 0.00000701, -0.00000165, -0.00000982, 0.00000953, -0.00000100, -0.00000101, -0.00000478, 0.00000048, -0.00000090, -0.00000153, 0.00000341, -0.00000441, -0.00000379, -0.00000809, 0.00000585, 0.00000647, -0.00000719, },
		{ -0.00000891, 0.00000717, 0.00000664, -0.00000425, -0.00000036, 0.00000156, 0.00000450, 0.00000647, -0.00000582, 0.00000439, 0.00000613, -0.00001673, 0.00001280, -0.00001025, 0.00000627, -0.00000086, -0.00001258, 0.00000304, -0.00001002, 0.00000461, -0.00000350, -0.00000511, 0.00000733, -0.00001476, 0.00000796, -0.00000597, 0.00000834, 0.00000818, -0.00000846, 0.00000230, -0.00000648, 0.00000458, 0.00000531, 0.00000047, -0.00000012, 0.00000249, -0.00000111, -0.00000341, 0.00000639, -0.00000514, -0.00000163, -0.00000816, -0.00000753, 0.00000587, -0.00000428, -0.00000368, -0.00000634, -0.00001147, -0.00001813, -0.00002782, -0.00012023, 0.00013100, 0.00016077, 0.00014350, -0.00010584, -0.00004699, -0.00001357, -0.00001332, 0.00000336, -0.00000906, -0.00000720, -0.00000371, -0.00000341, -0.00000219, 0.00000765, -0.00000423, -0.00000183, -0.00000427, -0.00000447, 0.00000107, -0.00000339, -0.00000623, -0.00000329, 0.00000082, 0.00000303, 0.00000288, -0.00000562, -0.00000029, -0.00000859, -0.00000026, -0.00001118, 0.00000369, -0.00002637, -0.00000369, 0.00000759, 0.00004708, -0.00000022, 0.00004233, 0.00001768, -0.00000019, -0.00002996, -0.00000754, 0.00000009, -0.00000846, 0.00000474, -0.00000017, 0.00000042, -0.00000921, -0.00000249, 0.00000167, -0.00000416, 0.00000196, -0.00001049, 0.00001012, 0.00000425, -0.00000119, 0.00000095, 0.00000116, 0.00000742, -0.00001328, 0.00001016, 0.00000495, -0.00000494, -0.00000545, 0.00000901, -0.00000283, -0.00000083, -0.00000542, 0.00000237, -0.00000283, 0.00000077, 0.00000116, -0.00000388, -0.00000582, -0.00000611, 0.00000962, 0.00000231, -0.00001001, },
		{ -0.00000539, 0.00001059, 0.00000220, -0.00000372, 0.00000056, 0.00000219, 0.00000525, 0.00000470, -0.00000784, 0.00001172, -0.00000437, -0.00001000, 0.00001122, -0.00001240, 0.00001247, -0.00001119, -0.00000436, -0.00000191, -0.00000682, 0.00000514, -0.00000662, -0.00000081, 0.00000345, -0.00001401, 0.00001120, -0.00001063, 0.00001730, -0.00000197, -0.00000272, -0.00000240, -0.00000234, 0.00000422, 0.00000614, -0.00000292, 0.00000365, -0.00000085, 0.00000106, -0.00000495, 0.00000924, -0.00001125, 0.00000481, -0.00001793, 0.00000498, -0.00000364, 0.00000385, -0.00001417, 0.00000502, -0.00002963, 0.00000415, -0.00007925, -0.00005697, 0.00015953, 0.00015026, 0.00009689, -0.00012936, -0.00001202, -0.00002861, 0.00000313, -0.00000743, -0.00000297, -0.00001208, 0.00000237, -0.00000893, 0.00000548, 0.00000191, -0.00000235, -0.00000324, -0.00000419, -0.00000387, 0.00000188, -0.00000544, -0.00000603, -0.00000152, 0.00000080, 0.00000447, -0.00000076, -0.00000294, -0.00000408, -0.00000492, -0.00000455, -0.00000640, -0.00000242, -0.00002581, -0.00000275, 0.00002179, 0.00004009, 0.00000235, 0.00004986, 0.00000597, -0.00000501, -0.00003463, 0.00000514, -0.00001007, -0.00000005, 0.00000042, 0.00000330, -0.00000437, -0.00000766, -0.00000142, 0.00000149, -0.00000455, 0.00000132, -0.00001018, 0.00001616, -0.00000373, 0.00000426, -0.00000425, 0.00000877, -0.00000164, -0.00000796, 0.00001074, 0.00000330, -0.00000774, 0.00000005, 0.00000575, -0.00000221, -0.00000270, -0.00000367, 0.00000211, -0.00000304, 0.00000161, -0.00000024, -0.00000398, -0.00000774, -0.00000217, 0.00001130, -0.00000293, -0.00001115, },
		{ -0.00000068, 0.00001142, -0.00000193, -0.00000203, 0.00000101, 0.00000279, 0.00000610, 0.00000115, -0.00000589, 0.00001537, -0.00001487, 0.00000099, 0.00000297, -0.00000648, 0.00001022, -0.00001529, 0.00000127, -0.00000655, -0.00000243, 0.00000377, -0.00000861, 0.00000394, -0.00000316, -0.00000837, 0.00000869, -0.00000921, 0.00002088, -0.00000947, 0.00000218, -0.00000660, 0.00000227, 0.00000353, 0.00000546, -0.00000392, 0.00000528, -0.00000226, 0.00000114, -0.00000357, 0.00000756, -0.00001179, 0.00000491, -0.00002085, 0.00001225, -0.00000926, 0.00000710, -0.00001837, 0.00000784, -0.00003673, 0.00001106, -0.00011908, 0.00002841, 0.00015145, 0.00015066, 0.00003348, -0.00012037, 0.00000307, -0.00003257, 0.00001110, -0.00001337, -0.00000063, -0.00001272, 0.00000410, -0.00001026, 0.00000999, -0.00000298, -0.00000025, -0.00000491, -0.00000405, -0.00000249, 0.00000141, -0.00000708, -0.00000530, 0.00000025, 0.00000056, 0.00000531, -0.00000389, -0.00000041, -0.00000800, -0.00000063, -0.00000986, -0.00000012, -0.00001150, -0.00002245, -0.00000078, 0.00003719, 0.00002618, 0.00001548, 0.00004643, -0.00000107, -0.00001522, -0.00002820, 0.00000813, -0.00001427, 0.00000520, -0.00000205, 0.00000482, -0.00000812, -0.00000619, 0.00000068, -0.00000093, -0.00000235, -0.00000257, -0.00000419, 0.00001523, -0.00000658, 0.00000601, -0.00000545, 0.00001269, -0.00000944, -0.00000122, 0.00000974, 0.00000107, -0.00000829, 0.00000402, 0.00000240, -0.00000129, -0.00000478, -0.00000105, 0.00000070, -0.00000222, 0.00000138, -0.00000083, -0.00000506, -0.00000860, 0.00000304, 0.00000998, -0.00000770, -0.00001082, },
		{ 0.00000464, 0.00000921, -0.00000422, -0.00000033, 0.00000143, 0.00000320, 0.00000681, -0.00000366, 0.00000059, 0.00001211, -0.00001978, 0.00001002, -0.00000640, 0.00000332, 0.00000127, -0.00001219, 0.00000174, -0.00000806, 0.00000091, 0.00000156, -0.00000913, 0.00000770, -0.00001037, 0.00000028, 0.00000142, -0.00000145, 0.00001722, -0.00001145, 0.00000343, -0.00000772, 0.00000524, 0.00000348, 0.00000306, -0.00000218, 0.00000414, -0.00000114, -0.00000121, 0.00000061, 0.00000173, -0.00000692, -0.00000189, -0.00001501, 0.00001116, -0.00000783, 0.00000270, -0.00001354, -0.00000122, -0.00002755, -0.00000862, -0.00012226, 0.00010312, 0.00012723, 0.00015152, -0.00003375, -0.00008792, -0.00000052, -0.00002428, 0.00000875, -0.00001247, -0.00000316, -0.00000858, 0.00000095, -0.00000637, 0.00000945, -0.00000477, 0.00000037, -0.00000600, -0.00000387, -0.00000069, -0.00000020, -0.00000818, -0.00000368, 0.00000111, 0.00000098, 0.00000458, -0.00000506, -0.00000002, -0.00000952, 0.00000136, -0.00001299, 0.00000419, -0.00002041, -0.00001864, 0.00000591, 0.00004710, 0.00001325, 0.00003261, 0.00003544, -0.00000618, -0.00002568, -0.00001519, 0.00000264, -0.00001144, 0.00000562, -0.00000138, 0.00000336, -0.00001003, -0.00000468, 0.00000259, -0.00000397, 0.00000058, -0.00000721, 0.00000482, 0.00000864, -0.00000405, 0.00000337, -0.00000162, 0.00001077, -0.00001287, 0.00000428, 0.00000842, -0.00000202, -0.00000610, 0.00000507, 0.00000078, -0.00000177, -0.00000530, 0.00000075, -0.00000046, -0.00000153, 0.00000104, -0.00000136, -0.00000680, -0.00000769, 0.00000823, 0.00000572, -0.00001082, -0.00000938, },
		{ 0.00000933, 0.00000473, -0.00000416, 0.00000068, 0.00000201, 0.00000358, 0.00000653, -0.00000771, 0.00000944, 0.00000215, -0.00001632, 0.00001214, -0.00001076, 0.00001044, -0.00000869, -0.00000528, -0.00000174, -0.00000637, 0.00000252, -0.00000113, -0.00000753, 0.00000848, -0.00001497, 0.00000802, -0.00000668, 0.00000933, 0.00000816, -0.00000829, 0.00000081, -0.00000523, 0.00000593, 0.00000396, 0.00000004, 0.00000093, 0.00000159, 0.00000103, -0.00000420, 0.00000514, -0.00000502, -0.00000041, -0.00001154, -0.00000383, 0.00000408, -0.00000162, -0.00000631, -0.00000379, -0.00001675, -0.00000888, -0.00004974, -0.00008052, 0.00014247, 0.00010967, 0.00013737, -0.00008655, -0.00004880, -0.00001303, -0.00000992, 0.00000027, -0.00000749, -0.00000799, -0.00000242, -0.00000441, 0.00000054, 0.00000491, -0.00000338, -0.00000101, -0.00000614, -0.00000352, 0.00000093, -0.00000257, -0.00000856, -0.00000136, 0.00000080, 0.00000231, 0.00000222, -0.00000405, -0.00000228, -0.00000770, -0.00000040, -0.00001203, 0.00000413, -0.00002692, -0.00001509, 0.00001827, 0.00004750, 0.00000809, 0.00004637, 0.00002159, -0.00001256, -0.00003041, -0.00000312, -0.00000570, -0.00000487, 0.00000297, 0.00000120, -0.00000021, -0.00001048, -0.00000243, 0.00000268, -0.00000539, 0.00000164, -0.00000902, 0.00001219, 0.00000071, 0.00000072, -0.00000107, 0.00000487, 0.00000426, -0.00001157, 0.00000759, 0.00000691, -0.00000500, -0.00000242, 0.00000381, 0.00000081, -0.00000376, -0.00000388, 0.00000104, -0.00000074, -0.00000156, 0.00000124, -0.00000255, -0.00000850, -0.00000471, 0.00001174, -0.00000027, -0.00001199, -0.00000681, },
		{ 0.00001178, -0.00000017, -0.00000247, 0.00000106, 0.00000253, 0.00000433, 0.00000425, -0.00000833, 0.00001631, -0.00001039, -0.00000638, 0.00000678, -0.00000748, 0.00001026, -0.00001396, 0.00000044, -0.00000576, -0.00000331, 0.00000300, -0.00000432, -0.00000350, 0.00000510, -0.00001439, 0.00001104, -0.00001093, 0.00001807, -0.00000193, -0.00000297, -0.00000350, -0.00000060, 0.00000505, 0.00000410, -0.00000197, 0.00000336, -0.00000027, 0.00000203, -0.00000537, 0.00000705, -0.00000886, 0.00000287, -0.00001809, 0.00000641, -0.00000336, 0.00000385, -0.00001363, 0.00000302, -0.00002824, 0.00000408, -0.00009189, -0.00000758, 0.00014172, 0.00011005, 0.00009957, -0.00011153, -0.00001859, -0.00002315, 0.00000216, -0.00000790, -0.00000304, -0.00001124, 0.00000194, -0.00000810, 0.00000667, -0.00000063, -0.00000050, -0.00000367, -0.00000544, -0.00000286, 0.00000184, -0.00000531, -0.00000782, 0.00000086, -0.00000008, 0.00000389, -0.00000079, -0.00000205, -0.00000590, -0.00000373, -0.00000506, -0.00000744, -0.00000062, -0.00003054, -0.00001030, 0.00003346, 0.00003927, 0.00001307, 0.00005211, 0.00000788, -0.00002062, -0.00002690, 0.00000246, -0.00001112, 0.00000101, 0.00000029, 0.00000342, -0.00000427, -0.00001002, 0.00000065, 0.00000037, -0.00000413, -0.00000039, -0.00000597, 0.00001428, -0.00000418, 0.00000385, -0.00000359, 0.00001003, -0.00000340, -0.00000738, 0.00000923, 0.00000464, -0.00000638, 0.00000069, 0.00000211, 0.00000101, -0.00000588, -0.00000155, 0.00000040, -0.00000054, -0.00000199, 0.00000181, -0.00000461, -0.00000942, 0.00000006, 0.00001218, -0.00000616, -0.00001168, -0.00000287, },
		{ 0.00001094, -0.00000352, -0.00000047, 0.00000132, 0.00000271, 0.00000550, -0.00000015, -0.00000386, 0.00001691, -0.00001953, 0.00000448, -0.00000194, 0.00000095, 0.00000295, -0.00001234, 0.00000174, -0.00000746, -0.00000083, 0.00000296, -0.00000753, 0.00000198, -0.00000166, -0.00000853, 0.00000793, -0.00000859, 0.00002079, -0.00000879, 0.00000099, -0.00000661, 0.00000389, 0.00000377, 0.00000329, -0.00000205, 0.00000391, -0.00000025, 0.00000070, -0.00000359, 0.00000505, -0.00000803, 0.00000036, -0.00001760, 0.00001077, -0.00000642, 0.00000419, -0.00001459, 0.00000124, -0.00002784, -0.00000220, -0.00011041, 0.00006828, 0.00011581, 0.00012334, 0.00004253, -0.00010686, -0.00000469, -0.00002444, 0.00000671, -0.00001139, -0.00000215, -0.00001061, 0.00000228, -0.00000762, 0.00000902, -0.00000420, 0.00000166, -0.00000635, -0.00000430, -0.00000186, 0.00000172, -0.00000795, -0.00000574, 0.00000203, -0.00000050, 0.00000463, -0.00000304, -0.00000086, -0.00000865, -0.00000014, -0.00001003, -0.00000156, -0.00000859, -0.00003187, -0.00000178, 0.00004622, 0.00002755, 0.00002566, 0.00004892, -0.00000509, -0.00002749, -0.00001765, 0.00000086, -0.00001118, 0.00000360, -0.00000046, 0.00000374, -0.00000774, -0.00000859, 0.00000349, -0.00000313, -0.00000125, -0.00000425, 0.00000091, 0.00001070, -0.00000434, 0.00000343, -0.00000211, 0.00001094, -0.00000891, -0.00000255, 0.00001000, 0.00000134, -0.00000543, 0.00000177, 0.00000159, -0.00000009, -0.00000656, 0.00000020, -0.00000007, -0.00000076, -0.00000194, 0.00000198, -0.00000720, -0.00000883, 0.00000553, 0.00000920, -0.00001046, -0.00001040, 0.00000226, },
		{ 0.00000711, -0.00000431, 0.00000082, 0.00000182, 0.00000265, 0.00000640, -0.00000524, 0.00000474, 0.00000971, -0.00002087, 0.00001048, -0.00000802, 0.00000875, -0.00000697, -0.00000599, -0.00000115, -0.00000619, 0.00000042, 0.00000229, -0.00000947, 0.00000666, -0.00000898, -0.00000006, 0.00000066, -0.00000048, 0.00001664, -0.00001045, 0.00000162, -0.00000668, 0.00000660, 0.00000281, 0.00000172, -0.00000057, 0.00000293, 0.00000108, -0.00000217, 0.00000010, 0.00000051, -0.00000411, -0.00000651, -0.00001052, 0.00000856, -0.00000437, -0.00000072, -0.00000963, -0.00000815, -0.00001668, -0.00002970, -0.00009059, 0.00011954, 0.00008821, 0.00013375, -0.00001867, -0.00008196, -0.00000486, -0.00001752, 0.00000390, -0.00000991, -0.00000466, -0.00000670, -0.00000077, -0.00000321, 0.00000695, -0.00000441, 0.00000161, -0.00000793, -0.00000315, -0.00000063, 0.00000040, -0.00000989, -0.00000259, 0.00000167, 0.00000022, 0.00000384, -0.00000364, -0.00000162, -0.00000889, 0.00000088, -0.00001271, 0.00000280, -0.00001783, -0.00003112, 0.00001166, 0.00005197, 0.00001869, 0.00004049, 0.00003834, -0.00001743, -0.00002946, -0.00000813, -0.00000436, -0.00000730, 0.00000323, 0.00000062, 0.00000207, -0.00001030, -0.00000575, 0.00000442, -0.00000562, 0.00000089, -0.00000695, 0.00000805, 0.00000428, -0.00000133, 0.00000055, 0.00000255, 0.00000734, -0.00001083, 0.00000161, 0.00001002, -0.00000221, -0.00000282, 0.00000096, 0.00000239, -0.00000276, -0.00000523, 0.00000060, 0.00000022, -0.00000191, -0.00000083, 0.00000097, -0.00000962, -0.00000632, 0.00001002, 0.00000375, -0.00001266, -0.00000818, 0.00000753, },
		{ 0.00000200, -0.00000294, 0.00000125, 0.00000240, 0.00000285, 0.00000578, -0.00000825, 0.00001373, -0.00000273, -0.00001418, 0.00000914, -0.00000755, 0.00001065, -0.00001372, 0.00000047, -0.00000542, -0.00000336, 0.00000097, 0.00000041, -0.00000880, 0.00000812, -0.00001339, 0.00000688, -0.00000655, 0.00000962, 0.00000808, -0.00000787, -0.00000070, -0.00000371, 0.00000708, 0.00000220, 0.00000026, 0.00000123, 0.00000177, 0.00000208, -0.00000455, 0.00000327, -0.00000364, -0.00000072, -0.00001338, -0.00000104, 0.00000300, -0.00000025, -0.00000715, -0.00000369, -0.00001833, -0.00000473, -0.00006601, -0.00003632, 0.00013317, 0.00007687, 0.00012600, -0.00006685, -0.00005124, -0.00001120, -0.00000770, -0.00000226, -0.00000627, -0.00000805, -0.00000213, -0.00000450, 0.00000254, 0.00000223, -0.00000192, -0.00000079, -0.00000788, -0.00000226, 0.00000063, -0.00000208, -0.00001038, 0.00000068, 0.00000025, 0.00000186, 0.00000178, -0.00000287, -0.00000410, -0.00000658, -0.00000125, -0.00001189, 0.00000374, -0.00002691, -0.00002730, 0.00002845, 0.00004928, 0.00001718, 0.00005213, 0.00002251, -0.00002804, -0.00002516, -0.00000296, -0.00000863, -0.00000269, 0.00000186, 0.00000216, -0.00000082, -0.00001188, -0.00000155, 0.00000259, -0.00000554, 0.00000049, -0.00000600, 0.00001177, -0.00000105, 0.00000165, -0.00000190, 0.00000744, 0.00000136, -0.00000975, 0.00000499, 0.00000877, -0.00000464, -0.00000026, -0.00000032, 0.00000334, -0.00000584, -0.00000266, 0.00000000, 0.00000092, -0.00000358, 0.00000119, -0.00000155, -0.00001104, -0.00000200, 0.00001197, -0.00000241, -0.00001314, -0.00000462, 0.00001115, },
		{ -0.00000217, -0.00000079, 0.00000134, 0.00000266, 0.00000371, 0.00000283, -0.00000669, 0.00001833, -0.00001490, -0.00000352, 0.00000260, -0.00000113, 0.00000522, -0.00001392, 0.00000317, -0.00000795, -0.00000092, 0.00000161, -0.00000283, -0.00000517, 0.00000535, -0.00001278, 0.00000910, -0.00000962, 0.00001714, -0.00000090, -0.00000380, -0.00000394, 0.00000085, 0.00000586, 0.00000171, -0.00000040, 0.00000233, 0.00000150, 0.00000146, -0.00000486, 0.00000407, -0.00000513, -0.00000090, -0.00001600, 0.00000598, -0.00000159, 0.00000200, -0.00001076, -0.00000206, -0.00002194, -0.00000418, -0.00009073, 0.00003208, 0.00011526, 0.00008538, 0.00009457, -0.00009132, -0.00002655, -0.00001601, -0.00000054, -0.00000724, -0.00000372, -0.00000965, 0.00000047, -0.00000612, 0.00000658, -0.00000229, 0.00000130, -0.00000450, -0.00000640, -0.00000169, 0.00000159, -0.00000543, -0.00000885, 0.00000282, -0.00000111, 0.00000348, -0.00000055, -0.00000189, -0.00000686, -0.00000321, -0.00000535, -0.00000816, 0.00000071, -0.00003495, -0.00001862, 0.00004461, 0.00004058, 0.00002392, 0.00005670, 0.00000348, -0.00003428, -0.00001689, -0.00000314, -0.00000942, 0.00000023, 0.00000120, 0.00000288, -0.00000419, -0.00001201, 0.00000280, -0.00000118, -0.00000320, -0.00000217, -0.00000128, 0.00001060, -0.00000287, 0.00000228, -0.00000155, 0.00000964, -0.00000418, -0.00000717, 0.00000792, 0.00000590, -0.00000491, 0.00000074, -0.00000052, 0.00000292, -0.00000771, -0.00000031, -0.00000056, 0.00000109, -0.00000476, 0.00000319, -0.00000526, -0.00001069, 0.00000321, 0.00001071, -0.00000765, -0.00001245, 0.00000042, 0.00001161, },
		{ -0.00000387, 0.00000084, 0.00000165, 0.00000245, 0.00000494, -0.00000184, -0.00000014, 0.00001542, -0.00002133, 0.00000534, -0.00000399, 0.00000672, -0.00000444, -0.00000833, 0.00000119, -0.00000739, 0.00000013, 0.00000236, -0.00000653, 0.00000013, -0.00000051, -0.00000764, 0.00000606, -0.00000685, 0.00001913, -0.00000695, -0.00000091, -0.00000586, 0.00000512, 0.00000383, 0.00000126, -0.00000021, 0.00000262, 0.00000204, -0.00000067, -0.00000310, 0.00000250, -0.00000400, -0.00000489, -0.00001300, 0.00000810, -0.00000292, 0.00000064, -0.00001011, -0.00000614, -0.00001709, -0.00002056, -0.00008714, 0.00008898, 0.00008488, 0.00010367, 0.00004596, -0.00009169, -0.00001280, -0.00001580, 0.00000146, -0.00000880, -0.00000374, -0.00000849, 0.00000020, -0.00000448, 0.00000709, -0.00000437, 0.00000309, -0.00000795, -0.00000418, -0.00000126, 0.00000177, -0.00000887, -0.00000542, 0.00000302, -0.00000134, 0.00000408, -0.00000212, -0.00000188, -0.00000834, -0.00000075, -0.00000933, -0.00000326, -0.00000587, -0.00004095, -0.00000401, 0.00005561, 0.00003073, 0.00003655, 0.00005219, -0.00001616, -0.00003383, -0.00000910, -0.00000623, -0.00000724, 0.00000118, 0.00000159, 0.00000237, -0.00000770, -0.00000997, 0.00000539, -0.00000472, -0.00000057, -0.00000489, 0.00000483, 0.00000607, -0.00000152, 0.00000061, 0.00000173, 0.00000807, -0.00000752, -0.00000419, 0.00001031, 0.00000190, -0.00000315, -0.00000013, 0.00000084, 0.00000052, -0.00000740, 0.00000076, -0.00000029, -0.00000002, -0.00000447, 0.00000399, -0.00000926, -0.00000812, 0.00000775, 0.00000684, -0.00001126, -0.00001061, 0.00000606, 0.00000869, },
		{ -0.00000308, 0.00000146, 0.00000222, 0.00000217, 0.00000547, -0.00000602, 0.00000895, 0.00000556, -0.00001966, 0.00000837, -0.00000599, 0.00001057, -0.00001288, -0.00000087, -0.00000344, -0.00000454, 0.00000015, 0.00000246, -0.00000906, 0.00000474, -0.00000668, -0.00000071, 0.00000003, 0.00000041, 0.00001546, -0.00000886, -0.00000047, -0.00000520, 0.00000767, 0.00000175, 0.00000102, 0.00000034, 0.00000272, 0.00000238, -0.00000296, -0.00000071, 0.00000019, -0.00000237, -0.00001017, -0.00000634, 0.00000617, -0.00000170, -0.00000300, -0.00000736, -0.00001244, -0.00000922, -0.00004800, -0.00005192, 0.00011656, 0.00006214, 0.00011608, -0.00000598, -0.00007569, -0.00000831, -0.00001163, -0.00000066, -0.00000750, -0.00000563, -0.00000552, -0.00000186, -0.00000051, 0.00000432, -0.00000334, 0.00000224, -0.00000975, -0.00000214, -0.00000067, 0.00000067, -0.00001126, -0.00000112, 0.00000148, -0.00000016, 0.00000326, -0.00000255, -0.00000322, -0.00000781, -0.00000046, -0.00001144, 0.00000083, -0.00001527, -0.00004320, 0.00001551, 0.00005867, 0.00002480, 0.00005067, 0.00003827, -0.00003273, -0.00002690, -0.00000536, -0.00000873, -0.00000426, 0.00000127, 0.00000230, 0.00000085, -0.00001097, -0.00000565, 0.00000490, -0.00000606, 0.00000032, -0.00000521, 0.00000906, 0.00000131, 0.00000068, -0.00000133, 0.00000596, 0.00000383, -0.00000852, -0.00000103, 0.00001142, -0.00000187, -0.00000066, -0.00000189, 0.00000293, -0.00000314, -0.00000518, 0.00000048, 0.00000077, -0.00000235, -0.00000239, 0.00000269, -0.00001232, -0.00000358, 0.00001017, 0.00000180, -0.00001340, -0.00000713, 0.00001050, 0.00000378, },
		{ -0.00000101, 0.00000148, 0.00000264, 0.00000240, 0.00000413, -0.00000696, 0.00001616, -0.00000712, -0.00001179, 0.00000555, -0.00000223, 0.00000761, -0.00001573, 0.00000410, -0.00000751, -0.00000137, 0.00000015, 0.00000108, -0.00000902, 0.00000654, -0.00001032, 0.00000464, -0.00000543, 0.00000881, 0.00000837, -0.00000753, -0.00000199, -0.00000210, 0.00000796, 0.00000022, 0.00000103, 0.00000084, 0.00000314, 0.00000161, -0.00000409, 0.00000078, -0.00000110, -0.00000261, -0.00001337, 0.00000040, 0.00000295, -0.00000032, -0.00000614, -0.00000609, -0.00001573, -0.00000741, -0.00007176, 0.00000306, 0.00011245, 0.00005782, 0.00011082, -0.00004763, -0.00005370, -0.00000836, -0.00000674, -0.00000402, -0.00000542, -0.00000756, -0.00000262, -0.00000371, 0.00000356, 0.00000015, -0.00000026, -0.00000114, -0.00000927, -0.00000091, 0.00000024, -0.00000199, -0.00001144, 0.00000248, -0.00000068, 0.00000179, 0.00000149, -0.00000230, -0.00000532, -0.00000575, -0.00000234, -0.00001102, 0.00000262, -0.00002658, -0.00003947, 0.00003667, 0.00005397, 0.00002608, 0.00006115, 0.00001646, -0.00004208, -0.00001684, -0.00000613, -0.00000870, -0.00000223, 0.00000164, 0.00000259, -0.00000149, -0.00001310, -0.00000022, 0.00000167, -0.00000482, -0.00000104, -0.00000224, 0.00000944, -0.00000111, 0.00000147, -0.00000136, 0.00000856, -0.00000086, -0.00000809, 0.00000259, 0.00001048, -0.00000402, 0.00000097, -0.00000303, 0.00000427, -0.00000652, -0.00000227, -0.00000036, 0.00000181, -0.00000501, 0.00000087, -0.00000092, -0.00001316, 0.00000182, 0.00000985, -0.00000304, -0.00001436, -0.00000177, 0.00001192, -0.00000080, },
		{ 0.00000083, 0.00000158, 0.00000255, 0.00000329, 0.00000076, -0.00000317, 0.00001748, -0.00001714, -0.00000245, 0.00000038, 0.00000450, -0.00000081, -0.00001210, 0.00000431, -0.00000875, 0.00000054, 0.00000074, -0.00000184, -0.00000624, 0.00000488, -0.00001005, 0.00000624, -0.00000740, 0.00001491, 0.00000091, -0.00000497, -0.00000397, 0.00000227, 0.00000620, -0.00000037, 0.00000107, 0.00000139, 0.00000365, -0.00000018, -0.00000375, 0.00000097, -0.00000104, -0.00000545, -0.00001258, 0.00000450, 0.00000073, -0.00000048, -0.00000722, -0.00000795, -0.00001356, -0.00001767, -0.00007667, 0.00005741, 0.00008836, 0.00006996, 0.00008549, -0.00007140, -0.00003373, -0.00000891, -0.00000377, -0.00000621, -0.00000436, -0.00000811, -0.00000116, -0.00000376, 0.00000568, -0.00000308, 0.00000283, -0.00000569, -0.00000687, -0.00000062, 0.00000122, -0.00000583, -0.00000896, 0.00000403, -0.00000206, 0.00000339, -0.00000036, -0.00000222, -0.00000707, -0.00000342, -0.00000532, -0.00000847, 0.00000115, -0.00003839, -0.00002804, 0.00005496, 0.00004455, 0.00003517, 0.00006314, -0.00000932, -0.00004172, -0.00000846, -0.00000898, -0.00000662, -0.00000137, 0.00000245, 0.00000223, -0.00000467, -0.00001292, 0.00000421, -0.00000240, -0.00000237, -0.00000327, 0.00000265, 0.00000648, -0.00000077, 0.00000031, 0.00000121, 0.00000805, -0.00000425, -0.00000709, 0.00000660, 0.00000742, -0.00000392, 0.00000083, -0.00000257, 0.00000373, -0.00000823, -0.00000001, -0.00000076, 0.00000181, -0.00000668, 0.00000396, -0.00000615, -0.00001097, 0.00000637, 0.00000742, -0.00000716, -0.00001389, 0.00000468, 0.00000968, -0.00000316, },
		{ 0.00000167, 0.00000203, 0.00000211, 0.00000414, -0.00000324, 0.00000432, 0.00001167, -0.00002067, 0.00000378, -0.00000274, 0.00000942, -0.00001021, -0.00000464, 0.00000051, -0.00000693, 0.00000085, 0.00000166, -0.00000534, -0.00000190, 0.00000081, -0.00000652, 0.00000409, -0.00000501, 0.00001674, -0.00000454, -0.00000300, -0.00000479, 0.00000628, 0.00000324, 0.00000006, 0.00000082, 0.00000239, 0.00000348, -0.00000204, -0.00000275, 0.00000070, -0.00000083, -0.00000945, -0.00000845, 0.00000549, -0.00000006, -0.00000203, -0.00000697, -0.00001132, -0.00000880, -0.00003787, -0.00005596, 0.00009288, 0.00006178, 0.00008763, 0.00004653, -0.00007704, -0.00001933, -0.00000836, -0.00000333, -0.00000647, -0.00000479, -0.00000703, -0.00000131, -0.00000175, 0.00000499, -0.00000383, 0.00000398, -0.00000957, -0.00000368, -0.00000080, 0.00000159, -0.00000969, -0.00000448, 0.00000315, -0.00000180, 0.00000378, -0.00000158, -0.00000291, -0.00000761, -0.00000204, -0.00000806, -0.00000484, -0.00000411, -0.00004843, -0.00000885, 0.00006630, 0.00003519, 0.00004974, 0.00005325, -0.00003292, -0.00003276, -0.00000509, -0.00001075, -0.00000429, -0.00000079, 0.00000317, 0.00000134, -0.00000846, -0.00000982, 0.00000577, -0.00000508, -0.00000070, -0.00000419, 0.00000678, 0.00000254, 0.00000079, -0.00000145, 0.00000510, 0.00000489, -0.00000587, -0.00000566, 0.00001028, 0.00000310, -0.00000208, -0.00000087, -0.00000058, 0.00000122, -0.00000778, 0.00000094, -0.00000025, 0.00000022, -0.00000633, 0.00000532, -0.00001143, -0.00000594, 0.00000857, 0.00000417, -0.00001077, -0.00001114, 0.00001025, 0.00000492, -0.00000286, },
		{ 0.00000169, 0.00000253, 0.00000186, 0.00000383, -0.00000544, 0.00001210, 0.00000099, -0.00001741, 0.00000511, -0.00000145, 0.00000896, -0.00001585, 0.00000242, -0.00000458, -0.00000348, 0.00000030, 0.00000205, -0.00000798, 0.00000216, -0.00000363, -0.00000177, -0.00000005, 0.00000059, 0.00001429, -0.00000710, -0.00000248, -0.00000353, 0.00000851, 0.00000029, 0.00000106, 0.00000038, 0.00000386, 0.00000209, -0.00000300, -0.00000207, 0.00000098, -0.00000211, -0.00001217, -0.00000326, 0.00000465, -0.00000036, -0.00000362, -0.00000739, -0.00001282, -0.00000779, -0.00005805, -0.00001503, 0.00010174, 0.00004652, 0.00009837, 0.00000460, -0.00006939, -0.00001046, -0.00000718, -0.00000435, -0.00000553, -0.00000605, -0.00000506, -0.00000218, 0.00000130, 0.00000215, -0.00000195, 0.00000222, -0.00001120, -0.00000102, -0.00000077, 0.00000053, -0.00001203, 0.00000038, 0.00000068, -0.00000004, 0.00000282, -0.00000205, -0.00000434, -0.00000685, -0.00000214, -0.00000956, -0.00000142, -0.00001325, -0.00005377, 0.00001595, 0.00006852, 0.00003084, 0.00006471, 0.00003113, -0.00004779, -0.00001993, -0.00000667, -0.00000992, -0.00000304, 0.00000029, 0.00000331, -0.00000018, -0.00001190, -0.00000456, 0.00000410, -0.00000531, -0.00000088, -0.00000251, 0.00000807, 0.00000002, 0.00000155, -0.00000185, 0.00000806, 0.00000080, -0.00000634, -0.00000333, 0.00001247, -0.00000101, 0.00000024, -0.00000307, 0.00000193, -0.00000237, -0.00000563, 0.00000071, 0.00000091, -0.00000268, -0.00000370, 0.00000375, -0.00001471, 0.00000046, 0.00000803, 0.00000111, -0.00001403, -0.00000556, 0.00001273, -0.00000000, -0.00000092, },
		{ 0.00000161, 0.00000263, 0.00000220, 0.00000177, -0.00000389, 0.00001618, -0.00001003, -0.00001030, 0.00000304, 0.00000322, 0.00000280, -0.00001537, 0.00000577, -0.00000811, -0.00000031, -0.00000012, 0.00000116, -0.00000864, 0.00000442, -0.00000657, 0.00000201, -0.00000378, 0.00000707, 0.00000905, -0.00000721, -0.00000313, -0.00000027, 0.00000819, -0.00000146, 0.00000182, 0.00000037, 0.00000512, -0.00000012, -0.00000295, -0.00000188, 0.00000191, -0.00000540, -0.00001196, 0.00000081, 0.00000348, -0.00000111, -0.00000438, -0.00000947, -0.00001066, -0.00001556, -0.00006598, 0.00003168, 0.00008856, 0.00004715, 0.00009444, -0.00003069, -0.00005478, -0.00000559, -0.00000640, -0.00000535, -0.00000459, -0.00000708, -0.00000325, -0.00000259, 0.00000382, -0.00000117, 0.00000129, -0.00000200, -0.00001006, 0.00000027, -0.00000014, -0.00000232, -0.00001160, 0.00000372, -0.00000179, 0.00000222, 0.00000106, -0.00000217, -0.00000593, -0.00000537, -0.00000346, -0.00000945, 0.00000057, -0.00002569, -0.00005145, 0.00004225, 0.00006223, 0.00003508, 0.00007319, 0.00000076, -0.00005000, -0.00000926, -0.00001027, -0.00000725, -0.00000277, 0.00000180, 0.00000290, -0.00000268, -0.00001350, 0.00000083, 0.00000059, -0.00000377, -0.00000228, 0.00000117, 0.00000634, -0.00000019, 0.00000064, 0.00000013, 0.00000840, -0.00000245, -0.00000655, 0.00000030, 0.00001225, -0.00000358, 0.00000173, -0.00000458, 0.00000376, -0.00000576, -0.00000282, -0.00000000, 0.00000191, -0.00000587, 0.00000044, -0.00000096, -0.00001425, 0.00000589, 0.00000580, -0.00000176, -0.00001621, 0.00000210, 0.00001104, -0.00000287, 0.00000108, },
		{ 0.00000189, 0.00000225, 0.00000283, -0.00000136, 0.00000146, 0.00001422, -0.00001716, -0.00000331, 0.00000084, 0.00000782, -0.00000610, -0.00000970, 0.00000454, -0.00000863, 0.00000139, 0.00000010, -0.00000106, -0.00000711, 0.00000435, -0.00000713, 0.00000351, -0.00000525, 0.00001213, 0.00000303, -0.00000608, -0.00000393, 0.00000394, 0.00000557, -0.00000144, 0.00000167, 0.00000148, 0.00000509, -0.00000203, -0.00000266, -0.00000144, 0.00000232, -0.00000952, -0.00000901, 0.00000296, 0.00000252, -0.00000225, -0.00000481, -0.00001202, -0.00000680, -0.00003149, -0.00005426, 0.00006840, 0.00006569, 0.00005925, 0.00007515, -0.00005382, -0.00003840, -0.00000332, -0.00000645, -0.00000546, -0.00000446, -0.00000719, -0.00000224, -0.00000180, 0.00000458, -0.00000319, 0.00000392, -0.00000706, -0.00000681, 0.00000021, 0.00000079, -0.00000637, -0.00000827, 0.00000438, -0.00000280, 0.00000375, -0.00000067, -0.00000251, -0.00000697, -0.00000402, -0.00000523, -0.00000802, 0.00000001, -0.00003985, -0.00003947, 0.00006495, 0.00005092, 0.00004839, 0.00006861, -0.00002983, -0.00004062, -0.00000470, -0.00001244, -0.00000465, -0.00000251, 0.00000310, 0.00000214, -0.00000623, -0.00001216, 0.00000416, -0.00000260, -0.00000212, -0.00000322, 0.00000488, 0.00000318, 0.00000100, -0.00000115, 0.00000382, 0.00000596, -0.00000405, -0.00000681, 0.00000500, 0.00000950, -0.00000398, 0.00000163, -0.00000462, 0.00000403, -0.00000786, -0.00000038, -0.00000044, 0.00000183, -0.00000791, 0.00000436, -0.00000759, -0.00000957, 0.00000828, 0.00000360, -0.00000531, -0.00001573, 0.00000958, 0.00000608, -0.00000284, 0.00000206, },
		{ 0.00000240, 0.00000180, 0.00000287, -0.00000371, 0.00000836, 0.00000683, -0.00001857, 0.00000092, 0.00000104, 0.00000887, -0.00001348, -0.00000211, 0.00000023, -0.00000643, 0.00000149, 0.00000079, -0.00000399, -0.00000402, 0.00000238, -0.00000556, 0.00000262, -0.00000383, 0.00001433, -0.00000206, -0.00000491, -0.00000367, 0.00000747, 0.00000189, -0.00000006, 0.00000072, 0.00000370, 0.00000325, -0.00000269, -0.00000296, 0.00000009, 0.00000079, -0.00001240, -0.00000491, 0.00000372, 0.00000148, -0.00000306, -0.00000616, -0.00001263, -0.00000597, -0.00004868, -0.00002453, 0.00008549, 0.00004603, 0.00007373, 0.00004535, -0.00006373, -0.00002353, -0.00000292, -0.00000698, -0.00000474, -0.00000521, -0.00000631, -0.00000201, 0.00000009, 0.00000329, -0.00000289, 0.00000422, -0.00001089, -0.00000297, -0.00000046, 0.00000112, -0.00001017, -0.00000321, 0.00000251, -0.00000178, 0.00000376, -0.00000174, -0.00000344, -0.00000707, -0.00000345, -0.00000666, -0.00000591, -0.00000397, -0.00005319, -0.00001756, 0.00007921, 0.00004044, 0.00006693, 0.00004777, -0.00005135, -0.00002550, -0.00000614, -0.00001164, -0.00000338, -0.00000153, 0.00000376, 0.00000090, -0.00001004, -0.00000812, 0.00000447, -0.00000404, -0.00000162, -0.00000236, 0.00000667, 0.00000070, 0.00000191, -0.00000216, 0.00000732, 0.00000203, -0.00000437, -0.00000667, 0.00000981, 0.00000495, -0.00000241, -0.00000008, -0.00000310, 0.00000250, -0.00000814, 0.00000105, -0.00000018, 0.00000014, -0.00000759, 0.00000600, -0.00001363, -0.00000213, 0.00000710, 0.00000259, -0.00001011, -0.00001122, 0.00001396, 0.00000039, -0.00000078, 0.00000205, },
		},
		{ { 0.00000444, 0.00000182, -0.00000242, -0.00000548, 0.00000587, -0.00000252, 0.00001576, -0.00000303, -0.00000225, -0.00000081, 0.00000365, 0.00000034, 0.00000136, 0.00000368, -0.00000759, 0.00000056, 0.00000719, -0.00000151, -0.00000468, 0.00000033, -0.00000615, 0.00000239, -0.00001511, 0.00000602, 0.00000225, -0.00000225, -0.00000694, 0.00000036, -0.00001467, 0.00000262, 0.00000093, -0.00000116, -0.00000189, -0.00000288, 0.00000721, -0.00000201, 0.00000219, 0.00000128, -0.00000408, -0.00000333, -0.00000349, -0.00000215, 0.00000133, -0.00000384, -0.00000159, -0.00000802, -0.00000365, -0.00001191, -0.00005855, 0.00000888, 0.00008342, 0.00003760, 0.00008525, 0.00001182, -0.00006194, -0.00001126, -0.00000877, -0.00000499, -0.00000574, -0.00000467, -0.00000683, -0.00000201, 0.00000121, -0.00000116, -0.00000158, -0.00000116, 0.00000121, -0.00000201, -0.00000683, -0.00000467, -0.00000574, -0.00000499, -0.00000877, -0.00001126, -0.00006194, 0.00001182, 0.00008525, 0.00003760, 0.00008342, 0.00000888, -0.00005855, -0.00001191, -0.00000365, -0.00000802, -0.00000159, -0.00000384, 0.00000133, -0.00000215, -0.00000349, -0.00000333, -0.00000408, 0.00000128, 0.00000219, -0.00000201, 0.00000721, -0.00000288, -0.00000189, -0.00000116, 0.00000093, 0.00000262, -0.00001467, 0.00000036, -0.00000694, -0.00000225, 0.00000225, 0.00000602, -0.00001511, 0.00000239, -0.00000615, 0.00000033, -0.00000468, -0.00000151, 0.00000719, 0.00000056, -0.00000759, 0.00000368, 0.00000136, 0.00000034, 0.00000365, -0.00000081, -0.00000225, -0.00000303, 0.00001576, -0.00000252, 0.00000587, -0.00000548, -0.00000242, 0.00000182, },
		{ 0.00000423, 0.00000137, -0.00000413, -0.00000295, 0.00000445, -0.00000001, 0.00001572, -0.00000696, 0.00000012, -0.00000184, 0.00000598, -0.00000249, 0.00000443, 0.00000043, -0.00000694, 0.00000202, 0.00000774, -0.00000505, -0.00000168, -0.00000260, -0.00000258, -0.00000230, -0.00001257, 0.00000849, 0.00000012, -0.00000309, -0.00000590, -0.00000096, -0.00001469, 0.00000571, -0.00000134, 0.00000046, -0.00000419, -0.00000028, 0.00000587, -0.00000165, 0.00000265, 0.00000020, -0.00000471, -0.00000234, -0.00000445, -0.00000053, 0.00000027, -0.00000363, -0.00000198, -0.00000903, -0.00000100, -0.00002249, -0.00005540, 0.00003675, 0.00007181, 0.00003857, 0.00008362, -0.00001144, -0.00005398, -0.00000721, -0.00000867, -0.00000462, -0.00000596, -0.00000453, -0.00000688, -0.00000033, 0.00000073, -0.00000148, -0.00000159, -0.00000060, 0.00000115, -0.00000357, -0.00000651, -0.00000505, -0.00000512, -0.00000603, -0.00000825, -0.00001794, -0.00006598, 0.00003794, 0.00008011, 0.00004401, 0.00008688, -0.00002152, -0.00005233, -0.00000536, -0.00000605, -0.00000661, -0.00000171, -0.00000354, 0.00000178, -0.00000340, -0.00000260, -0.00000452, -0.00000285, 0.00000202, 0.00000140, -0.00000184, 0.00000810, -0.00000527, 0.00000061, -0.00000323, 0.00000373, -0.00000168, -0.00001299, 0.00000043, -0.00000720, -0.00000165, 0.00000487, 0.00000123, -0.00001426, 0.00000481, -0.00000877, 0.00000284, -0.00000746, 0.00000258, 0.00000549, -0.00000056, -0.00000744, 0.00000609, -0.00000138, 0.00000326, 0.00000064, 0.00000097, -0.00000515, 0.00000253, 0.00001309, -0.00000297, 0.00000596, -0.00000718, -0.00000081, 0.00000237, },
		{ 0.00000365, 0.00000094, -0.00000560, -0.00000015, 0.00000236, 0.00000399, 0.00001300, -0.00000865, 0.00000128, -0.00000151, 0.00000682, -0.00000430, 0.00000699, -0.00000310, -0.00000562, 0.00000375, 0.00000683, -0.00000727, 0.00000057, -0.00000478, 0.00000041, -0.00000768, -0.00000754, 0.00000850, -0.00000114, -0.00000422, -0.00000427, -0.00000332, -0.00001307, 0.00000739, -0.00000279, 0.00000132, -0.00000591, 0.00000222, 0.00000435, -0.00000091, 0.00000270, -0.00000105, -0.00000471, -0.00000182, -0.00000516, 0.00000123, -0.00000137, -0.00000268, -0.00000337, -0.00000862, -0.00000007, -0.00003454, -0.00004270, 0.00005741, 0.00005689, 0.00004494, 0.00007501, -0.00002974, -0.00004415, -0.00000517, -0.00000811, -0.00000485, -0.00000568, -0.00000478, -0.00000640, 0.00000114, -0.00000011, -0.00000157, -0.00000157, 0.00000008, 0.00000054, -0.00000479, -0.00000617, -0.00000541, -0.00000434, -0.00000761, -0.00000724, -0.00002754, -0.00006403, 0.00006394, 0.00007008, 0.00005786, 0.00007821, -0.00004830, -0.00003963, -0.00000342, -0.00000703, -0.00000560, -0.00000181, -0.00000304, 0.00000176, -0.00000418, -0.00000209, -0.00000554, -0.00000125, 0.00000236, 0.00000033, -0.00000098, 0.00000822, -0.00000700, 0.00000279, -0.00000523, 0.00000650, -0.00000665, -0.00000990, -0.00000070, -0.00000669, -0.00000105, 0.00000719, -0.00000499, -0.00001012, 0.00000406, -0.00000919, 0.00000375, -0.00000888, 0.00000613, 0.00000329, -0.00000148, -0.00000640, 0.00000724, -0.00000308, 0.00000534, -0.00000204, 0.00000258, -0.00000757, 0.00000859, 0.00000840, -0.00000141, 0.00000447, -0.00000770, 0.00000049, 0.00000302, },
		{ 0.00000285, 0.00000041, -0.00000650, 0.00000231, 0.00000044, 0.00000844, 0.00000832, -0.00000824, 0.00000118, 0.00000029, 0.00000576, -0.00000441, 0.00000829, -0.00000629, -0.00000377, 0.00000548, 0.00000461, -0.00000789, 0.00000160, -0.00000556, 0.00000161, -0.00001198, -0.00000147, 0.00000660, -0.00000157, -0.00000551, -0.00000235, -0.00000632, -0.00001008, 0.00000768, -0.00000335, 0.00000125, -0.00000678, 0.00000433, 0.00000290, -0.00000001, 0.00000234, -0.00000229, -0.00000414, -0.00000197, -0.00000524, 0.00000264, -0.00000320, -0.00000122, -0.00000576, -0.00000632, -0.00000264, -0.00004451, -0.00002285, 0.00006857, 0.00004299, 0.00005381, 0.00006052, -0.00004202, -0.00003423, -0.00000433, -0.00000744, -0.00000535, -0.00000504, -0.00000542, -0.00000526, 0.00000209, -0.00000105, -0.00000148, -0.00000144, 0.00000066, -0.00000048, -0.00000560, -0.00000593, -0.00000559, -0.00000369, -0.00000945, -0.00000618, -0.00003980, -0.00005435, 0.00008629, 0.00005866, 0.00007658, 0.00005586, -0.00006562, -0.00002511, -0.00000469, -0.00000667, -0.00000510, -0.00000179, -0.00000246, 0.00000131, -0.00000439, -0.00000222, -0.00000594, 0.00000031, 0.00000246, -0.00000104, 0.00000075, 0.00000721, -0.00000757, 0.00000403, -0.00000646, 0.00000848, -0.00001146, -0.00000607, -0.00000267, -0.00000560, -0.00000011, 0.00000824, -0.00001107, -0.00000390, 0.00000036, -0.00000695, 0.00000240, -0.00000808, 0.00000812, 0.00000139, -0.00000246, -0.00000444, 0.00000694, -0.00000333, 0.00000593, -0.00000354, 0.00000317, -0.00000842, 0.00001368, 0.00000294, 0.00000150, 0.00000163, -0.00000700, 0.00000137, 0.00000367, },
		{ 0.00000201, -0.00000035, -0.00000658, 0.00000396, -0.00000052, 0.00001215, 0.00000285, -0.00000641, 0.00000030, 0.00000304, 0.00000305, -0.00000266, 0.00000786, -0.00000859, -0.00000157, 0.00000683, 0.00000157, -0.00000712, 0.00000147, -0.00000493, 0.00000053, -0.00001391, 0.00000408, 0.00000372, -0.00000150, -0.00000667, -0.00000056, -0.00000948, -0.00000613, 0.00000674, -0.00000306, 0.00000025, -0.00000663, 0.00000580, 0.00000178, 0.00000079, 0.00000169, -0.00000338, -0.00000314, -0.00000286, -0.00000432, 0.00000311, -0.00000449, -0.00000001, -0.00000835, -0.00000274, -0.00000925, -0.00004916, 0.00000025, 0.00007034, 0.00003311, 0.00006214, 0.00004219, -0.00004824, -0.00002541, -0.00000401, -0.00000703, -0.00000572, -0.00000437, -0.00000622, -0.00000351, 0.00000230, -0.00000180, -0.00000138, -0.00000109, 0.00000091, -0.00000160, -0.00000617, -0.00000574, -0.00000557, -0.00000337, -0.00001125, -0.00000570, -0.00005385, -0.00003597, 0.00010157, 0.00005043, 0.00009495, 0.00002199, -0.00006997, -0.00001347, -0.00000669, -0.00000601, -0.00000455, -0.00000199, -0.00000160, 0.00000036, -0.00000395, -0.00000320, -0.00000530, 0.00000132, 0.00000256, -0.00000272, 0.00000346, 0.00000477, -0.00000654, 0.00000372, -0.00000623, 0.00000880, -0.00001509, -0.00000238, -0.00000488, -0.00000429, 0.00000146, 0.00000711, -0.00001520, 0.00000243, -0.00000492, -0.00000272, -0.00000102, -0.00000495, 0.00000799, 0.00000037, -0.00000369, -0.00000168, 0.00000532, -0.00000216, 0.00000483, -0.00000346, 0.00000228, -0.00000698, 0.00001650, -0.00000182, 0.00000470, -0.00000191, -0.00000534, 0.00000191, 0.00000418, },
		{ 0.00000133, -0.00000141, -0.00000576, 0.00000457, -0.00000003, 0.00001414, -0.00000217, -0.00000413, -0.00000055, 0.00000583, -0.00000052, 0.00000055, 0.00000563, -0.00000963, 0.00000076, 0.00000747, -0.00000169, -0.00000544, 0.00000066, -0.00000349, -0.00000249, -0.00001293, 0.00000788, 0.00000084, -0.00000139, -0.00000735, 0.00000069, -0.00001228, -0.00000174, 0.00000489, -0.00000212, -0.00000155, -0.00000542, 0.00000648, 0.00000116, 0.00000124, 0.00000093, -0.00000426, -0.00000191, -0.00000439, -0.00000227, 0.00000214, -0.00000449, -0.00000004, -0.00000985, 0.00000046, -0.00001876, -0.00004679, 0.00002250, 0.00006461, 0.00002871, 0.00006733, 0.00002251, -0.00004928, -0.00001826, -0.00000385, -0.00000707, -0.00000568, -0.00000396, -0.00000686, -0.00000137, 0.00000171, -0.00000215, -0.00000139, -0.00000045, 0.00000059, -0.00000251, -0.00000672, -0.00000539, -0.00000550, -0.00000344, -0.00001273, -0.00000668, -0.00006802, -0.00000918, 0.00010740, 0.00004981, 0.00010599, -0.00001745, -0.00006159, -0.00000765, -0.00000723, -0.00000619, -0.00000330, -0.00000283, -0.00000024, -0.00000115, -0.00000291, -0.00000496, -0.00000348, 0.00000150, 0.00000279, -0.00000458, 0.00000696, 0.00000097, -0.00000381, 0.00000160, -0.00000411, 0.00000686, -0.00001665, 0.00000026, -0.00000668, -0.00000303, 0.00000365, 0.00000343, -0.00001592, 0.00000667, -0.00000973, 0.00000198, -0.00000535, -0.00000021, 0.00000595, 0.00000037, -0.00000511, 0.00000152, 0.00000280, 0.00000004, 0.00000237, -0.00000193, 0.00000000, -0.00000314, 0.00001630, -0.00000467, 0.00000703, -0.00000531, -0.00000317, 0.00000226, 0.00000437, },
		{ 0.00000089, -0.00000272, -0.00000417, 0.00000425, 0.00000191, 0.00001388, -0.00000577, -0.00000226, -0.00000055, 0.00000759, -0.00000382, 0.00000436, 0.00000202, -0.00000927, 0.00000297, 0.00000722, -0.00000457, -0.00000339, -0.00000030, -0.00000214, -0.00000644, -0.00000933, 0.00000930, -0.00000125, -0.00000172, -0.00000721, 0.00000095, -0.00001413, 0.00000249, 0.00000258, -0.00000090, -0.00000383, -0.00000331, 0.00000641, 0.00000104, 0.00000122, 0.00000018, -0.00000479, -0.00000083, -0.00000609, 0.00000058, -0.00000019, -0.00000294, -0.00000192, -0.00000909, 0.00000126, -0.00002868, -0.00003762, 0.00004065, 0.00005416, 0.00002981, 0.00006763, 0.00000398, -0.00004661, -0.00001269, -0.00000384, -0.00000742, -0.00000516, -0.00000402, -0.00000703, 0.00000079, 0.00000046, -0.00000201, -0.00000155, 0.00000044, -0.00000033, -0.00000304, -0.00000742, -0.00000475, -0.00000552, -0.00000396, -0.00001358, -0.00001016, -0.00007978, 0.00002398, 0.00010349, 0.00005940, 0.00010295, -0.00005396, -0.00004457, -0.00000771, -0.00000556, -0.00000759, -0.00000120, -0.00000434, 0.00000156, -0.00000303, -0.00000165, -0.00000703, -0.00000084, 0.00000094, 0.00000296, -0.00000611, 0.00001061, -0.00000365, 0.00000018, -0.00000203, -0.00000029, 0.00000258, -0.00001572, 0.00000123, -0.00000759, -0.00000192, 0.00000602, -0.00000235, -0.00001275, 0.00000729, -0.00001210, 0.00000524, -0.00000890, 0.00000471, 0.00000287, 0.00000101, -0.00000636, 0.00000459, 0.00000004, 0.00000254, -0.00000072, 0.00000041, -0.00000296, 0.00000243, 0.00001310, -0.00000495, 0.00000763, -0.00000773, -0.00000101, 0.00000260, 0.00000414, },
		{ 0.00000067, -0.00000411, -0.00000209, 0.00000336, 0.00000486, 0.00001148, -0.00000753, -0.00000128, 0.00000072, 0.00000754, -0.00000574, 0.00000766, -0.00000224, -0.00000768, 0.00000484, 0.00000605, -0.00000670, -0.00000136, -0.00000105, -0.00000164, -0.00001009, -0.00000406, 0.00000841, -0.00000217, -0.00000276, -0.00000604, -0.00000010, -0.00001452, 0.00000589, 0.00000040, 0.00000010, -0.00000604, -0.00000070, 0.00000583, 0.00000120, 0.00000086, -0.00000059, -0.00000480, -0.00000038, -0.00000723, 0.00000334, -0.00000315, -0.00000038, -0.00000534, -0.00000590, -0.00000172, -0.00003613, -0.00002347, 0.00005285, 0.00004198, 0.00003524, 0.00006242, -0.00001139, -0.00004185, -0.00000829, -0.00000426, -0.00000773, -0.00000437, -0.00000455, -0.00000649, 0.00000255, -0.00000113, -0.00000148, -0.00000177, 0.00000132, -0.00000161, -0.00000329, -0.00000812, -0.00000395, -0.00000557, -0.00000511, -0.00001345, -0.00001734, -0.00008582, 0.00005957, 0.00009234, 0.00007835, 0.00008186, -0.00007898, -0.00002553, -0.00001109, -0.00000273, -0.00000942, 0.00000104, -0.00000592, 0.00000325, -0.00000470, -0.00000085, -0.00000853, 0.00000180, 0.00000022, 0.00000256, -0.00000649, 0.00001332, -0.00000795, 0.00000439, -0.00000623, 0.00000439, -0.00000334, -0.00001255, 0.00000047, -0.00000748, -0.00000079, 0.00000768, -0.00000888, -0.00000650, 0.00000398, -0.00001092, 0.00000565, -0.00001011, 0.00000820, 0.00000002, 0.00000153, -0.00000684, 0.00000684, -0.00000222, 0.00000452, -0.00000351, 0.00000265, -0.00000553, 0.00000846, 0.00000778, -0.00000281, 0.00000618, -0.00000867, 0.00000071, 0.00000303, 0.00000350, },
		{ 0.00000053, -0.00000531, 0.00000009, 0.00000247, 0.00000803, 0.00000758, -0.00000762, -0.00000110, 0.00000313, 0.00000547, -0.00000557, 0.00000946, -0.00000623, -0.00000522, 0.00000623, 0.00000409, -0.00000783, 0.00000041, -0.00000150, -0.00000241, -0.00001227, 0.00000157, 0.00000581, -0.00000197, -0.00000447, -0.00000385, -0.00000256, -0.00001309, 0.00000790, -0.00000113, 0.00000034, -0.00000756, 0.00000186, 0.00000514, 0.00000126, 0.00000047, -0.00000153, -0.00000409, -0.00000102, -0.00000698, 0.00000489, -0.00000544, 0.00000188, -0.00000898, -0.00000140, -0.00000838, -0.00003901, -0.00000689, 0.00005848, 0.00003083, 0.00004298, 0.00005222, -0.00002251, -0.00003636, -0.00000463, -0.00000531, -0.00000760, -0.00000362, -0.00000538, -0.00000513, 0.00000347, -0.00000256, -0.00000090, -0.00000175, 0.00000177, -0.00000278, -0.00000368, -0.00000843, -0.00000341, -0.00000536, -0.00000719, -0.00001199, -0.00002920, -0.00008263, 0.00009234, 0.00007911, 0.00010166, 0.00004362, -0.00008687, -0.00001108, -0.00001396, -0.00000086, -0.00001013, 0.00000215, -0.00000647, 0.00000393, -0.00000539, -0.00000129, -0.00000851, 0.00000342, 0.00000010, 0.00000094, -0.00000481, 0.00001388, -0.00001058, 0.00000740, -0.00000961, 0.00000852, -0.00000960, -0.00000807, -0.00000157, -0.00000659, 0.00000069, 0.00000760, -0.00001423, 0.00000085, -0.00000206, -0.00000649, 0.00000285, -0.00000818, 0.00000912, -0.00000148, 0.00000118, -0.00000603, 0.00000765, -0.00000332, 0.00000528, -0.00000519, 0.00000393, -0.00000663, 0.00001343, 0.00000179, 0.00000090, 0.00000297, -0.00000801, 0.00000182, 0.00000352, 0.00000261, },
		{ 0.00000024, -0.00000604, 0.00000197, 0.00000209, 0.00001054, 0.00000315, -0.00000667, -0.00000119, 0.00000598, 0.00000187, -0.00000324, 0.00000915, -0.00000913, -0.00000241, 0.00000708, 0.00000151, -0.00000790, 0.00000182, -0.00000186, -0.00000435, -0.00001229, 0.00000628, 0.00000248, -0.00000114, -0.00000647, -0.00000106, -0.00000616, -0.00000984, 0.00000823, -0.00000164, -0.00000053, -0.00000789, 0.00000386, 0.00000470, 0.00000087, 0.00000036, -0.00000274, -0.00000263, -0.00000290, -0.00000486, 0.00000431, -0.00000583, 0.00000232, -0.00001099, 0.00000222, -0.00001697, -0.00003674, 0.00000978, 0.00005785, 0.00002292, 0.00005050, 0.00003868, -0.00002933, -0.00003085, -0.00000163, -0.00000696, -0.00000684, -0.00000320, -0.00000627, -0.00000302, 0.00000325, -0.00000334, -0.00000062, -0.00000122, 0.00000138, -0.00000332, -0.00000468, -0.00000780, -0.00000365, -0.00000450, -0.00001044, -0.00000920, -0.00004599, -0.00006740, 0.00011687, 0.00007034, 0.00012084, -0.00000533, -0.00007705, -0.00000531, -0.00001327, -0.00000182, -0.00000833, 0.00000098, -0.00000506, 0.00000287, -0.00000459, -0.00000340, -0.00000637, 0.00000326, 0.00000114, -0.00000222, -0.00000060, 0.00001139, -0.00001039, 0.00000796, -0.00001080, 0.00001063, -0.00001463, -0.00000358, -0.00000403, -0.00000532, 0.00000273, 0.00000501, -0.00001652, 0.00000677, -0.00000854, -0.00000042, -0.00000216, -0.00000352, 0.00000725, -0.00000112, -0.00000041, -0.00000368, 0.00000674, -0.00000296, 0.00000448, -0.00000530, 0.00000377, -0.00000553, 0.00001601, -0.00000327, 0.00000486, -0.00000115, -0.00000610, 0.00000236, 0.00000394, 0.00000170, },
		{ -0.00000038, -0.00000606, 0.00000332, 0.00000253, 0.00001169, -0.00000085, -0.00000543, -0.00000082, 0.00000823, -0.00000220, 0.00000063, 0.00000675, -0.00001046, 0.00000029, 0.00000737, -0.00000141, -0.00000691, 0.00000279, -0.00000242, -0.00000697, -0.00001001, 0.00000909, -0.00000054, -0.00000039, -0.00000809, 0.00000162, -0.00001022, -0.00000523, 0.00000701, -0.00000117, -0.00000250, -0.00000685, 0.00000508, 0.00000462, -0.00000006, 0.00000061, -0.00000411, -0.00000077, -0.00000562, -0.00000114, 0.00000146, -0.00000380, 0.00000005, -0.00000992, 0.00000270, -0.00002479, -0.00003021, 0.00002486, 0.00005190, 0.00001975, 0.00005534, 0.00002405, -0.00003271, -0.00002541, 0.00000046, -0.00000880, -0.00000554, -0.00000324, -0.00000692, -0.00000041, 0.00000185, -0.00000317, -0.00000093, -0.00000002, -0.00000004, -0.00000293, -0.00000656, -0.00000598, -0.00000500, -0.00000290, -0.00001470, -0.00000586, -0.00006645, -0.00003928, 0.00012944, 0.00007163, 0.00012633, -0.00005428, -0.00005445, -0.00000827, -0.00000826, -0.00000600, -0.00000383, -0.00000260, -0.00000155, 0.00000003, -0.00000244, -0.00000684, -0.00000233, 0.00000134, 0.00000320, -0.00000650, 0.00000579, 0.00000587, -0.00000703, 0.00000547, -0.00000905, 0.00000970, -0.00001718, -0.00000029, -0.00000610, -0.00000400, 0.00000516, -0.00000007, -0.00001469, 0.00000908, -0.00001291, 0.00000492, -0.00000741, 0.00000237, 0.00000346, 0.00000077, -0.00000299, -0.00000010, 0.00000430, -0.00000126, 0.00000225, -0.00000387, 0.00000222, -0.00000217, 0.00001548, -0.00000610, 0.00000774, -0.00000512, -0.00000355, 0.00000255, 0.00000411, 0.00000100, },
		{ -0.00000144, -0.00000530, 0.00000405, 0.00000380, 0.00001120, -0.00000375, -0.00000449, 0.00000054, 0.00000891, -0.00000545, 0.00000491, 0.00000284, -0.00001015, 0.00000260, 0.00000707, -0.00000437, -0.00000494, 0.00000319, -0.00000343, -0.00000952, -0.00000589, 0.00000955, -0.00000243, -0.00000043, -0.00000859, 0.00000325, -0.00001372, -0.00000017, 0.00000483, -0.00000020, -0.00000512, -0.00000470, 0.00000561, 0.00000466, -0.00000125, 0.00000097, -0.00000522, 0.00000077, -0.00000817, 0.00000311, -0.00000270, -0.00000002, -0.00000454, -0.00000564, -0.00000119, -0.00002931, -0.00002116, 0.00003729, 0.00004211, 0.00002173, 0.00005567, 0.00001063, -0.00003395, -0.00001976, 0.00000116, -0.00001021, -0.00000407, -0.00000371, -0.00000706, 0.00000221, -0.00000043, -0.00000211, -0.00000173, 0.00000161, -0.00000225, -0.00000183, -0.00000904, -0.00000327, -0.00000716, -0.00000107, -0.00001904, -0.00000383, -0.00008739, -0.00000033, 0.00012961, 0.00008513, 0.00011082, -0.00009138, -0.00002789, -0.00001612, -0.00000093, -0.00001176, 0.00000197, -0.00000742, 0.00000305, -0.00000370, 0.00000008, -0.00001044, 0.00000243, -0.00000142, 0.00000535, -0.00001063, 0.00001294, -0.00000156, -0.00000126, 0.00000042, -0.00000460, 0.00000567, -0.00001675, 0.00000113, -0.00000728, -0.00000265, 0.00000728, -0.00000660, -0.00000907, 0.00000690, -0.00001335, 0.00000737, -0.00001069, 0.00000744, -0.00000067, 0.00000313, -0.00000570, 0.00000389, 0.00000101, 0.00000119, -0.00000085, -0.00000144, -0.00000011, 0.00000277, 0.00001201, -0.00000610, 0.00000857, -0.00000796, -0.00000102, 0.00000266, 0.00000389, 0.00000066, },
		{ -0.00000283, -0.00000383, 0.00000433, 0.00000556, 0.00000924, -0.00000538, -0.00000398, 0.00000297, 0.00000752, -0.00000681, 0.00000833, -0.00000162, -0.00000854, 0.00000440, 0.00000609, -0.00000694, -0.00000222, 0.00000286, -0.00000496, -0.00001121, -0.00000079, 0.00000785, -0.00000282, -0.00000167, -0.00000744, 0.00000303, -0.00001558, 0.00000427, 0.00000256, 0.00000054, -0.00000761, -0.00000203, 0.00000588, 0.00000435, -0.00000216, 0.00000097, -0.00000551, 0.00000115, -0.00000928, 0.00000624, -0.00000644, 0.00000375, -0.00000968, 0.00000028, -0.00000880, -0.00002933, -0.00001117, 0.00004625, 0.00003052, 0.00002800, 0.00005088, -0.00000002, -0.00003415, -0.00001368, 0.00000015, -0.00001061, -0.00000284, -0.00000449, -0.00000642, 0.00000420, -0.00000290, -0.00000065, -0.00000259, 0.00000304, -0.00000450, -0.00000077, -0.00001123, -0.00000063, -0.00000917, -0.00000018, -0.00002189, -0.00000586, -0.00010379, 0.00004435, 0.00012098, 0.00010772, 0.00007257, -0.00010768, -0.00000676, -0.00002310, 0.00000504, -0.00001618, 0.00000655, -0.00001128, 0.00000689, -0.00000675, 0.00000146, -0.00001251, 0.00000618, -0.00000351, 0.00000611, -0.00001276, 0.00001871, -0.00000887, 0.00000516, -0.00000572, 0.00000129, -0.00000050, -0.00001377, 0.00000070, -0.00000753, -0.00000099, 0.00000801, -0.00001271, -0.00000141, 0.00000101, -0.00000961, 0.00000590, -0.00001050, 0.00000988, -0.00000344, 0.00000465, -0.00000742, 0.00000717, -0.00000215, 0.00000356, -0.00000395, 0.00000120, -0.00000224, 0.00000805, 0.00000657, -0.00000356, 0.00000708, -0.00000906, 0.00000095, 0.00000286, 0.00000329, 0.00000065, },
		{ -0.00000431, -0.00000191, 0.00000442, 0.00000729, 0.00000639, -0.00000599, -0.00000361, 0.00000597, 0.00000426, -0.00000581, 0.00000992, -0.00000561, -0.00000619, 0.00000578, 0.00000433, -0.00000867, 0.00000086, 0.00000169, -0.00000687, -0.00001144, 0.00000416, 0.00000469, -0.00000193, -0.00000401, -0.00000461, 0.00000060, -0.00001509, 0.00000715, 0.00000102, 0.00000032, -0.00000914, 0.00000048, 0.00000629, 0.00000321, -0.00000224, 0.00000017, -0.00000460, -0.00000022, -0.00000800, 0.00000678, -0.00000790, 0.00000538, -0.00001291, 0.00000505, -0.00001770, -0.00002539, -0.00000104, 0.00005091, 0.00001967, 0.00003646, 0.00004181, -0.00000743, -0.00003382, -0.00000740, -0.00000242, -0.00000979, -0.00000214, -0.00000544, -0.00000479, 0.00000491, -0.00000468, 0.00000048, -0.00000282, 0.00000345, -0.00000581, -0.00000079, -0.00001195, 0.00000059, -0.00000970, -0.00000167, -0.00002150, -0.00001484, -0.00010992, 0.00008767, 0.00011040, 0.00013103, 0.00001705, -0.00010035, 0.00000245, -0.00002428, 0.00000618, -0.00001642, 0.00000744, -0.00001201, 0.00000813, -0.00000763, 0.00000042, -0.00001161, 0.00000733, -0.00000351, 0.00000418, -0.00001123, 0.00002096, -0.00001375, 0.00001008, -0.00001094, 0.00000673, -0.00000713, -0.00000943, -0.00000094, -0.00000714, 0.00000126, 0.00000634, -0.00001627, 0.00000561, -0.00000630, -0.00000310, 0.00000108, -0.00000668, 0.00000891, -0.00000378, 0.00000433, -0.00000721, 0.00000872, -0.00000421, 0.00000497, -0.00000621, 0.00000324, -0.00000322, 0.00001222, 0.00000068, 0.00000051, 0.00000369, -0.00000838, 0.00000215, 0.00000317, 0.00000246, 0.00000078, },
		{ -0.00000553, 0.00000016, 0.00000461, 0.00000842, 0.00000335, -0.00000608, -0.00000281, 0.00000859, 0.00000001, -0.00000274, 0.00000930, -0.00000836, -0.00000367, 0.00000681, 0.00000174, -0.00000910, 0.00000374, -0.00000028, -0.00000886, -0.00000993, 0.00000788, 0.00000115, -0.00000048, -0.00000678, -0.00000073, -0.00000368, -0.00001215, 0.00000803, 0.00000060, -0.00000124, -0.00000921, 0.00000237, 0.00000702, 0.00000115, -0.00000131, -0.00000150, -0.00000256, -0.00000320, -0.00000422, 0.00000418, -0.00000608, 0.00000344, -0.00001223, 0.00000587, -0.00002460, -0.00001943, 0.00000924, 0.00005044, 0.00001220, 0.00004424, 0.00003050, -0.00001226, -0.00003271, -0.00000168, -0.00000587, -0.00000797, -0.00000203, -0.00000640, -0.00000218, 0.00000391, -0.00000507, 0.00000059, -0.00000188, 0.00000220, -0.00000542, -0.00000267, -0.00001030, -0.00000065, -0.00000777, -0.00000651, -0.00001695, -0.00003245, -0.00010100, 0.00012268, 0.00010547, 0.00014371, -0.00004389, -0.00007398, -0.00000140, -0.00001787, 0.00000121, -0.00001139, 0.00000359, -0.00000857, 0.00000587, -0.00000574, -0.00000337, -0.00000732, 0.00000523, -0.00000088, -0.00000082, -0.00000536, 0.00001838, -0.00001452, 0.00001171, -0.00001349, 0.00000996, -0.00001249, -0.00000513, -0.00000297, -0.00000643, 0.00000408, 0.00000194, -0.00001575, 0.00000940, -0.00001213, 0.00000359, -0.00000511, -0.00000052, 0.00000507, -0.00000165, 0.00000197, -0.00000477, 0.00000801, -0.00000453, 0.00000484, -0.00000704, 0.00000416, -0.00000243, 0.00001409, -0.00000405, 0.00000468, -0.00000062, -0.00000633, 0.00000263, 0.00000348, 0.00000164, 0.00000076, },
		{ -0.00000617, 0.00000209, 0.00000502, 0.00000861, 0.00000072, -0.00000613, -0.00000109, 0.00000982, -0.00000394, 0.00000143, 0.00000676, -0.00000961, -0.00000130, 0.00000746, -0.00000151, -0.00000793, 0.00000582, -0.00000287, -0.00001055, -0.00000676, 0.00000961, -0.00000168, 0.00000056, -0.00000897, 0.00000306, -0.00000879, -0.00000740, 0.00000716, 0.00000116, -0.00000395, -0.00000781, 0.00000362, 0.00000780, -0.00000146, 0.00000032, -0.00000360, -0.00000014, -0.00000679, 0.00000099, -0.00000074, -0.00000148, -0.00000189, -0.00000725, 0.00000131, -0.00002683, -0.00001366, 0.00001993, 0.00004457, 0.00001008, 0.00004857, 0.00001942, -0.00001581, -0.00003015, 0.00000244, -0.00000917, -0.00000571, -0.00000239, -0.00000714, 0.00000104, 0.00000130, -0.00000385, -0.00000051, 0.00000028, -0.00000072, -0.00000327, -0.00000641, -0.00000624, -0.00000447, -0.00000341, -0.00001438, -0.00000904, -0.00005786, -0.00007516, 0.00014507, 0.00011133, 0.00013538, -0.00009547, -0.00003897, -0.00001396, -0.00000608, -0.00000816, -0.00000249, -0.00000388, -0.00000181, 0.00000076, -0.00000187, -0.00000887, -0.00000069, 0.00000068, 0.00000356, -0.00000784, 0.00000399, 0.00001121, -0.00001083, 0.00000940, -0.00001255, 0.00000997, -0.00001536, -0.00000195, -0.00000471, -0.00000547, 0.00000687, -0.00000448, -0.00001100, 0.00000849, -0.00001408, 0.00000776, -0.00001006, 0.00000572, -0.00000001, 0.00000193, -0.00000173, -0.00000060, 0.00000521, -0.00000308, 0.00000308, -0.00000630, 0.00000393, 0.00000012, 0.00001316, -0.00000647, 0.00000760, -0.00000469, -0.00000362, 0.00000266, 0.00000363, 0.00000105, 0.00000035, },
		{ -0.00000603, 0.00000374, 0.00000561, 0.00000785, -0.00000123, -0.00000626, 0.00000167, 0.00000903, -0.00000637, 0.00000543, 0.00000312, -0.00000954, 0.00000082, 0.00000751, -0.00000498, -0.00000520, 0.00000656, -0.00000574, -0.00001151, -0.00000243, 0.00000909, -0.00000304, 0.00000038, -0.00000954, 0.00000543, -0.00001329, -0.00000206, 0.00000535, 0.00000200, -0.00000709, -0.00000542, 0.00000457, 0.00000804, -0.00000380, 0.00000186, -0.00000531, 0.00000157, -0.00000946, 0.00000573, -0.00000597, 0.00000391, -0.00000866, 0.00000031, -0.00000778, -0.00002360, -0.00000937, 0.00003069, 0.00003411, 0.00001385, 0.00004770, 0.00001064, -0.00001930, -0.00002560, 0.00000414, -0.00001136, -0.00000362, -0.00000309, -0.00000731, 0.00000415, -0.00000219, -0.00000149, -0.00000239, 0.00000301, -0.00000450, -0.00000024, -0.00001099, -0.00000094, -0.00000966, 0.00000190, -0.00002333, -0.00000072, -0.00008711, -0.00003454, 0.00015491, 0.00012776, 0.00010098, -0.00012487, -0.00000777, -0.00002754, 0.00000594, -0.00001776, 0.00000675, -0.00001197, 0.00000580, -0.00000511, 0.00000203, -0.00001390, 0.00000605, -0.00000432, 0.00000783, -0.00001458, 0.00001446, 0.00000126, -0.00000385, 0.00000387, -0.00000854, 0.00000689, -0.00001550, -0.00000032, -0.00000595, -0.00000398, 0.00000852, -0.00001112, -0.00000346, 0.00000321, -0.00001132, 0.00000774, -0.00001164, 0.00000973, -0.00000424, 0.00000532, -0.00000539, 0.00000409, 0.00000119, -0.00000039, 0.00000010, -0.00000430, 0.00000296, 0.00000378, 0.00000973, -0.00000617, 0.00000841, -0.00000754, -0.00000099, 0.00000255, 0.00000350, 0.00000081, -0.00000058, },
		{ -0.00000510, 0.00000507, 0.00000613, 0.00000644, -0.00000261, -0.00000625, 0.00000504, 0.00000622, -0.00000656, 0.00000813, -0.00000066, -0.00000859, 0.00000282, 0.00000664, -0.00000798, -0.00000135, 0.00000569, -0.00000852, -0.00001139, 0.00000224, 0.00000669, -0.00000277, -0.00000137, -0.00000792, 0.00000537, -0.00001585, 0.00000255, 0.00000363, 0.00000223, -0.00000971, -0.00000276, 0.00000567, 0.00000710, -0.00000500, 0.00000248, -0.00000580, 0.00000155, -0.00000975, 0.00000800, -0.00000913, 0.00000742, -0.00001386, 0.00000711, -0.00001833, -0.00001641, -0.00000627, 0.00004016, 0.00002130, 0.00002227, 0.00004159, 0.00000501, -0.00002323, -0.00001913, 0.00000319, -0.00001191, -0.00000212, -0.00000405, -0.00000647, 0.00000616, -0.00000539, 0.00000095, -0.00000402, 0.00000510, -0.00000772, 0.00000211, -0.00001457, 0.00000352, -0.00001400, 0.00000560, -0.00003019, 0.00000346, -0.00011362, 0.00001502, 0.00015680, 0.00014819, 0.00004368, -0.00012581, 0.00000980, -0.00003462, 0.00001267, -0.00002298, 0.00001228, -0.00001708, 0.00001116, -0.00000920, 0.00000368, -0.00001606, 0.00001035, -0.00000736, 0.00000956, -0.00001825, 0.00002291, -0.00000858, 0.00000411, -0.00000307, -0.00000294, 0.00000184, -0.00001352, -0.00000002, -0.00000681, -0.00000158, 0.00000780, -0.00001569, 0.00000417, -0.00000432, -0.00000493, 0.00000357, -0.00000906, 0.00001010, -0.00000596, 0.00000697, -0.00000754, 0.00000783, -0.00000278, 0.00000258, -0.00000333, -0.00000165, 0.00000197, 0.00000749, 0.00000482, -0.00000359, 0.00000696, -0.00000863, 0.00000104, 0.00000250, 0.00000309, 0.00000086, -0.00000195, },
		{ -0.00000351, 0.00000612, 0.00000629, 0.00000483, -0.00000375, -0.00000558, 0.00000812, 0.00000216, -0.00000457, 0.00000896, -0.00000378, -0.00000724, 0.00000478, 0.00000459, -0.00000972, 0.00000283, 0.00000330, -0.00001085, -0.00000996, 0.00000628, 0.00000327, -0.00000133, -0.00000430, -0.00000431, 0.00000260, -0.00001567, 0.00000545, 0.00000278, 0.00000117, -0.00001105, -0.00000042, 0.00000717, 0.00000471, -0.00000458, 0.00000172, -0.00000475, -0.00000051, -0.00000706, 0.00000667, -0.00000856, 0.00000694, -0.00001475, 0.00000960, -0.00002626, -0.00000836, -0.00000273, 0.00004612, 0.00000942, 0.00003248, 0.00003188, 0.00000205, -0.00002715, -0.00001162, 0.00000015, -0.00001088, -0.00000132, -0.00000524, -0.00000431, 0.00000620, -0.00000707, 0.00000229, -0.00000432, 0.00000530, -0.00000894, 0.00000219, -0.00001532, 0.00000497, -0.00001517, 0.00000508, -0.00003181, -0.00000111, -0.00012998, 0.00006582, 0.00015783, 0.00016134, -0.00002512, -0.00010092, 0.00001005, -0.00003120, 0.00001085, -0.00002094, 0.00001140, -0.00001666, 0.00001198, -0.00000969, 0.00000166, -0.00001382, 0.00001046, -0.00000678, 0.00000710, -0.00001671, 0.00002659, -0.00001542, 0.00001049, -0.00000925, 0.00000240, -0.00000359, -0.00001052, -0.00000050, -0.00000748, 0.00000182, 0.00000406, -0.00001629, 0.00000900, -0.00001098, 0.00000247, -0.00000295, -0.00000324, 0.00000689, -0.00000460, 0.00000605, -0.00000721, 0.00000943, -0.00000550, 0.00000479, -0.00000634, 0.00000099, 0.00000162, 0.00001009, -0.00000017, 0.00000020, 0.00000378, -0.00000798, 0.00000224, 0.00000259, 0.00000253, 0.00000099, -0.00000349, },
		{ -0.00000147, 0.00000692, 0.00000592, 0.00000343, -0.00000495, -0.00000380, 0.00000987, -0.00000193, -0.00000119, 0.00000799, -0.00000592, -0.00000574, 0.00000661, 0.00000138, -0.00000959, 0.00000640, -0.00000021, -0.00001245, -0.00000718, 0.00000879, -0.00000003, 0.00000030, -0.00000743, 0.00000033, -0.00000223, -0.00001287, 0.00000637, 0.00000302, -0.00000134, -0.00001077, 0.00000139, 0.00000882, 0.00000120, -0.00000269, -0.00000024, -0.00000259, -0.00000393, -0.00000210, 0.00000212, -0.00000430, 0.00000209, -0.00001033, 0.00000570, -0.00002822, -0.00000272, 0.00000321, 0.00004641, 0.00000188, 0.00004096, 0.00002126, 0.00000024, -0.00002994, -0.00000447, -0.00000383, -0.00000882, -0.00000111, -0.00000652, -0.00000094, 0.00000396, -0.00000652, 0.00000179, -0.00000267, 0.00000297, -0.00000740, -0.00000078, -0.00001228, 0.00000223, -0.00001188, -0.00000106, -0.00002657, -0.00001719, -0.00013029, 0.00011108, 0.00016412, 0.00015534, -0.00008899, -0.00006082, -0.00000359, -0.00001841, 0.00000119, -0.00001200, 0.00000418, -0.00001049, 0.00000794, -0.00000648, -0.00000378, -0.00000744, 0.00000633, -0.00000249, 0.00000043, -0.00000948, 0.00002416, -0.00001740, 0.00001340, -0.00001297, 0.00000594, -0.00000803, -0.00000752, -0.00000129, -0.00000785, 0.00000563, -0.00000225, -0.00001233, 0.00000908, -0.00001396, 0.00000782, -0.00000901, 0.00000357, 0.00000155, -0.00000088, 0.00000286, -0.00000433, 0.00000837, -0.00000619, 0.00000549, -0.00000824, 0.00000315, 0.00000224, 0.00001080, -0.00000391, 0.00000387, -0.00000014, -0.00000606, 0.00000268, 0.00000276, 0.00000199, 0.00000093, -0.00000489, },
		{ 0.00000080, 0.00000738, 0.00000507, 0.00000237, -0.00000626, -0.00000079, 0.00000959, -0.00000478, 0.00000240, 0.00000586, -0.00000717, -0.00000402, 0.00000794, -0.00000257, -0.00000741, 0.00000850, -0.00000423, -0.00001311, -0.00000333, 0.00000928, -0.00000221, 0.00000107, -0.00000950, 0.00000451, -0.00000777, -0.00000833, 0.00000578, 0.00000394, -0.00000484, -0.00000910, 0.00000289, 0.00001000, -0.00000251, -0.00000015, -0.00000259, -0.00000038, -0.00000722, 0.00000328, -0.00000371, 0.00000183, -0.00000538, -0.00000207, -0.00000398, -0.00002318, -0.00000129, 0.00001245, 0.00004008, 0.00000091, 0.00004476, 0.00001239, -0.00000208, -0.00003040, 0.00000089, -0.00000749, -0.00000644, -0.00000136, -0.00000753, 0.00000299, -0.00000004, -0.00000392, -0.00000042, 0.00000062, -0.00000144, -0.00000357, -0.00000623, -0.00000608, -0.00000415, -0.00000477, -0.00001202, -0.00001557, -0.00004416, -0.00011222, 0.00014749, 0.00017730, 0.00012273, -0.00013195, -0.00002011, -0.00002246, -0.00000184, -0.00001196, 0.00000026, -0.00000644, -0.00000090, 0.00000086, -0.00000132, -0.00001065, 0.00000105, -0.00000025, 0.00000379, -0.00000859, 0.00000190, 0.00001623, -0.00001433, 0.00001231, -0.00001363, 0.00000707, -0.00001081, -0.00000502, -0.00000230, -0.00000746, 0.00000862, -0.00000944, -0.00000498, 0.00000434, -0.00001193, 0.00000891, -0.00001200, 0.00000878, -0.00000370, 0.00000355, -0.00000137, 0.00000020, 0.00000505, -0.00000478, 0.00000438, -0.00000867, 0.00000469, 0.00000373, 0.00000941, -0.00000559, 0.00000631, -0.00000374, -0.00000355, 0.00000262, 0.00000291, 0.00000160, 0.00000050, -0.00000583, },
		{ 0.00000315, 0.00000731, 0.00000405, 0.00000146, -0.00000732, 0.00000299, 0.00000720, -0.00000559, 0.00000512, 0.00000340, -0.00000791, -0.00000188, 0.00000821, -0.00000641, -0.00000353, 0.00000861, -0.00000815, -0.00001268, 0.00000101, 0.00000779, -0.00000277, 0.00000029, -0.00000952, 0.00000682, -0.00001242, -0.00000343, 0.00000462, 0.00000469, -0.00000844, -0.00000663, 0.00000447, 0.00000992, -0.00000531, 0.00000197, -0.00000432, 0.00000069, -0.00000870, 0.00000683, -0.00000820, 0.00000691, -0.00001223, 0.00000651, -0.00001620, -0.00001297, -0.00000352, 0.00002405, 0.00002821, 0.00000673, 0.00004255, 0.00000695, -0.00000595, -0.00002794, 0.00000360, -0.00000992, -0.00000428, -0.00000205, -0.00000766, 0.00000632, -0.00000452, -0.00000036, -0.00000329, 0.00000428, -0.00000647, 0.00000093, -0.00001227, 0.00000112, -0.00001188, 0.00000355, -0.00002473, -0.00000273, -0.00007764, -0.00007765, 0.00017616, 0.00019257, 0.00006448, -0.00014420, 0.00000807, -0.00003679, 0.00001131, -0.00002262, 0.00001067, -0.00001595, 0.00000831, -0.00000613, 0.00000301, -0.00001599, 0.00000855, -0.00000639, 0.00000920, -0.00001681, 0.00001432, 0.00000517, -0.00000763, 0.00000799, -0.00001170, 0.00000610, -0.00001201, -0.00000303, -0.00000372, -0.00000565, 0.00000926, -0.00001503, 0.00000310, -0.00000327, -0.00000564, 0.00000531, -0.00001061, 0.00001045, -0.00000678, 0.00000682, -0.00000497, 0.00000488, 0.00000055, -0.00000185, 0.00000174, -0.00000767, 0.00000575, 0.00000553, 0.00000641, -0.00000510, 0.00000692, -0.00000623, -0.00000111, 0.00000233, 0.00000298, 0.00000138, -0.00000034, -0.00000614, },
		{ 0.00000544, 0.00000655, 0.00000329, 0.00000032, -0.00000752, 0.00000657, 0.00000344, -0.00000434, 0.00000637, 0.00000124, -0.00000845, 0.00000082, 0.00000693, -0.00000914, 0.00000118, 0.00000673, -0.00001145, -0.00001102, 0.00000501, 0.00000494, -0.00000188, -0.00000203, -0.00000717, 0.00000638, -0.00001490, 0.00000058, 0.00000383, 0.00000444, -0.00001126, -0.00000397, 0.00000638, 0.00000814, -0.00000636, 0.00000282, -0.00000475, -0.00000005, -0.00000741, 0.00000698, -0.00000921, 0.00000828, -0.00001514, 0.00001127, -0.00002622, -0.00000159, -0.00000674, 0.00003521, 0.00001386, 0.00001725, 0.00003503, 0.00000506, -0.00001135, -0.00002289, 0.00000365, -0.00001081, -0.00000258, -0.00000324, -0.00000634, 0.00000781, -0.00000787, 0.00000258, -0.00000533, 0.00000662, -0.00001015, 0.00000394, -0.00001640, 0.00000641, -0.00001775, 0.00000947, -0.00003489, 0.00000639, -0.00011072, -0.00003154, 0.00020131, 0.00019974, -0.00000906, -0.00012566, 0.00001684, -0.00003989, 0.00001559, -0.00002599, 0.00001485, -0.00002031, 0.00001358, -0.00001013, 0.00000406, -0.00001727, 0.00001230, -0.00000937, 0.00001103, -0.00002103, 0.00002420, -0.00000584, 0.00000038, 0.00000206, -0.00000835, 0.00000390, -0.00001218, -0.00000131, -0.00000570, -0.00000211, 0.00000648, -0.00001673, 0.00000873, -0.00001051, 0.00000231, -0.00000138, -0.00000535, 0.00000817, -0.00000663, 0.00000764, -0.00000653, 0.00000817, -0.00000372, 0.00000158, -0.00000181, -0.00000553, 0.00000660, 0.00000691, 0.00000273, -0.00000301, 0.00000575, -0.00000722, 0.00000081, 0.00000202, 0.00000298, 0.00000122, -0.00000150, -0.00000575, },
		{ 0.00000749, 0.00000507, 0.00000308, -0.00000136, -0.00000630, 0.00000883, -0.00000046, -0.00000176, 0.00000625, -0.00000045, -0.00000879, 0.00000390, 0.00000398, -0.00000990, 0.00000559, 0.00000325, -0.00001379, -0.00000813, 0.00000784, 0.00000175, -0.00000033, -0.00000518, -0.00000302, 0.00000326, -0.00001471, 0.00000298, 0.00000395, 0.00000272, -0.00001271, -0.00000144, 0.00000847, 0.00000483, -0.00000554, 0.00000223, -0.00000392, -0.00000229, -0.00000362, 0.00000364, -0.00000618, 0.00000493, -0.00001239, 0.00000935, -0.00002985, 0.00000656, -0.00000747, 0.00004236, 0.00000120, 0.00002888, 0.00002458, 0.00000549, -0.00001728, -0.00001637, 0.00000176, -0.00001041, -0.00000134, -0.00000489, -0.00000335, 0.00000668, -0.00000882, 0.00000360, -0.00000530, 0.00000631, -0.00001096, 0.00000378, -0.00001666, 0.00000744, -0.00001908, 0.00000984, -0.00003875, 0.00000657, -0.00013613, 0.00002054, 0.00022705, 0.00018705, -0.00008091, -0.00008595, 0.00000755, -0.00003084, 0.00000982, -0.00002070, 0.00001125, -0.00001778, 0.00001318, -0.00000981, 0.00000095, -0.00001359, 0.00001110, -0.00000794, 0.00000791, -0.00001929, 0.00002881, -0.00001397, 0.00000738, -0.00000383, -0.00000474, 0.00000133, -0.00001179, 0.00000020, -0.00000795, 0.00000269, 0.00000042, -0.00001350, 0.00000954, -0.00001419, 0.00000856, -0.00000828, 0.00000168, 0.00000309, -0.00000356, 0.00000581, -0.00000546, 0.00000903, -0.00000653, 0.00000443, -0.00000547, -0.00000263, 0.00000739, 0.00000729, -0.00000059, -0.00000024, 0.00000333, -0.00000681, 0.00000201, 0.00000183, 0.00000293, 0.00000098, -0.00000276, -0.00000468, },
		{ 0.00000898, 0.00000312, 0.00000342, -0.00000357, -0.00000349, 0.00000894, -0.00000316, 0.00000098, 0.00000539, -0.00000194, -0.00000859, 0.00000675, -0.00000019, -0.00000832, 0.00000866, -0.00000114, -0.00001497, -0.00000423, 0.00000887, -0.00000074, 0.00000084, -0.00000794, 0.00000164, -0.00000158, -0.00001223, 0.00000380, 0.00000491, -0.00000036, -0.00001265, 0.00000098, 0.00001011, 0.00000082, -0.00000343, 0.00000069, -0.00000260, -0.00000484, 0.00000109, -0.00000160, -0.00000057, -0.00000187, -0.00000488, 0.00000068, -0.00002534, 0.00000863, -0.00000317, 0.00004272, -0.00000604, 0.00003781, 0.00001420, 0.00000636, -0.00002231, -0.00000981, -0.00000101, -0.00000919, -0.00000057, -0.00000664, 0.00000083, 0.00000306, -0.00000707, 0.00000229, -0.00000291, 0.00000307, -0.00000858, 0.00000014, -0.00001265, 0.00000351, -0.00001498, 0.00000354, -0.00003478, -0.00000500, -0.00014859, 0.00007498, 0.00025370, 0.00014662, -0.00013319, -0.00004056, -0.00001157, -0.00001439, -0.00000257, -0.00000937, 0.00000181, -0.00000968, 0.00000797, -0.00000604, -0.00000512, -0.00000619, 0.00000585, -0.00000283, 0.00000055, -0.00001168, 0.00002720, -0.00001764, 0.00001187, -0.00000860, -0.00000153, -0.00000125, -0.00001089, 0.00000114, -0.00000964, 0.00000732, -0.00000732, -0.00000623, 0.00000507, -0.00001251, 0.00001044, -0.00001239, 0.00000767, -0.00000258, 0.00000094, 0.00000224, -0.00000219, 0.00000731, -0.00000721, 0.00000588, -0.00000851, 0.00000072, 0.00000808, 0.00000643, -0.00000277, 0.00000224, 0.00000046, -0.00000539, 0.00000251, 0.00000181, 0.00000284, 0.00000056, -0.00000393, -0.00000296, },
		{ 0.00000954, 0.00000124, 0.00000393, -0.00000585, 0.00000044, 0.00000680, -0.00000384, 0.00000285, 0.00000456, -0.00000368, -0.00000733, 0.00000852, -0.00000463, -0.00000471, 0.00000973, -0.00000571, -0.00001485, 0.00000018, 0.00000799, -0.00000184, 0.00000075, -0.00000915, 0.00000531, -0.00000668, -0.00000850, 0.00000364, 0.00000612, -0.00000422, -0.00001136, 0.00000348, 0.00001057, -0.00000275, -0.00000105, -0.00000090, -0.00000183, -0.00000612, 0.00000457, -0.00000616, 0.00000480, -0.00000902, 0.00000410, -0.00001178, -0.00001419, 0.00000463, 0.00000640, 0.00003560, -0.00000591, 0.00004134, 0.00000637, 0.00000605, -0.00002530, -0.00000439, -0.00000376, -0.00000756, -0.00000042, -0.00000776, 0.00000504, -0.00000190, -0.00000346, -0.00000062, 0.00000093, -0.00000203, -0.00000410, -0.00000569, -0.00000581, -0.00000393, -0.00000690, -0.00000798, -0.00002445, -0.00002771, -0.00014595, 0.00013143, 0.00027554, 0.00007914, -0.00015370, -0.00000484, -0.00002963, 0.00000174, -0.00001535, 0.00000279, -0.00000904, 0.00000039, 0.00000079, -0.00000134, -0.00001142, 0.00000212, -0.00000087, 0.00000349, -0.00000853, -0.00000029, 0.00002032, -0.00001679, 0.00001345, -0.00001192, 0.00000123, -0.00000394, -0.00000917, 0.00000083, -0.00000959, 0.00000986, -0.00001408, 0.00000242, -0.00000290, -0.00000597, 0.00000713, -0.00001195, 0.00001034, -0.00000655, 0.00000495, -0.00000151, 0.00000203, 0.00000366, -0.00000580, 0.00000557, -0.00001042, 0.00000426, 0.00000838, 0.00000461, -0.00000353, 0.00000377, -0.00000212, -0.00000347, 0.00000245, 0.00000201, 0.00000264, -0.00000004, -0.00000488, -0.00000062, },
		{ 0.00000883, 0.00000006, 0.00000398, -0.00000737, 0.00000442, 0.00000316, -0.00000253, 0.00000337, 0.00000424, -0.00000597, -0.00000468, 0.00000840, -0.00000807, 0.00000003, 0.00000868, -0.00000989, -0.00001330, 0.00000434, 0.00000566, -0.00000150, -0.00000086, -0.00000817, 0.00000684, -0.00001062, -0.00000472, 0.00000335, 0.00000680, -0.00000810, -0.00000920, 0.00000616, 0.00000938, -0.00000494, 0.00000066, -0.00000183, -0.00000230, -0.00000511, 0.00000511, -0.00000776, 0.00000708, -0.00001299, 0.00001021, -0.00002312, -0.00000062, -0.00000268, 0.00001894, 0.00002292, 0.00000104, 0.00003875, 0.00000227, 0.00000376, -0.00002581, -0.00000066, -0.00000599, -0.00000571, -0.00000111, -0.00000744, 0.00000782, -0.00000645, 0.00000040, -0.00000363, 0.00000453, -0.00000708, 0.00000039, -0.00001134, 0.00000113, -0.00001194, 0.00000195, -0.00002121, -0.00001172, -0.00005801, -0.00012867, 0.00019170, 0.00028154, -0.00000454, -0.00014048, 0.00001175, -0.00003803, 0.00001070, -0.00002262, 0.00001050, -0.00001657, 0.00000836, -0.00000507, 0.00000148, -0.00001505, 0.00000822, -0.00000606, 0.00000806, -0.00001606, 0.00001167, 0.00001046, -0.00001254, 0.00001248, -0.00001396, 0.00000376, -0.00000704, -0.00000630, -0.00000123, -0.00000698, 0.00000867, -0.00001708, 0.00000896, -0.00001093, 0.00000283, 0.00000001, -0.00000710, 0.00000887, -0.00000741, 0.00000691, -0.00000388, 0.00000569, -0.00000069, -0.00000290, 0.00000356, -0.00001089, 0.00000777, 0.00000792, 0.00000241, -0.00000309, 0.00000415, -0.00000395, -0.00000151, 0.00000204, 0.00000243, 0.00000224, -0.00000072, -0.00000553, 0.00000226, },
		{ 0.00000676, 0.00000006, 0.00000299, -0.00000730, 0.00000711, -0.00000058, -0.00000007, 0.00000282, 0.00000432, -0.00000859, -0.00000081, 0.00000605, -0.00000945, 0.00000467, 0.00000588, -0.00001326, -0.00001033, 0.00000740, 0.00000279, -0.00000032, -0.00000351, -0.00000519, 0.00000582, -0.00001255, -0.00000178, 0.00000355, 0.00000639, -0.00001136, -0.00000638, 0.00000871, 0.00000665, -0.00000544, 0.00000134, -0.00000201, -0.00000381, -0.00000196, 0.00000242, -0.00000555, 0.00000488, -0.00001161, 0.00001010, -0.00002864, 0.00001030, -0.00000902, 0.00003047, 0.00000848, 0.00001216, 0.00003126, 0.00000165, -0.00000033, -0.00002412, 0.00000146, -0.00000760, -0.00000371, -0.00000273, -0.00000519, 0.00000802, -0.00000898, 0.00000289, -0.00000517, 0.00000621, -0.00001020, 0.00000287, -0.00001450, 0.00000541, -0.00001740, 0.00000803, -0.00003215, -0.00000133, -0.00009109, -0.00009774, 0.00025667, 0.00025938, -0.00008592, -0.00010255, 0.00000866, -0.00003401, 0.00000972, -0.00002162, 0.00001099, -0.00001795, 0.00001157, -0.00000751, 0.00000082, -0.00001437, 0.00001016, -0.00000769, 0.00000873, -0.00001937, 0.00002113, 0.00000022, -0.00000643, 0.00000958, -0.00001481, 0.00000603, -0.00001045, -0.00000245, -0.00000483, -0.00000196, 0.00000339, -0.00001475, 0.00001053, -0.00001540, 0.00001021, -0.00000789, 0.00000015, 0.00000414, -0.00000514, 0.00000631, -0.00000400, 0.00000756, -0.00000448, 0.00000055, 0.00000028, -0.00000977, 0.00001094, 0.00000649, 0.00000054, -0.00000203, 0.00000359, -0.00000487, 0.00000018, 0.00000149, 0.00000301, 0.00000156, -0.00000139, -0.00000578, 0.00000545, },
		{ 0.00000362, 0.00000132, 0.00000081, -0.00000526, 0.00000748, -0.00000295, 0.00000220, 0.00000209, 0.00000415, -0.00001082, 0.00000350, 0.00000187, -0.00000824, 0.00000807, 0.00000197, -0.00001556, -0.00000611, 0.00000868, 0.00000042, 0.00000073, -0.00000620, -0.00000117, 0.00000277, -0.00001239, 0.00000003, 0.00000443, 0.00000470, -0.00001364, -0.00000294, 0.00001050, 0.00000318, -0.00000464, 0.00000129, -0.00000201, -0.00000530, 0.00000190, -0.00000207, -0.00000072, -0.00000089, -0.00000529, 0.00000309, -0.00002587, 0.00001480, -0.00001061, 0.00003707, -0.00000342, 0.00002374, 0.00002130, 0.00000328, -0.00000544, -0.00002093, 0.00000230, -0.00000862, -0.00000180, -0.00000489, -0.00000130, 0.00000544, -0.00000876, 0.00000316, -0.00000450, 0.00000519, -0.00001047, 0.00000238, -0.00001406, 0.00000555, -0.00001846, 0.00000908, -0.00003807, 0.00000306, -0.00012286, -0.00005247, 0.00032247, 0.00020137, -0.00014505, -0.00005620, -0.00000665, -0.00002118, 0.00000113, -0.00001384, 0.00000511, -0.00001344, 0.00000992, -0.00000656, -0.00000287, -0.00000979, 0.00000800, -0.00000563, 0.00000521, -0.00001745, 0.00002620, -0.00000843, 0.00000027, 0.00000513, -0.00001420, 0.00000746, -0.00001344, 0.00000148, -0.00000882, 0.00000405, -0.00000463, -0.00000759, 0.00000620, -0.00001405, 0.00001300, -0.00001322, 0.00000685, -0.00000169, -0.00000102, 0.00000375, -0.00000197, 0.00000713, -0.00000674, 0.00000362, -0.00000369, -0.00000703, 0.00001329, 0.00000423, -0.00000057, -0.00000095, 0.00000249, -0.00000496, 0.00000140, 0.00000100, 0.00000363, 0.00000063, -0.00000199, -0.00000546, 0.00000849, },
		{ 0.00000016, 0.00000334, -0.00000212, -0.00000161, 0.00000531, -0.00000298, 0.00000312, 0.00000214, 0.00000288, -0.00001165, 0.00000702, -0.00000303, -0.00000472, 0.00000952, -0.00000237, -0.00001654, -0.00000117, 0.00000800, -0.00000072, 0.00000077, -0.00000779, 0.00000258, -0.00000124, -0.00001074, 0.00000095, 0.00000570, 0.00000191, -0.00001482, 0.00000113, 0.00001080, 0.00000004, -0.00000342, 0.00000116, -0.00000270, -0.00000534, 0.00000439, -0.00000591, 0.00000405, -0.00000734, 0.00000300, -0.00000831, -0.00001575, 0.00001201, -0.00000588, 0.00003653, -0.00000963, 0.00003246, 0.00001142, 0.00000571, -0.00001071, -0.00001688, 0.00000217, -0.00000896, -0.00000042, -0.00000674, 0.00000310, 0.00000101, -0.00000629, 0.00000153, -0.00000206, 0.00000197, -0.00000837, -0.00000051, -0.00001064, 0.00000201, -0.00001536, 0.00000501, -0.00003848, -0.00000009, -0.00015083, 0.00001039, 0.00037826, 0.00010988, -0.00016796, -0.00001841, -0.00002291, -0.00000706, -0.00000929, -0.00000398, -0.00000333, -0.00000597, 0.00000558, -0.00000416, -0.00000744, -0.00000355, 0.00000362, -0.00000155, -0.00000097, -0.00001117, 0.00002657, -0.00001451, 0.00000679, -0.00000077, -0.00001151, 0.00000705, -0.00001478, 0.00000401, -0.00001143, 0.00000870, -0.00001254, 0.00000178, -0.00000239, -0.00000702, 0.00000992, -0.00001372, 0.00001039, -0.00000618, 0.00000304, 0.00000064, 0.00000124, 0.00000473, -0.00000713, 0.00000561, -0.00000771, -0.00000279, 0.00001429, 0.00000164, -0.00000086, -0.00000020, 0.00000127, -0.00000443, 0.00000206, 0.00000080, 0.00000404, -0.00000036, -0.00000260, -0.00000432, 0.00001070, },
		{ -0.00000262, 0.00000522, -0.00000481, 0.00000252, 0.00000138, -0.00000075, 0.00000225, 0.00000352, -0.00000001, -0.00001032, 0.00000850, -0.00000712, 0.00000008, 0.00000893, -0.00000664, -0.00001594, 0.00000366, 0.00000578, -0.00000049, -0.00000056, -0.00000754, 0.00000489, -0.00000496, -0.00000848, 0.00000152, 0.00000689, -0.00000161, -0.00001480, 0.00000561, 0.00000925, -0.00000191, -0.00000254, 0.00000146, -0.00000455, -0.00000303, 0.00000386, -0.00000677, 0.00000588, -0.00001095, 0.00000898, -0.00001939, -0.00000223, 0.00000424, 0.00000390, 0.00002922, -0.00000909, 0.00003639, 0.00000349, 0.00000782, -0.00001559, -0.00001232, 0.00000113, -0.00000839, -0.00000015, -0.00000720, 0.00000641, -0.00000356, -0.00000301, -0.00000071, 0.00000077, -0.00000196, -0.00000545, -0.00000407, -0.00000621, -0.00000319, -0.00001007, -0.00000240, -0.00003486, -0.00001048, -0.00017327, 0.00009496, 0.00040728, -0.00000065, -0.00015232, -0.00000014, -0.00003068, 0.00000118, -0.00001552, 0.00000274, -0.00000975, 0.00000064, 0.00000152, -0.00000279, -0.00001025, 0.00000157, -0.00000035, 0.00000206, -0.00000725, -0.00000262, 0.00002315, -0.00001780, 0.00001261, -0.00000791, -0.00000636, 0.00000394, -0.00001337, 0.00000381, -0.00001102, 0.00000968, -0.00001706, 0.00000950, -0.00001163, 0.00000303, 0.00000218, -0.00000920, 0.00000956, -0.00000768, 0.00000539, -0.00000151, 0.00000428, 0.00000126, -0.00000582, 0.00000612, -0.00001111, 0.00000263, 0.00001346, -0.00000056, -0.00000067, 0.00000016, 0.00000015, -0.00000347, 0.00000216, 0.00000109, 0.00000395, -0.00000116, -0.00000336, -0.00000213, 0.00001137, },
		{ -0.00000381, 0.00000598, -0.00000612, 0.00000556, -0.00000271, 0.00000268, 0.00000009, 0.00000592, -0.00000435, -0.00000670, 0.00000722, -0.00000901, 0.00000478, 0.00000672, -0.00001054, -0.00001355, 0.00000740, 0.00000293, 0.00000058, -0.00000292, -0.00000549, 0.00000523, -0.00000758, -0.00000631, 0.00000219, 0.00000756, -0.00000550, -0.00001335, 0.00000979, 0.00000620, -0.00000246, -0.00000220, 0.00000205, -0.00000712, 0.00000126, 0.00000015, -0.00000384, 0.00000331, -0.00000940, 0.00000915, -0.00002532, 0.00000953, -0.00000428, 0.00001528, 0.00001777, -0.00000295, 0.00003527, -0.00000163, 0.00000895, -0.00001978, -0.00000739, -0.00000084, -0.00000677, -0.00000131, -0.00000554, 0.00000731, -0.00000657, -0.00000050, -0.00000214, 0.00000248, -0.00000497, -0.00000334, -0.00000648, -0.00000295, -0.00000768, -0.00000505, -0.00001079, -0.00002926, -0.00002763, -0.00018686, 0.00020218, 0.00039193, -0.00010658, -0.00010865, -0.00000246, -0.00002664, 0.00000038, -0.00001454, 0.00000337, -0.00001125, 0.00000377, -0.00000028, -0.00000394, -0.00000974, 0.00000370, -0.00000201, 0.00000325, -0.00001136, 0.00000593, 0.00001722, -0.00001835, 0.00001699, -0.00001549, 0.00000082, -0.00000186, -0.00000897, 0.00000051, -0.00000703, 0.00000586, -0.00001592, 0.00001211, -0.00001737, 0.00001204, -0.00000707, -0.00000157, 0.00000504, -0.00000592, 0.00000533, -0.00000176, 0.00000597, -0.00000219, -0.00000336, 0.00000503, -0.00001324, 0.00000861, 0.00001071, -0.00000171, -0.00000055, 0.00000037, -0.00000083, -0.00000225, 0.00000175, 0.00000192, 0.00000318, -0.00000153, -0.00000431, 0.00000114, 0.00001001, },
		{ -0.00000295, 0.00000505, -0.00000529, 0.00000614, -0.00000512, 0.00000561, -0.00000205, 0.00000829, -0.00000912, -0.00000160, 0.00000339, -0.00000797, 0.00000811, 0.00000358, -0.00001384, -0.00000942, 0.00000918, 0.00000058, 0.00000154, -0.00000537, -0.00000236, 0.00000385, -0.00000885, -0.00000455, 0.00000313, 0.00000752, -0.00000955, -0.00001013, 0.00001263, 0.00000267, -0.00000218, -0.00000194, 0.00000211, -0.00000905, 0.00000582, -0.00000513, 0.00000159, -0.00000275, -0.00000298, 0.00000262, -0.00002335, 0.00001540, -0.00000938, 0.00002431, 0.00000586, 0.00000629, 0.00003008, -0.00000383, 0.00000877, -0.00002302, -0.00000238, -0.00000350, -0.00000437, -0.00000358, -0.00000193, 0.00000544, -0.00000720, 0.00000035, -0.00000204, 0.00000236, -0.00000627, -0.00000279, -0.00000693, -0.00000191, -0.00001015, -0.00000175, -0.00001881, -0.00002276, -0.00005247, -0.00018429, 0.00032500, 0.00032125, -0.00018237, -0.00005631, -0.00001708, -0.00001455, -0.00000711, -0.00000783, -0.00000128, -0.00000804, 0.00000334, 0.00000007, -0.00000711, -0.00000632, 0.00000286, -0.00000120, 0.00000165, -0.00001233, 0.00001305, 0.00000974, -0.00001595, 0.00001873, -0.00002181, 0.00000845, -0.00000905, -0.00000267, -0.00000486, -0.00000058, -0.00000187, -0.00000908, 0.00000812, -0.00001671, 0.00001626, -0.00001401, 0.00000604, -0.00000104, -0.00000206, 0.00000333, -0.00000009, 0.00000577, -0.00000474, -0.00000042, 0.00000243, -0.00001348, 0.00001420, 0.00000644, -0.00000153, -0.00000093, 0.00000073, -0.00000172, -0.00000091, 0.00000103, 0.00000316, 0.00000176, -0.00000144, -0.00000530, 0.00000515, 0.00000665, },
		{ -0.00000033, 0.00000257, -0.00000236, 0.00000376, -0.00000461, 0.00000648, -0.00000270, 0.00000919, -0.00001281, 0.00000346, -0.00000180, -0.00000425, 0.00000934, 0.00000012, -0.00001624, -0.00000399, 0.00000869, -0.00000043, 0.00000146, -0.00000684, 0.00000067, 0.00000157, -0.00000913, -0.00000308, 0.00000420, 0.00000676, -0.00001353, -0.00000498, 0.00001312, -0.00000003, -0.00000204, -0.00000096, 0.00000059, -0.00000863, 0.00000823, -0.00000917, 0.00000658, -0.00000926, 0.00000530, -0.00000824, -0.00001419, 0.00001409, -0.00000881, 0.00002820, -0.00000331, 0.00001590, 0.00002232, -0.00000336, 0.00000690, -0.00002476, 0.00000208, -0.00000619, -0.00000198, -0.00000586, 0.00000239, 0.00000172, -0.00000590, -0.00000011, -0.00000086, 0.00000090, -0.00000636, -0.00000314, -0.00000623, -0.00000243, -0.00001108, 0.00000012, -0.00002676, -0.00001487, -0.00008759, -0.00015374, 0.00044550, 0.00019790, -0.00020989, -0.00001579, -0.00003115, -0.00000269, -0.00001502, -0.00000032, -0.00000728, -0.00000310, 0.00000153, 0.00000057, -0.00001007, -0.00000220, 0.00000085, 0.00000056, -0.00000145, -0.00001071, 0.00001838, 0.00000116, -0.00001021, 0.00001637, -0.00002463, 0.00001405, -0.00001529, 0.00000331, -0.00001004, 0.00000577, -0.00001064, 0.00000100, -0.00000115, -0.00000938, 0.00001382, -0.00001584, 0.00001063, -0.00000607, 0.00000199, 0.00000071, 0.00000262, 0.00000388, -0.00000595, 0.00000238, -0.00000139, -0.00001135, 0.00001827, 0.00000156, -0.00000026, -0.00000192, 0.00000138, -0.00000254, 0.00000034, 0.00000031, 0.00000445, -0.00000001, -0.00000113, -0.00000587, 0.00000910, 0.00000198, },
		{ 0.00000308, -0.00000053, 0.00000162, -0.00000089, -0.00000108, 0.00000458, -0.00000092, 0.00000750, -0.00001401, 0.00000675, -0.00000655, 0.00000101, 0.00000848, -0.00000333, -0.00001728, 0.00000185, 0.00000627, 0.00000010, -0.00000007, -0.00000661, 0.00000261, -0.00000069, -0.00000895, -0.00000163, 0.00000517, 0.00000527, -0.00001696, 0.00000173, 0.00001080, -0.00000095, -0.00000282, 0.00000121, -0.00000305, -0.00000481, 0.00000660, -0.00000935, 0.00000793, -0.00001240, 0.00001099, -0.00001870, -0.00000162, 0.00000753, -0.00000312, 0.00002642, -0.00000806, 0.00002390, 0.00001332, -0.00000050, 0.00000289, -0.00002420, 0.00000501, -0.00000791, -0.00000072, -0.00000669, 0.00000559, -0.00000201, -0.00000414, -0.00000063, 0.00000006, -0.00000049, -0.00000666, -0.00000280, -0.00000621, -0.00000263, -0.00001221, 0.00000204, -0.00003590, -0.00000468, -0.00013527, -0.00008171, 0.00053632, 0.00004128, -0.00018576, -0.00000039, -0.00003386, 0.00000099, -0.00001691, 0.00000249, -0.00000992, -0.00000026, 0.00000129, -0.00000125, -0.00001022, -0.00000008, 0.00000012, 0.00000105, -0.00000396, -0.00000796, 0.00002224, -0.00000820, -0.00000115, 0.00000903, -0.00002210, 0.00001526, -0.00001813, 0.00000652, -0.00001251, 0.00000899, -0.00001668, 0.00001013, -0.00001190, 0.00000203, 0.00000563, -0.00001193, 0.00001058, -0.00000813, 0.00000444, -0.00000100, 0.00000510, 0.00000107, -0.00000586, 0.00000455, -0.00000594, -0.00000671, 0.00001979, -0.00000271, 0.00000136, -0.00000316, 0.00000220, -0.00000321, 0.00000128, -0.00000001, 0.00000527, -0.00000161, -0.00000101, -0.00000545, 0.00001193, -0.00000275, },
		{ 0.00000596, -0.00000291, 0.00000492, -0.00000602, 0.00000423, 0.00000047, 0.00000316, 0.00000301, -0.00001204, 0.00000701, -0.00000908, 0.00000615, 0.00000616, -0.00000673, -0.00001645, 0.00000686, 0.00000295, 0.00000161, -0.00000269, -0.00000469, 0.00000300, -0.00000234, -0.00000870, -0.00000003, 0.00000595, 0.00000285, -0.00001900, 0.00000889, 0.00000624, -0.00000005, -0.00000450, 0.00000418, -0.00000813, 0.00000193, 0.00000074, -0.00000473, 0.00000399, -0.00000964, 0.00001037, -0.00002380, 0.00000918, -0.00000025, 0.00000492, 0.00002054, -0.00000845, 0.00002929, 0.00000417, 0.00000433, -0.00000350, -0.00002070, 0.00000554, -0.00000786, -0.00000139, -0.00000499, 0.00000607, -0.00000397, -0.00000345, 0.00000010, -0.00000059, -0.00000049, -0.00000839, -0.00000046, -0.00000844, -0.00000086, -0.00001503, 0.00000501, -0.00004656, 0.00000661, -0.00019344, 0.00004076, 0.00056728, -0.00011568, -0.00012354, -0.00001099, -0.00002185, -0.00000648, -0.00000996, -0.00000210, -0.00000659, -0.00000177, 0.00000419, -0.00000652, -0.00000631, -0.00000147, 0.00000213, -0.00000116, -0.00000430, -0.00000533, 0.00002474, -0.00001741, 0.00001015, -0.00000278, -0.00001383, 0.00001095, -0.00001612, 0.00000546, -0.00001068, 0.00000706, -0.00001693, 0.00001414, -0.00001940, 0.00001307, -0.00000502, -0.00000403, 0.00000632, -0.00000672, 0.00000447, -0.00000083, 0.00000622, -0.00000169, -0.00000477, 0.00000573, -0.00001049, 0.00000006, 0.00001817, -0.00000523, 0.00000238, -0.00000405, 0.00000283, -0.00000352, 0.00000170, 0.00000037, 0.00000517, -0.00000252, -0.00000150, -0.00000352, 0.00001261, -0.00000616, },
		{ 0.00000719, -0.00000337, 0.00000581, -0.00000939, 0.00000918, -0.00000417, 0.00000823, -0.00000330, -0.00000739, 0.00000406, -0.00000830, 0.00000963, 0.00000328, -0.00001012, -0.00001341, 0.00000987, 0.00000005, 0.00000303, -0.00000541, -0.00000186, 0.00000216, -0.00000335, -0.00000837, 0.00000157, 0.00000668, -0.00000082, -0.00001856, 0.00001473, 0.00000097, 0.00000170, -0.00000615, 0.00000660, -0.00001275, 0.00000936, -0.00000734, 0.00000317, -0.00000422, -0.00000128, 0.00000249, -0.00002077, 0.00001408, -0.00000521, 0.00001187, 0.00001316, -0.00000559, 0.00003172, -0.00000409, 0.00001022, -0.00001180, -0.00001432, 0.00000350, -0.00000595, -0.00000387, -0.00000083, 0.00000348, -0.00000346, -0.00000443, 0.00000241, -0.00000304, 0.00000105, -0.00001144, 0.00000365, -0.00001279, 0.00000273, -0.00001909, 0.00000782, -0.00005613, 0.00001326, -0.00025158, 0.00021061, 0.00051585, -0.00023567, -0.00004927, -0.00003630, -0.00000082, -0.00002122, 0.00000308, -0.00001215, 0.00000144, -0.00000672, 0.00000916, -0.00001384, 0.00000045, -0.00000556, 0.00000627, -0.00000563, -0.00000248, -0.00000290, 0.00002512, -0.00002455, 0.00002111, -0.00001658, -0.00000155, 0.00000206, -0.00000974, 0.00000046, -0.00000494, 0.00000014, -0.00001059, 0.00001080, -0.00001999, 0.00001926, -0.00001387, 0.00000456, -0.00000006, -0.00000290, 0.00000246, 0.00000126, 0.00000546, -0.00000366, -0.00000306, 0.00000561, -0.00001409, 0.00000799, 0.00001358, -0.00000538, 0.00000203, -0.00000394, 0.00000283, -0.00000328, 0.00000150, 0.00000155, 0.00000394, -0.00000245, -0.00000275, 0.00000001, 0.00001059, -0.00000717, },
		{ 0.00000635, -0.00000149, 0.00000333, -0.00000919, 0.00001157, -0.00000721, 0.00001230, -0.00000961, -0.00000164, -0.00000106, -0.00000426, 0.00001054, 0.00000056, -0.00001336, -0.00000829, 0.00001017, -0.00000130, 0.00000323, -0.00000703, 0.00000070, 0.00000092, -0.00000419, -0.00000761, 0.00000281, 0.00000755, -0.00000594, -0.00001472, 0.00001741, -0.00000302, 0.00000270, -0.00000634, 0.00000669, -0.00001438, 0.00001424, -0.00001404, 0.00001068, -0.00001305, 0.00000925, -0.00001002, -0.00001048, 0.00001194, -0.00000523, 0.00001537, 0.00000651, -0.00000073, 0.00003099, -0.00001014, 0.00001552, -0.00002053, -0.00000621, -0.00000024, -0.00000315, -0.00000682, 0.00000430, -0.00000092, -0.00000138, -0.00000616, 0.00000512, -0.00000595, 0.00000260, -0.00001393, 0.00000725, -0.00001683, 0.00000557, -0.00002143, 0.00000648, -0.00005870, 0.00000456, -0.00028912, 0.00040728, 0.00037764, -0.00028952, 0.00000827, -0.00005851, 0.00001754, -0.00003428, 0.00001501, -0.00002182, 0.00000953, -0.00001138, 0.00001273, -0.00001964, 0.00000664, -0.00000942, 0.00000994, -0.00001004, -0.00000019, 0.00000046, 0.00002187, -0.00002696, 0.00002810, -0.00002840, 0.00001110, -0.00000846, -0.00000145, -0.00000619, 0.00000225, -0.00000915, 0.00000018, 0.00000098, -0.00001279, 0.00001805, -0.00001748, 0.00001041, -0.00000577, 0.00000136, -0.00000024, 0.00000437, 0.00000310, -0.00000453, -0.00000107, 0.00000397, -0.00001572, 0.00001559, 0.00000698, -0.00000331, 0.00000009, -0.00000251, 0.00000196, -0.00000245, 0.00000082, 0.00000329, 0.00000177, -0.00000152, -0.00000445, 0.00000464, 0.00000611, -0.00000551, },
		{ 0.00000400, 0.00000201, -0.00000214, -0.00000492, 0.00001010, -0.00000697, 0.00001349, -0.00001386, 0.00000313, -0.00000638, 0.00000181, 0.00000895, -0.00000172, -0.00001599, -0.00000186, 0.00000788, -0.00000066, 0.00000163, -0.00000671, 0.00000195, 0.00000018, -0.00000542, -0.00000605, 0.00000343, 0.00000853, -0.00001221, -0.00000736, 0.00001580, -0.00000411, 0.00000161, -0.00000396, 0.00000318, -0.00001112, 0.00001370, -0.00001574, 0.00001365, -0.00001782, 0.00001670, -0.00002178, 0.00000280, 0.00000502, -0.00000104, 0.00001515, 0.00000155, 0.00000544, 0.00002673, -0.00001243, 0.00001803, -0.00002736, 0.00000151, -0.00000393, -0.00000118, -0.00000819, 0.00000789, -0.00000467, 0.00000012, -0.00000659, 0.00000600, -0.00000691, 0.00000163, -0.00001289, 0.00000685, -0.00001675, 0.00000366, -0.00001774, -0.00000421, -0.00004727, -0.00003201, -0.00027864, 0.00059284, 0.00017253, -0.00026710, 0.00002878, -0.00006191, 0.00002172, -0.00003630, 0.00001794, -0.00002434, 0.00001206, -0.00001120, 0.00001091, -0.00001989, 0.00000830, -0.00000948, 0.00000987, -0.00001135, 0.00000018, 0.00000627, 0.00001357, -0.00002239, 0.00002775, -0.00003402, 0.00001983, -0.00001672, 0.00000530, -0.00001123, 0.00000733, -0.00001661, 0.00001093, -0.00001145, -0.00000022, 0.00000993, -0.00001468, 0.00001147, -0.00000857, 0.00000411, -0.00000201, 0.00000714, -0.00000003, -0.00000443, 0.00000097, 0.00000081, -0.00001454, 0.00002121, -0.00000003, 0.00000005, -0.00000299, 0.00000005, 0.00000029, -0.00000120, 0.00000004, 0.00000504, -0.00000074, -0.00000028, -0.00000588, 0.00000924, 0.00000026, -0.00000184, },
		{ 0.00000143, 0.00000545, -0.00000881, 0.00000232, 0.00000498, -0.00000303, 0.00001086, -0.00001465, 0.00000508, -0.00000972, 0.00000799, 0.00000571, -0.00000378, -0.00001723, 0.00000454, 0.00000399, 0.00000149, -0.00000152, -0.00000441, 0.00000151, 0.00000035, -0.00000725, -0.00000364, 0.00000351, 0.00000912, -0.00001851, 0.00000246, 0.00001015, -0.00000192, -0.00000185, 0.00000091, -0.00000367, -0.00000290, 0.00000685, -0.00001067, 0.00000953, -0.00001504, 0.00001646, -0.00002709, 0.00001330, -0.00000224, 0.00000443, 0.00001281, -0.00000205, 0.00001270, 0.00001867, -0.00000972, 0.00001574, -0.00002990, 0.00000657, -0.00000576, -0.00000162, -0.00000617, 0.00000771, -0.00000536, -0.00000110, -0.00000377, 0.00000306, -0.00000390, -0.00000387, -0.00000610, -0.00000023, -0.00000960, -0.00000600, -0.00000512, -0.00002726, -0.00001822, -0.00010439, -0.00019437, 0.00071952, -0.00005726, -0.00018224, 0.00000847, -0.00004064, 0.00000675, -0.00002284, 0.00000777, -0.00001592, 0.00000577, -0.00000369, 0.00000173, -0.00001256, 0.00000322, -0.00000367, 0.00000409, -0.00000761, -0.00000286, 0.00001506, 0.00000018, -0.00001036, 0.00001852, -0.00003081, 0.00002139, -0.00001953, 0.00000762, -0.00001194, 0.00000728, -0.00001833, 0.00001673, -0.00002113, 0.00001290, -0.00000182, -0.00000695, 0.00000781, -0.00000772, 0.00000434, -0.00000174, 0.00000832, -0.00000289, -0.00000372, 0.00000285, -0.00000359, -0.00001018, 0.00002346, -0.00000568, 0.00000328, -0.00000618, 0.00000304, -0.00000175, 0.00000010, -0.00000035, 0.00000610, -0.00000279, 0.00000054, -0.00000612, 0.00001240, -0.00000531, 0.00000248, },
		{ 0.00000009, 0.00000680, -0.00001408, 0.00001010, -0.00000205, 0.00000353, 0.00000489, -0.00001186, 0.00000346, -0.00000952, 0.00001230, 0.00000210, -0.00000608, -0.00001628, 0.00000938, 0.00000002, 0.00000395, -0.00000512, -0.00000101, -0.00000013, 0.00000113, -0.00000931, -0.00000084, 0.00000357, 0.00000835, -0.00002301, 0.00001245, 0.00000224, 0.00000240, -0.00000656, 0.00000668, -0.00001175, 0.00000798, -0.00000438, -0.00000011, -0.00000103, -0.00000454, 0.00000702, -0.00002266, 0.00001647, -0.00000553, 0.00000781, 0.00001061, -0.00000505, 0.00002072, 0.00000733, -0.00000195, 0.00000785, -0.00002686, 0.00000760, -0.00000483, -0.00000489, -0.00000041, 0.00000306, -0.00000207, -0.00000574, 0.00000269, -0.00000395, 0.00000313, -0.00001363, 0.00000600, -0.00001364, 0.00000426, -0.00002267, 0.00001483, -0.00005985, 0.00002376, -0.00020820, -0.00002386, 0.00074357, -0.00025776, -0.00006879, -0.00003857, -0.00000180, -0.00002245, 0.00000257, -0.00001293, 0.00000163, -0.00000792, 0.00000956, -0.00001299, 0.00000076, -0.00000738, 0.00000698, -0.00000649, 0.00000060, -0.00000863, 0.00002529, -0.00001595, 0.00000685, 0.00000192, -0.00001908, 0.00001514, -0.00001591, 0.00000460, -0.00000765, 0.00000129, -0.00001257, 0.00001440, -0.00002349, 0.00002137, -0.00001248, 0.00000229, 0.00000143, -0.00000409, 0.00000231, 0.00000071, 0.00000732, -0.00000466, -0.00000280, 0.00000427, -0.00000853, -0.00000303, 0.00002167, -0.00000855, 0.00000496, -0.00000826, 0.00000552, -0.00000354, 0.00000103, 0.00000004, 0.00000592, -0.00000369, 0.00000031, -0.00000449, 0.00001293, -0.00000899, 0.00000593, },
		{ 0.00000089, 0.00000460, -0.00001552, 0.00001554, -0.00000832, 0.00001044, -0.00000263, -0.00000674, -0.00000108, -0.00000544, 0.00001349, -0.00000079, -0.00000891, -0.00001274, 0.00001142, -0.00000248, 0.00000525, -0.00000768, 0.00000204, -0.00000182, 0.00000165, -0.00001077, 0.00000154, 0.00000439, 0.00000513, -0.00002370, 0.00001973, -0.00000510, 0.00000654, -0.00001031, 0.00001074, -0.00001769, 0.00001746, -0.00001569, 0.00001172, -0.00001399, 0.00000999, -0.00000886, -0.00000943, 0.00001127, -0.00000294, 0.00000721, 0.00000991, -0.00000761, 0.00002828, -0.00000536, 0.00000911, -0.00000441, -0.00001882, 0.00000493, -0.00000185, -0.00000961, 0.00000734, -0.00000443, 0.00000375, -0.00001217, 0.00001075, -0.00001266, 0.00001145, -0.00002425, 0.00001937, -0.00002897, 0.00001998, -0.00004060, 0.00003458, -0.00009161, 0.00006309, -0.00032167, 0.00022149, 0.00064121, -0.00037977, 0.00003109, -0.00008667, 0.00003747, -0.00005266, 0.00002923, -0.00003533, 0.00002106, -0.00002290, 0.00002293, -0.00002768, 0.00001483, -0.00001889, 0.00001822, -0.00001795, 0.00000988, -0.00001407, 0.00003337, -0.00003032, 0.00002428, -0.00001747, -0.00000238, 0.00000353, -0.00000768, -0.00000206, -0.00000029, -0.00000848, -0.00000095, 0.00000416, -0.00001685, 0.00002179, -0.00001795, 0.00000925, -0.00000476, 0.00000033, -0.00000067, 0.00000447, 0.00000435, -0.00000504, -0.00000189, 0.00000488, -0.00001295, 0.00000577, 0.00001618, -0.00000809, 0.00000418, -0.00000826, 0.00000661, -0.00000452, 0.00000134, 0.00000131, 0.00000433, -0.00000319, -0.00000114, -0.00000090, 0.00001034, -0.00000973, 0.00000747, },
		{ 0.00000365, -0.00000130, -0.00001188, 0.00001645, -0.00001131, 0.00001517, -0.00000929, -0.00000148, -0.00000663, 0.00000143, 0.00001142, -0.00000249, -0.00001210, -0.00000693, 0.00001025, -0.00000258, 0.00000439, -0.00000798, 0.00000337, -0.00000233, 0.00000088, -0.00001071, 0.00000271, 0.00000647, -0.00000106, -0.00001922, 0.00002185, -0.00000908, 0.00000802, -0.00001082, 0.00001057, -0.00001822, 0.00002122, -0.00002204, 0.00001931, -0.00002337, 0.00002217, -0.00002485, 0.00000750, 0.00000075, 0.00000405, 0.00000333, 0.00001030, -0.00000867, 0.00003314, -0.00001618, 0.00001993, -0.00001784, -0.00000836, 0.00000061, 0.00000097, -0.00001294, 0.00001362, -0.00001116, 0.00000866, -0.00001680, 0.00001640, -0.00001867, 0.00001620, -0.00003006, 0.00002737, -0.00003879, 0.00002933, -0.00005031, 0.00004237, -0.00010676, 0.00007576, -0.00040651, 0.00050207, 0.00042071, -0.00039562, 0.00008236, -0.00010983, 0.00005792, -0.00006825, 0.00004396, -0.00004813, 0.00003274, -0.00003102, 0.00002908, -0.00003527, 0.00002291, -0.00002507, 0.00002424, -0.00002481, 0.00001541, -0.00001495, 0.00003485, -0.00003757, 0.00003584, -0.00003337, 0.00001362, -0.00000877, 0.00000129, -0.00000876, 0.00000625, -0.00001760, 0.00001190, -0.00001017, -0.00000333, 0.00001411, -0.00001648, 0.00001143, -0.00000828, 0.00000349, -0.00000291, 0.00000812, 0.00000033, -0.00000423, -0.00000102, 0.00000428, -0.00001567, 0.00001448, 0.00000832, -0.00000478, 0.00000093, -0.00000585, 0.00000585, -0.00000438, 0.00000103, 0.00000318, 0.00000169, -0.00000158, -0.00000334, 0.00000396, 0.00000511, -0.00000751, 0.00000700, },
		{ 0.00000698, -0.00000941, -0.00000378, 0.00001229, -0.00000967, 0.00001590, -0.00001302, 0.00000168, -0.00001076, 0.00000898, 0.00000703, -0.00000325, -0.00001490, 0.00000005, 0.00000650, -0.00000039, 0.00000126, -0.00000567, 0.00000236, -0.00000099, -0.00000172, -0.00000869, 0.00000243, 0.00000958, -0.00000965, -0.00000971, 0.00001789, -0.00000812, 0.00000544, -0.00000701, 0.00000511, -0.00001181, 0.00001679, -0.00001992, 0.00001835, -0.00002395, 0.00002568, -0.00003368, 0.00002092, -0.00000934, 0.00001129, -0.00000096, 0.00000983, -0.00000639, 0.00003263, -0.00002147, 0.00002629, -0.00002824, 0.00000091, -0.00000249, 0.00000099, -0.00001183, 0.00001473, -0.00001315, 0.00000876, -0.00001572, 0.00001547, -0.00001755, 0.00001266, -0.00002569, 0.00002364, -0.00003584, 0.00002427, -0.00004276, 0.00002717, -0.00009047, 0.00003745, -0.00041693, 0.00075561, 0.00012507, -0.00030959, 0.00007024, -0.00009336, 0.00004753, -0.00005882, 0.00003754, -0.00004303, 0.00002941, -0.00002617, 0.00002271, -0.00003060, 0.00001989, -0.00002116, 0.00002053, -0.00002269, 0.00001328, -0.00000811, 0.00002663, -0.00003385, 0.00003664, -0.00004015, 0.00002332, -0.00001686, 0.00000679, -0.00001173, 0.00000799, -0.00002118, 0.00002020, -0.00002257, 0.00001194, 0.00000158, -0.00000926, 0.00000858, -0.00000812, 0.00000417, -0.00000312, 0.00001023, -0.00000356, -0.00000277, -0.00000012, 0.00000223, -0.00001565, 0.00002120, 0.00000005, -0.00000001, -0.00000389, -0.00000153, 0.00000339, -0.00000323, 0.00000043, 0.00000503, -0.00000121, 0.00000037, -0.00000534, 0.00000872, -0.00000135, -0.00000330, 0.00000540, },
		{ 0.00000881, -0.00001703, 0.00000644, 0.00000439, -0.00000382, 0.00001228, -0.00001294, 0.00000139, -0.00001141, 0.00001485, 0.00000194, -0.00000384, -0.00001627, 0.00000653, 0.00000165, 0.00000300, -0.00000316, -0.00000153, -0.00000053, 0.00000188, -0.00000579, -0.00000510, 0.00000128, 0.00001259, -0.00001875, 0.00000286, 0.00000900, -0.00000270, -0.00000073, 0.00000026, -0.00000434, 0.00000023, 0.00000490, -0.00000929, 0.00000800, -0.00001396, 0.00001737, -0.00003046, 0.00002482, -0.00001331, 0.00001419, -0.00000225, 0.00000639, 0.00000068, 0.00002506, -0.00001871, 0.00002497, -0.00003213, 0.00000595, -0.00000226, -0.00000327, -0.00000480, 0.00000875, -0.00000814, 0.00000193, -0.00000699, 0.00000604, -0.00000739, -0.00000095, -0.00000921, 0.00000569, -0.00001716, 0.00000160, -0.00001466, -0.00001481, -0.00003731, -0.00006469, -0.00031505, 0.00091277, -0.00017733, -0.00015764, 0.00000563, -0.00004069, 0.00000789, -0.00002488, 0.00000980, -0.00001934, 0.00001026, -0.00000787, 0.00000369, -0.00001346, 0.00000530, -0.00000660, 0.00000644, -0.00001074, 0.00000279, 0.00000654, 0.00000894, -0.00001880, 0.00002524, -0.00003525, 0.00002357, -0.00001775, 0.00000638, -0.00000899, 0.00000292, -0.00001631, 0.00001957, -0.00002735, 0.00002289, -0.00001063, 0.00000020, 0.00000258, -0.00000502, 0.00000248, -0.00000099, 0.00001000, -0.00000628, -0.00000126, 0.00000086, -0.00000115, -0.00001241, 0.00002438, -0.00000655, 0.00000433, -0.00000872, 0.00000352, -0.00000005, -0.00000150, -0.00000000, 0.00000612, -0.00000344, 0.00000169, -0.00000606, 0.00001181, -0.00000719, 0.00000137, 0.00000409, },
		{ 0.00000727, -0.00002117, 0.00001545, -0.00000445, 0.00000423, 0.00000562, -0.00000971, -0.00000222, -0.00000777, 0.00001726, -0.00000228, -0.00000507, -0.00001523, 0.00001082, -0.00000243, 0.00000589, -0.00000719, 0.00000279, -0.00000390, 0.00000502, -0.00001015, -0.00000108, 0.00000046, 0.00001368, -0.00002557, 0.00001512, -0.00000177, 0.00000476, -0.00000818, 0.00000824, -0.00001432, 0.00001382, -0.00001033, 0.00000602, -0.00000821, 0.00000362, -0.00000088, -0.00001521, 0.00001726, -0.00000843, 0.00001020, 0.00000132, -0.00000070, 0.00001218, 0.00001101, -0.00000798, 0.00001535, -0.00002837, 0.00000575, 0.00000140, -0.00001092, 0.00000680, -0.00000308, 0.00000274, -0.00001048, 0.00000762, -0.00000981, 0.00000928, -0.00002143, 0.00001562, -0.00002245, 0.00001297, -0.00003364, 0.00002751, -0.00007517, 0.00004182, -0.00022101, -0.00008812, 0.00091876, -0.00041215, 0.00000454, -0.00007967, 0.00002763, -0.00004554, 0.00002128, -0.00002926, 0.00001474, -0.00001786, 0.00001759, -0.00002186, 0.00001048, -0.00001580, 0.00001396, -0.00001373, 0.00000725, -0.00001256, 0.00002509, -0.00001392, 0.00000356, 0.00000470, -0.00002038, 0.00001504, -0.00001178, 0.00000063, -0.00000159, -0.00000762, -0.00000374, 0.00000923, -0.00002176, 0.00002514, -0.00001772, 0.00000777, -0.00000356, -0.00000089, -0.00000035, 0.00000277, 0.00000741, -0.00000725, -0.00000019, 0.00000185, -0.00000538, -0.00000620, 0.00002329, -0.00000993, 0.00000654, -0.00001188, 0.00000782, -0.00000345, 0.00000021, 0.00000016, 0.00000591, -0.00000429, 0.00000171, -0.00000476, 0.00001206, -0.00001080, 0.00000503, 0.00000428, },
		{ 0.00000161, -0.00001979, 0.00002021, -0.00001107, 0.00001162, -0.00000174, -0.00000527, -0.00000756, -0.00000057, 0.00001564, -0.00000465, -0.00000732, -0.00001143, 0.00001183, -0.00000424, 0.00000675, -0.00000910, 0.00000544, -0.00000593, 0.00000676, -0.00001317, 0.00000193, 0.00000130, 0.00001117, -0.00002736, 0.00002329, -0.00001054, 0.00001083, -0.00001363, 0.00001334, -0.00002044, 0.00002358, -0.00002280, 0.00001958, -0.00002355, 0.00002180, -0.00002232, 0.00000668, 0.00000154, 0.00000356, 0.00000040, 0.00000871, -0.00000969, 0.00002529, -0.00000612, 0.00000747, 0.00000002, -0.00001882, 0.00000198, 0.00000614, -0.00001856, 0.00001868, -0.00001615, 0.00001489, -0.00002353, 0.00002259, -0.00002597, 0.00002563, -0.00004082, 0.00003953, -0.00005045, 0.00004306, -0.00006821, 0.00006756, -0.00013305, 0.00011841, -0.00039464, 0.00023934, 0.00075378, -0.00052257, 0.00012087, -0.00014609, 0.00008271, -0.00008955, 0.00006029, -0.00006312, 0.00004511, -0.00004280, 0.00003930, -0.00004362, 0.00003147, -0.00003441, 0.00003209, -0.00003201, 0.00002403, -0.00002607, 0.00004071, -0.00003464, 0.00002586, -0.00001820, -0.00000111, 0.00000197, -0.00000247, -0.00000704, 0.00000659, -0.00001914, 0.00001195, -0.00000729, -0.00000737, 0.00001794, -0.00001729, 0.00001055, -0.00000723, 0.00000217, -0.00000269, 0.00000683, 0.00000319, -0.00000648, 0.00000025, 0.00000263, -0.00000961, 0.00000199, 0.00001823, -0.00000954, 0.00000560, -0.00001210, 0.00001008, -0.00000584, 0.00000139, 0.00000109, 0.00000431, -0.00000356, 0.00000031, -0.00000144, 0.00000913, -0.00001131, 0.00000694, 0.00000631, },
		{ -0.00000724, -0.00001264, 0.00001914, -0.00001317, 0.00001585, -0.00000736, -0.00000196, -0.00001216, 0.00000817, 0.00001074, -0.00000514, -0.00001027, -0.00000542, 0.00000951, -0.00000321, 0.00000484, -0.00000793, 0.00000516, -0.00000529, 0.00000576, -0.00001355, 0.00000285, 0.00000449, 0.00000434, -0.00002257, 0.00002468, -0.00001409, 0.00001251, -0.00001435, 0.00001278, -0.00001931, 0.00002502, -0.00002702, 0.00002507, -0.00003081, 0.00003237, -0.00003794, 0.00002641, -0.00001506, 0.00001702, -0.00001077, 0.00001623, -0.00001671, 0.00003533, -0.00002093, 0.00002213, -0.00001594, -0.00000768, -0.00000183, 0.00000823, -0.00002152, 0.00002517, -0.00002427, 0.00002191, -0.00003049, 0.00003062, -0.00003451, 0.00003299, -0.00004921, 0.00005097, -0.00006517, 0.00005834, -0.00008507, 0.00008462, -0.00016127, 0.00015319, -0.00052691, 0.00060460, 0.00044436, -0.00048835, 0.00015487, -0.00016267, 0.00009987, -0.00010336, 0.00007414, -0.00007588, 0.00005781, -0.00005246, 0.00004647, -0.00005140, 0.00003979, -0.00004141, 0.00003912, -0.00004009, 0.00003184, -0.00003063, 0.00004631, -0.00004560, 0.00004006, -0.00003550, 0.00001540, -0.00000966, 0.00000517, -0.00001203, 0.00001063, -0.00002575, 0.00002407, -0.00002329, 0.00001041, 0.00000446, -0.00001028, 0.00000805, -0.00000729, 0.00000296, -0.00000328, 0.00000982, -0.00000155, -0.00000438, 0.00000010, 0.00000290, -0.00001283, 0.00001058, 0.00001050, -0.00000594, 0.00000158, -0.00000905, 0.00000961, -0.00000659, 0.00000181, 0.00000257, 0.00000171, -0.00000166, -0.00000195, 0.00000307, 0.00000372, -0.00000896, 0.00000732, 0.00000929, },
		{ -0.00001681, -0.00000158, 0.00001272, -0.00001017, 0.00001569, -0.00000973, -0.00000155, -0.00001369, 0.00001585, 0.00000425, -0.00000449, -0.00001298, 0.00000146, 0.00000491, 0.00000007, 0.00000062, -0.00000394, 0.00000191, -0.00000174, 0.00000168, -0.00001096, 0.00000161, 0.00000954, -0.00000591, -0.00001172, 0.00001878, -0.00001122, 0.00000861, -0.00000953, 0.00000603, -0.00001029, 0.00001672, -0.00002058, 0.00001921, -0.00002569, 0.00002958, -0.00004033, 0.00003552, -0.00002448, 0.00002506, -0.00001759, 0.00001922, -0.00001753, 0.00003763, -0.00002803, 0.00003025, -0.00002701, 0.00000040, -0.00000212, 0.00000447, -0.00001621, 0.00002185, -0.00002238, 0.00001861, -0.00002602, 0.00002600, -0.00002936, 0.00002491, -0.00003931, 0.00004130, -0.00005650, 0.00004728, -0.00007089, 0.00006225, -0.00013752, 0.00011034, -0.00055357, 0.00091971, 0.00006008, -0.00033320, 0.00010234, -0.00011990, 0.00006988, -0.00007822, 0.00005456, -0.00005968, 0.00004559, -0.00004054, 0.00003362, -0.00004002, 0.00003031, -0.00003184, 0.00003021, -0.00003308, 0.00002602, -0.00002208, 0.00003778, -0.00004207, 0.00004067, -0.00004119, 0.00002334, -0.00001493, 0.00000713, -0.00001092, 0.00000706, -0.00002301, 0.00002666, -0.00003173, 0.00002447, -0.00000955, -0.00000043, 0.00000221, -0.00000450, 0.00000159, -0.00000177, 0.00001085, -0.00000569, -0.00000166, -0.00000042, 0.00000239, -0.00001412, 0.00001780, 0.00000206, -0.00000068, -0.00000437, -0.00000337, 0.00000655, -0.00000562, 0.00000158, 0.00000410, -0.00000109, 0.00000055, -0.00000408, 0.00000736, -0.00000262, -0.00000482, 0.00000717, 0.00001140, },
		{ -0.00002390, 0.00001011, 0.00000333, -0.00000338, 0.00001166, -0.00000885, -0.00000440, -0.00001083, 0.00002021, -0.00000184, -0.00000384, -0.00001424, 0.00000742, -0.00000020, 0.00000406, -0.00000441, 0.00000140, -0.00000303, 0.00000361, -0.00000451, -0.00000623, -0.00000077, 0.00001472, -0.00001702, 0.00000246, 0.00000763, -0.00000326, 0.00000041, -0.00000080, -0.00000462, 0.00000392, 0.00000127, -0.00000558, 0.00000358, -0.00000904, 0.00001303, -0.00002713, 0.00002974, -0.00002163, 0.00002282, -0.00001591, 0.00001458, -0.00000978, 0.00002977, -0.00002443, 0.00002832, -0.00002964, 0.00000262, 0.00000263, -0.00000565, -0.00000233, 0.00000806, -0.00000945, 0.00000394, -0.00000924, 0.00000793, -0.00000988, 0.00000103, -0.00001080, 0.00000970, -0.00002306, 0.00000816, -0.00002378, -0.00000222, -0.00005666, -0.00002541, -0.00042745, 0.00109602, -0.00030487, -0.00011707, -0.00000813, -0.00003364, 0.00000358, -0.00002201, 0.00000745, -0.00001889, 0.00001179, -0.00001004, 0.00000380, -0.00001226, 0.00000538, -0.00000774, 0.00000711, -0.00001226, 0.00000755, -0.00000158, 0.00001626, -0.00002456, 0.00002721, -0.00003378, 0.00002073, -0.00001215, 0.00000245, -0.00000346, -0.00000408, -0.00001022, 0.00001739, -0.00002828, 0.00002912, -0.00001833, 0.00000762, -0.00000372, -0.00000089, -0.00000072, 0.00000124, 0.00000969, -0.00000841, 0.00000094, -0.00000099, 0.00000097, -0.00001289, 0.00002213, -0.00000504, 0.00000421, -0.00001038, 0.00000341, 0.00000181, -0.00000338, 0.00000104, 0.00000510, -0.00000323, 0.00000214, -0.00000503, 0.00000996, -0.00000815, -0.00000032, 0.00000767, 0.00001062, },
		{ -0.00002579, 0.00001881, -0.00000569, 0.00000457, 0.00000564, -0.00000616, -0.00000936, -0.00000385, 0.00002014, -0.00000603, -0.00000413, -0.00001314, 0.00001089, -0.00000394, 0.00000692, -0.00000825, 0.00000595, -0.00000753, 0.00000870, -0.00001085, -0.00000112, -0.00000250, 0.00001762, -0.00002555, 0.00001582, -0.00000478, 0.00000635, -0.00000885, 0.00000825, -0.00001484, 0.00001801, -0.00001549, 0.00001197, -0.00001569, 0.00001295, -0.00001158, -0.00000257, 0.00001108, -0.00000695, 0.00001005, -0.00000547, 0.00000273, 0.00000532, 0.00001316, -0.00001117, 0.00001681, -0.00002388, -0.00000057, 0.00001068, -0.00001902, 0.00001610, -0.00001197, 0.00001031, -0.00001757, 0.00001481, -0.00001803, 0.00001767, -0.00003129, 0.00002784, -0.00003457, 0.00002515, -0.00004774, 0.00004284, -0.00009247, 0.00006148, -0.00023570, -0.00014004, 0.00107279, -0.00055986, 0.00008436, -0.00012714, 0.00006158, -0.00007214, 0.00004334, -0.00004880, 0.00003099, -0.00003037, 0.00002724, -0.00003196, 0.00002156, -0.00002552, 0.00002207, -0.00002197, 0.00001494, -0.00001679, 0.00002427, -0.00001172, 0.00000098, 0.00000461, -0.00001691, 0.00001020, -0.00000368, -0.00000619, 0.00000693, -0.00001869, 0.00000858, -0.00000102, -0.00001348, 0.00002253, -0.00001862, 0.00001042, -0.00000687, 0.00000141, -0.00000239, 0.00000461, 0.00000674, -0.00000927, 0.00000275, -0.00000130, -0.00000131, -0.00000905, 0.00002271, -0.00000918, 0.00000688, -0.00001448, 0.00000939, -0.00000322, -0.00000064, 0.00000064, 0.00000511, -0.00000410, 0.00000250, -0.00000417, 0.00000987, -0.00001144, 0.00000340, 0.00000946, 0.00000573, },
		{ -0.00002133, 0.00002200, -0.00001137, 0.00001090, 0.00000003, -0.00000385, -0.00001415, 0.00000550, 0.00001591, -0.00000769, -0.00000567, -0.00000947, 0.00001105, -0.00000504, 0.00000724, -0.00000927, 0.00000774, -0.00000944, 0.00001129, -0.00001515, 0.00000245, -0.00000187, 0.00001613, -0.00002835, 0.00002419, -0.00001404, 0.00001352, -0.00001532, 0.00001367, -0.00001997, 0.00002620, -0.00002676, 0.00002454, -0.00003026, 0.00003108, -0.00003443, 0.00002397, -0.00001265, 0.00001355, -0.00000845, 0.00000963, -0.00001213, 0.00002293, -0.00000698, 0.00000676, 0.00000009, -0.00001344, -0.00000558, 0.00001751, -0.00002965, 0.00003181, -0.00003025, 0.00002858, -0.00003710, 0.00003652, -0.00004141, 0.00004177, -0.00005894, 0.00006150, -0.00007448, 0.00006912, -0.00009863, 0.00010283, -0.00017580, 0.00017375, -0.00046610, 0.00026702, 0.00083991, -0.00064646, 0.00020590, -0.00020430, 0.00012700, -0.00012548, 0.00009083, -0.00009078, 0.00006954, -0.00006312, 0.00005545, -0.00005891, 0.00004754, -0.00004940, 0.00004533, -0.00004537, 0.00003763, -0.00003694, 0.00004573, -0.00003651, 0.00002526, -0.00001864, 0.00000219, -0.00000236, 0.00000532, -0.00001369, 0.00001460, -0.00003007, 0.00002598, -0.00002149, 0.00000724, 0.00000761, -0.00001105, 0.00000729, -0.00000612, 0.00000130, -0.00000236, 0.00000726, 0.00000270, -0.00000819, 0.00000331, -0.00000107, -0.00000420, -0.00000306, 0.00001949, -0.00000964, 0.00000618, -0.00001515, 0.00001289, -0.00000712, 0.00000175, 0.00000068, 0.00000401, -0.00000355, 0.00000154, -0.00000161, 0.00000699, -0.00001193, 0.00000596, 0.00001214, -0.00000297, },
		{ -0.00001150, 0.00001909, -0.00001217, 0.00001375, -0.00000337, -0.00000378, -0.00001632, 0.00001463, 0.00000901, -0.00000721, -0.00000801, -0.00000394, 0.00000819, -0.00000331, 0.00000469, -0.00000690, 0.00000587, -0.00000755, 0.00000993, -0.00001589, 0.00000325, 0.00000190, 0.00000951, -0.00002393, 0.00002497, -0.00001695, 0.00001518, -0.00001624, 0.00001289, -0.00001718, 0.00002473, -0.00002771, 0.00002633, -0.00003324, 0.00003702, -0.00004563, 0.00004167, -0.00003112, 0.00003075, -0.00002482, 0.00002260, -0.00002354, 0.00003627, -0.00002349, 0.00002226, -0.00001527, -0.00000392, -0.00000748, 0.00001789, -0.00003125, 0.00003718, -0.00003819, 0.00003625, -0.00004501, 0.00004550, -0.00005101, 0.00005024, -0.00006814, 0.00007419, -0.00009166, 0.00008790, -0.00012024, 0.00012664, -0.00021401, 0.00022611, -0.00063769, 0.00070657, 0.00044586, -0.00055723, 0.00021489, -0.00020801, 0.00013590, -0.00013324, 0.00009994, -0.00009997, 0.00008014, -0.00007180, 0.00006157, -0.00006493, 0.00005415, -0.00005530, 0.00005154, -0.00005285, 0.00004604, -0.00004375, 0.00005393, -0.00004915, 0.00003936, -0.00003398, 0.00001589, -0.00001057, 0.00000949, -0.00001515, 0.00001435, -0.00003188, 0.00003421, -0.00003541, 0.00002565, -0.00000920, 0.00000030, 0.00000044, -0.00000249, -0.00000080, -0.00000057, 0.00000858, -0.00000170, -0.00000545, 0.00000242, -0.00000020, -0.00000728, 0.00000413, 0.00001334, -0.00000679, 0.00000210, -0.00001190, 0.00001290, -0.00000887, 0.00000318, 0.00000124, 0.00000207, -0.00000193, -0.00000024, 0.00000186, 0.00000211, -0.00000994, 0.00000776, 0.00001431, -0.00001351, },
		{ 0.00000086, 0.00001162, -0.00000847, 0.00001285, -0.00000407, -0.00000666, -0.00001426, 0.00002103, 0.00000151, -0.00000564, -0.00001017, 0.00000208, 0.00000353, 0.00000034, 0.00000011, -0.00000193, 0.00000089, -0.00000211, 0.00000460, -0.00001287, 0.00000133, 0.00000814, -0.00000114, -0.00001312, 0.00001814, -0.00001280, 0.00001056, -0.00001120, 0.00000603, -0.00000688, 0.00001368, -0.00001769, 0.00001600, -0.00002240, 0.00002712, -0.00003951, 0.00004291, -0.00003593, 0.00003652, -0.00003168, 0.00002703, -0.00002577, 0.00003954, -0.00003024, 0.00002913, -0.00002349, -0.00000006, -0.00000275, 0.00000879, -0.00002039, 0.00002785, -0.00003062, 0.00002774, -0.00003543, 0.00003546, -0.00004014, 0.00003597, -0.00005092, 0.00005634, -0.00007464, 0.00006810, -0.00009670, 0.00009429, -0.00017959, 0.00017404, -0.00067093, 0.00106746, -0.00001348, -0.00033754, 0.00012117, -0.00013734, 0.00008480, -0.00009090, 0.00006564, -0.00007103, 0.00005741, -0.00005152, 0.00004137, -0.00004610, 0.00003746, -0.00003933, 0.00003674, -0.00004023, 0.00003582, -0.00003307, 0.00004472, -0.00004513, 0.00003838, -0.00003633, 0.00001945, -0.00001064, 0.00000604, -0.00000852, 0.00000426, -0.00002128, 0.00002862, -0.00003631, 0.00003423, -0.00002080, 0.00000984, -0.00000626, 0.00000149, -0.00000327, 0.00000207, 0.00000857, -0.00000580, -0.00000161, 0.00000030, 0.00000115, -0.00000995, 0.00001124, 0.00000576, -0.00000195, -0.00000427, -0.00000540, 0.00000948, -0.00000811, 0.00000339, 0.00000217, -0.00000017, 0.00000009, -0.00000202, 0.00000504, -0.00000341, -0.00000643, 0.00000960, 0.00001411, -0.00002287, },
		{ 0.00001212, 0.00000255, -0.00000226, 0.00000947, -0.00000306, -0.00001166, -0.00000783, 0.00002310, -0.00000471, -0.00000416, -0.00001108, 0.00000699, -0.00000117, 0.00000426, -0.00000474, 0.00000374, -0.00000534, 0.00000511, -0.00000322, -0.00000730, -0.00000197, 0.00001481, -0.00001294, 0.00000094, 0.00000634, -0.00000362, 0.00000151, -0.00000235, -0.00000402, 0.00000719, -0.00000288, -0.00000075, -0.00000248, -0.00000152, 0.00000442, -0.00001741, 0.00002674, -0.00002438, 0.00002741, -0.00002553, 0.00002007, -0.00001661, 0.00003065, -0.00002488, 0.00002481, -0.00002220, -0.00000326, 0.00000848, -0.00000845, 0.00000113, 0.00000542, -0.00000872, 0.00000420, -0.00000966, 0.00000785, -0.00001052, 0.00000124, -0.00000990, 0.00001035, -0.00002538, 0.00001139, -0.00002958, 0.00000663, -0.00007060, 0.00000371, -0.00051457, 0.00124746, -0.00042247, -0.00006952, -0.00002890, -0.00002138, -0.00000515, -0.00001482, 0.00000103, -0.00001448, 0.00000992, -0.00000972, 0.00000183, -0.00000890, 0.00000332, -0.00000677, 0.00000567, -0.00001149, 0.00001029, -0.00000785, 0.00002071, -0.00002637, 0.00002352, -0.00002614, 0.00001303, -0.00000314, -0.00000366, 0.00000397, -0.00001280, -0.00000078, 0.00001021, -0.00002284, 0.00002942, -0.00002251, 0.00001314, -0.00000925, 0.00000324, -0.00000426, 0.00000434, 0.00000756, -0.00000900, 0.00000246, -0.00000243, 0.00000255, -0.00001153, 0.00001692, -0.00000142, 0.00000299, -0.00001099, 0.00000265, 0.00000364, -0.00000525, 0.00000256, 0.00000312, -0.00000210, 0.00000179, -0.00000303, 0.00000680, -0.00000813, -0.00000241, 0.00001205, 0.00001013, -0.00002798, },
		{ 0.00001914, -0.00000497, 0.00000389, 0.00000564, -0.00000221, -0.00001673, 0.00000156, 0.00002062, -0.00000850, -0.00000358, -0.00001004, 0.00000956, -0.00000431, 0.00000676, -0.00000793, 0.00000785, -0.00001028, 0.00001152, -0.00001104, -0.00000134, -0.00000456, 0.00001923, -0.00002228, 0.00001392, -0.00000612, 0.00000672, -0.00000832, 0.00000651, -0.00001278, 0.00001942, -0.00001837, 0.00001595, -0.00002131, 0.00002093, -0.00002229, 0.00001249, -0.00000048, -0.00000087, 0.00000643, -0.00000860, 0.00000395, 0.00000124, 0.00001246, -0.00001006, 0.00001161, -0.00001358, -0.00001071, 0.00002189, -0.00002785, 0.00002612, -0.00002237, 0.00001958, -0.00002607, 0.00002336, -0.00002765, 0.00002720, -0.00004185, 0.00004109, -0.00004850, 0.00003949, -0.00006364, 0.00005949, -0.00011069, 0.00008236, -0.00025488, -0.00016917, 0.00118660, -0.00068108, 0.00015647, -0.00017161, 0.00009331, -0.00009703, 0.00006448, -0.00006790, 0.00004736, -0.00004316, 0.00003664, -0.00004142, 0.00003205, -0.00003472, 0.00002971, -0.00002979, 0.00002247, -0.00002071, 0.00002279, -0.00000960, -0.00000056, 0.00000145, -0.00000903, 0.00000141, 0.00000733, -0.00001463, 0.00001647, -0.00002994, 0.00002222, -0.00001439, 0.00000028, 0.00001326, -0.00001415, 0.00000892, -0.00000710, 0.00000153, -0.00000272, 0.00000536, 0.00000583, -0.00001069, 0.00000569, -0.00000481, 0.00000332, -0.00001133, 0.00002000, -0.00000655, 0.00000610, -0.00001589, 0.00001006, -0.00000290, -0.00000130, 0.00000117, 0.00000371, -0.00000322, 0.00000268, -0.00000284, 0.00000649, -0.00001103, 0.00000145, 0.00001498, 0.00000217, -0.00002699, },
		{ 0.00002041, -0.00000879, 0.00000791, 0.00000310, -0.00000332, -0.00001945, 0.00001150, 0.00001464, -0.00000966, -0.00000412, -0.00000704, 0.00000929, -0.00000495, 0.00000675, -0.00000814, 0.00000865, -0.00001178, 0.00001463, -0.00001636, 0.00000284, -0.00000452, 0.00001918, -0.00002605, 0.00002180, -0.00001488, 0.00001405, -0.00001508, 0.00001167, -0.00001614, 0.00002469, -0.00002653, 0.00002517, -0.00003216, 0.00003525, -0.00004187, 0.00003837, -0.00002756, 0.00002484, -0.00001815, 0.00001203, -0.00001472, 0.00002104, -0.00000817, 0.00000763, -0.00000429, -0.00000330, -0.00001665, 0.00003059, -0.00004087, 0.00004446, -0.00004431, 0.00004242, -0.00005051, 0.00005011, -0.00005646, 0.00005712, -0.00007542, 0.00008157, -0.00009700, 0.00009384, -0.00012706, 0.00013496, -0.00021364, 0.00022199, -0.00052961, 0.00030446, 0.00088930, -0.00073585, 0.00027236, -0.00025073, 0.00016198, -0.00015356, 0.00011496, -0.00011301, 0.00008966, -0.00007998, 0.00006816, -0.00007053, 0.00006002, -0.00006091, 0.00005520, -0.00005538, 0.00004819, -0.00004473, 0.00004720, -0.00003523, 0.00002215, -0.00001858, 0.00000692, -0.00000847, 0.00001447, -0.00002059, 0.00002212, -0.00003903, 0.00003829, -0.00003527, 0.00002405, -0.00000740, -0.00000002, -0.00000037, -0.00000114, -0.00000281, 0.00000093, 0.00000509, 0.00000341, -0.00001031, 0.00000701, -0.00000577, 0.00000272, -0.00000886, 0.00001982, -0.00000855, 0.00000606, -0.00001722, 0.00001474, -0.00000824, 0.00000249, -0.00000013, 0.00000367, -0.00000333, 0.00000263, -0.00000158, 0.00000415, -0.00001176, 0.00000507, 0.00001734, -0.00000838, -0.00002000, },
		{ 0.00001651, -0.00000847, 0.00000906, 0.00000250, -0.00000717, -0.00001808, 0.00001943, 0.00000700, -0.00000880, -0.00000534, -0.00000273, 0.00000662, -0.00000316, 0.00000421, -0.00000519, 0.00000561, -0.00000888, 0.00001304, -0.00001754, 0.00000387, -0.00000099, 0.00001386, -0.00002291, 0.00002236, -0.00001714, 0.00001561, -0.00001632, 0.00001114, -0.00001228, 0.00002062, -0.00002409, 0.00002279, -0.00002989, 0.00003476, -0.00004561, 0.00004962, -0.00004315, 0.00004181, -0.00003629, 0.00002749, -0.00002785, 0.00003484, -0.00002329, 0.00002045, -0.00001573, 0.00000229, -0.00001528, 0.00002841, -0.00004019, 0.00004731, -0.00005032, 0.00004891, -0.00005753, 0.00005811, -0.00006515, 0.00006470, -0.00008316, 0.00009262, -0.00011321, 0.00011242, -0.00014940, 0.00016083, -0.00025578, 0.00028279, -0.00072482, 0.00079699, 0.00042653, -0.00059554, 0.00025391, -0.00023834, 0.00015961, -0.00015236, 0.00011644, -0.00011579, 0.00009544, -0.00008534, 0.00007134, -0.00007311, 0.00006324, -0.00006402, 0.00005886, -0.00006047, 0.00005530, -0.00005193, 0.00005603, -0.00004721, 0.00003332, -0.00002879, 0.00001486, -0.00001096, 0.00001353, -0.00001728, 0.00001646, -0.00003446, 0.00004002, -0.00004342, 0.00003882, -0.00002385, 0.00001307, -0.00000989, 0.00000521, -0.00000734, 0.00000503, 0.00000426, 0.00000017, -0.00000756, 0.00000573, -0.00000456, 0.00000030, -0.00000409, 0.00001636, -0.00000726, 0.00000252, -0.00001423, 0.00001536, -0.00001091, 0.00000497, -0.00000081, 0.00000290, -0.00000248, 0.00000186, 0.00000021, 0.00000044, -0.00001057, 0.00000873, 0.00001758, -0.00001885, -0.00000907, },
		{ 0.00000966, -0.00000523, 0.00000810, 0.00000309, -0.00001308, -0.00001222, 0.00002352, -0.00000038, -0.00000692, -0.00000652, 0.00000176, 0.00000267, 0.00000011, 0.00000012, -0.00000016, -0.00000028, -0.00000226, 0.00000698, -0.00001445, 0.00000181, 0.00000539, 0.00000437, -0.00001381, 0.00001594, -0.00001268, 0.00001106, -0.00001200, 0.00000554, -0.00000253, 0.00000872, -0.00001230, 0.00000975, -0.00001498, 0.00001892, -0.00003110, 0.00004139, -0.00004057, 0.00004258, -0.00004058, 0.00003109, -0.00002951, 0.00003699, -0.00002733, 0.00002309, -0.00001793, -0.00000054, -0.00000407, 0.00001328, -0.00002339, 0.00003139, -0.00003617, 0.00003427, -0.00004192, 0.00004174, -0.00004769, 0.00004358, -0.00005796, 0.00006556, -0.00008640, 0.00008225, -0.00011479, 0.00011662, -0.00020808, 0.00021645, -0.00075192, 0.00118050, -0.00008386, -0.00032592, 0.00012588, -0.00014374, 0.00009023, -0.00009489, 0.00006888, -0.00007505, 0.00006272, -0.00005715, 0.00004443, -0.00004757, 0.00004002, -0.00004223, 0.00003866, -0.00004255, 0.00004083, -0.00003919, 0.00004610, -0.00004227, 0.00002976, -0.00002618, 0.00001228, -0.00000443, 0.00000384, -0.00000481, -0.00000000, -0.00001606, 0.00002547, -0.00003466, 0.00003848, -0.00002924, 0.00001898, -0.00001472, 0.00000825, -0.00000926, 0.00000759, 0.00000382, -0.00000387, -0.00000272, 0.00000194, -0.00000113, -0.00000381, 0.00000244, 0.00001044, -0.00000351, -0.00000366, -0.00000746, 0.00001179, -0.00001027, 0.00000549, -0.00000056, 0.00000149, -0.00000097, 0.00000077, 0.00000181, -0.00000368, -0.00000799, 0.00001270, 0.00001434, -0.00002626, 0.00000256, },
		{ 0.00000271, -0.00000117, 0.00000664, 0.00000318, -0.00001910, -0.00000302, 0.00002317, -0.00000609, -0.00000494, -0.00000693, 0.00000530, -0.00000115, 0.00000343, -0.00000389, 0.00000501, -0.00000688, 0.00000598, -0.00000178, -0.00000848, -0.00000196, 0.00001264, -0.00000663, -0.00000173, 0.00000528, -0.00000379, 0.00000245, -0.00000432, -0.00000222, 0.00000910, -0.00000621, 0.00000362, -0.00000828, 0.00000648, -0.00000614, -0.00000350, 0.00001678, -0.00002071, 0.00002644, -0.00002950, 0.00002135, -0.00001869, 0.00002675, -0.00001963, 0.00001506, -0.00001084, -0.00001075, 0.00001443, -0.00001097, 0.00000506, 0.00000116, -0.00000599, 0.00000254, -0.00000795, 0.00000548, -0.00000900, -0.00000054, -0.00000631, 0.00000711, -0.00002313, 0.00000991, -0.00003009, 0.00000887, -0.00007477, 0.00001493, -0.00056214, 0.00134747, -0.00051271, -0.00002569, -0.00005078, -0.00000756, -0.00001574, -0.00000529, -0.00000784, -0.00000729, 0.00000524, -0.00000727, -0.00000170, -0.00000404, -0.00000034, -0.00000406, 0.00000237, -0.00000846, 0.00001054, -0.00001143, 0.00002164, -0.00002402, 0.00001464, -0.00001361, 0.00000199, 0.00000784, -0.00001055, 0.00001186, -0.00002144, 0.00001019, -0.00000045, -0.00001171, 0.00002308, -0.00002166, 0.00001502, -0.00001232, 0.00000591, -0.00000691, 0.00000731, 0.00000426, -0.00000819, 0.00000315, -0.00000333, 0.00000363, -0.00000872, 0.00000954, 0.00000350, 0.00000109, -0.00001065, 0.00000137, 0.00000513, -0.00000671, 0.00000408, 0.00000056, -0.00000027, 0.00000083, -0.00000029, 0.00000260, -0.00000732, -0.00000447, 0.00001671, 0.00000721, -0.00002839, 0.00001163, },
		{ -0.00000203, 0.00000182, 0.00000622, 0.00000103, -0.00002280, 0.00000725, 0.00001902, -0.00000948, -0.00000347, -0.00000621, 0.00000709, -0.00000365, 0.00000550, -0.00000625, 0.00000826, -0.00001168, 0.00001304, -0.00001052, -0.00000190, -0.00000532, 0.00001823, -0.00001586, 0.00000944, -0.00000563, 0.00000581, -0.00000675, 0.00000333, -0.00000830, 0.00001770, -0.00001809, 0.00001668, -0.00002331, 0.00002527, -0.00003004, 0.00002638, -0.00001431, 0.00000831, -0.00000025, -0.00000799, 0.00000262, 0.00000018, 0.00000865, -0.00000471, 0.00000090, 0.00000078, -0.00002278, 0.00003305, -0.00003547, 0.00003489, -0.00003232, 0.00002874, -0.00003429, 0.00003166, -0.00003703, 0.00003612, -0.00005109, 0.00005300, -0.00006179, 0.00005354, -0.00007885, 0.00007504, -0.00012749, 0.00010249, -0.00027865, -0.00016908, 0.00124574, -0.00075907, 0.00020768, -0.00020379, 0.00011596, -0.00011478, 0.00008009, -0.00008247, 0.00006036, -0.00005346, 0.00004359, -0.00004816, 0.00003987, -0.00004143, 0.00003512, -0.00003555, 0.00002843, -0.00002351, 0.00002051, -0.00000762, -0.00000110, -0.00000418, 0.00000179, -0.00000943, 0.00001929, -0.00002271, 0.00002496, -0.00003883, 0.00003423, -0.00002771, 0.00001666, -0.00000102, -0.00000483, 0.00000317, -0.00000376, -0.00000111, -0.00000073, 0.00000439, 0.00000519, -0.00001162, 0.00000829, -0.00000823, 0.00000807, -0.00001290, 0.00001561, -0.00000268, 0.00000463, -0.00001620, 0.00000981, -0.00000259, -0.00000152, 0.00000140, 0.00000213, -0.00000195, 0.00000251, -0.00000109, 0.00000229, -0.00000987, -0.00000021, 0.00001986, -0.00000280, -0.00002468, 0.00001613, },
		{ -0.00000370, 0.00000295, 0.00000739, -0.00000416, -0.00002229, 0.00001610, 0.00001255, -0.00001063, -0.00000264, -0.00000441, 0.00000690, -0.00000425, 0.00000558, -0.00000600, 0.00000821, -0.00001272, 0.00001644, -0.00001660, 0.00000297, -0.00000631, 0.00002000, -0.00002059, 0.00001631, -0.00001305, 0.00001246, -0.00001319, 0.00000798, -0.00000976, 0.00001959, -0.00002224, 0.00002125, -0.00002861, 0.00003306, -0.00004248, 0.00004647, -0.00003930, 0.00003457, -0.00002685, 0.00001474, -0.00001686, 0.00001917, -0.00000951, 0.00000984, -0.00001195, 0.00000966, -0.00002901, 0.00004297, -0.00004972, 0.00005395, -0.00005551, 0.00005358, -0.00006095, 0.00006057, -0.00006824, 0.00006865, -0.00008688, 0.00009576, -0.00011358, 0.00011241, -0.00014819, 0.00015806, -0.00023995, 0.00025545, -0.00057733, 0.00034809, 0.00089693, -0.00078011, 0.00030972, -0.00027690, 0.00018093, -0.00016823, 0.00012779, -0.00012535, 0.00010141, -0.00008996, 0.00007466, -0.00007601, 0.00006648, -0.00006660, 0.00005952, -0.00005997, 0.00005374, -0.00004801, 0.00004450, -0.00003071, 0.00001670, -0.00001792, 0.00001215, -0.00001493, 0.00002335, -0.00002608, 0.00002744, -0.00004383, 0.00004607, -0.00004509, 0.00003912, -0.00002382, 0.00001347, -0.00001090, 0.00000675, -0.00000956, 0.00000681, 0.00000042, 0.00000543, -0.00001271, 0.00001081, -0.00001069, 0.00001023, -0.00001468, 0.00001905, -0.00000653, 0.00000546, -0.00001828, 0.00001545, -0.00000904, 0.00000354, -0.00000145, 0.00000354, -0.00000310, 0.00000376, -0.00000161, 0.00000094, -0.00001107, 0.00000479, 0.00002079, -0.00001348, -0.00001642, 0.00001583, },
		{ -0.00000297, 0.00000276, 0.00000936, -0.00001172, -0.00001702, 0.00002160, 0.00000544, -0.00001006, -0.00000231, -0.00000195, 0.00000509, -0.00000304, 0.00000374, -0.00000323, 0.00000470, -0.00000933, 0.00001490, -0.00001836, 0.00000466, -0.00000390, 0.00001708, -0.00001964, 0.00001716, -0.00001481, 0.00001395, -0.00001493, 0.00000831, -0.00000586, 0.00001401, -0.00001737, 0.00001555, -0.00002170, 0.00002616, -0.00003770, 0.00004856, -0.00004812, 0.00004737, -0.00004299, 0.00002934, -0.00002874, 0.00003047, -0.00002025, 0.00001692, -0.00001682, 0.00000966, -0.00002361, 0.00003787, -0.00004629, 0.00005333, -0.00005815, 0.00005727, -0.00006533, 0.00006572, -0.00007410, 0.00007345, -0.00009105, 0.00010242, -0.00012536, 0.00012680, -0.00016664, 0.00018036, -0.00027842, 0.00031286, -0.00077535, 0.00086393, 0.00039144, -0.00060135, 0.00026743, -0.00024919, 0.00016717, -0.00015746, 0.00012070, -0.00012049, 0.00010089, -0.00009060, 0.00007382, -0.00007425, 0.00006541, -0.00006593, 0.00005955, -0.00006137, 0.00005799, -0.00005388, 0.00005210, -0.00004001, 0.00002276, -0.00002085, 0.00001259, -0.00001057, 0.00001677, -0.00001786, 0.00001652, -0.00003271, 0.00004013, -0.00004498, 0.00004651, -0.00003594, 0.00002499, -0.00002048, 0.00001391, -0.00001518, 0.00001237, -0.00000236, 0.00000372, -0.00001039, 0.00000938, -0.00000924, 0.00000867, -0.00001282, 0.00001885, -0.00000717, 0.00000284, -0.00001586, 0.00001664, -0.00001233, 0.00000681, -0.00000337, 0.00000415, -0.00000331, 0.00000431, -0.00000191, -0.00000115, -0.00001085, 0.00001039, 0.00001826, -0.00002205, -0.00000629, 0.00001222, },
		{ -0.00000145, 0.00000268, 0.00001035, -0.00001959, -0.00000804, 0.00002301, -0.00000095, -0.00000838, -0.00000223, 0.00000060, 0.00000239, -0.00000072, 0.00000082, 0.00000100, -0.00000116, -0.00000238, 0.00000874, -0.00001563, 0.00000302, 0.00000157, 0.00001016, -0.00001374, 0.00001244, -0.00001102, 0.00001020, -0.00001208, 0.00000512, 0.00000158, 0.00000341, -0.00000616, 0.00000243, -0.00000557, 0.00000722, -0.00001711, 0.00003170, -0.00003741, 0.00004169, -0.00004291, 0.00003031, -0.00002825, 0.00002986, -0.00001982, 0.00001331, -0.00001109, -0.00000097, -0.00000576, 0.00001731, -0.00002450, 0.00003163, -0.00003790, 0.00003685, -0.00004406, 0.00004335, -0.00005050, 0.00004622, -0.00005885, 0.00006697, -0.00008918, 0.00008654, -0.00012126, 0.00012440, -0.00021684, 0.00022875, -0.00078340, 0.00124316, -0.00013896, -0.00030412, 0.00011840, -0.00013964, 0.00008631, -0.00009025, 0.00006428, -0.00007150, 0.00006093, -0.00005677, 0.00004242, -0.00004424, 0.00003778, -0.00004025, 0.00003569, -0.00003963, 0.00004032, -0.00003978, 0.00004181, -0.00003419, 0.00001631, -0.00001277, 0.00000347, 0.00000253, 0.00000145, -0.00000068, -0.00000461, -0.00000843, 0.00001801, -0.00002675, 0.00003575, -0.00003262, 0.00002460, -0.00002095, 0.00001385, -0.00001473, 0.00001333, -0.00000230, -0.00000049, -0.00000465, 0.00000400, -0.00000370, 0.00000312, -0.00000714, 0.00001499, -0.00000483, -0.00000276, -0.00000926, 0.00001309, -0.00001160, 0.00000732, -0.00000358, 0.00000356, -0.00000237, 0.00000397, -0.00000209, -0.00000360, -0.00000916, 0.00001608, 0.00001188, -0.00002624, 0.00000272, 0.00000764, },
		{ -0.00000073, 0.00000415, 0.00000847, -0.00002516, 0.00000245, 0.00002076, -0.00000589, -0.00000611, -0.00000218, 0.00000269, -0.00000033, 0.00000177, -0.00000198, 0.00000505, -0.00000734, 0.00000598, -0.00000019, -0.00000969, -0.00000078, 0.00000858, 0.00000124, -0.00000518, 0.00000445, -0.00000373, 0.00000307, -0.00000643, 0.00000072, 0.00000912, -0.00000769, 0.00000609, -0.00001195, 0.00001258, -0.00001570, 0.00001120, 0.00000270, -0.00001184, 0.00002012, -0.00002772, 0.00001826, -0.00001623, 0.00001857, -0.00000987, 0.00000121, 0.00000234, -0.00001829, 0.00001924, -0.00001228, 0.00000853, -0.00000386, -0.00000192, -0.00000056, -0.00000451, 0.00000118, -0.00000573, -0.00000376, -0.00000076, 0.00000066, -0.00001672, 0.00000382, -0.00002499, 0.00000364, -0.00006760, 0.00000462, -0.00056173, 0.00138244, -0.00056173, 0.00000462, -0.00006760, 0.00000364, -0.00002499, 0.00000382, -0.00001672, 0.00000066, -0.00000076, -0.00000376, -0.00000573, 0.00000118, -0.00000451, -0.00000056, -0.00000192, -0.00000386, 0.00000853, -0.00001228, 0.00001924, -0.00001829, 0.00000234, 0.00000121, -0.00000987, 0.00001857, -0.00001623, 0.00001826, -0.00002772, 0.00002012, -0.00001184, 0.00000270, 0.00001120, -0.00001570, 0.00001258, -0.00001195, 0.00000609, -0.00000769, 0.00000912, 0.00000072, -0.00000643, 0.00000307, -0.00000373, 0.00000445, -0.00000518, 0.00000124, 0.00000858, -0.00000078, -0.00000969, -0.00000019, 0.00000598, -0.00000734, 0.00000505, -0.00000198, 0.00000177, -0.00000033, 0.00000269, -0.00000218, -0.00000611, -0.00000589, 0.00002076, 0.00000245, -0.00002516, 0.00000847, 0.00000415, },
		{ -0.00000145, 0.00000764, 0.00000272, -0.00002624, 0.00001188, 0.00001608, -0.00000916, -0.00000360, -0.00000209, 0.00000397, -0.00000237, 0.00000356, -0.00000358, 0.00000732, -0.00001160, 0.00001309, -0.00000926, -0.00000276, -0.00000483, 0.00001499, -0.00000714, 0.00000312, -0.00000370, 0.00000400, -0.00000465, -0.00000049, -0.00000230, 0.00001333, -0.00001473, 0.00001385, -0.00002095, 0.00002460, -0.00003262, 0.00003575, -0.00002675, 0.00001801, -0.00000843, -0.00000461, -0.00000068, 0.00000145, 0.00000253, 0.00000347, -0.00001277, 0.00001631, -0.00003419, 0.00004181, -0.00003978, 0.00004032, -0.00003963, 0.00003569, -0.00004025, 0.00003778, -0.00004424, 0.00004242, -0.00005677, 0.00006093, -0.00007150, 0.00006428, -0.00009025, 0.00008631, -0.00013964, 0.00011840, -0.00030412, -0.00013896, 0.00124316, -0.00078340, 0.00022875, -0.00021684, 0.00012440, -0.00012126, 0.00008654, -0.00008918, 0.00006697, -0.00005885, 0.00004622, -0.00005050, 0.00004335, -0.00004406, 0.00003685, -0.00003790, 0.00003163, -0.00002450, 0.00001731, -0.00000576, -0.00000097, -0.00001109, 0.00001331, -0.00001982, 0.00002986, -0.00002825, 0.00003031, -0.00004291, 0.00004169, -0.00003741, 0.00003170, -0.00001711, 0.00000722, -0.00000557, 0.00000243, -0.00000616, 0.00000341, 0.00000158, 0.00000512, -0.00001208, 0.00001020, -0.00001102, 0.00001244, -0.00001374, 0.00001016, 0.00000157, 0.00000302, -0.00001563, 0.00000874, -0.00000238, -0.00000116, 0.00000100, 0.00000082, -0.00000072, 0.00000239, 0.00000060, -0.00000223, -0.00000838, -0.00000095, 0.00002301, -0.00000804, -0.00001959, 0.00001035, 0.00000268, },
		{ -0.00000297, 0.00001222, -0.00000629, -0.00002205, 0.00001826, 0.00001039, -0.00001085, -0.00000115, -0.00000191, 0.00000431, -0.00000331, 0.00000415, -0.00000337, 0.00000681, -0.00001233, 0.00001664, -0.00001586, 0.00000284, -0.00000717, 0.00001885, -0.00001282, 0.00000867, -0.00000924, 0.00000938, -0.00001039, 0.00000372, -0.00000236, 0.00001237, -0.00001518, 0.00001391, -0.00002048, 0.00002499, -0.00003594, 0.00004651, -0.00004498, 0.00004013, -0.00003271, 0.00001652, -0.00001786, 0.00001677, -0.00001057, 0.00001259, -0.00002085, 0.00002276, -0.00004001, 0.00005210, -0.00005388, 0.00005799, -0.00006137, 0.00005955, -0.00006593, 0.00006541, -0.00007425, 0.00007382, -0.00009060, 0.00010089, -0.00012049, 0.00012070, -0.00015746, 0.00016717, -0.00024919, 0.00026743, -0.00060135, 0.00039144, 0.00086393, -0.00077535, 0.00031286, -0.00027842, 0.00018036, -0.00016664, 0.00012680, -0.00012536, 0.00010242, -0.00009105, 0.00007345, -0.00007410, 0.00006572, -0.00006533, 0.00005727, -0.00005815, 0.00005333, -0.00004629, 0.00003787, -0.00002361, 0.00000966, -0.00001682, 0.00001692, -0.00002025, 0.00003047, -0.00002874, 0.00002934, -0.00004299, 0.00004737, -0.00004812, 0.00004856, -0.00003770, 0.00002616, -0.00002170, 0.00001555, -0.00001737, 0.00001401, -0.00000586, 0.00000831, -0.00001493, 0.00001395, -0.00001481, 0.00001716, -0.00001964, 0.00001708, -0.00000390, 0.00000466, -0.00001836, 0.00001490, -0.00000933, 0.00000470, -0.00000323, 0.00000374, -0.00000304, 0.00000509, -0.00000195, -0.00000231, -0.00001006, 0.00000544, 0.00002160, -0.00001702, -0.00001172, 0.00000936, 0.00000276, },
		{ -0.00000370, 0.00001583, -0.00001642, -0.00001348, 0.00002079, 0.00000479, -0.00001107, 0.00000094, -0.00000161, 0.00000376, -0.00000310, 0.00000354, -0.00000145, 0.00000354, -0.00000904, 0.00001545, -0.00001828, 0.00000546, -0.00000653, 0.00001905, -0.00001468, 0.00001023, -0.00001069, 0.00001081, -0.00001271, 0.00000543, 0.00000042, 0.00000681, -0.00000956, 0.00000675, -0.00001090, 0.00001347, -0.00002382, 0.00003912, -0.00004509, 0.00004607, -0.00004383, 0.00002744, -0.00002608, 0.00002335, -0.00001493, 0.00001215, -0.00001792, 0.00001670, -0.00003071, 0.00004450, -0.00004801, 0.00005374, -0.00005997, 0.00005952, -0.00006660, 0.00006648, -0.00007601, 0.00007466, -0.00008996, 0.00010141, -0.00012535, 0.00012779, -0.00016823, 0.00018093, -0.00027690, 0.00030972, -0.00078011, 0.00089693, 0.00034809, -0.00057733, 0.00025545, -0.00023995, 0.00015806, -0.00014819, 0.00011241, -0.00011358, 0.00009576, -0.00008688, 0.00006865, -0.00006824, 0.00006057, -0.00006095, 0.00005358, -0.00005551, 0.00005395, -0.00004972, 0.00004297, -0.00002901, 0.00000966, -0.00001195, 0.00000984, -0.00000951, 0.00001917, -0.00001686, 0.00001474, -0.00002685, 0.00003457, -0.00003930, 0.00004647, -0.00004248, 0.00003306, -0.00002861, 0.00002125, -0.00002224, 0.00001959, -0.00000976, 0.00000798, -0.00001319, 0.00001246, -0.00001305, 0.00001631, -0.00002059, 0.00002000, -0.00000631, 0.00000297, -0.00001660, 0.00001644, -0.00001272, 0.00000821, -0.00000600, 0.00000558, -0.00000425, 0.00000690, -0.00000441, -0.00000264, -0.00001063, 0.00001255, 0.00001610, -0.00002229, -0.00000416, 0.00000739, 0.00000295, },
		{ -0.00000203, 0.00001613, -0.00002468, -0.00000280, 0.00001986, -0.00000021, -0.00000987, 0.00000229, -0.00000109, 0.00000251, -0.00000195, 0.00000213, 0.00000140, -0.00000152, -0.00000259, 0.00000981, -0.00001620, 0.00000463, -0.00000268, 0.00001561, -0.00001290, 0.00000807, -0.00000823, 0.00000829, -0.00001162, 0.00000519, 0.00000439, -0.00000073, -0.00000111, -0.00000376, 0.00000317, -0.00000483, -0.00000102, 0.00001666, -0.00002771, 0.00003423, -0.00003883, 0.00002496, -0.00002271, 0.00001929, -0.00000943, 0.00000179, -0.00000418, -0.00000110, -0.00000762, 0.00002051, -0.00002351, 0.00002843, -0.00003555, 0.00003512, -0.00004143, 0.00003987, -0.00004816, 0.00004359, -0.00005346, 0.00006036, -0.00008247, 0.00008009, -0.00011478, 0.00011596, -0.00020379, 0.00020768, -0.00075907, 0.00124574, -0.00016908, -0.00027865, 0.00010249, -0.00012749, 0.00007504, -0.00007885, 0.00005354, -0.00006179, 0.00005300, -0.00005109, 0.00003612, -0.00003703, 0.00003166, -0.00003429, 0.00002874, -0.00003232, 0.00003489, -0.00003547, 0.00003305, -0.00002278, 0.00000078, 0.00000090, -0.00000471, 0.00000865, 0.00000018, 0.00000262, -0.00000799, -0.00000025, 0.00000831, -0.00001431, 0.00002638, -0.00003004, 0.00002527, -0.00002331, 0.00001668, -0.00001809, 0.00001770, -0.00000830, 0.00000333, -0.00000675, 0.00000581, -0.00000563, 0.00000944, -0.00001586, 0.00001823, -0.00000532, -0.00000190, -0.00001052, 0.00001304, -0.00001168, 0.00000826, -0.00000625, 0.00000550, -0.00000365, 0.00000709, -0.00000621, -0.00000347, -0.00000948, 0.00001902, 0.00000725, -0.00002280, 0.00000103, 0.00000622, 0.00000182, },
		{ 0.00000271, 0.00001163, -0.00002839, 0.00000721, 0.00001671, -0.00000447, -0.00000732, 0.00000260, -0.00000029, 0.00000083, -0.00000027, 0.00000056, 0.00000408, -0.00000671, 0.00000513, 0.00000137, -0.00001065, 0.00000109, 0.00000350, 0.00000954, -0.00000872, 0.00000363, -0.00000333, 0.00000315, -0.00000819, 0.00000426, 0.00000731, -0.00000691, 0.00000591, -0.00001232, 0.00001502, -0.00002166, 0.00002308, -0.00001171, -0.00000045, 0.00001019, -0.00002144, 0.00001186, -0.00001055, 0.00000784, 0.00000199, -0.00001361, 0.00001464, -0.00002402, 0.00002164, -0.00001143, 0.00001054, -0.00000846, 0.00000237, -0.00000406, -0.00000034, -0.00000404, -0.00000170, -0.00000727, 0.00000524, -0.00000729, -0.00000784, -0.00000529, -0.00001574, -0.00000756, -0.00005078, -0.00002569, -0.00051271, 0.00134747, -0.00056214, 0.00001493, -0.00007477, 0.00000887, -0.00003009, 0.00000991, -0.00002313, 0.00000711, -0.00000631, -0.00000054, -0.00000900, 0.00000548, -0.00000795, 0.00000254, -0.00000599, 0.00000116, 0.00000506, -0.00001097, 0.00001443, -0.00001075, -0.00001084, 0.00001506, -0.00001963, 0.00002675, -0.00001869, 0.00002135, -0.00002950, 0.00002644, -0.00002071, 0.00001678, -0.00000350, -0.00000614, 0.00000648, -0.00000828, 0.00000362, -0.00000621, 0.00000910, -0.00000222, -0.00000432, 0.00000245, -0.00000379, 0.00000528, -0.00000173, -0.00000663, 0.00001264, -0.00000196, -0.00000848, -0.00000178, 0.00000598, -0.00000688, 0.00000501, -0.00000389, 0.00000343, -0.00000115, 0.00000530, -0.00000693, -0.00000494, -0.00000609, 0.00002317, -0.00000302, -0.00001910, 0.00000318, 0.00000664, -0.00000117, },
		{ 0.00000966, 0.00000256, -0.00002626, 0.00001434, 0.00001270, -0.00000799, -0.00000368, 0.00000181, 0.00000077, -0.00000097, 0.00000149, -0.00000056, 0.00000549, -0.00001027, 0.00001179, -0.00000746, -0.00000366, -0.00000351, 0.00001044, 0.00000244, -0.00000381, -0.00000113, 0.00000194, -0.00000272, -0.00000387, 0.00000382, 0.00000759, -0.00000926, 0.00000825, -0.00001472, 0.00001898, -0.00002924, 0.00003848, -0.00003466, 0.00002547, -0.00001606, -0.00000000, -0.00000481, 0.00000384, -0.00000443, 0.00001228, -0.00002618, 0.00002976, -0.00004227, 0.00004610, -0.00003919, 0.00004083, -0.00004255, 0.00003866, -0.00004223, 0.00004002, -0.00004757, 0.00004443, -0.00005715, 0.00006272, -0.00007505, 0.00006888, -0.00009489, 0.00009023, -0.00014374, 0.00012588, -0.00032592, -0.00008386, 0.00118050, -0.00075192, 0.00021645, -0.00020808, 0.00011662, -0.00011479, 0.00008225, -0.00008640, 0.00006556, -0.00005796, 0.00004358, -0.00004769, 0.00004174, -0.00004192, 0.00003427, -0.00003617, 0.00003139, -0.00002339, 0.00001328, -0.00000407, -0.00000054, -0.00001793, 0.00002309, -0.00002733, 0.00003699, -0.00002951, 0.00003109, -0.00004058, 0.00004258, -0.00004057, 0.00004139, -0.00003110, 0.00001892, -0.00001498, 0.00000975, -0.00001230, 0.00000872, -0.00000253, 0.00000554, -0.00001200, 0.00001106, -0.00001268, 0.00001594, -0.00001381, 0.00000437, 0.00000539, 0.00000181, -0.00001445, 0.00000698, -0.00000226, -0.00000028, -0.00000016, 0.00000012, 0.00000011, 0.00000267, 0.00000176, -0.00000652, -0.00000692, -0.00000038, 0.00002352, -0.00001222, -0.00001308, 0.00000309, 0.00000810, -0.00000523, },
		{ 0.00001651, -0.00000907, -0.00001885, 0.00001758, 0.00000873, -0.00001057, 0.00000044, 0.00000021, 0.00000186, -0.00000248, 0.00000290, -0.00000081, 0.00000497, -0.00001091, 0.00001536, -0.00001423, 0.00000252, -0.00000726, 0.00001636, -0.00000409, 0.00000030, -0.00000456, 0.00000573, -0.00000756, 0.00000017, 0.00000426, 0.00000503, -0.00000734, 0.00000521, -0.00000989, 0.00001307, -0.00002385, 0.00003882, -0.00004342, 0.00004002, -0.00003446, 0.00001646, -0.00001728, 0.00001353, -0.00001096, 0.00001486, -0.00002879, 0.00003332, -0.00004721, 0.00005603, -0.00005193, 0.00005530, -0.00006047, 0.00005886, -0.00006402, 0.00006324, -0.00007311, 0.00007134, -0.00008534, 0.00009544, -0.00011579, 0.00011644, -0.00015236, 0.00015961, -0.00023834, 0.00025391, -0.00059554, 0.00042653, 0.00079699, -0.00072482, 0.00028279, -0.00025578, 0.00016083, -0.00014940, 0.00011242, -0.00011321, 0.00009262, -0.00008316, 0.00006470, -0.00006515, 0.00005811, -0.00005753, 0.00004891, -0.00005032, 0.00004731, -0.00004019, 0.00002841, -0.00001528, 0.00000229, -0.00001573, 0.00002045, -0.00002329, 0.00003484, -0.00002785, 0.00002749, -0.00003629, 0.00004181, -0.00004315, 0.00004962, -0.00004561, 0.00003476, -0.00002989, 0.00002279, -0.00002409, 0.00002062, -0.00001228, 0.00001114, -0.00001632, 0.00001561, -0.00001714, 0.00002236, -0.00002291, 0.00001386, -0.00000099, 0.00000387, -0.00001754, 0.00001304, -0.00000888, 0.00000561, -0.00000519, 0.00000421, -0.00000316, 0.00000662, -0.00000273, -0.00000534, -0.00000880, 0.00000700, 0.00001943, -0.00001808, -0.00000717, 0.00000250, 0.00000906, -0.00000847, },
		{ 0.00002041, -0.00002000, -0.00000838, 0.00001734, 0.00000507, -0.00001176, 0.00000415, -0.00000158, 0.00000263, -0.00000333, 0.00000367, -0.00000013, 0.00000249, -0.00000824, 0.00001474, -0.00001722, 0.00000606, -0.00000855, 0.00001982, -0.00000886, 0.00000272, -0.00000577, 0.00000701, -0.00001031, 0.00000341, 0.00000509, 0.00000093, -0.00000281, -0.00000114, -0.00000037, -0.00000002, -0.00000740, 0.00002405, -0.00003527, 0.00003829, -0.00003903, 0.00002212, -0.00002059, 0.00001447, -0.00000847, 0.00000692, -0.00001858, 0.00002215, -0.00003523, 0.00004720, -0.00004473, 0.00004819, -0.00005538, 0.00005520, -0.00006091, 0.00006002, -0.00007053, 0.00006816, -0.00007998, 0.00008966, -0.00011301, 0.00011496, -0.00015356, 0.00016198, -0.00025073, 0.00027236, -0.00073585, 0.00088930, 0.00030446, -0.00052961, 0.00022199, -0.00021364, 0.00013496, -0.00012706, 0.00009384, -0.00009700, 0.00008157, -0.00007542, 0.00005712, -0.00005646, 0.00005011, -0.00005051, 0.00004242, -0.00004431, 0.00004446, -0.00004087, 0.00003059, -0.00001665, -0.00000330, -0.00000429, 0.00000763, -0.00000817, 0.00002104, -0.00001472, 0.00001203, -0.00001815, 0.00002484, -0.00002756, 0.00003837, -0.00004187, 0.00003525, -0.00003216, 0.00002517, -0.00002653, 0.00002469, -0.00001614, 0.00001167, -0.00001508, 0.00001405, -0.00001488, 0.00002180, -0.00002605, 0.00001918, -0.00000452, 0.00000284, -0.00001636, 0.00001463, -0.00001178, 0.00000865, -0.00000814, 0.00000675, -0.00000495, 0.00000929, -0.00000704, -0.00000412, -0.00000966, 0.00001464, 0.00001150, -0.00001945, -0.00000332, 0.00000310, 0.00000791, -0.00000879, },
		{ 0.00001914, -0.00002699, 0.00000217, 0.00001498, 0.00000145, -0.00001103, 0.00000649, -0.00000284, 0.00000268, -0.00000322, 0.00000371, 0.00000117, -0.00000130, -0.00000290, 0.00001006, -0.00001589, 0.00000610, -0.00000655, 0.00002000, -0.00001133, 0.00000332, -0.00000481, 0.00000569, -0.00001069, 0.00000583, 0.00000536, -0.00000272, 0.00000153, -0.00000710, 0.00000892, -0.00001415, 0.00001326, 0.00000028, -0.00001439, 0.00002222, -0.00002994, 0.00001647, -0.00001463, 0.00000733, 0.00000141, -0.00000903, 0.00000145, -0.00000056, -0.00000960, 0.00002279, -0.00002071, 0.00002247, -0.00002979, 0.00002971, -0.00003472, 0.00003205, -0.00004142, 0.00003664, -0.00004316, 0.00004736, -0.00006790, 0.00006448, -0.00009703, 0.00009331, -0.00017161, 0.00015647, -0.00068108, 0.00118660, -0.00016917, -0.00025488, 0.00008236, -0.00011069, 0.00005949, -0.00006364, 0.00003949, -0.00004850, 0.00004109, -0.00004185, 0.00002720, -0.00002765, 0.00002336, -0.00002607, 0.00001958, -0.00002237, 0.00002612, -0.00002785, 0.00002189, -0.00001071, -0.00001358, 0.00001161, -0.00001006, 0.00001246, 0.00000124, 0.00000395, -0.00000860, 0.00000643, -0.00000087, -0.00000048, 0.00001249, -0.00002229, 0.00002093, -0.00002131, 0.00001595, -0.00001837, 0.00001942, -0.00001278, 0.00000651, -0.00000832, 0.00000672, -0.00000612, 0.00001392, -0.00002228, 0.00001923, -0.00000456, -0.00000134, -0.00001104, 0.00001152, -0.00001028, 0.00000785, -0.00000793, 0.00000676, -0.00000431, 0.00000956, -0.00001004, -0.00000358, -0.00000850, 0.00002062, 0.00000156, -0.00001673, -0.00000221, 0.00000564, 0.00000389, -0.00000497, },
		{ 0.00001212, -0.00002798, 0.00001013, 0.00001205, -0.00000241, -0.00000813, 0.00000680, -0.00000303, 0.00000179, -0.00000210, 0.00000312, 0.00000256, -0.00000525, 0.00000364, 0.00000265, -0.00001099, 0.00000299, -0.00000142, 0.00001692, -0.00001153, 0.00000255, -0.00000243, 0.00000246, -0.00000900, 0.00000756, 0.00000434, -0.00000426, 0.00000324, -0.00000925, 0.00001314, -0.00002251, 0.00002942, -0.00002284, 0.00001021, -0.00000078, -0.00001280, 0.00000397, -0.00000366, -0.00000314, 0.00001303, -0.00002614, 0.00002352, -0.00002637, 0.00002071, -0.00000785, 0.00001029, -0.00001149, 0.00000567, -0.00000677, 0.00000332, -0.00000890, 0.00000183, -0.00000972, 0.00000992, -0.00001448, 0.00000103, -0.00001482, -0.00000515, -0.00002138, -0.00002890, -0.00006952, -0.00042247, 0.00124746, -0.00051457, 0.00000371, -0.00007060, 0.00000663, -0.00002958, 0.00001139, -0.00002538, 0.00001035, -0.00000990, 0.00000124, -0.00001052, 0.00000785, -0.00000966, 0.00000420, -0.00000872, 0.00000542, 0.00000113, -0.00000845, 0.00000848, -0.00000326, -0.00002220, 0.00002481, -0.00002488, 0.00003065, -0.00001661, 0.00002007, -0.00002553, 0.00002741, -0.00002438, 0.00002674, -0.00001741, 0.00000442, -0.00000152, -0.00000248, -0.00000075, -0.00000288, 0.00000719, -0.00000402, -0.00000235, 0.00000151, -0.00000362, 0.00000634, 0.00000094, -0.00001294, 0.00001481, -0.00000197, -0.00000730, -0.00000322, 0.00000511, -0.00000534, 0.00000374, -0.00000474, 0.00000426, -0.00000117, 0.00000699, -0.00001108, -0.00000416, -0.00000471, 0.00002310, -0.00000783, -0.00001166, -0.00000306, 0.00000947, -0.00000226, 0.00000255, },
		{ 0.00000086, -0.00002287, 0.00001411, 0.00000960, -0.00000643, -0.00000341, 0.00000504, -0.00000202, 0.00000009, -0.00000017, 0.00000217, 0.00000339, -0.00000811, 0.00000948, -0.00000540, -0.00000427, -0.00000195, 0.00000576, 0.00001124, -0.00000995, 0.00000115, 0.00000030, -0.00000161, -0.00000580, 0.00000857, 0.00000207, -0.00000327, 0.00000149, -0.00000626, 0.00000984, -0.00002080, 0.00003423, -0.00003631, 0.00002862, -0.00002128, 0.00000426, -0.00000852, 0.00000604, -0.00001064, 0.00001945, -0.00003633, 0.00003838, -0.00004513, 0.00004472, -0.00003307, 0.00003582, -0.00004023, 0.00003674, -0.00003933, 0.00003746, -0.00004610, 0.00004137, -0.00005152, 0.00005741, -0.00007103, 0.00006564, -0.00009090, 0.00008480, -0.00013734, 0.00012117, -0.00033754, -0.00001348, 0.00106746, -0.00067093, 0.00017404, -0.00017959, 0.00009429, -0.00009670, 0.00006810, -0.00007464, 0.00005634, -0.00005092, 0.00003597, -0.00004014, 0.00003546, -0.00003543, 0.00002774, -0.00003062, 0.00002785, -0.00002039, 0.00000879, -0.00000275, -0.00000006, -0.00002349, 0.00002913, -0.00003024, 0.00003954, -0.00002577, 0.00002703, -0.00003168, 0.00003652, -0.00003593, 0.00004291, -0.00003951, 0.00002712, -0.00002240, 0.00001600, -0.00001769, 0.00001368, -0.00000688, 0.00000603, -0.00001120, 0.00001056, -0.00001280, 0.00001814, -0.00001312, -0.00000114, 0.00000814, 0.00000133, -0.00001287, 0.00000460, -0.00000211, 0.00000089, -0.00000193, 0.00000011, 0.00000034, 0.00000353, 0.00000208, -0.00001017, -0.00000564, 0.00000151, 0.00002103, -0.00001426, -0.00000666, -0.00000407, 0.00001285, -0.00000847, 0.00001162, },
		{ -0.00001150, -0.00001351, 0.00001431, 0.00000776, -0.00000994, 0.00000211, 0.00000186, -0.00000024, -0.00000193, 0.00000207, 0.00000124, 0.00000318, -0.00000887, 0.00001290, -0.00001190, 0.00000210, -0.00000679, 0.00001334, 0.00000413, -0.00000728, -0.00000020, 0.00000242, -0.00000545, -0.00000170, 0.00000858, -0.00000057, -0.00000080, -0.00000249, 0.00000044, 0.00000030, -0.00000920, 0.00002565, -0.00003541, 0.00003421, -0.00003188, 0.00001435, -0.00001515, 0.00000949, -0.00001057, 0.00001589, -0.00003398, 0.00003936, -0.00004915, 0.00005393, -0.00004375, 0.00004604, -0.00005285, 0.00005154, -0.00005530, 0.00005415, -0.00006493, 0.00006157, -0.00007180, 0.00008014, -0.00009997, 0.00009994, -0.00013324, 0.00013590, -0.00020801, 0.00021489, -0.00055723, 0.00044586, 0.00070657, -0.00063769, 0.00022611, -0.00021401, 0.00012664, -0.00012024, 0.00008790, -0.00009166, 0.00007419, -0.00006814, 0.00005024, -0.00005101, 0.00004550, -0.00004501, 0.00003625, -0.00003819, 0.00003718, -0.00003125, 0.00001789, -0.00000748, -0.00000392, -0.00001527, 0.00002226, -0.00002349, 0.00003627, -0.00002354, 0.00002260, -0.00002482, 0.00003075, -0.00003112, 0.00004167, -0.00004563, 0.00003702, -0.00003324, 0.00002633, -0.00002771, 0.00002473, -0.00001718, 0.00001289, -0.00001624, 0.00001518, -0.00001695, 0.00002497, -0.00002393, 0.00000951, 0.00000190, 0.00000325, -0.00001589, 0.00000993, -0.00000755, 0.00000587, -0.00000690, 0.00000469, -0.00000331, 0.00000819, -0.00000394, -0.00000801, -0.00000721, 0.00000901, 0.00001463, -0.00001632, -0.00000378, -0.00000337, 0.00001375, -0.00001217, 0.00001909, },
		{ -0.00002133, -0.00000297, 0.00001214, 0.00000596, -0.00001193, 0.00000699, -0.00000161, 0.00000154, -0.00000355, 0.00000401, 0.00000068, 0.00000175, -0.00000712, 0.00001289, -0.00001515, 0.00000618, -0.00000964, 0.00001949, -0.00000306, -0.00000420, -0.00000107, 0.00000331, -0.00000819, 0.00000270, 0.00000726, -0.00000236, 0.00000130, -0.00000612, 0.00000729, -0.00001105, 0.00000761, 0.00000724, -0.00002149, 0.00002598, -0.00003007, 0.00001460, -0.00001369, 0.00000532, -0.00000236, 0.00000219, -0.00001864, 0.00002526, -0.00003651, 0.00004573, -0.00003694, 0.00003763, -0.00004537, 0.00004533, -0.00004940, 0.00004754, -0.00005891, 0.00005545, -0.00006312, 0.00006954, -0.00009078, 0.00009083, -0.00012548, 0.00012700, -0.00020430, 0.00020590, -0.00064646, 0.00083991, 0.00026702, -0.00046610, 0.00017375, -0.00017580, 0.00010283, -0.00009863, 0.00006912, -0.00007448, 0.00006150, -0.00005894, 0.00004177, -0.00004141, 0.00003652, -0.00003710, 0.00002858, -0.00003025, 0.00003181, -0.00002965, 0.00001751, -0.00000558, -0.00001344, 0.00000009, 0.00000676, -0.00000698, 0.00002293, -0.00001213, 0.00000963, -0.00000845, 0.00001355, -0.00001265, 0.00002397, -0.00003443, 0.00003108, -0.00003026, 0.00002454, -0.00002676, 0.00002620, -0.00001997, 0.00001367, -0.00001532, 0.00001352, -0.00001404, 0.00002419, -0.00002835, 0.00001613, -0.00000187, 0.00000245, -0.00001515, 0.00001129, -0.00000944, 0.00000774, -0.00000927, 0.00000724, -0.00000504, 0.00001105, -0.00000947, -0.00000567, -0.00000769, 0.00001591, 0.00000550, -0.00001415, -0.00000385, 0.00000003, 0.00001090, -0.00001137, 0.00002200, },
		{ -0.00002579, 0.00000573, 0.00000946, 0.00000340, -0.00001144, 0.00000987, -0.00000417, 0.00000250, -0.00000410, 0.00000511, 0.00000064, -0.00000064, -0.00000322, 0.00000939, -0.00001448, 0.00000688, -0.00000918, 0.00002271, -0.00000905, -0.00000131, -0.00000130, 0.00000275, -0.00000927, 0.00000674, 0.00000461, -0.00000239, 0.00000141, -0.00000687, 0.00001042, -0.00001862, 0.00002253, -0.00001348, -0.00000102, 0.00000858, -0.00001869, 0.00000693, -0.00000619, -0.00000368, 0.00001020, -0.00001691, 0.00000461, 0.00000098, -0.00001172, 0.00002427, -0.00001679, 0.00001494, -0.00002197, 0.00002207, -0.00002552, 0.00002156, -0.00003196, 0.00002724, -0.00003037, 0.00003099, -0.00004880, 0.00004334, -0.00007214, 0.00006158, -0.00012714, 0.00008436, -0.00055986, 0.00107279, -0.00014004, -0.00023570, 0.00006148, -0.00009247, 0.00004284, -0.00004774, 0.00002515, -0.00003457, 0.00002784, -0.00003129, 0.00001767, -0.00001803, 0.00001481, -0.00001757, 0.00001031, -0.00001197, 0.00001610, -0.00001902, 0.00001068, -0.00000057, -0.00002388, 0.00001681, -0.00001117, 0.00001316, 0.00000532, 0.00000273, -0.00000547, 0.00001005, -0.00000695, 0.00001108, -0.00000257, -0.00001158, 0.00001295, -0.00001569, 0.00001197, -0.00001549, 0.00001801, -0.00001484, 0.00000825, -0.00000885, 0.00000635, -0.00000478, 0.00001582, -0.00002555, 0.00001762, -0.00000250, -0.00000112, -0.00001085, 0.00000870, -0.00000753, 0.00000595, -0.00000825, 0.00000692, -0.00000394, 0.00001089, -0.00001314, -0.00000413, -0.00000603, 0.00002014, -0.00000385, -0.00000936, -0.00000616, 0.00000564, 0.00000457, -0.00000569, 0.00001881, },
		{ -0.00002390, 0.00001062, 0.00000767, -0.00000032, -0.00000815, 0.00000996, -0.00000503, 0.00000214, -0.00000323, 0.00000510, 0.00000104, -0.00000338, 0.00000181, 0.00000341, -0.00001038, 0.00000421, -0.00000504, 0.00002213, -0.00001289, 0.00000097, -0.00000099, 0.00000094, -0.00000841, 0.00000969, 0.00000124, -0.00000072, -0.00000089, -0.00000372, 0.00000762, -0.00001833, 0.00002912, -0.00002828, 0.00001739, -0.00001022, -0.00000408, -0.00000346, 0.00000245, -0.00001215, 0.00002073, -0.00003378, 0.00002721, -0.00002456, 0.00001626, -0.00000158, 0.00000755, -0.00001226, 0.00000711, -0.00000774, 0.00000538, -0.00001226, 0.00000380, -0.00001004, 0.00001179, -0.00001889, 0.00000745, -0.00002201, 0.00000358, -0.00003364, -0.00000813, -0.00011707, -0.00030487, 0.00109602, -0.00042745, -0.00002541, -0.00005666, -0.00000222, -0.00002378, 0.00000816, -0.00002306, 0.00000970, -0.00001080, 0.00000103, -0.00000988, 0.00000793, -0.00000924, 0.00000394, -0.00000945, 0.00000806, -0.00000233, -0.00000565, 0.00000263, 0.00000262, -0.00002964, 0.00002832, -0.00002443, 0.00002977, -0.00000978, 0.00001458, -0.00001591, 0.00002282, -0.00002163, 0.00002974, -0.00002713, 0.00001303, -0.00000904, 0.00000358, -0.00000558, 0.00000127, 0.00000392, -0.00000462, -0.00000080, 0.00000041, -0.00000326, 0.00000763, 0.00000246, -0.00001702, 0.00001472, -0.00000077, -0.00000623, -0.00000451, 0.00000361, -0.00000303, 0.00000140, -0.00000441, 0.00000406, -0.00000020, 0.00000742, -0.00001424, -0.00000384, -0.00000184, 0.00002021, -0.00001083, -0.00000440, -0.00000885, 0.00001166, -0.00000338, 0.00000333, 0.00001011, },
		{ -0.00001681, 0.00001140, 0.00000717, -0.00000482, -0.00000262, 0.00000736, -0.00000408, 0.00000055, -0.00000109, 0.00000410, 0.00000158, -0.00000562, 0.00000655, -0.00000337, -0.00000437, -0.00000068, 0.00000206, 0.00001780, -0.00001412, 0.00000239, -0.00000042, -0.00000166, -0.00000569, 0.00001085, -0.00000177, 0.00000159, -0.00000450, 0.00000221, -0.00000043, -0.00000955, 0.00002447, -0.00003173, 0.00002666, -0.00002301, 0.00000706, -0.00001092, 0.00000713, -0.00001493, 0.00002334, -0.00004119, 0.00004067, -0.00004207, 0.00003778, -0.00002208, 0.00002602, -0.00003308, 0.00003021, -0.00003184, 0.00003031, -0.00004002, 0.00003362, -0.00004054, 0.00004559, -0.00005968, 0.00005456, -0.00007822, 0.00006988, -0.00011990, 0.00010234, -0.00033320, 0.00006008, 0.00091971, -0.00055357, 0.00011034, -0.00013752, 0.00006225, -0.00007089, 0.00004728, -0.00005650, 0.00004130, -0.00003931, 0.00002491, -0.00002936, 0.00002600, -0.00002602, 0.00001861, -0.00002238, 0.00002185, -0.00001621, 0.00000447, -0.00000212, 0.00000040, -0.00002701, 0.00003025, -0.00002803, 0.00003763, -0.00001753, 0.00001922, -0.00001759, 0.00002506, -0.00002448, 0.00003552, -0.00004033, 0.00002958, -0.00002569, 0.00001921, -0.00002058, 0.00001672, -0.00001029, 0.00000603, -0.00000953, 0.00000861, -0.00001122, 0.00001878, -0.00001172, -0.00000591, 0.00000954, 0.00000161, -0.00001096, 0.00000168, -0.00000174, 0.00000191, -0.00000394, 0.00000062, 0.00000007, 0.00000491, 0.00000146, -0.00001298, -0.00000449, 0.00000425, 0.00001585, -0.00001369, -0.00000155, -0.00000973, 0.00001569, -0.00001017, 0.00001272, -0.00000158, },
		{ -0.00000724, 0.00000929, 0.00000732, -0.00000896, 0.00000372, 0.00000307, -0.00000195, -0.00000166, 0.00000171, 0.00000257, 0.00000181, -0.00000659, 0.00000961, -0.00000905, 0.00000158, -0.00000594, 0.00001050, 0.00001058, -0.00001283, 0.00000290, 0.00000010, -0.00000438, -0.00000155, 0.00000982, -0.00000328, 0.00000296, -0.00000729, 0.00000805, -0.00001028, 0.00000446, 0.00001041, -0.00002329, 0.00002407, -0.00002575, 0.00001063, -0.00001203, 0.00000517, -0.00000966, 0.00001540, -0.00003550, 0.00004006, -0.00004560, 0.00004631, -0.00003063, 0.00003184, -0.00004009, 0.00003912, -0.00004141, 0.00003979, -0.00005140, 0.00004647, -0.00005246, 0.00005781, -0.00007588, 0.00007414, -0.00010336, 0.00009987, -0.00016267, 0.00015487, -0.00048835, 0.00044436, 0.00060460, -0.00052691, 0.00015319, -0.00016127, 0.00008462, -0.00008507, 0.00005834, -0.00006517, 0.00005097, -0.00004921, 0.00003299, -0.00003451, 0.00003062, -0.00003049, 0.00002191, -0.00002427, 0.00002517, -0.00002152, 0.00000823, -0.00000183, -0.00000768, -0.00001594, 0.00002213, -0.00002093, 0.00003533, -0.00001671, 0.00001623, -0.00001077, 0.00001702, -0.00001506, 0.00002641, -0.00003794, 0.00003237, -0.00003081, 0.00002507, -0.00002702, 0.00002502, -0.00001931, 0.00001278, -0.00001435, 0.00001251, -0.00001409, 0.00002468, -0.00002257, 0.00000434, 0.00000449, 0.00000285, -0.00001355, 0.00000576, -0.00000529, 0.00000516, -0.00000793, 0.00000484, -0.00000321, 0.00000951, -0.00000542, -0.00001027, -0.00000514, 0.00001074, 0.00000817, -0.00001216, -0.00000196, -0.00000736, 0.00001585, -0.00001317, 0.00001914, -0.00001264, },
		{ 0.00000161, 0.00000631, 0.00000694, -0.00001131, 0.00000913, -0.00000144, 0.00000031, -0.00000356, 0.00000431, 0.00000109, 0.00000139, -0.00000584, 0.00001008, -0.00001210, 0.00000560, -0.00000954, 0.00001823, 0.00000199, -0.00000961, 0.00000263, 0.00000025, -0.00000648, 0.00000319, 0.00000683, -0.00000269, 0.00000217, -0.00000723, 0.00001055, -0.00001729, 0.00001794, -0.00000737, -0.00000729, 0.00001195, -0.00001914, 0.00000659, -0.00000704, -0.00000247, 0.00000197, -0.00000111, -0.00001820, 0.00002586, -0.00003464, 0.00004071, -0.00002607, 0.00002403, -0.00003201, 0.00003209, -0.00003441, 0.00003147, -0.00004362, 0.00003930, -0.00004280, 0.00004511, -0.00006312, 0.00006029, -0.00008955, 0.00008271, -0.00014609, 0.00012087, -0.00052257, 0.00075378, 0.00023934, -0.00039464, 0.00011841, -0.00013305, 0.00006756, -0.00006821, 0.00004306, -0.00005045, 0.00003953, -0.00004082, 0.00002563, -0.00002597, 0.00002259, -0.00002353, 0.00001489, -0.00001615, 0.00001868, -0.00001856, 0.00000614, 0.00000198, -0.00001882, 0.00000002, 0.00000747, -0.00000612, 0.00002529, -0.00000969, 0.00000871, 0.00000040, 0.00000356, 0.00000154, 0.00000668, -0.00002232, 0.00002180, -0.00002355, 0.00001958, -0.00002280, 0.00002358, -0.00002044, 0.00001334, -0.00001363, 0.00001083, -0.00001054, 0.00002329, -0.00002736, 0.00001117, 0.00000130, 0.00000193, -0.00001317, 0.00000676, -0.00000593, 0.00000544, -0.00000910, 0.00000675, -0.00000424, 0.00001183, -0.00001143, -0.00000732, -0.00000465, 0.00001564, -0.00000057, -0.00000756, -0.00000527, -0.00000174, 0.00001162, -0.00001107, 0.00002021, -0.00001979, },
		{ 0.00000727, 0.00000428, 0.00000503, -0.00001080, 0.00001206, -0.00000476, 0.00000171, -0.00000429, 0.00000591, 0.00000016, 0.00000021, -0.00000345, 0.00000782, -0.00001188, 0.00000654, -0.00000993, 0.00002329, -0.00000620, -0.00000538, 0.00000185, -0.00000019, -0.00000725, 0.00000741, 0.00000277, -0.00000035, -0.00000089, -0.00000356, 0.00000777, -0.00001772, 0.00002514, -0.00002176, 0.00000923, -0.00000374, -0.00000762, -0.00000159, 0.00000063, -0.00001178, 0.00001504, -0.00002038, 0.00000470, 0.00000356, -0.00001392, 0.00002509, -0.00001256, 0.00000725, -0.00001373, 0.00001396, -0.00001580, 0.00001048, -0.00002186, 0.00001759, -0.00001786, 0.00001474, -0.00002926, 0.00002128, -0.00004554, 0.00002763, -0.00007967, 0.00000454, -0.00041215, 0.00091876, -0.00008812, -0.00022101, 0.00004182, -0.00007517, 0.00002751, -0.00003364, 0.00001297, -0.00002245, 0.00001562, -0.00002143, 0.00000928, -0.00000981, 0.00000762, -0.00001048, 0.00000274, -0.00000308, 0.00000680, -0.00001092, 0.00000140, 0.00000575, -0.00002837, 0.00001535, -0.00000798, 0.00001101, 0.00001218, -0.00000070, 0.00000132, 0.00001020, -0.00000843, 0.00001726, -0.00001521, -0.00000088, 0.00000362, -0.00000821, 0.00000602, -0.00001033, 0.00001382, -0.00001432, 0.00000824, -0.00000818, 0.00000476, -0.00000177, 0.00001512, -0.00002557, 0.00001368, 0.00000046, -0.00000108, -0.00001015, 0.00000502, -0.00000390, 0.00000279, -0.00000719, 0.00000589, -0.00000243, 0.00001082, -0.00001523, -0.00000507, -0.00000228, 0.00001726, -0.00000777, -0.00000222, -0.00000971, 0.00000562, 0.00000423, -0.00000445, 0.00001545, -0.00002117, },
		{ 0.00000881, 0.00000409, 0.00000137, -0.00000719, 0.00001181, -0.00000606, 0.00000169, -0.00000344, 0.00000612, -0.00000000, -0.00000150, -0.00000005, 0.00000352, -0.00000872, 0.00000433, -0.00000655, 0.00002438, -0.00001241, -0.00000115, 0.00000086, -0.00000126, -0.00000628, 0.00001000, -0.00000099, 0.00000248, -0.00000502, 0.00000258, 0.00000020, -0.00001063, 0.00002289, -0.00002735, 0.00001957, -0.00001631, 0.00000292, -0.00000899, 0.00000638, -0.00001775, 0.00002357, -0.00003525, 0.00002524, -0.00001880, 0.00000894, 0.00000654, 0.00000279, -0.00001074, 0.00000644, -0.00000660, 0.00000530, -0.00001346, 0.00000369, -0.00000787, 0.00001026, -0.00001934, 0.00000980, -0.00002488, 0.00000789, -0.00004069, 0.00000563, -0.00015764, -0.00017733, 0.00091277, -0.00031505, -0.00006469, -0.00003731, -0.00001481, -0.00001466, 0.00000160, -0.00001716, 0.00000569, -0.00000921, -0.00000095, -0.00000739, 0.00000604, -0.00000699, 0.00000193, -0.00000814, 0.00000875, -0.00000480, -0.00000327, -0.00000226, 0.00000595, -0.00003213, 0.00002497, -0.00001871, 0.00002506, 0.00000068, 0.00000639, -0.00000225, 0.00001419, -0.00001331, 0.00002482, -0.00003046, 0.00001737, -0.00001396, 0.00000800, -0.00000929, 0.00000490, 0.00000023, -0.00000434, 0.00000026, -0.00000073, -0.00000270, 0.00000900, 0.00000286, -0.00001875, 0.00001259, 0.00000128, -0.00000510, -0.00000579, 0.00000188, -0.00000053, -0.00000153, -0.00000316, 0.00000300, 0.00000165, 0.00000653, -0.00001627, -0.00000384, 0.00000194, 0.00001485, -0.00001141, 0.00000139, -0.00001294, 0.00001228, -0.00000382, 0.00000439, 0.00000644, -0.00001703, },
		{ 0.00000698, 0.00000540, -0.00000330, -0.00000135, 0.00000872, -0.00000534, 0.00000037, -0.00000121, 0.00000503, 0.00000043, -0.00000323, 0.00000339, -0.00000153, -0.00000389, -0.00000001, 0.00000005, 0.00002120, -0.00001565, 0.00000223, -0.00000012, -0.00000277, -0.00000356, 0.00001023, -0.00000312, 0.00000417, -0.00000812, 0.00000858, -0.00000926, 0.00000158, 0.00001194, -0.00002257, 0.00002020, -0.00002118, 0.00000799, -0.00001173, 0.00000679, -0.00001686, 0.00002332, -0.00004015, 0.00003664, -0.00003385, 0.00002663, -0.00000811, 0.00001328, -0.00002269, 0.00002053, -0.00002116, 0.00001989, -0.00003060, 0.00002271, -0.00002617, 0.00002941, -0.00004303, 0.00003754, -0.00005882, 0.00004753, -0.00009336, 0.00007024, -0.00030959, 0.00012507, 0.00075561, -0.00041693, 0.00003745, -0.00009047, 0.00002717, -0.00004276, 0.00002427, -0.00003584, 0.00002364, -0.00002569, 0.00001266, -0.00001755, 0.00001547, -0.00001572, 0.00000876, -0.00001315, 0.00001473, -0.00001183, 0.00000099, -0.00000249, 0.00000091, -0.00002824, 0.00002629, -0.00002147, 0.00003263, -0.00000639, 0.00000983, -0.00000096, 0.00001129, -0.00000934, 0.00002092, -0.00003368, 0.00002568, -0.00002395, 0.00001835, -0.00001992, 0.00001679, -0.00001181, 0.00000511, -0.00000701, 0.00000544, -0.00000812, 0.00001789, -0.00000971, -0.00000965, 0.00000958, 0.00000243, -0.00000869, -0.00000172, -0.00000099, 0.00000236, -0.00000567, 0.00000126, -0.00000039, 0.00000650, 0.00000005, -0.00001490, -0.00000325, 0.00000703, 0.00000898, -0.00001076, 0.00000168, -0.00001302, 0.00001590, -0.00000967, 0.00001229, -0.00000378, -0.00000941, },
		{ 0.00000365, 0.00000700, -0.00000751, 0.00000511, 0.00000396, -0.00000334, -0.00000158, 0.00000169, 0.00000318, 0.00000103, -0.00000438, 0.00000585, -0.00000585, 0.00000093, -0.00000478, 0.00000832, 0.00001448, -0.00001567, 0.00000428, -0.00000102, -0.00000423, 0.00000033, 0.00000812, -0.00000291, 0.00000349, -0.00000828, 0.00001143, -0.00001648, 0.00001411, -0.00000333, -0.00001017, 0.00001190, -0.00001760, 0.00000625, -0.00000876, 0.00000129, -0.00000877, 0.00001362, -0.00003337, 0.00003584, -0.00003757, 0.00003485, -0.00001495, 0.00001541, -0.00002481, 0.00002424, -0.00002507, 0.00002291, -0.00003527, 0.00002908, -0.00003102, 0.00003274, -0.00004813, 0.00004396, -0.00006825, 0.00005792, -0.00010983, 0.00008236, -0.00039562, 0.00042071, 0.00050207, -0.00040651, 0.00007576, -0.00010676, 0.00004237, -0.00005031, 0.00002933, -0.00003879, 0.00002737, -0.00003006, 0.00001620, -0.00001867, 0.00001640, -0.00001680, 0.00000866, -0.00001116, 0.00001362, -0.00001294, 0.00000097, 0.00000061, -0.00000836, -0.00001784, 0.00001993, -0.00001618, 0.00003314, -0.00000867, 0.00001030, 0.00000333, 0.00000405, 0.00000075, 0.00000750, -0.00002485, 0.00002217, -0.00002337, 0.00001931, -0.00002204, 0.00002122, -0.00001822, 0.00001057, -0.00001082, 0.00000802, -0.00000908, 0.00002185, -0.00001922, -0.00000106, 0.00000647, 0.00000271, -0.00001071, 0.00000088, -0.00000233, 0.00000337, -0.00000798, 0.00000439, -0.00000258, 0.00001025, -0.00000693, -0.00001210, -0.00000249, 0.00001142, 0.00000143, -0.00000663, -0.00000148, -0.00000929, 0.00001517, -0.00001131, 0.00001645, -0.00001188, -0.00000130, },
		{ 0.00000089, 0.00000747, -0.00000973, 0.00001034, -0.00000090, -0.00000114, -0.00000319, 0.00000433, 0.00000131, 0.00000134, -0.00000452, 0.00000661, -0.00000826, 0.00000418, -0.00000809, 0.00001618, 0.00000577, -0.00001295, 0.00000488, -0.00000189, -0.00000504, 0.00000435, 0.00000447, -0.00000067, 0.00000033, -0.00000476, 0.00000925, -0.00001795, 0.00002179, -0.00001685, 0.00000416, -0.00000095, -0.00000848, -0.00000029, -0.00000206, -0.00000768, 0.00000353, -0.00000238, -0.00001747, 0.00002428, -0.00003032, 0.00003337, -0.00001407, 0.00000988, -0.00001795, 0.00001822, -0.00001889, 0.00001483, -0.00002768, 0.00002293, -0.00002290, 0.00002106, -0.00003533, 0.00002923, -0.00005266, 0.00003747, -0.00008667, 0.00003109, -0.00037977, 0.00064121, 0.00022149, -0.00032167, 0.00006309, -0.00009161, 0.00003458, -0.00004060, 0.00001998, -0.00002897, 0.00001937, -0.00002425, 0.00001145, -0.00001266, 0.00001075, -0.00001217, 0.00000375, -0.00000443, 0.00000734, -0.00000961, -0.00000185, 0.00000493, -0.00001882, -0.00000441, 0.00000911, -0.00000536, 0.00002828, -0.00000761, 0.00000991, 0.00000721, -0.00000294, 0.00001127, -0.00000943, -0.00000886, 0.00000999, -0.00001399, 0.00001172, -0.00001569, 0.00001746, -0.00001769, 0.00001074, -0.00001031, 0.00000654, -0.00000510, 0.00001973, -0.00002370, 0.00000513, 0.00000439, 0.00000154, -0.00001077, 0.00000165, -0.00000182, 0.00000204, -0.00000768, 0.00000525, -0.00000248, 0.00001142, -0.00001274, -0.00000891, -0.00000079, 0.00001349, -0.00000544, -0.00000108, -0.00000674, -0.00000263, 0.00001044, -0.00000832, 0.00001554, -0.00001552, 0.00000460, },
		{ 0.00000009, 0.00000593, -0.00000899, 0.00001293, -0.00000449, 0.00000031, -0.00000369, 0.00000592, 0.00000004, 0.00000103, -0.00000354, 0.00000552, -0.00000826, 0.00000496, -0.00000855, 0.00002167, -0.00000303, -0.00000853, 0.00000427, -0.00000280, -0.00000466, 0.00000732, 0.00000071, 0.00000231, -0.00000409, 0.00000143, 0.00000229, -0.00001248, 0.00002137, -0.00002349, 0.00001440, -0.00001257, 0.00000129, -0.00000765, 0.00000460, -0.00001591, 0.00001514, -0.00001908, 0.00000192, 0.00000685, -0.00001595, 0.00002529, -0.00000863, 0.00000060, -0.00000649, 0.00000698, -0.00000738, 0.00000076, -0.00001299, 0.00000956, -0.00000792, 0.00000163, -0.00001293, 0.00000257, -0.00002245, -0.00000180, -0.00003857, -0.00006879, -0.00025776, 0.00074357, -0.00002386, -0.00020820, 0.00002376, -0.00005985, 0.00001483, -0.00002267, 0.00000426, -0.00001364, 0.00000600, -0.00001363, 0.00000313, -0.00000395, 0.00000269, -0.00000574, -0.00000207, 0.00000306, -0.00000041, -0.00000489, -0.00000483, 0.00000760, -0.00002686, 0.00000785, -0.00000195, 0.00000733, 0.00002072, -0.00000505, 0.00001061, 0.00000781, -0.00000553, 0.00001647, -0.00002266, 0.00000702, -0.00000454, -0.00000103, -0.00000011, -0.00000438, 0.00000798, -0.00001175, 0.00000668, -0.00000656, 0.00000240, 0.00000224, 0.00001245, -0.00002301, 0.00000835, 0.00000357, -0.00000084, -0.00000931, 0.00000113, -0.00000013, -0.00000101, -0.00000512, 0.00000395, 0.00000002, 0.00000938, -0.00001628, -0.00000608, 0.00000210, 0.00001230, -0.00000952, 0.00000346, -0.00001186, 0.00000489, 0.00000353, -0.00000205, 0.00001010, -0.00001408, 0.00000680, },
		{ 0.00000143, 0.00000248, -0.00000531, 0.00001240, -0.00000612, 0.00000054, -0.00000279, 0.00000610, -0.00000035, 0.00000010, -0.00000175, 0.00000304, -0.00000618, 0.00000328, -0.00000568, 0.00002346, -0.00001018, -0.00000359, 0.00000285, -0.00000372, -0.00000289, 0.00000832, -0.00000174, 0.00000434, -0.00000772, 0.00000781, -0.00000695, -0.00000182, 0.00001290, -0.00002113, 0.00001673, -0.00001833, 0.00000728, -0.00001194, 0.00000762, -0.00001953, 0.00002139, -0.00003081, 0.00001852, -0.00001036, 0.00000018, 0.00001506, -0.00000286, -0.00000761, 0.00000409, -0.00000367, 0.00000322, -0.00001256, 0.00000173, -0.00000369, 0.00000577, -0.00001592, 0.00000777, -0.00002284, 0.00000675, -0.00004064, 0.00000847, -0.00018224, -0.00005726, 0.00071952, -0.00019437, -0.00010439, -0.00001822, -0.00002726, -0.00000512, -0.00000600, -0.00000960, -0.00000023, -0.00000610, -0.00000387, -0.00000390, 0.00000306, -0.00000377, -0.00000110, -0.00000536, 0.00000771, -0.00000617, -0.00000162, -0.00000576, 0.00000657, -0.00002990, 0.00001574, -0.00000972, 0.00001867, 0.00001270, -0.00000205, 0.00001281, 0.00000443, -0.00000224, 0.00001330, -0.00002709, 0.00001646, -0.00001504, 0.00000953, -0.00001067, 0.00000685, -0.00000290, -0.00000367, 0.00000091, -0.00000185, -0.00000192, 0.00001015, 0.00000246, -0.00001851, 0.00000912, 0.00000351, -0.00000364, -0.00000725, 0.00000035, 0.00000151, -0.00000441, -0.00000152, 0.00000149, 0.00000399, 0.00000454, -0.00001723, -0.00000378, 0.00000571, 0.00000799, -0.00000972, 0.00000508, -0.00001465, 0.00001086, -0.00000303, 0.00000498, 0.00000232, -0.00000881, 0.00000545, },
		{ 0.00000400, -0.00000184, 0.00000026, 0.00000924, -0.00000588, -0.00000028, -0.00000074, 0.00000504, 0.00000004, -0.00000120, 0.00000029, 0.00000005, -0.00000299, 0.00000005, -0.00000003, 0.00002121, -0.00001454, 0.00000081, 0.00000097, -0.00000443, -0.00000003, 0.00000714, -0.00000201, 0.00000411, -0.00000857, 0.00001147, -0.00001468, 0.00000993, -0.00000022, -0.00001145, 0.00001093, -0.00001661, 0.00000733, -0.00001123, 0.00000530, -0.00001672, 0.00001983, -0.00003402, 0.00002775, -0.00002239, 0.00001357, 0.00000627, 0.00000018, -0.00001135, 0.00000987, -0.00000948, 0.00000830, -0.00001989, 0.00001091, -0.00001120, 0.00001206, -0.00002434, 0.00001794, -0.00003630, 0.00002172, -0.00006191, 0.00002878, -0.00026710, 0.00017253, 0.00059284, -0.00027864, -0.00003201, -0.00004727, -0.00000421, -0.00001774, 0.00000366, -0.00001675, 0.00000685, -0.00001289, 0.00000163, -0.00000691, 0.00000600, -0.00000659, 0.00000012, -0.00000467, 0.00000789, -0.00000819, -0.00000118, -0.00000393, 0.00000151, -0.00002736, 0.00001803, -0.00001243, 0.00002673, 0.00000544, 0.00000155, 0.00001515, -0.00000104, 0.00000502, 0.00000280, -0.00002178, 0.00001670, -0.00001782, 0.00001365, -0.00001574, 0.00001370, -0.00001112, 0.00000318, -0.00000396, 0.00000161, -0.00000411, 0.00001580, -0.00000736, -0.00001221, 0.00000853, 0.00000343, -0.00000605, -0.00000542, 0.00000018, 0.00000195, -0.00000671, 0.00000163, -0.00000066, 0.00000788, -0.00000186, -0.00001599, -0.00000172, 0.00000895, 0.00000181, -0.00000638, 0.00000313, -0.00001386, 0.00001349, -0.00000697, 0.00001010, -0.00000492, -0.00000214, 0.00000201, },
		{ 0.00000635, -0.00000551, 0.00000611, 0.00000464, -0.00000445, -0.00000152, 0.00000177, 0.00000329, 0.00000082, -0.00000245, 0.00000196, -0.00000251, 0.00000009, -0.00000331, 0.00000698, 0.00001559, -0.00001572, 0.00000397, -0.00000107, -0.00000453, 0.00000310, 0.00000437, -0.00000024, 0.00000136, -0.00000577, 0.00001041, -0.00001748, 0.00001805, -0.00001279, 0.00000098, 0.00000018, -0.00000915, 0.00000225, -0.00000619, -0.00000145, -0.00000846, 0.00001110, -0.00002840, 0.00002810, -0.00002696, 0.00002187, 0.00000046, -0.00000019, -0.00001004, 0.00000994, -0.00000942, 0.00000664, -0.00001964, 0.00001273, -0.00001138, 0.00000953, -0.00002182, 0.00001501, -0.00003428, 0.00001754, -0.00005851, 0.00000827, -0.00028952, 0.00037764, 0.00040728, -0.00028912, 0.00000456, -0.00005870, 0.00000648, -0.00002143, 0.00000557, -0.00001683, 0.00000725, -0.00001393, 0.00000260, -0.00000595, 0.00000512, -0.00000616, -0.00000138, -0.00000092, 0.00000430, -0.00000682, -0.00000315, -0.00000024, -0.00000621, -0.00002053, 0.00001552, -0.00001014, 0.00003099, -0.00000073, 0.00000651, 0.00001537, -0.00000523, 0.00001194, -0.00001048, -0.00001002, 0.00000925, -0.00001305, 0.00001068, -0.00001404, 0.00001424, -0.00001438, 0.00000669, -0.00000634, 0.00000270, -0.00000302, 0.00001741, -0.00001472, -0.00000594, 0.00000755, 0.00000281, -0.00000761, -0.00000419, 0.00000092, 0.00000070, -0.00000703, 0.00000323, -0.00000130, 0.00001017, -0.00000829, -0.00001336, 0.00000056, 0.00001054, -0.00000426, -0.00000106, -0.00000164, -0.00000961, 0.00001230, -0.00000721, 0.00001157, -0.00000919, 0.00000333, -0.00000149, },
		{ 0.00000719, -0.00000717, 0.00001059, 0.00000001, -0.00000275, -0.00000245, 0.00000394, 0.00000155, 0.00000150, -0.00000328, 0.00000283, -0.00000394, 0.00000203, -0.00000538, 0.00001358, 0.00000799, -0.00001409, 0.00000561, -0.00000306, -0.00000366, 0.00000546, 0.00000126, 0.00000246, -0.00000290, -0.00000006, 0.00000456, -0.00001387, 0.00001926, -0.00001999, 0.00001080, -0.00001059, 0.00000014, -0.00000494, 0.00000046, -0.00000974, 0.00000206, -0.00000155, -0.00001658, 0.00002111, -0.00002455, 0.00002512, -0.00000290, -0.00000248, -0.00000563, 0.00000627, -0.00000556, 0.00000045, -0.00001384, 0.00000916, -0.00000672, 0.00000144, -0.00001215, 0.00000308, -0.00002122, -0.00000082, -0.00003630, -0.00004927, -0.00023567, 0.00051585, 0.00021061, -0.00025158, 0.00001326, -0.00005613, 0.00000782, -0.00001909, 0.00000273, -0.00001279, 0.00000365, -0.00001144, 0.00000105, -0.00000304, 0.00000241, -0.00000443, -0.00000346, 0.00000348, -0.00000083, -0.00000387, -0.00000595, 0.00000350, -0.00001432, -0.00001180, 0.00001022, -0.00000409, 0.00003172, -0.00000559, 0.00001316, 0.00001187, -0.00000521, 0.00001408, -0.00002077, 0.00000249, -0.00000128, -0.00000422, 0.00000317, -0.00000734, 0.00000936, -0.00001275, 0.00000660, -0.00000615, 0.00000170, 0.00000097, 0.00001473, -0.00001856, -0.00000082, 0.00000668, 0.00000157, -0.00000837, -0.00000335, 0.00000216, -0.00000186, -0.00000541, 0.00000303, 0.00000005, 0.00000987, -0.00001341, -0.00001012, 0.00000328, 0.00000963, -0.00000830, 0.00000406, -0.00000739, -0.00000330, 0.00000823, -0.00000417, 0.00000918, -0.00000939, 0.00000581, -0.00000337, },
		{ 0.00000596, -0.00000616, 0.00001261, -0.00000352, -0.00000150, -0.00000252, 0.00000517, 0.00000037, 0.00000170, -0.00000352, 0.00000283, -0.00000405, 0.00000238, -0.00000523, 0.00001817, 0.00000006, -0.00001049, 0.00000573, -0.00000477, -0.00000169, 0.00000622, -0.00000083, 0.00000447, -0.00000672, 0.00000632, -0.00000403, -0.00000502, 0.00001307, -0.00001940, 0.00001414, -0.00001693, 0.00000706, -0.00001068, 0.00000546, -0.00001612, 0.00001095, -0.00001383, -0.00000278, 0.00001015, -0.00001741, 0.00002474, -0.00000533, -0.00000430, -0.00000116, 0.00000213, -0.00000147, -0.00000631, -0.00000652, 0.00000419, -0.00000177, -0.00000659, -0.00000210, -0.00000996, -0.00000648, -0.00002185, -0.00001099, -0.00012354, -0.00011568, 0.00056728, 0.00004076, -0.00019344, 0.00000661, -0.00004656, 0.00000501, -0.00001503, -0.00000086, -0.00000844, -0.00000046, -0.00000839, -0.00000049, -0.00000059, 0.00000010, -0.00000345, -0.00000397, 0.00000607, -0.00000499, -0.00000139, -0.00000786, 0.00000554, -0.00002070, -0.00000350, 0.00000433, 0.00000417, 0.00002929, -0.00000845, 0.00002054, 0.00000492, -0.00000025, 0.00000918, -0.00002380, 0.00001037, -0.00000964, 0.00000399, -0.00000473, 0.00000074, 0.00000193, -0.00000813, 0.00000418, -0.00000450, -0.00000005, 0.00000624, 0.00000889, -0.00001900, 0.00000285, 0.00000595, -0.00000003, -0.00000870, -0.00000234, 0.00000300, -0.00000469, -0.00000269, 0.00000161, 0.00000295, 0.00000686, -0.00001645, -0.00000673, 0.00000616, 0.00000615, -0.00000908, 0.00000701, -0.00001204, 0.00000301, 0.00000316, 0.00000047, 0.00000423, -0.00000602, 0.00000492, -0.00000291, },
		{ 0.00000308, -0.00000275, 0.00001193, -0.00000545, -0.00000101, -0.00000161, 0.00000527, -0.00000001, 0.00000128, -0.00000321, 0.00000220, -0.00000316, 0.00000136, -0.00000271, 0.00001979, -0.00000671, -0.00000594, 0.00000455, -0.00000586, 0.00000107, 0.00000510, -0.00000100, 0.00000444, -0.00000813, 0.00001058, -0.00001193, 0.00000563, 0.00000203, -0.00001190, 0.00001013, -0.00001668, 0.00000899, -0.00001251, 0.00000652, -0.00001813, 0.00001526, -0.00002210, 0.00000903, -0.00000115, -0.00000820, 0.00002224, -0.00000796, -0.00000396, 0.00000105, 0.00000012, -0.00000008, -0.00001022, -0.00000125, 0.00000129, -0.00000026, -0.00000992, 0.00000249, -0.00001691, 0.00000099, -0.00003386, -0.00000039, -0.00018576, 0.00004128, 0.00053632, -0.00008171, -0.00013527, -0.00000468, -0.00003590, 0.00000204, -0.00001221, -0.00000263, -0.00000621, -0.00000280, -0.00000666, -0.00000049, 0.00000006, -0.00000063, -0.00000414, -0.00000201, 0.00000559, -0.00000669, -0.00000072, -0.00000791, 0.00000501, -0.00002420, 0.00000289, -0.00000050, 0.00001332, 0.00002390, -0.00000806, 0.00002642, -0.00000312, 0.00000753, -0.00000162, -0.00001870, 0.00001099, -0.00001240, 0.00000793, -0.00000935, 0.00000660, -0.00000481, -0.00000305, 0.00000121, -0.00000282, -0.00000095, 0.00001080, 0.00000173, -0.00001696, 0.00000527, 0.00000517, -0.00000163, -0.00000895, -0.00000069, 0.00000261, -0.00000661, -0.00000007, 0.00000010, 0.00000627, 0.00000185, -0.00001728, -0.00000333, 0.00000848, 0.00000101, -0.00000655, 0.00000675, -0.00001401, 0.00000750, -0.00000092, 0.00000458, -0.00000108, -0.00000089, 0.00000162, -0.00000053, },
		{ -0.00000033, 0.00000198, 0.00000910, -0.00000587, -0.00000113, -0.00000001, 0.00000445, 0.00000031, 0.00000034, -0.00000254, 0.00000138, -0.00000192, -0.00000026, 0.00000156, 0.00001827, -0.00001135, -0.00000139, 0.00000238, -0.00000595, 0.00000388, 0.00000262, 0.00000071, 0.00000199, -0.00000607, 0.00001063, -0.00001584, 0.00001382, -0.00000938, -0.00000115, 0.00000100, -0.00001064, 0.00000577, -0.00001004, 0.00000331, -0.00001529, 0.00001405, -0.00002463, 0.00001637, -0.00001021, 0.00000116, 0.00001838, -0.00001071, -0.00000145, 0.00000056, 0.00000085, -0.00000220, -0.00001007, 0.00000057, 0.00000153, -0.00000310, -0.00000728, -0.00000032, -0.00001502, -0.00000269, -0.00003115, -0.00001579, -0.00020989, 0.00019790, 0.00044550, -0.00015374, -0.00008759, -0.00001487, -0.00002676, 0.00000012, -0.00001108, -0.00000243, -0.00000623, -0.00000314, -0.00000636, 0.00000090, -0.00000086, -0.00000011, -0.00000590, 0.00000172, 0.00000239, -0.00000586, -0.00000198, -0.00000619, 0.00000208, -0.00002476, 0.00000690, -0.00000336, 0.00002232, 0.00001590, -0.00000331, 0.00002820, -0.00000881, 0.00001409, -0.00001419, -0.00000824, 0.00000530, -0.00000926, 0.00000658, -0.00000917, 0.00000823, -0.00000863, 0.00000059, -0.00000096, -0.00000204, -0.00000003, 0.00001312, -0.00000498, -0.00001353, 0.00000676, 0.00000420, -0.00000308, -0.00000913, 0.00000157, 0.00000067, -0.00000684, 0.00000146, -0.00000043, 0.00000869, -0.00000399, -0.00001624, 0.00000012, 0.00000934, -0.00000425, -0.00000180, 0.00000346, -0.00001281, 0.00000919, -0.00000270, 0.00000648, -0.00000461, 0.00000376, -0.00000236, 0.00000257, },
		{ -0.00000295, 0.00000665, 0.00000515, -0.00000530, -0.00000144, 0.00000176, 0.00000316, 0.00000103, -0.00000091, -0.00000172, 0.00000073, -0.00000093, -0.00000153, 0.00000644, 0.00001420, -0.00001348, 0.00000243, -0.00000042, -0.00000474, 0.00000577, -0.00000009, 0.00000333, -0.00000206, -0.00000104, 0.00000604, -0.00001401, 0.00001626, -0.00001671, 0.00000812, -0.00000908, -0.00000187, -0.00000058, -0.00000486, -0.00000267, -0.00000905, 0.00000845, -0.00002181, 0.00001873, -0.00001595, 0.00000974, 0.00001305, -0.00001233, 0.00000165, -0.00000120, 0.00000286, -0.00000632, -0.00000711, 0.00000007, 0.00000334, -0.00000804, -0.00000128, -0.00000783, -0.00000711, -0.00001455, -0.00001708, -0.00005631, -0.00018237, 0.00032125, 0.00032500, -0.00018429, -0.00005247, -0.00002276, -0.00001881, -0.00000175, -0.00001015, -0.00000191, -0.00000693, -0.00000279, -0.00000627, 0.00000236, -0.00000204, 0.00000035, -0.00000720, 0.00000544, -0.00000193, -0.00000358, -0.00000437, -0.00000350, -0.00000238, -0.00002302, 0.00000877, -0.00000383, 0.00003008, 0.00000629, 0.00000586, 0.00002431, -0.00000938, 0.00001540, -0.00002335, 0.00000262, -0.00000298, -0.00000275, 0.00000159, -0.00000513, 0.00000582, -0.00000905, 0.00000211, -0.00000194, -0.00000218, 0.00000267, 0.00001263, -0.00001013, -0.00000955, 0.00000752, 0.00000313, -0.00000455, -0.00000885, 0.00000385, -0.00000236, -0.00000537, 0.00000154, 0.00000058, 0.00000918, -0.00000942, -0.00001384, 0.00000358, 0.00000811, -0.00000797, 0.00000339, -0.00000160, -0.00000912, 0.00000829, -0.00000205, 0.00000561, -0.00000512, 0.00000614, -0.00000529, 0.00000505, },
		{ -0.00000381, 0.00001001, 0.00000114, -0.00000431, -0.00000153, 0.00000318, 0.00000192, 0.00000175, -0.00000225, -0.00000083, 0.00000037, -0.00000055, -0.00000171, 0.00001071, 0.00000861, -0.00001324, 0.00000503, -0.00000336, -0.00000219, 0.00000597, -0.00000176, 0.00000533, -0.00000592, 0.00000504, -0.00000157, -0.00000707, 0.00001204, -0.00001737, 0.00001211, -0.00001592, 0.00000586, -0.00000703, 0.00000051, -0.00000897, -0.00000186, 0.00000082, -0.00001549, 0.00001699, -0.00001835, 0.00001722, 0.00000593, -0.00001136, 0.00000325, -0.00000201, 0.00000370, -0.00000974, -0.00000394, -0.00000028, 0.00000377, -0.00001125, 0.00000337, -0.00001454, 0.00000038, -0.00002664, -0.00000246, -0.00010865, -0.00010658, 0.00039193, 0.00020218, -0.00018686, -0.00002763, -0.00002926, -0.00001079, -0.00000505, -0.00000768, -0.00000295, -0.00000648, -0.00000334, -0.00000497, 0.00000248, -0.00000214, -0.00000050, -0.00000657, 0.00000731, -0.00000554, -0.00000131, -0.00000677, -0.00000084, -0.00000739, -0.00001978, 0.00000895, -0.00000163, 0.00003527, -0.00000295, 0.00001777, 0.00001528, -0.00000428, 0.00000953, -0.00002532, 0.00000915, -0.00000940, 0.00000331, -0.00000384, 0.00000015, 0.00000126, -0.00000712, 0.00000205, -0.00000220, -0.00000246, 0.00000620, 0.00000979, -0.00001335, -0.00000550, 0.00000756, 0.00000219, -0.00000631, -0.00000758, 0.00000523, -0.00000549, -0.00000292, 0.00000058, 0.00000293, 0.00000740, -0.00001355, -0.00001054, 0.00000672, 0.00000478, -0.00000901, 0.00000722, -0.00000670, -0.00000435, 0.00000592, 0.00000009, 0.00000268, -0.00000271, 0.00000556, -0.00000612, 0.00000598, },
		{ -0.00000262, 0.00001137, -0.00000213, -0.00000336, -0.00000116, 0.00000395, 0.00000109, 0.00000216, -0.00000347, 0.00000015, 0.00000016, -0.00000067, -0.00000056, 0.00001346, 0.00000263, -0.00001111, 0.00000612, -0.00000582, 0.00000126, 0.00000428, -0.00000151, 0.00000539, -0.00000768, 0.00000956, -0.00000920, 0.00000218, 0.00000303, -0.00001163, 0.00000950, -0.00001706, 0.00000968, -0.00001102, 0.00000381, -0.00001337, 0.00000394, -0.00000636, -0.00000791, 0.00001261, -0.00001780, 0.00002315, -0.00000262, -0.00000725, 0.00000206, -0.00000035, 0.00000157, -0.00001025, -0.00000279, 0.00000152, 0.00000064, -0.00000975, 0.00000274, -0.00001552, 0.00000118, -0.00003068, -0.00000014, -0.00015232, -0.00000065, 0.00040728, 0.00009496, -0.00017327, -0.00001048, -0.00003486, -0.00000240, -0.00001007, -0.00000319, -0.00000621, -0.00000407, -0.00000545, -0.00000196, 0.00000077, -0.00000071, -0.00000301, -0.00000356, 0.00000641, -0.00000720, -0.00000015, -0.00000839, 0.00000113, -0.00001232, -0.00001559, 0.00000782, 0.00000349, 0.00003639, -0.00000909, 0.00002922, 0.00000390, 0.00000424, -0.00000223, -0.00001939, 0.00000898, -0.00001095, 0.00000588, -0.00000677, 0.00000386, -0.00000303, -0.00000455, 0.00000146, -0.00000254, -0.00000191, 0.00000925, 0.00000561, -0.00001480, -0.00000161, 0.00000689, 0.00000152, -0.00000848, -0.00000496, 0.00000489, -0.00000754, -0.00000056, -0.00000049, 0.00000578, 0.00000366, -0.00001594, -0.00000664, 0.00000893, 0.00000008, -0.00000712, 0.00000850, -0.00001032, -0.00000001, 0.00000352, 0.00000225, -0.00000075, 0.00000138, 0.00000252, -0.00000481, 0.00000522, },
		{ 0.00000016, 0.00001070, -0.00000432, -0.00000260, -0.00000036, 0.00000404, 0.00000080, 0.00000206, -0.00000443, 0.00000127, -0.00000020, -0.00000086, 0.00000164, 0.00001429, -0.00000279, -0.00000771, 0.00000561, -0.00000713, 0.00000473, 0.00000124, 0.00000064, 0.00000304, -0.00000618, 0.00001039, -0.00001372, 0.00000992, -0.00000702, -0.00000239, 0.00000178, -0.00001254, 0.00000870, -0.00001143, 0.00000401, -0.00001478, 0.00000705, -0.00001151, -0.00000077, 0.00000679, -0.00001451, 0.00002657, -0.00001117, -0.00000097, -0.00000155, 0.00000362, -0.00000355, -0.00000744, -0.00000416, 0.00000558, -0.00000597, -0.00000333, -0.00000398, -0.00000929, -0.00000706, -0.00002291, -0.00001841, -0.00016796, 0.00010988, 0.00037826, 0.00001039, -0.00015083, -0.00000009, -0.00003848, 0.00000501, -0.00001536, 0.00000201, -0.00001064, -0.00000051, -0.00000837, 0.00000197, -0.00000206, 0.00000153, -0.00000629, 0.00000101, 0.00000310, -0.00000674, -0.00000042, -0.00000896, 0.00000217, -0.00001688, -0.00001071, 0.00000571, 0.00001142, 0.00003246, -0.00000963, 0.00003653, -0.00000588, 0.00001201, -0.00001575, -0.00000831, 0.00000300, -0.00000734, 0.00000405, -0.00000591, 0.00000439, -0.00000534, -0.00000270, 0.00000116, -0.00000342, 0.00000004, 0.00001080, 0.00000113, -0.00001482, 0.00000191, 0.00000570, 0.00000095, -0.00001074, -0.00000124, 0.00000258, -0.00000779, 0.00000077, -0.00000072, 0.00000800, -0.00000117, -0.00001654, -0.00000237, 0.00000952, -0.00000472, -0.00000303, 0.00000702, -0.00001165, 0.00000288, 0.00000214, 0.00000312, -0.00000298, 0.00000531, -0.00000161, -0.00000212, 0.00000334, },
		{ 0.00000362, 0.00000849, -0.00000546, -0.00000199, 0.00000063, 0.00000363, 0.00000100, 0.00000140, -0.00000496, 0.00000249, -0.00000095, -0.00000057, 0.00000423, 0.00001329, -0.00000703, -0.00000369, 0.00000362, -0.00000674, 0.00000713, -0.00000197, 0.00000375, -0.00000102, -0.00000169, 0.00000685, -0.00001322, 0.00001300, -0.00001405, 0.00000620, -0.00000759, -0.00000463, 0.00000405, -0.00000882, 0.00000148, -0.00001344, 0.00000746, -0.00001420, 0.00000513, 0.00000027, -0.00000843, 0.00002620, -0.00001745, 0.00000521, -0.00000563, 0.00000800, -0.00000979, -0.00000287, -0.00000656, 0.00000992, -0.00001344, 0.00000511, -0.00001384, 0.00000113, -0.00002118, -0.00000665, -0.00005620, -0.00014505, 0.00020137, 0.00032247, -0.00005247, -0.00012286, 0.00000306, -0.00003807, 0.00000908, -0.00001846, 0.00000555, -0.00001406, 0.00000238, -0.00001047, 0.00000519, -0.00000450, 0.00000316, -0.00000876, 0.00000544, -0.00000130, -0.00000489, -0.00000180, -0.00000862, 0.00000230, -0.00002093, -0.00000544, 0.00000328, 0.00002130, 0.00002374, -0.00000342, 0.00003707, -0.00001061, 0.00001480, -0.00002587, 0.00000309, -0.00000529, -0.00000089, -0.00000072, -0.00000207, 0.00000190, -0.00000530, -0.00000201, 0.00000129, -0.00000464, 0.00000318, 0.00001050, -0.00000294, -0.00001364, 0.00000470, 0.00000443, 0.00000003, -0.00001239, 0.00000277, -0.00000117, -0.00000620, 0.00000073, 0.00000042, 0.00000868, -0.00000611, -0.00001556, 0.00000197, 0.00000807, -0.00000824, 0.00000187, 0.00000350, -0.00001082, 0.00000415, 0.00000209, 0.00000220, -0.00000295, 0.00000748, -0.00000526, 0.00000081, 0.00000132, },
		{ 0.00000676, 0.00000545, -0.00000578, -0.00000139, 0.00000156, 0.00000301, 0.00000149, 0.00000018, -0.00000487, 0.00000359, -0.00000203, 0.00000054, 0.00000649, 0.00001094, -0.00000977, 0.00000028, 0.00000055, -0.00000448, 0.00000756, -0.00000400, 0.00000631, -0.00000514, 0.00000414, 0.00000015, -0.00000789, 0.00001021, -0.00001540, 0.00001053, -0.00001475, 0.00000339, -0.00000196, -0.00000483, -0.00000245, -0.00001045, 0.00000603, -0.00001481, 0.00000958, -0.00000643, 0.00000022, 0.00002113, -0.00001937, 0.00000873, -0.00000769, 0.00001016, -0.00001437, 0.00000082, -0.00000751, 0.00001157, -0.00001795, 0.00001099, -0.00002162, 0.00000972, -0.00003401, 0.00000866, -0.00010255, -0.00008592, 0.00025938, 0.00025667, -0.00009774, -0.00009109, -0.00000133, -0.00003215, 0.00000803, -0.00001740, 0.00000541, -0.00001450, 0.00000287, -0.00001020, 0.00000621, -0.00000517, 0.00000289, -0.00000898, 0.00000802, -0.00000519, -0.00000273, -0.00000371, -0.00000760, 0.00000146, -0.00002412, -0.00000033, 0.00000165, 0.00003126, 0.00001216, 0.00000848, 0.00003047, -0.00000902, 0.00001030, -0.00002864, 0.00001010, -0.00001161, 0.00000488, -0.00000555, 0.00000242, -0.00000196, -0.00000381, -0.00000201, 0.00000134, -0.00000544, 0.00000665, 0.00000871, -0.00000638, -0.00001136, 0.00000639, 0.00000355, -0.00000178, -0.00001255, 0.00000582, -0.00000519, -0.00000351, -0.00000032, 0.00000279, 0.00000740, -0.00001033, -0.00001326, 0.00000588, 0.00000467, -0.00000945, 0.00000605, -0.00000081, -0.00000859, 0.00000432, 0.00000282, -0.00000007, -0.00000058, 0.00000711, -0.00000730, 0.00000299, 0.00000006, },
		{ 0.00000883, 0.00000226, -0.00000553, -0.00000072, 0.00000224, 0.00000243, 0.00000204, -0.00000151, -0.00000395, 0.00000415, -0.00000309, 0.00000241, 0.00000792, 0.00000777, -0.00001089, 0.00000356, -0.00000290, -0.00000069, 0.00000569, -0.00000388, 0.00000691, -0.00000741, 0.00000887, -0.00000710, 0.00000001, 0.00000283, -0.00001093, 0.00000896, -0.00001708, 0.00000867, -0.00000698, -0.00000123, -0.00000630, -0.00000704, 0.00000376, -0.00001396, 0.00001248, -0.00001254, 0.00001046, 0.00001167, -0.00001606, 0.00000806, -0.00000606, 0.00000822, -0.00001505, 0.00000148, -0.00000507, 0.00000836, -0.00001657, 0.00001050, -0.00002262, 0.00001070, -0.00003803, 0.00001175, -0.00014048, -0.00000454, 0.00028154, 0.00019170, -0.00012867, -0.00005801, -0.00001172, -0.00002121, 0.00000195, -0.00001194, 0.00000113, -0.00001134, 0.00000039, -0.00000708, 0.00000453, -0.00000363, 0.00000040, -0.00000645, 0.00000782, -0.00000744, -0.00000111, -0.00000571, -0.00000599, -0.00000066, -0.00002581, 0.00000376, 0.00000227, 0.00003875, 0.00000104, 0.00002292, 0.00001894, -0.00000268, -0.00000062, -0.00002312, 0.00001021, -0.00001299, 0.00000708, -0.00000776, 0.00000511, -0.00000511, -0.00000230, -0.00000183, 0.00000066, -0.00000494, 0.00000938, 0.00000616, -0.00000920, -0.00000810, 0.00000680, 0.00000335, -0.00000472, -0.00001062, 0.00000684, -0.00000817, -0.00000086, -0.00000150, 0.00000566, 0.00000434, -0.00001330, -0.00000989, 0.00000868, 0.00000003, -0.00000807, 0.00000840, -0.00000468, -0.00000597, 0.00000424, 0.00000337, -0.00000253, 0.00000316, 0.00000442, -0.00000737, 0.00000398, 0.00000006, },
		{ 0.00000954, -0.00000062, -0.00000488, -0.00000004, 0.00000264, 0.00000201, 0.00000245, -0.00000347, -0.00000212, 0.00000377, -0.00000353, 0.00000461, 0.00000838, 0.00000426, -0.00001042, 0.00000557, -0.00000580, 0.00000366, 0.00000203, -0.00000151, 0.00000495, -0.00000655, 0.00001034, -0.00001195, 0.00000713, -0.00000597, -0.00000290, 0.00000242, -0.00001408, 0.00000986, -0.00000959, 0.00000083, -0.00000917, -0.00000394, 0.00000123, -0.00001192, 0.00001345, -0.00001679, 0.00002032, -0.00000029, -0.00000853, 0.00000349, -0.00000087, 0.00000212, -0.00001142, -0.00000134, 0.00000079, 0.00000039, -0.00000904, 0.00000279, -0.00001535, 0.00000174, -0.00002963, -0.00000484, -0.00015370, 0.00007914, 0.00027554, 0.00013143, -0.00014595, -0.00002771, -0.00002445, -0.00000798, -0.00000690, -0.00000393, -0.00000581, -0.00000569, -0.00000410, -0.00000203, 0.00000093, -0.00000062, -0.00000346, -0.00000190, 0.00000504, -0.00000776, -0.00000042, -0.00000756, -0.00000376, -0.00000439, -0.00002530, 0.00000605, 0.00000637, 0.00004134, -0.00000591, 0.00003560, 0.00000640, 0.00000463, -0.00001419, -0.00001178, 0.00000410, -0.00000902, 0.00000480, -0.00000616, 0.00000457, -0.00000612, -0.00000183, -0.00000090, -0.00000105, -0.00000275, 0.00001057, 0.00000348, -0.00001136, -0.00000422, 0.00000612, 0.00000364, -0.00000850, -0.00000668, 0.00000531, -0.00000915, 0.00000075, -0.00000184, 0.00000799, 0.00000018, -0.00001485, -0.00000571, 0.00000973, -0.00000471, -0.00000463, 0.00000852, -0.00000733, -0.00000368, 0.00000456, 0.00000285, -0.00000384, 0.00000680, 0.00000044, -0.00000585, 0.00000393, 0.00000124, },
		{ 0.00000898, -0.00000296, -0.00000393, 0.00000056, 0.00000284, 0.00000181, 0.00000251, -0.00000539, 0.00000046, 0.00000224, -0.00000277, 0.00000643, 0.00000808, 0.00000072, -0.00000851, 0.00000588, -0.00000721, 0.00000731, -0.00000219, 0.00000224, 0.00000094, -0.00000258, 0.00000767, -0.00001239, 0.00001044, -0.00001251, 0.00000507, -0.00000623, -0.00000732, 0.00000732, -0.00000964, 0.00000114, -0.00001089, -0.00000125, -0.00000153, -0.00000860, 0.00001187, -0.00001764, 0.00002720, -0.00001168, 0.00000055, -0.00000283, 0.00000585, -0.00000619, -0.00000512, -0.00000604, 0.00000797, -0.00000968, 0.00000181, -0.00000937, -0.00000257, -0.00001439, -0.00001157, -0.00004056, -0.00013319, 0.00014662, 0.00025370, 0.00007498, -0.00014859, -0.00000500, -0.00003478, 0.00000354, -0.00001498, 0.00000351, -0.00001265, 0.00000014, -0.00000858, 0.00000307, -0.00000291, 0.00000229, -0.00000707, 0.00000306, 0.00000083, -0.00000664, -0.00000057, -0.00000919, -0.00000101, -0.00000981, -0.00002231, 0.00000636, 0.00001420, 0.00003781, -0.00000604, 0.00004272, -0.00000317, 0.00000863, -0.00002534, 0.00000068, -0.00000488, -0.00000187, -0.00000057, -0.00000160, 0.00000109, -0.00000484, -0.00000260, 0.00000069, -0.00000343, 0.00000082, 0.00001011, 0.00000098, -0.00001265, -0.00000036, 0.00000491, 0.00000380, -0.00001223, -0.00000158, 0.00000164, -0.00000794, 0.00000084, -0.00000074, 0.00000887, -0.00000423, -0.00001497, -0.00000114, 0.00000866, -0.00000832, -0.00000019, 0.00000675, -0.00000859, -0.00000194, 0.00000539, 0.00000098, -0.00000316, 0.00000894, -0.00000349, -0.00000357, 0.00000342, 0.00000312, },
		{ 0.00000749, -0.00000468, -0.00000276, 0.00000098, 0.00000293, 0.00000183, 0.00000201, -0.00000681, 0.00000333, -0.00000024, -0.00000059, 0.00000729, 0.00000739, -0.00000263, -0.00000547, 0.00000443, -0.00000653, 0.00000903, -0.00000546, 0.00000581, -0.00000356, 0.00000309, 0.00000168, -0.00000828, 0.00000856, -0.00001419, 0.00000954, -0.00001350, 0.00000042, 0.00000269, -0.00000795, 0.00000020, -0.00001179, 0.00000133, -0.00000474, -0.00000383, 0.00000738, -0.00001397, 0.00002881, -0.00001929, 0.00000791, -0.00000794, 0.00001110, -0.00001359, 0.00000095, -0.00000981, 0.00001318, -0.00001778, 0.00001125, -0.00002070, 0.00000982, -0.00003084, 0.00000755, -0.00008595, -0.00008091, 0.00018705, 0.00022705, 0.00002054, -0.00013613, 0.00000657, -0.00003875, 0.00000984, -0.00001908, 0.00000744, -0.00001666, 0.00000378, -0.00001096, 0.00000631, -0.00000530, 0.00000360, -0.00000882, 0.00000668, -0.00000335, -0.00000489, -0.00000134, -0.00001041, 0.00000176, -0.00001637, -0.00001728, 0.00000549, 0.00002458, 0.00002888, 0.00000120, 0.00004236, -0.00000747, 0.00000656, -0.00002985, 0.00000935, -0.00001239, 0.00000493, -0.00000618, 0.00000364, -0.00000362, -0.00000229, -0.00000392, 0.00000223, -0.00000554, 0.00000483, 0.00000847, -0.00000144, -0.00001271, 0.00000272, 0.00000395, 0.00000298, -0.00001471, 0.00000326, -0.00000302, -0.00000518, -0.00000033, 0.00000175, 0.00000784, -0.00000813, -0.00001379, 0.00000325, 0.00000559, -0.00000990, 0.00000398, 0.00000390, -0.00000879, -0.00000045, 0.00000625, -0.00000176, -0.00000046, 0.00000883, -0.00000630, -0.00000136, 0.00000308, 0.00000507, },
		{ 0.00000544, -0.00000575, -0.00000150, 0.00000122, 0.00000298, 0.00000202, 0.00000081, -0.00000722, 0.00000575, -0.00000301, 0.00000273, 0.00000691, 0.00000660, -0.00000553, -0.00000181, 0.00000158, -0.00000372, 0.00000817, -0.00000653, 0.00000764, -0.00000663, 0.00000817, -0.00000535, -0.00000138, 0.00000231, -0.00001051, 0.00000873, -0.00001673, 0.00000648, -0.00000211, -0.00000570, -0.00000131, -0.00001218, 0.00000390, -0.00000835, 0.00000206, 0.00000038, -0.00000584, 0.00002420, -0.00002103, 0.00001103, -0.00000937, 0.00001230, -0.00001727, 0.00000406, -0.00001013, 0.00001358, -0.00002031, 0.00001485, -0.00002599, 0.00001559, -0.00003989, 0.00001684, -0.00012566, -0.00000906, 0.00019974, 0.00020131, -0.00003154, -0.00011072, 0.00000639, -0.00003489, 0.00000947, -0.00001775, 0.00000641, -0.00001640, 0.00000394, -0.00001015, 0.00000662, -0.00000533, 0.00000258, -0.00000787, 0.00000781, -0.00000634, -0.00000324, -0.00000258, -0.00001081, 0.00000365, -0.00002289, -0.00001135, 0.00000506, 0.00003503, 0.00001725, 0.00001386, 0.00003521, -0.00000674, -0.00000159, -0.00002622, 0.00001127, -0.00001514, 0.00000828, -0.00000921, 0.00000698, -0.00000741, -0.00000005, -0.00000475, 0.00000282, -0.00000636, 0.00000814, 0.00000638, -0.00000397, -0.00001126, 0.00000444, 0.00000383, 0.00000058, -0.00001490, 0.00000638, -0.00000717, -0.00000203, -0.00000188, 0.00000494, 0.00000501, -0.00001102, -0.00001145, 0.00000673, 0.00000118, -0.00000914, 0.00000693, 0.00000082, -0.00000845, 0.00000124, 0.00000637, -0.00000434, 0.00000344, 0.00000657, -0.00000752, 0.00000032, 0.00000329, 0.00000655, },
		{ 0.00000315, -0.00000614, -0.00000034, 0.00000138, 0.00000298, 0.00000233, -0.00000111, -0.00000623, 0.00000692, -0.00000510, 0.00000641, 0.00000553, 0.00000575, -0.00000767, 0.00000174, -0.00000185, 0.00000055, 0.00000488, -0.00000497, 0.00000682, -0.00000678, 0.00001045, -0.00001061, 0.00000531, -0.00000564, -0.00000327, 0.00000310, -0.00001503, 0.00000926, -0.00000565, -0.00000372, -0.00000303, -0.00001201, 0.00000610, -0.00001170, 0.00000799, -0.00000763, 0.00000517, 0.00001432, -0.00001681, 0.00000920, -0.00000639, 0.00000855, -0.00001599, 0.00000301, -0.00000613, 0.00000831, -0.00001595, 0.00001067, -0.00002262, 0.00001131, -0.00003679, 0.00000807, -0.00014420, 0.00006448, 0.00019257, 0.00017616, -0.00007765, -0.00007764, -0.00000273, -0.00002473, 0.00000355, -0.00001188, 0.00000112, -0.00001227, 0.00000093, -0.00000647, 0.00000428, -0.00000329, -0.00000036, -0.00000452, 0.00000632, -0.00000766, -0.00000205, -0.00000428, -0.00000992, 0.00000360, -0.00002794, -0.00000595, 0.00000695, 0.00004255, 0.00000673, 0.00002821, 0.00002405, -0.00000352, -0.00001297, -0.00001620, 0.00000651, -0.00001223, 0.00000691, -0.00000820, 0.00000683, -0.00000870, 0.00000069, -0.00000432, 0.00000197, -0.00000531, 0.00000992, 0.00000447, -0.00000663, -0.00000844, 0.00000469, 0.00000462, -0.00000343, -0.00001242, 0.00000682, -0.00000952, 0.00000029, -0.00000277, 0.00000779, 0.00000101, -0.00001268, -0.00000815, 0.00000861, -0.00000353, -0.00000641, 0.00000821, -0.00000188, -0.00000791, 0.00000340, 0.00000512, -0.00000559, 0.00000720, 0.00000299, -0.00000732, 0.00000146, 0.00000405, 0.00000731, },
		{ 0.00000080, -0.00000583, 0.00000050, 0.00000160, 0.00000291, 0.00000262, -0.00000355, -0.00000374, 0.00000631, -0.00000559, 0.00000941, 0.00000373, 0.00000469, -0.00000867, 0.00000438, -0.00000478, 0.00000505, 0.00000020, -0.00000137, 0.00000355, -0.00000370, 0.00000878, -0.00001200, 0.00000891, -0.00001193, 0.00000434, -0.00000498, -0.00000944, 0.00000862, -0.00000746, -0.00000230, -0.00000502, -0.00001081, 0.00000707, -0.00001363, 0.00001231, -0.00001433, 0.00001623, 0.00000190, -0.00000859, 0.00000379, -0.00000025, 0.00000105, -0.00001065, -0.00000132, 0.00000086, -0.00000090, -0.00000644, 0.00000026, -0.00001196, -0.00000184, -0.00002246, -0.00002011, -0.00013195, 0.00012273, 0.00017730, 0.00014749, -0.00011222, -0.00004416, -0.00001557, -0.00001202, -0.00000477, -0.00000415, -0.00000608, -0.00000623, -0.00000357, -0.00000144, 0.00000062, -0.00000042, -0.00000392, -0.00000004, 0.00000299, -0.00000753, -0.00000136, -0.00000644, -0.00000749, 0.00000089, -0.00003040, -0.00000208, 0.00001239, 0.00004476, 0.00000091, 0.00004008, 0.00001245, -0.00000129, -0.00002318, -0.00000398, -0.00000207, -0.00000538, 0.00000183, -0.00000371, 0.00000328, -0.00000722, -0.00000038, -0.00000259, -0.00000015, -0.00000251, 0.00001000, 0.00000289, -0.00000910, -0.00000484, 0.00000394, 0.00000578, -0.00000833, -0.00000777, 0.00000451, -0.00000950, 0.00000107, -0.00000221, 0.00000928, -0.00000333, -0.00001311, -0.00000423, 0.00000850, -0.00000741, -0.00000257, 0.00000794, -0.00000402, -0.00000717, 0.00000586, 0.00000240, -0.00000478, 0.00000959, -0.00000079, -0.00000626, 0.00000237, 0.00000507, 0.00000738, },
		{ -0.00000147, -0.00000489, 0.00000093, 0.00000199, 0.00000276, 0.00000268, -0.00000606, -0.00000014, 0.00000387, -0.00000391, 0.00001080, 0.00000224, 0.00000315, -0.00000824, 0.00000549, -0.00000619, 0.00000837, -0.00000433, 0.00000286, -0.00000088, 0.00000155, 0.00000357, -0.00000901, 0.00000782, -0.00001396, 0.00000908, -0.00001233, -0.00000225, 0.00000563, -0.00000785, -0.00000129, -0.00000752, -0.00000803, 0.00000594, -0.00001297, 0.00001340, -0.00001740, 0.00002416, -0.00000948, 0.00000043, -0.00000249, 0.00000633, -0.00000744, -0.00000378, -0.00000648, 0.00000794, -0.00001049, 0.00000418, -0.00001200, 0.00000119, -0.00001841, -0.00000359, -0.00006082, -0.00008899, 0.00015534, 0.00016412, 0.00011108, -0.00013029, -0.00001719, -0.00002657, -0.00000106, -0.00001188, 0.00000223, -0.00001228, -0.00000078, -0.00000740, 0.00000297, -0.00000267, 0.00000179, -0.00000652, 0.00000396, -0.00000094, -0.00000652, -0.00000111, -0.00000882, -0.00000383, -0.00000447, -0.00002994, 0.00000024, 0.00002126, 0.00004096, 0.00000188, 0.00004641, 0.00000321, -0.00000272, -0.00002822, 0.00000570, -0.00001033, 0.00000209, -0.00000430, 0.00000212, -0.00000210, -0.00000393, -0.00000259, -0.00000024, -0.00000269, 0.00000120, 0.00000882, 0.00000139, -0.00001077, -0.00000134, 0.00000302, 0.00000637, -0.00001287, -0.00000223, 0.00000033, -0.00000743, 0.00000030, -0.00000003, 0.00000879, -0.00000718, -0.00001245, -0.00000021, 0.00000640, -0.00000959, 0.00000138, 0.00000661, -0.00000574, -0.00000592, 0.00000799, -0.00000119, -0.00000193, 0.00000987, -0.00000380, -0.00000495, 0.00000343, 0.00000592, 0.00000692, },
		{ -0.00000351, -0.00000349, 0.00000099, 0.00000253, 0.00000259, 0.00000224, -0.00000798, 0.00000378, 0.00000020, -0.00000017, 0.00001009, 0.00000162, 0.00000099, -0.00000634, 0.00000479, -0.00000550, 0.00000943, -0.00000721, 0.00000605, -0.00000460, 0.00000689, -0.00000324, -0.00000295, 0.00000247, -0.00001098, 0.00000900, -0.00001629, 0.00000406, 0.00000182, -0.00000748, -0.00000050, -0.00001052, -0.00000359, 0.00000240, -0.00000925, 0.00001049, -0.00001542, 0.00002659, -0.00001671, 0.00000710, -0.00000678, 0.00001046, -0.00001382, 0.00000166, -0.00000969, 0.00001198, -0.00001666, 0.00001140, -0.00002094, 0.00001085, -0.00003120, 0.00001005, -0.00010092, -0.00002512, 0.00016134, 0.00015783, 0.00006582, -0.00012998, -0.00000111, -0.00003181, 0.00000508, -0.00001517, 0.00000497, -0.00001532, 0.00000219, -0.00000894, 0.00000530, -0.00000432, 0.00000229, -0.00000707, 0.00000620, -0.00000431, -0.00000524, -0.00000132, -0.00001088, 0.00000015, -0.00001162, -0.00002715, 0.00000205, 0.00003188, 0.00003248, 0.00000942, 0.00004612, -0.00000273, -0.00000836, -0.00002626, 0.00000960, -0.00001475, 0.00000694, -0.00000856, 0.00000667, -0.00000706, -0.00000051, -0.00000475, 0.00000172, -0.00000458, 0.00000471, 0.00000717, -0.00000042, -0.00001105, 0.00000117, 0.00000278, 0.00000545, -0.00001567, 0.00000260, -0.00000431, -0.00000430, -0.00000133, 0.00000327, 0.00000628, -0.00000996, -0.00001085, 0.00000330, 0.00000283, -0.00000972, 0.00000459, 0.00000478, -0.00000724, -0.00000378, 0.00000896, -0.00000457, 0.00000216, 0.00000812, -0.00000558, -0.00000375, 0.00000483, 0.00000629, 0.00000612, },
		{ -0.00000510, -0.00000195, 0.00000086, 0.00000309, 0.00000250, 0.00000104, -0.00000863, 0.00000696, -0.00000359, 0.00000482, 0.00000749, 0.00000197, -0.00000165, -0.00000333, 0.00000258, -0.00000278, 0.00000783, -0.00000754, 0.00000697, -0.00000596, 0.00001010, -0.00000906, 0.00000357, -0.00000493, -0.00000432, 0.00000417, -0.00001569, 0.00000780, -0.00000158, -0.00000681, -0.00000002, -0.00001352, 0.00000184, -0.00000294, -0.00000307, 0.00000411, -0.00000858, 0.00002291, -0.00001825, 0.00000956, -0.00000736, 0.00001035, -0.00001606, 0.00000368, -0.00000920, 0.00001116, -0.00001708, 0.00001228, -0.00002298, 0.00001267, -0.00003462, 0.00000980, -0.00012581, 0.00004368, 0.00014819, 0.00015680, 0.00001502, -0.00011362, 0.00000346, -0.00003019, 0.00000560, -0.00001400, 0.00000352, -0.00001457, 0.00000211, -0.00000772, 0.00000510, -0.00000402, 0.00000095, -0.00000539, 0.00000616, -0.00000647, -0.00000405, -0.00000212, -0.00001191, 0.00000319, -0.00001913, -0.00002323, 0.00000501, 0.00004159, 0.00002227, 0.00002130, 0.00004016, -0.00000627, -0.00001641, -0.00001833, 0.00000711, -0.00001386, 0.00000742, -0.00000913, 0.00000800, -0.00000975, 0.00000155, -0.00000580, 0.00000248, -0.00000500, 0.00000710, 0.00000567, -0.00000276, -0.00000971, 0.00000223, 0.00000363, 0.00000255, -0.00001585, 0.00000537, -0.00000792, -0.00000137, -0.00000277, 0.00000669, 0.00000224, -0.00001139, -0.00000852, 0.00000569, -0.00000135, -0.00000798, 0.00000664, 0.00000282, -0.00000859, -0.00000066, 0.00000813, -0.00000656, 0.00000622, 0.00000504, -0.00000625, -0.00000261, 0.00000644, 0.00000613, 0.00000507, },
		{ -0.00000603, -0.00000058, 0.00000081, 0.00000350, 0.00000255, -0.00000099, -0.00000754, 0.00000841, -0.00000617, 0.00000973, 0.00000378, 0.00000296, -0.00000430, 0.00000010, -0.00000039, 0.00000119, 0.00000409, -0.00000539, 0.00000532, -0.00000424, 0.00000973, -0.00001164, 0.00000774, -0.00001132, 0.00000321, -0.00000346, -0.00001112, 0.00000852, -0.00000398, -0.00000595, -0.00000032, -0.00001550, 0.00000689, -0.00000854, 0.00000387, -0.00000385, 0.00000126, 0.00001446, -0.00001458, 0.00000783, -0.00000432, 0.00000605, -0.00001390, 0.00000203, -0.00000511, 0.00000580, -0.00001197, 0.00000675, -0.00001776, 0.00000594, -0.00002754, -0.00000777, -0.00012487, 0.00010098, 0.00012776, 0.00015491, -0.00003454, -0.00008711, -0.00000072, -0.00002333, 0.00000190, -0.00000966, -0.00000094, -0.00001099, -0.00000024, -0.00000450, 0.00000301, -0.00000239, -0.00000149, -0.00000219, 0.00000415, -0.00000731, -0.00000309, -0.00000362, -0.00001136, 0.00000414, -0.00002560, -0.00001930, 0.00001064, 0.00004770, 0.00001385, 0.00003411, 0.00003069, -0.00000937, -0.00002360, -0.00000778, 0.00000031, -0.00000866, 0.00000391, -0.00000597, 0.00000573, -0.00000946, 0.00000157, -0.00000531, 0.00000186, -0.00000380, 0.00000804, 0.00000457, -0.00000542, -0.00000709, 0.00000200, 0.00000535, -0.00000206, -0.00001329, 0.00000543, -0.00000954, 0.00000038, -0.00000304, 0.00000909, -0.00000243, -0.00001151, -0.00000574, 0.00000656, -0.00000520, -0.00000498, 0.00000751, 0.00000082, -0.00000954, 0.00000312, 0.00000543, -0.00000637, 0.00000903, 0.00000167, -0.00000626, -0.00000123, 0.00000785, 0.00000561, 0.00000374, },
		{ -0.00000617, 0.00000035, 0.00000105, 0.00000363, 0.00000266, -0.00000362, -0.00000469, 0.00000760, -0.00000647, 0.00001316, 0.00000012, 0.00000393, -0.00000630, 0.00000308, -0.00000308, 0.00000521, -0.00000060, -0.00000173, 0.00000193, -0.00000001, 0.00000572, -0.00001006, 0.00000776, -0.00001408, 0.00000849, -0.00001100, -0.00000448, 0.00000687, -0.00000547, -0.00000471, -0.00000195, -0.00001536, 0.00000997, -0.00001255, 0.00000940, -0.00001083, 0.00001121, 0.00000399, -0.00000784, 0.00000356, 0.00000068, -0.00000069, -0.00000887, -0.00000187, 0.00000076, -0.00000181, -0.00000388, -0.00000249, -0.00000816, -0.00000608, -0.00001396, -0.00003897, -0.00009547, 0.00013538, 0.00011133, 0.00014507, -0.00007516, -0.00005786, -0.00000904, -0.00001438, -0.00000341, -0.00000447, -0.00000624, -0.00000641, -0.00000327, -0.00000072, 0.00000028, -0.00000051, -0.00000385, 0.00000130, 0.00000104, -0.00000714, -0.00000239, -0.00000571, -0.00000917, 0.00000244, -0.00003015, -0.00001581, 0.00001942, 0.00004857, 0.00001008, 0.00004457, 0.00001993, -0.00001366, -0.00002683, 0.00000131, -0.00000725, -0.00000189, -0.00000148, -0.00000074, 0.00000099, -0.00000679, -0.00000014, -0.00000360, 0.00000032, -0.00000146, 0.00000780, 0.00000362, -0.00000781, -0.00000395, 0.00000116, 0.00000716, -0.00000740, -0.00000879, 0.00000306, -0.00000897, 0.00000056, -0.00000168, 0.00000961, -0.00000676, -0.00001055, -0.00000287, 0.00000582, -0.00000793, -0.00000151, 0.00000746, -0.00000130, -0.00000961, 0.00000676, 0.00000143, -0.00000394, 0.00000982, -0.00000109, -0.00000613, 0.00000072, 0.00000861, 0.00000502, 0.00000209, },
		{ -0.00000553, 0.00000076, 0.00000164, 0.00000348, 0.00000263, -0.00000633, -0.00000062, 0.00000468, -0.00000405, 0.00001409, -0.00000243, 0.00000416, -0.00000704, 0.00000484, -0.00000453, 0.00000801, -0.00000477, 0.00000197, -0.00000165, 0.00000507, -0.00000052, -0.00000511, 0.00000359, -0.00001213, 0.00000940, -0.00001575, 0.00000194, 0.00000408, -0.00000643, -0.00000297, -0.00000513, -0.00001249, 0.00000996, -0.00001349, 0.00001171, -0.00001452, 0.00001838, -0.00000536, -0.00000082, -0.00000088, 0.00000523, -0.00000732, -0.00000337, -0.00000574, 0.00000587, -0.00000857, 0.00000359, -0.00001139, 0.00000121, -0.00001787, -0.00000140, -0.00007398, -0.00004389, 0.00014371, 0.00010547, 0.00012268, -0.00010100, -0.00003245, -0.00001695, -0.00000651, -0.00000777, -0.00000065, -0.00001030, -0.00000267, -0.00000542, 0.00000220, -0.00000188, 0.00000059, -0.00000507, 0.00000391, -0.00000218, -0.00000640, -0.00000203, -0.00000797, -0.00000587, -0.00000168, -0.00003271, -0.00001226, 0.00003050, 0.00004424, 0.00001220, 0.00005044, 0.00000924, -0.00001943, -0.00002460, 0.00000587, -0.00001223, 0.00000344, -0.00000608, 0.00000418, -0.00000422, -0.00000320, -0.00000256, -0.00000150, -0.00000131, 0.00000115, 0.00000702, 0.00000237, -0.00000921, -0.00000124, 0.00000060, 0.00000803, -0.00001215, -0.00000368, -0.00000073, -0.00000678, -0.00000048, 0.00000115, 0.00000788, -0.00000993, -0.00000886, -0.00000028, 0.00000374, -0.00000910, 0.00000174, 0.00000681, -0.00000367, -0.00000836, 0.00000930, -0.00000274, 0.00000001, 0.00000859, -0.00000281, -0.00000608, 0.00000335, 0.00000842, 0.00000461, 0.00000016, },
		{ -0.00000431, 0.00000078, 0.00000246, 0.00000317, 0.00000215, -0.00000838, 0.00000369, 0.00000051, 0.00000068, 0.00001222, -0.00000322, 0.00000324, -0.00000621, 0.00000497, -0.00000421, 0.00000872, -0.00000721, 0.00000433, -0.00000378, 0.00000891, -0.00000668, 0.00000108, -0.00000310, -0.00000630, 0.00000561, -0.00001627, 0.00000634, 0.00000126, -0.00000714, -0.00000094, -0.00000943, -0.00000713, 0.00000673, -0.00001094, 0.00001008, -0.00001375, 0.00002096, -0.00001123, 0.00000418, -0.00000351, 0.00000733, -0.00001161, 0.00000042, -0.00000763, 0.00000813, -0.00001201, 0.00000744, -0.00001642, 0.00000618, -0.00002428, 0.00000245, -0.00010035, 0.00001705, 0.00013103, 0.00011040, 0.00008767, -0.00010992, -0.00001484, -0.00002150, -0.00000167, -0.00000970, 0.00000059, -0.00001195, -0.00000079, -0.00000581, 0.00000345, -0.00000282, 0.00000048, -0.00000468, 0.00000491, -0.00000479, -0.00000544, -0.00000214, -0.00000979, -0.00000242, -0.00000740, -0.00003382, -0.00000743, 0.00004181, 0.00003646, 0.00001967, 0.00005091, -0.00000104, -0.00002539, -0.00001770, 0.00000505, -0.00001291, 0.00000538, -0.00000790, 0.00000678, -0.00000800, -0.00000022, -0.00000460, 0.00000017, -0.00000224, 0.00000321, 0.00000629, 0.00000048, -0.00000914, 0.00000032, 0.00000102, 0.00000715, -0.00001509, 0.00000060, -0.00000461, -0.00000401, -0.00000193, 0.00000469, 0.00000416, -0.00001144, -0.00000687, 0.00000169, 0.00000086, -0.00000867, 0.00000433, 0.00000578, -0.00000619, -0.00000561, 0.00000992, -0.00000581, 0.00000426, 0.00000597, -0.00000361, -0.00000599, 0.00000639, 0.00000729, 0.00000442, -0.00000191, },
		{ -0.00000283, 0.00000065, 0.00000329, 0.00000286, 0.00000095, -0.00000906, 0.00000708, -0.00000356, 0.00000657, 0.00000805, -0.00000224, 0.00000120, -0.00000395, 0.00000356, -0.00000215, 0.00000717, -0.00000742, 0.00000465, -0.00000344, 0.00000988, -0.00001050, 0.00000590, -0.00000961, 0.00000101, -0.00000141, -0.00001271, 0.00000801, -0.00000099, -0.00000753, 0.00000070, -0.00001377, -0.00000050, 0.00000129, -0.00000572, 0.00000516, -0.00000887, 0.00001871, -0.00001276, 0.00000611, -0.00000351, 0.00000618, -0.00001251, 0.00000146, -0.00000675, 0.00000689, -0.00001128, 0.00000655, -0.00001618, 0.00000504, -0.00002310, -0.00000676, -0.00010768, 0.00007257, 0.00010772, 0.00012098, 0.00004435, -0.00010379, -0.00000586, -0.00002189, -0.00000018, -0.00000917, -0.00000063, -0.00001123, -0.00000077, -0.00000450, 0.00000304, -0.00000259, -0.00000065, -0.00000290, 0.00000420, -0.00000642, -0.00000449, -0.00000284, -0.00001061, 0.00000015, -0.00001368, -0.00003415, -0.00000002, 0.00005088, 0.00002800, 0.00003052, 0.00004625, -0.00001117, -0.00002933, -0.00000880, 0.00000028, -0.00000968, 0.00000375, -0.00000644, 0.00000624, -0.00000928, 0.00000115, -0.00000551, 0.00000097, -0.00000216, 0.00000435, 0.00000588, -0.00000203, -0.00000761, 0.00000054, 0.00000256, 0.00000427, -0.00001558, 0.00000303, -0.00000744, -0.00000167, -0.00000282, 0.00000785, -0.00000079, -0.00001121, -0.00000496, 0.00000286, -0.00000222, -0.00000694, 0.00000609, 0.00000440, -0.00000854, -0.00000162, 0.00000833, -0.00000681, 0.00000752, 0.00000297, -0.00000398, -0.00000538, 0.00000924, 0.00000556, 0.00000433, -0.00000383, },
		{ -0.00000144, 0.00000066, 0.00000389, 0.00000266, -0.00000102, -0.00000796, 0.00000857, -0.00000610, 0.00001201, 0.00000277, -0.00000011, -0.00000144, -0.00000085, 0.00000119, 0.00000101, 0.00000389, -0.00000570, 0.00000313, -0.00000067, 0.00000744, -0.00001069, 0.00000737, -0.00001335, 0.00000690, -0.00000907, -0.00000660, 0.00000728, -0.00000265, -0.00000728, 0.00000113, -0.00001675, 0.00000567, -0.00000460, 0.00000042, -0.00000126, -0.00000156, 0.00001294, -0.00001063, 0.00000535, -0.00000142, 0.00000243, -0.00001044, 0.00000008, -0.00000370, 0.00000305, -0.00000742, 0.00000197, -0.00001176, -0.00000093, -0.00001612, -0.00002789, -0.00009138, 0.00011082, 0.00008513, 0.00012961, -0.00000033, -0.00008739, -0.00000383, -0.00001904, -0.00000107, -0.00000716, -0.00000327, -0.00000904, -0.00000183, -0.00000225, 0.00000161, -0.00000173, -0.00000211, -0.00000043, 0.00000221, -0.00000706, -0.00000371, -0.00000407, -0.00001021, 0.00000116, -0.00001976, -0.00003395, 0.00001063, 0.00005567, 0.00002173, 0.00004211, 0.00003729, -0.00002116, -0.00002931, -0.00000119, -0.00000564, -0.00000454, -0.00000002, -0.00000270, 0.00000311, -0.00000817, 0.00000077, -0.00000522, 0.00000097, -0.00000125, 0.00000466, 0.00000561, -0.00000470, -0.00000512, -0.00000020, 0.00000483, -0.00000017, -0.00001372, 0.00000325, -0.00000859, -0.00000043, -0.00000243, 0.00000955, -0.00000589, -0.00000952, -0.00000343, 0.00000319, -0.00000494, -0.00000437, 0.00000707, 0.00000260, -0.00001015, 0.00000284, 0.00000491, -0.00000545, 0.00000891, 0.00000054, -0.00000449, -0.00000375, 0.00001120, 0.00000380, 0.00000405, -0.00000530, },
		{ -0.00000038, 0.00000100, 0.00000411, 0.00000255, -0.00000355, -0.00000512, 0.00000774, -0.00000610, 0.00001548, -0.00000217, 0.00000222, -0.00000387, 0.00000225, -0.00000126, 0.00000430, -0.00000010, -0.00000299, 0.00000077, 0.00000346, 0.00000237, -0.00000741, 0.00000492, -0.00001291, 0.00000908, -0.00001469, -0.00000007, 0.00000516, -0.00000400, -0.00000610, -0.00000029, -0.00001718, 0.00000970, -0.00000905, 0.00000547, -0.00000703, 0.00000587, 0.00000579, -0.00000650, 0.00000320, 0.00000134, -0.00000233, -0.00000684, -0.00000244, 0.00000003, -0.00000155, -0.00000260, -0.00000383, -0.00000600, -0.00000826, -0.00000827, -0.00005445, -0.00005428, 0.00012633, 0.00007163, 0.00012944, -0.00003928, -0.00006645, -0.00000586, -0.00001470, -0.00000290, -0.00000500, -0.00000598, -0.00000656, -0.00000293, -0.00000004, -0.00000002, -0.00000093, -0.00000317, 0.00000185, -0.00000041, -0.00000692, -0.00000324, -0.00000554, -0.00000880, 0.00000046, -0.00002541, -0.00003271, 0.00002405, 0.00005534, 0.00001975, 0.00005190, 0.00002486, -0.00003021, -0.00002479, 0.00000270, -0.00000992, 0.00000005, -0.00000380, 0.00000146, -0.00000114, -0.00000562, -0.00000077, -0.00000411, 0.00000061, -0.00000006, 0.00000462, 0.00000508, -0.00000685, -0.00000250, -0.00000117, 0.00000701, -0.00000523, -0.00001022, 0.00000162, -0.00000809, -0.00000039, -0.00000054, 0.00000909, -0.00001001, -0.00000697, -0.00000242, 0.00000279, -0.00000691, -0.00000141, 0.00000737, 0.00000029, -0.00001046, 0.00000675, 0.00000063, -0.00000220, 0.00000823, -0.00000082, -0.00000543, -0.00000085, 0.00001169, 0.00000253, 0.00000332, -0.00000606, },
		{ 0.00000024, 0.00000170, 0.00000394, 0.00000236, -0.00000610, -0.00000115, 0.00000486, -0.00000327, 0.00001601, -0.00000553, 0.00000377, -0.00000530, 0.00000448, -0.00000296, 0.00000674, -0.00000368, -0.00000041, -0.00000112, 0.00000725, -0.00000352, -0.00000216, -0.00000042, -0.00000854, 0.00000677, -0.00001652, 0.00000501, 0.00000273, -0.00000532, -0.00000403, -0.00000358, -0.00001463, 0.00001063, -0.00001080, 0.00000796, -0.00001039, 0.00001139, -0.00000060, -0.00000222, 0.00000114, 0.00000326, -0.00000637, -0.00000340, -0.00000459, 0.00000287, -0.00000506, 0.00000098, -0.00000833, -0.00000182, -0.00001327, -0.00000531, -0.00007705, -0.00000533, 0.00012084, 0.00007034, 0.00011687, -0.00006740, -0.00004599, -0.00000920, -0.00001044, -0.00000450, -0.00000365, -0.00000780, -0.00000468, -0.00000332, 0.00000138, -0.00000122, -0.00000062, -0.00000334, 0.00000325, -0.00000302, -0.00000627, -0.00000320, -0.00000684, -0.00000696, -0.00000163, -0.00003085, -0.00002933, 0.00003868, 0.00005050, 0.00002292, 0.00005785, 0.00000978, -0.00003674, -0.00001697, 0.00000222, -0.00001099, 0.00000232, -0.00000583, 0.00000431, -0.00000486, -0.00000290, -0.00000263, -0.00000274, 0.00000036, 0.00000087, 0.00000470, 0.00000386, -0.00000789, -0.00000053, -0.00000164, 0.00000823, -0.00000984, -0.00000616, -0.00000106, -0.00000647, -0.00000114, 0.00000248, 0.00000628, -0.00001229, -0.00000435, -0.00000186, 0.00000182, -0.00000790, 0.00000151, 0.00000708, -0.00000241, -0.00000913, 0.00000915, -0.00000324, 0.00000187, 0.00000598, -0.00000119, -0.00000667, 0.00000315, 0.00001054, 0.00000209, 0.00000197, -0.00000604, },
		{ 0.00000053, 0.00000261, 0.00000352, 0.00000182, -0.00000801, 0.00000297, 0.00000090, 0.00000179, 0.00001343, -0.00000663, 0.00000393, -0.00000519, 0.00000528, -0.00000332, 0.00000765, -0.00000603, 0.00000118, -0.00000148, 0.00000912, -0.00000818, 0.00000285, -0.00000649, -0.00000206, 0.00000085, -0.00001423, 0.00000760, 0.00000069, -0.00000659, -0.00000157, -0.00000807, -0.00000960, 0.00000852, -0.00000961, 0.00000740, -0.00001058, 0.00001388, -0.00000481, 0.00000094, 0.00000010, 0.00000342, -0.00000851, -0.00000129, -0.00000539, 0.00000393, -0.00000647, 0.00000215, -0.00001013, -0.00000086, -0.00001396, -0.00001108, -0.00008687, 0.00004362, 0.00010166, 0.00007911, 0.00009234, -0.00008263, -0.00002920, -0.00001199, -0.00000719, -0.00000536, -0.00000341, -0.00000843, -0.00000368, -0.00000278, 0.00000177, -0.00000175, -0.00000090, -0.00000256, 0.00000347, -0.00000513, -0.00000538, -0.00000362, -0.00000760, -0.00000531, -0.00000463, -0.00003636, -0.00002251, 0.00005222, 0.00004298, 0.00003083, 0.00005848, -0.00000689, -0.00003901, -0.00000838, -0.00000140, -0.00000898, 0.00000188, -0.00000544, 0.00000489, -0.00000698, -0.00000102, -0.00000409, -0.00000153, 0.00000047, 0.00000126, 0.00000514, 0.00000186, -0.00000756, 0.00000034, -0.00000113, 0.00000790, -0.00001309, -0.00000256, -0.00000385, -0.00000447, -0.00000197, 0.00000581, 0.00000157, -0.00001227, -0.00000241, -0.00000150, 0.00000041, -0.00000783, 0.00000409, 0.00000623, -0.00000522, -0.00000623, 0.00000946, -0.00000557, 0.00000547, 0.00000313, -0.00000110, -0.00000762, 0.00000758, 0.00000803, 0.00000247, 0.00000009, -0.00000531, },
		{ 0.00000067, 0.00000350, 0.00000303, 0.00000071, -0.00000867, 0.00000618, -0.00000281, 0.00000778, 0.00000846, -0.00000553, 0.00000265, -0.00000351, 0.00000452, -0.00000222, 0.00000684, -0.00000684, 0.00000153, 0.00000002, 0.00000820, -0.00001011, 0.00000565, -0.00001092, 0.00000398, -0.00000650, -0.00000888, 0.00000768, -0.00000079, -0.00000748, 0.00000047, -0.00001255, -0.00000334, 0.00000439, -0.00000623, 0.00000439, -0.00000795, 0.00001332, -0.00000649, 0.00000256, 0.00000022, 0.00000180, -0.00000853, -0.00000085, -0.00000470, 0.00000325, -0.00000592, 0.00000104, -0.00000942, -0.00000273, -0.00001109, -0.00002553, -0.00007898, 0.00008186, 0.00007835, 0.00009234, 0.00005957, -0.00008582, -0.00001734, -0.00001345, -0.00000511, -0.00000557, -0.00000395, -0.00000812, -0.00000329, -0.00000161, 0.00000132, -0.00000177, -0.00000148, -0.00000113, 0.00000255, -0.00000649, -0.00000455, -0.00000437, -0.00000773, -0.00000426, -0.00000829, -0.00004185, -0.00001139, 0.00006242, 0.00003524, 0.00004198, 0.00005285, -0.00002347, -0.00003613, -0.00000172, -0.00000590, -0.00000534, -0.00000038, -0.00000315, 0.00000334, -0.00000723, -0.00000038, -0.00000480, -0.00000059, 0.00000086, 0.00000120, 0.00000583, -0.00000070, -0.00000604, 0.00000010, 0.00000040, 0.00000589, -0.00001452, -0.00000010, -0.00000604, -0.00000276, -0.00000217, 0.00000841, -0.00000406, -0.00001009, -0.00000164, -0.00000105, -0.00000136, -0.00000670, 0.00000605, 0.00000484, -0.00000768, -0.00000224, 0.00000766, -0.00000574, 0.00000754, 0.00000072, -0.00000128, -0.00000753, 0.00001148, 0.00000486, 0.00000336, -0.00000209, -0.00000411, },
		{ 0.00000089, 0.00000414, 0.00000260, -0.00000101, -0.00000773, 0.00000763, -0.00000495, 0.00001310, 0.00000243, -0.00000296, 0.00000041, -0.00000072, 0.00000254, 0.00000004, 0.00000459, -0.00000636, 0.00000101, 0.00000287, 0.00000471, -0.00000890, 0.00000524, -0.00001210, 0.00000729, -0.00001275, -0.00000235, 0.00000602, -0.00000192, -0.00000759, 0.00000123, -0.00001572, 0.00000258, -0.00000029, -0.00000203, 0.00000018, -0.00000365, 0.00001061, -0.00000611, 0.00000296, 0.00000094, -0.00000084, -0.00000703, -0.00000165, -0.00000303, 0.00000156, -0.00000434, -0.00000120, -0.00000759, -0.00000556, -0.00000771, -0.00004457, -0.00005396, 0.00010295, 0.00005940, 0.00010349, 0.00002398, -0.00007978, -0.00001016, -0.00001358, -0.00000396, -0.00000552, -0.00000475, -0.00000742, -0.00000304, -0.00000033, 0.00000044, -0.00000155, -0.00000201, 0.00000046, 0.00000079, -0.00000703, -0.00000402, -0.00000516, -0.00000742, -0.00000384, -0.00001269, -0.00004661, 0.00000398, 0.00006763, 0.00002981, 0.00005416, 0.00004065, -0.00003762, -0.00002868, 0.00000126, -0.00000909, -0.00000192, -0.00000294, -0.00000019, 0.00000058, -0.00000609, -0.00000083, -0.00000479, 0.00000018, 0.00000122, 0.00000104, 0.00000641, -0.00000331, -0.00000383, -0.00000090, 0.00000258, 0.00000249, -0.00001413, 0.00000095, -0.00000721, -0.00000172, -0.00000125, 0.00000930, -0.00000933, -0.00000644, -0.00000214, -0.00000030, -0.00000339, -0.00000457, 0.00000722, 0.00000297, -0.00000927, 0.00000202, 0.00000436, -0.00000382, 0.00000759, -0.00000055, -0.00000226, -0.00000577, 0.00001388, 0.00000191, 0.00000425, -0.00000417, -0.00000272, },
		{ 0.00000133, 0.00000437, 0.00000226, -0.00000317, -0.00000531, 0.00000703, -0.00000467, 0.00001630, -0.00000314, 0.00000000, -0.00000193, 0.00000237, 0.00000004, 0.00000280, 0.00000152, -0.00000511, 0.00000037, 0.00000595, -0.00000021, -0.00000535, 0.00000198, -0.00000973, 0.00000667, -0.00001592, 0.00000343, 0.00000365, -0.00000303, -0.00000668, 0.00000026, -0.00001665, 0.00000686, -0.00000411, 0.00000160, -0.00000381, 0.00000097, 0.00000696, -0.00000458, 0.00000279, 0.00000150, -0.00000348, -0.00000496, -0.00000291, -0.00000115, -0.00000024, -0.00000283, -0.00000330, -0.00000619, -0.00000723, -0.00000765, -0.00006159, -0.00001745, 0.00010599, 0.00004981, 0.00010740, -0.00000918, -0.00006802, -0.00000668, -0.00001273, -0.00000344, -0.00000550, -0.00000539, -0.00000672, -0.00000251, 0.00000059, -0.00000045, -0.00000139, -0.00000215, 0.00000171, -0.00000137, -0.00000686, -0.00000396, -0.00000568, -0.00000707, -0.00000385, -0.00001826, -0.00004928, 0.00002251, 0.00006733, 0.00002871, 0.00006461, 0.00002250, -0.00004679, -0.00001876, 0.00000046, -0.00000985, -0.00000004, -0.00000449, 0.00000214, -0.00000227, -0.00000439, -0.00000191, -0.00000426, 0.00000093, 0.00000124, 0.00000116, 0.00000648, -0.00000542, -0.00000155, -0.00000212, 0.00000489, -0.00000174, -0.00001228, 0.00000069, -0.00000735, -0.00000139, 0.00000084, 0.00000788, -0.00001293, -0.00000249, -0.00000349, 0.00000066, -0.00000544, -0.00000169, 0.00000747, 0.00000076, -0.00000963, 0.00000563, 0.00000055, -0.00000052, 0.00000583, -0.00000055, -0.00000413, -0.00000217, 0.00001414, -0.00000003, 0.00000457, -0.00000576, -0.00000141, },
		{ 0.00000201, 0.00000418, 0.00000191, -0.00000534, -0.00000191, 0.00000470, -0.00000182, 0.00001650, -0.00000698, 0.00000228, -0.00000346, 0.00000483, -0.00000216, 0.00000532, -0.00000168, -0.00000369, 0.00000037, 0.00000799, -0.00000495, -0.00000102, -0.00000272, -0.00000492, 0.00000243, -0.00001520, 0.00000711, 0.00000146, -0.00000429, -0.00000488, -0.00000238, -0.00001509, 0.00000880, -0.00000623, 0.00000372, -0.00000654, 0.00000477, 0.00000346, -0.00000272, 0.00000256, 0.00000132, -0.00000530, -0.00000320, -0.00000395, 0.00000036, -0.00000160, -0.00000199, -0.00000455, -0.00000601, -0.00000669, -0.00001347, -0.00006997, 0.00002199, 0.00009495, 0.00005043, 0.00010157, -0.00003597, -0.00005385, -0.00000570, -0.00001125, -0.00000337, -0.00000557, -0.00000574, -0.00000617, -0.00000160, 0.00000091, -0.00000109, -0.00000138, -0.00000180, 0.00000230, -0.00000351, -0.00000622, -0.00000437, -0.00000572, -0.00000703, -0.00000401, -0.00002541, -0.00004824, 0.00004219, 0.00006214, 0.00003311, 0.00007034, 0.00000025, -0.00004916, -0.00000925, -0.00000274, -0.00000835, -0.00000001, -0.00000449, 0.00000311, -0.00000432, -0.00000286, -0.00000314, -0.00000338, 0.00000169, 0.00000079, 0.00000178, 0.00000580, -0.00000663, 0.00000025, -0.00000306, 0.00000674, -0.00000613, -0.00000948, -0.00000056, -0.00000667, -0.00000150, 0.00000372, 0.00000408, -0.00001391, 0.00000053, -0.00000493, 0.00000147, -0.00000712, 0.00000157, 0.00000683, -0.00000157, -0.00000859, 0.00000786, -0.00000266, 0.00000305, 0.00000304, 0.00000030, -0.00000641, 0.00000285, 0.00001215, -0.00000052, 0.00000396, -0.00000658, -0.00000035, },
		{ 0.00000285, 0.00000367, 0.00000137, -0.00000700, 0.00000163, 0.00000150, 0.00000294, 0.00001368, -0.00000842, 0.00000317, -0.00000354, 0.00000593, -0.00000333, 0.00000694, -0.00000444, -0.00000246, 0.00000139, 0.00000812, -0.00000808, 0.00000240, -0.00000695, 0.00000036, -0.00000390, -0.00001107, 0.00000824, -0.00000011, -0.00000560, -0.00000267, -0.00000607, -0.00001146, 0.00000848, -0.00000646, 0.00000403, -0.00000757, 0.00000721, 0.00000075, -0.00000104, 0.00000246, 0.00000031, -0.00000594, -0.00000222, -0.00000439, 0.00000131, -0.00000246, -0.00000179, -0.00000510, -0.00000667, -0.00000469, -0.00002511, -0.00006562, 0.00005586, 0.00007658, 0.00005866, 0.00008629, -0.00005435, -0.00003980, -0.00000618, -0.00000945, -0.00000369, -0.00000559, -0.00000593, -0.00000560, -0.00000048, 0.00000066, -0.00000144, -0.00000148, -0.00000105, 0.00000209, -0.00000526, -0.00000542, -0.00000504, -0.00000535, -0.00000744, -0.00000433, -0.00003423, -0.00004202, 0.00006052, 0.00005381, 0.00004299, 0.00006857, -0.00002285, -0.00004451, -0.00000264, -0.00000632, -0.00000576, -0.00000122, -0.00000320, 0.00000264, -0.00000524, -0.00000197, -0.00000414, -0.00000229, 0.00000234, -0.00000001, 0.00000290, 0.00000433, -0.00000678, 0.00000125, -0.00000335, 0.00000768, -0.00001008, -0.00000632, -0.00000235, -0.00000551, -0.00000157, 0.00000660, -0.00000147, -0.00001198, 0.00000161, -0.00000556, 0.00000160, -0.00000789, 0.00000461, 0.00000548, -0.00000377, -0.00000629, 0.00000829, -0.00000441, 0.00000576, 0.00000029, 0.00000118, -0.00000824, 0.00000832, 0.00000844, 0.00000044, 0.00000231, -0.00000650, 0.00000041, },
		{ 0.00000365, 0.00000302, 0.00000049, -0.00000770, 0.00000447, -0.00000141, 0.00000840, 0.00000859, -0.00000757, 0.00000258, -0.00000204, 0.00000534, -0.00000308, 0.00000724, -0.00000640, -0.00000148, 0.00000329, 0.00000613, -0.00000888, 0.00000375, -0.00000919, 0.00000406, -0.00001012, -0.00000499, 0.00000719, -0.00000105, -0.00000669, -0.00000070, -0.00000990, -0.00000665, 0.00000650, -0.00000523, 0.00000279, -0.00000700, 0.00000822, -0.00000098, 0.00000033, 0.00000236, -0.00000125, -0.00000554, -0.00000209, -0.00000418, 0.00000176, -0.00000304, -0.00000181, -0.00000560, -0.00000703, -0.00000342, -0.00003963, -0.00004830, 0.00007821, 0.00005786, 0.00007008, 0.00006394, -0.00006403, -0.00002754, -0.00000724, -0.00000761, -0.00000434, -0.00000541, -0.00000617, -0.00000479, 0.00000054, 0.00000008, -0.00000157, -0.00000157, -0.00000011, 0.00000114, -0.00000640, -0.00000478, -0.00000568, -0.00000485, -0.00000811, -0.00000517, -0.00004415, -0.00002974, 0.00007501, 0.00004494, 0.00005689, 0.00005741, -0.00004270, -0.00003454, -0.00000007, -0.00000862, -0.00000337, -0.00000268, -0.00000137, 0.00000123, -0.00000516, -0.00000182, -0.00000471, -0.00000105, 0.00000270, -0.00000091, 0.00000435, 0.00000222, -0.00000591, 0.00000132, -0.00000279, 0.00000739, -0.00001307, -0.00000332, -0.00000427, -0.00000422, -0.00000114, 0.00000850, -0.00000754, -0.00000768, 0.00000041, -0.00000478, 0.00000057, -0.00000727, 0.00000683, 0.00000375, -0.00000562, -0.00000310, 0.00000699, -0.00000430, 0.00000682, -0.00000151, 0.00000128, -0.00000865, 0.00001300, 0.00000399, 0.00000236, -0.00000015, -0.00000560, 0.00000094, },
		{ 0.00000423, 0.00000237, -0.00000081, -0.00000718, 0.00000596, -0.00000297, 0.00001309, 0.00000253, -0.00000515, 0.00000097, 0.00000064, 0.00000326, -0.00000138, 0.00000609, -0.00000744, -0.00000056, 0.00000549, 0.00000258, -0.00000746, 0.00000284, -0.00000877, 0.00000481, -0.00001426, 0.00000123, 0.00000487, -0.00000165, -0.00000720, 0.00000043, -0.00001299, -0.00000168, 0.00000373, -0.00000323, 0.00000061, -0.00000527, 0.00000810, -0.00000184, 0.00000140, 0.00000202, -0.00000285, -0.00000452, -0.00000260, -0.00000340, 0.00000178, -0.00000354, -0.00000171, -0.00000661, -0.00000605, -0.00000536, -0.00005233, -0.00002152, 0.00008688, 0.00004401, 0.00008011, 0.00003794, -0.00006598, -0.00001794, -0.00000825, -0.00000603, -0.00000512, -0.00000505, -0.00000651, -0.00000357, 0.00000115, -0.00000060, -0.00000159, -0.00000148, 0.00000073, -0.00000033, -0.00000688, -0.00000453, -0.00000596, -0.00000462, -0.00000867, -0.00000721, -0.00005398, -0.00001144, 0.00008362, 0.00003857, 0.00007181, 0.00003675, -0.00005540, -0.00002249, -0.00000100, -0.00000903, -0.00000198, -0.00000363, 0.00000027, -0.00000053, -0.00000445, -0.00000234, -0.00000471, 0.00000020, 0.00000265, -0.00000165, 0.00000587, -0.00000028, -0.00000419, 0.00000046, -0.00000134, 0.00000571, -0.00001469, -0.00000096, -0.00000590, -0.00000309, 0.00000012, 0.00000849, -0.00001257, -0.00000230, -0.00000258, -0.00000260, -0.00000168, -0.00000505, 0.00000774, 0.00000202, -0.00000694, 0.00000043, 0.00000443, -0.00000249, 0.00000598, -0.00000184, 0.00000012, -0.00000696, 0.00001572, -0.00000001, 0.00000445, -0.00000295, -0.00000413, 0.00000137, },
		},
		{ { -0.00000693, -0.00000421, 0.00000393, 0.00000516, -0.00000521, 0.00000264, -0.00000600, 0.00000186, 0.00000045, 0.00000255, -0.00000393, 0.00000105, 0.00000026, -0.00000212, 0.00000367, -0.00000297, -0.00000064, 0.00000182, 0.00000374, 0.00000088, 0.00000446, -0.00000477, 0.00000427, -0.00000475, 0.00000414, -0.00000242, 0.00000527, -0.00000490, 0.00001260, -0.00000390, -0.00000095, -0.00000020, 0.00000197, 0.00000445, -0.00002337, 0.00000373, -0.00000163, -0.00000205, -0.00000414, 0.00000283, -0.00001211, 0.00000240, -0.00000353, -0.00000025, -0.00000556, -0.00000407, 0.00000637, -0.00000732, -0.00000440, -0.00000357, -0.00000642, -0.00001164, -0.00006068, 0.00001154, 0.00008677, 0.00003737, 0.00008159, 0.00000876, -0.00006021, -0.00001209, -0.00000830, -0.00000837, -0.00000410, -0.00000473, -0.00000423, -0.00000473, -0.00000410, -0.00000837, -0.00000830, -0.00001209, -0.00006021, 0.00000876, 0.00008159, 0.00003737, 0.00008677, 0.00001154, -0.00006068, -0.00001164, -0.00000642, -0.00000357, -0.00000440, -0.00000732, 0.00000637, -0.00000407, -0.00000556, -0.00000025, -0.00000353, 0.00000240, -0.00001211, 0.00000283, -0.00000414, -0.00000205, -0.00000163, 0.00000373, -0.00002337, 0.00000445, 0.00000197, -0.00000020, -0.00000095, -0.00000390, 0.00001260, -0.00000490, 0.00000527, -0.00000242, 0.00000414, -0.00000475, 0.00000427, -0.00000477, 0.00000446, 0.00000088, 0.00000374, 0.00000182, -0.00000064, -0.00000297, 0.00000367, -0.00000212, 0.00000026, 0.00000105, -0.00000393, 0.00000255, 0.00000045, 0.00000186, -0.00000600, 0.00000264, -0.00000521, 0.00000516, 0.00000393, -0.00000421, },
		{ -0.00000683, -0.00000368, 0.00000577, 0.00000300, -0.00000459, 0.00000247, -0.00000671, 0.00000354, -0.00000041, 0.00000320, -0.00000562, 0.00000313, -0.00000199, 0.00000032, 0.00000230, -0.00000280, -0.00000113, 0.00000326, 0.00000242, 0.00000250, 0.00000268, -0.00000396, 0.00000359, -0.00000434, 0.00000406, -0.00000161, 0.00000414, -0.00000380, 0.00001333, -0.00000757, 0.00000255, -0.00000310, 0.00000635, -0.00000180, -0.00002052, 0.00000490, -0.00000239, -0.00000256, -0.00000286, 0.00000162, -0.00001174, 0.00000294, -0.00000395, 0.00000040, -0.00000719, -0.00000200, 0.00000529, -0.00000774, -0.00000346, -0.00000484, -0.00000487, -0.00001859, -0.00005907, 0.00003344, 0.00007678, 0.00003712, 0.00007881, -0.00000498, -0.00005538, -0.00000861, -0.00000923, -0.00000748, -0.00000401, -0.00000472, -0.00000432, -0.00000453, -0.00000454, -0.00000888, -0.00000763, -0.00001631, -0.00006410, 0.00002463, 0.00008126, 0.00004126, 0.00009276, -0.00001274, -0.00005653, -0.00000731, -0.00000722, -0.00000274, -0.00000522, -0.00000683, 0.00000744, -0.00000642, -0.00000330, -0.00000173, -0.00000230, 0.00000110, -0.00001169, 0.00000334, -0.00000493, -0.00000187, -0.00000028, 0.00000106, -0.00002418, 0.00000957, -0.00000249, 0.00000315, -0.00000508, 0.00000117, 0.00000991, -0.00000434, 0.00000529, -0.00000256, 0.00000364, -0.00000448, 0.00000425, -0.00000478, 0.00000544, -0.00000019, 0.00000467, 0.00000063, -0.00000031, -0.00000267, 0.00000422, -0.00000395, 0.00000230, -0.00000111, -0.00000188, 0.00000140, 0.00000165, -0.00000036, -0.00000435, 0.00000172, -0.00000469, 0.00000646, 0.00000225, -0.00000473, },
		{ -0.00000658, -0.00000302, 0.00000750, 0.00000030, -0.00000299, 0.00000116, -0.00000623, 0.00000436, -0.00000063, 0.00000301, -0.00000645, 0.00000460, -0.00000386, 0.00000286, 0.00000046, -0.00000233, -0.00000158, 0.00000466, 0.00000101, 0.00000434, 0.00000036, -0.00000249, 0.00000232, -0.00000333, 0.00000355, -0.00000039, 0.00000221, -0.00000125, 0.00001194, -0.00000931, 0.00000473, -0.00000477, 0.00000964, -0.00000808, -0.00001639, 0.00000485, -0.00000270, -0.00000317, -0.00000135, -0.00000003, -0.00001085, 0.00000296, -0.00000380, 0.00000037, -0.00000823, -0.00000018, 0.00000414, -0.00000793, -0.00000275, -0.00000602, -0.00000338, -0.00002716, -0.00005159, 0.00005090, 0.00006503, 0.00003945, 0.00007330, -0.00001639, -0.00004994, -0.00000591, -0.00001026, -0.00000630, -0.00000425, -0.00000452, -0.00000455, -0.00000418, -0.00000528, -0.00000901, -0.00000732, -0.00002130, -0.00006655, 0.00004217, 0.00007785, 0.00004948, 0.00009255, -0.00003663, -0.00004775, -0.00000584, -0.00000685, -0.00000266, -0.00000572, -0.00000631, 0.00000842, -0.00000891, -0.00000052, -0.00000400, -0.00000023, -0.00000109, -0.00001037, 0.00000299, -0.00000506, -0.00000208, 0.00000159, -0.00000313, -0.00002249, 0.00001263, -0.00000595, 0.00000598, -0.00000891, 0.00000679, 0.00000583, -0.00000233, 0.00000419, -0.00000198, 0.00000259, -0.00000358, 0.00000358, -0.00000401, 0.00000555, -0.00000056, 0.00000502, -0.00000014, -0.00000024, -0.00000185, 0.00000381, -0.00000480, 0.00000364, -0.00000277, 0.00000001, 0.00000022, 0.00000276, -0.00000263, -0.00000220, 0.00000004, -0.00000317, 0.00000684, 0.00000086, -0.00000528, },
		{ -0.00000627, -0.00000208, 0.00000876, -0.00000250, -0.00000080, -0.00000104, -0.00000461, 0.00000425, -0.00000013, 0.00000186, -0.00000617, 0.00000506, -0.00000484, 0.00000492, -0.00000145, -0.00000183, -0.00000174, 0.00000573, -0.00000013, 0.00000600, -0.00000211, -0.00000067, 0.00000070, -0.00000196, 0.00000285, 0.00000086, 0.00000003, 0.00000223, 0.00000874, -0.00000906, 0.00000534, -0.00000480, 0.00001110, -0.00001340, -0.00001179, 0.00000403, -0.00000277, -0.00000366, 0.00000007, -0.00000181, -0.00000972, 0.00000275, -0.00000337, -0.00000013, -0.00000882, 0.00000146, 0.00000285, -0.00000772, -0.00000256, -0.00000660, -0.00000290, -0.00003586, -0.00003912, 0.00006283, 0.00005342, 0.00004320, 0.00006565, -0.00002550, -0.00004413, -0.00000403, -0.00001117, -0.00000503, -0.00000470, -0.00000419, -0.00000483, -0.00000383, -0.00000616, -0.00000890, -0.00000733, -0.00002718, -0.00006692, 0.00006053, 0.00007192, 0.00006199, 0.00008437, -0.00005708, -0.00003633, -0.00000670, -0.00000541, -0.00000329, -0.00000593, -0.00000565, 0.00000907, -0.00001120, 0.00000242, -0.00000675, 0.00000248, -0.00000402, -0.00000822, 0.00000184, -0.00000455, -0.00000259, 0.00000371, -0.00000850, -0.00001830, 0.00001310, -0.00000759, 0.00000741, -0.00001148, 0.00001192, 0.00000120, 0.00000062, 0.00000220, -0.00000082, 0.00000114, -0.00000223, 0.00000247, -0.00000266, 0.00000492, -0.00000026, 0.00000477, -0.00000046, -0.00000039, -0.00000064, 0.00000251, -0.00000453, 0.00000398, -0.00000354, 0.00000126, -0.00000056, 0.00000335, -0.00000441, -0.00000013, -0.00000191, -0.00000099, 0.00000640, -0.00000022, -0.00000580, },
		{ -0.00000601, -0.00000070, 0.00000922, -0.00000485, 0.00000144, -0.00000366, -0.00000217, 0.00000335, 0.00000093, -0.00000009, -0.00000481, 0.00000436, -0.00000461, 0.00000608, -0.00000303, -0.00000152, -0.00000143, 0.00000621, -0.00000067, 0.00000707, -0.00000431, 0.00000112, -0.00000094, -0.00000050, 0.00000224, 0.00000173, -0.00000178, 0.00000583, 0.00000440, -0.00000720, 0.00000457, -0.00000323, 0.00001044, -0.00001709, -0.00000746, 0.00000286, -0.00000280, -0.00000385, 0.00000120, -0.00000349, -0.00000856, 0.00000255, -0.00000291, -0.00000088, -0.00000908, 0.00000298, 0.00000140, -0.00000709, -0.00000303, -0.00000624, -0.00000422, -0.00004300, -0.00002326, 0.00006903, 0.00004326, 0.00004733, 0.00005649, -0.00003241, -0.00003816, -0.00000299, -0.00001176, -0.00000384, -0.00000520, -0.00000387, -0.00000503, -0.00000366, -0.00000693, -0.00000877, -0.00000753, -0.00003417, -0.00006436, 0.00007849, 0.00006468, 0.00007778, 0.00006744, -0.00007134, -0.00002467, -0.00000875, -0.00000347, -0.00000426, -0.00000609, -0.00000456, 0.00000898, -0.00001280, 0.00000497, -0.00000946, 0.00000540, -0.00000734, -0.00000555, 0.00000014, -0.00000363, -0.00000312, 0.00000553, -0.00001430, -0.00001215, 0.00001099, -0.00000703, 0.00000686, -0.00001203, 0.00001559, -0.00000304, 0.00000382, -0.00000024, 0.00000065, -0.00000046, -0.00000070, 0.00000121, -0.00000104, 0.00000382, 0.00000051, 0.00000406, -0.00000042, -0.00000062, 0.00000067, 0.00000061, -0.00000332, 0.00000331, -0.00000328, 0.00000160, -0.00000063, 0.00000311, -0.00000530, 0.00000132, -0.00000356, 0.00000137, 0.00000540, -0.00000109, -0.00000620, },
		{ -0.00000588, 0.00000120, 0.00000861, -0.00000629, 0.00000313, -0.00000609, 0.00000060, 0.00000203, 0.00000219, -0.00000247, -0.00000265, 0.00000266, -0.00000316, 0.00000611, -0.00000402, -0.00000155, -0.00000055, 0.00000596, -0.00000041, 0.00000724, -0.00000584, 0.00000254, -0.00000230, 0.00000079, 0.00000195, 0.00000187, -0.00000263, 0.00000873, -0.00000021, -0.00000440, 0.00000295, -0.00000056, 0.00000786, -0.00001886, -0.00000387, 0.00000164, -0.00000288, -0.00000369, 0.00000194, -0.00000497, -0.00000744, 0.00000245, -0.00000253, -0.00000177, -0.00000905, 0.00000433, -0.00000009, -0.00000621, -0.00000404, -0.00000500, -0.00000769, -0.00004716, -0.00000596, 0.00007009, 0.00003532, 0.00005102, 0.00004645, -0.00003729, -0.00003223, -0.00000275, -0.00001188, -0.00000293, -0.00000559, -0.00000370, -0.00000499, -0.00000386, -0.00000734, -0.00000894, -0.00000771, -0.00004250, -0.00005797, 0.00009456, 0.00005792, 0.00009479, 0.00004234, -0.00007752, -0.00001506, -0.00001062, -0.00000184, -0.00000504, -0.00000660, -0.00000271, 0.00000775, -0.00001317, 0.00000649, -0.00001149, 0.00000791, -0.00001051, -0.00000283, -0.00000169, -0.00000265, -0.00000328, 0.00000641, -0.00001954, -0.00000503, 0.00000688, -0.00000444, 0.00000428, -0.00001022, 0.00001714, -0.00000608, 0.00000656, -0.00000264, 0.00000212, -0.00000194, 0.00000073, 0.00000009, 0.00000050, 0.00000258, 0.00000149, 0.00000309, -0.00000018, -0.00000073, 0.00000175, -0.00000143, -0.00000155, 0.00000193, -0.00000212, 0.00000104, 0.00000005, 0.00000196, -0.00000508, 0.00000179, -0.00000444, 0.00000344, 0.00000417, -0.00000189, -0.00000639, },
		{ -0.00000586, 0.00000354, 0.00000690, -0.00000653, 0.00000380, -0.00000777, 0.00000315, 0.00000072, 0.00000323, -0.00000476, -0.00000022, 0.00000041, -0.00000073, 0.00000502, -0.00000434, -0.00000191, 0.00000082, 0.00000505, 0.00000067, 0.00000639, -0.00000648, 0.00000333, -0.00000315, 0.00000176, 0.00000206, 0.00000117, -0.00000216, 0.00001023, -0.00000419, -0.00000145, 0.00000121, 0.00000246, 0.00000395, -0.00001885, -0.00000120, 0.00000051, -0.00000299, -0.00000326, 0.00000233, -0.00000630, -0.00000632, 0.00000238, -0.00000224, -0.00000280, -0.00000865, 0.00000538, -0.00000140, -0.00000539, -0.00000517, -0.00000333, -0.00001307, -0.00004748, 0.00001093, 0.00006704, 0.00002988, 0.00005370, 0.00003606, -0.00004034, -0.00002654, -0.00000323, -0.00001146, -0.00000246, -0.00000572, -0.00000379, -0.00000462, -0.00000459, -0.00000717, -0.00000967, -0.00000770, -0.00005231, -0.00004695, 0.00010720, 0.00005373, 0.00011004, 0.00001114, -0.00007508, -0.00000916, -0.00001109, -0.00000130, -0.00000508, -0.00000777, 0.00000015, 0.00000506, -0.00001189, 0.00000645, -0.00001220, 0.00000941, -0.00001298, -0.00000056, -0.00000320, -0.00000199, -0.00000271, 0.00000576, -0.00002319, 0.00000179, 0.00000184, -0.00000052, 0.00000011, -0.00000623, 0.00001635, -0.00000741, 0.00000828, -0.00000456, 0.00000329, -0.00000308, 0.00000187, -0.00000068, 0.00000169, 0.00000151, 0.00000239, 0.00000211, 0.00000007, -0.00000054, 0.00000224, -0.00000310, 0.00000023, 0.00000031, -0.00000049, -0.00000016, 0.00000129, 0.00000009, -0.00000383, 0.00000117, -0.00000428, 0.00000486, 0.00000296, -0.00000270, -0.00000633, },
		{ -0.00000583, 0.00000609, 0.00000424, -0.00000553, 0.00000321, -0.00000828, 0.00000503, -0.00000020, 0.00000365, -0.00000644, 0.00000191, -0.00000180, 0.00000216, 0.00000308, -0.00000411, -0.00000247, 0.00000248, 0.00000367, 0.00000237, 0.00000460, -0.00000619, 0.00000341, -0.00000341, 0.00000236, 0.00000249, -0.00000025, -0.00000032, 0.00000994, -0.00000690, 0.00000097, 0.00000002, 0.00000500, -0.00000049, -0.00001748, 0.00000059, -0.00000051, -0.00000301, -0.00000272, 0.00000249, -0.00000756, -0.00000505, 0.00000220, -0.00000193, -0.00000407, -0.00000777, 0.00000593, -0.00000226, -0.00000504, -0.00000587, -0.00000200, -0.00001955, -0.00004380, 0.00002590, 0.00006108, 0.00002687, 0.00005499, 0.00002582, -0.00004177, -0.00002125, -0.00000426, -0.00001050, -0.00000247, -0.00000551, -0.00000421, -0.00000387, -0.00000588, -0.00000632, -0.00001111, -0.00000748, -0.00006348, -0.00003082, 0.00011515, 0.00005416, 0.00012005, -0.00002276, -0.00006500, -0.00000761, -0.00000951, -0.00000227, -0.00000412, -0.00000970, 0.00000402, 0.00000090, -0.00000880, 0.00000456, -0.00001120, 0.00000942, -0.00001430, 0.00000086, -0.00000406, -0.00000190, -0.00000124, 0.00000323, -0.00002440, 0.00000708, -0.00000289, 0.00000367, -0.00000476, -0.00000072, 0.00001347, -0.00000691, 0.00000870, -0.00000572, 0.00000399, -0.00000377, 0.00000262, -0.00000099, 0.00000235, 0.00000083, 0.00000298, 0.00000131, 0.00000023, 0.00000003, 0.00000194, -0.00000399, 0.00000152, -0.00000099, 0.00000108, -0.00000152, 0.00000268, -0.00000210, -0.00000190, -0.00000034, -0.00000309, 0.00000546, 0.00000195, -0.00000357, -0.00000607, },
		{ -0.00000559, 0.00000849, 0.00000103, -0.00000354, 0.00000144, -0.00000746, 0.00000596, -0.00000051, 0.00000321, -0.00000714, 0.00000323, -0.00000339, 0.00000493, 0.00000071, -0.00000354, -0.00000300, 0.00000413, 0.00000219, 0.00000434, 0.00000216, -0.00000512, 0.00000286, -0.00000315, 0.00000271, 0.00000299, -0.00000204, 0.00000256, 0.00000786, -0.00000800, 0.00000246, -0.00000019, 0.00000638, -0.00000464, -0.00001533, 0.00000169, -0.00000147, -0.00000285, -0.00000222, 0.00000253, -0.00000884, -0.00000352, 0.00000178, -0.00000150, -0.00000562, -0.00000631, 0.00000587, -0.00000251, -0.00000546, -0.00000566, -0.00000184, -0.00002595, -0.00003667, 0.00003803, 0.00005335, 0.00002604, 0.00005468, 0.00001614, -0.00004186, -0.00001651, -0.00000566, -0.00000908, -0.00000296, -0.00000495, -0.00000491, -0.00000283, -0.00000762, -0.00000489, -0.00001319, -0.00000731, -0.00007546, -0.00000968, 0.00011777, 0.00006059, 0.00012144, -0.00005511, -0.00004963, -0.00000990, -0.00000596, -0.00000468, -0.00000228, -0.00001214, 0.00000852, -0.00000438, -0.00000418, 0.00000090, -0.00000842, 0.00000779, -0.00001428, 0.00000126, -0.00000412, -0.00000239, 0.00000100, -0.00000113, -0.00002276, 0.00000989, -0.00000612, 0.00000697, -0.00000923, 0.00000533, 0.00000916, -0.00000486, 0.00000782, -0.00000602, 0.00000415, -0.00000401, 0.00000302, -0.00000089, 0.00000249, 0.00000061, 0.00000318, 0.00000077, 0.00000028, 0.00000091, 0.00000084, -0.00000390, 0.00000195, -0.00000152, 0.00000205, -0.00000254, 0.00000374, -0.00000406, 0.00000017, -0.00000231, -0.00000112, 0.00000528, 0.00000115, -0.00000441, -0.00000572, },
		{ -0.00000490, 0.00001031, -0.00000218, -0.00000105, -0.00000119, -0.00000545, 0.00000591, -0.00000019, 0.00000191, -0.00000673, 0.00000344, -0.00000387, 0.00000699, -0.00000162, -0.00000292, -0.00000324, 0.00000544, 0.00000097, 0.00000615, -0.00000053, -0.00000357, 0.00000190, -0.00000252, 0.00000298, 0.00000323, -0.00000365, 0.00000590, 0.00000441, -0.00000758, 0.00000299, 0.00000065, 0.00000629, -0.00000790, -0.00001291, 0.00000230, -0.00000240, -0.00000246, -0.00000184, 0.00000250, -0.00001014, -0.00000171, 0.00000107, -0.00000094, -0.00000742, -0.00000433, 0.00000523, -0.00000216, -0.00000671, -0.00000428, -0.00000349, -0.00003101, -0.00002712, 0.00004693, 0.00004480, 0.00002703, 0.00005269, 0.00000738, -0.00004087, -0.00001240, -0.00000722, -0.00000737, -0.00000381, -0.00000415, -0.00000574, -0.00000173, -0.00000952, -0.00000322, -0.00001556, -0.00000775, -0.00008720, 0.00001564, 0.00011535, 0.00007335, 0.00011169, -0.00008152, -0.00003225, -0.00001452, -0.00000127, -0.00000790, -0.00000008, -0.00001450, 0.00001295, -0.00001003, 0.00000133, -0.00000399, -0.00000421, 0.00000472, -0.00001304, 0.00000075, -0.00000352, -0.00000324, 0.00000356, -0.00000678, -0.00001848, 0.00000983, -0.00000708, 0.00000846, -0.00001224, 0.00001082, 0.00000433, -0.00000182, 0.00000588, -0.00000555, 0.00000387, -0.00000389, 0.00000319, -0.00000055, 0.00000224, 0.00000076, 0.00000303, 0.00000048, 0.00000032, 0.00000190, -0.00000084, -0.00000288, 0.00000143, -0.00000103, 0.00000205, -0.00000278, 0.00000402, -0.00000528, 0.00000180, -0.00000417, 0.00000121, 0.00000452, 0.00000049, -0.00000512, -0.00000546, },
		{ -0.00000356, 0.00001117, -0.00000484, 0.00000134, -0.00000412, -0.00000263, 0.00000507, 0.00000057, -0.00000001, -0.00000535, 0.00000254, -0.00000304, 0.00000794, -0.00000352, -0.00000251, -0.00000300, 0.00000616, 0.00000034, 0.00000737, -0.00000301, -0.00000190, 0.00000080, -0.00000176, 0.00000337, 0.00000290, -0.00000453, 0.00000892, 0.00000026, -0.00000605, 0.00000285, 0.00000226, 0.00000479, -0.00000998, -0.00001060, 0.00000256, -0.00000325, -0.00000191, -0.00000153, 0.00000231, -0.00001131, 0.00000023, 0.00000017, -0.00000038, -0.00000927, -0.00000200, 0.00000421, -0.00000147, -0.00000852, -0.00000192, -0.00000715, -0.00003370, -0.00001637, 0.00005266, 0.00003617, 0.00002944, 0.00004904, -0.00000016, -0.00003909, -0.00000899, -0.00000871, -0.00000560, -0.00000481, -0.00000332, -0.00000648, -0.00000090, -0.00001113, -0.00000185, -0.00001763, -0.00000969, -0.00009717, 0.00004353, 0.00010924, 0.00009126, 0.00008982, -0.00009836, -0.00001641, -0.00001940, 0.00000324, -0.00001093, 0.00000163, -0.00001592, 0.00001638, -0.00001506, 0.00000674, -0.00000926, 0.00000073, 0.00000072, -0.00001097, -0.00000031, -0.00000263, -0.00000398, 0.00000571, -0.00001275, -0.00001234, 0.00000718, -0.00000559, 0.00000766, -0.00001307, 0.00001482, -0.00000009, 0.00000153, 0.00000331, -0.00000452, 0.00000326, -0.00000355, 0.00000331, -0.00000021, 0.00000183, 0.00000113, 0.00000265, 0.00000032, 0.00000051, 0.00000267, -0.00000272, -0.00000127, 0.00000015, 0.00000042, 0.00000097, -0.00000205, 0.00000329, -0.00000542, 0.00000252, -0.00000539, 0.00000340, 0.00000349, -0.00000017, -0.00000554, -0.00000543, },
		{ -0.00000150, 0.00001081, -0.00000647, 0.00000304, -0.00000671, 0.00000048, 0.00000380, 0.00000142, -0.00000214, -0.00000341, 0.00000082, -0.00000099, 0.00000760, -0.00000473, -0.00000244, -0.00000220, 0.00000618, 0.00000050, 0.00000771, -0.00000492, -0.00000045, -0.00000017, -0.00000103, 0.00000399, 0.00000181, -0.00000422, 0.00001089, -0.00000378, -0.00000399, 0.00000253, 0.00000405, 0.00000229, -0.00001090, -0.00000854, 0.00000251, -0.00000387, -0.00000137, -0.00000115, 0.00000177, -0.00001210, 0.00000204, -0.00000069, -0.00000009, -0.00001087, 0.00000032, 0.00000319, -0.00000094, -0.00001031, 0.00000083, -0.00001245, -0.00003354, -0.00000552, 0.00005547, 0.00002809, 0.00003276, 0.00004391, -0.00000635, -0.00003678, -0.00000632, -0.00000992, -0.00000401, -0.00000570, -0.00000270, -0.00000684, -0.00000072, -0.00001194, -0.00000141, -0.00001870, -0.00001415, -0.00010351, 0.00007181, 0.00010172, 0.00011160, 0.00005685, -0.00010345, -0.00000509, -0.00002246, 0.00000615, -0.00001267, 0.00000197, -0.00001559, 0.00001785, -0.00001843, 0.00001100, -0.00001390, 0.00000552, -0.00000349, -0.00000866, -0.00000140, -0.00000192, -0.00000410, 0.00000666, -0.00001786, -0.00000560, 0.00000286, -0.00000210, 0.00000466, -0.00001149, 0.00001674, -0.00000339, 0.00000451, 0.00000056, -0.00000317, 0.00000248, -0.00000311, 0.00000351, -0.00000004, 0.00000149, 0.00000148, 0.00000221, 0.00000017, 0.00000103, 0.00000289, -0.00000430, 0.00000045, -0.00000143, 0.00000245, -0.00000096, -0.00000044, 0.00000158, -0.00000438, 0.00000208, -0.00000558, 0.00000503, 0.00000248, -0.00000097, -0.00000560, -0.00000570, },
		{ 0.00000117, 0.00000921, -0.00000682, 0.00000362, -0.00000842, 0.00000333, 0.00000249, 0.00000199, -0.00000398, -0.00000145, -0.00000119, 0.00000191, 0.00000608, -0.00000519, -0.00000271, -0.00000092, 0.00000556, 0.00000143, 0.00000705, -0.00000602, 0.00000056, -0.00000083, -0.00000044, 0.00000482, -0.00000004, -0.00000254, 0.00001128, -0.00000702, -0.00000199, 0.00000256, 0.00000537, -0.00000056, -0.00001097, -0.00000672, 0.00000209, -0.00000410, -0.00000106, -0.00000050, 0.00000067, -0.00001219, 0.00000337, -0.00000118, -0.00000037, -0.00001181, 0.00000223, 0.00000263, -0.00000113, -0.00001134, 0.00000302, -0.00001845, -0.00003066, 0.00000466, 0.00005569, 0.00002110, 0.00003643, 0.00003764, -0.00001116, -0.00003410, -0.00000441, -0.00001067, -0.00000285, -0.00000624, -0.00000253, -0.00000655, -0.00000150, -0.00001152, -0.00000247, -0.00001812, -0.00002208, -0.00010435, 0.00009806, 0.00009561, 0.00013036, 0.00001584, -0.00009663, -0.00000013, -0.00002219, 0.00000636, -0.00001228, 0.00000031, -0.00001291, 0.00001669, -0.00001930, 0.00001317, -0.00001698, 0.00000931, -0.00000720, -0.00000667, -0.00000206, -0.00000180, -0.00000318, 0.00000574, -0.00002097, 0.00000031, -0.00000183, 0.00000239, 0.00000015, -0.00000780, 0.00001648, -0.00000514, 0.00000666, -0.00000199, -0.00000171, 0.00000161, -0.00000263, 0.00000384, -0.00000018, 0.00000139, 0.00000166, 0.00000187, -0.00000007, 0.00000196, 0.00000229, -0.00000514, 0.00000173, -0.00000275, 0.00000448, -0.00000328, 0.00000168, -0.00000080, -0.00000240, 0.00000052, -0.00000459, 0.00000581, 0.00000170, -0.00000196, -0.00000528, -0.00000623, },
		{ 0.00000418, 0.00000658, -0.00000592, 0.00000291, -0.00000887, 0.00000551, 0.00000145, 0.00000200, -0.00000512, 0.00000000, -0.00000287, 0.00000507, 0.00000369, -0.00000502, -0.00000319, 0.00000064, 0.00000452, 0.00000292, 0.00000549, -0.00000626, 0.00000103, -0.00000112, 0.00000003, 0.00000570, -0.00000237, 0.00000037, 0.00000991, -0.00000899, -0.00000045, 0.00000330, 0.00000566, -0.00000310, -0.00001060, -0.00000503, 0.00000126, -0.00000380, -0.00000114, 0.00000054, -0.00000115, -0.00001134, 0.00000394, -0.00000106, -0.00000152, -0.00001177, 0.00000337, 0.00000285, -0.00000253, -0.00001089, 0.00000363, -0.00002390, -0.00002576, 0.00001384, 0.00005353, 0.00001571, 0.00003979, 0.00003072, -0.00001475, -0.00003118, -0.00000325, -0.00001082, -0.00000230, -0.00000622, -0.00000298, -0.00000548, -0.00000340, -0.00000964, -0.00000538, -0.00001552, -0.00003404, -0.00009819, 0.00012008, 0.00009371, 0.00014293, -0.00002837, -0.00007982, -0.00000178, -0.00001805, 0.00000350, -0.00000947, -0.00000350, -0.00000780, 0.00001271, -0.00001731, 0.00001275, -0.00001787, 0.00001145, -0.00000987, -0.00000540, -0.00000199, -0.00000247, -0.00000111, 0.00000266, -0.00002128, 0.00000412, -0.00000550, 0.00000662, -0.00000481, -0.00000278, 0.00001438, -0.00000533, 0.00000773, -0.00000408, -0.00000027, 0.00000070, -0.00000205, 0.00000422, -0.00000059, 0.00000159, 0.00000159, 0.00000168, -0.00000039, 0.00000327, 0.00000082, -0.00000493, 0.00000211, -0.00000323, 0.00000589, -0.00000536, 0.00000378, -0.00000332, 0.00000006, -0.00000181, -0.00000258, 0.00000572, 0.00000121, -0.00000308, -0.00000470, -0.00000683, },
		{ 0.00000712, 0.00000337, -0.00000407, 0.00000105, -0.00000798, 0.00000676, 0.00000085, 0.00000135, -0.00000534, 0.00000054, -0.00000368, 0.00000788, 0.00000091, -0.00000445, -0.00000368, 0.00000220, 0.00000339, 0.00000464, 0.00000330, -0.00000579, 0.00000105, -0.00000110, 0.00000049, 0.00000634, -0.00000473, 0.00000407, 0.00000696, -0.00000959, 0.00000049, 0.00000479, 0.00000464, -0.00000485, -0.00001016, -0.00000337, 0.00000003, -0.00000295, -0.00000166, 0.00000199, -0.00000369, -0.00000949, 0.00000362, -0.00000021, -0.00000361, -0.00001061, 0.00000363, 0.00000398, -0.00000532, -0.00000856, 0.00000188, -0.00002751, -0.00001988, 0.00002203, 0.00004915, 0.00001240, 0.00004217, 0.00002373, -0.00001737, -0.00002805, -0.00000281, -0.00001035, -0.00000243, -0.00000560, -0.00000406, -0.00000368, -0.00000633, -0.00000639, -0.00001006, -0.00001105, -0.00004995, -0.00008427, 0.00013647, 0.00009798, 0.00014493, -0.00007003, -0.00005669, -0.00000872, -0.00001067, -0.00000198, -0.00000467, -0.00000897, -0.00000076, 0.00000632, -0.00001268, 0.00000974, -0.00001642, 0.00001172, -0.00001131, -0.00000493, -0.00000121, -0.00000378, 0.00000183, -0.00000234, -0.00001859, 0.00000506, -0.00000703, 0.00000939, -0.00000905, 0.00000253, 0.00001117, -0.00000426, 0.00000776, -0.00000563, 0.00000108, -0.00000029, -0.00000127, 0.00000449, -0.00000114, 0.00000199, 0.00000131, 0.00000159, -0.00000067, 0.00000474, -0.00000138, -0.00000366, 0.00000141, -0.00000253, 0.00000613, -0.00000661, 0.00000526, -0.00000538, 0.00000236, -0.00000434, 0.00000007, 0.00000491, 0.00000093, -0.00000420, -0.00000406, -0.00000722, },
		{ 0.00000951, 0.00000013, -0.00000180, -0.00000158, -0.00000592, 0.00000711, 0.00000063, 0.00000013, -0.00000465, 0.00000003, -0.00000323, 0.00000978, -0.00000180, -0.00000377, -0.00000394, 0.00000347, 0.00000251, 0.00000618, 0.00000086, -0.00000487, 0.00000077, -0.00000091, 0.00000110, 0.00000641, -0.00000656, 0.00000786, 0.00000292, -0.00000901, 0.00000102, 0.00000678, 0.00000241, -0.00000558, -0.00000987, -0.00000175, -0.00000144, -0.00000170, -0.00000248, 0.00000364, -0.00000670, -0.00000681, 0.00000254, 0.00000122, -0.00000649, -0.00000843, 0.00000312, 0.00000575, -0.00000925, -0.00000446, -0.00000252, -0.00002834, -0.00001408, 0.00002943, 0.00004270, 0.00001150, 0.00004297, 0.00001727, -0.00001935, -0.00002475, -0.00000301, -0.00000932, -0.00000315, -0.00000447, -0.00000561, -0.00000143, -0.00000989, -0.00000227, -0.00001599, -0.00000545, -0.00006891, -0.00006285, 0.00014698, 0.00010896, 0.00013322, -0.00010343, -0.00003190, -0.00001840, -0.00000163, -0.00000885, 0.00000108, -0.00001507, 0.00000712, -0.00000151, -0.00000618, 0.00000470, -0.00001298, 0.00001032, -0.00001166, -0.00000502, -0.00000004, -0.00000530, 0.00000494, -0.00000843, -0.00001342, 0.00000306, -0.00000587, 0.00000986, -0.00001156, 0.00000710, 0.00000768, -0.00000248, 0.00000696, -0.00000664, 0.00000235, -0.00000141, -0.00000017, 0.00000443, -0.00000161, 0.00000244, 0.00000094, 0.00000149, -0.00000070, 0.00000604, -0.00000394, -0.00000158, -0.00000028, -0.00000058, 0.00000495, -0.00000662, 0.00000567, -0.00000644, 0.00000390, -0.00000644, 0.00000283, 0.00000369, 0.00000069, -0.00000510, -0.00000359, -0.00000712, },
		{ 0.00001095, -0.00000257, 0.00000030, -0.00000438, -0.00000311, 0.00000672, 0.00000063, -0.00000135, -0.00000335, -0.00000138, -0.00000145, 0.00001042, -0.00000403, -0.00000320, -0.00000382, 0.00000427, 0.00000215, 0.00000717, -0.00000142, -0.00000379, 0.00000041, -0.00000071, 0.00000200, 0.00000567, -0.00000733, 0.00001100, -0.00000150, -0.00000770, 0.00000152, 0.00000873, -0.00000060, -0.00000541, -0.00000975, -0.00000030, -0.00000290, -0.00000033, -0.00000331, 0.00000513, -0.00000976, -0.00000372, 0.00000104, 0.00000285, -0.00000975, -0.00000564, 0.00000229, 0.00000760, -0.00001363, 0.00000077, -0.00000920, -0.00002604, -0.00000914, 0.00003617, 0.00003450, 0.00001312, 0.00004178, 0.00001181, -0.00002098, -0.00002131, -0.00000371, -0.00000792, -0.00000425, -0.00000309, -0.00000727, 0.00000077, -0.00001342, 0.00000190, -0.00002217, -0.00000003, -0.00008920, -0.00003522, 0.00015270, 0.00012528, 0.00010676, -0.00012408, -0.00001017, -0.00002775, 0.00000694, -0.00001538, 0.00000632, -0.00002041, 0.00001446, -0.00000942, 0.00000097, -0.00000137, -0.00000832, 0.00000778, -0.00001139, -0.00000520, 0.00000097, -0.00000634, 0.00000728, -0.00001438, -0.00000693, -0.00000118, -0.00000221, 0.00000779, -0.00001178, 0.00001014, 0.00000468, -0.00000054, 0.00000564, -0.00000720, 0.00000353, -0.00000268, 0.00000134, 0.00000387, -0.00000176, 0.00000271, 0.00000068, 0.00000122, -0.00000025, 0.00000675, -0.00000631, 0.00000081, -0.00000255, 0.00000232, 0.00000240, -0.00000527, 0.00000480, -0.00000619, 0.00000422, -0.00000757, 0.00000516, 0.00000243, 0.00000028, -0.00000559, -0.00000349, -0.00000629, },
		{ 0.00001119, -0.00000430, 0.00000168, -0.00000676, -0.00000003, 0.00000589, 0.00000056, -0.00000268, -0.00000189, -0.00000326, 0.00000143, 0.00000969, -0.00000556, -0.00000288, -0.00000329, 0.00000453, 0.00000242, 0.00000738, -0.00000324, -0.00000282, 0.00000017, -0.00000059, 0.00000321, 0.00000402, -0.00000671, 0.00001284, -0.00000558, -0.00000614, 0.00000245, 0.00000997, -0.00000373, -0.00000470, -0.00000963, 0.00000076, -0.00000398, 0.00000075, -0.00000375, 0.00000600, -0.00001230, -0.00000077, -0.00000035, 0.00000414, -0.00001279, -0.00000278, 0.00000172, 0.00000872, -0.00001745, 0.00000595, -0.00001707, -0.00002099, -0.00000533, 0.00004213, 0.00002508, 0.00001699, 0.00003851, 0.00000767, -0.00002247, -0.00001780, -0.00000467, -0.00000641, -0.00000540, -0.00000183, -0.00000858, 0.00000234, -0.00001611, 0.00000509, -0.00002738, 0.00000354, -0.00010850, -0.00000354, 0.00015587, 0.00014369, 0.00006712, -0.00012959, 0.00000475, -0.00003392, 0.00001292, -0.00001986, 0.00000960, -0.00002363, 0.00001982, -0.00001599, 0.00000744, -0.00000731, -0.00000338, 0.00000482, -0.00001106, -0.00000493, 0.00000124, -0.00000626, 0.00000789, -0.00001880, -0.00000066, -0.00000635, 0.00000310, 0.00000362, -0.00000973, 0.00001133, 0.00000268, 0.00000108, 0.00000408, -0.00000737, 0.00000457, -0.00000402, 0.00000322, 0.00000273, -0.00000143, 0.00000263, 0.00000067, 0.00000064, 0.00000085, 0.00000652, -0.00000796, 0.00000290, -0.00000476, 0.00000560, -0.00000107, -0.00000278, 0.00000276, -0.00000465, 0.00000317, -0.00000739, 0.00000666, 0.00000142, -0.00000044, -0.00000557, -0.00000384, -0.00000464, },
		{ 0.00001021, -0.00000486, 0.00000198, -0.00000823, 0.00000282, 0.00000493, 0.00000018, -0.00000347, -0.00000081, -0.00000503, 0.00000496, 0.00000775, -0.00000631, -0.00000281, -0.00000242, 0.00000434, 0.00000328, 0.00000674, -0.00000442, -0.00000211, 0.00000012, -0.00000057, 0.00000460, 0.00000160, -0.00000463, 0.00001299, -0.00000878, -0.00000473, 0.00000411, 0.00000995, -0.00000628, -0.00000393, -0.00000927, 0.00000120, -0.00000438, 0.00000120, -0.00000343, 0.00000581, -0.00001378, 0.00000145, -0.00000109, 0.00000450, -0.00001499, -0.00000046, 0.00000201, 0.00000831, -0.00001960, 0.00000972, -0.00002459, -0.00001424, -0.00000237, 0.00004690, 0.00001530, 0.00002250, 0.00003338, 0.00000487, -0.00002387, -0.00001438, -0.00000567, -0.00000511, -0.00000625, -0.00000108, -0.00000912, 0.00000271, -0.00001724, 0.00000632, -0.00003038, 0.00000355, -0.00012430, 0.00002964, 0.00015935, 0.00015945, 0.00001841, -0.00012019, 0.00001085, -0.00003500, 0.00001482, -0.00002105, 0.00000988, -0.00002372, 0.00002213, -0.00002004, 0.00001210, -0.00001205, 0.00000095, 0.00000212, -0.00001113, -0.00000379, 0.00000038, -0.00000463, 0.00000609, -0.00002050, 0.00000379, -0.00001085, 0.00000870, -0.00000168, -0.00000597, 0.00001081, 0.00000183, 0.00000212, 0.00000246, -0.00000715, 0.00000534, -0.00000526, 0.00000532, 0.00000106, -0.00000059, 0.00000214, 0.00000096, -0.00000024, 0.00000264, 0.00000513, -0.00000845, 0.00000410, -0.00000624, 0.00000848, -0.00000480, 0.00000033, -0.00000002, -0.00000214, 0.00000092, -0.00000587, 0.00000711, 0.00000084, -0.00000152, -0.00000504, -0.00000460, -0.00000226, },
		{ 0.00000825, -0.00000433, 0.00000111, -0.00000848, 0.00000509, 0.00000403, -0.00000063, -0.00000348, -0.00000053, -0.00000608, 0.00000850, 0.00000496, -0.00000640, -0.00000289, -0.00000140, 0.00000392, 0.00000450, 0.00000537, -0.00000495, -0.00000171, 0.00000027, -0.00000056, 0.00000595, -0.00000122, -0.00000136, 0.00001140, -0.00001080, -0.00000365, 0.00000658, 0.00000836, -0.00000772, -0.00000352, -0.00000847, 0.00000090, -0.00000391, 0.00000081, -0.00000217, 0.00000431, -0.00001384, 0.00000253, -0.00000078, 0.00000352, -0.00001587, 0.00000090, 0.00000354, 0.00000584, -0.00001923, 0.00001079, -0.00003011, -0.00000725, 0.00000046, 0.00004978, 0.00000627, 0.00002871, 0.00002698, 0.00000320, -0.00002514, -0.00001120, -0.00000646, -0.00000428, -0.00000650, -0.00000114, -0.00000856, 0.00000153, -0.00001631, 0.00000491, -0.00003029, -0.00000136, -0.00013437, 0.00006201, 0.00016583, 0.00016723, -0.00003327, -0.00009873, 0.00000816, -0.00003051, 0.00001218, -0.00001854, 0.00000682, -0.00002036, 0.00002094, -0.00002098, 0.00001423, -0.00001491, 0.00000407, 0.00000006, -0.00001181, -0.00000169, -0.00000166, -0.00000146, 0.00000174, -0.00001882, 0.00000521, -0.00001318, 0.00001310, -0.00000681, -0.00000148, 0.00000913, 0.00000193, 0.00000257, 0.00000087, -0.00000651, 0.00000567, -0.00000614, 0.00000734, -0.00000090, 0.00000065, 0.00000130, 0.00000149, -0.00000132, 0.00000496, 0.00000261, -0.00000757, 0.00000402, -0.00000641, 0.00001023, -0.00000799, 0.00000335, -0.00000291, 0.00000075, -0.00000207, -0.00000326, 0.00000659, 0.00000068, -0.00000285, -0.00000414, -0.00000557, 0.00000057, },
		{ 0.00000574, -0.00000303, -0.00000072, -0.00000749, 0.00000659, 0.00000327, -0.00000181, -0.00000268, -0.00000128, -0.00000595, 0.00001144, 0.00000179, -0.00000603, -0.00000299, -0.00000044, 0.00000355, 0.00000577, 0.00000355, -0.00000497, -0.00000153, 0.00000052, -0.00000041, 0.00000691, -0.00000393, 0.00000255, 0.00000836, -0.00001168, -0.00000278, 0.00000959, 0.00000528, -0.00000782, -0.00000372, -0.00000719, -0.00000010, -0.00000262, -0.00000039, -0.00000002, 0.00000155, -0.00001241, 0.00000234, 0.00000070, 0.00000106, -0.00001522, 0.00000118, 0.00000631, 0.00000122, -0.00001598, 0.00000837, -0.00003223, -0.00000151, 0.00000405, 0.00004998, -0.00000086, 0.00003450, 0.00002010, 0.00000226, -0.00002610, -0.00000844, -0.00000690, -0.00000405, -0.00000602, -0.00000209, -0.00000688, -0.00000123, -0.00001328, 0.00000067, -0.00002683, -0.00001180, -0.00013723, 0.00009205, 0.00017697, 0.00016229, -0.00008089, -0.00007004, -0.00000130, -0.00002142, 0.00000570, -0.00001286, 0.00000094, -0.00001401, 0.00001653, -0.00001887, 0.00001377, -0.00001573, 0.00000582, -0.00000134, -0.00001292, 0.00000109, -0.00000448, 0.00000270, -0.00000463, -0.00001389, 0.00000309, -0.00001235, 0.00001506, -0.00001050, 0.00000263, 0.00000711, 0.00000253, 0.00000260, -0.00000074, -0.00000535, 0.00000536, -0.00000636, 0.00000891, -0.00000283, 0.00000204, 0.00000030, 0.00000207, -0.00000231, 0.00000747, -0.00000073, -0.00000542, 0.00000257, -0.00000499, 0.00001031, -0.00000995, 0.00000561, -0.00000524, 0.00000332, -0.00000512, -0.00000008, 0.00000534, 0.00000077, -0.00000422, -0.00000313, -0.00000645, 0.00000344, },
		{ 0.00000319, -0.00000148, -0.00000310, -0.00000550, 0.00000735, 0.00000256, -0.00000312, -0.00000128, -0.00000301, -0.00000439, 0.00001327, -0.00000131, -0.00000540, -0.00000297, 0.00000029, 0.00000348, 0.00000674, 0.00000159, -0.00000468, -0.00000142, 0.00000073, 0.00000005, 0.00000718, -0.00000595, 0.00000639, 0.00000443, -0.00001169, -0.00000179, 0.00001256, 0.00000116, -0.00000668, -0.00000452, -0.00000556, -0.00000153, -0.00000081, -0.00000209, 0.00000267, -0.00000212, -0.00000976, 0.00000108, 0.00000312, -0.00000265, -0.00001322, 0.00000061, 0.00000988, -0.00000505, -0.00001019, 0.00000245, -0.00003023, 0.00000186, 0.00000912, 0.00004690, -0.00000506, 0.00003878, 0.00001359, 0.00000157, -0.00002658, -0.00000619, -0.00000695, -0.00000440, -0.00000490, -0.00000378, -0.00000434, -0.00000517, -0.00000859, -0.00000591, -0.00002051, -0.00002746, -0.00013242, 0.00011951, 0.00019259, 0.00014162, -0.00011787, -0.00003996, -0.00001415, -0.00000991, -0.00000296, -0.00000540, -0.00000652, -0.00000584, 0.00000987, -0.00001444, 0.00001123, -0.00001486, 0.00000649, -0.00000241, -0.00001394, 0.00000388, -0.00000731, 0.00000685, -0.00001184, -0.00000665, -0.00000218, -0.00000814, 0.00001394, -0.00001187, 0.00000542, 0.00000554, 0.00000308, 0.00000246, -0.00000242, -0.00000364, 0.00000428, -0.00000568, 0.00000970, -0.00000436, 0.00000326, -0.00000060, 0.00000245, -0.00000290, 0.00000968, -0.00000440, -0.00000243, 0.00000001, -0.00000205, 0.00000850, -0.00001025, 0.00000660, -0.00000645, 0.00000494, -0.00000755, 0.00000308, 0.00000377, 0.00000087, -0.00000539, -0.00000225, -0.00000693, 0.00000590, },
		{ 0.00000110, -0.00000022, -0.00000546, -0.00000288, 0.00000752, 0.00000173, -0.00000421, 0.00000028, -0.00000537, -0.00000149, 0.00001372, -0.00000399, -0.00000469, -0.00000278, 0.00000069, 0.00000386, 0.00000714, -0.00000018, -0.00000432, -0.00000123, 0.00000079, 0.00000089, 0.00000655, -0.00000684, 0.00000948, 0.00000025, -0.00001124, -0.00000026, 0.00001479, -0.00000327, -0.00000475, -0.00000567, -0.00000390, -0.00000297, 0.00000106, -0.00000381, 0.00000532, -0.00000608, -0.00000647, -0.00000072, 0.00000591, -0.00000707, -0.00001036, -0.00000022, 0.00001340, -0.00001193, -0.00000291, -0.00000616, -0.00002428, 0.00000245, 0.00001593, 0.00004037, -0.00000570, 0.00004069, 0.00000819, 0.00000069, -0.00002644, -0.00000448, -0.00000672, -0.00000511, -0.00000341, -0.00000581, -0.00000148, -0.00000958, -0.00000316, -0.00001380, -0.00001254, -0.00004708, -0.00012058, 0.00014549, 0.00021038, 0.00010495, -0.00013934, -0.00001406, -0.00002653, 0.00000129, -0.00001160, 0.00000201, -0.00001387, 0.00000258, 0.00000237, -0.00000886, 0.00000751, -0.00001299, 0.00000663, -0.00000371, -0.00001419, 0.00000586, -0.00000921, 0.00000974, -0.00001829, 0.00000127, -0.00000938, -0.00000121, 0.00000983, -0.00001060, 0.00000633, 0.00000500, 0.00000313, 0.00000241, -0.00000421, -0.00000141, 0.00000241, -0.00000398, 0.00000944, -0.00000516, 0.00000401, -0.00000114, 0.00000239, -0.00000274, 0.00001109, -0.00000773, 0.00000078, -0.00000307, 0.00000192, 0.00000503, -0.00000882, 0.00000611, -0.00000625, 0.00000517, -0.00000881, 0.00000565, 0.00000229, 0.00000073, -0.00000611, -0.00000175, -0.00000678, 0.00000759, },
		{ -0.00000019, 0.00000031, -0.00000727, -0.00000009, 0.00000732, 0.00000059, -0.00000469, 0.00000148, -0.00000781, 0.00000239, 0.00001276, -0.00000603, -0.00000400, -0.00000244, 0.00000082, 0.00000470, 0.00000686, -0.00000155, -0.00000405, -0.00000085, 0.00000067, 0.00000209, 0.00000502, -0.00000634, 0.00001127, -0.00000358, -0.00001067, 0.00000216, 0.00001559, -0.00000717, -0.00000263, -0.00000678, -0.00000259, -0.00000393, 0.00000245, -0.00000497, 0.00000728, -0.00000958, -0.00000327, -0.00000236, 0.00000836, -0.00001144, -0.00000732, -0.00000059, 0.00001584, -0.00001808, 0.00000433, -0.00001590, -0.00001543, 0.00000060, 0.00002413, 0.00003079, -0.00000270, 0.00003978, 0.00000443, -0.00000070, -0.00002562, -0.00000322, -0.00000639, -0.00000587, -0.00000198, -0.00000767, 0.00000100, -0.00001354, 0.00000186, -0.00002163, -0.00000464, -0.00006876, -0.00010307, 0.00017209, 0.00022593, 0.00005510, -0.00014324, 0.00000352, -0.00003510, 0.00000963, -0.00001813, 0.00000759, -0.00001948, 0.00000968, -0.00000451, -0.00000337, 0.00000366, -0.00001095, 0.00000687, -0.00000581, -0.00001302, 0.00000629, -0.00000931, 0.00001023, -0.00002234, 0.00000795, -0.00001669, 0.00000704, 0.00000359, -0.00000706, 0.00000537, 0.00000571, 0.00000242, 0.00000256, -0.00000604, 0.00000116, -0.00000007, -0.00000136, 0.00000807, -0.00000506, 0.00000411, -0.00000114, 0.00000172, -0.00000160, 0.00001126, -0.00001010, 0.00000349, -0.00000591, 0.00000617, 0.00000049, -0.00000598, 0.00000429, -0.00000469, 0.00000392, -0.00000859, 0.00000722, 0.00000120, 0.00000017, -0.00000625, -0.00000174, -0.00000589, 0.00000829, },
		{ -0.00000058, -0.00000013, -0.00000814, 0.00000248, 0.00000694, -0.00000095, -0.00000430, 0.00000181, -0.00000969, 0.00000670, 0.00001060, -0.00000737, -0.00000334, -0.00000208, 0.00000083, 0.00000585, 0.00000592, -0.00000245, -0.00000393, -0.00000030, 0.00000042, 0.00000346, 0.00000281, -0.00000452, 0.00001151, -0.00000668, -0.00001017, 0.00000557, 0.00001453, -0.00000983, -0.00000092, -0.00000747, -0.00000196, -0.00000404, 0.00000292, -0.00000508, 0.00000796, -0.00001193, -0.00000088, -0.00000317, 0.00000972, -0.00001500, -0.00000476, 0.00000017, 0.00001626, -0.00002215, 0.00000983, -0.00002480, -0.00000540, -0.00000261, 0.00003275, 0.00001915, 0.00000341, 0.00003610, 0.00000248, -0.00000273, -0.00002415, -0.00000231, -0.00000620, -0.00000633, -0.00000102, -0.00000884, 0.00000247, -0.00001621, 0.00000533, -0.00002799, 0.00000141, -0.00009038, -0.00008147, 0.00020168, 0.00023355, -0.00000222, -0.00013070, 0.00001086, -0.00003782, 0.00001343, -0.00002111, 0.00001007, -0.00002219, 0.00001426, -0.00000961, 0.00000100, 0.00000050, -0.00000936, 0.00000761, -0.00000898, -0.00001010, 0.00000477, -0.00000719, 0.00000760, -0.00002274, 0.00001160, -0.00002214, 0.00001482, -0.00000336, -0.00000220, 0.00000304, 0.00000746, 0.00000101, 0.00000288, -0.00000771, 0.00000378, -0.00000284, 0.00000192, 0.00000573, -0.00000407, 0.00000350, -0.00000059, 0.00000045, 0.00000055, 0.00000996, -0.00001106, 0.00000508, -0.00000775, 0.00000980, -0.00000429, -0.00000235, 0.00000163, -0.00000216, 0.00000143, -0.00000694, 0.00000764, 0.00000065, -0.00000084, -0.00000581, -0.00000218, -0.00000436, 0.00000803, },
		{ -0.00000021, -0.00000152, -0.00000789, 0.00000456, 0.00000644, -0.00000283, -0.00000297, 0.00000097, -0.00001042, 0.00001078, 0.00000762, -0.00000808, -0.00000265, -0.00000186, 0.00000096, 0.00000704, 0.00000452, -0.00000295, -0.00000390, 0.00000034, 0.00000021, 0.00000472, 0.00000033, -0.00000173, 0.00001029, -0.00000894, -0.00000966, 0.00000974, 0.00001155, -0.00001084, -0.00000003, -0.00000751, -0.00000214, -0.00000313, 0.00000227, -0.00000392, 0.00000708, -0.00001270, 0.00000023, -0.00000267, 0.00000949, -0.00001719, -0.00000314, 0.00000243, 0.00001407, -0.00002320, 0.00001219, -0.00003096, 0.00000383, -0.00000566, 0.00004042, 0.00000690, 0.00001162, 0.00003014, 0.00000221, -0.00000537, -0.00002214, -0.00000164, -0.00000632, -0.00000625, -0.00000083, -0.00000899, 0.00000250, -0.00001702, 0.00000647, -0.00003184, 0.00000416, -0.00011015, -0.00005685, 0.00023592, 0.00022741, -0.00005931, -0.00010580, 0.00000859, -0.00003441, 0.00001234, -0.00002012, 0.00000910, -0.00002159, 0.00001585, -0.00001241, 0.00000376, -0.00000156, -0.00000846, 0.00000887, -0.00001308, -0.00000560, 0.00000143, -0.00000299, 0.00000190, -0.00001899, 0.00001103, -0.00002410, 0.00002034, -0.00000942, 0.00000272, 0.00000024, 0.00000970, -0.00000083, 0.00000318, -0.00000896, 0.00000604, -0.00000546, 0.00000542, 0.00000276, -0.00000239, 0.00000233, 0.00000039, -0.00000124, 0.00000351, 0.00000726, -0.00001041, 0.00000518, -0.00000801, 0.00001200, -0.00000840, 0.00000128, -0.00000121, 0.00000070, -0.00000175, -0.00000419, 0.00000702, 0.00000061, -0.00000221, -0.00000491, -0.00000290, -0.00000243, 0.00000706, },
		{ 0.00000055, -0.00000356, -0.00000658, 0.00000606, 0.00000577, -0.00000480, -0.00000089, -0.00000110, -0.00000963, 0.00001404, 0.00000424, -0.00000828, -0.00000190, -0.00000190, 0.00000144, 0.00000795, 0.00000295, -0.00000321, -0.00000384, 0.00000093, 0.00000020, 0.00000552, -0.00000188, 0.00000144, 0.00000797, -0.00001052, -0.00000883, 0.00001410, 0.00000703, -0.00001020, -0.00000009, -0.00000690, -0.00000302, -0.00000135, 0.00000061, -0.00000160, 0.00000467, -0.00001184, -0.00000004, -0.00000077, 0.00000752, -0.00001781, -0.00000250, 0.00000605, 0.00000931, -0.00002091, 0.00001066, -0.00003298, 0.00001051, -0.00000706, 0.00004568, -0.00000431, 0.00002058, 0.00002269, 0.00000327, -0.00000849, -0.00001967, -0.00000117, -0.00000675, -0.00000558, -0.00000144, -0.00000811, 0.00000110, -0.00001591, 0.00000507, -0.00003281, 0.00000293, -0.00012706, -0.00002919, 0.00027472, 0.00020303, -0.00010798, -0.00007468, -0.00000052, -0.00002630, 0.00000731, -0.00001590, 0.00000526, -0.00001818, 0.00001476, -0.00001305, 0.00000502, -0.00000269, -0.00000800, 0.00001023, -0.00001749, -0.00000020, -0.00000301, 0.00000244, -0.00000603, -0.00001154, 0.00000604, -0.00002166, 0.00002232, -0.00001321, 0.00000643, -0.00000203, 0.00001172, -0.00000265, 0.00000320, -0.00000949, 0.00000757, -0.00000745, 0.00000860, -0.00000038, -0.00000036, 0.00000087, 0.00000152, -0.00000299, 0.00000689, 0.00000348, -0.00000833, 0.00000375, -0.00000646, 0.00001220, -0.00001107, 0.00000415, -0.00000352, 0.00000317, -0.00000491, -0.00000089, 0.00000569, 0.00000088, -0.00000370, -0.00000379, -0.00000364, -0.00000045, 0.00000575, },
		{ 0.00000122, -0.00000580, -0.00000449, 0.00000703, 0.00000480, -0.00000648, 0.00000151, -0.00000413, -0.00000723, 0.00001604, 0.00000087, -0.00000809, -0.00000107, -0.00000227, 0.00000244, 0.00000830, 0.00000149, -0.00000342, -0.00000359, 0.00000131, 0.00000056, 0.00000560, -0.00000332, 0.00000434, 0.00000508, -0.00001171, -0.00000722, 0.00001786, 0.00000171, -0.00000828, -0.00000091, -0.00000590, -0.00000422, 0.00000089, -0.00000161, 0.00000141, 0.00000120, -0.00000969, -0.00000141, 0.00000223, 0.00000405, -0.00001704, -0.00000253, 0.00001044, 0.00000264, -0.00001579, 0.00000539, -0.00003035, 0.00001351, -0.00000568, 0.00004733, -0.00001298, 0.00002890, 0.00001471, 0.00000515, -0.00001188, -0.00001680, -0.00000098, -0.00000734, -0.00000452, -0.00000261, -0.00000652, -0.00000133, -0.00001335, 0.00000157, -0.00003128, -0.00000214, -0.00014105, 0.00000299, 0.00031552, 0.00015864, -0.00014118, -0.00004414, -0.00001237, -0.00001612, 0.00000033, -0.00001001, -0.00000008, -0.00001315, 0.00001194, -0.00001230, 0.00000542, -0.00000347, -0.00000733, 0.00001088, -0.00002123, 0.00000501, -0.00000740, 0.00000773, -0.00001460, -0.00000184, -0.00000254, -0.00001495, 0.00002024, -0.00001387, 0.00000799, -0.00000289, 0.00001281, -0.00000401, 0.00000269, -0.00000910, 0.00000805, -0.00000840, 0.00001099, -0.00000321, 0.00000161, -0.00000055, 0.00000247, -0.00000439, 0.00001013, -0.00000080, -0.00000527, 0.00000116, -0.00000327, 0.00001028, -0.00001186, 0.00000571, -0.00000474, 0.00000463, -0.00000737, 0.00000230, 0.00000407, 0.00000116, -0.00000504, -0.00000272, -0.00000414, 0.00000120, 0.00000456, },
		{ 0.00000134, -0.00000772, -0.00000199, 0.00000762, 0.00000334, -0.00000744, 0.00000366, -0.00000760, -0.00000344, 0.00001659, -0.00000221, -0.00000759, -0.00000026, -0.00000289, 0.00000396, 0.00000795, 0.00000036, -0.00000375, -0.00000304, 0.00000140, 0.00000132, 0.00000482, -0.00000363, 0.00000639, 0.00000217, -0.00001283, -0.00000442, 0.00002018, -0.00000345, -0.00000574, -0.00000206, -0.00000494, -0.00000521, 0.00000298, -0.00000376, 0.00000442, -0.00000263, -0.00000693, -0.00000326, 0.00000570, -0.00000030, -0.00001542, -0.00000256, 0.00001460, -0.00000470, -0.00000907, -0.00000252, -0.00002359, 0.00001265, -0.00000112, 0.00004479, -0.00001802, 0.00003539, 0.00000712, 0.00000735, -0.00001534, -0.00001357, -0.00000125, -0.00000778, -0.00000344, -0.00000388, -0.00000481, -0.00000403, -0.00001024, -0.00000306, -0.00002828, -0.00001014, -0.00015284, 0.00004241, 0.00035310, 0.00009613, -0.00015455, -0.00002015, -0.00002268, -0.00000689, -0.00000621, -0.00000443, -0.00000522, -0.00000799, 0.00000869, -0.00001120, 0.00000586, -0.00000473, -0.00000568, 0.00000995, -0.00002319, 0.00000882, -0.00001045, 0.00001132, -0.00002183, 0.00000803, -0.00001294, -0.00000510, 0.00001455, -0.00001132, 0.00000702, -0.00000185, 0.00001251, -0.00000456, 0.00000153, -0.00000772, 0.00000738, -0.00000808, 0.00001220, -0.00000530, 0.00000314, -0.00000159, 0.00000291, -0.00000502, 0.00001263, -0.00000489, -0.00000191, -0.00000195, 0.00000097, 0.00000652, -0.00001071, 0.00000578, -0.00000462, 0.00000470, -0.00000858, 0.00000483, 0.00000259, 0.00000120, -0.00000601, -0.00000189, -0.00000422, 0.00000228, 0.00000384, },
		{ 0.00000062, -0.00000887, 0.00000051, 0.00000796, 0.00000131, -0.00000734, 0.00000496, -0.00001081, 0.00000124, 0.00001575, -0.00000482, -0.00000679, 0.00000039, -0.00000357, 0.00000586, 0.00000689, -0.00000033, -0.00000426, -0.00000219, 0.00000121, 0.00000241, 0.00000329, -0.00000272, 0.00000723, -0.00000036, -0.00001405, -0.00000024, 0.00002042, -0.00000757, -0.00000331, -0.00000307, -0.00000446, -0.00000545, 0.00000430, -0.00000516, 0.00000667, -0.00000599, -0.00000434, -0.00000484, 0.00000884, -0.00000478, -0.00001358, -0.00000187, 0.00001741, -0.00001126, -0.00000242, -0.00001133, -0.00001411, 0.00000867, 0.00000618, 0.00003818, -0.00001898, 0.00003924, 0.00000064, 0.00000942, -0.00001869, -0.00000998, -0.00000217, -0.00000770, -0.00000281, -0.00000471, -0.00000365, -0.00000617, -0.00000763, -0.00000763, -0.00002514, -0.00001983, -0.00016340, 0.00009236, 0.00038006, 0.00002128, -0.00014744, -0.00000650, -0.00002818, -0.00000110, -0.00001026, -0.00000087, -0.00000866, -0.00000408, 0.00000619, -0.00001071, 0.00000711, -0.00000713, -0.00000241, 0.00000674, -0.00002251, 0.00001022, -0.00001108, 0.00001185, -0.00002583, 0.00001573, -0.00002283, 0.00000599, 0.00000652, -0.00000626, 0.00000380, 0.00000110, 0.00001066, -0.00000420, -0.00000025, -0.00000547, 0.00000563, -0.00000649, 0.00001209, -0.00000639, 0.00000399, -0.00000204, 0.00000263, -0.00000457, 0.00001393, -0.00000813, 0.00000102, -0.00000477, 0.00000542, 0.00000162, -0.00000801, 0.00000451, -0.00000324, 0.00000333, -0.00000831, 0.00000634, 0.00000156, 0.00000081, -0.00000644, -0.00000141, -0.00000384, 0.00000269, 0.00000379, },
		{ -0.00000102, -0.00000899, 0.00000273, 0.00000811, -0.00000124, -0.00000602, 0.00000493, -0.00001308, 0.00000618, 0.00001377, -0.00000692, -0.00000566, 0.00000067, -0.00000403, 0.00000786, 0.00000531, -0.00000063, -0.00000488, -0.00000111, 0.00000088, 0.00000357, 0.00000131, -0.00000080, 0.00000682, -0.00000236, -0.00001529, 0.00000518, 0.00001833, -0.00000999, -0.00000157, -0.00000354, -0.00000476, -0.00000461, 0.00000440, -0.00000530, 0.00000756, -0.00000820, -0.00000260, -0.00000545, 0.00001091, -0.00000867, -0.00001209, 0.00000010, 0.00001797, -0.00001571, 0.00000247, -0.00001899, -0.00000391, 0.00000304, 0.00001508, 0.00002837, -0.00001603, 0.00004012, -0.00000422, 0.00001099, -0.00002174, -0.00000610, -0.00000385, -0.00000686, -0.00000295, -0.00000470, -0.00000354, -0.00000710, -0.00000635, -0.00001113, -0.00002297, -0.00003027, -0.00017311, 0.00015558, 0.00038810, -0.00005692, -0.00012313, -0.00000407, -0.00002743, 0.00000001, -0.00001075, -0.00000025, -0.00000955, -0.00000220, 0.00000509, -0.00001132, 0.00000952, -0.00001091, 0.00000263, 0.00000108, -0.00001888, 0.00000876, -0.00000880, 0.00000860, -0.00002528, 0.00001929, -0.00002991, 0.00001609, -0.00000206, 0.00000002, -0.00000084, 0.00000548, 0.00000748, -0.00000305, -0.00000243, -0.00000267, 0.00000310, -0.00000383, 0.00001073, -0.00000645, 0.00000406, -0.00000184, 0.00000161, -0.00000297, 0.00001374, -0.00001007, 0.00000291, -0.00000655, 0.00000913, -0.00000349, -0.00000442, 0.00000239, -0.00000102, 0.00000085, -0.00000666, 0.00000671, 0.00000109, -0.00000004, -0.00000635, -0.00000124, -0.00000310, 0.00000255, 0.00000435, },
		{ -0.00000339, -0.00000803, 0.00000451, 0.00000802, -0.00000407, -0.00000359, 0.00000338, -0.00001385, 0.00001071, 0.00001100, -0.00000853, -0.00000419, 0.00000041, -0.00000395, 0.00000956, 0.00000350, -0.00000073, -0.00000546, 0.00000000, 0.00000061, 0.00000448, -0.00000063, 0.00000165, 0.00000540, -0.00000397, -0.00001618, 0.00001127, 0.00001414, -0.00001052, -0.00000084, -0.00000332, -0.00000587, -0.00000262, 0.00000318, -0.00000401, 0.00000684, -0.00000890, -0.00000213, -0.00000470, 0.00001147, -0.00001151, -0.00001121, 0.00000353, 0.00001587, -0.00001723, 0.00000437, -0.00002374, 0.00000496, -0.00000245, 0.00002400, 0.00001674, -0.00000992, 0.00003810, -0.00000723, 0.00001177, -0.00002426, -0.00000214, -0.00000618, -0.00000524, -0.00000392, -0.00000377, -0.00000456, -0.00000665, -0.00000671, -0.00001317, -0.00002221, -0.00004138, -0.00018091, 0.00023283, 0.00036970, -0.00012772, -0.00008821, -0.00001073, -0.00002111, -0.00000318, -0.00000789, -0.00000243, -0.00000802, -0.00000228, 0.00000531, -0.00001284, 0.00001278, -0.00001567, 0.00000893, -0.00000650, -0.00001269, 0.00000475, -0.00000392, 0.00000177, -0.00001984, 0.00001759, -0.00003240, 0.00002317, -0.00000930, 0.00000597, -0.00000572, 0.00001045, 0.00000346, -0.00000141, -0.00000467, 0.00000029, 0.00000023, -0.00000052, 0.00000839, -0.00000561, 0.00000345, -0.00000111, -0.00000001, -0.00000035, 0.00001206, -0.00001050, 0.00000344, -0.00000675, 0.00001126, -0.00000786, -0.00000077, 0.00000009, 0.00000139, -0.00000217, -0.00000404, 0.00000612, 0.00000112, -0.00000125, -0.00000584, -0.00000124, -0.00000227, 0.00000215, 0.00000525, },
		{ -0.00000610, -0.00000616, 0.00000585, 0.00000753, -0.00000677, -0.00000044, 0.00000042, -0.00001284, 0.00001428, 0.00000784, -0.00000970, -0.00000240, -0.00000051, -0.00000309, 0.00001062, 0.00000181, -0.00000086, -0.00000578, 0.00000094, 0.00000062, 0.00000480, -0.00000200, 0.00000400, 0.00000346, -0.00000549, -0.00001615, 0.00001717, 0.00000853, -0.00000940, -0.00000102, -0.00000260, -0.00000751, 0.00000018, 0.00000092, -0.00000154, 0.00000468, -0.00000815, -0.00000289, -0.00000260, 0.00001047, -0.00001324, -0.00001079, 0.00000801, 0.00001139, -0.00001579, 0.00000284, -0.00002451, 0.00001090, -0.00000617, 0.00003131, 0.00000490, -0.00000173, 0.00003358, -0.00000831, 0.00001149, -0.00002590, 0.00000150, -0.00000879, -0.00000316, -0.00000541, -0.00000229, -0.00000630, -0.00000528, -0.00000826, -0.00001421, -0.00002234, -0.00005432, -0.00018365, 0.00032167, 0.00032013, -0.00018065, -0.00005115, -0.00002209, -0.00001180, -0.00000880, -0.00000313, -0.00000617, -0.00000511, -0.00000342, 0.00000602, -0.00001441, 0.00001596, -0.00002038, 0.00001540, -0.00001485, -0.00000504, -0.00000078, 0.00000240, -0.00000742, -0.00001035, 0.00001070, -0.00002953, 0.00002591, -0.00001368, 0.00001018, -0.00000960, 0.00001507, -0.00000076, 0.00000032, -0.00000662, 0.00000297, -0.00000250, 0.00000295, 0.00000548, -0.00000415, 0.00000239, -0.00000008, -0.00000190, 0.00000294, 0.00000917, -0.00000954, 0.00000259, -0.00000520, 0.00001133, -0.00001070, 0.00000216, -0.00000170, 0.00000330, -0.00000502, -0.00000101, 0.00000492, 0.00000142, -0.00000259, -0.00000511, -0.00000120, -0.00000162, 0.00000188, 0.00000604, },
		{ -0.00000863, -0.00000374, 0.00000685, 0.00000643, -0.00000884, 0.00000284, -0.00000351, -0.00001009, 0.00001654, 0.00000460, -0.00001044, -0.00000042, -0.00000204, -0.00000135, 0.00001077, 0.00000054, -0.00000125, -0.00000565, 0.00000151, 0.00000107, 0.00000429, -0.00000234, 0.00000563, 0.00000157, -0.00000732, -0.00001463, 0.00002185, 0.00000253, -0.00000727, -0.00000173, -0.00000182, -0.00000909, 0.00000318, -0.00000178, 0.00000151, 0.00000165, -0.00000643, -0.00000445, 0.00000040, 0.00000826, -0.00001419, -0.00001028, 0.00001268, 0.00000549, -0.00001218, -0.00000158, -0.00002125, 0.00001315, -0.00000706, 0.00003576, -0.00000569, 0.00000740, 0.00002711, -0.00000749, 0.00000989, -0.00002626, 0.00000425, -0.00001105, -0.00000125, -0.00000675, -0.00000104, -0.00000785, -0.00000404, -0.00000989, -0.00001549, -0.00002188, -0.00007146, -0.00017593, 0.00041568, 0.00023911, -0.00020772, -0.00002042, -0.00003271, -0.00000312, -0.00001408, 0.00000128, -0.00000958, -0.00000249, -0.00000417, 0.00000590, -0.00001469, 0.00001768, -0.00002364, 0.00002049, -0.00002232, 0.00000245, -0.00000626, 0.00000843, -0.00001700, 0.00000129, -0.00000009, -0.00002176, 0.00002395, -0.00001443, 0.00001168, -0.00001151, 0.00001846, -0.00000454, 0.00000178, -0.00000799, 0.00000503, -0.00000468, 0.00000613, 0.00000244, -0.00000242, 0.00000117, 0.00000093, -0.00000367, 0.00000640, 0.00000555, -0.00000762, 0.00000075, -0.00000215, 0.00000923, -0.00001159, 0.00000386, -0.00000248, 0.00000415, -0.00000703, 0.00000179, 0.00000354, 0.00000171, -0.00000383, -0.00000436, -0.00000096, -0.00000143, 0.00000211, 0.00000626, },
		{ -0.00001046, -0.00000115, 0.00000763, 0.00000457, -0.00000981, 0.00000554, -0.00000773, -0.00000595, 0.00001738, 0.00000150, -0.00001069, 0.00000153, -0.00000396, 0.00000119, 0.00000991, -0.00000011, -0.00000204, -0.00000496, 0.00000161, 0.00000199, 0.00000290, -0.00000143, 0.00000607, 0.00000016, -0.00000973, -0.00001116, 0.00002439, -0.00000276, -0.00000496, -0.00000237, -0.00000152, -0.00000996, 0.00000560, -0.00000411, 0.00000431, -0.00000145, -0.00000449, -0.00000609, 0.00000359, 0.00000549, -0.00001491, -0.00000894, 0.00001636, -0.00000043, -0.00000783, -0.00000750, -0.00001493, 0.00001193, -0.00000491, 0.00003674, -0.00001399, 0.00001642, 0.00001927, -0.00000487, 0.00000676, -0.00002490, 0.00000547, -0.00001221, -0.00000028, -0.00000709, -0.00000098, -0.00000806, -0.00000428, -0.00001005, -0.00001874, -0.00001887, -0.00009585, -0.00015073, 0.00050454, 0.00013180, -0.00020521, -0.00000255, -0.00003768, 0.00000138, -0.00001620, 0.00000303, -0.00001068, -0.00000189, -0.00000300, 0.00000359, -0.00001237, 0.00001657, -0.00002400, 0.00002269, -0.00002727, 0.00000812, -0.00001003, 0.00001225, -0.00002459, 0.00001253, -0.00001251, -0.00001062, 0.00001802, -0.00001162, 0.00001020, -0.00001097, 0.00002005, -0.00000743, 0.00000271, -0.00000860, 0.00000627, -0.00000603, 0.00000864, -0.00000036, -0.00000073, 0.00000005, 0.00000165, -0.00000494, 0.00000951, 0.00000179, -0.00000530, -0.00000149, 0.00000177, 0.00000536, -0.00001053, 0.00000414, -0.00000205, 0.00000365, -0.00000777, 0.00000384, 0.00000239, 0.00000173, -0.00000475, -0.00000373, -0.00000044, -0.00000180, 0.00000308, 0.00000552, },
		{ -0.00001123, 0.00000127, 0.00000828, 0.00000193, -0.00000937, 0.00000704, -0.00001148, -0.00000099, 0.00001692, -0.00000137, -0.00001039, 0.00000315, -0.00000595, 0.00000425, 0.00000816, -0.00000013, -0.00000322, -0.00000374, 0.00000131, 0.00000324, 0.00000083, 0.00000068, 0.00000515, -0.00000052, -0.00001273, -0.00000568, 0.00002419, -0.00000643, -0.00000326, -0.00000239, -0.00000223, -0.00000951, 0.00000675, -0.00000532, 0.00000608, -0.00000377, -0.00000315, -0.00000703, 0.00000617, 0.00000287, -0.00001597, -0.00000605, 0.00001794, -0.00000489, -0.00000445, -0.00001303, -0.00000727, 0.00000839, -0.00000034, 0.00003437, -0.00001944, 0.00002456, 0.00001064, -0.00000064, 0.00000198, -0.00002152, 0.00000464, -0.00001162, -0.00000099, -0.00000567, -0.00000296, -0.00000592, -0.00000722, -0.00000738, -0.00002549, -0.00001161, -0.00013011, -0.00010066, 0.00057508, 0.00000873, -0.00017475, -0.00000065, -0.00003395, -0.00000060, -0.00001324, 0.00000053, -0.00000811, -0.00000449, 0.00000108, -0.00000180, -0.00000660, 0.00001173, -0.00002050, 0.00002089, -0.00002842, 0.00001062, -0.00001076, 0.00001233, -0.00002808, 0.00002071, -0.00002385, 0.00000158, 0.00000971, -0.00000621, 0.00000617, -0.00000803, 0.00001964, -0.00000919, 0.00000302, -0.00000848, 0.00000668, -0.00000648, 0.00001030, -0.00000266, 0.00000070, -0.00000079, 0.00000188, -0.00000543, 0.00001184, -0.00000156, -0.00000320, -0.00000338, 0.00000572, 0.00000046, -0.00000793, 0.00000319, -0.00000057, 0.00000186, -0.00000707, 0.00000486, 0.00000172, 0.00000135, -0.00000527, -0.00000326, 0.00000029, -0.00000265, 0.00000478, 0.00000365, },
		{ -0.00001078, 0.00000329, 0.00000876, -0.00000131, -0.00000746, 0.00000691, -0.00001404, 0.00000409, 0.00001543, -0.00000398, -0.00000942, 0.00000412, -0.00000757, 0.00000741, 0.00000582, 0.00000031, -0.00000465, -0.00000214, 0.00000076, 0.00000452, -0.00000151, 0.00000361, 0.00000306, -0.00000059, -0.00001599, 0.00000143, 0.00002120, -0.00000798, -0.00000268, -0.00000149, -0.00000415, -0.00000745, 0.00000622, -0.00000493, 0.00000627, -0.00000474, -0.00000297, -0.00000669, 0.00000756, 0.00000092, -0.00001771, -0.00000129, 0.00001670, -0.00000677, -0.00000349, -0.00001632, -0.00000031, 0.00000416, 0.00000543, 0.00002939, -0.00002203, 0.00003127, 0.00000177, 0.00000486, -0.00000430, -0.00001612, 0.00000157, -0.00000894, -0.00000370, -0.00000215, -0.00000736, -0.00000099, -0.00001341, -0.00000124, -0.00003637, 0.00000034, -0.00017513, -0.00001985, 0.00061328, -0.00011542, -0.00012327, -0.00001366, -0.00002130, -0.00000940, -0.00000490, -0.00000649, -0.00000166, -0.00001045, 0.00000814, -0.00001022, 0.00000260, 0.00000311, -0.00001298, 0.00001482, -0.00002533, 0.00000943, -0.00000794, 0.00000795, -0.00002616, 0.00002377, -0.00003155, 0.00001234, 0.00000108, 0.00000031, 0.00000061, -0.00000324, 0.00001742, -0.00000983, 0.00000277, -0.00000777, 0.00000640, -0.00000610, 0.00001106, -0.00000435, 0.00000175, -0.00000128, 0.00000159, -0.00000501, 0.00001311, -0.00000411, -0.00000178, -0.00000429, 0.00000877, -0.00000453, -0.00000448, 0.00000152, 0.00000152, -0.00000084, -0.00000514, 0.00000483, 0.00000160, 0.00000056, -0.00000541, -0.00000284, 0.00000101, -0.00000372, 0.00000698, 0.00000074, },
		{ -0.00000920, 0.00000488, 0.00000890, -0.00000476, -0.00000435, 0.00000504, -0.00001491, 0.00000866, 0.00001327, -0.00000635, -0.00000774, 0.00000419, -0.00000840, 0.00001016, 0.00000329, 0.00000093, -0.00000605, -0.00000044, 0.00000024, 0.00000545, -0.00000355, 0.00000670, 0.00000026, -0.00000041, -0.00001887, 0.00000933, 0.00001592, -0.00000744, -0.00000327, 0.00000028, -0.00000713, -0.00000392, 0.00000405, -0.00000293, 0.00000478, -0.00000417, -0.00000414, -0.00000490, 0.00000753, -0.00000024, -0.00001998, 0.00000510, 0.00001263, -0.00000567, -0.00000571, -0.00001605, 0.00000422, 0.00000095, 0.00001093, 0.00002287, -0.00002206, 0.00003621, -0.00000673, 0.00001106, -0.00001160, -0.00000912, -0.00000344, -0.00000442, -0.00000816, 0.00000313, -0.00001373, 0.00000621, -0.00002230, 0.00000772, -0.00005045, 0.00001536, -0.00022869, 0.00009393, 0.00060688, -0.00022413, -0.00006178, -0.00003661, -0.00000240, -0.00002304, 0.00000730, -0.00001672, 0.00000750, -0.00001866, 0.00001707, -0.00002055, 0.00001414, -0.00000829, -0.00000234, 0.00000523, -0.00001854, 0.00000496, -0.00000207, -0.00000041, -0.00001876, 0.00002076, -0.00003391, 0.00001956, -0.00000590, 0.00000630, -0.00000521, 0.00000251, 0.00001382, -0.00000951, 0.00000213, -0.00000668, 0.00000567, -0.00000509, 0.00001104, -0.00000544, 0.00000244, -0.00000147, 0.00000084, -0.00000374, 0.00001326, -0.00000569, -0.00000129, -0.00000381, 0.00001018, -0.00000864, -0.00000103, -0.00000022, 0.00000356, -0.00000382, -0.00000245, 0.00000398, 0.00000189, -0.00000049, -0.00000531, -0.00000231, 0.00000141, -0.00000458, 0.00000922, -0.00000285, },
		{ -0.00000679, 0.00000609, 0.00000846, -0.00000789, -0.00000058, 0.00000170, -0.00001391, 0.00001222, 0.00001075, -0.00000847, -0.00000542, 0.00000322, -0.00000812, 0.00001206, 0.00000104, 0.00000138, -0.00000710, 0.00000105, 0.00000002, 0.00000565, -0.00000470, 0.00000924, -0.00000258, -0.00000054, -0.00002055, 0.00001684, 0.00000937, -0.00000540, -0.00000468, 0.00000247, -0.00001060, 0.00000050, 0.00000076, 0.00000023, 0.00000202, -0.00000241, -0.00000634, -0.00000197, 0.00000636, -0.00000092, -0.00002216, 0.00001224, 0.00000654, -0.00000216, -0.00001083, -0.00001192, 0.00000535, -0.00000004, 0.00001494, 0.00001588, -0.00001997, 0.00003906, -0.00001413, 0.00001711, -0.00001904, -0.00000144, -0.00000949, 0.00000107, -0.00001342, 0.00000906, -0.00002076, 0.00001414, -0.00003213, 0.00001736, -0.00006499, 0.00002928, -0.00028462, 0.00023747, 0.00054822, -0.00030194, -0.00000323, -0.00006188, 0.00001774, -0.00003770, 0.00002021, -0.00002748, 0.00001697, -0.00002693, 0.00002578, -0.00003069, 0.00002595, -0.00002054, 0.00000960, -0.00000617, -0.00000956, -0.00000137, 0.00000533, -0.00001115, -0.00000715, 0.00001214, -0.00003042, 0.00002207, -0.00000978, 0.00001037, -0.00001007, 0.00000828, 0.00000943, -0.00000848, 0.00000123, -0.00000539, 0.00000469, -0.00000368, 0.00001038, -0.00000601, 0.00000281, -0.00000147, -0.00000016, -0.00000182, 0.00001245, -0.00000638, -0.00000166, -0.00000188, 0.00000954, -0.00001113, 0.00000169, -0.00000139, 0.00000489, -0.00000636, 0.00000038, 0.00000273, 0.00000234, -0.00000160, -0.00000514, -0.00000149, 0.00000119, -0.00000478, 0.00001095, -0.00000658, },
		{ -0.00000395, 0.00000707, 0.00000721, -0.00001012, 0.00000313, -0.00000254, -0.00001118, 0.00001452, 0.00000813, -0.00001026, -0.00000262, 0.00000128, -0.00000659, 0.00001276, -0.00000056, 0.00000134, -0.00000753, 0.00000208, 0.00000032, 0.00000488, -0.00000449, 0.00001052, -0.00000477, -0.00000153, -0.00002024, 0.00002271, 0.00000284, -0.00000279, -0.00000619, 0.00000433, -0.00001369, 0.00000488, -0.00000277, 0.00000370, -0.00000120, -0.00000022, -0.00000882, 0.00000137, 0.00000472, -0.00000183, -0.00002324, 0.00001870, -0.00000007, 0.00000235, -0.00001756, -0.00000478, 0.00000319, 0.00000155, 0.00001680, 0.00000922, -0.00001609, 0.00003948, -0.00001958, 0.00002194, -0.00002542, 0.00000558, -0.00001519, 0.00000610, -0.00001793, 0.00001385, -0.00002635, 0.00002035, -0.00004003, 0.00002429, -0.00007562, 0.00003561, -0.00033264, 0.00040105, 0.00043657, -0.00033749, 0.00004015, -0.00008104, 0.00003316, -0.00004863, 0.00002990, -0.00003537, 0.00002372, -0.00003247, 0.00003161, -0.00003801, 0.00003542, -0.00003112, 0.00002042, -0.00001710, -0.00000052, -0.00000756, 0.00001207, -0.00002176, 0.00000621, -0.00000025, -0.00002198, 0.00001984, -0.00001003, 0.00001171, -0.00001308, 0.00001328, 0.00000475, -0.00000694, 0.00000015, -0.00000400, 0.00000361, -0.00000202, 0.00000923, -0.00000613, 0.00000295, -0.00000139, -0.00000122, 0.00000045, 0.00001094, -0.00000643, -0.00000255, 0.00000118, 0.00000686, -0.00001164, 0.00000317, -0.00000154, 0.00000503, -0.00000785, 0.00000273, 0.00000153, 0.00000266, -0.00000253, -0.00000507, -0.00000025, 0.00000014, -0.00000394, 0.00001163, -0.00000984, },
		{ -0.00000110, 0.00000795, 0.00000500, -0.00001096, 0.00000602, -0.00000693, -0.00000716, 0.00001551, 0.00000553, -0.00001158, 0.00000034, -0.00000140, -0.00000390, 0.00001215, -0.00000127, 0.00000064, -0.00000715, 0.00000246, 0.00000120, 0.00000309, -0.00000272, 0.00001011, -0.00000584, -0.00000373, -0.00001740, 0.00002588, -0.00000239, -0.00000067, -0.00000695, 0.00000504, -0.00001546, 0.00000821, -0.00000550, 0.00000644, -0.00000385, 0.00000143, -0.00001062, 0.00000414, 0.00000350, -0.00000384, -0.00002207, 0.00002290, -0.00000533, 0.00000587, -0.00002382, 0.00000352, -0.00000109, 0.00000514, 0.00001658, 0.00000317, -0.00001058, 0.00003707, -0.00002219, 0.00002438, -0.00002941, 0.00001044, -0.00001893, 0.00000900, -0.00001988, 0.00001549, -0.00002815, 0.00002204, -0.00004273, 0.00002453, -0.00007724, 0.00002657, -0.00035935, 0.00056859, 0.00027941, -0.00032600, 0.00005952, -0.00008703, 0.00003855, -0.00005148, 0.00003270, -0.00003719, 0.00002490, -0.00003268, 0.00003206, -0.00004000, 0.00004002, -0.00003752, 0.00002770, -0.00002516, 0.00000634, -0.00001148, 0.00001584, -0.00002944, 0.00001826, -0.00001362, -0.00001063, 0.00001403, -0.00000704, 0.00001028, -0.00001389, 0.00001701, 0.00000016, -0.00000500, -0.00000113, -0.00000247, 0.00000243, -0.00000019, 0.00000766, -0.00000582, 0.00000286, -0.00000130, -0.00000216, 0.00000280, 0.00000905, -0.00000623, -0.00000347, 0.00000475, 0.00000258, -0.00001020, 0.00000327, -0.00000054, 0.00000377, -0.00000790, 0.00000413, 0.00000075, 0.00000262, -0.00000312, -0.00000516, 0.00000136, -0.00000175, -0.00000191, 0.00001088, -0.00001209, },
		{ 0.00000143, 0.00000874, 0.00000191, -0.00001017, 0.00000746, -0.00001065, -0.00000248, 0.00001537, 0.00000298, -0.00001227, 0.00000307, -0.00000438, -0.00000036, 0.00001036, -0.00000109, -0.00000077, -0.00000594, 0.00000221, 0.00000255, 0.00000050, 0.00000047, 0.00000790, -0.00000559, -0.00000714, -0.00001196, 0.00002575, -0.00000538, 0.00000009, -0.00000631, 0.00000395, -0.00001515, 0.00000961, -0.00000653, 0.00000755, -0.00000502, 0.00000162, -0.00001084, 0.00000547, 0.00000348, -0.00000761, -0.00001779, 0.00002349, -0.00000753, 0.00000643, -0.00002729, 0.00001062, -0.00000557, 0.00000938, 0.00001506, -0.00000238, -0.00000346, 0.00003152, -0.00002122, 0.00002343, -0.00002984, 0.00001178, -0.00001928, 0.00000828, -0.00001772, 0.00001224, -0.00002416, 0.00001680, -0.00003735, 0.00001462, -0.00006536, -0.00000518, -0.00035019, 0.00071923, 0.00009228, -0.00027052, 0.00005132, -0.00007600, 0.00003076, -0.00004360, 0.00002631, -0.00003095, 0.00001874, -0.00002590, 0.00002553, -0.00003500, 0.00003797, -0.00003793, 0.00002962, -0.00002855, 0.00000931, -0.00001154, 0.00001485, -0.00003184, 0.00002602, -0.00002484, 0.00000092, 0.00000657, -0.00000205, 0.00000672, -0.00001268, 0.00001931, -0.00000413, -0.00000270, -0.00000270, -0.00000071, 0.00000107, 0.00000182, 0.00000567, -0.00000507, 0.00000250, -0.00000121, -0.00000286, 0.00000501, 0.00000707, -0.00000609, -0.00000385, 0.00000800, -0.00000251, -0.00000726, 0.00000221, 0.00000140, 0.00000125, -0.00000648, 0.00000438, 0.00000060, 0.00000210, -0.00000333, -0.00000534, 0.00000316, -0.00000426, 0.00000120, 0.00000861, -0.00001301, },
		{ 0.00000349, 0.00000934, -0.00000176, -0.00000783, 0.00000713, -0.00001305, 0.00000219, 0.00001440, 0.00000047, -0.00001215, 0.00000516, -0.00000718, 0.00000354, 0.00000770, -0.00000018, -0.00000271, -0.00000406, 0.00000147, 0.00000409, -0.00000244, 0.00000461, 0.00000420, -0.00000423, -0.00001135, -0.00000441, 0.00002238, -0.00000573, -0.00000094, -0.00000395, 0.00000083, -0.00001244, 0.00000864, -0.00000539, 0.00000652, -0.00000418, -0.00000015, -0.00000898, 0.00000485, 0.00000505, -0.00001335, -0.00001013, 0.00001983, -0.00000562, 0.00000265, -0.00002608, 0.00001424, -0.00000815, 0.00001259, 0.00001334, -0.00000782, 0.00000523, 0.00002277, -0.00001635, 0.00001856, -0.00002608, 0.00000886, -0.00001540, 0.00000315, -0.00001066, 0.00000329, -0.00001344, 0.00000351, -0.00002251, -0.00000716, -0.00003767, -0.00006433, -0.00029236, 0.00083015, -0.00010299, -0.00018182, 0.00001820, -0.00004841, 0.00000992, -0.00002494, 0.00001065, -0.00001656, 0.00000516, -0.00001212, 0.00001196, -0.00002276, 0.00002882, -0.00003175, 0.00002546, -0.00002649, 0.00000764, -0.00000713, 0.00000842, -0.00002764, 0.00002738, -0.00003126, 0.00001001, -0.00000030, 0.00000330, 0.00000215, -0.00001001, 0.00002033, -0.00000804, -0.00000007, -0.00000467, 0.00000143, -0.00000058, 0.00000407, 0.00000326, -0.00000383, 0.00000179, -0.00000104, -0.00000334, 0.00000698, 0.00000514, -0.00000624, -0.00000328, 0.00001016, -0.00000749, -0.00000354, 0.00000050, 0.00000379, -0.00000205, -0.00000389, 0.00000355, 0.00000108, 0.00000117, -0.00000327, -0.00000543, 0.00000480, -0.00000694, 0.00000500, 0.00000503, -0.00001250, },
		{ 0.00000509, 0.00000950, -0.00000553, -0.00000432, 0.00000504, -0.00001372, 0.00000627, 0.00001291, -0.00000203, -0.00001113, 0.00000630, -0.00000929, 0.00000725, 0.00000466, 0.00000110, -0.00000486, -0.00000181, 0.00000052, 0.00000542, -0.00000513, 0.00000896, -0.00000037, -0.00000229, -0.00001557, 0.00000426, 0.00001649, -0.00000371, -0.00000362, -0.00000013, -0.00000403, -0.00000759, 0.00000547, -0.00000218, 0.00000342, -0.00000137, -0.00000382, -0.00000509, 0.00000236, 0.00000798, -0.00002054, 0.00000030, 0.00001227, 0.00000034, -0.00000575, -0.00001933, 0.00001290, -0.00000718, 0.00001328, 0.00001250, -0.00001345, 0.00001521, 0.00001124, -0.00000785, 0.00000998, -0.00001832, 0.00000181, -0.00000747, -0.00000613, 0.00000087, -0.00001077, 0.00000328, -0.00001706, 0.00000091, -0.00003971, 0.00000449, -0.00015081, -0.00017794, 0.00088045, -0.00028149, -0.00007658, -0.00003155, -0.00000917, -0.00002038, 0.00000166, -0.00001191, 0.00000386, -0.00001390, 0.00000687, -0.00000699, -0.00000472, 0.00001369, -0.00001983, 0.00001595, -0.00001956, 0.00000192, 0.00000099, -0.00000272, -0.00001712, 0.00002169, -0.00003136, 0.00001465, -0.00000467, 0.00000740, -0.00000223, -0.00000663, 0.00002029, -0.00001150, 0.00000282, -0.00000702, 0.00000397, -0.00000257, 0.00000656, 0.00000043, -0.00000209, 0.00000068, -0.00000069, -0.00000364, 0.00000870, 0.00000329, -0.00000668, -0.00000156, 0.00001061, -0.00001144, 0.00000013, -0.00000121, 0.00000596, -0.00000544, -0.00000068, 0.00000198, 0.00000197, 0.00000002, -0.00000312, -0.00000518, 0.00000584, -0.00000923, 0.00000887, 0.00000060, -0.00001073, },
		{ 0.00000633, 0.00000894, -0.00000875, -0.00000032, 0.00000158, -0.00001258, 0.00000935, 0.00001117, -0.00000449, -0.00000925, 0.00000627, -0.00001030, 0.00001023, 0.00000175, 0.00000233, -0.00000685, 0.00000045, -0.00000032, 0.00000613, -0.00000690, 0.00001267, -0.00000498, -0.00000050, -0.00001882, 0.00001268, 0.00000933, -0.00000020, -0.00000724, 0.00000440, -0.00000975, -0.00000144, 0.00000085, 0.00000238, -0.00000105, 0.00000272, -0.00000870, 0.00000009, -0.00000128, 0.00001141, -0.00002797, 0.00001200, 0.00000224, 0.00000911, -0.00001771, -0.00000757, 0.00000638, -0.00000201, 0.00001068, 0.00001314, -0.00001922, 0.00002576, -0.00000204, 0.00000322, -0.00000124, -0.00000770, -0.00000814, 0.00000319, -0.00001802, 0.00001503, -0.00002775, 0.00002341, -0.00004189, 0.00002934, -0.00007863, 0.00005529, -0.00025814, -0.00000678, 0.00085512, -0.00041933, 0.00002557, -0.00008575, 0.00003346, -0.00005368, 0.00003088, -0.00003682, 0.00002630, -0.00003475, 0.00002763, -0.00002807, 0.00001616, -0.00000482, -0.00000446, 0.00000310, -0.00000957, -0.00000611, 0.00001088, -0.00001639, -0.00000212, 0.00000996, -0.00002515, 0.00001409, -0.00000543, 0.00000915, -0.00000548, -0.00000319, 0.00001942, -0.00001436, 0.00000571, -0.00000954, 0.00000676, -0.00000477, 0.00000918, -0.00000267, 0.00000007, -0.00000082, -0.00000013, -0.00000384, 0.00001021, 0.00000144, -0.00000724, 0.00000119, 0.00000906, -0.00001367, 0.00000301, -0.00000229, 0.00000727, -0.00000817, 0.00000246, 0.00000016, 0.00000294, -0.00000109, -0.00000312, -0.00000433, 0.00000591, -0.00001054, 0.00001211, -0.00000400, -0.00000808, },
		{ 0.00000740, 0.00000741, -0.00001082, 0.00000337, -0.00000258, -0.00000989, 0.00001127, 0.00000936, -0.00000680, -0.00000666, 0.00000509, -0.00001000, 0.00001206, -0.00000056, 0.00000312, -0.00000826, 0.00000234, -0.00000076, 0.00000593, -0.00000725, 0.00001495, -0.00000881, 0.00000042, -0.00002015, 0.00001943, 0.00000243, 0.00000353, -0.00001073, 0.00000853, -0.00001514, 0.00000477, -0.00000402, 0.00000712, -0.00000570, 0.00000689, -0.00001356, 0.00000534, -0.00000485, 0.00001400, -0.00003392, 0.00002285, -0.00000798, 0.00001836, -0.00003087, 0.00000718, -0.00000404, 0.00000671, 0.00000505, 0.00001509, -0.00002454, 0.00003562, -0.00001542, 0.00001507, -0.00001323, 0.00000374, -0.00001883, 0.00001425, -0.00002990, 0.00002877, -0.00004405, 0.00004271, -0.00006600, 0.00005683, -0.00011643, 0.00010452, -0.00037296, 0.00021164, 0.00074853, -0.00049790, 0.00010585, -0.00013100, 0.00006967, -0.00008210, 0.00005613, -0.00005836, 0.00004566, -0.00005273, 0.00004573, -0.00004707, 0.00003594, -0.00002314, 0.00001116, -0.00001016, 0.00000079, -0.00001386, 0.00001976, -0.00002946, 0.00001419, -0.00000522, -0.00001423, 0.00000896, -0.00000260, 0.00000823, -0.00000712, -0.00000004, 0.00001775, -0.00001632, 0.00000815, -0.00001180, 0.00000943, -0.00000688, 0.00001163, -0.00000575, 0.00000239, -0.00000256, 0.00000061, -0.00000395, 0.00001156, -0.00000056, -0.00000759, 0.00000458, 0.00000564, -0.00001384, 0.00000461, -0.00000232, 0.00000723, -0.00000962, 0.00000485, -0.00000138, 0.00000366, -0.00000189, -0.00000343, -0.00000277, 0.00000474, -0.00001042, 0.00001408, -0.00000808, -0.00000502, },
		{ 0.00000839, 0.00000482, -0.00001130, 0.00000599, -0.00000662, -0.00000616, 0.00001205, 0.00000755, -0.00000880, -0.00000366, 0.00000293, -0.00000838, 0.00001253, -0.00000193, 0.00000315, -0.00000881, 0.00000358, -0.00000063, 0.00000468, -0.00000589, 0.00001526, -0.00001122, -0.00000003, -0.00001888, 0.00002332, -0.00000277, 0.00000618, -0.00001293, 0.00001108, -0.00001889, 0.00000965, -0.00000777, 0.00001069, -0.00000916, 0.00000978, -0.00001700, 0.00000923, -0.00000694, 0.00001427, -0.00003654, 0.00003050, -0.00001574, 0.00002524, -0.00004211, 0.00002183, -0.00001583, 0.00001710, -0.00000229, 0.00001741, -0.00002828, 0.00004316, -0.00002682, 0.00002545, -0.00002363, 0.00001345, -0.00002753, 0.00002277, -0.00003851, 0.00003840, -0.00005538, 0.00005609, -0.00008333, 0.00007600, -0.00014367, 0.00013886, -0.00047600, 0.00045799, 0.00056663, -0.00050739, 0.00014993, -0.00015571, 0.00009046, -0.00009821, 0.00007104, -0.00007098, 0.00005694, -0.00006321, 0.00005681, -0.00005977, 0.00005056, -0.00003753, 0.00002363, -0.00002071, 0.00000867, -0.00001864, 0.00002470, -0.00003849, 0.00002803, -0.00002030, -0.00000141, 0.00000111, 0.00000277, 0.00000514, -0.00000733, 0.00000282, 0.00001515, -0.00001694, 0.00000956, -0.00001321, 0.00001141, -0.00000840, 0.00001346, -0.00000838, 0.00000450, -0.00000427, 0.00000136, -0.00000388, 0.00001271, -0.00000278, -0.00000735, 0.00000799, 0.00000081, -0.00001202, 0.00000477, -0.00000116, 0.00000567, -0.00000942, 0.00000602, -0.00000223, 0.00000384, -0.00000221, -0.00000411, -0.00000051, 0.00000232, -0.00000867, 0.00001435, -0.00001106, -0.00000199, },
		{ 0.00000928, 0.00000134, -0.00001003, 0.00000699, -0.00000980, -0.00000202, 0.00001193, 0.00000573, -0.00001029, -0.00000062, 0.00000014, -0.00000566, 0.00001165, -0.00000226, 0.00000228, -0.00000833, 0.00000402, 0.00000010, 0.00000250, -0.00000288, 0.00001342, -0.00001190, -0.00000200, -0.00001484, 0.00002371, -0.00000522, 0.00000670, -0.00001289, 0.00001110, -0.00001995, 0.00001207, -0.00000925, 0.00001191, -0.00001025, 0.00001016, -0.00001777, 0.00001050, -0.00000635, 0.00001103, -0.00003437, 0.00003301, -0.00001869, 0.00002702, -0.00004819, 0.00003286, -0.00002576, 0.00002647, -0.00000927, 0.00001862, -0.00002907, 0.00004661, -0.00003414, 0.00003209, -0.00003007, 0.00001890, -0.00003156, 0.00002589, -0.00004075, 0.00004042, -0.00005767, 0.00005872, -0.00008808, 0.00007970, -0.00015096, 0.00014423, -0.00054462, 0.00070475, 0.00032711, -0.00044890, 0.00015077, -0.00015271, 0.00008988, -0.00009686, 0.00007110, -0.00007067, 0.00005652, -0.00006282, 0.00005760, -0.00006289, 0.00005681, -0.00004494, 0.00003017, -0.00002601, 0.00001178, -0.00001836, 0.00002348, -0.00004066, 0.00003593, -0.00003161, 0.00000998, -0.00000685, 0.00000885, 0.00000105, -0.00000672, 0.00000568, 0.00001135, -0.00001578, 0.00000933, -0.00001313, 0.00001208, -0.00000878, 0.00001414, -0.00001005, 0.00000593, -0.00000553, 0.00000183, -0.00000345, 0.00001350, -0.00000520, -0.00000624, 0.00001074, -0.00000467, -0.00000864, 0.00000367, 0.00000099, 0.00000276, -0.00000758, 0.00000579, -0.00000219, 0.00000337, -0.00000204, -0.00000507, 0.00000218, -0.00000108, -0.00000541, 0.00001280, -0.00001257, 0.00000070, },
		{ 0.00000990, -0.00000265, -0.00000721, 0.00000616, -0.00001153, 0.00000192, 0.00001117, 0.00000387, -0.00001112, 0.00000208, -0.00000283, -0.00000222, 0.00000966, -0.00000162, 0.00000058, -0.00000686, 0.00000366, 0.00000127, -0.00000028, 0.00000139, 0.00000963, -0.00001096, -0.00000519, -0.00000845, 0.00002064, -0.00000456, 0.00000463, -0.00001022, 0.00000819, -0.00001780, 0.00001140, -0.00000781, 0.00001014, -0.00000829, 0.00000737, -0.00001518, 0.00000846, -0.00000248, 0.00000382, -0.00002681, 0.00002936, -0.00001544, 0.00002190, -0.00004665, 0.00003722, -0.00003071, 0.00003201, -0.00001362, 0.00001712, -0.00002562, 0.00004458, -0.00003571, 0.00003321, -0.00003075, 0.00001824, -0.00002899, 0.00002164, -0.00003452, 0.00003253, -0.00004826, 0.00004735, -0.00007623, 0.00006283, -0.00013129, 0.00010889, -0.00055648, 0.00091972, 0.00005752, -0.00033449, 0.00011005, -0.00012100, 0.00006655, -0.00007659, 0.00005489, -0.00005610, 0.00004317, -0.00005036, 0.00004683, -0.00005501, 0.00005308, -0.00004379, 0.00002933, -0.00002479, 0.00000906, -0.00001219, 0.00001522, -0.00003449, 0.00003557, -0.00003620, 0.00001694, -0.00001232, 0.00001361, -0.00000258, -0.00000616, 0.00000894, 0.00000615, -0.00001255, 0.00000703, -0.00001106, 0.00001093, -0.00000752, 0.00001322, -0.00001031, 0.00000624, -0.00000597, 0.00000172, -0.00000240, 0.00001371, -0.00000765, -0.00000413, 0.00001219, -0.00000993, -0.00000438, 0.00000179, 0.00000366, -0.00000101, -0.00000442, 0.00000426, -0.00000128, 0.00000233, -0.00000150, -0.00000606, 0.00000491, -0.00000498, -0.00000108, 0.00000963, -0.00001254, 0.00000290, },
		{ 0.00000997, -0.00000654, -0.00000339, 0.00000368, -0.00001154, 0.00000516, 0.00001006, 0.00000195, -0.00001118, 0.00000413, -0.00000552, 0.00000144, 0.00000694, -0.00000027, -0.00000173, -0.00000460, 0.00000268, 0.00000263, -0.00000315, 0.00000626, 0.00000448, -0.00000885, -0.00000889, -0.00000066, 0.00001488, -0.00000118, 0.00000020, -0.00000517, 0.00000262, -0.00001266, 0.00000776, -0.00000357, 0.00000550, -0.00000342, 0.00000152, -0.00000934, 0.00000325, 0.00000436, -0.00000685, -0.00001443, 0.00001987, -0.00000602, 0.00000959, -0.00003654, 0.00003314, -0.00002852, 0.00003158, -0.00001352, 0.00001176, -0.00001729, 0.00003646, -0.00003081, 0.00002808, -0.00002503, 0.00001087, -0.00001928, 0.00000957, -0.00001949, 0.00001442, -0.00002681, 0.00002149, -0.00004698, 0.00002413, -0.00008239, 0.00002674, -0.00049404, 0.00107092, -0.00020867, -0.00018530, 0.00003773, -0.00006603, 0.00002425, -0.00004024, 0.00002464, -0.00002914, 0.00001852, -0.00002726, 0.00002569, -0.00003696, 0.00003988, -0.00003439, 0.00002142, -0.00001740, 0.00000103, -0.00000093, 0.00000081, -0.00002042, 0.00002646, -0.00003271, 0.00001754, -0.00001333, 0.00001538, -0.00000451, -0.00000640, 0.00001285, -0.00000042, -0.00000727, 0.00000259, -0.00000685, 0.00000775, -0.00000441, 0.00001048, -0.00000890, 0.00000513, -0.00000529, 0.00000077, -0.00000050, 0.00001304, -0.00000980, -0.00000111, 0.00001195, -0.00001415, -0.00000002, -0.00000027, 0.00000621, -0.00000495, -0.00000057, 0.00000185, 0.00000024, 0.00000095, -0.00000084, -0.00000675, 0.00000715, -0.00000870, 0.00000367, 0.00000537, -0.00001115, 0.00000465, },
		{ 0.00000918, -0.00000966, 0.00000067, 0.00000007, -0.00000991, 0.00000741, 0.00000881, 0.00000002, -0.00001048, 0.00000532, -0.00000756, 0.00000482, 0.00000396, 0.00000138, -0.00000427, -0.00000189, 0.00000134, 0.00000384, -0.00000552, 0.00001091, -0.00000123, -0.00000625, -0.00001219, 0.00000718, 0.00000776, 0.00000387, -0.00000568, 0.00000133, -0.00000463, -0.00000546, 0.00000203, 0.00000262, -0.00000113, 0.00000346, -0.00000641, -0.00000126, -0.00000409, 0.00001297, -0.00001943, 0.00000102, 0.00000627, 0.00000794, -0.00000842, -0.00001881, 0.00002074, -0.00001864, 0.00002435, -0.00000823, 0.00000232, -0.00000440, 0.00002280, -0.00001996, 0.00001731, -0.00001372, -0.00000220, -0.00000365, -0.00000883, 0.00000254, -0.00001172, 0.00000411, -0.00001599, -0.00000358, -0.00003270, -0.00000837, -0.00010003, -0.00034877, 0.00113205, -0.00043723, -0.00002782, -0.00005001, 0.00000137, -0.00002875, 0.00000552, -0.00001404, 0.00000533, -0.00001307, 0.00000252, -0.00000226, -0.00001184, 0.00001979, -0.00001893, 0.00000847, -0.00000582, -0.00001026, 0.00001308, -0.00001714, -0.00000084, 0.00001020, -0.00002169, 0.00001149, -0.00000915, 0.00001333, -0.00000407, -0.00000778, 0.00001733, -0.00000792, -0.00000043, -0.00000360, -0.00000080, 0.00000276, 0.00000039, 0.00000603, -0.00000586, 0.00000260, -0.00000342, -0.00000111, 0.00000236, 0.00001125, -0.00001126, 0.00000248, 0.00000989, -0.00001671, 0.00000371, -0.00000191, 0.00000802, -0.00000827, 0.00000324, -0.00000088, 0.00000198, -0.00000045, -0.00000036, -0.00000682, 0.00000842, -0.00001158, 0.00000805, 0.00000070, -0.00000881, 0.00000608, },
		{ 0.00000733, -0.00001138, 0.00000411, -0.00000392, -0.00000702, 0.00000861, 0.00000755, -0.00000186, -0.00000914, 0.00000559, -0.00000867, 0.00000751, 0.00000121, 0.00000293, -0.00000663, 0.00000089, -0.00000006, 0.00000459, -0.00000688, 0.00001454, -0.00000665, -0.00000392, -0.00001415, 0.00001366, 0.00000092, 0.00000911, -0.00001165, 0.00000790, -0.00001211, 0.00000231, -0.00000436, 0.00000931, -0.00000821, 0.00001076, -0.00001478, 0.00000737, -0.00001182, 0.00002144, -0.00003163, 0.00001684, -0.00000858, 0.00002344, -0.00002900, 0.00000366, 0.00000212, -0.00000246, 0.00001115, 0.00000161, -0.00001023, 0.00001154, 0.00000537, -0.00000503, 0.00000287, 0.00000093, -0.00001842, 0.00001498, -0.00003021, 0.00002767, -0.00004135, 0.00003921, -0.00005891, 0.00004674, -0.00009890, 0.00007981, -0.00025948, -0.00012428, 0.00108744, -0.00059886, 0.00011090, -0.00013413, 0.00006736, -0.00008145, 0.00005154, -0.00005333, 0.00004042, -0.00004536, 0.00003328, -0.00003177, 0.00001565, -0.00000309, -0.00000103, -0.00000623, 0.00000681, -0.00002166, 0.00002641, -0.00003478, 0.00002032, -0.00000983, -0.00000558, 0.00000026, -0.00000052, 0.00000779, -0.00000140, -0.00001006, 0.00002181, -0.00001550, 0.00000701, -0.00001062, 0.00000628, -0.00000335, 0.00000627, 0.00000041, -0.00000163, -0.00000103, -0.00000059, -0.00000374, 0.00000609, 0.00000826, -0.00001165, 0.00000615, 0.00000620, -0.00001728, 0.00000629, -0.00000268, 0.00000858, -0.00001033, 0.00000628, -0.00000333, 0.00000350, -0.00000154, -0.00000030, -0.00000604, 0.00000837, -0.00001306, 0.00001135, -0.00000365, -0.00000597, 0.00000733, },
		{ 0.00000437, -0.00001134, 0.00000620, -0.00000745, -0.00000343, 0.00000887, 0.00000633, -0.00000354, -0.00000737, 0.00000502, -0.00000876, 0.00000922, -0.00000092, 0.00000397, -0.00000840, 0.00000332, -0.00000124, 0.00000465, -0.00000689, 0.00001652, -0.00001104, -0.00000243, -0.00001408, 0.00001759, -0.00000411, 0.00001299, -0.00001623, 0.00001301, -0.00001819, 0.00000898, -0.00000972, 0.00001478, -0.00001399, 0.00001661, -0.00002164, 0.00001455, -0.00001787, 0.00002753, -0.00004083, 0.00002995, -0.00002125, 0.00003672, -0.00004798, 0.00002660, -0.00001891, 0.00001694, -0.00000555, 0.00001398, -0.00002378, 0.00002803, -0.00001302, 0.00001109, -0.00001218, 0.00001556, -0.00003405, 0.00003248, -0.00004992, 0.00005057, -0.00006828, 0.00007121, -0.00009872, 0.00009373, -0.00016174, 0.00016535, -0.00043037, 0.00016237, 0.00093566, -0.00067420, 0.00020815, -0.00019671, 0.00011811, -0.00012247, 0.00008816, -0.00008484, 0.00006869, -0.00007161, 0.00005874, -0.00005700, 0.00004016, -0.00002401, 0.00001515, -0.00001888, 0.00001695, -0.00002975, 0.00003537, -0.00004785, 0.00003841, -0.00002912, 0.00001179, -0.00001329, 0.00001057, 0.00000011, 0.00000252, -0.00001240, 0.00002534, -0.00002196, 0.00001374, -0.00001720, 0.00001318, -0.00000950, 0.00001226, -0.00000551, 0.00000302, -0.00000512, 0.00000266, -0.00000668, 0.00001031, 0.00000421, -0.00001074, 0.00000933, 0.00000134, -0.00001588, 0.00000746, -0.00000237, 0.00000767, -0.00001072, 0.00000798, -0.00000498, 0.00000444, -0.00000208, -0.00000079, -0.00000435, 0.00000688, -0.00001282, 0.00001307, -0.00000707, -0.00000308, 0.00000848, },
		{ 0.00000056, -0.00000950, 0.00000651, -0.00000983, 0.00000024, 0.00000841, 0.00000515, -0.00000492, -0.00000539, 0.00000381, -0.00000789, 0.00000980, -0.00000219, 0.00000423, -0.00000926, 0.00000508, -0.00000198, 0.00000398, -0.00000542, 0.00001652, -0.00001392, -0.00000208, -0.00001172, 0.00001829, -0.00000624, 0.00001429, -0.00001817, 0.00001537, -0.00002147, 0.00001306, -0.00001256, 0.00001750, -0.00001682, 0.00001932, -0.00002519, 0.00001839, -0.00002036, 0.00002928, -0.00004475, 0.00003758, -0.00002855, 0.00004410, -0.00006101, 0.00004524, -0.00003770, 0.00003538, -0.00002220, 0.00002592, -0.00003552, 0.00004198, -0.00002903, 0.00002498, -0.00002436, 0.00002644, -0.00004501, 0.00004436, -0.00006293, 0.00006555, -0.00008594, 0.00009237, -0.00012615, 0.00012610, -0.00020680, 0.00022823, -0.00058452, 0.00048180, 0.00069075, -0.00065716, 0.00024959, -0.00022462, 0.00014281, -0.00014262, 0.00010739, -0.00010153, 0.00008383, -0.00008600, 0.00007346, -0.00007280, 0.00005688, -0.00003868, 0.00002584, -0.00002615, 0.00002159, -0.00003169, 0.00003696, -0.00005275, 0.00004914, -0.00004311, 0.00002614, -0.00002554, 0.00002130, -0.00000762, 0.00000618, -0.00001362, 0.00002678, -0.00002600, 0.00001832, -0.00002189, 0.00001852, -0.00001439, 0.00001717, -0.00001064, 0.00000710, -0.00000879, 0.00000560, -0.00000927, 0.00001444, -0.00000050, -0.00000850, 0.00001150, -0.00000405, -0.00001282, 0.00000728, -0.00000103, 0.00000537, -0.00000935, 0.00000809, -0.00000553, 0.00000460, -0.00000197, -0.00000183, -0.00000191, 0.00000410, -0.00001090, 0.00001300, -0.00000916, -0.00000044, 0.00000946, },
		{ -0.00000362, -0.00000619, 0.00000497, -0.00001061, 0.00000345, 0.00000751, 0.00000400, -0.00000592, -0.00000344, 0.00000219, -0.00000623, 0.00000932, -0.00000249, 0.00000359, -0.00000903, 0.00000596, -0.00000218, 0.00000266, -0.00000264, 0.00001450, -0.00001514, -0.00000281, -0.00000733, 0.00001574, -0.00000502, 0.00001239, -0.00001681, 0.00001426, -0.00002111, 0.00001364, -0.00001193, 0.00001650, -0.00001572, 0.00001779, -0.00002427, 0.00001769, -0.00001808, 0.00002551, -0.00004199, 0.00003794, -0.00002829, 0.00004292, -0.00006467, 0.00005533, -0.00004972, 0.00004852, -0.00003494, 0.00003419, -0.00004254, 0.00005042, -0.00003954, 0.00003351, -0.00003055, 0.00003032, -0.00004784, 0.00004684, -0.00006510, 0.00006792, -0.00008889, 0.00009619, -0.00013331, 0.00013398, -0.00022099, 0.00024891, -0.00069109, 0.00079581, 0.00038089, -0.00055604, 0.00023176, -0.00021213, 0.00013608, -0.00013699, 0.00010473, -0.00009933, 0.00008213, -0.00008507, 0.00007411, -0.00007587, 0.00006263, -0.00004424, 0.00002865, -0.00002602, 0.00001903, -0.00002603, 0.00002971, -0.00004746, 0.00004961, -0.00004825, 0.00003371, -0.00003305, 0.00002881, -0.00001316, 0.00000794, -0.00001252, 0.00002511, -0.00002651, 0.00001956, -0.00002346, 0.00002107, -0.00001686, 0.00001990, -0.00001393, 0.00000965, -0.00001116, 0.00000741, -0.00001075, 0.00001774, -0.00000526, -0.00000515, 0.00001230, -0.00000927, -0.00000861, 0.00000601, 0.00000102, 0.00000204, -0.00000647, 0.00000665, -0.00000493, 0.00000396, -0.00000126, -0.00000321, 0.00000095, 0.00000044, -0.00000762, 0.00001128, -0.00000981, 0.00000180, 0.00001006, },
		{ -0.00000750, -0.00000207, 0.00000191, -0.00000969, 0.00000581, 0.00000637, 0.00000294, -0.00000654, -0.00000166, 0.00000041, -0.00000405, 0.00000794, -0.00000191, 0.00000206, -0.00000771, 0.00000586, -0.00000186, 0.00000094, 0.00000106, 0.00001078, -0.00001488, -0.00000419, -0.00000167, 0.00001063, -0.00000084, 0.00000746, -0.00001224, 0.00000972, -0.00001710, 0.00001058, -0.00000771, 0.00001168, -0.00001053, 0.00001187, -0.00001866, 0.00001221, -0.00001089, 0.00001617, -0.00003252, 0.00003077, -0.00001994, 0.00003226, -0.00005726, 0.00005425, -0.00005175, 0.00005294, -0.00004056, 0.00003615, -0.00004263, 0.00005120, -0.00004236, 0.00003459, -0.00002883, 0.00002533, -0.00004060, 0.00003790, -0.00005424, 0.00005525, -0.00007427, 0.00007916, -0.00011569, 0.00011142, -0.00019577, 0.00021263, -0.00072186, 0.00106285, 0.00004443, -0.00039195, 0.00016229, -0.00016202, 0.00009917, -0.00010605, 0.00008027, -0.00007811, 0.00006337, -0.00006850, 0.00006021, -0.00006553, 0.00005662, -0.00003996, 0.00002311, -0.00001831, 0.00000937, -0.00001322, 0.00001421, -0.00003212, 0.00003911, -0.00004289, 0.00003225, -0.00003339, 0.00003090, -0.00001477, 0.00000657, -0.00000832, 0.00001977, -0.00002294, 0.00001679, -0.00002115, 0.00002001, -0.00001608, 0.00001965, -0.00001460, 0.00000993, -0.00001152, 0.00000745, -0.00001047, 0.00001947, -0.00000932, -0.00000115, 0.00001161, -0.00001371, -0.00000388, 0.00000411, 0.00000331, -0.00000174, -0.00000262, 0.00000400, -0.00000335, 0.00000270, -0.00000018, -0.00000465, 0.00000376, -0.00000351, -0.00000354, 0.00000832, -0.00000917, 0.00000368, 0.00001000, },
		{ -0.00001038, 0.00000198, -0.00000199, -0.00000731, 0.00000718, 0.00000515, 0.00000201, -0.00000684, -0.00000016, -0.00000130, -0.00000165, 0.00000595, -0.00000065, -0.00000014, -0.00000543, 0.00000486, -0.00000110, -0.00000089, 0.00000512, 0.00000589, -0.00001355, -0.00000560, 0.00000416, 0.00000420, 0.00000526, 0.00000039, -0.00000533, 0.00000260, -0.00001021, 0.00000466, -0.00000068, 0.00000387, -0.00000214, 0.00000243, -0.00000924, 0.00000287, 0.00000018, 0.00000253, -0.00001773, 0.00001744, -0.00000484, 0.00001331, -0.00003940, 0.00004169, -0.00004262, 0.00004700, -0.00003734, 0.00003044, -0.00003490, 0.00004362, -0.00003683, 0.00002777, -0.00001907, 0.00001157, -0.00002358, 0.00001797, -0.00003092, 0.00002824, -0.00004288, 0.00004197, -0.00007375, 0.00005839, -0.00012988, 0.00011334, -0.00065681, 0.00124461, -0.00027593, -0.00019509, 0.00005811, -0.00008473, 0.00003964, -0.00005568, 0.00003872, -0.00004185, 0.00003099, -0.00003931, 0.00003437, -0.00004393, 0.00004054, -0.00002736, 0.00001082, -0.00000481, -0.00000537, 0.00000438, -0.00000690, -0.00000923, 0.00001938, -0.00002779, 0.00002160, -0.00002585, 0.00002667, -0.00001169, 0.00000162, -0.00000094, 0.00001095, -0.00001550, 0.00001013, -0.00001499, 0.00001526, -0.00001189, 0.00001620, -0.00001240, 0.00000764, -0.00000956, 0.00000538, -0.00000804, 0.00001911, -0.00001200, 0.00000290, 0.00000950, -0.00001689, 0.00000074, 0.00000205, 0.00000533, -0.00000528, 0.00000151, 0.00000069, -0.00000119, 0.00000112, 0.00000098, -0.00000582, 0.00000603, -0.00000711, 0.00000063, 0.00000473, -0.00000757, 0.00000528, 0.00000898, },
		{ -0.00001169, 0.00000511, -0.00000589, -0.00000396, 0.00000760, 0.00000393, 0.00000130, -0.00000696, 0.00000108, -0.00000281, 0.00000070, 0.00000368, 0.00000097, -0.00000271, -0.00000249, 0.00000313, -0.00000006, -0.00000246, 0.00000895, 0.00000049, -0.00001164, -0.00000637, 0.00000898, -0.00000203, 0.00001176, -0.00000737, 0.00000252, -0.00000565, -0.00000192, -0.00000268, 0.00000765, -0.00000533, 0.00000776, -0.00000874, 0.00000217, -0.00000846, 0.00001311, -0.00001308, -0.00000028, 0.00000080, 0.00001398, -0.00001072, -0.00001406, 0.00001981, -0.00002364, 0.00003133, -0.00002553, 0.00001749, -0.00002017, 0.00002873, -0.00002413, 0.00001446, -0.00000310, -0.00000866, 0.00000056, -0.00000988, 0.00000132, -0.00000905, 0.00000064, -0.00001011, -0.00001341, -0.00001839, -0.00003076, -0.00004359, -0.00048885, 0.00131245, -0.00053991, 0.00000075, -0.00005821, 0.00000389, -0.00003025, 0.00000417, -0.00001155, 0.00000226, -0.00000863, -0.00000322, 0.00000167, -0.00001551, 0.00001825, -0.00000988, -0.00000492, 0.00001116, -0.00002175, 0.00002298, -0.00002937, 0.00001680, -0.00000565, -0.00000598, 0.00000378, -0.00001159, 0.00001682, -0.00000442, -0.00000634, 0.00000882, -0.00000033, -0.00000524, 0.00000054, -0.00000579, 0.00000749, -0.00000484, 0.00001001, -0.00000769, 0.00000307, -0.00000549, 0.00000135, -0.00000350, 0.00001646, -0.00001279, 0.00000637, 0.00000628, -0.00001852, 0.00000476, 0.00000025, 0.00000663, -0.00000798, 0.00000520, -0.00000267, 0.00000110, -0.00000042, 0.00000193, -0.00000641, 0.00000739, -0.00000982, 0.00000424, 0.00000115, -0.00000541, 0.00000670, 0.00000684, },
		{ -0.00001113, 0.00000665, -0.00000898, -0.00000027, 0.00000724, 0.00000275, 0.00000088, -0.00000702, 0.00000208, -0.00000402, 0.00000276, 0.00000145, 0.00000260, -0.00000526, 0.00000072, 0.00000095, 0.00000107, -0.00000350, 0.00001200, -0.00000479, -0.00000962, -0.00000600, 0.00001177, -0.00000660, 0.00001705, -0.00001418, 0.00000961, -0.00001328, 0.00000597, -0.00000961, 0.00001537, -0.00001389, 0.00001702, -0.00001937, 0.00001318, -0.00001930, 0.00002525, -0.00002772, 0.00001651, -0.00001546, 0.00003239, -0.00003524, 0.00001402, -0.00000705, 0.00000157, 0.00000878, -0.00000744, -0.00000048, -0.00000091, 0.00000921, -0.00000706, -0.00000226, 0.00001552, -0.00003127, 0.00002714, -0.00004037, 0.00003648, -0.00004974, 0.00004833, -0.00006787, 0.00005457, -0.00010605, 0.00008570, -0.00024068, -0.00022649, 0.00125265, -0.00071616, 0.00016407, -0.00016305, 0.00008611, -0.00009621, 0.00006154, -0.00006034, 0.00004530, -0.00004754, 0.00003259, -0.00003134, 0.00001390, -0.00000519, 0.00000802, -0.00001993, 0.00002554, -0.00003568, 0.00003822, -0.00004825, 0.00004062, -0.00003068, 0.00001781, -0.00001740, 0.00000651, 0.00000347, 0.00000536, -0.00001577, 0.00001937, -0.00001235, 0.00000606, -0.00001029, 0.00000487, -0.00000191, 0.00000383, 0.00000216, -0.00000142, -0.00000294, -0.00000004, -0.00000399, 0.00000262, 0.00001176, -0.00001153, 0.00000877, 0.00000235, -0.00001853, 0.00000781, -0.00000098, 0.00000694, -0.00000938, 0.00000783, -0.00000547, 0.00000306, -0.00000159, 0.00000242, -0.00000627, 0.00000759, -0.00001126, 0.00000677, -0.00000185, -0.00000307, 0.00000796, 0.00000363, },
		{ -0.00000877, 0.00000625, -0.00001062, 0.00000315, 0.00000635, 0.00000165, 0.00000074, -0.00000709, 0.00000290, -0.00000488, 0.00000435, -0.00000042, 0.00000390, -0.00000737, 0.00000377, -0.00000137, 0.00000213, -0.00000380, 0.00001384, -0.00000941, -0.00000779, -0.00000429, 0.00001197, -0.00000844, 0.00001978, -0.00001859, 0.00001441, -0.00001866, 0.00001174, -0.00001435, 0.00002062, -0.00001980, 0.00002346, -0.00002714, 0.00002138, -0.00002718, 0.00003400, -0.00003848, 0.00002930, -0.00002758, 0.00004610, -0.00005522, 0.00003926, -0.00003329, 0.00002787, -0.00001614, 0.00001312, -0.00001994, 0.00001922, -0.00001111, 0.00001048, -0.00001831, 0.00003224, -0.00005111, 0.00005039, -0.00006702, 0.00006723, -0.00008544, 0.00009043, -0.00011984, 0.00011648, -0.00018761, 0.00019735, -0.00044991, 0.00010593, 0.00106924, -0.00078713, 0.00027109, -0.00023636, 0.00014595, -0.00014497, 0.00010510, -0.00009781, 0.00007860, -0.00007794, 0.00006099, -0.00005817, 0.00003844, -0.00002464, 0.00002186, -0.00003017, 0.00003451, -0.00004345, 0.00004621, -0.00005908, 0.00005705, -0.00005019, 0.00003825, -0.00003721, 0.00002457, -0.00001030, 0.00001512, -0.00002449, 0.00002863, -0.00002298, 0.00001626, -0.00002026, 0.00001502, -0.00001110, 0.00001247, -0.00000584, 0.00000505, -0.00000914, 0.00000566, -0.00000962, 0.00000939, 0.00000571, -0.00000844, 0.00000978, -0.00000184, -0.00001704, 0.00000970, -0.00000152, 0.00000618, -0.00000927, 0.00000901, -0.00000727, 0.00000436, -0.00000215, 0.00000233, -0.00000537, 0.00000664, -0.00001132, 0.00000799, -0.00000390, -0.00000082, 0.00000896, -0.00000033, },
		{ -0.00000509, 0.00000400, -0.00001049, 0.00000583, 0.00000514, 0.00000074, 0.00000077, -0.00000717, 0.00000355, -0.00000539, 0.00000534, -0.00000171, 0.00000459, -0.00000870, 0.00000622, -0.00000348, 0.00000295, -0.00000330, 0.00001427, -0.00001305, -0.00000625, -0.00000141, 0.00000954, -0.00000708, 0.00001920, -0.00001968, 0.00001589, -0.00002063, 0.00001416, -0.00001563, 0.00002206, -0.00002159, 0.00002548, -0.00003029, 0.00002488, -0.00003012, 0.00003734, -0.00004317, 0.00003558, -0.00003267, 0.00005164, -0.00006634, 0.00005647, -0.00005331, 0.00004967, -0.00003828, 0.00003160, -0.00003675, 0.00003613, -0.00002809, 0.00002438, -0.00002951, 0.00004256, -0.00006318, 0.00006472, -0.00008356, 0.00008649, -0.00010801, 0.00011739, -0.00015459, 0.00015836, -0.00024525, 0.00027952, -0.00063666, 0.00047095, 0.00078393, -0.00075169, 0.00030949, -0.00026543, 0.00017251, -0.00016707, 0.00012651, -0.00011659, 0.00009558, -0.00009386, 0.00007650, -0.00007371, 0.00005348, -0.00003610, 0.00002829, -0.00003274, 0.00003546, -0.00004267, 0.00004452, -0.00005890, 0.00006230, -0.00005974, 0.00005058, -0.00005108, 0.00003860, -0.00002118, 0.00002216, -0.00003020, 0.00003452, -0.00003014, 0.00002325, -0.00002731, 0.00002264, -0.00001820, 0.00001932, -0.00001238, 0.00001021, -0.00001416, 0.00001032, -0.00001429, 0.00001563, -0.00000068, -0.00000410, 0.00000939, -0.00000590, -0.00001426, 0.00001042, -0.00000137, 0.00000447, -0.00000772, 0.00000861, -0.00000784, 0.00000481, -0.00000200, 0.00000166, -0.00000385, 0.00000474, -0.00001016, 0.00000790, -0.00000486, 0.00000121, 0.00000945, -0.00000448, },
		{ -0.00000082, 0.00000041, -0.00000867, 0.00000753, 0.00000374, 0.00000014, 0.00000080, -0.00000716, 0.00000398, -0.00000552, 0.00000569, -0.00000228, 0.00000446, -0.00000896, 0.00000772, -0.00000506, 0.00000343, -0.00000203, 0.00001323, -0.00001553, -0.00000488, 0.00000211, 0.00000506, -0.00000280, 0.00001528, -0.00001723, 0.00001376, -0.00001882, 0.00001277, -0.00001296, 0.00001914, -0.00001867, 0.00002238, -0.00002801, 0.00002279, -0.00002721, 0.00003434, -0.00004083, 0.00003421, -0.00002934, 0.00004721, -0.00006599, 0.00006201, -0.00006268, 0.00006223, -0.00005298, 0.00004384, -0.00004721, 0.00004628, -0.00003825, 0.00003130, -0.00003264, 0.00004313, -0.00006379, 0.00006602, -0.00008541, 0.00008906, -0.00011143, 0.00012201, -0.00016332, 0.00016911, -0.00026410, 0.00030985, -0.00076528, 0.00082319, 0.00043303, -0.00062474, 0.00027990, -0.00024695, 0.00016198, -0.00015872, 0.00012216, -0.00011331, 0.00009316, -0.00009244, 0.00007639, -0.00007538, 0.00005664, -0.00003757, 0.00002583, -0.00002660, 0.00002766, -0.00003289, 0.00003284, -0.00004707, 0.00005491, -0.00005694, 0.00005173, -0.00005568, 0.00004544, -0.00002646, 0.00002428, -0.00003110, 0.00003546, -0.00003232, 0.00002550, -0.00002985, 0.00002616, -0.00002166, 0.00002292, -0.00001608, 0.00001280, -0.00001676, 0.00001277, -0.00001686, 0.00002013, -0.00000625, 0.00000062, 0.00000783, -0.00000955, -0.00001051, 0.00001006, -0.00000069, 0.00000213, -0.00000501, 0.00000678, -0.00000718, 0.00000441, -0.00000120, 0.00000053, -0.00000197, 0.00000224, -0.00000814, 0.00000679, -0.00000480, 0.00000301, 0.00000918, -0.00000814, },
		{ 0.00000312, -0.00000377, -0.00000553, 0.00000822, 0.00000224, 0.00000000, 0.00000058, -0.00000688, 0.00000410, -0.00000526, 0.00000540, -0.00000208, 0.00000346, -0.00000805, 0.00000802, -0.00000587, 0.00000351, -0.00000015, 0.00001086, -0.00001686, -0.00000346, 0.00000548, -0.00000046, 0.00000350, 0.00000872, -0.00001184, 0.00000851, -0.00001368, 0.00000799, -0.00000677, 0.00001234, -0.00001148, 0.00001455, -0.00002061, 0.00001539, -0.00001875, 0.00002542, -0.00003196, 0.00002575, -0.00001814, 0.00003318, -0.00005392, 0.00005466, -0.00005923, 0.00006271, -0.00005716, 0.00004702, -0.00004893, 0.00004752, -0.00003961, 0.00002953, -0.00002627, 0.00003265, -0.00005159, 0.00005277, -0.00007086, 0.00007299, -0.00009337, 0.00010133, -0.00014204, 0.00014320, -0.00023584, 0.00027314, -0.00080544, 0.00111628, 0.00006176, -0.00043408, 0.00019499, -0.00018714, 0.00011817, -0.00012244, 0.00009378, -0.00008923, 0.00007233, -0.00007447, 0.00006126, -0.00006355, 0.00004819, -0.00002946, 0.00001524, -0.00001283, 0.00001248, -0.00001582, 0.00001319, -0.00002551, 0.00003614, -0.00004210, 0.00004106, -0.00004966, 0.00004353, -0.00002474, 0.00002033, -0.00002634, 0.00003080, -0.00002893, 0.00002237, -0.00002720, 0.00002480, -0.00002070, 0.00002248, -0.00001619, 0.00001207, -0.00001622, 0.00001228, -0.00001652, 0.00002199, -0.00000998, 0.00000482, 0.00000550, -0.00001258, -0.00000611, 0.00000880, 0.00000033, -0.00000046, -0.00000164, 0.00000388, -0.00000552, 0.00000331, 0.00000010, -0.00000086, -0.00000002, -0.00000043, -0.00000572, 0.00000506, -0.00000395, 0.00000461, 0.00000789, -0.00001062, },
		{ 0.00000588, -0.00000763, -0.00000169, 0.00000804, 0.00000064, 0.00000047, -0.00000011, -0.00000615, 0.00000381, -0.00000461, 0.00000453, -0.00000121, 0.00000168, -0.00000600, 0.00000704, -0.00000577, 0.00000316, 0.00000214, 0.00000742, -0.00001713, -0.00000175, 0.00000792, -0.00000575, 0.00001047, 0.00000080, -0.00000469, 0.00000133, -0.00000640, 0.00000100, 0.00000169, 0.00000300, -0.00000142, 0.00000344, -0.00000955, 0.00000413, -0.00000625, 0.00001226, -0.00001849, 0.00001231, -0.00000143, 0.00001207, -0.00003241, 0.00003593, -0.00004353, 0.00005083, -0.00005001, 0.00004035, -0.00004143, 0.00003966, -0.00003218, 0.00001942, -0.00001125, 0.00001235, -0.00002801, 0.00002659, -0.00004167, 0.00004021, -0.00005594, 0.00005753, -0.00009288, 0.00008243, -0.00016120, 0.00016524, -0.00073798, 0.00131024, -0.00028281, -0.00021519, 0.00007621, -0.00009997, 0.00005136, -0.00006631, 0.00004790, -0.00004984, 0.00003784, -0.00004412, 0.00003477, -0.00004141, 0.00003091, -0.00001440, -0.00000073, 0.00000559, -0.00000692, 0.00000505, -0.00001049, 0.00000166, 0.00000974, -0.00001819, 0.00002051, -0.00003406, 0.00003331, -0.00001625, 0.00001060, -0.00001630, 0.00002108, -0.00002052, 0.00001434, -0.00001971, 0.00001882, -0.00001545, 0.00001807, -0.00001272, 0.00000802, -0.00001248, 0.00000872, -0.00001307, 0.00002077, -0.00001116, 0.00000771, 0.00000288, -0.00001490, -0.00000142, 0.00000689, 0.00000144, -0.00000286, 0.00000184, 0.00000042, -0.00000320, 0.00000179, 0.00000166, -0.00000230, 0.00000171, -0.00000288, -0.00000336, 0.00000318, -0.00000260, 0.00000604, 0.00000551, -0.00001143, },
		{ 0.00000688, -0.00001037, 0.00000219, 0.00000723, -0.00000100, 0.00000156, -0.00000143, -0.00000483, 0.00000302, -0.00000360, 0.00000321, 0.00000018, -0.00000066, -0.00000304, 0.00000488, -0.00000472, 0.00000244, 0.00000457, 0.00000321, -0.00001644, 0.00000038, 0.00000877, -0.00000962, 0.00001666, -0.00000695, 0.00000267, -0.00000620, 0.00000140, -0.00000650, 0.00001060, -0.00000696, 0.00000943, -0.00000879, 0.00000294, -0.00000871, 0.00000793, -0.00000259, -0.00000325, -0.00000288, 0.00001710, -0.00001197, -0.00000578, 0.00000980, -0.00001881, 0.00002898, -0.00003321, 0.00002525, -0.00002631, 0.00002452, -0.00001800, 0.00000344, 0.00000936, -0.00001413, 0.00000279, -0.00000795, -0.00000294, -0.00000354, -0.00000564, -0.00000206, -0.00002408, -0.00000381, -0.00005076, -0.00000481, -0.00055925, 0.00137808, -0.00055925, -0.00000481, -0.00005076, -0.00000381, -0.00002408, -0.00000206, -0.00000564, -0.00000354, -0.00000294, -0.00000795, 0.00000279, -0.00001413, 0.00000936, 0.00000344, -0.00001800, 0.00002452, -0.00002631, 0.00002525, -0.00003321, 0.00002898, -0.00001881, 0.00000980, -0.00000578, -0.00001197, 0.00001710, -0.00000288, -0.00000325, -0.00000259, 0.00000793, -0.00000871, 0.00000294, -0.00000879, 0.00000943, -0.00000696, 0.00001060, -0.00000650, 0.00000140, -0.00000620, 0.00000267, -0.00000695, 0.00001666, -0.00000962, 0.00000877, 0.00000038, -0.00001644, 0.00000321, 0.00000457, 0.00000244, -0.00000472, 0.00000488, -0.00000304, -0.00000066, 0.00000018, 0.00000321, -0.00000360, 0.00000302, -0.00000483, -0.00000143, 0.00000156, -0.00000100, 0.00000723, 0.00000219, -0.00001037, },
		{ 0.00000588, -0.00001143, 0.00000551, 0.00000604, -0.00000260, 0.00000318, -0.00000336, -0.00000288, 0.00000171, -0.00000230, 0.00000166, 0.00000179, -0.00000320, 0.00000042, 0.00000184, -0.00000286, 0.00000144, 0.00000689, -0.00000142, -0.00001490, 0.00000288, 0.00000771, -0.00001116, 0.00002077, -0.00001307, 0.00000872, -0.00001248, 0.00000802, -0.00001272, 0.00001807, -0.00001545, 0.00001882, -0.00001971, 0.00001434, -0.00002052, 0.00002108, -0.00001630, 0.00001060, -0.00001625, 0.00003331, -0.00003406, 0.00002051, -0.00001819, 0.00000974, 0.00000166, -0.00001049, 0.00000505, -0.00000692, 0.00000559, -0.00000073, -0.00001440, 0.00003091, -0.00004141, 0.00003477, -0.00004412, 0.00003784, -0.00004984, 0.00004790, -0.00006631, 0.00005136, -0.00009997, 0.00007621, -0.00021519, -0.00028281, 0.00131024, -0.00073798, 0.00016524, -0.00016120, 0.00008243, -0.00009288, 0.00005753, -0.00005594, 0.00004021, -0.00004167, 0.00002659, -0.00002801, 0.00001235, -0.00001125, 0.00001942, -0.00003218, 0.00003966, -0.00004143, 0.00004035, -0.00005001, 0.00005083, -0.00004353, 0.00003593, -0.00003241, 0.00001207, -0.00000143, 0.00001231, -0.00001849, 0.00001226, -0.00000625, 0.00000413, -0.00000955, 0.00000344, -0.00000142, 0.00000300, 0.00000169, 0.00000100, -0.00000640, 0.00000133, -0.00000469, 0.00000080, 0.00001047, -0.00000575, 0.00000792, -0.00000175, -0.00001713, 0.00000742, 0.00000214, 0.00000316, -0.00000577, 0.00000704, -0.00000600, 0.00000168, -0.00000121, 0.00000453, -0.00000461, 0.00000381, -0.00000615, -0.00000011, 0.00000047, 0.00000064, 0.00000804, -0.00000169, -0.00000763, },
		{ 0.00000312, -0.00001062, 0.00000789, 0.00000461, -0.00000395, 0.00000506, -0.00000572, -0.00000043, -0.00000002, -0.00000086, 0.00000010, 0.00000331, -0.00000552, 0.00000388, -0.00000164, -0.00000046, 0.00000033, 0.00000880, -0.00000611, -0.00001258, 0.00000550, 0.00000482, -0.00000998, 0.00002199, -0.00001652, 0.00001228, -0.00001622, 0.00001207, -0.00001619, 0.00002248, -0.00002070, 0.00002480, -0.00002720, 0.00002237, -0.00002893, 0.00003080, -0.00002634, 0.00002033, -0.00002474, 0.00004353, -0.00004966, 0.00004106, -0.00004210, 0.00003614, -0.00002551, 0.00001319, -0.00001582, 0.00001248, -0.00001283, 0.00001524, -0.00002946, 0.00004819, -0.00006355, 0.00006126, -0.00007447, 0.00007233, -0.00008923, 0.00009378, -0.00012244, 0.00011817, -0.00018714, 0.00019499, -0.00043408, 0.00006176, 0.00111628, -0.00080544, 0.00027314, -0.00023584, 0.00014320, -0.00014204, 0.00010133, -0.00009337, 0.00007299, -0.00007086, 0.00005277, -0.00005159, 0.00003265, -0.00002627, 0.00002953, -0.00003961, 0.00004752, -0.00004893, 0.00004702, -0.00005716, 0.00006271, -0.00005923, 0.00005466, -0.00005392, 0.00003318, -0.00001814, 0.00002575, -0.00003196, 0.00002542, -0.00001875, 0.00001539, -0.00002061, 0.00001455, -0.00001148, 0.00001234, -0.00000677, 0.00000799, -0.00001368, 0.00000851, -0.00001184, 0.00000872, 0.00000350, -0.00000046, 0.00000548, -0.00000346, -0.00001686, 0.00001086, -0.00000015, 0.00000351, -0.00000587, 0.00000802, -0.00000805, 0.00000346, -0.00000208, 0.00000540, -0.00000526, 0.00000410, -0.00000688, 0.00000058, 0.00000000, 0.00000224, 0.00000822, -0.00000553, -0.00000377, },
		{ -0.00000082, -0.00000814, 0.00000918, 0.00000301, -0.00000480, 0.00000679, -0.00000814, 0.00000224, -0.00000197, 0.00000053, -0.00000120, 0.00000441, -0.00000718, 0.00000678, -0.00000501, 0.00000213, -0.00000069, 0.00001006, -0.00001051, -0.00000955, 0.00000783, 0.00000062, -0.00000625, 0.00002013, -0.00001686, 0.00001277, -0.00001676, 0.00001280, -0.00001608, 0.00002292, -0.00002166, 0.00002616, -0.00002985, 0.00002550, -0.00003232, 0.00003546, -0.00003110, 0.00002428, -0.00002646, 0.00004544, -0.00005568, 0.00005173, -0.00005694, 0.00005491, -0.00004707, 0.00003284, -0.00003289, 0.00002766, -0.00002660, 0.00002583, -0.00003757, 0.00005664, -0.00007538, 0.00007639, -0.00009244, 0.00009316, -0.00011331, 0.00012216, -0.00015872, 0.00016198, -0.00024695, 0.00027990, -0.00062474, 0.00043303, 0.00082319, -0.00076528, 0.00030985, -0.00026410, 0.00016911, -0.00016332, 0.00012201, -0.00011143, 0.00008906, -0.00008541, 0.00006602, -0.00006379, 0.00004313, -0.00003264, 0.00003130, -0.00003825, 0.00004628, -0.00004721, 0.00004384, -0.00005298, 0.00006223, -0.00006268, 0.00006201, -0.00006599, 0.00004721, -0.00002934, 0.00003421, -0.00004083, 0.00003434, -0.00002721, 0.00002279, -0.00002801, 0.00002238, -0.00001867, 0.00001914, -0.00001296, 0.00001277, -0.00001882, 0.00001376, -0.00001723, 0.00001528, -0.00000280, 0.00000506, 0.00000211, -0.00000488, -0.00001553, 0.00001323, -0.00000203, 0.00000343, -0.00000506, 0.00000772, -0.00000896, 0.00000446, -0.00000228, 0.00000569, -0.00000552, 0.00000398, -0.00000716, 0.00000080, 0.00000014, 0.00000374, 0.00000753, -0.00000867, 0.00000041, },
		{ -0.00000509, -0.00000448, 0.00000945, 0.00000121, -0.00000486, 0.00000790, -0.00001016, 0.00000474, -0.00000385, 0.00000166, -0.00000200, 0.00000481, -0.00000784, 0.00000861, -0.00000772, 0.00000447, -0.00000137, 0.00001042, -0.00001426, -0.00000590, 0.00000939, -0.00000410, -0.00000068, 0.00001563, -0.00001429, 0.00001032, -0.00001416, 0.00001021, -0.00001238, 0.00001932, -0.00001820, 0.00002264, -0.00002731, 0.00002325, -0.00003014, 0.00003452, -0.00003020, 0.00002216, -0.00002118, 0.00003860, -0.00005108, 0.00005058, -0.00005974, 0.00006230, -0.00005890, 0.00004452, -0.00004267, 0.00003546, -0.00003274, 0.00002829, -0.00003610, 0.00005348, -0.00007371, 0.00007650, -0.00009386, 0.00009558, -0.00011659, 0.00012651, -0.00016707, 0.00017251, -0.00026543, 0.00030949, -0.00075169, 0.00078393, 0.00047095, -0.00063666, 0.00027952, -0.00024525, 0.00015836, -0.00015459, 0.00011739, -0.00010801, 0.00008649, -0.00008356, 0.00006472, -0.00006318, 0.00004256, -0.00002951, 0.00002438, -0.00002809, 0.00003613, -0.00003675, 0.00003160, -0.00003828, 0.00004967, -0.00005331, 0.00005647, -0.00006634, 0.00005164, -0.00003267, 0.00003558, -0.00004317, 0.00003734, -0.00003012, 0.00002488, -0.00003029, 0.00002548, -0.00002159, 0.00002206, -0.00001563, 0.00001416, -0.00002063, 0.00001589, -0.00001968, 0.00001920, -0.00000708, 0.00000954, -0.00000141, -0.00000625, -0.00001305, 0.00001427, -0.00000330, 0.00000295, -0.00000348, 0.00000622, -0.00000870, 0.00000459, -0.00000171, 0.00000534, -0.00000539, 0.00000355, -0.00000717, 0.00000077, 0.00000074, 0.00000514, 0.00000583, -0.00001049, 0.00000400, },
		{ -0.00000877, -0.00000033, 0.00000896, -0.00000082, -0.00000390, 0.00000799, -0.00001132, 0.00000664, -0.00000537, 0.00000233, -0.00000215, 0.00000436, -0.00000727, 0.00000901, -0.00000927, 0.00000618, -0.00000152, 0.00000970, -0.00001704, -0.00000184, 0.00000978, -0.00000844, 0.00000571, 0.00000939, -0.00000962, 0.00000566, -0.00000914, 0.00000505, -0.00000584, 0.00001247, -0.00001110, 0.00001502, -0.00002026, 0.00001626, -0.00002298, 0.00002863, -0.00002449, 0.00001512, -0.00001030, 0.00002457, -0.00003721, 0.00003825, -0.00005019, 0.00005705, -0.00005908, 0.00004621, -0.00004345, 0.00003451, -0.00003017, 0.00002186, -0.00002464, 0.00003844, -0.00005817, 0.00006099, -0.00007794, 0.00007860, -0.00009781, 0.00010510, -0.00014497, 0.00014595, -0.00023636, 0.00027109, -0.00078713, 0.00106924, 0.00010593, -0.00044991, 0.00019735, -0.00018761, 0.00011648, -0.00011984, 0.00009043, -0.00008544, 0.00006723, -0.00006702, 0.00005039, -0.00005111, 0.00003224, -0.00001831, 0.00001048, -0.00001111, 0.00001922, -0.00001994, 0.00001312, -0.00001614, 0.00002787, -0.00003329, 0.00003926, -0.00005522, 0.00004610, -0.00002758, 0.00002930, -0.00003848, 0.00003400, -0.00002718, 0.00002138, -0.00002714, 0.00002346, -0.00001980, 0.00002062, -0.00001435, 0.00001174, -0.00001866, 0.00001441, -0.00001859, 0.00001978, -0.00000844, 0.00001197, -0.00000429, -0.00000779, -0.00000941, 0.00001384, -0.00000380, 0.00000213, -0.00000137, 0.00000377, -0.00000737, 0.00000390, -0.00000042, 0.00000435, -0.00000488, 0.00000290, -0.00000709, 0.00000074, 0.00000165, 0.00000635, 0.00000315, -0.00001062, 0.00000625, },
		{ -0.00001113, 0.00000363, 0.00000796, -0.00000307, -0.00000185, 0.00000677, -0.00001126, 0.00000759, -0.00000627, 0.00000242, -0.00000159, 0.00000306, -0.00000547, 0.00000783, -0.00000938, 0.00000694, -0.00000098, 0.00000781, -0.00001853, 0.00000235, 0.00000877, -0.00001153, 0.00001176, 0.00000262, -0.00000399, -0.00000004, -0.00000294, -0.00000142, 0.00000216, 0.00000383, -0.00000191, 0.00000487, -0.00001029, 0.00000606, -0.00001235, 0.00001937, -0.00001577, 0.00000536, 0.00000347, 0.00000651, -0.00001740, 0.00001781, -0.00003068, 0.00004062, -0.00004825, 0.00003822, -0.00003568, 0.00002554, -0.00001993, 0.00000802, -0.00000519, 0.00001390, -0.00003134, 0.00003259, -0.00004754, 0.00004530, -0.00006034, 0.00006154, -0.00009621, 0.00008611, -0.00016305, 0.00016407, -0.00071616, 0.00125265, -0.00022649, -0.00024068, 0.00008570, -0.00010605, 0.00005457, -0.00006787, 0.00004833, -0.00004974, 0.00003648, -0.00004037, 0.00002714, -0.00003127, 0.00001552, -0.00000226, -0.00000706, 0.00000921, -0.00000091, -0.00000048, -0.00000744, 0.00000878, 0.00000157, -0.00000705, 0.00001402, -0.00003524, 0.00003239, -0.00001546, 0.00001651, -0.00002772, 0.00002525, -0.00001930, 0.00001318, -0.00001937, 0.00001702, -0.00001389, 0.00001537, -0.00000961, 0.00000597, -0.00001328, 0.00000961, -0.00001418, 0.00001705, -0.00000660, 0.00001177, -0.00000600, -0.00000962, -0.00000479, 0.00001200, -0.00000350, 0.00000107, 0.00000095, 0.00000072, -0.00000526, 0.00000260, 0.00000145, 0.00000276, -0.00000402, 0.00000208, -0.00000702, 0.00000088, 0.00000275, 0.00000724, -0.00000027, -0.00000898, 0.00000665, },
		{ -0.00001169, 0.00000684, 0.00000670, -0.00000541, 0.00000115, 0.00000424, -0.00000982, 0.00000739, -0.00000641, 0.00000193, -0.00000042, 0.00000110, -0.00000267, 0.00000520, -0.00000798, 0.00000663, 0.00000025, 0.00000476, -0.00001852, 0.00000628, 0.00000637, -0.00001279, 0.00001646, -0.00000350, 0.00000135, -0.00000549, 0.00000307, -0.00000769, 0.00001001, -0.00000484, 0.00000749, -0.00000579, 0.00000054, -0.00000524, -0.00000033, 0.00000882, -0.00000634, -0.00000442, 0.00001682, -0.00001159, 0.00000378, -0.00000598, -0.00000565, 0.00001680, -0.00002937, 0.00002298, -0.00002175, 0.00001116, -0.00000492, -0.00000988, 0.00001825, -0.00001551, 0.00000167, -0.00000322, -0.00000863, 0.00000226, -0.00001155, 0.00000417, -0.00003025, 0.00000389, -0.00005821, 0.00000075, -0.00053991, 0.00131245, -0.00048885, -0.00004359, -0.00003076, -0.00001839, -0.00001341, -0.00001011, 0.00000064, -0.00000905, 0.00000132, -0.00000988, 0.00000056, -0.00000866, -0.00000310, 0.00001446, -0.00002413, 0.00002873, -0.00002017, 0.00001749, -0.00002553, 0.00003133, -0.00002364, 0.00001981, -0.00001406, -0.00001072, 0.00001398, 0.00000080, -0.00000028, -0.00001308, 0.00001311, -0.00000846, 0.00000217, -0.00000874, 0.00000776, -0.00000533, 0.00000765, -0.00000268, -0.00000192, -0.00000565, 0.00000252, -0.00000737, 0.00001176, -0.00000203, 0.00000898, -0.00000637, -0.00001164, 0.00000049, 0.00000895, -0.00000246, -0.00000006, 0.00000313, -0.00000249, -0.00000271, 0.00000097, 0.00000368, 0.00000070, -0.00000281, 0.00000108, -0.00000696, 0.00000130, 0.00000393, 0.00000760, -0.00000396, -0.00000589, 0.00000511, },
		{ -0.00001038, 0.00000898, 0.00000528, -0.00000757, 0.00000473, 0.00000063, -0.00000711, 0.00000603, -0.00000582, 0.00000098, 0.00000112, -0.00000119, 0.00000069, 0.00000151, -0.00000528, 0.00000533, 0.00000205, 0.00000074, -0.00001689, 0.00000950, 0.00000290, -0.00001200, 0.00001911, -0.00000804, 0.00000538, -0.00000956, 0.00000764, -0.00001240, 0.00001620, -0.00001189, 0.00001526, -0.00001499, 0.00001013, -0.00001550, 0.00001095, -0.00000094, 0.00000162, -0.00001169, 0.00002667, -0.00002585, 0.00002160, -0.00002779, 0.00001938, -0.00000923, -0.00000690, 0.00000438, -0.00000537, -0.00000481, 0.00001082, -0.00002736, 0.00004054, -0.00004393, 0.00003437, -0.00003931, 0.00003099, -0.00004185, 0.00003872, -0.00005568, 0.00003964, -0.00008473, 0.00005811, -0.00019509, -0.00027593, 0.00124461, -0.00065681, 0.00011334, -0.00012988, 0.00005839, -0.00007375, 0.00004197, -0.00004288, 0.00002824, -0.00003092, 0.00001797, -0.00002358, 0.00001157, -0.00001907, 0.00002777, -0.00003683, 0.00004362, -0.00003490, 0.00003044, -0.00003734, 0.00004700, -0.00004262, 0.00004169, -0.00003940, 0.00001331, -0.00000484, 0.00001744, -0.00001773, 0.00000253, 0.00000018, 0.00000287, -0.00000924, 0.00000243, -0.00000214, 0.00000387, -0.00000068, 0.00000466, -0.00001021, 0.00000260, -0.00000533, 0.00000039, 0.00000526, 0.00000420, 0.00000416, -0.00000560, -0.00001355, 0.00000589, 0.00000512, -0.00000089, -0.00000110, 0.00000486, -0.00000543, -0.00000014, -0.00000065, 0.00000595, -0.00000165, -0.00000130, -0.00000016, -0.00000684, 0.00000201, 0.00000515, 0.00000718, -0.00000731, -0.00000199, 0.00000198, },
		{ -0.00000750, 0.00001000, 0.00000368, -0.00000917, 0.00000832, -0.00000354, -0.00000351, 0.00000376, -0.00000465, -0.00000018, 0.00000270, -0.00000335, 0.00000400, -0.00000262, -0.00000174, 0.00000331, 0.00000411, -0.00000388, -0.00001371, 0.00001161, -0.00000115, -0.00000932, 0.00001947, -0.00001047, 0.00000745, -0.00001152, 0.00000993, -0.00001460, 0.00001965, -0.00001608, 0.00002001, -0.00002115, 0.00001679, -0.00002294, 0.00001977, -0.00000832, 0.00000657, -0.00001477, 0.00003090, -0.00003339, 0.00003225, -0.00004289, 0.00003911, -0.00003212, 0.00001421, -0.00001322, 0.00000937, -0.00001831, 0.00002311, -0.00003996, 0.00005662, -0.00006553, 0.00006021, -0.00006850, 0.00006337, -0.00007811, 0.00008027, -0.00010605, 0.00009917, -0.00016202, 0.00016229, -0.00039195, 0.00004443, 0.00106285, -0.00072186, 0.00021263, -0.00019577, 0.00011142, -0.00011569, 0.00007916, -0.00007427, 0.00005525, -0.00005424, 0.00003790, -0.00004060, 0.00002533, -0.00002883, 0.00003459, -0.00004236, 0.00005120, -0.00004263, 0.00003615, -0.00004056, 0.00005294, -0.00005175, 0.00005425, -0.00005726, 0.00003226, -0.00001994, 0.00003077, -0.00003252, 0.00001617, -0.00001089, 0.00001221, -0.00001866, 0.00001187, -0.00001053, 0.00001168, -0.00000771, 0.00001058, -0.00001710, 0.00000972, -0.00001224, 0.00000746, -0.00000084, 0.00001063, -0.00000167, -0.00000419, -0.00001488, 0.00001078, 0.00000106, 0.00000094, -0.00000186, 0.00000586, -0.00000771, 0.00000206, -0.00000191, 0.00000794, -0.00000405, 0.00000041, -0.00000166, -0.00000654, 0.00000294, 0.00000637, 0.00000581, -0.00000969, 0.00000191, -0.00000207, },
		{ -0.00000362, 0.00001006, 0.00000180, -0.00000981, 0.00001128, -0.00000762, 0.00000044, 0.00000095, -0.00000321, -0.00000126, 0.00000396, -0.00000493, 0.00000665, -0.00000647, 0.00000204, 0.00000102, 0.00000601, -0.00000861, -0.00000927, 0.00001230, -0.00000515, -0.00000526, 0.00001774, -0.00001075, 0.00000741, -0.00001116, 0.00000965, -0.00001393, 0.00001990, -0.00001686, 0.00002107, -0.00002346, 0.00001956, -0.00002651, 0.00002511, -0.00001252, 0.00000794, -0.00001316, 0.00002881, -0.00003305, 0.00003371, -0.00004825, 0.00004961, -0.00004746, 0.00002971, -0.00002603, 0.00001903, -0.00002602, 0.00002865, -0.00004424, 0.00006263, -0.00007587, 0.00007411, -0.00008507, 0.00008213, -0.00009933, 0.00010473, -0.00013699, 0.00013608, -0.00021213, 0.00023176, -0.00055604, 0.00038089, 0.00079581, -0.00069109, 0.00024891, -0.00022099, 0.00013398, -0.00013331, 0.00009619, -0.00008889, 0.00006792, -0.00006510, 0.00004684, -0.00004784, 0.00003032, -0.00003055, 0.00003351, -0.00003954, 0.00005042, -0.00004254, 0.00003419, -0.00003494, 0.00004852, -0.00004972, 0.00005533, -0.00006467, 0.00004292, -0.00002829, 0.00003794, -0.00004199, 0.00002551, -0.00001808, 0.00001769, -0.00002427, 0.00001779, -0.00001572, 0.00001650, -0.00001193, 0.00001364, -0.00002111, 0.00001426, -0.00001681, 0.00001239, -0.00000502, 0.00001574, -0.00000733, -0.00000281, -0.00001514, 0.00001450, -0.00000264, 0.00000266, -0.00000218, 0.00000596, -0.00000903, 0.00000359, -0.00000249, 0.00000932, -0.00000623, 0.00000219, -0.00000344, -0.00000592, 0.00000400, 0.00000751, 0.00000345, -0.00001061, 0.00000497, -0.00000619, },
		{ 0.00000056, 0.00000946, -0.00000044, -0.00000916, 0.00001300, -0.00001090, 0.00000410, -0.00000191, -0.00000183, -0.00000197, 0.00000460, -0.00000553, 0.00000809, -0.00000935, 0.00000537, -0.00000103, 0.00000728, -0.00001282, -0.00000405, 0.00001150, -0.00000850, -0.00000050, 0.00001444, -0.00000927, 0.00000560, -0.00000879, 0.00000710, -0.00001064, 0.00001717, -0.00001439, 0.00001852, -0.00002189, 0.00001832, -0.00002600, 0.00002678, -0.00001362, 0.00000618, -0.00000762, 0.00002130, -0.00002554, 0.00002614, -0.00004311, 0.00004914, -0.00005275, 0.00003696, -0.00003169, 0.00002159, -0.00002615, 0.00002584, -0.00003868, 0.00005688, -0.00007280, 0.00007346, -0.00008600, 0.00008383, -0.00010153, 0.00010739, -0.00014262, 0.00014281, -0.00022462, 0.00024959, -0.00065716, 0.00069075, 0.00048180, -0.00058452, 0.00022823, -0.00020680, 0.00012610, -0.00012615, 0.00009237, -0.00008594, 0.00006555, -0.00006293, 0.00004436, -0.00004501, 0.00002644, -0.00002436, 0.00002498, -0.00002903, 0.00004198, -0.00003552, 0.00002592, -0.00002220, 0.00003538, -0.00003770, 0.00004524, -0.00006101, 0.00004410, -0.00002855, 0.00003758, -0.00004475, 0.00002928, -0.00002036, 0.00001839, -0.00002519, 0.00001932, -0.00001682, 0.00001750, -0.00001256, 0.00001306, -0.00002147, 0.00001537, -0.00001817, 0.00001429, -0.00000624, 0.00001829, -0.00001172, -0.00000208, -0.00001392, 0.00001652, -0.00000542, 0.00000398, -0.00000198, 0.00000508, -0.00000926, 0.00000423, -0.00000219, 0.00000980, -0.00000789, 0.00000381, -0.00000539, -0.00000492, 0.00000515, 0.00000841, 0.00000024, -0.00000983, 0.00000651, -0.00000950, },
		{ 0.00000437, 0.00000848, -0.00000308, -0.00000707, 0.00001307, -0.00001282, 0.00000688, -0.00000435, -0.00000079, -0.00000208, 0.00000444, -0.00000498, 0.00000798, -0.00001072, 0.00000767, -0.00000237, 0.00000746, -0.00001588, 0.00000134, 0.00000933, -0.00001074, 0.00000421, 0.00001031, -0.00000668, 0.00000266, -0.00000512, 0.00000302, -0.00000551, 0.00001226, -0.00000950, 0.00001318, -0.00001720, 0.00001374, -0.00002196, 0.00002534, -0.00001240, 0.00000252, 0.00000011, 0.00001057, -0.00001329, 0.00001179, -0.00002912, 0.00003841, -0.00004785, 0.00003537, -0.00002975, 0.00001695, -0.00001888, 0.00001515, -0.00002401, 0.00004016, -0.00005700, 0.00005874, -0.00007161, 0.00006869, -0.00008484, 0.00008816, -0.00012247, 0.00011811, -0.00019671, 0.00020815, -0.00067420, 0.00093566, 0.00016237, -0.00043037, 0.00016535, -0.00016174, 0.00009373, -0.00009872, 0.00007121, -0.00006828, 0.00005057, -0.00004992, 0.00003248, -0.00003405, 0.00001556, -0.00001218, 0.00001109, -0.00001302, 0.00002803, -0.00002378, 0.00001398, -0.00000555, 0.00001694, -0.00001891, 0.00002660, -0.00004798, 0.00003672, -0.00002125, 0.00002995, -0.00004083, 0.00002753, -0.00001787, 0.00001455, -0.00002164, 0.00001661, -0.00001399, 0.00001478, -0.00000972, 0.00000898, -0.00001819, 0.00001301, -0.00001623, 0.00001299, -0.00000411, 0.00001759, -0.00001408, -0.00000243, -0.00001104, 0.00001652, -0.00000689, 0.00000465, -0.00000124, 0.00000332, -0.00000840, 0.00000397, -0.00000092, 0.00000922, -0.00000876, 0.00000502, -0.00000737, -0.00000354, 0.00000633, 0.00000887, -0.00000343, -0.00000745, 0.00000620, -0.00001134, },
		{ 0.00000733, 0.00000733, -0.00000597, -0.00000365, 0.00001135, -0.00001306, 0.00000837, -0.00000604, -0.00000030, -0.00000154, 0.00000350, -0.00000333, 0.00000628, -0.00001033, 0.00000858, -0.00000268, 0.00000629, -0.00001728, 0.00000620, 0.00000615, -0.00001165, 0.00000826, 0.00000609, -0.00000374, -0.00000059, -0.00000103, -0.00000163, 0.00000041, 0.00000627, -0.00000335, 0.00000628, -0.00001062, 0.00000701, -0.00001550, 0.00002181, -0.00001006, -0.00000140, 0.00000779, -0.00000052, 0.00000026, -0.00000558, -0.00000983, 0.00002032, -0.00003478, 0.00002641, -0.00002166, 0.00000681, -0.00000623, -0.00000103, -0.00000309, 0.00001565, -0.00003177, 0.00003328, -0.00004536, 0.00004042, -0.00005333, 0.00005154, -0.00008145, 0.00006736, -0.00013413, 0.00011090, -0.00059886, 0.00108744, -0.00012428, -0.00025948, 0.00007981, -0.00009890, 0.00004674, -0.00005891, 0.00003921, -0.00004135, 0.00002767, -0.00003021, 0.00001498, -0.00001842, 0.00000093, 0.00000287, -0.00000503, 0.00000537, 0.00001154, -0.00001023, 0.00000161, 0.00001115, -0.00000246, 0.00000212, 0.00000366, -0.00002900, 0.00002344, -0.00000858, 0.00001684, -0.00003163, 0.00002144, -0.00001182, 0.00000737, -0.00001478, 0.00001076, -0.00000821, 0.00000931, -0.00000436, 0.00000231, -0.00001211, 0.00000790, -0.00001165, 0.00000911, 0.00000092, 0.00001366, -0.00001415, -0.00000392, -0.00000665, 0.00001454, -0.00000688, 0.00000459, -0.00000006, 0.00000089, -0.00000663, 0.00000293, 0.00000121, 0.00000751, -0.00000867, 0.00000559, -0.00000914, -0.00000186, 0.00000755, 0.00000861, -0.00000702, -0.00000392, 0.00000411, -0.00001138, },
		{ 0.00000918, 0.00000608, -0.00000881, 0.00000070, 0.00000805, -0.00001158, 0.00000842, -0.00000682, -0.00000036, -0.00000045, 0.00000198, -0.00000088, 0.00000324, -0.00000827, 0.00000802, -0.00000191, 0.00000371, -0.00001671, 0.00000989, 0.00000248, -0.00001126, 0.00001125, 0.00000236, -0.00000111, -0.00000342, 0.00000260, -0.00000586, 0.00000603, 0.00000039, 0.00000276, -0.00000080, -0.00000360, -0.00000043, -0.00000792, 0.00001733, -0.00000778, -0.00000407, 0.00001333, -0.00000915, 0.00001149, -0.00002169, 0.00001020, -0.00000084, -0.00001714, 0.00001308, -0.00001026, -0.00000582, 0.00000847, -0.00001893, 0.00001979, -0.00001184, -0.00000226, 0.00000252, -0.00001307, 0.00000533, -0.00001404, 0.00000552, -0.00002875, 0.00000137, -0.00005001, -0.00002782, -0.00043723, 0.00113205, -0.00034877, -0.00010003, -0.00000837, -0.00003270, -0.00000358, -0.00001599, 0.00000411, -0.00001172, 0.00000254, -0.00000883, -0.00000365, -0.00000220, -0.00001372, 0.00001731, -0.00001996, 0.00002280, -0.00000440, 0.00000232, -0.00000823, 0.00002435, -0.00001864, 0.00002074, -0.00001881, -0.00000842, 0.00000794, 0.00000627, 0.00000102, -0.00001943, 0.00001297, -0.00000409, -0.00000126, -0.00000641, 0.00000346, -0.00000113, 0.00000262, 0.00000203, -0.00000546, -0.00000463, 0.00000133, -0.00000568, 0.00000387, 0.00000776, 0.00000718, -0.00001219, -0.00000625, -0.00000123, 0.00001091, -0.00000552, 0.00000384, 0.00000134, -0.00000189, -0.00000427, 0.00000138, 0.00000396, 0.00000482, -0.00000756, 0.00000532, -0.00001048, 0.00000002, 0.00000881, 0.00000741, -0.00000991, 0.00000007, 0.00000067, -0.00000966, },
		{ 0.00000997, 0.00000465, -0.00001115, 0.00000537, 0.00000367, -0.00000870, 0.00000715, -0.00000675, -0.00000084, 0.00000095, 0.00000024, 0.00000185, -0.00000057, -0.00000495, 0.00000621, -0.00000027, -0.00000002, -0.00001415, 0.00001195, -0.00000111, -0.00000980, 0.00001304, -0.00000050, 0.00000077, -0.00000529, 0.00000513, -0.00000890, 0.00001048, -0.00000441, 0.00000775, -0.00000685, 0.00000259, -0.00000727, -0.00000042, 0.00001285, -0.00000640, -0.00000451, 0.00001538, -0.00001333, 0.00001754, -0.00003271, 0.00002646, -0.00002042, 0.00000081, -0.00000093, 0.00000103, -0.00001740, 0.00002142, -0.00003439, 0.00003988, -0.00003696, 0.00002569, -0.00002726, 0.00001852, -0.00002914, 0.00002464, -0.00004024, 0.00002425, -0.00006603, 0.00003773, -0.00018530, -0.00020867, 0.00107092, -0.00049404, 0.00002674, -0.00008239, 0.00002413, -0.00004698, 0.00002149, -0.00002681, 0.00001442, -0.00001949, 0.00000957, -0.00001928, 0.00001087, -0.00002503, 0.00002808, -0.00003081, 0.00003646, -0.00001729, 0.00001176, -0.00001352, 0.00003158, -0.00002852, 0.00003314, -0.00003654, 0.00000959, -0.00000602, 0.00001987, -0.00001443, -0.00000685, 0.00000436, 0.00000325, -0.00000934, 0.00000152, -0.00000342, 0.00000550, -0.00000357, 0.00000776, -0.00001266, 0.00000262, -0.00000517, 0.00000020, -0.00000118, 0.00001488, -0.00000066, -0.00000889, -0.00000885, 0.00000448, 0.00000626, -0.00000315, 0.00000263, 0.00000268, -0.00000460, -0.00000173, -0.00000027, 0.00000694, 0.00000144, -0.00000552, 0.00000413, -0.00001118, 0.00000195, 0.00001006, 0.00000516, -0.00001154, 0.00000368, -0.00000339, -0.00000654, },
		{ 0.00000990, 0.00000290, -0.00001254, 0.00000963, -0.00000108, -0.00000498, 0.00000491, -0.00000606, -0.00000150, 0.00000233, -0.00000128, 0.00000426, -0.00000442, -0.00000101, 0.00000366, 0.00000179, -0.00000438, -0.00000993, 0.00001219, -0.00000413, -0.00000765, 0.00001371, -0.00000240, 0.00000172, -0.00000597, 0.00000624, -0.00001031, 0.00001322, -0.00000752, 0.00001093, -0.00001106, 0.00000703, -0.00001255, 0.00000615, 0.00000894, -0.00000616, -0.00000258, 0.00001361, -0.00001232, 0.00001694, -0.00003620, 0.00003557, -0.00003449, 0.00001522, -0.00001219, 0.00000906, -0.00002479, 0.00002933, -0.00004379, 0.00005308, -0.00005501, 0.00004683, -0.00005036, 0.00004317, -0.00005610, 0.00005489, -0.00007659, 0.00006655, -0.00012100, 0.00011005, -0.00033449, 0.00005752, 0.00091972, -0.00055648, 0.00010889, -0.00013129, 0.00006283, -0.00007623, 0.00004735, -0.00004826, 0.00003253, -0.00003452, 0.00002164, -0.00002899, 0.00001824, -0.00003075, 0.00003321, -0.00003571, 0.00004458, -0.00002562, 0.00001712, -0.00001362, 0.00003201, -0.00003071, 0.00003722, -0.00004665, 0.00002190, -0.00001544, 0.00002936, -0.00002681, 0.00000382, -0.00000248, 0.00000846, -0.00001518, 0.00000737, -0.00000829, 0.00001014, -0.00000781, 0.00001140, -0.00001780, 0.00000819, -0.00001022, 0.00000463, -0.00000456, 0.00002064, -0.00000845, -0.00000519, -0.00001096, 0.00000963, 0.00000139, -0.00000028, 0.00000127, 0.00000366, -0.00000686, 0.00000058, -0.00000162, 0.00000966, -0.00000222, -0.00000283, 0.00000208, -0.00001112, 0.00000387, 0.00001117, 0.00000192, -0.00001153, 0.00000616, -0.00000721, -0.00000265, },
		{ 0.00000928, 0.00000070, -0.00001257, 0.00001280, -0.00000541, -0.00000108, 0.00000218, -0.00000507, -0.00000204, 0.00000337, -0.00000219, 0.00000579, -0.00000758, 0.00000276, 0.00000099, 0.00000367, -0.00000864, -0.00000467, 0.00001074, -0.00000624, -0.00000520, 0.00001350, -0.00000345, 0.00000183, -0.00000553, 0.00000593, -0.00001005, 0.00001414, -0.00000878, 0.00001208, -0.00001313, 0.00000933, -0.00001578, 0.00001135, 0.00000568, -0.00000672, 0.00000105, 0.00000885, -0.00000685, 0.00000998, -0.00003161, 0.00003593, -0.00004066, 0.00002348, -0.00001836, 0.00001178, -0.00002601, 0.00003017, -0.00004494, 0.00005681, -0.00006289, 0.00005760, -0.00006282, 0.00005652, -0.00007067, 0.00007110, -0.00009686, 0.00008988, -0.00015271, 0.00015077, -0.00044890, 0.00032711, 0.00070475, -0.00054462, 0.00014423, -0.00015096, 0.00007970, -0.00008808, 0.00005872, -0.00005767, 0.00004042, -0.00004075, 0.00002589, -0.00003156, 0.00001890, -0.00003007, 0.00003209, -0.00003414, 0.00004661, -0.00002907, 0.00001862, -0.00000927, 0.00002647, -0.00002576, 0.00003286, -0.00004819, 0.00002702, -0.00001869, 0.00003301, -0.00003437, 0.00001103, -0.00000635, 0.00001050, -0.00001777, 0.00001016, -0.00001025, 0.00001191, -0.00000925, 0.00001207, -0.00001995, 0.00001110, -0.00001289, 0.00000670, -0.00000522, 0.00002371, -0.00001484, -0.00000200, -0.00001190, 0.00001342, -0.00000288, 0.00000250, 0.00000010, 0.00000402, -0.00000833, 0.00000228, -0.00000226, 0.00001165, -0.00000566, 0.00000014, -0.00000062, -0.00001029, 0.00000573, 0.00001193, -0.00000202, -0.00000980, 0.00000699, -0.00001003, 0.00000134, },
		{ 0.00000839, -0.00000199, -0.00001106, 0.00001435, -0.00000867, 0.00000232, -0.00000051, -0.00000411, -0.00000221, 0.00000384, -0.00000223, 0.00000602, -0.00000942, 0.00000567, -0.00000116, 0.00000477, -0.00001202, 0.00000081, 0.00000799, -0.00000735, -0.00000278, 0.00001271, -0.00000388, 0.00000136, -0.00000427, 0.00000450, -0.00000838, 0.00001346, -0.00000840, 0.00001141, -0.00001321, 0.00000956, -0.00001694, 0.00001515, 0.00000282, -0.00000733, 0.00000514, 0.00000277, 0.00000111, -0.00000141, -0.00002030, 0.00002803, -0.00003849, 0.00002470, -0.00001864, 0.00000867, -0.00002071, 0.00002363, -0.00003753, 0.00005056, -0.00005977, 0.00005681, -0.00006321, 0.00005694, -0.00007098, 0.00007104, -0.00009821, 0.00009046, -0.00015571, 0.00014993, -0.00050739, 0.00056663, 0.00045799, -0.00047600, 0.00013886, -0.00014367, 0.00007600, -0.00008333, 0.00005609, -0.00005538, 0.00003840, -0.00003851, 0.00002277, -0.00002753, 0.00001345, -0.00002363, 0.00002545, -0.00002682, 0.00004316, -0.00002828, 0.00001741, -0.00000229, 0.00001710, -0.00001583, 0.00002183, -0.00004211, 0.00002524, -0.00001574, 0.00003050, -0.00003654, 0.00001427, -0.00000694, 0.00000923, -0.00001700, 0.00000978, -0.00000916, 0.00001069, -0.00000777, 0.00000965, -0.00001889, 0.00001108, -0.00001293, 0.00000618, -0.00000277, 0.00002332, -0.00001888, -0.00000003, -0.00001122, 0.00001526, -0.00000589, 0.00000468, -0.00000063, 0.00000358, -0.00000881, 0.00000315, -0.00000193, 0.00001253, -0.00000838, 0.00000293, -0.00000366, -0.00000880, 0.00000755, 0.00001205, -0.00000616, -0.00000662, 0.00000599, -0.00001130, 0.00000482, },
		{ 0.00000740, -0.00000502, -0.00000808, 0.00001408, -0.00001042, 0.00000474, -0.00000277, -0.00000343, -0.00000189, 0.00000366, -0.00000138, 0.00000485, -0.00000962, 0.00000723, -0.00000232, 0.00000461, -0.00001384, 0.00000564, 0.00000458, -0.00000759, -0.00000056, 0.00001156, -0.00000395, 0.00000061, -0.00000256, 0.00000239, -0.00000575, 0.00001163, -0.00000688, 0.00000943, -0.00001180, 0.00000815, -0.00001632, 0.00001775, -0.00000004, -0.00000712, 0.00000823, -0.00000260, 0.00000896, -0.00001423, -0.00000522, 0.00001419, -0.00002946, 0.00001976, -0.00001386, 0.00000079, -0.00001016, 0.00001116, -0.00002314, 0.00003594, -0.00004707, 0.00004573, -0.00005273, 0.00004566, -0.00005836, 0.00005613, -0.00008210, 0.00006967, -0.00013100, 0.00010585, -0.00049790, 0.00074853, 0.00021164, -0.00037296, 0.00010452, -0.00011643, 0.00005683, -0.00006600, 0.00004271, -0.00004405, 0.00002877, -0.00002990, 0.00001425, -0.00001883, 0.00000374, -0.00001323, 0.00001507, -0.00001542, 0.00003562, -0.00002454, 0.00001509, 0.00000505, 0.00000671, -0.00000404, 0.00000718, -0.00003087, 0.00001836, -0.00000798, 0.00002285, -0.00003392, 0.00001400, -0.00000485, 0.00000534, -0.00001356, 0.00000689, -0.00000570, 0.00000712, -0.00000402, 0.00000477, -0.00001514, 0.00000853, -0.00001073, 0.00000353, 0.00000243, 0.00001943, -0.00002015, 0.00000042, -0.00000881, 0.00001495, -0.00000725, 0.00000593, -0.00000076, 0.00000234, -0.00000826, 0.00000312, -0.00000056, 0.00001206, -0.00001000, 0.00000509, -0.00000666, -0.00000680, 0.00000936, 0.00001127, -0.00000989, -0.00000258, 0.00000337, -0.00001082, 0.00000741, },
		{ 0.00000633, -0.00000808, -0.00000400, 0.00001211, -0.00001054, 0.00000591, -0.00000433, -0.00000312, -0.00000109, 0.00000294, 0.00000016, 0.00000246, -0.00000817, 0.00000727, -0.00000229, 0.00000301, -0.00001367, 0.00000906, 0.00000119, -0.00000724, 0.00000144, 0.00001021, -0.00000384, -0.00000013, -0.00000082, 0.00000007, -0.00000267, 0.00000918, -0.00000477, 0.00000676, -0.00000954, 0.00000571, -0.00001436, 0.00001942, -0.00000319, -0.00000548, 0.00000915, -0.00000543, 0.00001409, -0.00002515, 0.00000996, -0.00000212, -0.00001639, 0.00001088, -0.00000611, -0.00000957, 0.00000310, -0.00000446, -0.00000482, 0.00001616, -0.00002807, 0.00002763, -0.00003475, 0.00002630, -0.00003682, 0.00003088, -0.00005368, 0.00003346, -0.00008575, 0.00002557, -0.00041933, 0.00085512, -0.00000678, -0.00025814, 0.00005529, -0.00007863, 0.00002934, -0.00004189, 0.00002341, -0.00002775, 0.00001503, -0.00001802, 0.00000319, -0.00000814, -0.00000770, -0.00000124, 0.00000322, -0.00000204, 0.00002576, -0.00001922, 0.00001314, 0.00001068, -0.00000201, 0.00000638, -0.00000757, -0.00001771, 0.00000911, 0.00000224, 0.00001200, -0.00002797, 0.00001141, -0.00000128, 0.00000009, -0.00000870, 0.00000272, -0.00000105, 0.00000238, 0.00000085, -0.00000144, -0.00000975, 0.00000440, -0.00000724, -0.00000020, 0.00000933, 0.00001268, -0.00001882, -0.00000050, -0.00000498, 0.00001267, -0.00000690, 0.00000613, -0.00000032, 0.00000045, -0.00000685, 0.00000233, 0.00000175, 0.00001023, -0.00001030, 0.00000627, -0.00000925, -0.00000449, 0.00001117, 0.00000935, -0.00001258, 0.00000158, -0.00000032, -0.00000875, 0.00000894, },
		{ 0.00000509, -0.00001073, 0.00000060, 0.00000887, -0.00000923, 0.00000584, -0.00000518, -0.00000312, 0.00000002, 0.00000197, 0.00000198, -0.00000068, -0.00000544, 0.00000596, -0.00000121, 0.00000013, -0.00001144, 0.00001061, -0.00000156, -0.00000668, 0.00000329, 0.00000870, -0.00000364, -0.00000069, 0.00000068, -0.00000209, 0.00000043, 0.00000656, -0.00000257, 0.00000397, -0.00000702, 0.00000282, -0.00001150, 0.00002029, -0.00000663, -0.00000223, 0.00000740, -0.00000467, 0.00001465, -0.00003136, 0.00002169, -0.00001712, -0.00000272, 0.00000099, 0.00000192, -0.00001956, 0.00001595, -0.00001983, 0.00001369, -0.00000472, -0.00000699, 0.00000687, -0.00001390, 0.00000386, -0.00001191, 0.00000166, -0.00002038, -0.00000917, -0.00003155, -0.00007658, -0.00028149, 0.00088045, -0.00017794, -0.00015081, 0.00000449, -0.00003971, 0.00000091, -0.00001706, 0.00000328, -0.00001077, 0.00000087, -0.00000613, -0.00000747, 0.00000181, -0.00001832, 0.00000998, -0.00000785, 0.00001124, 0.00001521, -0.00001345, 0.00001250, 0.00001328, -0.00000718, 0.00001290, -0.00001933, -0.00000575, 0.00000034, 0.00001227, 0.00000030, -0.00002054, 0.00000798, 0.00000236, -0.00000509, -0.00000382, -0.00000137, 0.00000342, -0.00000218, 0.00000547, -0.00000759, -0.00000403, -0.00000013, -0.00000362, -0.00000371, 0.00001649, 0.00000426, -0.00001557, -0.00000229, -0.00000037, 0.00000896, -0.00000513, 0.00000542, 0.00000052, -0.00000181, -0.00000486, 0.00000110, 0.00000466, 0.00000725, -0.00000929, 0.00000630, -0.00001113, -0.00000203, 0.00001291, 0.00000627, -0.00001372, 0.00000504, -0.00000432, -0.00000553, 0.00000950, },
		{ 0.00000349, -0.00001250, 0.00000503, 0.00000500, -0.00000694, 0.00000480, -0.00000543, -0.00000327, 0.00000117, 0.00000108, 0.00000355, -0.00000389, -0.00000205, 0.00000379, 0.00000050, -0.00000354, -0.00000749, 0.00001016, -0.00000328, -0.00000624, 0.00000514, 0.00000698, -0.00000334, -0.00000104, 0.00000179, -0.00000383, 0.00000326, 0.00000407, -0.00000058, 0.00000143, -0.00000467, -0.00000007, -0.00000804, 0.00002033, -0.00001001, 0.00000215, 0.00000330, -0.00000030, 0.00001001, -0.00003126, 0.00002738, -0.00002764, 0.00000842, -0.00000713, 0.00000764, -0.00002649, 0.00002546, -0.00003175, 0.00002882, -0.00002276, 0.00001196, -0.00001212, 0.00000516, -0.00001656, 0.00001065, -0.00002494, 0.00000992, -0.00004841, 0.00001820, -0.00018182, -0.00010299, 0.00083015, -0.00029236, -0.00006433, -0.00003767, -0.00000716, -0.00002251, 0.00000351, -0.00001344, 0.00000329, -0.00001066, 0.00000315, -0.00001540, 0.00000886, -0.00002608, 0.00001856, -0.00001635, 0.00002277, 0.00000523, -0.00000782, 0.00001334, 0.00001259, -0.00000815, 0.00001424, -0.00002608, 0.00000265, -0.00000562, 0.00001983, -0.00001013, -0.00001335, 0.00000505, 0.00000485, -0.00000898, -0.00000015, -0.00000418, 0.00000652, -0.00000539, 0.00000864, -0.00001244, 0.00000083, -0.00000395, -0.00000094, -0.00000573, 0.00002238, -0.00000441, -0.00001135, -0.00000423, 0.00000420, 0.00000461, -0.00000244, 0.00000409, 0.00000147, -0.00000406, -0.00000271, -0.00000018, 0.00000770, 0.00000354, -0.00000718, 0.00000516, -0.00001215, 0.00000047, 0.00001440, 0.00000219, -0.00001305, 0.00000713, -0.00000783, -0.00000176, 0.00000934, },
		{ 0.00000143, -0.00001301, 0.00000861, 0.00000120, -0.00000426, 0.00000316, -0.00000534, -0.00000333, 0.00000210, 0.00000060, 0.00000438, -0.00000648, 0.00000125, 0.00000140, 0.00000221, -0.00000726, -0.00000251, 0.00000800, -0.00000385, -0.00000609, 0.00000707, 0.00000501, -0.00000286, -0.00000121, 0.00000250, -0.00000507, 0.00000567, 0.00000182, 0.00000107, -0.00000071, -0.00000270, -0.00000270, -0.00000413, 0.00001931, -0.00001268, 0.00000672, -0.00000205, 0.00000657, 0.00000092, -0.00002484, 0.00002602, -0.00003184, 0.00001485, -0.00001154, 0.00000931, -0.00002855, 0.00002962, -0.00003793, 0.00003797, -0.00003500, 0.00002553, -0.00002590, 0.00001874, -0.00003095, 0.00002631, -0.00004360, 0.00003076, -0.00007600, 0.00005132, -0.00027052, 0.00009228, 0.00071923, -0.00035019, -0.00000518, -0.00006536, 0.00001462, -0.00003735, 0.00001680, -0.00002416, 0.00001224, -0.00001772, 0.00000828, -0.00001928, 0.00001178, -0.00002984, 0.00002343, -0.00002122, 0.00003152, -0.00000346, -0.00000238, 0.00001506, 0.00000938, -0.00000557, 0.00001062, -0.00002729, 0.00000643, -0.00000753, 0.00002349, -0.00001779, -0.00000761, 0.00000348, 0.00000547, -0.00001084, 0.00000162, -0.00000502, 0.00000755, -0.00000653, 0.00000961, -0.00001515, 0.00000395, -0.00000631, 0.00000009, -0.00000538, 0.00002575, -0.00001196, -0.00000714, -0.00000559, 0.00000790, 0.00000047, 0.00000050, 0.00000255, 0.00000221, -0.00000594, -0.00000077, -0.00000109, 0.00001036, -0.00000036, -0.00000438, 0.00000307, -0.00001227, 0.00000298, 0.00001537, -0.00000248, -0.00001065, 0.00000746, -0.00001017, 0.00000191, 0.00000874, },
		{ -0.00000110, -0.00001209, 0.00001088, -0.00000191, -0.00000175, 0.00000136, -0.00000516, -0.00000312, 0.00000262, 0.00000075, 0.00000413, -0.00000790, 0.00000377, -0.00000054, 0.00000327, -0.00001020, 0.00000258, 0.00000475, -0.00000347, -0.00000623, 0.00000905, 0.00000280, -0.00000216, -0.00000130, 0.00000286, -0.00000582, 0.00000766, -0.00000019, 0.00000243, -0.00000247, -0.00000113, -0.00000500, 0.00000016, 0.00001701, -0.00001389, 0.00001028, -0.00000704, 0.00001403, -0.00001063, -0.00001362, 0.00001826, -0.00002944, 0.00001584, -0.00001148, 0.00000634, -0.00002516, 0.00002770, -0.00003752, 0.00004002, -0.00004000, 0.00003206, -0.00003268, 0.00002490, -0.00003719, 0.00003270, -0.00005148, 0.00003855, -0.00008703, 0.00005952, -0.00032600, 0.00027941, 0.00056859, -0.00035935, 0.00002657, -0.00007724, 0.00002453, -0.00004273, 0.00002204, -0.00002815, 0.00001549, -0.00001988, 0.00000900, -0.00001893, 0.00001044, -0.00002941, 0.00002438, -0.00002219, 0.00003707, -0.00001058, 0.00000317, 0.00001658, 0.00000514, -0.00000109, 0.00000352, -0.00002382, 0.00000587, -0.00000533, 0.00002290, -0.00002207, -0.00000384, 0.00000350, 0.00000414, -0.00001062, 0.00000143, -0.00000385, 0.00000644, -0.00000550, 0.00000821, -0.00001546, 0.00000504, -0.00000695, -0.00000067, -0.00000239, 0.00002588, -0.00001740, -0.00000373, -0.00000584, 0.00001011, -0.00000272, 0.00000309, 0.00000120, 0.00000246, -0.00000715, 0.00000064, -0.00000127, 0.00001215, -0.00000390, -0.00000140, 0.00000034, -0.00001158, 0.00000553, 0.00001551, -0.00000716, -0.00000693, 0.00000602, -0.00001096, 0.00000500, 0.00000795, },
		{ -0.00000395, -0.00000984, 0.00001163, -0.00000394, 0.00000014, -0.00000025, -0.00000507, -0.00000253, 0.00000266, 0.00000153, 0.00000273, -0.00000785, 0.00000503, -0.00000154, 0.00000317, -0.00001164, 0.00000686, 0.00000118, -0.00000255, -0.00000643, 0.00001094, 0.00000045, -0.00000122, -0.00000139, 0.00000295, -0.00000613, 0.00000923, -0.00000202, 0.00000361, -0.00000400, 0.00000015, -0.00000694, 0.00000475, 0.00001328, -0.00001308, 0.00001171, -0.00001003, 0.00001984, -0.00002198, -0.00000025, 0.00000621, -0.00002176, 0.00001207, -0.00000756, -0.00000052, -0.00001710, 0.00002042, -0.00003112, 0.00003542, -0.00003801, 0.00003161, -0.00003247, 0.00002372, -0.00003537, 0.00002990, -0.00004863, 0.00003316, -0.00008104, 0.00004015, -0.00033749, 0.00043657, 0.00040105, -0.00033264, 0.00003561, -0.00007562, 0.00002429, -0.00004003, 0.00002035, -0.00002635, 0.00001385, -0.00001793, 0.00000610, -0.00001519, 0.00000558, -0.00002542, 0.00002194, -0.00001958, 0.00003948, -0.00001609, 0.00000922, 0.00001680, 0.00000155, 0.00000319, -0.00000478, -0.00001756, 0.00000235, -0.00000007, 0.00001870, -0.00002324, -0.00000183, 0.00000472, 0.00000137, -0.00000882, -0.00000022, -0.00000120, 0.00000370, -0.00000277, 0.00000488, -0.00001369, 0.00000433, -0.00000619, -0.00000279, 0.00000284, 0.00002271, -0.00002024, -0.00000153, -0.00000477, 0.00001052, -0.00000449, 0.00000488, 0.00000032, 0.00000208, -0.00000753, 0.00000134, -0.00000056, 0.00001276, -0.00000659, 0.00000128, -0.00000262, -0.00001026, 0.00000813, 0.00001452, -0.00001118, -0.00000254, 0.00000313, -0.00001012, 0.00000721, 0.00000707, },
		{ -0.00000679, -0.00000658, 0.00001095, -0.00000478, 0.00000119, -0.00000149, -0.00000514, -0.00000160, 0.00000234, 0.00000273, 0.00000038, -0.00000636, 0.00000489, -0.00000139, 0.00000169, -0.00001113, 0.00000954, -0.00000188, -0.00000166, -0.00000638, 0.00001245, -0.00000182, -0.00000016, -0.00000147, 0.00000281, -0.00000601, 0.00001038, -0.00000368, 0.00000469, -0.00000539, 0.00000123, -0.00000848, 0.00000943, 0.00000828, -0.00001007, 0.00001037, -0.00000978, 0.00002207, -0.00003042, 0.00001214, -0.00000715, -0.00001115, 0.00000533, -0.00000137, -0.00000956, -0.00000617, 0.00000960, -0.00002054, 0.00002595, -0.00003069, 0.00002578, -0.00002693, 0.00001697, -0.00002748, 0.00002021, -0.00003770, 0.00001774, -0.00006188, -0.00000323, -0.00030194, 0.00054822, 0.00023747, -0.00028462, 0.00002928, -0.00006499, 0.00001736, -0.00003213, 0.00001414, -0.00002076, 0.00000906, -0.00001342, 0.00000107, -0.00000949, -0.00000144, -0.00001904, 0.00001711, -0.00001413, 0.00003906, -0.00001997, 0.00001588, 0.00001494, -0.00000004, 0.00000535, -0.00001192, -0.00001083, -0.00000216, 0.00000654, 0.00001224, -0.00002216, -0.00000092, 0.00000636, -0.00000197, -0.00000634, -0.00000241, 0.00000202, 0.00000023, 0.00000076, 0.00000050, -0.00001060, 0.00000247, -0.00000468, -0.00000540, 0.00000937, 0.00001684, -0.00002055, -0.00000054, -0.00000258, 0.00000924, -0.00000470, 0.00000565, 0.00000002, 0.00000105, -0.00000710, 0.00000138, 0.00000104, 0.00001206, -0.00000812, 0.00000322, -0.00000542, -0.00000847, 0.00001075, 0.00001222, -0.00001391, 0.00000170, -0.00000058, -0.00000789, 0.00000846, 0.00000609, },
		{ -0.00000920, -0.00000285, 0.00000922, -0.00000458, 0.00000141, -0.00000231, -0.00000531, -0.00000049, 0.00000189, 0.00000398, -0.00000245, -0.00000382, 0.00000356, -0.00000022, -0.00000103, -0.00000864, 0.00001018, -0.00000381, -0.00000129, -0.00000569, 0.00001326, -0.00000374, 0.00000084, -0.00000147, 0.00000244, -0.00000544, 0.00001104, -0.00000509, 0.00000567, -0.00000668, 0.00000213, -0.00000951, 0.00001382, 0.00000251, -0.00000521, 0.00000630, -0.00000590, 0.00001956, -0.00003391, 0.00002076, -0.00001876, -0.00000041, -0.00000207, 0.00000496, -0.00001854, 0.00000523, -0.00000234, -0.00000829, 0.00001414, -0.00002055, 0.00001707, -0.00001866, 0.00000750, -0.00001672, 0.00000730, -0.00002304, -0.00000240, -0.00003661, -0.00006178, -0.00022413, 0.00060688, 0.00009393, -0.00022869, 0.00001536, -0.00005045, 0.00000772, -0.00002230, 0.00000621, -0.00001373, 0.00000313, -0.00000816, -0.00000442, -0.00000344, -0.00000912, -0.00001160, 0.00001106, -0.00000673, 0.00003621, -0.00002206, 0.00002287, 0.00001093, 0.00000095, 0.00000422, -0.00001605, -0.00000571, -0.00000567, 0.00001263, 0.00000510, -0.00001998, -0.00000024, 0.00000753, -0.00000490, -0.00000414, -0.00000417, 0.00000478, -0.00000293, 0.00000405, -0.00000392, -0.00000713, 0.00000028, -0.00000327, -0.00000744, 0.00001592, 0.00000933, -0.00001887, -0.00000041, 0.00000026, 0.00000670, -0.00000355, 0.00000545, 0.00000024, -0.00000044, -0.00000605, 0.00000093, 0.00000329, 0.00001016, -0.00000840, 0.00000419, -0.00000774, -0.00000635, 0.00001327, 0.00000866, -0.00001491, 0.00000504, -0.00000435, -0.00000476, 0.00000890, 0.00000488, },
		{ -0.00001078, 0.00000074, 0.00000698, -0.00000372, 0.00000101, -0.00000284, -0.00000541, 0.00000056, 0.00000160, 0.00000483, -0.00000514, -0.00000084, 0.00000152, 0.00000152, -0.00000448, -0.00000453, 0.00000877, -0.00000429, -0.00000178, -0.00000411, 0.00001311, -0.00000501, 0.00000159, -0.00000128, 0.00000175, -0.00000435, 0.00001106, -0.00000610, 0.00000640, -0.00000777, 0.00000277, -0.00000983, 0.00001742, -0.00000324, 0.00000061, 0.00000031, 0.00000108, 0.00001234, -0.00003155, 0.00002377, -0.00002616, 0.00000795, -0.00000794, 0.00000943, -0.00002533, 0.00001482, -0.00001298, 0.00000311, 0.00000260, -0.00001022, 0.00000814, -0.00001045, -0.00000166, -0.00000649, -0.00000490, -0.00000940, -0.00002130, -0.00001366, -0.00012327, -0.00011542, 0.00061328, -0.00001985, -0.00017513, 0.00000034, -0.00003637, -0.00000124, -0.00001341, -0.00000099, -0.00000736, -0.00000215, -0.00000370, -0.00000894, 0.00000157, -0.00001612, -0.00000430, 0.00000486, 0.00000177, 0.00003127, -0.00002203, 0.00002939, 0.00000543, 0.00000416, -0.00000031, -0.00001632, -0.00000349, -0.00000677, 0.00001670, -0.00000129, -0.00001771, 0.00000092, 0.00000756, -0.00000669, -0.00000297, -0.00000474, 0.00000627, -0.00000493, 0.00000622, -0.00000745, -0.00000415, -0.00000149, -0.00000268, -0.00000798, 0.00002120, 0.00000143, -0.00001599, -0.00000059, 0.00000306, 0.00000361, -0.00000151, 0.00000452, 0.00000076, -0.00000214, -0.00000465, 0.00000031, 0.00000582, 0.00000741, -0.00000757, 0.00000412, -0.00000942, -0.00000398, 0.00001543, 0.00000409, -0.00001404, 0.00000691, -0.00000746, -0.00000131, 0.00000876, 0.00000329, },
		{ -0.00001123, 0.00000365, 0.00000478, -0.00000265, 0.00000029, -0.00000326, -0.00000527, 0.00000135, 0.00000172, 0.00000486, -0.00000707, 0.00000186, -0.00000057, 0.00000319, -0.00000793, 0.00000046, 0.00000572, -0.00000338, -0.00000320, -0.00000156, 0.00001184, -0.00000543, 0.00000188, -0.00000079, 0.00000070, -0.00000266, 0.00001030, -0.00000648, 0.00000668, -0.00000848, 0.00000302, -0.00000919, 0.00001964, -0.00000803, 0.00000617, -0.00000621, 0.00000971, 0.00000158, -0.00002385, 0.00002071, -0.00002808, 0.00001233, -0.00001076, 0.00001062, -0.00002842, 0.00002089, -0.00002050, 0.00001173, -0.00000660, -0.00000180, 0.00000108, -0.00000449, -0.00000811, 0.00000053, -0.00001324, -0.00000060, -0.00003395, -0.00000065, -0.00017475, 0.00000873, 0.00057508, -0.00010066, -0.00013011, -0.00001161, -0.00002549, -0.00000738, -0.00000722, -0.00000592, -0.00000296, -0.00000567, -0.00000099, -0.00001162, 0.00000464, -0.00002152, 0.00000198, -0.00000064, 0.00001064, 0.00002456, -0.00001944, 0.00003437, -0.00000034, 0.00000839, -0.00000727, -0.00001303, -0.00000445, -0.00000489, 0.00001794, -0.00000605, -0.00001597, 0.00000287, 0.00000617, -0.00000703, -0.00000315, -0.00000377, 0.00000608, -0.00000532, 0.00000675, -0.00000951, -0.00000223, -0.00000239, -0.00000326, -0.00000643, 0.00002419, -0.00000568, -0.00001273, -0.00000052, 0.00000515, 0.00000068, 0.00000083, 0.00000324, 0.00000131, -0.00000374, -0.00000322, -0.00000013, 0.00000816, 0.00000425, -0.00000595, 0.00000315, -0.00001039, -0.00000137, 0.00001692, -0.00000099, -0.00001148, 0.00000704, -0.00000937, 0.00000193, 0.00000828, 0.00000127, },
		{ -0.00001046, 0.00000552, 0.00000308, -0.00000180, -0.00000044, -0.00000373, -0.00000475, 0.00000173, 0.00000239, 0.00000384, -0.00000777, 0.00000365, -0.00000205, 0.00000414, -0.00001053, 0.00000536, 0.00000177, -0.00000149, -0.00000530, 0.00000179, 0.00000951, -0.00000494, 0.00000165, 0.00000005, -0.00000073, -0.00000036, 0.00000864, -0.00000603, 0.00000627, -0.00000860, 0.00000271, -0.00000743, 0.00002005, -0.00001097, 0.00001020, -0.00001162, 0.00001802, -0.00001062, -0.00001251, 0.00001253, -0.00002459, 0.00001225, -0.00001003, 0.00000812, -0.00002727, 0.00002269, -0.00002400, 0.00001657, -0.00001237, 0.00000359, -0.00000300, -0.00000189, -0.00001068, 0.00000303, -0.00001620, 0.00000138, -0.00003768, -0.00000255, -0.00020521, 0.00013180, 0.00050454, -0.00015073, -0.00009585, -0.00001887, -0.00001874, -0.00001005, -0.00000428, -0.00000806, -0.00000098, -0.00000709, -0.00000028, -0.00001221, 0.00000547, -0.00002490, 0.00000676, -0.00000487, 0.00001927, 0.00001642, -0.00001399, 0.00003674, -0.00000491, 0.00001193, -0.00001493, -0.00000750, -0.00000783, -0.00000043, 0.00001636, -0.00000894, -0.00001491, 0.00000549, 0.00000359, -0.00000609, -0.00000449, -0.00000145, 0.00000431, -0.00000411, 0.00000560, -0.00000996, -0.00000152, -0.00000237, -0.00000496, -0.00000276, 0.00002439, -0.00001116, -0.00000973, 0.00000016, 0.00000607, -0.00000143, 0.00000290, 0.00000199, 0.00000161, -0.00000496, -0.00000204, -0.00000011, 0.00000991, 0.00000119, -0.00000396, 0.00000153, -0.00001069, 0.00000150, 0.00001738, -0.00000595, -0.00000773, 0.00000554, -0.00000981, 0.00000457, 0.00000763, -0.00000115, },
		{ -0.00000863, 0.00000626, 0.00000211, -0.00000143, -0.00000096, -0.00000436, -0.00000383, 0.00000171, 0.00000354, 0.00000179, -0.00000703, 0.00000415, -0.00000248, 0.00000386, -0.00001159, 0.00000923, -0.00000215, 0.00000075, -0.00000762, 0.00000555, 0.00000640, -0.00000367, 0.00000093, 0.00000117, -0.00000242, 0.00000244, 0.00000613, -0.00000468, 0.00000503, -0.00000799, 0.00000178, -0.00000454, 0.00001846, -0.00001151, 0.00001168, -0.00001443, 0.00002395, -0.00002176, -0.00000009, 0.00000129, -0.00001700, 0.00000843, -0.00000626, 0.00000245, -0.00002232, 0.00002049, -0.00002364, 0.00001768, -0.00001469, 0.00000590, -0.00000417, -0.00000249, -0.00000958, 0.00000128, -0.00001408, -0.00000312, -0.00003271, -0.00002042, -0.00020772, 0.00023911, 0.00041568, -0.00017593, -0.00007146, -0.00002188, -0.00001549, -0.00000989, -0.00000404, -0.00000785, -0.00000104, -0.00000675, -0.00000125, -0.00001105, 0.00000425, -0.00002626, 0.00000989, -0.00000749, 0.00002711, 0.00000740, -0.00000569, 0.00003576, -0.00000706, 0.00001315, -0.00002125, -0.00000158, -0.00001218, 0.00000549, 0.00001268, -0.00001028, -0.00001419, 0.00000826, 0.00000040, -0.00000445, -0.00000643, 0.00000165, 0.00000151, -0.00000178, 0.00000318, -0.00000909, -0.00000182, -0.00000173, -0.00000727, 0.00000253, 0.00002185, -0.00001463, -0.00000732, 0.00000157, 0.00000563, -0.00000234, 0.00000429, 0.00000107, 0.00000151, -0.00000565, -0.00000125, 0.00000054, 0.00001077, -0.00000135, -0.00000204, -0.00000042, -0.00001044, 0.00000460, 0.00001654, -0.00001009, -0.00000351, 0.00000284, -0.00000884, 0.00000643, 0.00000685, -0.00000374, },
		{ -0.00000610, 0.00000604, 0.00000188, -0.00000162, -0.00000120, -0.00000511, -0.00000259, 0.00000142, 0.00000492, -0.00000101, -0.00000502, 0.00000330, -0.00000170, 0.00000216, -0.00001070, 0.00001133, -0.00000520, 0.00000259, -0.00000954, 0.00000917, 0.00000294, -0.00000190, -0.00000008, 0.00000239, -0.00000415, 0.00000548, 0.00000295, -0.00000250, 0.00000297, -0.00000662, 0.00000032, -0.00000076, 0.00001507, -0.00000960, 0.00001018, -0.00001368, 0.00002591, -0.00002953, 0.00001070, -0.00001035, -0.00000742, 0.00000240, -0.00000078, -0.00000504, -0.00001485, 0.00001540, -0.00002038, 0.00001596, -0.00001441, 0.00000602, -0.00000342, -0.00000511, -0.00000617, -0.00000313, -0.00000880, -0.00001180, -0.00002209, -0.00005115, -0.00018065, 0.00032013, 0.00032167, -0.00018365, -0.00005432, -0.00002234, -0.00001421, -0.00000826, -0.00000528, -0.00000630, -0.00000229, -0.00000541, -0.00000316, -0.00000879, 0.00000150, -0.00002590, 0.00001149, -0.00000831, 0.00003358, -0.00000173, 0.00000490, 0.00003131, -0.00000617, 0.00001090, -0.00002451, 0.00000284, -0.00001579, 0.00001139, 0.00000801, -0.00001079, -0.00001324, 0.00001047, -0.00000260, -0.00000289, -0.00000815, 0.00000468, -0.00000154, 0.00000092, 0.00000018, -0.00000751, -0.00000260, -0.00000102, -0.00000940, 0.00000853, 0.00001717, -0.00001615, -0.00000549, 0.00000346, 0.00000400, -0.00000200, 0.00000480, 0.00000062, 0.00000094, -0.00000578, -0.00000086, 0.00000181, 0.00001062, -0.00000309, -0.00000051, -0.00000240, -0.00000970, 0.00000784, 0.00001428, -0.00001284, 0.00000042, -0.00000044, -0.00000677, 0.00000753, 0.00000585, -0.00000616, },
		{ -0.00000339, 0.00000525, 0.00000215, -0.00000227, -0.00000124, -0.00000584, -0.00000125, 0.00000112, 0.00000612, -0.00000404, -0.00000217, 0.00000139, 0.00000009, -0.00000077, -0.00000786, 0.00001126, -0.00000675, 0.00000344, -0.00001050, 0.00001206, -0.00000035, -0.00000001, -0.00000111, 0.00000345, -0.00000561, 0.00000839, -0.00000052, 0.00000023, 0.00000029, -0.00000467, -0.00000141, 0.00000346, 0.00001045, -0.00000572, 0.00000597, -0.00000930, 0.00002317, -0.00003240, 0.00001759, -0.00001984, 0.00000177, -0.00000392, 0.00000475, -0.00001269, -0.00000650, 0.00000893, -0.00001567, 0.00001278, -0.00001284, 0.00000531, -0.00000228, -0.00000802, -0.00000243, -0.00000789, -0.00000318, -0.00002111, -0.00001073, -0.00008821, -0.00012772, 0.00036970, 0.00023283, -0.00018091, -0.00004138, -0.00002221, -0.00001317, -0.00000671, -0.00000665, -0.00000456, -0.00000377, -0.00000392, -0.00000524, -0.00000618, -0.00000214, -0.00002426, 0.00001177, -0.00000723, 0.00003810, -0.00000992, 0.00001674, 0.00002400, -0.00000245, 0.00000496, -0.00002374, 0.00000437, -0.00001723, 0.00001587, 0.00000353, -0.00001121, -0.00001151, 0.00001147, -0.00000470, -0.00000213, -0.00000890, 0.00000684, -0.00000401, 0.00000318, -0.00000262, -0.00000587, -0.00000332, -0.00000084, -0.00001052, 0.00001414, 0.00001127, -0.00001618, -0.00000397, 0.00000540, 0.00000165, -0.00000063, 0.00000448, 0.00000061, 0.00000000, -0.00000546, -0.00000073, 0.00000350, 0.00000956, -0.00000395, 0.00000041, -0.00000419, -0.00000853, 0.00001100, 0.00001071, -0.00001385, 0.00000338, -0.00000359, -0.00000407, 0.00000802, 0.00000451, -0.00000803, },
		{ -0.00000102, 0.00000435, 0.00000255, -0.00000310, -0.00000124, -0.00000635, -0.00000004, 0.00000109, 0.00000671, -0.00000666, 0.00000085, -0.00000102, 0.00000239, -0.00000442, -0.00000349, 0.00000913, -0.00000655, 0.00000291, -0.00001007, 0.00001374, -0.00000297, 0.00000161, -0.00000184, 0.00000406, -0.00000645, 0.00001073, -0.00000383, 0.00000310, -0.00000267, -0.00000243, -0.00000305, 0.00000748, 0.00000548, -0.00000084, 0.00000002, -0.00000206, 0.00001609, -0.00002991, 0.00001929, -0.00002528, 0.00000860, -0.00000880, 0.00000876, -0.00001888, 0.00000108, 0.00000263, -0.00001091, 0.00000952, -0.00001132, 0.00000509, -0.00000220, -0.00000955, -0.00000025, -0.00001075, 0.00000001, -0.00002743, -0.00000407, -0.00012313, -0.00005692, 0.00038810, 0.00015558, -0.00017311, -0.00003027, -0.00002297, -0.00001113, -0.00000635, -0.00000710, -0.00000354, -0.00000470, -0.00000295, -0.00000686, -0.00000385, -0.00000610, -0.00002174, 0.00001099, -0.00000422, 0.00004012, -0.00001603, 0.00002837, 0.00001508, 0.00000304, -0.00000391, -0.00001899, 0.00000247, -0.00001571, 0.00001797, 0.00000010, -0.00001209, -0.00000867, 0.00001091, -0.00000545, -0.00000260, -0.00000820, 0.00000756, -0.00000530, 0.00000440, -0.00000461, -0.00000476, -0.00000354, -0.00000157, -0.00000999, 0.00001833, 0.00000518, -0.00001529, -0.00000236, 0.00000682, -0.00000080, 0.00000131, 0.00000357, 0.00000088, -0.00000111, -0.00000488, -0.00000063, 0.00000531, 0.00000786, -0.00000403, 0.00000067, -0.00000566, -0.00000692, 0.00001377, 0.00000618, -0.00001308, 0.00000493, -0.00000602, -0.00000124, 0.00000811, 0.00000273, -0.00000899, },
		{ 0.00000062, 0.00000379, 0.00000269, -0.00000384, -0.00000141, -0.00000644, 0.00000081, 0.00000156, 0.00000634, -0.00000831, 0.00000333, -0.00000324, 0.00000451, -0.00000801, 0.00000162, 0.00000542, -0.00000477, 0.00000102, -0.00000813, 0.00001393, -0.00000457, 0.00000263, -0.00000204, 0.00000399, -0.00000639, 0.00001209, -0.00000649, 0.00000563, -0.00000547, -0.00000025, -0.00000420, 0.00001066, 0.00000110, 0.00000380, -0.00000626, 0.00000652, 0.00000599, -0.00002283, 0.00001573, -0.00002583, 0.00001185, -0.00001108, 0.00001022, -0.00002251, 0.00000674, -0.00000241, -0.00000713, 0.00000711, -0.00001071, 0.00000619, -0.00000408, -0.00000866, -0.00000087, -0.00001026, -0.00000110, -0.00002818, -0.00000650, -0.00014744, 0.00002128, 0.00038006, 0.00009236, -0.00016340, -0.00001983, -0.00002514, -0.00000763, -0.00000763, -0.00000617, -0.00000365, -0.00000471, -0.00000281, -0.00000770, -0.00000217, -0.00000998, -0.00001869, 0.00000942, 0.00000064, 0.00003924, -0.00001898, 0.00003818, 0.00000618, 0.00000867, -0.00001411, -0.00001133, -0.00000242, -0.00001126, 0.00001741, -0.00000187, -0.00001358, -0.00000478, 0.00000884, -0.00000484, -0.00000434, -0.00000599, 0.00000667, -0.00000516, 0.00000430, -0.00000545, -0.00000446, -0.00000307, -0.00000331, -0.00000757, 0.00002042, -0.00000024, -0.00001405, -0.00000036, 0.00000723, -0.00000272, 0.00000329, 0.00000241, 0.00000121, -0.00000219, -0.00000426, -0.00000033, 0.00000689, 0.00000586, -0.00000357, 0.00000039, -0.00000679, -0.00000482, 0.00001575, 0.00000124, -0.00001081, 0.00000496, -0.00000734, 0.00000131, 0.00000796, 0.00000051, -0.00000887, },
		{ 0.00000134, 0.00000384, 0.00000228, -0.00000422, -0.00000189, -0.00000601, 0.00000120, 0.00000259, 0.00000483, -0.00000858, 0.00000470, -0.00000462, 0.00000578, -0.00001071, 0.00000652, 0.00000097, -0.00000195, -0.00000191, -0.00000489, 0.00001263, -0.00000502, 0.00000291, -0.00000159, 0.00000314, -0.00000530, 0.00001220, -0.00000808, 0.00000738, -0.00000772, 0.00000153, -0.00000456, 0.00001251, -0.00000185, 0.00000702, -0.00001132, 0.00001455, -0.00000510, -0.00001294, 0.00000803, -0.00002183, 0.00001132, -0.00001045, 0.00000882, -0.00002319, 0.00000995, -0.00000568, -0.00000473, 0.00000586, -0.00001120, 0.00000869, -0.00000799, -0.00000522, -0.00000443, -0.00000621, -0.00000689, -0.00002268, -0.00002015, -0.00015455, 0.00009613, 0.00035310, 0.00004241, -0.00015284, -0.00001014, -0.00002828, -0.00000306, -0.00001024, -0.00000403, -0.00000481, -0.00000388, -0.00000344, -0.00000778, -0.00000125, -0.00001357, -0.00001534, 0.00000735, 0.00000712, 0.00003539, -0.00001802, 0.00004479, -0.00000112, 0.00001265, -0.00002359, -0.00000252, -0.00000907, -0.00000470, 0.00001460, -0.00000256, -0.00001542, -0.00000030, 0.00000570, -0.00000326, -0.00000693, -0.00000263, 0.00000442, -0.00000376, 0.00000298, -0.00000521, -0.00000494, -0.00000206, -0.00000574, -0.00000345, 0.00002018, -0.00000442, -0.00001283, 0.00000217, 0.00000639, -0.00000363, 0.00000482, 0.00000132, 0.00000140, -0.00000304, -0.00000375, 0.00000036, 0.00000795, 0.00000396, -0.00000289, -0.00000026, -0.00000759, -0.00000221, 0.00001659, -0.00000344, -0.00000760, 0.00000366, -0.00000744, 0.00000334, 0.00000762, -0.00000199, -0.00000772, },
		{ 0.00000122, 0.00000456, 0.00000120, -0.00000414, -0.00000272, -0.00000504, 0.00000116, 0.00000407, 0.00000230, -0.00000737, 0.00000463, -0.00000474, 0.00000571, -0.00001186, 0.00001028, -0.00000327, 0.00000116, -0.00000527, -0.00000080, 0.00001013, -0.00000439, 0.00000247, -0.00000055, 0.00000161, -0.00000321, 0.00001099, -0.00000840, 0.00000805, -0.00000910, 0.00000269, -0.00000401, 0.00001281, -0.00000289, 0.00000799, -0.00001387, 0.00002024, -0.00001495, -0.00000254, -0.00000184, -0.00001460, 0.00000773, -0.00000740, 0.00000501, -0.00002123, 0.00001088, -0.00000733, -0.00000347, 0.00000542, -0.00001230, 0.00001194, -0.00001315, -0.00000008, -0.00001001, 0.00000033, -0.00001612, -0.00001237, -0.00004414, -0.00014118, 0.00015864, 0.00031552, 0.00000299, -0.00014105, -0.00000214, -0.00003128, 0.00000157, -0.00001335, -0.00000133, -0.00000652, -0.00000261, -0.00000452, -0.00000734, -0.00000098, -0.00001680, -0.00001188, 0.00000515, 0.00001471, 0.00002890, -0.00001298, 0.00004733, -0.00000568, 0.00001351, -0.00003035, 0.00000539, -0.00001579, 0.00000264, 0.00001044, -0.00000253, -0.00001704, 0.00000405, 0.00000223, -0.00000141, -0.00000969, 0.00000120, 0.00000141, -0.00000161, 0.00000089, -0.00000422, -0.00000590, -0.00000091, -0.00000828, 0.00000171, 0.00001786, -0.00000722, -0.00001171, 0.00000508, 0.00000434, -0.00000332, 0.00000560, 0.00000056, 0.00000131, -0.00000359, -0.00000342, 0.00000149, 0.00000830, 0.00000244, -0.00000227, -0.00000107, -0.00000809, 0.00000087, 0.00001604, -0.00000723, -0.00000413, 0.00000151, -0.00000648, 0.00000480, 0.00000703, -0.00000449, -0.00000580, },
		{ 0.00000055, 0.00000575, -0.00000045, -0.00000364, -0.00000379, -0.00000370, 0.00000088, 0.00000569, -0.00000089, -0.00000491, 0.00000317, -0.00000352, 0.00000415, -0.00001107, 0.00001220, -0.00000646, 0.00000375, -0.00000833, 0.00000348, 0.00000689, -0.00000299, 0.00000152, 0.00000087, -0.00000036, -0.00000038, 0.00000860, -0.00000745, 0.00000757, -0.00000949, 0.00000320, -0.00000265, 0.00001172, -0.00000203, 0.00000643, -0.00001321, 0.00002232, -0.00002166, 0.00000604, -0.00001154, -0.00000603, 0.00000244, -0.00000301, -0.00000020, -0.00001749, 0.00001023, -0.00000800, -0.00000269, 0.00000502, -0.00001305, 0.00001476, -0.00001818, 0.00000526, -0.00001590, 0.00000731, -0.00002630, -0.00000052, -0.00007468, -0.00010798, 0.00020303, 0.00027472, -0.00002919, -0.00012706, 0.00000293, -0.00003281, 0.00000507, -0.00001591, 0.00000110, -0.00000811, -0.00000144, -0.00000558, -0.00000675, -0.00000117, -0.00001967, -0.00000849, 0.00000327, 0.00002269, 0.00002058, -0.00000431, 0.00004568, -0.00000706, 0.00001051, -0.00003298, 0.00001066, -0.00002091, 0.00000931, 0.00000605, -0.00000250, -0.00001781, 0.00000752, -0.00000077, -0.00000004, -0.00001184, 0.00000467, -0.00000160, 0.00000061, -0.00000135, -0.00000302, -0.00000690, -0.00000009, -0.00001020, 0.00000703, 0.00001410, -0.00000883, -0.00001052, 0.00000797, 0.00000144, -0.00000188, 0.00000552, 0.00000020, 0.00000093, -0.00000384, -0.00000321, 0.00000295, 0.00000795, 0.00000144, -0.00000190, -0.00000190, -0.00000828, 0.00000424, 0.00001404, -0.00000963, -0.00000110, -0.00000089, -0.00000480, 0.00000577, 0.00000606, -0.00000658, -0.00000356, },
		{ -0.00000021, 0.00000706, -0.00000243, -0.00000290, -0.00000491, -0.00000221, 0.00000061, 0.00000702, -0.00000419, -0.00000175, 0.00000070, -0.00000121, 0.00000128, -0.00000840, 0.00001200, -0.00000801, 0.00000518, -0.00001041, 0.00000726, 0.00000351, -0.00000124, 0.00000039, 0.00000233, -0.00000239, 0.00000276, 0.00000542, -0.00000546, 0.00000604, -0.00000896, 0.00000318, -0.00000083, 0.00000970, 0.00000024, 0.00000272, -0.00000942, 0.00002034, -0.00002410, 0.00001103, -0.00001899, 0.00000190, -0.00000299, 0.00000143, -0.00000560, -0.00001308, 0.00000887, -0.00000846, -0.00000156, 0.00000376, -0.00001241, 0.00001585, -0.00002159, 0.00000910, -0.00002012, 0.00001234, -0.00003441, 0.00000859, -0.00010580, -0.00005931, 0.00022741, 0.00023592, -0.00005685, -0.00011015, 0.00000416, -0.00003184, 0.00000647, -0.00001702, 0.00000250, -0.00000899, -0.00000083, -0.00000625, -0.00000632, -0.00000164, -0.00002214, -0.00000537, 0.00000221, 0.00003014, 0.00001162, 0.00000690, 0.00004042, -0.00000566, 0.00000383, -0.00003096, 0.00001219, -0.00002320, 0.00001407, 0.00000243, -0.00000314, -0.00001719, 0.00000949, -0.00000267, 0.00000023, -0.00001270, 0.00000708, -0.00000392, 0.00000227, -0.00000313, -0.00000214, -0.00000751, -0.00000003, -0.00001084, 0.00001155, 0.00000974, -0.00000966, -0.00000894, 0.00001029, -0.00000173, 0.00000033, 0.00000472, 0.00000021, 0.00000034, -0.00000390, -0.00000295, 0.00000452, 0.00000704, 0.00000096, -0.00000186, -0.00000265, -0.00000808, 0.00000762, 0.00001078, -0.00001042, 0.00000097, -0.00000297, -0.00000283, 0.00000644, 0.00000456, -0.00000789, -0.00000152, },
		{ -0.00000058, 0.00000803, -0.00000436, -0.00000218, -0.00000581, -0.00000084, 0.00000065, 0.00000764, -0.00000694, 0.00000143, -0.00000216, 0.00000163, -0.00000235, -0.00000429, 0.00000980, -0.00000775, 0.00000508, -0.00001106, 0.00000996, 0.00000055, 0.00000045, -0.00000059, 0.00000350, -0.00000407, 0.00000573, 0.00000192, -0.00000284, 0.00000378, -0.00000771, 0.00000288, 0.00000101, 0.00000746, 0.00000304, -0.00000220, -0.00000336, 0.00001482, -0.00002214, 0.00001160, -0.00002274, 0.00000760, -0.00000719, 0.00000477, -0.00001010, -0.00000898, 0.00000761, -0.00000936, 0.00000050, 0.00000100, -0.00000961, 0.00001426, -0.00002219, 0.00001007, -0.00002111, 0.00001343, -0.00003782, 0.00001086, -0.00013070, -0.00000222, 0.00023355, 0.00020168, -0.00008147, -0.00009038, 0.00000141, -0.00002799, 0.00000533, -0.00001621, 0.00000247, -0.00000884, -0.00000102, -0.00000633, -0.00000620, -0.00000231, -0.00002415, -0.00000273, 0.00000248, 0.00003610, 0.00000341, 0.00001915, 0.00003275, -0.00000261, -0.00000540, -0.00002480, 0.00000983, -0.00002215, 0.00001626, 0.00000017, -0.00000476, -0.00001500, 0.00000972, -0.00000317, -0.00000088, -0.00001193, 0.00000796, -0.00000508, 0.00000292, -0.00000404, -0.00000196, -0.00000747, -0.00000092, -0.00000983, 0.00001453, 0.00000557, -0.00001017, -0.00000668, 0.00001151, -0.00000452, 0.00000281, 0.00000346, 0.00000042, -0.00000030, -0.00000393, -0.00000245, 0.00000592, 0.00000585, 0.00000083, -0.00000208, -0.00000334, -0.00000737, 0.00001060, 0.00000670, -0.00000969, 0.00000181, -0.00000430, -0.00000095, 0.00000694, 0.00000248, -0.00000814, -0.00000013, },
		{ -0.00000019, 0.00000829, -0.00000589, -0.00000174, -0.00000625, 0.00000017, 0.00000120, 0.00000722, -0.00000859, 0.00000392, -0.00000469, 0.00000429, -0.00000598, 0.00000049, 0.00000617, -0.00000591, 0.00000349, -0.00001010, 0.00001126, -0.00000160, 0.00000172, -0.00000114, 0.00000411, -0.00000506, 0.00000807, -0.00000136, -0.00000007, 0.00000116, -0.00000604, 0.00000256, 0.00000242, 0.00000571, 0.00000537, -0.00000706, 0.00000359, 0.00000704, -0.00001669, 0.00000795, -0.00002234, 0.00001023, -0.00000931, 0.00000629, -0.00001302, -0.00000581, 0.00000687, -0.00001095, 0.00000366, -0.00000337, -0.00000451, 0.00000968, -0.00001948, 0.00000759, -0.00001813, 0.00000963, -0.00003510, 0.00000352, -0.00014324, 0.00005510, 0.00022593, 0.00017209, -0.00010307, -0.00006876, -0.00000464, -0.00002163, 0.00000186, -0.00001354, 0.00000100, -0.00000767, -0.00000198, -0.00000587, -0.00000639, -0.00000322, -0.00002562, -0.00000070, 0.00000443, 0.00003978, -0.00000270, 0.00003079, 0.00002413, 0.00000060, -0.00001543, -0.00001590, 0.00000433, -0.00001808, 0.00001584, -0.00000059, -0.00000732, -0.00001144, 0.00000836, -0.00000236, -0.00000327, -0.00000958, 0.00000728, -0.00000497, 0.00000245, -0.00000393, -0.00000259, -0.00000678, -0.00000263, -0.00000717, 0.00001559, 0.00000216, -0.00001067, -0.00000358, 0.00001127, -0.00000634, 0.00000502, 0.00000209, 0.00000067, -0.00000085, -0.00000405, -0.00000155, 0.00000686, 0.00000470, 0.00000082, -0.00000244, -0.00000400, -0.00000603, 0.00001276, 0.00000239, -0.00000781, 0.00000148, -0.00000469, 0.00000059, 0.00000732, -0.00000009, -0.00000727, 0.00000031, },
		{ 0.00000110, 0.00000759, -0.00000678, -0.00000175, -0.00000611, 0.00000073, 0.00000229, 0.00000565, -0.00000881, 0.00000517, -0.00000625, 0.00000611, -0.00000882, 0.00000503, 0.00000192, -0.00000307, 0.00000078, -0.00000773, 0.00001109, -0.00000274, 0.00000239, -0.00000114, 0.00000401, -0.00000516, 0.00000944, -0.00000398, 0.00000241, -0.00000141, -0.00000421, 0.00000241, 0.00000313, 0.00000500, 0.00000633, -0.00001060, 0.00000983, -0.00000121, -0.00000938, 0.00000127, -0.00001829, 0.00000974, -0.00000921, 0.00000586, -0.00001419, -0.00000371, 0.00000663, -0.00001299, 0.00000751, -0.00000886, 0.00000237, 0.00000258, -0.00001387, 0.00000201, -0.00001160, 0.00000129, -0.00002653, -0.00001406, -0.00013934, 0.00010495, 0.00021038, 0.00014549, -0.00012058, -0.00004708, -0.00001254, -0.00001380, -0.00000316, -0.00000958, -0.00000148, -0.00000581, -0.00000341, -0.00000511, -0.00000672, -0.00000448, -0.00002644, 0.00000069, 0.00000819, 0.00004069, -0.00000570, 0.00004037, 0.00001593, 0.00000245, -0.00002428, -0.00000616, -0.00000291, -0.00001193, 0.00001340, -0.00000022, -0.00001036, -0.00000707, 0.00000591, -0.00000072, -0.00000647, -0.00000608, 0.00000532, -0.00000381, 0.00000106, -0.00000297, -0.00000390, -0.00000567, -0.00000475, -0.00000327, 0.00001479, -0.00000026, -0.00001124, 0.00000025, 0.00000948, -0.00000684, 0.00000655, 0.00000089, 0.00000079, -0.00000123, -0.00000432, -0.00000018, 0.00000714, 0.00000386, 0.00000069, -0.00000278, -0.00000469, -0.00000399, 0.00001372, -0.00000149, -0.00000537, 0.00000028, -0.00000421, 0.00000173, 0.00000752, -0.00000288, -0.00000546, -0.00000022, },
		{ 0.00000319, 0.00000590, -0.00000693, -0.00000225, -0.00000539, 0.00000087, 0.00000377, 0.00000308, -0.00000755, 0.00000494, -0.00000645, 0.00000660, -0.00001025, 0.00000850, -0.00000205, 0.00000001, -0.00000243, -0.00000440, 0.00000968, -0.00000290, 0.00000245, -0.00000060, 0.00000326, -0.00000436, 0.00000970, -0.00000568, 0.00000428, -0.00000364, -0.00000242, 0.00000246, 0.00000308, 0.00000554, 0.00000542, -0.00001187, 0.00001394, -0.00000814, -0.00000218, -0.00000665, -0.00001184, 0.00000685, -0.00000731, 0.00000388, -0.00001394, -0.00000241, 0.00000649, -0.00001486, 0.00001123, -0.00001444, 0.00000987, -0.00000584, -0.00000652, -0.00000540, -0.00000296, -0.00000991, -0.00001415, -0.00003996, -0.00011787, 0.00014162, 0.00019259, 0.00011951, -0.00013242, -0.00002746, -0.00002051, -0.00000591, -0.00000859, -0.00000517, -0.00000434, -0.00000378, -0.00000490, -0.00000440, -0.00000695, -0.00000619, -0.00002658, 0.00000157, 0.00001359, 0.00003878, -0.00000506, 0.00004690, 0.00000912, 0.00000186, -0.00003023, 0.00000245, -0.00001019, -0.00000505, 0.00000988, 0.00000061, -0.00001322, -0.00000265, 0.00000312, 0.00000108, -0.00000976, -0.00000212, 0.00000267, -0.00000209, -0.00000081, -0.00000153, -0.00000556, -0.00000452, -0.00000668, 0.00000116, 0.00001256, -0.00000179, -0.00001169, 0.00000443, 0.00000639, -0.00000595, 0.00000718, 0.00000005, 0.00000073, -0.00000142, -0.00000468, 0.00000159, 0.00000674, 0.00000348, 0.00000029, -0.00000297, -0.00000540, -0.00000131, 0.00001327, -0.00000439, -0.00000301, -0.00000128, -0.00000312, 0.00000256, 0.00000735, -0.00000550, -0.00000310, -0.00000148, },
		{ 0.00000574, 0.00000344, -0.00000645, -0.00000313, -0.00000422, 0.00000077, 0.00000534, -0.00000008, -0.00000512, 0.00000332, -0.00000524, 0.00000561, -0.00000995, 0.00001031, -0.00000499, 0.00000257, -0.00000542, -0.00000073, 0.00000747, -0.00000231, 0.00000207, 0.00000030, 0.00000204, -0.00000283, 0.00000891, -0.00000636, 0.00000536, -0.00000535, -0.00000074, 0.00000260, 0.00000253, 0.00000711, 0.00000263, -0.00001050, 0.00001506, -0.00001235, 0.00000309, -0.00001389, -0.00000463, 0.00000270, -0.00000448, 0.00000109, -0.00001292, -0.00000134, 0.00000582, -0.00001573, 0.00001377, -0.00001887, 0.00001653, -0.00001401, 0.00000094, -0.00001286, 0.00000570, -0.00002142, -0.00000130, -0.00007004, -0.00008089, 0.00016229, 0.00017697, 0.00009205, -0.00013723, -0.00001180, -0.00002683, 0.00000067, -0.00001328, -0.00000123, -0.00000688, -0.00000209, -0.00000602, -0.00000405, -0.00000690, -0.00000844, -0.00002610, 0.00000226, 0.00002010, 0.00003450, -0.00000086, 0.00004998, 0.00000405, -0.00000151, -0.00003223, 0.00000837, -0.00001598, 0.00000122, 0.00000631, 0.00000118, -0.00001522, 0.00000106, 0.00000070, 0.00000234, -0.00001241, 0.00000155, -0.00000002, -0.00000039, -0.00000262, -0.00000010, -0.00000719, -0.00000372, -0.00000782, 0.00000528, 0.00000959, -0.00000278, -0.00001168, 0.00000836, 0.00000255, -0.00000393, 0.00000691, -0.00000041, 0.00000052, -0.00000153, -0.00000497, 0.00000355, 0.00000577, 0.00000355, -0.00000044, -0.00000299, -0.00000603, 0.00000179, 0.00001144, -0.00000595, -0.00000128, -0.00000268, -0.00000181, 0.00000327, 0.00000659, -0.00000749, -0.00000072, -0.00000303, },
		{ 0.00000825, 0.00000057, -0.00000557, -0.00000414, -0.00000285, 0.00000068, 0.00000659, -0.00000326, -0.00000207, 0.00000075, -0.00000291, 0.00000335, -0.00000799, 0.00001023, -0.00000641, 0.00000402, -0.00000757, 0.00000261, 0.00000496, -0.00000132, 0.00000149, 0.00000130, 0.00000065, -0.00000090, 0.00000734, -0.00000614, 0.00000567, -0.00000651, 0.00000087, 0.00000257, 0.00000193, 0.00000913, -0.00000148, -0.00000681, 0.00001310, -0.00001318, 0.00000521, -0.00001882, 0.00000174, -0.00000146, -0.00000166, -0.00000169, -0.00001181, 0.00000006, 0.00000407, -0.00001491, 0.00001423, -0.00002098, 0.00002094, -0.00002036, 0.00000682, -0.00001854, 0.00001218, -0.00003051, 0.00000816, -0.00009873, -0.00003327, 0.00016723, 0.00016583, 0.00006201, -0.00013437, -0.00000136, -0.00003029, 0.00000491, -0.00001631, 0.00000153, -0.00000856, -0.00000114, -0.00000650, -0.00000428, -0.00000646, -0.00001120, -0.00002514, 0.00000320, 0.00002698, 0.00002871, 0.00000627, 0.00004978, 0.00000046, -0.00000725, -0.00003011, 0.00001079, -0.00001923, 0.00000584, 0.00000354, 0.00000090, -0.00001587, 0.00000352, -0.00000078, 0.00000253, -0.00001384, 0.00000431, -0.00000217, 0.00000081, -0.00000391, 0.00000090, -0.00000847, -0.00000352, -0.00000772, 0.00000836, 0.00000658, -0.00000365, -0.00001080, 0.00001140, -0.00000136, -0.00000122, 0.00000595, -0.00000056, 0.00000027, -0.00000171, -0.00000495, 0.00000537, 0.00000450, 0.00000392, -0.00000140, -0.00000289, -0.00000640, 0.00000496, 0.00000850, -0.00000608, -0.00000053, -0.00000348, -0.00000063, 0.00000403, 0.00000509, -0.00000848, 0.00000111, -0.00000433, },
		{ 0.00001021, -0.00000226, -0.00000460, -0.00000504, -0.00000152, 0.00000084, 0.00000711, -0.00000587, 0.00000092, -0.00000214, -0.00000002, 0.00000033, -0.00000480, 0.00000848, -0.00000624, 0.00000410, -0.00000845, 0.00000513, 0.00000264, -0.00000024, 0.00000096, 0.00000214, -0.00000059, 0.00000106, 0.00000532, -0.00000526, 0.00000534, -0.00000715, 0.00000246, 0.00000212, 0.00000183, 0.00001081, -0.00000597, -0.00000168, 0.00000870, -0.00001085, 0.00000379, -0.00002050, 0.00000609, -0.00000463, 0.00000038, -0.00000379, -0.00001113, 0.00000212, 0.00000095, -0.00001205, 0.00001210, -0.00002004, 0.00002213, -0.00002372, 0.00000988, -0.00002105, 0.00001482, -0.00003500, 0.00001085, -0.00012019, 0.00001841, 0.00015945, 0.00015935, 0.00002964, -0.00012430, 0.00000355, -0.00003038, 0.00000632, -0.00001724, 0.00000271, -0.00000912, -0.00000108, -0.00000625, -0.00000511, -0.00000567, -0.00001438, -0.00002387, 0.00000487, 0.00003338, 0.00002250, 0.00001530, 0.00004690, -0.00000237, -0.00001424, -0.00002459, 0.00000972, -0.00001960, 0.00000831, 0.00000201, -0.00000046, -0.00001499, 0.00000450, -0.00000109, 0.00000145, -0.00001378, 0.00000581, -0.00000343, 0.00000120, -0.00000438, 0.00000120, -0.00000927, -0.00000393, -0.00000628, 0.00000995, 0.00000411, -0.00000473, -0.00000878, 0.00001299, -0.00000463, 0.00000160, 0.00000460, -0.00000057, 0.00000012, -0.00000211, -0.00000442, 0.00000674, 0.00000328, 0.00000434, -0.00000242, -0.00000281, -0.00000631, 0.00000775, 0.00000496, -0.00000503, -0.00000081, -0.00000347, 0.00000018, 0.00000493, 0.00000282, -0.00000823, 0.00000198, -0.00000486, },
		{ 0.00001119, -0.00000464, -0.00000384, -0.00000557, -0.00000044, 0.00000142, 0.00000666, -0.00000739, 0.00000317, -0.00000465, 0.00000276, -0.00000278, -0.00000107, 0.00000560, -0.00000476, 0.00000290, -0.00000796, 0.00000652, 0.00000085, 0.00000064, 0.00000067, 0.00000263, -0.00000143, 0.00000273, 0.00000322, -0.00000402, 0.00000457, -0.00000737, 0.00000408, 0.00000108, 0.00000268, 0.00001133, -0.00000973, 0.00000362, 0.00000310, -0.00000635, -0.00000066, -0.00001880, 0.00000789, -0.00000626, 0.00000124, -0.00000493, -0.00001106, 0.00000482, -0.00000338, -0.00000731, 0.00000744, -0.00001599, 0.00001982, -0.00002363, 0.00000960, -0.00001986, 0.00001292, -0.00003392, 0.00000475, -0.00012959, 0.00006712, 0.00014369, 0.00015587, -0.00000354, -0.00010850, 0.00000354, -0.00002738, 0.00000509, -0.00001611, 0.00000234, -0.00000858, -0.00000183, -0.00000540, -0.00000641, -0.00000467, -0.00001780, -0.00002247, 0.00000767, 0.00003851, 0.00001699, 0.00002508, 0.00004213, -0.00000533, -0.00002099, -0.00001707, 0.00000595, -0.00001745, 0.00000872, 0.00000172, -0.00000278, -0.00001279, 0.00000414, -0.00000035, -0.00000077, -0.00001230, 0.00000600, -0.00000375, 0.00000075, -0.00000398, 0.00000076, -0.00000963, -0.00000470, -0.00000373, 0.00000997, 0.00000245, -0.00000614, -0.00000558, 0.00001284, -0.00000671, 0.00000402, 0.00000321, -0.00000059, 0.00000017, -0.00000282, -0.00000324, 0.00000738, 0.00000242, 0.00000453, -0.00000329, -0.00000288, -0.00000556, 0.00000969, 0.00000143, -0.00000326, -0.00000189, -0.00000268, 0.00000056, 0.00000589, -0.00000003, -0.00000676, 0.00000168, -0.00000430, },
		{ 0.00001095, -0.00000629, -0.00000349, -0.00000559, 0.00000028, 0.00000243, 0.00000516, -0.00000757, 0.00000422, -0.00000619, 0.00000480, -0.00000527, 0.00000240, 0.00000232, -0.00000255, 0.00000081, -0.00000631, 0.00000675, -0.00000025, 0.00000122, 0.00000068, 0.00000271, -0.00000176, 0.00000387, 0.00000134, -0.00000268, 0.00000353, -0.00000720, 0.00000564, -0.00000054, 0.00000468, 0.00001014, -0.00001178, 0.00000779, -0.00000221, -0.00000118, -0.00000693, -0.00001438, 0.00000728, -0.00000634, 0.00000097, -0.00000520, -0.00001139, 0.00000778, -0.00000832, -0.00000137, 0.00000097, -0.00000942, 0.00001446, -0.00002041, 0.00000632, -0.00001538, 0.00000694, -0.00002775, -0.00001017, -0.00012408, 0.00010676, 0.00012528, 0.00015270, -0.00003522, -0.00008920, -0.00000003, -0.00002217, 0.00000190, -0.00001342, 0.00000077, -0.00000727, -0.00000309, -0.00000425, -0.00000792, -0.00000371, -0.00002131, -0.00002098, 0.00001181, 0.00004178, 0.00001312, 0.00003450, 0.00003617, -0.00000914, -0.00002604, -0.00000920, 0.00000077, -0.00001363, 0.00000760, 0.00000229, -0.00000564, -0.00000975, 0.00000285, 0.00000104, -0.00000372, -0.00000976, 0.00000513, -0.00000331, -0.00000033, -0.00000290, -0.00000030, -0.00000975, -0.00000541, -0.00000060, 0.00000873, 0.00000152, -0.00000770, -0.00000150, 0.00001100, -0.00000733, 0.00000567, 0.00000200, -0.00000071, 0.00000041, -0.00000379, -0.00000142, 0.00000717, 0.00000215, 0.00000427, -0.00000382, -0.00000320, -0.00000403, 0.00001042, -0.00000145, -0.00000138, -0.00000335, -0.00000135, 0.00000063, 0.00000672, -0.00000311, -0.00000438, 0.00000030, -0.00000257, },
		{ 0.00000951, -0.00000712, -0.00000359, -0.00000510, 0.00000069, 0.00000369, 0.00000283, -0.00000644, 0.00000390, -0.00000644, 0.00000567, -0.00000662, 0.00000495, -0.00000058, -0.00000028, -0.00000158, -0.00000394, 0.00000604, -0.00000070, 0.00000149, 0.00000094, 0.00000244, -0.00000161, 0.00000443, -0.00000017, -0.00000141, 0.00000235, -0.00000664, 0.00000696, -0.00000248, 0.00000768, 0.00000710, -0.00001156, 0.00000986, -0.00000587, 0.00000306, -0.00001342, -0.00000843, 0.00000494, -0.00000530, -0.00000004, -0.00000502, -0.00001166, 0.00001032, -0.00001298, 0.00000470, -0.00000618, -0.00000151, 0.00000712, -0.00001507, 0.00000108, -0.00000885, -0.00000163, -0.00001840, -0.00003190, -0.00010343, 0.00013322, 0.00010896, 0.00014698, -0.00006285, -0.00006891, -0.00000545, -0.00001599, -0.00000227, -0.00000989, -0.00000143, -0.00000561, -0.00000447, -0.00000315, -0.00000932, -0.00000301, -0.00002475, -0.00001935, 0.00001727, 0.00004297, 0.00001150, 0.00004270, 0.00002943, -0.00001408, -0.00002834, -0.00000252, -0.00000446, -0.00000925, 0.00000575, 0.00000312, -0.00000843, -0.00000649, 0.00000122, 0.00000254, -0.00000681, -0.00000670, 0.00000364, -0.00000248, -0.00000170, -0.00000144, -0.00000175, -0.00000987, -0.00000558, 0.00000241, 0.00000678, 0.00000102, -0.00000901, 0.00000292, 0.00000786, -0.00000656, 0.00000641, 0.00000110, -0.00000091, 0.00000077, -0.00000487, 0.00000086, 0.00000618, 0.00000251, 0.00000347, -0.00000394, -0.00000377, -0.00000180, 0.00000978, -0.00000323, 0.00000003, -0.00000465, 0.00000013, 0.00000063, 0.00000711, -0.00000592, -0.00000158, -0.00000180, 0.00000013, },
		{ 0.00000712, -0.00000722, -0.00000406, -0.00000420, 0.00000093, 0.00000491, 0.00000007, -0.00000434, 0.00000236, -0.00000538, 0.00000526, -0.00000661, 0.00000613, -0.00000253, 0.00000141, -0.00000366, -0.00000138, 0.00000474, -0.00000067, 0.00000159, 0.00000131, 0.00000199, -0.00000114, 0.00000449, -0.00000127, -0.00000029, 0.00000108, -0.00000563, 0.00000776, -0.00000426, 0.00001117, 0.00000253, -0.00000905, 0.00000939, -0.00000703, 0.00000506, -0.00001859, -0.00000234, 0.00000183, -0.00000378, -0.00000121, -0.00000493, -0.00001131, 0.00001172, -0.00001642, 0.00000974, -0.00001268, 0.00000632, -0.00000076, -0.00000897, -0.00000467, -0.00000198, -0.00001067, -0.00000872, -0.00005669, -0.00007003, 0.00014493, 0.00009798, 0.00013647, -0.00008427, -0.00004995, -0.00001105, -0.00001006, -0.00000639, -0.00000633, -0.00000368, -0.00000406, -0.00000560, -0.00000243, -0.00001035, -0.00000281, -0.00002805, -0.00001737, 0.00002373, 0.00004217, 0.00001240, 0.00004915, 0.00002203, -0.00001988, -0.00002751, 0.00000188, -0.00000856, -0.00000532, 0.00000398, 0.00000363, -0.00001061, -0.00000361, -0.00000021, 0.00000362, -0.00000949, -0.00000369, 0.00000199, -0.00000166, -0.00000295, 0.00000003, -0.00000337, -0.00001016, -0.00000485, 0.00000464, 0.00000479, 0.00000049, -0.00000959, 0.00000696, 0.00000407, -0.00000473, 0.00000634, 0.00000049, -0.00000110, 0.00000105, -0.00000579, 0.00000330, 0.00000464, 0.00000339, 0.00000220, -0.00000368, -0.00000445, 0.00000091, 0.00000788, -0.00000368, 0.00000054, -0.00000534, 0.00000135, 0.00000085, 0.00000676, -0.00000798, 0.00000105, -0.00000407, 0.00000337, },
		{ 0.00000418, -0.00000683, -0.00000470, -0.00000308, 0.00000121, 0.00000572, -0.00000258, -0.00000181, 0.00000006, -0.00000332, 0.00000378, -0.00000536, 0.00000589, -0.00000323, 0.00000211, -0.00000493, 0.00000082, 0.00000327, -0.00000039, 0.00000168, 0.00000159, 0.00000159, -0.00000059, 0.00000422, -0.00000205, 0.00000070, -0.00000027, -0.00000408, 0.00000773, -0.00000533, 0.00001438, -0.00000278, -0.00000481, 0.00000662, -0.00000550, 0.00000412, -0.00002128, 0.00000266, -0.00000111, -0.00000247, -0.00000199, -0.00000540, -0.00000987, 0.00001145, -0.00001787, 0.00001275, -0.00001731, 0.00001271, -0.00000780, -0.00000350, -0.00000947, 0.00000350, -0.00001805, -0.00000178, -0.00007982, -0.00002837, 0.00014293, 0.00009371, 0.00012008, -0.00009819, -0.00003404, -0.00001552, -0.00000538, -0.00000964, -0.00000340, -0.00000548, -0.00000298, -0.00000622, -0.00000230, -0.00001082, -0.00000325, -0.00003118, -0.00001475, 0.00003072, 0.00003979, 0.00001571, 0.00005353, 0.00001384, -0.00002576, -0.00002390, 0.00000363, -0.00001089, -0.00000253, 0.00000285, 0.00000337, -0.00001177, -0.00000152, -0.00000106, 0.00000394, -0.00001134, -0.00000115, 0.00000054, -0.00000114, -0.00000380, 0.00000126, -0.00000503, -0.00001060, -0.00000310, 0.00000566, 0.00000330, -0.00000045, -0.00000899, 0.00000991, 0.00000037, -0.00000237, 0.00000570, 0.00000003, -0.00000112, 0.00000103, -0.00000626, 0.00000549, 0.00000292, 0.00000452, 0.00000064, -0.00000319, -0.00000502, 0.00000369, 0.00000507, -0.00000287, 0.00000000, -0.00000512, 0.00000200, 0.00000145, 0.00000551, -0.00000887, 0.00000291, -0.00000592, 0.00000658, },
		{ 0.00000117, -0.00000623, -0.00000528, -0.00000196, 0.00000170, 0.00000581, -0.00000459, 0.00000052, -0.00000240, -0.00000080, 0.00000168, -0.00000328, 0.00000448, -0.00000275, 0.00000173, -0.00000514, 0.00000229, 0.00000196, -0.00000007, 0.00000187, 0.00000166, 0.00000139, -0.00000018, 0.00000384, -0.00000263, 0.00000161, -0.00000171, -0.00000199, 0.00000666, -0.00000514, 0.00001648, -0.00000780, 0.00000015, 0.00000239, -0.00000183, 0.00000031, -0.00002097, 0.00000574, -0.00000318, -0.00000180, -0.00000206, -0.00000667, -0.00000720, 0.00000931, -0.00001698, 0.00001317, -0.00001930, 0.00001669, -0.00001291, 0.00000031, -0.00001228, 0.00000636, -0.00002219, -0.00000013, -0.00009663, 0.00001584, 0.00013036, 0.00009561, 0.00009806, -0.00010435, -0.00002208, -0.00001812, -0.00000247, -0.00001152, -0.00000150, -0.00000655, -0.00000253, -0.00000624, -0.00000285, -0.00001067, -0.00000441, -0.00003410, -0.00001116, 0.00003764, 0.00003643, 0.00002110, 0.00005569, 0.00000466, -0.00003066, -0.00001845, 0.00000302, -0.00001134, -0.00000113, 0.00000263, 0.00000223, -0.00001181, -0.00000037, -0.00000118, 0.00000337, -0.00001219, 0.00000067, -0.00000050, -0.00000106, -0.00000410, 0.00000209, -0.00000672, -0.00001097, -0.00000056, 0.00000537, 0.00000256, -0.00000199, -0.00000702, 0.00001128, -0.00000254, -0.00000004, 0.00000482, -0.00000044, -0.00000083, 0.00000056, -0.00000602, 0.00000705, 0.00000143, 0.00000556, -0.00000092, -0.00000271, -0.00000519, 0.00000608, 0.00000191, -0.00000119, -0.00000145, -0.00000398, 0.00000199, 0.00000249, 0.00000333, -0.00000842, 0.00000362, -0.00000682, 0.00000921, },
		{ -0.00000150, -0.00000570, -0.00000560, -0.00000097, 0.00000248, 0.00000503, -0.00000558, 0.00000208, -0.00000438, 0.00000158, -0.00000044, -0.00000096, 0.00000245, -0.00000143, 0.00000045, -0.00000430, 0.00000289, 0.00000103, 0.00000017, 0.00000221, 0.00000148, 0.00000149, -0.00000004, 0.00000351, -0.00000311, 0.00000248, -0.00000317, 0.00000056, 0.00000451, -0.00000339, 0.00001674, -0.00001149, 0.00000466, -0.00000210, 0.00000286, -0.00000560, -0.00001786, 0.00000666, -0.00000410, -0.00000192, -0.00000140, -0.00000866, -0.00000349, 0.00000552, -0.00001390, 0.00001100, -0.00001843, 0.00001785, -0.00001559, 0.00000197, -0.00001267, 0.00000615, -0.00002246, -0.00000509, -0.00010345, 0.00005685, 0.00011160, 0.00010172, 0.00007181, -0.00010351, -0.00001415, -0.00001870, -0.00000141, -0.00001194, -0.00000072, -0.00000684, -0.00000270, -0.00000570, -0.00000401, -0.00000992, -0.00000632, -0.00003678, -0.00000635, 0.00004391, 0.00003276, 0.00002809, 0.00005547, -0.00000552, -0.00003354, -0.00001245, 0.00000083, -0.00001031, -0.00000094, 0.00000319, 0.00000032, -0.00001087, -0.00000009, -0.00000069, 0.00000204, -0.00001210, 0.00000177, -0.00000115, -0.00000137, -0.00000387, 0.00000251, -0.00000854, -0.00001090, 0.00000229, 0.00000405, 0.00000253, -0.00000399, -0.00000378, 0.00001089, -0.00000422, 0.00000181, 0.00000399, -0.00000103, -0.00000017, -0.00000045, -0.00000492, 0.00000771, 0.00000050, 0.00000618, -0.00000220, -0.00000244, -0.00000473, 0.00000760, -0.00000099, 0.00000082, -0.00000341, -0.00000214, 0.00000142, 0.00000380, 0.00000048, -0.00000671, 0.00000304, -0.00000647, 0.00001081, },
		{ -0.00000356, -0.00000543, -0.00000554, -0.00000017, 0.00000349, 0.00000340, -0.00000539, 0.00000252, -0.00000542, 0.00000329, -0.00000205, 0.00000097, 0.00000042, 0.00000015, -0.00000127, -0.00000272, 0.00000267, 0.00000051, 0.00000032, 0.00000265, 0.00000113, 0.00000183, -0.00000021, 0.00000331, -0.00000355, 0.00000326, -0.00000452, 0.00000331, 0.00000153, -0.00000009, 0.00001482, -0.00001307, 0.00000766, -0.00000559, 0.00000718, -0.00001234, -0.00001275, 0.00000571, -0.00000398, -0.00000263, -0.00000031, -0.00001097, 0.00000072, 0.00000073, -0.00000926, 0.00000674, -0.00001506, 0.00001638, -0.00001592, 0.00000163, -0.00001093, 0.00000324, -0.00001940, -0.00001641, -0.00009836, 0.00008982, 0.00009126, 0.00010924, 0.00004353, -0.00009717, -0.00000969, -0.00001763, -0.00000185, -0.00001113, -0.00000090, -0.00000648, -0.00000332, -0.00000481, -0.00000560, -0.00000871, -0.00000899, -0.00003909, -0.00000016, 0.00004904, 0.00002944, 0.00003617, 0.00005266, -0.00001637, -0.00003370, -0.00000715, -0.00000192, -0.00000852, -0.00000147, 0.00000421, -0.00000200, -0.00000927, -0.00000038, 0.00000017, 0.00000023, -0.00001131, 0.00000231, -0.00000153, -0.00000191, -0.00000325, 0.00000256, -0.00001060, -0.00000998, 0.00000479, 0.00000226, 0.00000285, -0.00000605, 0.00000026, 0.00000892, -0.00000453, 0.00000290, 0.00000337, -0.00000176, 0.00000080, -0.00000190, -0.00000301, 0.00000737, 0.00000034, 0.00000616, -0.00000300, -0.00000251, -0.00000352, 0.00000794, -0.00000304, 0.00000254, -0.00000535, -0.00000001, 0.00000057, 0.00000507, -0.00000263, -0.00000412, 0.00000134, -0.00000484, 0.00001117, },
		{ -0.00000490, -0.00000546, -0.00000512, 0.00000049, 0.00000452, 0.00000121, -0.00000417, 0.00000180, -0.00000528, 0.00000402, -0.00000278, 0.00000205, -0.00000103, 0.00000143, -0.00000288, -0.00000084, 0.00000190, 0.00000032, 0.00000048, 0.00000303, 0.00000076, 0.00000224, -0.00000055, 0.00000319, -0.00000389, 0.00000387, -0.00000555, 0.00000588, -0.00000182, 0.00000433, 0.00001082, -0.00001224, 0.00000846, -0.00000708, 0.00000983, -0.00001848, -0.00000678, 0.00000356, -0.00000324, -0.00000352, 0.00000075, -0.00001304, 0.00000472, -0.00000421, -0.00000399, 0.00000133, -0.00001003, 0.00001295, -0.00001450, -0.00000008, -0.00000790, -0.00000127, -0.00001452, -0.00003225, -0.00008152, 0.00011169, 0.00007335, 0.00011535, 0.00001564, -0.00008720, -0.00000775, -0.00001556, -0.00000322, -0.00000952, -0.00000173, -0.00000574, -0.00000415, -0.00000381, -0.00000737, -0.00000722, -0.00001240, -0.00004087, 0.00000738, 0.00005269, 0.00002703, 0.00004480, 0.00004693, -0.00002712, -0.00003101, -0.00000349, -0.00000428, -0.00000671, -0.00000216, 0.00000523, -0.00000433, -0.00000742, -0.00000094, 0.00000107, -0.00000171, -0.00001014, 0.00000250, -0.00000184, -0.00000246, -0.00000240, 0.00000230, -0.00001291, -0.00000790, 0.00000629, 0.00000065, 0.00000299, -0.00000758, 0.00000441, 0.00000590, -0.00000365, 0.00000323, 0.00000298, -0.00000252, 0.00000190, -0.00000357, -0.00000053, 0.00000615, 0.00000097, 0.00000544, -0.00000324, -0.00000292, -0.00000162, 0.00000699, -0.00000387, 0.00000344, -0.00000673, 0.00000191, -0.00000019, 0.00000591, -0.00000545, -0.00000119, -0.00000105, -0.00000218, 0.00001031, },
		{ -0.00000559, -0.00000572, -0.00000441, 0.00000115, 0.00000528, -0.00000112, -0.00000231, 0.00000017, -0.00000406, 0.00000374, -0.00000254, 0.00000205, -0.00000152, 0.00000195, -0.00000390, 0.00000084, 0.00000091, 0.00000028, 0.00000077, 0.00000318, 0.00000061, 0.00000249, -0.00000089, 0.00000302, -0.00000401, 0.00000415, -0.00000602, 0.00000782, -0.00000486, 0.00000916, 0.00000533, -0.00000923, 0.00000697, -0.00000612, 0.00000989, -0.00002276, -0.00000113, 0.00000100, -0.00000239, -0.00000412, 0.00000126, -0.00001428, 0.00000779, -0.00000842, 0.00000090, -0.00000418, -0.00000438, 0.00000852, -0.00001214, -0.00000228, -0.00000468, -0.00000596, -0.00000990, -0.00004963, -0.00005511, 0.00012144, 0.00006059, 0.00011777, -0.00000968, -0.00007546, -0.00000731, -0.00001319, -0.00000489, -0.00000762, -0.00000283, -0.00000491, -0.00000495, -0.00000296, -0.00000908, -0.00000566, -0.00001651, -0.00004186, 0.00001614, 0.00005468, 0.00002604, 0.00005335, 0.00003803, -0.00003667, -0.00002595, -0.00000184, -0.00000566, -0.00000546, -0.00000251, 0.00000587, -0.00000631, -0.00000562, -0.00000150, 0.00000178, -0.00000352, -0.00000884, 0.00000253, -0.00000222, -0.00000285, -0.00000147, 0.00000169, -0.00001533, -0.00000464, 0.00000638, -0.00000019, 0.00000246, -0.00000800, 0.00000786, 0.00000256, -0.00000204, 0.00000299, 0.00000271, -0.00000315, 0.00000286, -0.00000512, 0.00000216, 0.00000434, 0.00000219, 0.00000413, -0.00000300, -0.00000354, 0.00000071, 0.00000493, -0.00000339, 0.00000323, -0.00000714, 0.00000321, -0.00000051, 0.00000596, -0.00000746, 0.00000144, -0.00000354, 0.00000103, 0.00000849, },
		{ -0.00000583, -0.00000607, -0.00000357, 0.00000195, 0.00000546, -0.00000309, -0.00000034, -0.00000190, -0.00000210, 0.00000268, -0.00000152, 0.00000108, -0.00000099, 0.00000152, -0.00000399, 0.00000194, 0.00000003, 0.00000023, 0.00000131, 0.00000298, 0.00000083, 0.00000235, -0.00000099, 0.00000262, -0.00000377, 0.00000399, -0.00000572, 0.00000870, -0.00000691, 0.00001347, -0.00000072, -0.00000476, 0.00000367, -0.00000289, 0.00000708, -0.00002440, 0.00000323, -0.00000124, -0.00000190, -0.00000406, 0.00000086, -0.00001430, 0.00000942, -0.00001120, 0.00000456, -0.00000880, 0.00000090, 0.00000402, -0.00000970, -0.00000412, -0.00000227, -0.00000951, -0.00000761, -0.00006500, -0.00002276, 0.00012005, 0.00005416, 0.00011515, -0.00003082, -0.00006348, -0.00000748, -0.00001111, -0.00000632, -0.00000588, -0.00000387, -0.00000421, -0.00000551, -0.00000247, -0.00001050, -0.00000426, -0.00002125, -0.00004177, 0.00002582, 0.00005499, 0.00002687, 0.00006108, 0.00002590, -0.00004380, -0.00001955, -0.00000200, -0.00000587, -0.00000504, -0.00000226, 0.00000593, -0.00000777, -0.00000407, -0.00000193, 0.00000220, -0.00000505, -0.00000756, 0.00000249, -0.00000272, -0.00000301, -0.00000051, 0.00000059, -0.00001748, -0.00000049, 0.00000500, 0.00000002, 0.00000097, -0.00000690, 0.00000994, -0.00000032, -0.00000025, 0.00000249, 0.00000236, -0.00000341, 0.00000341, -0.00000619, 0.00000460, 0.00000237, 0.00000367, 0.00000248, -0.00000247, -0.00000411, 0.00000308, 0.00000216, -0.00000180, 0.00000191, -0.00000644, 0.00000365, -0.00000020, 0.00000503, -0.00000828, 0.00000321, -0.00000553, 0.00000424, 0.00000609, },
		{ -0.00000586, -0.00000633, -0.00000270, 0.00000296, 0.00000486, -0.00000428, 0.00000117, -0.00000383, 0.00000009, 0.00000129, -0.00000016, -0.00000049, 0.00000031, 0.00000023, -0.00000310, 0.00000224, -0.00000054, 0.00000007, 0.00000211, 0.00000239, 0.00000151, 0.00000169, -0.00000068, 0.00000187, -0.00000308, 0.00000329, -0.00000456, 0.00000828, -0.00000741, 0.00001635, -0.00000623, 0.00000011, -0.00000052, 0.00000184, 0.00000179, -0.00002319, 0.00000576, -0.00000271, -0.00000199, -0.00000320, -0.00000056, -0.00001298, 0.00000941, -0.00001220, 0.00000645, -0.00001189, 0.00000506, 0.00000015, -0.00000777, -0.00000508, -0.00000130, -0.00001109, -0.00000916, -0.00007508, 0.00001114, 0.00011004, 0.00005373, 0.00010720, -0.00004695, -0.00005231, -0.00000770, -0.00000967, -0.00000717, -0.00000459, -0.00000462, -0.00000379, -0.00000572, -0.00000246, -0.00001146, -0.00000323, -0.00002654, -0.00004034, 0.00003606, 0.00005370, 0.00002988, 0.00006704, 0.00001093, -0.00004748, -0.00001307, -0.00000333, -0.00000517, -0.00000539, -0.00000140, 0.00000538, -0.00000865, -0.00000280, -0.00000224, 0.00000238, -0.00000632, -0.00000630, 0.00000233, -0.00000326, -0.00000299, 0.00000051, -0.00000120, -0.00001885, 0.00000395, 0.00000246, 0.00000121, -0.00000145, -0.00000419, 0.00001023, -0.00000216, 0.00000117, 0.00000206, 0.00000176, -0.00000315, 0.00000333, -0.00000648, 0.00000639, 0.00000067, 0.00000505, 0.00000082, -0.00000191, -0.00000434, 0.00000502, -0.00000073, 0.00000041, -0.00000022, -0.00000476, 0.00000323, 0.00000072, 0.00000315, -0.00000777, 0.00000380, -0.00000653, 0.00000690, 0.00000354, },
		{ -0.00000588, -0.00000639, -0.00000189, 0.00000417, 0.00000344, -0.00000444, 0.00000179, -0.00000508, 0.00000196, 0.00000005, 0.00000104, -0.00000212, 0.00000193, -0.00000155, -0.00000143, 0.00000175, -0.00000073, -0.00000018, 0.00000309, 0.00000149, 0.00000258, 0.00000050, 0.00000009, 0.00000073, -0.00000194, 0.00000212, -0.00000264, 0.00000656, -0.00000608, 0.00001714, -0.00001022, 0.00000428, -0.00000444, 0.00000688, -0.00000503, -0.00001954, 0.00000641, -0.00000328, -0.00000265, -0.00000169, -0.00000283, -0.00001051, 0.00000791, -0.00001149, 0.00000649, -0.00001317, 0.00000775, -0.00000271, -0.00000660, -0.00000504, -0.00000184, -0.00001062, -0.00001506, -0.00007752, 0.00004234, 0.00009479, 0.00005792, 0.00009456, -0.00005797, -0.00004250, -0.00000771, -0.00000894, -0.00000734, -0.00000386, -0.00000499, -0.00000370, -0.00000559, -0.00000293, -0.00001188, -0.00000275, -0.00003223, -0.00003729, 0.00004645, 0.00005102, 0.00003532, 0.00007009, -0.00000596, -0.00004716, -0.00000769, -0.00000500, -0.00000404, -0.00000621, -0.00000009, 0.00000433, -0.00000905, -0.00000177, -0.00000253, 0.00000245, -0.00000744, -0.00000497, 0.00000194, -0.00000369, -0.00000288, 0.00000164, -0.00000387, -0.00001886, 0.00000786, -0.00000056, 0.00000295, -0.00000440, -0.00000021, 0.00000873, -0.00000263, 0.00000187, 0.00000195, 0.00000079, -0.00000230, 0.00000254, -0.00000584, 0.00000724, -0.00000041, 0.00000596, -0.00000055, -0.00000155, -0.00000402, 0.00000611, -0.00000316, 0.00000266, -0.00000265, -0.00000247, 0.00000219, 0.00000203, 0.00000060, -0.00000609, 0.00000313, -0.00000629, 0.00000861, 0.00000120, },
		{ -0.00000601, -0.00000620, -0.00000109, 0.00000540, 0.00000137, -0.00000356, 0.00000132, -0.00000530, 0.00000311, -0.00000063, 0.00000160, -0.00000328, 0.00000331, -0.00000332, 0.00000061, 0.00000067, -0.00000062, -0.00000042, 0.00000406, 0.00000051, 0.00000382, -0.00000104, 0.00000121, -0.00000070, -0.00000046, 0.00000065, -0.00000024, 0.00000382, -0.00000304, 0.00001559, -0.00001203, 0.00000686, -0.00000703, 0.00001099, -0.00001215, -0.00001430, 0.00000553, -0.00000312, -0.00000363, 0.00000014, -0.00000555, -0.00000734, 0.00000540, -0.00000946, 0.00000497, -0.00001280, 0.00000898, -0.00000456, -0.00000609, -0.00000426, -0.00000347, -0.00000875, -0.00002467, -0.00007134, 0.00006744, 0.00007778, 0.00006468, 0.00007849, -0.00006436, -0.00003417, -0.00000753, -0.00000877, -0.00000693, -0.00000366, -0.00000503, -0.00000387, -0.00000520, -0.00000384, -0.00001176, -0.00000299, -0.00003816, -0.00003241, 0.00005649, 0.00004733, 0.00004326, 0.00006903, -0.00002326, -0.00004300, -0.00000422, -0.00000624, -0.00000303, -0.00000709, 0.00000140, 0.00000298, -0.00000908, -0.00000088, -0.00000291, 0.00000255, -0.00000856, -0.00000349, 0.00000120, -0.00000385, -0.00000280, 0.00000286, -0.00000746, -0.00001709, 0.00001044, -0.00000323, 0.00000457, -0.00000720, 0.00000440, 0.00000583, -0.00000178, 0.00000173, 0.00000224, -0.00000050, -0.00000094, 0.00000112, -0.00000431, 0.00000707, -0.00000067, 0.00000621, -0.00000143, -0.00000152, -0.00000303, 0.00000608, -0.00000461, 0.00000436, -0.00000481, -0.00000009, 0.00000093, 0.00000335, -0.00000217, -0.00000366, 0.00000144, -0.00000485, 0.00000922, -0.00000070, },
		{ -0.00000627, -0.00000580, -0.00000022, 0.00000640, -0.00000099, -0.00000191, -0.00000013, -0.00000441, 0.00000335, -0.00000056, 0.00000126, -0.00000354, 0.00000398, -0.00000453, 0.00000251, -0.00000064, -0.00000039, -0.00000046, 0.00000477, -0.00000026, 0.00000492, -0.00000266, 0.00000247, -0.00000223, 0.00000114, -0.00000082, 0.00000220, 0.00000062, 0.00000120, 0.00001192, -0.00001148, 0.00000741, -0.00000759, 0.00001310, -0.00001830, -0.00000850, 0.00000371, -0.00000259, -0.00000455, 0.00000184, -0.00000822, -0.00000402, 0.00000248, -0.00000675, 0.00000242, -0.00001120, 0.00000907, -0.00000565, -0.00000593, -0.00000329, -0.00000541, -0.00000670, -0.00003633, -0.00005708, 0.00008437, 0.00006199, 0.00007192, 0.00006053, -0.00006692, -0.00002718, -0.00000733, -0.00000890, -0.00000616, -0.00000383, -0.00000483, -0.00000419, -0.00000470, -0.00000503, -0.00001117, -0.00000403, -0.00004413, -0.00002550, 0.00006565, 0.00004320, 0.00005342, 0.00006283, -0.00003912, -0.00003586, -0.00000290, -0.00000660, -0.00000256, -0.00000772, 0.00000285, 0.00000146, -0.00000882, -0.00000013, -0.00000337, 0.00000275, -0.00000972, -0.00000181, 0.00000007, -0.00000366, -0.00000277, 0.00000403, -0.00001179, -0.00001340, 0.00001110, -0.00000480, 0.00000534, -0.00000906, 0.00000874, 0.00000223, 0.00000003, 0.00000086, 0.00000285, -0.00000196, 0.00000070, -0.00000067, -0.00000211, 0.00000600, -0.00000013, 0.00000573, -0.00000174, -0.00000183, -0.00000145, 0.00000492, -0.00000484, 0.00000506, -0.00000617, 0.00000186, -0.00000013, 0.00000425, -0.00000461, -0.00000104, -0.00000080, -0.00000250, 0.00000876, -0.00000208, },
		{ -0.00000658, -0.00000528, 0.00000086, 0.00000684, -0.00000317, 0.00000004, -0.00000220, -0.00000263, 0.00000276, 0.00000022, 0.00000001, -0.00000277, 0.00000364, -0.00000480, 0.00000381, -0.00000185, -0.00000024, -0.00000014, 0.00000502, -0.00000056, 0.00000555, -0.00000401, 0.00000358, -0.00000358, 0.00000259, -0.00000198, 0.00000419, -0.00000233, 0.00000583, 0.00000679, -0.00000891, 0.00000598, -0.00000595, 0.00001263, -0.00002249, -0.00000313, 0.00000159, -0.00000208, -0.00000506, 0.00000299, -0.00001037, -0.00000109, -0.00000023, -0.00000400, -0.00000052, -0.00000891, 0.00000842, -0.00000631, -0.00000572, -0.00000266, -0.00000685, -0.00000584, -0.00004775, -0.00003663, 0.00009255, 0.00004948, 0.00007785, 0.00004217, -0.00006655, -0.00002130, -0.00000732, -0.00000901, -0.00000528, -0.00000418, -0.00000455, -0.00000452, -0.00000425, -0.00000630, -0.00001026, -0.00000591, -0.00004994, -0.00001639, 0.00007330, 0.00003945, 0.00006503, 0.00005090, -0.00005159, -0.00002716, -0.00000338, -0.00000602, -0.00000275, -0.00000793, 0.00000414, -0.00000018, -0.00000823, 0.00000037, -0.00000380, 0.00000296, -0.00001085, -0.00000003, -0.00000135, -0.00000317, -0.00000270, 0.00000485, -0.00001639, -0.00000808, 0.00000964, -0.00000477, 0.00000473, -0.00000931, 0.00001194, -0.00000125, 0.00000221, -0.00000039, 0.00000355, -0.00000333, 0.00000232, -0.00000249, 0.00000036, 0.00000434, 0.00000101, 0.00000466, -0.00000158, -0.00000233, 0.00000046, 0.00000286, -0.00000386, 0.00000460, -0.00000645, 0.00000301, -0.00000063, 0.00000436, -0.00000623, 0.00000116, -0.00000299, 0.00000030, 0.00000750, -0.00000302, },
		{ -0.00000683, -0.00000473, 0.00000225, 0.00000646, -0.00000469, 0.00000172, -0.00000435, -0.00000036, 0.00000165, 0.00000140, -0.00000188, -0.00000111, 0.00000230, -0.00000395, 0.00000422, -0.00000267, -0.00000031, 0.00000063, 0.00000467, -0.00000019, 0.00000544, -0.00000478, 0.00000425, -0.00000448, 0.00000364, -0.00000256, 0.00000529, -0.00000434, 0.00000991, 0.00000117, -0.00000508, 0.00000315, -0.00000249, 0.00000957, -0.00002418, 0.00000106, -0.00000028, -0.00000187, -0.00000493, 0.00000334, -0.00001169, 0.00000110, -0.00000230, -0.00000173, -0.00000330, -0.00000642, 0.00000744, -0.00000683, -0.00000522, -0.00000274, -0.00000722, -0.00000731, -0.00005653, -0.00001274, 0.00009276, 0.00004126, 0.00008126, 0.00002463, -0.00006410, -0.00001631, -0.00000763, -0.00000888, -0.00000454, -0.00000453, -0.00000432, -0.00000472, -0.00000401, -0.00000748, -0.00000923, -0.00000861, -0.00005538, -0.00000498, 0.00007881, 0.00003712, 0.00007678, 0.00003344, -0.00005907, -0.00001859, -0.00000487, -0.00000484, -0.00000346, -0.00000774, 0.00000529, -0.00000200, -0.00000719, 0.00000040, -0.00000395, 0.00000294, -0.00001174, 0.00000162, -0.00000286, -0.00000256, -0.00000239, 0.00000490, -0.00002052, -0.00000180, 0.00000635, -0.00000310, 0.00000255, -0.00000757, 0.00001333, -0.00000380, 0.00000414, -0.00000161, 0.00000406, -0.00000434, 0.00000359, -0.00000396, 0.00000268, 0.00000250, 0.00000242, 0.00000326, -0.00000113, -0.00000280, 0.00000230, 0.00000032, -0.00000199, 0.00000313, -0.00000562, 0.00000320, -0.00000041, 0.00000354, -0.00000671, 0.00000247, -0.00000459, 0.00000300, 0.00000577, -0.00000368, },
		},
		{ { 0.00000484, 0.00000359, -0.00000349, -0.00000117, -0.00000186, -0.00000024, -0.00000005, -0.00000015, 0.00000042, -0.00000305, 0.00000309, -0.00000244, 0.00000055, 0.00000008, 0.00000316, -0.00000010, -0.00000124, -0.00000189, 0.00000055, 0.00000094, 0.00000008, 0.00000356, -0.00000654, 0.00000145, -0.00000586, 0.00000218, 0.00000153, 0.00000122, 0.00000330, -0.00000282, -0.00000588, -0.00000148, -0.00000047, -0.00000049, 0.00001278, -0.00000090, -0.00000674, 0.00000045, -0.00000109, -0.00000122, 0.00001323, -0.00000437, 0.00000037, -0.00000453, 0.00000781, 0.00000091, -0.00002095, 0.00000181, -0.00000210, -0.00000565, -0.00000013, -0.00000245, -0.00001782, -0.00000227, -0.00002009, -0.00001109, -0.00005960, 0.00000890, 0.00009720, 0.00003096, 0.00008009, 0.00001057, -0.00005912, -0.00001759, -0.00002172, -0.00001759, -0.00005912, 0.00001057, 0.00008009, 0.00003096, 0.00009720, 0.00000890, -0.00005960, -0.00001109, -0.00002009, -0.00000227, -0.00001782, -0.00000245, -0.00000013, -0.00000565, -0.00000210, 0.00000181, -0.00002095, 0.00000091, 0.00000781, -0.00000453, 0.00000037, -0.00000437, 0.00001323, -0.00000122, -0.00000109, 0.00000045, -0.00000674, -0.00000090, 0.00001278, -0.00000049, -0.00000047, -0.00000148, -0.00000588, -0.00000282, 0.00000330, 0.00000122, 0.00000153, 0.00000218, -0.00000586, 0.00000145, -0.00000654, 0.00000356, 0.00000008, 0.00000094, 0.00000055, -0.00000189, -0.00000124, -0.00000010, 0.00000316, 0.00000008, 0.00000055, -0.00000244, 0.00000309, -0.00000305, 0.00000042, -0.00000015, -0.00000005, -0.00000024, -0.00000186, -0.00000117, -0.00000349, 0.00000359, },
		{ 0.00000488, 0.00000299, -0.00000392, -0.00000067, -0.00000207, 0.00000020, -0.00000050, 0.00000039, -0.00000031, -0.00000241, 0.00000296, -0.00000275, 0.00000063, 0.00000027, 0.00000342, -0.00000081, -0.00000062, -0.00000242, 0.00000122, 0.00000038, 0.00000092, 0.00000240, -0.00000610, 0.00000131, -0.00000609, 0.00000302, 0.00000099, 0.00000179, 0.00000280, -0.00000278, -0.00000644, -0.00000052, -0.00000146, 0.00000117, 0.00001226, -0.00000179, -0.00000673, 0.00000107, -0.00000195, 0.00000036, 0.00001288, -0.00000580, 0.00000144, -0.00000552, 0.00000996, -0.00000244, -0.00001963, 0.00000281, -0.00000282, -0.00000599, 0.00000102, -0.00000374, -0.00001699, -0.00000350, -0.00001861, -0.00001422, -0.00005951, 0.00002354, 0.00008968, 0.00002929, 0.00007771, 0.00000416, -0.00005613, -0.00001692, -0.00002174, -0.00001835, -0.00006220, 0.00001781, 0.00008171, 0.00003396, 0.00010342, -0.00000731, -0.00005707, -0.00000932, -0.00002098, -0.00000132, -0.00001847, -0.00000132, -0.00000120, -0.00000529, -0.00000134, 0.00000041, -0.00002162, 0.00000395, 0.00000550, -0.00000321, -0.00000104, -0.00000240, 0.00001289, -0.00000236, -0.00000044, -0.00000009, -0.00000680, 0.00000022, 0.00001292, -0.00000194, 0.00000048, -0.00000246, -0.00000519, -0.00000303, 0.00000393, 0.00000053, 0.00000222, 0.00000110, -0.00000531, 0.00000124, -0.00000664, 0.00000451, -0.00000071, 0.00000147, -0.00000008, -0.00000137, -0.00000191, 0.00000080, 0.00000265, 0.00000009, 0.00000038, -0.00000199, 0.00000299, -0.00000349, 0.00000106, -0.00000066, 0.00000040, -0.00000072, -0.00000163, -0.00000172, -0.00000286, 0.00000399, },
		{ 0.00000500, 0.00000220, -0.00000412, -0.00000027, -0.00000220, 0.00000056, -0.00000089, 0.00000090, -0.00000109, -0.00000160, 0.00000261, -0.00000292, 0.00000062, 0.00000063, 0.00000341, -0.00000128, -0.00000015, -0.00000289, 0.00000185, -0.00000011, 0.00000171, 0.00000113, -0.00000542, 0.00000089, -0.00000599, 0.00000362, 0.00000062, 0.00000221, 0.00000243, -0.00000294, -0.00000680, 0.00000034, -0.00000238, 0.00000295, 0.00001141, -0.00000246, -0.00000677, 0.00000175, -0.00000292, 0.00000233, 0.00001185, -0.00000661, 0.00000204, -0.00000599, 0.00001169, -0.00000585, -0.00001786, 0.00000348, -0.00000356, -0.00000619, 0.00000212, -0.00000505, -0.00001617, -0.00000474, -0.00001691, -0.00001829, -0.00005691, 0.00003618, 0.00008144, 0.00002868, 0.00007465, -0.00000145, -0.00005324, -0.00001634, -0.00002180, -0.00001922, -0.00006534, 0.00002590, 0.00008249, 0.00003852, 0.00010768, -0.00002444, -0.00005201, -0.00000911, -0.00002101, -0.00000088, -0.00001877, -0.00000051, -0.00000204, -0.00000502, -0.00000047, -0.00000143, -0.00002152, 0.00000641, 0.00000332, -0.00000178, -0.00000262, -0.00000004, 0.00001191, -0.00000303, -0.00000002, -0.00000051, -0.00000693, 0.00000157, 0.00001263, -0.00000309, 0.00000129, -0.00000335, -0.00000447, -0.00000332, 0.00000462, -0.00000025, 0.00000303, -0.00000017, -0.00000444, 0.00000067, -0.00000636, 0.00000517, -0.00000135, 0.00000190, -0.00000062, -0.00000095, -0.00000254, 0.00000178, 0.00000194, 0.00000028, 0.00000011, -0.00000142, 0.00000267, -0.00000370, 0.00000154, -0.00000108, 0.00000081, -0.00000120, -0.00000140, -0.00000226, -0.00000208, 0.00000423, },
		{ 0.00000516, 0.00000124, -0.00000406, -0.00000001, -0.00000223, 0.00000079, -0.00000119, 0.00000134, -0.00000185, -0.00000069, 0.00000207, -0.00000296, 0.00000055, 0.00000114, 0.00000314, -0.00000148, 0.00000011, -0.00000321, 0.00000237, -0.00000049, 0.00000237, -0.00000013, -0.00000459, 0.00000025, -0.00000562, 0.00000399, 0.00000040, 0.00000249, 0.00000220, -0.00000332, -0.00000689, 0.00000101, -0.00000317, 0.00000473, 0.00001032, -0.00000294, -0.00000682, 0.00000241, -0.00000391, 0.00000456, 0.00001020, -0.00000680, 0.00000213, -0.00000583, 0.00001279, -0.00000907, -0.00001580, 0.00000393, -0.00000439, -0.00000614, 0.00000304, -0.00000623, -0.00001555, -0.00000573, -0.00001536, -0.00002282, -0.00005210, 0.00004657, 0.00007294, 0.00002888, 0.00007100, -0.00000630, -0.00005050, -0.00001582, -0.00002191, -0.00002022, -0.00006848, 0.00003484, 0.00008239, 0.00004484, 0.00010932, -0.00004168, -0.00004478, -0.00001054, -0.00002002, -0.00000110, -0.00001859, -0.00000011, -0.00000258, -0.00000490, 0.00000054, -0.00000374, -0.00002055, 0.00000808, 0.00000152, -0.00000049, -0.00000413, 0.00000253, 0.00001042, -0.00000325, 0.00000016, -0.00000080, -0.00000710, 0.00000311, 0.00001192, -0.00000390, 0.00000191, -0.00000407, -0.00000383, -0.00000362, 0.00000532, -0.00000106, 0.00000387, -0.00000158, -0.00000330, -0.00000024, -0.00000568, 0.00000547, -0.00000178, 0.00000218, -0.00000100, -0.00000070, -0.00000304, 0.00000277, 0.00000109, 0.00000061, -0.00000024, -0.00000076, 0.00000214, -0.00000368, 0.00000186, -0.00000138, 0.00000113, -0.00000164, -0.00000124, -0.00000272, -0.00000122, 0.00000433, },
		{ 0.00000531, 0.00000016, -0.00000375, 0.00000010, -0.00000214, 0.00000089, -0.00000135, 0.00000166, -0.00000253, 0.00000026, 0.00000139, -0.00000290, 0.00000045, 0.00000174, 0.00000265, -0.00000141, 0.00000013, -0.00000335, 0.00000275, -0.00000071, 0.00000281, -0.00000128, -0.00000373, -0.00000052, -0.00000503, 0.00000416, 0.00000032, 0.00000265, 0.00000207, -0.00000392, -0.00000668, 0.00000143, -0.00000372, 0.00000641, 0.00000906, -0.00000328, -0.00000684, 0.00000299, -0.00000482, 0.00000689, 0.00000805, -0.00000641, 0.00000172, -0.00000501, 0.00001313, -0.00001187, -0.00001362, 0.00000422, -0.00000536, -0.00000579, 0.00000368, -0.00000718, -0.00001525, -0.00000625, -0.00001432, -0.00002728, -0.00004554, 0.00005467, 0.00006455, 0.00002970, 0.00006685, -0.00001041, -0.00004793, -0.00001535, -0.00002206, -0.00002136, -0.00007157, 0.00004460, 0.00008145, 0.00005303, 0.00010767, -0.00005811, -0.00003593, -0.00001346, -0.00001800, -0.00000203, -0.00001788, -0.00000016, -0.00000277, -0.00000496, 0.00000165, -0.00000644, -0.00001873, 0.00000884, 0.00000034, 0.00000042, -0.00000536, 0.00000508, 0.00000857, -0.00000308, 0.00000010, -0.00000099, -0.00000728, 0.00000480, 0.00001080, -0.00000433, 0.00000226, -0.00000455, -0.00000336, -0.00000382, 0.00000594, -0.00000182, 0.00000467, -0.00000302, -0.00000198, -0.00000144, -0.00000460, 0.00000539, -0.00000196, 0.00000226, -0.00000118, -0.00000067, -0.00000331, 0.00000365, 0.00000019, 0.00000104, -0.00000062, -0.00000005, 0.00000145, -0.00000344, 0.00000198, -0.00000154, 0.00000132, -0.00000200, -0.00000115, -0.00000306, -0.00000036, 0.00000436, },
		{ 0.00000539, -0.00000098, -0.00000324, 0.00000004, -0.00000193, 0.00000083, -0.00000135, 0.00000180, -0.00000307, 0.00000118, 0.00000060, -0.00000276, 0.00000037, 0.00000237, 0.00000201, -0.00000109, -0.00000009, -0.00000329, 0.00000297, -0.00000072, 0.00000300, -0.00000223, -0.00000292, -0.00000132, -0.00000431, 0.00000418, 0.00000034, 0.00000274, 0.00000199, -0.00000468, -0.00000617, 0.00000156, -0.00000397, 0.00000791, 0.00000773, -0.00000353, -0.00000680, 0.00000342, -0.00000551, 0.00000917, 0.00000557, -0.00000557, 0.00000090, -0.00000359, 0.00001266, -0.00001413, -0.00001145, 0.00000439, -0.00000646, -0.00000510, 0.00000401, -0.00000786, -0.00001533, -0.00000617, -0.00001407, -0.00003116, -0.00003777, 0.00006058, 0.00005654, 0.00003094, 0.00006227, -0.00001380, -0.00004555, -0.00001491, -0.00002225, -0.00002271, -0.00007451, 0.00005510, 0.00007978, 0.00006307, 0.00010219, -0.00007275, -0.00002618, -0.00001754, -0.00001508, -0.00000364, -0.00001669, -0.00000061, -0.00000264, -0.00000517, 0.00000280, -0.00000942, -0.00001613, 0.00000864, -0.00000007, 0.00000078, -0.00000609, 0.00000739, 0.00000654, -0.00000263, -0.00000014, -0.00000109, -0.00000742, 0.00000656, 0.00000932, -0.00000437, 0.00000234, -0.00000472, -0.00000315, -0.00000382, 0.00000640, -0.00000243, 0.00000533, -0.00000439, -0.00000059, -0.00000281, -0.00000320, 0.00000494, -0.00000187, 0.00000213, -0.00000115, -0.00000091, -0.00000328, 0.00000433, -0.00000067, 0.00000151, -0.00000100, 0.00000067, 0.00000065, -0.00000301, 0.00000193, -0.00000154, 0.00000138, -0.00000227, -0.00000118, -0.00000323, 0.00000044, 0.00000438, },
		{ 0.00000534, -0.00000210, -0.00000255, -0.00000015, -0.00000161, 0.00000063, -0.00000119, 0.00000176, -0.00000342, 0.00000201, -0.00000024, -0.00000258, 0.00000034, 0.00000296, 0.00000128, -0.00000061, -0.00000052, -0.00000302, 0.00000301, -0.00000055, 0.00000290, -0.00000293, -0.00000226, -0.00000208, -0.00000352, 0.00000411, 0.00000042, 0.00000280, 0.00000191, -0.00000556, -0.00000538, 0.00000139, -0.00000388, 0.00000912, 0.00000642, -0.00000374, -0.00000665, 0.00000363, -0.00000588, 0.00001121, 0.00000296, -0.00000444, -0.00000017, -0.00000170, 0.00001140, -0.00001575, -0.00000937, 0.00000445, -0.00000767, -0.00000412, 0.00000403, -0.00000829, -0.00001576, -0.00000548, -0.00001476, -0.00003404, -0.00002933, 0.00006449, 0.00004906, 0.00003246, 0.00005735, -0.00001653, -0.00004338, -0.00001451, -0.00002244, -0.00002432, -0.00007715, 0.00006619, 0.00007758, 0.00007474, 0.00009253, -0.00008467, -0.00001634, -0.00002232, -0.00001151, -0.00000574, -0.00001513, -0.00000137, -0.00000228, -0.00000543, 0.00000386, -0.00001247, -0.00001293, 0.00000755, 0.00000033, 0.00000045, -0.00000618, 0.00000926, 0.00000451, -0.00000201, -0.00000051, -0.00000115, -0.00000746, 0.00000830, 0.00000755, -0.00000406, 0.00000214, -0.00000459, -0.00000327, -0.00000355, 0.00000660, -0.00000282, 0.00000576, -0.00000557, 0.00000074, -0.00000426, -0.00000157, 0.00000418, -0.00000152, 0.00000181, -0.00000092, -0.00000141, -0.00000293, 0.00000474, -0.00000142, 0.00000197, -0.00000133, 0.00000132, -0.00000020, -0.00000244, 0.00000172, -0.00000140, 0.00000130, -0.00000243, -0.00000132, -0.00000320, 0.00000113, 0.00000443, },
		{ 0.00000510, -0.00000311, -0.00000177, -0.00000044, -0.00000119, 0.00000031, -0.00000087, 0.00000152, -0.00000354, 0.00000269, -0.00000108, -0.00000237, 0.00000041, 0.00000345, 0.00000056, -0.00000002, -0.00000112, -0.00000258, 0.00000292, -0.00000022, 0.00000253, -0.00000334, -0.00000179, -0.00000271, -0.00000271, 0.00000398, 0.00000054, 0.00000285, 0.00000178, -0.00000646, -0.00000437, 0.00000094, -0.00000343, 0.00001000, 0.00000520, -0.00000398, -0.00000635, 0.00000360, -0.00000585, 0.00001284, 0.00000042, -0.00000321, -0.00000129, 0.00000046, 0.00000948, -0.00001675, -0.00000742, 0.00000437, -0.00000889, -0.00000292, 0.00000383, -0.00000858, -0.00001643, -0.00000427, -0.00001643, -0.00003561, -0.00002075, 0.00006667, 0.00004221, 0.00003412, 0.00005215, -0.00001863, -0.00004141, -0.00001417, -0.00002259, -0.00002629, -0.00007932, 0.00007766, 0.00007519, 0.00008764, 0.00007854, -0.00009305, -0.00000728, -0.00002722, -0.00000766, -0.00000811, -0.00001339, -0.00000226, -0.00000183, -0.00000563, 0.00000464, -0.00001536, -0.00000941, 0.00000575, 0.00000150, -0.00000060, -0.00000552, 0.00001051, 0.00000266, -0.00000133, -0.00000094, -0.00000121, -0.00000734, 0.00000992, 0.00000559, -0.00000345, 0.00000169, -0.00000415, -0.00000372, -0.00000295, 0.00000648, -0.00000291, 0.00000586, -0.00000644, 0.00000187, -0.00000562, 0.00000017, 0.00000320, -0.00000099, 0.00000133, -0.00000055, -0.00000213, -0.00000225, 0.00000483, -0.00000199, 0.00000235, -0.00000158, 0.00000187, -0.00000103, -0.00000178, 0.00000140, -0.00000115, 0.00000108, -0.00000249, -0.00000157, -0.00000296, 0.00000165, 0.00000457, },
		{ 0.00000463, -0.00000394, -0.00000096, -0.00000080, -0.00000073, -0.00000011, -0.00000042, 0.00000108, -0.00000342, 0.00000318, -0.00000188, -0.00000217, 0.00000060, 0.00000376, -0.00000006, 0.00000057, -0.00000180, -0.00000203, 0.00000275, 0.00000021, 0.00000194, -0.00000348, -0.00000154, -0.00000319, -0.00000195, 0.00000382, 0.00000066, 0.00000293, 0.00000152, -0.00000733, -0.00000320, 0.00000024, -0.00000263, 0.00001050, 0.00000414, -0.00000426, -0.00000590, 0.00000330, -0.00000536, 0.00001394, -0.00000187, -0.00000205, -0.00000226, 0.00000266, 0.00000709, -0.00001718, -0.00000558, 0.00000409, -0.00001000, -0.00000163, 0.00000354, -0.00000887, -0.00001713, -0.00000275, -0.00001892, -0.00003575, -0.00001246, 0.00006740, 0.00003608, 0.00003580, 0.00004680, -0.00002019, -0.00003960, -0.00001393, -0.00002262, -0.00002877, -0.00008076, 0.00008924, 0.00007303, 0.00010114, 0.00006041, -0.00009730, 0.00000023, -0.00003166, -0.00000395, -0.00001045, -0.00001171, -0.00000308, -0.00000148, -0.00000559, 0.00000495, -0.00001779, -0.00000588, 0.00000351, 0.00000329, -0.00000230, -0.00000409, 0.00001106, 0.00000113, -0.00000072, -0.00000136, -0.00000131, -0.00000701, 0.00001133, 0.00000355, -0.00000262, 0.00000107, -0.00000347, -0.00000451, -0.00000200, 0.00000601, -0.00000265, 0.00000556, -0.00000690, 0.00000268, -0.00000676, 0.00000187, 0.00000208, -0.00000032, 0.00000076, -0.00000011, -0.00000299, -0.00000127, 0.00000458, -0.00000234, 0.00000262, -0.00000171, 0.00000228, -0.00000179, -0.00000109, 0.00000101, -0.00000081, 0.00000076, -0.00000247, -0.00000189, -0.00000252, 0.00000200, 0.00000482, },
		{ 0.00000389, -0.00000452, -0.00000020, -0.00000117, -0.00000026, -0.00000057, 0.00000012, 0.00000049, -0.00000306, 0.00000345, -0.00000261, -0.00000197, 0.00000091, 0.00000387, -0.00000051, 0.00000107, -0.00000249, -0.00000142, 0.00000255, 0.00000068, 0.00000118, -0.00000338, -0.00000151, -0.00000348, -0.00000126, 0.00000367, 0.00000078, 0.00000304, 0.00000112, -0.00000806, -0.00000196, -0.00000066, -0.00000151, 0.00001060, 0.00000328, -0.00000463, -0.00000530, 0.00000273, -0.00000438, 0.00001442, -0.00000376, -0.00000113, -0.00000288, 0.00000463, 0.00000444, -0.00001715, -0.00000382, 0.00000353, -0.00001087, -0.00000041, 0.00000329, -0.00000936, -0.00001762, -0.00000122, -0.00002195, -0.00003449, -0.00000475, 0.00006693, 0.00003069, 0.00003736, 0.00004140, -0.00002130, -0.00003791, -0.00001387, -0.00002244, -0.00003191, -0.00008123, 0.00010062, 0.00007161, 0.00011440, 0.00003861, -0.00009711, 0.00000554, -0.00003509, -0.00000077, -0.00001245, -0.00001032, -0.00000362, -0.00000141, -0.00000516, 0.00000462, -0.00001950, -0.00000270, 0.00000115, 0.00000545, -0.00000450, -0.00000197, 0.00001086, 0.00000002, -0.00000025, -0.00000171, -0.00000148, -0.00000644, 0.00001244, 0.00000154, -0.00000164, 0.00000035, -0.00000263, -0.00000556, -0.00000072, 0.00000518, -0.00000202, 0.00000484, -0.00000688, 0.00000306, -0.00000755, 0.00000340, 0.00000097, 0.00000038, 0.00000018, 0.00000033, -0.00000392, -0.00000005, 0.00000402, -0.00000244, 0.00000275, -0.00000170, 0.00000250, -0.00000241, -0.00000043, 0.00000060, -0.00000043, 0.00000037, -0.00000241, -0.00000224, -0.00000193, 0.00000218, 0.00000518, },
		{ 0.00000290, -0.00000479, 0.00000045, -0.00000151, 0.00000017, -0.00000103, 0.00000071, -0.00000022, -0.00000249, 0.00000348, -0.00000324, -0.00000179, 0.00000134, 0.00000375, -0.00000074, 0.00000140, -0.00000310, -0.00000083, 0.00000239, 0.00000109, 0.00000034, -0.00000310, -0.00000166, -0.00000359, -0.00000066, 0.00000353, 0.00000089, 0.00000318, 0.00000055, -0.00000861, -0.00000072, -0.00000167, -0.00000012, 0.00001032, 0.00000264, -0.00000508, -0.00000456, 0.00000196, -0.00000296, 0.00001423, -0.00000516, -0.00000056, -0.00000301, 0.00000616, 0.00000176, -0.00001679, -0.00000208, 0.00000261, -0.00001137, 0.00000061, 0.00000325, -0.00001021, -0.00001764, -0.00000004, -0.00002516, -0.00003202, 0.00000220, 0.00006546, 0.00002611, 0.00003866, 0.00003609, -0.00002208, -0.00003624, -0.00001406, -0.00002193, -0.00003588, -0.00008043, 0.00011147, 0.00007149, 0.00012642, 0.00001398, -0.00009249, 0.00000822, -0.00003709, 0.00000152, -0.00001385, -0.00000943, -0.00000371, -0.00000176, -0.00000421, 0.00000349, -0.00002024, -0.00000022, -0.00000096, 0.00000768, -0.00000697, 0.00000069, 0.00000996, -0.00000062, 0.00000001, -0.00000197, -0.00000174, -0.00000561, 0.00001318, -0.00000034, -0.00000063, -0.00000037, -0.00000174, -0.00000680, 0.00000083, 0.00000404, -0.00000106, 0.00000369, -0.00000637, 0.00000295, -0.00000788, 0.00000462, -0.00000003, 0.00000102, -0.00000033, 0.00000067, -0.00000480, 0.00000129, 0.00000319, -0.00000231, 0.00000274, -0.00000155, 0.00000253, -0.00000287, 0.00000014, 0.00000023, -0.00000007, -0.00000006, -0.00000233, -0.00000258, -0.00000122, 0.00000222, 0.00000564, },
		{ 0.00000167, -0.00000475, 0.00000094, -0.00000176, 0.00000052, -0.00000143, 0.00000128, -0.00000100, -0.00000173, 0.00000328, -0.00000377, -0.00000161, 0.00000186, 0.00000342, -0.00000072, 0.00000150, -0.00000357, -0.00000032, 0.00000233, 0.00000137, -0.00000047, -0.00000271, -0.00000196, -0.00000353, -0.00000014, 0.00000340, 0.00000102, 0.00000333, -0.00000019, -0.00000892, 0.00000043, -0.00000271, 0.00000146, 0.00000970, 0.00000223, -0.00000562, -0.00000373, 0.00000103, -0.00000117, 0.00001340, -0.00000601, -0.00000040, -0.00000255, 0.00000705, -0.00000073, -0.00001622, -0.00000033, 0.00000130, -0.00001141, 0.00000130, 0.00000354, -0.00001160, -0.00001693, 0.00000045, -0.00002809, -0.00002866, 0.00000835, 0.00006314, 0.00002239, 0.00003956, 0.00003102, -0.00002267, -0.00003451, -0.00001460, -0.00002100, -0.00004081, -0.00007812, 0.00012149, 0.00007319, 0.00013609, -0.00001239, -0.00008379, 0.00000810, -0.00003736, 0.00000269, -0.00001447, -0.00000918, -0.00000322, -0.00000265, -0.00000266, 0.00000148, -0.00001984, 0.00000127, -0.00000246, 0.00000964, -0.00000944, 0.00000369, 0.00000845, -0.00000079, 0.00000004, -0.00000211, -0.00000208, -0.00000451, 0.00001350, -0.00000199, 0.00000032, -0.00000100, -0.00000093, -0.00000810, 0.00000256, 0.00000264, 0.00000019, 0.00000217, -0.00000537, 0.00000232, -0.00000769, 0.00000543, -0.00000081, 0.00000152, -0.00000071, 0.00000081, -0.00000551, 0.00000266, 0.00000217, -0.00000197, 0.00000259, -0.00000127, 0.00000236, -0.00000314, 0.00000060, -0.00000007, 0.00000024, -0.00000047, -0.00000228, -0.00000285, -0.00000048, 0.00000216, 0.00000615, },
		{ 0.00000026, -0.00000438, 0.00000123, -0.00000190, 0.00000076, -0.00000174, 0.00000179, -0.00000178, -0.00000083, 0.00000286, -0.00000418, -0.00000140, 0.00000243, 0.00000291, -0.00000044, 0.00000135, -0.00000383, 0.00000007, 0.00000242, 0.00000145, -0.00000118, -0.00000232, -0.00000233, -0.00000335, 0.00000031, 0.00000326, 0.00000118, 0.00000344, -0.00000107, -0.00000896, 0.00000143, -0.00000371, 0.00000314, 0.00000878, 0.00000202, -0.00000621, -0.00000285, 0.00000004, 0.00000089, 0.00001201, -0.00000635, -0.00000063, -0.00000148, 0.00000720, -0.00000286, -0.00001556, 0.00000146, -0.00000043, -0.00001093, 0.00000158, 0.00000421, -0.00001361, -0.00001535, -0.00000005, -0.00003030, -0.00002483, 0.00001379, 0.00006006, 0.00001958, 0.00003990, 0.00002636, -0.00002322, -0.00003264, -0.00001555, -0.00001957, -0.00004682, -0.00007407, 0.00013046, 0.00007716, 0.00014228, -0.00003916, -0.00007172, 0.00000530, -0.00003586, 0.00000262, -0.00001421, -0.00000964, -0.00000212, -0.00000409, -0.00000053, -0.00000140, -0.00001822, 0.00000155, -0.00000304, 0.00001101, -0.00001160, 0.00000676, 0.00000650, -0.00000056, -0.00000017, -0.00000213, -0.00000247, -0.00000317, 0.00001337, -0.00000334, 0.00000115, -0.00000143, -0.00000032, -0.00000931, 0.00000435, 0.00000111, 0.00000161, 0.00000037, -0.00000395, 0.00000121, -0.00000695, 0.00000578, -0.00000129, 0.00000182, -0.00000089, 0.00000070, -0.00000595, 0.00000392, 0.00000105, -0.00000147, 0.00000234, -0.00000090, 0.00000203, -0.00000321, 0.00000092, -0.00000027, 0.00000046, -0.00000084, -0.00000229, -0.00000300, 0.00000023, 0.00000208, 0.00000666, },
		{ -0.00000125, -0.00000372, 0.00000131, -0.00000190, 0.00000086, -0.00000192, 0.00000219, -0.00000250, 0.00000013, 0.00000226, -0.00000449, -0.00000115, 0.00000298, 0.00000228, 0.00000004, 0.00000094, -0.00000387, 0.00000031, 0.00000269, 0.00000127, -0.00000171, -0.00000198, -0.00000271, -0.00000309, 0.00000070, 0.00000310, 0.00000140, 0.00000348, -0.00000205, -0.00000872, 0.00000221, -0.00000458, 0.00000485, 0.00000765, 0.00000198, -0.00000684, -0.00000199, -0.00000091, 0.00000307, 0.00001017, -0.00000625, -0.00000116, 0.00000013, 0.00000657, -0.00000451, -0.00001489, 0.00000327, -0.00000253, -0.00000994, 0.00000147, 0.00000527, -0.00001624, -0.00001280, -0.00000175, -0.00003139, -0.00002093, 0.00001863, 0.00005626, 0.00001775, 0.00003956, 0.00002225, -0.00002385, -0.00003054, -0.00001696, -0.00001762, -0.00005396, -0.00006814, 0.00013825, 0.00008369, 0.00014392, -0.00006486, -0.00005723, 0.00000021, -0.00003272, 0.00000138, -0.00001313, -0.00001073, -0.00000045, -0.00000601, 0.00000210, -0.00000504, -0.00001543, 0.00000054, -0.00000249, 0.00001152, -0.00001317, 0.00000964, 0.00000430, -0.00000001, -0.00000057, -0.00000206, -0.00000288, -0.00000166, 0.00001281, -0.00000433, 0.00000178, -0.00000160, -0.00000003, -0.00001029, 0.00000604, -0.00000043, 0.00000309, -0.00000158, -0.00000223, -0.00000031, -0.00000571, 0.00000564, -0.00000144, 0.00000188, -0.00000085, 0.00000029, -0.00000604, 0.00000498, -0.00000006, -0.00000089, 0.00000203, -0.00000047, 0.00000156, -0.00000311, 0.00000109, -0.00000034, 0.00000056, -0.00000113, -0.00000237, -0.00000298, 0.00000085, 0.00000202, 0.00000708, },
		{ -0.00000277, -0.00000281, 0.00000118, -0.00000176, 0.00000082, -0.00000196, 0.00000244, -0.00000311, 0.00000111, 0.00000149, -0.00000470, -0.00000083, 0.00000346, 0.00000162, 0.00000067, 0.00000030, -0.00000370, 0.00000040, 0.00000312, 0.00000083, -0.00000200, -0.00000177, -0.00000304, -0.00000277, 0.00000105, 0.00000292, 0.00000169, 0.00000339, -0.00000308, -0.00000821, 0.00000273, -0.00000525, 0.00000647, 0.00000637, 0.00000208, -0.00000746, -0.00000119, -0.00000172, 0.00000522, 0.00000805, -0.00000583, -0.00000188, 0.00000215, 0.00000521, -0.00000564, -0.00001422, 0.00000504, -0.00000490, -0.00000851, 0.00000101, 0.00000661, -0.00001939, -0.00000936, -0.00000472, -0.00003108, -0.00001738, 0.00002307, 0.00005179, 0.00001692, 0.00003844, 0.00001879, -0.00002465, -0.00002819, -0.00001883, -0.00001518, -0.00006219, -0.00006032, 0.00014492, 0.00009283, 0.00014018, -0.00008800, -0.00004149, -0.00000657, -0.00002827, -0.00000083, -0.00001140, -0.00001231, 0.00000161, -0.00000826, 0.00000502, -0.00000922, -0.00001164, -0.00000173, -0.00000070, 0.00001097, -0.00001389, 0.00001205, 0.00000206, 0.00000075, -0.00000110, -0.00000194, -0.00000324, -0.00000003, 0.00001185, -0.00000496, 0.00000219, -0.00000147, -0.00000015, -0.00001091, 0.00000750, -0.00000184, 0.00000446, -0.00000353, -0.00000036, -0.00000211, -0.00000403, 0.00000505, -0.00000126, 0.00000169, -0.00000060, -0.00000042, -0.00000572, 0.00000572, -0.00000108, -0.00000029, 0.00000171, -0.00000003, 0.00000101, -0.00000286, 0.00000112, -0.00000029, 0.00000052, -0.00000134, -0.00000255, -0.00000277, 0.00000132, 0.00000207, 0.00000734, },
		{ -0.00000421, -0.00000174, 0.00000088, -0.00000150, 0.00000064, -0.00000183, 0.00000252, -0.00000357, 0.00000205, 0.00000062, -0.00000482, -0.00000042, 0.00000380, 0.00000100, 0.00000136, -0.00000051, -0.00000334, 0.00000037, 0.00000370, 0.00000013, -0.00000203, -0.00000172, -0.00000327, -0.00000245, 0.00000138, 0.00000272, 0.00000207, 0.00000314, -0.00000408, -0.00000747, 0.00000299, -0.00000567, 0.00000793, 0.00000504, 0.00000225, -0.00000804, -0.00000052, -0.00000229, 0.00000717, 0.00000582, -0.00000526, -0.00000259, 0.00000438, 0.00000324, -0.00000626, -0.00001355, 0.00000667, -0.00000740, -0.00000679, 0.00000035, 0.00000807, -0.00002284, -0.00000521, -0.00000887, -0.00002924, -0.00001448, 0.00002726, 0.00004669, 0.00001708, 0.00003648, 0.00001606, -0.00002569, -0.00002560, -0.00002109, -0.00001238, -0.00007137, -0.00005070, 0.00015070, 0.00010432, 0.00013047, -0.00010722, -0.00002576, -0.00001424, -0.00002300, -0.00000364, -0.00000930, -0.00001411, 0.00000383, -0.00001059, 0.00000797, -0.00001361, -0.00000717, -0.00000507, 0.00000231, 0.00000926, -0.00001361, 0.00001378, -0.00000001, 0.00000157, -0.00000171, -0.00000181, -0.00000352, 0.00000163, 0.00001055, -0.00000523, 0.00000238, -0.00000102, -0.00000074, -0.00001106, 0.00000860, -0.00000296, 0.00000559, -0.00000531, 0.00000150, -0.00000404, -0.00000206, 0.00000407, -0.00000077, 0.00000130, -0.00000017, -0.00000141, -0.00000499, 0.00000610, -0.00000192, 0.00000026, 0.00000142, 0.00000036, 0.00000043, -0.00000250, 0.00000105, -0.00000013, 0.00000036, -0.00000145, -0.00000282, -0.00000236, 0.00000159, 0.00000226, 0.00000737, },
		{ -0.00000546, -0.00000057, 0.00000043, -0.00000114, 0.00000035, -0.00000157, 0.00000242, -0.00000383, 0.00000288, -0.00000033, -0.00000486, 0.00000010, 0.00000396, 0.00000050, 0.00000202, -0.00000141, -0.00000284, 0.00000027, 0.00000437, -0.00000078, -0.00000181, -0.00000187, -0.00000338, -0.00000213, 0.00000168, 0.00000250, 0.00000251, 0.00000271, -0.00000501, -0.00000654, 0.00000297, -0.00000580, 0.00000915, 0.00000373, 0.00000245, -0.00000855, 0.00000000, -0.00000254, 0.00000880, 0.00000364, -0.00000468, -0.00000312, 0.00000659, 0.00000085, -0.00000645, -0.00001283, 0.00000803, -0.00000984, -0.00000495, -0.00000032, 0.00000941, -0.00002631, -0.00000066, -0.00001396, -0.00002589, -0.00001242, 0.00003130, 0.00004105, 0.00001817, 0.00003370, 0.00001407, -0.00002695, -0.00002283, -0.00002363, -0.00000939, -0.00008126, -0.00003948, 0.00015599, 0.00011761, 0.00011463, -0.00012136, -0.00001126, -0.00002195, -0.00001751, -0.00000665, -0.00000716, -0.00001581, 0.00000591, -0.00001271, 0.00001063, -0.00001784, -0.00000239, -0.00000916, 0.00000635, 0.00000644, -0.00001225, 0.00001467, -0.00000174, 0.00000234, -0.00000231, -0.00000171, -0.00000363, 0.00000322, 0.00000901, -0.00000519, 0.00000238, -0.00000029, -0.00000182, -0.00001066, 0.00000922, -0.00000369, 0.00000632, -0.00000676, 0.00000316, -0.00000591, 0.00000006, 0.00000282, -0.00000005, 0.00000074, 0.00000037, -0.00000259, -0.00000387, 0.00000608, -0.00000252, 0.00000070, 0.00000123, 0.00000065, -0.00000012, -0.00000209, 0.00000090, 0.00000010, 0.00000009, -0.00000148, -0.00000315, -0.00000175, 0.00000166, 0.00000264, 0.00000710, },
		{ -0.00000643, 0.00000059, -0.00000010, -0.00000072, -0.00000003, -0.00000118, 0.00000215, -0.00000389, 0.00000357, -0.00000133, -0.00000481, 0.00000071, 0.00000391, 0.00000021, 0.00000255, -0.00000230, -0.00000228, 0.00000015, 0.00000504, -0.00000185, -0.00000138, -0.00000220, -0.00000334, -0.00000184, 0.00000194, 0.00000230, 0.00000301, 0.00000208, -0.00000579, -0.00000548, 0.00000270, -0.00000562, 0.00001008, 0.00000251, 0.00000264, -0.00000897, 0.00000035, -0.00000243, 0.00000998, 0.00000167, -0.00000424, -0.00000328, 0.00000851, -0.00000174, -0.00000633, -0.00001197, 0.00000898, -0.00001202, -0.00000320, -0.00000078, 0.00001036, -0.00002944, 0.00000387, -0.00001958, -0.00002126, -0.00001126, 0.00003522, 0.00003499, 0.00002006, 0.00003019, 0.00001277, -0.00002840, -0.00001999, -0.00002625, -0.00000649, -0.00009154, -0.00002699, 0.00016139, 0.00013178, 0.00009293, -0.00012965, 0.00000091, -0.00002886, -0.00001241, -0.00000938, -0.00000538, -0.00001708, 0.00000752, -0.00001432, 0.00001265, -0.00002149, 0.00000221, -0.00001359, 0.00001113, 0.00000265, -0.00000983, 0.00001463, -0.00000299, 0.00000296, -0.00000286, -0.00000169, -0.00000353, 0.00000464, 0.00000731, -0.00000488, 0.00000224, 0.00000065, -0.00000335, -0.00000969, 0.00000930, -0.00000394, 0.00000657, -0.00000774, 0.00000446, -0.00000755, 0.00000215, 0.00000144, 0.00000081, 0.00000011, 0.00000092, -0.00000387, -0.00000244, 0.00000567, -0.00000284, 0.00000100, 0.00000115, 0.00000082, -0.00000059, -0.00000168, 0.00000073, 0.00000038, -0.00000027, -0.00000146, -0.00000352, -0.00000100, 0.00000152, 0.00000321, 0.00000648, },
		{ -0.00000706, 0.00000166, -0.00000065, -0.00000027, -0.00000045, -0.00000070, 0.00000173, -0.00000373, 0.00000408, -0.00000233, -0.00000468, 0.00000139, 0.00000364, 0.00000017, 0.00000287, -0.00000311, -0.00000174, 0.00000009, 0.00000565, -0.00000299, -0.00000078, -0.00000269, -0.00000318, -0.00000157, 0.00000216, 0.00000214, 0.00000351, 0.00000126, -0.00000639, -0.00000435, 0.00000221, -0.00000513, 0.00001066, 0.00000145, 0.00000277, -0.00000929, 0.00000052, -0.00000195, 0.00001065, 0.00000002, -0.00000404, -0.00000293, 0.00000993, -0.00000429, -0.00000603, -0.00001090, 0.00000939, -0.00001376, -0.00000176, -0.00000082, 0.00001064, -0.00003185, 0.00000793, -0.00002524, -0.00001570, -0.00001089, 0.00003897, 0.00002872, 0.00002255, 0.00002607, 0.00001204, -0.00002992, -0.00001725, -0.00002875, -0.00000397, -0.00010181, -0.00001362, 0.00016761, 0.00014563, 0.00006613, -0.00013171, 0.00000992, -0.00003423, -0.00000826, -0.00001141, -0.00000429, -0.00001760, 0.00000837, -0.00001511, 0.00001372, -0.00002416, 0.00000616, -0.00001787, 0.00001627, -0.00000184, -0.00000650, 0.00001367, -0.00000367, 0.00000333, -0.00000330, -0.00000176, -0.00000319, 0.00000580, 0.00000557, -0.00000440, 0.00000205, 0.00000170, -0.00000524, -0.00000818, 0.00000884, -0.00000367, 0.00000625, -0.00000814, 0.00000526, -0.00000879, 0.00000404, 0.00000006, 0.00000170, -0.00000051, 0.00000138, -0.00000513, -0.00000080, 0.00000491, -0.00000289, 0.00000113, 0.00000120, 0.00000082, -0.00000093, -0.00000132, 0.00000057, 0.00000065, -0.00000066, -0.00000140, -0.00000387, -0.00000014, 0.00000122, 0.00000396, 0.00000552, },
		{ -0.00000729, 0.00000258, -0.00000118, 0.00000015, -0.00000088, -0.00000018, 0.00000120, -0.00000337, 0.00000439, -0.00000332, -0.00000444, 0.00000211, 0.00000316, 0.00000040, 0.00000291, -0.00000373, -0.00000127, 0.00000016, 0.00000610, -0.00000410, -0.00000009, -0.00000328, -0.00000290, -0.00000131, 0.00000230, 0.00000207, 0.00000396, 0.00000029, -0.00000678, -0.00000321, 0.00000154, -0.00000435, 0.00001089, 0.00000059, 0.00000281, -0.00000950, 0.00000055, -0.00000114, 0.00001080, -0.00000126, -0.00000413, -0.00000197, 0.00001065, -0.00000658, -0.00000569, -0.00000955, 0.00000917, -0.00001489, -0.00000080, -0.00000026, 0.00001000, -0.00003320, 0.00001104, -0.00003037, -0.00000969, -0.00001110, 0.00004242, 0.00002247, 0.00002539, 0.00002156, 0.00001172, -0.00003137, -0.00001478, -0.00003088, -0.00000217, -0.00011166, 0.00000023, 0.00017540, 0.00015769, 0.00003544, -0.00012767, 0.00001527, -0.00003755, -0.00000548, -0.00001239, -0.00000417, -0.00001713, 0.00000822, -0.00001488, 0.00001358, -0.00002550, 0.00000899, -0.00002149, 0.00002129, -0.00000667, -0.00000248, 0.00001188, -0.00000379, 0.00000343, -0.00000361, -0.00000192, -0.00000259, 0.00000664, 0.00000389, -0.00000384, 0.00000190, 0.00000272, -0.00000737, -0.00000621, 0.00000786, -0.00000289, 0.00000537, -0.00000793, 0.00000548, -0.00000949, 0.00000559, -0.00000118, 0.00000251, -0.00000103, 0.00000166, -0.00000622, 0.00000092, 0.00000390, -0.00000269, 0.00000111, 0.00000138, 0.00000065, -0.00000111, -0.00000105, 0.00000048, 0.00000088, -0.00000107, -0.00000136, -0.00000415, 0.00000075, 0.00000079, 0.00000483, 0.00000422, },
		{ -0.00000712, 0.00000328, -0.00000163, 0.00000052, -0.00000126, 0.00000034, 0.00000061, -0.00000284, 0.00000450, -0.00000427, -0.00000408, 0.00000282, 0.00000252, 0.00000091, 0.00000263, -0.00000412, -0.00000094, 0.00000041, 0.00000631, -0.00000508, 0.00000060, -0.00000393, -0.00000254, -0.00000105, 0.00000236, 0.00000213, 0.00000429, -0.00000078, -0.00000695, -0.00000210, 0.00000075, -0.00000332, 0.00001076, -0.00000003, 0.00000275, -0.00000963, 0.00000049, -0.00000004, 0.00001044, -0.00000215, -0.00000452, -0.00000040, 0.00001056, -0.00000844, -0.00000540, -0.00000791, 0.00000828, -0.00001531, -0.00000043, 0.00000103, 0.00000830, -0.00003322, 0.00001279, -0.00003444, -0.00000378, -0.00001156, 0.00004535, 0.00001652, 0.00002830, 0.00001690, 0.00001162, -0.00003260, -0.00001278, -0.00003242, -0.00000140, -0.00012072, 0.00001420, 0.00018548, 0.00016638, 0.00000249, -0.00011813, 0.00001684, -0.00003853, -0.00000434, -0.00001213, -0.00000518, -0.00001552, 0.00000694, -0.00001348, 0.00001210, -0.00002527, 0.00001034, -0.00002398, 0.00002575, -0.00001145, 0.00000192, 0.00000940, -0.00000336, 0.00000325, -0.00000378, -0.00000218, -0.00000173, 0.00000709, 0.00000236, -0.00000326, 0.00000188, 0.00000358, -0.00000957, -0.00000391, 0.00000647, -0.00000168, 0.00000396, -0.00000710, 0.00000506, -0.00000959, 0.00000666, -0.00000215, 0.00000316, -0.00000138, 0.00000165, -0.00000702, 0.00000257, 0.00000273, -0.00000230, 0.00000095, 0.00000167, 0.00000033, -0.00000113, -0.00000090, 0.00000049, 0.00000101, -0.00000143, -0.00000135, -0.00000431, 0.00000161, 0.00000032, 0.00000577, 0.00000264, },
		{ -0.00000655, 0.00000373, -0.00000198, 0.00000080, -0.00000157, 0.00000082, -0.00000001, -0.00000218, 0.00000440, -0.00000517, -0.00000359, 0.00000346, 0.00000178, 0.00000164, 0.00000204, -0.00000424, -0.00000079, 0.00000088, 0.00000623, -0.00000587, 0.00000122, -0.00000458, -0.00000212, -0.00000078, 0.00000230, 0.00000235, 0.00000444, -0.00000188, -0.00000689, -0.00000109, -0.00000010, -0.00000209, 0.00001030, -0.00000041, 0.00000257, -0.00000969, 0.00000040, 0.00000124, 0.00000965, -0.00000270, -0.00000514, 0.00000173, 0.00000962, -0.00000976, -0.00000523, -0.00000598, 0.00000673, -0.00001501, -0.00000068, 0.00000307, 0.00000547, -0.00003176, 0.00001289, -0.00003695, 0.00000149, -0.00001192, 0.00004754, 0.00001117, 0.00003099, 0.00001233, 0.00001157, -0.00003346, -0.00001138, -0.00003319, -0.00000189, -0.00012867, 0.00002808, 0.00019842, 0.00017011, -0.00003083, -0.00010416, 0.00001491, -0.00003719, -0.00000487, -0.00001060, -0.00000732, -0.00001279, 0.00000454, -0.00001093, 0.00000926, -0.00002339, 0.00000995, -0.00002497, 0.00002919, -0.00001578, 0.00000639, 0.00000647, -0.00000249, 0.00000281, -0.00000383, -0.00000248, -0.00000064, 0.00000714, 0.00000108, -0.00000276, 0.00000208, 0.00000412, -0.00001166, -0.00000146, 0.00000480, -0.00000015, 0.00000213, -0.00000573, 0.00000405, -0.00000905, 0.00000720, -0.00000279, 0.00000358, -0.00000151, 0.00000131, -0.00000742, 0.00000401, 0.00000152, -0.00000179, 0.00000071, 0.00000204, -0.00000012, -0.00000099, -0.00000088, 0.00000062, 0.00000101, -0.00000173, -0.00000142, -0.00000431, 0.00000235, -0.00000011, 0.00000669, 0.00000084, },
		{ -0.00000563, 0.00000392, -0.00000219, 0.00000098, -0.00000178, 0.00000121, -0.00000059, -0.00000142, 0.00000411, -0.00000600, -0.00000295, 0.00000396, 0.00000102, 0.00000253, 0.00000116, -0.00000408, -0.00000083, 0.00000158, 0.00000583, -0.00000640, 0.00000170, -0.00000516, -0.00000168, -0.00000050, 0.00000215, 0.00000276, 0.00000436, -0.00000296, -0.00000663, -0.00000020, -0.00000094, -0.00000073, 0.00000954, -0.00000055, 0.00000230, -0.00000972, 0.00000036, 0.00000258, 0.00000855, -0.00000301, -0.00000585, 0.00000429, 0.00000787, -0.00001050, -0.00000519, -0.00000385, 0.00000462, -0.00001406, -0.00000150, 0.00000576, 0.00000158, -0.00002882, 0.00001118, -0.00003755, 0.00000565, -0.00001182, 0.00004876, 0.00000668, 0.00003318, 0.00000808, 0.00001138, -0.00003386, -0.00001068, -0.00003310, -0.00000379, -0.00013534, 0.00004182, 0.00021452, 0.00016744, -0.00006244, -0.00008716, 0.00001010, -0.00003385, -0.00000687, -0.00000795, -0.00001042, -0.00000908, 0.00000115, -0.00000738, 0.00000521, -0.00001992, 0.00000776, -0.00002421, 0.00003127, -0.00001926, 0.00001056, 0.00000332, -0.00000130, 0.00000217, -0.00000381, -0.00000279, 0.00000063, 0.00000678, 0.00000008, -0.00000237, 0.00000256, 0.00000421, -0.00001344, 0.00000097, 0.00000302, 0.00000154, 0.00000003, -0.00000393, 0.00000253, -0.00000791, 0.00000719, -0.00000305, 0.00000374, -0.00000140, 0.00000061, -0.00000734, 0.00000513, 0.00000039, -0.00000126, 0.00000044, 0.00000243, -0.00000066, -0.00000071, -0.00000098, 0.00000087, 0.00000088, -0.00000194, -0.00000157, -0.00000412, 0.00000292, -0.00000043, 0.00000748, -0.00000108, },
		{ -0.00000442, 0.00000385, -0.00000227, 0.00000104, -0.00000188, 0.00000150, -0.00000110, -0.00000062, 0.00000363, -0.00000675, -0.00000214, 0.00000427, 0.00000034, 0.00000349, 0.00000006, -0.00000368, -0.00000104, 0.00000248, 0.00000510, -0.00000664, 0.00000200, -0.00000565, -0.00000124, -0.00000021, 0.00000190, 0.00000335, 0.00000399, -0.00000393, -0.00000620, 0.00000052, -0.00000173, 0.00000069, 0.00000855, -0.00000048, 0.00000192, -0.00000975, 0.00000045, 0.00000386, 0.00000727, -0.00000318, -0.00000650, 0.00000705, 0.00000545, -0.00001069, -0.00000523, -0.00000166, 0.00000212, -0.00001259, -0.00000272, 0.00000894, -0.00000315, -0.00002456, 0.00000771, -0.00003607, 0.00000832, -0.00001097, 0.00004884, 0.00000325, 0.00003465, 0.00000432, 0.00001094, -0.00003374, -0.00001070, -0.00003214, -0.00000712, -0.00014071, 0.00005566, 0.00023369, 0.00015724, -0.00009033, -0.00006877, 0.00000334, -0.00002906, -0.00000996, -0.00000453, -0.00001418, -0.00000470, -0.00000294, -0.00000312, 0.00000024, -0.00001511, 0.00000385, -0.00002161, 0.00003176, -0.00002160, 0.00001412, 0.00000020, 0.00000007, 0.00000139, -0.00000374, -0.00000304, 0.00000199, 0.00000604, -0.00000061, -0.00000215, 0.00000336, 0.00000375, -0.00001475, 0.00000317, 0.00000132, 0.00000320, -0.00000217, -0.00000189, 0.00000063, -0.00000627, 0.00000666, -0.00000294, 0.00000365, -0.00000109, -0.00000043, -0.00000676, 0.00000583, -0.00000055, -0.00000077, 0.00000020, 0.00000280, -0.00000124, -0.00000033, -0.00000119, 0.00000124, 0.00000059, -0.00000203, -0.00000182, -0.00000371, 0.00000326, -0.00000058, 0.00000806, -0.00000299, },
		{ -0.00000303, 0.00000357, -0.00000223, 0.00000100, -0.00000187, 0.00000166, -0.00000150, 0.00000017, 0.00000300, -0.00000741, -0.00000117, 0.00000433, -0.00000016, 0.00000440, -0.00000118, -0.00000307, -0.00000137, 0.00000353, 0.00000408, -0.00000658, 0.00000210, -0.00000602, -0.00000080, 0.00000006, 0.00000159, 0.00000410, 0.00000333, -0.00000474, -0.00000564, 0.00000107, -0.00000240, 0.00000210, 0.00000738, -0.00000022, 0.00000147, -0.00000981, 0.00000072, 0.00000495, 0.00000595, -0.00000336, -0.00000690, 0.00000979, 0.00000256, -0.00001041, -0.00000526, 0.00000042, -0.00000053, -0.00001083, -0.00000412, 0.00001229, -0.00000839, -0.00001931, 0.00000272, -0.00003255, 0.00000933, -0.00000918, 0.00004766, 0.00000102, 0.00003526, 0.00000119, 0.00001021, -0.00003314, -0.00001135, -0.00003046, -0.00001174, -0.00014494, 0.00007009, 0.00025543, 0.00013884, -0.00011264, -0.00005071, -0.00000426, -0.00002357, -0.00001356, -0.00000079, -0.00001816, -0.00000004, -0.00000734, 0.00000147, -0.00000521, -0.00000935, -0.00000149, -0.00001726, 0.00003056, -0.00002261, 0.00001683, -0.00000265, 0.00000147, 0.00000057, -0.00000370, -0.00000317, 0.00000338, 0.00000498, -0.00000098, -0.00000208, 0.00000449, 0.00000268, -0.00001544, 0.00000496, -0.00000012, 0.00000465, -0.00000425, 0.00000019, -0.00000145, -0.00000426, 0.00000571, -0.00000250, 0.00000334, -0.00000064, -0.00000174, -0.00000569, 0.00000607, -0.00000123, -0.00000041, 0.00000005, 0.00000308, -0.00000180, 0.00000010, -0.00000146, 0.00000170, 0.00000016, -0.00000202, -0.00000214, -0.00000309, 0.00000335, -0.00000048, 0.00000834, -0.00000478, },
		{ -0.00000153, 0.00000312, -0.00000208, 0.00000088, -0.00000175, 0.00000170, -0.00000177, 0.00000092, 0.00000223, -0.00000795, -0.00000007, 0.00000412, -0.00000042, 0.00000517, -0.00000245, -0.00000235, -0.00000176, 0.00000468, 0.00000282, -0.00000624, 0.00000200, -0.00000626, -0.00000035, 0.00000029, 0.00000129, 0.00000495, 0.00000237, -0.00000533, -0.00000498, 0.00000144, -0.00000291, 0.00000340, 0.00000611, 0.00000019, 0.00000095, -0.00000990, 0.00000122, 0.00000574, 0.00000473, -0.00000367, -0.00000688, 0.00001224, -0.00000057, -0.00000982, -0.00000515, 0.00000218, -0.00000309, -0.00000906, -0.00000540, 0.00001548, -0.00001371, -0.00001352, -0.00000338, -0.00002720, 0.00000866, -0.00000639, 0.00004525, 0.00000001, 0.00003495, -0.00000130, 0.00000920, -0.00003217, -0.00001246, -0.00002828, -0.00001739, -0.00014837, 0.00008590, 0.00027871, 0.00011215, -0.00012791, -0.00003461, -0.00001152, -0.00001822, -0.00001704, 0.00000270, -0.00002186, 0.00000439, -0.00001159, 0.00000591, -0.00001065, -0.00000318, -0.00000782, -0.00001143, 0.00002777, -0.00002220, 0.00001850, -0.00000505, 0.00000277, -0.00000024, -0.00000370, -0.00000313, 0.00000469, 0.00000368, -0.00000106, -0.00000214, 0.00000588, 0.00000099, -0.00001542, 0.00000621, -0.00000114, 0.00000570, -0.00000602, 0.00000212, -0.00000353, -0.00000207, 0.00000446, -0.00000182, 0.00000291, -0.00000014, -0.00000321, -0.00000420, 0.00000583, -0.00000159, -0.00000023, 0.00000004, 0.00000323, -0.00000228, 0.00000054, -0.00000177, 0.00000223, -0.00000038, -0.00000192, -0.00000252, -0.00000230, 0.00000318, -0.00000013, 0.00000826, -0.00000635, },
		{ -0.00000003, 0.00000254, -0.00000186, 0.00000069, -0.00000155, 0.00000162, -0.00000188, 0.00000159, 0.00000135, -0.00000835, 0.00000114, 0.00000362, -0.00000037, 0.00000568, -0.00000364, -0.00000159, -0.00000214, 0.00000582, 0.00000140, -0.00000569, 0.00000173, -0.00000639, 0.00000012, 0.00000042, 0.00000105, 0.00000583, 0.00000116, -0.00000568, -0.00000428, 0.00000162, -0.00000321, 0.00000455, 0.00000482, 0.00000070, 0.00000038, -0.00001004, 0.00000199, 0.00000617, 0.00000370, -0.00000421, -0.00000627, 0.00001414, -0.00000367, -0.00000908, -0.00000479, 0.00000343, -0.00000528, -0.00000755, -0.00000626, 0.00001812, -0.00001865, -0.00000774, -0.00001002, -0.00002048, 0.00000650, -0.00000269, 0.00004172, 0.00000013, 0.00003378, -0.00000319, 0.00000802, -0.00003100, -0.00001378, -0.00002592, -0.00002368, -0.00015145, 0.00010413, 0.00030200, 0.00007775, -0.00013520, -0.00002185, -0.00001735, -0.00001383, -0.00001972, 0.00000540, -0.00002475, 0.00000811, -0.00001521, 0.00000970, -0.00001551, 0.00000280, -0.00001458, -0.00000453, 0.00002359, -0.00002044, 0.00001907, -0.00000686, 0.00000387, -0.00000098, -0.00000380, -0.00000285, 0.00000582, 0.00000223, -0.00000090, -0.00000227, 0.00000747, -0.00000127, -0.00001466, 0.00000681, -0.00000161, 0.00000623, -0.00000730, 0.00000370, -0.00000538, 0.00000011, 0.00000306, -0.00000100, 0.00000243, 0.00000031, -0.00000471, -0.00000240, 0.00000516, -0.00000162, -0.00000026, 0.00000020, 0.00000320, -0.00000264, 0.00000094, -0.00000206, 0.00000277, -0.00000101, -0.00000175, -0.00000291, -0.00000137, 0.00000279, 0.00000047, 0.00000776, -0.00000758, },
		{ 0.00000137, 0.00000191, -0.00000161, 0.00000049, -0.00000130, 0.00000145, -0.00000185, 0.00000214, 0.00000037, -0.00000857, 0.00000240, 0.00000285, 0.00000003, 0.00000585, -0.00000465, -0.00000089, -0.00000242, 0.00000687, -0.00000008, -0.00000498, 0.00000131, -0.00000642, 0.00000063, 0.00000042, 0.00000095, 0.00000666, -0.00000025, -0.00000578, -0.00000356, 0.00000163, -0.00000329, 0.00000547, 0.00000358, 0.00000126, -0.00000024, -0.00001020, 0.00000300, 0.00000620, 0.00000293, -0.00000502, -0.00000499, 0.00001529, -0.00000651, -0.00000833, -0.00000405, 0.00000400, -0.00000687, -0.00000655, -0.00000642, 0.00001985, -0.00002272, -0.00000256, -0.00001656, -0.00001294, 0.00000318, 0.00000170, 0.00003730, 0.00000122, 0.00003187, -0.00000460, 0.00000681, -0.00002987, -0.00001502, -0.00002378, -0.00003019, -0.00015473, 0.00012597, 0.00032334, 0.00003691, -0.00013422, -0.00001345, -0.00002086, -0.00001113, -0.00002104, 0.00000681, -0.00002637, 0.00001067, -0.00001776, 0.00001239, -0.00001927, 0.00000798, -0.00002114, 0.00000290, 0.00001839, -0.00001750, 0.00001855, -0.00000801, 0.00000469, -0.00000160, -0.00000399, -0.00000232, 0.00000670, 0.00000073, -0.00000058, -0.00000240, 0.00000912, -0.00000399, -0.00001318, 0.00000675, -0.00000149, 0.00000612, -0.00000798, 0.00000475, -0.00000682, 0.00000209, 0.00000167, -0.00000017, 0.00000200, 0.00000057, -0.00000608, -0.00000043, 0.00000414, -0.00000134, -0.00000050, 0.00000053, 0.00000299, -0.00000284, 0.00000124, -0.00000229, 0.00000327, -0.00000169, -0.00000153, -0.00000327, -0.00000037, 0.00000222, 0.00000130, 0.00000686, -0.00000842, },
		{ 0.00000261, 0.00000127, -0.00000135, 0.00000028, -0.00000104, 0.00000121, -0.00000168, 0.00000256, -0.00000068, -0.00000860, 0.00000366, 0.00000186, 0.00000077, 0.00000563, -0.00000540, -0.00000031, -0.00000254, 0.00000774, -0.00000153, -0.00000419, 0.00000081, -0.00000638, 0.00000120, 0.00000024, 0.00000106, 0.00000732, -0.00000176, -0.00000564, -0.00000286, 0.00000148, -0.00000313, 0.00000612, 0.00000245, 0.00000185, -0.00000089, -0.00001036, 0.00000422, 0.00000582, 0.00000243, -0.00000611, -0.00000301, 0.00001553, -0.00000889, -0.00000772, -0.00000287, 0.00000377, -0.00000768, -0.00000627, -0.00000565, 0.00002038, -0.00002552, 0.00000147, -0.00002232, -0.00000526, -0.00000082, 0.00000647, 0.00003230, 0.00000300, 0.00002941, -0.00000569, 0.00000575, -0.00002901, -0.00001588, -0.00002222, -0.00003650, -0.00015870, 0.00015268, 0.00034042, -0.00000844, -0.00012532, -0.00000988, -0.00002145, -0.00001062, -0.00002056, 0.00000655, -0.00002638, 0.00001173, -0.00001891, 0.00001361, -0.00002147, 0.00001179, -0.00002686, 0.00001029, 0.00001261, -0.00001368, 0.00001707, -0.00000851, 0.00000521, -0.00000210, -0.00000429, -0.00000152, 0.00000725, -0.00000072, -0.00000018, -0.00000242, 0.00001069, -0.00000701, -0.00001109, 0.00000606, -0.00000077, 0.00000537, -0.00000798, 0.00000519, -0.00000772, 0.00000371, 0.00000043, 0.00000057, 0.00000172, 0.00000055, -0.00000716, 0.00000154, 0.00000288, -0.00000082, -0.00000095, 0.00000104, 0.00000259, -0.00000285, 0.00000141, -0.00000242, 0.00000371, -0.00000237, -0.00000131, -0.00000355, 0.00000062, 0.00000155, 0.00000229, 0.00000557, -0.00000882, },
		{ 0.00000361, 0.00000067, -0.00000113, 0.00000012, -0.00000079, 0.00000094, -0.00000140, 0.00000284, -0.00000178, -0.00000839, 0.00000482, 0.00000073, 0.00000182, 0.00000501, -0.00000582, 0.00000007, -0.00000244, 0.00000835, -0.00000286, -0.00000340, 0.00000027, -0.00000630, 0.00000185, -0.00000014, 0.00000145, 0.00000772, -0.00000330, -0.00000528, -0.00000220, 0.00000121, -0.00000275, 0.00000648, 0.00000148, 0.00000242, -0.00000159, -0.00001047, 0.00000558, 0.00000510, 0.00000218, -0.00000740, -0.00000039, 0.00001482, -0.00001067, -0.00000731, -0.00000125, 0.00000273, -0.00000764, -0.00000682, -0.00000384, 0.00001952, -0.00002674, 0.00000391, -0.00002668, 0.00000189, -0.00000499, 0.00001121, 0.00002708, 0.00000519, 0.00002662, -0.00000663, 0.00000500, -0.00002862, -0.00001611, -0.00002156, -0.00004225, -0.00016374, 0.00018541, 0.00035075, -0.00005583, -0.00010959, -0.00001113, -0.00001892, -0.00001250, -0.00001809, 0.00000447, -0.00002461, 0.00001110, -0.00001845, 0.00001316, -0.00002182, 0.00001378, -0.00003117, 0.00001704, 0.00000675, -0.00000932, 0.00001481, -0.00000842, 0.00000543, -0.00000248, -0.00000466, -0.00000045, 0.00000741, -0.00000201, 0.00000023, -0.00000226, 0.00001201, -0.00001014, -0.00000854, 0.00000484, 0.00000047, 0.00000400, -0.00000730, 0.00000495, -0.00000798, 0.00000485, -0.00000053, 0.00000113, 0.00000165, 0.00000015, -0.00000781, 0.00000334, 0.00000152, -0.00000012, -0.00000154, 0.00000169, 0.00000200, -0.00000267, 0.00000143, -0.00000241, 0.00000402, -0.00000301, -0.00000112, -0.00000371, 0.00000152, 0.00000085, 0.00000336, 0.00000396, -0.00000875, },
		{ 0.00000436, 0.00000015, -0.00000095, 0.00000002, -0.00000059, 0.00000069, -0.00000102, 0.00000297, -0.00000292, -0.00000793, 0.00000583, -0.00000045, 0.00000311, 0.00000400, -0.00000589, 0.00000024, -0.00000209, 0.00000866, -0.00000400, -0.00000266, -0.00000027, -0.00000618, 0.00000256, -0.00000074, 0.00000216, 0.00000777, -0.00000474, -0.00000476, -0.00000161, 0.00000085, -0.00000217, 0.00000655, 0.00000071, 0.00000295, -0.00000235, -0.00001046, 0.00000697, 0.00000413, 0.00000209, -0.00000876, 0.00000275, 0.00001320, -0.00001180, -0.00000712, 0.00000075, 0.00000094, -0.00000676, -0.00000817, -0.00000099, 0.00001724, -0.00002625, 0.00000446, -0.00002917, 0.00000791, -0.00000879, 0.00001555, 0.00002200, 0.00000748, 0.00002374, -0.00000757, 0.00000468, -0.00002884, -0.00001557, -0.00002198, -0.00004730, -0.00016997, 0.00022503, 0.00035186, -0.00010236, -0.00008870, -0.00001658, -0.00001351, -0.00001669, -0.00001371, 0.00000063, -0.00002110, 0.00000879, -0.00001640, 0.00001102, -0.00002020, 0.00001368, -0.00003364, 0.00002263, 0.00000128, -0.00000480, 0.00001203, -0.00000783, 0.00000538, -0.00000275, -0.00000508, 0.00000087, 0.00000715, -0.00000306, 0.00000056, -0.00000181, 0.00001290, -0.00001313, -0.00000575, 0.00000326, 0.00000210, 0.00000214, -0.00000603, 0.00000408, -0.00000757, 0.00000544, -0.00000115, 0.00000145, 0.00000182, -0.00000067, -0.00000792, 0.00000481, 0.00000020, 0.00000065, -0.00000223, 0.00000244, 0.00000127, -0.00000231, 0.00000130, -0.00000225, 0.00000417, -0.00000359, -0.00000097, -0.00000370, 0.00000227, 0.00000023, 0.00000439, 0.00000212, -0.00000826, },
		{ 0.00000483, -0.00000028, -0.00000083, -0.00000002, -0.00000047, 0.00000048, -0.00000059, 0.00000295, -0.00000407, -0.00000720, 0.00000658, -0.00000158, 0.00000453, 0.00000267, -0.00000564, 0.00000018, -0.00000149, 0.00000865, -0.00000491, -0.00000201, -0.00000078, -0.00000604, 0.00000331, -0.00000152, 0.00000319, 0.00000740, -0.00000600, -0.00000411, -0.00000111, 0.00000045, -0.00000144, 0.00000634, 0.00000013, 0.00000344, -0.00000319, -0.00001027, 0.00000828, 0.00000302, 0.00000204, -0.00001002, 0.00000616, 0.00001080, -0.00001233, -0.00000708, 0.00000296, -0.00000144, -0.00000519, -0.00001020, 0.00000275, 0.00001368, -0.00002411, 0.00000305, -0.00002949, 0.00001231, -0.00001179, 0.00001913, 0.00001737, 0.00000963, 0.00002090, -0.00000860, 0.00000483, -0.00002968, -0.00001426, -0.00002348, -0.00005172, -0.00017713, 0.00027199, 0.00034153, -0.00014496, -0.00006480, -0.00002516, -0.00000583, -0.00002275, -0.00000775, -0.00000468, -0.00001610, 0.00000504, -0.00001296, 0.00000736, -0.00001672, 0.00001143, -0.00003401, 0.00002667, -0.00000335, -0.00000052, 0.00000897, -0.00000688, 0.00000511, -0.00000296, -0.00000549, 0.00000237, 0.00000645, -0.00000379, 0.00000074, -0.00000100, 0.00001321, -0.00001576, -0.00000295, 0.00000152, 0.00000394, -0.00000005, -0.00000429, 0.00000267, -0.00000656, 0.00000548, -0.00000140, 0.00000153, 0.00000221, -0.00000191, -0.00000744, 0.00000583, -0.00000094, 0.00000140, -0.00000293, 0.00000324, 0.00000042, -0.00000179, 0.00000101, -0.00000193, 0.00000415, -0.00000405, -0.00000088, -0.00000351, 0.00000281, -0.00000024, 0.00000529, 0.00000016, -0.00000739, },
		{ 0.00000504, -0.00000061, -0.00000077, 0.00000003, -0.00000043, 0.00000035, -0.00000014, 0.00000279, -0.00000519, -0.00000621, 0.00000703, -0.00000253, 0.00000596, 0.00000109, -0.00000509, -0.00000006, -0.00000068, 0.00000833, -0.00000558, -0.00000148, -0.00000125, -0.00000584, 0.00000407, -0.00000244, 0.00000452, 0.00000658, -0.00000701, -0.00000337, -0.00000071, 0.00000005, -0.00000061, 0.00000589, -0.00000024, 0.00000388, -0.00000414, -0.00000982, 0.00000939, 0.00000192, 0.00000189, -0.00001097, 0.00000959, 0.00000784, -0.00001238, -0.00000705, 0.00000517, -0.00000413, -0.00000316, -0.00001266, 0.00000709, 0.00000913, -0.00002058, -0.00000019, -0.00002757, 0.00001483, -0.00001366, 0.00002171, 0.00001342, 0.00001146, 0.00001822, -0.00000971, 0.00000534, -0.00003100, -0.00001236, -0.00002584, -0.00005588, -0.00018455, 0.00032611, 0.00031801, -0.00018061, -0.00004035, -0.00003539, 0.00000315, -0.00002995, -0.00000080, -0.00001093, -0.00001010, 0.00000024, -0.00000852, 0.00000257, -0.00001171, 0.00000720, -0.00003221, 0.00002892, -0.00000683, 0.00000322, 0.00000590, -0.00000569, 0.00000469, -0.00000313, -0.00000587, 0.00000402, 0.00000531, -0.00000416, 0.00000073, 0.00000022, 0.00001282, -0.00001781, -0.00000038, -0.00000016, 0.00000576, -0.00000234, -0.00000229, 0.00000088, -0.00000506, 0.00000503, -0.00000130, 0.00000139, 0.00000278, -0.00000350, -0.00000635, 0.00000632, -0.00000180, 0.00000202, -0.00000358, 0.00000402, -0.00000048, -0.00000114, 0.00000059, -0.00000145, 0.00000395, -0.00000441, -0.00000086, -0.00000312, 0.00000309, -0.00000048, 0.00000594, -0.00000178, -0.00000622, },
		{ 0.00000503, -0.00000085, -0.00000076, 0.00000014, -0.00000049, 0.00000031, 0.00000029, 0.00000248, -0.00000627, -0.00000495, 0.00000711, -0.00000321, 0.00000728, -0.00000060, -0.00000433, -0.00000043, 0.00000028, 0.00000774, -0.00000603, -0.00000106, -0.00000170, -0.00000556, 0.00000475, -0.00000341, 0.00000607, 0.00000533, -0.00000771, -0.00000260, -0.00000044, -0.00000028, 0.00000024, 0.00000527, -0.00000044, 0.00000427, -0.00000520, -0.00000906, 0.00001018, 0.00000096, 0.00000150, -0.00001138, 0.00001272, 0.00000461, -0.00001212, -0.00000685, 0.00000710, -0.00000682, -0.00000101, -0.00001519, 0.00001162, 0.00000403, -0.00001609, -0.00000486, -0.00002361, 0.00001539, -0.00001427, 0.00002319, 0.00001024, 0.00001291, 0.00001566, -0.00001077, 0.00000601, -0.00003252, -0.00001024, -0.00002858, -0.00006048, -0.00019102, 0.00038649, 0.00028024, -0.00020660, -0.00001789, -0.00004556, 0.00001223, -0.00003735, 0.00000635, -0.00001743, -0.00000372, -0.00000500, -0.00000362, -0.00000279, -0.00000570, 0.00000138, -0.00002843, 0.00002933, -0.00000898, 0.00000617, 0.00000301, -0.00000440, 0.00000415, -0.00000328, -0.00000615, 0.00000575, 0.00000377, -0.00000413, 0.00000050, 0.00000186, 0.00001163, -0.00001910, 0.00000173, -0.00000153, 0.00000732, -0.00000451, -0.00000026, -0.00000106, -0.00000323, 0.00000420, -0.00000094, 0.00000111, 0.00000343, -0.00000534, -0.00000472, 0.00000624, -0.00000230, 0.00000244, -0.00000412, 0.00000475, -0.00000138, -0.00000042, 0.00000006, -0.00000084, 0.00000356, -0.00000463, -0.00000090, -0.00000256, 0.00000309, -0.00000044, 0.00000626, -0.00000358, -0.00000486, },
		{ 0.00000484, -0.00000103, -0.00000078, 0.00000029, -0.00000062, 0.00000039, 0.00000068, 0.00000204, -0.00000726, -0.00000348, 0.00000680, -0.00000353, 0.00000835, -0.00000228, -0.00000345, -0.00000086, 0.00000131, 0.00000697, -0.00000631, -0.00000071, -0.00000216, -0.00000514, 0.00000527, -0.00000432, 0.00000774, 0.00000370, -0.00000807, -0.00000182, -0.00000032, -0.00000048, 0.00000102, 0.00000454, -0.00000051, 0.00000464, -0.00000640, -0.00000794, 0.00001054, 0.00000027, 0.00000073, -0.00001106, 0.00001523, 0.00000142, -0.00001178, -0.00000628, 0.00000846, -0.00000914, 0.00000089, -0.00001739, 0.00001587, -0.00000105, -0.00001126, -0.00001040, -0.00001805, 0.00001419, -0.00001368, 0.00002361, 0.00000781, 0.00001404, 0.00001309, -0.00001156, 0.00000650, -0.00003382, -0.00000845, -0.00003102, -0.00006653, -0.00019484, 0.00045143, 0.00022807, -0.00022079, 0.00000022, -0.00005391, 0.00002007, -0.00004392, 0.00001282, -0.00002338, 0.00000231, -0.00001003, 0.00000106, -0.00000803, 0.00000063, -0.00000543, -0.00002306, 0.00002804, -0.00000976, 0.00000820, 0.00000048, -0.00000310, 0.00000354, -0.00000342, -0.00000632, 0.00000750, 0.00000188, -0.00000369, 0.00000006, 0.00000389, 0.00000963, -0.00001948, 0.00000318, -0.00000239, 0.00000841, -0.00000631, 0.00000155, -0.00000293, -0.00000129, 0.00000316, -0.00000043, 0.00000080, 0.00000401, -0.00000726, -0.00000265, 0.00000562, -0.00000241, 0.00000260, -0.00000448, 0.00000535, -0.00000225, 0.00000035, -0.00000054, -0.00000012, 0.00000301, -0.00000474, -0.00000097, -0.00000185, 0.00000284, -0.00000011, 0.00000617, -0.00000513, -0.00000339, },
		{ 0.00000453, -0.00000116, -0.00000079, 0.00000047, -0.00000083, 0.00000058, 0.00000100, 0.00000149, -0.00000811, -0.00000183, 0.00000610, -0.00000342, 0.00000905, -0.00000383, -0.00000255, -0.00000126, 0.00000232, 0.00000610, -0.00000646, -0.00000040, -0.00000267, -0.00000453, 0.00000554, -0.00000505, 0.00000940, 0.00000177, -0.00000810, -0.00000104, -0.00000036, -0.00000049, 0.00000164, 0.00000381, -0.00000050, 0.00000500, -0.00000774, -0.00000643, 0.00001041, -0.00000006, -0.00000050, -0.00000985, 0.00001684, -0.00000142, -0.00001159, -0.00000512, 0.00000896, -0.00001072, 0.00000214, -0.00001883, 0.00001935, -0.00000552, -0.00000675, -0.00001608, -0.00001150, 0.00001163, -0.00001211, 0.00002318, 0.00000594, 0.00001503, 0.00001029, -0.00001176, 0.00000638, -0.00003435, -0.00000766, -0.00003231, -0.00007524, -0.00019385, 0.00051839, 0.00016235, -0.00022181, 0.00001197, -0.00005882, 0.00002543, -0.00004859, 0.00001772, -0.00002797, 0.00000724, -0.00001414, 0.00000485, -0.00001245, 0.00000652, -0.00001252, -0.00001664, 0.00002540, -0.00000930, 0.00000928, -0.00000164, -0.00000184, 0.00000287, -0.00000355, -0.00000636, 0.00000919, -0.00000032, -0.00000287, -0.00000058, 0.00000625, 0.00000684, -0.00001890, 0.00000385, -0.00000258, 0.00000885, -0.00000753, 0.00000292, -0.00000449, 0.00000055, 0.00000207, 0.00000010, 0.00000057, 0.00000439, -0.00000909, -0.00000031, 0.00000454, -0.00000214, 0.00000249, -0.00000466, 0.00000581, -0.00000302, 0.00000111, -0.00000117, 0.00000067, 0.00000231, -0.00000473, -0.00000105, -0.00000104, 0.00000237, 0.00000050, 0.00000564, -0.00000631, -0.00000193, },
		{ 0.00000415, -0.00000128, -0.00000078, 0.00000065, -0.00000107, 0.00000089, 0.00000121, 0.00000082, -0.00000880, -0.00000008, 0.00000506, -0.00000287, 0.00000931, -0.00000512, -0.00000174, -0.00000154, 0.00000320, 0.00000523, -0.00000658, -0.00000006, -0.00000327, -0.00000367, 0.00000547, -0.00000545, 0.00001087, -0.00000032, -0.00000786, -0.00000027, -0.00000058, -0.00000026, 0.00000203, 0.00000316, -0.00000047, 0.00000537, -0.00000921, -0.00000454, 0.00000974, 0.00000003, -0.00000222, -0.00000766, 0.00001733, -0.00000364, -0.00001176, -0.00000321, 0.00000836, -0.00001126, 0.00000239, -0.00001912, 0.00002157, -0.00000882, -0.00000327, -0.00002112, -0.00000474, 0.00000828, -0.00000997, 0.00002224, 0.00000434, 0.00001615, 0.00000695, -0.00001098, 0.00000515, -0.00003351, -0.00000863, -0.00003153, -0.00008800, -0.00018556, 0.00058408, 0.00008497, -0.00020922, 0.00001595, -0.00005900, 0.00002720, -0.00005046, 0.00002024, -0.00003047, 0.00001038, -0.00001669, 0.00000711, -0.00001533, 0.00001119, -0.00001908, -0.00000987, 0.00002186, -0.00000788, 0.00000951, -0.00000331, -0.00000065, 0.00000212, -0.00000364, -0.00000627, 0.00001077, -0.00000273, -0.00000169, -0.00000137, 0.00000882, 0.00000336, -0.00001737, 0.00000366, -0.00000201, 0.00000850, -0.00000801, 0.00000366, -0.00000552, 0.00000206, 0.00000113, 0.00000050, 0.00000056, 0.00000440, -0.00001063, 0.00000211, 0.00000313, -0.00000153, 0.00000211, -0.00000463, 0.00000611, -0.00000369, 0.00000183, -0.00000179, 0.00000150, 0.00000149, -0.00000463, -0.00000110, -0.00000021, 0.00000174, 0.00000135, 0.00000468, -0.00000706, -0.00000058, },
		{ 0.00000376, -0.00000141, -0.00000073, 0.00000079, -0.00000133, 0.00000130, 0.00000132, 0.00000007, -0.00000927, 0.00000168, 0.00000372, -0.00000187, 0.00000907, -0.00000606, -0.00000110, -0.00000162, 0.00000387, 0.00000444, -0.00000674, 0.00000036, -0.00000401, -0.00000252, 0.00000497, -0.00000541, 0.00001200, -0.00000243, -0.00000738, 0.00000049, -0.00000100, 0.00000024, 0.00000211, 0.00000267, -0.00000047, 0.00000576, -0.00001080, -0.00000230, 0.00000856, 0.00000054, -0.00000441, -0.00000450, 0.00001658, -0.00000502, -0.00001244, -0.00000045, 0.00000653, -0.00001054, 0.00000140, -0.00001796, 0.00002216, -0.00001043, -0.00000145, -0.00002473, 0.00000142, 0.00000481, -0.00000777, 0.00002122, 0.00000268, 0.00001771, 0.00000275, -0.00000884, 0.00000235, -0.00003075, -0.00001206, -0.00002777, -0.00010615, -0.00016730, 0.00064462, -0.00000111, -0.00018364, 0.00001149, -0.00005361, 0.00002465, -0.00004886, 0.00001982, -0.00003035, 0.00001121, -0.00001717, 0.00000732, -0.00001610, 0.00001394, -0.00002434, -0.00000346, 0.00001799, -0.00000584, 0.00000908, -0.00000461, 0.00000050, 0.00000127, -0.00000365, -0.00000606, 0.00001219, -0.00000529, -0.00000022, -0.00000224, 0.00001147, -0.00000064, -0.00001497, 0.00000263, -0.00000066, 0.00000732, -0.00000766, 0.00000363, -0.00000586, 0.00000306, 0.00000049, 0.00000065, 0.00000086, 0.00000393, -0.00001168, 0.00000440, 0.00000153, -0.00000068, 0.00000151, -0.00000441, 0.00000623, -0.00000422, 0.00000248, -0.00000238, 0.00000231, 0.00000059, -0.00000445, -0.00000109, 0.00000057, 0.00000103, 0.00000236, 0.00000333, -0.00000734, 0.00000059, },
		{ 0.00000339, -0.00000158, -0.00000060, 0.00000087, -0.00000157, 0.00000179, 0.00000130, -0.00000075, -0.00000949, 0.00000338, 0.00000219, -0.00000049, 0.00000832, -0.00000660, -0.00000069, -0.00000147, 0.00000426, 0.00000381, -0.00000698, 0.00000090, -0.00000491, -0.00000105, 0.00000399, -0.00000482, 0.00001264, -0.00000440, -0.00000676, 0.00000125, -0.00000162, 0.00000105, 0.00000183, 0.00000241, -0.00000055, 0.00000617, -0.00001245, 0.00000022, 0.00000691, 0.00000141, -0.00000696, -0.00000047, 0.00001457, -0.00000547, -0.00001369, 0.00000315, 0.00000341, -0.00000844, -0.00000099, -0.00001520, 0.00002091, -0.00001004, -0.00000177, -0.00002625, 0.00000619, 0.00000191, -0.00000606, 0.00002057, 0.00000056, 0.00002001, -0.00000261, -0.00000501, -0.00000241, -0.00002559, -0.00001853, -0.00002030, -0.00013084, -0.00013647, 0.00069568, -0.00009206, -0.00014668, -0.00000127, -0.00004242, 0.00001746, -0.00004348, 0.00001615, -0.00002731, 0.00000945, -0.00001530, 0.00000521, -0.00001438, 0.00001423, -0.00002764, 0.00000189, 0.00001438, -0.00000362, 0.00000823, -0.00000563, 0.00000164, 0.00000028, -0.00000353, -0.00000576, 0.00001339, -0.00000789, 0.00000148, -0.00000310, 0.00001402, -0.00000496, -0.00001187, 0.00000088, 0.00000141, 0.00000534, -0.00000646, 0.00000278, -0.00000541, 0.00000344, 0.00000027, 0.00000047, 0.00000154, 0.00000287, -0.00001207, 0.00000634, -0.00000007, 0.00000031, 0.00000074, -0.00000402, 0.00000621, -0.00000461, 0.00000305, -0.00000291, 0.00000309, -0.00000036, -0.00000422, -0.00000097, 0.00000124, 0.00000034, 0.00000343, 0.00000165, -0.00000716, 0.00000154, },
		{ 0.00000307, -0.00000178, -0.00000039, 0.00000087, -0.00000176, 0.00000233, 0.00000114, -0.00000162, -0.00000943, 0.00000491, 0.00000057, 0.00000118, 0.00000710, -0.00000672, -0.00000053, -0.00000104, 0.00000435, 0.00000339, -0.00000736, 0.00000158, -0.00000596, 0.00000072, 0.00000253, -0.00000362, 0.00001267, -0.00000610, -0.00000607, 0.00000202, -0.00000241, 0.00000215, 0.00000117, 0.00000243, -0.00000074, 0.00000660, -0.00001411, 0.00000292, 0.00000489, 0.00000254, -0.00000970, 0.00000424, 0.00001139, -0.00000497, -0.00001548, 0.00000750, -0.00000088, -0.00000501, -0.00000475, -0.00001084, 0.00001778, -0.00000752, -0.00000450, -0.00002520, 0.00000894, 0.00000020, -0.00000531, 0.00002071, -0.00000232, 0.00002331, -0.00000931, 0.00000070, -0.00000937, -0.00001778, -0.00002838, -0.00000872, -0.00016284, -0.00009078, 0.00073288, -0.00018335, -0.00010092, -0.00002134, -0.00002583, 0.00000581, -0.00003442, 0.00000929, -0.00002137, 0.00000508, -0.00001106, 0.00000071, -0.00001004, 0.00001177, -0.00002849, 0.00000560, 0.00001157, -0.00000164, 0.00000723, -0.00000650, 0.00000283, -0.00000089, -0.00000324, -0.00000539, 0.00001434, -0.00001045, 0.00000332, -0.00000386, 0.00001630, -0.00000934, -0.00000830, -0.00000144, 0.00000406, 0.00000268, -0.00000451, 0.00000115, -0.00000417, 0.00000312, 0.00000055, -0.00000011, 0.00000264, 0.00000119, -0.00001169, 0.00000777, -0.00000151, 0.00000132, -0.00000013, -0.00000352, 0.00000605, -0.00000487, 0.00000352, -0.00000335, 0.00000380, -0.00000132, -0.00000394, -0.00000074, 0.00000173, -0.00000023, 0.00000444, -0.00000023, -0.00000654, 0.00000222, },
		{ 0.00000281, -0.00000202, -0.00000011, 0.00000077, -0.00000188, 0.00000290, 0.00000083, -0.00000251, -0.00000908, 0.00000618, -0.00000101, 0.00000304, 0.00000547, -0.00000643, -0.00000063, -0.00000037, 0.00000414, 0.00000318, -0.00000786, 0.00000239, -0.00000714, 0.00000274, 0.00000063, -0.00000182, 0.00001201, -0.00000741, -0.00000539, 0.00000278, -0.00000334, 0.00000351, 0.00000013, 0.00000276, -0.00000106, 0.00000703, -0.00001571, 0.00000569, 0.00000264, 0.00000378, -0.00001241, 0.00000935, 0.00000726, -0.00000365, -0.00001767, 0.00001237, -0.00000612, -0.00000046, -0.00000969, -0.00000508, 0.00001295, -0.00000299, -0.00000965, -0.00002137, 0.00000925, 0.00000015, -0.00000592, 0.00002194, -0.00000618, 0.00002771, -0.00001741, 0.00000829, -0.00001851, -0.00000734, -0.00004158, 0.00000697, -0.00020234, -0.00002855, 0.00075200, -0.00027003, -0.00004972, -0.00004700, -0.00000485, -0.00000959, -0.00002218, -0.00000033, -0.00001291, -0.00000159, -0.00000472, -0.00000592, -0.00000326, 0.00000654, -0.00002665, 0.00000726, 0.00000999, -0.00000029, 0.00000634, -0.00000734, 0.00000411, -0.00000225, -0.00000274, -0.00000497, 0.00001499, -0.00001285, 0.00000522, -0.00000444, 0.00001812, -0.00001351, -0.00000453, -0.00000407, 0.00000706, -0.00000047, -0.00000197, -0.00000114, -0.00000219, 0.00000214, 0.00000132, -0.00000106, 0.00000410, -0.00000109, -0.00001050, 0.00000855, -0.00000264, 0.00000223, -0.00000102, -0.00000295, 0.00000579, -0.00000501, 0.00000389, -0.00000370, 0.00000440, -0.00000228, -0.00000363, -0.00000038, 0.00000199, -0.00000060, 0.00000527, -0.00000220, -0.00000556, 0.00000264, },
		{ 0.00000262, -0.00000228, 0.00000024, 0.00000057, -0.00000189, 0.00000347, 0.00000039, -0.00000338, -0.00000843, 0.00000713, -0.00000243, 0.00000493, 0.00000356, -0.00000581, -0.00000094, 0.00000050, 0.00000367, 0.00000317, -0.00000848, 0.00000329, -0.00000838, 0.00000492, -0.00000162, 0.00000054, 0.00001063, -0.00000823, -0.00000476, 0.00000353, -0.00000436, 0.00000506, -0.00000124, 0.00000336, -0.00000150, 0.00000741, -0.00001715, 0.00000837, 0.00000032, 0.00000495, -0.00001481, 0.00001448, 0.00000249, -0.00000172, -0.00002002, 0.00001743, -0.00001194, 0.00000487, -0.00001545, 0.00000170, 0.00000682, 0.00000316, -0.00001693, -0.00001486, 0.00000696, 0.00000198, -0.00000808, 0.00002439, -0.00001105, 0.00003316, -0.00002675, 0.00001755, -0.00002955, 0.00000537, -0.00005770, 0.00002617, -0.00024878, 0.00005107, 0.00074941, -0.00034710, 0.00000306, -0.00007588, 0.00001894, -0.00002759, -0.00000769, -0.00001193, -0.00000258, -0.00000997, 0.00000317, -0.00001414, 0.00000552, -0.00000115, -0.00002217, 0.00000671, 0.00000993, 0.00000015, 0.00000578, -0.00000826, 0.00000550, -0.00000380, -0.00000203, -0.00000453, 0.00001530, -0.00001499, 0.00000707, -0.00000475, 0.00001932, -0.00001719, -0.00000086, -0.00000675, 0.00001014, -0.00000383, 0.00000094, -0.00000388, 0.00000035, 0.00000058, 0.00000251, -0.00000232, 0.00000584, -0.00000386, -0.00000851, 0.00000862, -0.00000336, 0.00000295, -0.00000186, -0.00000234, 0.00000546, -0.00000504, 0.00000416, -0.00000394, 0.00000489, -0.00000321, -0.00000329, 0.00000010, 0.00000199, -0.00000070, 0.00000581, -0.00000411, -0.00000430, 0.00000283, },
		{ 0.00000246, -0.00000255, 0.00000064, 0.00000029, -0.00000179, 0.00000400, -0.00000019, -0.00000420, -0.00000751, 0.00000769, -0.00000357, 0.00000670, 0.00000150, -0.00000493, -0.00000137, 0.00000149, 0.00000301, 0.00000329, -0.00000915, 0.00000420, -0.00000958, 0.00000713, -0.00000409, 0.00000334, 0.00000858, -0.00000854, -0.00000422, 0.00000421, -0.00000537, 0.00000670, -0.00000285, 0.00000420, -0.00000201, 0.00000769, -0.00001833, 0.00001082, -0.00000190, 0.00000586, -0.00001663, 0.00001925, -0.00000251, 0.00000051, -0.00002222, 0.00002227, -0.00001786, 0.00001051, -0.00002155, 0.00000897, -0.00000003, 0.00001033, -0.00002576, -0.00000609, 0.00000227, 0.00000566, -0.00001174, 0.00002798, -0.00001680, 0.00003938, -0.00003694, 0.00002797, -0.00004191, 0.00001963, -0.00007584, 0.00004772, -0.00030076, 0.00014785, 0.00072238, -0.00040985, 0.00005330, -0.00010521, 0.00004356, -0.00004667, 0.00000783, -0.00002447, 0.00000867, -0.00001922, 0.00001181, -0.00002318, 0.00001554, -0.00001071, -0.00001540, 0.00000403, 0.00001146, -0.00000045, 0.00000567, -0.00000929, 0.00000698, -0.00000550, -0.00000114, -0.00000404, 0.00001522, -0.00001677, 0.00000879, -0.00000472, 0.00001976, -0.00002014, 0.00000240, -0.00000916, 0.00001302, -0.00000712, 0.00000391, -0.00000681, 0.00000325, -0.00000138, 0.00000401, -0.00000376, 0.00000771, -0.00000697, -0.00000583, 0.00000796, -0.00000360, 0.00000340, -0.00000259, -0.00000175, 0.00000508, -0.00000498, 0.00000435, -0.00000409, 0.00000524, -0.00000408, -0.00000290, 0.00000067, 0.00000172, -0.00000048, 0.00000596, -0.00000582, -0.00000288, 0.00000281, },
		{ 0.00000233, -0.00000281, 0.00000106, -0.00000008, -0.00000157, 0.00000447, -0.00000089, -0.00000492, -0.00000635, 0.00000784, -0.00000432, 0.00000821, -0.00000055, -0.00000393, -0.00000185, 0.00000249, 0.00000226, 0.00000345, -0.00000978, 0.00000504, -0.00001063, 0.00000920, -0.00000659, 0.00000640, 0.00000598, -0.00000835, -0.00000377, 0.00000479, -0.00000628, 0.00000830, -0.00000458, 0.00000520, -0.00000255, 0.00000780, -0.00001916, 0.00001289, -0.00000386, 0.00000634, -0.00001759, 0.00002324, -0.00000732, 0.00000267, -0.00002393, 0.00002645, -0.00002337, 0.00001590, -0.00002741, 0.00001607, -0.00000687, 0.00001773, -0.00003532, 0.00000424, -0.00000436, 0.00001087, -0.00001663, 0.00003243, -0.00002304, 0.00004589, -0.00004737, 0.00003882, -0.00005471, 0.00003435, -0.00009467, 0.00006984, -0.00035594, 0.00026029, 0.00066936, -0.00045424, 0.00009700, -0.00013203, 0.00006681, -0.00006508, 0.00002293, -0.00003667, 0.00001971, -0.00002831, 0.00002023, -0.00003206, 0.00002585, -0.00002130, -0.00000695, -0.00000043, 0.00001444, -0.00000211, 0.00000604, -0.00001042, 0.00000847, -0.00000726, -0.00000013, -0.00000349, 0.00001470, -0.00001808, 0.00001028, -0.00000431, 0.00001934, -0.00002215, 0.00000499, -0.00001103, 0.00001538, -0.00001002, 0.00000664, -0.00000962, 0.00000622, -0.00000353, 0.00000564, -0.00000521, 0.00000950, -0.00001019, -0.00000263, 0.00000665, -0.00000335, 0.00000354, -0.00000319, -0.00000119, 0.00000467, -0.00000484, 0.00000444, -0.00000412, 0.00000545, -0.00000488, -0.00000246, 0.00000132, 0.00000122, 0.00000007, 0.00000565, -0.00000721, -0.00000142, 0.00000266, },
		{ 0.00000219, -0.00000301, 0.00000148, -0.00000050, -0.00000122, 0.00000484, -0.00000171, -0.00000550, -0.00000501, 0.00000758, -0.00000463, 0.00000934, -0.00000246, -0.00000291, -0.00000226, 0.00000339, 0.00000155, 0.00000355, -0.00001026, 0.00000568, -0.00001138, 0.00001098, -0.00000891, 0.00000950, 0.00000298, -0.00000773, -0.00000337, 0.00000518, -0.00000696, 0.00000971, -0.00000628, 0.00000623, -0.00000302, 0.00000767, -0.00001954, 0.00001444, -0.00000539, 0.00000623, -0.00001747, 0.00002608, -0.00001151, 0.00000441, -0.00002476, 0.00002950, -0.00002791, 0.00002046, -0.00003240, 0.00002231, -0.00001290, 0.00002442, -0.00004463, 0.00001517, -0.00001215, 0.00001703, -0.00002222, 0.00003721, -0.00002921, 0.00005204, -0.00005724, 0.00004914, -0.00006680, 0.00004815, -0.00011246, 0.00009017, -0.00041111, 0.00038559, 0.00059025, -0.00047724, 0.00013059, -0.00015351, 0.00008646, -0.00008101, 0.00003609, -0.00004721, 0.00002936, -0.00003614, 0.00002738, -0.00003972, 0.00003537, -0.00003189, 0.00000233, -0.00000610, 0.00001854, -0.00000463, 0.00000680, -0.00001155, 0.00000985, -0.00000894, 0.00000092, -0.00000283, 0.00001369, -0.00001882, 0.00001146, -0.00000349, 0.00001801, -0.00002306, 0.00000667, -0.00001211, 0.00001697, -0.00001223, 0.00000881, -0.00001199, 0.00000896, -0.00000561, 0.00000718, -0.00000649, 0.00001100, -0.00001326, 0.00000086, 0.00000482, -0.00000268, 0.00000340, -0.00000364, -0.00000066, 0.00000423, -0.00000461, 0.00000443, -0.00000406, 0.00000551, -0.00000560, -0.00000197, 0.00000197, 0.00000052, 0.00000092, 0.00000485, -0.00000817, -0.00000002, 0.00000242, },
		{ 0.00000201, -0.00000314, 0.00000186, -0.00000096, -0.00000074, 0.00000509, -0.00000262, -0.00000592, -0.00000355, 0.00000693, -0.00000446, 0.00000997, -0.00000407, -0.00000201, -0.00000250, 0.00000407, 0.00000100, 0.00000347, -0.00001049, 0.00000603, -0.00001171, 0.00001226, -0.00001081, 0.00001240, -0.00000021, -0.00000680, -0.00000297, 0.00000531, -0.00000729, 0.00001078, -0.00000780, 0.00000716, -0.00000333, 0.00000722, -0.00001940, 0.00001537, -0.00000637, 0.00000544, -0.00001610, 0.00002746, -0.00001467, 0.00000536, -0.00002441, 0.00003101, -0.00003099, 0.00002365, -0.00003592, 0.00002702, -0.00001732, 0.00002948, -0.00005259, 0.00002560, -0.00002015, 0.00002335, -0.00002780, 0.00004165, -0.00003461, 0.00005701, -0.00006560, 0.00005784, -0.00007689, 0.00005944, -0.00012719, 0.00010595, -0.00046227, 0.00051964, 0.00048652, -0.00047708, 0.00015122, -0.00016718, 0.00010048, -0.00009276, 0.00004588, -0.00005482, 0.00003645, -0.00004165, 0.00003225, -0.00004511, 0.00004297, -0.00004135, 0.00001144, -0.00001221, 0.00002325, -0.00000773, 0.00000777, -0.00001254, 0.00001095, -0.00001038, 0.00000185, -0.00000200, 0.00001215, -0.00001889, 0.00001225, -0.00000227, 0.00001579, -0.00002279, 0.00000729, -0.00001220, 0.00001756, -0.00001351, 0.00001015, -0.00001365, 0.00001119, -0.00000737, 0.00000843, -0.00000740, 0.00001200, -0.00001591, 0.00000436, 0.00000266, -0.00000166, 0.00000301, -0.00000398, -0.00000014, 0.00000375, -0.00000429, 0.00000434, -0.00000389, 0.00000541, -0.00000621, -0.00000140, 0.00000258, -0.00000030, 0.00000203, 0.00000357, -0.00000863, 0.00000122, 0.00000217, },
		{ 0.00000176, -0.00000317, 0.00000218, -0.00000142, -0.00000014, 0.00000520, -0.00000360, -0.00000613, -0.00000205, 0.00000596, -0.00000380, 0.00001005, -0.00000529, -0.00000133, -0.00000247, 0.00000442, 0.00000071, 0.00000311, -0.00001037, 0.00000597, -0.00001150, 0.00001291, -0.00001209, 0.00001483, -0.00000333, -0.00000570, -0.00000248, 0.00000508, -0.00000715, 0.00001134, -0.00000895, 0.00000787, -0.00000339, 0.00000639, -0.00001866, 0.00001559, -0.00000673, 0.00000391, -0.00001342, 0.00002716, -0.00001650, 0.00000525, -0.00002262, 0.00003067, -0.00003219, 0.00002499, -0.00003748, 0.00002960, -0.00001944, 0.00003203, -0.00005817, 0.00003436, -0.00002727, 0.00002890, -0.00003253, 0.00004496, -0.00003843, 0.00005993, -0.00007147, 0.00006375, -0.00008363, 0.00006656, -0.00013675, 0.00011420, -0.00050485, 0.00065714, 0.00036124, -0.00045346, 0.00015706, -0.00017117, 0.00010725, -0.00009895, 0.00005111, -0.00005845, 0.00004001, -0.00004395, 0.00003395, -0.00004731, 0.00004762, -0.00004859, 0.00001935, -0.00001792, 0.00002794, -0.00001098, 0.00000870, -0.00001318, 0.00001159, -0.00001139, 0.00000254, -0.00000092, 0.00001002, -0.00001823, 0.00001259, -0.00000068, 0.00001274, -0.00002134, 0.00000679, -0.00001121, 0.00001702, -0.00001369, 0.00001045, -0.00001433, 0.00001264, -0.00000857, 0.00000918, -0.00000777, 0.00001228, -0.00001788, 0.00000760, 0.00000039, -0.00000045, 0.00000244, -0.00000422, 0.00000038, 0.00000322, -0.00000387, 0.00000413, -0.00000361, 0.00000515, -0.00000671, -0.00000075, 0.00000309, -0.00000116, 0.00000332, 0.00000185, -0.00000855, 0.00000221, 0.00000194, },
		{ 0.00000145, -0.00000309, 0.00000242, -0.00000187, 0.00000057, 0.00000512, -0.00000460, -0.00000611, -0.00000059, 0.00000475, -0.00000271, 0.00000955, -0.00000603, -0.00000095, -0.00000211, 0.00000438, 0.00000079, 0.00000238, -0.00000982, 0.00000544, -0.00001067, 0.00001278, -0.00001257, 0.00001655, -0.00000616, -0.00000459, -0.00000182, 0.00000442, -0.00000643, 0.00001126, -0.00000959, 0.00000820, -0.00000310, 0.00000513, -0.00001729, 0.00001506, -0.00000645, 0.00000167, -0.00000944, 0.00002507, -0.00001678, 0.00000390, -0.00001925, 0.00002828, -0.00003124, 0.00002418, -0.00003671, 0.00002966, -0.00001873, 0.00003138, -0.00006044, 0.00004038, -0.00003246, 0.00003272, -0.00003555, 0.00004634, -0.00003988, 0.00005997, -0.00007390, 0.00006580, -0.00008576, 0.00006794, -0.00013911, 0.00011193, -0.00053398, 0.00079187, 0.00021897, -0.00040759, 0.00014740, -0.00016442, 0.00010575, -0.00009864, 0.00005094, -0.00005734, 0.00003936, -0.00004238, 0.00003187, -0.00004563, 0.00004852, -0.00005265, 0.00002507, -0.00002236, 0.00003194, -0.00001391, 0.00000926, -0.00001325, 0.00001156, -0.00001177, 0.00000282, 0.00000047, 0.00000729, -0.00001678, 0.00001245, 0.00000123, 0.00000899, -0.00001881, 0.00000517, -0.00000912, 0.00001531, -0.00001268, 0.00000959, -0.00001388, 0.00001312, -0.00000902, 0.00000930, -0.00000747, 0.00001170, -0.00001893, 0.00001031, -0.00000175, 0.00000080, 0.00000180, -0.00000445, 0.00000095, 0.00000259, -0.00000332, 0.00000380, -0.00000322, 0.00000472, -0.00000707, -0.00000002, 0.00000344, -0.00000197, 0.00000467, -0.00000022, -0.00000794, 0.00000291, 0.00000178, },
		{ 0.00000106, -0.00000291, 0.00000256, -0.00000229, 0.00000137, 0.00000484, -0.00000559, -0.00000585, 0.00000074, 0.00000340, -0.00000127, 0.00000851, -0.00000626, -0.00000090, -0.00000139, 0.00000390, 0.00000127, 0.00000124, -0.00000879, 0.00000439, -0.00000918, 0.00001182, -0.00001210, 0.00001737, -0.00000845, -0.00000364, -0.00000092, 0.00000329, -0.00000507, 0.00001043, -0.00000957, 0.00000804, -0.00000239, 0.00000339, -0.00001531, 0.00001380, -0.00000554, -0.00000116, -0.00000432, 0.00002125, -0.00001545, 0.00000125, -0.00001430, 0.00002383, -0.00002804, 0.00002107, -0.00003346, 0.00002699, -0.00001491, 0.00002709, -0.00005872, 0.00004276, -0.00003476, 0.00003396, -0.00003607, 0.00004505, -0.00003828, 0.00005642, -0.00007211, 0.00006310, -0.00008224, 0.00006230, -0.00013255, 0.00009645, -0.00054483, 0.00091698, 0.00006559, -0.00034217, 0.00012276, -0.00014679, 0.00009562, -0.00009143, 0.00004503, -0.00005117, 0.00003420, -0.00003667, 0.00002571, -0.00003970, 0.00004512, -0.00005281, 0.00002779, -0.00002475, 0.00003459, -0.00001605, 0.00000917, -0.00001254, 0.00001068, -0.00001137, 0.00000257, 0.00000224, 0.00000395, -0.00001452, 0.00001181, 0.00000335, 0.00000471, -0.00001534, 0.00000255, -0.00000602, 0.00001248, -0.00001049, 0.00000754, -0.00001223, 0.00001253, -0.00000861, 0.00000866, -0.00000642, 0.00001018, -0.00001890, 0.00001223, -0.00000354, 0.00000192, 0.00000117, -0.00000471, 0.00000161, 0.00000185, -0.00000263, 0.00000334, -0.00000272, 0.00000412, -0.00000727, 0.00000077, 0.00000359, -0.00000264, 0.00000596, -0.00000253, -0.00000684, 0.00000328, 0.00000171, },
		{ 0.00000060, -0.00000262, 0.00000261, -0.00000265, 0.00000225, 0.00000434, -0.00000651, -0.00000537, 0.00000188, 0.00000202, 0.00000041, 0.00000698, -0.00000600, -0.00000119, -0.00000032, 0.00000298, 0.00000217, -0.00000033, -0.00000729, 0.00000285, -0.00000704, 0.00001002, -0.00001064, 0.00001714, -0.00001003, -0.00000298, 0.00000028, 0.00000165, -0.00000304, 0.00000878, -0.00000882, 0.00000732, -0.00000120, 0.00000119, -0.00001273, 0.00001185, -0.00000408, -0.00000442, 0.00000171, 0.00001587, -0.00001258, -0.00000261, -0.00000793, 0.00001746, -0.00002269, 0.00001574, -0.00002781, 0.00002165, -0.00000801, 0.00001904, -0.00005268, 0.00004090, -0.00003347, 0.00003192, -0.00003347, 0.00004057, -0.00003314, 0.00004879, -0.00006560, 0.00005508, -0.00007242, 0.00004879, -0.00011587, 0.00006569, -0.00053301, 0.00102545, -0.00009206, -0.00026124, 0.00008487, -0.00011905, 0.00007728, -0.00007758, 0.00003355, -0.00004009, 0.00002462, -0.00002693, 0.00001558, -0.00002955, 0.00003727, -0.00004870, 0.00002693, -0.00002447, 0.00003534, -0.00001697, 0.00000815, -0.00001088, 0.00000883, -0.00001005, 0.00000167, 0.00000440, 0.00000004, -0.00001147, 0.00001068, 0.00000559, 0.00000013, -0.00001117, -0.00000087, -0.00000207, 0.00000869, -0.00000727, 0.00000440, -0.00000940, 0.00001083, -0.00000729, 0.00000727, -0.00000464, 0.00000771, -0.00001771, 0.00001322, -0.00000480, 0.00000278, 0.00000068, -0.00000506, 0.00000240, 0.00000097, -0.00000177, 0.00000273, -0.00000211, 0.00000335, -0.00000730, 0.00000161, 0.00000352, -0.00000308, 0.00000706, -0.00000494, -0.00000533, 0.00000334, 0.00000174, },
		{ 0.00000009, -0.00000225, 0.00000256, -0.00000294, 0.00000319, 0.00000360, -0.00000731, -0.00000468, 0.00000277, 0.00000073, 0.00000220, 0.00000509, -0.00000531, -0.00000179, 0.00000104, 0.00000168, 0.00000345, -0.00000229, -0.00000535, 0.00000086, -0.00000435, 0.00000744, -0.00000822, 0.00001580, -0.00001076, -0.00000272, 0.00000179, -0.00000047, -0.00000037, 0.00000633, -0.00000728, 0.00000597, 0.00000048, -0.00000146, -0.00000965, 0.00000931, -0.00000221, -0.00000788, 0.00000832, 0.00000924, -0.00000838, -0.00000748, -0.00000043, 0.00000950, -0.00001549, 0.00000847, -0.00002006, 0.00001400, 0.00000165, 0.00000749, -0.00004236, 0.00003460, -0.00002820, 0.00002623, -0.00002740, 0.00003262, -0.00002425, 0.00003693, -0.00005421, 0.00004160, -0.00005612, 0.00002715, -0.00008861, 0.00001841, -0.00049492, 0.00111050, -0.00024657, -0.00016987, 0.00003651, -0.00008291, 0.00005189, -0.00005794, 0.00001722, -0.00002471, 0.00001120, -0.00001367, 0.00000199, -0.00001561, 0.00002523, -0.00004032, 0.00002226, -0.00002116, 0.00003382, -0.00001638, 0.00000602, -0.00000818, 0.00000595, -0.00000776, 0.00000007, 0.00000695, -0.00000436, -0.00000769, 0.00000910, 0.00000782, -0.00000451, -0.00000659, -0.00000483, 0.00000247, 0.00000418, -0.00000322, 0.00000035, -0.00000555, 0.00000811, -0.00000512, 0.00000516, -0.00000221, 0.00000439, -0.00001537, 0.00001316, -0.00000540, 0.00000324, 0.00000038, -0.00000555, 0.00000334, -0.00000006, -0.00000075, 0.00000196, -0.00000138, 0.00000242, -0.00000713, 0.00000247, 0.00000321, -0.00000323, 0.00000787, -0.00000729, -0.00000352, 0.00000312, 0.00000187, },
		{ -0.00000044, -0.00000182, 0.00000243, -0.00000315, 0.00000416, 0.00000262, -0.00000793, -0.00000381, 0.00000337, -0.00000038, 0.00000395, 0.00000296, -0.00000428, -0.00000261, 0.00000262, 0.00000008, 0.00000503, -0.00000454, -0.00000307, -0.00000146, -0.00000124, 0.00000425, -0.00000493, 0.00001339, -0.00001059, -0.00000290, 0.00000361, -0.00000299, 0.00000287, 0.00000313, -0.00000499, 0.00000403, 0.00000261, -0.00000448, -0.00000616, 0.00000631, -0.00000008, -0.00001127, 0.00001510, 0.00000176, -0.00000317, -0.00001301, 0.00000776, 0.00000042, -0.00000691, -0.00000025, -0.00001074, 0.00000460, 0.00001343, -0.00000697, -0.00002822, 0.00002405, -0.00001898, 0.00001683, -0.00001784, 0.00002124, -0.00001176, 0.00002107, -0.00003823, 0.00002298, -0.00003378, -0.00000218, -0.00005118, -0.00004547, -0.00042819, 0.00116613, -0.00039046, -0.00007386, -0.00001867, -0.00004080, 0.00002125, -0.00003393, -0.00000276, -0.00000611, -0.00000511, 0.00000220, -0.00001419, 0.00000128, 0.00000969, -0.00002807, 0.00001390, -0.00001474, 0.00002985, -0.00001412, 0.00000271, -0.00000445, 0.00000207, -0.00000451, -0.00000220, 0.00000983, -0.00000912, -0.00000329, 0.00000712, 0.00000991, -0.00000897, -0.00000190, -0.00000902, 0.00000728, -0.00000072, 0.00000134, -0.00000432, -0.00000092, 0.00000455, -0.00000224, 0.00000248, 0.00000072, 0.00000038, -0.00001199, 0.00001207, -0.00000526, 0.00000325, 0.00000033, -0.00000619, 0.00000443, -0.00000125, 0.00000043, 0.00000104, -0.00000056, 0.00000134, -0.00000674, 0.00000331, 0.00000267, -0.00000306, 0.00000827, -0.00000944, -0.00000152, 0.00000266, 0.00000206, },
		{ -0.00000097, -0.00000136, 0.00000224, -0.00000328, 0.00000515, 0.00000140, -0.00000832, -0.00000283, 0.00000366, -0.00000120, 0.00000553, 0.00000075, -0.00000302, -0.00000355, 0.00000428, -0.00000169, 0.00000678, -0.00000696, -0.00000059, -0.00000395, 0.00000211, 0.00000064, -0.00000099, 0.00001003, -0.00000954, -0.00000352, 0.00000565, -0.00000578, 0.00000651, -0.00000067, -0.00000205, 0.00000157, 0.00000512, -0.00000777, -0.00000242, 0.00000301, 0.00000213, -0.00001431, 0.00002161, -0.00000607, 0.00000260, -0.00001877, 0.00001611, -0.00000916, 0.00000243, -0.00000978, -0.00000055, -0.00000573, 0.00002643, -0.00002338, -0.00001110, 0.00000988, -0.00000621, 0.00000408, -0.00000514, 0.00000686, 0.00000382, 0.00000182, -0.00001839, 0.00000009, -0.00000640, -0.00003798, -0.00000496, -0.00012477, -0.00033191, 0.00118751, -0.00051663, 0.00002071, -0.00007649, 0.00000428, -0.00001234, -0.00000741, -0.00002479, 0.00001431, -0.00002301, 0.00001943, -0.00003173, 0.00001995, -0.00000832, -0.00001276, 0.00000235, -0.00000546, 0.00002354, -0.00001024, -0.00000168, 0.00000017, -0.00000266, -0.00000044, -0.00000507, 0.00001291, -0.00001407, 0.00000158, 0.00000484, 0.00001174, -0.00001298, 0.00000258, -0.00001310, 0.00001202, -0.00000566, 0.00000605, -0.00000927, 0.00000419, 0.00000042, 0.00000116, -0.00000058, 0.00000391, -0.00000406, -0.00000777, 0.00001004, -0.00000441, 0.00000278, 0.00000054, -0.00000697, 0.00000565, -0.00000256, 0.00000175, -0.00000000, 0.00000035, 0.00000014, -0.00000613, 0.00000407, 0.00000195, -0.00000254, 0.00000819, -0.00001124, 0.00000053, 0.00000203, 0.00000229, },
		{ -0.00000145, -0.00000090, 0.00000201, -0.00000332, 0.00000611, -0.00000005, -0.00000843, -0.00000180, 0.00000366, -0.00000169, 0.00000681, -0.00000138, -0.00000168, -0.00000449, 0.00000590, -0.00000347, 0.00000855, -0.00000940, 0.00000192, -0.00000642, 0.00000545, -0.00000312, 0.00000334, 0.00000593, -0.00000774, -0.00000449, 0.00000780, -0.00000868, 0.00001035, -0.00000489, 0.00000139, -0.00000130, 0.00000788, -0.00001118, 0.00000141, -0.00000039, 0.00000423, -0.00001672, 0.00002740, -0.00001373, 0.00000846, -0.00002427, 0.00002406, -0.00001858, 0.00001181, -0.00001937, 0.00000968, -0.00001605, 0.00003956, -0.00004055, 0.00000780, -0.00000690, 0.00000926, -0.00001129, 0.00000999, -0.00000977, 0.00002162, -0.00001979, 0.00000415, -0.00002572, 0.00002439, -0.00007832, 0.00004767, -0.00021694, -0.00020694, 0.00117142, -0.00061887, 0.00010790, -0.00013255, 0.00004901, -0.00004627, 0.00001948, -0.00004702, 0.00003488, -0.00004098, 0.00003661, -0.00004923, 0.00003898, -0.00002746, 0.00000447, -0.00001156, 0.00000614, 0.00001521, -0.00000496, -0.00000695, 0.00000546, -0.00000798, 0.00000425, -0.00000836, 0.00001602, -0.00001899, 0.00000674, 0.00000237, 0.00001317, -0.00001633, 0.00000654, -0.00001675, 0.00001634, -0.00001027, 0.00001053, -0.00001410, 0.00000940, -0.00000398, 0.00000479, -0.00000378, 0.00000710, -0.00000861, -0.00000301, 0.00000725, -0.00000293, 0.00000187, 0.00000098, -0.00000785, 0.00000696, -0.00000394, 0.00000317, -0.00000115, 0.00000130, -0.00000117, -0.00000529, 0.00000472, 0.00000108, -0.00000171, 0.00000759, -0.00001257, 0.00000250, 0.00000130, 0.00000253, },
		{ -0.00000187, -0.00000048, 0.00000175, -0.00000326, 0.00000701, -0.00000170, -0.00000824, -0.00000079, 0.00000340, -0.00000179, 0.00000767, -0.00000329, -0.00000039, -0.00000531, 0.00000733, -0.00000511, 0.00001019, -0.00001169, 0.00000430, -0.00000867, 0.00000854, -0.00000672, 0.00000774, 0.00000139, -0.00000536, -0.00000570, 0.00000990, -0.00001148, 0.00001416, -0.00000927, 0.00000511, -0.00000438, 0.00001073, -0.00001454, 0.00000513, -0.00000371, 0.00000604, -0.00001824, 0.00003206, -0.00002070, 0.00001391, -0.00002904, 0.00003100, -0.00002715, 0.00002051, -0.00002824, 0.00001909, -0.00002537, 0.00005162, -0.00005711, 0.00002706, -0.00002495, 0.00002628, -0.00002823, 0.00002652, -0.00002757, 0.00004043, -0.00004244, 0.00002787, -0.00005268, 0.00005648, -0.00012060, 0.00010346, -0.00031808, -0.00005596, 0.00111655, -0.00069229, 0.00018238, -0.00018261, 0.00009008, -0.00007785, 0.00004451, -0.00006752, 0.00005386, -0.00005743, 0.00005220, -0.00006515, 0.00005682, -0.00004619, 0.00002225, -0.00002672, 0.00001924, 0.00000544, 0.00000131, -0.00001273, 0.00001104, -0.00001357, 0.00000926, -0.00001185, 0.00001896, -0.00002363, 0.00001193, -0.00000016, 0.00001412, -0.00001882, 0.00000970, -0.00001965, 0.00001990, -0.00001418, 0.00001440, -0.00001842, 0.00001433, -0.00000831, 0.00000837, -0.00000683, 0.00000999, -0.00001293, 0.00000196, 0.00000395, -0.00000097, 0.00000063, 0.00000157, -0.00000875, 0.00000827, -0.00000533, 0.00000462, -0.00000234, 0.00000225, -0.00000252, -0.00000424, 0.00000520, 0.00000015, -0.00000061, 0.00000647, -0.00001336, 0.00000427, 0.00000056, 0.00000273, },
		{ -0.00000219, -0.00000010, 0.00000148, -0.00000311, 0.00000781, -0.00000349, -0.00000771, 0.00000014, 0.00000292, -0.00000153, 0.00000806, -0.00000484, 0.00000071, -0.00000588, 0.00000844, -0.00000645, 0.00001151, -0.00001365, 0.00000635, -0.00001050, 0.00001113, -0.00000988, 0.00001185, -0.00000328, -0.00000265, -0.00000698, 0.00001177, -0.00001394, 0.00001764, -0.00001353, 0.00000885, -0.00000746, 0.00001347, -0.00001768, 0.00000855, -0.00000677, 0.00000739, -0.00001869, 0.00003524, -0.00002650, 0.00001849, -0.00003261, 0.00003640, -0.00003423, 0.00002782, -0.00003564, 0.00002684, -0.00003271, 0.00006145, -0.00007167, 0.00004512, -0.00004275, 0.00004346, -0.00004544, 0.00004319, -0.00004523, 0.00005885, -0.00006453, 0.00005097, -0.00007873, 0.00008735, -0.00016174, 0.00015840, -0.00042307, 0.00011652, 0.00102366, -0.00073367, 0.00023973, -0.00022291, 0.00012443, -0.00010456, 0.00006556, -0.00008443, 0.00006959, -0.00007083, 0.00006475, -0.00007804, 0.00007194, -0.00006296, 0.00003908, -0.00004185, 0.00003285, -0.00000504, 0.00000801, -0.00001857, 0.00001649, -0.00001899, 0.00001422, -0.00001526, 0.00002146, -0.00002769, 0.00001692, -0.00000260, 0.00001448, -0.00002030, 0.00001185, -0.00002157, 0.00002245, -0.00001710, 0.00001732, -0.00002187, 0.00001860, -0.00001220, 0.00001159, -0.00000949, 0.00001230, -0.00001666, 0.00000676, 0.00000043, 0.00000126, -0.00000082, 0.00000222, -0.00000959, 0.00000949, -0.00000664, 0.00000602, -0.00000352, 0.00000315, -0.00000385, -0.00000301, 0.00000548, -0.00000079, 0.00000069, 0.00000486, -0.00001356, 0.00000576, -0.00000015, 0.00000287, },
		{ -0.00000242, 0.00000021, 0.00000121, -0.00000284, 0.00000846, -0.00000538, -0.00000685, 0.00000091, 0.00000230, -0.00000092, 0.00000792, -0.00000594, 0.00000152, -0.00000611, 0.00000910, -0.00000736, 0.00001239, -0.00001513, 0.00000793, -0.00001175, 0.00001301, -0.00001229, 0.00001535, -0.00000772, 0.00000013, -0.00000812, 0.00001320, -0.00001582, 0.00002053, -0.00001735, 0.00001234, -0.00001030, 0.00001589, -0.00002038, 0.00001148, -0.00000937, 0.00000818, -0.00001794, 0.00003667, -0.00003074, 0.00002180, -0.00003462, 0.00003983, -0.00003928, 0.00003315, -0.00004093, 0.00003224, -0.00003724, 0.00006800, -0.00008291, 0.00006047, -0.00005873, 0.00005928, -0.00006147, 0.00005860, -0.00006132, 0.00007535, -0.00008438, 0.00007156, -0.00010166, 0.00011436, -0.00019840, 0.00020803, -0.00052588, 0.00030433, 0.00089564, -0.00074164, 0.00027685, -0.00025049, 0.00014955, -0.00012427, 0.00008081, -0.00009618, 0.00008064, -0.00007986, 0.00007301, -0.00008661, 0.00008296, -0.00007627, 0.00005347, -0.00005561, 0.00004585, -0.00001537, 0.00001451, -0.00002393, 0.00002132, -0.00002380, 0.00001873, -0.00001827, 0.00002327, -0.00003091, 0.00002144, -0.00000483, 0.00001423, -0.00002071, 0.00001286, -0.00002233, 0.00002377, -0.00001880, 0.00001904, -0.00002413, 0.00002188, -0.00001536, 0.00001420, -0.00001151, 0.00001377, -0.00001949, 0.00001104, -0.00000298, 0.00000352, -0.00000230, 0.00000279, -0.00001025, 0.00001052, -0.00000776, 0.00000729, -0.00000461, 0.00000393, -0.00000511, -0.00000163, 0.00000554, -0.00000164, 0.00000208, 0.00000284, -0.00001314, 0.00000689, -0.00000076, 0.00000294, },
		{ -0.00000254, 0.00000045, 0.00000095, -0.00000246, 0.00000890, -0.00000728, -0.00000568, 0.00000148, 0.00000162, -0.00000002, 0.00000728, -0.00000654, 0.00000197, -0.00000593, 0.00000925, -0.00000773, 0.00001270, -0.00001602, 0.00000891, -0.00001227, 0.00001402, -0.00001374, 0.00001790, -0.00001158, 0.00000269, -0.00000892, 0.00001402, -0.00001692, 0.00002255, -0.00002045, 0.00001528, -0.00001264, 0.00001778, -0.00002247, 0.00001375, -0.00001139, 0.00000834, -0.00001596, 0.00003621, -0.00003315, 0.00002356, -0.00003481, 0.00004098, -0.00004191, 0.00003605, -0.00004365, 0.00003477, -0.00003835, 0.00007044, -0.00008972, 0.00007173, -0.00007138, 0.00007226, -0.00007490, 0.00007136, -0.00007444, 0.00008843, -0.00010034, 0.00008777, -0.00011934, 0.00013489, -0.00022721, 0.00024770, -0.00061983, 0.00049988, 0.00073738, -0.00071682, 0.00029209, -0.00026338, 0.00016364, -0.00013541, 0.00008894, -0.00010157, 0.00008593, -0.00008354, 0.00007605, -0.00008990, 0.00008880, -0.00008490, 0.00006410, -0.00006672, 0.00005714, -0.00002467, 0.00002013, -0.00002827, 0.00002505, -0.00002754, 0.00002239, -0.00002057, 0.00002414, -0.00003301, 0.00002522, -0.00000669, 0.00001332, -0.00002001, 0.00001265, -0.00002185, 0.00002377, -0.00001913, 0.00001936, -0.00002500, 0.00002391, -0.00001754, 0.00001596, -0.00001270, 0.00001424, -0.00002116, 0.00001446, -0.00000596, 0.00000557, -0.00000365, 0.00000317, -0.00001061, 0.00001126, -0.00000861, 0.00000832, -0.00000555, 0.00000453, -0.00000621, -0.00000017, 0.00000537, -0.00000235, 0.00000346, 0.00000051, -0.00001215, 0.00000762, -0.00000124, 0.00000292, },
		{ -0.00000257, 0.00000062, 0.00000070, -0.00000196, 0.00000908, -0.00000911, -0.00000425, 0.00000180, 0.00000096, 0.00000108, 0.00000617, -0.00000660, 0.00000201, -0.00000531, 0.00000883, -0.00000750, 0.00001239, -0.00001623, 0.00000921, -0.00001201, 0.00001404, -0.00001406, 0.00001927, -0.00001456, 0.00000477, -0.00000921, 0.00001406, -0.00001704, 0.00002349, -0.00002256, 0.00001742, -0.00001426, 0.00001895, -0.00002379, 0.00001520, -0.00001271, 0.00000784, -0.00001278, 0.00003384, -0.00003359, 0.00002362, -0.00003309, 0.00003973, -0.00004193, 0.00003627, -0.00004353, 0.00003416, -0.00003569, 0.00006827, -0.00009134, 0.00007782, -0.00007943, 0.00008109, -0.00008444, 0.00008020, -0.00008333, 0.00009675, -0.00011095, 0.00009801, -0.00012988, 0.00014661, -0.00024513, 0.00027299, -0.00069809, 0.00069465, 0.00055553, -0.00066175, 0.00028538, -0.00026080, 0.00016581, -0.00013713, 0.00008919, -0.00009996, 0.00008490, -0.00008135, 0.00007336, -0.00008736, 0.00008876, -0.00008795, 0.00006991, -0.00007409, 0.00006572, -0.00003212, 0.00002424, -0.00003109, 0.00002723, -0.00002979, 0.00002483, -0.00002189, 0.00002386, -0.00003376, 0.00002804, -0.00000808, 0.00001179, -0.00001827, 0.00001126, -0.00002013, 0.00002243, -0.00001808, 0.00001824, -0.00002436, 0.00002455, -0.00001856, 0.00001674, -0.00001297, 0.00001358, -0.00002148, 0.00001676, -0.00000825, 0.00000718, -0.00000468, 0.00000322, -0.00001058, 0.00001160, -0.00000909, 0.00000905, -0.00000625, 0.00000489, -0.00000709, 0.00000131, 0.00000498, -0.00000285, 0.00000471, -0.00000201, -0.00001064, 0.00000794, -0.00000157, 0.00000281, },
		{ -0.00000253, 0.00000075, 0.00000045, -0.00000132, 0.00000896, -0.00001077, -0.00000261, 0.00000185, 0.00000039, 0.00000228, 0.00000466, -0.00000617, 0.00000165, -0.00000426, 0.00000785, -0.00000667, 0.00001142, -0.00001575, 0.00000883, -0.00001095, 0.00001306, -0.00001318, 0.00001929, -0.00001643, 0.00000615, -0.00000883, 0.00001323, -0.00001608, 0.00002317, -0.00002348, 0.00001854, -0.00001497, 0.00001923, -0.00002420, 0.00001574, -0.00001326, 0.00000672, -0.00000855, 0.00002967, -0.00003205, 0.00002196, -0.00002949, 0.00003613, -0.00003934, 0.00003381, -0.00004057, 0.00003042, -0.00002927, 0.00006136, -0.00008740, 0.00007806, -0.00008196, 0.00008476, -0.00008906, 0.00008413, -0.00008699, 0.00009931, -0.00011512, 0.00010102, -0.00013185, 0.00014770, -0.00024969, 0.00028008, -0.00075413, 0.00087960, 0.00035813, -0.00058071, 0.00025817, -0.00024316, 0.00015610, -0.00012935, 0.00008148, -0.00009129, 0.00007747, -0.00007325, 0.00006494, -0.00007892, 0.00008264, -0.00008499, 0.00007022, -0.00007691, 0.00007080, -0.00003704, 0.00002633, -0.00003198, 0.00002754, -0.00003024, 0.00002574, -0.00002199, 0.00002228, -0.00003300, 0.00002971, -0.00000890, 0.00000969, -0.00001560, 0.00000878, -0.00001727, 0.00001986, -0.00001571, 0.00001572, -0.00002221, 0.00002373, -0.00001834, 0.00001648, -0.00001230, 0.00001180, -0.00002039, 0.00001776, -0.00000964, 0.00000817, -0.00000527, 0.00000285, -0.00001007, 0.00001147, -0.00000912, 0.00000939, -0.00000667, 0.00000495, -0.00000766, 0.00000272, 0.00000439, -0.00000311, 0.00000575, -0.00000456, -0.00000872, 0.00000788, -0.00000174, 0.00000262, },
		{ -0.00000244, 0.00000085, 0.00000019, -0.00000055, 0.00000849, -0.00001217, -0.00000086, 0.00000165, -0.00000002, 0.00000348, 0.00000287, -0.00000530, 0.00000092, -0.00000282, 0.00000638, -0.00000528, 0.00000985, -0.00001460, 0.00000780, -0.00000917, 0.00001115, -0.00001113, 0.00001792, -0.00001703, 0.00000668, -0.00000772, 0.00001148, -0.00001399, 0.00002154, -0.00002308, 0.00001851, -0.00001463, 0.00001852, -0.00002364, 0.00001531, -0.00001302, 0.00000506, -0.00000350, 0.00002393, -0.00002870, 0.00001873, -0.00002422, 0.00003042, -0.00003435, 0.00002887, -0.00003501, 0.00002385, -0.00001941, 0.00005000, -0.00007798, 0.00007226, -0.00007853, 0.00008268, -0.00008813, 0.00008255, -0.00008484, 0.00009551, -0.00011221, 0.00009611, -0.00012444, 0.00013709, -0.00023925, 0.00026607, -0.00078219, 0.00104579, 0.00015407, -0.00047938, 0.00021329, -0.00021201, 0.00013551, -0.00011278, 0.00006642, -0.00007608, 0.00006414, -0.00005972, 0.00005124, -0.00006503, 0.00007073, -0.00007607, 0.00006481, -0.00007475, 0.00007187, -0.00003896, 0.00002603, -0.00003071, 0.00002575, -0.00002869, 0.00002493, -0.00002075, 0.00001935, -0.00003065, 0.00003009, -0.00000909, 0.00000711, -0.00001215, 0.00000539, -0.00001345, 0.00001625, -0.00001221, 0.00001197, -0.00001869, 0.00002152, -0.00001692, 0.00001520, -0.00001074, 0.00000900, -0.00001793, 0.00001740, -0.00000999, 0.00000840, -0.00000531, 0.00000200, -0.00000903, 0.00001082, -0.00000866, 0.00000929, -0.00000676, 0.00000469, -0.00000787, 0.00000400, 0.00000366, -0.00000311, 0.00000647, -0.00000703, -0.00000649, 0.00000746, -0.00000177, 0.00000238, },
		{ -0.00000233, 0.00000094, -0.00000007, 0.00000035, 0.00000765, -0.00001321, 0.00000091, 0.00000121, -0.00000022, 0.00000457, 0.00000091, -0.00000408, -0.00000009, -0.00000109, 0.00000449, -0.00000341, 0.00000774, -0.00001289, 0.00000624, -0.00000679, 0.00000846, -0.00000805, 0.00001521, -0.00001633, 0.00000630, -0.00000585, 0.00000886, -0.00001083, 0.00001860, -0.00002134, 0.00001725, -0.00001318, 0.00001677, -0.00002208, 0.00001391, -0.00001201, 0.00000299, 0.00000211, 0.00001696, -0.00002378, 0.00001417, -0.00001763, 0.00002300, -0.00002736, 0.00002186, -0.00002730, 0.00001501, -0.00000678, 0.00003485, -0.00006363, 0.00006071, -0.00006915, 0.00007472, -0.00008146, 0.00007526, -0.00007675, 0.00008525, -0.00010213, 0.00008320, -0.00010755, 0.00011456, -0.00021326, 0.00022936, -0.00077776, 0.00118489, -0.00004738, -0.00036448, 0.00015470, -0.00016988, 0.00010584, -0.00008883, 0.00004518, -0.00005540, 0.00004590, -0.00004171, 0.00003322, -0.00004659, 0.00005382, -0.00006177, 0.00005393, -0.00006759, 0.00006876, -0.00003768, 0.00002321, -0.00002718, 0.00002184, -0.00002508, 0.00002234, -0.00001813, 0.00001508, -0.00002672, 0.00002914, -0.00000865, 0.00000418, -0.00000815, 0.00000131, -0.00000892, 0.00001184, -0.00000784, 0.00000725, -0.00001400, 0.00001810, -0.00001441, 0.00001302, -0.00000845, 0.00000536, -0.00001425, 0.00001573, -0.00000928, 0.00000783, -0.00000475, 0.00000065, -0.00000748, 0.00000966, -0.00000770, 0.00000874, -0.00000649, 0.00000408, -0.00000768, 0.00000505, 0.00000284, -0.00000286, 0.00000681, -0.00000926, -0.00000409, 0.00000676, -0.00000167, 0.00000211, },
		{ -0.00000223, 0.00000105, -0.00000035, 0.00000137, 0.00000643, -0.00001382, 0.00000260, 0.00000058, -0.00000019, 0.00000546, -0.00000108, -0.00000263, -0.00000131, 0.00000082, 0.00000231, -0.00000119, 0.00000524, -0.00001074, 0.00000428, -0.00000401, 0.00000520, -0.00000415, 0.00001132, -0.00001440, 0.00000503, -0.00000329, 0.00000547, -0.00000673, 0.00001448, -0.00001832, 0.00001482, -0.00001065, 0.00001401, -0.00001957, 0.00001160, -0.00001030, 0.00000068, 0.00000794, 0.00000915, -0.00001768, 0.00000865, -0.00001016, 0.00001439, -0.00001892, 0.00001334, -0.00001811, 0.00000465, 0.00000772, 0.00001691, -0.00004533, 0.00004419, -0.00005439, 0.00006126, -0.00006937, 0.00006260, -0.00006310, 0.00006898, -0.00008541, 0.00006291, -0.00008190, 0.00008084, -0.00017234, 0.00016981, -0.00073795, 0.00128980, -0.00023722, -0.00024321, 0.00008715, -0.00012007, 0.00006958, -0.00005948, 0.00001950, -0.00003079, 0.00002415, -0.00002054, 0.00001220, -0.00002491, 0.00003314, -0.00004310, 0.00003831, -0.00005585, 0.00006164, -0.00003329, 0.00001796, -0.00002154, 0.00001594, -0.00001954, 0.00001804, -0.00001423, 0.00000964, -0.00002132, 0.00002687, -0.00000760, 0.00000103, -0.00000383, -0.00000318, -0.00000396, 0.00000696, -0.00000294, 0.00000188, -0.00000847, 0.00001370, -0.00001103, 0.00001014, -0.00000566, 0.00000115, -0.00000961, 0.00001290, -0.00000757, 0.00000647, -0.00000363, -0.00000116, -0.00000544, 0.00000803, -0.00000627, 0.00000775, -0.00000589, 0.00000315, -0.00000707, 0.00000582, 0.00000201, -0.00000238, 0.00000674, -0.00001114, -0.00000165, 0.00000583, -0.00000148, 0.00000184, },
		{ -0.00000216, 0.00000119, -0.00000065, 0.00000247, 0.00000484, -0.00001394, 0.00000411, -0.00000016, 0.00000008, 0.00000607, -0.00000297, -0.00000106, -0.00000260, 0.00000279, -0.00000000, 0.00000123, 0.00000250, -0.00000831, 0.00000212, -0.00000105, 0.00000167, 0.00000026, 0.00000653, -0.00001139, 0.00000298, -0.00000018, 0.00000151, -0.00000191, 0.00000939, -0.00001420, 0.00001133, -0.00000715, 0.00001035, -0.00001623, 0.00000848, -0.00000799, -0.00000171, 0.00001365, 0.00000096, -0.00001083, 0.00000261, -0.00000234, 0.00000521, -0.00000969, 0.00000402, -0.00000818, -0.00000630, 0.00002299, -0.00000254, -0.00002439, 0.00002394, -0.00003526, 0.00004317, -0.00005263, 0.00004534, -0.00004475, 0.00004766, -0.00006312, 0.00003646, -0.00004896, 0.00003764, -0.00011835, 0.00008893, -0.00066182, 0.00135503, -0.00040723, -0.00012285, 0.00001577, -0.00006631, 0.00002964, -0.00002712, -0.00000859, -0.00000409, 0.00000057, 0.00000217, -0.00001025, -0.00000156, 0.00001023, -0.00002146, 0.00001906, -0.00004031, 0.00005106, -0.00002614, 0.00001059, -0.00001410, 0.00000838, -0.00001234, 0.00001225, -0.00000925, 0.00000327, -0.00001469, 0.00002340, -0.00000601, -0.00000216, 0.00000054, -0.00000778, 0.00000110, 0.00000194, 0.00000213, -0.00000374, -0.00000247, 0.00000866, -0.00000705, 0.00000679, -0.00000263, -0.00000328, -0.00000433, 0.00000915, -0.00000501, 0.00000442, -0.00000202, -0.00000335, -0.00000300, 0.00000599, -0.00000443, 0.00000637, -0.00000499, 0.00000193, -0.00000605, 0.00000626, 0.00000123, -0.00000172, 0.00000625, -0.00001258, 0.00000073, 0.00000477, -0.00000123, 0.00000158, },
		{ -0.00000213, 0.00000136, -0.00000094, 0.00000362, 0.00000292, -0.00001353, 0.00000535, -0.00000096, 0.00000057, 0.00000634, -0.00000466, 0.00000050, -0.00000387, 0.00000468, -0.00000230, 0.00000367, -0.00000030, -0.00000579, -0.00000004, 0.00000184, -0.00000184, 0.00000482, 0.00000118, -0.00000758, 0.00000035, 0.00000326, -0.00000276, 0.00000334, 0.00000362, -0.00000923, 0.00000700, -0.00000286, 0.00000597, -0.00001222, 0.00000473, -0.00000523, -0.00000400, 0.00001890, -0.00000713, -0.00000370, -0.00000351, 0.00000531, -0.00000391, -0.00000037, -0.00000536, 0.00000165, -0.00001687, 0.00003782, -0.00002210, -0.00000238, 0.00000147, -0.00001317, 0.00002171, -0.00003242, 0.00002470, -0.00002299, 0.00002271, -0.00003686, 0.00000570, -0.00001088, -0.00001250, -0.00005432, -0.00001017, -0.00055050, 0.00137717, -0.00055050, -0.00001017, -0.00005432, -0.00001250, -0.00001088, 0.00000570, -0.00003686, 0.00002271, -0.00002299, 0.00002470, -0.00003242, 0.00002171, -0.00001317, 0.00000147, -0.00000238, -0.00002210, 0.00003782, -0.00001687, 0.00000165, -0.00000536, -0.00000037, -0.00000391, 0.00000531, -0.00000351, -0.00000370, -0.00000713, 0.00001890, -0.00000400, -0.00000523, 0.00000473, -0.00001222, 0.00000597, -0.00000286, 0.00000700, -0.00000923, 0.00000362, 0.00000334, -0.00000276, 0.00000326, 0.00000035, -0.00000758, 0.00000118, 0.00000482, -0.00000184, 0.00000184, -0.00000004, -0.00000579, -0.00000030, 0.00000367, -0.00000230, 0.00000468, -0.00000387, 0.00000050, -0.00000466, 0.00000634, 0.00000057, -0.00000096, 0.00000535, -0.00001353, 0.00000292, 0.00000362, -0.00000094, 0.00000136, },
		{ -0.00000216, 0.00000158, -0.00000123, 0.00000477, 0.00000073, -0.00001258, 0.00000625, -0.00000172, 0.00000123, 0.00000626, -0.00000605, 0.00000193, -0.00000499, 0.00000637, -0.00000443, 0.00000599, -0.00000300, -0.00000335, -0.00000202, 0.00000442, -0.00000501, 0.00000915, -0.00000433, -0.00000328, -0.00000263, 0.00000679, -0.00000705, 0.00000866, -0.00000247, -0.00000374, 0.00000213, 0.00000194, 0.00000110, -0.00000778, 0.00000054, -0.00000216, -0.00000601, 0.00002340, -0.00001469, 0.00000327, -0.00000925, 0.00001225, -0.00001234, 0.00000838, -0.00001410, 0.00001059, -0.00002614, 0.00005106, -0.00004031, 0.00001906, -0.00002146, 0.00001023, -0.00000156, -0.00001025, 0.00000217, 0.00000057, -0.00000409, -0.00000859, -0.00002712, 0.00002964, -0.00006631, 0.00001577, -0.00012285, -0.00040723, 0.00135503, -0.00066182, 0.00008893, -0.00011835, 0.00003764, -0.00004896, 0.00003646, -0.00006312, 0.00004766, -0.00004475, 0.00004534, -0.00005263, 0.00004317, -0.00003526, 0.00002394, -0.00002439, -0.00000254, 0.00002299, -0.00000630, -0.00000818, 0.00000402, -0.00000969, 0.00000521, -0.00000234, 0.00000261, -0.00001083, 0.00000096, 0.00001365, -0.00000171, -0.00000799, 0.00000848, -0.00001623, 0.00001035, -0.00000715, 0.00001133, -0.00001420, 0.00000939, -0.00000191, 0.00000151, -0.00000018, 0.00000298, -0.00001139, 0.00000653, 0.00000026, 0.00000167, -0.00000105, 0.00000212, -0.00000831, 0.00000250, 0.00000123, -0.00000000, 0.00000279, -0.00000260, -0.00000106, -0.00000297, 0.00000607, 0.00000008, -0.00000016, 0.00000411, -0.00001394, 0.00000484, 0.00000247, -0.00000065, 0.00000119, },
		{ -0.00000223, 0.00000184, -0.00000148, 0.00000583, -0.00000165, -0.00001114, 0.00000674, -0.00000238, 0.00000201, 0.00000582, -0.00000707, 0.00000315, -0.00000589, 0.00000775, -0.00000627, 0.00000803, -0.00000544, -0.00000116, -0.00000363, 0.00000647, -0.00000757, 0.00001290, -0.00000961, 0.00000115, -0.00000566, 0.00001014, -0.00001103, 0.00001370, -0.00000847, 0.00000188, -0.00000294, 0.00000696, -0.00000396, -0.00000318, -0.00000383, 0.00000103, -0.00000760, 0.00002687, -0.00002132, 0.00000964, -0.00001423, 0.00001804, -0.00001954, 0.00001594, -0.00002154, 0.00001796, -0.00003329, 0.00006164, -0.00005585, 0.00003831, -0.00004310, 0.00003314, -0.00002491, 0.00001220, -0.00002054, 0.00002415, -0.00003079, 0.00001950, -0.00005948, 0.00006958, -0.00012007, 0.00008715, -0.00024321, -0.00023722, 0.00128980, -0.00073795, 0.00016981, -0.00017234, 0.00008084, -0.00008190, 0.00006291, -0.00008541, 0.00006898, -0.00006310, 0.00006260, -0.00006937, 0.00006126, -0.00005439, 0.00004419, -0.00004533, 0.00001691, 0.00000772, 0.00000465, -0.00001811, 0.00001334, -0.00001892, 0.00001439, -0.00001016, 0.00000865, -0.00001768, 0.00000915, 0.00000794, 0.00000068, -0.00001030, 0.00001160, -0.00001957, 0.00001401, -0.00001065, 0.00001482, -0.00001832, 0.00001448, -0.00000673, 0.00000547, -0.00000329, 0.00000503, -0.00001440, 0.00001132, -0.00000415, 0.00000520, -0.00000401, 0.00000428, -0.00001074, 0.00000524, -0.00000119, 0.00000231, 0.00000082, -0.00000131, -0.00000263, -0.00000108, 0.00000546, -0.00000019, 0.00000058, 0.00000260, -0.00001382, 0.00000643, 0.00000137, -0.00000035, 0.00000105, },
		{ -0.00000233, 0.00000211, -0.00000167, 0.00000676, -0.00000409, -0.00000926, 0.00000681, -0.00000286, 0.00000284, 0.00000505, -0.00000768, 0.00000408, -0.00000649, 0.00000874, -0.00000770, 0.00000966, -0.00000748, 0.00000065, -0.00000475, 0.00000783, -0.00000928, 0.00001573, -0.00001425, 0.00000536, -0.00000845, 0.00001302, -0.00001441, 0.00001810, -0.00001400, 0.00000725, -0.00000784, 0.00001184, -0.00000892, 0.00000131, -0.00000815, 0.00000418, -0.00000865, 0.00002914, -0.00002672, 0.00001508, -0.00001813, 0.00002234, -0.00002508, 0.00002184, -0.00002718, 0.00002321, -0.00003768, 0.00006876, -0.00006759, 0.00005393, -0.00006177, 0.00005382, -0.00004659, 0.00003322, -0.00004171, 0.00004590, -0.00005540, 0.00004518, -0.00008883, 0.00010584, -0.00016988, 0.00015470, -0.00036448, -0.00004738, 0.00118489, -0.00077776, 0.00022936, -0.00021326, 0.00011456, -0.00010755, 0.00008320, -0.00010213, 0.00008525, -0.00007675, 0.00007526, -0.00008146, 0.00007472, -0.00006915, 0.00006071, -0.00006363, 0.00003485, -0.00000678, 0.00001501, -0.00002730, 0.00002186, -0.00002736, 0.00002300, -0.00001763, 0.00001417, -0.00002378, 0.00001696, 0.00000211, 0.00000299, -0.00001201, 0.00001391, -0.00002208, 0.00001677, -0.00001318, 0.00001725, -0.00002134, 0.00001860, -0.00001083, 0.00000886, -0.00000585, 0.00000630, -0.00001633, 0.00001521, -0.00000805, 0.00000846, -0.00000679, 0.00000624, -0.00001289, 0.00000774, -0.00000341, 0.00000449, -0.00000109, -0.00000009, -0.00000408, 0.00000091, 0.00000457, -0.00000022, 0.00000121, 0.00000091, -0.00001321, 0.00000765, 0.00000035, -0.00000007, 0.00000094, },
		{ -0.00000244, 0.00000238, -0.00000177, 0.00000746, -0.00000649, -0.00000703, 0.00000647, -0.00000311, 0.00000366, 0.00000400, -0.00000787, 0.00000469, -0.00000676, 0.00000929, -0.00000866, 0.00001082, -0.00000903, 0.00000200, -0.00000531, 0.00000840, -0.00000999, 0.00001740, -0.00001793, 0.00000900, -0.00001074, 0.00001520, -0.00001692, 0.00002152, -0.00001869, 0.00001197, -0.00001221, 0.00001625, -0.00001345, 0.00000539, -0.00001215, 0.00000711, -0.00000909, 0.00003009, -0.00003065, 0.00001935, -0.00002075, 0.00002493, -0.00002869, 0.00002575, -0.00003071, 0.00002603, -0.00003896, 0.00007187, -0.00007475, 0.00006481, -0.00007607, 0.00007073, -0.00006503, 0.00005124, -0.00005972, 0.00006414, -0.00007608, 0.00006642, -0.00011278, 0.00013551, -0.00021201, 0.00021329, -0.00047938, 0.00015407, 0.00104579, -0.00078219, 0.00026607, -0.00023925, 0.00013709, -0.00012444, 0.00009611, -0.00011221, 0.00009551, -0.00008484, 0.00008255, -0.00008813, 0.00008268, -0.00007853, 0.00007226, -0.00007798, 0.00005000, -0.00001941, 0.00002385, -0.00003501, 0.00002887, -0.00003435, 0.00003042, -0.00002422, 0.00001873, -0.00002870, 0.00002393, -0.00000350, 0.00000506, -0.00001302, 0.00001531, -0.00002364, 0.00001852, -0.00001463, 0.00001851, -0.00002308, 0.00002154, -0.00001399, 0.00001148, -0.00000772, 0.00000668, -0.00001703, 0.00001792, -0.00001113, 0.00001115, -0.00000917, 0.00000780, -0.00001460, 0.00000985, -0.00000528, 0.00000638, -0.00000282, 0.00000092, -0.00000530, 0.00000287, 0.00000348, -0.00000002, 0.00000165, -0.00000086, -0.00001217, 0.00000849, -0.00000055, 0.00000019, 0.00000085, },
		{ -0.00000253, 0.00000262, -0.00000174, 0.00000788, -0.00000872, -0.00000456, 0.00000575, -0.00000311, 0.00000439, 0.00000272, -0.00000766, 0.00000495, -0.00000667, 0.00000939, -0.00000912, 0.00001147, -0.00001007, 0.00000285, -0.00000527, 0.00000817, -0.00000964, 0.00001776, -0.00002039, 0.00001180, -0.00001230, 0.00001648, -0.00001834, 0.00002373, -0.00002221, 0.00001572, -0.00001571, 0.00001986, -0.00001727, 0.00000878, -0.00001560, 0.00000969, -0.00000890, 0.00002971, -0.00003300, 0.00002228, -0.00002199, 0.00002574, -0.00003024, 0.00002754, -0.00003198, 0.00002633, -0.00003704, 0.00007080, -0.00007691, 0.00007022, -0.00008499, 0.00008264, -0.00007892, 0.00006494, -0.00007325, 0.00007747, -0.00009129, 0.00008148, -0.00012935, 0.00015610, -0.00024316, 0.00025817, -0.00058071, 0.00035813, 0.00087960, -0.00075413, 0.00028008, -0.00024969, 0.00014770, -0.00013185, 0.00010102, -0.00011512, 0.00009931, -0.00008699, 0.00008413, -0.00008906, 0.00008476, -0.00008196, 0.00007806, -0.00008740, 0.00006136, -0.00002927, 0.00003042, -0.00004057, 0.00003381, -0.00003934, 0.00003613, -0.00002949, 0.00002196, -0.00003205, 0.00002967, -0.00000855, 0.00000672, -0.00001326, 0.00001574, -0.00002420, 0.00001923, -0.00001497, 0.00001854, -0.00002348, 0.00002317, -0.00001608, 0.00001323, -0.00000883, 0.00000615, -0.00001643, 0.00001929, -0.00001318, 0.00001306, -0.00001095, 0.00000883, -0.00001575, 0.00001142, -0.00000667, 0.00000785, -0.00000426, 0.00000165, -0.00000617, 0.00000466, 0.00000228, 0.00000039, 0.00000185, -0.00000261, -0.00001077, 0.00000896, -0.00000132, 0.00000045, 0.00000075, },
		{ -0.00000257, 0.00000281, -0.00000157, 0.00000794, -0.00001064, -0.00000201, 0.00000471, -0.00000285, 0.00000498, 0.00000131, -0.00000709, 0.00000489, -0.00000625, 0.00000905, -0.00000909, 0.00001160, -0.00001058, 0.00000322, -0.00000468, 0.00000718, -0.00000825, 0.00001676, -0.00002148, 0.00001358, -0.00001297, 0.00001674, -0.00001856, 0.00002455, -0.00002436, 0.00001824, -0.00001808, 0.00002243, -0.00002013, 0.00001126, -0.00001827, 0.00001179, -0.00000808, 0.00002804, -0.00003376, 0.00002386, -0.00002189, 0.00002483, -0.00002979, 0.00002723, -0.00003109, 0.00002424, -0.00003212, 0.00006572, -0.00007409, 0.00006991, -0.00008795, 0.00008876, -0.00008736, 0.00007336, -0.00008135, 0.00008490, -0.00009996, 0.00008919, -0.00013713, 0.00016581, -0.00026080, 0.00028538, -0.00066175, 0.00055553, 0.00069465, -0.00069809, 0.00027299, -0.00024513, 0.00014661, -0.00012988, 0.00009801, -0.00011095, 0.00009675, -0.00008333, 0.00008020, -0.00008444, 0.00008109, -0.00007943, 0.00007782, -0.00009134, 0.00006827, -0.00003569, 0.00003416, -0.00004353, 0.00003627, -0.00004193, 0.00003973, -0.00003309, 0.00002362, -0.00003359, 0.00003384, -0.00001278, 0.00000784, -0.00001271, 0.00001520, -0.00002379, 0.00001895, -0.00001426, 0.00001742, -0.00002256, 0.00002349, -0.00001704, 0.00001406, -0.00000921, 0.00000477, -0.00001456, 0.00001927, -0.00001406, 0.00001404, -0.00001201, 0.00000921, -0.00001623, 0.00001239, -0.00000750, 0.00000883, -0.00000531, 0.00000201, -0.00000660, 0.00000617, 0.00000108, 0.00000096, 0.00000180, -0.00000425, -0.00000911, 0.00000908, -0.00000196, 0.00000070, 0.00000062, },
		{ -0.00000254, 0.00000292, -0.00000124, 0.00000762, -0.00001215, 0.00000051, 0.00000346, -0.00000235, 0.00000537, -0.00000017, -0.00000621, 0.00000453, -0.00000555, 0.00000832, -0.00000861, 0.00001126, -0.00001061, 0.00000317, -0.00000365, 0.00000557, -0.00000596, 0.00001446, -0.00002116, 0.00001424, -0.00001270, 0.00001596, -0.00001754, 0.00002391, -0.00002500, 0.00001936, -0.00001913, 0.00002377, -0.00002185, 0.00001265, -0.00002001, 0.00001332, -0.00000669, 0.00002522, -0.00003301, 0.00002414, -0.00002057, 0.00002239, -0.00002754, 0.00002505, -0.00002827, 0.00002013, -0.00002467, 0.00005714, -0.00006672, 0.00006410, -0.00008490, 0.00008880, -0.00008990, 0.00007605, -0.00008354, 0.00008593, -0.00010157, 0.00008894, -0.00013541, 0.00016364, -0.00026338, 0.00029209, -0.00071682, 0.00073738, 0.00049988, -0.00061983, 0.00024770, -0.00022721, 0.00013489, -0.00011934, 0.00008777, -0.00010034, 0.00008843, -0.00007444, 0.00007136, -0.00007490, 0.00007226, -0.00007138, 0.00007173, -0.00008972, 0.00007044, -0.00003835, 0.00003477, -0.00004365, 0.00003605, -0.00004191, 0.00004098, -0.00003481, 0.00002356, -0.00003315, 0.00003621, -0.00001596, 0.00000834, -0.00001139, 0.00001375, -0.00002247, 0.00001778, -0.00001264, 0.00001528, -0.00002045, 0.00002255, -0.00001692, 0.00001402, -0.00000892, 0.00000269, -0.00001158, 0.00001790, -0.00001374, 0.00001402, -0.00001227, 0.00000891, -0.00001602, 0.00001270, -0.00000773, 0.00000925, -0.00000593, 0.00000197, -0.00000654, 0.00000728, -0.00000002, 0.00000162, 0.00000148, -0.00000568, -0.00000728, 0.00000890, -0.00000246, 0.00000095, 0.00000045, },
		{ -0.00000242, 0.00000294, -0.00000076, 0.00000689, -0.00001314, 0.00000284, 0.00000208, -0.00000164, 0.00000554, -0.00000163, -0.00000511, 0.00000393, -0.00000461, 0.00000729, -0.00000776, 0.00001052, -0.00001025, 0.00000279, -0.00000230, 0.00000352, -0.00000298, 0.00001104, -0.00001949, 0.00001377, -0.00001151, 0.00001420, -0.00001536, 0.00002188, -0.00002413, 0.00001904, -0.00001880, 0.00002377, -0.00002233, 0.00001286, -0.00002071, 0.00001423, -0.00000483, 0.00002144, -0.00003091, 0.00002327, -0.00001827, 0.00001873, -0.00002380, 0.00002132, -0.00002393, 0.00001451, -0.00001537, 0.00004585, -0.00005561, 0.00005347, -0.00007627, 0.00008296, -0.00008661, 0.00007301, -0.00007986, 0.00008064, -0.00009618, 0.00008081, -0.00012427, 0.00014955, -0.00025049, 0.00027685, -0.00074164, 0.00089564, 0.00030433, -0.00052588, 0.00020803, -0.00019840, 0.00011436, -0.00010166, 0.00007156, -0.00008438, 0.00007535, -0.00006132, 0.00005860, -0.00006147, 0.00005928, -0.00005873, 0.00006047, -0.00008291, 0.00006800, -0.00003724, 0.00003224, -0.00004093, 0.00003315, -0.00003928, 0.00003983, -0.00003462, 0.00002180, -0.00003074, 0.00003667, -0.00001794, 0.00000818, -0.00000937, 0.00001148, -0.00002038, 0.00001589, -0.00001030, 0.00001234, -0.00001735, 0.00002053, -0.00001582, 0.00001320, -0.00000812, 0.00000013, -0.00000772, 0.00001535, -0.00001229, 0.00001301, -0.00001175, 0.00000793, -0.00001513, 0.00001239, -0.00000736, 0.00000910, -0.00000611, 0.00000152, -0.00000594, 0.00000792, -0.00000092, 0.00000230, 0.00000091, -0.00000685, -0.00000538, 0.00000846, -0.00000284, 0.00000121, 0.00000021, },
		{ -0.00000219, 0.00000287, -0.00000015, 0.00000576, -0.00001356, 0.00000486, 0.00000069, -0.00000079, 0.00000548, -0.00000301, -0.00000385, 0.00000315, -0.00000352, 0.00000602, -0.00000664, 0.00000949, -0.00000959, 0.00000222, -0.00000082, 0.00000126, 0.00000043, 0.00000676, -0.00001666, 0.00001230, -0.00000949, 0.00001159, -0.00001220, 0.00001860, -0.00002187, 0.00001732, -0.00001710, 0.00002245, -0.00002157, 0.00001185, -0.00002030, 0.00001448, -0.00000260, 0.00001692, -0.00002769, 0.00002146, -0.00001526, 0.00001422, -0.00001899, 0.00001649, -0.00001857, 0.00000801, -0.00000504, 0.00003285, -0.00004185, 0.00003908, -0.00006296, 0.00007194, -0.00007804, 0.00006475, -0.00007083, 0.00006959, -0.00008443, 0.00006556, -0.00010456, 0.00012443, -0.00022291, 0.00023973, -0.00073367, 0.00102366, 0.00011652, -0.00042307, 0.00015840, -0.00016174, 0.00008735, -0.00007873, 0.00005097, -0.00006453, 0.00005885, -0.00004523, 0.00004319, -0.00004544, 0.00004346, -0.00004275, 0.00004512, -0.00007167, 0.00006145, -0.00003271, 0.00002684, -0.00003564, 0.00002782, -0.00003423, 0.00003640, -0.00003261, 0.00001849, -0.00002650, 0.00003524, -0.00001869, 0.00000739, -0.00000677, 0.00000855, -0.00001768, 0.00001347, -0.00000746, 0.00000885, -0.00001353, 0.00001764, -0.00001394, 0.00001177, -0.00000698, -0.00000265, -0.00000328, 0.00001185, -0.00000988, 0.00001113, -0.00001050, 0.00000635, -0.00001365, 0.00001151, -0.00000645, 0.00000844, -0.00000588, 0.00000071, -0.00000484, 0.00000806, -0.00000153, 0.00000292, 0.00000014, -0.00000771, -0.00000349, 0.00000781, -0.00000311, 0.00000148, -0.00000010, },
		{ -0.00000187, 0.00000273, 0.00000056, 0.00000427, -0.00001336, 0.00000647, -0.00000061, 0.00000015, 0.00000520, -0.00000424, -0.00000252, 0.00000225, -0.00000234, 0.00000462, -0.00000533, 0.00000827, -0.00000875, 0.00000157, 0.00000063, -0.00000097, 0.00000395, 0.00000196, -0.00001293, 0.00000999, -0.00000683, 0.00000837, -0.00000831, 0.00001433, -0.00001842, 0.00001440, -0.00001418, 0.00001990, -0.00001965, 0.00000970, -0.00001882, 0.00001412, -0.00000016, 0.00001193, -0.00002363, 0.00001896, -0.00001185, 0.00000926, -0.00001357, 0.00001104, -0.00001273, 0.00000131, 0.00000544, 0.00001924, -0.00002672, 0.00002225, -0.00004619, 0.00005682, -0.00006515, 0.00005220, -0.00005743, 0.00005386, -0.00006752, 0.00004451, -0.00007785, 0.00009008, -0.00018261, 0.00018238, -0.00069229, 0.00111655, -0.00005596, -0.00031808, 0.00010346, -0.00012060, 0.00005648, -0.00005268, 0.00002787, -0.00004244, 0.00004043, -0.00002757, 0.00002652, -0.00002823, 0.00002628, -0.00002495, 0.00002706, -0.00005711, 0.00005162, -0.00002537, 0.00001909, -0.00002824, 0.00002051, -0.00002715, 0.00003100, -0.00002904, 0.00001391, -0.00002070, 0.00003206, -0.00001824, 0.00000604, -0.00000371, 0.00000513, -0.00001454, 0.00001073, -0.00000438, 0.00000511, -0.00000927, 0.00001416, -0.00001148, 0.00000990, -0.00000570, -0.00000536, 0.00000139, 0.00000774, -0.00000672, 0.00000854, -0.00000867, 0.00000430, -0.00001169, 0.00001019, -0.00000511, 0.00000733, -0.00000531, -0.00000039, -0.00000329, 0.00000767, -0.00000179, 0.00000340, -0.00000079, -0.00000824, -0.00000170, 0.00000701, -0.00000326, 0.00000175, -0.00000048, },
		{ -0.00000145, 0.00000253, 0.00000130, 0.00000250, -0.00001257, 0.00000759, -0.00000171, 0.00000108, 0.00000472, -0.00000529, -0.00000117, 0.00000130, -0.00000115, 0.00000317, -0.00000394, 0.00000696, -0.00000785, 0.00000098, 0.00000187, -0.00000293, 0.00000725, -0.00000301, -0.00000861, 0.00000710, -0.00000378, 0.00000479, -0.00000398, 0.00000940, -0.00001410, 0.00001053, -0.00001027, 0.00001634, -0.00001675, 0.00000654, -0.00001633, 0.00001317, 0.00000237, 0.00000674, -0.00001899, 0.00001602, -0.00000836, 0.00000425, -0.00000798, 0.00000546, -0.00000695, -0.00000496, 0.00001521, 0.00000614, -0.00001156, 0.00000447, -0.00002746, 0.00003898, -0.00004923, 0.00003661, -0.00004098, 0.00003488, -0.00004702, 0.00001948, -0.00004627, 0.00004901, -0.00013255, 0.00010790, -0.00061887, 0.00117142, -0.00020694, -0.00021694, 0.00004767, -0.00007832, 0.00002439, -0.00002572, 0.00000415, -0.00001979, 0.00002162, -0.00000977, 0.00000999, -0.00001129, 0.00000926, -0.00000690, 0.00000780, -0.00004055, 0.00003956, -0.00001605, 0.00000968, -0.00001937, 0.00001181, -0.00001858, 0.00002406, -0.00002427, 0.00000846, -0.00001373, 0.00002740, -0.00001672, 0.00000423, -0.00000039, 0.00000141, -0.00001118, 0.00000788, -0.00000130, 0.00000139, -0.00000489, 0.00001035, -0.00000868, 0.00000780, -0.00000449, -0.00000774, 0.00000593, 0.00000334, -0.00000312, 0.00000545, -0.00000642, 0.00000192, -0.00000940, 0.00000855, -0.00000347, 0.00000590, -0.00000449, -0.00000168, -0.00000138, 0.00000681, -0.00000169, 0.00000366, -0.00000180, -0.00000843, -0.00000005, 0.00000611, -0.00000332, 0.00000201, -0.00000090, },
		{ -0.00000097, 0.00000229, 0.00000203, 0.00000053, -0.00001124, 0.00000819, -0.00000254, 0.00000195, 0.00000407, -0.00000613, 0.00000014, 0.00000035, -0.00000000, 0.00000175, -0.00000256, 0.00000565, -0.00000697, 0.00000054, 0.00000278, -0.00000441, 0.00001004, -0.00000777, -0.00000406, 0.00000391, -0.00000058, 0.00000116, 0.00000042, 0.00000419, -0.00000927, 0.00000605, -0.00000566, 0.00001202, -0.00001310, 0.00000258, -0.00001298, 0.00001174, 0.00000484, 0.00000158, -0.00001407, 0.00001291, -0.00000507, -0.00000044, -0.00000266, 0.00000017, -0.00000168, -0.00001024, 0.00002354, -0.00000546, 0.00000235, -0.00001276, -0.00000832, 0.00001995, -0.00003173, 0.00001943, -0.00002301, 0.00001431, -0.00002479, -0.00000741, -0.00001234, 0.00000428, -0.00007649, 0.00002071, -0.00051663, 0.00118751, -0.00033191, -0.00012477, -0.00000496, -0.00003798, -0.00000640, 0.00000009, -0.00001839, 0.00000182, 0.00000382, 0.00000686, -0.00000514, 0.00000408, -0.00000621, 0.00000988, -0.00001110, -0.00002338, 0.00002643, -0.00000573, -0.00000055, -0.00000978, 0.00000243, -0.00000916, 0.00001611, -0.00001877, 0.00000260, -0.00000607, 0.00002161, -0.00001431, 0.00000213, 0.00000301, -0.00000242, -0.00000777, 0.00000512, 0.00000157, -0.00000205, -0.00000067, 0.00000651, -0.00000578, 0.00000565, -0.00000352, -0.00000954, 0.00001003, -0.00000099, 0.00000064, 0.00000211, -0.00000395, -0.00000059, -0.00000696, 0.00000678, -0.00000169, 0.00000428, -0.00000355, -0.00000302, 0.00000075, 0.00000553, -0.00000120, 0.00000366, -0.00000283, -0.00000832, 0.00000140, 0.00000515, -0.00000328, 0.00000224, -0.00000136, },
		{ -0.00000044, 0.00000206, 0.00000266, -0.00000152, -0.00000944, 0.00000827, -0.00000306, 0.00000267, 0.00000331, -0.00000674, 0.00000134, -0.00000056, 0.00000104, 0.00000043, -0.00000125, 0.00000443, -0.00000619, 0.00000033, 0.00000325, -0.00000526, 0.00001207, -0.00001199, 0.00000038, 0.00000072, 0.00000248, -0.00000224, 0.00000455, -0.00000092, -0.00000432, 0.00000134, -0.00000072, 0.00000728, -0.00000902, -0.00000190, -0.00000897, 0.00000991, 0.00000712, -0.00000329, -0.00000912, 0.00000983, -0.00000220, -0.00000451, 0.00000207, -0.00000445, 0.00000271, -0.00001412, 0.00002985, -0.00001474, 0.00001390, -0.00002807, 0.00000969, 0.00000128, -0.00001419, 0.00000220, -0.00000511, -0.00000611, -0.00000276, -0.00003393, 0.00002125, -0.00004080, -0.00001867, -0.00007386, -0.00039046, 0.00116613, -0.00042819, -0.00004547, -0.00005118, -0.00000218, -0.00003378, 0.00002298, -0.00003823, 0.00002107, -0.00001176, 0.00002124, -0.00001784, 0.00001683, -0.00001898, 0.00002405, -0.00002822, -0.00000697, 0.00001343, 0.00000460, -0.00001074, -0.00000025, -0.00000691, 0.00000042, 0.00000776, -0.00001301, -0.00000317, 0.00000176, 0.00001510, -0.00001127, -0.00000008, 0.00000631, -0.00000616, -0.00000448, 0.00000261, 0.00000403, -0.00000499, 0.00000313, 0.00000287, -0.00000299, 0.00000361, -0.00000290, -0.00001059, 0.00001339, -0.00000493, 0.00000425, -0.00000124, -0.00000146, -0.00000307, -0.00000454, 0.00000503, 0.00000008, 0.00000262, -0.00000261, -0.00000428, 0.00000296, 0.00000395, -0.00000038, 0.00000337, -0.00000381, -0.00000793, 0.00000262, 0.00000416, -0.00000315, 0.00000243, -0.00000182, },
		{ 0.00000009, 0.00000187, 0.00000312, -0.00000352, -0.00000729, 0.00000787, -0.00000323, 0.00000321, 0.00000247, -0.00000713, 0.00000242, -0.00000138, 0.00000196, -0.00000075, -0.00000006, 0.00000334, -0.00000555, 0.00000038, 0.00000324, -0.00000540, 0.00001316, -0.00001537, 0.00000439, -0.00000221, 0.00000516, -0.00000512, 0.00000811, -0.00000555, 0.00000035, -0.00000322, 0.00000418, 0.00000247, -0.00000483, -0.00000659, -0.00000451, 0.00000782, 0.00000910, -0.00000769, -0.00000436, 0.00000695, 0.00000007, -0.00000776, 0.00000595, -0.00000818, 0.00000602, -0.00001638, 0.00003382, -0.00002116, 0.00002226, -0.00004032, 0.00002523, -0.00001561, 0.00000199, -0.00001367, 0.00001120, -0.00002471, 0.00001722, -0.00005794, 0.00005189, -0.00008291, 0.00003651, -0.00016987, -0.00024657, 0.00111050, -0.00049492, 0.00001841, -0.00008861, 0.00002715, -0.00005612, 0.00004160, -0.00005421, 0.00003693, -0.00002425, 0.00003262, -0.00002740, 0.00002623, -0.00002820, 0.00003460, -0.00004236, 0.00000749, 0.00000165, 0.00001400, -0.00002006, 0.00000847, -0.00001549, 0.00000950, -0.00000043, -0.00000748, -0.00000838, 0.00000924, 0.00000832, -0.00000788, -0.00000221, 0.00000931, -0.00000965, -0.00000146, 0.00000048, 0.00000597, -0.00000728, 0.00000633, -0.00000037, -0.00000047, 0.00000179, -0.00000272, -0.00001076, 0.00001580, -0.00000822, 0.00000744, -0.00000435, 0.00000086, -0.00000535, -0.00000229, 0.00000345, 0.00000168, 0.00000104, -0.00000179, -0.00000531, 0.00000509, 0.00000220, 0.00000073, 0.00000277, -0.00000468, -0.00000731, 0.00000360, 0.00000319, -0.00000294, 0.00000256, -0.00000225, },
		{ 0.00000060, 0.00000174, 0.00000334, -0.00000533, -0.00000494, 0.00000706, -0.00000308, 0.00000352, 0.00000161, -0.00000730, 0.00000335, -0.00000211, 0.00000273, -0.00000177, 0.00000097, 0.00000240, -0.00000506, 0.00000068, 0.00000278, -0.00000480, 0.00001322, -0.00001771, 0.00000771, -0.00000464, 0.00000727, -0.00000729, 0.00001083, -0.00000940, 0.00000440, -0.00000727, 0.00000869, -0.00000207, -0.00000087, -0.00001117, 0.00000013, 0.00000559, 0.00001068, -0.00001147, 0.00000004, 0.00000440, 0.00000167, -0.00001005, 0.00000883, -0.00001088, 0.00000815, -0.00001697, 0.00003534, -0.00002447, 0.00002693, -0.00004870, 0.00003727, -0.00002955, 0.00001558, -0.00002693, 0.00002462, -0.00004009, 0.00003355, -0.00007758, 0.00007728, -0.00011905, 0.00008487, -0.00026124, -0.00009206, 0.00102545, -0.00053301, 0.00006569, -0.00011587, 0.00004879, -0.00007242, 0.00005508, -0.00006560, 0.00004879, -0.00003314, 0.00004057, -0.00003347, 0.00003192, -0.00003347, 0.00004090, -0.00005268, 0.00001904, -0.00000801, 0.00002165, -0.00002781, 0.00001574, -0.00002269, 0.00001746, -0.00000793, -0.00000261, -0.00001258, 0.00001587, 0.00000171, -0.00000442, -0.00000408, 0.00001185, -0.00001273, 0.00000119, -0.00000120, 0.00000732, -0.00000882, 0.00000878, -0.00000304, 0.00000165, 0.00000028, -0.00000298, -0.00001003, 0.00001714, -0.00001064, 0.00001002, -0.00000704, 0.00000285, -0.00000729, -0.00000033, 0.00000217, 0.00000298, -0.00000032, -0.00000119, -0.00000600, 0.00000698, 0.00000041, 0.00000202, 0.00000188, -0.00000537, -0.00000651, 0.00000434, 0.00000225, -0.00000265, 0.00000261, -0.00000262, },
		{ 0.00000106, 0.00000171, 0.00000328, -0.00000684, -0.00000253, 0.00000596, -0.00000264, 0.00000359, 0.00000077, -0.00000727, 0.00000412, -0.00000272, 0.00000334, -0.00000263, 0.00000185, 0.00000161, -0.00000471, 0.00000117, 0.00000192, -0.00000354, 0.00001223, -0.00001890, 0.00001018, -0.00000642, 0.00000866, -0.00000861, 0.00001253, -0.00001223, 0.00000754, -0.00001049, 0.00001248, -0.00000602, 0.00000255, -0.00001534, 0.00000471, 0.00000335, 0.00001181, -0.00001452, 0.00000395, 0.00000224, 0.00000257, -0.00001137, 0.00001068, -0.00001254, 0.00000917, -0.00001605, 0.00003459, -0.00002475, 0.00002779, -0.00005281, 0.00004512, -0.00003970, 0.00002571, -0.00003667, 0.00003420, -0.00005117, 0.00004503, -0.00009143, 0.00009562, -0.00014679, 0.00012276, -0.00034217, 0.00006559, 0.00091698, -0.00054483, 0.00009645, -0.00013255, 0.00006230, -0.00008224, 0.00006310, -0.00007211, 0.00005642, -0.00003828, 0.00004505, -0.00003607, 0.00003396, -0.00003476, 0.00004276, -0.00005872, 0.00002709, -0.00001491, 0.00002699, -0.00003346, 0.00002107, -0.00002804, 0.00002383, -0.00001430, 0.00000125, -0.00001545, 0.00002125, -0.00000432, -0.00000116, -0.00000554, 0.00001380, -0.00001531, 0.00000339, -0.00000239, 0.00000804, -0.00000957, 0.00001043, -0.00000507, 0.00000329, -0.00000092, -0.00000364, -0.00000845, 0.00001737, -0.00001210, 0.00001182, -0.00000918, 0.00000439, -0.00000879, 0.00000124, 0.00000127, 0.00000390, -0.00000139, -0.00000090, -0.00000626, 0.00000851, -0.00000127, 0.00000340, 0.00000074, -0.00000585, -0.00000559, 0.00000484, 0.00000137, -0.00000229, 0.00000256, -0.00000291, },
		{ 0.00000145, 0.00000178, 0.00000291, -0.00000794, -0.00000022, 0.00000467, -0.00000197, 0.00000344, -0.00000002, -0.00000707, 0.00000472, -0.00000322, 0.00000380, -0.00000332, 0.00000259, 0.00000095, -0.00000445, 0.00000180, 0.00000080, -0.00000175, 0.00001031, -0.00001893, 0.00001170, -0.00000747, 0.00000930, -0.00000902, 0.00001312, -0.00001388, 0.00000959, -0.00001268, 0.00001531, -0.00000912, 0.00000517, -0.00001881, 0.00000899, 0.00000123, 0.00001245, -0.00001678, 0.00000729, 0.00000047, 0.00000282, -0.00001177, 0.00001156, -0.00001325, 0.00000926, -0.00001391, 0.00003194, -0.00002236, 0.00002507, -0.00005265, 0.00004852, -0.00004563, 0.00003187, -0.00004238, 0.00003936, -0.00005734, 0.00005094, -0.00009864, 0.00010575, -0.00016442, 0.00014740, -0.00040759, 0.00021897, 0.00079187, -0.00053398, 0.00011193, -0.00013911, 0.00006794, -0.00008576, 0.00006580, -0.00007390, 0.00005997, -0.00003988, 0.00004634, -0.00003555, 0.00003272, -0.00003246, 0.00004038, -0.00006044, 0.00003138, -0.00001873, 0.00002966, -0.00003671, 0.00002418, -0.00003124, 0.00002828, -0.00001925, 0.00000390, -0.00001678, 0.00002507, -0.00000944, 0.00000167, -0.00000645, 0.00001506, -0.00001729, 0.00000513, -0.00000310, 0.00000820, -0.00000959, 0.00001126, -0.00000643, 0.00000442, -0.00000182, -0.00000459, -0.00000616, 0.00001655, -0.00001257, 0.00001278, -0.00001067, 0.00000544, -0.00000982, 0.00000238, 0.00000079, 0.00000438, -0.00000211, -0.00000095, -0.00000603, 0.00000955, -0.00000271, 0.00000475, -0.00000059, -0.00000611, -0.00000460, 0.00000512, 0.00000057, -0.00000187, 0.00000242, -0.00000309, },
		{ 0.00000176, 0.00000194, 0.00000221, -0.00000855, 0.00000185, 0.00000332, -0.00000116, 0.00000309, -0.00000075, -0.00000671, 0.00000515, -0.00000361, 0.00000413, -0.00000387, 0.00000322, 0.00000038, -0.00000422, 0.00000244, -0.00000045, 0.00000039, 0.00000760, -0.00001788, 0.00001228, -0.00000777, 0.00000918, -0.00000857, 0.00001264, -0.00001433, 0.00001045, -0.00001369, 0.00001702, -0.00001121, 0.00000679, -0.00002134, 0.00001274, -0.00000068, 0.00001259, -0.00001823, 0.00001002, -0.00000092, 0.00000254, -0.00001139, 0.00001159, -0.00001318, 0.00000870, -0.00001098, 0.00002794, -0.00001792, 0.00001935, -0.00004859, 0.00004762, -0.00004731, 0.00003395, -0.00004395, 0.00004001, -0.00005845, 0.00005111, -0.00009895, 0.00010725, -0.00017117, 0.00015706, -0.00045346, 0.00036124, 0.00065714, -0.00050485, 0.00011420, -0.00013675, 0.00006656, -0.00008363, 0.00006375, -0.00007147, 0.00005993, -0.00003843, 0.00004496, -0.00003253, 0.00002890, -0.00002727, 0.00003436, -0.00005817, 0.00003203, -0.00001944, 0.00002960, -0.00003748, 0.00002499, -0.00003219, 0.00003067, -0.00002262, 0.00000525, -0.00001650, 0.00002716, -0.00001342, 0.00000391, -0.00000673, 0.00001559, -0.00001866, 0.00000639, -0.00000339, 0.00000787, -0.00000895, 0.00001134, -0.00000715, 0.00000508, -0.00000248, -0.00000570, -0.00000333, 0.00001483, -0.00001209, 0.00001291, -0.00001150, 0.00000597, -0.00001037, 0.00000311, 0.00000071, 0.00000442, -0.00000247, -0.00000133, -0.00000529, 0.00001005, -0.00000380, 0.00000596, -0.00000205, -0.00000613, -0.00000360, 0.00000520, -0.00000014, -0.00000142, 0.00000218, -0.00000317, },
		{ 0.00000201, 0.00000217, 0.00000122, -0.00000863, 0.00000357, 0.00000203, -0.00000030, 0.00000258, -0.00000140, -0.00000621, 0.00000541, -0.00000389, 0.00000434, -0.00000429, 0.00000375, -0.00000014, -0.00000398, 0.00000301, -0.00000166, 0.00000266, 0.00000436, -0.00001591, 0.00001200, -0.00000740, 0.00000843, -0.00000737, 0.00001119, -0.00001365, 0.00001015, -0.00001351, 0.00001756, -0.00001220, 0.00000729, -0.00002279, 0.00001579, -0.00000227, 0.00001225, -0.00001889, 0.00001215, -0.00000200, 0.00000185, -0.00001038, 0.00001095, -0.00001254, 0.00000777, -0.00000773, 0.00002325, -0.00001221, 0.00001144, -0.00004135, 0.00004297, -0.00004511, 0.00003225, -0.00004165, 0.00003645, -0.00005482, 0.00004588, -0.00009276, 0.00010048, -0.00016718, 0.00015122, -0.00047708, 0.00048652, 0.00051964, -0.00046227, 0.00010595, -0.00012719, 0.00005944, -0.00007689, 0.00005784, -0.00006560, 0.00005701, -0.00003461, 0.00004165, -0.00002780, 0.00002335, -0.00002015, 0.00002560, -0.00005259, 0.00002948, -0.00001732, 0.00002702, -0.00003592, 0.00002365, -0.00003099, 0.00003101, -0.00002441, 0.00000536, -0.00001467, 0.00002746, -0.00001610, 0.00000544, -0.00000637, 0.00001537, -0.00001940, 0.00000722, -0.00000333, 0.00000716, -0.00000780, 0.00001078, -0.00000729, 0.00000531, -0.00000297, -0.00000680, -0.00000021, 0.00001240, -0.00001081, 0.00001226, -0.00001171, 0.00000603, -0.00001049, 0.00000347, 0.00000100, 0.00000407, -0.00000250, -0.00000201, -0.00000407, 0.00000997, -0.00000446, 0.00000693, -0.00000355, -0.00000592, -0.00000262, 0.00000509, -0.00000074, -0.00000096, 0.00000186, -0.00000314, },
		{ 0.00000219, 0.00000242, -0.00000002, -0.00000817, 0.00000485, 0.00000092, 0.00000052, 0.00000197, -0.00000197, -0.00000560, 0.00000551, -0.00000406, 0.00000443, -0.00000461, 0.00000423, -0.00000066, -0.00000364, 0.00000340, -0.00000268, 0.00000482, 0.00000086, -0.00001326, 0.00001100, -0.00000649, 0.00000718, -0.00000561, 0.00000896, -0.00001199, 0.00000881, -0.00001223, 0.00001697, -0.00001211, 0.00000667, -0.00002306, 0.00001801, -0.00000349, 0.00001146, -0.00001882, 0.00001369, -0.00000283, 0.00000092, -0.00000894, 0.00000985, -0.00001155, 0.00000680, -0.00000463, 0.00001854, -0.00000610, 0.00000233, -0.00003189, 0.00003537, -0.00003972, 0.00002738, -0.00003614, 0.00002936, -0.00004721, 0.00003609, -0.00008101, 0.00008646, -0.00015351, 0.00013059, -0.00047724, 0.00059025, 0.00038559, -0.00041111, 0.00009017, -0.00011246, 0.00004815, -0.00006680, 0.00004914, -0.00005724, 0.00005204, -0.00002921, 0.00003721, -0.00002222, 0.00001703, -0.00001215, 0.00001517, -0.00004463, 0.00002442, -0.00001290, 0.00002231, -0.00003240, 0.00002046, -0.00002791, 0.00002950, -0.00002476, 0.00000441, -0.00001151, 0.00002608, -0.00001747, 0.00000623, -0.00000539, 0.00001444, -0.00001954, 0.00000767, -0.00000302, 0.00000623, -0.00000628, 0.00000971, -0.00000696, 0.00000518, -0.00000337, -0.00000773, 0.00000298, 0.00000950, -0.00000891, 0.00001098, -0.00001138, 0.00000568, -0.00001026, 0.00000355, 0.00000155, 0.00000339, -0.00000226, -0.00000291, -0.00000246, 0.00000934, -0.00000463, 0.00000758, -0.00000501, -0.00000550, -0.00000171, 0.00000484, -0.00000122, -0.00000050, 0.00000148, -0.00000301, },
		{ 0.00000233, 0.00000266, -0.00000142, -0.00000721, 0.00000565, 0.00000007, 0.00000122, 0.00000132, -0.00000246, -0.00000488, 0.00000545, -0.00000412, 0.00000444, -0.00000484, 0.00000467, -0.00000119, -0.00000319, 0.00000354, -0.00000335, 0.00000665, -0.00000263, -0.00001019, 0.00000950, -0.00000521, 0.00000564, -0.00000353, 0.00000622, -0.00000962, 0.00000664, -0.00001002, 0.00001538, -0.00001103, 0.00000499, -0.00002215, 0.00001934, -0.00000431, 0.00001028, -0.00001808, 0.00001470, -0.00000349, -0.00000013, -0.00000726, 0.00000847, -0.00001042, 0.00000604, -0.00000211, 0.00001444, -0.00000043, -0.00000695, -0.00002130, 0.00002585, -0.00003206, 0.00002023, -0.00002831, 0.00001971, -0.00003667, 0.00002293, -0.00006508, 0.00006681, -0.00013203, 0.00009700, -0.00045424, 0.00066936, 0.00026029, -0.00035594, 0.00006984, -0.00009467, 0.00003435, -0.00005471, 0.00003882, -0.00004737, 0.00004589, -0.00002304, 0.00003243, -0.00001663, 0.00001087, -0.00000436, 0.00000424, -0.00003532, 0.00001773, -0.00000687, 0.00001607, -0.00002741, 0.00001590, -0.00002337, 0.00002645, -0.00002393, 0.00000267, -0.00000732, 0.00002324, -0.00001759, 0.00000634, -0.00000386, 0.00001289, -0.00001916, 0.00000780, -0.00000255, 0.00000520, -0.00000458, 0.00000830, -0.00000628, 0.00000479, -0.00000377, -0.00000835, 0.00000598, 0.00000640, -0.00000659, 0.00000920, -0.00001063, 0.00000504, -0.00000978, 0.00000345, 0.00000226, 0.00000249, -0.00000185, -0.00000393, -0.00000055, 0.00000821, -0.00000432, 0.00000784, -0.00000635, -0.00000492, -0.00000089, 0.00000447, -0.00000157, -0.00000008, 0.00000106, -0.00000281, },
		{ 0.00000246, 0.00000281, -0.00000288, -0.00000582, 0.00000596, -0.00000048, 0.00000172, 0.00000067, -0.00000290, -0.00000408, 0.00000524, -0.00000409, 0.00000435, -0.00000498, 0.00000508, -0.00000175, -0.00000259, 0.00000340, -0.00000360, 0.00000796, -0.00000583, -0.00000697, 0.00000771, -0.00000376, 0.00000401, -0.00000138, 0.00000325, -0.00000681, 0.00000391, -0.00000712, 0.00001302, -0.00000916, 0.00000240, -0.00002014, 0.00001976, -0.00000472, 0.00000879, -0.00001677, 0.00001522, -0.00000404, -0.00000114, -0.00000550, 0.00000698, -0.00000929, 0.00000567, -0.00000045, 0.00001146, 0.00000403, -0.00001540, -0.00001071, 0.00001554, -0.00002318, 0.00001181, -0.00001922, 0.00000867, -0.00002447, 0.00000783, -0.00004667, 0.00004356, -0.00010521, 0.00005330, -0.00040985, 0.00072238, 0.00014785, -0.00030076, 0.00004772, -0.00007584, 0.00001963, -0.00004191, 0.00002797, -0.00003694, 0.00003938, -0.00001680, 0.00002798, -0.00001174, 0.00000566, 0.00000227, -0.00000609, -0.00002576, 0.00001033, -0.00000003, 0.00000897, -0.00002155, 0.00001051, -0.00001786, 0.00002227, -0.00002222, 0.00000051, -0.00000251, 0.00001925, -0.00001663, 0.00000586, -0.00000190, 0.00001082, -0.00001833, 0.00000769, -0.00000201, 0.00000420, -0.00000285, 0.00000670, -0.00000537, 0.00000421, -0.00000422, -0.00000854, 0.00000858, 0.00000334, -0.00000409, 0.00000713, -0.00000958, 0.00000420, -0.00000915, 0.00000329, 0.00000301, 0.00000149, -0.00000137, -0.00000493, 0.00000150, 0.00000670, -0.00000357, 0.00000769, -0.00000751, -0.00000420, -0.00000019, 0.00000400, -0.00000179, 0.00000029, 0.00000064, -0.00000255, },
		{ 0.00000262, 0.00000283, -0.00000430, -0.00000411, 0.00000581, -0.00000070, 0.00000199, 0.00000010, -0.00000329, -0.00000321, 0.00000489, -0.00000394, 0.00000416, -0.00000504, 0.00000546, -0.00000234, -0.00000186, 0.00000295, -0.00000336, 0.00000862, -0.00000851, -0.00000386, 0.00000584, -0.00000232, 0.00000251, 0.00000058, 0.00000035, -0.00000388, 0.00000094, -0.00000383, 0.00001014, -0.00000675, -0.00000086, -0.00001719, 0.00001932, -0.00000475, 0.00000707, -0.00001499, 0.00001530, -0.00000453, -0.00000203, -0.00000380, 0.00000550, -0.00000826, 0.00000578, 0.00000015, 0.00000993, 0.00000671, -0.00002217, -0.00000115, 0.00000552, -0.00001414, 0.00000317, -0.00000997, -0.00000258, -0.00001193, -0.00000769, -0.00002759, 0.00001894, -0.00007588, 0.00000306, -0.00034710, 0.00074941, 0.00005107, -0.00024878, 0.00002617, -0.00005770, 0.00000537, -0.00002955, 0.00001755, -0.00002675, 0.00003316, -0.00001105, 0.00002439, -0.00000808, 0.00000198, 0.00000696, -0.00001486, -0.00001693, 0.00000316, 0.00000682, 0.00000170, -0.00001545, 0.00000487, -0.00001194, 0.00001743, -0.00002002, -0.00000172, 0.00000249, 0.00001448, -0.00001481, 0.00000495, 0.00000032, 0.00000837, -0.00001715, 0.00000741, -0.00000150, 0.00000336, -0.00000124, 0.00000506, -0.00000436, 0.00000353, -0.00000476, -0.00000823, 0.00001063, 0.00000054, -0.00000162, 0.00000492, -0.00000838, 0.00000329, -0.00000848, 0.00000317, 0.00000367, 0.00000050, -0.00000094, -0.00000581, 0.00000356, 0.00000493, -0.00000243, 0.00000713, -0.00000843, -0.00000338, 0.00000039, 0.00000347, -0.00000189, 0.00000057, 0.00000024, -0.00000228, },
		{ 0.00000281, 0.00000264, -0.00000556, -0.00000220, 0.00000527, -0.00000060, 0.00000199, -0.00000038, -0.00000363, -0.00000228, 0.00000440, -0.00000370, 0.00000389, -0.00000501, 0.00000579, -0.00000295, -0.00000102, 0.00000223, -0.00000264, 0.00000855, -0.00001050, -0.00000109, 0.00000410, -0.00000106, 0.00000132, 0.00000214, -0.00000219, -0.00000114, -0.00000197, -0.00000047, 0.00000706, -0.00000407, -0.00000453, -0.00001351, 0.00001812, -0.00000444, 0.00000522, -0.00001285, 0.00001499, -0.00000497, -0.00000274, -0.00000225, 0.00000411, -0.00000734, 0.00000634, -0.00000029, 0.00000999, 0.00000726, -0.00002665, 0.00000654, -0.00000326, -0.00000592, -0.00000472, -0.00000159, -0.00001291, -0.00000033, -0.00002218, -0.00000959, -0.00000485, -0.00004700, -0.00004972, -0.00027003, 0.00075200, -0.00002855, -0.00020234, 0.00000697, -0.00004158, -0.00000734, -0.00001851, 0.00000829, -0.00001741, 0.00002771, -0.00000618, 0.00002194, -0.00000592, 0.00000015, 0.00000925, -0.00002137, -0.00000965, -0.00000299, 0.00001295, -0.00000508, -0.00000969, -0.00000046, -0.00000612, 0.00001237, -0.00001767, -0.00000365, 0.00000726, 0.00000935, -0.00001241, 0.00000378, 0.00000264, 0.00000569, -0.00001571, 0.00000703, -0.00000106, 0.00000276, 0.00000013, 0.00000351, -0.00000334, 0.00000278, -0.00000539, -0.00000741, 0.00001201, -0.00000182, 0.00000063, 0.00000274, -0.00000714, 0.00000239, -0.00000786, 0.00000318, 0.00000414, -0.00000037, -0.00000063, -0.00000643, 0.00000547, 0.00000304, -0.00000101, 0.00000618, -0.00000908, -0.00000251, 0.00000083, 0.00000290, -0.00000188, 0.00000077, -0.00000011, -0.00000202, },
		{ 0.00000307, 0.00000222, -0.00000654, -0.00000023, 0.00000444, -0.00000023, 0.00000173, -0.00000074, -0.00000394, -0.00000132, 0.00000380, -0.00000335, 0.00000352, -0.00000487, 0.00000605, -0.00000352, -0.00000013, 0.00000132, -0.00000151, 0.00000777, -0.00001169, 0.00000119, 0.00000264, -0.00000011, 0.00000055, 0.00000312, -0.00000417, 0.00000115, -0.00000451, 0.00000268, 0.00000406, -0.00000144, -0.00000830, -0.00000934, 0.00001630, -0.00000386, 0.00000332, -0.00001045, 0.00001434, -0.00000539, -0.00000324, -0.00000089, 0.00000283, -0.00000650, 0.00000723, -0.00000164, 0.00001157, 0.00000560, -0.00002849, 0.00001177, -0.00001004, 0.00000071, -0.00001106, 0.00000508, -0.00002137, 0.00000929, -0.00003442, 0.00000581, -0.00002583, -0.00002134, -0.00010092, -0.00018335, 0.00073288, -0.00009078, -0.00016284, -0.00000872, -0.00002838, -0.00001778, -0.00000937, 0.00000070, -0.00000931, 0.00002331, -0.00000232, 0.00002071, -0.00000531, 0.00000020, 0.00000894, -0.00002520, -0.00000450, -0.00000752, 0.00001778, -0.00001084, -0.00000475, -0.00000501, -0.00000088, 0.00000750, -0.00001548, -0.00000497, 0.00001139, 0.00000424, -0.00000970, 0.00000254, 0.00000489, 0.00000292, -0.00001411, 0.00000660, -0.00000074, 0.00000243, 0.00000117, 0.00000215, -0.00000241, 0.00000202, -0.00000607, -0.00000610, 0.00001267, -0.00000362, 0.00000253, 0.00000072, -0.00000596, 0.00000158, -0.00000736, 0.00000339, 0.00000435, -0.00000104, -0.00000053, -0.00000672, 0.00000710, 0.00000118, 0.00000057, 0.00000491, -0.00000943, -0.00000162, 0.00000114, 0.00000233, -0.00000176, 0.00000087, -0.00000039, -0.00000178, },
		{ 0.00000339, 0.00000154, -0.00000716, 0.00000165, 0.00000343, 0.00000034, 0.00000124, -0.00000097, -0.00000422, -0.00000036, 0.00000309, -0.00000291, 0.00000305, -0.00000461, 0.00000621, -0.00000402, 0.00000074, 0.00000031, -0.00000007, 0.00000634, -0.00001207, 0.00000287, 0.00000154, 0.00000047, 0.00000027, 0.00000344, -0.00000541, 0.00000278, -0.00000646, 0.00000534, 0.00000141, 0.00000088, -0.00001187, -0.00000496, 0.00001402, -0.00000310, 0.00000148, -0.00000789, 0.00001339, -0.00000576, -0.00000353, 0.00000028, 0.00000164, -0.00000563, 0.00000823, -0.00000362, 0.00001438, 0.00000189, -0.00002764, 0.00001423, -0.00001438, 0.00000521, -0.00001530, 0.00000945, -0.00002731, 0.00001615, -0.00004348, 0.00001746, -0.00004242, -0.00000127, -0.00014668, -0.00009206, 0.00069568, -0.00013647, -0.00013084, -0.00002030, -0.00001853, -0.00002559, -0.00000241, -0.00000501, -0.00000261, 0.00002001, 0.00000056, 0.00002057, -0.00000606, 0.00000191, 0.00000619, -0.00002625, -0.00000177, -0.00001004, 0.00002091, -0.00001520, -0.00000099, -0.00000844, 0.00000341, 0.00000315, -0.00001369, -0.00000547, 0.00001457, -0.00000047, -0.00000696, 0.00000141, 0.00000691, 0.00000022, -0.00001245, 0.00000617, -0.00000055, 0.00000241, 0.00000183, 0.00000105, -0.00000162, 0.00000125, -0.00000676, -0.00000440, 0.00001264, -0.00000482, 0.00000399, -0.00000105, -0.00000491, 0.00000090, -0.00000698, 0.00000381, 0.00000426, -0.00000147, -0.00000069, -0.00000660, 0.00000832, -0.00000049, 0.00000219, 0.00000338, -0.00000949, -0.00000075, 0.00000130, 0.00000179, -0.00000157, 0.00000087, -0.00000060, -0.00000158, },
		{ 0.00000376, 0.00000059, -0.00000734, 0.00000333, 0.00000236, 0.00000103, 0.00000057, -0.00000109, -0.00000445, 0.00000059, 0.00000231, -0.00000238, 0.00000248, -0.00000422, 0.00000623, -0.00000441, 0.00000151, -0.00000068, 0.00000153, 0.00000440, -0.00001168, 0.00000393, 0.00000086, 0.00000065, 0.00000049, 0.00000306, -0.00000586, 0.00000363, -0.00000766, 0.00000732, -0.00000066, 0.00000263, -0.00001497, -0.00000064, 0.00001147, -0.00000224, -0.00000022, -0.00000529, 0.00001219, -0.00000606, -0.00000365, 0.00000127, 0.00000050, -0.00000461, 0.00000908, -0.00000584, 0.00001799, -0.00000346, -0.00002434, 0.00001394, -0.00001610, 0.00000732, -0.00001717, 0.00001121, -0.00003035, 0.00001982, -0.00004886, 0.00002465, -0.00005361, 0.00001149, -0.00018364, -0.00000111, 0.00064462, -0.00016730, -0.00010615, -0.00002777, -0.00001206, -0.00003075, 0.00000235, -0.00000884, 0.00000275, 0.00001771, 0.00000268, 0.00002122, -0.00000777, 0.00000481, 0.00000142, -0.00002473, -0.00000145, -0.00001043, 0.00002216, -0.00001796, 0.00000140, -0.00001054, 0.00000653, -0.00000045, -0.00001244, -0.00000502, 0.00001658, -0.00000450, -0.00000441, 0.00000054, 0.00000856, -0.00000230, -0.00001080, 0.00000576, -0.00000047, 0.00000267, 0.00000211, 0.00000024, -0.00000100, 0.00000049, -0.00000738, -0.00000243, 0.00001200, -0.00000541, 0.00000497, -0.00000252, -0.00000401, 0.00000036, -0.00000674, 0.00000444, 0.00000387, -0.00000162, -0.00000110, -0.00000606, 0.00000907, -0.00000187, 0.00000372, 0.00000168, -0.00000927, 0.00000007, 0.00000132, 0.00000130, -0.00000133, 0.00000079, -0.00000073, -0.00000141, },
		{ 0.00000415, -0.00000058, -0.00000706, 0.00000468, 0.00000135, 0.00000174, -0.00000021, -0.00000110, -0.00000463, 0.00000149, 0.00000150, -0.00000179, 0.00000183, -0.00000369, 0.00000611, -0.00000463, 0.00000211, -0.00000153, 0.00000313, 0.00000211, -0.00001063, 0.00000440, 0.00000056, 0.00000050, 0.00000113, 0.00000206, -0.00000552, 0.00000366, -0.00000801, 0.00000850, -0.00000201, 0.00000366, -0.00001737, 0.00000336, 0.00000882, -0.00000137, -0.00000169, -0.00000273, 0.00001077, -0.00000627, -0.00000364, 0.00000212, -0.00000065, -0.00000331, 0.00000951, -0.00000788, 0.00002186, -0.00000987, -0.00001908, 0.00001119, -0.00001533, 0.00000711, -0.00001669, 0.00001038, -0.00003047, 0.00002024, -0.00005046, 0.00002720, -0.00005900, 0.00001595, -0.00020922, 0.00008497, 0.00058408, -0.00018556, -0.00008800, -0.00003153, -0.00000863, -0.00003351, 0.00000515, -0.00001098, 0.00000695, 0.00001615, 0.00000434, 0.00002224, -0.00000997, 0.00000828, -0.00000474, -0.00002112, -0.00000327, -0.00000882, 0.00002157, -0.00001912, 0.00000239, -0.00001126, 0.00000836, -0.00000321, -0.00001176, -0.00000364, 0.00001733, -0.00000766, -0.00000222, 0.00000003, 0.00000974, -0.00000454, -0.00000921, 0.00000537, -0.00000047, 0.00000316, 0.00000203, -0.00000026, -0.00000058, -0.00000027, -0.00000786, -0.00000032, 0.00001087, -0.00000545, 0.00000547, -0.00000367, -0.00000327, -0.00000006, -0.00000658, 0.00000523, 0.00000320, -0.00000154, -0.00000174, -0.00000512, 0.00000931, -0.00000287, 0.00000506, -0.00000008, -0.00000880, 0.00000082, 0.00000121, 0.00000089, -0.00000107, 0.00000065, -0.00000078, -0.00000128, },
		{ 0.00000453, -0.00000193, -0.00000631, 0.00000564, 0.00000050, 0.00000237, -0.00000104, -0.00000105, -0.00000473, 0.00000231, 0.00000067, -0.00000117, 0.00000111, -0.00000302, 0.00000581, -0.00000466, 0.00000249, -0.00000214, 0.00000454, -0.00000031, -0.00000909, 0.00000439, 0.00000057, 0.00000010, 0.00000207, 0.00000055, -0.00000449, 0.00000292, -0.00000753, 0.00000885, -0.00000258, 0.00000385, -0.00001890, 0.00000684, 0.00000625, -0.00000058, -0.00000287, -0.00000032, 0.00000919, -0.00000636, -0.00000355, 0.00000287, -0.00000184, -0.00000164, 0.00000928, -0.00000930, 0.00002540, -0.00001664, -0.00001252, 0.00000652, -0.00001245, 0.00000485, -0.00001414, 0.00000724, -0.00002797, 0.00001772, -0.00004859, 0.00002543, -0.00005882, 0.00001197, -0.00022181, 0.00016235, 0.00051839, -0.00019385, -0.00007524, -0.00003231, -0.00000766, -0.00003435, 0.00000638, -0.00001176, 0.00001029, 0.00001503, 0.00000594, 0.00002318, -0.00001211, 0.00001163, -0.00001150, -0.00001608, -0.00000675, -0.00000552, 0.00001935, -0.00001883, 0.00000214, -0.00001072, 0.00000896, -0.00000512, -0.00001159, -0.00000142, 0.00001684, -0.00000985, -0.00000050, -0.00000006, 0.00001041, -0.00000643, -0.00000774, 0.00000500, -0.00000050, 0.00000381, 0.00000164, -0.00000049, -0.00000036, -0.00000104, -0.00000810, 0.00000177, 0.00000940, -0.00000505, 0.00000554, -0.00000453, -0.00000267, -0.00000040, -0.00000646, 0.00000610, 0.00000232, -0.00000126, -0.00000255, -0.00000383, 0.00000905, -0.00000342, 0.00000610, -0.00000183, -0.00000811, 0.00000149, 0.00000100, 0.00000058, -0.00000083, 0.00000047, -0.00000079, -0.00000116, },
		{ 0.00000484, -0.00000339, -0.00000513, 0.00000617, -0.00000011, 0.00000284, -0.00000185, -0.00000097, -0.00000474, 0.00000301, -0.00000012, -0.00000054, 0.00000035, -0.00000225, 0.00000535, -0.00000448, 0.00000260, -0.00000241, 0.00000562, -0.00000265, -0.00000726, 0.00000401, 0.00000080, -0.00000043, 0.00000316, -0.00000129, -0.00000293, 0.00000155, -0.00000631, 0.00000841, -0.00000239, 0.00000318, -0.00001948, 0.00000963, 0.00000389, 0.00000006, -0.00000369, 0.00000188, 0.00000750, -0.00000632, -0.00000342, 0.00000354, -0.00000310, 0.00000048, 0.00000820, -0.00000976, 0.00002804, -0.00002306, -0.00000543, 0.00000063, -0.00000803, 0.00000106, -0.00001003, 0.00000231, -0.00002338, 0.00001282, -0.00004392, 0.00002007, -0.00005391, 0.00000022, -0.00022079, 0.00022807, 0.00045143, -0.00019484, -0.00006653, -0.00003102, -0.00000845, -0.00003382, 0.00000650, -0.00001156, 0.00001309, 0.00001404, 0.00000781, 0.00002361, -0.00001368, 0.00001419, -0.00001805, -0.00001040, -0.00001126, -0.00000105, 0.00001587, -0.00001739, 0.00000089, -0.00000914, 0.00000846, -0.00000628, -0.00001178, 0.00000142, 0.00001523, -0.00001106, 0.00000073, 0.00000027, 0.00001054, -0.00000794, -0.00000640, 0.00000464, -0.00000051, 0.00000454, 0.00000102, -0.00000048, -0.00000032, -0.00000182, -0.00000807, 0.00000370, 0.00000774, -0.00000432, 0.00000527, -0.00000514, -0.00000216, -0.00000071, -0.00000631, 0.00000697, 0.00000131, -0.00000086, -0.00000345, -0.00000228, 0.00000835, -0.00000353, 0.00000680, -0.00000348, -0.00000726, 0.00000204, 0.00000068, 0.00000039, -0.00000062, 0.00000029, -0.00000078, -0.00000103, },
		{ 0.00000503, -0.00000486, -0.00000358, 0.00000626, -0.00000044, 0.00000309, -0.00000256, -0.00000090, -0.00000463, 0.00000356, -0.00000084, 0.00000006, -0.00000042, -0.00000138, 0.00000475, -0.00000412, 0.00000244, -0.00000230, 0.00000624, -0.00000472, -0.00000534, 0.00000343, 0.00000111, -0.00000094, 0.00000420, -0.00000323, -0.00000106, -0.00000026, -0.00000451, 0.00000732, -0.00000153, 0.00000173, -0.00001910, 0.00001163, 0.00000186, 0.00000050, -0.00000413, 0.00000377, 0.00000575, -0.00000615, -0.00000328, 0.00000415, -0.00000440, 0.00000301, 0.00000617, -0.00000898, 0.00002933, -0.00002843, 0.00000138, -0.00000570, -0.00000279, -0.00000362, -0.00000500, -0.00000372, -0.00001743, 0.00000635, -0.00003735, 0.00001223, -0.00004556, -0.00001789, -0.00020660, 0.00028024, 0.00038649, -0.00019102, -0.00006048, -0.00002858, -0.00001024, -0.00003252, 0.00000601, -0.00001077, 0.00001566, 0.00001291, 0.00001024, 0.00002319, -0.00001427, 0.00001539, -0.00002361, -0.00000486, -0.00001609, 0.00000403, 0.00001162, -0.00001519, -0.00000101, -0.00000682, 0.00000710, -0.00000685, -0.00001212, 0.00000461, 0.00001272, -0.00001138, 0.00000150, 0.00000096, 0.00001018, -0.00000906, -0.00000520, 0.00000427, -0.00000044, 0.00000527, 0.00000024, -0.00000028, -0.00000044, -0.00000260, -0.00000771, 0.00000533, 0.00000607, -0.00000341, 0.00000475, -0.00000556, -0.00000170, -0.00000106, -0.00000603, 0.00000774, 0.00000028, -0.00000043, -0.00000433, -0.00000060, 0.00000728, -0.00000321, 0.00000711, -0.00000495, -0.00000627, 0.00000248, 0.00000029, 0.00000031, -0.00000049, 0.00000014, -0.00000076, -0.00000085, },
		{ 0.00000504, -0.00000622, -0.00000178, 0.00000594, -0.00000048, 0.00000309, -0.00000312, -0.00000086, -0.00000441, 0.00000395, -0.00000145, 0.00000059, -0.00000114, -0.00000048, 0.00000402, -0.00000358, 0.00000202, -0.00000180, 0.00000632, -0.00000635, -0.00000350, 0.00000278, 0.00000139, -0.00000130, 0.00000503, -0.00000506, 0.00000088, -0.00000229, -0.00000234, 0.00000576, -0.00000016, -0.00000038, -0.00001781, 0.00001282, 0.00000022, 0.00000073, -0.00000416, 0.00000531, 0.00000402, -0.00000587, -0.00000313, 0.00000469, -0.00000569, 0.00000590, 0.00000322, -0.00000683, 0.00002892, -0.00003221, 0.00000720, -0.00001171, 0.00000257, -0.00000852, 0.00000024, -0.00001010, -0.00001093, -0.00000080, -0.00002995, 0.00000315, -0.00003539, -0.00004035, -0.00018061, 0.00031801, 0.00032611, -0.00018455, -0.00005588, -0.00002584, -0.00001236, -0.00003100, 0.00000534, -0.00000971, 0.00001822, 0.00001146, 0.00001342, 0.00002171, -0.00001366, 0.00001483, -0.00002757, -0.00000019, -0.00002058, 0.00000913, 0.00000709, -0.00001266, -0.00000316, -0.00000413, 0.00000517, -0.00000705, -0.00001238, 0.00000784, 0.00000959, -0.00001097, 0.00000189, 0.00000192, 0.00000939, -0.00000982, -0.00000414, 0.00000388, -0.00000024, 0.00000589, -0.00000061, 0.00000005, -0.00000071, -0.00000337, -0.00000701, 0.00000658, 0.00000452, -0.00000244, 0.00000407, -0.00000584, -0.00000125, -0.00000148, -0.00000558, 0.00000833, -0.00000068, -0.00000006, -0.00000509, 0.00000109, 0.00000596, -0.00000253, 0.00000703, -0.00000621, -0.00000519, 0.00000279, -0.00000014, 0.00000035, -0.00000043, 0.00000003, -0.00000077, -0.00000061, },
		{ 0.00000483, -0.00000739, 0.00000016, 0.00000529, -0.00000024, 0.00000281, -0.00000351, -0.00000088, -0.00000405, 0.00000415, -0.00000193, 0.00000101, -0.00000179, 0.00000042, 0.00000324, -0.00000293, 0.00000140, -0.00000094, 0.00000583, -0.00000744, -0.00000191, 0.00000221, 0.00000153, -0.00000140, 0.00000548, -0.00000656, 0.00000267, -0.00000429, -0.00000005, 0.00000394, 0.00000152, -0.00000295, -0.00001576, 0.00001321, -0.00000100, 0.00000074, -0.00000379, 0.00000645, 0.00000237, -0.00000549, -0.00000296, 0.00000511, -0.00000688, 0.00000897, -0.00000052, -0.00000335, 0.00002667, -0.00003401, 0.00001143, -0.00001672, 0.00000736, -0.00001296, 0.00000504, -0.00001610, -0.00000468, -0.00000775, -0.00002275, -0.00000583, -0.00002516, -0.00006480, -0.00014496, 0.00034153, 0.00027199, -0.00017713, -0.00005172, -0.00002348, -0.00001426, -0.00002968, 0.00000483, -0.00000860, 0.00002090, 0.00000963, 0.00001737, 0.00001913, -0.00001179, 0.00001231, -0.00002949, 0.00000305, -0.00002411, 0.00001368, 0.00000275, -0.00001020, -0.00000519, -0.00000144, 0.00000296, -0.00000708, -0.00001233, 0.00001080, 0.00000616, -0.00001002, 0.00000204, 0.00000302, 0.00000828, -0.00001027, -0.00000319, 0.00000344, 0.00000013, 0.00000634, -0.00000144, 0.00000045, -0.00000111, -0.00000411, -0.00000600, 0.00000740, 0.00000319, -0.00000152, 0.00000331, -0.00000604, -0.00000078, -0.00000201, -0.00000491, 0.00000865, -0.00000149, 0.00000018, -0.00000564, 0.00000267, 0.00000453, -0.00000158, 0.00000658, -0.00000720, -0.00000407, 0.00000295, -0.00000059, 0.00000048, -0.00000047, -0.00000002, -0.00000083, -0.00000028, },
		{ 0.00000436, -0.00000826, 0.00000212, 0.00000439, 0.00000023, 0.00000227, -0.00000370, -0.00000097, -0.00000359, 0.00000417, -0.00000225, 0.00000130, -0.00000231, 0.00000127, 0.00000244, -0.00000223, 0.00000065, 0.00000020, 0.00000481, -0.00000792, -0.00000067, 0.00000182, 0.00000145, -0.00000115, 0.00000544, -0.00000757, 0.00000408, -0.00000603, 0.00000214, 0.00000210, 0.00000326, -0.00000575, -0.00001313, 0.00001290, -0.00000181, 0.00000056, -0.00000306, 0.00000715, 0.00000087, -0.00000508, -0.00000275, 0.00000538, -0.00000783, 0.00001203, -0.00000480, 0.00000128, 0.00002263, -0.00003364, 0.00001368, -0.00002020, 0.00001102, -0.00001640, 0.00000879, -0.00002110, 0.00000063, -0.00001371, -0.00001669, -0.00001351, -0.00001658, -0.00008870, -0.00010236, 0.00035186, 0.00022503, -0.00016997, -0.00004730, -0.00002198, -0.00001557, -0.00002884, 0.00000468, -0.00000757, 0.00002374, 0.00000748, 0.00002200, 0.00001555, -0.00000879, 0.00000791, -0.00002917, 0.00000446, -0.00002625, 0.00001724, -0.00000099, -0.00000817, -0.00000676, 0.00000094, 0.00000075, -0.00000712, -0.00001180, 0.00001320, 0.00000275, -0.00000876, 0.00000209, 0.00000413, 0.00000697, -0.00001046, -0.00000235, 0.00000295, 0.00000071, 0.00000655, -0.00000217, 0.00000085, -0.00000161, -0.00000476, -0.00000474, 0.00000777, 0.00000216, -0.00000074, 0.00000256, -0.00000618, -0.00000027, -0.00000266, -0.00000400, 0.00000866, -0.00000209, 0.00000024, -0.00000589, 0.00000400, 0.00000311, -0.00000045, 0.00000583, -0.00000793, -0.00000292, 0.00000297, -0.00000102, 0.00000069, -0.00000059, 0.00000002, -0.00000095, 0.00000015, },
		{ 0.00000361, -0.00000875, 0.00000396, 0.00000336, 0.00000085, 0.00000152, -0.00000371, -0.00000112, -0.00000301, 0.00000402, -0.00000241, 0.00000143, -0.00000267, 0.00000200, 0.00000169, -0.00000154, -0.00000012, 0.00000152, 0.00000334, -0.00000781, 0.00000015, 0.00000165, 0.00000113, -0.00000053, 0.00000485, -0.00000798, 0.00000495, -0.00000730, 0.00000400, 0.00000047, 0.00000484, -0.00000854, -0.00001014, 0.00001201, -0.00000226, 0.00000023, -0.00000201, 0.00000741, -0.00000045, -0.00000466, -0.00000248, 0.00000543, -0.00000842, 0.00001481, -0.00000932, 0.00000675, 0.00001704, -0.00003117, 0.00001378, -0.00002182, 0.00001316, -0.00001845, 0.00001110, -0.00002461, 0.00000447, -0.00001809, -0.00001250, -0.00001892, -0.00001113, -0.00010959, -0.00005583, 0.00035075, 0.00018541, -0.00016374, -0.00004225, -0.00002156, -0.00001611, -0.00002862, 0.00000500, -0.00000663, 0.00002662, 0.00000519, 0.00002708, 0.00001121, -0.00000499, 0.00000189, -0.00002668, 0.00000391, -0.00002674, 0.00001952, -0.00000384, -0.00000682, -0.00000764, 0.00000273, -0.00000125, -0.00000731, -0.00001067, 0.00001482, -0.00000039, -0.00000740, 0.00000218, 0.00000510, 0.00000558, -0.00001047, -0.00000159, 0.00000242, 0.00000148, 0.00000648, -0.00000275, 0.00000121, -0.00000220, -0.00000528, -0.00000330, 0.00000772, 0.00000145, -0.00000014, 0.00000185, -0.00000630, 0.00000027, -0.00000340, -0.00000286, 0.00000835, -0.00000244, 0.00000007, -0.00000582, 0.00000501, 0.00000182, 0.00000073, 0.00000482, -0.00000839, -0.00000178, 0.00000284, -0.00000140, 0.00000094, -0.00000079, 0.00000012, -0.00000113, 0.00000067, },
		{ 0.00000261, -0.00000882, 0.00000557, 0.00000229, 0.00000155, 0.00000062, -0.00000355, -0.00000131, -0.00000237, 0.00000371, -0.00000242, 0.00000141, -0.00000285, 0.00000259, 0.00000104, -0.00000095, -0.00000082, 0.00000288, 0.00000154, -0.00000716, 0.00000055, 0.00000172, 0.00000057, 0.00000043, 0.00000371, -0.00000772, 0.00000519, -0.00000798, 0.00000537, -0.00000077, 0.00000606, -0.00001109, -0.00000701, 0.00001069, -0.00000242, -0.00000018, -0.00000072, 0.00000725, -0.00000152, -0.00000429, -0.00000210, 0.00000521, -0.00000851, 0.00001707, -0.00001368, 0.00001261, 0.00001029, -0.00002686, 0.00001179, -0.00002147, 0.00001361, -0.00001891, 0.00001173, -0.00002638, 0.00000655, -0.00002056, -0.00001062, -0.00002145, -0.00000988, -0.00012532, -0.00000844, 0.00034042, 0.00015268, -0.00015870, -0.00003650, -0.00002222, -0.00001588, -0.00002901, 0.00000575, -0.00000569, 0.00002941, 0.00000300, 0.00003230, 0.00000647, -0.00000082, -0.00000526, -0.00002232, 0.00000147, -0.00002552, 0.00002038, -0.00000565, -0.00000627, -0.00000768, 0.00000377, -0.00000287, -0.00000772, -0.00000889, 0.00001553, -0.00000301, -0.00000611, 0.00000243, 0.00000582, 0.00000422, -0.00001036, -0.00000089, 0.00000185, 0.00000245, 0.00000612, -0.00000313, 0.00000148, -0.00000286, -0.00000564, -0.00000176, 0.00000732, 0.00000106, 0.00000024, 0.00000120, -0.00000638, 0.00000081, -0.00000419, -0.00000153, 0.00000774, -0.00000254, -0.00000031, -0.00000540, 0.00000563, 0.00000077, 0.00000186, 0.00000366, -0.00000860, -0.00000068, 0.00000256, -0.00000168, 0.00000121, -0.00000104, 0.00000028, -0.00000135, 0.00000127, },
		{ 0.00000137, -0.00000842, 0.00000686, 0.00000130, 0.00000222, -0.00000037, -0.00000327, -0.00000153, -0.00000169, 0.00000327, -0.00000229, 0.00000124, -0.00000284, 0.00000299, 0.00000053, -0.00000050, -0.00000134, 0.00000414, -0.00000043, -0.00000608, 0.00000057, 0.00000200, -0.00000017, 0.00000167, 0.00000209, -0.00000682, 0.00000475, -0.00000798, 0.00000612, -0.00000149, 0.00000675, -0.00001318, -0.00000399, 0.00000912, -0.00000240, -0.00000058, 0.00000073, 0.00000670, -0.00000232, -0.00000399, -0.00000160, 0.00000469, -0.00000801, 0.00001855, -0.00001750, 0.00001839, 0.00000290, -0.00002114, 0.00000798, -0.00001927, 0.00001239, -0.00001776, 0.00001067, -0.00002637, 0.00000681, -0.00002104, -0.00001113, -0.00002086, -0.00001345, -0.00013422, 0.00003691, 0.00032334, 0.00012597, -0.00015473, -0.00003019, -0.00002378, -0.00001502, -0.00002987, 0.00000681, -0.00000460, 0.00003187, 0.00000122, 0.00003730, 0.00000170, 0.00000318, -0.00001294, -0.00001656, -0.00000256, -0.00002272, 0.00001985, -0.00000642, -0.00000655, -0.00000687, 0.00000400, -0.00000405, -0.00000833, -0.00000651, 0.00001529, -0.00000499, -0.00000502, 0.00000293, 0.00000620, 0.00000300, -0.00001020, -0.00000024, 0.00000126, 0.00000358, 0.00000547, -0.00000329, 0.00000163, -0.00000356, -0.00000578, -0.00000025, 0.00000666, 0.00000095, 0.00000042, 0.00000063, -0.00000642, 0.00000131, -0.00000498, -0.00000008, 0.00000687, -0.00000242, -0.00000089, -0.00000465, 0.00000585, 0.00000003, 0.00000285, 0.00000240, -0.00000857, 0.00000037, 0.00000214, -0.00000185, 0.00000145, -0.00000130, 0.00000049, -0.00000161, 0.00000191, },
		{ -0.00000003, -0.00000758, 0.00000776, 0.00000047, 0.00000279, -0.00000137, -0.00000291, -0.00000175, -0.00000101, 0.00000277, -0.00000206, 0.00000094, -0.00000264, 0.00000320, 0.00000020, -0.00000026, -0.00000162, 0.00000516, -0.00000240, -0.00000471, 0.00000031, 0.00000243, -0.00000100, 0.00000306, 0.00000011, -0.00000538, 0.00000370, -0.00000730, 0.00000623, -0.00000161, 0.00000681, -0.00001466, -0.00000127, 0.00000747, -0.00000227, -0.00000090, 0.00000223, 0.00000582, -0.00000285, -0.00000380, -0.00000098, 0.00000387, -0.00000686, 0.00001907, -0.00002044, 0.00002359, -0.00000453, -0.00001458, 0.00000280, -0.00001551, 0.00000970, -0.00001521, 0.00000811, -0.00002475, 0.00000540, -0.00001972, -0.00001383, -0.00001735, -0.00002185, -0.00013520, 0.00007775, 0.00030200, 0.00010413, -0.00015145, -0.00002368, -0.00002592, -0.00001378, -0.00003100, 0.00000802, -0.00000319, 0.00003378, 0.00000013, 0.00004172, -0.00000269, 0.00000650, -0.00002048, -0.00001002, -0.00000774, -0.00001865, 0.00001812, -0.00000626, -0.00000755, -0.00000528, 0.00000343, -0.00000479, -0.00000908, -0.00000367, 0.00001414, -0.00000627, -0.00000421, 0.00000370, 0.00000617, 0.00000199, -0.00001004, 0.00000038, 0.00000070, 0.00000482, 0.00000455, -0.00000321, 0.00000162, -0.00000428, -0.00000568, 0.00000116, 0.00000583, 0.00000105, 0.00000042, 0.00000012, -0.00000639, 0.00000173, -0.00000569, 0.00000140, 0.00000582, -0.00000214, -0.00000159, -0.00000364, 0.00000568, -0.00000037, 0.00000362, 0.00000114, -0.00000835, 0.00000135, 0.00000159, -0.00000188, 0.00000162, -0.00000155, 0.00000069, -0.00000186, 0.00000254, },
		{ -0.00000153, -0.00000635, 0.00000826, -0.00000013, 0.00000318, -0.00000230, -0.00000252, -0.00000192, -0.00000038, 0.00000223, -0.00000177, 0.00000054, -0.00000228, 0.00000323, 0.00000004, -0.00000023, -0.00000159, 0.00000583, -0.00000420, -0.00000321, -0.00000014, 0.00000291, -0.00000182, 0.00000446, -0.00000207, -0.00000353, 0.00000212, -0.00000602, 0.00000570, -0.00000114, 0.00000621, -0.00001542, 0.00000099, 0.00000588, -0.00000214, -0.00000106, 0.00000368, 0.00000469, -0.00000313, -0.00000370, -0.00000024, 0.00000277, -0.00000505, 0.00001850, -0.00002220, 0.00002777, -0.00001143, -0.00000782, -0.00000318, -0.00001065, 0.00000591, -0.00001159, 0.00000439, -0.00002186, 0.00000270, -0.00001704, -0.00001822, -0.00001152, -0.00003461, -0.00012791, 0.00011215, 0.00027871, 0.00008590, -0.00014837, -0.00001739, -0.00002828, -0.00001246, -0.00003217, 0.00000920, -0.00000130, 0.00003495, 0.00000001, 0.00004525, -0.00000639, 0.00000866, -0.00002720, -0.00000338, -0.00001352, -0.00001371, 0.00001548, -0.00000540, -0.00000906, -0.00000309, 0.00000218, -0.00000515, -0.00000982, -0.00000057, 0.00001224, -0.00000688, -0.00000367, 0.00000473, 0.00000574, 0.00000122, -0.00000990, 0.00000095, 0.00000019, 0.00000611, 0.00000340, -0.00000291, 0.00000144, -0.00000498, -0.00000533, 0.00000237, 0.00000495, 0.00000129, 0.00000029, -0.00000035, -0.00000626, 0.00000200, -0.00000624, 0.00000282, 0.00000468, -0.00000176, -0.00000235, -0.00000245, 0.00000517, -0.00000042, 0.00000412, -0.00000007, -0.00000795, 0.00000223, 0.00000092, -0.00000177, 0.00000170, -0.00000175, 0.00000088, -0.00000208, 0.00000312, },
		{ -0.00000303, -0.00000478, 0.00000834, -0.00000048, 0.00000335, -0.00000309, -0.00000214, -0.00000202, 0.00000016, 0.00000170, -0.00000146, 0.00000010, -0.00000180, 0.00000308, 0.00000005, -0.00000041, -0.00000123, 0.00000607, -0.00000569, -0.00000174, -0.00000064, 0.00000334, -0.00000250, 0.00000571, -0.00000426, -0.00000145, 0.00000019, -0.00000425, 0.00000465, -0.00000012, 0.00000496, -0.00001544, 0.00000268, 0.00000449, -0.00000208, -0.00000098, 0.00000498, 0.00000338, -0.00000317, -0.00000370, 0.00000057, 0.00000147, -0.00000265, 0.00001683, -0.00002261, 0.00003056, -0.00001726, -0.00000149, -0.00000935, -0.00000521, 0.00000147, -0.00000734, -0.00000004, -0.00001816, -0.00000079, -0.00001356, -0.00002357, -0.00000426, -0.00005071, -0.00011264, 0.00013884, 0.00025543, 0.00007009, -0.00014494, -0.00001174, -0.00003046, -0.00001135, -0.00003314, 0.00001021, 0.00000119, 0.00003526, 0.00000102, 0.00004766, -0.00000918, 0.00000933, -0.00003255, 0.00000272, -0.00001931, -0.00000839, 0.00001229, -0.00000412, -0.00001083, -0.00000053, 0.00000042, -0.00000526, -0.00001041, 0.00000256, 0.00000979, -0.00000690, -0.00000336, 0.00000595, 0.00000495, 0.00000072, -0.00000981, 0.00000147, -0.00000022, 0.00000738, 0.00000210, -0.00000240, 0.00000107, -0.00000564, -0.00000474, 0.00000333, 0.00000410, 0.00000159, 0.00000006, -0.00000080, -0.00000602, 0.00000210, -0.00000658, 0.00000408, 0.00000353, -0.00000137, -0.00000307, -0.00000118, 0.00000440, -0.00000016, 0.00000433, -0.00000117, -0.00000741, 0.00000300, 0.00000017, -0.00000150, 0.00000166, -0.00000187, 0.00000100, -0.00000223, 0.00000357, },
		{ -0.00000442, -0.00000299, 0.00000806, -0.00000058, 0.00000326, -0.00000371, -0.00000182, -0.00000203, 0.00000059, 0.00000124, -0.00000119, -0.00000033, -0.00000124, 0.00000280, 0.00000020, -0.00000077, -0.00000055, 0.00000583, -0.00000676, -0.00000043, -0.00000109, 0.00000365, -0.00000294, 0.00000666, -0.00000627, 0.00000063, -0.00000189, -0.00000217, 0.00000320, 0.00000132, 0.00000317, -0.00001475, 0.00000375, 0.00000336, -0.00000215, -0.00000061, 0.00000604, 0.00000199, -0.00000304, -0.00000374, 0.00000139, 0.00000007, 0.00000020, 0.00001412, -0.00002160, 0.00003176, -0.00002161, 0.00000385, -0.00001511, 0.00000024, -0.00000312, -0.00000294, -0.00000470, -0.00001418, -0.00000453, -0.00000996, -0.00002906, 0.00000334, -0.00006877, -0.00009033, 0.00015724, 0.00023369, 0.00005566, -0.00014071, -0.00000712, -0.00003214, -0.00001070, -0.00003374, 0.00001094, 0.00000432, 0.00003465, 0.00000325, 0.00004884, -0.00001097, 0.00000832, -0.00003607, 0.00000771, -0.00002456, -0.00000315, 0.00000894, -0.00000272, -0.00001259, 0.00000212, -0.00000166, -0.00000523, -0.00001069, 0.00000545, 0.00000705, -0.00000650, -0.00000318, 0.00000727, 0.00000386, 0.00000045, -0.00000975, 0.00000192, -0.00000048, 0.00000855, 0.00000069, -0.00000173, 0.00000052, -0.00000620, -0.00000393, 0.00000399, 0.00000335, 0.00000190, -0.00000021, -0.00000124, -0.00000565, 0.00000200, -0.00000664, 0.00000510, 0.00000248, -0.00000104, -0.00000368, 0.00000006, 0.00000349, 0.00000034, 0.00000427, -0.00000214, -0.00000675, 0.00000363, -0.00000062, -0.00000110, 0.00000150, -0.00000188, 0.00000104, -0.00000227, 0.00000385, },
		{ -0.00000563, -0.00000108, 0.00000748, -0.00000043, 0.00000292, -0.00000412, -0.00000157, -0.00000194, 0.00000088, 0.00000087, -0.00000098, -0.00000071, -0.00000066, 0.00000243, 0.00000044, -0.00000126, 0.00000039, 0.00000513, -0.00000734, 0.00000061, -0.00000140, 0.00000374, -0.00000305, 0.00000719, -0.00000791, 0.00000253, -0.00000393, 0.00000003, 0.00000154, 0.00000302, 0.00000097, -0.00001344, 0.00000421, 0.00000256, -0.00000237, 0.00000008, 0.00000678, 0.00000063, -0.00000279, -0.00000381, 0.00000217, -0.00000130, 0.00000332, 0.00001056, -0.00001926, 0.00003127, -0.00002421, 0.00000776, -0.00001992, 0.00000521, -0.00000738, 0.00000115, -0.00000908, -0.00001042, -0.00000795, -0.00000687, -0.00003385, 0.00001010, -0.00008716, -0.00006244, 0.00016744, 0.00021452, 0.00004182, -0.00013534, -0.00000379, -0.00003310, -0.00001068, -0.00003386, 0.00001138, 0.00000808, 0.00003318, 0.00000668, 0.00004876, -0.00001182, 0.00000565, -0.00003755, 0.00001118, -0.00002882, 0.00000158, 0.00000576, -0.00000150, -0.00001406, 0.00000462, -0.00000385, -0.00000519, -0.00001050, 0.00000787, 0.00000429, -0.00000585, -0.00000301, 0.00000855, 0.00000258, 0.00000036, -0.00000972, 0.00000230, -0.00000055, 0.00000954, -0.00000073, -0.00000094, -0.00000020, -0.00000663, -0.00000296, 0.00000436, 0.00000276, 0.00000215, -0.00000050, -0.00000168, -0.00000516, 0.00000170, -0.00000640, 0.00000583, 0.00000158, -0.00000083, -0.00000408, 0.00000116, 0.00000253, 0.00000102, 0.00000396, -0.00000295, -0.00000600, 0.00000411, -0.00000142, -0.00000059, 0.00000121, -0.00000178, 0.00000098, -0.00000219, 0.00000392, },
		{ -0.00000655, 0.00000084, 0.00000669, -0.00000011, 0.00000235, -0.00000431, -0.00000142, -0.00000173, 0.00000101, 0.00000062, -0.00000088, -0.00000099, -0.00000012, 0.00000204, 0.00000071, -0.00000179, 0.00000152, 0.00000401, -0.00000742, 0.00000131, -0.00000151, 0.00000358, -0.00000279, 0.00000720, -0.00000905, 0.00000405, -0.00000573, 0.00000213, -0.00000015, 0.00000480, -0.00000146, -0.00001166, 0.00000412, 0.00000208, -0.00000276, 0.00000108, 0.00000714, -0.00000064, -0.00000248, -0.00000383, 0.00000281, -0.00000249, 0.00000647, 0.00000639, -0.00001578, 0.00002919, -0.00002497, 0.00000995, -0.00002339, 0.00000926, -0.00001093, 0.00000454, -0.00001279, -0.00000732, -0.00001060, -0.00000487, -0.00003719, 0.00001491, -0.00010416, -0.00003083, 0.00017011, 0.00019842, 0.00002808, -0.00012867, -0.00000189, -0.00003319, -0.00001138, -0.00003346, 0.00001157, 0.00001233, 0.00003099, 0.00001117, 0.00004754, -0.00001192, 0.00000149, -0.00003695, 0.00001289, -0.00003176, 0.00000547, 0.00000307, -0.00000068, -0.00001501, 0.00000673, -0.00000598, -0.00000523, -0.00000976, 0.00000962, 0.00000173, -0.00000514, -0.00000270, 0.00000965, 0.00000124, 0.00000040, -0.00000969, 0.00000257, -0.00000041, 0.00001030, -0.00000209, -0.00000010, -0.00000109, -0.00000689, -0.00000188, 0.00000444, 0.00000235, 0.00000230, -0.00000078, -0.00000212, -0.00000458, 0.00000122, -0.00000587, 0.00000623, 0.00000088, -0.00000079, -0.00000424, 0.00000204, 0.00000164, 0.00000178, 0.00000346, -0.00000359, -0.00000517, 0.00000440, -0.00000218, -0.00000001, 0.00000082, -0.00000157, 0.00000080, -0.00000198, 0.00000373, },
		{ -0.00000712, 0.00000264, 0.00000577, 0.00000032, 0.00000161, -0.00000431, -0.00000135, -0.00000143, 0.00000101, 0.00000049, -0.00000090, -0.00000113, 0.00000033, 0.00000167, 0.00000095, -0.00000230, 0.00000273, 0.00000257, -0.00000702, 0.00000165, -0.00000138, 0.00000316, -0.00000215, 0.00000666, -0.00000959, 0.00000506, -0.00000710, 0.00000396, -0.00000168, 0.00000647, -0.00000391, -0.00000957, 0.00000358, 0.00000188, -0.00000326, 0.00000236, 0.00000709, -0.00000173, -0.00000218, -0.00000378, 0.00000325, -0.00000336, 0.00000940, 0.00000192, -0.00001145, 0.00002575, -0.00002398, 0.00001034, -0.00002527, 0.00001210, -0.00001348, 0.00000694, -0.00001552, -0.00000518, -0.00001213, -0.00000434, -0.00003853, 0.00001684, -0.00011813, 0.00000249, 0.00016638, 0.00018548, 0.00001420, -0.00012072, -0.00000140, -0.00003242, -0.00001278, -0.00003260, 0.00001162, 0.00001690, 0.00002830, 0.00001652, 0.00004535, -0.00001156, -0.00000378, -0.00003444, 0.00001279, -0.00003322, 0.00000830, 0.00000103, -0.00000043, -0.00001531, 0.00000828, -0.00000791, -0.00000540, -0.00000844, 0.00001056, -0.00000040, -0.00000452, -0.00000215, 0.00001044, -0.00000004, 0.00000049, -0.00000963, 0.00000275, -0.00000003, 0.00001076, -0.00000332, 0.00000075, -0.00000210, -0.00000695, -0.00000078, 0.00000429, 0.00000213, 0.00000236, -0.00000105, -0.00000254, -0.00000393, 0.00000060, -0.00000508, 0.00000631, 0.00000041, -0.00000094, -0.00000412, 0.00000263, 0.00000091, 0.00000252, 0.00000282, -0.00000408, -0.00000427, 0.00000450, -0.00000284, 0.00000061, 0.00000034, -0.00000126, 0.00000052, -0.00000163, 0.00000328, },
		{ -0.00000729, 0.00000422, 0.00000483, 0.00000079, 0.00000075, -0.00000415, -0.00000136, -0.00000107, 0.00000088, 0.00000048, -0.00000105, -0.00000111, 0.00000065, 0.00000138, 0.00000111, -0.00000269, 0.00000390, 0.00000092, -0.00000622, 0.00000166, -0.00000103, 0.00000251, -0.00000118, 0.00000559, -0.00000949, 0.00000548, -0.00000793, 0.00000537, -0.00000289, 0.00000786, -0.00000621, -0.00000737, 0.00000272, 0.00000190, -0.00000384, 0.00000389, 0.00000664, -0.00000259, -0.00000192, -0.00000361, 0.00000343, -0.00000379, 0.00001188, -0.00000248, -0.00000667, 0.00002129, -0.00002149, 0.00000899, -0.00002550, 0.00001358, -0.00001488, 0.00000822, -0.00001713, -0.00000417, -0.00001239, -0.00000548, -0.00003755, 0.00001527, -0.00012767, 0.00003544, 0.00015769, 0.00017540, 0.00000023, -0.00011166, -0.00000217, -0.00003088, -0.00001478, -0.00003137, 0.00001172, 0.00002156, 0.00002539, 0.00002247, 0.00004242, -0.00001110, -0.00000969, -0.00003037, 0.00001104, -0.00003320, 0.00001000, -0.00000026, -0.00000080, -0.00001489, 0.00000917, -0.00000955, -0.00000569, -0.00000658, 0.00001065, -0.00000197, -0.00000413, -0.00000126, 0.00001080, -0.00000114, 0.00000055, -0.00000950, 0.00000281, 0.00000059, 0.00001089, -0.00000435, 0.00000154, -0.00000321, -0.00000678, 0.00000029, 0.00000396, 0.00000207, 0.00000230, -0.00000131, -0.00000290, -0.00000328, -0.00000009, -0.00000410, 0.00000610, 0.00000016, -0.00000127, -0.00000373, 0.00000291, 0.00000040, 0.00000316, 0.00000211, -0.00000444, -0.00000332, 0.00000439, -0.00000337, 0.00000120, -0.00000018, -0.00000088, 0.00000015, -0.00000118, 0.00000258, },
		{ -0.00000706, 0.00000552, 0.00000396, 0.00000122, -0.00000014, -0.00000387, -0.00000140, -0.00000066, 0.00000065, 0.00000057, -0.00000132, -0.00000093, 0.00000082, 0.00000120, 0.00000113, -0.00000289, 0.00000491, -0.00000080, -0.00000513, 0.00000138, -0.00000051, 0.00000170, 0.00000006, 0.00000404, -0.00000879, 0.00000526, -0.00000814, 0.00000625, -0.00000367, 0.00000884, -0.00000818, -0.00000524, 0.00000170, 0.00000205, -0.00000440, 0.00000557, 0.00000580, -0.00000319, -0.00000176, -0.00000330, 0.00000333, -0.00000367, 0.00001367, -0.00000650, -0.00000184, 0.00001627, -0.00001787, 0.00000616, -0.00002416, 0.00001372, -0.00001511, 0.00000837, -0.00001760, -0.00000429, -0.00001141, -0.00000826, -0.00003423, 0.00000992, -0.00013171, 0.00006613, 0.00014563, 0.00016761, -0.00001362, -0.00010181, -0.00000397, -0.00002875, -0.00001725, -0.00002992, 0.00001204, 0.00002607, 0.00002255, 0.00002872, 0.00003897, -0.00001089, -0.00001570, -0.00002524, 0.00000793, -0.00003185, 0.00001064, -0.00000082, -0.00000176, -0.00001376, 0.00000939, -0.00001090, -0.00000603, -0.00000429, 0.00000993, -0.00000293, -0.00000404, 0.00000002, 0.00001065, -0.00000195, 0.00000052, -0.00000929, 0.00000277, 0.00000145, 0.00001066, -0.00000513, 0.00000221, -0.00000435, -0.00000639, 0.00000126, 0.00000351, 0.00000214, 0.00000216, -0.00000157, -0.00000318, -0.00000269, -0.00000078, -0.00000299, 0.00000565, 0.00000009, -0.00000174, -0.00000311, 0.00000287, 0.00000017, 0.00000364, 0.00000139, -0.00000468, -0.00000233, 0.00000408, -0.00000373, 0.00000173, -0.00000070, -0.00000045, -0.00000027, -0.00000065, 0.00000166, },
		{ -0.00000643, 0.00000648, 0.00000321, 0.00000152, -0.00000100, -0.00000352, -0.00000146, -0.00000027, 0.00000038, 0.00000073, -0.00000168, -0.00000059, 0.00000082, 0.00000115, 0.00000100, -0.00000284, 0.00000567, -0.00000244, -0.00000387, 0.00000092, 0.00000011, 0.00000081, 0.00000144, 0.00000215, -0.00000755, 0.00000446, -0.00000774, 0.00000657, -0.00000394, 0.00000930, -0.00000969, -0.00000335, 0.00000065, 0.00000224, -0.00000488, 0.00000731, 0.00000464, -0.00000353, -0.00000169, -0.00000286, 0.00000296, -0.00000299, 0.00001463, -0.00000983, 0.00000265, 0.00001113, -0.00001359, 0.00000221, -0.00002149, 0.00001265, -0.00001432, 0.00000752, -0.00001708, -0.00000538, -0.00000938, -0.00001241, -0.00002886, 0.00000091, -0.00012965, 0.00009293, 0.00013178, 0.00016139, -0.00002699, -0.00009154, -0.00000649, -0.00002625, -0.00001999, -0.00002840, 0.00001277, 0.00003019, 0.00002006, 0.00003499, 0.00003522, -0.00001126, -0.00002126, -0.00001958, 0.00000387, -0.00002944, 0.00001036, -0.00000078, -0.00000320, -0.00001202, 0.00000898, -0.00001197, -0.00000633, -0.00000174, 0.00000851, -0.00000328, -0.00000424, 0.00000167, 0.00000998, -0.00000243, 0.00000035, -0.00000897, 0.00000264, 0.00000251, 0.00001008, -0.00000562, 0.00000270, -0.00000548, -0.00000579, 0.00000208, 0.00000301, 0.00000230, 0.00000194, -0.00000184, -0.00000334, -0.00000220, -0.00000138, -0.00000185, 0.00000504, 0.00000015, -0.00000228, -0.00000230, 0.00000255, 0.00000021, 0.00000391, 0.00000071, -0.00000481, -0.00000133, 0.00000357, -0.00000389, 0.00000215, -0.00000118, -0.00000003, -0.00000072, -0.00000010, 0.00000059, },
		{ -0.00000546, 0.00000710, 0.00000264, 0.00000166, -0.00000175, -0.00000315, -0.00000148, 0.00000009, 0.00000010, 0.00000090, -0.00000209, -0.00000012, 0.00000065, 0.00000123, 0.00000070, -0.00000252, 0.00000608, -0.00000387, -0.00000259, 0.00000037, 0.00000074, -0.00000005, 0.00000282, 0.00000006, -0.00000591, 0.00000316, -0.00000676, 0.00000632, -0.00000369, 0.00000922, -0.00001066, -0.00000182, -0.00000029, 0.00000238, -0.00000519, 0.00000901, 0.00000322, -0.00000363, -0.00000171, -0.00000231, 0.00000234, -0.00000174, 0.00001467, -0.00001225, 0.00000644, 0.00000635, -0.00000916, -0.00000239, -0.00001784, 0.00001063, -0.00001271, 0.00000591, -0.00001581, -0.00000716, -0.00000665, -0.00001751, -0.00002195, -0.00001126, -0.00012136, 0.00011463, 0.00011761, 0.00015599, -0.00003948, -0.00008126, -0.00000939, -0.00002363, -0.00002283, -0.00002695, 0.00001407, 0.00003370, 0.00001817, 0.00004105, 0.00003130, -0.00001242, -0.00002589, -0.00001396, -0.00000066, -0.00002631, 0.00000941, -0.00000032, -0.00000495, -0.00000984, 0.00000803, -0.00001283, -0.00000645, 0.00000085, 0.00000659, -0.00000312, -0.00000468, 0.00000364, 0.00000880, -0.00000254, 0.00000000, -0.00000855, 0.00000245, 0.00000373, 0.00000915, -0.00000580, 0.00000297, -0.00000654, -0.00000501, 0.00000271, 0.00000251, 0.00000250, 0.00000168, -0.00000213, -0.00000338, -0.00000187, -0.00000181, -0.00000078, 0.00000437, 0.00000027, -0.00000284, -0.00000141, 0.00000202, 0.00000050, 0.00000396, 0.00000010, -0.00000486, -0.00000033, 0.00000288, -0.00000383, 0.00000242, -0.00000157, 0.00000035, -0.00000114, 0.00000043, -0.00000057, },
		{ -0.00000421, 0.00000737, 0.00000226, 0.00000159, -0.00000236, -0.00000282, -0.00000145, 0.00000036, -0.00000013, 0.00000105, -0.00000250, 0.00000043, 0.00000036, 0.00000142, 0.00000026, -0.00000192, 0.00000610, -0.00000499, -0.00000141, -0.00000017, 0.00000130, -0.00000077, 0.00000407, -0.00000206, -0.00000404, 0.00000150, -0.00000531, 0.00000559, -0.00000296, 0.00000860, -0.00001106, -0.00000074, -0.00000102, 0.00000238, -0.00000523, 0.00001055, 0.00000163, -0.00000352, -0.00000181, -0.00000171, 0.00000157, -0.00000001, 0.00001378, -0.00001361, 0.00000926, 0.00000231, -0.00000507, -0.00000717, -0.00001361, 0.00000797, -0.00001059, 0.00000383, -0.00001411, -0.00000930, -0.00000364, -0.00002300, -0.00001424, -0.00002576, -0.00010722, 0.00013047, 0.00010432, 0.00015070, -0.00005070, -0.00007137, -0.00001238, -0.00002109, -0.00002560, -0.00002569, 0.00001606, 0.00003648, 0.00001708, 0.00004669, 0.00002726, -0.00001448, -0.00002924, -0.00000887, -0.00000521, -0.00002284, 0.00000807, 0.00000035, -0.00000679, -0.00000740, 0.00000667, -0.00001355, -0.00000626, 0.00000324, 0.00000438, -0.00000259, -0.00000526, 0.00000582, 0.00000717, -0.00000229, -0.00000052, -0.00000804, 0.00000225, 0.00000504, 0.00000793, -0.00000567, 0.00000299, -0.00000747, -0.00000408, 0.00000314, 0.00000207, 0.00000272, 0.00000138, -0.00000245, -0.00000327, -0.00000172, -0.00000203, 0.00000013, 0.00000370, 0.00000037, -0.00000334, -0.00000051, 0.00000136, 0.00000100, 0.00000380, -0.00000042, -0.00000482, 0.00000062, 0.00000205, -0.00000357, 0.00000252, -0.00000183, 0.00000064, -0.00000150, 0.00000088, -0.00000174, },
		{ -0.00000277, 0.00000734, 0.00000207, 0.00000132, -0.00000277, -0.00000255, -0.00000134, 0.00000052, -0.00000029, 0.00000112, -0.00000286, 0.00000101, -0.00000003, 0.00000171, -0.00000029, -0.00000108, 0.00000572, -0.00000572, -0.00000042, -0.00000060, 0.00000169, -0.00000126, 0.00000505, -0.00000403, -0.00000211, -0.00000036, -0.00000353, 0.00000446, -0.00000184, 0.00000750, -0.00001091, -0.00000015, -0.00000147, 0.00000219, -0.00000496, 0.00001185, -0.00000003, -0.00000324, -0.00000194, -0.00000110, 0.00000075, 0.00000206, 0.00001205, -0.00001389, 0.00001097, -0.00000070, -0.00000173, -0.00001164, -0.00000922, 0.00000502, -0.00000826, 0.00000161, -0.00001231, -0.00001140, -0.00000083, -0.00002827, -0.00000657, -0.00004149, -0.00008800, 0.00014018, 0.00009283, 0.00014492, -0.00006032, -0.00006219, -0.00001518, -0.00001883, -0.00002819, -0.00002465, 0.00001879, 0.00003844, 0.00001692, 0.00005179, 0.00002307, -0.00001738, -0.00003108, -0.00000472, -0.00000936, -0.00001939, 0.00000661, 0.00000101, -0.00000851, -0.00000490, 0.00000504, -0.00001422, -0.00000564, 0.00000521, 0.00000215, -0.00000188, -0.00000583, 0.00000805, 0.00000522, -0.00000172, -0.00000119, -0.00000746, 0.00000208, 0.00000637, 0.00000647, -0.00000525, 0.00000273, -0.00000821, -0.00000308, 0.00000339, 0.00000169, 0.00000292, 0.00000105, -0.00000277, -0.00000304, -0.00000177, -0.00000200, 0.00000083, 0.00000312, 0.00000040, -0.00000370, 0.00000030, 0.00000067, 0.00000162, 0.00000346, -0.00000083, -0.00000470, 0.00000149, 0.00000111, -0.00000311, 0.00000244, -0.00000196, 0.00000082, -0.00000176, 0.00000118, -0.00000281, },
		{ -0.00000125, 0.00000708, 0.00000202, 0.00000085, -0.00000298, -0.00000237, -0.00000113, 0.00000056, -0.00000034, 0.00000109, -0.00000311, 0.00000156, -0.00000047, 0.00000203, -0.00000089, -0.00000006, 0.00000498, -0.00000604, 0.00000029, -0.00000085, 0.00000188, -0.00000144, 0.00000564, -0.00000571, -0.00000031, -0.00000223, -0.00000158, 0.00000309, -0.00000043, 0.00000604, -0.00001029, -0.00000003, -0.00000160, 0.00000178, -0.00000433, 0.00001281, -0.00000166, -0.00000288, -0.00000206, -0.00000057, -0.00000001, 0.00000430, 0.00000964, -0.00001317, 0.00001152, -0.00000249, 0.00000054, -0.00001543, -0.00000504, 0.00000210, -0.00000601, -0.00000045, -0.00001073, -0.00001313, 0.00000138, -0.00003272, 0.00000021, -0.00005723, -0.00006486, 0.00014392, 0.00008369, 0.00013825, -0.00006814, -0.00005396, -0.00001762, -0.00001696, -0.00003054, -0.00002385, 0.00002225, 0.00003956, 0.00001775, 0.00005626, 0.00001863, -0.00002093, -0.00003139, -0.00000175, -0.00001280, -0.00001624, 0.00000527, 0.00000147, -0.00000994, -0.00000253, 0.00000327, -0.00001489, -0.00000451, 0.00000657, 0.00000013, -0.00000116, -0.00000625, 0.00001017, 0.00000307, -0.00000091, -0.00000199, -0.00000684, 0.00000198, 0.00000765, 0.00000485, -0.00000458, 0.00000221, -0.00000872, -0.00000205, 0.00000348, 0.00000140, 0.00000310, 0.00000070, -0.00000309, -0.00000271, -0.00000198, -0.00000171, 0.00000127, 0.00000269, 0.00000031, -0.00000387, 0.00000094, 0.00000004, 0.00000228, 0.00000298, -0.00000115, -0.00000449, 0.00000226, 0.00000013, -0.00000250, 0.00000219, -0.00000192, 0.00000086, -0.00000190, 0.00000131, -0.00000372, },
		{ 0.00000026, 0.00000666, 0.00000208, 0.00000023, -0.00000300, -0.00000229, -0.00000084, 0.00000046, -0.00000027, 0.00000092, -0.00000321, 0.00000203, -0.00000090, 0.00000234, -0.00000147, 0.00000105, 0.00000392, -0.00000595, 0.00000070, -0.00000089, 0.00000182, -0.00000129, 0.00000578, -0.00000695, 0.00000121, -0.00000395, 0.00000037, 0.00000161, 0.00000111, 0.00000435, -0.00000931, -0.00000032, -0.00000143, 0.00000115, -0.00000334, 0.00001337, -0.00000317, -0.00000247, -0.00000213, -0.00000017, -0.00000056, 0.00000650, 0.00000676, -0.00001160, 0.00001101, -0.00000304, 0.00000155, -0.00001822, -0.00000140, -0.00000053, -0.00000409, -0.00000212, -0.00000964, -0.00001421, 0.00000262, -0.00003586, 0.00000530, -0.00007172, -0.00003916, 0.00014228, 0.00007716, 0.00013046, -0.00007407, -0.00004682, -0.00001957, -0.00001555, -0.00003264, -0.00002322, 0.00002636, 0.00003990, 0.00001958, 0.00006006, 0.00001379, -0.00002483, -0.00003030, -0.00000005, -0.00001535, -0.00001361, 0.00000421, 0.00000158, -0.00001093, -0.00000043, 0.00000146, -0.00001556, -0.00000286, 0.00000720, -0.00000148, -0.00000063, -0.00000635, 0.00001201, 0.00000089, 0.00000004, -0.00000285, -0.00000621, 0.00000202, 0.00000878, 0.00000314, -0.00000371, 0.00000143, -0.00000896, -0.00000107, 0.00000344, 0.00000118, 0.00000326, 0.00000031, -0.00000335, -0.00000233, -0.00000232, -0.00000118, 0.00000145, 0.00000242, 0.00000007, -0.00000383, 0.00000135, -0.00000044, 0.00000291, 0.00000243, -0.00000140, -0.00000418, 0.00000286, -0.00000083, -0.00000178, 0.00000179, -0.00000174, 0.00000076, -0.00000190, 0.00000123, -0.00000438, },
		{ 0.00000167, 0.00000615, 0.00000216, -0.00000048, -0.00000285, -0.00000228, -0.00000047, 0.00000024, -0.00000007, 0.00000060, -0.00000314, 0.00000236, -0.00000127, 0.00000259, -0.00000197, 0.00000217, 0.00000266, -0.00000551, 0.00000081, -0.00000071, 0.00000152, -0.00000081, 0.00000543, -0.00000769, 0.00000232, -0.00000537, 0.00000217, 0.00000019, 0.00000264, 0.00000256, -0.00000810, -0.00000093, -0.00000100, 0.00000032, -0.00000199, 0.00001350, -0.00000451, -0.00000208, -0.00000211, 0.00000004, -0.00000079, 0.00000845, 0.00000369, -0.00000944, 0.00000964, -0.00000246, 0.00000127, -0.00001984, 0.00000148, -0.00000266, -0.00000265, -0.00000322, -0.00000918, -0.00001447, 0.00000269, -0.00003736, 0.00000810, -0.00008379, -0.00001239, 0.00013609, 0.00007319, 0.00012149, -0.00007812, -0.00004081, -0.00002100, -0.00001460, -0.00003451, -0.00002267, 0.00003102, 0.00003956, 0.00002239, 0.00006314, 0.00000835, -0.00002866, -0.00002809, 0.00000045, -0.00001693, -0.00001160, 0.00000354, 0.00000130, -0.00001141, 0.00000130, -0.00000033, -0.00001622, -0.00000073, 0.00000705, -0.00000255, -0.00000040, -0.00000601, 0.00001340, -0.00000117, 0.00000103, -0.00000373, -0.00000562, 0.00000223, 0.00000970, 0.00000146, -0.00000271, 0.00000043, -0.00000892, -0.00000019, 0.00000333, 0.00000102, 0.00000340, -0.00000014, -0.00000353, -0.00000196, -0.00000271, -0.00000047, 0.00000137, 0.00000233, -0.00000032, -0.00000357, 0.00000150, -0.00000072, 0.00000342, 0.00000186, -0.00000161, -0.00000377, 0.00000328, -0.00000173, -0.00000100, 0.00000128, -0.00000143, 0.00000052, -0.00000176, 0.00000094, -0.00000475, },
		{ 0.00000290, 0.00000564, 0.00000222, -0.00000122, -0.00000258, -0.00000233, -0.00000006, -0.00000007, 0.00000023, 0.00000014, -0.00000287, 0.00000253, -0.00000155, 0.00000274, -0.00000231, 0.00000319, 0.00000129, -0.00000480, 0.00000067, -0.00000033, 0.00000102, -0.00000003, 0.00000462, -0.00000788, 0.00000295, -0.00000637, 0.00000369, -0.00000106, 0.00000404, 0.00000083, -0.00000680, -0.00000174, -0.00000037, -0.00000063, -0.00000034, 0.00001318, -0.00000561, -0.00000174, -0.00000197, 0.00000001, -0.00000062, 0.00000996, 0.00000069, -0.00000697, 0.00000768, -0.00000096, -0.00000022, -0.00002024, 0.00000349, -0.00000421, -0.00000176, -0.00000371, -0.00000943, -0.00001385, 0.00000152, -0.00003709, 0.00000822, -0.00009249, 0.00001398, 0.00012642, 0.00007149, 0.00011147, -0.00008043, -0.00003588, -0.00002193, -0.00001406, -0.00003624, -0.00002208, 0.00003609, 0.00003866, 0.00002611, 0.00006546, 0.00000220, -0.00003202, -0.00002516, -0.00000004, -0.00001764, -0.00001021, 0.00000325, 0.00000061, -0.00001137, 0.00000261, -0.00000208, -0.00001679, 0.00000176, 0.00000616, -0.00000301, -0.00000056, -0.00000516, 0.00001423, -0.00000296, 0.00000196, -0.00000456, -0.00000508, 0.00000264, 0.00001032, -0.00000012, -0.00000167, -0.00000072, -0.00000861, 0.00000055, 0.00000318, 0.00000089, 0.00000353, -0.00000066, -0.00000359, -0.00000166, -0.00000310, 0.00000034, 0.00000109, 0.00000239, -0.00000083, -0.00000310, 0.00000140, -0.00000074, 0.00000375, 0.00000134, -0.00000179, -0.00000324, 0.00000348, -0.00000249, -0.00000022, 0.00000071, -0.00000103, 0.00000017, -0.00000151, 0.00000045, -0.00000479, },
		{ 0.00000389, 0.00000518, 0.00000218, -0.00000193, -0.00000224, -0.00000241, 0.00000037, -0.00000043, 0.00000060, -0.00000043, -0.00000241, 0.00000250, -0.00000170, 0.00000275, -0.00000244, 0.00000402, -0.00000005, -0.00000392, 0.00000033, 0.00000018, 0.00000038, 0.00000097, 0.00000340, -0.00000755, 0.00000306, -0.00000688, 0.00000484, -0.00000202, 0.00000518, -0.00000072, -0.00000556, -0.00000263, 0.00000035, -0.00000164, 0.00000154, 0.00001244, -0.00000644, -0.00000148, -0.00000171, -0.00000025, 0.00000002, 0.00001086, -0.00000197, -0.00000450, 0.00000545, 0.00000115, -0.00000270, -0.00001950, 0.00000462, -0.00000516, -0.00000141, -0.00000362, -0.00001032, -0.00001245, -0.00000077, -0.00003509, 0.00000554, -0.00009711, 0.00003861, 0.00011440, 0.00007161, 0.00010062, -0.00008123, -0.00003191, -0.00002244, -0.00001387, -0.00003791, -0.00002130, 0.00004140, 0.00003736, 0.00003069, 0.00006693, -0.00000475, -0.00003449, -0.00002195, -0.00000122, -0.00001762, -0.00000936, 0.00000329, -0.00000041, -0.00001087, 0.00000353, -0.00000382, -0.00001715, 0.00000444, 0.00000463, -0.00000288, -0.00000113, -0.00000376, 0.00001442, -0.00000438, 0.00000273, -0.00000530, -0.00000463, 0.00000328, 0.00001060, -0.00000151, -0.00000066, -0.00000196, -0.00000806, 0.00000112, 0.00000304, 0.00000078, 0.00000367, -0.00000126, -0.00000348, -0.00000151, -0.00000338, 0.00000118, 0.00000068, 0.00000255, -0.00000142, -0.00000249, 0.00000107, -0.00000051, 0.00000387, 0.00000091, -0.00000197, -0.00000261, 0.00000345, -0.00000306, 0.00000049, 0.00000012, -0.00000057, -0.00000026, -0.00000117, -0.00000020, -0.00000452, },
		{ 0.00000463, 0.00000482, 0.00000200, -0.00000252, -0.00000189, -0.00000247, 0.00000076, -0.00000081, 0.00000101, -0.00000109, -0.00000179, 0.00000228, -0.00000171, 0.00000262, -0.00000234, 0.00000458, -0.00000127, -0.00000299, -0.00000011, 0.00000076, -0.00000032, 0.00000208, 0.00000187, -0.00000676, 0.00000268, -0.00000690, 0.00000556, -0.00000265, 0.00000601, -0.00000200, -0.00000451, -0.00000347, 0.00000107, -0.00000262, 0.00000355, 0.00001133, -0.00000701, -0.00000131, -0.00000136, -0.00000072, 0.00000113, 0.00001106, -0.00000409, -0.00000230, 0.00000329, 0.00000351, -0.00000588, -0.00001779, 0.00000495, -0.00000559, -0.00000148, -0.00000308, -0.00001171, -0.00001045, -0.00000395, -0.00003166, 0.00000023, -0.00009730, 0.00006041, 0.00010114, 0.00007303, 0.00008924, -0.00008076, -0.00002877, -0.00002262, -0.00001393, -0.00003960, -0.00002019, 0.00004680, 0.00003580, 0.00003608, 0.00006740, -0.00001246, -0.00003575, -0.00001892, -0.00000275, -0.00001713, -0.00000887, 0.00000354, -0.00000163, -0.00001000, 0.00000409, -0.00000558, -0.00001718, 0.00000709, 0.00000266, -0.00000226, -0.00000205, -0.00000187, 0.00001394, -0.00000536, 0.00000330, -0.00000590, -0.00000426, 0.00000414, 0.00001050, -0.00000263, 0.00000024, -0.00000320, -0.00000733, 0.00000152, 0.00000293, 0.00000066, 0.00000382, -0.00000195, -0.00000319, -0.00000154, -0.00000348, 0.00000194, 0.00000021, 0.00000275, -0.00000203, -0.00000180, 0.00000057, -0.00000006, 0.00000376, 0.00000060, -0.00000217, -0.00000188, 0.00000318, -0.00000342, 0.00000108, -0.00000042, -0.00000011, -0.00000073, -0.00000080, -0.00000096, -0.00000394, },
		{ 0.00000510, 0.00000457, 0.00000165, -0.00000296, -0.00000157, -0.00000249, 0.00000108, -0.00000115, 0.00000140, -0.00000178, -0.00000103, 0.00000187, -0.00000158, 0.00000235, -0.00000199, 0.00000483, -0.00000225, -0.00000213, -0.00000055, 0.00000133, -0.00000099, 0.00000320, 0.00000017, -0.00000562, 0.00000187, -0.00000644, 0.00000586, -0.00000291, 0.00000648, -0.00000295, -0.00000372, -0.00000415, 0.00000169, -0.00000345, 0.00000559, 0.00000992, -0.00000734, -0.00000121, -0.00000094, -0.00000133, 0.00000266, 0.00001051, -0.00000552, -0.00000060, 0.00000150, 0.00000575, -0.00000941, -0.00001536, 0.00000464, -0.00000563, -0.00000183, -0.00000226, -0.00001339, -0.00000811, -0.00000766, -0.00002722, -0.00000728, -0.00009305, 0.00007854, 0.00008764, 0.00007519, 0.00007766, -0.00007932, -0.00002629, -0.00002259, -0.00001417, -0.00004141, -0.00001863, 0.00005215, 0.00003412, 0.00004221, 0.00006667, -0.00002075, -0.00003561, -0.00001643, -0.00000427, -0.00001643, -0.00000858, 0.00000383, -0.00000292, -0.00000889, 0.00000437, -0.00000742, -0.00001675, 0.00000948, 0.00000046, -0.00000129, -0.00000321, 0.00000042, 0.00001284, -0.00000585, 0.00000360, -0.00000635, -0.00000398, 0.00000520, 0.00001000, -0.00000343, 0.00000094, -0.00000437, -0.00000646, 0.00000178, 0.00000285, 0.00000054, 0.00000398, -0.00000271, -0.00000271, -0.00000179, -0.00000334, 0.00000253, -0.00000022, 0.00000292, -0.00000258, -0.00000112, -0.00000002, 0.00000056, 0.00000345, 0.00000041, -0.00000237, -0.00000108, 0.00000269, -0.00000354, 0.00000152, -0.00000087, 0.00000031, -0.00000119, -0.00000044, -0.00000177, -0.00000311, },
		{ 0.00000534, 0.00000443, 0.00000113, -0.00000320, -0.00000132, -0.00000243, 0.00000130, -0.00000140, 0.00000172, -0.00000244, -0.00000020, 0.00000132, -0.00000133, 0.00000197, -0.00000142, 0.00000474, -0.00000293, -0.00000141, -0.00000092, 0.00000181, -0.00000152, 0.00000418, -0.00000157, -0.00000426, 0.00000074, -0.00000557, 0.00000576, -0.00000282, 0.00000660, -0.00000355, -0.00000327, -0.00000459, 0.00000214, -0.00000406, 0.00000755, 0.00000830, -0.00000746, -0.00000115, -0.00000051, -0.00000201, 0.00000451, 0.00000926, -0.00000618, 0.00000045, 0.00000033, 0.00000755, -0.00001293, -0.00001247, 0.00000386, -0.00000543, -0.00000228, -0.00000137, -0.00001513, -0.00000574, -0.00001151, -0.00002232, -0.00001634, -0.00008467, 0.00009253, 0.00007474, 0.00007758, 0.00006619, -0.00007715, -0.00002432, -0.00002244, -0.00001451, -0.00004338, -0.00001653, 0.00005735, 0.00003246, 0.00004906, 0.00006449, -0.00002933, -0.00003404, -0.00001476, -0.00000548, -0.00001576, -0.00000829, 0.00000403, -0.00000412, -0.00000767, 0.00000445, -0.00000937, -0.00001575, 0.00001140, -0.00000170, -0.00000017, -0.00000444, 0.00000296, 0.00001121, -0.00000588, 0.00000363, -0.00000665, -0.00000374, 0.00000642, 0.00000912, -0.00000388, 0.00000139, -0.00000538, -0.00000556, 0.00000191, 0.00000280, 0.00000042, 0.00000411, -0.00000352, -0.00000208, -0.00000226, -0.00000293, 0.00000290, -0.00000055, 0.00000301, -0.00000302, -0.00000052, -0.00000061, 0.00000128, 0.00000296, 0.00000034, -0.00000258, -0.00000024, 0.00000201, -0.00000342, 0.00000176, -0.00000119, 0.00000063, -0.00000161, -0.00000015, -0.00000255, -0.00000210, },
		{ 0.00000539, 0.00000438, 0.00000044, -0.00000323, -0.00000118, -0.00000227, 0.00000138, -0.00000154, 0.00000193, -0.00000301, 0.00000065, 0.00000067, -0.00000100, 0.00000151, -0.00000067, 0.00000433, -0.00000328, -0.00000091, -0.00000115, 0.00000213, -0.00000187, 0.00000494, -0.00000320, -0.00000281, -0.00000059, -0.00000439, 0.00000533, -0.00000243, 0.00000640, -0.00000382, -0.00000315, -0.00000472, 0.00000234, -0.00000437, 0.00000932, 0.00000656, -0.00000742, -0.00000109, -0.00000014, -0.00000263, 0.00000654, 0.00000739, -0.00000609, 0.00000078, -0.00000007, 0.00000864, -0.00001613, -0.00000942, 0.00000280, -0.00000517, -0.00000264, -0.00000061, -0.00001669, -0.00000364, -0.00001508, -0.00001754, -0.00002618, -0.00007275, 0.00010219, 0.00006307, 0.00007978, 0.00005510, -0.00007451, -0.00002271, -0.00002225, -0.00001491, -0.00004555, -0.00001380, 0.00006227, 0.00003094, 0.00005654, 0.00006058, -0.00003777, -0.00003116, -0.00001407, -0.00000617, -0.00001533, -0.00000786, 0.00000401, -0.00000510, -0.00000646, 0.00000439, -0.00001145, -0.00001413, 0.00001266, -0.00000359, 0.00000090, -0.00000557, 0.00000557, 0.00000917, -0.00000551, 0.00000342, -0.00000680, -0.00000353, 0.00000773, 0.00000791, -0.00000397, 0.00000156, -0.00000617, -0.00000468, 0.00000199, 0.00000274, 0.00000034, 0.00000418, -0.00000431, -0.00000132, -0.00000292, -0.00000223, 0.00000300, -0.00000072, 0.00000297, -0.00000329, -0.00000009, -0.00000109, 0.00000201, 0.00000237, 0.00000037, -0.00000276, 0.00000060, 0.00000118, -0.00000307, 0.00000180, -0.00000135, 0.00000083, -0.00000193, 0.00000004, -0.00000324, -0.00000098, },
		{ 0.00000531, 0.00000436, -0.00000036, -0.00000306, -0.00000115, -0.00000200, 0.00000132, -0.00000154, 0.00000198, -0.00000344, 0.00000145, -0.00000005, -0.00000062, 0.00000104, 0.00000019, 0.00000365, -0.00000331, -0.00000067, -0.00000118, 0.00000226, -0.00000196, 0.00000539, -0.00000460, -0.00000144, -0.00000198, -0.00000302, 0.00000467, -0.00000182, 0.00000594, -0.00000382, -0.00000336, -0.00000455, 0.00000226, -0.00000433, 0.00001080, 0.00000480, -0.00000728, -0.00000099, 0.00000010, -0.00000308, 0.00000857, 0.00000508, -0.00000536, 0.00000042, 0.00000034, 0.00000884, -0.00001873, -0.00000644, 0.00000165, -0.00000496, -0.00000277, -0.00000016, -0.00001788, -0.00000203, -0.00001800, -0.00001346, -0.00003593, -0.00005811, 0.00010767, 0.00005303, 0.00008145, 0.00004460, -0.00007157, -0.00002136, -0.00002206, -0.00001535, -0.00004793, -0.00001041, 0.00006685, 0.00002970, 0.00006455, 0.00005467, -0.00004554, -0.00002728, -0.00001432, -0.00000625, -0.00001525, -0.00000718, 0.00000368, -0.00000579, -0.00000536, 0.00000422, -0.00001362, -0.00001187, 0.00001313, -0.00000501, 0.00000172, -0.00000641, 0.00000805, 0.00000689, -0.00000482, 0.00000299, -0.00000684, -0.00000328, 0.00000906, 0.00000641, -0.00000372, 0.00000143, -0.00000668, -0.00000392, 0.00000207, 0.00000265, 0.00000032, 0.00000416, -0.00000503, -0.00000052, -0.00000373, -0.00000128, 0.00000281, -0.00000071, 0.00000275, -0.00000335, 0.00000013, -0.00000141, 0.00000265, 0.00000174, 0.00000045, -0.00000290, 0.00000139, 0.00000026, -0.00000253, 0.00000166, -0.00000135, 0.00000089, -0.00000214, 0.00000010, -0.00000375, 0.00000016, },
		{ 0.00000516, 0.00000433, -0.00000122, -0.00000272, -0.00000124, -0.00000164, 0.00000113, -0.00000138, 0.00000186, -0.00000368, 0.00000214, -0.00000076, -0.00000024, 0.00000061, 0.00000109, 0.00000277, -0.00000304, -0.00000070, -0.00000100, 0.00000218, -0.00000178, 0.00000547, -0.00000568, -0.00000024, -0.00000330, -0.00000158, 0.00000387, -0.00000106, 0.00000532, -0.00000362, -0.00000383, -0.00000407, 0.00000191, -0.00000390, 0.00001192, 0.00000311, -0.00000710, -0.00000080, 0.00000016, -0.00000325, 0.00001042, 0.00000253, -0.00000413, -0.00000049, 0.00000152, 0.00000808, -0.00002055, -0.00000374, 0.00000054, -0.00000490, -0.00000258, -0.00000011, -0.00001859, -0.00000110, -0.00002002, -0.00001054, -0.00004478, -0.00004168, 0.00010932, 0.00004484, 0.00008239, 0.00003484, -0.00006848, -0.00002022, -0.00002191, -0.00001582, -0.00005050, -0.00000630, 0.00007100, 0.00002888, 0.00007294, 0.00004657, -0.00005210, -0.00002282, -0.00001536, -0.00000573, -0.00001555, -0.00000623, 0.00000304, -0.00000614, -0.00000439, 0.00000393, -0.00001580, -0.00000907, 0.00001279, -0.00000583, 0.00000213, -0.00000680, 0.00001020, 0.00000456, -0.00000391, 0.00000241, -0.00000682, -0.00000294, 0.00001032, 0.00000473, -0.00000317, 0.00000101, -0.00000689, -0.00000332, 0.00000220, 0.00000249, 0.00000040, 0.00000399, -0.00000562, 0.00000025, -0.00000459, -0.00000013, 0.00000237, -0.00000049, 0.00000237, -0.00000321, 0.00000011, -0.00000148, 0.00000314, 0.00000114, 0.00000055, -0.00000296, 0.00000207, -0.00000069, -0.00000185, 0.00000134, -0.00000119, 0.00000079, -0.00000223, -0.00000001, -0.00000406, 0.00000124, },
		{ 0.00000500, 0.00000423, -0.00000208, -0.00000226, -0.00000140, -0.00000120, 0.00000081, -0.00000108, 0.00000154, -0.00000370, 0.00000267, -0.00000142, 0.00000011, 0.00000028, 0.00000194, 0.00000178, -0.00000254, -0.00000095, -0.00000062, 0.00000190, -0.00000135, 0.00000517, -0.00000636, 0.00000067, -0.00000444, -0.00000017, 0.00000303, -0.00000025, 0.00000462, -0.00000332, -0.00000447, -0.00000335, 0.00000129, -0.00000309, 0.00001263, 0.00000157, -0.00000693, -0.00000051, -0.00000002, -0.00000303, 0.00001191, -0.00000004, -0.00000262, -0.00000178, 0.00000332, 0.00000641, -0.00002152, -0.00000143, -0.00000047, -0.00000502, -0.00000204, -0.00000051, -0.00001877, -0.00000088, -0.00002101, -0.00000911, -0.00005201, -0.00002444, 0.00010768, 0.00003852, 0.00008249, 0.00002590, -0.00006534, -0.00001922, -0.00002180, -0.00001634, -0.00005324, -0.00000145, 0.00007465, 0.00002868, 0.00008144, 0.00003618, -0.00005691, -0.00001829, -0.00001691, -0.00000474, -0.00001617, -0.00000505, 0.00000212, -0.00000619, -0.00000356, 0.00000348, -0.00001786, -0.00000585, 0.00001169, -0.00000599, 0.00000204, -0.00000661, 0.00001185, 0.00000233, -0.00000292, 0.00000175, -0.00000677, -0.00000246, 0.00001141, 0.00000295, -0.00000238, 0.00000034, -0.00000680, -0.00000294, 0.00000243, 0.00000221, 0.00000062, 0.00000362, -0.00000599, 0.00000089, -0.00000542, 0.00000113, 0.00000171, -0.00000011, 0.00000185, -0.00000289, -0.00000015, -0.00000128, 0.00000341, 0.00000063, 0.00000062, -0.00000292, 0.00000261, -0.00000160, -0.00000109, 0.00000090, -0.00000089, 0.00000056, -0.00000220, -0.00000027, -0.00000412, 0.00000220, },
		{ 0.00000488, 0.00000399, -0.00000286, -0.00000172, -0.00000163, -0.00000072, 0.00000040, -0.00000066, 0.00000106, -0.00000349, 0.00000299, -0.00000199, 0.00000038, 0.00000009, 0.00000265, 0.00000080, -0.00000191, -0.00000137, -0.00000008, 0.00000147, -0.00000071, 0.00000451, -0.00000664, 0.00000124, -0.00000531, 0.00000110, 0.00000222, 0.00000053, 0.00000393, -0.00000303, -0.00000519, -0.00000246, 0.00000048, -0.00000194, 0.00001292, 0.00000022, -0.00000680, -0.00000009, -0.00000044, -0.00000236, 0.00001289, -0.00000240, -0.00000104, -0.00000321, 0.00000550, 0.00000395, -0.00002162, 0.00000041, -0.00000134, -0.00000529, -0.00000120, -0.00000132, -0.00001847, -0.00000132, -0.00002098, -0.00000932, -0.00005707, -0.00000731, 0.00010342, 0.00003396, 0.00008171, 0.00001781, -0.00006220, -0.00001835, -0.00002174, -0.00001692, -0.00005613, 0.00000416, 0.00007771, 0.00002929, 0.00008968, 0.00002354, -0.00005951, -0.00001422, -0.00001861, -0.00000350, -0.00001699, -0.00000374, 0.00000102, -0.00000599, -0.00000282, 0.00000281, -0.00001963, -0.00000244, 0.00000996, -0.00000552, 0.00000144, -0.00000580, 0.00001288, 0.00000036, -0.00000195, 0.00000107, -0.00000673, -0.00000179, 0.00001226, 0.00000117, -0.00000146, -0.00000052, -0.00000644, -0.00000278, 0.00000280, 0.00000179, 0.00000099, 0.00000302, -0.00000609, 0.00000131, -0.00000610, 0.00000240, 0.00000092, 0.00000038, 0.00000122, -0.00000242, -0.00000062, -0.00000081, 0.00000342, 0.00000027, 0.00000063, -0.00000275, 0.00000296, -0.00000241, -0.00000031, 0.00000039, -0.00000050, 0.00000020, -0.00000207, -0.00000067, -0.00000392, 0.00000299, },
		},
		{ { -0.00000216, -0.00000132, 0.00000007, 0.00000048, 0.00000077, 0.00000029, 0.00000031, -0.00000022, -0.00000044, 0.00000038, 0.00000112, 0.00000052, -0.00000219, -0.00000056, 0.00000018, 0.00000028, -0.00000017, -0.00000026, -0.00000033, -0.00000110, 0.00000119, -0.00000079, 0.00000073, 0.00000083, -0.00000037, -0.00000100, -0.00000052, 0.00000004, -0.00000386, 0.00000176, 0.00000267, 0.00000149, -0.00000110, -0.00000241, -0.00000171, -0.00000102, 0.00000526, -0.00000141, -0.00000100, -0.00000140, -0.00000228, -0.00000134, 0.00000144, 0.00000287, -0.00000845, 0.00000036, 0.00000073, 0.00000071, -0.00000110, -0.00000258, -0.00000253, -0.00000356, 0.00001124, -0.00000728, -0.00000363, -0.00000122, -0.00000557, -0.00001277, -0.00001557, -0.00001457, -0.00006803, -0.00000521, 0.00001940, 0.00005267, 0.00015504, 0.00005267, 0.00001940, -0.00000521, -0.00006803, -0.00001457, -0.00001557, -0.00001277, -0.00000557, -0.00000122, -0.00000363, -0.00000728, 0.00001124, -0.00000356, -0.00000253, -0.00000258, -0.00000110, 0.00000071, 0.00000073, 0.00000036, -0.00000845, 0.00000287, 0.00000144, -0.00000134, -0.00000228, -0.00000140, -0.00000100, -0.00000141, 0.00000526, -0.00000102, -0.00000171, -0.00000241, -0.00000110, 0.00000149, 0.00000267, 0.00000176, -0.00000386, 0.00000004, -0.00000052, -0.00000100, -0.00000037, 0.00000083, 0.00000073, -0.00000079, 0.00000119, -0.00000110, -0.00000033, -0.00000026, -0.00000017, 0.00000028, 0.00000018, -0.00000056, -0.00000219, 0.00000052, 0.00000112, 0.00000038, -0.00000044, -0.00000022, 0.00000031, 0.00000029, 0.00000077, 0.00000048, 0.00000007, -0.00000132, },
		{ -0.00000215, -0.00000128, 0.00000002, 0.00000057, 0.00000067, 0.00000028, 0.00000028, -0.00000028, -0.00000026, 0.00000029, 0.00000122, 0.00000044, -0.00000220, -0.00000040, 0.00000002, 0.00000033, -0.00000019, -0.00000029, -0.00000054, -0.00000090, 0.00000127, -0.00000083, 0.00000089, 0.00000061, -0.00000011, -0.00000124, -0.00000056, 0.00000005, -0.00000391, 0.00000220, 0.00000232, 0.00000138, -0.00000092, -0.00000254, -0.00000189, -0.00000050, 0.00000528, -0.00000152, -0.00000108, -0.00000175, -0.00000180, -0.00000167, 0.00000151, 0.00000269, -0.00000839, 0.00000106, 0.00000026, 0.00000052, -0.00000058, -0.00000276, -0.00000360, -0.00000213, 0.00001080, -0.00000731, -0.00000355, -0.00000240, -0.00000341, -0.00001486, -0.00001466, -0.00001553, -0.00006696, 0.00000100, 0.00001221, 0.00004981, 0.00015521, 0.00005547, 0.00002674, -0.00001211, -0.00006849, -0.00001391, -0.00001631, -0.00001070, -0.00000779, 0.00000011, -0.00000386, -0.00000711, 0.00001153, -0.00000491, -0.00000142, -0.00000247, -0.00000159, 0.00000090, 0.00000121, -0.00000038, -0.00000845, 0.00000303, 0.00000137, -0.00000100, -0.00000276, -0.00000104, -0.00000093, -0.00000129, 0.00000520, -0.00000151, -0.00000152, -0.00000227, -0.00000130, 0.00000164, 0.00000299, 0.00000130, -0.00000378, 0.00000000, -0.00000045, -0.00000078, -0.00000061, 0.00000105, 0.00000056, -0.00000073, 0.00000108, -0.00000126, -0.00000013, -0.00000023, -0.00000015, 0.00000024, 0.00000034, -0.00000074, -0.00000216, 0.00000058, 0.00000103, 0.00000047, -0.00000063, -0.00000015, 0.00000032, 0.00000032, 0.00000086, 0.00000039, 0.00000012, -0.00000138, },
		{ -0.00000213, -0.00000125, -0.00000002, 0.00000066, 0.00000056, 0.00000029, 0.00000024, -0.00000031, -0.00000010, 0.00000020, 0.00000132, 0.00000033, -0.00000217, -0.00000028, -0.00000012, 0.00000038, -0.00000023, -0.00000031, -0.00000076, -0.00000066, 0.00000130, -0.00000084, 0.00000103, 0.00000039, 0.00000016, -0.00000151, -0.00000057, 0.00000003, -0.00000393, 0.00000263, 0.00000196, 0.00000130, -0.00000076, -0.00000268, -0.00000207, 0.00000003, 0.00000526, -0.00000162, -0.00000117, -0.00000208, -0.00000133, -0.00000199, 0.00000158, 0.00000250, -0.00000828, 0.00000172, -0.00000022, 0.00000033, -0.00000004, -0.00000302, -0.00000462, -0.00000064, 0.00001021, -0.00000718, -0.00000363, -0.00000342, -0.00000139, -0.00001692, -0.00001366, -0.00001671, -0.00006534, 0.00000652, 0.00000522, 0.00004687, 0.00015573, 0.00005827, 0.00003415, -0.00001968, -0.00006831, -0.00001360, -0.00001683, -0.00000874, -0.00001002, 0.00000153, -0.00000420, -0.00000681, 0.00001167, -0.00000617, -0.00000028, -0.00000244, -0.00000203, 0.00000108, 0.00000170, -0.00000118, -0.00000839, 0.00000316, 0.00000132, -0.00000068, -0.00000324, -0.00000068, -0.00000087, -0.00000116, 0.00000512, -0.00000199, -0.00000131, -0.00000213, -0.00000150, 0.00000181, 0.00000328, 0.00000083, -0.00000367, -0.00000006, -0.00000034, -0.00000059, -0.00000083, 0.00000125, 0.00000039, -0.00000066, 0.00000095, -0.00000140, 0.00000006, -0.00000018, -0.00000014, 0.00000019, 0.00000052, -0.00000094, -0.00000210, 0.00000061, 0.00000096, 0.00000055, -0.00000081, -0.00000006, 0.00000031, 0.00000037, 0.00000094, 0.00000030, 0.00000019, -0.00000145, },
		{ -0.00000209, -0.00000124, -0.00000005, 0.00000075, 0.00000044, 0.00000031, 0.00000017, -0.00000032, 0.00000005, 0.00000012, 0.00000142, 0.00000022, -0.00000211, -0.00000018, -0.00000026, 0.00000042, -0.00000027, -0.00000032, -0.00000100, -0.00000038, 0.00000130, -0.00000082, 0.00000115, 0.00000017, 0.00000043, -0.00000180, -0.00000056, -0.00000001, -0.00000392, 0.00000304, 0.00000157, 0.00000125, -0.00000062, -0.00000280, -0.00000224, 0.00000059, 0.00000519, -0.00000170, -0.00000129, -0.00000238, -0.00000088, -0.00000230, 0.00000164, 0.00000231, -0.00000813, 0.00000234, -0.00000071, 0.00000016, 0.00000052, -0.00000334, -0.00000558, 0.00000090, 0.00000947, -0.00000689, -0.00000389, -0.00000424, 0.00000046, -0.00001889, -0.00001262, -0.00001805, -0.00006325, 0.00001136, -0.00000154, 0.00004381, 0.00015657, 0.00006111, 0.00004157, -0.00002790, -0.00006743, -0.00001371, -0.00001707, -0.00000693, -0.00001221, 0.00000299, -0.00000465, -0.00000639, 0.00001167, -0.00000734, 0.00000089, -0.00000249, -0.00000241, 0.00000124, 0.00000219, -0.00000204, -0.00000824, 0.00000325, 0.00000129, -0.00000038, -0.00000371, -0.00000031, -0.00000080, -0.00000104, 0.00000501, -0.00000246, -0.00000109, -0.00000201, -0.00000171, 0.00000201, 0.00000354, 0.00000036, -0.00000352, -0.00000014, -0.00000020, -0.00000045, -0.00000102, 0.00000142, 0.00000022, -0.00000058, 0.00000079, -0.00000150, 0.00000022, -0.00000012, -0.00000014, 0.00000015, 0.00000069, -0.00000117, -0.00000201, 0.00000062, 0.00000091, 0.00000061, -0.00000100, 0.00000004, 0.00000029, 0.00000043, 0.00000099, 0.00000022, 0.00000025, -0.00000153, },
		{ -0.00000203, -0.00000124, -0.00000006, 0.00000082, 0.00000031, 0.00000035, 0.00000009, -0.00000030, 0.00000017, 0.00000005, 0.00000153, 0.00000009, -0.00000204, -0.00000010, -0.00000037, 0.00000046, -0.00000031, -0.00000032, -0.00000125, -0.00000007, 0.00000125, -0.00000076, 0.00000124, -0.00000004, 0.00000069, -0.00000210, -0.00000052, -0.00000006, -0.00000387, 0.00000342, 0.00000117, 0.00000123, -0.00000051, -0.00000291, -0.00000241, 0.00000117, 0.00000508, -0.00000175, -0.00000142, -0.00000265, -0.00000046, -0.00000258, 0.00000168, 0.00000213, -0.00000796, 0.00000293, -0.00000121, 0.00000002, 0.00000107, -0.00000372, -0.00000647, 0.00000249, 0.00000858, -0.00000643, -0.00000433, -0.00000484, 0.00000209, -0.00002070, -0.00001162, -0.00001946, -0.00006077, 0.00001557, -0.00000804, 0.00004061, 0.00015774, 0.00006406, 0.00004889, -0.00003671, -0.00006585, -0.00001427, -0.00001698, -0.00000533, -0.00001430, 0.00000447, -0.00000516, -0.00000587, 0.00001153, -0.00000842, 0.00000207, -0.00000261, -0.00000274, 0.00000137, 0.00000269, -0.00000296, -0.00000801, 0.00000329, 0.00000129, -0.00000010, -0.00000414, 0.00000004, -0.00000073, -0.00000093, 0.00000489, -0.00000292, -0.00000084, -0.00000191, -0.00000190, 0.00000223, 0.00000377, -0.00000012, -0.00000335, -0.00000026, -0.00000003, -0.00000034, -0.00000117, 0.00000156, 0.00000007, -0.00000050, 0.00000062, -0.00000157, 0.00000037, -0.00000006, -0.00000015, 0.00000012, 0.00000086, -0.00000141, -0.00000188, 0.00000059, 0.00000088, 0.00000067, -0.00000118, 0.00000015, 0.00000025, 0.00000052, 0.00000103, 0.00000015, 0.00000031, -0.00000162, },
		{ -0.00000196, -0.00000126, -0.00000006, 0.00000088, 0.00000019, 0.00000040, -0.00000000, -0.00000026, 0.00000028, -0.00000000, 0.00000162, -0.00000005, -0.00000194, -0.00000006, -0.00000047, 0.00000048, -0.00000035, -0.00000033, -0.00000149, 0.00000027, 0.00000116, -0.00000066, 0.00000130, -0.00000022, 0.00000094, -0.00000240, -0.00000047, -0.00000014, -0.00000380, 0.00000378, 0.00000075, 0.00000126, -0.00000044, -0.00000300, -0.00000258, 0.00000178, 0.00000490, -0.00000177, -0.00000159, -0.00000289, -0.00000008, -0.00000283, 0.00000169, 0.00000197, -0.00000777, 0.00000350, -0.00000173, -0.00000009, 0.00000162, -0.00000415, -0.00000728, 0.00000410, 0.00000756, -0.00000581, -0.00000496, -0.00000518, 0.00000345, -0.00002230, -0.00001073, -0.00002087, -0.00005798, 0.00001917, -0.00001426, 0.00003726, 0.00015920, 0.00006718, 0.00005603, -0.00004606, -0.00006356, -0.00001531, -0.00001652, -0.00000400, -0.00001623, 0.00000591, -0.00000571, -0.00000528, 0.00001127, -0.00000939, 0.00000325, -0.00000280, -0.00000299, 0.00000147, 0.00000321, -0.00000394, -0.00000768, 0.00000327, 0.00000133, 0.00000015, -0.00000455, 0.00000038, -0.00000066, -0.00000084, 0.00000475, -0.00000337, -0.00000058, -0.00000183, -0.00000209, 0.00000247, 0.00000397, -0.00000060, -0.00000315, -0.00000039, 0.00000017, -0.00000027, -0.00000129, 0.00000167, -0.00000006, -0.00000042, 0.00000044, -0.00000161, 0.00000051, 0.00000002, -0.00000017, 0.00000010, 0.00000103, -0.00000167, -0.00000173, 0.00000054, 0.00000088, 0.00000070, -0.00000135, 0.00000027, 0.00000021, 0.00000062, 0.00000105, 0.00000008, 0.00000037, -0.00000171, },
		{ -0.00000188, -0.00000130, -0.00000004, 0.00000093, 0.00000006, 0.00000045, -0.00000011, -0.00000019, 0.00000036, -0.00000003, 0.00000170, -0.00000018, -0.00000183, -0.00000003, -0.00000055, 0.00000050, -0.00000039, -0.00000034, -0.00000174, 0.00000065, 0.00000102, -0.00000052, 0.00000132, -0.00000038, 0.00000117, -0.00000270, -0.00000040, -0.00000023, -0.00000369, 0.00000412, 0.00000033, 0.00000131, -0.00000040, -0.00000307, -0.00000274, 0.00000242, 0.00000467, -0.00000175, -0.00000179, -0.00000308, 0.00000026, -0.00000305, 0.00000168, 0.00000184, -0.00000758, 0.00000406, -0.00000227, -0.00000016, 0.00000213, -0.00000463, -0.00000802, 0.00000573, 0.00000641, -0.00000504, -0.00000578, -0.00000527, 0.00000451, -0.00002364, -0.00000998, -0.00002220, -0.00005496, 0.00002223, -0.00002020, 0.00003376, 0.00016094, 0.00007054, 0.00006286, -0.00005585, -0.00006058, -0.00001686, -0.00001566, -0.00000298, -0.00001794, 0.00000727, -0.00000628, -0.00000462, 0.00001090, -0.00001026, 0.00000442, -0.00000307, -0.00000318, 0.00000154, 0.00000373, -0.00000498, -0.00000725, 0.00000318, 0.00000142, 0.00000035, -0.00000492, 0.00000070, -0.00000057, -0.00000076, 0.00000461, -0.00000382, -0.00000029, -0.00000178, -0.00000225, 0.00000271, 0.00000415, -0.00000108, -0.00000293, -0.00000055, 0.00000040, -0.00000024, -0.00000137, 0.00000174, -0.00000017, -0.00000037, 0.00000025, -0.00000163, 0.00000063, 0.00000011, -0.00000021, 0.00000010, 0.00000119, -0.00000194, -0.00000155, 0.00000045, 0.00000090, 0.00000070, -0.00000150, 0.00000039, 0.00000016, 0.00000074, 0.00000105, 0.00000004, 0.00000042, -0.00000180, },
		{ -0.00000179, -0.00000135, -0.00000001, 0.00000097, -0.00000006, 0.00000051, -0.00000023, -0.00000009, 0.00000041, -0.00000005, 0.00000176, -0.00000032, -0.00000171, -0.00000003, -0.00000061, 0.00000050, -0.00000042, -0.00000035, -0.00000199, 0.00000105, 0.00000084, -0.00000033, 0.00000129, -0.00000050, 0.00000137, -0.00000299, -0.00000032, -0.00000034, -0.00000356, 0.00000442, -0.00000010, 0.00000140, -0.00000040, -0.00000311, -0.00000291, 0.00000308, 0.00000438, -0.00000170, -0.00000202, -0.00000323, 0.00000056, -0.00000322, 0.00000163, 0.00000175, -0.00000740, 0.00000460, -0.00000284, -0.00000018, 0.00000260, -0.00000515, -0.00000866, 0.00000735, 0.00000515, -0.00000412, -0.00000677, -0.00000509, 0.00000525, -0.00002468, -0.00000945, -0.00002338, -0.00005182, 0.00002479, -0.00002586, 0.00003010, 0.00016293, 0.00007422, 0.00006925, -0.00006599, -0.00005694, -0.00001892, -0.00001437, -0.00000231, -0.00001941, 0.00000850, -0.00000682, -0.00000393, 0.00001042, -0.00001102, 0.00000557, -0.00000339, -0.00000329, 0.00000157, 0.00000426, -0.00000607, -0.00000672, 0.00000302, 0.00000156, 0.00000051, -0.00000524, 0.00000099, -0.00000046, -0.00000071, 0.00000447, -0.00000426, 0.00000002, -0.00000176, -0.00000239, 0.00000297, 0.00000429, -0.00000155, -0.00000269, -0.00000072, 0.00000066, -0.00000026, -0.00000140, 0.00000176, -0.00000024, -0.00000033, 0.00000007, -0.00000164, 0.00000075, 0.00000021, -0.00000026, 0.00000010, 0.00000133, -0.00000222, -0.00000134, 0.00000033, 0.00000095, 0.00000069, -0.00000164, 0.00000050, 0.00000010, 0.00000087, 0.00000102, 0.00000000, 0.00000047, -0.00000190, },
		{ -0.00000170, -0.00000141, 0.00000004, 0.00000098, -0.00000017, 0.00000057, -0.00000036, 0.00000002, 0.00000044, -0.00000003, 0.00000180, -0.00000044, -0.00000158, -0.00000005, -0.00000065, 0.00000049, -0.00000045, -0.00000037, -0.00000223, 0.00000147, 0.00000061, -0.00000010, 0.00000122, -0.00000058, 0.00000153, -0.00000327, -0.00000024, -0.00000046, -0.00000340, 0.00000470, -0.00000053, 0.00000151, -0.00000043, -0.00000313, -0.00000309, 0.00000377, 0.00000403, -0.00000160, -0.00000228, -0.00000334, 0.00000081, -0.00000334, 0.00000154, 0.00000170, -0.00000723, 0.00000514, -0.00000343, -0.00000013, 0.00000302, -0.00000569, -0.00000922, 0.00000896, 0.00000380, -0.00000308, -0.00000794, -0.00000466, 0.00000565, -0.00002539, -0.00000918, -0.00002433, -0.00004863, 0.00002692, -0.00003126, 0.00002628, 0.00016514, 0.00007829, 0.00007509, -0.00007637, -0.00005269, -0.00002147, -0.00001266, -0.00000201, -0.00002058, 0.00000958, -0.00000732, -0.00000322, 0.00000986, -0.00001167, 0.00000669, -0.00000377, -0.00000333, 0.00000157, 0.00000480, -0.00000722, -0.00000609, 0.00000277, 0.00000175, 0.00000062, -0.00000551, 0.00000125, -0.00000035, -0.00000068, 0.00000433, -0.00000470, 0.00000036, -0.00000177, -0.00000250, 0.00000322, 0.00000439, -0.00000201, -0.00000244, -0.00000091, 0.00000094, -0.00000031, -0.00000139, 0.00000174, -0.00000029, -0.00000031, -0.00000011, -0.00000163, 0.00000085, 0.00000032, -0.00000033, 0.00000013, 0.00000146, -0.00000249, -0.00000111, 0.00000019, 0.00000102, 0.00000064, -0.00000175, 0.00000061, 0.00000004, 0.00000101, 0.00000098, -0.00000001, 0.00000049, -0.00000199, },
		{ -0.00000160, -0.00000149, 0.00000011, 0.00000098, -0.00000028, 0.00000064, -0.00000049, 0.00000016, 0.00000043, 0.00000001, 0.00000182, -0.00000055, -0.00000145, -0.00000009, -0.00000067, 0.00000046, -0.00000047, -0.00000040, -0.00000245, 0.00000190, 0.00000035, 0.00000017, 0.00000110, -0.00000063, 0.00000165, -0.00000352, -0.00000015, -0.00000058, -0.00000321, 0.00000494, -0.00000094, 0.00000165, -0.00000051, -0.00000311, -0.00000326, 0.00000447, 0.00000361, -0.00000145, -0.00000258, -0.00000339, 0.00000100, -0.00000342, 0.00000142, 0.00000170, -0.00000708, 0.00000568, -0.00000405, -0.00000003, 0.00000337, -0.00000626, -0.00000967, 0.00001052, 0.00000237, -0.00000192, -0.00000924, -0.00000397, 0.00000570, -0.00002573, -0.00000920, -0.00002502, -0.00004548, 0.00002868, -0.00003640, 0.00002231, 0.00016755, 0.00008283, 0.00008022, -0.00008685, -0.00004789, -0.00002449, -0.00001053, -0.00000211, -0.00002141, 0.00001046, -0.00000775, -0.00000252, 0.00000922, -0.00001220, 0.00000775, -0.00000418, -0.00000331, 0.00000153, 0.00000533, -0.00000841, -0.00000535, 0.00000245, 0.00000200, 0.00000067, -0.00000573, 0.00000148, -0.00000021, -0.00000067, 0.00000419, -0.00000514, 0.00000072, -0.00000182, -0.00000257, 0.00000347, 0.00000447, -0.00000245, -0.00000217, -0.00000111, 0.00000124, -0.00000039, -0.00000134, 0.00000168, -0.00000030, -0.00000033, -0.00000027, -0.00000161, 0.00000095, 0.00000043, -0.00000041, 0.00000017, 0.00000156, -0.00000277, -0.00000087, 0.00000002, 0.00000113, 0.00000056, -0.00000183, 0.00000070, -0.00000001, 0.00000116, 0.00000092, -0.00000001, 0.00000051, -0.00000207, },
		{ -0.00000150, -0.00000157, 0.00000019, 0.00000096, -0.00000037, 0.00000069, -0.00000062, 0.00000032, 0.00000040, 0.00000008, 0.00000181, -0.00000064, -0.00000132, -0.00000013, -0.00000068, 0.00000042, -0.00000048, -0.00000044, -0.00000265, 0.00000235, 0.00000006, 0.00000049, 0.00000093, -0.00000062, 0.00000172, -0.00000375, -0.00000008, -0.00000070, -0.00000301, 0.00000514, -0.00000135, 0.00000181, -0.00000061, -0.00000308, -0.00000342, 0.00000520, 0.00000314, -0.00000127, -0.00000292, -0.00000340, 0.00000114, -0.00000345, 0.00000126, 0.00000173, -0.00000695, 0.00000622, -0.00000470, 0.00000014, 0.00000365, -0.00000683, -0.00001002, 0.00001203, 0.00000088, -0.00000066, -0.00001068, -0.00000306, 0.00000540, -0.00002570, -0.00000954, -0.00002537, -0.00004245, 0.00003014, -0.00004131, 0.00001820, 0.00017014, 0.00008790, 0.00008452, -0.00009730, -0.00004262, -0.00002794, -0.00000799, -0.00000260, -0.00002190, 0.00001112, -0.00000807, -0.00000185, 0.00000853, -0.00001263, 0.00000876, -0.00000463, -0.00000322, 0.00000146, 0.00000584, -0.00000963, -0.00000452, 0.00000205, 0.00000231, 0.00000066, -0.00000588, 0.00000167, -0.00000006, -0.00000070, 0.00000405, -0.00000557, 0.00000110, -0.00000191, -0.00000260, 0.00000371, 0.00000452, -0.00000288, -0.00000189, -0.00000131, 0.00000155, -0.00000051, -0.00000125, 0.00000156, -0.00000027, -0.00000037, -0.00000040, -0.00000159, 0.00000104, 0.00000055, -0.00000049, 0.00000024, 0.00000164, -0.00000303, -0.00000062, -0.00000018, 0.00000126, 0.00000045, -0.00000188, 0.00000078, -0.00000006, 0.00000131, 0.00000084, 0.00000001, 0.00000050, -0.00000214, },
		{ -0.00000141, -0.00000166, 0.00000029, 0.00000091, -0.00000044, 0.00000074, -0.00000076, 0.00000049, 0.00000035, 0.00000018, 0.00000177, -0.00000071, -0.00000120, -0.00000019, -0.00000067, 0.00000036, -0.00000048, -0.00000049, -0.00000283, 0.00000279, -0.00000027, 0.00000085, 0.00000072, -0.00000057, 0.00000174, -0.00000393, -0.00000001, -0.00000082, -0.00000278, 0.00000531, -0.00000174, 0.00000198, -0.00000075, -0.00000301, -0.00000358, 0.00000593, 0.00000261, -0.00000104, -0.00000329, -0.00000336, 0.00000123, -0.00000343, 0.00000106, 0.00000182, -0.00000685, 0.00000676, -0.00000537, 0.00000038, 0.00000385, -0.00000740, -0.00001027, 0.00001346, -0.00000063, 0.00000065, -0.00001220, -0.00000193, 0.00000478, -0.00002530, -0.00001022, -0.00002536, -0.00003961, 0.00003136, -0.00004599, 0.00001394, 0.00017288, 0.00009358, 0.00008784, -0.00010757, -0.00003698, -0.00003176, -0.00000508, -0.00000348, -0.00002202, 0.00001153, -0.00000828, -0.00000122, 0.00000779, -0.00001294, 0.00000970, -0.00000509, -0.00000307, 0.00000136, 0.00000634, -0.00001087, -0.00000360, 0.00000157, 0.00000267, 0.00000060, -0.00000598, 0.00000182, 0.00000011, -0.00000074, 0.00000392, -0.00000600, 0.00000150, -0.00000203, -0.00000258, 0.00000394, 0.00000453, -0.00000329, -0.00000160, -0.00000152, 0.00000187, -0.00000064, -0.00000113, 0.00000140, -0.00000020, -0.00000045, -0.00000052, -0.00000157, 0.00000113, 0.00000067, -0.00000059, 0.00000032, 0.00000168, -0.00000328, -0.00000036, -0.00000039, 0.00000141, 0.00000032, -0.00000190, 0.00000085, -0.00000010, 0.00000147, 0.00000075, 0.00000005, 0.00000048, -0.00000220, },
		{ -0.00000132, -0.00000175, 0.00000040, 0.00000085, -0.00000050, 0.00000078, -0.00000088, 0.00000067, 0.00000027, 0.00000030, 0.00000170, -0.00000076, -0.00000110, -0.00000025, -0.00000065, 0.00000029, -0.00000046, -0.00000057, -0.00000298, 0.00000323, -0.00000061, 0.00000124, 0.00000046, -0.00000048, 0.00000171, -0.00000409, 0.00000005, -0.00000094, -0.00000254, 0.00000543, -0.00000210, 0.00000216, -0.00000091, -0.00000292, -0.00000373, 0.00000666, 0.00000202, -0.00000077, -0.00000368, -0.00000328, 0.00000127, -0.00000336, 0.00000083, 0.00000194, -0.00000677, 0.00000730, -0.00000607, 0.00000068, 0.00000396, -0.00000797, -0.00001040, 0.00001480, -0.00000216, 0.00000201, -0.00001379, -0.00000064, 0.00000384, -0.00002453, -0.00001124, -0.00002496, -0.00003702, 0.00003239, -0.00005048, 0.00000955, 0.00017577, 0.00009992, 0.00009004, -0.00011750, -0.00003108, -0.00003589, -0.00000183, -0.00000474, -0.00002176, 0.00001167, -0.00000836, -0.00000065, 0.00000702, -0.00001315, 0.00001056, -0.00000555, -0.00000289, 0.00000125, 0.00000680, -0.00001212, -0.00000260, 0.00000102, 0.00000309, 0.00000049, -0.00000601, 0.00000194, 0.00000029, -0.00000081, 0.00000380, -0.00000642, 0.00000191, -0.00000218, -0.00000252, 0.00000415, 0.00000452, -0.00000368, -0.00000132, -0.00000173, 0.00000218, -0.00000080, -0.00000097, 0.00000121, -0.00000010, -0.00000057, -0.00000060, -0.00000156, 0.00000122, 0.00000079, -0.00000070, 0.00000042, 0.00000170, -0.00000351, -0.00000010, -0.00000062, 0.00000158, 0.00000015, -0.00000189, 0.00000089, -0.00000012, 0.00000163, 0.00000064, 0.00000010, 0.00000044, -0.00000225, },
		{ -0.00000124, -0.00000183, 0.00000052, 0.00000077, -0.00000054, 0.00000081, -0.00000100, 0.00000086, 0.00000018, 0.00000045, 0.00000160, -0.00000078, -0.00000100, -0.00000032, -0.00000061, 0.00000020, -0.00000044, -0.00000066, -0.00000309, 0.00000365, -0.00000097, 0.00000166, 0.00000017, -0.00000034, 0.00000163, -0.00000419, 0.00000009, -0.00000105, -0.00000228, 0.00000552, -0.00000243, 0.00000233, -0.00000110, -0.00000282, -0.00000387, 0.00000740, 0.00000140, -0.00000046, -0.00000411, -0.00000316, 0.00000126, -0.00000325, 0.00000058, 0.00000210, -0.00000671, 0.00000784, -0.00000678, 0.00000104, 0.00000397, -0.00000852, -0.00001041, 0.00001603, -0.00000366, 0.00000337, -0.00001540, 0.00000080, 0.00000262, -0.00002340, -0.00001259, -0.00002416, -0.00003474, 0.00003329, -0.00005478, 0.00000501, 0.00017879, 0.00010697, 0.00009100, -0.00012693, -0.00002504, -0.00004025, 0.00000169, -0.00000634, -0.00002114, 0.00001155, -0.00000828, -0.00000016, 0.00000623, -0.00001324, 0.00001131, -0.00000600, -0.00000266, 0.00000113, 0.00000721, -0.00001335, -0.00000154, 0.00000039, 0.00000355, 0.00000032, -0.00000599, 0.00000201, 0.00000048, -0.00000090, 0.00000367, -0.00000683, 0.00000234, -0.00000237, -0.00000241, 0.00000434, 0.00000448, -0.00000404, -0.00000103, -0.00000192, 0.00000249, -0.00000097, -0.00000079, 0.00000097, 0.00000004, -0.00000072, -0.00000065, -0.00000155, 0.00000132, 0.00000091, -0.00000081, 0.00000054, 0.00000168, -0.00000371, 0.00000016, -0.00000085, 0.00000177, -0.00000004, -0.00000184, 0.00000090, -0.00000013, 0.00000178, 0.00000053, 0.00000017, 0.00000038, -0.00000227, },
		{ -0.00000118, -0.00000192, 0.00000064, 0.00000067, -0.00000056, 0.00000082, -0.00000110, 0.00000105, 0.00000007, 0.00000062, 0.00000147, -0.00000078, -0.00000093, -0.00000038, -0.00000057, 0.00000010, -0.00000040, -0.00000078, -0.00000316, 0.00000405, -0.00000133, 0.00000210, -0.00000016, -0.00000016, 0.00000148, -0.00000426, 0.00000011, -0.00000115, -0.00000200, 0.00000556, -0.00000273, 0.00000250, -0.00000130, -0.00000270, -0.00000398, 0.00000811, 0.00000073, -0.00000012, -0.00000455, -0.00000301, 0.00000120, -0.00000311, 0.00000030, 0.00000229, -0.00000667, 0.00000836, -0.00000749, 0.00000145, 0.00000389, -0.00000905, -0.00001031, 0.00001713, -0.00000512, 0.00000472, -0.00001699, 0.00000232, 0.00000116, -0.00002196, -0.00001425, -0.00002295, -0.00003279, 0.00003411, -0.00005890, 0.00000033, 0.00018196, 0.00011477, 0.00009058, -0.00013571, -0.00001897, -0.00004474, 0.00000543, -0.00000825, -0.00002015, 0.00001114, -0.00000806, 0.00000024, 0.00000544, -0.00001323, 0.00001196, -0.00000642, -0.00000241, 0.00000100, 0.00000756, -0.00001456, -0.00000042, -0.00000028, 0.00000406, 0.00000011, -0.00000591, 0.00000205, 0.00000068, -0.00000100, 0.00000355, -0.00000722, 0.00000276, -0.00000258, -0.00000225, 0.00000451, 0.00000440, -0.00000436, -0.00000076, -0.00000211, 0.00000279, -0.00000114, -0.00000059, 0.00000070, 0.00000020, -0.00000090, -0.00000067, -0.00000155, 0.00000142, 0.00000103, -0.00000092, 0.00000068, 0.00000162, -0.00000388, 0.00000040, -0.00000109, 0.00000197, -0.00000026, -0.00000176, 0.00000089, -0.00000012, 0.00000193, 0.00000040, 0.00000026, 0.00000029, -0.00000228, },
		{ -0.00000113, -0.00000200, 0.00000077, 0.00000055, -0.00000056, 0.00000081, -0.00000118, 0.00000123, -0.00000006, 0.00000081, 0.00000132, -0.00000074, -0.00000088, -0.00000044, -0.00000052, -0.00000001, -0.00000035, -0.00000092, -0.00000319, 0.00000441, -0.00000169, 0.00000255, -0.00000052, 0.00000006, 0.00000129, -0.00000428, 0.00000011, -0.00000124, -0.00000172, 0.00000556, -0.00000298, 0.00000266, -0.00000152, -0.00000256, -0.00000406, 0.00000881, 0.00000004, 0.00000025, -0.00000500, -0.00000284, 0.00000111, -0.00000293, 0.00000002, 0.00000250, -0.00000665, 0.00000887, -0.00000820, 0.00000191, 0.00000372, -0.00000955, -0.00001009, 0.00001808, -0.00000651, 0.00000601, -0.00001853, 0.00000388, -0.00000050, -0.00002023, -0.00001619, -0.00002135, -0.00003121, 0.00003487, -0.00006285, -0.00000452, 0.00018530, 0.00012333, 0.00008868, -0.00014366, -0.00001302, -0.00004928, 0.00000930, -0.00001042, -0.00001884, 0.00001046, -0.00000767, 0.00000055, 0.00000465, -0.00001311, 0.00001249, -0.00000680, -0.00000215, 0.00000089, 0.00000784, -0.00001571, 0.00000073, -0.00000101, 0.00000460, -0.00000014, -0.00000577, 0.00000206, 0.00000088, -0.00000111, 0.00000342, -0.00000758, 0.00000319, -0.00000282, -0.00000205, 0.00000465, 0.00000430, -0.00000466, -0.00000050, -0.00000227, 0.00000307, -0.00000131, -0.00000037, 0.00000041, 0.00000040, -0.00000112, -0.00000066, -0.00000156, 0.00000153, 0.00000113, -0.00000104, 0.00000083, 0.00000152, -0.00000401, 0.00000063, -0.00000132, 0.00000217, -0.00000049, -0.00000165, 0.00000086, -0.00000009, 0.00000206, 0.00000027, 0.00000036, 0.00000019, -0.00000226, },
		{ -0.00000110, -0.00000207, 0.00000090, 0.00000042, -0.00000054, 0.00000078, -0.00000124, 0.00000141, -0.00000019, 0.00000102, 0.00000113, -0.00000068, -0.00000085, -0.00000049, -0.00000046, -0.00000014, -0.00000028, -0.00000108, -0.00000316, 0.00000474, -0.00000203, 0.00000301, -0.00000090, 0.00000031, 0.00000104, -0.00000425, 0.00000009, -0.00000132, -0.00000143, 0.00000552, -0.00000319, 0.00000279, -0.00000173, -0.00000243, -0.00000411, 0.00000947, -0.00000068, 0.00000063, -0.00000546, -0.00000264, 0.00000099, -0.00000273, -0.00000027, 0.00000273, -0.00000664, 0.00000935, -0.00000889, 0.00000241, 0.00000345, -0.00001002, -0.00000975, 0.00001887, -0.00000780, 0.00000722, -0.00001996, 0.00000543, -0.00000230, -0.00001826, -0.00001837, -0.00001939, -0.00002999, 0.00003560, -0.00006662, -0.00000955, 0.00018882, 0.00013265, 0.00008521, -0.00015064, -0.00000731, -0.00005375, 0.00001322, -0.00001278, -0.00001722, 0.00000952, -0.00000714, 0.00000076, 0.00000387, -0.00001289, 0.00001289, -0.00000711, -0.00000189, 0.00000079, 0.00000803, -0.00001680, 0.00000190, -0.00000177, 0.00000516, -0.00000042, -0.00000559, 0.00000204, 0.00000107, -0.00000123, 0.00000328, -0.00000792, 0.00000359, -0.00000306, -0.00000180, 0.00000477, 0.00000417, -0.00000491, -0.00000025, -0.00000241, 0.00000332, -0.00000147, -0.00000016, 0.00000010, 0.00000061, -0.00000136, -0.00000061, -0.00000159, 0.00000164, 0.00000123, -0.00000116, 0.00000100, 0.00000139, -0.00000410, 0.00000083, -0.00000155, 0.00000237, -0.00000074, -0.00000151, 0.00000080, -0.00000004, 0.00000218, 0.00000014, 0.00000047, 0.00000007, -0.00000222, },
		{ -0.00000109, -0.00000212, 0.00000102, 0.00000027, -0.00000050, 0.00000073, -0.00000127, 0.00000157, -0.00000032, 0.00000124, 0.00000093, -0.00000059, -0.00000085, -0.00000052, -0.00000040, -0.00000027, -0.00000021, -0.00000127, -0.00000308, 0.00000502, -0.00000236, 0.00000346, -0.00000130, 0.00000059, 0.00000075, -0.00000417, 0.00000005, -0.00000138, -0.00000113, 0.00000543, -0.00000334, 0.00000288, -0.00000194, -0.00000230, -0.00000412, 0.00001008, -0.00000140, 0.00000103, -0.00000592, -0.00000243, 0.00000085, -0.00000252, -0.00000054, 0.00000295, -0.00000662, 0.00000979, -0.00000956, 0.00000293, 0.00000309, -0.00001046, -0.00000928, 0.00001948, -0.00000897, 0.00000831, -0.00002125, 0.00000691, -0.00000419, -0.00001612, -0.00002074, -0.00001711, -0.00002914, 0.00003631, -0.00007021, -0.00001480, 0.00019257, 0.00014273, 0.00008008, -0.00015650, -0.00000199, -0.00005804, 0.00001710, -0.00001527, -0.00001534, 0.00000834, -0.00000645, 0.00000088, 0.00000311, -0.00001257, 0.00001315, -0.00000734, -0.00000165, 0.00000073, 0.00000811, -0.00001779, 0.00000307, -0.00000254, 0.00000573, -0.00000073, -0.00000537, 0.00000200, 0.00000125, -0.00000134, 0.00000312, -0.00000822, 0.00000398, -0.00000332, -0.00000152, 0.00000487, 0.00000401, -0.00000513, -0.00000003, -0.00000253, 0.00000353, -0.00000160, 0.00000005, -0.00000023, 0.00000083, -0.00000163, -0.00000053, -0.00000163, 0.00000177, 0.00000132, -0.00000127, 0.00000118, 0.00000121, -0.00000415, 0.00000100, -0.00000176, 0.00000257, -0.00000100, -0.00000135, 0.00000071, 0.00000004, 0.00000228, 0.00000001, 0.00000059, -0.00000008, -0.00000216, },
		{ -0.00000110, -0.00000216, 0.00000115, 0.00000011, -0.00000044, 0.00000066, -0.00000128, 0.00000172, -0.00000044, 0.00000147, 0.00000070, -0.00000047, -0.00000088, -0.00000055, -0.00000034, -0.00000041, -0.00000012, -0.00000148, -0.00000295, 0.00000525, -0.00000265, 0.00000390, -0.00000169, 0.00000088, 0.00000042, -0.00000406, -0.00000002, -0.00000143, -0.00000082, 0.00000530, -0.00000344, 0.00000294, -0.00000214, -0.00000217, -0.00000407, 0.00001064, -0.00000212, 0.00000143, -0.00000636, -0.00000222, 0.00000069, -0.00000229, -0.00000080, 0.00000316, -0.00000658, 0.00001019, -0.00001018, 0.00000346, 0.00000266, -0.00001085, -0.00000869, 0.00001989, -0.00001000, 0.00000925, -0.00002235, 0.00000827, -0.00000610, -0.00001386, -0.00002322, -0.00001456, -0.00002863, 0.00003701, -0.00007359, -0.00002031, 0.00019661, 0.00015353, 0.00007323, -0.00016111, 0.00000282, -0.00006206, 0.00002086, -0.00001781, -0.00001325, 0.00000694, -0.00000564, 0.00000090, 0.00000237, -0.00001216, 0.00001325, -0.00000748, -0.00000144, 0.00000072, 0.00000806, -0.00001867, 0.00000421, -0.00000332, 0.00000630, -0.00000104, -0.00000511, 0.00000194, 0.00000142, -0.00000144, 0.00000294, -0.00000848, 0.00000434, -0.00000357, -0.00000120, 0.00000493, 0.00000382, -0.00000530, 0.00000017, -0.00000261, 0.00000371, -0.00000171, 0.00000024, -0.00000055, 0.00000106, -0.00000191, -0.00000043, -0.00000168, 0.00000190, 0.00000139, -0.00000138, 0.00000137, 0.00000100, -0.00000416, 0.00000114, -0.00000195, 0.00000276, -0.00000126, -0.00000116, 0.00000060, 0.00000014, 0.00000237, -0.00000011, 0.00000072, -0.00000024, -0.00000208, },
		{ -0.00000114, -0.00000218, 0.00000126, -0.00000005, -0.00000037, 0.00000056, -0.00000125, 0.00000185, -0.00000056, 0.00000170, 0.00000046, -0.00000033, -0.00000093, -0.00000056, -0.00000029, -0.00000056, -0.00000003, -0.00000171, -0.00000276, 0.00000541, -0.00000290, 0.00000431, -0.00000208, 0.00000119, 0.00000005, -0.00000390, -0.00000010, -0.00000146, -0.00000051, 0.00000512, -0.00000348, 0.00000295, -0.00000232, -0.00000206, -0.00000397, 0.00001113, -0.00000282, 0.00000181, -0.00000678, -0.00000203, 0.00000053, -0.00000207, -0.00000103, 0.00000335, -0.00000653, 0.00001053, -0.00001074, 0.00000398, 0.00000215, -0.00001121, -0.00000798, 0.00002010, -0.00001086, 0.00001002, -0.00002323, 0.00000945, -0.00000797, -0.00001154, -0.00002576, -0.00001181, -0.00002844, 0.00003767, -0.00007673, -0.00002612, 0.00020101, 0.00016498, 0.00006464, -0.00016435, 0.00000699, -0.00006568, 0.00002438, -0.00002033, -0.00001101, 0.00000536, -0.00000471, 0.00000083, 0.00000166, -0.00001165, 0.00001321, -0.00000751, -0.00000127, 0.00000076, 0.00000789, -0.00001941, 0.00000531, -0.00000409, 0.00000685, -0.00000136, -0.00000483, 0.00000187, 0.00000156, -0.00000153, 0.00000274, -0.00000869, 0.00000466, -0.00000381, -0.00000085, 0.00000497, 0.00000361, -0.00000543, 0.00000034, -0.00000266, 0.00000383, -0.00000178, 0.00000042, -0.00000086, 0.00000129, -0.00000220, -0.00000030, -0.00000174, 0.00000204, 0.00000146, -0.00000147, 0.00000157, 0.00000075, -0.00000411, 0.00000124, -0.00000211, 0.00000293, -0.00000153, -0.00000095, 0.00000047, 0.00000027, 0.00000243, -0.00000022, 0.00000084, -0.00000041, -0.00000197, },
		{ -0.00000119, -0.00000217, 0.00000136, -0.00000021, -0.00000028, 0.00000044, -0.00000120, 0.00000194, -0.00000067, 0.00000192, 0.00000021, -0.00000017, -0.00000101, -0.00000055, -0.00000024, -0.00000071, 0.00000008, -0.00000197, -0.00000250, 0.00000551, -0.00000311, 0.00000468, -0.00000245, 0.00000149, -0.00000035, -0.00000370, -0.00000021, -0.00000147, -0.00000020, 0.00000491, -0.00000345, 0.00000291, -0.00000246, -0.00000197, -0.00000380, 0.00001154, -0.00000349, 0.00000218, -0.00000718, -0.00000185, 0.00000037, -0.00000186, -0.00000121, 0.00000349, -0.00000643, 0.00001080, -0.00001122, 0.00000448, 0.00000158, -0.00001153, -0.00000714, 0.00002010, -0.00001154, 0.00001059, -0.00002384, 0.00001040, -0.00000973, -0.00000925, -0.00002828, -0.00000893, -0.00002851, 0.00003828, -0.00007959, -0.00003230, 0.00020585, 0.00017701, 0.00005429, -0.00016612, 0.00001039, -0.00006880, 0.00002758, -0.00002274, -0.00000868, 0.00000364, -0.00000368, 0.00000069, 0.00000098, -0.00001104, 0.00001300, -0.00000741, -0.00000117, 0.00000086, 0.00000756, -0.00002000, 0.00000634, -0.00000483, 0.00000737, -0.00000166, -0.00000454, 0.00000179, 0.00000167, -0.00000158, 0.00000250, -0.00000883, 0.00000492, -0.00000403, -0.00000048, 0.00000499, 0.00000336, -0.00000551, 0.00000048, -0.00000266, 0.00000391, -0.00000181, 0.00000056, -0.00000116, 0.00000151, -0.00000250, -0.00000015, -0.00000182, 0.00000219, 0.00000150, -0.00000156, 0.00000178, 0.00000046, -0.00000402, 0.00000129, -0.00000223, 0.00000307, -0.00000178, -0.00000073, 0.00000032, 0.00000042, 0.00000246, -0.00000032, 0.00000097, -0.00000060, -0.00000184, },
		{ -0.00000128, -0.00000215, 0.00000144, -0.00000038, -0.00000018, 0.00000031, -0.00000110, 0.00000201, -0.00000075, 0.00000213, -0.00000004, 0.00000001, -0.00000111, -0.00000053, -0.00000019, -0.00000087, 0.00000018, -0.00000224, -0.00000219, 0.00000554, -0.00000326, 0.00000501, -0.00000279, 0.00000178, -0.00000076, -0.00000347, -0.00000033, -0.00000147, 0.00000011, 0.00000465, -0.00000335, 0.00000281, -0.00000258, -0.00000191, -0.00000357, 0.00001186, -0.00000412, 0.00000252, -0.00000753, -0.00000169, 0.00000023, -0.00000167, -0.00000134, 0.00000359, -0.00000629, 0.00001099, -0.00001161, 0.00000493, 0.00000096, -0.00001181, -0.00000619, 0.00001988, -0.00001201, 0.00001093, -0.00002415, 0.00001106, -0.00001132, -0.00000705, -0.00003069, -0.00000600, -0.00002879, 0.00003880, -0.00008214, -0.00003891, 0.00021123, 0.00018951, 0.00004219, -0.00016633, 0.00001294, -0.00007131, 0.00003036, -0.00002496, -0.00000631, 0.00000181, -0.00000259, 0.00000048, 0.00000033, -0.00001035, 0.00001262, -0.00000718, -0.00000113, 0.00000103, 0.00000708, -0.00002041, 0.00000728, -0.00000552, 0.00000785, -0.00000194, -0.00000424, 0.00000172, 0.00000174, -0.00000161, 0.00000223, -0.00000892, 0.00000513, -0.00000421, -0.00000009, 0.00000499, 0.00000309, -0.00000554, 0.00000059, -0.00000263, 0.00000393, -0.00000179, 0.00000066, -0.00000144, 0.00000170, -0.00000278, 0.00000002, -0.00000190, 0.00000235, 0.00000154, -0.00000163, 0.00000198, 0.00000014, -0.00000387, 0.00000131, -0.00000232, 0.00000318, -0.00000202, -0.00000050, 0.00000015, 0.00000060, 0.00000248, -0.00000041, 0.00000109, -0.00000079, -0.00000170, },
		{ -0.00000138, -0.00000210, 0.00000150, -0.00000055, -0.00000006, 0.00000015, -0.00000098, 0.00000205, -0.00000081, 0.00000233, -0.00000029, 0.00000019, -0.00000123, -0.00000049, -0.00000015, -0.00000102, 0.00000029, -0.00000253, -0.00000183, 0.00000550, -0.00000334, 0.00000529, -0.00000309, 0.00000205, -0.00000119, -0.00000322, -0.00000047, -0.00000146, 0.00000042, 0.00000435, -0.00000319, 0.00000265, -0.00000265, -0.00000188, -0.00000325, 0.00001208, -0.00000470, 0.00000282, -0.00000783, -0.00000158, 0.00000011, -0.00000152, -0.00000140, 0.00000362, -0.00000609, 0.00001108, -0.00001189, 0.00000532, 0.00000030, -0.00001205, -0.00000513, 0.00001943, -0.00001227, 0.00001104, -0.00002413, 0.00001139, -0.00001267, -0.00000501, -0.00003292, -0.00000312, -0.00002921, 0.00003920, -0.00008432, -0.00004603, 0.00021727, 0.00020236, 0.00002839, -0.00016493, 0.00001452, -0.00007314, 0.00003263, -0.00002690, -0.00000399, -0.00000007, -0.00000147, 0.00000023, -0.00000030, -0.00000957, 0.00001208, -0.00000681, -0.00000116, 0.00000129, 0.00000642, -0.00002063, 0.00000811, -0.00000614, 0.00000827, -0.00000218, -0.00000395, 0.00000165, 0.00000176, -0.00000159, 0.00000191, -0.00000893, 0.00000527, -0.00000436, 0.00000030, 0.00000497, 0.00000279, -0.00000552, 0.00000065, -0.00000255, 0.00000389, -0.00000171, 0.00000071, -0.00000168, 0.00000187, -0.00000306, 0.00000019, -0.00000198, 0.00000251, 0.00000156, -0.00000169, 0.00000217, -0.00000021, -0.00000368, 0.00000127, -0.00000237, 0.00000326, -0.00000223, -0.00000027, -0.00000004, 0.00000079, 0.00000246, -0.00000048, 0.00000120, -0.00000099, -0.00000153, },
		{ -0.00000150, -0.00000202, 0.00000155, -0.00000070, 0.00000006, -0.00000003, -0.00000082, 0.00000204, -0.00000084, 0.00000251, -0.00000054, 0.00000039, -0.00000137, -0.00000043, -0.00000012, -0.00000118, 0.00000040, -0.00000283, -0.00000140, 0.00000537, -0.00000336, 0.00000550, -0.00000335, 0.00000229, -0.00000162, -0.00000295, -0.00000061, -0.00000144, 0.00000073, 0.00000401, -0.00000296, 0.00000242, -0.00000267, -0.00000189, -0.00000285, 0.00001219, -0.00000521, 0.00000307, -0.00000806, -0.00000151, 0.00000003, -0.00000140, -0.00000140, 0.00000358, -0.00000581, 0.00001106, -0.00001205, 0.00000562, -0.00000037, -0.00001225, -0.00000395, 0.00001876, -0.00001231, 0.00001088, -0.00002376, 0.00001135, -0.00001373, -0.00000320, -0.00003487, -0.00000037, -0.00002970, 0.00003941, -0.00008606, -0.00005374, 0.00022409, 0.00021542, 0.00001296, -0.00016187, 0.00001507, -0.00007419, 0.00003430, -0.00002849, -0.00000177, -0.00000196, -0.00000034, -0.00000006, -0.00000090, -0.00000872, 0.00001136, -0.00000628, -0.00000129, 0.00000163, 0.00000559, -0.00002064, 0.00000880, -0.00000667, 0.00000861, -0.00000237, -0.00000367, 0.00000160, 0.00000174, -0.00000153, 0.00000153, -0.00000886, 0.00000533, -0.00000445, 0.00000069, 0.00000493, 0.00000246, -0.00000544, 0.00000069, -0.00000243, 0.00000379, -0.00000157, 0.00000072, -0.00000188, 0.00000200, -0.00000331, 0.00000037, -0.00000207, 0.00000267, 0.00000157, -0.00000173, 0.00000236, -0.00000059, -0.00000343, 0.00000119, -0.00000237, 0.00000329, -0.00000242, -0.00000004, -0.00000023, 0.00000101, 0.00000242, -0.00000053, 0.00000130, -0.00000119, -0.00000136, },
		{ -0.00000165, -0.00000193, 0.00000157, -0.00000085, 0.00000018, -0.00000022, -0.00000062, 0.00000201, -0.00000084, 0.00000265, -0.00000077, 0.00000058, -0.00000152, -0.00000036, -0.00000010, -0.00000133, 0.00000051, -0.00000314, -0.00000092, 0.00000517, -0.00000330, 0.00000564, -0.00000355, 0.00000249, -0.00000204, -0.00000266, -0.00000076, -0.00000141, 0.00000103, 0.00000364, -0.00000266, 0.00000211, -0.00000264, -0.00000194, -0.00000236, 0.00001217, -0.00000564, 0.00000326, -0.00000823, -0.00000150, -0.00000002, -0.00000132, -0.00000131, 0.00000344, -0.00000545, 0.00001092, -0.00001206, 0.00000582, -0.00000105, -0.00001242, -0.00000266, 0.00001785, -0.00001211, 0.00001047, -0.00002302, 0.00001090, -0.00001444, -0.00000170, -0.00003648, 0.00000215, -0.00003017, 0.00003938, -0.00008731, -0.00006212, 0.00023182, 0.00022850, -0.00000401, -0.00015716, 0.00001453, -0.00007439, 0.00003530, -0.00002963, 0.00000027, -0.00000381, 0.00000075, -0.00000035, -0.00000149, -0.00000778, 0.00001048, -0.00000560, -0.00000152, 0.00000206, 0.00000458, -0.00002043, 0.00000935, -0.00000711, 0.00000887, -0.00000249, -0.00000342, 0.00000157, 0.00000165, -0.00000141, 0.00000110, -0.00000871, 0.00000531, -0.00000447, 0.00000107, 0.00000487, 0.00000210, -0.00000531, 0.00000068, -0.00000226, 0.00000362, -0.00000137, 0.00000066, -0.00000203, 0.00000208, -0.00000352, 0.00000054, -0.00000215, 0.00000283, 0.00000157, -0.00000175, 0.00000254, -0.00000099, -0.00000315, 0.00000105, -0.00000233, 0.00000328, -0.00000258, 0.00000018, -0.00000042, 0.00000124, 0.00000236, -0.00000056, 0.00000139, -0.00000139, -0.00000117, },
		{ -0.00000181, -0.00000180, 0.00000156, -0.00000099, 0.00000030, -0.00000042, -0.00000039, 0.00000193, -0.00000080, 0.00000277, -0.00000097, 0.00000077, -0.00000168, -0.00000027, -0.00000009, -0.00000147, 0.00000061, -0.00000346, -0.00000040, 0.00000490, -0.00000316, 0.00000571, -0.00000369, 0.00000264, -0.00000244, -0.00000237, -0.00000091, -0.00000137, 0.00000133, 0.00000324, -0.00000230, 0.00000173, -0.00000256, -0.00000203, -0.00000179, 0.00001203, -0.00000599, 0.00000338, -0.00000833, -0.00000156, -0.00000001, -0.00000130, -0.00000112, 0.00000321, -0.00000500, 0.00001065, -0.00001192, 0.00000590, -0.00000171, -0.00001256, -0.00000128, 0.00001673, -0.00001168, 0.00000978, -0.00002190, 0.00001001, -0.00001476, -0.00000056, -0.00003767, 0.00000434, -0.00003055, 0.00003904, -0.00008799, -0.00007127, 0.00024061, 0.00024143, -0.00002238, -0.00015080, 0.00001285, -0.00007370, 0.00003556, -0.00003027, 0.00000206, -0.00000557, 0.00000179, -0.00000063, -0.00000206, -0.00000677, 0.00000943, -0.00000475, -0.00000184, 0.00000258, 0.00000339, -0.00001999, 0.00000974, -0.00000743, 0.00000902, -0.00000254, -0.00000320, 0.00000156, 0.00000151, -0.00000124, 0.00000062, -0.00000847, 0.00000519, -0.00000442, 0.00000144, 0.00000481, 0.00000171, -0.00000513, 0.00000063, -0.00000204, 0.00000340, -0.00000111, 0.00000054, -0.00000213, 0.00000211, -0.00000370, 0.00000070, -0.00000222, 0.00000299, 0.00000155, -0.00000176, 0.00000270, -0.00000140, -0.00000282, 0.00000087, -0.00000224, 0.00000322, -0.00000270, 0.00000038, -0.00000061, 0.00000148, 0.00000227, -0.00000057, 0.00000145, -0.00000158, -0.00000097, },
		{ -0.00000199, -0.00000165, 0.00000153, -0.00000111, 0.00000042, -0.00000063, -0.00000014, 0.00000181, -0.00000073, 0.00000284, -0.00000116, 0.00000094, -0.00000185, -0.00000018, -0.00000009, -0.00000161, 0.00000070, -0.00000377, 0.00000017, 0.00000455, -0.00000295, 0.00000570, -0.00000375, 0.00000272, -0.00000281, -0.00000209, -0.00000105, -0.00000133, 0.00000163, 0.00000281, -0.00000187, 0.00000128, -0.00000241, -0.00000217, -0.00000112, 0.00001176, -0.00000623, 0.00000343, -0.00000833, -0.00000168, 0.00000005, -0.00000134, -0.00000085, 0.00000288, -0.00000445, 0.00001024, -0.00001161, 0.00000584, -0.00000235, -0.00001268, 0.00000019, 0.00001538, -0.00001102, 0.00000883, -0.00002039, 0.00000867, -0.00001465, 0.00000016, -0.00003835, 0.00000612, -0.00003073, 0.00003833, -0.00008804, -0.00008128, 0.00025061, 0.00025400, -0.00004201, -0.00014284, 0.00001003, -0.00007207, 0.00003503, -0.00003034, 0.00000355, -0.00000718, 0.00000271, -0.00000088, -0.00000264, -0.00000570, 0.00000823, -0.00000375, -0.00000227, 0.00000319, 0.00000201, -0.00001930, 0.00000995, -0.00000763, 0.00000906, -0.00000251, -0.00000302, 0.00000159, 0.00000130, -0.00000101, 0.00000007, -0.00000814, 0.00000497, -0.00000430, 0.00000177, 0.00000475, 0.00000129, -0.00000490, 0.00000054, -0.00000179, 0.00000311, -0.00000077, 0.00000036, -0.00000216, 0.00000208, -0.00000383, 0.00000084, -0.00000228, 0.00000314, 0.00000153, -0.00000174, 0.00000284, -0.00000182, -0.00000245, 0.00000064, -0.00000209, 0.00000311, -0.00000277, 0.00000057, -0.00000080, 0.00000173, 0.00000216, -0.00000055, 0.00000150, -0.00000176, -0.00000077, },
		{ -0.00000217, -0.00000149, 0.00000148, -0.00000121, 0.00000054, -0.00000085, 0.00000014, 0.00000165, -0.00000061, 0.00000288, -0.00000131, 0.00000110, -0.00000202, -0.00000007, -0.00000010, -0.00000174, 0.00000078, -0.00000407, 0.00000078, 0.00000412, -0.00000265, 0.00000560, -0.00000373, 0.00000274, -0.00000314, -0.00000181, -0.00000118, -0.00000128, 0.00000191, 0.00000235, -0.00000137, 0.00000075, -0.00000221, -0.00000236, -0.00000036, 0.00001135, -0.00000638, 0.00000339, -0.00000825, -0.00000189, 0.00000017, -0.00000143, -0.00000047, 0.00000243, -0.00000378, 0.00000968, -0.00001112, 0.00000562, -0.00000294, -0.00001277, 0.00000175, 0.00001383, -0.00001012, 0.00000762, -0.00001849, 0.00000685, -0.00001407, 0.00000043, -0.00003848, 0.00000740, -0.00003064, 0.00003719, -0.00008738, -0.00009223, 0.00026195, 0.00026599, -0.00006271, -0.00013335, 0.00000606, -0.00006948, 0.00003367, -0.00002979, 0.00000468, -0.00000860, 0.00000351, -0.00000107, -0.00000321, -0.00000456, 0.00000687, -0.00000258, -0.00000281, 0.00000389, 0.00000046, -0.00001837, 0.00000996, -0.00000768, 0.00000898, -0.00000238, -0.00000290, 0.00000166, 0.00000103, -0.00000071, -0.00000055, -0.00000772, 0.00000466, -0.00000408, 0.00000207, 0.00000468, 0.00000085, -0.00000461, 0.00000042, -0.00000149, 0.00000276, -0.00000038, 0.00000010, -0.00000214, 0.00000198, -0.00000390, 0.00000095, -0.00000231, 0.00000328, 0.00000150, -0.00000170, 0.00000296, -0.00000225, -0.00000205, 0.00000037, -0.00000190, 0.00000295, -0.00000280, 0.00000073, -0.00000098, 0.00000199, 0.00000203, -0.00000051, 0.00000152, -0.00000193, -0.00000057, },
		{ -0.00000237, -0.00000130, 0.00000139, -0.00000129, 0.00000064, -0.00000106, 0.00000045, 0.00000146, -0.00000045, 0.00000287, -0.00000142, 0.00000123, -0.00000218, 0.00000004, -0.00000011, -0.00000186, 0.00000085, -0.00000437, 0.00000141, 0.00000363, -0.00000228, 0.00000542, -0.00000364, 0.00000269, -0.00000343, -0.00000156, -0.00000130, -0.00000124, 0.00000220, 0.00000187, -0.00000083, 0.00000016, -0.00000194, -0.00000259, 0.00000048, 0.00001080, -0.00000641, 0.00000326, -0.00000807, -0.00000217, 0.00000036, -0.00000159, 0.00000000, 0.00000187, -0.00000301, 0.00000898, -0.00001044, 0.00000523, -0.00000347, -0.00001284, 0.00000337, 0.00001208, -0.00000902, 0.00000615, -0.00001622, 0.00000457, -0.00001300, 0.00000019, -0.00003800, 0.00000811, -0.00003020, 0.00003554, -0.00008597, -0.00010421, 0.00027481, 0.00027717, -0.00008428, -0.00012243, 0.00000097, -0.00006593, 0.00003146, -0.00002857, 0.00000538, -0.00000979, 0.00000413, -0.00000117, -0.00000380, -0.00000337, 0.00000537, -0.00000125, -0.00000344, 0.00000467, -0.00000126, -0.00001720, 0.00000978, -0.00000758, 0.00000877, -0.00000215, -0.00000283, 0.00000176, 0.00000069, -0.00000034, -0.00000122, -0.00000721, 0.00000423, -0.00000376, 0.00000232, 0.00000461, 0.00000038, -0.00000427, 0.00000026, -0.00000116, 0.00000235, 0.00000008, -0.00000022, -0.00000205, 0.00000182, -0.00000391, 0.00000104, -0.00000232, 0.00000340, 0.00000147, -0.00000164, 0.00000305, -0.00000268, -0.00000162, 0.00000005, -0.00000166, 0.00000273, -0.00000278, 0.00000085, -0.00000113, 0.00000224, 0.00000188, -0.00000045, 0.00000152, -0.00000208, -0.00000038, },
		{ -0.00000257, -0.00000109, 0.00000128, -0.00000134, 0.00000073, -0.00000128, 0.00000078, 0.00000123, -0.00000025, 0.00000282, -0.00000149, 0.00000134, -0.00000233, 0.00000016, -0.00000014, -0.00000197, 0.00000090, -0.00000464, 0.00000207, 0.00000308, -0.00000183, 0.00000515, -0.00000345, 0.00000255, -0.00000366, -0.00000133, -0.00000139, -0.00000121, 0.00000247, 0.00000138, -0.00000023, -0.00000051, -0.00000161, -0.00000287, 0.00000141, 0.00001012, -0.00000633, 0.00000303, -0.00000780, -0.00000254, 0.00000061, -0.00000182, 0.00000058, 0.00000119, -0.00000212, 0.00000813, -0.00000958, 0.00000465, -0.00000393, -0.00001289, 0.00000506, 0.00001016, -0.00000770, 0.00000445, -0.00001359, 0.00000182, -0.00001144, -0.00000056, -0.00003686, 0.00000816, -0.00002933, 0.00003332, -0.00008372, -0.00011730, 0.00028932, 0.00028731, -0.00010650, -0.00011019, -0.00000518, -0.00006143, 0.00002839, -0.00002665, 0.00000564, -0.00001071, 0.00000456, -0.00000117, -0.00000440, -0.00000214, 0.00000373, 0.00000022, -0.00000418, 0.00000551, -0.00000314, -0.00001579, 0.00000940, -0.00000733, 0.00000842, -0.00000181, -0.00000283, 0.00000191, 0.00000028, 0.00000009, -0.00000196, -0.00000660, 0.00000370, -0.00000335, 0.00000252, 0.00000454, -0.00000011, -0.00000389, 0.00000006, -0.00000079, 0.00000190, 0.00000060, -0.00000060, -0.00000189, 0.00000158, -0.00000385, 0.00000108, -0.00000231, 0.00000350, 0.00000144, -0.00000157, 0.00000311, -0.00000310, -0.00000117, -0.00000029, -0.00000138, 0.00000246, -0.00000270, 0.00000094, -0.00000127, 0.00000250, 0.00000172, -0.00000036, 0.00000149, -0.00000221, -0.00000019, },
		{ -0.00000277, -0.00000087, 0.00000115, -0.00000138, 0.00000081, -0.00000148, 0.00000112, 0.00000098, -0.00000002, 0.00000271, -0.00000152, 0.00000141, -0.00000247, 0.00000027, -0.00000016, -0.00000207, 0.00000093, -0.00000489, 0.00000274, 0.00000247, -0.00000131, 0.00000479, -0.00000318, 0.00000232, -0.00000382, -0.00000113, -0.00000146, -0.00000119, 0.00000274, 0.00000087, 0.00000041, -0.00000124, -0.00000122, -0.00000319, 0.00000242, 0.00000930, -0.00000613, 0.00000271, -0.00000742, -0.00000299, 0.00000093, -0.00000210, 0.00000125, 0.00000040, -0.00000112, 0.00000712, -0.00000852, 0.00000389, -0.00000429, -0.00001292, 0.00000679, 0.00000807, -0.00000621, 0.00000254, -0.00001064, -0.00000138, -0.00000939, -0.00000185, -0.00003504, 0.00000752, -0.00002796, 0.00003048, -0.00008059, -0.00013157, 0.00030562, 0.00029616, -0.00012913, -0.00009679, -0.00001232, -0.00005603, 0.00002447, -0.00002401, 0.00000539, -0.00001132, 0.00000476, -0.00000105, -0.00000503, -0.00000086, 0.00000199, 0.00000184, -0.00000500, 0.00000642, -0.00000515, -0.00001415, 0.00000881, -0.00000692, 0.00000793, -0.00000136, -0.00000289, 0.00000209, -0.00000019, 0.00000058, -0.00000276, -0.00000591, 0.00000307, -0.00000283, 0.00000266, 0.00000448, -0.00000063, -0.00000346, -0.00000017, -0.00000039, 0.00000140, 0.00000117, -0.00000105, -0.00000166, 0.00000127, -0.00000372, 0.00000107, -0.00000225, 0.00000357, 0.00000141, -0.00000147, 0.00000314, -0.00000351, -0.00000070, -0.00000067, -0.00000106, 0.00000214, -0.00000257, 0.00000100, -0.00000138, 0.00000274, 0.00000154, -0.00000025, 0.00000143, -0.00000232, -0.00000001, },
		{ -0.00000296, -0.00000064, 0.00000099, -0.00000139, 0.00000086, -0.00000167, 0.00000146, 0.00000069, 0.00000026, 0.00000256, -0.00000150, 0.00000144, -0.00000259, 0.00000038, -0.00000019, -0.00000216, 0.00000094, -0.00000512, 0.00000342, 0.00000182, -0.00000073, 0.00000436, -0.00000282, 0.00000201, -0.00000393, -0.00000097, -0.00000150, -0.00000117, 0.00000300, 0.00000036, 0.00000109, -0.00000203, -0.00000078, -0.00000354, 0.00000350, 0.00000835, -0.00000581, 0.00000230, -0.00000695, -0.00000353, 0.00000132, -0.00000245, 0.00000201, -0.00000051, -0.00000002, 0.00000597, -0.00000727, 0.00000292, -0.00000456, -0.00001295, 0.00000856, 0.00000583, -0.00000454, 0.00000045, -0.00000739, -0.00000499, -0.00000687, -0.00000367, -0.00003252, 0.00000613, -0.00002603, 0.00002698, -0.00007655, -0.00014708, 0.00032385, 0.00030347, -0.00015191, -0.00008238, -0.00002038, -0.00004977, 0.00001971, -0.00002066, 0.00000464, -0.00001161, 0.00000471, -0.00000079, -0.00000569, 0.00000044, 0.00000014, 0.00000358, -0.00000590, 0.00000737, -0.00000729, -0.00001229, 0.00000802, -0.00000636, 0.00000730, -0.00000079, -0.00000302, 0.00000232, -0.00000072, 0.00000113, -0.00000361, -0.00000514, 0.00000233, -0.00000220, 0.00000275, 0.00000444, -0.00000116, -0.00000298, -0.00000043, 0.00000003, 0.00000086, 0.00000179, -0.00000156, -0.00000138, 0.00000089, -0.00000352, 0.00000102, -0.00000216, 0.00000362, 0.00000139, -0.00000136, 0.00000313, -0.00000389, -0.00000023, -0.00000108, -0.00000070, 0.00000177, -0.00000238, 0.00000101, -0.00000146, 0.00000296, 0.00000136, -0.00000012, 0.00000135, -0.00000240, 0.00000015, },
		{ -0.00000315, -0.00000040, 0.00000081, -0.00000137, 0.00000089, -0.00000185, 0.00000182, 0.00000038, 0.00000057, 0.00000236, -0.00000143, 0.00000143, -0.00000268, 0.00000048, -0.00000023, -0.00000225, 0.00000093, -0.00000531, 0.00000409, 0.00000113, -0.00000009, 0.00000384, -0.00000238, 0.00000160, -0.00000395, -0.00000085, -0.00000150, -0.00000117, 0.00000325, -0.00000015, 0.00000180, -0.00000287, -0.00000029, -0.00000393, 0.00000464, 0.00000727, -0.00000538, 0.00000179, -0.00000638, -0.00000415, 0.00000177, -0.00000285, 0.00000286, -0.00000152, 0.00000120, 0.00000469, -0.00000583, 0.00000175, -0.00000471, -0.00001296, 0.00001034, 0.00000348, -0.00000273, -0.00000180, -0.00000388, -0.00000899, -0.00000388, -0.00000601, -0.00002931, 0.00000397, -0.00002351, 0.00002277, -0.00007155, -0.00016386, 0.00034413, 0.00030901, -0.00017457, -0.00006717, -0.00002922, -0.00004274, 0.00001418, -0.00001660, 0.00000335, -0.00001153, 0.00000440, -0.00000038, -0.00000638, 0.00000177, -0.00000179, 0.00000543, -0.00000687, 0.00000834, -0.00000953, -0.00001022, 0.00000703, -0.00000563, 0.00000654, -0.00000011, -0.00000323, 0.00000258, -0.00000131, 0.00000174, -0.00000451, -0.00000428, 0.00000150, -0.00000147, 0.00000276, 0.00000440, -0.00000172, -0.00000247, -0.00000071, 0.00000046, 0.00000029, 0.00000246, -0.00000212, -0.00000103, 0.00000044, -0.00000324, 0.00000092, -0.00000203, 0.00000363, 0.00000137, -0.00000123, 0.00000308, -0.00000425, 0.00000025, -0.00000150, -0.00000031, 0.00000135, -0.00000214, 0.00000097, -0.00000151, 0.00000317, 0.00000118, 0.00000002, 0.00000124, -0.00000245, 0.00000028, },
		{ -0.00000332, -0.00000016, 0.00000061, -0.00000133, 0.00000090, -0.00000200, 0.00000217, 0.00000006, 0.00000091, 0.00000212, -0.00000131, 0.00000137, -0.00000275, 0.00000056, -0.00000026, -0.00000232, 0.00000090, -0.00000546, 0.00000475, 0.00000042, 0.00000061, 0.00000326, -0.00000185, 0.00000111, -0.00000391, -0.00000078, -0.00000146, -0.00000119, 0.00000350, -0.00000066, 0.00000253, -0.00000376, 0.00000024, -0.00000433, 0.00000583, 0.00000608, -0.00000484, 0.00000118, -0.00000572, -0.00000485, 0.00000228, -0.00000330, 0.00000378, -0.00000262, 0.00000250, 0.00000327, -0.00000422, 0.00000038, -0.00000475, -0.00001296, 0.00001213, 0.00000102, -0.00000081, -0.00000417, -0.00000016, -0.00001332, -0.00000048, -0.00000884, -0.00002541, 0.00000102, -0.00002036, 0.00001784, -0.00006559, -0.00018194, 0.00036655, 0.00031255, -0.00019684, -0.00005134, -0.00003872, -0.00003501, 0.00000792, -0.00001187, 0.00000154, -0.00001110, 0.00000381, 0.00000020, -0.00000711, 0.00000311, -0.00000379, 0.00000737, -0.00000790, 0.00000933, -0.00001184, -0.00000797, 0.00000587, -0.00000475, 0.00000563, 0.00000068, -0.00000350, 0.00000289, -0.00000195, 0.00000239, -0.00000545, -0.00000337, 0.00000057, -0.00000063, 0.00000271, 0.00000437, -0.00000229, -0.00000191, -0.00000102, 0.00000091, -0.00000030, 0.00000315, -0.00000272, -0.00000064, -0.00000008, -0.00000289, 0.00000076, -0.00000185, 0.00000361, 0.00000137, -0.00000109, 0.00000300, -0.00000458, 0.00000072, -0.00000193, 0.00000011, 0.00000089, -0.00000184, 0.00000089, -0.00000151, 0.00000336, 0.00000099, 0.00000018, 0.00000110, -0.00000248, 0.00000040, },
		{ -0.00000348, 0.00000007, 0.00000039, -0.00000126, 0.00000089, -0.00000213, 0.00000251, -0.00000028, 0.00000127, 0.00000183, -0.00000114, 0.00000127, -0.00000278, 0.00000063, -0.00000028, -0.00000239, 0.00000084, -0.00000556, 0.00000539, -0.00000031, 0.00000134, 0.00000261, -0.00000125, 0.00000053, -0.00000378, -0.00000076, -0.00000139, -0.00000123, 0.00000373, -0.00000116, 0.00000327, -0.00000467, 0.00000081, -0.00000475, 0.00000706, 0.00000478, -0.00000419, 0.00000049, -0.00000497, -0.00000563, 0.00000285, -0.00000379, 0.00000477, -0.00000381, 0.00000389, 0.00000175, -0.00000244, -0.00000119, -0.00000467, -0.00001295, 0.00001390, -0.00000150, 0.00000120, -0.00000661, 0.00000372, -0.00001793, 0.00000329, -0.00001212, -0.00002087, -0.00000270, -0.00001656, 0.00001217, -0.00005867, -0.00020132, 0.00039119, 0.00031386, -0.00021844, -0.00003511, -0.00004873, -0.00002670, 0.00000101, -0.00000650, -0.00000079, -0.00001030, 0.00000294, 0.00000095, -0.00000789, 0.00000445, -0.00000583, 0.00000939, -0.00000897, 0.00001030, -0.00001421, -0.00000557, 0.00000453, -0.00000374, 0.00000460, 0.00000158, -0.00000384, 0.00000322, -0.00000264, 0.00000308, -0.00000643, -0.00000239, -0.00000044, 0.00000030, 0.00000259, 0.00000436, -0.00000286, -0.00000133, -0.00000134, 0.00000137, -0.00000091, 0.00000386, -0.00000337, -0.00000019, -0.00000066, -0.00000246, 0.00000055, -0.00000163, 0.00000355, 0.00000139, -0.00000094, 0.00000288, -0.00000488, 0.00000117, -0.00000236, 0.00000054, 0.00000040, -0.00000150, 0.00000077, -0.00000148, 0.00000351, 0.00000082, 0.00000035, 0.00000093, -0.00000247, 0.00000049, },
		{ -0.00000362, 0.00000030, 0.00000017, -0.00000117, 0.00000084, -0.00000224, 0.00000284, -0.00000062, 0.00000166, 0.00000151, -0.00000093, 0.00000112, -0.00000277, 0.00000068, -0.00000030, -0.00000245, 0.00000076, -0.00000562, 0.00000600, -0.00000105, 0.00000210, 0.00000190, -0.00000058, -0.00000013, -0.00000358, -0.00000079, -0.00000127, -0.00000129, 0.00000394, -0.00000164, 0.00000402, -0.00000561, 0.00000140, -0.00000517, 0.00000832, 0.00000340, -0.00000344, -0.00000028, -0.00000414, -0.00000646, 0.00000346, -0.00000431, 0.00000580, -0.00000506, 0.00000534, 0.00000012, -0.00000051, -0.00000295, -0.00000446, -0.00001292, 0.00001564, -0.00000407, 0.00000327, -0.00000910, 0.00000771, -0.00002277, 0.00000738, -0.00001581, -0.00001572, -0.00000718, -0.00001212, 0.00000577, -0.00005080, -0.00022200, 0.00041810, 0.00031274, -0.00023910, -0.00001872, -0.00005910, -0.00001794, -0.00000645, -0.00000054, -0.00000362, -0.00000914, 0.00000178, 0.00000189, -0.00000870, 0.00000577, -0.00000789, 0.00001147, -0.00001005, 0.00001124, -0.00001658, -0.00000303, 0.00000303, -0.00000259, 0.00000346, 0.00000257, -0.00000424, 0.00000357, -0.00000335, 0.00000381, -0.00000742, -0.00000137, -0.00000151, 0.00000133, 0.00000240, 0.00000436, -0.00000345, -0.00000073, -0.00000167, 0.00000182, -0.00000151, 0.00000459, -0.00000403, 0.00000029, -0.00000129, -0.00000196, 0.00000029, -0.00000136, 0.00000345, 0.00000143, -0.00000079, 0.00000272, -0.00000513, 0.00000160, -0.00000279, 0.00000098, -0.00000012, -0.00000111, 0.00000060, -0.00000141, 0.00000364, 0.00000065, 0.00000052, 0.00000074, -0.00000242, 0.00000055, },
		{ -0.00000373, 0.00000053, -0.00000007, -0.00000106, 0.00000077, -0.00000231, 0.00000315, -0.00000096, 0.00000206, 0.00000115, -0.00000067, 0.00000092, -0.00000273, 0.00000071, -0.00000032, -0.00000251, 0.00000065, -0.00000562, 0.00000656, -0.00000177, 0.00000287, 0.00000116, 0.00000015, -0.00000087, -0.00000330, -0.00000088, -0.00000110, -0.00000136, 0.00000415, -0.00000210, 0.00000475, -0.00000656, 0.00000201, -0.00000559, 0.00000958, 0.00000193, -0.00000261, -0.00000112, -0.00000324, -0.00000736, 0.00000412, -0.00000484, 0.00000687, -0.00000637, 0.00000684, -0.00000158, 0.00000155, -0.00000489, -0.00000413, -0.00001288, 0.00001733, -0.00000665, 0.00000536, -0.00001158, 0.00001174, -0.00002775, 0.00001172, -0.00001984, -0.00001003, -0.00001238, -0.00000705, -0.00000133, -0.00004202, -0.00024392, 0.00044733, 0.00030901, -0.00025854, -0.00000239, -0.00006965, -0.00000884, -0.00001437, 0.00000592, -0.00000691, -0.00000763, 0.00000034, 0.00000299, -0.00000956, 0.00000706, -0.00000996, 0.00001357, -0.00001114, 0.00001212, -0.00001895, -0.00000038, 0.00000141, -0.00000132, 0.00000220, 0.00000366, -0.00000470, 0.00000395, -0.00000409, 0.00000455, -0.00000843, -0.00000031, -0.00000265, 0.00000244, 0.00000214, 0.00000437, -0.00000404, -0.00000010, -0.00000201, 0.00000226, -0.00000212, 0.00000532, -0.00000472, 0.00000080, -0.00000197, -0.00000141, -0.00000003, -0.00000104, 0.00000331, 0.00000149, -0.00000063, 0.00000252, -0.00000533, 0.00000201, -0.00000321, 0.00000142, -0.00000066, -0.00000069, 0.00000039, -0.00000130, 0.00000373, 0.00000049, 0.00000070, 0.00000053, -0.00000235, 0.00000058, },
		{ -0.00000381, 0.00000073, -0.00000031, -0.00000093, 0.00000068, -0.00000234, 0.00000344, -0.00000129, 0.00000247, 0.00000076, -0.00000036, 0.00000067, -0.00000264, 0.00000071, -0.00000032, -0.00000256, 0.00000052, -0.00000557, 0.00000707, -0.00000248, 0.00000366, 0.00000038, 0.00000093, -0.00000168, -0.00000295, -0.00000101, -0.00000089, -0.00000146, 0.00000434, -0.00000254, 0.00000547, -0.00000752, 0.00000263, -0.00000598, 0.00001084, 0.00000041, -0.00000170, -0.00000202, -0.00000228, -0.00000829, 0.00000480, -0.00000538, 0.00000795, -0.00000771, 0.00000838, -0.00000334, 0.00000372, -0.00000699, -0.00000367, -0.00001282, 0.00001895, -0.00000922, 0.00000745, -0.00001402, 0.00001575, -0.00003282, 0.00001624, -0.00002412, -0.00000387, -0.00001823, -0.00000140, -0.00000910, -0.00003239, -0.00026703, 0.00047886, 0.00030251, -0.00027650, 0.00001365, -0.00008021, 0.00000044, -0.00002263, 0.00001282, -0.00001062, -0.00000580, -0.00000136, 0.00000427, -0.00001045, 0.00000832, -0.00001199, 0.00001569, -0.00001222, 0.00001292, -0.00002127, 0.00000233, -0.00000033, 0.00000004, 0.00000086, 0.00000483, -0.00000520, 0.00000433, -0.00000485, 0.00000531, -0.00000943, 0.00000076, -0.00000383, 0.00000363, 0.00000181, 0.00000439, -0.00000462, 0.00000054, -0.00000235, 0.00000269, -0.00000270, 0.00000603, -0.00000541, 0.00000133, -0.00000269, -0.00000079, -0.00000040, -0.00000068, 0.00000312, 0.00000157, -0.00000047, 0.00000229, -0.00000549, 0.00000238, -0.00000360, 0.00000185, -0.00000121, -0.00000022, 0.00000013, -0.00000114, 0.00000379, 0.00000035, 0.00000088, 0.00000030, -0.00000225, 0.00000057, },
		{ -0.00000386, 0.00000092, -0.00000055, -0.00000079, 0.00000056, -0.00000234, 0.00000370, -0.00000161, 0.00000288, 0.00000034, -0.00000002, 0.00000038, -0.00000252, 0.00000069, -0.00000031, -0.00000261, 0.00000037, -0.00000547, 0.00000752, -0.00000315, 0.00000443, -0.00000041, 0.00000175, -0.00000255, -0.00000254, -0.00000120, -0.00000065, -0.00000159, 0.00000452, -0.00000294, 0.00000616, -0.00000846, 0.00000324, -0.00000634, 0.00001207, -0.00000115, -0.00000073, -0.00000298, -0.00000127, -0.00000927, 0.00000549, -0.00000592, 0.00000903, -0.00000907, 0.00000992, -0.00000513, 0.00000597, -0.00000924, -0.00000310, -0.00001273, 0.00002049, -0.00001175, 0.00000949, -0.00001636, 0.00001969, -0.00003789, 0.00002085, -0.00002859, 0.00000266, -0.00002467, 0.00000479, -0.00001747, -0.00002198, -0.00029122, 0.00051266, 0.00029311, -0.00029273, 0.00002915, -0.00009059, 0.00000976, -0.00003109, 0.00002005, -0.00001469, -0.00000368, -0.00000330, 0.00000572, -0.00001137, 0.00000952, -0.00001398, 0.00001779, -0.00001326, 0.00001362, -0.00002351, 0.00000508, -0.00000215, 0.00000148, -0.00000056, 0.00000607, -0.00000575, 0.00000471, -0.00000560, 0.00000605, -0.00001041, 0.00000183, -0.00000504, 0.00000488, 0.00000143, 0.00000441, -0.00000520, 0.00000119, -0.00000269, 0.00000309, -0.00000326, 0.00000672, -0.00000609, 0.00000186, -0.00000343, -0.00000012, -0.00000082, -0.00000027, 0.00000289, 0.00000168, -0.00000031, 0.00000201, -0.00000560, 0.00000271, -0.00000397, 0.00000227, -0.00000176, 0.00000027, -0.00000016, -0.00000094, 0.00000381, 0.00000023, 0.00000106, 0.00000006, -0.00000212, 0.00000054, },
		{ -0.00000388, 0.00000108, -0.00000079, -0.00000063, 0.00000041, -0.00000230, 0.00000392, -0.00000191, 0.00000328, -0.00000008, 0.00000036, 0.00000004, -0.00000235, 0.00000063, -0.00000028, -0.00000266, 0.00000020, -0.00000530, 0.00000791, -0.00000379, 0.00000518, -0.00000121, 0.00000259, -0.00000346, -0.00000206, -0.00000143, -0.00000036, -0.00000173, 0.00000468, -0.00000330, 0.00000681, -0.00000938, 0.00000384, -0.00000667, 0.00001326, -0.00000273, 0.00000028, -0.00000396, -0.00000023, -0.00001026, 0.00000619, -0.00000643, 0.00001009, -0.00001041, 0.00001145, -0.00000693, 0.00000828, -0.00001161, -0.00000242, -0.00001262, 0.00002192, -0.00001420, 0.00001146, -0.00001857, 0.00002349, -0.00004289, 0.00002548, -0.00003315, 0.00000948, -0.00003160, 0.00001144, -0.00002637, -0.00001090, -0.00031638, 0.00054867, 0.00028070, -0.00030702, 0.00004391, -0.00010061, 0.00001895, -0.00003964, 0.00002751, -0.00001906, -0.00000130, -0.00000546, 0.00000731, -0.00001232, 0.00001066, -0.00001590, 0.00001985, -0.00001424, 0.00001419, -0.00002563, 0.00000783, -0.00000403, 0.00000298, -0.00000203, 0.00000735, -0.00000632, 0.00000508, -0.00000634, 0.00000678, -0.00001136, 0.00000289, -0.00000626, 0.00000618, 0.00000099, 0.00000444, -0.00000576, 0.00000184, -0.00000302, 0.00000347, -0.00000379, 0.00000738, -0.00000675, 0.00000239, -0.00000418, 0.00000059, -0.00000127, 0.00000017, 0.00000262, 0.00000182, -0.00000016, 0.00000171, -0.00000566, 0.00000300, -0.00000430, 0.00000267, -0.00000232, 0.00000077, -0.00000048, -0.00000070, 0.00000378, 0.00000014, 0.00000122, -0.00000020, -0.00000195, 0.00000047, },
		{ -0.00000386, 0.00000121, -0.00000102, -0.00000046, 0.00000024, -0.00000223, 0.00000411, -0.00000218, 0.00000367, -0.00000052, 0.00000076, -0.00000034, -0.00000214, 0.00000055, -0.00000024, -0.00000272, 0.00000001, -0.00000507, 0.00000823, -0.00000437, 0.00000590, -0.00000201, 0.00000344, -0.00000441, -0.00000153, -0.00000171, -0.00000003, -0.00000189, 0.00000482, -0.00000362, 0.00000742, -0.00001026, 0.00000442, -0.00000693, 0.00001439, -0.00000432, 0.00000133, -0.00000497, 0.00000084, -0.00001125, 0.00000688, -0.00000690, 0.00001111, -0.00001173, 0.00001294, -0.00000870, 0.00001061, -0.00001409, -0.00000165, -0.00001246, 0.00002323, -0.00001656, 0.00001333, -0.00002061, 0.00002708, -0.00004773, 0.00003003, -0.00003770, 0.00001646, -0.00003892, 0.00001847, -0.00003570, 0.00000074, -0.00034237, 0.00058679, 0.00026523, -0.00031915, 0.00005769, -0.00011010, 0.00002787, -0.00004813, 0.00003510, -0.00002365, 0.00000127, -0.00000780, 0.00000903, -0.00001328, 0.00001171, -0.00001773, 0.00002184, -0.00001514, 0.00001462, -0.00002761, 0.00001054, -0.00000593, 0.00000451, -0.00000354, 0.00000867, -0.00000691, 0.00000543, -0.00000706, 0.00000748, -0.00001227, 0.00000393, -0.00000747, 0.00000752, 0.00000051, 0.00000447, -0.00000631, 0.00000248, -0.00000333, 0.00000380, -0.00000427, 0.00000800, -0.00000738, 0.00000289, -0.00000494, 0.00000132, -0.00000176, 0.00000066, 0.00000231, 0.00000199, -0.00000002, 0.00000137, -0.00000566, 0.00000323, -0.00000459, 0.00000303, -0.00000285, 0.00000130, -0.00000082, -0.00000043, 0.00000372, 0.00000007, 0.00000138, -0.00000046, -0.00000177, 0.00000036, },
		{ -0.00000381, 0.00000132, -0.00000124, -0.00000028, 0.00000005, -0.00000212, 0.00000425, -0.00000242, 0.00000403, -0.00000095, 0.00000117, -0.00000075, -0.00000190, 0.00000044, -0.00000018, -0.00000278, -0.00000020, -0.00000479, 0.00000847, -0.00000489, 0.00000657, -0.00000278, 0.00000428, -0.00000539, -0.00000096, -0.00000202, 0.00000033, -0.00000206, 0.00000494, -0.00000389, 0.00000797, -0.00001111, 0.00000496, -0.00000713, 0.00001545, -0.00000589, 0.00000238, -0.00000599, 0.00000191, -0.00001224, 0.00000755, -0.00000733, 0.00001207, -0.00001299, 0.00001437, -0.00001043, 0.00001293, -0.00001663, -0.00000079, -0.00001226, 0.00002440, -0.00001879, 0.00001505, -0.00002243, 0.00003041, -0.00005234, 0.00003442, -0.00004214, 0.00002349, -0.00004652, 0.00002578, -0.00004537, 0.00001281, -0.00036902, 0.00062688, 0.00024665, -0.00032894, 0.00007030, -0.00011887, 0.00003635, -0.00005641, 0.00004270, -0.00002839, 0.00000400, -0.00001030, 0.00001086, -0.00001423, 0.00001266, -0.00001943, 0.00002375, -0.00001596, 0.00001487, -0.00002940, 0.00001318, -0.00000783, 0.00000605, -0.00000506, 0.00001001, -0.00000751, 0.00000575, -0.00000773, 0.00000814, -0.00001311, 0.00000491, -0.00000866, 0.00000888, -0.00000001, 0.00000450, -0.00000682, 0.00000312, -0.00000363, 0.00000410, -0.00000470, 0.00000855, -0.00000796, 0.00000337, -0.00000568, 0.00000208, -0.00000228, 0.00000117, 0.00000196, 0.00000219, 0.00000011, 0.00000101, -0.00000561, 0.00000341, -0.00000483, 0.00000335, -0.00000337, 0.00000182, -0.00000119, -0.00000013, 0.00000362, 0.00000003, 0.00000152, -0.00000073, -0.00000156, 0.00000023, },
		{ -0.00000372, 0.00000139, -0.00000144, -0.00000011, -0.00000016, -0.00000196, 0.00000435, -0.00000262, 0.00000437, -0.00000138, 0.00000160, -0.00000118, -0.00000162, 0.00000030, -0.00000011, -0.00000284, -0.00000042, -0.00000445, 0.00000863, -0.00000533, 0.00000718, -0.00000351, 0.00000511, -0.00000637, -0.00000035, -0.00000236, 0.00000072, -0.00000226, 0.00000504, -0.00000411, 0.00000845, -0.00001189, 0.00000545, -0.00000725, 0.00001640, -0.00000742, 0.00000343, -0.00000700, 0.00000296, -0.00001321, 0.00000818, -0.00000769, 0.00001294, -0.00001418, 0.00001572, -0.00001208, 0.00001520, -0.00001922, 0.00000013, -0.00001201, 0.00002542, -0.00002087, 0.00001662, -0.00002399, 0.00003342, -0.00005664, 0.00003856, -0.00004637, 0.00003045, -0.00005427, 0.00003324, -0.00005524, 0.00002515, -0.00039613, 0.00066878, 0.00022499, -0.00033624, 0.00008156, -0.00012677, 0.00004425, -0.00006435, 0.00005019, -0.00003318, 0.00000682, -0.00001289, 0.00001277, -0.00001518, 0.00001351, -0.00002100, 0.00002555, -0.00001665, 0.00001493, -0.00003098, 0.00001570, -0.00000969, 0.00000758, -0.00000657, 0.00001134, -0.00000810, 0.00000603, -0.00000836, 0.00000873, -0.00001389, 0.00000583, -0.00000980, 0.00001025, -0.00000057, 0.00000451, -0.00000730, 0.00000373, -0.00000390, 0.00000435, -0.00000507, 0.00000904, -0.00000849, 0.00000380, -0.00000640, 0.00000284, -0.00000281, 0.00000171, 0.00000158, 0.00000241, 0.00000022, 0.00000062, -0.00000550, 0.00000354, -0.00000502, 0.00000363, -0.00000385, 0.00000234, -0.00000157, 0.00000021, 0.00000349, 0.00000003, 0.00000163, -0.00000099, -0.00000133, 0.00000007, },
		{ -0.00000360, 0.00000143, -0.00000162, 0.00000007, -0.00000038, -0.00000178, 0.00000441, -0.00000277, 0.00000467, -0.00000179, 0.00000204, -0.00000164, -0.00000131, 0.00000013, -0.00000001, -0.00000291, -0.00000065, -0.00000405, 0.00000871, -0.00000570, 0.00000772, -0.00000420, 0.00000590, -0.00000735, 0.00000028, -0.00000272, 0.00000112, -0.00000246, 0.00000512, -0.00000427, 0.00000886, -0.00001261, 0.00000589, -0.00000729, 0.00001725, -0.00000889, 0.00000445, -0.00000798, 0.00000399, -0.00001414, 0.00000877, -0.00000796, 0.00001372, -0.00001527, 0.00001695, -0.00001361, 0.00001740, -0.00002181, 0.00000110, -0.00001169, 0.00002626, -0.00002278, 0.00001799, -0.00002528, 0.00003606, -0.00006055, 0.00004236, -0.00005028, 0.00003721, -0.00006204, 0.00004075, -0.00006519, 0.00003759, -0.00042348, 0.00071229, 0.00020029, -0.00034094, 0.00009129, -0.00013364, 0.00005142, -0.00007182, 0.00005744, -0.00003794, 0.00000967, -0.00001555, 0.00001473, -0.00001609, 0.00001422, -0.00002240, 0.00002721, -0.00001723, 0.00001478, -0.00003232, 0.00001808, -0.00001149, 0.00000905, -0.00000805, 0.00001266, -0.00000867, 0.00000625, -0.00000892, 0.00000926, -0.00001457, 0.00000667, -0.00001088, 0.00001160, -0.00000114, 0.00000450, -0.00000775, 0.00000432, -0.00000414, 0.00000455, -0.00000537, 0.00000946, -0.00000896, 0.00000418, -0.00000707, 0.00000359, -0.00000334, 0.00000227, 0.00000118, 0.00000267, 0.00000032, 0.00000021, -0.00000534, 0.00000361, -0.00000516, 0.00000385, -0.00000429, 0.00000284, -0.00000195, 0.00000057, 0.00000332, 0.00000005, 0.00000173, -0.00000125, -0.00000109, -0.00000012, },
		{ -0.00000344, 0.00000143, -0.00000177, 0.00000024, -0.00000061, -0.00000156, 0.00000442, -0.00000288, 0.00000493, -0.00000218, 0.00000246, -0.00000212, -0.00000097, -0.00000006, 0.00000010, -0.00000299, -0.00000089, -0.00000360, 0.00000869, -0.00000598, 0.00000819, -0.00000482, 0.00000665, -0.00000831, 0.00000091, -0.00000309, 0.00000155, -0.00000267, 0.00000517, -0.00000438, 0.00000919, -0.00001325, 0.00000626, -0.00000722, 0.00001797, -0.00001027, 0.00000543, -0.00000892, 0.00000498, -0.00001502, 0.00000929, -0.00000815, 0.00001438, -0.00001623, 0.00001805, -0.00001502, 0.00001949, -0.00002438, 0.00000211, -0.00001131, 0.00002692, -0.00002448, 0.00001915, -0.00002624, 0.00003828, -0.00006402, 0.00004573, -0.00005378, 0.00004364, -0.00006968, 0.00004815, -0.00007506, 0.00004996, -0.00045084, 0.00075716, 0.00017263, -0.00034296, 0.00009935, -0.00013933, 0.00005773, -0.00007867, 0.00006434, -0.00004258, 0.00001248, -0.00001821, 0.00001672, -0.00001696, 0.00001480, -0.00002363, 0.00002872, -0.00001765, 0.00001441, -0.00003339, 0.00002028, -0.00001319, 0.00001046, -0.00000947, 0.00001393, -0.00000920, 0.00000641, -0.00000940, 0.00000970, -0.00001515, 0.00000741, -0.00001189, 0.00001292, -0.00000172, 0.00000448, -0.00000814, 0.00000487, -0.00000434, 0.00000469, -0.00000560, 0.00000979, -0.00000934, 0.00000449, -0.00000769, 0.00000433, -0.00000388, 0.00000283, 0.00000074, 0.00000295, 0.00000039, -0.00000021, -0.00000513, 0.00000362, -0.00000523, 0.00000402, -0.00000468, 0.00000333, -0.00000234, 0.00000094, 0.00000311, 0.00000010, 0.00000180, -0.00000151, -0.00000084, -0.00000034, },
		{ -0.00000325, 0.00000140, -0.00000190, 0.00000039, -0.00000085, -0.00000130, 0.00000438, -0.00000294, 0.00000514, -0.00000254, 0.00000288, -0.00000259, -0.00000061, -0.00000028, 0.00000023, -0.00000308, -0.00000113, -0.00000310, 0.00000860, -0.00000616, 0.00000856, -0.00000537, 0.00000733, -0.00000923, 0.00000155, -0.00000347, 0.00000198, -0.00000288, 0.00000519, -0.00000443, 0.00000944, -0.00001380, 0.00000657, -0.00000705, 0.00001855, -0.00001156, 0.00000635, -0.00000980, 0.00000592, -0.00001582, 0.00000974, -0.00000823, 0.00001490, -0.00001706, 0.00001899, -0.00001625, 0.00002144, -0.00002690, 0.00000312, -0.00001084, 0.00002738, -0.00002597, 0.00002008, -0.00002688, 0.00004004, -0.00006697, 0.00004860, -0.00005677, 0.00004962, -0.00007706, 0.00005532, -0.00008471, 0.00006205, -0.00047794, 0.00080312, 0.00014214, -0.00034223, 0.00010564, -0.00014374, 0.00006305, -0.00008478, 0.00007077, -0.00004700, 0.00001518, -0.00002083, 0.00001869, -0.00001777, 0.00001523, -0.00002465, 0.00003006, -0.00001793, 0.00001381, -0.00003418, 0.00002227, -0.00001475, 0.00001177, -0.00001081, 0.00001514, -0.00000969, 0.00000650, -0.00000980, 0.00001005, -0.00001562, 0.00000805, -0.00001279, 0.00001419, -0.00000229, 0.00000443, -0.00000849, 0.00000539, -0.00000451, 0.00000478, -0.00000576, 0.00001003, -0.00000965, 0.00000472, -0.00000825, 0.00000503, -0.00000440, 0.00000339, 0.00000030, 0.00000326, 0.00000044, -0.00000064, -0.00000488, 0.00000357, -0.00000524, 0.00000412, -0.00000502, 0.00000378, -0.00000271, 0.00000133, 0.00000287, 0.00000019, 0.00000184, -0.00000174, -0.00000058, -0.00000058, },
		{ -0.00000303, 0.00000132, -0.00000200, 0.00000054, -0.00000109, -0.00000102, 0.00000430, -0.00000294, 0.00000529, -0.00000285, 0.00000327, -0.00000307, -0.00000024, -0.00000052, 0.00000037, -0.00000318, -0.00000138, -0.00000255, 0.00000841, -0.00000626, 0.00000884, -0.00000583, 0.00000794, -0.00001011, 0.00000217, -0.00000384, 0.00000241, -0.00000309, 0.00000519, -0.00000442, 0.00000960, -0.00001426, 0.00000679, -0.00000677, 0.00001897, -0.00001273, 0.00000719, -0.00001061, 0.00000678, -0.00001655, 0.00001010, -0.00000820, 0.00001528, -0.00001772, 0.00001976, -0.00001730, 0.00002320, -0.00002932, 0.00000413, -0.00001029, 0.00002763, -0.00002722, 0.00002076, -0.00002716, 0.00004130, -0.00006935, 0.00005091, -0.00005915, 0.00005502, -0.00008403, 0.00006211, -0.00009398, 0.00007366, -0.00050451, 0.00084989, 0.00010900, -0.00033876, 0.00011005, -0.00014677, 0.00006729, -0.00009004, 0.00007662, -0.00005112, 0.00001771, -0.00002335, 0.00002061, -0.00001850, 0.00001549, -0.00002546, 0.00003120, -0.00001804, 0.00001297, -0.00003465, 0.00002401, -0.00001617, 0.00001296, -0.00001205, 0.00001628, -0.00001012, 0.00000650, -0.00001010, 0.00001029, -0.00001596, 0.00000856, -0.00001357, 0.00001540, -0.00000285, 0.00000434, -0.00000877, 0.00000587, -0.00000464, 0.00000480, -0.00000583, 0.00001017, -0.00000986, 0.00000487, -0.00000872, 0.00000569, -0.00000489, 0.00000395, -0.00000016, 0.00000358, 0.00000047, -0.00000108, -0.00000458, 0.00000346, -0.00000518, 0.00000415, -0.00000529, 0.00000419, -0.00000307, 0.00000172, 0.00000261, 0.00000030, 0.00000185, -0.00000196, -0.00000032, -0.00000083, },
		{ -0.00000278, 0.00000121, -0.00000207, 0.00000067, -0.00000133, -0.00000072, 0.00000417, -0.00000290, 0.00000539, -0.00000312, 0.00000363, -0.00000354, 0.00000014, -0.00000077, 0.00000053, -0.00000329, -0.00000161, -0.00000197, 0.00000815, -0.00000625, 0.00000901, -0.00000621, 0.00000846, -0.00001091, 0.00000276, -0.00000419, 0.00000283, -0.00000329, 0.00000515, -0.00000434, 0.00000966, -0.00001462, 0.00000692, -0.00000637, 0.00001923, -0.00001377, 0.00000793, -0.00001133, 0.00000755, -0.00001718, 0.00001037, -0.00000804, 0.00001549, -0.00001821, 0.00002032, -0.00001813, 0.00002475, -0.00003162, 0.00000510, -0.00000965, 0.00002766, -0.00002823, 0.00002117, -0.00002707, 0.00004205, -0.00007111, 0.00005257, -0.00006085, 0.00005974, -0.00009046, 0.00006838, -0.00010270, 0.00008460, -0.00053026, 0.00089713, 0.00007340, -0.00033256, 0.00011254, -0.00014833, 0.00007034, -0.00009435, 0.00008178, -0.00005484, 0.00002000, -0.00002572, 0.00002245, -0.00001913, 0.00001558, -0.00002605, 0.00003215, -0.00001797, 0.00001187, -0.00003479, 0.00002549, -0.00001740, 0.00001401, -0.00001317, 0.00001731, -0.00001047, 0.00000641, -0.00001030, 0.00001041, -0.00001617, 0.00000893, -0.00001423, 0.00001652, -0.00000339, 0.00000421, -0.00000900, 0.00000629, -0.00000473, 0.00000477, -0.00000582, 0.00001022, -0.00000997, 0.00000492, -0.00000911, 0.00000629, -0.00000535, 0.00000448, -0.00000062, 0.00000393, 0.00000047, -0.00000152, -0.00000424, 0.00000330, -0.00000507, 0.00000412, -0.00000550, 0.00000455, -0.00000340, 0.00000212, 0.00000233, 0.00000045, 0.00000183, -0.00000216, -0.00000007, -0.00000109, },
		{ -0.00000250, 0.00000107, -0.00000211, 0.00000078, -0.00000157, -0.00000039, 0.00000400, -0.00000279, 0.00000542, -0.00000334, 0.00000395, -0.00000399, 0.00000053, -0.00000104, 0.00000069, -0.00000341, -0.00000184, -0.00000136, 0.00000780, -0.00000614, 0.00000908, -0.00000648, 0.00000889, -0.00001164, 0.00000331, -0.00000452, 0.00000324, -0.00000348, 0.00000509, -0.00000420, 0.00000963, -0.00001488, 0.00000697, -0.00000585, 0.00001932, -0.00001465, 0.00000857, -0.00001195, 0.00000823, -0.00001771, 0.00001054, -0.00000775, 0.00001554, -0.00001851, 0.00002067, -0.00001873, 0.00002606, -0.00003376, 0.00000603, -0.00000890, 0.00002748, -0.00002898, 0.00002132, -0.00002660, 0.00004226, -0.00007223, 0.00005356, -0.00006180, 0.00006365, -0.00009620, 0.00007399, -0.00011072, 0.00009465, -0.00055491, 0.00094451, 0.00003558, -0.00032371, 0.00011306, -0.00014837, 0.00007215, -0.00009761, 0.00008615, -0.00005809, 0.00002198, -0.00002788, 0.00002416, -0.00001965, 0.00001549, -0.00002640, 0.00003288, -0.00001774, 0.00001053, -0.00003460, 0.00002667, -0.00001842, 0.00001490, -0.00001415, 0.00001824, -0.00001074, 0.00000623, -0.00001038, 0.00001042, -0.00001624, 0.00000916, -0.00001474, 0.00001755, -0.00000389, 0.00000404, -0.00000915, 0.00000666, -0.00000477, 0.00000467, -0.00000573, 0.00001016, -0.00000998, 0.00000487, -0.00000940, 0.00000682, -0.00000576, 0.00000499, -0.00000108, 0.00000428, 0.00000044, -0.00000196, -0.00000387, 0.00000308, -0.00000489, 0.00000401, -0.00000564, 0.00000486, -0.00000370, 0.00000250, 0.00000204, 0.00000062, 0.00000177, -0.00000233, 0.00000018, -0.00000137, },
		{ -0.00000221, 0.00000089, -0.00000211, 0.00000086, -0.00000179, -0.00000005, 0.00000379, -0.00000264, 0.00000539, -0.00000350, 0.00000423, -0.00000441, 0.00000091, -0.00000131, 0.00000087, -0.00000354, -0.00000206, -0.00000071, 0.00000738, -0.00000594, 0.00000904, -0.00000664, 0.00000920, -0.00001228, 0.00000381, -0.00000481, 0.00000363, -0.00000366, 0.00000499, -0.00000400, 0.00000949, -0.00001503, 0.00000693, -0.00000521, 0.00001922, -0.00001538, 0.00000908, -0.00001246, 0.00000880, -0.00001812, 0.00001059, -0.00000733, 0.00001540, -0.00001861, 0.00002080, -0.00001908, 0.00002711, -0.00003572, 0.00000688, -0.00000804, 0.00002706, -0.00002946, 0.00002120, -0.00002577, 0.00004192, -0.00007267, 0.00005381, -0.00006192, 0.00006668, -0.00010115, 0.00007880, -0.00011788, 0.00010361, -0.00057814, 0.00099165, -0.00000418, -0.00031230, 0.00011164, -0.00014687, 0.00007266, -0.00009976, 0.00008966, -0.00006078, 0.00002359, -0.00002979, 0.00002571, -0.00002003, 0.00001521, -0.00002650, 0.00003339, -0.00001732, 0.00000895, -0.00003406, 0.00002755, -0.00001922, 0.00001560, -0.00001496, 0.00001903, -0.00001092, 0.00000594, -0.00001035, 0.00001029, -0.00001616, 0.00000923, -0.00001509, 0.00001846, -0.00000435, 0.00000382, -0.00000923, 0.00000697, -0.00000477, 0.00000452, -0.00000556, 0.00001000, -0.00000988, 0.00000471, -0.00000958, 0.00000727, -0.00000612, 0.00000547, -0.00000153, 0.00000465, 0.00000039, -0.00000239, -0.00000347, 0.00000282, -0.00000465, 0.00000384, -0.00000570, 0.00000512, -0.00000396, 0.00000287, 0.00000173, 0.00000081, 0.00000169, -0.00000248, 0.00000042, -0.00000164, },
		{ -0.00000190, 0.00000067, -0.00000208, 0.00000092, -0.00000200, 0.00000030, 0.00000355, -0.00000243, 0.00000530, -0.00000359, 0.00000445, -0.00000479, 0.00000127, -0.00000158, 0.00000106, -0.00000369, -0.00000226, -0.00000005, 0.00000689, -0.00000564, 0.00000889, -0.00000669, 0.00000940, -0.00001282, 0.00000424, -0.00000505, 0.00000398, -0.00000381, 0.00000486, -0.00000373, 0.00000927, -0.00001507, 0.00000679, -0.00000444, 0.00001895, -0.00001593, 0.00000946, -0.00001284, 0.00000924, -0.00001841, 0.00001053, -0.00000677, 0.00001508, -0.00001849, 0.00002069, -0.00001916, 0.00002786, -0.00003745, 0.00000764, -0.00000707, 0.00002641, -0.00002966, 0.00002081, -0.00002456, 0.00004104, -0.00007242, 0.00005332, -0.00006119, 0.00006873, -0.00010518, 0.00008269, -0.00012405, 0.00011129, -0.00059968, 0.00103819, -0.00004558, -0.00029847, 0.00010828, -0.00014382, 0.00007184, -0.00010074, 0.00009224, -0.00006286, 0.00002478, -0.00003140, 0.00002707, -0.00002027, 0.00001474, -0.00002636, 0.00003367, -0.00001673, 0.00000713, -0.00003318, 0.00002810, -0.00001977, 0.00001610, -0.00001560, 0.00001968, -0.00001100, 0.00000555, -0.00001019, 0.00001004, -0.00001594, 0.00000915, -0.00001527, 0.00001924, -0.00000474, 0.00000354, -0.00000923, 0.00000722, -0.00000472, 0.00000431, -0.00000531, 0.00000973, -0.00000968, 0.00000444, -0.00000965, 0.00000764, -0.00000641, 0.00000589, -0.00000195, 0.00000502, 0.00000031, -0.00000280, -0.00000305, 0.00000251, -0.00000436, 0.00000359, -0.00000568, 0.00000530, -0.00000417, 0.00000321, 0.00000142, 0.00000103, 0.00000157, -0.00000259, 0.00000063, -0.00000191, },
		{ -0.00000158, 0.00000044, -0.00000202, 0.00000095, -0.00000219, 0.00000066, 0.00000327, -0.00000217, 0.00000514, -0.00000363, 0.00000461, -0.00000514, 0.00000162, -0.00000184, 0.00000124, -0.00000384, -0.00000245, 0.00000062, 0.00000634, -0.00000526, 0.00000863, -0.00000663, 0.00000948, -0.00001325, 0.00000460, -0.00000524, 0.00000429, -0.00000394, 0.00000469, -0.00000341, 0.00000894, -0.00001500, 0.00000656, -0.00000356, 0.00001848, -0.00001630, 0.00000969, -0.00001309, 0.00000956, -0.00001856, 0.00001034, -0.00000606, 0.00001457, -0.00001817, 0.00002034, -0.00001896, 0.00002830, -0.00003894, 0.00000829, -0.00000598, 0.00002553, -0.00002960, 0.00002014, -0.00002301, 0.00003962, -0.00007148, 0.00005205, -0.00005957, 0.00006974, -0.00010819, 0.00008554, -0.00012907, 0.00011748, -0.00061922, 0.00108373, -0.00008828, -0.00028239, 0.00010307, -0.00013925, 0.00006970, -0.00010052, 0.00009382, -0.00006426, 0.00002550, -0.00003266, 0.00002821, -0.00002034, 0.00001408, -0.00002595, 0.00003372, -0.00001597, 0.00000508, -0.00003195, 0.00002832, -0.00002007, 0.00001640, -0.00001605, 0.00002017, -0.00001097, 0.00000504, -0.00000991, 0.00000965, -0.00001556, 0.00000890, -0.00001528, 0.00001988, -0.00000507, 0.00000321, -0.00000916, 0.00000741, -0.00000462, 0.00000404, -0.00000498, 0.00000937, -0.00000937, 0.00000407, -0.00000960, 0.00000790, -0.00000662, 0.00000626, -0.00000235, 0.00000538, 0.00000021, -0.00000320, -0.00000261, 0.00000217, -0.00000401, 0.00000328, -0.00000559, 0.00000542, -0.00000434, 0.00000353, 0.00000110, 0.00000126, 0.00000142, -0.00000268, 0.00000083, -0.00000218, },
		{ -0.00000125, 0.00000017, -0.00000192, 0.00000095, -0.00000235, 0.00000101, 0.00000297, -0.00000186, 0.00000492, -0.00000359, 0.00000471, -0.00000543, 0.00000194, -0.00000210, 0.00000142, -0.00000400, -0.00000261, 0.00000130, 0.00000573, -0.00000478, 0.00000826, -0.00000645, 0.00000943, -0.00001355, 0.00000487, -0.00000536, 0.00000455, -0.00000403, 0.00000449, -0.00000304, 0.00000853, -0.00001482, 0.00000624, -0.00000256, 0.00001784, -0.00001648, 0.00000977, -0.00001320, 0.00000974, -0.00001858, 0.00001004, -0.00000523, 0.00001387, -0.00001762, 0.00001974, -0.00001848, 0.00002842, -0.00004016, 0.00000881, -0.00000478, 0.00002443, -0.00002927, 0.00001922, -0.00002111, 0.00003769, -0.00006986, 0.00005002, -0.00005704, 0.00006967, -0.00011011, 0.00008724, -0.00013283, 0.00012203, -0.00063647, 0.00112788, -0.00013195, -0.00026424, 0.00009609, -0.00013321, 0.00006624, -0.00009909, 0.00009437, -0.00006495, 0.00002571, -0.00003353, 0.00002908, -0.00002024, 0.00001321, -0.00002530, 0.00003354, -0.00001505, 0.00000283, -0.00003039, 0.00002821, -0.00002010, 0.00001647, -0.00001630, 0.00002050, -0.00001082, 0.00000443, -0.00000951, 0.00000914, -0.00001503, 0.00000849, -0.00001510, 0.00002038, -0.00000533, 0.00000281, -0.00000900, 0.00000752, -0.00000448, 0.00000373, -0.00000459, 0.00000891, -0.00000897, 0.00000358, -0.00000943, 0.00000807, -0.00000676, 0.00000657, -0.00000271, 0.00000574, 0.00000008, -0.00000358, -0.00000216, 0.00000179, -0.00000362, 0.00000291, -0.00000542, 0.00000547, -0.00000445, 0.00000381, 0.00000080, 0.00000150, 0.00000124, -0.00000273, 0.00000101, -0.00000243, },
		{ -0.00000092, -0.00000011, -0.00000179, 0.00000092, -0.00000249, 0.00000136, 0.00000265, -0.00000151, 0.00000463, -0.00000349, 0.00000475, -0.00000567, 0.00000223, -0.00000234, 0.00000160, -0.00000417, -0.00000275, 0.00000198, 0.00000508, -0.00000423, 0.00000779, -0.00000616, 0.00000925, -0.00001374, 0.00000506, -0.00000541, 0.00000476, -0.00000409, 0.00000426, -0.00000261, 0.00000802, -0.00001454, 0.00000584, -0.00000146, 0.00001701, -0.00001648, 0.00000969, -0.00001317, 0.00000979, -0.00001846, 0.00000961, -0.00000425, 0.00001300, -0.00001687, 0.00001890, -0.00001771, 0.00002820, -0.00004110, 0.00000919, -0.00000345, 0.00002311, -0.00002867, 0.00001806, -0.00001891, 0.00003526, -0.00006757, 0.00004723, -0.00005360, 0.00006848, -0.00011086, 0.00008773, -0.00013522, 0.00012477, -0.00065118, 0.00117025, -0.00017620, -0.00024427, 0.00008747, -0.00012575, 0.00006152, -0.00009646, 0.00009386, -0.00006487, 0.00002537, -0.00003399, 0.00002968, -0.00001995, 0.00001215, -0.00002440, 0.00003313, -0.00001398, 0.00000038, -0.00002851, 0.00002776, -0.00001986, 0.00001631, -0.00001635, 0.00002066, -0.00001056, 0.00000371, -0.00000899, 0.00000849, -0.00001436, 0.00000793, -0.00001474, 0.00002071, -0.00000551, 0.00000235, -0.00000876, 0.00000757, -0.00000429, 0.00000337, -0.00000413, 0.00000837, -0.00000846, 0.00000299, -0.00000914, 0.00000812, -0.00000680, 0.00000681, -0.00000302, 0.00000608, -0.00000008, -0.00000393, -0.00000170, 0.00000139, -0.00000319, 0.00000248, -0.00000517, 0.00000544, -0.00000449, 0.00000405, 0.00000051, 0.00000175, 0.00000104, -0.00000274, 0.00000115, -0.00000267, },
		{ -0.00000060, -0.00000041, -0.00000164, 0.00000086, -0.00000261, 0.00000170, 0.00000231, -0.00000113, 0.00000428, -0.00000332, 0.00000471, -0.00000584, 0.00000248, -0.00000256, 0.00000177, -0.00000434, -0.00000286, 0.00000264, 0.00000440, -0.00000361, 0.00000723, -0.00000576, 0.00000895, -0.00001379, 0.00000515, -0.00000539, 0.00000491, -0.00000411, 0.00000400, -0.00000213, 0.00000743, -0.00001415, 0.00000535, -0.00000026, 0.00001601, -0.00001628, 0.00000946, -0.00001299, 0.00000969, -0.00001820, 0.00000906, -0.00000316, 0.00001195, -0.00001590, 0.00001782, -0.00001666, 0.00002763, -0.00004173, 0.00000942, -0.00000201, 0.00002157, -0.00002782, 0.00001666, -0.00001642, 0.00003238, -0.00006464, 0.00004370, -0.00004926, 0.00006615, -0.00011039, 0.00008692, -0.00013615, 0.00012556, -0.00066309, 0.00121046, -0.00022065, -0.00022271, 0.00007734, -0.00011700, 0.00005559, -0.00009264, 0.00009230, -0.00006403, 0.00002446, -0.00003400, 0.00002997, -0.00001946, 0.00001091, -0.00002325, 0.00003250, -0.00001277, -0.00000223, -0.00002633, 0.00002698, -0.00001935, 0.00001593, -0.00001619, 0.00002064, -0.00001018, 0.00000288, -0.00000835, 0.00000773, -0.00001354, 0.00000720, -0.00001419, 0.00002089, -0.00000560, 0.00000183, -0.00000843, 0.00000755, -0.00000405, 0.00000296, -0.00000361, 0.00000774, -0.00000787, 0.00000231, -0.00000872, 0.00000807, -0.00000676, 0.00000697, -0.00000329, 0.00000640, -0.00000025, -0.00000426, -0.00000125, 0.00000097, -0.00000272, 0.00000200, -0.00000485, 0.00000535, -0.00000448, 0.00000425, 0.00000024, 0.00000200, 0.00000081, -0.00000272, 0.00000127, -0.00000288, },
		{ -0.00000029, -0.00000072, -0.00000146, 0.00000077, -0.00000269, 0.00000203, 0.00000196, -0.00000071, 0.00000388, -0.00000308, 0.00000460, -0.00000595, 0.00000268, -0.00000276, 0.00000192, -0.00000451, -0.00000294, 0.00000329, 0.00000369, -0.00000293, 0.00000657, -0.00000526, 0.00000852, -0.00001371, 0.00000514, -0.00000529, 0.00000500, -0.00000409, 0.00000370, -0.00000161, 0.00000677, -0.00001366, 0.00000480, 0.00000103, 0.00001485, -0.00001590, 0.00000906, -0.00001267, 0.00000946, -0.00001780, 0.00000840, -0.00000195, 0.00001073, -0.00001474, 0.00001651, -0.00001533, 0.00002672, -0.00004204, 0.00000948, -0.00000046, 0.00001984, -0.00002673, 0.00001506, -0.00001369, 0.00002909, -0.00006112, 0.00003948, -0.00004407, 0.00006271, -0.00010869, 0.00008478, -0.00013556, 0.00012430, -0.00067197, 0.00124812, -0.00026493, -0.00019985, 0.00006587, -0.00010705, 0.00004854, -0.00008770, 0.00008968, -0.00006240, 0.00002297, -0.00003355, 0.00002994, -0.00001878, 0.00000948, -0.00002188, 0.00003166, -0.00001143, -0.00000499, -0.00002387, 0.00002588, -0.00001858, 0.00001532, -0.00001582, 0.00002044, -0.00000969, 0.00000196, -0.00000760, 0.00000684, -0.00001260, 0.00000633, -0.00001346, 0.00002089, -0.00000560, 0.00000124, -0.00000802, 0.00000746, -0.00000378, 0.00000253, -0.00000304, 0.00000704, -0.00000719, 0.00000152, -0.00000819, 0.00000790, -0.00000662, 0.00000706, -0.00000349, 0.00000669, -0.00000044, -0.00000455, -0.00000080, 0.00000054, -0.00000223, 0.00000147, -0.00000447, 0.00000518, -0.00000440, 0.00000440, -0.00000001, 0.00000224, 0.00000056, -0.00000267, 0.00000135, -0.00000307, },
		{ 0.00000000, -0.00000104, -0.00000126, 0.00000066, -0.00000273, 0.00000233, 0.00000162, -0.00000027, 0.00000343, -0.00000278, 0.00000442, -0.00000600, 0.00000283, -0.00000292, 0.00000206, -0.00000468, -0.00000300, 0.00000391, 0.00000296, -0.00000220, 0.00000584, -0.00000466, 0.00000797, -0.00001350, 0.00000502, -0.00000510, 0.00000502, -0.00000403, 0.00000337, -0.00000105, 0.00000603, -0.00001309, 0.00000418, 0.00000239, 0.00001354, -0.00001534, 0.00000851, -0.00001220, 0.00000909, -0.00001726, 0.00000764, -0.00000063, 0.00000937, -0.00001339, 0.00001499, -0.00001374, 0.00002547, -0.00004203, 0.00000937, 0.00000120, 0.00001792, -0.00002540, 0.00001327, -0.00001074, 0.00002543, -0.00005706, 0.00003462, -0.00003807, 0.00005816, -0.00010574, 0.00008128, -0.00013340, 0.00012089, -0.00067762, 0.00128290, -0.00030863, -0.00017596, 0.00005325, -0.00009606, 0.00004046, -0.00008168, 0.00008604, -0.00006000, 0.00002089, -0.00003262, 0.00002957, -0.00001790, 0.00000788, -0.00002028, 0.00003061, -0.00001000, -0.00000786, -0.00002116, 0.00002448, -0.00001755, 0.00001448, -0.00001524, 0.00002007, -0.00000908, 0.00000095, -0.00000675, 0.00000585, -0.00001153, 0.00000532, -0.00001255, 0.00002073, -0.00000552, 0.00000059, -0.00000753, 0.00000730, -0.00000347, 0.00000206, -0.00000244, 0.00000628, -0.00000644, 0.00000066, -0.00000754, 0.00000762, -0.00000638, 0.00000705, -0.00000364, 0.00000695, -0.00000064, -0.00000481, -0.00000037, 0.00000010, -0.00000172, 0.00000091, -0.00000402, 0.00000494, -0.00000426, 0.00000449, -0.00000023, 0.00000248, 0.00000030, -0.00000258, 0.00000140, -0.00000322, },
		{ 0.00000027, -0.00000135, -0.00000104, 0.00000051, -0.00000275, 0.00000260, 0.00000127, 0.00000019, 0.00000294, -0.00000242, 0.00000417, -0.00000596, 0.00000293, -0.00000305, 0.00000217, -0.00000485, -0.00000302, 0.00000450, 0.00000223, -0.00000143, 0.00000505, -0.00000398, 0.00000731, -0.00001316, 0.00000480, -0.00000483, 0.00000496, -0.00000392, 0.00000302, -0.00000047, 0.00000523, -0.00001243, 0.00000350, 0.00000381, 0.00001208, -0.00001461, 0.00000780, -0.00001159, 0.00000859, -0.00001659, 0.00000677, 0.00000077, 0.00000788, -0.00001187, 0.00001327, -0.00001191, 0.00002389, -0.00004169, 0.00000909, 0.00000296, 0.00001583, -0.00002387, 0.00001131, -0.00000763, 0.00002146, -0.00005252, 0.00002918, -0.00003133, 0.00005257, -0.00010157, 0.00007642, -0.00012965, 0.00011528, -0.00067988, 0.00131444, -0.00035137, -0.00015134, 0.00003968, -0.00008416, 0.00003148, -0.00007469, 0.00008143, -0.00005684, 0.00001824, -0.00003122, 0.00002887, -0.00001681, 0.00000612, -0.00001847, 0.00002938, -0.00000848, -0.00001081, -0.00001823, 0.00002280, -0.00001627, 0.00001344, -0.00001446, 0.00001952, -0.00000836, -0.00000015, -0.00000581, 0.00000477, -0.00001035, 0.00000418, -0.00001147, 0.00002040, -0.00000534, -0.00000012, -0.00000697, 0.00000707, -0.00000312, 0.00000157, -0.00000180, 0.00000546, -0.00000563, -0.00000028, -0.00000679, 0.00000724, -0.00000604, 0.00000696, -0.00000372, 0.00000717, -0.00000086, -0.00000504, 0.00000004, -0.00000033, -0.00000120, 0.00000031, -0.00000352, 0.00000463, -0.00000405, 0.00000453, -0.00000041, 0.00000271, 0.00000002, -0.00000246, 0.00000141, -0.00000335, },
		{ 0.00000052, -0.00000166, -0.00000081, 0.00000034, -0.00000272, 0.00000285, 0.00000094, 0.00000066, 0.00000241, -0.00000201, 0.00000385, -0.00000586, 0.00000298, -0.00000314, 0.00000226, -0.00000501, -0.00000301, 0.00000505, 0.00000150, -0.00000064, 0.00000419, -0.00000322, 0.00000654, -0.00001269, 0.00000448, -0.00000447, 0.00000484, -0.00000376, 0.00000263, 0.00000015, 0.00000438, -0.00001169, 0.00000278, 0.00000527, 0.00001051, -0.00001372, 0.00000696, -0.00001085, 0.00000796, -0.00001580, 0.00000581, 0.00000225, 0.00000627, -0.00001020, 0.00001137, -0.00000985, 0.00002198, -0.00004102, 0.00000863, 0.00000481, 0.00001359, -0.00002215, 0.00000923, -0.00000439, 0.00001724, -0.00004757, 0.00002323, -0.00002391, 0.00004599, -0.00009621, 0.00007022, -0.00012433, 0.00010744, -0.00067860, 0.00134245, -0.00039276, -0.00012629, 0.00002539, -0.00007154, 0.00002174, -0.00006682, 0.00007590, -0.00005296, 0.00001504, -0.00002935, 0.00002782, -0.00001553, 0.00000421, -0.00001648, 0.00002797, -0.00000691, -0.00001381, -0.00001512, 0.00002087, -0.00001476, 0.00001219, -0.00001349, 0.00001880, -0.00000754, -0.00000132, -0.00000479, 0.00000360, -0.00000908, 0.00000292, -0.00001024, 0.00001991, -0.00000507, -0.00000088, -0.00000633, 0.00000678, -0.00000274, 0.00000107, -0.00000114, 0.00000460, -0.00000477, -0.00000128, -0.00000595, 0.00000675, -0.00000561, 0.00000679, -0.00000372, 0.00000735, -0.00000108, -0.00000523, 0.00000043, -0.00000075, -0.00000069, -0.00000030, -0.00000297, 0.00000427, -0.00000378, 0.00000451, -0.00000056, 0.00000292, -0.00000026, -0.00000231, 0.00000139, -0.00000343, },
		{ 0.00000074, -0.00000196, -0.00000057, 0.00000015, -0.00000266, 0.00000307, 0.00000062, 0.00000113, 0.00000185, -0.00000154, 0.00000346, -0.00000569, 0.00000296, -0.00000318, 0.00000232, -0.00000516, -0.00000298, 0.00000555, 0.00000079, 0.00000017, 0.00000330, -0.00000240, 0.00000567, -0.00001211, 0.00000405, -0.00000402, 0.00000464, -0.00000355, 0.00000223, 0.00000078, 0.00000348, -0.00001089, 0.00000202, 0.00000676, 0.00000883, -0.00001269, 0.00000598, -0.00000998, 0.00000721, -0.00001490, 0.00000478, 0.00000378, 0.00000456, -0.00000840, 0.00000931, -0.00000758, 0.00001978, -0.00004004, 0.00000800, 0.00000674, 0.00001121, -0.00002027, 0.00000704, -0.00000108, 0.00001284, -0.00004227, 0.00001686, -0.00001593, 0.00003852, -0.00008973, 0.00006273, -0.00011747, 0.00009737, -0.00067369, 0.00136664, -0.00043245, -0.00010113, 0.00001058, -0.00005837, 0.00001139, -0.00005818, 0.00006954, -0.00004841, 0.00001130, -0.00002702, 0.00002643, -0.00001406, 0.00000217, -0.00001432, 0.00002641, -0.00000529, -0.00001683, -0.00001185, 0.00001870, -0.00001305, 0.00001077, -0.00001235, 0.00001793, -0.00000662, -0.00000255, -0.00000371, 0.00000237, -0.00000773, 0.00000158, -0.00000886, 0.00001926, -0.00000471, -0.00000170, -0.00000562, 0.00000643, -0.00000233, 0.00000056, -0.00000048, 0.00000372, -0.00000387, -0.00000233, -0.00000502, 0.00000617, -0.00000509, 0.00000652, -0.00000366, 0.00000748, -0.00000131, -0.00000539, 0.00000079, -0.00000115, -0.00000018, -0.00000092, -0.00000239, 0.00000385, -0.00000345, 0.00000444, -0.00000066, 0.00000311, -0.00000055, -0.00000214, 0.00000134, -0.00000348, },
		{ 0.00000092, -0.00000225, -0.00000032, -0.00000006, -0.00000257, 0.00000325, 0.00000033, 0.00000159, 0.00000127, -0.00000104, 0.00000302, -0.00000544, 0.00000288, -0.00000318, 0.00000234, -0.00000529, -0.00000291, 0.00000601, 0.00000011, 0.00000098, 0.00000237, -0.00000153, 0.00000473, -0.00001142, 0.00000353, -0.00000349, 0.00000437, -0.00000330, 0.00000180, 0.00000142, 0.00000255, -0.00001004, 0.00000125, 0.00000827, 0.00000707, -0.00001152, 0.00000487, -0.00000901, 0.00000636, -0.00001389, 0.00000368, 0.00000536, 0.00000278, -0.00000649, 0.00000713, -0.00000514, 0.00001730, -0.00003874, 0.00000720, 0.00000873, 0.00000873, -0.00001824, 0.00000478, 0.00000227, 0.00000832, -0.00003671, 0.00001016, -0.00000747, 0.00003024, -0.00008219, 0.00005402, -0.00010913, 0.00008513, -0.00066509, 0.00138678, -0.00047009, -0.00007614, -0.00000450, -0.00004484, 0.00000057, -0.00004892, 0.00006242, -0.00004324, 0.00000708, -0.00002426, 0.00002472, -0.00001241, 0.00000003, -0.00001201, 0.00002471, -0.00000366, -0.00001982, -0.00000848, 0.00001634, -0.00001114, 0.00000917, -0.00001104, 0.00001690, -0.00000563, -0.00000383, -0.00000258, 0.00000108, -0.00000632, 0.00000015, -0.00000734, 0.00001846, -0.00000427, -0.00000256, -0.00000485, 0.00000602, -0.00000190, 0.00000005, 0.00000019, 0.00000282, -0.00000296, -0.00000342, -0.00000402, 0.00000550, -0.00000448, 0.00000617, -0.00000351, 0.00000756, -0.00000154, -0.00000550, 0.00000112, -0.00000153, 0.00000032, -0.00000154, -0.00000177, 0.00000338, -0.00000306, 0.00000430, -0.00000073, 0.00000326, -0.00000084, -0.00000195, 0.00000125, -0.00000350, },
		{ 0.00000106, -0.00000251, -0.00000007, -0.00000029, -0.00000244, 0.00000339, 0.00000006, 0.00000205, 0.00000068, -0.00000051, 0.00000252, -0.00000513, 0.00000274, -0.00000314, 0.00000233, -0.00000541, -0.00000281, 0.00000641, -0.00000055, 0.00000178, 0.00000143, -0.00000062, 0.00000371, -0.00001063, 0.00000292, -0.00000289, 0.00000403, -0.00000300, 0.00000135, 0.00000206, 0.00000160, -0.00000914, 0.00000046, 0.00000976, 0.00000525, -0.00001024, 0.00000366, -0.00000793, 0.00000541, -0.00001280, 0.00000253, 0.00000695, 0.00000096, -0.00000451, 0.00000485, -0.00000257, 0.00001457, -0.00003715, 0.00000624, 0.00001076, 0.00000616, -0.00001609, 0.00000248, 0.00000559, 0.00000375, -0.00003098, 0.00000322, 0.00000135, 0.00002127, -0.00007371, 0.00004417, -0.00009940, 0.00007077, -0.00065278, 0.00140266, -0.00050534, -0.00005164, -0.00001964, -0.00003114, -0.00001053, -0.00003915, 0.00005467, -0.00003753, 0.00000242, -0.00002110, 0.00002269, -0.00001060, -0.00000220, -0.00000958, 0.00002290, -0.00000204, -0.00002275, -0.00000505, 0.00001382, -0.00000909, 0.00000744, -0.00000958, 0.00001575, -0.00000456, -0.00000514, -0.00000141, -0.00000024, -0.00000487, -0.00000133, -0.00000572, 0.00001752, -0.00000375, -0.00000346, -0.00000402, 0.00000556, -0.00000146, -0.00000046, 0.00000084, 0.00000192, -0.00000203, -0.00000452, -0.00000296, 0.00000475, -0.00000380, 0.00000574, -0.00000330, 0.00000757, -0.00000176, -0.00000559, 0.00000141, -0.00000188, 0.00000079, -0.00000215, -0.00000114, 0.00000288, -0.00000263, 0.00000412, -0.00000075, 0.00000339, -0.00000112, -0.00000173, 0.00000113, -0.00000347, },
		{ 0.00000117, -0.00000275, 0.00000017, -0.00000052, -0.00000229, 0.00000349, -0.00000018, 0.00000248, 0.00000009, 0.00000005, 0.00000198, -0.00000476, 0.00000255, -0.00000304, 0.00000228, -0.00000551, -0.00000269, 0.00000676, -0.00000115, 0.00000256, 0.00000049, 0.00000031, 0.00000264, -0.00000975, 0.00000223, -0.00000222, 0.00000363, -0.00000266, 0.00000089, 0.00000270, 0.00000063, -0.00000822, -0.00000032, 0.00001123, 0.00000339, -0.00000886, 0.00000236, -0.00000677, 0.00000438, -0.00001162, 0.00000134, 0.00000854, -0.00000089, -0.00000248, 0.00000250, 0.00000012, 0.00001163, -0.00003529, 0.00000514, 0.00001283, 0.00000354, -0.00001385, 0.00000016, 0.00000885, -0.00000081, -0.00002515, -0.00000384, 0.00001040, 0.00001175, -0.00006441, 0.00003331, -0.00008838, 0.00005441, -0.00063678, 0.00141412, -0.00053792, -0.00002790, -0.00003459, -0.00001746, -0.00002175, -0.00002905, 0.00004638, -0.00003135, -0.00000262, -0.00001757, 0.00002038, -0.00000865, -0.00000450, -0.00000705, 0.00002099, -0.00000045, -0.00002558, -0.00000159, 0.00001118, -0.00000690, 0.00000558, -0.00000800, 0.00001447, -0.00000343, -0.00000646, -0.00000022, -0.00000158, -0.00000339, -0.00000285, -0.00000400, 0.00001646, -0.00000317, -0.00000439, -0.00000315, 0.00000505, -0.00000099, -0.00000095, 0.00000148, 0.00000102, -0.00000111, -0.00000562, -0.00000186, 0.00000394, -0.00000304, 0.00000523, -0.00000300, 0.00000754, -0.00000198, -0.00000563, 0.00000167, -0.00000220, 0.00000123, -0.00000275, -0.00000049, 0.00000234, -0.00000215, 0.00000388, -0.00000072, 0.00000349, -0.00000139, -0.00000150, 0.00000098, -0.00000340, },
		{ 0.00000123, -0.00000297, 0.00000040, -0.00000077, -0.00000210, 0.00000355, -0.00000037, 0.00000288, -0.00000050, 0.00000062, 0.00000139, -0.00000433, 0.00000229, -0.00000290, 0.00000219, -0.00000558, -0.00000254, 0.00000705, -0.00000171, 0.00000330, -0.00000045, 0.00000125, 0.00000153, -0.00000879, 0.00000147, -0.00000148, 0.00000317, -0.00000228, 0.00000042, 0.00000332, -0.00000034, -0.00000726, -0.00000109, 0.00001265, 0.00000151, -0.00000742, 0.00000099, -0.00000553, 0.00000329, -0.00001039, 0.00000014, 0.00001011, -0.00000274, -0.00000043, 0.00000012, 0.00000287, 0.00000850, -0.00003317, 0.00000390, 0.00001490, 0.00000087, -0.00001154, -0.00000214, 0.00001200, -0.00000528, -0.00001932, -0.00001094, 0.00001956, 0.00000180, -0.00005440, 0.00002156, -0.00007623, 0.00003617, -0.00061717, 0.00142104, -0.00056757, -0.00000519, -0.00004914, -0.00000400, -0.00003291, -0.00001875, 0.00003768, -0.00002479, -0.00000798, -0.00001372, 0.00001781, -0.00000657, -0.00000684, -0.00000445, 0.00001900, 0.00000108, -0.00002829, 0.00000185, 0.00000844, -0.00000461, 0.00000363, -0.00000632, 0.00001309, -0.00000227, -0.00000779, 0.00000097, -0.00000292, -0.00000191, -0.00000439, -0.00000221, 0.00001528, -0.00000252, -0.00000534, -0.00000224, 0.00000451, -0.00000052, -0.00000142, 0.00000208, 0.00000015, -0.00000021, -0.00000671, -0.00000074, 0.00000308, -0.00000222, 0.00000465, -0.00000264, 0.00000743, -0.00000218, -0.00000565, 0.00000188, -0.00000247, 0.00000163, -0.00000331, 0.00000015, 0.00000178, -0.00000163, 0.00000359, -0.00000065, 0.00000355, -0.00000165, -0.00000126, 0.00000081, -0.00000329, },
		{ 0.00000126, -0.00000315, 0.00000061, -0.00000102, -0.00000188, 0.00000357, -0.00000053, 0.00000326, -0.00000107, 0.00000120, 0.00000078, -0.00000384, 0.00000199, -0.00000271, 0.00000205, -0.00000563, -0.00000237, 0.00000727, -0.00000221, 0.00000400, -0.00000135, 0.00000218, 0.00000040, -0.00000778, 0.00000066, -0.00000069, 0.00000265, -0.00000187, -0.00000005, 0.00000393, -0.00000130, -0.00000630, -0.00000182, 0.00001400, -0.00000036, -0.00000592, -0.00000044, -0.00000425, 0.00000215, -0.00000911, -0.00000107, 0.00001163, -0.00000455, 0.00000162, -0.00000226, 0.00000566, 0.00000523, -0.00003083, 0.00000254, 0.00001697, -0.00000180, -0.00000920, -0.00000439, 0.00001500, -0.00000961, -0.00001357, -0.00001796, 0.00002870, -0.00000841, -0.00004385, 0.00000907, -0.00006309, 0.00001624, -0.00059405, 0.00142336, -0.00059405, 0.00001624, -0.00006309, 0.00000907, -0.00004385, -0.00000841, 0.00002870, -0.00001796, -0.00001357, -0.00000961, 0.00001500, -0.00000439, -0.00000920, -0.00000180, 0.00001697, 0.00000254, -0.00003083, 0.00000523, 0.00000566, -0.00000226, 0.00000162, -0.00000455, 0.00001163, -0.00000107, -0.00000911, 0.00000215, -0.00000425, -0.00000044, -0.00000592, -0.00000036, 0.00001400, -0.00000182, -0.00000630, -0.00000130, 0.00000393, -0.00000005, -0.00000187, 0.00000265, -0.00000069, 0.00000066, -0.00000778, 0.00000040, 0.00000218, -0.00000135, 0.00000400, -0.00000221, 0.00000727, -0.00000237, -0.00000563, 0.00000205, -0.00000271, 0.00000199, -0.00000384, 0.00000078, 0.00000120, -0.00000107, 0.00000326, -0.00000053, 0.00000357, -0.00000188, -0.00000102, 0.00000061, -0.00000315, },
		{ 0.00000123, -0.00000329, 0.00000081, -0.00000126, -0.00000165, 0.00000355, -0.00000065, 0.00000359, -0.00000163, 0.00000178, 0.00000015, -0.00000331, 0.00000163, -0.00000247, 0.00000188, -0.00000565, -0.00000218, 0.00000743, -0.00000264, 0.00000465, -0.00000222, 0.00000308, -0.00000074, -0.00000671, -0.00000021, 0.00000015, 0.00000208, -0.00000142, -0.00000052, 0.00000451, -0.00000224, -0.00000534, -0.00000252, 0.00001528, -0.00000221, -0.00000439, -0.00000191, -0.00000292, 0.00000097, -0.00000779, -0.00000227, 0.00001309, -0.00000632, 0.00000363, -0.00000461, 0.00000844, 0.00000185, -0.00002829, 0.00000108, 0.00001900, -0.00000445, -0.00000684, -0.00000657, 0.00001781, -0.00001372, -0.00000798, -0.00002479, 0.00003768, -0.00001875, -0.00003291, -0.00000400, -0.00004914, -0.00000519, -0.00056757, 0.00142104, -0.00061717, 0.00003617, -0.00007623, 0.00002156, -0.00005440, 0.00000180, 0.00001956, -0.00001094, -0.00001932, -0.00000528, 0.00001200, -0.00000214, -0.00001154, 0.00000087, 0.00001490, 0.00000390, -0.00003317, 0.00000850, 0.00000287, 0.00000012, -0.00000043, -0.00000274, 0.00001011, 0.00000014, -0.00001039, 0.00000329, -0.00000553, 0.00000099, -0.00000742, 0.00000151, 0.00001265, -0.00000109, -0.00000726, -0.00000034, 0.00000332, 0.00000042, -0.00000228, 0.00000317, -0.00000148, 0.00000147, -0.00000879, 0.00000153, 0.00000125, -0.00000045, 0.00000330, -0.00000171, 0.00000705, -0.00000254, -0.00000558, 0.00000219, -0.00000290, 0.00000229, -0.00000433, 0.00000139, 0.00000062, -0.00000050, 0.00000288, -0.00000037, 0.00000355, -0.00000210, -0.00000077, 0.00000040, -0.00000297, },
		{ 0.00000117, -0.00000340, 0.00000098, -0.00000150, -0.00000139, 0.00000349, -0.00000072, 0.00000388, -0.00000215, 0.00000234, -0.00000049, -0.00000275, 0.00000123, -0.00000220, 0.00000167, -0.00000563, -0.00000198, 0.00000754, -0.00000300, 0.00000523, -0.00000304, 0.00000394, -0.00000186, -0.00000562, -0.00000111, 0.00000102, 0.00000148, -0.00000095, -0.00000099, 0.00000505, -0.00000315, -0.00000439, -0.00000317, 0.00001646, -0.00000400, -0.00000285, -0.00000339, -0.00000158, -0.00000022, -0.00000646, -0.00000343, 0.00001447, -0.00000800, 0.00000558, -0.00000690, 0.00001118, -0.00000159, -0.00002558, -0.00000045, 0.00002099, -0.00000705, -0.00000450, -0.00000865, 0.00002038, -0.00001757, -0.00000262, -0.00003135, 0.00004638, -0.00002905, -0.00002175, -0.00001746, -0.00003459, -0.00002790, -0.00053792, 0.00141412, -0.00063678, 0.00005441, -0.00008838, 0.00003331, -0.00006441, 0.00001175, 0.00001040, -0.00000384, -0.00002515, -0.00000081, 0.00000885, 0.00000016, -0.00001385, 0.00000354, 0.00001283, 0.00000514, -0.00003529, 0.00001163, 0.00000012, 0.00000250, -0.00000248, -0.00000089, 0.00000854, 0.00000134, -0.00001162, 0.00000438, -0.00000677, 0.00000236, -0.00000886, 0.00000339, 0.00001123, -0.00000032, -0.00000822, 0.00000063, 0.00000270, 0.00000089, -0.00000266, 0.00000363, -0.00000222, 0.00000223, -0.00000975, 0.00000264, 0.00000031, 0.00000049, 0.00000256, -0.00000115, 0.00000676, -0.00000269, -0.00000551, 0.00000228, -0.00000304, 0.00000255, -0.00000476, 0.00000198, 0.00000005, 0.00000009, 0.00000248, -0.00000018, 0.00000349, -0.00000229, -0.00000052, 0.00000017, -0.00000275, },
		{ 0.00000106, -0.00000347, 0.00000113, -0.00000173, -0.00000112, 0.00000339, -0.00000075, 0.00000412, -0.00000263, 0.00000288, -0.00000114, -0.00000215, 0.00000079, -0.00000188, 0.00000141, -0.00000559, -0.00000176, 0.00000757, -0.00000330, 0.00000574, -0.00000380, 0.00000475, -0.00000296, -0.00000452, -0.00000203, 0.00000192, 0.00000084, -0.00000046, -0.00000146, 0.00000556, -0.00000402, -0.00000346, -0.00000375, 0.00001752, -0.00000572, -0.00000133, -0.00000487, -0.00000024, -0.00000141, -0.00000514, -0.00000456, 0.00001575, -0.00000958, 0.00000744, -0.00000909, 0.00001382, -0.00000505, -0.00002275, -0.00000204, 0.00002290, -0.00000958, -0.00000220, -0.00001060, 0.00002269, -0.00002110, 0.00000242, -0.00003753, 0.00005467, -0.00003915, -0.00001053, -0.00003114, -0.00001964, -0.00005164, -0.00050534, 0.00140266, -0.00065278, 0.00007077, -0.00009940, 0.00004417, -0.00007371, 0.00002127, 0.00000135, 0.00000322, -0.00003098, 0.00000375, 0.00000559, 0.00000248, -0.00001609, 0.00000616, 0.00001076, 0.00000624, -0.00003715, 0.00001457, -0.00000257, 0.00000485, -0.00000451, 0.00000096, 0.00000695, 0.00000253, -0.00001280, 0.00000541, -0.00000793, 0.00000366, -0.00001024, 0.00000525, 0.00000976, 0.00000046, -0.00000914, 0.00000160, 0.00000206, 0.00000135, -0.00000300, 0.00000403, -0.00000289, 0.00000292, -0.00001063, 0.00000371, -0.00000062, 0.00000143, 0.00000178, -0.00000055, 0.00000641, -0.00000281, -0.00000541, 0.00000233, -0.00000314, 0.00000274, -0.00000513, 0.00000252, -0.00000051, 0.00000068, 0.00000205, 0.00000006, 0.00000339, -0.00000244, -0.00000029, -0.00000007, -0.00000251, },
		{ 0.00000092, -0.00000350, 0.00000125, -0.00000195, -0.00000084, 0.00000326, -0.00000073, 0.00000430, -0.00000306, 0.00000338, -0.00000177, -0.00000154, 0.00000032, -0.00000153, 0.00000112, -0.00000550, -0.00000154, 0.00000756, -0.00000351, 0.00000617, -0.00000448, 0.00000550, -0.00000402, -0.00000342, -0.00000296, 0.00000282, 0.00000019, 0.00000005, -0.00000190, 0.00000602, -0.00000485, -0.00000256, -0.00000427, 0.00001846, -0.00000734, 0.00000015, -0.00000632, 0.00000108, -0.00000258, -0.00000383, -0.00000563, 0.00001690, -0.00001104, 0.00000917, -0.00001114, 0.00001634, -0.00000848, -0.00001982, -0.00000366, 0.00002471, -0.00001201, 0.00000003, -0.00001241, 0.00002472, -0.00002426, 0.00000708, -0.00004324, 0.00006242, -0.00004892, 0.00000057, -0.00004484, -0.00000450, -0.00007614, -0.00047009, 0.00138678, -0.00066509, 0.00008513, -0.00010913, 0.00005402, -0.00008219, 0.00003024, -0.00000747, 0.00001016, -0.00003671, 0.00000832, 0.00000227, 0.00000478, -0.00001824, 0.00000873, 0.00000873, 0.00000720, -0.00003874, 0.00001730, -0.00000514, 0.00000713, -0.00000649, 0.00000278, 0.00000536, 0.00000368, -0.00001389, 0.00000636, -0.00000901, 0.00000487, -0.00001152, 0.00000707, 0.00000827, 0.00000125, -0.00001004, 0.00000255, 0.00000142, 0.00000180, -0.00000330, 0.00000437, -0.00000349, 0.00000353, -0.00001142, 0.00000473, -0.00000153, 0.00000237, 0.00000098, 0.00000011, 0.00000601, -0.00000291, -0.00000529, 0.00000234, -0.00000318, 0.00000288, -0.00000544, 0.00000302, -0.00000104, 0.00000127, 0.00000159, 0.00000033, 0.00000325, -0.00000257, -0.00000006, -0.00000032, -0.00000225, },
		{ 0.00000074, -0.00000348, 0.00000134, -0.00000214, -0.00000055, 0.00000311, -0.00000066, 0.00000444, -0.00000345, 0.00000385, -0.00000239, -0.00000092, -0.00000018, -0.00000115, 0.00000079, -0.00000539, -0.00000131, 0.00000748, -0.00000366, 0.00000652, -0.00000509, 0.00000617, -0.00000502, -0.00000233, -0.00000387, 0.00000372, -0.00000048, 0.00000056, -0.00000233, 0.00000643, -0.00000562, -0.00000170, -0.00000471, 0.00001926, -0.00000886, 0.00000158, -0.00000773, 0.00000237, -0.00000371, -0.00000255, -0.00000662, 0.00001793, -0.00001235, 0.00001077, -0.00001305, 0.00001870, -0.00001185, -0.00001683, -0.00000529, 0.00002641, -0.00001432, 0.00000217, -0.00001406, 0.00002643, -0.00002702, 0.00001130, -0.00004841, 0.00006954, -0.00005818, 0.00001139, -0.00005837, 0.00001058, -0.00010113, -0.00043245, 0.00136664, -0.00067369, 0.00009737, -0.00011747, 0.00006273, -0.00008973, 0.00003852, -0.00001593, 0.00001686, -0.00004227, 0.00001284, -0.00000108, 0.00000704, -0.00002027, 0.00001121, 0.00000674, 0.00000800, -0.00004004, 0.00001978, -0.00000758, 0.00000931, -0.00000840, 0.00000456, 0.00000378, 0.00000478, -0.00001490, 0.00000721, -0.00000998, 0.00000598, -0.00001269, 0.00000883, 0.00000676, 0.00000202, -0.00001089, 0.00000348, 0.00000078, 0.00000223, -0.00000355, 0.00000464, -0.00000402, 0.00000405, -0.00001211, 0.00000567, -0.00000240, 0.00000330, 0.00000017, 0.00000079, 0.00000555, -0.00000298, -0.00000516, 0.00000232, -0.00000318, 0.00000296, -0.00000569, 0.00000346, -0.00000154, 0.00000185, 0.00000113, 0.00000062, 0.00000307, -0.00000266, 0.00000015, -0.00000057, -0.00000196, },
		{ 0.00000052, -0.00000343, 0.00000139, -0.00000231, -0.00000026, 0.00000292, -0.00000056, 0.00000451, -0.00000378, 0.00000427, -0.00000297, -0.00000030, -0.00000069, -0.00000075, 0.00000043, -0.00000523, -0.00000108, 0.00000735, -0.00000372, 0.00000679, -0.00000561, 0.00000675, -0.00000595, -0.00000128, -0.00000477, 0.00000460, -0.00000114, 0.00000107, -0.00000274, 0.00000678, -0.00000633, -0.00000088, -0.00000507, 0.00001991, -0.00001024, 0.00000292, -0.00000908, 0.00000360, -0.00000479, -0.00000132, -0.00000754, 0.00001880, -0.00001349, 0.00001219, -0.00001476, 0.00002087, -0.00001512, -0.00001381, -0.00000691, 0.00002797, -0.00001648, 0.00000421, -0.00001553, 0.00002782, -0.00002935, 0.00001504, -0.00005296, 0.00007590, -0.00006682, 0.00002174, -0.00007154, 0.00002539, -0.00012629, -0.00039276, 0.00134245, -0.00067860, 0.00010744, -0.00012433, 0.00007022, -0.00009621, 0.00004599, -0.00002391, 0.00002323, -0.00004757, 0.00001724, -0.00000439, 0.00000923, -0.00002215, 0.00001359, 0.00000481, 0.00000863, -0.00004102, 0.00002198, -0.00000985, 0.00001137, -0.00001020, 0.00000627, 0.00000225, 0.00000581, -0.00001580, 0.00000796, -0.00001085, 0.00000696, -0.00001372, 0.00001051, 0.00000527, 0.00000278, -0.00001169, 0.00000438, 0.00000015, 0.00000263, -0.00000376, 0.00000484, -0.00000447, 0.00000448, -0.00001269, 0.00000654, -0.00000322, 0.00000419, -0.00000064, 0.00000150, 0.00000505, -0.00000301, -0.00000501, 0.00000226, -0.00000314, 0.00000298, -0.00000586, 0.00000385, -0.00000201, 0.00000241, 0.00000066, 0.00000094, 0.00000285, -0.00000272, 0.00000034, -0.00000081, -0.00000166, },
		{ 0.00000027, -0.00000335, 0.00000141, -0.00000246, 0.00000002, 0.00000271, -0.00000041, 0.00000453, -0.00000405, 0.00000463, -0.00000352, 0.00000031, -0.00000120, -0.00000033, 0.00000004, -0.00000504, -0.00000086, 0.00000717, -0.00000372, 0.00000696, -0.00000604, 0.00000724, -0.00000679, -0.00000028, -0.00000563, 0.00000546, -0.00000180, 0.00000157, -0.00000312, 0.00000707, -0.00000697, -0.00000012, -0.00000534, 0.00002040, -0.00001147, 0.00000418, -0.00001035, 0.00000477, -0.00000581, -0.00000015, -0.00000836, 0.00001952, -0.00001446, 0.00001344, -0.00001627, 0.00002280, -0.00001823, -0.00001081, -0.00000848, 0.00002938, -0.00001847, 0.00000612, -0.00001681, 0.00002887, -0.00003122, 0.00001824, -0.00005684, 0.00008143, -0.00007469, 0.00003148, -0.00008416, 0.00003968, -0.00015134, -0.00035137, 0.00131444, -0.00067988, 0.00011528, -0.00012965, 0.00007642, -0.00010157, 0.00005257, -0.00003133, 0.00002918, -0.00005252, 0.00002146, -0.00000763, 0.00001131, -0.00002387, 0.00001583, 0.00000296, 0.00000909, -0.00004169, 0.00002389, -0.00001191, 0.00001327, -0.00001187, 0.00000788, 0.00000077, 0.00000677, -0.00001659, 0.00000859, -0.00001159, 0.00000780, -0.00001461, 0.00001208, 0.00000381, 0.00000350, -0.00001243, 0.00000523, -0.00000047, 0.00000302, -0.00000392, 0.00000496, -0.00000483, 0.00000480, -0.00001316, 0.00000731, -0.00000398, 0.00000505, -0.00000143, 0.00000223, 0.00000450, -0.00000302, -0.00000485, 0.00000217, -0.00000305, 0.00000293, -0.00000596, 0.00000417, -0.00000242, 0.00000294, 0.00000019, 0.00000127, 0.00000260, -0.00000275, 0.00000051, -0.00000104, -0.00000135, },
		{ 0.00000000, -0.00000322, 0.00000140, -0.00000258, 0.00000030, 0.00000248, -0.00000023, 0.00000449, -0.00000426, 0.00000494, -0.00000402, 0.00000091, -0.00000172, 0.00000010, -0.00000037, -0.00000481, -0.00000064, 0.00000695, -0.00000364, 0.00000705, -0.00000638, 0.00000762, -0.00000754, 0.00000066, -0.00000644, 0.00000628, -0.00000244, 0.00000206, -0.00000347, 0.00000730, -0.00000753, 0.00000059, -0.00000552, 0.00002073, -0.00001255, 0.00000532, -0.00001153, 0.00000585, -0.00000675, 0.00000095, -0.00000908, 0.00002007, -0.00001524, 0.00001448, -0.00001755, 0.00002448, -0.00002116, -0.00000786, -0.00001000, 0.00003061, -0.00002028, 0.00000788, -0.00001790, 0.00002957, -0.00003262, 0.00002089, -0.00006000, 0.00008604, -0.00008168, 0.00004046, -0.00009606, 0.00005325, -0.00017596, -0.00030863, 0.00128290, -0.00067762, 0.00012089, -0.00013340, 0.00008128, -0.00010574, 0.00005816, -0.00003807, 0.00003462, -0.00005706, 0.00002543, -0.00001074, 0.00001327, -0.00002540, 0.00001792, 0.00000120, 0.00000937, -0.00004203, 0.00002547, -0.00001374, 0.00001499, -0.00001339, 0.00000937, -0.00000063, 0.00000764, -0.00001726, 0.00000909, -0.00001220, 0.00000851, -0.00001534, 0.00001354, 0.00000239, 0.00000418, -0.00001309, 0.00000603, -0.00000105, 0.00000337, -0.00000403, 0.00000502, -0.00000510, 0.00000502, -0.00001350, 0.00000797, -0.00000466, 0.00000584, -0.00000220, 0.00000296, 0.00000391, -0.00000300, -0.00000468, 0.00000206, -0.00000292, 0.00000283, -0.00000600, 0.00000442, -0.00000278, 0.00000343, -0.00000027, 0.00000162, 0.00000233, -0.00000273, 0.00000066, -0.00000126, -0.00000104, },
		{ -0.00000029, -0.00000307, 0.00000135, -0.00000267, 0.00000056, 0.00000224, -0.00000001, 0.00000440, -0.00000440, 0.00000518, -0.00000447, 0.00000147, -0.00000223, 0.00000054, -0.00000080, -0.00000455, -0.00000044, 0.00000669, -0.00000349, 0.00000706, -0.00000662, 0.00000790, -0.00000819, 0.00000152, -0.00000719, 0.00000704, -0.00000304, 0.00000253, -0.00000378, 0.00000746, -0.00000802, 0.00000124, -0.00000560, 0.00002089, -0.00001346, 0.00000633, -0.00001260, 0.00000684, -0.00000760, 0.00000196, -0.00000969, 0.00002044, -0.00001582, 0.00001532, -0.00001858, 0.00002588, -0.00002387, -0.00000499, -0.00001143, 0.00003166, -0.00002188, 0.00000948, -0.00001878, 0.00002994, -0.00003355, 0.00002297, -0.00006240, 0.00008968, -0.00008770, 0.00004854, -0.00010705, 0.00006587, -0.00019985, -0.00026493, 0.00124812, -0.00067197, 0.00012430, -0.00013556, 0.00008478, -0.00010869, 0.00006271, -0.00004407, 0.00003948, -0.00006112, 0.00002909, -0.00001369, 0.00001506, -0.00002673, 0.00001984, -0.00000046, 0.00000948, -0.00004204, 0.00002672, -0.00001533, 0.00001651, -0.00001474, 0.00001073, -0.00000195, 0.00000840, -0.00001780, 0.00000946, -0.00001267, 0.00000906, -0.00001590, 0.00001485, 0.00000103, 0.00000480, -0.00001366, 0.00000677, -0.00000161, 0.00000370, -0.00000409, 0.00000500, -0.00000529, 0.00000514, -0.00001371, 0.00000852, -0.00000526, 0.00000657, -0.00000293, 0.00000369, 0.00000329, -0.00000294, -0.00000451, 0.00000192, -0.00000276, 0.00000268, -0.00000595, 0.00000460, -0.00000308, 0.00000388, -0.00000071, 0.00000196, 0.00000203, -0.00000269, 0.00000077, -0.00000146, -0.00000072, },
		{ -0.00000060, -0.00000288, 0.00000127, -0.00000272, 0.00000081, 0.00000200, 0.00000024, 0.00000425, -0.00000448, 0.00000535, -0.00000485, 0.00000200, -0.00000272, 0.00000097, -0.00000125, -0.00000426, -0.00000025, 0.00000640, -0.00000329, 0.00000697, -0.00000676, 0.00000807, -0.00000872, 0.00000231, -0.00000787, 0.00000774, -0.00000361, 0.00000296, -0.00000405, 0.00000755, -0.00000843, 0.00000183, -0.00000560, 0.00002089, -0.00001419, 0.00000720, -0.00001354, 0.00000773, -0.00000835, 0.00000288, -0.00001018, 0.00002064, -0.00001619, 0.00001593, -0.00001935, 0.00002698, -0.00002633, -0.00000223, -0.00001277, 0.00003250, -0.00002325, 0.00001091, -0.00001946, 0.00002997, -0.00003400, 0.00002446, -0.00006403, 0.00009230, -0.00009264, 0.00005559, -0.00011700, 0.00007734, -0.00022271, -0.00022065, 0.00121046, -0.00066309, 0.00012556, -0.00013615, 0.00008692, -0.00011039, 0.00006615, -0.00004926, 0.00004370, -0.00006464, 0.00003238, -0.00001642, 0.00001666, -0.00002782, 0.00002157, -0.00000201, 0.00000942, -0.00004173, 0.00002763, -0.00001666, 0.00001782, -0.00001590, 0.00001195, -0.00000316, 0.00000906, -0.00001820, 0.00000969, -0.00001299, 0.00000946, -0.00001628, 0.00001601, -0.00000026, 0.00000535, -0.00001415, 0.00000743, -0.00000213, 0.00000400, -0.00000411, 0.00000491, -0.00000539, 0.00000515, -0.00001379, 0.00000895, -0.00000576, 0.00000723, -0.00000361, 0.00000440, 0.00000264, -0.00000286, -0.00000434, 0.00000177, -0.00000256, 0.00000248, -0.00000584, 0.00000471, -0.00000332, 0.00000428, -0.00000113, 0.00000231, 0.00000170, -0.00000261, 0.00000086, -0.00000164, -0.00000041, },
		{ -0.00000092, -0.00000267, 0.00000115, -0.00000274, 0.00000104, 0.00000175, 0.00000051, 0.00000405, -0.00000449, 0.00000544, -0.00000517, 0.00000248, -0.00000319, 0.00000139, -0.00000170, -0.00000393, -0.00000008, 0.00000608, -0.00000302, 0.00000681, -0.00000680, 0.00000812, -0.00000914, 0.00000299, -0.00000846, 0.00000837, -0.00000413, 0.00000337, -0.00000429, 0.00000757, -0.00000876, 0.00000235, -0.00000551, 0.00002071, -0.00001474, 0.00000793, -0.00001436, 0.00000849, -0.00000899, 0.00000371, -0.00001056, 0.00002066, -0.00001635, 0.00001631, -0.00001986, 0.00002776, -0.00002851, 0.00000038, -0.00001398, 0.00003313, -0.00002440, 0.00001215, -0.00001995, 0.00002968, -0.00003399, 0.00002537, -0.00006487, 0.00009386, -0.00009646, 0.00006152, -0.00012575, 0.00008747, -0.00024427, -0.00017620, 0.00117025, -0.00065118, 0.00012477, -0.00013522, 0.00008773, -0.00011086, 0.00006848, -0.00005360, 0.00004723, -0.00006757, 0.00003526, -0.00001891, 0.00001806, -0.00002867, 0.00002311, -0.00000345, 0.00000919, -0.00004110, 0.00002820, -0.00001771, 0.00001890, -0.00001687, 0.00001300, -0.00000425, 0.00000961, -0.00001846, 0.00000979, -0.00001317, 0.00000969, -0.00001648, 0.00001701, -0.00000146, 0.00000584, -0.00001454, 0.00000802, -0.00000261, 0.00000426, -0.00000409, 0.00000476, -0.00000541, 0.00000506, -0.00001374, 0.00000925, -0.00000616, 0.00000779, -0.00000423, 0.00000508, 0.00000198, -0.00000275, -0.00000417, 0.00000160, -0.00000234, 0.00000223, -0.00000567, 0.00000475, -0.00000349, 0.00000463, -0.00000151, 0.00000265, 0.00000136, -0.00000249, 0.00000092, -0.00000179, -0.00000011, },
		{ -0.00000125, -0.00000243, 0.00000101, -0.00000273, 0.00000124, 0.00000150, 0.00000080, 0.00000381, -0.00000445, 0.00000547, -0.00000542, 0.00000291, -0.00000362, 0.00000179, -0.00000216, -0.00000358, 0.00000008, 0.00000574, -0.00000271, 0.00000657, -0.00000676, 0.00000807, -0.00000943, 0.00000358, -0.00000897, 0.00000891, -0.00000459, 0.00000373, -0.00000448, 0.00000752, -0.00000900, 0.00000281, -0.00000533, 0.00002038, -0.00001510, 0.00000849, -0.00001503, 0.00000914, -0.00000951, 0.00000443, -0.00001082, 0.00002050, -0.00001630, 0.00001647, -0.00002010, 0.00002821, -0.00003039, 0.00000283, -0.00001505, 0.00003354, -0.00002530, 0.00001321, -0.00002024, 0.00002908, -0.00003353, 0.00002571, -0.00006495, 0.00009437, -0.00009909, 0.00006624, -0.00013321, 0.00009609, -0.00026424, -0.00013195, 0.00112788, -0.00063647, 0.00012203, -0.00013283, 0.00008724, -0.00011011, 0.00006967, -0.00005704, 0.00005002, -0.00006986, 0.00003769, -0.00002111, 0.00001922, -0.00002927, 0.00002443, -0.00000478, 0.00000881, -0.00004016, 0.00002842, -0.00001848, 0.00001974, -0.00001762, 0.00001387, -0.00000523, 0.00001004, -0.00001858, 0.00000974, -0.00001320, 0.00000977, -0.00001648, 0.00001784, -0.00000256, 0.00000624, -0.00001482, 0.00000853, -0.00000304, 0.00000449, -0.00000403, 0.00000455, -0.00000536, 0.00000487, -0.00001355, 0.00000943, -0.00000645, 0.00000826, -0.00000478, 0.00000573, 0.00000130, -0.00000261, -0.00000400, 0.00000142, -0.00000210, 0.00000194, -0.00000543, 0.00000471, -0.00000359, 0.00000492, -0.00000186, 0.00000297, 0.00000101, -0.00000235, 0.00000095, -0.00000192, 0.00000017, },
		{ -0.00000158, -0.00000218, 0.00000083, -0.00000268, 0.00000142, 0.00000126, 0.00000110, 0.00000353, -0.00000434, 0.00000542, -0.00000559, 0.00000328, -0.00000401, 0.00000217, -0.00000261, -0.00000320, 0.00000021, 0.00000538, -0.00000235, 0.00000626, -0.00000662, 0.00000790, -0.00000960, 0.00000407, -0.00000937, 0.00000937, -0.00000498, 0.00000404, -0.00000462, 0.00000741, -0.00000916, 0.00000321, -0.00000507, 0.00001988, -0.00001528, 0.00000890, -0.00001556, 0.00000965, -0.00000991, 0.00000504, -0.00001097, 0.00002017, -0.00001605, 0.00001640, -0.00002007, 0.00002832, -0.00003195, 0.00000508, -0.00001597, 0.00003372, -0.00002595, 0.00001408, -0.00002034, 0.00002821, -0.00003266, 0.00002550, -0.00006426, 0.00009382, -0.00010052, 0.00006970, -0.00013925, 0.00010307, -0.00028239, -0.00008828, 0.00108373, -0.00061922, 0.00011748, -0.00012907, 0.00008554, -0.00010819, 0.00006974, -0.00005957, 0.00005205, -0.00007148, 0.00003962, -0.00002301, 0.00002014, -0.00002960, 0.00002553, -0.00000598, 0.00000829, -0.00003894, 0.00002830, -0.00001896, 0.00002034, -0.00001817, 0.00001457, -0.00000606, 0.00001034, -0.00001856, 0.00000956, -0.00001309, 0.00000969, -0.00001630, 0.00001848, -0.00000356, 0.00000656, -0.00001500, 0.00000894, -0.00000341, 0.00000469, -0.00000394, 0.00000429, -0.00000524, 0.00000460, -0.00001325, 0.00000948, -0.00000663, 0.00000863, -0.00000526, 0.00000634, 0.00000062, -0.00000245, -0.00000384, 0.00000124, -0.00000184, 0.00000162, -0.00000514, 0.00000461, -0.00000363, 0.00000514, -0.00000217, 0.00000327, 0.00000066, -0.00000219, 0.00000095, -0.00000202, 0.00000044, },
		{ -0.00000190, -0.00000191, 0.00000063, -0.00000259, 0.00000157, 0.00000103, 0.00000142, 0.00000321, -0.00000417, 0.00000530, -0.00000568, 0.00000359, -0.00000436, 0.00000251, -0.00000305, -0.00000280, 0.00000031, 0.00000502, -0.00000195, 0.00000589, -0.00000641, 0.00000764, -0.00000965, 0.00000444, -0.00000968, 0.00000973, -0.00000531, 0.00000431, -0.00000472, 0.00000722, -0.00000923, 0.00000354, -0.00000474, 0.00001924, -0.00001527, 0.00000915, -0.00001594, 0.00001004, -0.00001019, 0.00000555, -0.00001100, 0.00001968, -0.00001560, 0.00001610, -0.00001977, 0.00002810, -0.00003318, 0.00000713, -0.00001673, 0.00003367, -0.00002636, 0.00001474, -0.00002027, 0.00002707, -0.00003140, 0.00002478, -0.00006286, 0.00009224, -0.00010074, 0.00007184, -0.00014382, 0.00010828, -0.00029847, -0.00004558, 0.00103819, -0.00059968, 0.00011129, -0.00012405, 0.00008269, -0.00010518, 0.00006873, -0.00006119, 0.00005332, -0.00007242, 0.00004104, -0.00002456, 0.00002081, -0.00002966, 0.00002641, -0.00000707, 0.00000764, -0.00003745, 0.00002786, -0.00001916, 0.00002069, -0.00001849, 0.00001508, -0.00000677, 0.00001053, -0.00001841, 0.00000924, -0.00001284, 0.00000946, -0.00001593, 0.00001895, -0.00000444, 0.00000679, -0.00001507, 0.00000927, -0.00000373, 0.00000486, -0.00000381, 0.00000398, -0.00000505, 0.00000424, -0.00001282, 0.00000940, -0.00000669, 0.00000889, -0.00000564, 0.00000689, -0.00000005, -0.00000226, -0.00000369, 0.00000106, -0.00000158, 0.00000127, -0.00000479, 0.00000445, -0.00000359, 0.00000530, -0.00000243, 0.00000355, 0.00000030, -0.00000200, 0.00000092, -0.00000208, 0.00000067, },
		{ -0.00000221, -0.00000164, 0.00000042, -0.00000248, 0.00000169, 0.00000081, 0.00000173, 0.00000287, -0.00000396, 0.00000512, -0.00000570, 0.00000384, -0.00000465, 0.00000282, -0.00000347, -0.00000239, 0.00000039, 0.00000465, -0.00000153, 0.00000547, -0.00000612, 0.00000727, -0.00000958, 0.00000471, -0.00000988, 0.00001000, -0.00000556, 0.00000452, -0.00000477, 0.00000697, -0.00000923, 0.00000382, -0.00000435, 0.00001846, -0.00001509, 0.00000923, -0.00001616, 0.00001029, -0.00001035, 0.00000594, -0.00001092, 0.00001903, -0.00001496, 0.00001560, -0.00001922, 0.00002755, -0.00003406, 0.00000895, -0.00001732, 0.00003339, -0.00002650, 0.00001521, -0.00002003, 0.00002571, -0.00002979, 0.00002359, -0.00006078, 0.00008966, -0.00009976, 0.00007266, -0.00014687, 0.00011164, -0.00031230, -0.00000418, 0.00099165, -0.00057814, 0.00010361, -0.00011788, 0.00007880, -0.00010115, 0.00006668, -0.00006192, 0.00005381, -0.00007267, 0.00004192, -0.00002577, 0.00002120, -0.00002946, 0.00002706, -0.00000804, 0.00000688, -0.00003572, 0.00002711, -0.00001908, 0.00002080, -0.00001861, 0.00001540, -0.00000733, 0.00001059, -0.00001812, 0.00000880, -0.00001246, 0.00000908, -0.00001538, 0.00001922, -0.00000521, 0.00000693, -0.00001503, 0.00000949, -0.00000400, 0.00000499, -0.00000366, 0.00000363, -0.00000481, 0.00000381, -0.00001228, 0.00000920, -0.00000664, 0.00000904, -0.00000594, 0.00000738, -0.00000071, -0.00000206, -0.00000354, 0.00000087, -0.00000131, 0.00000091, -0.00000441, 0.00000423, -0.00000350, 0.00000539, -0.00000264, 0.00000379, -0.00000005, -0.00000179, 0.00000086, -0.00000211, 0.00000089, },
		{ -0.00000250, -0.00000137, 0.00000018, -0.00000233, 0.00000177, 0.00000062, 0.00000204, 0.00000250, -0.00000370, 0.00000486, -0.00000564, 0.00000401, -0.00000489, 0.00000308, -0.00000387, -0.00000196, 0.00000044, 0.00000428, -0.00000108, 0.00000499, -0.00000576, 0.00000682, -0.00000940, 0.00000487, -0.00000998, 0.00001016, -0.00000573, 0.00000467, -0.00000477, 0.00000666, -0.00000915, 0.00000404, -0.00000389, 0.00001755, -0.00001474, 0.00000916, -0.00001624, 0.00001042, -0.00001038, 0.00000623, -0.00001074, 0.00001824, -0.00001415, 0.00001490, -0.00001842, 0.00002667, -0.00003460, 0.00001053, -0.00001774, 0.00003288, -0.00002640, 0.00001549, -0.00001965, 0.00002416, -0.00002788, 0.00002198, -0.00005809, 0.00008615, -0.00009761, 0.00007215, -0.00014837, 0.00011306, -0.00032371, 0.00003558, 0.00094451, -0.00055491, 0.00009465, -0.00011072, 0.00007399, -0.00009620, 0.00006365, -0.00006180, 0.00005356, -0.00007223, 0.00004226, -0.00002660, 0.00002132, -0.00002898, 0.00002748, -0.00000890, 0.00000603, -0.00003376, 0.00002606, -0.00001873, 0.00002067, -0.00001851, 0.00001554, -0.00000775, 0.00001054, -0.00001771, 0.00000823, -0.00001195, 0.00000857, -0.00001465, 0.00001932, -0.00000585, 0.00000697, -0.00001488, 0.00000963, -0.00000420, 0.00000509, -0.00000348, 0.00000324, -0.00000452, 0.00000331, -0.00001164, 0.00000889, -0.00000648, 0.00000908, -0.00000614, 0.00000780, -0.00000136, -0.00000184, -0.00000341, 0.00000069, -0.00000104, 0.00000053, -0.00000399, 0.00000395, -0.00000334, 0.00000542, -0.00000279, 0.00000400, -0.00000039, -0.00000157, 0.00000078, -0.00000211, 0.00000107, },
		{ -0.00000278, -0.00000109, -0.00000007, -0.00000216, 0.00000183, 0.00000045, 0.00000233, 0.00000212, -0.00000340, 0.00000455, -0.00000550, 0.00000412, -0.00000507, 0.00000330, -0.00000424, -0.00000152, 0.00000047, 0.00000393, -0.00000062, 0.00000448, -0.00000535, 0.00000629, -0.00000911, 0.00000492, -0.00000997, 0.00001022, -0.00000582, 0.00000477, -0.00000473, 0.00000629, -0.00000900, 0.00000421, -0.00000339, 0.00001652, -0.00001423, 0.00000893, -0.00001617, 0.00001041, -0.00001030, 0.00000641, -0.00001047, 0.00001731, -0.00001317, 0.00001401, -0.00001740, 0.00002549, -0.00003479, 0.00001187, -0.00001797, 0.00003215, -0.00002605, 0.00001558, -0.00001913, 0.00002245, -0.00002572, 0.00002000, -0.00005484, 0.00008178, -0.00009435, 0.00007034, -0.00014833, 0.00011254, -0.00033256, 0.00007340, 0.00089713, -0.00053026, 0.00008460, -0.00010270, 0.00006838, -0.00009046, 0.00005974, -0.00006085, 0.00005257, -0.00007111, 0.00004205, -0.00002707, 0.00002117, -0.00002823, 0.00002766, -0.00000965, 0.00000510, -0.00003162, 0.00002475, -0.00001813, 0.00002032, -0.00001821, 0.00001549, -0.00000804, 0.00001037, -0.00001718, 0.00000755, -0.00001133, 0.00000793, -0.00001377, 0.00001923, -0.00000637, 0.00000692, -0.00001462, 0.00000966, -0.00000434, 0.00000515, -0.00000329, 0.00000283, -0.00000419, 0.00000276, -0.00001091, 0.00000846, -0.00000621, 0.00000901, -0.00000625, 0.00000815, -0.00000197, -0.00000161, -0.00000329, 0.00000053, -0.00000077, 0.00000014, -0.00000354, 0.00000363, -0.00000312, 0.00000539, -0.00000290, 0.00000417, -0.00000072, -0.00000133, 0.00000067, -0.00000207, 0.00000121, },
		{ -0.00000303, -0.00000083, -0.00000032, -0.00000196, 0.00000185, 0.00000030, 0.00000261, 0.00000172, -0.00000307, 0.00000419, -0.00000529, 0.00000415, -0.00000518, 0.00000346, -0.00000458, -0.00000108, 0.00000047, 0.00000358, -0.00000016, 0.00000395, -0.00000489, 0.00000569, -0.00000872, 0.00000487, -0.00000986, 0.00001017, -0.00000583, 0.00000480, -0.00000464, 0.00000587, -0.00000877, 0.00000434, -0.00000285, 0.00001540, -0.00001357, 0.00000856, -0.00001596, 0.00001029, -0.00001010, 0.00000650, -0.00001012, 0.00001628, -0.00001205, 0.00001296, -0.00001617, 0.00002401, -0.00003465, 0.00001297, -0.00001804, 0.00003120, -0.00002546, 0.00001549, -0.00001850, 0.00002061, -0.00002335, 0.00001771, -0.00005112, 0.00007662, -0.00009004, 0.00006729, -0.00014677, 0.00011005, -0.00033876, 0.00010900, 0.00084989, -0.00050451, 0.00007366, -0.00009398, 0.00006211, -0.00008403, 0.00005502, -0.00005915, 0.00005091, -0.00006935, 0.00004130, -0.00002716, 0.00002076, -0.00002722, 0.00002763, -0.00001029, 0.00000413, -0.00002932, 0.00002320, -0.00001730, 0.00001976, -0.00001772, 0.00001528, -0.00000820, 0.00001010, -0.00001655, 0.00000678, -0.00001061, 0.00000719, -0.00001273, 0.00001897, -0.00000677, 0.00000679, -0.00001426, 0.00000960, -0.00000442, 0.00000519, -0.00000309, 0.00000241, -0.00000384, 0.00000217, -0.00001011, 0.00000794, -0.00000583, 0.00000884, -0.00000626, 0.00000841, -0.00000255, -0.00000138, -0.00000318, 0.00000037, -0.00000052, -0.00000024, -0.00000307, 0.00000327, -0.00000285, 0.00000529, -0.00000294, 0.00000430, -0.00000102, -0.00000109, 0.00000054, -0.00000200, 0.00000132, },
		{ -0.00000325, -0.00000058, -0.00000058, -0.00000174, 0.00000184, 0.00000019, 0.00000287, 0.00000133, -0.00000271, 0.00000378, -0.00000502, 0.00000412, -0.00000524, 0.00000357, -0.00000488, -0.00000064, 0.00000044, 0.00000326, 0.00000030, 0.00000339, -0.00000440, 0.00000503, -0.00000825, 0.00000472, -0.00000965, 0.00001003, -0.00000576, 0.00000478, -0.00000451, 0.00000539, -0.00000849, 0.00000443, -0.00000229, 0.00001419, -0.00001279, 0.00000805, -0.00001562, 0.00001005, -0.00000980, 0.00000650, -0.00000969, 0.00001514, -0.00001081, 0.00001177, -0.00001475, 0.00002227, -0.00003418, 0.00001381, -0.00001793, 0.00003006, -0.00002465, 0.00001523, -0.00001777, 0.00001869, -0.00002083, 0.00001518, -0.00004700, 0.00007077, -0.00008478, 0.00006305, -0.00014374, 0.00010564, -0.00034223, 0.00014214, 0.00080312, -0.00047794, 0.00006205, -0.00008471, 0.00005532, -0.00007706, 0.00004962, -0.00005677, 0.00004860, -0.00006697, 0.00004004, -0.00002688, 0.00002008, -0.00002597, 0.00002738, -0.00001084, 0.00000312, -0.00002690, 0.00002144, -0.00001625, 0.00001899, -0.00001706, 0.00001490, -0.00000823, 0.00000974, -0.00001582, 0.00000592, -0.00000980, 0.00000635, -0.00001156, 0.00001855, -0.00000705, 0.00000657, -0.00001380, 0.00000944, -0.00000443, 0.00000519, -0.00000288, 0.00000198, -0.00000347, 0.00000155, -0.00000923, 0.00000733, -0.00000537, 0.00000856, -0.00000616, 0.00000860, -0.00000310, -0.00000113, -0.00000308, 0.00000023, -0.00000028, -0.00000061, -0.00000259, 0.00000288, -0.00000254, 0.00000514, -0.00000294, 0.00000438, -0.00000130, -0.00000085, 0.00000039, -0.00000190, 0.00000140, },
		{ -0.00000344, -0.00000034, -0.00000084, -0.00000151, 0.00000180, 0.00000010, 0.00000311, 0.00000094, -0.00000234, 0.00000333, -0.00000468, 0.00000402, -0.00000523, 0.00000362, -0.00000513, -0.00000021, 0.00000039, 0.00000295, 0.00000074, 0.00000283, -0.00000388, 0.00000433, -0.00000769, 0.00000449, -0.00000934, 0.00000979, -0.00000560, 0.00000469, -0.00000434, 0.00000487, -0.00000814, 0.00000448, -0.00000172, 0.00001292, -0.00001189, 0.00000741, -0.00001515, 0.00000970, -0.00000940, 0.00000641, -0.00000920, 0.00001393, -0.00000947, 0.00001046, -0.00001319, 0.00002028, -0.00003339, 0.00001441, -0.00001765, 0.00002872, -0.00002363, 0.00001480, -0.00001696, 0.00001672, -0.00001821, 0.00001248, -0.00004258, 0.00006434, -0.00007867, 0.00005773, -0.00013933, 0.00009935, -0.00034296, 0.00017263, 0.00075716, -0.00045084, 0.00004996, -0.00007506, 0.00004815, -0.00006968, 0.00004364, -0.00005378, 0.00004573, -0.00006402, 0.00003828, -0.00002624, 0.00001915, -0.00002448, 0.00002692, -0.00001131, 0.00000211, -0.00002438, 0.00001949, -0.00001502, 0.00001805, -0.00001623, 0.00001438, -0.00000815, 0.00000929, -0.00001502, 0.00000498, -0.00000892, 0.00000543, -0.00001027, 0.00001797, -0.00000722, 0.00000626, -0.00001325, 0.00000919, -0.00000438, 0.00000517, -0.00000267, 0.00000155, -0.00000309, 0.00000091, -0.00000831, 0.00000665, -0.00000482, 0.00000819, -0.00000598, 0.00000869, -0.00000360, -0.00000089, -0.00000299, 0.00000010, -0.00000006, -0.00000097, -0.00000212, 0.00000246, -0.00000218, 0.00000493, -0.00000288, 0.00000442, -0.00000156, -0.00000061, 0.00000024, -0.00000177, 0.00000143, },
		{ -0.00000360, -0.00000012, -0.00000109, -0.00000125, 0.00000173, 0.00000005, 0.00000332, 0.00000057, -0.00000195, 0.00000284, -0.00000429, 0.00000385, -0.00000516, 0.00000361, -0.00000534, 0.00000021, 0.00000032, 0.00000267, 0.00000118, 0.00000227, -0.00000334, 0.00000359, -0.00000707, 0.00000418, -0.00000896, 0.00000946, -0.00000537, 0.00000455, -0.00000414, 0.00000432, -0.00000775, 0.00000450, -0.00000114, 0.00001160, -0.00001088, 0.00000667, -0.00001457, 0.00000926, -0.00000892, 0.00000625, -0.00000867, 0.00001266, -0.00000805, 0.00000905, -0.00001149, 0.00001808, -0.00003232, 0.00001478, -0.00001723, 0.00002721, -0.00002240, 0.00001422, -0.00001609, 0.00001473, -0.00001555, 0.00000967, -0.00003794, 0.00005744, -0.00007182, 0.00005142, -0.00013364, 0.00009129, -0.00034094, 0.00020029, 0.00071229, -0.00042348, 0.00003759, -0.00006519, 0.00004075, -0.00006204, 0.00003721, -0.00005028, 0.00004236, -0.00006055, 0.00003606, -0.00002528, 0.00001799, -0.00002278, 0.00002626, -0.00001169, 0.00000110, -0.00002181, 0.00001740, -0.00001361, 0.00001695, -0.00001527, 0.00001372, -0.00000796, 0.00000877, -0.00001414, 0.00000399, -0.00000798, 0.00000445, -0.00000889, 0.00001725, -0.00000729, 0.00000589, -0.00001261, 0.00000886, -0.00000427, 0.00000512, -0.00000246, 0.00000112, -0.00000272, 0.00000028, -0.00000735, 0.00000590, -0.00000420, 0.00000772, -0.00000570, 0.00000871, -0.00000405, -0.00000065, -0.00000291, -0.00000001, 0.00000013, -0.00000131, -0.00000164, 0.00000204, -0.00000179, 0.00000467, -0.00000277, 0.00000441, -0.00000178, -0.00000038, 0.00000007, -0.00000162, 0.00000143, },
		{ -0.00000372, 0.00000007, -0.00000133, -0.00000099, 0.00000163, 0.00000003, 0.00000349, 0.00000021, -0.00000157, 0.00000234, -0.00000385, 0.00000363, -0.00000502, 0.00000354, -0.00000550, 0.00000062, 0.00000022, 0.00000241, 0.00000158, 0.00000171, -0.00000281, 0.00000284, -0.00000640, 0.00000380, -0.00000849, 0.00000904, -0.00000507, 0.00000435, -0.00000390, 0.00000373, -0.00000730, 0.00000451, -0.00000057, 0.00001025, -0.00000980, 0.00000583, -0.00001389, 0.00000873, -0.00000836, 0.00000603, -0.00000810, 0.00001134, -0.00000657, 0.00000758, -0.00000969, 0.00001570, -0.00003098, 0.00001493, -0.00001665, 0.00002555, -0.00002100, 0.00001351, -0.00001518, 0.00001277, -0.00001289, 0.00000682, -0.00003318, 0.00005019, -0.00006435, 0.00004425, -0.00012677, 0.00008156, -0.00033624, 0.00022499, 0.00066878, -0.00039613, 0.00002515, -0.00005524, 0.00003324, -0.00005427, 0.00003045, -0.00004637, 0.00003856, -0.00005664, 0.00003342, -0.00002399, 0.00001662, -0.00002087, 0.00002542, -0.00001201, 0.00000013, -0.00001922, 0.00001520, -0.00001208, 0.00001572, -0.00001418, 0.00001294, -0.00000769, 0.00000818, -0.00001321, 0.00000296, -0.00000700, 0.00000343, -0.00000742, 0.00001640, -0.00000725, 0.00000545, -0.00001189, 0.00000845, -0.00000411, 0.00000504, -0.00000226, 0.00000072, -0.00000236, -0.00000035, -0.00000637, 0.00000511, -0.00000351, 0.00000718, -0.00000533, 0.00000863, -0.00000445, -0.00000042, -0.00000284, -0.00000011, 0.00000030, -0.00000162, -0.00000118, 0.00000160, -0.00000138, 0.00000437, -0.00000262, 0.00000435, -0.00000196, -0.00000016, -0.00000011, -0.00000144, 0.00000139, },
		{ -0.00000381, 0.00000023, -0.00000156, -0.00000073, 0.00000152, 0.00000003, 0.00000362, -0.00000013, -0.00000119, 0.00000182, -0.00000337, 0.00000335, -0.00000483, 0.00000341, -0.00000561, 0.00000101, 0.00000011, 0.00000219, 0.00000196, 0.00000117, -0.00000228, 0.00000208, -0.00000568, 0.00000337, -0.00000796, 0.00000855, -0.00000470, 0.00000410, -0.00000363, 0.00000312, -0.00000682, 0.00000450, -0.00000001, 0.00000888, -0.00000866, 0.00000491, -0.00001311, 0.00000814, -0.00000773, 0.00000575, -0.00000751, 0.00001001, -0.00000506, 0.00000605, -0.00000783, 0.00001318, -0.00002940, 0.00001487, -0.00001596, 0.00002375, -0.00001943, 0.00001266, -0.00001423, 0.00001086, -0.00001030, 0.00000400, -0.00002839, 0.00004270, -0.00005641, 0.00003635, -0.00011887, 0.00007030, -0.00032894, 0.00024665, 0.00062688, -0.00036902, 0.00001281, -0.00004537, 0.00002578, -0.00004652, 0.00002349, -0.00004214, 0.00003442, -0.00005234, 0.00003041, -0.00002243, 0.00001505, -0.00001879, 0.00002440, -0.00001226, -0.00000079, -0.00001663, 0.00001293, -0.00001043, 0.00001437, -0.00001299, 0.00001207, -0.00000733, 0.00000755, -0.00001224, 0.00000191, -0.00000599, 0.00000238, -0.00000589, 0.00001545, -0.00000713, 0.00000496, -0.00001111, 0.00000797, -0.00000389, 0.00000494, -0.00000206, 0.00000033, -0.00000202, -0.00000096, -0.00000539, 0.00000428, -0.00000278, 0.00000657, -0.00000489, 0.00000847, -0.00000479, -0.00000020, -0.00000278, -0.00000018, 0.00000044, -0.00000190, -0.00000075, 0.00000117, -0.00000095, 0.00000403, -0.00000242, 0.00000425, -0.00000212, 0.00000005, -0.00000028, -0.00000124, 0.00000132, },
		{ -0.00000386, 0.00000036, -0.00000177, -0.00000046, 0.00000138, 0.00000007, 0.00000372, -0.00000043, -0.00000082, 0.00000130, -0.00000285, 0.00000303, -0.00000459, 0.00000323, -0.00000566, 0.00000137, -0.00000002, 0.00000199, 0.00000231, 0.00000066, -0.00000176, 0.00000132, -0.00000494, 0.00000289, -0.00000738, 0.00000800, -0.00000427, 0.00000380, -0.00000333, 0.00000248, -0.00000631, 0.00000447, 0.00000051, 0.00000752, -0.00000747, 0.00000393, -0.00001227, 0.00000748, -0.00000706, 0.00000543, -0.00000691, 0.00000867, -0.00000354, 0.00000451, -0.00000593, 0.00001054, -0.00002761, 0.00001462, -0.00001514, 0.00002184, -0.00001773, 0.00001171, -0.00001328, 0.00000903, -0.00000780, 0.00000127, -0.00002365, 0.00003510, -0.00004813, 0.00002787, -0.00011010, 0.00005769, -0.00031915, 0.00026523, 0.00058679, -0.00034237, 0.00000074, -0.00003570, 0.00001847, -0.00003892, 0.00001646, -0.00003770, 0.00003003, -0.00004773, 0.00002708, -0.00002061, 0.00001333, -0.00001656, 0.00002323, -0.00001246, -0.00000165, -0.00001409, 0.00001061, -0.00000870, 0.00001294, -0.00001173, 0.00001111, -0.00000690, 0.00000688, -0.00001125, 0.00000084, -0.00000497, 0.00000133, -0.00000432, 0.00001439, -0.00000693, 0.00000442, -0.00001026, 0.00000742, -0.00000362, 0.00000482, -0.00000189, -0.00000003, -0.00000171, -0.00000153, -0.00000441, 0.00000344, -0.00000201, 0.00000590, -0.00000437, 0.00000823, -0.00000507, 0.00000001, -0.00000272, -0.00000024, 0.00000055, -0.00000214, -0.00000034, 0.00000076, -0.00000052, 0.00000367, -0.00000218, 0.00000411, -0.00000223, 0.00000024, -0.00000046, -0.00000102, 0.00000121, },
		{ -0.00000388, 0.00000047, -0.00000195, -0.00000020, 0.00000122, 0.00000014, 0.00000378, -0.00000070, -0.00000048, 0.00000077, -0.00000232, 0.00000267, -0.00000430, 0.00000300, -0.00000566, 0.00000171, -0.00000016, 0.00000182, 0.00000262, 0.00000017, -0.00000127, 0.00000059, -0.00000418, 0.00000239, -0.00000675, 0.00000738, -0.00000379, 0.00000347, -0.00000302, 0.00000184, -0.00000576, 0.00000444, 0.00000099, 0.00000618, -0.00000626, 0.00000289, -0.00001136, 0.00000678, -0.00000634, 0.00000508, -0.00000632, 0.00000735, -0.00000203, 0.00000298, -0.00000403, 0.00000783, -0.00002563, 0.00001419, -0.00001424, 0.00001985, -0.00001590, 0.00001066, -0.00001232, 0.00000731, -0.00000546, -0.00000130, -0.00001906, 0.00002751, -0.00003964, 0.00001895, -0.00010061, 0.00004391, -0.00030702, 0.00028070, 0.00054867, -0.00031638, -0.00001090, -0.00002637, 0.00001144, -0.00003160, 0.00000948, -0.00003315, 0.00002548, -0.00004289, 0.00002349, -0.00001857, 0.00001146, -0.00001420, 0.00002192, -0.00001262, -0.00000242, -0.00001161, 0.00000828, -0.00000693, 0.00001145, -0.00001041, 0.00001009, -0.00000643, 0.00000619, -0.00001026, -0.00000023, -0.00000396, 0.00000028, -0.00000273, 0.00001326, -0.00000667, 0.00000384, -0.00000938, 0.00000681, -0.00000330, 0.00000468, -0.00000173, -0.00000036, -0.00000143, -0.00000206, -0.00000346, 0.00000259, -0.00000121, 0.00000518, -0.00000379, 0.00000791, -0.00000530, 0.00000020, -0.00000266, -0.00000028, 0.00000063, -0.00000235, 0.00000004, 0.00000036, -0.00000008, 0.00000328, -0.00000191, 0.00000392, -0.00000230, 0.00000041, -0.00000063, -0.00000079, 0.00000108, },
		{ -0.00000386, 0.00000054, -0.00000212, 0.00000006, 0.00000106, 0.00000023, 0.00000381, -0.00000094, -0.00000016, 0.00000027, -0.00000176, 0.00000227, -0.00000397, 0.00000271, -0.00000560, 0.00000201, -0.00000031, 0.00000168, 0.00000289, -0.00000027, -0.00000082, -0.00000012, -0.00000343, 0.00000186, -0.00000609, 0.00000672, -0.00000326, 0.00000309, -0.00000269, 0.00000119, -0.00000520, 0.00000441, 0.00000143, 0.00000488, -0.00000504, 0.00000183, -0.00001041, 0.00000605, -0.00000560, 0.00000471, -0.00000575, 0.00000607, -0.00000056, 0.00000148, -0.00000215, 0.00000508, -0.00002351, 0.00001362, -0.00001326, 0.00001779, -0.00001398, 0.00000952, -0.00001137, 0.00000572, -0.00000330, -0.00000368, -0.00001469, 0.00002005, -0.00003109, 0.00000976, -0.00009059, 0.00002915, -0.00029273, 0.00029311, 0.00051266, -0.00029122, -0.00002198, -0.00001747, 0.00000479, -0.00002467, 0.00000266, -0.00002859, 0.00002085, -0.00003789, 0.00001969, -0.00001636, 0.00000949, -0.00001175, 0.00002049, -0.00001273, -0.00000310, -0.00000924, 0.00000597, -0.00000513, 0.00000992, -0.00000907, 0.00000903, -0.00000592, 0.00000549, -0.00000927, -0.00000127, -0.00000298, -0.00000073, -0.00000115, 0.00001207, -0.00000634, 0.00000324, -0.00000846, 0.00000616, -0.00000294, 0.00000452, -0.00000159, -0.00000065, -0.00000120, -0.00000254, -0.00000255, 0.00000175, -0.00000041, 0.00000443, -0.00000315, 0.00000752, -0.00000547, 0.00000037, -0.00000261, -0.00000031, 0.00000069, -0.00000252, 0.00000038, -0.00000002, 0.00000034, 0.00000288, -0.00000161, 0.00000370, -0.00000234, 0.00000056, -0.00000079, -0.00000055, 0.00000092, },
		{ -0.00000381, 0.00000057, -0.00000225, 0.00000030, 0.00000088, 0.00000035, 0.00000379, -0.00000114, 0.00000013, -0.00000022, -0.00000121, 0.00000185, -0.00000360, 0.00000238, -0.00000549, 0.00000229, -0.00000047, 0.00000157, 0.00000312, -0.00000068, -0.00000040, -0.00000079, -0.00000269, 0.00000133, -0.00000541, 0.00000603, -0.00000270, 0.00000269, -0.00000235, 0.00000054, -0.00000462, 0.00000439, 0.00000181, 0.00000363, -0.00000383, 0.00000076, -0.00000943, 0.00000531, -0.00000485, 0.00000433, -0.00000520, 0.00000483, 0.00000086, 0.00000004, -0.00000033, 0.00000233, -0.00002127, 0.00001292, -0.00001222, 0.00001569, -0.00001199, 0.00000832, -0.00001045, 0.00000427, -0.00000136, -0.00000580, -0.00001062, 0.00001282, -0.00002263, 0.00000044, -0.00008021, 0.00001365, -0.00027650, 0.00030251, 0.00047886, -0.00026703, -0.00003239, -0.00000910, -0.00000140, -0.00001823, -0.00000387, -0.00002412, 0.00001624, -0.00003282, 0.00001575, -0.00001402, 0.00000745, -0.00000922, 0.00001895, -0.00001282, -0.00000367, -0.00000699, 0.00000372, -0.00000334, 0.00000838, -0.00000771, 0.00000795, -0.00000538, 0.00000480, -0.00000829, -0.00000228, -0.00000202, -0.00000170, 0.00000041, 0.00001084, -0.00000598, 0.00000263, -0.00000752, 0.00000547, -0.00000254, 0.00000434, -0.00000146, -0.00000089, -0.00000101, -0.00000295, -0.00000168, 0.00000093, 0.00000038, 0.00000366, -0.00000248, 0.00000707, -0.00000557, 0.00000052, -0.00000256, -0.00000032, 0.00000071, -0.00000264, 0.00000067, -0.00000036, 0.00000076, 0.00000247, -0.00000129, 0.00000344, -0.00000234, 0.00000068, -0.00000093, -0.00000031, 0.00000073, },
		{ -0.00000373, 0.00000058, -0.00000235, 0.00000053, 0.00000070, 0.00000049, 0.00000373, -0.00000130, 0.00000039, -0.00000069, -0.00000066, 0.00000142, -0.00000321, 0.00000201, -0.00000533, 0.00000252, -0.00000063, 0.00000149, 0.00000331, -0.00000104, -0.00000003, -0.00000141, -0.00000197, 0.00000080, -0.00000472, 0.00000532, -0.00000212, 0.00000226, -0.00000201, -0.00000010, -0.00000404, 0.00000437, 0.00000214, 0.00000244, -0.00000265, -0.00000031, -0.00000843, 0.00000455, -0.00000409, 0.00000395, -0.00000470, 0.00000366, 0.00000220, -0.00000132, 0.00000141, -0.00000038, -0.00001895, 0.00001212, -0.00001114, 0.00001357, -0.00000996, 0.00000706, -0.00000956, 0.00000299, 0.00000034, -0.00000763, -0.00000691, 0.00000592, -0.00001437, -0.00000884, -0.00006965, -0.00000239, -0.00025854, 0.00030901, 0.00044733, -0.00024392, -0.00004202, -0.00000133, -0.00000705, -0.00001238, -0.00001003, -0.00001984, 0.00001172, -0.00002775, 0.00001174, -0.00001158, 0.00000536, -0.00000665, 0.00001733, -0.00001288, -0.00000413, -0.00000489, 0.00000155, -0.00000158, 0.00000684, -0.00000637, 0.00000687, -0.00000484, 0.00000412, -0.00000736, -0.00000324, -0.00000112, -0.00000261, 0.00000193, 0.00000958, -0.00000559, 0.00000201, -0.00000656, 0.00000475, -0.00000210, 0.00000415, -0.00000136, -0.00000110, -0.00000088, -0.00000330, -0.00000087, 0.00000015, 0.00000116, 0.00000287, -0.00000177, 0.00000656, -0.00000562, 0.00000065, -0.00000251, -0.00000032, 0.00000071, -0.00000273, 0.00000092, -0.00000067, 0.00000115, 0.00000206, -0.00000096, 0.00000315, -0.00000231, 0.00000077, -0.00000106, -0.00000007, 0.00000053, },
		{ -0.00000362, 0.00000055, -0.00000242, 0.00000074, 0.00000052, 0.00000065, 0.00000364, -0.00000141, 0.00000060, -0.00000111, -0.00000012, 0.00000098, -0.00000279, 0.00000160, -0.00000513, 0.00000272, -0.00000079, 0.00000143, 0.00000345, -0.00000136, 0.00000029, -0.00000196, -0.00000129, 0.00000029, -0.00000403, 0.00000459, -0.00000151, 0.00000182, -0.00000167, -0.00000073, -0.00000345, 0.00000436, 0.00000240, 0.00000133, -0.00000151, -0.00000137, -0.00000742, 0.00000381, -0.00000335, 0.00000357, -0.00000424, 0.00000257, 0.00000346, -0.00000259, 0.00000303, -0.00000303, -0.00001658, 0.00001124, -0.00001005, 0.00001147, -0.00000789, 0.00000577, -0.00000870, 0.00000189, 0.00000178, -0.00000914, -0.00000362, -0.00000054, -0.00000645, -0.00001794, -0.00005910, -0.00001872, -0.00023910, 0.00031274, 0.00041810, -0.00022200, -0.00005080, 0.00000577, -0.00001212, -0.00000718, -0.00001572, -0.00001581, 0.00000738, -0.00002277, 0.00000771, -0.00000910, 0.00000327, -0.00000407, 0.00001564, -0.00001292, -0.00000446, -0.00000295, -0.00000051, 0.00000012, 0.00000534, -0.00000506, 0.00000580, -0.00000431, 0.00000346, -0.00000646, -0.00000414, -0.00000028, -0.00000344, 0.00000340, 0.00000832, -0.00000517, 0.00000140, -0.00000561, 0.00000402, -0.00000164, 0.00000394, -0.00000129, -0.00000127, -0.00000079, -0.00000358, -0.00000013, -0.00000058, 0.00000190, 0.00000210, -0.00000105, 0.00000600, -0.00000562, 0.00000076, -0.00000245, -0.00000030, 0.00000068, -0.00000277, 0.00000112, -0.00000093, 0.00000151, 0.00000166, -0.00000062, 0.00000284, -0.00000224, 0.00000084, -0.00000117, 0.00000017, 0.00000030, },
		{ -0.00000348, 0.00000049, -0.00000247, 0.00000093, 0.00000035, 0.00000082, 0.00000351, -0.00000148, 0.00000077, -0.00000150, 0.00000040, 0.00000054, -0.00000236, 0.00000117, -0.00000488, 0.00000288, -0.00000094, 0.00000139, 0.00000355, -0.00000163, 0.00000055, -0.00000246, -0.00000066, -0.00000019, -0.00000337, 0.00000386, -0.00000091, 0.00000137, -0.00000134, -0.00000133, -0.00000286, 0.00000436, 0.00000259, 0.00000030, -0.00000044, -0.00000239, -0.00000643, 0.00000308, -0.00000264, 0.00000322, -0.00000384, 0.00000158, 0.00000460, -0.00000374, 0.00000453, -0.00000557, -0.00001421, 0.00001030, -0.00000897, 0.00000939, -0.00000583, 0.00000445, -0.00000789, 0.00000095, 0.00000294, -0.00001030, -0.00000079, -0.00000650, 0.00000101, -0.00002670, -0.00004873, -0.00003511, -0.00021844, 0.00031386, 0.00039119, -0.00020132, -0.00005867, 0.00001217, -0.00001656, -0.00000270, -0.00002087, -0.00001212, 0.00000329, -0.00001793, 0.00000372, -0.00000661, 0.00000120, -0.00000150, 0.00001390, -0.00001295, -0.00000467, -0.00000119, -0.00000244, 0.00000175, 0.00000389, -0.00000381, 0.00000477, -0.00000379, 0.00000285, -0.00000563, -0.00000497, 0.00000049, -0.00000419, 0.00000478, 0.00000706, -0.00000475, 0.00000081, -0.00000467, 0.00000327, -0.00000116, 0.00000373, -0.00000123, -0.00000139, -0.00000076, -0.00000378, 0.00000053, -0.00000125, 0.00000261, 0.00000134, -0.00000031, 0.00000539, -0.00000556, 0.00000084, -0.00000239, -0.00000028, 0.00000063, -0.00000278, 0.00000127, -0.00000114, 0.00000183, 0.00000127, -0.00000028, 0.00000251, -0.00000213, 0.00000089, -0.00000126, 0.00000039, 0.00000007, },
		{ -0.00000332, 0.00000040, -0.00000248, 0.00000110, 0.00000018, 0.00000099, 0.00000336, -0.00000151, 0.00000089, -0.00000184, 0.00000089, 0.00000011, -0.00000193, 0.00000072, -0.00000458, 0.00000300, -0.00000109, 0.00000137, 0.00000361, -0.00000185, 0.00000076, -0.00000289, -0.00000008, -0.00000064, -0.00000272, 0.00000315, -0.00000030, 0.00000091, -0.00000102, -0.00000191, -0.00000229, 0.00000437, 0.00000271, -0.00000063, 0.00000057, -0.00000337, -0.00000545, 0.00000239, -0.00000195, 0.00000289, -0.00000350, 0.00000068, 0.00000563, -0.00000475, 0.00000587, -0.00000797, -0.00001184, 0.00000933, -0.00000790, 0.00000737, -0.00000379, 0.00000311, -0.00000711, 0.00000020, 0.00000381, -0.00001110, 0.00000154, -0.00001187, 0.00000792, -0.00003501, -0.00003872, -0.00005134, -0.00019684, 0.00031255, 0.00036655, -0.00018194, -0.00006559, 0.00001784, -0.00002036, 0.00000102, -0.00002541, -0.00000884, -0.00000048, -0.00001332, -0.00000016, -0.00000417, -0.00000081, 0.00000102, 0.00001213, -0.00001296, -0.00000475, 0.00000038, -0.00000422, 0.00000327, 0.00000250, -0.00000262, 0.00000378, -0.00000330, 0.00000228, -0.00000485, -0.00000572, 0.00000118, -0.00000484, 0.00000608, 0.00000583, -0.00000433, 0.00000024, -0.00000376, 0.00000253, -0.00000066, 0.00000350, -0.00000119, -0.00000146, -0.00000078, -0.00000391, 0.00000111, -0.00000185, 0.00000326, 0.00000061, 0.00000042, 0.00000475, -0.00000546, 0.00000090, -0.00000232, -0.00000026, 0.00000056, -0.00000275, 0.00000137, -0.00000131, 0.00000212, 0.00000091, 0.00000006, 0.00000217, -0.00000200, 0.00000090, -0.00000133, 0.00000061, -0.00000016, },
		{ -0.00000315, 0.00000028, -0.00000245, 0.00000124, 0.00000002, 0.00000118, 0.00000317, -0.00000151, 0.00000097, -0.00000214, 0.00000135, -0.00000031, -0.00000150, 0.00000025, -0.00000425, 0.00000308, -0.00000123, 0.00000137, 0.00000363, -0.00000203, 0.00000092, -0.00000324, 0.00000044, -0.00000103, -0.00000212, 0.00000246, 0.00000029, 0.00000046, -0.00000071, -0.00000247, -0.00000172, 0.00000440, 0.00000276, -0.00000147, 0.00000150, -0.00000428, -0.00000451, 0.00000174, -0.00000131, 0.00000258, -0.00000323, -0.00000011, 0.00000654, -0.00000563, 0.00000703, -0.00001022, -0.00000953, 0.00000834, -0.00000687, 0.00000543, -0.00000179, 0.00000177, -0.00000638, -0.00000038, 0.00000440, -0.00001153, 0.00000335, -0.00001660, 0.00001418, -0.00004274, -0.00002922, -0.00006717, -0.00017457, 0.00030901, 0.00034413, -0.00016386, -0.00007155, 0.00002277, -0.00002351, 0.00000397, -0.00002931, -0.00000601, -0.00000388, -0.00000899, -0.00000388, -0.00000180, -0.00000273, 0.00000348, 0.00001034, -0.00001296, -0.00000471, 0.00000175, -0.00000583, 0.00000469, 0.00000120, -0.00000152, 0.00000286, -0.00000285, 0.00000177, -0.00000415, -0.00000638, 0.00000179, -0.00000538, 0.00000727, 0.00000464, -0.00000393, -0.00000029, -0.00000287, 0.00000180, -0.00000015, 0.00000325, -0.00000117, -0.00000150, -0.00000085, -0.00000395, 0.00000160, -0.00000238, 0.00000384, -0.00000009, 0.00000113, 0.00000409, -0.00000531, 0.00000093, -0.00000225, -0.00000023, 0.00000048, -0.00000268, 0.00000143, -0.00000143, 0.00000236, 0.00000057, 0.00000038, 0.00000182, -0.00000185, 0.00000089, -0.00000137, 0.00000081, -0.00000040, },
		{ -0.00000296, 0.00000015, -0.00000240, 0.00000135, -0.00000012, 0.00000136, 0.00000296, -0.00000146, 0.00000101, -0.00000238, 0.00000177, -0.00000070, -0.00000108, -0.00000023, -0.00000389, 0.00000313, -0.00000136, 0.00000139, 0.00000362, -0.00000216, 0.00000102, -0.00000352, 0.00000089, -0.00000138, -0.00000156, 0.00000179, 0.00000086, 0.00000003, -0.00000043, -0.00000298, -0.00000116, 0.00000444, 0.00000275, -0.00000220, 0.00000233, -0.00000514, -0.00000361, 0.00000113, -0.00000072, 0.00000232, -0.00000302, -0.00000079, 0.00000730, -0.00000636, 0.00000802, -0.00001229, -0.00000729, 0.00000737, -0.00000590, 0.00000358, 0.00000014, 0.00000044, -0.00000569, -0.00000079, 0.00000471, -0.00001161, 0.00000464, -0.00002066, 0.00001971, -0.00004977, -0.00002038, -0.00008238, -0.00015191, 0.00030347, 0.00032385, -0.00014708, -0.00007655, 0.00002698, -0.00002603, 0.00000613, -0.00003252, -0.00000367, -0.00000687, -0.00000499, -0.00000739, 0.00000045, -0.00000454, 0.00000583, 0.00000856, -0.00001295, -0.00000456, 0.00000292, -0.00000727, 0.00000597, -0.00000002, -0.00000051, 0.00000201, -0.00000245, 0.00000132, -0.00000353, -0.00000695, 0.00000230, -0.00000581, 0.00000835, 0.00000350, -0.00000354, -0.00000078, -0.00000203, 0.00000109, 0.00000036, 0.00000300, -0.00000117, -0.00000150, -0.00000097, -0.00000393, 0.00000201, -0.00000282, 0.00000436, -0.00000073, 0.00000182, 0.00000342, -0.00000512, 0.00000094, -0.00000216, -0.00000019, 0.00000038, -0.00000259, 0.00000144, -0.00000150, 0.00000256, 0.00000026, 0.00000069, 0.00000146, -0.00000167, 0.00000086, -0.00000139, 0.00000099, -0.00000064, },
		{ -0.00000277, -0.00000001, -0.00000232, 0.00000143, -0.00000025, 0.00000154, 0.00000274, -0.00000138, 0.00000100, -0.00000257, 0.00000214, -0.00000106, -0.00000067, -0.00000070, -0.00000351, 0.00000314, -0.00000147, 0.00000141, 0.00000357, -0.00000225, 0.00000107, -0.00000372, 0.00000127, -0.00000166, -0.00000105, 0.00000117, 0.00000140, -0.00000039, -0.00000017, -0.00000346, -0.00000063, 0.00000448, 0.00000266, -0.00000283, 0.00000307, -0.00000591, -0.00000276, 0.00000058, -0.00000019, 0.00000209, -0.00000289, -0.00000136, 0.00000793, -0.00000692, 0.00000881, -0.00001415, -0.00000515, 0.00000642, -0.00000500, 0.00000184, 0.00000199, -0.00000086, -0.00000503, -0.00000105, 0.00000476, -0.00001132, 0.00000539, -0.00002401, 0.00002447, -0.00005603, -0.00001232, -0.00009679, -0.00012913, 0.00029616, 0.00030562, -0.00013157, -0.00008059, 0.00003048, -0.00002796, 0.00000752, -0.00003504, -0.00000185, -0.00000939, -0.00000138, -0.00001064, 0.00000254, -0.00000621, 0.00000807, 0.00000679, -0.00001292, -0.00000429, 0.00000389, -0.00000852, 0.00000712, -0.00000112, 0.00000040, 0.00000125, -0.00000210, 0.00000093, -0.00000299, -0.00000742, 0.00000271, -0.00000613, 0.00000930, 0.00000242, -0.00000319, -0.00000122, -0.00000124, 0.00000041, 0.00000087, 0.00000274, -0.00000119, -0.00000146, -0.00000113, -0.00000382, 0.00000232, -0.00000318, 0.00000479, -0.00000131, 0.00000247, 0.00000274, -0.00000489, 0.00000093, -0.00000207, -0.00000016, 0.00000027, -0.00000247, 0.00000141, -0.00000152, 0.00000271, -0.00000002, 0.00000098, 0.00000112, -0.00000148, 0.00000081, -0.00000138, 0.00000115, -0.00000087, },
		{ -0.00000257, -0.00000019, -0.00000221, 0.00000149, -0.00000036, 0.00000172, 0.00000250, -0.00000127, 0.00000094, -0.00000270, 0.00000246, -0.00000138, -0.00000029, -0.00000117, -0.00000310, 0.00000311, -0.00000157, 0.00000144, 0.00000350, -0.00000231, 0.00000108, -0.00000385, 0.00000158, -0.00000189, -0.00000060, 0.00000060, 0.00000190, -0.00000079, 0.00000006, -0.00000389, -0.00000011, 0.00000454, 0.00000252, -0.00000335, 0.00000370, -0.00000660, -0.00000196, 0.00000009, 0.00000028, 0.00000191, -0.00000283, -0.00000181, 0.00000842, -0.00000733, 0.00000940, -0.00001579, -0.00000314, 0.00000551, -0.00000418, 0.00000022, 0.00000373, -0.00000214, -0.00000440, -0.00000117, 0.00000456, -0.00001071, 0.00000564, -0.00002665, 0.00002839, -0.00006143, -0.00000518, -0.00011019, -0.00010650, 0.00028731, 0.00028932, -0.00011730, -0.00008372, 0.00003332, -0.00002933, 0.00000816, -0.00003686, -0.00000056, -0.00001144, 0.00000182, -0.00001359, 0.00000445, -0.00000770, 0.00001016, 0.00000506, -0.00001289, -0.00000393, 0.00000465, -0.00000958, 0.00000813, -0.00000212, 0.00000119, 0.00000058, -0.00000182, 0.00000061, -0.00000254, -0.00000780, 0.00000303, -0.00000633, 0.00001012, 0.00000141, -0.00000287, -0.00000161, -0.00000051, -0.00000023, 0.00000138, 0.00000247, -0.00000121, -0.00000139, -0.00000133, -0.00000366, 0.00000255, -0.00000345, 0.00000515, -0.00000183, 0.00000308, 0.00000207, -0.00000464, 0.00000090, -0.00000197, -0.00000014, 0.00000016, -0.00000233, 0.00000134, -0.00000149, 0.00000282, -0.00000025, 0.00000123, 0.00000078, -0.00000128, 0.00000073, -0.00000134, 0.00000128, -0.00000109, },
		{ -0.00000237, -0.00000038, -0.00000208, 0.00000152, -0.00000045, 0.00000188, 0.00000224, -0.00000113, 0.00000085, -0.00000278, 0.00000273, -0.00000166, 0.00000005, -0.00000162, -0.00000268, 0.00000305, -0.00000164, 0.00000147, 0.00000340, -0.00000232, 0.00000104, -0.00000391, 0.00000182, -0.00000205, -0.00000022, 0.00000008, 0.00000235, -0.00000116, 0.00000026, -0.00000427, 0.00000038, 0.00000461, 0.00000232, -0.00000376, 0.00000423, -0.00000721, -0.00000122, -0.00000034, 0.00000069, 0.00000176, -0.00000283, -0.00000215, 0.00000877, -0.00000758, 0.00000978, -0.00001720, -0.00000126, 0.00000467, -0.00000344, -0.00000125, 0.00000537, -0.00000337, -0.00000380, -0.00000117, 0.00000413, -0.00000979, 0.00000538, -0.00002857, 0.00003146, -0.00006593, 0.00000097, -0.00012243, -0.00008428, 0.00027717, 0.00027481, -0.00010421, -0.00008597, 0.00003554, -0.00003020, 0.00000811, -0.00003800, 0.00000019, -0.00001300, 0.00000457, -0.00001622, 0.00000615, -0.00000902, 0.00001208, 0.00000337, -0.00001284, -0.00000347, 0.00000523, -0.00001044, 0.00000898, -0.00000301, 0.00000187, 0.00000000, -0.00000159, 0.00000036, -0.00000217, -0.00000807, 0.00000326, -0.00000641, 0.00001080, 0.00000048, -0.00000259, -0.00000194, 0.00000016, -0.00000083, 0.00000187, 0.00000220, -0.00000124, -0.00000130, -0.00000156, -0.00000343, 0.00000269, -0.00000364, 0.00000542, -0.00000228, 0.00000363, 0.00000141, -0.00000437, 0.00000085, -0.00000186, -0.00000011, 0.00000004, -0.00000218, 0.00000123, -0.00000142, 0.00000287, -0.00000045, 0.00000146, 0.00000045, -0.00000106, 0.00000064, -0.00000129, 0.00000139, -0.00000130, },
		{ -0.00000217, -0.00000057, -0.00000193, 0.00000152, -0.00000051, 0.00000203, 0.00000199, -0.00000098, 0.00000073, -0.00000280, 0.00000295, -0.00000190, 0.00000037, -0.00000205, -0.00000225, 0.00000296, -0.00000170, 0.00000150, 0.00000328, -0.00000231, 0.00000095, -0.00000390, 0.00000198, -0.00000214, 0.00000010, -0.00000038, 0.00000276, -0.00000149, 0.00000042, -0.00000461, 0.00000085, 0.00000468, 0.00000207, -0.00000408, 0.00000466, -0.00000772, -0.00000055, -0.00000071, 0.00000103, 0.00000166, -0.00000290, -0.00000238, 0.00000898, -0.00000768, 0.00000996, -0.00001837, 0.00000046, 0.00000389, -0.00000281, -0.00000258, 0.00000687, -0.00000456, -0.00000321, -0.00000107, 0.00000351, -0.00000860, 0.00000468, -0.00002979, 0.00003367, -0.00006948, 0.00000606, -0.00013335, -0.00006271, 0.00026599, 0.00026195, -0.00009223, -0.00008738, 0.00003719, -0.00003064, 0.00000740, -0.00003848, 0.00000043, -0.00001407, 0.00000685, -0.00001849, 0.00000762, -0.00001012, 0.00001383, 0.00000175, -0.00001277, -0.00000294, 0.00000562, -0.00001112, 0.00000968, -0.00000378, 0.00000243, -0.00000047, -0.00000143, 0.00000017, -0.00000189, -0.00000825, 0.00000339, -0.00000638, 0.00001135, -0.00000036, -0.00000236, -0.00000221, 0.00000075, -0.00000137, 0.00000235, 0.00000191, -0.00000128, -0.00000118, -0.00000181, -0.00000314, 0.00000274, -0.00000373, 0.00000560, -0.00000265, 0.00000412, 0.00000078, -0.00000407, 0.00000078, -0.00000174, -0.00000010, -0.00000007, -0.00000202, 0.00000110, -0.00000131, 0.00000288, -0.00000061, 0.00000165, 0.00000014, -0.00000085, 0.00000054, -0.00000121, 0.00000148, -0.00000149, },
		{ -0.00000199, -0.00000077, -0.00000176, 0.00000150, -0.00000055, 0.00000216, 0.00000173, -0.00000080, 0.00000057, -0.00000277, 0.00000311, -0.00000209, 0.00000064, -0.00000245, -0.00000182, 0.00000284, -0.00000174, 0.00000153, 0.00000314, -0.00000228, 0.00000084, -0.00000383, 0.00000208, -0.00000216, 0.00000036, -0.00000077, 0.00000311, -0.00000179, 0.00000054, -0.00000490, 0.00000129, 0.00000475, 0.00000177, -0.00000430, 0.00000497, -0.00000814, 0.00000007, -0.00000101, 0.00000130, 0.00000159, -0.00000302, -0.00000251, 0.00000906, -0.00000763, 0.00000995, -0.00001930, 0.00000201, 0.00000319, -0.00000227, -0.00000375, 0.00000823, -0.00000570, -0.00000264, -0.00000088, 0.00000271, -0.00000718, 0.00000355, -0.00003034, 0.00003503, -0.00007207, 0.00001003, -0.00014284, -0.00004201, 0.00025400, 0.00025061, -0.00008128, -0.00008804, 0.00003833, -0.00003073, 0.00000612, -0.00003835, 0.00000016, -0.00001465, 0.00000867, -0.00002039, 0.00000883, -0.00001102, 0.00001538, 0.00000019, -0.00001268, -0.00000235, 0.00000584, -0.00001161, 0.00001024, -0.00000445, 0.00000288, -0.00000085, -0.00000134, 0.00000005, -0.00000168, -0.00000833, 0.00000343, -0.00000623, 0.00001176, -0.00000112, -0.00000217, -0.00000241, 0.00000128, -0.00000187, 0.00000281, 0.00000163, -0.00000133, -0.00000105, -0.00000209, -0.00000281, 0.00000272, -0.00000375, 0.00000570, -0.00000295, 0.00000455, 0.00000017, -0.00000377, 0.00000070, -0.00000161, -0.00000009, -0.00000018, -0.00000185, 0.00000094, -0.00000116, 0.00000284, -0.00000073, 0.00000181, -0.00000014, -0.00000063, 0.00000042, -0.00000111, 0.00000153, -0.00000165, },
		{ -0.00000181, -0.00000097, -0.00000158, 0.00000145, -0.00000057, 0.00000227, 0.00000148, -0.00000061, 0.00000038, -0.00000270, 0.00000322, -0.00000224, 0.00000087, -0.00000282, -0.00000140, 0.00000270, -0.00000176, 0.00000155, 0.00000299, -0.00000222, 0.00000070, -0.00000370, 0.00000211, -0.00000213, 0.00000054, -0.00000111, 0.00000340, -0.00000204, 0.00000063, -0.00000513, 0.00000171, 0.00000481, 0.00000144, -0.00000442, 0.00000519, -0.00000847, 0.00000062, -0.00000124, 0.00000151, 0.00000156, -0.00000320, -0.00000254, 0.00000902, -0.00000743, 0.00000974, -0.00001999, 0.00000339, 0.00000258, -0.00000184, -0.00000475, 0.00000943, -0.00000677, -0.00000206, -0.00000063, 0.00000179, -0.00000557, 0.00000206, -0.00003027, 0.00003556, -0.00007370, 0.00001285, -0.00015080, -0.00002238, 0.00024143, 0.00024061, -0.00007127, -0.00008799, 0.00003904, -0.00003055, 0.00000434, -0.00003767, -0.00000056, -0.00001476, 0.00001001, -0.00002190, 0.00000978, -0.00001168, 0.00001673, -0.00000128, -0.00001256, -0.00000171, 0.00000590, -0.00001192, 0.00001065, -0.00000500, 0.00000321, -0.00000112, -0.00000130, -0.00000001, -0.00000156, -0.00000833, 0.00000338, -0.00000599, 0.00001203, -0.00000179, -0.00000203, -0.00000256, 0.00000173, -0.00000230, 0.00000324, 0.00000133, -0.00000137, -0.00000091, -0.00000237, -0.00000244, 0.00000264, -0.00000369, 0.00000571, -0.00000316, 0.00000490, -0.00000040, -0.00000346, 0.00000061, -0.00000147, -0.00000009, -0.00000027, -0.00000168, 0.00000077, -0.00000097, 0.00000277, -0.00000080, 0.00000193, -0.00000039, -0.00000042, 0.00000030, -0.00000099, 0.00000156, -0.00000180, },
		{ -0.00000165, -0.00000117, -0.00000139, 0.00000139, -0.00000056, 0.00000236, 0.00000124, -0.00000042, 0.00000018, -0.00000258, 0.00000328, -0.00000233, 0.00000105, -0.00000315, -0.00000099, 0.00000254, -0.00000175, 0.00000157, 0.00000283, -0.00000215, 0.00000054, -0.00000352, 0.00000208, -0.00000203, 0.00000066, -0.00000137, 0.00000362, -0.00000226, 0.00000068, -0.00000531, 0.00000210, 0.00000487, 0.00000107, -0.00000447, 0.00000531, -0.00000871, 0.00000110, -0.00000141, 0.00000165, 0.00000157, -0.00000342, -0.00000249, 0.00000887, -0.00000711, 0.00000935, -0.00002043, 0.00000458, 0.00000206, -0.00000152, -0.00000560, 0.00001048, -0.00000778, -0.00000149, -0.00000035, 0.00000075, -0.00000381, 0.00000027, -0.00002963, 0.00003530, -0.00007439, 0.00001453, -0.00015716, -0.00000401, 0.00022850, 0.00023182, -0.00006212, -0.00008731, 0.00003938, -0.00003017, 0.00000215, -0.00003648, -0.00000170, -0.00001444, 0.00001090, -0.00002302, 0.00001047, -0.00001211, 0.00001785, -0.00000266, -0.00001242, -0.00000105, 0.00000582, -0.00001206, 0.00001092, -0.00000545, 0.00000344, -0.00000131, -0.00000132, -0.00000002, -0.00000150, -0.00000823, 0.00000326, -0.00000564, 0.00001217, -0.00000236, -0.00000194, -0.00000264, 0.00000211, -0.00000266, 0.00000364, 0.00000103, -0.00000141, -0.00000076, -0.00000266, -0.00000204, 0.00000249, -0.00000355, 0.00000564, -0.00000330, 0.00000517, -0.00000092, -0.00000314, 0.00000051, -0.00000133, -0.00000010, -0.00000036, -0.00000152, 0.00000058, -0.00000077, 0.00000265, -0.00000084, 0.00000201, -0.00000062, -0.00000022, 0.00000018, -0.00000085, 0.00000157, -0.00000193, },
		{ -0.00000150, -0.00000136, -0.00000119, 0.00000130, -0.00000053, 0.00000242, 0.00000101, -0.00000023, -0.00000004, -0.00000242, 0.00000329, -0.00000237, 0.00000119, -0.00000343, -0.00000059, 0.00000236, -0.00000173, 0.00000157, 0.00000267, -0.00000207, 0.00000037, -0.00000331, 0.00000200, -0.00000188, 0.00000072, -0.00000157, 0.00000379, -0.00000243, 0.00000069, -0.00000544, 0.00000246, 0.00000493, 0.00000069, -0.00000445, 0.00000533, -0.00000886, 0.00000153, -0.00000153, 0.00000174, 0.00000160, -0.00000367, -0.00000237, 0.00000861, -0.00000667, 0.00000880, -0.00002064, 0.00000559, 0.00000163, -0.00000129, -0.00000628, 0.00001136, -0.00000872, -0.00000090, -0.00000006, -0.00000034, -0.00000196, -0.00000177, -0.00002849, 0.00003430, -0.00007419, 0.00001507, -0.00016187, 0.00001296, 0.00021542, 0.00022409, -0.00005374, -0.00008606, 0.00003941, -0.00002970, -0.00000037, -0.00003487, -0.00000320, -0.00001373, 0.00001135, -0.00002376, 0.00001088, -0.00001231, 0.00001876, -0.00000395, -0.00001225, -0.00000037, 0.00000562, -0.00001205, 0.00001106, -0.00000581, 0.00000358, -0.00000140, -0.00000140, 0.00000003, -0.00000151, -0.00000806, 0.00000307, -0.00000521, 0.00001219, -0.00000285, -0.00000189, -0.00000267, 0.00000242, -0.00000296, 0.00000401, 0.00000073, -0.00000144, -0.00000061, -0.00000295, -0.00000162, 0.00000229, -0.00000335, 0.00000550, -0.00000336, 0.00000537, -0.00000140, -0.00000283, 0.00000040, -0.00000118, -0.00000012, -0.00000043, -0.00000137, 0.00000039, -0.00000054, 0.00000251, -0.00000084, 0.00000204, -0.00000082, -0.00000003, 0.00000006, -0.00000070, 0.00000155, -0.00000202, },
		{ -0.00000138, -0.00000153, -0.00000099, 0.00000120, -0.00000048, 0.00000246, 0.00000079, -0.00000004, -0.00000027, -0.00000223, 0.00000326, -0.00000237, 0.00000127, -0.00000368, -0.00000021, 0.00000217, -0.00000169, 0.00000156, 0.00000251, -0.00000198, 0.00000019, -0.00000306, 0.00000187, -0.00000168, 0.00000071, -0.00000171, 0.00000389, -0.00000255, 0.00000065, -0.00000552, 0.00000279, 0.00000497, 0.00000030, -0.00000436, 0.00000527, -0.00000893, 0.00000191, -0.00000159, 0.00000176, 0.00000165, -0.00000395, -0.00000218, 0.00000827, -0.00000614, 0.00000811, -0.00002063, 0.00000642, 0.00000129, -0.00000116, -0.00000681, 0.00001208, -0.00000957, -0.00000030, 0.00000023, -0.00000147, -0.00000007, -0.00000399, -0.00002690, 0.00003263, -0.00007314, 0.00001452, -0.00016493, 0.00002839, 0.00020236, 0.00021727, -0.00004603, -0.00008432, 0.00003920, -0.00002921, -0.00000312, -0.00003292, -0.00000501, -0.00001267, 0.00001139, -0.00002413, 0.00001104, -0.00001227, 0.00001943, -0.00000513, -0.00001205, 0.00000030, 0.00000532, -0.00001189, 0.00001108, -0.00000609, 0.00000362, -0.00000140, -0.00000152, 0.00000011, -0.00000158, -0.00000783, 0.00000282, -0.00000470, 0.00001208, -0.00000325, -0.00000188, -0.00000265, 0.00000265, -0.00000319, 0.00000435, 0.00000042, -0.00000146, -0.00000047, -0.00000322, -0.00000119, 0.00000205, -0.00000309, 0.00000529, -0.00000334, 0.00000550, -0.00000183, -0.00000253, 0.00000029, -0.00000102, -0.00000015, -0.00000049, -0.00000123, 0.00000019, -0.00000029, 0.00000233, -0.00000081, 0.00000205, -0.00000098, 0.00000015, -0.00000006, -0.00000055, 0.00000150, -0.00000210, },
		{ -0.00000128, -0.00000170, -0.00000079, 0.00000109, -0.00000041, 0.00000248, 0.00000060, 0.00000015, -0.00000050, -0.00000202, 0.00000318, -0.00000232, 0.00000131, -0.00000387, 0.00000014, 0.00000198, -0.00000163, 0.00000154, 0.00000235, -0.00000190, 0.00000002, -0.00000278, 0.00000170, -0.00000144, 0.00000066, -0.00000179, 0.00000393, -0.00000263, 0.00000059, -0.00000554, 0.00000309, 0.00000499, -0.00000009, -0.00000421, 0.00000513, -0.00000892, 0.00000223, -0.00000161, 0.00000174, 0.00000172, -0.00000424, -0.00000194, 0.00000785, -0.00000552, 0.00000728, -0.00002041, 0.00000708, 0.00000103, -0.00000113, -0.00000718, 0.00001262, -0.00001035, 0.00000033, 0.00000048, -0.00000259, 0.00000181, -0.00000631, -0.00002496, 0.00003036, -0.00007131, 0.00001294, -0.00016633, 0.00004219, 0.00018951, 0.00021123, -0.00003891, -0.00008214, 0.00003880, -0.00002879, -0.00000600, -0.00003069, -0.00000705, -0.00001132, 0.00001106, -0.00002415, 0.00001093, -0.00001201, 0.00001988, -0.00000619, -0.00001181, 0.00000096, 0.00000493, -0.00001161, 0.00001099, -0.00000629, 0.00000359, -0.00000134, -0.00000167, 0.00000023, -0.00000169, -0.00000753, 0.00000252, -0.00000412, 0.00001186, -0.00000357, -0.00000191, -0.00000258, 0.00000281, -0.00000335, 0.00000465, 0.00000011, -0.00000147, -0.00000033, -0.00000347, -0.00000076, 0.00000178, -0.00000279, 0.00000501, -0.00000326, 0.00000554, -0.00000219, -0.00000224, 0.00000018, -0.00000087, -0.00000019, -0.00000053, -0.00000111, 0.00000001, -0.00000004, 0.00000213, -0.00000075, 0.00000201, -0.00000110, 0.00000031, -0.00000018, -0.00000038, 0.00000144, -0.00000215, },
		{ -0.00000119, -0.00000184, -0.00000060, 0.00000097, -0.00000032, 0.00000246, 0.00000042, 0.00000032, -0.00000073, -0.00000178, 0.00000307, -0.00000223, 0.00000129, -0.00000402, 0.00000046, 0.00000178, -0.00000156, 0.00000150, 0.00000219, -0.00000182, -0.00000015, -0.00000250, 0.00000151, -0.00000116, 0.00000056, -0.00000181, 0.00000391, -0.00000266, 0.00000048, -0.00000551, 0.00000336, 0.00000499, -0.00000048, -0.00000403, 0.00000492, -0.00000883, 0.00000250, -0.00000158, 0.00000167, 0.00000179, -0.00000454, -0.00000166, 0.00000737, -0.00000483, 0.00000634, -0.00002000, 0.00000756, 0.00000086, -0.00000117, -0.00000741, 0.00001300, -0.00001104, 0.00000098, 0.00000069, -0.00000368, 0.00000364, -0.00000868, -0.00002274, 0.00002758, -0.00006880, 0.00001039, -0.00016612, 0.00005429, 0.00017701, 0.00020585, -0.00003230, -0.00007959, 0.00003828, -0.00002851, -0.00000893, -0.00002828, -0.00000925, -0.00000973, 0.00001040, -0.00002384, 0.00001059, -0.00001154, 0.00002010, -0.00000714, -0.00001153, 0.00000158, 0.00000448, -0.00001122, 0.00001080, -0.00000643, 0.00000349, -0.00000121, -0.00000186, 0.00000037, -0.00000185, -0.00000718, 0.00000218, -0.00000349, 0.00001154, -0.00000380, -0.00000197, -0.00000246, 0.00000291, -0.00000345, 0.00000491, -0.00000020, -0.00000147, -0.00000021, -0.00000370, -0.00000035, 0.00000149, -0.00000245, 0.00000468, -0.00000311, 0.00000551, -0.00000250, -0.00000197, 0.00000008, -0.00000071, -0.00000024, -0.00000055, -0.00000101, -0.00000017, 0.00000021, 0.00000192, -0.00000067, 0.00000194, -0.00000120, 0.00000044, -0.00000028, -0.00000021, 0.00000136, -0.00000217, },
		{ -0.00000114, -0.00000197, -0.00000041, 0.00000084, -0.00000022, 0.00000243, 0.00000027, 0.00000047, -0.00000095, -0.00000153, 0.00000293, -0.00000211, 0.00000124, -0.00000411, 0.00000075, 0.00000157, -0.00000147, 0.00000146, 0.00000204, -0.00000174, -0.00000030, -0.00000220, 0.00000129, -0.00000086, 0.00000042, -0.00000178, 0.00000383, -0.00000266, 0.00000034, -0.00000543, 0.00000361, 0.00000497, -0.00000085, -0.00000381, 0.00000466, -0.00000869, 0.00000274, -0.00000153, 0.00000156, 0.00000187, -0.00000483, -0.00000136, 0.00000685, -0.00000409, 0.00000531, -0.00001941, 0.00000789, 0.00000076, -0.00000127, -0.00000751, 0.00001321, -0.00001165, 0.00000166, 0.00000083, -0.00000471, 0.00000536, -0.00001101, -0.00002033, 0.00002438, -0.00006568, 0.00000699, -0.00016435, 0.00006464, 0.00016498, 0.00020101, -0.00002612, -0.00007673, 0.00003767, -0.00002844, -0.00001181, -0.00002576, -0.00001154, -0.00000797, 0.00000945, -0.00002323, 0.00001002, -0.00001086, 0.00002010, -0.00000798, -0.00001121, 0.00000215, 0.00000398, -0.00001074, 0.00001053, -0.00000653, 0.00000335, -0.00000103, -0.00000207, 0.00000053, -0.00000203, -0.00000678, 0.00000181, -0.00000282, 0.00001113, -0.00000397, -0.00000206, -0.00000232, 0.00000295, -0.00000348, 0.00000512, -0.00000051, -0.00000146, -0.00000010, -0.00000390, 0.00000005, 0.00000119, -0.00000208, 0.00000431, -0.00000290, 0.00000541, -0.00000276, -0.00000171, -0.00000003, -0.00000056, -0.00000029, -0.00000056, -0.00000093, -0.00000033, 0.00000046, 0.00000170, -0.00000056, 0.00000185, -0.00000125, 0.00000056, -0.00000037, -0.00000005, 0.00000126, -0.00000218, },
		{ -0.00000110, -0.00000208, -0.00000024, 0.00000072, -0.00000011, 0.00000237, 0.00000014, 0.00000060, -0.00000116, -0.00000126, 0.00000276, -0.00000195, 0.00000114, -0.00000416, 0.00000100, 0.00000137, -0.00000138, 0.00000139, 0.00000190, -0.00000168, -0.00000043, -0.00000191, 0.00000106, -0.00000055, 0.00000024, -0.00000171, 0.00000371, -0.00000261, 0.00000017, -0.00000530, 0.00000382, 0.00000493, -0.00000120, -0.00000357, 0.00000434, -0.00000848, 0.00000294, -0.00000144, 0.00000142, 0.00000194, -0.00000511, -0.00000104, 0.00000630, -0.00000332, 0.00000421, -0.00001867, 0.00000806, 0.00000072, -0.00000144, -0.00000748, 0.00001325, -0.00001216, 0.00000237, 0.00000090, -0.00000564, 0.00000694, -0.00001325, -0.00001781, 0.00002086, -0.00006206, 0.00000282, -0.00016111, 0.00007323, 0.00015353, 0.00019661, -0.00002031, -0.00007359, 0.00003701, -0.00002863, -0.00001456, -0.00002322, -0.00001386, -0.00000610, 0.00000827, -0.00002235, 0.00000925, -0.00001000, 0.00001989, -0.00000869, -0.00001085, 0.00000266, 0.00000346, -0.00001018, 0.00001019, -0.00000658, 0.00000316, -0.00000080, -0.00000229, 0.00000069, -0.00000222, -0.00000636, 0.00000143, -0.00000212, 0.00001064, -0.00000407, -0.00000217, -0.00000214, 0.00000294, -0.00000344, 0.00000530, -0.00000082, -0.00000143, -0.00000002, -0.00000406, 0.00000042, 0.00000088, -0.00000169, 0.00000390, -0.00000265, 0.00000525, -0.00000295, -0.00000148, -0.00000012, -0.00000041, -0.00000034, -0.00000055, -0.00000088, -0.00000047, 0.00000070, 0.00000147, -0.00000044, 0.00000172, -0.00000128, 0.00000066, -0.00000044, 0.00000011, 0.00000115, -0.00000216, },
		{ -0.00000109, -0.00000216, -0.00000008, 0.00000059, 0.00000001, 0.00000228, 0.00000004, 0.00000071, -0.00000135, -0.00000100, 0.00000257, -0.00000176, 0.00000100, -0.00000415, 0.00000121, 0.00000118, -0.00000127, 0.00000132, 0.00000177, -0.00000163, -0.00000053, -0.00000163, 0.00000083, -0.00000023, 0.00000005, -0.00000160, 0.00000353, -0.00000253, -0.00000003, -0.00000513, 0.00000401, 0.00000487, -0.00000152, -0.00000332, 0.00000398, -0.00000822, 0.00000312, -0.00000134, 0.00000125, 0.00000200, -0.00000537, -0.00000073, 0.00000573, -0.00000254, 0.00000307, -0.00001779, 0.00000811, 0.00000073, -0.00000165, -0.00000734, 0.00001315, -0.00001257, 0.00000311, 0.00000088, -0.00000645, 0.00000834, -0.00001534, -0.00001527, 0.00001710, -0.00005804, -0.00000199, -0.00015650, 0.00008008, 0.00014273, 0.00019257, -0.00001480, -0.00007021, 0.00003631, -0.00002914, -0.00001711, -0.00002074, -0.00001612, -0.00000419, 0.00000691, -0.00002125, 0.00000831, -0.00000897, 0.00001948, -0.00000928, -0.00001046, 0.00000309, 0.00000293, -0.00000956, 0.00000979, -0.00000662, 0.00000295, -0.00000054, -0.00000252, 0.00000085, -0.00000243, -0.00000592, 0.00000103, -0.00000140, 0.00001008, -0.00000412, -0.00000230, -0.00000194, 0.00000288, -0.00000334, 0.00000543, -0.00000113, -0.00000138, 0.00000005, -0.00000417, 0.00000075, 0.00000059, -0.00000130, 0.00000346, -0.00000236, 0.00000502, -0.00000308, -0.00000127, -0.00000021, -0.00000027, -0.00000040, -0.00000052, -0.00000085, -0.00000059, 0.00000093, 0.00000124, -0.00000032, 0.00000157, -0.00000127, 0.00000073, -0.00000050, 0.00000027, 0.00000102, -0.00000212, },
		{ -0.00000110, -0.00000222, 0.00000007, 0.00000047, 0.00000014, 0.00000218, -0.00000004, 0.00000080, -0.00000151, -0.00000074, 0.00000237, -0.00000155, 0.00000083, -0.00000410, 0.00000139, 0.00000100, -0.00000116, 0.00000123, 0.00000164, -0.00000159, -0.00000061, -0.00000136, 0.00000061, 0.00000010, -0.00000016, -0.00000147, 0.00000332, -0.00000241, -0.00000025, -0.00000491, 0.00000417, 0.00000477, -0.00000180, -0.00000306, 0.00000359, -0.00000792, 0.00000328, -0.00000123, 0.00000107, 0.00000204, -0.00000559, -0.00000042, 0.00000516, -0.00000177, 0.00000190, -0.00001680, 0.00000803, 0.00000079, -0.00000189, -0.00000711, 0.00001289, -0.00001289, 0.00000387, 0.00000076, -0.00000714, 0.00000952, -0.00001722, -0.00001278, 0.00001322, -0.00005375, -0.00000731, -0.00015064, 0.00008521, 0.00013265, 0.00018882, -0.00000955, -0.00006662, 0.00003560, -0.00002999, -0.00001939, -0.00001837, -0.00001826, -0.00000230, 0.00000543, -0.00001996, 0.00000722, -0.00000780, 0.00001887, -0.00000975, -0.00001002, 0.00000345, 0.00000241, -0.00000889, 0.00000935, -0.00000664, 0.00000273, -0.00000027, -0.00000273, 0.00000099, -0.00000264, -0.00000546, 0.00000063, -0.00000068, 0.00000947, -0.00000411, -0.00000243, -0.00000173, 0.00000279, -0.00000319, 0.00000552, -0.00000143, -0.00000132, 0.00000009, -0.00000425, 0.00000104, 0.00000031, -0.00000090, 0.00000301, -0.00000203, 0.00000474, -0.00000316, -0.00000108, -0.00000028, -0.00000014, -0.00000046, -0.00000049, -0.00000085, -0.00000068, 0.00000113, 0.00000102, -0.00000019, 0.00000141, -0.00000124, 0.00000078, -0.00000054, 0.00000042, 0.00000090, -0.00000207, },
		{ -0.00000113, -0.00000226, 0.00000019, 0.00000036, 0.00000027, 0.00000206, -0.00000009, 0.00000086, -0.00000165, -0.00000049, 0.00000217, -0.00000132, 0.00000063, -0.00000401, 0.00000152, 0.00000083, -0.00000104, 0.00000113, 0.00000153, -0.00000156, -0.00000066, -0.00000112, 0.00000040, 0.00000041, -0.00000037, -0.00000131, 0.00000307, -0.00000227, -0.00000050, -0.00000466, 0.00000430, 0.00000465, -0.00000205, -0.00000282, 0.00000319, -0.00000758, 0.00000342, -0.00000111, 0.00000088, 0.00000206, -0.00000577, -0.00000014, 0.00000460, -0.00000101, 0.00000073, -0.00001571, 0.00000784, 0.00000089, -0.00000215, -0.00000680, 0.00001249, -0.00001311, 0.00000465, 0.00000055, -0.00000767, 0.00001046, -0.00001884, -0.00001042, 0.00000930, -0.00004928, -0.00001302, -0.00014366, 0.00008868, 0.00012333, 0.00018530, -0.00000452, -0.00006285, 0.00003487, -0.00003121, -0.00002135, -0.00001619, -0.00002023, -0.00000050, 0.00000388, -0.00001853, 0.00000601, -0.00000651, 0.00001808, -0.00001009, -0.00000955, 0.00000372, 0.00000191, -0.00000820, 0.00000887, -0.00000665, 0.00000250, 0.00000002, -0.00000293, 0.00000111, -0.00000284, -0.00000500, 0.00000025, 0.00000004, 0.00000881, -0.00000406, -0.00000256, -0.00000152, 0.00000266, -0.00000298, 0.00000556, -0.00000172, -0.00000124, 0.00000011, -0.00000428, 0.00000129, 0.00000006, -0.00000052, 0.00000255, -0.00000169, 0.00000441, -0.00000319, -0.00000092, -0.00000035, -0.00000001, -0.00000052, -0.00000044, -0.00000088, -0.00000074, 0.00000132, 0.00000081, -0.00000006, 0.00000123, -0.00000118, 0.00000081, -0.00000056, 0.00000055, 0.00000077, -0.00000200, },
		{ -0.00000118, -0.00000228, 0.00000029, 0.00000026, 0.00000040, 0.00000193, -0.00000012, 0.00000089, -0.00000176, -0.00000026, 0.00000197, -0.00000109, 0.00000040, -0.00000388, 0.00000162, 0.00000068, -0.00000092, 0.00000103, 0.00000142, -0.00000155, -0.00000067, -0.00000090, 0.00000020, 0.00000070, -0.00000059, -0.00000114, 0.00000279, -0.00000211, -0.00000076, -0.00000436, 0.00000440, 0.00000451, -0.00000225, -0.00000258, 0.00000276, -0.00000722, 0.00000355, -0.00000100, 0.00000068, 0.00000205, -0.00000591, 0.00000011, 0.00000406, -0.00000028, -0.00000042, -0.00001456, 0.00000756, 0.00000100, -0.00000241, -0.00000642, 0.00001196, -0.00001323, 0.00000544, 0.00000024, -0.00000806, 0.00001114, -0.00002015, -0.00000825, 0.00000543, -0.00004474, -0.00001897, -0.00013571, 0.00009058, 0.00011477, 0.00018196, 0.00000033, -0.00005890, 0.00003411, -0.00003279, -0.00002295, -0.00001425, -0.00002196, 0.00000116, 0.00000232, -0.00001699, 0.00000472, -0.00000512, 0.00001713, -0.00001031, -0.00000905, 0.00000389, 0.00000145, -0.00000749, 0.00000836, -0.00000667, 0.00000229, 0.00000030, -0.00000311, 0.00000120, -0.00000301, -0.00000455, -0.00000012, 0.00000073, 0.00000811, -0.00000398, -0.00000270, -0.00000130, 0.00000250, -0.00000273, 0.00000556, -0.00000200, -0.00000115, 0.00000011, -0.00000426, 0.00000148, -0.00000016, -0.00000016, 0.00000210, -0.00000133, 0.00000405, -0.00000316, -0.00000078, -0.00000040, 0.00000010, -0.00000057, -0.00000038, -0.00000093, -0.00000078, 0.00000147, 0.00000062, 0.00000007, 0.00000105, -0.00000110, 0.00000082, -0.00000056, 0.00000067, 0.00000064, -0.00000192, },
		{ -0.00000124, -0.00000227, 0.00000038, 0.00000017, 0.00000053, 0.00000178, -0.00000013, 0.00000090, -0.00000184, -0.00000004, 0.00000177, -0.00000085, 0.00000016, -0.00000371, 0.00000168, 0.00000054, -0.00000081, 0.00000091, 0.00000132, -0.00000155, -0.00000065, -0.00000072, 0.00000004, 0.00000097, -0.00000079, -0.00000097, 0.00000249, -0.00000192, -0.00000103, -0.00000404, 0.00000448, 0.00000434, -0.00000241, -0.00000237, 0.00000234, -0.00000683, 0.00000367, -0.00000090, 0.00000048, 0.00000201, -0.00000599, 0.00000032, 0.00000355, 0.00000039, -0.00000154, -0.00001335, 0.00000721, 0.00000113, -0.00000266, -0.00000600, 0.00001131, -0.00001324, 0.00000623, -0.00000016, -0.00000828, 0.00001155, -0.00002114, -0.00000634, 0.00000169, -0.00004025, -0.00002504, -0.00012693, 0.00009100, 0.00010697, 0.00017879, 0.00000501, -0.00005478, 0.00003329, -0.00003474, -0.00002416, -0.00001259, -0.00002340, 0.00000262, 0.00000080, -0.00001540, 0.00000337, -0.00000366, 0.00001603, -0.00001041, -0.00000852, 0.00000397, 0.00000104, -0.00000678, 0.00000784, -0.00000671, 0.00000210, 0.00000058, -0.00000325, 0.00000126, -0.00000316, -0.00000411, -0.00000046, 0.00000140, 0.00000740, -0.00000387, -0.00000282, -0.00000110, 0.00000233, -0.00000243, 0.00000552, -0.00000228, -0.00000105, 0.00000009, -0.00000419, 0.00000163, -0.00000034, 0.00000017, 0.00000166, -0.00000097, 0.00000365, -0.00000309, -0.00000066, -0.00000044, 0.00000020, -0.00000061, -0.00000032, -0.00000100, -0.00000078, 0.00000160, 0.00000045, 0.00000018, 0.00000086, -0.00000100, 0.00000081, -0.00000054, 0.00000077, 0.00000052, -0.00000183, },
		{ -0.00000132, -0.00000225, 0.00000044, 0.00000010, 0.00000064, 0.00000163, -0.00000012, 0.00000089, -0.00000189, 0.00000015, 0.00000158, -0.00000062, -0.00000010, -0.00000351, 0.00000170, 0.00000042, -0.00000070, 0.00000079, 0.00000122, -0.00000156, -0.00000060, -0.00000057, -0.00000010, 0.00000121, -0.00000097, -0.00000080, 0.00000218, -0.00000173, -0.00000132, -0.00000368, 0.00000452, 0.00000415, -0.00000252, -0.00000218, 0.00000191, -0.00000642, 0.00000380, -0.00000081, 0.00000029, 0.00000194, -0.00000601, 0.00000049, 0.00000309, 0.00000102, -0.00000260, -0.00001212, 0.00000680, 0.00000125, -0.00000289, -0.00000555, 0.00001056, -0.00001315, 0.00000702, -0.00000065, -0.00000836, 0.00001167, -0.00002176, -0.00000474, -0.00000183, -0.00003589, -0.00003108, -0.00011750, 0.00009004, 0.00009992, 0.00017577, 0.00000955, -0.00005048, 0.00003239, -0.00003702, -0.00002496, -0.00001124, -0.00002453, 0.00000384, -0.00000064, -0.00001379, 0.00000201, -0.00000216, 0.00001480, -0.00001040, -0.00000797, 0.00000396, 0.00000068, -0.00000607, 0.00000730, -0.00000677, 0.00000194, 0.00000083, -0.00000336, 0.00000127, -0.00000328, -0.00000368, -0.00000077, 0.00000202, 0.00000666, -0.00000373, -0.00000292, -0.00000091, 0.00000216, -0.00000210, 0.00000543, -0.00000254, -0.00000094, 0.00000005, -0.00000409, 0.00000171, -0.00000048, 0.00000046, 0.00000124, -0.00000061, 0.00000323, -0.00000298, -0.00000057, -0.00000046, 0.00000029, -0.00000065, -0.00000025, -0.00000110, -0.00000076, 0.00000170, 0.00000030, 0.00000027, 0.00000067, -0.00000088, 0.00000078, -0.00000050, 0.00000085, 0.00000040, -0.00000175, },
		{ -0.00000141, -0.00000220, 0.00000048, 0.00000005, 0.00000075, 0.00000147, -0.00000010, 0.00000085, -0.00000190, 0.00000032, 0.00000141, -0.00000039, -0.00000036, -0.00000328, 0.00000168, 0.00000032, -0.00000059, 0.00000067, 0.00000113, -0.00000157, -0.00000052, -0.00000045, -0.00000020, 0.00000140, -0.00000113, -0.00000064, 0.00000187, -0.00000152, -0.00000160, -0.00000329, 0.00000453, 0.00000394, -0.00000258, -0.00000203, 0.00000150, -0.00000600, 0.00000392, -0.00000074, 0.00000011, 0.00000182, -0.00000598, 0.00000060, 0.00000267, 0.00000157, -0.00000360, -0.00001087, 0.00000634, 0.00000136, -0.00000307, -0.00000509, 0.00000970, -0.00001294, 0.00000779, -0.00000122, -0.00000828, 0.00001153, -0.00002202, -0.00000348, -0.00000508, -0.00003176, -0.00003698, -0.00010757, 0.00008784, 0.00009358, 0.00017288, 0.00001394, -0.00004599, 0.00003136, -0.00003961, -0.00002536, -0.00001022, -0.00002530, 0.00000478, -0.00000193, -0.00001220, 0.00000065, -0.00000063, 0.00001346, -0.00001027, -0.00000740, 0.00000385, 0.00000038, -0.00000537, 0.00000676, -0.00000685, 0.00000182, 0.00000106, -0.00000343, 0.00000123, -0.00000336, -0.00000329, -0.00000104, 0.00000261, 0.00000593, -0.00000358, -0.00000301, -0.00000075, 0.00000198, -0.00000174, 0.00000531, -0.00000278, -0.00000082, -0.00000001, -0.00000393, 0.00000174, -0.00000057, 0.00000072, 0.00000085, -0.00000027, 0.00000279, -0.00000283, -0.00000049, -0.00000048, 0.00000036, -0.00000067, -0.00000019, -0.00000120, -0.00000071, 0.00000177, 0.00000018, 0.00000035, 0.00000049, -0.00000076, 0.00000074, -0.00000044, 0.00000091, 0.00000029, -0.00000166, },
		{ -0.00000150, -0.00000214, 0.00000050, 0.00000001, 0.00000084, 0.00000131, -0.00000006, 0.00000078, -0.00000188, 0.00000045, 0.00000126, -0.00000018, -0.00000062, -0.00000303, 0.00000164, 0.00000024, -0.00000049, 0.00000055, 0.00000104, -0.00000159, -0.00000040, -0.00000037, -0.00000027, 0.00000156, -0.00000125, -0.00000051, 0.00000155, -0.00000131, -0.00000189, -0.00000288, 0.00000452, 0.00000371, -0.00000260, -0.00000191, 0.00000110, -0.00000557, 0.00000405, -0.00000070, -0.00000006, 0.00000167, -0.00000588, 0.00000066, 0.00000231, 0.00000205, -0.00000452, -0.00000963, 0.00000584, 0.00000146, -0.00000322, -0.00000463, 0.00000876, -0.00001263, 0.00000853, -0.00000185, -0.00000807, 0.00001112, -0.00002190, -0.00000260, -0.00000799, -0.00002794, -0.00004262, -0.00009730, 0.00008452, 0.00008790, 0.00017014, 0.00001820, -0.00004131, 0.00003014, -0.00004245, -0.00002537, -0.00000954, -0.00002570, 0.00000540, -0.00000306, -0.00001068, -0.00000066, 0.00000088, 0.00001203, -0.00001002, -0.00000683, 0.00000365, 0.00000014, -0.00000470, 0.00000622, -0.00000695, 0.00000173, 0.00000126, -0.00000345, 0.00000114, -0.00000340, -0.00000292, -0.00000127, 0.00000314, 0.00000520, -0.00000342, -0.00000308, -0.00000061, 0.00000181, -0.00000135, 0.00000514, -0.00000301, -0.00000070, -0.00000008, -0.00000375, 0.00000172, -0.00000062, 0.00000093, 0.00000049, 0.00000006, 0.00000235, -0.00000265, -0.00000044, -0.00000048, 0.00000042, -0.00000068, -0.00000013, -0.00000132, -0.00000064, 0.00000181, 0.00000008, 0.00000040, 0.00000032, -0.00000062, 0.00000069, -0.00000037, 0.00000096, 0.00000019, -0.00000157, },
		{ -0.00000160, -0.00000207, 0.00000051, -0.00000001, 0.00000092, 0.00000116, -0.00000001, 0.00000070, -0.00000183, 0.00000056, 0.00000113, 0.00000002, -0.00000087, -0.00000277, 0.00000156, 0.00000017, -0.00000041, 0.00000043, 0.00000095, -0.00000161, -0.00000027, -0.00000033, -0.00000030, 0.00000168, -0.00000134, -0.00000039, 0.00000124, -0.00000111, -0.00000217, -0.00000245, 0.00000447, 0.00000347, -0.00000257, -0.00000182, 0.00000072, -0.00000514, 0.00000419, -0.00000067, -0.00000021, 0.00000148, -0.00000573, 0.00000067, 0.00000200, 0.00000245, -0.00000535, -0.00000841, 0.00000533, 0.00000153, -0.00000331, -0.00000418, 0.00000775, -0.00001220, 0.00000922, -0.00000252, -0.00000775, 0.00001046, -0.00002141, -0.00000211, -0.00001053, -0.00002449, -0.00004789, -0.00008685, 0.00008022, 0.00008283, 0.00016755, 0.00002231, -0.00003640, 0.00002868, -0.00004548, -0.00002502, -0.00000920, -0.00002573, 0.00000570, -0.00000397, -0.00000924, -0.00000192, 0.00000237, 0.00001052, -0.00000967, -0.00000626, 0.00000337, -0.00000003, -0.00000405, 0.00000568, -0.00000708, 0.00000170, 0.00000142, -0.00000342, 0.00000100, -0.00000339, -0.00000258, -0.00000145, 0.00000361, 0.00000447, -0.00000326, -0.00000311, -0.00000051, 0.00000165, -0.00000094, 0.00000494, -0.00000321, -0.00000058, -0.00000015, -0.00000352, 0.00000165, -0.00000063, 0.00000110, 0.00000017, 0.00000035, 0.00000190, -0.00000245, -0.00000040, -0.00000047, 0.00000046, -0.00000067, -0.00000009, -0.00000145, -0.00000055, 0.00000182, 0.00000001, 0.00000043, 0.00000016, -0.00000049, 0.00000064, -0.00000028, 0.00000098, 0.00000011, -0.00000149, },
		{ -0.00000170, -0.00000199, 0.00000049, -0.00000001, 0.00000098, 0.00000101, 0.00000004, 0.00000061, -0.00000175, 0.00000064, 0.00000102, 0.00000019, -0.00000111, -0.00000249, 0.00000146, 0.00000013, -0.00000033, 0.00000032, 0.00000085, -0.00000163, -0.00000011, -0.00000031, -0.00000029, 0.00000174, -0.00000139, -0.00000031, 0.00000094, -0.00000091, -0.00000244, -0.00000201, 0.00000439, 0.00000322, -0.00000250, -0.00000177, 0.00000036, -0.00000470, 0.00000433, -0.00000068, -0.00000035, 0.00000125, -0.00000551, 0.00000062, 0.00000175, 0.00000277, -0.00000609, -0.00000722, 0.00000480, 0.00000157, -0.00000333, -0.00000377, 0.00000669, -0.00001167, 0.00000986, -0.00000322, -0.00000732, 0.00000958, -0.00002058, -0.00000201, -0.00001266, -0.00002147, -0.00005269, -0.00007637, 0.00007509, 0.00007829, 0.00016514, 0.00002628, -0.00003126, 0.00002692, -0.00004863, -0.00002433, -0.00000918, -0.00002539, 0.00000565, -0.00000466, -0.00000794, -0.00000308, 0.00000380, 0.00000896, -0.00000922, -0.00000569, 0.00000302, -0.00000013, -0.00000343, 0.00000514, -0.00000723, 0.00000170, 0.00000154, -0.00000334, 0.00000081, -0.00000334, -0.00000228, -0.00000160, 0.00000403, 0.00000377, -0.00000309, -0.00000313, -0.00000043, 0.00000151, -0.00000053, 0.00000470, -0.00000340, -0.00000046, -0.00000024, -0.00000327, 0.00000153, -0.00000058, 0.00000122, -0.00000010, 0.00000061, 0.00000147, -0.00000223, -0.00000037, -0.00000045, 0.00000049, -0.00000065, -0.00000005, -0.00000158, -0.00000044, 0.00000180, -0.00000003, 0.00000044, 0.00000002, -0.00000036, 0.00000057, -0.00000017, 0.00000098, 0.00000004, -0.00000141, },
		{ -0.00000179, -0.00000190, 0.00000047, 0.00000000, 0.00000102, 0.00000087, 0.00000010, 0.00000050, -0.00000164, 0.00000069, 0.00000095, 0.00000033, -0.00000134, -0.00000222, 0.00000133, 0.00000010, -0.00000026, 0.00000021, 0.00000075, -0.00000164, 0.00000007, -0.00000033, -0.00000024, 0.00000176, -0.00000140, -0.00000026, 0.00000066, -0.00000072, -0.00000269, -0.00000155, 0.00000429, 0.00000297, -0.00000239, -0.00000176, 0.00000002, -0.00000426, 0.00000447, -0.00000071, -0.00000046, 0.00000099, -0.00000524, 0.00000051, 0.00000156, 0.00000302, -0.00000672, -0.00000607, 0.00000426, 0.00000157, -0.00000329, -0.00000339, 0.00000557, -0.00001102, 0.00001042, -0.00000393, -0.00000682, 0.00000850, -0.00001941, -0.00000231, -0.00001437, -0.00001892, -0.00005694, -0.00006599, 0.00006925, 0.00007422, 0.00016293, 0.00003010, -0.00002586, 0.00002479, -0.00005182, -0.00002338, -0.00000945, -0.00002468, 0.00000525, -0.00000509, -0.00000677, -0.00000412, 0.00000515, 0.00000735, -0.00000866, -0.00000515, 0.00000260, -0.00000018, -0.00000284, 0.00000460, -0.00000740, 0.00000175, 0.00000163, -0.00000322, 0.00000056, -0.00000323, -0.00000202, -0.00000170, 0.00000438, 0.00000308, -0.00000291, -0.00000311, -0.00000040, 0.00000140, -0.00000010, 0.00000442, -0.00000356, -0.00000034, -0.00000032, -0.00000299, 0.00000137, -0.00000050, 0.00000129, -0.00000033, 0.00000084, 0.00000105, -0.00000199, -0.00000035, -0.00000042, 0.00000050, -0.00000061, -0.00000003, -0.00000171, -0.00000032, 0.00000176, -0.00000005, 0.00000041, -0.00000009, -0.00000023, 0.00000051, -0.00000006, 0.00000097, -0.00000001, -0.00000135, },
		{ -0.00000188, -0.00000180, 0.00000042, 0.00000004, 0.00000105, 0.00000074, 0.00000016, 0.00000039, -0.00000150, 0.00000070, 0.00000090, 0.00000045, -0.00000155, -0.00000194, 0.00000119, 0.00000010, -0.00000021, 0.00000011, 0.00000063, -0.00000163, 0.00000025, -0.00000037, -0.00000017, 0.00000174, -0.00000137, -0.00000024, 0.00000040, -0.00000055, -0.00000293, -0.00000108, 0.00000415, 0.00000271, -0.00000225, -0.00000178, -0.00000029, -0.00000382, 0.00000461, -0.00000076, -0.00000057, 0.00000070, -0.00000492, 0.00000035, 0.00000142, 0.00000318, -0.00000725, -0.00000498, 0.00000373, 0.00000154, -0.00000318, -0.00000307, 0.00000442, -0.00001026, 0.00001090, -0.00000462, -0.00000628, 0.00000727, -0.00001794, -0.00000298, -0.00001566, -0.00001686, -0.00006058, -0.00005585, 0.00006286, 0.00007054, 0.00016094, 0.00003376, -0.00002020, 0.00002223, -0.00005496, -0.00002220, -0.00000998, -0.00002364, 0.00000451, -0.00000527, -0.00000578, -0.00000504, 0.00000641, 0.00000573, -0.00000802, -0.00000463, 0.00000213, -0.00000016, -0.00000227, 0.00000406, -0.00000758, 0.00000184, 0.00000168, -0.00000305, 0.00000026, -0.00000308, -0.00000179, -0.00000175, 0.00000467, 0.00000242, -0.00000274, -0.00000307, -0.00000040, 0.00000131, 0.00000033, 0.00000412, -0.00000369, -0.00000023, -0.00000040, -0.00000270, 0.00000117, -0.00000038, 0.00000132, -0.00000052, 0.00000102, 0.00000065, -0.00000174, -0.00000034, -0.00000039, 0.00000050, -0.00000055, -0.00000003, -0.00000183, -0.00000018, 0.00000170, -0.00000003, 0.00000036, -0.00000019, -0.00000011, 0.00000045, 0.00000006, 0.00000093, -0.00000004, -0.00000130, },
		{ -0.00000196, -0.00000171, 0.00000037, 0.00000008, 0.00000105, 0.00000062, 0.00000021, 0.00000027, -0.00000135, 0.00000070, 0.00000088, 0.00000054, -0.00000173, -0.00000167, 0.00000103, 0.00000010, -0.00000017, 0.00000002, 0.00000051, -0.00000161, 0.00000044, -0.00000042, -0.00000006, 0.00000167, -0.00000129, -0.00000027, 0.00000017, -0.00000039, -0.00000315, -0.00000060, 0.00000397, 0.00000247, -0.00000209, -0.00000183, -0.00000058, -0.00000337, 0.00000475, -0.00000084, -0.00000066, 0.00000038, -0.00000455, 0.00000015, 0.00000133, 0.00000327, -0.00000768, -0.00000394, 0.00000321, 0.00000147, -0.00000299, -0.00000280, 0.00000325, -0.00000939, 0.00001127, -0.00000528, -0.00000571, 0.00000591, -0.00001623, -0.00000400, -0.00001652, -0.00001531, -0.00006356, -0.00004606, 0.00005603, 0.00006718, 0.00015920, 0.00003726, -0.00001426, 0.00001917, -0.00005798, -0.00002087, -0.00001073, -0.00002230, 0.00000345, -0.00000518, -0.00000496, -0.00000581, 0.00000756, 0.00000410, -0.00000728, -0.00000415, 0.00000162, -0.00000009, -0.00000173, 0.00000350, -0.00000777, 0.00000197, 0.00000169, -0.00000283, -0.00000008, -0.00000289, -0.00000159, -0.00000177, 0.00000490, 0.00000178, -0.00000258, -0.00000300, -0.00000044, 0.00000126, 0.00000075, 0.00000378, -0.00000380, -0.00000014, -0.00000047, -0.00000240, 0.00000094, -0.00000022, 0.00000130, -0.00000066, 0.00000116, 0.00000027, -0.00000149, -0.00000033, -0.00000035, 0.00000048, -0.00000047, -0.00000006, -0.00000194, -0.00000005, 0.00000162, -0.00000000, 0.00000028, -0.00000026, -0.00000000, 0.00000040, 0.00000019, 0.00000088, -0.00000006, -0.00000126, },
		{ -0.00000203, -0.00000162, 0.00000031, 0.00000015, 0.00000103, 0.00000052, 0.00000025, 0.00000015, -0.00000118, 0.00000067, 0.00000088, 0.00000059, -0.00000188, -0.00000141, 0.00000086, 0.00000012, -0.00000015, -0.00000006, 0.00000037, -0.00000157, 0.00000062, -0.00000050, 0.00000007, 0.00000156, -0.00000117, -0.00000034, -0.00000003, -0.00000026, -0.00000335, -0.00000012, 0.00000377, 0.00000223, -0.00000190, -0.00000191, -0.00000084, -0.00000292, 0.00000489, -0.00000093, -0.00000073, 0.00000004, -0.00000414, -0.00000010, 0.00000129, 0.00000329, -0.00000801, -0.00000296, 0.00000269, 0.00000137, -0.00000274, -0.00000261, 0.00000207, -0.00000842, 0.00001153, -0.00000587, -0.00000516, 0.00000447, -0.00001430, -0.00000533, -0.00001698, -0.00001427, -0.00006585, -0.00003671, 0.00004889, 0.00006406, 0.00015774, 0.00004061, -0.00000804, 0.00001557, -0.00006077, -0.00001946, -0.00001162, -0.00002070, 0.00000209, -0.00000484, -0.00000433, -0.00000643, 0.00000858, 0.00000249, -0.00000647, -0.00000372, 0.00000107, 0.00000002, -0.00000121, 0.00000293, -0.00000796, 0.00000213, 0.00000168, -0.00000258, -0.00000046, -0.00000265, -0.00000142, -0.00000175, 0.00000508, 0.00000117, -0.00000241, -0.00000291, -0.00000051, 0.00000123, 0.00000117, 0.00000342, -0.00000387, -0.00000006, -0.00000052, -0.00000210, 0.00000069, -0.00000004, 0.00000124, -0.00000076, 0.00000125, -0.00000007, -0.00000125, -0.00000032, -0.00000031, 0.00000046, -0.00000037, -0.00000010, -0.00000204, 0.00000009, 0.00000153, 0.00000005, 0.00000017, -0.00000030, 0.00000009, 0.00000035, 0.00000031, 0.00000082, -0.00000006, -0.00000124, },
		{ -0.00000209, -0.00000153, 0.00000025, 0.00000022, 0.00000099, 0.00000043, 0.00000029, 0.00000004, -0.00000100, 0.00000061, 0.00000091, 0.00000062, -0.00000201, -0.00000117, 0.00000069, 0.00000015, -0.00000014, -0.00000012, 0.00000022, -0.00000150, 0.00000079, -0.00000058, 0.00000022, 0.00000142, -0.00000102, -0.00000045, -0.00000020, -0.00000014, -0.00000352, 0.00000036, 0.00000354, 0.00000201, -0.00000171, -0.00000201, -0.00000109, -0.00000246, 0.00000501, -0.00000104, -0.00000080, -0.00000031, -0.00000371, -0.00000038, 0.00000129, 0.00000325, -0.00000824, -0.00000204, 0.00000219, 0.00000124, -0.00000241, -0.00000249, 0.00000089, -0.00000734, 0.00001167, -0.00000639, -0.00000465, 0.00000299, -0.00001221, -0.00000693, -0.00001707, -0.00001371, -0.00006743, -0.00002790, 0.00004157, 0.00006111, 0.00015657, 0.00004381, -0.00000154, 0.00001136, -0.00006325, -0.00001805, -0.00001262, -0.00001889, 0.00000046, -0.00000424, -0.00000389, -0.00000689, 0.00000947, 0.00000090, -0.00000558, -0.00000334, 0.00000052, 0.00000016, -0.00000071, 0.00000234, -0.00000813, 0.00000231, 0.00000164, -0.00000230, -0.00000088, -0.00000238, -0.00000129, -0.00000170, 0.00000519, 0.00000059, -0.00000224, -0.00000280, -0.00000062, 0.00000125, 0.00000157, 0.00000304, -0.00000392, -0.00000001, -0.00000056, -0.00000180, 0.00000043, 0.00000017, 0.00000115, -0.00000082, 0.00000130, -0.00000038, -0.00000100, -0.00000032, -0.00000027, 0.00000042, -0.00000026, -0.00000018, -0.00000211, 0.00000022, 0.00000142, 0.00000012, 0.00000005, -0.00000032, 0.00000017, 0.00000031, 0.00000044, 0.00000075, -0.00000005, -0.00000124, },
		{ -0.00000213, -0.00000145, 0.00000019, 0.00000030, 0.00000094, 0.00000037, 0.00000031, -0.00000006, -0.00000081, 0.00000055, 0.00000096, 0.00000061, -0.00000210, -0.00000094, 0.00000052, 0.00000019, -0.00000014, -0.00000018, 0.00000006, -0.00000140, 0.00000095, -0.00000066, 0.00000039, 0.00000125, -0.00000083, -0.00000059, -0.00000034, -0.00000006, -0.00000367, 0.00000083, 0.00000328, 0.00000181, -0.00000150, -0.00000213, -0.00000131, -0.00000199, 0.00000512, -0.00000116, -0.00000087, -0.00000068, -0.00000324, -0.00000068, 0.00000132, 0.00000316, -0.00000839, -0.00000118, 0.00000170, 0.00000108, -0.00000203, -0.00000244, -0.00000028, -0.00000617, 0.00001167, -0.00000681, -0.00000420, 0.00000153, -0.00001002, -0.00000874, -0.00001683, -0.00001360, -0.00006831, -0.00001968, 0.00003415, 0.00005827, 0.00015573, 0.00004687, 0.00000522, 0.00000652, -0.00006534, -0.00001671, -0.00001366, -0.00001692, -0.00000139, -0.00000342, -0.00000363, -0.00000718, 0.00001021, -0.00000064, -0.00000462, -0.00000302, -0.00000004, 0.00000033, -0.00000022, 0.00000172, -0.00000828, 0.00000250, 0.00000158, -0.00000199, -0.00000133, -0.00000208, -0.00000117, -0.00000162, 0.00000526, 0.00000003, -0.00000207, -0.00000268, -0.00000076, 0.00000130, 0.00000196, 0.00000263, -0.00000393, 0.00000003, -0.00000057, -0.00000151, 0.00000016, 0.00000039, 0.00000103, -0.00000084, 0.00000130, -0.00000066, -0.00000076, -0.00000031, -0.00000023, 0.00000038, -0.00000012, -0.00000028, -0.00000217, 0.00000033, 0.00000132, 0.00000020, -0.00000010, -0.00000031, 0.00000024, 0.00000029, 0.00000056, 0.00000066, -0.00000002, -0.00000125, },
		{ -0.00000215, -0.00000138, 0.00000012, 0.00000039, 0.00000086, 0.00000032, 0.00000032, -0.00000015, -0.00000063, 0.00000047, 0.00000103, 0.00000058, -0.00000216, -0.00000074, 0.00000034, 0.00000024, -0.00000015, -0.00000023, -0.00000013, -0.00000126, 0.00000108, -0.00000073, 0.00000056, 0.00000105, -0.00000061, -0.00000078, -0.00000045, 0.00000000, -0.00000378, 0.00000130, 0.00000299, 0.00000164, -0.00000130, -0.00000227, -0.00000152, -0.00000151, 0.00000520, -0.00000129, -0.00000093, -0.00000104, -0.00000276, -0.00000100, 0.00000137, 0.00000303, -0.00000845, -0.00000038, 0.00000121, 0.00000090, -0.00000159, -0.00000247, -0.00000142, -0.00000491, 0.00001153, -0.00000711, -0.00000386, 0.00000011, -0.00000779, -0.00001070, -0.00001631, -0.00001391, -0.00006849, -0.00001211, 0.00002674, 0.00005547, 0.00015521, 0.00004981, 0.00001221, 0.00000100, -0.00006696, -0.00001553, -0.00001466, -0.00001486, -0.00000341, -0.00000240, -0.00000355, -0.00000731, 0.00001080, -0.00000213, -0.00000360, -0.00000276, -0.00000058, 0.00000052, 0.00000026, 0.00000106, -0.00000839, 0.00000269, 0.00000151, -0.00000167, -0.00000180, -0.00000175, -0.00000108, -0.00000152, 0.00000528, -0.00000050, -0.00000189, -0.00000254, -0.00000092, 0.00000138, 0.00000232, 0.00000220, -0.00000391, 0.00000005, -0.00000056, -0.00000124, -0.00000011, 0.00000061, 0.00000089, -0.00000083, 0.00000127, -0.00000090, -0.00000054, -0.00000029, -0.00000019, 0.00000033, 0.00000002, -0.00000040, -0.00000220, 0.00000044, 0.00000122, 0.00000029, -0.00000026, -0.00000028, 0.00000028, 0.00000028, 0.00000067, 0.00000057, 0.00000002, -0.00000128, },
		},
		{ { -0.00000049, -0.00000069, 0.00000364, -0.00000022, -0.00000070, -0.00000112, -0.00000118, -0.00000079, 0.00000053, 0.00000069, -0.00000610, 0.00000076, 0.00000349, 0.00000226, -0.00000076, 0.00000153, 0.00000080, -0.00000074, -0.00000367, -0.00000169, -0.00000364, -0.00000086, 0.00000767, -0.00000089, 0.00000058, 0.00000090, 0.00000057, -0.00000180, 0.00000184, -0.00000321, -0.00000002, -0.00000125, 0.00000152, 0.00000286, -0.00001110, 0.00000288, -0.00000281, 0.00000323, 0.00000224, -0.00000023, 0.00000035, 0.00000055, -0.00000655, -0.00000311, 0.00000564, -0.00000607, 0.00001644, -0.00000581, -0.00000126, 0.00000164, -0.00000509, -0.00000138, -0.00001894, -0.00000234, 0.00000397, -0.00001395, -0.00001371, -0.00000807, -0.00007370, 0.00001040, 0.00008053, 0.00003414, 0.00007498, -0.00000489, -0.00010526, -0.00000489, 0.00007498, 0.00003414, 0.00008053, 0.00001040, -0.00007370, -0.00000807, -0.00001371, -0.00001395, 0.00000397, -0.00000234, -0.00001894, -0.00000138, -0.00000509, 0.00000164, -0.00000126, -0.00000581, 0.00001644, -0.00000607, 0.00000564, -0.00000311, -0.00000655, 0.00000055, 0.00000035, -0.00000023, 0.00000224, 0.00000323, -0.00000281, 0.00000288, -0.00001110, 0.00000286, 0.00000152, -0.00000125, -0.00000002, -0.00000321, 0.00000184, -0.00000180, 0.00000057, 0.00000090, 0.00000058, -0.00000089, 0.00000767, -0.00000086, -0.00000364, -0.00000169, -0.00000367, -0.00000074, 0.00000080, 0.00000153, -0.00000076, 0.00000226, 0.00000349, 0.00000076, -0.00000610, 0.00000069, 0.00000053, -0.00000079, -0.00000118, -0.00000112, -0.00000070, -0.00000022, 0.00000364, -0.00000069, },
		{ -0.00000049, -0.00000098, 0.00000356, 0.00000026, -0.00000096, -0.00000092, -0.00000138, -0.00000065, 0.00000029, 0.00000114, -0.00000613, 0.00000029, 0.00000340, 0.00000246, -0.00000062, 0.00000128, 0.00000110, -0.00000074, -0.00000363, -0.00000189, -0.00000313, -0.00000184, 0.00000797, -0.00000048, 0.00000035, 0.00000080, 0.00000060, -0.00000169, 0.00000177, -0.00000264, -0.00000074, -0.00000067, 0.00000050, 0.00000394, -0.00001128, 0.00000242, -0.00000233, 0.00000257, 0.00000277, -0.00000081, 0.00000106, -0.00000007, -0.00000550, -0.00000447, 0.00000657, -0.00000665, 0.00001593, -0.00000412, -0.00000255, 0.00000270, -0.00000651, 0.00000061, -0.00001958, -0.00000327, 0.00000426, -0.00001273, -0.00001558, -0.00000539, -0.00007352, -0.00000017, 0.00008554, 0.00003630, 0.00007905, -0.00000233, -0.00010538, -0.00000719, 0.00007087, 0.00003242, 0.00007519, 0.00001981, -0.00007245, -0.00001135, -0.00001163, -0.00001519, 0.00000370, -0.00000160, -0.00001800, -0.00000346, -0.00000382, 0.00000080, -0.00000016, -0.00000722, 0.00001654, -0.00000513, 0.00000446, -0.00000164, -0.00000759, 0.00000113, -0.00000028, 0.00000019, 0.00000184, 0.00000377, -0.00000312, 0.00000311, -0.00001067, 0.00000162, 0.00000257, -0.00000180, 0.00000062, -0.00000364, 0.00000179, -0.00000184, 0.00000047, 0.00000104, 0.00000077, -0.00000119, 0.00000722, 0.00000016, -0.00000408, -0.00000159, -0.00000366, -0.00000074, 0.00000051, 0.00000172, -0.00000082, 0.00000199, 0.00000358, 0.00000119, -0.00000599, 0.00000021, 0.00000077, -0.00000095, -0.00000098, -0.00000132, -0.00000046, -0.00000063, 0.00000364, -0.00000038, },
		{ -0.00000047, -0.00000123, 0.00000339, 0.00000079, -0.00000121, -0.00000073, -0.00000156, -0.00000053, 0.00000005, 0.00000155, -0.00000608, -0.00000022, 0.00000332, 0.00000261, -0.00000040, 0.00000098, 0.00000140, -0.00000074, -0.00000355, -0.00000216, -0.00000258, -0.00000275, 0.00000809, 0.00000006, 0.00000007, 0.00000075, 0.00000057, -0.00000152, 0.00000159, -0.00000195, -0.00000148, -0.00000011, -0.00000043, 0.00000479, -0.00001116, 0.00000173, -0.00000170, 0.00000180, 0.00000341, -0.00000152, 0.00000184, -0.00000069, -0.00000448, -0.00000570, 0.00000723, -0.00000685, 0.00001504, -0.00000221, -0.00000396, 0.00000389, -0.00000796, 0.00000237, -0.00001983, -0.00000446, 0.00000464, -0.00001167, -0.00001710, -0.00000350, -0.00007184, -0.00001176, 0.00008999, 0.00003898, 0.00008303, 0.00000053, -0.00010575, -0.00000924, 0.00006673, 0.00003104, 0.00006970, 0.00002796, -0.00006991, -0.00001504, -0.00000953, -0.00001633, 0.00000334, -0.00000096, -0.00001687, -0.00000549, -0.00000282, 0.00000026, 0.00000070, -0.00000830, 0.00001622, -0.00000383, 0.00000309, -0.00000013, -0.00000858, 0.00000164, -0.00000077, 0.00000045, 0.00000155, 0.00000421, -0.00000327, 0.00000314, -0.00001002, 0.00000027, 0.00000357, -0.00000227, 0.00000114, -0.00000389, 0.00000162, -0.00000180, 0.00000028, 0.00000124, 0.00000090, -0.00000135, 0.00000664, 0.00000119, -0.00000443, -0.00000157, -0.00000361, -0.00000072, 0.00000024, 0.00000185, -0.00000080, 0.00000169, 0.00000366, 0.00000160, -0.00000580, -0.00000029, 0.00000100, -0.00000110, -0.00000077, -0.00000150, -0.00000025, -0.00000097, 0.00000354, -0.00000004, },
		{ -0.00000044, -0.00000144, 0.00000313, 0.00000136, -0.00000145, -0.00000056, -0.00000170, -0.00000044, -0.00000016, 0.00000189, -0.00000592, -0.00000076, 0.00000325, 0.00000268, -0.00000011, 0.00000064, 0.00000170, -0.00000073, -0.00000342, -0.00000250, -0.00000200, -0.00000356, 0.00000805, 0.00000070, -0.00000023, 0.00000074, 0.00000048, -0.00000130, 0.00000133, -0.00000119, -0.00000220, 0.00000036, -0.00000122, 0.00000539, -0.00001074, 0.00000080, -0.00000091, 0.00000094, 0.00000415, -0.00000233, 0.00000263, -0.00000126, -0.00000354, -0.00000674, 0.00000758, -0.00000669, 0.00001382, -0.00000016, -0.00000542, 0.00000511, -0.00000929, 0.00000376, -0.00001959, -0.00000598, 0.00000520, -0.00001087, -0.00001812, -0.00000251, -0.00006866, -0.00002414, 0.00009360, 0.00004230, 0.00008689, 0.00000372, -0.00010638, -0.00001106, 0.00006256, 0.00002995, 0.00006420, 0.00003483, -0.00006626, -0.00001888, -0.00000759, -0.00001724, 0.00000280, -0.00000034, -0.00001567, -0.00000737, -0.00000219, 0.00000006, 0.00000129, -0.00000904, 0.00001547, -0.00000224, 0.00000158, 0.00000138, -0.00000948, 0.00000204, -0.00000113, 0.00000055, 0.00000135, 0.00000456, -0.00000328, 0.00000298, -0.00000921, -0.00000110, 0.00000446, -0.00000259, 0.00000150, -0.00000395, 0.00000131, -0.00000167, 0.00000000, 0.00000150, 0.00000097, -0.00000141, 0.00000596, 0.00000220, -0.00000466, -0.00000164, -0.00000354, -0.00000070, -0.00000001, 0.00000191, -0.00000072, 0.00000136, 0.00000371, 0.00000198, -0.00000556, -0.00000080, 0.00000120, -0.00000125, -0.00000058, -0.00000165, -0.00000008, -0.00000123, 0.00000337, 0.00000032, },
		{ -0.00000040, -0.00000160, 0.00000279, 0.00000195, -0.00000166, -0.00000043, -0.00000181, -0.00000039, -0.00000035, 0.00000217, -0.00000567, -0.00000134, 0.00000320, 0.00000268, 0.00000025, 0.00000028, 0.00000199, -0.00000071, -0.00000324, -0.00000289, -0.00000144, -0.00000424, 0.00000782, 0.00000144, -0.00000054, 0.00000073, 0.00000036, -0.00000105, 0.00000101, -0.00000040, -0.00000282, 0.00000070, -0.00000182, 0.00000569, -0.00001002, -0.00000033, -0.00000001, 0.00000000, 0.00000498, -0.00000321, 0.00000338, -0.00000174, -0.00000272, -0.00000756, 0.00000762, -0.00000620, 0.00001235, 0.00000196, -0.00000683, 0.00000625, -0.00001039, 0.00000468, -0.00001878, -0.00000787, 0.00000600, -0.00001040, -0.00001858, -0.00000247, -0.00006408, -0.00003699, 0.00009609, 0.00004637, 0.00009058, 0.00000730, -0.00010729, -0.00001263, 0.00005838, 0.00002910, 0.00005878, 0.00004048, -0.00006175, -0.00002264, -0.00000598, -0.00001779, 0.00000201, 0.00000033, -0.00001450, -0.00000896, -0.00000199, 0.00000025, 0.00000159, -0.00000944, 0.00001433, -0.00000039, -0.00000001, 0.00000281, -0.00001024, 0.00000231, -0.00000133, 0.00000051, 0.00000123, 0.00000483, -0.00000315, 0.00000267, -0.00000829, -0.00000244, 0.00000519, -0.00000274, 0.00000168, -0.00000382, 0.00000088, -0.00000148, -0.00000035, 0.00000181, 0.00000098, -0.00000135, 0.00000519, 0.00000315, -0.00000477, -0.00000180, -0.00000345, -0.00000065, -0.00000024, 0.00000192, -0.00000059, 0.00000101, 0.00000373, 0.00000235, -0.00000526, -0.00000129, 0.00000138, -0.00000138, -0.00000041, -0.00000176, 0.00000004, -0.00000142, 0.00000313, 0.00000068, },
		{ -0.00000035, -0.00000171, 0.00000238, 0.00000255, -0.00000182, -0.00000035, -0.00000187, -0.00000037, -0.00000050, 0.00000238, -0.00000531, -0.00000196, 0.00000317, 0.00000261, 0.00000067, -0.00000010, 0.00000226, -0.00000068, -0.00000303, -0.00000331, -0.00000092, -0.00000478, 0.00000742, 0.00000224, -0.00000084, 0.00000073, 0.00000022, -0.00000079, 0.00000065, 0.00000036, -0.00000331, 0.00000086, -0.00000220, 0.00000568, -0.00000900, -0.00000164, 0.00000097, -0.00000098, 0.00000586, -0.00000411, 0.00000405, -0.00000209, -0.00000203, -0.00000815, 0.00000736, -0.00000541, 0.00001070, 0.00000404, -0.00000810, 0.00000719, -0.00001112, 0.00000504, -0.00001738, -0.00001013, 0.00000704, -0.00001031, -0.00001842, -0.00000339, -0.00005826, -0.00004996, 0.00009715, 0.00005129, 0.00009409, 0.00001129, -0.00010851, -0.00001395, 0.00005415, 0.00002848, 0.00005349, 0.00004499, -0.00005664, -0.00002606, -0.00000485, -0.00001790, 0.00000091, 0.00000110, -0.00001342, -0.00001021, -0.00000226, 0.00000080, 0.00000164, -0.00000952, 0.00001284, 0.00000163, -0.00000159, 0.00000413, -0.00001084, 0.00000244, -0.00000138, 0.00000035, 0.00000116, 0.00000504, -0.00000292, 0.00000225, -0.00000731, -0.00000369, 0.00000572, -0.00000268, 0.00000167, -0.00000349, 0.00000034, -0.00000122, -0.00000078, 0.00000216, 0.00000095, -0.00000121, 0.00000438, 0.00000403, -0.00000474, -0.00000203, -0.00000336, -0.00000059, -0.00000044, 0.00000187, -0.00000041, 0.00000067, 0.00000371, 0.00000271, -0.00000492, -0.00000177, 0.00000153, -0.00000149, -0.00000028, -0.00000183, 0.00000010, -0.00000152, 0.00000282, 0.00000103, },
		{ -0.00000029, -0.00000177, 0.00000191, 0.00000313, -0.00000191, -0.00000032, -0.00000187, -0.00000039, -0.00000062, 0.00000250, -0.00000486, -0.00000261, 0.00000314, 0.00000249, 0.00000113, -0.00000048, 0.00000249, -0.00000064, -0.00000278, -0.00000375, -0.00000046, -0.00000516, 0.00000686, 0.00000308, -0.00000109, 0.00000071, 0.00000008, -0.00000054, 0.00000029, 0.00000107, -0.00000362, 0.00000082, -0.00000234, 0.00000537, -0.00000772, -0.00000308, 0.00000199, -0.00000196, 0.00000675, -0.00000498, 0.00000459, -0.00000229, -0.00000152, -0.00000850, 0.00000681, -0.00000439, 0.00000895, 0.00000600, -0.00000914, 0.00000784, -0.00001141, 0.00000477, -0.00001538, -0.00001275, 0.00000832, -0.00001061, -0.00001767, -0.00000518, -0.00005144, -0.00006265, 0.00009646, 0.00005712, 0.00009739, 0.00001575, -0.00011006, -0.00001502, 0.00004987, 0.00002807, 0.00004836, 0.00004851, -0.00005119, -0.00002894, -0.00000433, -0.00001750, -0.00000053, 0.00000201, -0.00001249, -0.00001106, -0.00000302, 0.00000168, 0.00000147, -0.00000930, 0.00001106, 0.00000376, -0.00000310, 0.00000528, -0.00001126, 0.00000241, -0.00000129, 0.00000008, 0.00000110, 0.00000521, -0.00000261, 0.00000177, -0.00000635, -0.00000478, 0.00000599, -0.00000241, 0.00000146, -0.00000299, -0.00000027, -0.00000092, -0.00000126, 0.00000253, 0.00000089, -0.00000100, 0.00000354, 0.00000480, -0.00000458, -0.00000231, -0.00000329, -0.00000051, -0.00000062, 0.00000177, -0.00000020, 0.00000035, 0.00000363, 0.00000306, -0.00000456, -0.00000221, 0.00000163, -0.00000157, -0.00000019, -0.00000184, 0.00000012, -0.00000155, 0.00000247, 0.00000138, },
		{ -0.00000022, -0.00000179, 0.00000139, 0.00000367, -0.00000193, -0.00000035, -0.00000182, -0.00000046, -0.00000070, 0.00000254, -0.00000431, -0.00000327, 0.00000313, 0.00000231, 0.00000161, -0.00000083, 0.00000267, -0.00000058, -0.00000251, -0.00000417, -0.00000008, -0.00000538, 0.00000616, 0.00000394, -0.00000129, 0.00000065, -0.00000005, -0.00000031, -0.00000004, 0.00000168, -0.00000372, 0.00000055, -0.00000224, 0.00000478, -0.00000623, -0.00000458, 0.00000300, -0.00000291, 0.00000760, -0.00000576, 0.00000493, -0.00000228, -0.00000118, -0.00000860, 0.00000600, -0.00000320, 0.00000719, 0.00000776, -0.00000988, 0.00000813, -0.00001117, 0.00000386, -0.00001281, -0.00001566, 0.00000981, -0.00001124, -0.00001639, -0.00000770, -0.00004391, -0.00007462, 0.00009375, 0.00006390, 0.00010049, 0.00002069, -0.00011195, -0.00001581, 0.00004552, 0.00002787, 0.00004337, 0.00005118, -0.00004567, -0.00003108, -0.00000448, -0.00001658, -0.00000231, 0.00000304, -0.00001174, -0.00001149, -0.00000422, 0.00000284, 0.00000115, -0.00000886, 0.00000905, 0.00000590, -0.00000448, 0.00000622, -0.00001146, 0.00000222, -0.00000107, -0.00000025, 0.00000103, 0.00000537, -0.00000224, 0.00000125, -0.00000544, -0.00000567, 0.00000599, -0.00000191, 0.00000107, -0.00000235, -0.00000094, -0.00000060, -0.00000178, 0.00000292, 0.00000081, -0.00000075, 0.00000272, 0.00000544, -0.00000428, -0.00000264, -0.00000324, -0.00000040, -0.00000077, 0.00000163, 0.00000003, 0.00000006, 0.00000349, 0.00000340, -0.00000416, -0.00000262, 0.00000168, -0.00000162, -0.00000015, -0.00000180, 0.00000007, -0.00000152, 0.00000207, 0.00000170, },
		{ -0.00000015, -0.00000175, 0.00000083, 0.00000416, -0.00000186, -0.00000045, -0.00000172, -0.00000056, -0.00000074, 0.00000250, -0.00000368, -0.00000393, 0.00000310, 0.00000210, 0.00000210, -0.00000114, 0.00000279, -0.00000050, -0.00000222, -0.00000455, 0.00000018, -0.00000543, 0.00000533, 0.00000478, -0.00000140, 0.00000056, -0.00000015, -0.00000012, -0.00000034, 0.00000216, -0.00000359, 0.00000006, -0.00000192, 0.00000394, -0.00000458, -0.00000609, 0.00000394, -0.00000378, 0.00000838, -0.00000641, 0.00000505, -0.00000207, -0.00000104, -0.00000848, 0.00000498, -0.00000190, 0.00000551, 0.00000922, -0.00001025, 0.00000800, -0.00001039, 0.00000233, -0.00000976, -0.00001877, 0.00001141, -0.00001213, -0.00001469, -0.00001078, -0.00003603, -0.00008544, 0.00008877, 0.00007162, 0.00010344, 0.00002612, -0.00011419, -0.00001634, 0.00004107, 0.00002789, 0.00003853, 0.00005315, -0.00004032, -0.00003237, -0.00000533, -0.00001516, -0.00000437, 0.00000418, -0.00001117, -0.00001152, -0.00000583, 0.00000419, 0.00000073, -0.00000825, 0.00000691, 0.00000797, -0.00000564, 0.00000691, -0.00001145, 0.00000189, -0.00000074, -0.00000061, 0.00000092, 0.00000552, -0.00000183, 0.00000074, -0.00000462, -0.00000634, 0.00000572, -0.00000121, 0.00000054, -0.00000161, -0.00000161, -0.00000028, -0.00000232, 0.00000329, 0.00000074, -0.00000047, 0.00000193, 0.00000594, -0.00000386, -0.00000300, -0.00000323, -0.00000026, -0.00000088, 0.00000144, 0.00000026, -0.00000018, 0.00000329, 0.00000373, -0.00000375, -0.00000299, 0.00000169, -0.00000163, -0.00000015, -0.00000171, -0.00000002, -0.00000142, 0.00000166, 0.00000199, },
		{ -0.00000006, -0.00000167, 0.00000025, 0.00000458, -0.00000169, -0.00000061, -0.00000157, -0.00000069, -0.00000075, 0.00000238, -0.00000298, -0.00000457, 0.00000306, 0.00000187, 0.00000256, -0.00000138, 0.00000284, -0.00000038, -0.00000194, -0.00000489, 0.00000032, -0.00000531, 0.00000440, 0.00000557, -0.00000142, 0.00000041, -0.00000022, 0.00000001, -0.00000057, 0.00000249, -0.00000322, -0.00000063, -0.00000140, 0.00000292, -0.00000286, -0.00000753, 0.00000474, -0.00000452, 0.00000903, -0.00000686, 0.00000490, -0.00000163, -0.00000107, -0.00000814, 0.00000379, -0.00000056, 0.00000399, 0.00001033, -0.00001021, 0.00000742, -0.00000906, 0.00000023, -0.00000633, -0.00002195, 0.00001302, -0.00001318, -0.00001271, -0.00001417, -0.00002816, -0.00009468, 0.00008134, 0.00008019, 0.00010630, 0.00003203, -0.00011676, -0.00001661, 0.00003653, 0.00002813, 0.00003380, 0.00005458, -0.00003531, -0.00003272, -0.00000685, -0.00001332, -0.00000664, 0.00000538, -0.00001074, -0.00001118, -0.00000774, 0.00000562, 0.00000032, -0.00000754, 0.00000470, 0.00000990, -0.00000654, 0.00000733, -0.00001123, 0.00000143, -0.00000033, -0.00000098, 0.00000075, 0.00000568, -0.00000140, 0.00000027, -0.00000394, -0.00000675, 0.00000518, -0.00000034, -0.00000010, -0.00000081, -0.00000226, 0.00000002, -0.00000286, 0.00000363, 0.00000069, -0.00000019, 0.00000121, 0.00000629, -0.00000332, -0.00000335, -0.00000327, -0.00000010, -0.00000097, 0.00000123, 0.00000048, -0.00000036, 0.00000304, 0.00000404, -0.00000331, -0.00000331, 0.00000165, -0.00000161, -0.00000021, -0.00000156, -0.00000015, -0.00000129, 0.00000123, 0.00000225, },
		{ 0.00000003, -0.00000156, -0.00000033, 0.00000491, -0.00000142, -0.00000083, -0.00000138, -0.00000084, -0.00000072, 0.00000219, -0.00000224, -0.00000517, 0.00000297, 0.00000164, 0.00000299, -0.00000155, 0.00000282, -0.00000022, -0.00000168, -0.00000514, 0.00000032, -0.00000505, 0.00000341, 0.00000627, -0.00000132, 0.00000020, -0.00000025, 0.00000010, -0.00000073, 0.00000265, -0.00000261, -0.00000151, -0.00000074, 0.00000177, -0.00000114, -0.00000883, 0.00000536, -0.00000509, 0.00000952, -0.00000708, 0.00000446, -0.00000098, -0.00000126, -0.00000762, 0.00000248, 0.00000073, 0.00000269, 0.00001105, -0.00000974, 0.00000639, -0.00000722, -0.00000232, -0.00000268, -0.00002506, 0.00001451, -0.00001425, -0.00001065, -0.00001761, -0.00002068, -0.00010197, 0.00007137, 0.00008946, 0.00010920, 0.00003838, -0.00011963, -0.00001665, 0.00003189, 0.00002857, 0.00002917, 0.00005560, -0.00003081, -0.00003212, -0.00000895, -0.00001115, -0.00000904, 0.00000655, -0.00001041, -0.00001055, -0.00000986, 0.00000703, -0.00000001, -0.00000681, 0.00000252, 0.00001159, -0.00000712, 0.00000746, -0.00001080, 0.00000086, 0.00000013, -0.00000132, 0.00000051, 0.00000585, -0.00000096, -0.00000015, -0.00000340, -0.00000691, 0.00000440, 0.00000065, -0.00000080, -0.00000000, -0.00000284, 0.00000027, -0.00000337, 0.00000391, 0.00000068, 0.00000007, 0.00000057, 0.00000648, -0.00000268, -0.00000369, -0.00000337, 0.00000007, -0.00000102, 0.00000099, 0.00000068, -0.00000049, 0.00000272, 0.00000434, -0.00000286, -0.00000358, 0.00000156, -0.00000155, -0.00000031, -0.00000137, -0.00000032, -0.00000112, 0.00000081, 0.00000247, },
		{ 0.00000013, -0.00000141, -0.00000090, 0.00000514, -0.00000104, -0.00000110, -0.00000116, -0.00000101, -0.00000067, 0.00000194, -0.00000146, -0.00000571, 0.00000284, 0.00000143, 0.00000335, -0.00000163, 0.00000271, -0.00000003, -0.00000144, -0.00000532, 0.00000018, -0.00000465, 0.00000238, 0.00000687, -0.00000109, -0.00000006, -0.00000023, 0.00000012, -0.00000082, 0.00000264, -0.00000180, -0.00000251, 0.00000000, 0.00000057, 0.00000051, -0.00000991, 0.00000574, -0.00000545, 0.00000980, -0.00000703, 0.00000373, -0.00000013, -0.00000159, -0.00000696, 0.00000111, 0.00000191, 0.00000168, 0.00001134, -0.00000885, 0.00000494, -0.00000494, -0.00000520, 0.00000102, -0.00002793, 0.00001574, -0.00001517, -0.00000869, -0.00002085, -0.00001395, -0.00010699, 0.00005884, 0.00009919, 0.00011231, 0.00004510, -0.00012274, -0.00001651, 0.00002718, 0.00002919, 0.00002465, 0.00005631, -0.00002691, -0.00003062, -0.00001150, -0.00000880, -0.00001141, 0.00000760, -0.00001010, -0.00000972, -0.00001206, 0.00000829, -0.00000016, -0.00000612, 0.00000044, 0.00001298, -0.00000736, 0.00000730, -0.00001019, 0.00000020, 0.00000062, -0.00000160, 0.00000018, 0.00000602, -0.00000051, -0.00000051, -0.00000303, -0.00000683, 0.00000342, 0.00000172, -0.00000151, 0.00000076, -0.00000332, 0.00000044, -0.00000384, 0.00000412, 0.00000073, 0.00000029, 0.00000003, 0.00000652, -0.00000196, -0.00000398, -0.00000353, 0.00000026, -0.00000103, 0.00000073, 0.00000085, -0.00000055, 0.00000237, 0.00000460, -0.00000239, -0.00000380, 0.00000143, -0.00000146, -0.00000046, -0.00000115, -0.00000051, -0.00000093, 0.00000041, 0.00000264, },
		{ 0.00000024, -0.00000123, -0.00000144, 0.00000526, -0.00000058, -0.00000141, -0.00000092, -0.00000119, -0.00000061, 0.00000164, -0.00000068, -0.00000617, 0.00000264, 0.00000126, 0.00000363, -0.00000160, 0.00000251, 0.00000020, -0.00000123, -0.00000539, -0.00000009, -0.00000414, 0.00000136, 0.00000734, -0.00000074, -0.00000037, -0.00000019, 0.00000010, -0.00000083, 0.00000248, -0.00000080, -0.00000360, 0.00000077, -0.00000060, 0.00000200, -0.00001072, 0.00000584, -0.00000557, 0.00000985, -0.00000670, 0.00000270, 0.00000090, -0.00000202, -0.00000619, -0.00000026, 0.00000293, 0.00000099, 0.00001121, -0.00000757, 0.00000314, -0.00000234, -0.00000825, 0.00000460, -0.00003038, 0.00001655, -0.00001581, -0.00000703, -0.00002363, -0.00000830, -0.00010951, 0.00004388, 0.00010907, 0.00011584, 0.00005210, -0.00012602, -0.00001625, 0.00002245, 0.00002995, 0.00002024, 0.00005677, -0.00002363, -0.00002835, -0.00001431, -0.00000644, -0.00001363, 0.00000843, -0.00000973, -0.00000877, -0.00001422, 0.00000929, -0.00000007, -0.00000553, -0.00000148, 0.00001402, -0.00000723, 0.00000687, -0.00000941, -0.00000052, 0.00000111, -0.00000181, -0.00000023, 0.00000619, -0.00000005, -0.00000079, -0.00000281, -0.00000654, 0.00000229, 0.00000280, -0.00000215, 0.00000143, -0.00000366, 0.00000052, -0.00000425, 0.00000423, 0.00000085, 0.00000045, -0.00000040, 0.00000640, -0.00000118, -0.00000423, -0.00000375, 0.00000044, -0.00000099, 0.00000045, 0.00000099, -0.00000055, 0.00000197, 0.00000482, -0.00000190, -0.00000397, 0.00000126, -0.00000134, -0.00000065, -0.00000089, -0.00000071, -0.00000074, 0.00000004, 0.00000276, },
		{ 0.00000035, -0.00000104, -0.00000194, 0.00000526, -0.00000002, -0.00000174, -0.00000069, -0.00000135, -0.00000054, 0.00000132, 0.00000008, -0.00000653, 0.00000235, 0.00000113, 0.00000382, -0.00000147, 0.00000223, 0.00000047, -0.00000106, -0.00000535, -0.00000049, -0.00000355, 0.00000038, 0.00000765, -0.00000027, -0.00000071, -0.00000012, 0.00000003, -0.00000077, 0.00000217, 0.00000033, -0.00000469, 0.00000148, -0.00000168, 0.00000325, -0.00001120, 0.00000564, -0.00000544, 0.00000965, -0.00000607, 0.00000140, 0.00000207, -0.00000252, -0.00000537, -0.00000158, 0.00000374, 0.00000065, 0.00001067, -0.00000596, 0.00000107, 0.00000046, -0.00001129, 0.00000784, -0.00003225, 0.00001683, -0.00001602, -0.00000585, -0.00002571, -0.00000398, -0.00010940, 0.00002670, 0.00011869, 0.00012001, 0.00005927, -0.00012937, -0.00001594, 0.00001775, 0.00003078, 0.00001599, 0.00005703, -0.00002095, -0.00002545, -0.00001719, -0.00000425, -0.00001556, 0.00000894, -0.00000924, -0.00000781, -0.00001623, 0.00000993, 0.00000033, -0.00000510, -0.00000317, 0.00001468, -0.00000673, 0.00000618, -0.00000851, -0.00000127, 0.00000158, -0.00000194, -0.00000071, 0.00000633, 0.00000042, -0.00000099, -0.00000274, -0.00000607, 0.00000108, 0.00000382, -0.00000269, 0.00000196, -0.00000383, 0.00000050, -0.00000458, 0.00000424, 0.00000105, 0.00000055, -0.00000071, 0.00000615, -0.00000037, -0.00000440, -0.00000403, 0.00000062, -0.00000091, 0.00000016, 0.00000109, -0.00000050, 0.00000155, 0.00000499, -0.00000139, -0.00000409, 0.00000106, -0.00000119, -0.00000086, -0.00000063, -0.00000090, -0.00000056, -0.00000030, 0.00000284, },
		{ 0.00000046, -0.00000083, -0.00000238, 0.00000514, 0.00000061, -0.00000208, -0.00000046, -0.00000149, -0.00000048, 0.00000097, 0.00000079, -0.00000676, 0.00000197, 0.00000107, 0.00000392, -0.00000123, 0.00000188, 0.00000076, -0.00000093, -0.00000521, -0.00000100, -0.00000290, -0.00000053, 0.00000779, 0.00000032, -0.00000108, -0.00000004, -0.00000007, -0.00000066, 0.00000175, 0.00000156, -0.00000574, 0.00000208, -0.00000259, 0.00000422, -0.00001132, 0.00000512, -0.00000505, 0.00000920, -0.00000515, -0.00000015, 0.00000331, -0.00000303, -0.00000454, -0.00000280, 0.00000429, 0.00000066, 0.00000976, -0.00000409, -0.00000117, 0.00000330, -0.00001414, 0.00001057, -0.00003339, 0.00001644, -0.00001568, -0.00000529, -0.00002692, -0.00000120, -0.00010664, 0.00000766, 0.00012759, 0.00012510, 0.00006651, -0.00013269, -0.00001569, 0.00001316, 0.00003161, 0.00001195, 0.00005707, -0.00001882, -0.00002213, -0.00001990, -0.00000242, -0.00001704, 0.00000903, -0.00000853, -0.00000693, -0.00001797, 0.00001013, 0.00000108, -0.00000484, -0.00000461, 0.00001492, -0.00000587, 0.00000527, -0.00000751, -0.00000202, 0.00000200, -0.00000199, -0.00000125, 0.00000643, 0.00000092, -0.00000112, -0.00000279, -0.00000546, -0.00000017, 0.00000473, -0.00000307, 0.00000231, -0.00000382, 0.00000037, -0.00000483, 0.00000413, 0.00000134, 0.00000058, -0.00000091, 0.00000578, 0.00000045, -0.00000448, -0.00000437, 0.00000077, -0.00000077, -0.00000013, 0.00000116, -0.00000039, 0.00000113, 0.00000510, -0.00000085, -0.00000417, 0.00000082, -0.00000103, -0.00000110, -0.00000036, -0.00000108, -0.00000040, -0.00000059, 0.00000286, },
		{ 0.00000058, -0.00000062, -0.00000275, 0.00000489, 0.00000130, -0.00000241, -0.00000026, -0.00000161, -0.00000042, 0.00000063, 0.00000145, -0.00000685, 0.00000150, 0.00000107, 0.00000390, -0.00000088, 0.00000145, 0.00000109, -0.00000084, -0.00000497, -0.00000160, -0.00000224, -0.00000134, 0.00000776, 0.00000100, -0.00000145, 0.00000003, -0.00000019, -0.00000050, 0.00000123, 0.00000281, -0.00000667, 0.00000250, -0.00000327, 0.00000485, -0.00001105, 0.00000429, -0.00000441, 0.00000851, -0.00000397, -0.00000188, 0.00000458, -0.00000350, -0.00000375, -0.00000389, 0.00000457, 0.00000100, 0.00000855, -0.00000206, -0.00000345, 0.00000604, -0.00001660, 0.00001261, -0.00003368, 0.00001530, -0.00001469, -0.00000547, -0.00002712, -0.00000004, -0.00010135, -0.00001280, 0.00013525, 0.00013137, 0.00007368, -0.00013588, -0.00001559, 0.00000875, 0.00003236, 0.00000819, 0.00005687, -0.00001711, -0.00001861, -0.00002222, -0.00000110, -0.00001798, 0.00000863, -0.00000757, -0.00000620, -0.00001936, 0.00000984, 0.00000218, -0.00000476, -0.00000577, 0.00001475, -0.00000470, 0.00000420, -0.00000647, -0.00000274, 0.00000236, -0.00000194, -0.00000182, 0.00000646, 0.00000145, -0.00000120, -0.00000295, -0.00000476, -0.00000137, 0.00000545, -0.00000325, 0.00000247, -0.00000363, 0.00000014, -0.00000498, 0.00000391, 0.00000171, 0.00000054, -0.00000100, 0.00000530, 0.00000126, -0.00000448, -0.00000474, 0.00000088, -0.00000058, -0.00000043, 0.00000119, -0.00000025, 0.00000072, 0.00000513, -0.00000029, -0.00000421, 0.00000056, -0.00000085, -0.00000135, -0.00000010, -0.00000124, -0.00000026, -0.00000083, 0.00000283, },
		{ 0.00000070, -0.00000042, -0.00000305, 0.00000453, 0.00000203, -0.00000270, -0.00000010, -0.00000168, -0.00000039, 0.00000029, 0.00000202, -0.00000680, 0.00000092, 0.00000113, 0.00000379, -0.00000044, 0.00000097, 0.00000142, -0.00000078, -0.00000463, -0.00000227, -0.00000159, -0.00000202, 0.00000755, 0.00000177, -0.00000180, 0.00000009, -0.00000031, -0.00000033, 0.00000065, 0.00000404, -0.00000741, 0.00000269, -0.00000368, 0.00000511, -0.00001040, 0.00000316, -0.00000357, 0.00000762, -0.00000256, -0.00000375, 0.00000580, -0.00000388, -0.00000303, -0.00000479, 0.00000457, 0.00000165, 0.00000711, 0.00000003, -0.00000565, 0.00000852, -0.00001853, 0.00001383, -0.00003304, 0.00001338, -0.00001301, -0.00000645, -0.00002626, -0.00000052, -0.00009373, -0.00003413, 0.00014109, 0.00013908, 0.00008071, -0.00013883, -0.00001574, 0.00000463, 0.00003292, 0.00000480, 0.00005637, -0.00001569, -0.00001515, -0.00002394, -0.00000045, -0.00001826, 0.00000769, -0.00000630, -0.00000569, -0.00002032, 0.00000902, 0.00000364, -0.00000486, -0.00000665, 0.00001418, -0.00000324, 0.00000300, -0.00000543, -0.00000340, 0.00000265, -0.00000181, -0.00000240, 0.00000640, 0.00000202, -0.00000123, -0.00000318, -0.00000403, -0.00000248, 0.00000596, -0.00000320, 0.00000242, -0.00000325, -0.00000019, -0.00000506, 0.00000357, 0.00000216, 0.00000045, -0.00000099, 0.00000475, 0.00000202, -0.00000437, -0.00000515, 0.00000094, -0.00000033, -0.00000073, 0.00000118, -0.00000008, 0.00000033, 0.00000508, 0.00000028, -0.00000420, 0.00000029, -0.00000066, -0.00000160, 0.00000013, -0.00000136, -0.00000016, -0.00000102, 0.00000276, },
		{ 0.00000082, -0.00000023, -0.00000326, 0.00000406, 0.00000278, -0.00000294, 0.00000000, -0.00000170, -0.00000039, -0.00000002, 0.00000250, -0.00000658, 0.00000025, 0.00000126, 0.00000358, 0.00000008, 0.00000046, 0.00000176, -0.00000076, -0.00000422, -0.00000297, -0.00000100, -0.00000256, 0.00000717, 0.00000259, -0.00000210, 0.00000011, -0.00000042, -0.00000015, 0.00000005, 0.00000519, -0.00000792, 0.00000259, -0.00000378, 0.00000501, -0.00000939, 0.00000177, -0.00000254, 0.00000655, -0.00000098, -0.00000567, 0.00000691, -0.00000413, -0.00000242, -0.00000550, 0.00000428, 0.00000257, 0.00000551, 0.00000208, -0.00000765, 0.00001059, -0.00001976, 0.00001414, -0.00003142, 0.00001067, -0.00001066, -0.00000822, -0.00002435, -0.00000255, -0.00008413, -0.00005567, 0.00014457, 0.00014845, 0.00008751, -0.00014146, -0.00001622, 0.00000085, 0.00003321, 0.00000185, 0.00005550, -0.00001441, -0.00001195, -0.00002487, -0.00000057, -0.00001785, 0.00000620, -0.00000472, -0.00000543, -0.00002084, 0.00000768, 0.00000540, -0.00000510, -0.00000726, 0.00001326, -0.00000156, 0.00000175, -0.00000443, -0.00000398, 0.00000286, -0.00000162, -0.00000296, 0.00000622, 0.00000265, -0.00000122, -0.00000345, -0.00000331, -0.00000343, 0.00000620, -0.00000290, 0.00000216, -0.00000271, -0.00000059, -0.00000506, 0.00000312, 0.00000267, 0.00000030, -0.00000090, 0.00000415, 0.00000273, -0.00000417, -0.00000557, 0.00000093, -0.00000001, -0.00000103, 0.00000113, 0.00000011, -0.00000002, 0.00000495, 0.00000087, -0.00000415, 0.00000000, -0.00000048, -0.00000185, 0.00000034, -0.00000143, -0.00000011, -0.00000115, 0.00000265, },
		{ 0.00000094, -0.00000005, -0.00000338, 0.00000349, 0.00000353, -0.00000311, 0.00000004, -0.00000167, -0.00000042, -0.00000029, 0.00000286, -0.00000621, -0.00000051, 0.00000144, 0.00000330, 0.00000065, -0.00000006, 0.00000208, -0.00000075, -0.00000374, -0.00000368, -0.00000049, -0.00000292, 0.00000663, 0.00000343, -0.00000233, 0.00000007, -0.00000050, 0.00000001, -0.00000054, 0.00000622, -0.00000814, 0.00000220, -0.00000358, 0.00000454, -0.00000806, 0.00000019, -0.00000140, 0.00000537, 0.00000071, -0.00000756, 0.00000783, -0.00000419, -0.00000196, -0.00000598, 0.00000372, 0.00000369, 0.00000386, 0.00000397, -0.00000932, 0.00001213, -0.00002021, 0.00001347, -0.00002886, 0.00000724, -0.00000768, -0.00001073, -0.00002148, -0.00000594, -0.00007299, -0.00007674, 0.00014510, 0.00015965, 0.00009407, -0.00014371, -0.00001711, -0.00000249, 0.00003314, -0.00000057, 0.00005419, -0.00001311, -0.00000924, -0.00002487, -0.00000151, -0.00001673, 0.00000417, -0.00000286, -0.00000541, -0.00002090, 0.00000586, 0.00000741, -0.00000544, -0.00000765, 0.00001202, 0.00000027, 0.00000050, -0.00000352, -0.00000446, 0.00000299, -0.00000137, -0.00000346, 0.00000590, 0.00000333, -0.00000119, -0.00000374, -0.00000264, -0.00000418, 0.00000615, -0.00000236, 0.00000171, -0.00000203, -0.00000105, -0.00000500, 0.00000258, 0.00000323, 0.00000012, -0.00000075, 0.00000352, 0.00000335, -0.00000386, -0.00000599, 0.00000084, 0.00000036, -0.00000132, 0.00000106, 0.00000030, -0.00000031, 0.00000473, 0.00000147, -0.00000405, -0.00000030, -0.00000030, -0.00000208, 0.00000050, -0.00000146, -0.00000009, -0.00000123, 0.00000251, },
		{ 0.00000105, 0.00000010, -0.00000342, 0.00000284, 0.00000425, -0.00000320, 0.00000001, -0.00000159, -0.00000048, -0.00000053, 0.00000311, -0.00000569, -0.00000133, 0.00000165, 0.00000295, 0.00000126, -0.00000057, 0.00000237, -0.00000075, -0.00000322, -0.00000437, -0.00000009, -0.00000311, 0.00000594, 0.00000427, -0.00000247, -0.00000003, -0.00000053, 0.00000014, -0.00000110, 0.00000707, -0.00000805, 0.00000150, -0.00000309, 0.00000376, -0.00000647, -0.00000151, -0.00000021, 0.00000414, 0.00000244, -0.00000936, 0.00000850, -0.00000404, -0.00000166, -0.00000624, 0.00000291, 0.00000496, 0.00000224, 0.00000560, -0.00001057, 0.00001304, -0.00001981, 0.00001184, -0.00002542, 0.00000319, -0.00000419, -0.00001386, -0.00001781, -0.00001042, -0.00006082, -0.00009662, 0.00014220, 0.00017278, 0.00010042, -0.00014558, -0.00001845, -0.00000536, 0.00003264, -0.00000240, 0.00005239, -0.00001167, -0.00000718, -0.00002385, -0.00000329, -0.00001494, 0.00000166, -0.00000075, -0.00000562, -0.00002053, 0.00000363, 0.00000957, -0.00000580, -0.00000785, 0.00001054, 0.00000218, -0.00000067, -0.00000272, -0.00000482, 0.00000304, -0.00000109, -0.00000389, 0.00000543, 0.00000408, -0.00000114, -0.00000403, -0.00000206, -0.00000470, 0.00000581, -0.00000160, 0.00000110, -0.00000124, -0.00000153, -0.00000490, 0.00000196, 0.00000382, -0.00000008, -0.00000056, 0.00000289, 0.00000388, -0.00000347, -0.00000639, 0.00000067, 0.00000078, -0.00000158, 0.00000096, 0.00000049, -0.00000055, 0.00000442, 0.00000207, -0.00000390, -0.00000060, -0.00000014, -0.00000228, 0.00000062, -0.00000143, -0.00000012, -0.00000127, 0.00000233, },
		{ 0.00000116, 0.00000022, -0.00000337, 0.00000213, 0.00000491, -0.00000318, -0.00000009, -0.00000145, -0.00000057, -0.00000071, 0.00000323, -0.00000503, -0.00000220, 0.00000188, 0.00000256, 0.00000187, -0.00000105, 0.00000260, -0.00000074, -0.00000268, -0.00000500, 0.00000017, -0.00000312, 0.00000514, 0.00000507, -0.00000250, -0.00000020, -0.00000053, 0.00000023, -0.00000159, 0.00000772, -0.00000764, 0.00000051, -0.00000232, 0.00000270, -0.00000469, -0.00000327, 0.00000096, 0.00000291, 0.00000413, -0.00001096, 0.00000887, -0.00000364, -0.00000153, -0.00000627, 0.00000189, 0.00000631, 0.00000074, 0.00000690, -0.00001133, 0.00001328, -0.00001854, 0.00000930, -0.00002123, -0.00000131, -0.00000035, -0.00001743, -0.00001357, -0.00001564, -0.00004821, -0.00011457, 0.00013543, 0.00018782, 0.00010668, -0.00014707, -0.00002025, -0.00000772, 0.00003167, -0.00000360, 0.00005005, -0.00000998, -0.00000590, -0.00002180, -0.00000583, -0.00001258, -0.00000121, 0.00000149, -0.00000599, -0.00001980, 0.00000108, 0.00001178, -0.00000612, -0.00000793, 0.00000887, 0.00000409, -0.00000171, -0.00000209, -0.00000506, 0.00000302, -0.00000081, -0.00000422, 0.00000480, 0.00000486, -0.00000107, -0.00000428, -0.00000160, -0.00000498, 0.00000518, -0.00000063, 0.00000036, -0.00000040, -0.00000199, -0.00000478, 0.00000130, 0.00000441, -0.00000027, -0.00000035, 0.00000228, 0.00000430, -0.00000299, -0.00000675, 0.00000041, 0.00000124, -0.00000183, 0.00000083, 0.00000067, -0.00000071, 0.00000403, 0.00000265, -0.00000370, -0.00000091, 0.00000001, -0.00000246, 0.00000068, -0.00000135, -0.00000019, -0.00000126, 0.00000214, },
		{ 0.00000127, 0.00000031, -0.00000324, 0.00000138, 0.00000551, -0.00000306, -0.00000027, -0.00000126, -0.00000069, -0.00000086, 0.00000324, -0.00000425, -0.00000309, 0.00000209, 0.00000217, 0.00000246, -0.00000147, 0.00000277, -0.00000072, -0.00000215, -0.00000555, 0.00000027, -0.00000296, 0.00000424, 0.00000580, -0.00000241, -0.00000045, -0.00000048, 0.00000028, -0.00000201, 0.00000815, -0.00000690, -0.00000074, -0.00000135, 0.00000144, -0.00000281, -0.00000498, 0.00000204, 0.00000177, 0.00000570, -0.00001231, 0.00000887, -0.00000298, -0.00000158, -0.00000610, 0.00000070, 0.00000766, -0.00000057, 0.00000779, -0.00001155, 0.00001281, -0.00001644, 0.00000597, -0.00001648, -0.00000606, 0.00000364, -0.00002123, -0.00000904, -0.00002122, -0.00003574, -0.00012992, 0.00012448, 0.00020464, 0.00011303, -0.00014828, -0.00002247, -0.00000960, 0.00003021, -0.00000414, 0.00004716, -0.00000796, -0.00000545, -0.00001874, -0.00000904, -0.00000978, -0.00000431, 0.00000378, -0.00000644, -0.00001880, -0.00000167, 0.00001392, -0.00000630, -0.00000793, 0.00000710, 0.00000590, -0.00000256, -0.00000165, -0.00000516, 0.00000294, -0.00000052, -0.00000442, 0.00000402, 0.00000568, -0.00000097, -0.00000450, -0.00000128, -0.00000500, 0.00000430, 0.00000047, -0.00000043, 0.00000044, -0.00000240, -0.00000466, 0.00000060, 0.00000497, -0.00000043, -0.00000014, 0.00000172, 0.00000460, -0.00000243, -0.00000705, 0.00000005, 0.00000174, -0.00000204, 0.00000067, 0.00000083, -0.00000081, 0.00000357, 0.00000321, -0.00000344, -0.00000122, 0.00000014, -0.00000260, 0.00000068, -0.00000122, -0.00000030, -0.00000122, 0.00000192, },
		{ 0.00000138, 0.00000037, -0.00000304, 0.00000060, 0.00000601, -0.00000282, -0.00000051, -0.00000103, -0.00000084, -0.00000095, 0.00000313, -0.00000337, -0.00000397, 0.00000227, 0.00000179, 0.00000300, -0.00000180, 0.00000286, -0.00000066, -0.00000164, -0.00000598, 0.00000020, -0.00000263, 0.00000328, 0.00000644, -0.00000218, -0.00000077, -0.00000038, 0.00000028, -0.00000235, 0.00000835, -0.00000585, -0.00000221, -0.00000023, 0.00000007, -0.00000093, -0.00000656, 0.00000294, 0.00000078, 0.00000709, -0.00001332, 0.00000847, -0.00000206, -0.00000180, -0.00000573, -0.00000061, 0.00000897, -0.00000161, 0.00000822, -0.00001121, 0.00001165, -0.00001360, 0.00000201, -0.00001139, -0.00001083, 0.00000756, -0.00002501, -0.00000451, -0.00002672, -0.00002404, -0.00014207, 0.00010915, 0.00022298, 0.00011975, -0.00014934, -0.00002506, -0.00001104, 0.00002829, -0.00000407, 0.00004374, -0.00000559, -0.00000584, -0.00001480, -0.00001274, -0.00000671, -0.00000749, 0.00000597, -0.00000688, -0.00001762, -0.00000449, 0.00001588, -0.00000628, -0.00000792, 0.00000529, 0.00000756, -0.00000317, -0.00000141, -0.00000514, 0.00000281, -0.00000026, -0.00000450, 0.00000310, 0.00000650, -0.00000084, -0.00000468, -0.00000109, -0.00000479, 0.00000319, 0.00000167, -0.00000124, 0.00000124, -0.00000272, -0.00000457, -0.00000010, 0.00000549, -0.00000054, 0.00000004, 0.00000123, 0.00000478, -0.00000183, -0.00000730, -0.00000040, 0.00000226, -0.00000220, 0.00000049, 0.00000096, -0.00000083, 0.00000306, 0.00000372, -0.00000312, -0.00000152, 0.00000025, -0.00000271, 0.00000063, -0.00000104, -0.00000044, -0.00000115, 0.00000170, },
		{ 0.00000148, 0.00000039, -0.00000278, -0.00000017, 0.00000639, -0.00000246, -0.00000082, -0.00000077, -0.00000100, -0.00000101, 0.00000293, -0.00000242, -0.00000481, 0.00000238, 0.00000145, 0.00000347, -0.00000204, 0.00000285, -0.00000056, -0.00000117, -0.00000627, -0.00000004, -0.00000216, 0.00000229, 0.00000695, -0.00000181, -0.00000114, -0.00000026, 0.00000023, -0.00000259, 0.00000832, -0.00000454, -0.00000382, 0.00000096, -0.00000132, 0.00000086, -0.00000792, 0.00000361, 0.00000000, 0.00000823, -0.00001394, 0.00000766, -0.00000089, -0.00000216, -0.00000520, -0.00000201, 0.00001016, -0.00000234, 0.00000819, -0.00001033, 0.00000988, -0.00001014, -0.00000237, -0.00000621, -0.00001535, 0.00001114, -0.00002850, -0.00000032, -0.00003174, -0.00001365, -0.00015050, 0.00008944, 0.00024244, 0.00012720, -0.00015042, -0.00002789, -0.00001212, 0.00002597, -0.00000342, 0.00003985, -0.00000290, -0.00000701, -0.00001014, -0.00001671, -0.00000358, -0.00001055, 0.00000792, -0.00000720, -0.00001638, -0.00000724, 0.00001751, -0.00000597, -0.00000796, 0.00000353, 0.00000899, -0.00000349, -0.00000139, -0.00000499, 0.00000264, -0.00000004, -0.00000445, 0.00000207, 0.00000729, -0.00000065, -0.00000482, -0.00000104, -0.00000436, 0.00000192, 0.00000290, -0.00000199, 0.00000193, -0.00000291, -0.00000452, -0.00000078, 0.00000593, -0.00000059, 0.00000017, 0.00000082, 0.00000484, -0.00000117, -0.00000746, -0.00000093, 0.00000277, -0.00000231, 0.00000028, 0.00000107, -0.00000079, 0.00000251, 0.00000417, -0.00000273, -0.00000182, 0.00000034, -0.00000277, 0.00000052, -0.00000082, -0.00000060, -0.00000106, 0.00000148, },
		{ 0.00000157, 0.00000038, -0.00000246, -0.00000092, 0.00000665, -0.00000199, -0.00000117, -0.00000049, -0.00000116, -0.00000104, 0.00000264, -0.00000144, -0.00000557, 0.00000240, 0.00000117, 0.00000384, -0.00000215, 0.00000274, -0.00000040, -0.00000077, -0.00000642, -0.00000046, -0.00000157, 0.00000131, 0.00000733, -0.00000131, -0.00000157, -0.00000011, 0.00000016, -0.00000275, 0.00000808, -0.00000301, -0.00000551, 0.00000213, -0.00000265, 0.00000246, -0.00000898, 0.00000397, -0.00000052, 0.00000907, -0.00001413, 0.00000643, 0.00000051, -0.00000265, -0.00000454, -0.00000344, 0.00001118, -0.00000273, 0.00000770, -0.00000896, 0.00000759, -0.00000625, -0.00000693, -0.00000122, -0.00001936, 0.00001416, -0.00003144, 0.00000323, -0.00003585, -0.00000508, -0.00015486, 0.00006549, 0.00026246, 0.00013583, -0.00015175, -0.00003082, -0.00001295, 0.00002333, -0.00000229, 0.00003559, 0.00000005, -0.00000883, -0.00000497, -0.00002071, -0.00000062, -0.00001330, 0.00000949, -0.00000728, -0.00001521, -0.00000979, 0.00001873, -0.00000532, -0.00000809, 0.00000188, 0.00001012, -0.00000350, -0.00000160, -0.00000472, 0.00000244, 0.00000015, -0.00000428, 0.00000095, 0.00000802, -0.00000039, -0.00000493, -0.00000111, -0.00000376, 0.00000055, 0.00000407, -0.00000262, 0.00000248, -0.00000295, -0.00000454, -0.00000141, 0.00000627, -0.00000055, 0.00000025, 0.00000050, 0.00000478, -0.00000050, -0.00000753, -0.00000153, 0.00000326, -0.00000235, 0.00000005, 0.00000115, -0.00000070, 0.00000193, 0.00000454, -0.00000228, -0.00000212, 0.00000039, -0.00000279, 0.00000036, -0.00000057, -0.00000076, -0.00000096, 0.00000126, },
		{ 0.00000166, 0.00000035, -0.00000211, -0.00000163, 0.00000677, -0.00000141, -0.00000155, -0.00000021, -0.00000131, -0.00000104, 0.00000229, -0.00000044, -0.00000622, 0.00000231, 0.00000098, 0.00000409, -0.00000213, 0.00000253, -0.00000019, -0.00000044, -0.00000641, -0.00000104, -0.00000089, 0.00000036, 0.00000754, -0.00000068, -0.00000202, 0.00000004, 0.00000007, -0.00000285, 0.00000766, -0.00000132, -0.00000718, 0.00000320, -0.00000382, 0.00000378, -0.00000967, 0.00000398, -0.00000074, 0.00000956, -0.00001388, 0.00000480, 0.00000209, -0.00000322, -0.00000379, -0.00000485, 0.00001200, -0.00000276, 0.00000677, -0.00000717, 0.00000491, -0.00000211, -0.00001141, 0.00000331, -0.00002260, 0.00001638, -0.00003359, 0.00000584, -0.00003873, 0.00000126, -0.00015493, 0.00003766, 0.00028236, 0.00014615, -0.00015360, -0.00003369, -0.00001367, 0.00002049, -0.00000082, 0.00003108, 0.00000314, -0.00001114, 0.00000044, -0.00002448, 0.00000195, -0.00001556, 0.00001053, -0.00000702, -0.00001421, -0.00001201, 0.00001943, -0.00000428, -0.00000835, 0.00000038, 0.00001092, -0.00000318, -0.00000201, -0.00000436, 0.00000222, 0.00000029, -0.00000400, -0.00000021, 0.00000864, -0.00000003, -0.00000502, -0.00000128, -0.00000303, -0.00000085, 0.00000512, -0.00000308, 0.00000284, -0.00000282, -0.00000464, -0.00000199, 0.00000650, -0.00000041, 0.00000026, 0.00000028, 0.00000461, 0.00000019, -0.00000751, -0.00000219, 0.00000371, -0.00000232, -0.00000021, 0.00000120, -0.00000056, 0.00000136, 0.00000482, -0.00000175, -0.00000240, 0.00000042, -0.00000277, 0.00000015, -0.00000030, -0.00000093, -0.00000087, 0.00000106, },
		{ 0.00000173, 0.00000028, -0.00000173, -0.00000228, 0.00000674, -0.00000074, -0.00000195, 0.00000005, -0.00000144, -0.00000103, 0.00000189, 0.00000052, -0.00000672, 0.00000208, 0.00000089, 0.00000421, -0.00000199, 0.00000221, 0.00000008, -0.00000019, -0.00000623, -0.00000176, -0.00000015, -0.00000051, 0.00000758, 0.00000005, -0.00000247, 0.00000016, -0.00000002, -0.00000288, 0.00000709, 0.00000045, -0.00000875, 0.00000407, -0.00000475, 0.00000476, -0.00000995, 0.00000361, -0.00000064, 0.00000970, -0.00001318, 0.00000281, 0.00000380, -0.00000383, -0.00000298, -0.00000621, 0.00001258, -0.00000244, 0.00000547, -0.00000508, 0.00000199, 0.00000205, -0.00001556, 0.00000711, -0.00002486, 0.00001759, -0.00003475, 0.00000729, -0.00004007, 0.00000508, -0.00015070, 0.00000649, 0.00030132, 0.00015871, -0.00015623, -0.00003630, -0.00001444, 0.00001760, 0.00000086, 0.00002649, 0.00000623, -0.00001372, 0.00000580, -0.00002773, 0.00000391, -0.00001716, 0.00001092, -0.00000632, -0.00001348, -0.00001380, 0.00001955, -0.00000284, -0.00000875, -0.00000092, 0.00001136, -0.00000254, -0.00000262, -0.00000392, 0.00000200, 0.00000038, -0.00000364, -0.00000137, 0.00000910, 0.00000044, -0.00000512, -0.00000151, -0.00000222, -0.00000222, 0.00000599, -0.00000331, 0.00000298, -0.00000251, -0.00000481, -0.00000249, 0.00000660, -0.00000017, 0.00000019, 0.00000016, 0.00000434, 0.00000087, -0.00000740, -0.00000290, 0.00000410, -0.00000219, -0.00000050, 0.00000124, -0.00000038, 0.00000080, 0.00000499, -0.00000116, -0.00000266, 0.00000042, -0.00000271, -0.00000009, -0.00000002, -0.00000109, -0.00000078, 0.00000086, },
		{ 0.00000180, 0.00000020, -0.00000135, -0.00000285, 0.00000657, 0.00000001, -0.00000233, 0.00000029, -0.00000154, -0.00000102, 0.00000147, 0.00000143, -0.00000706, 0.00000171, 0.00000091, 0.00000420, -0.00000171, 0.00000180, 0.00000040, -0.00000003, -0.00000590, -0.00000261, 0.00000062, -0.00000128, 0.00000745, 0.00000087, -0.00000289, 0.00000025, -0.00000010, -0.00000289, 0.00000642, 0.00000224, -0.00001013, 0.00000467, -0.00000537, 0.00000533, -0.00000978, 0.00000287, -0.00000021, 0.00000947, -0.00001205, 0.00000051, 0.00000557, -0.00000443, -0.00000216, -0.00000748, 0.00001291, -0.00000180, 0.00000386, -0.00000279, -0.00000100, 0.00000602, -0.00001911, 0.00000994, -0.00002593, 0.00001767, -0.00003477, 0.00000741, -0.00003970, 0.00000624, -0.00014231, -0.00002730, 0.00031840, 0.00017411, -0.00015992, -0.00003848, -0.00001542, 0.00001479, 0.00000258, 0.00002200, 0.00000913, -0.00001635, 0.00001082, -0.00003021, 0.00000506, -0.00001795, 0.00001058, -0.00000513, -0.00001311, -0.00001508, 0.00001905, -0.00000098, -0.00000930, -0.00000200, 0.00001142, -0.00000159, -0.00000339, -0.00000342, 0.00000178, 0.00000044, -0.00000322, -0.00000247, 0.00000938, 0.00000106, -0.00000523, -0.00000178, -0.00000139, -0.00000348, 0.00000661, -0.00000328, 0.00000288, -0.00000201, -0.00000506, -0.00000291, 0.00000656, 0.00000018, 0.00000005, 0.00000013, 0.00000398, 0.00000153, -0.00000719, -0.00000361, 0.00000440, -0.00000198, -0.00000082, 0.00000126, -0.00000019, 0.00000029, 0.00000505, -0.00000050, -0.00000291, 0.00000038, -0.00000262, -0.00000036, 0.00000026, -0.00000123, -0.00000070, 0.00000069, },
		{ 0.00000186, 0.00000010, -0.00000096, -0.00000332, 0.00000626, 0.00000082, -0.00000269, 0.00000048, -0.00000160, -0.00000102, 0.00000104, 0.00000225, -0.00000721, 0.00000118, 0.00000104, 0.00000406, -0.00000131, 0.00000130, 0.00000077, 0.00000004, -0.00000542, -0.00000355, 0.00000137, -0.00000193, 0.00000714, 0.00000175, -0.00000325, 0.00000027, -0.00000013, -0.00000288, 0.00000567, 0.00000398, -0.00001125, 0.00000494, -0.00000562, 0.00000547, -0.00000916, 0.00000175, 0.00000051, 0.00000891, -0.00001053, -0.00000202, 0.00000734, -0.00000497, -0.00000136, -0.00000862, 0.00001299, -0.00000087, 0.00000205, -0.00000045, -0.00000387, 0.00000956, -0.00002186, 0.00001163, -0.00002571, 0.00001652, -0.00003357, 0.00000611, -0.00003754, 0.00000474, -0.00013013, -0.00006281, 0.00033262, 0.00019292, -0.00016491, -0.00004006, -0.00001676, 0.00001221, 0.00000417, 0.00001777, 0.00001167, -0.00001879, 0.00001523, -0.00003170, 0.00000525, -0.00001785, 0.00000946, -0.00000341, -0.00001312, -0.00001581, 0.00001792, 0.00000124, -0.00000996, -0.00000286, 0.00001111, -0.00000036, -0.00000429, -0.00000289, 0.00000157, 0.00000047, -0.00000278, -0.00000347, 0.00000941, 0.00000183, -0.00000537, -0.00000206, -0.00000058, -0.00000457, 0.00000693, -0.00000296, 0.00000254, -0.00000134, -0.00000538, -0.00000325, 0.00000638, 0.00000063, -0.00000017, 0.00000019, 0.00000357, 0.00000214, -0.00000690, -0.00000433, 0.00000460, -0.00000166, -0.00000116, 0.00000126, 0.00000001, -0.00000016, 0.00000497, 0.00000022, -0.00000312, 0.00000031, -0.00000249, -0.00000066, 0.00000053, -0.00000134, -0.00000065, 0.00000054, },
		{ 0.00000191, -0.00000001, -0.00000060, -0.00000369, 0.00000580, 0.00000166, -0.00000299, 0.00000061, -0.00000161, -0.00000104, 0.00000062, 0.00000295, -0.00000716, 0.00000050, 0.00000128, 0.00000380, -0.00000080, 0.00000074, 0.00000117, 0.00000004, -0.00000481, -0.00000455, 0.00000207, -0.00000242, 0.00000667, 0.00000266, -0.00000353, 0.00000021, -0.00000011, -0.00000288, 0.00000490, 0.00000559, -0.00001204, 0.00000484, -0.00000549, 0.00000517, -0.00000811, 0.00000030, 0.00000150, 0.00000804, -0.00000869, -0.00000469, 0.00000901, -0.00000540, -0.00000062, -0.00000960, 0.00001280, 0.00000030, 0.00000012, 0.00000181, -0.00000647, 0.00001250, -0.00002360, 0.00001204, -0.00002412, 0.00001414, -0.00003115, 0.00000339, -0.00003362, 0.00000073, -0.00011469, -0.00009902, 0.00034294, 0.00021567, -0.00017139, -0.00004093, -0.00001857, 0.00000999, 0.00000549, 0.00001398, 0.00001367, -0.00002080, 0.00001878, -0.00003202, 0.00000439, -0.00001680, 0.00000755, -0.00000118, -0.00001353, -0.00001597, 0.00001620, 0.00000377, -0.00001072, -0.00000350, 0.00001045, 0.00000111, -0.00000527, -0.00000235, 0.00000139, 0.00000048, -0.00000235, -0.00000433, 0.00000919, 0.00000276, -0.00000555, -0.00000232, 0.00000016, -0.00000544, 0.00000693, -0.00000234, 0.00000197, -0.00000052, -0.00000575, -0.00000352, 0.00000607, 0.00000117, -0.00000044, 0.00000032, 0.00000310, 0.00000271, -0.00000652, -0.00000502, 0.00000467, -0.00000125, -0.00000152, 0.00000125, 0.00000020, -0.00000054, 0.00000477, 0.00000099, -0.00000330, 0.00000021, -0.00000233, -0.00000096, 0.00000077, -0.00000141, -0.00000063, 0.00000042, },
		{ 0.00000194, -0.00000012, -0.00000027, -0.00000395, 0.00000523, 0.00000251, -0.00000322, 0.00000068, -0.00000157, -0.00000108, 0.00000023, 0.00000352, -0.00000690, -0.00000031, 0.00000160, 0.00000343, -0.00000022, 0.00000013, 0.00000158, -0.00000002, -0.00000410, -0.00000556, 0.00000268, -0.00000275, 0.00000606, 0.00000356, -0.00000370, 0.00000005, -0.00000003, -0.00000290, 0.00000415, 0.00000703, -0.00001246, 0.00000432, -0.00000496, 0.00000445, -0.00000667, -0.00000142, 0.00000269, 0.00000694, -0.00000662, -0.00000740, 0.00001050, -0.00000567, 0.00000003, -0.00001041, 0.00001237, 0.00000165, -0.00000184, 0.00000388, -0.00000865, 0.00001467, -0.00002422, 0.00001111, -0.00002119, 0.00001059, -0.00002760, -0.00000063, -0.00002810, -0.00000550, -0.00009666, -0.00013478, 0.00034832, 0.00024284, -0.00017948, -0.00004100, -0.00002095, 0.00000824, 0.00000640, 0.00001079, 0.00001498, -0.00002220, 0.00002128, -0.00003106, 0.00000243, -0.00001482, 0.00000490, 0.00000151, -0.00001431, -0.00001561, 0.00001396, 0.00000652, -0.00001149, -0.00000396, 0.00000948, 0.00000276, -0.00000628, -0.00000185, 0.00000123, 0.00000049, -0.00000195, -0.00000500, 0.00000869, 0.00000384, -0.00000577, -0.00000253, 0.00000078, -0.00000605, 0.00000658, -0.00000145, 0.00000120, 0.00000042, -0.00000615, -0.00000371, 0.00000564, 0.00000177, -0.00000076, 0.00000051, 0.00000261, 0.00000321, -0.00000607, -0.00000566, 0.00000461, -0.00000074, -0.00000190, 0.00000123, 0.00000037, -0.00000084, 0.00000445, 0.00000178, -0.00000343, 0.00000006, -0.00000215, -0.00000126, 0.00000097, -0.00000144, -0.00000063, 0.00000031, },
		{ 0.00000197, -0.00000023, 0.00000001, -0.00000409, 0.00000454, 0.00000334, -0.00000336, 0.00000066, -0.00000147, -0.00000114, -0.00000012, 0.00000395, -0.00000644, -0.00000126, 0.00000199, 0.00000299, 0.00000042, -0.00000048, 0.00000199, -0.00000012, -0.00000331, -0.00000656, 0.00000316, -0.00000290, 0.00000531, 0.00000444, -0.00000373, -0.00000022, 0.00000011, -0.00000296, 0.00000344, 0.00000823, -0.00001246, 0.00000341, -0.00000406, 0.00000334, -0.00000491, -0.00000333, 0.00000401, 0.00000567, -0.00000439, -0.00001005, 0.00001173, -0.00000573, 0.00000055, -0.00001103, 0.00001171, 0.00000312, -0.00000371, 0.00000564, -0.00001028, 0.00001595, -0.00002364, 0.00000885, -0.00001702, 0.00000603, -0.00002308, -0.00000577, -0.00002124, -0.00001350, -0.00007689, -0.00016889, 0.00034778, 0.00027475, -0.00018917, -0.00004029, -0.00002392, 0.00000699, 0.00000681, 0.00000830, 0.00001549, -0.00002285, 0.00002258, -0.00002877, -0.00000057, -0.00001198, 0.00000161, 0.00000458, -0.00001539, -0.00001479, 0.00001130, 0.00000938, -0.00001222, -0.00000427, 0.00000825, 0.00000451, -0.00000726, -0.00000142, 0.00000111, 0.00000049, -0.00000162, -0.00000546, 0.00000790, 0.00000507, -0.00000602, -0.00000270, 0.00000127, -0.00000639, 0.00000589, -0.00000030, 0.00000028, 0.00000143, -0.00000653, -0.00000386, 0.00000511, 0.00000243, -0.00000109, 0.00000074, 0.00000211, 0.00000364, -0.00000556, -0.00000623, 0.00000441, -0.00000014, -0.00000228, 0.00000121, 0.00000052, -0.00000104, 0.00000400, 0.00000259, -0.00000349, -0.00000012, -0.00000194, -0.00000156, 0.00000114, -0.00000143, -0.00000066, 0.00000024, },
		{ 0.00000199, -0.00000033, 0.00000025, -0.00000412, 0.00000376, 0.00000413, -0.00000340, 0.00000057, -0.00000131, -0.00000123, -0.00000041, 0.00000423, -0.00000578, -0.00000229, 0.00000243, 0.00000250, 0.00000108, -0.00000108, 0.00000237, -0.00000025, -0.00000248, -0.00000748, 0.00000349, -0.00000286, 0.00000444, 0.00000525, -0.00000362, -0.00000059, 0.00000033, -0.00000306, 0.00000279, 0.00000917, -0.00001205, 0.00000212, -0.00000283, 0.00000193, -0.00000292, -0.00000535, 0.00000537, 0.00000433, -0.00000212, -0.00001252, 0.00001262, -0.00000553, 0.00000092, -0.00001144, 0.00001084, 0.00000466, -0.00000543, 0.00000699, -0.00001128, 0.00001629, -0.00002188, 0.00000537, -0.00001179, 0.00000068, -0.00001785, -0.00001174, -0.00001342, -0.00002272, -0.00005629, -0.00020013, 0.00034044, 0.00031161, -0.00020035, -0.00003891, -0.00002743, 0.00000627, 0.00000667, 0.00000658, 0.00001513, -0.00002264, 0.00002264, -0.00002521, -0.00000453, -0.00000843, -0.00000217, 0.00000788, -0.00001666, -0.00001361, 0.00000835, 0.00001221, -0.00001281, -0.00000450, 0.00000684, 0.00000630, -0.00000816, -0.00000107, 0.00000105, 0.00000051, -0.00000138, -0.00000569, 0.00000684, 0.00000641, -0.00000629, -0.00000281, 0.00000161, -0.00000644, 0.00000489, 0.00000104, -0.00000075, 0.00000245, -0.00000687, -0.00000398, 0.00000450, 0.00000311, -0.00000143, 0.00000097, 0.00000163, 0.00000398, -0.00000500, -0.00000671, 0.00000407, 0.00000053, -0.00000265, 0.00000117, 0.00000063, -0.00000115, 0.00000345, 0.00000339, -0.00000349, -0.00000034, -0.00000173, -0.00000184, 0.00000125, -0.00000136, -0.00000072, 0.00000018, },
		{ 0.00000200, -0.00000041, 0.00000042, -0.00000403, 0.00000292, 0.00000485, -0.00000332, 0.00000040, -0.00000111, -0.00000134, -0.00000066, 0.00000435, -0.00000496, -0.00000338, 0.00000286, 0.00000200, 0.00000172, -0.00000164, 0.00000269, -0.00000038, -0.00000165, -0.00000830, 0.00000363, -0.00000264, 0.00000349, 0.00000597, -0.00000335, -0.00000106, 0.00000060, -0.00000320, 0.00000222, 0.00000984, -0.00001123, 0.00000050, -0.00000136, 0.00000029, -0.00000080, -0.00000735, 0.00000668, 0.00000300, 0.00000008, -0.00001470, 0.00001309, -0.00000505, 0.00000112, -0.00001164, 0.00000977, 0.00000621, -0.00000693, 0.00000788, -0.00001160, 0.00001567, -0.00001900, 0.00000083, -0.00000576, -0.00000516, -0.00001221, -0.00001818, -0.00000508, -0.00003251, -0.00003587, -0.00022730, 0.00032556, 0.00035343, -0.00021274, -0.00003704, -0.00003137, 0.00000599, 0.00000601, 0.00000563, 0.00001389, -0.00002156, 0.00002148, -0.00002049, -0.00000924, -0.00000436, -0.00000624, 0.00001125, -0.00001800, -0.00001221, 0.00000527, 0.00001488, -0.00001318, -0.00000470, 0.00000531, 0.00000803, -0.00000891, -0.00000086, 0.00000104, 0.00000053, -0.00000123, -0.00000568, 0.00000553, 0.00000782, -0.00000654, -0.00000288, 0.00000180, -0.00000623, 0.00000362, 0.00000253, -0.00000181, 0.00000343, -0.00000713, -0.00000409, 0.00000384, 0.00000378, -0.00000173, 0.00000120, 0.00000118, 0.00000423, -0.00000440, -0.00000710, 0.00000359, 0.00000125, -0.00000299, 0.00000112, 0.00000072, -0.00000117, 0.00000282, 0.00000416, -0.00000339, -0.00000060, -0.00000150, -0.00000209, 0.00000131, -0.00000125, -0.00000081, 0.00000014, },
		{ 0.00000199, -0.00000047, 0.00000053, -0.00000383, 0.00000203, 0.00000548, -0.00000313, 0.00000016, -0.00000086, -0.00000147, -0.00000086, 0.00000433, -0.00000400, -0.00000449, 0.00000327, 0.00000152, 0.00000232, -0.00000211, 0.00000294, -0.00000049, -0.00000085, -0.00000897, 0.00000357, -0.00000224, 0.00000248, 0.00000658, -0.00000293, -0.00000161, 0.00000091, -0.00000336, 0.00000174, 0.00001024, -0.00001005, -0.00000138, 0.00000028, -0.00000146, 0.00000132, -0.00000924, 0.00000783, 0.00000180, 0.00000211, -0.00001648, 0.00001307, -0.00000426, 0.00000114, -0.00001161, 0.00000854, 0.00000773, -0.00000815, 0.00000828, -0.00001125, 0.00001416, -0.00001515, -0.00000452, 0.00000075, -0.00001114, -0.00000655, -0.00002466, 0.00000326, -0.00004218, -0.00001662, -0.00024931, 0.00030259, 0.00039999, -0.00022590, -0.00003501, -0.00003553, 0.00000602, 0.00000491, 0.00000540, 0.00001184, -0.00001968, 0.00001921, -0.00001483, -0.00001447, -0.00000001, -0.00001038, 0.00001448, -0.00001923, -0.00001075, 0.00000223, 0.00001724, -0.00001323, -0.00000494, 0.00000375, 0.00000964, -0.00000945, -0.00000081, 0.00000110, 0.00000057, -0.00000118, -0.00000546, 0.00000402, 0.00000925, -0.00000674, -0.00000292, 0.00000186, -0.00000578, 0.00000214, 0.00000407, -0.00000283, 0.00000430, -0.00000726, -0.00000423, 0.00000317, 0.00000440, -0.00000199, 0.00000139, 0.00000080, 0.00000439, -0.00000378, -0.00000737, 0.00000298, 0.00000200, -0.00000330, 0.00000105, 0.00000077, -0.00000111, 0.00000212, 0.00000488, -0.00000320, -0.00000091, -0.00000127, -0.00000231, 0.00000131, -0.00000110, -0.00000091, 0.00000011, },
		{ 0.00000198, -0.00000051, 0.00000056, -0.00000354, 0.00000111, 0.00000600, -0.00000281, -0.00000015, -0.00000060, -0.00000160, -0.00000101, 0.00000418, -0.00000293, -0.00000558, 0.00000360, 0.00000110, 0.00000282, -0.00000247, 0.00000308, -0.00000055, -0.00000011, -0.00000947, 0.00000329, -0.00000168, 0.00000143, 0.00000706, -0.00000236, -0.00000221, 0.00000123, -0.00000352, 0.00000132, 0.00001038, -0.00000857, -0.00000342, 0.00000198, -0.00000320, 0.00000332, -0.00001088, 0.00000872, 0.00000081, 0.00000385, -0.00001778, 0.00001252, -0.00000314, 0.00000095, -0.00001136, 0.00000715, 0.00000919, -0.00000909, 0.00000820, -0.00001027, 0.00001186, -0.00001054, -0.00001036, 0.00000737, -0.00001685, -0.00000125, -0.00003076, 0.00001105, -0.00005101, 0.00000047, -0.00026518, 0.00027123, 0.00045085, -0.00023924, -0.00003324, -0.00003965, 0.00000618, 0.00000349, 0.00000576, 0.00000912, -0.00001711, 0.00001603, -0.00000850, -0.00001991, 0.00000432, -0.00001432, 0.00001737, -0.00002019, -0.00000939, -0.00000059, 0.00001912, -0.00001288, -0.00000529, 0.00000224, 0.00001103, -0.00000973, -0.00000094, 0.00000125, 0.00000061, -0.00000121, -0.00000504, 0.00000236, 0.00001063, -0.00000685, -0.00000296, 0.00000182, -0.00000514, 0.00000053, 0.00000560, -0.00000374, 0.00000501, -0.00000723, -0.00000441, 0.00000250, 0.00000495, -0.00000216, 0.00000153, 0.00000049, 0.00000444, -0.00000313, -0.00000754, 0.00000227, 0.00000275, -0.00000355, 0.00000096, 0.00000079, -0.00000098, 0.00000139, 0.00000550, -0.00000291, -0.00000125, -0.00000103, -0.00000249, 0.00000126, -0.00000091, -0.00000103, 0.00000009, },
		{ 0.00000195, -0.00000052, 0.00000053, -0.00000317, 0.00000020, 0.00000640, -0.00000237, -0.00000050, -0.00000032, -0.00000173, -0.00000112, 0.00000393, -0.00000180, -0.00000659, 0.00000381, 0.00000078, 0.00000321, -0.00000270, 0.00000311, -0.00000053, 0.00000053, -0.00000975, 0.00000279, -0.00000096, 0.00000036, 0.00000740, -0.00000166, -0.00000284, 0.00000154, -0.00000367, 0.00000096, 0.00001030, -0.00000684, -0.00000551, 0.00000362, -0.00000482, 0.00000508, -0.00001218, 0.00000926, 0.00000012, 0.00000520, -0.00001850, 0.00001141, -0.00000170, 0.00000057, -0.00001088, 0.00000562, 0.00001058, -0.00000974, 0.00000769, -0.00000876, 0.00000895, -0.00000545, -0.00001635, 0.00001369, -0.00002188, 0.00000326, -0.00003599, 0.00001774, -0.00005829, 0.00001451, -0.00027414, 0.00023142, 0.00050534, -0.00025200, -0.00003224, -0.00004335, 0.00000618, 0.00000198, 0.00000652, 0.00000590, -0.00001408, 0.00001220, -0.00000184, -0.00002522, 0.00000832, -0.00001779, 0.00001968, -0.00002070, -0.00000830, -0.00000302, 0.00002041, -0.00001204, -0.00000579, 0.00000086, 0.00001212, -0.00000969, -0.00000129, 0.00000150, 0.00000063, -0.00000131, -0.00000447, 0.00000063, 0.00001188, -0.00000681, -0.00000303, 0.00000171, -0.00000437, -0.00000114, 0.00000700, -0.00000444, 0.00000550, -0.00000700, -0.00000466, 0.00000188, 0.00000539, -0.00000222, 0.00000158, 0.00000027, 0.00000439, -0.00000248, -0.00000760, 0.00000147, 0.00000347, -0.00000371, 0.00000082, 0.00000080, -0.00000079, 0.00000064, 0.00000602, -0.00000249, -0.00000163, -0.00000081, -0.00000264, 0.00000116, -0.00000069, -0.00000115, 0.00000008, },
		{ 0.00000191, -0.00000049, 0.00000043, -0.00000272, -0.00000069, 0.00000665, -0.00000182, -0.00000087, -0.00000005, -0.00000184, -0.00000121, 0.00000358, -0.00000065, -0.00000747, 0.00000387, 0.00000058, 0.00000345, -0.00000277, 0.00000300, -0.00000043, 0.00000105, -0.00000980, 0.00000206, -0.00000010, -0.00000069, 0.00000761, -0.00000087, -0.00000346, 0.00000178, -0.00000376, 0.00000062, 0.00001005, -0.00000498, -0.00000755, 0.00000507, -0.00000618, 0.00000649, -0.00001303, 0.00000937, -0.00000018, 0.00000610, -0.00001859, 0.00000973, 0.00000006, -0.00000002, -0.00001017, 0.00000396, 0.00001189, -0.00001013, 0.00000681, -0.00000686, 0.00000564, -0.00000018, -0.00002210, 0.00001929, -0.00002583, 0.00000657, -0.00003992, 0.00002282, -0.00006338, 0.00002473, -0.00027562, 0.00018342, 0.00056250, -0.00026325, -0.00003263, -0.00004623, 0.00000572, 0.00000063, 0.00000744, 0.00000246, -0.00001085, 0.00000805, 0.00000476, -0.00003002, 0.00001167, -0.00002052, 0.00002122, -0.00002059, -0.00000766, -0.00000489, 0.00002098, -0.00001066, -0.00000652, -0.00000032, 0.00001285, -0.00000930, -0.00000187, 0.00000187, 0.00000061, -0.00000145, -0.00000380, -0.00000110, 0.00001293, -0.00000657, -0.00000318, 0.00000158, -0.00000354, -0.00000277, 0.00000820, -0.00000488, 0.00000572, -0.00000656, -0.00000501, 0.00000133, 0.00000568, -0.00000216, 0.00000154, 0.00000016, 0.00000423, -0.00000182, -0.00000756, 0.00000062, 0.00000412, -0.00000377, 0.00000063, 0.00000081, -0.00000058, -0.00000007, 0.00000640, -0.00000196, -0.00000204, -0.00000059, -0.00000274, 0.00000100, -0.00000045, -0.00000128, 0.00000007, },
		{ 0.00000187, -0.00000043, 0.00000027, -0.00000223, -0.00000153, 0.00000676, -0.00000118, -0.00000125, 0.00000019, -0.00000191, -0.00000129, 0.00000318, 0.00000048, -0.00000819, 0.00000375, 0.00000054, 0.00000352, -0.00000267, 0.00000274, -0.00000022, 0.00000141, -0.00000961, 0.00000111, 0.00000087, -0.00000172, 0.00000769, -0.00000002, -0.00000400, 0.00000192, -0.00000376, 0.00000027, 0.00000968, -0.00000307, -0.00000941, 0.00000622, -0.00000718, 0.00000743, -0.00001336, 0.00000899, -0.00000004, 0.00000646, -0.00001801, 0.00000747, 0.00000210, -0.00000079, -0.00000922, 0.00000216, 0.00001313, -0.00001032, 0.00000568, -0.00000473, 0.00000217, 0.00000494, -0.00002721, 0.00002377, -0.00002831, 0.00000832, -0.00004216, 0.00002581, -0.00006572, 0.00003050, -0.00026931, 0.00012777, 0.00062117, -0.00027196, -0.00003508, -0.00004782, 0.00000444, -0.00000027, 0.00000823, -0.00000091, -0.00000773, 0.00000395, 0.00001089, -0.00003395, 0.00001408, -0.00002225, 0.00002177, -0.00001971, -0.00000759, -0.00000606, 0.00002073, -0.00000869, -0.00000749, -0.00000124, 0.00001316, -0.00000852, -0.00000271, 0.00000237, 0.00000054, -0.00000158, -0.00000310, -0.00000273, 0.00001368, -0.00000608, -0.00000345, 0.00000149, -0.00000272, -0.00000426, 0.00000911, -0.00000499, 0.00000562, -0.00000587, -0.00000545, 0.00000088, 0.00000582, -0.00000195, 0.00000138, 0.00000016, 0.00000397, -0.00000115, -0.00000743, -0.00000025, 0.00000467, -0.00000370, 0.00000039, 0.00000081, -0.00000036, -0.00000074, 0.00000662, -0.00000130, -0.00000248, -0.00000039, -0.00000280, 0.00000081, -0.00000020, -0.00000139, 0.00000006, },
		{ 0.00000181, -0.00000034, 0.00000006, -0.00000170, -0.00000230, 0.00000672, -0.00000046, -0.00000161, 0.00000038, -0.00000194, -0.00000137, 0.00000276, 0.00000153, -0.00000870, 0.00000342, 0.00000067, 0.00000341, -0.00000239, 0.00000233, 0.00000010, 0.00000161, -0.00000916, -0.00000004, 0.00000192, -0.00000270, 0.00000765, 0.00000086, -0.00000444, 0.00000192, -0.00000364, -0.00000012, 0.00000925, -0.00000122, -0.00001097, 0.00000696, -0.00000772, 0.00000782, -0.00001309, 0.00000807, 0.00000058, 0.00000626, -0.00001676, 0.00000468, 0.00000440, -0.00000175, -0.00000804, 0.00000022, 0.00001433, -0.00001037, 0.00000442, -0.00000255, -0.00000119, 0.00000956, -0.00003131, 0.00002675, -0.00002899, 0.00000820, -0.00004238, 0.00002636, -0.00006488, 0.00003140, -0.00025521, 0.00006532, 0.00067996, -0.00027697, -0.00004029, -0.00004764, 0.00000197, -0.00000040, 0.00000858, -0.00000387, -0.00000507, 0.00000029, 0.00001614, -0.00003665, 0.00001524, -0.00002277, 0.00002119, -0.00001794, -0.00000823, -0.00000643, 0.00001961, -0.00000612, -0.00000874, -0.00000183, 0.00001299, -0.00000732, -0.00000380, 0.00000301, 0.00000037, -0.00000166, -0.00000245, -0.00000416, 0.00001407, -0.00000528, -0.00000388, 0.00000148, -0.00000199, -0.00000554, 0.00000965, -0.00000471, 0.00000519, -0.00000494, -0.00000601, 0.00000055, 0.00000576, -0.00000158, 0.00000111, 0.00000029, 0.00000361, -0.00000049, -0.00000723, -0.00000111, 0.00000508, -0.00000349, 0.00000008, 0.00000084, -0.00000016, -0.00000132, 0.00000668, -0.00000052, -0.00000293, -0.00000022, -0.00000281, 0.00000058, 0.00000006, -0.00000148, 0.00000005, },
		{ 0.00000175, -0.00000022, -0.00000021, -0.00000115, -0.00000298, 0.00000654, 0.00000032, -0.00000193, 0.00000050, -0.00000190, -0.00000147, 0.00000233, 0.00000247, -0.00000897, 0.00000286, 0.00000098, 0.00000311, -0.00000192, 0.00000176, 0.00000054, 0.00000163, -0.00000846, -0.00000137, 0.00000303, -0.00000362, 0.00000750, 0.00000171, -0.00000471, 0.00000174, -0.00000336, -0.00000061, 0.00000885, 0.00000047, -0.00001212, 0.00000720, -0.00000773, 0.00000761, -0.00001219, 0.00000659, 0.00000171, 0.00000548, -0.00001483, 0.00000140, 0.00000691, -0.00000287, -0.00000663, -0.00000186, 0.00001553, -0.00001037, 0.00000318, -0.00000054, -0.00000417, 0.00001335, -0.00003405, 0.00002792, -0.00002760, 0.00000599, -0.00004033, 0.00002420, -0.00006060, 0.00002722, -0.00023358, -0.00000279, 0.00073728, -0.00027707, -0.00004900, -0.00004521, -0.00000207, 0.00000057, 0.00000817, -0.00000609, -0.00000322, -0.00000254, 0.00002013, -0.00003782, 0.00001494, -0.00002192, 0.00001937, -0.00001519, -0.00000966, -0.00000591, 0.00001759, -0.00000297, -0.00001027, -0.00000208, 0.00001233, -0.00000571, -0.00000516, 0.00000382, 0.00000009, -0.00000164, -0.00000190, -0.00000532, 0.00001403, -0.00000413, -0.00000451, 0.00000160, -0.00000141, -0.00000652, 0.00000975, -0.00000402, 0.00000442, -0.00000377, -0.00000668, 0.00000035, 0.00000551, -0.00000104, 0.00000072, 0.00000053, 0.00000315, 0.00000017, -0.00000698, -0.00000191, 0.00000533, -0.00000313, -0.00000030, 0.00000090, 0.00000002, -0.00000180, 0.00000654, 0.00000037, -0.00000339, -0.00000007, -0.00000279, 0.00000032, 0.00000030, -0.00000155, 0.00000002, },
		{ 0.00000168, -0.00000007, -0.00000051, -0.00000061, -0.00000357, 0.00000621, 0.00000112, -0.00000217, 0.00000054, -0.00000178, -0.00000160, 0.00000193, 0.00000326, -0.00000897, 0.00000207, 0.00000147, 0.00000262, -0.00000129, 0.00000106, 0.00000110, 0.00000147, -0.00000752, -0.00000285, 0.00000416, -0.00000447, 0.00000727, 0.00000249, -0.00000477, 0.00000133, -0.00000290, -0.00000121, 0.00000851, 0.00000191, -0.00001277, 0.00000686, -0.00000714, 0.00000676, -0.00001067, 0.00000457, 0.00000333, 0.00000413, -0.00001228, -0.00000230, 0.00000957, -0.00000413, -0.00000498, -0.00000409, 0.00001675, -0.00001041, 0.00000210, 0.00000110, -0.00000651, 0.00001602, -0.00003512, 0.00002704, -0.00002397, 0.00000154, -0.00003591, 0.00001920, -0.00005275, 0.00001800, -0.00020498, -0.00007514, 0.00079143, -0.00027104, -0.00006187, -0.00004010, -0.00000803, 0.00000293, 0.00000668, -0.00000723, -0.00000252, -0.00000418, 0.00002250, -0.00003719, 0.00001300, -0.00001957, 0.00001624, -0.00001145, -0.00001191, -0.00000449, 0.00001468, 0.00000072, -0.00001206, -0.00000195, 0.00001115, -0.00000368, -0.00000677, 0.00000480, -0.00000034, -0.00000147, -0.00000155, -0.00000611, 0.00001349, -0.00000262, -0.00000536, 0.00000191, -0.00000105, -0.00000713, 0.00000938, -0.00000290, 0.00000330, -0.00000238, -0.00000746, 0.00000030, 0.00000505, -0.00000034, 0.00000020, 0.00000089, 0.00000261, 0.00000083, -0.00000670, -0.00000262, 0.00000537, -0.00000260, -0.00000076, 0.00000100, 0.00000013, -0.00000214, 0.00000622, 0.00000136, -0.00000384, 0.00000005, -0.00000273, 0.00000005, 0.00000052, -0.00000159, -0.00000002, },
		{ 0.00000160, 0.00000010, -0.00000083, -0.00000008, -0.00000405, 0.00000576, 0.00000192, -0.00000233, 0.00000048, -0.00000159, -0.00000177, 0.00000159, 0.00000387, -0.00000868, 0.00000104, 0.00000213, 0.00000196, -0.00000049, 0.00000023, 0.00000176, 0.00000114, -0.00000636, -0.00000445, 0.00000530, -0.00000524, 0.00000699, 0.00000316, -0.00000459, 0.00000068, -0.00000221, -0.00000197, 0.00000831, 0.00000301, -0.00001286, 0.00000591, -0.00000594, 0.00000528, -0.00000854, 0.00000203, 0.00000541, 0.00000224, -0.00000916, -0.00000633, 0.00001230, -0.00000550, -0.00000311, -0.00000648, 0.00001805, -0.00001059, 0.00000133, 0.00000219, -0.00000798, 0.00001730, -0.00003433, 0.00002395, -0.00001805, -0.00000514, -0.00002911, 0.00001138, -0.00004141, 0.00000399, -0.00017025, -0.00015009, 0.00084060, -0.00025769, -0.00007951, -0.00003194, -0.00001618, 0.00000696, 0.00000385, -0.00000700, -0.00000327, -0.00000431, 0.00002300, -0.00003460, 0.00000933, -0.00001570, 0.00001181, -0.00000672, -0.00001497, -0.00000218, 0.00001094, 0.00000488, -0.00001408, -0.00000144, 0.00000947, -0.00000127, -0.00000862, 0.00000594, -0.00000094, -0.00000112, -0.00000145, -0.00000647, 0.00001241, -0.00000072, -0.00000644, 0.00000242, -0.00000095, -0.00000733, 0.00000851, -0.00000136, 0.00000188, -0.00000079, -0.00000832, 0.00000038, 0.00000438, 0.00000051, -0.00000044, 0.00000137, 0.00000197, 0.00000149, -0.00000642, -0.00000320, 0.00000520, -0.00000191, -0.00000129, 0.00000115, 0.00000017, -0.00000234, 0.00000570, 0.00000244, -0.00000427, 0.00000013, -0.00000264, -0.00000022, 0.00000071, -0.00000159, -0.00000008, },
		{ 0.00000152, 0.00000030, -0.00000116, 0.00000040, -0.00000441, 0.00000520, 0.00000269, -0.00000236, 0.00000032, -0.00000131, -0.00000199, 0.00000131, 0.00000428, -0.00000810, -0.00000020, 0.00000295, 0.00000114, 0.00000044, -0.00000070, 0.00000252, 0.00000064, -0.00000501, -0.00000612, 0.00000639, -0.00000593, 0.00000667, 0.00000366, -0.00000412, -0.00000023, -0.00000131, -0.00000290, 0.00000829, 0.00000371, -0.00001233, 0.00000433, -0.00000414, 0.00000320, -0.00000585, -0.00000095, 0.00000789, -0.00000013, -0.00000557, -0.00001058, 0.00001503, -0.00000696, -0.00000103, -0.00000902, 0.00001945, -0.00001098, 0.00000100, 0.00000257, -0.00000839, 0.00001701, -0.00003153, 0.00001863, -0.00000987, -0.00001395, -0.00002005, 0.00000092, -0.00002684, -0.00001431, -0.00013047, -0.00022584, 0.00088297, -0.00023594, -0.00010240, -0.00002047, -0.00002673, 0.00001285, -0.00000055, -0.00000516, -0.00000572, -0.00000269, 0.00002142, -0.00002995, 0.00000393, -0.00001034, 0.00000617, -0.00000109, -0.00001878, 0.00000096, 0.00000647, 0.00000941, -0.00001628, -0.00000058, 0.00000730, 0.00000149, -0.00001069, 0.00000726, -0.00000173, -0.00000052, -0.00000166, -0.00000634, 0.00001077, 0.00000157, -0.00000776, 0.00000317, -0.00000115, -0.00000709, 0.00000714, 0.00000058, 0.00000018, 0.00000095, -0.00000925, 0.00000061, 0.00000352, 0.00000150, -0.00000118, 0.00000194, 0.00000127, 0.00000215, -0.00000615, -0.00000362, 0.00000480, -0.00000105, -0.00000190, 0.00000135, 0.00000013, -0.00000239, 0.00000500, 0.00000358, -0.00000466, 0.00000016, -0.00000252, -0.00000048, 0.00000085, -0.00000155, -0.00000015, },
		{ 0.00000143, 0.00000051, -0.00000150, 0.00000084, -0.00000464, 0.00000454, 0.00000341, -0.00000227, 0.00000004, -0.00000095, -0.00000226, 0.00000113, 0.00000447, -0.00000725, -0.00000163, 0.00000389, 0.00000021, 0.00000147, -0.00000171, 0.00000335, 0.00000001, -0.00000350, -0.00000782, 0.00000743, -0.00000653, 0.00000636, 0.00000398, -0.00000335, -0.00000142, -0.00000018, -0.00000400, 0.00000847, 0.00000397, -0.00001116, 0.00000214, -0.00000178, 0.00000058, -0.00000270, -0.00000429, 0.00001070, -0.00000288, -0.00000162, -0.00001493, 0.00001769, -0.00000847, 0.00000124, -0.00001171, 0.00002099, -0.00001166, 0.00000121, 0.00000211, -0.00000760, 0.00001504, -0.00002669, 0.00001114, 0.00000039, -0.00002467, -0.00000901, -0.00001188, -0.00000947, -0.00003621, -0.00008694, -0.00030045, 0.00091674, -0.00020484, -0.00013085, -0.00000554, -0.00003979, 0.00002070, -0.00000667, -0.00000155, -0.00001002, 0.00000082, 0.00001769, -0.00002328, -0.00000312, -0.00000363, -0.00000056, 0.00000531, -0.00002323, 0.00000482, 0.00000141, 0.00001419, -0.00001858, 0.00000061, 0.00000469, 0.00000456, -0.00001293, 0.00000874, -0.00000273, 0.00000032, -0.00000221, -0.00000569, 0.00000859, 0.00000420, -0.00000930, 0.00000414, -0.00000166, -0.00000641, 0.00000530, 0.00000285, -0.00000173, 0.00000279, -0.00001022, 0.00000096, 0.00000248, 0.00000261, -0.00000200, 0.00000260, 0.00000050, 0.00000282, -0.00000593, -0.00000387, 0.00000417, -0.00000005, -0.00000257, 0.00000161, -0.00000000, -0.00000227, 0.00000414, 0.00000475, -0.00000498, 0.00000015, -0.00000238, -0.00000073, 0.00000095, -0.00000146, -0.00000023, },
		{ 0.00000133, 0.00000072, -0.00000182, 0.00000120, -0.00000476, 0.00000381, 0.00000405, -0.00000204, -0.00000035, -0.00000050, -0.00000259, 0.00000104, 0.00000443, -0.00000613, -0.00000322, 0.00000492, -0.00000081, 0.00000257, -0.00000276, 0.00000422, -0.00000073, -0.00000188, -0.00000952, 0.00000837, -0.00000705, 0.00000607, 0.00000406, -0.00000227, -0.00000286, 0.00000115, -0.00000528, 0.00000888, 0.00000377, -0.00000938, -0.00000060, 0.00000106, -0.00000248, 0.00000082, -0.00000787, 0.00001372, -0.00000592, 0.00000258, -0.00001926, 0.00002019, -0.00000999, 0.00000369, -0.00001453, 0.00002268, -0.00001268, 0.00000203, 0.00000072, -0.00000555, 0.00001137, -0.00001989, 0.00000167, 0.00001244, -0.00003695, 0.00000365, -0.00002656, 0.00001010, -0.00006082, -0.00004114, -0.00037193, 0.00094025, -0.00016366, -0.00016500, 0.00001281, -0.00005534, 0.00003052, -0.00001454, 0.00000391, -0.00001624, 0.00000628, 0.00001184, -0.00001470, -0.00001163, 0.00000424, -0.00000817, 0.00001230, -0.00002817, 0.00000925, -0.00000409, 0.00001906, -0.00002090, 0.00000208, 0.00000170, 0.00000786, -0.00001529, 0.00001035, -0.00000392, 0.00000145, -0.00000313, -0.00000453, 0.00000588, 0.00000713, -0.00001102, 0.00000533, -0.00000248, -0.00000530, 0.00000304, 0.00000540, -0.00000378, 0.00000468, -0.00001119, 0.00000142, 0.00000128, 0.00000381, -0.00000288, 0.00000333, -0.00000031, 0.00000349, -0.00000575, -0.00000392, 0.00000330, 0.00000109, -0.00000328, 0.00000191, -0.00000023, -0.00000201, 0.00000313, 0.00000593, -0.00000524, 0.00000010, -0.00000222, -0.00000095, 0.00000099, -0.00000133, -0.00000033, },
		{ 0.00000123, 0.00000094, -0.00000211, 0.00000150, -0.00000476, 0.00000303, 0.00000458, -0.00000166, -0.00000084, 0.00000001, -0.00000296, 0.00000104, 0.00000419, -0.00000478, -0.00000491, 0.00000598, -0.00000188, 0.00000369, -0.00000382, 0.00000511, -0.00000155, -0.00000019, -0.00001114, 0.00000919, -0.00000749, 0.00000583, 0.00000391, -0.00000089, -0.00000453, 0.00000267, -0.00000672, 0.00000951, 0.00000311, -0.00000703, -0.00000381, 0.00000427, -0.00000587, 0.00000456, -0.00001158, 0.00001684, -0.00000913, 0.00000688, -0.00002344, 0.00002244, -0.00001149, 0.00000626, -0.00001746, 0.00002453, -0.00001408, 0.00000351, -0.00000160, -0.00000224, 0.00000607, -0.00001131, -0.00000947, 0.00002589, -0.00005034, 0.00001742, -0.00004255, 0.00003117, -0.00008713, 0.00000539, -0.00043832, 0.00095201, -0.00011192, -0.00020473, 0.00003440, -0.00007321, 0.00004221, -0.00002410, 0.00001118, -0.00002433, 0.00001361, 0.00000400, -0.00000447, -0.00002133, 0.00001299, -0.00001638, 0.00001964, -0.00003342, 0.00001406, -0.00000983, 0.00002386, -0.00002315, 0.00000376, -0.00000159, 0.00001131, -0.00001773, 0.00001208, -0.00000531, 0.00000284, -0.00000442, -0.00000286, 0.00000271, 0.00001030, -0.00001287, 0.00000671, -0.00000357, -0.00000381, 0.00000041, 0.00000814, -0.00000588, 0.00000654, -0.00001212, 0.00000195, -0.00000004, 0.00000507, -0.00000379, 0.00000409, -0.00000115, 0.00000415, -0.00000564, -0.00000377, 0.00000222, 0.00000233, -0.00000401, 0.00000226, -0.00000054, -0.00000160, 0.00000200, 0.00000708, -0.00000540, -0.00000001, -0.00000205, -0.00000114, 0.00000097, -0.00000116, -0.00000044, },
		{ 0.00000113, 0.00000115, -0.00000237, 0.00000171, -0.00000465, 0.00000221, 0.00000499, -0.00000113, -0.00000142, 0.00000057, -0.00000336, 0.00000113, 0.00000374, -0.00000325, -0.00000665, 0.00000702, -0.00000293, 0.00000479, -0.00000484, 0.00000598, -0.00000242, 0.00000153, -0.00001266, 0.00000987, -0.00000784, 0.00000564, 0.00000351, 0.00000077, -0.00000639, 0.00000433, -0.00000827, 0.00001035, 0.00000203, -0.00000418, -0.00000738, 0.00000773, -0.00000943, 0.00000839, -0.00001528, 0.00001995, -0.00001238, 0.00001114, -0.00002733, 0.00002437, -0.00001291, 0.00000894, -0.00002046, 0.00002651, -0.00001584, 0.00000564, -0.00000484, 0.00000225, -0.00000070, -0.00000124, -0.00002190, 0.00004026, -0.00006431, 0.00003175, -0.00005919, 0.00005290, -0.00011400, 0.00005100, -0.00049772, 0.00095077, -0.00004945, -0.00024968, 0.00005882, -0.00009310, 0.00005551, -0.00003520, 0.00002012, -0.00003415, 0.00002264, -0.00000556, 0.00000710, -0.00003187, 0.00002229, -0.00002490, 0.00002707, -0.00003873, 0.00001905, -0.00001560, 0.00002843, -0.00002522, 0.00000559, -0.00000511, 0.00001483, -0.00002018, 0.00001388, -0.00000687, 0.00000448, -0.00000606, -0.00000073, -0.00000084, 0.00001361, -0.00001479, 0.00000822, -0.00000489, -0.00000200, -0.00000247, 0.00001095, -0.00000796, 0.00000831, -0.00001297, 0.00000254, -0.00000144, 0.00000634, -0.00000470, 0.00000488, -0.00000200, 0.00000481, -0.00000560, -0.00000342, 0.00000094, 0.00000364, -0.00000475, 0.00000263, -0.00000093, -0.00000107, 0.00000080, 0.00000817, -0.00000545, -0.00000017, -0.00000188, -0.00000129, 0.00000089, -0.00000095, -0.00000055, },
		{ 0.00000103, 0.00000134, -0.00000258, 0.00000183, -0.00000444, 0.00000140, 0.00000527, -0.00000046, -0.00000207, 0.00000116, -0.00000378, 0.00000131, 0.00000312, -0.00000157, -0.00000838, 0.00000801, -0.00000394, 0.00000583, -0.00000579, 0.00000680, -0.00000330, 0.00000321, -0.00001403, 0.00001038, -0.00000811, 0.00000553, 0.00000288, 0.00000267, -0.00000839, 0.00000606, -0.00000990, 0.00001135, 0.00000058, -0.00000093, -0.00001117, 0.00001128, -0.00001302, 0.00001214, -0.00001881, 0.00002290, -0.00001554, 0.00001521, -0.00003081, 0.00002591, -0.00001423, 0.00001165, -0.00002347, 0.00002859, -0.00001794, 0.00000836, -0.00000890, 0.00000778, -0.00000869, 0.00000996, -0.00003511, 0.00005498, -0.00007827, 0.00004599, -0.00007577, 0.00007441, -0.00014027, 0.00009411, -0.00054842, 0.00093558, 0.00002360, -0.00029921, 0.00008551, -0.00011454, 0.00007009, -0.00004754, 0.00003048, -0.00004542, 0.00003306, -0.00001648, 0.00001957, -0.00004283, 0.00003176, -0.00003337, 0.00003429, -0.00004389, 0.00002399, -0.00002118, 0.00003260, -0.00002701, 0.00000748, -0.00000875, 0.00001833, -0.00002256, 0.00001569, -0.00000856, 0.00000633, -0.00000800, 0.00000179, -0.00000467, 0.00001695, -0.00001669, 0.00000980, -0.00000639, 0.00000006, -0.00000551, 0.00001374, -0.00000990, 0.00000992, -0.00001370, 0.00000315, -0.00000289, 0.00000759, -0.00000559, 0.00000566, -0.00000284, 0.00000546, -0.00000563, -0.00000288, -0.00000049, 0.00000497, -0.00000545, 0.00000301, -0.00000138, -0.00000045, -0.00000045, 0.00000916, -0.00000538, -0.00000038, -0.00000172, -0.00000138, 0.00000075, -0.00000070, -0.00000067, },
		{ 0.00000093, 0.00000152, -0.00000273, 0.00000187, -0.00000415, 0.00000060, 0.00000539, 0.00000033, -0.00000276, 0.00000176, -0.00000420, 0.00000154, 0.00000236, 0.00000018, -0.00001004, 0.00000887, -0.00000485, 0.00000676, -0.00000663, 0.00000754, -0.00000414, 0.00000482, -0.00001519, 0.00001070, -0.00000830, 0.00000549, 0.00000202, 0.00000475, -0.00001045, 0.00000781, -0.00001154, 0.00001245, -0.00000116, 0.00000259, -0.00001503, 0.00001477, -0.00001647, 0.00001567, -0.00002204, 0.00002557, -0.00001846, 0.00001896, -0.00003375, 0.00002698, -0.00001538, 0.00001436, -0.00002642, 0.00003070, -0.00002031, 0.00001157, -0.00001362, 0.00001413, -0.00001759, 0.00002183, -0.00004858, 0.00006945, -0.00009157, 0.00005948, -0.00009153, 0.00009480, -0.00016477, 0.00013319, -0.00058893, 0.00090587, 0.00010669, -0.00035242, 0.00011368, -0.00013695, 0.00008545, -0.00006074, 0.00004190, -0.00005778, 0.00004447, -0.00002829, 0.00003244, -0.00005373, 0.00004096, -0.00004144, 0.00004100, -0.00004862, 0.00002865, -0.00002635, 0.00003621, -0.00002843, 0.00000935, -0.00001240, 0.00002168, -0.00002480, 0.00001748, -0.00001032, 0.00000834, -0.00001017, 0.00000459, -0.00000865, 0.00002021, -0.00001848, 0.00001138, -0.00000797, 0.00000225, -0.00000860, 0.00001639, -0.00001163, 0.00001130, -0.00001428, 0.00000374, -0.00000434, 0.00000876, -0.00000641, 0.00000641, -0.00000365, 0.00000608, -0.00000573, -0.00000217, -0.00000204, 0.00000628, -0.00000609, 0.00000339, -0.00000187, 0.00000024, -0.00000170, 0.00001002, -0.00000518, -0.00000064, -0.00000156, -0.00000143, 0.00000055, -0.00000043, -0.00000079, },
		{ 0.00000083, 0.00000166, -0.00000282, 0.00000182, -0.00000377, -0.00000016, 0.00000537, 0.00000123, -0.00000346, 0.00000233, -0.00000460, 0.00000182, 0.00000150, 0.00000195, -0.00001156, 0.00000956, -0.00000561, 0.00000754, -0.00000733, 0.00000816, -0.00000491, 0.00000629, -0.00001612, 0.00001083, -0.00000841, 0.00000552, 0.00000098, 0.00000697, -0.00001250, 0.00000950, -0.00001313, 0.00001360, -0.00000310, 0.00000624, -0.00001879, 0.00001803, -0.00001963, 0.00001881, -0.00002483, 0.00002784, -0.00002104, 0.00002224, -0.00003605, 0.00002753, -0.00001634, 0.00001698, -0.00002924, 0.00003278, -0.00002286, 0.00001514, -0.00001882, 0.00002102, -0.00002700, 0.00003387, -0.00006170, 0.00008304, -0.00010360, 0.00007156, -0.00010572, 0.00011317, -0.00018638, 0.00016687, -0.00061804, 0.00086145, 0.00019894, -0.00040814, 0.00014239, -0.00015958, 0.00010102, -0.00007431, 0.00005393, -0.00007077, 0.00005638, -0.00004046, 0.00004517, -0.00006409, 0.00004949, -0.00004873, 0.00004691, -0.00005270, 0.00003279, -0.00003091, 0.00003909, -0.00002939, 0.00001111, -0.00001596, 0.00002480, -0.00002682, 0.00001916, -0.00001209, 0.00001043, -0.00001248, 0.00000757, -0.00001264, 0.00002325, -0.00002007, 0.00001288, -0.00000957, 0.00000449, -0.00001160, 0.00001877, -0.00001305, 0.00001241, -0.00001466, 0.00000429, -0.00000576, 0.00000982, -0.00000713, 0.00000709, -0.00000441, 0.00000666, -0.00000589, -0.00000132, -0.00000365, 0.00000752, -0.00000664, 0.00000373, -0.00000238, 0.00000096, -0.00000291, 0.00001071, -0.00000483, -0.00000094, -0.00000142, -0.00000143, 0.00000030, -0.00000013, -0.00000090, },
		{ 0.00000073, 0.00000178, -0.00000285, 0.00000168, -0.00000335, -0.00000087, 0.00000519, 0.00000222, -0.00000415, 0.00000286, -0.00000496, 0.00000211, 0.00000057, 0.00000369, -0.00001288, 0.00001003, -0.00000620, 0.00000814, -0.00000785, 0.00000863, -0.00000557, 0.00000760, -0.00001679, 0.00001075, -0.00000843, 0.00000560, -0.00000021, 0.00000922, -0.00001445, 0.00001104, -0.00001458, 0.00001472, -0.00000514, 0.00000986, -0.00002227, 0.00002089, -0.00002233, 0.00002142, -0.00002706, 0.00002959, -0.00002316, 0.00002494, -0.00003763, 0.00002752, -0.00001706, 0.00001944, -0.00003184, 0.00003471, -0.00002546, 0.00001890, -0.00002426, 0.00002813, -0.00003651, 0.00004554, -0.00007387, 0.00009513, -0.00011372, 0.00008160, -0.00011765, 0.00012871, -0.00020411, 0.00019395, -0.00063488, 0.00080254, 0.00029907, -0.00046496, 0.00017056, -0.00018163, 0.00011615, -0.00008769, 0.00006604, -0.00008384, 0.00006826, -0.00005241, 0.00005719, -0.00007341, 0.00005691, -0.00005491, 0.00005175, -0.00005589, 0.00003622, -0.00003467, 0.00004114, -0.00002981, 0.00001269, -0.00001932, 0.00002757, -0.00002854, 0.00002068, -0.00001381, 0.00001252, -0.00001484, 0.00001060, -0.00001649, 0.00002593, -0.00002135, 0.00001422, -0.00001107, 0.00000665, -0.00001440, 0.00002080, -0.00001408, 0.00001318, -0.00001482, 0.00000477, -0.00000708, 0.00001073, -0.00000771, 0.00000768, -0.00000509, 0.00000718, -0.00000608, -0.00000036, -0.00000527, 0.00000864, -0.00000706, 0.00000401, -0.00000287, 0.00000167, -0.00000404, 0.00001122, -0.00000433, -0.00000127, -0.00000130, -0.00000137, 0.00000001, 0.00000017, -0.00000100, },
		{ 0.00000063, 0.00000186, -0.00000281, 0.00000147, -0.00000288, -0.00000151, 0.00000488, 0.00000325, -0.00000479, 0.00000331, -0.00000524, 0.00000238, -0.00000036, 0.00000531, -0.00001395, 0.00001024, -0.00000657, 0.00000854, -0.00000818, 0.00000894, -0.00000610, 0.00000870, -0.00001716, 0.00001045, -0.00000838, 0.00000573, -0.00000148, 0.00001144, -0.00001620, 0.00001236, -0.00001583, 0.00001573, -0.00000715, 0.00001331, -0.00002532, 0.00002321, -0.00002445, 0.00002340, -0.00002863, 0.00003075, -0.00002471, 0.00002696, -0.00003842, 0.00002693, -0.00001752, 0.00002168, -0.00003411, 0.00003640, -0.00002798, 0.00002264, -0.00002965, 0.00003511, -0.00004564, 0.00005631, -0.00008453, 0.00010513, -0.00012138, 0.00008905, -0.00012671, 0.00014069, -0.00021709, 0.00021352, -0.00063896, 0.00072981, 0.00040549, -0.00052127, 0.00019698, -0.00020219, 0.00013010, -0.00010026, 0.00007765, -0.00009642, 0.00007950, -0.00006354, 0.00006793, -0.00008121, 0.00006284, -0.00005966, 0.00005526, -0.00005800, 0.00003875, -0.00003746, 0.00004224, -0.00002964, 0.00001402, -0.00002237, 0.00002990, -0.00002988, 0.00002196, -0.00001538, 0.00001451, -0.00001712, 0.00001352, -0.00002004, 0.00002814, -0.00002224, 0.00001530, -0.00001240, 0.00000864, -0.00001687, 0.00002235, -0.00001467, 0.00001359, -0.00001474, 0.00000514, -0.00000828, 0.00001144, -0.00000814, 0.00000815, -0.00000567, 0.00000763, -0.00000628, 0.00000066, -0.00000684, 0.00000959, -0.00000731, 0.00000422, -0.00000334, 0.00000235, -0.00000504, 0.00001151, -0.00000369, -0.00000163, -0.00000121, -0.00000127, -0.00000032, 0.00000047, -0.00000109, },
		{ 0.00000054, 0.00000190, -0.00000270, 0.00000119, -0.00000238, -0.00000207, 0.00000443, 0.00000431, -0.00000535, 0.00000365, -0.00000543, 0.00000263, -0.00000127, 0.00000677, -0.00001471, 0.00001017, -0.00000672, 0.00000871, -0.00000828, 0.00000906, -0.00000646, 0.00000955, -0.00001722, 0.00000994, -0.00000823, 0.00000586, -0.00000278, 0.00001353, -0.00001767, 0.00001337, -0.00001681, 0.00001654, -0.00000902, 0.00001644, -0.00002778, 0.00002485, -0.00002586, 0.00002463, -0.00002947, 0.00003125, -0.00002564, 0.00002824, -0.00003837, 0.00002573, -0.00001767, 0.00002361, -0.00003595, 0.00003772, -0.00003026, 0.00002616, -0.00003472, 0.00004159, -0.00005395, 0.00006564, -0.00009312, 0.00011255, -0.00012613, 0.00009344, -0.00013238, 0.00014855, -0.00022466, 0.00022492, -0.00063016, 0.00064433, 0.00051627, -0.00057529, 0.00022039, -0.00022033, 0.00014214, -0.00011139, 0.00008816, -0.00010791, 0.00008951, -0.00007326, 0.00007685, -0.00008705, 0.00006695, -0.00006271, 0.00005726, -0.00005887, 0.00004024, -0.00003917, 0.00004232, -0.00002884, 0.00001502, -0.00002502, 0.00003169, -0.00003076, 0.00002293, -0.00001673, 0.00001630, -0.00001921, 0.00001621, -0.00002314, 0.00002975, -0.00002264, 0.00001606, -0.00001346, 0.00001034, -0.00001893, 0.00002337, -0.00001476, 0.00001361, -0.00001441, 0.00000539, -0.00000931, 0.00001192, -0.00000838, 0.00000849, -0.00000612, 0.00000798, -0.00000648, 0.00000169, -0.00000830, 0.00001031, -0.00000738, 0.00000433, -0.00000374, 0.00000295, -0.00000589, 0.00001157, -0.00000291, -0.00000200, -0.00000115, -0.00000112, -0.00000067, 0.00000076, -0.00000115, },
		{ 0.00000045, 0.00000191, -0.00000253, 0.00000085, -0.00000188, -0.00000254, 0.00000386, 0.00000535, -0.00000578, 0.00000386, -0.00000552, 0.00000281, -0.00000210, 0.00000802, -0.00001513, 0.00000979, -0.00000662, 0.00000864, -0.00000817, 0.00000898, -0.00000663, 0.00001013, -0.00001695, 0.00000922, -0.00000800, 0.00000598, -0.00000405, 0.00001540, -0.00001878, 0.00001400, -0.00001743, 0.00001708, -0.00001065, 0.00001909, -0.00002952, 0.00002571, -0.00002649, 0.00002506, -0.00002952, 0.00003103, -0.00002588, 0.00002870, -0.00003746, 0.00002394, -0.00001749, 0.00002516, -0.00003726, 0.00003856, -0.00003214, 0.00002922, -0.00003918, 0.00004722, -0.00006102, 0.00007307, -0.00009920, 0.00011697, -0.00012759, 0.00009444, -0.00013431, 0.00015186, -0.00022638, 0.00022780, -0.00060878, 0.00054759, 0.00062924, -0.00062514, 0.00023950, -0.00023512, 0.00015154, -0.00012046, 0.00009698, -0.00011771, 0.00009771, -0.00008100, 0.00008347, -0.00009058, 0.00006897, -0.00006388, 0.00005760, -0.00005840, 0.00004058, -0.00003972, 0.00004136, -0.00002739, 0.00001564, -0.00002717, 0.00003287, -0.00003113, 0.00002352, -0.00001777, 0.00001779, -0.00002099, 0.00001851, -0.00002566, 0.00003065, -0.00002249, 0.00001642, -0.00001418, 0.00001166, -0.00002046, 0.00002377, -0.00001433, 0.00001324, -0.00001383, 0.00000549, -0.00001014, 0.00001215, -0.00000842, 0.00000866, -0.00000644, 0.00000822, -0.00000663, 0.00000269, -0.00000958, 0.00001077, -0.00000724, 0.00000432, -0.00000405, 0.00000345, -0.00000656, 0.00001139, -0.00000200, -0.00000238, -0.00000112, -0.00000094, -0.00000102, 0.00000102, -0.00000119, },
		{ 0.00000037, 0.00000186, -0.00000230, 0.00000047, -0.00000138, -0.00000291, 0.00000320, 0.00000634, -0.00000607, 0.00000393, -0.00000547, 0.00000291, -0.00000283, 0.00000900, -0.00001518, 0.00000910, -0.00000627, 0.00000833, -0.00000782, 0.00000870, -0.00000661, 0.00001044, -0.00001637, 0.00000830, -0.00000767, 0.00000605, -0.00000521, 0.00001697, -0.00001943, 0.00001418, -0.00001765, 0.00001728, -0.00001194, 0.00002116, -0.00003044, 0.00002572, -0.00002629, 0.00002465, -0.00002877, 0.00003009, -0.00002541, 0.00002834, -0.00003571, 0.00002158, -0.00001697, 0.00002626, -0.00003796, 0.00003881, -0.00003347, 0.00003161, -0.00004275, 0.00005165, -0.00006643, 0.00007817, -0.00010239, 0.00011808, -0.00012553, 0.00009185, -0.00013227, 0.00015040, -0.00022204, 0.00022213, -0.00057551, 0.00044143, 0.00074199, -0.00066892, 0.00025306, -0.00024571, 0.00015761, -0.00012686, 0.00010355, -0.00012529, 0.00010357, -0.00008629, 0.00008740, -0.00009153, 0.00006870, -0.00006304, 0.00005621, -0.00005654, 0.00003972, -0.00003906, 0.00003934, -0.00002530, 0.00001586, -0.00002876, 0.00003336, -0.00003092, 0.00002367, -0.00001842, 0.00001888, -0.00002234, 0.00002031, -0.00002746, 0.00003077, -0.00002175, 0.00001634, -0.00001449, 0.00001254, -0.00002141, 0.00002354, -0.00001337, 0.00001249, -0.00001300, 0.00000545, -0.00001075, 0.00001209, -0.00000823, 0.00000867, -0.00000660, 0.00000833, -0.00000672, 0.00000361, -0.00001065, 0.00001093, -0.00000688, 0.00000419, -0.00000426, 0.00000382, -0.00000702, 0.00001098, -0.00000098, -0.00000274, -0.00000113, -0.00000072, -0.00000138, 0.00000126, -0.00000120, },
		{ 0.00000029, 0.00000178, -0.00000202, 0.00000006, -0.00000091, -0.00000319, 0.00000246, 0.00000725, -0.00000619, 0.00000383, -0.00000530, 0.00000290, -0.00000341, 0.00000969, -0.00001485, 0.00000811, -0.00000570, 0.00000779, -0.00000726, 0.00000821, -0.00000637, 0.00001045, -0.00001547, 0.00000719, -0.00000725, 0.00000603, -0.00000621, 0.00001817, -0.00001959, 0.00001388, -0.00001743, 0.00001710, -0.00001280, 0.00002254, -0.00003046, 0.00002484, -0.00002523, 0.00002340, -0.00002724, 0.00002845, -0.00002423, 0.00002716, -0.00003314, 0.00001868, -0.00001610, 0.00002684, -0.00003795, 0.00003837, -0.00003410, 0.00003314, -0.00004518, 0.00005461, -0.00006988, 0.00008065, -0.00010244, 0.00011574, -0.00011988, 0.00008561, -0.00012623, 0.00014415, -0.00021166, 0.00020820, -0.00053139, 0.00032798, 0.00085203, -0.00070472, 0.00025991, -0.00025130, 0.00015976, -0.00013010, 0.00010738, -0.00013015, 0.00010664, -0.00008872, 0.00008834, -0.00008971, 0.00006608, -0.00006016, 0.00005310, -0.00005329, 0.00003766, -0.00003721, 0.00003632, -0.00002261, 0.00001565, -0.00002973, 0.00003312, -0.00003011, 0.00002334, -0.00001861, 0.00001949, -0.00002317, 0.00002148, -0.00002847, 0.00003006, -0.00002038, 0.00001579, -0.00001435, 0.00001291, -0.00002174, 0.00002264, -0.00001190, 0.00001138, -0.00001194, 0.00000524, -0.00001111, 0.00001174, -0.00000782, 0.00000849, -0.00000659, 0.00000829, -0.00000671, 0.00000440, -0.00001145, 0.00001076, -0.00000629, 0.00000392, -0.00000435, 0.00000404, -0.00000725, 0.00001033, 0.00000012, -0.00000308, -0.00000118, -0.00000048, -0.00000173, 0.00000145, -0.00000118, },
		{ 0.00000022, 0.00000166, -0.00000169, -0.00000037, -0.00000047, -0.00000337, 0.00000168, 0.00000803, -0.00000611, 0.00000356, -0.00000498, 0.00000279, -0.00000382, 0.00001005, -0.00001413, 0.00000683, -0.00000491, 0.00000704, -0.00000649, 0.00000754, -0.00000594, 0.00001017, -0.00001427, 0.00000592, -0.00000673, 0.00000591, -0.00000697, 0.00001892, -0.00001919, 0.00001307, -0.00001675, 0.00001648, -0.00001317, 0.00002316, -0.00002955, 0.00002306, -0.00002334, 0.00002136, -0.00002496, 0.00002614, -0.00002238, 0.00002519, -0.00002981, 0.00001531, -0.00001488, 0.00002686, -0.00003716, 0.00003716, -0.00003391, 0.00003364, -0.00004627, 0.00005586, -0.00007111, 0.00008029, -0.00009924, 0.00010992, -0.00011070, 0.00007584, -0.00011630, 0.00013326, -0.00019554, 0.00018659, -0.00047780, 0.00020963, 0.00095678, -0.00073078, 0.00025905, -0.00025127, 0.00015750, -0.00012976, 0.00010808, -0.00013191, 0.00010658, -0.00008802, 0.00008613, -0.00008508, 0.00006110, -0.00005529, 0.00004834, -0.00004873, 0.00003446, -0.00003422, 0.00003237, -0.00001937, 0.00001500, -0.00003003, 0.00003212, -0.00002868, 0.00002248, -0.00001828, 0.00001953, -0.00002338, 0.00002194, -0.00002860, 0.00002848, -0.00001839, 0.00001476, -0.00001373, 0.00001274, -0.00002141, 0.00002111, -0.00000996, 0.00000995, -0.00001069, 0.00000488, -0.00001120, 0.00001109, -0.00000718, 0.00000812, -0.00000640, 0.00000809, -0.00000658, 0.00000502, -0.00001194, 0.00001026, -0.00000547, 0.00000351, -0.00000431, 0.00000409, -0.00000726, 0.00000947, 0.00000129, -0.00000337, -0.00000127, -0.00000023, -0.00000204, 0.00000159, -0.00000112, },
		{ 0.00000016, 0.00000151, -0.00000133, -0.00000081, -0.00000007, -0.00000347, 0.00000088, 0.00000868, -0.00000583, 0.00000312, -0.00000452, 0.00000255, -0.00000404, 0.00001010, -0.00001306, 0.00000531, -0.00000394, 0.00000610, -0.00000553, 0.00000669, -0.00000530, 0.00000962, -0.00001280, 0.00000450, -0.00000611, 0.00000564, -0.00000746, 0.00001918, -0.00001821, 0.00001174, -0.00001559, 0.00001543, -0.00001302, 0.00002299, -0.00002770, 0.00002044, -0.00002067, 0.00001858, -0.00002202, 0.00002323, -0.00001991, 0.00002249, -0.00002580, 0.00001153, -0.00001333, 0.00002628, -0.00003556, 0.00003512, -0.00003282, 0.00003297, -0.00004587, 0.00005523, -0.00006999, 0.00007700, -0.00009282, 0.00010075, -0.00009821, 0.00006281, -0.00010280, 0.00011811, -0.00017419, 0.00015818, -0.00041639, 0.00008893, 0.00105372, -0.00074547, 0.00024969, -0.00024517, 0.00015052, -0.00012556, 0.00010537, -0.00013032, 0.00010318, -0.00008403, 0.00008072, -0.00007771, 0.00005389, -0.00004857, 0.00004207, -0.00004298, 0.00003020, -0.00003019, 0.00002759, -0.00001566, 0.00001393, -0.00002965, 0.00003036, -0.00002661, 0.00002109, -0.00001739, 0.00001896, -0.00002292, 0.00002164, -0.00002782, 0.00002606, -0.00001581, 0.00001327, -0.00001263, 0.00001203, -0.00002045, 0.00001896, -0.00000761, 0.00000826, -0.00000926, 0.00000437, -0.00001104, 0.00001014, -0.00000632, 0.00000757, -0.00000602, 0.00000772, -0.00000630, 0.00000545, -0.00001210, 0.00000941, -0.00000444, 0.00000298, -0.00000413, 0.00000397, -0.00000705, 0.00000841, 0.00000249, -0.00000360, -0.00000140, 0.00000002, -0.00000232, 0.00000168, -0.00000103, },
		{ 0.00000010, 0.00000132, -0.00000094, -0.00000123, 0.00000027, -0.00000348, 0.00000008, 0.00000915, -0.00000533, 0.00000251, -0.00000393, 0.00000220, -0.00000408, 0.00000981, -0.00001165, 0.00000357, -0.00000283, 0.00000501, -0.00000443, 0.00000569, -0.00000449, 0.00000880, -0.00001109, 0.00000297, -0.00000540, 0.00000521, -0.00000764, 0.00001892, -0.00001666, 0.00000991, -0.00001399, 0.00001394, -0.00001233, 0.00002203, -0.00002496, 0.00001703, -0.00001733, 0.00001517, -0.00001852, 0.00001981, -0.00001689, 0.00001913, -0.00002121, 0.00000743, -0.00001147, 0.00002508, -0.00003311, 0.00003222, -0.00003077, 0.00003107, -0.00004389, 0.00005266, -0.00006645, 0.00007082, -0.00008332, 0.00008851, -0.00008277, 0.00004693, -0.00008616, 0.00009923, -0.00014830, 0.00012406, -0.00034904, -0.00003150, 0.00114042, -0.00074744, 0.00023126, -0.00023274, 0.00013866, -0.00011737, 0.00009911, -0.00012524, 0.00009635, -0.00007676, 0.00007223, -0.00006778, 0.00004467, -0.00004021, 0.00003450, -0.00003623, 0.00002505, -0.00002525, 0.00002213, -0.00001157, 0.00001246, -0.00002859, 0.00002783, -0.00002393, 0.00001915, -0.00001592, 0.00001774, -0.00002176, 0.00002055, -0.00002615, 0.00002283, -0.00001269, 0.00001134, -0.00001107, 0.00001078, -0.00001888, 0.00001628, -0.00000491, 0.00000637, -0.00000772, 0.00000372, -0.00001061, 0.00000891, -0.00000527, 0.00000685, -0.00000547, 0.00000718, -0.00000588, 0.00000565, -0.00001191, 0.00000824, -0.00000321, 0.00000234, -0.00000383, 0.00000368, -0.00000662, 0.00000718, 0.00000369, -0.00000376, -0.00000156, 0.00000027, -0.00000255, 0.00000171, -0.00000091, },
		{ 0.00000006, 0.00000110, -0.00000055, -0.00000162, 0.00000055, -0.00000343, -0.00000068, 0.00000942, -0.00000462, 0.00000176, -0.00000323, 0.00000174, -0.00000392, 0.00000922, -0.00000994, 0.00000168, -0.00000162, 0.00000379, -0.00000320, 0.00000456, -0.00000352, 0.00000776, -0.00000918, 0.00000136, -0.00000461, 0.00000459, -0.00000747, 0.00001813, -0.00001455, 0.00000761, -0.00001197, 0.00001205, -0.00001111, 0.00002029, -0.00002141, 0.00001297, -0.00001343, 0.00001126, -0.00001458, 0.00001599, -0.00001341, 0.00001523, -0.00001615, 0.00000311, -0.00000934, 0.00002325, -0.00002983, 0.00002847, -0.00002774, 0.00002789, -0.00004031, 0.00004815, -0.00006057, 0.00006193, -0.00007106, 0.00007361, -0.00006488, 0.00002873, -0.00006698, 0.00007729, -0.00011877, 0.00008549, -0.00027775, -0.00014904, 0.00121469, -0.00073563, 0.00020353, -0.00021398, 0.00012196, -0.00010524, 0.00008932, -0.00011671, 0.00008617, -0.00006635, 0.00006088, -0.00005560, 0.00003374, -0.00003053, 0.00002590, -0.00002869, 0.00001917, -0.00001958, 0.00001613, -0.00000722, 0.00001063, -0.00002686, 0.00002460, -0.00002067, 0.00001669, -0.00001388, 0.00001587, -0.00001989, 0.00001867, -0.00002361, 0.00001887, -0.00000911, 0.00000904, -0.00000908, 0.00000906, -0.00001676, 0.00001315, -0.00000197, 0.00000436, -0.00000609, 0.00000296, -0.00000993, 0.00000742, -0.00000403, 0.00000597, -0.00000475, 0.00000646, -0.00000528, 0.00000561, -0.00001137, 0.00000676, -0.00000182, 0.00000160, -0.00000340, 0.00000323, -0.00000601, 0.00000583, 0.00000485, -0.00000383, -0.00000176, 0.00000049, -0.00000272, 0.00000167, -0.00000075, },
		{ 0.00000002, 0.00000087, -0.00000015, -0.00000198, 0.00000076, -0.00000331, -0.00000140, 0.00000950, -0.00000372, 0.00000087, -0.00000244, 0.00000118, -0.00000358, 0.00000834, -0.00000799, -0.00000030, -0.00000037, 0.00000251, -0.00000189, 0.00000335, -0.00000242, 0.00000652, -0.00000711, -0.00000030, -0.00000374, 0.00000378, -0.00000694, 0.00001681, -0.00001192, 0.00000492, -0.00000959, 0.00000979, -0.00000940, 0.00001785, -0.00001715, 0.00000838, -0.00000911, 0.00000699, -0.00001033, 0.00001190, -0.00000960, 0.00001089, -0.00001077, -0.00000134, -0.00000698, 0.00002083, -0.00002575, 0.00002393, -0.00002377, 0.00002348, -0.00003520, 0.00004180, -0.00005251, 0.00005059, -0.00005645, 0.00005657, -0.00004513, 0.00000885, -0.00004593, 0.00005312, -0.00008660, 0.00004390, -0.00020460, -0.00026116, 0.00127460, -0.00070934, 0.00016654, -0.00018914, 0.00010067, -0.00008938, 0.00007619, -0.00010488, 0.00007285, -0.00005309, 0.00004705, -0.00004157, 0.00002151, -0.00001986, 0.00001659, -0.00002062, 0.00001277, -0.00001337, 0.00000978, -0.00000272, 0.00000850, -0.00002450, 0.00002070, -0.00001691, 0.00001374, -0.00001128, 0.00001337, -0.00001735, 0.00001604, -0.00002028, 0.00001430, -0.00000519, 0.00000645, -0.00000674, 0.00000691, -0.00001418, 0.00000966, 0.00000112, 0.00000230, -0.00000444, 0.00000209, -0.00000902, 0.00000569, -0.00000264, 0.00000495, -0.00000386, 0.00000558, -0.00000452, 0.00000532, -0.00001049, 0.00000501, -0.00000032, 0.00000079, -0.00000286, 0.00000264, -0.00000522, 0.00000439, 0.00000595, -0.00000380, -0.00000198, 0.00000069, -0.00000283, 0.00000157, -0.00000056, },
		{ -0.00000000, 0.00000062, 0.00000023, -0.00000228, 0.00000091, -0.00000314, -0.00000205, 0.00000938, -0.00000263, -0.00000011, -0.00000158, 0.00000055, -0.00000309, 0.00000722, -0.00000587, -0.00000231, 0.00000088, 0.00000119, -0.00000054, 0.00000208, -0.00000122, 0.00000512, -0.00000493, -0.00000196, -0.00000281, 0.00000279, -0.00000607, 0.00001501, -0.00000885, 0.00000190, -0.00000695, 0.00000724, -0.00000727, 0.00001480, -0.00001232, 0.00000343, -0.00000455, 0.00000252, -0.00000593, 0.00000766, -0.00000556, 0.00000624, -0.00000519, -0.00000580, -0.00000444, 0.00001785, -0.00002094, 0.00001866, -0.00001894, 0.00001791, -0.00002867, 0.00003380, -0.00004254, 0.00003722, -0.00004002, 0.00003802, -0.00002419, -0.00001199, -0.00002378, 0.00002757, -0.00005290, 0.00000076, -0.00013166, -0.00036551, 0.00131857, -0.00066828, 0.00012070, -0.00015871, 0.00007523, -0.00007015, 0.00006002, -0.00009009, 0.00005675, -0.00003738, 0.00003120, -0.00002620, 0.00000842, -0.00000861, 0.00000690, -0.00001229, 0.00000609, -0.00000681, 0.00000326, 0.00000181, 0.00000611, -0.00002157, 0.00001623, -0.00001271, 0.00001037, -0.00000817, 0.00001030, -0.00001418, 0.00001275, -0.00001627, 0.00000926, -0.00000105, 0.00000366, -0.00000412, 0.00000441, -0.00001123, 0.00000593, 0.00000425, 0.00000027, -0.00000279, 0.00000116, -0.00000790, 0.00000377, -0.00000114, 0.00000381, -0.00000284, 0.00000456, -0.00000361, 0.00000481, -0.00000930, 0.00000303, 0.00000126, -0.00000005, -0.00000224, 0.00000192, -0.00000431, 0.00000291, 0.00000694, -0.00000367, -0.00000221, 0.00000084, -0.00000286, 0.00000140, -0.00000035, },
		{ -0.00000002, 0.00000037, 0.00000058, -0.00000253, 0.00000099, -0.00000293, -0.00000261, 0.00000904, -0.00000139, -0.00000116, -0.00000067, -0.00000014, -0.00000245, 0.00000590, -0.00000363, -0.00000428, 0.00000207, -0.00000012, 0.00000082, 0.00000079, 0.00000004, 0.00000361, -0.00000271, -0.00000359, -0.00000183, 0.00000161, -0.00000487, 0.00001278, -0.00000542, -0.00000134, -0.00000412, 0.00000447, -0.00000479, 0.00001124, -0.00000708, -0.00000169, 0.00000009, -0.00000199, -0.00000153, 0.00000341, -0.00000143, 0.00000143, 0.00000044, -0.00001016, -0.00000179, 0.00001437, -0.00001550, 0.00001279, -0.00001336, 0.00001134, -0.00002092, 0.00002440, -0.00003101, 0.00002228, -0.00002237, 0.00001864, -0.00000281, -0.00003304, -0.00000133, 0.00000158, -0.00001878, -0.00004245, -0.00006093, -0.00045999, 0.00134543, -0.00061258, 0.00006675, -0.00012344, 0.00004628, -0.00004812, 0.00004130, -0.00007279, 0.00003837, -0.00001977, 0.00001392, -0.00001003, -0.00000504, 0.00000280, -0.00000282, -0.00000401, -0.00000065, -0.00000013, -0.00000324, 0.00000625, 0.00000356, -0.00001814, 0.00001129, -0.00000818, 0.00000665, -0.00000462, 0.00000672, -0.00001048, 0.00000890, -0.00001173, 0.00000389, 0.00000317, 0.00000078, -0.00000131, 0.00000167, -0.00000802, 0.00000209, 0.00000730, -0.00000164, -0.00000121, 0.00000018, -0.00000660, 0.00000170, 0.00000043, 0.00000259, -0.00000171, 0.00000340, -0.00000255, 0.00000407, -0.00000783, 0.00000089, 0.00000286, -0.00000090, -0.00000157, 0.00000111, -0.00000330, 0.00000142, 0.00000781, -0.00000342, -0.00000246, 0.00000095, -0.00000282, 0.00000118, -0.00000012, },
		{ -0.00000002, 0.00000012, 0.00000090, -0.00000271, 0.00000100, -0.00000270, -0.00000307, 0.00000852, -0.00000003, -0.00000224, 0.00000023, -0.00000085, -0.00000171, 0.00000443, -0.00000135, -0.00000614, 0.00000315, -0.00000138, 0.00000215, -0.00000049, 0.00000132, 0.00000203, -0.00000048, -0.00000515, -0.00000083, 0.00000027, -0.00000338, 0.00001018, -0.00000174, -0.00000469, -0.00000121, 0.00000159, -0.00000207, 0.00000731, -0.00000162, -0.00000681, 0.00000463, -0.00000637, 0.00000274, -0.00000072, 0.00000269, -0.00000342, 0.00000598, -0.00001431, 0.00000090, 0.00001047, -0.00000955, 0.00000647, -0.00000720, 0.00000396, -0.00001220, 0.00001394, -0.00001835, 0.00000633, -0.00000416, -0.00000086, 0.00001829, -0.00005355, 0.00002062, -0.00002394, 0.00001463, -0.00008429, 0.00000575, -0.00054282, 0.00135446, -0.00054282, 0.00000575, -0.00008429, 0.00001463, -0.00002394, 0.00002062, -0.00005355, 0.00001829, -0.00000086, -0.00000416, 0.00000633, -0.00001835, 0.00001394, -0.00001220, 0.00000396, -0.00000720, 0.00000647, -0.00000955, 0.00001047, 0.00000090, -0.00001431, 0.00000598, -0.00000342, 0.00000269, -0.00000072, 0.00000274, -0.00000637, 0.00000463, -0.00000681, -0.00000162, 0.00000731, -0.00000207, 0.00000159, -0.00000121, -0.00000469, -0.00000174, 0.00001018, -0.00000338, 0.00000027, -0.00000083, -0.00000515, -0.00000048, 0.00000203, 0.00000132, -0.00000049, 0.00000215, -0.00000138, 0.00000315, -0.00000614, -0.00000135, 0.00000443, -0.00000171, -0.00000085, 0.00000023, -0.00000224, -0.00000003, 0.00000852, -0.00000307, -0.00000270, 0.00000100, -0.00000271, 0.00000090, 0.00000012, },
		{ -0.00000002, -0.00000012, 0.00000118, -0.00000282, 0.00000095, -0.00000246, -0.00000342, 0.00000781, 0.00000142, -0.00000330, 0.00000111, -0.00000157, -0.00000090, 0.00000286, 0.00000089, -0.00000783, 0.00000407, -0.00000255, 0.00000340, -0.00000171, 0.00000259, 0.00000043, 0.00000170, -0.00000660, 0.00000018, -0.00000121, -0.00000164, 0.00000730, 0.00000209, -0.00000802, 0.00000167, -0.00000131, 0.00000078, 0.00000317, 0.00000389, -0.00001173, 0.00000890, -0.00001048, 0.00000672, -0.00000462, 0.00000665, -0.00000818, 0.00001129, -0.00001814, 0.00000356, 0.00000625, -0.00000324, -0.00000013, -0.00000065, -0.00000401, -0.00000282, 0.00000280, -0.00000504, -0.00001003, 0.00001392, -0.00001977, 0.00003837, -0.00007279, 0.00004130, -0.00004812, 0.00004628, -0.00012344, 0.00006675, -0.00061258, 0.00134543, -0.00045999, -0.00006093, -0.00004245, -0.00001878, 0.00000158, -0.00000133, -0.00003304, -0.00000281, 0.00001864, -0.00002237, 0.00002228, -0.00003101, 0.00002440, -0.00002092, 0.00001134, -0.00001336, 0.00001279, -0.00001550, 0.00001437, -0.00000179, -0.00001016, 0.00000044, 0.00000143, -0.00000143, 0.00000341, -0.00000153, -0.00000199, 0.00000009, -0.00000169, -0.00000708, 0.00001124, -0.00000479, 0.00000447, -0.00000412, -0.00000134, -0.00000542, 0.00001278, -0.00000487, 0.00000161, -0.00000183, -0.00000359, -0.00000271, 0.00000361, 0.00000004, 0.00000079, 0.00000082, -0.00000012, 0.00000207, -0.00000428, -0.00000363, 0.00000590, -0.00000245, -0.00000014, -0.00000067, -0.00000116, -0.00000139, 0.00000904, -0.00000261, -0.00000293, 0.00000099, -0.00000253, 0.00000058, 0.00000037, },
		{ -0.00000000, -0.00000035, 0.00000140, -0.00000286, 0.00000084, -0.00000221, -0.00000367, 0.00000694, 0.00000291, -0.00000431, 0.00000192, -0.00000224, -0.00000005, 0.00000126, 0.00000303, -0.00000930, 0.00000481, -0.00000361, 0.00000456, -0.00000284, 0.00000381, -0.00000114, 0.00000377, -0.00000790, 0.00000116, -0.00000279, 0.00000027, 0.00000425, 0.00000593, -0.00001123, 0.00000441, -0.00000412, 0.00000366, -0.00000105, 0.00000926, -0.00001627, 0.00001275, -0.00001418, 0.00001030, -0.00000817, 0.00001037, -0.00001271, 0.00001623, -0.00002157, 0.00000611, 0.00000181, 0.00000326, -0.00000681, 0.00000609, -0.00001229, 0.00000690, -0.00000861, 0.00000842, -0.00002620, 0.00003120, -0.00003738, 0.00005675, -0.00009009, 0.00006002, -0.00007015, 0.00007523, -0.00015871, 0.00012070, -0.00066828, 0.00131857, -0.00036551, -0.00013166, 0.00000076, -0.00005290, 0.00002757, -0.00002378, -0.00001199, -0.00002419, 0.00003802, -0.00004002, 0.00003722, -0.00004254, 0.00003380, -0.00002867, 0.00001791, -0.00001894, 0.00001866, -0.00002094, 0.00001785, -0.00000444, -0.00000580, -0.00000519, 0.00000624, -0.00000556, 0.00000766, -0.00000593, 0.00000252, -0.00000455, 0.00000343, -0.00001232, 0.00001480, -0.00000727, 0.00000724, -0.00000695, 0.00000190, -0.00000885, 0.00001501, -0.00000607, 0.00000279, -0.00000281, -0.00000196, -0.00000493, 0.00000512, -0.00000122, 0.00000208, -0.00000054, 0.00000119, 0.00000088, -0.00000231, -0.00000587, 0.00000722, -0.00000309, 0.00000055, -0.00000158, -0.00000011, -0.00000263, 0.00000938, -0.00000205, -0.00000314, 0.00000091, -0.00000228, 0.00000023, 0.00000062, },
		{ 0.00000002, -0.00000056, 0.00000157, -0.00000283, 0.00000069, -0.00000198, -0.00000380, 0.00000595, 0.00000439, -0.00000522, 0.00000264, -0.00000286, 0.00000079, -0.00000032, 0.00000501, -0.00001049, 0.00000532, -0.00000452, 0.00000558, -0.00000386, 0.00000495, -0.00000264, 0.00000569, -0.00000902, 0.00000209, -0.00000444, 0.00000230, 0.00000112, 0.00000966, -0.00001418, 0.00000691, -0.00000674, 0.00000645, -0.00000519, 0.00001430, -0.00002028, 0.00001604, -0.00001735, 0.00001337, -0.00001128, 0.00001374, -0.00001691, 0.00002070, -0.00002450, 0.00000850, -0.00000272, 0.00000978, -0.00001337, 0.00001277, -0.00002062, 0.00001659, -0.00001986, 0.00002151, -0.00004157, 0.00004705, -0.00005309, 0.00007285, -0.00010488, 0.00007619, -0.00008938, 0.00010067, -0.00018914, 0.00016654, -0.00070934, 0.00127460, -0.00026116, -0.00020460, 0.00004390, -0.00008660, 0.00005312, -0.00004593, 0.00000885, -0.00004513, 0.00005657, -0.00005645, 0.00005059, -0.00005251, 0.00004180, -0.00003520, 0.00002348, -0.00002377, 0.00002393, -0.00002575, 0.00002083, -0.00000698, -0.00000134, -0.00001077, 0.00001089, -0.00000960, 0.00001190, -0.00001033, 0.00000699, -0.00000911, 0.00000838, -0.00001715, 0.00001785, -0.00000940, 0.00000979, -0.00000959, 0.00000492, -0.00001192, 0.00001681, -0.00000694, 0.00000378, -0.00000374, -0.00000030, -0.00000711, 0.00000652, -0.00000242, 0.00000335, -0.00000189, 0.00000251, -0.00000037, -0.00000030, -0.00000799, 0.00000834, -0.00000358, 0.00000118, -0.00000244, 0.00000087, -0.00000372, 0.00000950, -0.00000140, -0.00000331, 0.00000076, -0.00000198, -0.00000015, 0.00000087, },
		{ 0.00000006, -0.00000075, 0.00000167, -0.00000272, 0.00000049, -0.00000176, -0.00000383, 0.00000485, 0.00000583, -0.00000601, 0.00000323, -0.00000340, 0.00000160, -0.00000182, 0.00000676, -0.00001137, 0.00000561, -0.00000528, 0.00000646, -0.00000475, 0.00000597, -0.00000403, 0.00000742, -0.00000993, 0.00000296, -0.00000609, 0.00000436, -0.00000197, 0.00001315, -0.00001676, 0.00000906, -0.00000908, 0.00000904, -0.00000911, 0.00001887, -0.00002361, 0.00001867, -0.00001989, 0.00001587, -0.00001388, 0.00001669, -0.00002067, 0.00002460, -0.00002686, 0.00001063, -0.00000722, 0.00001613, -0.00001958, 0.00001917, -0.00002869, 0.00002590, -0.00003053, 0.00003374, -0.00005560, 0.00006088, -0.00006635, 0.00008617, -0.00011671, 0.00008932, -0.00010524, 0.00012196, -0.00021398, 0.00020353, -0.00073563, 0.00121469, -0.00014904, -0.00027775, 0.00008549, -0.00011877, 0.00007729, -0.00006698, 0.00002873, -0.00006488, 0.00007361, -0.00007106, 0.00006193, -0.00006057, 0.00004815, -0.00004031, 0.00002789, -0.00002774, 0.00002847, -0.00002983, 0.00002325, -0.00000934, 0.00000311, -0.00001615, 0.00001523, -0.00001341, 0.00001599, -0.00001458, 0.00001126, -0.00001343, 0.00001297, -0.00002141, 0.00002029, -0.00001111, 0.00001205, -0.00001197, 0.00000761, -0.00001455, 0.00001813, -0.00000747, 0.00000459, -0.00000461, 0.00000136, -0.00000918, 0.00000776, -0.00000352, 0.00000456, -0.00000320, 0.00000379, -0.00000162, 0.00000168, -0.00000994, 0.00000922, -0.00000392, 0.00000174, -0.00000323, 0.00000176, -0.00000462, 0.00000942, -0.00000068, -0.00000343, 0.00000055, -0.00000162, -0.00000055, 0.00000110, },
		{ 0.00000010, -0.00000091, 0.00000171, -0.00000255, 0.00000027, -0.00000156, -0.00000376, 0.00000369, 0.00000718, -0.00000662, 0.00000368, -0.00000383, 0.00000234, -0.00000321, 0.00000824, -0.00001191, 0.00000565, -0.00000588, 0.00000718, -0.00000547, 0.00000685, -0.00000527, 0.00000891, -0.00001061, 0.00000372, -0.00000772, 0.00000637, -0.00000491, 0.00001628, -0.00001888, 0.00001078, -0.00001107, 0.00001134, -0.00001269, 0.00002283, -0.00002615, 0.00002055, -0.00002176, 0.00001774, -0.00001592, 0.00001915, -0.00002393, 0.00002783, -0.00002859, 0.00001246, -0.00001157, 0.00002213, -0.00002525, 0.00002505, -0.00003623, 0.00003450, -0.00004021, 0.00004467, -0.00006778, 0.00007223, -0.00007676, 0.00009635, -0.00012524, 0.00009911, -0.00011737, 0.00013866, -0.00023274, 0.00023126, -0.00074744, 0.00114042, -0.00003150, -0.00034904, 0.00012406, -0.00014830, 0.00009923, -0.00008616, 0.00004693, -0.00008277, 0.00008851, -0.00008332, 0.00007082, -0.00006645, 0.00005266, -0.00004389, 0.00003107, -0.00003077, 0.00003222, -0.00003311, 0.00002508, -0.00001147, 0.00000743, -0.00002121, 0.00001913, -0.00001689, 0.00001981, -0.00001852, 0.00001517, -0.00001733, 0.00001703, -0.00002496, 0.00002203, -0.00001233, 0.00001394, -0.00001399, 0.00000991, -0.00001666, 0.00001892, -0.00000764, 0.00000521, -0.00000540, 0.00000297, -0.00001109, 0.00000880, -0.00000449, 0.00000569, -0.00000443, 0.00000501, -0.00000283, 0.00000357, -0.00001165, 0.00000981, -0.00000408, 0.00000220, -0.00000393, 0.00000251, -0.00000533, 0.00000915, 0.00000008, -0.00000348, 0.00000027, -0.00000123, -0.00000094, 0.00000132, },
		{ 0.00000016, -0.00000103, 0.00000168, -0.00000232, 0.00000002, -0.00000140, -0.00000360, 0.00000249, 0.00000841, -0.00000705, 0.00000397, -0.00000413, 0.00000298, -0.00000444, 0.00000941, -0.00001210, 0.00000545, -0.00000630, 0.00000772, -0.00000602, 0.00000757, -0.00000632, 0.00001014, -0.00001104, 0.00000437, -0.00000926, 0.00000826, -0.00000761, 0.00001896, -0.00002045, 0.00001203, -0.00001263, 0.00001327, -0.00001581, 0.00002606, -0.00002782, 0.00002164, -0.00002292, 0.00001896, -0.00001739, 0.00002109, -0.00002661, 0.00003036, -0.00002965, 0.00001393, -0.00001566, 0.00002759, -0.00003019, 0.00003020, -0.00004298, 0.00004207, -0.00004857, 0.00005389, -0.00007771, 0.00008072, -0.00008403, 0.00010318, -0.00013032, 0.00010537, -0.00012556, 0.00015052, -0.00024517, 0.00024969, -0.00074547, 0.00105372, 0.00008893, -0.00041639, 0.00015818, -0.00017419, 0.00011811, -0.00010280, 0.00006281, -0.00009821, 0.00010075, -0.00009282, 0.00007700, -0.00006999, 0.00005523, -0.00004587, 0.00003297, -0.00003282, 0.00003512, -0.00003556, 0.00002628, -0.00001333, 0.00001153, -0.00002580, 0.00002249, -0.00001991, 0.00002323, -0.00002202, 0.00001858, -0.00002067, 0.00002044, -0.00002770, 0.00002299, -0.00001302, 0.00001543, -0.00001559, 0.00001174, -0.00001821, 0.00001918, -0.00000746, 0.00000564, -0.00000611, 0.00000450, -0.00001280, 0.00000962, -0.00000530, 0.00000669, -0.00000553, 0.00000610, -0.00000394, 0.00000531, -0.00001306, 0.00001010, -0.00000404, 0.00000255, -0.00000452, 0.00000312, -0.00000583, 0.00000868, 0.00000088, -0.00000347, -0.00000007, -0.00000081, -0.00000133, 0.00000151, },
		{ 0.00000022, -0.00000112, 0.00000159, -0.00000204, -0.00000023, -0.00000127, -0.00000337, 0.00000129, 0.00000947, -0.00000726, 0.00000409, -0.00000431, 0.00000351, -0.00000547, 0.00001026, -0.00001194, 0.00000502, -0.00000658, 0.00000809, -0.00000640, 0.00000812, -0.00000718, 0.00001109, -0.00001120, 0.00000488, -0.00001069, 0.00000995, -0.00000996, 0.00002111, -0.00002141, 0.00001274, -0.00001373, 0.00001476, -0.00001839, 0.00002848, -0.00002860, 0.00002194, -0.00002338, 0.00001953, -0.00001828, 0.00002248, -0.00002868, 0.00003212, -0.00003003, 0.00001500, -0.00001937, 0.00003237, -0.00003422, 0.00003446, -0.00004873, 0.00004834, -0.00005529, 0.00006110, -0.00008508, 0.00008613, -0.00008802, 0.00010658, -0.00013191, 0.00010808, -0.00012976, 0.00015750, -0.00025127, 0.00025905, -0.00073078, 0.00095678, 0.00020963, -0.00047780, 0.00018659, -0.00019554, 0.00013326, -0.00011630, 0.00007584, -0.00011070, 0.00010992, -0.00009924, 0.00008029, -0.00007111, 0.00005586, -0.00004627, 0.00003364, -0.00003391, 0.00003716, -0.00003716, 0.00002686, -0.00001488, 0.00001531, -0.00002981, 0.00002519, -0.00002238, 0.00002614, -0.00002496, 0.00002136, -0.00002334, 0.00002306, -0.00002955, 0.00002316, -0.00001317, 0.00001648, -0.00001675, 0.00001307, -0.00001919, 0.00001892, -0.00000697, 0.00000591, -0.00000673, 0.00000592, -0.00001427, 0.00001017, -0.00000594, 0.00000754, -0.00000649, 0.00000704, -0.00000491, 0.00000683, -0.00001413, 0.00001005, -0.00000382, 0.00000279, -0.00000498, 0.00000356, -0.00000611, 0.00000803, 0.00000168, -0.00000337, -0.00000047, -0.00000037, -0.00000169, 0.00000166, },
		{ 0.00000029, -0.00000118, 0.00000145, -0.00000173, -0.00000048, -0.00000118, -0.00000308, 0.00000012, 0.00001033, -0.00000725, 0.00000404, -0.00000435, 0.00000392, -0.00000629, 0.00001076, -0.00001145, 0.00000440, -0.00000671, 0.00000829, -0.00000659, 0.00000849, -0.00000782, 0.00001174, -0.00001111, 0.00000524, -0.00001194, 0.00001138, -0.00001190, 0.00002264, -0.00002174, 0.00001291, -0.00001435, 0.00001579, -0.00002038, 0.00003006, -0.00002847, 0.00002148, -0.00002317, 0.00001949, -0.00001861, 0.00002334, -0.00003011, 0.00003312, -0.00002973, 0.00001565, -0.00002261, 0.00003632, -0.00003721, 0.00003766, -0.00005329, 0.00005310, -0.00006016, 0.00006608, -0.00008971, 0.00008834, -0.00008872, 0.00010664, -0.00013015, 0.00010738, -0.00013010, 0.00015976, -0.00025130, 0.00025991, -0.00070472, 0.00085203, 0.00032798, -0.00053139, 0.00020820, -0.00021166, 0.00014415, -0.00012623, 0.00008561, -0.00011988, 0.00011574, -0.00010244, 0.00008065, -0.00006988, 0.00005461, -0.00004518, 0.00003314, -0.00003410, 0.00003837, -0.00003795, 0.00002684, -0.00001610, 0.00001868, -0.00003314, 0.00002716, -0.00002423, 0.00002845, -0.00002724, 0.00002340, -0.00002523, 0.00002484, -0.00003046, 0.00002254, -0.00001280, 0.00001710, -0.00001743, 0.00001388, -0.00001959, 0.00001817, -0.00000621, 0.00000603, -0.00000725, 0.00000719, -0.00001547, 0.00001045, -0.00000637, 0.00000821, -0.00000726, 0.00000779, -0.00000570, 0.00000811, -0.00001485, 0.00000969, -0.00000341, 0.00000290, -0.00000530, 0.00000383, -0.00000619, 0.00000725, 0.00000246, -0.00000319, -0.00000091, 0.00000006, -0.00000202, 0.00000178, },
		{ 0.00000037, -0.00000120, 0.00000126, -0.00000138, -0.00000072, -0.00000113, -0.00000274, -0.00000098, 0.00001098, -0.00000702, 0.00000382, -0.00000426, 0.00000419, -0.00000688, 0.00001093, -0.00001065, 0.00000361, -0.00000672, 0.00000833, -0.00000660, 0.00000867, -0.00000823, 0.00001209, -0.00001075, 0.00000545, -0.00001300, 0.00001249, -0.00001337, 0.00002354, -0.00002141, 0.00001254, -0.00001449, 0.00001634, -0.00002175, 0.00003077, -0.00002746, 0.00002031, -0.00002234, 0.00001888, -0.00001842, 0.00002367, -0.00003092, 0.00003336, -0.00002876, 0.00001586, -0.00002530, 0.00003934, -0.00003906, 0.00003972, -0.00005654, 0.00005621, -0.00006304, 0.00006870, -0.00009153, 0.00008740, -0.00008629, 0.00010357, -0.00012529, 0.00010355, -0.00012686, 0.00015761, -0.00024571, 0.00025306, -0.00066892, 0.00074199, 0.00044143, -0.00057551, 0.00022213, -0.00022204, 0.00015040, -0.00013227, 0.00009185, -0.00012553, 0.00011808, -0.00010239, 0.00007817, -0.00006643, 0.00005165, -0.00004275, 0.00003161, -0.00003347, 0.00003881, -0.00003796, 0.00002626, -0.00001697, 0.00002158, -0.00003571, 0.00002834, -0.00002541, 0.00003009, -0.00002877, 0.00002465, -0.00002629, 0.00002572, -0.00003044, 0.00002116, -0.00001194, 0.00001728, -0.00001765, 0.00001418, -0.00001943, 0.00001697, -0.00000521, 0.00000605, -0.00000767, 0.00000830, -0.00001637, 0.00001044, -0.00000661, 0.00000870, -0.00000782, 0.00000833, -0.00000627, 0.00000910, -0.00001518, 0.00000900, -0.00000283, 0.00000291, -0.00000547, 0.00000393, -0.00000607, 0.00000634, 0.00000320, -0.00000291, -0.00000138, 0.00000047, -0.00000230, 0.00000186, },
		{ 0.00000045, -0.00000119, 0.00000102, -0.00000102, -0.00000094, -0.00000112, -0.00000238, -0.00000200, 0.00001139, -0.00000656, 0.00000345, -0.00000405, 0.00000432, -0.00000724, 0.00001077, -0.00000958, 0.00000269, -0.00000663, 0.00000822, -0.00000644, 0.00000866, -0.00000842, 0.00001215, -0.00001014, 0.00000549, -0.00001383, 0.00001324, -0.00001433, 0.00002377, -0.00002046, 0.00001166, -0.00001418, 0.00001642, -0.00002249, 0.00003065, -0.00002566, 0.00001851, -0.00002099, 0.00001779, -0.00001777, 0.00002352, -0.00003113, 0.00003287, -0.00002717, 0.00001564, -0.00002739, 0.00004136, -0.00003972, 0.00004058, -0.00005840, 0.00005760, -0.00006388, 0.00006897, -0.00009058, 0.00008347, -0.00008100, 0.00009771, -0.00011771, 0.00009698, -0.00012046, 0.00015154, -0.00023512, 0.00023950, -0.00062514, 0.00062924, 0.00054759, -0.00060878, 0.00022780, -0.00022638, 0.00015186, -0.00013431, 0.00009444, -0.00012759, 0.00011697, -0.00009920, 0.00007307, -0.00006102, 0.00004722, -0.00003918, 0.00002922, -0.00003214, 0.00003856, -0.00003726, 0.00002516, -0.00001749, 0.00002394, -0.00003746, 0.00002870, -0.00002588, 0.00003103, -0.00002952, 0.00002506, -0.00002649, 0.00002571, -0.00002952, 0.00001909, -0.00001065, 0.00001708, -0.00001743, 0.00001400, -0.00001878, 0.00001540, -0.00000405, 0.00000598, -0.00000800, 0.00000922, -0.00001695, 0.00001013, -0.00000663, 0.00000898, -0.00000817, 0.00000864, -0.00000662, 0.00000979, -0.00001513, 0.00000802, -0.00000210, 0.00000281, -0.00000552, 0.00000386, -0.00000578, 0.00000535, 0.00000386, -0.00000254, -0.00000188, 0.00000085, -0.00000253, 0.00000191, },
		{ 0.00000054, -0.00000115, 0.00000076, -0.00000067, -0.00000112, -0.00000115, -0.00000200, -0.00000291, 0.00001157, -0.00000589, 0.00000295, -0.00000374, 0.00000433, -0.00000738, 0.00001031, -0.00000830, 0.00000169, -0.00000648, 0.00000798, -0.00000612, 0.00000849, -0.00000838, 0.00001192, -0.00000931, 0.00000539, -0.00001441, 0.00001361, -0.00001476, 0.00002337, -0.00001893, 0.00001034, -0.00001346, 0.00001606, -0.00002264, 0.00002975, -0.00002314, 0.00001621, -0.00001921, 0.00001630, -0.00001673, 0.00002293, -0.00003076, 0.00003169, -0.00002502, 0.00001502, -0.00002884, 0.00004232, -0.00003917, 0.00004024, -0.00005887, 0.00005726, -0.00006271, 0.00006695, -0.00008705, 0.00007685, -0.00007326, 0.00008951, -0.00010791, 0.00008816, -0.00011139, 0.00014214, -0.00022033, 0.00022039, -0.00057529, 0.00051627, 0.00064433, -0.00063016, 0.00022492, -0.00022466, 0.00014855, -0.00013238, 0.00009344, -0.00012613, 0.00011255, -0.00009312, 0.00006564, -0.00005395, 0.00004159, -0.00003472, 0.00002616, -0.00003026, 0.00003772, -0.00003595, 0.00002361, -0.00001767, 0.00002573, -0.00003837, 0.00002824, -0.00002564, 0.00003125, -0.00002947, 0.00002463, -0.00002586, 0.00002485, -0.00002778, 0.00001644, -0.00000902, 0.00001654, -0.00001681, 0.00001337, -0.00001767, 0.00001353, -0.00000278, 0.00000586, -0.00000823, 0.00000994, -0.00001722, 0.00000955, -0.00000646, 0.00000906, -0.00000828, 0.00000871, -0.00000672, 0.00001017, -0.00001471, 0.00000677, -0.00000127, 0.00000263, -0.00000543, 0.00000365, -0.00000535, 0.00000431, 0.00000443, -0.00000207, -0.00000238, 0.00000119, -0.00000270, 0.00000190, },
		{ 0.00000063, -0.00000109, 0.00000047, -0.00000032, -0.00000127, -0.00000121, -0.00000163, -0.00000369, 0.00001151, -0.00000504, 0.00000235, -0.00000334, 0.00000422, -0.00000731, 0.00000959, -0.00000684, 0.00000066, -0.00000628, 0.00000763, -0.00000567, 0.00000815, -0.00000814, 0.00001144, -0.00000828, 0.00000514, -0.00001474, 0.00001359, -0.00001467, 0.00002235, -0.00001687, 0.00000864, -0.00001240, 0.00001530, -0.00002224, 0.00002814, -0.00002004, 0.00001352, -0.00001712, 0.00001451, -0.00001538, 0.00002196, -0.00002988, 0.00002990, -0.00002237, 0.00001402, -0.00002964, 0.00004224, -0.00003746, 0.00003875, -0.00005800, 0.00005526, -0.00005966, 0.00006284, -0.00008121, 0.00006793, -0.00006354, 0.00007950, -0.00009642, 0.00007765, -0.00010026, 0.00013010, -0.00020219, 0.00019698, -0.00052127, 0.00040549, 0.00072981, -0.00063896, 0.00021352, -0.00021709, 0.00014069, -0.00012671, 0.00008905, -0.00012138, 0.00010513, -0.00008453, 0.00005631, -0.00004564, 0.00003511, -0.00002965, 0.00002264, -0.00002798, 0.00003640, -0.00003411, 0.00002168, -0.00001752, 0.00002693, -0.00003842, 0.00002696, -0.00002471, 0.00003075, -0.00002863, 0.00002340, -0.00002445, 0.00002321, -0.00002532, 0.00001331, -0.00000715, 0.00001573, -0.00001583, 0.00001236, -0.00001620, 0.00001144, -0.00000148, 0.00000573, -0.00000838, 0.00001045, -0.00001716, 0.00000870, -0.00000610, 0.00000894, -0.00000818, 0.00000854, -0.00000657, 0.00001024, -0.00001395, 0.00000531, -0.00000036, 0.00000238, -0.00000524, 0.00000331, -0.00000479, 0.00000325, 0.00000488, -0.00000151, -0.00000288, 0.00000147, -0.00000281, 0.00000186, },
		{ 0.00000073, -0.00000100, 0.00000017, 0.00000001, -0.00000137, -0.00000130, -0.00000127, -0.00000433, 0.00001122, -0.00000404, 0.00000167, -0.00000287, 0.00000401, -0.00000706, 0.00000864, -0.00000527, -0.00000036, -0.00000608, 0.00000718, -0.00000509, 0.00000768, -0.00000771, 0.00001073, -0.00000708, 0.00000477, -0.00001482, 0.00001318, -0.00001408, 0.00002080, -0.00001440, 0.00000665, -0.00001107, 0.00001422, -0.00002135, 0.00002593, -0.00001649, 0.00001060, -0.00001484, 0.00001252, -0.00001381, 0.00002068, -0.00002854, 0.00002757, -0.00001932, 0.00001269, -0.00002981, 0.00004114, -0.00003467, 0.00003622, -0.00005589, 0.00005175, -0.00005491, 0.00005691, -0.00007341, 0.00005719, -0.00005241, 0.00006826, -0.00008384, 0.00006604, -0.00008769, 0.00011615, -0.00018163, 0.00017056, -0.00046496, 0.00029907, 0.00080254, -0.00063488, 0.00019395, -0.00020411, 0.00012871, -0.00011765, 0.00008160, -0.00011372, 0.00009513, -0.00007387, 0.00004554, -0.00003651, 0.00002813, -0.00002426, 0.00001890, -0.00002546, 0.00003471, -0.00003184, 0.00001944, -0.00001706, 0.00002752, -0.00003763, 0.00002494, -0.00002316, 0.00002959, -0.00002706, 0.00002142, -0.00002233, 0.00002089, -0.00002227, 0.00000986, -0.00000514, 0.00001472, -0.00001458, 0.00001104, -0.00001445, 0.00000922, -0.00000021, 0.00000560, -0.00000843, 0.00001075, -0.00001679, 0.00000760, -0.00000557, 0.00000863, -0.00000785, 0.00000814, -0.00000620, 0.00001003, -0.00001288, 0.00000369, 0.00000057, 0.00000211, -0.00000496, 0.00000286, -0.00000415, 0.00000222, 0.00000519, -0.00000087, -0.00000335, 0.00000168, -0.00000285, 0.00000178, },
		{ 0.00000083, -0.00000090, -0.00000013, 0.00000030, -0.00000143, -0.00000142, -0.00000094, -0.00000483, 0.00001071, -0.00000291, 0.00000096, -0.00000238, 0.00000373, -0.00000664, 0.00000752, -0.00000365, -0.00000132, -0.00000589, 0.00000666, -0.00000441, 0.00000709, -0.00000713, 0.00000982, -0.00000576, 0.00000429, -0.00001466, 0.00001241, -0.00001305, 0.00001877, -0.00001160, 0.00000449, -0.00000957, 0.00001288, -0.00002007, 0.00002325, -0.00001264, 0.00000757, -0.00001248, 0.00001043, -0.00001209, 0.00001916, -0.00002682, 0.00002480, -0.00001596, 0.00001111, -0.00002939, 0.00003909, -0.00003091, 0.00003279, -0.00005270, 0.00004691, -0.00004873, 0.00004949, -0.00006409, 0.00004517, -0.00004046, 0.00005638, -0.00007077, 0.00005393, -0.00007431, 0.00010102, -0.00015958, 0.00014239, -0.00040814, 0.00019894, 0.00086145, -0.00061804, 0.00016687, -0.00018638, 0.00011317, -0.00010572, 0.00007156, -0.00010360, 0.00008304, -0.00006170, 0.00003387, -0.00002700, 0.00002102, -0.00001882, 0.00001514, -0.00002286, 0.00003278, -0.00002924, 0.00001698, -0.00001634, 0.00002753, -0.00003605, 0.00002224, -0.00002104, 0.00002784, -0.00002483, 0.00001881, -0.00001963, 0.00001803, -0.00001879, 0.00000624, -0.00000310, 0.00001360, -0.00001313, 0.00000950, -0.00001250, 0.00000697, 0.00000098, 0.00000552, -0.00000841, 0.00001083, -0.00001612, 0.00000629, -0.00000491, 0.00000816, -0.00000733, 0.00000754, -0.00000561, 0.00000956, -0.00001156, 0.00000195, 0.00000150, 0.00000182, -0.00000460, 0.00000233, -0.00000346, 0.00000123, 0.00000537, -0.00000016, -0.00000377, 0.00000182, -0.00000282, 0.00000166, },
		{ 0.00000093, -0.00000079, -0.00000043, 0.00000055, -0.00000143, -0.00000156, -0.00000064, -0.00000518, 0.00001002, -0.00000170, 0.00000024, -0.00000187, 0.00000339, -0.00000609, 0.00000628, -0.00000204, -0.00000217, -0.00000573, 0.00000608, -0.00000365, 0.00000641, -0.00000641, 0.00000876, -0.00000434, 0.00000374, -0.00001428, 0.00001130, -0.00001163, 0.00001639, -0.00000860, 0.00000225, -0.00000797, 0.00001138, -0.00001848, 0.00002021, -0.00000865, 0.00000459, -0.00001017, 0.00000834, -0.00001032, 0.00001748, -0.00002480, 0.00002168, -0.00001240, 0.00000935, -0.00002843, 0.00003621, -0.00002635, 0.00002865, -0.00004862, 0.00004100, -0.00004144, 0.00004096, -0.00005373, 0.00003244, -0.00002829, 0.00004447, -0.00005778, 0.00004190, -0.00006074, 0.00008545, -0.00013695, 0.00011368, -0.00035242, 0.00010669, 0.00090587, -0.00058893, 0.00013319, -0.00016477, 0.00009480, -0.00009153, 0.00005948, -0.00009157, 0.00006945, -0.00004858, 0.00002183, -0.00001759, 0.00001413, -0.00001362, 0.00001157, -0.00002031, 0.00003070, -0.00002642, 0.00001436, -0.00001538, 0.00002698, -0.00003375, 0.00001896, -0.00001846, 0.00002557, -0.00002204, 0.00001567, -0.00001647, 0.00001477, -0.00001503, 0.00000259, -0.00000116, 0.00001245, -0.00001154, 0.00000781, -0.00001045, 0.00000475, 0.00000202, 0.00000549, -0.00000830, 0.00001070, -0.00001519, 0.00000482, -0.00000414, 0.00000754, -0.00000663, 0.00000676, -0.00000485, 0.00000887, -0.00001004, 0.00000018, 0.00000236, 0.00000154, -0.00000420, 0.00000176, -0.00000276, 0.00000033, 0.00000539, 0.00000060, -0.00000415, 0.00000187, -0.00000273, 0.00000152, },
		{ 0.00000103, -0.00000067, -0.00000070, 0.00000075, -0.00000138, -0.00000172, -0.00000038, -0.00000538, 0.00000916, -0.00000045, -0.00000045, -0.00000138, 0.00000301, -0.00000545, 0.00000497, -0.00000049, -0.00000288, -0.00000563, 0.00000546, -0.00000284, 0.00000566, -0.00000559, 0.00000759, -0.00000289, 0.00000315, -0.00001370, 0.00000992, -0.00000990, 0.00001374, -0.00000551, 0.00000006, -0.00000639, 0.00000980, -0.00001669, 0.00001695, -0.00000467, 0.00000179, -0.00000800, 0.00000633, -0.00000856, 0.00001569, -0.00002256, 0.00001833, -0.00000875, 0.00000748, -0.00002701, 0.00003260, -0.00002118, 0.00002399, -0.00004389, 0.00003429, -0.00003337, 0.00003176, -0.00004283, 0.00001957, -0.00001648, 0.00003306, -0.00004542, 0.00003048, -0.00004754, 0.00007009, -0.00011454, 0.00008551, -0.00029921, 0.00002360, 0.00093558, -0.00054842, 0.00009411, -0.00014027, 0.00007441, -0.00007577, 0.00004599, -0.00007827, 0.00005498, -0.00003511, 0.00000996, -0.00000869, 0.00000778, -0.00000890, 0.00000836, -0.00001794, 0.00002859, -0.00002347, 0.00001165, -0.00001423, 0.00002591, -0.00003081, 0.00001521, -0.00001554, 0.00002290, -0.00001881, 0.00001214, -0.00001302, 0.00001128, -0.00001117, -0.00000093, 0.00000058, 0.00001135, -0.00000990, 0.00000606, -0.00000839, 0.00000267, 0.00000288, 0.00000553, -0.00000811, 0.00001038, -0.00001403, 0.00000321, -0.00000330, 0.00000680, -0.00000579, 0.00000583, -0.00000394, 0.00000801, -0.00000838, -0.00000157, 0.00000312, 0.00000131, -0.00000378, 0.00000116, -0.00000207, -0.00000046, 0.00000527, 0.00000140, -0.00000444, 0.00000183, -0.00000258, 0.00000134, },
		{ 0.00000113, -0.00000055, -0.00000095, 0.00000089, -0.00000129, -0.00000188, -0.00000017, -0.00000545, 0.00000817, 0.00000080, -0.00000107, -0.00000093, 0.00000263, -0.00000475, 0.00000364, 0.00000094, -0.00000342, -0.00000560, 0.00000481, -0.00000200, 0.00000488, -0.00000470, 0.00000634, -0.00000144, 0.00000254, -0.00001297, 0.00000831, -0.00000796, 0.00001095, -0.00000247, -0.00000200, -0.00000489, 0.00000822, -0.00001479, 0.00001361, -0.00000084, -0.00000073, -0.00000606, 0.00000448, -0.00000687, 0.00001388, -0.00002018, 0.00001483, -0.00000511, 0.00000559, -0.00002522, 0.00002843, -0.00001560, 0.00001905, -0.00003873, 0.00002707, -0.00002490, 0.00002229, -0.00003187, 0.00000710, -0.00000556, 0.00002264, -0.00003415, 0.00002012, -0.00003520, 0.00005551, -0.00009310, 0.00005882, -0.00024968, -0.00004945, 0.00095077, -0.00049772, 0.00005100, -0.00011400, 0.00005290, -0.00005919, 0.00003175, -0.00006431, 0.00004026, -0.00002190, -0.00000124, -0.00000070, 0.00000225, -0.00000484, 0.00000564, -0.00001584, 0.00002651, -0.00002046, 0.00000894, -0.00001291, 0.00002437, -0.00002733, 0.00001114, -0.00001238, 0.00001995, -0.00001528, 0.00000839, -0.00000943, 0.00000773, -0.00000738, -0.00000418, 0.00000203, 0.00001035, -0.00000827, 0.00000433, -0.00000639, 0.00000077, 0.00000351, 0.00000564, -0.00000784, 0.00000987, -0.00001266, 0.00000153, -0.00000242, 0.00000598, -0.00000484, 0.00000479, -0.00000293, 0.00000702, -0.00000665, -0.00000325, 0.00000374, 0.00000113, -0.00000336, 0.00000057, -0.00000142, -0.00000113, 0.00000499, 0.00000221, -0.00000465, 0.00000171, -0.00000237, 0.00000115, },
		{ 0.00000123, -0.00000044, -0.00000116, 0.00000097, -0.00000114, -0.00000205, -0.00000001, -0.00000540, 0.00000708, 0.00000200, -0.00000160, -0.00000054, 0.00000226, -0.00000401, 0.00000233, 0.00000222, -0.00000377, -0.00000564, 0.00000415, -0.00000115, 0.00000409, -0.00000379, 0.00000507, -0.00000004, 0.00000195, -0.00001212, 0.00000654, -0.00000588, 0.00000814, 0.00000041, -0.00000381, -0.00000357, 0.00000671, -0.00001287, 0.00001030, 0.00000271, -0.00000286, -0.00000442, 0.00000284, -0.00000531, 0.00001208, -0.00001773, 0.00001131, -0.00000159, 0.00000376, -0.00002315, 0.00002386, -0.00000983, 0.00001406, -0.00003342, 0.00001964, -0.00001638, 0.00001299, -0.00002133, -0.00000447, 0.00000400, 0.00001361, -0.00002433, 0.00001118, -0.00002410, 0.00004221, -0.00007321, 0.00003440, -0.00020473, -0.00011192, 0.00095201, -0.00043832, 0.00000539, -0.00008713, 0.00003117, -0.00004255, 0.00001742, -0.00005034, 0.00002589, -0.00000947, -0.00001131, 0.00000607, -0.00000224, -0.00000160, 0.00000351, -0.00001408, 0.00002453, -0.00001746, 0.00000626, -0.00001149, 0.00002244, -0.00002344, 0.00000688, -0.00000913, 0.00001684, -0.00001158, 0.00000456, -0.00000587, 0.00000427, -0.00000381, -0.00000703, 0.00000311, 0.00000951, -0.00000672, 0.00000267, -0.00000453, -0.00000089, 0.00000391, 0.00000583, -0.00000749, 0.00000919, -0.00001114, -0.00000019, -0.00000155, 0.00000511, -0.00000382, 0.00000369, -0.00000188, 0.00000598, -0.00000491, -0.00000478, 0.00000419, 0.00000104, -0.00000296, 0.00000001, -0.00000084, -0.00000166, 0.00000458, 0.00000303, -0.00000476, 0.00000150, -0.00000211, 0.00000094, },
		{ 0.00000133, -0.00000033, -0.00000133, 0.00000099, -0.00000095, -0.00000222, 0.00000010, -0.00000524, 0.00000593, 0.00000313, -0.00000201, -0.00000023, 0.00000191, -0.00000328, 0.00000109, 0.00000330, -0.00000392, -0.00000575, 0.00000349, -0.00000031, 0.00000333, -0.00000288, 0.00000381, 0.00000128, 0.00000142, -0.00001119, 0.00000468, -0.00000378, 0.00000540, 0.00000304, -0.00000530, -0.00000248, 0.00000533, -0.00001102, 0.00000713, 0.00000588, -0.00000453, -0.00000313, 0.00000145, -0.00000392, 0.00001035, -0.00001529, 0.00000786, 0.00000170, 0.00000208, -0.00002090, 0.00001906, -0.00000409, 0.00000925, -0.00002817, 0.00001230, -0.00000817, 0.00000424, -0.00001163, -0.00001470, 0.00001184, 0.00000628, -0.00001624, 0.00000391, -0.00001454, 0.00003052, -0.00005534, 0.00001281, -0.00016500, -0.00016366, 0.00094025, -0.00037193, -0.00004114, -0.00006082, 0.00001010, -0.00002656, 0.00000365, -0.00003695, 0.00001244, 0.00000167, -0.00001989, 0.00001137, -0.00000555, 0.00000072, 0.00000203, -0.00001268, 0.00002268, -0.00001453, 0.00000369, -0.00000999, 0.00002019, -0.00001926, 0.00000258, -0.00000592, 0.00001372, -0.00000787, 0.00000082, -0.00000248, 0.00000106, -0.00000060, -0.00000938, 0.00000377, 0.00000888, -0.00000528, 0.00000115, -0.00000286, -0.00000227, 0.00000406, 0.00000607, -0.00000705, 0.00000837, -0.00000952, -0.00000188, -0.00000073, 0.00000422, -0.00000276, 0.00000257, -0.00000081, 0.00000492, -0.00000322, -0.00000613, 0.00000443, 0.00000104, -0.00000259, -0.00000050, -0.00000035, -0.00000204, 0.00000405, 0.00000381, -0.00000476, 0.00000120, -0.00000182, 0.00000072, },
		{ 0.00000143, -0.00000023, -0.00000146, 0.00000095, -0.00000073, -0.00000238, 0.00000015, -0.00000498, 0.00000475, 0.00000414, -0.00000227, -0.00000000, 0.00000161, -0.00000257, -0.00000005, 0.00000417, -0.00000387, -0.00000593, 0.00000282, 0.00000050, 0.00000260, -0.00000200, 0.00000261, 0.00000248, 0.00000096, -0.00001022, 0.00000279, -0.00000173, 0.00000285, 0.00000530, -0.00000641, -0.00000166, 0.00000414, -0.00000930, 0.00000420, 0.00000859, -0.00000569, -0.00000221, 0.00000032, -0.00000273, 0.00000874, -0.00001293, 0.00000456, 0.00000469, 0.00000061, -0.00001858, 0.00001419, 0.00000141, 0.00000482, -0.00002323, 0.00000531, -0.00000056, -0.00000363, -0.00000312, -0.00002328, 0.00001769, 0.00000082, -0.00001002, -0.00000155, -0.00000667, 0.00002070, -0.00003979, -0.00000554, -0.00013085, -0.00020484, 0.00091674, -0.00030045, -0.00008694, -0.00003621, -0.00000947, -0.00001188, -0.00000901, -0.00002467, 0.00000039, 0.00001114, -0.00002669, 0.00001504, -0.00000760, 0.00000211, 0.00000121, -0.00001166, 0.00002099, -0.00001171, 0.00000124, -0.00000847, 0.00001769, -0.00001493, -0.00000162, -0.00000288, 0.00001070, -0.00000429, -0.00000270, 0.00000058, -0.00000178, 0.00000214, -0.00001116, 0.00000397, 0.00000847, -0.00000400, -0.00000018, -0.00000142, -0.00000335, 0.00000398, 0.00000636, -0.00000653, 0.00000743, -0.00000782, -0.00000350, 0.00000001, 0.00000335, -0.00000171, 0.00000147, 0.00000021, 0.00000389, -0.00000163, -0.00000725, 0.00000447, 0.00000113, -0.00000226, -0.00000095, 0.00000004, -0.00000227, 0.00000341, 0.00000454, -0.00000464, 0.00000084, -0.00000150, 0.00000051, },
		{ 0.00000152, -0.00000015, -0.00000155, 0.00000085, -0.00000048, -0.00000252, 0.00000016, -0.00000466, 0.00000358, 0.00000500, -0.00000239, 0.00000013, 0.00000135, -0.00000190, -0.00000105, 0.00000480, -0.00000362, -0.00000615, 0.00000215, 0.00000127, 0.00000194, -0.00000118, 0.00000150, 0.00000352, 0.00000061, -0.00000925, 0.00000095, 0.00000018, 0.00000058, 0.00000714, -0.00000709, -0.00000115, 0.00000317, -0.00000776, 0.00000157, 0.00001077, -0.00000634, -0.00000166, -0.00000052, -0.00000173, 0.00000726, -0.00001069, 0.00000149, 0.00000730, -0.00000058, -0.00001628, 0.00000941, 0.00000647, 0.00000096, -0.00001878, -0.00000109, 0.00000617, -0.00001034, 0.00000393, -0.00002995, 0.00002142, -0.00000269, -0.00000572, -0.00000516, -0.00000055, 0.00001285, -0.00002673, -0.00002047, -0.00010240, -0.00023594, 0.00088297, -0.00022584, -0.00013047, -0.00001431, -0.00002684, 0.00000092, -0.00002005, -0.00001395, -0.00000987, 0.00001863, -0.00003153, 0.00001701, -0.00000839, 0.00000257, 0.00000100, -0.00001098, 0.00001945, -0.00000902, -0.00000103, -0.00000696, 0.00001503, -0.00001058, -0.00000557, -0.00000013, 0.00000789, -0.00000095, -0.00000585, 0.00000320, -0.00000414, 0.00000433, -0.00001233, 0.00000371, 0.00000829, -0.00000290, -0.00000131, -0.00000023, -0.00000412, 0.00000366, 0.00000667, -0.00000593, 0.00000639, -0.00000612, -0.00000501, 0.00000064, 0.00000252, -0.00000070, 0.00000044, 0.00000114, 0.00000295, -0.00000020, -0.00000810, 0.00000428, 0.00000131, -0.00000199, -0.00000131, 0.00000032, -0.00000236, 0.00000269, 0.00000520, -0.00000441, 0.00000040, -0.00000116, 0.00000030, },
		{ 0.00000160, -0.00000008, -0.00000159, 0.00000071, -0.00000022, -0.00000264, 0.00000013, -0.00000427, 0.00000244, 0.00000570, -0.00000234, 0.00000017, 0.00000115, -0.00000129, -0.00000191, 0.00000520, -0.00000320, -0.00000642, 0.00000149, 0.00000197, 0.00000137, -0.00000044, 0.00000051, 0.00000438, 0.00000038, -0.00000832, -0.00000079, 0.00000188, -0.00000136, 0.00000851, -0.00000733, -0.00000095, 0.00000242, -0.00000644, -0.00000072, 0.00001241, -0.00000647, -0.00000145, -0.00000112, -0.00000094, 0.00000594, -0.00000862, -0.00000127, 0.00000947, -0.00000144, -0.00001408, 0.00000488, 0.00001094, -0.00000218, -0.00001497, -0.00000672, 0.00001181, -0.00001570, 0.00000933, -0.00003460, 0.00002300, -0.00000431, -0.00000327, -0.00000700, 0.00000385, 0.00000696, -0.00001618, -0.00003194, -0.00007951, -0.00025769, 0.00084060, -0.00015009, -0.00017025, 0.00000399, -0.00004141, 0.00001138, -0.00002911, -0.00000514, -0.00001805, 0.00002395, -0.00003433, 0.00001730, -0.00000798, 0.00000219, 0.00000133, -0.00001059, 0.00001805, -0.00000648, -0.00000311, -0.00000550, 0.00001230, -0.00000633, -0.00000916, 0.00000224, 0.00000541, 0.00000203, -0.00000854, 0.00000528, -0.00000594, 0.00000591, -0.00001286, 0.00000301, 0.00000831, -0.00000197, -0.00000221, 0.00000068, -0.00000459, 0.00000316, 0.00000699, -0.00000524, 0.00000530, -0.00000445, -0.00000636, 0.00000114, 0.00000176, 0.00000023, -0.00000049, 0.00000196, 0.00000213, 0.00000104, -0.00000868, 0.00000387, 0.00000159, -0.00000177, -0.00000159, 0.00000048, -0.00000233, 0.00000192, 0.00000576, -0.00000405, -0.00000008, -0.00000083, 0.00000010, },
		{ 0.00000168, -0.00000002, -0.00000159, 0.00000052, 0.00000005, -0.00000273, 0.00000005, -0.00000384, 0.00000136, 0.00000622, -0.00000214, 0.00000013, 0.00000100, -0.00000076, -0.00000260, 0.00000537, -0.00000262, -0.00000670, 0.00000083, 0.00000261, 0.00000089, 0.00000020, -0.00000034, 0.00000505, 0.00000030, -0.00000746, -0.00000238, 0.00000330, -0.00000290, 0.00000938, -0.00000713, -0.00000105, 0.00000191, -0.00000536, -0.00000262, 0.00001349, -0.00000611, -0.00000155, -0.00000147, -0.00000034, 0.00000480, -0.00000677, -0.00000368, 0.00001115, -0.00000195, -0.00001206, 0.00000072, 0.00001468, -0.00000449, -0.00001191, -0.00001145, 0.00001624, -0.00001957, 0.00001300, -0.00003719, 0.00002250, -0.00000418, -0.00000252, -0.00000723, 0.00000668, 0.00000293, -0.00000803, -0.00004010, -0.00006187, -0.00027104, 0.00079143, -0.00007514, -0.00020498, 0.00001800, -0.00005275, 0.00001920, -0.00003591, 0.00000154, -0.00002397, 0.00002704, -0.00003512, 0.00001602, -0.00000651, 0.00000110, 0.00000210, -0.00001041, 0.00001675, -0.00000409, -0.00000498, -0.00000413, 0.00000957, -0.00000230, -0.00001228, 0.00000413, 0.00000333, 0.00000457, -0.00001067, 0.00000676, -0.00000714, 0.00000686, -0.00001277, 0.00000191, 0.00000851, -0.00000121, -0.00000290, 0.00000133, -0.00000477, 0.00000249, 0.00000727, -0.00000447, 0.00000416, -0.00000285, -0.00000752, 0.00000147, 0.00000110, 0.00000106, -0.00000129, 0.00000262, 0.00000147, 0.00000207, -0.00000897, 0.00000326, 0.00000193, -0.00000160, -0.00000178, 0.00000054, -0.00000217, 0.00000112, 0.00000621, -0.00000357, -0.00000061, -0.00000051, -0.00000007, },
		{ 0.00000175, 0.00000002, -0.00000155, 0.00000030, 0.00000032, -0.00000279, -0.00000007, -0.00000339, 0.00000037, 0.00000654, -0.00000180, 0.00000002, 0.00000090, -0.00000030, -0.00000313, 0.00000533, -0.00000191, -0.00000698, 0.00000017, 0.00000315, 0.00000053, 0.00000072, -0.00000104, 0.00000551, 0.00000035, -0.00000668, -0.00000377, 0.00000442, -0.00000402, 0.00000975, -0.00000652, -0.00000141, 0.00000160, -0.00000451, -0.00000413, 0.00001403, -0.00000532, -0.00000190, -0.00000164, 0.00000009, 0.00000382, -0.00000516, -0.00000571, 0.00001233, -0.00000208, -0.00001027, -0.00000297, 0.00001759, -0.00000591, -0.00000966, -0.00001519, 0.00001937, -0.00002192, 0.00001494, -0.00003782, 0.00002013, -0.00000254, -0.00000322, -0.00000609, 0.00000817, 0.00000057, -0.00000207, -0.00004521, -0.00004900, -0.00027707, 0.00073728, -0.00000279, -0.00023358, 0.00002722, -0.00006060, 0.00002420, -0.00004033, 0.00000599, -0.00002760, 0.00002792, -0.00003405, 0.00001335, -0.00000417, -0.00000054, 0.00000318, -0.00001037, 0.00001553, -0.00000186, -0.00000663, -0.00000287, 0.00000691, 0.00000140, -0.00001483, 0.00000548, 0.00000171, 0.00000659, -0.00001219, 0.00000761, -0.00000773, 0.00000720, -0.00001212, 0.00000047, 0.00000885, -0.00000061, -0.00000336, 0.00000174, -0.00000471, 0.00000171, 0.00000750, -0.00000362, 0.00000303, -0.00000137, -0.00000846, 0.00000163, 0.00000054, 0.00000176, -0.00000192, 0.00000311, 0.00000098, 0.00000286, -0.00000897, 0.00000247, 0.00000233, -0.00000147, -0.00000190, 0.00000050, -0.00000193, 0.00000032, 0.00000654, -0.00000298, -0.00000115, -0.00000021, -0.00000022, },
		{ 0.00000181, 0.00000005, -0.00000148, 0.00000006, 0.00000058, -0.00000281, -0.00000022, -0.00000293, -0.00000052, 0.00000668, -0.00000132, -0.00000016, 0.00000084, 0.00000008, -0.00000349, 0.00000508, -0.00000111, -0.00000723, -0.00000049, 0.00000361, 0.00000029, 0.00000111, -0.00000158, 0.00000576, 0.00000055, -0.00000601, -0.00000494, 0.00000519, -0.00000471, 0.00000965, -0.00000554, -0.00000199, 0.00000148, -0.00000388, -0.00000528, 0.00001407, -0.00000416, -0.00000245, -0.00000166, 0.00000037, 0.00000301, -0.00000380, -0.00000732, 0.00001299, -0.00000183, -0.00000874, -0.00000612, 0.00001961, -0.00000643, -0.00000823, -0.00001794, 0.00002119, -0.00002277, 0.00001524, -0.00003665, 0.00001614, 0.00000029, -0.00000507, -0.00000387, 0.00000858, -0.00000040, 0.00000197, -0.00004764, -0.00004029, -0.00027697, 0.00067996, 0.00006532, -0.00025521, 0.00003140, -0.00006488, 0.00002636, -0.00004238, 0.00000820, -0.00002899, 0.00002675, -0.00003131, 0.00000956, -0.00000119, -0.00000255, 0.00000442, -0.00001037, 0.00001433, 0.00000022, -0.00000804, -0.00000175, 0.00000440, 0.00000468, -0.00001676, 0.00000626, 0.00000058, 0.00000807, -0.00001309, 0.00000782, -0.00000772, 0.00000696, -0.00001097, -0.00000122, 0.00000925, -0.00000012, -0.00000364, 0.00000192, -0.00000444, 0.00000086, 0.00000765, -0.00000270, 0.00000192, -0.00000004, -0.00000916, 0.00000161, 0.00000010, 0.00000233, -0.00000239, 0.00000341, 0.00000067, 0.00000342, -0.00000870, 0.00000153, 0.00000276, -0.00000137, -0.00000194, 0.00000038, -0.00000161, -0.00000046, 0.00000672, -0.00000230, -0.00000170, 0.00000006, -0.00000034, },
		{ 0.00000187, 0.00000006, -0.00000139, -0.00000020, 0.00000081, -0.00000280, -0.00000039, -0.00000248, -0.00000130, 0.00000662, -0.00000074, -0.00000036, 0.00000081, 0.00000039, -0.00000370, 0.00000467, -0.00000025, -0.00000743, -0.00000115, 0.00000397, 0.00000016, 0.00000138, -0.00000195, 0.00000582, 0.00000088, -0.00000545, -0.00000587, 0.00000562, -0.00000499, 0.00000911, -0.00000426, -0.00000272, 0.00000149, -0.00000345, -0.00000608, 0.00001368, -0.00000273, -0.00000310, -0.00000158, 0.00000054, 0.00000237, -0.00000271, -0.00000852, 0.00001316, -0.00000124, -0.00000749, -0.00000869, 0.00002073, -0.00000606, -0.00000759, -0.00001971, 0.00002177, -0.00002225, 0.00001408, -0.00003395, 0.00001089, 0.00000395, -0.00000773, -0.00000091, 0.00000823, -0.00000027, 0.00000444, -0.00004782, -0.00003508, -0.00027196, 0.00062117, 0.00012777, -0.00026931, 0.00003050, -0.00006572, 0.00002581, -0.00004216, 0.00000832, -0.00002831, 0.00002377, -0.00002721, 0.00000494, 0.00000217, -0.00000473, 0.00000568, -0.00001032, 0.00001313, 0.00000216, -0.00000922, -0.00000079, 0.00000210, 0.00000747, -0.00001801, 0.00000646, -0.00000004, 0.00000899, -0.00001336, 0.00000743, -0.00000718, 0.00000622, -0.00000941, -0.00000307, 0.00000968, 0.00000027, -0.00000376, 0.00000192, -0.00000400, -0.00000002, 0.00000769, -0.00000172, 0.00000087, 0.00000111, -0.00000961, 0.00000141, -0.00000022, 0.00000274, -0.00000267, 0.00000352, 0.00000054, 0.00000375, -0.00000819, 0.00000048, 0.00000318, -0.00000129, -0.00000191, 0.00000019, -0.00000125, -0.00000118, 0.00000676, -0.00000153, -0.00000223, 0.00000027, -0.00000043, },
		{ 0.00000191, 0.00000007, -0.00000128, -0.00000045, 0.00000100, -0.00000274, -0.00000059, -0.00000204, -0.00000196, 0.00000640, -0.00000007, -0.00000058, 0.00000081, 0.00000063, -0.00000377, 0.00000412, 0.00000062, -0.00000756, -0.00000182, 0.00000423, 0.00000016, 0.00000154, -0.00000216, 0.00000568, 0.00000133, -0.00000501, -0.00000656, 0.00000572, -0.00000488, 0.00000820, -0.00000277, -0.00000354, 0.00000158, -0.00000318, -0.00000657, 0.00001293, -0.00000110, -0.00000380, -0.00000145, 0.00000061, 0.00000187, -0.00000187, -0.00000930, 0.00001285, -0.00000032, -0.00000652, -0.00001066, 0.00002098, -0.00000489, -0.00000766, -0.00002059, 0.00002122, -0.00002052, 0.00001167, -0.00003002, 0.00000476, 0.00000805, -0.00001085, 0.00000246, 0.00000744, 0.00000063, 0.00000572, -0.00004623, -0.00003263, -0.00026325, 0.00056250, 0.00018342, -0.00027562, 0.00002473, -0.00006338, 0.00002282, -0.00003992, 0.00000657, -0.00002583, 0.00001929, -0.00002210, -0.00000018, 0.00000564, -0.00000686, 0.00000681, -0.00001013, 0.00001189, 0.00000396, -0.00001017, -0.00000002, 0.00000006, 0.00000973, -0.00001859, 0.00000610, -0.00000018, 0.00000937, -0.00001303, 0.00000649, -0.00000618, 0.00000507, -0.00000755, -0.00000498, 0.00001005, 0.00000062, -0.00000376, 0.00000178, -0.00000346, -0.00000087, 0.00000761, -0.00000069, -0.00000010, 0.00000206, -0.00000980, 0.00000105, -0.00000043, 0.00000300, -0.00000277, 0.00000345, 0.00000058, 0.00000387, -0.00000747, -0.00000065, 0.00000358, -0.00000121, -0.00000184, -0.00000005, -0.00000087, -0.00000182, 0.00000665, -0.00000069, -0.00000272, 0.00000043, -0.00000049, },
		{ 0.00000195, 0.00000008, -0.00000115, -0.00000069, 0.00000116, -0.00000264, -0.00000081, -0.00000163, -0.00000249, 0.00000602, 0.00000064, -0.00000079, 0.00000080, 0.00000082, -0.00000371, 0.00000347, 0.00000147, -0.00000760, -0.00000248, 0.00000439, 0.00000027, 0.00000158, -0.00000222, 0.00000539, 0.00000188, -0.00000466, -0.00000700, 0.00000550, -0.00000444, 0.00000700, -0.00000114, -0.00000437, 0.00000171, -0.00000303, -0.00000681, 0.00001188, 0.00000063, -0.00000447, -0.00000131, 0.00000063, 0.00000150, -0.00000129, -0.00000969, 0.00001212, 0.00000086, -0.00000579, -0.00001204, 0.00002041, -0.00000302, -0.00000830, -0.00002070, 0.00001968, -0.00001779, 0.00000832, -0.00002522, -0.00000184, 0.00001220, -0.00001408, 0.00000590, 0.00000652, 0.00000198, 0.00000618, -0.00004335, -0.00003224, -0.00025200, 0.00050534, 0.00023142, -0.00027414, 0.00001451, -0.00005829, 0.00001774, -0.00003599, 0.00000326, -0.00002188, 0.00001369, -0.00001635, -0.00000545, 0.00000895, -0.00000876, 0.00000769, -0.00000974, 0.00001058, 0.00000562, -0.00001088, 0.00000057, -0.00000170, 0.00001141, -0.00001850, 0.00000520, 0.00000012, 0.00000926, -0.00001218, 0.00000508, -0.00000482, 0.00000362, -0.00000551, -0.00000684, 0.00001030, 0.00000096, -0.00000367, 0.00000154, -0.00000284, -0.00000166, 0.00000740, 0.00000036, -0.00000096, 0.00000279, -0.00000975, 0.00000053, -0.00000053, 0.00000311, -0.00000270, 0.00000321, 0.00000078, 0.00000381, -0.00000659, -0.00000180, 0.00000393, -0.00000112, -0.00000173, -0.00000032, -0.00000050, -0.00000237, 0.00000640, 0.00000020, -0.00000317, 0.00000053, -0.00000052, },
		{ 0.00000198, 0.00000009, -0.00000103, -0.00000091, 0.00000126, -0.00000249, -0.00000103, -0.00000125, -0.00000291, 0.00000550, 0.00000139, -0.00000098, 0.00000079, 0.00000096, -0.00000355, 0.00000275, 0.00000227, -0.00000754, -0.00000313, 0.00000444, 0.00000049, 0.00000153, -0.00000216, 0.00000495, 0.00000250, -0.00000441, -0.00000723, 0.00000501, -0.00000374, 0.00000560, 0.00000053, -0.00000514, 0.00000182, -0.00000296, -0.00000685, 0.00001063, 0.00000236, -0.00000504, -0.00000121, 0.00000061, 0.00000125, -0.00000094, -0.00000973, 0.00001103, 0.00000224, -0.00000529, -0.00001288, 0.00001912, -0.00000059, -0.00000939, -0.00002019, 0.00001737, -0.00001432, 0.00000432, -0.00001991, -0.00000850, 0.00001603, -0.00001711, 0.00000912, 0.00000576, 0.00000349, 0.00000618, -0.00003965, -0.00003324, -0.00023924, 0.00045085, 0.00027123, -0.00026518, 0.00000047, -0.00005101, 0.00001105, -0.00003076, -0.00000125, -0.00001685, 0.00000737, -0.00001036, -0.00001054, 0.00001186, -0.00001027, 0.00000820, -0.00000909, 0.00000919, 0.00000715, -0.00001136, 0.00000095, -0.00000314, 0.00001252, -0.00001778, 0.00000385, 0.00000081, 0.00000872, -0.00001088, 0.00000332, -0.00000320, 0.00000198, -0.00000342, -0.00000857, 0.00001038, 0.00000132, -0.00000352, 0.00000123, -0.00000221, -0.00000236, 0.00000706, 0.00000143, -0.00000168, 0.00000329, -0.00000947, -0.00000011, -0.00000055, 0.00000308, -0.00000247, 0.00000282, 0.00000110, 0.00000360, -0.00000558, -0.00000293, 0.00000418, -0.00000101, -0.00000160, -0.00000060, -0.00000015, -0.00000281, 0.00000600, 0.00000111, -0.00000354, 0.00000056, -0.00000051, },
		{ 0.00000199, 0.00000011, -0.00000091, -0.00000110, 0.00000131, -0.00000231, -0.00000127, -0.00000091, -0.00000320, 0.00000488, 0.00000212, -0.00000111, 0.00000077, 0.00000105, -0.00000330, 0.00000200, 0.00000298, -0.00000737, -0.00000378, 0.00000439, 0.00000080, 0.00000139, -0.00000199, 0.00000440, 0.00000317, -0.00000423, -0.00000726, 0.00000430, -0.00000283, 0.00000407, 0.00000214, -0.00000578, 0.00000186, -0.00000292, -0.00000674, 0.00000925, 0.00000402, -0.00000546, -0.00000118, 0.00000057, 0.00000110, -0.00000081, -0.00000945, 0.00000964, 0.00000375, -0.00000494, -0.00001323, 0.00001724, 0.00000223, -0.00001075, -0.00001923, 0.00001448, -0.00001038, -0.00000001, -0.00001447, -0.00001483, 0.00001921, -0.00001968, 0.00001184, 0.00000540, 0.00000491, 0.00000602, -0.00003553, -0.00003501, -0.00022590, 0.00039999, 0.00030259, -0.00024931, -0.00001662, -0.00004218, 0.00000326, -0.00002466, -0.00000655, -0.00001114, 0.00000075, -0.00000452, -0.00001515, 0.00001416, -0.00001125, 0.00000828, -0.00000815, 0.00000773, 0.00000854, -0.00001161, 0.00000114, -0.00000426, 0.00001307, -0.00001648, 0.00000211, 0.00000180, 0.00000783, -0.00000924, 0.00000132, -0.00000146, 0.00000028, -0.00000138, -0.00001005, 0.00001024, 0.00000174, -0.00000336, 0.00000091, -0.00000161, -0.00000293, 0.00000658, 0.00000248, -0.00000224, 0.00000357, -0.00000897, -0.00000085, -0.00000049, 0.00000294, -0.00000211, 0.00000232, 0.00000152, 0.00000327, -0.00000449, -0.00000400, 0.00000433, -0.00000086, -0.00000147, -0.00000086, 0.00000016, -0.00000313, 0.00000548, 0.00000203, -0.00000383, 0.00000053, -0.00000047, },
		{ 0.00000200, 0.00000014, -0.00000081, -0.00000125, 0.00000131, -0.00000209, -0.00000150, -0.00000060, -0.00000339, 0.00000416, 0.00000282, -0.00000117, 0.00000072, 0.00000112, -0.00000299, 0.00000125, 0.00000359, -0.00000710, -0.00000440, 0.00000423, 0.00000118, 0.00000120, -0.00000173, 0.00000378, 0.00000384, -0.00000409, -0.00000713, 0.00000343, -0.00000181, 0.00000253, 0.00000362, -0.00000623, 0.00000180, -0.00000288, -0.00000654, 0.00000782, 0.00000553, -0.00000568, -0.00000123, 0.00000053, 0.00000104, -0.00000086, -0.00000891, 0.00000803, 0.00000531, -0.00000470, -0.00001318, 0.00001488, 0.00000527, -0.00001221, -0.00001800, 0.00001125, -0.00000624, -0.00000436, -0.00000924, -0.00002049, 0.00002148, -0.00002156, 0.00001389, 0.00000563, 0.00000601, 0.00000599, -0.00003137, -0.00003704, -0.00021274, 0.00035343, 0.00032556, -0.00022730, -0.00003587, -0.00003251, -0.00000508, -0.00001818, -0.00001221, -0.00000516, -0.00000576, 0.00000083, -0.00001900, 0.00001567, -0.00001160, 0.00000788, -0.00000693, 0.00000621, 0.00000977, -0.00001164, 0.00000112, -0.00000505, 0.00001309, -0.00001470, 0.00000008, 0.00000300, 0.00000668, -0.00000735, -0.00000080, 0.00000029, -0.00000136, 0.00000050, -0.00001123, 0.00000984, 0.00000222, -0.00000320, 0.00000060, -0.00000106, -0.00000335, 0.00000597, 0.00000349, -0.00000264, 0.00000363, -0.00000830, -0.00000165, -0.00000038, 0.00000269, -0.00000164, 0.00000172, 0.00000200, 0.00000286, -0.00000338, -0.00000496, 0.00000435, -0.00000066, -0.00000134, -0.00000111, 0.00000040, -0.00000332, 0.00000485, 0.00000292, -0.00000403, 0.00000042, -0.00000041, },
		{ 0.00000199, 0.00000018, -0.00000072, -0.00000136, 0.00000125, -0.00000184, -0.00000173, -0.00000034, -0.00000349, 0.00000339, 0.00000345, -0.00000115, 0.00000063, 0.00000117, -0.00000265, 0.00000053, 0.00000407, -0.00000671, -0.00000500, 0.00000398, 0.00000163, 0.00000097, -0.00000143, 0.00000311, 0.00000450, -0.00000398, -0.00000687, 0.00000245, -0.00000075, 0.00000104, 0.00000489, -0.00000644, 0.00000161, -0.00000281, -0.00000629, 0.00000641, 0.00000684, -0.00000569, -0.00000138, 0.00000051, 0.00000105, -0.00000107, -0.00000816, 0.00000630, 0.00000684, -0.00000450, -0.00001281, 0.00001221, 0.00000835, -0.00001361, -0.00001666, 0.00000788, -0.00000217, -0.00000843, -0.00000453, -0.00002521, 0.00002264, -0.00002264, 0.00001513, 0.00000658, 0.00000667, 0.00000627, -0.00002743, -0.00003891, -0.00020035, 0.00031161, 0.00034044, -0.00020013, -0.00005629, -0.00002272, -0.00001342, -0.00001174, -0.00001785, 0.00000068, -0.00001179, 0.00000537, -0.00002188, 0.00001629, -0.00001128, 0.00000699, -0.00000543, 0.00000466, 0.00001084, -0.00001144, 0.00000092, -0.00000553, 0.00001262, -0.00001252, -0.00000212, 0.00000433, 0.00000537, -0.00000535, -0.00000292, 0.00000193, -0.00000283, 0.00000212, -0.00001205, 0.00000917, 0.00000279, -0.00000306, 0.00000033, -0.00000059, -0.00000362, 0.00000525, 0.00000444, -0.00000286, 0.00000349, -0.00000748, -0.00000248, -0.00000025, 0.00000237, -0.00000108, 0.00000108, 0.00000250, 0.00000243, -0.00000229, -0.00000578, 0.00000423, -0.00000041, -0.00000123, -0.00000131, 0.00000057, -0.00000340, 0.00000413, 0.00000376, -0.00000412, 0.00000025, -0.00000033, },
		{ 0.00000197, 0.00000024, -0.00000066, -0.00000143, 0.00000114, -0.00000156, -0.00000194, -0.00000012, -0.00000349, 0.00000259, 0.00000400, -0.00000104, 0.00000052, 0.00000121, -0.00000228, -0.00000014, 0.00000441, -0.00000623, -0.00000556, 0.00000364, 0.00000211, 0.00000074, -0.00000109, 0.00000243, 0.00000511, -0.00000386, -0.00000653, 0.00000143, 0.00000028, -0.00000030, 0.00000589, -0.00000639, 0.00000127, -0.00000270, -0.00000602, 0.00000507, 0.00000790, -0.00000546, -0.00000162, 0.00000049, 0.00000111, -0.00000142, -0.00000726, 0.00000451, 0.00000825, -0.00000427, -0.00001222, 0.00000938, 0.00001130, -0.00001479, -0.00001539, 0.00000458, 0.00000161, -0.00001198, -0.00000057, -0.00002877, 0.00002258, -0.00002285, 0.00001549, 0.00000830, 0.00000681, 0.00000699, -0.00002392, -0.00004029, -0.00018917, 0.00027475, 0.00034778, -0.00016889, -0.00007689, -0.00001350, -0.00002124, -0.00000577, -0.00002308, 0.00000603, -0.00001702, 0.00000885, -0.00002364, 0.00001595, -0.00001028, 0.00000564, -0.00000371, 0.00000312, 0.00001171, -0.00001103, 0.00000055, -0.00000573, 0.00001173, -0.00001005, -0.00000439, 0.00000567, 0.00000401, -0.00000333, -0.00000491, 0.00000334, -0.00000406, 0.00000341, -0.00001246, 0.00000823, 0.00000344, -0.00000296, 0.00000011, -0.00000022, -0.00000373, 0.00000444, 0.00000531, -0.00000290, 0.00000316, -0.00000656, -0.00000331, -0.00000012, 0.00000199, -0.00000048, 0.00000042, 0.00000299, 0.00000199, -0.00000126, -0.00000644, 0.00000395, -0.00000012, -0.00000114, -0.00000147, 0.00000066, -0.00000336, 0.00000334, 0.00000454, -0.00000409, 0.00000001, -0.00000023, },
		{ 0.00000194, 0.00000031, -0.00000063, -0.00000144, 0.00000097, -0.00000126, -0.00000215, 0.00000006, -0.00000343, 0.00000178, 0.00000445, -0.00000084, 0.00000037, 0.00000123, -0.00000190, -0.00000074, 0.00000461, -0.00000566, -0.00000607, 0.00000321, 0.00000261, 0.00000051, -0.00000076, 0.00000177, 0.00000564, -0.00000371, -0.00000615, 0.00000042, 0.00000120, -0.00000145, 0.00000658, -0.00000605, 0.00000078, -0.00000253, -0.00000577, 0.00000384, 0.00000869, -0.00000500, -0.00000195, 0.00000049, 0.00000123, -0.00000185, -0.00000628, 0.00000276, 0.00000948, -0.00000396, -0.00001149, 0.00000652, 0.00001396, -0.00001561, -0.00001431, 0.00000151, 0.00000490, -0.00001482, 0.00000243, -0.00003106, 0.00002128, -0.00002220, 0.00001498, 0.00001079, 0.00000640, 0.00000824, -0.00002095, -0.00004100, -0.00017948, 0.00024284, 0.00034832, -0.00013478, -0.00009666, -0.00000550, -0.00002810, -0.00000063, -0.00002760, 0.00001059, -0.00002119, 0.00001111, -0.00002422, 0.00001467, -0.00000865, 0.00000388, -0.00000184, 0.00000165, 0.00001237, -0.00001041, 0.00000003, -0.00000567, 0.00001050, -0.00000740, -0.00000662, 0.00000694, 0.00000269, -0.00000142, -0.00000667, 0.00000445, -0.00000496, 0.00000432, -0.00001246, 0.00000703, 0.00000415, -0.00000290, -0.00000003, 0.00000005, -0.00000370, 0.00000356, 0.00000606, -0.00000275, 0.00000268, -0.00000556, -0.00000410, -0.00000002, 0.00000158, 0.00000013, -0.00000022, 0.00000343, 0.00000160, -0.00000031, -0.00000690, 0.00000352, 0.00000023, -0.00000108, -0.00000157, 0.00000068, -0.00000322, 0.00000251, 0.00000523, -0.00000395, -0.00000027, -0.00000012, },
		{ 0.00000191, 0.00000042, -0.00000063, -0.00000141, 0.00000077, -0.00000096, -0.00000233, 0.00000021, -0.00000330, 0.00000099, 0.00000477, -0.00000054, 0.00000020, 0.00000125, -0.00000152, -0.00000125, 0.00000467, -0.00000502, -0.00000652, 0.00000271, 0.00000310, 0.00000032, -0.00000044, 0.00000117, 0.00000607, -0.00000352, -0.00000575, -0.00000052, 0.00000197, -0.00000234, 0.00000693, -0.00000544, 0.00000016, -0.00000232, -0.00000555, 0.00000276, 0.00000919, -0.00000433, -0.00000235, 0.00000048, 0.00000139, -0.00000235, -0.00000527, 0.00000111, 0.00001045, -0.00000350, -0.00001072, 0.00000377, 0.00001620, -0.00001597, -0.00001353, -0.00000118, 0.00000755, -0.00001680, 0.00000439, -0.00003202, 0.00001878, -0.00002080, 0.00001367, 0.00001398, 0.00000549, 0.00000999, -0.00001857, -0.00004093, -0.00017139, 0.00021567, 0.00034294, -0.00009902, -0.00011469, 0.00000073, -0.00003362, 0.00000339, -0.00003115, 0.00001414, -0.00002412, 0.00001204, -0.00002360, 0.00001250, -0.00000647, 0.00000181, 0.00000012, 0.00000030, 0.00001280, -0.00000960, -0.00000062, -0.00000540, 0.00000901, -0.00000469, -0.00000869, 0.00000804, 0.00000150, 0.00000030, -0.00000811, 0.00000517, -0.00000549, 0.00000484, -0.00001204, 0.00000559, 0.00000490, -0.00000288, -0.00000011, 0.00000021, -0.00000353, 0.00000266, 0.00000667, -0.00000242, 0.00000207, -0.00000455, -0.00000481, 0.00000004, 0.00000117, 0.00000074, -0.00000080, 0.00000380, 0.00000128, 0.00000050, -0.00000716, 0.00000295, 0.00000062, -0.00000104, -0.00000161, 0.00000061, -0.00000299, 0.00000166, 0.00000580, -0.00000369, -0.00000060, -0.00000001, },
		{ 0.00000186, 0.00000054, -0.00000065, -0.00000134, 0.00000053, -0.00000066, -0.00000249, 0.00000031, -0.00000312, 0.00000022, 0.00000497, -0.00000016, 0.00000001, 0.00000126, -0.00000116, -0.00000166, 0.00000460, -0.00000433, -0.00000690, 0.00000214, 0.00000357, 0.00000019, -0.00000017, 0.00000063, 0.00000638, -0.00000325, -0.00000538, -0.00000134, 0.00000254, -0.00000296, 0.00000693, -0.00000457, -0.00000058, -0.00000206, -0.00000537, 0.00000183, 0.00000941, -0.00000347, -0.00000278, 0.00000047, 0.00000157, -0.00000289, -0.00000429, -0.00000036, 0.00001111, -0.00000286, -0.00000996, 0.00000124, 0.00001792, -0.00001581, -0.00001312, -0.00000341, 0.00000946, -0.00001785, 0.00000525, -0.00003170, 0.00001523, -0.00001879, 0.00001167, 0.00001777, 0.00000417, 0.00001221, -0.00001676, -0.00004006, -0.00016491, 0.00019292, 0.00033262, -0.00006281, -0.00013013, 0.00000474, -0.00003754, 0.00000611, -0.00003357, 0.00001652, -0.00002571, 0.00001163, -0.00002186, 0.00000956, -0.00000387, -0.00000045, 0.00000205, -0.00000087, 0.00001299, -0.00000862, -0.00000136, -0.00000497, 0.00000734, -0.00000202, -0.00001053, 0.00000891, 0.00000051, 0.00000175, -0.00000916, 0.00000547, -0.00000562, 0.00000494, -0.00001125, 0.00000398, 0.00000567, -0.00000288, -0.00000013, 0.00000027, -0.00000325, 0.00000175, 0.00000714, -0.00000193, 0.00000137, -0.00000355, -0.00000542, 0.00000004, 0.00000077, 0.00000130, -0.00000131, 0.00000406, 0.00000104, 0.00000118, -0.00000721, 0.00000225, 0.00000104, -0.00000102, -0.00000160, 0.00000048, -0.00000269, 0.00000082, 0.00000626, -0.00000332, -0.00000096, 0.00000010, },
		{ 0.00000180, 0.00000069, -0.00000070, -0.00000123, 0.00000026, -0.00000036, -0.00000262, 0.00000038, -0.00000291, -0.00000050, 0.00000505, 0.00000029, -0.00000019, 0.00000126, -0.00000082, -0.00000198, 0.00000440, -0.00000361, -0.00000719, 0.00000153, 0.00000398, 0.00000013, 0.00000005, 0.00000018, 0.00000656, -0.00000291, -0.00000506, -0.00000201, 0.00000288, -0.00000328, 0.00000661, -0.00000348, -0.00000139, -0.00000178, -0.00000523, 0.00000106, 0.00000938, -0.00000247, -0.00000322, 0.00000044, 0.00000178, -0.00000342, -0.00000339, -0.00000159, 0.00001142, -0.00000200, -0.00000930, -0.00000098, 0.00001905, -0.00001508, -0.00001311, -0.00000513, 0.00001058, -0.00001795, 0.00000506, -0.00003021, 0.00001082, -0.00001635, 0.00000913, 0.00002200, 0.00000258, 0.00001479, -0.00001542, -0.00003848, -0.00015992, 0.00017411, 0.00031840, -0.00002730, -0.00014231, 0.00000624, -0.00003970, 0.00000741, -0.00003477, 0.00001767, -0.00002593, 0.00000994, -0.00001911, 0.00000602, -0.00000100, -0.00000279, 0.00000386, -0.00000180, 0.00001291, -0.00000748, -0.00000216, -0.00000443, 0.00000557, 0.00000051, -0.00001205, 0.00000947, -0.00000021, 0.00000287, -0.00000978, 0.00000533, -0.00000537, 0.00000467, -0.00001013, 0.00000224, 0.00000642, -0.00000289, -0.00000010, 0.00000025, -0.00000289, 0.00000087, 0.00000745, -0.00000128, 0.00000062, -0.00000261, -0.00000590, -0.00000003, 0.00000040, 0.00000180, -0.00000171, 0.00000420, 0.00000091, 0.00000171, -0.00000706, 0.00000143, 0.00000147, -0.00000102, -0.00000154, 0.00000029, -0.00000233, 0.00000001, 0.00000657, -0.00000285, -0.00000135, 0.00000020, },
		{ 0.00000173, 0.00000086, -0.00000078, -0.00000109, -0.00000002, -0.00000009, -0.00000271, 0.00000042, -0.00000266, -0.00000116, 0.00000499, 0.00000080, -0.00000038, 0.00000124, -0.00000050, -0.00000219, 0.00000410, -0.00000290, -0.00000740, 0.00000087, 0.00000434, 0.00000016, 0.00000019, -0.00000017, 0.00000660, -0.00000249, -0.00000481, -0.00000251, 0.00000298, -0.00000331, 0.00000599, -0.00000222, -0.00000222, -0.00000151, -0.00000512, 0.00000044, 0.00000910, -0.00000137, -0.00000364, 0.00000038, 0.00000200, -0.00000392, -0.00000262, -0.00000254, 0.00001136, -0.00000092, -0.00000875, -0.00000284, 0.00001955, -0.00001380, -0.00001348, -0.00000632, 0.00001092, -0.00001716, 0.00000391, -0.00002773, 0.00000580, -0.00001372, 0.00000623, 0.00002649, 0.00000086, 0.00001760, -0.00001444, -0.00003630, -0.00015623, 0.00015871, 0.00030132, 0.00000649, -0.00015070, 0.00000508, -0.00004007, 0.00000729, -0.00003475, 0.00001759, -0.00002486, 0.00000711, -0.00001556, 0.00000205, 0.00000199, -0.00000508, 0.00000547, -0.00000244, 0.00001258, -0.00000621, -0.00000298, -0.00000383, 0.00000380, 0.00000281, -0.00001318, 0.00000970, -0.00000064, 0.00000361, -0.00000995, 0.00000476, -0.00000475, 0.00000407, -0.00000875, 0.00000045, 0.00000709, -0.00000288, -0.00000002, 0.00000016, -0.00000247, 0.00000005, 0.00000758, -0.00000051, -0.00000015, -0.00000176, -0.00000623, -0.00000019, 0.00000008, 0.00000221, -0.00000199, 0.00000421, 0.00000089, 0.00000208, -0.00000672, 0.00000052, 0.00000189, -0.00000103, -0.00000144, 0.00000005, -0.00000195, -0.00000074, 0.00000674, -0.00000228, -0.00000173, 0.00000028, },
		{ 0.00000166, 0.00000106, -0.00000087, -0.00000093, -0.00000030, 0.00000015, -0.00000277, 0.00000042, -0.00000240, -0.00000175, 0.00000482, 0.00000136, -0.00000056, 0.00000120, -0.00000021, -0.00000232, 0.00000371, -0.00000219, -0.00000751, 0.00000019, 0.00000461, 0.00000028, 0.00000026, -0.00000041, 0.00000650, -0.00000199, -0.00000464, -0.00000282, 0.00000284, -0.00000308, 0.00000512, -0.00000085, -0.00000303, -0.00000128, -0.00000502, -0.00000003, 0.00000864, -0.00000021, -0.00000400, 0.00000029, 0.00000222, -0.00000436, -0.00000201, -0.00000318, 0.00001092, 0.00000038, -0.00000835, -0.00000428, 0.00001943, -0.00001201, -0.00001421, -0.00000702, 0.00001053, -0.00001556, 0.00000195, -0.00002448, 0.00000044, -0.00001114, 0.00000314, 0.00003108, -0.00000082, 0.00002049, -0.00001367, -0.00003369, -0.00015360, 0.00014615, 0.00028236, 0.00003766, -0.00015493, 0.00000126, -0.00003873, 0.00000584, -0.00003359, 0.00001638, -0.00002260, 0.00000331, -0.00001141, -0.00000211, 0.00000491, -0.00000717, 0.00000677, -0.00000276, 0.00001200, -0.00000485, -0.00000379, -0.00000322, 0.00000209, 0.00000480, -0.00001388, 0.00000956, -0.00000074, 0.00000398, -0.00000967, 0.00000378, -0.00000382, 0.00000320, -0.00000718, -0.00000132, 0.00000766, -0.00000285, 0.00000007, 0.00000004, -0.00000202, -0.00000068, 0.00000754, 0.00000036, -0.00000089, -0.00000104, -0.00000641, -0.00000044, -0.00000019, 0.00000253, -0.00000213, 0.00000409, 0.00000098, 0.00000231, -0.00000622, -0.00000044, 0.00000229, -0.00000104, -0.00000131, -0.00000021, -0.00000155, -0.00000141, 0.00000677, -0.00000163, -0.00000211, 0.00000035, },
		{ 0.00000157, 0.00000126, -0.00000096, -0.00000076, -0.00000057, 0.00000036, -0.00000279, 0.00000039, -0.00000212, -0.00000228, 0.00000454, 0.00000193, -0.00000070, 0.00000115, 0.00000005, -0.00000235, 0.00000326, -0.00000153, -0.00000753, -0.00000050, 0.00000478, 0.00000050, 0.00000025, -0.00000055, 0.00000627, -0.00000141, -0.00000454, -0.00000295, 0.00000248, -0.00000262, 0.00000407, 0.00000055, -0.00000376, -0.00000111, -0.00000493, -0.00000039, 0.00000802, 0.00000095, -0.00000428, 0.00000015, 0.00000244, -0.00000472, -0.00000160, -0.00000350, 0.00001012, 0.00000188, -0.00000809, -0.00000532, 0.00001873, -0.00000979, -0.00001521, -0.00000728, 0.00000949, -0.00001330, -0.00000062, -0.00002071, -0.00000497, -0.00000883, 0.00000005, 0.00003559, -0.00000229, 0.00002333, -0.00001295, -0.00003082, -0.00015175, 0.00013583, 0.00026246, 0.00006549, -0.00015486, -0.00000508, -0.00003585, 0.00000323, -0.00003144, 0.00001416, -0.00001936, -0.00000122, -0.00000693, -0.00000625, 0.00000759, -0.00000896, 0.00000770, -0.00000273, 0.00001118, -0.00000344, -0.00000454, -0.00000265, 0.00000051, 0.00000643, -0.00001413, 0.00000907, -0.00000052, 0.00000397, -0.00000898, 0.00000246, -0.00000265, 0.00000213, -0.00000551, -0.00000301, 0.00000808, -0.00000275, 0.00000016, -0.00000011, -0.00000157, -0.00000131, 0.00000733, 0.00000131, -0.00000157, -0.00000046, -0.00000642, -0.00000077, -0.00000040, 0.00000274, -0.00000215, 0.00000384, 0.00000117, 0.00000240, -0.00000557, -0.00000144, 0.00000264, -0.00000104, -0.00000116, -0.00000049, -0.00000117, -0.00000199, 0.00000665, -0.00000092, -0.00000246, 0.00000038, },
		{ 0.00000148, 0.00000148, -0.00000106, -0.00000060, -0.00000082, 0.00000052, -0.00000277, 0.00000034, -0.00000182, -0.00000273, 0.00000417, 0.00000251, -0.00000079, 0.00000107, 0.00000028, -0.00000231, 0.00000277, -0.00000093, -0.00000746, -0.00000117, 0.00000484, 0.00000082, 0.00000017, -0.00000059, 0.00000593, -0.00000078, -0.00000452, -0.00000291, 0.00000193, -0.00000199, 0.00000290, 0.00000192, -0.00000436, -0.00000104, -0.00000482, -0.00000065, 0.00000729, 0.00000207, -0.00000445, -0.00000004, 0.00000264, -0.00000499, -0.00000139, -0.00000349, 0.00000899, 0.00000353, -0.00000796, -0.00000597, 0.00001751, -0.00000724, -0.00001638, -0.00000720, 0.00000792, -0.00001055, -0.00000358, -0.00001671, -0.00001014, -0.00000701, -0.00000290, 0.00003985, -0.00000342, 0.00002597, -0.00001212, -0.00002789, -0.00015042, 0.00012720, 0.00024244, 0.00008944, -0.00015050, -0.00001365, -0.00003174, -0.00000032, -0.00002850, 0.00001114, -0.00001535, -0.00000621, -0.00000237, -0.00001014, 0.00000988, -0.00001033, 0.00000819, -0.00000234, 0.00001016, -0.00000201, -0.00000520, -0.00000216, -0.00000089, 0.00000766, -0.00001394, 0.00000823, 0.00000000, 0.00000361, -0.00000792, 0.00000086, -0.00000132, 0.00000096, -0.00000382, -0.00000454, 0.00000832, -0.00000259, 0.00000023, -0.00000026, -0.00000114, -0.00000181, 0.00000695, 0.00000229, -0.00000216, -0.00000004, -0.00000627, -0.00000117, -0.00000056, 0.00000285, -0.00000204, 0.00000347, 0.00000145, 0.00000238, -0.00000481, -0.00000242, 0.00000293, -0.00000101, -0.00000100, -0.00000077, -0.00000082, -0.00000246, 0.00000639, -0.00000017, -0.00000278, 0.00000039, },
		{ 0.00000138, 0.00000170, -0.00000115, -0.00000044, -0.00000104, 0.00000063, -0.00000271, 0.00000025, -0.00000152, -0.00000312, 0.00000372, 0.00000306, -0.00000083, 0.00000096, 0.00000049, -0.00000220, 0.00000226, -0.00000040, -0.00000730, -0.00000183, 0.00000478, 0.00000123, 0.00000004, -0.00000054, 0.00000549, -0.00000010, -0.00000457, -0.00000272, 0.00000124, -0.00000124, 0.00000167, 0.00000319, -0.00000479, -0.00000109, -0.00000468, -0.00000084, 0.00000650, 0.00000310, -0.00000450, -0.00000026, 0.00000281, -0.00000514, -0.00000141, -0.00000317, 0.00000756, 0.00000529, -0.00000792, -0.00000628, 0.00001588, -0.00000449, -0.00001762, -0.00000688, 0.00000597, -0.00000749, -0.00000671, -0.00001274, -0.00001480, -0.00000584, -0.00000559, 0.00004374, -0.00000407, 0.00002829, -0.00001104, -0.00002506, -0.00014934, 0.00011975, 0.00022298, 0.00010915, -0.00014207, -0.00002404, -0.00002672, -0.00000451, -0.00002501, 0.00000756, -0.00001083, -0.00001139, 0.00000201, -0.00001360, 0.00001165, -0.00001121, 0.00000822, -0.00000161, 0.00000897, -0.00000061, -0.00000573, -0.00000180, -0.00000206, 0.00000847, -0.00001332, 0.00000709, 0.00000078, 0.00000294, -0.00000656, -0.00000093, 0.00000007, -0.00000023, -0.00000221, -0.00000585, 0.00000835, -0.00000235, 0.00000028, -0.00000038, -0.00000077, -0.00000218, 0.00000644, 0.00000328, -0.00000263, 0.00000020, -0.00000598, -0.00000164, -0.00000066, 0.00000286, -0.00000180, 0.00000300, 0.00000179, 0.00000227, -0.00000397, -0.00000337, 0.00000313, -0.00000095, -0.00000084, -0.00000103, -0.00000051, -0.00000282, 0.00000601, 0.00000060, -0.00000304, 0.00000037, },
		{ 0.00000127, 0.00000192, -0.00000122, -0.00000030, -0.00000122, 0.00000068, -0.00000260, 0.00000014, -0.00000122, -0.00000344, 0.00000321, 0.00000357, -0.00000081, 0.00000083, 0.00000067, -0.00000204, 0.00000174, 0.00000005, -0.00000705, -0.00000243, 0.00000460, 0.00000172, -0.00000014, -0.00000043, 0.00000497, 0.00000060, -0.00000466, -0.00000240, 0.00000044, -0.00000043, 0.00000047, 0.00000430, -0.00000500, -0.00000128, -0.00000450, -0.00000097, 0.00000568, 0.00000402, -0.00000442, -0.00000052, 0.00000294, -0.00000516, -0.00000165, -0.00000256, 0.00000590, 0.00000710, -0.00000793, -0.00000630, 0.00001392, -0.00000167, -0.00001880, -0.00000644, 0.00000378, -0.00000431, -0.00000978, -0.00000904, -0.00001874, -0.00000545, -0.00000796, 0.00004716, -0.00000414, 0.00003021, -0.00000960, -0.00002247, -0.00014828, 0.00011303, 0.00020464, 0.00012448, -0.00012992, -0.00003574, -0.00002122, -0.00000904, -0.00002123, 0.00000364, -0.00000606, -0.00001648, 0.00000597, -0.00001644, 0.00001281, -0.00001155, 0.00000779, -0.00000057, 0.00000766, 0.00000070, -0.00000610, -0.00000158, -0.00000298, 0.00000887, -0.00001231, 0.00000570, 0.00000177, 0.00000204, -0.00000498, -0.00000281, 0.00000144, -0.00000135, -0.00000074, -0.00000690, 0.00000815, -0.00000201, 0.00000028, -0.00000048, -0.00000045, -0.00000241, 0.00000580, 0.00000424, -0.00000296, 0.00000027, -0.00000555, -0.00000215, -0.00000072, 0.00000277, -0.00000147, 0.00000246, 0.00000217, 0.00000209, -0.00000309, -0.00000425, 0.00000324, -0.00000086, -0.00000069, -0.00000126, -0.00000027, -0.00000306, 0.00000551, 0.00000138, -0.00000324, 0.00000031, },
		{ 0.00000116, 0.00000214, -0.00000126, -0.00000019, -0.00000135, 0.00000068, -0.00000246, 0.00000001, -0.00000091, -0.00000370, 0.00000265, 0.00000403, -0.00000071, 0.00000067, 0.00000083, -0.00000183, 0.00000124, 0.00000041, -0.00000675, -0.00000299, 0.00000430, 0.00000228, -0.00000035, -0.00000027, 0.00000441, 0.00000130, -0.00000478, -0.00000199, -0.00000040, 0.00000036, -0.00000063, 0.00000518, -0.00000498, -0.00000160, -0.00000428, -0.00000107, 0.00000486, 0.00000480, -0.00000422, -0.00000081, 0.00000302, -0.00000506, -0.00000209, -0.00000171, 0.00000409, 0.00000887, -0.00000793, -0.00000612, 0.00001178, 0.00000108, -0.00001980, -0.00000599, 0.00000149, -0.00000121, -0.00001258, -0.00000583, -0.00002180, -0.00000590, -0.00000998, 0.00005005, -0.00000360, 0.00003167, -0.00000772, -0.00002025, -0.00014707, 0.00010668, 0.00018782, 0.00013543, -0.00011457, -0.00004821, -0.00001564, -0.00001357, -0.00001743, -0.00000035, -0.00000131, -0.00002123, 0.00000930, -0.00001854, 0.00001328, -0.00001133, 0.00000690, 0.00000074, 0.00000631, 0.00000189, -0.00000627, -0.00000153, -0.00000364, 0.00000887, -0.00001096, 0.00000413, 0.00000291, 0.00000096, -0.00000327, -0.00000469, 0.00000270, -0.00000232, 0.00000051, -0.00000764, 0.00000772, -0.00000159, 0.00000023, -0.00000053, -0.00000020, -0.00000250, 0.00000507, 0.00000514, -0.00000312, 0.00000017, -0.00000500, -0.00000268, -0.00000074, 0.00000260, -0.00000105, 0.00000187, 0.00000256, 0.00000188, -0.00000220, -0.00000503, 0.00000323, -0.00000071, -0.00000057, -0.00000145, -0.00000009, -0.00000318, 0.00000491, 0.00000213, -0.00000337, 0.00000022, },
		{ 0.00000105, 0.00000233, -0.00000127, -0.00000012, -0.00000143, 0.00000062, -0.00000228, -0.00000014, -0.00000060, -0.00000390, 0.00000207, 0.00000442, -0.00000055, 0.00000049, 0.00000096, -0.00000158, 0.00000078, 0.00000067, -0.00000639, -0.00000347, 0.00000388, 0.00000289, -0.00000056, -0.00000008, 0.00000382, 0.00000196, -0.00000490, -0.00000153, -0.00000124, 0.00000110, -0.00000160, 0.00000581, -0.00000470, -0.00000206, -0.00000403, -0.00000114, 0.00000408, 0.00000543, -0.00000389, -0.00000109, 0.00000304, -0.00000482, -0.00000272, -0.00000067, 0.00000218, 0.00001054, -0.00000785, -0.00000580, 0.00000957, 0.00000363, -0.00002053, -0.00000562, -0.00000075, 0.00000166, -0.00001494, -0.00000329, -0.00002385, -0.00000718, -0.00001167, 0.00005239, -0.00000240, 0.00003264, -0.00000536, -0.00001845, -0.00014558, 0.00010042, 0.00017278, 0.00014220, -0.00009662, -0.00006082, -0.00001042, -0.00001781, -0.00001386, -0.00000419, 0.00000319, -0.00002542, 0.00001184, -0.00001981, 0.00001304, -0.00001057, 0.00000560, 0.00000224, 0.00000496, 0.00000291, -0.00000624, -0.00000166, -0.00000404, 0.00000850, -0.00000936, 0.00000244, 0.00000414, -0.00000021, -0.00000151, -0.00000647, 0.00000376, -0.00000309, 0.00000150, -0.00000805, 0.00000707, -0.00000110, 0.00000014, -0.00000053, -0.00000003, -0.00000247, 0.00000427, 0.00000594, -0.00000311, -0.00000009, -0.00000437, -0.00000322, -0.00000075, 0.00000237, -0.00000057, 0.00000126, 0.00000295, 0.00000165, -0.00000133, -0.00000569, 0.00000311, -0.00000053, -0.00000048, -0.00000159, 0.00000001, -0.00000320, 0.00000425, 0.00000284, -0.00000342, 0.00000010, },
		{ 0.00000094, 0.00000251, -0.00000123, -0.00000009, -0.00000146, 0.00000050, -0.00000208, -0.00000030, -0.00000030, -0.00000405, 0.00000147, 0.00000473, -0.00000031, 0.00000030, 0.00000106, -0.00000132, 0.00000036, 0.00000084, -0.00000599, -0.00000386, 0.00000335, 0.00000352, -0.00000075, 0.00000012, 0.00000323, 0.00000258, -0.00000500, -0.00000105, -0.00000203, 0.00000171, -0.00000236, 0.00000615, -0.00000418, -0.00000264, -0.00000374, -0.00000119, 0.00000333, 0.00000590, -0.00000346, -0.00000137, 0.00000299, -0.00000446, -0.00000352, 0.00000050, 0.00000027, 0.00001202, -0.00000765, -0.00000544, 0.00000741, 0.00000586, -0.00002090, -0.00000541, -0.00000286, 0.00000417, -0.00001673, -0.00000151, -0.00002487, -0.00000924, -0.00001311, 0.00005419, -0.00000057, 0.00003314, -0.00000249, -0.00001711, -0.00014371, 0.00009407, 0.00015965, 0.00014510, -0.00007674, -0.00007299, -0.00000594, -0.00002148, -0.00001073, -0.00000768, 0.00000724, -0.00002886, 0.00001347, -0.00002021, 0.00001213, -0.00000932, 0.00000397, 0.00000386, 0.00000369, 0.00000372, -0.00000598, -0.00000196, -0.00000419, 0.00000783, -0.00000756, 0.00000071, 0.00000537, -0.00000140, 0.00000019, -0.00000806, 0.00000454, -0.00000358, 0.00000220, -0.00000814, 0.00000622, -0.00000054, 0.00000001, -0.00000050, 0.00000007, -0.00000233, 0.00000343, 0.00000663, -0.00000292, -0.00000049, -0.00000368, -0.00000374, -0.00000075, 0.00000208, -0.00000006, 0.00000065, 0.00000330, 0.00000144, -0.00000051, -0.00000621, 0.00000286, -0.00000029, -0.00000042, -0.00000167, 0.00000004, -0.00000311, 0.00000353, 0.00000349, -0.00000338, -0.00000005, },
		{ 0.00000082, 0.00000265, -0.00000115, -0.00000011, -0.00000143, 0.00000034, -0.00000185, -0.00000048, 0.00000000, -0.00000415, 0.00000087, 0.00000495, -0.00000002, 0.00000011, 0.00000113, -0.00000103, -0.00000001, 0.00000093, -0.00000557, -0.00000417, 0.00000273, 0.00000415, -0.00000090, 0.00000030, 0.00000267, 0.00000312, -0.00000506, -0.00000059, -0.00000271, 0.00000216, -0.00000290, 0.00000620, -0.00000343, -0.00000331, -0.00000345, -0.00000122, 0.00000265, 0.00000622, -0.00000296, -0.00000162, 0.00000286, -0.00000398, -0.00000443, 0.00000175, -0.00000156, 0.00001326, -0.00000726, -0.00000510, 0.00000540, 0.00000768, -0.00002084, -0.00000543, -0.00000472, 0.00000620, -0.00001785, -0.00000057, -0.00002487, -0.00001195, -0.00001441, 0.00005550, 0.00000185, 0.00003321, 0.00000085, -0.00001622, -0.00014146, 0.00008751, 0.00014845, 0.00014457, -0.00005567, -0.00008413, -0.00000255, -0.00002435, -0.00000822, -0.00001066, 0.00001067, -0.00003142, 0.00001414, -0.00001976, 0.00001059, -0.00000765, 0.00000208, 0.00000551, 0.00000257, 0.00000428, -0.00000550, -0.00000242, -0.00000413, 0.00000691, -0.00000567, -0.00000098, 0.00000655, -0.00000254, 0.00000177, -0.00000939, 0.00000501, -0.00000378, 0.00000259, -0.00000792, 0.00000519, 0.00000005, -0.00000015, -0.00000042, 0.00000011, -0.00000210, 0.00000259, 0.00000717, -0.00000256, -0.00000100, -0.00000297, -0.00000422, -0.00000076, 0.00000176, 0.00000046, 0.00000008, 0.00000358, 0.00000126, 0.00000025, -0.00000658, 0.00000250, -0.00000002, -0.00000039, -0.00000170, 0.00000000, -0.00000294, 0.00000278, 0.00000406, -0.00000326, -0.00000023, },
		{ 0.00000070, 0.00000276, -0.00000102, -0.00000016, -0.00000136, 0.00000013, -0.00000160, -0.00000066, 0.00000029, -0.00000420, 0.00000028, 0.00000508, 0.00000033, -0.00000008, 0.00000118, -0.00000073, -0.00000033, 0.00000094, -0.00000515, -0.00000437, 0.00000202, 0.00000475, -0.00000099, 0.00000045, 0.00000216, 0.00000357, -0.00000506, -0.00000019, -0.00000325, 0.00000242, -0.00000320, 0.00000596, -0.00000248, -0.00000403, -0.00000318, -0.00000123, 0.00000202, 0.00000640, -0.00000240, -0.00000181, 0.00000265, -0.00000340, -0.00000543, 0.00000300, -0.00000324, 0.00001418, -0.00000665, -0.00000486, 0.00000364, 0.00000902, -0.00002032, -0.00000569, -0.00000630, 0.00000769, -0.00001826, -0.00000045, -0.00002394, -0.00001515, -0.00001569, 0.00005637, 0.00000480, 0.00003292, 0.00000463, -0.00001574, -0.00013883, 0.00008071, 0.00013908, 0.00014109, -0.00003413, -0.00009373, -0.00000052, -0.00002626, -0.00000645, -0.00001301, 0.00001338, -0.00003304, 0.00001383, -0.00001853, 0.00000852, -0.00000565, 0.00000003, 0.00000711, 0.00000165, 0.00000457, -0.00000479, -0.00000303, -0.00000388, 0.00000580, -0.00000375, -0.00000256, 0.00000762, -0.00000357, 0.00000316, -0.00001040, 0.00000511, -0.00000368, 0.00000269, -0.00000741, 0.00000404, 0.00000065, -0.00000033, -0.00000031, 0.00000009, -0.00000180, 0.00000177, 0.00000755, -0.00000202, -0.00000159, -0.00000227, -0.00000463, -0.00000078, 0.00000142, 0.00000097, -0.00000044, 0.00000379, 0.00000113, 0.00000092, -0.00000680, 0.00000202, 0.00000029, -0.00000039, -0.00000168, -0.00000010, -0.00000270, 0.00000203, 0.00000453, -0.00000305, -0.00000042, },
		{ 0.00000058, 0.00000283, -0.00000083, -0.00000026, -0.00000124, -0.00000010, -0.00000135, -0.00000085, 0.00000056, -0.00000421, -0.00000029, 0.00000513, 0.00000072, -0.00000025, 0.00000119, -0.00000043, -0.00000058, 0.00000088, -0.00000474, -0.00000448, 0.00000126, 0.00000530, -0.00000100, 0.00000054, 0.00000171, 0.00000391, -0.00000498, 0.00000014, -0.00000363, 0.00000247, -0.00000325, 0.00000545, -0.00000137, -0.00000476, -0.00000295, -0.00000120, 0.00000145, 0.00000646, -0.00000182, -0.00000194, 0.00000236, -0.00000274, -0.00000647, 0.00000420, -0.00000470, 0.00001475, -0.00000577, -0.00000476, 0.00000218, 0.00000984, -0.00001936, -0.00000620, -0.00000757, 0.00000863, -0.00001798, -0.00000110, -0.00002222, -0.00001861, -0.00001711, 0.00005687, 0.00000819, 0.00003236, 0.00000875, -0.00001559, -0.00013588, 0.00007368, 0.00013137, 0.00013525, -0.00001280, -0.00010135, -0.00000004, -0.00002712, -0.00000547, -0.00001469, 0.00001530, -0.00003368, 0.00001261, -0.00001660, 0.00000604, -0.00000345, -0.00000206, 0.00000855, 0.00000100, 0.00000457, -0.00000389, -0.00000375, -0.00000350, 0.00000458, -0.00000188, -0.00000397, 0.00000851, -0.00000441, 0.00000429, -0.00001105, 0.00000485, -0.00000327, 0.00000250, -0.00000667, 0.00000281, 0.00000123, -0.00000050, -0.00000019, 0.00000003, -0.00000145, 0.00000100, 0.00000776, -0.00000134, -0.00000224, -0.00000160, -0.00000497, -0.00000084, 0.00000109, 0.00000145, -0.00000088, 0.00000390, 0.00000107, 0.00000150, -0.00000685, 0.00000145, 0.00000063, -0.00000042, -0.00000161, -0.00000026, -0.00000241, 0.00000130, 0.00000489, -0.00000275, -0.00000062, },
		{ 0.00000046, 0.00000286, -0.00000059, -0.00000040, -0.00000108, -0.00000036, -0.00000110, -0.00000103, 0.00000082, -0.00000417, -0.00000085, 0.00000510, 0.00000113, -0.00000039, 0.00000116, -0.00000013, -0.00000077, 0.00000077, -0.00000437, -0.00000448, 0.00000045, 0.00000578, -0.00000091, 0.00000058, 0.00000134, 0.00000413, -0.00000483, 0.00000037, -0.00000382, 0.00000231, -0.00000307, 0.00000473, -0.00000017, -0.00000546, -0.00000279, -0.00000112, 0.00000092, 0.00000643, -0.00000125, -0.00000199, 0.00000200, -0.00000202, -0.00000751, 0.00000527, -0.00000587, 0.00001492, -0.00000461, -0.00000484, 0.00000108, 0.00001013, -0.00001797, -0.00000693, -0.00000853, 0.00000903, -0.00001704, -0.00000242, -0.00001990, -0.00002213, -0.00001882, 0.00005707, 0.00001195, 0.00003161, 0.00001316, -0.00001569, -0.00013269, 0.00006651, 0.00012510, 0.00012759, 0.00000766, -0.00010664, -0.00000120, -0.00002692, -0.00000529, -0.00001568, 0.00001644, -0.00003339, 0.00001057, -0.00001414, 0.00000330, -0.00000117, -0.00000409, 0.00000976, 0.00000066, 0.00000429, -0.00000280, -0.00000454, -0.00000303, 0.00000331, -0.00000015, -0.00000515, 0.00000920, -0.00000505, 0.00000512, -0.00001132, 0.00000422, -0.00000259, 0.00000208, -0.00000574, 0.00000156, 0.00000175, -0.00000066, -0.00000007, -0.00000004, -0.00000108, 0.00000032, 0.00000779, -0.00000053, -0.00000290, -0.00000100, -0.00000521, -0.00000093, 0.00000076, 0.00000188, -0.00000123, 0.00000392, 0.00000107, 0.00000197, -0.00000676, 0.00000079, 0.00000097, -0.00000048, -0.00000149, -0.00000046, -0.00000208, 0.00000061, 0.00000514, -0.00000238, -0.00000083, },
		{ 0.00000035, 0.00000284, -0.00000030, -0.00000056, -0.00000090, -0.00000063, -0.00000086, -0.00000119, 0.00000106, -0.00000409, -0.00000139, 0.00000499, 0.00000155, -0.00000050, 0.00000109, 0.00000016, -0.00000091, 0.00000062, -0.00000403, -0.00000440, -0.00000037, 0.00000615, -0.00000071, 0.00000055, 0.00000105, 0.00000424, -0.00000458, 0.00000050, -0.00000383, 0.00000196, -0.00000269, 0.00000382, 0.00000108, -0.00000607, -0.00000274, -0.00000099, 0.00000042, 0.00000633, -0.00000071, -0.00000194, 0.00000158, -0.00000127, -0.00000851, 0.00000618, -0.00000673, 0.00001468, -0.00000317, -0.00000510, 0.00000033, 0.00000993, -0.00001623, -0.00000781, -0.00000924, 0.00000894, -0.00001556, -0.00000425, -0.00001719, -0.00002545, -0.00002095, 0.00005703, 0.00001599, 0.00003078, 0.00001775, -0.00001594, -0.00012937, 0.00005927, 0.00012001, 0.00011869, 0.00002670, -0.00010940, -0.00000398, -0.00002571, -0.00000585, -0.00001602, 0.00001683, -0.00003225, 0.00000784, -0.00001129, 0.00000046, 0.00000107, -0.00000596, 0.00001067, 0.00000065, 0.00000374, -0.00000158, -0.00000537, -0.00000252, 0.00000207, 0.00000140, -0.00000607, 0.00000965, -0.00000544, 0.00000564, -0.00001120, 0.00000325, -0.00000168, 0.00000148, -0.00000469, 0.00000033, 0.00000217, -0.00000077, 0.00000003, -0.00000012, -0.00000071, -0.00000027, 0.00000765, 0.00000038, -0.00000355, -0.00000049, -0.00000535, -0.00000106, 0.00000047, 0.00000223, -0.00000147, 0.00000382, 0.00000113, 0.00000235, -0.00000653, 0.00000008, 0.00000132, -0.00000054, -0.00000135, -0.00000069, -0.00000174, -0.00000002, 0.00000526, -0.00000194, -0.00000104, },
		{ 0.00000024, 0.00000276, 0.00000004, -0.00000074, -0.00000071, -0.00000089, -0.00000065, -0.00000134, 0.00000126, -0.00000397, -0.00000190, 0.00000482, 0.00000197, -0.00000055, 0.00000099, 0.00000045, -0.00000099, 0.00000044, -0.00000375, -0.00000423, -0.00000118, 0.00000640, -0.00000040, 0.00000045, 0.00000085, 0.00000423, -0.00000425, 0.00000052, -0.00000366, 0.00000143, -0.00000215, 0.00000280, 0.00000229, -0.00000654, -0.00000281, -0.00000079, -0.00000005, 0.00000619, -0.00000023, -0.00000181, 0.00000111, -0.00000052, -0.00000941, 0.00000687, -0.00000723, 0.00001402, -0.00000148, -0.00000553, -0.00000007, 0.00000929, -0.00001422, -0.00000877, -0.00000973, 0.00000843, -0.00001363, -0.00000644, -0.00001431, -0.00002835, -0.00002363, 0.00005677, 0.00002024, 0.00002995, 0.00002245, -0.00001625, -0.00012602, 0.00005210, 0.00011584, 0.00010907, 0.00004388, -0.00010951, -0.00000830, -0.00002363, -0.00000703, -0.00001581, 0.00001655, -0.00003038, 0.00000460, -0.00000825, -0.00000234, 0.00000314, -0.00000757, 0.00001121, 0.00000099, 0.00000293, -0.00000026, -0.00000619, -0.00000202, 0.00000090, 0.00000270, -0.00000670, 0.00000985, -0.00000557, 0.00000584, -0.00001072, 0.00000200, -0.00000060, 0.00000077, -0.00000360, -0.00000080, 0.00000248, -0.00000083, 0.00000010, -0.00000019, -0.00000037, -0.00000074, 0.00000734, 0.00000136, -0.00000414, -0.00000009, -0.00000539, -0.00000123, 0.00000020, 0.00000251, -0.00000160, 0.00000363, 0.00000126, 0.00000264, -0.00000617, -0.00000068, 0.00000164, -0.00000061, -0.00000119, -0.00000092, -0.00000141, -0.00000058, 0.00000526, -0.00000144, -0.00000123, },
		{ 0.00000013, 0.00000264, 0.00000041, -0.00000093, -0.00000051, -0.00000115, -0.00000046, -0.00000146, 0.00000143, -0.00000380, -0.00000239, 0.00000460, 0.00000237, -0.00000055, 0.00000085, 0.00000073, -0.00000103, 0.00000026, -0.00000353, -0.00000398, -0.00000196, 0.00000652, 0.00000003, 0.00000029, 0.00000073, 0.00000412, -0.00000384, 0.00000044, -0.00000332, 0.00000076, -0.00000151, 0.00000172, 0.00000342, -0.00000683, -0.00000303, -0.00000051, -0.00000051, 0.00000602, 0.00000018, -0.00000160, 0.00000062, 0.00000020, -0.00001019, 0.00000730, -0.00000736, 0.00001298, 0.00000044, -0.00000612, -0.00000016, 0.00000829, -0.00001206, -0.00000972, -0.00001010, 0.00000760, -0.00001141, -0.00000880, -0.00001150, -0.00003062, -0.00002691, 0.00005631, 0.00002465, 0.00002919, 0.00002718, -0.00001651, -0.00012274, 0.00004510, 0.00011231, 0.00009919, 0.00005884, -0.00010699, -0.00001395, -0.00002085, -0.00000869, -0.00001517, 0.00001574, -0.00002793, 0.00000102, -0.00000520, -0.00000494, 0.00000494, -0.00000885, 0.00001134, 0.00000168, 0.00000191, 0.00000111, -0.00000696, -0.00000159, -0.00000013, 0.00000373, -0.00000703, 0.00000980, -0.00000545, 0.00000574, -0.00000991, 0.00000051, 0.00000057, 0.00000000, -0.00000251, -0.00000180, 0.00000264, -0.00000082, 0.00000012, -0.00000023, -0.00000006, -0.00000109, 0.00000687, 0.00000238, -0.00000465, 0.00000018, -0.00000532, -0.00000144, -0.00000003, 0.00000271, -0.00000163, 0.00000335, 0.00000143, 0.00000284, -0.00000571, -0.00000146, 0.00000194, -0.00000067, -0.00000101, -0.00000116, -0.00000110, -0.00000104, 0.00000514, -0.00000090, -0.00000141, },
		{ 0.00000003, 0.00000247, 0.00000081, -0.00000112, -0.00000032, -0.00000137, -0.00000031, -0.00000155, 0.00000156, -0.00000358, -0.00000286, 0.00000434, 0.00000272, -0.00000049, 0.00000068, 0.00000099, -0.00000102, 0.00000007, -0.00000337, -0.00000369, -0.00000268, 0.00000648, 0.00000057, 0.00000007, 0.00000068, 0.00000391, -0.00000337, 0.00000027, -0.00000284, -0.00000000, -0.00000080, 0.00000065, 0.00000440, -0.00000691, -0.00000340, -0.00000015, -0.00000096, 0.00000585, 0.00000051, -0.00000132, 0.00000013, 0.00000086, -0.00001080, 0.00000746, -0.00000712, 0.00001159, 0.00000252, -0.00000681, -0.00000001, 0.00000703, -0.00000986, -0.00001055, -0.00001041, 0.00000655, -0.00000904, -0.00001115, -0.00000895, -0.00003212, -0.00003081, 0.00005560, 0.00002917, 0.00002857, 0.00003189, -0.00001665, -0.00011963, 0.00003838, 0.00010920, 0.00008946, 0.00007137, -0.00010197, -0.00002068, -0.00001761, -0.00001065, -0.00001425, 0.00001451, -0.00002506, -0.00000268, -0.00000232, -0.00000722, 0.00000639, -0.00000974, 0.00001105, 0.00000269, 0.00000073, 0.00000248, -0.00000762, -0.00000126, -0.00000098, 0.00000446, -0.00000708, 0.00000952, -0.00000509, 0.00000536, -0.00000883, -0.00000114, 0.00000177, -0.00000074, -0.00000151, -0.00000261, 0.00000265, -0.00000073, 0.00000010, -0.00000025, 0.00000020, -0.00000132, 0.00000627, 0.00000341, -0.00000505, 0.00000032, -0.00000514, -0.00000168, -0.00000022, 0.00000282, -0.00000155, 0.00000299, 0.00000164, 0.00000297, -0.00000517, -0.00000224, 0.00000219, -0.00000072, -0.00000084, -0.00000138, -0.00000083, -0.00000142, 0.00000491, -0.00000033, -0.00000156, },
		{ -0.00000006, 0.00000225, 0.00000123, -0.00000129, -0.00000015, -0.00000156, -0.00000021, -0.00000161, 0.00000165, -0.00000331, -0.00000331, 0.00000404, 0.00000304, -0.00000036, 0.00000048, 0.00000123, -0.00000097, -0.00000010, -0.00000327, -0.00000335, -0.00000332, 0.00000629, 0.00000121, -0.00000019, 0.00000069, 0.00000363, -0.00000286, 0.00000002, -0.00000226, -0.00000081, -0.00000010, -0.00000034, 0.00000518, -0.00000675, -0.00000394, 0.00000027, -0.00000140, 0.00000568, 0.00000075, -0.00000098, -0.00000033, 0.00000143, -0.00001123, 0.00000733, -0.00000654, 0.00000990, 0.00000470, -0.00000754, 0.00000032, 0.00000562, -0.00000774, -0.00001118, -0.00001074, 0.00000538, -0.00000664, -0.00001332, -0.00000685, -0.00003272, -0.00003531, 0.00005458, 0.00003380, 0.00002813, 0.00003653, -0.00001661, -0.00011676, 0.00003203, 0.00010630, 0.00008019, 0.00008134, -0.00009468, -0.00002816, -0.00001417, -0.00001271, -0.00001318, 0.00001302, -0.00002195, -0.00000633, 0.00000023, -0.00000906, 0.00000742, -0.00001021, 0.00001033, 0.00000399, -0.00000056, 0.00000379, -0.00000814, -0.00000107, -0.00000163, 0.00000490, -0.00000686, 0.00000903, -0.00000452, 0.00000474, -0.00000753, -0.00000286, 0.00000292, -0.00000140, -0.00000063, -0.00000322, 0.00000249, -0.00000057, 0.00000001, -0.00000022, 0.00000041, -0.00000142, 0.00000557, 0.00000440, -0.00000531, 0.00000032, -0.00000489, -0.00000194, -0.00000038, 0.00000284, -0.00000138, 0.00000256, 0.00000187, 0.00000306, -0.00000457, -0.00000298, 0.00000238, -0.00000075, -0.00000069, -0.00000157, -0.00000061, -0.00000169, 0.00000458, 0.00000025, -0.00000167, },
		{ -0.00000015, 0.00000199, 0.00000166, -0.00000142, -0.00000002, -0.00000171, -0.00000015, -0.00000163, 0.00000169, -0.00000299, -0.00000375, 0.00000373, 0.00000329, -0.00000018, 0.00000026, 0.00000144, -0.00000088, -0.00000026, -0.00000323, -0.00000300, -0.00000386, 0.00000594, 0.00000193, -0.00000047, 0.00000074, 0.00000329, -0.00000232, -0.00000028, -0.00000161, -0.00000161, 0.00000054, -0.00000121, 0.00000572, -0.00000634, -0.00000462, 0.00000074, -0.00000183, 0.00000552, 0.00000092, -0.00000061, -0.00000074, 0.00000189, -0.00001145, 0.00000691, -0.00000564, 0.00000797, 0.00000691, -0.00000825, 0.00000073, 0.00000419, -0.00000583, -0.00001152, -0.00001117, 0.00000418, -0.00000437, -0.00001516, -0.00000533, -0.00003237, -0.00004032, 0.00005315, 0.00003853, 0.00002789, 0.00004107, -0.00001634, -0.00011419, 0.00002612, 0.00010344, 0.00007162, 0.00008877, -0.00008544, -0.00003603, -0.00001078, -0.00001469, -0.00001213, 0.00001141, -0.00001877, -0.00000976, 0.00000233, -0.00001039, 0.00000800, -0.00001025, 0.00000922, 0.00000551, -0.00000190, 0.00000498, -0.00000848, -0.00000104, -0.00000207, 0.00000505, -0.00000641, 0.00000838, -0.00000378, 0.00000394, -0.00000609, -0.00000458, 0.00000394, -0.00000192, 0.00000006, -0.00000359, 0.00000216, -0.00000034, -0.00000012, -0.00000015, 0.00000056, -0.00000140, 0.00000478, 0.00000533, -0.00000543, 0.00000018, -0.00000455, -0.00000222, -0.00000050, 0.00000279, -0.00000114, 0.00000210, 0.00000210, 0.00000310, -0.00000393, -0.00000368, 0.00000250, -0.00000074, -0.00000056, -0.00000172, -0.00000045, -0.00000186, 0.00000416, 0.00000083, -0.00000175, },
		{ -0.00000022, 0.00000170, 0.00000207, -0.00000152, 0.00000007, -0.00000180, -0.00000015, -0.00000162, 0.00000168, -0.00000262, -0.00000416, 0.00000340, 0.00000349, 0.00000006, 0.00000003, 0.00000163, -0.00000077, -0.00000040, -0.00000324, -0.00000264, -0.00000428, 0.00000544, 0.00000272, -0.00000075, 0.00000081, 0.00000292, -0.00000178, -0.00000060, -0.00000094, -0.00000235, 0.00000107, -0.00000191, 0.00000599, -0.00000567, -0.00000544, 0.00000125, -0.00000224, 0.00000537, 0.00000103, -0.00000025, -0.00000107, 0.00000222, -0.00001146, 0.00000622, -0.00000448, 0.00000590, 0.00000905, -0.00000886, 0.00000115, 0.00000284, -0.00000422, -0.00001149, -0.00001174, 0.00000304, -0.00000231, -0.00001658, -0.00000448, -0.00003108, -0.00004567, 0.00005118, 0.00004337, 0.00002787, 0.00004552, -0.00001581, -0.00011195, 0.00002069, 0.00010049, 0.00006390, 0.00009375, -0.00007462, -0.00004391, -0.00000770, -0.00001639, -0.00001124, 0.00000981, -0.00001566, -0.00001281, 0.00000386, -0.00001117, 0.00000813, -0.00000988, 0.00000776, 0.00000719, -0.00000320, 0.00000600, -0.00000860, -0.00000118, -0.00000228, 0.00000493, -0.00000576, 0.00000760, -0.00000291, 0.00000300, -0.00000458, -0.00000623, 0.00000478, -0.00000224, 0.00000055, -0.00000372, 0.00000168, -0.00000004, -0.00000031, -0.00000005, 0.00000065, -0.00000129, 0.00000394, 0.00000616, -0.00000538, -0.00000008, -0.00000417, -0.00000251, -0.00000058, 0.00000267, -0.00000083, 0.00000161, 0.00000231, 0.00000313, -0.00000327, -0.00000431, 0.00000254, -0.00000070, -0.00000046, -0.00000182, -0.00000035, -0.00000193, 0.00000367, 0.00000139, -0.00000179, },
		{ -0.00000029, 0.00000138, 0.00000247, -0.00000155, 0.00000012, -0.00000184, -0.00000019, -0.00000157, 0.00000163, -0.00000221, -0.00000456, 0.00000306, 0.00000363, 0.00000035, -0.00000020, 0.00000177, -0.00000062, -0.00000051, -0.00000329, -0.00000231, -0.00000458, 0.00000480, 0.00000354, -0.00000100, 0.00000089, 0.00000253, -0.00000126, -0.00000092, -0.00000027, -0.00000299, 0.00000146, -0.00000241, 0.00000599, -0.00000478, -0.00000635, 0.00000177, -0.00000261, 0.00000521, 0.00000110, 0.00000008, -0.00000129, 0.00000241, -0.00001126, 0.00000528, -0.00000310, 0.00000376, 0.00001106, -0.00000930, 0.00000147, 0.00000168, -0.00000302, -0.00001106, -0.00001249, 0.00000201, -0.00000053, -0.00001750, -0.00000433, -0.00002894, -0.00005119, 0.00004851, 0.00004836, 0.00002807, 0.00004987, -0.00001502, -0.00011006, 0.00001575, 0.00009739, 0.00005712, 0.00009646, -0.00006265, -0.00005144, -0.00000518, -0.00001767, -0.00001061, 0.00000832, -0.00001275, -0.00001538, 0.00000477, -0.00001141, 0.00000784, -0.00000914, 0.00000600, 0.00000895, -0.00000439, 0.00000681, -0.00000850, -0.00000152, -0.00000229, 0.00000459, -0.00000498, 0.00000675, -0.00000196, 0.00000199, -0.00000308, -0.00000772, 0.00000537, -0.00000234, 0.00000082, -0.00000362, 0.00000107, 0.00000029, -0.00000054, 0.00000008, 0.00000071, -0.00000109, 0.00000308, 0.00000686, -0.00000516, -0.00000046, -0.00000375, -0.00000278, -0.00000064, 0.00000249, -0.00000048, 0.00000113, 0.00000249, 0.00000314, -0.00000261, -0.00000486, 0.00000250, -0.00000062, -0.00000039, -0.00000187, -0.00000032, -0.00000191, 0.00000313, 0.00000191, -0.00000177, },
		{ -0.00000035, 0.00000103, 0.00000282, -0.00000152, 0.00000010, -0.00000183, -0.00000028, -0.00000149, 0.00000153, -0.00000177, -0.00000492, 0.00000271, 0.00000371, 0.00000067, -0.00000041, 0.00000187, -0.00000044, -0.00000059, -0.00000336, -0.00000203, -0.00000474, 0.00000403, 0.00000438, -0.00000121, 0.00000095, 0.00000216, -0.00000078, -0.00000122, 0.00000034, -0.00000349, 0.00000167, -0.00000268, 0.00000572, -0.00000369, -0.00000731, 0.00000225, -0.00000292, 0.00000504, 0.00000116, 0.00000035, -0.00000138, 0.00000244, -0.00001084, 0.00000413, -0.00000159, 0.00000163, 0.00001284, -0.00000952, 0.00000164, 0.00000080, -0.00000226, -0.00001021, -0.00001342, 0.00000110, 0.00000091, -0.00001790, -0.00000485, -0.00002606, -0.00005664, 0.00004499, 0.00005349, 0.00002848, 0.00005415, -0.00001395, -0.00010851, 0.00001129, 0.00009409, 0.00005129, 0.00009715, -0.00004996, -0.00005826, -0.00000339, -0.00001842, -0.00001031, 0.00000704, -0.00001013, -0.00001738, 0.00000504, -0.00001112, 0.00000719, -0.00000810, 0.00000404, 0.00001070, -0.00000541, 0.00000736, -0.00000815, -0.00000203, -0.00000209, 0.00000405, -0.00000411, 0.00000586, -0.00000098, 0.00000097, -0.00000164, -0.00000900, 0.00000568, -0.00000220, 0.00000086, -0.00000331, 0.00000036, 0.00000065, -0.00000079, 0.00000022, 0.00000073, -0.00000084, 0.00000224, 0.00000742, -0.00000478, -0.00000092, -0.00000331, -0.00000303, -0.00000068, 0.00000226, -0.00000010, 0.00000067, 0.00000261, 0.00000317, -0.00000196, -0.00000531, 0.00000238, -0.00000050, -0.00000037, -0.00000187, -0.00000035, -0.00000182, 0.00000255, 0.00000238, -0.00000171, },
		{ -0.00000040, 0.00000068, 0.00000313, -0.00000142, 0.00000004, -0.00000176, -0.00000041, -0.00000138, 0.00000138, -0.00000129, -0.00000526, 0.00000235, 0.00000373, 0.00000101, -0.00000059, 0.00000192, -0.00000024, -0.00000065, -0.00000345, -0.00000180, -0.00000477, 0.00000315, 0.00000519, -0.00000135, 0.00000098, 0.00000181, -0.00000035, -0.00000148, 0.00000088, -0.00000382, 0.00000168, -0.00000274, 0.00000519, -0.00000244, -0.00000829, 0.00000267, -0.00000315, 0.00000483, 0.00000123, 0.00000051, -0.00000133, 0.00000231, -0.00001024, 0.00000281, -0.00000001, -0.00000039, 0.00001433, -0.00000944, 0.00000159, 0.00000025, -0.00000199, -0.00000896, -0.00001450, 0.00000033, 0.00000201, -0.00001779, -0.00000598, -0.00002264, -0.00006175, 0.00004048, 0.00005878, 0.00002910, 0.00005838, -0.00001263, -0.00010729, 0.00000730, 0.00009058, 0.00004637, 0.00009609, -0.00003699, -0.00006408, -0.00000247, -0.00001858, -0.00001040, 0.00000600, -0.00000787, -0.00001878, 0.00000468, -0.00001039, 0.00000625, -0.00000683, 0.00000196, 0.00001235, -0.00000620, 0.00000762, -0.00000756, -0.00000272, -0.00000174, 0.00000338, -0.00000321, 0.00000498, 0.00000000, -0.00000001, -0.00000033, -0.00001002, 0.00000569, -0.00000182, 0.00000070, -0.00000282, -0.00000040, 0.00000101, -0.00000105, 0.00000036, 0.00000073, -0.00000054, 0.00000144, 0.00000782, -0.00000424, -0.00000144, -0.00000289, -0.00000324, -0.00000071, 0.00000199, 0.00000028, 0.00000025, 0.00000268, 0.00000320, -0.00000134, -0.00000567, 0.00000217, -0.00000035, -0.00000039, -0.00000181, -0.00000043, -0.00000166, 0.00000195, 0.00000279, -0.00000160, },
		{ -0.00000044, 0.00000032, 0.00000337, -0.00000123, -0.00000008, -0.00000165, -0.00000058, -0.00000125, 0.00000120, -0.00000080, -0.00000556, 0.00000198, 0.00000371, 0.00000136, -0.00000072, 0.00000191, -0.00000001, -0.00000070, -0.00000354, -0.00000164, -0.00000466, 0.00000220, 0.00000596, -0.00000141, 0.00000097, 0.00000150, 0.00000000, -0.00000167, 0.00000131, -0.00000395, 0.00000150, -0.00000259, 0.00000446, -0.00000110, -0.00000921, 0.00000298, -0.00000328, 0.00000456, 0.00000135, 0.00000055, -0.00000113, 0.00000204, -0.00000948, 0.00000138, 0.00000158, -0.00000224, 0.00001547, -0.00000904, 0.00000129, 0.00000006, -0.00000219, -0.00000737, -0.00001567, -0.00000034, 0.00000280, -0.00001724, -0.00000759, -0.00001888, -0.00006626, 0.00003483, 0.00006420, 0.00002995, 0.00006256, -0.00001106, -0.00010638, 0.00000372, 0.00008689, 0.00004230, 0.00009360, -0.00002414, -0.00006866, -0.00000251, -0.00001812, -0.00001087, 0.00000520, -0.00000598, -0.00001959, 0.00000376, -0.00000929, 0.00000511, -0.00000542, -0.00000016, 0.00001382, -0.00000669, 0.00000758, -0.00000674, -0.00000354, -0.00000126, 0.00000263, -0.00000233, 0.00000415, 0.00000094, -0.00000091, 0.00000080, -0.00001074, 0.00000539, -0.00000122, 0.00000036, -0.00000220, -0.00000119, 0.00000133, -0.00000130, 0.00000048, 0.00000074, -0.00000023, 0.00000070, 0.00000805, -0.00000356, -0.00000200, -0.00000250, -0.00000342, -0.00000073, 0.00000170, 0.00000064, -0.00000011, 0.00000268, 0.00000325, -0.00000076, -0.00000592, 0.00000189, -0.00000016, -0.00000044, -0.00000170, -0.00000056, -0.00000145, 0.00000136, 0.00000313, -0.00000144, },
		{ -0.00000047, -0.00000004, 0.00000354, -0.00000097, -0.00000025, -0.00000150, -0.00000077, -0.00000110, 0.00000100, -0.00000029, -0.00000580, 0.00000160, 0.00000366, 0.00000169, -0.00000080, 0.00000185, 0.00000024, -0.00000072, -0.00000361, -0.00000157, -0.00000443, 0.00000119, 0.00000664, -0.00000135, 0.00000090, 0.00000124, 0.00000028, -0.00000180, 0.00000162, -0.00000389, 0.00000114, -0.00000227, 0.00000357, 0.00000027, -0.00001002, 0.00000314, -0.00000327, 0.00000421, 0.00000155, 0.00000045, -0.00000077, 0.00000164, -0.00000858, -0.00000013, 0.00000309, -0.00000383, 0.00001622, -0.00000830, 0.00000070, 0.00000026, -0.00000282, -0.00000549, -0.00001687, -0.00000096, 0.00000334, -0.00001633, -0.00000953, -0.00001504, -0.00006991, 0.00002796, 0.00006970, 0.00003104, 0.00006673, -0.00000924, -0.00010575, 0.00000053, 0.00008303, 0.00003898, 0.00008999, -0.00001176, -0.00007184, -0.00000350, -0.00001710, -0.00001167, 0.00000464, -0.00000446, -0.00001983, 0.00000237, -0.00000796, 0.00000389, -0.00000396, -0.00000221, 0.00001504, -0.00000685, 0.00000723, -0.00000570, -0.00000448, -0.00000069, 0.00000184, -0.00000152, 0.00000341, 0.00000180, -0.00000170, 0.00000173, -0.00001116, 0.00000479, -0.00000043, -0.00000011, -0.00000148, -0.00000195, 0.00000159, -0.00000152, 0.00000057, 0.00000075, 0.00000007, 0.00000006, 0.00000809, -0.00000275, -0.00000258, -0.00000216, -0.00000355, -0.00000074, 0.00000140, 0.00000098, -0.00000040, 0.00000261, 0.00000332, -0.00000022, -0.00000608, 0.00000155, 0.00000005, -0.00000053, -0.00000156, -0.00000073, -0.00000121, 0.00000079, 0.00000339, -0.00000123, },
		{ -0.00000049, -0.00000038, 0.00000364, -0.00000063, -0.00000046, -0.00000132, -0.00000098, -0.00000095, 0.00000077, 0.00000021, -0.00000599, 0.00000119, 0.00000358, 0.00000199, -0.00000082, 0.00000172, 0.00000051, -0.00000074, -0.00000366, -0.00000159, -0.00000408, 0.00000016, 0.00000722, -0.00000119, 0.00000077, 0.00000104, 0.00000047, -0.00000184, 0.00000179, -0.00000364, 0.00000062, -0.00000180, 0.00000257, 0.00000162, -0.00001067, 0.00000311, -0.00000312, 0.00000377, 0.00000184, 0.00000019, -0.00000028, 0.00000113, -0.00000759, -0.00000164, 0.00000446, -0.00000513, 0.00001654, -0.00000722, -0.00000016, 0.00000080, -0.00000382, -0.00000346, -0.00001800, -0.00000160, 0.00000370, -0.00001519, -0.00001163, -0.00001135, -0.00007245, 0.00001981, 0.00007519, 0.00003242, 0.00007087, -0.00000719, -0.00010538, -0.00000233, 0.00007905, 0.00003630, 0.00008554, -0.00000017, -0.00007352, -0.00000539, -0.00001558, -0.00001273, 0.00000426, -0.00000327, -0.00001958, 0.00000061, -0.00000651, 0.00000270, -0.00000255, -0.00000412, 0.00001593, -0.00000665, 0.00000657, -0.00000447, -0.00000550, -0.00000007, 0.00000106, -0.00000081, 0.00000277, 0.00000257, -0.00000233, 0.00000242, -0.00001128, 0.00000394, 0.00000050, -0.00000067, -0.00000074, -0.00000264, 0.00000177, -0.00000169, 0.00000060, 0.00000080, 0.00000035, -0.00000048, 0.00000797, -0.00000184, -0.00000313, -0.00000189, -0.00000363, -0.00000074, 0.00000110, 0.00000128, -0.00000062, 0.00000246, 0.00000340, 0.00000029, -0.00000613, 0.00000114, 0.00000029, -0.00000065, -0.00000138, -0.00000092, -0.00000096, 0.00000026, 0.00000356, -0.00000098, },
		},
		{ { 0.00000243, 0.00000166, -0.00000270, -0.00000191, -0.00000178, -0.00000132, 0.00000721, -0.00000037, -0.00000386, -0.00000254, 0.00000431, 0.00000055, -0.00000013, 0.00000041, -0.00000841, -0.00000131, 0.00000068, 0.00000284, 0.00000212, 0.00000233, 0.00000413, 0.00000085, -0.00000949, -0.00000044, 0.00000588, -0.00000095, -0.00000775, 0.00000318, 0.00000003, 0.00000428, 0.00000174, 0.00000027, -0.00000301, -0.00000608, 0.00002121, -0.00000749, 0.00000064, -0.00000385, 0.00000225, 0.00000091, -0.00001395, 0.00000130, 0.00000289, -0.00000228, -0.00001080, 0.00000202, -0.00001126, -0.00000060, -0.00000429, -0.00000561, -0.00000695, -0.00000813, 0.00000426, -0.00001552, -0.00006281, 0.00001270, 0.00008599, 0.00003798, 0.00008187, 0.00000955, -0.00006042, -0.00001182, -0.00001289, -0.00000789, -0.00000652, -0.00000789, -0.00001289, -0.00001182, -0.00006042, 0.00000955, 0.00008187, 0.00003798, 0.00008599, 0.00001270, -0.00006281, -0.00001552, 0.00000426, -0.00000813, -0.00000695, -0.00000561, -0.00000429, -0.00000060, -0.00001126, 0.00000202, -0.00001080, -0.00000228, 0.00000289, 0.00000130, -0.00001395, 0.00000091, 0.00000225, -0.00000385, 0.00000064, -0.00000749, 0.00002121, -0.00000608, -0.00000301, 0.00000027, 0.00000174, 0.00000428, 0.00000003, 0.00000318, -0.00000775, -0.00000095, 0.00000588, -0.00000044, -0.00000949, 0.00000085, 0.00000413, 0.00000233, 0.00000212, 0.00000284, 0.00000068, -0.00000131, -0.00000841, 0.00000041, -0.00000013, 0.00000055, 0.00000431, -0.00000254, -0.00000386, -0.00000037, 0.00000721, -0.00000132, -0.00000178, -0.00000191, -0.00000270, 0.00000166, },
		{ 0.00000231, 0.00000204, -0.00000277, -0.00000193, -0.00000139, -0.00000211, 0.00000652, 0.00000135, -0.00000439, -0.00000256, 0.00000355, 0.00000141, -0.00000110, 0.00000180, -0.00000811, -0.00000252, 0.00000069, 0.00000250, 0.00000257, 0.00000190, 0.00000440, 0.00000156, -0.00000907, -0.00000205, 0.00000605, 0.00000034, -0.00000783, 0.00000191, 0.00000078, 0.00000311, 0.00000328, -0.00000098, -0.00000060, -0.00000900, 0.00002099, -0.00000430, -0.00000173, -0.00000199, -0.00000047, 0.00000410, -0.00001467, -0.00000010, 0.00000294, -0.00000074, -0.00001171, 0.00000194, -0.00001103, -0.00000090, -0.00000400, -0.00000598, -0.00000637, -0.00000835, 0.00000294, -0.00001041, -0.00006112, -0.00000658, 0.00009153, 0.00004150, 0.00008238, 0.00002146, -0.00006437, -0.00001389, -0.00001287, -0.00000849, -0.00000647, -0.00000747, -0.00001271, -0.00001032, -0.00005601, -0.00000106, 0.00007972, 0.00003666, 0.00007809, 0.00002991, -0.00006044, -0.00002234, 0.00000585, -0.00000799, -0.00000739, -0.00000548, -0.00000436, -0.00000040, -0.00001141, 0.00000191, -0.00000966, -0.00000380, 0.00000274, 0.00000231, -0.00001244, -0.00000262, 0.00000463, -0.00000517, 0.00000238, -0.00000954, 0.00001997, -0.00000225, -0.00000563, 0.00000155, 0.00000025, 0.00000529, -0.00000056, 0.00000414, -0.00000727, -0.00000226, 0.00000548, 0.00000114, -0.00000964, 0.00000003, 0.00000394, 0.00000254, 0.00000193, 0.00000290, 0.00000085, -0.00000045, -0.00000805, -0.00000143, 0.00000094, -0.00000033, 0.00000504, -0.00000252, -0.00000328, -0.00000186, 0.00000735, -0.00000015, -0.00000223, -0.00000186, -0.00000276, 0.00000149, },
		{ 0.00000198, 0.00000257, -0.00000289, -0.00000197, -0.00000106, -0.00000258, 0.00000544, 0.00000308, -0.00000467, -0.00000271, 0.00000289, 0.00000211, -0.00000180, 0.00000259, -0.00000712, -0.00000403, 0.00000084, 0.00000194, 0.00000320, 0.00000136, 0.00000462, 0.00000225, -0.00000846, -0.00000359, 0.00000592, 0.00000160, -0.00000755, 0.00000045, 0.00000155, 0.00000195, 0.00000468, -0.00000201, 0.00000132, -0.00001081, 0.00001932, -0.00000021, -0.00000439, 0.00000003, -0.00000308, 0.00000655, -0.00001442, -0.00000193, 0.00000294, 0.00000069, -0.00001224, 0.00000153, -0.00001058, -0.00000151, -0.00000327, -0.00000679, -0.00000540, -0.00000897, 0.00000237, -0.00000743, -0.00005573, -0.00002655, 0.00009352, 0.00004757, 0.00008127, 0.00003448, -0.00006762, -0.00001662, -0.00001262, -0.00000927, -0.00000635, -0.00000721, -0.00001236, -0.00000926, -0.00005140, -0.00001024, 0.00007602, 0.00003710, 0.00006895, 0.00004411, -0.00005421, -0.00003008, 0.00000706, -0.00000748, -0.00000798, -0.00000534, -0.00000447, -0.00000008, -0.00001166, 0.00000173, -0.00000842, -0.00000515, 0.00000243, 0.00000297, -0.00001038, -0.00000606, 0.00000628, -0.00000568, 0.00000331, -0.00001036, 0.00001741, 0.00000215, -0.00000815, 0.00000260, -0.00000100, 0.00000598, -0.00000086, 0.00000471, -0.00000641, -0.00000356, 0.00000487, 0.00000256, -0.00000945, -0.00000095, 0.00000390, 0.00000250, 0.00000200, 0.00000273, 0.00000119, 0.00000007, -0.00000713, -0.00000349, 0.00000186, -0.00000104, 0.00000558, -0.00000234, -0.00000281, -0.00000295, 0.00000687, 0.00000135, -0.00000272, -0.00000175, -0.00000301, 0.00000154, },
		{ 0.00000153, 0.00000312, -0.00000294, -0.00000210, -0.00000074, -0.00000277, 0.00000413, 0.00000460, -0.00000455, -0.00000310, 0.00000244, 0.00000252, -0.00000212, 0.00000274, -0.00000553, -0.00000569, 0.00000104, 0.00000126, 0.00000388, 0.00000088, 0.00000467, 0.00000300, -0.00000777, -0.00000495, 0.00000548, 0.00000279, -0.00000696, -0.00000107, 0.00000219, 0.00000092, 0.00000580, -0.00000264, 0.00000260, -0.00001143, 0.00001640, 0.00000436, -0.00000691, 0.00000179, -0.00000513, 0.00000795, -0.00001312, -0.00000413, 0.00000292, 0.00000190, -0.00001229, 0.00000071, -0.00000976, -0.00000262, -0.00000196, -0.00000817, -0.00000393, -0.00001017, 0.00000275, -0.00000655, -0.00004750, -0.00004557, 0.00009081, 0.00005634, 0.00007869, 0.00004833, -0.00006989, -0.00002008, -0.00001217, -0.00001018, -0.00000621, -0.00000705, -0.00001193, -0.00000850, -0.00004682, -0.00001790, 0.00007094, 0.00003881, 0.00005950, 0.00005481, -0.00004473, -0.00003775, 0.00000721, -0.00000623, -0.00000897, -0.00000500, -0.00000480, 0.00000051, -0.00001209, 0.00000153, -0.00000718, -0.00000625, 0.00000193, 0.00000336, -0.00000802, -0.00000906, 0.00000694, -0.00000530, 0.00000343, -0.00001008, 0.00001386, 0.00000668, -0.00001020, 0.00000319, -0.00000184, 0.00000626, -0.00000080, 0.00000485, -0.00000519, -0.00000479, 0.00000410, 0.00000377, -0.00000888, -0.00000211, 0.00000402, 0.00000223, 0.00000224, 0.00000240, 0.00000162, 0.00000033, -0.00000583, -0.00000550, 0.00000242, -0.00000144, 0.00000580, -0.00000191, -0.00000258, -0.00000354, 0.00000578, 0.00000309, -0.00000317, -0.00000158, -0.00000341, 0.00000176, },
		{ 0.00000109, 0.00000358, -0.00000281, -0.00000241, -0.00000040, -0.00000280, 0.00000279, 0.00000571, -0.00000392, -0.00000378, 0.00000223, 0.00000262, -0.00000203, 0.00000231, -0.00000352, -0.00000729, 0.00000111, 0.00000059, 0.00000444, 0.00000062, 0.00000441, 0.00000388, -0.00000707, -0.00000606, 0.00000471, 0.00000389, -0.00000613, -0.00000251, 0.00000259, 0.00000013, 0.00000655, -0.00000280, 0.00000320, -0.00001099, 0.00001257, 0.00000894, -0.00000885, 0.00000290, -0.00000631, 0.00000815, -0.00001086, -0.00000655, 0.00000284, 0.00000284, -0.00001181, -0.00000055, -0.00000852, -0.00000426, -0.00000007, -0.00001007, -0.00000201, -0.00001189, 0.00000394, -0.00000731, -0.00003771, -0.00006186, 0.00008252, 0.00006764, 0.00007500, 0.00006265, -0.00007089, -0.00002435, -0.00001155, -0.00001117, -0.00000610, -0.00000692, -0.00001151, -0.00000789, -0.00004245, -0.00002399, 0.00006464, 0.00004134, 0.00005044, 0.00006194, -0.00003295, -0.00004431, 0.00000576, -0.00000403, -0.00001043, -0.00000439, -0.00000541, 0.00000140, -0.00001267, 0.00000129, -0.00000597, -0.00000708, 0.00000127, 0.00000351, -0.00000558, -0.00001134, 0.00000650, -0.00000411, 0.00000289, -0.00000893, 0.00000975, 0.00001089, -0.00001148, 0.00000315, -0.00000217, 0.00000605, -0.00000035, 0.00000460, -0.00000372, -0.00000588, 0.00000320, 0.00000469, -0.00000795, -0.00000339, 0.00000425, 0.00000181, 0.00000254, 0.00000205, 0.00000204, 0.00000044, -0.00000435, -0.00000722, 0.00000245, -0.00000143, 0.00000566, -0.00000116, -0.00000265, -0.00000360, 0.00000417, 0.00000491, -0.00000349, -0.00000140, -0.00000392, 0.00000204, },
		{ 0.00000080, 0.00000380, -0.00000240, -0.00000291, -0.00000003, -0.00000272, 0.00000157, 0.00000630, -0.00000279, -0.00000471, 0.00000222, 0.00000245, -0.00000161, 0.00000144, -0.00000136, -0.00000858, 0.00000090, 0.00000010, 0.00000473, 0.00000073, 0.00000375, 0.00000492, -0.00000639, -0.00000686, 0.00000365, 0.00000487, -0.00000511, -0.00000377, 0.00000268, -0.00000037, 0.00000689, -0.00000248, 0.00000315, -0.00000967, 0.00000827, 0.00001303, -0.00000983, 0.00000310, -0.00000644, 0.00000719, -0.00000788, -0.00000896, 0.00000259, 0.00000352, -0.00001087, -0.00000216, -0.00000694, -0.00000634, 0.00000225, -0.00001225, 0.00000011, -0.00001383, 0.00000552, -0.00000894, -0.00002784, -0.00007382, 0.00006824, 0.00008086, 0.00007077, 0.00007695, -0.00007025, -0.00002949, -0.00001080, -0.00001220, -0.00000609, -0.00000676, -0.00001118, -0.00000730, -0.00003847, -0.00002855, 0.00005738, 0.00004427, 0.00004227, 0.00006572, -0.00001998, -0.00004887, 0.00000243, -0.00000092, -0.00001225, -0.00000359, -0.00000625, 0.00000249, -0.00001326, 0.00000090, -0.00000473, -0.00000768, 0.00000052, 0.00000342, -0.00000320, -0.00001277, 0.00000505, -0.00000236, 0.00000198, -0.00000729, 0.00000554, 0.00001435, -0.00001173, 0.00000240, -0.00000198, 0.00000537, 0.00000044, 0.00000402, -0.00000211, -0.00000676, 0.00000219, 0.00000531, -0.00000671, -0.00000474, 0.00000452, 0.00000138, 0.00000276, 0.00000179, 0.00000235, 0.00000052, -0.00000289, -0.00000844, 0.00000185, -0.00000100, 0.00000515, -0.00000012, -0.00000298, -0.00000321, 0.00000223, 0.00000660, -0.00000354, -0.00000128, -0.00000444, 0.00000224, },
		{ 0.00000079, 0.00000371, -0.00000169, -0.00000360, 0.00000037, -0.00000261, 0.00000057, 0.00000634, -0.00000121, -0.00000576, 0.00000232, 0.00000212, -0.00000100, 0.00000038, 0.00000066, -0.00000932, 0.00000024, -0.00000008, 0.00000460, 0.00000130, 0.00000269, 0.00000607, -0.00000574, -0.00000738, 0.00000236, 0.00000569, -0.00000395, -0.00000477, 0.00000243, -0.00000060, 0.00000685, -0.00000175, 0.00000261, -0.00000779, 0.00000395, 0.00001618, -0.00000959, 0.00000228, -0.00000557, 0.00000532, -0.00000455, -0.00001102, 0.00000203, 0.00000404, -0.00000959, -0.00000397, -0.00000518, -0.00000865, 0.00000468, -0.00001436, 0.00000202, -0.00001554, 0.00000689, -0.00001046, -0.00001937, -0.00008020, 0.00004813, 0.00009495, 0.00006681, 0.00009063, -0.00006759, -0.00003559, -0.00000992, -0.00001324, -0.00000617, -0.00000654, -0.00001099, -0.00000666, -0.00003495, -0.00003166, 0.00004942, 0.00004723, 0.00003530, 0.00006660, -0.00000684, -0.00005081, -0.00000271, 0.00000278, -0.00001412, -0.00000286, -0.00000709, 0.00000356, -0.00001364, 0.00000019, -0.00000336, -0.00000817, -0.00000018, 0.00000306, -0.00000096, -0.00001335, 0.00000281, -0.00000040, 0.00000105, -0.00000552, 0.00000167, 0.00001672, -0.00001086, 0.00000094, -0.00000131, 0.00000430, 0.00000148, 0.00000324, -0.00000050, -0.00000734, 0.00000109, 0.00000563, -0.00000522, -0.00000604, 0.00000471, 0.00000106, 0.00000279, 0.00000170, 0.00000249, 0.00000064, -0.00000159, -0.00000903, 0.00000062, -0.00000023, 0.00000437, 0.00000111, -0.00000349, -0.00000253, 0.00000018, 0.00000794, -0.00000322, -0.00000128, -0.00000486, 0.00000224, },
		{ 0.00000110, 0.00000328, -0.00000070, -0.00000443, 0.00000074, -0.00000247, -0.00000017, 0.00000585, 0.00000065, -0.00000677, 0.00000235, 0.00000180, -0.00000039, -0.00000065, 0.00000227, -0.00000933, -0.00000094, 0.00000012, 0.00000400, 0.00000234, 0.00000130, 0.00000724, -0.00000505, -0.00000765, 0.00000094, 0.00000626, -0.00000270, -0.00000547, 0.00000188, -0.00000062, 0.00000653, -0.00000074, 0.00000176, -0.00000564, 0.00000003, 0.00001810, -0.00000806, 0.00000055, -0.00000395, 0.00000293, -0.00000137, -0.00001240, 0.00000100, 0.00000452, -0.00000817, -0.00000575, -0.00000346, -0.00001090, 0.00000686, -0.00001595, 0.00000325, -0.00001647, 0.00000738, -0.00001094, -0.00001352, -0.00008038, 0.00002311, 0.00010842, 0.00006408, 0.00010303, -0.00006249, -0.00004274, -0.00000892, -0.00001432, -0.00000633, -0.00000629, -0.00001091, -0.00000596, -0.00003190, -0.00003351, 0.00004108, 0.00004984, 0.00002973, 0.00006511, 0.00000559, -0.00004995, -0.00000922, 0.00000652, -0.00001557, -0.00000254, -0.00000766, 0.00000432, -0.00001354, -0.00000104, -0.00000176, -0.00000866, -0.00000071, 0.00000238, 0.00000108, -0.00001316, 0.00000011, 0.00000136, 0.00000043, -0.00000395, -0.00000152, 0.00001781, -0.00000889, -0.00000107, -0.00000031, 0.00000298, 0.00000262, 0.00000242, 0.00000097, -0.00000758, -0.00000008, 0.00000566, -0.00000360, -0.00000719, 0.00000471, 0.00000097, 0.00000253, 0.00000183, 0.00000245, 0.00000083, -0.00000053, -0.00000898, -0.00000111, 0.00000071, 0.00000348, 0.00000239, -0.00000402, -0.00000173, -0.00000172, 0.00000875, -0.00000243, -0.00000147, -0.00000510, 0.00000193, },
		{ 0.00000175, 0.00000257, 0.00000049, -0.00000526, 0.00000101, -0.00000229, -0.00000067, 0.00000498, 0.00000257, -0.00000752, 0.00000216, 0.00000163, 0.00000005, -0.00000140, 0.00000324, -0.00000851, -0.00000264, 0.00000068, 0.00000297, 0.00000375, -0.00000028, 0.00000827, -0.00000424, -0.00000775, -0.00000049, 0.00000653, -0.00000137, -0.00000586, 0.00000109, -0.00000054, 0.00000603, 0.00000041, 0.00000079, -0.00000353, -0.00000318, 0.00001864, -0.00000539, -0.00000178, -0.00000198, 0.00000055, 0.00000119, -0.00001279, -0.00000063, 0.00000506, -0.00000678, -0.00000727, -0.00000205, -0.00001277, 0.00000837, -0.00001661, 0.00000339, -0.00001618, 0.00000644, -0.00000967, -0.00001103, -0.00007449, -0.00000523, 0.00011942, 0.00006361, 0.00011352, -0.00005458, -0.00005098, -0.00000776, -0.00001550, -0.00000650, -0.00000609, -0.00001088, -0.00000528, -0.00002921, -0.00003436, 0.00003275, 0.00005177, 0.00002568, 0.00006173, 0.00001673, -0.00004646, -0.00001636, 0.00000958, -0.00001610, -0.00000298, -0.00000766, 0.00000450, -0.00001272, -0.00000291, 0.00000016, -0.00000925, -0.00000097, 0.00000137, 0.00000293, -0.00001235, -0.00000271, 0.00000256, 0.00000037, -0.00000280, -0.00000384, 0.00001762, -0.00000600, -0.00000341, 0.00000082, 0.00000161, 0.00000370, 0.00000169, 0.00000217, -0.00000744, -0.00000129, 0.00000543, -0.00000194, -0.00000809, 0.00000445, 0.00000118, 0.00000197, 0.00000214, 0.00000226, 0.00000108, 0.00000027, -0.00000835, -0.00000317, 0.00000161, 0.00000266, 0.00000355, -0.00000442, -0.00000103, -0.00000325, 0.00000887, -0.00000115, -0.00000183, -0.00000511, 0.00000128, },
		{ 0.00000266, 0.00000169, 0.00000171, -0.00000597, 0.00000108, -0.00000203, -0.00000097, 0.00000387, 0.00000434, -0.00000782, 0.00000160, 0.00000176, 0.00000017, -0.00000174, 0.00000348, -0.00000688, -0.00000475, 0.00000148, 0.00000163, 0.00000535, -0.00000180, 0.00000896, -0.00000320, -0.00000778, -0.00000180, 0.00000645, 0.00000002, -0.00000597, 0.00000016, -0.00000044, 0.00000545, 0.00000158, -0.00000011, -0.00000167, -0.00000550, 0.00001787, -0.00000188, -0.00000428, -0.00000015, -0.00000131, 0.00000271, -0.00001201, -0.00000288, 0.00000574, -0.00000559, -0.00000835, -0.00000114, -0.00001399, 0.00000893, -0.00001605, 0.00000216, -0.00001437, 0.00000376, -0.00000632, -0.00001204, -0.00006344, -0.00003467, 0.00012597, 0.00006629, 0.00012160, -0.00004359, -0.00006029, -0.00000648, -0.00001682, -0.00000658, -0.00000604, -0.00001078, -0.00000472, -0.00002674, -0.00003454, 0.00002482, 0.00005268, 0.00002321, 0.00005689, 0.00002627, -0.00004087, -0.00002321, 0.00001124, -0.00001526, -0.00000445, -0.00000690, 0.00000392, -0.00001105, -0.00000548, 0.00000240, -0.00000995, -0.00000092, 0.00000007, 0.00000454, -0.00001108, -0.00000530, 0.00000293, 0.00000099, -0.00000215, -0.00000525, 0.00001627, -0.00000250, -0.00000575, 0.00000183, 0.00000035, 0.00000457, 0.00000119, 0.00000304, -0.00000693, -0.00000249, 0.00000496, -0.00000033, -0.00000866, 0.00000387, 0.00000170, 0.00000118, 0.00000254, 0.00000203, 0.00000133, 0.00000086, -0.00000727, -0.00000529, 0.00000222, 0.00000207, 0.00000444, -0.00000453, -0.00000058, -0.00000426, 0.00000824, 0.00000055, -0.00000233, -0.00000488, 0.00000029, },
		{ 0.00000368, 0.00000081, 0.00000281, -0.00000640, 0.00000088, -0.00000166, -0.00000117, 0.00000271, 0.00000572, -0.00000754, 0.00000060, 0.00000224, -0.00000008, -0.00000159, 0.00000299, -0.00000461, -0.00000704, 0.00000235, 0.00000019, 0.00000691, -0.00000302, 0.00000913, -0.00000184, -0.00000784, -0.00000290, 0.00000597, 0.00000143, -0.00000582, -0.00000081, -0.00000044, 0.00000488, 0.00000266, -0.00000085, -0.00000020, -0.00000689, 0.00001600, 0.00000200, -0.00000645, 0.00000104, -0.00000221, 0.00000292, -0.00001006, -0.00000565, 0.00000650, -0.00000465, -0.00000892, -0.00000083, -0.00001442, 0.00000836, -0.00001414, -0.00000050, -0.00001106, -0.00000063, -0.00000100, -0.00001605, -0.00004881, -0.00006263, 0.00012620, 0.00007273, 0.00012698, -0.00002945, -0.00007053, -0.00000516, -0.00001830, -0.00000649, -0.00000624, -0.00001048, -0.00000444, -0.00002428, -0.00003439, 0.00001767, 0.00005231, 0.00002236, 0.00005094, 0.00003417, -0.00003393, -0.00002886, 0.00001094, -0.00001281, -0.00000704, -0.00000534, 0.00000253, -0.00000855, -0.00000865, 0.00000481, -0.00001070, -0.00000062, -0.00000143, 0.00000586, -0.00000950, -0.00000743, 0.00000237, 0.00000223, -0.00000191, -0.00000587, 0.00001406, 0.00000122, -0.00000780, 0.00000252, -0.00000061, 0.00000509, 0.00000100, 0.00000353, -0.00000611, -0.00000360, 0.00000427, 0.00000116, -0.00000886, 0.00000299, 0.00000247, 0.00000026, 0.00000289, 0.00000187, 0.00000149, 0.00000129, -0.00000592, -0.00000724, 0.00000236, 0.00000186, 0.00000497, -0.00000427, -0.00000047, -0.00000467, 0.00000693, 0.00000256, -0.00000286, -0.00000448, -0.00000095, },
		{ 0.00000465, 0.00000011, 0.00000362, -0.00000644, 0.00000035, -0.00000115, -0.00000133, 0.00000166, 0.00000658, -0.00000661, -0.00000083, 0.00000305, -0.00000068, -0.00000102, 0.00000191, -0.00000195, -0.00000924, 0.00000303, -0.00000111, 0.00000815, -0.00000369, 0.00000863, -0.00000011, -0.00000796, -0.00000370, 0.00000511, 0.00000282, -0.00000545, -0.00000171, -0.00000060, 0.00000438, 0.00000359, -0.00000134, 0.00000084, -0.00000743, 0.00001338, 0.00000575, -0.00000780, 0.00000122, -0.00000189, 0.00000180, -0.00000712, -0.00000868, 0.00000720, -0.00000390, -0.00000899, -0.00000109, -0.00001406, 0.00000670, -0.00001101, -0.00000435, -0.00000652, -0.00000631, 0.00000567, -0.00002200, -0.00003270, -0.00008646, 0.00011865, 0.00008302, 0.00012976, -0.00001233, -0.00008134, -0.00000398, -0.00001991, -0.00000623, -0.00000676, -0.00000990, -0.00000454, -0.00002167, -0.00003422, 0.00001161, 0.00005049, 0.00002307, 0.00004416, 0.00004055, -0.00002644, -0.00003254, 0.00000838, -0.00000876, -0.00001059, -0.00000315, 0.00000051, -0.00000540, -0.00001216, 0.00000717, -0.00001134, -0.00000021, -0.00000294, 0.00000678, -0.00000768, -0.00000895, 0.00000091, 0.00000388, -0.00000188, -0.00000594, 0.00001134, 0.00000476, -0.00000924, 0.00000271, -0.00000119, 0.00000522, 0.00000113, 0.00000368, -0.00000505, -0.00000456, 0.00000337, 0.00000247, -0.00000871, 0.00000186, 0.00000339, -0.00000064, 0.00000305, 0.00000190, 0.00000150, 0.00000166, -0.00000446, -0.00000877, 0.00000187, 0.00000208, 0.00000509, -0.00000363, -0.00000073, -0.00000450, 0.00000506, 0.00000465, -0.00000328, -0.00000400, -0.00000230, },
		{ 0.00000536, -0.00000024, 0.00000401, -0.00000602, -0.00000050, -0.00000051, -0.00000150, 0.00000083, 0.00000683, -0.00000506, -0.00000259, 0.00000408, -0.00000150, -0.00000016, 0.00000048, 0.00000076, -0.00001100, 0.00000326, -0.00000204, 0.00000884, -0.00000363, 0.00000738, 0.00000197, -0.00000815, -0.00000419, 0.00000392, 0.00000412, -0.00000488, -0.00000249, -0.00000096, 0.00000395, 0.00000436, -0.00000161, 0.00000149, -0.00000732, 0.00001042, 0.00000888, -0.00000798, 0.00000019, -0.00000036, -0.00000043, -0.00000361, -0.00001158, 0.00000760, -0.00000321, -0.00000873, -0.00000177, -0.00001309, 0.00000418, -0.00000700, -0.00000896, -0.00000132, -0.00001258, 0.00001277, -0.00002847, -0.00001736, -0.00010383, 0.00010254, 0.00009663, 0.00013042, 0.00000731, -0.00009219, -0.00000329, -0.00002149, -0.00000584, -0.00000757, -0.00000901, -0.00000510, -0.00001879, -0.00003425, 0.00000684, 0.00004715, 0.00002518, 0.00003685, 0.00004558, -0.00001912, -0.00003379, 0.00000365, -0.00000351, -0.00001466, -0.00000069, -0.00000183, -0.00000197, -0.00001562, 0.00000912, -0.00001161, 0.00000009, -0.00000425, 0.00000718, -0.00000566, -0.00000983, -0.00000126, 0.00000563, -0.00000177, -0.00000575, 0.00000851, 0.00000775, -0.00000988, 0.00000233, -0.00000134, 0.00000495, 0.00000157, 0.00000355, -0.00000388, -0.00000528, 0.00000229, 0.00000357, -0.00000821, 0.00000057, 0.00000434, -0.00000133, 0.00000287, 0.00000218, 0.00000131, 0.00000202, -0.00000305, -0.00000974, 0.00000073, 0.00000267, 0.00000486, -0.00000266, -0.00000129, -0.00000388, 0.00000285, 0.00000660, -0.00000344, -0.00000355, -0.00000360, },
		{ 0.00000567, -0.00000013, 0.00000393, -0.00000516, -0.00000162, 0.00000019, -0.00000170, 0.00000028, 0.00000650, -0.00000305, -0.00000449, 0.00000513, -0.00000235, 0.00000076, -0.00000101, 0.00000316, -0.00001203, 0.00000284, -0.00000239, 0.00000880, -0.00000274, 0.00000542, 0.00000429, -0.00000835, -0.00000441, 0.00000249, 0.00000524, -0.00000414, -0.00000309, -0.00000150, 0.00000356, 0.00000502, -0.00000169, 0.00000185, -0.00000680, 0.00000751, 0.00001103, -0.00000682, -0.00000200, 0.00000216, -0.00000331, -0.00000008, -0.00001386, 0.00000739, -0.00000235, -0.00000837, -0.00000259, -0.00001177, 0.00000117, -0.00000262, -0.00001368, 0.00000380, -0.00001855, 0.00001915, -0.00003390, -0.00000493, -0.00011301, 0.00007803, 0.00011231, 0.00012986, 0.00002874, -0.00010234, -0.00000353, -0.00002280, -0.00000548, -0.00000858, -0.00000787, -0.00000606, -0.00001562, -0.00003460, 0.00000340, 0.00004243, 0.00002842, 0.00002934, 0.00004941, -0.00001246, -0.00003255, -0.00000275, 0.00000228, -0.00001863, 0.00000151, -0.00000403, 0.00000127, -0.00001856, 0.00001029, -0.00001125, 0.00000006, -0.00000510, 0.00000693, -0.00000346, -0.00001016, -0.00000383, 0.00000708, -0.00000130, -0.00000556, 0.00000589, 0.00000993, -0.00000961, 0.00000139, -0.00000111, 0.00000433, 0.00000221, 0.00000326, -0.00000273, -0.00000569, 0.00000104, 0.00000443, -0.00000744, -0.00000079, 0.00000517, -0.00000169, 0.00000228, 0.00000275, 0.00000094, 0.00000237, -0.00000178, -0.00001007, -0.00000100, 0.00000351, 0.00000443, -0.00000150, -0.00000201, -0.00000299, 0.00000057, 0.00000816, -0.00000318, -0.00000324, -0.00000472, },
		{ 0.00000547, 0.00000048, 0.00000341, -0.00000395, -0.00000288, 0.00000087, -0.00000190, -0.00000001, 0.00000568, -0.00000077, -0.00000630, 0.00000600, -0.00000302, 0.00000150, -0.00000227, 0.00000494, -0.00001210, 0.00000164, -0.00000210, 0.00000799, -0.00000105, 0.00000290, 0.00000666, -0.00000846, -0.00000443, 0.00000095, 0.00000609, -0.00000321, -0.00000351, -0.00000217, 0.00000313, 0.00000561, -0.00000165, 0.00000205, -0.00000609, 0.00000498, 0.00001197, -0.00000436, -0.00000507, 0.00000519, -0.00000622, 0.00000281, -0.00001505, 0.00000627, -0.00000109, -0.00000817, -0.00000328, -0.00001046, -0.00000188, 0.00000153, -0.00001783, 0.00000806, -0.00002328, 0.00002371, -0.00003690, 0.00000296, -0.00011321, 0.00004630, 0.00012811, 0.00012928, 0.00005103, -0.00011091, -0.00000523, -0.00002353, -0.00000538, -0.00000960, -0.00000665, -0.00000731, -0.00001227, -0.00003521, 0.00000116, 0.00003662, 0.00003235, 0.00002204, 0.00005205, -0.00000664, -0.00002916, -0.00001000, 0.00000773, -0.00002175, 0.00000285, -0.00000557, 0.00000385, -0.00002053, 0.00001033, -0.00001002, -0.00000053, -0.00000531, 0.00000593, -0.00000109, -0.00001003, -0.00000647, 0.00000789, -0.00000022, -0.00000558, 0.00000371, 0.00001115, -0.00000847, -0.00000000, -0.00000061, 0.00000350, 0.00000295, 0.00000293, -0.00000171, -0.00000577, -0.00000031, 0.00000502, -0.00000644, -0.00000210, 0.00000575, -0.00000162, 0.00000126, 0.00000354, 0.00000048, 0.00000269, -0.00000067, -0.00000979, -0.00000316, 0.00000439, 0.00000397, -0.00000032, -0.00000272, -0.00000204, -0.00000150, 0.00000911, -0.00000242, -0.00000313, -0.00000554, },
		{ 0.00000474, 0.00000153, 0.00000256, -0.00000255, -0.00000411, 0.00000141, -0.00000204, -0.00000011, 0.00000454, 0.00000151, -0.00000775, 0.00000647, -0.00000331, 0.00000185, -0.00000304, 0.00000586, -0.00001112, -0.00000033, -0.00000118, 0.00000648, 0.00000126, 0.00000008, 0.00000885, -0.00000833, -0.00000439, -0.00000053, 0.00000657, -0.00000211, -0.00000375, -0.00000290, 0.00000260, 0.00000620, -0.00000158, 0.00000218, -0.00000538, 0.00000305, 0.00001168, -0.00000090, -0.00000853, 0.00000810, -0.00000846, 0.00000446, -0.00001475, 0.00000406, 0.00000074, -0.00000836, -0.00000356, -0.00000946, -0.00000454, 0.00000493, -0.00002080, 0.00001078, -0.00002603, 0.00002559, -0.00003649, 0.00000539, -0.00010469, 0.00000953, 0.00014160, 0.00013004, 0.00007319, -0.00011699, -0.00000894, -0.00002335, -0.00000583, -0.00001036, -0.00000557, -0.00000863, -0.00000893, -0.00003592, -0.00000015, 0.00003015, 0.00003642, 0.00001543, 0.00005341, -0.00000159, -0.00002432, -0.00001705, 0.00001187, -0.00002330, 0.00000281, -0.00000603, 0.00000536, -0.00002117, 0.00000901, -0.00000781, -0.00000179, -0.00000477, 0.00000419, 0.00000142, -0.00000958, -0.00000886, 0.00000777, 0.00000156, -0.00000588, 0.00000209, 0.00001143, -0.00000662, -0.00000164, -0.00000001, 0.00000259, 0.00000365, 0.00000268, -0.00000091, -0.00000550, -0.00000169, 0.00000532, -0.00000528, -0.00000329, 0.00000603, -0.00000110, -0.00000011, 0.00000441, 0.00000004, 0.00000289, 0.00000029, -0.00000900, -0.00000552, 0.00000506, 0.00000367, 0.00000070, -0.00000324, -0.00000123, -0.00000316, 0.00000931, -0.00000114, -0.00000324, -0.00000600, },
		{ 0.00000354, 0.00000288, 0.00000157, -0.00000115, -0.00000514, 0.00000169, -0.00000205, -0.00000010, 0.00000328, 0.00000353, -0.00000864, 0.00000636, -0.00000307, 0.00000167, -0.00000318, 0.00000583, -0.00000912, -0.00000293, 0.00000020, 0.00000446, 0.00000393, -0.00000272, 0.00001057, -0.00000780, -0.00000439, -0.00000181, 0.00000659, -0.00000084, -0.00000383, -0.00000360, 0.00000189, 0.00000683, -0.00000150, 0.00000231, -0.00000479, 0.00000176, 0.00001036, 0.00000310, -0.00001175, 0.00001022, -0.00000939, 0.00000438, -0.00001276, 0.00000070, 0.00000317, -0.00000905, -0.00000328, -0.00000899, -0.00000650, 0.00000717, -0.00002217, 0.00001155, -0.00002634, 0.00002432, -0.00003226, 0.00000231, -0.00008881, -0.00002927, 0.00015013, 0.00013339, 0.00009431, -0.00011974, -0.00001510, -0.00002197, -0.00000707, -0.00001063, -0.00000489, -0.00000975, -0.00000587, -0.00003646, -0.00000094, 0.00002358, 0.00004004, 0.00001004, 0.00005328, 0.00000302, -0.00001893, -0.00002280, 0.00001386, -0.00002272, 0.00000105, -0.00000515, 0.00000554, -0.00002031, 0.00000626, -0.00000463, -0.00000369, -0.00000350, 0.00000181, 0.00000395, -0.00000889, -0.00001076, 0.00000659, 0.00000401, -0.00000641, 0.00000098, 0.00001091, -0.00000433, -0.00000328, 0.00000052, 0.00000174, 0.00000420, 0.00000259, -0.00000037, -0.00000493, -0.00000303, 0.00000531, -0.00000400, -0.00000429, 0.00000596, -0.00000016, -0.00000169, 0.00000520, -0.00000022, 0.00000288, 0.00000116, -0.00000785, -0.00000781, 0.00000529, 0.00000370, 0.00000142, -0.00000344, -0.00000073, -0.00000425, 0.00000873, 0.00000059, -0.00000351, -0.00000612, },
		{ 0.00000200, 0.00000432, 0.00000067, 0.00000003, -0.00000579, 0.00000162, -0.00000189, -0.00000008, 0.00000208, 0.00000507, -0.00000880, 0.00000560, -0.00000226, 0.00000093, -0.00000267, 0.00000490, -0.00000631, -0.00000587, 0.00000177, 0.00000222, 0.00000660, -0.00000513, 0.00001156, -0.00000675, -0.00000455, -0.00000276, 0.00000612, 0.00000057, -0.00000378, -0.00000419, 0.00000094, 0.00000752, -0.00000144, 0.00000248, -0.00000434, 0.00000106, 0.00000832, 0.00000706, -0.00001409, 0.00001093, -0.00000853, 0.00000237, -0.00000913, -0.00000362, 0.00000608, -0.00001018, -0.00000246, -0.00000908, -0.00000764, 0.00000810, -0.00002179, 0.00001026, -0.00002416, 0.00001999, -0.00002451, -0.00000542, -0.00006782, -0.00006664, 0.00015117, 0.00014017, 0.00011378, -0.00011856, -0.00002398, -0.00001926, -0.00000926, -0.00001020, -0.00000481, -0.00001045, -0.00000335, -0.00003656, -0.00000163, 0.00001749, 0.00004258, 0.00000633, 0.00005144, 0.00000761, -0.00001391, -0.00002631, 0.00001313, -0.00001979, -0.00000249, -0.00000293, 0.00000440, -0.00001798, 0.00000227, -0.00000071, -0.00000606, -0.00000166, -0.00000100, 0.00000634, -0.00000799, -0.00001203, 0.00000437, 0.00000692, -0.00000698, 0.00000023, 0.00000984, -0.00000190, -0.00000466, 0.00000081, 0.00000106, 0.00000452, 0.00000271, -0.00000008, -0.00000415, -0.00000420, 0.00000494, -0.00000264, -0.00000508, 0.00000556, 0.00000109, -0.00000327, 0.00000570, -0.00000015, 0.00000257, 0.00000201, -0.00000651, -0.00000977, 0.00000488, 0.00000417, 0.00000176, -0.00000325, -0.00000059, -0.00000471, 0.00000743, 0.00000263, -0.00000382, -0.00000597, },
		{ 0.00000032, 0.00000562, 0.00000008, 0.00000081, -0.00000596, 0.00000117, -0.00000154, -0.00000013, 0.00000111, 0.00000600, -0.00000819, 0.00000423, -0.00000092, -0.00000031, -0.00000160, 0.00000328, -0.00000300, -0.00000880, 0.00000320, 0.00000011, 0.00000892, -0.00000681, 0.00001163, -0.00000511, -0.00000493, -0.00000331, 0.00000516, 0.00000205, -0.00000359, -0.00000462, -0.00000024, 0.00000822, -0.00000136, 0.00000263, -0.00000398, 0.00000078, 0.00000600, 0.00001040, -0.00001499, 0.00000984, -0.00000571, -0.00000148, -0.00000419, -0.00000849, 0.00000913, -0.00001153, -0.00000126, -0.00000961, -0.00000803, 0.00000778, -0.00001984, 0.00000717, -0.00001987, 0.00001315, -0.00001412, -0.00001626, -0.00004459, -0.00009903, 0.00014278, 0.00015062, 0.00013142, -0.00011316, -0.00003555, -0.00001530, -0.00001242, -0.00000900, -0.00000546, -0.00001054, -0.00000158, -0.00003593, -0.00000262, 0.00001237, 0.00004356, 0.00000468, 0.00004771, 0.00001259, -0.00001005, -0.00002695, 0.00000953, -0.00001466, -0.00000752, 0.00000035, 0.00000220, -0.00001450, -0.00000260, 0.00000356, -0.00000858, 0.00000045, -0.00000391, 0.00000833, -0.00000684, -0.00001266, 0.00000132, 0.00000994, -0.00000731, -0.00000039, 0.00000851, 0.00000034, -0.00000557, 0.00000075, 0.00000062, 0.00000458, 0.00000303, 0.00000001, -0.00000328, -0.00000509, 0.00000422, -0.00000125, -0.00000564, 0.00000490, 0.00000248, -0.00000464, 0.00000570, 0.00000037, 0.00000193, 0.00000287, -0.00000513, -0.00001120, 0.00000374, 0.00000508, 0.00000174, -0.00000270, -0.00000081, -0.00000461, 0.00000558, 0.00000473, -0.00000402, -0.00000568, },
		{ -0.00000133, 0.00000655, -0.00000003, 0.00000107, -0.00000561, 0.00000037, -0.00000103, -0.00000028, 0.00000043, 0.00000626, -0.00000687, 0.00000238, 0.00000077, -0.00000186, -0.00000019, 0.00000125, 0.00000038, -0.00001132, 0.00000416, -0.00000156, 0.00001056, -0.00000752, 0.00001067, -0.00000292, -0.00000552, -0.00000344, 0.00000378, 0.00000352, -0.00000328, -0.00000486, -0.00000163, 0.00000886, -0.00000117, 0.00000270, -0.00000361, 0.00000067, 0.00000383, 0.00001259, -0.00001410, 0.00000681, -0.00000105, -0.00000672, 0.00000142, -0.00001330, 0.00001186, -0.00001273, -0.00000004, -0.00001027, -0.00000795, 0.00000655, -0.00001675, 0.00000286, -0.00001418, 0.00000479, -0.00000248, -0.00002818, -0.00002220, -0.00012333, 0.00012392, 0.00016410, 0.00014759, -0.00010364, -0.00004947, -0.00001040, -0.00001637, -0.00000712, -0.00000682, -0.00000999, -0.00000068, -0.00003444, -0.00000422, 0.00000860, 0.00004266, 0.00000525, 0.00004209, 0.00001821, -0.00000785, -0.00002457, 0.00000338, -0.00000790, -0.00001345, 0.00000416, -0.00000057, -0.00001037, -0.00000779, 0.00000767, -0.00001083, 0.00000246, -0.00000652, 0.00000965, -0.00000537, -0.00001273, -0.00000227, 0.00001267, -0.00000710, -0.00000110, 0.00000720, 0.00000214, -0.00000589, 0.00000032, 0.00000042, 0.00000440, 0.00000350, -0.00000000, -0.00000246, -0.00000562, 0.00000313, 0.00000015, -0.00000599, 0.00000407, 0.00000385, -0.00000560, 0.00000510, 0.00000138, 0.00000098, 0.00000375, -0.00000381, -0.00001199, 0.00000188, 0.00000634, 0.00000146, -0.00000187, -0.00000130, -0.00000409, 0.00000343, 0.00000665, -0.00000394, -0.00000538, },
		{ -0.00000274, 0.00000694, 0.00000044, 0.00000081, -0.00000480, -0.00000069, -0.00000043, -0.00000051, 0.00000007, 0.00000592, -0.00000503, 0.00000027, 0.00000258, -0.00000347, 0.00000127, -0.00000082, 0.00000343, -0.00001304, 0.00000436, -0.00000253, 0.00001128, -0.00000714, 0.00000873, -0.00000030, -0.00000624, -0.00000322, 0.00000210, 0.00000484, -0.00000280, -0.00000490, -0.00000316, 0.00000934, -0.00000080, 0.00000259, -0.00000310, 0.00000050, 0.00000219, 0.00001331, -0.00001130, 0.00000205, 0.00000495, -0.00001264, 0.00000689, -0.00001735, 0.00001373, -0.00001334, 0.00000079, -0.00001065, -0.00000779, 0.00000490, -0.00001314, -0.00000191, -0.00000804, -0.00000389, 0.00000885, -0.00003904, -0.00000346, -0.00013733, 0.00009479, 0.00017903, 0.00016314, -0.00009047, -0.00006508, -0.00000512, -0.00002074, -0.00000480, -0.00000872, -0.00000889, -0.00000060, -0.00003205, -0.00000655, 0.00000634, 0.00003979, 0.00000796, 0.00003479, 0.00002445, -0.00000742, -0.00001948, -0.00000456, -0.00000041, -0.00001941, 0.00000776, -0.00000330, -0.00000623, -0.00001265, 0.00001104, -0.00001236, 0.00000395, -0.00000847, 0.00001007, -0.00000350, -0.00001241, -0.00000601, 0.00001469, -0.00000609, -0.00000209, 0.00000614, 0.00000333, -0.00000557, -0.00000043, 0.00000038, 0.00000407, 0.00000401, -0.00000001, -0.00000179, -0.00000571, 0.00000172, 0.00000149, -0.00000615, 0.00000315, 0.00000505, -0.00000600, 0.00000382, 0.00000282, -0.00000020, 0.00000458, -0.00000260, -0.00001211, -0.00000058, 0.00000775, 0.00000108, -0.00000093, -0.00000188, -0.00000336, 0.00000125, 0.00000813, -0.00000343, -0.00000519, },
		{ -0.00000381, 0.00000670, 0.00000150, 0.00000010, -0.00000368, -0.00000182, 0.00000012, -0.00000076, -0.00000001, 0.00000511, -0.00000290, -0.00000181, 0.00000424, -0.00000488, 0.00000250, -0.00000258, 0.00000576, -0.00001370, 0.00000362, -0.00000266, 0.00001102, -0.00000569, 0.00000599, 0.00000253, -0.00000696, -0.00000274, 0.00000029, 0.00000589, -0.00000213, -0.00000479, -0.00000472, 0.00000955, -0.00000014, 0.00000222, -0.00000234, 0.00000007, 0.00000135, 0.00001244, -0.00000682, -0.00000396, 0.00001153, -0.00001831, 0.00001132, -0.00001991, 0.00001422, -0.00001293, 0.00000081, -0.00001036, -0.00000799, 0.00000334, -0.00000970, -0.00000632, -0.00000242, -0.00001167, 0.00001836, -0.00004697, 0.00000949, -0.00014000, 0.00005695, 0.00019300, 0.00017931, -0.00007437, -0.00008150, -0.00000018, -0.00002503, -0.00000244, -0.00001085, -0.00000747, -0.00000118, -0.00002891, -0.00000956, 0.00000555, 0.00003514, 0.00001245, 0.00002628, 0.00003104, -0.00000845, -0.00001248, -0.00001315, 0.00000669, -0.00002445, 0.00001039, -0.00000530, -0.00000270, -0.00001657, 0.00001316, -0.00001279, 0.00000459, -0.00000944, 0.00000941, -0.00000118, -0.00001184, -0.00000952, 0.00001564, -0.00000413, -0.00000342, 0.00000541, 0.00000387, -0.00000471, -0.00000136, 0.00000040, 0.00000367, 0.00000448, 0.00000010, -0.00000140, -0.00000534, 0.00000005, 0.00000271, -0.00000612, 0.00000221, 0.00000596, -0.00000576, 0.00000193, 0.00000455, -0.00000143, 0.00000525, -0.00000147, -0.00001164, -0.00000342, 0.00000906, 0.00000082, -0.00000006, -0.00000237, -0.00000262, -0.00000068, 0.00000896, -0.00000240, -0.00000518, },
		{ -0.00000448, 0.00000582, 0.00000302, -0.00000086, -0.00000247, -0.00000284, 0.00000050, -0.00000095, 0.00000007, 0.00000403, -0.00000077, -0.00000358, 0.00000550, -0.00000582, 0.00000324, -0.00000375, 0.00000712, -0.00001315, 0.00000194, -0.00000199, 0.00000985, -0.00000336, 0.00000275, 0.00000526, -0.00000748, -0.00000217, -0.00000145, 0.00000653, -0.00000124, -0.00000455, -0.00000619, 0.00000936, 0.00000087, 0.00000153, -0.00000128, -0.00000072, 0.00000140, 0.00001010, -0.00000110, -0.00001045, 0.00001776, -0.00002274, 0.00001385, -0.00002039, 0.00001296, -0.00001119, -0.00000027, -0.00000908, -0.00000888, 0.00000234, -0.00000701, -0.00000969, 0.00000188, -0.00001756, 0.00002487, -0.00005064, 0.00001548, -0.00013175, 0.00001322, 0.00020295, 0.00019748, -0.00005613, -0.00009774, 0.00000363, -0.00002867, -0.00000047, -0.00001283, -0.00000604, -0.00000215, -0.00002525, -0.00001305, 0.00000596, 0.00002912, 0.00001814, 0.00001722, 0.00003746, -0.00001032, -0.00000464, -0.00002107, 0.00001224, -0.00002769, 0.00001137, -0.00000602, -0.00000034, -0.00001904, 0.00001367, -0.00001188, 0.00000418, -0.00000926, 0.00000764, 0.00000156, -0.00001115, -0.00001251, 0.00001531, -0.00000123, -0.00000504, 0.00000501, 0.00000385, -0.00000349, -0.00000230, 0.00000033, 0.00000333, 0.00000479, 0.00000040, -0.00000131, -0.00000455, -0.00000175, 0.00000372, -0.00000591, 0.00000132, 0.00000652, -0.00000489, -0.00000043, 0.00000634, -0.00000250, 0.00000563, -0.00000037, -0.00001070, -0.00000638, 0.00000999, 0.00000089, 0.00000054, -0.00000260, -0.00000209, -0.00000216, 0.00000901, -0.00000085, -0.00000537, },
		{ -0.00000478, 0.00000442, 0.00000481, -0.00000184, -0.00000141, -0.00000354, 0.00000057, -0.00000102, 0.00000021, 0.00000289, 0.00000108, -0.00000480, 0.00000615, -0.00000614, 0.00000333, -0.00000417, 0.00000741, -0.00001143, -0.00000055, -0.00000068, 0.00000798, -0.00000048, -0.00000058, 0.00000757, -0.00000763, -0.00000166, -0.00000293, 0.00000664, -0.00000011, -0.00000423, -0.00000748, 0.00000870, 0.00000226, 0.00000054, 0.00000007, -0.00000188, 0.00000227, 0.00000664, 0.00000516, -0.00001654, 0.00002268, -0.00002506, 0.00001388, -0.00001848, 0.00000981, -0.00000804, -0.00000258, -0.00000671, -0.00001062, 0.00000215, -0.00000543, -0.00001160, 0.00000436, -0.00002098, 0.00002777, -0.00004954, 0.00001448, -0.00011435, -0.00003254, 0.00020560, 0.00021883, -0.00003642, -0.00011290, 0.00000559, -0.00003116, 0.00000069, -0.00001431, -0.00000494, -0.00000319, -0.00002141, -0.00001671, 0.00000717, 0.00002231, 0.00002429, 0.00000842, 0.00004300, -0.00001219, 0.00000278, -0.00002708, 0.00001527, -0.00002846, 0.00001026, -0.00000514, 0.00000056, -0.00001981, 0.00001242, -0.00000960, 0.00000269, -0.00000794, 0.00000487, 0.00000455, -0.00001037, -0.00001479, 0.00001366, 0.00000243, -0.00000676, 0.00000478, 0.00000345, -0.00000216, -0.00000301, 0.00000005, 0.00000314, 0.00000489, 0.00000095, -0.00000153, -0.00000344, -0.00000355, 0.00000444, -0.00000549, 0.00000050, 0.00000672, -0.00000350, -0.00000301, 0.00000792, -0.00000318, 0.00000559, 0.00000077, -0.00000946, -0.00000916, 0.00001030, 0.00000144, 0.00000076, -0.00000246, -0.00000189, -0.00000306, 0.00000828, 0.00000112, -0.00000567, },
		{ -0.00000481, 0.00000266, 0.00000659, -0.00000256, -0.00000073, -0.00000376, 0.00000027, -0.00000090, 0.00000031, 0.00000188, 0.00000246, -0.00000531, 0.00000612, -0.00000578, 0.00000274, -0.00000382, 0.00000668, -0.00000876, -0.00000355, 0.00000095, 0.00000577, 0.00000254, -0.00000359, 0.00000914, -0.00000723, -0.00000137, -0.00000399, 0.00000615, 0.00000125, -0.00000387, -0.00000847, 0.00000751, 0.00000399, -0.00000069, 0.00000165, -0.00000332, 0.00000374, 0.00000257, 0.00001118, -0.00002135, 0.00002547, -0.00002465, 0.00001110, -0.00001419, 0.00000495, -0.00000366, -0.00000596, -0.00000337, -0.00001312, 0.00000278, -0.00000501, -0.00001197, 0.00000495, -0.00002187, 0.00002709, -0.00004398, 0.00000757, -0.00009069, -0.00007591, 0.00019794, 0.00024396, -0.00001550, -0.00012630, 0.00000520, -0.00003218, 0.00000074, -0.00001498, -0.00000442, -0.00000402, -0.00001770, -0.00002019, 0.00000872, 0.00001535, 0.00003011, 0.00000070, 0.00004695, -0.00001319, 0.00000866, -0.00003016, 0.00001519, -0.00002648, 0.00000697, -0.00000261, -0.00000004, -0.00001890, 0.00000957, -0.00000616, 0.00000034, -0.00000570, 0.00000141, 0.00000754, -0.00000946, -0.00001631, 0.00001085, 0.00000650, -0.00000825, 0.00000452, 0.00000293, -0.00000100, -0.00000332, -0.00000055, 0.00000314, 0.00000474, 0.00000172, -0.00000200, -0.00000215, -0.00000517, 0.00000477, -0.00000486, -0.00000027, 0.00000661, -0.00000176, -0.00000553, 0.00000900, -0.00000327, 0.00000503, 0.00000202, -0.00000808, -0.00001150, 0.00000980, 0.00000256, 0.00000056, -0.00000191, -0.00000205, -0.00000336, 0.00000686, 0.00000333, -0.00000597, },
		{ -0.00000471, 0.00000079, 0.00000809, -0.00000277, -0.00000063, -0.00000341, -0.00000042, -0.00000060, 0.00000030, 0.00000112, 0.00000324, -0.00000509, 0.00000545, -0.00000481, 0.00000155, -0.00000283, 0.00000516, -0.00000550, -0.00000665, 0.00000254, 0.00000360, 0.00000528, -0.00000588, 0.00000973, -0.00000620, -0.00000137, -0.00000452, 0.00000505, 0.00000277, -0.00000348, -0.00000912, 0.00000581, 0.00000599, -0.00000206, 0.00000330, -0.00000484, 0.00000549, -0.00000153, 0.00001619, -0.00002414, 0.00002557, -0.00002128, 0.00000565, -0.00000795, -0.00000112, 0.00000149, -0.00000999, 0.00000053, -0.00001604, 0.00000395, -0.00000551, -0.00001111, 0.00000401, -0.00002063, 0.00002349, -0.00003508, -0.00000328, -0.00006441, -0.00011251, 0.00017779, 0.00027254, 0.00000695, -0.00013766, 0.00000228, -0.00003166, -0.00000040, -0.00001475, -0.00000463, -0.00000448, -0.00001436, -0.00002320, 0.00001016, 0.00000885, 0.00003489, -0.00000523, 0.00004872, -0.00001260, 0.00001212, -0.00002974, 0.00001188, -0.00002195, 0.00000183, 0.00000123, -0.00000186, -0.00001663, 0.00000550, -0.00000198, -0.00000251, -0.00000292, -0.00000232, 0.00001016, -0.00000826, -0.00001717, 0.00000718, 0.00001053, -0.00000917, 0.00000397, 0.00000255, -0.00000024, -0.00000308, -0.00000149, 0.00000334, 0.00000438, 0.00000266, -0.00000260, -0.00000085, -0.00000644, 0.00000462, -0.00000400, -0.00000098, 0.00000627, 0.00000015, -0.00000772, 0.00000936, -0.00000265, 0.00000392, 0.00000337, -0.00000670, -0.00001323, 0.00000843, 0.00000420, -0.00000002, -0.00000102, -0.00000253, -0.00000317, 0.00000496, 0.00000554, -0.00000607, },
		{ -0.00000463, -0.00000098, 0.00000905, -0.00000229, -0.00000116, -0.00000251, -0.00000141, -0.00000018, 0.00000019, 0.00000068, 0.00000340, -0.00000422, 0.00000427, -0.00000340, -0.00000003, -0.00000144, 0.00000319, -0.00000207, -0.00000940, 0.00000368, 0.00000185, 0.00000735, -0.00000715, 0.00000924, -0.00000455, -0.00000169, -0.00000450, 0.00000342, 0.00000435, -0.00000304, -0.00000940, 0.00000367, 0.00000811, -0.00000339, 0.00000485, -0.00000624, 0.00000715, -0.00000509, 0.00001957, -0.00002444, 0.00002282, -0.00001515, -0.00000189, -0.00000053, -0.00000759, 0.00000673, -0.00001405, 0.00000441, -0.00001881, 0.00000515, -0.00000639, -0.00000962, 0.00000222, -0.00001812, 0.00001811, -0.00002447, -0.00001559, -0.00003928, -0.00013856, 0.00014425, 0.00030311, 0.00003188, -0.00014709, -0.00000300, -0.00002982, -0.00000257, -0.00001370, -0.00000551, -0.00000454, -0.00001151, -0.00002555, 0.00001119, 0.00000328, 0.00003810, -0.00000890, 0.00004795, -0.00001000, 0.00001274, -0.00002582, 0.00000577, -0.00001551, -0.00000449, 0.00000577, -0.00000433, -0.00001353, 0.00000082, 0.00000233, -0.00000533, -0.00000008, -0.00000581, 0.00001204, -0.00000662, -0.00001754, 0.00000307, 0.00001401, -0.00000917, 0.00000292, 0.00000251, -0.00000005, -0.00000227, -0.00000273, 0.00000365, 0.00000388, 0.00000366, -0.00000320, 0.00000029, -0.00000723, 0.00000396, -0.00000292, -0.00000166, 0.00000580, 0.00000201, -0.00000933, 0.00000884, -0.00000130, 0.00000235, 0.00000476, -0.00000537, -0.00001426, 0.00000623, 0.00000623, -0.00000082, 0.00000009, -0.00000318, -0.00000265, 0.00000285, 0.00000745, -0.00000580, },
		{ -0.00000470, -0.00000243, 0.00000929, -0.00000104, -0.00000231, -0.00000119, -0.00000255, 0.00000025, 0.00000000, 0.00000053, 0.00000303, -0.00000290, 0.00000285, -0.00000182, -0.00000172, 0.00000003, 0.00000114, 0.00000106, -0.00001139, 0.00000404, 0.00000082, 0.00000850, -0.00000725, 0.00000768, -0.00000240, -0.00000226, -0.00000397, 0.00000139, 0.00000586, -0.00000252, -0.00000934, 0.00000119, 0.00001018, -0.00000450, 0.00000611, -0.00000728, 0.00000840, -0.00000763, 0.00002089, -0.00002213, 0.00001744, -0.00000689, -0.00001058, 0.00000702, -0.00001354, 0.00001123, -0.00001742, 0.00000755, -0.00002077, 0.00000575, -0.00000698, -0.00000829, 0.00000047, -0.00001537, 0.00001231, -0.00001399, -0.00002678, -0.00001871, -0.00015155, 0.00009809, 0.00033293, 0.00006089, -0.00015508, -0.00001014, -0.00002717, -0.00000540, -0.00001213, -0.00000684, -0.00000433, -0.00000910, -0.00002721, 0.00001161, -0.00000107, 0.00003944, -0.00001013, 0.00004463, -0.00000534, 0.00001060, -0.00001893, -0.00000220, -0.00000818, -0.00001104, 0.00001019, -0.00000676, -0.00001031, -0.00000376, 0.00000613, -0.00000758, 0.00000229, -0.00000855, 0.00001281, -0.00000435, -0.00001764, -0.00000104, 0.00001648, -0.00000801, 0.00000125, 0.00000293, -0.00000049, -0.00000094, -0.00000412, 0.00000395, 0.00000337, 0.00000460, -0.00000364, 0.00000112, -0.00000744, 0.00000281, -0.00000169, -0.00000230, 0.00000526, 0.00000365, -0.00001019, 0.00000741, 0.00000068, 0.00000046, 0.00000609, -0.00000409, -0.00001459, 0.00000336, 0.00000839, -0.00000163, 0.00000121, -0.00000382, -0.00000203, 0.00000082, 0.00000882, -0.00000503, },
		{ -0.00000500, -0.00000345, 0.00000872, 0.00000092, -0.00000390, 0.00000032, -0.00000362, 0.00000054, -0.00000017, 0.00000059, 0.00000232, -0.00000139, 0.00000147, -0.00000035, -0.00000322, 0.00000126, -0.00000060, 0.00000350, -0.00001232, 0.00000343, 0.00000065, 0.00000863, -0.00000621, 0.00000526, 0.00000002, -0.00000295, -0.00000306, -0.00000086, 0.00000713, -0.00000185, -0.00000899, -0.00000147, 0.00001201, -0.00000523, 0.00000693, -0.00000779, 0.00000896, -0.00000887, 0.00002005, -0.00001743, 0.00001009, 0.00000253, -0.00001925, 0.00001354, -0.00001797, 0.00001422, -0.00001939, 0.00000929, -0.00002130, 0.00000515, -0.00000661, -0.00000786, -0.00000038, -0.00001340, 0.00000740, -0.00000534, -0.00003468, -0.00000515, -0.00015058, 0.00004188, 0.00035817, 0.00009593, -0.00016231, -0.00001845, -0.00002438, -0.00000834, -0.00001047, -0.00000826, -0.00000413, -0.00000699, -0.00002827, 0.00001143, -0.00000415, 0.00003890, -0.00000905, 0.00003907, 0.00000101, 0.00000625, -0.00001010, -0.00001076, -0.00000118, -0.00001678, 0.00001365, -0.00000841, -0.00000766, -0.00000756, 0.00000882, -0.00000877, 0.00000375, -0.00001014, 0.00001224, -0.00000138, -0.00001766, -0.00000477, 0.00001760, -0.00000560, -0.00000100, 0.00000378, -0.00000146, 0.00000072, -0.00000548, 0.00000410, 0.00000298, 0.00000533, -0.00000381, 0.00000151, -0.00000704, 0.00000122, -0.00000037, -0.00000288, 0.00000473, 0.00000497, -0.00001024, 0.00000516, 0.00000307, -0.00000149, 0.00000716, -0.00000280, -0.00001432, 0.00000009, 0.00001040, -0.00000220, 0.00000212, -0.00000423, -0.00000154, -0.00000086, 0.00000945, -0.00000367, },
		{ -0.00000550, -0.00000400, 0.00000739, 0.00000340, -0.00000567, 0.00000174, -0.00000436, 0.00000052, -0.00000023, 0.00000074, 0.00000147, 0.00000001, 0.00000041, 0.00000075, -0.00000427, 0.00000200, -0.00000175, 0.00000498, -0.00001203, 0.00000182, 0.00000133, 0.00000783, -0.00000424, 0.00000229, 0.00000239, -0.00000357, -0.00000193, -0.00000309, 0.00000799, -0.00000099, -0.00000841, -0.00000415, 0.00001342, -0.00000544, 0.00000718, -0.00000764, 0.00000871, -0.00000869, 0.00001719, -0.00001092, 0.00000166, 0.00001192, -0.00002666, 0.00001796, -0.00002008, 0.00001508, -0.00001944, 0.00000914, -0.00001995, 0.00000291, -0.00000480, -0.00000890, 0.00000030, -0.00001294, 0.00000433, 0.00000026, -0.00003792, 0.00000025, -0.00013658, -0.00002014, 0.00037424, 0.00013900, -0.00016936, -0.00002723, -0.00002213, -0.00001088, -0.00000917, -0.00000938, -0.00000423, -0.00000492, -0.00002891, 0.00001077, -0.00000607, 0.00003672, -0.00000608, 0.00003188, 0.00000836, 0.00000061, -0.00000064, -0.00001850, 0.00000428, -0.00002077, 0.00001540, -0.00000868, -0.00000615, -0.00001004, 0.00000997, -0.00000860, 0.00000404, -0.00001035, 0.00001025, 0.00000225, -0.00001768, -0.00000783, 0.00001723, -0.00000204, -0.00000366, 0.00000492, -0.00000278, 0.00000246, -0.00000656, 0.00000391, 0.00000282, 0.00000577, -0.00000362, 0.00000143, -0.00000607, -0.00000066, 0.00000091, -0.00000336, 0.00000421, 0.00000590, -0.00000952, 0.00000231, 0.00000558, -0.00000321, 0.00000779, -0.00000142, -0.00001361, -0.00000329, 0.00001195, -0.00000231, 0.00000263, -0.00000426, -0.00000137, -0.00000200, 0.00000922, -0.00000175, },
		{ -0.00000611, -0.00000413, 0.00000544, 0.00000614, -0.00000730, 0.00000276, -0.00000456, 0.00000009, -0.00000010, 0.00000085, 0.00000069, 0.00000106, -0.00000010, 0.00000131, -0.00000473, 0.00000209, -0.00000217, 0.00000541, -0.00001057, -0.00000064, 0.00000265, 0.00000634, -0.00000170, -0.00000078, 0.00000438, -0.00000391, -0.00000077, -0.00000507, 0.00000828, 0.00000011, -0.00000769, -0.00000669, 0.00001424, -0.00000502, 0.00000681, -0.00000682, 0.00000762, -0.00000721, 0.00001275, -0.00000341, -0.00000674, 0.00002008, -0.00003168, 0.00001946, -0.00001939, 0.00001353, -0.00001735, 0.00000691, -0.00001659, -0.00000108, -0.00000138, -0.00001163, 0.00000273, -0.00001430, 0.00000353, 0.00000234, -0.00003611, -0.00000218, -0.00011227, -0.00008244, 0.00037637, 0.00019161, -0.00017637, -0.00003607, -0.00002091, -0.00001259, -0.00000860, -0.00000987, -0.00000492, -0.00000269, -0.00002935, 0.00000981, -0.00000705, 0.00003329, -0.00000182, 0.00002387, 0.00001585, -0.00000521, 0.00000807, -0.00002411, 0.00000724, -0.00002234, 0.00001500, -0.00000724, -0.00000606, -0.00001095, 0.00000943, -0.00000700, 0.00000309, -0.00000918, 0.00000699, 0.00000633, -0.00001769, -0.00001012, 0.00001541, 0.00000236, -0.00000638, 0.00000607, -0.00000416, 0.00000395, -0.00000712, 0.00000325, 0.00000300, 0.00000584, -0.00000303, 0.00000089, -0.00000466, -0.00000267, 0.00000201, -0.00000366, 0.00000371, 0.00000647, -0.00000817, -0.00000085, 0.00000784, -0.00000441, 0.00000781, 0.00000013, -0.00001263, -0.00000648, 0.00001276, -0.00000178, 0.00000264, -0.00000379, -0.00000161, -0.00000251, 0.00000818, 0.00000060, },
		{ -0.00000667, -0.00000397, 0.00000313, 0.00000879, -0.00000843, 0.00000312, -0.00000406, -0.00000082, 0.00000026, 0.00000084, 0.00000014, 0.00000157, 0.00000003, 0.00000126, -0.00000457, 0.00000153, -0.00000186, 0.00000488, -0.00000816, -0.00000361, 0.00000426, 0.00000455, 0.00000096, -0.00000349, 0.00000564, -0.00000380, 0.00000023, -0.00000659, 0.00000790, 0.00000146, -0.00000690, -0.00000892, 0.00001436, -0.00000396, 0.00000585, -0.00000538, 0.00000582, -0.00000470, 0.00000732, 0.00000420, -0.00001405, 0.00002598, -0.00003355, 0.00001764, -0.00001583, 0.00000969, -0.00001327, 0.00000279, -0.00001146, -0.00000657, 0.00000342, -0.00001581, 0.00000665, -0.00001719, 0.00000475, 0.00000125, -0.00003000, -0.00001073, -0.00008177, -0.00013887, 0.00036031, 0.00025423, -0.00018273, -0.00004498, -0.00002088, -0.00001337, -0.00000889, -0.00000958, -0.00000631, -0.00000021, -0.00002971, 0.00000871, -0.00000737, 0.00002903, 0.00000306, 0.00001586, 0.00002256, -0.00001014, 0.00001480, -0.00002664, 0.00000716, -0.00002127, 0.00001241, -0.00000412, -0.00000739, -0.00001032, 0.00000734, -0.00000419, 0.00000113, -0.00000687, 0.00000283, 0.00001051, -0.00001752, -0.00001170, 0.00001243, 0.00000712, -0.00000873, 0.00000690, -0.00000528, 0.00000491, -0.00000695, 0.00000204, 0.00000352, 0.00000555, -0.00000208, -0.00000001, -0.00000297, -0.00000460, 0.00000279, -0.00000373, 0.00000318, 0.00000675, -0.00000639, -0.00000398, 0.00000952, -0.00000484, 0.00000711, 0.00000189, -0.00001151, -0.00000922, 0.00001268, -0.00000056, 0.00000211, -0.00000283, -0.00000228, -0.00000241, 0.00000646, 0.00000317, },
		{ -0.00000700, -0.00000368, 0.00000071, 0.00001100, -0.00000878, 0.00000262, -0.00000281, -0.00000216, 0.00000080, 0.00000067, -0.00000008, 0.00000149, 0.00000080, 0.00000070, -0.00000387, 0.00000044, -0.00000099, 0.00000364, -0.00000521, -0.00000665, 0.00000574, 0.00000290, 0.00000327, -0.00000542, 0.00000592, -0.00000312, 0.00000094, -0.00000748, 0.00000679, 0.00000306, -0.00000608, -0.00001073, 0.00001374, -0.00000228, 0.00000439, -0.00000345, 0.00000348, -0.00000152, 0.00000157, 0.00001101, -0.00001939, 0.00002894, -0.00003195, 0.00001262, -0.00000985, 0.00000411, -0.00000778, -0.00000265, -0.00000516, -0.00001292, 0.00000897, -0.00002078, 0.00001133, -0.00002083, 0.00000713, -0.00000190, -0.00002121, -0.00002261, -0.00005006, -0.00018340, 0.00032310, 0.00032586, -0.00018683, -0.00005462, -0.00002171, -0.00001348, -0.00000985, -0.00000866, -0.00000828, 0.00000244, -0.00002997, 0.00000752, -0.00000720, 0.00002436, 0.00000797, 0.00000858, 0.00002775, -0.00001335, 0.00001876, -0.00002565, 0.00000405, -0.00001784, 0.00000802, 0.00000029, -0.00000975, -0.00000854, 0.00000414, -0.00000063, -0.00000141, -0.00000389, -0.00000171, 0.00001430, -0.00001690, -0.00001278, 0.00000873, 0.00001164, -0.00001024, 0.00000709, -0.00000583, 0.00000507, -0.00000596, 0.00000027, 0.00000436, 0.00000496, -0.00000085, -0.00000115, -0.00000119, -0.00000624, 0.00000313, -0.00000351, 0.00000259, 0.00000681, -0.00000440, -0.00000675, 0.00001035, -0.00000438, 0.00000570, 0.00000384, -0.00001036, -0.00001136, 0.00001164, 0.00000128, 0.00000114, -0.00000148, -0.00000327, -0.00000184, 0.00000430, 0.00000567, },
		{ -0.00000688, -0.00000346, -0.00000154, 0.00001247, -0.00000816, 0.00000121, -0.00000089, -0.00000379, 0.00000143, 0.00000039, 0.00000003, 0.00000087, 0.00000206, -0.00000019, -0.00000288, -0.00000093, 0.00000015, 0.00000206, -0.00000219, -0.00000927, 0.00000665, 0.00000181, 0.00000482, -0.00000626, 0.00000511, -0.00000187, 0.00000129, -0.00000766, 0.00000498, 0.00000481, -0.00000528, -0.00001204, 0.00001236, -0.00000008, 0.00000256, -0.00000120, 0.00000088, 0.00000191, -0.00000386, 0.00001629, -0.00002220, 0.00002871, -0.00002707, 0.00000505, -0.00000234, -0.00000228, -0.00000175, -0.00000852, 0.00000137, -0.00001921, 0.00001432, -0.00002552, 0.00001565, -0.00002404, 0.00000936, -0.00000548, -0.00001202, -0.00003446, -0.00002218, -0.00021104, 0.00026374, 0.00040363, -0.00018590, -0.00006631, -0.00002256, -0.00001359, -0.00001096, -0.00000751, -0.00001046, 0.00000496, -0.00002995, 0.00000613, -0.00000659, 0.00001950, 0.00001250, 0.00000253, 0.00003098, -0.00001438, 0.00001965, -0.00002135, -0.00000147, -0.00001283, 0.00000260, 0.00000526, -0.00001250, -0.00000624, 0.00000050, 0.00000301, -0.00000394, -0.00000083, -0.00000597, 0.00001716, -0.00001552, -0.00001367, 0.00000483, 0.00001532, -0.00001048, 0.00000637, -0.00000559, 0.00000430, -0.00000412, -0.00000197, 0.00000541, 0.00000417, 0.00000053, -0.00000236, 0.00000047, -0.00000743, 0.00000296, -0.00000301, 0.00000193, 0.00000675, -0.00000241, -0.00000890, 0.00001018, -0.00000303, 0.00000368, 0.00000584, -0.00000919, -0.00001284, 0.00000974, 0.00000356, -0.00000007, 0.00000008, -0.00000440, -0.00000102, 0.00000202, 0.00000779, },
		{ -0.00000617, -0.00000342, -0.00000339, 0.00001301, -0.00000650, -0.00000100, 0.00000149, -0.00000548, 0.00000196, 0.00000007, 0.00000038, -0.00000014, 0.00000353, -0.00000110, -0.00000187, -0.00000225, 0.00000122, 0.00000056, 0.00000044, -0.00001102, 0.00000662, 0.00000159, 0.00000534, -0.00000584, 0.00000323, -0.00000014, 0.00000129, -0.00000714, 0.00000257, 0.00000661, -0.00000448, -0.00001282, 0.00001028, 0.00000250, 0.00000052, 0.00000116, -0.00000176, 0.00000519, -0.00000843, 0.00001958, -0.00002228, 0.00002553, -0.00001964, -0.00000399, 0.00000548, -0.00000833, 0.00000375, -0.00001378, 0.00000709, -0.00002434, 0.00001839, -0.00002889, 0.00001837, -0.00002546, 0.00000995, -0.00000765, -0.00000483, -0.00004296, -0.00000250, -0.00021858, 0.00018370, 0.00048268, -0.00017617, -0.00008186, -0.00002224, -0.00001470, -0.00001144, -0.00000679, -0.00001226, 0.00000684, -0.00002925, 0.00000429, -0.00000543, 0.00001454, 0.00001647, -0.00000212, 0.00003217, -0.00001325, 0.00001775, -0.00001450, -0.00000830, -0.00000736, -0.00000280, 0.00000986, -0.00001483, -0.00000419, -0.00000281, 0.00000601, -0.00000583, 0.00000169, -0.00000932, 0.00001859, -0.00001309, -0.00001468, 0.00000126, 0.00001764, -0.00000917, 0.00000461, -0.00000447, 0.00000259, -0.00000156, -0.00000449, 0.00000650, 0.00000333, 0.00000190, -0.00000348, 0.00000184, -0.00000805, 0.00000226, -0.00000225, 0.00000122, 0.00000664, -0.00000061, -0.00001027, 0.00000898, -0.00000092, 0.00000130, 0.00000771, -0.00000793, -0.00001370, 0.00000715, 0.00000601, -0.00000128, 0.00000161, -0.00000542, -0.00000021, -0.00000008, 0.00000926, },
		{ -0.00000480, -0.00000366, -0.00000469, 0.00001253, -0.00000389, -0.00000379, 0.00000402, -0.00000692, 0.00000220, -0.00000014, 0.00000083, -0.00000130, 0.00000492, -0.00000171, -0.00000117, -0.00000323, 0.00000188, -0.00000048, 0.00000225, -0.00001157, 0.00000543, 0.00000239, 0.00000474, -0.00000423, 0.00000049, 0.00000184, 0.00000106, -0.00000598, -0.00000026, 0.00000831, -0.00000367, -0.00001305, 0.00000760, 0.00000531, -0.00000156, 0.00000348, -0.00000421, 0.00000804, -0.00001180, 0.00002072, -0.00001989, 0.00002010, -0.00001077, -0.00001313, 0.00001226, -0.00001283, 0.00000768, -0.00001743, 0.00001095, -0.00002731, 0.00002015, -0.00002979, 0.00001831, -0.00002388, 0.00000759, -0.00000681, -0.00000171, -0.00004545, 0.00000598, -0.00020516, 0.00008708, 0.00055643, -0.00015321, -0.00010332, -0.00001939, -0.00001793, -0.00001038, -0.00000730, -0.00001298, 0.00000747, -0.00002741, 0.00000161, -0.00000350, 0.00000936, 0.00001993, -0.00000548, 0.00003161, -0.00001041, 0.00001384, -0.00000635, -0.00001502, -0.00000276, -0.00000706, 0.00001318, -0.00001592, -0.00000313, -0.00000505, 0.00000773, -0.00000650, 0.00000312, -0.00001124, 0.00001823, -0.00000943, -0.00001607, -0.00000154, 0.00001822, -0.00000622, 0.00000188, -0.00000255, 0.00000007, 0.00000149, -0.00000705, 0.00000744, 0.00000258, 0.00000313, -0.00000435, 0.00000279, -0.00000806, 0.00000110, -0.00000132, 0.00000049, 0.00000649, 0.00000092, -0.00001081, 0.00000690, 0.00000166, -0.00000113, 0.00000919, -0.00000649, -0.00001407, 0.00000417, 0.00000828, -0.00000220, 0.00000286, -0.00000613, 0.00000034, -0.00000172, 0.00000989, },
		{ -0.00000281, -0.00000414, -0.00000542, 0.00001110, -0.00000059, -0.00000679, 0.00000630, -0.00000781, 0.00000195, -0.00000011, 0.00000122, -0.00000236, 0.00000591, -0.00000173, -0.00000103, -0.00000361, 0.00000188, -0.00000079, 0.00000299, -0.00001079, 0.00000307, 0.00000418, 0.00000313, -0.00000166, -0.00000275, 0.00000377, 0.00000073, -0.00000435, -0.00000329, 0.00000975, -0.00000281, -0.00001278, 0.00000447, 0.00000815, -0.00000351, 0.00000559, -0.00000630, 0.00001025, -0.00001383, 0.00001986, -0.00001559, 0.00001340, -0.00000176, -0.00002096, 0.00001674, -0.00001480, 0.00000919, -0.00001867, 0.00001219, -0.00002738, 0.00001888, -0.00002745, 0.00001469, -0.00001852, 0.00000143, -0.00000192, -0.00000394, -0.00004046, 0.00000215, -0.00017256, -0.00001951, 0.00061698, -0.00011247, -0.00013246, -0.00001286, -0.00002425, -0.00000697, -0.00000974, -0.00001196, 0.00000629, -0.00002396, -0.00000226, -0.00000051, 0.00000377, 0.00002310, -0.00000791, 0.00002987, -0.00000665, 0.00000905, 0.00000164, -0.00002016, -0.00000023, -0.00000924, 0.00001443, -0.00001516, -0.00000365, -0.00000570, 0.00000772, -0.00000562, 0.00000312, -0.00001142, 0.00001596, -0.00000455, -0.00001792, -0.00000332, 0.00001696, -0.00000180, -0.00000159, -0.00000005, -0.00000298, 0.00000471, -0.00000936, 0.00000802, 0.00000207, 0.00000411, -0.00000490, 0.00000327, -0.00000750, -0.00000039, -0.00000037, -0.00000016, 0.00000631, 0.00000216, -0.00001059, 0.00000419, 0.00000434, -0.00000324, 0.00001004, -0.00000475, -0.00001411, 0.00000112, 0.00001006, -0.00000258, 0.00000362, -0.00000632, 0.00000045, -0.00000271, 0.00000959, },
		{ -0.00000036, -0.00000477, -0.00000563, 0.00000890, 0.00000308, -0.00000959, 0.00000797, -0.00000787, 0.00000108, 0.00000024, 0.00000141, -0.00000309, 0.00000624, -0.00000097, -0.00000158, -0.00000326, 0.00000110, -0.00000025, 0.00000262, -0.00000875, -0.00000026, 0.00000669, 0.00000084, 0.00000146, -0.00000601, 0.00000529, 0.00000050, -0.00000244, -0.00000626, 0.00001073, -0.00000185, -0.00001207, 0.00000105, 0.00001085, -0.00000521, 0.00000739, -0.00000795, 0.00001174, -0.00001458, 0.00001738, -0.00001020, 0.00000659, 0.00000606, -0.00002625, 0.00001798, -0.00001361, 0.00000785, -0.00001711, 0.00001046, -0.00002423, 0.00001430, -0.00002162, 0.00000729, -0.00000924, -0.00000863, 0.00000718, -0.00001170, -0.00002798, -0.00001301, -0.00012509, -0.00012749, 0.00065604, -0.00005009, -0.00017026, -0.00000207, -0.00003420, -0.00000077, -0.00001454, -0.00000878, 0.00000291, -0.00001862, -0.00000760, 0.00000372, -0.00000241, 0.00002624, -0.00000986, 0.00002766, -0.00000289, 0.00000458, 0.00000805, -0.00002246, -0.00000067, -0.00000874, 0.00001323, -0.00001224, -0.00000598, -0.00000454, 0.00000584, -0.00000311, 0.00000163, -0.00000983, 0.00001191, 0.00000132, -0.00002016, -0.00000402, 0.00001399, 0.00000370, -0.00000535, 0.00000267, -0.00000619, 0.00000771, -0.00001111, 0.00000808, 0.00000191, 0.00000477, -0.00000508, 0.00000330, -0.00000648, -0.00000202, 0.00000044, -0.00000064, 0.00000604, 0.00000313, -0.00000978, 0.00000120, 0.00000671, -0.00000467, 0.00001005, -0.00000261, -0.00001399, -0.00000173, 0.00001107, -0.00000226, 0.00000378, -0.00000591, 0.00000002, -0.00000298, 0.00000841, },
		{ 0.00000233, -0.00000537, -0.00000547, 0.00000621, 0.00000670, -0.00001177, 0.00000868, -0.00000695, -0.00000045, 0.00000095, 0.00000131, -0.00000335, 0.00000578, 0.00000062, -0.00000284, -0.00000221, -0.00000040, 0.00000104, 0.00000129, -0.00000574, -0.00000417, 0.00000949, -0.00000168, 0.00000458, -0.00000880, 0.00000608, 0.00000055, -0.00000046, -0.00000890, 0.00001111, -0.00000076, -0.00001101, -0.00000248, 0.00001321, -0.00000651, 0.00000879, -0.00000910, 0.00001252, -0.00001422, 0.00001384, -0.00000462, 0.00000074, 0.00001162, -0.00002818, 0.00001556, -0.00000919, 0.00000376, -0.00001291, 0.00000592, -0.00001812, 0.00000674, -0.00001268, -0.00000341, 0.00000332, -0.00002183, 0.00001967, -0.00002398, -0.00000959, -0.00003652, -0.00006914, -0.00022722, 0.00066586, 0.00003629, -0.00021632, 0.00001255, -0.00004756, 0.00000808, -0.00002165, -0.00000341, -0.00000269, -0.00001138, -0.00001439, 0.00000921, -0.00000923, 0.00002954, -0.00001173, 0.00002564, 0.00000003, 0.00000150, 0.00001176, -0.00002113, -0.00000443, -0.00000550, 0.00000965, -0.00000731, -0.00000998, -0.00000174, 0.00000233, 0.00000074, -0.00000111, -0.00000676, 0.00000649, 0.00000772, -0.00002251, -0.00000385, 0.00000972, 0.00000968, -0.00000886, 0.00000518, -0.00000915, 0.00001011, -0.00001205, 0.00000748, 0.00000214, 0.00000508, -0.00000491, 0.00000294, -0.00000516, -0.00000360, 0.00000095, -0.00000086, 0.00000563, 0.00000393, -0.00000860, -0.00000170, 0.00000840, -0.00000516, 0.00000910, -0.00000007, -0.00001382, -0.00000413, 0.00001117, -0.00000119, 0.00000334, -0.00000489, -0.00000093, -0.00000258, 0.00000651, },
		{ 0.00000493, -0.00000573, -0.00000513, 0.00000333, 0.00000987, -0.00001299, 0.00000824, -0.00000500, -0.00000256, 0.00000193, 0.00000092, -0.00000308, 0.00000453, 0.00000296, -0.00000461, -0.00000066, -0.00000240, 0.00000283, -0.00000067, -0.00000223, -0.00000812, 0.00001206, -0.00000389, 0.00000717, -0.00001062, 0.00000585, 0.00000101, 0.00000139, -0.00001097, 0.00001077, 0.00000051, -0.00000971, -0.00000590, 0.00001508, -0.00000734, 0.00000975, -0.00000974, 0.00001265, -0.00001301, 0.00000980, 0.00000033, -0.00000330, 0.00001428, -0.00002649, 0.00000970, -0.00000208, -0.00000243, -0.00000674, -0.00000069, -0.00000987, -0.00000289, -0.00000162, -0.00001623, 0.00001779, -0.00003661, 0.00003380, -0.00003865, 0.00001183, -0.00006387, -0.00001234, -0.00030910, 0.00064041, 0.00014674, -0.00026841, 0.00002935, -0.00006321, 0.00001871, -0.00003040, 0.00000362, -0.00001004, -0.00000267, -0.00002228, 0.00001565, -0.00001648, 0.00003299, -0.00001375, 0.00002424, 0.00000152, 0.00000048, 0.00001221, -0.00001607, -0.00001117, -0.00000005, 0.00000428, -0.00000094, -0.00001510, 0.00000218, -0.00000225, 0.00000537, -0.00000456, -0.00000277, 0.00000037, 0.00001400, -0.00002453, -0.00000318, 0.00000475, 0.00001539, -0.00001153, 0.00000702, -0.00001145, 0.00001159, -0.00001200, 0.00000618, 0.00000275, 0.00000507, -0.00000444, 0.00000231, -0.00000372, -0.00000492, 0.00000104, -0.00000075, 0.00000504, 0.00000464, -0.00000728, -0.00000418, 0.00000913, -0.00000458, 0.00000723, 0.00000280, -0.00001364, -0.00000597, 0.00001033, 0.00000053, 0.00000239, -0.00000341, -0.00000227, -0.00000169, 0.00000418, },
		{ 0.00000713, -0.00000565, -0.00000479, 0.00000057, 0.00001224, -0.00001302, 0.00000659, -0.00000216, -0.00000505, 0.00000302, 0.00000033, -0.00000236, 0.00000264, 0.00000576, -0.00000660, 0.00000107, -0.00000455, 0.00000474, -0.00000279, 0.00000121, -0.00001153, 0.00001385, -0.00000527, 0.00000873, -0.00001111, 0.00000445, 0.00000193, 0.00000298, -0.00001229, 0.00000963, 0.00000194, -0.00000829, -0.00000902, 0.00001628, -0.00000760, 0.00001022, -0.00000985, 0.00001219, -0.00001122, 0.00000577, 0.00000406, -0.00000508, 0.00001391, -0.00002155, 0.00000123, 0.00000661, -0.00000955, 0.00000027, -0.00000818, -0.00000077, -0.00001321, 0.00001003, -0.00002946, 0.00003220, -0.00005078, 0.00004714, -0.00005280, 0.00003251, -0.00008971, 0.00003744, -0.00036479, 0.00057639, 0.00027837, -0.00032221, 0.00004532, -0.00007903, 0.00002944, -0.00003949, 0.00001126, -0.00001822, 0.00000665, -0.00003052, 0.00002239, -0.00002366, 0.00003627, -0.00001584, 0.00002357, 0.00000144, 0.00000167, 0.00000950, -0.00000787, -0.00001993, 0.00000652, -0.00000185, 0.00000589, -0.00002045, 0.00000635, -0.00000703, 0.00000996, -0.00000798, 0.00000141, -0.00000563, 0.00001944, -0.00002568, -0.00000253, -0.00000021, 0.00002007, -0.00001278, 0.00000781, -0.00001277, 0.00001191, -0.00001088, 0.00000421, 0.00000369, 0.00000482, -0.00000376, 0.00000152, -0.00000234, -0.00000584, 0.00000065, -0.00000032, 0.00000426, 0.00000533, -0.00000603, -0.00000598, 0.00000875, -0.00000296, 0.00000461, 0.00000581, -0.00001339, -0.00000725, 0.00000867, 0.00000270, 0.00000117, -0.00000168, -0.00000377, -0.00000055, 0.00000174, },
		{ 0.00000865, -0.00000498, -0.00000459, -0.00000182, 0.00001358, -0.00001179, 0.00000388, 0.00000128, -0.00000757, 0.00000397, -0.00000028, -0.00000135, 0.00000036, 0.00000867, -0.00000839, 0.00000256, -0.00000642, 0.00000630, -0.00000455, 0.00000400, -0.00001383, 0.00001438, -0.00000541, 0.00000891, -0.00001008, 0.00000190, 0.00000327, 0.00000421, -0.00001274, 0.00000770, 0.00000354, -0.00000685, -0.00001164, 0.00001669, -0.00000723, 0.00001015, -0.00000942, 0.00001120, -0.00000903, 0.00000211, 0.00000627, -0.00000462, 0.00001099, -0.00001431, -0.00000849, 0.00001536, -0.00001614, 0.00000668, -0.00001512, 0.00000765, -0.00002257, 0.00002048, -0.00004109, 0.00004430, -0.00006190, 0.00005700, -0.00006325, 0.00004844, -0.00010870, 0.00007316, -0.00038835, 0.00047409, 0.00042510, -0.00037138, 0.00005628, -0.00009207, 0.00003799, -0.00004709, 0.00001795, -0.00002583, 0.00001535, -0.00003803, 0.00002848, -0.00002997, 0.00003879, -0.00001758, 0.00002333, 0.00000009, 0.00000463, 0.00000440, 0.00000220, -0.00002917, 0.00001273, -0.00000740, 0.00001199, -0.00002494, 0.00000976, -0.00001105, 0.00001361, -0.00001057, 0.00000494, -0.00001067, 0.00002329, -0.00002544, -0.00000242, -0.00000445, 0.00002304, -0.00001222, 0.00000732, -0.00001292, 0.00001097, -0.00000872, 0.00000169, 0.00000482, 0.00000441, -0.00000295, 0.00000070, -0.00000113, -0.00000627, -0.00000021, 0.00000038, 0.00000336, 0.00000603, -0.00000496, -0.00000698, 0.00000732, -0.00000051, 0.00000154, 0.00000867, -0.00001294, -0.00000806, 0.00000642, 0.00000499, -0.00000004, 0.00000005, -0.00000517, 0.00000054, -0.00000047, },
		{ 0.00000929, -0.00000366, -0.00000462, -0.00000368, 0.00001382, -0.00000941, 0.00000040, 0.00000491, -0.00000976, 0.00000451, -0.00000073, -0.00000029, -0.00000196, 0.00001125, -0.00000955, 0.00000342, -0.00000761, 0.00000711, -0.00000549, 0.00000568, -0.00001462, 0.00001335, -0.00000409, 0.00000758, -0.00000753, -0.00000164, 0.00000487, 0.00000507, -0.00001228, 0.00000503, 0.00000526, -0.00000550, -0.00001360, 0.00001621, -0.00000618, 0.00000948, -0.00000840, 0.00000968, -0.00000655, -0.00000104, 0.00000702, -0.00000236, 0.00000642, -0.00000618, -0.00001777, 0.00002248, -0.00002064, 0.00001105, -0.00002001, 0.00001382, -0.00002930, 0.00002796, -0.00004915, 0.00005193, -0.00006764, 0.00006085, -0.00006710, 0.00005609, -0.00011640, 0.00008970, -0.00037715, 0.00033779, 0.00057768, -0.00040791, 0.00005726, -0.00009895, 0.00004171, -0.00005107, 0.00002190, -0.00003129, 0.00002195, -0.00004351, 0.00003275, -0.00003441, 0.00003973, -0.00001832, 0.00002290, -0.00000177, 0.00000843, -0.00000177, 0.00001241, -0.00003706, 0.00001694, -0.00001100, 0.00001618, -0.00002752, 0.00001144, -0.00001337, 0.00001548, -0.00001158, 0.00000711, -0.00001400, 0.00002497, -0.00002342, -0.00000329, -0.00000737, 0.00002382, -0.00000965, 0.00000550, -0.00001186, 0.00000882, -0.00000569, -0.00000118, 0.00000603, 0.00000393, -0.00000210, -0.00000009, -0.00000016, -0.00000624, -0.00000143, 0.00000120, 0.00000243, 0.00000668, -0.00000410, -0.00000721, 0.00000502, 0.00000241, -0.00000154, 0.00001103, -0.00001209, -0.00000860, 0.00000392, 0.00000705, -0.00000093, 0.00000148, -0.00000622, 0.00000132, -0.00000216, },
		{ 0.00000899, -0.00000174, -0.00000483, -0.00000498, 0.00001302, -0.00000617, -0.00000341, 0.00000826, -0.00001121, 0.00000437, -0.00000079, 0.00000059, -0.00000398, 0.00001309, -0.00000971, 0.00000331, -0.00000781, 0.00000683, -0.00000530, 0.00000593, -0.00001369, 0.00001068, -0.00000131, 0.00000484, -0.00000373, -0.00000581, 0.00000650, 0.00000559, -0.00001093, 0.00000176, 0.00000702, -0.00000433, -0.00001477, 0.00001477, -0.00000442, 0.00000817, -0.00000674, 0.00000760, -0.00000380, -0.00000372, 0.00000664, 0.00000090, 0.00000147, 0.00000126, -0.00002491, 0.00002640, -0.00002171, 0.00001216, -0.00002165, 0.00001647, -0.00003208, 0.00003102, -0.00005206, 0.00005342, -0.00006629, 0.00005691, -0.00006230, 0.00005305, -0.00011004, 0.00008457, -0.00033239, 0.00017576, 0.00072433, -0.00042290, 0.00004323, -0.00009632, 0.00003806, -0.00004940, 0.00002136, -0.00003298, 0.00002501, -0.00004566, 0.00003403, -0.00003594, 0.00003819, -0.00001722, 0.00002141, -0.00000315, 0.00001181, -0.00000739, 0.00002086, -0.00004178, 0.00001767, -0.00001147, 0.00001747, -0.00002732, 0.00001062, -0.00001330, 0.00001497, -0.00001049, 0.00000737, -0.00001510, 0.00002412, -0.00001942, -0.00000541, -0.00000859, 0.00002220, -0.00000514, 0.00000251, -0.00000975, 0.00000566, -0.00000203, -0.00000419, 0.00000715, 0.00000346, -0.00000124, -0.00000081, 0.00000056, -0.00000582, -0.00000283, 0.00000194, 0.00000162, 0.00000722, -0.00000342, -0.00000681, 0.00000220, 0.00000534, -0.00000418, 0.00001255, -0.00001066, -0.00000907, 0.00000149, 0.00000854, -0.00000124, 0.00000241, -0.00000672, 0.00000158, -0.00000315, },
		{ 0.00000779, 0.00000061, -0.00000514, -0.00000577, 0.00001136, -0.00000243, -0.00000707, 0.00001084, -0.00001157, 0.00000337, -0.00000035, 0.00000106, -0.00000539, 0.00001387, -0.00000861, 0.00000205, -0.00000685, 0.00000534, -0.00000384, 0.00000468, -0.00001111, 0.00000655, 0.00000267, 0.00000101, 0.00000090, -0.00001014, 0.00000786, 0.00000588, -0.00000881, -0.00000193, 0.00000873, -0.00000341, -0.00001506, 0.00001237, -0.00000196, 0.00000620, -0.00000439, 0.00000492, -0.00000073, -0.00000612, 0.00000567, 0.00000417, -0.00000253, 0.00000649, -0.00002846, 0.00002597, -0.00001849, 0.00000932, -0.00001934, 0.00001486, -0.00003015, 0.00002888, -0.00004896, 0.00004789, -0.00005706, 0.00004444, -0.00004808, 0.00003854, -0.00008901, 0.00005838, -0.00025917, -0.00000034, 0.00085169, -0.00040765, 0.00000990, -0.00008159, 0.00002509, -0.00004053, 0.00001499, -0.00002968, 0.00002338, -0.00004345, 0.00003138, -0.00003370, 0.00003341, -0.00001355, 0.00001803, -0.00000296, 0.00001345, -0.00001086, 0.00002579, -0.00004184, 0.00001388, -0.00000810, 0.00001532, -0.00002391, 0.00000691, -0.00001051, 0.00001184, -0.00000712, 0.00000553, -0.00001380, 0.00002073, -0.00001353, -0.00000879, -0.00000800, 0.00001826, 0.00000095, -0.00000127, -0.00000688, 0.00000178, 0.00000192, -0.00000706, 0.00000807, 0.00000305, -0.00000039, -0.00000148, 0.00000110, -0.00000518, -0.00000418, 0.00000240, 0.00000110, 0.00000752, -0.00000280, -0.00000603, -0.00000071, 0.00000779, -0.00000595, 0.00001296, -0.00000852, -0.00000965, -0.00000056, 0.00000918, -0.00000081, 0.00000271, -0.00000655, 0.00000124, -0.00000337, },
		{ 0.00000589, 0.00000313, -0.00000535, -0.00000620, 0.00000913, 0.00000134, -0.00001010, 0.00001227, -0.00001065, 0.00000145, 0.00000066, 0.00000103, -0.00000600, 0.00001342, -0.00000618, -0.00000035, -0.00000478, 0.00000272, -0.00000123, 0.00000211, -0.00000717, 0.00000139, 0.00000738, -0.00000338, 0.00000577, -0.00001409, 0.00000863, 0.00000606, -0.00000608, -0.00000581, 0.00001026, -0.00000275, -0.00001445, 0.00000909, 0.00000112, 0.00000359, -0.00000135, 0.00000162, 0.00000272, -0.00000846, 0.00000472, 0.00000648, -0.00000442, 0.00000835, -0.00002753, 0.00002068, -0.00001083, 0.00000249, -0.00001307, 0.00000899, -0.00002351, 0.00002153, -0.00003990, 0.00003551, -0.00004032, 0.00002402, -0.00002521, 0.00001362, -0.00005511, 0.00001477, -0.00016593, -0.00017645, 0.00094620, -0.00035481, -0.00004533, -0.00005355, 0.00000199, -0.00002385, 0.00000223, -0.00002080, 0.00001650, -0.00003636, 0.00002429, -0.00002723, 0.00002495, -0.00000681, 0.00001211, -0.00000034, 0.00001227, -0.00001089, 0.00002598, -0.00003643, 0.00000523, -0.00000084, 0.00000978, -0.00001739, 0.00000045, -0.00000514, 0.00000628, -0.00000169, 0.00000177, -0.00001029, 0.00001511, -0.00000615, -0.00001318, -0.00000580, 0.00001242, 0.00000801, -0.00000531, -0.00000366, -0.00000241, 0.00000581, -0.00000957, 0.00000866, 0.00000271, 0.00000047, -0.00000215, 0.00000157, -0.00000449, -0.00000523, 0.00000238, 0.00000101, 0.00000747, -0.00000208, -0.00000515, -0.00000328, 0.00000933, -0.00000652, 0.00001209, -0.00000562, -0.00001046, -0.00000203, 0.00000884, 0.00000039, 0.00000237, -0.00000575, 0.00000033, -0.00000290, },
		{ 0.00000359, 0.00000552, -0.00000526, -0.00000644, 0.00000665, 0.00000472, -0.00001209, 0.00001230, -0.00000840, -0.00000130, 0.00000213, 0.00000049, -0.00000577, 0.00001175, -0.00000257, -0.00000366, -0.00000185, -0.00000076, 0.00000221, -0.00000137, -0.00000241, -0.00000420, 0.00001220, -0.00000770, 0.00001024, -0.00001713, 0.00000854, 0.00000623, -0.00000293, -0.00000961, 0.00001148, -0.00000234, -0.00001303, 0.00000507, 0.00000468, 0.00000043, 0.00000230, -0.00000224, 0.00000654, -0.00001095, 0.00000431, 0.00000708, -0.00000343, 0.00000627, -0.00002195, 0.00001085, 0.00000063, -0.00000758, -0.00000359, -0.00000038, -0.00001299, 0.00000990, -0.00002589, 0.00001754, -0.00001766, -0.00000240, 0.00000402, -0.00001881, -0.00001227, -0.00004000, -0.00006359, -0.00033763, 0.00099570, -0.00025975, -0.00012252, -0.00001296, -0.00003053, 0.00000001, -0.00001647, -0.00000665, 0.00000459, -0.00002459, 0.00001291, -0.00001664, 0.00001288, 0.00000302, 0.00000345, 0.00000514, 0.00000766, -0.00000680, 0.00002094, -0.00002556, -0.00000779, 0.00000963, 0.00000155, -0.00000842, -0.00000813, 0.00000217, -0.00000104, 0.00000518, -0.00000338, -0.00000513, 0.00000793, 0.00000206, -0.00001808, -0.00000245, 0.00000537, 0.00001527, -0.00000899, -0.00000055, -0.00000651, 0.00000928, -0.00001150, 0.00000888, 0.00000244, 0.00000138, -0.00000287, 0.00000206, -0.00000395, -0.00000580, 0.00000172, 0.00000143, 0.00000700, -0.00000116, -0.00000444, -0.00000515, 0.00000964, -0.00000573, 0.00000997, -0.00000208, -0.00001149, -0.00000283, 0.00000753, 0.00000227, 0.00000153, -0.00000443, -0.00000102, -0.00000192, },
		{ 0.00000121, 0.00000745, -0.00000470, -0.00000665, 0.00000422, 0.00000735, -0.00001281, 0.00001089, -0.00000499, -0.00000461, 0.00000387, -0.00000044, -0.00000481, 0.00000905, 0.00000189, -0.00000747, 0.00000150, -0.00000462, 0.00000595, -0.00000517, 0.00000249, -0.00000950, 0.00001644, -0.00001128, 0.00001369, -0.00001881, 0.00000739, 0.00000649, 0.00000042, -0.00001305, 0.00001223, -0.00000211, -0.00001093, 0.00000057, 0.00000848, -0.00000306, 0.00000636, -0.00000646, 0.00001061, -0.00001364, 0.00000474, 0.00000558, 0.00000066, 0.00000042, -0.00001238, -0.00000239, 0.00001450, -0.00001942, 0.00000768, -0.00001178, -0.00000016, -0.00000434, -0.00000881, -0.00000382, 0.00000830, -0.00003177, 0.00003606, -0.00005446, 0.00003388, -0.00009795, 0.00003582, -0.00046966, 0.00099096, -0.00012160, -0.00021854, 0.00003711, -0.00006997, 0.00002901, -0.00003946, 0.00001142, -0.00001119, -0.00000915, -0.00000188, -0.00000273, -0.00000208, 0.00001539, -0.00000760, 0.00001333, -0.00000029, 0.00000127, 0.00001111, -0.00001021, -0.00002384, 0.00002189, -0.00000807, 0.00000182, -0.00001772, 0.00001037, -0.00000912, 0.00001253, -0.00000905, 0.00000086, 0.00000010, 0.00001023, -0.00002280, 0.00000138, -0.00000205, 0.00002183, -0.00001168, 0.00000201, -0.00001012, 0.00001204, -0.00001270, 0.00000867, 0.00000223, 0.00000235, -0.00000369, 0.00000266, -0.00000368, -0.00000576, 0.00000040, 0.00000235, 0.00000611, 0.00000005, -0.00000409, -0.00000606, 0.00000861, -0.00000363, 0.00000681, 0.00000188, -0.00001263, -0.00000302, 0.00000541, 0.00000456, 0.00000041, -0.00000283, -0.00000257, -0.00000069, },
		{ -0.00000090, 0.00000866, -0.00000354, -0.00000696, 0.00000208, 0.00000900, -0.00001218, 0.00000818, -0.00000075, -0.00000809, 0.00000554, -0.00000151, -0.00000336, 0.00000570, 0.00000670, -0.00001125, 0.00000472, -0.00000831, 0.00000942, -0.00000864, 0.00000682, -0.00001382, 0.00001945, -0.00001355, 0.00001562, -0.00001881, 0.00000513, 0.00000687, 0.00000376, -0.00001587, 0.00001235, -0.00000194, -0.00000839, -0.00000408, 0.00001220, -0.00000660, 0.00001052, -0.00001073, 0.00001467, -0.00001640, 0.00000603, 0.00000204, 0.00000745, -0.00000832, -0.00000023, -0.00001721, 0.00002877, -0.00003108, 0.00001882, -0.00002326, 0.00001293, -0.00001895, 0.00000886, -0.00002572, 0.00003424, -0.00006032, 0.00006662, -0.00008819, 0.00007694, -0.00015049, 0.00012062, -0.00056072, 0.00092710, 0.00005595, -0.00032686, 0.00009124, -0.00011215, 0.00005975, -0.00006397, 0.00003112, -0.00002886, 0.00000817, -0.00001848, 0.00001305, -0.00001862, 0.00002913, -0.00002008, 0.00002339, -0.00001079, 0.00001238, -0.00000219, 0.00000782, -0.00004087, 0.00003402, -0.00001741, 0.00001181, -0.00002692, 0.00001814, -0.00001670, 0.00001922, -0.00001423, 0.00000665, -0.00000733, 0.00001742, -0.00002659, 0.00000495, -0.00000893, 0.00002689, -0.00001283, 0.00000366, -0.00001295, 0.00001385, -0.00001306, 0.00000806, 0.00000206, 0.00000338, -0.00000459, 0.00000338, -0.00000373, -0.00000510, -0.00000152, 0.00000367, 0.00000488, 0.00000153, -0.00000419, -0.00000596, 0.00000631, -0.00000049, 0.00000298, 0.00000588, -0.00001366, -0.00000278, 0.00000278, 0.00000692, -0.00000066, -0.00000122, -0.00000405, 0.00000048, },
		{ -0.00000247, 0.00000897, -0.00000175, -0.00000740, 0.00000034, 0.00000961, -0.00001033, 0.00000454, 0.00000381, -0.00001125, 0.00000679, -0.00000243, -0.00000174, 0.00000213, 0.00001130, -0.00001439, 0.00000725, -0.00001124, 0.00001200, -0.00001116, 0.00000993, -0.00001655, 0.00002074, -0.00001410, 0.00001573, -0.00001702, 0.00000181, 0.00000735, 0.00000688, -0.00001780, 0.00001173, -0.00000170, -0.00000571, -0.00000848, 0.00001546, -0.00000982, 0.00001438, -0.00001461, 0.00001829, -0.00001891, 0.00000788, -0.00000301, 0.00001594, -0.00001846, 0.00001250, -0.00003133, 0.00004114, -0.00004038, 0.00002772, -0.00003268, 0.00002398, -0.00003150, 0.00002441, -0.00004505, 0.00005665, -0.00008410, 0.00009125, -0.00011483, 0.00011061, -0.00018964, 0.00018100, -0.00060291, 0.00080455, 0.00026433, -0.00043765, 0.00014205, -0.00015152, 0.00008782, -0.00008643, 0.00004943, -0.00004574, 0.00002498, -0.00003474, 0.00002870, -0.00003489, 0.00004258, -0.00003249, 0.00003394, -0.00002236, 0.00002488, -0.00001689, 0.00002607, -0.00005642, 0.00004383, -0.00002459, 0.00001992, -0.00003425, 0.00002413, -0.00002251, 0.00002413, -0.00001790, 0.00001126, -0.00001336, 0.00002275, -0.00002876, 0.00000753, -0.00001443, 0.00002974, -0.00001209, 0.00000419, -0.00001477, 0.00001457, -0.00001257, 0.00000705, 0.00000194, 0.00000444, -0.00000552, 0.00000417, -0.00000403, -0.00000394, -0.00000383, 0.00000517, 0.00000349, 0.00000316, -0.00000470, -0.00000494, 0.00000304, 0.00000324, -0.00000101, 0.00000949, -0.00001429, -0.00000236, 0.00000002, 0.00000895, -0.00000138, 0.00000011, -0.00000520, 0.00000133, },
		{ -0.00000333, 0.00000835, 0.00000056, -0.00000791, -0.00000098, 0.00000926, -0.00000759, 0.00000045, 0.00000811, -0.00001359, 0.00000726, -0.00000291, -0.00000032, -0.00000116, 0.00001511, -0.00001636, 0.00000859, -0.00001292, 0.00001323, -0.00001224, 0.00001136, -0.00001732, 0.00002004, -0.00001273, 0.00001394, -0.00001354, -0.00000233, 0.00000787, 0.00000961, -0.00001865, 0.00001024, -0.00000124, -0.00000322, -0.00001222, 0.00001785, -0.00001229, 0.00001745, -0.00001759, 0.00002097, -0.00002068, 0.00000972, -0.00000866, 0.00002475, -0.00002809, 0.00002358, -0.00004239, 0.00004937, -0.00004529, 0.00003245, -0.00003806, 0.00003094, -0.00003973, 0.00003536, -0.00005899, 0.00007234, -0.00009963, 0.00010612, -0.00012998, 0.00012969, -0.00020921, 0.00021026, -0.00059318, 0.00062948, 0.00049046, -0.00053859, 0.00018087, -0.00018179, 0.00010832, -0.00010287, 0.00006299, -0.00005890, 0.00003862, -0.00004822, 0.00004198, -0.00004879, 0.00005380, -0.00004301, 0.00004320, -0.00003311, 0.00003661, -0.00003052, 0.00004186, -0.00006793, 0.00004922, -0.00002793, 0.00002470, -0.00003840, 0.00002711, -0.00002547, 0.00002629, -0.00001922, 0.00001384, -0.00001715, 0.00002553, -0.00002878, 0.00000854, -0.00001790, 0.00002994, -0.00000933, 0.00000356, -0.00001552, 0.00001417, -0.00001125, 0.00000569, 0.00000189, 0.00000541, -0.00000635, 0.00000489, -0.00000446, -0.00000249, -0.00000624, 0.00000655, 0.00000218, 0.00000478, -0.00000546, -0.00000327, -0.00000075, 0.00000700, -0.00000458, 0.00001225, -0.00001423, -0.00000207, -0.00000248, 0.00001028, -0.00000146, 0.00000093, -0.00000580, 0.00000166, },
		{ -0.00000344, 0.00000689, 0.00000320, -0.00000833, -0.00000197, 0.00000817, -0.00000437, -0.00000351, 0.00001158, -0.00001468, 0.00000667, -0.00000273, 0.00000061, -0.00000374, 0.00001765, -0.00001673, 0.00000838, -0.00001304, 0.00001280, -0.00001159, 0.00001087, -0.00001599, 0.00001734, -0.00000954, 0.00001044, -0.00000870, -0.00000695, 0.00000829, 0.00001182, -0.00001827, 0.00000783, -0.00000040, -0.00000123, -0.00001490, 0.00001901, -0.00001360, 0.00001925, -0.00001914, 0.00002216, -0.00002115, 0.00001084, -0.00001379, 0.00003226, -0.00003524, 0.00003086, -0.00004832, 0.00005165, -0.00004428, 0.00003163, -0.00003798, 0.00003227, -0.00004199, 0.00003985, -0.00006543, 0.00007899, -0.00010442, 0.00010856, -0.00013070, 0.00013093, -0.00020568, 0.00020565, -0.00053392, 0.00041347, 0.00071766, -0.00061593, 0.00019886, -0.00019687, 0.00011658, -0.00010956, 0.00006866, -0.00006553, 0.00004655, -0.00005663, 0.00005070, -0.00005825, 0.00006084, -0.00004981, 0.00004929, -0.00004101, 0.00004532, -0.00004058, 0.00005265, -0.00007325, 0.00004855, -0.00002622, 0.00002512, -0.00003845, 0.00002629, -0.00002485, 0.00002510, -0.00001767, 0.00001385, -0.00001818, 0.00002540, -0.00002639, 0.00000766, -0.00001898, 0.00002738, -0.00000468, 0.00000191, -0.00001525, 0.00001271, -0.00000919, 0.00000402, 0.00000198, 0.00000617, -0.00000692, 0.00000538, -0.00000478, -0.00000106, -0.00000838, 0.00000748, 0.00000124, 0.00000614, -0.00000623, -0.00000131, -0.00000451, 0.00001017, -0.00000719, 0.00001377, -0.00001326, -0.00000215, -0.00000438, 0.00001063, -0.00000074, 0.00000112, -0.00000575, 0.00000139, },
		{ -0.00000289, 0.00000481, 0.00000587, -0.00000848, -0.00000277, 0.00000665, -0.00000116, -0.00000681, 0.00001375, -0.00001422, 0.00000489, -0.00000175, 0.00000083, -0.00000533, 0.00001863, -0.00001530, 0.00000653, -0.00001153, 0.00001066, -0.00000922, 0.00000852, -0.00001276, 0.00001292, -0.00000486, 0.00000563, -0.00000297, -0.00001164, 0.00000848, 0.00001340, -0.00001660, 0.00000453, 0.00000089, -0.00000001, -0.00001621, 0.00001863, -0.00001344, 0.00001939, -0.00001882, 0.00002140, -0.00001980, 0.00001052, -0.00001730, 0.00003698, -0.00003822, 0.00003265, -0.00004770, 0.00004698, -0.00003666, 0.00002469, -0.00003185, 0.00002731, -0.00003753, 0.00003699, -0.00006332, 0.00007547, -0.00009740, 0.00009759, -0.00011603, 0.00011349, -0.00017871, 0.00016876, -0.00043268, 0.00017266, 0.00092709, -0.00065608, 0.00018824, -0.00019178, 0.00010899, -0.00010368, 0.00006405, -0.00006350, 0.00004682, -0.00005814, 0.00005312, -0.00006161, 0.00006212, -0.00005132, 0.00005058, -0.00004427, 0.00004905, -0.00004499, 0.00005649, -0.00007093, 0.00004097, -0.00001895, 0.00002083, -0.00003408, 0.00002140, -0.00002045, 0.00002045, -0.00001316, 0.00001118, -0.00001632, 0.00002234, -0.00002167, 0.00000486, -0.00001761, 0.00002225, 0.00000146, -0.00000045, -0.00001412, 0.00001032, -0.00000653, 0.00000207, 0.00000228, 0.00000656, -0.00000702, 0.00000540, -0.00000476, 0.00000004, -0.00000990, 0.00000763, 0.00000091, 0.00000702, -0.00000674, 0.00000051, -0.00000768, 0.00001222, -0.00000840, 0.00001381, -0.00001126, -0.00000276, -0.00000545, 0.00000985, 0.00000081, 0.00000069, -0.00000508, 0.00000056, },
		{ -0.00000187, 0.00000242, 0.00000825, -0.00000813, -0.00000358, 0.00000502, 0.00000159, -0.00000900, 0.00001432, -0.00001212, 0.00000198, -0.00000003, 0.00000030, -0.00000578, 0.00001796, -0.00001213, 0.00000316, -0.00000856, 0.00000703, -0.00000539, 0.00000465, -0.00000804, 0.00000730, 0.00000076, 0.00000009, 0.00000304, -0.00001591, 0.00000828, 0.00001431, -0.00001366, 0.00000043, 0.00000268, 0.00000024, -0.00001596, 0.00001659, -0.00001162, 0.00001762, -0.00001636, 0.00001841, -0.00001631, 0.00000822, -0.00001830, 0.00003774, -0.00003587, 0.00002807, -0.00004007, 0.00003533, -0.00002274, 0.00001202, -0.00002005, 0.00001640, -0.00002667, 0.00002704, -0.00005291, 0.00006216, -0.00007917, 0.00007409, -0.00008720, 0.00007920, -0.00013126, 0.00010522, -0.00030138, -0.00007384, 0.00109969, -0.00064732, 0.00014361, -0.00016367, 0.00008367, -0.00008386, 0.00004803, -0.00005179, 0.00003845, -0.00005181, 0.00004829, -0.00005790, 0.00005671, -0.00004658, 0.00004603, -0.00004167, 0.00004647, -0.00004241, 0.00005236, -0.00006054, 0.00002660, -0.00000651, 0.00001220, -0.00002566, 0.00001279, -0.00001265, 0.00001274, -0.00000607, 0.00000611, -0.00001189, 0.00001677, -0.00001502, 0.00000042, -0.00001408, 0.00001507, 0.00000847, -0.00000310, -0.00001238, 0.00000721, -0.00000340, -0.00000012, 0.00000288, 0.00000643, -0.00000649, 0.00000478, -0.00000417, 0.00000052, -0.00001050, 0.00000678, 0.00000137, 0.00000725, -0.00000675, 0.00000181, -0.00000981, 0.00001275, -0.00000799, 0.00001230, -0.00000827, -0.00000395, -0.00000559, 0.00000796, 0.00000308, -0.00000024, -0.00000392, -0.00000069, },
		{ -0.00000067, 0.00000005, 0.00001001, -0.00000711, -0.00000455, 0.00000356, 0.00000353, -0.00000981, 0.00001320, -0.00000853, -0.00000182, 0.00000224, -0.00000088, -0.00000518, 0.00001580, -0.00000750, -0.00000134, -0.00000453, 0.00000237, -0.00000058, -0.00000018, -0.00000250, 0.00000115, 0.00000665, -0.00000552, 0.00000871, -0.00001934, 0.00000758, 0.00001453, -0.00000959, -0.00000425, 0.00000488, -0.00000050, -0.00001415, 0.00001292, -0.00000816, 0.00001393, -0.00001176, 0.00001318, -0.00001061, 0.00000371, -0.00001629, 0.00003400, -0.00002784, 0.00001717, -0.00002603, 0.00001780, -0.00000388, -0.00000498, -0.00000394, 0.00000089, -0.00001076, 0.00001141, -0.00003570, 0.00004082, -0.00005192, 0.00004075, -0.00004750, 0.00003227, -0.00006913, 0.00002396, -0.00015493, -0.00030589, 0.00121824, -0.00058147, 0.00006313, -0.00011256, 0.00004109, -0.00005071, 0.00002111, -0.00003076, 0.00002171, -0.00003782, 0.00003629, -0.00004713, 0.00004455, -0.00003550, 0.00003541, -0.00003284, 0.00003717, -0.00003259, 0.00004041, -0.00004282, 0.00000663, 0.00000982, 0.00000039, -0.00001420, 0.00000141, -0.00000237, 0.00000287, 0.00000276, -0.00000062, -0.00000557, 0.00000941, -0.00000716, -0.00000511, -0.00000895, 0.00000660, 0.00001559, -0.00000554, -0.00001031, 0.00000359, 0.00000002, -0.00000247, 0.00000383, 0.00000568, -0.00000522, 0.00000340, -0.00000286, 0.00000019, -0.00000999, 0.00000483, 0.00000267, 0.00000674, -0.00000611, 0.00000233, -0.00001057, 0.00001157, -0.00000594, 0.00000942, -0.00000450, -0.00000561, -0.00000488, 0.00000515, 0.00000579, -0.00000142, -0.00000249, -0.00000213, },
		{ 0.00000045, -0.00000196, 0.00001092, -0.00000534, -0.00000575, 0.00000248, 0.00000445, -0.00000919, 0.00001056, -0.00000383, -0.00000606, 0.00000471, -0.00000247, -0.00000376, 0.00001248, -0.00000192, -0.00000639, -0.00000005, -0.00000272, 0.00000454, -0.00000526, 0.00000311, -0.00000476, 0.00001208, -0.00001051, 0.00001344, -0.00002154, 0.00000630, 0.00001409, -0.00000461, -0.00000923, 0.00000732, -0.00000213, -0.00001097, 0.00000790, -0.00000332, 0.00000856, -0.00000529, 0.00000603, -0.00000300, -0.00000286, -0.00001131, 0.00002589, -0.00001467, 0.00000107, -0.00000723, -0.00000355, 0.00001769, -0.00002408, 0.00001431, -0.00001703, 0.00000792, -0.00000753, -0.00001427, 0.00001441, -0.00001914, 0.00000169, -0.00000181, -0.00002127, -0.00000015, -0.00006404, -0.00000947, -0.00050446, 0.00126935, -0.00045534, -0.00005070, -0.00004183, -0.00001567, -0.00000689, -0.00001444, -0.00000233, -0.00000175, -0.00001759, 0.00001829, -0.00003035, 0.00002662, -0.00001892, 0.00001942, -0.00001841, 0.00002182, -0.00001640, 0.00002197, -0.00001963, -0.00001684, 0.00002804, -0.00001287, -0.00000123, -0.00001129, 0.00000904, -0.00000788, 0.00001220, -0.00000802, 0.00000168, 0.00000127, 0.00000102, -0.00001100, -0.00000298, -0.00000227, 0.00002204, -0.00000731, -0.00000819, -0.00000029, 0.00000353, -0.00000493, 0.00000514, 0.00000427, -0.00000319, 0.00000124, -0.00000081, -0.00000100, -0.00000836, 0.00000185, 0.00000472, 0.00000557, -0.00000479, 0.00000193, -0.00000988, 0.00000876, -0.00000252, 0.00000553, -0.00000032, -0.00000748, -0.00000353, 0.00000174, 0.00000858, -0.00000254, -0.00000108, -0.00000348, },
		{ 0.00000121, -0.00000336, 0.00001082, -0.00000282, -0.00000716, 0.00000184, 0.00000433, -0.00000730, 0.00000679, 0.00000144, -0.00001020, 0.00000695, -0.00000411, -0.00000190, 0.00000853, 0.00000396, -0.00001132, 0.00000422, -0.00000756, 0.00000928, -0.00000987, 0.00000805, -0.00000973, 0.00001643, -0.00001432, 0.00001677, -0.00002229, 0.00000441, 0.00001305, 0.00000097, -0.00001410, 0.00000973, -0.00000440, -0.00000679, 0.00000198, 0.00000243, 0.00000203, 0.00000246, -0.00000238, 0.00000587, -0.00001087, -0.00000396, 0.00001429, 0.00000225, -0.00001827, 0.00001382, -0.00002593, 0.00003912, -0.00004252, 0.00003202, -0.00003465, 0.00002657, -0.00002682, 0.00000813, -0.00001333, 0.00001485, -0.00003811, 0.00004405, -0.00007442, 0.00006692, -0.00014718, 0.00011955, -0.00065370, 0.00124521, -0.00027167, -0.00019068, 0.00004182, -0.00008106, 0.00004293, -0.00005474, 0.00003021, -0.00002908, 0.00000638, -0.00000346, -0.00000957, 0.00000479, 0.00000141, -0.00000030, 0.00000006, 0.00000209, 0.00000423, -0.00000060, 0.00000627, -0.00004098, 0.00004564, -0.00002546, 0.00001139, -0.00002364, 0.00001999, -0.00001806, 0.00002093, -0.00001494, 0.00000875, -0.00000655, 0.00000854, -0.00001641, 0.00000298, -0.00001059, 0.00002705, -0.00000800, -0.00000625, -0.00000420, 0.00000695, -0.00000737, 0.00000675, 0.00000230, -0.00000052, -0.00000155, 0.00000185, -0.00000293, -0.00000579, -0.00000191, 0.00000725, 0.00000392, -0.00000291, 0.00000068, -0.00000787, 0.00000465, 0.00000181, 0.00000117, 0.00000380, -0.00000925, -0.00000184, -0.00000186, 0.00001103, -0.00000326, 0.00000005, -0.00000449, },
		{ 0.00000144, -0.00000400, 0.00000972, 0.00000028, -0.00000866, 0.00000158, 0.00000331, -0.00000450, 0.00000245, 0.00000662, -0.00001365, 0.00000851, -0.00000542, -0.00000004, 0.00000452, 0.00000947, -0.00001543, 0.00000762, -0.00001148, 0.00001296, -0.00001332, 0.00001165, -0.00001316, 0.00001918, -0.00001653, 0.00001839, -0.00002148, 0.00000197, 0.00001154, 0.00000675, -0.00001844, 0.00001178, -0.00000694, -0.00000215, -0.00000418, 0.00000842, -0.00000496, 0.00001065, -0.00001116, 0.00001504, -0.00001938, 0.00000470, 0.00000065, 0.00002085, -0.00003821, 0.00003414, -0.00004620, 0.00005736, -0.00005746, 0.00004642, -0.00004916, 0.00004226, -0.00004334, 0.00002806, -0.00003851, 0.00004556, -0.00007352, 0.00008419, -0.00012022, 0.00012363, -0.00021483, 0.00021914, -0.00074270, 0.00114466, -0.00003945, -0.00034524, 0.00012884, -0.00014753, 0.00009262, -0.00009469, 0.00006252, -0.00005651, 0.00003074, -0.00002600, 0.00001244, -0.00001833, 0.00002300, -0.00002140, 0.00002023, -0.00001955, 0.00002653, -0.00002420, 0.00003160, -0.00006270, 0.00005999, -0.00003523, 0.00002179, -0.00003391, 0.00002892, -0.00002622, 0.00002770, -0.00002025, 0.00001457, -0.00001302, 0.00001450, -0.00002062, 0.00000813, -0.00001751, 0.00003003, -0.00000737, -0.00000464, -0.00000791, 0.00001007, -0.00000965, 0.00000851, -0.00000006, 0.00000256, -0.00000471, 0.00000486, -0.00000533, -0.00000262, -0.00000605, 0.00000988, 0.00000210, -0.00000069, -0.00000122, -0.00000487, -0.00000026, 0.00000640, -0.00000302, 0.00000734, -0.00001056, -0.00000017, -0.00000519, 0.00001274, -0.00000333, 0.00000068, -0.00000498, },
		{ 0.00000109, -0.00000390, 0.00000777, 0.00000371, -0.00001004, 0.00000156, 0.00000169, -0.00000129, -0.00000181, 0.00001102, -0.00001585, 0.00000901, -0.00000606, 0.00000142, 0.00000098, 0.00001397, -0.00001813, 0.00000963, -0.00001398, 0.00001510, -0.00001515, 0.00001348, -0.00001470, 0.00002008, -0.00001693, 0.00001820, -0.00001918, -0.00000091, 0.00000967, 0.00001230, -0.00002178, 0.00001307, -0.00000928, 0.00000232, -0.00000988, 0.00001391, -0.00001154, 0.00001832, -0.00001922, 0.00002338, -0.00002718, 0.00001326, -0.00001317, 0.00003871, -0.00005589, 0.00005066, -0.00006138, 0.00006968, -0.00006639, 0.00005510, -0.00005812, 0.00005242, -0.00005433, 0.00004246, -0.00005758, 0.00006897, -0.00010001, 0.00011350, -0.00015274, 0.00016299, -0.00025869, 0.00028028, -0.00076661, 0.00097364, 0.00022656, -0.00049926, 0.00020779, -0.00020642, 0.00013528, -0.00012865, 0.00008983, -0.00007987, 0.00005180, -0.00004599, 0.00003258, -0.00003976, 0.00004302, -0.00004112, 0.00003934, -0.00004019, 0.00004738, -0.00004549, 0.00005301, -0.00007904, 0.00006874, -0.00004036, 0.00002841, -0.00004066, 0.00003449, -0.00003118, 0.00003148, -0.00002304, 0.00001823, -0.00001728, 0.00001820, -0.00002300, 0.00001183, -0.00002238, 0.00003061, -0.00000532, -0.00000345, -0.00001126, 0.00001268, -0.00001158, 0.00001020, -0.00000250, 0.00000569, -0.00000787, 0.00000783, -0.00000779, 0.00000070, -0.00001005, 0.00001215, 0.00000049, 0.00000153, -0.00000342, -0.00000136, -0.00000530, 0.00001056, -0.00000642, 0.00000984, -0.00001110, 0.00000114, -0.00000787, 0.00001341, -0.00000255, 0.00000070, -0.00000485, },
		{ 0.00000022, -0.00000316, 0.00000524, 0.00000710, -0.00001101, 0.00000154, -0.00000015, 0.00000178, -0.00000537, 0.00001409, -0.00001642, 0.00000819, -0.00000581, 0.00000215, -0.00000165, 0.00001699, -0.00001902, 0.00000991, -0.00001475, 0.00001540, -0.00001510, 0.00001335, -0.00001425, 0.00001910, -0.00001555, 0.00001631, -0.00001559, -0.00000407, 0.00000761, 0.00001719, -0.00002368, 0.00001327, -0.00001097, 0.00000598, -0.00001438, 0.00001813, -0.00001686, 0.00002448, -0.00002548, 0.00002971, -0.00003303, 0.00002024, -0.00002519, 0.00005339, -0.00006862, 0.00006071, -0.00006906, 0.00007402, -0.00006754, 0.00005641, -0.00005981, 0.00005525, -0.00005781, 0.00004909, -0.00006797, 0.00008213, -0.00011432, 0.00012835, -0.00016790, 0.00018043, -0.00027378, 0.00029883, -0.00072715, 0.00074479, 0.00050694, -0.00063564, 0.00026671, -0.00024911, 0.00016422, -0.00015124, 0.00010759, -0.00009522, 0.00006605, -0.00006024, 0.00004783, -0.00005660, 0.00005867, -0.00005674, 0.00005461, -0.00005689, 0.00006370, -0.00006132, 0.00006759, -0.00008765, 0.00007022, -0.00003961, 0.00003020, -0.00004292, 0.00003583, -0.00003217, 0.00003163, -0.00002277, 0.00001919, -0.00001880, 0.00001924, -0.00002325, 0.00001364, -0.00002481, 0.00002868, -0.00000197, -0.00000267, -0.00001408, 0.00001458, -0.00001293, 0.00001156, -0.00000468, 0.00000846, -0.00001059, 0.00001031, -0.00000984, 0.00000364, -0.00001336, 0.00001357, -0.00000051, 0.00000341, -0.00000552, 0.00000212, -0.00000979, 0.00001365, -0.00000850, 0.00001096, -0.00001067, 0.00000186, -0.00000961, 0.00001285, -0.00000091, 0.00000011, -0.00000414, },
		{ -0.00000102, -0.00000201, 0.00000246, 0.00001007, -0.00001130, 0.00000127, -0.00000185, 0.00000420, -0.00000772, 0.00001541, -0.00001515, 0.00000603, -0.00000460, 0.00000199, -0.00000309, 0.00001828, -0.00001795, 0.00000840, -0.00001376, 0.00001387, -0.00001321, 0.00001135, -0.00001199, 0.00001646, -0.00001264, 0.00001300, -0.00001104, -0.00000732, 0.00000551, 0.00002100, -0.00002381, 0.00001211, -0.00001161, 0.00000826, -0.00001707, 0.00002044, -0.00002018, 0.00002826, -0.00002899, 0.00003307, -0.00003586, 0.00002431, -0.00003370, 0.00006283, -0.00007431, 0.00006248, -0.00006786, 0.00006941, -0.00006023, 0.00004974, -0.00005360, 0.00005000, -0.00005296, 0.00004691, -0.00006839, 0.00008359, -0.00011489, 0.00012713, -0.00016402, 0.00017436, -0.00025903, 0.00027582, -0.00063227, 0.00047625, 0.00077947, -0.00073706, 0.00029471, -0.00026825, 0.00017398, -0.00015815, 0.00011226, -0.00009946, 0.00007075, -0.00006624, 0.00005579, -0.00006651, 0.00006764, -0.00006596, 0.00006371, -0.00006721, 0.00007298, -0.00006929, 0.00007331, -0.00008716, 0.00006368, -0.00003260, 0.00002684, -0.00004038, 0.00003265, -0.00002899, 0.00002803, -0.00001932, 0.00001729, -0.00001743, 0.00001756, -0.00002132, 0.00001345, -0.00002469, 0.00002441, 0.00000240, -0.00000221, -0.00001625, 0.00001559, -0.00001351, 0.00001229, -0.00000621, 0.00001044, -0.00001241, 0.00001186, -0.00001105, 0.00000571, -0.00001551, 0.00001376, -0.00000059, 0.00000463, -0.00000714, 0.00000502, -0.00001315, 0.00001513, -0.00000891, 0.00001053, -0.00000920, 0.00000188, -0.00001026, 0.00001107, 0.00000150, -0.00000095, -0.00000300, },
		{ -0.00000238, -0.00000075, -0.00000018, 0.00001228, -0.00001071, 0.00000057, -0.00000308, 0.00000558, -0.00000852, 0.00001483, -0.00001212, 0.00000267, -0.00000253, 0.00000093, -0.00000329, 0.00001782, -0.00001505, 0.00000531, -0.00001124, 0.00001075, -0.00000977, 0.00000783, -0.00000830, 0.00001258, -0.00000860, 0.00000868, -0.00000592, -0.00001044, 0.00000354, 0.00002341, -0.00002197, 0.00000944, -0.00001093, 0.00000876, -0.00001753, 0.00002042, -0.00002099, 0.00002907, -0.00002915, 0.00003280, -0.00003492, 0.00002454, -0.00003744, 0.00006566, -0.00007179, 0.00005526, -0.00005760, 0.00005613, -0.00004499, 0.00003567, -0.00004001, 0.00003716, -0.00004016, 0.00003623, -0.00005905, 0.00007357, -0.00010210, 0.00011041, -0.00014202, 0.00014634, -0.00021722, 0.00021714, -0.00049512, 0.00018983, 0.00102130, -0.00078810, 0.00028350, -0.00025893, 0.00016123, -0.00014694, 0.00010184, -0.00009092, 0.00006441, -0.00006258, 0.00005503, -0.00006802, 0.00006849, -0.00006732, 0.00006511, -0.00006954, 0.00007364, -0.00006806, 0.00006934, -0.00007741, 0.00004949, -0.00001990, 0.00001881, -0.00003340, 0.00002532, -0.00002201, 0.00002106, -0.00001305, 0.00001280, -0.00001345, 0.00001349, -0.00001749, 0.00001139, -0.00002223, 0.00001821, 0.00000739, -0.00000193, -0.00001769, 0.00001555, -0.00001312, 0.00001213, -0.00000674, 0.00001125, -0.00001299, 0.00001214, -0.00001105, 0.00000653, -0.00001612, 0.00001245, 0.00000042, 0.00000499, -0.00000798, 0.00000694, -0.00001493, 0.00001473, -0.00000756, 0.00000864, -0.00000682, 0.00000122, -0.00000985, 0.00000823, 0.00000440, -0.00000224, -0.00000164, },
		{ -0.00000360, 0.00000034, -0.00000236, 0.00001345, -0.00000911, -0.00000066, -0.00000362, 0.00000570, -0.00000770, 0.00001248, -0.00000764, -0.00000149, 0.00000009, -0.00000084, -0.00000241, 0.00001585, -0.00001066, 0.00000105, -0.00000761, 0.00000651, -0.00000527, 0.00000332, -0.00000377, 0.00000801, -0.00000396, 0.00000385, -0.00000067, -0.00001325, 0.00000181, 0.00002421, -0.00001812, 0.00000533, -0.00000885, 0.00000731, -0.00001564, 0.00001795, -0.00001911, 0.00002670, -0.00002574, 0.00002871, -0.00003001, 0.00002055, -0.00003589, 0.00006143, -0.00006102, 0.00003961, -0.00003943, 0.00003573, -0.00002355, 0.00001590, -0.00002069, 0.00001836, -0.00002107, 0.00001866, -0.00004161, 0.00005393, -0.00007812, 0.00008087, -0.00010520, 0.00010071, -0.00015452, 0.00013250, -0.00033239, -0.00009121, 0.00121140, -0.00077717, 0.00022883, -0.00021968, 0.00012547, -0.00011752, 0.00007639, -0.00006967, 0.00004709, -0.00004925, 0.00004541, -0.00006089, 0.00006090, -0.00006047, 0.00005837, -0.00006341, 0.00006535, -0.00005760, 0.00005619, -0.00005948, 0.00002911, -0.00000293, 0.00000732, -0.00002301, 0.00001477, -0.00001212, 0.00001160, -0.00000473, 0.00000638, -0.00000746, 0.00000766, -0.00001230, 0.00000786, -0.00001787, 0.00001069, 0.00001248, -0.00000167, -0.00001835, 0.00001438, -0.00001163, 0.00001089, -0.00000605, 0.00001066, -0.00001210, 0.00001094, -0.00000964, 0.00000588, -0.00001506, 0.00000960, 0.00000255, 0.00000443, -0.00000788, 0.00000762, -0.00001492, 0.00001242, -0.00000462, 0.00000557, -0.00000382, 0.00000006, -0.00000855, 0.00000463, 0.00000743, -0.00000345, -0.00000034, },
		{ -0.00000444, 0.00000102, -0.00000382, 0.00001345, -0.00000654, -0.00000240, -0.00000342, 0.00000455, -0.00000542, 0.00000874, -0.00000226, -0.00000593, 0.00000287, -0.00000299, -0.00000077, 0.00001279, -0.00000535, -0.00000376, -0.00000347, 0.00000174, -0.00000034, -0.00000153, 0.00000096, 0.00000336, 0.00000075, -0.00000099, 0.00000430, -0.00001557, 0.00000041, 0.00002334, -0.00001246, 0.00000002, -0.00000549, 0.00000400, -0.00001156, 0.00001325, -0.00001474, 0.00002136, -0.00001905, 0.00002114, -0.00002148, 0.00001265, -0.00002932, 0.00005069, -0.00004313, 0.00001732, -0.00001566, 0.00001081, 0.00000145, -0.00000697, 0.00000182, -0.00000387, 0.00000174, -0.00000313, -0.00001888, 0.00002782, -0.00004663, 0.00004279, -0.00005872, 0.00004396, -0.00007945, 0.00003410, -0.00016234, -0.00034464, 0.00133282, -0.00069817, 0.00013141, -0.00015294, 0.00006929, -0.00007235, 0.00003811, -0.00003767, 0.00002049, -0.00002773, 0.00002812, -0.00004611, 0.00004577, -0.00004620, 0.00004421, -0.00004955, 0.00004902, -0.00003922, 0.00003565, -0.00003562, 0.00000489, 0.00001617, -0.00000589, -0.00001071, 0.00000238, -0.00000064, 0.00000085, 0.00000458, -0.00000105, -0.00000035, 0.00000092, -0.00000648, 0.00000344, -0.00001223, 0.00000258, 0.00001715, -0.00000127, -0.00001820, 0.00001205, -0.00000903, 0.00000849, -0.00000404, 0.00000862, -0.00000973, 0.00000827, -0.00000685, 0.00000380, -0.00001240, 0.00000539, 0.00000560, 0.00000308, -0.00000687, 0.00000703, -0.00001319, 0.00000846, -0.00000051, 0.00000180, -0.00000061, -0.00000131, -0.00000665, 0.00000069, 0.00001020, -0.00000428, 0.00000065, },
		{ -0.00000474, 0.00000113, -0.00000445, 0.00001231, -0.00000317, -0.00000449, -0.00000257, 0.00000233, -0.00000207, 0.00000418, 0.00000335, -0.00001003, 0.00000533, -0.00000513, 0.00000121, 0.00000918, 0.00000022, -0.00000846, 0.00000053, -0.00000290, 0.00000437, -0.00000607, 0.00000527, -0.00000080, 0.00000503, -0.00000541, 0.00000863, -0.00001725, -0.00000061, 0.00002090, -0.00000537, -0.00000604, -0.00000122, -0.00000079, -0.00000579, 0.00000688, -0.00000844, 0.00001366, -0.00000981, 0.00001094, -0.00001023, 0.00000175, -0.00001877, 0.00003488, -0.00002022, -0.00000887, 0.00001055, -0.00001535, 0.00002675, -0.00002983, 0.00002448, -0.00002643, 0.00002510, -0.00002586, 0.00000559, -0.00000079, -0.00001217, 0.00000145, -0.00000881, -0.00001629, -0.00000169, -0.00006509, -0.00000268, -0.00055155, 0.00137456, -0.00055155, -0.00000268, -0.00006509, -0.00000169, -0.00001629, -0.00000881, 0.00000145, -0.00001217, -0.00000079, 0.00000559, -0.00002586, 0.00002510, -0.00002643, 0.00002448, -0.00002983, 0.00002675, -0.00001535, 0.00001055, -0.00000887, -0.00002022, 0.00003488, -0.00001877, 0.00000175, -0.00001023, 0.00001094, -0.00000981, 0.00001366, -0.00000844, 0.00000688, -0.00000579, -0.00000079, -0.00000122, -0.00000604, -0.00000537, 0.00002090, -0.00000061, -0.00001725, 0.00000863, -0.00000541, 0.00000503, -0.00000080, 0.00000527, -0.00000607, 0.00000437, -0.00000290, 0.00000053, -0.00000846, 0.00000022, 0.00000918, 0.00000121, -0.00000513, 0.00000533, -0.00001003, 0.00000335, 0.00000418, -0.00000207, 0.00000233, -0.00000257, -0.00000449, -0.00000317, 0.00001231, -0.00000445, 0.00000113, },
		{ -0.00000444, 0.00000065, -0.00000428, 0.00001020, 0.00000069, -0.00000665, -0.00000131, -0.00000061, 0.00000180, -0.00000051, 0.00000846, -0.00001319, 0.00000703, -0.00000687, 0.00000308, 0.00000560, 0.00000539, -0.00001240, 0.00000380, -0.00000685, 0.00000827, -0.00000973, 0.00000862, -0.00000404, 0.00000849, -0.00000903, 0.00001205, -0.00001820, -0.00000127, 0.00001715, 0.00000258, -0.00001223, 0.00000344, -0.00000648, 0.00000092, -0.00000035, -0.00000105, 0.00000458, 0.00000085, -0.00000064, 0.00000238, -0.00001071, -0.00000589, 0.00001617, 0.00000489, -0.00003562, 0.00003565, -0.00003922, 0.00004902, -0.00004955, 0.00004421, -0.00004620, 0.00004577, -0.00004611, 0.00002812, -0.00002773, 0.00002049, -0.00003767, 0.00003811, -0.00007235, 0.00006929, -0.00015294, 0.00013141, -0.00069817, 0.00133282, -0.00034464, -0.00016234, 0.00003410, -0.00007945, 0.00004396, -0.00005872, 0.00004279, -0.00004663, 0.00002782, -0.00001888, -0.00000313, 0.00000174, -0.00000387, 0.00000182, -0.00000697, 0.00000145, 0.00001081, -0.00001566, 0.00001732, -0.00004313, 0.00005069, -0.00002932, 0.00001265, -0.00002148, 0.00002114, -0.00001905, 0.00002136, -0.00001474, 0.00001325, -0.00001156, 0.00000400, -0.00000549, 0.00000002, -0.00001246, 0.00002334, 0.00000041, -0.00001557, 0.00000430, -0.00000099, 0.00000075, 0.00000336, 0.00000096, -0.00000153, -0.00000034, 0.00000174, -0.00000347, -0.00000376, -0.00000535, 0.00001279, -0.00000077, -0.00000299, 0.00000287, -0.00000593, -0.00000226, 0.00000874, -0.00000542, 0.00000455, -0.00000342, -0.00000240, -0.00000654, 0.00001345, -0.00000382, 0.00000102, },
		{ -0.00000360, -0.00000034, -0.00000345, 0.00000743, 0.00000463, -0.00000855, 0.00000006, -0.00000382, 0.00000557, -0.00000462, 0.00001242, -0.00001492, 0.00000762, -0.00000788, 0.00000443, 0.00000255, 0.00000960, -0.00001506, 0.00000588, -0.00000964, 0.00001094, -0.00001210, 0.00001066, -0.00000605, 0.00001089, -0.00001163, 0.00001438, -0.00001835, -0.00000167, 0.00001248, 0.00001069, -0.00001787, 0.00000786, -0.00001230, 0.00000766, -0.00000746, 0.00000638, -0.00000473, 0.00001160, -0.00001212, 0.00001477, -0.00002301, 0.00000732, -0.00000293, 0.00002911, -0.00005948, 0.00005619, -0.00005760, 0.00006535, -0.00006341, 0.00005837, -0.00006047, 0.00006090, -0.00006089, 0.00004541, -0.00004925, 0.00004709, -0.00006967, 0.00007639, -0.00011752, 0.00012547, -0.00021968, 0.00022883, -0.00077717, 0.00121140, -0.00009121, -0.00033239, 0.00013250, -0.00015452, 0.00010071, -0.00010520, 0.00008087, -0.00007812, 0.00005393, -0.00004161, 0.00001866, -0.00002107, 0.00001836, -0.00002069, 0.00001590, -0.00002355, 0.00003573, -0.00003943, 0.00003961, -0.00006102, 0.00006143, -0.00003589, 0.00002055, -0.00003001, 0.00002871, -0.00002574, 0.00002670, -0.00001911, 0.00001795, -0.00001564, 0.00000731, -0.00000885, 0.00000533, -0.00001812, 0.00002421, 0.00000181, -0.00001325, -0.00000067, 0.00000385, -0.00000396, 0.00000801, -0.00000377, 0.00000332, -0.00000527, 0.00000651, -0.00000761, 0.00000105, -0.00001066, 0.00001585, -0.00000241, -0.00000084, 0.00000009, -0.00000149, -0.00000764, 0.00001248, -0.00000770, 0.00000570, -0.00000362, -0.00000066, -0.00000911, 0.00001345, -0.00000236, 0.00000034, },
		{ -0.00000238, -0.00000164, -0.00000224, 0.00000440, 0.00000823, -0.00000985, 0.00000122, -0.00000682, 0.00000864, -0.00000756, 0.00001473, -0.00001493, 0.00000694, -0.00000798, 0.00000499, 0.00000042, 0.00001245, -0.00001612, 0.00000653, -0.00001105, 0.00001214, -0.00001299, 0.00001125, -0.00000674, 0.00001213, -0.00001312, 0.00001555, -0.00001769, -0.00000193, 0.00000739, 0.00001821, -0.00002223, 0.00001139, -0.00001749, 0.00001349, -0.00001345, 0.00001280, -0.00001305, 0.00002106, -0.00002201, 0.00002532, -0.00003340, 0.00001881, -0.00001990, 0.00004949, -0.00007741, 0.00006934, -0.00006806, 0.00007364, -0.00006954, 0.00006511, -0.00006732, 0.00006849, -0.00006802, 0.00005503, -0.00006258, 0.00006441, -0.00009092, 0.00010184, -0.00014694, 0.00016123, -0.00025893, 0.00028350, -0.00078810, 0.00102130, 0.00018983, -0.00049512, 0.00021714, -0.00021722, 0.00014634, -0.00014202, 0.00011041, -0.00010210, 0.00007357, -0.00005905, 0.00003623, -0.00004016, 0.00003716, -0.00004001, 0.00003567, -0.00004499, 0.00005613, -0.00005760, 0.00005526, -0.00007179, 0.00006566, -0.00003744, 0.00002454, -0.00003492, 0.00003280, -0.00002915, 0.00002907, -0.00002099, 0.00002042, -0.00001753, 0.00000876, -0.00001093, 0.00000944, -0.00002197, 0.00002341, 0.00000354, -0.00001044, -0.00000592, 0.00000868, -0.00000860, 0.00001258, -0.00000830, 0.00000783, -0.00000977, 0.00001075, -0.00001124, 0.00000531, -0.00001505, 0.00001782, -0.00000329, 0.00000093, -0.00000253, 0.00000267, -0.00001212, 0.00001483, -0.00000852, 0.00000558, -0.00000308, 0.00000057, -0.00001071, 0.00001228, -0.00000018, -0.00000075, },
		{ -0.00000102, -0.00000300, -0.00000095, 0.00000150, 0.00001107, -0.00001026, 0.00000188, -0.00000920, 0.00001053, -0.00000891, 0.00001513, -0.00001315, 0.00000502, -0.00000714, 0.00000463, -0.00000059, 0.00001376, -0.00001551, 0.00000571, -0.00001105, 0.00001186, -0.00001241, 0.00001044, -0.00000621, 0.00001229, -0.00001351, 0.00001559, -0.00001625, -0.00000221, 0.00000240, 0.00002441, -0.00002469, 0.00001345, -0.00002132, 0.00001756, -0.00001743, 0.00001729, -0.00001932, 0.00002803, -0.00002899, 0.00003265, -0.00004038, 0.00002684, -0.00003260, 0.00006368, -0.00008716, 0.00007331, -0.00006929, 0.00007298, -0.00006721, 0.00006371, -0.00006596, 0.00006764, -0.00006651, 0.00005579, -0.00006624, 0.00007075, -0.00009946, 0.00011226, -0.00015815, 0.00017398, -0.00026825, 0.00029471, -0.00073706, 0.00077947, 0.00047625, -0.00063227, 0.00027582, -0.00025903, 0.00017436, -0.00016402, 0.00012713, -0.00011489, 0.00008359, -0.00006839, 0.00004691, -0.00005296, 0.00005000, -0.00005360, 0.00004974, -0.00006023, 0.00006941, -0.00006786, 0.00006248, -0.00007431, 0.00006283, -0.00003370, 0.00002431, -0.00003586, 0.00003307, -0.00002899, 0.00002826, -0.00002018, 0.00002044, -0.00001707, 0.00000826, -0.00001161, 0.00001211, -0.00002381, 0.00002100, 0.00000551, -0.00000732, -0.00001104, 0.00001300, -0.00001264, 0.00001646, -0.00001199, 0.00001135, -0.00001321, 0.00001387, -0.00001376, 0.00000840, -0.00001795, 0.00001828, -0.00000309, 0.00000199, -0.00000460, 0.00000603, -0.00001515, 0.00001541, -0.00000772, 0.00000420, -0.00000185, 0.00000127, -0.00001130, 0.00001007, 0.00000246, -0.00000201, },
		{ 0.00000022, -0.00000414, 0.00000011, -0.00000091, 0.00001285, -0.00000961, 0.00000186, -0.00001067, 0.00001096, -0.00000850, 0.00001365, -0.00000979, 0.00000212, -0.00000552, 0.00000341, -0.00000051, 0.00001357, -0.00001336, 0.00000364, -0.00000984, 0.00001031, -0.00001059, 0.00000846, -0.00000468, 0.00001156, -0.00001293, 0.00001458, -0.00001408, -0.00000267, -0.00000197, 0.00002868, -0.00002481, 0.00001364, -0.00002325, 0.00001924, -0.00001880, 0.00001919, -0.00002277, 0.00003163, -0.00003217, 0.00003583, -0.00004292, 0.00003020, -0.00003961, 0.00007022, -0.00008765, 0.00006759, -0.00006132, 0.00006370, -0.00005689, 0.00005461, -0.00005674, 0.00005867, -0.00005660, 0.00004783, -0.00006024, 0.00006605, -0.00009522, 0.00010759, -0.00015124, 0.00016422, -0.00024911, 0.00026671, -0.00063564, 0.00050694, 0.00074479, -0.00072715, 0.00029883, -0.00027378, 0.00018043, -0.00016790, 0.00012835, -0.00011432, 0.00008213, -0.00006797, 0.00004909, -0.00005781, 0.00005525, -0.00005981, 0.00005641, -0.00006754, 0.00007402, -0.00006906, 0.00006071, -0.00006862, 0.00005339, -0.00002519, 0.00002024, -0.00003303, 0.00002971, -0.00002548, 0.00002448, -0.00001686, 0.00001813, -0.00001438, 0.00000598, -0.00001097, 0.00001327, -0.00002368, 0.00001719, 0.00000761, -0.00000407, -0.00001559, 0.00001631, -0.00001555, 0.00001910, -0.00001425, 0.00001335, -0.00001510, 0.00001540, -0.00001475, 0.00000991, -0.00001902, 0.00001699, -0.00000165, 0.00000215, -0.00000581, 0.00000819, -0.00001642, 0.00001409, -0.00000537, 0.00000178, -0.00000015, 0.00000154, -0.00001101, 0.00000710, 0.00000524, -0.00000316, },
		{ 0.00000109, -0.00000485, 0.00000070, -0.00000255, 0.00001341, -0.00000787, 0.00000114, -0.00001110, 0.00000984, -0.00000642, 0.00001056, -0.00000530, -0.00000136, -0.00000342, 0.00000153, 0.00000049, 0.00001215, -0.00001005, 0.00000070, -0.00000779, 0.00000783, -0.00000787, 0.00000569, -0.00000250, 0.00001020, -0.00001158, 0.00001268, -0.00001126, -0.00000345, -0.00000532, 0.00003061, -0.00002238, 0.00001183, -0.00002300, 0.00001820, -0.00001728, 0.00001823, -0.00002304, 0.00003148, -0.00003118, 0.00003449, -0.00004066, 0.00002841, -0.00004036, 0.00006874, -0.00007904, 0.00005301, -0.00004549, 0.00004738, -0.00004019, 0.00003934, -0.00004112, 0.00004302, -0.00003976, 0.00003258, -0.00004599, 0.00005180, -0.00007987, 0.00008983, -0.00012865, 0.00013528, -0.00020642, 0.00020779, -0.00049926, 0.00022656, 0.00097364, -0.00076661, 0.00028028, -0.00025869, 0.00016299, -0.00015274, 0.00011350, -0.00010001, 0.00006897, -0.00005758, 0.00004246, -0.00005433, 0.00005242, -0.00005812, 0.00005510, -0.00006639, 0.00006968, -0.00006138, 0.00005066, -0.00005589, 0.00003871, -0.00001317, 0.00001326, -0.00002718, 0.00002338, -0.00001922, 0.00001832, -0.00001154, 0.00001391, -0.00000988, 0.00000232, -0.00000928, 0.00001307, -0.00002178, 0.00001230, 0.00000967, -0.00000091, -0.00001918, 0.00001820, -0.00001693, 0.00002008, -0.00001470, 0.00001348, -0.00001515, 0.00001510, -0.00001398, 0.00000963, -0.00001813, 0.00001397, 0.00000098, 0.00000142, -0.00000606, 0.00000901, -0.00001585, 0.00001102, -0.00000181, -0.00000129, 0.00000169, 0.00000156, -0.00001004, 0.00000371, 0.00000777, -0.00000390, },
		{ 0.00000144, -0.00000498, 0.00000068, -0.00000333, 0.00001274, -0.00000519, -0.00000017, -0.00001056, 0.00000734, -0.00000302, 0.00000640, -0.00000026, -0.00000487, -0.00000122, -0.00000069, 0.00000210, 0.00000988, -0.00000605, -0.00000262, -0.00000533, 0.00000486, -0.00000471, 0.00000256, -0.00000006, 0.00000851, -0.00000965, 0.00001007, -0.00000791, -0.00000464, -0.00000737, 0.00003003, -0.00001751, 0.00000813, -0.00002062, 0.00001450, -0.00001302, 0.00001457, -0.00002025, 0.00002770, -0.00002622, 0.00002892, -0.00003391, 0.00002179, -0.00003523, 0.00005999, -0.00006270, 0.00003160, -0.00002420, 0.00002653, -0.00001955, 0.00002023, -0.00002140, 0.00002300, -0.00001833, 0.00001244, -0.00002600, 0.00003074, -0.00005651, 0.00006252, -0.00009469, 0.00009262, -0.00014753, 0.00012884, -0.00034524, -0.00003945, 0.00114466, -0.00074270, 0.00021914, -0.00021483, 0.00012363, -0.00012022, 0.00008419, -0.00007352, 0.00004556, -0.00003851, 0.00002806, -0.00004334, 0.00004226, -0.00004916, 0.00004642, -0.00005746, 0.00005736, -0.00004620, 0.00003414, -0.00003821, 0.00002085, 0.00000065, 0.00000470, -0.00001938, 0.00001504, -0.00001116, 0.00001065, -0.00000496, 0.00000842, -0.00000418, -0.00000215, -0.00000694, 0.00001178, -0.00001844, 0.00000675, 0.00001154, 0.00000197, -0.00002148, 0.00001839, -0.00001653, 0.00001918, -0.00001316, 0.00001165, -0.00001332, 0.00001296, -0.00001148, 0.00000762, -0.00001543, 0.00000947, 0.00000452, -0.00000004, -0.00000542, 0.00000851, -0.00001365, 0.00000662, 0.00000245, -0.00000450, 0.00000331, 0.00000158, -0.00000866, 0.00000028, 0.00000972, -0.00000400, },
		{ 0.00000121, -0.00000449, 0.00000005, -0.00000326, 0.00001103, -0.00000186, -0.00000184, -0.00000925, 0.00000380, 0.00000117, 0.00000181, 0.00000465, -0.00000787, 0.00000068, -0.00000291, 0.00000392, 0.00000725, -0.00000191, -0.00000579, -0.00000293, 0.00000185, -0.00000155, -0.00000052, 0.00000230, 0.00000675, -0.00000737, 0.00000695, -0.00000420, -0.00000625, -0.00000800, 0.00002705, -0.00001059, 0.00000298, -0.00001641, 0.00000854, -0.00000655, 0.00000875, -0.00001494, 0.00002093, -0.00001806, 0.00001999, -0.00002364, 0.00001139, -0.00002546, 0.00004564, -0.00004098, 0.00000627, -0.00000060, 0.00000423, 0.00000209, 0.00000006, -0.00000030, 0.00000141, 0.00000479, -0.00000957, -0.00000346, 0.00000638, -0.00002908, 0.00003021, -0.00005474, 0.00004293, -0.00008106, 0.00004182, -0.00019068, -0.00027167, 0.00124521, -0.00065370, 0.00011955, -0.00014718, 0.00006692, -0.00007442, 0.00004405, -0.00003811, 0.00001485, -0.00001333, 0.00000813, -0.00002682, 0.00002657, -0.00003465, 0.00003202, -0.00004252, 0.00003912, -0.00002593, 0.00001382, -0.00001827, 0.00000225, 0.00001429, -0.00000396, -0.00001087, 0.00000587, -0.00000238, 0.00000246, 0.00000203, 0.00000243, 0.00000198, -0.00000679, -0.00000440, 0.00000973, -0.00001410, 0.00000097, 0.00001305, 0.00000441, -0.00002229, 0.00001677, -0.00001432, 0.00001643, -0.00000973, 0.00000805, -0.00000987, 0.00000928, -0.00000756, 0.00000422, -0.00001132, 0.00000396, 0.00000853, -0.00000190, -0.00000411, 0.00000695, -0.00001020, 0.00000144, 0.00000679, -0.00000730, 0.00000433, 0.00000184, -0.00000716, -0.00000282, 0.00001082, -0.00000336, },
		{ 0.00000045, -0.00000348, -0.00000108, -0.00000254, 0.00000858, 0.00000174, -0.00000353, -0.00000748, -0.00000032, 0.00000553, -0.00000252, 0.00000876, -0.00000988, 0.00000193, -0.00000479, 0.00000557, 0.00000472, 0.00000185, -0.00000836, -0.00000100, -0.00000081, 0.00000124, -0.00000319, 0.00000427, 0.00000514, -0.00000493, 0.00000353, -0.00000029, -0.00000819, -0.00000731, 0.00002204, -0.00000227, -0.00000298, -0.00001100, 0.00000102, 0.00000127, 0.00000168, -0.00000802, 0.00001220, -0.00000788, 0.00000904, -0.00001129, -0.00000123, -0.00001287, 0.00002804, -0.00001684, -0.00001963, 0.00002197, -0.00001640, 0.00002182, -0.00001841, 0.00001942, -0.00001892, 0.00002662, -0.00003035, 0.00001829, -0.00001759, -0.00000175, -0.00000233, -0.00001444, -0.00000689, -0.00001567, -0.00004183, -0.00005070, -0.00045534, 0.00126935, -0.00050446, -0.00000947, -0.00006404, -0.00000015, -0.00002127, -0.00000181, 0.00000169, -0.00001914, 0.00001441, -0.00001427, -0.00000753, 0.00000792, -0.00001703, 0.00001431, -0.00002408, 0.00001769, -0.00000355, -0.00000723, 0.00000107, -0.00001467, 0.00002589, -0.00001131, -0.00000286, -0.00000300, 0.00000603, -0.00000529, 0.00000856, -0.00000332, 0.00000790, -0.00001097, -0.00000213, 0.00000732, -0.00000923, -0.00000461, 0.00001409, 0.00000630, -0.00002154, 0.00001344, -0.00001051, 0.00001208, -0.00000476, 0.00000311, -0.00000526, 0.00000454, -0.00000272, -0.00000005, -0.00000639, -0.00000192, 0.00001248, -0.00000376, -0.00000247, 0.00000471, -0.00000606, -0.00000383, 0.00001056, -0.00000919, 0.00000445, 0.00000248, -0.00000575, -0.00000534, 0.00001092, -0.00000196, },
		{ -0.00000067, -0.00000213, -0.00000249, -0.00000142, 0.00000579, 0.00000515, -0.00000488, -0.00000561, -0.00000450, 0.00000942, -0.00000594, 0.00001157, -0.00001057, 0.00000233, -0.00000611, 0.00000674, 0.00000267, 0.00000483, -0.00000999, 0.00000019, -0.00000286, 0.00000340, -0.00000522, 0.00000568, 0.00000383, -0.00000247, 0.00000002, 0.00000359, -0.00001031, -0.00000554, 0.00001559, 0.00000660, -0.00000895, -0.00000511, -0.00000716, 0.00000941, -0.00000557, -0.00000062, 0.00000276, 0.00000287, -0.00000237, 0.00000141, -0.00001420, 0.00000039, 0.00000982, 0.00000663, -0.00004282, 0.00004041, -0.00003259, 0.00003717, -0.00003284, 0.00003541, -0.00003550, 0.00004455, -0.00004713, 0.00003629, -0.00003782, 0.00002171, -0.00003076, 0.00002111, -0.00005071, 0.00004109, -0.00011256, 0.00006313, -0.00058147, 0.00121824, -0.00030589, -0.00015493, 0.00002396, -0.00006913, 0.00003227, -0.00004750, 0.00004075, -0.00005192, 0.00004082, -0.00003570, 0.00001141, -0.00001076, 0.00000089, -0.00000394, -0.00000498, -0.00000388, 0.00001780, -0.00002603, 0.00001717, -0.00002784, 0.00003400, -0.00001629, 0.00000371, -0.00001061, 0.00001318, -0.00001176, 0.00001393, -0.00000816, 0.00001292, -0.00001415, -0.00000050, 0.00000488, -0.00000425, -0.00000959, 0.00001453, 0.00000758, -0.00001934, 0.00000871, -0.00000552, 0.00000665, 0.00000115, -0.00000250, -0.00000018, -0.00000058, 0.00000237, -0.00000453, -0.00000134, -0.00000750, 0.00001580, -0.00000518, -0.00000088, 0.00000224, -0.00000182, -0.00000853, 0.00001320, -0.00000981, 0.00000353, 0.00000356, -0.00000455, -0.00000711, 0.00001001, 0.00000005, },
		{ -0.00000187, -0.00000069, -0.00000392, -0.00000024, 0.00000308, 0.00000796, -0.00000559, -0.00000395, -0.00000827, 0.00001230, -0.00000799, 0.00001275, -0.00000981, 0.00000181, -0.00000675, 0.00000725, 0.00000137, 0.00000678, -0.00001050, 0.00000052, -0.00000417, 0.00000478, -0.00000649, 0.00000643, 0.00000288, -0.00000012, -0.00000340, 0.00000721, -0.00001238, -0.00000310, 0.00000847, 0.00001507, -0.00001408, 0.00000042, -0.00001502, 0.00001677, -0.00001189, 0.00000611, -0.00000607, 0.00001274, -0.00001265, 0.00001279, -0.00002566, 0.00001220, -0.00000651, 0.00002660, -0.00006054, 0.00005236, -0.00004241, 0.00004647, -0.00004167, 0.00004603, -0.00004658, 0.00005671, -0.00005790, 0.00004829, -0.00005181, 0.00003845, -0.00005179, 0.00004803, -0.00008386, 0.00008367, -0.00016367, 0.00014361, -0.00064732, 0.00109969, -0.00007384, -0.00030138, 0.00010522, -0.00013126, 0.00007920, -0.00008720, 0.00007409, -0.00007917, 0.00006216, -0.00005291, 0.00002704, -0.00002667, 0.00001640, -0.00002005, 0.00001202, -0.00002274, 0.00003533, -0.00004007, 0.00002807, -0.00003587, 0.00003774, -0.00001830, 0.00000822, -0.00001631, 0.00001841, -0.00001636, 0.00001762, -0.00001162, 0.00001659, -0.00001596, 0.00000024, 0.00000268, 0.00000043, -0.00001366, 0.00001431, 0.00000828, -0.00001591, 0.00000304, 0.00000009, 0.00000076, 0.00000730, -0.00000804, 0.00000465, -0.00000539, 0.00000703, -0.00000856, 0.00000316, -0.00001213, 0.00001796, -0.00000578, 0.00000030, -0.00000003, 0.00000198, -0.00001212, 0.00001432, -0.00000900, 0.00000159, 0.00000502, -0.00000358, -0.00000813, 0.00000825, 0.00000242, },
		{ -0.00000289, 0.00000056, -0.00000508, 0.00000069, 0.00000081, 0.00000985, -0.00000545, -0.00000276, -0.00001126, 0.00001381, -0.00000840, 0.00001222, -0.00000768, 0.00000051, -0.00000674, 0.00000702, 0.00000091, 0.00000763, -0.00000990, 0.00000004, -0.00000476, 0.00000540, -0.00000702, 0.00000656, 0.00000228, 0.00000207, -0.00000653, 0.00001032, -0.00001412, -0.00000045, 0.00000146, 0.00002225, -0.00001761, 0.00000486, -0.00002167, 0.00002234, -0.00001632, 0.00001118, -0.00001316, 0.00002045, -0.00002045, 0.00002140, -0.00003408, 0.00002083, -0.00001895, 0.00004097, -0.00007093, 0.00005649, -0.00004499, 0.00004905, -0.00004427, 0.00005058, -0.00005132, 0.00006212, -0.00006161, 0.00005312, -0.00005814, 0.00004682, -0.00006350, 0.00006405, -0.00010368, 0.00010899, -0.00019178, 0.00018824, -0.00065608, 0.00092709, 0.00017266, -0.00043268, 0.00016876, -0.00017871, 0.00011349, -0.00011603, 0.00009759, -0.00009740, 0.00007547, -0.00006332, 0.00003699, -0.00003753, 0.00002731, -0.00003185, 0.00002469, -0.00003666, 0.00004698, -0.00004770, 0.00003265, -0.00003822, 0.00003698, -0.00001730, 0.00001052, -0.00001980, 0.00002140, -0.00001882, 0.00001939, -0.00001344, 0.00001863, -0.00001621, -0.00000001, 0.00000089, 0.00000453, -0.00001660, 0.00001340, 0.00000848, -0.00001164, -0.00000297, 0.00000563, -0.00000486, 0.00001292, -0.00001276, 0.00000852, -0.00000922, 0.00001066, -0.00001153, 0.00000653, -0.00001530, 0.00001863, -0.00000533, 0.00000083, -0.00000175, 0.00000489, -0.00001422, 0.00001375, -0.00000681, -0.00000116, 0.00000665, -0.00000277, -0.00000848, 0.00000587, 0.00000481, },
		{ -0.00000344, 0.00000139, -0.00000575, 0.00000112, -0.00000074, 0.00001063, -0.00000438, -0.00000215, -0.00001326, 0.00001377, -0.00000719, 0.00001017, -0.00000451, -0.00000131, -0.00000623, 0.00000614, 0.00000124, 0.00000748, -0.00000838, -0.00000106, -0.00000478, 0.00000538, -0.00000692, 0.00000617, 0.00000198, 0.00000402, -0.00000919, 0.00001271, -0.00001525, 0.00000191, -0.00000468, 0.00002738, -0.00001898, 0.00000766, -0.00002639, 0.00002540, -0.00001818, 0.00001385, -0.00001767, 0.00002510, -0.00002485, 0.00002629, -0.00003845, 0.00002512, -0.00002622, 0.00004855, -0.00007325, 0.00005265, -0.00004058, 0.00004532, -0.00004101, 0.00004929, -0.00004981, 0.00006084, -0.00005825, 0.00005070, -0.00005663, 0.00004655, -0.00006553, 0.00006866, -0.00010956, 0.00011658, -0.00019687, 0.00019886, -0.00061593, 0.00071766, 0.00041347, -0.00053392, 0.00020565, -0.00020568, 0.00013093, -0.00013070, 0.00010856, -0.00010442, 0.00007899, -0.00006543, 0.00003985, -0.00004199, 0.00003227, -0.00003798, 0.00003163, -0.00004428, 0.00005165, -0.00004832, 0.00003086, -0.00003524, 0.00003226, -0.00001379, 0.00001084, -0.00002115, 0.00002216, -0.00001914, 0.00001925, -0.00001360, 0.00001901, -0.00001490, -0.00000123, -0.00000040, 0.00000783, -0.00001827, 0.00001182, 0.00000829, -0.00000695, -0.00000870, 0.00001044, -0.00000954, 0.00001734, -0.00001599, 0.00001087, -0.00001159, 0.00001280, -0.00001304, 0.00000838, -0.00001673, 0.00001765, -0.00000374, 0.00000061, -0.00000273, 0.00000667, -0.00001468, 0.00001158, -0.00000351, -0.00000437, 0.00000817, -0.00000197, -0.00000833, 0.00000320, 0.00000689, },
		{ -0.00000333, 0.00000166, -0.00000580, 0.00000093, -0.00000146, 0.00001028, -0.00000248, -0.00000207, -0.00001423, 0.00001225, -0.00000458, 0.00000700, -0.00000075, -0.00000327, -0.00000546, 0.00000478, 0.00000218, 0.00000655, -0.00000624, -0.00000249, -0.00000446, 0.00000489, -0.00000635, 0.00000541, 0.00000189, 0.00000569, -0.00001125, 0.00001417, -0.00001552, 0.00000356, -0.00000933, 0.00002994, -0.00001790, 0.00000854, -0.00002878, 0.00002553, -0.00001715, 0.00001384, -0.00001922, 0.00002629, -0.00002547, 0.00002711, -0.00003840, 0.00002470, -0.00002793, 0.00004922, -0.00006793, 0.00004186, -0.00003052, 0.00003661, -0.00003311, 0.00004320, -0.00004301, 0.00005380, -0.00004879, 0.00004198, -0.00004822, 0.00003862, -0.00005890, 0.00006299, -0.00010287, 0.00010832, -0.00018179, 0.00018087, -0.00053859, 0.00049046, 0.00062948, -0.00059318, 0.00021026, -0.00020921, 0.00012969, -0.00012998, 0.00010612, -0.00009963, 0.00007234, -0.00005899, 0.00003536, -0.00003973, 0.00003094, -0.00003806, 0.00003245, -0.00004529, 0.00004937, -0.00004239, 0.00002358, -0.00002809, 0.00002475, -0.00000866, 0.00000972, -0.00002068, 0.00002097, -0.00001759, 0.00001745, -0.00001229, 0.00001785, -0.00001222, -0.00000322, -0.00000124, 0.00001024, -0.00001865, 0.00000961, 0.00000787, -0.00000233, -0.00001354, 0.00001394, -0.00001273, 0.00002004, -0.00001732, 0.00001136, -0.00001224, 0.00001323, -0.00001292, 0.00000859, -0.00001636, 0.00001511, -0.00000116, -0.00000032, -0.00000291, 0.00000726, -0.00001359, 0.00000811, 0.00000045, -0.00000759, 0.00000926, -0.00000098, -0.00000791, 0.00000056, 0.00000835, },
		{ -0.00000247, 0.00000133, -0.00000520, 0.00000011, -0.00000138, 0.00000895, 0.00000002, -0.00000236, -0.00001429, 0.00000949, -0.00000101, 0.00000324, 0.00000304, -0.00000494, -0.00000470, 0.00000316, 0.00000349, 0.00000517, -0.00000383, -0.00000394, -0.00000403, 0.00000417, -0.00000552, 0.00000444, 0.00000194, 0.00000705, -0.00001257, 0.00001457, -0.00001477, 0.00000419, -0.00001209, 0.00002974, -0.00001443, 0.00000753, -0.00002876, 0.00002275, -0.00001336, 0.00001126, -0.00001790, 0.00002413, -0.00002251, 0.00002413, -0.00003425, 0.00001992, -0.00002459, 0.00004383, -0.00005642, 0.00002607, -0.00001689, 0.00002488, -0.00002236, 0.00003394, -0.00003249, 0.00004258, -0.00003489, 0.00002870, -0.00003474, 0.00002498, -0.00004574, 0.00004943, -0.00008643, 0.00008782, -0.00015152, 0.00014205, -0.00043765, 0.00026433, 0.00080455, -0.00060291, 0.00018100, -0.00018964, 0.00011061, -0.00011483, 0.00009125, -0.00008410, 0.00005665, -0.00004505, 0.00002441, -0.00003150, 0.00002398, -0.00003268, 0.00002772, -0.00004038, 0.00004114, -0.00003133, 0.00001250, -0.00001846, 0.00001594, -0.00000301, 0.00000788, -0.00001891, 0.00001829, -0.00001461, 0.00001438, -0.00000982, 0.00001546, -0.00000848, -0.00000571, -0.00000170, 0.00001173, -0.00001780, 0.00000688, 0.00000735, 0.00000181, -0.00001702, 0.00001573, -0.00001410, 0.00002074, -0.00001655, 0.00000993, -0.00001116, 0.00001200, -0.00001124, 0.00000725, -0.00001439, 0.00001130, 0.00000213, -0.00000174, -0.00000243, 0.00000679, -0.00001125, 0.00000381, 0.00000454, -0.00001033, 0.00000961, 0.00000034, -0.00000740, -0.00000175, 0.00000897, },
		{ -0.00000090, 0.00000048, -0.00000405, -0.00000122, -0.00000066, 0.00000692, 0.00000278, -0.00000278, -0.00001366, 0.00000588, 0.00000298, -0.00000049, 0.00000631, -0.00000596, -0.00000419, 0.00000153, 0.00000488, 0.00000367, -0.00000152, -0.00000510, -0.00000373, 0.00000338, -0.00000459, 0.00000338, 0.00000206, 0.00000806, -0.00001306, 0.00001385, -0.00001295, 0.00000366, -0.00001283, 0.00002689, -0.00000893, 0.00000495, -0.00002659, 0.00001742, -0.00000733, 0.00000665, -0.00001423, 0.00001922, -0.00001670, 0.00001814, -0.00002692, 0.00001181, -0.00001741, 0.00003402, -0.00004087, 0.00000782, -0.00000219, 0.00001238, -0.00001079, 0.00002339, -0.00002008, 0.00002913, -0.00001862, 0.00001305, -0.00001848, 0.00000817, -0.00002886, 0.00003112, -0.00006397, 0.00005975, -0.00011215, 0.00009124, -0.00032686, 0.00005595, 0.00092710, -0.00056072, 0.00012062, -0.00015049, 0.00007694, -0.00008819, 0.00006662, -0.00006032, 0.00003424, -0.00002572, 0.00000886, -0.00001895, 0.00001293, -0.00002326, 0.00001882, -0.00003108, 0.00002877, -0.00001721, -0.00000023, -0.00000832, 0.00000745, 0.00000204, 0.00000603, -0.00001640, 0.00001467, -0.00001073, 0.00001052, -0.00000660, 0.00001220, -0.00000408, -0.00000839, -0.00000194, 0.00001235, -0.00001587, 0.00000376, 0.00000687, 0.00000513, -0.00001881, 0.00001562, -0.00001355, 0.00001945, -0.00001382, 0.00000682, -0.00000864, 0.00000942, -0.00000831, 0.00000472, -0.00001125, 0.00000670, 0.00000570, -0.00000336, -0.00000151, 0.00000554, -0.00000809, -0.00000075, 0.00000818, -0.00001218, 0.00000900, 0.00000208, -0.00000696, -0.00000354, 0.00000866, },
		{ 0.00000121, -0.00000069, -0.00000257, -0.00000283, 0.00000041, 0.00000456, 0.00000541, -0.00000302, -0.00001263, 0.00000188, 0.00000681, -0.00000363, 0.00000861, -0.00000606, -0.00000409, 0.00000005, 0.00000611, 0.00000235, 0.00000040, -0.00000576, -0.00000368, 0.00000266, -0.00000369, 0.00000235, 0.00000223, 0.00000867, -0.00001270, 0.00001204, -0.00001012, 0.00000201, -0.00001168, 0.00002183, -0.00000205, 0.00000138, -0.00002280, 0.00001023, 0.00000010, 0.00000086, -0.00000905, 0.00001253, -0.00000912, 0.00001037, -0.00001772, 0.00000182, -0.00000807, 0.00002189, -0.00002384, -0.00001021, 0.00001111, 0.00000127, -0.00000029, 0.00001333, -0.00000760, 0.00001539, -0.00000208, -0.00000273, -0.00000188, -0.00000915, -0.00001119, 0.00001142, -0.00003946, 0.00002901, -0.00006997, 0.00003711, -0.00021854, -0.00012160, 0.00099096, -0.00046966, 0.00003582, -0.00009795, 0.00003388, -0.00005446, 0.00003606, -0.00003177, 0.00000830, -0.00000382, -0.00000881, -0.00000434, -0.00000016, -0.00001178, 0.00000768, -0.00001942, 0.00001450, -0.00000239, -0.00001238, 0.00000042, 0.00000066, 0.00000558, 0.00000474, -0.00001364, 0.00001061, -0.00000646, 0.00000636, -0.00000306, 0.00000848, 0.00000057, -0.00001093, -0.00000211, 0.00001223, -0.00001305, 0.00000042, 0.00000649, 0.00000739, -0.00001881, 0.00001369, -0.00001128, 0.00001644, -0.00000950, 0.00000249, -0.00000517, 0.00000595, -0.00000462, 0.00000150, -0.00000747, 0.00000189, 0.00000905, -0.00000481, -0.00000044, 0.00000387, -0.00000461, -0.00000499, 0.00001089, -0.00001281, 0.00000735, 0.00000422, -0.00000665, -0.00000470, 0.00000745, },
		{ 0.00000359, -0.00000192, -0.00000102, -0.00000443, 0.00000153, 0.00000227, 0.00000753, -0.00000283, -0.00001149, -0.00000208, 0.00000997, -0.00000573, 0.00000964, -0.00000515, -0.00000444, -0.00000116, 0.00000700, 0.00000143, 0.00000172, -0.00000580, -0.00000395, 0.00000206, -0.00000287, 0.00000138, 0.00000244, 0.00000888, -0.00001150, 0.00000928, -0.00000651, -0.00000055, -0.00000899, 0.00001527, 0.00000537, -0.00000245, -0.00001808, 0.00000206, 0.00000793, -0.00000513, -0.00000338, 0.00000518, -0.00000104, 0.00000217, -0.00000813, -0.00000842, 0.00000155, 0.00000963, -0.00000779, -0.00002556, 0.00002094, -0.00000680, 0.00000766, 0.00000514, 0.00000345, 0.00000302, 0.00001288, -0.00001664, 0.00001291, -0.00002459, 0.00000459, -0.00000665, -0.00001647, 0.00000001, -0.00003053, -0.00001296, -0.00012252, -0.00025975, 0.00099570, -0.00033763, -0.00006359, -0.00004000, -0.00001227, -0.00001881, 0.00000402, -0.00000240, -0.00001766, 0.00001754, -0.00002589, 0.00000990, -0.00001299, -0.00000038, -0.00000359, -0.00000758, 0.00000063, 0.00001085, -0.00002195, 0.00000627, -0.00000343, 0.00000708, 0.00000431, -0.00001095, 0.00000654, -0.00000224, 0.00000230, 0.00000043, 0.00000468, 0.00000507, -0.00001303, -0.00000234, 0.00001148, -0.00000961, -0.00000293, 0.00000623, 0.00000854, -0.00001713, 0.00001024, -0.00000770, 0.00001220, -0.00000420, -0.00000241, -0.00000137, 0.00000221, -0.00000076, -0.00000185, -0.00000366, -0.00000257, 0.00001175, -0.00000577, 0.00000049, 0.00000213, -0.00000130, -0.00000840, 0.00001230, -0.00001209, 0.00000472, 0.00000665, -0.00000644, -0.00000526, 0.00000552, },
		{ 0.00000589, -0.00000290, 0.00000033, -0.00000575, 0.00000237, 0.00000039, 0.00000884, -0.00000203, -0.00001046, -0.00000562, 0.00001209, -0.00000652, 0.00000933, -0.00000328, -0.00000515, -0.00000208, 0.00000747, 0.00000101, 0.00000238, -0.00000523, -0.00000449, 0.00000157, -0.00000215, 0.00000047, 0.00000271, 0.00000866, -0.00000957, 0.00000581, -0.00000241, -0.00000366, -0.00000531, 0.00000801, 0.00001242, -0.00000580, -0.00001318, -0.00000615, 0.00001511, -0.00001029, 0.00000177, -0.00000169, 0.00000628, -0.00000514, 0.00000045, -0.00001739, 0.00000978, -0.00000084, 0.00000523, -0.00003643, 0.00002598, -0.00001089, 0.00001227, -0.00000034, 0.00001211, -0.00000681, 0.00002495, -0.00002723, 0.00002429, -0.00003636, 0.00001650, -0.00002080, 0.00000223, -0.00002385, 0.00000199, -0.00005355, -0.00004533, -0.00035481, 0.00094620, -0.00017645, -0.00016593, 0.00001477, -0.00005511, 0.00001362, -0.00002521, 0.00002402, -0.00004032, 0.00003551, -0.00003990, 0.00002153, -0.00002351, 0.00000899, -0.00001307, 0.00000249, -0.00001083, 0.00002068, -0.00002753, 0.00000835, -0.00000442, 0.00000648, 0.00000472, -0.00000846, 0.00000272, 0.00000162, -0.00000135, 0.00000359, 0.00000112, 0.00000909, -0.00001445, -0.00000275, 0.00001026, -0.00000581, -0.00000608, 0.00000606, 0.00000863, -0.00001409, 0.00000577, -0.00000338, 0.00000738, 0.00000139, -0.00000717, 0.00000211, -0.00000123, 0.00000272, -0.00000478, -0.00000035, -0.00000618, 0.00001342, -0.00000600, 0.00000103, 0.00000066, 0.00000145, -0.00001065, 0.00001227, -0.00001010, 0.00000134, 0.00000913, -0.00000620, -0.00000535, 0.00000313, },
		{ 0.00000779, -0.00000337, 0.00000124, -0.00000655, 0.00000271, -0.00000081, 0.00000918, -0.00000056, -0.00000965, -0.00000852, 0.00001296, -0.00000595, 0.00000779, -0.00000071, -0.00000603, -0.00000280, 0.00000752, 0.00000110, 0.00000240, -0.00000418, -0.00000518, 0.00000110, -0.00000148, -0.00000039, 0.00000305, 0.00000807, -0.00000706, 0.00000192, 0.00000178, -0.00000688, -0.00000127, 0.00000095, 0.00001826, -0.00000800, -0.00000879, -0.00001353, 0.00002073, -0.00001380, 0.00000553, -0.00000712, 0.00001184, -0.00001051, 0.00000691, -0.00002391, 0.00001532, -0.00000810, 0.00001388, -0.00004184, 0.00002579, -0.00001086, 0.00001345, -0.00000296, 0.00001803, -0.00001355, 0.00003341, -0.00003370, 0.00003138, -0.00004345, 0.00002338, -0.00002968, 0.00001499, -0.00004053, 0.00002509, -0.00008159, 0.00000990, -0.00040765, 0.00085169, -0.00000034, -0.00025917, 0.00005838, -0.00008901, 0.00003854, -0.00004808, 0.00004444, -0.00005706, 0.00004789, -0.00004896, 0.00002888, -0.00003015, 0.00001486, -0.00001934, 0.00000932, -0.00001849, 0.00002597, -0.00002846, 0.00000649, -0.00000253, 0.00000417, 0.00000567, -0.00000612, -0.00000073, 0.00000492, -0.00000439, 0.00000620, -0.00000196, 0.00001237, -0.00001506, -0.00000341, 0.00000873, -0.00000193, -0.00000881, 0.00000588, 0.00000786, -0.00001014, 0.00000090, 0.00000101, 0.00000267, 0.00000655, -0.00001111, 0.00000468, -0.00000384, 0.00000534, -0.00000685, 0.00000205, -0.00000861, 0.00001387, -0.00000539, 0.00000106, -0.00000035, 0.00000337, -0.00001157, 0.00001084, -0.00000707, -0.00000243, 0.00001136, -0.00000577, -0.00000514, 0.00000061, },
		{ 0.00000899, -0.00000315, 0.00000158, -0.00000672, 0.00000241, -0.00000124, 0.00000854, 0.00000149, -0.00000907, -0.00001066, 0.00001255, -0.00000418, 0.00000534, 0.00000220, -0.00000681, -0.00000342, 0.00000722, 0.00000162, 0.00000194, -0.00000283, -0.00000582, 0.00000056, -0.00000081, -0.00000124, 0.00000346, 0.00000715, -0.00000419, -0.00000203, 0.00000566, -0.00000975, 0.00000251, -0.00000514, 0.00002220, -0.00000859, -0.00000541, -0.00001942, 0.00002412, -0.00001510, 0.00000737, -0.00001049, 0.00001497, -0.00001330, 0.00001062, -0.00002732, 0.00001747, -0.00001147, 0.00001767, -0.00004178, 0.00002086, -0.00000739, 0.00001181, -0.00000315, 0.00002141, -0.00001722, 0.00003819, -0.00003594, 0.00003403, -0.00004566, 0.00002501, -0.00003298, 0.00002136, -0.00004940, 0.00003806, -0.00009632, 0.00004323, -0.00042290, 0.00072433, 0.00017576, -0.00033239, 0.00008457, -0.00011004, 0.00005305, -0.00006230, 0.00005691, -0.00006629, 0.00005342, -0.00005206, 0.00003102, -0.00003208, 0.00001647, -0.00002165, 0.00001216, -0.00002171, 0.00002640, -0.00002491, 0.00000126, 0.00000147, 0.00000090, 0.00000664, -0.00000372, -0.00000380, 0.00000760, -0.00000674, 0.00000817, -0.00000442, 0.00001477, -0.00001477, -0.00000433, 0.00000702, 0.00000176, -0.00001093, 0.00000559, 0.00000650, -0.00000581, -0.00000373, 0.00000484, -0.00000131, 0.00001068, -0.00001369, 0.00000593, -0.00000530, 0.00000683, -0.00000781, 0.00000331, -0.00000971, 0.00001309, -0.00000398, 0.00000059, -0.00000079, 0.00000437, -0.00001121, 0.00000826, -0.00000341, -0.00000617, 0.00001302, -0.00000498, -0.00000483, -0.00000174, },
		{ 0.00000929, -0.00000216, 0.00000132, -0.00000622, 0.00000148, -0.00000093, 0.00000705, 0.00000392, -0.00000860, -0.00001209, 0.00001103, -0.00000154, 0.00000241, 0.00000502, -0.00000721, -0.00000410, 0.00000668, 0.00000243, 0.00000120, -0.00000143, -0.00000624, -0.00000016, -0.00000009, -0.00000210, 0.00000393, 0.00000603, -0.00000118, -0.00000569, 0.00000882, -0.00001186, 0.00000550, -0.00000965, 0.00002382, -0.00000737, -0.00000329, -0.00002342, 0.00002497, -0.00001400, 0.00000711, -0.00001158, 0.00001548, -0.00001337, 0.00001144, -0.00002752, 0.00001618, -0.00001100, 0.00001694, -0.00003706, 0.00001241, -0.00000177, 0.00000843, -0.00000177, 0.00002290, -0.00001832, 0.00003973, -0.00003441, 0.00003275, -0.00004351, 0.00002195, -0.00003129, 0.00002190, -0.00005107, 0.00004171, -0.00009895, 0.00005726, -0.00040791, 0.00057768, 0.00033779, -0.00037715, 0.00008970, -0.00011640, 0.00005609, -0.00006710, 0.00006085, -0.00006764, 0.00005193, -0.00004915, 0.00002796, -0.00002930, 0.00001382, -0.00002001, 0.00001105, -0.00002064, 0.00002248, -0.00001777, -0.00000618, 0.00000642, -0.00000236, 0.00000702, -0.00000104, -0.00000655, 0.00000968, -0.00000840, 0.00000948, -0.00000618, 0.00001621, -0.00001360, -0.00000550, 0.00000526, 0.00000503, -0.00001228, 0.00000507, 0.00000487, -0.00000164, -0.00000753, 0.00000758, -0.00000409, 0.00001335, -0.00001462, 0.00000568, -0.00000549, 0.00000711, -0.00000761, 0.00000342, -0.00000955, 0.00001125, -0.00000196, -0.00000029, -0.00000073, 0.00000451, -0.00000976, 0.00000491, 0.00000040, -0.00000941, 0.00001382, -0.00000368, -0.00000462, -0.00000366, },
		{ 0.00000865, -0.00000047, 0.00000054, -0.00000517, 0.00000005, -0.00000004, 0.00000499, 0.00000642, -0.00000806, -0.00001294, 0.00000867, 0.00000154, -0.00000051, 0.00000732, -0.00000698, -0.00000496, 0.00000603, 0.00000336, 0.00000038, -0.00000021, -0.00000627, -0.00000113, 0.00000070, -0.00000295, 0.00000441, 0.00000482, 0.00000169, -0.00000872, 0.00001097, -0.00001292, 0.00000732, -0.00001222, 0.00002304, -0.00000445, -0.00000242, -0.00002544, 0.00002329, -0.00001067, 0.00000494, -0.00001057, 0.00001361, -0.00001105, 0.00000976, -0.00002494, 0.00001199, -0.00000740, 0.00001273, -0.00002917, 0.00000220, 0.00000440, 0.00000463, 0.00000009, 0.00002333, -0.00001758, 0.00003879, -0.00002997, 0.00002848, -0.00003803, 0.00001535, -0.00002583, 0.00001795, -0.00004709, 0.00003799, -0.00009207, 0.00005628, -0.00037138, 0.00042510, 0.00047409, -0.00038835, 0.00007316, -0.00010870, 0.00004844, -0.00006325, 0.00005700, -0.00006190, 0.00004430, -0.00004109, 0.00002048, -0.00002257, 0.00000765, -0.00001512, 0.00000668, -0.00001614, 0.00001536, -0.00000849, -0.00001431, 0.00001099, -0.00000462, 0.00000627, 0.00000211, -0.00000903, 0.00001120, -0.00000942, 0.00001015, -0.00000723, 0.00001669, -0.00001164, -0.00000685, 0.00000354, 0.00000770, -0.00001274, 0.00000421, 0.00000327, 0.00000190, -0.00001008, 0.00000891, -0.00000541, 0.00001438, -0.00001383, 0.00000400, -0.00000455, 0.00000630, -0.00000642, 0.00000256, -0.00000839, 0.00000867, 0.00000036, -0.00000135, -0.00000028, 0.00000397, -0.00000757, 0.00000128, 0.00000388, -0.00001179, 0.00001358, -0.00000182, -0.00000459, -0.00000498, },
		{ 0.00000713, 0.00000174, -0.00000055, -0.00000377, -0.00000168, 0.00000117, 0.00000270, 0.00000867, -0.00000725, -0.00001339, 0.00000581, 0.00000461, -0.00000296, 0.00000875, -0.00000598, -0.00000603, 0.00000533, 0.00000426, -0.00000032, 0.00000065, -0.00000584, -0.00000234, 0.00000152, -0.00000376, 0.00000482, 0.00000369, 0.00000421, -0.00001088, 0.00001191, -0.00001277, 0.00000781, -0.00001278, 0.00002007, -0.00000021, -0.00000253, -0.00002568, 0.00001944, -0.00000563, 0.00000141, -0.00000798, 0.00000996, -0.00000703, 0.00000635, -0.00002045, 0.00000589, -0.00000185, 0.00000652, -0.00001993, -0.00000787, 0.00000950, 0.00000167, 0.00000144, 0.00002357, -0.00001584, 0.00003627, -0.00002366, 0.00002239, -0.00003052, 0.00000665, -0.00001822, 0.00001126, -0.00003949, 0.00002944, -0.00007903, 0.00004532, -0.00032221, 0.00027837, 0.00057639, -0.00036479, 0.00003744, -0.00008971, 0.00003251, -0.00005280, 0.00004714, -0.00005078, 0.00003220, -0.00002946, 0.00001003, -0.00001321, -0.00000077, -0.00000818, 0.00000027, -0.00000955, 0.00000661, 0.00000123, -0.00002155, 0.00001391, -0.00000508, 0.00000406, 0.00000577, -0.00001122, 0.00001219, -0.00000985, 0.00001022, -0.00000760, 0.00001628, -0.00000902, -0.00000829, 0.00000194, 0.00000963, -0.00001229, 0.00000298, 0.00000193, 0.00000445, -0.00001111, 0.00000873, -0.00000527, 0.00001385, -0.00001153, 0.00000121, -0.00000279, 0.00000474, -0.00000455, 0.00000107, -0.00000660, 0.00000576, 0.00000264, -0.00000236, 0.00000033, 0.00000302, -0.00000505, -0.00000216, 0.00000659, -0.00001302, 0.00001224, 0.00000057, -0.00000479, -0.00000565, },
		{ 0.00000493, 0.00000418, -0.00000169, -0.00000227, -0.00000341, 0.00000239, 0.00000053, 0.00001033, -0.00000597, -0.00001364, 0.00000280, 0.00000723, -0.00000458, 0.00000913, -0.00000418, -0.00000728, 0.00000464, 0.00000504, -0.00000075, 0.00000104, -0.00000492, -0.00000372, 0.00000231, -0.00000444, 0.00000507, 0.00000275, 0.00000618, -0.00001200, 0.00001159, -0.00001145, 0.00000702, -0.00001153, 0.00001539, 0.00000475, -0.00000318, -0.00002453, 0.00001400, 0.00000037, -0.00000277, -0.00000456, 0.00000537, -0.00000225, 0.00000218, -0.00001510, -0.00000094, 0.00000428, -0.00000005, -0.00001117, -0.00001607, 0.00001221, 0.00000048, 0.00000152, 0.00002424, -0.00001375, 0.00003299, -0.00001648, 0.00001565, -0.00002228, -0.00000267, -0.00001004, 0.00000362, -0.00003040, 0.00001871, -0.00006321, 0.00002935, -0.00026841, 0.00014674, 0.00064041, -0.00030910, -0.00001234, -0.00006387, 0.00001183, -0.00003865, 0.00003380, -0.00003661, 0.00001779, -0.00001623, -0.00000162, -0.00000289, -0.00000987, -0.00000069, -0.00000674, -0.00000243, -0.00000208, 0.00000970, -0.00002649, 0.00001428, -0.00000330, 0.00000033, 0.00000980, -0.00001301, 0.00001265, -0.00000974, 0.00000975, -0.00000734, 0.00001508, -0.00000590, -0.00000971, 0.00000051, 0.00001077, -0.00001097, 0.00000139, 0.00000101, 0.00000585, -0.00001062, 0.00000717, -0.00000389, 0.00001206, -0.00000812, -0.00000223, -0.00000067, 0.00000283, -0.00000240, -0.00000066, -0.00000461, 0.00000296, 0.00000453, -0.00000308, 0.00000092, 0.00000193, -0.00000256, -0.00000500, 0.00000824, -0.00001299, 0.00000987, 0.00000333, -0.00000513, -0.00000573, },
		{ 0.00000233, 0.00000651, -0.00000258, -0.00000093, -0.00000489, 0.00000334, -0.00000119, 0.00001117, -0.00000413, -0.00001382, -0.00000007, 0.00000910, -0.00000516, 0.00000840, -0.00000170, -0.00000860, 0.00000393, 0.00000563, -0.00000086, 0.00000095, -0.00000360, -0.00000516, 0.00000294, -0.00000491, 0.00000508, 0.00000214, 0.00000748, -0.00001205, 0.00001011, -0.00000915, 0.00000518, -0.00000886, 0.00000968, 0.00000972, -0.00000385, -0.00002251, 0.00000772, 0.00000649, -0.00000676, -0.00000111, 0.00000074, 0.00000233, -0.00000174, -0.00000998, -0.00000731, 0.00000965, -0.00000550, -0.00000443, -0.00002113, 0.00001176, 0.00000150, 0.00000003, 0.00002564, -0.00001173, 0.00002954, -0.00000923, 0.00000921, -0.00001439, -0.00001138, -0.00000269, -0.00000341, -0.00002165, 0.00000808, -0.00004756, 0.00001255, -0.00021632, 0.00003629, 0.00066586, -0.00022722, -0.00006914, -0.00003652, -0.00000959, -0.00002398, 0.00001967, -0.00002183, 0.00000332, -0.00000341, -0.00001268, 0.00000674, -0.00001812, 0.00000592, -0.00001291, 0.00000376, -0.00000919, 0.00001556, -0.00002818, 0.00001162, 0.00000074, -0.00000462, 0.00001384, -0.00001422, 0.00001252, -0.00000910, 0.00000879, -0.00000651, 0.00001321, -0.00000248, -0.00001101, -0.00000076, 0.00001111, -0.00000890, -0.00000046, 0.00000055, 0.00000608, -0.00000880, 0.00000458, -0.00000168, 0.00000949, -0.00000417, -0.00000574, 0.00000129, 0.00000104, -0.00000040, -0.00000221, -0.00000284, 0.00000062, 0.00000578, -0.00000335, 0.00000131, 0.00000095, -0.00000045, -0.00000695, 0.00000868, -0.00001177, 0.00000670, 0.00000621, -0.00000547, -0.00000537, },
		{ -0.00000036, 0.00000841, -0.00000298, 0.00000002, -0.00000591, 0.00000378, -0.00000226, 0.00001107, -0.00000173, -0.00001399, -0.00000261, 0.00001005, -0.00000467, 0.00000671, 0.00000120, -0.00000978, 0.00000313, 0.00000604, -0.00000064, 0.00000044, -0.00000202, -0.00000648, 0.00000330, -0.00000508, 0.00000477, 0.00000191, 0.00000808, -0.00001111, 0.00000771, -0.00000619, 0.00000267, -0.00000535, 0.00000370, 0.00001399, -0.00000402, -0.00002016, 0.00000132, 0.00001191, -0.00000983, 0.00000163, -0.00000311, 0.00000584, -0.00000454, -0.00000598, -0.00001224, 0.00001323, -0.00000874, -0.00000067, -0.00002246, 0.00000805, 0.00000458, -0.00000289, 0.00002766, -0.00000986, 0.00002624, -0.00000241, 0.00000372, -0.00000760, -0.00001862, 0.00000291, -0.00000878, -0.00001454, -0.00000077, -0.00003420, -0.00000207, -0.00017026, -0.00005009, 0.00065604, -0.00012749, -0.00012509, -0.00001301, -0.00002798, -0.00001170, 0.00000718, -0.00000863, -0.00000924, 0.00000729, -0.00002162, 0.00001430, -0.00002423, 0.00001046, -0.00001711, 0.00000785, -0.00001361, 0.00001798, -0.00002625, 0.00000606, 0.00000659, -0.00001020, 0.00001738, -0.00001458, 0.00001174, -0.00000795, 0.00000739, -0.00000521, 0.00001085, 0.00000105, -0.00001207, -0.00000185, 0.00001073, -0.00000626, -0.00000244, 0.00000050, 0.00000529, -0.00000601, 0.00000146, 0.00000084, 0.00000669, -0.00000026, -0.00000875, 0.00000262, -0.00000025, 0.00000110, -0.00000326, -0.00000158, -0.00000097, 0.00000624, -0.00000309, 0.00000141, 0.00000024, 0.00000108, -0.00000787, 0.00000797, -0.00000959, 0.00000308, 0.00000890, -0.00000563, -0.00000477, },
		{ -0.00000281, 0.00000959, -0.00000271, 0.00000045, -0.00000632, 0.00000362, -0.00000258, 0.00001006, 0.00000112, -0.00001411, -0.00000475, 0.00001004, -0.00000324, 0.00000434, 0.00000419, -0.00001059, 0.00000216, 0.00000631, -0.00000016, -0.00000037, -0.00000039, -0.00000750, 0.00000327, -0.00000490, 0.00000411, 0.00000207, 0.00000802, -0.00000936, 0.00000471, -0.00000298, -0.00000005, -0.00000159, -0.00000180, 0.00001696, -0.00000332, -0.00001792, -0.00000455, 0.00001596, -0.00001142, 0.00000312, -0.00000562, 0.00000772, -0.00000570, -0.00000365, -0.00001516, 0.00001443, -0.00000924, -0.00000023, -0.00002016, 0.00000164, 0.00000905, -0.00000665, 0.00002987, -0.00000791, 0.00002310, 0.00000377, -0.00000051, -0.00000226, -0.00002396, 0.00000629, -0.00001196, -0.00000974, -0.00000697, -0.00002425, -0.00001286, -0.00013246, -0.00011247, 0.00061698, -0.00001951, -0.00017256, 0.00000215, -0.00004046, -0.00000394, -0.00000192, 0.00000143, -0.00001852, 0.00001469, -0.00002745, 0.00001888, -0.00002738, 0.00001219, -0.00001867, 0.00000919, -0.00001480, 0.00001674, -0.00002096, -0.00000176, 0.00001340, -0.00001559, 0.00001986, -0.00001383, 0.00001025, -0.00000630, 0.00000559, -0.00000351, 0.00000815, 0.00000447, -0.00001278, -0.00000281, 0.00000975, -0.00000329, -0.00000435, 0.00000073, 0.00000377, -0.00000275, -0.00000166, 0.00000313, 0.00000418, 0.00000307, -0.00001079, 0.00000299, -0.00000079, 0.00000188, -0.00000361, -0.00000103, -0.00000173, 0.00000591, -0.00000236, 0.00000122, -0.00000011, 0.00000195, -0.00000781, 0.00000630, -0.00000679, -0.00000059, 0.00001110, -0.00000542, -0.00000414, },
		{ -0.00000480, 0.00000989, -0.00000172, 0.00000034, -0.00000613, 0.00000286, -0.00000220, 0.00000828, 0.00000417, -0.00001407, -0.00000649, 0.00000919, -0.00000113, 0.00000166, 0.00000690, -0.00001081, 0.00000092, 0.00000649, 0.00000049, -0.00000132, 0.00000110, -0.00000806, 0.00000279, -0.00000435, 0.00000313, 0.00000258, 0.00000744, -0.00000705, 0.00000149, 0.00000007, -0.00000255, 0.00000188, -0.00000622, 0.00001822, -0.00000154, -0.00001607, -0.00000943, 0.00001823, -0.00001124, 0.00000312, -0.00000650, 0.00000773, -0.00000505, -0.00000313, -0.00001592, 0.00001318, -0.00000706, -0.00000276, -0.00001502, -0.00000635, 0.00001384, -0.00001041, 0.00003161, -0.00000548, 0.00001993, 0.00000936, -0.00000350, 0.00000161, -0.00002741, 0.00000747, -0.00001298, -0.00000730, -0.00001038, -0.00001793, -0.00001939, -0.00010332, -0.00015321, 0.00055643, 0.00008708, -0.00020516, 0.00000598, -0.00004545, -0.00000171, -0.00000681, 0.00000759, -0.00002388, 0.00001831, -0.00002979, 0.00002015, -0.00002731, 0.00001095, -0.00001743, 0.00000768, -0.00001283, 0.00001226, -0.00001313, -0.00001077, 0.00002010, -0.00001989, 0.00002072, -0.00001180, 0.00000804, -0.00000421, 0.00000348, -0.00000156, 0.00000531, 0.00000760, -0.00001305, -0.00000367, 0.00000831, -0.00000026, -0.00000598, 0.00000106, 0.00000184, 0.00000049, -0.00000423, 0.00000474, 0.00000239, 0.00000543, -0.00001157, 0.00000225, -0.00000048, 0.00000188, -0.00000323, -0.00000117, -0.00000171, 0.00000492, -0.00000130, 0.00000083, -0.00000014, 0.00000220, -0.00000692, 0.00000402, -0.00000379, -0.00000389, 0.00001253, -0.00000469, -0.00000366, },
		{ -0.00000617, 0.00000926, -0.00000008, -0.00000021, -0.00000542, 0.00000161, -0.00000128, 0.00000601, 0.00000715, -0.00001370, -0.00000793, 0.00000771, 0.00000130, -0.00000092, 0.00000898, -0.00001027, -0.00000061, 0.00000664, 0.00000122, -0.00000225, 0.00000226, -0.00000805, 0.00000184, -0.00000348, 0.00000190, 0.00000333, 0.00000650, -0.00000449, -0.00000156, 0.00000259, -0.00000447, 0.00000461, -0.00000917, 0.00001764, 0.00000126, -0.00001468, -0.00001309, 0.00001859, -0.00000932, 0.00000169, -0.00000583, 0.00000601, -0.00000281, -0.00000419, -0.00001483, 0.00000986, -0.00000280, -0.00000736, -0.00000830, -0.00001450, 0.00001775, -0.00001325, 0.00003217, -0.00000212, 0.00001647, 0.00001454, -0.00000543, 0.00000429, -0.00002925, 0.00000684, -0.00001226, -0.00000679, -0.00001144, -0.00001470, -0.00002224, -0.00008186, -0.00017617, 0.00048268, 0.00018370, -0.00021858, -0.00000250, -0.00004296, -0.00000483, -0.00000765, 0.00000995, -0.00002546, 0.00001837, -0.00002889, 0.00001839, -0.00002434, 0.00000709, -0.00001378, 0.00000375, -0.00000833, 0.00000548, -0.00000399, -0.00001964, 0.00002553, -0.00002228, 0.00001958, -0.00000843, 0.00000519, -0.00000176, 0.00000116, 0.00000052, 0.00000250, 0.00001028, -0.00001282, -0.00000448, 0.00000661, 0.00000257, -0.00000714, 0.00000129, -0.00000014, 0.00000323, -0.00000584, 0.00000534, 0.00000159, 0.00000662, -0.00001102, 0.00000044, 0.00000056, 0.00000122, -0.00000225, -0.00000187, -0.00000110, 0.00000353, -0.00000014, 0.00000038, 0.00000007, 0.00000196, -0.00000548, 0.00000149, -0.00000100, -0.00000650, 0.00001301, -0.00000339, -0.00000342, },
		{ -0.00000688, 0.00000779, 0.00000202, -0.00000102, -0.00000440, 0.00000008, -0.00000007, 0.00000356, 0.00000974, -0.00001284, -0.00000919, 0.00000584, 0.00000368, -0.00000303, 0.00001018, -0.00000890, -0.00000241, 0.00000675, 0.00000193, -0.00000301, 0.00000296, -0.00000743, 0.00000047, -0.00000236, 0.00000053, 0.00000417, 0.00000541, -0.00000197, -0.00000412, 0.00000430, -0.00000559, 0.00000637, -0.00001048, 0.00001532, 0.00000483, -0.00001367, -0.00001552, 0.00001716, -0.00000597, -0.00000083, -0.00000394, 0.00000301, 0.00000050, -0.00000624, -0.00001250, 0.00000526, 0.00000260, -0.00001283, -0.00000147, -0.00002135, 0.00001965, -0.00001438, 0.00003098, 0.00000253, 0.00001250, 0.00001950, -0.00000659, 0.00000613, -0.00002995, 0.00000496, -0.00001046, -0.00000751, -0.00001096, -0.00001359, -0.00002256, -0.00006631, -0.00018590, 0.00040363, 0.00026374, -0.00021104, -0.00002218, -0.00003446, -0.00001202, -0.00000548, 0.00000936, -0.00002404, 0.00001565, -0.00002552, 0.00001432, -0.00001921, 0.00000137, -0.00000852, -0.00000175, -0.00000228, -0.00000234, 0.00000505, -0.00002707, 0.00002871, -0.00002220, 0.00001629, -0.00000386, 0.00000191, 0.00000088, -0.00000120, 0.00000256, -0.00000008, 0.00001236, -0.00001204, -0.00000528, 0.00000481, 0.00000498, -0.00000766, 0.00000129, -0.00000187, 0.00000511, -0.00000626, 0.00000482, 0.00000181, 0.00000665, -0.00000927, -0.00000219, 0.00000206, 0.00000015, -0.00000093, -0.00000288, -0.00000019, 0.00000206, 0.00000087, 0.00000003, 0.00000039, 0.00000143, -0.00000379, -0.00000089, 0.00000121, -0.00000816, 0.00001247, -0.00000154, -0.00000346, },
		{ -0.00000700, 0.00000567, 0.00000430, -0.00000184, -0.00000327, -0.00000148, 0.00000114, 0.00000128, 0.00001164, -0.00001136, -0.00001036, 0.00000384, 0.00000570, -0.00000438, 0.00001035, -0.00000675, -0.00000440, 0.00000681, 0.00000259, -0.00000351, 0.00000313, -0.00000624, -0.00000119, -0.00000115, -0.00000085, 0.00000496, 0.00000436, 0.00000027, -0.00000596, 0.00000507, -0.00000583, 0.00000709, -0.00001024, 0.00001164, 0.00000873, -0.00001278, -0.00001690, 0.00001430, -0.00000171, -0.00000389, -0.00000141, -0.00000063, 0.00000414, -0.00000854, -0.00000975, 0.00000029, 0.00000802, -0.00001784, 0.00000405, -0.00002565, 0.00001876, -0.00001335, 0.00002775, 0.00000858, 0.00000797, 0.00002436, -0.00000720, 0.00000752, -0.00002997, 0.00000244, -0.00000828, -0.00000866, -0.00000985, -0.00001348, -0.00002171, -0.00005462, -0.00018683, 0.00032586, 0.00032310, -0.00018340, -0.00005006, -0.00002261, -0.00002121, -0.00000190, 0.00000713, -0.00002083, 0.00001133, -0.00002078, 0.00000897, -0.00001292, -0.00000516, -0.00000265, -0.00000778, 0.00000411, -0.00000985, 0.00001262, -0.00003195, 0.00002894, -0.00001939, 0.00001101, 0.00000157, -0.00000152, 0.00000348, -0.00000345, 0.00000439, -0.00000228, 0.00001374, -0.00001073, -0.00000608, 0.00000306, 0.00000679, -0.00000748, 0.00000094, -0.00000312, 0.00000592, -0.00000542, 0.00000327, 0.00000290, 0.00000574, -0.00000665, -0.00000521, 0.00000364, -0.00000099, 0.00000044, -0.00000387, 0.00000070, 0.00000080, 0.00000149, -0.00000008, 0.00000067, 0.00000080, -0.00000216, -0.00000281, 0.00000262, -0.00000878, 0.00001100, 0.00000071, -0.00000368, },
		{ -0.00000667, 0.00000317, 0.00000646, -0.00000241, -0.00000228, -0.00000283, 0.00000211, -0.00000056, 0.00001268, -0.00000922, -0.00001151, 0.00000189, 0.00000711, -0.00000484, 0.00000952, -0.00000398, -0.00000639, 0.00000675, 0.00000318, -0.00000373, 0.00000279, -0.00000460, -0.00000297, -0.00000001, -0.00000208, 0.00000555, 0.00000352, 0.00000204, -0.00000695, 0.00000491, -0.00000528, 0.00000690, -0.00000873, 0.00000712, 0.00001243, -0.00001170, -0.00001752, 0.00001051, 0.00000283, -0.00000687, 0.00000113, -0.00000419, 0.00000734, -0.00001032, -0.00000739, -0.00000412, 0.00001241, -0.00002127, 0.00000716, -0.00002664, 0.00001480, -0.00001014, 0.00002256, 0.00001586, 0.00000306, 0.00002903, -0.00000737, 0.00000871, -0.00002971, -0.00000021, -0.00000631, -0.00000958, -0.00000889, -0.00001337, -0.00002088, -0.00004498, -0.00018273, 0.00025423, 0.00036031, -0.00013887, -0.00008177, -0.00001073, -0.00003000, 0.00000125, 0.00000475, -0.00001719, 0.00000665, -0.00001581, 0.00000342, -0.00000657, -0.00001146, 0.00000279, -0.00001327, 0.00000969, -0.00001583, 0.00001764, -0.00003355, 0.00002598, -0.00001405, 0.00000420, 0.00000732, -0.00000470, 0.00000582, -0.00000538, 0.00000585, -0.00000396, 0.00001436, -0.00000892, -0.00000690, 0.00000146, 0.00000790, -0.00000659, 0.00000023, -0.00000380, 0.00000564, -0.00000349, 0.00000096, 0.00000455, 0.00000426, -0.00000361, -0.00000816, 0.00000488, -0.00000186, 0.00000153, -0.00000457, 0.00000126, 0.00000003, 0.00000157, 0.00000014, 0.00000084, 0.00000026, -0.00000082, -0.00000406, 0.00000312, -0.00000843, 0.00000879, 0.00000313, -0.00000397, },
		{ -0.00000611, 0.00000060, 0.00000818, -0.00000251, -0.00000161, -0.00000379, 0.00000264, -0.00000178, 0.00001276, -0.00000648, -0.00001263, 0.00000013, 0.00000781, -0.00000441, 0.00000784, -0.00000085, -0.00000817, 0.00000647, 0.00000371, -0.00000366, 0.00000201, -0.00000267, -0.00000466, 0.00000089, -0.00000303, 0.00000584, 0.00000300, 0.00000325, -0.00000712, 0.00000395, -0.00000416, 0.00000607, -0.00000638, 0.00000236, 0.00001541, -0.00001012, -0.00001769, 0.00000633, 0.00000699, -0.00000918, 0.00000309, -0.00000700, 0.00000943, -0.00001095, -0.00000606, -0.00000724, 0.00001500, -0.00002234, 0.00000724, -0.00002411, 0.00000807, -0.00000521, 0.00001585, 0.00002387, -0.00000182, 0.00003329, -0.00000705, 0.00000981, -0.00002935, -0.00000269, -0.00000492, -0.00000987, -0.00000860, -0.00001259, -0.00002091, -0.00003607, -0.00017637, 0.00019161, 0.00037637, -0.00008244, -0.00011227, -0.00000218, -0.00003611, 0.00000234, 0.00000353, -0.00001430, 0.00000273, -0.00001163, -0.00000138, -0.00000108, -0.00001659, 0.00000691, -0.00001735, 0.00001353, -0.00001939, 0.00001946, -0.00003168, 0.00002008, -0.00000674, -0.00000341, 0.00001275, -0.00000721, 0.00000762, -0.00000682, 0.00000681, -0.00000502, 0.00001424, -0.00000669, -0.00000769, 0.00000011, 0.00000828, -0.00000507, -0.00000077, -0.00000391, 0.00000438, -0.00000078, -0.00000170, 0.00000634, 0.00000265, -0.00000064, -0.00001057, 0.00000541, -0.00000217, 0.00000209, -0.00000473, 0.00000131, -0.00000010, 0.00000106, 0.00000069, 0.00000085, -0.00000010, 0.00000009, -0.00000456, 0.00000276, -0.00000730, 0.00000614, 0.00000544, -0.00000413, },
		{ -0.00000550, -0.00000175, 0.00000922, -0.00000200, -0.00000137, -0.00000426, 0.00000263, -0.00000231, 0.00001195, -0.00000329, -0.00001361, -0.00000142, 0.00000779, -0.00000321, 0.00000558, 0.00000231, -0.00000952, 0.00000590, 0.00000421, -0.00000336, 0.00000091, -0.00000066, -0.00000607, 0.00000143, -0.00000362, 0.00000577, 0.00000282, 0.00000391, -0.00000656, 0.00000246, -0.00000278, 0.00000492, -0.00000366, -0.00000204, 0.00001723, -0.00000783, -0.00001768, 0.00000225, 0.00001025, -0.00001035, 0.00000404, -0.00000860, 0.00000997, -0.00001004, -0.00000615, -0.00000868, 0.00001540, -0.00002077, 0.00000428, -0.00001850, -0.00000064, 0.00000061, 0.00000836, 0.00003188, -0.00000608, 0.00003672, -0.00000607, 0.00001077, -0.00002891, -0.00000492, -0.00000423, -0.00000938, -0.00000917, -0.00001088, -0.00002213, -0.00002723, -0.00016936, 0.00013900, 0.00037424, -0.00002014, -0.00013658, 0.00000025, -0.00003792, 0.00000026, 0.00000433, -0.00001294, 0.00000030, -0.00000890, -0.00000480, 0.00000291, -0.00001995, 0.00000914, -0.00001944, 0.00001508, -0.00002008, 0.00001796, -0.00002666, 0.00001192, 0.00000166, -0.00001092, 0.00001719, -0.00000869, 0.00000871, -0.00000764, 0.00000718, -0.00000544, 0.00001342, -0.00000415, -0.00000841, -0.00000099, 0.00000799, -0.00000309, -0.00000193, -0.00000357, 0.00000239, 0.00000229, -0.00000424, 0.00000783, 0.00000133, 0.00000182, -0.00001203, 0.00000498, -0.00000175, 0.00000200, -0.00000427, 0.00000075, 0.00000041, 0.00000001, 0.00000147, 0.00000074, -0.00000023, 0.00000052, -0.00000436, 0.00000174, -0.00000567, 0.00000340, 0.00000739, -0.00000400, },
		{ -0.00000500, -0.00000367, 0.00000945, -0.00000086, -0.00000154, -0.00000423, 0.00000212, -0.00000220, 0.00001040, 0.00000009, -0.00001432, -0.00000280, 0.00000716, -0.00000149, 0.00000307, 0.00000516, -0.00001024, 0.00000497, 0.00000473, -0.00000288, -0.00000037, 0.00000122, -0.00000704, 0.00000151, -0.00000381, 0.00000533, 0.00000298, 0.00000410, -0.00000548, 0.00000072, -0.00000146, 0.00000378, -0.00000100, -0.00000560, 0.00001760, -0.00000477, -0.00001766, -0.00000138, 0.00001224, -0.00001014, 0.00000375, -0.00000877, 0.00000882, -0.00000756, -0.00000766, -0.00000841, 0.00001365, -0.00001678, -0.00000118, -0.00001076, -0.00001010, 0.00000625, 0.00000101, 0.00003907, -0.00000905, 0.00003890, -0.00000415, 0.00001143, -0.00002827, -0.00000699, -0.00000413, -0.00000826, -0.00001047, -0.00000834, -0.00002438, -0.00001845, -0.00016231, 0.00009593, 0.00035817, 0.00004188, -0.00015058, -0.00000515, -0.00003468, -0.00000534, 0.00000740, -0.00001340, -0.00000038, -0.00000786, -0.00000661, 0.00000515, -0.00002130, 0.00000929, -0.00001939, 0.00001422, -0.00001797, 0.00001354, -0.00001925, 0.00000253, 0.00001009, -0.00001743, 0.00002005, -0.00000887, 0.00000896, -0.00000779, 0.00000693, -0.00000523, 0.00001201, -0.00000147, -0.00000899, -0.00000185, 0.00000713, -0.00000086, -0.00000306, -0.00000295, 0.00000002, 0.00000526, -0.00000621, 0.00000863, 0.00000065, 0.00000343, -0.00001232, 0.00000350, -0.00000060, 0.00000126, -0.00000322, -0.00000035, 0.00000147, -0.00000139, 0.00000232, 0.00000059, -0.00000017, 0.00000054, -0.00000362, 0.00000032, -0.00000390, 0.00000092, 0.00000872, -0.00000345, },
		{ -0.00000470, -0.00000503, 0.00000882, 0.00000082, -0.00000203, -0.00000382, 0.00000121, -0.00000163, 0.00000839, 0.00000336, -0.00001459, -0.00000409, 0.00000609, 0.00000046, 0.00000068, 0.00000741, -0.00001019, 0.00000365, 0.00000526, -0.00000230, -0.00000169, 0.00000281, -0.00000744, 0.00000112, -0.00000364, 0.00000460, 0.00000337, 0.00000395, -0.00000412, -0.00000094, -0.00000049, 0.00000293, 0.00000125, -0.00000801, 0.00001648, -0.00000104, -0.00001764, -0.00000435, 0.00001281, -0.00000855, 0.00000229, -0.00000758, 0.00000613, -0.00000376, -0.00001031, -0.00000676, 0.00001019, -0.00001104, -0.00000818, -0.00000220, -0.00001893, 0.00001060, -0.00000534, 0.00004463, -0.00001013, 0.00003944, -0.00000107, 0.00001161, -0.00002721, -0.00000910, -0.00000433, -0.00000684, -0.00001213, -0.00000540, -0.00002717, -0.00001014, -0.00015508, 0.00006089, 0.00033293, 0.00009809, -0.00015155, -0.00001871, -0.00002678, -0.00001399, 0.00001231, -0.00001537, 0.00000047, -0.00000829, -0.00000698, 0.00000575, -0.00002077, 0.00000755, -0.00001742, 0.00001123, -0.00001354, 0.00000702, -0.00001058, -0.00000689, 0.00001744, -0.00002213, 0.00002089, -0.00000763, 0.00000840, -0.00000728, 0.00000611, -0.00000450, 0.00001018, 0.00000119, -0.00000934, -0.00000252, 0.00000586, 0.00000139, -0.00000397, -0.00000226, -0.00000240, 0.00000768, -0.00000725, 0.00000850, 0.00000082, 0.00000404, -0.00001139, 0.00000106, 0.00000114, 0.00000003, -0.00000172, -0.00000182, 0.00000285, -0.00000290, 0.00000303, 0.00000053, 0.00000000, 0.00000025, -0.00000255, -0.00000119, -0.00000231, -0.00000104, 0.00000929, -0.00000243, },
		{ -0.00000463, -0.00000580, 0.00000745, 0.00000285, -0.00000265, -0.00000318, 0.00000009, -0.00000082, 0.00000623, 0.00000623, -0.00001426, -0.00000537, 0.00000476, 0.00000235, -0.00000130, 0.00000884, -0.00000933, 0.00000201, 0.00000580, -0.00000166, -0.00000292, 0.00000396, -0.00000723, 0.00000029, -0.00000320, 0.00000366, 0.00000388, 0.00000365, -0.00000273, -0.00000227, -0.00000005, 0.00000251, 0.00000292, -0.00000917, 0.00001401, 0.00000307, -0.00001754, -0.00000662, 0.00001204, -0.00000581, -0.00000008, -0.00000533, 0.00000233, 0.00000082, -0.00001353, -0.00000433, 0.00000577, -0.00000449, -0.00001551, 0.00000577, -0.00002582, 0.00001274, -0.00001000, 0.00004795, -0.00000890, 0.00003810, 0.00000328, 0.00001119, -0.00002555, -0.00001151, -0.00000454, -0.00000551, -0.00001370, -0.00000257, -0.00002982, -0.00000300, -0.00014709, 0.00003188, 0.00030311, 0.00014425, -0.00013856, -0.00003928, -0.00001559, -0.00002447, 0.00001811, -0.00001812, 0.00000222, -0.00000962, -0.00000639, 0.00000515, -0.00001881, 0.00000441, -0.00001405, 0.00000673, -0.00000759, -0.00000053, -0.00000189, -0.00001515, 0.00002282, -0.00002444, 0.00001957, -0.00000509, 0.00000715, -0.00000624, 0.00000485, -0.00000339, 0.00000811, 0.00000367, -0.00000940, -0.00000304, 0.00000435, 0.00000342, -0.00000450, -0.00000169, -0.00000455, 0.00000924, -0.00000715, 0.00000735, 0.00000185, 0.00000368, -0.00000940, -0.00000207, 0.00000319, -0.00000144, -0.00000003, -0.00000340, 0.00000427, -0.00000422, 0.00000340, 0.00000068, 0.00000019, -0.00000018, -0.00000141, -0.00000251, -0.00000116, -0.00000229, 0.00000905, -0.00000098, },
		{ -0.00000471, -0.00000607, 0.00000554, 0.00000496, -0.00000317, -0.00000253, -0.00000102, -0.00000002, 0.00000420, 0.00000843, -0.00001323, -0.00000670, 0.00000337, 0.00000392, -0.00000265, 0.00000936, -0.00000772, 0.00000015, 0.00000627, -0.00000098, -0.00000400, 0.00000462, -0.00000644, -0.00000085, -0.00000260, 0.00000266, 0.00000438, 0.00000334, -0.00000149, -0.00000308, -0.00000024, 0.00000255, 0.00000397, -0.00000917, 0.00001053, 0.00000718, -0.00001717, -0.00000826, 0.00001016, -0.00000232, -0.00000292, -0.00000251, -0.00000198, 0.00000550, -0.00001663, -0.00000186, 0.00000123, 0.00000183, -0.00002195, 0.00001188, -0.00002974, 0.00001212, -0.00001260, 0.00004872, -0.00000523, 0.00003489, 0.00000885, 0.00001016, -0.00002320, -0.00001436, -0.00000448, -0.00000463, -0.00001475, -0.00000040, -0.00003166, 0.00000228, -0.00013766, 0.00000695, 0.00027254, 0.00017779, -0.00011251, -0.00006441, -0.00000328, -0.00003508, 0.00002349, -0.00002063, 0.00000401, -0.00001111, -0.00000551, 0.00000395, -0.00001604, 0.00000053, -0.00000999, 0.00000149, -0.00000112, -0.00000795, 0.00000565, -0.00002128, 0.00002557, -0.00002414, 0.00001619, -0.00000153, 0.00000549, -0.00000484, 0.00000330, -0.00000206, 0.00000599, 0.00000581, -0.00000912, -0.00000348, 0.00000277, 0.00000505, -0.00000452, -0.00000137, -0.00000620, 0.00000973, -0.00000588, 0.00000528, 0.00000360, 0.00000254, -0.00000665, -0.00000550, 0.00000516, -0.00000283, 0.00000155, -0.00000481, 0.00000545, -0.00000509, 0.00000324, 0.00000112, 0.00000030, -0.00000060, -0.00000042, -0.00000341, -0.00000063, -0.00000277, 0.00000809, 0.00000079, },
		{ -0.00000481, -0.00000597, 0.00000333, 0.00000686, -0.00000336, -0.00000205, -0.00000191, 0.00000056, 0.00000256, 0.00000980, -0.00001150, -0.00000808, 0.00000202, 0.00000503, -0.00000327, 0.00000900, -0.00000553, -0.00000176, 0.00000661, -0.00000027, -0.00000486, 0.00000477, -0.00000517, -0.00000215, -0.00000200, 0.00000172, 0.00000474, 0.00000314, -0.00000055, -0.00000332, -0.00000100, 0.00000293, 0.00000452, -0.00000825, 0.00000650, 0.00001085, -0.00001631, -0.00000946, 0.00000754, 0.00000141, -0.00000570, 0.00000034, -0.00000616, 0.00000957, -0.00001890, -0.00000004, -0.00000261, 0.00000697, -0.00002648, 0.00001519, -0.00003016, 0.00000866, -0.00001319, 0.00004695, 0.00000070, 0.00003011, 0.00001535, 0.00000872, -0.00002019, -0.00001770, -0.00000402, -0.00000442, -0.00001498, 0.00000074, -0.00003218, 0.00000520, -0.00012630, -0.00001550, 0.00024396, 0.00019794, -0.00007591, -0.00009069, 0.00000757, -0.00004398, 0.00002709, -0.00002187, 0.00000495, -0.00001197, -0.00000501, 0.00000278, -0.00001312, -0.00000337, -0.00000596, -0.00000366, 0.00000495, -0.00001419, 0.00001110, -0.00002465, 0.00002547, -0.00002135, 0.00001118, 0.00000257, 0.00000374, -0.00000332, 0.00000165, -0.00000069, 0.00000399, 0.00000751, -0.00000847, -0.00000387, 0.00000125, 0.00000615, -0.00000399, -0.00000137, -0.00000723, 0.00000914, -0.00000359, 0.00000254, 0.00000577, 0.00000095, -0.00000355, -0.00000876, 0.00000668, -0.00000382, 0.00000274, -0.00000578, 0.00000612, -0.00000531, 0.00000246, 0.00000188, 0.00000031, -0.00000090, 0.00000027, -0.00000376, -0.00000073, -0.00000256, 0.00000659, 0.00000266, },
		{ -0.00000478, -0.00000567, 0.00000112, 0.00000828, -0.00000306, -0.00000189, -0.00000246, 0.00000076, 0.00000144, 0.00001030, -0.00000916, -0.00000946, 0.00000077, 0.00000559, -0.00000318, 0.00000792, -0.00000301, -0.00000350, 0.00000672, 0.00000050, -0.00000549, 0.00000444, -0.00000355, -0.00000344, -0.00000153, 0.00000095, 0.00000489, 0.00000314, 0.00000005, -0.00000301, -0.00000216, 0.00000345, 0.00000478, -0.00000676, 0.00000243, 0.00001366, -0.00001479, -0.00001037, 0.00000455, 0.00000487, -0.00000794, 0.00000269, -0.00000960, 0.00001242, -0.00001981, 0.00000056, -0.00000514, 0.00001026, -0.00002846, 0.00001527, -0.00002708, 0.00000278, -0.00001219, 0.00004300, 0.00000842, 0.00002429, 0.00002231, 0.00000717, -0.00001671, -0.00002141, -0.00000319, -0.00000494, -0.00001431, 0.00000069, -0.00003116, 0.00000559, -0.00011290, -0.00003642, 0.00021883, 0.00020560, -0.00003254, -0.00011435, 0.00001448, -0.00004954, 0.00002777, -0.00002098, 0.00000436, -0.00001160, -0.00000543, 0.00000215, -0.00001062, -0.00000671, -0.00000258, -0.00000804, 0.00000981, -0.00001848, 0.00001388, -0.00002506, 0.00002268, -0.00001654, 0.00000516, 0.00000664, 0.00000227, -0.00000188, 0.00000007, 0.00000054, 0.00000226, 0.00000870, -0.00000748, -0.00000423, -0.00000011, 0.00000664, -0.00000293, -0.00000166, -0.00000763, 0.00000757, -0.00000058, -0.00000048, 0.00000798, -0.00000068, -0.00000055, -0.00001143, 0.00000741, -0.00000417, 0.00000333, -0.00000614, 0.00000615, -0.00000480, 0.00000108, 0.00000289, 0.00000021, -0.00000102, 0.00000057, -0.00000354, -0.00000141, -0.00000184, 0.00000481, 0.00000442, },
		{ -0.00000448, -0.00000537, -0.00000085, 0.00000901, -0.00000216, -0.00000209, -0.00000260, 0.00000054, 0.00000089, 0.00000999, -0.00000638, -0.00001070, -0.00000037, 0.00000563, -0.00000250, 0.00000634, -0.00000043, -0.00000489, 0.00000652, 0.00000132, -0.00000591, 0.00000372, -0.00000175, -0.00000455, -0.00000131, 0.00000040, 0.00000479, 0.00000333, 0.00000033, -0.00000230, -0.00000349, 0.00000385, 0.00000501, -0.00000504, -0.00000123, 0.00001531, -0.00001251, -0.00001115, 0.00000156, 0.00000764, -0.00000926, 0.00000418, -0.00001188, 0.00001367, -0.00001904, -0.00000034, -0.00000602, 0.00001137, -0.00002769, 0.00001224, -0.00002107, -0.00000464, -0.00001032, 0.00003746, 0.00001722, 0.00001814, 0.00002912, 0.00000596, -0.00001305, -0.00002525, -0.00000215, -0.00000604, -0.00001283, -0.00000047, -0.00002867, 0.00000363, -0.00009774, -0.00005613, 0.00019748, 0.00020295, 0.00001322, -0.00013175, 0.00001548, -0.00005064, 0.00002487, -0.00001756, 0.00000188, -0.00000969, -0.00000701, 0.00000234, -0.00000888, -0.00000908, -0.00000027, -0.00001119, 0.00001296, -0.00002039, 0.00001385, -0.00002274, 0.00001776, -0.00001045, -0.00000110, 0.00001010, 0.00000140, -0.00000072, -0.00000128, 0.00000153, 0.00000087, 0.00000936, -0.00000619, -0.00000455, -0.00000124, 0.00000653, -0.00000145, -0.00000217, -0.00000748, 0.00000526, 0.00000275, -0.00000336, 0.00000985, -0.00000199, 0.00000194, -0.00001315, 0.00000712, -0.00000375, 0.00000324, -0.00000582, 0.00000550, -0.00000358, -0.00000077, 0.00000403, 0.00000007, -0.00000095, 0.00000050, -0.00000284, -0.00000247, -0.00000086, 0.00000302, 0.00000582, },
		{ -0.00000381, -0.00000518, -0.00000240, 0.00000896, -0.00000068, -0.00000262, -0.00000237, -0.00000006, 0.00000082, 0.00000906, -0.00000342, -0.00001164, -0.00000147, 0.00000525, -0.00000143, 0.00000455, 0.00000193, -0.00000576, 0.00000596, 0.00000221, -0.00000612, 0.00000271, 0.00000005, -0.00000534, -0.00000140, 0.00000010, 0.00000448, 0.00000367, 0.00000040, -0.00000136, -0.00000471, 0.00000387, 0.00000541, -0.00000342, -0.00000413, 0.00001564, -0.00000952, -0.00001184, -0.00000118, 0.00000941, -0.00000944, 0.00000459, -0.00001279, 0.00001316, -0.00001657, -0.00000270, -0.00000530, 0.00001039, -0.00002445, 0.00000669, -0.00001315, -0.00001248, -0.00000845, 0.00003104, 0.00002628, 0.00001245, 0.00003514, 0.00000555, -0.00000956, -0.00002891, -0.00000118, -0.00000747, -0.00001085, -0.00000244, -0.00002503, -0.00000018, -0.00008150, -0.00007437, 0.00017931, 0.00019300, 0.00005695, -0.00014000, 0.00000949, -0.00004697, 0.00001836, -0.00001167, -0.00000242, -0.00000632, -0.00000970, 0.00000334, -0.00000799, -0.00001036, 0.00000081, -0.00001293, 0.00001422, -0.00001991, 0.00001132, -0.00001831, 0.00001153, -0.00000396, -0.00000682, 0.00001244, 0.00000135, 0.00000007, -0.00000234, 0.00000222, -0.00000014, 0.00000955, -0.00000472, -0.00000479, -0.00000213, 0.00000589, 0.00000029, -0.00000274, -0.00000696, 0.00000253, 0.00000599, -0.00000569, 0.00001102, -0.00000266, 0.00000362, -0.00001370, 0.00000576, -0.00000258, 0.00000250, -0.00000488, 0.00000424, -0.00000181, -0.00000290, 0.00000511, -0.00000001, -0.00000076, 0.00000012, -0.00000182, -0.00000368, 0.00000010, 0.00000150, 0.00000670, },
		{ -0.00000274, -0.00000519, -0.00000343, 0.00000813, 0.00000125, -0.00000336, -0.00000188, -0.00000093, 0.00000108, 0.00000775, -0.00000058, -0.00001211, -0.00000260, 0.00000458, -0.00000020, 0.00000282, 0.00000382, -0.00000600, 0.00000505, 0.00000315, -0.00000615, 0.00000149, 0.00000172, -0.00000571, -0.00000179, -0.00000001, 0.00000401, 0.00000407, 0.00000038, -0.00000043, -0.00000557, 0.00000333, 0.00000614, -0.00000209, -0.00000609, 0.00001469, -0.00000601, -0.00001241, -0.00000350, 0.00001007, -0.00000847, 0.00000395, -0.00001236, 0.00001104, -0.00001265, -0.00000623, -0.00000330, 0.00000776, -0.00001941, -0.00000041, -0.00000456, -0.00001948, -0.00000742, 0.00002445, 0.00003479, 0.00000796, 0.00003979, 0.00000634, -0.00000655, -0.00003205, -0.00000060, -0.00000889, -0.00000872, -0.00000480, -0.00002074, -0.00000512, -0.00006508, -0.00009047, 0.00016314, 0.00017903, 0.00009479, -0.00013733, -0.00000346, -0.00003904, 0.00000885, -0.00000389, -0.00000804, -0.00000191, -0.00001314, 0.00000490, -0.00000779, -0.00001065, 0.00000079, -0.00001334, 0.00001373, -0.00001735, 0.00000689, -0.00001264, 0.00000495, 0.00000205, -0.00001130, 0.00001331, 0.00000219, 0.00000050, -0.00000310, 0.00000259, -0.00000080, 0.00000934, -0.00000316, -0.00000490, -0.00000280, 0.00000484, 0.00000210, -0.00000322, -0.00000624, -0.00000030, 0.00000873, -0.00000714, 0.00001128, -0.00000253, 0.00000436, -0.00001304, 0.00000343, -0.00000082, 0.00000127, -0.00000347, 0.00000258, 0.00000027, -0.00000503, 0.00000592, 0.00000007, -0.00000051, -0.00000043, -0.00000069, -0.00000480, 0.00000081, 0.00000044, 0.00000694, },
		{ -0.00000133, -0.00000538, -0.00000394, 0.00000665, 0.00000343, -0.00000409, -0.00000130, -0.00000187, 0.00000146, 0.00000634, 0.00000188, -0.00001199, -0.00000381, 0.00000375, 0.00000098, 0.00000138, 0.00000510, -0.00000560, 0.00000385, 0.00000407, -0.00000599, 0.00000015, 0.00000313, -0.00000562, -0.00000246, -0.00000000, 0.00000350, 0.00000440, 0.00000042, 0.00000032, -0.00000589, 0.00000214, 0.00000720, -0.00000110, -0.00000710, 0.00001267, -0.00000227, -0.00001273, -0.00000537, 0.00000965, -0.00000652, 0.00000246, -0.00001083, 0.00000767, -0.00000779, -0.00001037, -0.00000057, 0.00000416, -0.00001345, -0.00000790, 0.00000338, -0.00002457, -0.00000785, 0.00001821, 0.00004209, 0.00000525, 0.00004266, 0.00000860, -0.00000422, -0.00003444, -0.00000068, -0.00000999, -0.00000682, -0.00000712, -0.00001637, -0.00001040, -0.00004947, -0.00010364, 0.00014759, 0.00016410, 0.00012392, -0.00012333, -0.00002220, -0.00002818, -0.00000248, 0.00000479, -0.00001418, 0.00000286, -0.00001675, 0.00000655, -0.00000795, -0.00001027, -0.00000004, -0.00001273, 0.00001186, -0.00001330, 0.00000142, -0.00000672, -0.00000105, 0.00000681, -0.00001410, 0.00001259, 0.00000383, 0.00000067, -0.00000361, 0.00000270, -0.00000117, 0.00000886, -0.00000163, -0.00000486, -0.00000328, 0.00000352, 0.00000378, -0.00000344, -0.00000552, -0.00000292, 0.00001067, -0.00000752, 0.00001056, -0.00000156, 0.00000416, -0.00001132, 0.00000038, 0.00000125, -0.00000019, -0.00000186, 0.00000077, 0.00000238, -0.00000687, 0.00000626, 0.00000043, -0.00000028, -0.00000103, 0.00000037, -0.00000561, 0.00000107, -0.00000003, 0.00000655, },
		{ 0.00000032, -0.00000568, -0.00000402, 0.00000473, 0.00000558, -0.00000461, -0.00000081, -0.00000270, 0.00000174, 0.00000508, 0.00000374, -0.00001120, -0.00000513, 0.00000287, 0.00000193, 0.00000037, 0.00000570, -0.00000464, 0.00000248, 0.00000490, -0.00000564, -0.00000125, 0.00000422, -0.00000509, -0.00000328, 0.00000001, 0.00000303, 0.00000458, 0.00000062, 0.00000075, -0.00000557, 0.00000034, 0.00000851, -0.00000039, -0.00000731, 0.00000994, 0.00000132, -0.00001266, -0.00000684, 0.00000833, -0.00000391, 0.00000045, -0.00000858, 0.00000356, -0.00000260, -0.00001450, 0.00000220, 0.00000035, -0.00000752, -0.00001466, 0.00000953, -0.00002695, -0.00001005, 0.00001259, 0.00004771, 0.00000468, 0.00004356, 0.00001237, -0.00000262, -0.00003593, -0.00000158, -0.00001054, -0.00000546, -0.00000900, -0.00001242, -0.00001530, -0.00003555, -0.00011316, 0.00013142, 0.00015062, 0.00014278, -0.00009903, -0.00004459, -0.00001626, -0.00001412, 0.00001315, -0.00001987, 0.00000717, -0.00001984, 0.00000778, -0.00000803, -0.00000961, -0.00000126, -0.00001153, 0.00000913, -0.00000849, -0.00000419, -0.00000148, -0.00000571, 0.00000984, -0.00001499, 0.00001040, 0.00000600, 0.00000078, -0.00000398, 0.00000263, -0.00000136, 0.00000822, -0.00000024, -0.00000462, -0.00000359, 0.00000205, 0.00000516, -0.00000331, -0.00000493, -0.00000511, 0.00001163, -0.00000681, 0.00000892, 0.00000011, 0.00000320, -0.00000880, -0.00000300, 0.00000328, -0.00000160, -0.00000031, -0.00000092, 0.00000423, -0.00000819, 0.00000600, 0.00000111, -0.00000013, -0.00000154, 0.00000117, -0.00000596, 0.00000081, 0.00000008, 0.00000562, },
		{ 0.00000200, -0.00000597, -0.00000382, 0.00000263, 0.00000743, -0.00000471, -0.00000059, -0.00000325, 0.00000176, 0.00000417, 0.00000488, -0.00000977, -0.00000651, 0.00000201, 0.00000257, -0.00000015, 0.00000570, -0.00000327, 0.00000109, 0.00000556, -0.00000508, -0.00000264, 0.00000494, -0.00000420, -0.00000415, -0.00000008, 0.00000271, 0.00000452, 0.00000106, 0.00000081, -0.00000466, -0.00000190, 0.00000984, 0.00000023, -0.00000698, 0.00000692, 0.00000437, -0.00001203, -0.00000799, 0.00000634, -0.00000100, -0.00000166, -0.00000606, -0.00000071, 0.00000227, -0.00001798, 0.00000440, -0.00000293, -0.00000249, -0.00001979, 0.00001313, -0.00002631, -0.00001391, 0.00000761, 0.00005144, 0.00000633, 0.00004258, 0.00001749, -0.00000163, -0.00003656, -0.00000335, -0.00001045, -0.00000481, -0.00001020, -0.00000926, -0.00001926, -0.00002398, -0.00011856, 0.00011378, 0.00014017, 0.00015117, -0.00006664, -0.00006782, -0.00000542, -0.00002451, 0.00001999, -0.00002416, 0.00001026, -0.00002179, 0.00000810, -0.00000764, -0.00000908, -0.00000246, -0.00001018, 0.00000608, -0.00000362, -0.00000913, 0.00000237, -0.00000853, 0.00001093, -0.00001409, 0.00000706, 0.00000832, 0.00000106, -0.00000434, 0.00000248, -0.00000144, 0.00000752, 0.00000094, -0.00000419, -0.00000378, 0.00000057, 0.00000612, -0.00000276, -0.00000455, -0.00000675, 0.00001156, -0.00000513, 0.00000660, 0.00000222, 0.00000177, -0.00000587, -0.00000631, 0.00000490, -0.00000267, 0.00000093, -0.00000226, 0.00000560, -0.00000880, 0.00000507, 0.00000208, -0.00000008, -0.00000189, 0.00000162, -0.00000579, 0.00000003, 0.00000067, 0.00000432, },
		{ 0.00000354, -0.00000612, -0.00000351, 0.00000059, 0.00000873, -0.00000425, -0.00000073, -0.00000344, 0.00000142, 0.00000370, 0.00000529, -0.00000781, -0.00000785, 0.00000116, 0.00000288, -0.00000022, 0.00000520, -0.00000169, -0.00000016, 0.00000596, -0.00000429, -0.00000400, 0.00000531, -0.00000303, -0.00000493, -0.00000037, 0.00000259, 0.00000420, 0.00000174, 0.00000052, -0.00000328, -0.00000433, 0.00001091, 0.00000098, -0.00000641, 0.00000401, 0.00000659, -0.00001076, -0.00000889, 0.00000395, 0.00000181, -0.00000350, -0.00000369, -0.00000463, 0.00000626, -0.00002031, 0.00000554, -0.00000515, 0.00000105, -0.00002272, 0.00001386, -0.00002280, -0.00001893, 0.00000302, 0.00005328, 0.00001004, 0.00004004, 0.00002358, -0.00000094, -0.00003646, -0.00000587, -0.00000975, -0.00000489, -0.00001063, -0.00000707, -0.00002197, -0.00001510, -0.00011974, 0.00009431, 0.00013339, 0.00015013, -0.00002927, -0.00008881, 0.00000231, -0.00003226, 0.00002432, -0.00002634, 0.00001155, -0.00002217, 0.00000717, -0.00000650, -0.00000899, -0.00000328, -0.00000905, 0.00000317, 0.00000070, -0.00001276, 0.00000438, -0.00000939, 0.00001022, -0.00001175, 0.00000310, 0.00001036, 0.00000176, -0.00000479, 0.00000231, -0.00000150, 0.00000683, 0.00000189, -0.00000360, -0.00000383, -0.00000084, 0.00000659, -0.00000181, -0.00000439, -0.00000780, 0.00001057, -0.00000272, 0.00000393, 0.00000446, 0.00000020, -0.00000293, -0.00000912, 0.00000583, -0.00000318, 0.00000167, -0.00000307, 0.00000636, -0.00000864, 0.00000353, 0.00000328, -0.00000010, -0.00000205, 0.00000169, -0.00000514, -0.00000115, 0.00000157, 0.00000288, },
		{ 0.00000474, -0.00000600, -0.00000324, -0.00000114, 0.00000931, -0.00000316, -0.00000123, -0.00000324, 0.00000070, 0.00000367, 0.00000506, -0.00000552, -0.00000900, 0.00000029, 0.00000289, 0.00000004, 0.00000441, -0.00000011, -0.00000110, 0.00000603, -0.00000329, -0.00000528, 0.00000532, -0.00000169, -0.00000550, -0.00000091, 0.00000268, 0.00000365, 0.00000259, -0.00000001, -0.00000164, -0.00000662, 0.00001143, 0.00000209, -0.00000588, 0.00000156, 0.00000777, -0.00000886, -0.00000958, 0.00000142, 0.00000419, -0.00000477, -0.00000179, -0.00000781, 0.00000901, -0.00002117, 0.00000536, -0.00000603, 0.00000281, -0.00002330, 0.00001187, -0.00001705, -0.00002432, -0.00000159, 0.00005341, 0.00001543, 0.00003642, 0.00003015, -0.00000015, -0.00003592, -0.00000893, -0.00000863, -0.00000557, -0.00001036, -0.00000583, -0.00002335, -0.00000894, -0.00011699, 0.00007319, 0.00013004, 0.00014160, 0.00000953, -0.00010469, 0.00000539, -0.00003649, 0.00002559, -0.00002603, 0.00001078, -0.00002080, 0.00000493, -0.00000454, -0.00000946, -0.00000356, -0.00000836, 0.00000074, 0.00000406, -0.00001475, 0.00000446, -0.00000846, 0.00000810, -0.00000853, -0.00000090, 0.00001168, 0.00000305, -0.00000538, 0.00000218, -0.00000158, 0.00000620, 0.00000260, -0.00000290, -0.00000375, -0.00000211, 0.00000657, -0.00000053, -0.00000439, -0.00000833, 0.00000885, 0.00000008, 0.00000126, 0.00000648, -0.00000118, -0.00000033, -0.00001112, 0.00000586, -0.00000304, 0.00000185, -0.00000331, 0.00000647, -0.00000775, 0.00000151, 0.00000454, -0.00000011, -0.00000204, 0.00000141, -0.00000411, -0.00000255, 0.00000256, 0.00000153, },
		{ 0.00000547, -0.00000554, -0.00000313, -0.00000242, 0.00000911, -0.00000150, -0.00000204, -0.00000272, -0.00000032, 0.00000397, 0.00000439, -0.00000316, -0.00000979, -0.00000067, 0.00000269, 0.00000048, 0.00000354, 0.00000126, -0.00000162, 0.00000575, -0.00000210, -0.00000644, 0.00000502, -0.00000031, -0.00000577, -0.00000171, 0.00000293, 0.00000295, 0.00000350, -0.00000061, -0.00000000, -0.00000847, 0.00001115, 0.00000371, -0.00000558, -0.00000022, 0.00000789, -0.00000647, -0.00001003, -0.00000109, 0.00000593, -0.00000531, -0.00000053, -0.00001002, 0.00001033, -0.00002053, 0.00000385, -0.00000557, 0.00000285, -0.00002175, 0.00000773, -0.00001000, -0.00002916, -0.00000664, 0.00005205, 0.00002204, 0.00003235, 0.00003662, 0.00000116, -0.00003521, -0.00001227, -0.00000731, -0.00000665, -0.00000960, -0.00000538, -0.00002353, -0.00000523, -0.00011091, 0.00005103, 0.00012928, 0.00012811, 0.00004630, -0.00011321, 0.00000296, -0.00003690, 0.00002371, -0.00002328, 0.00000806, -0.00001783, 0.00000153, -0.00000188, -0.00001046, -0.00000328, -0.00000817, -0.00000109, 0.00000627, -0.00001505, 0.00000281, -0.00000622, 0.00000519, -0.00000507, -0.00000436, 0.00001197, 0.00000498, -0.00000609, 0.00000205, -0.00000165, 0.00000561, 0.00000313, -0.00000217, -0.00000351, -0.00000321, 0.00000609, 0.00000095, -0.00000443, -0.00000846, 0.00000666, 0.00000290, -0.00000105, 0.00000799, -0.00000210, 0.00000164, -0.00001210, 0.00000494, -0.00000227, 0.00000150, -0.00000302, 0.00000600, -0.00000630, -0.00000077, 0.00000568, -0.00000001, -0.00000190, 0.00000087, -0.00000288, -0.00000395, 0.00000341, 0.00000048, },
		{ 0.00000567, -0.00000472, -0.00000324, -0.00000318, 0.00000816, 0.00000057, -0.00000299, -0.00000201, -0.00000150, 0.00000443, 0.00000351, -0.00000100, -0.00001007, -0.00000178, 0.00000237, 0.00000094, 0.00000275, 0.00000228, -0.00000169, 0.00000517, -0.00000079, -0.00000744, 0.00000443, 0.00000104, -0.00000569, -0.00000273, 0.00000326, 0.00000221, 0.00000433, -0.00000111, 0.00000139, -0.00000961, 0.00000993, 0.00000589, -0.00000556, -0.00000130, 0.00000708, -0.00000383, -0.00001016, -0.00000346, 0.00000693, -0.00000510, 0.00000006, -0.00001125, 0.00001029, -0.00001856, 0.00000127, -0.00000403, 0.00000151, -0.00001863, 0.00000228, -0.00000275, -0.00003255, -0.00001246, 0.00004941, 0.00002934, 0.00002842, 0.00004243, 0.00000340, -0.00003460, -0.00001562, -0.00000606, -0.00000787, -0.00000858, -0.00000548, -0.00002280, -0.00000353, -0.00010234, 0.00002874, 0.00012986, 0.00011231, 0.00007803, -0.00011301, -0.00000493, -0.00003390, 0.00001915, -0.00001855, 0.00000380, -0.00001368, -0.00000262, 0.00000117, -0.00001177, -0.00000259, -0.00000837, -0.00000235, 0.00000739, -0.00001386, -0.00000008, -0.00000331, 0.00000216, -0.00000200, -0.00000682, 0.00001103, 0.00000751, -0.00000680, 0.00000185, -0.00000169, 0.00000502, 0.00000356, -0.00000150, -0.00000309, -0.00000414, 0.00000524, 0.00000249, -0.00000441, -0.00000835, 0.00000429, 0.00000542, -0.00000274, 0.00000880, -0.00000239, 0.00000284, -0.00001203, 0.00000316, -0.00000101, 0.00000076, -0.00000235, 0.00000513, -0.00000449, -0.00000305, 0.00000650, 0.00000028, -0.00000170, 0.00000019, -0.00000162, -0.00000516, 0.00000393, -0.00000013, },
		{ 0.00000536, -0.00000360, -0.00000355, -0.00000344, 0.00000660, 0.00000285, -0.00000388, -0.00000129, -0.00000266, 0.00000486, 0.00000267, 0.00000073, -0.00000974, -0.00000305, 0.00000202, 0.00000131, 0.00000218, 0.00000287, -0.00000133, 0.00000434, 0.00000057, -0.00000821, 0.00000357, 0.00000229, -0.00000528, -0.00000388, 0.00000355, 0.00000157, 0.00000495, -0.00000134, 0.00000233, -0.00000988, 0.00000775, 0.00000851, -0.00000575, -0.00000177, 0.00000563, -0.00000126, -0.00000983, -0.00000566, 0.00000718, -0.00000425, 0.00000009, -0.00001161, 0.00000912, -0.00001562, -0.00000197, -0.00000183, -0.00000069, -0.00001466, -0.00000351, 0.00000365, -0.00003379, -0.00001912, 0.00004558, 0.00003685, 0.00002518, 0.00004715, 0.00000684, -0.00003425, -0.00001879, -0.00000510, -0.00000901, -0.00000757, -0.00000584, -0.00002149, -0.00000329, -0.00009219, 0.00000731, 0.00013042, 0.00009663, 0.00010254, -0.00010383, -0.00001736, -0.00002847, 0.00001277, -0.00001258, -0.00000132, -0.00000896, -0.00000700, 0.00000418, -0.00001309, -0.00000177, -0.00000873, -0.00000321, 0.00000760, -0.00001158, -0.00000361, -0.00000043, -0.00000036, 0.00000019, -0.00000798, 0.00000888, 0.00001042, -0.00000732, 0.00000149, -0.00000161, 0.00000436, 0.00000395, -0.00000096, -0.00000249, -0.00000488, 0.00000412, 0.00000392, -0.00000419, -0.00000815, 0.00000197, 0.00000738, -0.00000363, 0.00000884, -0.00000204, 0.00000326, -0.00001100, 0.00000076, 0.00000048, -0.00000016, -0.00000150, 0.00000408, -0.00000259, -0.00000506, 0.00000683, 0.00000083, -0.00000150, -0.00000051, -0.00000050, -0.00000602, 0.00000401, -0.00000024, },
		{ 0.00000465, -0.00000230, -0.00000400, -0.00000328, 0.00000465, 0.00000506, -0.00000450, -0.00000073, -0.00000363, 0.00000509, 0.00000208, 0.00000187, -0.00000877, -0.00000446, 0.00000166, 0.00000150, 0.00000190, 0.00000305, -0.00000064, 0.00000339, 0.00000186, -0.00000871, 0.00000247, 0.00000337, -0.00000456, -0.00000505, 0.00000368, 0.00000113, 0.00000522, -0.00000119, 0.00000271, -0.00000924, 0.00000476, 0.00001134, -0.00000594, -0.00000188, 0.00000388, 0.00000091, -0.00000895, -0.00000768, 0.00000678, -0.00000294, -0.00000021, -0.00001134, 0.00000717, -0.00001216, -0.00000540, 0.00000051, -0.00000315, -0.00001059, -0.00000876, 0.00000838, -0.00003254, -0.00002644, 0.00004055, 0.00004416, 0.00002307, 0.00005049, 0.00001161, -0.00003422, -0.00002167, -0.00000454, -0.00000990, -0.00000676, -0.00000623, -0.00001991, -0.00000398, -0.00008134, -0.00001233, 0.00012976, 0.00008302, 0.00011865, -0.00008646, -0.00003270, -0.00002200, 0.00000567, -0.00000631, -0.00000652, -0.00000435, -0.00001101, 0.00000670, -0.00001406, -0.00000109, -0.00000899, -0.00000390, 0.00000720, -0.00000868, -0.00000712, 0.00000180, -0.00000189, 0.00000122, -0.00000780, 0.00000575, 0.00001338, -0.00000743, 0.00000084, -0.00000134, 0.00000359, 0.00000438, -0.00000060, -0.00000171, -0.00000545, 0.00000282, 0.00000511, -0.00000370, -0.00000796, -0.00000011, 0.00000863, -0.00000369, 0.00000815, -0.00000111, 0.00000303, -0.00000924, -0.00000195, 0.00000191, -0.00000102, -0.00000068, 0.00000305, -0.00000083, -0.00000661, 0.00000658, 0.00000166, -0.00000133, -0.00000115, 0.00000035, -0.00000644, 0.00000362, 0.00000011, },
		{ 0.00000368, -0.00000095, -0.00000448, -0.00000286, 0.00000256, 0.00000693, -0.00000467, -0.00000047, -0.00000427, 0.00000497, 0.00000186, 0.00000236, -0.00000724, -0.00000592, 0.00000129, 0.00000149, 0.00000187, 0.00000289, 0.00000026, 0.00000247, 0.00000299, -0.00000886, 0.00000116, 0.00000427, -0.00000360, -0.00000611, 0.00000353, 0.00000100, 0.00000509, -0.00000061, 0.00000252, -0.00000780, 0.00000122, 0.00001406, -0.00000587, -0.00000191, 0.00000223, 0.00000237, -0.00000743, -0.00000950, 0.00000586, -0.00000143, -0.00000062, -0.00001070, 0.00000481, -0.00000865, -0.00000855, 0.00000253, -0.00000534, -0.00000704, -0.00001281, 0.00001094, -0.00002886, -0.00003393, 0.00003417, 0.00005094, 0.00002236, 0.00005231, 0.00001767, -0.00003439, -0.00002428, -0.00000444, -0.00001048, -0.00000624, -0.00000649, -0.00001830, -0.00000516, -0.00007053, -0.00002945, 0.00012698, 0.00007273, 0.00012620, -0.00006263, -0.00004881, -0.00001605, -0.00000100, -0.00000063, -0.00001106, -0.00000050, -0.00001414, 0.00000836, -0.00001442, -0.00000083, -0.00000892, -0.00000465, 0.00000650, -0.00000565, -0.00001006, 0.00000292, -0.00000221, 0.00000104, -0.00000645, 0.00000200, 0.00001600, -0.00000689, -0.00000020, -0.00000085, 0.00000266, 0.00000488, -0.00000044, -0.00000081, -0.00000582, 0.00000143, 0.00000597, -0.00000290, -0.00000784, -0.00000184, 0.00000913, -0.00000302, 0.00000691, 0.00000019, 0.00000235, -0.00000704, -0.00000461, 0.00000299, -0.00000159, -0.00000008, 0.00000224, 0.00000060, -0.00000754, 0.00000572, 0.00000271, -0.00000117, -0.00000166, 0.00000088, -0.00000640, 0.00000281, 0.00000081, },
		{ 0.00000266, 0.00000029, -0.00000488, -0.00000233, 0.00000055, 0.00000824, -0.00000426, -0.00000058, -0.00000453, 0.00000444, 0.00000207, 0.00000222, -0.00000529, -0.00000727, 0.00000086, 0.00000133, 0.00000203, 0.00000254, 0.00000118, 0.00000170, 0.00000387, -0.00000866, -0.00000033, 0.00000496, -0.00000249, -0.00000693, 0.00000304, 0.00000119, 0.00000457, 0.00000035, 0.00000183, -0.00000575, -0.00000250, 0.00001627, -0.00000525, -0.00000215, 0.00000099, 0.00000293, -0.00000530, -0.00001108, 0.00000454, 0.00000007, -0.00000092, -0.00000995, 0.00000240, -0.00000548, -0.00001105, 0.00000392, -0.00000690, -0.00000445, -0.00001526, 0.00001124, -0.00002321, -0.00004087, 0.00002627, 0.00005689, 0.00002321, 0.00005268, 0.00002482, -0.00003454, -0.00002674, -0.00000472, -0.00001078, -0.00000604, -0.00000658, -0.00001682, -0.00000648, -0.00006029, -0.00004359, 0.00012160, 0.00006629, 0.00012597, -0.00003467, -0.00006344, -0.00001204, -0.00000632, 0.00000376, -0.00001437, 0.00000216, -0.00001605, 0.00000893, -0.00001399, -0.00000114, -0.00000835, -0.00000559, 0.00000574, -0.00000288, -0.00001201, 0.00000271, -0.00000131, -0.00000015, -0.00000428, -0.00000188, 0.00001787, -0.00000550, -0.00000167, -0.00000011, 0.00000158, 0.00000545, -0.00000044, 0.00000016, -0.00000597, 0.00000002, 0.00000645, -0.00000180, -0.00000778, -0.00000320, 0.00000896, -0.00000180, 0.00000535, 0.00000163, 0.00000148, -0.00000475, -0.00000688, 0.00000348, -0.00000174, 0.00000017, 0.00000176, 0.00000160, -0.00000782, 0.00000434, 0.00000387, -0.00000097, -0.00000203, 0.00000108, -0.00000597, 0.00000171, 0.00000169, },
		{ 0.00000175, 0.00000128, -0.00000511, -0.00000183, -0.00000115, 0.00000887, -0.00000325, -0.00000103, -0.00000442, 0.00000355, 0.00000266, 0.00000161, -0.00000317, -0.00000835, 0.00000027, 0.00000108, 0.00000226, 0.00000214, 0.00000197, 0.00000118, 0.00000445, -0.00000809, -0.00000194, 0.00000543, -0.00000129, -0.00000744, 0.00000217, 0.00000169, 0.00000370, 0.00000161, 0.00000082, -0.00000341, -0.00000600, 0.00001762, -0.00000384, -0.00000280, 0.00000037, 0.00000256, -0.00000271, -0.00001235, 0.00000293, 0.00000137, -0.00000097, -0.00000925, 0.00000016, -0.00000291, -0.00001272, 0.00000450, -0.00000766, -0.00000298, -0.00001610, 0.00000958, -0.00001636, -0.00004646, 0.00001673, 0.00006173, 0.00002568, 0.00005177, 0.00003275, -0.00003436, -0.00002921, -0.00000528, -0.00001088, -0.00000609, -0.00000650, -0.00001550, -0.00000776, -0.00005098, -0.00005458, 0.00011352, 0.00006361, 0.00011942, -0.00000523, -0.00007449, -0.00001103, -0.00000967, 0.00000644, -0.00001618, 0.00000339, -0.00001661, 0.00000837, -0.00001277, -0.00000205, -0.00000727, -0.00000678, 0.00000506, -0.00000063, -0.00001279, 0.00000119, 0.00000055, -0.00000198, -0.00000178, -0.00000539, 0.00001864, -0.00000318, -0.00000353, 0.00000079, 0.00000041, 0.00000603, -0.00000054, 0.00000109, -0.00000586, -0.00000137, 0.00000653, -0.00000049, -0.00000775, -0.00000424, 0.00000827, -0.00000028, 0.00000375, 0.00000297, 0.00000068, -0.00000264, -0.00000851, 0.00000324, -0.00000140, 0.00000005, 0.00000163, 0.00000216, -0.00000752, 0.00000257, 0.00000498, -0.00000067, -0.00000229, 0.00000101, -0.00000526, 0.00000049, 0.00000257, },
		{ 0.00000110, 0.00000193, -0.00000510, -0.00000147, -0.00000243, 0.00000875, -0.00000172, -0.00000173, -0.00000402, 0.00000239, 0.00000348, 0.00000071, -0.00000111, -0.00000898, -0.00000053, 0.00000083, 0.00000245, 0.00000183, 0.00000253, 0.00000097, 0.00000471, -0.00000719, -0.00000360, 0.00000566, -0.00000008, -0.00000758, 0.00000097, 0.00000242, 0.00000262, 0.00000298, -0.00000031, -0.00000107, -0.00000889, 0.00001781, -0.00000152, -0.00000395, 0.00000043, 0.00000136, 0.00000011, -0.00001316, 0.00000108, 0.00000238, -0.00000071, -0.00000866, -0.00000176, -0.00000104, -0.00001354, 0.00000432, -0.00000766, -0.00000254, -0.00001557, 0.00000652, -0.00000922, -0.00004995, 0.00000559, 0.00006511, 0.00002973, 0.00004984, 0.00004108, -0.00003351, -0.00003190, -0.00000596, -0.00001091, -0.00000629, -0.00000633, -0.00001432, -0.00000892, -0.00004274, -0.00006249, 0.00010303, 0.00006408, 0.00010842, 0.00002311, -0.00008038, -0.00001352, -0.00001094, 0.00000738, -0.00001647, 0.00000325, -0.00001595, 0.00000686, -0.00001090, -0.00000346, -0.00000575, -0.00000817, 0.00000452, 0.00000100, -0.00001240, -0.00000137, 0.00000293, -0.00000395, 0.00000055, -0.00000806, 0.00001810, 0.00000003, -0.00000564, 0.00000176, -0.00000074, 0.00000653, -0.00000062, 0.00000188, -0.00000547, -0.00000270, 0.00000626, 0.00000094, -0.00000765, -0.00000505, 0.00000724, 0.00000130, 0.00000234, 0.00000400, 0.00000012, -0.00000094, -0.00000933, 0.00000227, -0.00000065, -0.00000039, 0.00000180, 0.00000235, -0.00000677, 0.00000065, 0.00000585, -0.00000017, -0.00000247, 0.00000074, -0.00000443, -0.00000070, 0.00000328, },
		{ 0.00000079, 0.00000224, -0.00000486, -0.00000128, -0.00000322, 0.00000794, 0.00000018, -0.00000253, -0.00000349, 0.00000111, 0.00000437, -0.00000023, 0.00000062, -0.00000903, -0.00000159, 0.00000064, 0.00000249, 0.00000170, 0.00000279, 0.00000106, 0.00000471, -0.00000604, -0.00000522, 0.00000563, 0.00000109, -0.00000734, -0.00000050, 0.00000324, 0.00000148, 0.00000430, -0.00000131, 0.00000094, -0.00001086, 0.00001672, 0.00000167, -0.00000552, 0.00000105, -0.00000040, 0.00000281, -0.00001335, -0.00000096, 0.00000306, -0.00000018, -0.00000817, -0.00000336, 0.00000019, -0.00001364, 0.00000356, -0.00000709, -0.00000286, -0.00001412, 0.00000278, -0.00000271, -0.00005081, -0.00000684, 0.00006660, 0.00003530, 0.00004723, 0.00004942, -0.00003166, -0.00003495, -0.00000666, -0.00001099, -0.00000654, -0.00000617, -0.00001324, -0.00000992, -0.00003559, -0.00006759, 0.00009063, 0.00006681, 0.00009495, 0.00004813, -0.00008020, -0.00001937, -0.00001046, 0.00000689, -0.00001554, 0.00000202, -0.00001436, 0.00000468, -0.00000865, -0.00000518, -0.00000397, -0.00000959, 0.00000404, 0.00000203, -0.00001102, -0.00000455, 0.00000532, -0.00000557, 0.00000228, -0.00000959, 0.00001618, 0.00000395, -0.00000779, 0.00000261, -0.00000175, 0.00000685, -0.00000060, 0.00000243, -0.00000477, -0.00000395, 0.00000569, 0.00000236, -0.00000738, -0.00000574, 0.00000607, 0.00000269, 0.00000130, 0.00000460, -0.00000008, 0.00000024, -0.00000932, 0.00000066, 0.00000038, -0.00000100, 0.00000212, 0.00000232, -0.00000576, -0.00000121, 0.00000634, 0.00000057, -0.00000261, 0.00000037, -0.00000360, -0.00000169, 0.00000371, },
		{ 0.00000080, 0.00000224, -0.00000444, -0.00000128, -0.00000354, 0.00000660, 0.00000223, -0.00000321, -0.00000298, -0.00000012, 0.00000515, -0.00000100, 0.00000185, -0.00000844, -0.00000289, 0.00000052, 0.00000235, 0.00000179, 0.00000276, 0.00000138, 0.00000452, -0.00000474, -0.00000671, 0.00000531, 0.00000219, -0.00000676, -0.00000211, 0.00000402, 0.00000044, 0.00000537, -0.00000198, 0.00000240, -0.00001173, 0.00001435, 0.00000554, -0.00000729, 0.00000198, -0.00000236, 0.00000505, -0.00001277, -0.00000320, 0.00000342, 0.00000052, -0.00000768, -0.00000473, 0.00000090, -0.00001326, 0.00000249, -0.00000625, -0.00000359, -0.00001225, -0.00000092, 0.00000243, -0.00004887, -0.00001998, 0.00006572, 0.00004227, 0.00004427, 0.00005738, -0.00002855, -0.00003847, -0.00000730, -0.00001118, -0.00000676, -0.00000609, -0.00001220, -0.00001080, -0.00002949, -0.00007025, 0.00007695, 0.00007077, 0.00008086, 0.00006824, -0.00007382, -0.00002784, -0.00000894, 0.00000552, -0.00001383, 0.00000011, -0.00001225, 0.00000225, -0.00000634, -0.00000694, -0.00000216, -0.00001087, 0.00000352, 0.00000259, -0.00000896, -0.00000788, 0.00000719, -0.00000644, 0.00000310, -0.00000983, 0.00001303, 0.00000827, -0.00000967, 0.00000315, -0.00000248, 0.00000689, -0.00000037, 0.00000268, -0.00000377, -0.00000511, 0.00000487, 0.00000365, -0.00000686, -0.00000639, 0.00000492, 0.00000375, 0.00000073, 0.00000473, 0.00000010, 0.00000090, -0.00000858, -0.00000136, 0.00000144, -0.00000161, 0.00000245, 0.00000222, -0.00000471, -0.00000279, 0.00000630, 0.00000157, -0.00000272, -0.00000003, -0.00000291, -0.00000240, 0.00000380, },
		{ 0.00000109, 0.00000204, -0.00000392, -0.00000140, -0.00000349, 0.00000491, 0.00000417, -0.00000360, -0.00000265, -0.00000116, 0.00000566, -0.00000143, 0.00000245, -0.00000722, -0.00000435, 0.00000044, 0.00000204, 0.00000205, 0.00000254, 0.00000181, 0.00000425, -0.00000339, -0.00000795, 0.00000469, 0.00000320, -0.00000588, -0.00000372, 0.00000460, -0.00000035, 0.00000605, -0.00000217, 0.00000315, -0.00001148, 0.00001089, 0.00000975, -0.00000893, 0.00000289, -0.00000411, 0.00000650, -0.00001134, -0.00000558, 0.00000351, 0.00000127, -0.00000708, -0.00000597, 0.00000129, -0.00001267, 0.00000140, -0.00000541, -0.00000439, -0.00001043, -0.00000403, 0.00000576, -0.00004431, -0.00003295, 0.00006194, 0.00005044, 0.00004134, 0.00006464, -0.00002399, -0.00004245, -0.00000789, -0.00001151, -0.00000692, -0.00000610, -0.00001117, -0.00001155, -0.00002435, -0.00007089, 0.00006265, 0.00007500, 0.00006764, 0.00008252, -0.00006186, -0.00003771, -0.00000731, 0.00000394, -0.00001189, -0.00000201, -0.00001007, -0.00000007, -0.00000426, -0.00000852, -0.00000055, -0.00001181, 0.00000284, 0.00000284, -0.00000655, -0.00001086, 0.00000815, -0.00000631, 0.00000290, -0.00000885, 0.00000894, 0.00001257, -0.00001099, 0.00000320, -0.00000280, 0.00000655, 0.00000013, 0.00000259, -0.00000251, -0.00000613, 0.00000389, 0.00000471, -0.00000606, -0.00000707, 0.00000388, 0.00000441, 0.00000062, 0.00000444, 0.00000059, 0.00000111, -0.00000729, -0.00000352, 0.00000231, -0.00000203, 0.00000262, 0.00000223, -0.00000378, -0.00000392, 0.00000571, 0.00000279, -0.00000280, -0.00000040, -0.00000241, -0.00000281, 0.00000358, },
		{ 0.00000153, 0.00000176, -0.00000341, -0.00000158, -0.00000317, 0.00000309, 0.00000578, -0.00000354, -0.00000258, -0.00000191, 0.00000580, -0.00000144, 0.00000242, -0.00000550, -0.00000583, 0.00000033, 0.00000162, 0.00000240, 0.00000224, 0.00000223, 0.00000402, -0.00000211, -0.00000888, 0.00000377, 0.00000410, -0.00000479, -0.00000519, 0.00000485, -0.00000080, 0.00000626, -0.00000184, 0.00000319, -0.00001020, 0.00000668, 0.00001386, -0.00001008, 0.00000343, -0.00000530, 0.00000694, -0.00000906, -0.00000802, 0.00000336, 0.00000193, -0.00000625, -0.00000718, 0.00000153, -0.00001209, 0.00000051, -0.00000480, -0.00000500, -0.00000897, -0.00000623, 0.00000721, -0.00003775, -0.00004473, 0.00005481, 0.00005950, 0.00003881, 0.00007094, -0.00001790, -0.00004682, -0.00000850, -0.00001193, -0.00000705, -0.00000621, -0.00001018, -0.00001217, -0.00002008, -0.00006989, 0.00004833, 0.00007869, 0.00005634, 0.00009081, -0.00004557, -0.00004750, -0.00000655, 0.00000275, -0.00001017, -0.00000393, -0.00000817, -0.00000196, -0.00000262, -0.00000976, 0.00000071, -0.00001229, 0.00000190, 0.00000292, -0.00000413, -0.00001312, 0.00000795, -0.00000513, 0.00000179, -0.00000691, 0.00000436, 0.00001640, -0.00001143, 0.00000260, -0.00000264, 0.00000580, 0.00000092, 0.00000219, -0.00000107, -0.00000696, 0.00000279, 0.00000548, -0.00000495, -0.00000777, 0.00000300, 0.00000467, 0.00000088, 0.00000388, 0.00000126, 0.00000104, -0.00000569, -0.00000553, 0.00000274, -0.00000212, 0.00000252, 0.00000244, -0.00000310, -0.00000455, 0.00000460, 0.00000413, -0.00000277, -0.00000074, -0.00000210, -0.00000294, 0.00000312, },
		{ 0.00000198, 0.00000154, -0.00000301, -0.00000175, -0.00000272, 0.00000135, 0.00000687, -0.00000295, -0.00000281, -0.00000234, 0.00000558, -0.00000104, 0.00000186, -0.00000349, -0.00000713, 0.00000007, 0.00000119, 0.00000273, 0.00000200, 0.00000250, 0.00000390, -0.00000095, -0.00000945, 0.00000256, 0.00000487, -0.00000356, -0.00000641, 0.00000471, -0.00000086, 0.00000598, -0.00000100, 0.00000260, -0.00000815, 0.00000215, 0.00001741, -0.00001036, 0.00000331, -0.00000568, 0.00000628, -0.00000606, -0.00001038, 0.00000297, 0.00000243, -0.00000515, -0.00000842, 0.00000173, -0.00001166, -0.00000008, -0.00000447, -0.00000534, -0.00000798, -0.00000748, 0.00000706, -0.00003008, -0.00005421, 0.00004411, 0.00006895, 0.00003710, 0.00007602, -0.00001024, -0.00005140, -0.00000926, -0.00001236, -0.00000721, -0.00000635, -0.00000927, -0.00001262, -0.00001662, -0.00006762, 0.00003448, 0.00008127, 0.00004757, 0.00009352, -0.00002655, -0.00005573, -0.00000743, 0.00000237, -0.00000897, -0.00000540, -0.00000679, -0.00000327, -0.00000151, -0.00001058, 0.00000153, -0.00001224, 0.00000069, 0.00000294, -0.00000193, -0.00001442, 0.00000655, -0.00000308, 0.00000003, -0.00000439, -0.00000021, 0.00001932, -0.00001081, 0.00000132, -0.00000201, 0.00000468, 0.00000195, 0.00000155, 0.00000045, -0.00000755, 0.00000160, 0.00000592, -0.00000359, -0.00000846, 0.00000225, 0.00000462, 0.00000136, 0.00000320, 0.00000194, 0.00000084, -0.00000403, -0.00000712, 0.00000259, -0.00000180, 0.00000211, 0.00000289, -0.00000271, -0.00000467, 0.00000308, 0.00000544, -0.00000258, -0.00000106, -0.00000197, -0.00000289, 0.00000257, },
		{ 0.00000231, 0.00000149, -0.00000276, -0.00000186, -0.00000223, -0.00000015, 0.00000735, -0.00000186, -0.00000328, -0.00000252, 0.00000504, -0.00000033, 0.00000094, -0.00000143, -0.00000805, -0.00000045, 0.00000085, 0.00000290, 0.00000193, 0.00000254, 0.00000394, 0.00000003, -0.00000964, 0.00000114, 0.00000548, -0.00000226, -0.00000727, 0.00000414, -0.00000056, 0.00000529, 0.00000025, 0.00000155, -0.00000563, -0.00000225, 0.00001997, -0.00000954, 0.00000238, -0.00000517, 0.00000463, -0.00000262, -0.00001244, 0.00000231, 0.00000274, -0.00000380, -0.00000966, 0.00000191, -0.00001141, -0.00000040, -0.00000436, -0.00000548, -0.00000739, -0.00000799, 0.00000585, -0.00002234, -0.00006044, 0.00002991, 0.00007809, 0.00003666, 0.00007972, -0.00000106, -0.00005601, -0.00001032, -0.00001271, -0.00000747, -0.00000647, -0.00000849, -0.00001287, -0.00001389, -0.00006437, 0.00002146, 0.00008238, 0.00004150, 0.00009153, -0.00000658, -0.00006112, -0.00001041, 0.00000294, -0.00000835, -0.00000637, -0.00000598, -0.00000400, -0.00000090, -0.00001103, 0.00000194, -0.00001171, -0.00000074, 0.00000294, -0.00000010, -0.00001467, 0.00000410, -0.00000047, -0.00000199, -0.00000173, -0.00000430, 0.00002099, -0.00000900, -0.00000060, -0.00000098, 0.00000328, 0.00000311, 0.00000078, 0.00000191, -0.00000783, 0.00000034, 0.00000605, -0.00000205, -0.00000907, 0.00000156, 0.00000440, 0.00000190, 0.00000257, 0.00000250, 0.00000069, -0.00000252, -0.00000811, 0.00000180, -0.00000110, 0.00000141, 0.00000355, -0.00000256, -0.00000439, 0.00000135, 0.00000652, -0.00000211, -0.00000139, -0.00000193, -0.00000277, 0.00000204, },
		},
		{ { -0.00000075, -0.00000105, -0.00000188, -0.00000012, 0.00001140, 0.00000172, -0.00001683, 0.00000243, 0.00000687, 0.00000558, -0.00000116, -0.00000241, -0.00000557, -0.00000342, 0.00001051, 0.00000036, -0.00000106, -0.00000164, 0.00000440, 0.00000089, -0.00000727, -0.00000366, 0.00001597, -0.00000272, -0.00000737, 0.00000088, 0.00000499, 0.00000134, -0.00001613, 0.00000010, -0.00000349, -0.00000440, -0.00000222, 0.00000177, -0.00000685, 0.00000232, 0.00000081, -0.00000329, -0.00000144, -0.00000544, 0.00001098, -0.00000320, -0.00000401, 0.00000021, 0.00000312, -0.00000329, -0.00001114, -0.00000521, -0.00000785, -0.00000948, -0.00005948, 0.00001149, 0.00008451, 0.00003825, 0.00008679, 0.00001005, -0.00006230, -0.00001493, -0.00000988, -0.00000546, -0.00000547, -0.00000281, -0.00000515, -0.00000216, -0.00000147, -0.00000216, -0.00000515, -0.00000281, -0.00000547, -0.00000546, -0.00000988, -0.00001493, -0.00006230, 0.00001005, 0.00008679, 0.00003825, 0.00008451, 0.00001149, -0.00005948, -0.00000948, -0.00000785, -0.00000521, -0.00001114, -0.00000329, 0.00000312, 0.00000021, -0.00000401, -0.00000320, 0.00001098, -0.00000544, -0.00000144, -0.00000329, 0.00000081, 0.00000232, -0.00000685, 0.00000177, -0.00000222, -0.00000440, -0.00000349, 0.00000010, -0.00001613, 0.00000134, 0.00000499, 0.00000088, -0.00000737, -0.00000272, 0.00001597, -0.00000366, -0.00000727, 0.00000089, 0.00000440, -0.00000164, -0.00000106, 0.00000036, 0.00001051, -0.00000342, -0.00000557, -0.00000241, -0.00000116, 0.00000558, 0.00000687, 0.00000243, -0.00001683, 0.00000172, 0.00001140, -0.00000012, -0.00000188, -0.00000105, },
		{ -0.00000055, -0.00000135, -0.00000100, -0.00000125, 0.00000914, 0.00000628, -0.00001645, -0.00000146, 0.00000773, 0.00000542, 0.00000077, -0.00000341, -0.00000381, -0.00000606, 0.00001040, 0.00000150, -0.00000073, -0.00000220, 0.00000361, 0.00000239, -0.00000582, -0.00000676, 0.00001418, 0.00000253, -0.00000993, 0.00000175, 0.00000243, 0.00000530, -0.00001607, -0.00000282, -0.00000152, -0.00000508, -0.00000254, 0.00000145, -0.00000627, 0.00000080, 0.00000249, -0.00000442, -0.00000001, -0.00000663, 0.00000974, -0.00000058, -0.00000533, 0.00000033, 0.00000207, -0.00000047, -0.00001219, -0.00000463, -0.00000914, -0.00000460, -0.00005407, -0.00001631, 0.00008994, 0.00004257, 0.00008424, 0.00003185, -0.00006655, -0.00001972, -0.00001027, -0.00000573, -0.00000602, -0.00000285, -0.00000481, -0.00000300, -0.00000140, -0.00000170, -0.00000485, -0.00000340, -0.00000455, -0.00000546, -0.00000929, -0.00001164, -0.00005588, -0.00000903, 0.00008431, 0.00003928, 0.00007341, 0.00003650, -0.00005717, -0.00001796, -0.00000552, -0.00000661, -0.00000940, -0.00000621, 0.00000379, 0.00000021, -0.00000266, -0.00000539, 0.00001123, -0.00000331, -0.00000331, -0.00000202, -0.00000084, 0.00000369, -0.00000733, 0.00000191, -0.00000176, -0.00000373, -0.00000500, 0.00000150, -0.00001413, -0.00000357, 0.00000726, 0.00000035, -0.00000504, -0.00000663, 0.00001538, 0.00000071, -0.00000852, -0.00000056, 0.00000447, -0.00000050, -0.00000160, -0.00000039, 0.00000980, -0.00000032, -0.00000717, -0.00000163, -0.00000262, 0.00000527, 0.00000616, 0.00000529, -0.00001532, -0.00000345, 0.00001267, 0.00000180, -0.00000267, -0.00000115, },
		{ -0.00000006, -0.00000181, -0.00000035, -0.00000153, 0.00000634, 0.00000968, -0.00001418, -0.00000591, 0.00000833, 0.00000510, 0.00000274, -0.00000418, -0.00000239, -0.00000781, 0.00000921, 0.00000318, -0.00000081, -0.00000201, 0.00000211, 0.00000395, -0.00000451, -0.00000827, 0.00001035, 0.00000826, -0.00001188, 0.00000216, 0.00000055, 0.00000742, -0.00001369, -0.00000691, 0.00000047, -0.00000552, -0.00000284, 0.00000103, -0.00000560, -0.00000090, 0.00000418, -0.00000536, 0.00000098, -0.00000707, 0.00000790, 0.00000198, -0.00000619, 0.00000016, 0.00000118, 0.00000156, -0.00001186, -0.00000550, -0.00000862, -0.00000375, -0.00004275, -0.00004268, 0.00008658, 0.00005281, 0.00007737, 0.00005475, -0.00006727, -0.00002640, -0.00001028, -0.00000652, -0.00000599, -0.00000359, -0.00000392, -0.00000407, -0.00000126, -0.00000158, -0.00000406, -0.00000429, -0.00000363, -0.00000541, -0.00000875, -0.00000937, -0.00004851, -0.00002434, 0.00007692, 0.00004418, 0.00005996, 0.00005564, -0.00004713, -0.00002827, -0.00000359, -0.00000780, -0.00000786, -0.00000841, 0.00000353, 0.00000069, -0.00000166, -0.00000674, 0.00001022, -0.00000026, -0.00000548, -0.00000074, -0.00000239, 0.00000486, -0.00000758, 0.00000169, -0.00000101, -0.00000328, -0.00000580, 0.00000141, -0.00001073, -0.00000838, 0.00000836, 0.00000074, -0.00000347, -0.00000880, 0.00001263, 0.00000562, -0.00000905, -0.00000210, 0.00000399, 0.00000090, -0.00000203, -0.00000099, 0.00000862, 0.00000272, -0.00000814, -0.00000145, -0.00000334, 0.00000434, 0.00000580, 0.00000697, -0.00001226, -0.00000851, 0.00001258, 0.00000431, -0.00000306, -0.00000174, },
		{ 0.00000048, -0.00000211, -0.00000017, -0.00000105, 0.00000348, 0.00001157, -0.00001036, -0.00001024, 0.00000823, 0.00000498, 0.00000435, -0.00000431, -0.00000167, -0.00000840, 0.00000691, 0.00000534, -0.00000129, -0.00000114, 0.00000014, 0.00000540, -0.00000341, -0.00000835, 0.00000534, 0.00001330, -0.00001230, 0.00000136, 0.00000007, 0.00000732, -0.00000929, -0.00001145, 0.00000188, -0.00000546, -0.00000323, 0.00000057, -0.00000480, -0.00000283, 0.00000586, -0.00000605, 0.00000154, -0.00000692, 0.00000579, 0.00000414, -0.00000636, -0.00000047, 0.00000075, 0.00000246, -0.00001000, -0.00000775, -0.00000652, -0.00000595, -0.00002874, -0.00006312, 0.00007243, 0.00006814, 0.00006793, 0.00007666, -0.00006313, -0.00003518, -0.00000987, -0.00000795, -0.00000536, -0.00000491, -0.00000274, -0.00000508, -0.00000124, -0.00000163, -0.00000311, -0.00000507, -0.00000312, -0.00000498, -0.00000850, -0.00000768, -0.00004112, -0.00003541, 0.00006544, 0.00005107, 0.00004701, 0.00006737, -0.00003096, -0.00003790, -0.00000358, -0.00000791, -0.00000724, -0.00000941, 0.00000213, 0.00000173, -0.00000116, -0.00000710, 0.00000802, 0.00000341, -0.00000757, 0.00000026, -0.00000362, 0.00000563, -0.00000739, 0.00000091, 0.00000016, -0.00000316, -0.00000582, 0.00000015, -0.00000677, -0.00001216, 0.00000777, 0.00000216, -0.00000270, -0.00000937, 0.00000843, 0.00001009, -0.00000836, -0.00000384, 0.00000317, 0.00000218, -0.00000207, -0.00000164, 0.00000728, 0.00000531, -0.00000821, -0.00000199, -0.00000331, 0.00000287, 0.00000576, 0.00000759, -0.00000826, -0.00001269, 0.00001091, 0.00000713, -0.00000286, -0.00000274, },
		{ 0.00000074, -0.00000199, -0.00000058, -0.00000001, 0.00000100, 0.00001190, -0.00000558, -0.00001373, 0.00000703, 0.00000529, 0.00000531, -0.00000358, -0.00000178, -0.00000784, 0.00000373, 0.00000770, -0.00000195, 0.00000010, -0.00000189, 0.00000642, -0.00000235, -0.00000749, 0.00000021, 0.00001655, -0.00001052, -0.00000103, 0.00000120, 0.00000525, -0.00000375, -0.00001538, 0.00000206, -0.00000467, -0.00000379, 0.00000011, -0.00000387, -0.00000492, 0.00000734, -0.00000630, 0.00000157, -0.00000623, 0.00000356, 0.00000582, -0.00000592, -0.00000142, 0.00000067, 0.00000244, -0.00000714, -0.00001055, -0.00000394, -0.00000923, -0.00001576, -0.00007404, 0.00004743, 0.00008602, 0.00005858, 0.00009519, -0.00005295, -0.00004599, -0.00000920, -0.00000994, -0.00000429, -0.00000650, -0.00000167, -0.00000572, -0.00000154, -0.00000164, -0.00000230, -0.00000536, -0.00000327, -0.00000404, -0.00000861, -0.00000629, -0.00003434, -0.00004229, 0.00005109, 0.00005807, 0.00003662, 0.00007163, -0.00001132, -0.00004439, -0.00000644, -0.00000666, -0.00000763, -0.00000925, -0.00000015, 0.00000305, -0.00000096, -0.00000670, 0.00000507, 0.00000710, -0.00000900, 0.00000056, -0.00000422, 0.00000574, -0.00000649, -0.00000061, 0.00000177, -0.00000336, -0.00000518, -0.00000180, -0.00000307, -0.00001427, 0.00000547, 0.00000430, -0.00000241, -0.00000884, 0.00000373, 0.00001323, -0.00000615, -0.00000575, 0.00000215, 0.00000307, -0.00000155, -0.00000241, 0.00000594, 0.00000724, -0.00000733, -0.00000316, -0.00000273, 0.00000114, 0.00000581, 0.00000752, -0.00000405, -0.00001537, 0.00000771, 0.00000988, -0.00000194, -0.00000391, },
		{ 0.00000048, -0.00000134, -0.00000147, 0.00000121, -0.00000075, 0.00001082, -0.00000055, -0.00001579, 0.00000460, 0.00000606, 0.00000556, -0.00000204, -0.00000256, -0.00000644, 0.00000015, 0.00000980, -0.00000239, 0.00000126, -0.00000345, 0.00000667, -0.00000107, -0.00000631, -0.00000400, 0.00001725, -0.00000642, -0.00000479, 0.00000351, 0.00000210, 0.00000167, -0.00001764, 0.00000056, -0.00000311, -0.00000451, -0.00000032, -0.00000286, -0.00000698, 0.00000832, -0.00000584, 0.00000089, -0.00000493, 0.00000115, 0.00000719, -0.00000521, -0.00000230, 0.00000055, 0.00000205, -0.00000419, -0.00001273, -0.00000232, -0.00001145, -0.00000693, -0.00007377, 0.00001399, 0.00010232, 0.00005250, 0.00010807, -0.00003604, -0.00005834, -0.00000863, -0.00001227, -0.00000309, -0.00000801, -0.00000104, -0.00000577, -0.00000225, -0.00000150, -0.00000185, -0.00000500, -0.00000412, -0.00000268, -0.00000895, -0.00000516, -0.00002840, -0.00004552, 0.00003536, 0.00006352, 0.00002996, 0.00006946, 0.00000892, -0.00004603, -0.00001207, -0.00000460, -0.00000854, -0.00000854, -0.00000262, 0.00000401, -0.00000062, -0.00000602, 0.00000201, 0.00001006, -0.00000918, -0.00000023, -0.00000394, 0.00000497, -0.00000471, -0.00000287, 0.00000370, -0.00000375, -0.00000414, -0.00000392, -0.00000025, -0.00001451, 0.00000187, 0.00000652, -0.00000210, -0.00000784, -0.00000056, 0.00001441, -0.00000252, -0.00000754, 0.00000096, 0.00000346, -0.00000054, -0.00000322, 0.00000463, 0.00000849, -0.00000567, -0.00000465, -0.00000198, -0.00000046, 0.00000566, 0.00000716, -0.00000028, -0.00001620, 0.00000329, 0.00001211, -0.00000029, -0.00000497, },
		{ -0.00000040, -0.00000029, -0.00000255, 0.00000220, -0.00000152, 0.00000863, 0.00000405, -0.00001608, 0.00000109, 0.00000708, 0.00000529, 0.00000001, -0.00000365, -0.00000469, -0.00000323, 0.00001111, -0.00000222, 0.00000190, -0.00000415, 0.00000590, 0.00000062, -0.00000533, -0.00000664, 0.00001527, -0.00000053, -0.00000911, 0.00000602, -0.00000089, 0.00000563, -0.00001740, -0.00000275, -0.00000094, -0.00000528, -0.00000074, -0.00000189, -0.00000873, 0.00000842, -0.00000436, -0.00000068, -0.00000292, -0.00000152, 0.00000848, -0.00000460, -0.00000269, -0.00000002, 0.00000192, -0.00000205, -0.00001325, -0.00000276, -0.00001113, -0.00000381, -0.00006322, -0.00002314, 0.00011199, 0.00005267, 0.00011373, -0.00001250, -0.00007120, -0.00000877, -0.00001461, -0.00000212, -0.00000909, -0.00000107, -0.00000521, -0.00000327, -0.00000127, -0.00000175, -0.00000408, -0.00000540, -0.00000127, -0.00000918, -0.00000443, -0.00002322, -0.00004595, 0.00001978, 0.00006614, 0.00002742, 0.00006250, 0.00002731, -0.00004236, -0.00001933, -0.00000290, -0.00000903, -0.00000812, -0.00000445, 0.00000399, 0.00000034, -0.00000556, -0.00000051, 0.00001165, -0.00000770, -0.00000227, -0.00000273, 0.00000331, -0.00000212, -0.00000565, 0.00000563, -0.00000401, -0.00000305, -0.00000571, 0.00000132, -0.00001308, -0.00000234, 0.00000814, -0.00000130, -0.00000688, -0.00000383, 0.00001351, 0.00000204, -0.00000876, -0.00000049, 0.00000339, 0.00000076, -0.00000382, 0.00000325, 0.00000918, -0.00000355, -0.00000605, -0.00000144, -0.00000161, 0.00000508, 0.00000684, 0.00000263, -0.00001520, -0.00000176, 0.00001335, 0.00000202, -0.00000566, },
		{ -0.00000181, 0.00000086, -0.00000342, 0.00000259, -0.00000125, 0.00000581, 0.00000766, -0.00001457, -0.00000307, 0.00000795, 0.00000480, 0.00000215, -0.00000456, -0.00000312, -0.00000586, 0.00001125, -0.00000116, 0.00000177, -0.00000378, 0.00000412, 0.00000271, -0.00000478, -0.00000755, 0.00001114, 0.00000603, -0.00001280, 0.00000755, -0.00000249, 0.00000716, -0.00001440, -0.00000747, 0.00000134, -0.00000585, -0.00000118, -0.00000104, -0.00000990, 0.00000734, -0.00000173, -0.00000313, -0.00000029, -0.00000443, 0.00000977, -0.00000428, -0.00000242, -0.00000122, 0.00000237, -0.00000121, -0.00001169, -0.00000551, -0.00000804, -0.00000597, -0.00004578, -0.00005772, 0.00011026, 0.00006092, 0.00011184, 0.00001649, -0.00008292, -0.00001049, -0.00001649, -0.00000175, -0.00000951, -0.00000181, -0.00000421, -0.00000434, -0.00000113, -0.00000182, -0.00000292, -0.00000662, -0.00000029, -0.00000888, -0.00000436, -0.00001851, -0.00004460, 0.00000572, 0.00006512, 0.00002877, 0.00005255, 0.00004218, -0.00003404, -0.00002639, -0.00000290, -0.00000825, -0.00000869, -0.00000508, 0.00000265, 0.00000209, -0.00000556, -0.00000212, 0.00001160, -0.00000459, -0.00000528, -0.00000090, 0.00000108, 0.00000086, -0.00000842, 0.00000702, -0.00000375, -0.00000230, -0.00000676, 0.00000152, -0.00001040, -0.00000640, 0.00000864, 0.00000016, -0.00000618, -0.00000588, 0.00001088, 0.00000670, -0.00000888, -0.00000228, 0.00000297, 0.00000206, -0.00000399, 0.00000171, 0.00000948, -0.00000128, -0.00000703, -0.00000137, -0.00000214, 0.00000402, 0.00000667, 0.00000456, -0.00001273, -0.00000668, 0.00001321, 0.00000484, -0.00000577, },
		{ -0.00000349, 0.00000173, -0.00000374, 0.00000219, -0.00000011, 0.00000286, 0.00000990, -0.00001152, -0.00000728, 0.00000823, 0.00000447, 0.00000395, -0.00000488, -0.00000214, -0.00000737, 0.00001004, 0.00000080, 0.00000086, -0.00000245, 0.00000160, 0.00000494, -0.00000456, -0.00000707, 0.00000587, 0.00001190, -0.00001465, 0.00000712, -0.00000190, 0.00000597, -0.00000913, -0.00001268, 0.00000300, -0.00000588, -0.00000176, -0.00000035, -0.00001031, 0.00000502, 0.00000189, -0.00000614, 0.00000261, -0.00000722, 0.00001083, -0.00000406, -0.00000172, -0.00000279, 0.00000322, -0.00000147, -0.00000851, -0.00000978, -0.00000332, -0.00001117, -0.00002639, -0.00008343, 0.00009397, 0.00007697, 0.00010377, 0.00004856, -0.00009126, -0.00001481, -0.00001737, -0.00000237, -0.00000911, -0.00000316, -0.00000305, -0.00000517, -0.00000131, -0.00000183, -0.00000191, -0.00000729, -0.00000022, -0.00000769, -0.00000516, -0.00001399, -0.00004237, -0.00000581, 0.00006026, 0.00003325, 0.00004137, 0.00005266, -0.00002253, -0.00003139, -0.00000545, -0.00000594, -0.00001036, -0.00000450, 0.00000018, 0.00000435, -0.00000585, -0.00000291, 0.00001015, -0.00000038, -0.00000858, 0.00000088, -0.00000109, 0.00000360, -0.00001048, 0.00000730, -0.00000259, -0.00000220, -0.00000683, 0.00000045, -0.00000706, -0.00000969, 0.00000783, 0.00000216, -0.00000566, -0.00000687, 0.00000721, 0.00001056, -0.00000751, -0.00000440, 0.00000230, 0.00000311, -0.00000358, -0.00000000, 0.00000946, 0.00000091, -0.00000742, -0.00000184, -0.00000211, 0.00000263, 0.00000658, 0.00000568, -0.00000944, -0.00001070, 0.00001146, 0.00000791, -0.00000521, },
		{ -0.00000514, 0.00000200, -0.00000331, 0.00000103, 0.00000154, 0.00000032, 0.00001063, -0.00000742, -0.00001088, 0.00000758, 0.00000452, 0.00000513, -0.00000438, -0.00000193, -0.00000770, 0.00000765, 0.00000339, -0.00000055, -0.00000054, -0.00000113, 0.00000686, -0.00000435, -0.00000585, 0.00000067, 0.00001582, -0.00001382, 0.00000430, 0.00000089, 0.00000262, -0.00000279, -0.00001709, 0.00000323, -0.00000504, -0.00000262, 0.00000021, -0.00001000, 0.00000168, 0.00000598, -0.00000908, 0.00000510, -0.00000931, 0.00001113, -0.00000343, -0.00000118, -0.00000409, 0.00000385, -0.00000209, -0.00000487, -0.00001396, 0.00000099, -0.00001628, -0.00001022, -0.00009557, 0.00006290, 0.00009798, 0.00009263, 0.00008034, -0.00009366, -0.00002276, -0.00001679, -0.00000428, -0.00000783, -0.00000498, -0.00000198, -0.00000554, -0.00000193, -0.00000163, -0.00000133, -0.00000709, -0.00000125, -0.00000557, -0.00000677, -0.00000957, -0.00003987, -0.00001436, 0.00005196, 0.00003964, 0.00003061, 0.00005851, -0.00000949, -0.00003307, -0.00001049, -0.00000263, -0.00001254, -0.00000339, -0.00000263, 0.00000637, -0.00000587, -0.00000336, 0.00000795, 0.00000401, -0.00001120, 0.00000178, -0.00000248, 0.00000537, -0.00001118, 0.00000601, -0.00000032, -0.00000286, -0.00000593, -0.00000157, -0.00000364, -0.00001179, 0.00000583, 0.00000431, -0.00000504, -0.00000718, 0.00000332, 0.00001286, -0.00000456, -0.00000659, 0.00000136, 0.00000377, -0.00000260, -0.00000176, 0.00000911, 0.00000292, -0.00000722, -0.00000272, -0.00000176, 0.00000122, 0.00000634, 0.00000630, -0.00000599, -0.00001321, 0.00000817, 0.00001086, -0.00000393, },
		{ -0.00000648, 0.00000150, -0.00000222, -0.00000058, 0.00000320, -0.00000136, 0.00000988, -0.00000285, -0.00001336, 0.00000585, 0.00000501, 0.00000562, -0.00000310, -0.00000238, -0.00000705, 0.00000448, 0.00000614, -0.00000200, 0.00000141, -0.00000350, 0.00000801, -0.00000376, -0.00000459, -0.00000341, 0.00001700, -0.00001019, -0.00000052, 0.00000511, -0.00000162, 0.00000302, -0.00001935, 0.00000141, -0.00000316, -0.00000380, 0.00000069, -0.00000910, -0.00000221, 0.00000980, -0.00001111, 0.00000642, -0.00000999, 0.00001005, -0.00000179, -0.00000148, -0.00000437, 0.00000350, -0.00000208, -0.00000217, -0.00001632, 0.00000283, -0.00001840, -0.00000108, -0.00009242, 0.00002032, 0.00011869, 0.00008268, 0.00010803, -0.00008770, -0.00003503, -0.00001449, -0.00000763, -0.00000571, -0.00000709, -0.00000117, -0.00000543, -0.00000289, -0.00000126, -0.00000125, -0.00000601, -0.00000321, -0.00000281, -0.00000884, -0.00000544, -0.00003728, -0.00002009, 0.00004124, 0.00004641, 0.00002178, 0.00005978, 0.00000375, -0.00003120, -0.00001691, 0.00000035, -0.00001407, -0.00000281, -0.00000471, 0.00000720, -0.00000489, -0.00000413, 0.00000586, 0.00000756, -0.00001219, 0.00000109, -0.00000253, 0.00000565, -0.00001010, 0.00000303, 0.00000295, -0.00000413, -0.00000431, -0.00000404, -0.00000070, -0.00001256, 0.00000300, 0.00000613, -0.00000404, -0.00000717, -0.00000012, 0.00001324, -0.00000036, -0.00000840, 0.00000005, 0.00000404, -0.00000122, -0.00000332, 0.00000832, 0.00000479, -0.00000656, -0.00000372, -0.00000141, 0.00000011, 0.00000572, 0.00000675, -0.00000294, -0.00001396, 0.00000371, 0.00001322, -0.00000193, },
		{ -0.00000733, 0.00000030, -0.00000078, -0.00000218, 0.00000435, -0.00000189, 0.00000792, 0.00000157, -0.00001437, 0.00000313, 0.00000576, 0.00000555, -0.00000126, -0.00000321, -0.00000582, 0.00000108, 0.00000849, -0.00000301, 0.00000288, -0.00000499, 0.00000805, -0.00000249, -0.00000384, -0.00000577, 0.00001538, -0.00000440, -0.00000626, 0.00000938, -0.00000510, 0.00000671, -0.00001856, -0.00000261, -0.00000044, -0.00000519, 0.00000109, -0.00000784, -0.00000607, 0.00001254, -0.00001146, 0.00000591, -0.00000872, 0.00000718, 0.00000128, -0.00000307, -0.00000317, 0.00000166, -0.00000074, -0.00000140, -0.00001569, 0.00000098, -0.00001602, -0.00000033, -0.00007584, -0.00002729, 0.00013240, 0.00007834, 0.00012830, -0.00007178, -0.00005158, -0.00001068, -0.00001228, -0.00000298, -0.00000929, -0.00000068, -0.00000498, -0.00000392, -0.00000095, -0.00000147, -0.00000438, -0.00000558, -0.00000005, -0.00001072, -0.00000211, -0.00003436, -0.00002367, 0.00002956, 0.00005184, 0.00001615, 0.00005671, 0.00001637, -0.00002649, -0.00002290, 0.00000142, -0.00001374, -0.00000382, -0.00000512, 0.00000613, -0.00000245, -0.00000570, 0.00000451, 0.00000951, -0.00001100, -0.00000146, -0.00000108, 0.00000437, -0.00000733, -0.00000129, 0.00000669, -0.00000554, -0.00000244, -0.00000639, 0.00000129, -0.00001201, -0.00000020, 0.00000723, -0.00000255, -0.00000705, -0.00000276, 0.00001179, 0.00000438, -0.00000925, -0.00000176, 0.00000400, 0.00000029, -0.00000441, 0.00000700, 0.00000654, -0.00000563, -0.00000460, -0.00000133, -0.00000050, 0.00000466, 0.00000717, -0.00000057, -0.00001309, -0.00000123, 0.00001445, 0.00000076, },
		{ -0.00000760, -0.00000134, 0.00000055, -0.00000326, 0.00000460, -0.00000123, 0.00000518, 0.00000527, -0.00001381, -0.00000025, 0.00000649, 0.00000518, 0.00000078, -0.00000402, -0.00000447, -0.00000201, 0.00000998, -0.00000320, 0.00000354, -0.00000534, 0.00000690, -0.00000055, -0.00000377, -0.00000634, 0.00001158, 0.00000229, -0.00001143, 0.00001216, -0.00000637, 0.00000725, -0.00001452, -0.00000830, 0.00000252, -0.00000641, 0.00000129, -0.00000637, -0.00000942, 0.00001358, -0.00000970, 0.00000335, -0.00000546, 0.00000257, 0.00000564, -0.00000585, -0.00000060, -0.00000155, 0.00000196, -0.00000275, -0.00001201, -0.00000438, -0.00000950, -0.00000656, -0.00005091, -0.00007153, 0.00013262, 0.00008273, 0.00013925, -0.00004574, -0.00007122, -0.00000628, -0.00001763, -0.00000011, -0.00001131, -0.00000051, -0.00000445, -0.00000467, -0.00000101, -0.00000166, -0.00000269, -0.00000765, 0.00000196, -0.00001168, -0.00000012, -0.00003071, -0.00002607, 0.00001849, 0.00005443, 0.00001461, 0.00004970, 0.00002798, -0.00002026, -0.00002656, -0.00000068, -0.00001084, -0.00000688, -0.00000353, 0.00000306, 0.00000141, -0.00000803, 0.00000408, 0.00000968, -0.00000770, -0.00000551, 0.00000140, 0.00000200, -0.00000348, -0.00000612, 0.00001006, -0.00000638, -0.00000096, -0.00000800, 0.00000201, -0.00001032, -0.00000336, 0.00000745, -0.00000069, -0.00000683, -0.00000454, 0.00000899, 0.00000875, -0.00000860, -0.00000411, 0.00000371, 0.00000168, -0.00000482, 0.00000515, 0.00000815, -0.00000452, -0.00000521, -0.00000162, -0.00000061, 0.00000326, 0.00000754, 0.00000115, -0.00001107, -0.00000581, 0.00001412, 0.00000404, },
		{ -0.00000726, -0.00000308, 0.00000132, -0.00000350, 0.00000382, 0.00000039, 0.00000224, 0.00000783, -0.00001181, -0.00000384, 0.00000686, 0.00000476, 0.00000267, -0.00000447, -0.00000336, -0.00000438, 0.00001035, -0.00000245, 0.00000328, -0.00000459, 0.00000480, 0.00000184, -0.00000422, -0.00000558, 0.00000669, 0.00000838, -0.00001459, 0.00001228, -0.00000465, 0.00000456, -0.00000801, -0.00001445, 0.00000474, -0.00000694, 0.00000104, -0.00000472, -0.00001194, 0.00001269, -0.00000591, -0.00000091, -0.00000074, -0.00000309, 0.00001056, -0.00000907, 0.00000257, -0.00000533, 0.00000527, -0.00000541, -0.00000642, -0.00001161, -0.00000101, -0.00001615, -0.00002453, -0.00010412, 0.00011503, 0.00009638, 0.00014122, -0.00001126, -0.00009153, -0.00000291, -0.00002264, 0.00000217, -0.00001276, -0.00000072, -0.00000404, -0.00000488, -0.00000166, -0.00000154, -0.00000144, -0.00000879, 0.00000259, -0.00001118, 0.00000010, -0.00002595, -0.00002816, 0.00000933, 0.00005316, 0.00001737, 0.00003953, 0.00003830, -0.00001375, -0.00002666, -0.00000617, -0.00000566, -0.00001157, -0.00000047, -0.00000134, 0.00000591, -0.00001054, 0.00000418, 0.00000850, -0.00000304, -0.00001011, 0.00000397, -0.00000053, 0.00000043, -0.00001033, 0.00001205, -0.00000594, -0.00000044, -0.00000841, 0.00000132, -0.00000773, -0.00000616, 0.00000685, 0.00000124, -0.00000632, -0.00000566, 0.00000556, 0.00001191, -0.00000621, -0.00000680, 0.00000311, 0.00000280, -0.00000452, 0.00000291, 0.00000949, -0.00000323, -0.00000554, -0.00000220, -0.00000042, 0.00000183, 0.00000761, 0.00000250, -0.00000859, -0.00000924, 0.00001200, 0.00000767, },
		{ -0.00000627, -0.00000461, 0.00000128, -0.00000284, 0.00000223, 0.00000245, -0.00000027, 0.00000895, -0.00000867, -0.00000715, 0.00000662, 0.00000449, 0.00000417, -0.00000436, -0.00000269, -0.00000587, 0.00000958, -0.00000088, 0.00000230, -0.00000303, 0.00000220, 0.00000421, -0.00000478, -0.00000419, 0.00000193, 0.00001258, -0.00001483, 0.00000935, -0.00000016, -0.00000041, -0.00000064, -0.00001943, 0.00000517, -0.00000624, 0.00000009, -0.00000287, -0.00001356, 0.00001005, -0.00000071, -0.00000597, 0.00000436, -0.00000858, 0.00001480, -0.00001151, 0.00000506, -0.00000837, 0.00000789, -0.00000792, -0.00000088, -0.00001821, 0.00000632, -0.00002459, -0.00000349, -0.00011910, 0.00007909, 0.00011650, 0.00013690, 0.00002823, -0.00010901, -0.00000272, -0.00002599, 0.00000293, -0.00001315, -0.00000146, -0.00000377, -0.00000451, -0.00000288, -0.00000103, -0.00000090, -0.00000866, 0.00000160, -0.00000909, -0.00000145, -0.00002009, -0.00003044, 0.00000277, 0.00004789, 0.00002379, 0.00002752, 0.00004682, -0.00000765, -0.00002311, -0.00001404, 0.00000043, -0.00001653, 0.00000273, -0.00000571, 0.00000977, -0.00001215, 0.00000393, 0.00000692, 0.00000180, -0.00001394, 0.00000544, -0.00000216, 0.00000330, -0.00001284, 0.00001184, -0.00000375, -0.00000121, -0.00000742, -0.00000066, -0.00000463, -0.00000841, 0.00000565, 0.00000286, -0.00000535, -0.00000636, 0.00000216, 0.00001328, -0.00000223, -0.00000937, 0.00000205, 0.00000365, -0.00000366, 0.00000060, 0.00001030, -0.00000167, -0.00000573, -0.00000282, -0.00000026, 0.00000073, 0.00000714, 0.00000378, -0.00000627, -0.00001101, 0.00000822, 0.00001124, },
		{ -0.00000454, -0.00000576, 0.00000044, -0.00000157, 0.00000033, 0.00000430, -0.00000181, 0.00000857, -0.00000482, -0.00000978, 0.00000565, 0.00000442, 0.00000519, -0.00000368, -0.00000246, -0.00000653, 0.00000789, 0.00000119, 0.00000094, -0.00000113, -0.00000038, 0.00000612, -0.00000504, -0.00000288, -0.00000171, 0.00001414, -0.00001205, 0.00000392, 0.00000592, -0.00000592, 0.00000559, -0.00002163, 0.00000299, -0.00000403, -0.00000168, -0.00000080, -0.00001437, 0.00000617, 0.00000493, -0.00001061, 0.00000852, -0.00001255, 0.00001700, -0.00001186, 0.00000562, -0.00000940, 0.00000850, -0.00000870, 0.00000260, -0.00002173, 0.00000964, -0.00002807, 0.00000751, -0.00011452, 0.00002880, 0.00013714, 0.00013080, 0.00006818, -0.00011974, -0.00000785, -0.00002641, 0.00000132, -0.00001201, -0.00000293, -0.00000350, -0.00000381, -0.00000436, -0.00000036, -0.00000099, -0.00000736, -0.00000074, -0.00000580, -0.00000430, -0.00001356, -0.00003285, -0.00000128, 0.00003938, 0.00003235, 0.00001556, 0.00005260, -0.00000182, -0.00001704, -0.00002218, 0.00000534, -0.00001991, 0.00000442, -0.00000849, 0.00001161, -0.00001180, 0.00000246, 0.00000593, 0.00000556, -0.00001579, 0.00000482, -0.00000203, 0.00000431, -0.00001295, 0.00000909, 0.00000020, -0.00000316, -0.00000526, -0.00000352, -0.00000150, -0.00000993, 0.00000410, 0.00000391, -0.00000389, -0.00000679, -0.00000074, 0.00001274, 0.00000276, -0.00001117, 0.00000028, 0.00000432, -0.00000250, -0.00000141, 0.00001035, 0.00000030, -0.00000594, -0.00000324, -0.00000043, 0.00000025, 0.00000598, 0.00000520, -0.00000452, -0.00001100, 0.00000331, 0.00001416, },
		{ -0.00000197, -0.00000652, -0.00000092, -0.00000022, -0.00000124, 0.00000533, -0.00000202, 0.00000684, -0.00000078, -0.00001141, 0.00000402, 0.00000446, 0.00000576, -0.00000254, -0.00000252, -0.00000654, 0.00000563, 0.00000338, -0.00000043, 0.00000070, -0.00000248, 0.00000722, -0.00000469, -0.00000215, -0.00000373, 0.00001310, -0.00000701, -0.00000265, 0.00001180, -0.00000993, 0.00000884, -0.00002002, -0.00000195, -0.00000051, -0.00000406, 0.00000136, -0.00001445, 0.00000166, 0.00000998, -0.00001368, 0.00001060, -0.00001389, 0.00001619, -0.00000932, 0.00000348, -0.00000766, 0.00000635, -0.00000682, 0.00000275, -0.00002073, 0.00000742, -0.00002480, 0.00000709, -0.00009305, -0.00002785, 0.00015048, 0.00012790, 0.00010409, -0.00012030, -0.00001978, -0.00002315, -0.00000311, -0.00000915, -0.00000529, -0.00000299, -0.00000318, -0.00000559, 0.00000006, -0.00000135, -0.00000538, -0.00000373, -0.00000212, -0.00000755, -0.00000722, -0.00003479, -0.00000363, 0.00002917, 0.00004093, 0.00000582, 0.00005447, 0.00000452, -0.00001037, -0.00002806, 0.00000694, -0.00002006, 0.00000322, -0.00000848, 0.00001050, -0.00000884, -0.00000078, 0.00000621, 0.00000742, -0.00001493, 0.00000170, 0.00000009, 0.00000328, -0.00001064, 0.00000417, 0.00000531, -0.00000570, -0.00000251, -0.00000657, 0.00000109, -0.00001056, 0.00000235, 0.00000432, -0.00000212, -0.00000692, -0.00000299, 0.00001060, 0.00000785, -0.00001147, -0.00000236, 0.00000488, -0.00000133, -0.00000279, 0.00000947, 0.00000268, -0.00000624, -0.00000332, -0.00000106, 0.00000043, 0.00000420, 0.00000671, -0.00000338, -0.00000957, -0.00000187, 0.00001576, },
		{ 0.00000144, -0.00000696, -0.00000237, 0.00000062, -0.00000193, 0.00000518, -0.00000091, 0.00000421, 0.00000291, -0.00001187, 0.00000188, 0.00000449, 0.00000598, -0.00000112, -0.00000269, -0.00000613, 0.00000313, 0.00000532, -0.00000148, 0.00000213, -0.00000385, 0.00000741, -0.00000368, -0.00000212, -0.00000418, 0.00001014, -0.00000106, -0.00000862, 0.00001564, -0.00001080, 0.00000808, -0.00001456, -0.00000893, 0.00000349, -0.00000648, 0.00000324, -0.00001380, -0.00000297, 0.00001361, -0.00001442, 0.00000999, -0.00001217, 0.00001218, -0.00000388, -0.00000123, -0.00000338, 0.00000165, -0.00000238, -0.00000043, -0.00001535, 0.00000016, -0.00001560, -0.00000251, -0.00006132, -0.00008059, 0.00014897, 0.00013189, 0.00013277, -0.00010869, -0.00003866, -0.00001649, -0.00001011, -0.00000480, -0.00000842, -0.00000209, -0.00000305, -0.00000608, -0.00000021, -0.00000151, -0.00000340, -0.00000648, 0.00000096, -0.00001017, -0.00000208, -0.00003542, -0.00000547, 0.00001916, 0.00004725, 0.00000031, 0.00005143, 0.00001227, -0.00000508, -0.00002960, 0.00000392, -0.00001625, -0.00000131, -0.00000542, 0.00000639, -0.00000347, -0.00000554, 0.00000773, 0.00000727, -0.00001148, -0.00000351, 0.00000367, 0.00000074, -0.00000664, -0.00000193, 0.00001046, -0.00000787, -0.00000003, -0.00000899, 0.00000257, -0.00001013, 0.00000042, 0.00000428, -0.00000041, -0.00000657, -0.00000466, 0.00000747, 0.00001203, -0.00000979, -0.00000582, 0.00000530, -0.00000033, -0.00000339, 0.00000773, 0.00000531, -0.00000652, -0.00000314, -0.00000204, 0.00000107, 0.00000217, 0.00000803, -0.00000257, -0.00000742, -0.00000632, 0.00001548, },
		{ 0.00000553, -0.00000713, -0.00000349, 0.00000051, -0.00000150, 0.00000385, 0.00000117, 0.00000130, 0.00000572, -0.00001111, -0.00000054, 0.00000439, 0.00000594, 0.00000040, -0.00000284, -0.00000549, 0.00000065, 0.00000681, -0.00000208, 0.00000304, -0.00000445, 0.00000680, -0.00000218, -0.00000262, -0.00000353, 0.00000634, 0.00000428, -0.00001244, 0.00001614, -0.00000783, 0.00000352, -0.00000640, -0.00001637, 0.00000670, -0.00000808, 0.00000428, -0.00001226, -0.00000740, 0.00001542, -0.00001268, 0.00000685, -0.00000780, 0.00000569, 0.00000346, -0.00000739, 0.00000227, -0.00000433, 0.00000339, -0.00000563, -0.00000733, -0.00000983, -0.00000353, -0.00001640, -0.00002803, -0.00011948, 0.00012765, 0.00014358, 0.00015335, -0.00008502, -0.00006290, -0.00000789, -0.00001857, 0.00000017, -0.00001185, -0.00000090, -0.00000356, -0.00000558, -0.00000141, -0.00000115, -0.00000199, -0.00000821, 0.00000262, -0.00001132, 0.00000104, -0.00003402, -0.00000792, 0.00001107, 0.00004952, 0.00000030, 0.00004321, 0.00002182, -0.00000241, -0.00002595, -0.00000359, -0.00000910, -0.00000838, -0.00000019, 0.00000031, 0.00000320, -0.00001085, 0.00000976, 0.00000578, -0.00000634, -0.00000964, 0.00000752, -0.00000215, -0.00000221, -0.00000776, 0.00001430, -0.00000863, 0.00000130, -0.00001005, 0.00000251, -0.00000857, -0.00000170, 0.00000410, 0.00000081, -0.00000555, -0.00000594, 0.00000406, 0.00001452, -0.00000600, -0.00000969, 0.00000535, 0.00000053, -0.00000330, 0.00000544, 0.00000782, -0.00000653, -0.00000293, -0.00000306, 0.00000175, 0.00000039, 0.00000870, -0.00000161, -0.00000538, -0.00000917, 0.00001305, },
		{ 0.00000985, -0.00000697, -0.00000405, -0.00000065, -0.00000012, 0.00000179, 0.00000355, -0.00000115, 0.00000723, -0.00000921, -0.00000301, 0.00000409, 0.00000565, 0.00000194, -0.00000288, -0.00000469, -0.00000164, 0.00000775, -0.00000221, 0.00000345, -0.00000441, 0.00000567, -0.00000050, -0.00000332, -0.00000245, 0.00000283, 0.00000781, -0.00001322, 0.00001302, -0.00000156, -0.00000344, 0.00000236, -0.00002221, 0.00000767, -0.00000799, 0.00000394, -0.00000967, -0.00001151, 0.00001543, -0.00000894, 0.00000199, -0.00000194, -0.00000184, 0.00001102, -0.00001321, 0.00000745, -0.00000974, 0.00000858, -0.00001069, 0.00000065, -0.00001920, 0.00000731, -0.00002881, -0.00000146, -0.00013756, 0.00008612, 0.00016000, 0.00016757, -0.00005155, -0.00008926, 0.00000010, -0.00002666, 0.00000443, -0.00001476, 0.00000017, -0.00000456, -0.00000422, -0.00000338, -0.00000026, -0.00000140, -0.00000855, 0.00000252, -0.00001066, 0.00000182, -0.00003030, -0.00001166, 0.00000593, 0.00004693, 0.00000584, 0.00003065, 0.00003257, -0.00000231, -0.00001791, -0.00001384, -0.00000057, -0.00001607, 0.00000536, -0.00000591, 0.00000940, -0.00001515, 0.00001104, 0.00000413, -0.00000095, -0.00001509, 0.00001016, -0.00000404, 0.00000125, -0.00001188, 0.00001565, -0.00000719, 0.00000091, -0.00000933, 0.00000080, -0.00000599, -0.00000401, 0.00000404, 0.00000125, -0.00000385, -0.00000694, 0.00000091, 0.00001493, -0.00000053, -0.00001326, 0.00000462, 0.00000139, -0.00000281, 0.00000306, 0.00000975, -0.00000591, -0.00000304, -0.00000372, 0.00000197, -0.00000061, 0.00000834, -0.00000011, -0.00000409, -0.00000993, 0.00000865, },
		{ 0.00001367, -0.00000626, -0.00000409, -0.00000257, 0.00000164, -0.00000027, 0.00000541, -0.00000247, 0.00000721, -0.00000637, -0.00000530, 0.00000361, 0.00000510, 0.00000345, -0.00000282, -0.00000377, -0.00000368, 0.00000815, -0.00000195, 0.00000349, -0.00000393, 0.00000430, 0.00000108, -0.00000390, -0.00000151, 0.00000037, 0.00000898, -0.00001101, 0.00000711, 0.00000639, -0.00001050, 0.00000925, -0.00002452, 0.00000536, -0.00000573, 0.00000197, -0.00000599, -0.00001522, 0.00001394, -0.00000404, -0.00000334, 0.00000393, -0.00000862, 0.00001689, -0.00001682, 0.00001031, -0.00001271, 0.00001130, -0.00001347, 0.00000596, -0.00002483, 0.00001326, -0.00003500, 0.00001283, -0.00013280, 0.00002927, 0.00017477, 0.00017909, -0.00001203, -0.00011351, 0.00000446, -0.00003232, 0.00000651, -0.00001618, 0.00000051, -0.00000556, -0.00000250, -0.00000562, 0.00000080, -0.00000148, -0.00000762, 0.00000087, -0.00000843, 0.00000053, -0.00002460, -0.00001661, 0.00000381, 0.00003990, 0.00001561, 0.00001580, 0.00004282, -0.00000345, -0.00000770, -0.00002397, 0.00000662, -0.00002196, 0.00000906, -0.00001026, 0.00001330, -0.00001694, 0.00001036, 0.00000355, 0.00000326, -0.00001838, 0.00001032, -0.00000385, 0.00000269, -0.00001332, 0.00001392, -0.00000336, -0.00000122, -0.00000695, -0.00000226, -0.00000279, -0.00000632, 0.00000417, 0.00000093, -0.00000171, -0.00000758, -0.00000170, 0.00001336, 0.00000574, -0.00001559, 0.00000271, 0.00000246, -0.00000228, 0.00000108, 0.00001067, -0.00000437, -0.00000371, -0.00000371, 0.00000141, -0.00000056, 0.00000679, 0.00000209, -0.00000383, -0.00000867, 0.00000297, },
		{ 0.00001613, -0.00000467, -0.00000388, -0.00000464, 0.00000295, -0.00000150, 0.00000606, -0.00000227, 0.00000572, -0.00000295, -0.00000721, 0.00000301, 0.00000426, 0.00000489, -0.00000268, -0.00000276, -0.00000540, 0.00000804, -0.00000138, 0.00000324, -0.00000317, 0.00000288, 0.00000241, -0.00000419, -0.00000101, -0.00000077, 0.00000803, -0.00000676, 0.00000008, 0.00001375, -0.00001521, 0.00001216, -0.00002211, -0.00000044, -0.00000149, -0.00000138, -0.00000158, -0.00001831, 0.00001128, 0.00000112, -0.00000796, 0.00000842, -0.00001315, 0.00001960, -0.00001689, 0.00000967, -0.00001215, 0.00001043, -0.00001268, 0.00000699, -0.00002491, 0.00001245, -0.00003287, 0.00001341, -0.00010869, -0.00003349, 0.00017964, 0.00019198, 0.00002954, -0.00013147, 0.00000251, -0.00003384, 0.00000525, -0.00001539, -0.00000044, -0.00000601, -0.00000111, -0.00000741, 0.00000147, -0.00000178, -0.00000594, -0.00000165, -0.00000538, -0.00000200, -0.00001782, -0.00002207, 0.00000391, 0.00002995, 0.00002725, 0.00000153, 0.00005023, -0.00000366, 0.00000172, -0.00003083, 0.00000987, -0.00002397, 0.00000916, -0.00001121, 0.00001363, -0.00001529, 0.00000702, 0.00000477, 0.00000535, -0.00001863, 0.00000742, -0.00000121, 0.00000178, -0.00001190, 0.00000933, 0.00000230, -0.00000448, -0.00000354, -0.00000591, 0.00000039, -0.00000825, 0.00000430, 0.00000017, 0.00000038, -0.00000761, -0.00000380, 0.00001036, 0.00001166, -0.00001581, -0.00000063, 0.00000387, -0.00000200, -0.00000012, 0.00001036, -0.00000187, -0.00000500, -0.00000297, 0.00000003, 0.00000051, 0.00000429, 0.00000476, -0.00000439, -0.00000602, -0.00000289, },
		{ 0.00001645, -0.00000196, -0.00000376, -0.00000617, 0.00000307, -0.00000129, 0.00000514, -0.00000055, 0.00000314, 0.00000056, -0.00000855, 0.00000232, 0.00000317, 0.00000617, -0.00000243, -0.00000173, -0.00000672, 0.00000740, -0.00000053, 0.00000272, -0.00000217, 0.00000146, 0.00000351, -0.00000424, -0.00000096, -0.00000082, 0.00000577, -0.00000194, -0.00000611, 0.00001839, -0.00001573, 0.00001006, -0.00001512, -0.00000880, 0.00000370, -0.00000524, 0.00000282, -0.00002035, 0.00000763, 0.00000587, -0.00001106, 0.00001070, -0.00001464, 0.00001856, -0.00001314, 0.00000548, -0.00000814, 0.00000613, -0.00000839, 0.00000374, -0.00001965, 0.00000544, -0.00002344, 0.00000310, -0.00007314, -0.00009048, 0.00016703, 0.00020877, 0.00007027, -0.00014005, -0.00000720, -0.00003059, 0.00000030, -0.00001224, -0.00000286, -0.00000553, -0.00000062, -0.00000814, 0.00000120, -0.00000178, -0.00000419, -0.00000416, -0.00000245, -0.00000475, -0.00001111, -0.00002694, 0.00000484, 0.00001920, 0.00003788, -0.00000914, 0.00005252, -0.00000088, 0.00000773, -0.00003208, 0.00000768, -0.00002116, 0.00000507, -0.00000839, 0.00001025, -0.00001031, 0.00000128, 0.00000770, 0.00000523, -0.00001588, 0.00000184, 0.00000335, -0.00000095, -0.00000832, 0.00000293, 0.00000857, -0.00000775, -0.00000013, -0.00000918, 0.00000272, -0.00000930, 0.00000406, -0.00000048, 0.00000179, -0.00000672, -0.00000556, 0.00000669, 0.00001612, -0.00001338, -0.00000529, 0.00000545, -0.00000203, -0.00000046, 0.00000893, 0.00000135, -0.00000665, -0.00000176, -0.00000188, 0.00000215, 0.00000141, 0.00000733, -0.00000518, -0.00000297, -0.00000768, },
		{ 0.00001421, 0.00000188, -0.00000400, -0.00000661, 0.00000155, 0.00000049, 0.00000284, 0.00000218, 0.00000017, 0.00000357, -0.00000909, 0.00000151, 0.00000196, 0.00000712, -0.00000194, -0.00000084, -0.00000746, 0.00000617, 0.00000065, 0.00000189, -0.00000092, -0.00000001, 0.00000452, -0.00000419, -0.00000114, -0.00000032, 0.00000329, 0.00000194, -0.00000983, 0.00001896, -0.00001147, 0.00000345, -0.00000508, -0.00001774, 0.00000816, -0.00000836, 0.00000617, -0.00002071, 0.00000301, 0.00000989, -0.00001237, 0.00001068, -0.00001320, 0.00001423, -0.00000643, -0.00000110, -0.00000198, -0.00000020, -0.00000202, -0.00000222, -0.00001106, -0.00000503, -0.00001030, -0.00001234, -0.00003617, -0.00013087, 0.00013277, 0.00022874, 0.00010940, -0.00013803, -0.00002440, -0.00002330, -0.00000759, -0.00000733, -0.00000642, -0.00000412, -0.00000125, -0.00000752, -0.00000020, -0.00000119, -0.00000293, -0.00000593, -0.00000044, -0.00000680, -0.00000555, -0.00003013, 0.00000518, 0.00000980, 0.00004495, -0.00001392, 0.00004831, 0.00000603, 0.00000893, -0.00002702, 0.00000030, -0.00001420, -0.00000234, -0.00000274, 0.00000422, -0.00000322, -0.00000569, 0.00001138, 0.00000368, -0.00001108, -0.00000518, 0.00000852, -0.00000424, -0.00000396, -0.00000367, 0.00001385, -0.00000973, 0.00000214, -0.00001107, 0.00000349, -0.00000904, 0.00000303, -0.00000045, 0.00000202, -0.00000476, -0.00000716, 0.00000306, 0.00001833, -0.00000831, -0.00001070, 0.00000678, -0.00000220, -0.00000018, 0.00000678, 0.00000476, -0.00000815, -0.00000057, -0.00000374, 0.00000370, -0.00000107, 0.00000907, -0.00000543, -0.00000059, -0.00001037, },
		{ 0.00000955, 0.00000645, -0.00000455, -0.00000579, -0.00000151, 0.00000338, -0.00000013, 0.00000506, -0.00000233, 0.00000551, -0.00000866, 0.00000053, 0.00000086, 0.00000749, -0.00000107, -0.00000030, -0.00000744, 0.00000428, 0.00000221, 0.00000069, 0.00000064, -0.00000161, 0.00000556, -0.00000420, -0.00000131, 0.00000016, 0.00000154, 0.00000387, -0.00001033, 0.00001537, -0.00000335, -0.00000572, 0.00000537, -0.00002473, 0.00001004, -0.00000941, 0.00000746, -0.00001884, -0.00000265, 0.00001314, -0.00001217, 0.00000890, -0.00000971, 0.00000792, 0.00000151, -0.00000814, 0.00000426, -0.00000643, 0.00000422, -0.00000839, -0.00000223, -0.00001521, 0.00000186, -0.00002632, -0.00000697, -0.00014766, 0.00007807, 0.00024724, 0.00014842, -0.00012605, -0.00004712, -0.00001401, -0.00001677, -0.00000199, -0.00001026, -0.00000226, -0.00000276, -0.00000578, -0.00000252, -0.00000007, -0.00000234, -0.00000660, 0.00000026, -0.00000767, -0.00000177, -0.00003099, 0.00000393, 0.00000325, 0.00004688, -0.00001190, 0.00003774, 0.00001668, 0.00000571, -0.00001691, -0.00001026, -0.00000526, -0.00001098, 0.00000372, -0.00000248, 0.00000405, -0.00001209, 0.00001434, 0.00000202, -0.00000574, -0.00001194, 0.00001264, -0.00000656, -0.00000035, -0.00000884, 0.00001672, -0.00000936, 0.00000248, -0.00001094, 0.00000233, -0.00000728, 0.00000101, 0.00000055, 0.00000091, -0.00000188, -0.00000854, -0.00000008, 0.00001804, -0.00000124, -0.00001583, 0.00000721, -0.00000210, 0.00000028, 0.00000454, 0.00000768, -0.00000889, 0.00000001, -0.00000493, 0.00000447, -0.00000243, 0.00000937, -0.00000453, 0.00000033, -0.00001040, },
		{ 0.00000327, 0.00001095, -0.00000508, -0.00000398, -0.00000546, 0.00000643, -0.00000272, 0.00000708, -0.00000357, 0.00000597, -0.00000718, -0.00000072, 0.00000015, 0.00000707, 0.00000033, -0.00000028, -0.00000656, 0.00000173, 0.00000411, -0.00000083, 0.00000243, -0.00000329, 0.00000660, -0.00000427, -0.00000137, 0.00000027, 0.00000099, 0.00000361, -0.00000792, 0.00000878, 0.00000648, -0.00001459, 0.00001324, -0.00002743, 0.00000796, -0.00000759, 0.00000614, -0.00001455, -0.00000921, 0.00001565, -0.00001091, 0.00000621, -0.00000536, 0.00000127, 0.00000873, -0.00001357, 0.00000853, -0.00001051, 0.00000819, -0.00001239, 0.00000396, -0.00002167, 0.00000909, -0.00003378, 0.00000875, -0.00013975, 0.00001002, 0.00025635, 0.00019010, -0.00010586, -0.00007233, -0.00000544, -0.00002511, 0.00000219, -0.00001331, -0.00000068, -0.00000455, -0.00000352, -0.00000515, 0.00000114, -0.00000224, -0.00000629, -0.00000020, -0.00000738, 0.00000016, -0.00002945, 0.00000076, 0.00000013, 0.00004344, -0.00000379, 0.00002262, 0.00002910, 0.00000011, -0.00000458, -0.00002080, 0.00000272, -0.00001821, 0.00000861, -0.00000765, 0.00000942, -0.00001614, 0.00001513, 0.00000156, -0.00000135, -0.00001680, 0.00001426, -0.00000671, 0.00000128, -0.00001139, 0.00001633, -0.00000626, 0.00000068, -0.00000876, -0.00000059, -0.00000429, -0.00000181, 0.00000242, -0.00000125, 0.00000135, -0.00000942, -0.00000262, 0.00001557, 0.00000666, -0.00001949, 0.00000607, -0.00000134, 0.00000041, 0.00000283, 0.00000950, -0.00000837, -0.00000047, -0.00000500, 0.00000398, -0.00000227, 0.00000798, -0.00000225, -0.00000047, -0.00000794, },
		{ -0.00000335, 0.00001434, -0.00000504, -0.00000182, -0.00000927, 0.00000845, -0.00000386, 0.00000740, -0.00000304, 0.00000483, -0.00000478, -0.00000223, -0.00000002, 0.00000575, 0.00000227, -0.00000081, -0.00000487, -0.00000131, 0.00000618, -0.00000249, 0.00000424, -0.00000484, 0.00000746, -0.00000420, -0.00000142, -0.00000000, 0.00000154, 0.00000171, -0.00000383, 0.00000124, 0.00001530, -0.00002025, 0.00001610, -0.00002449, 0.00000162, -0.00000301, 0.00000243, -0.00000830, -0.00001608, 0.00001725, -0.00000896, 0.00000333, -0.00000121, -0.00000433, 0.00001368, -0.00001594, 0.00000953, -0.00001121, 0.00000870, -0.00001284, 0.00000587, -0.00002264, 0.00000955, -0.00003272, 0.00001009, -0.00011221, -0.00005968, 0.00024711, 0.00023663, -0.00007900, -0.00009700, -0.00000015, -0.00003078, 0.00000389, -0.00001468, -0.00000008, -0.00000595, -0.00000151, -0.00000734, 0.00000187, -0.00000220, -0.00000543, -0.00000133, -0.00000639, 0.00000063, -0.00002599, -0.00000393, 0.00000003, 0.00003567, 0.00000828, 0.00000597, 0.00004032, -0.00000492, 0.00000651, -0.00002796, 0.00000704, -0.00002180, 0.00001007, -0.00000966, 0.00001148, -0.00001672, 0.00001295, 0.00000306, 0.00000112, -0.00001878, 0.00001268, -0.00000418, 0.00000051, -0.00001101, 0.00001278, -0.00000086, -0.00000268, -0.00000514, -0.00000454, -0.00000079, -0.00000488, 0.00000464, -0.00000370, 0.00000412, -0.00000933, -0.00000474, 0.00001168, 0.00001397, -0.00002055, 0.00000292, 0.00000031, -0.00000013, 0.00000208, 0.00000989, -0.00000641, -0.00000212, -0.00000385, 0.00000220, -0.00000065, 0.00000515, 0.00000109, -0.00000261, -0.00000383, },
		{ -0.00000879, 0.00001560, -0.00000386, -0.00000003, -0.00001187, 0.00000842, -0.00000282, 0.00000566, -0.00000075, 0.00000240, -0.00000181, -0.00000389, 0.00000039, 0.00000364, 0.00000456, -0.00000173, -0.00000261, -0.00000447, 0.00000804, -0.00000391, 0.00000568, -0.00000590, 0.00000778, -0.00000365, -0.00000176, -0.00000038, 0.00000266, -0.00000073, 0.00000023, -0.00000502, 0.00002066, -0.00002067, 0.00001294, -0.00001618, -0.00000789, 0.00000312, -0.00000259, -0.00000126, -0.00002223, 0.00001742, -0.00000636, 0.00000059, 0.00000219, -0.00000818, 0.00001572, -0.00001489, 0.00000718, -0.00000867, 0.00000597, -0.00000993, 0.00000367, -0.00001853, 0.00000413, -0.00002462, 0.00000077, -0.00007467, -0.00011772, 0.00021255, 0.00028746, -0.00004551, -0.00011907, 0.00000032, -0.00003291, 0.00000261, -0.00001409, -0.00000072, -0.00000654, -0.00000031, -0.00000849, 0.00000167, -0.00000184, -0.00000451, -0.00000253, -0.00000528, 0.00000025, -0.00002141, -0.00000933, 0.00000187, 0.00002545, 0.00002139, -0.00000873, 0.00004728, -0.00000659, 0.00001343, -0.00002943, 0.00000620, -0.00002078, 0.00000746, -0.00000806, 0.00000996, -0.00001382, 0.00000798, 0.00000639, 0.00000160, -0.00001783, 0.00000815, 0.00000053, -0.00000216, -0.00000833, 0.00000704, 0.00000559, -0.00000644, -0.00000121, -0.00000841, 0.00000223, -0.00000737, 0.00000637, -0.00000550, 0.00000553, -0.00000785, -0.00000672, 0.00000728, 0.00001940, -0.00001833, -0.00000217, 0.00000270, -0.00000136, 0.00000232, 0.00000892, -0.00000323, -0.00000464, -0.00000182, -0.00000047, 0.00000188, 0.00000161, 0.00000472, -0.00000521, 0.00000063, },
		{ -0.00001180, 0.00001409, -0.00000128, 0.00000079, -0.00001245, 0.00000581, 0.00000047, 0.00000217, 0.00000272, -0.00000062, 0.00000115, -0.00000540, 0.00000119, 0.00000109, 0.00000680, -0.00000268, -0.00000026, -0.00000722, 0.00000921, -0.00000464, 0.00000632, -0.00000605, 0.00000716, -0.00000226, -0.00000268, -0.00000051, 0.00000364, -0.00000254, 0.00000275, -0.00000831, 0.00002118, -0.00001538, 0.00000452, -0.00000441, -0.00001819, 0.00000875, -0.00000718, 0.00000493, -0.00002630, 0.00001543, -0.00000283, -0.00000216, 0.00000493, -0.00001035, 0.00001520, -0.00001119, 0.00000263, -0.00000423, 0.00000147, -0.00000521, -0.00000084, -0.00001168, -0.00000412, -0.00001358, -0.00001263, -0.00003833, -0.00015287, 0.00015078, 0.00033771, -0.00000318, -0.00013792, -0.00000415, -0.00003185, -0.00000113, -0.00001206, -0.00000231, -0.00000636, -0.00000008, -0.00000847, 0.00000049, -0.00000108, -0.00000381, -0.00000338, -0.00000445, -0.00000046, -0.00001647, -0.00001454, 0.00000439, 0.00001490, 0.00003267, -0.00001858, 0.00004786, -0.00000327, 0.00001476, -0.00002467, 0.00000053, -0.00001582, 0.00000160, -0.00000374, 0.00000583, -0.00000853, 0.00000141, 0.00001054, 0.00000086, -0.00001483, 0.00000183, 0.00000615, -0.00000548, -0.00000469, 0.00000071, 0.00001144, -0.00000915, 0.00000174, -0.00001101, 0.00000380, -0.00000848, 0.00000685, -0.00000581, 0.00000499, -0.00000484, -0.00000869, 0.00000313, 0.00002209, -0.00001285, -0.00000856, 0.00000527, -0.00000295, 0.00000323, 0.00000707, 0.00000054, -0.00000738, 0.00000040, -0.00000325, 0.00000442, -0.00000164, 0.00000766, -0.00000722, 0.00000408, },
		{ -0.00001167, 0.00000983, 0.00000251, 0.00000051, -0.00001078, 0.00000082, 0.00000531, -0.00000210, 0.00000632, -0.00000330, 0.00000344, -0.00000638, 0.00000201, -0.00000137, 0.00000846, -0.00000319, 0.00000166, -0.00000900, 0.00000922, -0.00000427, 0.00000579, -0.00000499, 0.00000536, 0.00000013, -0.00000431, -0.00000014, 0.00000390, -0.00000287, 0.00000286, -0.00000803, 0.00001698, -0.00000567, -0.00000668, 0.00000765, -0.00002629, 0.00001158, -0.00000955, 0.00000865, -0.00002706, 0.00001067, 0.00000196, -0.00000527, 0.00000742, -0.00001145, 0.00001313, -0.00000635, -0.00000231, 0.00000010, -0.00000267, -0.00000094, -0.00000513, -0.00000520, -0.00001133, -0.00000439, -0.00002327, -0.00001255, -0.00015935, 0.00006685, 0.00037794, 0.00005197, -0.00015402, -0.00001254, -0.00002901, -0.00000600, -0.00000971, -0.00000408, -0.00000586, -0.00000051, -0.00000757, -0.00000131, -0.00000017, -0.00000330, -0.00000380, -0.00000400, -0.00000127, -0.00001168, -0.00001888, 0.00000651, 0.00000573, 0.00004004, -0.00002195, 0.00004157, 0.00000491, 0.00001108, -0.00001517, -0.00000792, -0.00000899, -0.00000550, 0.00000139, 0.00000093, -0.00000271, -0.00000500, 0.00001399, 0.00000017, -0.00001112, -0.00000467, 0.00001103, -0.00000795, -0.00000161, -0.00000453, 0.00001515, -0.00000962, 0.00000275, -0.00001154, 0.00000332, -0.00000776, 0.00000560, -0.00000424, 0.00000242, -0.00000065, -0.00001051, -0.00000034, 0.00002183, -0.00000486, -0.00001507, 0.00000717, -0.00000428, 0.00000419, 0.00000506, 0.00000411, -0.00000954, 0.00000201, -0.00000532, 0.00000609, -0.00000369, 0.00000906, -0.00000772, 0.00000551, },
		{ -0.00000856, 0.00000358, 0.00000677, -0.00000058, -0.00000727, -0.00000556, 0.00001037, -0.00000579, 0.00000883, -0.00000468, 0.00000447, -0.00000649, 0.00000245, -0.00000318, 0.00000902, -0.00000283, 0.00000265, -0.00000938, 0.00000779, -0.00000260, 0.00000398, -0.00000271, 0.00000245, 0.00000339, -0.00000646, 0.00000066, 0.00000330, -0.00000154, 0.00000065, -0.00000479, 0.00000958, 0.00000581, -0.00001725, 0.00001652, -0.00002943, 0.00000989, -0.00000848, 0.00000887, -0.00002388, 0.00000301, 0.00000801, -0.00000893, 0.00001004, -0.00001211, 0.00001061, -0.00000190, -0.00000588, 0.00000254, -0.00000466, 0.00000104, -0.00000706, -0.00000170, -0.00001447, -0.00000060, -0.00002660, -0.00000206, -0.00013875, -0.00002730, 0.00039574, 0.00012391, -0.00016764, -0.00002354, -0.00002607, -0.00001050, -0.00000823, -0.00000521, -0.00000565, -0.00000107, -0.00000641, -0.00000309, 0.00000040, -0.00000269, -0.00000403, -0.00000368, -0.00000224, -0.00000726, -0.00002204, 0.00000761, -0.00000101, 0.00004257, -0.00001883, 0.00002969, 0.00001607, 0.00000464, -0.00000389, -0.00001605, -0.00000303, -0.00001141, 0.00000516, -0.00000277, 0.00000177, -0.00000952, 0.00001536, 0.00000071, -0.00000803, -0.00000984, 0.00001375, -0.00000836, -0.00000029, -0.00000747, 0.00001579, -0.00000732, 0.00000149, -0.00000980, 0.00000078, -0.00000532, 0.00000277, -0.00000103, -0.00000158, 0.00000391, -0.00001169, -0.00000313, 0.00001906, 0.00000429, -0.00002026, 0.00000749, -0.00000475, 0.00000454, 0.00000362, 0.00000672, -0.00001041, 0.00000236, -0.00000603, 0.00000623, -0.00000395, 0.00000847, -0.00000634, 0.00000457, },
		{ -0.00000339, -0.00000329, 0.00001040, -0.00000176, -0.00000287, -0.00001190, 0.00001405, -0.00000752, 0.00000925, -0.00000415, 0.00000393, -0.00000556, 0.00000220, -0.00000391, 0.00000817, -0.00000138, 0.00000248, -0.00000821, 0.00000492, 0.00000024, 0.00000110, 0.00000049, -0.00000118, 0.00000700, -0.00000862, 0.00000150, 0.00000211, 0.00000102, -0.00000297, -0.00000017, 0.00000143, 0.00001590, -0.00002388, 0.00001948, -0.00002614, 0.00000325, -0.00000395, 0.00000565, -0.00001716, -0.00000681, 0.00001466, -0.00001279, 0.00001267, -0.00001255, 0.00000828, 0.00000123, -0.00000715, 0.00000230, -0.00000382, 0.00000006, -0.00000585, -0.00000210, -0.00001258, -0.00000304, -0.00002186, -0.00000579, -0.00009968, -0.00011514, 0.00037903, 0.00021398, -0.00017722, -0.00003676, -0.00002404, -0.00001375, -0.00000828, -0.00000530, -0.00000605, -0.00000137, -0.00000549, -0.00000430, 0.00000022, -0.00000169, -0.00000435, -0.00000318, -0.00000356, -0.00000321, -0.00002397, 0.00000751, -0.00000504, 0.00004047, -0.00001062, 0.00001485, 0.00002724, -0.00000148, 0.00000580, -0.00002090, -0.00000023, -0.00001425, 0.00000604, -0.00000393, 0.00000370, -0.00001117, 0.00001396, 0.00000303, -0.00000629, -0.00001283, 0.00001365, -0.00000625, -0.00000113, -0.00000780, 0.00001337, -0.00000262, -0.00000155, -0.00000639, -0.00000311, -0.00000186, -0.00000096, 0.00000299, -0.00000591, 0.00000768, -0.00001159, -0.00000551, 0.00001467, 0.00001299, -0.00002280, 0.00000559, -0.00000395, 0.00000385, 0.00000324, 0.00000791, -0.00000963, 0.00000118, -0.00000517, 0.00000472, -0.00000240, 0.00000606, -0.00000331, 0.00000164, },
		{ 0.00000236, -0.00000913, 0.00001224, -0.00000224, 0.00000130, -0.00001666, 0.00001496, -0.00000635, 0.00000711, -0.00000163, 0.00000196, -0.00000370, 0.00000117, -0.00000337, 0.00000590, 0.00000107, 0.00000124, -0.00000569, 0.00000100, 0.00000377, -0.00000227, 0.00000396, -0.00000480, 0.00001021, -0.00001006, 0.00000180, 0.00000089, 0.00000391, -0.00000659, 0.00000389, -0.00000499, 0.00002187, -0.00002439, 0.00001554, -0.00001680, -0.00000707, 0.00000272, 0.00000023, -0.00000838, -0.00001713, 0.00002056, -0.00001586, 0.00001455, -0.00001228, 0.00000600, 0.00000308, -0.00000638, -0.00000009, -0.00000094, -0.00000294, -0.00000249, -0.00000531, -0.00000718, -0.00000956, -0.00001214, -0.00001774, -0.00005524, -0.00017972, 0.00032014, 0.00031816, -0.00017791, -0.00005359, -0.00002257, -0.00001604, -0.00000947, -0.00000476, -0.00000676, -0.00000149, -0.00000496, -0.00000476, -0.00000078, -0.00000030, -0.00000487, -0.00000234, -0.00000523, 0.00000038, -0.00002469, 0.00000625, -0.00000657, 0.00003476, 0.00000043, 0.00000017, 0.00003541, -0.00000450, 0.00001135, -0.00002070, -0.00000157, -0.00001349, 0.00000374, -0.00000242, 0.00000305, -0.00001005, 0.00001012, 0.00000677, -0.00000576, -0.00001367, 0.00001098, -0.00000212, -0.00000358, -0.00000610, 0.00000881, 0.00000328, -0.00000516, -0.00000245, -0.00000718, 0.00000148, -0.00000450, 0.00000663, -0.00000922, 0.00000952, -0.00000968, -0.00000783, 0.00000968, 0.00001977, -0.00002181, 0.00000135, -0.00000190, 0.00000203, 0.00000399, 0.00000767, -0.00000733, -0.00000131, -0.00000304, 0.00000195, 0.00000039, 0.00000255, 0.00000056, -0.00000225, },
		{ 0.00000709, -0.00001250, 0.00001149, -0.00000146, 0.00000427, -0.00001866, 0.00001235, -0.00000217, 0.00000276, 0.00000230, -0.00000083, -0.00000135, -0.00000044, -0.00000176, 0.00000258, 0.00000408, -0.00000067, -0.00000237, -0.00000327, 0.00000726, -0.00000538, 0.00000691, -0.00000762, 0.00001221, -0.00001008, 0.00000100, 0.00000024, 0.00000618, -0.00000885, 0.00000578, -0.00000794, 0.00002237, -0.00001845, 0.00000579, -0.00000371, -0.00001833, 0.00000911, -0.00000525, 0.00000024, -0.00002575, 0.00002390, -0.00001671, 0.00001447, -0.00001033, 0.00000298, 0.00000454, -0.00000483, -0.00000296, 0.00000208, -0.00000587, 0.00000078, -0.00000872, -0.00000153, -0.00001601, -0.00000286, -0.00002951, -0.00001903, -0.00020841, 0.00021953, 0.00042544, -0.00016132, -0.00007714, -0.00001993, -0.00001890, -0.00001041, -0.00000483, -0.00000688, -0.00000203, -0.00000440, -0.00000485, -0.00000213, 0.00000104, -0.00000532, -0.00000137, -0.00000695, 0.00000319, -0.00002411, 0.00000387, -0.00000602, 0.00002672, 0.00001199, -0.00001163, 0.00003847, -0.00000297, 0.00001186, -0.00001555, -0.00000625, -0.00001011, -0.00000066, 0.00000065, 0.00000088, -0.00000732, 0.00000510, 0.00001076, -0.00000555, -0.00001322, 0.00000681, 0.00000280, -0.00000644, -0.00000365, 0.00000364, 0.00000875, -0.00000790, 0.00000066, -0.00001014, 0.00000355, -0.00000675, 0.00000872, -0.00001040, 0.00000872, -0.00000585, -0.00001021, 0.00000493, 0.00002364, -0.00001719, -0.00000468, 0.00000088, -0.00000054, 0.00000552, 0.00000645, -0.00000406, -0.00000447, -0.00000034, -0.00000128, 0.00000350, -0.00000100, 0.00000417, -0.00000584, },
		{ 0.00000959, -0.00001255, 0.00000799, 0.00000063, 0.00000556, -0.00001745, 0.00000640, 0.00000423, -0.00000271, 0.00000652, -0.00000351, 0.00000085, -0.00000220, 0.00000045, -0.00000112, 0.00000697, -0.00000258, 0.00000098, -0.00000704, 0.00000987, -0.00000744, 0.00000860, -0.00000892, 0.00001240, -0.00000826, -0.00000115, 0.00000051, 0.00000719, -0.00000893, 0.00000474, -0.00000696, 0.00001773, -0.00000774, -0.00000683, 0.00000946, -0.00002709, 0.00001250, -0.00000849, 0.00000636, -0.00003051, 0.00002302, -0.00001405, 0.00001126, -0.00000571, -0.00000171, 0.00000681, -0.00000415, -0.00000432, 0.00000302, -0.00000638, 0.00000143, -0.00000945, 0.00000085, -0.00001810, 0.00000069, -0.00003361, -0.00000088, -0.00019688, 0.00008772, 0.00051825, -0.00011689, -0.00011104, -0.00001392, -0.00002439, -0.00000923, -0.00000709, -0.00000521, -0.00000388, -0.00000307, -0.00000533, -0.00000302, 0.00000160, -0.00000510, -0.00000083, -0.00000809, 0.00000468, -0.00002196, 0.00000034, -0.00000374, 0.00001754, 0.00002224, -0.00001896, 0.00003592, 0.00000275, 0.00000839, -0.00000739, -0.00001197, -0.00000627, -0.00000510, 0.00000343, -0.00000107, -0.00000469, 0.00000061, 0.00001351, -0.00000447, -0.00001263, 0.00000256, 0.00000699, -0.00000831, -0.00000188, -0.00000057, 0.00001229, -0.00000852, 0.00000189, -0.00001105, 0.00000355, -0.00000699, 0.00000857, -0.00000892, 0.00000524, -0.00000054, -0.00001241, 0.00000085, 0.00002433, -0.00000962, -0.00001135, 0.00000351, -0.00000319, 0.00000714, 0.00000494, -0.00000060, -0.00000746, 0.00000203, -0.00000402, 0.00000590, -0.00000354, 0.00000649, -0.00000801, },
		{ 0.00000934, -0.00000935, 0.00000240, 0.00000352, 0.00000529, -0.00001341, -0.00000178, 0.00001126, -0.00000769, 0.00000968, -0.00000508, 0.00000223, -0.00000361, 0.00000264, -0.00000439, 0.00000897, -0.00000383, 0.00000360, -0.00000955, 0.00001094, -0.00000787, 0.00000854, -0.00000835, 0.00001056, -0.00000463, -0.00000447, 0.00000161, 0.00000684, -0.00000679, 0.00000107, -0.00000294, 0.00000983, 0.00000468, -0.00001845, 0.00001881, -0.00003027, 0.00001086, -0.00000790, 0.00000842, -0.00003009, 0.00001711, -0.00000732, 0.00000448, 0.00000191, -0.00000842, 0.00001058, -0.00000545, -0.00000281, 0.00000047, -0.00000306, -0.00000202, -0.00000575, -0.00000218, -0.00001336, -0.00000433, -0.00002642, -0.00000382, -0.00015086, -0.00005555, 0.00057533, -0.00003506, -0.00015711, -0.00000344, -0.00003382, -0.00000468, -0.00001257, -0.00000101, -0.00000765, -0.00000035, -0.00000694, -0.00000262, 0.00000063, -0.00000360, -0.00000134, -0.00000794, 0.00000425, -0.00001796, -0.00000444, -0.00000004, 0.00000810, 0.00003011, -0.00002149, 0.00002879, 0.00001076, 0.00000341, 0.00000075, -0.00001585, -0.00000440, -0.00000752, 0.00000410, -0.00000121, -0.00000365, -0.00000194, 0.00001385, -0.00000163, -0.00001285, -0.00000064, 0.00000930, -0.00000817, -0.00000175, -0.00000277, 0.00001303, -0.00000651, 0.00000087, -0.00000965, 0.00000137, -0.00000519, 0.00000620, -0.00000508, -0.00000015, 0.00000521, -0.00001384, -0.00000255, 0.00002223, -0.00000044, -0.00001718, 0.00000500, -0.00000517, 0.00000813, 0.00000391, 0.00000227, -0.00000949, 0.00000330, -0.00000551, 0.00000686, -0.00000438, 0.00000693, -0.00000816, },
		{ 0.00000665, -0.00000379, -0.00000398, 0.00000621, 0.00000411, -0.00000759, -0.00001044, 0.00001700, -0.00001055, 0.00001060, -0.00000482, 0.00000235, -0.00000425, 0.00000419, -0.00000650, 0.00000949, -0.00000388, 0.00000494, -0.00001029, 0.00001012, -0.00000645, 0.00000666, -0.00000598, 0.00000697, 0.00000031, -0.00000834, 0.00000303, 0.00000551, -0.00000308, -0.00000401, 0.00000228, 0.00000140, 0.00001527, -0.00002544, 0.00002146, -0.00002630, 0.00000370, -0.00000328, 0.00000618, -0.00002457, 0.00000670, 0.00000274, -0.00000513, 0.00001171, -0.00001645, 0.00001548, -0.00000867, 0.00000155, -0.00000538, 0.00000375, -0.00000913, 0.00000196, -0.00001020, -0.00000245, -0.00001683, -0.00000959, -0.00002322, -0.00008502, -0.00018510, 0.00057682, 0.00008854, -0.00021266, 0.00000968, -0.00004646, 0.00000285, -0.00002091, 0.00000536, -0.00001309, 0.00000369, -0.00000983, -0.00000065, -0.00000209, -0.00000065, -0.00000319, -0.00000611, 0.00000168, -0.00001213, -0.00001031, 0.00000468, -0.00000092, 0.00003521, -0.00001997, 0.00001917, 0.00001836, -0.00000018, 0.00000595, -0.00001559, -0.00000606, -0.00000673, 0.00000177, 0.00000116, -0.00000482, -0.00000201, 0.00001150, 0.00000306, -0.00001408, -0.00000236, 0.00000935, -0.00000582, -0.00000344, -0.00000279, 0.00001106, -0.00000224, -0.00000189, -0.00000650, -0.00000235, -0.00000202, 0.00000238, 0.00000008, -0.00000602, 0.00001000, -0.00001377, -0.00000555, 0.00001823, 0.00000871, -0.00002079, 0.00000458, -0.00000593, 0.00000794, 0.00000384, 0.00000405, -0.00001009, 0.00000310, -0.00000543, 0.00000613, -0.00000339, 0.00000551, -0.00000639, },
		{ 0.00000252, 0.00000258, -0.00000949, 0.00000757, 0.00000289, -0.00000140, -0.00001768, 0.00001970, -0.00001011, 0.00000870, -0.00000254, 0.00000115, -0.00000397, 0.00000474, -0.00000704, 0.00000826, -0.00000258, 0.00000478, -0.00000915, 0.00000749, -0.00000343, 0.00000334, -0.00000234, 0.00000233, 0.00000566, -0.00001186, 0.00000403, 0.00000395, 0.00000111, -0.00000891, 0.00000659, -0.00000491, 0.00002125, -0.00002564, 0.00001655, -0.00001578, -0.00000751, 0.00000389, 0.00000097, -0.00001554, -0.00000621, 0.00001408, -0.00001551, 0.00002161, -0.00002385, 0.00001991, -0.00001236, 0.00000711, -0.00001257, 0.00001178, -0.00001740, 0.00001100, -0.00002013, 0.00001083, -0.00003197, 0.00001048, -0.00004857, -0.00001903, -0.00027626, 0.00051016, 0.00024910, -0.00026842, 0.00001965, -0.00005875, 0.00001069, -0.00002990, 0.00001197, -0.00001867, 0.00000795, -0.00001324, 0.00000228, -0.00000596, 0.00000322, -0.00000602, -0.00000289, -0.00000262, -0.00000507, -0.00001667, 0.00000971, -0.00000880, 0.00003745, -0.00001558, 0.00000934, 0.00002318, -0.00000027, 0.00000657, -0.00001050, -0.00001122, -0.00000306, -0.00000302, 0.00000538, -0.00000758, -0.00000026, 0.00000728, 0.00000868, -0.00001562, -0.00000308, 0.00000770, -0.00000199, -0.00000621, -0.00000140, 0.00000736, 0.00000304, -0.00000519, -0.00000278, -0.00000638, 0.00000131, -0.00000163, 0.00000508, -0.00001077, 0.00001251, -0.00001161, -0.00000846, 0.00001333, 0.00001633, -0.00002121, 0.00000199, -0.00000535, 0.00000643, 0.00000486, 0.00000465, -0.00000926, 0.00000149, -0.00000397, 0.00000399, -0.00000105, 0.00000285, -0.00000341, },
		{ -0.00000176, 0.00000813, -0.00001266, 0.00000678, 0.00000233, 0.00000389, -0.00002195, 0.00001832, -0.00000610, 0.00000423, 0.00000128, -0.00000093, -0.00000295, 0.00000421, -0.00000594, 0.00000545, -0.00000013, 0.00000333, -0.00000645, 0.00000353, 0.00000059, -0.00000070, 0.00000178, -0.00000244, 0.00001042, -0.00001410, 0.00000391, 0.00000285, 0.00000471, -0.00001219, 0.00000838, -0.00000744, 0.00002138, -0.00001902, 0.00000558, -0.00000151, -0.00001959, 0.00001082, -0.00000471, -0.00000573, -0.00001863, 0.00002373, -0.00002379, 0.00002876, -0.00002798, 0.00002147, -0.00001410, 0.00001114, -0.00001796, 0.00001754, -0.00002305, 0.00001726, -0.00002733, 0.00002086, -0.00004288, 0.00002512, -0.00006727, 0.00002811, -0.00031150, 0.00037529, 0.00043030, -0.00030833, 0.00001711, -0.00006475, 0.00001426, -0.00003570, 0.00001564, -0.00002186, 0.00001040, -0.00001551, 0.00000468, -0.00000954, 0.00000668, -0.00000870, 0.00000066, -0.00000746, 0.00000192, -0.00002231, 0.00001392, -0.00001462, 0.00003673, -0.00000936, 0.00000094, 0.00002402, 0.00000358, 0.00000296, -0.00000193, -0.00001799, 0.00000156, -0.00000838, 0.00000965, -0.00001022, 0.00000167, 0.00000289, 0.00001358, -0.00001612, -0.00000395, 0.00000557, 0.00000195, -0.00000871, 0.00000009, 0.00000349, 0.00000771, -0.00000753, 0.00000009, -0.00000936, 0.00000350, -0.00000451, 0.00000848, -0.00001300, 0.00001189, -0.00000724, -0.00001135, 0.00000836, 0.00002135, -0.00001818, -0.00000239, -0.00000377, 0.00000389, 0.00000662, 0.00000441, -0.00000737, -0.00000103, -0.00000171, 0.00000119, 0.00000180, -0.00000009, -0.00000025, },
		{ -0.00000501, 0.00001155, -0.00001262, 0.00000361, 0.00000272, 0.00000747, -0.00002249, 0.00001287, 0.00000075, -0.00000171, 0.00000558, -0.00000312, -0.00000164, 0.00000288, -0.00000356, 0.00000160, 0.00000289, 0.00000110, -0.00000284, -0.00000096, 0.00000477, -0.00000462, 0.00000548, -0.00000642, 0.00001371, -0.00001442, 0.00000231, 0.00000261, 0.00000704, -0.00001306, 0.00000702, -0.00000595, 0.00001635, -0.00000770, -0.00000796, 0.00001232, -0.00002861, 0.00001436, -0.00000807, 0.00000187, -0.00002742, 0.00002874, -0.00002714, 0.00003041, -0.00002635, 0.00001792, -0.00001155, 0.00001090, -0.00001843, 0.00001767, -0.00002252, 0.00001690, -0.00002736, 0.00002242, -0.00004336, 0.00002681, -0.00006918, 0.00004374, -0.00028563, 0.00018720, 0.00060499, -0.00031181, -0.00000859, -0.00005789, 0.00000848, -0.00003413, 0.00001290, -0.00001986, 0.00000876, -0.00001461, 0.00000463, -0.00001092, 0.00000797, -0.00000966, 0.00000302, -0.00001115, 0.00000714, -0.00002569, 0.00001587, -0.00001729, 0.00003278, -0.00000188, -0.00000543, 0.00002126, 0.00001002, -0.00000267, 0.00000710, -0.00002326, 0.00000435, -0.00001174, 0.00001164, -0.00001070, 0.00000187, 0.00000026, 0.00001599, -0.00001417, -0.00000620, 0.00000433, 0.00000455, -0.00000958, 0.00000031, 0.00000091, 0.00001031, -0.00000768, 0.00000105, -0.00001028, 0.00000363, -0.00000533, 0.00000935, -0.00001208, 0.00000813, -0.00000122, -0.00001394, 0.00000380, 0.00002341, -0.00001225, -0.00000761, -0.00000192, 0.00000090, 0.00000854, 0.00000392, -0.00000506, -0.00000374, 0.00000054, -0.00000143, 0.00000421, -0.00000233, 0.00000210, },
		{ -0.00000654, 0.00001223, -0.00000938, -0.00000135, 0.00000372, 0.00000918, -0.00001946, 0.00000439, 0.00000881, -0.00000749, 0.00000904, -0.00000452, -0.00000059, 0.00000117, -0.00000051, -0.00000248, 0.00000575, -0.00000122, 0.00000090, -0.00000513, 0.00000828, -0.00000762, 0.00000804, -0.00000893, 0.00001499, -0.00001261, -0.00000071, 0.00000318, 0.00000800, -0.00001158, 0.00000303, -0.00000161, 0.00000841, 0.00000479, -0.00001973, 0.00002142, -0.00003117, 0.00001214, -0.00000712, 0.00000508, -0.00003037, 0.00002717, -0.00002381, 0.00002502, -0.00001773, 0.00000822, -0.00000353, 0.00000489, -0.00001227, 0.00001041, -0.00001405, 0.00000805, -0.00001808, 0.00001294, -0.00003056, 0.00001239, -0.00005042, 0.00002538, -0.00020798, -0.00002539, 0.00073943, -0.00025864, -0.00006553, -0.00003388, -0.00000999, -0.00002241, 0.00000151, -0.00001091, 0.00000147, -0.00000907, 0.00000058, -0.00000854, 0.00000562, -0.00000753, 0.00000271, -0.00001207, 0.00000897, -0.00002540, 0.00001430, -0.00001588, 0.00002533, 0.00000670, -0.00001016, 0.00001657, 0.00001641, -0.00000706, 0.00001300, -0.00002387, 0.00000277, -0.00001098, 0.00000956, -0.00000749, -0.00000104, 0.00000069, 0.00001480, -0.00000899, -0.00001049, 0.00000478, 0.00000491, -0.00000801, -0.00000150, 0.00000047, 0.00001014, -0.00000519, -0.00000027, -0.00000883, 0.00000148, -0.00000392, 0.00000762, -0.00000827, 0.00000208, 0.00000531, -0.00001559, -0.00000030, 0.00002276, -0.00000454, -0.00001243, -0.00000064, -0.00000185, 0.00000994, 0.00000376, -0.00000293, -0.00000598, 0.00000210, -0.00000316, 0.00000549, -0.00000323, 0.00000301, },
		{ -0.00000634, 0.00001040, -0.00000382, -0.00000689, 0.00000455, 0.00000945, -0.00001384, -0.00000530, 0.00001601, -0.00001135, 0.00001045, -0.00000443, -0.00000024, -0.00000044, 0.00000251, -0.00000597, 0.00000771, -0.00000297, 0.00000402, -0.00000822, 0.00001044, -0.00000911, 0.00000900, -0.00000966, 0.00001419, -0.00000895, -0.00000461, 0.00000408, 0.00000799, -0.00000846, -0.00000225, 0.00000353, 0.00000051, 0.00001469, -0.00002595, 0.00002282, -0.00002570, 0.00000368, -0.00000156, 0.00000336, -0.00002701, 0.00001891, -0.00001397, 0.00001299, -0.00000292, -0.00000661, 0.00000909, -0.00000622, -0.00000018, -0.00000335, 0.00000124, -0.00000801, -0.00000083, -0.00000606, -0.00000651, -0.00001529, -0.00001507, -0.00001846, -0.00010063, -0.00022532, 0.00080055, -0.00013538, -0.00015463, 0.00000629, -0.00004050, -0.00000101, -0.00001806, 0.00000452, -0.00001115, 0.00000108, -0.00000767, -0.00000202, -0.00000074, -0.00000189, -0.00000088, -0.00000940, 0.00000658, -0.00002078, 0.00000864, -0.00001013, 0.00001457, 0.00001623, -0.00001404, 0.00001206, 0.00001994, -0.00000723, 0.00001306, -0.00001795, -0.00000424, -0.00000543, 0.00000301, -0.00000046, -0.00000711, 0.00000423, 0.00001013, -0.00000095, -0.00001647, 0.00000680, 0.00000311, -0.00000419, -0.00000513, 0.00000202, 0.00000747, -0.00000058, -0.00000330, -0.00000560, -0.00000229, -0.00000094, 0.00000405, -0.00000273, -0.00000468, 0.00001083, -0.00001552, -0.00000413, 0.00002011, 0.00000352, -0.00001563, -0.00000057, -0.00000386, 0.00001033, 0.00000433, -0.00000136, -0.00000735, 0.00000259, -0.00000367, 0.00000538, -0.00000265, 0.00000247, },
		{ -0.00000496, 0.00000695, 0.00000259, -0.00001146, 0.00000428, 0.00000900, -0.00000703, -0.00001416, 0.00002043, -0.00001200, 0.00000916, -0.00000261, -0.00000072, -0.00000164, 0.00000491, -0.00000820, 0.00000828, -0.00000369, 0.00000599, -0.00000974, 0.00001089, -0.00000888, 0.00000826, -0.00000865, 0.00001158, -0.00000406, -0.00000862, 0.00000467, 0.00000763, -0.00000473, -0.00000723, 0.00000738, -0.00000472, 0.00001932, -0.00002465, 0.00001598, -0.00001316, -0.00000923, 0.00000692, -0.00000190, -0.00001888, 0.00000590, 0.00000016, -0.00000310, 0.00001508, -0.00002342, 0.00002327, -0.00001943, 0.00001461, -0.00001994, 0.00001927, -0.00002680, 0.00001953, -0.00002900, 0.00002204, -0.00004784, 0.00002599, -0.00007096, 0.00000720, -0.00037533, 0.00076457, 0.00005830, -0.00026612, 0.00005425, -0.00007703, 0.00002535, -0.00004183, 0.00002305, -0.00002640, 0.00001370, -0.00001846, 0.00000726, -0.00000989, 0.00000624, -0.00000706, -0.00000359, 0.00000048, -0.00001244, -0.00000046, -0.00000083, 0.00000150, 0.00002610, -0.00001757, 0.00000934, 0.00001884, -0.00000180, 0.00000660, -0.00000594, -0.00001555, 0.00000357, -0.00000651, 0.00000885, -0.00001484, 0.00000947, 0.00000349, 0.00000831, -0.00002264, 0.00000917, 0.00000025, 0.00000069, -0.00000944, 0.00000448, 0.00000354, 0.00000476, -0.00000670, -0.00000186, -0.00000643, 0.00000236, 0.00000003, 0.00000290, -0.00001035, 0.00001391, -0.00001319, -0.00000787, 0.00001629, 0.00001060, -0.00001638, -0.00000194, -0.00000496, 0.00000958, 0.00000566, -0.00000038, -0.00000781, 0.00000204, -0.00000309, 0.00000415, -0.00000100, 0.00000097, },
		{ -0.00000329, 0.00000307, 0.00000825, -0.00001370, 0.00000222, 0.00000849, -0.00000044, -0.00002041, 0.00002082, -0.00000900, 0.00000531, 0.00000059, -0.00000182, -0.00000233, 0.00000633, -0.00000881, 0.00000726, -0.00000324, 0.00000659, -0.00000951, 0.00000959, -0.00000702, 0.00000604, -0.00000625, 0.00000773, 0.00000127, -0.00001197, 0.00000442, 0.00000743, -0.00000129, -0.00001066, 0.00000855, -0.00000592, 0.00001786, -0.00001637, 0.00000303, 0.00000301, -0.00002282, 0.00001503, -0.00000782, -0.00000908, -0.00000833, 0.00001476, -0.00001908, 0.00003175, -0.00003760, 0.00003454, -0.00003022, 0.00002716, -0.00003392, 0.00003410, -0.00004191, 0.00003599, -0.00004792, 0.00004564, -0.00007391, 0.00005864, -0.00011239, 0.00008696, -0.00044733, 0.00062470, 0.00030551, -0.00037836, 0.00009445, -0.00010865, 0.00004813, -0.00006273, 0.00003879, -0.00003943, 0.00002479, -0.00002839, 0.00001626, -0.00001904, 0.00001438, -0.00001372, 0.00000344, -0.00000741, -0.00000236, -0.00001107, 0.00000998, -0.00001185, 0.00003486, -0.00002039, 0.00000868, 0.00001322, 0.00000833, -0.00000448, 0.00000924, -0.00002785, 0.00001284, -0.00001594, 0.00001750, -0.00002152, 0.00001389, -0.00000264, 0.00001631, -0.00002682, 0.00001005, -0.00000189, 0.00000485, -0.00001271, 0.00000622, 0.00000008, 0.00000908, -0.00000888, 0.00000090, -0.00000948, 0.00000455, -0.00000298, 0.00000699, -0.00001342, 0.00001366, -0.00000854, -0.00001149, 0.00001201, 0.00001575, -0.00001446, -0.00000446, -0.00000538, 0.00000788, 0.00000743, 0.00000030, -0.00000763, 0.00000082, -0.00000190, 0.00000238, 0.00000099, -0.00000068, },
		{ -0.00000211, -0.00000015, 0.00001193, -0.00001287, -0.00000171, 0.00000821, 0.00000494, -0.00002305, 0.00001696, -0.00000292, -0.00000016, 0.00000429, -0.00000299, -0.00000265, 0.00000668, -0.00000780, 0.00000483, -0.00000176, 0.00000587, -0.00000769, 0.00000682, -0.00000387, 0.00000276, -0.00000295, 0.00000326, 0.00000635, -0.00001412, 0.00000312, 0.00000755, 0.00000147, -0.00001206, 0.00000680, -0.00000336, 0.00001161, -0.00000392, -0.00001178, 0.00001788, -0.00003249, 0.00001903, -0.00001104, -0.00000120, -0.00001979, 0.00002557, -0.00003043, 0.00004234, -0.00004449, 0.00003840, -0.00003397, 0.00003248, -0.00003984, 0.00003990, -0.00004717, 0.00004184, -0.00005519, 0.00005542, -0.00008311, 0.00007041, -0.00012654, 0.00011872, -0.00042995, 0.00039545, 0.00057213, -0.00046040, 0.00010714, -0.00012208, 0.00005712, -0.00007243, 0.00004484, -0.00004456, 0.00002952, -0.00003319, 0.00002109, -0.00002455, 0.00001923, -0.00001790, 0.00000890, -0.00001423, 0.00000655, -0.00002035, 0.00001942, -0.00002272, 0.00004041, -0.00002123, 0.00000896, 0.00000509, 0.00002005, -0.00001615, 0.00002293, -0.00003664, 0.00001836, -0.00002163, 0.00002216, -0.00002418, 0.00001480, -0.00000564, 0.00002056, -0.00002688, 0.00000767, -0.00000163, 0.00000656, -0.00001333, 0.00000572, -0.00000135, 0.00001087, -0.00000856, 0.00000151, -0.00001030, 0.00000458, -0.00000389, 0.00000847, -0.00001311, 0.00001005, -0.00000219, -0.00001459, 0.00000764, 0.00001861, -0.00001030, -0.00000739, -0.00000560, 0.00000568, 0.00000916, 0.00000109, -0.00000722, -0.00000060, -0.00000068, 0.00000075, 0.00000257, -0.00000171, },
		{ -0.00000188, -0.00000213, 0.00001309, -0.00000907, -0.00000690, 0.00000793, 0.00000874, -0.00002199, 0.00000970, 0.00000475, -0.00000573, 0.00000734, -0.00000361, -0.00000284, 0.00000609, -0.00000546, 0.00000145, 0.00000037, 0.00000415, -0.00000467, 0.00000305, 0.00000005, -0.00000107, 0.00000073, -0.00000127, 0.00001064, -0.00001488, 0.00000096, 0.00000773, 0.00000367, -0.00001175, 0.00000303, 0.00000132, 0.00000341, 0.00000852, -0.00002350, 0.00002661, -0.00003437, 0.00001612, -0.00000912, 0.00000196, -0.00002532, 0.00002934, -0.00003374, 0.00004341, -0.00004094, 0.00003201, -0.00002778, 0.00002738, -0.00003430, 0.00003316, -0.00003894, 0.00003320, -0.00004642, 0.00004635, -0.00006986, 0.00005509, -0.00010608, 0.00009673, -0.00033215, 0.00011190, 0.00081167, -0.00047827, 0.00007384, -0.00010597, 0.00004388, -0.00006417, 0.00003575, -0.00003731, 0.00002401, -0.00002930, 0.00001835, -0.00002321, 0.00001785, -0.00001687, 0.00001003, -0.00001708, 0.00001134, -0.00002538, 0.00002454, -0.00002833, 0.00004057, -0.00001839, 0.00000808, -0.00000228, 0.00002915, -0.00002362, 0.00003025, -0.00003773, 0.00001676, -0.00002071, 0.00002035, -0.00002072, 0.00001037, -0.00000378, 0.00001950, -0.00002164, 0.00000112, 0.00000194, 0.00000490, -0.00001048, 0.00000225, 0.00000004, 0.00000945, -0.00000529, -0.00000038, -0.00000858, 0.00000218, -0.00000245, 0.00000717, -0.00000970, 0.00000398, 0.00000465, -0.00001649, 0.00000328, 0.00001932, -0.00000481, -0.00000979, -0.00000614, 0.00000343, 0.00001037, 0.00000235, -0.00000688, -0.00000182, 0.00000011, -0.00000023, 0.00000325, -0.00000169, },
		{ -0.00000249, -0.00000289, 0.00001197, -0.00000323, -0.00001210, 0.00000702, 0.00001113, -0.00001797, 0.00000069, 0.00001207, -0.00000979, 0.00000872, -0.00000317, -0.00000313, 0.00000480, -0.00000228, -0.00000226, 0.00000263, 0.00000185, -0.00000097, -0.00000112, 0.00000416, -0.00000487, 0.00000427, -0.00000538, 0.00001384, -0.00001434, -0.00000158, 0.00000746, 0.00000577, -0.00001057, -0.00000124, 0.00000578, -0.00000344, 0.00001690, -0.00002812, 0.00002608, -0.00002679, 0.00000571, -0.00000159, -0.00000049, -0.00002380, 0.00002495, -0.00002792, 0.00003413, -0.00002662, 0.00001543, -0.00001179, 0.00001191, -0.00001743, 0.00001425, -0.00001784, 0.00001084, -0.00002240, 0.00001943, -0.00003578, 0.00001527, -0.00005521, 0.00003100, -0.00018157, -0.00017661, 0.00097451, -0.00040385, -0.00001580, -0.00005591, 0.00000550, -0.00003581, 0.00000994, -0.00001651, 0.00000711, -0.00001550, 0.00000671, -0.00001364, 0.00000893, -0.00000932, 0.00000530, -0.00001417, 0.00001011, -0.00002426, 0.00002335, -0.00002684, 0.00003392, -0.00001059, 0.00000404, -0.00000572, 0.00003181, -0.00002304, 0.00002791, -0.00002893, 0.00000677, -0.00001236, 0.00001158, -0.00001094, 0.00000058, 0.00000294, 0.00001322, -0.00001149, -0.00000903, 0.00000834, 0.00000023, -0.00000460, -0.00000375, 0.00000386, 0.00000530, 0.00000030, -0.00000416, -0.00000489, -0.00000203, 0.00000072, 0.00000385, -0.00000432, -0.00000294, 0.00001046, -0.00001649, -0.00000114, 0.00001836, 0.00000094, -0.00001086, -0.00000730, 0.00000138, 0.00001078, 0.00000417, -0.00000668, -0.00000272, 0.00000034, -0.00000046, 0.00000297, -0.00000069, },
		{ -0.00000340, -0.00000294, 0.00000943, 0.00000319, -0.00001588, 0.00000474, 0.00001259, -0.00001223, -0.00000810, 0.00001716, -0.00001110, 0.00000784, -0.00000153, -0.00000354, 0.00000307, 0.00000119, -0.00000567, 0.00000454, -0.00000053, 0.00000282, -0.00000506, 0.00000784, -0.00000812, 0.00000720, -0.00000867, 0.00001577, -0.00001273, -0.00000400, 0.00000631, 0.00000821, -0.00000937, -0.00000454, 0.00000802, -0.00000643, 0.00001869, -0.00002395, 0.00001611, -0.00001112, -0.00001004, 0.00000960, -0.00000711, -0.00001653, 0.00001387, -0.00001471, 0.00001671, -0.00000444, -0.00000794, 0.00001047, -0.00001024, 0.00000672, -0.00001223, 0.00001097, -0.00001958, 0.00001061, -0.00001811, 0.00001029, -0.00003796, 0.00001147, -0.00005580, -0.00001771, -0.00041708, 0.00101961, -0.00022434, -0.00015791, 0.00002181, -0.00005269, 0.00000816, -0.00002859, 0.00001440, -0.00001835, 0.00000601, -0.00001208, 0.00000271, -0.00000626, 0.00000376, -0.00000469, -0.00000573, 0.00000293, -0.00001695, 0.00001568, -0.00001819, 0.00002070, 0.00000209, -0.00000398, -0.00000345, 0.00002617, -0.00001311, 0.00001559, -0.00001119, -0.00000995, 0.00000158, -0.00000221, 0.00000310, -0.00001253, 0.00001262, 0.00000365, 0.00000146, -0.00002069, 0.00001576, -0.00000579, 0.00000268, -0.00001070, 0.00000864, -0.00000008, 0.00000662, -0.00000837, -0.00000059, -0.00000672, 0.00000431, -0.00000007, 0.00000131, -0.00000888, 0.00001380, -0.00001411, -0.00000563, 0.00001629, 0.00000598, -0.00001017, -0.00000902, -0.00000048, 0.00001045, 0.00000636, -0.00000638, -0.00000347, 0.00000017, -0.00000018, 0.00000209, 0.00000082, },
		{ -0.00000390, -0.00000303, 0.00000655, 0.00000867, -0.00001701, 0.00000067, 0.00001355, -0.00000604, -0.00001496, 0.00001877, -0.00000919, 0.00000484, 0.00000097, -0.00000393, 0.00000109, 0.00000444, -0.00000822, 0.00000567, -0.00000257, 0.00000609, -0.00000815, 0.00001053, -0.00001028, 0.00000906, -0.00001076, 0.00001626, -0.00001022, -0.00000600, 0.00000416, 0.00001102, -0.00000857, -0.00000607, 0.00000710, -0.00000472, 0.00001375, -0.00001227, -0.00000033, 0.00000840, -0.00002663, 0.00002053, -0.00001453, -0.00000686, -0.00000018, 0.00000177, -0.00000408, 0.00002006, -0.00003211, 0.00003284, -0.00003265, 0.00003116, -0.00003861, 0.00003912, -0.00004898, 0.00004263, -0.00005484, 0.00005480, -0.00008824, 0.00007348, -0.00013519, 0.00011820, -0.00056525, 0.00092562, 0.00005078, -0.00033124, 0.00010900, -0.00011680, 0.00005647, -0.00007035, 0.00004737, -0.00004568, 0.00002958, -0.00003313, 0.00002146, -0.00002376, 0.00001890, -0.00001706, 0.00000575, -0.00000792, -0.00000556, 0.00000348, -0.00000441, 0.00000310, 0.00001778, -0.00001495, 0.00000401, 0.00001321, 0.00000422, -0.00000359, 0.00001133, -0.00002896, 0.00001688, -0.00001696, 0.00001752, -0.00002532, 0.00002190, -0.00000596, 0.00001392, -0.00003078, 0.00002147, -0.00001068, 0.00000895, -0.00001634, 0.00001231, -0.00000460, 0.00001168, -0.00001122, 0.00000266, -0.00001030, 0.00000679, -0.00000303, 0.00000550, -0.00001229, 0.00001385, -0.00000941, -0.00000996, 0.00001352, 0.00000972, -0.00000777, -0.00001082, -0.00000245, 0.00000963, 0.00000845, -0.00000560, -0.00000437, -0.00000004, 0.00000014, 0.00000121, 0.00000212, },
		{ -0.00000341, -0.00000374, 0.00000424, 0.00001210, -0.00001494, -0.00000506, 0.00001412, -0.00000034, -0.00001882, 0.00001658, -0.00000453, 0.00000050, 0.00000365, -0.00000397, -0.00000098, 0.00000705, -0.00000953, 0.00000575, -0.00000390, 0.00000835, -0.00000985, 0.00001174, -0.00001091, 0.00000943, -0.00001127, 0.00001509, -0.00000680, -0.00000765, 0.00000141, 0.00001371, -0.00000788, -0.00000604, 0.00000352, 0.00000058, 0.00000435, 0.00000301, -0.00001797, 0.00002585, -0.00003853, 0.00002661, -0.00001872, 0.00000098, -0.00001254, 0.00001634, -0.00002235, 0.00004034, -0.00005023, 0.00004837, -0.00004807, 0.00004808, -0.00005645, 0.00005757, -0.00006765, 0.00006304, -0.00007868, 0.00008372, -0.00011913, 0.00011100, -0.00018149, 0.00019399, -0.00059653, 0.00069836, 0.00038749, -0.00049862, 0.00017835, -0.00016701, 0.00009351, -0.00010245, 0.00007165, -0.00006586, 0.00004743, -0.00004954, 0.00003637, -0.00003788, 0.00003103, -0.00002731, 0.00001616, -0.00001850, 0.00000610, -0.00000957, 0.00001068, -0.00001498, 0.00003299, -0.00002601, 0.00001384, -0.00000334, 0.00002408, -0.00002378, 0.00003227, -0.00004421, 0.00002814, -0.00002777, 0.00002775, -0.00003362, 0.00002700, -0.00001202, 0.00002243, -0.00003612, 0.00002275, -0.00001200, 0.00001189, -0.00001853, 0.00001292, -0.00000638, 0.00001370, -0.00001121, 0.00000353, -0.00001149, 0.00000698, -0.00000383, 0.00000711, -0.00001239, 0.00001060, -0.00000309, -0.00001361, 0.00001025, 0.00001199, -0.00000423, -0.00001200, -0.00000489, 0.00000869, 0.00000998, -0.00000393, -0.00000573, 0.00000007, 0.00000006, 0.00000085, 0.00000261, },
		{ -0.00000179, -0.00000522, 0.00000289, 0.00001311, -0.00000997, -0.00001161, 0.00001398, 0.00000444, -0.00001946, 0.00001124, 0.00000166, -0.00000393, 0.00000564, -0.00000336, -0.00000303, 0.00000870, -0.00000939, 0.00000467, -0.00000428, 0.00000920, -0.00000979, 0.00001109, -0.00000965, 0.00000799, -0.00000989, 0.00001207, -0.00000236, -0.00000930, -0.00000117, 0.00001535, -0.00000648, -0.00000549, -0.00000106, 0.00000693, -0.00000567, 0.00001673, -0.00003086, 0.00003546, -0.00004104, 0.00002429, -0.00001651, 0.00000351, -0.00001919, 0.00002448, -0.00003299, 0.00005083, -0.00005670, 0.00005157, -0.00005086, 0.00005144, -0.00005933, 0.00005960, -0.00006852, 0.00006417, -0.00008099, 0.00008721, -0.00011952, 0.00011131, -0.00017958, 0.00019487, -0.00051234, 0.00037208, 0.00073132, -0.00061365, 0.00019988, -0.00018304, 0.00010381, -0.00011236, 0.00007694, -0.00007032, 0.00005211, -0.00005460, 0.00004126, -0.00004294, 0.00003501, -0.00003078, 0.00002106, -0.00002445, 0.00001374, -0.00001921, 0.00002263, -0.00002899, 0.00004350, -0.00003332, 0.00002184, -0.00001817, 0.00004020, -0.00003821, 0.00004500, -0.00005004, 0.00003067, -0.00003051, 0.00003013, -0.00003420, 0.00002511, -0.00001191, 0.00002446, -0.00003459, 0.00001792, -0.00000826, 0.00001011, -0.00001606, 0.00000940, -0.00000441, 0.00001185, -0.00000774, 0.00000149, -0.00000980, 0.00000445, -0.00000211, 0.00000590, -0.00000940, 0.00000494, 0.00000362, -0.00001586, 0.00000648, 0.00001297, -0.00000035, -0.00001188, -0.00000803, 0.00000784, 0.00001066, -0.00000126, -0.00000756, 0.00000057, -0.00000058, 0.00000125, 0.00000206, },
		{ 0.00000060, -0.00000711, 0.00000228, 0.00001209, -0.00000321, -0.00001761, 0.00001249, 0.00000835, -0.00001741, 0.00000415, 0.00000769, -0.00000712, 0.00000625, -0.00000197, -0.00000483, 0.00000918, -0.00000777, 0.00000250, -0.00000361, 0.00000842, -0.00000780, 0.00000850, -0.00000644, 0.00000464, -0.00000655, 0.00000719, 0.00000314, -0.00001132, -0.00000275, 0.00001497, -0.00000340, -0.00000577, -0.00000455, 0.00001137, -0.00001233, 0.00002419, -0.00003443, 0.00003366, -0.00003207, 0.00001255, -0.00000702, -0.00000065, -0.00001826, 0.00002391, -0.00003329, 0.00004872, -0.00004907, 0.00004035, -0.00003898, 0.00003914, -0.00004518, 0.00004326, -0.00004977, 0.00004413, -0.00005965, 0.00006302, -0.00008746, 0.00007306, -0.00012932, 0.00012696, -0.00033994, 0.00000359, 0.00101755, -0.00063144, 0.00014988, -0.00015099, 0.00007743, -0.00009241, 0.00005718, -0.00005408, 0.00003922, -0.00004422, 0.00003229, -0.00003536, 0.00002762, -0.00002446, 0.00001740, -0.00002259, 0.00001411, -0.00002203, 0.00002774, -0.00003513, 0.00004577, -0.00003342, 0.00002391, -0.00002615, 0.00004696, -0.00004142, 0.00004489, -0.00004311, 0.00002216, -0.00002343, 0.00002324, -0.00002601, 0.00001546, -0.00000493, 0.00001942, -0.00002593, 0.00000704, 0.00000037, 0.00000377, -0.00000916, 0.00000208, 0.00000099, 0.00000653, -0.00000138, -0.00000291, -0.00000573, -0.00000026, 0.00000155, 0.00000260, -0.00000443, -0.00000157, 0.00000924, -0.00001604, 0.00000222, 0.00001300, 0.00000308, -0.00001005, -0.00001174, 0.00000700, 0.00001050, 0.00000219, -0.00000956, 0.00000123, -0.00000160, 0.00000223, 0.00000072, },
		{ 0.00000299, -0.00000864, 0.00000177, 0.00000999, 0.00000376, -0.00002156, 0.00000902, 0.00001165, -0.00001364, -0.00000302, 0.00001200, -0.00000813, 0.00000517, 0.00000006, -0.00000613, 0.00000837, -0.00000484, -0.00000053, -0.00000199, 0.00000608, -0.00000406, 0.00000421, -0.00000154, -0.00000030, -0.00000156, 0.00000082, 0.00000942, -0.00001383, -0.00000284, 0.00001206, 0.00000186, -0.00000777, -0.00000540, 0.00001173, -0.00001297, 0.00002279, -0.00002705, 0.00002036, -0.00001309, -0.00000634, 0.00000773, -0.00001012, -0.00001067, 0.00001534, -0.00002390, 0.00003498, -0.00002896, 0.00001693, -0.00001495, 0.00001394, -0.00001721, 0.00001236, -0.00001574, 0.00000778, -0.00002007, 0.00001761, -0.00003116, 0.00000696, -0.00004550, 0.00001491, -0.00012564, -0.00033966, 0.00118545, -0.00052113, 0.00001992, -0.00006958, 0.00001460, -0.00004336, 0.00001348, -0.00001828, 0.00000962, -0.00001895, 0.00000979, -0.00001539, 0.00000907, -0.00000838, 0.00000489, -0.00001230, 0.00000633, -0.00001686, 0.00002442, -0.00003169, 0.00003825, -0.00002469, 0.00001773, -0.00002422, 0.00004136, -0.00003119, 0.00003102, -0.00002392, 0.00000387, -0.00000797, 0.00000863, -0.00001075, -0.00000016, 0.00000721, 0.00000898, -0.00001199, -0.00000785, 0.00001195, -0.00000535, 0.00000038, -0.00000731, 0.00000821, -0.00000062, 0.00000624, -0.00000820, -0.00000064, -0.00000580, 0.00000585, -0.00000139, 0.00000085, -0.00000718, 0.00001249, -0.00001380, -0.00000241, 0.00001239, 0.00000551, -0.00000659, -0.00001547, 0.00000578, 0.00000985, 0.00000583, -0.00001108, 0.00000152, -0.00000252, 0.00000323, -0.00000077, },
		{ 0.00000451, -0.00000901, 0.00000062, 0.00000790, 0.00000942, -0.00002230, 0.00000331, 0.00001455, -0.00000917, -0.00000879, 0.00001354, -0.00000667, 0.00000267, 0.00000227, -0.00000659, 0.00000631, -0.00000095, -0.00000401, 0.00000028, 0.00000255, 0.00000088, -0.00000113, 0.00000431, -0.00000608, 0.00000428, -0.00000620, 0.00001573, -0.00001642, -0.00000161, 0.00000692, 0.00000892, -0.00001146, -0.00000325, 0.00000750, -0.00000723, 0.00001302, -0.00001074, -0.00000072, 0.00001093, -0.00002720, 0.00002320, -0.00002108, 0.00000005, 0.00000232, -0.00000859, 0.00001404, -0.00000175, -0.00001259, 0.00001469, -0.00001712, 0.00001681, -0.00002450, 0.00002408, -0.00003449, 0.00002622, -0.00003568, 0.00003341, -0.00006743, 0.00004685, -0.00010563, 0.00007700, -0.00059482, 0.00119307, -0.00027659, -0.00017695, 0.00004645, -0.00007227, 0.00002420, -0.00004490, 0.00002906, -0.00002993, 0.00001544, -0.00002118, 0.00001238, -0.00001646, 0.00001379, -0.00001343, 0.00000390, -0.00000748, -0.00000536, 0.00001389, -0.00001980, 0.00002220, -0.00000822, 0.00000379, -0.00001261, 0.00002429, -0.00000960, 0.00000678, 0.00000307, -0.00001948, 0.00001139, -0.00000944, 0.00000746, -0.00001779, 0.00002085, -0.00000337, 0.00000367, -0.00002322, 0.00002323, -0.00001428, 0.00000975, -0.00001611, 0.00001480, -0.00000725, 0.00001289, -0.00001242, 0.00000371, -0.00001048, 0.00000916, -0.00000446, 0.00000478, -0.00001044, 0.00001264, -0.00000936, -0.00000698, 0.00001125, 0.00000679, -0.00000207, -0.00001835, 0.00000361, 0.00000925, 0.00000894, -0.00001141, 0.00000084, -0.00000273, 0.00000363, -0.00000166, },
		{ 0.00000456, -0.00000769, -0.00000168, 0.00000662, 0.00001274, -0.00001939, -0.00000429, 0.00001689, -0.00000471, -0.00001228, 0.00001210, -0.00000320, -0.00000051, 0.00000403, -0.00000599, 0.00000321, 0.00000336, -0.00000737, 0.00000269, -0.00000150, 0.00000612, -0.00000648, 0.00001000, -0.00001150, 0.00000972, -0.00001255, 0.00002087, -0.00001813, 0.00000004, 0.00000072, 0.00001642, -0.00001563, 0.00000076, 0.00000012, 0.00000276, -0.00000169, 0.00000943, -0.00002311, 0.00003292, -0.00004345, 0.00003380, -0.00002863, 0.00000907, -0.00001007, 0.00000706, -0.00000763, 0.00002508, -0.00004003, 0.00004132, -0.00004484, 0.00004687, -0.00005641, 0.00005789, -0.00006991, 0.00006510, -0.00008071, 0.00008733, -0.00012766, 0.00012026, -0.00019847, 0.00022079, -0.00071828, 0.00102873, 0.00007785, -0.00040520, 0.00016754, -0.00016028, 0.00009163, -0.00010225, 0.00007462, -0.00006809, 0.00004910, -0.00005179, 0.00003994, -0.00004168, 0.00003555, -0.00003185, 0.00002090, -0.00002268, 0.00000829, -0.00000006, -0.00000322, 0.00000157, 0.00001223, -0.00001447, 0.00000509, 0.00000039, 0.00001746, -0.00002088, 0.00003046, -0.00004093, 0.00002842, -0.00002528, 0.00002327, -0.00003245, 0.00003145, -0.00001337, 0.00001683, -0.00003508, 0.00003062, -0.00001978, 0.00001595, -0.00002160, 0.00001829, -0.00001105, 0.00001648, -0.00001381, 0.00000573, -0.00001281, 0.00001013, -0.00000534, 0.00000623, -0.00001064, 0.00000979, -0.00000349, -0.00001083, 0.00000950, 0.00000712, 0.00000262, -0.00001946, 0.00000000, 0.00000914, 0.00001087, -0.00001001, -0.00000119, -0.00000187, 0.00000293, -0.00000138, },
		{ 0.00000304, -0.00000470, -0.00000505, 0.00000637, 0.00001347, -0.00001333, -0.00001270, 0.00001815, -0.00000056, -0.00001333, 0.00000828, 0.00000120, -0.00000335, 0.00000475, -0.00000427, -0.00000047, 0.00000740, -0.00000994, 0.00000458, -0.00000522, 0.00001055, -0.00001062, 0.00001423, -0.00001519, 0.00001332, -0.00001673, 0.00002340, -0.00001771, 0.00000073, -0.00000474, 0.00002233, -0.00001828, 0.00000442, -0.00000758, 0.00001317, -0.00001614, 0.00002689, -0.00003957, 0.00004589, -0.00004919, 0.00003484, -0.00002858, 0.00001206, -0.00001703, 0.00001759, -0.00002361, 0.00004421, -0.00005761, 0.00005687, -0.00006071, 0.00006381, -0.00007356, 0.00007527, -0.00008732, 0.00008433, -0.00010359, 0.00011464, -0.00015541, 0.00015317, -0.00023722, 0.00027653, -0.00069583, 0.00071589, 0.00048902, -0.00061230, 0.00025547, -0.00022169, 0.00013706, -0.00014056, 0.00010336, -0.00009213, 0.00007085, -0.00007199, 0.00005812, -0.00005824, 0.00004943, -0.00004365, 0.00003252, -0.00003349, 0.00001874, -0.00001230, 0.00001272, -0.00001809, 0.00003116, -0.00003169, 0.00002297, -0.00002328, 0.00004198, -0.00004352, 0.00005014, -0.00005345, 0.00003717, -0.00003360, 0.00003185, -0.00003973, 0.00003508, -0.00001738, 0.00002395, -0.00004015, 0.00003136, -0.00001948, 0.00001681, -0.00002185, 0.00001703, -0.00001051, 0.00001577, -0.00001143, 0.00000461, -0.00001205, 0.00000814, -0.00000355, 0.00000493, -0.00000799, 0.00000481, 0.00000258, -0.00001320, 0.00000700, 0.00000691, 0.00000653, -0.00001810, -0.00000515, 0.00000966, 0.00001136, -0.00000677, -0.00000451, 0.00000004, 0.00000111, 0.00000018, },
		{ 0.00000043, -0.00000068, -0.00000885, 0.00000672, 0.00001217, -0.00000542, -0.00002035, 0.00001755, 0.00000338, -0.00001238, 0.00000326, 0.00000519, -0.00000493, 0.00000410, -0.00000169, -0.00000411, 0.00001042, -0.00001106, 0.00000530, -0.00000770, 0.00001308, -0.00001242, 0.00001582, -0.00001594, 0.00001382, -0.00001744, 0.00002214, -0.00001406, -0.00000086, -0.00000769, 0.00002464, -0.00001737, 0.00000535, -0.00001243, 0.00001988, -0.00002512, 0.00003579, -0.00004446, 0.00004526, -0.00004134, 0.00002435, -0.00001919, 0.00000685, -0.00001582, 0.00001956, -0.00002963, 0.00005080, -0.00006038, 0.00005648, -0.00005968, 0.00006230, -0.00007043, 0.00007055, -0.00008083, 0.00007752, -0.00009713, 0.00010737, -0.00014209, 0.00013646, -0.00021241, 0.00023963, -0.00054563, 0.00030952, 0.00088424, -0.00073991, 0.00027348, -0.00023202, 0.00014208, -0.00014514, 0.00010332, -0.00009203, 0.00007189, -0.00007380, 0.00005961, -0.00005952, 0.00004949, -0.00004345, 0.00003367, -0.00003506, 0.00002132, -0.00001811, 0.00002292, -0.00003136, 0.00004311, -0.00004230, 0.00003477, -0.00003951, 0.00005631, -0.00005384, 0.00005587, -0.00005229, 0.00003398, -0.00003133, 0.00003046, -0.00003729, 0.00002979, -0.00001360, 0.00002331, -0.00003699, 0.00002441, -0.00001264, 0.00001174, -0.00001645, 0.00001077, -0.00000552, 0.00001079, -0.00000550, 0.00000059, -0.00000840, 0.00000351, 0.00000047, 0.00000152, -0.00000353, -0.00000087, 0.00000752, -0.00001350, 0.00000371, 0.00000654, 0.00000892, -0.00001410, -0.00001143, 0.00001045, 0.00001062, -0.00000214, -0.00000854, 0.00000248, -0.00000139, 0.00000258, },
		{ -0.00000232, 0.00000328, -0.00001204, 0.00000682, 0.00000996, 0.00000262, -0.00002560, 0.00001439, 0.00000731, -0.00001023, -0.00000162, 0.00000761, -0.00000471, 0.00000220, 0.00000120, -0.00000697, 0.00001176, -0.00001028, 0.00000438, -0.00000830, 0.00001296, -0.00001117, 0.00001409, -0.00001311, 0.00001061, -0.00001411, 0.00001664, -0.00000678, -0.00000539, -0.00000704, 0.00002215, -0.00001171, 0.00000196, -0.00001222, 0.00002000, -0.00002526, 0.00003294, -0.00003564, 0.00003048, -0.00002093, 0.00000411, -0.00000206, -0.00000558, -0.00000682, 0.00001276, -0.00002492, 0.00004396, -0.00004784, 0.00004001, -0.00004183, 0.00004258, -0.00004759, 0.00004472, -0.00005180, 0.00004627, -0.00006315, 0.00006795, -0.00009149, 0.00007593, -0.00013352, 0.00012991, -0.00031298, -0.00011540, 0.00118654, -0.00073848, 0.00019765, -0.00017866, 0.00009835, -0.00010995, 0.00007002, -0.00006426, 0.00004906, -0.00005425, 0.00004169, -0.00004308, 0.00003363, -0.00002929, 0.00002234, -0.00002528, 0.00001381, -0.00001497, 0.00002429, -0.00003472, 0.00004449, -0.00004247, 0.00003602, -0.00004325, 0.00005569, -0.00004806, 0.00004538, -0.00003667, 0.00001891, -0.00001877, 0.00001950, -0.00002567, 0.00001632, -0.00000280, 0.00001558, -0.00002647, 0.00001098, -0.00000056, 0.00000204, -0.00000675, 0.00000090, 0.00000261, 0.00000293, 0.00000256, -0.00000510, -0.00000304, -0.00000255, 0.00000550, -0.00000268, 0.00000122, -0.00000566, 0.00001023, -0.00001151, -0.00000017, 0.00000620, 0.00000954, -0.00000796, -0.00001784, 0.00001077, 0.00000929, 0.00000298, -0.00001228, 0.00000453, -0.00000374, 0.00000498, },
		{ -0.00000416, 0.00000605, -0.00001356, 0.00000577, 0.00000801, 0.00000913, -0.00002718, 0.00000843, 0.00001131, -0.00000763, -0.00000524, 0.00000781, -0.00000275, -0.00000041, 0.00000366, -0.00000838, 0.00001103, -0.00000745, 0.00000173, -0.00000679, 0.00001001, -0.00000687, 0.00000916, -0.00000693, 0.00000401, -0.00000717, 0.00000749, 0.00000344, -0.00001242, -0.00000296, 0.00001506, -0.00000158, -0.00000569, -0.00000662, 0.00001303, -0.00001626, 0.00001898, -0.00001536, 0.00000549, 0.00000688, -0.00002052, 0.00001810, -0.00002125, 0.00000652, 0.00000027, -0.00001238, 0.00002693, -0.00002398, 0.00001221, -0.00001240, 0.00001046, -0.00001161, 0.00000522, -0.00000850, -0.00000025, -0.00001192, 0.00000851, -0.00001847, -0.00000970, -0.00002535, -0.00001522, -0.00005846, -0.00048053, 0.00133239, -0.00058141, 0.00002645, -0.00006692, 0.00001183, -0.00004080, 0.00000903, -0.00001384, 0.00000665, -0.00001705, 0.00000767, -0.00001206, 0.00000478, -0.00000377, 0.00000068, -0.00000591, -0.00000240, -0.00000362, 0.00001682, -0.00002778, 0.00003481, -0.00003151, 0.00002561, -0.00003336, 0.00003983, -0.00002730, 0.00002127, -0.00001033, -0.00000406, 0.00000034, 0.00000246, -0.00000826, -0.00000198, 0.00001194, 0.00000364, -0.00001154, -0.00000585, 0.00001373, -0.00000947, 0.00000456, -0.00001002, 0.00001145, -0.00000549, 0.00001059, -0.00001055, 0.00000233, -0.00000840, 0.00000993, -0.00000609, 0.00000472, -0.00000835, 0.00001020, -0.00000758, -0.00000411, 0.00000579, 0.00000859, -0.00000072, -0.00002305, 0.00000968, 0.00000824, 0.00000752, -0.00001461, 0.00000525, -0.00000502, 0.00000642, },
		{ -0.00000433, 0.00000680, -0.00001274, 0.00000300, 0.00000718, 0.00001301, -0.00002464, 0.00000010, 0.00001505, -0.00000505, -0.00000702, 0.00000587, 0.00000028, -0.00000286, 0.00000496, -0.00000796, 0.00000824, -0.00000290, -0.00000226, -0.00000353, 0.00000479, -0.00000033, 0.00000204, 0.00000143, -0.00000461, 0.00000186, -0.00000355, 0.00001480, -0.00002034, 0.00000307, 0.00000509, 0.00001105, -0.00001571, 0.00000254, 0.00000110, -0.00000101, -0.00000178, 0.00001037, -0.00002235, 0.00003421, -0.00004214, 0.00003486, -0.00003451, 0.00001889, -0.00001268, 0.00000257, 0.00000590, 0.00000394, -0.00001881, 0.00001983, -0.00002450, 0.00002695, -0.00003634, 0.00003635, -0.00004817, 0.00004107, -0.00005305, 0.00005572, -0.00009480, 0.00007992, -0.00015194, 0.00015745, -0.00072204, 0.00128718, -0.00027488, -0.00021492, 0.00007851, -0.00009703, 0.00004497, -0.00006460, 0.00004623, -0.00004413, 0.00002796, -0.00003350, 0.00002534, -0.00002948, 0.00002630, -0.00002528, 0.00001765, -0.00002256, 0.00001194, 0.00000378, -0.00001349, 0.00001705, -0.00001230, 0.00000627, -0.00001300, 0.00001314, 0.00000265, -0.00000953, 0.00001944, -0.00002812, 0.00001994, -0.00001517, 0.00000975, -0.00002020, 0.00002609, -0.00000830, 0.00000362, -0.00002193, 0.00002638, -0.00001927, 0.00001420, -0.00001895, 0.00001826, -0.00001191, 0.00001630, -0.00001385, 0.00000600, -0.00001240, 0.00001228, -0.00000743, 0.00000594, -0.00000837, 0.00000767, -0.00000259, -0.00000735, 0.00000505, 0.00000667, 0.00000627, -0.00002572, 0.00000639, 0.00000815, 0.00001053, -0.00001468, 0.00000395, -0.00000460, 0.00000623, },
		{ -0.00000264, 0.00000536, -0.00000955, -0.00000141, 0.00000767, 0.00001403, -0.00001850, -0.00000944, 0.00001785, -0.00000256, -0.00000703, 0.00000249, 0.00000335, -0.00000429, 0.00000465, -0.00000576, 0.00000390, 0.00000256, -0.00000675, 0.00000056, -0.00000152, 0.00000694, -0.00000556, 0.00001003, -0.00001315, 0.00001067, -0.00001397, 0.00002471, -0.00002672, 0.00000859, -0.00000494, 0.00002302, -0.00002489, 0.00001182, -0.00001173, 0.00001534, -0.00002270, 0.00003361, -0.00004448, 0.00005278, -0.00005358, 0.00004223, -0.00003996, 0.00002497, -0.00002064, 0.00001394, -0.00001215, 0.00002783, -0.00004415, 0.00004533, -0.00005200, 0.00005684, -0.00006775, 0.00006960, -0.00008321, 0.00007997, -0.00009860, 0.00010997, -0.00015462, 0.00015245, -0.00024242, 0.00028917, -0.00080474, 0.00105448, 0.00013989, -0.00047714, 0.00021803, -0.00019776, 0.00012259, -0.00013006, 0.00009825, -0.00008811, 0.00006733, -0.00006967, 0.00005795, -0.00005899, 0.00005177, -0.00004731, 0.00003791, -0.00003990, 0.00002575, -0.00000947, 0.00000285, -0.00000334, 0.00000962, -0.00001630, 0.00001137, -0.00001663, 0.00003301, -0.00003791, 0.00004396, -0.00004572, 0.00003359, -0.00002765, 0.00002305, -0.00003336, 0.00003516, -0.00001612, 0.00001495, -0.00003333, 0.00003381, -0.00002421, 0.00001930, -0.00002334, 0.00002080, -0.00001436, 0.00001804, -0.00001370, 0.00000682, -0.00001350, 0.00001171, -0.00000610, 0.00000460, -0.00000598, 0.00000352, 0.00000230, -0.00000916, 0.00000367, 0.00000446, 0.00001174, -0.00002496, 0.00000068, 0.00000926, 0.00001160, -0.00001220, 0.00000055, -0.00000242, 0.00000427, },
		{ 0.00000045, 0.00000225, -0.00000472, -0.00000669, 0.00000895, 0.00001279, -0.00001012, -0.00001850, 0.00001878, 0.00000007, -0.00000578, -0.00000125, 0.00000537, -0.00000409, 0.00000273, -0.00000232, -0.00000104, 0.00000780, -0.00001059, 0.00000425, -0.00000738, 0.00001312, -0.00001164, 0.00001670, -0.00001928, 0.00001678, -0.00002112, 0.00003055, -0.00002903, 0.00001096, -0.00001195, 0.00003094, -0.00002969, 0.00001735, -0.00002083, 0.00002715, -0.00003703, 0.00004700, -0.00005384, 0.00005662, -0.00005037, 0.00003681, -0.00003452, 0.00002150, -0.00001990, 0.00001734, -0.00002177, 0.00004129, -0.00005686, 0.00005678, -0.00006417, 0.00006960, -0.00008001, 0.00008172, -0.00009514, 0.00009350, -0.00011528, 0.00012967, -0.00017269, 0.00017342, -0.00026478, 0.00031614, -0.00072921, 0.00067635, 0.00059319, -0.00069662, 0.00030679, -0.00025821, 0.00016691, -0.00016677, 0.00012510, -0.00011071, 0.00008819, -0.00008915, 0.00007514, -0.00007409, 0.00006399, -0.00005761, 0.00004770, -0.00004786, 0.00003187, -0.00001729, 0.00001540, -0.00002013, 0.00002780, -0.00003528, 0.00003204, -0.00004082, 0.00005477, -0.00005527, 0.00005585, -0.00005105, 0.00003661, -0.00003091, 0.00002788, -0.00003803, 0.00003614, -0.00001707, 0.00001971, -0.00003746, 0.00003387, -0.00002252, 0.00001836, -0.00002192, 0.00001806, -0.00001201, 0.00001526, -0.00000981, 0.00000455, -0.00001150, 0.00000821, -0.00000235, 0.00000125, -0.00000212, -0.00000096, 0.00000593, -0.00000904, 0.00000154, 0.00000253, 0.00001483, -0.00002055, -0.00000697, 0.00001116, 0.00001093, -0.00000759, -0.00000433, 0.00000095, 0.00000105, },
		{ 0.00000394, -0.00000144, 0.00000051, -0.00001162, 0.00001005, 0.00001045, -0.00000130, -0.00002533, 0.00001705, 0.00000311, -0.00000399, -0.00000424, 0.00000560, -0.00000217, -0.00000028, 0.00000144, -0.00000543, 0.00001162, -0.00001266, 0.00000633, -0.00001127, 0.00001651, -0.00001445, 0.00001961, -0.00002111, 0.00001827, -0.00002304, 0.00003042, -0.00002555, 0.00000832, -0.00001366, 0.00003232, -0.00002761, 0.00001628, -0.00002273, 0.00003024, -0.00004011, 0.00004614, -0.00004714, 0.00004404, -0.00003233, 0.00001917, -0.00001866, 0.00000849, -0.00000995, 0.00001152, -0.00002080, 0.00004157, -0.00005408, 0.00005132, -0.00005804, 0.00006209, -0.00007001, 0.00006966, -0.00008087, 0.00007831, -0.00009933, 0.00011089, -0.00014536, 0.00013986, -0.00021795, 0.00024603, -0.00052994, 0.00022435, 0.00100107, -0.00081043, 0.00030728, -0.00025443, 0.00016037, -0.00016098, 0.00011579, -0.00010274, 0.00008244, -0.00008460, 0.00007026, -0.00006889, 0.00005778, -0.00005150, 0.00004269, -0.00004243, 0.00002660, -0.00001591, 0.00001985, -0.00002848, 0.00003703, -0.00004497, 0.00004262, -0.00005252, 0.00006143, -0.00005630, 0.00005143, -0.00004195, 0.00002770, -0.00002394, 0.00002328, -0.00003334, 0.00002845, -0.00001064, 0.00001731, -0.00003383, 0.00002641, -0.00001433, 0.00001164, -0.00001511, 0.00001058, -0.00000550, 0.00000874, -0.00000303, -0.00000006, -0.00000712, 0.00000264, 0.00000283, -0.00000296, 0.00000188, -0.00000450, 0.00000751, -0.00000696, -0.00000109, 0.00000110, 0.00001531, -0.00001316, -0.00001534, 0.00001290, 0.00000932, -0.00000188, -0.00000955, 0.00000442, -0.00000242, },
		{ 0.00000666, -0.00000444, 0.00000477, -0.00001493, 0.00000989, 0.00000832, 0.00000620, -0.00002854, 0.00001230, 0.00000665, -0.00000229, -0.00000571, 0.00000383, 0.00000095, -0.00000345, 0.00000441, -0.00000817, 0.00001306, -0.00001215, 0.00000598, -0.00001222, 0.00001611, -0.00001306, 0.00001788, -0.00001784, 0.00001440, -0.00001907, 0.00002390, -0.00001605, 0.00000034, -0.00000949, 0.00002650, -0.00001807, 0.00000790, -0.00001638, 0.00002336, -0.00003092, 0.00003108, -0.00002608, 0.00001836, -0.00000394, -0.00000601, 0.00000347, -0.00001055, 0.00000625, -0.00000122, -0.00001096, 0.00003027, -0.00003776, 0.00003131, -0.00003626, 0.00003737, -0.00004137, 0.00003767, -0.00004526, 0.00003978, -0.00005685, 0.00006107, -0.00008231, 0.00006468, -0.00012038, 0.00011019, -0.00026526, -0.00021664, 0.00128397, -0.00077260, 0.00020134, -0.00017926, 0.00009934, -0.00011085, 0.00006954, -0.00006393, 0.00004978, -0.00005573, 0.00004315, -0.00004345, 0.00003339, -0.00002929, 0.00002308, -0.00002381, 0.00001000, -0.00000488, 0.00001493, -0.00002648, 0.00003505, -0.00004272, 0.00004000, -0.00004875, 0.00005106, -0.00004063, 0.00003193, -0.00002078, 0.00000946, -0.00000910, 0.00001131, -0.00002128, 0.00001412, 0.00000127, 0.00000939, -0.00002410, 0.00001332, -0.00000164, 0.00000109, -0.00000484, 0.00000030, 0.00000331, 0.00000030, 0.00000490, -0.00000551, -0.00000174, -0.00000359, 0.00000801, -0.00000664, 0.00000476, -0.00000618, 0.00000682, -0.00000347, -0.00000366, 0.00000005, 0.00001355, -0.00000413, -0.00002276, 0.00001326, 0.00000781, 0.00000363, -0.00001373, 0.00000675, -0.00000499, },
		{ 0.00000768, -0.00000577, 0.00000698, -0.00001570, 0.00000778, 0.00000734, 0.00001118, -0.00002756, 0.00000488, 0.00001037, -0.00000097, -0.00000542, 0.00000050, 0.00000433, -0.00000570, 0.00000567, -0.00000859, 0.00001175, -0.00000893, 0.00000306, -0.00001005, 0.00001189, -0.00000768, 0.00001191, -0.00001008, 0.00000598, -0.00001022, 0.00001228, -0.00000208, -0.00001148, -0.00000079, 0.00001503, -0.00000291, -0.00000593, -0.00000363, 0.00000863, -0.00001241, 0.00000628, 0.00000321, -0.00001297, 0.00002685, -0.00003130, 0.00002532, -0.00002973, 0.00002327, -0.00001585, 0.00000297, 0.00001253, -0.00001381, 0.00000338, -0.00000608, 0.00000346, -0.00000308, -0.00000422, 0.00000064, -0.00000990, -0.00000154, -0.00000369, -0.00000300, -0.00002806, -0.00000326, -0.00004659, -0.00000194, -0.00056755, 0.00138512, -0.00056755, -0.00000194, -0.00004659, -0.00000326, -0.00002806, -0.00000300, -0.00000369, -0.00000154, -0.00000990, 0.00000064, -0.00000422, -0.00000308, 0.00000346, -0.00000608, 0.00000338, -0.00001381, 0.00001253, 0.00000297, -0.00001585, 0.00002327, -0.00002973, 0.00002532, -0.00003130, 0.00002685, -0.00001297, 0.00000321, 0.00000628, -0.00001241, 0.00000863, -0.00000363, -0.00000593, -0.00000291, 0.00001503, -0.00000079, -0.00001148, -0.00000208, 0.00001228, -0.00001022, 0.00000598, -0.00001008, 0.00001191, -0.00000768, 0.00001189, -0.00001005, 0.00000306, -0.00000893, 0.00001175, -0.00000859, 0.00000567, -0.00000570, 0.00000433, 0.00000050, -0.00000542, -0.00000097, 0.00001037, 0.00000488, -0.00002756, 0.00001118, 0.00000734, 0.00000778, -0.00001570, 0.00000698, -0.00000577, },
		{ 0.00000666, -0.00000499, 0.00000675, -0.00001373, 0.00000363, 0.00000781, 0.00001326, -0.00002276, -0.00000413, 0.00001355, 0.00000005, -0.00000366, -0.00000347, 0.00000682, -0.00000618, 0.00000476, -0.00000664, 0.00000801, -0.00000359, -0.00000174, -0.00000551, 0.00000490, 0.00000030, 0.00000331, 0.00000030, -0.00000484, 0.00000109, -0.00000164, 0.00001332, -0.00002410, 0.00000939, 0.00000127, 0.00001412, -0.00002128, 0.00001131, -0.00000910, 0.00000946, -0.00002078, 0.00003193, -0.00004063, 0.00005106, -0.00004875, 0.00004000, -0.00004272, 0.00003505, -0.00002648, 0.00001493, -0.00000488, 0.00001000, -0.00002381, 0.00002308, -0.00002929, 0.00003339, -0.00004345, 0.00004315, -0.00005573, 0.00004978, -0.00006393, 0.00006954, -0.00011085, 0.00009934, -0.00017926, 0.00020134, -0.00077260, 0.00128397, -0.00021664, -0.00026526, 0.00011019, -0.00012038, 0.00006468, -0.00008231, 0.00006107, -0.00005685, 0.00003978, -0.00004526, 0.00003767, -0.00004137, 0.00003737, -0.00003626, 0.00003131, -0.00003776, 0.00003027, -0.00001096, -0.00000122, 0.00000625, -0.00001055, 0.00000347, -0.00000601, -0.00000394, 0.00001836, -0.00002608, 0.00003108, -0.00003092, 0.00002336, -0.00001638, 0.00000790, -0.00001807, 0.00002650, -0.00000949, 0.00000034, -0.00001605, 0.00002390, -0.00001907, 0.00001440, -0.00001784, 0.00001788, -0.00001306, 0.00001611, -0.00001222, 0.00000598, -0.00001215, 0.00001306, -0.00000817, 0.00000441, -0.00000345, 0.00000095, 0.00000383, -0.00000571, -0.00000229, 0.00000665, 0.00001230, -0.00002854, 0.00000620, 0.00000832, 0.00000989, -0.00001493, 0.00000477, -0.00000444, },
		{ 0.00000394, -0.00000242, 0.00000442, -0.00000955, -0.00000188, 0.00000932, 0.00001290, -0.00001534, -0.00001316, 0.00001531, 0.00000110, -0.00000109, -0.00000696, 0.00000751, -0.00000450, 0.00000188, -0.00000296, 0.00000283, 0.00000264, -0.00000712, -0.00000006, -0.00000303, 0.00000874, -0.00000550, 0.00001058, -0.00001511, 0.00001164, -0.00001433, 0.00002641, -0.00003383, 0.00001731, -0.00001064, 0.00002845, -0.00003334, 0.00002328, -0.00002394, 0.00002770, -0.00004195, 0.00005143, -0.00005630, 0.00006143, -0.00005252, 0.00004262, -0.00004497, 0.00003703, -0.00002848, 0.00001985, -0.00001591, 0.00002660, -0.00004243, 0.00004269, -0.00005150, 0.00005778, -0.00006889, 0.00007026, -0.00008460, 0.00008244, -0.00010274, 0.00011579, -0.00016098, 0.00016037, -0.00025443, 0.00030728, -0.00081043, 0.00100107, 0.00022435, -0.00052994, 0.00024603, -0.00021795, 0.00013986, -0.00014536, 0.00011089, -0.00009933, 0.00007831, -0.00008087, 0.00006966, -0.00007001, 0.00006209, -0.00005804, 0.00005132, -0.00005408, 0.00004157, -0.00002080, 0.00001152, -0.00000995, 0.00000849, -0.00001866, 0.00001917, -0.00003233, 0.00004404, -0.00004714, 0.00004614, -0.00004011, 0.00003024, -0.00002273, 0.00001628, -0.00002761, 0.00003232, -0.00001366, 0.00000832, -0.00002555, 0.00003042, -0.00002304, 0.00001827, -0.00002111, 0.00001961, -0.00001445, 0.00001651, -0.00001127, 0.00000633, -0.00001266, 0.00001162, -0.00000543, 0.00000144, -0.00000028, -0.00000217, 0.00000560, -0.00000424, -0.00000399, 0.00000311, 0.00001705, -0.00002533, -0.00000130, 0.00001045, 0.00001005, -0.00001162, 0.00000051, -0.00000144, },
		{ 0.00000045, 0.00000105, 0.00000095, -0.00000433, -0.00000759, 0.00001093, 0.00001116, -0.00000697, -0.00002055, 0.00001483, 0.00000253, 0.00000154, -0.00000904, 0.00000593, -0.00000096, -0.00000212, 0.00000125, -0.00000235, 0.00000821, -0.00001150, 0.00000455, -0.00000981, 0.00001526, -0.00001201, 0.00001806, -0.00002192, 0.00001836, -0.00002252, 0.00003387, -0.00003746, 0.00001971, -0.00001707, 0.00003614, -0.00003803, 0.00002788, -0.00003091, 0.00003661, -0.00005105, 0.00005585, -0.00005527, 0.00005477, -0.00004082, 0.00003204, -0.00003528, 0.00002780, -0.00002013, 0.00001540, -0.00001729, 0.00003187, -0.00004786, 0.00004770, -0.00005761, 0.00006399, -0.00007409, 0.00007514, -0.00008915, 0.00008819, -0.00011071, 0.00012510, -0.00016677, 0.00016691, -0.00025821, 0.00030679, -0.00069662, 0.00059319, 0.00067635, -0.00072921, 0.00031614, -0.00026478, 0.00017342, -0.00017269, 0.00012967, -0.00011528, 0.00009350, -0.00009514, 0.00008172, -0.00008001, 0.00006960, -0.00006417, 0.00005678, -0.00005686, 0.00004129, -0.00002177, 0.00001734, -0.00001990, 0.00002150, -0.00003452, 0.00003681, -0.00005037, 0.00005662, -0.00005384, 0.00004700, -0.00003703, 0.00002715, -0.00002083, 0.00001735, -0.00002969, 0.00003094, -0.00001195, 0.00001096, -0.00002903, 0.00003055, -0.00002112, 0.00001678, -0.00001928, 0.00001670, -0.00001164, 0.00001312, -0.00000738, 0.00000425, -0.00001059, 0.00000780, -0.00000104, -0.00000232, 0.00000273, -0.00000409, 0.00000537, -0.00000125, -0.00000578, 0.00000007, 0.00001878, -0.00001850, -0.00001012, 0.00001279, 0.00000895, -0.00000669, -0.00000472, 0.00000225, },
		{ -0.00000264, 0.00000427, -0.00000242, 0.00000055, -0.00001220, 0.00001160, 0.00000926, 0.00000068, -0.00002496, 0.00001174, 0.00000446, 0.00000367, -0.00000916, 0.00000230, 0.00000352, -0.00000598, 0.00000460, -0.00000610, 0.00001171, -0.00001350, 0.00000682, -0.00001370, 0.00001804, -0.00001436, 0.00002080, -0.00002334, 0.00001930, -0.00002421, 0.00003381, -0.00003333, 0.00001495, -0.00001612, 0.00003516, -0.00003336, 0.00002305, -0.00002765, 0.00003359, -0.00004572, 0.00004396, -0.00003791, 0.00003301, -0.00001663, 0.00001137, -0.00001630, 0.00000962, -0.00000334, 0.00000285, -0.00000947, 0.00002575, -0.00003990, 0.00003791, -0.00004731, 0.00005177, -0.00005899, 0.00005795, -0.00006967, 0.00006733, -0.00008811, 0.00009825, -0.00013006, 0.00012259, -0.00019776, 0.00021803, -0.00047714, 0.00013989, 0.00105448, -0.00080474, 0.00028917, -0.00024242, 0.00015245, -0.00015462, 0.00010997, -0.00009860, 0.00007997, -0.00008321, 0.00006960, -0.00006775, 0.00005684, -0.00005200, 0.00004533, -0.00004415, 0.00002783, -0.00001215, 0.00001394, -0.00002064, 0.00002497, -0.00003996, 0.00004223, -0.00005358, 0.00005278, -0.00004448, 0.00003361, -0.00002270, 0.00001534, -0.00001173, 0.00001182, -0.00002489, 0.00002302, -0.00000494, 0.00000859, -0.00002672, 0.00002471, -0.00001397, 0.00001067, -0.00001315, 0.00001003, -0.00000556, 0.00000694, -0.00000152, 0.00000056, -0.00000675, 0.00000256, 0.00000390, -0.00000576, 0.00000465, -0.00000429, 0.00000335, 0.00000249, -0.00000703, -0.00000256, 0.00001785, -0.00000944, -0.00001850, 0.00001403, 0.00000767, -0.00000141, -0.00000955, 0.00000536, },
		{ -0.00000433, 0.00000623, -0.00000460, 0.00000395, -0.00001468, 0.00001053, 0.00000815, 0.00000639, -0.00002572, 0.00000627, 0.00000667, 0.00000505, -0.00000735, -0.00000259, 0.00000767, -0.00000837, 0.00000594, -0.00000743, 0.00001228, -0.00001240, 0.00000600, -0.00001385, 0.00001630, -0.00001191, 0.00001826, -0.00001895, 0.00001420, -0.00001927, 0.00002638, -0.00002193, 0.00000362, -0.00000830, 0.00002609, -0.00002020, 0.00000975, -0.00001517, 0.00001994, -0.00002812, 0.00001944, -0.00000953, 0.00000265, 0.00001314, -0.00001300, 0.00000627, -0.00001230, 0.00001705, -0.00001349, 0.00000378, 0.00001194, -0.00002256, 0.00001765, -0.00002528, 0.00002630, -0.00002948, 0.00002534, -0.00003350, 0.00002796, -0.00004413, 0.00004623, -0.00006460, 0.00004497, -0.00009703, 0.00007851, -0.00021492, -0.00027488, 0.00128718, -0.00072204, 0.00015745, -0.00015194, 0.00007992, -0.00009480, 0.00005572, -0.00005305, 0.00004107, -0.00004817, 0.00003635, -0.00003634, 0.00002695, -0.00002450, 0.00001983, -0.00001881, 0.00000394, 0.00000590, 0.00000257, -0.00001268, 0.00001889, -0.00003451, 0.00003486, -0.00004214, 0.00003421, -0.00002235, 0.00001037, -0.00000178, -0.00000101, 0.00000110, 0.00000254, -0.00001571, 0.00001105, 0.00000509, 0.00000307, -0.00002034, 0.00001480, -0.00000355, 0.00000186, -0.00000461, 0.00000143, 0.00000204, -0.00000033, 0.00000479, -0.00000353, -0.00000226, -0.00000290, 0.00000824, -0.00000796, 0.00000496, -0.00000286, 0.00000028, 0.00000587, -0.00000702, -0.00000505, 0.00001505, 0.00000010, -0.00002464, 0.00001301, 0.00000718, 0.00000300, -0.00001274, 0.00000680, },
		{ -0.00000416, 0.00000642, -0.00000502, 0.00000525, -0.00001461, 0.00000752, 0.00000824, 0.00000968, -0.00002305, -0.00000072, 0.00000859, 0.00000579, -0.00000411, -0.00000758, 0.00001020, -0.00000835, 0.00000472, -0.00000609, 0.00000993, -0.00000840, 0.00000233, -0.00001055, 0.00001059, -0.00000549, 0.00001145, -0.00001002, 0.00000456, -0.00000947, 0.00001373, -0.00000585, -0.00001154, 0.00000364, 0.00001194, -0.00000198, -0.00000826, 0.00000246, 0.00000034, -0.00000406, -0.00001033, 0.00002127, -0.00002730, 0.00003983, -0.00003336, 0.00002561, -0.00003151, 0.00003481, -0.00002778, 0.00001682, -0.00000362, -0.00000240, -0.00000591, 0.00000068, -0.00000377, 0.00000478, -0.00001206, 0.00000767, -0.00001705, 0.00000665, -0.00001384, 0.00000903, -0.00004080, 0.00001183, -0.00006692, 0.00002645, -0.00058141, 0.00133239, -0.00048053, -0.00005846, -0.00001522, -0.00002535, -0.00000970, -0.00001847, 0.00000851, -0.00001192, -0.00000025, -0.00000850, 0.00000522, -0.00001161, 0.00001046, -0.00001240, 0.00001221, -0.00002398, 0.00002693, -0.00001238, 0.00000027, 0.00000652, -0.00002125, 0.00001810, -0.00002052, 0.00000688, 0.00000549, -0.00001536, 0.00001898, -0.00001626, 0.00001303, -0.00000662, -0.00000569, -0.00000158, 0.00001506, -0.00000296, -0.00001242, 0.00000344, 0.00000749, -0.00000717, 0.00000401, -0.00000693, 0.00000916, -0.00000687, 0.00001001, -0.00000679, 0.00000173, -0.00000745, 0.00001103, -0.00000838, 0.00000366, -0.00000041, -0.00000275, 0.00000781, -0.00000524, -0.00000763, 0.00001131, 0.00000843, -0.00002718, 0.00000913, 0.00000801, 0.00000577, -0.00001356, 0.00000605, },
		{ -0.00000232, 0.00000498, -0.00000374, 0.00000453, -0.00001228, 0.00000298, 0.00000929, 0.00001077, -0.00001784, -0.00000796, 0.00000954, 0.00000620, -0.00000017, -0.00001151, 0.00001023, -0.00000566, 0.00000122, -0.00000268, 0.00000550, -0.00000255, -0.00000304, -0.00000510, 0.00000256, 0.00000293, 0.00000261, 0.00000090, -0.00000675, 0.00000204, -0.00000056, 0.00001098, -0.00002647, 0.00001558, -0.00000280, 0.00001632, -0.00002567, 0.00001950, -0.00001877, 0.00001891, -0.00003667, 0.00004538, -0.00004806, 0.00005569, -0.00004325, 0.00003602, -0.00004247, 0.00004449, -0.00003472, 0.00002429, -0.00001497, 0.00001381, -0.00002528, 0.00002234, -0.00002929, 0.00003363, -0.00004308, 0.00004169, -0.00005425, 0.00004906, -0.00006426, 0.00007002, -0.00010995, 0.00009835, -0.00017866, 0.00019765, -0.00073848, 0.00118654, -0.00011540, -0.00031298, 0.00012991, -0.00013352, 0.00007593, -0.00009149, 0.00006795, -0.00006315, 0.00004627, -0.00005180, 0.00004472, -0.00004759, 0.00004258, -0.00004183, 0.00004001, -0.00004784, 0.00004396, -0.00002492, 0.00001276, -0.00000682, -0.00000558, -0.00000206, 0.00000411, -0.00002093, 0.00003048, -0.00003564, 0.00003294, -0.00002526, 0.00002000, -0.00001222, 0.00000196, -0.00001171, 0.00002215, -0.00000704, -0.00000539, -0.00000678, 0.00001664, -0.00001411, 0.00001061, -0.00001311, 0.00001409, -0.00001117, 0.00001296, -0.00000830, 0.00000438, -0.00001028, 0.00001176, -0.00000697, 0.00000120, 0.00000220, -0.00000471, 0.00000761, -0.00000162, -0.00001023, 0.00000731, 0.00001439, -0.00002560, 0.00000262, 0.00000996, 0.00000682, -0.00001204, 0.00000328, },
		{ 0.00000043, 0.00000258, -0.00000139, 0.00000248, -0.00000854, -0.00000214, 0.00001062, 0.00001045, -0.00001143, -0.00001410, 0.00000892, 0.00000654, 0.00000371, -0.00001350, 0.00000752, -0.00000087, -0.00000353, 0.00000152, 0.00000047, 0.00000351, -0.00000840, 0.00000059, -0.00000550, 0.00001079, -0.00000552, 0.00001077, -0.00001645, 0.00001174, -0.00001264, 0.00002441, -0.00003699, 0.00002331, -0.00001360, 0.00002979, -0.00003729, 0.00003046, -0.00003133, 0.00003398, -0.00005229, 0.00005587, -0.00005384, 0.00005631, -0.00003951, 0.00003477, -0.00004230, 0.00004311, -0.00003136, 0.00002292, -0.00001811, 0.00002132, -0.00003506, 0.00003367, -0.00004345, 0.00004949, -0.00005952, 0.00005961, -0.00007380, 0.00007189, -0.00009203, 0.00010332, -0.00014514, 0.00014208, -0.00023202, 0.00027348, -0.00073991, 0.00088424, 0.00030952, -0.00054563, 0.00023963, -0.00021241, 0.00013646, -0.00014209, 0.00010737, -0.00009713, 0.00007752, -0.00008083, 0.00007055, -0.00007043, 0.00006230, -0.00005968, 0.00005648, -0.00006038, 0.00005080, -0.00002963, 0.00001956, -0.00001582, 0.00000685, -0.00001919, 0.00002435, -0.00004134, 0.00004526, -0.00004446, 0.00003579, -0.00002512, 0.00001988, -0.00001243, 0.00000535, -0.00001737, 0.00002464, -0.00000769, -0.00000086, -0.00001406, 0.00002214, -0.00001744, 0.00001382, -0.00001594, 0.00001582, -0.00001242, 0.00001308, -0.00000770, 0.00000530, -0.00001106, 0.00001042, -0.00000411, -0.00000169, 0.00000410, -0.00000493, 0.00000519, 0.00000326, -0.00001238, 0.00000338, 0.00001755, -0.00002035, -0.00000542, 0.00001217, 0.00000672, -0.00000885, -0.00000068, },
		{ 0.00000304, 0.00000018, 0.00000111, 0.00000004, -0.00000451, -0.00000677, 0.00001136, 0.00000966, -0.00000515, -0.00001810, 0.00000653, 0.00000691, 0.00000700, -0.00001320, 0.00000258, 0.00000481, -0.00000799, 0.00000493, -0.00000355, 0.00000814, -0.00001205, 0.00000461, -0.00001143, 0.00001577, -0.00001051, 0.00001703, -0.00002185, 0.00001681, -0.00001948, 0.00003136, -0.00004015, 0.00002395, -0.00001738, 0.00003508, -0.00003973, 0.00003185, -0.00003360, 0.00003717, -0.00005345, 0.00005014, -0.00004352, 0.00004198, -0.00002328, 0.00002297, -0.00003169, 0.00003116, -0.00001809, 0.00001272, -0.00001230, 0.00001874, -0.00003349, 0.00003252, -0.00004365, 0.00004943, -0.00005824, 0.00005812, -0.00007199, 0.00007085, -0.00009213, 0.00010336, -0.00014056, 0.00013706, -0.00022169, 0.00025547, -0.00061230, 0.00048902, 0.00071589, -0.00069583, 0.00027653, -0.00023722, 0.00015317, -0.00015541, 0.00011464, -0.00010359, 0.00008433, -0.00008732, 0.00007527, -0.00007356, 0.00006381, -0.00006071, 0.00005687, -0.00005761, 0.00004421, -0.00002361, 0.00001759, -0.00001703, 0.00001206, -0.00002858, 0.00003484, -0.00004919, 0.00004589, -0.00003957, 0.00002689, -0.00001614, 0.00001317, -0.00000758, 0.00000442, -0.00001828, 0.00002233, -0.00000474, 0.00000073, -0.00001771, 0.00002340, -0.00001673, 0.00001332, -0.00001519, 0.00001423, -0.00001062, 0.00001055, -0.00000522, 0.00000458, -0.00000994, 0.00000740, -0.00000047, -0.00000427, 0.00000475, -0.00000335, 0.00000120, 0.00000828, -0.00001333, -0.00000056, 0.00001815, -0.00001270, -0.00001333, 0.00001347, 0.00000637, -0.00000505, -0.00000470, },
		{ 0.00000456, -0.00000138, 0.00000293, -0.00000187, -0.00000119, -0.00001001, 0.00001087, 0.00000914, 0.00000000, -0.00001946, 0.00000262, 0.00000712, 0.00000950, -0.00001083, -0.00000349, 0.00000979, -0.00001064, 0.00000623, -0.00000534, 0.00001013, -0.00001281, 0.00000573, -0.00001381, 0.00001648, -0.00001105, 0.00001829, -0.00002160, 0.00001595, -0.00001978, 0.00003062, -0.00003508, 0.00001683, -0.00001337, 0.00003145, -0.00003245, 0.00002327, -0.00002528, 0.00002842, -0.00004093, 0.00003046, -0.00002088, 0.00001746, 0.00000039, 0.00000509, -0.00001447, 0.00001223, 0.00000157, -0.00000322, -0.00000006, 0.00000829, -0.00002268, 0.00002090, -0.00003185, 0.00003555, -0.00004168, 0.00003994, -0.00005179, 0.00004910, -0.00006809, 0.00007462, -0.00010225, 0.00009163, -0.00016028, 0.00016754, -0.00040520, 0.00007785, 0.00102873, -0.00071828, 0.00022079, -0.00019847, 0.00012026, -0.00012766, 0.00008733, -0.00008071, 0.00006510, -0.00006991, 0.00005789, -0.00005641, 0.00004687, -0.00004484, 0.00004132, -0.00004003, 0.00002508, -0.00000763, 0.00000706, -0.00001007, 0.00000907, -0.00002863, 0.00003380, -0.00004345, 0.00003292, -0.00002311, 0.00000943, -0.00000169, 0.00000276, 0.00000012, 0.00000076, -0.00001563, 0.00001642, 0.00000072, 0.00000004, -0.00001813, 0.00002087, -0.00001255, 0.00000972, -0.00001150, 0.00001000, -0.00000648, 0.00000612, -0.00000150, 0.00000269, -0.00000737, 0.00000336, 0.00000321, -0.00000599, 0.00000403, -0.00000051, -0.00000320, 0.00001210, -0.00001228, -0.00000471, 0.00001689, -0.00000429, -0.00001939, 0.00001274, 0.00000662, -0.00000168, -0.00000769, },
		{ 0.00000451, -0.00000166, 0.00000363, -0.00000273, 0.00000084, -0.00001141, 0.00000894, 0.00000925, 0.00000361, -0.00001835, -0.00000207, 0.00000679, 0.00001125, -0.00000698, -0.00000936, 0.00001264, -0.00001044, 0.00000478, -0.00000446, 0.00000916, -0.00001048, 0.00000371, -0.00001242, 0.00001289, -0.00000725, 0.00001480, -0.00001611, 0.00000975, -0.00001428, 0.00002323, -0.00002322, 0.00000367, -0.00000337, 0.00002085, -0.00001779, 0.00000746, -0.00000944, 0.00001139, -0.00001948, 0.00000307, 0.00000678, -0.00000960, 0.00002429, -0.00001261, 0.00000379, -0.00000822, 0.00002220, -0.00001980, 0.00001389, -0.00000536, -0.00000748, 0.00000390, -0.00001343, 0.00001379, -0.00001646, 0.00001238, -0.00002118, 0.00001544, -0.00002993, 0.00002906, -0.00004490, 0.00002420, -0.00007227, 0.00004645, -0.00017695, -0.00027659, 0.00119307, -0.00059482, 0.00007700, -0.00010563, 0.00004685, -0.00006743, 0.00003341, -0.00003568, 0.00002622, -0.00003449, 0.00002408, -0.00002450, 0.00001681, -0.00001712, 0.00001469, -0.00001259, -0.00000175, 0.00001404, -0.00000859, 0.00000232, 0.00000005, -0.00002108, 0.00002320, -0.00002720, 0.00001093, -0.00000072, -0.00001074, 0.00001302, -0.00000723, 0.00000750, -0.00000325, -0.00001146, 0.00000892, 0.00000692, -0.00000161, -0.00001642, 0.00001573, -0.00000620, 0.00000428, -0.00000608, 0.00000431, -0.00000113, 0.00000088, 0.00000255, 0.00000028, -0.00000401, -0.00000095, 0.00000631, -0.00000659, 0.00000227, 0.00000267, -0.00000667, 0.00001354, -0.00000879, -0.00000917, 0.00001455, 0.00000331, -0.00002230, 0.00000942, 0.00000790, 0.00000062, -0.00000901, },
		{ 0.00000299, -0.00000077, 0.00000323, -0.00000252, 0.00000152, -0.00001108, 0.00000583, 0.00000985, 0.00000578, -0.00001547, -0.00000659, 0.00000551, 0.00001239, -0.00000241, -0.00001380, 0.00001249, -0.00000718, 0.00000085, -0.00000139, 0.00000585, -0.00000580, -0.00000064, -0.00000820, 0.00000624, -0.00000062, 0.00000821, -0.00000731, 0.00000038, -0.00000535, 0.00001195, -0.00000785, -0.00001199, 0.00000898, 0.00000721, -0.00000016, -0.00001075, 0.00000863, -0.00000797, 0.00000387, -0.00002392, 0.00003102, -0.00003119, 0.00004136, -0.00002422, 0.00001773, -0.00002469, 0.00003825, -0.00003169, 0.00002442, -0.00001686, 0.00000633, -0.00001230, 0.00000489, -0.00000838, 0.00000907, -0.00001539, 0.00000979, -0.00001895, 0.00000962, -0.00001828, 0.00001348, -0.00004336, 0.00001460, -0.00006958, 0.00001992, -0.00052113, 0.00118545, -0.00033966, -0.00012564, 0.00001491, -0.00004550, 0.00000696, -0.00003116, 0.00001761, -0.00002007, 0.00000778, -0.00001574, 0.00001236, -0.00001721, 0.00001394, -0.00001495, 0.00001693, -0.00002896, 0.00003498, -0.00002390, 0.00001534, -0.00001067, -0.00001012, 0.00000773, -0.00000634, -0.00001309, 0.00002036, -0.00002705, 0.00002279, -0.00001297, 0.00001173, -0.00000540, -0.00000777, 0.00000186, 0.00001206, -0.00000284, -0.00001383, 0.00000942, 0.00000082, -0.00000156, -0.00000030, -0.00000154, 0.00000421, -0.00000406, 0.00000608, -0.00000199, -0.00000053, -0.00000484, 0.00000837, -0.00000613, 0.00000006, 0.00000517, -0.00000813, 0.00001200, -0.00000302, -0.00001364, 0.00001165, 0.00000902, -0.00002156, 0.00000376, 0.00000999, 0.00000177, -0.00000864, },
		{ 0.00000060, 0.00000072, 0.00000223, -0.00000160, 0.00000123, -0.00000956, 0.00000219, 0.00001050, 0.00000700, -0.00001174, -0.00001005, 0.00000308, 0.00001300, 0.00000222, -0.00001604, 0.00000924, -0.00000157, -0.00000443, 0.00000260, 0.00000155, -0.00000026, -0.00000573, -0.00000291, -0.00000138, 0.00000653, 0.00000099, 0.00000208, -0.00000916, 0.00000377, 0.00000037, 0.00000704, -0.00002593, 0.00001942, -0.00000493, 0.00001546, -0.00002601, 0.00002324, -0.00002343, 0.00002216, -0.00004311, 0.00004489, -0.00004142, 0.00004696, -0.00002615, 0.00002391, -0.00003342, 0.00004577, -0.00003513, 0.00002774, -0.00002203, 0.00001411, -0.00002259, 0.00001740, -0.00002446, 0.00002762, -0.00003536, 0.00003229, -0.00004422, 0.00003922, -0.00005408, 0.00005718, -0.00009241, 0.00007743, -0.00015099, 0.00014988, -0.00063144, 0.00101755, 0.00000359, -0.00033994, 0.00012696, -0.00012932, 0.00007306, -0.00008746, 0.00006302, -0.00005965, 0.00004413, -0.00004977, 0.00004326, -0.00004518, 0.00003914, -0.00003898, 0.00004035, -0.00004907, 0.00004872, -0.00003329, 0.00002391, -0.00001826, -0.00000065, -0.00000702, 0.00001255, -0.00003207, 0.00003366, -0.00003443, 0.00002419, -0.00001233, 0.00001137, -0.00000455, -0.00000577, -0.00000340, 0.00001497, -0.00000275, -0.00001132, 0.00000314, 0.00000719, -0.00000655, 0.00000464, -0.00000644, 0.00000850, -0.00000780, 0.00000842, -0.00000361, 0.00000250, -0.00000777, 0.00000918, -0.00000483, -0.00000197, 0.00000625, -0.00000712, 0.00000769, 0.00000415, -0.00001741, 0.00000835, 0.00001249, -0.00001761, -0.00000321, 0.00001209, 0.00000228, -0.00000711, },
		{ -0.00000179, 0.00000206, 0.00000125, -0.00000058, 0.00000057, -0.00000756, -0.00000126, 0.00001066, 0.00000784, -0.00000803, -0.00001188, -0.00000035, 0.00001297, 0.00000648, -0.00001586, 0.00000362, 0.00000494, -0.00000940, 0.00000590, -0.00000211, 0.00000445, -0.00000980, 0.00000149, -0.00000774, 0.00001185, -0.00000441, 0.00000940, -0.00001606, 0.00001011, -0.00000826, 0.00001792, -0.00003459, 0.00002446, -0.00001191, 0.00002511, -0.00003420, 0.00003013, -0.00003051, 0.00003067, -0.00005004, 0.00004500, -0.00003821, 0.00004020, -0.00001817, 0.00002184, -0.00003332, 0.00004350, -0.00002899, 0.00002263, -0.00001921, 0.00001374, -0.00002445, 0.00002106, -0.00003078, 0.00003501, -0.00004294, 0.00004126, -0.00005460, 0.00005211, -0.00007032, 0.00007694, -0.00011236, 0.00010381, -0.00018304, 0.00019988, -0.00061365, 0.00073132, 0.00037208, -0.00051234, 0.00019487, -0.00017958, 0.00011131, -0.00011952, 0.00008721, -0.00008099, 0.00006417, -0.00006852, 0.00005960, -0.00005933, 0.00005144, -0.00005086, 0.00005157, -0.00005670, 0.00005083, -0.00003299, 0.00002448, -0.00001919, 0.00000351, -0.00001651, 0.00002429, -0.00004104, 0.00003546, -0.00003086, 0.00001673, -0.00000567, 0.00000693, -0.00000106, -0.00000549, -0.00000648, 0.00001535, -0.00000117, -0.00000930, -0.00000236, 0.00001207, -0.00000989, 0.00000799, -0.00000965, 0.00001109, -0.00000979, 0.00000920, -0.00000428, 0.00000467, -0.00000939, 0.00000870, -0.00000303, -0.00000336, 0.00000564, -0.00000393, 0.00000166, 0.00001124, -0.00001946, 0.00000444, 0.00001398, -0.00001161, -0.00000997, 0.00001311, 0.00000289, -0.00000522, },
		{ -0.00000341, 0.00000261, 0.00000085, 0.00000006, 0.00000007, -0.00000573, -0.00000393, 0.00000998, 0.00000869, -0.00000489, -0.00001200, -0.00000423, 0.00001199, 0.00001025, -0.00001361, -0.00000309, 0.00001060, -0.00001239, 0.00000711, -0.00000383, 0.00000698, -0.00001149, 0.00000353, -0.00001121, 0.00001370, -0.00000638, 0.00001292, -0.00001853, 0.00001189, -0.00001200, 0.00002275, -0.00003612, 0.00002243, -0.00001202, 0.00002700, -0.00003362, 0.00002775, -0.00002777, 0.00002814, -0.00004421, 0.00003227, -0.00002378, 0.00002408, -0.00000334, 0.00001384, -0.00002601, 0.00003299, -0.00001498, 0.00001068, -0.00000957, 0.00000610, -0.00001850, 0.00001616, -0.00002731, 0.00003103, -0.00003788, 0.00003637, -0.00004954, 0.00004743, -0.00006586, 0.00007165, -0.00010245, 0.00009351, -0.00016701, 0.00017835, -0.00049862, 0.00038749, 0.00069836, -0.00059653, 0.00019399, -0.00018149, 0.00011100, -0.00011913, 0.00008372, -0.00007868, 0.00006304, -0.00006765, 0.00005757, -0.00005645, 0.00004808, -0.00004807, 0.00004837, -0.00005023, 0.00004034, -0.00002235, 0.00001634, -0.00001254, 0.00000098, -0.00001872, 0.00002661, -0.00003853, 0.00002585, -0.00001797, 0.00000301, 0.00000435, 0.00000058, 0.00000352, -0.00000604, -0.00000788, 0.00001371, 0.00000141, -0.00000765, -0.00000680, 0.00001509, -0.00001127, 0.00000943, -0.00001091, 0.00001174, -0.00000985, 0.00000835, -0.00000390, 0.00000575, -0.00000953, 0.00000705, -0.00000098, -0.00000397, 0.00000365, 0.00000050, -0.00000453, 0.00001658, -0.00001882, -0.00000034, 0.00001412, -0.00000506, -0.00001494, 0.00001210, 0.00000424, -0.00000374, },
		{ -0.00000390, 0.00000212, 0.00000121, 0.00000014, -0.00000004, -0.00000437, -0.00000560, 0.00000845, 0.00000963, -0.00000245, -0.00001082, -0.00000777, 0.00000972, 0.00001352, -0.00000996, -0.00000941, 0.00001385, -0.00001229, 0.00000550, -0.00000303, 0.00000679, -0.00001030, 0.00000266, -0.00001122, 0.00001168, -0.00000460, 0.00001231, -0.00001634, 0.00000895, -0.00001068, 0.00002147, -0.00003078, 0.00001392, -0.00000596, 0.00002190, -0.00002532, 0.00001752, -0.00001696, 0.00001688, -0.00002896, 0.00001133, -0.00000359, 0.00000422, 0.00001321, 0.00000401, -0.00001495, 0.00001778, 0.00000310, -0.00000441, 0.00000348, -0.00000556, -0.00000792, 0.00000575, -0.00001706, 0.00001890, -0.00002376, 0.00002146, -0.00003313, 0.00002958, -0.00004568, 0.00004737, -0.00007035, 0.00005647, -0.00011680, 0.00010900, -0.00033124, 0.00005078, 0.00092562, -0.00056525, 0.00011820, -0.00013519, 0.00007348, -0.00008824, 0.00005480, -0.00005484, 0.00004263, -0.00004898, 0.00003912, -0.00003861, 0.00003116, -0.00003265, 0.00003284, -0.00003211, 0.00002006, -0.00000408, 0.00000177, -0.00000018, -0.00000686, -0.00001453, 0.00002053, -0.00002663, 0.00000840, -0.00000033, -0.00001227, 0.00001375, -0.00000472, 0.00000710, -0.00000607, -0.00000857, 0.00001102, 0.00000416, -0.00000600, -0.00001022, 0.00001626, -0.00001076, 0.00000906, -0.00001028, 0.00001053, -0.00000815, 0.00000609, -0.00000257, 0.00000567, -0.00000822, 0.00000444, 0.00000109, -0.00000393, 0.00000097, 0.00000484, -0.00000919, 0.00001877, -0.00001496, -0.00000604, 0.00001355, 0.00000067, -0.00001701, 0.00000867, 0.00000655, -0.00000303, },
		{ -0.00000340, 0.00000082, 0.00000209, -0.00000018, 0.00000017, -0.00000347, -0.00000638, 0.00000636, 0.00001045, -0.00000048, -0.00000902, -0.00001017, 0.00000598, 0.00001629, -0.00000563, -0.00001411, 0.00001380, -0.00000888, 0.00000131, -0.00000007, 0.00000431, -0.00000672, -0.00000059, -0.00000837, 0.00000662, -0.00000008, 0.00000864, -0.00001070, 0.00000268, -0.00000579, 0.00001576, -0.00002069, 0.00000146, 0.00000365, 0.00001262, -0.00001253, 0.00000310, -0.00000221, 0.00000158, -0.00000995, -0.00001119, 0.00001559, -0.00001311, 0.00002617, -0.00000345, -0.00000398, 0.00000209, 0.00002070, -0.00001819, 0.00001568, -0.00001695, 0.00000293, -0.00000573, -0.00000469, 0.00000376, -0.00000626, 0.00000271, -0.00001208, 0.00000601, -0.00001835, 0.00001440, -0.00002859, 0.00000816, -0.00005269, 0.00002181, -0.00015791, -0.00022434, 0.00101961, -0.00041708, -0.00001771, -0.00005580, 0.00001147, -0.00003796, 0.00001029, -0.00001811, 0.00001061, -0.00001958, 0.00001097, -0.00001223, 0.00000672, -0.00001024, 0.00001047, -0.00000794, -0.00000444, 0.00001671, -0.00001471, 0.00001387, -0.00001653, -0.00000711, 0.00000960, -0.00001004, -0.00001112, 0.00001611, -0.00002395, 0.00001869, -0.00000643, 0.00000802, -0.00000454, -0.00000937, 0.00000821, 0.00000631, -0.00000400, -0.00001273, 0.00001577, -0.00000867, 0.00000720, -0.00000812, 0.00000784, -0.00000506, 0.00000282, -0.00000053, 0.00000454, -0.00000567, 0.00000119, 0.00000307, -0.00000354, -0.00000153, 0.00000784, -0.00001110, 0.00001716, -0.00000810, -0.00001223, 0.00001259, 0.00000474, -0.00001588, 0.00000319, 0.00000943, -0.00000294, },
		{ -0.00000249, -0.00000069, 0.00000297, -0.00000046, 0.00000034, -0.00000272, -0.00000668, 0.00000417, 0.00001078, 0.00000138, -0.00000730, -0.00001086, 0.00000094, 0.00001836, -0.00000114, -0.00001649, 0.00001046, -0.00000294, -0.00000432, 0.00000385, 0.00000072, -0.00000203, -0.00000489, -0.00000416, 0.00000030, 0.00000530, 0.00000386, -0.00000375, -0.00000460, 0.00000023, 0.00000834, -0.00000903, -0.00001149, 0.00001322, 0.00000294, 0.00000058, -0.00001094, 0.00001158, -0.00001236, 0.00000677, -0.00002893, 0.00002791, -0.00002304, 0.00003181, -0.00000572, 0.00000404, -0.00001059, 0.00003392, -0.00002684, 0.00002335, -0.00002426, 0.00001011, -0.00001417, 0.00000530, -0.00000932, 0.00000893, -0.00001364, 0.00000671, -0.00001550, 0.00000711, -0.00001651, 0.00000994, -0.00003581, 0.00000550, -0.00005591, -0.00001580, -0.00040385, 0.00097451, -0.00017661, -0.00018157, 0.00003100, -0.00005521, 0.00001527, -0.00003578, 0.00001943, -0.00002240, 0.00001084, -0.00001784, 0.00001425, -0.00001743, 0.00001191, -0.00001179, 0.00001543, -0.00002662, 0.00003413, -0.00002792, 0.00002495, -0.00002380, -0.00000049, -0.00000159, 0.00000571, -0.00002679, 0.00002608, -0.00002812, 0.00001690, -0.00000344, 0.00000578, -0.00000124, -0.00001057, 0.00000577, 0.00000746, -0.00000158, -0.00001434, 0.00001384, -0.00000538, 0.00000427, -0.00000487, 0.00000416, -0.00000112, -0.00000097, 0.00000185, 0.00000263, -0.00000226, -0.00000228, 0.00000480, -0.00000313, -0.00000317, 0.00000872, -0.00000979, 0.00001207, 0.00000069, -0.00001797, 0.00001113, 0.00000702, -0.00001210, -0.00000323, 0.00001197, -0.00000289, },
		{ -0.00000188, -0.00000169, 0.00000325, -0.00000023, 0.00000011, -0.00000182, -0.00000688, 0.00000235, 0.00001037, 0.00000343, -0.00000614, -0.00000979, -0.00000481, 0.00001932, 0.00000328, -0.00001649, 0.00000465, 0.00000398, -0.00000970, 0.00000717, -0.00000245, 0.00000218, -0.00000858, -0.00000038, -0.00000529, 0.00000945, 0.00000004, 0.00000225, -0.00001048, 0.00000490, 0.00000194, 0.00000112, -0.00002164, 0.00001950, -0.00000378, 0.00001037, -0.00002072, 0.00002035, -0.00002071, 0.00001676, -0.00003773, 0.00003025, -0.00002362, 0.00002915, -0.00000228, 0.00000808, -0.00001839, 0.00004057, -0.00002833, 0.00002454, -0.00002538, 0.00001134, -0.00001708, 0.00001003, -0.00001687, 0.00001785, -0.00002321, 0.00001835, -0.00002930, 0.00002401, -0.00003731, 0.00003575, -0.00006417, 0.00004388, -0.00010597, 0.00007384, -0.00047827, 0.00081167, 0.00011190, -0.00033215, 0.00009673, -0.00010608, 0.00005509, -0.00006986, 0.00004635, -0.00004642, 0.00003320, -0.00003894, 0.00003316, -0.00003430, 0.00002738, -0.00002778, 0.00003201, -0.00004094, 0.00004341, -0.00003374, 0.00002934, -0.00002532, 0.00000196, -0.00000912, 0.00001612, -0.00003437, 0.00002661, -0.00002350, 0.00000852, 0.00000341, 0.00000132, 0.00000303, -0.00001175, 0.00000367, 0.00000773, 0.00000096, -0.00001488, 0.00001064, -0.00000127, 0.00000073, -0.00000107, 0.00000005, 0.00000305, -0.00000467, 0.00000415, 0.00000037, 0.00000145, -0.00000546, 0.00000609, -0.00000284, -0.00000361, 0.00000734, -0.00000573, 0.00000475, 0.00000970, -0.00002199, 0.00000874, 0.00000793, -0.00000690, -0.00000907, 0.00001309, -0.00000213, },
		{ -0.00000211, -0.00000171, 0.00000257, 0.00000075, -0.00000068, -0.00000060, -0.00000722, 0.00000109, 0.00000916, 0.00000568, -0.00000560, -0.00000739, -0.00001030, 0.00001861, 0.00000764, -0.00001459, -0.00000219, 0.00001005, -0.00001311, 0.00000847, -0.00000389, 0.00000458, -0.00001030, 0.00000151, -0.00000856, 0.00001087, -0.00000135, 0.00000572, -0.00001333, 0.00000656, -0.00000163, 0.00000767, -0.00002688, 0.00002056, -0.00000564, 0.00001480, -0.00002418, 0.00002216, -0.00002163, 0.00001836, -0.00003664, 0.00002293, -0.00001615, 0.00002005, 0.00000509, 0.00000896, -0.00002123, 0.00004041, -0.00002272, 0.00001942, -0.00002035, 0.00000655, -0.00001423, 0.00000890, -0.00001790, 0.00001923, -0.00002455, 0.00002109, -0.00003319, 0.00002952, -0.00004456, 0.00004484, -0.00007243, 0.00005712, -0.00012208, 0.00010714, -0.00046040, 0.00057213, 0.00039545, -0.00042995, 0.00011872, -0.00012654, 0.00007041, -0.00008311, 0.00005542, -0.00005519, 0.00004184, -0.00004717, 0.00003990, -0.00003984, 0.00003248, -0.00003397, 0.00003840, -0.00004449, 0.00004234, -0.00003043, 0.00002557, -0.00001979, -0.00000120, -0.00001104, 0.00001903, -0.00003249, 0.00001788, -0.00001178, -0.00000392, 0.00001161, -0.00000336, 0.00000680, -0.00001206, 0.00000147, 0.00000755, 0.00000312, -0.00001412, 0.00000635, 0.00000326, -0.00000295, 0.00000276, -0.00000387, 0.00000682, -0.00000769, 0.00000587, -0.00000176, 0.00000483, -0.00000780, 0.00000668, -0.00000265, -0.00000299, 0.00000429, -0.00000016, -0.00000292, 0.00001696, -0.00002305, 0.00000494, 0.00000821, -0.00000171, -0.00001287, 0.00001193, -0.00000015, },
		{ -0.00000329, -0.00000068, 0.00000099, 0.00000238, -0.00000190, 0.00000082, -0.00000763, 0.00000030, 0.00000743, 0.00000788, -0.00000538, -0.00000446, -0.00001446, 0.00001575, 0.00001201, -0.00001149, -0.00000854, 0.00001366, -0.00001342, 0.00000699, -0.00000298, 0.00000455, -0.00000948, 0.00000090, -0.00000888, 0.00000908, 0.00000008, 0.00000622, -0.00001271, 0.00000485, -0.00000189, 0.00001005, -0.00002682, 0.00001631, -0.00000264, 0.00001389, -0.00002152, 0.00001750, -0.00001594, 0.00001284, -0.00002785, 0.00000924, -0.00000448, 0.00000833, 0.00001322, 0.00000868, -0.00002039, 0.00003486, -0.00001185, 0.00000998, -0.00001107, -0.00000236, -0.00000741, 0.00000344, -0.00001372, 0.00001438, -0.00001904, 0.00001626, -0.00002839, 0.00002479, -0.00003943, 0.00003879, -0.00006273, 0.00004813, -0.00010865, 0.00009445, -0.00037836, 0.00030551, 0.00062470, -0.00044733, 0.00008696, -0.00011239, 0.00005864, -0.00007391, 0.00004564, -0.00004792, 0.00003599, -0.00004191, 0.00003410, -0.00003392, 0.00002716, -0.00003022, 0.00003454, -0.00003760, 0.00003175, -0.00001908, 0.00001476, -0.00000833, -0.00000908, -0.00000782, 0.00001503, -0.00002282, 0.00000301, 0.00000303, -0.00001637, 0.00001786, -0.00000592, 0.00000855, -0.00001066, -0.00000129, 0.00000743, 0.00000442, -0.00001197, 0.00000127, 0.00000773, -0.00000625, 0.00000604, -0.00000702, 0.00000959, -0.00000951, 0.00000659, -0.00000324, 0.00000726, -0.00000881, 0.00000633, -0.00000233, -0.00000182, 0.00000059, 0.00000531, -0.00000900, 0.00002082, -0.00002041, -0.00000044, 0.00000849, 0.00000222, -0.00001370, 0.00000825, 0.00000307, },
		{ -0.00000496, 0.00000097, -0.00000100, 0.00000415, -0.00000309, 0.00000204, -0.00000781, -0.00000038, 0.00000566, 0.00000958, -0.00000496, -0.00000194, -0.00001638, 0.00001060, 0.00001629, -0.00000787, -0.00001319, 0.00001391, -0.00001035, 0.00000290, 0.00000003, 0.00000236, -0.00000643, -0.00000186, -0.00000670, 0.00000476, 0.00000354, 0.00000448, -0.00000944, 0.00000069, 0.00000025, 0.00000917, -0.00002264, 0.00000831, 0.00000349, 0.00000947, -0.00001484, 0.00000885, -0.00000651, 0.00000357, -0.00001555, -0.00000594, 0.00000660, -0.00000180, 0.00001884, 0.00000934, -0.00001757, 0.00002610, 0.00000150, -0.00000083, -0.00000046, -0.00001244, 0.00000048, -0.00000359, -0.00000706, 0.00000624, -0.00000989, 0.00000726, -0.00001846, 0.00001370, -0.00002640, 0.00002305, -0.00004183, 0.00002535, -0.00007703, 0.00005425, -0.00026612, 0.00005830, 0.00076457, -0.00037533, 0.00000720, -0.00007096, 0.00002599, -0.00004784, 0.00002204, -0.00002900, 0.00001953, -0.00002680, 0.00001927, -0.00001994, 0.00001461, -0.00001943, 0.00002327, -0.00002342, 0.00001508, -0.00000310, 0.00000016, 0.00000590, -0.00001888, -0.00000190, 0.00000692, -0.00000923, -0.00001316, 0.00001598, -0.00002465, 0.00001932, -0.00000472, 0.00000738, -0.00000723, -0.00000473, 0.00000763, 0.00000467, -0.00000862, -0.00000406, 0.00001158, -0.00000865, 0.00000826, -0.00000888, 0.00001089, -0.00000974, 0.00000599, -0.00000369, 0.00000828, -0.00000820, 0.00000491, -0.00000164, -0.00000072, -0.00000261, 0.00000916, -0.00001200, 0.00002043, -0.00001416, -0.00000703, 0.00000900, 0.00000428, -0.00001146, 0.00000259, 0.00000695, },
		{ -0.00000634, 0.00000247, -0.00000265, 0.00000538, -0.00000367, 0.00000259, -0.00000735, -0.00000136, 0.00000433, 0.00001033, -0.00000386, -0.00000057, -0.00001563, 0.00000352, 0.00002011, -0.00000413, -0.00001552, 0.00001083, -0.00000468, -0.00000273, 0.00000405, -0.00000094, -0.00000229, -0.00000560, -0.00000330, -0.00000058, 0.00000747, 0.00000202, -0.00000513, -0.00000419, 0.00000311, 0.00000680, -0.00001647, -0.00000095, 0.00001013, 0.00000423, -0.00000711, -0.00000046, 0.00000301, -0.00000543, -0.00000424, -0.00001795, 0.00001306, -0.00000723, 0.00001994, 0.00001206, -0.00001404, 0.00001623, 0.00001457, -0.00001013, 0.00000864, -0.00002078, 0.00000658, -0.00000940, -0.00000088, -0.00000189, -0.00000074, -0.00000202, -0.00000767, 0.00000108, -0.00001115, 0.00000452, -0.00001806, -0.00000101, -0.00004050, 0.00000629, -0.00015463, -0.00013538, 0.00080055, -0.00022532, -0.00010063, -0.00001846, -0.00001507, -0.00001529, -0.00000651, -0.00000606, -0.00000083, -0.00000801, 0.00000124, -0.00000335, -0.00000018, -0.00000622, 0.00000909, -0.00000661, -0.00000292, 0.00001299, -0.00001397, 0.00001891, -0.00002701, 0.00000336, -0.00000156, 0.00000368, -0.00002570, 0.00002282, -0.00002595, 0.00001469, 0.00000051, 0.00000353, -0.00000225, -0.00000846, 0.00000799, 0.00000408, -0.00000461, -0.00000895, 0.00001419, -0.00000966, 0.00000900, -0.00000911, 0.00001044, -0.00000822, 0.00000402, -0.00000297, 0.00000771, -0.00000597, 0.00000251, -0.00000044, -0.00000024, -0.00000443, 0.00001045, -0.00001135, 0.00001601, -0.00000530, -0.00001384, 0.00000945, 0.00000455, -0.00000689, -0.00000382, 0.00001040, },
		{ -0.00000654, 0.00000301, -0.00000323, 0.00000549, -0.00000316, 0.00000210, -0.00000598, -0.00000293, 0.00000376, 0.00000994, -0.00000185, -0.00000064, -0.00001243, -0.00000454, 0.00002276, -0.00000030, -0.00001559, 0.00000531, 0.00000208, -0.00000827, 0.00000762, -0.00000392, 0.00000148, -0.00000883, -0.00000027, -0.00000519, 0.00001014, 0.00000047, -0.00000150, -0.00000801, 0.00000491, 0.00000478, -0.00001049, -0.00000899, 0.00001480, 0.00000069, -0.00000104, -0.00000749, 0.00000956, -0.00001098, 0.00000277, -0.00002387, 0.00001300, -0.00000706, 0.00001641, 0.00001657, -0.00001016, 0.00000670, 0.00002533, -0.00001588, 0.00001430, -0.00002540, 0.00000897, -0.00001207, 0.00000271, -0.00000753, 0.00000562, -0.00000854, 0.00000058, -0.00000907, 0.00000147, -0.00001091, 0.00000151, -0.00002241, -0.00000999, -0.00003388, -0.00006553, -0.00025864, 0.00073943, -0.00002539, -0.00020798, 0.00002538, -0.00005042, 0.00001239, -0.00003056, 0.00001294, -0.00001808, 0.00000805, -0.00001405, 0.00001041, -0.00001227, 0.00000489, -0.00000353, 0.00000822, -0.00001773, 0.00002502, -0.00002381, 0.00002717, -0.00003037, 0.00000508, -0.00000712, 0.00001214, -0.00003117, 0.00002142, -0.00001973, 0.00000479, 0.00000841, -0.00000161, 0.00000303, -0.00001158, 0.00000800, 0.00000318, -0.00000071, -0.00001261, 0.00001499, -0.00000893, 0.00000804, -0.00000762, 0.00000828, -0.00000513, 0.00000090, -0.00000122, 0.00000575, -0.00000248, -0.00000051, 0.00000117, -0.00000059, -0.00000452, 0.00000904, -0.00000749, 0.00000881, 0.00000439, -0.00001946, 0.00000918, 0.00000372, -0.00000135, -0.00000938, 0.00001223, },
		{ -0.00000501, 0.00000210, -0.00000233, 0.00000421, -0.00000143, 0.00000054, -0.00000374, -0.00000506, 0.00000392, 0.00000854, 0.00000090, -0.00000192, -0.00000761, -0.00001225, 0.00002341, 0.00000380, -0.00001394, -0.00000122, 0.00000813, -0.00001208, 0.00000935, -0.00000533, 0.00000363, -0.00001028, 0.00000105, -0.00000768, 0.00001031, 0.00000091, 0.00000031, -0.00000958, 0.00000455, 0.00000433, -0.00000620, -0.00001417, 0.00001599, 0.00000026, 0.00000187, -0.00001070, 0.00001164, -0.00001174, 0.00000435, -0.00002326, 0.00000710, -0.00000267, 0.00001002, 0.00002126, -0.00000543, -0.00000188, 0.00003278, -0.00001729, 0.00001587, -0.00002569, 0.00000714, -0.00001115, 0.00000302, -0.00000966, 0.00000797, -0.00001092, 0.00000463, -0.00001461, 0.00000876, -0.00001986, 0.00001290, -0.00003413, 0.00000848, -0.00005789, -0.00000859, -0.00031181, 0.00060499, 0.00018720, -0.00028563, 0.00004374, -0.00006918, 0.00002681, -0.00004336, 0.00002242, -0.00002736, 0.00001690, -0.00002252, 0.00001767, -0.00001843, 0.00001090, -0.00001155, 0.00001792, -0.00002635, 0.00003041, -0.00002714, 0.00002874, -0.00002742, 0.00000187, -0.00000807, 0.00001436, -0.00002861, 0.00001232, -0.00000796, -0.00000770, 0.00001635, -0.00000595, 0.00000702, -0.00001306, 0.00000704, 0.00000261, 0.00000231, -0.00001442, 0.00001371, -0.00000642, 0.00000548, -0.00000462, 0.00000477, -0.00000096, -0.00000284, 0.00000110, 0.00000289, 0.00000160, -0.00000356, 0.00000288, -0.00000164, -0.00000312, 0.00000558, -0.00000171, 0.00000075, 0.00001287, -0.00002249, 0.00000747, 0.00000272, 0.00000361, -0.00001262, 0.00001155, },
		{ -0.00000176, -0.00000025, -0.00000009, 0.00000180, 0.00000119, -0.00000171, -0.00000103, -0.00000737, 0.00000441, 0.00000662, 0.00000389, -0.00000377, -0.00000239, -0.00001818, 0.00002135, 0.00000836, -0.00001135, -0.00000724, 0.00001189, -0.00001300, 0.00000848, -0.00000451, 0.00000350, -0.00000936, 0.00000009, -0.00000753, 0.00000771, 0.00000349, 0.00000009, -0.00000871, 0.00000195, 0.00000557, -0.00000395, -0.00001612, 0.00001358, 0.00000289, 0.00000167, -0.00001022, 0.00000965, -0.00000838, 0.00000156, -0.00001799, -0.00000193, 0.00000296, 0.00000358, 0.00002402, 0.00000094, -0.00000936, 0.00003673, -0.00001462, 0.00001392, -0.00002231, 0.00000192, -0.00000746, 0.00000066, -0.00000870, 0.00000668, -0.00000954, 0.00000468, -0.00001551, 0.00001040, -0.00002186, 0.00001564, -0.00003570, 0.00001426, -0.00006475, 0.00001711, -0.00030833, 0.00043030, 0.00037529, -0.00031150, 0.00002811, -0.00006727, 0.00002512, -0.00004288, 0.00002086, -0.00002733, 0.00001726, -0.00002305, 0.00001754, -0.00001796, 0.00001114, -0.00001410, 0.00002147, -0.00002798, 0.00002876, -0.00002379, 0.00002373, -0.00001863, -0.00000573, -0.00000471, 0.00001082, -0.00001959, -0.00000151, 0.00000558, -0.00001902, 0.00002138, -0.00000744, 0.00000838, -0.00001219, 0.00000471, 0.00000285, 0.00000391, -0.00001410, 0.00001042, -0.00000244, 0.00000178, -0.00000070, 0.00000059, 0.00000353, -0.00000645, 0.00000333, -0.00000013, 0.00000545, -0.00000594, 0.00000421, -0.00000295, -0.00000093, 0.00000128, 0.00000423, -0.00000610, 0.00001832, -0.00002195, 0.00000389, 0.00000233, 0.00000678, -0.00001266, 0.00000813, },
		{ 0.00000252, -0.00000341, 0.00000285, -0.00000105, 0.00000399, -0.00000397, 0.00000149, -0.00000926, 0.00000465, 0.00000486, 0.00000643, -0.00000535, 0.00000199, -0.00002121, 0.00001633, 0.00001333, -0.00000846, -0.00001161, 0.00001251, -0.00001077, 0.00000508, -0.00000163, 0.00000131, -0.00000638, -0.00000278, -0.00000519, 0.00000304, 0.00000736, -0.00000140, -0.00000621, -0.00000199, 0.00000770, -0.00000308, -0.00001562, 0.00000868, 0.00000728, -0.00000026, -0.00000758, 0.00000538, -0.00000302, -0.00000306, -0.00001122, -0.00001050, 0.00000657, -0.00000027, 0.00002318, 0.00000934, -0.00001558, 0.00003745, -0.00000880, 0.00000971, -0.00001667, -0.00000507, -0.00000262, -0.00000289, -0.00000602, 0.00000322, -0.00000596, 0.00000228, -0.00001324, 0.00000795, -0.00001867, 0.00001197, -0.00002990, 0.00001069, -0.00005875, 0.00001965, -0.00026842, 0.00024910, 0.00051016, -0.00027626, -0.00001903, -0.00004857, 0.00001048, -0.00003197, 0.00001083, -0.00002013, 0.00001100, -0.00001740, 0.00001178, -0.00001257, 0.00000711, -0.00001236, 0.00001991, -0.00002385, 0.00002161, -0.00001551, 0.00001408, -0.00000621, -0.00001554, 0.00000097, 0.00000389, -0.00000751, -0.00001578, 0.00001655, -0.00002564, 0.00002125, -0.00000491, 0.00000659, -0.00000891, 0.00000111, 0.00000395, 0.00000403, -0.00001186, 0.00000566, 0.00000233, -0.00000234, 0.00000334, -0.00000343, 0.00000749, -0.00000915, 0.00000478, -0.00000258, 0.00000826, -0.00000704, 0.00000474, -0.00000397, 0.00000115, -0.00000254, 0.00000870, -0.00001011, 0.00001970, -0.00001768, -0.00000140, 0.00000289, 0.00000757, -0.00000949, 0.00000258, },
		{ 0.00000665, -0.00000639, 0.00000551, -0.00000339, 0.00000613, -0.00000543, 0.00000310, -0.00001009, 0.00000405, 0.00000384, 0.00000794, -0.00000593, 0.00000458, -0.00002079, 0.00000871, 0.00001823, -0.00000555, -0.00001377, 0.00001000, -0.00000602, 0.00000008, 0.00000238, -0.00000202, -0.00000235, -0.00000650, -0.00000189, -0.00000224, 0.00001106, -0.00000279, -0.00000344, -0.00000582, 0.00000935, -0.00000236, -0.00001408, 0.00000306, 0.00001150, -0.00000201, -0.00000482, 0.00000116, 0.00000177, -0.00000673, -0.00000606, -0.00001559, 0.00000595, -0.00000018, 0.00001836, 0.00001917, -0.00001997, 0.00003521, -0.00000092, 0.00000468, -0.00001031, -0.00001213, 0.00000168, -0.00000611, -0.00000319, -0.00000065, -0.00000209, -0.00000065, -0.00000983, 0.00000369, -0.00001309, 0.00000536, -0.00002091, 0.00000285, -0.00004646, 0.00000968, -0.00021266, 0.00008854, 0.00057682, -0.00018510, -0.00008502, -0.00002322, -0.00000959, -0.00001683, -0.00000245, -0.00001020, 0.00000196, -0.00000913, 0.00000375, -0.00000538, 0.00000155, -0.00000867, 0.00001548, -0.00001645, 0.00001171, -0.00000513, 0.00000274, 0.00000670, -0.00002457, 0.00000618, -0.00000328, 0.00000370, -0.00002630, 0.00002146, -0.00002544, 0.00001527, 0.00000140, 0.00000228, -0.00000401, -0.00000308, 0.00000551, 0.00000303, -0.00000834, 0.00000031, 0.00000697, -0.00000598, 0.00000666, -0.00000645, 0.00001012, -0.00001029, 0.00000494, -0.00000388, 0.00000949, -0.00000650, 0.00000419, -0.00000425, 0.00000235, -0.00000482, 0.00001060, -0.00001055, 0.00001700, -0.00001044, -0.00000759, 0.00000411, 0.00000621, -0.00000398, -0.00000379, },
		{ 0.00000934, -0.00000816, 0.00000693, -0.00000438, 0.00000686, -0.00000551, 0.00000330, -0.00000949, 0.00000227, 0.00000391, 0.00000813, -0.00000517, 0.00000500, -0.00001718, -0.00000044, 0.00002223, -0.00000255, -0.00001384, 0.00000521, -0.00000015, -0.00000508, 0.00000620, -0.00000519, 0.00000137, -0.00000965, 0.00000087, -0.00000651, 0.00001303, -0.00000277, -0.00000175, -0.00000817, 0.00000930, -0.00000064, -0.00001285, -0.00000163, 0.00001385, -0.00000194, -0.00000365, -0.00000121, 0.00000410, -0.00000752, -0.00000440, -0.00001585, 0.00000075, 0.00000341, 0.00001076, 0.00002879, -0.00002149, 0.00003011, 0.00000810, -0.00000004, -0.00000444, -0.00001796, 0.00000425, -0.00000794, -0.00000134, -0.00000360, 0.00000063, -0.00000262, -0.00000694, -0.00000035, -0.00000765, -0.00000101, -0.00001257, -0.00000468, -0.00003382, -0.00000344, -0.00015711, -0.00003506, 0.00057533, -0.00005555, -0.00015086, -0.00000382, -0.00002642, -0.00000433, -0.00001336, -0.00000218, -0.00000575, -0.00000202, -0.00000306, 0.00000047, -0.00000281, -0.00000545, 0.00001058, -0.00000842, 0.00000191, 0.00000448, -0.00000732, 0.00001711, -0.00003009, 0.00000842, -0.00000790, 0.00001086, -0.00003027, 0.00001881, -0.00001845, 0.00000468, 0.00000983, -0.00000294, 0.00000107, -0.00000679, 0.00000684, 0.00000161, -0.00000447, -0.00000463, 0.00001056, -0.00000835, 0.00000854, -0.00000787, 0.00001094, -0.00000955, 0.00000360, -0.00000383, 0.00000897, -0.00000439, 0.00000264, -0.00000361, 0.00000223, -0.00000508, 0.00000968, -0.00000769, 0.00001126, -0.00000178, -0.00001341, 0.00000529, 0.00000352, 0.00000240, -0.00000935, },
		{ 0.00000959, -0.00000801, 0.00000649, -0.00000354, 0.00000590, -0.00000402, 0.00000203, -0.00000746, -0.00000060, 0.00000494, 0.00000714, -0.00000319, 0.00000351, -0.00001135, -0.00000962, 0.00002433, 0.00000085, -0.00001241, -0.00000054, 0.00000524, -0.00000892, 0.00000857, -0.00000699, 0.00000355, -0.00001105, 0.00000189, -0.00000852, 0.00001229, -0.00000057, -0.00000188, -0.00000831, 0.00000699, 0.00000256, -0.00001263, -0.00000447, 0.00001351, 0.00000061, -0.00000469, -0.00000107, 0.00000343, -0.00000510, -0.00000627, -0.00001197, -0.00000739, 0.00000839, 0.00000275, 0.00003592, -0.00001896, 0.00002224, 0.00001754, -0.00000374, 0.00000034, -0.00002196, 0.00000468, -0.00000809, -0.00000083, -0.00000510, 0.00000160, -0.00000302, -0.00000533, -0.00000307, -0.00000388, -0.00000521, -0.00000709, -0.00000923, -0.00002439, -0.00001392, -0.00011104, -0.00011689, 0.00051825, 0.00008772, -0.00019688, -0.00000088, -0.00003361, 0.00000069, -0.00001810, 0.00000085, -0.00000945, 0.00000143, -0.00000638, 0.00000302, -0.00000432, -0.00000415, 0.00000681, -0.00000171, -0.00000571, 0.00001126, -0.00001405, 0.00002302, -0.00003051, 0.00000636, -0.00000849, 0.00001250, -0.00002709, 0.00000946, -0.00000683, -0.00000774, 0.00001773, -0.00000696, 0.00000474, -0.00000893, 0.00000719, 0.00000051, -0.00000115, -0.00000826, 0.00001240, -0.00000892, 0.00000860, -0.00000744, 0.00000987, -0.00000704, 0.00000098, -0.00000258, 0.00000697, -0.00000112, 0.00000045, -0.00000220, 0.00000085, -0.00000351, 0.00000652, -0.00000271, 0.00000423, 0.00000640, -0.00001745, 0.00000556, 0.00000063, 0.00000799, -0.00001255, },
		{ 0.00000709, -0.00000584, 0.00000417, -0.00000100, 0.00000350, -0.00000128, -0.00000034, -0.00000447, -0.00000406, 0.00000645, 0.00000552, -0.00000054, 0.00000088, -0.00000468, -0.00001719, 0.00002364, 0.00000493, -0.00001021, -0.00000585, 0.00000872, -0.00001040, 0.00000872, -0.00000675, 0.00000355, -0.00001014, 0.00000066, -0.00000790, 0.00000875, 0.00000364, -0.00000365, -0.00000644, 0.00000280, 0.00000681, -0.00001322, -0.00000555, 0.00001076, 0.00000510, -0.00000732, 0.00000088, 0.00000065, -0.00000066, -0.00001011, -0.00000625, -0.00001555, 0.00001186, -0.00000297, 0.00003847, -0.00001163, 0.00001199, 0.00002672, -0.00000602, 0.00000387, -0.00002411, 0.00000319, -0.00000695, -0.00000137, -0.00000532, 0.00000104, -0.00000213, -0.00000485, -0.00000440, -0.00000203, -0.00000688, -0.00000483, -0.00001041, -0.00001890, -0.00001993, -0.00007714, -0.00016132, 0.00042544, 0.00021953, -0.00020841, -0.00001903, -0.00002951, -0.00000286, -0.00001601, -0.00000153, -0.00000872, 0.00000078, -0.00000587, 0.00000208, -0.00000296, -0.00000483, 0.00000454, 0.00000298, -0.00001033, 0.00001447, -0.00001671, 0.00002390, -0.00002575, 0.00000024, -0.00000525, 0.00000911, -0.00001833, -0.00000371, 0.00000579, -0.00001845, 0.00002237, -0.00000794, 0.00000578, -0.00000885, 0.00000618, 0.00000024, 0.00000100, -0.00001008, 0.00001221, -0.00000762, 0.00000691, -0.00000538, 0.00000726, -0.00000327, -0.00000237, -0.00000067, 0.00000408, 0.00000258, -0.00000176, -0.00000044, -0.00000135, -0.00000083, 0.00000230, 0.00000276, -0.00000217, 0.00001235, -0.00001866, 0.00000427, -0.00000146, 0.00001149, -0.00001250, },
		{ 0.00000236, -0.00000225, 0.00000056, 0.00000255, 0.00000039, 0.00000195, -0.00000304, -0.00000131, -0.00000733, 0.00000767, 0.00000399, 0.00000203, -0.00000190, 0.00000135, -0.00002181, 0.00001977, 0.00000968, -0.00000783, -0.00000968, 0.00000952, -0.00000922, 0.00000663, -0.00000450, 0.00000148, -0.00000718, -0.00000245, -0.00000516, 0.00000328, 0.00000881, -0.00000610, -0.00000358, -0.00000212, 0.00001098, -0.00001367, -0.00000576, 0.00000677, 0.00001012, -0.00001005, 0.00000305, -0.00000242, 0.00000374, -0.00001349, -0.00000157, -0.00002070, 0.00001135, -0.00000450, 0.00003541, 0.00000017, 0.00000043, 0.00003476, -0.00000657, 0.00000625, -0.00002469, 0.00000038, -0.00000523, -0.00000234, -0.00000487, -0.00000030, -0.00000078, -0.00000476, -0.00000496, -0.00000149, -0.00000676, -0.00000476, -0.00000947, -0.00001604, -0.00002257, -0.00005359, -0.00017791, 0.00031816, 0.00032014, -0.00017972, -0.00005524, -0.00001774, -0.00001214, -0.00000956, -0.00000718, -0.00000531, -0.00000249, -0.00000294, -0.00000094, -0.00000009, -0.00000638, 0.00000308, 0.00000600, -0.00001228, 0.00001455, -0.00001586, 0.00002056, -0.00001713, -0.00000838, 0.00000023, 0.00000272, -0.00000707, -0.00001680, 0.00001554, -0.00002439, 0.00002187, -0.00000499, 0.00000389, -0.00000659, 0.00000391, 0.00000089, 0.00000180, -0.00001006, 0.00001021, -0.00000480, 0.00000396, -0.00000227, 0.00000377, 0.00000100, -0.00000569, 0.00000124, 0.00000107, 0.00000590, -0.00000337, 0.00000117, -0.00000370, 0.00000196, -0.00000163, 0.00000711, -0.00000635, 0.00001496, -0.00001666, 0.00000130, -0.00000224, 0.00001224, -0.00000913, },
		{ -0.00000339, 0.00000164, -0.00000331, 0.00000606, -0.00000240, 0.00000472, -0.00000517, 0.00000118, -0.00000963, 0.00000791, 0.00000324, 0.00000385, -0.00000395, 0.00000559, -0.00002280, 0.00001299, 0.00001467, -0.00000551, -0.00001159, 0.00000768, -0.00000591, 0.00000299, -0.00000096, -0.00000186, -0.00000311, -0.00000639, -0.00000155, -0.00000262, 0.00001337, -0.00000780, -0.00000113, -0.00000625, 0.00001365, -0.00001283, -0.00000629, 0.00000303, 0.00001396, -0.00001117, 0.00000370, -0.00000393, 0.00000604, -0.00001425, -0.00000023, -0.00002090, 0.00000580, -0.00000148, 0.00002724, 0.00001485, -0.00001062, 0.00004047, -0.00000504, 0.00000751, -0.00002397, -0.00000321, -0.00000356, -0.00000318, -0.00000435, -0.00000169, 0.00000022, -0.00000430, -0.00000549, -0.00000137, -0.00000605, -0.00000530, -0.00000828, -0.00001375, -0.00002404, -0.00003676, -0.00017722, 0.00021398, 0.00037903, -0.00011514, -0.00009968, -0.00000579, -0.00002186, -0.00000304, -0.00001258, -0.00000210, -0.00000585, 0.00000006, -0.00000382, 0.00000230, -0.00000715, 0.00000123, 0.00000828, -0.00001255, 0.00001267, -0.00001279, 0.00001466, -0.00000681, -0.00001716, 0.00000565, -0.00000395, 0.00000325, -0.00002614, 0.00001948, -0.00002388, 0.00001590, 0.00000143, -0.00000017, -0.00000297, 0.00000102, 0.00000211, 0.00000150, -0.00000862, 0.00000700, -0.00000118, 0.00000049, 0.00000110, 0.00000024, 0.00000492, -0.00000821, 0.00000248, -0.00000138, 0.00000817, -0.00000391, 0.00000220, -0.00000556, 0.00000393, -0.00000415, 0.00000925, -0.00000752, 0.00001405, -0.00001190, -0.00000287, -0.00000176, 0.00001040, -0.00000329, },
		{ -0.00000856, 0.00000457, -0.00000634, 0.00000847, -0.00000395, 0.00000623, -0.00000603, 0.00000236, -0.00001041, 0.00000672, 0.00000362, 0.00000454, -0.00000475, 0.00000749, -0.00002026, 0.00000429, 0.00001906, -0.00000313, -0.00001169, 0.00000391, -0.00000158, -0.00000103, 0.00000277, -0.00000532, 0.00000078, -0.00000980, 0.00000149, -0.00000732, 0.00001579, -0.00000747, -0.00000029, -0.00000836, 0.00001375, -0.00000984, -0.00000803, 0.00000071, 0.00001536, -0.00000952, 0.00000177, -0.00000277, 0.00000516, -0.00001141, -0.00000303, -0.00001605, -0.00000389, 0.00000464, 0.00001607, 0.00002969, -0.00001883, 0.00004257, -0.00000101, 0.00000761, -0.00002204, -0.00000726, -0.00000224, -0.00000368, -0.00000403, -0.00000269, 0.00000040, -0.00000309, -0.00000641, -0.00000107, -0.00000565, -0.00000521, -0.00000823, -0.00001050, -0.00002607, -0.00002354, -0.00016764, 0.00012391, 0.00039574, -0.00002730, -0.00013875, -0.00000206, -0.00002660, -0.00000060, -0.00001447, -0.00000170, -0.00000706, 0.00000104, -0.00000466, 0.00000254, -0.00000588, -0.00000190, 0.00001061, -0.00001211, 0.00001004, -0.00000893, 0.00000801, 0.00000301, -0.00002388, 0.00000887, -0.00000848, 0.00000989, -0.00002943, 0.00001652, -0.00001725, 0.00000581, 0.00000958, -0.00000479, 0.00000065, -0.00000154, 0.00000330, 0.00000066, -0.00000646, 0.00000339, 0.00000245, -0.00000271, 0.00000398, -0.00000260, 0.00000779, -0.00000938, 0.00000265, -0.00000283, 0.00000902, -0.00000318, 0.00000245, -0.00000649, 0.00000447, -0.00000468, 0.00000883, -0.00000579, 0.00001037, -0.00000556, -0.00000727, -0.00000058, 0.00000677, 0.00000358, },
		{ -0.00001167, 0.00000551, -0.00000772, 0.00000906, -0.00000369, 0.00000609, -0.00000532, 0.00000201, -0.00000954, 0.00000411, 0.00000506, 0.00000419, -0.00000428, 0.00000717, -0.00001507, -0.00000486, 0.00002183, -0.00000034, -0.00001051, -0.00000065, 0.00000242, -0.00000424, 0.00000560, -0.00000776, 0.00000332, -0.00001154, 0.00000275, -0.00000962, 0.00001515, -0.00000453, -0.00000161, -0.00000795, 0.00001103, -0.00000467, -0.00001112, 0.00000017, 0.00001399, -0.00000500, -0.00000271, 0.00000093, 0.00000139, -0.00000550, -0.00000899, -0.00000792, -0.00001517, 0.00001108, 0.00000491, 0.00004157, -0.00002195, 0.00004004, 0.00000573, 0.00000651, -0.00001888, -0.00001168, -0.00000127, -0.00000400, -0.00000380, -0.00000330, -0.00000017, -0.00000131, -0.00000757, -0.00000051, -0.00000586, -0.00000408, -0.00000971, -0.00000600, -0.00002901, -0.00001254, -0.00015402, 0.00005197, 0.00037794, 0.00006685, -0.00015935, -0.00001255, -0.00002327, -0.00000439, -0.00001133, -0.00000520, -0.00000513, -0.00000094, -0.00000267, 0.00000010, -0.00000231, -0.00000635, 0.00001313, -0.00001145, 0.00000742, -0.00000527, 0.00000196, 0.00001067, -0.00002706, 0.00000865, -0.00000955, 0.00001158, -0.00002629, 0.00000765, -0.00000668, -0.00000567, 0.00001698, -0.00000803, 0.00000286, -0.00000287, 0.00000390, -0.00000014, -0.00000431, 0.00000013, 0.00000536, -0.00000499, 0.00000579, -0.00000427, 0.00000922, -0.00000900, 0.00000166, -0.00000319, 0.00000846, -0.00000137, 0.00000201, -0.00000638, 0.00000344, -0.00000330, 0.00000632, -0.00000210, 0.00000531, 0.00000082, -0.00001078, 0.00000051, 0.00000251, 0.00000983, },
		{ -0.00001180, 0.00000408, -0.00000722, 0.00000766, -0.00000164, 0.00000442, -0.00000325, 0.00000040, -0.00000738, 0.00000054, 0.00000707, 0.00000323, -0.00000295, 0.00000527, -0.00000856, -0.00001285, 0.00002209, 0.00000313, -0.00000869, -0.00000484, 0.00000499, -0.00000581, 0.00000685, -0.00000848, 0.00000380, -0.00001101, 0.00000174, -0.00000915, 0.00001144, 0.00000071, -0.00000469, -0.00000548, 0.00000615, 0.00000183, -0.00001483, 0.00000086, 0.00001054, 0.00000141, -0.00000853, 0.00000583, -0.00000374, 0.00000160, -0.00001582, 0.00000053, -0.00002467, 0.00001476, -0.00000327, 0.00004786, -0.00001858, 0.00003267, 0.00001490, 0.00000439, -0.00001454, -0.00001647, -0.00000046, -0.00000445, -0.00000338, -0.00000381, -0.00000108, 0.00000049, -0.00000847, -0.00000008, -0.00000636, -0.00000231, -0.00001206, -0.00000113, -0.00003185, -0.00000415, -0.00013792, -0.00000318, 0.00033771, 0.00015078, -0.00015287, -0.00003833, -0.00001263, -0.00001358, -0.00000412, -0.00001168, -0.00000084, -0.00000521, 0.00000147, -0.00000423, 0.00000263, -0.00001119, 0.00001520, -0.00001035, 0.00000493, -0.00000216, -0.00000283, 0.00001543, -0.00002630, 0.00000493, -0.00000718, 0.00000875, -0.00001819, -0.00000441, 0.00000452, -0.00001538, 0.00002118, -0.00000831, 0.00000275, -0.00000254, 0.00000364, -0.00000051, -0.00000268, -0.00000226, 0.00000716, -0.00000605, 0.00000632, -0.00000464, 0.00000921, -0.00000722, -0.00000026, -0.00000268, 0.00000680, 0.00000109, 0.00000119, -0.00000540, 0.00000115, -0.00000062, 0.00000272, 0.00000217, 0.00000047, 0.00000581, -0.00001245, 0.00000079, -0.00000128, 0.00001409, },
		{ -0.00000879, 0.00000063, -0.00000521, 0.00000472, 0.00000161, 0.00000188, -0.00000047, -0.00000182, -0.00000464, -0.00000323, 0.00000892, 0.00000232, -0.00000136, 0.00000270, -0.00000217, -0.00001833, 0.00001940, 0.00000728, -0.00000672, -0.00000785, 0.00000553, -0.00000550, 0.00000637, -0.00000737, 0.00000223, -0.00000841, -0.00000121, -0.00000644, 0.00000559, 0.00000704, -0.00000833, -0.00000216, 0.00000053, 0.00000815, -0.00001783, 0.00000160, 0.00000639, 0.00000798, -0.00001382, 0.00000996, -0.00000806, 0.00000746, -0.00002078, 0.00000620, -0.00002943, 0.00001343, -0.00000659, 0.00004728, -0.00000873, 0.00002139, 0.00002545, 0.00000187, -0.00000933, -0.00002141, 0.00000025, -0.00000528, -0.00000253, -0.00000451, -0.00000184, 0.00000167, -0.00000849, -0.00000031, -0.00000654, -0.00000072, -0.00001409, 0.00000261, -0.00003291, 0.00000032, -0.00011907, -0.00004551, 0.00028746, 0.00021255, -0.00011772, -0.00007467, 0.00000077, -0.00002462, 0.00000413, -0.00001853, 0.00000367, -0.00000993, 0.00000597, -0.00000867, 0.00000718, -0.00001489, 0.00001572, -0.00000818, 0.00000219, 0.00000059, -0.00000636, 0.00001742, -0.00002223, -0.00000126, -0.00000259, 0.00000312, -0.00000789, -0.00001618, 0.00001294, -0.00002067, 0.00002066, -0.00000502, 0.00000023, -0.00000073, 0.00000266, -0.00000038, -0.00000176, -0.00000365, 0.00000778, -0.00000590, 0.00000568, -0.00000391, 0.00000804, -0.00000447, -0.00000261, -0.00000173, 0.00000456, 0.00000364, 0.00000039, -0.00000389, -0.00000181, 0.00000240, -0.00000075, 0.00000566, -0.00000282, 0.00000842, -0.00001187, -0.00000003, -0.00000386, 0.00001560, },
		{ -0.00000335, -0.00000383, -0.00000261, 0.00000109, 0.00000515, -0.00000065, 0.00000220, -0.00000385, -0.00000212, -0.00000641, 0.00000989, 0.00000208, -0.00000013, 0.00000031, 0.00000292, -0.00002055, 0.00001397, 0.00001168, -0.00000474, -0.00000933, 0.00000412, -0.00000370, 0.00000464, -0.00000488, -0.00000079, -0.00000454, -0.00000514, -0.00000268, -0.00000086, 0.00001278, -0.00001101, 0.00000051, -0.00000418, 0.00001268, -0.00001878, 0.00000112, 0.00000306, 0.00001295, -0.00001672, 0.00001148, -0.00000966, 0.00001007, -0.00002180, 0.00000704, -0.00002796, 0.00000651, -0.00000492, 0.00004032, 0.00000597, 0.00000828, 0.00003567, 0.00000003, -0.00000393, -0.00002599, 0.00000063, -0.00000639, -0.00000133, -0.00000543, -0.00000220, 0.00000187, -0.00000734, -0.00000151, -0.00000595, -0.00000008, -0.00001468, 0.00000389, -0.00003078, -0.00000015, -0.00009700, -0.00007900, 0.00023663, 0.00024711, -0.00005968, -0.00011221, 0.00001009, -0.00003272, 0.00000955, -0.00002264, 0.00000587, -0.00001284, 0.00000870, -0.00001121, 0.00000953, -0.00001594, 0.00001368, -0.00000433, -0.00000121, 0.00000333, -0.00000896, 0.00001725, -0.00001608, -0.00000830, 0.00000243, -0.00000301, 0.00000162, -0.00002449, 0.00001610, -0.00002025, 0.00001530, 0.00000124, -0.00000383, 0.00000171, 0.00000154, -0.00000000, -0.00000142, -0.00000420, 0.00000746, -0.00000484, 0.00000424, -0.00000249, 0.00000618, -0.00000131, -0.00000487, -0.00000081, 0.00000227, 0.00000575, -0.00000002, -0.00000223, -0.00000478, 0.00000483, -0.00000304, 0.00000740, -0.00000386, 0.00000845, -0.00000927, -0.00000182, -0.00000504, 0.00001434, },
		{ 0.00000327, -0.00000794, -0.00000047, -0.00000225, 0.00000798, -0.00000227, 0.00000398, -0.00000500, -0.00000047, -0.00000837, 0.00000950, 0.00000283, 0.00000041, -0.00000134, 0.00000607, -0.00001949, 0.00000666, 0.00001557, -0.00000262, -0.00000942, 0.00000135, -0.00000125, 0.00000242, -0.00000181, -0.00000429, -0.00000059, -0.00000876, 0.00000068, -0.00000626, 0.00001633, -0.00001139, 0.00000128, -0.00000671, 0.00001426, -0.00001680, -0.00000135, 0.00000156, 0.00001513, -0.00001614, 0.00000942, -0.00000765, 0.00000861, -0.00001821, 0.00000272, -0.00002080, -0.00000458, 0.00000011, 0.00002910, 0.00002262, -0.00000379, 0.00004344, 0.00000013, 0.00000076, -0.00002945, 0.00000016, -0.00000738, -0.00000020, -0.00000629, -0.00000224, 0.00000114, -0.00000515, -0.00000352, -0.00000455, -0.00000068, -0.00001331, 0.00000219, -0.00002511, -0.00000544, -0.00007233, -0.00010586, 0.00019010, 0.00025635, 0.00001002, -0.00013975, 0.00000875, -0.00003378, 0.00000909, -0.00002167, 0.00000396, -0.00001239, 0.00000819, -0.00001051, 0.00000853, -0.00001357, 0.00000873, 0.00000127, -0.00000536, 0.00000621, -0.00001091, 0.00001565, -0.00000921, -0.00001455, 0.00000614, -0.00000759, 0.00000796, -0.00002743, 0.00001324, -0.00001459, 0.00000648, 0.00000878, -0.00000792, 0.00000361, 0.00000099, 0.00000027, -0.00000137, -0.00000427, 0.00000660, -0.00000329, 0.00000243, -0.00000083, 0.00000411, 0.00000173, -0.00000656, -0.00000028, 0.00000033, 0.00000707, 0.00000015, -0.00000072, -0.00000718, 0.00000597, -0.00000357, 0.00000708, -0.00000272, 0.00000643, -0.00000546, -0.00000398, -0.00000508, 0.00001095, },
		{ 0.00000955, -0.00001040, 0.00000033, -0.00000453, 0.00000937, -0.00000243, 0.00000447, -0.00000493, 0.00000001, -0.00000889, 0.00000768, 0.00000454, 0.00000028, -0.00000210, 0.00000721, -0.00001583, -0.00000124, 0.00001804, -0.00000008, -0.00000854, -0.00000188, 0.00000091, 0.00000055, 0.00000101, -0.00000728, 0.00000233, -0.00001094, 0.00000248, -0.00000936, 0.00001672, -0.00000884, -0.00000035, -0.00000656, 0.00001264, -0.00001194, -0.00000574, 0.00000202, 0.00001434, -0.00001209, 0.00000405, -0.00000248, 0.00000372, -0.00001098, -0.00000526, -0.00001026, -0.00001691, 0.00000571, 0.00001668, 0.00003774, -0.00001190, 0.00004688, 0.00000325, 0.00000393, -0.00003099, -0.00000177, -0.00000767, 0.00000026, -0.00000660, -0.00000234, -0.00000007, -0.00000252, -0.00000578, -0.00000276, -0.00000226, -0.00001026, -0.00000199, -0.00001677, -0.00001401, -0.00004712, -0.00012605, 0.00014842, 0.00024724, 0.00007807, -0.00014766, -0.00000697, -0.00002632, 0.00000186, -0.00001521, -0.00000223, -0.00000839, 0.00000422, -0.00000643, 0.00000426, -0.00000814, 0.00000151, 0.00000792, -0.00000971, 0.00000890, -0.00001217, 0.00001314, -0.00000265, -0.00001884, 0.00000746, -0.00000941, 0.00001004, -0.00002473, 0.00000537, -0.00000572, -0.00000335, 0.00001537, -0.00001033, 0.00000387, 0.00000154, 0.00000016, -0.00000131, -0.00000420, 0.00000556, -0.00000161, 0.00000064, 0.00000069, 0.00000221, 0.00000428, -0.00000744, -0.00000030, -0.00000107, 0.00000749, 0.00000086, 0.00000053, -0.00000866, 0.00000551, -0.00000233, 0.00000506, -0.00000013, 0.00000338, -0.00000151, -0.00000579, -0.00000455, 0.00000645, },
		{ 0.00001421, -0.00001037, -0.00000059, -0.00000543, 0.00000907, -0.00000107, 0.00000370, -0.00000374, -0.00000057, -0.00000815, 0.00000476, 0.00000678, -0.00000018, -0.00000220, 0.00000678, -0.00001070, -0.00000831, 0.00001833, 0.00000306, -0.00000716, -0.00000476, 0.00000202, -0.00000045, 0.00000303, -0.00000904, 0.00000349, -0.00001107, 0.00000214, -0.00000973, 0.00001385, -0.00000367, -0.00000396, -0.00000424, 0.00000852, -0.00000518, -0.00001108, 0.00000368, 0.00001138, -0.00000569, -0.00000322, 0.00000422, -0.00000274, -0.00000234, -0.00001420, 0.00000030, -0.00002702, 0.00000893, 0.00000603, 0.00004831, -0.00001392, 0.00004495, 0.00000980, 0.00000518, -0.00003013, -0.00000555, -0.00000680, -0.00000044, -0.00000593, -0.00000293, -0.00000119, -0.00000020, -0.00000752, -0.00000125, -0.00000412, -0.00000642, -0.00000733, -0.00000759, -0.00002330, -0.00002440, -0.00013803, 0.00010940, 0.00022874, 0.00013277, -0.00013087, -0.00003617, -0.00001234, -0.00001030, -0.00000503, -0.00001106, -0.00000222, -0.00000202, -0.00000020, -0.00000198, -0.00000110, -0.00000643, 0.00001423, -0.00001320, 0.00001068, -0.00001237, 0.00000989, 0.00000301, -0.00002071, 0.00000617, -0.00000836, 0.00000816, -0.00001774, -0.00000508, 0.00000345, -0.00001147, 0.00001896, -0.00000983, 0.00000194, 0.00000329, -0.00000032, -0.00000114, -0.00000419, 0.00000452, -0.00000001, -0.00000092, 0.00000189, 0.00000065, 0.00000617, -0.00000746, -0.00000084, -0.00000194, 0.00000712, 0.00000196, 0.00000151, -0.00000909, 0.00000357, 0.00000017, 0.00000218, 0.00000284, 0.00000049, 0.00000155, -0.00000661, -0.00000400, 0.00000188, },
		{ 0.00001645, -0.00000768, -0.00000297, -0.00000518, 0.00000733, 0.00000141, 0.00000215, -0.00000188, -0.00000176, -0.00000665, 0.00000135, 0.00000893, -0.00000046, -0.00000203, 0.00000545, -0.00000529, -0.00001338, 0.00001612, 0.00000669, -0.00000556, -0.00000672, 0.00000179, -0.00000048, 0.00000406, -0.00000930, 0.00000272, -0.00000918, -0.00000013, -0.00000775, 0.00000857, 0.00000293, -0.00000832, -0.00000095, 0.00000335, 0.00000184, -0.00001588, 0.00000523, 0.00000770, 0.00000128, -0.00001031, 0.00001025, -0.00000839, 0.00000507, -0.00002116, 0.00000768, -0.00003208, 0.00000773, -0.00000088, 0.00005252, -0.00000914, 0.00003788, 0.00001920, 0.00000484, -0.00002694, -0.00001111, -0.00000475, -0.00000245, -0.00000416, -0.00000419, -0.00000178, 0.00000120, -0.00000814, -0.00000062, -0.00000553, -0.00000286, -0.00001224, 0.00000030, -0.00003059, -0.00000720, -0.00014005, 0.00007027, 0.00020877, 0.00016703, -0.00009048, -0.00007314, 0.00000310, -0.00002344, 0.00000544, -0.00001965, 0.00000374, -0.00000839, 0.00000613, -0.00000814, 0.00000548, -0.00001314, 0.00001856, -0.00001464, 0.00001070, -0.00001106, 0.00000587, 0.00000763, -0.00002035, 0.00000282, -0.00000524, 0.00000370, -0.00000880, -0.00001512, 0.00001006, -0.00001573, 0.00001839, -0.00000611, -0.00000194, 0.00000577, -0.00000082, -0.00000096, -0.00000424, 0.00000351, 0.00000146, -0.00000217, 0.00000272, -0.00000053, 0.00000740, -0.00000672, -0.00000173, -0.00000243, 0.00000617, 0.00000317, 0.00000232, -0.00000855, 0.00000056, 0.00000314, -0.00000055, 0.00000514, -0.00000129, 0.00000307, -0.00000617, -0.00000376, -0.00000196, },
		{ 0.00001613, -0.00000289, -0.00000602, -0.00000439, 0.00000476, 0.00000429, 0.00000051, 0.00000003, -0.00000297, -0.00000500, -0.00000187, 0.00001036, -0.00000012, -0.00000200, 0.00000387, -0.00000063, -0.00001581, 0.00001166, 0.00001036, -0.00000380, -0.00000761, 0.00000038, 0.00000017, 0.00000430, -0.00000825, 0.00000039, -0.00000591, -0.00000354, -0.00000448, 0.00000230, 0.00000933, -0.00001190, 0.00000178, -0.00000121, 0.00000742, -0.00001863, 0.00000535, 0.00000477, 0.00000702, -0.00001529, 0.00001363, -0.00001121, 0.00000916, -0.00002397, 0.00000987, -0.00003083, 0.00000172, -0.00000366, 0.00005023, 0.00000153, 0.00002725, 0.00002995, 0.00000391, -0.00002207, -0.00001782, -0.00000200, -0.00000538, -0.00000165, -0.00000594, -0.00000178, 0.00000147, -0.00000741, -0.00000111, -0.00000601, -0.00000044, -0.00001539, 0.00000525, -0.00003384, 0.00000251, -0.00013147, 0.00002954, 0.00019198, 0.00017964, -0.00003349, -0.00010869, 0.00001341, -0.00003287, 0.00001245, -0.00002491, 0.00000699, -0.00001268, 0.00001043, -0.00001215, 0.00000967, -0.00001689, 0.00001960, -0.00001315, 0.00000842, -0.00000796, 0.00000112, 0.00001128, -0.00001831, -0.00000158, -0.00000138, -0.00000149, -0.00000044, -0.00002211, 0.00001216, -0.00001521, 0.00001375, 0.00000008, -0.00000676, 0.00000803, -0.00000077, -0.00000101, -0.00000419, 0.00000241, 0.00000288, -0.00000317, 0.00000324, -0.00000138, 0.00000804, -0.00000540, -0.00000276, -0.00000268, 0.00000489, 0.00000426, 0.00000301, -0.00000721, -0.00000295, 0.00000572, -0.00000227, 0.00000606, -0.00000150, 0.00000295, -0.00000464, -0.00000388, -0.00000467, },
		{ 0.00001367, 0.00000297, -0.00000867, -0.00000383, 0.00000209, 0.00000679, -0.00000056, 0.00000141, -0.00000371, -0.00000371, -0.00000437, 0.00001067, 0.00000108, -0.00000228, 0.00000246, 0.00000271, -0.00001559, 0.00000574, 0.00001336, -0.00000170, -0.00000758, -0.00000171, 0.00000093, 0.00000417, -0.00000632, -0.00000279, -0.00000226, -0.00000695, -0.00000122, -0.00000336, 0.00001392, -0.00001332, 0.00000269, -0.00000385, 0.00001032, -0.00001838, 0.00000326, 0.00000355, 0.00001036, -0.00001694, 0.00001330, -0.00001026, 0.00000906, -0.00002196, 0.00000662, -0.00002397, -0.00000770, -0.00000345, 0.00004282, 0.00001580, 0.00001561, 0.00003990, 0.00000381, -0.00001661, -0.00002460, 0.00000053, -0.00000843, 0.00000087, -0.00000762, -0.00000148, 0.00000080, -0.00000562, -0.00000250, -0.00000556, 0.00000051, -0.00001618, 0.00000651, -0.00003232, 0.00000446, -0.00011351, -0.00001203, 0.00017909, 0.00017477, 0.00002927, -0.00013280, 0.00001283, -0.00003500, 0.00001326, -0.00002483, 0.00000596, -0.00001347, 0.00001130, -0.00001271, 0.00001031, -0.00001682, 0.00001689, -0.00000862, 0.00000393, -0.00000334, -0.00000404, 0.00001394, -0.00001522, -0.00000599, 0.00000197, -0.00000573, 0.00000536, -0.00002452, 0.00000925, -0.00001050, 0.00000639, 0.00000711, -0.00001101, 0.00000898, 0.00000037, -0.00000151, -0.00000390, 0.00000108, 0.00000430, -0.00000393, 0.00000349, -0.00000195, 0.00000815, -0.00000368, -0.00000377, -0.00000282, 0.00000345, 0.00000510, 0.00000361, -0.00000530, -0.00000637, 0.00000721, -0.00000247, 0.00000541, -0.00000027, 0.00000164, -0.00000257, -0.00000409, -0.00000626, },
		{ 0.00000985, 0.00000865, -0.00000993, -0.00000409, -0.00000011, 0.00000834, -0.00000061, 0.00000197, -0.00000372, -0.00000304, -0.00000591, 0.00000975, 0.00000306, -0.00000281, 0.00000139, 0.00000462, -0.00001326, -0.00000053, 0.00001493, 0.00000091, -0.00000694, -0.00000385, 0.00000125, 0.00000404, -0.00000401, -0.00000599, 0.00000080, -0.00000933, 0.00000091, -0.00000719, 0.00001565, -0.00001188, 0.00000125, -0.00000404, 0.00001016, -0.00001509, -0.00000095, 0.00000413, 0.00001104, -0.00001515, 0.00000940, -0.00000591, 0.00000536, -0.00001607, -0.00000057, -0.00001384, -0.00001791, -0.00000231, 0.00003257, 0.00003065, 0.00000584, 0.00004693, 0.00000593, -0.00001166, -0.00003030, 0.00000182, -0.00001066, 0.00000252, -0.00000855, -0.00000140, -0.00000026, -0.00000338, -0.00000422, -0.00000456, 0.00000017, -0.00001476, 0.00000443, -0.00002666, 0.00000010, -0.00008926, -0.00005155, 0.00016757, 0.00016000, 0.00008612, -0.00013756, -0.00000146, -0.00002881, 0.00000731, -0.00001920, 0.00000065, -0.00001069, 0.00000858, -0.00000974, 0.00000745, -0.00001321, 0.00001102, -0.00000184, -0.00000194, 0.00000199, -0.00000894, 0.00001543, -0.00001151, -0.00000967, 0.00000394, -0.00000799, 0.00000767, -0.00002221, 0.00000236, -0.00000344, -0.00000156, 0.00001302, -0.00001322, 0.00000781, 0.00000283, -0.00000245, -0.00000332, -0.00000050, 0.00000567, -0.00000441, 0.00000345, -0.00000221, 0.00000775, -0.00000164, -0.00000469, -0.00000288, 0.00000194, 0.00000565, 0.00000409, -0.00000301, -0.00000921, 0.00000723, -0.00000115, 0.00000355, 0.00000179, -0.00000012, -0.00000065, -0.00000405, -0.00000697, },
		{ 0.00000553, 0.00001305, -0.00000917, -0.00000538, -0.00000161, 0.00000870, 0.00000039, 0.00000175, -0.00000306, -0.00000293, -0.00000653, 0.00000782, 0.00000544, -0.00000330, 0.00000053, 0.00000535, -0.00000969, -0.00000600, 0.00001452, 0.00000406, -0.00000594, -0.00000555, 0.00000081, 0.00000410, -0.00000170, -0.00000857, 0.00000251, -0.00001005, 0.00000130, -0.00000863, 0.00001430, -0.00000776, -0.00000221, -0.00000215, 0.00000752, -0.00000964, -0.00000634, 0.00000578, 0.00000976, -0.00001085, 0.00000320, 0.00000031, -0.00000019, -0.00000838, -0.00000910, -0.00000359, -0.00002595, -0.00000241, 0.00002182, 0.00004321, 0.00000030, 0.00004952, 0.00001107, -0.00000792, -0.00003402, 0.00000104, -0.00001132, 0.00000262, -0.00000821, -0.00000199, -0.00000115, -0.00000141, -0.00000558, -0.00000356, -0.00000090, -0.00001185, 0.00000017, -0.00001857, -0.00000789, -0.00006290, -0.00008502, 0.00015335, 0.00014358, 0.00012765, -0.00011948, -0.00002803, -0.00001640, -0.00000353, -0.00000983, -0.00000733, -0.00000563, 0.00000339, -0.00000433, 0.00000227, -0.00000739, 0.00000346, 0.00000569, -0.00000780, 0.00000685, -0.00001268, 0.00001542, -0.00000740, -0.00001226, 0.00000428, -0.00000808, 0.00000670, -0.00001637, -0.00000640, 0.00000352, -0.00000783, 0.00001614, -0.00001244, 0.00000428, 0.00000634, -0.00000353, -0.00000262, -0.00000218, 0.00000680, -0.00000445, 0.00000304, -0.00000208, 0.00000681, 0.00000065, -0.00000549, -0.00000284, 0.00000040, 0.00000594, 0.00000439, -0.00000054, -0.00001111, 0.00000572, 0.00000130, 0.00000117, 0.00000385, -0.00000150, 0.00000051, -0.00000349, -0.00000713, },
		{ 0.00000144, 0.00001548, -0.00000632, -0.00000742, -0.00000257, 0.00000803, 0.00000217, 0.00000107, -0.00000204, -0.00000314, -0.00000652, 0.00000531, 0.00000773, -0.00000339, -0.00000033, 0.00000530, -0.00000582, -0.00000979, 0.00001203, 0.00000747, -0.00000466, -0.00000657, -0.00000041, 0.00000428, 0.00000042, -0.00001013, 0.00000257, -0.00000899, -0.00000003, -0.00000787, 0.00001046, -0.00000193, -0.00000664, 0.00000074, 0.00000367, -0.00000351, -0.00001148, 0.00000727, 0.00000773, -0.00000554, -0.00000347, 0.00000639, -0.00000542, -0.00000131, -0.00001625, 0.00000392, -0.00002960, -0.00000508, 0.00001227, 0.00005143, 0.00000031, 0.00004725, 0.00001916, -0.00000547, -0.00003542, -0.00000208, -0.00001017, 0.00000096, -0.00000648, -0.00000340, -0.00000151, -0.00000021, -0.00000608, -0.00000305, -0.00000209, -0.00000842, -0.00000480, -0.00001011, -0.00001649, -0.00003866, -0.00010869, 0.00013277, 0.00013189, 0.00014897, -0.00008059, -0.00006132, -0.00000251, -0.00001560, 0.00000016, -0.00001535, -0.00000043, -0.00000238, 0.00000165, -0.00000338, -0.00000123, -0.00000388, 0.00001218, -0.00001217, 0.00000999, -0.00001442, 0.00001361, -0.00000297, -0.00001380, 0.00000324, -0.00000648, 0.00000349, -0.00000893, -0.00001456, 0.00000808, -0.00001080, 0.00001564, -0.00000862, -0.00000106, 0.00001014, -0.00000418, -0.00000212, -0.00000368, 0.00000741, -0.00000385, 0.00000213, -0.00000148, 0.00000532, 0.00000313, -0.00000613, -0.00000269, -0.00000112, 0.00000598, 0.00000449, 0.00000188, -0.00001187, 0.00000291, 0.00000421, -0.00000091, 0.00000518, -0.00000193, 0.00000062, -0.00000237, -0.00000696, },
		{ -0.00000197, 0.00001576, -0.00000187, -0.00000957, -0.00000338, 0.00000671, 0.00000420, 0.00000043, -0.00000106, -0.00000332, -0.00000624, 0.00000268, 0.00000947, -0.00000279, -0.00000133, 0.00000488, -0.00000236, -0.00001147, 0.00000785, 0.00001060, -0.00000299, -0.00000692, -0.00000212, 0.00000432, 0.00000235, -0.00001056, 0.00000109, -0.00000657, -0.00000251, -0.00000570, 0.00000531, 0.00000417, -0.00001064, 0.00000328, 0.00000009, 0.00000170, -0.00001493, 0.00000742, 0.00000621, -0.00000078, -0.00000884, 0.00001050, -0.00000848, 0.00000322, -0.00002006, 0.00000694, -0.00002806, -0.00001037, 0.00000452, 0.00005447, 0.00000582, 0.00004093, 0.00002917, -0.00000363, -0.00003479, -0.00000722, -0.00000755, -0.00000212, -0.00000373, -0.00000538, -0.00000135, 0.00000006, -0.00000559, -0.00000318, -0.00000299, -0.00000529, -0.00000915, -0.00000311, -0.00002315, -0.00001978, -0.00012030, 0.00010409, 0.00012790, 0.00015048, -0.00002785, -0.00009305, 0.00000709, -0.00002480, 0.00000742, -0.00002073, 0.00000275, -0.00000682, 0.00000635, -0.00000766, 0.00000348, -0.00000932, 0.00001619, -0.00001389, 0.00001060, -0.00001368, 0.00000998, 0.00000166, -0.00001445, 0.00000136, -0.00000406, -0.00000051, -0.00000195, -0.00002002, 0.00000884, -0.00000993, 0.00001180, -0.00000265, -0.00000701, 0.00001310, -0.00000373, -0.00000215, -0.00000469, 0.00000722, -0.00000248, 0.00000070, -0.00000043, 0.00000338, 0.00000563, -0.00000654, -0.00000252, -0.00000254, 0.00000576, 0.00000446, 0.00000402, -0.00001141, -0.00000078, 0.00000684, -0.00000202, 0.00000533, -0.00000124, -0.00000022, -0.00000092, -0.00000652, },
		{ -0.00000454, 0.00001416, 0.00000331, -0.00001100, -0.00000452, 0.00000520, 0.00000598, 0.00000025, -0.00000043, -0.00000324, -0.00000594, 0.00000030, 0.00001035, -0.00000141, -0.00000250, 0.00000432, 0.00000028, -0.00001117, 0.00000276, 0.00001274, -0.00000074, -0.00000679, -0.00000389, 0.00000391, 0.00000410, -0.00000993, -0.00000150, -0.00000352, -0.00000526, -0.00000316, 0.00000020, 0.00000909, -0.00001295, 0.00000431, -0.00000203, 0.00000482, -0.00001579, 0.00000556, 0.00000593, 0.00000246, -0.00001180, 0.00001161, -0.00000849, 0.00000442, -0.00001991, 0.00000534, -0.00002218, -0.00001704, -0.00000182, 0.00005260, 0.00001556, 0.00003235, 0.00003938, -0.00000128, -0.00003285, -0.00001356, -0.00000430, -0.00000580, -0.00000074, -0.00000736, -0.00000099, -0.00000036, -0.00000436, -0.00000381, -0.00000350, -0.00000293, -0.00001201, 0.00000132, -0.00002641, -0.00000785, -0.00011974, 0.00006818, 0.00013080, 0.00013714, 0.00002880, -0.00011452, 0.00000751, -0.00002807, 0.00000964, -0.00002173, 0.00000260, -0.00000870, 0.00000850, -0.00000940, 0.00000562, -0.00001186, 0.00001700, -0.00001255, 0.00000852, -0.00001061, 0.00000493, 0.00000617, -0.00001437, -0.00000080, -0.00000168, -0.00000403, 0.00000299, -0.00002163, 0.00000559, -0.00000592, 0.00000592, 0.00000392, -0.00001205, 0.00001414, -0.00000171, -0.00000288, -0.00000504, 0.00000612, -0.00000038, -0.00000113, 0.00000094, 0.00000119, 0.00000789, -0.00000653, -0.00000246, -0.00000368, 0.00000519, 0.00000442, 0.00000565, -0.00000978, -0.00000482, 0.00000857, -0.00000181, 0.00000430, 0.00000033, -0.00000157, 0.00000044, -0.00000576, },
		{ -0.00000627, 0.00001124, 0.00000822, -0.00001101, -0.00000627, 0.00000378, 0.00000714, 0.00000073, -0.00000026, -0.00000282, -0.00000573, -0.00000167, 0.00001030, 0.00000060, -0.00000366, 0.00000365, 0.00000205, -0.00000937, -0.00000223, 0.00001328, 0.00000216, -0.00000636, -0.00000535, 0.00000286, 0.00000565, -0.00000841, -0.00000463, -0.00000066, -0.00000742, -0.00000121, -0.00000375, 0.00001184, -0.00001284, 0.00000330, -0.00000216, 0.00000544, -0.00001394, 0.00000180, 0.00000692, 0.00000393, -0.00001215, 0.00000977, -0.00000571, 0.00000273, -0.00001653, 0.00000043, -0.00001404, -0.00002311, -0.00000765, 0.00004682, 0.00002752, 0.00002379, 0.00004789, 0.00000277, -0.00003044, -0.00002009, -0.00000145, -0.00000909, 0.00000160, -0.00000866, -0.00000090, -0.00000103, -0.00000288, -0.00000451, -0.00000377, -0.00000146, -0.00001315, 0.00000293, -0.00002599, -0.00000272, -0.00010901, 0.00002823, 0.00013690, 0.00011650, 0.00007909, -0.00011910, -0.00000349, -0.00002459, 0.00000632, -0.00001821, -0.00000088, -0.00000792, 0.00000789, -0.00000837, 0.00000506, -0.00001151, 0.00001480, -0.00000858, 0.00000436, -0.00000597, -0.00000071, 0.00001005, -0.00001356, -0.00000287, 0.00000009, -0.00000624, 0.00000517, -0.00001943, -0.00000064, -0.00000041, -0.00000016, 0.00000935, -0.00001483, 0.00001258, 0.00000193, -0.00000419, -0.00000478, 0.00000421, 0.00000220, -0.00000303, 0.00000230, -0.00000088, 0.00000958, -0.00000587, -0.00000269, -0.00000436, 0.00000417, 0.00000449, 0.00000662, -0.00000715, -0.00000867, 0.00000895, -0.00000027, 0.00000245, 0.00000223, -0.00000284, 0.00000128, -0.00000461, },
		{ -0.00000726, 0.00000767, 0.00001200, -0.00000924, -0.00000859, 0.00000250, 0.00000761, 0.00000183, -0.00000042, -0.00000220, -0.00000554, -0.00000323, 0.00000949, 0.00000291, -0.00000452, 0.00000280, 0.00000311, -0.00000680, -0.00000621, 0.00001191, 0.00000556, -0.00000566, -0.00000632, 0.00000124, 0.00000685, -0.00000616, -0.00000773, 0.00000132, -0.00000841, -0.00000044, -0.00000594, 0.00001205, -0.00001033, 0.00000043, -0.00000053, 0.00000397, -0.00001011, -0.00000304, 0.00000850, 0.00000418, -0.00001054, 0.00000591, -0.00000134, -0.00000047, -0.00001157, -0.00000566, -0.00000617, -0.00002666, -0.00001375, 0.00003830, 0.00003953, 0.00001737, 0.00005316, 0.00000933, -0.00002816, -0.00002595, 0.00000010, -0.00001118, 0.00000259, -0.00000879, -0.00000144, -0.00000154, -0.00000166, -0.00000488, -0.00000404, -0.00000072, -0.00001276, 0.00000217, -0.00002264, -0.00000291, -0.00009153, -0.00001126, 0.00014122, 0.00009638, 0.00011503, -0.00010412, -0.00002453, -0.00001615, -0.00000101, -0.00001161, -0.00000642, -0.00000541, 0.00000527, -0.00000533, 0.00000257, -0.00000907, 0.00001056, -0.00000309, -0.00000074, -0.00000091, -0.00000591, 0.00001269, -0.00001194, -0.00000472, 0.00000104, -0.00000694, 0.00000474, -0.00001445, -0.00000801, 0.00000456, -0.00000465, 0.00001228, -0.00001459, 0.00000838, 0.00000669, -0.00000558, -0.00000422, 0.00000184, 0.00000480, -0.00000459, 0.00000328, -0.00000245, 0.00001035, -0.00000438, -0.00000336, -0.00000447, 0.00000267, 0.00000476, 0.00000686, -0.00000384, -0.00001181, 0.00000783, 0.00000224, 0.00000039, 0.00000382, -0.00000350, 0.00000132, -0.00000308, },
		{ -0.00000760, 0.00000404, 0.00001412, -0.00000581, -0.00001107, 0.00000115, 0.00000754, 0.00000326, -0.00000061, -0.00000162, -0.00000521, -0.00000452, 0.00000815, 0.00000515, -0.00000482, 0.00000168, 0.00000371, -0.00000411, -0.00000860, 0.00000875, 0.00000899, -0.00000454, -0.00000683, -0.00000069, 0.00000745, -0.00000336, -0.00001032, 0.00000201, -0.00000800, -0.00000096, -0.00000638, 0.00001006, -0.00000612, -0.00000348, 0.00000200, 0.00000140, -0.00000551, -0.00000770, 0.00000968, 0.00000408, -0.00000803, 0.00000141, 0.00000306, -0.00000353, -0.00000688, -0.00001084, -0.00000068, -0.00002656, -0.00002026, 0.00002798, 0.00004970, 0.00001461, 0.00005443, 0.00001849, -0.00002607, -0.00003071, -0.00000012, -0.00001168, 0.00000196, -0.00000765, -0.00000269, -0.00000166, -0.00000101, -0.00000467, -0.00000445, -0.00000051, -0.00001131, -0.00000011, -0.00001763, -0.00000628, -0.00007122, -0.00004574, 0.00013925, 0.00008273, 0.00013262, -0.00007153, -0.00005091, -0.00000656, -0.00000950, -0.00000438, -0.00001201, -0.00000275, 0.00000196, -0.00000155, -0.00000060, -0.00000585, 0.00000564, 0.00000257, -0.00000546, 0.00000335, -0.00000970, 0.00001358, -0.00000942, -0.00000637, 0.00000129, -0.00000641, 0.00000252, -0.00000830, -0.00001452, 0.00000725, -0.00000637, 0.00001216, -0.00001143, 0.00000229, 0.00001158, -0.00000634, -0.00000377, -0.00000055, 0.00000690, -0.00000534, 0.00000354, -0.00000320, 0.00000998, -0.00000201, -0.00000447, -0.00000402, 0.00000078, 0.00000518, 0.00000649, -0.00000025, -0.00001381, 0.00000527, 0.00000518, -0.00000123, 0.00000460, -0.00000326, 0.00000055, -0.00000134, },
		{ -0.00000733, 0.00000076, 0.00001445, -0.00000123, -0.00001309, -0.00000057, 0.00000717, 0.00000466, -0.00000050, -0.00000133, -0.00000460, -0.00000563, 0.00000654, 0.00000700, -0.00000441, 0.00000029, 0.00000400, -0.00000176, -0.00000925, 0.00000438, 0.00001179, -0.00000276, -0.00000705, -0.00000255, 0.00000723, -0.00000020, -0.00001201, 0.00000129, -0.00000639, -0.00000244, -0.00000554, 0.00000669, -0.00000129, -0.00000733, 0.00000437, -0.00000108, -0.00000146, -0.00001100, 0.00000951, 0.00000451, -0.00000570, -0.00000245, 0.00000613, -0.00000512, -0.00000382, -0.00001374, 0.00000142, -0.00002290, -0.00002649, 0.00001637, 0.00005671, 0.00001615, 0.00005184, 0.00002956, -0.00002367, -0.00003436, -0.00000211, -0.00001072, -0.00000005, -0.00000558, -0.00000438, -0.00000147, -0.00000095, -0.00000392, -0.00000498, -0.00000068, -0.00000929, -0.00000298, -0.00001228, -0.00001068, -0.00005158, -0.00007178, 0.00012830, 0.00007834, 0.00013240, -0.00002729, -0.00007584, -0.00000033, -0.00001602, 0.00000098, -0.00001569, -0.00000140, -0.00000074, 0.00000166, -0.00000317, -0.00000307, 0.00000128, 0.00000718, -0.00000872, 0.00000591, -0.00001146, 0.00001254, -0.00000607, -0.00000784, 0.00000109, -0.00000519, -0.00000044, -0.00000261, -0.00001856, 0.00000671, -0.00000510, 0.00000938, -0.00000626, -0.00000440, 0.00001538, -0.00000577, -0.00000384, -0.00000249, 0.00000805, -0.00000499, 0.00000288, -0.00000301, 0.00000849, 0.00000108, -0.00000582, -0.00000321, -0.00000126, 0.00000555, 0.00000576, 0.00000313, -0.00001437, 0.00000157, 0.00000792, -0.00000189, 0.00000435, -0.00000218, -0.00000078, 0.00000030, },
		{ -0.00000648, -0.00000193, 0.00001322, 0.00000371, -0.00001396, -0.00000294, 0.00000675, 0.00000572, 0.00000011, -0.00000141, -0.00000372, -0.00000656, 0.00000479, 0.00000832, -0.00000332, -0.00000122, 0.00000404, 0.00000005, -0.00000840, -0.00000036, 0.00001324, -0.00000012, -0.00000717, -0.00000404, 0.00000613, 0.00000300, -0.00001256, -0.00000070, -0.00000404, -0.00000431, -0.00000413, 0.00000295, 0.00000303, -0.00001010, 0.00000565, -0.00000253, 0.00000109, -0.00001219, 0.00000756, 0.00000586, -0.00000413, -0.00000489, 0.00000720, -0.00000471, -0.00000281, -0.00001407, 0.00000035, -0.00001691, -0.00003120, 0.00000375, 0.00005978, 0.00002178, 0.00004641, 0.00004124, -0.00002009, -0.00003728, -0.00000544, -0.00000884, -0.00000281, -0.00000321, -0.00000601, -0.00000125, -0.00000126, -0.00000289, -0.00000543, -0.00000117, -0.00000709, -0.00000571, -0.00000763, -0.00001449, -0.00003503, -0.00008770, 0.00010803, 0.00008268, 0.00011869, 0.00002032, -0.00009242, -0.00000108, -0.00001840, 0.00000283, -0.00001632, -0.00000217, -0.00000208, 0.00000350, -0.00000437, -0.00000148, -0.00000179, 0.00001005, -0.00000999, 0.00000642, -0.00001111, 0.00000980, -0.00000221, -0.00000910, 0.00000069, -0.00000380, -0.00000316, 0.00000141, -0.00001935, 0.00000302, -0.00000162, 0.00000511, -0.00000052, -0.00001019, 0.00001700, -0.00000341, -0.00000459, -0.00000376, 0.00000801, -0.00000350, 0.00000141, -0.00000200, 0.00000614, 0.00000448, -0.00000705, -0.00000238, -0.00000310, 0.00000562, 0.00000501, 0.00000585, -0.00001336, -0.00000285, 0.00000988, -0.00000136, 0.00000320, -0.00000058, -0.00000222, 0.00000150, },
		{ -0.00000514, -0.00000393, 0.00001086, 0.00000817, -0.00001321, -0.00000599, 0.00000630, 0.00000634, 0.00000122, -0.00000176, -0.00000272, -0.00000722, 0.00000292, 0.00000911, -0.00000176, -0.00000260, 0.00000377, 0.00000136, -0.00000659, -0.00000456, 0.00001286, 0.00000332, -0.00000718, -0.00000504, 0.00000431, 0.00000583, -0.00001179, -0.00000364, -0.00000157, -0.00000593, -0.00000286, -0.00000032, 0.00000601, -0.00001118, 0.00000537, -0.00000248, 0.00000178, -0.00001120, 0.00000401, 0.00000795, -0.00000336, -0.00000587, 0.00000637, -0.00000263, -0.00000339, -0.00001254, -0.00000263, -0.00001049, -0.00003307, -0.00000949, 0.00005851, 0.00003061, 0.00003964, 0.00005196, -0.00001436, -0.00003987, -0.00000957, -0.00000677, -0.00000557, -0.00000125, -0.00000709, -0.00000133, -0.00000163, -0.00000193, -0.00000554, -0.00000198, -0.00000498, -0.00000783, -0.00000428, -0.00001679, -0.00002276, -0.00009366, 0.00008034, 0.00009263, 0.00009798, 0.00006290, -0.00009557, -0.00001022, -0.00001628, 0.00000099, -0.00001396, -0.00000487, -0.00000209, 0.00000385, -0.00000409, -0.00000118, -0.00000343, 0.00001113, -0.00000931, 0.00000510, -0.00000908, 0.00000598, 0.00000168, -0.00001000, 0.00000021, -0.00000262, -0.00000504, 0.00000323, -0.00001709, -0.00000279, 0.00000262, 0.00000089, 0.00000430, -0.00001382, 0.00001582, 0.00000067, -0.00000585, -0.00000435, 0.00000686, -0.00000113, -0.00000054, -0.00000055, 0.00000339, 0.00000765, -0.00000770, -0.00000193, -0.00000438, 0.00000513, 0.00000452, 0.00000758, -0.00001088, -0.00000742, 0.00001063, 0.00000032, 0.00000154, 0.00000103, -0.00000331, 0.00000200, },
		{ -0.00000349, -0.00000521, 0.00000791, 0.00001146, -0.00001070, -0.00000944, 0.00000568, 0.00000658, 0.00000263, -0.00000211, -0.00000184, -0.00000742, 0.00000091, 0.00000946, -0.00000000, -0.00000358, 0.00000311, 0.00000230, -0.00000440, -0.00000751, 0.00001056, 0.00000721, -0.00000687, -0.00000566, 0.00000216, 0.00000783, -0.00000969, -0.00000706, 0.00000045, -0.00000683, -0.00000220, -0.00000259, 0.00000730, -0.00001048, 0.00000360, -0.00000109, 0.00000088, -0.00000858, -0.00000038, 0.00001015, -0.00000291, -0.00000585, 0.00000435, 0.00000018, -0.00000450, -0.00001036, -0.00000594, -0.00000545, -0.00003139, -0.00002253, 0.00005266, 0.00004137, 0.00003325, 0.00006026, -0.00000581, -0.00004237, -0.00001399, -0.00000516, -0.00000769, -0.00000022, -0.00000729, -0.00000191, -0.00000183, -0.00000131, -0.00000517, -0.00000305, -0.00000316, -0.00000911, -0.00000237, -0.00001737, -0.00001481, -0.00009126, 0.00004856, 0.00010377, 0.00007697, 0.00009397, -0.00008343, -0.00002639, -0.00001117, -0.00000332, -0.00000978, -0.00000851, -0.00000147, 0.00000322, -0.00000279, -0.00000172, -0.00000406, 0.00001083, -0.00000722, 0.00000261, -0.00000614, 0.00000189, 0.00000502, -0.00001031, -0.00000035, -0.00000176, -0.00000588, 0.00000300, -0.00001268, -0.00000913, 0.00000597, -0.00000190, 0.00000712, -0.00001465, 0.00001190, 0.00000587, -0.00000707, -0.00000456, 0.00000494, 0.00000160, -0.00000245, 0.00000086, 0.00000080, 0.00001004, -0.00000737, -0.00000214, -0.00000488, 0.00000395, 0.00000447, 0.00000823, -0.00000728, -0.00001152, 0.00000990, 0.00000286, -0.00000011, 0.00000219, -0.00000374, 0.00000173, },
		{ -0.00000181, -0.00000577, 0.00000484, 0.00001321, -0.00000668, -0.00001273, 0.00000456, 0.00000667, 0.00000402, -0.00000214, -0.00000137, -0.00000703, -0.00000128, 0.00000948, 0.00000171, -0.00000399, 0.00000206, 0.00000297, -0.00000228, -0.00000888, 0.00000670, 0.00001088, -0.00000588, -0.00000618, 0.00000016, 0.00000864, -0.00000640, -0.00001040, 0.00000152, -0.00000676, -0.00000230, -0.00000375, 0.00000702, -0.00000842, 0.00000086, 0.00000108, -0.00000090, -0.00000528, -0.00000459, 0.00001160, -0.00000212, -0.00000556, 0.00000209, 0.00000265, -0.00000508, -0.00000869, -0.00000825, -0.00000290, -0.00002639, -0.00003404, 0.00004218, 0.00005255, 0.00002877, 0.00006512, 0.00000572, -0.00004460, -0.00001851, -0.00000436, -0.00000888, -0.00000029, -0.00000662, -0.00000292, -0.00000182, -0.00000113, -0.00000434, -0.00000421, -0.00000181, -0.00000951, -0.00000175, -0.00001649, -0.00001049, -0.00008292, 0.00001649, 0.00011184, 0.00006092, 0.00011026, -0.00005772, -0.00004578, -0.00000597, -0.00000804, -0.00000551, -0.00001169, -0.00000121, 0.00000237, -0.00000122, -0.00000242, -0.00000428, 0.00000977, -0.00000443, -0.00000029, -0.00000313, -0.00000173, 0.00000734, -0.00000990, -0.00000104, -0.00000118, -0.00000585, 0.00000134, -0.00000747, -0.00001440, 0.00000716, -0.00000249, 0.00000755, -0.00001280, 0.00000603, 0.00001114, -0.00000755, -0.00000478, 0.00000271, 0.00000412, -0.00000378, 0.00000177, -0.00000116, 0.00001125, -0.00000586, -0.00000312, -0.00000456, 0.00000215, 0.00000480, 0.00000795, -0.00000307, -0.00001457, 0.00000766, 0.00000581, -0.00000125, 0.00000259, -0.00000342, 0.00000086, },
		{ -0.00000040, -0.00000566, 0.00000202, 0.00001335, -0.00000176, -0.00001520, 0.00000263, 0.00000684, 0.00000508, -0.00000161, -0.00000144, -0.00000605, -0.00000355, 0.00000918, 0.00000325, -0.00000382, 0.00000076, 0.00000339, -0.00000049, -0.00000876, 0.00000204, 0.00001351, -0.00000383, -0.00000688, -0.00000130, 0.00000814, -0.00000234, -0.00001308, 0.00000132, -0.00000571, -0.00000305, -0.00000401, 0.00000563, -0.00000565, -0.00000212, 0.00000331, -0.00000273, -0.00000227, -0.00000770, 0.00001165, -0.00000051, -0.00000556, 0.00000034, 0.00000399, -0.00000445, -0.00000812, -0.00000903, -0.00000290, -0.00001933, -0.00004236, 0.00002731, 0.00006250, 0.00002742, 0.00006614, 0.00001978, -0.00004595, -0.00002322, -0.00000443, -0.00000918, -0.00000127, -0.00000540, -0.00000408, -0.00000175, -0.00000127, -0.00000327, -0.00000521, -0.00000107, -0.00000909, -0.00000212, -0.00001461, -0.00000877, -0.00007120, -0.00001250, 0.00011373, 0.00005267, 0.00011199, -0.00002314, -0.00006322, -0.00000381, -0.00001113, -0.00000276, -0.00001325, -0.00000205, 0.00000192, -0.00000002, -0.00000269, -0.00000460, 0.00000848, -0.00000152, -0.00000292, -0.00000068, -0.00000436, 0.00000842, -0.00000873, -0.00000189, -0.00000074, -0.00000528, -0.00000094, -0.00000275, -0.00001740, 0.00000563, -0.00000089, 0.00000602, -0.00000911, -0.00000053, 0.00001527, -0.00000664, -0.00000533, 0.00000062, 0.00000590, -0.00000415, 0.00000190, -0.00000222, 0.00001111, -0.00000323, -0.00000469, -0.00000365, 0.00000001, 0.00000529, 0.00000708, 0.00000109, -0.00001608, 0.00000405, 0.00000863, -0.00000152, 0.00000220, -0.00000255, -0.00000029, },
		{ 0.00000048, -0.00000497, -0.00000029, 0.00001211, 0.00000329, -0.00001620, -0.00000028, 0.00000716, 0.00000566, -0.00000046, -0.00000198, -0.00000465, -0.00000567, 0.00000849, 0.00000463, -0.00000322, -0.00000054, 0.00000346, 0.00000096, -0.00000754, -0.00000252, 0.00001441, -0.00000056, -0.00000784, -0.00000210, 0.00000652, 0.00000187, -0.00001451, -0.00000025, -0.00000392, -0.00000414, -0.00000375, 0.00000370, -0.00000287, -0.00000471, 0.00000497, -0.00000394, -0.00000023, -0.00000918, 0.00001006, 0.00000201, -0.00000602, -0.00000062, 0.00000401, -0.00000262, -0.00000854, -0.00000854, -0.00000460, -0.00001207, -0.00004603, 0.00000892, 0.00006946, 0.00002996, 0.00006352, 0.00003536, -0.00004552, -0.00002840, -0.00000516, -0.00000895, -0.00000268, -0.00000412, -0.00000500, -0.00000185, -0.00000150, -0.00000225, -0.00000577, -0.00000104, -0.00000801, -0.00000309, -0.00001227, -0.00000863, -0.00005834, -0.00003604, 0.00010807, 0.00005250, 0.00010232, 0.00001399, -0.00007377, -0.00000693, -0.00001145, -0.00000232, -0.00001273, -0.00000419, 0.00000205, 0.00000055, -0.00000230, -0.00000521, 0.00000719, 0.00000115, -0.00000493, 0.00000089, -0.00000584, 0.00000832, -0.00000698, -0.00000286, -0.00000032, -0.00000451, -0.00000311, 0.00000056, -0.00001764, 0.00000167, 0.00000210, 0.00000351, -0.00000479, -0.00000642, 0.00001725, -0.00000400, -0.00000631, -0.00000107, 0.00000667, -0.00000345, 0.00000126, -0.00000239, 0.00000980, 0.00000015, -0.00000644, -0.00000256, -0.00000204, 0.00000556, 0.00000606, 0.00000460, -0.00001579, -0.00000055, 0.00001082, -0.00000075, 0.00000121, -0.00000147, -0.00000134, },
		{ 0.00000074, -0.00000391, -0.00000194, 0.00000988, 0.00000771, -0.00001537, -0.00000405, 0.00000752, 0.00000581, 0.00000114, -0.00000273, -0.00000316, -0.00000733, 0.00000724, 0.00000594, -0.00000241, -0.00000155, 0.00000307, 0.00000215, -0.00000575, -0.00000615, 0.00001323, 0.00000373, -0.00000884, -0.00000241, 0.00000430, 0.00000547, -0.00001427, -0.00000307, -0.00000180, -0.00000518, -0.00000336, 0.00000177, -0.00000061, -0.00000649, 0.00000574, -0.00000422, 0.00000056, -0.00000900, 0.00000710, 0.00000507, -0.00000670, -0.00000096, 0.00000305, -0.00000015, -0.00000925, -0.00000763, -0.00000666, -0.00000644, -0.00004439, -0.00001132, 0.00007163, 0.00003662, 0.00005807, 0.00005109, -0.00004229, -0.00003434, -0.00000629, -0.00000861, -0.00000404, -0.00000327, -0.00000536, -0.00000230, -0.00000164, -0.00000154, -0.00000572, -0.00000167, -0.00000650, -0.00000429, -0.00000994, -0.00000920, -0.00004599, -0.00005295, 0.00009519, 0.00005858, 0.00008602, 0.00004743, -0.00007404, -0.00001576, -0.00000923, -0.00000394, -0.00001055, -0.00000714, 0.00000244, 0.00000067, -0.00000142, -0.00000592, 0.00000582, 0.00000356, -0.00000623, 0.00000157, -0.00000630, 0.00000734, -0.00000492, -0.00000387, 0.00000011, -0.00000379, -0.00000467, 0.00000206, -0.00001538, -0.00000375, 0.00000525, 0.00000120, -0.00000103, -0.00001052, 0.00001655, 0.00000021, -0.00000749, -0.00000235, 0.00000642, -0.00000189, 0.00000010, -0.00000195, 0.00000770, 0.00000373, -0.00000784, -0.00000178, -0.00000358, 0.00000531, 0.00000529, 0.00000703, -0.00001373, -0.00000558, 0.00001190, 0.00000100, -0.00000001, -0.00000058, -0.00000199, },
		{ 0.00000048, -0.00000274, -0.00000286, 0.00000713, 0.00001091, -0.00001269, -0.00000826, 0.00000759, 0.00000576, 0.00000287, -0.00000331, -0.00000199, -0.00000821, 0.00000531, 0.00000728, -0.00000164, -0.00000207, 0.00000218, 0.00000317, -0.00000384, -0.00000836, 0.00001009, 0.00000843, -0.00000937, -0.00000270, 0.00000216, 0.00000777, -0.00001216, -0.00000677, 0.00000015, -0.00000582, -0.00000316, 0.00000016, 0.00000091, -0.00000739, 0.00000563, -0.00000362, 0.00000026, -0.00000757, 0.00000341, 0.00000802, -0.00000710, -0.00000116, 0.00000173, 0.00000213, -0.00000941, -0.00000724, -0.00000791, -0.00000358, -0.00003790, -0.00003096, 0.00006737, 0.00004701, 0.00005107, 0.00006544, -0.00003541, -0.00004112, -0.00000768, -0.00000850, -0.00000498, -0.00000312, -0.00000507, -0.00000311, -0.00000163, -0.00000124, -0.00000508, -0.00000274, -0.00000491, -0.00000536, -0.00000795, -0.00000987, -0.00003518, -0.00006313, 0.00007666, 0.00006793, 0.00006814, 0.00007243, -0.00006312, -0.00002874, -0.00000595, -0.00000652, -0.00000775, -0.00001000, 0.00000246, 0.00000075, -0.00000047, -0.00000636, 0.00000414, 0.00000579, -0.00000692, 0.00000154, -0.00000605, 0.00000586, -0.00000283, -0.00000480, 0.00000057, -0.00000323, -0.00000546, 0.00000188, -0.00001145, -0.00000929, 0.00000732, 0.00000007, 0.00000136, -0.00001230, 0.00001330, 0.00000534, -0.00000835, -0.00000341, 0.00000540, 0.00000014, -0.00000114, -0.00000129, 0.00000534, 0.00000691, -0.00000840, -0.00000167, -0.00000431, 0.00000435, 0.00000498, 0.00000823, -0.00001024, -0.00001036, 0.00001157, 0.00000348, -0.00000105, -0.00000017, -0.00000211, },
		{ -0.00000006, -0.00000174, -0.00000306, 0.00000431, 0.00001258, -0.00000851, -0.00001226, 0.00000697, 0.00000580, 0.00000434, -0.00000334, -0.00000145, -0.00000814, 0.00000272, 0.00000862, -0.00000099, -0.00000203, 0.00000090, 0.00000399, -0.00000210, -0.00000905, 0.00000562, 0.00001263, -0.00000880, -0.00000347, 0.00000074, 0.00000836, -0.00000838, -0.00001073, 0.00000141, -0.00000580, -0.00000328, -0.00000101, 0.00000169, -0.00000758, 0.00000486, -0.00000239, -0.00000074, -0.00000548, -0.00000026, 0.00001022, -0.00000674, -0.00000166, 0.00000069, 0.00000353, -0.00000841, -0.00000786, -0.00000780, -0.00000359, -0.00002827, -0.00004713, 0.00005564, 0.00005996, 0.00004418, 0.00007692, -0.00002434, -0.00004851, -0.00000937, -0.00000875, -0.00000541, -0.00000363, -0.00000429, -0.00000406, -0.00000158, -0.00000126, -0.00000407, -0.00000392, -0.00000359, -0.00000599, -0.00000652, -0.00001028, -0.00002640, -0.00006727, 0.00005475, 0.00007737, 0.00005281, 0.00008658, -0.00004268, -0.00004275, -0.00000375, -0.00000862, -0.00000550, -0.00001186, 0.00000156, 0.00000118, 0.00000016, -0.00000619, 0.00000198, 0.00000790, -0.00000707, 0.00000098, -0.00000536, 0.00000418, -0.00000090, -0.00000560, 0.00000103, -0.00000284, -0.00000552, 0.00000047, -0.00000691, -0.00001369, 0.00000742, 0.00000055, 0.00000216, -0.00001188, 0.00000826, 0.00001035, -0.00000827, -0.00000451, 0.00000395, 0.00000211, -0.00000201, -0.00000081, 0.00000318, 0.00000921, -0.00000781, -0.00000239, -0.00000418, 0.00000274, 0.00000510, 0.00000833, -0.00000591, -0.00001418, 0.00000968, 0.00000634, -0.00000153, -0.00000035, -0.00000181, },
		{ -0.00000055, -0.00000115, -0.00000267, 0.00000180, 0.00001267, -0.00000345, -0.00001532, 0.00000529, 0.00000616, 0.00000527, -0.00000262, -0.00000163, -0.00000717, -0.00000032, 0.00000980, -0.00000039, -0.00000160, -0.00000050, 0.00000447, -0.00000056, -0.00000852, 0.00000071, 0.00001538, -0.00000663, -0.00000504, 0.00000035, 0.00000726, -0.00000357, -0.00001413, 0.00000150, -0.00000500, -0.00000373, -0.00000176, 0.00000191, -0.00000733, 0.00000369, -0.00000084, -0.00000202, -0.00000331, -0.00000331, 0.00001123, -0.00000539, -0.00000266, 0.00000021, 0.00000379, -0.00000621, -0.00000940, -0.00000661, -0.00000552, -0.00001796, -0.00005717, 0.00003650, 0.00007341, 0.00003928, 0.00008431, -0.00000903, -0.00005588, -0.00001164, -0.00000929, -0.00000546, -0.00000455, -0.00000340, -0.00000485, -0.00000170, -0.00000140, -0.00000300, -0.00000481, -0.00000285, -0.00000602, -0.00000573, -0.00001027, -0.00001972, -0.00006655, 0.00003185, 0.00008424, 0.00004257, 0.00008994, -0.00001631, -0.00005407, -0.00000460, -0.00000914, -0.00000463, -0.00001219, -0.00000047, 0.00000207, 0.00000033, -0.00000533, -0.00000058, 0.00000974, -0.00000663, -0.00000001, -0.00000442, 0.00000249, 0.00000080, -0.00000627, 0.00000145, -0.00000254, -0.00000508, -0.00000152, -0.00000282, -0.00001607, 0.00000530, 0.00000243, 0.00000175, -0.00000993, 0.00000253, 0.00001418, -0.00000676, -0.00000582, 0.00000239, 0.00000361, -0.00000220, -0.00000073, 0.00000150, 0.00001040, -0.00000606, -0.00000381, -0.00000341, 0.00000077, 0.00000542, 0.00000773, -0.00000146, -0.00001645, 0.00000628, 0.00000914, -0.00000125, -0.00000100, -0.00000135, },
		},
		{ { -0.00000136, 0.00000010, 0.00000017, 0.00000252, -0.00000913, 0.00000042, 0.00001608, -0.00000079, -0.00000494, -0.00000760, 0.00001654, -0.00000030, 0.00000249, 0.00000023, -0.00000391, 0.00000033, -0.00000629, 0.00000136, -0.00000584, -0.00000545, 0.00000341, 0.00000122, -0.00001268, 0.00000284, 0.00000130, 0.00000010, -0.00001041, -0.00000269, 0.00001501, -0.00000091, -0.00000345, 0.00000189, 0.00000092, -0.00000021, -0.00000699, -0.00000084, -0.00000213, 0.00000049, -0.00000263, 0.00000015, -0.00000235, -0.00000408, -0.00000733, -0.00000462, -0.00001192, -0.00000935, -0.00005587, 0.00001219, 0.00008162, 0.00003656, 0.00008605, 0.00001057, -0.00005983, -0.00001194, -0.00001330, -0.00000655, -0.00000266, -0.00000080, -0.00000340, -0.00000290, -0.00000196, -0.00000562, 0.00000832, -0.00000363, -0.00000301, -0.00000363, 0.00000832, -0.00000562, -0.00000196, -0.00000290, -0.00000340, -0.00000080, -0.00000266, -0.00000655, -0.00001330, -0.00001194, -0.00005983, 0.00001057, 0.00008605, 0.00003656, 0.00008162, 0.00001219, -0.00005587, -0.00000935, -0.00001192, -0.00000462, -0.00000733, -0.00000408, -0.00000235, 0.00000015, -0.00000263, 0.00000049, -0.00000213, -0.00000084, -0.00000699, -0.00000021, 0.00000092, 0.00000189, -0.00000345, -0.00000091, 0.00001501, -0.00000269, -0.00001041, 0.00000010, 0.00000130, 0.00000284, -0.00001268, 0.00000122, 0.00000341, -0.00000545, -0.00000584, 0.00000136, -0.00000629, 0.00000033, -0.00000391, 0.00000023, 0.00000249, -0.00000030, 0.00001654, -0.00000760, -0.00000494, -0.00000079, 0.00001608, 0.00000042, -0.00000913, 0.00000252, 0.00000017, 0.00000010, },
		{ -0.00000138, -0.00000000, -0.00000030, 0.00000264, -0.00000608, -0.00000493, 0.00001620, 0.00000226, -0.00000349, -0.00000983, 0.00001204, 0.00000661, -0.00000202, 0.00000490, -0.00000721, 0.00000258, -0.00000716, 0.00000087, -0.00000445, -0.00000575, 0.00000052, 0.00000436, -0.00001255, 0.00000032, 0.00000201, 0.00000047, -0.00000707, -0.00000847, 0.00001472, 0.00000288, -0.00000451, 0.00000139, 0.00000095, 0.00000094, -0.00000683, -0.00000166, -0.00000192, 0.00000009, -0.00000233, -0.00000039, -0.00000152, -0.00000336, -0.00000752, -0.00000403, -0.00001217, -0.00000597, -0.00004610, -0.00002343, 0.00008681, 0.00004180, 0.00008050, 0.00004003, -0.00006307, -0.00001980, -0.00001134, -0.00000963, -0.00000213, -0.00000216, -0.00000155, -0.00000508, 0.00000055, -0.00000806, 0.00000743, -0.00000094, -0.00000351, -0.00000454, 0.00000654, -0.00000128, -0.00000510, -0.00000068, -0.00000507, 0.00000030, -0.00000327, -0.00000369, -0.00001442, -0.00000750, -0.00005157, -0.00001488, 0.00008205, 0.00004097, 0.00006641, 0.00004431, -0.00005322, -0.00001909, -0.00000944, -0.00000668, -0.00000612, -0.00000535, -0.00000269, -0.00000006, -0.00000231, 0.00000032, -0.00000176, -0.00000066, -0.00000635, -0.00000187, 0.00000129, 0.00000149, -0.00000134, -0.00000437, 0.00001378, 0.00000275, -0.00001123, -0.00000223, 0.00000150, 0.00000396, -0.00001070, -0.00000333, 0.00000625, -0.00000478, -0.00000653, 0.00000021, -0.00000359, -0.00000326, -0.00000012, -0.00000389, 0.00000571, -0.00000438, 0.00001645, -0.00000208, -0.00000733, -0.00000235, 0.00001324, 0.00000695, -0.00001133, 0.00000154, 0.00000077, 0.00000029, },
		{ -0.00000136, -0.00000007, -0.00000061, 0.00000222, -0.00000309, -0.00000819, 0.00001355, 0.00000641, -0.00000301, -0.00000906, 0.00000472, 0.00001354, -0.00000503, 0.00000750, -0.00000769, 0.00000181, -0.00000534, -0.00000152, -0.00000222, -0.00000627, -0.00000139, 0.00000522, -0.00000998, -0.00000369, 0.00000385, -0.00000106, -0.00000221, -0.00001326, 0.00001247, 0.00000675, -0.00000453, 0.00000035, 0.00000107, 0.00000170, -0.00000595, -0.00000303, -0.00000118, -0.00000086, -0.00000141, -0.00000165, -0.00000021, -0.00000338, -0.00000636, -0.00000526, -0.00000974, -0.00000827, -0.00002952, -0.00005264, 0.00007549, 0.00005743, 0.00006783, 0.00006921, -0.00005857, -0.00003098, -0.00000927, -0.00001239, -0.00000208, -0.00000330, -0.00000023, -0.00000631, 0.00000140, -0.00000788, 0.00000395, 0.00000294, -0.00000453, -0.00000408, 0.00000310, 0.00000341, -0.00000740, 0.00000052, -0.00000584, 0.00000063, -0.00000344, -0.00000167, -0.00001412, -0.00000611, -0.00004084, -0.00003382, 0.00006916, 0.00005092, 0.00004915, 0.00006542, -0.00003714, -0.00003219, -0.00000723, -0.00000870, -0.00000495, -0.00000643, -0.00000301, -0.00000062, -0.00000177, -0.00000015, -0.00000104, -0.00000092, -0.00000511, -0.00000389, 0.00000197, 0.00000040, 0.00000120, -0.00000686, 0.00001119, 0.00000725, -0.00000974, -0.00000546, 0.00000158, 0.00000447, -0.00000767, -0.00000779, 0.00000765, -0.00000297, -0.00000705, -0.00000163, -0.00000072, -0.00000605, 0.00000188, -0.00000543, 0.00000594, -0.00000449, 0.00001198, 0.00000537, -0.00000953, -0.00000300, 0.00000871, 0.00001309, -0.00001156, -0.00000065, 0.00000155, 0.00000050, },
		{ -0.00000117, -0.00000026, -0.00000067, 0.00000143, -0.00000059, -0.00000935, 0.00000905, 0.00001051, -0.00000264, -0.00000676, -0.00000262, 0.00001745, -0.00000422, 0.00000641, -0.00000446, -0.00000200, -0.00000146, -0.00000491, 0.00000016, -0.00000686, -0.00000223, 0.00000407, -0.00000579, -0.00000823, 0.00000592, -0.00000321, 0.00000222, -0.00001534, 0.00000769, 0.00001072, -0.00000393, -0.00000053, 0.00000078, 0.00000235, -0.00000466, -0.00000456, -0.00000039, -0.00000186, -0.00000041, -0.00000297, 0.00000092, -0.00000355, -0.00000463, -0.00000722, -0.00000626, -0.00001265, -0.00001388, -0.00006708, 0.00004613, 0.00007913, 0.00005375, 0.00009257, -0.00004389, -0.00004487, -0.00000785, -0.00001458, -0.00000262, -0.00000405, 0.00000026, -0.00000626, 0.00000045, -0.00000557, -0.00000085, 0.00000654, -0.00000498, -0.00000331, -0.00000046, 0.00000673, -0.00000764, 0.00000005, -0.00000536, -0.00000004, -0.00000293, -0.00000079, -0.00001232, -0.00000698, -0.00002981, -0.00004540, 0.00005005, 0.00006134, 0.00003620, 0.00007268, -0.00001158, -0.00004316, -0.00000826, -0.00000923, -0.00000477, -0.00000671, -0.00000368, -0.00000114, -0.00000136, -0.00000063, -0.00000024, -0.00000135, -0.00000356, -0.00000589, 0.00000256, -0.00000075, 0.00000323, -0.00000765, 0.00000727, 0.00001069, -0.00000681, -0.00000813, 0.00000049, 0.00000510, -0.00000465, -0.00001086, 0.00000670, 0.00000023, -0.00000762, -0.00000335, 0.00000101, -0.00000666, 0.00000114, -0.00000407, 0.00000341, -0.00000155, 0.00000522, 0.00001219, -0.00000988, -0.00000379, 0.00000402, 0.00001712, -0.00000896, -0.00000408, 0.00000245, 0.00000065, },
		{ -0.00000074, -0.00000068, -0.00000036, 0.00000033, 0.00000131, -0.00000893, 0.00000402, 0.00001326, -0.00000137, -0.00000461, -0.00000755, 0.00001654, 0.00000087, 0.00000210, 0.00000126, -0.00000694, 0.00000236, -0.00000743, 0.00000139, -0.00000675, -0.00000272, 0.00000213, -0.00000157, -0.00001171, 0.00000670, -0.00000418, 0.00000418, -0.00001347, 0.00000040, 0.00001459, -0.00000307, -0.00000074, -0.00000017, 0.00000305, -0.00000330, -0.00000572, -0.00000014, -0.00000222, -0.00000008, -0.00000350, 0.00000100, -0.00000296, -0.00000354, -0.00000820, -0.00000416, -0.00001503, -0.00000531, -0.00006373, 0.00000408, 0.00009786, 0.00004578, 0.00010485, -0.00001797, -0.00005996, -0.00000788, -0.00001617, -0.00000358, -0.00000453, 0.00000004, -0.00000527, -0.00000159, -0.00000245, -0.00000517, 0.00000826, -0.00000384, -0.00000320, -0.00000286, 0.00000763, -0.00000549, -0.00000190, -0.00000391, -0.00000144, -0.00000187, -0.00000091, -0.00000946, -0.00000893, -0.00002021, -0.00004995, 0.00002810, 0.00006802, 0.00003064, 0.00006798, 0.00001660, -0.00004672, -0.00001402, -0.00000812, -0.00000556, -0.00000628, -0.00000460, -0.00000162, -0.00000112, -0.00000112, 0.00000055, -0.00000181, -0.00000199, -0.00000745, 0.00000256, -0.00000128, 0.00000398, -0.00000640, 0.00000236, 0.00001300, -0.00000334, -0.00000923, -0.00000206, 0.00000593, -0.00000205, -0.00001205, 0.00000356, 0.00000415, -0.00000783, -0.00000475, 0.00000123, -0.00000512, -0.00000163, -0.00000118, -0.00000013, 0.00000215, -0.00000092, 0.00001582, -0.00000714, -0.00000544, 0.00000040, 0.00001798, -0.00000352, -0.00000817, 0.00000308, 0.00000079, },
		{ -0.00000013, -0.00000131, 0.00000025, -0.00000092, 0.00000262, -0.00000757, -0.00000042, 0.00001388, 0.00000123, -0.00000350, -0.00000913, 0.00001120, 0.00000840, -0.00000298, 0.00000662, -0.00001013, 0.00000367, -0.00000734, 0.00000042, -0.00000526, -0.00000369, 0.00000067, 0.00000118, -0.00001280, 0.00000503, -0.00000270, 0.00000267, -0.00000780, -0.00000822, 0.00001746, -0.00000179, -0.00000039, -0.00000144, 0.00000354, -0.00000195, -0.00000622, -0.00000080, -0.00000161, -0.00000077, -0.00000281, -0.00000036, -0.00000120, -0.00000380, -0.00000717, -0.00000483, -0.00001335, -0.00000552, -0.00004643, -0.00003942, 0.00010272, 0.00005013, 0.00010348, 0.00001756, -0.00007323, -0.00001055, -0.00001702, -0.00000493, -0.00000486, -0.00000061, -0.00000392, -0.00000369, -0.00000000, -0.00000747, 0.00000719, -0.00000086, -0.00000400, -0.00000369, 0.00000614, -0.00000166, -0.00000441, -0.00000227, -0.00000297, -0.00000073, -0.00000152, -0.00000630, -0.00001078, -0.00001308, -0.00004870, 0.00000658, 0.00006856, 0.00003232, 0.00005612, 0.00004055, -0.00004024, -0.00002329, -0.00000691, -0.00000637, -0.00000588, -0.00000523, -0.00000238, -0.00000079, -0.00000178, 0.00000136, -0.00000221, -0.00000065, -0.00000823, 0.00000162, -0.00000083, 0.00000321, -0.00000348, -0.00000267, 0.00001386, 0.00000020, -0.00000862, -0.00000542, 0.00000625, 0.00000041, -0.00001174, -0.00000072, 0.00000745, -0.00000687, -0.00000613, 0.00000040, -0.00000255, -0.00000467, 0.00000117, -0.00000255, 0.00000438, -0.00000418, 0.00001503, -0.00000142, -0.00000772, -0.00000181, 0.00001572, 0.00000369, -0.00001170, 0.00000281, 0.00000111, },
		{ 0.00000049, -0.00000195, 0.00000091, -0.00000199, 0.00000318, -0.00000564, -0.00000373, 0.00001245, 0.00000470, -0.00000324, -0.00000804, 0.00000366, 0.00001516, -0.00000564, 0.00000877, -0.00000928, 0.00000114, -0.00000420, -0.00000266, -0.00000257, -0.00000533, 0.00000024, 0.00000192, -0.00001123, 0.00000090, 0.00000105, -0.00000157, -0.00000018, -0.00001583, 0.00001778, 0.00000067, -0.00000023, -0.00000219, 0.00000324, -0.00000038, -0.00000615, -0.00000218, -0.00000033, -0.00000211, -0.00000134, -0.00000256, 0.00000111, -0.00000491, -0.00000467, -0.00000752, -0.00000869, -0.00001139, -0.00002423, -0.00007132, 0.00008574, 0.00006807, 0.00009062, 0.00005755, -0.00008005, -0.00001756, -0.00001662, -0.00000698, -0.00000486, -0.00000158, -0.00000267, -0.00000513, 0.00000091, -0.00000725, 0.00000366, 0.00000313, -0.00000507, -0.00000346, 0.00000325, 0.00000242, -0.00000622, -0.00000131, -0.00000402, 0.00000002, -0.00000211, -0.00000356, -0.00001173, -0.00000870, -0.00004349, -0.00001185, 0.00006237, 0.00003891, 0.00004251, 0.00005578, -0.00002473, -0.00003251, -0.00000784, -0.00000607, -0.00000619, -0.00000513, -0.00000362, -0.00000027, -0.00000263, 0.00000208, -0.00000236, 0.00000020, -0.00000797, -0.00000035, 0.00000053, 0.00000135, 0.00000012, -0.00000667, 0.00001284, 0.00000370, -0.00000687, -0.00000836, 0.00000515, 0.00000316, -0.00001065, -0.00000483, 0.00000887, -0.00000420, -0.00000769, -0.00000084, -0.00000026, -0.00000634, 0.00000147, -0.00000271, 0.00000421, -0.00000403, 0.00001048, 0.00000563, -0.00000945, -0.00000315, 0.00001145, 0.00001083, -0.00001319, 0.00000104, 0.00000178, },
		{ 0.00000102, -0.00000233, 0.00000125, -0.00000246, 0.00000281, -0.00000331, -0.00000596, 0.00000973, 0.00000802, -0.00000293, -0.00000589, -0.00000318, 0.00001813, -0.00000378, 0.00000651, -0.00000424, -0.00000434, 0.00000055, -0.00000632, 0.00000009, -0.00000687, 0.00000032, 0.00000130, -0.00000801, -0.00000439, 0.00000551, -0.00000632, 0.00000635, -0.00001965, 0.00001403, 0.00000503, -0.00000107, -0.00000156, 0.00000164, 0.00000164, -0.00000587, -0.00000364, 0.00000076, -0.00000310, -0.00000019, -0.00000430, 0.00000259, -0.00000546, -0.00000244, -0.00000993, -0.00000431, -0.00001720, -0.00000710, -0.00008232, 0.00004638, 0.00009374, 0.00007351, 0.00009400, -0.00007504, -0.00003053, -0.00001433, -0.00001034, -0.00000413, -0.00000306, -0.00000155, -0.00000579, 0.00000038, -0.00000521, -0.00000088, 0.00000652, -0.00000526, -0.00000318, 0.00000032, 0.00000531, -0.00000635, -0.00000148, -0.00000431, 0.00000015, -0.00000234, -0.00000162, -0.00001151, -0.00000660, -0.00003631, -0.00002550, 0.00005042, 0.00004720, 0.00003144, 0.00006107, -0.00000407, -0.00003760, -0.00001232, -0.00000458, -0.00000710, -0.00000456, -0.00000501, 0.00000014, -0.00000332, 0.00000229, -0.00000189, 0.00000023, -0.00000656, -0.00000317, 0.00000235, -0.00000082, 0.00000322, -0.00000866, 0.00000981, 0.00000709, -0.00000467, -0.00000983, 0.00000220, 0.00000620, -0.00000915, -0.00000796, 0.00000799, -0.00000013, -0.00000909, -0.00000219, 0.00000099, -0.00000601, -0.00000037, -0.00000102, 0.00000231, -0.00000164, 0.00000434, 0.00001155, -0.00000912, -0.00000451, 0.00000667, 0.00001604, -0.00001159, -0.00000236, 0.00000275, },
		{ 0.00000151, -0.00000229, 0.00000099, -0.00000207, 0.00000145, -0.00000070, -0.00000732, 0.00000660, 0.00001023, -0.00000173, -0.00000420, -0.00000725, 0.00001606, 0.00000239, 0.00000112, 0.00000280, -0.00000998, 0.00000414, -0.00000828, 0.00000106, -0.00000711, -0.00000014, 0.00000061, -0.00000477, -0.00000894, 0.00000849, -0.00000897, 0.00000903, -0.00001783, 0.00000585, 0.00001116, -0.00000304, 0.00000057, -0.00000112, 0.00000398, -0.00000558, -0.00000458, 0.00000081, -0.00000281, -0.00000041, -0.00000439, 0.00000201, -0.00000412, -0.00000219, -0.00000977, -0.00000329, -0.00001822, -0.00000120, -0.00007109, -0.00000674, 0.00011511, 0.00006220, 0.00011850, -0.00005410, -0.00004967, -0.00001025, -0.00001526, -0.00000246, -0.00000517, -0.00000042, -0.00000603, -0.00000084, -0.00000276, -0.00000459, 0.00000774, -0.00000364, -0.00000363, -0.00000162, 0.00000618, -0.00000461, -0.00000263, -0.00000400, -0.00000026, -0.00000220, -0.00000053, -0.00001034, -0.00000595, -0.00002888, -0.00003382, 0.00003477, 0.00005397, 0.00002539, 0.00005789, 0.00001691, -0.00003594, -0.00001967, -0.00000323, -0.00000754, -0.00000440, -0.00000573, -0.00000024, -0.00000321, 0.00000143, -0.00000045, -0.00000073, -0.00000411, -0.00000636, 0.00000395, -0.00000243, 0.00000491, -0.00000825, 0.00000521, 0.00000997, -0.00000240, -0.00000952, -0.00000207, 0.00000877, -0.00000702, -0.00001005, 0.00000532, 0.00000420, -0.00000947, -0.00000383, 0.00000124, -0.00000421, -0.00000320, 0.00000098, 0.00000029, 0.00000101, -0.00000084, 0.00001423, -0.00000594, -0.00000634, 0.00000255, 0.00001817, -0.00000684, -0.00000673, 0.00000353, },
		{ 0.00000214, -0.00000182, 0.00000012, -0.00000087, -0.00000059, 0.00000181, -0.00000788, 0.00000361, 0.00001093, 0.00000056, -0.00000354, -0.00000806, 0.00000994, 0.00001037, -0.00000422, 0.00000834, -0.00001248, 0.00000407, -0.00000693, -0.00000060, -0.00000544, -0.00000184, 0.00000082, -0.00000274, -0.00001130, 0.00000846, -0.00000793, 0.00000670, -0.00001061, -0.00000522, 0.00001748, -0.00000508, 0.00000322, -0.00000399, 0.00000582, -0.00000502, -0.00000485, -0.00000041, -0.00000105, -0.00000215, -0.00000265, -0.00000070, -0.00000084, -0.00000423, -0.00000651, -0.00000636, -0.00001344, -0.00000619, -0.00004510, -0.00005826, 0.00011840, 0.00006503, 0.00012595, -0.00001696, -0.00007233, -0.00000609, -0.00002086, -0.00000047, -0.00000749, 0.00000060, -0.00000616, -0.00000194, -0.00000115, -0.00000616, 0.00000616, -0.00000024, -0.00000481, -0.00000235, 0.00000512, -0.00000162, -0.00000407, -0.00000352, -0.00000093, -0.00000182, -0.00000011, -0.00000859, -0.00000603, -0.00002217, -0.00003742, 0.00001810, 0.00005668, 0.00002505, 0.00004910, 0.00003431, -0.00002750, -0.00002718, -0.00000422, -0.00000625, -0.00000556, -0.00000511, -0.00000190, -0.00000191, -0.00000069, 0.00000192, -0.00000244, -0.00000113, -0.00000917, 0.00000460, -0.00000284, 0.00000484, -0.00000582, 0.00000008, 0.00001167, -0.00000005, -0.00000788, -0.00000644, 0.00000973, -0.00000380, -0.00001149, 0.00000196, 0.00000740, -0.00000803, -0.00000598, 0.00000091, -0.00000201, -0.00000544, 0.00000173, -0.00000045, 0.00000242, -0.00000348, 0.00001301, -0.00000041, -0.00000823, -0.00000058, 0.00001720, -0.00000000, -0.00001076, 0.00000340, },
		{ 0.00000304, -0.00000102, -0.00000116, 0.00000071, -0.00000270, 0.00000364, -0.00000747, 0.00000084, 0.00001038, 0.00000338, -0.00000352, -0.00000668, 0.00000240, 0.00001664, -0.00000616, 0.00000955, -0.00000995, -0.00000032, -0.00000251, -0.00000425, -0.00000248, -0.00000440, 0.00000183, -0.00000204, -0.00001128, 0.00000532, -0.00000345, 0.00000052, -0.00000055, -0.00001589, 0.00002128, -0.00000527, 0.00000468, -0.00000537, 0.00000603, -0.00000357, -0.00000489, -0.00000230, 0.00000131, -0.00000440, -0.00000022, -0.00000418, 0.00000291, -0.00000708, -0.00000189, -0.00001128, -0.00000595, -0.00001610, -0.00001706, -0.00009211, 0.00009454, 0.00008377, 0.00011759, 0.00003120, -0.00009230, -0.00000555, -0.00002492, 0.00000016, -0.00000890, 0.00000090, -0.00000601, -0.00000271, -0.00000080, -0.00000552, 0.00000244, 0.00000381, -0.00000586, -0.00000246, 0.00000303, 0.00000149, -0.00000494, -0.00000332, -0.00000164, -0.00000133, -0.00000019, -0.00000653, -0.00000644, -0.00001639, -0.00003774, 0.00000303, 0.00005387, 0.00002960, 0.00003792, 0.00004603, -0.00001453, -0.00003147, -0.00000906, -0.00000295, -0.00000799, -0.00000331, -0.00000454, 0.00000019, -0.00000352, 0.00000447, -0.00000406, 0.00000139, -0.00001069, 0.00000365, -0.00000179, 0.00000322, -0.00000229, -0.00000437, 0.00001161, 0.00000252, -0.00000566, -0.00000959, 0.00000826, 0.00000063, -0.00001252, -0.00000122, 0.00000859, -0.00000468, -0.00000840, 0.00000023, -0.00000028, -0.00000611, 0.00000058, 0.00000036, 0.00000232, -0.00000354, 0.00000888, 0.00000570, -0.00000904, -0.00000308, 0.00001405, 0.00000714, -0.00001292, 0.00000176, },
		{ 0.00000401, 0.00000005, -0.00000249, 0.00000210, -0.00000412, 0.00000418, -0.00000588, -0.00000187, 0.00000922, 0.00000588, -0.00000324, -0.00000477, -0.00000373, 0.00001846, -0.00000305, 0.00000593, -0.00000315, -0.00000715, 0.00000268, -0.00000775, 0.00000002, -0.00000647, 0.00000256, -0.00000177, -0.00000984, 0.00000041, 0.00000250, -0.00000661, 0.00000845, -0.00002233, 0.00001995, -0.00000190, 0.00000343, -0.00000400, 0.00000382, -0.00000076, -0.00000531, -0.00000382, 0.00000284, -0.00000552, 0.00000110, -0.00000634, 0.00000488, -0.00000830, 0.00000115, -0.00001429, -0.00000074, -0.00002307, 0.00000101, -0.00009852, 0.00004449, 0.00011105, 0.00010154, 0.00008045, -0.00010112, -0.00001311, -0.00002472, -0.00000263, -0.00000797, -0.00000044, -0.00000502, -0.00000354, -0.00000117, -0.00000374, -0.00000167, 0.00000676, -0.00000561, -0.00000285, 0.00000100, 0.00000361, -0.00000464, -0.00000357, -0.00000232, -0.00000078, -0.00000065, -0.00000432, -0.00000712, -0.00001129, -0.00003631, -0.00000862, 0.00004560, 0.00003700, 0.00002732, 0.00005153, -0.00000002, -0.00003035, -0.00001723, 0.00000090, -0.00001031, -0.00000159, -0.00000693, 0.00000191, -0.00000576, 0.00000587, -0.00000439, 0.00000238, -0.00001009, 0.00000082, 0.00000055, 0.00000070, 0.00000126, -0.00000718, 0.00000963, 0.00000519, -0.00000341, -0.00001078, 0.00000447, 0.00000555, -0.00001277, -0.00000404, 0.00000791, -0.00000029, -0.00001018, -0.00000102, 0.00000080, -0.00000525, -0.00000183, 0.00000178, 0.00000160, -0.00000218, 0.00000386, 0.00001025, -0.00000769, -0.00000545, 0.00000995, 0.00001286, -0.00001222, -0.00000146, },
		{ 0.00000452, 0.00000141, -0.00000353, 0.00000275, -0.00000430, 0.00000313, -0.00000316, -0.00000452, 0.00000792, 0.00000743, -0.00000208, -0.00000359, -0.00000682, 0.00001521, 0.00000413, -0.00000035, 0.00000474, -0.00001283, 0.00000544, -0.00000863, 0.00000024, -0.00000667, 0.00000184, -0.00000079, -0.00000829, -0.00000441, 0.00000741, -0.00001145, 0.00001287, -0.00002195, 0.00001245, 0.00000524, -0.00000066, 0.00000008, -0.00000047, 0.00000309, -0.00000628, -0.00000423, 0.00000248, -0.00000433, 0.00000001, -0.00000571, 0.00000354, -0.00000621, 0.00000045, -0.00001265, -0.00000137, -0.00002204, 0.00000346, -0.00007852, -0.00001953, 0.00013218, 0.00008965, 0.00011951, -0.00009109, -0.00003170, -0.00001877, -0.00000991, -0.00000406, -0.00000387, -0.00000277, -0.00000502, -0.00000133, -0.00000232, -0.00000435, 0.00000715, -0.00000332, -0.00000403, -0.00000023, 0.00000423, -0.00000316, -0.00000406, -0.00000304, -0.00000021, -0.00000137, -0.00000209, -0.00000800, -0.00000670, -0.00003408, -0.00001642, 0.00003354, 0.00004435, 0.00001991, 0.00005112, 0.00001370, -0.00002404, -0.00002587, 0.00000244, -0.00001042, -0.00000168, -0.00000754, 0.00000181, -0.00000603, 0.00000492, -0.00000252, 0.00000117, -0.00000716, -0.00000357, 0.00000352, -0.00000189, 0.00000395, -0.00000800, 0.00000625, 0.00000745, -0.00000124, -0.00001016, -0.00000056, 0.00000954, -0.00001142, -0.00000688, 0.00000624, 0.00000369, -0.00001023, -0.00000322, 0.00000146, -0.00000366, -0.00000419, 0.00000243, 0.00000145, -0.00000082, -0.00000012, 0.00001178, -0.00000396, -0.00000768, 0.00000575, 0.00001613, -0.00000864, -0.00000553, },
		{ 0.00000392, 0.00000307, -0.00000402, 0.00000239, -0.00000314, 0.00000075, 0.00000013, -0.00000675, 0.00000649, 0.00000801, -0.00000009, -0.00000339, -0.00000690, 0.00000850, 0.00001227, -0.00000553, 0.00000981, -0.00001398, 0.00000353, -0.00000571, -0.00000236, -0.00000466, -0.00000063, 0.00000123, -0.00000732, -0.00000786, 0.00000950, -0.00001203, 0.00001123, -0.00001475, 0.00000032, 0.00001421, -0.00000587, 0.00000517, -0.00000515, 0.00000666, -0.00000713, -0.00000366, 0.00000024, -0.00000100, -0.00000325, -0.00000255, -0.00000075, -0.00000120, -0.00000377, -0.00000656, -0.00000757, -0.00001347, -0.00000690, -0.00004349, -0.00007814, 0.00013123, 0.00009166, 0.00014063, -0.00005912, -0.00006004, -0.00000863, -0.00002049, 0.00000171, -0.00000857, 0.00000031, -0.00000714, -0.00000080, -0.00000217, -0.00000468, 0.00000470, 0.00000065, -0.00000570, -0.00000072, 0.00000353, -0.00000102, -0.00000436, -0.00000388, 0.00000026, -0.00000209, -0.00000014, -0.00000874, -0.00000293, -0.00003104, -0.00002138, 0.00002048, 0.00004850, 0.00001758, 0.00004550, 0.00002565, -0.00001485, -0.00003108, -0.00000097, -0.00000690, -0.00000450, -0.00000561, -0.00000075, -0.00000379, 0.00000135, 0.00000145, -0.00000187, -0.00000265, -0.00000845, 0.00000605, -0.00000364, 0.00000523, -0.00000705, 0.00000236, 0.00000863, 0.00000101, -0.00000840, -0.00000530, 0.00001118, -0.00000786, -0.00001006, 0.00000446, 0.00000609, -0.00000802, -0.00000628, 0.00000187, -0.00000212, -0.00000540, 0.00000146, 0.00000235, -0.00000014, -0.00000222, 0.00001026, 0.00000117, -0.00000907, 0.00000170, 0.00001688, -0.00000314, -0.00000919, },
		{ 0.00000184, 0.00000484, -0.00000372, 0.00000101, -0.00000098, -0.00000219, 0.00000307, -0.00000792, 0.00000461, 0.00000808, 0.00000212, -0.00000359, -0.00000531, 0.00000121, 0.00001774, -0.00000637, 0.00000962, -0.00000944, -0.00000277, -0.00000012, -0.00000643, -0.00000167, -0.00000379, 0.00000347, -0.00000656, -0.00000979, 0.00000847, -0.00000845, 0.00000468, -0.00000354, -0.00001259, 0.00002146, -0.00000925, 0.00000858, -0.00000783, 0.00000815, -0.00000667, -0.00000308, -0.00000267, 0.00000290, -0.00000687, 0.00000118, -0.00000572, 0.00000428, -0.00000883, 0.00000081, -0.00001532, -0.00000270, -0.00002105, -0.00000977, -0.00011280, 0.00009890, 0.00010940, 0.00014356, -0.00000919, -0.00009143, 0.00000052, -0.00003054, 0.00000633, -0.00001236, 0.00000275, -0.00000900, -0.00000003, -0.00000307, -0.00000315, 0.00000059, 0.00000481, -0.00000678, -0.00000112, 0.00000229, 0.00000094, -0.00000407, -0.00000479, 0.00000039, -0.00000243, 0.00000110, -0.00000878, -0.00000068, -0.00002656, -0.00002517, 0.00000931, 0.00004716, 0.00002094, 0.00003589, 0.00003557, -0.00000563, -0.00003006, -0.00000985, -0.00000044, -0.00000922, -0.00000194, -0.00000492, 0.00000004, -0.00000363, 0.00000609, -0.00000522, 0.00000178, -0.00001215, 0.00000685, -0.00000378, 0.00000487, -0.00000482, -0.00000117, 0.00000838, 0.00000336, -0.00000618, -0.00000854, 0.00000984, -0.00000235, -0.00001326, 0.00000281, 0.00000667, -0.00000409, -0.00000936, 0.00000167, -0.00000088, -0.00000522, -0.00000082, 0.00000373, 0.00000023, -0.00000282, 0.00000694, 0.00000592, -0.00000866, -0.00000229, 0.00001567, 0.00000284, -0.00001118, },
		{ -0.00000143, 0.00000618, -0.00000246, -0.00000111, 0.00000149, -0.00000474, 0.00000473, -0.00000746, 0.00000200, 0.00000812, 0.00000388, -0.00000339, -0.00000365, -0.00000408, 0.00001828, -0.00000190, 0.00000455, -0.00000108, -0.00001049, 0.00000499, -0.00000928, 0.00000011, -0.00000583, 0.00000440, -0.00000497, -0.00001091, 0.00000532, -0.00000255, -0.00000366, 0.00000732, -0.00002172, 0.00002336, -0.00000809, 0.00000805, -0.00000669, 0.00000629, -0.00000401, -0.00000356, -0.00000464, 0.00000526, -0.00000848, 0.00000290, -0.00000848, 0.00000706, -0.00001124, 0.00000522, -0.00001938, 0.00000355, -0.00002919, 0.00000904, -0.00011398, 0.00003828, 0.00013417, 0.00013607, 0.00004833, -0.00011513, 0.00000147, -0.00003507, 0.00000603, -0.00001259, 0.00000265, -0.00000916, -0.00000020, -0.00000382, -0.00000136, -0.00000314, 0.00000726, -0.00000607, -0.00000219, 0.00000131, 0.00000207, -0.00000305, -0.00000552, -0.00000012, -0.00000210, 0.00000125, -0.00000760, -0.00000058, -0.00002009, -0.00002911, 0.00000174, 0.00003991, 0.00002870, 0.00002455, 0.00004296, 0.00000200, -0.00002282, -0.00002163, 0.00000587, -0.00001319, 0.00000117, -0.00000859, 0.00000336, -0.00000778, 0.00000913, -0.00000680, 0.00000417, -0.00001308, 0.00000497, -0.00000191, 0.00000300, -0.00000189, -0.00000382, 0.00000690, 0.00000542, -0.00000376, -0.00000992, 0.00000608, 0.00000376, -0.00001533, 0.00000071, 0.00000625, 0.00000009, -0.00001105, 0.00000021, 0.00000034, -0.00000429, -0.00000325, 0.00000432, 0.00000116, -0.00000288, 0.00000347, 0.00000865, -0.00000594, -0.00000612, 0.00001315, 0.00000804, -0.00001085, },
		{ -0.00000500, 0.00000640, -0.00000014, -0.00000353, 0.00000352, -0.00000605, 0.00000455, -0.00000532, -0.00000124, 0.00000824, 0.00000491, -0.00000240, -0.00000287, -0.00000623, 0.00001398, 0.00000608, -0.00000229, 0.00000706, -0.00001544, 0.00000622, -0.00000849, -0.00000100, -0.00000540, 0.00000299, -0.00000190, -0.00001184, 0.00000139, 0.00000329, -0.00001039, 0.00001389, -0.00002383, 0.00001815, -0.00000161, 0.00000316, -0.00000169, 0.00000139, 0.00000060, -0.00000537, -0.00000472, 0.00000472, -0.00000666, 0.00000111, -0.00000734, 0.00000533, -0.00000891, 0.00000405, -0.00001661, 0.00000142, -0.00002620, 0.00000818, -0.00008551, -0.00003480, 0.00014962, 0.00013009, 0.00010071, -0.00012034, -0.00001168, -0.00003054, -0.00000154, -0.00000782, -0.00000104, -0.00000667, -0.00000231, -0.00000320, -0.00000094, -0.00000468, 0.00000669, -0.00000307, -0.00000415, 0.00000094, 0.00000219, -0.00000154, -0.00000573, -0.00000140, -0.00000105, 0.00000031, -0.00000515, -0.00000259, -0.00001195, -0.00003321, -0.00000247, 0.00002866, 0.00003767, 0.00001472, 0.00004646, 0.00000847, -0.00001233, -0.00003147, 0.00000783, -0.00001334, 0.00000124, -0.00000950, 0.00000398, -0.00000892, 0.00000863, -0.00000506, 0.00000331, -0.00001050, 0.00000031, 0.00000160, 0.00000015, 0.00000118, -0.00000551, 0.00000485, 0.00000653, -0.00000104, -0.00000986, 0.00000134, 0.00000857, -0.00001491, -0.00000266, 0.00000596, 0.00000296, -0.00001023, -0.00000287, 0.00000180, -0.00000341, -0.00000463, 0.00000318, 0.00000314, -0.00000304, 0.00000097, 0.00000875, -0.00000145, -0.00000913, 0.00000962, 0.00001185, -0.00000845, },
		{ -0.00000769, 0.00000497, 0.00000299, -0.00000563, 0.00000446, -0.00000572, 0.00000259, -0.00000203, -0.00000447, 0.00000807, 0.00000548, -0.00000087, -0.00000295, -0.00000576, 0.00000701, 0.00001395, -0.00000680, 0.00001095, -0.00001453, 0.00000198, -0.00000365, -0.00000488, -0.00000279, -0.00000036, 0.00000210, -0.00001240, -0.00000255, 0.00000738, -0.00001332, 0.00001432, -0.00001850, 0.00000703, 0.00000832, -0.00000409, 0.00000508, -0.00000450, 0.00000552, -0.00000757, -0.00000327, 0.00000156, -0.00000189, -0.00000357, -0.00000294, -0.00000015, -0.00000264, -0.00000194, -0.00000793, -0.00000773, -0.00001414, -0.00000694, -0.00004258, -0.00009771, 0.00013927, 0.00013496, 0.00013853, -0.00010091, -0.00003985, -0.00001756, -0.00001541, 0.00000078, -0.00000742, -0.00000205, -0.00000621, -0.00000096, -0.00000245, -0.00000354, 0.00000327, 0.00000148, -0.00000635, 0.00000085, 0.00000166, -0.00000007, -0.00000516, -0.00000328, 0.00000035, -0.00000123, -0.00000201, -0.00000579, -0.00000360, -0.00003605, -0.00000535, 0.00001693, 0.00004371, 0.00000987, 0.00004443, 0.00001553, -0.00000290, -0.00003455, 0.00000250, -0.00000815, -0.00000273, -0.00000675, 0.00000103, -0.00000627, 0.00000420, -0.00000004, -0.00000050, -0.00000516, -0.00000594, 0.00000558, -0.00000271, 0.00000376, -0.00000630, 0.00000292, 0.00000622, 0.00000205, -0.00000900, -0.00000280, 0.00001055, -0.00001130, -0.00000751, 0.00000634, 0.00000382, -0.00000689, -0.00000701, 0.00000318, -0.00000283, -0.00000456, 0.00000043, 0.00000563, -0.00000305, -0.00000055, 0.00000689, 0.00000330, -0.00001022, 0.00000506, 0.00001426, -0.00000484, },
		{ -0.00000871, 0.00000195, 0.00000626, -0.00000668, 0.00000386, -0.00000378, -0.00000052, 0.00000141, -0.00000678, 0.00000704, 0.00000610, 0.00000066, -0.00000324, -0.00000414, 0.00000029, 0.00001825, -0.00000613, 0.00000886, -0.00000766, -0.00000625, 0.00000300, -0.00000928, 0.00000006, -0.00000382, 0.00000532, -0.00001155, -0.00000652, 0.00000922, -0.00001222, 0.00000942, -0.00000824, -0.00000624, 0.00001778, -0.00001013, 0.00001029, -0.00000844, 0.00000843, -0.00000848, -0.00000175, -0.00000245, 0.00000357, -0.00000857, 0.00000196, -0.00000633, 0.00000426, -0.00000906, 0.00000212, -0.00001822, -0.00000042, -0.00002461, -0.00000444, -0.00013072, 0.00009531, 0.00015121, 0.00016003, -0.00005845, -0.00007720, -0.00000179, -0.00003082, 0.00000923, -0.00001354, 0.00000258, -0.00001033, 0.00000173, -0.00000496, -0.00000088, -0.00000135, 0.00000574, -0.00000749, 0.00000029, 0.00000117, 0.00000082, -0.00000380, -0.00000527, 0.00000142, -0.00000258, 0.00000079, -0.00000865, 0.00000288, -0.00003557, -0.00000951, 0.00000831, 0.00004350, 0.00001205, 0.00003628, 0.00002445, 0.00000234, -0.00002881, -0.00000965, 0.00000080, -0.00000921, -0.00000167, -0.00000426, -0.00000113, -0.00000252, 0.00000634, -0.00000525, 0.00000078, -0.00001161, 0.00000824, -0.00000436, 0.00000517, -0.00000616, 0.00000141, 0.00000467, 0.00000509, -0.00000763, -0.00000553, 0.00000941, -0.00000520, -0.00001283, 0.00000681, 0.00000332, -0.00000238, -0.00001062, 0.00000347, -0.00000210, -0.00000366, -0.00000271, 0.00000724, -0.00000200, -0.00000180, 0.00000450, 0.00000653, -0.00000859, -0.00000038, 0.00001545, -0.00000100, },
		{ -0.00000804, -0.00000182, 0.00000860, -0.00000598, 0.00000165, -0.00000075, -0.00000385, 0.00000397, -0.00000748, 0.00000482, 0.00000704, 0.00000176, -0.00000313, -0.00000278, -0.00000401, 0.00001740, -0.00000021, 0.00000230, 0.00000204, -0.00001437, 0.00000753, -0.00001102, 0.00000061, -0.00000522, 0.00000595, -0.00000829, -0.00001078, 0.00000914, -0.00000830, 0.00000183, 0.00000282, -0.00001697, 0.00002268, -0.00001161, 0.00001109, -0.00000815, 0.00000765, -0.00000676, -0.00000171, -0.00000516, 0.00000699, -0.00001090, 0.00000420, -0.00000971, 0.00000795, -0.00001290, 0.00000821, -0.00002358, 0.00000687, -0.00003364, 0.00001460, -0.00012579, 0.00002413, 0.00016870, 0.00017132, -0.00000171, -0.00011254, 0.00000791, -0.00004105, 0.00001240, -0.00001573, 0.00000444, -0.00001238, 0.00000292, -0.00000659, 0.00000121, -0.00000478, 0.00000769, -0.00000639, -0.00000136, 0.00000120, 0.00000090, -0.00000195, -0.00000674, 0.00000151, -0.00000297, 0.00000226, -0.00000974, 0.00000571, -0.00003034, -0.00001655, 0.00000465, 0.00003619, 0.00002057, 0.00002377, 0.00003439, 0.00000332, -0.00001632, -0.00002431, 0.00000911, -0.00001458, 0.00000269, -0.00000910, 0.00000369, -0.00000850, 0.00001104, -0.00000815, 0.00000467, -0.00001440, 0.00000796, -0.00000377, 0.00000486, -0.00000492, 0.00000007, 0.00000271, 0.00000717, -0.00000552, -0.00000695, 0.00000616, 0.00000147, -0.00001672, 0.00000591, 0.00000300, 0.00000124, -0.00001185, 0.00000160, -0.00000051, -0.00000301, -0.00000470, 0.00000661, 0.00000081, -0.00000343, 0.00000278, 0.00000721, -0.00000436, -0.00000591, 0.00001531, 0.00000257, },
		{ -0.00000638, -0.00000508, 0.00000895, -0.00000321, -0.00000183, 0.00000258, -0.00000642, 0.00000491, -0.00000637, 0.00000166, 0.00000808, 0.00000250, -0.00000249, -0.00000227, -0.00000534, 0.00001227, 0.00000827, -0.00000473, 0.00000978, -0.00001783, 0.00000658, -0.00000802, -0.00000242, -0.00000356, 0.00000337, -0.00000274, -0.00001492, 0.00000748, -0.00000317, -0.00000550, 0.00001090, -0.00002173, 0.00002076, -0.00000722, 0.00000677, -0.00000343, 0.00000329, -0.00000246, -0.00000363, -0.00000538, 0.00000676, -0.00000895, 0.00000213, -0.00000846, 0.00000641, -0.00001110, 0.00000739, -0.00002040, 0.00000386, -0.00002928, 0.00001120, -0.00009006, -0.00005489, 0.00017070, 0.00018190, 0.00005775, -0.00013364, 0.00000344, -0.00004079, 0.00000672, -0.00001171, 0.00000179, -0.00001080, 0.00000110, -0.00000578, 0.00000097, -0.00000533, 0.00000625, -0.00000281, -0.00000398, 0.00000171, 0.00000037, -0.00000018, -0.00000709, 0.00000028, -0.00000214, 0.00000200, -0.00000842, 0.00000439, -0.00002062, -0.00002590, 0.00000505, 0.00002407, 0.00003179, 0.00001103, 0.00004225, 0.00000306, -0.00000243, -0.00003504, 0.00001158, -0.00001506, 0.00000328, -0.00001067, 0.00000538, -0.00001090, 0.00001151, -0.00000708, 0.00000473, -0.00001300, 0.00000412, -0.00000079, 0.00000278, -0.00000252, -0.00000154, 0.00000139, 0.00000736, -0.00000229, -0.00000775, 0.00000253, 0.00000644, -0.00001729, 0.00000236, 0.00000425, 0.00000243, -0.00000973, -0.00000262, 0.00000203, -0.00000322, -0.00000468, 0.00000344, 0.00000500, -0.00000518, 0.00000194, 0.00000564, 0.00000110, -0.00001008, 0.00001333, 0.00000595, },
		{ -0.00000470, -0.00000667, 0.00000683, 0.00000128, -0.00000570, 0.00000522, -0.00000751, 0.00000405, -0.00000388, -0.00000172, 0.00000862, 0.00000333, -0.00000166, -0.00000236, -0.00000460, 0.00000544, 0.00001541, -0.00000803, 0.00001178, -0.00001419, -0.00000045, -0.00000094, -0.00000792, 0.00000001, -0.00000106, 0.00000340, -0.00001752, 0.00000405, 0.00000203, -0.00001064, 0.00001409, -0.00001972, 0.00001269, 0.00000152, -0.00000076, 0.00000359, -0.00000254, 0.00000274, -0.00000648, -0.00000357, 0.00000331, -0.00000346, -0.00000326, -0.00000362, 0.00000074, -0.00000473, 0.00000081, -0.00001031, -0.00000706, -0.00001512, -0.00000676, -0.00004202, -0.00011719, 0.00014266, 0.00019763, 0.00011121, -0.00013231, -0.00001836, -0.00002935, -0.00000744, -0.00000240, -0.00000465, -0.00000595, -0.00000361, -0.00000244, -0.00000185, -0.00000294, 0.00000204, 0.00000208, -0.00000655, 0.00000204, -0.00000014, 0.00000090, -0.00000609, -0.00000207, -0.00000043, 0.00000039, -0.00000518, -0.00000003, -0.00000862, -0.00003492, 0.00000635, 0.00001161, 0.00004044, 0.00000308, 0.00004418, 0.00000567, 0.00000703, -0.00003650, 0.00000521, -0.00000917, -0.00000093, -0.00000795, 0.00000282, -0.00000865, 0.00000711, -0.00000191, 0.00000107, -0.00000796, -0.00000230, 0.00000360, -0.00000032, 0.00000061, -0.00000356, 0.00000131, 0.00000536, 0.00000197, -0.00000842, -0.00000013, 0.00000828, -0.00001388, -0.00000385, 0.00000715, 0.00000114, -0.00000496, -0.00000794, 0.00000458, -0.00000387, -0.00000315, -0.00000109, 0.00000897, -0.00000587, 0.00000117, 0.00000324, 0.00000566, -0.00001132, 0.00000906, 0.00000943, },
		{ -0.00000348, -0.00000628, 0.00000273, 0.00000632, -0.00000868, 0.00000621, -0.00000670, 0.00000165, -0.00000082, -0.00000442, 0.00000806, 0.00000463, -0.00000102, -0.00000250, -0.00000324, -0.00000033, 0.00001816, -0.00000545, 0.00000733, -0.00000475, -0.00001059, 0.00000668, -0.00001262, 0.00000267, -0.00000463, 0.00000745, -0.00001677, -0.00000157, 0.00000681, -0.00001301, 0.00001264, -0.00001275, 0.00000161, 0.00001101, -0.00000792, 0.00000947, -0.00000682, 0.00000639, -0.00000836, -0.00000149, -0.00000122, 0.00000287, -0.00000894, 0.00000164, -0.00000560, 0.00000245, -0.00000721, 0.00000127, -0.00001905, -0.00000007, -0.00002573, -0.00000251, -0.00014396, 0.00008112, 0.00021519, 0.00015661, -0.00010740, -0.00005363, -0.00001172, -0.00002527, 0.00000797, -0.00001171, -0.00000021, -0.00000926, 0.00000177, -0.00000574, 0.00000067, -0.00000284, 0.00000618, -0.00000760, 0.00000136, -0.00000003, 0.00000091, -0.00000394, -0.00000482, 0.00000132, -0.00000160, -0.00000132, -0.00000537, 0.00000235, -0.00004003, 0.00000470, 0.00000338, 0.00004205, 0.00000342, 0.00003797, 0.00001356, 0.00000893, -0.00002742, -0.00000872, 0.00000090, -0.00000799, -0.00000249, -0.00000264, -0.00000315, -0.00000034, 0.00000525, -0.00000414, -0.00000154, -0.00000900, 0.00000748, -0.00000306, 0.00000355, -0.00000552, 0.00000217, 0.00000193, 0.00000631, -0.00000869, -0.00000153, 0.00000713, -0.00000761, -0.00001110, 0.00001016, -0.00000095, 0.00000021, -0.00001201, 0.00000539, -0.00000373, -0.00000157, -0.00000489, 0.00001062, -0.00000413, -0.00000060, 0.00000162, 0.00000756, -0.00000883, 0.00000272, 0.00001289, },
		{ -0.00000237, -0.00000463, -0.00000189, 0.00001021, -0.00000941, 0.00000489, -0.00000405, -0.00000166, 0.00000196, -0.00000583, 0.00000622, 0.00000636, -0.00000060, -0.00000238, -0.00000230, -0.00000346, 0.00001579, 0.00000199, -0.00000072, 0.00000604, -0.00001867, 0.00001021, -0.00001295, 0.00000166, -0.00000491, 0.00000737, -0.00001173, -0.00000911, 0.00001066, -0.00001286, 0.00000814, -0.00000394, -0.00000853, 0.00001744, -0.00001134, 0.00001132, -0.00000724, 0.00000669, -0.00000775, -0.00000089, -0.00000443, 0.00000701, -0.00001167, 0.00000394, -0.00000898, 0.00000642, -0.00001193, 0.00000851, -0.00002506, 0.00000734, -0.00003430, 0.00001501, -0.00013069, -0.00000220, 0.00022189, 0.00019768, -0.00006397, -0.00009298, 0.00000336, -0.00003960, 0.00001379, -0.00001534, 0.00000325, -0.00001309, 0.00000437, -0.00000831, 0.00000303, -0.00000586, 0.00000748, -0.00000606, -0.00000078, 0.00000091, 0.00000000, -0.00000134, -0.00000701, 0.00000214, -0.00000289, 0.00000166, -0.00000929, 0.00000926, -0.00003855, -0.00000225, 0.00000173, 0.00003515, 0.00001204, 0.00002490, 0.00002558, 0.00000460, -0.00001149, -0.00002468, 0.00000998, -0.00001381, 0.00000236, -0.00000795, 0.00000247, -0.00000744, 0.00001102, -0.00000782, 0.00000330, -0.00001330, 0.00000881, -0.00000398, 0.00000514, -0.00000652, 0.00000294, -0.00000136, 0.00000917, -0.00000767, -0.00000240, 0.00000450, -0.00000090, -0.00001671, 0.00001095, -0.00000148, 0.00000313, -0.00001259, 0.00000308, -0.00000174, -0.00000137, -0.00000617, 0.00000863, 0.00000046, -0.00000361, 0.00000152, 0.00000630, -0.00000332, -0.00000431, 0.00001547, },
		{ -0.00000045, -0.00000310, -0.00000521, 0.00001137, -0.00000710, 0.00000123, -0.00000011, -0.00000506, 0.00000380, -0.00000582, 0.00000351, 0.00000795, 0.00000001, -0.00000216, -0.00000197, -0.00000404, 0.00001001, 0.00001072, -0.00000762, 0.00001284, -0.00002000, 0.00000661, -0.00000739, -0.00000369, -0.00000150, 0.00000323, -0.00000348, -0.00001690, 0.00001253, -0.00001044, 0.00000230, 0.00000392, -0.00001487, 0.00001867, -0.00000961, 0.00000838, -0.00000359, 0.00000372, -0.00000459, -0.00000233, -0.00000509, 0.00000736, -0.00000992, 0.00000187, -0.00000785, 0.00000539, -0.00001112, 0.00000858, -0.00002203, 0.00000387, -0.00002895, 0.00000969, -0.00008882, -0.00008394, 0.00020167, 0.00023878, -0.00000898, -0.00012563, 0.00000745, -0.00004453, 0.00001105, -0.00001299, 0.00000243, -0.00001319, 0.00000355, -0.00000783, 0.00000244, -0.00000559, 0.00000529, -0.00000211, -0.00000383, 0.00000219, -0.00000118, 0.00000080, -0.00000778, 0.00000144, -0.00000283, 0.00000279, -0.00001034, 0.00001061, -0.00003009, -0.00001406, 0.00000561, 0.00002214, 0.00002506, 0.00000974, 0.00003719, -0.00000099, 0.00000433, -0.00003542, 0.00001254, -0.00001447, 0.00000344, -0.00001016, 0.00000517, -0.00001104, 0.00001253, -0.00000759, 0.00000447, -0.00001353, 0.00000654, -0.00000238, 0.00000472, -0.00000577, 0.00000247, -0.00000289, 0.00000918, -0.00000459, -0.00000375, 0.00000231, 0.00000372, -0.00001837, 0.00000776, 0.00000115, 0.00000236, -0.00000898, -0.00000228, 0.00000195, -0.00000286, -0.00000459, 0.00000344, 0.00000666, -0.00000677, 0.00000229, 0.00000305, 0.00000309, -0.00000983, 0.00001581, },
		{ 0.00000308, -0.00000287, -0.00000600, 0.00000918, -0.00000198, -0.00000387, 0.00000402, -0.00000765, 0.00000432, -0.00000466, 0.00000068, 0.00000867, 0.00000134, -0.00000223, -0.00000185, -0.00000328, 0.00000367, 0.00001673, -0.00000926, 0.00001237, -0.00001318, -0.00000343, 0.00000206, -0.00001105, 0.00000334, -0.00000247, 0.00000488, -0.00002203, 0.00001083, -0.00000570, -0.00000384, 0.00000942, -0.00001661, 0.00001499, -0.00000388, 0.00000233, 0.00000212, -0.00000060, -0.00000045, -0.00000480, -0.00000375, 0.00000454, -0.00000473, -0.00000323, -0.00000361, 0.00000080, -0.00000613, 0.00000302, -0.00001224, -0.00000701, -0.00001482, -0.00000828, -0.00003968, -0.00013908, 0.00014490, 0.00027821, 0.00005336, -0.00014381, -0.00000360, -0.00003853, -0.00000026, -0.00000540, -0.00000215, -0.00000976, -0.00000060, -0.00000439, -0.00000096, -0.00000249, 0.00000075, 0.00000267, -0.00000641, 0.00000280, -0.00000172, 0.00000165, -0.00000672, -0.00000078, -0.00000149, 0.00000200, -0.00000839, 0.00000697, -0.00001682, -0.00002749, 0.00001115, 0.00000813, 0.00003642, -0.00000128, 0.00004286, -0.00000208, 0.00001346, -0.00003565, 0.00000588, -0.00000877, -0.00000014, -0.00000824, 0.00000370, -0.00000990, 0.00000899, -0.00000315, 0.00000187, -0.00000990, 0.00000125, 0.00000120, 0.00000252, -0.00000322, 0.00000030, -0.00000197, 0.00000606, 0.00000043, -0.00000592, 0.00000164, 0.00000501, -0.00001547, 0.00000072, 0.00000651, -0.00000137, -0.00000267, -0.00000872, 0.00000567, -0.00000485, -0.00000154, -0.00000272, 0.00001188, -0.00000806, 0.00000237, -0.00000010, 0.00000772, -0.00001170, 0.00001279, },
		{ 0.00000806, -0.00000413, -0.00000429, 0.00000441, 0.00000446, -0.00000874, 0.00000692, -0.00000857, 0.00000337, -0.00000280, -0.00000159, 0.00000806, 0.00000354, -0.00000268, -0.00000158, -0.00000246, -0.00000079, 0.00001759, -0.00000446, 0.00000528, -0.00000106, -0.00001537, 0.00001052, -0.00001605, 0.00000586, -0.00000605, 0.00000966, -0.00002171, 0.00000445, 0.00000122, -0.00000969, 0.00001238, -0.00001476, 0.00000857, 0.00000308, -0.00000378, 0.00000690, -0.00000363, 0.00000251, -0.00000653, -0.00000213, 0.00000071, 0.00000108, -0.00000820, 0.00000048, -0.00000395, -0.00000065, -0.00000377, -0.00000141, -0.00001800, -0.00000139, -0.00002474, -0.00000381, -0.00015256, 0.00005625, 0.00030478, 0.00012348, -0.00014462, -0.00002822, -0.00002527, -0.00001600, 0.00000360, -0.00000762, -0.00000500, -0.00000611, 0.00000012, -0.00000528, 0.00000130, -0.00000371, 0.00000598, -0.00000692, 0.00000179, -0.00000100, 0.00000091, -0.00000411, -0.00000382, 0.00000041, 0.00000005, -0.00000456, 0.00000056, -0.00000263, -0.00003798, 0.00001356, -0.00000150, 0.00004067, -0.00000349, 0.00003916, 0.00000445, 0.00001308, -0.00002496, -0.00000800, 0.00000071, -0.00000628, -0.00000372, -0.00000072, -0.00000527, 0.00000214, 0.00000349, -0.00000253, -0.00000443, -0.00000502, 0.00000507, -0.00000026, 0.00000028, -0.00000297, 0.00000076, 0.00000099, 0.00000601, -0.00000811, 0.00000218, 0.00000359, -0.00000948, -0.00000797, 0.00001221, -0.00000538, 0.00000317, -0.00001315, 0.00000704, -0.00000538, 0.00000070, -0.00000701, 0.00001344, -0.00000576, 0.00000043, -0.00000134, 0.00000871, -0.00000903, 0.00000654, },
		{ 0.00001308, -0.00000579, -0.00000146, -0.00000095, 0.00000999, -0.00001143, 0.00000731, -0.00000727, 0.00000101, -0.00000067, -0.00000302, 0.00000630, 0.00000609, -0.00000306, -0.00000127, -0.00000212, -0.00000255, 0.00001361, 0.00000451, -0.00000424, 0.00001056, -0.00002300, 0.00001275, -0.00001484, 0.00000316, -0.00000496, 0.00000867, -0.00001509, -0.00000593, 0.00000908, -0.00001443, 0.00001301, -0.00001082, 0.00000184, 0.00000859, -0.00000736, 0.00000851, -0.00000366, 0.00000298, -0.00000633, -0.00000172, -0.00000198, 0.00000479, -0.00001016, 0.00000157, -0.00000584, 0.00000196, -0.00000766, 0.00000516, -0.00002269, 0.00000373, -0.00002983, 0.00000849, -0.00012635, -0.00004391, 0.00030076, 0.00020335, -0.00012814, -0.00006087, -0.00001142, -0.00003015, 0.00000917, -0.00001055, -0.00000166, -0.00001039, 0.00000319, -0.00000806, 0.00000332, -0.00000575, 0.00000613, -0.00000462, -0.00000094, 0.00000086, -0.00000094, -0.00000090, -0.00000655, 0.00000184, -0.00000190, -0.00000054, -0.00000577, 0.00000846, -0.00004169, 0.00000954, -0.00000368, 0.00003563, 0.00000383, 0.00002687, 0.00001719, 0.00000557, -0.00000808, -0.00002306, 0.00000872, -0.00001104, 0.00000031, -0.00000532, -0.00000007, -0.00000481, 0.00000913, -0.00000587, 0.00000017, -0.00000978, 0.00000727, -0.00000214, 0.00000335, -0.00000594, 0.00000366, -0.00000386, 0.00001000, -0.00000881, 0.00000255, 0.00000145, -0.00000321, -0.00001505, 0.00001505, -0.00000650, 0.00000541, -0.00001306, 0.00000440, -0.00000312, 0.00000045, -0.00000752, 0.00001019, 0.00000023, -0.00000345, -0.00000032, 0.00000601, -0.00000287, -0.00000125, },
		{ 0.00001601, -0.00000608, 0.00000046, -0.00000460, 0.00001252, -0.00001053, 0.00000460, -0.00000380, -0.00000235, 0.00000142, -0.00000372, 0.00000407, 0.00000807, -0.00000257, -0.00000142, -0.00000201, -0.00000238, 0.00000729, 0.00001330, -0.00001073, 0.00001599, -0.00002182, 0.00000630, -0.00000667, -0.00000462, 0.00000045, 0.00000281, -0.00000428, -0.00001730, 0.00001522, -0.00001650, 0.00001107, -0.00000572, -0.00000383, 0.00001147, -0.00000767, 0.00000674, -0.00000099, 0.00000140, -0.00000445, -0.00000271, -0.00000278, 0.00000538, -0.00000832, -0.00000088, -0.00000433, 0.00000096, -0.00000739, 0.00000571, -0.00001938, -0.00000076, -0.00002273, 0.00000056, -0.00007805, -0.00012686, 0.00025086, 0.00029001, -0.00009325, -0.00009567, -0.00000301, -0.00003818, 0.00000835, -0.00000925, -0.00000115, -0.00001191, 0.00000331, -0.00000788, 0.00000231, -0.00000454, 0.00000309, -0.00000030, -0.00000431, 0.00000282, -0.00000276, 0.00000159, -0.00000782, 0.00000189, -0.00000285, 0.00000218, -0.00000970, 0.00001382, -0.00003712, -0.00000118, 0.00000103, 0.00002322, 0.00001708, 0.00001089, 0.00003084, -0.00000301, 0.00000747, -0.00003222, 0.00001025, -0.00001108, 0.00000123, -0.00000756, 0.00000297, -0.00000893, 0.00001113, -0.00000602, 0.00000200, -0.00001139, 0.00000664, -0.00000211, 0.00000485, -0.00000726, 0.00000496, -0.00000636, 0.00001060, -0.00000676, 0.00000139, 0.00000061, 0.00000070, -0.00001788, 0.00001285, -0.00000296, 0.00000276, -0.00000805, -0.00000185, 0.00000148, -0.00000221, -0.00000449, 0.00000332, 0.00000790, -0.00000750, 0.00000167, 0.00000140, 0.00000412, -0.00000782, },
		{ 0.00001510, -0.00000358, -0.00000008, -0.00000509, 0.00001112, -0.00000596, -0.00000061, 0.00000093, -0.00000591, 0.00000307, -0.00000392, 0.00000214, 0.00000864, -0.00000066, -0.00000241, -0.00000160, -0.00000172, 0.00000168, 0.00001792, -0.00001045, 0.00001276, -0.00001165, -0.00000667, 0.00000500, -0.00001381, 0.00000668, -0.00000417, 0.00000617, -0.00002513, 0.00001635, -0.00001413, 0.00000598, 0.00000043, -0.00000842, 0.00001226, -0.00000580, 0.00000328, 0.00000237, -0.00000034, -0.00000242, -0.00000396, -0.00000253, 0.00000389, -0.00000427, -0.00000487, -0.00000150, -0.00000161, -0.00000493, 0.00000252, -0.00001141, -0.00000996, -0.00001058, -0.00001473, -0.00003177, -0.00016813, 0.00015304, 0.00036987, -0.00003489, -0.00012911, -0.00000273, -0.00003934, 0.00000174, -0.00000496, -0.00000257, -0.00001113, 0.00000085, -0.00000527, -0.00000096, -0.00000130, -0.00000129, 0.00000387, -0.00000644, 0.00000343, -0.00000324, 0.00000225, -0.00000692, 0.00000022, -0.00000237, 0.00000283, -0.00001020, 0.00001299, -0.00002563, -0.00001575, 0.00000891, 0.00000845, 0.00003011, -0.00000214, 0.00003910, -0.00000632, 0.00001535, -0.00003109, 0.00000353, -0.00000595, -0.00000133, -0.00000676, 0.00000281, -0.00000911, 0.00000885, -0.00000277, 0.00000093, -0.00000986, 0.00000340, -0.00000031, 0.00000449, -0.00000631, 0.00000378, -0.00000558, 0.00000741, -0.00000198, -0.00000160, 0.00000191, 0.00000126, -0.00001588, 0.00000579, 0.00000442, -0.00000332, -0.00000041, -0.00000907, 0.00000615, -0.00000540, -0.00000012, -0.00000408, 0.00001389, -0.00000910, 0.00000251, -0.00000252, 0.00000886, -0.00001059, },
		{ 0.00001014, 0.00000173, -0.00000314, -0.00000261, 0.00000662, 0.00000084, -0.00000660, 0.00000536, -0.00000854, 0.00000380, -0.00000372, 0.00000087, 0.00000761, 0.00000250, -0.00000399, -0.00000080, -0.00000158, -0.00000130, 0.00001676, -0.00000338, 0.00000300, 0.00000296, -0.00002004, 0.00001398, -0.00001900, 0.00000908, -0.00000768, 0.00001156, -0.00002561, 0.00001039, -0.00000660, -0.00000223, 0.00000752, -0.00001236, 0.00001215, -0.00000355, 0.00000033, 0.00000418, -0.00000035, -0.00000182, -0.00000411, -0.00000271, 0.00000229, -0.00000067, -0.00000739, -0.00000040, -0.00000268, -0.00000350, -0.00000041, -0.00000427, -0.00001669, -0.00000259, -0.00002380, -0.00000592, -0.00015941, 0.00002453, 0.00041871, 0.00005429, -0.00015918, -0.00001015, -0.00003621, -0.00000723, -0.00000099, -0.00000354, -0.00000987, -0.00000237, -0.00000221, -0.00000431, 0.00000152, -0.00000435, 0.00000551, -0.00000571, 0.00000167, -0.00000171, 0.00000071, -0.00000401, -0.00000274, -0.00000077, 0.00000161, -0.00000765, 0.00000750, -0.00001076, -0.00002956, 0.00001500, -0.00000311, 0.00003716, -0.00000705, 0.00003801, -0.00000126, 0.00001358, -0.00002029, -0.00000870, 0.00000143, -0.00000520, -0.00000434, 0.00000047, -0.00000642, 0.00000378, 0.00000213, -0.00000146, -0.00000670, -0.00000097, 0.00000217, 0.00000294, -0.00000355, 0.00000054, -0.00000218, 0.00000171, 0.00000397, -0.00000529, 0.00000453, -0.00000051, -0.00001074, -0.00000356, 0.00001250, -0.00000923, 0.00000594, -0.00001362, 0.00000796, -0.00000657, 0.00000268, -0.00000862, 0.00001522, -0.00000638, 0.00000076, -0.00000384, 0.00000946, -0.00000849, },
		{ 0.00000281, 0.00000822, -0.00000720, 0.00000106, 0.00000120, 0.00000740, -0.00001105, 0.00000772, -0.00000916, 0.00000308, -0.00000295, 0.00000008, 0.00000560, 0.00000592, -0.00000523, -0.00000016, -0.00000195, -0.00000161, 0.00001119, 0.00000700, -0.00000777, 0.00001501, -0.00002681, 0.00001467, -0.00001622, 0.00000474, -0.00000514, 0.00000966, -0.00001787, -0.00000181, 0.00000438, -0.00001185, 0.00001433, -0.00001531, 0.00001153, -0.00000177, -0.00000103, 0.00000357, 0.00000183, -0.00000288, -0.00000282, -0.00000397, 0.00000184, 0.00000072, -0.00000667, -0.00000256, -0.00000078, -0.00000481, -0.00000064, -0.00000148, -0.00001666, -0.00000369, -0.00002055, -0.00000478, -0.00011281, -0.00010131, 0.00040967, 0.00017691, -0.00018125, -0.00002467, -0.00003194, -0.00001509, -0.00000017, -0.00000227, -0.00000960, -0.00000465, -0.00000060, -0.00000573, 0.00000189, -0.00000429, 0.00000354, -0.00000190, -0.00000210, 0.00000133, -0.00000223, -0.00000022, -0.00000585, 0.00000104, -0.00000057, -0.00000350, 0.00000012, 0.00000319, -0.00003819, 0.00001541, -0.00000785, 0.00003541, -0.00000249, 0.00002799, 0.00001054, 0.00000529, -0.00000511, -0.00002052, 0.00000634, -0.00000717, -0.00000263, -0.00000201, -0.00000308, -0.00000147, 0.00000615, -0.00000304, -0.00000386, -0.00000467, 0.00000394, 0.00000129, -0.00000017, -0.00000333, 0.00000197, -0.00000408, 0.00000871, -0.00000773, 0.00000658, -0.00000230, -0.00000546, -0.00001157, 0.00001743, -0.00001119, 0.00000753, -0.00001285, 0.00000516, -0.00000426, 0.00000213, -0.00000847, 0.00001099, 0.00000042, -0.00000326, -0.00000244, 0.00000607, -0.00000265, },
		{ -0.00000398, 0.00001335, -0.00000987, 0.00000343, -0.00000258, 0.00001135, -0.00001213, 0.00000687, -0.00000721, 0.00000074, -0.00000146, -0.00000071, 0.00000364, 0.00000827, -0.00000496, -0.00000052, -0.00000215, -0.00000066, 0.00000443, 0.00001566, -0.00001363, 0.00001875, -0.00002290, 0.00000541, -0.00000568, -0.00000525, 0.00000219, 0.00000222, -0.00000505, -0.00001603, 0.00001468, -0.00001944, 0.00001832, -0.00001565, 0.00000941, 0.00000026, -0.00000157, 0.00000168, 0.00000464, -0.00000408, -0.00000136, -0.00000543, 0.00000208, 0.00000029, -0.00000352, -0.00000670, 0.00000264, -0.00000748, 0.00000088, -0.00000233, -0.00001110, -0.00001147, -0.00000904, -0.00001827, -0.00005504, -0.00018759, 0.00032669, 0.00032204, -0.00018355, -0.00004862, -0.00002720, -0.00002108, -0.00000254, 0.00000055, -0.00000987, -0.00000591, -0.00000079, -0.00000488, -0.00000033, -0.00000146, -0.00000094, 0.00000320, -0.00000602, 0.00000414, -0.00000483, 0.00000271, -0.00000763, 0.00000190, -0.00000244, 0.00000043, -0.00000628, 0.00001270, -0.00003902, 0.00000884, -0.00000539, 0.00002588, 0.00000873, 0.00001347, 0.00002379, -0.00000335, 0.00000747, -0.00002604, 0.00000520, -0.00000525, -0.00000305, -0.00000315, -0.00000090, -0.00000484, 0.00000751, -0.00000253, -0.00000246, -0.00000666, 0.00000421, 0.00000033, 0.00000269, -0.00000628, 0.00000487, -0.00000765, 0.00001023, -0.00000736, 0.00000639, -0.00000207, -0.00000255, -0.00001549, 0.00001674, -0.00000731, 0.00000330, -0.00000679, -0.00000150, 0.00000080, -0.00000134, -0.00000435, 0.00000304, 0.00000871, -0.00000736, 0.00000003, 0.00000081, 0.00000412, },
		{ -0.00000747, 0.00001491, -0.00000922, 0.00000272, -0.00000314, 0.00001148, -0.00000928, 0.00000288, -0.00000311, -0.00000280, 0.00000061, -0.00000185, 0.00000254, 0.00000865, -0.00000262, -0.00000223, -0.00000156, -0.00000006, -0.00000041, 0.00001882, -0.00001141, 0.00001276, -0.00000960, -0.00001005, 0.00000770, -0.00001591, 0.00000963, -0.00000578, 0.00000699, -0.00002625, 0.00001924, -0.00002111, 0.00001665, -0.00001135, 0.00000416, 0.00000415, -0.00000309, 0.00000078, 0.00000561, -0.00000313, -0.00000168, -0.00000536, 0.00000131, 0.00000007, -0.00000069, -0.00000953, 0.00000407, -0.00000801, 0.00000074, -0.00000291, -0.00000540, -0.00001847, 0.00000050, -0.00002966, -0.00001390, -0.00021027, 0.00017685, 0.00046053, -0.00014686, -0.00008722, -0.00001973, -0.00002769, -0.00000503, 0.00000176, -0.00000829, -0.00000780, -0.00000134, -0.00000335, -0.00000323, 0.00000175, -0.00000501, 0.00000655, -0.00000747, 0.00000459, -0.00000522, 0.00000316, -0.00000695, 0.00000096, -0.00000297, 0.00000261, -0.00000963, 0.00001598, -0.00003203, -0.00000312, 0.00000167, 0.00001251, 0.00002133, 0.00000047, 0.00003254, -0.00000668, 0.00001255, -0.00002270, -0.00000209, -0.00000036, -0.00000488, -0.00000312, -0.00000021, -0.00000589, 0.00000617, -0.00000025, -0.00000225, -0.00000715, 0.00000320, 0.00000012, 0.00000445, -0.00000735, 0.00000535, -0.00000788, 0.00000796, -0.00000400, 0.00000364, 0.00000078, -0.00000273, -0.00001477, 0.00001061, 0.00000126, -0.00000462, 0.00000165, -0.00000890, 0.00000588, -0.00000530, 0.00000087, -0.00000494, 0.00001474, -0.00000871, 0.00000126, -0.00000360, 0.00000869, },
		{ -0.00000648, 0.00001231, -0.00000489, -0.00000123, -0.00000065, 0.00000839, -0.00000359, -0.00000283, 0.00000175, -0.00000646, 0.00000263, -0.00000318, 0.00000241, 0.00000715, 0.00000131, -0.00000474, -0.00000035, -0.00000049, -0.00000198, 0.00001579, -0.00000230, 0.00000067, 0.00000692, -0.00002416, 0.00001662, -0.00002095, 0.00001175, -0.00000900, 0.00001269, -0.00002764, 0.00001480, -0.00001489, 0.00000815, -0.00000177, -0.00000476, 0.00001057, -0.00000662, 0.00000232, 0.00000318, 0.00000116, -0.00000477, -0.00000271, -0.00000177, 0.00000195, -0.00000074, -0.00000813, 0.00000068, -0.00000371, -0.00000386, 0.00000035, -0.00000467, -0.00001789, -0.00000058, -0.00002666, -0.00000439, -0.00017016, -0.00000595, 0.00054964, -0.00005138, -0.00014358, -0.00000768, -0.00003795, -0.00000405, -0.00000199, -0.00000256, -0.00001207, -0.00000036, -0.00000338, -0.00000422, 0.00000244, -0.00000568, 0.00000563, -0.00000471, 0.00000154, -0.00000253, 0.00000058, -0.00000367, -0.00000180, -0.00000185, 0.00000242, -0.00000924, 0.00001330, -0.00001948, -0.00001692, 0.00000931, 0.00000004, 0.00003018, -0.00000627, 0.00003333, -0.00000251, 0.00000967, -0.00001263, -0.00001179, 0.00000407, -0.00000576, -0.00000338, -0.00000000, -0.00000563, 0.00000356, 0.00000221, -0.00000196, -0.00000714, 0.00000177, 0.00000018, 0.00000516, -0.00000650, 0.00000347, -0.00000525, 0.00000305, 0.00000095, -0.00000048, 0.00000507, -0.00000469, -0.00001119, 0.00000171, 0.00001094, -0.00001202, 0.00000805, -0.00001309, 0.00000781, -0.00000690, 0.00000396, -0.00000936, 0.00001555, -0.00000557, -0.00000011, -0.00000529, 0.00000922, },
		{ -0.00000198, 0.00000687, 0.00000167, -0.00000677, 0.00000304, 0.00000409, 0.00000273, -0.00000808, 0.00000555, -0.00000886, 0.00000367, -0.00000404, 0.00000261, 0.00000478, 0.00000541, -0.00000668, 0.00000054, -0.00000148, -0.00000098, 0.00000890, 0.00000908, -0.00001086, 0.00001877, -0.00002947, 0.00001541, -0.00001654, 0.00000585, -0.00000501, 0.00000989, -0.00001936, 0.00000213, -0.00000245, -0.00000513, 0.00001091, -0.00001533, 0.00001784, -0.00001093, 0.00000547, -0.00000170, 0.00000753, -0.00000935, 0.00000151, -0.00000652, 0.00000579, -0.00000391, -0.00000246, -0.00000712, 0.00000466, -0.00001213, 0.00000725, -0.00000953, -0.00000886, -0.00001284, -0.00000900, -0.00002198, -0.00009345, -0.00017032, 0.00054790, 0.00011049, -0.00021061, 0.00000513, -0.00005111, 0.00000071, -0.00001086, 0.00000687, -0.00001834, 0.00000241, -0.00000577, -0.00000217, -0.00000059, -0.00000204, 0.00000021, 0.00000168, -0.00000408, 0.00000226, -0.00000390, 0.00000098, -0.00000534, 0.00000030, 0.00000031, -0.00000590, 0.00000665, -0.00000504, -0.00002850, 0.00001379, -0.00000796, 0.00003236, -0.00000528, 0.00002656, 0.00000706, 0.00000262, -0.00000133, -0.00001847, 0.00000423, -0.00000350, -0.00000525, 0.00000081, -0.00000538, 0.00000137, 0.00000327, -0.00000049, -0.00000748, 0.00000065, 0.00000008, 0.00000512, -0.00000427, 0.00000026, -0.00000135, -0.00000234, 0.00000530, -0.00000397, 0.00000870, -0.00000607, -0.00000756, -0.00000636, 0.00001758, -0.00001476, 0.00000888, -0.00001158, 0.00000494, -0.00000470, 0.00000322, -0.00000863, 0.00001065, 0.00000143, -0.00000351, -0.00000430, 0.00000598, },
		{ 0.00000349, 0.00000118, 0.00000794, -0.00001133, 0.00000542, 0.00000096, 0.00000741, -0.00001092, 0.00000680, -0.00000898, 0.00000299, -0.00000376, 0.00000224, 0.00000288, 0.00000808, -0.00000666, 0.00000007, -0.00000198, 0.00000066, 0.00000191, 0.00001732, -0.00001576, 0.00002050, -0.00002272, 0.00000352, -0.00000402, -0.00000593, 0.00000396, 0.00000125, -0.00000538, -0.00001364, 0.00001094, -0.00001807, 0.00002183, -0.00002313, 0.00002204, -0.00001264, 0.00000716, -0.00000576, 0.00001245, -0.00001206, 0.00000430, -0.00001022, 0.00000904, -0.00000749, 0.00000411, -0.00001518, 0.00001247, -0.00001924, 0.00001313, -0.00001516, 0.00000257, -0.00002799, 0.00001179, -0.00004664, -0.00001967, -0.00026705, 0.00043496, 0.00032138, -0.00026507, 0.00000603, -0.00006010, 0.00000406, -0.00002013, 0.00001539, -0.00002297, 0.00000442, -0.00000869, 0.00000131, -0.00000563, 0.00000378, -0.00000690, 0.00000853, -0.00000925, 0.00000646, -0.00000778, 0.00000466, -0.00000782, 0.00000205, -0.00000231, -0.00000148, -0.00000100, 0.00000743, -0.00003460, 0.00001300, -0.00001013, 0.00002780, 0.00000172, 0.00001580, 0.00001736, -0.00000315, 0.00000567, -0.00001827, -0.00000134, 0.00000187, -0.00000837, 0.00000211, -0.00000549, 0.00000025, 0.00000256, 0.00000204, -0.00000797, -0.00000027, -0.00000004, 0.00000440, -0.00000131, -0.00000309, 0.00000210, -0.00000621, 0.00000726, -0.00000528, 0.00000994, -0.00000500, -0.00000605, -0.00001099, 0.00001863, -0.00001103, 0.00000347, -0.00000487, -0.00000160, 0.00000027, -0.00000055, -0.00000397, 0.00000238, 0.00000937, -0.00000667, -0.00000230, 0.00000099, },
		{ 0.00000720, -0.00000229, 0.00001167, -0.00001277, 0.00000469, 0.00000044, 0.00000923, -0.00001048, 0.00000509, -0.00000669, 0.00000051, -0.00000215, 0.00000081, 0.00000232, 0.00000842, -0.00000417, -0.00000205, -0.00000140, 0.00000133, -0.00000221, 0.00001908, -0.00001170, 0.00001196, -0.00000678, -0.00001369, 0.00001028, -0.00001738, 0.00001211, -0.00000721, 0.00000737, -0.00002530, 0.00001866, -0.00002471, 0.00002555, -0.00002337, 0.00001904, -0.00000818, 0.00000404, -0.00000539, 0.00001215, -0.00000947, 0.00000252, -0.00000972, 0.00000833, -0.00000752, 0.00000673, -0.00001790, 0.00001375, -0.00001915, 0.00001185, -0.00001460, 0.00000739, -0.00003426, 0.00002035, -0.00005567, 0.00001741, -0.00027072, 0.00022669, 0.00053464, -0.00026977, -0.00002327, -0.00005399, -0.00000269, -0.00002195, 0.00001604, -0.00002051, 0.00000153, -0.00000850, 0.00000262, -0.00000882, 0.00000751, -0.00001114, 0.00001145, -0.00001026, 0.00000698, -0.00000831, 0.00000506, -0.00000744, 0.00000192, -0.00000374, 0.00000179, -0.00000669, 0.00001490, -0.00003361, 0.00000675, -0.00000720, 0.00001855, 0.00001122, 0.00000546, 0.00002397, -0.00000385, 0.00000598, -0.00001138, -0.00001029, 0.00000749, -0.00001040, 0.00000234, -0.00000500, -0.00000057, 0.00000120, 0.00000421, -0.00000742, -0.00000189, 0.00000050, 0.00000277, 0.00000198, -0.00000572, 0.00000405, -0.00000760, 0.00000637, -0.00000416, 0.00000845, -0.00000119, -0.00000696, -0.00001179, 0.00001426, -0.00000221, -0.00000555, 0.00000364, -0.00000838, 0.00000503, -0.00000464, 0.00000141, -0.00000532, 0.00001451, -0.00000696, -0.00000135, -0.00000328, },
		{ 0.00000765, -0.00000247, 0.00001198, -0.00001043, 0.00000065, 0.00000235, 0.00000849, -0.00000731, 0.00000121, -0.00000283, -0.00000300, 0.00000028, -0.00000141, 0.00000300, 0.00000673, 0.00000005, -0.00000499, -0.00000015, 0.00000065, -0.00000270, 0.00001457, -0.00000110, -0.00000175, 0.00001071, -0.00002752, 0.00001824, -0.00002162, 0.00001352, -0.00000966, 0.00001276, -0.00002727, 0.00001633, -0.00002149, 0.00001916, -0.00001386, 0.00000736, 0.00000341, -0.00000485, 0.00000070, 0.00000535, -0.00000064, -0.00000467, -0.00000389, 0.00000207, -0.00000177, 0.00000244, -0.00001202, 0.00000535, -0.00000891, 0.00000049, -0.00000408, 0.00000028, -0.00002466, 0.00000814, -0.00003764, 0.00000622, -0.00019222, -0.00002367, 0.00068423, -0.00018667, -0.00009612, -0.00002569, -0.00002564, -0.00001070, 0.00000405, -0.00000759, -0.00000890, -0.00000251, -0.00000134, -0.00000669, 0.00000543, -0.00000890, 0.00000745, -0.00000500, 0.00000227, -0.00000422, 0.00000121, -0.00000357, -0.00000078, -0.00000292, 0.00000237, -0.00000855, 0.00001608, -0.00002589, -0.00000345, -0.00000121, 0.00000745, 0.00001983, -0.00000132, 0.00002501, 0.00000071, 0.00000149, -0.00000200, -0.00001749, 0.00000902, -0.00000842, -0.00000038, -0.00000256, -0.00000246, 0.00000092, 0.00000430, -0.00000457, -0.00000502, 0.00000221, 0.00000016, 0.00000525, -0.00000725, 0.00000432, -0.00000680, 0.00000350, -0.00000165, 0.00000527, 0.00000402, -0.00000881, -0.00001043, 0.00000701, 0.00000797, -0.00001364, 0.00000937, -0.00001154, 0.00000651, -0.00000619, 0.00000431, -0.00000905, 0.00001431, -0.00000316, -0.00000242, -0.00000532, },
		{ 0.00000516, -0.00000007, 0.00000969, -0.00000531, -0.00000523, 0.00000510, 0.00000667, -0.00000292, -0.00000326, 0.00000113, -0.00000632, 0.00000249, -0.00000354, 0.00000400, 0.00000428, 0.00000433, -0.00000716, 0.00000063, -0.00000054, -0.00000087, 0.00000695, 0.00001056, -0.00001319, 0.00002140, -0.00003056, 0.00001460, -0.00001533, 0.00000598, -0.00000411, 0.00000895, -0.00001879, 0.00000450, -0.00000963, 0.00000443, 0.00000294, -0.00000997, 0.00001882, -0.00001641, 0.00000983, -0.00000520, 0.00001140, -0.00001437, 0.00000483, -0.00000771, 0.00000809, -0.00000722, 0.00000057, -0.00001011, 0.00000819, -0.00001736, 0.00001306, -0.00001589, -0.00000253, -0.00002004, 0.00000039, -0.00003775, -0.00007499, -0.00024191, 0.00070808, 0.00000255, -0.00020707, 0.00001881, -0.00006126, 0.00001130, -0.00001836, 0.00001318, -0.00002466, 0.00000805, -0.00001015, 0.00000089, -0.00000272, -0.00000020, -0.00000270, 0.00000514, -0.00000612, 0.00000306, -0.00000550, 0.00000248, -0.00000514, -0.00000012, 0.00000018, -0.00000647, 0.00001169, -0.00001365, -0.00001494, 0.00000538, -0.00000295, 0.00002536, -0.00000364, 0.00002155, 0.00000760, -0.00000317, 0.00000420, -0.00001806, 0.00000361, -0.00000126, -0.00000640, 0.00000205, -0.00000588, 0.00000253, 0.00000165, 0.00000062, -0.00000936, 0.00000469, -0.00000283, 0.00000788, -0.00000743, 0.00000324, -0.00000477, 0.00000014, 0.00000070, 0.00000201, 0.00000861, -0.00000947, -0.00000925, 0.00000007, 0.00001547, -0.00001674, 0.00000911, -0.00000903, 0.00000348, -0.00000409, 0.00000337, -0.00000773, 0.00000891, 0.00000360, -0.00000467, -0.00000533, },
		{ 0.00000138, 0.00000300, 0.00000659, 0.00000061, -0.00001067, 0.00000661, 0.00000544, 0.00000111, -0.00000680, 0.00000388, -0.00000828, 0.00000346, -0.00000458, 0.00000421, 0.00000251, 0.00000703, -0.00000717, -0.00000010, -0.00000111, 0.00000114, 0.00000018, 0.00001785, -0.00001660, 0.00002064, -0.00002077, 0.00000022, -0.00000120, -0.00000716, 0.00000615, -0.00000048, -0.00000457, -0.00001121, 0.00000499, -0.00001246, 0.00002032, -0.00002594, 0.00003115, -0.00002431, 0.00001617, -0.00001364, 0.00002060, -0.00002085, 0.00001112, -0.00001581, 0.00001673, -0.00001646, 0.00001311, -0.00002476, 0.00002338, -0.00003234, 0.00002723, -0.00003067, 0.00001949, -0.00004791, 0.00003689, -0.00008138, 0.00002445, -0.00036142, 0.00057626, 0.00027800, -0.00032295, 0.00005625, -0.00009365, 0.00003166, -0.00004050, 0.00003224, -0.00003791, 0.00001710, -0.00001896, 0.00000963, -0.00001276, 0.00001044, -0.00001409, 0.00001517, -0.00001372, 0.00000964, -0.00001154, 0.00000766, -0.00000882, 0.00000289, -0.00000325, -0.00000220, 0.00000434, -0.00000037, -0.00002458, 0.00001015, -0.00001066, 0.00002687, -0.00000225, 0.00001616, 0.00001302, -0.00000380, 0.00000367, -0.00001065, -0.00000768, 0.00000875, -0.00001332, 0.00000689, -0.00000948, 0.00000497, -0.00000239, 0.00000636, -0.00001309, 0.00000630, -0.00000476, 0.00000884, -0.00000593, 0.00000116, -0.00000241, -0.00000256, 0.00000189, -0.00000023, 0.00001117, -0.00000754, -0.00000969, -0.00000451, 0.00001801, -0.00001334, 0.00000277, -0.00000199, -0.00000251, 0.00000029, -0.00000022, -0.00000306, 0.00000104, 0.00001028, -0.00000596, -0.00000472, },
		{ -0.00000193, 0.00000489, 0.00000434, 0.00000551, -0.00001377, 0.00000548, 0.00000566, 0.00000399, -0.00000863, 0.00000474, -0.00000838, 0.00000273, -0.00000405, 0.00000304, 0.00000222, 0.00000740, -0.00000474, -0.00000239, -0.00000065, 0.00000191, -0.00000320, 0.00001819, -0.00001079, 0.00000981, -0.00000279, -0.00001792, 0.00001311, -0.00001858, 0.00001435, -0.00000866, 0.00000774, -0.00002284, 0.00001470, -0.00002380, 0.00003038, -0.00003277, 0.00003326, -0.00002231, 0.00001402, -0.00001415, 0.00002113, -0.00001868, 0.00000979, -0.00001676, 0.00001812, -0.00001841, 0.00001758, -0.00002956, 0.00002699, -0.00003441, 0.00002794, -0.00003227, 0.00002686, -0.00005721, 0.00004885, -0.00009294, 0.00006292, -0.00035171, 0.00031057, 0.00057616, -0.00038686, 0.00005222, -0.00010009, 0.00003274, -0.00004746, 0.00003688, -0.00003849, 0.00001649, -0.00002064, 0.00001277, -0.00001791, 0.00001612, -0.00001980, 0.00001876, -0.00001518, 0.00001100, -0.00001294, 0.00000855, -0.00000908, 0.00000375, -0.00000548, 0.00000146, -0.00000246, 0.00001013, -0.00002945, 0.00001115, -0.00001428, 0.00002409, 0.00000171, 0.00001096, 0.00001494, 0.00000061, -0.00000263, 0.00000118, -0.00001958, 0.00001630, -0.00001676, 0.00000857, -0.00001055, 0.00000578, -0.00000506, 0.00000969, -0.00001359, 0.00000497, -0.00000398, 0.00000707, -0.00000244, -0.00000176, -0.00000007, -0.00000435, 0.00000196, -0.00000142, 0.00001151, -0.00000321, -0.00001148, -0.00000669, 0.00001589, -0.00000509, -0.00000659, 0.00000584, -0.00000790, 0.00000398, -0.00000370, 0.00000172, -0.00000546, 0.00001354, -0.00000431, -0.00000489, },
		{ -0.00000396, 0.00000480, 0.00000353, 0.00000861, -0.00001382, 0.00000170, 0.00000694, 0.00000590, -0.00000881, 0.00000389, -0.00000683, 0.00000055, -0.00000224, 0.00000076, 0.00000315, 0.00000598, -0.00000085, -0.00000515, 0.00000018, 0.00000142, -0.00000306, 0.00001277, 0.00000059, -0.00000461, 0.00001451, -0.00003028, 0.00001906, -0.00002116, 0.00001439, -0.00000964, 0.00001189, -0.00002450, 0.00001430, -0.00002451, 0.00002815, -0.00002614, 0.00002186, -0.00000819, 0.00000162, -0.00000483, 0.00001115, -0.00000638, -0.00000065, -0.00000853, 0.00000954, -0.00000962, 0.00000981, -0.00001999, 0.00001456, -0.00001933, 0.00001092, -0.00001545, 0.00001262, -0.00003928, 0.00002621, -0.00006023, 0.00003063, -0.00023261, -0.00001599, 0.00080490, -0.00033785, -0.00002185, -0.00006407, 0.00000213, -0.00002904, 0.00001887, -0.00002025, 0.00000127, -0.00001048, 0.00000542, -0.00001302, 0.00001159, -0.00001496, 0.00001196, -0.00000757, 0.00000478, -0.00000761, 0.00000345, -0.00000452, 0.00000096, -0.00000457, 0.00000210, -0.00000578, 0.00001500, -0.00002781, 0.00000732, -0.00001306, 0.00001702, 0.00000754, 0.00000649, 0.00001411, 0.00000736, -0.00000978, 0.00001064, -0.00002512, 0.00001580, -0.00001282, 0.00000436, -0.00000689, 0.00000276, -0.00000390, 0.00000823, -0.00000912, -0.00000040, 0.00000015, 0.00000244, 0.00000278, -0.00000521, 0.00000218, -0.00000560, 0.00000169, -0.00000230, 0.00001045, 0.00000209, -0.00001298, -0.00000798, 0.00001134, 0.00000439, -0.00001441, 0.00001007, -0.00000923, 0.00000432, -0.00000459, 0.00000380, -0.00000782, 0.00001171, 0.00000061, -0.00000605, },
		{ -0.00000519, 0.00000324, 0.00000360, 0.00001016, -0.00001130, -0.00000355, 0.00000803, 0.00000762, -0.00000790, 0.00000191, -0.00000430, -0.00000232, 0.00000004, -0.00000175, 0.00000431, 0.00000408, 0.00000280, -0.00000679, 0.00000021, 0.00000070, -0.00000114, 0.00000529, 0.00001153, -0.00001482, 0.00002283, -0.00003001, 0.00001224, -0.00001252, 0.00000493, -0.00000218, 0.00000666, -0.00001574, 0.00000407, -0.00001505, 0.00001452, -0.00000797, 0.00000013, 0.00001392, -0.00001676, 0.00001035, -0.00000529, 0.00001168, -0.00001609, 0.00000537, -0.00000593, 0.00000703, -0.00000728, 0.00000046, -0.00000948, 0.00000752, -0.00001779, 0.00001369, -0.00001709, -0.00000147, -0.00002095, 0.00000224, -0.00004536, -0.00006642, -0.00030303, 0.00087613, -0.00014058, -0.00016755, 0.00001010, -0.00005640, 0.00001149, -0.00001843, 0.00001308, -0.00002531, 0.00000935, -0.00001126, 0.00000138, -0.00000278, -0.00000032, -0.00000375, 0.00000710, -0.00000707, 0.00000276, -0.00000611, 0.00000360, -0.00000482, -0.00000048, -0.00000088, -0.00000464, 0.00001355, -0.00001985, -0.00000094, -0.00000726, 0.00000631, 0.00001472, 0.00000217, 0.00001300, 0.00001205, -0.00001196, 0.00001166, -0.00001976, 0.00000509, -0.00000110, -0.00000544, 0.00000114, -0.00000404, 0.00000136, 0.00000200, -0.00000039, -0.00000867, 0.00000635, -0.00000366, 0.00000839, -0.00000832, 0.00000408, -0.00000660, 0.00000172, -0.00000348, 0.00000893, 0.00000674, -0.00001250, -0.00000997, 0.00000678, 0.00001160, -0.00001712, 0.00000819, -0.00000531, 0.00000081, -0.00000236, 0.00000248, -0.00000574, 0.00000577, 0.00000700, -0.00000696, },
		{ -0.00000661, 0.00000132, 0.00000351, 0.00001091, -0.00000728, -0.00000867, 0.00000767, 0.00000968, -0.00000624, -0.00000067, -0.00000144, -0.00000511, 0.00000191, -0.00000354, 0.00000469, 0.00000294, 0.00000486, -0.00000626, -0.00000125, 0.00000062, 0.00000057, -0.00000052, 0.00001705, -0.00001591, 0.00001886, -0.00001681, -0.00000470, 0.00000309, -0.00000933, 0.00000919, -0.00000336, -0.00000192, -0.00001011, -0.00000135, -0.00000398, 0.00001399, -0.00002305, 0.00003465, -0.00003212, 0.00002293, -0.00001968, 0.00002687, -0.00002824, 0.00001709, -0.00002033, 0.00002312, -0.00002444, 0.00002125, -0.00003310, 0.00003276, -0.00004377, 0.00003987, -0.00004532, 0.00003589, -0.00006698, 0.00006115, -0.00011638, 0.00006985, -0.00046146, 0.00074408, 0.00018836, -0.00034573, 0.00009046, -0.00011884, 0.00005474, -0.00005829, 0.00004675, -0.00005096, 0.00002894, -0.00002901, 0.00001798, -0.00001998, 0.00001700, -0.00002109, 0.00002186, -0.00001844, 0.00001292, -0.00001549, 0.00001132, -0.00001057, 0.00000454, -0.00000567, -0.00000060, 0.00000769, -0.00000808, -0.00001141, 0.00000124, -0.00000619, 0.00002200, -0.00000219, 0.00001348, 0.00001174, -0.00000632, 0.00000283, -0.00000469, -0.00001233, 0.00001392, -0.00001655, 0.00000996, -0.00001165, 0.00000803, -0.00000603, 0.00000917, -0.00001639, 0.00001147, -0.00000841, 0.00001211, -0.00000959, 0.00000480, -0.00000700, 0.00000187, -0.00000470, 0.00000711, 0.00000995, -0.00000945, -0.00001307, 0.00000331, 0.00001492, -0.00001386, 0.00000092, 0.00000197, -0.00000443, 0.00000118, -0.00000065, -0.00000139, -0.00000120, 0.00001178, -0.00000576, },
		{ -0.00000865, -0.00000023, 0.00000263, 0.00001125, -0.00000263, -0.00001249, 0.00000537, 0.00001181, -0.00000367, -0.00000369, 0.00000136, -0.00000733, 0.00000287, -0.00000416, 0.00000388, 0.00000297, 0.00000507, -0.00000380, -0.00000374, 0.00000113, 0.00000119, -0.00000289, 0.00001575, -0.00000824, 0.00000587, 0.00000273, -0.00002309, 0.00001669, -0.00002002, 0.00001685, -0.00001062, 0.00000888, -0.00002007, 0.00000851, -0.00001856, 0.00002990, -0.00003724, 0.00004381, -0.00003521, 0.00002439, -0.00002351, 0.00003069, -0.00002903, 0.00001867, -0.00002514, 0.00002915, -0.00003100, 0.00003030, -0.00004287, 0.00004193, -0.00005173, 0.00004681, -0.00005372, 0.00005077, -0.00008489, 0.00008314, -0.00013840, 0.00011941, -0.00045055, 0.00043124, 0.00057225, -0.00048081, 0.00012564, -0.00014903, 0.00007238, -0.00007718, 0.00006094, -0.00005947, 0.00003488, -0.00003612, 0.00002591, -0.00002953, 0.00002672, -0.00003018, 0.00002790, -0.00002188, 0.00001645, -0.00001900, 0.00001378, -0.00001226, 0.00000697, -0.00000892, 0.00000296, 0.00000120, 0.00000342, -0.00002047, 0.00000918, -0.00001717, 0.00002685, -0.00000526, 0.00001508, 0.00000720, 0.00000492, -0.00001126, 0.00001297, -0.00002813, 0.00002440, -0.00002244, 0.00001419, -0.00001547, 0.00001173, -0.00001122, 0.00001482, -0.00001927, 0.00001192, -0.00000886, 0.00001172, -0.00000752, 0.00000330, -0.00000593, 0.00000115, -0.00000489, 0.00000437, 0.00001203, -0.00000470, -0.00001615, 0.00000035, 0.00001501, -0.00000665, -0.00000821, 0.00000856, -0.00000784, 0.00000319, -0.00000291, 0.00000214, -0.00000575, 0.00001236, -0.00000143, },
		{ -0.00001047, -0.00000166, 0.00000126, 0.00001086, 0.00000236, -0.00001462, 0.00000164, 0.00001305, 0.00000032, -0.00000712, 0.00000388, -0.00000872, 0.00000278, -0.00000371, 0.00000224, 0.00000367, 0.00000429, -0.00000074, -0.00000583, 0.00000119, 0.00000124, -0.00000252, 0.00000990, 0.00000329, -0.00000839, 0.00001871, -0.00003286, 0.00001962, -0.00002011, 0.00001483, -0.00000935, 0.00001069, -0.00002009, 0.00000896, -0.00002298, 0.00003285, -0.00003572, 0.00003578, -0.00002181, 0.00001127, -0.00001333, 0.00001981, -0.00001550, 0.00000701, -0.00001642, 0.00002027, -0.00002125, 0.00002110, -0.00003180, 0.00002795, -0.00003463, 0.00002724, -0.00003381, 0.00003262, -0.00006212, 0.00005412, -0.00009549, 0.00007090, -0.00029559, 0.00002778, 0.00089298, -0.00048649, 0.00006812, -0.00011738, 0.00004256, -0.00005772, 0.00004172, -0.00004002, 0.00001826, -0.00002444, 0.00001723, -0.00002353, 0.00002119, -0.00002412, 0.00001953, -0.00001295, 0.00000961, -0.00001324, 0.00000802, -0.00000731, 0.00000418, -0.00000770, 0.00000282, -0.00000219, 0.00001063, -0.00002451, 0.00001288, -0.00002281, 0.00002602, -0.00000449, 0.00001510, 0.00000245, 0.00001557, -0.00002221, 0.00002356, -0.00003324, 0.00002319, -0.00001792, 0.00000988, -0.00001205, 0.00000903, -0.00000994, 0.00001313, -0.00001465, 0.00000582, -0.00000371, 0.00000643, -0.00000174, -0.00000078, -0.00000278, -0.00000138, -0.00000298, 0.00000003, 0.00001372, 0.00000016, -0.00001734, -0.00000353, 0.00001376, 0.00000117, -0.00001488, 0.00001055, -0.00000669, 0.00000175, -0.00000251, 0.00000278, -0.00000607, 0.00000828, 0.00000515, },
		{ -0.00001024, -0.00000388, 0.00000031, 0.00000907, 0.00000768, -0.00001520, -0.00000252, 0.00001237, 0.00000589, -0.00001055, 0.00000565, -0.00000903, 0.00000170, -0.00000259, 0.00000044, 0.00000414, 0.00000372, 0.00000136, -0.00000608, -0.00000023, 0.00000173, -0.00000137, 0.00000352, 0.00001240, -0.00001613, 0.00002339, -0.00002822, 0.00000880, -0.00000849, 0.00000295, 0.00000055, 0.00000333, -0.00001040, 0.00000030, -0.00001692, 0.00002249, -0.00001919, 0.00001292, 0.00000417, -0.00001197, 0.00000653, -0.00000131, 0.00000756, -0.00001342, 0.00000205, -0.00000021, 0.00000156, -0.00000287, -0.00000411, -0.00000389, 0.00000111, -0.00001156, 0.00000683, -0.00001044, -0.00000873, -0.00001198, -0.00000812, -0.00003852, -0.00007683, -0.00033888, 0.00103303, -0.00030500, -0.00009629, -0.00002197, -0.00003417, -0.00000143, -0.00000853, 0.00000432, -0.00001812, 0.00000418, -0.00000713, -0.00000237, 0.00000072, -0.00000372, -0.00000161, 0.00000626, -0.00000576, 0.00000030, -0.00000466, 0.00000331, -0.00000355, -0.00000158, -0.00000201, -0.00000084, 0.00001160, -0.00002166, 0.00001013, -0.00002059, 0.00001748, 0.00000205, 0.00001074, 0.00000206, 0.00001926, -0.00002269, 0.00002034, -0.00002314, 0.00000840, -0.00000287, -0.00000249, -0.00000176, -0.00000004, -0.00000202, 0.00000428, -0.00000346, -0.00000536, 0.00000538, -0.00000204, 0.00000619, -0.00000636, 0.00000192, -0.00000559, 0.00000096, -0.00000558, 0.00001514, 0.00000398, -0.00001545, -0.00000910, 0.00001260, 0.00000692, -0.00001635, 0.00000645, -0.00000086, -0.00000266, 0.00000021, 0.00000073, -0.00000291, 0.00000157, 0.00001132, },
		{ -0.00000639, -0.00000751, 0.00000039, 0.00000565, 0.00001294, -0.00001426, -0.00000635, 0.00000946, 0.00001223, -0.00001283, 0.00000580, -0.00000778, -0.00000040, -0.00000109, -0.00000102, 0.00000383, 0.00000405, 0.00000183, -0.00000415, -0.00000318, 0.00000300, -0.00000089, -0.00000033, 0.00001513, -0.00001387, 0.00001531, -0.00001108, -0.00001098, 0.00000874, -0.00001249, 0.00001318, -0.00000747, 0.00000292, -0.00001150, -0.00000595, 0.00000501, 0.00000449, -0.00001505, 0.00003187, -0.00003444, 0.00002558, -0.00002204, 0.00002936, -0.00003217, 0.00002023, -0.00002189, 0.00002617, -0.00002916, 0.00002618, -0.00003773, 0.00003784, -0.00005035, 0.00004738, -0.00005502, 0.00004728, -0.00008003, 0.00007822, -0.00014374, 0.00010631, -0.00055327, 0.00092464, 0.00005726, -0.00032824, 0.00010050, -0.00012823, 0.00006694, -0.00006828, 0.00005457, -0.00005822, 0.00003643, -0.00003582, 0.00002383, -0.00002510, 0.00002151, -0.00002606, 0.00002693, -0.00002201, 0.00001506, -0.00001847, 0.00001471, -0.00001241, 0.00000649, -0.00000924, 0.00000366, 0.00000755, -0.00001309, 0.00000158, -0.00001104, 0.00000221, 0.00001355, 0.00000172, 0.00000771, 0.00001360, -0.00001096, 0.00000373, -0.00000118, -0.00001444, 0.00001654, -0.00001726, 0.00001058, -0.00001131, 0.00000864, -0.00000750, 0.00000964, -0.00001692, 0.00001409, -0.00000982, 0.00001301, -0.00001092, 0.00000624, -0.00000986, 0.00000525, -0.00001072, 0.00001525, 0.00000692, -0.00001091, -0.00001550, 0.00001142, 0.00001001, -0.00001272, -0.00000195, 0.00000678, -0.00000723, 0.00000294, -0.00000190, 0.00000103, -0.00000431, 0.00001399, },
		{ 0.00000099, -0.00001187, 0.00000105, 0.00000140, 0.00001701, -0.00001135, -0.00000976, 0.00000512, 0.00001757, -0.00001229, 0.00000330, -0.00000450, -0.00000361, 0.00000071, -0.00000214, 0.00000292, 0.00000491, 0.00000126, -0.00000117, -0.00000629, 0.00000409, -0.00000100, -0.00000112, 0.00001166, -0.00000392, -0.00000001, 0.00000984, -0.00002931, 0.00002129, -0.00002214, 0.00002006, -0.00001352, 0.00001155, -0.00001842, 0.00000202, -0.00001051, 0.00002437, -0.00003575, 0.00004856, -0.00004415, 0.00003254, -0.00003111, 0.00003853, -0.00003822, 0.00002702, -0.00003280, 0.00003922, -0.00004289, 0.00004245, -0.00005526, 0.00005572, -0.00006788, 0.00006490, -0.00007514, 0.00007505, -0.00011269, 0.00011778, -0.00018566, 0.00017942, -0.00055875, 0.00058411, 0.00051664, -0.00053748, 0.00018343, -0.00019047, 0.00010830, -0.00010502, 0.00008332, -0.00007926, 0.00005312, -0.00005201, 0.00003962, -0.00004162, 0.00003752, -0.00004066, 0.00003742, -0.00002924, 0.00002242, -0.00002571, 0.00002020, -0.00001675, 0.00001153, -0.00001431, 0.00000725, 0.00000235, -0.00000270, -0.00000911, 0.00000192, -0.00001519, 0.00002565, -0.00000851, 0.00001651, 0.00000202, 0.00000761, -0.00001820, 0.00002226, -0.00003428, 0.00003024, -0.00002603, 0.00001787, -0.00001848, 0.00001590, -0.00001586, 0.00001824, -0.00002301, 0.00001735, -0.00001260, 0.00001516, -0.00001163, 0.00000770, -0.00001173, 0.00000728, -0.00001281, 0.00001235, 0.00001006, -0.00000550, -0.00002049, 0.00000867, 0.00001170, -0.00000652, -0.00001063, 0.00001192, -0.00000846, 0.00000303, -0.00000263, 0.00000295, -0.00000649, 0.00001149, },
		{ 0.00000963, -0.00001487, 0.00000083, -0.00000203, 0.00001830, -0.00000582, -0.00001313, 0.00000077, 0.00001988, -0.00000761, -0.00000226, 0.00000069, -0.00000765, 0.00000267, -0.00000315, 0.00000210, 0.00000530, 0.00000110, 0.00000093, -0.00000756, 0.00000339, -0.00000043, -0.00000060, 0.00000556, 0.00000740, -0.00001352, 0.00002372, -0.00003577, 0.00002060, -0.00001920, 0.00001553, -0.00000950, 0.00001016, -0.00001543, 0.00000169, -0.00001728, 0.00003214, -0.00004033, 0.00004615, -0.00003446, 0.00002168, -0.00002283, 0.00002952, -0.00002630, 0.00001685, -0.00002623, 0.00003279, -0.00003508, 0.00003480, -0.00004589, 0.00004379, -0.00005287, 0.00004741, -0.00005712, 0.00005830, -0.00009106, 0.00008947, -0.00014101, 0.00012494, -0.00038040, 0.00011319, 0.00093269, -0.00061301, 0.00015897, -0.00017608, 0.00008920, -0.00009225, 0.00006924, -0.00006414, 0.00003989, -0.00004276, 0.00003286, -0.00003726, 0.00003339, -0.00003564, 0.00002959, -0.00002083, 0.00001655, -0.00002100, 0.00001501, -0.00001230, 0.00000936, -0.00001302, 0.00000584, 0.00000033, 0.00000482, -0.00001702, 0.00001249, -0.00002815, 0.00003204, -0.00001411, 0.00002242, -0.00000799, 0.00002326, -0.00003365, 0.00003479, -0.00004008, 0.00002957, -0.00002227, 0.00001485, -0.00001667, 0.00001492, -0.00001583, 0.00001767, -0.00001975, 0.00001219, -0.00000807, 0.00001085, -0.00000699, 0.00000474, -0.00000938, 0.00000501, -0.00001004, 0.00000544, 0.00001433, -0.00000118, -0.00002178, 0.00000280, 0.00001370, -0.00000088, -0.00001560, 0.00001126, -0.00000445, -0.00000058, -0.00000052, 0.00000171, -0.00000433, 0.00000474, },
		{ 0.00001590, -0.00001397, -0.00000185, -0.00000308, 0.00001569, 0.00000239, -0.00001656, -0.00000246, 0.00001803, 0.00000114, -0.00000987, 0.00000654, -0.00001146, 0.00000423, -0.00000406, 0.00000186, 0.00000440, 0.00000246, 0.00000079, -0.00000577, -0.00000003, 0.00000186, -0.00000090, 0.00000106, 0.00001381, -0.00001778, 0.00002370, -0.00002598, 0.00000518, -0.00000408, 0.00000080, 0.00000339, -0.00000017, -0.00000375, -0.00000629, -0.00001447, 0.00002619, -0.00002790, 0.00002549, -0.00000809, -0.00000343, -0.00000082, 0.00000613, -0.00000054, -0.00000653, -0.00000507, 0.00000930, -0.00000822, 0.00000607, -0.00001323, 0.00000673, -0.00001113, 0.00000163, -0.00000810, 0.00000500, -0.00002548, 0.00000834, -0.00003299, -0.00001348, -0.00011271, -0.00033616, 0.00115803, -0.00047137, -0.00000424, -0.00007261, 0.00000332, -0.00002674, 0.00001114, -0.00001308, -0.00000283, -0.00000802, 0.00000291, -0.00001105, 0.00000836, -0.00001095, 0.00000436, 0.00000186, -0.00000145, -0.00000504, -0.00000030, 0.00000067, -0.00000040, -0.00000447, -0.00000184, 0.00000347, 0.00000664, -0.00001856, 0.00001597, -0.00003125, 0.00002771, -0.00001033, 0.00001998, -0.00000938, 0.00002738, -0.00003360, 0.00002880, -0.00002702, 0.00001254, -0.00000568, 0.00000164, -0.00000580, 0.00000521, -0.00000679, 0.00000774, -0.00000775, -0.00000021, 0.00000237, 0.00000147, 0.00000182, -0.00000195, -0.00000306, -0.00000152, -0.00000277, -0.00000453, 0.00001918, 0.00000135, -0.00001860, -0.00000620, 0.00001645, 0.00000255, -0.00001513, 0.00000451, 0.00000363, -0.00000627, 0.00000308, -0.00000142, 0.00000022, -0.00000303, },
		{ 0.00001667, -0.00000780, -0.00000736, -0.00000143, 0.00000952, 0.00001199, -0.00001915, -0.00000456, 0.00001257, 0.00001196, -0.00001690, 0.00001069, -0.00001328, 0.00000429, -0.00000431, 0.00000197, 0.00000238, 0.00000519, -0.00000137, -0.00000151, -0.00000526, 0.00000537, -0.00000264, 0.00000020, 0.00001269, -0.00001113, 0.00001066, -0.00000434, -0.00001785, 0.00001511, -0.00001616, 0.00001770, -0.00001232, 0.00000962, -0.00001579, -0.00000715, 0.00001190, -0.00000559, -0.00000394, 0.00002376, -0.00003113, 0.00002345, -0.00001992, 0.00002705, -0.00003139, 0.00001921, -0.00001926, 0.00002450, -0.00002901, 0.00002616, -0.00003685, 0.00003658, -0.00004963, 0.00004686, -0.00005633, 0.00004971, -0.00008246, 0.00008223, -0.00015571, 0.00012305, -0.00062060, 0.00109848, -0.00010223, -0.00027311, 0.00008374, -0.00011807, 0.00006437, -0.00006697, 0.00005264, -0.00005681, 0.00003680, -0.00003687, 0.00002495, -0.00002619, 0.00002221, -0.00002731, 0.00002873, -0.00002294, 0.00001472, -0.00001914, 0.00001653, -0.00001323, 0.00000768, -0.00001284, 0.00000984, 0.00000358, -0.00001366, 0.00001123, -0.00002324, 0.00001219, 0.00000309, 0.00000820, -0.00000016, 0.00001768, -0.00001704, 0.00000606, -0.00000001, -0.00001402, 0.00001656, -0.00001541, 0.00000878, -0.00000854, 0.00000676, -0.00000669, 0.00000771, -0.00001454, 0.00001382, -0.00000859, 0.00001106, -0.00000933, 0.00000463, -0.00000972, 0.00000605, -0.00001435, 0.00002235, 0.00000317, -0.00001238, -0.00001619, 0.00001847, 0.00000436, -0.00001050, -0.00000533, 0.00001187, -0.00001045, 0.00000526, -0.00000373, 0.00000390, -0.00000797, },
		{ 0.00001118, 0.00000247, -0.00001399, 0.00000140, 0.00000192, 0.00002033, -0.00001907, -0.00000667, 0.00000565, 0.00002140, -0.00001994, 0.00001053, -0.00001135, 0.00000173, -0.00000308, 0.00000153, 0.00000044, 0.00000778, -0.00000367, 0.00000280, -0.00000973, 0.00000790, -0.00000438, 0.00000183, 0.00000628, 0.00000176, -0.00000743, 0.00001808, -0.00003622, 0.00002678, -0.00002494, 0.00002406, -0.00001747, 0.00001617, -0.00001912, -0.00000262, -0.00000208, 0.00001551, -0.00002888, 0.00004679, -0.00004745, 0.00003652, -0.00003501, 0.00004263, -0.00004404, 0.00003266, -0.00003766, 0.00004604, -0.00005154, 0.00005136, -0.00006395, 0.00006526, -0.00007930, 0.00007816, -0.00009196, 0.00009465, -0.00013488, 0.00014580, -0.00022621, 0.00023222, -0.00066258, 0.00075675, 0.00041179, -0.00054913, 0.00021532, -0.00021543, 0.00013294, -0.00012480, 0.00009891, -0.00009333, 0.00006713, -0.00006476, 0.00005089, -0.00005162, 0.00004620, -0.00004906, 0.00004535, -0.00003556, 0.00002740, -0.00003163, 0.00002658, -0.00002161, 0.00001667, -0.00002105, 0.00001440, -0.00000025, -0.00000580, 0.00000147, -0.00000802, -0.00000941, 0.00002074, -0.00000796, 0.00001495, -0.00000021, 0.00000804, -0.00002252, 0.00002807, -0.00003709, 0.00003290, -0.00002666, 0.00001885, -0.00001881, 0.00001743, -0.00001807, 0.00001909, -0.00002381, 0.00002011, -0.00001402, 0.00001616, -0.00001339, 0.00000982, -0.00001556, 0.00001208, -0.00001985, 0.00002086, 0.00000639, -0.00000595, -0.00002370, 0.00001708, 0.00000667, -0.00000492, -0.00001367, 0.00001568, -0.00000968, 0.00000355, -0.00000298, 0.00000421, -0.00000779, },
		{ 0.00000175, 0.00001324, -0.00001842, 0.00000270, -0.00000405, 0.00002438, -0.00001445, -0.00001022, -0.00000010, 0.00002605, -0.00001637, 0.00000467, -0.00000506, -0.00000372, 0.00000003, -0.00000029, -0.00000002, 0.00000832, -0.00000373, 0.00000435, -0.00001055, 0.00000691, -0.00000374, 0.00000286, -0.00000029, 0.00001299, -0.00002009, 0.00002962, -0.00003927, 0.00002248, -0.00001904, 0.00001714, -0.00001074, 0.00001128, -0.00001216, -0.00000541, -0.00000913, 0.00002643, -0.00003903, 0.00005086, -0.00004343, 0.00003024, -0.00003100, 0.00003801, -0.00003645, 0.00002670, -0.00003585, 0.00004482, -0.00004867, 0.00004842, -0.00005968, 0.00005922, -0.00007083, 0.00006788, -0.00008155, 0.00008558, -0.00012160, 0.00012701, -0.00019079, 0.00018605, -0.00048013, 0.00023821, 0.00091475, -0.00070068, 0.00023544, -0.00022885, 0.00013344, -0.00012609, 0.00009621, -0.00008816, 0.00006230, -0.00006229, 0.00004974, -0.00005202, 0.00004629, -0.00004776, 0.00004063, -0.00002997, 0.00002459, -0.00002994, 0.00002374, -0.00001913, 0.00001634, -0.00002065, 0.00001203, -0.00000006, 0.00000002, -0.00000774, 0.00000728, -0.00002844, 0.00003379, -0.00001991, 0.00002695, -0.00001569, 0.00002876, -0.00004230, 0.00004266, -0.00004410, 0.00003351, -0.00002451, 0.00001799, -0.00001952, 0.00001906, -0.00002017, 0.00002054, -0.00002299, 0.00001715, -0.00001140, 0.00001416, -0.00001134, 0.00000956, -0.00001573, 0.00001184, -0.00001802, 0.00001296, 0.00001237, -0.00000181, -0.00002582, 0.00001035, 0.00001121, -0.00000136, -0.00001685, 0.00001241, -0.00000291, -0.00000221, 0.00000093, 0.00000100, -0.00000308, },
		{ -0.00000722, 0.00001990, -0.00001737, 0.00000015, -0.00000600, 0.00002227, -0.00000479, -0.00001562, -0.00000307, 0.00002422, -0.00000596, -0.00000584, 0.00000406, -0.00001068, 0.00000419, -0.00000338, 0.00000156, 0.00000588, -0.00000035, 0.00000176, -0.00000653, 0.00000145, 0.00000048, 0.00000086, -0.00000258, 0.00001628, -0.00002036, 0.00002444, -0.00002413, 0.00000239, -0.00000035, -0.00000054, 0.00000536, -0.00000257, 0.00000258, -0.00001416, -0.00000832, 0.00002467, -0.00003192, 0.00003497, -0.00001995, 0.00000651, -0.00000985, 0.00001535, -0.00001102, 0.00000324, -0.00001479, 0.00002124, -0.00002091, 0.00001825, -0.00002562, 0.00002091, -0.00002760, 0.00002010, -0.00002951, 0.00002775, -0.00005025, 0.00003847, -0.00007109, 0.00002923, -0.00017559, -0.00028658, 0.00123280, -0.00061853, 0.00009325, -0.00013233, 0.00004984, -0.00006026, 0.00003754, -0.00003664, 0.00001857, -0.00002583, 0.00001787, -0.00002389, 0.00001951, -0.00002145, 0.00001375, -0.00000592, 0.00000575, -0.00001320, 0.00000713, -0.00000475, 0.00000520, -0.00000995, 0.00000104, 0.00000627, 0.00000062, -0.00001172, 0.00001601, -0.00003706, 0.00003457, -0.00002019, 0.00002775, -0.00001934, 0.00003438, -0.00004273, 0.00003535, -0.00002991, 0.00001610, -0.00000826, 0.00000570, -0.00000975, 0.00001036, -0.00001163, 0.00001116, -0.00001196, 0.00000541, -0.00000140, 0.00000563, -0.00000344, 0.00000358, -0.00000961, 0.00000472, -0.00000898, -0.00000021, 0.00002029, -0.00000051, -0.00002196, -0.00000125, 0.00001765, -0.00000058, -0.00001416, 0.00000297, 0.00000743, -0.00000929, 0.00000561, -0.00000339, 0.00000301, },
		{ -0.00001153, 0.00001927, -0.00000958, -0.00000652, -0.00000366, 0.00001457, 0.00000812, -0.00002140, -0.00000369, 0.00001696, 0.00000849, -0.00001733, 0.00001248, -0.00001626, 0.00000733, -0.00000635, 0.00000416, 0.00000140, 0.00000544, -0.00000376, 0.00000072, -0.00000664, 0.00000688, -0.00000385, 0.00000030, 0.00001060, -0.00000862, 0.00000589, 0.00000223, -0.00002408, 0.00002111, -0.00001936, 0.00002184, -0.00001685, 0.00001706, -0.00002234, -0.00000435, 0.00001464, -0.00001347, 0.00000746, 0.00001254, -0.00002347, 0.00001720, -0.00001375, 0.00002022, -0.00002586, 0.00001387, -0.00001232, 0.00001791, -0.00002357, 0.00002068, -0.00002995, 0.00002840, -0.00004081, 0.00003711, -0.00004758, 0.00004114, -0.00007173, 0.00006957, -0.00014692, 0.00011158, -0.00064824, 0.00124372, -0.00027084, -0.00019030, 0.00004373, -0.00008934, 0.00004680, -0.00005370, 0.00004027, -0.00004589, 0.00002896, -0.00003097, 0.00002021, -0.00002220, 0.00001815, -0.00002385, 0.00002619, -0.00002025, 0.00001097, -0.00001640, 0.00001558, -0.00001184, 0.00000681, -0.00001481, 0.00001619, -0.00000314, -0.00000954, 0.00001555, -0.00003225, 0.00002095, -0.00000714, 0.00001533, -0.00000845, 0.00002240, -0.00002318, 0.00000878, -0.00000029, -0.00001189, 0.00001471, -0.00001171, 0.00000530, -0.00000413, 0.00000312, -0.00000426, 0.00000411, -0.00000995, 0.00001120, -0.00000525, 0.00000678, -0.00000522, 0.00000013, -0.00000632, 0.00000357, -0.00001445, 0.00002689, -0.00000017, -0.00001430, -0.00001454, 0.00002324, -0.00000078, -0.00000805, -0.00000848, 0.00001645, -0.00001338, 0.00000757, -0.00000567, 0.00000669, },
		{ -0.00000942, 0.00001136, 0.00000293, -0.00001484, 0.00000070, 0.00000439, 0.00002043, -0.00002455, -0.00000415, 0.00000758, 0.00002210, -0.00002474, 0.00001600, -0.00001723, 0.00000701, -0.00000714, 0.00000570, -0.00000265, 0.00001073, -0.00000881, 0.00000728, -0.00001329, 0.00001177, -0.00000818, 0.00000523, 0.00000046, 0.00000792, -0.00001544, 0.00002645, -0.00004290, 0.00003244, -0.00002788, 0.00002842, -0.00002198, 0.00002242, -0.00002256, -0.00000376, 0.00000408, 0.00000597, -0.00001865, 0.00003942, -0.00004490, 0.00003554, -0.00003430, 0.00004185, -0.00004510, 0.00003409, -0.00003820, 0.00004792, -0.00005500, 0.00005494, -0.00006675, 0.00006811, -0.00008314, 0.00008315, -0.00010014, 0.00010495, -0.00014609, 0.00016045, -0.00025126, 0.00026605, -0.00074534, 0.00093031, 0.00026904, -0.00051621, 0.00021670, -0.00021861, 0.00014120, -0.00013205, 0.00010385, -0.00009810, 0.00007349, -0.00007127, 0.00005699, -0.00005709, 0.00005059, -0.00005332, 0.00004971, -0.00003908, 0.00002978, -0.00003510, 0.00003129, -0.00002540, 0.00002098, -0.00002768, 0.00002335, -0.00000663, -0.00000445, 0.00000842, -0.00001739, -0.00000207, 0.00001353, -0.00000465, 0.00001112, -0.00000004, 0.00000652, -0.00002423, 0.00003036, -0.00003650, 0.00003229, -0.00002424, 0.00001707, -0.00001647, 0.00001619, -0.00001767, 0.00001726, -0.00002149, 0.00001984, -0.00001274, 0.00001428, -0.00001217, 0.00000878, -0.00001612, 0.00001392, -0.00002411, 0.00002798, 0.00000224, -0.00000663, -0.00002489, 0.00002418, 0.00000109, -0.00000250, -0.00001678, 0.00001931, -0.00001108, 0.00000452, -0.00000379, 0.00000572, },
		{ -0.00000255, -0.00000028, 0.00001541, -0.00002052, 0.00000338, -0.00000404, 0.00002776, -0.00002197, -0.00000699, -0.00000005, 0.00002998, -0.00002393, 0.00001183, -0.00001183, 0.00000193, -0.00000438, 0.00000430, -0.00000371, 0.00001234, -0.00000974, 0.00000912, -0.00001445, 0.00001137, -0.00000833, 0.00000748, -0.00000722, 0.00001959, -0.00002763, 0.00003597, -0.00004313, 0.00002530, -0.00001977, 0.00001989, -0.00001332, 0.00001462, -0.00001159, -0.00001009, -0.00000111, 0.00001759, -0.00003248, 0.00004926, -0.00004704, 0.00003499, -0.00003590, 0.00004321, -0.00004371, 0.00003427, -0.00004305, 0.00005392, -0.00005933, 0.00005918, -0.00007025, 0.00007100, -0.00008482, 0.00008440, -0.00010234, 0.00010923, -0.00014782, 0.00015965, -0.00023625, 0.00024377, -0.00058209, 0.00039246, 0.00083892, -0.00073921, 0.00028532, -0.00026542, 0.00016682, -0.00015238, 0.00011699, -0.00010709, 0.00008101, -0.00007911, 0.00006455, -0.00006495, 0.00005733, -0.00005813, 0.00005047, -0.00003851, 0.00003202, -0.00003842, 0.00003273, -0.00002661, 0.00002411, -0.00002976, 0.00002122, -0.00000443, -0.00000152, 0.00000036, -0.00000023, -0.00002473, 0.00003139, -0.00002164, 0.00002817, -0.00002007, 0.00003123, -0.00004719, 0.00004626, -0.00004445, 0.00003414, -0.00002384, 0.00001854, -0.00001981, 0.00002058, -0.00002206, 0.00002088, -0.00002341, 0.00001959, -0.00001259, 0.00001516, -0.00001339, 0.00001205, -0.00001992, 0.00001707, -0.00002489, 0.00002087, 0.00000874, -0.00000213, -0.00002865, 0.00001795, 0.00000701, -0.00000045, -0.00001852, 0.00001391, -0.00000211, -0.00000294, 0.00000159, 0.00000085, },
		{ 0.00000493, -0.00001032, 0.00002253, -0.00001965, 0.00000137, -0.00000747, 0.00002737, -0.00001244, -0.00001322, -0.00000359, 0.00002957, -0.00001389, 0.00000044, -0.00000117, -0.00000682, 0.00000142, -0.00000027, -0.00000082, 0.00000888, -0.00000490, 0.00000453, -0.00000865, 0.00000440, -0.00000264, 0.00000411, -0.00000756, 0.00002000, -0.00002412, 0.00002596, -0.00002327, 0.00000123, 0.00000188, -0.00000019, 0.00000556, -0.00000266, 0.00000679, -0.00002102, -0.00000039, 0.00001855, -0.00003045, 0.00003935, -0.00002880, 0.00001534, -0.00001831, 0.00002412, -0.00002164, 0.00001376, -0.00002505, 0.00003343, -0.00003417, 0.00003131, -0.00003884, 0.00003598, -0.00004526, 0.00004052, -0.00005378, 0.00005407, -0.00007901, 0.00007373, -0.00011722, 0.00008415, -0.00026052, -0.00018949, 0.00124540, -0.00072774, 0.00017984, -0.00018939, 0.00009664, -0.00009540, 0.00006548, -0.00006191, 0.00004227, -0.00004613, 0.00003521, -0.00003877, 0.00003232, -0.00003352, 0.00002508, -0.00001586, 0.00001481, -0.00002320, 0.00001689, -0.00001254, 0.00001303, -0.00001809, 0.00000747, 0.00000573, -0.00000397, -0.00000353, 0.00001150, -0.00003754, 0.00003681, -0.00002610, 0.00003256, -0.00002632, 0.00003867, -0.00004853, 0.00003861, -0.00003048, 0.00001780, -0.00000935, 0.00000852, -0.00001239, 0.00001412, -0.00001525, 0.00001330, -0.00001474, 0.00001002, -0.00000452, 0.00000894, -0.00000795, 0.00000845, -0.00001586, 0.00001137, -0.00001618, 0.00000639, 0.00001859, -0.00000145, -0.00002508, 0.00000513, 0.00001618, -0.00000200, -0.00001384, 0.00000220, 0.00001006, -0.00001115, 0.00000723, -0.00000469, },
		{ 0.00000878, -0.00001429, 0.00002104, -0.00001105, -0.00000563, -0.00000545, 0.00001970, 0.00000218, -0.00002121, -0.00000350, 0.00002189, 0.00000245, -0.00001409, 0.00001064, -0.00001576, 0.00000759, -0.00000610, 0.00000450, 0.00000181, 0.00000401, -0.00000447, 0.00000167, -0.00000654, 0.00000674, -0.00000376, -0.00000078, 0.00000964, -0.00000723, 0.00000205, 0.00000741, -0.00002837, 0.00002548, -0.00002087, 0.00002441, -0.00001962, 0.00002348, -0.00002947, 0.00000165, 0.00001246, -0.00001711, 0.00001637, 0.00000105, -0.00001370, 0.00000856, -0.00000510, 0.00001032, -0.00001677, 0.00000525, -0.00000216, 0.00000749, -0.00001381, 0.00001069, -0.00001810, 0.00001449, -0.00002508, 0.00001917, -0.00002959, 0.00002221, -0.00004842, 0.00004034, -0.00011630, 0.00006829, -0.00062561, 0.00134032, -0.00042672, -0.00009492, -0.00001074, -0.00004756, 0.00001775, -0.00003094, 0.00001945, -0.00002699, 0.00001389, -0.00001870, 0.00001001, -0.00001345, 0.00000961, -0.00001582, 0.00001929, -0.00001386, 0.00000370, -0.00000995, 0.00001125, -0.00000749, 0.00000288, -0.00001356, 0.00002052, -0.00001046, -0.00000210, 0.00001425, -0.00003619, 0.00002626, -0.00001503, 0.00002116, -0.00001539, 0.00002602, -0.00002782, 0.00001056, -0.00000074, -0.00000930, 0.00001221, -0.00000731, 0.00000133, 0.00000066, -0.00000108, -0.00000133, 0.00000007, -0.00000444, 0.00000745, -0.00000102, 0.00000146, 0.00000011, -0.00000601, -0.00000057, -0.00000169, -0.00001088, 0.00002806, -0.00000236, -0.00001675, -0.00001045, 0.00002496, -0.00000438, -0.00000619, -0.00001071, 0.00001977, -0.00001528, 0.00000921, -0.00000715, },
		{ 0.00000705, -0.00001089, 0.00001162, 0.00000266, -0.00001471, -0.00000063, 0.00000832, 0.00001733, -0.00002720, -0.00000262, 0.00001095, 0.00001927, -0.00002555, 0.00001814, -0.00002039, 0.00001051, -0.00000994, 0.00000897, -0.00000511, 0.00001265, -0.00001299, 0.00001114, -0.00001595, 0.00001466, -0.00001137, 0.00000805, -0.00000472, 0.00001334, -0.00002274, 0.00003354, -0.00004802, 0.00003711, -0.00002990, 0.00003212, -0.00002591, 0.00002915, -0.00002804, -0.00000092, 0.00000602, -0.00000152, -0.00000783, 0.00002864, -0.00003768, 0.00003015, -0.00002928, 0.00003633, -0.00004116, 0.00003089, -0.00003403, 0.00004434, -0.00005250, 0.00005234, -0.00006277, 0.00006325, -0.00007807, 0.00007811, -0.00009732, 0.00010311, -0.00014287, 0.00015708, -0.00025421, 0.00027101, -0.00079070, 0.00108274, 0.00010680, -0.00044713, 0.00018954, -0.00019968, 0.00013173, -0.00012519, 0.00009668, -0.00009205, 0.00007050, -0.00006978, 0.00005627, -0.00005657, 0.00004936, -0.00005211, 0.00004915, -0.00003852, 0.00002832, -0.00003478, 0.00003284, -0.00002661, 0.00002286, -0.00003227, 0.00003205, -0.00001543, 0.00000130, 0.00001011, -0.00002343, 0.00000423, 0.00000607, -0.00000024, 0.00000638, 0.00000126, 0.00000406, -0.00002416, 0.00002994, -0.00003329, 0.00002911, -0.00001947, 0.00001325, -0.00001220, 0.00001283, -0.00001524, 0.00001341, -0.00001671, 0.00001707, -0.00000927, 0.00001002, -0.00000838, 0.00000474, -0.00001316, 0.00001210, -0.00002440, 0.00003213, -0.00000119, -0.00000807, -0.00002337, 0.00002866, -0.00000374, -0.00000016, -0.00001922, 0.00002206, -0.00001204, 0.00000542, -0.00000465, },
		{ 0.00000115, -0.00000263, -0.00000134, 0.00001601, -0.00002100, 0.00000274, -0.00000178, 0.00002769, -0.00002712, -0.00000437, 0.00000156, 0.00003049, -0.00002843, 0.00001714, -0.00001761, 0.00000754, -0.00000911, 0.00000937, -0.00000791, 0.00001639, -0.00001588, 0.00001428, -0.00001841, 0.00001590, -0.00001331, 0.00001240, -0.00001427, 0.00002617, -0.00003512, 0.00004178, -0.00004656, 0.00002844, -0.00002092, 0.00002338, -0.00001682, 0.00001995, -0.00001418, -0.00001075, 0.00000412, 0.00000834, -0.00002273, 0.00004229, -0.00004507, 0.00003509, -0.00003650, 0.00004390, -0.00004652, 0.00003790, -0.00004612, 0.00005820, -0.00006491, 0.00006477, -0.00007519, 0.00007643, -0.00009165, 0.00009321, -0.00011503, 0.00012420, -0.00016391, 0.00018017, -0.00026817, 0.00028605, -0.00066988, 0.00055842, 0.00071443, -0.00072671, 0.00030209, -0.00027907, 0.00018318, -0.00016578, 0.00012704, -0.00011676, 0.00009207, -0.00008965, 0.00007414, -0.00007329, 0.00006405, -0.00006438, 0.00005687, -0.00004442, 0.00003696, -0.00004449, 0.00004004, -0.00003299, 0.00003098, -0.00003863, 0.00003207, -0.00001252, 0.00000137, 0.00000474, -0.00000707, -0.00001917, 0.00002614, -0.00002012, 0.00002658, -0.00002146, 0.00003083, -0.00004833, 0.00004567, -0.00004117, 0.00003144, -0.00002019, 0.00001646, -0.00001752, 0.00001933, -0.00002129, 0.00001858, -0.00002083, 0.00001918, -0.00001126, 0.00001340, -0.00001252, 0.00001131, -0.00002061, 0.00001907, -0.00002875, 0.00002723, 0.00000473, -0.00000280, -0.00002931, 0.00002416, 0.00000224, 0.00000121, -0.00002013, 0.00001532, -0.00000177, -0.00000293, 0.00000155, },
		{ -0.00000507, 0.00000570, -0.00001192, 0.00002326, -0.00002030, 0.00000128, -0.00000677, 0.00002975, -0.00001903, -0.00001041, -0.00000315, 0.00003257, -0.00002065, 0.00000710, -0.00000759, -0.00000107, -0.00000328, 0.00000464, -0.00000481, 0.00001296, -0.00001071, 0.00000874, -0.00001184, 0.00000850, -0.00000717, 0.00000853, -0.00001350, 0.00002459, -0.00002874, 0.00002807, -0.00002348, 0.00000195, 0.00000227, 0.00000225, 0.00000357, 0.00000028, 0.00000748, -0.00002461, 0.00000663, 0.00000984, -0.00002426, 0.00003807, -0.00003299, 0.00002108, -0.00002426, 0.00003038, -0.00003010, 0.00002272, -0.00003356, 0.00004340, -0.00004532, 0.00004248, -0.00005001, 0.00004872, -0.00006043, 0.00005864, -0.00007618, 0.00007880, -0.00010599, 0.00010728, -0.00016320, 0.00014158, -0.00035627, -0.00005304, 0.00119301, -0.00078647, 0.00024183, -0.00023263, 0.00013472, -0.00012492, 0.00008902, -0.00008368, 0.00006355, -0.00006484, 0.00005147, -0.00005273, 0.00004415, -0.00004472, 0.00003609, -0.00002598, 0.00002388, -0.00003323, 0.00002733, -0.00002136, 0.00002196, -0.00002796, 0.00001717, 0.00000089, -0.00000501, 0.00000314, 0.00000491, -0.00003382, 0.00003457, -0.00002775, 0.00003387, -0.00002973, 0.00003951, -0.00005023, 0.00003793, -0.00002805, 0.00001688, -0.00000819, 0.00000933, -0.00001296, 0.00001561, -0.00001676, 0.00001334, -0.00001515, 0.00001256, -0.00000588, 0.00001020, -0.00001033, 0.00001105, -0.00001987, 0.00001638, -0.00002232, 0.00001353, 0.00001503, -0.00000192, -0.00002714, 0.00001180, 0.00001268, -0.00000181, -0.00001416, 0.00000222, 0.00001136, -0.00001163, 0.00000768, },
		{ -0.00000770, 0.00000968, -0.00001563, 0.00002133, -0.00001157, -0.00000538, -0.00000590, 0.00002356, -0.00000449, -0.00001933, -0.00000331, 0.00002607, -0.00000466, -0.00000825, 0.00000575, -0.00001193, 0.00000491, -0.00000329, 0.00000262, 0.00000396, 0.00000058, -0.00000310, 0.00000093, -0.00000459, 0.00000458, -0.00000196, -0.00000339, 0.00001024, -0.00000749, -0.00000018, 0.00001028, -0.00002972, 0.00002723, -0.00001969, 0.00002434, -0.00001927, 0.00002700, -0.00003482, 0.00000877, 0.00000616, -0.00001581, 0.00002089, -0.00000818, -0.00000427, -0.00000031, 0.00000391, -0.00000048, -0.00000616, -0.00000474, 0.00000917, -0.00000461, -0.00000186, -0.00000162, -0.00000374, -0.00000241, -0.00000548, -0.00000372, -0.00000560, -0.00000363, -0.00001645, -0.00000125, -0.00006797, -0.00000379, -0.00054976, 0.00137421, -0.00054976, -0.00000379, -0.00006797, -0.00000125, -0.00001645, -0.00000363, -0.00000560, -0.00000372, -0.00000548, -0.00000241, -0.00000374, -0.00000162, -0.00000186, -0.00000461, 0.00000917, -0.00000474, -0.00000616, -0.00000048, 0.00000391, -0.00000031, -0.00000427, -0.00000818, 0.00002089, -0.00001581, 0.00000616, 0.00000877, -0.00003482, 0.00002700, -0.00001927, 0.00002434, -0.00001969, 0.00002723, -0.00002972, 0.00001028, -0.00000018, -0.00000749, 0.00001024, -0.00000339, -0.00000196, 0.00000458, -0.00000459, 0.00000093, -0.00000310, 0.00000058, 0.00000396, 0.00000262, -0.00000329, 0.00000491, -0.00001193, 0.00000575, -0.00000825, -0.00000466, 0.00002607, -0.00000331, -0.00001933, -0.00000449, 0.00002356, -0.00000590, -0.00000538, -0.00001157, 0.00002133, -0.00001563, 0.00000968, },
		{ -0.00000507, 0.00000768, -0.00001163, 0.00001136, 0.00000222, -0.00001416, -0.00000181, 0.00001268, 0.00001180, -0.00002714, -0.00000192, 0.00001503, 0.00001353, -0.00002232, 0.00001638, -0.00001987, 0.00001105, -0.00001033, 0.00001020, -0.00000588, 0.00001256, -0.00001515, 0.00001334, -0.00001676, 0.00001561, -0.00001296, 0.00000933, -0.00000819, 0.00001688, -0.00002805, 0.00003793, -0.00005023, 0.00003951, -0.00002973, 0.00003387, -0.00002775, 0.00003457, -0.00003382, 0.00000491, 0.00000314, -0.00000501, 0.00000089, 0.00001717, -0.00002796, 0.00002196, -0.00002136, 0.00002733, -0.00003323, 0.00002388, -0.00002598, 0.00003609, -0.00004472, 0.00004415, -0.00005273, 0.00005147, -0.00006484, 0.00006355, -0.00008368, 0.00008902, -0.00012492, 0.00013472, -0.00023263, 0.00024183, -0.00078647, 0.00119301, -0.00005304, -0.00035627, 0.00014158, -0.00016320, 0.00010728, -0.00010599, 0.00007880, -0.00007618, 0.00005864, -0.00006043, 0.00004872, -0.00005001, 0.00004248, -0.00004532, 0.00004340, -0.00003356, 0.00002272, -0.00003010, 0.00003038, -0.00002426, 0.00002108, -0.00003299, 0.00003807, -0.00002426, 0.00000984, 0.00000663, -0.00002461, 0.00000748, 0.00000028, 0.00000357, 0.00000225, 0.00000227, 0.00000195, -0.00002348, 0.00002807, -0.00002874, 0.00002459, -0.00001350, 0.00000853, -0.00000717, 0.00000850, -0.00001184, 0.00000874, -0.00001071, 0.00001296, -0.00000481, 0.00000464, -0.00000328, -0.00000107, -0.00000759, 0.00000710, -0.00002065, 0.00003257, -0.00000315, -0.00001041, -0.00001903, 0.00002975, -0.00000677, 0.00000128, -0.00002030, 0.00002326, -0.00001192, 0.00000570, },
		{ 0.00000115, 0.00000155, -0.00000293, -0.00000177, 0.00001532, -0.00002013, 0.00000121, 0.00000224, 0.00002416, -0.00002931, -0.00000280, 0.00000473, 0.00002723, -0.00002875, 0.00001907, -0.00002061, 0.00001131, -0.00001252, 0.00001340, -0.00001126, 0.00001918, -0.00002083, 0.00001858, -0.00002129, 0.00001933, -0.00001752, 0.00001646, -0.00002019, 0.00003144, -0.00004117, 0.00004567, -0.00004833, 0.00003083, -0.00002146, 0.00002658, -0.00002012, 0.00002614, -0.00001917, -0.00000707, 0.00000474, 0.00000137, -0.00001252, 0.00003207, -0.00003863, 0.00003098, -0.00003299, 0.00004004, -0.00004449, 0.00003696, -0.00004442, 0.00005687, -0.00006438, 0.00006405, -0.00007329, 0.00007414, -0.00008965, 0.00009207, -0.00011676, 0.00012704, -0.00016578, 0.00018318, -0.00027907, 0.00030209, -0.00072671, 0.00071443, 0.00055842, -0.00066988, 0.00028605, -0.00026817, 0.00018017, -0.00016391, 0.00012420, -0.00011503, 0.00009321, -0.00009165, 0.00007643, -0.00007519, 0.00006477, -0.00006491, 0.00005820, -0.00004612, 0.00003790, -0.00004652, 0.00004390, -0.00003650, 0.00003509, -0.00004507, 0.00004229, -0.00002273, 0.00000834, 0.00000412, -0.00001075, -0.00001418, 0.00001995, -0.00001682, 0.00002338, -0.00002092, 0.00002844, -0.00004656, 0.00004178, -0.00003512, 0.00002617, -0.00001427, 0.00001240, -0.00001331, 0.00001590, -0.00001841, 0.00001428, -0.00001588, 0.00001639, -0.00000791, 0.00000937, -0.00000911, 0.00000754, -0.00001761, 0.00001714, -0.00002843, 0.00003049, 0.00000156, -0.00000437, -0.00002712, 0.00002769, -0.00000178, 0.00000274, -0.00002100, 0.00001601, -0.00000134, -0.00000263, },
		{ 0.00000705, -0.00000465, 0.00000542, -0.00001204, 0.00002206, -0.00001922, -0.00000016, -0.00000374, 0.00002866, -0.00002337, -0.00000807, -0.00000119, 0.00003213, -0.00002440, 0.00001210, -0.00001316, 0.00000474, -0.00000838, 0.00001002, -0.00000927, 0.00001707, -0.00001671, 0.00001341, -0.00001524, 0.00001283, -0.00001220, 0.00001325, -0.00001947, 0.00002911, -0.00003329, 0.00002994, -0.00002416, 0.00000406, 0.00000126, 0.00000638, -0.00000024, 0.00000607, 0.00000423, -0.00002343, 0.00001011, 0.00000130, -0.00001543, 0.00003205, -0.00003227, 0.00002286, -0.00002661, 0.00003284, -0.00003478, 0.00002832, -0.00003852, 0.00004915, -0.00005211, 0.00004936, -0.00005657, 0.00005627, -0.00006978, 0.00007050, -0.00009205, 0.00009668, -0.00012519, 0.00013173, -0.00019968, 0.00018954, -0.00044713, 0.00010680, 0.00108274, -0.00079070, 0.00027101, -0.00025421, 0.00015708, -0.00014287, 0.00010311, -0.00009732, 0.00007811, -0.00007807, 0.00006325, -0.00006277, 0.00005234, -0.00005250, 0.00004434, -0.00003403, 0.00003089, -0.00004116, 0.00003633, -0.00002928, 0.00003015, -0.00003768, 0.00002864, -0.00000783, -0.00000152, 0.00000602, -0.00000092, -0.00002804, 0.00002915, -0.00002591, 0.00003212, -0.00002990, 0.00003711, -0.00004802, 0.00003354, -0.00002274, 0.00001334, -0.00000472, 0.00000805, -0.00001137, 0.00001466, -0.00001595, 0.00001114, -0.00001299, 0.00001265, -0.00000511, 0.00000897, -0.00000994, 0.00001051, -0.00002039, 0.00001814, -0.00002555, 0.00001927, 0.00001095, -0.00000262, -0.00002720, 0.00001733, 0.00000832, -0.00000063, -0.00001471, 0.00000266, 0.00001162, -0.00001089, },
		{ 0.00000878, -0.00000715, 0.00000921, -0.00001528, 0.00001977, -0.00001071, -0.00000619, -0.00000438, 0.00002496, -0.00001045, -0.00001675, -0.00000236, 0.00002806, -0.00001088, -0.00000169, -0.00000057, -0.00000601, 0.00000011, 0.00000146, -0.00000102, 0.00000745, -0.00000444, 0.00000007, -0.00000133, -0.00000108, 0.00000066, 0.00000133, -0.00000731, 0.00001221, -0.00000930, -0.00000074, 0.00001056, -0.00002782, 0.00002602, -0.00001539, 0.00002116, -0.00001503, 0.00002626, -0.00003619, 0.00001425, -0.00000210, -0.00001046, 0.00002052, -0.00001356, 0.00000288, -0.00000749, 0.00001125, -0.00000995, 0.00000370, -0.00001386, 0.00001929, -0.00001582, 0.00000961, -0.00001345, 0.00001001, -0.00001870, 0.00001389, -0.00002699, 0.00001945, -0.00003094, 0.00001775, -0.00004756, -0.00001074, -0.00009492, -0.00042672, 0.00134032, -0.00062561, 0.00006829, -0.00011630, 0.00004034, -0.00004842, 0.00002221, -0.00002959, 0.00001917, -0.00002508, 0.00001449, -0.00001810, 0.00001069, -0.00001381, 0.00000749, -0.00000216, 0.00000525, -0.00001677, 0.00001032, -0.00000510, 0.00000856, -0.00001370, 0.00000105, 0.00001637, -0.00001711, 0.00001246, 0.00000165, -0.00002947, 0.00002348, -0.00001962, 0.00002441, -0.00002087, 0.00002548, -0.00002837, 0.00000741, 0.00000205, -0.00000723, 0.00000964, -0.00000078, -0.00000376, 0.00000674, -0.00000654, 0.00000167, -0.00000447, 0.00000401, 0.00000181, 0.00000450, -0.00000610, 0.00000759, -0.00001576, 0.00001064, -0.00001409, 0.00000245, 0.00002189, -0.00000350, -0.00002121, 0.00000218, 0.00001970, -0.00000545, -0.00000563, -0.00001105, 0.00002104, -0.00001429, },
		{ 0.00000493, -0.00000469, 0.00000723, -0.00001115, 0.00001006, 0.00000220, -0.00001384, -0.00000200, 0.00001618, 0.00000513, -0.00002508, -0.00000145, 0.00001859, 0.00000639, -0.00001618, 0.00001137, -0.00001586, 0.00000845, -0.00000795, 0.00000894, -0.00000452, 0.00001002, -0.00001474, 0.00001330, -0.00001525, 0.00001412, -0.00001239, 0.00000852, -0.00000935, 0.00001780, -0.00003048, 0.00003861, -0.00004853, 0.00003867, -0.00002632, 0.00003256, -0.00002610, 0.00003681, -0.00003754, 0.00001150, -0.00000353, -0.00000397, 0.00000573, 0.00000747, -0.00001809, 0.00001303, -0.00001254, 0.00001689, -0.00002320, 0.00001481, -0.00001586, 0.00002508, -0.00003352, 0.00003232, -0.00003877, 0.00003521, -0.00004613, 0.00004227, -0.00006191, 0.00006548, -0.00009540, 0.00009664, -0.00018939, 0.00017984, -0.00072774, 0.00124540, -0.00018949, -0.00026052, 0.00008415, -0.00011722, 0.00007373, -0.00007901, 0.00005407, -0.00005378, 0.00004052, -0.00004526, 0.00003598, -0.00003884, 0.00003131, -0.00003417, 0.00003343, -0.00002505, 0.00001376, -0.00002164, 0.00002412, -0.00001831, 0.00001534, -0.00002880, 0.00003935, -0.00003045, 0.00001855, -0.00000039, -0.00002102, 0.00000679, -0.00000266, 0.00000556, -0.00000019, 0.00000188, 0.00000123, -0.00002327, 0.00002596, -0.00002412, 0.00002000, -0.00000756, 0.00000411, -0.00000264, 0.00000440, -0.00000865, 0.00000453, -0.00000490, 0.00000888, -0.00000082, -0.00000027, 0.00000142, -0.00000682, -0.00000117, 0.00000044, -0.00001389, 0.00002957, -0.00000359, -0.00001322, -0.00001244, 0.00002737, -0.00000747, 0.00000137, -0.00001965, 0.00002253, -0.00001032, },
		{ -0.00000255, 0.00000085, 0.00000159, -0.00000294, -0.00000211, 0.00001391, -0.00001852, -0.00000045, 0.00000701, 0.00001795, -0.00002865, -0.00000213, 0.00000874, 0.00002087, -0.00002489, 0.00001707, -0.00001992, 0.00001205, -0.00001339, 0.00001516, -0.00001259, 0.00001959, -0.00002341, 0.00002088, -0.00002206, 0.00002058, -0.00001981, 0.00001854, -0.00002384, 0.00003414, -0.00004445, 0.00004626, -0.00004719, 0.00003123, -0.00002007, 0.00002817, -0.00002164, 0.00003139, -0.00002473, -0.00000023, 0.00000036, -0.00000152, -0.00000443, 0.00002122, -0.00002976, 0.00002411, -0.00002661, 0.00003273, -0.00003842, 0.00003202, -0.00003851, 0.00005047, -0.00005813, 0.00005733, -0.00006495, 0.00006455, -0.00007911, 0.00008101, -0.00010709, 0.00011699, -0.00015238, 0.00016682, -0.00026542, 0.00028532, -0.00073921, 0.00083892, 0.00039246, -0.00058209, 0.00024377, -0.00023625, 0.00015965, -0.00014782, 0.00010923, -0.00010234, 0.00008440, -0.00008482, 0.00007100, -0.00007025, 0.00005918, -0.00005933, 0.00005392, -0.00004305, 0.00003427, -0.00004371, 0.00004321, -0.00003590, 0.00003499, -0.00004704, 0.00004926, -0.00003248, 0.00001759, -0.00000111, -0.00001009, -0.00001159, 0.00001462, -0.00001332, 0.00001989, -0.00001977, 0.00002530, -0.00004313, 0.00003597, -0.00002763, 0.00001959, -0.00000722, 0.00000748, -0.00000833, 0.00001137, -0.00001445, 0.00000912, -0.00000974, 0.00001234, -0.00000371, 0.00000430, -0.00000438, 0.00000193, -0.00001183, 0.00001183, -0.00002393, 0.00002998, -0.00000005, -0.00000699, -0.00002197, 0.00002776, -0.00000404, 0.00000338, -0.00002052, 0.00001541, -0.00000028, },
		{ -0.00000942, 0.00000572, -0.00000379, 0.00000452, -0.00001108, 0.00001931, -0.00001678, -0.00000250, 0.00000109, 0.00002418, -0.00002489, -0.00000663, 0.00000224, 0.00002798, -0.00002411, 0.00001392, -0.00001612, 0.00000878, -0.00001217, 0.00001428, -0.00001274, 0.00001984, -0.00002149, 0.00001726, -0.00001767, 0.00001619, -0.00001647, 0.00001707, -0.00002424, 0.00003229, -0.00003650, 0.00003036, -0.00002423, 0.00000652, -0.00000004, 0.00001112, -0.00000465, 0.00001353, -0.00000207, -0.00001739, 0.00000842, -0.00000445, -0.00000663, 0.00002335, -0.00002768, 0.00002098, -0.00002540, 0.00003129, -0.00003510, 0.00002978, -0.00003908, 0.00004971, -0.00005332, 0.00005059, -0.00005709, 0.00005699, -0.00007127, 0.00007349, -0.00009810, 0.00010385, -0.00013205, 0.00014120, -0.00021861, 0.00021670, -0.00051621, 0.00026904, 0.00093031, -0.00074534, 0.00026605, -0.00025126, 0.00016045, -0.00014609, 0.00010495, -0.00010014, 0.00008315, -0.00008314, 0.00006811, -0.00006675, 0.00005494, -0.00005500, 0.00004792, -0.00003820, 0.00003409, -0.00004510, 0.00004185, -0.00003430, 0.00003554, -0.00004490, 0.00003942, -0.00001865, 0.00000597, 0.00000408, -0.00000376, -0.00002256, 0.00002242, -0.00002198, 0.00002842, -0.00002788, 0.00003244, -0.00004290, 0.00002645, -0.00001544, 0.00000792, 0.00000046, 0.00000523, -0.00000818, 0.00001177, -0.00001329, 0.00000728, -0.00000881, 0.00001073, -0.00000265, 0.00000570, -0.00000714, 0.00000701, -0.00001723, 0.00001600, -0.00002474, 0.00002210, 0.00000758, -0.00000415, -0.00002455, 0.00002043, 0.00000439, 0.00000070, -0.00001484, 0.00000293, 0.00001136, },
		{ -0.00001153, 0.00000669, -0.00000567, 0.00000757, -0.00001338, 0.00001645, -0.00000848, -0.00000805, -0.00000078, 0.00002324, -0.00001454, -0.00001430, -0.00000017, 0.00002689, -0.00001445, 0.00000357, -0.00000632, 0.00000013, -0.00000522, 0.00000678, -0.00000525, 0.00001120, -0.00000995, 0.00000411, -0.00000426, 0.00000312, -0.00000413, 0.00000530, -0.00001171, 0.00001471, -0.00001189, -0.00000029, 0.00000878, -0.00002318, 0.00002240, -0.00000845, 0.00001533, -0.00000714, 0.00002095, -0.00003225, 0.00001555, -0.00000954, -0.00000314, 0.00001619, -0.00001481, 0.00000681, -0.00001184, 0.00001558, -0.00001640, 0.00001097, -0.00002025, 0.00002619, -0.00002385, 0.00001815, -0.00002220, 0.00002021, -0.00003097, 0.00002896, -0.00004589, 0.00004027, -0.00005370, 0.00004680, -0.00008934, 0.00004373, -0.00019030, -0.00027084, 0.00124372, -0.00064824, 0.00011158, -0.00014692, 0.00006957, -0.00007173, 0.00004114, -0.00004758, 0.00003711, -0.00004081, 0.00002840, -0.00002995, 0.00002068, -0.00002357, 0.00001791, -0.00001232, 0.00001387, -0.00002586, 0.00002022, -0.00001375, 0.00001720, -0.00002347, 0.00001254, 0.00000746, -0.00001347, 0.00001464, -0.00000435, -0.00002234, 0.00001706, -0.00001685, 0.00002184, -0.00001936, 0.00002111, -0.00002408, 0.00000223, 0.00000589, -0.00000862, 0.00001060, 0.00000030, -0.00000385, 0.00000688, -0.00000664, 0.00000072, -0.00000376, 0.00000544, 0.00000140, 0.00000416, -0.00000635, 0.00000733, -0.00001626, 0.00001248, -0.00001733, 0.00000849, 0.00001696, -0.00000369, -0.00002140, 0.00000812, 0.00001457, -0.00000366, -0.00000652, -0.00000958, 0.00001927, },
		{ -0.00000722, 0.00000301, -0.00000339, 0.00000561, -0.00000929, 0.00000743, 0.00000297, -0.00001416, -0.00000058, 0.00001765, -0.00000125, -0.00002196, -0.00000051, 0.00002029, -0.00000021, -0.00000898, 0.00000472, -0.00000961, 0.00000358, -0.00000344, 0.00000563, -0.00000140, 0.00000541, -0.00001196, 0.00001116, -0.00001163, 0.00001036, -0.00000975, 0.00000570, -0.00000826, 0.00001610, -0.00002991, 0.00003535, -0.00004273, 0.00003438, -0.00001934, 0.00002775, -0.00002019, 0.00003457, -0.00003706, 0.00001601, -0.00001172, 0.00000062, 0.00000627, 0.00000104, -0.00000995, 0.00000520, -0.00000475, 0.00000713, -0.00001320, 0.00000575, -0.00000592, 0.00001375, -0.00002145, 0.00001951, -0.00002389, 0.00001787, -0.00002583, 0.00001857, -0.00003664, 0.00003754, -0.00006026, 0.00004984, -0.00013233, 0.00009325, -0.00061853, 0.00123280, -0.00028658, -0.00017559, 0.00002923, -0.00007109, 0.00003847, -0.00005025, 0.00002775, -0.00002951, 0.00002010, -0.00002760, 0.00002091, -0.00002562, 0.00001825, -0.00002091, 0.00002124, -0.00001479, 0.00000324, -0.00001102, 0.00001535, -0.00000985, 0.00000651, -0.00001995, 0.00003497, -0.00003192, 0.00002467, -0.00000832, -0.00001416, 0.00000258, -0.00000257, 0.00000536, -0.00000054, -0.00000035, 0.00000239, -0.00002413, 0.00002444, -0.00002036, 0.00001628, -0.00000258, 0.00000086, 0.00000048, 0.00000145, -0.00000653, 0.00000176, -0.00000035, 0.00000588, 0.00000156, -0.00000338, 0.00000419, -0.00001068, 0.00000406, -0.00000584, -0.00000596, 0.00002422, -0.00000307, -0.00001562, -0.00000479, 0.00002227, -0.00000600, 0.00000015, -0.00001737, 0.00001990, },
		{ 0.00000175, -0.00000308, 0.00000100, 0.00000093, -0.00000221, -0.00000291, 0.00001241, -0.00001685, -0.00000136, 0.00001121, 0.00001035, -0.00002582, -0.00000181, 0.00001237, 0.00001296, -0.00001802, 0.00001184, -0.00001573, 0.00000956, -0.00001134, 0.00001416, -0.00001140, 0.00001715, -0.00002299, 0.00002054, -0.00002017, 0.00001906, -0.00001952, 0.00001799, -0.00002451, 0.00003351, -0.00004410, 0.00004266, -0.00004230, 0.00002876, -0.00001569, 0.00002695, -0.00001991, 0.00003379, -0.00002844, 0.00000728, -0.00000774, 0.00000002, -0.00000006, 0.00001203, -0.00002065, 0.00001634, -0.00001913, 0.00002374, -0.00002994, 0.00002459, -0.00002997, 0.00004063, -0.00004776, 0.00004629, -0.00005202, 0.00004974, -0.00006229, 0.00006230, -0.00008816, 0.00009621, -0.00012609, 0.00013344, -0.00022885, 0.00023544, -0.00070068, 0.00091475, 0.00023821, -0.00048013, 0.00018605, -0.00019079, 0.00012701, -0.00012160, 0.00008558, -0.00008155, 0.00006788, -0.00007083, 0.00005922, -0.00005968, 0.00004842, -0.00004867, 0.00004482, -0.00003585, 0.00002670, -0.00003645, 0.00003801, -0.00003100, 0.00003024, -0.00004343, 0.00005086, -0.00003903, 0.00002643, -0.00000913, -0.00000541, -0.00001216, 0.00001128, -0.00001074, 0.00001714, -0.00001904, 0.00002248, -0.00003927, 0.00002962, -0.00002009, 0.00001299, -0.00000029, 0.00000286, -0.00000374, 0.00000691, -0.00001055, 0.00000435, -0.00000373, 0.00000832, -0.00000002, -0.00000029, 0.00000003, -0.00000372, -0.00000506, 0.00000467, -0.00001637, 0.00002605, -0.00000010, -0.00001022, -0.00001445, 0.00002438, -0.00000405, 0.00000270, -0.00001842, 0.00001324, },
		{ 0.00001118, -0.00000779, 0.00000421, -0.00000298, 0.00000355, -0.00000968, 0.00001568, -0.00001367, -0.00000492, 0.00000667, 0.00001708, -0.00002370, -0.00000595, 0.00000639, 0.00002086, -0.00001985, 0.00001208, -0.00001556, 0.00000982, -0.00001339, 0.00001616, -0.00001402, 0.00002011, -0.00002381, 0.00001909, -0.00001807, 0.00001743, -0.00001881, 0.00001885, -0.00002666, 0.00003290, -0.00003709, 0.00002807, -0.00002252, 0.00000804, -0.00000021, 0.00001495, -0.00000796, 0.00002074, -0.00000941, -0.00000802, 0.00000147, -0.00000580, -0.00000025, 0.00001440, -0.00002105, 0.00001667, -0.00002161, 0.00002658, -0.00003163, 0.00002740, -0.00003556, 0.00004535, -0.00004906, 0.00004620, -0.00005162, 0.00005089, -0.00006476, 0.00006713, -0.00009333, 0.00009891, -0.00012480, 0.00013294, -0.00021543, 0.00021532, -0.00054913, 0.00041179, 0.00075675, -0.00066258, 0.00023222, -0.00022621, 0.00014580, -0.00013488, 0.00009465, -0.00009196, 0.00007816, -0.00007930, 0.00006526, -0.00006395, 0.00005136, -0.00005154, 0.00004604, -0.00003766, 0.00003266, -0.00004404, 0.00004263, -0.00003501, 0.00003652, -0.00004745, 0.00004679, -0.00002888, 0.00001551, -0.00000208, -0.00000262, -0.00001912, 0.00001617, -0.00001747, 0.00002406, -0.00002494, 0.00002678, -0.00003622, 0.00001808, -0.00000743, 0.00000176, 0.00000628, 0.00000183, -0.00000438, 0.00000790, -0.00000973, 0.00000280, -0.00000367, 0.00000778, 0.00000044, 0.00000153, -0.00000308, 0.00000173, -0.00001135, 0.00001053, -0.00001994, 0.00002140, 0.00000565, -0.00000667, -0.00001907, 0.00002033, 0.00000192, 0.00000140, -0.00001399, 0.00000247, },
		{ 0.00001667, -0.00000797, 0.00000390, -0.00000373, 0.00000526, -0.00001045, 0.00001187, -0.00000533, -0.00001050, 0.00000436, 0.00001847, -0.00001619, -0.00001238, 0.00000317, 0.00002235, -0.00001435, 0.00000605, -0.00000972, 0.00000463, -0.00000933, 0.00001106, -0.00000859, 0.00001382, -0.00001454, 0.00000771, -0.00000669, 0.00000676, -0.00000854, 0.00000878, -0.00001541, 0.00001656, -0.00001402, -0.00000001, 0.00000606, -0.00001704, 0.00001768, -0.00000016, 0.00000820, 0.00000309, 0.00001219, -0.00002324, 0.00001123, -0.00001366, 0.00000358, 0.00000984, -0.00001284, 0.00000768, -0.00001323, 0.00001653, -0.00001914, 0.00001472, -0.00002294, 0.00002873, -0.00002731, 0.00002221, -0.00002619, 0.00002495, -0.00003687, 0.00003680, -0.00005681, 0.00005264, -0.00006697, 0.00006437, -0.00011807, 0.00008374, -0.00027311, -0.00010223, 0.00109848, -0.00062060, 0.00012305, -0.00015571, 0.00008223, -0.00008246, 0.00004971, -0.00005633, 0.00004686, -0.00004963, 0.00003658, -0.00003685, 0.00002616, -0.00002901, 0.00002450, -0.00001926, 0.00001921, -0.00003139, 0.00002705, -0.00001992, 0.00002345, -0.00003113, 0.00002376, -0.00000394, -0.00000559, 0.00001190, -0.00000715, -0.00001579, 0.00000962, -0.00001232, 0.00001770, -0.00001616, 0.00001511, -0.00001785, -0.00000434, 0.00001066, -0.00001113, 0.00001269, 0.00000020, -0.00000264, 0.00000537, -0.00000526, -0.00000151, -0.00000137, 0.00000519, 0.00000238, 0.00000197, -0.00000431, 0.00000429, -0.00001328, 0.00001069, -0.00001690, 0.00001196, 0.00001257, -0.00000456, -0.00001915, 0.00001199, 0.00000952, -0.00000143, -0.00000736, -0.00000780, },
		{ 0.00001590, -0.00000303, 0.00000022, -0.00000142, 0.00000308, -0.00000627, 0.00000363, 0.00000451, -0.00001513, 0.00000255, 0.00001645, -0.00000620, -0.00001860, 0.00000135, 0.00001918, -0.00000453, -0.00000277, -0.00000152, -0.00000306, -0.00000195, 0.00000182, 0.00000147, 0.00000237, -0.00000021, -0.00000775, 0.00000774, -0.00000679, 0.00000521, -0.00000580, 0.00000164, -0.00000568, 0.00001254, -0.00002702, 0.00002880, -0.00003360, 0.00002738, -0.00000938, 0.00001998, -0.00001033, 0.00002771, -0.00003125, 0.00001597, -0.00001856, 0.00000664, 0.00000347, -0.00000184, -0.00000447, -0.00000040, 0.00000067, -0.00000030, -0.00000504, -0.00000145, 0.00000186, 0.00000436, -0.00001095, 0.00000836, -0.00001105, 0.00000291, -0.00000802, -0.00000283, -0.00001308, 0.00001114, -0.00002674, 0.00000332, -0.00007261, -0.00000424, -0.00047137, 0.00115803, -0.00033616, -0.00011271, -0.00001348, -0.00003299, 0.00000834, -0.00002548, 0.00000500, -0.00000810, 0.00000163, -0.00001113, 0.00000673, -0.00001323, 0.00000607, -0.00000822, 0.00000930, -0.00000507, -0.00000653, -0.00000054, 0.00000613, -0.00000082, -0.00000343, -0.00000809, 0.00002549, -0.00002790, 0.00002619, -0.00001447, -0.00000629, -0.00000375, -0.00000017, 0.00000339, 0.00000080, -0.00000408, 0.00000518, -0.00002598, 0.00002370, -0.00001778, 0.00001381, 0.00000106, -0.00000090, 0.00000186, -0.00000003, -0.00000577, 0.00000079, 0.00000246, 0.00000440, 0.00000186, -0.00000406, 0.00000423, -0.00001146, 0.00000654, -0.00000987, 0.00000114, 0.00001803, -0.00000246, -0.00001656, 0.00000239, 0.00001569, -0.00000308, -0.00000185, -0.00001397, },
		{ 0.00000963, 0.00000474, -0.00000433, 0.00000171, -0.00000052, -0.00000058, -0.00000445, 0.00001126, -0.00001560, -0.00000088, 0.00001370, 0.00000280, -0.00002178, -0.00000118, 0.00001433, 0.00000544, -0.00001004, 0.00000501, -0.00000938, 0.00000474, -0.00000699, 0.00001085, -0.00000807, 0.00001219, -0.00001975, 0.00001767, -0.00001583, 0.00001492, -0.00001667, 0.00001485, -0.00002227, 0.00002957, -0.00004008, 0.00003479, -0.00003365, 0.00002326, -0.00000799, 0.00002242, -0.00001411, 0.00003204, -0.00002815, 0.00001249, -0.00001702, 0.00000482, 0.00000033, 0.00000584, -0.00001302, 0.00000936, -0.00001230, 0.00001501, -0.00002100, 0.00001655, -0.00002083, 0.00002959, -0.00003564, 0.00003339, -0.00003726, 0.00003286, -0.00004276, 0.00003989, -0.00006414, 0.00006924, -0.00009225, 0.00008920, -0.00017608, 0.00015897, -0.00061301, 0.00093269, 0.00011319, -0.00038040, 0.00012494, -0.00014101, 0.00008947, -0.00009106, 0.00005830, -0.00005712, 0.00004741, -0.00005287, 0.00004379, -0.00004589, 0.00003480, -0.00003508, 0.00003279, -0.00002623, 0.00001685, -0.00002630, 0.00002952, -0.00002283, 0.00002168, -0.00003446, 0.00004615, -0.00004033, 0.00003214, -0.00001728, 0.00000169, -0.00001543, 0.00001016, -0.00000950, 0.00001553, -0.00001920, 0.00002060, -0.00003577, 0.00002372, -0.00001352, 0.00000740, 0.00000556, -0.00000060, -0.00000043, 0.00000339, -0.00000756, 0.00000093, 0.00000110, 0.00000530, 0.00000210, -0.00000315, 0.00000267, -0.00000765, 0.00000069, -0.00000226, -0.00000761, 0.00001988, 0.00000077, -0.00001313, -0.00000582, 0.00001830, -0.00000203, 0.00000083, -0.00001487, },
		{ 0.00000099, 0.00001149, -0.00000649, 0.00000295, -0.00000263, 0.00000303, -0.00000846, 0.00001192, -0.00001063, -0.00000652, 0.00001170, 0.00000867, -0.00002049, -0.00000550, 0.00001006, 0.00001235, -0.00001281, 0.00000728, -0.00001173, 0.00000770, -0.00001163, 0.00001516, -0.00001260, 0.00001735, -0.00002301, 0.00001824, -0.00001586, 0.00001590, -0.00001848, 0.00001787, -0.00002603, 0.00003024, -0.00003428, 0.00002226, -0.00001820, 0.00000761, 0.00000202, 0.00001651, -0.00000851, 0.00002565, -0.00001519, 0.00000192, -0.00000911, -0.00000270, 0.00000235, 0.00000725, -0.00001431, 0.00001153, -0.00001675, 0.00002020, -0.00002571, 0.00002242, -0.00002924, 0.00003742, -0.00004066, 0.00003752, -0.00004162, 0.00003962, -0.00005201, 0.00005312, -0.00007926, 0.00008332, -0.00010502, 0.00010830, -0.00019047, 0.00018343, -0.00053748, 0.00051664, 0.00058411, -0.00055875, 0.00017942, -0.00018566, 0.00011778, -0.00011269, 0.00007505, -0.00007514, 0.00006490, -0.00006788, 0.00005572, -0.00005526, 0.00004245, -0.00004289, 0.00003922, -0.00003280, 0.00002702, -0.00003822, 0.00003853, -0.00003111, 0.00003254, -0.00004415, 0.00004856, -0.00003575, 0.00002437, -0.00001051, 0.00000202, -0.00001842, 0.00001155, -0.00001352, 0.00002006, -0.00002214, 0.00002129, -0.00002931, 0.00000984, -0.00000001, -0.00000392, 0.00001166, -0.00000112, -0.00000100, 0.00000409, -0.00000629, -0.00000117, 0.00000126, 0.00000491, 0.00000292, -0.00000214, 0.00000071, -0.00000361, -0.00000450, 0.00000330, -0.00001229, 0.00001757, 0.00000512, -0.00000976, -0.00001135, 0.00001701, 0.00000140, 0.00000105, -0.00001187, },
		{ -0.00000639, 0.00001399, -0.00000431, 0.00000103, -0.00000190, 0.00000294, -0.00000723, 0.00000678, -0.00000195, -0.00001272, 0.00001001, 0.00001142, -0.00001550, -0.00001091, 0.00000692, 0.00001525, -0.00001072, 0.00000525, -0.00000986, 0.00000624, -0.00001092, 0.00001301, -0.00000982, 0.00001409, -0.00001692, 0.00000964, -0.00000750, 0.00000864, -0.00001131, 0.00001058, -0.00001726, 0.00001654, -0.00001444, -0.00000118, 0.00000373, -0.00001096, 0.00001360, 0.00000771, 0.00000172, 0.00001355, 0.00000221, -0.00001104, 0.00000158, -0.00001309, 0.00000755, 0.00000366, -0.00000924, 0.00000649, -0.00001241, 0.00001471, -0.00001847, 0.00001506, -0.00002201, 0.00002693, -0.00002606, 0.00002151, -0.00002510, 0.00002383, -0.00003582, 0.00003643, -0.00005822, 0.00005457, -0.00006828, 0.00006694, -0.00012823, 0.00010050, -0.00032824, 0.00005726, 0.00092464, -0.00055327, 0.00010631, -0.00014374, 0.00007822, -0.00008003, 0.00004728, -0.00005502, 0.00004738, -0.00005035, 0.00003784, -0.00003773, 0.00002618, -0.00002916, 0.00002617, -0.00002189, 0.00002023, -0.00003217, 0.00002936, -0.00002204, 0.00002558, -0.00003444, 0.00003187, -0.00001505, 0.00000449, 0.00000501, -0.00000595, -0.00001150, 0.00000292, -0.00000747, 0.00001318, -0.00001249, 0.00000874, -0.00001098, -0.00001108, 0.00001531, -0.00001387, 0.00001513, -0.00000033, -0.00000089, 0.00000300, -0.00000318, -0.00000415, 0.00000183, 0.00000405, 0.00000383, -0.00000102, -0.00000109, -0.00000040, -0.00000778, 0.00000580, -0.00001283, 0.00001223, 0.00000946, -0.00000635, -0.00001426, 0.00001294, 0.00000565, 0.00000039, -0.00000751, },
		{ -0.00001024, 0.00001132, 0.00000157, -0.00000291, 0.00000073, 0.00000021, -0.00000266, -0.00000086, 0.00000645, -0.00001635, 0.00000692, 0.00001260, -0.00000910, -0.00001545, 0.00000398, 0.00001514, -0.00000558, 0.00000096, -0.00000559, 0.00000192, -0.00000636, 0.00000619, -0.00000204, 0.00000538, -0.00000536, -0.00000346, 0.00000428, -0.00000202, -0.00000004, -0.00000176, -0.00000249, -0.00000287, 0.00000840, -0.00002314, 0.00002034, -0.00002269, 0.00001926, 0.00000206, 0.00001074, 0.00000205, 0.00001748, -0.00002059, 0.00001013, -0.00002166, 0.00001160, -0.00000084, -0.00000201, -0.00000158, -0.00000355, 0.00000331, -0.00000466, 0.00000030, -0.00000576, 0.00000626, -0.00000161, -0.00000372, 0.00000072, -0.00000237, -0.00000713, 0.00000418, -0.00001812, 0.00000432, -0.00000853, -0.00000143, -0.00003417, -0.00002197, -0.00009629, -0.00030500, 0.00103303, -0.00033888, -0.00007683, -0.00003852, -0.00000812, -0.00001198, -0.00000873, -0.00001044, 0.00000683, -0.00001156, 0.00000111, -0.00000389, -0.00000411, -0.00000287, 0.00000156, -0.00000021, 0.00000205, -0.00001342, 0.00000756, -0.00000131, 0.00000653, -0.00001197, 0.00000417, 0.00001292, -0.00001919, 0.00002249, -0.00001692, 0.00000030, -0.00001040, 0.00000333, 0.00000055, 0.00000295, -0.00000849, 0.00000880, -0.00002822, 0.00002339, -0.00001613, 0.00001240, 0.00000352, -0.00000137, 0.00000173, -0.00000023, -0.00000608, 0.00000136, 0.00000372, 0.00000414, 0.00000044, -0.00000259, 0.00000170, -0.00000903, 0.00000565, -0.00001055, 0.00000589, 0.00001237, -0.00000252, -0.00001520, 0.00000768, 0.00000907, 0.00000031, -0.00000388, },
		{ -0.00001047, 0.00000515, 0.00000828, -0.00000607, 0.00000278, -0.00000251, 0.00000175, -0.00000669, 0.00001055, -0.00001488, 0.00000117, 0.00001376, -0.00000353, -0.00001734, 0.00000016, 0.00001372, 0.00000003, -0.00000298, -0.00000138, -0.00000278, -0.00000078, -0.00000174, 0.00000643, -0.00000371, 0.00000582, -0.00001465, 0.00001313, -0.00000994, 0.00000903, -0.00001205, 0.00000988, -0.00001792, 0.00002319, -0.00003324, 0.00002356, -0.00002221, 0.00001557, 0.00000245, 0.00001510, -0.00000449, 0.00002602, -0.00002281, 0.00001288, -0.00002451, 0.00001063, -0.00000219, 0.00000282, -0.00000770, 0.00000418, -0.00000731, 0.00000802, -0.00001324, 0.00000961, -0.00001295, 0.00001953, -0.00002412, 0.00002119, -0.00002353, 0.00001723, -0.00002444, 0.00001826, -0.00004002, 0.00004172, -0.00005772, 0.00004256, -0.00011738, 0.00006812, -0.00048649, 0.00089298, 0.00002778, -0.00029559, 0.00007090, -0.00009549, 0.00005412, -0.00006212, 0.00003262, -0.00003381, 0.00002724, -0.00003463, 0.00002795, -0.00003180, 0.00002110, -0.00002125, 0.00002027, -0.00001642, 0.00000701, -0.00001550, 0.00001981, -0.00001333, 0.00001127, -0.00002181, 0.00003578, -0.00003572, 0.00003285, -0.00002298, 0.00000896, -0.00002009, 0.00001069, -0.00000935, 0.00001483, -0.00002011, 0.00001962, -0.00003286, 0.00001871, -0.00000839, 0.00000329, 0.00000990, -0.00000252, 0.00000124, 0.00000119, -0.00000583, -0.00000074, 0.00000429, 0.00000367, 0.00000224, -0.00000371, 0.00000278, -0.00000872, 0.00000388, -0.00000712, 0.00000032, 0.00001305, 0.00000164, -0.00001462, 0.00000236, 0.00001086, 0.00000126, -0.00000166, },
		{ -0.00000865, -0.00000143, 0.00001236, -0.00000575, 0.00000214, -0.00000291, 0.00000319, -0.00000784, 0.00000856, -0.00000821, -0.00000665, 0.00001501, 0.00000035, -0.00001615, -0.00000470, 0.00001203, 0.00000437, -0.00000489, 0.00000115, -0.00000593, 0.00000330, -0.00000752, 0.00001172, -0.00000886, 0.00001192, -0.00001927, 0.00001482, -0.00001122, 0.00001173, -0.00001547, 0.00001419, -0.00002244, 0.00002440, -0.00002813, 0.00001297, -0.00001126, 0.00000492, 0.00000720, 0.00001508, -0.00000526, 0.00002685, -0.00001717, 0.00000918, -0.00002047, 0.00000342, 0.00000120, 0.00000296, -0.00000892, 0.00000697, -0.00001226, 0.00001378, -0.00001900, 0.00001645, -0.00002188, 0.00002790, -0.00003018, 0.00002672, -0.00002953, 0.00002591, -0.00003612, 0.00003488, -0.00005947, 0.00006094, -0.00007718, 0.00007238, -0.00014903, 0.00012564, -0.00048081, 0.00057225, 0.00043124, -0.00045055, 0.00011941, -0.00013840, 0.00008314, -0.00008489, 0.00005077, -0.00005372, 0.00004681, -0.00005173, 0.00004193, -0.00004287, 0.00003030, -0.00003100, 0.00002915, -0.00002514, 0.00001867, -0.00002903, 0.00003069, -0.00002351, 0.00002439, -0.00003521, 0.00004381, -0.00003724, 0.00002990, -0.00001856, 0.00000851, -0.00002007, 0.00000888, -0.00001062, 0.00001685, -0.00002002, 0.00001669, -0.00002309, 0.00000273, 0.00000587, -0.00000824, 0.00001575, -0.00000289, 0.00000119, 0.00000113, -0.00000374, -0.00000380, 0.00000507, 0.00000297, 0.00000388, -0.00000416, 0.00000287, -0.00000733, 0.00000136, -0.00000369, -0.00000367, 0.00001181, 0.00000537, -0.00001249, -0.00000263, 0.00001125, 0.00000263, -0.00000023, },
		{ -0.00000661, -0.00000576, 0.00001178, -0.00000120, -0.00000139, -0.00000065, 0.00000118, -0.00000443, 0.00000197, 0.00000092, -0.00001386, 0.00001492, 0.00000331, -0.00001307, -0.00000945, 0.00000995, 0.00000711, -0.00000470, 0.00000187, -0.00000700, 0.00000480, -0.00000959, 0.00001211, -0.00000841, 0.00001147, -0.00001639, 0.00000917, -0.00000603, 0.00000803, -0.00001165, 0.00000996, -0.00001655, 0.00001392, -0.00001233, -0.00000469, 0.00000283, -0.00000632, 0.00001174, 0.00001348, -0.00000219, 0.00002200, -0.00000619, 0.00000124, -0.00001141, -0.00000808, 0.00000769, -0.00000060, -0.00000567, 0.00000454, -0.00001057, 0.00001132, -0.00001549, 0.00001292, -0.00001844, 0.00002186, -0.00002109, 0.00001700, -0.00001998, 0.00001798, -0.00002901, 0.00002894, -0.00005096, 0.00004675, -0.00005829, 0.00005474, -0.00011884, 0.00009046, -0.00034573, 0.00018836, 0.00074408, -0.00046146, 0.00006985, -0.00011638, 0.00006115, -0.00006698, 0.00003589, -0.00004532, 0.00003987, -0.00004377, 0.00003276, -0.00003310, 0.00002125, -0.00002444, 0.00002312, -0.00002033, 0.00001709, -0.00002824, 0.00002687, -0.00001968, 0.00002293, -0.00003212, 0.00003465, -0.00002305, 0.00001399, -0.00000398, -0.00000135, -0.00001011, -0.00000192, -0.00000336, 0.00000919, -0.00000933, 0.00000309, -0.00000470, -0.00001681, 0.00001886, -0.00001591, 0.00001705, -0.00000052, 0.00000057, 0.00000062, -0.00000125, -0.00000626, 0.00000486, 0.00000294, 0.00000469, -0.00000354, 0.00000191, -0.00000511, -0.00000144, -0.00000067, -0.00000624, 0.00000968, 0.00000767, -0.00000867, -0.00000728, 0.00001091, 0.00000351, 0.00000132, },
		{ -0.00000519, -0.00000696, 0.00000700, 0.00000577, -0.00000574, 0.00000248, -0.00000236, 0.00000081, -0.00000531, 0.00000819, -0.00001712, 0.00001160, 0.00000678, -0.00000997, -0.00001250, 0.00000674, 0.00000893, -0.00000348, 0.00000172, -0.00000660, 0.00000408, -0.00000832, 0.00000839, -0.00000366, 0.00000635, -0.00000867, -0.00000039, 0.00000200, 0.00000136, -0.00000404, 0.00000114, -0.00000544, -0.00000110, 0.00000509, -0.00001976, 0.00001166, -0.00001196, 0.00001205, 0.00001300, 0.00000217, 0.00001472, 0.00000631, -0.00000726, -0.00000094, -0.00001985, 0.00001355, -0.00000464, -0.00000088, -0.00000048, -0.00000482, 0.00000360, -0.00000611, 0.00000276, -0.00000707, 0.00000710, -0.00000375, -0.00000032, -0.00000278, 0.00000138, -0.00001126, 0.00000935, -0.00002531, 0.00001308, -0.00001843, 0.00001149, -0.00005640, 0.00001010, -0.00016755, -0.00014058, 0.00087613, -0.00030303, -0.00006642, -0.00004536, 0.00000224, -0.00002095, -0.00000147, -0.00001709, 0.00001369, -0.00001779, 0.00000752, -0.00000948, 0.00000046, -0.00000728, 0.00000703, -0.00000593, 0.00000537, -0.00001609, 0.00001168, -0.00000529, 0.00001035, -0.00001676, 0.00001392, 0.00000013, -0.00000797, 0.00001452, -0.00001505, 0.00000407, -0.00001574, 0.00000666, -0.00000218, 0.00000493, -0.00001252, 0.00001224, -0.00003001, 0.00002283, -0.00001482, 0.00001153, 0.00000529, -0.00000114, 0.00000070, 0.00000021, -0.00000679, 0.00000280, 0.00000408, 0.00000431, -0.00000175, 0.00000004, -0.00000232, -0.00000430, 0.00000191, -0.00000790, 0.00000762, 0.00000803, -0.00000355, -0.00001130, 0.00001016, 0.00000360, 0.00000324, },
		{ -0.00000396, -0.00000605, 0.00000061, 0.00001171, -0.00000782, 0.00000380, -0.00000459, 0.00000432, -0.00000923, 0.00001007, -0.00001441, 0.00000439, 0.00001134, -0.00000798, -0.00001298, 0.00000209, 0.00001045, -0.00000230, 0.00000169, -0.00000560, 0.00000218, -0.00000521, 0.00000278, 0.00000244, 0.00000015, -0.00000040, -0.00000912, 0.00000823, -0.00000390, 0.00000276, -0.00000689, 0.00000436, -0.00001282, 0.00001580, -0.00002512, 0.00001064, -0.00000978, 0.00000736, 0.00001411, 0.00000649, 0.00000754, 0.00001702, -0.00001306, 0.00000732, -0.00002781, 0.00001500, -0.00000578, 0.00000210, -0.00000457, 0.00000096, -0.00000452, 0.00000345, -0.00000761, 0.00000478, -0.00000757, 0.00001196, -0.00001496, 0.00001159, -0.00001302, 0.00000542, -0.00001048, 0.00000127, -0.00002025, 0.00001887, -0.00002904, 0.00000213, -0.00006407, -0.00002185, -0.00033785, 0.00080490, -0.00001599, -0.00023261, 0.00003063, -0.00006023, 0.00002621, -0.00003928, 0.00001262, -0.00001545, 0.00001092, -0.00001933, 0.00001456, -0.00001999, 0.00000981, -0.00000962, 0.00000954, -0.00000853, -0.00000065, -0.00000638, 0.00001115, -0.00000483, 0.00000162, -0.00000819, 0.00002186, -0.00002614, 0.00002815, -0.00002451, 0.00001430, -0.00002450, 0.00001189, -0.00000964, 0.00001439, -0.00002116, 0.00001906, -0.00003028, 0.00001451, -0.00000461, 0.00000059, 0.00001277, -0.00000306, 0.00000142, 0.00000018, -0.00000515, -0.00000085, 0.00000598, 0.00000315, 0.00000076, -0.00000224, 0.00000055, -0.00000683, 0.00000389, -0.00000881, 0.00000590, 0.00000694, 0.00000170, -0.00001382, 0.00000861, 0.00000353, 0.00000480, },
		{ -0.00000193, -0.00000489, -0.00000431, 0.00001354, -0.00000546, 0.00000172, -0.00000370, 0.00000398, -0.00000790, 0.00000584, -0.00000659, -0.00000509, 0.00001589, -0.00000669, -0.00001148, -0.00000321, 0.00001151, -0.00000142, 0.00000196, -0.00000435, -0.00000007, -0.00000176, -0.00000244, 0.00000707, -0.00000398, 0.00000497, -0.00001359, 0.00000969, -0.00000506, 0.00000578, -0.00001055, 0.00000857, -0.00001676, 0.00001630, -0.00001958, 0.00000118, -0.00000263, 0.00000061, 0.00001494, 0.00001096, 0.00000171, 0.00002409, -0.00001428, 0.00001115, -0.00002945, 0.00001013, -0.00000246, 0.00000146, -0.00000548, 0.00000375, -0.00000908, 0.00000855, -0.00001294, 0.00001100, -0.00001518, 0.00001876, -0.00001980, 0.00001612, -0.00001791, 0.00001277, -0.00002064, 0.00001649, -0.00003849, 0.00003688, -0.00004746, 0.00003274, -0.00010009, 0.00005222, -0.00038686, 0.00057616, 0.00031057, -0.00035171, 0.00006292, -0.00009294, 0.00004885, -0.00005721, 0.00002686, -0.00003227, 0.00002794, -0.00003441, 0.00002699, -0.00002956, 0.00001758, -0.00001841, 0.00001812, -0.00001676, 0.00000979, -0.00001868, 0.00002113, -0.00001415, 0.00001402, -0.00002231, 0.00003326, -0.00003277, 0.00003038, -0.00002380, 0.00001470, -0.00002284, 0.00000774, -0.00000866, 0.00001435, -0.00001858, 0.00001311, -0.00001792, -0.00000279, 0.00000981, -0.00001079, 0.00001819, -0.00000320, 0.00000191, -0.00000065, -0.00000239, -0.00000474, 0.00000740, 0.00000222, 0.00000304, -0.00000405, 0.00000273, -0.00000838, 0.00000474, -0.00000863, 0.00000399, 0.00000566, 0.00000548, -0.00001377, 0.00000551, 0.00000434, 0.00000489, },
		{ 0.00000138, -0.00000472, -0.00000596, 0.00001028, 0.00000104, -0.00000306, -0.00000022, 0.00000029, -0.00000251, -0.00000199, 0.00000277, -0.00001334, 0.00001801, -0.00000451, -0.00000969, -0.00000754, 0.00001117, -0.00000023, 0.00000189, -0.00000256, -0.00000241, 0.00000116, -0.00000593, 0.00000884, -0.00000476, 0.00000630, -0.00001309, 0.00000636, -0.00000239, 0.00000497, -0.00000948, 0.00000689, -0.00001332, 0.00000875, -0.00000768, -0.00001065, 0.00000367, -0.00000380, 0.00001302, 0.00001616, -0.00000225, 0.00002687, -0.00001066, 0.00001015, -0.00002458, -0.00000037, 0.00000434, -0.00000220, -0.00000325, 0.00000289, -0.00000882, 0.00000766, -0.00001154, 0.00000964, -0.00001372, 0.00001517, -0.00001409, 0.00001044, -0.00001276, 0.00000963, -0.00001896, 0.00001710, -0.00003791, 0.00003224, -0.00004050, 0.00003166, -0.00009365, 0.00005625, -0.00032295, 0.00027800, 0.00057626, -0.00036142, 0.00002445, -0.00008138, 0.00003689, -0.00004791, 0.00001949, -0.00003067, 0.00002723, -0.00003234, 0.00002338, -0.00002476, 0.00001311, -0.00001646, 0.00001673, -0.00001581, 0.00001112, -0.00002085, 0.00002060, -0.00001364, 0.00001617, -0.00002431, 0.00003115, -0.00002594, 0.00002032, -0.00001246, 0.00000499, -0.00001121, -0.00000457, -0.00000048, 0.00000615, -0.00000716, -0.00000120, 0.00000022, -0.00002077, 0.00002064, -0.00001660, 0.00001785, 0.00000018, 0.00000114, -0.00000111, -0.00000010, -0.00000717, 0.00000703, 0.00000251, 0.00000421, -0.00000458, 0.00000346, -0.00000828, 0.00000388, -0.00000680, 0.00000111, 0.00000544, 0.00000661, -0.00001067, 0.00000061, 0.00000659, 0.00000300, },
		{ 0.00000516, -0.00000533, -0.00000467, 0.00000360, 0.00000891, -0.00000773, 0.00000337, -0.00000409, 0.00000348, -0.00000903, 0.00000911, -0.00001674, 0.00001547, 0.00000007, -0.00000925, -0.00000947, 0.00000861, 0.00000201, 0.00000070, 0.00000014, -0.00000477, 0.00000324, -0.00000743, 0.00000788, -0.00000283, 0.00000469, -0.00000936, 0.00000062, 0.00000165, 0.00000253, -0.00000588, 0.00000205, -0.00000640, -0.00000126, 0.00000361, -0.00001806, 0.00000420, -0.00000317, 0.00000760, 0.00002155, -0.00000364, 0.00002536, -0.00000295, 0.00000538, -0.00001494, -0.00001365, 0.00001169, -0.00000647, 0.00000018, -0.00000012, -0.00000514, 0.00000248, -0.00000550, 0.00000306, -0.00000612, 0.00000514, -0.00000270, -0.00000020, -0.00000272, 0.00000089, -0.00001015, 0.00000805, -0.00002466, 0.00001318, -0.00001836, 0.00001130, -0.00006126, 0.00001881, -0.00020707, 0.00000255, 0.00070808, -0.00024191, -0.00007499, -0.00003775, 0.00000039, -0.00002004, -0.00000253, -0.00001589, 0.00001306, -0.00001736, 0.00000819, -0.00001011, 0.00000057, -0.00000722, 0.00000809, -0.00000771, 0.00000483, -0.00001437, 0.00001140, -0.00000520, 0.00000983, -0.00001641, 0.00001882, -0.00000997, 0.00000294, 0.00000443, -0.00000963, 0.00000450, -0.00001879, 0.00000895, -0.00000411, 0.00000598, -0.00001533, 0.00001460, -0.00003056, 0.00002140, -0.00001319, 0.00001056, 0.00000695, -0.00000087, -0.00000054, 0.00000063, -0.00000716, 0.00000433, 0.00000428, 0.00000400, -0.00000354, 0.00000249, -0.00000632, 0.00000113, -0.00000326, -0.00000292, 0.00000667, 0.00000510, -0.00000523, -0.00000531, 0.00000969, -0.00000007, },
		{ 0.00000765, -0.00000532, -0.00000242, -0.00000316, 0.00001431, -0.00000905, 0.00000431, -0.00000619, 0.00000651, -0.00001154, 0.00000937, -0.00001364, 0.00000797, 0.00000701, -0.00001043, -0.00000881, 0.00000402, 0.00000527, -0.00000165, 0.00000350, -0.00000680, 0.00000432, -0.00000725, 0.00000525, 0.00000016, 0.00000221, -0.00000502, -0.00000457, 0.00000430, 0.00000092, -0.00000246, -0.00000256, -0.00000038, -0.00000842, 0.00000902, -0.00001749, -0.00000200, 0.00000149, 0.00000071, 0.00002501, -0.00000132, 0.00001983, 0.00000745, -0.00000121, -0.00000345, -0.00002589, 0.00001608, -0.00000855, 0.00000237, -0.00000292, -0.00000078, -0.00000357, 0.00000121, -0.00000422, 0.00000227, -0.00000500, 0.00000745, -0.00000890, 0.00000543, -0.00000669, -0.00000134, -0.00000251, -0.00000890, -0.00000759, 0.00000405, -0.00001070, -0.00002564, -0.00002569, -0.00009612, -0.00018667, 0.00068423, -0.00002367, -0.00019222, 0.00000622, -0.00003764, 0.00000814, -0.00002466, 0.00000028, -0.00000408, 0.00000049, -0.00000891, 0.00000535, -0.00001202, 0.00000244, -0.00000177, 0.00000207, -0.00000389, -0.00000467, -0.00000064, 0.00000535, 0.00000070, -0.00000485, 0.00000341, 0.00000736, -0.00001386, 0.00001916, -0.00002149, 0.00001633, -0.00002727, 0.00001276, -0.00000966, 0.00001352, -0.00002162, 0.00001824, -0.00002752, 0.00001071, -0.00000175, -0.00000110, 0.00001457, -0.00000270, 0.00000065, -0.00000015, -0.00000499, 0.00000005, 0.00000673, 0.00000300, -0.00000141, 0.00000028, -0.00000300, -0.00000283, 0.00000121, -0.00000731, 0.00000849, 0.00000235, 0.00000065, -0.00001043, 0.00001198, -0.00000247, },
		{ 0.00000720, -0.00000328, -0.00000135, -0.00000696, 0.00001451, -0.00000532, 0.00000141, -0.00000464, 0.00000503, -0.00000838, 0.00000364, -0.00000555, -0.00000221, 0.00001426, -0.00001179, -0.00000696, -0.00000119, 0.00000845, -0.00000416, 0.00000637, -0.00000760, 0.00000405, -0.00000572, 0.00000198, 0.00000277, 0.00000050, -0.00000189, -0.00000742, 0.00000421, 0.00000120, -0.00000057, -0.00000500, 0.00000234, -0.00001040, 0.00000749, -0.00001029, -0.00001138, 0.00000598, -0.00000385, 0.00002397, 0.00000546, 0.00001122, 0.00001855, -0.00000720, 0.00000675, -0.00003361, 0.00001490, -0.00000669, 0.00000179, -0.00000374, 0.00000192, -0.00000744, 0.00000506, -0.00000831, 0.00000698, -0.00001026, 0.00001145, -0.00001114, 0.00000751, -0.00000882, 0.00000262, -0.00000850, 0.00000153, -0.00002051, 0.00001604, -0.00002195, -0.00000269, -0.00005399, -0.00002327, -0.00026977, 0.00053464, 0.00022669, -0.00027072, 0.00001741, -0.00005567, 0.00002035, -0.00003426, 0.00000739, -0.00001460, 0.00001185, -0.00001915, 0.00001375, -0.00001790, 0.00000673, -0.00000752, 0.00000833, -0.00000972, 0.00000252, -0.00000947, 0.00001215, -0.00000539, 0.00000404, -0.00000818, 0.00001904, -0.00002337, 0.00002555, -0.00002471, 0.00001866, -0.00002530, 0.00000737, -0.00000721, 0.00001211, -0.00001738, 0.00001028, -0.00001369, -0.00000678, 0.00001196, -0.00001170, 0.00001908, -0.00000221, 0.00000133, -0.00000140, -0.00000205, -0.00000417, 0.00000842, 0.00000232, 0.00000081, -0.00000215, 0.00000051, -0.00000669, 0.00000509, -0.00001048, 0.00000923, 0.00000044, 0.00000469, -0.00001277, 0.00001167, -0.00000229, },
		{ 0.00000349, 0.00000099, -0.00000230, -0.00000667, 0.00000937, 0.00000238, -0.00000397, -0.00000055, 0.00000027, -0.00000160, -0.00000487, 0.00000347, -0.00001103, 0.00001863, -0.00001099, -0.00000605, -0.00000500, 0.00000994, -0.00000528, 0.00000726, -0.00000621, 0.00000210, -0.00000309, -0.00000131, 0.00000440, -0.00000004, -0.00000027, -0.00000797, 0.00000204, 0.00000256, 0.00000025, -0.00000549, 0.00000211, -0.00000837, 0.00000187, -0.00000134, -0.00001827, 0.00000567, -0.00000315, 0.00001736, 0.00001580, 0.00000172, 0.00002780, -0.00001013, 0.00001300, -0.00003460, 0.00000743, -0.00000100, -0.00000148, -0.00000231, 0.00000205, -0.00000782, 0.00000466, -0.00000778, 0.00000646, -0.00000925, 0.00000853, -0.00000690, 0.00000378, -0.00000563, 0.00000131, -0.00000869, 0.00000442, -0.00002297, 0.00001539, -0.00002013, 0.00000406, -0.00006010, 0.00000603, -0.00026507, 0.00032138, 0.00043496, -0.00026705, -0.00001967, -0.00004664, 0.00001179, -0.00002799, 0.00000257, -0.00001516, 0.00001313, -0.00001924, 0.00001247, -0.00001518, 0.00000411, -0.00000749, 0.00000904, -0.00001022, 0.00000430, -0.00001206, 0.00001245, -0.00000576, 0.00000716, -0.00001264, 0.00002204, -0.00002313, 0.00002183, -0.00001807, 0.00001094, -0.00001364, -0.00000538, 0.00000125, 0.00000396, -0.00000593, -0.00000402, 0.00000352, -0.00002272, 0.00002050, -0.00001576, 0.00001732, 0.00000191, 0.00000066, -0.00000198, 0.00000007, -0.00000666, 0.00000808, 0.00000288, 0.00000224, -0.00000376, 0.00000299, -0.00000898, 0.00000680, -0.00001092, 0.00000741, 0.00000096, 0.00000542, -0.00001133, 0.00000794, 0.00000118, },
		{ -0.00000198, 0.00000598, -0.00000430, -0.00000351, 0.00000143, 0.00001065, -0.00000863, 0.00000322, -0.00000470, 0.00000494, -0.00001158, 0.00000888, -0.00001476, 0.00001758, -0.00000636, -0.00000756, -0.00000607, 0.00000870, -0.00000397, 0.00000530, -0.00000234, -0.00000135, 0.00000026, -0.00000427, 0.00000512, 0.00000008, 0.00000065, -0.00000748, -0.00000049, 0.00000327, 0.00000137, -0.00000538, 0.00000081, -0.00000525, -0.00000350, 0.00000423, -0.00001847, -0.00000133, 0.00000262, 0.00000706, 0.00002656, -0.00000528, 0.00003236, -0.00000796, 0.00001379, -0.00002850, -0.00000504, 0.00000665, -0.00000590, 0.00000031, 0.00000030, -0.00000534, 0.00000098, -0.00000390, 0.00000226, -0.00000408, 0.00000168, 0.00000021, -0.00000204, -0.00000059, -0.00000217, -0.00000577, 0.00000241, -0.00001834, 0.00000687, -0.00001086, 0.00000071, -0.00005111, 0.00000513, -0.00021061, 0.00011049, 0.00054790, -0.00017032, -0.00009345, -0.00002198, -0.00000900, -0.00001284, -0.00000886, -0.00000953, 0.00000725, -0.00001213, 0.00000466, -0.00000712, -0.00000246, -0.00000391, 0.00000579, -0.00000652, 0.00000151, -0.00000935, 0.00000753, -0.00000170, 0.00000547, -0.00001093, 0.00001784, -0.00001533, 0.00001091, -0.00000513, -0.00000245, 0.00000213, -0.00001936, 0.00000989, -0.00000501, 0.00000585, -0.00001654, 0.00001541, -0.00002947, 0.00001877, -0.00001086, 0.00000908, 0.00000890, -0.00000098, -0.00000148, 0.00000054, -0.00000668, 0.00000541, 0.00000478, 0.00000261, -0.00000404, 0.00000367, -0.00000886, 0.00000555, -0.00000808, 0.00000273, 0.00000409, 0.00000304, -0.00000677, 0.00000167, 0.00000687, },
		{ -0.00000648, 0.00000922, -0.00000529, -0.00000011, -0.00000557, 0.00001555, -0.00000936, 0.00000396, -0.00000690, 0.00000781, -0.00001309, 0.00000805, -0.00001202, 0.00001094, 0.00000171, -0.00001119, -0.00000469, 0.00000507, -0.00000048, 0.00000095, 0.00000305, -0.00000525, 0.00000347, -0.00000650, 0.00000516, 0.00000018, 0.00000177, -0.00000714, -0.00000196, 0.00000221, 0.00000356, -0.00000563, -0.00000000, -0.00000338, -0.00000576, 0.00000407, -0.00001179, -0.00001263, 0.00000967, -0.00000251, 0.00003333, -0.00000627, 0.00003018, 0.00000004, 0.00000931, -0.00001692, -0.00001948, 0.00001330, -0.00000924, 0.00000242, -0.00000185, -0.00000180, -0.00000367, 0.00000058, -0.00000253, 0.00000154, -0.00000471, 0.00000563, -0.00000568, 0.00000244, -0.00000422, -0.00000338, -0.00000036, -0.00001207, -0.00000256, -0.00000199, -0.00000405, -0.00003795, -0.00000768, -0.00014358, -0.00005138, 0.00054964, -0.00000595, -0.00017016, -0.00000439, -0.00002666, -0.00000058, -0.00001789, -0.00000467, 0.00000035, -0.00000386, -0.00000371, 0.00000068, -0.00000813, -0.00000074, 0.00000195, -0.00000177, -0.00000271, -0.00000477, 0.00000116, 0.00000318, 0.00000232, -0.00000662, 0.00001057, -0.00000476, -0.00000177, 0.00000815, -0.00001489, 0.00001480, -0.00002764, 0.00001269, -0.00000900, 0.00001175, -0.00002095, 0.00001662, -0.00002416, 0.00000692, 0.00000067, -0.00000230, 0.00001579, -0.00000198, -0.00000049, -0.00000035, -0.00000474, 0.00000131, 0.00000715, 0.00000241, -0.00000318, 0.00000263, -0.00000646, 0.00000175, -0.00000283, -0.00000359, 0.00000839, -0.00000065, -0.00000123, -0.00000489, 0.00001231, },
		{ -0.00000747, 0.00000869, -0.00000360, 0.00000126, -0.00000871, 0.00001474, -0.00000494, 0.00000087, -0.00000530, 0.00000588, -0.00000890, 0.00000165, -0.00000462, 0.00000126, 0.00001061, -0.00001477, -0.00000273, 0.00000078, 0.00000364, -0.00000400, 0.00000796, -0.00000788, 0.00000535, -0.00000735, 0.00000445, 0.00000012, 0.00000320, -0.00000715, -0.00000225, -0.00000025, 0.00000617, -0.00000589, -0.00000021, -0.00000312, -0.00000488, -0.00000036, -0.00000209, -0.00002270, 0.00001255, -0.00000668, 0.00003254, 0.00000047, 0.00002133, 0.00001251, 0.00000167, -0.00000312, -0.00003203, 0.00001598, -0.00000963, 0.00000261, -0.00000297, 0.00000096, -0.00000695, 0.00000316, -0.00000522, 0.00000459, -0.00000747, 0.00000655, -0.00000501, 0.00000175, -0.00000323, -0.00000335, -0.00000134, -0.00000780, -0.00000829, 0.00000176, -0.00000503, -0.00002769, -0.00001973, -0.00008722, -0.00014686, 0.00046053, 0.00017685, -0.00021027, -0.00001390, -0.00002966, 0.00000050, -0.00001847, -0.00000540, -0.00000291, 0.00000074, -0.00000801, 0.00000407, -0.00000953, -0.00000069, 0.00000007, 0.00000131, -0.00000536, -0.00000168, -0.00000313, 0.00000561, 0.00000078, -0.00000309, 0.00000415, 0.00000416, -0.00001135, 0.00001665, -0.00002111, 0.00001924, -0.00002625, 0.00000699, -0.00000578, 0.00000963, -0.00001591, 0.00000770, -0.00001005, -0.00000960, 0.00001276, -0.00001141, 0.00001882, -0.00000041, -0.00000006, -0.00000156, -0.00000223, -0.00000262, 0.00000865, 0.00000254, -0.00000185, 0.00000061, -0.00000280, -0.00000311, 0.00000288, -0.00000928, 0.00001148, -0.00000314, 0.00000272, -0.00000922, 0.00001491, },
		{ -0.00000398, 0.00000412, 0.00000081, 0.00000003, -0.00000736, 0.00000871, 0.00000304, -0.00000435, -0.00000134, 0.00000080, -0.00000150, -0.00000679, 0.00000330, -0.00000731, 0.00001674, -0.00001549, -0.00000255, -0.00000207, 0.00000639, -0.00000736, 0.00001023, -0.00000765, 0.00000487, -0.00000628, 0.00000269, 0.00000033, 0.00000421, -0.00000666, -0.00000246, -0.00000253, 0.00000751, -0.00000484, -0.00000090, -0.00000315, -0.00000305, -0.00000525, 0.00000520, -0.00002604, 0.00000747, -0.00000335, 0.00002379, 0.00001347, 0.00000873, 0.00002588, -0.00000539, 0.00000884, -0.00003902, 0.00001270, -0.00000628, 0.00000043, -0.00000244, 0.00000190, -0.00000763, 0.00000271, -0.00000483, 0.00000414, -0.00000602, 0.00000320, -0.00000094, -0.00000146, -0.00000033, -0.00000488, -0.00000079, -0.00000591, -0.00000987, 0.00000055, -0.00000254, -0.00002108, -0.00002720, -0.00004862, -0.00018355, 0.00032204, 0.00032669, -0.00018759, -0.00005504, -0.00001827, -0.00000904, -0.00001147, -0.00001110, -0.00000233, 0.00000088, -0.00000748, 0.00000264, -0.00000670, -0.00000352, 0.00000029, 0.00000208, -0.00000543, -0.00000136, -0.00000408, 0.00000464, 0.00000168, -0.00000157, 0.00000026, 0.00000941, -0.00001565, 0.00001832, -0.00001944, 0.00001468, -0.00001603, -0.00000505, 0.00000222, 0.00000219, -0.00000525, -0.00000568, 0.00000541, -0.00002290, 0.00001875, -0.00001363, 0.00001566, 0.00000443, -0.00000066, -0.00000215, -0.00000052, -0.00000496, 0.00000827, 0.00000364, -0.00000071, -0.00000146, 0.00000074, -0.00000721, 0.00000687, -0.00001213, 0.00001135, -0.00000258, 0.00000343, -0.00000987, 0.00001335, },
		{ 0.00000281, -0.00000265, 0.00000607, -0.00000244, -0.00000326, 0.00000042, 0.00001099, -0.00000847, 0.00000213, -0.00000426, 0.00000516, -0.00001285, 0.00000753, -0.00001119, 0.00001743, -0.00001157, -0.00000546, -0.00000230, 0.00000658, -0.00000773, 0.00000871, -0.00000408, 0.00000197, -0.00000333, -0.00000017, 0.00000129, 0.00000394, -0.00000467, -0.00000386, -0.00000304, 0.00000615, -0.00000147, -0.00000308, -0.00000201, -0.00000263, -0.00000717, 0.00000634, -0.00002052, -0.00000511, 0.00000529, 0.00001054, 0.00002799, -0.00000249, 0.00003541, -0.00000785, 0.00001541, -0.00003819, 0.00000319, 0.00000012, -0.00000350, -0.00000057, 0.00000104, -0.00000585, -0.00000022, -0.00000223, 0.00000133, -0.00000210, -0.00000190, 0.00000354, -0.00000429, 0.00000189, -0.00000573, -0.00000060, -0.00000465, -0.00000960, -0.00000227, -0.00000017, -0.00001509, -0.00003194, -0.00002467, -0.00018125, 0.00017691, 0.00040967, -0.00010131, -0.00011281, -0.00000478, -0.00002055, -0.00000369, -0.00001666, -0.00000148, -0.00000064, -0.00000481, -0.00000078, -0.00000256, -0.00000667, 0.00000072, 0.00000184, -0.00000397, -0.00000282, -0.00000288, 0.00000183, 0.00000357, -0.00000103, -0.00000177, 0.00001153, -0.00001531, 0.00001433, -0.00001185, 0.00000438, -0.00000181, -0.00001787, 0.00000966, -0.00000514, 0.00000474, -0.00001622, 0.00001467, -0.00002681, 0.00001501, -0.00000777, 0.00000700, 0.00001119, -0.00000161, -0.00000195, -0.00000016, -0.00000523, 0.00000592, 0.00000560, 0.00000008, -0.00000295, 0.00000308, -0.00000916, 0.00000772, -0.00001105, 0.00000740, 0.00000120, 0.00000106, -0.00000720, 0.00000822, },
		{ 0.00001014, -0.00000849, 0.00000946, -0.00000384, 0.00000076, -0.00000638, 0.00001522, -0.00000862, 0.00000268, -0.00000657, 0.00000796, -0.00001362, 0.00000594, -0.00000923, 0.00001250, -0.00000356, -0.00001074, -0.00000051, 0.00000453, -0.00000529, 0.00000397, 0.00000171, -0.00000218, 0.00000054, -0.00000355, 0.00000294, 0.00000217, -0.00000097, -0.00000670, -0.00000146, 0.00000213, 0.00000378, -0.00000642, 0.00000047, -0.00000434, -0.00000520, 0.00000143, -0.00000870, -0.00002029, 0.00001358, -0.00000126, 0.00003801, -0.00000705, 0.00003716, -0.00000311, 0.00001500, -0.00002956, -0.00001076, 0.00000750, -0.00000765, 0.00000161, -0.00000077, -0.00000274, -0.00000401, 0.00000071, -0.00000171, 0.00000167, -0.00000571, 0.00000551, -0.00000435, 0.00000152, -0.00000431, -0.00000221, -0.00000237, -0.00000987, -0.00000354, -0.00000099, -0.00000723, -0.00003621, -0.00001015, -0.00015918, 0.00005429, 0.00041871, 0.00002453, -0.00015941, -0.00000592, -0.00002380, -0.00000259, -0.00001669, -0.00000427, -0.00000041, -0.00000350, -0.00000268, -0.00000040, -0.00000739, -0.00000067, 0.00000229, -0.00000271, -0.00000411, -0.00000182, -0.00000035, 0.00000418, 0.00000033, -0.00000355, 0.00001215, -0.00001236, 0.00000752, -0.00000223, -0.00000660, 0.00001039, -0.00002561, 0.00001156, -0.00000768, 0.00000908, -0.00001900, 0.00001398, -0.00002004, 0.00000296, 0.00000300, -0.00000338, 0.00001676, -0.00000130, -0.00000158, -0.00000080, -0.00000399, 0.00000250, 0.00000761, 0.00000087, -0.00000372, 0.00000380, -0.00000854, 0.00000536, -0.00000660, 0.00000084, 0.00000662, -0.00000261, -0.00000314, 0.00000173, },
		{ 0.00001510, -0.00001059, 0.00000886, -0.00000252, 0.00000251, -0.00000910, 0.00001389, -0.00000408, -0.00000012, -0.00000540, 0.00000615, -0.00000907, -0.00000041, -0.00000332, 0.00000442, 0.00000579, -0.00001588, 0.00000126, 0.00000191, -0.00000160, -0.00000198, 0.00000741, -0.00000558, 0.00000378, -0.00000631, 0.00000449, -0.00000031, 0.00000340, -0.00000986, 0.00000093, -0.00000277, 0.00000885, -0.00000911, 0.00000281, -0.00000676, -0.00000133, -0.00000595, 0.00000353, -0.00003109, 0.00001535, -0.00000632, 0.00003910, -0.00000214, 0.00003011, 0.00000845, 0.00000891, -0.00001575, -0.00002563, 0.00001299, -0.00001020, 0.00000283, -0.00000237, 0.00000022, -0.00000692, 0.00000225, -0.00000324, 0.00000343, -0.00000644, 0.00000387, -0.00000129, -0.00000130, -0.00000096, -0.00000527, 0.00000085, -0.00001113, -0.00000257, -0.00000496, 0.00000174, -0.00003934, -0.00000273, -0.00012911, -0.00003489, 0.00036987, 0.00015304, -0.00016813, -0.00003177, -0.00001473, -0.00001058, -0.00000996, -0.00001141, 0.00000252, -0.00000493, -0.00000161, -0.00000150, -0.00000487, -0.00000427, 0.00000389, -0.00000253, -0.00000396, -0.00000242, -0.00000034, 0.00000237, 0.00000328, -0.00000580, 0.00001226, -0.00000842, 0.00000043, 0.00000598, -0.00001413, 0.00001635, -0.00002513, 0.00000617, -0.00000417, 0.00000668, -0.00001381, 0.00000500, -0.00000667, -0.00001165, 0.00001276, -0.00001045, 0.00001792, 0.00000168, -0.00000172, -0.00000160, -0.00000241, -0.00000066, 0.00000864, 0.00000214, -0.00000392, 0.00000307, -0.00000591, 0.00000093, -0.00000061, -0.00000596, 0.00001112, -0.00000509, -0.00000008, -0.00000358, },
		{ 0.00001601, -0.00000782, 0.00000412, 0.00000140, 0.00000167, -0.00000750, 0.00000790, 0.00000332, -0.00000449, -0.00000221, 0.00000148, -0.00000185, -0.00000805, 0.00000276, -0.00000296, 0.00001285, -0.00001788, 0.00000070, 0.00000061, 0.00000139, -0.00000676, 0.00001060, -0.00000636, 0.00000496, -0.00000726, 0.00000485, -0.00000211, 0.00000664, -0.00001139, 0.00000200, -0.00000602, 0.00001113, -0.00000893, 0.00000297, -0.00000756, 0.00000123, -0.00001108, 0.00001025, -0.00003222, 0.00000747, -0.00000301, 0.00003084, 0.00001089, 0.00001708, 0.00002322, 0.00000103, -0.00000118, -0.00003712, 0.00001382, -0.00000970, 0.00000218, -0.00000285, 0.00000189, -0.00000782, 0.00000159, -0.00000276, 0.00000282, -0.00000431, -0.00000030, 0.00000309, -0.00000454, 0.00000231, -0.00000788, 0.00000331, -0.00001191, -0.00000115, -0.00000925, 0.00000835, -0.00003818, -0.00000301, -0.00009567, -0.00009325, 0.00029001, 0.00025086, -0.00012686, -0.00007805, 0.00000056, -0.00002273, -0.00000076, -0.00001938, 0.00000571, -0.00000739, 0.00000096, -0.00000433, -0.00000088, -0.00000832, 0.00000538, -0.00000278, -0.00000271, -0.00000445, 0.00000140, -0.00000099, 0.00000674, -0.00000767, 0.00001147, -0.00000383, -0.00000572, 0.00001107, -0.00001650, 0.00001522, -0.00001730, -0.00000428, 0.00000281, 0.00000045, -0.00000462, -0.00000667, 0.00000630, -0.00002182, 0.00001599, -0.00001073, 0.00001330, 0.00000729, -0.00000238, -0.00000201, -0.00000142, -0.00000257, 0.00000807, 0.00000407, -0.00000372, 0.00000142, -0.00000235, -0.00000380, 0.00000460, -0.00001053, 0.00001252, -0.00000460, 0.00000046, -0.00000608, },
		{ 0.00001308, -0.00000125, -0.00000287, 0.00000601, -0.00000032, -0.00000345, 0.00000023, 0.00001019, -0.00000752, 0.00000045, -0.00000312, 0.00000440, -0.00001306, 0.00000541, -0.00000650, 0.00001505, -0.00001505, -0.00000321, 0.00000145, 0.00000255, -0.00000881, 0.00001000, -0.00000386, 0.00000366, -0.00000594, 0.00000335, -0.00000214, 0.00000727, -0.00000978, 0.00000017, -0.00000587, 0.00000913, -0.00000481, -0.00000007, -0.00000532, 0.00000031, -0.00001104, 0.00000872, -0.00002306, -0.00000808, 0.00000557, 0.00001719, 0.00002687, 0.00000383, 0.00003563, -0.00000368, 0.00000954, -0.00004169, 0.00000846, -0.00000577, -0.00000054, -0.00000190, 0.00000184, -0.00000655, -0.00000090, -0.00000094, 0.00000086, -0.00000094, -0.00000462, 0.00000613, -0.00000575, 0.00000332, -0.00000806, 0.00000319, -0.00001039, -0.00000166, -0.00001055, 0.00000917, -0.00003015, -0.00001142, -0.00006087, -0.00012814, 0.00020335, 0.00030076, -0.00004391, -0.00012635, 0.00000849, -0.00002983, 0.00000373, -0.00002269, 0.00000516, -0.00000766, 0.00000196, -0.00000584, 0.00000157, -0.00001016, 0.00000479, -0.00000198, -0.00000172, -0.00000633, 0.00000298, -0.00000366, 0.00000851, -0.00000736, 0.00000859, 0.00000184, -0.00001082, 0.00001301, -0.00001443, 0.00000908, -0.00000593, -0.00001509, 0.00000867, -0.00000496, 0.00000316, -0.00001484, 0.00001275, -0.00002300, 0.00001056, -0.00000424, 0.00000451, 0.00001361, -0.00000255, -0.00000212, -0.00000127, -0.00000306, 0.00000609, 0.00000630, -0.00000302, -0.00000067, 0.00000101, -0.00000727, 0.00000731, -0.00001143, 0.00000999, -0.00000095, -0.00000146, -0.00000579, },
		{ 0.00000806, 0.00000654, -0.00000903, 0.00000871, -0.00000134, 0.00000043, -0.00000576, 0.00001344, -0.00000701, 0.00000070, -0.00000538, 0.00000704, -0.00001315, 0.00000317, -0.00000538, 0.00001221, -0.00000797, -0.00000948, 0.00000359, 0.00000218, -0.00000811, 0.00000601, 0.00000099, 0.00000076, -0.00000297, 0.00000028, -0.00000026, 0.00000507, -0.00000502, -0.00000443, -0.00000253, 0.00000349, 0.00000214, -0.00000527, -0.00000072, -0.00000372, -0.00000628, 0.00000071, -0.00000800, -0.00002496, 0.00001308, 0.00000445, 0.00003916, -0.00000349, 0.00004067, -0.00000150, 0.00001356, -0.00003798, -0.00000263, 0.00000056, -0.00000456, 0.00000005, 0.00000041, -0.00000382, -0.00000411, 0.00000091, -0.00000100, 0.00000179, -0.00000692, 0.00000598, -0.00000371, 0.00000130, -0.00000528, 0.00000012, -0.00000611, -0.00000500, -0.00000762, 0.00000360, -0.00001600, -0.00002527, -0.00002822, -0.00014462, 0.00012348, 0.00030478, 0.00005625, -0.00015256, -0.00000381, -0.00002474, -0.00000139, -0.00001800, -0.00000141, -0.00000377, -0.00000065, -0.00000395, 0.00000048, -0.00000820, 0.00000108, 0.00000071, -0.00000213, -0.00000653, 0.00000251, -0.00000363, 0.00000690, -0.00000378, 0.00000308, 0.00000857, -0.00001476, 0.00001238, -0.00000969, 0.00000122, 0.00000445, -0.00002171, 0.00000966, -0.00000605, 0.00000586, -0.00001605, 0.00001052, -0.00001537, -0.00000106, 0.00000528, -0.00000446, 0.00001759, -0.00000079, -0.00000246, -0.00000158, -0.00000268, 0.00000354, 0.00000806, -0.00000159, -0.00000280, 0.00000337, -0.00000857, 0.00000692, -0.00000874, 0.00000446, 0.00000441, -0.00000429, -0.00000413, },
		{ 0.00000308, 0.00001279, -0.00001170, 0.00000772, -0.00000010, 0.00000237, -0.00000806, 0.00001188, -0.00000272, -0.00000154, -0.00000485, 0.00000567, -0.00000872, -0.00000267, -0.00000137, 0.00000651, 0.00000072, -0.00001547, 0.00000501, 0.00000164, -0.00000592, 0.00000043, 0.00000606, -0.00000197, 0.00000030, -0.00000322, 0.00000252, 0.00000120, 0.00000125, -0.00000990, 0.00000187, -0.00000315, 0.00000899, -0.00000990, 0.00000370, -0.00000824, -0.00000014, -0.00000877, 0.00000588, -0.00003565, 0.00001346, -0.00000208, 0.00004286, -0.00000128, 0.00003642, 0.00000813, 0.00001115, -0.00002749, -0.00001682, 0.00000697, -0.00000839, 0.00000200, -0.00000149, -0.00000078, -0.00000672, 0.00000165, -0.00000172, 0.00000280, -0.00000641, 0.00000267, 0.00000075, -0.00000249, -0.00000096, -0.00000439, -0.00000060, -0.00000976, -0.00000215, -0.00000540, -0.00000026, -0.00003853, -0.00000360, -0.00014381, 0.00005336, 0.00027821, 0.00014490, -0.00013908, -0.00003968, -0.00000828, -0.00001482, -0.00000701, -0.00001224, 0.00000302, -0.00000613, 0.00000080, -0.00000361, -0.00000323, -0.00000473, 0.00000454, -0.00000375, -0.00000480, -0.00000045, -0.00000060, 0.00000212, 0.00000233, -0.00000388, 0.00001499, -0.00001661, 0.00000942, -0.00000384, -0.00000570, 0.00001083, -0.00002203, 0.00000488, -0.00000247, 0.00000334, -0.00001105, 0.00000206, -0.00000343, -0.00001318, 0.00001237, -0.00000926, 0.00001673, 0.00000367, -0.00000328, -0.00000185, -0.00000223, 0.00000134, 0.00000867, 0.00000068, -0.00000466, 0.00000432, -0.00000765, 0.00000402, -0.00000387, -0.00000198, 0.00000918, -0.00000600, -0.00000287, },
		{ -0.00000045, 0.00001581, -0.00000983, 0.00000309, 0.00000305, 0.00000229, -0.00000677, 0.00000666, 0.00000344, -0.00000459, -0.00000286, 0.00000195, -0.00000228, -0.00000898, 0.00000236, 0.00000115, 0.00000776, -0.00001837, 0.00000372, 0.00000231, -0.00000375, -0.00000459, 0.00000918, -0.00000289, 0.00000247, -0.00000577, 0.00000472, -0.00000238, 0.00000654, -0.00001353, 0.00000447, -0.00000759, 0.00001253, -0.00001104, 0.00000517, -0.00001016, 0.00000344, -0.00001447, 0.00001254, -0.00003542, 0.00000433, -0.00000099, 0.00003719, 0.00000974, 0.00002506, 0.00002214, 0.00000561, -0.00001406, -0.00003009, 0.00001061, -0.00001034, 0.00000279, -0.00000283, 0.00000144, -0.00000778, 0.00000080, -0.00000118, 0.00000219, -0.00000383, -0.00000211, 0.00000529, -0.00000559, 0.00000244, -0.00000783, 0.00000355, -0.00001319, 0.00000243, -0.00001299, 0.00001105, -0.00004453, 0.00000745, -0.00012563, -0.00000898, 0.00023878, 0.00020167, -0.00008394, -0.00008882, 0.00000969, -0.00002895, 0.00000387, -0.00002203, 0.00000858, -0.00001112, 0.00000539, -0.00000785, 0.00000187, -0.00000992, 0.00000736, -0.00000509, -0.00000233, -0.00000459, 0.00000372, -0.00000359, 0.00000838, -0.00000961, 0.00001867, -0.00001487, 0.00000392, 0.00000230, -0.00001044, 0.00001253, -0.00001690, -0.00000348, 0.00000323, -0.00000150, -0.00000369, -0.00000739, 0.00000661, -0.00002000, 0.00001284, -0.00000762, 0.00001072, 0.00001001, -0.00000404, -0.00000197, -0.00000216, 0.00000001, 0.00000795, 0.00000351, -0.00000582, 0.00000380, -0.00000506, -0.00000011, 0.00000123, -0.00000710, 0.00001137, -0.00000521, -0.00000310, },
		{ -0.00000237, 0.00001547, -0.00000431, -0.00000332, 0.00000630, 0.00000152, -0.00000361, 0.00000046, 0.00000863, -0.00000617, -0.00000137, -0.00000174, 0.00000308, -0.00001259, 0.00000313, -0.00000148, 0.00001095, -0.00001671, -0.00000090, 0.00000450, -0.00000240, -0.00000767, 0.00000917, -0.00000136, 0.00000294, -0.00000652, 0.00000514, -0.00000398, 0.00000881, -0.00001330, 0.00000330, -0.00000782, 0.00001102, -0.00000744, 0.00000247, -0.00000795, 0.00000236, -0.00001381, 0.00000998, -0.00002468, -0.00001149, 0.00000460, 0.00002558, 0.00002490, 0.00001204, 0.00003515, 0.00000173, -0.00000225, -0.00003855, 0.00000926, -0.00000929, 0.00000166, -0.00000289, 0.00000214, -0.00000701, -0.00000134, 0.00000000, 0.00000091, -0.00000078, -0.00000606, 0.00000748, -0.00000586, 0.00000303, -0.00000831, 0.00000437, -0.00001309, 0.00000325, -0.00001534, 0.00001379, -0.00003960, 0.00000336, -0.00009298, -0.00006397, 0.00019768, 0.00022189, -0.00000220, -0.00013069, 0.00001501, -0.00003430, 0.00000734, -0.00002506, 0.00000851, -0.00001193, 0.00000642, -0.00000898, 0.00000394, -0.00001167, 0.00000701, -0.00000443, -0.00000089, -0.00000775, 0.00000669, -0.00000724, 0.00001132, -0.00001134, 0.00001744, -0.00000853, -0.00000394, 0.00000814, -0.00001286, 0.00001066, -0.00000911, -0.00001173, 0.00000737, -0.00000491, 0.00000166, -0.00001295, 0.00001021, -0.00001867, 0.00000604, -0.00000072, 0.00000199, 0.00001579, -0.00000346, -0.00000230, -0.00000238, -0.00000060, 0.00000636, 0.00000622, -0.00000583, 0.00000196, -0.00000166, -0.00000405, 0.00000489, -0.00000941, 0.00001021, -0.00000189, -0.00000463, },
		{ -0.00000348, 0.00001289, 0.00000272, -0.00000883, 0.00000756, 0.00000162, -0.00000060, -0.00000413, 0.00001062, -0.00000489, -0.00000157, -0.00000373, 0.00000539, -0.00001201, 0.00000021, -0.00000095, 0.00001016, -0.00001110, -0.00000761, 0.00000713, -0.00000153, -0.00000869, 0.00000631, 0.00000193, 0.00000217, -0.00000552, 0.00000355, -0.00000306, 0.00000748, -0.00000900, -0.00000154, -0.00000414, 0.00000525, -0.00000034, -0.00000315, -0.00000264, -0.00000249, -0.00000799, 0.00000090, -0.00000872, -0.00002742, 0.00000893, 0.00001356, 0.00003797, 0.00000342, 0.00004205, 0.00000338, 0.00000470, -0.00004003, 0.00000235, -0.00000537, -0.00000132, -0.00000160, 0.00000132, -0.00000482, -0.00000394, 0.00000091, -0.00000003, 0.00000136, -0.00000760, 0.00000618, -0.00000284, 0.00000067, -0.00000574, 0.00000177, -0.00000926, -0.00000021, -0.00001171, 0.00000797, -0.00002527, -0.00001172, -0.00005363, -0.00010740, 0.00015661, 0.00021519, 0.00008112, -0.00014396, -0.00000251, -0.00002573, -0.00000007, -0.00001905, 0.00000127, -0.00000721, 0.00000245, -0.00000560, 0.00000164, -0.00000894, 0.00000287, -0.00000122, -0.00000149, -0.00000836, 0.00000639, -0.00000682, 0.00000947, -0.00000792, 0.00001101, 0.00000161, -0.00001275, 0.00001264, -0.00001301, 0.00000681, -0.00000157, -0.00001677, 0.00000745, -0.00000463, 0.00000267, -0.00001262, 0.00000668, -0.00001059, -0.00000475, 0.00000733, -0.00000545, 0.00001816, -0.00000033, -0.00000324, -0.00000250, -0.00000102, 0.00000463, 0.00000806, -0.00000442, -0.00000082, 0.00000165, -0.00000670, 0.00000621, -0.00000868, 0.00000632, 0.00000273, -0.00000628, },
		{ -0.00000470, 0.00000943, 0.00000906, -0.00001132, 0.00000566, 0.00000324, 0.00000117, -0.00000587, 0.00000897, -0.00000109, -0.00000315, -0.00000387, 0.00000458, -0.00000794, -0.00000496, 0.00000114, 0.00000715, -0.00000385, -0.00001388, 0.00000828, -0.00000013, -0.00000842, 0.00000197, 0.00000536, 0.00000131, -0.00000356, 0.00000061, -0.00000032, 0.00000360, -0.00000230, -0.00000796, 0.00000107, -0.00000191, 0.00000711, -0.00000865, 0.00000282, -0.00000795, -0.00000093, -0.00000917, 0.00000521, -0.00003650, 0.00000703, 0.00000567, 0.00004418, 0.00000308, 0.00004044, 0.00001161, 0.00000635, -0.00003492, -0.00000862, -0.00000003, -0.00000518, 0.00000039, -0.00000043, -0.00000207, -0.00000609, 0.00000090, -0.00000014, 0.00000204, -0.00000655, 0.00000208, 0.00000204, -0.00000294, -0.00000185, -0.00000244, -0.00000361, -0.00000595, -0.00000465, -0.00000240, -0.00000744, -0.00002935, -0.00001836, -0.00013231, 0.00011121, 0.00019763, 0.00014266, -0.00011719, -0.00004202, -0.00000676, -0.00001512, -0.00000706, -0.00001031, 0.00000081, -0.00000473, 0.00000074, -0.00000362, -0.00000326, -0.00000346, 0.00000331, -0.00000357, -0.00000648, 0.00000274, -0.00000254, 0.00000359, -0.00000076, 0.00000152, 0.00001269, -0.00001972, 0.00001409, -0.00001064, 0.00000203, 0.00000405, -0.00001752, 0.00000340, -0.00000106, 0.00000001, -0.00000792, -0.00000094, -0.00000045, -0.00001419, 0.00001178, -0.00000803, 0.00001541, 0.00000544, -0.00000460, -0.00000236, -0.00000166, 0.00000333, 0.00000862, -0.00000172, -0.00000388, 0.00000405, -0.00000751, 0.00000522, -0.00000570, 0.00000128, 0.00000683, -0.00000667, },
		{ -0.00000638, 0.00000595, 0.00001333, -0.00001008, 0.00000110, 0.00000564, 0.00000194, -0.00000518, 0.00000500, 0.00000344, -0.00000468, -0.00000322, 0.00000203, -0.00000262, -0.00000973, 0.00000243, 0.00000425, 0.00000236, -0.00001729, 0.00000644, 0.00000253, -0.00000775, -0.00000229, 0.00000736, 0.00000139, -0.00000154, -0.00000252, 0.00000278, -0.00000079, 0.00000412, -0.00001300, 0.00000473, -0.00000708, 0.00001151, -0.00001090, 0.00000538, -0.00001067, 0.00000328, -0.00001506, 0.00001158, -0.00003504, -0.00000243, 0.00000306, 0.00004225, 0.00001103, 0.00003179, 0.00002407, 0.00000505, -0.00002590, -0.00002062, 0.00000439, -0.00000842, 0.00000200, -0.00000214, 0.00000028, -0.00000709, -0.00000018, 0.00000037, 0.00000171, -0.00000398, -0.00000281, 0.00000625, -0.00000533, 0.00000097, -0.00000578, 0.00000110, -0.00001080, 0.00000179, -0.00001171, 0.00000672, -0.00004079, 0.00000344, -0.00013364, 0.00005775, 0.00018190, 0.00017070, -0.00005489, -0.00009006, 0.00001120, -0.00002928, 0.00000386, -0.00002040, 0.00000739, -0.00001110, 0.00000641, -0.00000846, 0.00000213, -0.00000895, 0.00000676, -0.00000538, -0.00000363, -0.00000246, 0.00000329, -0.00000343, 0.00000677, -0.00000722, 0.00002076, -0.00002173, 0.00001090, -0.00000550, -0.00000317, 0.00000748, -0.00001492, -0.00000274, 0.00000337, -0.00000356, -0.00000242, -0.00000802, 0.00000658, -0.00001783, 0.00000978, -0.00000473, 0.00000827, 0.00001227, -0.00000534, -0.00000227, -0.00000249, 0.00000250, 0.00000808, 0.00000166, -0.00000637, 0.00000491, -0.00000642, 0.00000258, -0.00000183, -0.00000321, 0.00000895, -0.00000508, },
		{ -0.00000804, 0.00000257, 0.00001531, -0.00000591, -0.00000436, 0.00000721, 0.00000278, -0.00000343, 0.00000081, 0.00000661, -0.00000470, -0.00000301, -0.00000051, 0.00000160, -0.00001185, 0.00000124, 0.00000300, 0.00000591, -0.00001672, 0.00000147, 0.00000616, -0.00000695, -0.00000552, 0.00000717, 0.00000271, 0.00000007, -0.00000492, 0.00000486, -0.00000377, 0.00000796, -0.00001440, 0.00000467, -0.00000815, 0.00001104, -0.00000850, 0.00000369, -0.00000910, 0.00000269, -0.00001458, 0.00000911, -0.00002431, -0.00001632, 0.00000332, 0.00003439, 0.00002377, 0.00002057, 0.00003619, 0.00000465, -0.00001655, -0.00003034, 0.00000571, -0.00000974, 0.00000226, -0.00000297, 0.00000151, -0.00000674, -0.00000195, 0.00000090, 0.00000120, -0.00000136, -0.00000639, 0.00000769, -0.00000478, 0.00000121, -0.00000659, 0.00000292, -0.00001238, 0.00000444, -0.00001573, 0.00001240, -0.00004105, 0.00000791, -0.00011254, -0.00000171, 0.00017132, 0.00016870, 0.00002413, -0.00012579, 0.00001460, -0.00003364, 0.00000687, -0.00002358, 0.00000821, -0.00001290, 0.00000795, -0.00000971, 0.00000420, -0.00001090, 0.00000699, -0.00000516, -0.00000171, -0.00000676, 0.00000765, -0.00000815, 0.00001109, -0.00001161, 0.00002268, -0.00001697, 0.00000282, 0.00000183, -0.00000830, 0.00000914, -0.00001078, -0.00000829, 0.00000595, -0.00000522, 0.00000061, -0.00001102, 0.00000753, -0.00001437, 0.00000204, 0.00000230, -0.00000021, 0.00001740, -0.00000401, -0.00000278, -0.00000313, 0.00000176, 0.00000704, 0.00000482, -0.00000748, 0.00000397, -0.00000385, -0.00000075, 0.00000165, -0.00000598, 0.00000860, -0.00000182, },
		{ -0.00000871, -0.00000100, 0.00001545, -0.00000038, -0.00000859, 0.00000653, 0.00000450, -0.00000180, -0.00000200, 0.00000724, -0.00000271, -0.00000366, -0.00000210, 0.00000347, -0.00001062, -0.00000238, 0.00000332, 0.00000681, -0.00001283, -0.00000520, 0.00000941, -0.00000553, -0.00000763, 0.00000509, 0.00000467, 0.00000141, -0.00000616, 0.00000517, -0.00000436, 0.00000824, -0.00001161, 0.00000078, -0.00000525, 0.00000634, -0.00000252, -0.00000113, -0.00000426, -0.00000167, -0.00000921, 0.00000080, -0.00000965, -0.00002881, 0.00000234, 0.00002445, 0.00003628, 0.00001205, 0.00004350, 0.00000831, -0.00000951, -0.00003557, 0.00000288, -0.00000865, 0.00000079, -0.00000258, 0.00000142, -0.00000527, -0.00000380, 0.00000082, 0.00000117, 0.00000029, -0.00000749, 0.00000574, -0.00000135, -0.00000088, -0.00000496, 0.00000173, -0.00001033, 0.00000258, -0.00001354, 0.00000923, -0.00003082, -0.00000179, -0.00007720, -0.00005845, 0.00016003, 0.00015121, 0.00009531, -0.00013072, -0.00000444, -0.00002461, -0.00000042, -0.00001822, 0.00000212, -0.00000906, 0.00000426, -0.00000633, 0.00000196, -0.00000857, 0.00000357, -0.00000245, -0.00000175, -0.00000848, 0.00000843, -0.00000844, 0.00001029, -0.00001013, 0.00001778, -0.00000624, -0.00000824, 0.00000942, -0.00001222, 0.00000922, -0.00000652, -0.00001155, 0.00000532, -0.00000382, 0.00000006, -0.00000928, 0.00000300, -0.00000625, -0.00000766, 0.00000886, -0.00000613, 0.00001825, 0.00000029, -0.00000414, -0.00000324, 0.00000066, 0.00000610, 0.00000704, -0.00000678, 0.00000141, -0.00000052, -0.00000378, 0.00000386, -0.00000668, 0.00000626, 0.00000195, },
		{ -0.00000769, -0.00000484, 0.00001426, 0.00000506, -0.00001022, 0.00000330, 0.00000689, -0.00000055, -0.00000305, 0.00000563, 0.00000043, -0.00000456, -0.00000283, 0.00000318, -0.00000701, -0.00000689, 0.00000382, 0.00000634, -0.00000751, -0.00001130, 0.00001055, -0.00000280, -0.00000900, 0.00000205, 0.00000622, 0.00000292, -0.00000630, 0.00000376, -0.00000271, 0.00000558, -0.00000594, -0.00000516, -0.00000050, -0.00000004, 0.00000420, -0.00000627, 0.00000103, -0.00000675, -0.00000273, -0.00000815, 0.00000250, -0.00003455, -0.00000290, 0.00001553, 0.00004443, 0.00000987, 0.00004371, 0.00001693, -0.00000535, -0.00003605, -0.00000360, -0.00000579, -0.00000201, -0.00000123, 0.00000035, -0.00000328, -0.00000516, -0.00000007, 0.00000166, 0.00000085, -0.00000635, 0.00000148, 0.00000327, -0.00000354, -0.00000245, -0.00000096, -0.00000621, -0.00000205, -0.00000742, 0.00000078, -0.00001541, -0.00001756, -0.00003985, -0.00010091, 0.00013853, 0.00013496, 0.00013927, -0.00009771, -0.00004258, -0.00000694, -0.00001414, -0.00000773, -0.00000793, -0.00000194, -0.00000264, -0.00000015, -0.00000294, -0.00000357, -0.00000189, 0.00000156, -0.00000327, -0.00000757, 0.00000552, -0.00000450, 0.00000508, -0.00000409, 0.00000832, 0.00000703, -0.00001850, 0.00001432, -0.00001332, 0.00000738, -0.00000255, -0.00001240, 0.00000210, -0.00000036, -0.00000279, -0.00000488, -0.00000365, 0.00000198, -0.00001453, 0.00001095, -0.00000680, 0.00001395, 0.00000701, -0.00000576, -0.00000295, -0.00000087, 0.00000548, 0.00000807, -0.00000447, -0.00000203, 0.00000259, -0.00000572, 0.00000446, -0.00000563, 0.00000299, 0.00000497, },
		{ -0.00000500, -0.00000845, 0.00001185, 0.00000962, -0.00000913, -0.00000145, 0.00000875, 0.00000097, -0.00000304, 0.00000314, 0.00000318, -0.00000463, -0.00000341, 0.00000180, -0.00000287, -0.00001023, 0.00000296, 0.00000596, -0.00000266, -0.00001491, 0.00000857, 0.00000134, -0.00000986, -0.00000104, 0.00000653, 0.00000485, -0.00000551, 0.00000118, 0.00000015, 0.00000160, 0.00000031, -0.00001050, 0.00000331, -0.00000506, 0.00000863, -0.00000892, 0.00000398, -0.00000950, 0.00000124, -0.00001334, 0.00000783, -0.00003147, -0.00001233, 0.00000847, 0.00004646, 0.00001472, 0.00003767, 0.00002866, -0.00000247, -0.00003321, -0.00001195, -0.00000259, -0.00000515, 0.00000031, -0.00000105, -0.00000140, -0.00000573, -0.00000154, 0.00000219, 0.00000094, -0.00000415, -0.00000307, 0.00000669, -0.00000468, -0.00000094, -0.00000320, -0.00000231, -0.00000667, -0.00000104, -0.00000782, -0.00000154, -0.00003054, -0.00001168, -0.00012034, 0.00010071, 0.00013009, 0.00014962, -0.00003480, -0.00008551, 0.00000818, -0.00002620, 0.00000142, -0.00001661, 0.00000405, -0.00000891, 0.00000533, -0.00000734, 0.00000111, -0.00000666, 0.00000472, -0.00000472, -0.00000537, 0.00000060, 0.00000139, -0.00000169, 0.00000316, -0.00000161, 0.00001815, -0.00002383, 0.00001389, -0.00001039, 0.00000329, 0.00000139, -0.00001184, -0.00000190, 0.00000299, -0.00000540, -0.00000100, -0.00000849, 0.00000622, -0.00001544, 0.00000706, -0.00000229, 0.00000608, 0.00001398, -0.00000623, -0.00000287, -0.00000240, 0.00000491, 0.00000824, -0.00000124, -0.00000532, 0.00000455, -0.00000605, 0.00000352, -0.00000353, -0.00000014, 0.00000640, },
		{ -0.00000143, -0.00001085, 0.00000804, 0.00001315, -0.00000612, -0.00000594, 0.00000865, 0.00000347, -0.00000288, 0.00000116, 0.00000432, -0.00000325, -0.00000429, 0.00000034, 0.00000021, -0.00001105, 0.00000009, 0.00000625, 0.00000071, -0.00001533, 0.00000376, 0.00000608, -0.00000992, -0.00000376, 0.00000542, 0.00000690, -0.00000382, -0.00000189, 0.00000300, -0.00000191, 0.00000497, -0.00001308, 0.00000417, -0.00000680, 0.00000913, -0.00000778, 0.00000336, -0.00000859, 0.00000117, -0.00001319, 0.00000587, -0.00002163, -0.00002282, 0.00000200, 0.00004296, 0.00002455, 0.00002870, 0.00003991, 0.00000174, -0.00002911, -0.00002009, -0.00000058, -0.00000760, 0.00000125, -0.00000210, -0.00000012, -0.00000552, -0.00000305, 0.00000207, 0.00000131, -0.00000219, -0.00000607, 0.00000726, -0.00000314, -0.00000136, -0.00000382, -0.00000020, -0.00000916, 0.00000265, -0.00001259, 0.00000603, -0.00003507, 0.00000147, -0.00011513, 0.00004833, 0.00013607, 0.00013417, 0.00003828, -0.00011398, 0.00000904, -0.00002919, 0.00000355, -0.00001938, 0.00000522, -0.00001124, 0.00000706, -0.00000848, 0.00000290, -0.00000848, 0.00000526, -0.00000464, -0.00000356, -0.00000401, 0.00000629, -0.00000669, 0.00000805, -0.00000809, 0.00002336, -0.00002172, 0.00000732, -0.00000366, -0.00000255, 0.00000532, -0.00001091, -0.00000497, 0.00000440, -0.00000583, 0.00000011, -0.00000928, 0.00000499, -0.00001049, -0.00000108, 0.00000455, -0.00000190, 0.00001828, -0.00000408, -0.00000365, -0.00000339, 0.00000388, 0.00000812, 0.00000200, -0.00000746, 0.00000473, -0.00000474, 0.00000149, -0.00000111, -0.00000246, 0.00000618, },
		{ 0.00000184, -0.00001118, 0.00000284, 0.00001567, -0.00000229, -0.00000866, 0.00000592, 0.00000694, -0.00000282, 0.00000023, 0.00000373, -0.00000082, -0.00000522, -0.00000088, 0.00000167, -0.00000936, -0.00000409, 0.00000667, 0.00000281, -0.00001326, -0.00000235, 0.00000984, -0.00000854, -0.00000618, 0.00000336, 0.00000838, -0.00000117, -0.00000482, 0.00000487, -0.00000378, 0.00000685, -0.00001215, 0.00000178, -0.00000522, 0.00000609, -0.00000363, 0.00000004, -0.00000492, -0.00000194, -0.00000922, -0.00000044, -0.00000985, -0.00003006, -0.00000563, 0.00003557, 0.00003589, 0.00002094, 0.00004716, 0.00000931, -0.00002517, -0.00002656, -0.00000068, -0.00000878, 0.00000110, -0.00000243, 0.00000039, -0.00000479, -0.00000407, 0.00000094, 0.00000229, -0.00000112, -0.00000678, 0.00000481, 0.00000059, -0.00000315, -0.00000307, -0.00000003, -0.00000900, 0.00000275, -0.00001236, 0.00000633, -0.00003054, 0.00000052, -0.00009143, -0.00000919, 0.00014356, 0.00010940, 0.00009890, -0.00011280, -0.00000977, -0.00002105, -0.00000270, -0.00001532, 0.00000081, -0.00000883, 0.00000428, -0.00000572, 0.00000118, -0.00000687, 0.00000290, -0.00000267, -0.00000308, -0.00000667, 0.00000815, -0.00000783, 0.00000858, -0.00000925, 0.00002146, -0.00001259, -0.00000354, 0.00000468, -0.00000845, 0.00000847, -0.00000979, -0.00000656, 0.00000347, -0.00000379, -0.00000167, -0.00000643, -0.00000012, -0.00000277, -0.00000944, 0.00000962, -0.00000637, 0.00001774, 0.00000121, -0.00000531, -0.00000359, 0.00000212, 0.00000808, 0.00000461, -0.00000792, 0.00000307, -0.00000219, -0.00000098, 0.00000101, -0.00000372, 0.00000484, },
		{ 0.00000392, -0.00000919, -0.00000314, 0.00001688, 0.00000170, -0.00000907, 0.00000117, 0.00001026, -0.00000222, -0.00000014, 0.00000235, 0.00000146, -0.00000540, -0.00000212, 0.00000187, -0.00000628, -0.00000802, 0.00000609, 0.00000446, -0.00001006, -0.00000786, 0.00001118, -0.00000530, -0.00000840, 0.00000101, 0.00000863, 0.00000236, -0.00000705, 0.00000523, -0.00000364, 0.00000605, -0.00000845, -0.00000265, -0.00000187, 0.00000145, 0.00000135, -0.00000379, -0.00000075, -0.00000561, -0.00000450, -0.00000690, -0.00000097, -0.00003108, -0.00001485, 0.00002565, 0.00004550, 0.00001758, 0.00004850, 0.00002048, -0.00002138, -0.00003104, -0.00000293, -0.00000874, -0.00000014, -0.00000209, 0.00000026, -0.00000388, -0.00000436, -0.00000102, 0.00000353, -0.00000072, -0.00000570, 0.00000065, 0.00000470, -0.00000468, -0.00000217, -0.00000080, -0.00000714, 0.00000031, -0.00000857, 0.00000171, -0.00002049, -0.00000863, -0.00006004, -0.00005912, 0.00014063, 0.00009166, 0.00013123, -0.00007814, -0.00004349, -0.00000690, -0.00001347, -0.00000757, -0.00000656, -0.00000377, -0.00000120, -0.00000075, -0.00000255, -0.00000325, -0.00000100, 0.00000024, -0.00000366, -0.00000713, 0.00000666, -0.00000515, 0.00000517, -0.00000587, 0.00001421, 0.00000032, -0.00001475, 0.00001123, -0.00001203, 0.00000950, -0.00000786, -0.00000732, 0.00000123, -0.00000063, -0.00000466, -0.00000236, -0.00000571, 0.00000353, -0.00001398, 0.00000981, -0.00000553, 0.00001227, 0.00000850, -0.00000690, -0.00000339, -0.00000009, 0.00000801, 0.00000649, -0.00000675, 0.00000013, 0.00000075, -0.00000314, 0.00000239, -0.00000402, 0.00000307, },
		{ 0.00000452, -0.00000553, -0.00000864, 0.00001613, 0.00000575, -0.00000768, -0.00000396, 0.00001178, -0.00000012, -0.00000082, 0.00000145, 0.00000243, -0.00000419, -0.00000366, 0.00000146, -0.00000322, -0.00001023, 0.00000369, 0.00000624, -0.00000688, -0.00001142, 0.00000954, -0.00000056, -0.00001016, -0.00000124, 0.00000745, 0.00000625, -0.00000800, 0.00000395, -0.00000189, 0.00000352, -0.00000357, -0.00000716, 0.00000117, -0.00000252, 0.00000492, -0.00000603, 0.00000181, -0.00000754, -0.00000168, -0.00001042, 0.00000244, -0.00002587, -0.00002404, 0.00001370, 0.00005112, 0.00001991, 0.00004435, 0.00003354, -0.00001642, -0.00003408, -0.00000670, -0.00000800, -0.00000209, -0.00000137, -0.00000021, -0.00000304, -0.00000406, -0.00000316, 0.00000423, -0.00000023, -0.00000403, -0.00000332, 0.00000715, -0.00000435, -0.00000232, -0.00000133, -0.00000502, -0.00000277, -0.00000387, -0.00000406, -0.00000991, -0.00001877, -0.00003170, -0.00009109, 0.00011951, 0.00008965, 0.00013218, -0.00001953, -0.00007852, 0.00000346, -0.00002204, -0.00000137, -0.00001265, 0.00000045, -0.00000621, 0.00000354, -0.00000571, 0.00000001, -0.00000433, 0.00000248, -0.00000423, -0.00000628, 0.00000309, -0.00000047, 0.00000008, -0.00000066, 0.00000524, 0.00001245, -0.00002195, 0.00001287, -0.00001145, 0.00000741, -0.00000441, -0.00000829, -0.00000079, 0.00000184, -0.00000667, 0.00000024, -0.00000863, 0.00000544, -0.00001283, 0.00000474, -0.00000035, 0.00000413, 0.00001521, -0.00000682, -0.00000359, -0.00000208, 0.00000743, 0.00000792, -0.00000452, -0.00000316, 0.00000313, -0.00000430, 0.00000275, -0.00000353, 0.00000141, },
		{ 0.00000401, -0.00000146, -0.00001222, 0.00001286, 0.00000995, -0.00000545, -0.00000769, 0.00001025, 0.00000386, -0.00000218, 0.00000160, 0.00000178, -0.00000183, -0.00000525, 0.00000080, -0.00000102, -0.00001018, -0.00000029, 0.00000791, -0.00000404, -0.00001277, 0.00000555, 0.00000447, -0.00001078, -0.00000341, 0.00000519, 0.00000963, -0.00000718, 0.00000126, 0.00000070, 0.00000055, 0.00000082, -0.00001009, 0.00000238, -0.00000439, 0.00000587, -0.00000576, 0.00000191, -0.00000693, -0.00000159, -0.00001031, 0.00000090, -0.00001723, -0.00003035, -0.00000002, 0.00005153, 0.00002732, 0.00003700, 0.00004560, -0.00000862, -0.00003631, -0.00001129, -0.00000712, -0.00000432, -0.00000065, -0.00000078, -0.00000232, -0.00000357, -0.00000464, 0.00000361, 0.00000100, -0.00000285, -0.00000561, 0.00000676, -0.00000167, -0.00000374, -0.00000117, -0.00000354, -0.00000502, -0.00000044, -0.00000797, -0.00000263, -0.00002472, -0.00001311, -0.00010112, 0.00008045, 0.00010154, 0.00011105, 0.00004449, -0.00009852, 0.00000101, -0.00002307, -0.00000074, -0.00001429, 0.00000115, -0.00000830, 0.00000488, -0.00000634, 0.00000110, -0.00000552, 0.00000284, -0.00000382, -0.00000531, -0.00000076, 0.00000382, -0.00000400, 0.00000343, -0.00000190, 0.00001995, -0.00002233, 0.00000845, -0.00000661, 0.00000250, 0.00000041, -0.00000984, -0.00000177, 0.00000256, -0.00000647, 0.00000002, -0.00000775, 0.00000268, -0.00000715, -0.00000315, 0.00000593, -0.00000305, 0.00001846, -0.00000373, -0.00000477, -0.00000324, 0.00000588, 0.00000922, -0.00000187, -0.00000588, 0.00000418, -0.00000412, 0.00000210, -0.00000249, 0.00000005, },
		{ 0.00000304, 0.00000176, -0.00001292, 0.00000714, 0.00001405, -0.00000308, -0.00000904, 0.00000570, 0.00000888, -0.00000354, 0.00000232, 0.00000036, 0.00000058, -0.00000611, -0.00000028, 0.00000023, -0.00000840, -0.00000468, 0.00000859, -0.00000122, -0.00001252, 0.00000063, 0.00000826, -0.00000959, -0.00000566, 0.00000252, 0.00001161, -0.00000437, -0.00000229, 0.00000322, -0.00000179, 0.00000365, -0.00001069, 0.00000139, -0.00000406, 0.00000447, -0.00000352, 0.00000019, -0.00000454, -0.00000331, -0.00000799, -0.00000295, -0.00000906, -0.00003147, -0.00001453, 0.00004603, 0.00003792, 0.00002960, 0.00005387, 0.00000303, -0.00003774, -0.00001639, -0.00000644, -0.00000653, -0.00000019, -0.00000133, -0.00000164, -0.00000332, -0.00000494, 0.00000149, 0.00000303, -0.00000246, -0.00000586, 0.00000381, 0.00000244, -0.00000552, -0.00000080, -0.00000271, -0.00000601, 0.00000090, -0.00000890, 0.00000016, -0.00002492, -0.00000555, -0.00009230, 0.00003120, 0.00011759, 0.00008377, 0.00009454, -0.00009211, -0.00001706, -0.00001610, -0.00000595, -0.00001128, -0.00000189, -0.00000708, 0.00000291, -0.00000418, -0.00000022, -0.00000440, 0.00000131, -0.00000230, -0.00000489, -0.00000357, 0.00000603, -0.00000537, 0.00000468, -0.00000527, 0.00002128, -0.00001589, -0.00000055, 0.00000052, -0.00000345, 0.00000532, -0.00001128, -0.00000204, 0.00000183, -0.00000440, -0.00000248, -0.00000425, -0.00000251, -0.00000032, -0.00000995, 0.00000955, -0.00000616, 0.00001664, 0.00000240, -0.00000668, -0.00000352, 0.00000338, 0.00001038, 0.00000084, -0.00000747, 0.00000364, -0.00000270, 0.00000071, -0.00000116, -0.00000102, },
		{ 0.00000214, 0.00000340, -0.00001076, -0.00000000, 0.00001720, -0.00000058, -0.00000823, -0.00000041, 0.00001301, -0.00000348, 0.00000242, -0.00000045, 0.00000173, -0.00000544, -0.00000201, 0.00000091, -0.00000598, -0.00000803, 0.00000740, 0.00000196, -0.00001149, -0.00000380, 0.00000973, -0.00000644, -0.00000788, -0.00000005, 0.00001167, 0.00000008, -0.00000582, 0.00000484, -0.00000284, 0.00000460, -0.00000917, -0.00000113, -0.00000244, 0.00000192, -0.00000069, -0.00000191, -0.00000190, -0.00000511, -0.00000556, -0.00000625, -0.00000422, -0.00002718, -0.00002750, 0.00003431, 0.00004910, 0.00002505, 0.00005668, 0.00001810, -0.00003742, -0.00002217, -0.00000603, -0.00000859, -0.00000011, -0.00000182, -0.00000093, -0.00000352, -0.00000407, -0.00000162, 0.00000512, -0.00000235, -0.00000481, -0.00000024, 0.00000616, -0.00000616, -0.00000115, -0.00000194, -0.00000616, 0.00000060, -0.00000749, -0.00000047, -0.00002086, -0.00000609, -0.00007233, -0.00001696, 0.00012595, 0.00006503, 0.00011840, -0.00005826, -0.00004510, -0.00000619, -0.00001344, -0.00000636, -0.00000651, -0.00000423, -0.00000084, -0.00000070, -0.00000265, -0.00000215, -0.00000105, -0.00000041, -0.00000485, -0.00000502, 0.00000582, -0.00000399, 0.00000322, -0.00000508, 0.00001748, -0.00000522, -0.00001061, 0.00000670, -0.00000793, 0.00000846, -0.00001130, -0.00000274, 0.00000082, -0.00000184, -0.00000544, -0.00000060, -0.00000693, 0.00000407, -0.00001248, 0.00000834, -0.00000422, 0.00001037, 0.00000994, -0.00000806, -0.00000354, 0.00000056, 0.00001093, 0.00000361, -0.00000788, 0.00000181, -0.00000059, -0.00000087, 0.00000012, -0.00000182, },
		{ 0.00000151, 0.00000353, -0.00000673, -0.00000684, 0.00001817, 0.00000255, -0.00000634, -0.00000594, 0.00001423, -0.00000084, 0.00000101, 0.00000029, 0.00000098, -0.00000320, -0.00000421, 0.00000124, -0.00000383, -0.00000947, 0.00000420, 0.00000532, -0.00001005, -0.00000702, 0.00000877, -0.00000207, -0.00000952, -0.00000240, 0.00000997, 0.00000521, -0.00000825, 0.00000491, -0.00000243, 0.00000395, -0.00000636, -0.00000411, -0.00000073, -0.00000045, 0.00000143, -0.00000321, -0.00000024, -0.00000573, -0.00000440, -0.00000754, -0.00000323, -0.00001967, -0.00003594, 0.00001691, 0.00005789, 0.00002539, 0.00005397, 0.00003477, -0.00003382, -0.00002888, -0.00000595, -0.00001034, -0.00000053, -0.00000220, -0.00000026, -0.00000400, -0.00000263, -0.00000461, 0.00000618, -0.00000162, -0.00000363, -0.00000364, 0.00000774, -0.00000459, -0.00000276, -0.00000084, -0.00000603, -0.00000042, -0.00000517, -0.00000246, -0.00001526, -0.00001025, -0.00004967, -0.00005410, 0.00011850, 0.00006220, 0.00011511, -0.00000674, -0.00007109, -0.00000120, -0.00001822, -0.00000329, -0.00000977, -0.00000219, -0.00000412, 0.00000201, -0.00000439, -0.00000041, -0.00000281, 0.00000081, -0.00000458, -0.00000558, 0.00000398, -0.00000112, 0.00000057, -0.00000304, 0.00001116, 0.00000585, -0.00001783, 0.00000903, -0.00000897, 0.00000849, -0.00000894, -0.00000477, 0.00000061, -0.00000014, -0.00000711, 0.00000106, -0.00000828, 0.00000414, -0.00000998, 0.00000280, 0.00000112, 0.00000239, 0.00001606, -0.00000725, -0.00000420, -0.00000173, 0.00001023, 0.00000660, -0.00000732, -0.00000070, 0.00000145, -0.00000207, 0.00000099, -0.00000229, },
		{ 0.00000102, 0.00000275, -0.00000236, -0.00001159, 0.00001604, 0.00000667, -0.00000451, -0.00000912, 0.00001155, 0.00000434, -0.00000164, 0.00000231, -0.00000102, -0.00000037, -0.00000601, 0.00000099, -0.00000219, -0.00000909, -0.00000013, 0.00000799, -0.00000796, -0.00000915, 0.00000620, 0.00000220, -0.00000983, -0.00000467, 0.00000709, 0.00000981, -0.00000866, 0.00000322, -0.00000082, 0.00000235, -0.00000317, -0.00000656, 0.00000023, -0.00000189, 0.00000229, -0.00000332, 0.00000014, -0.00000501, -0.00000456, -0.00000710, -0.00000458, -0.00001232, -0.00003760, -0.00000407, 0.00006107, 0.00003144, 0.00004720, 0.00005042, -0.00002550, -0.00003631, -0.00000660, -0.00001151, -0.00000162, -0.00000234, 0.00000015, -0.00000431, -0.00000148, -0.00000635, 0.00000531, 0.00000032, -0.00000318, -0.00000526, 0.00000652, -0.00000088, -0.00000521, 0.00000038, -0.00000579, -0.00000155, -0.00000306, -0.00000413, -0.00001034, -0.00001433, -0.00003053, -0.00007504, 0.00009400, 0.00007351, 0.00009374, 0.00004638, -0.00008232, -0.00000710, -0.00001720, -0.00000431, -0.00000993, -0.00000244, -0.00000546, 0.00000259, -0.00000430, -0.00000019, -0.00000310, 0.00000076, -0.00000364, -0.00000587, 0.00000164, 0.00000164, -0.00000156, -0.00000107, 0.00000503, 0.00001403, -0.00001965, 0.00000635, -0.00000632, 0.00000551, -0.00000439, -0.00000801, 0.00000130, 0.00000032, -0.00000687, 0.00000009, -0.00000632, 0.00000055, -0.00000434, -0.00000424, 0.00000651, -0.00000378, 0.00001813, -0.00000318, -0.00000589, -0.00000293, 0.00000802, 0.00000973, -0.00000596, -0.00000331, 0.00000281, -0.00000246, 0.00000125, -0.00000233, },
		{ 0.00000049, 0.00000178, 0.00000104, -0.00001319, 0.00001083, 0.00001145, -0.00000315, -0.00000945, 0.00000563, 0.00001048, -0.00000403, 0.00000421, -0.00000271, 0.00000147, -0.00000634, -0.00000026, -0.00000084, -0.00000769, -0.00000420, 0.00000887, -0.00000483, -0.00001065, 0.00000316, 0.00000515, -0.00000836, -0.00000687, 0.00000370, 0.00001284, -0.00000667, 0.00000012, 0.00000135, 0.00000053, -0.00000035, -0.00000797, 0.00000020, -0.00000236, 0.00000208, -0.00000263, -0.00000027, -0.00000362, -0.00000513, -0.00000619, -0.00000607, -0.00000784, -0.00003251, -0.00002473, 0.00005578, 0.00004251, 0.00003891, 0.00006237, -0.00001185, -0.00004349, -0.00000870, -0.00001173, -0.00000356, -0.00000211, 0.00000002, -0.00000402, -0.00000131, -0.00000622, 0.00000242, 0.00000325, -0.00000346, -0.00000507, 0.00000313, 0.00000366, -0.00000725, 0.00000091, -0.00000513, -0.00000267, -0.00000158, -0.00000486, -0.00000698, -0.00001662, -0.00001756, -0.00008005, 0.00005755, 0.00009062, 0.00006807, 0.00008574, -0.00007132, -0.00002423, -0.00001139, -0.00000869, -0.00000752, -0.00000467, -0.00000491, 0.00000111, -0.00000256, -0.00000134, -0.00000211, -0.00000033, -0.00000218, -0.00000615, -0.00000038, 0.00000324, -0.00000219, -0.00000023, 0.00000067, 0.00001778, -0.00001583, -0.00000018, -0.00000157, 0.00000105, 0.00000090, -0.00001123, 0.00000192, 0.00000024, -0.00000533, -0.00000257, -0.00000266, -0.00000420, 0.00000114, -0.00000928, 0.00000877, -0.00000564, 0.00001516, 0.00000366, -0.00000804, -0.00000324, 0.00000470, 0.00001245, -0.00000373, -0.00000564, 0.00000318, -0.00000199, 0.00000091, -0.00000195, },
		{ -0.00000013, 0.00000111, 0.00000281, -0.00001170, 0.00000369, 0.00001572, -0.00000181, -0.00000772, -0.00000142, 0.00001503, -0.00000418, 0.00000438, -0.00000255, 0.00000117, -0.00000467, -0.00000255, 0.00000040, -0.00000613, -0.00000687, 0.00000745, -0.00000072, -0.00001174, 0.00000041, 0.00000625, -0.00000542, -0.00000862, 0.00000020, 0.00001386, -0.00000267, -0.00000348, 0.00000321, -0.00000083, 0.00000162, -0.00000823, -0.00000065, -0.00000221, 0.00000136, -0.00000178, -0.00000079, -0.00000238, -0.00000523, -0.00000588, -0.00000637, -0.00000691, -0.00002329, -0.00004024, 0.00004055, 0.00005612, 0.00003232, 0.00006856, 0.00000658, -0.00004870, -0.00001308, -0.00001078, -0.00000630, -0.00000152, -0.00000073, -0.00000297, -0.00000227, -0.00000441, -0.00000166, 0.00000614, -0.00000369, -0.00000400, -0.00000086, 0.00000719, -0.00000747, -0.00000000, -0.00000369, -0.00000392, -0.00000061, -0.00000486, -0.00000493, -0.00001702, -0.00001055, -0.00007323, 0.00001756, 0.00010348, 0.00005013, 0.00010272, -0.00003942, -0.00004643, -0.00000552, -0.00001335, -0.00000483, -0.00000717, -0.00000380, -0.00000120, -0.00000036, -0.00000281, -0.00000077, -0.00000161, -0.00000080, -0.00000622, -0.00000195, 0.00000354, -0.00000144, -0.00000039, -0.00000179, 0.00001746, -0.00000822, -0.00000780, 0.00000267, -0.00000270, 0.00000503, -0.00001280, 0.00000118, 0.00000067, -0.00000369, -0.00000526, 0.00000042, -0.00000734, 0.00000367, -0.00001013, 0.00000662, -0.00000298, 0.00000840, 0.00001120, -0.00000913, -0.00000350, 0.00000123, 0.00001388, -0.00000042, -0.00000757, 0.00000262, -0.00000092, 0.00000025, -0.00000131, },
		{ -0.00000074, 0.00000079, 0.00000308, -0.00000817, -0.00000352, 0.00001798, 0.00000040, -0.00000544, -0.00000714, 0.00001582, -0.00000092, 0.00000215, -0.00000013, -0.00000118, -0.00000163, -0.00000512, 0.00000123, -0.00000475, -0.00000783, 0.00000415, 0.00000356, -0.00001205, -0.00000205, 0.00000593, -0.00000206, -0.00000923, -0.00000334, 0.00001300, 0.00000236, -0.00000640, 0.00000398, -0.00000128, 0.00000256, -0.00000745, -0.00000199, -0.00000181, 0.00000055, -0.00000112, -0.00000112, -0.00000162, -0.00000460, -0.00000628, -0.00000556, -0.00000812, -0.00001402, -0.00004672, 0.00001660, 0.00006798, 0.00003064, 0.00006802, 0.00002810, -0.00004995, -0.00002021, -0.00000893, -0.00000946, -0.00000091, -0.00000187, -0.00000144, -0.00000391, -0.00000190, -0.00000549, 0.00000763, -0.00000286, -0.00000320, -0.00000384, 0.00000826, -0.00000517, -0.00000245, -0.00000159, -0.00000527, 0.00000004, -0.00000453, -0.00000358, -0.00001617, -0.00000788, -0.00005996, -0.00001797, 0.00010485, 0.00004578, 0.00009786, 0.00000408, -0.00006373, -0.00000531, -0.00001503, -0.00000416, -0.00000820, -0.00000354, -0.00000296, 0.00000100, -0.00000350, -0.00000008, -0.00000222, -0.00000014, -0.00000572, -0.00000330, 0.00000305, -0.00000017, -0.00000074, -0.00000307, 0.00001459, 0.00000040, -0.00001347, 0.00000418, -0.00000418, 0.00000670, -0.00001171, -0.00000157, 0.00000213, -0.00000272, -0.00000675, 0.00000139, -0.00000743, 0.00000236, -0.00000694, 0.00000126, 0.00000210, 0.00000087, 0.00001654, -0.00000755, -0.00000461, -0.00000137, 0.00001326, 0.00000402, -0.00000893, 0.00000131, 0.00000033, -0.00000036, -0.00000068, },
		{ -0.00000117, 0.00000065, 0.00000245, -0.00000408, -0.00000896, 0.00001712, 0.00000402, -0.00000379, -0.00000988, 0.00001219, 0.00000522, -0.00000155, 0.00000341, -0.00000407, 0.00000114, -0.00000666, 0.00000101, -0.00000335, -0.00000762, 0.00000023, 0.00000670, -0.00001086, -0.00000465, 0.00000510, 0.00000049, -0.00000813, -0.00000681, 0.00001069, 0.00000727, -0.00000765, 0.00000323, -0.00000075, 0.00000256, -0.00000589, -0.00000356, -0.00000135, -0.00000024, -0.00000063, -0.00000136, -0.00000114, -0.00000368, -0.00000671, -0.00000477, -0.00000923, -0.00000826, -0.00004316, -0.00001158, 0.00007268, 0.00003620, 0.00006134, 0.00005005, -0.00004540, -0.00002981, -0.00000698, -0.00001232, -0.00000079, -0.00000293, -0.00000004, -0.00000536, 0.00000005, -0.00000764, 0.00000673, -0.00000046, -0.00000331, -0.00000498, 0.00000654, -0.00000085, -0.00000557, 0.00000045, -0.00000626, 0.00000026, -0.00000405, -0.00000262, -0.00001458, -0.00000785, -0.00004487, -0.00004389, 0.00009257, 0.00005375, 0.00007913, 0.00004613, -0.00006708, -0.00001388, -0.00001265, -0.00000626, -0.00000722, -0.00000463, -0.00000355, 0.00000092, -0.00000297, -0.00000041, -0.00000186, -0.00000039, -0.00000456, -0.00000466, 0.00000235, 0.00000078, -0.00000053, -0.00000393, 0.00001072, 0.00000769, -0.00001534, 0.00000222, -0.00000321, 0.00000592, -0.00000823, -0.00000579, 0.00000407, -0.00000223, -0.00000686, 0.00000016, -0.00000491, -0.00000146, -0.00000200, -0.00000446, 0.00000641, -0.00000422, 0.00001745, -0.00000262, -0.00000676, -0.00000264, 0.00001051, 0.00000905, -0.00000935, -0.00000059, 0.00000143, -0.00000067, -0.00000026, },
		{ -0.00000136, 0.00000050, 0.00000155, -0.00000065, -0.00001156, 0.00001309, 0.00000871, -0.00000300, -0.00000953, 0.00000537, 0.00001198, -0.00000449, 0.00000594, -0.00000543, 0.00000188, -0.00000605, -0.00000072, -0.00000163, -0.00000705, -0.00000297, 0.00000765, -0.00000779, -0.00000767, 0.00000447, 0.00000158, -0.00000546, -0.00000974, 0.00000725, 0.00001119, -0.00000686, 0.00000120, 0.00000040, 0.00000197, -0.00000389, -0.00000511, -0.00000092, -0.00000104, -0.00000015, -0.00000177, -0.00000062, -0.00000301, -0.00000643, -0.00000495, -0.00000870, -0.00000723, -0.00003219, -0.00003714, 0.00006542, 0.00004915, 0.00005092, 0.00006916, -0.00003382, -0.00004084, -0.00000611, -0.00001412, -0.00000167, -0.00000344, 0.00000063, -0.00000584, 0.00000052, -0.00000740, 0.00000341, 0.00000310, -0.00000408, -0.00000453, 0.00000294, 0.00000395, -0.00000788, 0.00000140, -0.00000631, -0.00000023, -0.00000330, -0.00000208, -0.00001239, -0.00000927, -0.00003098, -0.00005857, 0.00006921, 0.00006783, 0.00005743, 0.00007549, -0.00005264, -0.00002952, -0.00000827, -0.00000974, -0.00000526, -0.00000636, -0.00000338, -0.00000021, -0.00000165, -0.00000141, -0.00000086, -0.00000118, -0.00000303, -0.00000595, 0.00000170, 0.00000107, 0.00000035, -0.00000453, 0.00000675, 0.00001247, -0.00001326, -0.00000221, -0.00000106, 0.00000385, -0.00000369, -0.00000998, 0.00000522, -0.00000139, -0.00000627, -0.00000222, -0.00000152, -0.00000534, 0.00000181, -0.00000769, 0.00000750, -0.00000503, 0.00001354, 0.00000472, -0.00000906, -0.00000301, 0.00000641, 0.00001355, -0.00000819, -0.00000309, 0.00000222, -0.00000061, -0.00000007, },
		{ -0.00000138, 0.00000029, 0.00000077, 0.00000154, -0.00001133, 0.00000695, 0.00001324, -0.00000235, -0.00000733, -0.00000208, 0.00001645, -0.00000438, 0.00000571, -0.00000389, -0.00000012, -0.00000326, -0.00000359, 0.00000021, -0.00000653, -0.00000478, 0.00000625, -0.00000333, -0.00001070, 0.00000396, 0.00000150, -0.00000223, -0.00001123, 0.00000275, 0.00001378, -0.00000437, -0.00000134, 0.00000149, 0.00000129, -0.00000187, -0.00000635, -0.00000066, -0.00000176, 0.00000032, -0.00000231, -0.00000006, -0.00000269, -0.00000535, -0.00000612, -0.00000668, -0.00000944, -0.00001909, -0.00005322, 0.00004431, 0.00006641, 0.00004097, 0.00008205, -0.00001488, -0.00005157, -0.00000750, -0.00001442, -0.00000369, -0.00000327, 0.00000030, -0.00000507, -0.00000068, -0.00000510, -0.00000128, 0.00000654, -0.00000454, -0.00000351, -0.00000094, 0.00000743, -0.00000806, 0.00000055, -0.00000508, -0.00000155, -0.00000216, -0.00000213, -0.00000963, -0.00001134, -0.00001980, -0.00006307, 0.00004003, 0.00008050, 0.00004180, 0.00008681, -0.00002343, -0.00004610, -0.00000597, -0.00001217, -0.00000403, -0.00000752, -0.00000336, -0.00000152, -0.00000039, -0.00000233, 0.00000009, -0.00000192, -0.00000166, -0.00000683, 0.00000094, 0.00000095, 0.00000139, -0.00000451, 0.00000288, 0.00001472, -0.00000847, -0.00000707, 0.00000047, 0.00000201, 0.00000032, -0.00001255, 0.00000436, 0.00000052, -0.00000575, -0.00000445, 0.00000087, -0.00000716, 0.00000258, -0.00000721, 0.00000490, -0.00000202, 0.00000661, 0.00001204, -0.00000983, -0.00000349, 0.00000226, 0.00001620, -0.00000493, -0.00000608, 0.00000264, -0.00000030, -0.00000000, },
		},
		{ { -0.00000148, -0.00000281, 0.00001540, -0.00000047, -0.00001351, 0.00000445, -0.00001033, -0.00000309, -0.00000067, 0.00000324, -0.00001189, 0.00000031, 0.00000320, -0.00000038, 0.00000031, -0.00000307, 0.00000399, -0.00000331, 0.00000079, 0.00000224, 0.00000130, 0.00000071, -0.00000194, -0.00000067, -0.00000083, -0.00000018, 0.00001001, -0.00000107, -0.00000583, -0.00000256, 0.00000819, 0.00000277, 0.00000143, -0.00000212, -0.00000940, -0.00000494, 0.00000487, -0.00000612, 0.00000000, -0.00000785, -0.00001019, -0.00001140, -0.00005635, 0.00001145, 0.00008531, 0.00003905, 0.00007504, 0.00001274, -0.00005534, -0.00001035, -0.00000864, -0.00000732, -0.00001095, -0.00000176, -0.00000188, 0.00000063, -0.00000057, -0.00000249, -0.00000070, -0.00000281, 0.00000102, 0.00000108, -0.00000726, -0.00000123, -0.00000230, -0.00000123, -0.00000726, 0.00000108, 0.00000102, -0.00000281, -0.00000070, -0.00000249, -0.00000057, 0.00000063, -0.00000188, -0.00000176, -0.00001095, -0.00000732, -0.00000864, -0.00001035, -0.00005534, 0.00001274, 0.00007504, 0.00003905, 0.00008531, 0.00001145, -0.00005635, -0.00001140, -0.00001019, -0.00000785, 0.00000000, -0.00000612, 0.00000487, -0.00000494, -0.00000940, -0.00000212, 0.00000143, 0.00000277, 0.00000819, -0.00000256, -0.00000583, -0.00000107, 0.00001001, -0.00000018, -0.00000083, -0.00000067, -0.00000194, 0.00000071, 0.00000130, 0.00000224, 0.00000079, -0.00000331, 0.00000399, -0.00000307, 0.00000031, -0.00000038, 0.00000320, 0.00000031, -0.00001189, 0.00000324, -0.00000067, -0.00000309, -0.00001033, 0.00000445, -0.00001351, -0.00000047, 0.00001540, -0.00000281, },
		{ -0.00000279, -0.00000297, 0.00000837, 0.00001055, -0.00001629, 0.00000205, -0.00000611, -0.00000664, -0.00000061, 0.00000308, -0.00000811, -0.00000573, 0.00000593, -0.00000202, 0.00000298, -0.00000493, 0.00000417, -0.00000223, -0.00000072, 0.00000222, 0.00000191, 0.00000077, -0.00000110, -0.00000194, 0.00000055, -0.00000266, 0.00000930, 0.00000241, -0.00000513, -0.00000503, 0.00000611, 0.00000522, 0.00000091, -0.00000011, -0.00000753, -0.00000847, 0.00000393, -0.00000268, -0.00000360, -0.00000172, -0.00001452, -0.00000444, -0.00004589, -0.00002907, 0.00008404, 0.00005232, 0.00006569, 0.00004645, -0.00005677, -0.00002038, -0.00000780, -0.00000694, -0.00001087, -0.00000478, -0.00000122, -0.00000083, 0.00000091, -0.00000269, -0.00000095, -0.00000200, -0.00000123, 0.00000324, -0.00000672, -0.00000255, -0.00000184, -0.00000115, -0.00000610, -0.00000211, 0.00000296, -0.00000285, -0.00000100, -0.00000189, -0.00000181, 0.00000106, -0.00000158, -0.00000054, -0.00000869, -0.00000888, -0.00000880, -0.00000554, -0.00004471, -0.00001816, 0.00007331, 0.00003897, 0.00007015, 0.00004852, -0.00004881, -0.00002719, -0.00000422, -0.00001277, 0.00000090, -0.00000640, 0.00000242, -0.00000019, -0.00001009, -0.00000464, 0.00000153, 0.00000128, 0.00000805, 0.00000110, -0.00000600, -0.00000375, 0.00000841, 0.00000356, -0.00000241, 0.00000062, -0.00000271, 0.00000079, 0.00000061, 0.00000216, 0.00000169, -0.00000293, 0.00000215, -0.00000004, -0.00000254, 0.00000129, 0.00000015, 0.00000519, -0.00001285, 0.00000107, 0.00000042, -0.00000122, -0.00001076, 0.00000134, -0.00000571, -0.00001126, 0.00001693, 0.00000197, },
		{ -0.00000482, -0.00000128, 0.00000111, 0.00001596, -0.00001121, -0.00000533, -0.00000010, -0.00001001, -0.00000058, 0.00000171, -0.00000330, -0.00001083, 0.00000674, -0.00000240, 0.00000409, -0.00000451, 0.00000223, 0.00000017, -0.00000249, 0.00000206, 0.00000222, 0.00000118, -0.00000042, -0.00000274, 0.00000111, -0.00000340, 0.00000659, 0.00000599, -0.00000394, -0.00000599, 0.00000245, 0.00000747, 0.00000095, 0.00000091, -0.00000504, -0.00001004, 0.00000038, 0.00000153, -0.00000640, 0.00000124, -0.00001299, -0.00000729, -0.00002610, -0.00005782, 0.00005943, 0.00007524, 0.00005268, 0.00007422, -0.00004432, -0.00003546, -0.00000652, -0.00000831, -0.00000855, -0.00000874, -0.00000050, -0.00000238, 0.00000173, -0.00000207, -0.00000172, -0.00000097, -0.00000286, 0.00000359, -0.00000436, -0.00000479, -0.00000087, -0.00000183, -0.00000412, -0.00000511, 0.00000365, -0.00000188, -0.00000165, -0.00000138, -0.00000229, 0.00000033, -0.00000047, -0.00000066, -0.00000537, -0.00001016, -0.00000896, -0.00000444, -0.00003065, -0.00003958, 0.00005842, 0.00004843, 0.00004996, 0.00007076, -0.00002398, -0.00004345, -0.00000369, -0.00001227, -0.00000286, -0.00000328, -0.00000225, 0.00000406, -0.00000901, -0.00000709, 0.00000050, 0.00000117, 0.00000606, 0.00000498, -0.00000532, -0.00000534, 0.00000498, 0.00000742, -0.00000327, 0.00000133, -0.00000295, 0.00000061, 0.00000021, 0.00000183, 0.00000208, -0.00000163, -0.00000022, 0.00000266, -0.00000418, 0.00000185, -0.00000183, 0.00000734, -0.00001022, -0.00000340, 0.00000260, -0.00000117, -0.00000805, -0.00000455, 0.00000191, -0.00001633, 0.00001077, 0.00001022, },
		{ -0.00000368, -0.00000140, -0.00000210, 0.00001356, -0.00000024, -0.00001337, 0.00000372, -0.00001058, -0.00000211, 0.00000056, 0.00000057, -0.00001311, 0.00000456, -0.00000077, 0.00000293, -0.00000160, -0.00000130, 0.00000305, -0.00000390, 0.00000161, 0.00000214, 0.00000193, -0.00000001, -0.00000284, 0.00000058, -0.00000250, 0.00000271, 0.00000886, -0.00000228, -0.00000561, -0.00000164, 0.00000839, 0.00000203, 0.00000110, -0.00000270, -0.00000955, -0.00000422, 0.00000418, -0.00000633, 0.00000001, -0.00000698, -0.00001386, -0.00000972, -0.00006414, 0.00001496, 0.00009611, 0.00004653, 0.00008829, -0.00001659, -0.00005269, -0.00000643, -0.00001068, -0.00000545, -0.00001193, -0.00000096, -0.00000312, 0.00000131, -0.00000067, -0.00000254, -0.00000036, -0.00000326, 0.00000213, -0.00000095, -0.00000690, -0.00000029, -0.00000237, -0.00000243, -0.00000680, 0.00000258, -0.00000004, -0.00000232, -0.00000118, -0.00000214, -0.00000090, 0.00000050, -0.00000086, -0.00000269, -0.00000974, -0.00000978, -0.00000514, -0.00001797, -0.00004855, 0.00003341, 0.00006000, 0.00003468, 0.00007487, 0.00000928, -0.00004991, -0.00001265, -0.00000656, -0.00000914, 0.00000041, -0.00000608, 0.00000551, -0.00000567, -0.00000912, -0.00000150, 0.00000172, 0.00000358, 0.00000761, -0.00000331, -0.00000611, 0.00000090, 0.00000988, -0.00000238, 0.00000082, -0.00000218, -0.00000028, 0.00000049, 0.00000103, 0.00000234, -0.00000028, -0.00000189, 0.00000379, -0.00000377, 0.00000078, -0.00000204, 0.00000638, -0.00000466, -0.00000875, 0.00000457, -0.00000164, -0.00000441, -0.00000956, 0.00000521, -0.00001393, -0.00000044, 0.00001717, },
		{ 0.00000263, -0.00000420, -0.00000117, 0.00000634, 0.00001057, -0.00001634, 0.00000195, -0.00000703, -0.00000557, 0.00000032, 0.00000236, -0.00001184, -0.00000051, 0.00000252, -0.00000008, 0.00000274, -0.00000495, 0.00000508, -0.00000416, 0.00000068, 0.00000177, 0.00000280, 0.00000026, -0.00000227, -0.00000083, -0.00000060, -0.00000120, 0.00001020, -0.00000008, -0.00000439, -0.00000500, 0.00000738, 0.00000390, 0.00000109, -0.00000115, -0.00000741, -0.00000831, 0.00000420, -0.00000374, -0.00000318, -0.00000117, -0.00001670, -0.00000500, -0.00004892, -0.00003435, 0.00009909, 0.00005563, 0.00008581, 0.00002264, -0.00006616, -0.00001040, -0.00001234, -0.00000366, -0.00001270, -0.00000328, -0.00000280, -0.00000015, 0.00000086, -0.00000276, -0.00000063, -0.00000242, -0.00000038, 0.00000220, -0.00000771, -0.00000091, -0.00000208, -0.00000174, -0.00000677, 0.00000001, 0.00000204, -0.00000264, -0.00000124, -0.00000185, -0.00000172, 0.00000043, -0.00000026, -0.00000155, -0.00000745, -0.00001099, -0.00000649, -0.00000924, -0.00004625, 0.00000478, 0.00006597, 0.00002966, 0.00006466, 0.00003977, -0.00004115, -0.00002804, -0.00000113, -0.00001281, 0.00000057, -0.00000611, 0.00000286, -0.00000042, -0.00001041, -0.00000379, 0.00000171, 0.00000208, 0.00000797, 0.00000013, -0.00000640, -0.00000251, 0.00000976, 0.00000065, -0.00000096, -0.00000041, -0.00000189, 0.00000144, -0.00000018, 0.00000266, 0.00000055, -0.00000221, 0.00000301, -0.00000151, -0.00000147, -0.00000087, 0.00000322, 0.00000182, -0.00001271, 0.00000459, -0.00000109, -0.00000194, -0.00001133, 0.00000319, -0.00000655, -0.00001106, 0.00001791, },
		{ 0.00001189, -0.00000669, 0.00000033, -0.00000000, 0.00001515, -0.00001108, -0.00000529, -0.00000101, -0.00000936, 0.00000022, 0.00000234, -0.00000787, -0.00000687, 0.00000587, -0.00000322, 0.00000639, -0.00000666, 0.00000486, -0.00000266, -0.00000090, 0.00000142, 0.00000330, 0.00000078, -0.00000145, -0.00000239, 0.00000114, -0.00000378, 0.00000936, 0.00000280, -0.00000305, -0.00000663, 0.00000439, 0.00000599, 0.00000136, -0.00000040, -0.00000468, -0.00001057, 0.00000162, -0.00000025, -0.00000551, 0.00000095, -0.00001298, -0.00001063, -0.00002460, -0.00006851, 0.00007315, 0.00007964, 0.00007205, 0.00006384, -0.00006783, -0.00002142, -0.00001160, -0.00000487, -0.00001040, -0.00000716, -0.00000196, -0.00000196, 0.00000176, -0.00000204, -0.00000162, -0.00000100, -0.00000271, 0.00000382, -0.00000650, -0.00000290, -0.00000102, -0.00000198, -0.00000543, -0.00000311, 0.00000347, -0.00000224, -0.00000148, -0.00000175, -0.00000179, -0.00000064, 0.00000087, -0.00000141, -0.00000452, -0.00001136, -0.00000840, -0.00000457, -0.00003695, -0.00001997, 0.00006112, 0.00003502, 0.00004757, 0.00005950, -0.00001951, -0.00004128, -0.00000283, -0.00001011, -0.00000434, -0.00000230, -0.00000252, 0.00000472, -0.00001005, -0.00000599, 0.00000053, 0.00000199, 0.00000626, 0.00000415, -0.00000603, -0.00000467, 0.00000705, 0.00000499, -0.00000308, 0.00000153, -0.00000343, 0.00000230, -0.00000114, 0.00000269, 0.00000100, -0.00000151, 0.00000112, 0.00000135, -0.00000358, 0.00000040, -0.00000021, 0.00000667, -0.00001317, 0.00000150, 0.00000123, -0.00000146, -0.00000968, -0.00000226, 0.00000124, -0.00001619, 0.00001107, },
		{ 0.00001879, -0.00000469, -0.00000117, -0.00000162, 0.00001176, 0.00000017, -0.00001356, 0.00000328, -0.00001061, -0.00000125, 0.00000175, -0.00000320, -0.00001200, 0.00000698, -0.00000430, 0.00000710, -0.00000483, 0.00000172, 0.00000061, -0.00000306, 0.00000142, 0.00000290, 0.00000199, -0.00000098, -0.00000320, 0.00000155, -0.00000404, 0.00000609, 0.00000630, -0.00000220, -0.00000606, -0.00000005, 0.00000769, 0.00000191, 0.00000020, -0.00000269, -0.00001014, -0.00000289, 0.00000271, -0.00000587, -0.00000038, -0.00000591, -0.00001807, -0.00000704, -0.00007416, 0.00002105, 0.00010609, 0.00005973, 0.00009443, -0.00005048, -0.00004053, -0.00000822, -0.00000933, -0.00000581, -0.00001136, -0.00000147, -0.00000347, 0.00000172, -0.00000070, -0.00000258, -0.00000006, -0.00000372, 0.00000322, -0.00000342, -0.00000560, 0.00000015, -0.00000248, -0.00000372, -0.00000563, 0.00000346, -0.00000095, -0.00000187, -0.00000177, -0.00000150, -0.00000182, 0.00000142, -0.00000094, -0.00000263, -0.00000972, -0.00001096, -0.00000281, -0.00002569, -0.00003536, 0.00004472, 0.00004638, 0.00003150, 0.00006568, 0.00000751, -0.00004413, -0.00001383, -0.00000276, -0.00001109, 0.00000165, -0.00000677, 0.00000666, -0.00000689, -0.00000824, -0.00000131, 0.00000238, 0.00000402, 0.00000709, -0.00000431, -0.00000589, 0.00000303, 0.00000875, -0.00000382, 0.00000222, -0.00000368, 0.00000197, -0.00000090, 0.00000176, 0.00000168, -0.00000076, -0.00000050, 0.00000318, -0.00000406, 0.00000038, -0.00000208, 0.00000802, -0.00000932, -0.00000429, 0.00000456, -0.00000220, -0.00000617, -0.00000799, 0.00000562, -0.00001439, -0.00000020, },
		{ 0.00001877, 0.00000314, -0.00000552, 0.00000055, 0.00000418, 0.00001083, -0.00001662, 0.00000190, -0.00000751, -0.00000478, 0.00000154, 0.00000028, -0.00001378, 0.00000431, -0.00000208, 0.00000402, 0.00000048, -0.00000340, 0.00000452, -0.00000504, 0.00000161, 0.00000160, 0.00000377, -0.00000098, -0.00000283, 0.00000022, -0.00000203, 0.00000119, 0.00000967, -0.00000172, -0.00000382, -0.00000470, 0.00000824, 0.00000263, 0.00000124, -0.00000212, -0.00000728, -0.00000798, 0.00000421, -0.00000477, -0.00000256, -0.00000081, -0.00001917, -0.00000480, -0.00005348, -0.00003863, 0.00011467, 0.00006245, 0.00010500, -0.00001241, -0.00006404, -0.00000493, -0.00001505, -0.00000126, -0.00001408, -0.00000215, -0.00000437, 0.00000093, 0.00000047, -0.00000261, -0.00000036, -0.00000295, 0.00000070, 0.00000046, -0.00000780, 0.00000047, -0.00000249, -0.00000252, -0.00000675, 0.00000182, 0.00000095, -0.00000223, -0.00000175, -0.00000142, -0.00000221, 0.00000058, 0.00000057, -0.00000242, -0.00000626, -0.00001341, -0.00000312, -0.00001592, -0.00004005, 0.00002096, 0.00005656, 0.00002281, 0.00005968, 0.00003217, -0.00003416, -0.00002909, 0.00000236, -0.00001365, 0.00000100, -0.00000631, 0.00000355, -0.00000090, -0.00001042, -0.00000301, 0.00000190, 0.00000288, 0.00000763, -0.00000088, -0.00000669, -0.00000065, 0.00001003, -0.00000191, 0.00000077, -0.00000194, -0.00000007, 0.00000087, -0.00000028, 0.00000289, -0.00000061, -0.00000093, 0.00000309, -0.00000237, -0.00000133, -0.00000180, 0.00000579, -0.00000248, -0.00001058, 0.00000686, -0.00000243, -0.00000300, -0.00001137, 0.00000512, -0.00000776, -0.00001052, },
		{ 0.00001129, 0.00001346, -0.00000859, 0.00000201, -0.00000137, 0.00001460, -0.00001109, -0.00000538, -0.00000143, -0.00000892, 0.00000133, 0.00000194, -0.00001180, -0.00000171, 0.00000254, -0.00000141, 0.00000689, -0.00000786, 0.00000685, -0.00000544, 0.00000118, 0.00000025, 0.00000517, -0.00000073, -0.00000187, -0.00000197, 0.00000079, -0.00000336, 0.00001131, -0.00000052, -0.00000148, -0.00000769, 0.00000669, 0.00000379, 0.00000241, -0.00000228, -0.00000370, -0.00001153, 0.00000332, -0.00000279, -0.00000372, -0.00000010, -0.00001267, -0.00001407, -0.00002315, -0.00008034, 0.00008921, 0.00008464, 0.00009614, 0.00003883, -0.00008224, -0.00000753, -0.00001816, 0.00000002, -0.00001351, -0.00000459, -0.00000452, -0.00000037, 0.00000101, -0.00000155, -0.00000170, -0.00000108, -0.00000237, 0.00000350, -0.00000820, -0.00000071, -0.00000164, -0.00000217, -0.00000638, -0.00000094, 0.00000275, -0.00000222, -0.00000166, -0.00000175, -0.00000166, -0.00000132, 0.00000246, -0.00000288, -0.00000273, -0.00001406, -0.00000544, -0.00000869, -0.00003674, -0.00000293, 0.00005848, 0.00002455, 0.00004545, 0.00004963, -0.00001536, -0.00003958, -0.00000150, -0.00000858, -0.00000549, -0.00000132, -0.00000300, 0.00000553, -0.00001123, -0.00000453, 0.00000022, 0.00000314, 0.00000601, 0.00000344, -0.00000691, -0.00000317, 0.00000834, 0.00000233, -0.00000214, 0.00000099, -0.00000293, 0.00000318, -0.00000245, 0.00000393, -0.00000071, -0.00000043, 0.00000160, 0.00000054, -0.00000364, -0.00000047, 0.00000188, 0.00000433, -0.00001422, 0.00000588, -0.00000059, -0.00000171, -0.00001127, 0.00000064, -0.00000009, -0.00001565, },
		{ 0.00000006, 0.00002033, -0.00000583, -0.00000076, -0.00000141, 0.00001014, 0.00000057, -0.00001395, 0.00000323, -0.00001066, -0.00000030, 0.00000251, -0.00000752, -0.00000860, 0.00000672, -0.00000592, 0.00001076, -0.00000846, 0.00000543, -0.00000299, -0.00000081, 0.00000003, 0.00000491, 0.00000080, -0.00000149, -0.00000339, 0.00000229, -0.00000528, 0.00000960, 0.00000259, -0.00000074, -0.00000747, 0.00000249, 0.00000575, 0.00000280, -0.00000160, -0.00000172, -0.00001154, -0.00000061, 0.00000001, -0.00000414, -0.00000184, -0.00000409, -0.00002315, -0.00000311, -0.00008684, 0.00003044, 0.00011482, 0.00008055, 0.00008742, -0.00008205, -0.00002219, -0.00001514, -0.00000436, -0.00000885, -0.00000877, -0.00000384, -0.00000223, 0.00000115, -0.00000019, -0.00000288, 0.00000036, -0.00000421, 0.00000418, -0.00000612, -0.00000340, -0.00000016, -0.00000247, -0.00000503, -0.00000386, 0.00000362, -0.00000151, -0.00000165, -0.00000222, -0.00000087, -0.00000299, 0.00000322, -0.00000230, -0.00000116, -0.00001155, -0.00000967, -0.00000371, -0.00002979, -0.00002037, 0.00004875, 0.00003500, 0.00002891, 0.00005779, 0.00000586, -0.00003886, -0.00001484, 0.00000090, -0.00001335, 0.00000351, -0.00000817, 0.00000838, -0.00000866, -0.00000666, -0.00000169, 0.00000351, 0.00000397, 0.00000671, -0.00000563, -0.00000487, 0.00000485, 0.00000689, -0.00000439, 0.00000305, -0.00000463, 0.00000415, -0.00000305, 0.00000348, 0.00000001, -0.00000018, 0.00000024, 0.00000278, -0.00000466, -0.00000001, -0.00000108, 0.00000797, -0.00001290, 0.00000075, 0.00000343, -0.00000236, -0.00000835, -0.00000546, 0.00000532, -0.00001458, },
		{ -0.00000973, 0.00001950, 0.00000351, -0.00000671, 0.00000209, 0.00000205, 0.00001124, -0.00001716, 0.00000209, -0.00000779, -0.00000412, 0.00000287, -0.00000309, -0.00001333, 0.00000739, -0.00000645, 0.00000938, -0.00000366, -0.00000006, 0.00000202, -0.00000416, 0.00000126, 0.00000259, 0.00000382, -0.00000217, -0.00000309, 0.00000124, -0.00000367, 0.00000446, 0.00000744, -0.00000186, -0.00000416, -0.00000346, 0.00000799, 0.00000212, 0.00000071, -0.00000235, -0.00000779, -0.00000678, 0.00000315, -0.00000480, -0.00000269, 0.00000014, -0.00002261, -0.00000306, -0.00006071, -0.00004039, 0.00012917, 0.00007627, 0.00011686, -0.00005413, -0.00005010, -0.00000644, -0.00001363, -0.00000163, -0.00001347, -0.00000275, -0.00000467, 0.00000140, 0.00000036, -0.00000263, 0.00000007, -0.00000370, 0.00000205, -0.00000201, -0.00000671, 0.00000119, -0.00000283, -0.00000347, -0.00000602, 0.00000308, -0.00000009, -0.00000177, -0.00000242, -0.00000065, -0.00000332, 0.00000189, 0.00000010, -0.00000210, -0.00000640, -0.00001438, -0.00000105, -0.00002235, -0.00002878, 0.00002980, 0.00004765, 0.00001706, 0.00005582, 0.00002496, -0.00002738, -0.00003053, 0.00000635, -0.00001543, 0.00000251, -0.00000748, 0.00000498, -0.00000208, -0.00000970, -0.00000273, 0.00000248, 0.00000333, 0.00000736, -0.00000221, -0.00000639, 0.00000135, 0.00000938, -0.00000389, 0.00000247, -0.00000363, 0.00000249, -0.00000109, 0.00000090, 0.00000210, -0.00000099, 0.00000015, 0.00000302, -0.00000324, -0.00000154, -0.00000158, 0.00000722, -0.00000683, -0.00000690, 0.00000772, -0.00000335, -0.00000460, -0.00001049, 0.00000665, -0.00000907, },
		{ -0.00001464, 0.00001131, 0.00001496, -0.00001052, 0.00000372, -0.00000290, 0.00001431, -0.00001133, -0.00000536, -0.00000153, -0.00000872, 0.00000277, 0.00000019, -0.00001416, 0.00000342, -0.00000245, 0.00000311, 0.00000478, -0.00000705, 0.00000710, -0.00000701, 0.00000265, -0.00000047, 0.00000694, -0.00000288, -0.00000174, -0.00000151, 0.00000002, -0.00000187, 0.00001181, -0.00000304, -0.00000008, -0.00000849, 0.00000867, 0.00000151, 0.00000353, -0.00000413, -0.00000270, -0.00001229, 0.00000473, -0.00000526, -0.00000174, -0.00000136, -0.00001261, -0.00001681, -0.00002291, -0.00009186, 0.00010601, 0.00009347, 0.00012069, -0.00000060, -0.00008290, 0.00000114, -0.00002265, 0.00000388, -0.00001602, -0.00000254, -0.00000692, 0.00000155, -0.00000009, -0.00000086, -0.00000179, -0.00000135, -0.00000168, 0.00000240, -0.00000900, 0.00000132, -0.00000255, -0.00000241, -0.00000686, 0.00000116, 0.00000164, -0.00000185, -0.00000223, -0.00000124, -0.00000225, -0.00000098, 0.00000327, -0.00000398, -0.00000118, -0.00001685, -0.00000168, -0.00001524, -0.00003005, 0.00000854, 0.00005407, 0.00001555, 0.00004439, 0.00004023, -0.00001095, -0.00003874, 0.00000068, -0.00000821, -0.00000557, -0.00000111, -0.00000299, 0.00000590, -0.00001200, -0.00000320, -0.00000003, 0.00000421, 0.00000565, 0.00000248, -0.00000750, -0.00000119, 0.00000881, -0.00000030, -0.00000062, -0.00000031, -0.00000120, 0.00000254, -0.00000274, 0.00000464, -0.00000226, 0.00000101, 0.00000160, -0.00000006, -0.00000423, -0.00000036, 0.00000358, 0.00000099, -0.00001336, 0.00000920, -0.00000242, -0.00000225, -0.00001239, 0.00000376, -0.00000198, },
		{ -0.00001422, 0.00000014, 0.00002187, -0.00000712, -0.00000021, -0.00000147, 0.00000879, 0.00000081, -0.00001436, 0.00000354, -0.00001084, 0.00000090, 0.00000239, -0.00001156, -0.00000356, 0.00000356, -0.00000450, 0.00001232, -0.00001107, 0.00000861, -0.00000671, 0.00000206, -0.00000199, 0.00000782, -0.00000172, -0.00000119, -0.00000365, 0.00000276, -0.00000589, 0.00001259, -0.00000156, 0.00000167, -0.00000958, 0.00000574, 0.00000275, 0.00000461, -0.00000418, 0.00000005, -0.00001356, 0.00000237, -0.00000373, -0.00000118, -0.00000445, -0.00000153, -0.00002886, 0.00000157, -0.00010161, 0.00004253, 0.00012378, 0.00010838, 0.00006274, -0.00010353, -0.00000339, -0.00002418, 0.00000252, -0.00001348, -0.00000493, -0.00000755, 0.00000044, -0.00000047, 0.00000097, -0.00000351, 0.00000083, -0.00000461, 0.00000475, -0.00000863, -0.00000059, -0.00000121, -0.00000224, -0.00000628, -0.00000163, 0.00000304, -0.00000158, -0.00000186, -0.00000220, -0.00000075, -0.00000367, 0.00000509, -0.00000429, 0.00000120, -0.00001479, -0.00000615, -0.00000824, -0.00002805, -0.00000791, 0.00004884, 0.00002508, 0.00002735, 0.00005045, 0.00000472, -0.00003445, -0.00001524, 0.00000380, -0.00001515, 0.00000528, -0.00000971, 0.00001024, -0.00001060, -0.00000469, -0.00000240, 0.00000489, 0.00000364, 0.00000623, -0.00000697, -0.00000316, 0.00000611, 0.00000453, -0.00000403, 0.00000299, -0.00000446, 0.00000534, -0.00000500, 0.00000553, -0.00000228, 0.00000129, 0.00000026, 0.00000273, -0.00000587, 0.00000030, 0.00000016, 0.00000652, -0.00001494, 0.00000566, 0.00000156, -0.00000233, -0.00001067, -0.00000208, 0.00000412, },
		{ -0.00001007, -0.00000893, 0.00002016, 0.00000371, -0.00000775, 0.00000342, 0.00000009, 0.00001161, -0.00001780, 0.00000259, -0.00000804, -0.00000344, 0.00000396, -0.00000733, -0.00001031, 0.00000763, -0.00000887, 0.00001436, -0.00000859, 0.00000443, -0.00000206, -0.00000153, -0.00000058, 0.00000517, 0.00000225, -0.00000256, -0.00000346, 0.00000231, -0.00000535, 0.00000829, 0.00000365, -0.00000028, -0.00000581, -0.00000088, 0.00000619, 0.00000287, -0.00000094, -0.00000141, -0.00000936, -0.00000431, 0.00000046, -0.00000314, -0.00000435, 0.00000235, -0.00002751, 0.00000066, -0.00007125, -0.00003877, 0.00014188, 0.00009993, 0.00011387, -0.00009451, -0.00002794, -0.00001447, -0.00000770, -0.00000532, -0.00001031, -0.00000568, -0.00000290, 0.00000051, 0.00000111, -0.00000316, 0.00000082, -0.00000473, 0.00000357, -0.00000506, -0.00000437, 0.00000090, -0.00000281, -0.00000466, -0.00000453, 0.00000356, -0.00000079, -0.00000154, -0.00000291, 0.00000005, -0.00000463, 0.00000409, -0.00000174, -0.00000026, -0.00000850, -0.00001285, -0.00000216, -0.00002522, -0.00001651, 0.00003317, 0.00003976, 0.00001223, 0.00005285, 0.00001808, -0.00002085, -0.00003210, 0.00001041, -0.00001769, 0.00000481, -0.00000954, 0.00000716, -0.00000398, -0.00000815, -0.00000307, 0.00000349, 0.00000338, 0.00000715, -0.00000386, -0.00000532, 0.00000312, 0.00000794, -0.00000504, 0.00000370, -0.00000485, 0.00000502, -0.00000390, 0.00000327, 0.00000010, -0.00000020, 0.00000049, 0.00000331, -0.00000469, -0.00000132, -0.00000076, 0.00000738, -0.00001055, -0.00000226, 0.00000726, -0.00000375, -0.00000678, -0.00000845, 0.00000735, },
		{ -0.00000409, -0.00001327, 0.00001111, 0.00001640, -0.00001244, 0.00000535, -0.00000444, 0.00001414, -0.00001177, -0.00000511, -0.00000144, -0.00000866, 0.00000444, -0.00000297, -0.00001414, 0.00000701, -0.00000756, 0.00000941, 0.00000014, -0.00000387, 0.00000498, -0.00000650, 0.00000257, 0.00000027, 0.00000749, -0.00000474, -0.00000149, -0.00000086, -0.00000125, 0.00000080, 0.00001038, -0.00000391, 0.00000026, -0.00000818, 0.00000947, 0.00000000, 0.00000397, -0.00000519, -0.00000245, -0.00001200, 0.00000487, -0.00000655, -0.00000071, -0.00000182, -0.00001374, -0.00001779, -0.00002535, -0.00010124, 0.00012135, 0.00011072, 0.00013817, -0.00004958, -0.00006893, 0.00000142, -0.00002225, 0.00000443, -0.00001612, -0.00000259, -0.00000770, 0.00000259, -0.00000071, -0.00000049, -0.00000155, -0.00000206, -0.00000047, 0.00000030, -0.00000844, 0.00000258, -0.00000333, -0.00000286, -0.00000668, 0.00000284, 0.00000043, -0.00000132, -0.00000303, -0.00000033, -0.00000356, 0.00000068, 0.00000257, -0.00000369, -0.00000113, -0.00001804, 0.00000062, -0.00002125, -0.00001934, 0.00001410, 0.00004984, 0.00000716, 0.00004476, 0.00003071, -0.00000593, -0.00003889, 0.00000373, -0.00000910, -0.00000426, -0.00000217, -0.00000197, 0.00000538, -0.00001190, -0.00000244, 0.00000011, 0.00000494, 0.00000541, 0.00000106, -0.00000738, 0.00000086, 0.00000847, -0.00000260, 0.00000114, -0.00000198, 0.00000149, 0.00000024, -0.00000141, 0.00000404, -0.00000282, 0.00000203, 0.00000176, -0.00000108, -0.00000472, 0.00000051, 0.00000445, -0.00000268, -0.00001080, 0.00001115, -0.00000388, -0.00000336, -0.00001256, 0.00000656, },
		{ 0.00000211, -0.00001322, -0.00000012, 0.00002347, -0.00000856, 0.00000045, -0.00000170, 0.00000768, 0.00000081, -0.00001460, 0.00000409, -0.00001122, 0.00000246, 0.00000101, -0.00001431, 0.00000189, -0.00000179, 0.00000023, 0.00001069, -0.00001132, 0.00000989, -0.00000916, 0.00000429, -0.00000349, 0.00001060, -0.00000490, -0.00000026, -0.00000389, 0.00000277, -0.00000543, 0.00001424, -0.00000508, 0.00000405, -0.00001158, 0.00000891, -0.00000059, 0.00000673, -0.00000691, 0.00000197, -0.00001547, 0.00000517, -0.00000757, 0.00000216, -0.00000748, 0.00000104, -0.00003421, 0.00000558, -0.00011677, 0.00005524, 0.00013693, 0.00013866, 0.00001984, -0.00010892, 0.00000971, -0.00003116, 0.00000839, -0.00001754, -0.00000161, -0.00001112, 0.00000351, -0.00000252, 0.00000242, -0.00000416, 0.00000107, -0.00000460, 0.00000453, -0.00001032, 0.00000222, -0.00000272, -0.00000188, -0.00000729, 0.00000078, 0.00000186, -0.00000115, -0.00000258, -0.00000155, -0.00000150, -0.00000314, 0.00000598, -0.00000576, 0.00000323, -0.00001804, -0.00000192, -0.00001491, -0.00002062, -0.00000045, 0.00004793, 0.00001543, 0.00002762, 0.00004271, 0.00000472, -0.00003134, -0.00001455, 0.00000532, -0.00001569, 0.00000610, -0.00001063, 0.00001159, -0.00001212, -0.00000285, -0.00000307, 0.00000616, 0.00000332, 0.00000533, -0.00000785, -0.00000110, 0.00000675, 0.00000190, -0.00000281, 0.00000200, -0.00000286, 0.00000475, -0.00000566, 0.00000687, -0.00000431, 0.00000297, 0.00000004, 0.00000260, -0.00000737, 0.00000150, 0.00000100, 0.00000413, -0.00001524, 0.00000982, -0.00000054, -0.00000245, -0.00001267, 0.00000177, },
		{ 0.00000694, -0.00001041, -0.00000839, 0.00002089, 0.00000371, -0.00000864, 0.00000452, -0.00000164, 0.00001178, -0.00001837, 0.00000333, -0.00000838, -0.00000255, 0.00000439, -0.00001167, -0.00000519, 0.00000456, -0.00000791, 0.00001708, -0.00001260, 0.00000867, -0.00000655, 0.00000197, -0.00000328, 0.00000890, -0.00000096, -0.00000185, -0.00000411, 0.00000332, -0.00000668, 0.00001197, -0.00000098, 0.00000258, -0.00000852, 0.00000281, 0.00000293, 0.00000486, -0.00000363, 0.00000058, -0.00001180, -0.00000083, -0.00000364, 0.00000018, -0.00000761, 0.00000582, -0.00003361, 0.00000584, -0.00008455, -0.00003444, 0.00015429, 0.00013351, 0.00008992, -0.00012418, -0.00000419, -0.00002567, 0.00000088, -0.00001118, -0.00000546, -0.00001035, 0.00000070, -0.00000181, 0.00000279, -0.00000420, 0.00000180, -0.00000583, 0.00000495, -0.00000825, -0.00000109, -0.00000045, -0.00000230, -0.00000608, -0.00000234, 0.00000315, -0.00000093, -0.00000180, -0.00000288, 0.00000015, -0.00000533, 0.00000629, -0.00000425, 0.00000256, -0.00001207, -0.00000903, -0.00000682, -0.00002254, -0.00000706, 0.00003404, 0.00003201, 0.00000904, 0.00004986, 0.00001210, -0.00001510, -0.00003322, 0.00001378, -0.00001955, 0.00000709, -0.00001184, 0.00000962, -0.00000620, -0.00000611, -0.00000380, 0.00000474, 0.00000325, 0.00000672, -0.00000545, -0.00000361, 0.00000449, 0.00000581, -0.00000516, 0.00000402, -0.00000490, 0.00000647, -0.00000647, 0.00000604, -0.00000263, 0.00000153, 0.00000013, 0.00000394, -0.00000685, -0.00000014, -0.00000003, 0.00000647, -0.00001313, 0.00000268, 0.00000579, -0.00000380, -0.00000932, -0.00000528, },
		{ 0.00000849, -0.00000596, -0.00001182, 0.00001079, 0.00001775, -0.00001431, 0.00000686, -0.00000585, 0.00001388, -0.00001223, -0.00000463, -0.00000129, -0.00000864, 0.00000609, -0.00000731, -0.00001131, 0.00000768, -0.00001079, 0.00001564, -0.00000582, 0.00000105, 0.00000077, -0.00000384, 0.00000064, 0.00000298, 0.00000601, -0.00000549, -0.00000160, 0.00000010, -0.00000292, 0.00000445, 0.00000701, -0.00000275, -0.00000103, -0.00000621, 0.00000843, -0.00000005, 0.00000313, -0.00000492, -0.00000348, -0.00001016, 0.00000305, -0.00000578, -0.00000159, -0.00000062, -0.00001674, -0.00001632, -0.00003140, -0.00010730, 0.00013396, 0.00014054, 0.00013926, -0.00009897, -0.00004407, -0.00000654, -0.00001627, 0.00000087, -0.00001304, -0.00000554, -0.00000591, 0.00000185, -0.00000016, -0.00000089, -0.00000074, -0.00000333, 0.00000127, -0.00000275, -0.00000629, 0.00000260, -0.00000357, -0.00000370, -0.00000563, 0.00000375, -0.00000049, -0.00000098, -0.00000366, 0.00000050, -0.00000507, 0.00000332, 0.00000030, -0.00000167, -0.00000319, -0.00001661, -0.00000031, -0.00002363, -0.00000840, 0.00001562, 0.00004592, -0.00000041, 0.00004612, 0.00002118, -0.00000054, -0.00003966, 0.00000717, -0.00001074, -0.00000190, -0.00000437, 0.00000005, 0.00000391, -0.00001084, -0.00000236, 0.00000068, 0.00000532, 0.00000521, -0.00000072, -0.00000643, 0.00000270, 0.00000735, -0.00000419, 0.00000259, -0.00000332, 0.00000427, -0.00000311, 0.00000140, 0.00000195, -0.00000202, 0.00000214, 0.00000251, -0.00000303, -0.00000432, 0.00000156, 0.00000443, -0.00000609, -0.00000698, 0.00001166, -0.00000478, -0.00000516, -0.00001144, },
		{ 0.00000547, -0.00000034, -0.00001167, -0.00000074, 0.00002514, -0.00001013, 0.00000113, -0.00000194, 0.00000663, 0.00000069, -0.00001463, 0.00000480, -0.00001177, 0.00000437, -0.00000188, -0.00001476, 0.00000610, -0.00000770, 0.00000736, 0.00000584, -0.00000860, 0.00000841, -0.00000926, 0.00000473, -0.00000328, 0.00001191, -0.00000767, 0.00000069, -0.00000380, 0.00000208, -0.00000357, 0.00001382, -0.00000691, 0.00000530, -0.00001243, 0.00001093, -0.00000325, 0.00000824, -0.00000889, 0.00000310, -0.00001626, 0.00000660, -0.00001014, 0.00000450, -0.00000967, 0.00000241, -0.00003779, 0.00000708, -0.00013002, 0.00006561, 0.00016014, 0.00016273, -0.00003640, -0.00009639, 0.00001292, -0.00003252, 0.00001047, -0.00001881, -0.00000074, -0.00001275, 0.00000550, -0.00000398, 0.00000342, -0.00000432, 0.00000067, -0.00000384, 0.00000310, -0.00001052, 0.00000423, -0.00000412, -0.00000168, -0.00000772, 0.00000293, 0.00000043, -0.00000047, -0.00000358, -0.00000046, -0.00000306, -0.00000112, 0.00000514, -0.00000567, 0.00000365, -0.00001964, 0.00000081, -0.00002090, -0.00000988, 0.00000193, 0.00004744, 0.00000549, 0.00002999, 0.00003404, 0.00000615, -0.00002969, -0.00001271, 0.00000534, -0.00001460, 0.00000542, -0.00001034, 0.00001190, -0.00001271, -0.00000158, -0.00000336, 0.00000710, 0.00000318, 0.00000383, -0.00000791, 0.00000095, 0.00000676, -0.00000066, -0.00000108, 0.00000039, -0.00000007, 0.00000220, -0.00000436, 0.00000664, -0.00000517, 0.00000404, 0.00000024, 0.00000172, -0.00000840, 0.00000322, 0.00000117, 0.00000129, -0.00001386, 0.00001275, -0.00000243, -0.00000308, -0.00001380, },
		{ -0.00000138, 0.00000546, -0.00000992, -0.00000831, 0.00002179, 0.00000348, -0.00000943, 0.00000550, -0.00000318, 0.00001174, -0.00001871, 0.00000425, -0.00000886, -0.00000132, 0.00000368, -0.00001506, 0.00000085, -0.00000128, -0.00000305, 0.00001610, -0.00001399, 0.00001098, -0.00000985, 0.00000491, -0.00000552, 0.00001257, -0.00000500, -0.00000036, -0.00000483, 0.00000391, -0.00000704, 0.00001452, -0.00000525, 0.00000566, -0.00001138, 0.00000676, -0.00000102, 0.00000745, -0.00000668, 0.00000287, -0.00001428, 0.00000266, -0.00000805, 0.00000413, -0.00001160, 0.00000964, -0.00003972, 0.00001085, -0.00009858, -0.00003001, 0.00017104, 0.00017249, 0.00004313, -0.00013532, 0.00001365, -0.00003505, 0.00000867, -0.00001663, -0.00000100, -0.00001500, 0.00000486, -0.00000474, 0.00000487, -0.00000534, 0.00000257, -0.00000654, 0.00000559, -0.00001077, 0.00000242, -0.00000247, -0.00000144, -0.00000748, 0.00000027, 0.00000197, -0.00000048, -0.00000262, -0.00000214, -0.00000073, -0.00000467, 0.00000739, -0.00000621, 0.00000509, -0.00001573, -0.00000444, -0.00001359, -0.00001454, -0.00000256, 0.00003496, 0.00002327, 0.00000844, 0.00004574, 0.00000776, -0.00001072, -0.00003327, 0.00001576, -0.00002010, 0.00000837, -0.00001351, 0.00001161, -0.00000808, -0.00000411, -0.00000454, 0.00000588, 0.00000323, 0.00000571, -0.00000647, -0.00000165, 0.00000544, 0.00000315, -0.00000420, 0.00000323, -0.00000335, 0.00000594, -0.00000757, 0.00000803, -0.00000508, 0.00000342, -0.00000035, 0.00000436, -0.00000925, 0.00000203, 0.00000013, 0.00000482, -0.00001429, 0.00000724, 0.00000379, -0.00000382, -0.00001175, },
		{ -0.00000898, 0.00000917, -0.00000714, -0.00001062, 0.00001052, 0.00001892, -0.00001612, 0.00000823, -0.00000705, 0.00001340, -0.00001250, -0.00000401, -0.00000116, -0.00000851, 0.00000732, -0.00001215, -0.00000594, 0.00000467, -0.00001039, 0.00001956, -0.00001113, 0.00000612, -0.00000426, 0.00000005, -0.00000246, 0.00000718, 0.00000251, -0.00000478, -0.00000225, 0.00000126, -0.00000456, 0.00000839, 0.00000237, 0.00000006, -0.00000377, -0.00000269, 0.00000556, 0.00000145, 0.00000102, -0.00000338, -0.00000563, -0.00000698, -0.00000055, -0.00000296, -0.00000451, 0.00000231, -0.00002144, -0.00001285, -0.00004055, -0.00011071, 0.00014510, 0.00018421, 0.00011571, -0.00013746, -0.00001590, -0.00001910, -0.00000666, -0.00000559, -0.00000768, -0.00001076, -0.00000177, -0.00000077, 0.00000166, -0.00000207, 0.00000053, -0.00000491, 0.00000315, -0.00000630, -0.00000284, 0.00000124, -0.00000305, -0.00000503, -0.00000369, 0.00000366, -0.00000081, -0.00000116, -0.00000373, 0.00000067, -0.00000592, 0.00000599, -0.00000275, 0.00000150, -0.00000692, -0.00001272, -0.00000486, -0.00002052, -0.00000073, 0.00001596, 0.00004118, -0.00000607, 0.00004730, 0.00001240, 0.00000456, -0.00004033, 0.00001014, -0.00001215, 0.00000060, -0.00000701, 0.00000256, 0.00000193, -0.00000915, -0.00000275, 0.00000147, 0.00000560, 0.00000469, -0.00000241, -0.00000488, 0.00000426, 0.00000542, -0.00000472, 0.00000316, -0.00000350, 0.00000598, -0.00000629, 0.00000480, -0.00000106, -0.00000015, 0.00000148, 0.00000372, -0.00000593, -0.00000259, 0.00000213, 0.00000370, -0.00000879, -0.00000247, 0.00001091, -0.00000519, -0.00000751, },
		{ -0.00001360, 0.00000837, -0.00000269, -0.00000990, -0.00000159, 0.00002684, -0.00001179, 0.00000174, -0.00000204, 0.00000542, 0.00000065, -0.00001451, 0.00000561, -0.00001244, 0.00000648, -0.00000622, -0.00001225, 0.00000742, -0.00001196, 0.00001488, -0.00000101, -0.00000349, 0.00000429, -0.00000679, 0.00000294, -0.00000071, 0.00001083, -0.00000900, 0.00000104, -0.00000314, 0.00000063, -0.00000033, 0.00001112, -0.00000645, 0.00000467, -0.00001134, 0.00001092, -0.00000433, 0.00000840, -0.00000941, 0.00000278, -0.00001523, 0.00000578, -0.00001036, 0.00000461, -0.00000988, 0.00000158, -0.00003857, 0.00000474, -0.00013969, 0.00007141, 0.00019935, 0.00016922, -0.00009577, -0.00006954, 0.00000581, -0.00002702, 0.00000750, -0.00001616, -0.00000341, -0.00001124, 0.00000523, -0.00000392, 0.00000337, -0.00000364, -0.00000059, -0.00000218, 0.00000032, -0.00000883, 0.00000477, -0.00000481, -0.00000202, -0.00000725, 0.00000434, -0.00000078, 0.00000005, -0.00000443, 0.00000057, -0.00000485, 0.00000202, 0.00000253, -0.00000366, 0.00000181, -0.00001848, 0.00000025, -0.00002321, 0.00000053, 0.00000126, 0.00004708, -0.00000414, 0.00003387, 0.00002466, 0.00000875, -0.00002913, -0.00001020, 0.00000435, -0.00001221, 0.00000335, -0.00000882, 0.00001109, -0.00001225, -0.00000103, -0.00000325, 0.00000774, 0.00000308, 0.00000188, -0.00000707, 0.00000279, 0.00000604, -0.00000269, 0.00000056, -0.00000106, 0.00000301, -0.00000168, -0.00000123, 0.00000461, -0.00000445, 0.00000401, 0.00000130, -0.00000047, -0.00000813, 0.00000475, 0.00000073, -0.00000158, -0.00001103, 0.00001423, -0.00000381, -0.00000445, },
		{ -0.00001345, 0.00000263, 0.00000328, -0.00000871, -0.00000872, 0.00002289, 0.00000302, -0.00001020, 0.00000653, -0.00000475, 0.00001164, -0.00001884, 0.00000528, -0.00000953, 0.00000035, 0.00000144, -0.00001634, 0.00000606, -0.00000825, 0.00000494, 0.00001114, -0.00001186, 0.00001017, -0.00001057, 0.00000587, -0.00000584, 0.00001465, -0.00000857, 0.00000114, -0.00000515, 0.00000365, -0.00000585, 0.00001496, -0.00000786, 0.00000763, -0.00001312, 0.00000966, -0.00000442, 0.00000956, -0.00000903, 0.00000436, -0.00001558, 0.00000472, -0.00001117, 0.00000704, -0.00001473, 0.00001226, -0.00004404, 0.00001335, -0.00011043, -0.00002949, 0.00019945, 0.00020722, -0.00002207, -0.00012465, 0.00002001, -0.00003808, 0.00001216, -0.00001891, 0.00000067, -0.00001744, 0.00000777, -0.00000697, 0.00000641, -0.00000586, 0.00000251, -0.00000627, 0.00000484, -0.00001172, 0.00000515, -0.00000443, -0.00000069, -0.00000844, 0.00000278, 0.00000040, 0.00000030, -0.00000377, -0.00000091, -0.00000251, -0.00000237, 0.00000659, -0.00000653, 0.00000601, -0.00001777, -0.00000126, -0.00001970, -0.00000359, -0.00000280, 0.00003698, 0.00001317, 0.00001077, 0.00003988, 0.00000550, -0.00000802, -0.00003207, 0.00001610, -0.00001886, 0.00000801, -0.00001388, 0.00001253, -0.00000907, -0.00000264, -0.00000497, 0.00000669, 0.00000347, 0.00000393, -0.00000653, 0.00000021, 0.00000592, 0.00000027, -0.00000247, 0.00000162, -0.00000039, 0.00000314, -0.00000638, 0.00000823, -0.00000625, 0.00000456, -0.00000019, 0.00000376, -0.00001095, 0.00000461, -0.00000036, 0.00000273, -0.00001388, 0.00001082, 0.00000178, -0.00000423, },
		{ -0.00000996, -0.00000551, 0.00000856, -0.00000739, -0.00000994, 0.00001046, 0.00001988, -0.00001791, 0.00000958, -0.00000813, 0.00001269, -0.00001247, -0.00000332, -0.00000111, -0.00000809, 0.00000762, -0.00001647, 0.00000093, -0.00000171, -0.00000550, 0.00001936, -0.00001370, 0.00000937, -0.00000826, 0.00000353, -0.00000525, 0.00001153, -0.00000207, -0.00000306, -0.00000313, 0.00000212, -0.00000542, 0.00001148, -0.00000224, 0.00000338, -0.00000697, 0.00000144, 0.00000165, 0.00000383, -0.00000166, -0.00000137, -0.00000798, -0.00000361, -0.00000471, 0.00000077, -0.00000844, 0.00000591, -0.00002644, -0.00000926, -0.00005047, -0.00011463, 0.00015982, 0.00023782, 0.00006366, -0.00015492, 0.00000650, -0.00003038, 0.00000257, -0.00001198, -0.00000245, -0.00001621, 0.00000322, -0.00000426, 0.00000407, -0.00000345, 0.00000164, -0.00000615, 0.00000444, -0.00000942, 0.00000112, -0.00000103, -0.00000195, -0.00000658, -0.00000111, 0.00000266, -0.00000049, -0.00000194, -0.00000304, -0.00000021, -0.00000531, 0.00000751, -0.00000535, 0.00000455, -0.00001092, -0.00000785, -0.00001166, -0.00001218, 0.00000175, 0.00001745, 0.00003435, -0.00000861, 0.00004696, 0.00000535, 0.00000852, -0.00004013, 0.00001179, -0.00001229, 0.00000215, -0.00000911, 0.00000473, 0.00000015, -0.00000739, -0.00000321, 0.00000214, 0.00000611, 0.00000344, -0.00000344, -0.00000315, 0.00000556, 0.00000273, -0.00000401, 0.00000253, -0.00000198, 0.00000555, -0.00000789, 0.00000743, -0.00000388, 0.00000192, 0.00000072, 0.00000471, -0.00000915, 0.00000034, 0.00000187, 0.00000250, -0.00001042, 0.00000206, 0.00000935, -0.00000539, },
		{ -0.00000648, -0.00001193, 0.00000997, -0.00000442, -0.00000831, -0.00000247, 0.00002848, -0.00001351, 0.00000230, -0.00000193, 0.00000391, 0.00000094, -0.00001436, 0.00000650, -0.00001312, 0.00000845, -0.00001148, -0.00000676, 0.00000476, -0.00001247, 0.00002024, -0.00000766, 0.00000234, -0.00000117, -0.00000263, -0.00000042, 0.00000379, 0.00000728, -0.00000853, 0.00000056, -0.00000199, -0.00000115, 0.00000365, 0.00000671, -0.00000394, 0.00000218, -0.00000829, 0.00000870, -0.00000358, 0.00000705, -0.00000843, 0.00000105, -0.00001252, 0.00000279, -0.00000808, 0.00000220, -0.00000786, -0.00000143, -0.00003682, -0.00000113, -0.00014620, 0.00007320, 0.00025772, 0.00014716, -0.00014483, -0.00003684, -0.00000782, -0.00001660, 0.00000066, -0.00001043, -0.00000906, -0.00000673, 0.00000255, -0.00000217, 0.00000220, -0.00000220, -0.00000248, 0.00000001, -0.00000337, -0.00000541, 0.00000362, -0.00000448, -0.00000308, -0.00000569, 0.00000468, -0.00000137, 0.00000004, -0.00000467, 0.00000089, -0.00000596, 0.00000525, -0.00000107, -0.00000027, -0.00000191, -0.00001467, -0.00000406, -0.00002002, 0.00000722, 0.00000039, 0.00004537, -0.00001202, 0.00003784, 0.00001551, 0.00001168, -0.00002884, -0.00000794, 0.00000342, -0.00000947, 0.00000060, -0.00000658, 0.00000958, -0.00001104, -0.00000099, -0.00000294, 0.00000836, 0.00000259, 0.00000001, -0.00000561, 0.00000441, 0.00000443, -0.00000370, 0.00000140, -0.00000145, 0.00000507, -0.00000554, 0.00000277, 0.00000137, -0.00000247, 0.00000305, 0.00000304, -0.00000394, -0.00000611, 0.00000543, -0.00000007, -0.00000411, -0.00000718, 0.00001429, -0.00000461, },
		{ -0.00000534, -0.00001388, 0.00000598, 0.00000097, -0.00000710, -0.00000952, 0.00002415, 0.00000238, -0.00001101, 0.00000778, -0.00000659, 0.00001174, -0.00001889, 0.00000645, -0.00001040, 0.00000240, -0.00000242, -0.00001460, 0.00000855, -0.00001399, 0.00001397, 0.00000335, -0.00000656, 0.00000613, -0.00000824, 0.00000420, -0.00000341, 0.00001394, -0.00001043, 0.00000183, -0.00000463, 0.00000229, -0.00000290, 0.00001280, -0.00000793, 0.00000746, -0.00001269, 0.00001036, -0.00000614, 0.00001022, -0.00000978, 0.00000416, -0.00001475, 0.00000418, -0.00001154, 0.00000733, -0.00001545, 0.00001223, -0.00004502, 0.00001140, -0.00011759, -0.00003650, 0.00024772, 0.00022416, -0.00009466, -0.00009519, 0.00001346, -0.00003271, 0.00000948, -0.00001636, -0.00000199, -0.00001606, 0.00000788, -0.00000727, 0.00000653, -0.00000515, 0.00000119, -0.00000469, 0.00000233, -0.00001045, 0.00000618, -0.00000553, -0.00000059, -0.00000843, 0.00000457, -0.00000098, 0.00000092, -0.00000475, 0.00000023, -0.00000453, 0.00000116, 0.00000383, -0.00000477, 0.00000458, -0.00001703, -0.00000133, -0.00002221, 0.00000673, -0.00000561, 0.00003940, 0.00000252, 0.00001536, 0.00003248, 0.00000506, -0.00000668, -0.00003003, 0.00001524, -0.00001610, 0.00000602, -0.00001282, 0.00001220, -0.00000894, -0.00000188, -0.00000501, 0.00000723, 0.00000379, 0.00000159, -0.00000558, 0.00000182, 0.00000577, -0.00000230, -0.00000058, -0.00000006, 0.00000310, -0.00000131, -0.00000297, 0.00000631, -0.00000562, 0.00000440, 0.00000110, 0.00000149, -0.00001099, 0.00000675, -0.00000128, 0.00000043, -0.00001195, 0.00001303, 0.00000017, },
		{ -0.00000583, -0.00001187, -0.00000168, 0.00000699, -0.00000673, -0.00000987, 0.00001069, 0.00002059, -0.00001974, 0.00001108, -0.00000931, 0.00001195, -0.00001217, -0.00000260, -0.00000120, -0.00000725, 0.00000651, -0.00001901, 0.00000752, -0.00000995, 0.00000332, 0.00001441, -0.00001220, 0.00000920, -0.00000952, 0.00000492, -0.00000600, 0.00001421, -0.00000620, -0.00000130, -0.00000356, 0.00000202, -0.00000462, 0.00001246, -0.00000526, 0.00000566, -0.00000913, 0.00000467, -0.00000187, 0.00000584, -0.00000367, -0.00000022, -0.00000904, -0.00000177, -0.00000746, 0.00000337, -0.00001142, 0.00000818, -0.00002942, -0.00000850, -0.00005739, -0.00012429, 0.00018695, 0.00029115, -0.00001366, -0.00014592, 0.00001565, -0.00003453, 0.00000700, -0.00001478, -0.00000035, -0.00001918, 0.00000677, -0.00000694, 0.00000581, -0.00000404, 0.00000173, -0.00000621, 0.00000417, -0.00001091, 0.00000433, -0.00000332, -0.00000090, -0.00000780, 0.00000151, 0.00000119, 0.00000019, -0.00000305, -0.00000186, -0.00000204, -0.00000294, 0.00000703, -0.00000631, 0.00000610, -0.00001348, -0.00000418, -0.00001802, -0.00000094, -0.00000057, 0.00002093, 0.00002505, -0.00000747, 0.00004426, 0.00000067, 0.00001088, -0.00003871, 0.00001173, -0.00001053, 0.00000197, -0.00000986, 0.00000584, -0.00000080, -0.00000609, -0.00000339, 0.00000247, 0.00000699, 0.00000124, -0.00000336, -0.00000162, 0.00000659, -0.00000046, -0.00000226, 0.00000085, 0.00000120, 0.00000254, -0.00000691, 0.00000811, -0.00000536, 0.00000315, 0.00000076, 0.00000450, -0.00001157, 0.00000373, 0.00000085, 0.00000097, -0.00001069, 0.00000593, 0.00000753, },
		{ -0.00000506, -0.00000881, -0.00000901, 0.00001014, -0.00000517, -0.00000722, -0.00000317, 0.00002998, -0.00001534, 0.00000292, -0.00000170, 0.00000209, 0.00000168, -0.00001431, 0.00000744, -0.00001366, 0.00000978, -0.00001670, 0.00000084, -0.00000189, -0.00000797, 0.00002117, -0.00001160, 0.00000660, -0.00000590, 0.00000141, -0.00000374, 0.00000874, 0.00000228, -0.00000671, -0.00000037, -0.00000095, -0.00000240, 0.00000715, 0.00000192, -0.00000057, -0.00000113, -0.00000423, 0.00000509, -0.00000165, 0.00000492, -0.00000682, -0.00000106, -0.00000938, -0.00000106, -0.00000456, -0.00000158, -0.00000480, -0.00000510, -0.00003461, -0.00000794, -0.00015306, 0.00007600, 0.00033277, 0.00008978, -0.00017051, -0.00000924, -0.00002112, -0.00000585, -0.00000667, -0.00000438, -0.00001537, -0.00000095, -0.00000140, 0.00000045, 0.00000063, -0.00000075, -0.00000421, 0.00000183, -0.00000694, -0.00000116, 0.00000125, -0.00000332, -0.00000463, -0.00000328, 0.00000397, -0.00000126, -0.00000059, -0.00000413, 0.00000009, -0.00000554, 0.00000734, -0.00000436, 0.00000324, -0.00000617, -0.00000961, -0.00001087, -0.00001150, 0.00000837, 0.00000170, 0.00004079, -0.00001656, 0.00004027, 0.00000781, 0.00001393, -0.00002790, -0.00000693, 0.00000370, -0.00000759, -0.00000173, -0.00000453, 0.00000812, -0.00000966, -0.00000106, -0.00000280, 0.00000934, 0.00000121, -0.00000110, -0.00000403, 0.00000594, 0.00000187, -0.00000334, 0.00000098, -0.00000003, 0.00000490, -0.00000780, 0.00000612, -0.00000185, -0.00000018, 0.00000193, 0.00000465, -0.00000793, -0.00000257, 0.00000495, -0.00000105, -0.00000593, -0.00000294, 0.00001330, },
		{ -0.00000086, -0.00000718, -0.00001276, 0.00000808, -0.00000083, -0.00000553, -0.00001047, 0.00002549, 0.00000155, -0.00001186, 0.00000932, -0.00000887, 0.00001233, -0.00001904, 0.00000780, -0.00001141, 0.00000458, -0.00000759, -0.00000956, 0.00000690, -0.00001603, 0.00002120, -0.00000487, 0.00000018, 0.00000009, -0.00000367, 0.00000043, 0.00000150, 0.00001024, -0.00001004, 0.00000133, -0.00000326, 0.00000014, 0.00000134, 0.00000840, -0.00000550, 0.00000499, -0.00000982, 0.00000845, -0.00000570, 0.00000907, -0.00000870, 0.00000215, -0.00001175, 0.00000089, -0.00000880, 0.00000443, -0.00001321, 0.00000921, -0.00004253, 0.00000481, -0.00011981, -0.00005193, 0.00032173, 0.00020900, -0.00015895, -0.00005605, -0.00000236, -0.00002055, 0.00000150, -0.00000958, -0.00000863, -0.00001078, 0.00000482, -0.00000528, 0.00000504, -0.00000323, -0.00000125, -0.00000203, -0.00000165, -0.00000695, 0.00000513, -0.00000529, -0.00000150, -0.00000713, 0.00000518, -0.00000169, 0.00000095, -0.00000503, 0.00000058, -0.00000583, 0.00000486, -0.00000010, -0.00000142, 0.00000109, -0.00001349, -0.00000526, -0.00001920, 0.00001301, -0.00000796, 0.00004039, -0.00000689, 0.00002063, 0.00002460, 0.00000558, -0.00000587, -0.00002811, 0.00001426, -0.00001279, 0.00000313, -0.00001083, 0.00001101, -0.00000798, -0.00000167, -0.00000491, 0.00000784, 0.00000367, -0.00000071, -0.00000397, 0.00000323, 0.00000474, -0.00000393, 0.00000068, -0.00000081, 0.00000572, -0.00000600, 0.00000167, 0.00000283, -0.00000348, 0.00000309, 0.00000335, -0.00000243, -0.00000887, 0.00000770, -0.00000229, -0.00000182, -0.00000878, 0.00001378, },
		{ 0.00000593, -0.00000688, -0.00001263, 0.00000172, 0.00000506, -0.00000550, -0.00001032, 0.00001120, 0.00002106, -0.00002168, 0.00001291, -0.00001088, 0.00001149, -0.00001179, -0.00000177, -0.00000149, -0.00000593, 0.00000371, -0.00001864, 0.00001177, -0.00001757, 0.00001402, 0.00000565, -0.00000670, 0.00000517, -0.00000724, 0.00000323, -0.00000356, 0.00001372, -0.00000835, -0.00000059, -0.00000286, 0.00000044, -0.00000164, 0.00001050, -0.00000548, 0.00000552, -0.00000883, 0.00000549, -0.00000355, 0.00000621, -0.00000380, -0.00000115, -0.00000750, -0.00000306, -0.00000693, 0.00000274, -0.00001133, 0.00000705, -0.00002812, -0.00001344, -0.00005760, -0.00014525, 0.00023730, 0.00032855, -0.00010503, -0.00011213, 0.00000857, -0.00002827, 0.00000375, -0.00001158, -0.00000360, -0.00001747, 0.00000685, -0.00000716, 0.00000565, -0.00000292, 0.00000005, -0.00000442, 0.00000166, -0.00000979, 0.00000560, -0.00000456, -0.00000066, -0.00000797, 0.00000339, -0.00000009, 0.00000065, -0.00000389, -0.00000083, -0.00000410, 0.00000073, 0.00000447, -0.00000514, 0.00000531, -0.00001332, -0.00000363, -0.00002097, 0.00000962, -0.00000537, 0.00002546, 0.00001417, -0.00000324, 0.00003929, -0.00000175, 0.00001179, -0.00003635, 0.00001030, -0.00000703, -0.00000006, -0.00000901, 0.00000560, -0.00000059, -0.00000552, -0.00000321, 0.00000252, 0.00000803, -0.00000170, -0.00000212, -0.00000045, 0.00000717, -0.00000363, -0.00000005, -0.00000117, 0.00000520, -0.00000249, -0.00000329, 0.00000635, -0.00000484, 0.00000289, 0.00000219, 0.00000232, -0.00001207, 0.00000655, -0.00000055, -0.00000082, -0.00000949, 0.00000860, },
		{ 0.00001179, -0.00000560, -0.00001080, -0.00000548, 0.00000912, -0.00000490, -0.00000682, -0.00000356, 0.00003129, -0.00001731, 0.00000377, -0.00000160, 0.00000017, 0.00000283, -0.00001438, 0.00000836, -0.00001388, 0.00000993, -0.00002060, 0.00000884, -0.00001120, 0.00000132, 0.00001659, -0.00001102, 0.00000725, -0.00000783, 0.00000331, -0.00000500, 0.00001206, -0.00000238, -0.00000477, -0.00000085, -0.00000092, -0.00000196, 0.00000861, -0.00000141, 0.00000188, -0.00000352, -0.00000089, 0.00000168, 0.00000001, 0.00000346, -0.00000614, -0.00000178, -0.00000786, -0.00000343, -0.00000225, -0.00000431, -0.00000319, -0.00000652, -0.00003561, -0.00001098, -0.00016678, 0.00008977, 0.00041453, -0.00000193, -0.00016432, 0.00000336, -0.00002664, -0.00000035, -0.00001009, -0.00000173, -0.00001901, 0.00000332, -0.00000449, 0.00000233, -0.00000002, -0.00000049, -0.00000458, 0.00000200, -0.00000891, 0.00000246, -0.00000125, -0.00000210, -0.00000596, -0.00000071, 0.00000272, -0.00000081, -0.00000153, -0.00000308, -0.00000169, -0.00000328, 0.00000741, -0.00000611, 0.00000545, -0.00000924, -0.00000543, -0.00001757, 0.00000009, 0.00000439, 0.00000598, 0.00003271, -0.00001687, 0.00004002, 0.00000244, 0.00001479, -0.00002578, -0.00000772, 0.00000598, -0.00000754, -0.00000270, -0.00000352, 0.00000746, -0.00000872, -0.00000086, -0.00000308, 0.00001086, -0.00000134, -0.00000095, -0.00000278, 0.00000741, -0.00000150, -0.00000167, -0.00000069, 0.00000327, 0.00000187, -0.00000723, 0.00000744, -0.00000373, 0.00000126, 0.00000168, 0.00000499, -0.00001116, 0.00000160, 0.00000350, -0.00000217, -0.00000668, 0.00000099, },
		{ 0.00001329, -0.00000147, -0.00000946, -0.00001025, 0.00000866, -0.00000169, -0.00000441, -0.00001133, 0.00002681, 0.00000052, -0.00001264, 0.00001104, -0.00001162, 0.00001358, -0.00001946, 0.00000933, -0.00001245, 0.00000645, -0.00001321, -0.00000183, 0.00000075, -0.00001270, 0.00002390, -0.00001067, 0.00000581, -0.00000571, 0.00000123, -0.00000374, 0.00000734, 0.00000459, -0.00000789, 0.00000003, -0.00000166, -0.00000187, 0.00000559, 0.00000310, -0.00000171, 0.00000119, -0.00000541, 0.00000466, -0.00000367, 0.00000679, -0.00000665, -0.00000062, -0.00000781, -0.00000383, -0.00000419, -0.00000054, -0.00000918, 0.00000475, -0.00003865, -0.00000380, -0.00012048, -0.00007161, 0.00042156, 0.00015109, -0.00019842, -0.00002016, -0.00001965, -0.00000672, -0.00000804, -0.00000157, -0.00001676, -0.00000346, -0.00000016, -0.00000190, 0.00000273, -0.00000093, -0.00000388, 0.00000068, -0.00000599, -0.00000213, 0.00000244, -0.00000386, -0.00000323, -0.00000471, 0.00000457, -0.00000161, 0.00000027, -0.00000443, -0.00000029, -0.00000551, 0.00000744, -0.00000389, 0.00000229, -0.00000319, -0.00000841, -0.00001195, -0.00001072, 0.00001339, -0.00000733, 0.00003805, -0.00001307, 0.00002467, 0.00001765, 0.00000587, -0.00000457, -0.00002738, 0.00001444, -0.00001023, 0.00000049, -0.00000885, 0.00000972, -0.00000676, -0.00000158, -0.00000503, 0.00000893, 0.00000253, -0.00000222, -0.00000225, 0.00000467, 0.00000260, -0.00000415, 0.00000067, 0.00000027, 0.00000608, -0.00000915, 0.00000589, -0.00000089, -0.00000084, 0.00000148, 0.00000568, -0.00000717, -0.00000484, 0.00000714, -0.00000318, -0.00000371, -0.00000489, },
		{ 0.00000984, 0.00000484, -0.00000867, -0.00001183, 0.00000396, 0.00000346, -0.00000414, -0.00001107, 0.00001192, 0.00002129, -0.00002370, 0.00001514, -0.00001305, 0.00001166, -0.00001158, -0.00000073, -0.00000199, -0.00000430, -0.00000064, -0.00001482, 0.00001195, -0.00002184, 0.00002357, -0.00000445, 0.00000118, -0.00000169, -0.00000200, -0.00000123, 0.00000202, 0.00000961, -0.00000772, -0.00000153, -0.00000081, -0.00000253, 0.00000329, 0.00000564, -0.00000261, 0.00000248, -0.00000551, 0.00000316, -0.00000257, 0.00000427, -0.00000153, -0.00000457, -0.00000296, -0.00000802, -0.00000226, -0.00000219, -0.00000704, 0.00000154, -0.00002165, -0.00002519, -0.00004949, -0.00018068, 0.00032030, 0.00033206, -0.00019271, -0.00006282, -0.00001179, -0.00001249, -0.00000690, -0.00000241, -0.00001234, -0.00001055, 0.00000268, -0.00000419, 0.00000310, 0.00000014, -0.00000347, -0.00000081, -0.00000306, -0.00000580, 0.00000426, -0.00000402, -0.00000180, -0.00000655, 0.00000388, -0.00000055, 0.00000036, -0.00000387, -0.00000075, -0.00000533, 0.00000459, 0.00000061, -0.00000224, 0.00000238, -0.00001030, -0.00000693, -0.00001850, 0.00001604, -0.00000943, 0.00002887, 0.00000374, 0.00000244, 0.00003307, -0.00000273, 0.00001200, -0.00003397, 0.00000853, -0.00000271, -0.00000328, -0.00000699, 0.00000431, 0.00000056, -0.00000554, -0.00000289, 0.00000265, 0.00000869, -0.00000475, -0.00000007, 0.00000047, 0.00000693, -0.00000606, 0.00000176, -0.00000244, 0.00000857, -0.00000810, 0.00000200, 0.00000265, -0.00000252, 0.00000122, 0.00000490, -0.00000190, -0.00001003, 0.00000790, -0.00000183, -0.00000278, -0.00000691, },
		{ 0.00000396, 0.00001030, -0.00000664, -0.00001176, -0.00000216, 0.00000752, -0.00000389, -0.00000706, -0.00000359, 0.00003236, -0.00001948, 0.00000504, -0.00000195, -0.00000147, 0.00000417, -0.00001447, 0.00000917, -0.00001367, 0.00000849, -0.00002191, 0.00001510, -0.00002073, 0.00001352, 0.00000714, -0.00000556, 0.00000339, -0.00000570, 0.00000167, -0.00000264, 0.00001185, -0.00000479, -0.00000412, 0.00000012, -0.00000277, 0.00000107, 0.00000677, -0.00000166, 0.00000162, -0.00000318, -0.00000019, 0.00000034, -0.00000025, 0.00000426, -0.00000805, 0.00000072, -0.00001013, -0.00000180, -0.00000393, -0.00000314, -0.00000595, -0.00000241, -0.00004393, -0.00000490, -0.00019521, 0.00012800, 0.00048584, -0.00011755, -0.00012560, -0.00000439, -0.00001927, -0.00000432, -0.00000606, -0.00000564, -0.00001696, 0.00000330, -0.00000446, 0.00000171, 0.00000163, -0.00000261, -0.00000251, -0.00000061, -0.00000786, 0.00000385, -0.00000248, -0.00000188, -0.00000612, 0.00000106, 0.00000175, -0.00000070, -0.00000207, -0.00000232, -0.00000364, 0.00000031, 0.00000533, -0.00000576, 0.00000545, -0.00000978, -0.00000411, -0.00002121, 0.00001119, -0.00000231, 0.00001217, 0.00002200, -0.00001329, 0.00003692, -0.00000042, 0.00001418, -0.00002259, -0.00001013, 0.00001030, -0.00000960, -0.00000187, -0.00000397, 0.00000800, -0.00000855, -0.00000025, -0.00000372, 0.00001273, -0.00000488, 0.00000060, -0.00000205, 0.00000865, -0.00000517, 0.00000079, -0.00000300, 0.00000772, -0.00000360, -0.00000360, 0.00000601, -0.00000344, 0.00000103, 0.00000307, 0.00000312, -0.00001231, 0.00000517, 0.00000163, -0.00000352, -0.00000607, },
		{ -0.00000097, 0.00001205, -0.00000202, -0.00001157, -0.00000689, 0.00000790, -0.00000146, -0.00000399, -0.00001189, 0.00002803, -0.00000076, -0.00001315, 0.00001268, -0.00001460, 0.00001546, -0.00002022, 0.00001103, -0.00001333, 0.00000744, -0.00001804, 0.00000704, -0.00000889, -0.00000379, 0.00002048, -0.00001168, 0.00000784, -0.00000880, 0.00000412, -0.00000599, 0.00001176, -0.00000098, -0.00000537, -0.00000095, -0.00000100, -0.00000233, 0.00000804, -0.00000106, 0.00000143, -0.00000199, -0.00000139, 0.00000075, -0.00000166, 0.00000504, -0.00000543, -0.00000211, -0.00000537, -0.00000726, -0.00000053, -0.00000475, -0.00000630, 0.00000228, -0.00003772, -0.00000892, -0.00012705, -0.00008520, 0.00053876, 0.00004778, -0.00020051, -0.00000041, -0.00002910, 0.00000192, -0.00001376, 0.00000313, -0.00002237, 0.00000256, -0.00000449, 0.00000086, 0.00000128, 0.00000019, -0.00000517, 0.00000182, -0.00000892, 0.00000230, -0.00000062, -0.00000217, -0.00000496, -0.00000192, 0.00000331, -0.00000115, -0.00000073, -0.00000330, -0.00000222, -0.00000325, 0.00000796, -0.00000627, 0.00000492, -0.00000657, -0.00000386, -0.00001895, 0.00000113, 0.00000820, -0.00000280, 0.00003142, -0.00001472, 0.00002596, 0.00001283, 0.00000500, -0.00000206, -0.00002858, 0.00001676, -0.00000956, -0.00000079, -0.00000787, 0.00000918, -0.00000597, -0.00000120, -0.00000567, 0.00001077, 0.00000001, -0.00000230, -0.00000099, 0.00000626, -0.00000063, -0.00000283, -0.00000078, 0.00000344, 0.00000335, -0.00000931, 0.00000807, -0.00000336, 0.00000099, 0.00000074, 0.00000677, -0.00001131, 0.00000009, 0.00000533, -0.00000397, -0.00000483, },
		{ -0.00000335, 0.00000973, 0.00000422, -0.00001100, -0.00000953, 0.00000469, 0.00000272, -0.00000314, -0.00001185, 0.00001275, 0.00002123, -0.00002568, 0.00001767, -0.00001586, 0.00001267, -0.00001178, 0.00000059, -0.00000263, -0.00000274, -0.00000584, -0.00000786, 0.00000731, -0.00002064, 0.00002883, -0.00001272, 0.00000871, -0.00000905, 0.00000428, -0.00000678, 0.00000926, 0.00000283, -0.00000469, -0.00000381, 0.00000202, -0.00000600, 0.00000898, -0.00000089, 0.00000237, -0.00000264, 0.00000015, -0.00000182, 0.00000070, 0.00000055, 0.00000242, -0.00000959, 0.00000347, -0.00001550, 0.00000551, -0.00001057, 0.00000056, -0.00000712, -0.00001170, -0.00004168, -0.00003554, -0.00022841, 0.00043986, 0.00028617, -0.00025640, -0.00001272, -0.00003703, 0.00000750, -0.00002115, 0.00000971, -0.00002408, -0.00000024, -0.00000460, 0.00000114, -0.00000110, 0.00000433, -0.00000793, 0.00000362, -0.00000894, 0.00000026, 0.00000063, -0.00000175, -0.00000427, -0.00000357, 0.00000286, 0.00000002, -0.00000088, -0.00000273, -0.00000213, -0.00000475, 0.00000728, -0.00000322, 0.00000119, -0.00000153, -0.00000553, -0.00001323, -0.00001043, 0.00001630, -0.00000999, 0.00002884, -0.00000387, 0.00000743, 0.00002718, -0.00000352, 0.00001262, -0.00003272, 0.00000781, 0.00000106, -0.00000650, -0.00000476, 0.00000274, 0.00000207, -0.00000575, -0.00000281, 0.00000335, 0.00000827, -0.00000704, 0.00000213, 0.00000147, 0.00000552, -0.00000710, 0.00000239, -0.00000197, 0.00000974, -0.00001237, 0.00000720, -0.00000165, 0.00000056, -0.00000098, 0.00000797, -0.00000734, -0.00000565, 0.00000739, -0.00000273, -0.00000468, },
		{ -0.00000404, 0.00000560, 0.00000921, -0.00000840, -0.00001123, 0.00000017, 0.00000603, -0.00000258, -0.00000776, -0.00000328, 0.00003316, -0.00002181, 0.00000682, -0.00000298, -0.00000246, 0.00000539, -0.00001445, 0.00000984, -0.00001312, 0.00000546, -0.00001983, 0.00001762, -0.00002755, 0.00002549, -0.00000483, 0.00000342, -0.00000427, 0.00000023, -0.00000353, 0.00000366, 0.00000729, -0.00000380, -0.00000569, 0.00000306, -0.00000674, 0.00000687, 0.00000126, 0.00000191, -0.00000227, 0.00000095, -0.00000336, 0.00000235, -0.00000353, 0.00000831, -0.00001346, 0.00000738, -0.00001737, 0.00000538, -0.00001147, 0.00000397, -0.00001507, 0.00000931, -0.00006219, 0.00001397, -0.00024461, 0.00020424, 0.00052511, -0.00023808, -0.00006275, -0.00003106, 0.00000146, -0.00001867, 0.00000637, -0.00001711, -0.00000793, -0.00000245, -0.00000000, -0.00000241, 0.00000628, -0.00000759, 0.00000240, -0.00000639, -0.00000309, 0.00000194, -0.00000136, -0.00000356, -0.00000427, 0.00000113, 0.00000187, -0.00000163, -0.00000144, -0.00000277, -0.00000457, 0.00000404, 0.00000191, -0.00000366, 0.00000334, -0.00000751, -0.00000656, -0.00001964, 0.00001828, -0.00000834, 0.00001785, 0.00001075, -0.00000742, 0.00003186, -0.00000146, 0.00001276, -0.00001915, -0.00001319, 0.00001582, -0.00001324, 0.00000044, -0.00000572, 0.00000967, -0.00000910, 0.00000060, -0.00000435, 0.00001436, -0.00000878, 0.00000318, -0.00000171, 0.00000927, -0.00000840, 0.00000317, -0.00000487, 0.00001187, -0.00001010, 0.00000221, 0.00000221, -0.00000104, -0.00000090, 0.00000609, -0.00000118, -0.00001058, 0.00000708, -0.00000001, -0.00000516, },
		{ -0.00000477, 0.00000227, 0.00001063, -0.00000296, -0.00001282, -0.00000351, 0.00000633, -0.00000034, -0.00000432, -0.00001204, 0.00002905, -0.00000233, -0.00001320, 0.00001392, -0.00001745, 0.00001774, -0.00002122, 0.00001281, -0.00001394, 0.00000708, -0.00002073, 0.00001491, -0.00001971, 0.00000911, 0.00001112, -0.00000698, 0.00000480, -0.00000744, 0.00000312, -0.00000430, 0.00001242, -0.00000418, -0.00000428, -0.00000026, -0.00000251, 0.00000012, 0.00000696, -0.00000196, 0.00000170, -0.00000226, -0.00000013, -0.00000097, -0.00000175, 0.00000586, -0.00000724, 0.00000012, -0.00000724, -0.00000616, -0.00000144, -0.00000444, -0.00000850, 0.00000626, -0.00004526, -0.00000340, -0.00014964, -0.00007707, 0.00065582, -0.00009285, -0.00016085, -0.00000543, -0.00002317, -0.00000070, -0.00001064, 0.00000014, -0.00002138, 0.00000359, -0.00000513, 0.00000060, 0.00000268, -0.00000161, -0.00000349, -0.00000027, -0.00000850, 0.00000431, -0.00000227, -0.00000160, -0.00000544, 0.00000001, 0.00000239, -0.00000114, -0.00000120, -0.00000254, -0.00000428, 0.00000038, 0.00000632, -0.00000663, 0.00000555, -0.00000768, -0.00000185, -0.00002332, 0.00001288, -0.00000017, 0.00000462, 0.00002111, -0.00001181, 0.00002398, 0.00001061, 0.00000265, 0.00000177, -0.00003174, 0.00002145, -0.00001130, -0.00000007, -0.00000850, 0.00000995, -0.00000604, -0.00000030, -0.00000683, 0.00001320, -0.00000384, -0.00000073, -0.00000046, 0.00000790, -0.00000453, -0.00000037, -0.00000317, 0.00000811, -0.00000226, -0.00000600, 0.00000725, -0.00000352, 0.00000102, 0.00000183, 0.00000548, -0.00001332, 0.00000451, 0.00000298, -0.00000488, },
		{ -0.00000610, 0.00000055, 0.00000869, 0.00000375, -0.00001331, -0.00000621, 0.00000395, 0.00000308, -0.00000282, -0.00001240, 0.00001360, 0.00002084, -0.00002742, 0.00002024, -0.00001909, 0.00001453, -0.00001249, 0.00000221, -0.00000327, -0.00000177, -0.00001073, 0.00000090, -0.00000152, -0.00001328, 0.00002762, -0.00001624, 0.00001295, -0.00001410, 0.00000895, -0.00001084, 0.00001559, -0.00000443, -0.00000079, -0.00000619, 0.00000430, -0.00000836, 0.00001343, -0.00000690, 0.00000732, -0.00000777, 0.00000611, -0.00000760, 0.00000458, -0.00000322, 0.00000609, -0.00001394, 0.00000915, -0.00002257, 0.00001331, -0.00001931, 0.00000828, -0.00001521, -0.00000294, -0.00005707, -0.00002334, -0.00027903, 0.00059075, 0.00018305, -0.00027857, 0.00002194, -0.00005585, 0.00002356, -0.00003268, 0.00001949, -0.00003422, 0.00000965, -0.00001201, 0.00000650, -0.00000495, 0.00000779, -0.00001148, 0.00000697, -0.00001395, 0.00000652, -0.00000390, 0.00000126, -0.00000718, 0.00000016, 0.00000087, 0.00000119, -0.00000262, -0.00000086, -0.00000478, -0.00000202, 0.00000781, -0.00000571, 0.00000372, -0.00000476, -0.00000100, -0.00002022, 0.00000133, 0.00001054, -0.00000588, 0.00002398, -0.00000690, 0.00000984, 0.00002309, -0.00000527, 0.00001454, -0.00003353, 0.00000928, 0.00000294, -0.00000845, -0.00000341, 0.00000183, 0.00000322, -0.00000566, -0.00000332, 0.00000496, 0.00000629, -0.00000780, 0.00000380, 0.00000277, 0.00000282, -0.00000645, 0.00000147, 0.00000074, 0.00000767, -0.00001359, 0.00001049, -0.00000488, 0.00000296, -0.00000239, 0.00000996, -0.00001244, -0.00000003, 0.00000533, -0.00000326, },
		{ -0.00000698, -0.00000084, 0.00000585, 0.00000858, -0.00001075, -0.00000915, 0.00000103, 0.00000528, -0.00000149, -0.00000863, -0.00000271, 0.00003362, -0.00002415, 0.00000899, -0.00000477, -0.00000253, 0.00000622, -0.00001420, 0.00001040, -0.00001252, 0.00000137, -0.00001441, 0.00001528, -0.00002915, 0.00003378, -0.00001591, 0.00001313, -0.00001347, 0.00000818, -0.00001079, 0.00001262, -0.00000093, 0.00000048, -0.00000941, 0.00000762, -0.00001216, 0.00001447, -0.00000702, 0.00000865, -0.00000934, 0.00000855, -0.00001031, 0.00000775, -0.00000972, 0.00001526, -0.00002186, 0.00001752, -0.00002891, 0.00001763, -0.00002466, 0.00001684, -0.00003016, 0.00002796, -0.00008963, 0.00004509, -0.00031614, 0.00032746, 0.00051141, -0.00033962, 0.00000831, -0.00006834, 0.00003067, -0.00004004, 0.00002459, -0.00003408, 0.00000669, -0.00001319, 0.00000842, -0.00000957, 0.00001333, -0.00001473, 0.00000935, -0.00001453, 0.00000486, -0.00000322, 0.00000234, -0.00000732, -0.00000022, -0.00000085, 0.00000347, -0.00000394, 0.00000074, -0.00000509, -0.00000336, 0.00000643, -0.00000147, -0.00000100, 0.00000021, -0.00000343, -0.00001208, -0.00001247, 0.00001913, -0.00001067, 0.00002036, 0.00000163, -0.00000151, 0.00002646, -0.00000194, 0.00001160, -0.00001664, -0.00001552, 0.00002115, -0.00001726, 0.00000331, -0.00000803, 0.00001192, -0.00000998, 0.00000127, -0.00000445, 0.00001501, -0.00001209, 0.00000607, -0.00000137, 0.00000878, -0.00001041, 0.00000456, -0.00000515, 0.00001403, -0.00001558, 0.00000840, -0.00000264, 0.00000246, -0.00000365, 0.00000983, -0.00000714, -0.00000609, 0.00000682, -0.00000102, },
		{ -0.00000614, -0.00000332, 0.00000432, 0.00000947, -0.00000457, -0.00001266, -0.00000093, 0.00000467, 0.00000125, -0.00000526, -0.00001175, 0.00002980, -0.00000411, -0.00001272, 0.00001449, -0.00001976, 0.00002010, -0.00002234, 0.00001463, -0.00001431, 0.00000523, -0.00002030, 0.00001961, -0.00002872, 0.00002296, -0.00000208, 0.00000246, -0.00000306, -0.00000137, -0.00000229, 0.00000215, 0.00000797, -0.00000333, -0.00000602, 0.00000308, -0.00000692, 0.00000618, 0.00000140, 0.00000171, -0.00000244, 0.00000206, -0.00000351, 0.00000142, -0.00000570, 0.00001090, -0.00001371, 0.00000782, -0.00001550, 0.00000204, -0.00000988, 0.00000364, -0.00001915, 0.00002040, -0.00006594, 0.00001909, -0.00019815, -0.00002952, 0.00074836, -0.00025192, -0.00008553, -0.00003652, 0.00000088, -0.00001731, 0.00000378, -0.00001287, -0.00001133, -0.00000285, 0.00000011, -0.00000447, 0.00000830, -0.00000747, 0.00000219, -0.00000661, -0.00000345, 0.00000228, -0.00000093, -0.00000350, -0.00000344, -0.00000012, 0.00000292, -0.00000256, -0.00000014, -0.00000335, -0.00000508, 0.00000404, 0.00000335, -0.00000533, 0.00000420, -0.00000613, -0.00000338, -0.00002284, 0.00002095, -0.00000826, 0.00001241, 0.00000917, -0.00000571, 0.00001939, 0.00001055, -0.00000075, 0.00000625, -0.00003606, 0.00002786, -0.00001509, 0.00000251, -0.00001072, 0.00001209, -0.00000704, 0.00000099, -0.00000817, 0.00001567, -0.00000838, 0.00000218, -0.00000053, 0.00000917, -0.00000836, 0.00000241, -0.00000549, 0.00001289, -0.00000941, 0.00000005, 0.00000361, -0.00000126, -0.00000094, 0.00000489, 0.00000140, -0.00001219, 0.00000716, 0.00000089, },
		{ -0.00000376, -0.00000658, 0.00000404, 0.00000728, 0.00000299, -0.00001492, -0.00000264, 0.00000221, 0.00000445, -0.00000332, -0.00001253, 0.00001432, 0.00002010, -0.00002873, 0.00002250, -0.00002239, 0.00001705, -0.00001369, 0.00000410, -0.00000378, -0.00000190, -0.00001395, 0.00000941, -0.00001248, -0.00000092, 0.00001958, -0.00001347, 0.00001183, -0.00001455, 0.00000964, -0.00001092, 0.00001821, -0.00000947, 0.00000181, -0.00000694, 0.00000439, -0.00000781, 0.00001445, -0.00000984, 0.00000957, -0.00001019, 0.00000956, -0.00001144, 0.00000654, -0.00000450, 0.00000675, -0.00001459, 0.00001058, -0.00002523, 0.00001675, -0.00002380, 0.00001123, -0.00001718, -0.00000219, -0.00006265, -0.00002474, -0.00031596, 0.00075694, 0.00002649, -0.00024965, 0.00002831, -0.00005779, 0.00002745, -0.00003476, 0.00002114, -0.00003748, 0.00001443, -0.00001567, 0.00000874, -0.00000582, 0.00000816, -0.00001187, 0.00000697, -0.00001564, 0.00001047, -0.00000715, 0.00000327, -0.00000899, 0.00000319, -0.00000086, 0.00000192, -0.00000383, 0.00000066, -0.00000766, 0.00000223, 0.00000607, -0.00000623, 0.00000440, -0.00000594, 0.00000133, -0.00002505, 0.00001347, 0.00000102, 0.00000209, 0.00001454, -0.00000487, 0.00000875, 0.00002161, -0.00000860, 0.00001809, -0.00003666, 0.00001346, 0.00000218, -0.00000826, -0.00000376, 0.00000230, 0.00000343, -0.00000497, -0.00000451, 0.00000742, 0.00000269, -0.00000667, 0.00000458, 0.00000438, -0.00000091, -0.00000433, -0.00000068, 0.00000528, 0.00000233, -0.00001101, 0.00001061, -0.00000576, 0.00000350, -0.00000187, 0.00000951, -0.00001549, 0.00000529, 0.00000251, },
		{ -0.00000143, -0.00000861, 0.00000297, 0.00000503, 0.00000815, -0.00001321, -0.00000589, 0.00000037, 0.00000560, -0.00000101, -0.00000940, -0.00000198, 0.00003367, -0.00002628, 0.00001130, -0.00000712, -0.00000163, 0.00000648, -0.00001365, 0.00001096, -0.00001240, -0.00000274, -0.00000672, 0.00000834, -0.00002437, 0.00003568, -0.00002273, 0.00002028, -0.00002117, 0.00001533, -0.00001740, 0.00002157, -0.00001058, 0.00000651, -0.00001404, 0.00001244, -0.00001736, 0.00002188, -0.00001587, 0.00001646, -0.00001756, 0.00001767, -0.00001927, 0.00001496, -0.00001713, 0.00002318, -0.00003101, 0.00002853, -0.00004189, 0.00003201, -0.00004057, 0.00003250, -0.00004773, 0.00004919, -0.00012097, 0.00008198, -0.00040298, 0.00049749, 0.00043035, -0.00039909, 0.00006811, -0.00010332, 0.00005928, -0.00006148, 0.00004290, -0.00005161, 0.00002285, -0.00002579, 0.00001835, -0.00001767, 0.00002082, -0.00002207, 0.00001632, -0.00002293, 0.00001410, -0.00001026, 0.00000765, -0.00001235, 0.00000533, -0.00000440, 0.00000617, -0.00000723, 0.00000411, -0.00000915, 0.00000043, 0.00000637, -0.00000342, 0.00000079, -0.00000233, 0.00000060, -0.00001859, -0.00000135, 0.00001364, -0.00000779, 0.00001783, -0.00000316, 0.00000224, 0.00002240, -0.00000316, 0.00001176, -0.00001615, -0.00001577, 0.00002481, -0.00002027, 0.00000555, -0.00000990, 0.00001396, -0.00001068, 0.00000137, -0.00000359, 0.00001403, -0.00001390, 0.00000849, -0.00000067, 0.00000689, -0.00001068, 0.00000442, -0.00000317, 0.00001291, -0.00001811, 0.00001293, -0.00000672, 0.00000553, -0.00000582, 0.00001276, -0.00001314, 0.00000007, 0.00000465, },
		{ -0.00000040, -0.00000776, -0.00000058, 0.00000500, 0.00000872, -0.00000679, -0.00001094, 0.00000027, 0.00000356, 0.00000281, -0.00000658, -0.00001102, 0.00003014, -0.00000592, -0.00001184, 0.00001432, -0.00002124, 0.00002225, -0.00002344, 0.00001653, -0.00001472, 0.00000232, -0.00001656, 0.00001972, -0.00003312, 0.00003408, -0.00001568, 0.00001400, -0.00001375, 0.00000793, -0.00001095, 0.00001246, -0.00000130, 0.00000186, -0.00001091, 0.00000910, -0.00001396, 0.00001534, -0.00000825, 0.00000985, -0.00001109, 0.00001123, -0.00001205, 0.00000858, -0.00001398, 0.00002042, -0.00002498, 0.00002105, -0.00003035, 0.00001782, -0.00002675, 0.00002059, -0.00003931, 0.00004575, -0.00010118, 0.00006102, -0.00027815, 0.00007225, 0.00079007, -0.00040290, 0.00000963, -0.00008662, 0.00003952, -0.00004577, 0.00002799, -0.00003480, 0.00000728, -0.00001682, 0.00001150, -0.00001451, 0.00001812, -0.00001718, 0.00001154, -0.00001687, 0.00000622, -0.00000427, 0.00000407, -0.00000848, 0.00000164, -0.00000305, 0.00000558, -0.00000599, 0.00000310, -0.00000659, -0.00000326, 0.00000606, 0.00000058, -0.00000356, 0.00000194, -0.00000284, -0.00000796, -0.00001691, 0.00002294, -0.00001278, 0.00001760, -0.00000158, 0.00000132, 0.00001370, 0.00001149, -0.00000418, 0.00001024, -0.00004017, 0.00003466, -0.00001983, 0.00000614, -0.00001392, 0.00001514, -0.00000864, 0.00000231, -0.00000914, 0.00001738, -0.00001264, 0.00000570, -0.00000080, 0.00000954, -0.00001126, 0.00000452, -0.00000651, 0.00001600, -0.00001600, 0.00000705, -0.00000158, 0.00000250, -0.00000409, 0.00000911, -0.00000481, -0.00000789, 0.00000737, },
		{ -0.00000009, -0.00000459, -0.00000575, 0.00000653, 0.00000598, 0.00000163, -0.00001523, 0.00000032, 0.00000016, 0.00000648, -0.00000458, -0.00001212, 0.00001478, 0.00001913, -0.00002956, 0.00002421, -0.00002536, 0.00001993, -0.00001527, 0.00000627, -0.00000420, -0.00000333, -0.00001442, 0.00001544, -0.00002261, 0.00001359, 0.00000639, -0.00000480, 0.00000498, -0.00000953, 0.00000522, -0.00000558, 0.00001517, -0.00001002, 0.00000117, -0.00000454, 0.00000098, -0.00000302, 0.00001040, -0.00000768, 0.00000688, -0.00000749, 0.00000787, -0.00001069, 0.00000412, -0.00000089, 0.00000192, -0.00000883, 0.00000462, -0.00001975, 0.00001156, -0.00001936, 0.00000445, -0.00000743, -0.00001638, -0.00004924, -0.00005312, -0.00031815, 0.00091290, -0.00016684, -0.00017151, 0.00000104, -0.00003688, 0.00001391, -0.00002285, 0.00001055, -0.00002986, 0.00001025, -0.00001220, 0.00000506, -0.00000140, 0.00000345, -0.00000729, 0.00000184, -0.00001195, 0.00000981, -0.00000696, 0.00000240, -0.00000799, 0.00000389, -0.00000088, 0.00000094, -0.00000321, 0.00000038, -0.00000912, 0.00000648, 0.00000300, -0.00000516, 0.00000326, -0.00000466, 0.00000034, -0.00002545, 0.00002243, -0.00000879, 0.00001145, 0.00000234, 0.00000122, 0.00000443, 0.00002262, -0.00001330, 0.00002285, -0.00004153, 0.00001992, -0.00000112, -0.00000582, -0.00000603, 0.00000439, 0.00000250, -0.00000368, -0.00000611, 0.00001023, -0.00000200, -0.00000387, 0.00000449, 0.00000594, -0.00000499, -0.00000148, -0.00000314, 0.00001038, -0.00000510, -0.00000513, 0.00000750, -0.00000394, 0.00000179, 0.00000094, 0.00000598, -0.00001522, 0.00000882, },
		{ 0.00000148, -0.00000159, -0.00000955, 0.00000665, 0.00000367, 0.00000753, -0.00001508, -0.00000221, -0.00000144, 0.00000699, -0.00000135, -0.00000981, -0.00000125, 0.00003330, -0.00002801, 0.00001339, -0.00000970, 0.00000009, 0.00000616, -0.00001276, 0.00001162, -0.00001325, -0.00000553, 0.00000133, -0.00000148, -0.00001387, 0.00003021, -0.00002308, 0.00002221, -0.00002454, 0.00001885, -0.00002044, 0.00002735, -0.00001862, 0.00001184, -0.00001755, 0.00001554, -0.00002017, 0.00002650, -0.00002241, 0.00002235, -0.00002389, 0.00002518, -0.00002725, 0.00002116, -0.00002293, 0.00002896, -0.00003743, 0.00003644, -0.00005178, 0.00004356, -0.00005401, 0.00004545, -0.00006157, 0.00006544, -0.00014671, 0.00011230, -0.00048947, 0.00070208, 0.00027933, -0.00040040, 0.00009899, -0.00012238, 0.00007685, -0.00007460, 0.00005412, -0.00006330, 0.00003491, -0.00003559, 0.00002596, -0.00002346, 0.00002579, -0.00002679, 0.00002052, -0.00002866, 0.00002173, -0.00001672, 0.00001257, -0.00001688, 0.00001074, -0.00000815, 0.00000881, -0.00001033, 0.00000749, -0.00001383, 0.00000615, 0.00000374, -0.00000327, 0.00000074, -0.00000289, 0.00000262, -0.00002334, 0.00001060, 0.00000383, -0.00000024, 0.00001004, -0.00000261, 0.00000244, 0.00002083, -0.00000600, 0.00001381, -0.00001819, -0.00001316, 0.00002577, -0.00002119, 0.00000615, -0.00001046, 0.00001512, -0.00001078, 0.00000072, -0.00000171, 0.00001123, -0.00001370, 0.00000993, 0.00000051, 0.00000370, -0.00000926, 0.00000285, 0.00000089, 0.00000823, -0.00001660, 0.00001427, -0.00000853, 0.00000676, -0.00000605, 0.00001331, -0.00001728, 0.00000622, },
		{ 0.00000534, -0.00000083, -0.00000947, 0.00000300, 0.00000453, 0.00000819, -0.00000919, -0.00000798, -0.00000003, 0.00000342, 0.00000392, -0.00000799, -0.00000993, 0.00002996, -0.00000749, -0.00001083, 0.00001359, -0.00002183, 0.00002401, -0.00002443, 0.00001864, -0.00001567, -0.00000071, -0.00001035, 0.00001513, -0.00003130, 0.00003934, -0.00002595, 0.00002415, -0.00002415, 0.00001764, -0.00002021, 0.00002410, -0.00001335, 0.00001008, -0.00001791, 0.00001660, -0.00002213, 0.00002579, -0.00002007, 0.00002068, -0.00002256, 0.00002367, -0.00002441, 0.00001936, -0.00002536, 0.00003292, -0.00003923, 0.00003768, -0.00004946, 0.00003888, -0.00004987, 0.00004429, -0.00006652, 0.00007924, -0.00014741, 0.00011852, -0.00038722, 0.00023472, 0.00075924, -0.00051736, 0.00010272, -0.00014211, 0.00008359, -0.00007925, 0.00005646, -0.00006097, 0.00003068, -0.00003550, 0.00002687, -0.00002768, 0.00003035, -0.00002895, 0.00002274, -0.00002917, 0.00001886, -0.00001421, 0.00001205, -0.00001597, 0.00000933, -0.00000850, 0.00001019, -0.00001120, 0.00000833, -0.00001234, 0.00000182, 0.00000515, -0.00000049, -0.00000284, 0.00000038, 0.00000040, -0.00001366, -0.00000693, 0.00001847, -0.00001191, 0.00001792, -0.00000856, 0.00000685, 0.00000873, 0.00001204, -0.00000650, 0.00001255, -0.00004262, 0.00004028, -0.00002407, 0.00000959, -0.00001710, 0.00001834, -0.00001031, 0.00000322, -0.00000924, 0.00001757, -0.00001561, 0.00000895, -0.00000079, 0.00000854, -0.00001254, 0.00000525, -0.00000537, 0.00001596, -0.00001989, 0.00001279, -0.00000646, 0.00000616, -0.00000696, 0.00001288, -0.00001153, -0.00000145, },
		{ 0.00000971, -0.00000177, -0.00000592, -0.00000356, 0.00000773, 0.00000500, -0.00000033, -0.00001402, 0.00000206, -0.00000155, 0.00000870, -0.00000645, -0.00001110, 0.00001475, 0.00001816, -0.00003006, 0.00002532, -0.00002774, 0.00002290, -0.00001713, 0.00000879, -0.00000484, -0.00000573, -0.00001176, 0.00001736, -0.00002898, 0.00002656, -0.00000866, 0.00000743, -0.00000607, -0.00000003, -0.00000367, 0.00000484, 0.00000614, -0.00000517, -0.00000362, 0.00000160, -0.00000624, 0.00000617, 0.00000096, 0.00000022, -0.00000161, 0.00000142, -0.00000032, -0.00000382, -0.00000408, 0.00000882, -0.00000954, 0.00000451, -0.00001017, -0.00000424, -0.00000472, -0.00000312, -0.00001514, 0.00001744, -0.00004988, -0.00001060, -0.00011909, -0.00026502, 0.00102805, -0.00036962, -0.00005791, -0.00005572, 0.00000595, -0.00001714, 0.00000344, -0.00001285, -0.00001035, -0.00000428, -0.00000017, -0.00000575, 0.00000921, -0.00000700, 0.00000277, -0.00000892, -0.00000212, 0.00000332, -0.00000195, -0.00000259, -0.00000300, 0.00000103, 0.00000193, -0.00000275, 0.00000025, -0.00000297, -0.00000745, 0.00000877, 0.00000031, -0.00000379, 0.00000138, -0.00000177, -0.00000355, -0.00002068, 0.00002565, -0.00001548, 0.00001908, -0.00000979, 0.00000925, -0.00000181, 0.00002517, -0.00001854, 0.00002779, -0.00004692, 0.00002750, -0.00000605, -0.00000177, -0.00000976, 0.00000781, 0.00000063, -0.00000211, -0.00000760, 0.00001261, -0.00000681, -0.00000010, 0.00000393, 0.00000687, -0.00000851, 0.00000111, -0.00000468, 0.00001424, -0.00001253, 0.00000235, 0.00000226, -0.00000020, -0.00000151, 0.00000536, -0.00000016, -0.00001142, },
		{ 0.00001067, -0.00000160, -0.00000202, -0.00000930, 0.00000953, 0.00000228, 0.00000637, -0.00001578, 0.00000105, -0.00000376, 0.00000918, -0.00000256, -0.00000966, -0.00000073, 0.00003266, -0.00002931, 0.00001503, -0.00001213, 0.00000238, 0.00000529, -0.00001149, 0.00001231, -0.00001531, -0.00000585, 0.00000752, -0.00001139, -0.00000016, 0.00001852, -0.00001670, 0.00001785, -0.00002205, 0.00001694, -0.00001781, 0.00002734, -0.00002217, 0.00001395, -0.00001787, 0.00001502, -0.00001862, 0.00002601, -0.00002390, 0.00002340, -0.00002521, 0.00002773, -0.00003084, 0.00002316, -0.00002392, 0.00002928, -0.00003746, 0.00003700, -0.00005368, 0.00004673, -0.00005890, 0.00004917, -0.00006432, 0.00006770, -0.00015527, 0.00012007, -0.00055287, 0.00091697, 0.00007023, -0.00033918, 0.00009022, -0.00011553, 0.00007505, -0.00007237, 0.00005209, -0.00006336, 0.00003752, -0.00003790, 0.00002732, -0.00002362, 0.00002532, -0.00002617, 0.00001929, -0.00002884, 0.00002472, -0.00001975, 0.00001467, -0.00001867, 0.00001392, -0.00001028, 0.00000974, -0.00001160, 0.00000909, -0.00001720, 0.00001208, -0.00000043, -0.00000144, -0.00000113, -0.00000106, 0.00000151, -0.00002405, 0.00001988, -0.00000693, 0.00000972, -0.00000154, 0.00000273, -0.00000105, 0.00002202, -0.00001058, 0.00001761, -0.00002250, -0.00000781, 0.00002384, -0.00001960, 0.00000463, -0.00000924, 0.00001502, -0.00001017, -0.00000051, 0.00000082, 0.00000701, -0.00001155, 0.00001029, 0.00000192, -0.00000022, -0.00000671, 0.00000060, 0.00000594, 0.00000094, -0.00001131, 0.00001203, -0.00000743, 0.00000554, -0.00000375, 0.00001057, -0.00001806, },
		{ 0.00000535, 0.00000198, -0.00000080, -0.00001071, 0.00000670, 0.00000339, 0.00000751, -0.00001111, -0.00000445, -0.00000154, 0.00000432, 0.00000430, -0.00000922, -0.00000862, 0.00002922, -0.00000861, -0.00001006, 0.00001263, -0.00002165, 0.00002531, -0.00002528, 0.00002105, -0.00001769, -0.00000254, -0.00000343, 0.00000724, -0.00002347, 0.00003717, -0.00003010, 0.00002974, -0.00003104, 0.00002459, -0.00002685, 0.00003365, -0.00002495, 0.00001875, -0.00002488, 0.00002348, -0.00002908, 0.00003479, -0.00003101, 0.00003108, -0.00003372, 0.00003621, -0.00003755, 0.00003094, -0.00003692, 0.00004521, -0.00005288, 0.00005360, -0.00006819, 0.00006021, -0.00007403, 0.00006929, -0.00009465, 0.00011349, -0.00019561, 0.00018046, -0.00051282, 0.00045181, 0.00064508, -0.00057190, 0.00017130, -0.00018659, 0.00012088, -0.00010808, 0.00008108, -0.00008421, 0.00005266, -0.00005376, 0.00004200, -0.00004036, 0.00004172, -0.00003964, 0.00003268, -0.00004032, 0.00003138, -0.00002496, 0.00002092, -0.00002416, 0.00001800, -0.00001513, 0.00001557, -0.00001701, 0.00001434, -0.00001949, 0.00000959, 0.00000105, 0.00000093, -0.00000422, 0.00000096, 0.00000162, -0.00001781, 0.00000421, 0.00000929, -0.00000579, 0.00001264, -0.00001027, 0.00000912, 0.00000593, 0.00001110, -0.00000692, 0.00001240, -0.00004238, 0.00004345, -0.00002653, 0.00001170, -0.00001924, 0.00002089, -0.00001157, 0.00000347, -0.00000830, 0.00001590, -0.00001659, 0.00001127, -0.00000024, 0.00000612, -0.00001202, 0.00000446, -0.00000201, 0.00001220, -0.00001970, 0.00001548, -0.00000925, 0.00000814, -0.00000800, 0.00001446, -0.00001670, },
		{ -0.00000500, 0.00000803, -0.00000237, -0.00000755, -0.00000036, 0.00000774, 0.00000424, -0.00000250, -0.00001153, 0.00000238, -0.00000249, 0.00001070, -0.00000869, -0.00000946, 0.00001404, 0.00001754, -0.00003057, 0.00002602, -0.00002948, 0.00002583, -0.00001926, 0.00001187, -0.00000623, -0.00000817, -0.00000665, 0.00001476, -0.00002965, 0.00003456, -0.00002166, 0.00001977, -0.00001832, 0.00001137, -0.00001469, 0.00001818, -0.00000725, 0.00000420, -0.00001197, 0.00001071, -0.00001621, 0.00001842, -0.00001250, 0.00001268, -0.00001488, 0.00001565, -0.00001444, 0.00000880, -0.00001757, 0.00002390, -0.00002665, 0.00002423, -0.00003253, 0.00002021, -0.00003131, 0.00002438, -0.00004693, 0.00005660, -0.00010196, 0.00005336, -0.00022601, -0.00014255, 0.00107349, -0.00054931, 0.00006840, -0.00012917, 0.00006273, -0.00006003, 0.00003968, -0.00004539, 0.00001821, -0.00002722, 0.00001906, -0.00002254, 0.00002484, -0.00002196, 0.00001702, -0.00002397, 0.00001278, -0.00000846, 0.00000772, -0.00001165, 0.00000600, -0.00000559, 0.00000779, -0.00000928, 0.00000673, -0.00000980, -0.00000167, 0.00000749, -0.00000026, -0.00000380, 0.00000048, 0.00000092, -0.00000854, -0.00001191, 0.00002253, -0.00001691, 0.00002235, -0.00001904, 0.00001671, -0.00000814, 0.00002789, -0.00002318, 0.00003171, -0.00005134, 0.00003463, -0.00001118, 0.00000274, -0.00001402, 0.00001186, -0.00000171, -0.00000065, -0.00000845, 0.00001373, -0.00001066, 0.00000374, 0.00000345, 0.00000662, -0.00001067, 0.00000259, -0.00000430, 0.00001522, -0.00001767, 0.00000912, -0.00000318, 0.00000387, -0.00000491, 0.00000979, -0.00000734, },
		{ -0.00001500, 0.00001234, -0.00000370, -0.00000302, -0.00000768, 0.00001121, 0.00000112, 0.00000463, -0.00001501, 0.00000327, -0.00000596, 0.00001180, -0.00000459, -0.00000876, -0.00000072, 0.00003206, -0.00003037, 0.00001622, -0.00001423, 0.00000505, 0.00000387, -0.00000964, 0.00001269, -0.00001828, -0.00000323, 0.00001011, -0.00001841, 0.00001318, 0.00000359, -0.00000541, 0.00000813, -0.00001404, 0.00000956, -0.00000913, 0.00002048, -0.00001959, 0.00001122, -0.00001369, 0.00000985, -0.00001176, 0.00001923, -0.00001881, 0.00001778, -0.00001944, 0.00002293, -0.00002751, 0.00001853, -0.00001782, 0.00002179, -0.00002859, 0.00002734, -0.00004406, 0.00003727, -0.00005039, 0.00003834, -0.00004995, 0.00004847, -0.00013650, 0.00008991, -0.00056797, 0.00110948, -0.00017154, -0.00022470, 0.00004124, -0.00007979, 0.00005063, -0.00005174, 0.00003392, -0.00004879, 0.00002750, -0.00002973, 0.00001985, -0.00001601, 0.00001762, -0.00001864, 0.00001109, -0.00002172, 0.00002091, -0.00001716, 0.00001191, -0.00001585, 0.00001303, -0.00000910, 0.00000749, -0.00000954, 0.00000720, -0.00001714, 0.00001592, -0.00000420, 0.00000062, -0.00000364, 0.00000231, -0.00000243, -0.00001987, 0.00002389, -0.00001516, 0.00001894, -0.00001425, 0.00001103, -0.00000725, 0.00002532, -0.00001631, 0.00002237, -0.00002808, -0.00000063, 0.00001966, -0.00001590, 0.00000121, -0.00000633, 0.00001368, -0.00000903, -0.00000189, 0.00000324, 0.00000225, -0.00000805, 0.00000991, 0.00000301, -0.00000400, -0.00000399, -0.00000114, 0.00001024, -0.00000696, -0.00000376, 0.00000712, -0.00000400, 0.00000235, 0.00000061, 0.00000479, },
		{ -0.00001858, 0.00001063, -0.00000145, -0.00000092, -0.00001093, 0.00000980, 0.00000210, 0.00000633, -0.00001193, -0.00000119, -0.00000371, 0.00000609, 0.00000380, -0.00000998, -0.00000735, 0.00002808, -0.00000924, -0.00000976, 0.00001179, -0.00002089, 0.00002613, -0.00002585, 0.00002368, -0.00002105, -0.00000199, 0.00000205, -0.00000139, -0.00001170, 0.00002815, -0.00002718, 0.00002902, -0.00003222, 0.00002638, -0.00002818, 0.00003792, -0.00003272, 0.00002487, -0.00002927, 0.00002732, -0.00003230, 0.00003943, -0.00003780, 0.00003759, -0.00004089, 0.00004494, -0.00004755, 0.00003956, -0.00004492, 0.00005330, -0.00006154, 0.00006378, -0.00008079, 0.00007533, -0.00009220, 0.00008804, -0.00011521, 0.00013827, -0.00023293, 0.00022995, -0.00063297, 0.00070344, 0.00045204, -0.00055424, 0.00019844, -0.00020576, 0.00013989, -0.00012278, 0.00009364, -0.00009704, 0.00006640, -0.00006570, 0.00005193, -0.00004831, 0.00004844, -0.00004572, 0.00003790, -0.00004672, 0.00004012, -0.00003314, 0.00002775, -0.00003040, 0.00002520, -0.00002080, 0.00001983, -0.00002151, 0.00001910, -0.00002593, 0.00001822, -0.00000516, 0.00000444, -0.00000769, 0.00000412, -0.00000033, -0.00001815, 0.00001309, -0.00000139, 0.00000358, 0.00000279, -0.00000674, 0.00000748, 0.00000595, 0.00000821, -0.00000526, 0.00000975, -0.00003922, 0.00004366, -0.00002650, 0.00001173, -0.00001966, 0.00002226, -0.00001218, 0.00000309, -0.00000656, 0.00001260, -0.00001547, 0.00001242, 0.00000072, 0.00000266, -0.00001008, 0.00000273, 0.00000268, 0.00000540, -0.00001534, 0.00001438, -0.00000905, 0.00000757, -0.00000636, 0.00001267, },
		{ -0.00001338, 0.00000222, 0.00000469, -0.00000228, -0.00000896, 0.00000318, 0.00000700, 0.00000332, -0.00000425, -0.00000842, 0.00000150, -0.00000249, 0.00001218, -0.00001103, -0.00000728, 0.00001254, 0.00001756, -0.00003152, 0.00002671, -0.00003074, 0.00002866, -0.00002163, 0.00001565, -0.00000899, -0.00000935, -0.00000070, 0.00000874, -0.00002440, 0.00003553, -0.00002937, 0.00002870, -0.00002833, 0.00002137, -0.00002452, 0.00003100, -0.00002184, 0.00001563, -0.00002186, 0.00002080, -0.00002668, 0.00003111, -0.00002712, 0.00002685, -0.00003005, 0.00003234, -0.00003179, 0.00002473, -0.00003384, 0.00004154, -0.00004625, 0.00004667, -0.00005836, 0.00004924, -0.00006378, 0.00005864, -0.00008592, 0.00010413, -0.00016563, 0.00013465, -0.00036664, 0.00005124, 0.00102934, -0.00067546, 0.00018123, -0.00020126, 0.00012054, -0.00010477, 0.00007760, -0.00007988, 0.00004970, -0.00005360, 0.00004144, -0.00004186, 0.00004235, -0.00003838, 0.00003245, -0.00004024, 0.00002984, -0.00002314, 0.00002020, -0.00002318, 0.00001762, -0.00001485, 0.00001574, -0.00001763, 0.00001519, -0.00001921, 0.00000790, 0.00000212, 0.00000233, -0.00000641, 0.00000210, 0.00000138, -0.00001201, -0.00000187, 0.00001455, -0.00001286, 0.00002008, -0.00002347, 0.00002153, -0.00001292, 0.00002952, -0.00002625, 0.00003365, -0.00005362, 0.00003991, -0.00001512, 0.00000643, -0.00001776, 0.00001572, -0.00000400, 0.00000039, -0.00000838, 0.00001313, -0.00001273, 0.00000684, 0.00000343, 0.00000493, -0.00001106, 0.00000261, -0.00000170, 0.00001246, -0.00001885, 0.00001312, -0.00000688, 0.00000654, -0.00000673, 0.00001234, },
		{ -0.00000257, -0.00000886, 0.00001115, -0.00000453, -0.00000451, -0.00000492, 0.00001172, 0.00000012, 0.00000266, -0.00001299, 0.00000418, -0.00000761, 0.00001447, -0.00000729, -0.00000697, -0.00000157, 0.00003195, -0.00003166, 0.00001726, -0.00001602, 0.00000805, 0.00000183, -0.00000691, 0.00001211, -0.00002122, 0.00000179, 0.00000847, -0.00002041, 0.00002258, -0.00001069, 0.00000754, -0.00000433, -0.00000264, -0.00000144, 0.00000400, 0.00000776, -0.00001110, 0.00000358, -0.00000514, 0.00000041, -0.00000018, 0.00000672, -0.00000743, 0.00000561, -0.00000650, 0.00001034, -0.00001654, 0.00000663, -0.00000419, 0.00000623, -0.00001066, 0.00000715, -0.00002227, 0.00001399, -0.00002676, 0.00001101, -0.00001626, 0.00000475, -0.00008548, 0.00001211, -0.00051344, 0.00124496, -0.00041161, -0.00007808, -0.00003799, -0.00002035, 0.00000680, -0.00001500, 0.00000138, -0.00002065, 0.00000519, -0.00001102, 0.00000343, -0.00000065, 0.00000288, -0.00000450, -0.00000369, -0.00000742, 0.00000995, -0.00000825, 0.00000357, -0.00000766, 0.00000720, -0.00000379, 0.00000137, -0.00000346, 0.00000089, -0.00001216, 0.00001561, -0.00000545, 0.00000099, -0.00000491, 0.00000532, -0.00000740, -0.00001182, 0.00002181, -0.00001851, 0.00002460, -0.00002518, 0.00001987, -0.00001451, 0.00002952, -0.00002217, 0.00002695, -0.00003355, 0.00000694, 0.00001450, -0.00001107, -0.00000336, -0.00000226, 0.00001145, -0.00000768, -0.00000291, 0.00000474, -0.00000196, -0.00000412, 0.00000935, 0.00000316, -0.00000674, -0.00000203, -0.00000127, 0.00001201, -0.00001309, 0.00000372, 0.00000142, 0.00000025, -0.00000139, 0.00000550, },
		{ 0.00000747, -0.00001639, 0.00001310, -0.00000390, -0.00000164, -0.00000979, 0.00001181, 0.00000096, 0.00000458, -0.00001133, 0.00000115, -0.00000594, 0.00000845, 0.00000238, -0.00000998, -0.00000653, 0.00002693, -0.00000965, -0.00000988, 0.00001120, -0.00001967, 0.00002638, -0.00002588, 0.00002613, -0.00002548, 0.00000151, 0.00000427, -0.00000785, 0.00000063, 0.00001482, -0.00001838, 0.00002221, -0.00002732, 0.00002221, -0.00002305, 0.00003507, -0.00003425, 0.00002609, -0.00002908, 0.00002634, -0.00003001, 0.00003767, -0.00003798, 0.00003746, -0.00004103, 0.00004647, -0.00005084, 0.00004178, -0.00004599, 0.00005367, -0.00006140, 0.00006388, -0.00008214, 0.00007825, -0.00009761, 0.00009311, -0.00011977, 0.00014328, -0.00024596, 0.00024808, -0.00071999, 0.00095784, 0.00020052, -0.00046664, 0.00017727, -0.00019171, 0.00013352, -0.00011714, 0.00008859, -0.00009413, 0.00006671, -0.00006657, 0.00005251, -0.00004801, 0.00004744, -0.00004440, 0.00003573, -0.00004551, 0.00004193, -0.00003565, 0.00002972, -0.00003207, 0.00002841, -0.00002326, 0.00002094, -0.00002270, 0.00002042, -0.00002910, 0.00002504, -0.00001130, 0.00000844, -0.00001200, 0.00000898, -0.00000521, -0.00001375, 0.00001708, -0.00001014, 0.00001316, -0.00000927, 0.00000064, 0.00000249, 0.00000853, 0.00000365, -0.00000202, 0.00000529, -0.00003375, 0.00004124, -0.00002405, 0.00000960, -0.00001821, 0.00002227, -0.00001214, 0.00000242, -0.00000467, 0.00000844, -0.00001273, 0.00001262, 0.00000162, -0.00000103, -0.00000759, 0.00000112, 0.00000708, -0.00000260, -0.00000811, 0.00001012, -0.00000616, 0.00000467, -0.00000227, },
		{ 0.00001156, -0.00001633, 0.00000820, 0.00000104, -0.00000220, -0.00000950, 0.00000630, 0.00000602, 0.00000190, -0.00000493, -0.00000550, -0.00000007, -0.00000168, 0.00001294, -0.00001318, -0.00000482, 0.00001036, 0.00001830, -0.00003317, 0.00002773, -0.00003172, 0.00003138, -0.00002413, 0.00002013, -0.00001348, -0.00000798, 0.00000397, 0.00000164, -0.00001492, 0.00002951, -0.00003022, 0.00003187, -0.00003342, 0.00002707, -0.00003005, 0.00003973, -0.00003393, 0.00002587, -0.00003049, 0.00002918, -0.00003483, 0.00004104, -0.00003935, 0.00003897, -0.00004320, 0.00004736, -0.00004829, 0.00004006, -0.00004897, 0.00005753, -0.00006365, 0.00006652, -0.00008161, 0.00007611, -0.00009489, 0.00009196, -0.00012354, 0.00014982, -0.00022839, 0.00021760, -0.00052248, 0.00030331, 0.00089055, -0.00072694, 0.00025748, -0.00025385, 0.00016529, -0.00013996, 0.00010750, -0.00010762, 0.00007632, -0.00007671, 0.00006131, -0.00005886, 0.00005739, -0.00005220, 0.00004510, -0.00005369, 0.00004502, -0.00003707, 0.00003231, -0.00003435, 0.00002927, -0.00002451, 0.00002377, -0.00002581, 0.00002352, -0.00002904, 0.00001946, -0.00000636, 0.00000779, -0.00001174, 0.00000681, -0.00000166, -0.00001167, 0.00000613, 0.00000472, -0.00000503, 0.00001283, -0.00002260, 0.00002262, -0.00001514, 0.00002933, -0.00002732, 0.00003326, -0.00005320, 0.00004252, -0.00001692, 0.00000838, -0.00002013, 0.00001871, -0.00000589, 0.00000095, -0.00000755, 0.00001086, -0.00001272, 0.00000882, 0.00000388, 0.00000203, -0.00000988, 0.00000154, 0.00000248, 0.00000637, -0.00001562, 0.00001328, -0.00000764, 0.00000665, -0.00000582, },
		{ 0.00000892, -0.00000932, -0.00000143, 0.00000796, -0.00000444, -0.00000601, -0.00000163, 0.00001142, -0.00000104, 0.00000110, -0.00001032, 0.00000399, -0.00000854, 0.00001692, -0.00001048, -0.00000427, -0.00000351, 0.00003277, -0.00003370, 0.00001861, -0.00001775, 0.00001144, -0.00000093, -0.00000298, 0.00000982, -0.00002275, 0.00000767, 0.00000348, -0.00001685, 0.00002568, -0.00002082, 0.00001853, -0.00001611, 0.00000897, -0.00001295, 0.00001842, -0.00000803, 0.00000128, -0.00000736, 0.00000619, -0.00001149, 0.00001398, -0.00000924, 0.00000799, -0.00001092, 0.00001150, -0.00000818, 0.00000049, -0.00001088, 0.00001506, -0.00001519, 0.00001374, -0.00002067, 0.00000870, -0.00002030, 0.00000932, -0.00003001, 0.00003348, -0.00005992, -0.00000546, -0.00011300, -0.00037827, 0.00129475, -0.00061399, 0.00007267, -0.00012972, 0.00005074, -0.00004746, 0.00003058, -0.00003930, 0.00001585, -0.00002530, 0.00001502, -0.00001924, 0.00001996, -0.00001647, 0.00001380, -0.00002255, 0.00001175, -0.00000639, 0.00000571, -0.00000932, 0.00000506, -0.00000298, 0.00000515, -0.00000805, 0.00000602, -0.00000938, -0.00000214, 0.00001019, -0.00000277, -0.00000187, -0.00000316, 0.00000578, -0.00001078, -0.00000253, 0.00001494, -0.00001647, 0.00002513, -0.00003209, 0.00002699, -0.00002109, 0.00003333, -0.00002716, 0.00003036, -0.00003766, 0.00001347, 0.00000975, -0.00000635, -0.00000806, 0.00000219, 0.00000881, -0.00000643, -0.00000326, 0.00000476, -0.00000469, -0.00000061, 0.00000906, 0.00000199, -0.00000789, -0.00000134, 0.00000069, 0.00001019, -0.00001553, 0.00000887, -0.00000301, 0.00000345, -0.00000389, },
		{ 0.00000288, 0.00000013, -0.00001071, 0.00001244, -0.00000493, -0.00000303, -0.00000741, 0.00001266, -0.00000005, 0.00000257, -0.00000948, 0.00000231, -0.00000781, 0.00001110, 0.00000005, -0.00000892, -0.00000663, 0.00002636, -0.00001041, -0.00001002, 0.00001069, -0.00001793, 0.00002586, -0.00002490, 0.00002763, -0.00003001, 0.00000749, 0.00000248, -0.00000989, 0.00000997, 0.00000088, -0.00000659, 0.00001140, -0.00001784, 0.00001325, -0.00001231, 0.00002528, -0.00002893, 0.00002155, -0.00002359, 0.00002007, -0.00002190, 0.00002905, -0.00003080, 0.00002969, -0.00003288, 0.00003912, -0.00004546, 0.00003568, -0.00003833, 0.00004460, -0.00005067, 0.00005182, -0.00006958, 0.00006554, -0.00008604, 0.00007972, -0.00010286, 0.00012156, -0.00022493, 0.00021927, -0.00074673, 0.00117726, -0.00007653, -0.00032571, 0.00011266, -0.00014524, 0.00010131, -0.00009034, 0.00006500, -0.00007416, 0.00005183, -0.00005443, 0.00004196, -0.00003797, 0.00003754, -0.00003469, 0.00002530, -0.00003562, 0.00003529, -0.00003074, 0.00002512, -0.00002749, 0.00002596, -0.00002093, 0.00001752, -0.00001923, 0.00001674, -0.00002690, 0.00002748, -0.00001484, 0.00001069, -0.00001502, 0.00001352, -0.00001120, -0.00000554, 0.00001527, -0.00001448, 0.00002006, -0.00002069, 0.00000963, -0.00000446, 0.00001272, -0.00000180, 0.00000189, 0.00000015, -0.00002714, 0.00003717, -0.00001993, 0.00000583, -0.00001519, 0.00002108, -0.00001168, 0.00000189, -0.00000342, 0.00000447, -0.00000921, 0.00001238, 0.00000183, -0.00000404, -0.00000547, 0.00000077, 0.00000940, -0.00000942, -0.00000026, 0.00000446, -0.00000199, 0.00000077, },
		{ -0.00000215, 0.00000710, -0.00001510, 0.00001138, -0.00000168, -0.00000272, -0.00000873, 0.00000843, 0.00000516, -0.00000015, -0.00000417, -0.00000343, -0.00000177, -0.00000027, 0.00001284, -0.00001474, -0.00000255, 0.00000795, 0.00001946, -0.00003545, 0.00002918, -0.00003249, 0.00003386, -0.00002645, 0.00002488, -0.00001949, -0.00000339, 0.00000548, -0.00000370, -0.00000432, 0.00001858, -0.00002478, 0.00002883, -0.00003247, 0.00002700, -0.00002941, 0.00004182, -0.00004045, 0.00003195, -0.00003525, 0.00003343, -0.00003821, 0.00004542, -0.00004598, 0.00004555, -0.00005053, 0.00005655, -0.00005959, 0.00005057, -0.00005876, 0.00006747, -0.00007407, 0.00007834, -0.00009597, 0.00009355, -0.00011649, 0.00011545, -0.00014976, 0.00018164, -0.00027501, 0.00028191, -0.00066629, 0.00058617, 0.00066942, -0.00069707, 0.00028299, -0.00027374, 0.00018590, -0.00015624, 0.00012122, -0.00012107, 0.00009099, -0.00009016, 0.00007315, -0.00006875, 0.00006573, -0.00005953, 0.00005125, -0.00006047, 0.00005428, -0.00004636, 0.00004056, -0.00004193, 0.00003791, -0.00003183, 0.00002942, -0.00003140, 0.00002914, -0.00003641, 0.00003010, -0.00001563, 0.00001446, -0.00001852, 0.00001378, -0.00000803, -0.00000648, 0.00000944, -0.00000361, 0.00000375, 0.00000260, -0.00001735, 0.00002007, -0.00001467, 0.00002724, -0.00002654, 0.00003090, -0.00005036, 0.00004247, -0.00001634, 0.00000822, -0.00002070, 0.00002048, -0.00000725, 0.00000124, -0.00000647, 0.00000757, -0.00001095, 0.00000977, 0.00000443, -0.00000138, -0.00000783, 0.00000029, 0.00000676, -0.00000144, -0.00000900, 0.00000992, -0.00000546, 0.00000419, },
		{ -0.00000374, 0.00000917, -0.00001327, 0.00000511, 0.00000419, -0.00000419, -0.00000682, 0.00000141, 0.00001084, -0.00000277, 0.00000064, -0.00000784, 0.00000316, -0.00000883, 0.00001895, -0.00001388, -0.00000083, -0.00000652, 0.00003470, -0.00003686, 0.00002073, -0.00001973, 0.00001530, -0.00000442, 0.00000227, 0.00000526, -0.00002146, 0.00001227, -0.00000268, -0.00000911, 0.00002205, -0.00002475, 0.00002480, -0.00002418, 0.00001763, -0.00002166, 0.00003045, -0.00002315, 0.00001430, -0.00001886, 0.00001768, -0.00002305, 0.00002756, -0.00002515, 0.00002385, -0.00002804, 0.00003064, -0.00002881, 0.00001999, -0.00003044, 0.00003605, -0.00003814, 0.00003968, -0.00005052, 0.00004272, -0.00005907, 0.00005109, -0.00007727, 0.00009056, -0.00013517, 0.00009132, -0.00027280, -0.00016643, 0.00124332, -0.00074911, 0.00019954, -0.00021332, 0.00011810, -0.00009996, 0.00007480, -0.00007912, 0.00005273, -0.00005721, 0.00004271, -0.00004329, 0.00004147, -0.00003633, 0.00003227, -0.00004146, 0.00003180, -0.00002458, 0.00002171, -0.00002410, 0.00001990, -0.00001546, 0.00001589, -0.00001891, 0.00001696, -0.00002164, 0.00001139, 0.00000032, 0.00000387, -0.00000836, 0.00000246, 0.00000220, -0.00001012, 0.00000475, 0.00000610, -0.00001044, 0.00002067, -0.00003406, 0.00003096, -0.00002574, 0.00003580, -0.00003061, 0.00003203, -0.00003968, 0.00001798, 0.00000652, -0.00000280, -0.00001194, 0.00000631, 0.00000615, -0.00000537, -0.00000299, 0.00000324, -0.00000551, 0.00000195, 0.00000919, -0.00000043, -0.00000742, -0.00000179, 0.00000434, 0.00000488, -0.00001350, 0.00001028, -0.00000468, 0.00000424, },
		{ -0.00000229, 0.00000706, -0.00000716, -0.00000323, 0.00000943, -0.00000481, -0.00000465, -0.00000431, 0.00001267, -0.00000126, 0.00000096, -0.00000698, 0.00000247, -0.00000920, 0.00001385, -0.00000315, -0.00000662, -0.00000808, 0.00002697, -0.00001220, -0.00000953, 0.00000987, -0.00001541, 0.00002426, -0.00002238, 0.00002722, -0.00003313, 0.00001445, -0.00000260, -0.00000678, 0.00001377, -0.00001008, 0.00000469, -0.00000023, -0.00000663, 0.00000219, 0.00000142, 0.00001073, -0.00001812, 0.00001214, -0.00001373, 0.00000963, -0.00000933, 0.00001502, -0.00001756, 0.00001549, -0.00001749, 0.00002361, -0.00003177, 0.00002162, -0.00002254, 0.00002689, -0.00003040, 0.00002874, -0.00004411, 0.00003777, -0.00005758, 0.00004773, -0.00006425, 0.00007235, -0.00016749, 0.00013647, -0.00069362, 0.00132604, -0.00034033, -0.00015837, 0.00001970, -0.00007562, 0.00004974, -0.00004742, 0.00002705, -0.00004048, 0.00002412, -0.00003094, 0.00002161, -0.00001944, 0.00002003, -0.00001797, 0.00000806, -0.00001832, 0.00002098, -0.00001872, 0.00001408, -0.00001668, 0.00001764, -0.00001360, 0.00000951, -0.00001109, 0.00000787, -0.00001856, 0.00002395, -0.00001382, 0.00000921, -0.00001463, 0.00001530, -0.00001559, 0.00000391, 0.00000878, -0.00001369, 0.00002246, -0.00002908, 0.00001799, -0.00001173, 0.00001735, -0.00000721, 0.00000552, -0.00000447, -0.00002075, 0.00003276, -0.00001523, 0.00000129, -0.00001123, 0.00001906, -0.00001100, 0.00000180, -0.00000337, 0.00000163, -0.00000579, 0.00001222, 0.00000088, -0.00000568, -0.00000436, 0.00000230, 0.00000836, -0.00001293, 0.00000578, -0.00000046, 0.00000159, },
		{ 0.00000027, 0.00000310, -0.00000001, -0.00001000, 0.00001125, -0.00000284, -0.00000406, -0.00000660, 0.00000941, 0.00000443, -0.00000252, -0.00000210, -0.00000250, -0.00000324, 0.00000156, 0.00001175, -0.00001528, -0.00000107, 0.00000600, 0.00002037, -0.00003786, 0.00003081, -0.00003287, 0.00003576, -0.00002798, 0.00002905, -0.00002606, 0.00000405, 0.00000295, -0.00000501, 0.00000397, 0.00000615, -0.00001543, 0.00002103, -0.00002633, 0.00002158, -0.00002265, 0.00003660, -0.00003992, 0.00003216, -0.00003460, 0.00003222, -0.00003556, 0.00004279, -0.00004520, 0.00004450, -0.00004962, 0.00005703, -0.00006249, 0.00005308, -0.00006011, 0.00006820, -0.00007415, 0.00007830, -0.00009686, 0.00009602, -0.00012208, 0.00012176, -0.00015619, 0.00018924, -0.00029180, 0.00030768, -0.00076771, 0.00086250, 0.00039393, -0.00059509, 0.00025573, -0.00025621, 0.00017752, -0.00014917, 0.00011470, -0.00011620, 0.00008953, -0.00008993, 0.00007331, -0.00006841, 0.00006468, -0.00005799, 0.00004867, -0.00005822, 0.00005491, -0.00004815, 0.00004220, -0.00004330, 0.00004098, -0.00003450, 0.00003064, -0.00003238, 0.00002982, -0.00003860, 0.00003671, -0.00002274, 0.00001980, -0.00002442, 0.00002085, -0.00001594, 0.00000271, 0.00000705, -0.00000789, 0.00001061, -0.00000797, -0.00000962, 0.00001490, -0.00001208, 0.00002373, -0.00002454, 0.00002741, -0.00004595, 0.00004044, -0.00001381, 0.00000617, -0.00001955, 0.00002097, -0.00000814, 0.00000159, -0.00000584, 0.00000420, -0.00000817, 0.00001012, 0.00000449, -0.00000442, -0.00000581, -0.00000006, 0.00000939, -0.00000867, -0.00000111, 0.00000457, -0.00000153, },
		{ 0.00000228, -0.00000061, 0.00000558, -0.00001306, 0.00000898, 0.00000124, -0.00000452, -0.00000635, 0.00000355, 0.00001038, -0.00000522, 0.00000169, -0.00000624, 0.00000225, -0.00000866, 0.00002038, -0.00001709, 0.00000288, -0.00001021, 0.00003750, -0.00004113, 0.00002376, -0.00002207, 0.00001952, -0.00000835, 0.00000852, -0.00000151, -0.00001646, 0.00001360, -0.00000725, -0.00000002, 0.00001336, -0.00002250, 0.00002528, -0.00002689, 0.00002134, -0.00002515, 0.00003702, -0.00003408, 0.00002462, -0.00002794, 0.00002652, -0.00003140, 0.00003732, -0.00003740, 0.00003625, -0.00004157, 0.00004639, -0.00004688, 0.00003743, -0.00004750, 0.00005401, -0.00005739, 0.00006124, -0.00007561, 0.00007202, -0.00009377, 0.00008926, -0.00012044, 0.00014273, -0.00020593, 0.00018564, -0.00044362, 0.00010204, 0.00109267, -0.00080042, 0.00028119, -0.00027088, 0.00016738, -0.00013888, 0.00010754, -0.00010891, 0.00008158, -0.00008305, 0.00006557, -0.00006314, 0.00005890, -0.00005210, 0.00004651, -0.00005595, 0.00004815, -0.00004026, 0.00003584, -0.00003719, 0.00003356, -0.00002731, 0.00002580, -0.00002862, 0.00002658, -0.00003289, 0.00002548, -0.00001170, 0.00001289, -0.00001732, 0.00001124, -0.00000544, -0.00000425, 0.00000729, -0.00000141, -0.00000303, 0.00001284, -0.00003155, 0.00003145, -0.00002797, 0.00003656, -0.00003241, 0.00003200, -0.00003956, 0.00002015, 0.00000532, -0.00000102, -0.00001441, 0.00000957, 0.00000371, -0.00000439, -0.00000253, 0.00000067, -0.00000458, 0.00000351, 0.00000948, -0.00000357, -0.00000587, -0.00000262, 0.00000842, -0.00000258, -0.00000769, 0.00000793, -0.00000327, },
		{ 0.00000299, -0.00000292, 0.00000835, -0.00001206, 0.00000379, 0.00000553, -0.00000433, -0.00000577, -0.00000130, 0.00001237, -0.00000309, 0.00000047, -0.00000469, 0.00000211, -0.00001021, 0.00001656, -0.00000702, -0.00000309, -0.00001104, 0.00002914, -0.00001561, -0.00000778, 0.00000823, -0.00001186, 0.00002138, -0.00001795, 0.00002408, -0.00003320, 0.00002016, -0.00000892, 0.00000033, 0.00001130, -0.00001570, 0.00001248, -0.00000958, 0.00000317, -0.00000773, 0.00001460, -0.00000511, -0.00000471, 0.00000026, -0.00000174, -0.00000257, 0.00000494, -0.00000144, -0.00000131, -0.00000203, 0.00000194, 0.00000297, -0.00001249, 0.00000231, -0.00000157, 0.00000389, -0.00000443, -0.00000099, -0.00001044, -0.00000022, -0.00001706, 0.00000226, -0.00000977, 0.00000230, -0.00008069, 0.00000473, -0.00055466, 0.00137871, -0.00055466, 0.00000473, -0.00008069, 0.00000230, -0.00000977, 0.00000226, -0.00001706, -0.00000022, -0.00001044, -0.00000099, -0.00000443, 0.00000389, -0.00000157, 0.00000231, -0.00001249, 0.00000297, 0.00000194, -0.00000203, -0.00000131, -0.00000144, 0.00000494, -0.00000257, -0.00000174, 0.00000026, -0.00000471, -0.00000511, 0.00001460, -0.00000773, 0.00000317, -0.00000958, 0.00001248, -0.00001570, 0.00001130, 0.00000033, -0.00000892, 0.00002016, -0.00003320, 0.00002408, -0.00001795, 0.00002138, -0.00001186, 0.00000823, -0.00000778, -0.00001561, 0.00002914, -0.00001104, -0.00000309, -0.00000702, 0.00001656, -0.00001021, 0.00000211, -0.00000469, 0.00000047, -0.00000309, 0.00001237, -0.00000130, -0.00000577, -0.00000433, 0.00000553, 0.00000379, -0.00001206, 0.00000835, -0.00000292, },
		{ 0.00000228, -0.00000327, 0.00000793, -0.00000769, -0.00000258, 0.00000842, -0.00000262, -0.00000587, -0.00000357, 0.00000948, 0.00000351, -0.00000458, 0.00000067, -0.00000253, -0.00000439, 0.00000371, 0.00000957, -0.00001441, -0.00000102, 0.00000532, 0.00002015, -0.00003956, 0.00003200, -0.00003241, 0.00003656, -0.00002797, 0.00003145, -0.00003155, 0.00001284, -0.00000303, -0.00000141, 0.00000729, -0.00000425, -0.00000544, 0.00001124, -0.00001732, 0.00001289, -0.00001170, 0.00002548, -0.00003289, 0.00002658, -0.00002862, 0.00002580, -0.00002731, 0.00003356, -0.00003719, 0.00003584, -0.00004026, 0.00004815, -0.00005595, 0.00004651, -0.00005210, 0.00005890, -0.00006314, 0.00006557, -0.00008305, 0.00008158, -0.00010891, 0.00010754, -0.00013888, 0.00016738, -0.00027088, 0.00028119, -0.00080042, 0.00109267, 0.00010204, -0.00044362, 0.00018564, -0.00020593, 0.00014273, -0.00012044, 0.00008926, -0.00009377, 0.00007202, -0.00007561, 0.00006124, -0.00005739, 0.00005401, -0.00004750, 0.00003743, -0.00004688, 0.00004639, -0.00004157, 0.00003625, -0.00003740, 0.00003732, -0.00003140, 0.00002652, -0.00002794, 0.00002462, -0.00003408, 0.00003702, -0.00002515, 0.00002134, -0.00002689, 0.00002528, -0.00002250, 0.00001336, -0.00000002, -0.00000725, 0.00001360, -0.00001646, -0.00000151, 0.00000852, -0.00000835, 0.00001952, -0.00002207, 0.00002376, -0.00004113, 0.00003750, -0.00001021, 0.00000288, -0.00001709, 0.00002038, -0.00000866, 0.00000225, -0.00000624, 0.00000169, -0.00000522, 0.00001038, 0.00000355, -0.00000635, -0.00000452, 0.00000124, 0.00000898, -0.00001306, 0.00000558, -0.00000061, },
		{ 0.00000027, -0.00000153, 0.00000457, -0.00000111, -0.00000867, 0.00000939, -0.00000006, -0.00000581, -0.00000442, 0.00000449, 0.00001012, -0.00000817, 0.00000420, -0.00000584, 0.00000159, -0.00000814, 0.00002097, -0.00001955, 0.00000617, -0.00001381, 0.00004044, -0.00004595, 0.00002741, -0.00002454, 0.00002373, -0.00001208, 0.00001490, -0.00000962, -0.00000797, 0.00001061, -0.00000789, 0.00000705, 0.00000271, -0.00001594, 0.00002085, -0.00002442, 0.00001980, -0.00002274, 0.00003671, -0.00003860, 0.00002982, -0.00003238, 0.00003064, -0.00003450, 0.00004098, -0.00004330, 0.00004220, -0.00004815, 0.00005491, -0.00005822, 0.00004867, -0.00005799, 0.00006468, -0.00006841, 0.00007331, -0.00008993, 0.00008953, -0.00011620, 0.00011470, -0.00014917, 0.00017752, -0.00025621, 0.00025573, -0.00059509, 0.00039393, 0.00086250, -0.00076771, 0.00030768, -0.00029180, 0.00018924, -0.00015619, 0.00012176, -0.00012208, 0.00009602, -0.00009686, 0.00007830, -0.00007415, 0.00006820, -0.00006011, 0.00005308, -0.00006249, 0.00005703, -0.00004962, 0.00004450, -0.00004520, 0.00004279, -0.00003556, 0.00003222, -0.00003460, 0.00003216, -0.00003992, 0.00003660, -0.00002265, 0.00002158, -0.00002633, 0.00002103, -0.00001543, 0.00000615, 0.00000397, -0.00000501, 0.00000295, 0.00000405, -0.00002606, 0.00002905, -0.00002798, 0.00003576, -0.00003287, 0.00003081, -0.00003786, 0.00002037, 0.00000600, -0.00000107, -0.00001528, 0.00001175, 0.00000156, -0.00000324, -0.00000250, -0.00000210, -0.00000252, 0.00000443, 0.00000941, -0.00000660, -0.00000406, -0.00000284, 0.00001125, -0.00001000, -0.00000001, 0.00000310, },
		{ -0.00000229, 0.00000159, -0.00000046, 0.00000578, -0.00001293, 0.00000836, 0.00000230, -0.00000436, -0.00000568, 0.00000088, 0.00001222, -0.00000579, 0.00000163, -0.00000337, 0.00000180, -0.00001100, 0.00001906, -0.00001123, 0.00000129, -0.00001523, 0.00003276, -0.00002075, -0.00000447, 0.00000552, -0.00000721, 0.00001735, -0.00001173, 0.00001799, -0.00002908, 0.00002246, -0.00001369, 0.00000878, 0.00000391, -0.00001559, 0.00001530, -0.00001463, 0.00000921, -0.00001382, 0.00002395, -0.00001856, 0.00000787, -0.00001109, 0.00000951, -0.00001360, 0.00001764, -0.00001668, 0.00001408, -0.00001872, 0.00002098, -0.00001832, 0.00000806, -0.00001797, 0.00002003, -0.00001944, 0.00002161, -0.00003094, 0.00002412, -0.00004048, 0.00002705, -0.00004742, 0.00004974, -0.00007562, 0.00001970, -0.00015837, -0.00034033, 0.00132604, -0.00069362, 0.00013647, -0.00016749, 0.00007235, -0.00006425, 0.00004773, -0.00005758, 0.00003777, -0.00004411, 0.00002874, -0.00003040, 0.00002689, -0.00002254, 0.00002162, -0.00003177, 0.00002361, -0.00001749, 0.00001549, -0.00001756, 0.00001502, -0.00000933, 0.00000963, -0.00001373, 0.00001214, -0.00001812, 0.00001073, 0.00000142, 0.00000219, -0.00000663, -0.00000023, 0.00000469, -0.00001008, 0.00001377, -0.00000678, -0.00000260, 0.00001445, -0.00003313, 0.00002722, -0.00002238, 0.00002426, -0.00001541, 0.00000987, -0.00000953, -0.00001220, 0.00002697, -0.00000808, -0.00000662, -0.00000315, 0.00001385, -0.00000920, 0.00000247, -0.00000698, 0.00000096, -0.00000126, 0.00001267, -0.00000431, -0.00000465, -0.00000481, 0.00000943, -0.00000323, -0.00000716, 0.00000706, },
		{ -0.00000374, 0.00000424, -0.00000468, 0.00001028, -0.00001350, 0.00000488, 0.00000434, -0.00000179, -0.00000742, -0.00000043, 0.00000919, 0.00000195, -0.00000551, 0.00000324, -0.00000299, -0.00000537, 0.00000615, 0.00000631, -0.00001194, -0.00000280, 0.00000652, 0.00001798, -0.00003968, 0.00003203, -0.00003061, 0.00003580, -0.00002574, 0.00003096, -0.00003406, 0.00002067, -0.00001044, 0.00000610, 0.00000475, -0.00001012, 0.00000220, 0.00000246, -0.00000836, 0.00000387, 0.00000032, 0.00001139, -0.00002164, 0.00001696, -0.00001891, 0.00001589, -0.00001546, 0.00001990, -0.00002410, 0.00002171, -0.00002458, 0.00003180, -0.00004146, 0.00003227, -0.00003633, 0.00004147, -0.00004329, 0.00004271, -0.00005721, 0.00005273, -0.00007912, 0.00007480, -0.00009996, 0.00011810, -0.00021332, 0.00019954, -0.00074911, 0.00124332, -0.00016643, -0.00027280, 0.00009132, -0.00013517, 0.00009056, -0.00007727, 0.00005109, -0.00005907, 0.00004272, -0.00005052, 0.00003968, -0.00003814, 0.00003605, -0.00003044, 0.00001999, -0.00002881, 0.00003064, -0.00002804, 0.00002385, -0.00002515, 0.00002756, -0.00002305, 0.00001768, -0.00001886, 0.00001430, -0.00002315, 0.00003045, -0.00002166, 0.00001763, -0.00002418, 0.00002480, -0.00002475, 0.00002205, -0.00000911, -0.00000268, 0.00001227, -0.00002146, 0.00000526, 0.00000227, -0.00000442, 0.00001530, -0.00001973, 0.00002073, -0.00003686, 0.00003470, -0.00000652, -0.00000083, -0.00001388, 0.00001895, -0.00000883, 0.00000316, -0.00000784, 0.00000064, -0.00000277, 0.00001084, 0.00000141, -0.00000682, -0.00000419, 0.00000419, 0.00000511, -0.00001327, 0.00000917, },
		{ -0.00000215, 0.00000419, -0.00000546, 0.00000992, -0.00000900, -0.00000144, 0.00000676, 0.00000029, -0.00000783, -0.00000138, 0.00000443, 0.00000977, -0.00001095, 0.00000757, -0.00000647, 0.00000124, -0.00000725, 0.00002048, -0.00002070, 0.00000822, -0.00001634, 0.00004247, -0.00005036, 0.00003090, -0.00002654, 0.00002724, -0.00001467, 0.00002007, -0.00001735, 0.00000260, 0.00000375, -0.00000361, 0.00000944, -0.00000648, -0.00000803, 0.00001378, -0.00001852, 0.00001446, -0.00001563, 0.00003010, -0.00003641, 0.00002914, -0.00003140, 0.00002942, -0.00003183, 0.00003791, -0.00004193, 0.00004056, -0.00004636, 0.00005428, -0.00006047, 0.00005125, -0.00005953, 0.00006573, -0.00006875, 0.00007315, -0.00009016, 0.00009099, -0.00012107, 0.00012122, -0.00015624, 0.00018590, -0.00027374, 0.00028299, -0.00069707, 0.00066942, 0.00058617, -0.00066629, 0.00028191, -0.00027501, 0.00018164, -0.00014976, 0.00011545, -0.00011649, 0.00009355, -0.00009597, 0.00007834, -0.00007407, 0.00006747, -0.00005876, 0.00005057, -0.00005959, 0.00005655, -0.00005053, 0.00004555, -0.00004598, 0.00004542, -0.00003821, 0.00003343, -0.00003525, 0.00003195, -0.00004045, 0.00004182, -0.00002941, 0.00002700, -0.00003247, 0.00002883, -0.00002478, 0.00001858, -0.00000432, -0.00000370, 0.00000548, -0.00000339, -0.00001949, 0.00002488, -0.00002645, 0.00003386, -0.00003249, 0.00002918, -0.00003545, 0.00001946, 0.00000795, -0.00000255, -0.00001474, 0.00001284, -0.00000027, -0.00000177, -0.00000343, -0.00000417, -0.00000015, 0.00000516, 0.00000843, -0.00000873, -0.00000272, -0.00000168, 0.00001138, -0.00001510, 0.00000710, },
		{ 0.00000288, 0.00000077, -0.00000199, 0.00000446, -0.00000026, -0.00000942, 0.00000940, 0.00000077, -0.00000547, -0.00000404, 0.00000183, 0.00001238, -0.00000921, 0.00000447, -0.00000342, 0.00000189, -0.00001168, 0.00002108, -0.00001519, 0.00000583, -0.00001993, 0.00003717, -0.00002714, 0.00000015, 0.00000189, -0.00000180, 0.00001272, -0.00000446, 0.00000963, -0.00002069, 0.00002006, -0.00001448, 0.00001527, -0.00000554, -0.00001120, 0.00001352, -0.00001502, 0.00001069, -0.00001484, 0.00002748, -0.00002690, 0.00001674, -0.00001923, 0.00001752, -0.00002093, 0.00002596, -0.00002749, 0.00002512, -0.00003074, 0.00003529, -0.00003562, 0.00002530, -0.00003469, 0.00003754, -0.00003797, 0.00004196, -0.00005443, 0.00005183, -0.00007416, 0.00006500, -0.00009034, 0.00010131, -0.00014524, 0.00011266, -0.00032571, -0.00007653, 0.00117726, -0.00074673, 0.00021927, -0.00022493, 0.00012156, -0.00010286, 0.00007972, -0.00008604, 0.00006554, -0.00006958, 0.00005182, -0.00005067, 0.00004460, -0.00003833, 0.00003568, -0.00004546, 0.00003912, -0.00003288, 0.00002969, -0.00003080, 0.00002905, -0.00002190, 0.00002007, -0.00002359, 0.00002155, -0.00002893, 0.00002528, -0.00001231, 0.00001325, -0.00001784, 0.00001140, -0.00000659, 0.00000088, 0.00000997, -0.00000989, 0.00000248, 0.00000749, -0.00003001, 0.00002763, -0.00002490, 0.00002586, -0.00001793, 0.00001069, -0.00001002, -0.00001041, 0.00002636, -0.00000663, -0.00000892, 0.00000005, 0.00001110, -0.00000781, 0.00000231, -0.00000948, 0.00000257, -0.00000005, 0.00001266, -0.00000741, -0.00000303, -0.00000493, 0.00001244, -0.00001071, 0.00000013, },
		{ 0.00000892, -0.00000389, 0.00000345, -0.00000301, 0.00000887, -0.00001553, 0.00001019, 0.00000069, -0.00000134, -0.00000789, 0.00000199, 0.00000906, -0.00000061, -0.00000469, 0.00000476, -0.00000326, -0.00000643, 0.00000881, 0.00000219, -0.00000806, -0.00000635, 0.00000975, 0.00001347, -0.00003766, 0.00003036, -0.00002716, 0.00003333, -0.00002109, 0.00002699, -0.00003209, 0.00002513, -0.00001647, 0.00001494, -0.00000253, -0.00001078, 0.00000578, -0.00000316, -0.00000187, -0.00000277, 0.00001019, -0.00000214, -0.00000938, 0.00000602, -0.00000805, 0.00000515, -0.00000298, 0.00000506, -0.00000932, 0.00000571, -0.00000639, 0.00001175, -0.00002255, 0.00001380, -0.00001647, 0.00001996, -0.00001924, 0.00001502, -0.00002530, 0.00001585, -0.00003930, 0.00003058, -0.00004746, 0.00005074, -0.00012972, 0.00007267, -0.00061399, 0.00129475, -0.00037827, -0.00011300, -0.00000546, -0.00005992, 0.00003348, -0.00003001, 0.00000932, -0.00002030, 0.00000870, -0.00002067, 0.00001374, -0.00001519, 0.00001506, -0.00001088, 0.00000049, -0.00000818, 0.00001150, -0.00001092, 0.00000799, -0.00000924, 0.00001398, -0.00001149, 0.00000619, -0.00000736, 0.00000128, -0.00000803, 0.00001842, -0.00001295, 0.00000897, -0.00001611, 0.00001853, -0.00002082, 0.00002568, -0.00001685, 0.00000348, 0.00000767, -0.00002275, 0.00000982, -0.00000298, -0.00000093, 0.00001144, -0.00001775, 0.00001861, -0.00003370, 0.00003277, -0.00000351, -0.00000427, -0.00001048, 0.00001692, -0.00000854, 0.00000399, -0.00001032, 0.00000110, -0.00000104, 0.00001142, -0.00000163, -0.00000601, -0.00000444, 0.00000796, -0.00000143, -0.00000932, },
		{ 0.00001156, -0.00000582, 0.00000665, -0.00000764, 0.00001328, -0.00001562, 0.00000637, 0.00000248, 0.00000154, -0.00000988, 0.00000203, 0.00000388, 0.00000882, -0.00001272, 0.00001086, -0.00000755, 0.00000095, -0.00000589, 0.00001871, -0.00002013, 0.00000838, -0.00001692, 0.00004252, -0.00005320, 0.00003326, -0.00002732, 0.00002933, -0.00001514, 0.00002262, -0.00002260, 0.00001283, -0.00000503, 0.00000472, 0.00000613, -0.00001167, -0.00000166, 0.00000681, -0.00001174, 0.00000779, -0.00000636, 0.00001946, -0.00002904, 0.00002352, -0.00002581, 0.00002377, -0.00002451, 0.00002927, -0.00003435, 0.00003231, -0.00003707, 0.00004502, -0.00005369, 0.00004510, -0.00005220, 0.00005739, -0.00005886, 0.00006131, -0.00007671, 0.00007632, -0.00010762, 0.00010750, -0.00013996, 0.00016529, -0.00025385, 0.00025748, -0.00072694, 0.00089055, 0.00030331, -0.00052248, 0.00021760, -0.00022839, 0.00014982, -0.00012354, 0.00009196, -0.00009489, 0.00007611, -0.00008161, 0.00006652, -0.00006365, 0.00005753, -0.00004897, 0.00004006, -0.00004829, 0.00004736, -0.00004320, 0.00003897, -0.00003935, 0.00004104, -0.00003483, 0.00002918, -0.00003049, 0.00002587, -0.00003393, 0.00003973, -0.00003005, 0.00002707, -0.00003342, 0.00003187, -0.00003022, 0.00002951, -0.00001492, 0.00000164, 0.00000397, -0.00000798, -0.00001348, 0.00002013, -0.00002413, 0.00003138, -0.00003172, 0.00002773, -0.00003317, 0.00001830, 0.00001036, -0.00000482, -0.00001318, 0.00001294, -0.00000168, -0.00000007, -0.00000550, -0.00000493, 0.00000190, 0.00000602, 0.00000630, -0.00000950, -0.00000220, 0.00000104, 0.00000820, -0.00001633, },
		{ 0.00000747, -0.00000227, 0.00000467, -0.00000616, 0.00001012, -0.00000811, -0.00000260, 0.00000708, 0.00000112, -0.00000759, -0.00000103, 0.00000162, 0.00001262, -0.00001273, 0.00000844, -0.00000467, 0.00000242, -0.00001214, 0.00002227, -0.00001821, 0.00000960, -0.00002405, 0.00004124, -0.00003375, 0.00000529, -0.00000202, 0.00000365, 0.00000853, 0.00000249, 0.00000064, -0.00000927, 0.00001316, -0.00001014, 0.00001708, -0.00001375, -0.00000521, 0.00000898, -0.00001200, 0.00000844, -0.00001130, 0.00002504, -0.00002910, 0.00002042, -0.00002270, 0.00002094, -0.00002326, 0.00002841, -0.00003207, 0.00002972, -0.00003565, 0.00004193, -0.00004551, 0.00003573, -0.00004440, 0.00004744, -0.00004801, 0.00005251, -0.00006657, 0.00006671, -0.00009413, 0.00008859, -0.00011714, 0.00013352, -0.00019171, 0.00017727, -0.00046664, 0.00020052, 0.00095784, -0.00071999, 0.00024808, -0.00024596, 0.00014328, -0.00011977, 0.00009311, -0.00009761, 0.00007825, -0.00008214, 0.00006388, -0.00006140, 0.00005367, -0.00004599, 0.00004178, -0.00005084, 0.00004647, -0.00004103, 0.00003746, -0.00003798, 0.00003767, -0.00003001, 0.00002634, -0.00002908, 0.00002609, -0.00003425, 0.00003507, -0.00002305, 0.00002221, -0.00002732, 0.00002221, -0.00001838, 0.00001482, 0.00000063, -0.00000785, 0.00000427, 0.00000151, -0.00002548, 0.00002613, -0.00002588, 0.00002638, -0.00001967, 0.00001120, -0.00000988, -0.00000965, 0.00002693, -0.00000653, -0.00000998, 0.00000238, 0.00000845, -0.00000594, 0.00000115, -0.00001133, 0.00000458, 0.00000096, 0.00001181, -0.00000979, -0.00000164, -0.00000390, 0.00001310, -0.00001639, },
		{ -0.00000257, 0.00000550, -0.00000139, 0.00000025, 0.00000142, 0.00000372, -0.00001309, 0.00001201, -0.00000127, -0.00000203, -0.00000674, 0.00000316, 0.00000935, -0.00000412, -0.00000196, 0.00000474, -0.00000291, -0.00000768, 0.00001145, -0.00000226, -0.00000336, -0.00001107, 0.00001450, 0.00000694, -0.00003355, 0.00002695, -0.00002217, 0.00002952, -0.00001451, 0.00001987, -0.00002518, 0.00002460, -0.00001851, 0.00002181, -0.00001182, -0.00000740, 0.00000532, -0.00000491, 0.00000099, -0.00000545, 0.00001561, -0.00001216, 0.00000089, -0.00000346, 0.00000137, -0.00000379, 0.00000720, -0.00000766, 0.00000357, -0.00000825, 0.00000995, -0.00000742, -0.00000369, -0.00000450, 0.00000288, -0.00000065, 0.00000343, -0.00001102, 0.00000519, -0.00002065, 0.00000138, -0.00001500, 0.00000680, -0.00002035, -0.00003799, -0.00007808, -0.00041161, 0.00124496, -0.00051344, 0.00001211, -0.00008548, 0.00000475, -0.00001626, 0.00001101, -0.00002676, 0.00001399, -0.00002227, 0.00000715, -0.00001066, 0.00000623, -0.00000419, 0.00000663, -0.00001654, 0.00001034, -0.00000650, 0.00000561, -0.00000743, 0.00000672, -0.00000018, 0.00000041, -0.00000514, 0.00000358, -0.00001110, 0.00000776, 0.00000400, -0.00000144, -0.00000264, -0.00000433, 0.00000754, -0.00001069, 0.00002258, -0.00002041, 0.00000847, 0.00000179, -0.00002122, 0.00001211, -0.00000691, 0.00000183, 0.00000805, -0.00001602, 0.00001726, -0.00003166, 0.00003195, -0.00000157, -0.00000697, -0.00000729, 0.00001447, -0.00000761, 0.00000418, -0.00001299, 0.00000266, 0.00000012, 0.00001172, -0.00000492, -0.00000451, -0.00000453, 0.00001115, -0.00000886, },
		{ -0.00001338, 0.00001234, -0.00000673, 0.00000654, -0.00000688, 0.00001312, -0.00001885, 0.00001246, -0.00000170, 0.00000261, -0.00001106, 0.00000493, 0.00000343, 0.00000684, -0.00001273, 0.00001313, -0.00000838, 0.00000039, -0.00000400, 0.00001572, -0.00001776, 0.00000643, -0.00001512, 0.00003991, -0.00005362, 0.00003365, -0.00002625, 0.00002952, -0.00001292, 0.00002153, -0.00002347, 0.00002008, -0.00001286, 0.00001455, -0.00000187, -0.00001201, 0.00000138, 0.00000210, -0.00000641, 0.00000233, 0.00000212, 0.00000790, -0.00001921, 0.00001519, -0.00001763, 0.00001574, -0.00001485, 0.00001762, -0.00002318, 0.00002020, -0.00002314, 0.00002984, -0.00004024, 0.00003245, -0.00003838, 0.00004235, -0.00004186, 0.00004144, -0.00005360, 0.00004970, -0.00007988, 0.00007760, -0.00010477, 0.00012054, -0.00020126, 0.00018123, -0.00067546, 0.00102934, 0.00005124, -0.00036664, 0.00013465, -0.00016563, 0.00010413, -0.00008592, 0.00005864, -0.00006378, 0.00004924, -0.00005836, 0.00004667, -0.00004625, 0.00004154, -0.00003384, 0.00002473, -0.00003179, 0.00003234, -0.00003005, 0.00002685, -0.00002712, 0.00003111, -0.00002668, 0.00002080, -0.00002186, 0.00001563, -0.00002184, 0.00003100, -0.00002452, 0.00002137, -0.00002833, 0.00002870, -0.00002937, 0.00003553, -0.00002440, 0.00000874, -0.00000070, -0.00000935, -0.00000899, 0.00001565, -0.00002163, 0.00002866, -0.00003074, 0.00002671, -0.00003152, 0.00001756, 0.00001254, -0.00000728, -0.00001103, 0.00001218, -0.00000249, 0.00000150, -0.00000842, -0.00000425, 0.00000332, 0.00000700, 0.00000318, -0.00000896, -0.00000228, 0.00000469, 0.00000222, },
		{ -0.00001858, 0.00001267, -0.00000636, 0.00000757, -0.00000905, 0.00001438, -0.00001534, 0.00000540, 0.00000268, 0.00000273, -0.00001008, 0.00000266, 0.00000072, 0.00001242, -0.00001547, 0.00001260, -0.00000656, 0.00000309, -0.00001218, 0.00002226, -0.00001966, 0.00001173, -0.00002650, 0.00004366, -0.00003922, 0.00000975, -0.00000526, 0.00000821, 0.00000595, 0.00000748, -0.00000674, 0.00000279, 0.00000358, -0.00000139, 0.00001309, -0.00001815, -0.00000033, 0.00000412, -0.00000769, 0.00000444, -0.00000516, 0.00001822, -0.00002593, 0.00001910, -0.00002151, 0.00001983, -0.00002080, 0.00002520, -0.00003040, 0.00002775, -0.00003314, 0.00004012, -0.00004672, 0.00003790, -0.00004572, 0.00004844, -0.00004831, 0.00005193, -0.00006570, 0.00006640, -0.00009704, 0.00009364, -0.00012278, 0.00013989, -0.00020576, 0.00019844, -0.00055424, 0.00045204, 0.00070344, -0.00063297, 0.00022995, -0.00023293, 0.00013827, -0.00011521, 0.00008804, -0.00009220, 0.00007533, -0.00008079, 0.00006378, -0.00006154, 0.00005330, -0.00004492, 0.00003956, -0.00004755, 0.00004494, -0.00004089, 0.00003759, -0.00003780, 0.00003943, -0.00003230, 0.00002732, -0.00002927, 0.00002487, -0.00003272, 0.00003792, -0.00002818, 0.00002638, -0.00003222, 0.00002902, -0.00002718, 0.00002815, -0.00001170, -0.00000139, 0.00000205, -0.00000199, -0.00002105, 0.00002368, -0.00002585, 0.00002613, -0.00002089, 0.00001179, -0.00000976, -0.00000924, 0.00002808, -0.00000735, -0.00000998, 0.00000380, 0.00000609, -0.00000371, -0.00000119, -0.00001193, 0.00000633, 0.00000210, 0.00000980, -0.00001093, -0.00000092, -0.00000145, 0.00001063, },
		{ -0.00001500, 0.00000479, 0.00000061, 0.00000235, -0.00000400, 0.00000712, -0.00000376, -0.00000696, 0.00001024, -0.00000114, -0.00000399, -0.00000400, 0.00000301, 0.00000991, -0.00000805, 0.00000225, 0.00000324, -0.00000189, -0.00000903, 0.00001368, -0.00000633, 0.00000121, -0.00001590, 0.00001966, -0.00000063, -0.00002808, 0.00002237, -0.00001631, 0.00002532, -0.00000725, 0.00001103, -0.00001425, 0.00001894, -0.00001516, 0.00002389, -0.00001987, -0.00000243, 0.00000231, -0.00000364, 0.00000062, -0.00000420, 0.00001592, -0.00001714, 0.00000720, -0.00000954, 0.00000749, -0.00000910, 0.00001303, -0.00001585, 0.00001191, -0.00001716, 0.00002091, -0.00002172, 0.00001109, -0.00001864, 0.00001762, -0.00001601, 0.00001985, -0.00002973, 0.00002750, -0.00004879, 0.00003392, -0.00005174, 0.00005063, -0.00007979, 0.00004124, -0.00022470, -0.00017154, 0.00110948, -0.00056797, 0.00008991, -0.00013650, 0.00004847, -0.00004995, 0.00003834, -0.00005039, 0.00003727, -0.00004406, 0.00002734, -0.00002859, 0.00002179, -0.00001782, 0.00001853, -0.00002751, 0.00002293, -0.00001944, 0.00001778, -0.00001881, 0.00001923, -0.00001176, 0.00000985, -0.00001369, 0.00001122, -0.00001959, 0.00002048, -0.00000913, 0.00000956, -0.00001404, 0.00000813, -0.00000541, 0.00000359, 0.00001318, -0.00001841, 0.00001011, -0.00000323, -0.00001828, 0.00001269, -0.00000964, 0.00000387, 0.00000505, -0.00001423, 0.00001622, -0.00003037, 0.00003206, -0.00000072, -0.00000876, -0.00000459, 0.00001180, -0.00000596, 0.00000327, -0.00001501, 0.00000463, 0.00000112, 0.00001121, -0.00000768, -0.00000302, -0.00000370, 0.00001234, },
		{ -0.00000500, -0.00000734, 0.00000979, -0.00000491, 0.00000387, -0.00000318, 0.00000912, -0.00001767, 0.00001522, -0.00000430, 0.00000259, -0.00001067, 0.00000662, 0.00000345, 0.00000374, -0.00001066, 0.00001373, -0.00000845, -0.00000065, -0.00000171, 0.00001186, -0.00001402, 0.00000274, -0.00001118, 0.00003463, -0.00005134, 0.00003171, -0.00002318, 0.00002789, -0.00000814, 0.00001671, -0.00001904, 0.00002235, -0.00001691, 0.00002253, -0.00001191, -0.00000854, 0.00000092, 0.00000048, -0.00000380, -0.00000026, 0.00000749, -0.00000167, -0.00000980, 0.00000673, -0.00000928, 0.00000779, -0.00000559, 0.00000600, -0.00001165, 0.00000772, -0.00000846, 0.00001278, -0.00002397, 0.00001702, -0.00002196, 0.00002484, -0.00002254, 0.00001906, -0.00002722, 0.00001821, -0.00004539, 0.00003968, -0.00006003, 0.00006273, -0.00012917, 0.00006840, -0.00054931, 0.00107349, -0.00014255, -0.00022601, 0.00005336, -0.00010196, 0.00005660, -0.00004693, 0.00002438, -0.00003131, 0.00002021, -0.00003253, 0.00002423, -0.00002665, 0.00002390, -0.00001757, 0.00000880, -0.00001444, 0.00001565, -0.00001488, 0.00001268, -0.00001250, 0.00001842, -0.00001621, 0.00001071, -0.00001197, 0.00000420, -0.00000725, 0.00001818, -0.00001469, 0.00001137, -0.00001832, 0.00001977, -0.00002166, 0.00003456, -0.00002965, 0.00001476, -0.00000665, -0.00000817, -0.00000623, 0.00001187, -0.00001926, 0.00002583, -0.00002948, 0.00002602, -0.00003057, 0.00001754, 0.00001404, -0.00000946, -0.00000869, 0.00001070, -0.00000249, 0.00000238, -0.00001153, -0.00000250, 0.00000424, 0.00000774, -0.00000036, -0.00000755, -0.00000237, 0.00000803, },
		{ 0.00000535, -0.00001670, 0.00001446, -0.00000800, 0.00000814, -0.00000925, 0.00001548, -0.00001970, 0.00001220, -0.00000201, 0.00000446, -0.00001202, 0.00000612, -0.00000024, 0.00001127, -0.00001659, 0.00001590, -0.00000830, 0.00000347, -0.00001157, 0.00002089, -0.00001924, 0.00001170, -0.00002653, 0.00004345, -0.00004238, 0.00001240, -0.00000692, 0.00001110, 0.00000593, 0.00000912, -0.00001027, 0.00001264, -0.00000579, 0.00000929, 0.00000421, -0.00001781, 0.00000162, 0.00000096, -0.00000422, 0.00000093, 0.00000105, 0.00000959, -0.00001949, 0.00001434, -0.00001701, 0.00001557, -0.00001513, 0.00001800, -0.00002416, 0.00002092, -0.00002496, 0.00003138, -0.00004032, 0.00003268, -0.00003964, 0.00004172, -0.00004036, 0.00004200, -0.00005376, 0.00005266, -0.00008421, 0.00008108, -0.00010808, 0.00012088, -0.00018659, 0.00017130, -0.00057190, 0.00064508, 0.00045181, -0.00051282, 0.00018046, -0.00019561, 0.00011349, -0.00009465, 0.00006929, -0.00007403, 0.00006021, -0.00006819, 0.00005360, -0.00005288, 0.00004521, -0.00003692, 0.00003094, -0.00003755, 0.00003621, -0.00003372, 0.00003108, -0.00003101, 0.00003479, -0.00002908, 0.00002348, -0.00002488, 0.00001875, -0.00002495, 0.00003365, -0.00002685, 0.00002459, -0.00003104, 0.00002974, -0.00003010, 0.00003717, -0.00002347, 0.00000724, -0.00000343, -0.00000254, -0.00001769, 0.00002105, -0.00002528, 0.00002531, -0.00002165, 0.00001263, -0.00001006, -0.00000861, 0.00002922, -0.00000862, -0.00000922, 0.00000430, 0.00000432, -0.00000154, -0.00000445, -0.00001111, 0.00000751, 0.00000339, 0.00000670, -0.00001071, -0.00000080, 0.00000198, },
		{ 0.00001067, -0.00001806, 0.00001057, -0.00000375, 0.00000554, -0.00000743, 0.00001203, -0.00001131, 0.00000094, 0.00000594, 0.00000060, -0.00000671, -0.00000022, 0.00000192, 0.00001029, -0.00001155, 0.00000701, 0.00000082, -0.00000051, -0.00001017, 0.00001502, -0.00000924, 0.00000463, -0.00001960, 0.00002384, -0.00000781, -0.00002250, 0.00001761, -0.00001058, 0.00002202, -0.00000105, 0.00000273, -0.00000154, 0.00000972, -0.00000693, 0.00001988, -0.00002405, 0.00000151, -0.00000106, -0.00000113, -0.00000144, -0.00000043, 0.00001208, -0.00001720, 0.00000909, -0.00001160, 0.00000974, -0.00001028, 0.00001392, -0.00001867, 0.00001467, -0.00001975, 0.00002472, -0.00002884, 0.00001929, -0.00002617, 0.00002532, -0.00002362, 0.00002732, -0.00003790, 0.00003752, -0.00006336, 0.00005209, -0.00007237, 0.00007505, -0.00011553, 0.00009022, -0.00033918, 0.00007023, 0.00091697, -0.00055287, 0.00012007, -0.00015527, 0.00006770, -0.00006432, 0.00004917, -0.00005890, 0.00004673, -0.00005368, 0.00003700, -0.00003746, 0.00002928, -0.00002392, 0.00002316, -0.00003084, 0.00002773, -0.00002521, 0.00002340, -0.00002390, 0.00002601, -0.00001862, 0.00001502, -0.00001787, 0.00001395, -0.00002217, 0.00002734, -0.00001781, 0.00001694, -0.00002205, 0.00001785, -0.00001670, 0.00001852, -0.00000016, -0.00001139, 0.00000752, -0.00000585, -0.00001531, 0.00001231, -0.00001149, 0.00000529, 0.00000238, -0.00001213, 0.00001503, -0.00002931, 0.00003266, -0.00000073, -0.00000966, -0.00000256, 0.00000918, -0.00000376, 0.00000105, -0.00001578, 0.00000637, 0.00000228, 0.00000953, -0.00000930, -0.00000202, -0.00000160, },
		{ 0.00000971, -0.00001142, -0.00000016, 0.00000536, -0.00000151, -0.00000020, 0.00000226, 0.00000235, -0.00001253, 0.00001424, -0.00000468, 0.00000111, -0.00000851, 0.00000687, 0.00000393, -0.00000010, -0.00000681, 0.00001261, -0.00000760, -0.00000211, 0.00000063, 0.00000781, -0.00000976, -0.00000177, -0.00000605, 0.00002750, -0.00004692, 0.00002779, -0.00001854, 0.00002517, -0.00000181, 0.00000925, -0.00000979, 0.00001908, -0.00001548, 0.00002565, -0.00002068, -0.00000355, -0.00000177, 0.00000138, -0.00000379, 0.00000031, 0.00000877, -0.00000745, -0.00000297, 0.00000025, -0.00000275, 0.00000193, 0.00000103, -0.00000300, -0.00000259, -0.00000195, 0.00000332, -0.00000212, -0.00000892, 0.00000277, -0.00000700, 0.00000921, -0.00000575, -0.00000017, -0.00000428, -0.00001035, -0.00001285, 0.00000344, -0.00001714, 0.00000595, -0.00005572, -0.00005791, -0.00036962, 0.00102805, -0.00026502, -0.00011909, -0.00001060, -0.00004988, 0.00001744, -0.00001514, -0.00000312, -0.00000472, -0.00000424, -0.00001017, 0.00000451, -0.00000954, 0.00000882, -0.00000408, -0.00000382, -0.00000032, 0.00000142, -0.00000161, 0.00000022, 0.00000096, 0.00000617, -0.00000624, 0.00000160, -0.00000362, -0.00000517, 0.00000614, 0.00000484, -0.00000367, -0.00000003, -0.00000607, 0.00000743, -0.00000866, 0.00002656, -0.00002898, 0.00001736, -0.00001176, -0.00000573, -0.00000484, 0.00000879, -0.00001713, 0.00002290, -0.00002774, 0.00002532, -0.00003006, 0.00001816, 0.00001475, -0.00001110, -0.00000645, 0.00000870, -0.00000155, 0.00000206, -0.00001402, -0.00000033, 0.00000500, 0.00000773, -0.00000356, -0.00000592, -0.00000177, },
		{ 0.00000534, -0.00000145, -0.00001153, 0.00001288, -0.00000696, 0.00000616, -0.00000646, 0.00001279, -0.00001989, 0.00001596, -0.00000537, 0.00000525, -0.00001254, 0.00000854, -0.00000079, 0.00000895, -0.00001561, 0.00001757, -0.00000924, 0.00000322, -0.00001031, 0.00001834, -0.00001710, 0.00000959, -0.00002407, 0.00004028, -0.00004262, 0.00001255, -0.00000650, 0.00001204, 0.00000873, 0.00000685, -0.00000856, 0.00001792, -0.00001191, 0.00001847, -0.00000693, -0.00001366, 0.00000040, 0.00000038, -0.00000284, -0.00000049, 0.00000515, 0.00000182, -0.00001234, 0.00000833, -0.00001120, 0.00001019, -0.00000850, 0.00000933, -0.00001597, 0.00001205, -0.00001421, 0.00001886, -0.00002917, 0.00002274, -0.00002895, 0.00003035, -0.00002768, 0.00002687, -0.00003550, 0.00003068, -0.00006097, 0.00005646, -0.00007925, 0.00008359, -0.00014211, 0.00010272, -0.00051736, 0.00075924, 0.00023472, -0.00038722, 0.00011852, -0.00014741, 0.00007924, -0.00006652, 0.00004429, -0.00004987, 0.00003888, -0.00004946, 0.00003768, -0.00003923, 0.00003292, -0.00002536, 0.00001936, -0.00002441, 0.00002367, -0.00002256, 0.00002068, -0.00002007, 0.00002579, -0.00002213, 0.00001660, -0.00001791, 0.00001008, -0.00001335, 0.00002410, -0.00002021, 0.00001764, -0.00002415, 0.00002415, -0.00002595, 0.00003934, -0.00003130, 0.00001513, -0.00001035, -0.00000071, -0.00001567, 0.00001864, -0.00002443, 0.00002401, -0.00002183, 0.00001359, -0.00001083, -0.00000749, 0.00002996, -0.00000993, -0.00000799, 0.00000392, 0.00000342, -0.00000003, -0.00000798, -0.00000919, 0.00000819, 0.00000453, 0.00000300, -0.00000947, -0.00000083, },
		{ 0.00000148, 0.00000622, -0.00001728, 0.00001331, -0.00000605, 0.00000676, -0.00000853, 0.00001427, -0.00001660, 0.00000823, 0.00000089, 0.00000285, -0.00000926, 0.00000370, 0.00000051, 0.00000993, -0.00001370, 0.00001123, -0.00000171, 0.00000072, -0.00001078, 0.00001512, -0.00001046, 0.00000615, -0.00002119, 0.00002577, -0.00001316, -0.00001819, 0.00001381, -0.00000600, 0.00002083, 0.00000244, -0.00000261, 0.00001004, -0.00000024, 0.00000383, 0.00001060, -0.00002334, 0.00000262, -0.00000289, 0.00000074, -0.00000327, 0.00000374, 0.00000615, -0.00001383, 0.00000749, -0.00001033, 0.00000881, -0.00000815, 0.00001074, -0.00001688, 0.00001257, -0.00001672, 0.00002173, -0.00002866, 0.00002052, -0.00002679, 0.00002579, -0.00002346, 0.00002596, -0.00003559, 0.00003491, -0.00006330, 0.00005412, -0.00007460, 0.00007685, -0.00012238, 0.00009899, -0.00040040, 0.00027933, 0.00070208, -0.00048947, 0.00011230, -0.00014671, 0.00006544, -0.00006157, 0.00004545, -0.00005401, 0.00004356, -0.00005178, 0.00003644, -0.00003743, 0.00002896, -0.00002293, 0.00002116, -0.00002725, 0.00002518, -0.00002389, 0.00002235, -0.00002241, 0.00002650, -0.00002017, 0.00001554, -0.00001755, 0.00001184, -0.00001862, 0.00002735, -0.00002044, 0.00001885, -0.00002454, 0.00002221, -0.00002308, 0.00003021, -0.00001387, -0.00000148, 0.00000133, -0.00000553, -0.00001325, 0.00001162, -0.00001276, 0.00000616, 0.00000009, -0.00000970, 0.00001339, -0.00002801, 0.00003330, -0.00000125, -0.00000981, -0.00000135, 0.00000699, -0.00000144, -0.00000221, -0.00001508, 0.00000753, 0.00000367, 0.00000665, -0.00000955, -0.00000159, },
		{ -0.00000009, 0.00000882, -0.00001522, 0.00000598, 0.00000094, 0.00000179, -0.00000394, 0.00000750, -0.00000513, -0.00000510, 0.00001038, -0.00000314, -0.00000148, -0.00000499, 0.00000594, 0.00000449, -0.00000387, -0.00000200, 0.00001023, -0.00000611, -0.00000368, 0.00000250, 0.00000439, -0.00000603, -0.00000582, -0.00000112, 0.00001992, -0.00004153, 0.00002285, -0.00001330, 0.00002262, 0.00000443, 0.00000122, 0.00000234, 0.00001145, -0.00000879, 0.00002243, -0.00002545, 0.00000034, -0.00000466, 0.00000326, -0.00000516, 0.00000300, 0.00000648, -0.00000912, 0.00000038, -0.00000321, 0.00000094, -0.00000088, 0.00000389, -0.00000799, 0.00000240, -0.00000696, 0.00000981, -0.00001195, 0.00000184, -0.00000729, 0.00000345, -0.00000140, 0.00000506, -0.00001220, 0.00001025, -0.00002986, 0.00001055, -0.00002285, 0.00001391, -0.00003688, 0.00000104, -0.00017151, -0.00016684, 0.00091290, -0.00031815, -0.00005312, -0.00004924, -0.00001638, -0.00000743, 0.00000445, -0.00001936, 0.00001156, -0.00001975, 0.00000462, -0.00000883, 0.00000192, -0.00000089, 0.00000412, -0.00001069, 0.00000787, -0.00000749, 0.00000688, -0.00000768, 0.00001040, -0.00000302, 0.00000098, -0.00000454, 0.00000117, -0.00001002, 0.00001517, -0.00000558, 0.00000522, -0.00000953, 0.00000498, -0.00000480, 0.00000639, 0.00001359, -0.00002261, 0.00001544, -0.00001442, -0.00000333, -0.00000420, 0.00000627, -0.00001527, 0.00001993, -0.00002536, 0.00002421, -0.00002956, 0.00001913, 0.00001478, -0.00001212, -0.00000458, 0.00000648, 0.00000016, 0.00000032, -0.00001523, 0.00000163, 0.00000598, 0.00000653, -0.00000575, -0.00000459, },
		{ -0.00000040, 0.00000737, -0.00000789, -0.00000481, 0.00000911, -0.00000409, 0.00000250, -0.00000158, 0.00000705, -0.00001600, 0.00001600, -0.00000651, 0.00000452, -0.00001126, 0.00000954, -0.00000080, 0.00000570, -0.00001264, 0.00001738, -0.00000914, 0.00000231, -0.00000864, 0.00001514, -0.00001392, 0.00000614, -0.00001983, 0.00003466, -0.00004017, 0.00001024, -0.00000418, 0.00001149, 0.00001370, 0.00000132, -0.00000158, 0.00001760, -0.00001278, 0.00002294, -0.00001691, -0.00000796, -0.00000284, 0.00000194, -0.00000356, 0.00000058, 0.00000606, -0.00000326, -0.00000659, 0.00000310, -0.00000599, 0.00000558, -0.00000305, 0.00000164, -0.00000848, 0.00000407, -0.00000427, 0.00000622, -0.00001687, 0.00001154, -0.00001718, 0.00001812, -0.00001451, 0.00001150, -0.00001682, 0.00000728, -0.00003480, 0.00002799, -0.00004577, 0.00003952, -0.00008662, 0.00000963, -0.00040290, 0.00079007, 0.00007225, -0.00027815, 0.00006102, -0.00010118, 0.00004575, -0.00003931, 0.00002059, -0.00002675, 0.00001782, -0.00003035, 0.00002105, -0.00002498, 0.00002042, -0.00001398, 0.00000858, -0.00001205, 0.00001123, -0.00001109, 0.00000985, -0.00000825, 0.00001534, -0.00001396, 0.00000910, -0.00001091, 0.00000186, -0.00000130, 0.00001246, -0.00001095, 0.00000793, -0.00001375, 0.00001400, -0.00001568, 0.00003408, -0.00003312, 0.00001972, -0.00001656, 0.00000232, -0.00001472, 0.00001653, -0.00002344, 0.00002225, -0.00002124, 0.00001432, -0.00001184, -0.00000592, 0.00003014, -0.00001102, -0.00000658, 0.00000281, 0.00000356, 0.00000027, -0.00001094, -0.00000679, 0.00000872, 0.00000500, -0.00000058, -0.00000776, },
		{ -0.00000143, 0.00000465, 0.00000007, -0.00001314, 0.00001276, -0.00000582, 0.00000553, -0.00000672, 0.00001293, -0.00001811, 0.00001291, -0.00000317, 0.00000442, -0.00001068, 0.00000689, -0.00000067, 0.00000849, -0.00001390, 0.00001403, -0.00000359, 0.00000137, -0.00001068, 0.00001396, -0.00000990, 0.00000555, -0.00002027, 0.00002481, -0.00001577, -0.00001615, 0.00001176, -0.00000316, 0.00002240, 0.00000224, -0.00000316, 0.00001783, -0.00000779, 0.00001364, -0.00000135, -0.00001859, 0.00000060, -0.00000233, 0.00000079, -0.00000342, 0.00000637, 0.00000043, -0.00000915, 0.00000411, -0.00000723, 0.00000617, -0.00000440, 0.00000533, -0.00001235, 0.00000765, -0.00001026, 0.00001410, -0.00002293, 0.00001632, -0.00002207, 0.00002082, -0.00001767, 0.00001835, -0.00002579, 0.00002285, -0.00005161, 0.00004290, -0.00006148, 0.00005928, -0.00010332, 0.00006811, -0.00039909, 0.00043035, 0.00049749, -0.00040298, 0.00008198, -0.00012097, 0.00004919, -0.00004773, 0.00003250, -0.00004057, 0.00003201, -0.00004189, 0.00002853, -0.00003101, 0.00002318, -0.00001713, 0.00001496, -0.00001927, 0.00001767, -0.00001756, 0.00001646, -0.00001587, 0.00002188, -0.00001736, 0.00001244, -0.00001404, 0.00000651, -0.00001058, 0.00002157, -0.00001740, 0.00001533, -0.00002117, 0.00002028, -0.00002273, 0.00003568, -0.00002437, 0.00000834, -0.00000672, -0.00000274, -0.00001240, 0.00001096, -0.00001365, 0.00000648, -0.00000163, -0.00000712, 0.00001130, -0.00002628, 0.00003367, -0.00000198, -0.00000940, -0.00000101, 0.00000560, 0.00000037, -0.00000589, -0.00001321, 0.00000815, 0.00000503, 0.00000297, -0.00000861, },
		{ -0.00000376, 0.00000251, 0.00000529, -0.00001549, 0.00000951, -0.00000187, 0.00000350, -0.00000576, 0.00001061, -0.00001101, 0.00000233, 0.00000528, -0.00000068, -0.00000433, -0.00000091, 0.00000438, 0.00000458, -0.00000667, 0.00000269, 0.00000742, -0.00000451, -0.00000497, 0.00000343, 0.00000230, -0.00000376, -0.00000826, 0.00000218, 0.00001346, -0.00003666, 0.00001809, -0.00000860, 0.00002161, 0.00000875, -0.00000487, 0.00001454, 0.00000209, 0.00000102, 0.00001347, -0.00002505, 0.00000133, -0.00000594, 0.00000440, -0.00000623, 0.00000607, 0.00000223, -0.00000766, 0.00000066, -0.00000383, 0.00000192, -0.00000086, 0.00000319, -0.00000899, 0.00000327, -0.00000715, 0.00001047, -0.00001564, 0.00000697, -0.00001187, 0.00000816, -0.00000582, 0.00000874, -0.00001567, 0.00001443, -0.00003748, 0.00002114, -0.00003476, 0.00002745, -0.00005779, 0.00002831, -0.00024965, 0.00002649, 0.00075694, -0.00031596, -0.00002474, -0.00006265, -0.00000219, -0.00001718, 0.00001123, -0.00002380, 0.00001675, -0.00002523, 0.00001058, -0.00001459, 0.00000675, -0.00000450, 0.00000654, -0.00001144, 0.00000956, -0.00001019, 0.00000957, -0.00000984, 0.00001445, -0.00000781, 0.00000439, -0.00000694, 0.00000181, -0.00000947, 0.00001821, -0.00001092, 0.00000964, -0.00001455, 0.00001183, -0.00001347, 0.00001958, -0.00000092, -0.00001248, 0.00000941, -0.00001395, -0.00000190, -0.00000378, 0.00000410, -0.00001369, 0.00001705, -0.00002239, 0.00002250, -0.00002873, 0.00002010, 0.00001432, -0.00001253, -0.00000332, 0.00000445, 0.00000221, -0.00000264, -0.00001492, 0.00000299, 0.00000728, 0.00000404, -0.00000658, },
		{ -0.00000614, 0.00000089, 0.00000716, -0.00001219, 0.00000140, 0.00000489, -0.00000094, -0.00000126, 0.00000361, 0.00000005, -0.00000941, 0.00001289, -0.00000549, 0.00000241, -0.00000836, 0.00000917, -0.00000053, 0.00000218, -0.00000838, 0.00001567, -0.00000817, 0.00000099, -0.00000704, 0.00001209, -0.00001072, 0.00000251, -0.00001509, 0.00002786, -0.00003606, 0.00000625, -0.00000075, 0.00001055, 0.00001939, -0.00000571, 0.00000917, 0.00001241, -0.00000826, 0.00002095, -0.00002284, -0.00000338, -0.00000613, 0.00000420, -0.00000533, 0.00000335, 0.00000404, -0.00000508, -0.00000335, -0.00000014, -0.00000256, 0.00000292, -0.00000012, -0.00000344, -0.00000350, -0.00000093, 0.00000228, -0.00000345, -0.00000661, 0.00000219, -0.00000747, 0.00000830, -0.00000447, 0.00000011, -0.00000285, -0.00001133, -0.00001287, 0.00000378, -0.00001731, 0.00000088, -0.00003652, -0.00008553, -0.00025192, 0.00074836, -0.00002952, -0.00019815, 0.00001909, -0.00006594, 0.00002040, -0.00001915, 0.00000364, -0.00000988, 0.00000204, -0.00001550, 0.00000782, -0.00001371, 0.00001090, -0.00000570, 0.00000142, -0.00000351, 0.00000206, -0.00000244, 0.00000171, 0.00000140, 0.00000618, -0.00000692, 0.00000308, -0.00000602, -0.00000333, 0.00000797, 0.00000215, -0.00000229, -0.00000137, -0.00000306, 0.00000246, -0.00000208, 0.00002296, -0.00002872, 0.00001961, -0.00002030, 0.00000523, -0.00001431, 0.00001463, -0.00002234, 0.00002010, -0.00001976, 0.00001449, -0.00001272, -0.00000411, 0.00002980, -0.00001175, -0.00000526, 0.00000125, 0.00000467, -0.00000093, -0.00001266, -0.00000457, 0.00000947, 0.00000432, -0.00000332, },
		{ -0.00000698, -0.00000102, 0.00000682, -0.00000609, -0.00000714, 0.00000983, -0.00000365, 0.00000246, -0.00000264, 0.00000840, -0.00001558, 0.00001403, -0.00000515, 0.00000456, -0.00001041, 0.00000878, -0.00000137, 0.00000607, -0.00001209, 0.00001501, -0.00000445, 0.00000127, -0.00000998, 0.00001192, -0.00000803, 0.00000331, -0.00001726, 0.00002115, -0.00001552, -0.00001664, 0.00001160, -0.00000194, 0.00002646, -0.00000151, 0.00000163, 0.00002036, -0.00001067, 0.00001913, -0.00001247, -0.00001208, -0.00000343, 0.00000021, -0.00000100, -0.00000147, 0.00000643, -0.00000336, -0.00000509, 0.00000074, -0.00000394, 0.00000347, -0.00000085, -0.00000022, -0.00000732, 0.00000234, -0.00000322, 0.00000486, -0.00001453, 0.00000935, -0.00001473, 0.00001333, -0.00000957, 0.00000842, -0.00001319, 0.00000669, -0.00003408, 0.00002459, -0.00004004, 0.00003067, -0.00006834, 0.00000831, -0.00033962, 0.00051141, 0.00032746, -0.00031614, 0.00004509, -0.00008963, 0.00002796, -0.00003016, 0.00001684, -0.00002466, 0.00001763, -0.00002891, 0.00001752, -0.00002186, 0.00001526, -0.00000972, 0.00000775, -0.00001031, 0.00000855, -0.00000934, 0.00000865, -0.00000702, 0.00001447, -0.00001216, 0.00000762, -0.00000941, 0.00000048, -0.00000093, 0.00001262, -0.00001079, 0.00000818, -0.00001347, 0.00001313, -0.00001591, 0.00003378, -0.00002915, 0.00001528, -0.00001441, 0.00000137, -0.00001252, 0.00001040, -0.00001420, 0.00000622, -0.00000253, -0.00000477, 0.00000899, -0.00002415, 0.00003362, -0.00000271, -0.00000863, -0.00000149, 0.00000528, 0.00000103, -0.00000915, -0.00001075, 0.00000858, 0.00000585, -0.00000084, },
		{ -0.00000610, -0.00000326, 0.00000533, -0.00000003, -0.00001244, 0.00000996, -0.00000239, 0.00000296, -0.00000488, 0.00001049, -0.00001359, 0.00000767, 0.00000074, 0.00000147, -0.00000645, 0.00000282, 0.00000277, 0.00000380, -0.00000780, 0.00000629, 0.00000496, -0.00000332, -0.00000566, 0.00000322, 0.00000183, -0.00000341, -0.00000845, 0.00000294, 0.00000928, -0.00003353, 0.00001454, -0.00000527, 0.00002309, 0.00000984, -0.00000690, 0.00002398, -0.00000588, 0.00001054, 0.00000133, -0.00002022, -0.00000100, -0.00000476, 0.00000372, -0.00000571, 0.00000781, -0.00000202, -0.00000478, -0.00000086, -0.00000262, 0.00000119, 0.00000087, 0.00000016, -0.00000718, 0.00000126, -0.00000390, 0.00000652, -0.00001395, 0.00000697, -0.00001148, 0.00000779, -0.00000495, 0.00000650, -0.00001201, 0.00000965, -0.00003422, 0.00001949, -0.00003268, 0.00002356, -0.00005585, 0.00002194, -0.00027857, 0.00018305, 0.00059075, -0.00027903, -0.00002334, -0.00005707, -0.00000294, -0.00001521, 0.00000828, -0.00001931, 0.00001331, -0.00002257, 0.00000915, -0.00001394, 0.00000609, -0.00000322, 0.00000458, -0.00000760, 0.00000611, -0.00000777, 0.00000732, -0.00000690, 0.00001343, -0.00000836, 0.00000430, -0.00000619, -0.00000079, -0.00000443, 0.00001559, -0.00001084, 0.00000895, -0.00001410, 0.00001295, -0.00001624, 0.00002762, -0.00001328, -0.00000152, 0.00000090, -0.00001073, -0.00000177, -0.00000327, 0.00000221, -0.00001249, 0.00001453, -0.00001909, 0.00002024, -0.00002742, 0.00002084, 0.00001360, -0.00001240, -0.00000282, 0.00000308, 0.00000395, -0.00000621, -0.00001331, 0.00000375, 0.00000869, 0.00000055, },
		{ -0.00000477, -0.00000488, 0.00000298, 0.00000451, -0.00001332, 0.00000548, 0.00000183, 0.00000102, -0.00000352, 0.00000725, -0.00000600, -0.00000226, 0.00000811, -0.00000317, -0.00000037, -0.00000453, 0.00000790, -0.00000046, -0.00000073, -0.00000384, 0.00001320, -0.00000683, -0.00000030, -0.00000604, 0.00000995, -0.00000850, -0.00000007, -0.00001130, 0.00002145, -0.00003174, 0.00000177, 0.00000265, 0.00001061, 0.00002398, -0.00001181, 0.00002111, 0.00000462, -0.00000017, 0.00001288, -0.00002332, -0.00000185, -0.00000768, 0.00000555, -0.00000663, 0.00000632, 0.00000038, -0.00000428, -0.00000254, -0.00000120, -0.00000114, 0.00000239, 0.00000001, -0.00000544, -0.00000160, -0.00000227, 0.00000431, -0.00000850, -0.00000027, -0.00000349, -0.00000161, 0.00000268, 0.00000060, -0.00000513, 0.00000359, -0.00002138, 0.00000014, -0.00001064, -0.00000070, -0.00002317, -0.00000543, -0.00016085, -0.00009285, 0.00065582, -0.00007707, -0.00014964, -0.00000340, -0.00004526, 0.00000626, -0.00000850, -0.00000444, -0.00000144, -0.00000616, -0.00000724, 0.00000012, -0.00000724, 0.00000586, -0.00000175, -0.00000097, -0.00000013, -0.00000226, 0.00000170, -0.00000196, 0.00000696, 0.00000012, -0.00000251, -0.00000026, -0.00000428, -0.00000418, 0.00001242, -0.00000430, 0.00000312, -0.00000744, 0.00000480, -0.00000698, 0.00001112, 0.00000911, -0.00001971, 0.00001491, -0.00002073, 0.00000708, -0.00001394, 0.00001281, -0.00002122, 0.00001774, -0.00001745, 0.00001392, -0.00001320, -0.00000233, 0.00002905, -0.00001204, -0.00000432, -0.00000034, 0.00000633, -0.00000351, -0.00001282, -0.00000296, 0.00001063, 0.00000227, },
		{ -0.00000404, -0.00000516, -0.00000001, 0.00000708, -0.00001058, -0.00000118, 0.00000609, -0.00000090, -0.00000104, 0.00000221, 0.00000221, -0.00001010, 0.00001187, -0.00000487, 0.00000317, -0.00000840, 0.00000927, -0.00000171, 0.00000318, -0.00000878, 0.00001436, -0.00000435, 0.00000060, -0.00000910, 0.00000967, -0.00000572, 0.00000044, -0.00001324, 0.00001582, -0.00001319, -0.00001915, 0.00001276, -0.00000146, 0.00003186, -0.00000742, 0.00001075, 0.00001785, -0.00000834, 0.00001828, -0.00001964, -0.00000656, -0.00000751, 0.00000334, -0.00000366, 0.00000191, 0.00000404, -0.00000457, -0.00000277, -0.00000144, -0.00000163, 0.00000187, 0.00000113, -0.00000427, -0.00000356, -0.00000136, 0.00000194, -0.00000309, -0.00000639, 0.00000240, -0.00000759, 0.00000628, -0.00000241, -0.00000000, -0.00000245, -0.00000793, -0.00001711, 0.00000637, -0.00001867, 0.00000146, -0.00003106, -0.00006275, -0.00023808, 0.00052511, 0.00020424, -0.00024461, 0.00001397, -0.00006219, 0.00000931, -0.00001507, 0.00000397, -0.00001147, 0.00000538, -0.00001737, 0.00000738, -0.00001346, 0.00000831, -0.00000353, 0.00000235, -0.00000336, 0.00000095, -0.00000227, 0.00000191, 0.00000126, 0.00000687, -0.00000674, 0.00000306, -0.00000569, -0.00000380, 0.00000729, 0.00000366, -0.00000353, 0.00000023, -0.00000427, 0.00000342, -0.00000483, 0.00002549, -0.00002755, 0.00001762, -0.00001983, 0.00000546, -0.00001312, 0.00000984, -0.00001445, 0.00000539, -0.00000246, -0.00000298, 0.00000682, -0.00002181, 0.00003316, -0.00000328, -0.00000776, -0.00000258, 0.00000603, 0.00000017, -0.00001123, -0.00000840, 0.00000921, 0.00000560, },
		{ -0.00000335, -0.00000468, -0.00000273, 0.00000739, -0.00000565, -0.00000734, 0.00000797, -0.00000098, 0.00000056, -0.00000165, 0.00000720, -0.00001237, 0.00000974, -0.00000197, 0.00000239, -0.00000710, 0.00000552, 0.00000147, 0.00000213, -0.00000704, 0.00000827, 0.00000335, -0.00000281, -0.00000575, 0.00000207, 0.00000274, -0.00000476, -0.00000650, 0.00000106, 0.00000781, -0.00003272, 0.00001262, -0.00000352, 0.00002718, 0.00000743, -0.00000387, 0.00002884, -0.00000999, 0.00001630, -0.00001043, -0.00001323, -0.00000553, -0.00000153, 0.00000119, -0.00000322, 0.00000728, -0.00000475, -0.00000213, -0.00000273, -0.00000088, 0.00000002, 0.00000286, -0.00000357, -0.00000427, -0.00000175, 0.00000063, 0.00000026, -0.00000894, 0.00000362, -0.00000793, 0.00000433, -0.00000110, 0.00000114, -0.00000460, -0.00000024, -0.00002408, 0.00000971, -0.00002115, 0.00000750, -0.00003703, -0.00001272, -0.00025640, 0.00028617, 0.00043986, -0.00022841, -0.00003554, -0.00004168, -0.00001170, -0.00000712, 0.00000056, -0.00001057, 0.00000551, -0.00001550, 0.00000347, -0.00000959, 0.00000242, 0.00000055, 0.00000070, -0.00000182, 0.00000015, -0.00000264, 0.00000237, -0.00000089, 0.00000898, -0.00000600, 0.00000202, -0.00000381, -0.00000469, 0.00000283, 0.00000926, -0.00000678, 0.00000428, -0.00000905, 0.00000871, -0.00001272, 0.00002883, -0.00002064, 0.00000731, -0.00000786, -0.00000584, -0.00000274, -0.00000263, 0.00000059, -0.00001178, 0.00001267, -0.00001586, 0.00001767, -0.00002568, 0.00002123, 0.00001275, -0.00001185, -0.00000314, 0.00000272, 0.00000469, -0.00000953, -0.00001100, 0.00000422, 0.00000973, },
		{ -0.00000097, -0.00000483, -0.00000397, 0.00000533, 0.00000009, -0.00001131, 0.00000677, 0.00000074, 0.00000099, -0.00000336, 0.00000807, -0.00000931, 0.00000335, 0.00000344, -0.00000078, -0.00000283, -0.00000063, 0.00000626, -0.00000099, -0.00000230, 0.00000001, 0.00001077, -0.00000567, -0.00000120, -0.00000597, 0.00000918, -0.00000787, -0.00000079, -0.00000956, 0.00001676, -0.00002858, -0.00000206, 0.00000500, 0.00001283, 0.00002596, -0.00001472, 0.00003142, -0.00000280, 0.00000820, 0.00000113, -0.00001895, -0.00000386, -0.00000657, 0.00000492, -0.00000627, 0.00000796, -0.00000325, -0.00000222, -0.00000330, -0.00000073, -0.00000115, 0.00000331, -0.00000192, -0.00000496, -0.00000217, -0.00000062, 0.00000230, -0.00000892, 0.00000182, -0.00000517, 0.00000019, 0.00000128, 0.00000086, -0.00000449, 0.00000256, -0.00002237, 0.00000313, -0.00001376, 0.00000192, -0.00002910, -0.00000041, -0.00020051, 0.00004778, 0.00053876, -0.00008520, -0.00012705, -0.00000892, -0.00003772, 0.00000228, -0.00000630, -0.00000475, -0.00000053, -0.00000726, -0.00000537, -0.00000211, -0.00000543, 0.00000504, -0.00000166, 0.00000075, -0.00000139, -0.00000199, 0.00000143, -0.00000106, 0.00000804, -0.00000233, -0.00000100, -0.00000095, -0.00000537, -0.00000098, 0.00001176, -0.00000599, 0.00000412, -0.00000880, 0.00000784, -0.00001168, 0.00002048, -0.00000379, -0.00000889, 0.00000704, -0.00001804, 0.00000744, -0.00001333, 0.00001103, -0.00002022, 0.00001546, -0.00001460, 0.00001268, -0.00001315, -0.00000076, 0.00002803, -0.00001189, -0.00000399, -0.00000146, 0.00000790, -0.00000689, -0.00001157, -0.00000202, 0.00001205, },
		{ 0.00000396, -0.00000607, -0.00000352, 0.00000163, 0.00000517, -0.00001231, 0.00000312, 0.00000307, 0.00000103, -0.00000344, 0.00000601, -0.00000360, -0.00000360, 0.00000772, -0.00000300, 0.00000079, -0.00000517, 0.00000865, -0.00000205, 0.00000060, -0.00000488, 0.00001273, -0.00000372, -0.00000025, -0.00000855, 0.00000800, -0.00000397, -0.00000187, -0.00000960, 0.00001030, -0.00001013, -0.00002259, 0.00001418, -0.00000042, 0.00003692, -0.00001329, 0.00002200, 0.00001217, -0.00000231, 0.00001119, -0.00002121, -0.00000411, -0.00000978, 0.00000545, -0.00000576, 0.00000533, 0.00000031, -0.00000364, -0.00000232, -0.00000207, -0.00000070, 0.00000175, 0.00000106, -0.00000612, -0.00000188, -0.00000248, 0.00000385, -0.00000786, -0.00000061, -0.00000251, -0.00000261, 0.00000163, 0.00000171, -0.00000446, 0.00000330, -0.00001696, -0.00000564, -0.00000606, -0.00000432, -0.00001927, -0.00000439, -0.00012560, -0.00011755, 0.00048584, 0.00012800, -0.00019521, -0.00000490, -0.00004393, -0.00000241, -0.00000595, -0.00000314, -0.00000393, -0.00000180, -0.00001013, 0.00000072, -0.00000805, 0.00000426, -0.00000025, 0.00000034, -0.00000019, -0.00000318, 0.00000162, -0.00000166, 0.00000677, 0.00000107, -0.00000277, 0.00000012, -0.00000412, -0.00000479, 0.00001185, -0.00000264, 0.00000167, -0.00000570, 0.00000339, -0.00000556, 0.00000714, 0.00001352, -0.00002073, 0.00001510, -0.00002191, 0.00000849, -0.00001367, 0.00000917, -0.00001447, 0.00000417, -0.00000147, -0.00000195, 0.00000504, -0.00001948, 0.00003236, -0.00000359, -0.00000706, -0.00000389, 0.00000752, -0.00000216, -0.00001176, -0.00000664, 0.00001030, },
		{ 0.00000984, -0.00000691, -0.00000278, -0.00000183, 0.00000790, -0.00001003, -0.00000190, 0.00000490, 0.00000122, -0.00000252, 0.00000265, 0.00000200, -0.00000810, 0.00000857, -0.00000244, 0.00000176, -0.00000606, 0.00000693, 0.00000047, -0.00000007, -0.00000475, 0.00000869, 0.00000265, -0.00000289, -0.00000554, 0.00000056, 0.00000431, -0.00000699, -0.00000328, -0.00000271, 0.00000853, -0.00003397, 0.00001200, -0.00000273, 0.00003307, 0.00000244, 0.00000374, 0.00002887, -0.00000943, 0.00001604, -0.00001850, -0.00000693, -0.00001030, 0.00000238, -0.00000224, 0.00000061, 0.00000459, -0.00000533, -0.00000075, -0.00000387, 0.00000036, -0.00000055, 0.00000388, -0.00000655, -0.00000180, -0.00000402, 0.00000426, -0.00000580, -0.00000306, -0.00000081, -0.00000347, 0.00000014, 0.00000310, -0.00000419, 0.00000268, -0.00001055, -0.00001234, -0.00000241, -0.00000690, -0.00001249, -0.00001179, -0.00006282, -0.00019271, 0.00033206, 0.00032030, -0.00018068, -0.00004949, -0.00002519, -0.00002165, 0.00000154, -0.00000704, -0.00000219, -0.00000226, -0.00000802, -0.00000296, -0.00000457, -0.00000153, 0.00000427, -0.00000257, 0.00000316, -0.00000551, 0.00000248, -0.00000261, 0.00000564, 0.00000329, -0.00000253, -0.00000081, -0.00000153, -0.00000772, 0.00000961, 0.00000202, -0.00000123, -0.00000200, -0.00000169, 0.00000118, -0.00000445, 0.00002357, -0.00002184, 0.00001195, -0.00001482, -0.00000064, -0.00000430, -0.00000199, -0.00000073, -0.00001158, 0.00001166, -0.00001305, 0.00001514, -0.00002370, 0.00002129, 0.00001192, -0.00001107, -0.00000414, 0.00000346, 0.00000396, -0.00001183, -0.00000867, 0.00000484, },
		{ 0.00001329, -0.00000489, -0.00000371, -0.00000318, 0.00000714, -0.00000484, -0.00000717, 0.00000568, 0.00000148, -0.00000084, -0.00000089, 0.00000589, -0.00000915, 0.00000608, 0.00000027, 0.00000067, -0.00000415, 0.00000260, 0.00000467, -0.00000225, -0.00000222, 0.00000253, 0.00000893, -0.00000503, -0.00000158, -0.00000676, 0.00000972, -0.00000885, 0.00000049, -0.00001023, 0.00001444, -0.00002738, -0.00000457, 0.00000587, 0.00001765, 0.00002467, -0.00001307, 0.00003805, -0.00000733, 0.00001339, -0.00001072, -0.00001195, -0.00000841, -0.00000319, 0.00000229, -0.00000389, 0.00000744, -0.00000551, -0.00000029, -0.00000443, 0.00000027, -0.00000161, 0.00000457, -0.00000471, -0.00000323, -0.00000386, 0.00000244, -0.00000213, -0.00000599, 0.00000068, -0.00000388, -0.00000093, 0.00000273, -0.00000190, -0.00000016, -0.00000346, -0.00001676, -0.00000157, -0.00000804, -0.00000672, -0.00001965, -0.00002016, -0.00019842, 0.00015109, 0.00042156, -0.00007161, -0.00012048, -0.00000380, -0.00003865, 0.00000475, -0.00000918, -0.00000054, -0.00000419, -0.00000383, -0.00000781, -0.00000062, -0.00000665, 0.00000679, -0.00000367, 0.00000466, -0.00000541, 0.00000119, -0.00000171, 0.00000310, 0.00000559, -0.00000187, -0.00000166, 0.00000003, -0.00000789, 0.00000459, 0.00000734, -0.00000374, 0.00000123, -0.00000571, 0.00000581, -0.00001067, 0.00002390, -0.00001270, 0.00000075, -0.00000183, -0.00001321, 0.00000645, -0.00001245, 0.00000933, -0.00001946, 0.00001358, -0.00001162, 0.00001104, -0.00001264, 0.00000052, 0.00002681, -0.00001133, -0.00000441, -0.00000169, 0.00000866, -0.00001025, -0.00000946, -0.00000147, },
		{ 0.00001179, 0.00000099, -0.00000668, -0.00000217, 0.00000350, 0.00000160, -0.00001116, 0.00000499, 0.00000168, 0.00000126, -0.00000373, 0.00000744, -0.00000723, 0.00000187, 0.00000327, -0.00000069, -0.00000167, -0.00000150, 0.00000741, -0.00000278, -0.00000095, -0.00000134, 0.00001086, -0.00000308, -0.00000086, -0.00000872, 0.00000746, -0.00000352, -0.00000270, -0.00000754, 0.00000598, -0.00000772, -0.00002578, 0.00001479, 0.00000244, 0.00004002, -0.00001687, 0.00003271, 0.00000598, 0.00000439, 0.00000009, -0.00001757, -0.00000543, -0.00000924, 0.00000545, -0.00000611, 0.00000741, -0.00000328, -0.00000169, -0.00000308, -0.00000153, -0.00000081, 0.00000272, -0.00000071, -0.00000596, -0.00000210, -0.00000125, 0.00000246, -0.00000891, 0.00000200, -0.00000458, -0.00000049, -0.00000002, 0.00000233, -0.00000449, 0.00000332, -0.00001901, -0.00000173, -0.00001009, -0.00000035, -0.00002664, 0.00000336, -0.00016432, -0.00000193, 0.00041453, 0.00008977, -0.00016678, -0.00001098, -0.00003561, -0.00000652, -0.00000319, -0.00000431, -0.00000225, -0.00000343, -0.00000786, -0.00000178, -0.00000614, 0.00000346, 0.00000001, 0.00000168, -0.00000089, -0.00000352, 0.00000188, -0.00000141, 0.00000861, -0.00000196, -0.00000092, -0.00000085, -0.00000477, -0.00000238, 0.00001206, -0.00000500, 0.00000331, -0.00000783, 0.00000725, -0.00001102, 0.00001659, 0.00000132, -0.00001120, 0.00000884, -0.00002060, 0.00000993, -0.00001388, 0.00000836, -0.00001438, 0.00000283, 0.00000017, -0.00000160, 0.00000377, -0.00001731, 0.00003129, -0.00000356, -0.00000682, -0.00000490, 0.00000912, -0.00000548, -0.00001080, -0.00000560, },
		{ 0.00000593, 0.00000860, -0.00000949, -0.00000082, -0.00000055, 0.00000655, -0.00001207, 0.00000232, 0.00000219, 0.00000289, -0.00000484, 0.00000635, -0.00000329, -0.00000249, 0.00000520, -0.00000117, -0.00000005, -0.00000363, 0.00000717, -0.00000045, -0.00000212, -0.00000170, 0.00000803, 0.00000252, -0.00000321, -0.00000552, -0.00000059, 0.00000560, -0.00000901, -0.00000006, -0.00000703, 0.00001030, -0.00003635, 0.00001179, -0.00000175, 0.00003929, -0.00000324, 0.00001417, 0.00002546, -0.00000537, 0.00000962, -0.00002097, -0.00000363, -0.00001332, 0.00000531, -0.00000514, 0.00000447, 0.00000073, -0.00000410, -0.00000083, -0.00000389, 0.00000065, -0.00000009, 0.00000339, -0.00000797, -0.00000066, -0.00000456, 0.00000560, -0.00000979, 0.00000166, -0.00000442, 0.00000005, -0.00000292, 0.00000565, -0.00000716, 0.00000685, -0.00001747, -0.00000360, -0.00001158, 0.00000375, -0.00002827, 0.00000857, -0.00011213, -0.00010503, 0.00032855, 0.00023730, -0.00014525, -0.00005760, -0.00001344, -0.00002812, 0.00000705, -0.00001133, 0.00000274, -0.00000693, -0.00000306, -0.00000750, -0.00000115, -0.00000380, 0.00000621, -0.00000355, 0.00000549, -0.00000883, 0.00000552, -0.00000548, 0.00001050, -0.00000164, 0.00000044, -0.00000286, -0.00000059, -0.00000835, 0.00001372, -0.00000356, 0.00000323, -0.00000724, 0.00000517, -0.00000670, 0.00000565, 0.00001402, -0.00001757, 0.00001177, -0.00001864, 0.00000371, -0.00000593, -0.00000149, -0.00000177, -0.00001179, 0.00001149, -0.00001088, 0.00001291, -0.00002168, 0.00002106, 0.00001120, -0.00001032, -0.00000550, 0.00000506, 0.00000172, -0.00001263, -0.00000688, },
		{ -0.00000086, 0.00001378, -0.00000878, -0.00000182, -0.00000229, 0.00000770, -0.00000887, -0.00000243, 0.00000335, 0.00000309, -0.00000348, 0.00000283, 0.00000167, -0.00000600, 0.00000572, -0.00000081, 0.00000068, -0.00000393, 0.00000474, 0.00000323, -0.00000397, -0.00000071, 0.00000367, 0.00000784, -0.00000491, -0.00000167, -0.00000798, 0.00001101, -0.00001083, 0.00000313, -0.00001279, 0.00001426, -0.00002811, -0.00000587, 0.00000558, 0.00002460, 0.00002063, -0.00000689, 0.00004039, -0.00000796, 0.00001301, -0.00001920, -0.00000526, -0.00001349, 0.00000109, -0.00000142, -0.00000010, 0.00000486, -0.00000583, 0.00000058, -0.00000503, 0.00000095, -0.00000169, 0.00000518, -0.00000713, -0.00000150, -0.00000529, 0.00000513, -0.00000695, -0.00000165, -0.00000203, -0.00000125, -0.00000323, 0.00000504, -0.00000528, 0.00000482, -0.00001078, -0.00000863, -0.00000958, 0.00000150, -0.00002055, -0.00000236, -0.00005605, -0.00015895, 0.00020900, 0.00032173, -0.00005193, -0.00011981, 0.00000481, -0.00004253, 0.00000921, -0.00001321, 0.00000443, -0.00000880, 0.00000089, -0.00001175, 0.00000215, -0.00000870, 0.00000907, -0.00000570, 0.00000845, -0.00000982, 0.00000499, -0.00000550, 0.00000840, 0.00000134, 0.00000014, -0.00000326, 0.00000133, -0.00001004, 0.00001024, 0.00000150, 0.00000043, -0.00000367, 0.00000009, 0.00000018, -0.00000487, 0.00002120, -0.00001603, 0.00000690, -0.00000956, -0.00000759, 0.00000458, -0.00001141, 0.00000780, -0.00001904, 0.00001233, -0.00000887, 0.00000932, -0.00001186, 0.00000155, 0.00002549, -0.00001047, -0.00000553, -0.00000083, 0.00000808, -0.00001276, -0.00000718, },
		{ -0.00000506, 0.00001330, -0.00000294, -0.00000593, -0.00000105, 0.00000495, -0.00000257, -0.00000793, 0.00000465, 0.00000193, -0.00000018, -0.00000185, 0.00000612, -0.00000780, 0.00000490, -0.00000003, 0.00000098, -0.00000334, 0.00000187, 0.00000594, -0.00000403, -0.00000110, 0.00000121, 0.00000934, -0.00000280, -0.00000106, -0.00000966, 0.00000812, -0.00000453, -0.00000173, -0.00000759, 0.00000370, -0.00000693, -0.00002790, 0.00001393, 0.00000781, 0.00004027, -0.00001656, 0.00004079, 0.00000170, 0.00000837, -0.00001150, -0.00001087, -0.00000961, -0.00000617, 0.00000324, -0.00000436, 0.00000734, -0.00000554, 0.00000009, -0.00000413, -0.00000059, -0.00000126, 0.00000397, -0.00000328, -0.00000463, -0.00000332, 0.00000125, -0.00000116, -0.00000694, 0.00000183, -0.00000421, -0.00000075, 0.00000063, 0.00000045, -0.00000140, -0.00000095, -0.00001537, -0.00000438, -0.00000667, -0.00000585, -0.00002112, -0.00000924, -0.00017051, 0.00008978, 0.00033277, 0.00007600, -0.00015306, -0.00000794, -0.00003461, -0.00000510, -0.00000480, -0.00000158, -0.00000456, -0.00000106, -0.00000938, -0.00000106, -0.00000682, 0.00000492, -0.00000165, 0.00000509, -0.00000423, -0.00000113, -0.00000057, 0.00000192, 0.00000715, -0.00000240, -0.00000095, -0.00000037, -0.00000671, 0.00000228, 0.00000874, -0.00000374, 0.00000141, -0.00000590, 0.00000660, -0.00001160, 0.00002117, -0.00000797, -0.00000189, 0.00000084, -0.00001670, 0.00000978, -0.00001366, 0.00000744, -0.00001431, 0.00000168, 0.00000209, -0.00000170, 0.00000292, -0.00001534, 0.00002998, -0.00000317, -0.00000722, -0.00000517, 0.00001014, -0.00000901, -0.00000881, },
		{ -0.00000583, 0.00000753, 0.00000593, -0.00001069, 0.00000097, 0.00000085, 0.00000373, -0.00001157, 0.00000450, 0.00000076, 0.00000315, -0.00000536, 0.00000811, -0.00000691, 0.00000254, 0.00000120, 0.00000085, -0.00000226, -0.00000046, 0.00000659, -0.00000162, -0.00000336, 0.00000124, 0.00000699, 0.00000247, -0.00000339, -0.00000609, -0.00000080, 0.00000584, -0.00000986, 0.00000197, -0.00001053, 0.00001173, -0.00003871, 0.00001088, 0.00000067, 0.00004426, -0.00000747, 0.00002505, 0.00002093, -0.00000057, -0.00000094, -0.00001802, -0.00000418, -0.00001348, 0.00000610, -0.00000631, 0.00000703, -0.00000294, -0.00000204, -0.00000186, -0.00000305, 0.00000019, 0.00000119, 0.00000151, -0.00000780, -0.00000090, -0.00000332, 0.00000433, -0.00001091, 0.00000417, -0.00000621, 0.00000173, -0.00000404, 0.00000581, -0.00000694, 0.00000677, -0.00001918, -0.00000035, -0.00001478, 0.00000700, -0.00003453, 0.00001565, -0.00014592, -0.00001366, 0.00029115, 0.00018695, -0.00012429, -0.00005739, -0.00000850, -0.00002942, 0.00000818, -0.00001142, 0.00000337, -0.00000746, -0.00000177, -0.00000904, -0.00000022, -0.00000367, 0.00000584, -0.00000187, 0.00000467, -0.00000913, 0.00000566, -0.00000526, 0.00001246, -0.00000462, 0.00000202, -0.00000356, -0.00000130, -0.00000620, 0.00001421, -0.00000600, 0.00000492, -0.00000952, 0.00000920, -0.00001220, 0.00001441, 0.00000332, -0.00000995, 0.00000752, -0.00001901, 0.00000651, -0.00000725, -0.00000120, -0.00000260, -0.00001217, 0.00001195, -0.00000931, 0.00001108, -0.00001974, 0.00002059, 0.00001069, -0.00000987, -0.00000673, 0.00000699, -0.00000168, -0.00001187, },
		{ -0.00000534, 0.00000017, 0.00001303, -0.00001195, 0.00000043, -0.00000128, 0.00000675, -0.00001099, 0.00000149, 0.00000110, 0.00000440, -0.00000562, 0.00000631, -0.00000297, -0.00000131, 0.00000310, -0.00000006, -0.00000058, -0.00000230, 0.00000577, 0.00000182, -0.00000558, 0.00000159, 0.00000379, 0.00000723, -0.00000501, -0.00000188, -0.00000894, 0.00001220, -0.00001282, 0.00000602, -0.00001610, 0.00001524, -0.00003003, -0.00000668, 0.00000506, 0.00003248, 0.00001536, 0.00000252, 0.00003940, -0.00000561, 0.00000673, -0.00002221, -0.00000133, -0.00001703, 0.00000458, -0.00000477, 0.00000383, 0.00000116, -0.00000453, 0.00000023, -0.00000475, 0.00000092, -0.00000098, 0.00000457, -0.00000843, -0.00000059, -0.00000553, 0.00000618, -0.00001045, 0.00000233, -0.00000469, 0.00000119, -0.00000515, 0.00000653, -0.00000727, 0.00000788, -0.00001606, -0.00000199, -0.00001636, 0.00000948, -0.00003271, 0.00001346, -0.00009519, -0.00009466, 0.00022416, 0.00024772, -0.00003650, -0.00011759, 0.00001140, -0.00004502, 0.00001223, -0.00001545, 0.00000733, -0.00001154, 0.00000418, -0.00001475, 0.00000416, -0.00000978, 0.00001022, -0.00000614, 0.00001036, -0.00001269, 0.00000746, -0.00000793, 0.00001280, -0.00000290, 0.00000229, -0.00000463, 0.00000183, -0.00001043, 0.00001394, -0.00000341, 0.00000420, -0.00000824, 0.00000613, -0.00000656, 0.00000335, 0.00001397, -0.00001399, 0.00000855, -0.00001460, -0.00000242, 0.00000240, -0.00001040, 0.00000645, -0.00001889, 0.00001174, -0.00000659, 0.00000778, -0.00001101, 0.00000238, 0.00002415, -0.00000952, -0.00000710, 0.00000097, 0.00000598, -0.00001388, },
		{ -0.00000648, -0.00000461, 0.00001429, -0.00000718, -0.00000411, -0.00000007, 0.00000543, -0.00000611, -0.00000394, 0.00000304, 0.00000305, -0.00000247, 0.00000137, 0.00000277, -0.00000554, 0.00000507, -0.00000145, 0.00000140, -0.00000370, 0.00000443, 0.00000441, -0.00000561, 0.00000001, 0.00000259, 0.00000836, -0.00000294, -0.00000099, -0.00001104, 0.00000958, -0.00000658, 0.00000060, -0.00000947, 0.00000342, -0.00000794, -0.00002884, 0.00001168, 0.00001551, 0.00003784, -0.00001202, 0.00004537, 0.00000039, 0.00000722, -0.00002002, -0.00000406, -0.00001467, -0.00000191, -0.00000027, -0.00000107, 0.00000525, -0.00000596, 0.00000089, -0.00000467, 0.00000004, -0.00000137, 0.00000468, -0.00000569, -0.00000308, -0.00000448, 0.00000362, -0.00000541, -0.00000337, 0.00000001, -0.00000248, -0.00000220, 0.00000220, -0.00000217, 0.00000255, -0.00000673, -0.00000906, -0.00001043, 0.00000066, -0.00001660, -0.00000782, -0.00003684, -0.00014483, 0.00014716, 0.00025772, 0.00007320, -0.00014620, -0.00000113, -0.00003682, -0.00000143, -0.00000786, 0.00000220, -0.00000808, 0.00000279, -0.00001252, 0.00000105, -0.00000843, 0.00000705, -0.00000358, 0.00000870, -0.00000829, 0.00000218, -0.00000394, 0.00000671, 0.00000365, -0.00000115, -0.00000199, 0.00000056, -0.00000853, 0.00000728, 0.00000379, -0.00000042, -0.00000263, -0.00000117, 0.00000234, -0.00000766, 0.00002024, -0.00001247, 0.00000476, -0.00000676, -0.00001148, 0.00000845, -0.00001312, 0.00000650, -0.00001436, 0.00000094, 0.00000391, -0.00000193, 0.00000230, -0.00001351, 0.00002848, -0.00000247, -0.00000831, -0.00000442, 0.00000997, -0.00001193, },
		{ -0.00000996, -0.00000539, 0.00000935, 0.00000206, -0.00001042, 0.00000250, 0.00000187, 0.00000034, -0.00000915, 0.00000471, 0.00000072, 0.00000192, -0.00000388, 0.00000743, -0.00000789, 0.00000555, -0.00000198, 0.00000253, -0.00000401, 0.00000273, 0.00000556, -0.00000315, -0.00000344, 0.00000344, 0.00000611, 0.00000214, -0.00000321, -0.00000739, 0.00000015, 0.00000473, -0.00000911, 0.00000215, -0.00001229, 0.00001179, -0.00004013, 0.00000852, 0.00000535, 0.00004696, -0.00000861, 0.00003435, 0.00001745, 0.00000175, -0.00001218, -0.00001166, -0.00000785, -0.00001092, 0.00000455, -0.00000535, 0.00000751, -0.00000531, -0.00000021, -0.00000304, -0.00000194, -0.00000049, 0.00000266, -0.00000111, -0.00000658, -0.00000195, -0.00000103, 0.00000112, -0.00000942, 0.00000444, -0.00000615, 0.00000164, -0.00000345, 0.00000407, -0.00000426, 0.00000322, -0.00001621, -0.00000245, -0.00001198, 0.00000257, -0.00003038, 0.00000650, -0.00015492, 0.00006366, 0.00023782, 0.00015982, -0.00011463, -0.00005047, -0.00000926, -0.00002644, 0.00000591, -0.00000844, 0.00000077, -0.00000471, -0.00000361, -0.00000798, -0.00000137, -0.00000166, 0.00000383, 0.00000165, 0.00000144, -0.00000697, 0.00000338, -0.00000224, 0.00001148, -0.00000542, 0.00000212, -0.00000313, -0.00000306, -0.00000207, 0.00001153, -0.00000525, 0.00000353, -0.00000826, 0.00000937, -0.00001370, 0.00001936, -0.00000550, -0.00000171, 0.00000093, -0.00001647, 0.00000762, -0.00000809, -0.00000111, -0.00000332, -0.00001247, 0.00001269, -0.00000813, 0.00000958, -0.00001791, 0.00001988, 0.00001046, -0.00000994, -0.00000739, 0.00000856, -0.00000551, },
		{ -0.00001345, -0.00000423, 0.00000178, 0.00001082, -0.00001388, 0.00000273, -0.00000036, 0.00000461, -0.00001095, 0.00000376, -0.00000019, 0.00000456, -0.00000625, 0.00000823, -0.00000638, 0.00000314, -0.00000039, 0.00000162, -0.00000247, 0.00000027, 0.00000592, 0.00000021, -0.00000653, 0.00000393, 0.00000347, 0.00000669, -0.00000497, -0.00000264, -0.00000907, 0.00001253, -0.00001388, 0.00000801, -0.00001886, 0.00001610, -0.00003207, -0.00000802, 0.00000550, 0.00003988, 0.00001077, 0.00001317, 0.00003698, -0.00000280, -0.00000359, -0.00001970, -0.00000126, -0.00001777, 0.00000601, -0.00000653, 0.00000659, -0.00000237, -0.00000251, -0.00000091, -0.00000377, 0.00000030, 0.00000040, 0.00000278, -0.00000844, -0.00000069, -0.00000443, 0.00000515, -0.00001172, 0.00000484, -0.00000627, 0.00000251, -0.00000586, 0.00000641, -0.00000697, 0.00000777, -0.00001744, 0.00000067, -0.00001891, 0.00001216, -0.00003808, 0.00002001, -0.00012465, -0.00002207, 0.00020722, 0.00019945, -0.00002949, -0.00011043, 0.00001335, -0.00004404, 0.00001226, -0.00001473, 0.00000704, -0.00001117, 0.00000472, -0.00001558, 0.00000436, -0.00000903, 0.00000956, -0.00000442, 0.00000966, -0.00001312, 0.00000763, -0.00000786, 0.00001496, -0.00000585, 0.00000365, -0.00000515, 0.00000114, -0.00000857, 0.00001465, -0.00000584, 0.00000587, -0.00001057, 0.00001017, -0.00001186, 0.00001114, 0.00000494, -0.00000825, 0.00000606, -0.00001634, 0.00000144, 0.00000035, -0.00000953, 0.00000528, -0.00001884, 0.00001164, -0.00000475, 0.00000653, -0.00001020, 0.00000302, 0.00002289, -0.00000872, -0.00000871, 0.00000328, 0.00000263, },
		{ -0.00001360, -0.00000445, -0.00000381, 0.00001423, -0.00001103, -0.00000158, 0.00000073, 0.00000475, -0.00000813, -0.00000047, 0.00000130, 0.00000401, -0.00000445, 0.00000461, -0.00000123, -0.00000168, 0.00000301, -0.00000106, 0.00000056, -0.00000269, 0.00000604, 0.00000279, -0.00000707, 0.00000188, 0.00000308, 0.00000774, -0.00000325, -0.00000103, -0.00001225, 0.00001109, -0.00000882, 0.00000335, -0.00001221, 0.00000435, -0.00001020, -0.00002913, 0.00000875, 0.00002466, 0.00003387, -0.00000414, 0.00004708, 0.00000126, 0.00000053, -0.00002321, 0.00000025, -0.00001848, 0.00000181, -0.00000366, 0.00000253, 0.00000202, -0.00000485, 0.00000057, -0.00000443, 0.00000005, -0.00000078, 0.00000434, -0.00000725, -0.00000202, -0.00000481, 0.00000477, -0.00000883, 0.00000032, -0.00000218, -0.00000059, -0.00000364, 0.00000337, -0.00000392, 0.00000523, -0.00001124, -0.00000341, -0.00001616, 0.00000750, -0.00002702, 0.00000581, -0.00006954, -0.00009577, 0.00016922, 0.00019935, 0.00007141, -0.00013969, 0.00000474, -0.00003857, 0.00000158, -0.00000988, 0.00000461, -0.00001036, 0.00000578, -0.00001523, 0.00000278, -0.00000941, 0.00000840, -0.00000433, 0.00001092, -0.00001134, 0.00000467, -0.00000645, 0.00001112, -0.00000033, 0.00000063, -0.00000314, 0.00000104, -0.00000900, 0.00001083, -0.00000071, 0.00000294, -0.00000679, 0.00000429, -0.00000349, -0.00000101, 0.00001488, -0.00001196, 0.00000742, -0.00001225, -0.00000622, 0.00000648, -0.00001244, 0.00000561, -0.00001451, 0.00000065, 0.00000542, -0.00000204, 0.00000174, -0.00001179, 0.00002684, -0.00000159, -0.00000990, -0.00000269, 0.00000837, },
		{ -0.00000898, -0.00000751, -0.00000519, 0.00001091, -0.00000247, -0.00000879, 0.00000370, 0.00000213, -0.00000259, -0.00000593, 0.00000372, 0.00000148, -0.00000015, -0.00000106, 0.00000480, -0.00000629, 0.00000598, -0.00000350, 0.00000316, -0.00000472, 0.00000542, 0.00000426, -0.00000488, -0.00000241, 0.00000469, 0.00000560, 0.00000147, -0.00000275, -0.00000915, 0.00000193, 0.00000256, -0.00000701, 0.00000060, -0.00001215, 0.00001014, -0.00004033, 0.00000456, 0.00001240, 0.00004730, -0.00000607, 0.00004118, 0.00001596, -0.00000073, -0.00002052, -0.00000486, -0.00001272, -0.00000692, 0.00000150, -0.00000275, 0.00000599, -0.00000592, 0.00000067, -0.00000373, -0.00000116, -0.00000081, 0.00000366, -0.00000369, -0.00000503, -0.00000305, 0.00000124, -0.00000284, -0.00000630, 0.00000315, -0.00000491, 0.00000053, -0.00000207, 0.00000166, -0.00000077, -0.00000177, -0.00001076, -0.00000768, -0.00000559, -0.00000666, -0.00001910, -0.00001590, -0.00013746, 0.00011571, 0.00018421, 0.00014510, -0.00011071, -0.00004055, -0.00001285, -0.00002144, 0.00000231, -0.00000451, -0.00000296, -0.00000055, -0.00000698, -0.00000563, -0.00000338, 0.00000102, 0.00000145, 0.00000556, -0.00000269, -0.00000377, 0.00000006, 0.00000237, 0.00000839, -0.00000456, 0.00000126, -0.00000225, -0.00000478, 0.00000251, 0.00000718, -0.00000246, 0.00000005, -0.00000426, 0.00000612, -0.00001113, 0.00001956, -0.00001039, 0.00000467, -0.00000594, -0.00001215, 0.00000732, -0.00000851, -0.00000116, -0.00000401, -0.00001250, 0.00001340, -0.00000705, 0.00000823, -0.00001612, 0.00001892, 0.00001052, -0.00001062, -0.00000714, 0.00000917, },
		{ -0.00000138, -0.00001175, -0.00000382, 0.00000379, 0.00000724, -0.00001429, 0.00000482, 0.00000013, 0.00000203, -0.00000925, 0.00000436, -0.00000035, 0.00000342, -0.00000508, 0.00000803, -0.00000757, 0.00000594, -0.00000335, 0.00000323, -0.00000420, 0.00000315, 0.00000544, -0.00000165, -0.00000647, 0.00000571, 0.00000323, 0.00000588, -0.00000454, -0.00000411, -0.00000808, 0.00001161, -0.00001351, 0.00000837, -0.00002010, 0.00001576, -0.00003327, -0.00001072, 0.00000776, 0.00004574, 0.00000844, 0.00002327, 0.00003496, -0.00000256, -0.00001454, -0.00001359, -0.00000444, -0.00001573, 0.00000509, -0.00000621, 0.00000739, -0.00000467, -0.00000073, -0.00000214, -0.00000262, -0.00000048, 0.00000197, 0.00000027, -0.00000748, -0.00000144, -0.00000247, 0.00000242, -0.00001077, 0.00000559, -0.00000654, 0.00000257, -0.00000534, 0.00000487, -0.00000474, 0.00000486, -0.00001500, -0.00000100, -0.00001663, 0.00000867, -0.00003505, 0.00001365, -0.00013532, 0.00004313, 0.00017249, 0.00017104, -0.00003001, -0.00009858, 0.00001085, -0.00003972, 0.00000964, -0.00001160, 0.00000413, -0.00000805, 0.00000266, -0.00001428, 0.00000287, -0.00000668, 0.00000745, -0.00000102, 0.00000676, -0.00001138, 0.00000566, -0.00000525, 0.00001452, -0.00000704, 0.00000391, -0.00000483, -0.00000036, -0.00000500, 0.00001257, -0.00000552, 0.00000491, -0.00000985, 0.00001098, -0.00001399, 0.00001610, -0.00000305, -0.00000128, 0.00000085, -0.00001506, 0.00000368, -0.00000132, -0.00000886, 0.00000425, -0.00001871, 0.00001174, -0.00000318, 0.00000550, -0.00000943, 0.00000348, 0.00002179, -0.00000831, -0.00000992, 0.00000546, },
		{ 0.00000547, -0.00001380, -0.00000308, -0.00000243, 0.00001275, -0.00001386, 0.00000129, 0.00000117, 0.00000322, -0.00000840, 0.00000172, 0.00000024, 0.00000404, -0.00000517, 0.00000664, -0.00000436, 0.00000220, -0.00000007, 0.00000039, -0.00000108, -0.00000066, 0.00000676, 0.00000095, -0.00000791, 0.00000383, 0.00000318, 0.00000710, -0.00000336, -0.00000158, -0.00001271, 0.00001190, -0.00001034, 0.00000542, -0.00001460, 0.00000534, -0.00001271, -0.00002969, 0.00000615, 0.00003404, 0.00002999, 0.00000549, 0.00004744, 0.00000193, -0.00000988, -0.00002090, 0.00000081, -0.00001964, 0.00000365, -0.00000567, 0.00000514, -0.00000112, -0.00000306, -0.00000046, -0.00000358, -0.00000047, 0.00000043, 0.00000293, -0.00000772, -0.00000168, -0.00000412, 0.00000423, -0.00001052, 0.00000310, -0.00000384, 0.00000067, -0.00000432, 0.00000342, -0.00000398, 0.00000550, -0.00001275, -0.00000074, -0.00001881, 0.00001047, -0.00003252, 0.00001292, -0.00009639, -0.00003640, 0.00016273, 0.00016014, 0.00006561, -0.00013002, 0.00000708, -0.00003779, 0.00000241, -0.00000967, 0.00000450, -0.00001014, 0.00000660, -0.00001626, 0.00000310, -0.00000889, 0.00000824, -0.00000325, 0.00001093, -0.00001243, 0.00000530, -0.00000691, 0.00001382, -0.00000357, 0.00000208, -0.00000380, 0.00000069, -0.00000767, 0.00001191, -0.00000328, 0.00000473, -0.00000926, 0.00000841, -0.00000860, 0.00000584, 0.00000736, -0.00000770, 0.00000610, -0.00001476, -0.00000188, 0.00000437, -0.00001177, 0.00000480, -0.00001463, 0.00000069, 0.00000663, -0.00000194, 0.00000113, -0.00001013, 0.00002514, -0.00000074, -0.00001167, -0.00000034, },
		{ 0.00000849, -0.00001144, -0.00000516, -0.00000478, 0.00001166, -0.00000698, -0.00000609, 0.00000443, 0.00000156, -0.00000432, -0.00000303, 0.00000251, 0.00000214, -0.00000202, 0.00000195, 0.00000140, -0.00000311, 0.00000427, -0.00000332, 0.00000259, -0.00000419, 0.00000735, 0.00000270, -0.00000643, -0.00000072, 0.00000521, 0.00000532, 0.00000068, -0.00000236, -0.00001084, 0.00000391, 0.00000005, -0.00000437, -0.00000190, -0.00001074, 0.00000717, -0.00003966, -0.00000054, 0.00002118, 0.00004612, -0.00000041, 0.00004592, 0.00001562, -0.00000840, -0.00002363, -0.00000031, -0.00001661, -0.00000319, -0.00000167, 0.00000030, 0.00000332, -0.00000507, 0.00000050, -0.00000366, -0.00000098, -0.00000049, 0.00000375, -0.00000563, -0.00000370, -0.00000357, 0.00000260, -0.00000629, -0.00000275, 0.00000127, -0.00000333, -0.00000074, -0.00000089, -0.00000016, 0.00000185, -0.00000591, -0.00000554, -0.00001304, 0.00000087, -0.00001627, -0.00000654, -0.00004407, -0.00009897, 0.00013926, 0.00014054, 0.00013396, -0.00010730, -0.00003140, -0.00001632, -0.00001674, -0.00000062, -0.00000159, -0.00000578, 0.00000305, -0.00001016, -0.00000348, -0.00000492, 0.00000313, -0.00000005, 0.00000843, -0.00000621, -0.00000103, -0.00000275, 0.00000701, 0.00000445, -0.00000292, 0.00000010, -0.00000160, -0.00000549, 0.00000601, 0.00000298, 0.00000064, -0.00000384, 0.00000077, 0.00000105, -0.00000582, 0.00001564, -0.00001079, 0.00000768, -0.00001131, -0.00000731, 0.00000609, -0.00000864, -0.00000129, -0.00000463, -0.00001223, 0.00001388, -0.00000585, 0.00000686, -0.00001431, 0.00001775, 0.00001079, -0.00001182, -0.00000596, },
		{ 0.00000694, -0.00000528, -0.00000932, -0.00000380, 0.00000579, 0.00000268, -0.00001313, 0.00000647, -0.00000003, -0.00000014, -0.00000685, 0.00000394, 0.00000013, 0.00000153, -0.00000263, 0.00000604, -0.00000647, 0.00000647, -0.00000490, 0.00000402, -0.00000516, 0.00000581, 0.00000449, -0.00000361, -0.00000545, 0.00000672, 0.00000325, 0.00000474, -0.00000380, -0.00000611, -0.00000620, 0.00000962, -0.00001184, 0.00000709, -0.00001955, 0.00001378, -0.00003322, -0.00001510, 0.00001210, 0.00004986, 0.00000904, 0.00003201, 0.00003404, -0.00000706, -0.00002254, -0.00000682, -0.00000903, -0.00001207, 0.00000256, -0.00000425, 0.00000629, -0.00000533, 0.00000015, -0.00000288, -0.00000180, -0.00000093, 0.00000315, -0.00000234, -0.00000608, -0.00000230, -0.00000045, -0.00000109, -0.00000825, 0.00000495, -0.00000583, 0.00000180, -0.00000420, 0.00000279, -0.00000181, 0.00000070, -0.00001035, -0.00000546, -0.00001118, 0.00000088, -0.00002567, -0.00000419, -0.00012418, 0.00008992, 0.00013351, 0.00015429, -0.00003444, -0.00008455, 0.00000584, -0.00003361, 0.00000582, -0.00000761, 0.00000018, -0.00000364, -0.00000083, -0.00001180, 0.00000058, -0.00000363, 0.00000486, 0.00000293, 0.00000281, -0.00000852, 0.00000258, -0.00000098, 0.00001197, -0.00000668, 0.00000332, -0.00000411, -0.00000185, -0.00000096, 0.00000890, -0.00000328, 0.00000197, -0.00000655, 0.00000867, -0.00001260, 0.00001708, -0.00000791, 0.00000456, -0.00000519, -0.00001167, 0.00000439, -0.00000255, -0.00000838, 0.00000333, -0.00001837, 0.00001178, -0.00000164, 0.00000452, -0.00000864, 0.00000371, 0.00002089, -0.00000839, -0.00001041, },
		{ 0.00000211, 0.00000177, -0.00001267, -0.00000245, -0.00000054, 0.00000982, -0.00001524, 0.00000413, 0.00000100, 0.00000150, -0.00000737, 0.00000260, 0.00000004, 0.00000297, -0.00000431, 0.00000687, -0.00000566, 0.00000475, -0.00000286, 0.00000200, -0.00000281, 0.00000190, 0.00000675, -0.00000110, -0.00000785, 0.00000533, 0.00000332, 0.00000616, -0.00000307, -0.00000285, -0.00001212, 0.00001159, -0.00001063, 0.00000610, -0.00001569, 0.00000532, -0.00001455, -0.00003134, 0.00000472, 0.00004271, 0.00002762, 0.00001543, 0.00004793, -0.00000045, -0.00002062, -0.00001491, -0.00000192, -0.00001804, 0.00000323, -0.00000576, 0.00000598, -0.00000314, -0.00000150, -0.00000155, -0.00000258, -0.00000115, 0.00000186, 0.00000078, -0.00000729, -0.00000188, -0.00000272, 0.00000222, -0.00001032, 0.00000453, -0.00000460, 0.00000107, -0.00000416, 0.00000242, -0.00000252, 0.00000351, -0.00001112, -0.00000161, -0.00001754, 0.00000839, -0.00003116, 0.00000971, -0.00010892, 0.00001984, 0.00013866, 0.00013693, 0.00005524, -0.00011677, 0.00000558, -0.00003421, 0.00000104, -0.00000748, 0.00000216, -0.00000757, 0.00000517, -0.00001547, 0.00000197, -0.00000691, 0.00000673, -0.00000059, 0.00000891, -0.00001158, 0.00000405, -0.00000508, 0.00001424, -0.00000543, 0.00000277, -0.00000389, -0.00000026, -0.00000490, 0.00001060, -0.00000349, 0.00000429, -0.00000916, 0.00000989, -0.00001132, 0.00001069, 0.00000023, -0.00000179, 0.00000189, -0.00001431, 0.00000101, 0.00000246, -0.00001122, 0.00000409, -0.00001460, 0.00000081, 0.00000768, -0.00000170, 0.00000045, -0.00000856, 0.00002347, -0.00000012, -0.00001322, },
		{ -0.00000409, 0.00000656, -0.00001256, -0.00000336, -0.00000388, 0.00001115, -0.00001080, -0.00000268, 0.00000445, 0.00000051, -0.00000472, -0.00000108, 0.00000176, 0.00000203, -0.00000282, 0.00000404, -0.00000141, 0.00000024, 0.00000149, -0.00000198, 0.00000114, -0.00000260, 0.00000847, 0.00000086, -0.00000738, 0.00000106, 0.00000541, 0.00000494, 0.00000011, -0.00000244, -0.00001190, 0.00000538, -0.00000197, -0.00000217, -0.00000426, -0.00000910, 0.00000373, -0.00003889, -0.00000593, 0.00003071, 0.00004476, 0.00000716, 0.00004984, 0.00001410, -0.00001934, -0.00002125, 0.00000062, -0.00001804, -0.00000113, -0.00000369, 0.00000257, 0.00000068, -0.00000356, -0.00000033, -0.00000303, -0.00000132, 0.00000043, 0.00000284, -0.00000668, -0.00000286, -0.00000333, 0.00000258, -0.00000844, 0.00000030, -0.00000047, -0.00000206, -0.00000155, -0.00000049, -0.00000071, 0.00000259, -0.00000770, -0.00000259, -0.00001612, 0.00000443, -0.00002225, 0.00000142, -0.00006893, -0.00004958, 0.00013817, 0.00011072, 0.00012135, -0.00010124, -0.00002535, -0.00001779, -0.00001374, -0.00000182, -0.00000071, -0.00000655, 0.00000487, -0.00001200, -0.00000245, -0.00000519, 0.00000397, 0.00000000, 0.00000947, -0.00000818, 0.00000026, -0.00000391, 0.00001038, 0.00000080, -0.00000125, -0.00000086, -0.00000149, -0.00000474, 0.00000749, 0.00000027, 0.00000257, -0.00000650, 0.00000498, -0.00000387, 0.00000014, 0.00000941, -0.00000756, 0.00000701, -0.00001414, -0.00000297, 0.00000444, -0.00000866, -0.00000144, -0.00000511, -0.00001177, 0.00001414, -0.00000444, 0.00000535, -0.00001244, 0.00001640, 0.00001111, -0.00001327, },
		{ -0.00001007, 0.00000735, -0.00000845, -0.00000678, -0.00000375, 0.00000726, -0.00000226, -0.00001055, 0.00000738, -0.00000076, -0.00000132, -0.00000469, 0.00000331, 0.00000049, -0.00000020, 0.00000010, 0.00000327, -0.00000390, 0.00000502, -0.00000485, 0.00000370, -0.00000504, 0.00000794, 0.00000312, -0.00000532, -0.00000386, 0.00000715, 0.00000338, 0.00000349, -0.00000307, -0.00000815, -0.00000398, 0.00000716, -0.00000954, 0.00000481, -0.00001769, 0.00001041, -0.00003210, -0.00002085, 0.00001808, 0.00005285, 0.00001223, 0.00003976, 0.00003317, -0.00001651, -0.00002522, -0.00000216, -0.00001285, -0.00000850, -0.00000026, -0.00000174, 0.00000409, -0.00000463, 0.00000005, -0.00000291, -0.00000154, -0.00000079, 0.00000356, -0.00000453, -0.00000466, -0.00000281, 0.00000090, -0.00000437, -0.00000506, 0.00000357, -0.00000473, 0.00000082, -0.00000316, 0.00000111, 0.00000051, -0.00000290, -0.00000568, -0.00001031, -0.00000532, -0.00000770, -0.00001447, -0.00002794, -0.00009451, 0.00011387, 0.00009993, 0.00014188, -0.00003877, -0.00007125, 0.00000066, -0.00002751, 0.00000235, -0.00000435, -0.00000314, 0.00000046, -0.00000431, -0.00000936, -0.00000141, -0.00000094, 0.00000287, 0.00000619, -0.00000088, -0.00000581, -0.00000028, 0.00000365, 0.00000829, -0.00000535, 0.00000231, -0.00000346, -0.00000256, 0.00000225, 0.00000517, -0.00000058, -0.00000153, -0.00000206, 0.00000443, -0.00000859, 0.00001436, -0.00000887, 0.00000763, -0.00001031, -0.00000733, 0.00000396, -0.00000344, -0.00000804, 0.00000259, -0.00001780, 0.00001161, 0.00000009, 0.00000342, -0.00000775, 0.00000371, 0.00002016, -0.00000893, },
		{ -0.00001422, 0.00000412, -0.00000208, -0.00001067, -0.00000233, 0.00000156, 0.00000566, -0.00001494, 0.00000652, 0.00000016, 0.00000030, -0.00000587, 0.00000273, 0.00000026, 0.00000129, -0.00000228, 0.00000553, -0.00000500, 0.00000534, -0.00000446, 0.00000299, -0.00000403, 0.00000453, 0.00000611, -0.00000316, -0.00000697, 0.00000623, 0.00000364, 0.00000489, -0.00000240, -0.00000469, -0.00001060, 0.00001024, -0.00000971, 0.00000528, -0.00001515, 0.00000380, -0.00001524, -0.00003445, 0.00000472, 0.00005045, 0.00002735, 0.00002508, 0.00004884, -0.00000791, -0.00002805, -0.00000824, -0.00000615, -0.00001479, 0.00000120, -0.00000429, 0.00000509, -0.00000367, -0.00000075, -0.00000220, -0.00000186, -0.00000158, 0.00000304, -0.00000163, -0.00000628, -0.00000224, -0.00000121, -0.00000059, -0.00000863, 0.00000475, -0.00000461, 0.00000083, -0.00000351, 0.00000097, -0.00000047, 0.00000044, -0.00000755, -0.00000493, -0.00001348, 0.00000252, -0.00002418, -0.00000339, -0.00010353, 0.00006274, 0.00010838, 0.00012378, 0.00004253, -0.00010161, 0.00000157, -0.00002886, -0.00000153, -0.00000445, -0.00000118, -0.00000373, 0.00000237, -0.00001356, 0.00000005, -0.00000418, 0.00000461, 0.00000275, 0.00000574, -0.00000958, 0.00000167, -0.00000156, 0.00001259, -0.00000589, 0.00000276, -0.00000365, -0.00000119, -0.00000172, 0.00000782, -0.00000199, 0.00000206, -0.00000671, 0.00000861, -0.00001107, 0.00001232, -0.00000450, 0.00000356, -0.00000356, -0.00001156, 0.00000239, 0.00000090, -0.00001084, 0.00000354, -0.00001436, 0.00000081, 0.00000879, -0.00000147, -0.00000021, -0.00000712, 0.00002187, 0.00000014, },
		{ -0.00001464, -0.00000198, 0.00000376, -0.00001239, -0.00000225, -0.00000242, 0.00000920, -0.00001336, 0.00000099, 0.00000358, -0.00000036, -0.00000423, -0.00000006, 0.00000160, 0.00000101, -0.00000226, 0.00000464, -0.00000274, 0.00000254, -0.00000120, -0.00000031, -0.00000062, -0.00000030, 0.00000881, -0.00000119, -0.00000750, 0.00000248, 0.00000565, 0.00000421, -0.00000003, -0.00000320, -0.00001200, 0.00000590, -0.00000299, -0.00000111, -0.00000557, -0.00000821, 0.00000068, -0.00003874, -0.00001095, 0.00004023, 0.00004439, 0.00001555, 0.00005407, 0.00000854, -0.00003005, -0.00001524, -0.00000168, -0.00001685, -0.00000118, -0.00000398, 0.00000327, -0.00000098, -0.00000225, -0.00000124, -0.00000223, -0.00000185, 0.00000164, 0.00000116, -0.00000686, -0.00000241, -0.00000255, 0.00000132, -0.00000900, 0.00000240, -0.00000168, -0.00000135, -0.00000179, -0.00000086, -0.00000009, 0.00000155, -0.00000692, -0.00000254, -0.00001602, 0.00000388, -0.00002265, 0.00000114, -0.00008290, -0.00000060, 0.00012069, 0.00009347, 0.00010601, -0.00009186, -0.00002291, -0.00001681, -0.00001261, -0.00000136, -0.00000174, -0.00000526, 0.00000473, -0.00001229, -0.00000270, -0.00000413, 0.00000353, 0.00000151, 0.00000867, -0.00000849, -0.00000008, -0.00000304, 0.00001181, -0.00000187, 0.00000002, -0.00000151, -0.00000174, -0.00000288, 0.00000694, -0.00000047, 0.00000265, -0.00000701, 0.00000710, -0.00000705, 0.00000478, 0.00000311, -0.00000245, 0.00000342, -0.00001416, 0.00000019, 0.00000277, -0.00000872, -0.00000153, -0.00000536, -0.00001133, 0.00001431, -0.00000290, 0.00000372, -0.00001052, 0.00001496, 0.00001131, },
		{ -0.00000973, -0.00000907, 0.00000665, -0.00001049, -0.00000460, -0.00000335, 0.00000772, -0.00000690, -0.00000683, 0.00000722, -0.00000158, -0.00000154, -0.00000324, 0.00000302, 0.00000015, -0.00000099, 0.00000210, 0.00000090, -0.00000109, 0.00000249, -0.00000363, 0.00000247, -0.00000389, 0.00000938, 0.00000135, -0.00000639, -0.00000221, 0.00000736, 0.00000333, 0.00000248, -0.00000273, -0.00000970, -0.00000208, 0.00000498, -0.00000748, 0.00000251, -0.00001543, 0.00000635, -0.00003053, -0.00002738, 0.00002496, 0.00005582, 0.00001706, 0.00004765, 0.00002980, -0.00002878, -0.00002235, -0.00000105, -0.00001438, -0.00000640, -0.00000210, 0.00000010, 0.00000189, -0.00000332, -0.00000065, -0.00000242, -0.00000177, -0.00000009, 0.00000308, -0.00000602, -0.00000347, -0.00000283, 0.00000119, -0.00000671, -0.00000201, 0.00000205, -0.00000370, 0.00000007, -0.00000263, 0.00000036, 0.00000140, -0.00000467, -0.00000275, -0.00001347, -0.00000163, -0.00001363, -0.00000644, -0.00005010, -0.00005413, 0.00011686, 0.00007627, 0.00012917, -0.00004039, -0.00006071, -0.00000306, -0.00002261, 0.00000014, -0.00000269, -0.00000480, 0.00000315, -0.00000678, -0.00000779, -0.00000235, 0.00000071, 0.00000212, 0.00000799, -0.00000346, -0.00000416, -0.00000186, 0.00000744, 0.00000446, -0.00000367, 0.00000124, -0.00000309, -0.00000217, 0.00000382, 0.00000259, 0.00000126, -0.00000416, 0.00000202, -0.00000006, -0.00000366, 0.00000938, -0.00000645, 0.00000739, -0.00001333, -0.00000309, 0.00000287, -0.00000412, -0.00000779, 0.00000209, -0.00001716, 0.00001124, 0.00000205, 0.00000209, -0.00000671, 0.00000351, 0.00001950, },
		{ 0.00000006, -0.00001458, 0.00000532, -0.00000546, -0.00000835, -0.00000236, 0.00000343, 0.00000075, -0.00001290, 0.00000797, -0.00000108, -0.00000001, -0.00000466, 0.00000278, 0.00000024, -0.00000018, 0.00000001, 0.00000348, -0.00000305, 0.00000415, -0.00000463, 0.00000305, -0.00000439, 0.00000689, 0.00000485, -0.00000487, -0.00000563, 0.00000671, 0.00000397, 0.00000351, -0.00000169, -0.00000666, -0.00000866, 0.00000838, -0.00000817, 0.00000351, -0.00001335, 0.00000090, -0.00001484, -0.00003886, 0.00000586, 0.00005779, 0.00002891, 0.00003500, 0.00004875, -0.00002037, -0.00002979, -0.00000371, -0.00000967, -0.00001155, -0.00000116, -0.00000230, 0.00000322, -0.00000299, -0.00000087, -0.00000222, -0.00000165, -0.00000151, 0.00000362, -0.00000386, -0.00000503, -0.00000247, -0.00000016, -0.00000340, -0.00000612, 0.00000418, -0.00000421, 0.00000036, -0.00000288, -0.00000019, 0.00000115, -0.00000223, -0.00000384, -0.00000877, -0.00000885, -0.00000436, -0.00001514, -0.00002219, -0.00008205, 0.00008742, 0.00008055, 0.00011482, 0.00003044, -0.00008684, -0.00000311, -0.00002315, -0.00000409, -0.00000184, -0.00000414, 0.00000001, -0.00000061, -0.00001154, -0.00000172, -0.00000160, 0.00000280, 0.00000575, 0.00000249, -0.00000747, -0.00000074, 0.00000259, 0.00000960, -0.00000528, 0.00000229, -0.00000339, -0.00000149, 0.00000080, 0.00000491, 0.00000003, -0.00000081, -0.00000299, 0.00000543, -0.00000846, 0.00001076, -0.00000592, 0.00000672, -0.00000860, -0.00000752, 0.00000251, -0.00000030, -0.00001066, 0.00000323, -0.00001395, 0.00000057, 0.00001014, -0.00000141, -0.00000076, -0.00000583, 0.00002033, },
		{ 0.00001129, -0.00001565, -0.00000009, 0.00000064, -0.00001127, -0.00000171, -0.00000059, 0.00000588, -0.00001422, 0.00000433, 0.00000188, -0.00000047, -0.00000364, 0.00000054, 0.00000160, -0.00000043, -0.00000071, 0.00000393, -0.00000245, 0.00000318, -0.00000293, 0.00000099, -0.00000214, 0.00000233, 0.00000834, -0.00000317, -0.00000691, 0.00000344, 0.00000601, 0.00000314, 0.00000022, -0.00000453, -0.00001123, 0.00000553, -0.00000300, -0.00000132, -0.00000549, -0.00000858, -0.00000150, -0.00003958, -0.00001536, 0.00004963, 0.00004545, 0.00002455, 0.00005848, -0.00000293, -0.00003674, -0.00000869, -0.00000544, -0.00001406, -0.00000273, -0.00000288, 0.00000246, -0.00000132, -0.00000166, -0.00000175, -0.00000166, -0.00000222, 0.00000275, -0.00000094, -0.00000638, -0.00000217, -0.00000164, -0.00000071, -0.00000820, 0.00000350, -0.00000237, -0.00000108, -0.00000170, -0.00000155, 0.00000101, -0.00000037, -0.00000452, -0.00000459, -0.00001351, 0.00000002, -0.00001816, -0.00000753, -0.00008224, 0.00003883, 0.00009614, 0.00008464, 0.00008921, -0.00008034, -0.00002315, -0.00001407, -0.00001267, -0.00000010, -0.00000372, -0.00000279, 0.00000332, -0.00001153, -0.00000370, -0.00000228, 0.00000241, 0.00000379, 0.00000669, -0.00000769, -0.00000148, -0.00000052, 0.00001131, -0.00000336, 0.00000079, -0.00000197, -0.00000187, -0.00000073, 0.00000517, 0.00000025, 0.00000118, -0.00000544, 0.00000685, -0.00000786, 0.00000689, -0.00000141, 0.00000254, -0.00000171, -0.00001180, 0.00000194, 0.00000133, -0.00000892, -0.00000143, -0.00000538, -0.00001109, 0.00001460, -0.00000137, 0.00000201, -0.00000859, 0.00001346, },
		{ 0.00001877, -0.00001052, -0.00000776, 0.00000512, -0.00001137, -0.00000300, -0.00000243, 0.00000686, -0.00001058, -0.00000248, 0.00000579, -0.00000180, -0.00000133, -0.00000237, 0.00000309, -0.00000093, -0.00000061, 0.00000289, -0.00000028, 0.00000087, -0.00000007, -0.00000194, 0.00000077, -0.00000191, 0.00001003, -0.00000065, -0.00000669, -0.00000088, 0.00000763, 0.00000288, 0.00000190, -0.00000301, -0.00001042, -0.00000090, 0.00000355, -0.00000631, 0.00000100, -0.00001365, 0.00000236, -0.00002909, -0.00003416, 0.00003217, 0.00005968, 0.00002281, 0.00005656, 0.00002096, -0.00004005, -0.00001592, -0.00000312, -0.00001341, -0.00000626, -0.00000242, 0.00000057, 0.00000058, -0.00000221, -0.00000142, -0.00000175, -0.00000223, 0.00000095, 0.00000182, -0.00000675, -0.00000252, -0.00000249, 0.00000047, -0.00000780, 0.00000046, 0.00000070, -0.00000295, -0.00000036, -0.00000261, 0.00000047, 0.00000093, -0.00000437, -0.00000215, -0.00001408, -0.00000126, -0.00001505, -0.00000493, -0.00006404, -0.00001241, 0.00010500, 0.00006245, 0.00011467, -0.00003863, -0.00005348, -0.00000480, -0.00001917, -0.00000081, -0.00000256, -0.00000477, 0.00000421, -0.00000798, -0.00000728, -0.00000212, 0.00000124, 0.00000263, 0.00000824, -0.00000470, -0.00000382, -0.00000172, 0.00000967, 0.00000119, -0.00000203, 0.00000022, -0.00000283, -0.00000098, 0.00000377, 0.00000160, 0.00000161, -0.00000504, 0.00000452, -0.00000340, 0.00000048, 0.00000402, -0.00000208, 0.00000431, -0.00001378, 0.00000028, 0.00000154, -0.00000478, -0.00000751, 0.00000190, -0.00001662, 0.00001083, 0.00000418, 0.00000055, -0.00000552, 0.00000314, },
		{ 0.00001879, -0.00000020, -0.00001439, 0.00000562, -0.00000799, -0.00000617, -0.00000220, 0.00000456, -0.00000429, -0.00000932, 0.00000802, -0.00000208, 0.00000038, -0.00000406, 0.00000318, -0.00000050, -0.00000076, 0.00000168, 0.00000176, -0.00000090, 0.00000197, -0.00000368, 0.00000222, -0.00000382, 0.00000875, 0.00000303, -0.00000589, -0.00000431, 0.00000709, 0.00000402, 0.00000238, -0.00000131, -0.00000824, -0.00000689, 0.00000666, -0.00000677, 0.00000165, -0.00001109, -0.00000276, -0.00001383, -0.00004413, 0.00000751, 0.00006568, 0.00003150, 0.00004638, 0.00004472, -0.00003536, -0.00002569, -0.00000281, -0.00001096, -0.00000972, -0.00000263, -0.00000094, 0.00000142, -0.00000182, -0.00000150, -0.00000177, -0.00000187, -0.00000095, 0.00000346, -0.00000563, -0.00000372, -0.00000248, 0.00000015, -0.00000560, -0.00000342, 0.00000322, -0.00000372, -0.00000006, -0.00000258, -0.00000070, 0.00000172, -0.00000347, -0.00000147, -0.00001136, -0.00000581, -0.00000933, -0.00000822, -0.00004053, -0.00005048, 0.00009443, 0.00005973, 0.00010609, 0.00002105, -0.00007416, -0.00000704, -0.00001807, -0.00000591, -0.00000038, -0.00000587, 0.00000271, -0.00000289, -0.00001014, -0.00000269, 0.00000020, 0.00000191, 0.00000769, -0.00000005, -0.00000606, -0.00000220, 0.00000630, 0.00000609, -0.00000404, 0.00000155, -0.00000320, -0.00000098, 0.00000199, 0.00000290, 0.00000142, -0.00000306, 0.00000061, 0.00000172, -0.00000483, 0.00000710, -0.00000430, 0.00000698, -0.00001200, -0.00000320, 0.00000175, -0.00000125, -0.00001061, 0.00000328, -0.00001356, 0.00000017, 0.00001176, -0.00000162, -0.00000117, -0.00000469, },
		{ 0.00001189, 0.00001107, -0.00001619, 0.00000124, -0.00000226, -0.00000968, -0.00000146, 0.00000123, 0.00000150, -0.00001317, 0.00000667, -0.00000021, 0.00000040, -0.00000358, 0.00000135, 0.00000112, -0.00000151, 0.00000100, 0.00000269, -0.00000114, 0.00000230, -0.00000343, 0.00000153, -0.00000308, 0.00000499, 0.00000705, -0.00000467, -0.00000603, 0.00000415, 0.00000626, 0.00000199, 0.00000053, -0.00000599, -0.00001005, 0.00000472, -0.00000252, -0.00000230, -0.00000434, -0.00001011, -0.00000283, -0.00004128, -0.00001951, 0.00005950, 0.00004757, 0.00003502, 0.00006112, -0.00001997, -0.00003695, -0.00000457, -0.00000840, -0.00001136, -0.00000452, -0.00000141, 0.00000087, -0.00000064, -0.00000179, -0.00000175, -0.00000148, -0.00000224, 0.00000347, -0.00000311, -0.00000543, -0.00000198, -0.00000102, -0.00000290, -0.00000650, 0.00000382, -0.00000271, -0.00000100, -0.00000162, -0.00000204, 0.00000176, -0.00000196, -0.00000196, -0.00000716, -0.00001040, -0.00000487, -0.00001160, -0.00002142, -0.00006783, 0.00006384, 0.00007205, 0.00007964, 0.00007315, -0.00006851, -0.00002460, -0.00001063, -0.00001298, 0.00000095, -0.00000551, -0.00000025, 0.00000162, -0.00001057, -0.00000468, -0.00000040, 0.00000136, 0.00000599, 0.00000439, -0.00000663, -0.00000305, 0.00000280, 0.00000936, -0.00000378, 0.00000114, -0.00000239, -0.00000145, 0.00000078, 0.00000330, 0.00000142, -0.00000090, -0.00000266, 0.00000486, -0.00000666, 0.00000639, -0.00000322, 0.00000587, -0.00000687, -0.00000787, 0.00000234, 0.00000022, -0.00000936, -0.00000101, -0.00000529, -0.00001108, 0.00001515, -0.00000000, 0.00000033, -0.00000669, },
		{ 0.00000263, 0.00001791, -0.00001106, -0.00000655, 0.00000319, -0.00001133, -0.00000194, -0.00000109, 0.00000459, -0.00001271, 0.00000182, 0.00000322, -0.00000087, -0.00000147, -0.00000151, 0.00000301, -0.00000221, 0.00000055, 0.00000266, -0.00000018, 0.00000144, -0.00000189, -0.00000041, -0.00000096, 0.00000065, 0.00000976, -0.00000251, -0.00000640, 0.00000013, 0.00000797, 0.00000208, 0.00000171, -0.00000379, -0.00001041, -0.00000042, 0.00000286, -0.00000611, 0.00000057, -0.00001281, -0.00000113, -0.00002804, -0.00004115, 0.00003977, 0.00006466, 0.00002966, 0.00006597, 0.00000478, -0.00004625, -0.00000924, -0.00000649, -0.00001099, -0.00000745, -0.00000155, -0.00000026, 0.00000043, -0.00000172, -0.00000185, -0.00000124, -0.00000264, 0.00000204, 0.00000001, -0.00000677, -0.00000174, -0.00000208, -0.00000091, -0.00000771, 0.00000220, -0.00000038, -0.00000242, -0.00000063, -0.00000276, 0.00000086, -0.00000015, -0.00000280, -0.00000328, -0.00001270, -0.00000366, -0.00001234, -0.00001040, -0.00006616, 0.00002264, 0.00008581, 0.00005563, 0.00009909, -0.00003435, -0.00004892, -0.00000500, -0.00001670, -0.00000117, -0.00000318, -0.00000374, 0.00000420, -0.00000831, -0.00000741, -0.00000115, 0.00000109, 0.00000390, 0.00000738, -0.00000500, -0.00000439, -0.00000008, 0.00001020, -0.00000120, -0.00000060, -0.00000083, -0.00000227, 0.00000026, 0.00000280, 0.00000177, 0.00000068, -0.00000416, 0.00000508, -0.00000495, 0.00000274, -0.00000008, 0.00000252, -0.00000051, -0.00001184, 0.00000236, 0.00000032, -0.00000557, -0.00000703, 0.00000195, -0.00001634, 0.00001057, 0.00000634, -0.00000117, -0.00000420, },
		{ -0.00000368, 0.00001717, -0.00000044, -0.00001393, 0.00000521, -0.00000956, -0.00000441, -0.00000164, 0.00000457, -0.00000875, -0.00000466, 0.00000638, -0.00000204, 0.00000078, -0.00000377, 0.00000379, -0.00000189, -0.00000028, 0.00000234, 0.00000103, 0.00000049, -0.00000028, -0.00000218, 0.00000082, -0.00000238, 0.00000988, 0.00000090, -0.00000611, -0.00000331, 0.00000761, 0.00000358, 0.00000172, -0.00000150, -0.00000912, -0.00000567, 0.00000551, -0.00000608, 0.00000041, -0.00000914, -0.00000656, -0.00001265, -0.00004991, 0.00000928, 0.00007487, 0.00003468, 0.00006000, 0.00003341, -0.00004855, -0.00001797, -0.00000514, -0.00000978, -0.00000974, -0.00000269, -0.00000086, 0.00000050, -0.00000090, -0.00000214, -0.00000118, -0.00000232, -0.00000004, 0.00000258, -0.00000680, -0.00000243, -0.00000237, -0.00000029, -0.00000690, -0.00000095, 0.00000213, -0.00000326, -0.00000036, -0.00000254, -0.00000067, 0.00000131, -0.00000312, -0.00000096, -0.00001193, -0.00000545, -0.00001068, -0.00000643, -0.00005269, -0.00001659, 0.00008829, 0.00004653, 0.00009611, 0.00001496, -0.00006414, -0.00000972, -0.00001386, -0.00000698, 0.00000001, -0.00000633, 0.00000418, -0.00000422, -0.00000955, -0.00000270, 0.00000110, 0.00000203, 0.00000839, -0.00000164, -0.00000561, -0.00000228, 0.00000886, 0.00000271, -0.00000250, 0.00000058, -0.00000284, -0.00000001, 0.00000193, 0.00000214, 0.00000161, -0.00000390, 0.00000305, -0.00000130, -0.00000160, 0.00000293, -0.00000077, 0.00000456, -0.00001311, 0.00000057, 0.00000056, -0.00000211, -0.00001058, 0.00000372, -0.00001337, -0.00000024, 0.00001356, -0.00000210, -0.00000140, },
		{ -0.00000482, 0.00001022, 0.00001077, -0.00001633, 0.00000191, -0.00000455, -0.00000805, -0.00000117, 0.00000260, -0.00000340, -0.00001022, 0.00000734, -0.00000183, 0.00000185, -0.00000418, 0.00000266, -0.00000022, -0.00000163, 0.00000208, 0.00000183, 0.00000021, 0.00000061, -0.00000295, 0.00000133, -0.00000327, 0.00000742, 0.00000498, -0.00000534, -0.00000532, 0.00000498, 0.00000606, 0.00000117, 0.00000050, -0.00000709, -0.00000901, 0.00000406, -0.00000225, -0.00000328, -0.00000286, -0.00001227, -0.00000369, -0.00004345, -0.00002398, 0.00007076, 0.00004996, 0.00004843, 0.00005842, -0.00003958, -0.00003065, -0.00000444, -0.00000896, -0.00001016, -0.00000537, -0.00000066, -0.00000047, 0.00000033, -0.00000229, -0.00000138, -0.00000165, -0.00000188, 0.00000365, -0.00000511, -0.00000412, -0.00000183, -0.00000087, -0.00000479, -0.00000436, 0.00000359, -0.00000286, -0.00000097, -0.00000172, -0.00000207, 0.00000173, -0.00000238, -0.00000050, -0.00000874, -0.00000855, -0.00000831, -0.00000652, -0.00003546, -0.00004432, 0.00007422, 0.00005268, 0.00007524, 0.00005943, -0.00005782, -0.00002610, -0.00000729, -0.00001299, 0.00000124, -0.00000640, 0.00000153, 0.00000038, -0.00001004, -0.00000504, 0.00000091, 0.00000095, 0.00000747, 0.00000245, -0.00000599, -0.00000394, 0.00000599, 0.00000659, -0.00000340, 0.00000111, -0.00000274, -0.00000042, 0.00000118, 0.00000222, 0.00000206, -0.00000249, 0.00000017, 0.00000223, -0.00000451, 0.00000409, -0.00000240, 0.00000674, -0.00001083, -0.00000330, 0.00000171, -0.00000058, -0.00001001, -0.00000010, -0.00000533, -0.00001121, 0.00001596, 0.00000111, -0.00000128, },
		{ -0.00000279, 0.00000197, 0.00001693, -0.00001126, -0.00000571, 0.00000134, -0.00001076, -0.00000122, 0.00000042, 0.00000107, -0.00001285, 0.00000519, 0.00000015, 0.00000129, -0.00000254, -0.00000004, 0.00000215, -0.00000293, 0.00000169, 0.00000216, 0.00000061, 0.00000079, -0.00000271, 0.00000062, -0.00000241, 0.00000356, 0.00000841, -0.00000375, -0.00000600, 0.00000110, 0.00000805, 0.00000128, 0.00000153, -0.00000464, -0.00001009, -0.00000019, 0.00000242, -0.00000640, 0.00000090, -0.00001277, -0.00000422, -0.00002719, -0.00004881, 0.00004852, 0.00007015, 0.00003897, 0.00007331, -0.00001816, -0.00004471, -0.00000554, -0.00000880, -0.00000888, -0.00000869, -0.00000054, -0.00000158, 0.00000106, -0.00000181, -0.00000189, -0.00000100, -0.00000285, 0.00000296, -0.00000211, -0.00000610, -0.00000115, -0.00000184, -0.00000255, -0.00000672, 0.00000324, -0.00000123, -0.00000200, -0.00000095, -0.00000269, 0.00000091, -0.00000083, -0.00000122, -0.00000478, -0.00001087, -0.00000694, -0.00000780, -0.00002038, -0.00005677, 0.00004645, 0.00006569, 0.00005232, 0.00008404, -0.00002907, -0.00004589, -0.00000444, -0.00001452, -0.00000172, -0.00000360, -0.00000268, 0.00000393, -0.00000847, -0.00000753, -0.00000011, 0.00000091, 0.00000522, 0.00000611, -0.00000503, -0.00000513, 0.00000241, 0.00000930, -0.00000266, 0.00000055, -0.00000194, -0.00000110, 0.00000077, 0.00000191, 0.00000222, -0.00000072, -0.00000223, 0.00000417, -0.00000493, 0.00000298, -0.00000202, 0.00000593, -0.00000573, -0.00000811, 0.00000308, -0.00000061, -0.00000664, -0.00000611, 0.00000205, -0.00001629, 0.00001055, 0.00000837, -0.00000297, },
		},
		{ { -0.00000082, 0.00000106, -0.00001443, 0.00000022, 0.00001237, -0.00000682, 0.00000134, 0.00000351, -0.00001241, 0.00000002, 0.00000191, -0.00000271, -0.00000697, -0.00000336, 0.00001654, -0.00000101, 0.00000150, 0.00000187, -0.00000384, -0.00000050, 0.00000048, 0.00000094, 0.00000256, 0.00000100, -0.00000235, -0.00000163, -0.00000103, -0.00000127, 0.00000232, -0.00000199, -0.00000881, -0.00000714, -0.00000558, -0.00000180, 0.00000864, -0.00000122, -0.00001583, -0.00000818, -0.00006519, 0.00001224, 0.00008439, 0.00003891, 0.00008027, 0.00001164, -0.00006090, -0.00001326, 0.00000108, -0.00000841, -0.00000654, -0.00000449, -0.00000643, 0.00000017, 0.00000587, -0.00000231, -0.00000117, -0.00000237, 0.00000443, -0.00000310, -0.00000022, -0.00000013, -0.00000139, -0.00000039, -0.00000988, 0.00000319, 0.00000476, 0.00000319, -0.00000988, -0.00000039, -0.00000139, -0.00000013, -0.00000022, -0.00000310, 0.00000443, -0.00000237, -0.00000117, -0.00000231, 0.00000587, 0.00000017, -0.00000643, -0.00000449, -0.00000654, -0.00000841, 0.00000108, -0.00001326, -0.00006090, 0.00001164, 0.00008027, 0.00003891, 0.00008439, 0.00001224, -0.00006519, -0.00000818, -0.00001583, -0.00000122, 0.00000864, -0.00000180, -0.00000558, -0.00000714, -0.00000881, -0.00000199, 0.00000232, -0.00000127, -0.00000103, -0.00000163, -0.00000235, 0.00000100, 0.00000256, 0.00000094, 0.00000048, -0.00000050, -0.00000384, 0.00000187, 0.00000150, -0.00000101, 0.00001654, -0.00000336, -0.00000697, -0.00000271, 0.00000191, 0.00000002, -0.00001241, 0.00000351, 0.00000134, -0.00000682, 0.00001237, 0.00000022, -0.00001443, 0.00000106, },
		{ 0.00000059, 0.00000083, -0.00000883, -0.00000934, 0.00001417, -0.00000203, -0.00000357, 0.00000637, -0.00000784, -0.00000755, 0.00000470, -0.00000203, -0.00000431, -0.00000891, 0.00001234, 0.00000736, -0.00000308, 0.00000486, -0.00000371, -0.00000147, 0.00000013, 0.00000084, 0.00000168, 0.00000249, -0.00000165, -0.00000209, -0.00000117, -0.00000130, 0.00000089, 0.00000113, -0.00000789, -0.00000774, -0.00000643, -0.00000343, 0.00000425, 0.00000754, -0.00001728, -0.00000360, -0.00004649, -0.00004102, 0.00008441, 0.00005140, 0.00006858, 0.00005321, -0.00005712, -0.00003145, 0.00000219, -0.00000622, -0.00000817, -0.00000484, -0.00000582, -0.00000360, 0.00000635, -0.00000027, -0.00000166, -0.00000244, 0.00000192, 0.00000133, -0.00000397, 0.00000279, -0.00000393, 0.00000221, -0.00000856, -0.00000208, 0.00000521, 0.00000500, -0.00000555, -0.00000655, 0.00000224, -0.00000304, 0.00000248, -0.00000486, 0.00000342, 0.00000002, -0.00000175, -0.00000264, 0.00000331, 0.00000389, -0.00000577, -0.00000502, -0.00000501, -0.00000853, -0.00000295, -0.00000106, -0.00005001, -0.00002550, 0.00007292, 0.00004641, 0.00006173, 0.00005923, -0.00005554, -0.00002831, -0.00000817, -0.00001150, 0.00001020, 0.00000095, -0.00000429, -0.00000693, -0.00000827, -0.00000557, 0.00000264, -0.00000052, -0.00000114, -0.00000132, -0.00000214, -0.00000090, 0.00000290, 0.00000125, 0.00000091, -0.00000020, -0.00000241, -0.00000178, 0.00000462, -0.00000395, 0.00001214, 0.00000701, -0.00001027, -0.00000159, -0.00000277, 0.00000509, -0.00001077, -0.00000348, 0.00000596, -0.00000785, 0.00000626, 0.00000925, -0.00001513, -0.00000289, },
		{ 0.00000298, -0.00000104, -0.00000271, -0.00001468, 0.00000979, 0.00000526, -0.00000685, 0.00000537, -0.00000133, -0.00001153, 0.00000170, 0.00000184, -0.00000376, -0.00000888, 0.00000293, 0.00001478, -0.00000349, 0.00000370, -0.00000050, -0.00000360, 0.00000028, 0.00000033, 0.00000128, 0.00000259, 0.00000018, -0.00000266, -0.00000129, -0.00000133, -0.00000022, 0.00000226, -0.00000491, -0.00000896, -0.00000648, -0.00000502, 0.00000055, 0.00000993, -0.00000942, -0.00001154, -0.00001853, -0.00007186, 0.00004839, 0.00007915, 0.00005070, 0.00008383, -0.00003284, -0.00005206, -0.00000153, -0.00000324, -0.00000849, -0.00000645, -0.00000465, -0.00000621, 0.00000435, 0.00000292, -0.00000259, -0.00000126, -0.00000150, 0.00000445, -0.00000467, 0.00000203, -0.00000224, -0.00000029, -0.00000295, -0.00000786, 0.00000488, 0.00000473, 0.00000062, -0.00001061, 0.00000185, -0.00000213, 0.00000135, -0.00000253, -0.00000068, 0.00000327, -0.00000232, -0.00000206, 0.00000011, 0.00000621, -0.00000367, -0.00000562, -0.00000457, -0.00000668, -0.00000738, 0.00000403, -0.00003224, -0.00004828, 0.00004685, 0.00006245, 0.00003949, 0.00007872, -0.00001773, -0.00005080, -0.00000566, -0.00001570, 0.00000544, 0.00000557, -0.00000331, -0.00000611, -0.00000776, -0.00000771, 0.00000071, 0.00000123, -0.00000161, -0.00000092, -0.00000185, -0.00000193, 0.00000179, 0.00000227, 0.00000081, 0.00000044, -0.00000145, -0.00000315, 0.00000350, -0.00000085, 0.00000300, 0.00001564, -0.00000892, -0.00000308, -0.00000390, 0.00000334, -0.00000292, -0.00001127, 0.00000726, -0.00000488, -0.00000109, 0.00001382, -0.00000949, -0.00000990, },
		{ 0.00000256, -0.00000095, -0.00000005, -0.00001345, 0.00000038, 0.00001221, -0.00000713, 0.00000219, 0.00000328, -0.00000970, -0.00000523, 0.00000527, -0.00000260, -0.00000700, -0.00000512, 0.00001523, 0.00000255, -0.00000098, 0.00000381, -0.00000496, -0.00000023, 0.00000004, 0.00000100, 0.00000201, 0.00000191, -0.00000235, -0.00000182, -0.00000127, -0.00000089, 0.00000191, -0.00000144, -0.00000917, -0.00000700, -0.00000587, -0.00000220, 0.00000758, 0.00000128, -0.00001856, -0.00000306, -0.00006619, -0.00001241, 0.00009891, 0.00004655, 0.00009083, 0.00001064, -0.00006782, -0.00001165, -0.00000102, -0.00000700, -0.00000865, -0.00000400, -0.00000704, 0.00000064, 0.00000560, -0.00000241, -0.00000076, -0.00000299, 0.00000373, -0.00000130, -0.00000184, 0.00000169, -0.00000402, 0.00000148, -0.00000940, 0.00000122, 0.00000502, 0.00000417, -0.00000868, -0.00000374, 0.00000177, -0.00000181, 0.00000108, -0.00000430, 0.00000439, -0.00000127, -0.00000175, -0.00000222, 0.00000624, -0.00000046, -0.00000571, -0.00000508, -0.00000440, -0.00000968, 0.00000273, -0.00001435, -0.00005434, 0.00001186, 0.00007139, 0.00003299, 0.00006968, 0.00002923, -0.00005661, -0.00001628, -0.00001210, -0.00000383, 0.00000954, -0.00000159, -0.00000501, -0.00000735, -0.00000834, -0.00000256, 0.00000264, -0.00000149, -0.00000088, -0.00000153, -0.00000205, -0.00000009, 0.00000307, 0.00000085, 0.00000105, -0.00000093, -0.00000269, 0.00000025, 0.00000333, -0.00000320, 0.00001584, -0.00000009, -0.00000827, -0.00000089, -0.00000294, 0.00000525, -0.00001395, 0.00000274, 0.00000103, -0.00000687, 0.00001269, 0.00000004, -0.00001567, },
		{ -0.00000275, 0.00000213, -0.00000121, -0.00000762, -0.00000957, 0.00001488, -0.00000333, -0.00000194, 0.00000512, -0.00000463, -0.00001060, 0.00000383, 0.00000131, -0.00000592, -0.00000843, 0.00000864, 0.00001103, -0.00000406, 0.00000498, -0.00000297, -0.00000259, 0.00000069, 0.00000013, 0.00000182, 0.00000228, -0.00000054, -0.00000290, -0.00000090, -0.00000162, 0.00000137, 0.00000058, -0.00000700, -0.00000880, -0.00000545, -0.00000480, 0.00000420, 0.00000755, -0.00001569, -0.00000611, -0.00003623, -0.00006565, 0.00008390, 0.00006691, 0.00007532, 0.00006073, -0.00006779, -0.00002940, -0.00000040, -0.00000477, -0.00000982, -0.00000493, -0.00000622, -0.00000335, 0.00000622, -0.00000032, -0.00000167, -0.00000209, 0.00000023, 0.00000315, -0.00000481, 0.00000337, -0.00000432, 0.00000091, -0.00000550, -0.00000483, 0.00000563, 0.00000462, -0.00000261, -0.00000992, 0.00000357, -0.00000251, 0.00000192, -0.00000416, 0.00000179, 0.00000158, -0.00000198, -0.00000300, 0.00000401, 0.00000317, -0.00000514, -0.00000557, -0.00000345, -0.00000863, -0.00000260, -0.00000102, -0.00004700, -0.00002008, 0.00006394, 0.00004258, 0.00004724, 0.00006250, -0.00003681, -0.00003522, -0.00000791, -0.00001097, 0.00000856, 0.00000251, -0.00000463, -0.00000624, -0.00000855, -0.00000526, 0.00000207, 0.00000001, -0.00000160, -0.00000088, -0.00000206, -0.00000130, 0.00000237, 0.00000190, 0.00000081, 0.00000011, -0.00000227, -0.00000173, 0.00000403, -0.00000259, 0.00000787, 0.00001167, -0.00001196, 0.00000098, -0.00000668, 0.00000655, -0.00000807, -0.00000646, 0.00000667, -0.00000878, 0.00000703, 0.00000902, -0.00001599, },
		{ -0.00001099, 0.00000536, -0.00000287, -0.00000212, -0.00001466, 0.00001053, 0.00000420, -0.00000597, 0.00000507, -0.00000011, -0.00001068, -0.00000265, 0.00000586, -0.00000453, -0.00000849, 0.00000006, 0.00001536, -0.00000116, 0.00000144, 0.00000182, -0.00000540, 0.00000112, -0.00000057, 0.00000169, 0.00000179, 0.00000156, -0.00000330, -0.00000102, -0.00000196, 0.00000067, 0.00000128, -0.00000368, -0.00001014, -0.00000545, -0.00000619, 0.00000070, 0.00000851, -0.00000597, -0.00001599, -0.00000920, -0.00008185, 0.00002807, 0.00009885, 0.00005618, 0.00009655, -0.00004190, -0.00005350, -0.00000212, -0.00000359, -0.00000853, -0.00000767, -0.00000455, -0.00000639, 0.00000447, 0.00000282, -0.00000276, -0.00000067, -0.00000278, 0.00000479, -0.00000368, 0.00000087, -0.00000070, -0.00000305, -0.00000014, -0.00000881, 0.00000381, 0.00000465, 0.00000256, -0.00001100, -0.00000043, 0.00000088, -0.00000060, -0.00000086, -0.00000253, 0.00000406, -0.00000170, -0.00000273, 0.00000062, 0.00000604, -0.00000380, -0.00000539, -0.00000434, -0.00000523, -0.00000832, 0.00000524, -0.00003209, -0.00004066, 0.00004090, 0.00005750, 0.00002975, 0.00007037, 0.00000081, -0.00004806, -0.00001187, -0.00001173, 0.00000206, 0.00000757, -0.00000400, -0.00000495, -0.00000839, -0.00000693, -0.00000023, 0.00000194, -0.00000214, -0.00000047, -0.00000197, -0.00000160, 0.00000050, 0.00000322, 0.00000037, 0.00000127, -0.00000208, -0.00000196, 0.00000167, 0.00000166, -0.00000091, 0.00001748, -0.00000759, -0.00000313, -0.00000364, -0.00000024, 0.00000270, -0.00001427, 0.00000741, -0.00000561, -0.00000064, 0.00001381, -0.00000986, },
		{ -0.00001722, 0.00000441, -0.00000131, -0.00000084, -0.00001258, 0.00000035, 0.00001234, -0.00000779, 0.00000345, 0.00000238, -0.00000651, -0.00000958, 0.00000654, -0.00000064, -0.00000793, -0.00000584, 0.00001261, 0.00000650, -0.00000310, 0.00000513, -0.00000521, -0.00000082, 0.00000028, 0.00000050, 0.00000195, 0.00000211, -0.00000163, -0.00000249, -0.00000115, -0.00000091, 0.00000205, -0.00000161, -0.00000867, -0.00000774, -0.00000507, -0.00000365, 0.00000742, 0.00000221, -0.00001849, -0.00000282, -0.00005887, -0.00004262, 0.00010983, 0.00005780, 0.00010227, 0.00000970, -0.00007585, -0.00000901, -0.00000381, -0.00000530, -0.00001085, -0.00000345, -0.00000775, 0.00000116, 0.00000518, -0.00000227, -0.00000071, -0.00000306, 0.00000242, 0.00000075, -0.00000331, 0.00000304, -0.00000549, 0.00000157, -0.00000742, -0.00000141, 0.00000560, 0.00000427, -0.00000616, -0.00000778, 0.00000452, -0.00000261, 0.00000157, -0.00000461, 0.00000351, 0.00000018, -0.00000230, -0.00000228, 0.00000681, -0.00000129, -0.00000485, -0.00000579, -0.00000222, -0.00001085, 0.00000412, -0.00001534, -0.00004753, 0.00001074, 0.00006461, 0.00002761, 0.00005662, 0.00003864, -0.00004302, -0.00002446, -0.00000956, -0.00000499, 0.00000918, -0.00000068, -0.00000506, -0.00000683, -0.00000848, -0.00000248, 0.00000229, -0.00000108, -0.00000121, -0.00000117, -0.00000191, -0.00000083, 0.00000301, 0.00000122, 0.00000118, -0.00000088, -0.00000226, -0.00000034, 0.00000386, -0.00000350, 0.00001294, 0.00000457, -0.00001068, 0.00000224, -0.00000816, 0.00000932, -0.00001364, 0.00000075, 0.00000128, -0.00000719, 0.00001301, -0.00000000, },
		{ -0.00001718, -0.00000244, 0.00000368, -0.00000323, -0.00000613, -0.00001024, 0.00001592, -0.00000485, -0.00000022, 0.00000384, -0.00000198, -0.00001196, 0.00000129, 0.00000499, -0.00000684, -0.00000837, 0.00000547, 0.00001311, -0.00000318, 0.00000346, -0.00000075, -0.00000483, 0.00000204, -0.00000125, 0.00000263, 0.00000122, 0.00000128, -0.00000419, -0.00000009, -0.00000267, 0.00000276, -0.00000099, -0.00000504, -0.00001089, -0.00000330, -0.00000724, 0.00000532, 0.00000557, -0.00001135, -0.00001198, -0.00002289, -0.00008704, 0.00007373, 0.00008671, 0.00008277, 0.00006972, -0.00008050, -0.00002611, -0.00000363, -0.00000304, -0.00001174, -0.00000466, -0.00000704, -0.00000272, 0.00000564, 0.00000010, -0.00000213, -0.00000135, -0.00000157, 0.00000457, -0.00000492, 0.00000308, -0.00000357, -0.00000135, -0.00000241, -0.00000683, 0.00000539, 0.00000430, 0.00000013, -0.00001207, 0.00000297, -0.00000040, 0.00000036, -0.00000270, -0.00000020, 0.00000295, -0.00000185, -0.00000362, 0.00000484, 0.00000240, -0.00000456, -0.00000609, -0.00000197, -0.00000867, -0.00000219, -0.00000145, -0.00004309, -0.00001626, 0.00005636, 0.00003938, 0.00003507, 0.00006110, -0.00001954, -0.00003686, -0.00001172, -0.00000757, 0.00000502, 0.00000522, -0.00000610, -0.00000439, -0.00000975, -0.00000406, 0.00000069, 0.00000122, -0.00000253, -0.00000006, -0.00000244, -0.00000095, 0.00000099, 0.00000320, 0.00000004, 0.00000120, -0.00000310, -0.00000059, 0.00000236, -0.00000007, 0.00000293, 0.00001584, -0.00001296, 0.00000277, -0.00000873, 0.00000548, -0.00000352, -0.00000990, 0.00000717, -0.00000938, 0.00000738, 0.00000915, },
		{ -0.00001030, -0.00001220, 0.00000797, -0.00000487, -0.00000114, -0.00001496, 0.00001129, 0.00000328, -0.00000533, 0.00000508, 0.00000031, -0.00000877, -0.00000693, 0.00000831, -0.00000344, -0.00000918, -0.00000157, 0.00001406, 0.00000252, -0.00000145, 0.00000438, -0.00000703, 0.00000161, -0.00000119, 0.00000193, 0.00000107, 0.00000266, -0.00000355, -0.00000093, -0.00000258, 0.00000159, 0.00000035, -0.00000263, -0.00001100, -0.00000474, -0.00000683, 0.00000041, 0.00000714, -0.00000288, -0.00001911, -0.00000352, -0.00008268, -0.00000164, 0.00011833, 0.00006467, 0.00011029, -0.00005192, -0.00005506, -0.00000230, -0.00000430, -0.00000850, -0.00000873, -0.00000464, -0.00000642, 0.00000439, 0.00000285, -0.00000295, -0.00000022, -0.00000365, 0.00000445, -0.00000214, -0.00000058, 0.00000088, -0.00000517, 0.00000116, -0.00000810, 0.00000172, 0.00000513, 0.00000347, -0.00000960, -0.00000427, 0.00000441, -0.00000239, 0.00000051, -0.00000380, 0.00000408, -0.00000059, -0.00000352, 0.00000105, 0.00000604, -0.00000414, -0.00000500, -0.00000426, -0.00000373, -0.00000913, 0.00000604, -0.00003146, -0.00003352, 0.00003458, 0.00005433, 0.00002084, 0.00006197, 0.00001301, -0.00003911, -0.00002040, -0.00000713, -0.00000117, 0.00000918, -0.00000462, -0.00000380, -0.00000883, -0.00000628, -0.00000097, 0.00000238, -0.00000235, -0.00000023, -0.00000207, -0.00000107, -0.00000086, 0.00000401, 0.00000004, 0.00000205, -0.00000267, -0.00000095, 0.00000029, 0.00000346, -0.00000338, 0.00001713, -0.00000443, -0.00000487, -0.00000127, -0.00000534, 0.00000830, -0.00001603, 0.00000639, -0.00000558, -0.00000071, 0.00001403, },
		{ 0.00000019, -0.00001904, 0.00000648, -0.00000174, -0.00000145, -0.00001165, -0.00000004, 0.00001280, -0.00000873, 0.00000493, 0.00000112, -0.00000340, -0.00001251, 0.00000578, 0.00000246, -0.00000886, -0.00000634, 0.00000967, 0.00000973, -0.00000389, 0.00000503, -0.00000395, -0.00000270, 0.00000152, -0.00000083, 0.00000268, 0.00000134, 0.00000001, -0.00000384, -0.00000037, -0.00000173, 0.00000319, -0.00000295, -0.00000688, -0.00000978, -0.00000277, -0.00000651, 0.00000872, 0.00000081, -0.00001488, -0.00000776, -0.00004468, -0.00007423, 0.00011277, 0.00007455, 0.00011395, 0.00000956, -0.00008561, -0.00000503, -0.00000746, -0.00000320, -0.00001331, -0.00000258, -0.00000882, 0.00000200, 0.00000436, -0.00000166, -0.00000118, -0.00000258, 0.00000067, 0.00000276, -0.00000431, 0.00000360, -0.00000567, 0.00000020, -0.00000461, -0.00000399, 0.00000601, 0.00000397, -0.00000308, -0.00001144, 0.00000585, -0.00000190, 0.00000093, -0.00000387, 0.00000185, 0.00000181, -0.00000269, -0.00000255, 0.00000753, -0.00000223, -0.00000399, -0.00000647, -0.00000016, -0.00001183, 0.00000534, -0.00001649, -0.00004021, 0.00000823, 0.00005974, 0.00002250, 0.00004554, 0.00004223, -0.00002789, -0.00002980, -0.00001005, -0.00000357, 0.00000688, 0.00000146, -0.00000625, -0.00000511, -0.00000960, -0.00000145, 0.00000102, 0.00000014, -0.00000214, -0.00000037, -0.00000218, -0.00000087, 0.00000208, 0.00000233, 0.00000059, -0.00000001, -0.00000285, 0.00000033, 0.00000297, -0.00000199, 0.00000835, 0.00000991, -0.00001338, 0.00000552, -0.00001234, 0.00001115, -0.00001108, -0.00000242, 0.00000194, -0.00000758, 0.00001316, },
		{ 0.00000961, -0.00001855, -0.00000210, 0.00000535, -0.00000532, -0.00000432, -0.00001133, 0.00001713, -0.00000637, 0.00000141, 0.00000267, -0.00000003, -0.00001191, -0.00000201, 0.00000787, -0.00000621, -0.00000923, 0.00000335, 0.00001320, -0.00000039, 0.00000051, 0.00000255, -0.00000766, 0.00000367, -0.00000287, 0.00000374, -0.00000026, 0.00000335, -0.00000558, 0.00000095, -0.00000415, 0.00000471, -0.00000314, -0.00000257, -0.00001339, -0.00000071, -0.00000996, 0.00000685, 0.00000254, -0.00000570, -0.00001888, -0.00001029, -0.00010062, 0.00005165, 0.00010989, 0.00009213, 0.00007970, -0.00009481, -0.00002209, -0.00000711, -0.00000124, -0.00001386, -0.00000400, -0.00000829, -0.00000169, 0.00000458, 0.00000094, -0.00000291, -0.00000041, -0.00000327, 0.00000546, -0.00000438, 0.00000218, -0.00000225, -0.00000363, -0.00000025, -0.00000739, 0.00000409, 0.00000451, 0.00000200, -0.00001234, 0.00000025, 0.00000303, -0.00000180, -0.00000088, -0.00000218, 0.00000385, -0.00000131, -0.00000439, 0.00000562, 0.00000179, -0.00000422, -0.00000635, -0.00000076, -0.00000852, -0.00000179, -0.00000222, -0.00003864, -0.00001334, 0.00004933, 0.00003741, 0.00002403, 0.00005791, -0.00000653, -0.00003242, -0.00001909, -0.00000216, 0.00000046, 0.00000841, -0.00000816, -0.00000195, -0.00001131, -0.00000249, -0.00000099, 0.00000258, -0.00000351, 0.00000082, -0.00000305, -0.00000009, -0.00000088, 0.00000477, -0.00000101, 0.00000264, -0.00000436, 0.00000096, 0.00000042, 0.00000262, -0.00000148, 0.00001841, -0.00001228, 0.00000279, -0.00000825, 0.00000194, 0.00000224, -0.00001306, 0.00000697, -0.00000934, 0.00000718, },
		{ 0.00001435, -0.00001069, -0.00001355, 0.00001074, -0.00000693, 0.00000009, -0.00001538, 0.00001183, 0.00000265, -0.00000496, 0.00000535, 0.00000018, -0.00000640, -0.00001035, 0.00000874, -0.00000069, -0.00001043, -0.00000245, 0.00001160, 0.00000653, -0.00000393, 0.00000618, -0.00000756, 0.00000102, -0.00000094, 0.00000145, 0.00000097, 0.00000295, -0.00000298, -0.00000145, -0.00000268, 0.00000191, 0.00000018, -0.00000254, -0.00001072, -0.00000521, -0.00000607, -0.00000120, 0.00000696, -0.00000170, -0.00001896, -0.00000356, -0.00007360, -0.00003814, 0.00013331, 0.00007880, 0.00012388, -0.00006187, -0.00005765, -0.00000146, -0.00000571, -0.00000820, -0.00000987, -0.00000465, -0.00000657, 0.00000440, 0.00000270, -0.00000286, -0.00000014, -0.00000400, 0.00000353, -0.00000033, -0.00000190, 0.00000197, -0.00000611, 0.00000077, -0.00000597, -0.00000100, 0.00000592, 0.00000351, -0.00000683, -0.00000879, 0.00000741, -0.00000320, 0.00000104, -0.00000414, 0.00000320, 0.00000093, -0.00000429, 0.00000129, 0.00000626, -0.00000468, -0.00000446, -0.00000428, -0.00000225, -0.00000979, 0.00000655, -0.00003073, -0.00002635, 0.00002743, 0.00005313, 0.00001210, 0.00005480, 0.00001919, -0.00002636, -0.00002855, -0.00000398, -0.00000273, 0.00000932, -0.00000430, -0.00000351, -0.00000830, -0.00000643, -0.00000093, 0.00000201, -0.00000181, -0.00000057, -0.00000183, -0.00000077, -0.00000178, 0.00000410, 0.00000032, 0.00000223, -0.00000260, -0.00000079, 0.00000001, 0.00000392, -0.00000386, 0.00001439, 0.00000042, -0.00000796, 0.00000244, -0.00001068, 0.00001259, -0.00001575, 0.00000388, -0.00000470, -0.00000124, },
		{ 0.00001294, 0.00000064, -0.00002102, 0.00000858, -0.00000210, -0.00000202, -0.00001051, -0.00000085, 0.00001357, -0.00000981, 0.00000644, -0.00000026, -0.00000069, -0.00001396, 0.00000356, 0.00000562, -0.00000880, -0.00000738, 0.00000724, 0.00001138, -0.00000275, 0.00000313, -0.00000105, -0.00000580, 0.00000360, -0.00000287, 0.00000411, -0.00000030, 0.00000239, -0.00000574, 0.00000103, -0.00000338, 0.00000535, -0.00000542, -0.00000391, -0.00001295, 0.00000074, -0.00001047, 0.00001122, -0.00000268, -0.00000833, -0.00001642, -0.00002704, -0.00010175, 0.00010341, 0.00009780, 0.00012607, 0.00001032, -0.00009716, 0.00000018, -0.00001180, -0.00000075, -0.00001612, -0.00000124, -0.00001042, 0.00000335, 0.00000295, -0.00000045, -0.00000218, -0.00000162, -0.00000131, 0.00000450, -0.00000472, 0.00000340, -0.00000484, -0.00000192, -0.00000194, -0.00000571, 0.00000564, 0.00000385, -0.00000031, -0.00001366, 0.00000500, 0.00000057, -0.00000084, -0.00000223, -0.00000035, 0.00000339, -0.00000281, -0.00000297, 0.00000825, -0.00000303, -0.00000335, -0.00000689, 0.00000155, -0.00001251, 0.00000636, -0.00001779, -0.00003253, 0.00000479, 0.00005606, 0.00001812, 0.00003572, 0.00004258, -0.00001459, -0.00003021, -0.00001434, 0.00000056, 0.00000276, 0.00000475, -0.00000855, -0.00000231, -0.00001155, 0.00000036, -0.00000096, 0.00000193, -0.00000345, 0.00000073, -0.00000286, -0.00000020, 0.00000032, 0.00000408, -0.00000065, 0.00000159, -0.00000432, 0.00000192, 0.00000112, 0.00000063, 0.00000307, 0.00001476, -0.00001519, 0.00000761, -0.00001424, 0.00001010, -0.00000633, -0.00000635, 0.00000262, -0.00000774, },
		{ 0.00000634, 0.00001034, -0.00001997, -0.00000183, 0.00000712, -0.00000740, -0.00000228, -0.00001268, 0.00001830, -0.00000770, 0.00000281, 0.00000169, 0.00000129, -0.00001092, -0.00000518, 0.00000909, -0.00000373, -0.00001099, 0.00000225, 0.00001157, 0.00000370, -0.00000307, 0.00000603, -0.00001015, 0.00000478, -0.00000404, 0.00000452, -0.00000153, 0.00000503, -0.00000649, 0.00000172, -0.00000552, 0.00000654, -0.00000507, -0.00000046, -0.00001532, 0.00000127, -0.00001183, 0.00000764, -0.00000015, -0.00000074, -0.00002403, -0.00000214, -0.00010305, 0.00001773, 0.00013340, 0.00010597, 0.00008938, -0.00010968, -0.00001829, -0.00001012, 0.00000019, -0.00001593, -0.00000314, -0.00000978, -0.00000046, 0.00000329, 0.00000195, -0.00000374, 0.00000044, -0.00000465, 0.00000577, -0.00000334, 0.00000105, -0.00000098, -0.00000514, 0.00000034, -0.00000628, 0.00000171, 0.00000535, 0.00000274, -0.00001065, -0.00000417, 0.00000702, -0.00000381, 0.00000071, -0.00000366, 0.00000401, -0.00000030, -0.00000520, 0.00000619, 0.00000150, -0.00000424, -0.00000628, 0.00000011, -0.00000815, -0.00000145, -0.00000317, -0.00003408, -0.00001062, 0.00004205, 0.00003722, 0.00001316, 0.00005496, 0.00000108, -0.00002287, -0.00002819, 0.00000352, -0.00000368, 0.00001097, -0.00000987, 0.00000018, -0.00001239, -0.00000133, -0.00000224, 0.00000347, -0.00000398, 0.00000132, -0.00000351, 0.00000084, -0.00000270, 0.00000602, -0.00000177, 0.00000383, -0.00000534, 0.00000209, -0.00000091, 0.00000450, -0.00000431, 0.00001854, -0.00000937, 0.00000067, -0.00000535, -0.00000329, 0.00000807, -0.00001503, 0.00000555, -0.00000839, },
		{ -0.00000203, 0.00001459, -0.00001095, -0.00001497, 0.00001348, -0.00000886, 0.00000137, -0.00001570, 0.00001199, 0.00000240, -0.00000484, 0.00000575, -0.00000030, -0.00000401, -0.00001250, 0.00000728, 0.00000294, -0.00001134, -0.00000334, 0.00000890, 0.00000978, -0.00000500, 0.00000637, -0.00000634, -0.00000107, 0.00000047, -0.00000000, 0.00000179, 0.00000214, -0.00000137, -0.00000278, -0.00000196, 0.00000123, 0.00000124, -0.00000385, -0.00000886, -0.00000727, -0.00000354, -0.00000448, 0.00000848, -0.00000319, -0.00001478, -0.00000960, -0.00005639, -0.00007641, 0.00013827, 0.00010125, 0.00013638, -0.00007094, -0.00006192, 0.00000075, -0.00000801, -0.00000742, -0.00001132, -0.00000429, -0.00000716, 0.00000481, 0.00000206, -0.00000222, -0.00000063, -0.00000375, 0.00000212, 0.00000152, -0.00000282, 0.00000231, -0.00000581, -0.00000098, -0.00000314, -0.00000363, 0.00000647, 0.00000318, -0.00000351, -0.00001283, 0.00000880, -0.00000237, 0.00000035, -0.00000334, 0.00000141, 0.00000272, -0.00000492, 0.00000134, 0.00000660, -0.00000527, -0.00000397, -0.00000421, -0.00000099, -0.00001019, 0.00000683, -0.00003010, -0.00001895, 0.00001951, 0.00005349, 0.00000365, 0.00004880, 0.00002114, -0.00001305, -0.00003362, -0.00000398, -0.00000159, 0.00000737, -0.00000253, -0.00000458, -0.00000643, -0.00000769, 0.00000016, 0.00000065, -0.00000041, -0.00000154, -0.00000110, -0.00000095, -0.00000192, 0.00000321, 0.00000146, 0.00000156, -0.00000158, -0.00000175, 0.00000098, 0.00000297, -0.00000248, 0.00000978, 0.00000615, -0.00001144, 0.00000616, -0.00001474, 0.00001438, -0.00001294, -0.00000011, -0.00000312, },
		{ -0.00000731, 0.00001218, 0.00000137, -0.00002302, 0.00001057, -0.00000230, -0.00000261, -0.00000915, -0.00000204, 0.00001448, -0.00001082, 0.00000782, -0.00000160, 0.00000150, -0.00001415, 0.00000070, 0.00000777, -0.00000703, -0.00000934, 0.00000576, 0.00001114, 0.00000031, -0.00000029, 0.00000302, -0.00000947, 0.00000589, -0.00000505, 0.00000576, -0.00000229, 0.00000498, -0.00000771, 0.00000266, -0.00000547, 0.00000805, -0.00000843, -0.00000046, -0.00001644, 0.00000457, -0.00001452, 0.00001396, -0.00000709, -0.00000059, -0.00002601, -0.00001049, -0.00011963, 0.00007883, 0.00012665, 0.00014027, 0.00001113, -0.00010982, 0.00000592, -0.00001627, 0.00000174, -0.00001910, 0.00000052, -0.00001252, 0.00000516, 0.00000101, 0.00000123, -0.00000356, -0.00000042, -0.00000327, 0.00000582, -0.00000455, 0.00000265, -0.00000355, -0.00000391, -0.00000032, -0.00000593, 0.00000412, 0.00000436, 0.00000148, -0.00001381, 0.00000180, 0.00000449, -0.00000328, -0.00000010, -0.00000266, 0.00000458, -0.00000256, -0.00000346, 0.00000877, -0.00000350, -0.00000313, -0.00000686, 0.00000275, -0.00001275, 0.00000711, -0.00001913, -0.00002485, 0.00000113, 0.00005259, 0.00001524, 0.00002607, 0.00004213, -0.00000544, -0.00002513, -0.00002184, 0.00000650, -0.00000225, 0.00000843, -0.00001132, 0.00000092, -0.00001368, 0.00000235, -0.00000307, 0.00000372, -0.00000467, 0.00000178, -0.00000368, 0.00000091, -0.00000188, 0.00000602, -0.00000209, 0.00000344, -0.00000605, 0.00000372, -0.00000083, 0.00000333, -0.00000161, 0.00001787, -0.00001501, 0.00000745, -0.00001315, 0.00000622, -0.00000007, -0.00001027, 0.00000276, },
		{ -0.00000639, 0.00000490, 0.00001122, -0.00002134, -0.00000162, 0.00000886, -0.00000932, -0.00000019, -0.00001403, 0.00001917, -0.00000864, 0.00000391, 0.00000088, 0.00000212, -0.00000945, -0.00000755, 0.00000833, 0.00000025, -0.00001302, 0.00000162, 0.00000893, 0.00000806, -0.00000617, 0.00000860, -0.00001130, 0.00000454, -0.00000409, 0.00000438, -0.00000199, 0.00000572, -0.00000640, 0.00000170, -0.00000616, 0.00000754, -0.00000596, 0.00000039, -0.00001576, 0.00000166, -0.00001187, 0.00000669, -0.00000122, 0.00000180, -0.00002522, -0.00000088, -0.00009342, -0.00002501, 0.00015192, 0.00012787, 0.00009701, -0.00012373, -0.00001584, -0.00001192, 0.00000077, -0.00001757, -0.00000244, -0.00001113, 0.00000058, 0.00000214, 0.00000272, -0.00000424, 0.00000088, -0.00000550, 0.00000549, -0.00000202, 0.00000003, -0.00000024, -0.00000543, -0.00000074, -0.00000378, -0.00000131, 0.00000651, 0.00000257, -0.00000750, -0.00000932, 0.00001042, -0.00000478, 0.00000145, -0.00000418, 0.00000322, 0.00000115, -0.00000595, 0.00000647, 0.00000151, -0.00000455, -0.00000596, 0.00000075, -0.00000771, -0.00000108, -0.00000421, -0.00002977, -0.00000760, 0.00003411, 0.00003889, 0.00000213, 0.00005299, 0.00000392, -0.00001060, -0.00003642, 0.00000741, -0.00000587, 0.00001179, -0.00001033, 0.00000112, -0.00001220, -0.00000124, -0.00000247, 0.00000335, -0.00000353, 0.00000112, -0.00000349, 0.00000141, -0.00000393, 0.00000641, -0.00000173, 0.00000421, -0.00000542, 0.00000214, -0.00000101, 0.00000494, -0.00000501, 0.00001602, -0.00000439, -0.00000318, -0.00000092, -0.00000879, 0.00001257, -0.00001488, 0.00000248, },
		{ -0.00000125, -0.00000275, 0.00001455, -0.00001103, -0.00001634, 0.00001600, -0.00001049, 0.00000253, -0.00001577, 0.00001166, 0.00000249, -0.00000485, 0.00000615, -0.00000093, -0.00000187, -0.00001332, 0.00000452, 0.00000636, -0.00001097, -0.00000495, 0.00000677, 0.00001146, -0.00000404, 0.00000456, -0.00000325, -0.00000454, 0.00000284, -0.00000223, 0.00000337, 0.00000041, 0.00000107, -0.00000477, -0.00000048, -0.00000045, 0.00000348, -0.00000647, -0.00000564, -0.00001065, 0.00000036, -0.00000896, 0.00001134, -0.00000697, -0.00000735, -0.00001986, -0.00003510, -0.00010980, 0.00012771, 0.00013364, 0.00014793, -0.00007909, -0.00006788, 0.00000417, -0.00001102, -0.00000623, -0.00001314, -0.00000340, -0.00000837, 0.00000581, 0.00000076, -0.00000092, -0.00000172, -0.00000299, 0.00000042, 0.00000319, -0.00000322, 0.00000192, -0.00000457, -0.00000333, -0.00000055, -0.00000532, 0.00000618, 0.00000311, -0.00000056, -0.00001528, 0.00000781, 0.00000037, -0.00000155, -0.00000153, -0.00000105, 0.00000454, -0.00000531, 0.00000123, 0.00000688, -0.00000566, -0.00000377, -0.00000381, -0.00000019, -0.00001015, 0.00000683, -0.00002956, -0.00001149, 0.00001132, 0.00005453, -0.00000375, 0.00004304, 0.00002126, -0.00000217, -0.00003377, -0.00000781, 0.00000235, 0.00000347, 0.00000055, -0.00000689, -0.00000338, -0.00000983, 0.00000205, -0.00000143, 0.00000155, -0.00000290, -0.00000006, -0.00000157, -0.00000130, 0.00000143, 0.00000330, 0.00000018, 0.00000022, -0.00000357, 0.00000282, 0.00000112, 0.00000004, 0.00000441, 0.00001148, -0.00001396, 0.00000839, -0.00001612, 0.00001294, -0.00000760, -0.00000519, },
		{ 0.00000172, -0.00000594, 0.00001074, 0.00000233, -0.00002494, 0.00001236, -0.00000237, -0.00000313, -0.00000768, -0.00000339, 0.00001527, -0.00001155, 0.00000892, -0.00000280, 0.00000317, -0.00001347, -0.00000198, 0.00000819, -0.00000341, -0.00001204, 0.00000508, 0.00000931, 0.00000465, -0.00000436, 0.00000722, -0.00001266, 0.00000747, -0.00000660, 0.00000690, -0.00000391, 0.00000703, -0.00000907, 0.00000389, -0.00000729, 0.00001046, -0.00001102, 0.00000239, -0.00001908, 0.00000742, -0.00001732, 0.00001554, -0.00001074, 0.00000600, -0.00003324, 0.00000053, -0.00012378, 0.00003880, 0.00015755, 0.00015989, 0.00001008, -0.00012213, 0.00001094, -0.00001997, 0.00000365, -0.00002185, 0.00000235, -0.00001478, 0.00000710, -0.00000110, 0.00000302, -0.00000494, 0.00000067, -0.00000491, 0.00000660, -0.00000391, 0.00000169, -0.00000238, -0.00000502, -0.00000030, -0.00000446, 0.00000147, 0.00000554, 0.00000207, -0.00001185, -0.00000326, 0.00000900, -0.00000559, 0.00000184, -0.00000452, 0.00000503, -0.00000183, -0.00000396, 0.00000898, -0.00000354, -0.00000339, -0.00000634, 0.00000342, -0.00001260, 0.00000763, -0.00002038, -0.00001761, -0.00000202, 0.00004849, 0.00001442, 0.00001580, 0.00004243, -0.00000116, -0.00001566, -0.00003070, 0.00001247, -0.00000667, 0.00001132, -0.00001354, 0.00000360, -0.00001509, 0.00000371, -0.00000454, 0.00000484, -0.00000521, 0.00000231, -0.00000423, 0.00000199, -0.00000392, 0.00000752, -0.00000309, 0.00000487, -0.00000728, 0.00000486, -0.00000200, 0.00000510, -0.00000460, 0.00001832, -0.00001220, 0.00000465, -0.00000923, 0.00000040, 0.00000646, -0.00001312, },
		{ -0.00000231, -0.00000347, 0.00000305, 0.00001209, -0.00002259, -0.00000139, 0.00001038, -0.00001085, 0.00000170, -0.00001512, 0.00001951, -0.00000908, 0.00000470, 0.00000021, 0.00000259, -0.00000784, -0.00000878, 0.00000583, 0.00000485, -0.00001436, 0.00000068, 0.00000622, 0.00001158, -0.00000773, 0.00000930, -0.00001038, 0.00000251, -0.00000274, 0.00000303, -0.00000132, 0.00000513, -0.00000508, 0.00000064, -0.00000571, 0.00000722, -0.00000529, -0.00000055, -0.00001419, -0.00000002, -0.00000968, 0.00000361, -0.00000009, 0.00000106, -0.00002158, -0.00000684, -0.00007371, -0.00007064, 0.00015862, 0.00016160, 0.00010109, -0.00013587, -0.00001559, -0.00001198, 0.00000018, -0.00001849, -0.00000225, -0.00001194, 0.00000100, 0.00000156, 0.00000286, -0.00000407, 0.00000067, -0.00000569, 0.00000467, -0.00000058, -0.00000066, -0.00000023, -0.00000450, -0.00000307, -0.00000068, -0.00000419, 0.00000739, 0.00000206, -0.00000379, -0.00001392, 0.00001205, -0.00000397, 0.00000089, -0.00000351, 0.00000142, 0.00000295, -0.00000657, 0.00000652, 0.00000169, -0.00000494, -0.00000562, 0.00000140, -0.00000742, -0.00000051, -0.00000533, -0.00002586, -0.00000413, 0.00002561, 0.00004186, -0.00000858, 0.00005156, 0.00000385, 0.00000133, -0.00004120, 0.00000788, -0.00000511, 0.00001024, -0.00000902, 0.00000039, -0.00001040, -0.00000249, -0.00000148, 0.00000208, -0.00000209, 0.00000015, -0.00000288, 0.00000139, -0.00000425, 0.00000566, -0.00000066, 0.00000355, -0.00000436, 0.00000091, 0.00000024, 0.00000389, -0.00000371, 0.00001141, 0.00000179, -0.00000764, 0.00000352, -0.00001285, 0.00001439, -0.00001198, },
		{ -0.00001051, 0.00000026, -0.00000296, 0.00001401, -0.00001090, -0.00001756, 0.00001815, -0.00001173, 0.00000354, -0.00001557, 0.00001093, 0.00000275, -0.00000479, 0.00000640, -0.00000160, -0.00000010, -0.00001306, 0.00000129, 0.00000857, -0.00000870, -0.00000753, 0.00000552, 0.00001128, -0.00000109, 0.00000109, 0.00000114, -0.00000865, 0.00000541, -0.00000459, 0.00000513, -0.00000165, 0.00000370, -0.00000684, 0.00000127, -0.00000261, 0.00000630, -0.00000967, -0.00000189, -0.00001436, 0.00000451, -0.00001342, 0.00001432, -0.00001160, 0.00000121, -0.00003101, -0.00001520, -0.00013152, 0.00009762, 0.00017539, 0.00016041, -0.00008752, -0.00007458, 0.00000789, -0.00001404, -0.00000507, -0.00001509, -0.00000212, -0.00001008, 0.00000727, -0.00000107, 0.00000086, -0.00000317, -0.00000198, -0.00000130, 0.00000451, -0.00000310, 0.00000103, -0.00000294, -0.00000542, 0.00000089, -0.00000548, 0.00000466, 0.00000371, 0.00000134, -0.00001549, 0.00000425, 0.00000470, -0.00000419, 0.00000084, -0.00000371, 0.00000601, -0.00000531, 0.00000105, 0.00000693, -0.00000566, -0.00000404, -0.00000289, -0.00000000, -0.00000960, 0.00000650, -0.00002900, -0.00000436, 0.00000365, 0.00005516, -0.00000916, 0.00003642, 0.00002170, 0.00000432, -0.00002861, -0.00001482, 0.00000811, -0.00000139, 0.00000412, -0.00000972, 0.00000009, -0.00001210, 0.00000406, -0.00000356, 0.00000345, -0.00000409, 0.00000088, -0.00000230, -0.00000023, -0.00000081, 0.00000535, -0.00000140, 0.00000224, -0.00000559, 0.00000474, -0.00000075, 0.00000261, -0.00000039, 0.00001502, -0.00001425, 0.00000795, -0.00001409, 0.00000832, -0.00000045, },
		{ -0.00001383, -0.00000104, -0.00000340, 0.00000863, 0.00000345, -0.00002671, 0.00001401, -0.00000246, -0.00000339, -0.00000634, -0.00000461, 0.00001564, -0.00001177, 0.00000963, -0.00000383, 0.00000439, -0.00001229, -0.00000391, 0.00000668, 0.00000157, -0.00001472, 0.00000456, 0.00000668, 0.00000917, -0.00000786, 0.00001033, -0.00001423, 0.00000740, -0.00000674, 0.00000683, -0.00000445, 0.00000783, -0.00000918, 0.00000407, -0.00000812, 0.00001170, -0.00001220, 0.00000358, -0.00001973, 0.00000811, -0.00001767, 0.00001474, -0.00001208, 0.00000933, -0.00003544, 0.00000297, -0.00011278, -0.00001344, 0.00018422, 0.00018964, 0.00000464, -0.00013221, 0.00001374, -0.00002184, 0.00000426, -0.00002378, 0.00000371, -0.00001665, 0.00000860, -0.00000285, 0.00000438, -0.00000581, 0.00000123, -0.00000594, 0.00000671, -0.00000289, 0.00000077, -0.00000171, -0.00000491, -0.00000191, -0.00000168, -0.00000180, 0.00000703, 0.00000174, -0.00000834, -0.00000907, 0.00001286, -0.00000677, 0.00000287, -0.00000539, 0.00000444, -0.00000057, -0.00000443, 0.00000888, -0.00000322, -0.00000399, -0.00000550, 0.00000374, -0.00001224, 0.00000802, -0.00002149, -0.00001111, -0.00000418, 0.00004336, 0.00001567, 0.00000488, 0.00004380, -0.00000087, -0.00000424, -0.00003834, 0.00001640, -0.00000892, 0.00001227, -0.00001424, 0.00000480, -0.00001495, 0.00000374, -0.00000477, 0.00000476, -0.00000465, 0.00000198, -0.00000417, 0.00000258, -0.00000523, 0.00000801, -0.00000312, 0.00000528, -0.00000739, 0.00000469, -0.00000177, 0.00000531, -0.00000529, 0.00001585, -0.00000689, -0.00000033, -0.00000350, -0.00000583, 0.00001170, },
		{ -0.00000628, -0.00000845, -0.00000005, 0.00000114, 0.00001272, -0.00002366, -0.00000103, 0.00001148, -0.00001181, 0.00000321, -0.00001578, 0.00001923, -0.00000904, 0.00000527, -0.00000042, 0.00000281, -0.00000631, -0.00000891, 0.00000225, 0.00000896, -0.00001406, -0.00000122, 0.00000420, 0.00001343, -0.00000710, 0.00000775, -0.00000728, -0.00000118, -0.00000021, 0.00000072, 0.00000029, 0.00000343, -0.00000273, -0.00000129, -0.00000426, 0.00000561, -0.00000311, -0.00000313, -0.00001087, -0.00000343, -0.00000570, -0.00000105, 0.00000279, -0.00000251, -0.00001400, -0.00001801, -0.00004868, -0.00011119, 0.00014639, 0.00020985, 0.00010136, -0.00014598, -0.00001758, -0.00001044, -0.00000143, -0.00001869, -0.00000267, -0.00001200, 0.00000060, 0.00000176, 0.00000216, -0.00000310, -0.00000024, -0.00000528, 0.00000345, 0.00000081, -0.00000093, -0.00000087, -0.00000272, -0.00000586, 0.00000205, -0.00000606, 0.00000733, 0.00000187, -0.00000049, -0.00001680, 0.00001107, -0.00000108, -0.00000099, -0.00000170, -0.00000118, 0.00000486, -0.00000695, 0.00000639, 0.00000182, -0.00000512, -0.00000556, 0.00000235, -0.00000756, 0.00000043, -0.00000661, -0.00002232, -0.00000041, 0.00001713, 0.00004511, -0.00001795, 0.00004953, 0.00000324, 0.00001012, -0.00004086, 0.00000428, -0.00000130, 0.00000649, -0.00000609, -0.00000186, -0.00000721, -0.00000478, 0.00000045, -0.00000002, 0.00000001, -0.00000127, -0.00000187, 0.00000083, -0.00000369, 0.00000387, 0.00000126, 0.00000203, -0.00000237, -0.00000128, 0.00000240, 0.00000191, -0.00000117, 0.00000584, 0.00000775, -0.00001120, 0.00000629, -0.00001397, 0.00001270, },
		{ 0.00000720, -0.00001491, 0.00000063, -0.00000220, 0.00001277, -0.00001055, -0.00001857, 0.00001992, -0.00001264, 0.00000447, -0.00001524, 0.00001005, 0.00000290, -0.00000441, 0.00000636, -0.00000220, 0.00000129, -0.00001202, -0.00000170, 0.00000893, -0.00000452, -0.00001075, 0.00000491, 0.00000962, 0.00000312, -0.00000308, 0.00000567, -0.00001217, 0.00000711, -0.00000613, 0.00000621, -0.00000317, 0.00000564, -0.00000816, 0.00000249, -0.00000438, 0.00000868, -0.00001226, 0.00000105, -0.00001696, 0.00000733, -0.00001621, 0.00001570, -0.00001499, 0.00000806, -0.00003909, -0.00000215, -0.00013624, 0.00004704, 0.00022278, 0.00017787, -0.00009880, -0.00008008, 0.00001031, -0.00001587, -0.00000477, -0.00001656, -0.00000097, -0.00001179, 0.00000873, -0.00000295, 0.00000261, -0.00000450, -0.00000116, -0.00000265, 0.00000524, -0.00000245, -0.00000014, -0.00000144, -0.00000658, 0.00000071, -0.00000396, 0.00000200, 0.00000501, 0.00000201, -0.00001346, -0.00000129, 0.00000964, -0.00000668, 0.00000304, -0.00000593, 0.00000670, -0.00000475, 0.00000079, 0.00000669, -0.00000520, -0.00000479, -0.00000148, -0.00000038, -0.00000860, 0.00000590, -0.00002832, 0.00000204, -0.00000280, 0.00005449, -0.00001194, 0.00002826, 0.00002366, 0.00000602, -0.00001939, -0.00002310, 0.00001387, -0.00000566, 0.00000693, -0.00001197, 0.00000290, -0.00001353, 0.00000530, -0.00000495, 0.00000457, -0.00000452, 0.00000124, -0.00000272, 0.00000077, -0.00000286, 0.00000692, -0.00000249, 0.00000375, -0.00000696, 0.00000582, -0.00000172, 0.00000420, -0.00000341, 0.00001573, -0.00001156, 0.00000438, -0.00000883, 0.00000145, },
		{ 0.00001415, -0.00001162, -0.00000573, 0.00000024, 0.00000599, 0.00000459, -0.00002830, 0.00001561, -0.00000279, -0.00000314, -0.00000541, -0.00000538, 0.00001534, -0.00001134, 0.00000992, -0.00000468, 0.00000518, -0.00001083, -0.00000490, 0.00000361, 0.00000689, -0.00001633, 0.00000339, 0.00000422, 0.00001268, -0.00000963, 0.00001127, -0.00001334, 0.00000513, -0.00000509, 0.00000521, -0.00000352, 0.00000702, -0.00000770, 0.00000285, -0.00000745, 0.00001116, -0.00001131, 0.00000242, -0.00001776, 0.00000601, -0.00001500, 0.00001101, -0.00001034, 0.00000830, -0.00003144, -0.00000381, -0.00008865, -0.00007114, 0.00019829, 0.00023474, -0.00000759, -0.00013847, 0.00001313, -0.00002110, 0.00000303, -0.00002439, 0.00000407, -0.00001754, 0.00000912, -0.00000367, 0.00000478, -0.00000574, 0.00000090, -0.00000612, 0.00000611, -0.00000157, 0.00000001, -0.00000166, -0.00000367, -0.00000462, 0.00000156, -0.00000482, 0.00000813, 0.00000114, -0.00000428, -0.00001425, 0.00001475, -0.00000598, 0.00000248, -0.00000492, 0.00000269, 0.00000118, -0.00000486, 0.00000856, -0.00000273, -0.00000465, -0.00000463, 0.00000402, -0.00001198, 0.00000853, -0.00002252, -0.00000545, -0.00000526, 0.00003734, 0.00001838, -0.00000596, 0.00004544, -0.00000254, 0.00000615, -0.00004225, 0.00001663, -0.00000797, 0.00001059, -0.00001284, 0.00000398, -0.00001290, 0.00000216, -0.00000354, 0.00000332, -0.00000293, 0.00000076, -0.00000339, 0.00000244, -0.00000549, 0.00000718, -0.00000192, 0.00000444, -0.00000611, 0.00000303, -0.00000006, 0.00000392, -0.00000382, 0.00001099, -0.00000002, -0.00000623, 0.00000231, -0.00001056, },
		{ 0.00000750, 0.00000155, -0.00001506, 0.00000315, -0.00000029, 0.00001285, -0.00002450, -0.00000047, 0.00001204, -0.00001214, 0.00000429, -0.00001599, 0.00001843, -0.00000871, 0.00000581, -0.00000111, 0.00000286, -0.00000493, -0.00000818, -0.00000160, 0.00001156, -0.00001158, -0.00000427, 0.00000312, 0.00001339, -0.00000435, 0.00000429, -0.00000263, -0.00000568, 0.00000265, -0.00000182, 0.00000217, 0.00000131, -0.00000008, -0.00000344, -0.00000238, 0.00000333, -0.00000016, -0.00000648, -0.00000681, -0.00000739, -0.00000130, -0.00000584, 0.00000595, -0.00000714, -0.00000500, -0.00003046, -0.00002487, -0.00013822, 0.00010966, 0.00027285, 0.00009964, -0.00015545, -0.00002069, -0.00000837, -0.00000323, -0.00001871, -0.00000340, -0.00001152, -0.00000045, 0.00000258, 0.00000081, -0.00000156, -0.00000159, -0.00000454, 0.00000213, 0.00000194, -0.00000075, -0.00000197, -0.00000064, -0.00000825, 0.00000351, -0.00000632, 0.00000596, 0.00000243, 0.00000169, -0.00001726, 0.00000731, 0.00000355, -0.00000370, 0.00000076, -0.00000409, 0.00000648, -0.00000694, 0.00000615, 0.00000175, -0.00000491, -0.00000594, 0.00000377, -0.00000827, 0.00000183, -0.00000809, -0.00001903, 0.00000314, 0.00000945, 0.00004748, -0.00002484, 0.00004568, 0.00000407, 0.00001409, -0.00003513, -0.00000267, 0.00000454, 0.00000154, -0.00000242, -0.00000485, -0.00000346, -0.00000731, 0.00000256, -0.00000224, 0.00000208, -0.00000256, -0.00000091, 0.00000008, -0.00000259, 0.00000154, 0.00000350, 0.00000022, -0.00000009, -0.00000368, 0.00000458, -0.00000008, 0.00000147, 0.00000076, 0.00001197, -0.00001238, 0.00000603, -0.00001130, },
		{ -0.00000597, 0.00001363, -0.00001713, -0.00000081, -0.00000016, 0.00001075, -0.00000998, -0.00001937, 0.00002138, -0.00001342, 0.00000558, -0.00001509, 0.00000937, 0.00000256, -0.00000346, 0.00000588, -0.00000263, 0.00000226, -0.00001048, -0.00000399, 0.00000739, 0.00000093, -0.00001373, 0.00000418, 0.00000737, 0.00000741, -0.00000661, 0.00000893, -0.00001379, 0.00000684, -0.00000594, 0.00000577, -0.00000330, 0.00000605, -0.00000794, 0.00000235, -0.00000482, 0.00000951, -0.00001301, 0.00000184, -0.00001700, 0.00000732, -0.00001578, 0.00001388, -0.00001518, 0.00001054, -0.00004068, 0.00000006, -0.00012166, -0.00002096, 0.00026871, 0.00020625, -0.00011647, -0.00008184, 0.00000942, -0.00001505, -0.00000637, -0.00001666, -0.00000075, -0.00001272, 0.00000941, -0.00000413, 0.00000362, -0.00000506, -0.00000111, -0.00000320, 0.00000516, -0.00000127, -0.00000141, -0.00000035, -0.00000657, -0.00000100, -0.00000124, -0.00000122, 0.00000653, 0.00000180, -0.00000983, -0.00000761, 0.00001383, -0.00000792, 0.00000423, -0.00000707, 0.00000621, -0.00000350, 0.00000044, 0.00000621, -0.00000443, -0.00000582, 0.00000021, -0.00000109, -0.00000741, 0.00000521, -0.00002752, 0.00000742, -0.00000757, 0.00005212, -0.00001208, 0.00001867, 0.00002710, 0.00000409, -0.00000858, -0.00003007, 0.00001753, -0.00000771, 0.00000773, -0.00001255, 0.00000403, -0.00001323, 0.00000502, -0.00000493, 0.00000431, -0.00000372, 0.00000065, -0.00000246, 0.00000121, -0.00000410, 0.00000738, -0.00000247, 0.00000409, -0.00000700, 0.00000540, -0.00000113, 0.00000412, -0.00000402, 0.00001329, -0.00000597, -0.00000181, -0.00000147, },
		{ -0.00001213, 0.00001284, -0.00000632, -0.00001202, 0.00000458, 0.00000320, 0.00000554, -0.00002966, 0.00001729, -0.00000357, -0.00000219, -0.00000511, -0.00000552, 0.00001425, -0.00001027, 0.00000985, -0.00000536, 0.00000558, -0.00000923, -0.00000506, -0.00000027, 0.00001137, -0.00001590, 0.00000090, 0.00000268, 0.00001434, -0.00000895, 0.00000958, -0.00000982, 0.00000080, -0.00000190, 0.00000227, -0.00000131, 0.00000476, -0.00000485, 0.00000036, -0.00000533, 0.00000884, -0.00000837, -0.00000097, -0.00001336, 0.00000145, -0.00000978, 0.00000489, -0.00000594, 0.00000336, -0.00002212, -0.00001767, -0.00005675, -0.00012487, 0.00019073, 0.00029943, -0.00002771, -0.00014038, 0.00000880, -0.00001772, -0.00000005, -0.00002355, 0.00000317, -0.00001714, 0.00000830, -0.00000325, 0.00000392, -0.00000451, -0.00000043, -0.00000543, 0.00000485, -0.00000004, -0.00000055, -0.00000210, -0.00000176, -0.00000756, 0.00000420, -0.00000667, 0.00000812, 0.00000101, -0.00000075, -0.00001751, 0.00001376, -0.00000286, 0.00000057, -0.00000314, -0.00000008, 0.00000322, -0.00000515, 0.00000813, -0.00000234, -0.00000506, -0.00000407, 0.00000460, -0.00001211, 0.00000936, -0.00002357, -0.00000056, -0.00000553, 0.00003106, 0.00002139, -0.00001545, 0.00004592, -0.00000375, 0.00001283, -0.00004081, 0.00001254, -0.00000370, 0.00000642, -0.00000948, 0.00000130, -0.00000918, -0.00000069, -0.00000119, 0.00000088, -0.00000043, -0.00000103, -0.00000210, 0.00000165, -0.00000472, 0.00000516, 0.00000032, 0.00000254, -0.00000371, 0.00000027, 0.00000264, 0.00000152, -0.00000096, 0.00000497, 0.00000689, -0.00001133, 0.00000633, },
		{ -0.00000501, -0.00000024, 0.00001009, -0.00002086, 0.00000523, -0.00000071, 0.00001221, -0.00002503, 0.00000030, 0.00001212, -0.00001195, 0.00000508, -0.00001596, 0.00001740, -0.00000844, 0.00000662, -0.00000204, 0.00000287, -0.00000376, -0.00000686, -0.00000504, 0.00001214, -0.00000703, -0.00000799, 0.00000261, 0.00001189, -0.00000025, -0.00000002, 0.00000227, -0.00000961, 0.00000459, -0.00000347, 0.00000329, -0.00000020, 0.00000183, -0.00000480, -0.00000105, 0.00000145, 0.00000233, -0.00000920, -0.00000360, -0.00001013, 0.00000161, -0.00000871, 0.00000725, -0.00001029, 0.00000196, -0.00003933, -0.00000907, -0.00014469, 0.00004612, 0.00034718, 0.00010046, -0.00016744, -0.00002248, -0.00000776, -0.00000373, -0.00001962, -0.00000360, -0.00001122, -0.00000148, 0.00000337, -0.00000056, -0.00000010, -0.00000276, -0.00000404, 0.00000118, 0.00000250, -0.00000004, -0.00000335, 0.00000129, -0.00000964, 0.00000331, -0.00000488, 0.00000340, 0.00000373, 0.00000258, -0.00001532, 0.00000139, 0.00000887, -0.00000628, 0.00000307, -0.00000659, 0.00000730, -0.00000630, 0.00000576, 0.00000143, -0.00000428, -0.00000676, 0.00000561, -0.00000947, 0.00000358, -0.00000968, -0.00001589, 0.00000611, 0.00000332, 0.00004801, -0.00002853, 0.00003933, 0.00000730, 0.00001310, -0.00002536, -0.00001107, 0.00001054, -0.00000297, 0.00000068, -0.00000738, -0.00000032, -0.00000900, 0.00000389, -0.00000368, 0.00000333, -0.00000306, -0.00000052, -0.00000037, -0.00000152, -0.00000066, 0.00000529, -0.00000111, 0.00000167, -0.00000537, 0.00000583, -0.00000109, 0.00000311, -0.00000255, 0.00001328, -0.00001028, 0.00000220, },
		{ 0.00000610, -0.00001187, 0.00001716, -0.00001622, -0.00000442, 0.00000322, 0.00000805, -0.00000927, -0.00001999, 0.00002272, -0.00001435, 0.00000717, -0.00001545, 0.00000926, 0.00000145, -0.00000177, 0.00000489, -0.00000282, 0.00000275, -0.00000856, -0.00000549, 0.00000447, 0.00000646, -0.00001527, 0.00000243, 0.00000558, 0.00001049, -0.00000816, 0.00000969, -0.00001251, 0.00000390, -0.00000352, 0.00000333, -0.00000153, 0.00000440, -0.00000567, 0.00000032, -0.00000325, 0.00000797, -0.00001104, -0.00000044, -0.00001360, 0.00000358, -0.00001129, 0.00000799, -0.00001100, 0.00000713, -0.00003407, -0.00000985, -0.00008943, -0.00009906, 0.00030326, 0.00025245, -0.00014415, -0.00007743, 0.00000344, -0.00001036, -0.00001080, -0.00001458, -0.00000230, -0.00001201, 0.00000850, -0.00000384, 0.00000315, -0.00000420, -0.00000233, -0.00000257, 0.00000403, 0.00000053, -0.00000279, 0.00000033, -0.00000562, -0.00000359, 0.00000172, -0.00000401, 0.00000749, 0.00000146, -0.00000574, -0.00001316, 0.00001583, -0.00000694, 0.00000378, -0.00000667, 0.00000432, -0.00000153, -0.00000006, 0.00000565, -0.00000362, -0.00000679, 0.00000179, -0.00000174, -0.00000640, 0.00000471, -0.00002670, 0.00001175, -0.00001065, 0.00004825, -0.00001030, 0.00000864, 0.00003090, 0.00000076, 0.00000078, -0.00003320, 0.00001736, -0.00000640, 0.00000575, -0.00001081, 0.00000288, -0.00001076, 0.00000292, -0.00000325, 0.00000253, -0.00000161, -0.00000094, -0.00000137, 0.00000083, -0.00000419, 0.00000639, -0.00000107, 0.00000300, -0.00000546, 0.00000328, 0.00000108, 0.00000233, -0.00000231, 0.00000821, 0.00000156, -0.00000924, },
		{ 0.00000701, -0.00000954, 0.00000765, 0.00000144, -0.00001905, 0.00000885, 0.00000083, 0.00000599, -0.00003072, 0.00001910, -0.00000488, -0.00000049, -0.00000544, -0.00000501, 0.00001249, -0.00000878, 0.00000962, -0.00000598, 0.00000564, -0.00000752, -0.00000463, -0.00000413, 0.00001400, -0.00001290, -0.00000304, 0.00000229, 0.00001391, -0.00000587, 0.00000559, -0.00000435, -0.00000471, 0.00000193, -0.00000118, 0.00000145, 0.00000180, -0.00000138, -0.00000268, -0.00000242, 0.00000543, -0.00000420, -0.00000564, -0.00000767, -0.00000424, -0.00000353, -0.00000195, -0.00000060, -0.00000348, -0.00001035, -0.00003409, -0.00002479, -0.00016423, 0.00015378, 0.00038538, -0.00005443, -0.00013922, 0.00000188, -0.00001275, -0.00000413, -0.00002179, 0.00000133, -0.00001566, 0.00000634, -0.00000175, 0.00000200, -0.00000237, -0.00000248, -0.00000415, 0.00000321, 0.00000148, -0.00000089, -0.00000281, 0.00000025, -0.00000985, 0.00000536, -0.00000670, 0.00000659, 0.00000181, 0.00000149, -0.00001809, 0.00000965, 0.00000226, -0.00000237, -0.00000051, -0.00000335, 0.00000513, -0.00000515, 0.00000764, -0.00000222, -0.00000499, -0.00000403, 0.00000566, -0.00001281, 0.00001062, -0.00002471, 0.00000371, -0.00000545, 0.00002539, 0.00002343, -0.00002225, 0.00004388, -0.00000257, 0.00001426, -0.00003383, 0.00000481, 0.00000288, 0.00000077, -0.00000505, -0.00000240, -0.00000469, -0.00000395, 0.00000148, -0.00000179, 0.00000214, -0.00000275, -0.00000077, 0.00000057, -0.00000330, 0.00000245, 0.00000301, 0.00000023, -0.00000090, -0.00000278, 0.00000538, -0.00000091, 0.00000208, -0.00000069, 0.00001215, -0.00001395, },
		{ -0.00000469, 0.00000368, -0.00000873, 0.00001751, -0.00002443, 0.00000534, 0.00000036, 0.00001062, -0.00002517, 0.00000118, 0.00001187, -0.00001152, 0.00000591, -0.00001605, 0.00001658, -0.00000864, 0.00000799, -0.00000338, 0.00000298, -0.00000291, -0.00000510, -0.00000775, 0.00001077, -0.00000123, -0.00001133, 0.00000181, 0.00000996, 0.00000389, -0.00000366, 0.00000581, -0.00001144, 0.00000430, -0.00000313, 0.00000263, -0.00000006, 0.00000195, -0.00000435, -0.00000135, 0.00000118, 0.00000294, -0.00000970, -0.00000297, -0.00000984, 0.00000109, -0.00000765, 0.00000456, -0.00000933, 0.00000345, -0.00004020, -0.00000664, -0.00012685, -0.00004195, 0.00042534, 0.00011085, -0.00018658, -0.00001961, -0.00001123, -0.00000097, -0.00002290, -0.00000204, -0.00001221, -0.00000142, 0.00000311, -0.00000096, 0.00000035, -0.00000290, -0.00000451, 0.00000119, 0.00000207, 0.00000137, -0.00000499, 0.00000288, -0.00000989, 0.00000164, -0.00000230, 0.00000032, 0.00000519, 0.00000262, -0.00001173, -0.00000536, 0.00001339, -0.00000750, 0.00000430, -0.00000793, 0.00000682, -0.00000482, 0.00000512, 0.00000099, -0.00000343, -0.00000776, 0.00000759, -0.00001087, 0.00000538, -0.00001117, -0.00001294, 0.00000826, -0.00000085, 0.00004631, -0.00002899, 0.00003073, 0.00001261, 0.00000854, -0.00001412, -0.00001825, 0.00001452, -0.00000534, 0.00000184, -0.00000824, 0.00000104, -0.00000886, 0.00000360, -0.00000361, 0.00000311, -0.00000225, -0.00000114, -0.00000011, -0.00000103, -0.00000205, 0.00000594, -0.00000127, 0.00000217, -0.00000560, 0.00000540, -0.00000043, 0.00000300, -0.00000334, 0.00001122, -0.00000490, },
		{ -0.00001600, 0.00001175, -0.00001437, 0.00001662, -0.00001184, -0.00001000, 0.00000757, 0.00000502, -0.00000863, -0.00002043, 0.00002410, -0.00001567, 0.00000946, -0.00001655, 0.00000995, -0.00000060, 0.00000069, 0.00000342, -0.00000274, 0.00000272, -0.00000628, -0.00000643, 0.00000102, 0.00001078, -0.00001431, -0.00000109, 0.00000506, 0.00001142, -0.00000693, 0.00000734, -0.00000804, -0.00000166, 0.00000100, -0.00000095, 0.00000205, 0.00000077, -0.00000146, -0.00000358, 0.00000042, 0.00000394, -0.00000622, -0.00000583, -0.00000679, -0.00000371, -0.00000305, -0.00000158, -0.00000271, -0.00000198, -0.00001988, -0.00002993, -0.00004526, -0.00017645, 0.00031521, 0.00032281, -0.00018429, -0.00006554, -0.00000849, -0.00000135, -0.00001840, -0.00000989, -0.00000612, -0.00000912, 0.00000547, -0.00000156, 0.00000074, -0.00000157, -0.00000507, -0.00000060, 0.00000180, 0.00000296, -0.00000435, 0.00000084, -0.00000429, -0.00000608, 0.00000377, -0.00000533, 0.00000705, 0.00000184, -0.00000238, -0.00001650, 0.00001455, -0.00000326, 0.00000152, -0.00000466, 0.00000110, 0.00000103, -0.00000067, 0.00000517, -0.00000308, -0.00000731, 0.00000287, -0.00000193, -0.00000593, 0.00000468, -0.00002601, 0.00001516, -0.00001237, 0.00004359, -0.00000785, -0.00000033, 0.00003334, -0.00000143, 0.00000605, -0.00003085, 0.00001264, -0.00000155, 0.00000104, -0.00000681, -0.00000045, -0.00000635, -0.00000068, -0.00000024, -0.00000043, 0.00000143, -0.00000321, 0.00000033, -0.00000031, -0.00000312, 0.00000406, 0.00000155, 0.00000068, -0.00000261, -0.00000010, 0.00000436, -0.00000056, 0.00000092, 0.00000173, 0.00000938, },
		{ -0.00001302, 0.00000366, -0.00000291, -0.00000044, 0.00001025, -0.00002544, 0.00001203, -0.00000047, 0.00000560, -0.00003134, 0.00002099, -0.00000662, 0.00000176, -0.00000622, -0.00000413, 0.00001040, -0.00000725, 0.00000955, -0.00000671, 0.00000549, -0.00000580, -0.00000386, -0.00000734, 0.00001430, -0.00000753, -0.00000784, 0.00000257, 0.00001194, -0.00000124, 0.00000046, 0.00000165, -0.00000984, 0.00000499, -0.00000386, 0.00000356, -0.00000069, 0.00000152, -0.00000511, 0.00000013, 0.00000219, -0.00000024, -0.00000991, -0.00000259, -0.00000899, 0.00000146, -0.00000709, 0.00000310, -0.00000910, -0.00000032, -0.00004720, -0.00000117, -0.00017997, 0.00008306, 0.00049022, -0.00008326, -0.00013852, -0.00000484, -0.00000852, -0.00000745, -0.00002042, -0.00000041, -0.00001396, 0.00000405, 0.00000005, -0.00000022, -0.00000006, -0.00000454, -0.00000292, 0.00000175, 0.00000261, -0.00000085, -0.00000366, 0.00000194, -0.00001093, 0.00000465, -0.00000486, 0.00000367, 0.00000349, 0.00000228, -0.00001599, 0.00000308, 0.00000830, -0.00000533, 0.00000213, -0.00000637, 0.00000634, -0.00000458, 0.00000703, -0.00000238, -0.00000445, -0.00000446, 0.00000715, -0.00001398, 0.00001219, -0.00002586, 0.00000748, -0.00000546, 0.00002109, 0.00002348, -0.00002550, 0.00003857, 0.00000182, 0.00001048, -0.00002271, -0.00000461, 0.00000987, -0.00000468, -0.00000094, -0.00000586, -0.00000068, -0.00000645, 0.00000343, -0.00000372, 0.00000389, -0.00000369, 0.00000001, -0.00000027, -0.00000185, -0.00000021, 0.00000534, -0.00000165, 0.00000141, -0.00000510, 0.00000712, -0.00000230, 0.00000410, -0.00000458, 0.00001442, },
		{ 0.00000279, -0.00001317, 0.00001272, -0.00001570, 0.00002191, -0.00002462, 0.00000295, 0.00000307, 0.00000810, -0.00002493, 0.00000207, 0.00001158, -0.00001124, 0.00000716, -0.00001668, 0.00001642, -0.00000971, 0.00001017, -0.00000528, 0.00000335, -0.00000254, -0.00000290, -0.00000978, 0.00000811, 0.00000447, -0.00001293, -0.00000038, 0.00000876, 0.00000661, -0.00000512, 0.00000652, -0.00000993, 0.00000088, -0.00000007, -0.00000051, 0.00000246, -0.00000047, -0.00000135, -0.00000408, 0.00000350, 0.00000056, -0.00000677, -0.00000619, -0.00000520, -0.00000420, -0.00000136, -0.00000350, -0.00000253, -0.00000279, -0.00003041, -0.00002000, -0.00008538, -0.00014719, 0.00049614, 0.00013954, -0.00021800, -0.00000857, -0.00002140, 0.00000694, -0.00003000, 0.00000258, -0.00001573, 0.00000096, 0.00000064, 0.00000069, -0.00000124, -0.00000111, -0.00000669, 0.00000277, 0.00000025, 0.00000374, -0.00000704, 0.00000431, -0.00000939, -0.00000070, 0.00000029, -0.00000214, 0.00000590, 0.00000269, -0.00000774, -0.00001121, 0.00001548, -0.00000626, 0.00000367, -0.00000751, 0.00000467, -0.00000234, 0.00000412, 0.00000066, -0.00000271, -0.00000850, 0.00000925, -0.00001198, 0.00000680, -0.00001223, -0.00001031, 0.00000957, -0.00000307, 0.00004261, -0.00002698, 0.00002103, 0.00001857, 0.00000287, -0.00000451, -0.00002162, 0.00001463, -0.00000429, 0.00000016, -0.00000663, -0.00000005, -0.00000635, 0.00000132, -0.00000174, 0.00000122, -0.00000001, -0.00000284, 0.00000103, -0.00000142, -0.00000220, 0.00000505, 0.00000007, 0.00000109, -0.00000406, 0.00000311, 0.00000198, 0.00000108, -0.00000166, 0.00000627, },
		{ 0.00001566, -0.00002019, 0.00001372, -0.00001232, 0.00001177, -0.00000454, -0.00001671, 0.00001213, 0.00000224, -0.00000834, -0.00002062, 0.00002560, -0.00001749, 0.00001249, -0.00001842, 0.00001145, -0.00000346, 0.00000371, 0.00000167, -0.00000244, 0.00000215, -0.00000363, -0.00000710, -0.00000208, 0.00001281, -0.00001029, -0.00000648, 0.00000602, 0.00000995, -0.00000287, 0.00000215, -0.00000103, -0.00000894, 0.00000666, -0.00000624, 0.00000667, -0.00000410, 0.00000394, -0.00000864, 0.00000554, -0.00000193, 0.00000064, -0.00001336, 0.00000224, -0.00001318, 0.00000736, -0.00001310, 0.00000790, -0.00001465, -0.00000117, -0.00005522, 0.00000207, -0.00024061, 0.00029412, 0.00042114, -0.00023680, -0.00004687, -0.00002555, 0.00001115, -0.00002850, -0.00000298, -0.00001201, -0.00000414, 0.00000037, 0.00000265, -0.00000352, 0.00000266, -0.00000914, 0.00000252, -0.00000134, 0.00000588, -0.00000612, 0.00000146, -0.00000324, -0.00000750, 0.00000394, -0.00000447, 0.00000472, 0.00000348, -0.00000061, -0.00001677, 0.00000971, 0.00000282, -0.00000213, -0.00000145, -0.00000297, 0.00000378, -0.00000123, 0.00000485, -0.00000302, -0.00000713, 0.00000320, -0.00000142, -0.00000619, 0.00000523, -0.00002557, 0.00001785, -0.00001326, 0.00003909, -0.00000613, -0.00000670, 0.00003282, -0.00000051, 0.00000573, -0.00002281, 0.00000401, 0.00000589, -0.00000542, -0.00000143, -0.00000511, -0.00000091, -0.00000490, 0.00000324, -0.00000375, 0.00000462, -0.00000549, 0.00000215, -0.00000183, -0.00000130, 0.00000091, 0.00000475, -0.00000219, 0.00000078, -0.00000386, 0.00000772, -0.00000350, 0.00000440, -0.00000450, },
		{ 0.00001311, -0.00000839, -0.00000325, 0.00000650, -0.00000956, 0.00001810, -0.00002962, 0.00001313, -0.00000011, 0.00000411, -0.00003138, 0.00002280, -0.00000854, 0.00000420, -0.00000702, -0.00000331, 0.00000853, -0.00000621, 0.00001004, -0.00000780, 0.00000533, -0.00000425, -0.00000280, -0.00000969, 0.00001248, -0.00000079, -0.00001227, 0.00000249, 0.00000954, 0.00000349, -0.00000411, 0.00000634, -0.00001282, 0.00000571, -0.00000443, 0.00000377, -0.00000140, 0.00000251, -0.00000562, 0.00000095, 0.00000066, 0.00000172, -0.00001183, -0.00000026, -0.00001047, 0.00000271, -0.00000791, 0.00000237, -0.00001014, 0.00000361, -0.00005126, 0.00000692, -0.00016613, -0.00002077, 0.00060677, -0.00010646, -0.00014382, -0.00000722, -0.00000828, -0.00000752, -0.00002136, -0.00000046, -0.00001351, 0.00000282, 0.00000084, -0.00000151, 0.00000121, -0.00000553, -0.00000267, 0.00000123, 0.00000278, -0.00000009, -0.00000478, 0.00000325, -0.00001077, 0.00000240, -0.00000182, 0.00000016, 0.00000542, 0.00000213, -0.00001207, -0.00000448, 0.00001360, -0.00000696, 0.00000372, -0.00000826, 0.00000623, -0.00000311, 0.00000614, -0.00000264, -0.00000368, -0.00000506, 0.00000874, -0.00001526, 0.00001372, -0.00002677, 0.00001073, -0.00000584, 0.00001862, 0.00002109, -0.00002511, 0.00003028, 0.00000891, 0.00000307, -0.00001010, -0.00001302, 0.00001499, -0.00000811, 0.00000136, -0.00000771, 0.00000155, -0.00000708, 0.00000371, -0.00000410, 0.00000412, -0.00000328, -0.00000023, -0.00000039, -0.00000095, -0.00000206, 0.00000651, -0.00000235, 0.00000242, -0.00000586, 0.00000704, -0.00000191, 0.00000429, -0.00000586, },
		{ -0.00000029, 0.00001107, -0.00002114, 0.00001975, -0.00001905, 0.00002193, -0.00002090, -0.00000188, 0.00000718, 0.00000495, -0.00002444, 0.00000288, 0.00001150, -0.00001143, 0.00000914, -0.00001816, 0.00001722, -0.00001188, 0.00001325, -0.00000776, 0.00000410, -0.00000284, -0.00000016, -0.00001147, 0.00000519, 0.00000860, -0.00001156, -0.00000480, 0.00000924, 0.00000683, -0.00000337, 0.00000358, -0.00000459, -0.00000583, 0.00000573, -0.00000614, 0.00000744, -0.00000554, 0.00000432, -0.00000945, 0.00000874, -0.00000522, 0.00000000, -0.00001364, 0.00000407, -0.00001441, 0.00001022, -0.00001696, 0.00001034, -0.00001711, -0.00000996, -0.00004786, -0.00002578, -0.00025781, 0.00054616, 0.00019523, -0.00026597, 0.00001321, -0.00003999, 0.00002115, -0.00004184, 0.00001115, -0.00002272, 0.00000659, -0.00000491, 0.00000523, -0.00000561, 0.00000321, -0.00001106, 0.00000630, -0.00000325, 0.00000729, -0.00000975, 0.00000597, -0.00000889, -0.00000254, 0.00000160, -0.00000279, 0.00000485, 0.00000377, -0.00000474, -0.00001454, 0.00001388, -0.00000192, 0.00000086, -0.00000512, 0.00000084, 0.00000106, 0.00000269, 0.00000068, -0.00000250, -0.00000854, 0.00001013, -0.00001235, 0.00000739, -0.00001253, -0.00000820, 0.00001021, -0.00000376, 0.00003771, -0.00002387, 0.00001196, 0.00002317, -0.00000120, 0.00000075, -0.00001944, 0.00001003, 0.00000049, -0.00000444, -0.00000253, -0.00000360, -0.00000166, -0.00000270, 0.00000165, -0.00000203, 0.00000329, -0.00000532, 0.00000289, -0.00000266, -0.00000109, 0.00000270, 0.00000276, -0.00000137, -0.00000104, -0.00000062, 0.00000559, -0.00000203, 0.00000166, },
		{ -0.00000895, 0.00001886, -0.00002051, 0.00001116, -0.00000572, 0.00000352, 0.00000428, -0.00002314, 0.00001579, 0.00000046, -0.00000880, -0.00002040, 0.00002714, -0.00001968, 0.00001605, -0.00002081, 0.00001346, -0.00000675, 0.00000685, -0.00000003, -0.00000212, 0.00000116, -0.00000068, -0.00000773, -0.00000419, 0.00001211, -0.00000348, -0.00001310, 0.00000795, 0.00000665, 0.00000312, -0.00000465, 0.00000700, -0.00001629, 0.00001194, -0.00001115, 0.00001104, -0.00000889, 0.00000921, -0.00001357, 0.00001082, -0.00000818, 0.00000789, -0.00002117, 0.00001140, -0.00002248, 0.00001733, -0.00002377, 0.00001784, -0.00002727, 0.00001721, -0.00007882, 0.00004238, -0.00027962, 0.00023275, 0.00054701, -0.00029797, -0.00002481, -0.00004494, 0.00002477, -0.00003927, 0.00000477, -0.00001892, 0.00000203, -0.00000593, 0.00000797, -0.00000881, 0.00000768, -0.00001377, 0.00000610, -0.00000480, 0.00000888, -0.00000793, 0.00000230, -0.00000289, -0.00000730, 0.00000184, -0.00000137, 0.00000062, 0.00000635, -0.00000062, -0.00001393, 0.00000194, 0.00001021, -0.00000613, 0.00000208, -0.00000711, 0.00000609, -0.00000145, 0.00000463, -0.00000345, -0.00000625, 0.00000282, -0.00000027, -0.00000710, 0.00000627, -0.00002529, 0.00001998, -0.00001386, 0.00003563, -0.00000629, -0.00000939, 0.00002847, 0.00000434, -0.00000007, -0.00001048, -0.00000660, 0.00001403, -0.00001193, 0.00000391, -0.00000976, 0.00000420, -0.00000848, 0.00000609, -0.00000643, 0.00000705, -0.00000702, 0.00000346, -0.00000317, 0.00000064, -0.00000226, 0.00000766, -0.00000471, 0.00000373, -0.00000690, 0.00001002, -0.00000534, 0.00000683, },
		{ -0.00000450, 0.00000864, -0.00000101, -0.00001218, 0.00001649, -0.00001727, 0.00002295, -0.00003025, 0.00001141, 0.00000224, 0.00000144, -0.00003077, 0.00002432, -0.00001024, 0.00000631, -0.00000733, -0.00000310, 0.00000743, -0.00000616, 0.00001142, -0.00000948, 0.00000542, -0.00000317, -0.00000131, -0.00001137, 0.00000941, 0.00000577, -0.00001477, 0.00000077, 0.00000810, 0.00000668, -0.00000636, 0.00000798, -0.00001213, 0.00000290, -0.00000186, 0.00000109, 0.00000066, 0.00000056, -0.00000316, -0.00000112, 0.00000221, 0.00000016, -0.00000971, -0.00000248, -0.00000682, -0.00000169, -0.00000244, -0.00000492, -0.00000395, -0.00000194, -0.00004227, -0.00000470, -0.00012166, -0.00015120, 0.00072319, -0.00011376, -0.00016183, -0.00000051, -0.00001562, -0.00000167, -0.00002672, 0.00000307, -0.00001607, 0.00000431, -0.00000094, -0.00000038, 0.00000008, -0.00000425, -0.00000442, 0.00000254, 0.00000131, 0.00000185, -0.00000651, 0.00000453, -0.00000995, -0.00000042, 0.00000115, -0.00000266, 0.00000649, 0.00000207, -0.00000773, -0.00001113, 0.00001635, -0.00000598, 0.00000332, -0.00000825, 0.00000431, -0.00000048, 0.00000473, -0.00000268, -0.00000313, -0.00000533, 0.00000990, -0.00001612, 0.00001471, -0.00002707, 0.00001331, -0.00000661, 0.00001795, 0.00001652, -0.00002187, 0.00002027, 0.00001701, -0.00000531, 0.00000079, -0.00001770, 0.00001626, -0.00000810, 0.00000079, -0.00000702, 0.00000118, -0.00000521, 0.00000185, -0.00000254, 0.00000253, -0.00000132, -0.00000165, 0.00000045, -0.00000100, -0.00000262, 0.00000608, -0.00000143, 0.00000172, -0.00000468, 0.00000491, 0.00000039, 0.00000252, },
		{ 0.00000531, -0.00000675, 0.00001793, -0.00002647, 0.00002276, -0.00001743, 0.00001715, -0.00001366, -0.00000841, 0.00001193, 0.00000172, -0.00002399, 0.00000364, 0.00001176, -0.00001226, 0.00001195, -0.00002056, 0.00001905, -0.00001506, 0.00001706, -0.00001068, 0.00000523, -0.00000389, 0.00000325, -0.00001325, 0.00000301, 0.00000997, -0.00000657, -0.00001165, 0.00001169, 0.00000418, 0.00000174, -0.00000287, 0.00000403, -0.00001500, 0.00001341, -0.00001351, 0.00001421, -0.00001262, 0.00001199, -0.00001687, 0.00001638, -0.00001388, 0.00000996, -0.00002449, 0.00001695, -0.00002833, 0.00002566, -0.00003423, 0.00002767, -0.00003756, 0.00001826, -0.00008522, 0.00004243, -0.00035929, 0.00056201, 0.00028455, -0.00033231, 0.00004616, -0.00006700, 0.00004140, -0.00005823, 0.00002371, -0.00003335, 0.00001570, -0.00001378, 0.00001285, -0.00001290, 0.00001011, -0.00001763, 0.00001176, -0.00000841, 0.00001203, -0.00001331, 0.00000828, -0.00000918, -0.00000278, 0.00000050, -0.00000076, 0.00000141, 0.00000657, -0.00000373, -0.00001430, 0.00000816, 0.00000535, -0.00000379, -0.00000109, -0.00000430, 0.00000506, 0.00000094, 0.00000118, -0.00000309, -0.00000757, 0.00000991, -0.00001167, 0.00000693, -0.00001190, -0.00000673, 0.00001044, -0.00000353, 0.00003268, -0.00002123, 0.00000531, 0.00002454, -0.00000155, 0.00000009, -0.00001140, 0.00000122, 0.00000817, -0.00001107, 0.00000329, -0.00000881, 0.00000432, -0.00000755, 0.00000572, -0.00000582, 0.00000688, -0.00000791, 0.00000495, -0.00000438, 0.00000087, -0.00000061, 0.00000619, -0.00000453, 0.00000267, -0.00000484, 0.00000933, -0.00000524, },
		{ 0.00000731, -0.00001065, 0.00001825, -0.00001666, 0.00000442, 0.00000417, -0.00000620, 0.00001256, -0.00002760, 0.00001742, 0.00000038, -0.00001036, -0.00001959, 0.00002851, -0.00002193, 0.00001967, -0.00002323, 0.00001544, -0.00000984, 0.00000949, -0.00000119, -0.00000207, 0.00000002, 0.00000229, -0.00000831, -0.00000520, 0.00000901, 0.00000500, -0.00001956, 0.00000965, 0.00000276, 0.00000947, -0.00001122, 0.00001400, -0.00002168, 0.00001503, -0.00001406, 0.00001364, -0.00001207, 0.00001277, -0.00001680, 0.00001459, -0.00001297, 0.00001343, -0.00002691, 0.00001808, -0.00002879, 0.00002383, -0.00003038, 0.00002361, -0.00003561, 0.00002981, -0.00009353, 0.00006628, -0.00028464, 0.00012914, 0.00069458, -0.00036011, -0.00000539, -0.00006197, 0.00003574, -0.00004781, 0.00001110, -0.00002489, 0.00000761, -0.00001177, 0.00001283, -0.00001366, 0.00001211, -0.00001770, 0.00000905, -0.00000767, 0.00001125, -0.00000931, 0.00000314, -0.00000321, -0.00000553, -0.00000212, 0.00000325, -0.00000440, 0.00000973, -0.00000188, -0.00000887, -0.00000722, 0.00001715, -0.00000902, 0.00000479, -0.00001032, 0.00000724, -0.00000091, 0.00000424, -0.00000413, -0.00000498, 0.00000207, 0.00000115, -0.00000826, 0.00000739, -0.00002493, 0.00002155, -0.00001450, 0.00003372, -0.00000889, -0.00000822, 0.00002050, 0.00001250, -0.00000966, 0.00000346, -0.00001647, 0.00002049, -0.00001659, 0.00000762, -0.00001293, 0.00000760, -0.00001020, 0.00000727, -0.00000753, 0.00000791, -0.00000715, 0.00000373, -0.00000378, 0.00000202, -0.00000465, 0.00000944, -0.00000602, 0.00000532, -0.00000827, 0.00001036, -0.00000527, },
		{ -0.00000059, -0.00000132, 0.00000227, 0.00000734, -0.00002088, 0.00002449, -0.00002127, 0.00002319, -0.00002656, 0.00000676, 0.00000643, -0.00000222, -0.00002955, 0.00002539, -0.00001138, 0.00000765, -0.00000673, -0.00000393, 0.00000753, -0.00000744, 0.00001389, -0.00001189, 0.00000602, -0.00000287, 0.00000091, -0.00001293, 0.00000627, 0.00001045, -0.00001388, -0.00000362, 0.00000873, 0.00000702, -0.00000501, 0.00000543, -0.00000698, -0.00000387, 0.00000411, -0.00000476, 0.00000588, -0.00000475, 0.00000272, -0.00000664, 0.00000755, -0.00000576, -0.00000270, -0.00001009, 0.00000277, -0.00001250, 0.00000999, -0.00001950, 0.00001049, -0.00001824, -0.00001904, -0.00003605, -0.00005103, -0.00029622, 0.00082433, -0.00009405, -0.00019891, 0.00001940, -0.00003347, 0.00001221, -0.00003822, 0.00001176, -0.00002316, 0.00001002, -0.00000668, 0.00000444, -0.00000459, 0.00000027, -0.00000901, 0.00000635, -0.00000238, 0.00000549, -0.00000936, 0.00000640, -0.00000945, -0.00000241, 0.00000251, -0.00000337, 0.00000549, 0.00000328, -0.00000456, -0.00001501, 0.00001508, -0.00000154, 0.00000043, -0.00000597, 0.00000037, 0.00000337, 0.00000265, -0.00000215, -0.00000330, -0.00000471, 0.00001006, -0.00001598, 0.00001465, -0.00002638, 0.00001498, -0.00000757, 0.00001864, 0.00001061, -0.00001723, 0.00001048, 0.00002385, -0.00001175, 0.00000713, -0.00001677, 0.00001263, -0.00000415, -0.00000288, -0.00000359, -0.00000191, -0.00000088, -0.00000198, 0.00000075, -0.00000061, 0.00000187, -0.00000396, 0.00000209, -0.00000200, -0.00000182, 0.00000404, 0.00000099, -0.00000055, -0.00000182, 0.00000113, 0.00000405, },
		{ -0.00000732, 0.00000789, -0.00001142, 0.00002166, -0.00002762, 0.00002069, -0.00001069, 0.00000836, -0.00000427, -0.00001523, 0.00001620, -0.00000077, -0.00002399, 0.00000451, 0.00001231, -0.00001366, 0.00001539, -0.00002365, 0.00002160, -0.00001886, 0.00002115, -0.00001370, 0.00000660, -0.00000564, 0.00000728, -0.00001537, 0.00000222, 0.00000809, 0.00000179, -0.00002027, 0.00001558, -0.00000079, 0.00000939, -0.00001164, 0.00001442, -0.00002496, 0.00002139, -0.00002117, 0.00002142, -0.00002036, 0.00002034, -0.00002503, 0.00002511, -0.00002393, 0.00002142, -0.00003681, 0.00003124, -0.00004349, 0.00004222, -0.00005236, 0.00004626, -0.00006026, 0.00004898, -0.00012444, 0.00010746, -0.00043687, 0.00053299, 0.00040992, -0.00041496, 0.00008776, -0.00010009, 0.00006565, -0.00007758, 0.00003906, -0.00004669, 0.00002751, -0.00002521, 0.00002284, -0.00002239, 0.00001886, -0.00002569, 0.00001854, -0.00001472, 0.00001763, -0.00001759, 0.00001144, -0.00001078, -0.00000077, -0.00000346, 0.00000411, -0.00000439, 0.00001113, -0.00000502, -0.00001035, -0.00000113, 0.00001449, -0.00000929, 0.00000375, -0.00000996, 0.00000903, -0.00000081, 0.00000212, -0.00000449, -0.00000556, 0.00000861, -0.00000996, 0.00000547, -0.00001046, -0.00000587, 0.00001047, -0.00000300, 0.00002851, -0.00002037, 0.00000237, 0.00002158, 0.00000272, -0.00000640, 0.00000119, -0.00000996, 0.00001690, -0.00001807, 0.00000937, -0.00001430, 0.00001023, -0.00001194, 0.00000929, -0.00000908, 0.00000980, -0.00000981, 0.00000656, -0.00000598, 0.00000303, -0.00000404, 0.00000943, -0.00000742, 0.00000600, -0.00000836, 0.00001198, },
		{ -0.00000308, 0.00000565, -0.00000950, 0.00001413, -0.00000952, -0.00000491, 0.00001505, -0.00001483, 0.00001803, -0.00002852, 0.00001609, 0.00000250, -0.00001322, -0.00001800, 0.00002943, -0.00002375, 0.00002275, -0.00002504, 0.00001673, -0.00001205, 0.00001102, -0.00000136, -0.00000263, -0.00000091, 0.00000486, -0.00000851, -0.00000547, 0.00000452, 0.00001338, -0.00002406, 0.00000963, -0.00000012, 0.00001432, -0.00001552, 0.00001791, -0.00002327, 0.00001442, -0.00001366, 0.00001317, -0.00001229, 0.00001323, -0.00001689, 0.00001529, -0.00001450, 0.00001523, -0.00002836, 0.00001991, -0.00002960, 0.00002422, -0.00003014, 0.00002216, -0.00003595, 0.00003190, -0.00009354, 0.00006724, -0.00025194, -0.00001215, 0.00085240, -0.00041206, 0.00000346, -0.00007080, 0.00003956, -0.00005064, 0.00001319, -0.00002747, 0.00001036, -0.00001502, 0.00001525, -0.00001621, 0.00001422, -0.00001941, 0.00001005, -0.00000880, 0.00001204, -0.00000953, 0.00000341, -0.00000367, -0.00000295, -0.00000679, 0.00000793, -0.00000887, 0.00001237, -0.00000318, -0.00000318, -0.00001565, 0.00002163, -0.00000934, 0.00000557, -0.00001168, 0.00000658, 0.00000078, 0.00000337, -0.00000464, -0.00000385, 0.00000155, 0.00000223, -0.00000905, 0.00000803, -0.00002404, 0.00002241, -0.00001524, 0.00003341, -0.00001370, -0.00000392, 0.00001025, 0.00002212, -0.00002024, 0.00001572, -0.00002276, 0.00002316, -0.00001780, 0.00000846, -0.00001352, 0.00000830, -0.00000929, 0.00000618, -0.00000656, 0.00000683, -0.00000562, 0.00000274, -0.00000338, 0.00000240, -0.00000569, 0.00000953, -0.00000560, 0.00000507, -0.00000752, 0.00000844, },
		{ 0.00000797, -0.00000389, 0.00000233, -0.00000419, 0.00001449, -0.00002698, 0.00002823, -0.00001994, 0.00001819, -0.00001878, -0.00000021, 0.00001179, -0.00000632, -0.00002798, 0.00002596, -0.00001173, 0.00000794, -0.00000500, -0.00000599, 0.00000899, -0.00001006, 0.00001739, -0.00001498, 0.00000725, -0.00000362, 0.00000412, -0.00001493, 0.00000427, 0.00001188, -0.00000876, -0.00001108, 0.00001192, 0.00000393, 0.00000035, -0.00000140, 0.00000233, -0.00001398, 0.00001285, -0.00001322, 0.00001379, -0.00001301, 0.00001163, -0.00001532, 0.00001648, -0.00001583, 0.00000891, -0.00002266, 0.00001769, -0.00002893, 0.00002841, -0.00004029, 0.00003214, -0.00004400, 0.00001623, -0.00008279, 0.00003625, -0.00043988, 0.00089397, -0.00003757, -0.00025933, 0.00005467, -0.00006306, 0.00003486, -0.00005650, 0.00002632, -0.00003556, 0.00002070, -0.00001708, 0.00001359, -0.00001334, 0.00000845, -0.00001674, 0.00001297, -0.00000854, 0.00001111, -0.00001376, 0.00000954, -0.00001030, -0.00000225, 0.00000092, -0.00000087, 0.00000157, 0.00000667, -0.00000380, -0.00001482, 0.00000910, 0.00000636, -0.00000479, -0.00000158, -0.00000533, 0.00000822, -0.00000010, -0.00000082, -0.00000454, -0.00000281, 0.00000883, -0.00001447, 0.00001322, -0.00002448, 0.00001559, -0.00000844, 0.00002001, 0.00000453, -0.00001292, 0.00000293, 0.00002725, -0.00001392, 0.00000720, -0.00000974, 0.00000444, 0.00000306, -0.00000893, 0.00000190, -0.00000702, 0.00000506, -0.00000693, 0.00000496, -0.00000451, 0.00000554, -0.00000653, 0.00000408, -0.00000361, -0.00000002, 0.00000090, 0.00000433, -0.00000368, 0.00000191, -0.00000329, },
		{ 0.00001256, -0.00000734, 0.00000839, -0.00001321, 0.00002133, -0.00002411, 0.00001378, -0.00000004, -0.00000249, 0.00000516, -0.00002052, 0.00001869, -0.00000169, -0.00002489, 0.00000573, 0.00001291, -0.00001525, 0.00001896, -0.00002687, 0.00002428, -0.00002263, 0.00002487, -0.00001631, 0.00000787, -0.00000779, 0.00001166, -0.00001777, 0.00000292, 0.00000333, 0.00001232, -0.00002919, 0.00001964, -0.00000675, 0.00001791, -0.00002078, 0.00002436, -0.00003349, 0.00002767, -0.00002733, 0.00002738, -0.00002705, 0.00002760, -0.00003214, 0.00003304, -0.00003328, 0.00003200, -0.00004797, 0.00004403, -0.00005669, 0.00005640, -0.00006761, 0.00006199, -0.00008023, 0.00007575, -0.00015682, 0.00015755, -0.00047833, 0.00045357, 0.00056795, -0.00050730, 0.00013220, -0.00013456, 0.00009005, -0.00009687, 0.00005482, -0.00006072, 0.00004017, -0.00003747, 0.00003354, -0.00003251, 0.00002798, -0.00003391, 0.00002549, -0.00002119, 0.00002330, -0.00002207, 0.00001517, -0.00001364, 0.00000342, -0.00000989, 0.00001107, -0.00001168, 0.00001672, -0.00000806, -0.00000359, -0.00001240, 0.00002368, -0.00001412, 0.00000818, -0.00001509, 0.00001217, -0.00000209, 0.00000318, -0.00000644, -0.00000285, 0.00000660, -0.00000763, 0.00000344, -0.00000862, -0.00000531, 0.00001031, -0.00000258, 0.00002585, -0.00002198, 0.00000353, 0.00001440, 0.00001100, -0.00001701, 0.00001567, -0.00002078, 0.00002427, -0.00002347, 0.00001407, -0.00001851, 0.00001455, -0.00001457, 0.00001125, -0.00001081, 0.00001116, -0.00001030, 0.00000715, -0.00000688, 0.00000467, -0.00000672, 0.00001157, -0.00000909, 0.00000796, -0.00001014, },
		{ 0.00000465, 0.00000069, 0.00000242, -0.00000627, 0.00000775, -0.00000092, -0.00001444, 0.00002412, -0.00001982, 0.00001887, -0.00002491, 0.00001147, 0.00000681, -0.00001729, -0.00001560, 0.00002964, -0.00002469, 0.00002470, -0.00002566, 0.00001675, -0.00001279, 0.00001096, -0.00000020, -0.00000408, -0.00000122, 0.00000653, -0.00000787, -0.00000574, 0.00000010, 0.00001970, -0.00002488, 0.00000659, -0.00000063, 0.00001606, -0.00001593, 0.00001724, -0.00001990, 0.00000930, -0.00000930, 0.00000897, -0.00000880, 0.00000975, -0.00001294, 0.00001188, -0.00001152, 0.00001189, -0.00002405, 0.00001535, -0.00002335, 0.00001693, -0.00002149, 0.00001177, -0.00002610, 0.00002083, -0.00007598, 0.00004312, -0.00018394, -0.00018005, 0.00100448, -0.00044075, -0.00000667, -0.00006564, 0.00003199, -0.00004459, 0.00000842, -0.00002426, 0.00000800, -0.00001356, 0.00001325, -0.00001464, 0.00001236, -0.00001746, 0.00000786, -0.00000710, 0.00001032, -0.00000777, 0.00000230, -0.00000336, -0.00000080, -0.00001051, 0.00001090, -0.00001112, 0.00001281, -0.00000312, 0.00000132, -0.00002123, 0.00002190, -0.00000602, 0.00000369, -0.00001060, 0.00000371, 0.00000388, 0.00000171, -0.00000449, -0.00000348, 0.00000193, 0.00000231, -0.00000880, 0.00000759, -0.00002220, 0.00002230, -0.00001588, 0.00003424, -0.00001989, 0.00000201, -0.00000023, 0.00003068, -0.00002868, 0.00002329, -0.00002342, 0.00002080, -0.00001490, 0.00000600, -0.00001115, 0.00000602, -0.00000569, 0.00000288, -0.00000366, 0.00000399, -0.00000267, 0.00000070, -0.00000203, 0.00000170, -0.00000525, 0.00000787, -0.00000350, 0.00000306, -0.00000487, },
		{ -0.00000710, 0.00001083, -0.00000610, 0.00000494, -0.00000892, 0.00001853, -0.00002867, 0.00002641, -0.00001296, 0.00000859, -0.00000816, -0.00000811, 0.00001720, -0.00001005, -0.00002647, 0.00002619, -0.00001130, 0.00000719, -0.00000227, -0.00000907, 0.00001156, -0.00001371, 0.00002151, -0.00001846, 0.00000902, -0.00000544, 0.00000841, -0.00001776, 0.00000419, 0.00000943, 0.00000045, -0.00002104, 0.00001723, -0.00000216, 0.00000905, -0.00001148, 0.00001438, -0.00002585, 0.00002284, -0.00002296, 0.00002315, -0.00002299, 0.00002236, -0.00002596, 0.00002781, -0.00002873, 0.00002358, -0.00003844, 0.00003590, -0.00004866, 0.00005022, -0.00006448, 0.00005796, -0.00007549, 0.00005837, -0.00013706, 0.00012732, -0.00056486, 0.00091772, 0.00006158, -0.00034351, 0.00010435, -0.00010315, 0.00006503, -0.00008061, 0.00004614, -0.00005283, 0.00003601, -0.00003178, 0.00002668, -0.00002575, 0.00001981, -0.00002717, 0.00002202, -0.00001691, 0.00001856, -0.00001976, 0.00001431, -0.00001321, 0.00000093, -0.00000426, 0.00000518, -0.00000542, 0.00001247, -0.00000595, -0.00001024, -0.00000117, 0.00001679, -0.00001143, 0.00000418, -0.00001211, 0.00001351, -0.00000325, 0.00000131, -0.00000695, 0.00000046, 0.00000616, -0.00001156, 0.00001045, -0.00002145, 0.00001519, -0.00000899, 0.00002136, -0.00000059, -0.00001041, -0.00000083, 0.00002587, -0.00001077, 0.00000098, 0.00000224, -0.00000663, 0.00001179, -0.00001576, 0.00000807, -0.00001278, 0.00001124, -0.00001169, 0.00000889, -0.00000808, 0.00000867, -0.00000853, 0.00000573, -0.00000522, 0.00000210, -0.00000250, 0.00000762, -0.00000668, 0.00000539, },
		{ -0.00000965, 0.00001067, -0.00000504, 0.00000652, -0.00001188, 0.00001712, -0.00001672, 0.00000359, 0.00001214, -0.00001275, 0.00001223, -0.00002254, 0.00001832, -0.00000042, -0.00002699, 0.00000757, 0.00001320, -0.00001651, 0.00002198, -0.00002949, 0.00002632, -0.00002555, 0.00002747, -0.00001792, 0.00000861, -0.00000988, 0.00001583, -0.00002002, 0.00000461, -0.00000312, 0.00002308, -0.00003639, 0.00002217, -0.00001194, 0.00002525, -0.00002801, 0.00003152, -0.00003844, 0.00003043, -0.00003035, 0.00003046, -0.00003100, 0.00003202, -0.00003641, 0.00003819, -0.00003969, 0.00003919, -0.00005521, 0.00005232, -0.00006474, 0.00006482, -0.00007636, 0.00007087, -0.00009259, 0.00009241, -0.00017460, 0.00018341, -0.00047644, 0.00032515, 0.00074888, -0.00059831, 0.00017099, -0.00016386, 0.00010945, -0.00011207, 0.00006768, -0.00007251, 0.00005102, -0.00004805, 0.00004263, -0.00004108, 0.00003544, -0.00004050, 0.00003104, -0.00002646, 0.00002789, -0.00002584, 0.00001875, -0.00001708, 0.00000890, -0.00001748, 0.00001856, -0.00001886, 0.00002195, -0.00001156, 0.00000429, -0.00002340, 0.00003075, -0.00001666, 0.00001096, -0.00001862, 0.00001368, -0.00000239, 0.00000396, -0.00000842, -0.00000005, 0.00000454, -0.00000534, 0.00000153, -0.00000701, -0.00000460, 0.00000979, -0.00000238, 0.00002480, -0.00002597, 0.00000815, 0.00000427, 0.00002135, -0.00002885, 0.00002871, -0.00002835, 0.00002805, -0.00002556, 0.00001600, -0.00002020, 0.00001617, -0.00001453, 0.00001087, -0.00001041, 0.00001049, -0.00000906, 0.00000642, -0.00000672, 0.00000527, -0.00000801, 0.00001196, -0.00000896, 0.00000795, },
		{ -0.00000111, -0.00000002, 0.00000491, -0.00000120, -0.00000237, 0.00000101, 0.00000683, -0.00002156, 0.00002887, -0.00001931, 0.00001418, -0.00001677, 0.00000402, 0.00001277, -0.00002211, -0.00001257, 0.00002902, -0.00002443, 0.00002511, -0.00002473, 0.00001517, -0.00001179, 0.00000912, 0.00000236, -0.00000652, -0.00000063, 0.00000686, -0.00000589, -0.00000681, -0.00000283, 0.00002233, -0.00002092, -0.00000011, 0.00000199, 0.00001384, -0.00001171, 0.00001154, -0.00001153, -0.00000004, -0.00000128, 0.00000126, -0.00000171, 0.00000237, -0.00000494, 0.00000421, -0.00000377, 0.00000313, -0.00001373, 0.00000425, -0.00001001, 0.00000206, -0.00000465, -0.00000735, -0.00000609, -0.00000331, -0.00004164, -0.00000325, -0.00008907, -0.00035825, 0.00113247, -0.00043352, -0.00004267, -0.00004231, 0.00001021, -0.00002762, -0.00000493, -0.00001363, -0.00000103, -0.00000588, 0.00000548, -0.00000771, 0.00000545, -0.00001097, 0.00000172, -0.00000188, 0.00000542, -0.00000337, -0.00000093, -0.00000129, -0.00000046, -0.00001164, 0.00001050, -0.00000975, 0.00000990, -0.00000051, 0.00000310, -0.00002237, 0.00001700, 0.00000122, -0.00000092, -0.00000700, -0.00000132, 0.00000830, -0.00000087, -0.00000332, -0.00000437, 0.00000373, 0.00000086, -0.00000701, 0.00000565, -0.00001910, 0.00002101, -0.00001612, 0.00003551, -0.00002620, 0.00000774, -0.00000869, 0.00003576, -0.00003244, 0.00002432, -0.00001773, 0.00001350, -0.00000835, 0.00000082, -0.00000635, 0.00000135, -0.00000011, -0.00000185, 0.00000043, 0.00000013, 0.00000096, -0.00000178, -0.00000018, 0.00000024, -0.00000364, 0.00000493, -0.00000028, -0.00000004, },
		{ 0.00000751, -0.00000915, 0.00001157, -0.00000616, 0.00000537, -0.00001066, 0.00001833, -0.00002527, 0.00001909, -0.00000144, -0.00000374, 0.00000319, -0.00001510, 0.00002127, -0.00001249, -0.00002553, 0.00002635, -0.00001035, 0.00000577, 0.00000097, -0.00001264, 0.00001464, -0.00001769, 0.00002559, -0.00002183, 0.00001105, -0.00000810, 0.00001352, -0.00002138, 0.00000616, 0.00000344, 0.00001257, -0.00003200, 0.00002338, -0.00000995, 0.00001944, -0.00002284, 0.00002692, -0.00003720, 0.00003204, -0.00003210, 0.00003220, -0.00003293, 0.00003309, -0.00003674, 0.00003959, -0.00004228, 0.00003886, -0.00005467, 0.00005435, -0.00006831, 0.00007173, -0.00008809, 0.00008354, -0.00010730, 0.00010027, -0.00018906, 0.00020833, -0.00065548, 0.00088591, 0.00020353, -0.00044697, 0.00016377, -0.00014967, 0.00009932, -0.00010786, 0.00006905, -0.00007313, 0.00005426, -0.00004919, 0.00004222, -0.00004038, 0.00003298, -0.00003907, 0.00003243, -0.00002657, 0.00002714, -0.00002693, 0.00002053, -0.00001824, 0.00000715, -0.00001275, 0.00001415, -0.00001470, 0.00002002, -0.00001056, -0.00000207, -0.00001422, 0.00002798, -0.00001803, 0.00001013, -0.00001890, 0.00001844, -0.00000634, 0.00000396, -0.00001027, 0.00000479, 0.00000239, -0.00000761, 0.00000675, -0.00001770, 0.00001404, -0.00000919, 0.00002223, -0.00000399, -0.00001059, -0.00000019, 0.00001961, -0.00000280, -0.00000989, 0.00001661, -0.00001792, 0.00001966, -0.00002140, 0.00001323, -0.00001760, 0.00001611, -0.00001489, 0.00001137, -0.00001025, 0.00001035, -0.00000920, 0.00000643, -0.00000620, 0.00000377, -0.00000524, 0.00000989, -0.00000852, },
		{ 0.00000599, -0.00000700, 0.00000664, -0.00000091, 0.00000282, -0.00000818, 0.00001025, -0.00000728, -0.00000736, 0.00002289, -0.00001964, 0.00001484, -0.00002004, 0.00001451, 0.00000324, -0.00003035, 0.00001017, 0.00001283, -0.00001687, 0.00002373, -0.00003079, 0.00002699, -0.00002689, 0.00002834, -0.00001807, 0.00000841, -0.00001137, 0.00001914, -0.00002143, 0.00000633, -0.00000959, 0.00003184, -0.00003989, 0.00002159, -0.00001469, 0.00002951, -0.00003133, 0.00003401, -0.00003826, 0.00002847, -0.00002918, 0.00002960, -0.00003105, 0.00003232, -0.00003645, 0.00003903, -0.00004137, 0.00004100, -0.00005640, 0.00005385, -0.00006527, 0.00006501, -0.00007605, 0.00007000, -0.00009376, 0.00009456, -0.00017271, 0.00018008, -0.00043050, 0.00015653, 0.00093718, -0.00067393, 0.00019425, -0.00018087, 0.00011845, -0.00011901, 0.00007415, -0.00007892, 0.00005711, -0.00005421, 0.00004756, -0.00004575, 0.00003912, -0.00004358, 0.00003355, -0.00002907, 0.00003011, -0.00002774, 0.00002108, -0.00001993, 0.00001416, -0.00002435, 0.00002455, -0.00002403, 0.00002513, -0.00001387, 0.00001123, -0.00003176, 0.00003370, -0.00001559, 0.00001112, -0.00001973, 0.00001297, -0.00000128, 0.00000397, -0.00000982, 0.00000207, 0.00000324, -0.00000390, 0.00000053, -0.00000633, -0.00000320, 0.00000860, -0.00000220, 0.00002498, -0.00003151, 0.00001474, -0.00000666, 0.00003110, -0.00003866, 0.00003718, -0.00003042, 0.00002680, -0.00002345, 0.00001456, -0.00001880, 0.00001465, -0.00001161, 0.00000811, -0.00000790, 0.00000789, -0.00000625, 0.00000451, -0.00000550, 0.00000468, -0.00000771, 0.00001047, -0.00000698, },
		{ -0.00000318, 0.00000256, -0.00000423, 0.00000819, -0.00000382, 0.00000060, -0.00000415, 0.00001160, -0.00002420, 0.00002771, -0.00001274, 0.00000439, -0.00000517, -0.00000500, 0.00001931, -0.00002689, -0.00000931, 0.00002767, -0.00002291, 0.00002393, -0.00002229, 0.00001208, -0.00000918, 0.00000575, 0.00000603, -0.00000976, 0.00000093, 0.00000565, -0.00000231, -0.00000925, -0.00000328, 0.00002040, -0.00001208, -0.00001013, 0.00000750, 0.00000783, -0.00000326, 0.00000157, 0.00000067, -0.00001226, 0.00000912, -0.00000884, 0.00000801, -0.00000796, 0.00000619, -0.00000686, 0.00000792, -0.00001013, 0.00000148, -0.00001209, 0.00000887, -0.00001857, 0.00001833, -0.00003300, 0.00002168, -0.00003750, 0.00000508, -0.00006465, 0.00001967, -0.00052765, 0.00121866, -0.00038071, -0.00010799, 0.00000039, -0.00002612, 0.00000026, -0.00002696, 0.00000468, -0.00001707, 0.00000835, -0.00000834, 0.00000476, -0.00000659, 0.00000008, -0.00000836, 0.00000691, -0.00000274, 0.00000386, -0.00000667, 0.00000321, -0.00000292, -0.00000901, 0.00000578, -0.00000410, 0.00000321, 0.00000514, 0.00000138, -0.00001845, 0.00000709, 0.00001166, -0.00000756, -0.00000146, -0.00000797, 0.00001365, -0.00000426, -0.00000101, -0.00000675, 0.00000716, -0.00000229, -0.00000354, 0.00000212, -0.00001474, 0.00001854, -0.00001570, 0.00003647, -0.00003141, 0.00001158, -0.00001332, 0.00003571, -0.00003026, 0.00001857, -0.00000663, 0.00000275, 0.00000026, -0.00000564, -0.00000040, -0.00000441, 0.00000609, -0.00000674, 0.00000455, -0.00000365, 0.00000429, -0.00000386, 0.00000151, -0.00000139, -0.00000155, 0.00000155, 0.00000308, },
		{ -0.00000852, 0.00000760, -0.00000880, 0.00000966, -0.00000371, 0.00000323, -0.00000906, 0.00001390, -0.00001739, 0.00000778, 0.00001223, -0.00001607, 0.00001273, -0.00001923, 0.00002273, -0.00001286, -0.00002559, 0.00002680, -0.00000929, 0.00000423, 0.00000399, -0.00001590, 0.00001739, -0.00002115, 0.00002885, -0.00002446, 0.00001286, -0.00001109, 0.00001888, -0.00002531, 0.00000959, -0.00000483, 0.00002554, -0.00004185, 0.00002853, -0.00001753, 0.00002934, -0.00003302, 0.00003740, -0.00004565, 0.00003836, -0.00003876, 0.00003908, -0.00004089, 0.00004181, -0.00004557, 0.00004957, -0.00005393, 0.00005188, -0.00006821, 0.00006961, -0.00008420, 0.00008896, -0.00010685, 0.00010408, -0.00013359, 0.00013448, -0.00022919, 0.00026710, -0.00070052, 0.00079575, 0.00038144, -0.00056003, 0.00022480, -0.00019611, 0.00013248, -0.00013410, 0.00009163, -0.00009343, 0.00007267, -0.00006669, 0.00005772, -0.00005489, 0.00004582, -0.00005052, 0.00004254, -0.00003608, 0.00003562, -0.00003426, 0.00002743, -0.00002467, 0.00001548, -0.00002322, 0.00002445, -0.00002465, 0.00002790, -0.00001630, 0.00000798, -0.00002773, 0.00003765, -0.00002286, 0.00001493, -0.00002454, 0.00002211, -0.00000876, 0.00000664, -0.00001396, 0.00000951, -0.00000175, -0.00000335, 0.00000283, -0.00001391, 0.00001267, -0.00000922, 0.00002246, -0.00000546, -0.00001350, 0.00000437, 0.00000963, 0.00000806, -0.00002253, 0.00003004, -0.00002649, 0.00002434, -0.00002403, 0.00001589, -0.00002009, 0.00001841, -0.00001550, 0.00001156, -0.00001032, 0.00001001, -0.00000812, 0.00000581, -0.00000613, 0.00000442, -0.00000661, 0.00001041, },
		{ -0.00000367, 0.00000271, -0.00000301, 0.00000150, 0.00000393, -0.00000152, -0.00000353, 0.00000267, 0.00000175, -0.00001624, 0.00002940, -0.00002104, 0.00001182, -0.00001274, 0.00000752, 0.00000892, -0.00003464, 0.00001346, 0.00001157, -0.00001591, 0.00002369, -0.00003025, 0.00002581, -0.00002623, 0.00002714, -0.00001653, 0.00000700, -0.00001179, 0.00002093, -0.00002130, 0.00000703, -0.00001439, 0.00003665, -0.00003824, 0.00001692, -0.00001395, 0.00002946, -0.00002963, 0.00003095, -0.00003247, 0.00002163, -0.00002371, 0.00002459, -0.00002684, 0.00002805, -0.00003173, 0.00003482, -0.00003744, 0.00003646, -0.00005056, 0.00004766, -0.00005737, 0.00005613, -0.00006582, 0.00005840, -0.00008241, 0.00008067, -0.00014994, 0.00014794, -0.00034664, -0.00003700, 0.00111358, -0.00071926, 0.00019270, -0.00017938, 0.00011260, -0.00011431, 0.00007135, -0.00007729, 0.00005598, -0.00005364, 0.00004623, -0.00004461, 0.00003734, -0.00004170, 0.00003174, -0.00002782, 0.00002885, -0.00002672, 0.00002104, -0.00002084, 0.00001749, -0.00002849, 0.00002705, -0.00002543, 0.00002478, -0.00001351, 0.00001538, -0.00003552, 0.00003123, -0.00001035, 0.00000835, -0.00001810, 0.00000985, 0.00000138, 0.00000292, -0.00001010, 0.00000286, 0.00000339, -0.00000397, 0.00000106, -0.00000716, -0.00000068, 0.00000647, -0.00000170, 0.00002564, -0.00003733, 0.00002138, -0.00001596, 0.00003757, -0.00004367, 0.00003901, -0.00002603, 0.00002035, -0.00001740, 0.00001010, -0.00001466, 0.00001043, -0.00000641, 0.00000362, -0.00000394, 0.00000404, -0.00000254, 0.00000199, -0.00000362, 0.00000318, -0.00000610, 0.00000752, },
		{ 0.00000556, -0.00000529, 0.00000498, -0.00000673, 0.00000935, -0.00000428, 0.00000136, -0.00000628, 0.00001203, -0.00002139, 0.00002048, -0.00000105, -0.00000875, 0.00000783, -0.00001371, 0.00002499, -0.00003067, -0.00000634, 0.00002591, -0.00002039, 0.00002145, -0.00001875, 0.00000797, -0.00000551, 0.00000147, 0.00001022, -0.00001339, 0.00000325, 0.00000302, 0.00000274, -0.00001317, -0.00000098, 0.00001413, 0.00000057, -0.00002210, 0.00001473, -0.00000078, 0.00000792, -0.00001083, 0.00001456, -0.00002516, 0.00001991, -0.00001958, 0.00001869, -0.00001959, 0.00001878, -0.00001963, 0.00002175, -0.00002586, 0.00001928, -0.00003107, 0.00003037, -0.00004175, 0.00004393, -0.00006134, 0.00005297, -0.00007647, 0.00005706, -0.00013092, 0.00012706, -0.00066956, 0.00124910, -0.00027814, -0.00020108, 0.00005973, -0.00007429, 0.00003667, -0.00005579, 0.00002922, -0.00003888, 0.00002798, -0.00002709, 0.00002168, -0.00002267, 0.00001463, -0.00002145, 0.00001848, -0.00001353, 0.00001346, -0.00001472, 0.00001022, -0.00000849, -0.00000237, -0.00000320, 0.00000541, -0.00000671, 0.00001338, -0.00000358, -0.00001007, -0.00000650, 0.00002366, -0.00001487, 0.00000496, -0.00001529, 0.00001919, -0.00000804, 0.00000225, -0.00001044, 0.00001200, -0.00000686, 0.00000132, -0.00000265, -0.00000947, 0.00001512, -0.00001456, 0.00003659, -0.00003464, 0.00001250, -0.00001327, 0.00003017, -0.00002248, 0.00000754, 0.00000748, -0.00000890, 0.00000862, -0.00001142, 0.00000501, -0.00000966, 0.00001135, -0.00001037, 0.00000744, -0.00000625, 0.00000632, -0.00000474, 0.00000237, -0.00000251, 0.00000022, -0.00000132, },
		{ 0.00000898, -0.00000653, 0.00000557, -0.00000610, 0.00000538, 0.00000083, -0.00000104, -0.00000476, 0.00000639, -0.00000688, -0.00000492, 0.00002494, -0.00002544, 0.00001815, -0.00001902, 0.00002079, -0.00001075, -0.00002687, 0.00002782, -0.00000860, 0.00000326, 0.00000598, -0.00001801, 0.00001895, -0.00002323, 0.00003051, -0.00002576, 0.00001390, -0.00001375, 0.00002367, -0.00002870, 0.00001338, -0.00001352, 0.00003693, -0.00004834, 0.00003085, -0.00002301, 0.00003656, -0.00003974, 0.00004357, -0.00004927, 0.00004024, -0.00004152, 0.00004234, -0.00004532, 0.00004684, -0.00005064, 0.00005571, -0.00006137, 0.00006007, -0.00007630, 0.00007870, -0.00009317, 0.00009860, -0.00011719, 0.00011555, -0.00014939, 0.00015488, -0.00025004, 0.00029544, -0.00069530, 0.00065247, 0.00058174, -0.00066890, 0.00027697, -0.00023455, 0.00015838, -0.00015448, 0.00010978, -0.00011002, 0.00008781, -0.00008107, 0.00007024, -0.00006656, 0.00005583, -0.00005932, 0.00005042, -0.00004369, 0.00004243, -0.00004038, 0.00003369, -0.00003113, 0.00002424, -0.00003359, 0.00003387, -0.00003316, 0.00003424, -0.00002136, 0.00001766, -0.00003912, 0.00004358, -0.00002442, 0.00001745, -0.00002805, 0.00002375, -0.00000994, 0.00000876, -0.00001728, 0.00001377, -0.00000536, 0.00000035, -0.00000041, -0.00001087, 0.00001170, -0.00000945, 0.00002225, -0.00000534, -0.00001838, 0.00001143, -0.00000190, 0.00001903, -0.00003356, 0.00003931, -0.00002993, 0.00002419, -0.00002259, 0.00001524, -0.00001954, 0.00001755, -0.00001319, 0.00000928, -0.00000820, 0.00000764, -0.00000540, 0.00000396, -0.00000495, 0.00000383, -0.00000633, },
		{ 0.00000394, 0.00000000, -0.00000084, 0.00000111, -0.00000339, 0.00000800, -0.00000502, 0.00000034, -0.00000353, 0.00000801, -0.00002068, 0.00002979, -0.00001610, 0.00000332, -0.00000149, -0.00000157, 0.00001563, -0.00003915, 0.00001715, 0.00000946, -0.00001350, 0.00002168, -0.00002778, 0.00002271, -0.00002355, 0.00002395, -0.00001342, 0.00000438, -0.00001088, 0.00002077, -0.00001911, 0.00000589, -0.00001623, 0.00003637, -0.00003105, 0.00000821, -0.00000962, 0.00002494, -0.00002292, 0.00002271, -0.00002186, 0.00001092, -0.00001485, 0.00001615, -0.00001894, 0.00001972, -0.00002268, 0.00002594, -0.00002818, 0.00002592, -0.00003817, 0.00003438, -0.00004187, 0.00003916, -0.00004685, 0.00003728, -0.00005974, 0.00005240, -0.00010912, 0.00009259, -0.00023674, -0.00023587, 0.00125794, -0.00072130, 0.00015986, -0.00015581, 0.00008969, -0.00009644, 0.00005790, -0.00006625, 0.00004630, -0.00004512, 0.00003757, -0.00003675, 0.00002932, -0.00003425, 0.00002506, -0.00002218, 0.00002353, -0.00002211, 0.00001780, -0.00001873, 0.00001748, -0.00002836, 0.00002471, -0.00002203, 0.00002011, -0.00000964, 0.00001562, -0.00003375, 0.00002319, -0.00000137, 0.00000306, -0.00001405, 0.00000462, 0.00000538, 0.00000075, -0.00000896, 0.00000191, 0.00000536, -0.00000588, 0.00000345, -0.00000970, 0.00000307, 0.00000329, -0.00000058, 0.00002599, -0.00004212, 0.00002627, -0.00002161, 0.00003883, -0.00004238, 0.00003374, -0.00001587, 0.00000997, -0.00000882, 0.00000392, -0.00000895, 0.00000470, -0.00000021, -0.00000136, 0.00000033, 0.00000000, 0.00000107, -0.00000031, -0.00000174, 0.00000136, -0.00000383, },
		{ -0.00000288, 0.00000635, -0.00000556, 0.00000533, -0.00000675, 0.00000774, -0.00000207, -0.00000066, -0.00000485, 0.00000792, -0.00001358, 0.00000857, 0.00001342, -0.00002253, 0.00001964, -0.00002007, 0.00002841, -0.00003254, -0.00000421, 0.00002415, -0.00001734, 0.00001830, -0.00001487, 0.00000365, -0.00000165, -0.00000284, 0.00001410, -0.00001677, 0.00000586, -0.00000055, 0.00000872, -0.00001808, 0.00000351, 0.00000475, 0.00001501, -0.00003385, 0.00002184, -0.00001010, 0.00001964, -0.00002316, 0.00002751, -0.00003626, 0.00002893, -0.00002897, 0.00002839, -0.00003052, 0.00003073, -0.00003194, 0.00003537, -0.00004142, 0.00003670, -0.00004942, 0.00005092, -0.00006357, 0.00006790, -0.00008779, 0.00008260, -0.00011383, 0.00010609, -0.00019112, 0.00021848, -0.00076883, 0.00121634, -0.00012868, -0.00031479, 0.00012895, -0.00012858, 0.00007690, -0.00008767, 0.00005686, -0.00006367, 0.00005043, -0.00004831, 0.00004075, -0.00004063, 0.00003066, -0.00003574, 0.00003127, -0.00002559, 0.00002430, -0.00002417, 0.00001903, -0.00001657, 0.00000749, -0.00001524, 0.00001732, -0.00001831, 0.00002286, -0.00001056, 0.00000114, -0.00002149, 0.00003497, -0.00002114, 0.00001093, -0.00002213, 0.00002401, -0.00001160, 0.00000597, -0.00001492, 0.00001761, -0.00001217, 0.00000687, -0.00000795, -0.00000394, 0.00001128, -0.00001287, 0.00003566, -0.00003555, 0.00001027, -0.00000880, 0.00002016, -0.00001093, -0.00000596, 0.00002135, -0.00001854, 0.00001434, -0.00001467, 0.00000833, -0.00001291, 0.00001431, -0.00001163, 0.00000821, -0.00000686, 0.00000641, -0.00000395, 0.00000197, -0.00000265, 0.00000104, },
		{ -0.00000462, 0.00000594, -0.00000304, 0.00000202, -0.00000180, -0.00000029, 0.00000633, -0.00000623, 0.00000079, -0.00000221, 0.00000373, -0.00001598, 0.00003364, -0.00002948, 0.00001797, -0.00001389, 0.00001543, -0.00000637, -0.00002919, 0.00002949, -0.00000863, 0.00000339, 0.00000628, -0.00001832, 0.00001869, -0.00002331, 0.00003009, -0.00002532, 0.00001377, -0.00001542, 0.00002703, -0.00003062, 0.00001623, -0.00002065, 0.00004449, -0.00004970, 0.00002904, -0.00002499, 0.00003954, -0.00004145, 0.00004413, -0.00004716, 0.00003708, -0.00003982, 0.00004132, -0.00004537, 0.00004720, -0.00005084, 0.00005670, -0.00006306, 0.00006177, -0.00007718, 0.00007983, -0.00009337, 0.00009873, -0.00011710, 0.00011563, -0.00015178, 0.00015796, -0.00024789, 0.00029060, -0.00064251, 0.00046886, 0.00078571, -0.00075757, 0.00030938, -0.00025725, 0.00017123, -0.00016449, 0.00011963, -0.00011938, 0.00009642, -0.00008930, 0.00007699, -0.00007286, 0.00006077, -0.00006351, 0.00005429, -0.00004775, 0.00004605, -0.00004383, 0.00003784, -0.00003596, 0.00003140, -0.00004157, 0.00004011, -0.00003816, 0.00003724, -0.00002398, 0.00002484, -0.00004613, 0.00004421, -0.00002185, 0.00001708, -0.00002886, 0.00002296, -0.00000955, 0.00000984, -0.00001953, 0.00001673, -0.00000762, 0.00000265, -0.00000220, -0.00000927, 0.00001166, -0.00001025, 0.00002198, -0.00000437, -0.00002398, 0.00001904, -0.00001245, 0.00002723, -0.00004004, 0.00004216, -0.00002705, 0.00001882, -0.00001712, 0.00001147, -0.00001609, 0.00001383, -0.00000841, 0.00000509, -0.00000446, 0.00000386, -0.00000163, 0.00000137, -0.00000300, 0.00000222, },
		{ -0.00000174, 0.00000074, 0.00000294, -0.00000348, 0.00000411, -0.00000665, 0.00000998, -0.00000639, 0.00000200, -0.00000670, 0.00000990, -0.00001942, 0.00002364, -0.00000553, -0.00000913, 0.00001181, -0.00001095, 0.00002200, -0.00004294, 0.00002071, 0.00000675, -0.00000982, 0.00001794, -0.00002370, 0.00001809, -0.00001929, 0.00001929, -0.00000923, 0.00000088, -0.00000871, 0.00001862, -0.00001480, 0.00000261, -0.00001464, 0.00003095, -0.00001912, -0.00000339, -0.00000265, 0.00001689, -0.00001242, 0.00001087, -0.00000834, -0.00000174, -0.00000432, 0.00000579, -0.00000875, 0.00000868, -0.00001066, 0.00001371, -0.00001501, 0.00001098, -0.00002109, 0.00001612, -0.00002114, 0.00001678, -0.00002207, 0.00000981, -0.00002930, 0.00001423, -0.00005654, 0.00002354, -0.00011625, -0.00041939, 0.00135263, -0.00067177, 0.00009387, -0.00011041, 0.00005060, -0.00006634, 0.00003448, -0.00004624, 0.00002842, -0.00002900, 0.00002198, -0.00002261, 0.00001558, -0.00002175, 0.00001396, -0.00001243, 0.00001436, -0.00001399, 0.00001122, -0.00001317, 0.00001346, -0.00002333, 0.00001724, -0.00001387, 0.00001135, -0.00000240, 0.00001189, -0.00002671, 0.00001062, 0.00000995, -0.00000359, -0.00000843, -0.00000195, 0.00001014, -0.00000227, -0.00000645, -0.00000076, 0.00000907, -0.00000952, 0.00000753, -0.00001376, 0.00000785, -0.00000080, 0.00000132, 0.00002540, -0.00004485, 0.00002812, -0.00002246, 0.00003424, -0.00003489, 0.00002266, -0.00000210, -0.00000199, 0.00000010, -0.00000211, -0.00000329, -0.00000094, 0.00000544, -0.00000543, 0.00000366, -0.00000308, 0.00000358, -0.00000158, -0.00000055, -0.00000011, },
		{ 0.00000061, -0.00000277, 0.00000515, -0.00000378, 0.00000345, -0.00000424, 0.00000354, 0.00000248, -0.00000508, -0.00000039, 0.00000033, -0.00000251, -0.00000539, 0.00002748, -0.00003388, 0.00002776, -0.00002237, 0.00002859, -0.00003195, -0.00000323, 0.00002279, -0.00001434, 0.00001520, -0.00001153, 0.00000005, 0.00000147, -0.00000623, 0.00001685, -0.00001923, 0.00000817, -0.00000434, 0.00001473, -0.00002306, 0.00000900, -0.00000578, 0.00002865, -0.00004298, 0.00002680, -0.00001805, 0.00002954, -0.00003293, 0.00003703, -0.00004337, 0.00003434, -0.00003531, 0.00003539, -0.00003891, 0.00004008, -0.00004165, 0.00004640, -0.00005410, 0.00005076, -0.00006391, 0.00006701, -0.00008031, 0.00008626, -0.00010804, 0.00010564, -0.00014353, 0.00014455, -0.00023570, 0.00028252, -0.00081639, 0.00112106, 0.00005738, -0.00043684, 0.00019806, -0.00018113, 0.00011472, -0.00011762, 0.00008335, -0.00008761, 0.00007214, -0.00006868, 0.00005885, -0.00005758, 0.00004553, -0.00004892, 0.00004324, -0.00003708, 0.00003471, -0.00003355, 0.00002828, -0.00002574, 0.00001883, -0.00002822, 0.00002935, -0.00002942, 0.00003165, -0.00001770, 0.00001287, -0.00003520, 0.00004326, -0.00002476, 0.00001523, -0.00002738, 0.00002725, -0.00001431, 0.00000952, -0.00001941, 0.00002309, -0.00001727, 0.00001219, -0.00001288, 0.00000100, 0.00000765, -0.00001103, 0.00003387, -0.00003443, 0.00000552, -0.00000121, 0.00000777, 0.00000158, -0.00001851, 0.00003172, -0.00002363, 0.00001564, -0.00001414, 0.00000859, -0.00001329, 0.00001425, -0.00001003, 0.00000654, -0.00000530, 0.00000447, -0.00000149, 0.00000033, -0.00000166, },
		{ 0.00000016, -0.00000212, 0.00000211, 0.00000100, -0.00000201, 0.00000286, -0.00000591, 0.00001128, -0.00001078, 0.00000577, -0.00000967, 0.00001187, -0.00002274, 0.00003613, -0.00002702, 0.00001203, -0.00000439, 0.00000755, -0.00000052, -0.00003200, 0.00003165, -0.00000949, 0.00000488, 0.00000458, -0.00001654, 0.00001635, -0.00002123, 0.00002750, -0.00002310, 0.00001226, -0.00001566, 0.00002832, -0.00003028, 0.00001704, -0.00002466, 0.00004672, -0.00004516, 0.00002277, -0.00002302, 0.00003771, -0.00003773, 0.00003900, -0.00003964, 0.00002942, -0.00003411, 0.00003626, -0.00004113, 0.00004286, -0.00004604, 0.00005228, -0.00005863, 0.00005660, -0.00007055, 0.00007275, -0.00008468, 0.00008931, -0.00010658, 0.00010420, -0.00014042, 0.00014356, -0.00022338, 0.00025554, -0.00055158, 0.00026341, 0.00097222, -0.00081069, 0.00031294, -0.00025841, 0.00016701, -0.00016108, 0.00011850, -0.00011898, 0.00009614, -0.00008926, 0.00007608, -0.00007210, 0.00005915, -0.00006182, 0.00005301, -0.00004716, 0.00004538, -0.00004349, 0.00003862, -0.00003765, 0.00003514, -0.00004518, 0.00004140, -0.00003821, 0.00003573, -0.00002295, 0.00002803, -0.00004746, 0.00003899, -0.00001527, 0.00001398, -0.00002699, 0.00001977, -0.00000755, 0.00000962, -0.00002022, 0.00001782, -0.00000794, 0.00000304, -0.00000205, -0.00000950, 0.00001285, -0.00001188, 0.00002208, -0.00000342, -0.00002889, 0.00002525, -0.00001974, 0.00003046, -0.00004020, 0.00003788, -0.00001824, 0.00000920, -0.00000880, 0.00000568, -0.00001076, 0.00000830, -0.00000232, 0.00000012, -0.00000016, -0.00000033, 0.00000223, -0.00000116, -0.00000091, },
		{ -0.00000066, -0.00000052, -0.00000125, 0.00000427, -0.00000434, 0.00000506, -0.00000739, 0.00000912, -0.00000497, 0.00000072, -0.00000610, 0.00000697, -0.00001265, 0.00001216, 0.00000845, -0.00002289, 0.00002458, -0.00001854, 0.00002655, -0.00004507, 0.00002360, 0.00000388, -0.00000538, 0.00001306, -0.00001872, 0.00001271, -0.00001429, 0.00001400, -0.00000472, -0.00000296, -0.00000564, 0.00001484, -0.00000879, -0.00000247, -0.00001003, 0.00002149, -0.00000436, -0.00001585, 0.00000521, 0.00000707, -0.00000020, -0.00000218, 0.00000556, -0.00001392, 0.00000576, -0.00000458, 0.00000191, -0.00000321, 0.00000238, 0.00000017, -0.00000013, -0.00000584, -0.00000214, -0.00000401, 0.00000137, -0.00000726, 0.00000445, -0.00001959, 0.00000390, -0.00002757, -0.00000047, -0.00004780, -0.00000137, -0.00056910, 0.00138562, -0.00056910, -0.00000137, -0.00004780, -0.00000047, -0.00002757, 0.00000390, -0.00001959, 0.00000445, -0.00000726, 0.00000137, -0.00000401, -0.00000214, -0.00000584, -0.00000013, 0.00000017, 0.00000238, -0.00000321, 0.00000191, -0.00000458, 0.00000576, -0.00001392, 0.00000556, -0.00000218, -0.00000020, 0.00000707, 0.00000521, -0.00001585, -0.00000436, 0.00002149, -0.00001003, -0.00000247, -0.00000879, 0.00001484, -0.00000564, -0.00000296, -0.00000472, 0.00001400, -0.00001429, 0.00001271, -0.00001872, 0.00001306, -0.00000538, 0.00000388, 0.00002360, -0.00004507, 0.00002655, -0.00001854, 0.00002458, -0.00002289, 0.00000845, 0.00001216, -0.00001265, 0.00000697, -0.00000610, 0.00000072, -0.00000497, 0.00000912, -0.00000739, 0.00000506, -0.00000434, 0.00000427, -0.00000125, -0.00000052, },
		{ 0.00000016, -0.00000091, -0.00000116, 0.00000223, -0.00000033, -0.00000016, 0.00000012, -0.00000232, 0.00000830, -0.00001076, 0.00000568, -0.00000880, 0.00000920, -0.00001824, 0.00003788, -0.00004020, 0.00003046, -0.00001974, 0.00002525, -0.00002889, -0.00000342, 0.00002208, -0.00001188, 0.00001285, -0.00000950, -0.00000205, 0.00000304, -0.00000794, 0.00001782, -0.00002022, 0.00000962, -0.00000755, 0.00001977, -0.00002699, 0.00001398, -0.00001527, 0.00003899, -0.00004746, 0.00002803, -0.00002295, 0.00003573, -0.00003821, 0.00004140, -0.00004518, 0.00003514, -0.00003765, 0.00003862, -0.00004349, 0.00004538, -0.00004716, 0.00005301, -0.00006182, 0.00005915, -0.00007210, 0.00007608, -0.00008926, 0.00009614, -0.00011898, 0.00011850, -0.00016108, 0.00016701, -0.00025841, 0.00031294, -0.00081069, 0.00097222, 0.00026341, -0.00055158, 0.00025554, -0.00022338, 0.00014356, -0.00014042, 0.00010420, -0.00010658, 0.00008931, -0.00008468, 0.00007275, -0.00007055, 0.00005660, -0.00005863, 0.00005228, -0.00004604, 0.00004286, -0.00004113, 0.00003626, -0.00003411, 0.00002942, -0.00003964, 0.00003900, -0.00003773, 0.00003771, -0.00002302, 0.00002277, -0.00004516, 0.00004672, -0.00002466, 0.00001704, -0.00003028, 0.00002832, -0.00001566, 0.00001226, -0.00002310, 0.00002750, -0.00002123, 0.00001635, -0.00001654, 0.00000458, 0.00000488, -0.00000949, 0.00003165, -0.00003200, -0.00000052, 0.00000755, -0.00000439, 0.00001203, -0.00002702, 0.00003613, -0.00002274, 0.00001187, -0.00000967, 0.00000577, -0.00001078, 0.00001128, -0.00000591, 0.00000286, -0.00000201, 0.00000100, 0.00000211, -0.00000212, },
		{ 0.00000061, -0.00000166, 0.00000033, -0.00000149, 0.00000447, -0.00000530, 0.00000654, -0.00001003, 0.00001425, -0.00001329, 0.00000859, -0.00001414, 0.00001564, -0.00002363, 0.00003172, -0.00001851, 0.00000158, 0.00000777, -0.00000121, 0.00000552, -0.00003443, 0.00003387, -0.00001103, 0.00000765, 0.00000100, -0.00001288, 0.00001219, -0.00001727, 0.00002309, -0.00001941, 0.00000952, -0.00001431, 0.00002725, -0.00002738, 0.00001523, -0.00002476, 0.00004326, -0.00003520, 0.00001287, -0.00001770, 0.00003165, -0.00002942, 0.00002935, -0.00002822, 0.00001883, -0.00002574, 0.00002828, -0.00003355, 0.00003471, -0.00003708, 0.00004324, -0.00004892, 0.00004553, -0.00005758, 0.00005885, -0.00006868, 0.00007214, -0.00008761, 0.00008335, -0.00011762, 0.00011472, -0.00018113, 0.00019806, -0.00043684, 0.00005738, 0.00112106, -0.00081639, 0.00028252, -0.00023570, 0.00014455, -0.00014353, 0.00010564, -0.00010804, 0.00008626, -0.00008031, 0.00006701, -0.00006391, 0.00005076, -0.00005410, 0.00004640, -0.00004165, 0.00004008, -0.00003891, 0.00003539, -0.00003531, 0.00003434, -0.00004337, 0.00003703, -0.00003293, 0.00002954, -0.00001805, 0.00002680, -0.00004298, 0.00002865, -0.00000578, 0.00000900, -0.00002306, 0.00001473, -0.00000434, 0.00000817, -0.00001923, 0.00001685, -0.00000623, 0.00000147, 0.00000005, -0.00001153, 0.00001520, -0.00001434, 0.00002279, -0.00000323, -0.00003195, 0.00002859, -0.00002237, 0.00002776, -0.00003388, 0.00002748, -0.00000539, -0.00000251, 0.00000033, -0.00000039, -0.00000508, 0.00000248, 0.00000354, -0.00000424, 0.00000345, -0.00000378, 0.00000515, -0.00000277, },
		{ -0.00000174, -0.00000011, -0.00000055, -0.00000158, 0.00000358, -0.00000308, 0.00000366, -0.00000543, 0.00000544, -0.00000094, -0.00000329, -0.00000211, 0.00000010, -0.00000199, -0.00000210, 0.00002266, -0.00003489, 0.00003424, -0.00002246, 0.00002812, -0.00004485, 0.00002540, 0.00000132, -0.00000080, 0.00000785, -0.00001376, 0.00000753, -0.00000952, 0.00000907, -0.00000076, -0.00000645, -0.00000227, 0.00001014, -0.00000195, -0.00000843, -0.00000359, 0.00000995, 0.00001062, -0.00002671, 0.00001189, -0.00000240, 0.00001135, -0.00001387, 0.00001724, -0.00002333, 0.00001346, -0.00001317, 0.00001122, -0.00001399, 0.00001436, -0.00001243, 0.00001396, -0.00002175, 0.00001558, -0.00002261, 0.00002198, -0.00002900, 0.00002842, -0.00004624, 0.00003448, -0.00006634, 0.00005060, -0.00011041, 0.00009387, -0.00067177, 0.00135263, -0.00041939, -0.00011625, 0.00002354, -0.00005654, 0.00001423, -0.00002930, 0.00000981, -0.00002207, 0.00001678, -0.00002114, 0.00001612, -0.00002109, 0.00001098, -0.00001501, 0.00001371, -0.00001066, 0.00000868, -0.00000875, 0.00000579, -0.00000432, -0.00000174, -0.00000834, 0.00001087, -0.00001242, 0.00001689, -0.00000265, -0.00000339, -0.00001912, 0.00003095, -0.00001464, 0.00000261, -0.00001480, 0.00001862, -0.00000871, 0.00000088, -0.00000923, 0.00001929, -0.00001929, 0.00001809, -0.00002370, 0.00001794, -0.00000982, 0.00000675, 0.00002071, -0.00004294, 0.00002200, -0.00001095, 0.00001181, -0.00000913, -0.00000553, 0.00002364, -0.00001942, 0.00000990, -0.00000670, 0.00000200, -0.00000639, 0.00000998, -0.00000665, 0.00000411, -0.00000348, 0.00000294, 0.00000074, },
		{ -0.00000462, 0.00000222, -0.00000300, 0.00000137, -0.00000163, 0.00000386, -0.00000446, 0.00000509, -0.00000841, 0.00001383, -0.00001609, 0.00001147, -0.00001712, 0.00001882, -0.00002705, 0.00004216, -0.00004004, 0.00002723, -0.00001245, 0.00001904, -0.00002398, -0.00000437, 0.00002198, -0.00001025, 0.00001166, -0.00000927, -0.00000220, 0.00000265, -0.00000762, 0.00001673, -0.00001953, 0.00000984, -0.00000955, 0.00002296, -0.00002886, 0.00001708, -0.00002185, 0.00004421, -0.00004613, 0.00002484, -0.00002398, 0.00003724, -0.00003816, 0.00004011, -0.00004157, 0.00003140, -0.00003596, 0.00003784, -0.00004383, 0.00004605, -0.00004775, 0.00005429, -0.00006351, 0.00006077, -0.00007286, 0.00007699, -0.00008930, 0.00009642, -0.00011938, 0.00011963, -0.00016449, 0.00017123, -0.00025725, 0.00030938, -0.00075757, 0.00078571, 0.00046886, -0.00064251, 0.00029060, -0.00024789, 0.00015796, -0.00015178, 0.00011563, -0.00011710, 0.00009873, -0.00009337, 0.00007983, -0.00007718, 0.00006177, -0.00006306, 0.00005670, -0.00005084, 0.00004720, -0.00004537, 0.00004132, -0.00003982, 0.00003708, -0.00004716, 0.00004413, -0.00004145, 0.00003954, -0.00002499, 0.00002904, -0.00004970, 0.00004449, -0.00002065, 0.00001623, -0.00003062, 0.00002703, -0.00001542, 0.00001377, -0.00002532, 0.00003009, -0.00002331, 0.00001869, -0.00001832, 0.00000628, 0.00000339, -0.00000863, 0.00002949, -0.00002919, -0.00000637, 0.00001543, -0.00001389, 0.00001797, -0.00002948, 0.00003364, -0.00001598, 0.00000373, -0.00000221, 0.00000079, -0.00000623, 0.00000633, -0.00000029, -0.00000180, 0.00000202, -0.00000304, 0.00000594, },
		{ -0.00000288, 0.00000104, -0.00000265, 0.00000197, -0.00000395, 0.00000641, -0.00000686, 0.00000821, -0.00001163, 0.00001431, -0.00001291, 0.00000833, -0.00001467, 0.00001434, -0.00001854, 0.00002135, -0.00000596, -0.00001093, 0.00002016, -0.00000880, 0.00001027, -0.00003555, 0.00003566, -0.00001287, 0.00001128, -0.00000394, -0.00000795, 0.00000687, -0.00001217, 0.00001761, -0.00001492, 0.00000597, -0.00001160, 0.00002401, -0.00002213, 0.00001093, -0.00002114, 0.00003497, -0.00002149, 0.00000114, -0.00001056, 0.00002286, -0.00001831, 0.00001732, -0.00001524, 0.00000749, -0.00001657, 0.00001903, -0.00002417, 0.00002430, -0.00002559, 0.00003127, -0.00003574, 0.00003066, -0.00004063, 0.00004075, -0.00004831, 0.00005043, -0.00006367, 0.00005686, -0.00008767, 0.00007690, -0.00012858, 0.00012895, -0.00031479, -0.00012868, 0.00121634, -0.00076883, 0.00021848, -0.00019112, 0.00010609, -0.00011383, 0.00008260, -0.00008779, 0.00006790, -0.00006357, 0.00005092, -0.00004942, 0.00003670, -0.00004142, 0.00003537, -0.00003194, 0.00003073, -0.00003052, 0.00002839, -0.00002897, 0.00002893, -0.00003626, 0.00002751, -0.00002316, 0.00001964, -0.00001010, 0.00002184, -0.00003385, 0.00001501, 0.00000475, 0.00000351, -0.00001808, 0.00000872, -0.00000055, 0.00000586, -0.00001677, 0.00001410, -0.00000284, -0.00000165, 0.00000365, -0.00001487, 0.00001830, -0.00001734, 0.00002415, -0.00000421, -0.00003254, 0.00002841, -0.00002007, 0.00001964, -0.00002253, 0.00001342, 0.00000857, -0.00001358, 0.00000792, -0.00000485, -0.00000066, -0.00000207, 0.00000774, -0.00000675, 0.00000533, -0.00000556, 0.00000635, },
		{ 0.00000394, -0.00000383, 0.00000136, -0.00000174, -0.00000031, 0.00000107, 0.00000000, 0.00000033, -0.00000136, -0.00000021, 0.00000470, -0.00000895, 0.00000392, -0.00000882, 0.00000997, -0.00001587, 0.00003374, -0.00004238, 0.00003883, -0.00002161, 0.00002627, -0.00004212, 0.00002599, -0.00000058, 0.00000329, 0.00000307, -0.00000970, 0.00000345, -0.00000588, 0.00000536, 0.00000191, -0.00000896, 0.00000075, 0.00000538, 0.00000462, -0.00001405, 0.00000306, -0.00000137, 0.00002319, -0.00003375, 0.00001562, -0.00000964, 0.00002011, -0.00002203, 0.00002471, -0.00002836, 0.00001748, -0.00001873, 0.00001780, -0.00002211, 0.00002353, -0.00002218, 0.00002506, -0.00003425, 0.00002932, -0.00003675, 0.00003757, -0.00004512, 0.00004630, -0.00006625, 0.00005790, -0.00009644, 0.00008969, -0.00015581, 0.00015986, -0.00072130, 0.00125794, -0.00023587, -0.00023674, 0.00009259, -0.00010912, 0.00005240, -0.00005974, 0.00003728, -0.00004685, 0.00003916, -0.00004187, 0.00003438, -0.00003817, 0.00002592, -0.00002818, 0.00002594, -0.00002268, 0.00001972, -0.00001894, 0.00001615, -0.00001485, 0.00001092, -0.00002186, 0.00002271, -0.00002292, 0.00002494, -0.00000962, 0.00000821, -0.00003105, 0.00003637, -0.00001623, 0.00000589, -0.00001911, 0.00002077, -0.00001088, 0.00000438, -0.00001342, 0.00002395, -0.00002355, 0.00002271, -0.00002778, 0.00002168, -0.00001350, 0.00000946, 0.00001715, -0.00003915, 0.00001563, -0.00000157, -0.00000149, 0.00000332, -0.00001610, 0.00002979, -0.00002068, 0.00000801, -0.00000353, 0.00000034, -0.00000502, 0.00000800, -0.00000339, 0.00000111, -0.00000084, 0.00000000, },
		{ 0.00000898, -0.00000633, 0.00000383, -0.00000495, 0.00000396, -0.00000540, 0.00000764, -0.00000820, 0.00000928, -0.00001319, 0.00001755, -0.00001954, 0.00001524, -0.00002259, 0.00002419, -0.00002993, 0.00003931, -0.00003356, 0.00001903, -0.00000190, 0.00001143, -0.00001838, -0.00000534, 0.00002225, -0.00000945, 0.00001170, -0.00001087, -0.00000041, 0.00000035, -0.00000536, 0.00001377, -0.00001728, 0.00000876, -0.00000994, 0.00002375, -0.00002805, 0.00001745, -0.00002442, 0.00004358, -0.00003912, 0.00001766, -0.00002136, 0.00003424, -0.00003316, 0.00003387, -0.00003359, 0.00002424, -0.00003113, 0.00003369, -0.00004038, 0.00004243, -0.00004369, 0.00005042, -0.00005932, 0.00005583, -0.00006656, 0.00007024, -0.00008107, 0.00008781, -0.00011002, 0.00010978, -0.00015448, 0.00015838, -0.00023455, 0.00027697, -0.00066890, 0.00058174, 0.00065247, -0.00069530, 0.00029544, -0.00025004, 0.00015488, -0.00014939, 0.00011555, -0.00011719, 0.00009860, -0.00009317, 0.00007870, -0.00007630, 0.00006007, -0.00006137, 0.00005571, -0.00005064, 0.00004684, -0.00004532, 0.00004234, -0.00004152, 0.00004024, -0.00004927, 0.00004357, -0.00003974, 0.00003656, -0.00002301, 0.00003085, -0.00004834, 0.00003693, -0.00001352, 0.00001338, -0.00002870, 0.00002367, -0.00001375, 0.00001390, -0.00002576, 0.00003051, -0.00002323, 0.00001895, -0.00001801, 0.00000598, 0.00000326, -0.00000860, 0.00002782, -0.00002687, -0.00001075, 0.00002079, -0.00001902, 0.00001815, -0.00002544, 0.00002494, -0.00000492, -0.00000688, 0.00000639, -0.00000476, -0.00000104, 0.00000083, 0.00000538, -0.00000610, 0.00000557, -0.00000653, },
		{ 0.00000556, -0.00000132, 0.00000022, -0.00000251, 0.00000237, -0.00000474, 0.00000632, -0.00000625, 0.00000744, -0.00001037, 0.00001135, -0.00000966, 0.00000501, -0.00001142, 0.00000862, -0.00000890, 0.00000748, 0.00000754, -0.00002248, 0.00003017, -0.00001327, 0.00001250, -0.00003464, 0.00003659, -0.00001456, 0.00001512, -0.00000947, -0.00000265, 0.00000132, -0.00000686, 0.00001200, -0.00001044, 0.00000225, -0.00000804, 0.00001919, -0.00001529, 0.00000496, -0.00001487, 0.00002366, -0.00000650, -0.00001007, -0.00000358, 0.00001338, -0.00000671, 0.00000541, -0.00000320, -0.00000237, -0.00000849, 0.00001022, -0.00001472, 0.00001346, -0.00001353, 0.00001848, -0.00002145, 0.00001463, -0.00002267, 0.00002168, -0.00002709, 0.00002798, -0.00003888, 0.00002922, -0.00005579, 0.00003667, -0.00007429, 0.00005973, -0.00020108, -0.00027814, 0.00124910, -0.00066956, 0.00012706, -0.00013092, 0.00005706, -0.00007647, 0.00005297, -0.00006134, 0.00004393, -0.00004175, 0.00003037, -0.00003107, 0.00001928, -0.00002586, 0.00002175, -0.00001963, 0.00001878, -0.00001959, 0.00001869, -0.00001958, 0.00001991, -0.00002516, 0.00001456, -0.00001083, 0.00000792, -0.00000078, 0.00001473, -0.00002210, 0.00000057, 0.00001413, -0.00000098, -0.00001317, 0.00000274, 0.00000302, 0.00000325, -0.00001339, 0.00001022, 0.00000147, -0.00000551, 0.00000797, -0.00001875, 0.00002145, -0.00002039, 0.00002591, -0.00000634, -0.00003067, 0.00002499, -0.00001371, 0.00000783, -0.00000875, -0.00000105, 0.00002048, -0.00002139, 0.00001203, -0.00000628, 0.00000136, -0.00000428, 0.00000935, -0.00000673, 0.00000498, -0.00000529, },
		{ -0.00000367, 0.00000752, -0.00000610, 0.00000318, -0.00000362, 0.00000199, -0.00000254, 0.00000404, -0.00000394, 0.00000362, -0.00000641, 0.00001043, -0.00001466, 0.00001010, -0.00001740, 0.00002035, -0.00002603, 0.00003901, -0.00004367, 0.00003757, -0.00001596, 0.00002138, -0.00003733, 0.00002564, -0.00000170, 0.00000647, -0.00000068, -0.00000716, 0.00000106, -0.00000397, 0.00000339, 0.00000286, -0.00001010, 0.00000292, 0.00000138, 0.00000985, -0.00001810, 0.00000835, -0.00001035, 0.00003123, -0.00003552, 0.00001538, -0.00001351, 0.00002478, -0.00002543, 0.00002705, -0.00002849, 0.00001749, -0.00002084, 0.00002104, -0.00002672, 0.00002885, -0.00002782, 0.00003174, -0.00004170, 0.00003734, -0.00004461, 0.00004623, -0.00005364, 0.00005598, -0.00007729, 0.00007135, -0.00011431, 0.00011260, -0.00017938, 0.00019270, -0.00071926, 0.00111358, -0.00003700, -0.00034664, 0.00014794, -0.00014994, 0.00008067, -0.00008241, 0.00005840, -0.00006582, 0.00005613, -0.00005737, 0.00004766, -0.00005056, 0.00003646, -0.00003744, 0.00003482, -0.00003173, 0.00002805, -0.00002684, 0.00002459, -0.00002371, 0.00002163, -0.00003247, 0.00003095, -0.00002963, 0.00002946, -0.00001395, 0.00001692, -0.00003824, 0.00003665, -0.00001439, 0.00000703, -0.00002130, 0.00002093, -0.00001179, 0.00000700, -0.00001653, 0.00002714, -0.00002623, 0.00002581, -0.00003025, 0.00002369, -0.00001591, 0.00001157, 0.00001346, -0.00003464, 0.00000892, 0.00000752, -0.00001274, 0.00001182, -0.00002104, 0.00002940, -0.00001624, 0.00000175, 0.00000267, -0.00000353, -0.00000152, 0.00000393, 0.00000150, -0.00000301, 0.00000271, },
		{ -0.00000852, 0.00001041, -0.00000661, 0.00000442, -0.00000613, 0.00000581, -0.00000812, 0.00001001, -0.00001032, 0.00001156, -0.00001550, 0.00001841, -0.00002009, 0.00001589, -0.00002403, 0.00002434, -0.00002649, 0.00003004, -0.00002253, 0.00000806, 0.00000963, 0.00000437, -0.00001350, -0.00000546, 0.00002246, -0.00000922, 0.00001267, -0.00001391, 0.00000283, -0.00000335, -0.00000175, 0.00000951, -0.00001396, 0.00000664, -0.00000876, 0.00002211, -0.00002454, 0.00001493, -0.00002286, 0.00003765, -0.00002773, 0.00000798, -0.00001630, 0.00002790, -0.00002465, 0.00002445, -0.00002322, 0.00001548, -0.00002467, 0.00002743, -0.00003426, 0.00003562, -0.00003608, 0.00004254, -0.00005052, 0.00004582, -0.00005489, 0.00005772, -0.00006669, 0.00007267, -0.00009343, 0.00009163, -0.00013410, 0.00013248, -0.00019611, 0.00022480, -0.00056003, 0.00038144, 0.00079575, -0.00070052, 0.00026710, -0.00022919, 0.00013448, -0.00013359, 0.00010408, -0.00010685, 0.00008896, -0.00008420, 0.00006961, -0.00006821, 0.00005188, -0.00005393, 0.00004957, -0.00004557, 0.00004181, -0.00004089, 0.00003908, -0.00003876, 0.00003836, -0.00004565, 0.00003740, -0.00003302, 0.00002934, -0.00001753, 0.00002853, -0.00004185, 0.00002554, -0.00000483, 0.00000959, -0.00002531, 0.00001888, -0.00001109, 0.00001286, -0.00002446, 0.00002885, -0.00002115, 0.00001739, -0.00001590, 0.00000399, 0.00000423, -0.00000929, 0.00002680, -0.00002559, -0.00001286, 0.00002273, -0.00001923, 0.00001273, -0.00001607, 0.00001223, 0.00000778, -0.00001739, 0.00001390, -0.00000906, 0.00000323, -0.00000371, 0.00000966, -0.00000880, 0.00000760, },
		{ -0.00000318, 0.00000308, 0.00000155, -0.00000155, -0.00000139, 0.00000151, -0.00000386, 0.00000429, -0.00000365, 0.00000455, -0.00000674, 0.00000609, -0.00000441, -0.00000040, -0.00000564, 0.00000026, 0.00000275, -0.00000663, 0.00001857, -0.00003026, 0.00003571, -0.00001332, 0.00001158, -0.00003141, 0.00003647, -0.00001570, 0.00001854, -0.00001474, 0.00000212, -0.00000354, -0.00000229, 0.00000716, -0.00000675, -0.00000101, -0.00000426, 0.00001365, -0.00000797, -0.00000146, -0.00000756, 0.00001166, 0.00000709, -0.00001845, 0.00000138, 0.00000514, 0.00000321, -0.00000410, 0.00000578, -0.00000901, -0.00000292, 0.00000321, -0.00000667, 0.00000386, -0.00000274, 0.00000691, -0.00000836, 0.00000008, -0.00000659, 0.00000476, -0.00000834, 0.00000835, -0.00001707, 0.00000468, -0.00002696, 0.00000026, -0.00002612, 0.00000039, -0.00010799, -0.00038071, 0.00121866, -0.00052765, 0.00001967, -0.00006465, 0.00000508, -0.00003750, 0.00002168, -0.00003300, 0.00001833, -0.00001857, 0.00000887, -0.00001209, 0.00000148, -0.00001013, 0.00000792, -0.00000686, 0.00000619, -0.00000796, 0.00000801, -0.00000884, 0.00000912, -0.00001226, 0.00000067, 0.00000157, -0.00000326, 0.00000783, 0.00000750, -0.00001013, -0.00001208, 0.00002040, -0.00000328, -0.00000925, -0.00000231, 0.00000565, 0.00000093, -0.00000976, 0.00000603, 0.00000575, -0.00000918, 0.00001208, -0.00002229, 0.00002393, -0.00002291, 0.00002767, -0.00000931, -0.00002689, 0.00001931, -0.00000500, -0.00000517, 0.00000439, -0.00001274, 0.00002771, -0.00002420, 0.00001160, -0.00000415, 0.00000060, -0.00000382, 0.00000819, -0.00000423, 0.00000256, },
		{ 0.00000599, -0.00000698, 0.00001047, -0.00000771, 0.00000468, -0.00000550, 0.00000451, -0.00000625, 0.00000789, -0.00000790, 0.00000811, -0.00001161, 0.00001465, -0.00001880, 0.00001456, -0.00002345, 0.00002680, -0.00003042, 0.00003718, -0.00003866, 0.00003110, -0.00000666, 0.00001474, -0.00003151, 0.00002498, -0.00000220, 0.00000860, -0.00000320, -0.00000633, 0.00000053, -0.00000390, 0.00000324, 0.00000207, -0.00000982, 0.00000397, -0.00000128, 0.00001297, -0.00001973, 0.00001112, -0.00001559, 0.00003370, -0.00003176, 0.00001123, -0.00001387, 0.00002513, -0.00002403, 0.00002455, -0.00002435, 0.00001416, -0.00001993, 0.00002108, -0.00002774, 0.00003011, -0.00002907, 0.00003355, -0.00004358, 0.00003912, -0.00004575, 0.00004756, -0.00005421, 0.00005711, -0.00007892, 0.00007415, -0.00011901, 0.00011845, -0.00018087, 0.00019425, -0.00067393, 0.00093718, 0.00015653, -0.00043050, 0.00018008, -0.00017271, 0.00009456, -0.00009376, 0.00007000, -0.00007605, 0.00006501, -0.00006527, 0.00005385, -0.00005640, 0.00004100, -0.00004137, 0.00003903, -0.00003645, 0.00003232, -0.00003105, 0.00002960, -0.00002918, 0.00002847, -0.00003826, 0.00003401, -0.00003133, 0.00002951, -0.00001469, 0.00002159, -0.00003989, 0.00003184, -0.00000959, 0.00000633, -0.00002143, 0.00001914, -0.00001137, 0.00000841, -0.00001807, 0.00002834, -0.00002689, 0.00002699, -0.00003079, 0.00002373, -0.00001687, 0.00001283, 0.00001017, -0.00003035, 0.00000324, 0.00001451, -0.00002004, 0.00001484, -0.00001964, 0.00002289, -0.00000736, -0.00000728, 0.00001025, -0.00000818, 0.00000282, -0.00000091, 0.00000664, -0.00000700, },
		{ 0.00000751, -0.00000852, 0.00000989, -0.00000524, 0.00000377, -0.00000620, 0.00000643, -0.00000920, 0.00001035, -0.00001025, 0.00001137, -0.00001489, 0.00001611, -0.00001760, 0.00001323, -0.00002140, 0.00001966, -0.00001792, 0.00001661, -0.00000989, -0.00000280, 0.00001961, -0.00000019, -0.00001059, -0.00000399, 0.00002223, -0.00000919, 0.00001404, -0.00001770, 0.00000675, -0.00000761, 0.00000239, 0.00000479, -0.00001027, 0.00000396, -0.00000634, 0.00001844, -0.00001890, 0.00001013, -0.00001803, 0.00002798, -0.00001422, -0.00000207, -0.00001056, 0.00002002, -0.00001470, 0.00001415, -0.00001275, 0.00000715, -0.00001824, 0.00002053, -0.00002693, 0.00002714, -0.00002657, 0.00003243, -0.00003907, 0.00003298, -0.00004038, 0.00004222, -0.00004919, 0.00005426, -0.00007313, 0.00006905, -0.00010786, 0.00009932, -0.00014967, 0.00016377, -0.00044697, 0.00020353, 0.00088591, -0.00065548, 0.00020833, -0.00018906, 0.00010027, -0.00010730, 0.00008354, -0.00008809, 0.00007173, -0.00006831, 0.00005435, -0.00005467, 0.00003886, -0.00004228, 0.00003959, -0.00003674, 0.00003309, -0.00003293, 0.00003220, -0.00003210, 0.00003204, -0.00003720, 0.00002692, -0.00002284, 0.00001944, -0.00000995, 0.00002338, -0.00003200, 0.00001257, 0.00000344, 0.00000616, -0.00002138, 0.00001352, -0.00000810, 0.00001105, -0.00002183, 0.00002559, -0.00001769, 0.00001464, -0.00001264, 0.00000097, 0.00000577, -0.00001035, 0.00002635, -0.00002553, -0.00001249, 0.00002127, -0.00001510, 0.00000319, -0.00000374, -0.00000144, 0.00001909, -0.00002527, 0.00001833, -0.00001066, 0.00000537, -0.00000616, 0.00001157, -0.00000915, },
		{ -0.00000111, -0.00000004, -0.00000028, 0.00000493, -0.00000364, 0.00000024, -0.00000018, -0.00000178, 0.00000096, 0.00000013, 0.00000043, -0.00000185, -0.00000011, 0.00000135, -0.00000635, 0.00000082, -0.00000835, 0.00001350, -0.00001773, 0.00002432, -0.00003244, 0.00003576, -0.00000869, 0.00000774, -0.00002620, 0.00003551, -0.00001612, 0.00002101, -0.00001910, 0.00000565, -0.00000701, 0.00000086, 0.00000373, -0.00000437, -0.00000332, -0.00000087, 0.00000830, -0.00000132, -0.00000700, -0.00000092, 0.00000122, 0.00001700, -0.00002237, 0.00000310, -0.00000051, 0.00000990, -0.00000975, 0.00001050, -0.00001164, -0.00000046, -0.00000129, -0.00000093, -0.00000337, 0.00000542, -0.00000188, 0.00000172, -0.00001097, 0.00000545, -0.00000771, 0.00000548, -0.00000588, -0.00000103, -0.00001363, -0.00000493, -0.00002762, 0.00001021, -0.00004231, -0.00004267, -0.00043352, 0.00113247, -0.00035825, -0.00008907, -0.00000325, -0.00004164, -0.00000331, -0.00000609, -0.00000735, -0.00000465, 0.00000206, -0.00001001, 0.00000425, -0.00001373, 0.00000313, -0.00000377, 0.00000421, -0.00000494, 0.00000237, -0.00000171, 0.00000126, -0.00000128, -0.00000004, -0.00001153, 0.00001154, -0.00001171, 0.00001384, 0.00000199, -0.00000011, -0.00002092, 0.00002233, -0.00000283, -0.00000681, -0.00000589, 0.00000686, -0.00000063, -0.00000652, 0.00000236, 0.00000912, -0.00001179, 0.00001517, -0.00002473, 0.00002511, -0.00002443, 0.00002902, -0.00001257, -0.00002211, 0.00001277, 0.00000402, -0.00001677, 0.00001418, -0.00001931, 0.00002887, -0.00002156, 0.00000683, 0.00000101, -0.00000237, -0.00000120, 0.00000491, -0.00000002, },
		{ -0.00000965, 0.00000795, -0.00000896, 0.00001196, -0.00000801, 0.00000527, -0.00000672, 0.00000642, -0.00000906, 0.00001049, -0.00001041, 0.00001087, -0.00001453, 0.00001617, -0.00002020, 0.00001600, -0.00002556, 0.00002805, -0.00002835, 0.00002871, -0.00002885, 0.00002135, 0.00000427, 0.00000815, -0.00002597, 0.00002480, -0.00000238, 0.00000979, -0.00000460, -0.00000701, 0.00000153, -0.00000534, 0.00000454, -0.00000005, -0.00000842, 0.00000396, -0.00000239, 0.00001368, -0.00001862, 0.00001096, -0.00001666, 0.00003075, -0.00002340, 0.00000429, -0.00001156, 0.00002195, -0.00001886, 0.00001856, -0.00001748, 0.00000890, -0.00001708, 0.00001875, -0.00002584, 0.00002789, -0.00002646, 0.00003104, -0.00004050, 0.00003544, -0.00004108, 0.00004263, -0.00004805, 0.00005102, -0.00007251, 0.00006768, -0.00011207, 0.00010945, -0.00016386, 0.00017099, -0.00059831, 0.00074888, 0.00032515, -0.00047644, 0.00018341, -0.00017460, 0.00009241, -0.00009259, 0.00007087, -0.00007636, 0.00006482, -0.00006474, 0.00005232, -0.00005521, 0.00003919, -0.00003969, 0.00003819, -0.00003641, 0.00003202, -0.00003100, 0.00003046, -0.00003035, 0.00003043, -0.00003844, 0.00003152, -0.00002801, 0.00002525, -0.00001194, 0.00002217, -0.00003639, 0.00002308, -0.00000312, 0.00000461, -0.00002002, 0.00001583, -0.00000988, 0.00000861, -0.00001792, 0.00002747, -0.00002555, 0.00002632, -0.00002949, 0.00002198, -0.00001651, 0.00001320, 0.00000757, -0.00002699, -0.00000042, 0.00001832, -0.00002254, 0.00001223, -0.00001275, 0.00001214, 0.00000359, -0.00001672, 0.00001712, -0.00001188, 0.00000652, -0.00000504, 0.00001067, },
		{ -0.00000710, 0.00000539, -0.00000668, 0.00000762, -0.00000250, 0.00000210, -0.00000522, 0.00000573, -0.00000853, 0.00000867, -0.00000808, 0.00000889, -0.00001169, 0.00001124, -0.00001278, 0.00000807, -0.00001576, 0.00001179, -0.00000663, 0.00000224, 0.00000098, -0.00001077, 0.00002587, -0.00000083, -0.00001041, -0.00000059, 0.00002136, -0.00000899, 0.00001519, -0.00002145, 0.00001045, -0.00001156, 0.00000616, 0.00000046, -0.00000695, 0.00000131, -0.00000325, 0.00001351, -0.00001211, 0.00000418, -0.00001143, 0.00001679, -0.00000117, -0.00001024, -0.00000595, 0.00001247, -0.00000542, 0.00000518, -0.00000426, 0.00000093, -0.00001321, 0.00001431, -0.00001976, 0.00001856, -0.00001691, 0.00002202, -0.00002717, 0.00001981, -0.00002575, 0.00002668, -0.00003178, 0.00003601, -0.00005283, 0.00004614, -0.00008061, 0.00006503, -0.00010315, 0.00010435, -0.00034351, 0.00006158, 0.00091772, -0.00056486, 0.00012732, -0.00013706, 0.00005837, -0.00007549, 0.00005796, -0.00006448, 0.00005022, -0.00004866, 0.00003590, -0.00003844, 0.00002358, -0.00002873, 0.00002781, -0.00002596, 0.00002236, -0.00002299, 0.00002315, -0.00002296, 0.00002284, -0.00002585, 0.00001438, -0.00001148, 0.00000905, -0.00000216, 0.00001723, -0.00002104, 0.00000045, 0.00000943, 0.00000419, -0.00001776, 0.00000841, -0.00000544, 0.00000902, -0.00001846, 0.00002151, -0.00001371, 0.00001156, -0.00000907, -0.00000227, 0.00000719, -0.00001130, 0.00002619, -0.00002647, -0.00001005, 0.00001720, -0.00000811, -0.00000816, 0.00000859, -0.00001296, 0.00002641, -0.00002867, 0.00001853, -0.00000892, 0.00000494, -0.00000610, 0.00001083, },
		{ 0.00000465, -0.00000487, 0.00000306, -0.00000350, 0.00000787, -0.00000525, 0.00000170, -0.00000203, 0.00000070, -0.00000267, 0.00000399, -0.00000366, 0.00000288, -0.00000569, 0.00000602, -0.00001115, 0.00000600, -0.00001490, 0.00002080, -0.00002342, 0.00002329, -0.00002868, 0.00003068, -0.00000023, 0.00000201, -0.00001989, 0.00003424, -0.00001588, 0.00002230, -0.00002220, 0.00000759, -0.00000880, 0.00000231, 0.00000193, -0.00000348, -0.00000449, 0.00000171, 0.00000388, 0.00000371, -0.00001060, 0.00000369, -0.00000602, 0.00002190, -0.00002123, 0.00000132, -0.00000312, 0.00001281, -0.00001112, 0.00001090, -0.00001051, -0.00000080, -0.00000336, 0.00000230, -0.00000777, 0.00001032, -0.00000710, 0.00000786, -0.00001746, 0.00001236, -0.00001464, 0.00001325, -0.00001356, 0.00000800, -0.00002426, 0.00000842, -0.00004459, 0.00003199, -0.00006564, -0.00000667, -0.00044075, 0.00100448, -0.00018005, -0.00018394, 0.00004312, -0.00007598, 0.00002083, -0.00002610, 0.00001177, -0.00002149, 0.00001693, -0.00002335, 0.00001535, -0.00002405, 0.00001189, -0.00001152, 0.00001188, -0.00001294, 0.00000975, -0.00000880, 0.00000897, -0.00000930, 0.00000930, -0.00001990, 0.00001724, -0.00001593, 0.00001606, -0.00000063, 0.00000659, -0.00002488, 0.00001970, 0.00000010, -0.00000574, -0.00000787, 0.00000653, -0.00000122, -0.00000408, -0.00000020, 0.00001096, -0.00001279, 0.00001675, -0.00002566, 0.00002470, -0.00002469, 0.00002964, -0.00001560, -0.00001729, 0.00000681, 0.00001147, -0.00002491, 0.00001887, -0.00001982, 0.00002412, -0.00001444, -0.00000092, 0.00000775, -0.00000627, 0.00000242, 0.00000069, },
		{ 0.00001256, -0.00001014, 0.00000796, -0.00000909, 0.00001157, -0.00000672, 0.00000467, -0.00000688, 0.00000715, -0.00001030, 0.00001116, -0.00001081, 0.00001125, -0.00001457, 0.00001455, -0.00001851, 0.00001407, -0.00002347, 0.00002427, -0.00002078, 0.00001567, -0.00001701, 0.00001100, 0.00001440, 0.00000353, -0.00002198, 0.00002585, -0.00000258, 0.00001031, -0.00000531, -0.00000862, 0.00000344, -0.00000763, 0.00000660, -0.00000285, -0.00000644, 0.00000318, -0.00000209, 0.00001217, -0.00001509, 0.00000818, -0.00001412, 0.00002368, -0.00001240, -0.00000359, -0.00000806, 0.00001672, -0.00001168, 0.00001107, -0.00000989, 0.00000342, -0.00001364, 0.00001517, -0.00002207, 0.00002330, -0.00002119, 0.00002549, -0.00003391, 0.00002798, -0.00003251, 0.00003354, -0.00003747, 0.00004017, -0.00006072, 0.00005482, -0.00009687, 0.00009005, -0.00013456, 0.00013220, -0.00050730, 0.00056795, 0.00045357, -0.00047833, 0.00015755, -0.00015682, 0.00007575, -0.00008023, 0.00006199, -0.00006761, 0.00005640, -0.00005669, 0.00004403, -0.00004797, 0.00003200, -0.00003328, 0.00003304, -0.00003214, 0.00002760, -0.00002705, 0.00002738, -0.00002733, 0.00002767, -0.00003349, 0.00002436, -0.00002078, 0.00001791, -0.00000675, 0.00001964, -0.00002919, 0.00001232, 0.00000333, 0.00000292, -0.00001777, 0.00001166, -0.00000779, 0.00000787, -0.00001631, 0.00002487, -0.00002263, 0.00002428, -0.00002687, 0.00001896, -0.00001525, 0.00001291, 0.00000573, -0.00002489, -0.00000169, 0.00001869, -0.00002052, 0.00000516, -0.00000249, -0.00000004, 0.00001378, -0.00002411, 0.00002133, -0.00001321, 0.00000839, -0.00000734, },
		{ 0.00000797, -0.00000329, 0.00000191, -0.00000368, 0.00000433, 0.00000090, -0.00000002, -0.00000361, 0.00000408, -0.00000653, 0.00000554, -0.00000451, 0.00000496, -0.00000693, 0.00000506, -0.00000702, 0.00000190, -0.00000893, 0.00000306, 0.00000444, -0.00000974, 0.00000720, -0.00001392, 0.00002725, 0.00000293, -0.00001292, 0.00000453, 0.00002001, -0.00000844, 0.00001559, -0.00002448, 0.00001322, -0.00001447, 0.00000883, -0.00000281, -0.00000454, -0.00000082, -0.00000010, 0.00000822, -0.00000533, -0.00000158, -0.00000479, 0.00000636, 0.00000910, -0.00001482, -0.00000380, 0.00000667, 0.00000157, -0.00000087, 0.00000092, -0.00000225, -0.00001030, 0.00000954, -0.00001376, 0.00001111, -0.00000854, 0.00001297, -0.00001674, 0.00000845, -0.00001334, 0.00001359, -0.00001708, 0.00002070, -0.00003556, 0.00002632, -0.00005650, 0.00003486, -0.00006306, 0.00005467, -0.00025933, -0.00003757, 0.00089397, -0.00043988, 0.00003625, -0.00008279, 0.00001623, -0.00004400, 0.00003214, -0.00004029, 0.00002841, -0.00002893, 0.00001769, -0.00002266, 0.00000891, -0.00001583, 0.00001648, -0.00001532, 0.00001163, -0.00001301, 0.00001379, -0.00001322, 0.00001285, -0.00001398, 0.00000233, -0.00000140, 0.00000035, 0.00000393, 0.00001192, -0.00001108, -0.00000876, 0.00001188, 0.00000427, -0.00001493, 0.00000412, -0.00000362, 0.00000725, -0.00001498, 0.00001739, -0.00001006, 0.00000899, -0.00000599, -0.00000500, 0.00000794, -0.00001173, 0.00002596, -0.00002798, -0.00000632, 0.00001179, -0.00000021, -0.00001878, 0.00001819, -0.00001994, 0.00002823, -0.00002698, 0.00001449, -0.00000419, 0.00000233, -0.00000389, },
		{ -0.00000308, 0.00000844, -0.00000752, 0.00000507, -0.00000560, 0.00000953, -0.00000569, 0.00000240, -0.00000338, 0.00000274, -0.00000562, 0.00000683, -0.00000656, 0.00000618, -0.00000929, 0.00000830, -0.00001352, 0.00000846, -0.00001780, 0.00002316, -0.00002276, 0.00001572, -0.00002024, 0.00002212, 0.00001025, -0.00000392, -0.00001370, 0.00003341, -0.00001524, 0.00002241, -0.00002404, 0.00000803, -0.00000905, 0.00000223, 0.00000155, -0.00000385, -0.00000464, 0.00000337, 0.00000078, 0.00000658, -0.00001168, 0.00000557, -0.00000934, 0.00002163, -0.00001565, -0.00000318, -0.00000318, 0.00001237, -0.00000887, 0.00000793, -0.00000679, -0.00000295, -0.00000367, 0.00000341, -0.00000953, 0.00001204, -0.00000880, 0.00001005, -0.00001941, 0.00001422, -0.00001621, 0.00001525, -0.00001502, 0.00001036, -0.00002747, 0.00001319, -0.00005064, 0.00003956, -0.00007080, 0.00000346, -0.00041206, 0.00085240, -0.00001215, -0.00025194, 0.00006724, -0.00009354, 0.00003190, -0.00003595, 0.00002216, -0.00003014, 0.00002422, -0.00002960, 0.00001991, -0.00002836, 0.00001523, -0.00001450, 0.00001529, -0.00001689, 0.00001323, -0.00001229, 0.00001317, -0.00001366, 0.00001442, -0.00002327, 0.00001791, -0.00001552, 0.00001432, -0.00000012, 0.00000963, -0.00002406, 0.00001338, 0.00000452, -0.00000547, -0.00000851, 0.00000486, -0.00000091, -0.00000263, -0.00000136, 0.00001102, -0.00001205, 0.00001673, -0.00002504, 0.00002275, -0.00002375, 0.00002943, -0.00001800, -0.00001322, 0.00000250, 0.00001609, -0.00002852, 0.00001803, -0.00001483, 0.00001505, -0.00000491, -0.00000952, 0.00001413, -0.00000950, 0.00000565, },
		{ -0.00000732, 0.00001198, -0.00000836, 0.00000600, -0.00000742, 0.00000943, -0.00000404, 0.00000303, -0.00000598, 0.00000656, -0.00000981, 0.00000980, -0.00000908, 0.00000929, -0.00001194, 0.00001023, -0.00001430, 0.00000937, -0.00001807, 0.00001690, -0.00000996, 0.00000119, -0.00000640, 0.00000272, 0.00002158, 0.00000237, -0.00002037, 0.00002851, -0.00000300, 0.00001047, -0.00000587, -0.00001046, 0.00000547, -0.00000996, 0.00000861, -0.00000556, -0.00000449, 0.00000212, -0.00000081, 0.00000903, -0.00000996, 0.00000375, -0.00000929, 0.00001449, -0.00000113, -0.00001035, -0.00000502, 0.00001113, -0.00000439, 0.00000411, -0.00000346, -0.00000077, -0.00001078, 0.00001144, -0.00001759, 0.00001763, -0.00001472, 0.00001854, -0.00002569, 0.00001886, -0.00002239, 0.00002284, -0.00002521, 0.00002751, -0.00004669, 0.00003906, -0.00007758, 0.00006565, -0.00010009, 0.00008776, -0.00041496, 0.00040992, 0.00053299, -0.00043687, 0.00010746, -0.00012444, 0.00004898, -0.00006026, 0.00004626, -0.00005236, 0.00004222, -0.00004349, 0.00003124, -0.00003681, 0.00002142, -0.00002393, 0.00002511, -0.00002503, 0.00002034, -0.00002036, 0.00002142, -0.00002117, 0.00002139, -0.00002496, 0.00001442, -0.00001164, 0.00000939, -0.00000079, 0.00001558, -0.00002027, 0.00000179, 0.00000809, 0.00000222, -0.00001537, 0.00000728, -0.00000564, 0.00000660, -0.00001370, 0.00002115, -0.00001886, 0.00002160, -0.00002365, 0.00001539, -0.00001366, 0.00001231, 0.00000451, -0.00002399, -0.00000077, 0.00001620, -0.00001523, -0.00000427, 0.00000836, -0.00001069, 0.00002069, -0.00002762, 0.00002166, -0.00001142, 0.00000789, },
		{ -0.00000059, 0.00000405, 0.00000113, -0.00000182, -0.00000055, 0.00000099, 0.00000404, -0.00000182, -0.00000200, 0.00000209, -0.00000396, 0.00000187, -0.00000061, 0.00000075, -0.00000198, -0.00000088, -0.00000191, -0.00000359, -0.00000288, -0.00000415, 0.00001263, -0.00001677, 0.00000713, -0.00001175, 0.00002385, 0.00001048, -0.00001723, 0.00001061, 0.00001864, -0.00000757, 0.00001498, -0.00002638, 0.00001465, -0.00001598, 0.00001006, -0.00000471, -0.00000330, -0.00000215, 0.00000265, 0.00000337, 0.00000037, -0.00000597, 0.00000043, -0.00000154, 0.00001508, -0.00001501, -0.00000456, 0.00000328, 0.00000549, -0.00000337, 0.00000251, -0.00000241, -0.00000945, 0.00000640, -0.00000936, 0.00000549, -0.00000238, 0.00000635, -0.00000901, 0.00000027, -0.00000459, 0.00000444, -0.00000668, 0.00001002, -0.00002316, 0.00001176, -0.00003822, 0.00001221, -0.00003347, 0.00001940, -0.00019891, -0.00009405, 0.00082433, -0.00029622, -0.00005103, -0.00003605, -0.00001904, -0.00001824, 0.00001049, -0.00001950, 0.00000999, -0.00001250, 0.00000277, -0.00001009, -0.00000270, -0.00000576, 0.00000755, -0.00000664, 0.00000272, -0.00000475, 0.00000588, -0.00000476, 0.00000411, -0.00000387, -0.00000698, 0.00000543, -0.00000501, 0.00000702, 0.00000873, -0.00000362, -0.00001388, 0.00001045, 0.00000627, -0.00001293, 0.00000091, -0.00000287, 0.00000602, -0.00001189, 0.00001389, -0.00000744, 0.00000753, -0.00000393, -0.00000673, 0.00000765, -0.00001138, 0.00002539, -0.00002955, -0.00000222, 0.00000643, 0.00000676, -0.00002656, 0.00002319, -0.00002127, 0.00002449, -0.00002088, 0.00000734, 0.00000227, -0.00000132, },
		{ 0.00000731, -0.00000527, 0.00001036, -0.00000827, 0.00000532, -0.00000602, 0.00000944, -0.00000465, 0.00000202, -0.00000378, 0.00000373, -0.00000715, 0.00000791, -0.00000753, 0.00000727, -0.00001020, 0.00000760, -0.00001293, 0.00000762, -0.00001659, 0.00002049, -0.00001647, 0.00000346, -0.00000966, 0.00001250, 0.00002050, -0.00000822, -0.00000889, 0.00003372, -0.00001450, 0.00002155, -0.00002493, 0.00000739, -0.00000826, 0.00000115, 0.00000207, -0.00000498, -0.00000413, 0.00000424, -0.00000091, 0.00000724, -0.00001032, 0.00000479, -0.00000902, 0.00001715, -0.00000722, -0.00000887, -0.00000188, 0.00000973, -0.00000440, 0.00000325, -0.00000212, -0.00000553, -0.00000321, 0.00000314, -0.00000931, 0.00001125, -0.00000767, 0.00000905, -0.00001770, 0.00001211, -0.00001366, 0.00001283, -0.00001177, 0.00000761, -0.00002489, 0.00001110, -0.00004781, 0.00003574, -0.00006197, -0.00000539, -0.00036011, 0.00069458, 0.00012914, -0.00028464, 0.00006628, -0.00009353, 0.00002981, -0.00003561, 0.00002361, -0.00003038, 0.00002383, -0.00002879, 0.00001808, -0.00002691, 0.00001343, -0.00001297, 0.00001459, -0.00001680, 0.00001277, -0.00001207, 0.00001364, -0.00001406, 0.00001503, -0.00002168, 0.00001400, -0.00001122, 0.00000947, 0.00000276, 0.00000965, -0.00001956, 0.00000500, 0.00000901, -0.00000520, -0.00000831, 0.00000229, 0.00000002, -0.00000207, -0.00000119, 0.00000949, -0.00000984, 0.00001544, -0.00002323, 0.00001967, -0.00002193, 0.00002851, -0.00001959, -0.00001036, 0.00000038, 0.00001742, -0.00002760, 0.00001256, -0.00000620, 0.00000417, 0.00000442, -0.00001666, 0.00001825, -0.00001065, },
		{ 0.00000531, -0.00000524, 0.00000933, -0.00000484, 0.00000267, -0.00000453, 0.00000619, -0.00000061, 0.00000087, -0.00000438, 0.00000495, -0.00000791, 0.00000688, -0.00000582, 0.00000572, -0.00000755, 0.00000432, -0.00000881, 0.00000329, -0.00001107, 0.00000817, 0.00000122, -0.00001140, 0.00000009, -0.00000155, 0.00002454, 0.00000531, -0.00002123, 0.00003268, -0.00000353, 0.00001044, -0.00000673, -0.00001190, 0.00000693, -0.00001167, 0.00000991, -0.00000757, -0.00000309, 0.00000118, 0.00000094, 0.00000506, -0.00000430, -0.00000109, -0.00000379, 0.00000535, 0.00000816, -0.00001430, -0.00000373, 0.00000657, 0.00000141, -0.00000076, 0.00000050, -0.00000278, -0.00000918, 0.00000828, -0.00001331, 0.00001203, -0.00000841, 0.00001176, -0.00001763, 0.00001011, -0.00001290, 0.00001285, -0.00001378, 0.00001570, -0.00003335, 0.00002371, -0.00005823, 0.00004140, -0.00006700, 0.00004616, -0.00033231, 0.00028455, 0.00056201, -0.00035929, 0.00004243, -0.00008522, 0.00001826, -0.00003756, 0.00002767, -0.00003423, 0.00002566, -0.00002833, 0.00001695, -0.00002449, 0.00000996, -0.00001388, 0.00001638, -0.00001687, 0.00001199, -0.00001262, 0.00001421, -0.00001351, 0.00001341, -0.00001500, 0.00000403, -0.00000287, 0.00000174, 0.00000418, 0.00001169, -0.00001165, -0.00000657, 0.00000997, 0.00000301, -0.00001325, 0.00000325, -0.00000389, 0.00000523, -0.00001068, 0.00001706, -0.00001506, 0.00001905, -0.00002056, 0.00001195, -0.00001226, 0.00001176, 0.00000364, -0.00002399, 0.00000172, 0.00001193, -0.00000841, -0.00001366, 0.00001715, -0.00001743, 0.00002276, -0.00002647, 0.00001793, -0.00000675, },
		{ -0.00000450, 0.00000252, 0.00000039, 0.00000491, -0.00000468, 0.00000172, -0.00000143, 0.00000608, -0.00000262, -0.00000100, 0.00000045, -0.00000165, -0.00000132, 0.00000253, -0.00000254, 0.00000185, -0.00000521, 0.00000118, -0.00000702, 0.00000079, -0.00000810, 0.00001626, -0.00001770, 0.00000079, -0.00000531, 0.00001701, 0.00002027, -0.00002187, 0.00001652, 0.00001795, -0.00000661, 0.00001331, -0.00002707, 0.00001471, -0.00001612, 0.00000990, -0.00000533, -0.00000313, -0.00000268, 0.00000473, -0.00000048, 0.00000431, -0.00000825, 0.00000332, -0.00000598, 0.00001635, -0.00001113, -0.00000773, 0.00000207, 0.00000649, -0.00000266, 0.00000115, -0.00000042, -0.00000995, 0.00000453, -0.00000651, 0.00000185, 0.00000131, 0.00000254, -0.00000442, -0.00000425, 0.00000008, -0.00000038, -0.00000094, 0.00000431, -0.00001607, 0.00000307, -0.00002672, -0.00000167, -0.00001562, -0.00000051, -0.00016183, -0.00011376, 0.00072319, -0.00015120, -0.00012166, -0.00000470, -0.00004227, -0.00000194, -0.00000395, -0.00000492, -0.00000244, -0.00000169, -0.00000682, -0.00000248, -0.00000971, 0.00000016, 0.00000221, -0.00000112, -0.00000316, 0.00000056, 0.00000066, 0.00000109, -0.00000186, 0.00000290, -0.00001213, 0.00000798, -0.00000636, 0.00000668, 0.00000810, 0.00000077, -0.00001477, 0.00000577, 0.00000941, -0.00001137, -0.00000131, -0.00000317, 0.00000542, -0.00000948, 0.00001142, -0.00000616, 0.00000743, -0.00000310, -0.00000733, 0.00000631, -0.00001024, 0.00002432, -0.00003077, 0.00000144, 0.00000224, 0.00001141, -0.00003025, 0.00002295, -0.00001727, 0.00001649, -0.00001218, -0.00000101, 0.00000864, },
		{ -0.00000895, 0.00000683, -0.00000534, 0.00001002, -0.00000690, 0.00000373, -0.00000471, 0.00000766, -0.00000226, 0.00000064, -0.00000317, 0.00000346, -0.00000702, 0.00000705, -0.00000643, 0.00000609, -0.00000848, 0.00000420, -0.00000976, 0.00000391, -0.00001193, 0.00001403, -0.00000660, -0.00001048, -0.00000007, 0.00000434, 0.00002847, -0.00000939, -0.00000629, 0.00003563, -0.00001386, 0.00001998, -0.00002529, 0.00000627, -0.00000710, -0.00000027, 0.00000282, -0.00000625, -0.00000345, 0.00000463, -0.00000145, 0.00000609, -0.00000711, 0.00000208, -0.00000613, 0.00001021, 0.00000194, -0.00001393, -0.00000062, 0.00000635, 0.00000062, -0.00000137, 0.00000184, -0.00000730, -0.00000289, 0.00000230, -0.00000793, 0.00000888, -0.00000480, 0.00000610, -0.00001377, 0.00000768, -0.00000881, 0.00000797, -0.00000593, 0.00000203, -0.00001892, 0.00000477, -0.00003927, 0.00002477, -0.00004494, -0.00002481, -0.00029797, 0.00054701, 0.00023275, -0.00027962, 0.00004238, -0.00007882, 0.00001721, -0.00002727, 0.00001784, -0.00002377, 0.00001733, -0.00002248, 0.00001140, -0.00002117, 0.00000789, -0.00000818, 0.00001082, -0.00001357, 0.00000921, -0.00000889, 0.00001104, -0.00001115, 0.00001194, -0.00001629, 0.00000700, -0.00000465, 0.00000312, 0.00000665, 0.00000795, -0.00001310, -0.00000348, 0.00001211, -0.00000419, -0.00000773, -0.00000068, 0.00000116, -0.00000212, -0.00000003, 0.00000685, -0.00000675, 0.00001346, -0.00002081, 0.00001605, -0.00001968, 0.00002714, -0.00002040, -0.00000880, 0.00000046, 0.00001579, -0.00002314, 0.00000428, 0.00000352, -0.00000572, 0.00001116, -0.00002051, 0.00001886, },
		{ -0.00000029, 0.00000166, -0.00000203, 0.00000559, -0.00000062, -0.00000104, -0.00000137, 0.00000276, 0.00000270, -0.00000109, -0.00000266, 0.00000289, -0.00000532, 0.00000329, -0.00000203, 0.00000165, -0.00000270, -0.00000166, -0.00000360, -0.00000253, -0.00000444, 0.00000049, 0.00001003, -0.00001944, 0.00000075, -0.00000120, 0.00002317, 0.00001196, -0.00002387, 0.00003771, -0.00000376, 0.00001021, -0.00000820, -0.00001253, 0.00000739, -0.00001235, 0.00001013, -0.00000854, -0.00000250, 0.00000068, 0.00000269, 0.00000106, 0.00000084, -0.00000512, 0.00000086, -0.00000192, 0.00001388, -0.00001454, -0.00000474, 0.00000377, 0.00000485, -0.00000279, 0.00000160, -0.00000254, -0.00000889, 0.00000597, -0.00000975, 0.00000729, -0.00000325, 0.00000630, -0.00001106, 0.00000321, -0.00000561, 0.00000523, -0.00000491, 0.00000659, -0.00002272, 0.00001115, -0.00004184, 0.00002115, -0.00003999, 0.00001321, -0.00026597, 0.00019523, 0.00054616, -0.00025781, -0.00002578, -0.00004786, -0.00000996, -0.00001711, 0.00001034, -0.00001696, 0.00001022, -0.00001441, 0.00000407, -0.00001364, 0.00000000, -0.00000522, 0.00000874, -0.00000945, 0.00000432, -0.00000554, 0.00000744, -0.00000614, 0.00000573, -0.00000583, -0.00000459, 0.00000358, -0.00000337, 0.00000683, 0.00000924, -0.00000480, -0.00001156, 0.00000860, 0.00000519, -0.00001147, -0.00000016, -0.00000284, 0.00000410, -0.00000776, 0.00001325, -0.00001188, 0.00001722, -0.00001816, 0.00000914, -0.00001143, 0.00001150, 0.00000288, -0.00002444, 0.00000495, 0.00000718, -0.00000188, -0.00002090, 0.00002193, -0.00001905, 0.00001975, -0.00002114, 0.00001107, },
		{ 0.00001311, -0.00000586, 0.00000429, -0.00000191, 0.00000704, -0.00000586, 0.00000242, -0.00000235, 0.00000651, -0.00000206, -0.00000095, -0.00000039, -0.00000023, -0.00000328, 0.00000412, -0.00000410, 0.00000371, -0.00000708, 0.00000155, -0.00000771, 0.00000136, -0.00000811, 0.00001499, -0.00001302, -0.00001010, 0.00000307, 0.00000891, 0.00003028, -0.00002511, 0.00002109, 0.00001862, -0.00000584, 0.00001073, -0.00002677, 0.00001372, -0.00001526, 0.00000874, -0.00000506, -0.00000368, -0.00000264, 0.00000614, -0.00000311, 0.00000623, -0.00000826, 0.00000372, -0.00000696, 0.00001360, -0.00000448, -0.00001207, 0.00000213, 0.00000542, 0.00000016, -0.00000182, 0.00000240, -0.00001077, 0.00000325, -0.00000478, -0.00000009, 0.00000278, 0.00000123, -0.00000267, -0.00000553, 0.00000121, -0.00000151, 0.00000084, 0.00000282, -0.00001351, -0.00000046, -0.00002136, -0.00000752, -0.00000828, -0.00000722, -0.00014382, -0.00010646, 0.00060677, -0.00002077, -0.00016613, 0.00000692, -0.00005126, 0.00000361, -0.00001014, 0.00000237, -0.00000791, 0.00000271, -0.00001047, -0.00000026, -0.00001183, 0.00000172, 0.00000066, 0.00000095, -0.00000562, 0.00000251, -0.00000140, 0.00000377, -0.00000443, 0.00000571, -0.00001282, 0.00000634, -0.00000411, 0.00000349, 0.00000954, 0.00000249, -0.00001227, -0.00000079, 0.00001248, -0.00000969, -0.00000280, -0.00000425, 0.00000533, -0.00000780, 0.00001004, -0.00000621, 0.00000853, -0.00000331, -0.00000702, 0.00000420, -0.00000854, 0.00002280, -0.00003138, 0.00000411, -0.00000011, 0.00001313, -0.00002962, 0.00001810, -0.00000956, 0.00000650, -0.00000325, -0.00000839, },
		{ 0.00001566, -0.00000450, 0.00000440, -0.00000350, 0.00000772, -0.00000386, 0.00000078, -0.00000219, 0.00000475, 0.00000091, -0.00000130, -0.00000183, 0.00000215, -0.00000549, 0.00000462, -0.00000375, 0.00000324, -0.00000490, -0.00000091, -0.00000511, -0.00000143, -0.00000542, 0.00000589, 0.00000401, -0.00002281, 0.00000573, -0.00000051, 0.00003282, -0.00000670, -0.00000613, 0.00003909, -0.00001326, 0.00001785, -0.00002557, 0.00000523, -0.00000619, -0.00000142, 0.00000320, -0.00000713, -0.00000302, 0.00000485, -0.00000123, 0.00000378, -0.00000297, -0.00000145, -0.00000213, 0.00000282, 0.00000971, -0.00001677, -0.00000061, 0.00000348, 0.00000472, -0.00000447, 0.00000394, -0.00000750, -0.00000324, 0.00000146, -0.00000612, 0.00000588, -0.00000134, 0.00000252, -0.00000914, 0.00000266, -0.00000352, 0.00000265, 0.00000037, -0.00000414, -0.00001201, -0.00000298, -0.00002850, 0.00001115, -0.00002555, -0.00004687, -0.00023680, 0.00042114, 0.00029412, -0.00024061, 0.00000207, -0.00005522, -0.00000117, -0.00001465, 0.00000790, -0.00001310, 0.00000736, -0.00001318, 0.00000224, -0.00001336, 0.00000064, -0.00000193, 0.00000554, -0.00000864, 0.00000394, -0.00000410, 0.00000667, -0.00000624, 0.00000666, -0.00000894, -0.00000103, 0.00000215, -0.00000287, 0.00000995, 0.00000602, -0.00000648, -0.00001029, 0.00001281, -0.00000208, -0.00000710, -0.00000363, 0.00000215, -0.00000244, 0.00000167, 0.00000371, -0.00000346, 0.00001145, -0.00001842, 0.00001249, -0.00001749, 0.00002560, -0.00002062, -0.00000834, 0.00000224, 0.00001213, -0.00001671, -0.00000454, 0.00001177, -0.00001232, 0.00001372, -0.00002019, },
		{ 0.00000279, 0.00000627, -0.00000166, 0.00000108, 0.00000198, 0.00000311, -0.00000406, 0.00000109, 0.00000007, 0.00000505, -0.00000220, -0.00000142, 0.00000103, -0.00000284, -0.00000001, 0.00000122, -0.00000174, 0.00000132, -0.00000635, -0.00000005, -0.00000663, 0.00000016, -0.00000429, 0.00001463, -0.00002162, -0.00000451, 0.00000287, 0.00001857, 0.00002103, -0.00002698, 0.00004261, -0.00000307, 0.00000957, -0.00001031, -0.00001223, 0.00000680, -0.00001198, 0.00000925, -0.00000850, -0.00000271, 0.00000066, 0.00000412, -0.00000234, 0.00000467, -0.00000751, 0.00000367, -0.00000626, 0.00001548, -0.00001121, -0.00000774, 0.00000269, 0.00000590, -0.00000214, 0.00000029, -0.00000070, -0.00000939, 0.00000431, -0.00000704, 0.00000374, 0.00000025, 0.00000277, -0.00000669, -0.00000111, -0.00000124, 0.00000069, 0.00000064, 0.00000096, -0.00001573, 0.00000258, -0.00003000, 0.00000694, -0.00002140, -0.00000857, -0.00021800, 0.00013954, 0.00049614, -0.00014719, -0.00008538, -0.00002000, -0.00003041, -0.00000279, -0.00000253, -0.00000350, -0.00000136, -0.00000420, -0.00000520, -0.00000619, -0.00000677, 0.00000056, 0.00000350, -0.00000408, -0.00000135, -0.00000047, 0.00000246, -0.00000051, -0.00000007, 0.00000088, -0.00000993, 0.00000652, -0.00000512, 0.00000661, 0.00000876, -0.00000038, -0.00001293, 0.00000447, 0.00000811, -0.00000978, -0.00000290, -0.00000254, 0.00000335, -0.00000528, 0.00001017, -0.00000971, 0.00001642, -0.00001668, 0.00000716, -0.00001124, 0.00001158, 0.00000207, -0.00002493, 0.00000810, 0.00000307, 0.00000295, -0.00002462, 0.00002191, -0.00001570, 0.00001272, -0.00001317, },
		{ -0.00001302, 0.00001442, -0.00000458, 0.00000410, -0.00000230, 0.00000712, -0.00000510, 0.00000141, -0.00000165, 0.00000534, -0.00000021, -0.00000185, -0.00000027, 0.00000001, -0.00000369, 0.00000389, -0.00000372, 0.00000343, -0.00000645, -0.00000068, -0.00000586, -0.00000094, -0.00000468, 0.00000987, -0.00000461, -0.00002271, 0.00001048, 0.00000182, 0.00003857, -0.00002550, 0.00002348, 0.00002109, -0.00000546, 0.00000748, -0.00002586, 0.00001219, -0.00001398, 0.00000715, -0.00000446, -0.00000445, -0.00000238, 0.00000703, -0.00000458, 0.00000634, -0.00000637, 0.00000213, -0.00000533, 0.00000830, 0.00000308, -0.00001599, 0.00000228, 0.00000349, 0.00000367, -0.00000486, 0.00000465, -0.00001093, 0.00000194, -0.00000366, -0.00000085, 0.00000261, 0.00000175, -0.00000292, -0.00000454, -0.00000006, -0.00000022, 0.00000005, 0.00000405, -0.00001396, -0.00000041, -0.00002042, -0.00000745, -0.00000852, -0.00000484, -0.00013852, -0.00008326, 0.00049022, 0.00008306, -0.00017997, -0.00000117, -0.00004720, -0.00000032, -0.00000910, 0.00000310, -0.00000709, 0.00000146, -0.00000899, -0.00000259, -0.00000991, -0.00000024, 0.00000219, 0.00000013, -0.00000511, 0.00000152, -0.00000069, 0.00000356, -0.00000386, 0.00000499, -0.00000984, 0.00000165, 0.00000046, -0.00000124, 0.00001194, 0.00000257, -0.00000784, -0.00000753, 0.00001430, -0.00000734, -0.00000386, -0.00000580, 0.00000549, -0.00000671, 0.00000955, -0.00000725, 0.00001040, -0.00000413, -0.00000622, 0.00000176, -0.00000662, 0.00002099, -0.00003134, 0.00000560, -0.00000047, 0.00001203, -0.00002544, 0.00001025, -0.00000044, -0.00000291, 0.00000366, },
		{ -0.00001600, 0.00000938, 0.00000173, 0.00000092, -0.00000056, 0.00000436, -0.00000010, -0.00000261, 0.00000068, 0.00000155, 0.00000406, -0.00000312, -0.00000031, 0.00000033, -0.00000321, 0.00000143, -0.00000043, -0.00000024, -0.00000068, -0.00000635, -0.00000045, -0.00000681, 0.00000104, -0.00000155, 0.00001264, -0.00003085, 0.00000605, -0.00000143, 0.00003334, -0.00000033, -0.00000785, 0.00004359, -0.00001237, 0.00001516, -0.00002601, 0.00000468, -0.00000593, -0.00000193, 0.00000287, -0.00000731, -0.00000308, 0.00000517, -0.00000067, 0.00000103, 0.00000110, -0.00000466, 0.00000152, -0.00000326, 0.00001455, -0.00001650, -0.00000238, 0.00000184, 0.00000705, -0.00000533, 0.00000377, -0.00000608, -0.00000429, 0.00000084, -0.00000435, 0.00000296, 0.00000180, -0.00000060, -0.00000507, -0.00000157, 0.00000074, -0.00000156, 0.00000547, -0.00000912, -0.00000612, -0.00000989, -0.00001840, -0.00000135, -0.00000849, -0.00006554, -0.00018429, 0.00032281, 0.00031521, -0.00017645, -0.00004526, -0.00002993, -0.00001988, -0.00000198, -0.00000271, -0.00000158, -0.00000305, -0.00000371, -0.00000679, -0.00000583, -0.00000622, 0.00000394, 0.00000042, -0.00000358, -0.00000146, 0.00000077, 0.00000205, -0.00000095, 0.00000100, -0.00000166, -0.00000804, 0.00000734, -0.00000693, 0.00001142, 0.00000506, -0.00000109, -0.00001431, 0.00001078, 0.00000102, -0.00000643, -0.00000628, 0.00000272, -0.00000274, 0.00000342, 0.00000069, -0.00000060, 0.00000995, -0.00001655, 0.00000946, -0.00001567, 0.00002410, -0.00002043, -0.00000863, 0.00000502, 0.00000757, -0.00001000, -0.00001184, 0.00001662, -0.00001437, 0.00001175, },
		{ -0.00000469, -0.00000490, 0.00001122, -0.00000334, 0.00000300, -0.00000043, 0.00000540, -0.00000560, 0.00000217, -0.00000127, 0.00000594, -0.00000205, -0.00000103, -0.00000011, -0.00000114, -0.00000225, 0.00000311, -0.00000361, 0.00000360, -0.00000886, 0.00000104, -0.00000824, 0.00000184, -0.00000534, 0.00001452, -0.00001825, -0.00001412, 0.00000854, 0.00001261, 0.00003073, -0.00002899, 0.00004631, -0.00000085, 0.00000826, -0.00001294, -0.00001117, 0.00000538, -0.00001087, 0.00000759, -0.00000776, -0.00000343, 0.00000099, 0.00000512, -0.00000482, 0.00000682, -0.00000793, 0.00000430, -0.00000750, 0.00001339, -0.00000536, -0.00001173, 0.00000262, 0.00000519, 0.00000032, -0.00000230, 0.00000164, -0.00000989, 0.00000288, -0.00000499, 0.00000137, 0.00000207, 0.00000119, -0.00000451, -0.00000290, 0.00000035, -0.00000096, 0.00000311, -0.00000142, -0.00001221, -0.00000204, -0.00002290, -0.00000097, -0.00001123, -0.00001961, -0.00018658, 0.00011085, 0.00042534, -0.00004195, -0.00012685, -0.00000664, -0.00004020, 0.00000345, -0.00000933, 0.00000456, -0.00000765, 0.00000109, -0.00000984, -0.00000297, -0.00000970, 0.00000294, 0.00000118, -0.00000135, -0.00000435, 0.00000195, -0.00000006, 0.00000263, -0.00000313, 0.00000430, -0.00001144, 0.00000581, -0.00000366, 0.00000389, 0.00000996, 0.00000181, -0.00001133, -0.00000123, 0.00001077, -0.00000775, -0.00000510, -0.00000291, 0.00000298, -0.00000338, 0.00000799, -0.00000864, 0.00001658, -0.00001605, 0.00000591, -0.00001152, 0.00001187, 0.00000118, -0.00002517, 0.00001062, 0.00000036, 0.00000534, -0.00002443, 0.00001751, -0.00000873, 0.00000368, },
		{ 0.00000701, -0.00001395, 0.00001215, -0.00000069, 0.00000208, -0.00000091, 0.00000538, -0.00000278, -0.00000090, 0.00000023, 0.00000301, 0.00000245, -0.00000330, 0.00000057, -0.00000077, -0.00000275, 0.00000214, -0.00000179, 0.00000148, -0.00000395, -0.00000469, -0.00000240, -0.00000505, 0.00000077, 0.00000288, 0.00000481, -0.00003383, 0.00001426, -0.00000257, 0.00004388, -0.00002225, 0.00002343, 0.00002539, -0.00000545, 0.00000371, -0.00002471, 0.00001062, -0.00001281, 0.00000566, -0.00000403, -0.00000499, -0.00000222, 0.00000764, -0.00000515, 0.00000513, -0.00000335, -0.00000051, -0.00000237, 0.00000226, 0.00000965, -0.00001809, 0.00000149, 0.00000181, 0.00000659, -0.00000670, 0.00000536, -0.00000985, 0.00000025, -0.00000281, -0.00000089, 0.00000148, 0.00000321, -0.00000415, -0.00000248, -0.00000237, 0.00000200, -0.00000175, 0.00000634, -0.00001566, 0.00000133, -0.00002179, -0.00000413, -0.00001275, 0.00000188, -0.00013922, -0.00005443, 0.00038538, 0.00015378, -0.00016423, -0.00002479, -0.00003409, -0.00001035, -0.00000348, -0.00000060, -0.00000195, -0.00000353, -0.00000424, -0.00000767, -0.00000564, -0.00000420, 0.00000543, -0.00000242, -0.00000268, -0.00000138, 0.00000180, 0.00000145, -0.00000118, 0.00000193, -0.00000471, -0.00000435, 0.00000559, -0.00000587, 0.00001391, 0.00000229, -0.00000304, -0.00001290, 0.00001400, -0.00000413, -0.00000463, -0.00000752, 0.00000564, -0.00000598, 0.00000962, -0.00000878, 0.00001249, -0.00000501, -0.00000544, -0.00000049, -0.00000488, 0.00001910, -0.00003072, 0.00000599, 0.00000083, 0.00000885, -0.00001905, 0.00000144, 0.00000765, -0.00000954, },
		{ 0.00000610, -0.00000924, 0.00000156, 0.00000821, -0.00000231, 0.00000233, 0.00000108, 0.00000328, -0.00000546, 0.00000300, -0.00000107, 0.00000639, -0.00000419, 0.00000083, -0.00000137, -0.00000094, -0.00000161, 0.00000253, -0.00000325, 0.00000292, -0.00001076, 0.00000288, -0.00001081, 0.00000575, -0.00000640, 0.00001736, -0.00003320, 0.00000078, 0.00000076, 0.00003090, 0.00000864, -0.00001030, 0.00004825, -0.00001065, 0.00001175, -0.00002670, 0.00000471, -0.00000640, -0.00000174, 0.00000179, -0.00000679, -0.00000362, 0.00000565, -0.00000006, -0.00000153, 0.00000432, -0.00000667, 0.00000378, -0.00000694, 0.00001583, -0.00001316, -0.00000574, 0.00000146, 0.00000749, -0.00000401, 0.00000172, -0.00000359, -0.00000562, 0.00000033, -0.00000279, 0.00000053, 0.00000403, -0.00000257, -0.00000233, -0.00000420, 0.00000315, -0.00000384, 0.00000850, -0.00001201, -0.00000230, -0.00001458, -0.00001080, -0.00001036, 0.00000344, -0.00007743, -0.00014415, 0.00025245, 0.00030326, -0.00009906, -0.00008943, -0.00000985, -0.00003407, 0.00000713, -0.00001100, 0.00000799, -0.00001129, 0.00000358, -0.00001360, -0.00000044, -0.00001104, 0.00000797, -0.00000325, 0.00000032, -0.00000567, 0.00000440, -0.00000153, 0.00000333, -0.00000352, 0.00000390, -0.00001251, 0.00000969, -0.00000816, 0.00001049, 0.00000558, 0.00000243, -0.00001527, 0.00000646, 0.00000447, -0.00000549, -0.00000856, 0.00000275, -0.00000282, 0.00000489, -0.00000177, 0.00000145, 0.00000926, -0.00001545, 0.00000717, -0.00001435, 0.00002272, -0.00001999, -0.00000927, 0.00000805, 0.00000322, -0.00000442, -0.00001622, 0.00001716, -0.00001187, },
		{ -0.00000501, 0.00000220, -0.00001028, 0.00001328, -0.00000255, 0.00000311, -0.00000109, 0.00000583, -0.00000537, 0.00000167, -0.00000111, 0.00000529, -0.00000066, -0.00000152, -0.00000037, -0.00000052, -0.00000306, 0.00000333, -0.00000368, 0.00000389, -0.00000900, -0.00000032, -0.00000738, 0.00000068, -0.00000297, 0.00001054, -0.00001107, -0.00002536, 0.00001310, 0.00000730, 0.00003933, -0.00002853, 0.00004801, 0.00000332, 0.00000611, -0.00001589, -0.00000968, 0.00000358, -0.00000947, 0.00000561, -0.00000676, -0.00000428, 0.00000143, 0.00000576, -0.00000630, 0.00000730, -0.00000659, 0.00000307, -0.00000628, 0.00000887, 0.00000139, -0.00001532, 0.00000258, 0.00000373, 0.00000340, -0.00000488, 0.00000331, -0.00000964, 0.00000129, -0.00000335, -0.00000004, 0.00000250, 0.00000118, -0.00000404, -0.00000276, -0.00000010, -0.00000056, 0.00000337, -0.00000148, -0.00001122, -0.00000360, -0.00001962, -0.00000373, -0.00000776, -0.00002248, -0.00016744, 0.00010046, 0.00034718, 0.00004612, -0.00014469, -0.00000907, -0.00003933, 0.00000196, -0.00001029, 0.00000725, -0.00000871, 0.00000161, -0.00001013, -0.00000360, -0.00000920, 0.00000233, 0.00000145, -0.00000105, -0.00000480, 0.00000183, -0.00000020, 0.00000329, -0.00000347, 0.00000459, -0.00000961, 0.00000227, -0.00000002, -0.00000025, 0.00001189, 0.00000261, -0.00000799, -0.00000703, 0.00001214, -0.00000504, -0.00000686, -0.00000376, 0.00000287, -0.00000204, 0.00000662, -0.00000844, 0.00001740, -0.00001596, 0.00000508, -0.00001195, 0.00001212, 0.00000030, -0.00002503, 0.00001221, -0.00000071, 0.00000523, -0.00002086, 0.00001009, -0.00000024, },
		{ -0.00001213, 0.00000633, -0.00001133, 0.00000689, 0.00000497, -0.00000096, 0.00000152, 0.00000264, 0.00000027, -0.00000371, 0.00000254, 0.00000032, 0.00000516, -0.00000472, 0.00000165, -0.00000210, -0.00000103, -0.00000043, 0.00000088, -0.00000119, -0.00000069, -0.00000918, 0.00000130, -0.00000948, 0.00000642, -0.00000370, 0.00001254, -0.00004081, 0.00001283, -0.00000375, 0.00004592, -0.00001545, 0.00002139, 0.00003106, -0.00000553, -0.00000056, -0.00002357, 0.00000936, -0.00001211, 0.00000460, -0.00000407, -0.00000506, -0.00000234, 0.00000813, -0.00000515, 0.00000322, -0.00000008, -0.00000314, 0.00000057, -0.00000286, 0.00001376, -0.00001751, -0.00000075, 0.00000101, 0.00000812, -0.00000667, 0.00000420, -0.00000756, -0.00000176, -0.00000210, -0.00000055, -0.00000004, 0.00000485, -0.00000543, -0.00000043, -0.00000451, 0.00000392, -0.00000325, 0.00000830, -0.00001714, 0.00000317, -0.00002355, -0.00000005, -0.00001772, 0.00000880, -0.00014038, -0.00002771, 0.00029943, 0.00019073, -0.00012487, -0.00005675, -0.00001767, -0.00002212, 0.00000336, -0.00000594, 0.00000489, -0.00000978, 0.00000145, -0.00001336, -0.00000097, -0.00000837, 0.00000884, -0.00000533, 0.00000036, -0.00000485, 0.00000476, -0.00000131, 0.00000227, -0.00000190, 0.00000080, -0.00000982, 0.00000958, -0.00000895, 0.00001434, 0.00000268, 0.00000090, -0.00001590, 0.00001137, -0.00000027, -0.00000506, -0.00000923, 0.00000558, -0.00000536, 0.00000985, -0.00001027, 0.00001425, -0.00000552, -0.00000511, -0.00000219, -0.00000357, 0.00001729, -0.00002966, 0.00000554, 0.00000320, 0.00000458, -0.00001202, -0.00000632, 0.00001284, },
		{ -0.00000597, -0.00000147, -0.00000181, -0.00000597, 0.00001329, -0.00000402, 0.00000412, -0.00000113, 0.00000540, -0.00000700, 0.00000409, -0.00000247, 0.00000738, -0.00000410, 0.00000121, -0.00000246, 0.00000065, -0.00000372, 0.00000431, -0.00000493, 0.00000502, -0.00001323, 0.00000403, -0.00001255, 0.00000773, -0.00000771, 0.00001753, -0.00003007, -0.00000858, 0.00000409, 0.00002710, 0.00001867, -0.00001208, 0.00005212, -0.00000757, 0.00000742, -0.00002752, 0.00000521, -0.00000741, -0.00000109, 0.00000021, -0.00000582, -0.00000443, 0.00000621, 0.00000044, -0.00000350, 0.00000621, -0.00000707, 0.00000423, -0.00000792, 0.00001383, -0.00000761, -0.00000983, 0.00000180, 0.00000653, -0.00000122, -0.00000124, -0.00000100, -0.00000657, -0.00000035, -0.00000141, -0.00000127, 0.00000516, -0.00000320, -0.00000111, -0.00000506, 0.00000362, -0.00000413, 0.00000941, -0.00001272, -0.00000075, -0.00001666, -0.00000637, -0.00001505, 0.00000942, -0.00008184, -0.00011647, 0.00020625, 0.00026871, -0.00002096, -0.00012166, 0.00000006, -0.00004068, 0.00001054, -0.00001518, 0.00001388, -0.00001578, 0.00000732, -0.00001700, 0.00000184, -0.00001301, 0.00000951, -0.00000482, 0.00000235, -0.00000794, 0.00000605, -0.00000330, 0.00000577, -0.00000594, 0.00000684, -0.00001379, 0.00000893, -0.00000661, 0.00000741, 0.00000737, 0.00000418, -0.00001373, 0.00000093, 0.00000739, -0.00000399, -0.00001048, 0.00000226, -0.00000263, 0.00000588, -0.00000346, 0.00000256, 0.00000937, -0.00001509, 0.00000558, -0.00001342, 0.00002138, -0.00001937, -0.00000998, 0.00001075, -0.00000016, -0.00000081, -0.00001713, 0.00001363, },
		{ 0.00000750, -0.00001130, 0.00000603, -0.00001238, 0.00001197, 0.00000076, 0.00000147, -0.00000008, 0.00000458, -0.00000368, -0.00000009, 0.00000022, 0.00000350, 0.00000154, -0.00000259, 0.00000008, -0.00000091, -0.00000256, 0.00000208, -0.00000224, 0.00000256, -0.00000731, -0.00000346, -0.00000485, -0.00000242, 0.00000154, 0.00000454, -0.00000267, -0.00003513, 0.00001409, 0.00000407, 0.00004568, -0.00002484, 0.00004748, 0.00000945, 0.00000314, -0.00001903, -0.00000809, 0.00000183, -0.00000827, 0.00000377, -0.00000594, -0.00000491, 0.00000175, 0.00000615, -0.00000694, 0.00000648, -0.00000409, 0.00000076, -0.00000370, 0.00000355, 0.00000731, -0.00001726, 0.00000169, 0.00000243, 0.00000596, -0.00000632, 0.00000351, -0.00000825, -0.00000064, -0.00000197, -0.00000075, 0.00000194, 0.00000213, -0.00000454, -0.00000159, -0.00000156, 0.00000081, 0.00000258, -0.00000045, -0.00001152, -0.00000340, -0.00001871, -0.00000323, -0.00000837, -0.00002069, -0.00015545, 0.00009964, 0.00027285, 0.00010966, -0.00013822, -0.00002487, -0.00003046, -0.00000500, -0.00000714, 0.00000595, -0.00000584, -0.00000130, -0.00000739, -0.00000681, -0.00000648, -0.00000016, 0.00000333, -0.00000238, -0.00000344, -0.00000008, 0.00000131, 0.00000217, -0.00000182, 0.00000265, -0.00000568, -0.00000263, 0.00000429, -0.00000435, 0.00001339, 0.00000312, -0.00000427, -0.00001158, 0.00001156, -0.00000160, -0.00000818, -0.00000493, 0.00000286, -0.00000111, 0.00000581, -0.00000871, 0.00001843, -0.00001599, 0.00000429, -0.00001214, 0.00001204, -0.00000047, -0.00002450, 0.00001285, -0.00000029, 0.00000315, -0.00001506, 0.00000155, },
		{ 0.00001415, -0.00001056, 0.00000231, -0.00000623, -0.00000002, 0.00001099, -0.00000382, 0.00000392, -0.00000006, 0.00000303, -0.00000611, 0.00000444, -0.00000192, 0.00000718, -0.00000549, 0.00000244, -0.00000339, 0.00000076, -0.00000293, 0.00000332, -0.00000354, 0.00000216, -0.00001290, 0.00000398, -0.00001284, 0.00001059, -0.00000797, 0.00001663, -0.00004225, 0.00000615, -0.00000254, 0.00004544, -0.00000596, 0.00001838, 0.00003734, -0.00000526, -0.00000545, -0.00002252, 0.00000853, -0.00001198, 0.00000402, -0.00000463, -0.00000465, -0.00000273, 0.00000856, -0.00000486, 0.00000118, 0.00000269, -0.00000492, 0.00000248, -0.00000598, 0.00001475, -0.00001425, -0.00000428, 0.00000114, 0.00000813, -0.00000482, 0.00000156, -0.00000462, -0.00000367, -0.00000166, 0.00000001, -0.00000157, 0.00000611, -0.00000612, 0.00000090, -0.00000574, 0.00000478, -0.00000367, 0.00000912, -0.00001754, 0.00000407, -0.00002439, 0.00000303, -0.00002110, 0.00001313, -0.00013847, -0.00000759, 0.00023474, 0.00019829, -0.00007114, -0.00008865, -0.00000381, -0.00003144, 0.00000830, -0.00001034, 0.00001101, -0.00001500, 0.00000601, -0.00001776, 0.00000242, -0.00001131, 0.00001116, -0.00000745, 0.00000285, -0.00000770, 0.00000702, -0.00000352, 0.00000521, -0.00000509, 0.00000513, -0.00001334, 0.00001127, -0.00000963, 0.00001268, 0.00000422, 0.00000339, -0.00001633, 0.00000689, 0.00000361, -0.00000490, -0.00001083, 0.00000518, -0.00000468, 0.00000992, -0.00001134, 0.00001534, -0.00000538, -0.00000541, -0.00000314, -0.00000279, 0.00001561, -0.00002830, 0.00000459, 0.00000599, 0.00000024, -0.00000573, -0.00001162, },
		{ 0.00000720, 0.00000145, -0.00000883, 0.00000438, -0.00001156, 0.00001573, -0.00000341, 0.00000420, -0.00000172, 0.00000582, -0.00000696, 0.00000375, -0.00000249, 0.00000692, -0.00000286, 0.00000077, -0.00000272, 0.00000124, -0.00000452, 0.00000457, -0.00000495, 0.00000530, -0.00001353, 0.00000290, -0.00001197, 0.00000693, -0.00000566, 0.00001387, -0.00002310, -0.00001939, 0.00000602, 0.00002366, 0.00002826, -0.00001194, 0.00005449, -0.00000280, 0.00000204, -0.00002832, 0.00000590, -0.00000860, -0.00000038, -0.00000148, -0.00000479, -0.00000520, 0.00000669, 0.00000079, -0.00000475, 0.00000670, -0.00000593, 0.00000304, -0.00000668, 0.00000964, -0.00000129, -0.00001346, 0.00000201, 0.00000501, 0.00000200, -0.00000396, 0.00000071, -0.00000658, -0.00000144, -0.00000014, -0.00000245, 0.00000524, -0.00000265, -0.00000116, -0.00000450, 0.00000261, -0.00000295, 0.00000873, -0.00001179, -0.00000097, -0.00001656, -0.00000477, -0.00001587, 0.00001031, -0.00008008, -0.00009880, 0.00017787, 0.00022278, 0.00004704, -0.00013624, -0.00000215, -0.00003909, 0.00000806, -0.00001499, 0.00001570, -0.00001621, 0.00000733, -0.00001696, 0.00000105, -0.00001226, 0.00000868, -0.00000438, 0.00000249, -0.00000816, 0.00000564, -0.00000317, 0.00000621, -0.00000613, 0.00000711, -0.00001217, 0.00000567, -0.00000308, 0.00000312, 0.00000962, 0.00000491, -0.00001075, -0.00000452, 0.00000893, -0.00000170, -0.00001202, 0.00000129, -0.00000220, 0.00000636, -0.00000441, 0.00000290, 0.00001005, -0.00001524, 0.00000447, -0.00001264, 0.00001992, -0.00001857, -0.00001055, 0.00001277, -0.00000220, 0.00000063, -0.00001491, },
		{ -0.00000628, 0.00001270, -0.00001397, 0.00000629, -0.00001120, 0.00000775, 0.00000584, -0.00000117, 0.00000191, 0.00000240, -0.00000128, -0.00000237, 0.00000203, 0.00000126, 0.00000387, -0.00000369, 0.00000083, -0.00000187, -0.00000127, 0.00000001, -0.00000002, 0.00000045, -0.00000478, -0.00000721, -0.00000186, -0.00000609, 0.00000649, -0.00000130, 0.00000428, -0.00004086, 0.00001012, 0.00000324, 0.00004953, -0.00001795, 0.00004511, 0.00001713, -0.00000041, -0.00002232, -0.00000661, 0.00000043, -0.00000756, 0.00000235, -0.00000556, -0.00000512, 0.00000182, 0.00000639, -0.00000695, 0.00000486, -0.00000118, -0.00000170, -0.00000099, -0.00000108, 0.00001107, -0.00001680, -0.00000049, 0.00000187, 0.00000733, -0.00000606, 0.00000205, -0.00000586, -0.00000272, -0.00000087, -0.00000093, 0.00000081, 0.00000345, -0.00000528, -0.00000024, -0.00000310, 0.00000216, 0.00000176, 0.00000060, -0.00001200, -0.00000267, -0.00001869, -0.00000143, -0.00001044, -0.00001758, -0.00014598, 0.00010136, 0.00020985, 0.00014639, -0.00011119, -0.00004868, -0.00001801, -0.00001400, -0.00000251, 0.00000279, -0.00000105, -0.00000570, -0.00000343, -0.00001087, -0.00000313, -0.00000311, 0.00000561, -0.00000426, -0.00000129, -0.00000273, 0.00000343, 0.00000029, 0.00000072, -0.00000021, -0.00000118, -0.00000728, 0.00000775, -0.00000710, 0.00001343, 0.00000420, -0.00000122, -0.00001406, 0.00000896, 0.00000225, -0.00000891, -0.00000631, 0.00000281, -0.00000042, 0.00000527, -0.00000904, 0.00001923, -0.00001578, 0.00000321, -0.00001181, 0.00001148, -0.00000103, -0.00002366, 0.00001272, 0.00000114, -0.00000005, -0.00000845, },
		{ -0.00001383, 0.00001170, -0.00000583, -0.00000350, -0.00000033, -0.00000689, 0.00001585, -0.00000529, 0.00000531, -0.00000177, 0.00000469, -0.00000739, 0.00000528, -0.00000312, 0.00000801, -0.00000523, 0.00000258, -0.00000417, 0.00000198, -0.00000465, 0.00000476, -0.00000477, 0.00000374, -0.00001495, 0.00000480, -0.00001424, 0.00001227, -0.00000892, 0.00001640, -0.00003834, -0.00000424, -0.00000087, 0.00004380, 0.00000488, 0.00001567, 0.00004336, -0.00000418, -0.00001111, -0.00002149, 0.00000802, -0.00001224, 0.00000374, -0.00000550, -0.00000399, -0.00000322, 0.00000888, -0.00000443, -0.00000057, 0.00000444, -0.00000539, 0.00000287, -0.00000677, 0.00001286, -0.00000907, -0.00000834, 0.00000174, 0.00000703, -0.00000180, -0.00000168, -0.00000191, -0.00000491, -0.00000171, 0.00000077, -0.00000289, 0.00000671, -0.00000594, 0.00000123, -0.00000581, 0.00000438, -0.00000285, 0.00000860, -0.00001665, 0.00000371, -0.00002378, 0.00000426, -0.00002184, 0.00001374, -0.00013221, 0.00000464, 0.00018964, 0.00018422, -0.00001344, -0.00011278, 0.00000297, -0.00003544, 0.00000933, -0.00001208, 0.00001474, -0.00001767, 0.00000811, -0.00001973, 0.00000358, -0.00001220, 0.00001170, -0.00000812, 0.00000407, -0.00000918, 0.00000783, -0.00000445, 0.00000683, -0.00000674, 0.00000740, -0.00001423, 0.00001033, -0.00000786, 0.00000917, 0.00000668, 0.00000456, -0.00001472, 0.00000157, 0.00000668, -0.00000391, -0.00001229, 0.00000439, -0.00000383, 0.00000963, -0.00001177, 0.00001564, -0.00000461, -0.00000634, -0.00000339, -0.00000246, 0.00001401, -0.00002671, 0.00000345, 0.00000863, -0.00000340, -0.00000104, },
		{ -0.00001051, -0.00000045, 0.00000832, -0.00001409, 0.00000795, -0.00001425, 0.00001502, -0.00000039, 0.00000261, -0.00000075, 0.00000474, -0.00000559, 0.00000224, -0.00000140, 0.00000535, -0.00000081, -0.00000023, -0.00000230, 0.00000088, -0.00000409, 0.00000345, -0.00000356, 0.00000406, -0.00001210, 0.00000009, -0.00000972, 0.00000412, -0.00000139, 0.00000811, -0.00001482, -0.00002861, 0.00000432, 0.00002170, 0.00003642, -0.00000916, 0.00005516, 0.00000365, -0.00000436, -0.00002900, 0.00000650, -0.00000960, -0.00000000, -0.00000289, -0.00000404, -0.00000566, 0.00000693, 0.00000105, -0.00000531, 0.00000601, -0.00000371, 0.00000084, -0.00000419, 0.00000470, 0.00000425, -0.00001549, 0.00000134, 0.00000371, 0.00000466, -0.00000548, 0.00000089, -0.00000542, -0.00000294, 0.00000103, -0.00000310, 0.00000451, -0.00000130, -0.00000198, -0.00000317, 0.00000086, -0.00000107, 0.00000727, -0.00001008, -0.00000212, -0.00001509, -0.00000507, -0.00001404, 0.00000789, -0.00007458, -0.00008752, 0.00016041, 0.00017539, 0.00009762, -0.00013152, -0.00001520, -0.00003101, 0.00000121, -0.00001160, 0.00001432, -0.00001342, 0.00000451, -0.00001436, -0.00000189, -0.00000967, 0.00000630, -0.00000261, 0.00000127, -0.00000684, 0.00000370, -0.00000165, 0.00000513, -0.00000459, 0.00000541, -0.00000865, 0.00000114, 0.00000109, -0.00000109, 0.00001128, 0.00000552, -0.00000753, -0.00000870, 0.00000857, 0.00000129, -0.00001306, -0.00000010, -0.00000160, 0.00000640, -0.00000479, 0.00000275, 0.00001093, -0.00001557, 0.00000354, -0.00001173, 0.00001815, -0.00001756, -0.00001090, 0.00001401, -0.00000296, 0.00000026, },
		{ -0.00000231, -0.00001198, 0.00001439, -0.00001285, 0.00000352, -0.00000764, 0.00000179, 0.00001141, -0.00000371, 0.00000389, 0.00000024, 0.00000091, -0.00000436, 0.00000355, -0.00000066, 0.00000566, -0.00000425, 0.00000139, -0.00000288, 0.00000015, -0.00000209, 0.00000208, -0.00000148, -0.00000249, -0.00001040, 0.00000039, -0.00000902, 0.00001024, -0.00000511, 0.00000788, -0.00004120, 0.00000133, 0.00000385, 0.00005156, -0.00000858, 0.00004186, 0.00002561, -0.00000413, -0.00002586, -0.00000533, -0.00000051, -0.00000742, 0.00000140, -0.00000562, -0.00000494, 0.00000169, 0.00000652, -0.00000657, 0.00000295, 0.00000142, -0.00000351, 0.00000089, -0.00000397, 0.00001205, -0.00001392, -0.00000379, 0.00000206, 0.00000739, -0.00000419, -0.00000068, -0.00000307, -0.00000450, -0.00000023, -0.00000066, -0.00000058, 0.00000467, -0.00000569, 0.00000067, -0.00000407, 0.00000286, 0.00000156, 0.00000100, -0.00001194, -0.00000225, -0.00001849, 0.00000018, -0.00001198, -0.00001559, -0.00013587, 0.00010109, 0.00016160, 0.00015862, -0.00007064, -0.00007371, -0.00000684, -0.00002158, 0.00000106, -0.00000009, 0.00000361, -0.00000968, -0.00000002, -0.00001419, -0.00000055, -0.00000529, 0.00000722, -0.00000571, 0.00000064, -0.00000508, 0.00000513, -0.00000132, 0.00000303, -0.00000274, 0.00000251, -0.00001038, 0.00000930, -0.00000773, 0.00001158, 0.00000622, 0.00000068, -0.00001436, 0.00000485, 0.00000583, -0.00000878, -0.00000784, 0.00000259, 0.00000021, 0.00000470, -0.00000908, 0.00001951, -0.00001512, 0.00000170, -0.00001085, 0.00001038, -0.00000139, -0.00002259, 0.00001209, 0.00000305, -0.00000347, },
		{ 0.00000172, -0.00001312, 0.00000646, 0.00000040, -0.00000923, 0.00000465, -0.00001220, 0.00001832, -0.00000460, 0.00000510, -0.00000200, 0.00000486, -0.00000728, 0.00000487, -0.00000309, 0.00000752, -0.00000392, 0.00000199, -0.00000423, 0.00000231, -0.00000521, 0.00000484, -0.00000454, 0.00000371, -0.00001509, 0.00000360, -0.00001354, 0.00001132, -0.00000667, 0.00001247, -0.00003070, -0.00001566, -0.00000116, 0.00004243, 0.00001580, 0.00001442, 0.00004849, -0.00000202, -0.00001761, -0.00002038, 0.00000763, -0.00001260, 0.00000342, -0.00000634, -0.00000339, -0.00000354, 0.00000898, -0.00000396, -0.00000183, 0.00000503, -0.00000452, 0.00000184, -0.00000559, 0.00000900, -0.00000326, -0.00001185, 0.00000207, 0.00000554, 0.00000147, -0.00000446, -0.00000030, -0.00000502, -0.00000238, 0.00000169, -0.00000391, 0.00000660, -0.00000491, 0.00000067, -0.00000494, 0.00000302, -0.00000110, 0.00000710, -0.00001478, 0.00000235, -0.00002185, 0.00000365, -0.00001997, 0.00001094, -0.00012213, 0.00001008, 0.00015989, 0.00015755, 0.00003880, -0.00012378, 0.00000053, -0.00003324, 0.00000600, -0.00001074, 0.00001554, -0.00001732, 0.00000742, -0.00001908, 0.00000239, -0.00001102, 0.00001046, -0.00000729, 0.00000389, -0.00000907, 0.00000703, -0.00000391, 0.00000690, -0.00000660, 0.00000747, -0.00001266, 0.00000722, -0.00000436, 0.00000465, 0.00000931, 0.00000508, -0.00001204, -0.00000341, 0.00000819, -0.00000198, -0.00001347, 0.00000317, -0.00000280, 0.00000892, -0.00001155, 0.00001527, -0.00000339, -0.00000768, -0.00000313, -0.00000237, 0.00001236, -0.00002494, 0.00000233, 0.00001074, -0.00000594, },
		{ -0.00000125, -0.00000519, -0.00000760, 0.00001294, -0.00001612, 0.00000839, -0.00001396, 0.00001148, 0.00000441, 0.00000004, 0.00000112, 0.00000282, -0.00000357, 0.00000022, 0.00000018, 0.00000330, 0.00000143, -0.00000130, -0.00000157, -0.00000006, -0.00000290, 0.00000155, -0.00000143, 0.00000205, -0.00000983, -0.00000338, -0.00000689, 0.00000055, 0.00000347, 0.00000235, -0.00000781, -0.00003377, -0.00000217, 0.00002126, 0.00004304, -0.00000375, 0.00005453, 0.00001132, -0.00001149, -0.00002956, 0.00000683, -0.00001015, -0.00000019, -0.00000381, -0.00000377, -0.00000566, 0.00000688, 0.00000123, -0.00000531, 0.00000454, -0.00000105, -0.00000153, -0.00000155, 0.00000037, 0.00000781, -0.00001528, -0.00000056, 0.00000311, 0.00000618, -0.00000532, -0.00000055, -0.00000333, -0.00000457, 0.00000192, -0.00000322, 0.00000319, 0.00000042, -0.00000299, -0.00000172, -0.00000092, 0.00000076, 0.00000581, -0.00000837, -0.00000340, -0.00001314, -0.00000623, -0.00001102, 0.00000417, -0.00006788, -0.00007909, 0.00014793, 0.00013364, 0.00012771, -0.00010980, -0.00003510, -0.00001986, -0.00000735, -0.00000697, 0.00001134, -0.00000896, 0.00000036, -0.00001065, -0.00000564, -0.00000647, 0.00000348, -0.00000045, -0.00000048, -0.00000477, 0.00000107, 0.00000041, 0.00000337, -0.00000223, 0.00000284, -0.00000454, -0.00000325, 0.00000456, -0.00000404, 0.00001146, 0.00000677, -0.00000495, -0.00001097, 0.00000636, 0.00000452, -0.00001332, -0.00000187, -0.00000093, 0.00000615, -0.00000485, 0.00000249, 0.00001166, -0.00001577, 0.00000253, -0.00001049, 0.00001600, -0.00001634, -0.00001103, 0.00001455, -0.00000275, },
		{ -0.00000639, 0.00000248, -0.00001488, 0.00001257, -0.00000879, -0.00000092, -0.00000318, -0.00000439, 0.00001602, -0.00000501, 0.00000494, -0.00000101, 0.00000214, -0.00000542, 0.00000421, -0.00000173, 0.00000641, -0.00000393, 0.00000141, -0.00000349, 0.00000112, -0.00000353, 0.00000335, -0.00000247, -0.00000124, -0.00001220, 0.00000112, -0.00001033, 0.00001179, -0.00000587, 0.00000741, -0.00003642, -0.00001060, 0.00000392, 0.00005299, 0.00000213, 0.00003889, 0.00003411, -0.00000760, -0.00002977, -0.00000421, -0.00000108, -0.00000771, 0.00000075, -0.00000596, -0.00000455, 0.00000151, 0.00000647, -0.00000595, 0.00000115, 0.00000322, -0.00000418, 0.00000145, -0.00000478, 0.00001042, -0.00000932, -0.00000750, 0.00000257, 0.00000651, -0.00000131, -0.00000378, -0.00000074, -0.00000543, -0.00000024, 0.00000003, -0.00000202, 0.00000549, -0.00000550, 0.00000088, -0.00000424, 0.00000272, 0.00000214, 0.00000058, -0.00001113, -0.00000244, -0.00001757, 0.00000077, -0.00001192, -0.00001584, -0.00012373, 0.00009701, 0.00012787, 0.00015192, -0.00002501, -0.00009342, -0.00000088, -0.00002522, 0.00000180, -0.00000122, 0.00000669, -0.00001187, 0.00000166, -0.00001576, 0.00000039, -0.00000596, 0.00000754, -0.00000616, 0.00000170, -0.00000640, 0.00000572, -0.00000199, 0.00000438, -0.00000409, 0.00000454, -0.00001130, 0.00000860, -0.00000617, 0.00000806, 0.00000893, 0.00000162, -0.00001302, 0.00000025, 0.00000833, -0.00000755, -0.00000945, 0.00000212, 0.00000088, 0.00000391, -0.00000864, 0.00001917, -0.00001403, -0.00000019, -0.00000932, 0.00000886, -0.00000162, -0.00002134, 0.00001122, 0.00000490, },
		{ -0.00000731, 0.00000276, -0.00001027, -0.00000007, 0.00000622, -0.00001315, 0.00000745, -0.00001501, 0.00001787, -0.00000161, 0.00000333, -0.00000083, 0.00000372, -0.00000605, 0.00000344, -0.00000209, 0.00000602, -0.00000188, 0.00000091, -0.00000368, 0.00000178, -0.00000467, 0.00000372, -0.00000307, 0.00000235, -0.00001368, 0.00000092, -0.00001132, 0.00000843, -0.00000225, 0.00000650, -0.00002184, -0.00002513, -0.00000544, 0.00004213, 0.00002607, 0.00001524, 0.00005259, 0.00000113, -0.00002485, -0.00001913, 0.00000711, -0.00001275, 0.00000275, -0.00000686, -0.00000313, -0.00000350, 0.00000877, -0.00000346, -0.00000256, 0.00000458, -0.00000266, -0.00000010, -0.00000328, 0.00000449, 0.00000180, -0.00001381, 0.00000148, 0.00000436, 0.00000412, -0.00000593, -0.00000032, -0.00000391, -0.00000355, 0.00000265, -0.00000455, 0.00000582, -0.00000327, -0.00000042, -0.00000356, 0.00000123, 0.00000101, 0.00000516, -0.00001252, 0.00000052, -0.00001910, 0.00000174, -0.00001627, 0.00000592, -0.00010982, 0.00001113, 0.00014027, 0.00012665, 0.00007883, -0.00011963, -0.00001049, -0.00002601, -0.00000059, -0.00000709, 0.00001396, -0.00001452, 0.00000457, -0.00001644, -0.00000046, -0.00000843, 0.00000805, -0.00000547, 0.00000266, -0.00000771, 0.00000498, -0.00000229, 0.00000576, -0.00000505, 0.00000589, -0.00000947, 0.00000302, -0.00000029, 0.00000031, 0.00001114, 0.00000576, -0.00000934, -0.00000703, 0.00000777, 0.00000070, -0.00001415, 0.00000150, -0.00000160, 0.00000782, -0.00001082, 0.00001448, -0.00000204, -0.00000915, -0.00000261, -0.00000230, 0.00001057, -0.00002302, 0.00000137, 0.00001218, },
		{ -0.00000203, -0.00000312, -0.00000011, -0.00001294, 0.00001438, -0.00001474, 0.00000616, -0.00001144, 0.00000615, 0.00000978, -0.00000248, 0.00000297, 0.00000098, -0.00000175, -0.00000158, 0.00000156, 0.00000146, 0.00000321, -0.00000192, -0.00000095, -0.00000110, -0.00000154, -0.00000041, 0.00000065, 0.00000016, -0.00000769, -0.00000643, -0.00000458, -0.00000253, 0.00000737, -0.00000159, -0.00000398, -0.00003362, -0.00001305, 0.00002114, 0.00004880, 0.00000365, 0.00005349, 0.00001951, -0.00001895, -0.00003010, 0.00000683, -0.00001019, -0.00000099, -0.00000421, -0.00000397, -0.00000527, 0.00000660, 0.00000134, -0.00000492, 0.00000272, 0.00000141, -0.00000334, 0.00000035, -0.00000237, 0.00000880, -0.00001283, -0.00000351, 0.00000318, 0.00000647, -0.00000363, -0.00000314, -0.00000098, -0.00000581, 0.00000231, -0.00000282, 0.00000152, 0.00000212, -0.00000375, -0.00000063, -0.00000222, 0.00000206, 0.00000481, -0.00000716, -0.00000429, -0.00001132, -0.00000742, -0.00000801, 0.00000075, -0.00006192, -0.00007094, 0.00013638, 0.00010125, 0.00013827, -0.00007641, -0.00005639, -0.00000960, -0.00001478, -0.00000319, 0.00000848, -0.00000448, -0.00000354, -0.00000727, -0.00000886, -0.00000385, 0.00000124, 0.00000123, -0.00000196, -0.00000278, -0.00000137, 0.00000214, 0.00000179, -0.00000000, 0.00000047, -0.00000107, -0.00000634, 0.00000637, -0.00000500, 0.00000978, 0.00000890, -0.00000334, -0.00001134, 0.00000294, 0.00000728, -0.00001250, -0.00000401, -0.00000030, 0.00000575, -0.00000484, 0.00000240, 0.00001199, -0.00001570, 0.00000137, -0.00000886, 0.00001348, -0.00001497, -0.00001095, 0.00001459, },
		{ 0.00000634, -0.00000839, 0.00000555, -0.00001503, 0.00000807, -0.00000329, -0.00000535, 0.00000067, -0.00000937, 0.00001854, -0.00000431, 0.00000450, -0.00000091, 0.00000209, -0.00000534, 0.00000383, -0.00000177, 0.00000602, -0.00000270, 0.00000084, -0.00000351, 0.00000132, -0.00000398, 0.00000347, -0.00000224, -0.00000133, -0.00001239, 0.00000018, -0.00000987, 0.00001097, -0.00000368, 0.00000352, -0.00002819, -0.00002287, 0.00000108, 0.00005496, 0.00001316, 0.00003722, 0.00004205, -0.00001062, -0.00003408, -0.00000317, -0.00000145, -0.00000815, 0.00000011, -0.00000628, -0.00000424, 0.00000150, 0.00000619, -0.00000520, -0.00000030, 0.00000401, -0.00000366, 0.00000071, -0.00000381, 0.00000702, -0.00000417, -0.00001065, 0.00000274, 0.00000535, 0.00000171, -0.00000628, 0.00000034, -0.00000514, -0.00000098, 0.00000105, -0.00000334, 0.00000577, -0.00000465, 0.00000044, -0.00000374, 0.00000195, 0.00000329, -0.00000046, -0.00000978, -0.00000314, -0.00001593, 0.00000019, -0.00001012, -0.00001829, -0.00010968, 0.00008938, 0.00010597, 0.00013340, 0.00001773, -0.00010305, -0.00000214, -0.00002403, -0.00000074, -0.00000015, 0.00000764, -0.00001183, 0.00000127, -0.00001532, -0.00000046, -0.00000507, 0.00000654, -0.00000552, 0.00000172, -0.00000649, 0.00000503, -0.00000153, 0.00000452, -0.00000404, 0.00000478, -0.00001015, 0.00000603, -0.00000307, 0.00000370, 0.00001157, 0.00000225, -0.00001099, -0.00000373, 0.00000909, -0.00000518, -0.00001092, 0.00000129, 0.00000169, 0.00000281, -0.00000770, 0.00001830, -0.00001268, -0.00000228, -0.00000740, 0.00000712, -0.00000183, -0.00001997, 0.00001034, },
		{ 0.00001294, -0.00000774, 0.00000262, -0.00000635, -0.00000633, 0.00001010, -0.00001424, 0.00000761, -0.00001519, 0.00001476, 0.00000307, 0.00000063, 0.00000112, 0.00000192, -0.00000432, 0.00000159, -0.00000065, 0.00000408, 0.00000032, -0.00000020, -0.00000286, 0.00000073, -0.00000345, 0.00000193, -0.00000096, 0.00000036, -0.00001155, -0.00000231, -0.00000855, 0.00000475, 0.00000276, 0.00000056, -0.00001434, -0.00003021, -0.00001459, 0.00004258, 0.00003572, 0.00001812, 0.00005606, 0.00000479, -0.00003253, -0.00001779, 0.00000636, -0.00001251, 0.00000155, -0.00000689, -0.00000335, -0.00000303, 0.00000825, -0.00000297, -0.00000281, 0.00000339, -0.00000035, -0.00000223, -0.00000084, 0.00000057, 0.00000500, -0.00001366, -0.00000031, 0.00000385, 0.00000564, -0.00000571, -0.00000194, -0.00000192, -0.00000484, 0.00000340, -0.00000472, 0.00000450, -0.00000131, -0.00000162, -0.00000218, -0.00000045, 0.00000295, 0.00000335, -0.00001042, -0.00000124, -0.00001612, -0.00000075, -0.00001180, 0.00000018, -0.00009716, 0.00001032, 0.00012607, 0.00009780, 0.00010341, -0.00010175, -0.00002704, -0.00001642, -0.00000833, -0.00000268, 0.00001122, -0.00001047, 0.00000074, -0.00001295, -0.00000391, -0.00000542, 0.00000535, -0.00000338, 0.00000103, -0.00000574, 0.00000239, -0.00000030, 0.00000411, -0.00000287, 0.00000360, -0.00000580, -0.00000105, 0.00000313, -0.00000275, 0.00001138, 0.00000724, -0.00000738, -0.00000880, 0.00000562, 0.00000356, -0.00001396, -0.00000069, -0.00000026, 0.00000644, -0.00000981, 0.00001357, -0.00000085, -0.00001051, -0.00000202, -0.00000210, 0.00000858, -0.00002102, 0.00000064, },
		{ 0.00001435, -0.00000124, -0.00000470, 0.00000388, -0.00001575, 0.00001259, -0.00001068, 0.00000244, -0.00000796, 0.00000042, 0.00001439, -0.00000386, 0.00000392, 0.00000001, -0.00000079, -0.00000260, 0.00000223, 0.00000032, 0.00000410, -0.00000178, -0.00000077, -0.00000183, -0.00000057, -0.00000181, 0.00000201, -0.00000093, -0.00000643, -0.00000830, -0.00000351, -0.00000430, 0.00000932, -0.00000273, -0.00000398, -0.00002855, -0.00002636, 0.00001919, 0.00005480, 0.00001210, 0.00005313, 0.00002743, -0.00002635, -0.00003073, 0.00000655, -0.00000979, -0.00000225, -0.00000428, -0.00000446, -0.00000468, 0.00000626, 0.00000129, -0.00000429, 0.00000093, 0.00000320, -0.00000414, 0.00000104, -0.00000320, 0.00000741, -0.00000879, -0.00000683, 0.00000351, 0.00000592, -0.00000100, -0.00000597, 0.00000077, -0.00000611, 0.00000197, -0.00000190, -0.00000033, 0.00000353, -0.00000400, -0.00000014, -0.00000286, 0.00000270, 0.00000440, -0.00000657, -0.00000465, -0.00000987, -0.00000820, -0.00000571, -0.00000146, -0.00005765, -0.00006187, 0.00012388, 0.00007880, 0.00013331, -0.00003814, -0.00007360, -0.00000356, -0.00001896, -0.00000170, 0.00000696, -0.00000120, -0.00000607, -0.00000521, -0.00001072, -0.00000254, 0.00000018, 0.00000191, -0.00000268, -0.00000145, -0.00000298, 0.00000295, 0.00000097, 0.00000145, -0.00000094, 0.00000102, -0.00000756, 0.00000618, -0.00000393, 0.00000653, 0.00001160, -0.00000245, -0.00001043, -0.00000069, 0.00000874, -0.00001035, -0.00000640, 0.00000018, 0.00000535, -0.00000496, 0.00000265, 0.00001183, -0.00001538, 0.00000009, -0.00000693, 0.00001074, -0.00001355, -0.00001069, },
		{ 0.00000961, 0.00000718, -0.00000934, 0.00000697, -0.00001306, 0.00000224, 0.00000194, -0.00000825, 0.00000279, -0.00001228, 0.00001841, -0.00000148, 0.00000262, 0.00000042, 0.00000096, -0.00000436, 0.00000264, -0.00000101, 0.00000477, -0.00000088, -0.00000009, -0.00000305, 0.00000082, -0.00000351, 0.00000258, -0.00000099, -0.00000249, -0.00001131, -0.00000195, -0.00000816, 0.00000841, 0.00000046, -0.00000216, -0.00001909, -0.00003242, -0.00000653, 0.00005791, 0.00002403, 0.00003741, 0.00004933, -0.00001334, -0.00003864, -0.00000222, -0.00000179, -0.00000852, -0.00000076, -0.00000635, -0.00000422, 0.00000179, 0.00000562, -0.00000439, -0.00000131, 0.00000385, -0.00000218, -0.00000088, -0.00000180, 0.00000303, 0.00000025, -0.00001234, 0.00000200, 0.00000451, 0.00000409, -0.00000739, -0.00000025, -0.00000363, -0.00000225, 0.00000218, -0.00000438, 0.00000546, -0.00000327, -0.00000041, -0.00000291, 0.00000094, 0.00000458, -0.00000169, -0.00000829, -0.00000400, -0.00001386, -0.00000124, -0.00000711, -0.00002209, -0.00009481, 0.00007970, 0.00009213, 0.00010989, 0.00005165, -0.00010062, -0.00001029, -0.00001888, -0.00000570, 0.00000254, 0.00000685, -0.00000996, -0.00000071, -0.00001339, -0.00000257, -0.00000314, 0.00000471, -0.00000415, 0.00000095, -0.00000558, 0.00000335, -0.00000026, 0.00000374, -0.00000287, 0.00000367, -0.00000766, 0.00000255, 0.00000051, -0.00000039, 0.00001320, 0.00000335, -0.00000923, -0.00000621, 0.00000787, -0.00000201, -0.00001191, -0.00000003, 0.00000267, 0.00000141, -0.00000637, 0.00001713, -0.00001133, -0.00000432, -0.00000532, 0.00000535, -0.00000210, -0.00001855, },
		{ 0.00000019, 0.00001316, -0.00000758, 0.00000194, -0.00000242, -0.00001108, 0.00001115, -0.00001234, 0.00000552, -0.00001338, 0.00000991, 0.00000835, -0.00000199, 0.00000297, 0.00000033, -0.00000285, -0.00000001, 0.00000059, 0.00000233, 0.00000208, -0.00000087, -0.00000218, -0.00000037, -0.00000214, 0.00000014, 0.00000102, -0.00000145, -0.00000960, -0.00000511, -0.00000625, 0.00000146, 0.00000688, -0.00000357, -0.00001005, -0.00002980, -0.00002789, 0.00004223, 0.00004554, 0.00002250, 0.00005974, 0.00000823, -0.00004021, -0.00001649, 0.00000534, -0.00001183, -0.00000016, -0.00000647, -0.00000399, -0.00000223, 0.00000753, -0.00000255, -0.00000269, 0.00000181, 0.00000185, -0.00000387, 0.00000093, -0.00000190, 0.00000585, -0.00001144, -0.00000308, 0.00000397, 0.00000601, -0.00000399, -0.00000461, 0.00000020, -0.00000567, 0.00000360, -0.00000431, 0.00000276, 0.00000067, -0.00000258, -0.00000118, -0.00000166, 0.00000436, 0.00000200, -0.00000882, -0.00000258, -0.00001331, -0.00000320, -0.00000746, -0.00000503, -0.00008561, 0.00000956, 0.00011395, 0.00007455, 0.00011277, -0.00007423, -0.00004468, -0.00000776, -0.00001488, 0.00000081, 0.00000872, -0.00000651, -0.00000277, -0.00000978, -0.00000688, -0.00000295, 0.00000319, -0.00000173, -0.00000037, -0.00000384, 0.00000001, 0.00000134, 0.00000268, -0.00000083, 0.00000152, -0.00000270, -0.00000395, 0.00000503, -0.00000389, 0.00000973, 0.00000967, -0.00000634, -0.00000886, 0.00000246, 0.00000578, -0.00001251, -0.00000340, 0.00000112, 0.00000493, -0.00000873, 0.00001280, -0.00000004, -0.00001165, -0.00000145, -0.00000174, 0.00000648, -0.00001904, },
		{ -0.00001030, 0.00001403, -0.00000071, -0.00000558, 0.00000639, -0.00001603, 0.00000830, -0.00000534, -0.00000127, -0.00000487, -0.00000443, 0.00001713, -0.00000338, 0.00000346, 0.00000029, -0.00000095, -0.00000267, 0.00000205, 0.00000004, 0.00000401, -0.00000086, -0.00000107, -0.00000207, -0.00000023, -0.00000235, 0.00000238, -0.00000097, -0.00000628, -0.00000883, -0.00000380, -0.00000462, 0.00000918, -0.00000117, -0.00000713, -0.00002040, -0.00003911, 0.00001301, 0.00006197, 0.00002084, 0.00005433, 0.00003458, -0.00003352, -0.00003146, 0.00000604, -0.00000913, -0.00000373, -0.00000426, -0.00000500, -0.00000414, 0.00000604, 0.00000105, -0.00000352, -0.00000059, 0.00000408, -0.00000380, 0.00000051, -0.00000239, 0.00000441, -0.00000427, -0.00000960, 0.00000347, 0.00000513, 0.00000172, -0.00000810, 0.00000116, -0.00000517, 0.00000088, -0.00000058, -0.00000214, 0.00000445, -0.00000365, -0.00000022, -0.00000295, 0.00000285, 0.00000439, -0.00000642, -0.00000464, -0.00000873, -0.00000850, -0.00000430, -0.00000230, -0.00005506, -0.00005192, 0.00011029, 0.00006467, 0.00011833, -0.00000164, -0.00008268, -0.00000352, -0.00001911, -0.00000288, 0.00000714, 0.00000041, -0.00000683, -0.00000474, -0.00001100, -0.00000263, 0.00000035, 0.00000159, -0.00000258, -0.00000093, -0.00000355, 0.00000266, 0.00000107, 0.00000193, -0.00000119, 0.00000161, -0.00000703, 0.00000438, -0.00000145, 0.00000252, 0.00001406, -0.00000157, -0.00000918, -0.00000344, 0.00000831, -0.00000693, -0.00000877, 0.00000031, 0.00000508, -0.00000533, 0.00000328, 0.00001129, -0.00001496, -0.00000114, -0.00000487, 0.00000797, -0.00001220, },
		{ -0.00001718, 0.00000915, 0.00000738, -0.00000938, 0.00000717, -0.00000990, -0.00000352, 0.00000548, -0.00000873, 0.00000277, -0.00001296, 0.00001584, 0.00000293, -0.00000007, 0.00000236, -0.00000059, -0.00000310, 0.00000120, 0.00000004, 0.00000320, 0.00000099, -0.00000095, -0.00000244, -0.00000006, -0.00000253, 0.00000122, 0.00000069, -0.00000406, -0.00000975, -0.00000439, -0.00000610, 0.00000522, 0.00000502, -0.00000757, -0.00001172, -0.00003686, -0.00001954, 0.00006110, 0.00003507, 0.00003938, 0.00005636, -0.00001626, -0.00004309, -0.00000145, -0.00000219, -0.00000867, -0.00000197, -0.00000609, -0.00000456, 0.00000240, 0.00000484, -0.00000362, -0.00000185, 0.00000295, -0.00000020, -0.00000270, 0.00000036, -0.00000040, 0.00000297, -0.00001207, 0.00000013, 0.00000430, 0.00000539, -0.00000683, -0.00000241, -0.00000135, -0.00000357, 0.00000308, -0.00000492, 0.00000457, -0.00000157, -0.00000135, -0.00000213, 0.00000010, 0.00000564, -0.00000272, -0.00000704, -0.00000466, -0.00001174, -0.00000304, -0.00000363, -0.00002611, -0.00008050, 0.00006972, 0.00008277, 0.00008671, 0.00007373, -0.00008704, -0.00002289, -0.00001198, -0.00001135, 0.00000557, 0.00000532, -0.00000724, -0.00000330, -0.00001089, -0.00000504, -0.00000099, 0.00000276, -0.00000267, -0.00000009, -0.00000419, 0.00000128, 0.00000122, 0.00000263, -0.00000125, 0.00000204, -0.00000483, -0.00000075, 0.00000346, -0.00000318, 0.00001311, 0.00000547, -0.00000837, -0.00000684, 0.00000499, 0.00000129, -0.00001196, -0.00000198, 0.00000384, -0.00000022, -0.00000485, 0.00001592, -0.00001024, -0.00000613, -0.00000323, 0.00000368, -0.00000244, },
		{ -0.00001722, -0.00000000, 0.00001301, -0.00000719, 0.00000128, 0.00000075, -0.00001364, 0.00000932, -0.00000816, 0.00000224, -0.00001068, 0.00000457, 0.00001294, -0.00000350, 0.00000386, -0.00000034, -0.00000226, -0.00000088, 0.00000118, 0.00000122, 0.00000301, -0.00000083, -0.00000191, -0.00000117, -0.00000121, -0.00000108, 0.00000229, -0.00000248, -0.00000848, -0.00000683, -0.00000506, -0.00000068, 0.00000918, -0.00000499, -0.00000956, -0.00002446, -0.00004302, 0.00003864, 0.00005662, 0.00002761, 0.00006461, 0.00001074, -0.00004753, -0.00001534, 0.00000412, -0.00001085, -0.00000222, -0.00000579, -0.00000485, -0.00000129, 0.00000681, -0.00000228, -0.00000230, 0.00000018, 0.00000351, -0.00000461, 0.00000157, -0.00000261, 0.00000452, -0.00000778, -0.00000616, 0.00000427, 0.00000560, -0.00000141, -0.00000742, 0.00000157, -0.00000549, 0.00000304, -0.00000331, 0.00000075, 0.00000242, -0.00000306, -0.00000071, -0.00000227, 0.00000518, 0.00000116, -0.00000775, -0.00000345, -0.00001085, -0.00000530, -0.00000381, -0.00000901, -0.00007585, 0.00000970, 0.00010227, 0.00005780, 0.00010983, -0.00004262, -0.00005887, -0.00000282, -0.00001849, 0.00000221, 0.00000742, -0.00000365, -0.00000507, -0.00000774, -0.00000867, -0.00000161, 0.00000205, -0.00000091, -0.00000115, -0.00000249, -0.00000163, 0.00000211, 0.00000195, 0.00000050, 0.00000028, -0.00000082, -0.00000521, 0.00000513, -0.00000310, 0.00000650, 0.00001261, -0.00000584, -0.00000793, -0.00000064, 0.00000654, -0.00000958, -0.00000651, 0.00000238, 0.00000345, -0.00000779, 0.00001234, 0.00000035, -0.00001258, -0.00000084, -0.00000131, 0.00000441, },
		{ -0.00001099, -0.00000986, 0.00001381, -0.00000064, -0.00000561, 0.00000741, -0.00001427, 0.00000270, -0.00000024, -0.00000364, -0.00000313, -0.00000759, 0.00001748, -0.00000091, 0.00000166, 0.00000167, -0.00000196, -0.00000208, 0.00000127, 0.00000037, 0.00000322, 0.00000050, -0.00000160, -0.00000197, -0.00000047, -0.00000214, 0.00000194, -0.00000023, -0.00000693, -0.00000839, -0.00000495, -0.00000400, 0.00000757, 0.00000206, -0.00001173, -0.00001187, -0.00004806, 0.00000081, 0.00007037, 0.00002975, 0.00005750, 0.00004090, -0.00004066, -0.00003209, 0.00000524, -0.00000832, -0.00000523, -0.00000434, -0.00000539, -0.00000380, 0.00000604, 0.00000062, -0.00000273, -0.00000170, 0.00000406, -0.00000253, -0.00000086, -0.00000060, 0.00000088, -0.00000043, -0.00001100, 0.00000256, 0.00000465, 0.00000381, -0.00000881, -0.00000014, -0.00000305, -0.00000070, 0.00000087, -0.00000368, 0.00000479, -0.00000278, -0.00000067, -0.00000276, 0.00000282, 0.00000447, -0.00000639, -0.00000455, -0.00000767, -0.00000853, -0.00000359, -0.00000212, -0.00005350, -0.00004190, 0.00009655, 0.00005618, 0.00009885, 0.00002807, -0.00008185, -0.00000920, -0.00001599, -0.00000597, 0.00000851, 0.00000070, -0.00000619, -0.00000545, -0.00001014, -0.00000368, 0.00000128, 0.00000067, -0.00000196, -0.00000102, -0.00000330, 0.00000156, 0.00000179, 0.00000169, -0.00000057, 0.00000112, -0.00000540, 0.00000182, 0.00000144, -0.00000116, 0.00001536, 0.00000006, -0.00000849, -0.00000453, 0.00000586, -0.00000265, -0.00001068, -0.00000011, 0.00000507, -0.00000597, 0.00000420, 0.00001053, -0.00001466, -0.00000212, -0.00000287, 0.00000536, },
		{ -0.00000275, -0.00001599, 0.00000902, 0.00000703, -0.00000878, 0.00000667, -0.00000646, -0.00000807, 0.00000655, -0.00000668, 0.00000098, -0.00001196, 0.00001167, 0.00000787, -0.00000259, 0.00000403, -0.00000173, -0.00000227, 0.00000011, 0.00000081, 0.00000190, 0.00000237, -0.00000130, -0.00000206, -0.00000088, -0.00000160, 0.00000001, 0.00000207, -0.00000526, -0.00000855, -0.00000624, -0.00000463, 0.00000251, 0.00000856, -0.00001097, -0.00000791, -0.00003522, -0.00003681, 0.00006250, 0.00004724, 0.00004258, 0.00006394, -0.00002008, -0.00004700, -0.00000102, -0.00000260, -0.00000863, -0.00000345, -0.00000557, -0.00000514, 0.00000317, 0.00000401, -0.00000300, -0.00000198, 0.00000158, 0.00000179, -0.00000416, 0.00000192, -0.00000251, 0.00000357, -0.00000992, -0.00000261, 0.00000462, 0.00000563, -0.00000483, -0.00000550, 0.00000091, -0.00000432, 0.00000337, -0.00000481, 0.00000315, 0.00000023, -0.00000209, -0.00000167, -0.00000032, 0.00000622, -0.00000335, -0.00000622, -0.00000493, -0.00000982, -0.00000477, -0.00000040, -0.00002940, -0.00006779, 0.00006073, 0.00007532, 0.00006691, 0.00008390, -0.00006565, -0.00003623, -0.00000611, -0.00001569, 0.00000755, 0.00000420, -0.00000480, -0.00000545, -0.00000880, -0.00000700, 0.00000058, 0.00000137, -0.00000162, -0.00000090, -0.00000290, -0.00000054, 0.00000228, 0.00000182, 0.00000013, 0.00000069, -0.00000259, -0.00000297, 0.00000498, -0.00000406, 0.00001103, 0.00000864, -0.00000843, -0.00000592, 0.00000131, 0.00000383, -0.00001060, -0.00000463, 0.00000512, -0.00000194, -0.00000333, 0.00001488, -0.00000957, -0.00000762, -0.00000121, 0.00000213, },
		{ 0.00000256, -0.00001567, 0.00000004, 0.00001269, -0.00000687, 0.00000103, 0.00000274, -0.00001395, 0.00000525, -0.00000294, -0.00000089, -0.00000827, -0.00000009, 0.00001584, -0.00000320, 0.00000333, 0.00000025, -0.00000269, -0.00000093, 0.00000105, 0.00000085, 0.00000307, -0.00000009, -0.00000205, -0.00000153, -0.00000088, -0.00000149, 0.00000264, -0.00000256, -0.00000834, -0.00000735, -0.00000501, -0.00000159, 0.00000954, -0.00000383, -0.00001210, -0.00001628, -0.00005661, 0.00002923, 0.00006968, 0.00003299, 0.00007139, 0.00001186, -0.00005434, -0.00001435, 0.00000273, -0.00000968, -0.00000440, -0.00000508, -0.00000571, -0.00000046, 0.00000624, -0.00000222, -0.00000175, -0.00000127, 0.00000439, -0.00000430, 0.00000108, -0.00000181, 0.00000177, -0.00000374, -0.00000868, 0.00000417, 0.00000502, 0.00000122, -0.00000940, 0.00000148, -0.00000402, 0.00000169, -0.00000184, -0.00000130, 0.00000373, -0.00000299, -0.00000076, -0.00000241, 0.00000560, 0.00000064, -0.00000704, -0.00000400, -0.00000865, -0.00000700, -0.00000102, -0.00001165, -0.00006782, 0.00001064, 0.00009083, 0.00004655, 0.00009891, -0.00001241, -0.00006619, -0.00000306, -0.00001856, 0.00000128, 0.00000758, -0.00000220, -0.00000587, -0.00000700, -0.00000917, -0.00000144, 0.00000191, -0.00000089, -0.00000127, -0.00000182, -0.00000235, 0.00000191, 0.00000201, 0.00000100, 0.00000004, -0.00000023, -0.00000496, 0.00000381, -0.00000098, 0.00000255, 0.00001523, -0.00000512, -0.00000700, -0.00000260, 0.00000527, -0.00000523, -0.00000970, 0.00000328, 0.00000219, -0.00000713, 0.00001221, 0.00000038, -0.00001345, -0.00000005, -0.00000095, },
		{ 0.00000298, -0.00000990, -0.00000949, 0.00001382, -0.00000109, -0.00000488, 0.00000726, -0.00001127, -0.00000292, 0.00000334, -0.00000390, -0.00000308, -0.00000892, 0.00001564, 0.00000300, -0.00000085, 0.00000350, -0.00000315, -0.00000145, 0.00000044, 0.00000081, 0.00000227, 0.00000179, -0.00000193, -0.00000185, -0.00000092, -0.00000161, 0.00000123, 0.00000071, -0.00000771, -0.00000776, -0.00000611, -0.00000331, 0.00000557, 0.00000544, -0.00001570, -0.00000566, -0.00005080, -0.00001773, 0.00007872, 0.00003949, 0.00006245, 0.00004685, -0.00004828, -0.00003224, 0.00000403, -0.00000738, -0.00000668, -0.00000457, -0.00000562, -0.00000367, 0.00000621, 0.00000011, -0.00000206, -0.00000232, 0.00000327, -0.00000068, -0.00000253, 0.00000135, -0.00000213, 0.00000185, -0.00001061, 0.00000062, 0.00000473, 0.00000488, -0.00000786, -0.00000295, -0.00000029, -0.00000224, 0.00000203, -0.00000467, 0.00000445, -0.00000150, -0.00000126, -0.00000259, 0.00000292, 0.00000435, -0.00000621, -0.00000465, -0.00000645, -0.00000849, -0.00000324, -0.00000153, -0.00005206, -0.00003284, 0.00008383, 0.00005070, 0.00007915, 0.00004839, -0.00007186, -0.00001853, -0.00001154, -0.00000942, 0.00000993, 0.00000055, -0.00000502, -0.00000648, -0.00000896, -0.00000491, 0.00000226, -0.00000022, -0.00000133, -0.00000129, -0.00000266, 0.00000018, 0.00000259, 0.00000128, 0.00000033, 0.00000028, -0.00000360, -0.00000050, 0.00000370, -0.00000349, 0.00001478, 0.00000293, -0.00000888, -0.00000376, 0.00000184, 0.00000170, -0.00001153, -0.00000133, 0.00000537, -0.00000685, 0.00000526, 0.00000979, -0.00001468, -0.00000271, -0.00000104, },
		{ 0.00000059, -0.00000289, -0.00001513, 0.00000925, 0.00000626, -0.00000785, 0.00000596, -0.00000348, -0.00001077, 0.00000509, -0.00000277, -0.00000159, -0.00001027, 0.00000701, 0.00001214, -0.00000395, 0.00000462, -0.00000178, -0.00000241, -0.00000020, 0.00000091, 0.00000125, 0.00000290, -0.00000090, -0.00000214, -0.00000132, -0.00000114, -0.00000052, 0.00000264, -0.00000557, -0.00000827, -0.00000693, -0.00000429, 0.00000095, 0.00001020, -0.00001150, -0.00000817, -0.00002831, -0.00005554, 0.00005923, 0.00006173, 0.00004641, 0.00007292, -0.00002550, -0.00005001, -0.00000106, -0.00000295, -0.00000853, -0.00000501, -0.00000502, -0.00000577, 0.00000389, 0.00000331, -0.00000264, -0.00000175, 0.00000002, 0.00000342, -0.00000486, 0.00000248, -0.00000304, 0.00000224, -0.00000655, -0.00000555, 0.00000500, 0.00000521, -0.00000208, -0.00000856, 0.00000221, -0.00000393, 0.00000279, -0.00000397, 0.00000133, 0.00000192, -0.00000244, -0.00000166, -0.00000027, 0.00000635, -0.00000360, -0.00000582, -0.00000484, -0.00000817, -0.00000622, 0.00000219, -0.00003145, -0.00005712, 0.00005321, 0.00006858, 0.00005140, 0.00008441, -0.00004102, -0.00004649, -0.00000360, -0.00001728, 0.00000754, 0.00000425, -0.00000343, -0.00000643, -0.00000774, -0.00000789, 0.00000113, 0.00000089, -0.00000130, -0.00000117, -0.00000209, -0.00000165, 0.00000249, 0.00000168, 0.00000084, 0.00000013, -0.00000147, -0.00000371, 0.00000486, -0.00000308, 0.00000736, 0.00001234, -0.00000891, -0.00000431, -0.00000203, 0.00000470, -0.00000755, -0.00000784, 0.00000637, -0.00000357, -0.00000203, 0.00001417, -0.00000934, -0.00000883, 0.00000083, },
		},
		{ { 0.00000270, 0.00000088, 0.00000008, -0.00000272, -0.00000695, 0.00000128, 0.00000428, -0.00000288, 0.00001263, -0.00000155, -0.00000245, 0.00000123, -0.00000166, 0.00000170, -0.00001524, -0.00000044, -0.00000367, -0.00000480, 0.00000897, 0.00000177, -0.00001453, 0.00000116, -0.00000269, -0.00000388, -0.00000207, -0.00000477, 0.00001202, -0.00000642, -0.00000500, -0.00000119, 0.00000196, -0.00000271, -0.00000844, -0.00001129, -0.00006457, 0.00001059, 0.00008323, 0.00003715, 0.00008514, 0.00001352, -0.00005711, -0.00001027, -0.00001431, -0.00000587, -0.00000274, -0.00000413, -0.00000594, -0.00000094, 0.00000033, 0.00000035, -0.00000516, 0.00000015, 0.00000955, -0.00000220, -0.00000372, -0.00000137, -0.00000527, 0.00000383, -0.00000384, 0.00000039, 0.00000351, -0.00000101, 0.00000683, -0.00000285, -0.00000364, -0.00000285, 0.00000683, -0.00000101, 0.00000351, 0.00000039, -0.00000384, 0.00000383, -0.00000527, -0.00000137, -0.00000372, -0.00000220, 0.00000955, 0.00000015, -0.00000516, 0.00000035, 0.00000033, -0.00000094, -0.00000594, -0.00000413, -0.00000274, -0.00000587, -0.00001431, -0.00001027, -0.00005711, 0.00001352, 0.00008514, 0.00003715, 0.00008323, 0.00001059, -0.00006457, -0.00001129, -0.00000844, -0.00000271, 0.00000196, -0.00000119, -0.00000500, -0.00000642, 0.00001202, -0.00000477, -0.00000207, -0.00000388, -0.00000269, 0.00000116, -0.00001453, 0.00000177, 0.00000897, -0.00000480, -0.00000367, -0.00000044, -0.00001524, 0.00000170, -0.00000166, 0.00000123, -0.00000245, -0.00000155, 0.00001263, -0.00000288, 0.00000428, 0.00000128, -0.00000695, -0.00000272, 0.00000008, 0.00000088, },
		{ 0.00000214, 0.00000195, -0.00000005, -0.00000079, -0.00000556, -0.00000402, 0.00000604, -0.00000255, 0.00000702, 0.00000733, -0.00000717, 0.00000369, -0.00000331, 0.00000389, -0.00001095, -0.00000780, 0.00000125, -0.00000794, 0.00000397, 0.00000865, -0.00001021, -0.00000797, 0.00000315, -0.00000672, 0.00000039, -0.00000778, 0.00000694, 0.00000392, -0.00000973, -0.00000068, -0.00000096, 0.00000274, -0.00001022, -0.00000330, -0.00004415, -0.00004669, 0.00007756, 0.00005416, 0.00006535, 0.00006586, -0.00005425, -0.00002688, -0.00000916, -0.00001261, -0.00000181, -0.00000415, -0.00000517, -0.00000400, 0.00000106, 0.00000031, -0.00000211, -0.00000522, 0.00000794, 0.00000408, -0.00000544, -0.00000077, -0.00000529, 0.00000037, 0.00000128, -0.00000506, 0.00000584, -0.00000148, 0.00000538, 0.00000100, -0.00000403, -0.00000414, 0.00000391, 0.00000348, -0.00000134, 0.00000517, -0.00000533, 0.00000268, -0.00000179, -0.00000389, -0.00000100, -0.00000551, 0.00000552, 0.00000674, -0.00000550, -0.00000125, 0.00000013, 0.00000030, -0.00000442, -0.00000504, -0.00000358, -0.00000221, -0.00001317, -0.00000780, -0.00003857, -0.00003247, 0.00007628, 0.00004889, 0.00005754, 0.00006214, -0.00005007, -0.00003608, -0.00000280, -0.00000916, 0.00000343, -0.00000163, -0.00000084, -0.00001005, 0.00000618, 0.00000527, -0.00000733, 0.00000090, -0.00000795, 0.00000422, -0.00000978, -0.00000803, 0.00000946, 0.00000160, -0.00000748, 0.00000207, -0.00001207, -0.00000608, 0.00000303, -0.00000331, 0.00000260, -0.00000687, 0.00001135, 0.00000269, -0.00000070, 0.00000581, -0.00000527, -0.00000535, -0.00000058, 0.00000056, },
		{ 0.00000126, 0.00000263, 0.00000043, -0.00000040, -0.00000276, -0.00000676, 0.00000300, 0.00000181, -0.00000006, 0.00001263, -0.00000556, 0.00000028, 0.00000009, 0.00000067, -0.00000272, -0.00001427, 0.00000212, -0.00000607, -0.00000248, 0.00000967, -0.00000058, -0.00001433, 0.00000192, -0.00000333, -0.00000236, -0.00000416, -0.00000258, 0.00001077, -0.00000725, -0.00000469, -0.00000058, 0.00000187, -0.00000385, -0.00000880, -0.00001546, -0.00007113, 0.00002718, 0.00008575, 0.00004423, 0.00009402, -0.00001682, -0.00005319, -0.00000482, -0.00001697, -0.00000371, -0.00000394, -0.00000387, -0.00000620, 0.00000010, 0.00000061, 0.00000006, -0.00000592, 0.00000162, 0.00000914, -0.00000330, -0.00000284, -0.00000277, -0.00000432, 0.00000450, -0.00000556, 0.00000232, 0.00000246, 0.00000125, 0.00000516, -0.00000402, -0.00000372, -0.00000084, 0.00000734, -0.00000303, 0.00000461, -0.00000150, -0.00000171, 0.00000221, -0.00000522, -0.00000039, -0.00000459, -0.00000091, 0.00000967, -0.00000124, -0.00000420, 0.00000042, -0.00000003, -0.00000193, -0.00000561, -0.00000419, -0.00000174, -0.00000808, -0.00001091, -0.00001964, -0.00005120, 0.00003768, 0.00007148, 0.00003473, 0.00007793, -0.00000313, -0.00005719, -0.00000826, -0.00000801, -0.00000156, 0.00000168, -0.00000155, -0.00000546, -0.00000515, 0.00001277, -0.00000672, 0.00000008, -0.00000589, -0.00000143, -0.00000009, -0.00001421, 0.00000398, 0.00000757, -0.00000639, -0.00000123, -0.00000438, -0.00001375, 0.00000467, -0.00000411, 0.00000265, -0.00000519, 0.00000362, 0.00001024, -0.00000391, 0.00000616, -0.00000062, -0.00000696, -0.00000262, 0.00000067, },
		{ 0.00000115, 0.00000219, 0.00000169, -0.00000068, -0.00000078, -0.00000585, -0.00000240, 0.00000515, -0.00000285, 0.00001049, 0.00000239, -0.00000594, 0.00000396, -0.00000221, 0.00000257, -0.00001408, -0.00000333, -0.00000077, -0.00000732, 0.00000606, 0.00000723, -0.00001180, -0.00000641, 0.00000226, -0.00000532, -0.00000067, -0.00000764, 0.00000813, 0.00000180, -0.00000930, -0.00000049, -0.00000029, 0.00000168, -0.00001095, -0.00000449, -0.00005293, -0.00004020, 0.00009223, 0.00005137, 0.00008457, 0.00004353, -0.00006809, -0.00001392, -0.00001336, -0.00001016, -0.00000272, -0.00000390, -0.00000569, -0.00000314, 0.00000153, 0.00000049, -0.00000316, -0.00000455, 0.00000883, 0.00000262, -0.00000514, -0.00000105, -0.00000607, 0.00000249, -0.00000075, -0.00000396, 0.00000594, -0.00000083, 0.00000614, -0.00000162, -0.00000372, -0.00000378, 0.00000667, 0.00000064, -0.00000032, 0.00000382, -0.00000461, 0.00000305, -0.00000343, -0.00000228, -0.00000164, -0.00000531, 0.00000691, 0.00000525, -0.00000533, -0.00000071, 0.00000004, -0.00000077, -0.00000437, -0.00000524, -0.00000232, -0.00000392, -0.00001142, -0.00001091, -0.00004472, -0.00000824, 0.00007688, 0.00003664, 0.00006015, 0.00004676, -0.00005064, -0.00002909, -0.00000270, -0.00000784, 0.00000367, -0.00000213, -0.00000111, -0.00001011, 0.00000841, 0.00000325, -0.00000633, 0.00000072, -0.00000875, 0.00000520, -0.00001164, -0.00000535, 0.00000943, -0.00000063, -0.00000618, 0.00000135, -0.00001514, -0.00000043, 0.00000019, -0.00000178, 0.00000028, -0.00000448, 0.00001332, -0.00000129, 0.00000197, 0.00000462, -0.00000609, -0.00000523, -0.00000015, },
		{ 0.00000134, 0.00000153, 0.00000246, -0.00000008, -0.00000069, -0.00000284, -0.00000624, 0.00000409, -0.00000045, 0.00000381, 0.00001034, -0.00000788, 0.00000250, -0.00000023, 0.00000181, -0.00000732, -0.00001147, 0.00000306, -0.00000795, 0.00000037, 0.00000944, -0.00000263, -0.00001399, 0.00000254, -0.00000359, -0.00000127, -0.00000618, -0.00000012, 0.00000894, -0.00000748, -0.00000478, 0.00000057, 0.00000101, -0.00000444, -0.00000972, -0.00002031, -0.00007682, 0.00004670, 0.00008510, 0.00005762, 0.00009317, -0.00004802, -0.00004070, -0.00000485, -0.00001717, -0.00000287, -0.00000444, -0.00000398, -0.00000601, 0.00000088, 0.00000100, -0.00000062, -0.00000630, 0.00000299, 0.00000857, -0.00000424, -0.00000210, -0.00000432, -0.00000255, 0.00000408, -0.00000652, 0.00000398, 0.00000174, 0.00000327, 0.00000278, -0.00000443, -0.00000392, 0.00000195, 0.00000638, -0.00000425, 0.00000534, -0.00000292, 0.00000030, 0.00000006, -0.00000427, 0.00000004, -0.00000539, 0.00000061, 0.00000936, -0.00000229, -0.00000333, 0.00000048, -0.00000071, -0.00000251, -0.00000556, -0.00000352, -0.00000187, -0.00000883, -0.00000928, -0.00002946, -0.00003738, 0.00005326, 0.00005693, 0.00003517, 0.00007048, -0.00001404, -0.00004930, -0.00000654, -0.00000703, -0.00000090, 0.00000149, -0.00000200, -0.00000579, -0.00000370, 0.00001314, -0.00000844, 0.00000223, -0.00000835, 0.00000056, -0.00000207, -0.00001300, 0.00000556, 0.00000589, -0.00000724, 0.00000078, -0.00000918, -0.00000967, 0.00000514, -0.00000541, 0.00000338, -0.00000699, 0.00000888, 0.00000609, -0.00000300, 0.00000694, -0.00000232, -0.00000696, -0.00000238, },
		{ 0.00000045, 0.00000169, 0.00000206, 0.00000143, -0.00000131, -0.00000058, -0.00000617, -0.00000063, 0.00000337, -0.00000136, 0.00001216, -0.00000236, -0.00000392, 0.00000410, -0.00000093, -0.00000035, -0.00001514, 0.00000057, -0.00000357, -0.00000545, 0.00000732, 0.00000569, -0.00001311, -0.00000490, 0.00000147, -0.00000378, -0.00000236, -0.00000670, 0.00000854, 0.00000023, -0.00000921, 0.00000023, -0.00000023, 0.00000090, -0.00001196, -0.00000601, -0.00006262, -0.00002976, 0.00010472, 0.00005169, 0.00010271, 0.00001025, -0.00007016, -0.00000427, -0.00001763, -0.00000731, -0.00000403, -0.00000335, -0.00000623, -0.00000216, 0.00000195, 0.00000045, -0.00000414, -0.00000372, 0.00000956, 0.00000109, -0.00000457, -0.00000189, -0.00000597, 0.00000404, -0.00000289, -0.00000233, 0.00000569, 0.00000035, 0.00000576, -0.00000364, -0.00000343, -0.00000267, 0.00000867, -0.00000261, 0.00000142, 0.00000191, -0.00000299, 0.00000221, -0.00000412, -0.00000084, -0.00000253, -0.00000479, 0.00000812, 0.00000362, -0.00000482, -0.00000041, -0.00000012, -0.00000189, -0.00000413, -0.00000532, -0.00000109, -0.00000600, -0.00000847, -0.00001802, -0.00004313, 0.00001342, 0.00007115, 0.00002834, 0.00006153, 0.00003179, -0.00004847, -0.00002382, -0.00000225, -0.00000681, 0.00000387, -0.00000273, -0.00000123, -0.00001008, 0.00001070, 0.00000078, -0.00000463, -0.00000040, -0.00000868, 0.00000560, -0.00001307, -0.00000262, 0.00000877, -0.00000256, -0.00000426, -0.00000115, -0.00001597, 0.00000452, -0.00000362, 0.00000090, -0.00000310, -0.00000001, 0.00001261, -0.00000378, 0.00000454, 0.00000289, -0.00000659, -0.00000512, },
		{ -0.00000195, 0.00000207, 0.00000159, 0.00000228, -0.00000069, -0.00000071, -0.00000306, -0.00000533, 0.00000426, -0.00000171, 0.00000766, 0.00000643, -0.00000861, 0.00000429, 0.00000015, 0.00000133, -0.00001108, -0.00000758, 0.00000238, -0.00000878, 0.00000319, 0.00000849, -0.00000428, -0.00001367, 0.00000317, -0.00000358, -0.00000068, -0.00000793, 0.00000237, 0.00000677, -0.00000726, -0.00000511, 0.00000206, -0.00000056, -0.00000441, -0.00001147, -0.00002566, -0.00008118, 0.00006862, 0.00008190, 0.00007746, 0.00007825, -0.00007313, -0.00002388, -0.00000939, -0.00001477, -0.00000367, -0.00000390, -0.00000466, -0.00000527, 0.00000133, 0.00000149, -0.00000161, -0.00000632, 0.00000424, 0.00000783, -0.00000496, -0.00000165, -0.00000564, -0.00000037, 0.00000273, -0.00000662, 0.00000524, 0.00000147, 0.00000470, 0.00000010, -0.00000426, -0.00000402, 0.00000523, 0.00000386, -0.00000424, 0.00000528, -0.00000345, 0.00000159, -0.00000202, -0.00000273, -0.00000008, -0.00000602, 0.00000234, 0.00000854, -0.00000286, -0.00000272, 0.00000066, -0.00000182, -0.00000257, -0.00000588, -0.00000205, -0.00000342, -0.00000733, -0.00001138, -0.00003580, -0.00001968, 0.00006071, 0.00004358, 0.00003730, 0.00006114, -0.00002179, -0.00004170, -0.00000567, -0.00000564, -0.00000075, 0.00000147, -0.00000257, -0.00000592, -0.00000209, 0.00001309, -0.00000978, 0.00000412, -0.00001092, 0.00000302, -0.00000465, -0.00001089, 0.00000625, 0.00000427, -0.00000738, 0.00000184, -0.00001388, -0.00000360, 0.00000276, -0.00000460, 0.00000248, -0.00000678, 0.00001282, 0.00000137, -0.00000063, 0.00000653, -0.00000359, -0.00000701, },
		{ -0.00000492, 0.00000119, 0.00000212, 0.00000179, 0.00000116, -0.00000188, -0.00000028, -0.00000639, 0.00000091, 0.00000126, 0.00000158, 0.00001167, -0.00000615, -0.00000164, 0.00000458, -0.00000038, -0.00000394, -0.00001430, 0.00000338, -0.00000659, -0.00000244, 0.00000739, 0.00000456, -0.00001456, -0.00000311, 0.00000056, -0.00000201, -0.00000476, -0.00000476, 0.00000785, -0.00000036, -0.00000993, 0.00000202, -0.00000148, 0.00000123, -0.00001417, -0.00000709, -0.00007375, -0.00001430, 0.00011340, 0.00005824, 0.00011377, -0.00002924, -0.00005995, -0.00000101, -0.00001941, -0.00000577, -0.00000459, -0.00000325, -0.00000627, -0.00000150, 0.00000262, -0.00000008, -0.00000477, -0.00000294, 0.00001027, -0.00000060, -0.00000379, -0.00000319, -0.00000491, 0.00000466, -0.00000467, -0.00000051, 0.00000530, 0.00000173, 0.00000441, -0.00000491, -0.00000331, -0.00000062, 0.00000932, -0.00000556, 0.00000341, -0.00000018, -0.00000087, 0.00000036, -0.00000372, 0.00000013, -0.00000355, -0.00000390, 0.00000894, 0.00000211, -0.00000422, -0.00000014, -0.00000060, -0.00000272, -0.00000407, -0.00000473, -0.00000085, -0.00000696, -0.00000649, -0.00002658, -0.00003438, 0.00002954, 0.00006228, 0.00002325, 0.00006149, 0.00001765, -0.00004389, -0.00002064, -0.00000083, -0.00000666, 0.00000453, -0.00000380, -0.00000089, -0.00001022, 0.00001320, -0.00000224, -0.00000215, -0.00000262, -0.00000740, 0.00000497, -0.00001352, -0.00000044, 0.00000790, -0.00000420, -0.00000198, -0.00000534, -0.00001378, 0.00000719, -0.00000682, 0.00000363, -0.00000661, 0.00000576, 0.00000933, -0.00000415, 0.00000622, 0.00000113, -0.00000697, },
		{ -0.00000710, -0.00000134, 0.00000283, 0.00000138, 0.00000219, -0.00000143, -0.00000042, -0.00000344, -0.00000410, 0.00000347, -0.00000147, 0.00001045, 0.00000186, -0.00000817, 0.00000601, 0.00000059, -0.00000053, -0.00001345, -0.00000331, 0.00000021, -0.00000807, 0.00000520, 0.00000744, -0.00000590, -0.00001311, 0.00000365, -0.00000314, -0.00000090, -0.00000893, 0.00000435, 0.00000486, -0.00000724, -0.00000500, 0.00000313, -0.00000205, -0.00000450, -0.00001344, -0.00003215, -0.00008294, 0.00009134, 0.00007846, 0.00010112, 0.00004769, -0.00008585, -0.00000844, -0.00001550, -0.00001120, -0.00000531, -0.00000279, -0.00000557, -0.00000431, 0.00000173, 0.00000181, -0.00000264, -0.00000619, 0.00000554, 0.00000679, -0.00000534, -0.00000169, -0.00000635, 0.00000168, 0.00000085, -0.00000595, 0.00000609, 0.00000161, 0.00000534, -0.00000248, -0.00000376, -0.00000375, 0.00000845, 0.00000005, -0.00000282, 0.00000421, -0.00000282, 0.00000167, -0.00000340, -0.00000105, -0.00000065, -0.00000640, 0.00000415, 0.00000730, -0.00000297, -0.00000238, 0.00000088, -0.00000320, -0.00000225, -0.00000624, -0.00000036, -0.00000573, -0.00000412, -0.00001756, -0.00003594, -0.00000252, 0.00006224, 0.00003198, 0.00004057, 0.00004991, -0.00002584, -0.00003556, -0.00000461, -0.00000456, -0.00000056, 0.00000122, -0.00000290, -0.00000616, -0.00000009, 0.00001230, -0.00001037, 0.00000524, -0.00001298, 0.00000537, -0.00000731, -0.00000832, 0.00000625, 0.00000276, -0.00000675, 0.00000134, -0.00001718, 0.00000306, -0.00000182, -0.00000157, -0.00000043, -0.00000394, 0.00001422, -0.00000250, 0.00000226, 0.00000526, -0.00000448, },
		{ -0.00000738, -0.00000454, 0.00000204, 0.00000233, 0.00000149, 0.00000083, -0.00000227, 0.00000001, -0.00000639, 0.00000188, -0.00000046, 0.00000519, 0.00000925, -0.00000869, 0.00000077, 0.00000539, -0.00000108, -0.00000728, -0.00001196, 0.00000470, -0.00000910, 0.00000115, 0.00000635, 0.00000401, -0.00001632, -0.00000089, -0.00000056, -0.00000009, -0.00000773, -0.00000191, 0.00000608, 0.00000005, -0.00001153, 0.00000496, -0.00000430, 0.00000309, -0.00001816, -0.00000720, -0.00008653, 0.00000675, 0.00011740, 0.00007327, 0.00011112, -0.00006713, -0.00004085, -0.00000440, -0.00001769, -0.00000640, -0.00000372, -0.00000406, -0.00000554, -0.00000134, 0.00000363, -0.00000115, -0.00000498, -0.00000228, 0.00001096, -0.00000239, -0.00000290, -0.00000470, -0.00000309, 0.00000426, -0.00000579, 0.00000127, 0.00000493, 0.00000305, 0.00000232, -0.00000538, -0.00000337, 0.00000239, 0.00000810, -0.00000738, 0.00000500, -0.00000187, 0.00000112, -0.00000198, -0.00000237, 0.00000049, -0.00000458, -0.00000260, 0.00000921, 0.00000096, -0.00000378, 0.00000029, -0.00000159, -0.00000301, -0.00000440, -0.00000322, -0.00000218, -0.00000565, -0.00000772, -0.00003302, -0.00002139, 0.00003970, 0.00005221, 0.00002092, 0.00005953, 0.00000524, -0.00003787, -0.00001918, 0.00000153, -0.00000749, 0.00000578, -0.00000544, 0.00000001, -0.00001057, 0.00001585, -0.00000569, 0.00000095, -0.00000579, -0.00000494, 0.00000318, -0.00001277, 0.00000081, 0.00000720, -0.00000558, 0.00000020, -0.00001066, -0.00000853, 0.00000649, -0.00000784, 0.00000497, -0.00000886, 0.00001138, 0.00000441, -0.00000250, 0.00000660, -0.00000029, },
		{ -0.00000516, -0.00000713, -0.00000057, 0.00000357, 0.00000089, 0.00000224, -0.00000231, 0.00000018, -0.00000398, -0.00000275, 0.00000201, 0.00000034, 0.00001150, -0.00000253, -0.00000695, 0.00000804, 0.00000030, -0.00000309, -0.00001432, 0.00000073, -0.00000310, -0.00000558, 0.00000572, 0.00000702, -0.00000807, -0.00001179, 0.00000354, -0.00000194, -0.00000236, -0.00000860, 0.00000513, 0.00000397, -0.00000824, -0.00000351, 0.00000263, -0.00000216, -0.00000610, -0.00001433, -0.00004109, -0.00008010, 0.00011239, 0.00007865, 0.00012269, 0.00000431, -0.00008291, 0.00000062, -0.00001972, -0.00000860, -0.00000633, -0.00000209, -0.00000600, -0.00000370, 0.00000250, 0.00000155, -0.00000331, -0.00000622, 0.00000711, 0.00000525, -0.00000523, -0.00000234, -0.00000615, 0.00000310, -0.00000110, -0.00000475, 0.00000660, 0.00000207, 0.00000514, -0.00000463, -0.00000320, -0.00000274, 0.00001089, -0.00000434, -0.00000027, 0.00000222, -0.00000111, 0.00000039, -0.00000363, 0.00000032, -0.00000149, -0.00000643, 0.00000583, 0.00000587, -0.00000280, -0.00000214, 0.00000091, -0.00000451, -0.00000192, -0.00000607, 0.00000059, -0.00000739, -0.00000105, -0.00002609, -0.00002951, 0.00001102, 0.00006026, 0.00002208, 0.00004457, 0.00003701, -0.00002593, -0.00003186, -0.00000233, -0.00000471, 0.00000042, 0.00000012, -0.00000243, -0.00000699, 0.00000271, 0.00001038, -0.00000980, 0.00000511, -0.00001388, 0.00000688, -0.00000931, -0.00000599, 0.00000604, 0.00000120, -0.00000531, -0.00000117, -0.00001776, 0.00000839, -0.00000698, 0.00000277, -0.00000484, 0.00000136, 0.00001252, -0.00000438, 0.00000464, 0.00000367, },
		{ -0.00000126, -0.00000786, -0.00000397, 0.00000301, 0.00000221, 0.00000131, 0.00000032, -0.00000234, 0.00000015, -0.00000612, 0.00000209, -0.00000113, 0.00000850, 0.00000557, -0.00001008, 0.00000354, 0.00000601, -0.00000293, -0.00000982, -0.00000858, 0.00000424, -0.00001025, 0.00000438, 0.00000487, 0.00000366, -0.00001809, 0.00000153, -0.00000163, 0.00000154, -0.00001072, 0.00000131, 0.00000381, 0.00000090, -0.00001338, 0.00000837, -0.00000802, 0.00000593, -0.00002356, -0.00000677, -0.00010013, 0.00003268, 0.00011773, 0.00009644, 0.00008996, -0.00009464, -0.00001920, -0.00001189, -0.00001347, -0.00000870, -0.00000167, -0.00000558, -0.00000426, -0.00000149, 0.00000473, -0.00000249, -0.00000497, -0.00000153, 0.00001143, -0.00000408, -0.00000216, -0.00000602, -0.00000092, 0.00000302, -0.00000613, 0.00000283, 0.00000466, 0.00000407, -0.00000019, -0.00000517, -0.00000346, 0.00000606, 0.00000489, -0.00000746, 0.00000553, -0.00000257, 0.00000225, -0.00000410, -0.00000048, 0.00000028, -0.00000549, -0.00000095, 0.00000889, 0.00000029, -0.00000363, 0.00000092, -0.00000309, -0.00000274, -0.00000496, -0.00000114, -0.00000478, -0.00000203, -0.00001324, -0.00003424, -0.00000809, 0.00004535, 0.00004153, 0.00002142, 0.00005486, -0.00000410, -0.00003206, -0.00001825, 0.00000405, -0.00000872, 0.00000718, -0.00000722, 0.00000111, -0.00001082, 0.00001824, -0.00000912, 0.00000415, -0.00000932, -0.00000179, 0.00000062, -0.00001110, 0.00000120, 0.00000671, -0.00000653, 0.00000160, -0.00001589, -0.00000115, 0.00000237, -0.00000581, 0.00000380, -0.00000853, 0.00001517, -0.00000064, 0.00000040, 0.00000581, },
		{ 0.00000232, -0.00000588, -0.00000697, 0.00000032, 0.00000418, 0.00000022, 0.00000240, -0.00000324, 0.00000104, -0.00000461, -0.00000156, 0.00000042, 0.00000328, 0.00001064, -0.00000621, -0.00000512, 0.00001033, -0.00000123, -0.00000557, -0.00001381, 0.00000393, -0.00000685, -0.00000160, 0.00000452, 0.00000770, -0.00001122, -0.00000931, 0.00000251, 0.00000023, -0.00000530, -0.00000656, 0.00000426, 0.00000463, -0.00001084, 0.00000004, -0.00000036, 0.00000022, -0.00001055, -0.00001273, -0.00005388, -0.00007063, 0.00012933, 0.00008664, 0.00013403, -0.00004450, -0.00006562, 0.00000098, -0.00001973, -0.00000864, -0.00000547, -0.00000269, -0.00000535, -0.00000392, 0.00000398, 0.00000046, -0.00000338, -0.00000660, 0.00000909, 0.00000310, -0.00000458, -0.00000355, -0.00000498, 0.00000358, -0.00000267, -0.00000327, 0.00000685, 0.00000272, 0.00000409, -0.00000609, -0.00000279, -0.00000068, 0.00001171, -0.00000833, 0.00000274, -0.00000019, 0.00000123, -0.00000201, -0.00000259, 0.00000108, -0.00000246, -0.00000604, 0.00000714, 0.00000453, -0.00000264, -0.00000177, 0.00000047, -0.00000540, -0.00000188, -0.00000496, -0.00000005, -0.00000693, -0.00000053, -0.00003375, -0.00001861, 0.00002008, 0.00005626, 0.00001411, 0.00004846, 0.00002332, -0.00002263, -0.00003078, 0.00000166, -0.00000657, 0.00000268, -0.00000221, -0.00000077, -0.00000872, 0.00000650, 0.00000714, -0.00000787, 0.00000349, -0.00001327, 0.00000708, -0.00001014, -0.00000452, 0.00000614, -0.00000061, -0.00000315, -0.00000570, -0.00001477, 0.00001058, -0.00001069, 0.00000673, -0.00000933, 0.00000799, 0.00000812, -0.00000383, 0.00000576, },
		{ 0.00000437, -0.00000192, -0.00000834, -0.00000323, 0.00000408, 0.00000166, 0.00000140, -0.00000048, -0.00000197, 0.00000002, -0.00000561, 0.00000157, -0.00000043, 0.00001067, 0.00000139, -0.00001054, 0.00000692, 0.00000566, -0.00000533, -0.00001137, -0.00000461, 0.00000198, -0.00000933, 0.00000614, 0.00000397, 0.00000270, -0.00001913, 0.00000349, -0.00000204, 0.00000214, -0.00001282, 0.00000391, 0.00000205, 0.00000108, -0.00001425, 0.00001080, -0.00001111, 0.00000813, -0.00002879, -0.00000742, -0.00011240, 0.00006119, 0.00011780, 0.00012342, 0.00004978, -0.00010502, -0.00000200, -0.00001926, -0.00000918, -0.00001101, 0.00000042, -0.00000696, -0.00000314, -0.00000136, 0.00000539, -0.00000360, -0.00000517, -0.00000033, 0.00001129, -0.00000535, -0.00000187, -0.00000672, 0.00000104, 0.00000135, -0.00000578, 0.00000409, 0.00000454, 0.00000462, -0.00000279, -0.00000449, -0.00000325, 0.00000978, 0.00000005, -0.00000558, 0.00000466, -0.00000190, 0.00000195, -0.00000526, 0.00000140, -0.00000036, -0.00000620, 0.00000089, 0.00000810, 0.00000003, -0.00000369, 0.00000159, -0.00000485, -0.00000218, -0.00000524, 0.00000062, -0.00000744, 0.00000251, -0.00002218, -0.00002903, 0.00000234, 0.00004849, 0.00003023, 0.00002486, 0.00004682, -0.00000915, -0.00002819, -0.00001631, 0.00000546, -0.00000929, 0.00000784, -0.00000833, 0.00000170, -0.00001031, 0.00001972, -0.00001185, 0.00000670, -0.00001240, 0.00000123, -0.00000192, -0.00000922, 0.00000124, 0.00000613, -0.00000670, 0.00000136, -0.00001944, 0.00000651, -0.00000398, -0.00000102, -0.00000010, -0.00000498, 0.00001585, -0.00000425, 0.00000340, },
		{ 0.00000540, 0.00000151, -0.00000677, -0.00000656, 0.00000135, 0.00000450, -0.00000044, 0.00000252, -0.00000406, 0.00000199, -0.00000524, -0.00000076, -0.00000068, 0.00000657, 0.00000820, -0.00000906, -0.00000249, 0.00001239, -0.00000398, -0.00000747, -0.00001212, 0.00000573, -0.00001002, 0.00000293, 0.00000209, 0.00000929, -0.00001525, -0.00000572, 0.00000065, 0.00000316, -0.00000945, -0.00000302, 0.00000188, 0.00000667, -0.00001486, 0.00000549, -0.00000582, 0.00000531, -0.00001831, -0.00000811, -0.00007101, -0.00005356, 0.00014124, 0.00010485, 0.00012713, -0.00008835, -0.00003978, -0.00000610, -0.00001548, -0.00001157, -0.00000248, -0.00000478, -0.00000353, -0.00000500, 0.00000610, -0.00000138, -0.00000291, -0.00000722, 0.00001134, 0.00000046, -0.00000354, -0.00000510, -0.00000303, 0.00000303, -0.00000358, -0.00000175, 0.00000691, 0.00000346, 0.00000229, -0.00000672, -0.00000262, 0.00000248, 0.00001029, -0.00001082, 0.00000528, -0.00000227, 0.00000343, -0.00000489, -0.00000051, 0.00000111, -0.00000342, -0.00000519, 0.00000793, 0.00000352, -0.00000270, -0.00000111, -0.00000059, -0.00000572, -0.00000216, -0.00000297, -0.00000243, -0.00000367, -0.00000436, -0.00003710, -0.00000681, 0.00002580, 0.00005043, 0.00000891, 0.00005074, 0.00001053, -0.00001751, -0.00003144, 0.00000683, -0.00000984, 0.00000601, -0.00000558, 0.00000192, -0.00001120, 0.00001109, 0.00000279, -0.00000482, 0.00000065, -0.00001129, 0.00000604, -0.00000976, -0.00000407, 0.00000676, -0.00000259, -0.00000079, -0.00001154, -0.00000833, 0.00000861, -0.00001120, 0.00000851, -0.00001212, 0.00001411, 0.00000231, -0.00000123, },
		{ 0.00000538, 0.00000296, -0.00000258, -0.00000866, -0.00000237, 0.00000518, 0.00000071, 0.00000183, -0.00000160, -0.00000109, -0.00000044, -0.00000499, 0.00000063, 0.00000155, 0.00001124, -0.00000270, -0.00001015, 0.00001091, 0.00000375, -0.00000756, -0.00001198, -0.00000056, -0.00000175, -0.00000612, 0.00000568, 0.00000461, 0.00000025, -0.00001860, 0.00000421, -0.00000109, 0.00000091, -0.00001308, 0.00000488, 0.00000188, -0.00000056, -0.00001277, 0.00001062, -0.00001165, 0.00000755, -0.00003162, -0.00001155, -0.00012036, 0.00008903, 0.00012272, 0.00014601, -0.00000425, -0.00009608, 0.00000566, -0.00002261, -0.00000750, -0.00001137, 0.00000112, -0.00000715, -0.00000302, -0.00000029, 0.00000503, -0.00000395, -0.00000599, 0.00000171, 0.00001017, -0.00000587, -0.00000227, -0.00000646, 0.00000227, -0.00000027, -0.00000494, 0.00000502, 0.00000460, 0.00000453, -0.00000511, -0.00000365, -0.00000230, 0.00001263, -0.00000553, -0.00000214, 0.00000248, 0.00000009, 0.00000006, -0.00000499, 0.00000275, -0.00000120, -0.00000662, 0.00000270, 0.00000709, -0.00000005, -0.00000374, 0.00000202, -0.00000654, -0.00000165, -0.00000474, 0.00000103, -0.00000849, 0.00000560, -0.00003183, -0.00001863, 0.00000864, 0.00005039, 0.00001861, 0.00003079, 0.00003550, -0.00000946, -0.00002738, -0.00001216, 0.00000465, -0.00000818, 0.00000690, -0.00000794, 0.00000104, -0.00000841, 0.00001965, -0.00001327, 0.00000801, -0.00001431, 0.00000333, -0.00000366, -0.00000795, 0.00000165, 0.00000503, -0.00000577, -0.00000116, -0.00001977, 0.00001214, -0.00001040, 0.00000505, -0.00000581, 0.00000137, 0.00001302, -0.00000533, },
		{ 0.00000231, 0.00000394, 0.00000120, -0.00000780, -0.00000591, 0.00000254, 0.00000435, -0.00000085, 0.00000239, -0.00000454, 0.00000283, -0.00000582, -0.00000045, -0.00000082, 0.00000940, 0.00000475, -0.00001111, 0.00000121, 0.00001337, -0.00000737, -0.00000862, -0.00000942, 0.00000561, -0.00001149, 0.00000665, -0.00000042, 0.00001092, -0.00001937, -0.00000170, -0.00000140, 0.00000603, -0.00001391, 0.00000115, -0.00000108, 0.00000913, -0.00001918, 0.00001158, -0.00001249, 0.00001189, -0.00002835, -0.00000144, -0.00009135, -0.00002987, 0.00014991, 0.00013176, 0.00009724, -0.00011693, -0.00001378, -0.00001636, -0.00000920, -0.00001589, 0.00000167, -0.00000762, -0.00000119, -0.00000636, 0.00000834, -0.00000345, -0.00000237, -0.00000761, 0.00001339, -0.00000227, -0.00000245, -0.00000654, -0.00000077, 0.00000171, -0.00000375, -0.00000040, 0.00000688, 0.00000412, -0.00000008, -0.00000656, -0.00000257, 0.00000647, 0.00000645, -0.00001104, 0.00000648, -0.00000324, 0.00000460, -0.00000737, 0.00000206, 0.00000052, -0.00000427, -0.00000397, 0.00000817, 0.00000291, -0.00000304, -0.00000017, -0.00000222, -0.00000552, -0.00000241, -0.00000081, -0.00000571, 0.00000170, -0.00001267, -0.00003402, 0.00000238, 0.00003025, 0.00004221, 0.00000756, 0.00004973, 0.00000074, -0.00001280, -0.00003207, 0.00001182, -0.00001341, 0.00000951, -0.00000918, 0.00000494, -0.00001380, 0.00001582, -0.00000201, -0.00000132, -0.00000271, -0.00000865, 0.00000438, -0.00000874, -0.00000423, 0.00000760, -0.00000434, 0.00000088, -0.00001724, 0.00000025, 0.00000282, -0.00000782, 0.00000689, -0.00001167, 0.00001783, -0.00000310, },
		{ -0.00000367, 0.00000568, 0.00000215, -0.00000351, -0.00000866, -0.00000141, 0.00000616, -0.00000052, 0.00000252, -0.00000296, 0.00000025, -0.00000127, -0.00000439, -0.00000027, 0.00000435, 0.00001018, -0.00000637, -0.00000874, 0.00001502, 0.00000019, -0.00000905, -0.00001169, 0.00000289, -0.00000607, -0.00000117, 0.00000298, 0.00000709, -0.00000409, -0.00001605, 0.00000328, 0.00000154, -0.00000250, -0.00001102, 0.00000370, 0.00000384, -0.00000463, -0.00000822, 0.00000679, -0.00000832, 0.00000248, -0.00003011, -0.00002127, -0.00012138, 0.00011351, 0.00013731, 0.00015393, -0.00006172, -0.00007138, 0.00000259, -0.00002011, -0.00000986, -0.00000860, -0.00000049, -0.00000549, -0.00000440, 0.00000208, 0.00000339, -0.00000334, -0.00000759, 0.00000470, 0.00000792, -0.00000551, -0.00000340, -0.00000514, 0.00000245, -0.00000141, -0.00000388, 0.00000563, 0.00000487, 0.00000366, -0.00000680, -0.00000294, -0.00000024, 0.00001361, -0.00001057, 0.00000197, -0.00000039, 0.00000286, -0.00000311, -0.00000319, 0.00000325, -0.00000204, -0.00000669, 0.00000426, 0.00000612, -0.00000020, -0.00000356, 0.00000194, -0.00000785, -0.00000133, -0.00000324, -0.00000052, -0.00000665, 0.00000483, -0.00003858, -0.00000629, 0.00001172, 0.00005097, 0.00000787, 0.00003773, 0.00002219, -0.00000585, -0.00002956, -0.00000560, 0.00000127, -0.00000500, 0.00000399, -0.00000570, -0.00000121, -0.00000485, 0.00001777, -0.00001313, 0.00000785, -0.00001476, 0.00000419, -0.00000421, -0.00000777, 0.00000289, 0.00000314, -0.00000379, -0.00000598, -0.00001603, 0.00001381, -0.00001445, 0.00001020, -0.00001144, 0.00000898, 0.00000741, },
		{ -0.00000643, 0.00000542, 0.00000212, 0.00000109, -0.00000881, -0.00000510, 0.00000390, 0.00000362, -0.00000081, 0.00000183, -0.00000446, 0.00000335, -0.00000630, -0.00000055, 0.00000019, 0.00001111, 0.00000085, -0.00001237, 0.00000608, 0.00001248, -0.00001056, -0.00000915, -0.00000595, 0.00000340, -0.00001039, 0.00000822, -0.00000160, 0.00001131, -0.00002238, 0.00000164, -0.00000259, 0.00000769, -0.00001738, 0.00000462, -0.00000324, 0.00001046, -0.00002205, 0.00001627, -0.00001813, 0.00001753, -0.00003819, 0.00000478, -0.00011206, -0.00000245, 0.00016010, 0.00016076, 0.00004564, -0.00012352, 0.00000435, -0.00002445, -0.00000435, -0.00001909, 0.00000510, -0.00000981, 0.00000052, -0.00000706, 0.00000985, -0.00000499, -0.00000242, -0.00000715, 0.00001460, -0.00000454, -0.00000177, -0.00000735, 0.00000119, 0.00000007, -0.00000328, 0.00000066, 0.00000688, 0.00000451, -0.00000269, -0.00000582, -0.00000227, 0.00001053, 0.00000067, -0.00000876, 0.00000586, -0.00000260, 0.00000411, -0.00000860, 0.00000443, -0.00000045, -0.00000496, -0.00000249, 0.00000797, 0.00000261, -0.00000356, 0.00000086, -0.00000417, -0.00000507, -0.00000211, 0.00000045, -0.00000826, 0.00000711, -0.00002358, -0.00002471, 0.00000706, 0.00003501, 0.00003130, 0.00001069, 0.00004426, -0.00000438, -0.00001058, -0.00003082, 0.00001496, -0.00001581, 0.00001192, -0.00001182, 0.00000724, -0.00001558, 0.00001979, -0.00000645, 0.00000184, -0.00000570, -0.00000623, 0.00000300, -0.00000800, -0.00000415, 0.00000801, -0.00000519, 0.00000077, -0.00002090, 0.00000867, -0.00000507, -0.00000132, 0.00000192, -0.00000743, 0.00001784, },
		{ 0.00000070, -0.00000003, 0.00000433, 0.00000214, -0.00000482, -0.00000831, -0.00000032, 0.00000681, -0.00000175, 0.00000326, -0.00000435, 0.00000189, -0.00000240, -0.00000393, -0.00000067, 0.00000727, 0.00000781, -0.00000956, -0.00000591, 0.00001832, -0.00000450, -0.00000966, -0.00001049, 0.00000495, -0.00000967, 0.00000421, -0.00000102, 0.00001085, -0.00000988, -0.00001178, 0.00000098, 0.00000544, -0.00000768, -0.00000697, 0.00000069, 0.00000754, -0.00001067, -0.00000107, -0.00000036, -0.00000120, -0.00000730, -0.00002375, -0.00003721, -0.00011451, 0.00013435, 0.00016337, 0.00013819, -0.00010972, -0.00003929, -0.00000813, -0.00001294, -0.00001561, -0.00000303, -0.00000420, -0.00000221, -0.00000705, 0.00000548, 0.00000073, -0.00000202, -0.00000966, 0.00000834, 0.00000477, -0.00000445, -0.00000502, -0.00000299, 0.00000157, -0.00000178, -0.00000286, 0.00000598, 0.00000535, 0.00000199, -0.00000762, -0.00000250, 0.00000306, 0.00001199, -0.00001374, 0.00000553, -0.00000300, 0.00000549, -0.00000676, -0.00000022, 0.00000283, -0.00000277, -0.00000641, 0.00000542, 0.00000539, -0.00000062, -0.00000301, 0.00000124, -0.00000869, -0.00000106, -0.00000116, -0.00000369, -0.00000178, -0.00000101, -0.00003939, 0.00000414, 0.00001390, 0.00004899, -0.00000010, 0.00004340, 0.00000919, -0.00000033, -0.00003338, 0.00000242, -0.00000397, -0.00000027, -0.00000043, -0.00000192, -0.00000476, 0.00000008, 0.00001433, -0.00001174, 0.00000660, -0.00001412, 0.00000411, -0.00000385, -0.00000851, 0.00000490, 0.00000071, -0.00000138, -0.00001221, -0.00000859, 0.00001068, -0.00001436, 0.00001230, -0.00001481, 0.00001566, },
		{ 0.00001398, -0.00000671, 0.00000696, 0.00000076, 0.00000075, -0.00000957, -0.00000420, 0.00000537, 0.00000233, -0.00000023, 0.00000074, -0.00000368, 0.00000335, -0.00000669, -0.00000080, 0.00000217, 0.00001141, -0.00000310, -0.00001262, 0.00001173, 0.00000935, -0.00001283, -0.00000918, -0.00000221, -0.00000047, -0.00000661, 0.00000688, -0.00000039, 0.00000938, -0.00002318, 0.00000326, -0.00000194, 0.00000705, -0.00001865, 0.00000614, -0.00000334, 0.00000927, -0.00002183, 0.00001756, -0.00002035, 0.00001949, -0.00004487, 0.00000739, -0.00012965, 0.00002512, 0.00017808, 0.00018089, -0.00001943, -0.00010765, 0.00000992, -0.00002631, -0.00000390, -0.00001883, 0.00000599, -0.00000996, 0.00000048, -0.00000620, 0.00000987, -0.00000537, -0.00000360, -0.00000531, 0.00001442, -0.00000585, -0.00000186, -0.00000712, 0.00000230, -0.00000136, -0.00000241, 0.00000136, 0.00000707, 0.00000434, -0.00000511, -0.00000484, -0.00000121, 0.00001360, -0.00000593, -0.00000451, 0.00000354, -0.00000039, 0.00000176, -0.00000807, 0.00000601, -0.00000149, -0.00000547, -0.00000095, 0.00000755, 0.00000240, -0.00000405, 0.00000173, -0.00000614, -0.00000458, -0.00000088, -0.00000013, -0.00000836, 0.00000986, -0.00003363, -0.00001173, 0.00000758, 0.00004022, 0.00001843, 0.00001768, 0.00003441, -0.00000433, -0.00001192, -0.00002649, 0.00001508, -0.00001590, 0.00001221, -0.00001255, 0.00000793, -0.00001577, 0.00002226, -0.00000988, 0.00000409, -0.00000771, -0.00000471, 0.00000262, -0.00000834, -0.00000305, 0.00000737, -0.00000465, -0.00000190, -0.00002084, 0.00001436, -0.00001237, 0.00000624, -0.00000501, -0.00000014, },
		{ 0.00001934, -0.00000506, 0.00000420, 0.00000198, 0.00000271, -0.00000639, -0.00000766, 0.00000097, 0.00000691, -0.00000266, 0.00000374, -0.00000542, 0.00000352, -0.00000377, -0.00000362, -0.00000026, 0.00000963, 0.00000459, -0.00001221, -0.00000141, 0.00001976, -0.00000937, -0.00000960, -0.00000829, 0.00000495, -0.00001119, 0.00000829, -0.00000470, 0.00001453, -0.00001591, -0.00000686, -0.00000167, 0.00000947, -0.00001342, -0.00000212, -0.00000290, 0.00001159, -0.00001707, 0.00000690, -0.00000897, 0.00000783, -0.00002010, -0.00001409, -0.00005782, -0.00010134, 0.00015490, 0.00019728, 0.00009492, -0.00013676, -0.00001019, -0.00002054, -0.00000468, -0.00002220, 0.00000348, -0.00000862, 0.00000154, -0.00001002, 0.00000906, -0.00000214, -0.00000075, -0.00001143, 0.00001189, 0.00000131, -0.00000316, -0.00000660, -0.00000054, -0.00000008, -0.00000134, -0.00000210, 0.00000621, 0.00000589, -0.00000036, -0.00000755, -0.00000221, 0.00000726, 0.00000759, -0.00001413, 0.00000746, -0.00000439, 0.00000697, -0.00000986, 0.00000322, 0.00000170, -0.00000332, -0.00000580, 0.00000616, 0.00000494, -0.00000133, -0.00000211, -0.00000003, -0.00000915, -0.00000039, 0.00000053, -0.00000717, 0.00000472, -0.00001117, -0.00003316, 0.00000982, 0.00001747, 0.00004312, -0.00000340, 0.00004545, -0.00000093, 0.00000450, -0.00003670, 0.00001010, -0.00000955, 0.00000477, -0.00000523, 0.00000240, -0.00000878, 0.00000560, 0.00001004, -0.00000975, 0.00000503, -0.00001320, 0.00000393, -0.00000346, -0.00000939, 0.00000700, -0.00000154, 0.00000031, -0.00001811, 0.00000083, 0.00000354, -0.00000974, 0.00001022, -0.00001429, },
		{ 0.00000823, 0.00000736, -0.00000402, 0.00000643, 0.00000040, -0.00000011, -0.00000990, -0.00000323, 0.00000677, 0.00000065, 0.00000079, -0.00000078, -0.00000220, 0.00000272, -0.00000698, -0.00000084, 0.00000464, 0.00001037, -0.00000694, -0.00001140, 0.00001725, 0.00000433, -0.00001389, -0.00000864, 0.00000102, -0.00000491, -0.00000098, 0.00000293, 0.00000333, 0.00000482, -0.00002134, 0.00000273, 0.00000088, 0.00000373, -0.00001722, 0.00000522, -0.00000090, 0.00000505, -0.00001784, 0.00001447, -0.00001783, 0.00001597, -0.00004622, 0.00000401, -0.00014155, 0.00005066, 0.00020882, 0.00017971, -0.00008419, -0.00007564, 0.00000329, -0.00002102, -0.00000874, -0.00001420, 0.00000353, -0.00000747, -0.00000176, -0.00000345, 0.00000819, -0.00000443, -0.00000599, -0.00000199, 0.00001267, -0.00000601, -0.00000284, -0.00000573, 0.00000222, -0.00000212, -0.00000144, 0.00000168, 0.00000757, 0.00000338, -0.00000689, -0.00000397, 0.00000104, 0.00001453, -0.00001180, 0.00000054, 0.00000028, 0.00000274, -0.00000203, -0.00000575, 0.00000646, -0.00000236, -0.00000577, 0.00000049, 0.00000712, 0.00000205, -0.00000426, 0.00000218, -0.00000794, -0.00000403, 0.00000118, -0.00000279, -0.00000505, 0.00000779, -0.00003915, 0.00000094, 0.00000627, 0.00004442, 0.00000565, 0.00002655, 0.00002180, -0.00000017, -0.00001646, -0.00001917, 0.00001204, -0.00001343, 0.00001013, -0.00001103, 0.00000667, -0.00001406, 0.00002296, -0.00001211, 0.00000539, -0.00000873, -0.00000413, 0.00000331, -0.00000999, -0.00000059, 0.00000545, -0.00000277, -0.00000706, -0.00001632, 0.00001533, -0.00001640, 0.00001220, -0.00001154, },
		{ -0.00000952, 0.00001877, -0.00000811, 0.00000746, -0.00000041, 0.00000339, -0.00000796, -0.00000684, 0.00000250, 0.00000629, -0.00000297, 0.00000369, -0.00000587, 0.00000482, -0.00000523, -0.00000335, 0.00000101, 0.00001098, 0.00000090, -0.00001409, 0.00000461, 0.00001861, -0.00001343, -0.00000917, -0.00000524, 0.00000269, -0.00000972, 0.00000956, -0.00000635, 0.00001645, -0.00002058, -0.00000276, -0.00000327, 0.00001214, -0.00001812, 0.00000196, -0.00000543, 0.00001413, -0.00002168, 0.00001317, -0.00001625, 0.00001572, -0.00003278, -0.00000427, -0.00007986, -0.00008538, 0.00018183, 0.00022913, 0.00002814, -0.00013689, 0.00000730, -0.00002837, 0.00000034, -0.00002628, 0.00000833, -0.00001176, 0.00000414, -0.00001195, 0.00001166, -0.00000423, -0.00000041, -0.00001203, 0.00001446, -0.00000170, -0.00000226, -0.00000752, 0.00000152, -0.00000195, -0.00000026, -0.00000178, 0.00000654, 0.00000622, -0.00000305, -0.00000681, -0.00000169, 0.00001149, 0.00000104, -0.00001155, 0.00000718, -0.00000396, 0.00000655, -0.00001147, 0.00000629, 0.00000023, -0.00000373, -0.00000496, 0.00000657, 0.00000469, -0.00000222, -0.00000101, -0.00000169, -0.00000941, 0.00000113, 0.00000069, -0.00000907, 0.00001028, -0.00002288, -0.00002136, 0.00001009, 0.00002343, 0.00003297, -0.00000118, 0.00004245, -0.00000630, 0.00000650, -0.00003754, 0.00001557, -0.00001381, 0.00000865, -0.00000909, 0.00000608, -0.00001222, 0.00001077, 0.00000566, -0.00000788, 0.00000392, -0.00001287, 0.00000460, -0.00000409, -0.00000929, 0.00000823, -0.00000271, -0.00000001, -0.00002153, 0.00000952, -0.00000532, -0.00000190, 0.00000436, },
		{ -0.00001522, 0.00001518, 0.00000001, 0.00000054, 0.00000410, 0.00000110, -0.00000158, -0.00000978, -0.00000213, 0.00000783, -0.00000108, 0.00000196, -0.00000245, -0.00000025, 0.00000149, -0.00000716, -0.00000037, 0.00000701, 0.00000828, -0.00001061, -0.00000821, 0.00002138, -0.00000164, -0.00001392, -0.00000730, 0.00000276, -0.00000833, 0.00000476, -0.00000225, 0.00000857, -0.00000161, -0.00001742, 0.00000058, 0.00000526, -0.00000166, -0.00001362, 0.00000247, 0.00000332, -0.00000133, -0.00001100, 0.00000777, -0.00001114, 0.00000720, -0.00004211, -0.00000560, -0.00014769, 0.00007563, 0.00025258, 0.00014777, -0.00013315, -0.00003870, -0.00001036, -0.00001120, -0.00001714, -0.00000636, -0.00000148, -0.00000298, -0.00000567, 0.00000067, 0.00000530, -0.00000267, -0.00000904, 0.00000230, 0.00000971, -0.00000529, -0.00000438, -0.00000346, 0.00000100, -0.00000194, -0.00000070, 0.00000173, 0.00000843, 0.00000148, -0.00000773, -0.00000340, 0.00000462, 0.00001251, -0.00001541, 0.00000494, -0.00000277, 0.00000574, -0.00000631, -0.00000213, 0.00000579, -0.00000295, -0.00000587, 0.00000170, 0.00000685, 0.00000140, -0.00000407, 0.00000211, -0.00000953, -0.00000309, 0.00000332, -0.00000672, 0.00000119, 0.00000034, -0.00003770, 0.00000953, 0.00000615, 0.00004536, -0.00000432, 0.00003444, 0.00000910, 0.00000578, -0.00002249, -0.00001020, 0.00000687, -0.00000917, 0.00000632, -0.00000777, 0.00000386, -0.00001080, 0.00002217, -0.00001351, 0.00000627, -0.00000937, -0.00000384, 0.00000439, -0.00001235, 0.00000277, 0.00000273, -0.00000037, -0.00001359, -0.00000805, 0.00001108, -0.00001548, 0.00001428, },
		{ -0.00000338, -0.00000164, 0.00001410, -0.00000767, 0.00000863, -0.00000192, 0.00000365, -0.00000923, -0.00000594, 0.00000421, 0.00000497, -0.00000259, 0.00000298, -0.00000549, 0.00000549, -0.00000666, -0.00000292, 0.00000289, 0.00001117, -0.00000305, -0.00001469, 0.00001140, 0.00001483, -0.00001609, -0.00000848, -0.00000188, -0.00000125, -0.00000534, 0.00000738, -0.00000491, 0.00001545, -0.00002265, -0.00000067, -0.00000273, 0.00001222, -0.00002046, 0.00000397, -0.00000560, 0.00001372, -0.00002279, 0.00001563, -0.00001964, 0.00001947, -0.00004200, 0.00000228, -0.00009982, -0.00007024, 0.00022277, 0.00024449, -0.00004992, -0.00011216, 0.00000943, -0.00002783, -0.00000085, -0.00002534, 0.00000947, -0.00001201, 0.00000427, -0.00001179, 0.00001242, -0.00000481, -0.00000163, -0.00001082, 0.00001538, -0.00000363, -0.00000221, -0.00000731, 0.00000259, -0.00000344, 0.00000117, -0.00000197, 0.00000724, 0.00000595, -0.00000556, -0.00000579, -0.00000035, 0.00001453, -0.00000630, -0.00000670, 0.00000489, -0.00000178, 0.00000408, -0.00001103, 0.00000835, -0.00000119, -0.00000403, -0.00000401, 0.00000683, 0.00000442, -0.00000307, 0.00000004, -0.00000355, -0.00000949, 0.00000366, -0.00000140, -0.00000776, 0.00001210, -0.00003226, -0.00000754, 0.00000673, 0.00003080, 0.00001981, 0.00000573, 0.00003459, -0.00000654, 0.00000478, -0.00003489, 0.00001775, -0.00001564, 0.00001037, -0.00001101, 0.00000820, -0.00001430, 0.00001492, 0.00000169, -0.00000642, 0.00000358, -0.00001353, 0.00000661, -0.00000643, -0.00000741, 0.00000782, -0.00000217, -0.00000316, -0.00002074, 0.00001471, -0.00001279, 0.00000653, },
		{ 0.00001155, -0.00001441, 0.00001852, -0.00000578, 0.00000526, 0.00000091, 0.00000247, -0.00000343, -0.00000932, -0.00000079, 0.00000826, -0.00000245, 0.00000289, -0.00000387, 0.00000180, -0.00000022, -0.00000717, 0.00000076, 0.00000876, 0.00000547, -0.00001392, -0.00000288, 0.00002297, -0.00000733, -0.00001342, -0.00000490, 0.00000220, -0.00000920, 0.00000858, -0.00000660, 0.00001350, -0.00000827, -0.00001289, -0.00000182, 0.00000970, -0.00000760, -0.00000933, -0.00000057, 0.00000754, -0.00000780, -0.00000363, -0.00000005, -0.00000282, -0.00000441, -0.00003479, -0.00001917, -0.00015099, 0.00010613, 0.00030229, 0.00008298, -0.00015413, -0.00000904, -0.00002331, -0.00000189, -0.00002533, 0.00000188, -0.00000694, 0.00000177, -0.00000980, 0.00000493, 0.00000232, -0.00000115, -0.00001169, 0.00000655, 0.00000635, -0.00000430, -0.00000585, -0.00000096, -0.00000096, -0.00000084, -0.00000044, 0.00000174, 0.00000947, -0.00000126, -0.00000755, -0.00000303, 0.00000913, 0.00000742, -0.00001576, 0.00000741, -0.00000448, 0.00000745, -0.00000989, 0.00000193, 0.00000434, -0.00000328, -0.00000577, 0.00000266, 0.00000677, 0.00000042, -0.00000348, 0.00000153, -0.00001097, -0.00000128, 0.00000437, -0.00001019, 0.00000835, -0.00001084, -0.00002924, 0.00001204, 0.00000941, 0.00004128, -0.00000915, 0.00003867, -0.00000094, 0.00001086, -0.00002771, -0.00000161, 0.00000128, -0.00000454, 0.00000207, -0.00000386, 0.00000040, -0.00000678, 0.00002054, -0.00001469, 0.00000755, -0.00001060, -0.00000274, 0.00000467, -0.00001420, 0.00000597, 0.00000027, 0.00000113, -0.00001943, 0.00000181, 0.00000293, -0.00000968, },
		{ 0.00001252, -0.00001021, 0.00000699, 0.00000685, -0.00000416, 0.00000739, -0.00000207, 0.00000317, -0.00001003, -0.00000498, 0.00000588, 0.00000319, -0.00000167, 0.00000173, -0.00000430, 0.00000537, -0.00000794, -0.00000209, 0.00000492, 0.00001033, -0.00000723, -0.00001333, 0.00001765, 0.00000916, -0.00001735, -0.00000730, 0.00000080, -0.00000550, 0.00000070, 0.00000246, -0.00000057, 0.00001142, -0.00002185, -0.00000087, 0.00000017, 0.00000948, -0.00002011, 0.00000365, -0.00000326, 0.00001019, -0.00002007, 0.00001367, -0.00001819, 0.00001763, -0.00004596, 0.00000368, -0.00011615, -0.00005670, 0.00028250, 0.00022850, -0.00012139, -0.00007238, -0.00000123, -0.00001934, -0.00000824, -0.00001906, 0.00000644, -0.00000899, 0.00000166, -0.00000934, 0.00001126, -0.00000390, -0.00000432, -0.00000782, 0.00001452, -0.00000432, -0.00000308, -0.00000588, 0.00000236, -0.00000407, 0.00000256, -0.00000268, 0.00000853, 0.00000468, -0.00000733, -0.00000494, 0.00000231, 0.00001510, -0.00001262, -0.00000102, 0.00000152, 0.00000139, 0.00000010, -0.00000863, 0.00000907, -0.00000231, -0.00000426, -0.00000308, 0.00000711, 0.00000392, -0.00000366, 0.00000085, -0.00000551, -0.00000916, 0.00000675, -0.00000556, -0.00000274, 0.00000843, -0.00003584, 0.00000395, 0.00000303, 0.00003707, 0.00000637, 0.00001494, 0.00002372, -0.00000296, 0.00000007, -0.00002918, 0.00001677, -0.00001501, 0.00000986, -0.00001083, 0.00000856, -0.00001484, 0.00001792, -0.00000192, -0.00000508, 0.00000362, -0.00001479, 0.00000965, -0.00001033, -0.00000370, 0.00000572, -0.00000007, -0.00000893, -0.00001521, 0.00001455, -0.00001609, },
		{ -0.00000090, 0.00000488, -0.00000905, 0.00001721, -0.00000831, 0.00000847, -0.00000199, 0.00000388, -0.00000532, -0.00000869, 0.00000086, 0.00000789, -0.00000315, 0.00000324, -0.00000463, 0.00000351, -0.00000219, -0.00000691, 0.00000248, 0.00000962, 0.00000214, -0.00001640, 0.00000415, 0.00002148, -0.00001176, -0.00001276, -0.00000157, -0.00000079, -0.00000667, 0.00000891, -0.00000822, 0.00001621, -0.00001330, -0.00000944, -0.00000284, 0.00001249, -0.00001231, -0.00000609, -0.00000207, 0.00000969, -0.00001195, 0.00000146, -0.00000580, 0.00000356, -0.00001511, -0.00002798, -0.00003289, -0.00015615, 0.00015182, 0.00034343, -0.00000669, -0.00014276, 0.00000457, -0.00002868, 0.00000189, -0.00002926, 0.00000726, -0.00001029, 0.00000474, -0.00001248, 0.00000792, 0.00000046, -0.00000096, -0.00001284, 0.00000967, 0.00000353, -0.00000381, -0.00000652, 0.00000100, -0.00000302, 0.00000094, -0.00000085, 0.00000203, 0.00001031, -0.00000443, -0.00000659, -0.00000241, 0.00001357, 0.00000007, -0.00001275, 0.00000736, -0.00000423, 0.00000711, -0.00001177, 0.00000548, 0.00000257, -0.00000346, -0.00000551, 0.00000344, 0.00000676, -0.00000075, -0.00000260, 0.00000053, -0.00001231, 0.00000175, 0.00000322, -0.00001118, 0.00001349, -0.00002233, -0.00001635, 0.00000906, 0.00001626, 0.00003201, -0.00000800, 0.00003780, -0.00000653, 0.00001310, -0.00003028, 0.00000480, -0.00000306, -0.00000100, -0.00000128, -0.00000049, -0.00000273, -0.00000279, 0.00001865, -0.00001614, 0.00000980, -0.00001318, 0.00000010, 0.00000295, -0.00001423, 0.00000776, -0.00000080, 0.00000030, -0.00002224, 0.00001021, -0.00000627, },
		{ -0.00001380, 0.00001382, -0.00001348, 0.00001356, -0.00000059, 0.00000094, 0.00000434, -0.00000097, 0.00000195, -0.00001036, -0.00000384, 0.00000718, 0.00000136, -0.00000057, 0.00000029, -0.00000257, 0.00000448, -0.00000898, -0.00000067, 0.00000660, 0.00000874, -0.00001155, -0.00000947, 0.00002188, 0.00000293, -0.00001773, -0.00000512, 0.00000160, -0.00000815, 0.00000625, -0.00000329, 0.00000520, 0.00000554, -0.00001914, -0.00000245, 0.00000444, 0.00000486, -0.00001798, 0.00000204, 0.00000033, 0.00000500, -0.00001510, 0.00000884, -0.00001335, 0.00001137, -0.00004555, 0.00000092, -0.00013079, -0.00004031, 0.00035876, 0.00017150, -0.00016803, -0.00003199, -0.00001697, -0.00000743, -0.00001866, -0.00000969, 0.00000081, -0.00000395, -0.00000268, -0.00000551, 0.00000903, -0.00000235, -0.00000761, -0.00000382, 0.00001246, -0.00000420, -0.00000441, -0.00000366, 0.00000099, -0.00000365, 0.00000360, -0.00000374, 0.00001047, 0.00000217, -0.00000795, -0.00000450, 0.00000644, 0.00001233, -0.00001621, 0.00000378, -0.00000159, 0.00000434, -0.00000427, -0.00000492, 0.00000857, -0.00000304, -0.00000439, -0.00000225, 0.00000753, 0.00000304, -0.00000383, 0.00000130, -0.00000757, -0.00000797, 0.00000934, -0.00001049, 0.00000487, -0.00000053, -0.00003195, 0.00000976, 0.00000233, 0.00003935, -0.00000412, 0.00002326, 0.00001269, 0.00000200, -0.00000569, -0.00002205, 0.00001391, -0.00001291, 0.00000798, -0.00000924, 0.00000766, -0.00001426, 0.00002015, -0.00000571, -0.00000304, 0.00000304, -0.00001555, 0.00001253, -0.00001470, 0.00000101, 0.00000270, 0.00000250, -0.00001587, -0.00000606, 0.00000898, },
		{ -0.00001312, 0.00000769, -0.00000327, -0.00000076, 0.00001176, -0.00000705, 0.00000915, -0.00000364, 0.00000472, -0.00000690, -0.00000800, 0.00000274, 0.00000675, -0.00000309, 0.00000290, -0.00000452, 0.00000453, -0.00000419, -0.00000622, 0.00000449, 0.00000955, -0.00000169, -0.00001728, 0.00001176, 0.00001721, -0.00001454, -0.00001193, 0.00000199, -0.00000537, -0.00000113, 0.00000541, -0.00000619, 0.00001556, -0.00001551, -0.00000821, -0.00000141, 0.00001248, -0.00001456, -0.00000502, -0.00000095, 0.00000852, -0.00001224, 0.00000232, -0.00000712, 0.00000513, -0.00002170, -0.00002491, -0.00004365, -0.00016684, 0.00022244, 0.00035684, -0.00010371, -0.00010463, 0.00000013, -0.00002347, -0.00000243, -0.00002655, 0.00000769, -0.00000986, 0.00000455, -0.00001259, 0.00000880, 0.00000038, -0.00000268, -0.00001186, 0.00001095, 0.00000190, -0.00000431, -0.00000591, 0.00000178, -0.00000451, 0.00000304, -0.00000203, 0.00000302, 0.00001039, -0.00000740, -0.00000535, -0.00000089, 0.00001655, -0.00000791, -0.00000733, 0.00000510, -0.00000214, 0.00000462, -0.00001145, 0.00000786, 0.00000094, -0.00000358, -0.00000516, 0.00000419, 0.00000662, -0.00000192, -0.00000161, -0.00000082, -0.00001338, 0.00000586, -0.00000055, -0.00000829, 0.00001394, -0.00003007, -0.00000327, 0.00000352, 0.00002474, 0.00001937, -0.00000200, 0.00003227, -0.00000762, 0.00001203, -0.00002958, 0.00000824, -0.00000533, 0.00000071, -0.00000298, 0.00000164, -0.00000496, 0.00000076, 0.00001662, -0.00001774, 0.00001293, -0.00001718, 0.00000491, -0.00000124, -0.00001169, 0.00000735, 0.00000019, -0.00000368, -0.00002034, 0.00001428, },
		{ -0.00000009, -0.00000656, 0.00000986, -0.00001171, 0.00001591, -0.00000582, 0.00000569, 0.00000066, 0.00000081, 0.00000030, -0.00001040, -0.00000235, 0.00000776, -0.00000005, 0.00000023, -0.00000083, -0.00000080, 0.00000305, -0.00000972, 0.00000139, 0.00000752, 0.00000661, -0.00001570, -0.00000306, 0.00002292, -0.00000246, -0.00001794, -0.00000157, -0.00000036, -0.00000754, 0.00000899, -0.00000743, 0.00001012, -0.00000015, -0.00001635, -0.00000364, 0.00000821, 0.00000026, -0.00001591, 0.00000114, 0.00000284, 0.00000083, -0.00001089, 0.00000441, -0.00000861, 0.00000423, -0.00004427, -0.00000212, -0.00014932, -0.00001066, 0.00043972, 0.00007401, -0.00017725, -0.00000506, -0.00002823, 0.00000133, -0.00002699, -0.00000116, -0.00000440, 0.00000068, -0.00000675, -0.00000197, 0.00000722, -0.00000156, -0.00001010, -0.00000014, 0.00001028, -0.00000416, -0.00000538, -0.00000146, -0.00000095, -0.00000228, 0.00000399, -0.00000486, 0.00001287, -0.00000150, -0.00000727, -0.00000439, 0.00001155, 0.00000621, -0.00001604, 0.00000631, -0.00000318, 0.00000580, -0.00000773, -0.00000091, 0.00000733, -0.00000351, -0.00000440, -0.00000157, 0.00000813, 0.00000174, -0.00000354, 0.00000134, -0.00000976, -0.00000541, 0.00001013, -0.00001414, 0.00001255, -0.00001235, -0.00002165, 0.00000890, 0.00000635, 0.00003585, -0.00000922, 0.00002800, 0.00000415, 0.00000583, -0.00001028, -0.00001554, 0.00001097, -0.00001085, 0.00000607, -0.00000737, 0.00000643, -0.00001332, 0.00002220, -0.00001028, 0.00000060, 0.00000066, -0.00001443, 0.00001370, -0.00001792, 0.00000522, 0.00000013, 0.00000388, -0.00002162, 0.00000406, },
		{ 0.00001245, -0.00001485, 0.00001284, -0.00000974, 0.00000767, 0.00000413, -0.00000281, 0.00000726, -0.00000368, 0.00000471, -0.00000804, -0.00000719, 0.00000457, 0.00000515, -0.00000252, 0.00000204, -0.00000361, 0.00000472, -0.00000607, -0.00000492, 0.00000632, 0.00000878, -0.00000606, -0.00001569, 0.00001829, 0.00001136, -0.00001606, -0.00001038, 0.00000450, -0.00000977, 0.00000566, -0.00000072, -0.00000117, 0.00001186, -0.00001500, -0.00000911, 0.00000230, 0.00000979, -0.00001446, -0.00000586, 0.00000237, 0.00000451, -0.00000908, -0.00000082, -0.00000394, 0.00000143, -0.00002341, -0.00002618, -0.00005169, -0.00018198, 0.00032274, 0.00032441, -0.00018510, -0.00005402, -0.00001684, -0.00001010, -0.00001346, -0.00001790, 0.00000343, -0.00000581, 0.00000134, -0.00001032, 0.00000783, 0.00000173, -0.00000593, -0.00000904, 0.00001050, 0.00000143, -0.00000569, -0.00000408, 0.00000119, -0.00000501, 0.00000507, -0.00000399, 0.00000506, 0.00000913, -0.00000943, -0.00000437, 0.00000213, 0.00001667, -0.00001445, -0.00000122, 0.00000177, 0.00000086, 0.00000073, -0.00000916, 0.00000884, -0.00000033, -0.00000367, -0.00000482, 0.00000507, 0.00000612, -0.00000283, -0.00000071, -0.00000250, -0.00001375, 0.00001028, -0.00000631, -0.00000155, 0.00000845, -0.00003101, 0.00000563, -0.00000065, 0.00003152, 0.00000665, 0.00000608, 0.00002418, -0.00000580, 0.00000880, -0.00002649, 0.00000927, -0.00000588, 0.00000085, -0.00000319, 0.00000260, -0.00000637, 0.00000406, 0.00001401, -0.00001869, 0.00001604, -0.00002168, 0.00001089, -0.00000734, -0.00000683, 0.00000487, 0.00000290, -0.00001035, -0.00001351, },
		{ 0.00001184, -0.00000931, 0.00000308, 0.00000207, -0.00000527, 0.00001326, -0.00000724, 0.00000836, -0.00000233, 0.00000250, -0.00000135, -0.00001037, -0.00000042, 0.00000743, -0.00000070, 0.00000036, -0.00000117, 0.00000050, 0.00000139, -0.00001007, 0.00000395, 0.00000752, 0.00000400, -0.00001900, 0.00000518, 0.00002046, -0.00000609, -0.00001836, 0.00000321, -0.00000513, -0.00000292, 0.00000738, -0.00000792, 0.00001212, -0.00000362, -0.00001535, -0.00000264, 0.00000958, -0.00000239, -0.00001576, 0.00000290, 0.00000203, 0.00000033, -0.00001057, 0.00000401, -0.00000808, 0.00000065, -0.00004669, -0.00000047, -0.00017877, 0.00004656, 0.00050451, -0.00005076, -0.00014703, -0.00000006, -0.00002771, 0.00000131, -0.00002855, 0.00000265, -0.00000613, 0.00000243, -0.00000856, -0.00000035, 0.00000723, -0.00000284, -0.00001048, 0.00000196, 0.00000910, -0.00000510, -0.00000510, -0.00000021, -0.00000267, -0.00000033, 0.00000356, -0.00000558, 0.00001521, -0.00000587, -0.00000559, -0.00000411, 0.00001655, -0.00000226, -0.00001215, 0.00000597, -0.00000262, 0.00000501, -0.00000922, 0.00000237, 0.00000591, -0.00000389, -0.00000426, -0.00000099, 0.00000877, 0.00000014, -0.00000284, 0.00000094, -0.00001196, -0.00000130, 0.00000814, -0.00001449, 0.00001717, -0.00002312, -0.00000841, 0.00000318, 0.00001442, 0.00002683, -0.00000837, 0.00002801, -0.00000052, 0.00000704, -0.00001223, -0.00001117, 0.00000933, -0.00001005, 0.00000529, -0.00000625, 0.00000566, -0.00001258, 0.00002438, -0.00001584, 0.00000623, -0.00000420, -0.00001047, 0.00001188, -0.00001849, 0.00000747, -0.00000064, 0.00000250, -0.00002371, },
		{ -0.00000334, 0.00000518, -0.00001047, 0.00001244, -0.00001179, 0.00001324, -0.00000296, 0.00000254, 0.00000371, -0.00000243, 0.00000398, -0.00000885, -0.00000605, 0.00000592, 0.00000362, -0.00000191, 0.00000118, -0.00000233, 0.00000421, -0.00000774, -0.00000286, 0.00000748, 0.00000759, -0.00001092, -0.00001106, 0.00002203, 0.00000562, -0.00001722, -0.00000732, 0.00000445, -0.00001173, 0.00001104, -0.00000694, 0.00000475, 0.00000684, -0.00001328, -0.00001064, 0.00000669, 0.00000598, -0.00001357, -0.00000682, 0.00000574, 0.00000011, -0.00000515, -0.00000517, 0.00000086, -0.00000475, -0.00002303, -0.00002841, -0.00006228, -0.00019303, 0.00044752, 0.00023599, -0.00022916, -0.00000936, -0.00003544, 0.00000449, -0.00002607, -0.00000704, -0.00000278, -0.00000030, -0.00000313, -0.00000719, 0.00000643, 0.00000311, -0.00000930, -0.00000561, 0.00000930, 0.00000131, -0.00000720, -0.00000171, -0.00000037, -0.00000449, 0.00000676, -0.00000661, 0.00000828, 0.00000611, -0.00000993, -0.00000404, 0.00000682, 0.00001299, -0.00001767, 0.00000361, -0.00000105, 0.00000333, -0.00000325, -0.00000577, 0.00000872, -0.00000129, -0.00000365, -0.00000463, 0.00000624, 0.00000506, -0.00000325, -0.00000006, -0.00000453, -0.00001287, 0.00001372, -0.00001237, 0.00000737, -0.00000201, -0.00002439, 0.00000768, -0.00000023, 0.00003344, -0.00000267, 0.00001296, 0.00001639, -0.00000355, 0.00000553, -0.00002290, 0.00000945, -0.00000600, 0.00000059, -0.00000284, 0.00000313, -0.00000763, 0.00000775, 0.00000997, -0.00001779, 0.00001762, -0.00002510, 0.00001639, -0.00001380, -0.00000092, 0.00000144, 0.00000597, -0.00001791, },
		{ -0.00002058, 0.00001443, -0.00001503, 0.00001148, -0.00000652, 0.00000372, 0.00000660, -0.00000469, 0.00000812, -0.00000373, 0.00000347, -0.00000266, -0.00001034, 0.00000183, 0.00000627, -0.00000056, -0.00000024, -0.00000056, 0.00000097, -0.00000022, -0.00000992, 0.00000665, 0.00000670, 0.00000082, -0.00002045, 0.00001375, 0.00001521, -0.00000792, -0.00001873, 0.00000825, -0.00001152, 0.00000489, 0.00000145, -0.00000417, 0.00001026, -0.00000387, -0.00001708, 0.00000142, 0.00000760, -0.00000213, -0.00001834, 0.00000806, -0.00000300, 0.00000465, -0.00001567, 0.00000960, -0.00001430, 0.00000356, -0.00005579, 0.00000881, -0.00022363, 0.00014463, 0.00052755, -0.00017850, -0.00008757, -0.00001656, -0.00001375, -0.00000926, -0.00002142, -0.00000008, -0.00000286, 0.00000006, -0.00000712, -0.00000135, 0.00000954, -0.00000660, -0.00000824, 0.00000186, 0.00000948, -0.00000744, -0.00000312, -0.00000051, -0.00000350, 0.00000180, 0.00000217, -0.00000534, 0.00001672, -0.00001012, -0.00000349, -0.00000294, 0.00001986, -0.00001117, -0.00000577, 0.00000329, -0.00000019, 0.00000206, -0.00000840, 0.00000431, 0.00000475, -0.00000431, -0.00000404, -0.00000035, 0.00000924, -0.00000153, -0.00000188, 0.00000007, -0.00001383, 0.00000395, 0.00000325, -0.00001046, 0.00001629, -0.00002882, 0.00000307, -0.00000359, 0.00002363, 0.00001476, -0.00000316, 0.00002415, -0.00000179, 0.00000579, -0.00001155, -0.00000916, 0.00000932, -0.00001081, 0.00000599, -0.00000621, 0.00000556, -0.00001204, 0.00002638, -0.00002184, 0.00001337, -0.00001127, -0.00000371, 0.00000679, -0.00001578, 0.00000701, 0.00000097, -0.00000235, },
		{ -0.00002371, 0.00000708, -0.00000498, -0.00000129, 0.00000642, -0.00000821, 0.00001384, -0.00000702, 0.00000684, -0.00000004, -0.00000068, 0.00000297, -0.00000962, -0.00000432, 0.00000635, 0.00000273, -0.00000186, 0.00000094, -0.00000133, 0.00000339, -0.00000924, 0.00000003, 0.00000759, 0.00000618, -0.00001588, -0.00000343, 0.00002186, 0.00000154, -0.00001895, -0.00000216, 0.00000080, -0.00000935, 0.00001236, -0.00001034, 0.00000882, 0.00000305, -0.00001270, -0.00001048, 0.00000936, 0.00000349, -0.00001432, -0.00000523, 0.00000606, -0.00000113, -0.00000448, -0.00000620, 0.00000226, -0.00000804, -0.00002624, -0.00002490, -0.00008518, -0.00018367, 0.00057910, 0.00009511, -0.00022273, 0.00001306, -0.00004385, 0.00001177, -0.00003351, 0.00000067, -0.00000678, 0.00000332, -0.00000609, -0.00000549, 0.00000665, 0.00000257, -0.00001091, -0.00000335, 0.00000886, 0.00000028, -0.00000760, 0.00000004, -0.00000202, -0.00000329, 0.00000791, -0.00000956, 0.00001253, 0.00000132, -0.00000863, -0.00000436, 0.00001271, 0.00000558, -0.00001651, 0.00000564, -0.00000199, 0.00000393, -0.00000586, -0.00000249, 0.00000813, -0.00000218, -0.00000340, -0.00000468, 0.00000775, 0.00000335, -0.00000306, 0.00000020, -0.00000684, -0.00001031, 0.00001486, -0.00001649, 0.00001554, -0.00001432, -0.00001227, 0.00000299, 0.00000584, 0.00002897, -0.00000634, 0.00001618, 0.00001122, -0.00000305, 0.00000425, -0.00002079, 0.00001055, -0.00000721, 0.00000132, -0.00000317, 0.00000420, -0.00000948, 0.00001244, 0.00000378, -0.00001393, 0.00001625, -0.00002575, 0.00001950, -0.00001863, 0.00000419, -0.00000129, 0.00000748, },
		{ -0.00000992, -0.00001240, 0.00001063, -0.00001466, 0.00001526, -0.00001239, 0.00001201, -0.00000162, 0.00000056, 0.00000527, -0.00000338, 0.00000357, -0.00000362, -0.00001013, 0.00000401, 0.00000473, -0.00000002, -0.00000122, 0.00000076, 0.00000058, -0.00000166, -0.00000909, 0.00000899, 0.00000536, -0.00000310, -0.00001906, 0.00002067, 0.00000885, -0.00000878, -0.00001811, 0.00001188, -0.00001727, 0.00001354, -0.00000692, 0.00000258, 0.00000543, -0.00000160, -0.00002082, 0.00000776, 0.00000303, 0.00000023, -0.00002260, 0.00001531, -0.00001079, 0.00001231, -0.00002476, 0.00001991, -0.00002628, 0.00001214, -0.00007033, 0.00002336, -0.00028125, 0.00028929, 0.00048596, -0.00027960, -0.00001882, -0.00004516, 0.00000856, -0.00002700, -0.00000778, -0.00000801, 0.00000438, -0.00000557, -0.00000326, -0.00000409, 0.00001322, -0.00001192, -0.00000416, 0.00000010, 0.00001099, -0.00001076, 0.00000021, -0.00000232, -0.00000316, 0.00000382, -0.00000028, -0.00000359, 0.00001656, -0.00001328, -0.00000171, -0.00000017, 0.00001993, -0.00001817, 0.00000104, -0.00000033, 0.00000292, -0.00000207, -0.00000573, 0.00000488, 0.00000397, -0.00000471, -0.00000388, 0.00000058, 0.00000921, -0.00000295, -0.00000087, -0.00000128, -0.00001483, 0.00000931, -0.00000353, -0.00000256, 0.00000925, -0.00002697, 0.00000868, -0.00000713, 0.00003007, 0.00000339, 0.00000329, 0.00001883, -0.00000163, 0.00000377, -0.00000972, -0.00000836, 0.00001003, -0.00001233, 0.00000752, -0.00000674, 0.00000560, -0.00001105, 0.00002722, -0.00002694, 0.00002055, -0.00001917, 0.00000462, -0.00000064, -0.00001034, 0.00000420, 0.00000443, },
		{ 0.00000581, -0.00002523, 0.00001421, -0.00001421, 0.00000950, -0.00000315, 0.00000013, 0.00000911, -0.00000668, 0.00000839, -0.00000258, 0.00000068, 0.00000219, -0.00001059, -0.00000187, 0.00000564, 0.00000282, -0.00000276, 0.00000184, -0.00000125, 0.00000271, -0.00001061, 0.00000363, 0.00000650, 0.00000460, -0.00002015, 0.00000625, 0.00001765, -0.00000007, -0.00002164, 0.00000494, -0.00000643, -0.00000200, 0.00000809, -0.00000879, 0.00000875, 0.00000270, -0.00001537, -0.00000657, 0.00000810, 0.00000454, -0.00001885, 0.00000118, 0.00000068, 0.00000392, -0.00001078, 0.00000044, -0.00000481, -0.00000274, -0.00003927, -0.00000833, -0.00013143, -0.00013305, 0.00068867, -0.00007890, -0.00016648, 0.00000511, -0.00003431, 0.00000547, -0.00003035, 0.00000064, -0.00000489, 0.00000204, -0.00000512, -0.00000721, 0.00001016, -0.00000145, -0.00000917, -0.00000380, 0.00001055, -0.00000284, -0.00000579, 0.00000000, -0.00000276, -0.00000196, 0.00000844, -0.00001235, 0.00001722, -0.00000477, -0.00000578, -0.00000487, 0.00001861, -0.00000439, -0.00001117, 0.00000434, -0.00000043, 0.00000193, -0.00000606, -0.00000040, 0.00000778, -0.00000328, -0.00000282, -0.00000493, 0.00000949, 0.00000108, -0.00000223, -0.00000009, -0.00000915, -0.00000609, 0.00001289, -0.00001676, 0.00001981, -0.00002429, 0.00000114, -0.00000554, 0.00001589, 0.00001903, -0.00000433, 0.00001512, 0.00000942, -0.00000510, 0.00000583, -0.00002120, 0.00001357, -0.00001043, 0.00000397, -0.00000499, 0.00000642, -0.00001230, 0.00001823, -0.00000459, -0.00000681, 0.00001124, -0.00002262, 0.00001889, -0.00002023, 0.00000691, -0.00000185, },
		{ 0.00000808, -0.00001805, -0.00000128, 0.00000151, -0.00000776, 0.00001259, -0.00001282, 0.00001643, -0.00000839, 0.00000627, 0.00000110, -0.00000187, 0.00000312, -0.00000449, -0.00000938, 0.00000552, 0.00000349, 0.00000024, -0.00000186, 0.00000212, -0.00000030, -0.00000302, -0.00000731, 0.00001039, 0.00000390, -0.00000779, -0.00001413, 0.00002402, 0.00000345, -0.00001004, -0.00001527, 0.00001217, -0.00001964, 0.00001976, -0.00001448, 0.00000948, 0.00000013, 0.00000089, -0.00002423, 0.00001393, -0.00000167, 0.00000218, -0.00002564, 0.00002127, -0.00001756, 0.00001912, -0.00003330, 0.00003003, -0.00003886, 0.00002089, -0.00008367, 0.00003377, -0.00033863, 0.00047289, 0.00036800, -0.00032761, 0.00003566, -0.00007085, 0.00002909, -0.00004424, 0.00000659, -0.00001672, 0.00001206, -0.00001154, 0.00000053, -0.00000625, 0.00001602, -0.00001662, -0.00000018, -0.00000167, 0.00001221, -0.00001373, 0.00000372, -0.00000483, -0.00000196, 0.00000563, -0.00000381, 0.00000003, 0.00001402, -0.00001446, -0.00000085, 0.00000451, 0.00001574, -0.00002117, 0.00000602, -0.00000301, 0.00000504, -0.00000576, -0.00000242, 0.00000465, 0.00000331, -0.00000487, -0.00000404, 0.00000205, 0.00000836, -0.00000375, -0.00000010, -0.00000306, -0.00001441, 0.00001342, -0.00001022, 0.00000707, -0.00000227, -0.00001774, 0.00000658, -0.00000456, 0.00003056, -0.00000374, 0.00000771, 0.00001486, -0.00000256, 0.00000331, -0.00000897, -0.00000684, 0.00000975, -0.00001302, 0.00000855, -0.00000674, 0.00000479, -0.00000861, 0.00002566, -0.00002947, 0.00002584, -0.00002587, 0.00001248, -0.00000846, -0.00000382, 0.00000049, },
		{ 0.00000022, 0.00000044, -0.00002112, 0.00001534, -0.00001876, 0.00001784, -0.00001267, 0.00001051, 0.00000053, -0.00000218, 0.00000661, -0.00000310, 0.00000107, 0.00000192, -0.00001179, 0.00000114, 0.00000393, 0.00000385, -0.00000459, 0.00000400, -0.00000243, 0.00000252, -0.00001180, 0.00000761, 0.00000435, 0.00000267, -0.00002264, 0.00001619, 0.00001055, 0.00000052, -0.00002475, 0.00001278, -0.00001574, 0.00000911, -0.00000134, -0.00000201, 0.00000391, 0.00000653, -0.00002194, 0.00000175, 0.00000220, 0.00000982, -0.00002768, 0.00001287, -0.00001099, 0.00001609, -0.00002527, 0.00001646, -0.00002272, 0.00001410, -0.00006535, 0.00002504, -0.00020795, -0.00002234, 0.00074229, -0.00025359, -0.00007601, -0.00002988, -0.00000678, -0.00001520, -0.00001530, -0.00000858, 0.00000417, -0.00000515, 0.00000053, -0.00001279, 0.00001724, -0.00000917, -0.00000374, -0.00000739, 0.00001480, -0.00000841, -0.00000133, -0.00000242, -0.00000186, -0.00000096, 0.00000819, -0.00001430, 0.00002139, -0.00001121, -0.00000206, -0.00000478, 0.00002284, -0.00001475, -0.00000316, 0.00000043, 0.00000315, -0.00000239, -0.00000368, -0.00000002, 0.00000804, -0.00000471, -0.00000197, -0.00000522, 0.00001119, -0.00000150, -0.00000089, -0.00000102, -0.00001099, -0.00000083, 0.00000802, -0.00001247, 0.00001808, -0.00002814, 0.00001094, -0.00001337, 0.00002616, 0.00000679, 0.00000118, 0.00001124, 0.00000984, -0.00000876, 0.00000948, -0.00002364, 0.00001822, -0.00001543, 0.00000842, -0.00000823, 0.00000959, -0.00001567, 0.00002442, -0.00001422, 0.00000276, 0.00000310, -0.00001596, 0.00001443, -0.00001809, 0.00000660, },
		{ -0.00000227, 0.00000984, -0.00002388, 0.00000933, -0.00000979, 0.00000368, 0.00000395, -0.00000696, 0.00001504, -0.00001117, 0.00001005, -0.00000265, -0.00000024, 0.00000275, -0.00000574, -0.00000766, 0.00000578, 0.00000328, -0.00000054, -0.00000130, 0.00000258, -0.00000098, -0.00000458, -0.00000437, 0.00001032, 0.00000266, -0.00001292, -0.00000574, 0.00002262, 0.00000078, -0.00001297, -0.00000928, 0.00000776, -0.00001641, 0.00002043, -0.00001772, 0.00001309, -0.00000241, 0.00000048, -0.00002417, 0.00001664, -0.00000317, 0.00000028, -0.00002371, 0.00002162, -0.00001837, 0.00001943, -0.00003486, 0.00003271, -0.00004403, 0.00002088, -0.00008499, 0.00002470, -0.00037255, 0.00067168, 0.00017956, -0.00030790, 0.00005646, -0.00007875, 0.00003672, -0.00005200, 0.00001439, -0.00002045, 0.00001549, -0.00001395, 0.00000092, -0.00000487, 0.00001515, -0.00001801, 0.00000119, -0.00000123, 0.00001118, -0.00001451, 0.00000568, -0.00000665, -0.00000067, 0.00000738, -0.00000830, 0.00000554, 0.00000885, -0.00001314, -0.00000114, 0.00001073, 0.00000734, -0.00001905, 0.00000747, -0.00000321, 0.00000464, -0.00000735, 0.00000007, 0.00000454, 0.00000227, -0.00000446, -0.00000475, 0.00000424, 0.00000648, -0.00000364, 0.00000011, -0.00000502, -0.00001232, 0.00001506, -0.00001459, 0.00001529, -0.00001469, -0.00000403, -0.00000204, 0.00000433, 0.00002406, -0.00000477, 0.00000803, 0.00001414, -0.00000638, 0.00000622, -0.00001120, -0.00000280, 0.00000684, -0.00001133, 0.00000768, -0.00000508, 0.00000223, -0.00000394, 0.00002079, -0.00002809, 0.00002752, -0.00002945, 0.00001773, -0.00001444, 0.00000172, },
		{ 0.00000389, 0.00000413, -0.00000756, -0.00001201, 0.00000976, -0.00001635, 0.00002033, -0.00001834, 0.00001929, -0.00000895, 0.00000471, 0.00000220, -0.00000172, 0.00000052, 0.00000202, -0.00001291, 0.00000416, 0.00000184, 0.00000524, -0.00000679, 0.00000694, -0.00000470, 0.00000284, -0.00001257, 0.00001139, 0.00000154, 0.00000014, -0.00002221, 0.00002413, 0.00000265, 0.00000194, -0.00002692, 0.00001923, -0.00002436, 0.00002104, -0.00001334, 0.00000806, -0.00000416, 0.00001314, -0.00003099, 0.00001297, -0.00000685, 0.00001771, -0.00003885, 0.00002753, -0.00002644, 0.00003275, -0.00004523, 0.00003917, -0.00004897, 0.00004007, -0.00010162, 0.00007130, -0.00031176, 0.00015718, 0.00071027, -0.00039107, 0.00002193, -0.00007786, 0.00003040, -0.00004440, 0.00000759, -0.00002416, 0.00001817, -0.00001641, 0.00000919, -0.00002055, 0.00002618, -0.00001898, 0.00000396, -0.00001301, 0.00002065, -0.00001554, 0.00000504, -0.00000688, 0.00000068, -0.00000045, 0.00000692, -0.00001460, 0.00002391, -0.00001672, 0.00000159, -0.00000327, 0.00002370, -0.00002288, 0.00000509, -0.00000431, 0.00000726, -0.00000770, 0.00000047, -0.00000109, 0.00000883, -0.00000627, -0.00000115, -0.00000519, 0.00001240, -0.00000393, 0.00000067, -0.00000257, -0.00001181, 0.00000431, 0.00000154, -0.00000456, 0.00001031, -0.00002404, 0.00001347, -0.00001614, 0.00003235, -0.00000362, 0.00000674, 0.00000730, 0.00001001, -0.00001168, 0.00001301, -0.00002626, 0.00002289, -0.00002076, 0.00001346, -0.00001189, 0.00001274, -0.00001849, 0.00002958, -0.00002335, 0.00001298, -0.00000647, -0.00000727, 0.00000739, -0.00001307, },
		{ 0.00000467, -0.00000263, 0.00000811, -0.00002493, 0.00001579, -0.00001909, 0.00001586, -0.00000790, 0.00000380, 0.00000786, -0.00000884, 0.00001026, -0.00000446, 0.00000048, 0.00000308, -0.00000772, -0.00000473, 0.00000447, 0.00000453, -0.00000286, 0.00000112, 0.00000131, -0.00000072, -0.00000666, -0.00000015, 0.00000849, 0.00000180, -0.00001779, 0.00000509, 0.00001649, 0.00000165, -0.00001805, -0.00000016, -0.00000140, -0.00000687, 0.00001393, -0.00001430, 0.00001085, 0.00000037, -0.00000532, -0.00001808, 0.00001317, 0.00000130, -0.00000819, -0.00001383, 0.00001291, -0.00000915, 0.00000846, -0.00002378, 0.00002120, -0.00003391, 0.00000309, -0.00006312, -0.00002066, -0.00035402, 0.00084805, -0.00005383, -0.00022332, 0.00003520, -0.00006014, 0.00002405, -0.00004372, 0.00000998, -0.00001453, 0.00001081, -0.00000966, -0.00000464, 0.00000228, 0.00000850, -0.00001402, -0.00000206, 0.00000326, 0.00000627, -0.00001150, 0.00000449, -0.00000634, -0.00000022, 0.00000921, -0.00001339, 0.00001246, 0.00000136, -0.00000936, -0.00000231, 0.00001737, -0.00000404, -0.00001210, 0.00000488, -0.00000030, 0.00000100, -0.00000577, 0.00000056, 0.00000541, 0.00000040, -0.00000326, -0.00000608, 0.00000713, 0.00000353, -0.00000247, -0.00000053, -0.00000672, -0.00000881, 0.00001370, -0.00001498, 0.00001912, -0.00002372, 0.00000939, -0.00001335, 0.00001690, 0.00001224, -0.00000032, 0.00000427, 0.00001671, -0.00001318, 0.00001278, -0.00001696, 0.00000443, 0.00000063, -0.00000650, 0.00000419, -0.00000125, -0.00000235, 0.00000300, 0.00001257, -0.00002247, 0.00002489, -0.00002889, 0.00001907, -0.00001702, },
		{ -0.00000707, 0.00000046, 0.00000796, -0.00001572, -0.00000046, -0.00000086, -0.00000713, 0.00001552, -0.00001751, 0.00002315, -0.00001604, 0.00001134, -0.00000312, 0.00000064, -0.00000038, 0.00000198, -0.00001333, 0.00000639, 0.00000028, 0.00000603, -0.00000831, 0.00000961, -0.00000732, 0.00000326, -0.00001246, 0.00001422, -0.00000135, -0.00000314, -0.00001807, 0.00002794, -0.00000352, 0.00000230, -0.00002635, 0.00002188, -0.00002901, 0.00002979, -0.00002387, 0.00001769, -0.00001205, 0.00001924, -0.00003913, 0.00002355, -0.00001547, 0.00002447, -0.00004813, 0.00004033, -0.00004026, 0.00004787, -0.00006402, 0.00006129, -0.00007571, 0.00006660, -0.00013785, 0.00011694, -0.00042620, 0.00039700, 0.00057724, -0.00045835, 0.00009729, -0.00011871, 0.00006318, -0.00007132, 0.00002990, -0.00003955, 0.00003180, -0.00002726, 0.00001696, -0.00002692, 0.00003356, -0.00002764, 0.00001101, -0.00001807, 0.00002579, -0.00002211, 0.00001146, -0.00001192, 0.00000406, -0.00000008, 0.00000428, -0.00001255, 0.00002365, -0.00002002, 0.00000422, 0.00000019, 0.00002004, -0.00002644, 0.00001097, -0.00000770, 0.00000988, -0.00001200, 0.00000477, -0.00000262, 0.00000956, -0.00000747, -0.00000081, -0.00000439, 0.00001259, -0.00000564, 0.00000195, -0.00000451, -0.00001120, 0.00000802, -0.00000449, 0.00000452, -0.00000127, -0.00001299, 0.00000775, -0.00001150, 0.00003142, -0.00000871, 0.00000908, 0.00000614, 0.00000730, -0.00001122, 0.00001376, -0.00002662, 0.00002538, -0.00002435, 0.00001728, -0.00001442, 0.00001449, -0.00001939, 0.00003208, -0.00002995, 0.00002157, -0.00001519, 0.00000118, -0.00000008, },
		{ -0.00001590, 0.00000460, -0.00000033, 0.00000266, -0.00002020, 0.00001569, -0.00002268, 0.00002444, -0.00001890, 0.00001618, -0.00000310, -0.00000205, 0.00000659, -0.00000367, -0.00000023, 0.00000437, -0.00001046, -0.00000076, 0.00000177, 0.00000713, -0.00000666, 0.00000546, -0.00000207, 0.00000094, -0.00000937, 0.00000514, 0.00000492, 0.00000128, -0.00002135, 0.00001647, 0.00000699, 0.00000553, -0.00002462, 0.00001078, -0.00001369, 0.00000754, 0.00000092, -0.00000412, 0.00000224, 0.00000910, -0.00001712, -0.00000534, 0.00000279, 0.00001242, -0.00002377, 0.00000450, -0.00000554, 0.00001102, -0.00001518, 0.00000198, -0.00000741, -0.00000467, -0.00003713, -0.00001197, -0.00011366, -0.00025650, 0.00095819, -0.00029097, -0.00009493, -0.00002210, -0.00001634, -0.00000891, -0.00001855, -0.00000790, 0.00000221, -0.00000295, 0.00000205, -0.00001657, 0.00001541, -0.00000410, -0.00000435, -0.00001035, 0.00001223, -0.00000298, -0.00000417, -0.00000059, -0.00000305, -0.00000127, 0.00001113, -0.00001840, 0.00001980, -0.00000747, -0.00000377, -0.00000364, 0.00002278, -0.00001606, -0.00000204, -0.00000085, 0.00000509, -0.00000541, -0.00000102, -0.00000139, 0.00000764, -0.00000242, -0.00000131, -0.00000787, 0.00001044, -0.00000021, -0.00000033, -0.00000219, -0.00000755, -0.00000474, 0.00000989, -0.00001114, 0.00001695, -0.00002601, 0.00001772, -0.00002233, 0.00002870, -0.00000108, 0.00000683, -0.00000148, 0.00002070, -0.00002119, 0.00002140, -0.00002510, 0.00001397, -0.00000816, 0.00000096, -0.00000153, 0.00000430, -0.00000829, 0.00001126, 0.00000213, -0.00001355, 0.00001858, -0.00002456, 0.00001658, },
		{ -0.00000573, -0.00000223, -0.00000174, 0.00000958, -0.00002107, 0.00000980, -0.00001286, 0.00000625, 0.00000449, -0.00000971, 0.00002034, -0.00001816, 0.00001513, -0.00000731, 0.00000263, -0.00000074, 0.00000108, -0.00001237, 0.00000699, 0.00000023, 0.00000517, -0.00000795, 0.00001069, -0.00000915, 0.00000309, -0.00001087, 0.00001525, -0.00000366, -0.00000711, -0.00001016, 0.00002628, -0.00000588, -0.00000013, -0.00002162, 0.00001887, -0.00002697, 0.00003171, -0.00002871, 0.00002265, -0.00001562, 0.00002075, -0.00004213, 0.00002907, -0.00001909, 0.00002536, -0.00005029, 0.00004531, -0.00004571, 0.00005376, -0.00007285, 0.00007278, -0.00009159, 0.00008079, -0.00015836, 0.00014053, -0.00052131, 0.00066845, 0.00034970, -0.00043752, 0.00012612, -0.00013347, 0.00007694, -0.00008398, 0.00004179, -0.00004676, 0.00003843, -0.00003216, 0.00001908, -0.00002758, 0.00003536, -0.00003130, 0.00001381, -0.00001930, 0.00002728, -0.00002535, 0.00001535, -0.00001540, 0.00000680, 0.00000082, -0.00000005, -0.00000770, 0.00001991, -0.00002014, 0.00000521, 0.00000549, 0.00001174, -0.00002417, 0.00001255, -0.00000791, 0.00000923, -0.00001335, 0.00000738, -0.00000332, 0.00000938, -0.00000769, -0.00000143, -0.00000240, 0.00001129, -0.00000608, 0.00000239, -0.00000632, -0.00000922, 0.00000940, -0.00000804, 0.00001167, -0.00001282, 0.00000142, -0.00000405, -0.00000007, 0.00002290, -0.00000705, 0.00000653, 0.00000926, 0.00000022, -0.00000572, 0.00000981, -0.00002279, 0.00002383, -0.00002439, 0.00001823, -0.00001447, 0.00001372, -0.00001736, 0.00003081, -0.00003251, 0.00002672, -0.00002110, 0.00000726, },
		{ 0.00001172, -0.00001224, 0.00000245, 0.00000352, -0.00000507, -0.00001078, 0.00000883, -0.00001870, 0.00002614, -0.00002504, 0.00002620, -0.00001448, 0.00000696, 0.00000016, -0.00000067, -0.00000217, 0.00000632, -0.00001347, 0.00000356, -0.00000156, 0.00001033, -0.00001112, 0.00001102, -0.00000719, 0.00000391, -0.00001245, 0.00001093, 0.00000004, 0.00000085, -0.00002247, 0.00002598, -0.00000339, 0.00001065, -0.00003092, 0.00002110, -0.00002598, 0.00002339, -0.00001545, 0.00001029, -0.00001057, 0.00002175, -0.00003283, 0.00001195, -0.00001239, 0.00002793, -0.00004379, 0.00002823, -0.00003043, 0.00003866, -0.00004785, 0.00003875, -0.00004962, 0.00004037, -0.00009622, 0.00006450, -0.00025316, -0.00006588, 0.00096347, -0.00048439, 0.00004314, -0.00009650, 0.00004097, -0.00005386, 0.00001752, -0.00003488, 0.00002632, -0.00002289, 0.00001857, -0.00003234, 0.00003207, -0.00002037, 0.00000898, -0.00002196, 0.00002419, -0.00001522, 0.00000641, -0.00000887, 0.00000302, -0.00000377, 0.00001273, -0.00002234, 0.00002617, -0.00001614, 0.00000252, -0.00000424, 0.00002517, -0.00002592, 0.00000838, -0.00000763, 0.00001115, -0.00001292, 0.00000573, -0.00000520, 0.00001090, -0.00000584, 0.00000097, -0.00000961, 0.00001356, -0.00000416, 0.00000236, -0.00000476, -0.00000703, -0.00000128, 0.00000514, -0.00000439, 0.00000934, -0.00002046, 0.00001788, -0.00002477, 0.00003520, -0.00001144, 0.00001285, -0.00000601, 0.00002310, -0.00002740, 0.00002918, -0.00003308, 0.00002358, -0.00001755, 0.00000937, -0.00000806, 0.00001022, -0.00001418, 0.00001916, -0.00000858, -0.00000331, 0.00001040, -0.00001814, },
		{ 0.00000987, -0.00000820, 0.00000007, -0.00000145, 0.00000757, -0.00002132, 0.00001521, -0.00002189, 0.00001952, -0.00000945, 0.00000315, 0.00001122, -0.00001441, 0.00001447, -0.00000874, 0.00000303, 0.00000026, -0.00000126, -0.00000958, 0.00000546, 0.00000230, 0.00000196, -0.00000480, 0.00000899, -0.00000903, 0.00000154, -0.00000730, 0.00001382, -0.00000489, -0.00001135, 0.00000066, 0.00001908, -0.00000347, -0.00000615, -0.00001238, 0.00000979, -0.00001725, 0.00002481, -0.00002505, 0.00001981, -0.00001169, 0.00001441, -0.00003660, 0.00002590, -0.00001400, 0.00001654, -0.00004112, 0.00003768, -0.00003725, 0.00004393, -0.00006401, 0.00006441, -0.00008568, 0.00006974, -0.00014690, 0.00011831, -0.00055935, 0.00092558, 0.00005804, -0.00033213, 0.00009845, -0.00011129, 0.00006230, -0.00007400, 0.00003590, -0.00003958, 0.00003286, -0.00002672, 0.00001188, -0.00001928, 0.00002851, -0.00002703, 0.00000954, -0.00001406, 0.00002265, -0.00002288, 0.00001440, -0.00001519, 0.00000732, 0.00000297, -0.00000614, -0.00000014, 0.00001261, -0.00001670, 0.00000444, 0.00001182, -0.00000006, -0.00001629, 0.00000920, -0.00000416, 0.00000447, -0.00001051, 0.00000689, -0.00000203, 0.00000755, -0.00000644, -0.00000332, 0.00000098, 0.00000826, -0.00000486, 0.00000144, -0.00000728, -0.00000649, 0.00000835, -0.00000793, 0.00001430, -0.00002027, 0.00001423, -0.00001743, 0.00001469, 0.00000917, 0.00000010, -0.00000026, 0.00001607, -0.00001074, 0.00000464, 0.00000091, -0.00001428, 0.00001764, -0.00002015, 0.00001558, -0.00001150, 0.00001011, -0.00001233, 0.00002567, -0.00003066, 0.00002773, -0.00002328, },
		{ -0.00001185, 0.00000672, -0.00000807, 0.00000009, 0.00000682, -0.00001236, 0.00000033, -0.00000250, -0.00000794, 0.00001995, -0.00002386, 0.00003050, -0.00002272, 0.00001517, -0.00000691, 0.00000326, -0.00000444, 0.00000811, -0.00001589, 0.00000720, -0.00000433, 0.00001283, -0.00001486, 0.00001636, -0.00001290, 0.00000743, -0.00001518, 0.00001623, -0.00000535, 0.00000018, -0.00002027, 0.00003132, -0.00001179, 0.00001458, -0.00003465, 0.00002789, -0.00003448, 0.00003612, -0.00003037, 0.00002436, -0.00002328, 0.00003404, -0.00004806, 0.00002918, -0.00002766, 0.00004287, -0.00006272, 0.00005114, -0.00005486, 0.00006605, -0.00008100, 0.00007709, -0.00009518, 0.00008999, -0.00016110, 0.00014990, -0.00042083, 0.00021115, 0.00084236, -0.00059276, 0.00015329, -0.00016214, 0.00009334, -0.00009642, 0.00005314, -0.00006197, 0.00005041, -0.00004276, 0.00003444, -0.00004691, 0.00004750, -0.00003585, 0.00002190, -0.00003321, 0.00003584, -0.00002745, 0.00001754, -0.00001822, 0.00001040, -0.00000690, 0.00001325, -0.00002407, 0.00003000, -0.00002294, 0.00000820, -0.00000334, 0.00002324, -0.00003106, 0.00001621, -0.00001290, 0.00001551, -0.00001914, 0.00001239, -0.00000943, 0.00001421, -0.00000906, 0.00000285, -0.00001059, 0.00001571, -0.00000748, 0.00000487, -0.00000772, -0.00000505, 0.00000050, 0.00000140, 0.00000269, -0.00000109, -0.00000880, 0.00000974, -0.00001890, 0.00003363, -0.00001547, 0.00001450, -0.00000646, 0.00002110, -0.00002883, 0.00003303, -0.00003795, 0.00003058, -0.00002503, 0.00001654, -0.00001356, 0.00001487, -0.00001841, 0.00002489, -0.00001743, 0.00000597, 0.00000260, },
		{ -0.00002231, 0.00000939, -0.00000759, -0.00000026, 0.00000125, 0.00000211, -0.00001593, 0.00001379, -0.00002447, 0.00002824, -0.00002187, 0.00001739, -0.00000174, -0.00000582, 0.00000929, -0.00000683, 0.00000130, 0.00000292, -0.00000511, -0.00000506, 0.00000191, 0.00000642, -0.00000359, 0.00000131, 0.00000398, -0.00000623, -0.00000184, -0.00000159, 0.00000964, -0.00000474, -0.00001513, 0.00001262, 0.00000776, 0.00000318, -0.00001522, 0.00000026, -0.00000388, -0.00000117, 0.00000974, -0.00001274, 0.00000855, 0.00000053, -0.00000062, -0.00002156, 0.00001291, 0.00000096, -0.00000301, -0.00001955, 0.00001607, -0.00001313, 0.00001604, -0.00003430, 0.00003178, -0.00005213, 0.00002599, -0.00009330, 0.00003262, -0.00050469, 0.00111433, -0.00024843, -0.00016754, 0.00002190, -0.00005393, 0.00001926, -0.00004030, 0.00001060, -0.00001639, 0.00001365, -0.00000983, -0.00000543, -0.00000147, 0.00001249, -0.00001417, -0.00000253, -0.00000158, 0.00001108, -0.00001378, 0.00000748, -0.00001002, 0.00000456, 0.00000674, -0.00001360, 0.00000935, 0.00000249, -0.00001008, 0.00000230, 0.00001774, -0.00001311, -0.00000459, 0.00000188, 0.00000284, -0.00000387, -0.00000356, 0.00000288, 0.00000171, 0.00000381, -0.00000364, -0.00000640, 0.00000553, 0.00000370, -0.00000199, -0.00000115, -0.00000667, -0.00000408, 0.00000582, -0.00000442, 0.00001145, -0.00002087, 0.00002094, -0.00002720, 0.00002786, -0.00000543, 0.00000943, -0.00000865, 0.00002410, -0.00002314, 0.00001768, -0.00001128, -0.00000236, 0.00000781, -0.00001234, 0.00000986, -0.00000608, 0.00000442, -0.00000527, 0.00001780, -0.00002537, 0.00002533, },
		{ -0.00000306, -0.00000758, 0.00000361, -0.00000539, -0.00000061, 0.00000747, -0.00001626, 0.00000927, -0.00001427, 0.00000636, 0.00000754, -0.00001525, 0.00002747, -0.00002527, 0.00001982, -0.00001229, 0.00000642, -0.00000580, 0.00000871, -0.00001672, 0.00000901, -0.00000527, 0.00001324, -0.00001623, 0.00001964, -0.00001749, 0.00001027, -0.00001648, 0.00001979, -0.00001020, -0.00000102, -0.00001444, 0.00003092, -0.00001571, 0.00001504, -0.00003376, 0.00002873, -0.00003593, 0.00004142, -0.00003888, 0.00003297, -0.00003078, 0.00004078, -0.00005735, 0.00004064, -0.00003711, 0.00005101, -0.00007369, 0.00006551, -0.00007014, 0.00008335, -0.00010336, 0.00010410, -0.00012939, 0.00012733, -0.00021140, 0.00021742, -0.00058162, 0.00054353, 0.00059920, -0.00059263, 0.00020544, -0.00019484, 0.00012190, -0.00012106, 0.00007537, -0.00007846, 0.00006550, -0.00005493, 0.00004301, -0.00005428, 0.00005611, -0.00004527, 0.00002949, -0.00003957, 0.00004298, -0.00003575, 0.00002565, -0.00002550, 0.00001673, -0.00000919, 0.00001170, -0.00002254, 0.00002996, -0.00002628, 0.00001211, -0.00000063, 0.00001662, -0.00002998, 0.00001921, -0.00001449, 0.00001604, -0.00002171, 0.00001666, -0.00001228, 0.00001627, -0.00001104, 0.00000348, -0.00001006, 0.00001607, -0.00000928, 0.00000628, -0.00001021, -0.00000210, 0.00000015, 0.00000029, 0.00000725, -0.00001048, 0.00000496, -0.00000376, -0.00000613, 0.00002408, -0.00001216, 0.00001046, -0.00000161, 0.00001339, -0.00002385, 0.00003091, -0.00003760, 0.00003287, -0.00002856, 0.00002059, -0.00001646, 0.00001698, -0.00001984, 0.00002727, -0.00002291, 0.00001258, },
		{ 0.00002129, -0.00001992, 0.00000749, -0.00000537, -0.00000189, 0.00000486, -0.00000495, -0.00000657, 0.00000601, -0.00001932, 0.00002946, -0.00002878, 0.00002883, -0.00001491, 0.00000505, 0.00000101, -0.00000218, -0.00000219, 0.00000684, -0.00000995, 0.00000049, -0.00000282, 0.00001173, -0.00001054, 0.00000954, -0.00000389, -0.00000081, -0.00000689, 0.00000581, 0.00000299, -0.00000328, -0.00001754, 0.00002343, -0.00000514, 0.00001213, -0.00002544, 0.00001381, -0.00001897, 0.00001766, -0.00001011, 0.00000541, -0.00000869, 0.00001872, -0.00002200, 0.00000064, -0.00000760, 0.00002329, -0.00003046, 0.00001126, -0.00001611, 0.00002300, -0.00002614, 0.00001255, -0.00002170, 0.00000612, -0.00004781, 0.00000043, -0.00011961, -0.00033553, 0.00118579, -0.00051250, 0.00001531, -0.00008047, 0.00002424, -0.00004195, 0.00000968, -0.00002861, 0.00001848, -0.00001559, 0.00001532, -0.00002981, 0.00002297, -0.00001005, 0.00000487, -0.00002033, 0.00001636, -0.00000589, 0.00000072, -0.00000462, -0.00000020, -0.00000147, 0.00001178, -0.00002145, 0.00001934, -0.00000891, -0.00000138, -0.00000039, 0.00002160, -0.00002457, 0.00000804, -0.00000717, 0.00001114, -0.00001388, 0.00000607, -0.00000386, 0.00000742, -0.00000140, 0.00000022, -0.00001012, 0.00001059, -0.00000175, 0.00000204, -0.00000519, -0.00000408, -0.00000318, 0.00000350, 0.00000074, 0.00000427, -0.00001430, 0.00001923, -0.00002951, 0.00003489, -0.00001627, 0.00001684, -0.00001511, 0.00002991, -0.00003348, 0.00002998, -0.00002375, 0.00001035, -0.00000336, -0.00000287, 0.00000265, 0.00000036, -0.00000188, 0.00000212, 0.00000913, -0.00001856, },
		{ 0.00001607, -0.00000604, -0.00000596, 0.00000327, -0.00000520, 0.00000127, 0.00000443, -0.00001484, 0.00001276, -0.00002214, 0.00001966, -0.00000737, -0.00000154, 0.00001782, -0.00002139, 0.00001931, -0.00001418, 0.00000736, -0.00000518, 0.00000724, -0.00001520, 0.00000822, -0.00000348, 0.00001054, -0.00001395, 0.00001921, -0.00001924, 0.00001102, -0.00001527, 0.00002038, -0.00001343, -0.00000285, -0.00000548, 0.00002445, -0.00001382, 0.00001069, -0.00002730, 0.00002262, -0.00002880, 0.00003678, -0.00003755, 0.00003227, -0.00002902, 0.00003776, -0.00005621, 0.00004152, -0.00003572, 0.00004707, -0.00007093, 0.00006481, -0.00006884, 0.00008190, -0.00010473, 0.00010761, -0.00013774, 0.00013509, -0.00022540, 0.00023637, -0.00069000, 0.00087911, 0.00026651, -0.00048572, 0.00018616, -0.00018030, 0.00011421, -0.00011665, 0.00007437, -0.00007591, 0.00006436, -0.00005320, 0.00003897, -0.00004968, 0.00005344, -0.00004440, 0.00002772, -0.00003721, 0.00004200, -0.00003664, 0.00002737, -0.00002760, 0.00001945, -0.00000902, 0.00000727, -0.00001715, 0.00002534, -0.00002516, 0.00001349, 0.00000358, 0.00000614, -0.00002267, 0.00001649, -0.00001136, 0.00001163, -0.00001914, 0.00001679, -0.00001219, 0.00001592, -0.00001091, 0.00000221, -0.00000750, 0.00001410, -0.00000886, 0.00000573, -0.00001113, 0.00000075, -0.00000182, 0.00000232, 0.00000734, -0.00001522, 0.00001597, -0.00001774, 0.00000953, 0.00000947, -0.00000327, 0.00000181, 0.00000755, 0.00000085, -0.00001290, 0.00002269, -0.00003157, 0.00002983, -0.00002735, 0.00002074, -0.00001617, 0.00001618, -0.00001837, 0.00002622, -0.00002479, },
		{ -0.00001228, 0.00001762, -0.00001816, 0.00000634, -0.00000366, -0.00000280, 0.00000711, -0.00001065, 0.00000339, -0.00000527, -0.00000791, 0.00002272, -0.00002801, 0.00003419, -0.00002463, 0.00001493, -0.00000792, 0.00000360, -0.00000630, 0.00001096, -0.00001469, 0.00000583, -0.00000737, 0.00001670, -0.00001720, 0.00001810, -0.00001311, 0.00000617, -0.00001270, 0.00001378, -0.00000520, -0.00000091, -0.00001771, 0.00003075, -0.00001650, 0.00002059, -0.00003418, 0.00002507, -0.00003144, 0.00003435, -0.00002948, 0.00002425, -0.00002694, 0.00003791, -0.00004456, 0.00002468, -0.00003013, 0.00004721, -0.00005939, 0.00004418, -0.00005099, 0.00006239, -0.00007288, 0.00006581, -0.00008437, 0.00007638, -0.00013724, 0.00011720, -0.00032124, -0.00005374, 0.00110997, -0.00068337, 0.00017098, -0.00017720, 0.00010066, -0.00010370, 0.00006188, -0.00007047, 0.00005572, -0.00004696, 0.00004180, -0.00005490, 0.00004812, -0.00003359, 0.00002505, -0.00003925, 0.00003556, -0.00002445, 0.00001727, -0.00001912, 0.00001223, -0.00000943, 0.00001694, -0.00002818, 0.00002794, -0.00001955, 0.00000782, -0.00000270, 0.00002204, -0.00003180, 0.00001843, -0.00001509, 0.00001806, -0.00002283, 0.00001589, -0.00001148, 0.00001379, -0.00000699, 0.00000417, -0.00001351, 0.00001512, -0.00000700, 0.00000624, -0.00000989, 0.00000026, -0.00000453, 0.00000314, 0.00000496, -0.00000428, -0.00000295, 0.00000980, -0.00002327, 0.00003341, -0.00002018, 0.00001918, -0.00001676, 0.00003059, -0.00003855, 0.00003815, -0.00003327, 0.00002091, -0.00001318, 0.00000590, -0.00000409, 0.00000610, -0.00000718, 0.00000813, 0.00000159, },
		{ -0.00002408, 0.00001736, -0.00000725, -0.00000614, 0.00000503, -0.00000699, 0.00000512, -0.00000160, -0.00000820, 0.00000945, -0.00002312, 0.00002770, -0.00002021, 0.00001337, 0.00000424, -0.00001233, 0.00001372, -0.00001201, 0.00000544, -0.00000212, 0.00000343, -0.00001119, 0.00000473, 0.00000117, 0.00000454, -0.00000761, 0.00001421, -0.00001696, 0.00000869, -0.00001084, 0.00001718, -0.00001419, -0.00000515, 0.00000529, 0.00001304, -0.00000639, 0.00000169, -0.00001587, 0.00001053, -0.00001401, 0.00002249, -0.00002582, 0.00002117, -0.00001665, 0.00002345, -0.00004287, 0.00002981, -0.00002139, 0.00002896, -0.00005211, 0.00004634, -0.00004773, 0.00005770, -0.00007994, 0.00008086, -0.00011142, 0.00010200, -0.00018773, 0.00018188, -0.00070123, 0.00115488, -0.00010058, -0.00029929, 0.00010257, -0.00011911, 0.00006898, -0.00008087, 0.00004730, -0.00005150, 0.00004443, -0.00003540, 0.00002054, -0.00003157, 0.00003804, -0.00003178, 0.00001508, -0.00002460, 0.00003129, -0.00002845, 0.00002085, -0.00002256, 0.00001679, -0.00000533, -0.00000032, -0.00000810, 0.00001632, -0.00001941, 0.00001217, 0.00000838, -0.00000631, -0.00001074, 0.00000883, -0.00000402, 0.00000270, -0.00001136, 0.00001217, -0.00000845, 0.00001261, -0.00000834, -0.00000113, -0.00000288, 0.00000976, -0.00000602, 0.00000277, -0.00000956, 0.00000208, -0.00000394, 0.00000655, 0.00000280, -0.00001338, 0.00002034, -0.00002719, 0.00002295, -0.00000555, 0.00000754, -0.00000844, 0.00001816, -0.00001373, 0.00000156, 0.00001026, -0.00002124, 0.00002252, -0.00002215, 0.00001752, -0.00001323, 0.00001319, -0.00001492, 0.00002286, },
		{ -0.00000364, -0.00000778, 0.00001559, -0.00001840, 0.00000758, -0.00000419, -0.00000129, 0.00000605, -0.00001217, 0.00001009, -0.00001576, 0.00000603, 0.00001021, -0.00001987, 0.00003211, -0.00002844, 0.00002096, -0.00001492, 0.00000852, -0.00000947, 0.00001394, -0.00001801, 0.00000954, -0.00001015, 0.00001958, -0.00002152, 0.00002468, -0.00002144, 0.00001292, -0.00001772, 0.00002065, -0.00001347, 0.00000173, -0.00001511, 0.00003282, -0.00002351, 0.00002576, -0.00003883, 0.00003115, -0.00003759, 0.00004414, -0.00004278, 0.00003794, -0.00004025, 0.00005193, -0.00006184, 0.00004373, -0.00004760, 0.00006520, -0.00008157, 0.00007004, -0.00007833, 0.00009346, -0.00011095, 0.00011034, -0.00013891, 0.00013883, -0.00021865, 0.00022681, -0.00053502, 0.00031070, 0.00088591, -0.00072996, 0.00026225, -0.00023839, 0.00015195, -0.00014670, 0.00010012, -0.00010145, 0.00008390, -0.00007064, 0.00006102, -0.00007291, 0.00006676, -0.00005140, 0.00004009, -0.00005343, 0.00005059, -0.00003955, 0.00003120, -0.00003187, 0.00002394, -0.00001697, 0.00002049, -0.00003208, 0.00003316, -0.00002725, 0.00001576, -0.00000388, 0.00001840, -0.00003305, 0.00002400, -0.00001936, 0.00002111, -0.00002797, 0.00002314, -0.00001768, 0.00001904, -0.00001152, 0.00000697, -0.00001547, 0.00001798, -0.00001081, 0.00000935, -0.00001396, 0.00000537, -0.00000800, 0.00000580, 0.00000582, -0.00001054, 0.00000898, -0.00000384, -0.00001053, 0.00002402, -0.00001655, 0.00001538, -0.00001257, 0.00002493, -0.00003671, 0.00004009, -0.00003760, 0.00002714, -0.00001954, 0.00001208, -0.00000880, 0.00000993, -0.00001047, 0.00001176, },
		{ 0.00001891, -0.00002258, 0.00001817, -0.00000853, -0.00000612, 0.00000702, -0.00000900, 0.00000899, -0.00000826, 0.00000138, 0.00000056, -0.00001669, 0.00002807, -0.00002728, 0.00002535, -0.00000954, -0.00000084, 0.00000464, -0.00000651, 0.00000106, 0.00000300, -0.00000227, -0.00000532, -0.00000073, 0.00000785, -0.00000390, 0.00000201, 0.00000500, -0.00001054, 0.00000314, -0.00000327, 0.00001010, -0.00001211, -0.00000749, 0.00001598, -0.00000100, 0.00000480, -0.00001034, -0.00000160, -0.00000475, 0.00000516, 0.00000174, -0.00000637, 0.00000194, 0.00000419, -0.00000005, -0.00001937, 0.00000746, 0.00000383, -0.00000102, -0.00001979, 0.00001279, -0.00000967, 0.00001363, -0.00003167, 0.00002593, -0.00005161, 0.00002814, -0.00009655, 0.00004480, -0.00058465, 0.00131186, -0.00043662, -0.00007914, -0.00001991, -0.00002676, -0.00000289, -0.00002156, -0.00000016, -0.00000959, 0.00000926, -0.00000462, -0.00000936, -0.00000270, 0.00001242, -0.00000964, -0.00000647, -0.00000336, 0.00001220, -0.00001216, 0.00000668, -0.00001045, 0.00000848, 0.00000188, -0.00001040, 0.00000347, 0.00000415, -0.00000983, 0.00000860, 0.00001253, -0.00001816, 0.00000305, -0.00000158, 0.00000560, -0.00000883, 0.00000009, 0.00000365, -0.00000151, 0.00000671, -0.00000371, -0.00000607, 0.00000321, 0.00000366, -0.00000122, -0.00000240, -0.00000512, 0.00000069, -0.00000435, 0.00001080, -0.00000455, -0.00000553, 0.00001663, -0.00002886, 0.00002987, -0.00001643, 0.00001614, -0.00001666, 0.00002666, -0.00002667, 0.00001592, -0.00000325, -0.00000930, 0.00001324, -0.00001484, 0.00001245, -0.00000900, 0.00000936, -0.00001104, },
		{ 0.00001317, -0.00000666, -0.00000504, 0.00001510, -0.00002060, 0.00001141, -0.00000738, 0.00000307, 0.00000111, -0.00000811, 0.00001069, -0.00002139, 0.00001790, -0.00000404, -0.00000698, 0.00002354, -0.00002574, 0.00002158, -0.00001809, 0.00001089, -0.00001035, 0.00001456, -0.00001884, 0.00001051, -0.00000993, 0.00001897, -0.00002180, 0.00002712, -0.00002653, 0.00001738, -0.00002021, 0.00002458, -0.00002009, 0.00000389, -0.00000974, 0.00002894, -0.00002438, 0.00002570, -0.00003772, 0.00003037, -0.00003531, 0.00004393, -0.00004595, 0.00004184, -0.00004367, 0.00005553, -0.00006817, 0.00005173, -0.00005363, 0.00007050, -0.00008958, 0.00008056, -0.00008881, 0.00010546, -0.00012778, 0.00013123, -0.00016759, 0.00017227, -0.00026560, 0.00029311, -0.00071020, 0.00070301, 0.00054459, -0.00064931, 0.00027025, -0.00024505, 0.00016174, -0.00015648, 0.00011159, -0.00011042, 0.00009335, -0.00007826, 0.00006565, -0.00007724, 0.00007276, -0.00005773, 0.00004450, -0.00005766, 0.00005648, -0.00004645, 0.00003798, -0.00003863, 0.00003126, -0.00002142, 0.00002061, -0.00003168, 0.00003349, -0.00003021, 0.00002093, -0.00000356, 0.00001103, -0.00002797, 0.00002353, -0.00001856, 0.00001886, -0.00002745, 0.00002562, -0.00002041, 0.00002151, -0.00001366, 0.00000754, -0.00001505, 0.00001822, -0.00001212, 0.00001013, -0.00001591, 0.00000960, -0.00001236, 0.00001119, 0.00000216, -0.00001157, 0.00001700, -0.00001674, 0.00000434, 0.00001003, -0.00000745, 0.00000680, -0.00000373, 0.00001397, -0.00002851, 0.00003571, -0.00003631, 0.00002841, -0.00002169, 0.00001491, -0.00001094, 0.00001157, -0.00001175, },
		{ -0.00000809, 0.00001593, -0.00002190, 0.00002016, -0.00001171, -0.00000380, 0.00000698, -0.00000900, 0.00001039, -0.00001236, 0.00001015, -0.00001045, -0.00000506, 0.00002100, -0.00002688, 0.00003140, -0.00002001, 0.00000980, -0.00000536, 0.00000058, -0.00000452, 0.00000902, -0.00000870, 0.00000116, -0.00000673, 0.00001497, -0.00001294, 0.00001298, -0.00000676, -0.00000113, -0.00000470, 0.00000637, 0.00000001, -0.00000748, -0.00000927, 0.00002451, -0.00001465, 0.00001694, -0.00002274, 0.00001243, -0.00001934, 0.00002391, -0.00002019, 0.00001556, -0.00002027, 0.00002830, -0.00002739, 0.00000880, -0.00001990, 0.00003397, -0.00003633, 0.00001885, -0.00002794, 0.00003654, -0.00004055, 0.00002935, -0.00004555, 0.00002926, -0.00007281, 0.00003281, -0.00016172, -0.00033490, 0.00131045, -0.00068162, 0.00012269, -0.00014560, 0.00007128, -0.00008159, 0.00004542, -0.00005526, 0.00003927, -0.00003221, 0.00003132, -0.00004364, 0.00003040, -0.00001737, 0.00001644, -0.00003182, 0.00002184, -0.00001106, 0.00000849, -0.00001202, 0.00000636, -0.00000390, 0.00001131, -0.00002138, 0.00001560, -0.00000904, 0.00000194, 0.00000368, 0.00001479, -0.00002688, 0.00001551, -0.00001181, 0.00001474, -0.00002010, 0.00001255, -0.00000671, 0.00000714, -0.00000055, 0.00000187, -0.00001153, 0.00000962, -0.00000301, 0.00000441, -0.00000878, 0.00000168, -0.00000388, -0.00000146, 0.00001247, -0.00001149, 0.00000537, 0.00000638, -0.00002242, 0.00002844, -0.00002033, 0.00001955, -0.00002010, 0.00003019, -0.00003477, 0.00002682, -0.00001466, 0.00000141, 0.00000455, -0.00000765, 0.00000735, -0.00000497, 0.00000604, },
		{ -0.00001196, 0.00001353, -0.00000873, -0.00000277, 0.00001473, -0.00002323, 0.00001647, -0.00001205, 0.00000905, -0.00000635, 0.00000072, 0.00000464, -0.00001985, 0.00002389, -0.00001564, 0.00000674, 0.00001125, -0.00001790, 0.00001690, -0.00001677, 0.00000991, -0.00000821, 0.00001225, -0.00001669, 0.00000833, -0.00000628, 0.00001434, -0.00001722, 0.00002409, -0.00002665, 0.00001791, -0.00001878, 0.00002405, -0.00002348, 0.00000496, -0.00000224, 0.00001968, -0.00001884, 0.00001988, -0.00003069, 0.00002292, -0.00002480, 0.00003339, -0.00003777, 0.00003414, -0.00003503, 0.00004625, -0.00006074, 0.00004552, -0.00004486, 0.00005956, -0.00007953, 0.00007134, -0.00007737, 0.00009233, -0.00011588, 0.00011899, -0.00015832, 0.00016131, -0.00025753, 0.00028466, -0.00079473, 0.00105439, 0.00014320, -0.00046723, 0.00019921, -0.00019479, 0.00012611, -0.00012837, 0.00009129, -0.00009256, 0.00007967, -0.00006590, 0.00005230, -0.00006483, 0.00006327, -0.00004985, 0.00003564, -0.00004926, 0.00005053, -0.00004241, 0.00003476, -0.00003646, 0.00003139, -0.00002068, 0.00001611, -0.00002631, 0.00002832, -0.00002750, 0.00002238, -0.00000189, 0.00000125, -0.00001774, 0.00001740, -0.00001284, 0.00001138, -0.00002094, 0.00002244, -0.00001859, 0.00002026, -0.00001266, 0.00000535, -0.00001189, 0.00001547, -0.00001043, 0.00000786, -0.00001451, 0.00001117, -0.00001559, 0.00001761, -0.00000528, -0.00000642, 0.00001808, -0.00002438, 0.00001637, -0.00000391, 0.00000333, -0.00000347, 0.00000683, 0.00000056, -0.00001642, 0.00002685, -0.00003071, 0.00002574, -0.00002035, 0.00001493, -0.00001104, 0.00001172, },
		{ 0.00000390, -0.00000664, 0.00001453, -0.00002270, 0.00002397, -0.00001759, 0.00000200, 0.00000350, -0.00000548, 0.00000757, -0.00001125, 0.00001432, -0.00001915, 0.00000763, 0.00000924, -0.00001982, 0.00003056, -0.00002494, 0.00001683, -0.00001362, 0.00000710, -0.00000953, 0.00001439, -0.00001438, 0.00000667, -0.00001151, 0.00002056, -0.00002028, 0.00002266, -0.00001844, 0.00000907, -0.00001296, 0.00001613, -0.00001136, -0.00000117, -0.00000986, 0.00002910, -0.00002496, 0.00002696, -0.00003262, 0.00002308, -0.00002942, 0.00003736, -0.00003751, 0.00003376, -0.00003904, 0.00004902, -0.00005153, 0.00003423, -0.00004452, 0.00006082, -0.00006801, 0.00005393, -0.00006488, 0.00007852, -0.00009072, 0.00008697, -0.00011515, 0.00011015, -0.00017605, 0.00016961, -0.00040171, 0.00002271, 0.00114184, -0.00079586, 0.00026182, -0.00023944, 0.00014778, -0.00014470, 0.00010124, -0.00010213, 0.00008161, -0.00006838, 0.00006223, -0.00007308, 0.00005923, -0.00004350, 0.00003912, -0.00005408, 0.00004459, -0.00003256, 0.00002803, -0.00003029, 0.00002373, -0.00001725, 0.00002060, -0.00003104, 0.00002586, -0.00002064, 0.00001371, -0.00000150, 0.00001434, -0.00003062, 0.00002390, -0.00001912, 0.00002071, -0.00002812, 0.00002293, -0.00001623, 0.00001535, -0.00000740, 0.00000688, -0.00001610, 0.00001492, -0.00000874, 0.00000930, -0.00001473, 0.00000937, -0.00001090, 0.00000519, 0.00000962, -0.00001462, 0.00001506, -0.00000649, -0.00001052, 0.00001979, -0.00001706, 0.00001700, -0.00001790, 0.00002772, -0.00003656, 0.00003229, -0.00002191, 0.00000891, -0.00000168, -0.00000221, 0.00000349, -0.00000209, },
		{ 0.00001157, -0.00001247, 0.00001444, -0.00001155, 0.00000087, 0.00001235, -0.00002386, 0.00002028, -0.00001584, 0.00001414, -0.00001342, 0.00001133, -0.00000599, -0.00001155, 0.00002237, -0.00002141, 0.00001749, -0.00000121, -0.00000761, 0.00000854, -0.00001167, 0.00000586, -0.00000329, 0.00000724, -0.00001192, 0.00000348, 0.00000023, 0.00000632, -0.00000832, 0.00001569, -0.00002131, 0.00001386, -0.00001296, 0.00001841, -0.00002265, 0.00000462, 0.00000623, 0.00000681, -0.00000816, 0.00000933, -0.00001914, 0.00001081, -0.00000865, 0.00001510, -0.00002039, 0.00001655, -0.00001581, 0.00002543, -0.00004073, 0.00002614, -0.00002231, 0.00003354, -0.00005265, 0.00004364, -0.00004531, 0.00005526, -0.00007597, 0.00007348, -0.00010949, 0.00010246, -0.00018772, 0.00018557, -0.00074962, 0.00129788, -0.00024774, -0.00023109, 0.00007408, -0.00010208, 0.00005489, -0.00006923, 0.00004394, -0.00005126, 0.00004544, -0.00003576, 0.00002299, -0.00003754, 0.00003995, -0.00002923, 0.00001475, -0.00002920, 0.00003341, -0.00002779, 0.00002155, -0.00002487, 0.00002347, -0.00001408, 0.00000694, -0.00001643, 0.00001827, -0.00001933, 0.00001992, 0.00000056, -0.00000897, -0.00000471, 0.00000748, -0.00000382, 0.00000034, -0.00000975, 0.00001424, -0.00001242, 0.00001541, -0.00000869, 0.00000062, -0.00000633, 0.00001014, -0.00000602, 0.00000263, -0.00000933, 0.00000875, -0.00001553, 0.00002237, -0.00001395, 0.00000343, 0.00001178, -0.00002433, 0.00002183, -0.00001357, 0.00001189, -0.00001197, 0.00001572, -0.00001181, -0.00000384, 0.00001645, -0.00002326, 0.00002122, -0.00001725, 0.00001351, -0.00001032, },
		{ -0.00000034, 0.00000260, -0.00000607, 0.00001403, -0.00002424, 0.00002875, -0.00002534, 0.00001077, -0.00000326, 0.00000148, 0.00000047, -0.00000412, 0.00001162, -0.00002188, 0.00001690, -0.00000302, -0.00000897, 0.00002402, -0.00002396, 0.00001878, -0.00001819, 0.00001121, -0.00001240, 0.00001763, -0.00001791, 0.00000981, -0.00001357, 0.00002288, -0.00002385, 0.00002847, -0.00002718, 0.00001745, -0.00001933, 0.00002361, -0.00002170, 0.00000548, -0.00000878, 0.00002869, -0.00002975, 0.00003238, -0.00003772, 0.00002815, -0.00003243, 0.00004203, -0.00004567, 0.00004304, -0.00004877, 0.00006030, -0.00006594, 0.00004993, -0.00005901, 0.00007651, -0.00008747, 0.00007594, -0.00008741, 0.00010412, -0.00012272, 0.00012449, -0.00016303, 0.00016738, -0.00025238, 0.00027497, -0.00062292, 0.00043499, 0.00083173, -0.00076930, 0.00031290, -0.00027729, 0.00018231, -0.00017441, 0.00013014, -0.00012703, 0.00010544, -0.00008876, 0.00007886, -0.00008929, 0.00007608, -0.00005876, 0.00005161, -0.00006677, 0.00005871, -0.00004642, 0.00004084, -0.00004283, 0.00003691, -0.00002783, 0.00002698, -0.00003704, 0.00003195, -0.00002813, 0.00002323, -0.00000589, 0.00001104, -0.00002869, 0.00002663, -0.00002174, 0.00002173, -0.00003071, 0.00002866, -0.00002244, 0.00002106, -0.00001213, 0.00000984, -0.00001849, 0.00001784, -0.00001215, 0.00001191, -0.00001837, 0.00001579, -0.00001846, 0.00001451, 0.00000192, -0.00001178, 0.00001959, -0.00001718, 0.00000236, 0.00000745, -0.00000890, 0.00000999, -0.00001138, 0.00002036, -0.00003264, 0.00003236, -0.00002468, 0.00001277, -0.00000494, 0.00000099, 0.00000115, },
		{ -0.00000992, 0.00001211, -0.00001424, 0.00001701, -0.00001639, 0.00000726, 0.00000626, -0.00002032, 0.00002044, -0.00001627, 0.00001563, -0.00001664, 0.00001959, -0.00001715, 0.00000049, 0.00001443, -0.00002036, 0.00002284, -0.00001073, 0.00000209, -0.00000104, -0.00000450, -0.00000003, 0.00000335, 0.00000054, -0.00000563, -0.00000277, 0.00000811, -0.00000337, 0.00000306, 0.00000350, -0.00001154, 0.00000597, -0.00000354, 0.00000823, -0.00001751, 0.00000299, 0.00001413, -0.00000712, 0.00000520, -0.00000371, -0.00000563, -0.00000269, 0.00000895, -0.00000616, 0.00000139, -0.00000619, 0.00000919, -0.00000203, -0.00001314, -0.00000132, 0.00000866, -0.00000175, -0.00001528, 0.00000440, -0.00000038, 0.00000284, -0.00001749, 0.00000479, -0.00003166, 0.00000681, -0.00006770, 0.00000312, -0.00056125, 0.00138495, -0.00056125, 0.00000312, -0.00006770, 0.00000681, -0.00003166, 0.00000479, -0.00001749, 0.00000284, -0.00000038, 0.00000440, -0.00001528, -0.00000175, 0.00000866, -0.00000132, -0.00001314, -0.00000203, 0.00000919, -0.00000619, 0.00000139, -0.00000616, 0.00000895, -0.00000269, -0.00000563, -0.00000371, 0.00000520, -0.00000712, 0.00001413, 0.00000299, -0.00001751, 0.00000823, -0.00000354, 0.00000597, -0.00001154, 0.00000350, 0.00000306, -0.00000337, 0.00000811, -0.00000277, -0.00000563, 0.00000054, 0.00000335, -0.00000003, -0.00000450, -0.00000104, 0.00000209, -0.00001073, 0.00002284, -0.00002036, 0.00001443, 0.00000049, -0.00001715, 0.00001959, -0.00001664, 0.00001563, -0.00001627, 0.00002044, -0.00002032, 0.00000626, 0.00000726, -0.00001639, 0.00001701, -0.00001424, 0.00001211, },
		{ -0.00000034, 0.00000115, 0.00000099, -0.00000494, 0.00001277, -0.00002468, 0.00003236, -0.00003264, 0.00002036, -0.00001138, 0.00000999, -0.00000890, 0.00000745, 0.00000236, -0.00001718, 0.00001959, -0.00001178, 0.00000192, 0.00001451, -0.00001846, 0.00001579, -0.00001837, 0.00001191, -0.00001215, 0.00001784, -0.00001849, 0.00000984, -0.00001213, 0.00002106, -0.00002244, 0.00002866, -0.00003071, 0.00002173, -0.00002174, 0.00002663, -0.00002869, 0.00001104, -0.00000589, 0.00002323, -0.00002813, 0.00003195, -0.00003704, 0.00002698, -0.00002783, 0.00003691, -0.00004283, 0.00004084, -0.00004642, 0.00005871, -0.00006677, 0.00005161, -0.00005876, 0.00007608, -0.00008929, 0.00007886, -0.00008876, 0.00010544, -0.00012703, 0.00013014, -0.00017441, 0.00018231, -0.00027729, 0.00031290, -0.00076930, 0.00083173, 0.00043499, -0.00062292, 0.00027497, -0.00025238, 0.00016738, -0.00016303, 0.00012449, -0.00012272, 0.00010412, -0.00008741, 0.00007594, -0.00008747, 0.00007651, -0.00005901, 0.00004993, -0.00006594, 0.00006030, -0.00004877, 0.00004304, -0.00004567, 0.00004203, -0.00003243, 0.00002815, -0.00003772, 0.00003238, -0.00002975, 0.00002869, -0.00000878, 0.00000548, -0.00002170, 0.00002361, -0.00001933, 0.00001745, -0.00002718, 0.00002847, -0.00002385, 0.00002288, -0.00001357, 0.00000981, -0.00001791, 0.00001763, -0.00001240, 0.00001121, -0.00001819, 0.00001878, -0.00002396, 0.00002402, -0.00000897, -0.00000302, 0.00001690, -0.00002188, 0.00001162, -0.00000412, 0.00000047, 0.00000148, -0.00000326, 0.00001077, -0.00002534, 0.00002875, -0.00002424, 0.00001403, -0.00000607, 0.00000260, },
		{ 0.00001157, -0.00001032, 0.00001351, -0.00001725, 0.00002122, -0.00002326, 0.00001645, -0.00000384, -0.00001181, 0.00001572, -0.00001197, 0.00001189, -0.00001357, 0.00002183, -0.00002433, 0.00001178, 0.00000343, -0.00001395, 0.00002237, -0.00001553, 0.00000875, -0.00000933, 0.00000263, -0.00000602, 0.00001014, -0.00000633, 0.00000062, -0.00000869, 0.00001541, -0.00001242, 0.00001424, -0.00000975, 0.00000034, -0.00000382, 0.00000748, -0.00000471, -0.00000897, 0.00000056, 0.00001992, -0.00001933, 0.00001827, -0.00001643, 0.00000694, -0.00001408, 0.00002347, -0.00002487, 0.00002155, -0.00002779, 0.00003341, -0.00002920, 0.00001475, -0.00002923, 0.00003995, -0.00003754, 0.00002299, -0.00003576, 0.00004544, -0.00005126, 0.00004394, -0.00006923, 0.00005489, -0.00010208, 0.00007408, -0.00023109, -0.00024774, 0.00129788, -0.00074962, 0.00018557, -0.00018772, 0.00010246, -0.00010949, 0.00007348, -0.00007597, 0.00005526, -0.00004531, 0.00004364, -0.00005265, 0.00003354, -0.00002231, 0.00002614, -0.00004073, 0.00002543, -0.00001581, 0.00001655, -0.00002039, 0.00001510, -0.00000865, 0.00001081, -0.00001914, 0.00000933, -0.00000816, 0.00000681, 0.00000623, 0.00000462, -0.00002265, 0.00001841, -0.00001296, 0.00001386, -0.00002131, 0.00001569, -0.00000832, 0.00000632, 0.00000023, 0.00000348, -0.00001192, 0.00000724, -0.00000329, 0.00000586, -0.00001167, 0.00000854, -0.00000761, -0.00000121, 0.00001749, -0.00002141, 0.00002237, -0.00001155, -0.00000599, 0.00001133, -0.00001342, 0.00001414, -0.00001584, 0.00002028, -0.00002386, 0.00001235, 0.00000087, -0.00001155, 0.00001444, -0.00001247, },
		{ 0.00000390, -0.00000209, 0.00000349, -0.00000221, -0.00000168, 0.00000891, -0.00002191, 0.00003229, -0.00003656, 0.00002772, -0.00001790, 0.00001700, -0.00001706, 0.00001979, -0.00001052, -0.00000649, 0.00001506, -0.00001462, 0.00000962, 0.00000519, -0.00001090, 0.00000937, -0.00001473, 0.00000930, -0.00000874, 0.00001492, -0.00001610, 0.00000688, -0.00000740, 0.00001535, -0.00001623, 0.00002293, -0.00002812, 0.00002071, -0.00001912, 0.00002390, -0.00003062, 0.00001434, -0.00000150, 0.00001371, -0.00002064, 0.00002586, -0.00003104, 0.00002060, -0.00001725, 0.00002373, -0.00003029, 0.00002803, -0.00003256, 0.00004459, -0.00005408, 0.00003912, -0.00004350, 0.00005923, -0.00007308, 0.00006223, -0.00006838, 0.00008161, -0.00010213, 0.00010124, -0.00014470, 0.00014778, -0.00023944, 0.00026182, -0.00079586, 0.00114184, 0.00002271, -0.00040171, 0.00016961, -0.00017605, 0.00011015, -0.00011515, 0.00008697, -0.00009072, 0.00007852, -0.00006488, 0.00005393, -0.00006801, 0.00006082, -0.00004452, 0.00003423, -0.00005153, 0.00004902, -0.00003904, 0.00003376, -0.00003751, 0.00003736, -0.00002942, 0.00002308, -0.00003262, 0.00002696, -0.00002496, 0.00002910, -0.00000986, -0.00000117, -0.00001136, 0.00001613, -0.00001296, 0.00000907, -0.00001844, 0.00002266, -0.00002028, 0.00002056, -0.00001151, 0.00000667, -0.00001438, 0.00001439, -0.00000953, 0.00000710, -0.00001362, 0.00001683, -0.00002494, 0.00003056, -0.00001982, 0.00000924, 0.00000763, -0.00001915, 0.00001432, -0.00001125, 0.00000757, -0.00000548, 0.00000350, 0.00000200, -0.00001759, 0.00002397, -0.00002270, 0.00001453, -0.00000664, },
		{ -0.00001196, 0.00001172, -0.00001104, 0.00001493, -0.00002035, 0.00002574, -0.00003071, 0.00002685, -0.00001642, 0.00000056, 0.00000683, -0.00000347, 0.00000333, -0.00000391, 0.00001637, -0.00002438, 0.00001808, -0.00000642, -0.00000528, 0.00001761, -0.00001559, 0.00001117, -0.00001451, 0.00000786, -0.00001043, 0.00001547, -0.00001189, 0.00000535, -0.00001266, 0.00002026, -0.00001859, 0.00002244, -0.00002094, 0.00001138, -0.00001284, 0.00001740, -0.00001774, 0.00000125, -0.00000189, 0.00002238, -0.00002750, 0.00002832, -0.00002631, 0.00001611, -0.00002068, 0.00003139, -0.00003646, 0.00003476, -0.00004241, 0.00005053, -0.00004926, 0.00003564, -0.00004985, 0.00006327, -0.00006483, 0.00005230, -0.00006590, 0.00007967, -0.00009256, 0.00009129, -0.00012837, 0.00012611, -0.00019479, 0.00019921, -0.00046723, 0.00014320, 0.00105439, -0.00079473, 0.00028466, -0.00025753, 0.00016131, -0.00015832, 0.00011899, -0.00011588, 0.00009233, -0.00007737, 0.00007134, -0.00007953, 0.00005956, -0.00004486, 0.00004552, -0.00006074, 0.00004625, -0.00003503, 0.00003414, -0.00003777, 0.00003339, -0.00002480, 0.00002292, -0.00003069, 0.00001988, -0.00001884, 0.00001968, -0.00000224, 0.00000496, -0.00002348, 0.00002405, -0.00001878, 0.00001791, -0.00002665, 0.00002409, -0.00001722, 0.00001434, -0.00000628, 0.00000833, -0.00001669, 0.00001225, -0.00000821, 0.00000991, -0.00001677, 0.00001690, -0.00001790, 0.00001125, 0.00000674, -0.00001564, 0.00002389, -0.00001985, 0.00000464, 0.00000072, -0.00000635, 0.00000905, -0.00001205, 0.00001647, -0.00002323, 0.00001473, -0.00000277, -0.00000873, 0.00001353, },
		{ -0.00000809, 0.00000604, -0.00000497, 0.00000735, -0.00000765, 0.00000455, 0.00000141, -0.00001466, 0.00002682, -0.00003477, 0.00003019, -0.00002010, 0.00001955, -0.00002033, 0.00002844, -0.00002242, 0.00000638, 0.00000537, -0.00001149, 0.00001247, -0.00000146, -0.00000388, 0.00000168, -0.00000878, 0.00000441, -0.00000301, 0.00000962, -0.00001153, 0.00000187, -0.00000055, 0.00000714, -0.00000671, 0.00001255, -0.00002010, 0.00001474, -0.00001181, 0.00001551, -0.00002688, 0.00001479, 0.00000368, 0.00000194, -0.00000904, 0.00001560, -0.00002138, 0.00001131, -0.00000390, 0.00000636, -0.00001202, 0.00000849, -0.00001106, 0.00002184, -0.00003182, 0.00001644, -0.00001737, 0.00003040, -0.00004364, 0.00003132, -0.00003221, 0.00003927, -0.00005526, 0.00004542, -0.00008159, 0.00007128, -0.00014560, 0.00012269, -0.00068162, 0.00131045, -0.00033490, -0.00016172, 0.00003281, -0.00007281, 0.00002926, -0.00004555, 0.00002935, -0.00004055, 0.00003654, -0.00002794, 0.00001885, -0.00003633, 0.00003397, -0.00001990, 0.00000880, -0.00002739, 0.00002830, -0.00002027, 0.00001556, -0.00002019, 0.00002391, -0.00001934, 0.00001243, -0.00002274, 0.00001694, -0.00001465, 0.00002451, -0.00000927, -0.00000748, 0.00000001, 0.00000637, -0.00000470, -0.00000113, -0.00000676, 0.00001298, -0.00001294, 0.00001497, -0.00000673, 0.00000116, -0.00000870, 0.00000902, -0.00000452, 0.00000058, -0.00000536, 0.00000980, -0.00002001, 0.00003140, -0.00002688, 0.00002100, -0.00000506, -0.00001045, 0.00001015, -0.00001236, 0.00001039, -0.00000900, 0.00000698, -0.00000380, -0.00001171, 0.00002016, -0.00002190, 0.00001593, },
		{ 0.00001317, -0.00001175, 0.00001157, -0.00001094, 0.00001491, -0.00002169, 0.00002841, -0.00003631, 0.00003571, -0.00002851, 0.00001397, -0.00000373, 0.00000680, -0.00000745, 0.00001003, 0.00000434, -0.00001674, 0.00001700, -0.00001157, 0.00000216, 0.00001119, -0.00001236, 0.00000960, -0.00001591, 0.00001013, -0.00001212, 0.00001822, -0.00001505, 0.00000754, -0.00001366, 0.00002151, -0.00002041, 0.00002562, -0.00002745, 0.00001886, -0.00001856, 0.00002353, -0.00002797, 0.00001103, -0.00000356, 0.00002093, -0.00003021, 0.00003349, -0.00003168, 0.00002061, -0.00002142, 0.00003126, -0.00003863, 0.00003798, -0.00004645, 0.00005648, -0.00005766, 0.00004450, -0.00005773, 0.00007276, -0.00007724, 0.00006565, -0.00007826, 0.00009335, -0.00011042, 0.00011159, -0.00015648, 0.00016174, -0.00024505, 0.00027025, -0.00064931, 0.00054459, 0.00070301, -0.00071020, 0.00029311, -0.00026560, 0.00017227, -0.00016759, 0.00013123, -0.00012778, 0.00010546, -0.00008881, 0.00008056, -0.00008958, 0.00007050, -0.00005363, 0.00005173, -0.00006817, 0.00005553, -0.00004367, 0.00004184, -0.00004595, 0.00004393, -0.00003531, 0.00003037, -0.00003772, 0.00002570, -0.00002438, 0.00002894, -0.00000974, 0.00000389, -0.00002009, 0.00002458, -0.00002021, 0.00001738, -0.00002653, 0.00002712, -0.00002180, 0.00001897, -0.00000993, 0.00001051, -0.00001884, 0.00001456, -0.00001035, 0.00001089, -0.00001809, 0.00002158, -0.00002574, 0.00002354, -0.00000698, -0.00000404, 0.00001790, -0.00002139, 0.00001069, -0.00000811, 0.00000111, 0.00000307, -0.00000738, 0.00001141, -0.00002060, 0.00001510, -0.00000504, -0.00000666, },
		{ 0.00001891, -0.00001104, 0.00000936, -0.00000900, 0.00001245, -0.00001484, 0.00001324, -0.00000930, -0.00000325, 0.00001592, -0.00002667, 0.00002666, -0.00001666, 0.00001614, -0.00001643, 0.00002987, -0.00002886, 0.00001663, -0.00000553, -0.00000455, 0.00001080, -0.00000435, 0.00000069, -0.00000512, -0.00000240, -0.00000122, 0.00000366, 0.00000321, -0.00000607, -0.00000371, 0.00000671, -0.00000151, 0.00000365, 0.00000009, -0.00000883, 0.00000560, -0.00000158, 0.00000305, -0.00001816, 0.00001253, 0.00000860, -0.00000983, 0.00000415, 0.00000347, -0.00001040, 0.00000188, 0.00000848, -0.00001045, 0.00000668, -0.00001216, 0.00001220, -0.00000336, -0.00000647, -0.00000964, 0.00001242, -0.00000270, -0.00000936, -0.00000462, 0.00000926, -0.00000959, -0.00000016, -0.00002156, -0.00000289, -0.00002676, -0.00001991, -0.00007914, -0.00043662, 0.00131186, -0.00058465, 0.00004480, -0.00009655, 0.00002814, -0.00005161, 0.00002593, -0.00003167, 0.00001363, -0.00000967, 0.00001279, -0.00001979, -0.00000102, 0.00000383, 0.00000746, -0.00001937, -0.00000005, 0.00000419, 0.00000194, -0.00000637, 0.00000174, 0.00000516, -0.00000475, -0.00000160, -0.00001034, 0.00000480, -0.00000100, 0.00001598, -0.00000749, -0.00001211, 0.00001010, -0.00000327, 0.00000314, -0.00001054, 0.00000500, 0.00000201, -0.00000390, 0.00000785, -0.00000073, -0.00000532, -0.00000227, 0.00000300, 0.00000106, -0.00000651, 0.00000464, -0.00000084, -0.00000954, 0.00002535, -0.00002728, 0.00002807, -0.00001669, 0.00000056, 0.00000138, -0.00000826, 0.00000899, -0.00000900, 0.00000702, -0.00000612, -0.00000853, 0.00001817, -0.00002258, },
		{ -0.00000364, 0.00001176, -0.00001047, 0.00000993, -0.00000880, 0.00001208, -0.00001954, 0.00002714, -0.00003760, 0.00004009, -0.00003671, 0.00002493, -0.00001257, 0.00001538, -0.00001655, 0.00002402, -0.00001053, -0.00000384, 0.00000898, -0.00001054, 0.00000582, 0.00000580, -0.00000800, 0.00000537, -0.00001396, 0.00000935, -0.00001081, 0.00001798, -0.00001547, 0.00000697, -0.00001152, 0.00001904, -0.00001768, 0.00002314, -0.00002797, 0.00002111, -0.00001936, 0.00002400, -0.00003305, 0.00001840, -0.00000388, 0.00001576, -0.00002725, 0.00003316, -0.00003208, 0.00002049, -0.00001697, 0.00002394, -0.00003187, 0.00003120, -0.00003955, 0.00005059, -0.00005343, 0.00004009, -0.00005140, 0.00006676, -0.00007291, 0.00006102, -0.00007064, 0.00008390, -0.00010145, 0.00010012, -0.00014670, 0.00015195, -0.00023839, 0.00026225, -0.00072996, 0.00088591, 0.00031070, -0.00053502, 0.00022681, -0.00021865, 0.00013883, -0.00013891, 0.00011034, -0.00011095, 0.00009346, -0.00007833, 0.00007004, -0.00008157, 0.00006520, -0.00004760, 0.00004373, -0.00006184, 0.00005193, -0.00004025, 0.00003794, -0.00004278, 0.00004414, -0.00003759, 0.00003115, -0.00003883, 0.00002576, -0.00002351, 0.00003282, -0.00001511, 0.00000173, -0.00001347, 0.00002065, -0.00001772, 0.00001292, -0.00002144, 0.00002468, -0.00002152, 0.00001958, -0.00001015, 0.00000954, -0.00001801, 0.00001394, -0.00000947, 0.00000852, -0.00001492, 0.00002096, -0.00002844, 0.00003211, -0.00001987, 0.00001021, 0.00000603, -0.00001576, 0.00001009, -0.00001217, 0.00000605, -0.00000129, -0.00000419, 0.00000758, -0.00001840, 0.00001559, -0.00000778, },
		{ -0.00002408, 0.00002286, -0.00001492, 0.00001319, -0.00001323, 0.00001752, -0.00002215, 0.00002252, -0.00002124, 0.00001026, 0.00000156, -0.00001373, 0.00001816, -0.00000844, 0.00000754, -0.00000555, 0.00002295, -0.00002719, 0.00002034, -0.00001338, 0.00000280, 0.00000655, -0.00000394, 0.00000208, -0.00000956, 0.00000277, -0.00000602, 0.00000976, -0.00000288, -0.00000113, -0.00000834, 0.00001261, -0.00000845, 0.00001217, -0.00001136, 0.00000270, -0.00000402, 0.00000883, -0.00001074, -0.00000631, 0.00000838, 0.00001217, -0.00001941, 0.00001632, -0.00000810, -0.00000032, -0.00000533, 0.00001679, -0.00002256, 0.00002085, -0.00002845, 0.00003129, -0.00002460, 0.00001508, -0.00003178, 0.00003804, -0.00003157, 0.00002054, -0.00003540, 0.00004443, -0.00005150, 0.00004730, -0.00008087, 0.00006898, -0.00011911, 0.00010257, -0.00029929, -0.00010058, 0.00115488, -0.00070123, 0.00018188, -0.00018773, 0.00010200, -0.00011142, 0.00008086, -0.00007994, 0.00005770, -0.00004773, 0.00004634, -0.00005211, 0.00002896, -0.00002139, 0.00002981, -0.00004287, 0.00002345, -0.00001665, 0.00002117, -0.00002582, 0.00002249, -0.00001401, 0.00001053, -0.00001587, 0.00000169, -0.00000639, 0.00001304, 0.00000529, -0.00000515, -0.00001419, 0.00001718, -0.00001084, 0.00000869, -0.00001696, 0.00001421, -0.00000761, 0.00000454, 0.00000117, 0.00000473, -0.00001119, 0.00000343, -0.00000212, 0.00000544, -0.00001201, 0.00001372, -0.00001233, 0.00000424, 0.00001337, -0.00002021, 0.00002770, -0.00002312, 0.00000945, -0.00000820, -0.00000160, 0.00000512, -0.00000699, 0.00000503, -0.00000614, -0.00000725, 0.00001736, },
		{ -0.00001228, 0.00000159, 0.00000813, -0.00000718, 0.00000610, -0.00000409, 0.00000590, -0.00001318, 0.00002091, -0.00003327, 0.00003815, -0.00003855, 0.00003059, -0.00001676, 0.00001918, -0.00002018, 0.00003341, -0.00002327, 0.00000980, -0.00000295, -0.00000428, 0.00000496, 0.00000314, -0.00000453, 0.00000026, -0.00000989, 0.00000624, -0.00000700, 0.00001512, -0.00001351, 0.00000417, -0.00000699, 0.00001379, -0.00001148, 0.00001589, -0.00002283, 0.00001806, -0.00001509, 0.00001843, -0.00003180, 0.00002204, -0.00000270, 0.00000782, -0.00001955, 0.00002794, -0.00002818, 0.00001694, -0.00000943, 0.00001223, -0.00001912, 0.00001727, -0.00002445, 0.00003556, -0.00003925, 0.00002505, -0.00003359, 0.00004812, -0.00005490, 0.00004180, -0.00004696, 0.00005572, -0.00007047, 0.00006188, -0.00010370, 0.00010066, -0.00017720, 0.00017098, -0.00068337, 0.00110997, -0.00005374, -0.00032124, 0.00011720, -0.00013724, 0.00007638, -0.00008437, 0.00006581, -0.00007288, 0.00006239, -0.00005099, 0.00004418, -0.00005939, 0.00004721, -0.00003013, 0.00002468, -0.00004456, 0.00003791, -0.00002694, 0.00002425, -0.00002948, 0.00003435, -0.00003144, 0.00002507, -0.00003418, 0.00002059, -0.00001650, 0.00003075, -0.00001771, -0.00000091, -0.00000520, 0.00001378, -0.00001270, 0.00000617, -0.00001311, 0.00001810, -0.00001720, 0.00001670, -0.00000737, 0.00000583, -0.00001469, 0.00001096, -0.00000630, 0.00000360, -0.00000792, 0.00001493, -0.00002463, 0.00003419, -0.00002801, 0.00002272, -0.00000791, -0.00000527, 0.00000339, -0.00001065, 0.00000711, -0.00000280, -0.00000366, 0.00000634, -0.00001816, 0.00001762, },
		{ 0.00001607, -0.00002479, 0.00002622, -0.00001837, 0.00001618, -0.00001617, 0.00002074, -0.00002735, 0.00002983, -0.00003157, 0.00002269, -0.00001290, 0.00000085, 0.00000755, 0.00000181, -0.00000327, 0.00000947, 0.00000953, -0.00001774, 0.00001597, -0.00001522, 0.00000734, 0.00000232, -0.00000182, 0.00000075, -0.00001113, 0.00000573, -0.00000886, 0.00001410, -0.00000750, 0.00000221, -0.00001091, 0.00001592, -0.00001219, 0.00001679, -0.00001914, 0.00001163, -0.00001136, 0.00001649, -0.00002267, 0.00000614, 0.00000358, 0.00001349, -0.00002516, 0.00002534, -0.00001715, 0.00000727, -0.00000902, 0.00001945, -0.00002760, 0.00002737, -0.00003664, 0.00004200, -0.00003721, 0.00002772, -0.00004440, 0.00005344, -0.00004968, 0.00003897, -0.00005320, 0.00006436, -0.00007591, 0.00007437, -0.00011665, 0.00011421, -0.00018030, 0.00018616, -0.00048572, 0.00026651, 0.00087911, -0.00069000, 0.00023637, -0.00022540, 0.00013509, -0.00013774, 0.00010761, -0.00010473, 0.00008190, -0.00006884, 0.00006481, -0.00007093, 0.00004707, -0.00003572, 0.00004152, -0.00005621, 0.00003776, -0.00002902, 0.00003227, -0.00003755, 0.00003678, -0.00002880, 0.00002262, -0.00002730, 0.00001069, -0.00001382, 0.00002445, -0.00000548, -0.00000285, -0.00001343, 0.00002038, -0.00001527, 0.00001102, -0.00001924, 0.00001921, -0.00001395, 0.00001054, -0.00000348, 0.00000822, -0.00001520, 0.00000724, -0.00000518, 0.00000736, -0.00001418, 0.00001931, -0.00002139, 0.00001782, -0.00000154, -0.00000737, 0.00001966, -0.00002214, 0.00001276, -0.00001484, 0.00000443, 0.00000127, -0.00000520, 0.00000327, -0.00000596, -0.00000604, },
		{ 0.00002129, -0.00001856, 0.00000913, 0.00000212, -0.00000188, 0.00000036, 0.00000265, -0.00000287, -0.00000336, 0.00001035, -0.00002375, 0.00002998, -0.00003348, 0.00002991, -0.00001511, 0.00001684, -0.00001627, 0.00003489, -0.00002951, 0.00001923, -0.00001430, 0.00000427, 0.00000074, 0.00000350, -0.00000318, -0.00000408, -0.00000519, 0.00000204, -0.00000175, 0.00001059, -0.00001012, 0.00000022, -0.00000140, 0.00000742, -0.00000386, 0.00000607, -0.00001388, 0.00001114, -0.00000717, 0.00000804, -0.00002457, 0.00002160, -0.00000039, -0.00000138, -0.00000891, 0.00001934, -0.00002145, 0.00001178, -0.00000147, -0.00000020, -0.00000462, 0.00000072, -0.00000589, 0.00001636, -0.00002033, 0.00000487, -0.00001005, 0.00002297, -0.00002981, 0.00001532, -0.00001559, 0.00001848, -0.00002861, 0.00000968, -0.00004195, 0.00002424, -0.00008047, 0.00001531, -0.00051250, 0.00118579, -0.00033553, -0.00011961, 0.00000043, -0.00004781, 0.00000612, -0.00002170, 0.00001255, -0.00002614, 0.00002300, -0.00001611, 0.00001126, -0.00003046, 0.00002329, -0.00000760, 0.00000064, -0.00002200, 0.00001872, -0.00000869, 0.00000541, -0.00001011, 0.00001766, -0.00001897, 0.00001381, -0.00002544, 0.00001213, -0.00000514, 0.00002343, -0.00001754, -0.00000328, 0.00000299, 0.00000581, -0.00000689, -0.00000081, -0.00000389, 0.00000954, -0.00001054, 0.00001173, -0.00000282, 0.00000049, -0.00000995, 0.00000684, -0.00000219, -0.00000218, 0.00000101, 0.00000505, -0.00001491, 0.00002883, -0.00002878, 0.00002946, -0.00001932, 0.00000601, -0.00000657, -0.00000495, 0.00000486, -0.00000189, -0.00000537, 0.00000749, -0.00001992, },
		{ -0.00000306, 0.00001258, -0.00002291, 0.00002727, -0.00001984, 0.00001698, -0.00001646, 0.00002059, -0.00002856, 0.00003287, -0.00003760, 0.00003091, -0.00002385, 0.00001339, -0.00000161, 0.00001046, -0.00001216, 0.00002408, -0.00000613, -0.00000376, 0.00000496, -0.00001048, 0.00000725, 0.00000029, 0.00000015, -0.00000210, -0.00001021, 0.00000628, -0.00000928, 0.00001607, -0.00001006, 0.00000348, -0.00001104, 0.00001627, -0.00001228, 0.00001666, -0.00002171, 0.00001604, -0.00001449, 0.00001921, -0.00002998, 0.00001662, -0.00000063, 0.00001211, -0.00002628, 0.00002996, -0.00002254, 0.00001170, -0.00000919, 0.00001673, -0.00002550, 0.00002565, -0.00003575, 0.00004298, -0.00003957, 0.00002949, -0.00004527, 0.00005611, -0.00005428, 0.00004301, -0.00005493, 0.00006550, -0.00007846, 0.00007537, -0.00012106, 0.00012190, -0.00019484, 0.00020544, -0.00059263, 0.00059920, 0.00054353, -0.00058162, 0.00021742, -0.00021140, 0.00012733, -0.00012939, 0.00010410, -0.00010336, 0.00008335, -0.00007014, 0.00006551, -0.00007369, 0.00005101, -0.00003711, 0.00004064, -0.00005735, 0.00004078, -0.00003078, 0.00003297, -0.00003888, 0.00004142, -0.00003593, 0.00002873, -0.00003376, 0.00001504, -0.00001571, 0.00003092, -0.00001444, -0.00000102, -0.00001020, 0.00001979, -0.00001648, 0.00001027, -0.00001749, 0.00001964, -0.00001623, 0.00001324, -0.00000527, 0.00000901, -0.00001672, 0.00000871, -0.00000580, 0.00000642, -0.00001229, 0.00001982, -0.00002527, 0.00002747, -0.00001525, 0.00000754, 0.00000636, -0.00001427, 0.00000927, -0.00001626, 0.00000747, -0.00000061, -0.00000539, 0.00000361, -0.00000758, },
		{ -0.00002231, 0.00002533, -0.00002537, 0.00001780, -0.00000527, 0.00000442, -0.00000608, 0.00000986, -0.00001234, 0.00000781, -0.00000236, -0.00001128, 0.00001768, -0.00002314, 0.00002410, -0.00000865, 0.00000943, -0.00000543, 0.00002786, -0.00002720, 0.00002094, -0.00002087, 0.00001145, -0.00000442, 0.00000582, -0.00000408, -0.00000667, -0.00000115, -0.00000199, 0.00000370, 0.00000553, -0.00000640, -0.00000364, 0.00000381, 0.00000171, 0.00000288, -0.00000356, -0.00000387, 0.00000284, 0.00000188, -0.00000459, -0.00001311, 0.00001774, 0.00000230, -0.00001008, 0.00000249, 0.00000935, -0.00001360, 0.00000674, 0.00000456, -0.00001002, 0.00000748, -0.00001378, 0.00001108, -0.00000158, -0.00000253, -0.00001417, 0.00001249, -0.00000147, -0.00000543, -0.00000983, 0.00001365, -0.00001639, 0.00001060, -0.00004030, 0.00001926, -0.00005393, 0.00002190, -0.00016754, -0.00024843, 0.00111433, -0.00050469, 0.00003262, -0.00009330, 0.00002599, -0.00005213, 0.00003178, -0.00003430, 0.00001604, -0.00001313, 0.00001607, -0.00001955, -0.00000301, 0.00000096, 0.00001291, -0.00002156, -0.00000062, 0.00000053, 0.00000855, -0.00001274, 0.00000974, -0.00000117, -0.00000388, 0.00000026, -0.00001522, 0.00000318, 0.00000776, 0.00001262, -0.00001513, -0.00000474, 0.00000964, -0.00000159, -0.00000184, -0.00000623, 0.00000398, 0.00000131, -0.00000359, 0.00000642, 0.00000191, -0.00000506, -0.00000511, 0.00000292, 0.00000130, -0.00000683, 0.00000929, -0.00000582, -0.00000174, 0.00001739, -0.00002187, 0.00002824, -0.00002447, 0.00001379, -0.00001593, 0.00000211, 0.00000125, -0.00000026, -0.00000759, 0.00000939, },
		{ -0.00001185, 0.00000260, 0.00000597, -0.00001743, 0.00002489, -0.00001841, 0.00001487, -0.00001356, 0.00001654, -0.00002503, 0.00003058, -0.00003795, 0.00003303, -0.00002883, 0.00002110, -0.00000646, 0.00001450, -0.00001547, 0.00003363, -0.00001890, 0.00000974, -0.00000880, -0.00000109, 0.00000269, 0.00000140, 0.00000050, -0.00000505, -0.00000772, 0.00000487, -0.00000748, 0.00001571, -0.00001059, 0.00000285, -0.00000906, 0.00001421, -0.00000943, 0.00001239, -0.00001914, 0.00001551, -0.00001290, 0.00001621, -0.00003106, 0.00002324, -0.00000334, 0.00000820, -0.00002294, 0.00003000, -0.00002407, 0.00001325, -0.00000690, 0.00001040, -0.00001822, 0.00001754, -0.00002745, 0.00003584, -0.00003321, 0.00002190, -0.00003585, 0.00004750, -0.00004691, 0.00003444, -0.00004276, 0.00005041, -0.00006197, 0.00005314, -0.00009642, 0.00009334, -0.00016214, 0.00015329, -0.00059276, 0.00084236, 0.00021115, -0.00042083, 0.00014990, -0.00016110, 0.00008999, -0.00009518, 0.00007709, -0.00008100, 0.00006605, -0.00005486, 0.00005114, -0.00006272, 0.00004287, -0.00002766, 0.00002918, -0.00004806, 0.00003404, -0.00002328, 0.00002436, -0.00003037, 0.00003612, -0.00003448, 0.00002789, -0.00003465, 0.00001458, -0.00001179, 0.00003132, -0.00002027, 0.00000018, -0.00000535, 0.00001623, -0.00001518, 0.00000743, -0.00001290, 0.00001636, -0.00001486, 0.00001283, -0.00000433, 0.00000720, -0.00001589, 0.00000811, -0.00000444, 0.00000326, -0.00000691, 0.00001517, -0.00002272, 0.00003050, -0.00002386, 0.00001995, -0.00000794, -0.00000250, 0.00000033, -0.00001236, 0.00000682, 0.00000009, -0.00000807, 0.00000672, },
		{ 0.00000987, -0.00002328, 0.00002773, -0.00003066, 0.00002567, -0.00001233, 0.00001011, -0.00001150, 0.00001558, -0.00002015, 0.00001764, -0.00001428, 0.00000091, 0.00000464, -0.00001074, 0.00001607, -0.00000026, 0.00000010, 0.00000917, 0.00001469, -0.00001743, 0.00001423, -0.00002027, 0.00001430, -0.00000793, 0.00000835, -0.00000649, -0.00000728, 0.00000144, -0.00000486, 0.00000826, 0.00000098, -0.00000332, -0.00000644, 0.00000755, -0.00000203, 0.00000689, -0.00001051, 0.00000447, -0.00000416, 0.00000920, -0.00001629, -0.00000006, 0.00001182, 0.00000444, -0.00001670, 0.00001261, -0.00000014, -0.00000614, 0.00000297, 0.00000732, -0.00001519, 0.00001440, -0.00002288, 0.00002265, -0.00001406, 0.00000954, -0.00002703, 0.00002851, -0.00001928, 0.00001188, -0.00002672, 0.00003286, -0.00003958, 0.00003590, -0.00007400, 0.00006230, -0.00011129, 0.00009845, -0.00033213, 0.00005804, 0.00092558, -0.00055935, 0.00011831, -0.00014690, 0.00006974, -0.00008568, 0.00006441, -0.00006401, 0.00004393, -0.00003725, 0.00003768, -0.00004112, 0.00001654, -0.00001400, 0.00002590, -0.00003660, 0.00001441, -0.00001169, 0.00001981, -0.00002505, 0.00002481, -0.00001725, 0.00000979, -0.00001238, -0.00000615, -0.00000347, 0.00001908, 0.00000066, -0.00001135, -0.00000489, 0.00001382, -0.00000730, 0.00000154, -0.00000903, 0.00000899, -0.00000480, 0.00000196, 0.00000230, 0.00000546, -0.00000958, -0.00000126, 0.00000026, 0.00000303, -0.00000874, 0.00001447, -0.00001441, 0.00001122, 0.00000315, -0.00000945, 0.00001952, -0.00002189, 0.00001521, -0.00002132, 0.00000757, -0.00000145, 0.00000007, -0.00000820, },
		{ 0.00001172, -0.00001814, 0.00001040, -0.00000331, -0.00000858, 0.00001916, -0.00001418, 0.00001022, -0.00000806, 0.00000937, -0.00001755, 0.00002358, -0.00003308, 0.00002918, -0.00002740, 0.00002310, -0.00000601, 0.00001285, -0.00001144, 0.00003520, -0.00002477, 0.00001788, -0.00002046, 0.00000934, -0.00000439, 0.00000514, -0.00000128, -0.00000703, -0.00000476, 0.00000236, -0.00000416, 0.00001356, -0.00000961, 0.00000097, -0.00000584, 0.00001090, -0.00000520, 0.00000573, -0.00001292, 0.00001115, -0.00000763, 0.00000838, -0.00002592, 0.00002517, -0.00000424, 0.00000252, -0.00001614, 0.00002617, -0.00002234, 0.00001273, -0.00000377, 0.00000302, -0.00000887, 0.00000641, -0.00001522, 0.00002419, -0.00002196, 0.00000898, -0.00002037, 0.00003207, -0.00003234, 0.00001857, -0.00002289, 0.00002632, -0.00003488, 0.00001752, -0.00005386, 0.00004097, -0.00009650, 0.00004314, -0.00048439, 0.00096347, -0.00006588, -0.00025316, 0.00006450, -0.00009622, 0.00004037, -0.00004962, 0.00003875, -0.00004785, 0.00003866, -0.00003043, 0.00002823, -0.00004379, 0.00002793, -0.00001239, 0.00001195, -0.00003283, 0.00002175, -0.00001057, 0.00001029, -0.00001545, 0.00002339, -0.00002598, 0.00002110, -0.00003092, 0.00001065, -0.00000339, 0.00002598, -0.00002247, 0.00000085, 0.00000004, 0.00001093, -0.00001245, 0.00000391, -0.00000719, 0.00001102, -0.00001112, 0.00001033, -0.00000156, 0.00000356, -0.00001347, 0.00000632, -0.00000217, -0.00000067, 0.00000016, 0.00000696, -0.00001448, 0.00002620, -0.00002504, 0.00002614, -0.00001870, 0.00000883, -0.00001078, -0.00000507, 0.00000352, 0.00000245, -0.00001224, },
		{ -0.00000573, 0.00000726, -0.00002110, 0.00002672, -0.00003251, 0.00003081, -0.00001736, 0.00001372, -0.00001447, 0.00001823, -0.00002439, 0.00002383, -0.00002279, 0.00000981, -0.00000572, 0.00000022, 0.00000926, 0.00000653, -0.00000705, 0.00002290, -0.00000007, -0.00000405, 0.00000142, -0.00001282, 0.00001167, -0.00000804, 0.00000940, -0.00000922, -0.00000632, 0.00000239, -0.00000608, 0.00001129, -0.00000240, -0.00000143, -0.00000769, 0.00000938, -0.00000332, 0.00000738, -0.00001335, 0.00000923, -0.00000791, 0.00001255, -0.00002417, 0.00001174, 0.00000549, 0.00000521, -0.00002014, 0.00001991, -0.00000770, -0.00000005, 0.00000082, 0.00000680, -0.00001540, 0.00001535, -0.00002535, 0.00002728, -0.00001930, 0.00001381, -0.00003130, 0.00003536, -0.00002758, 0.00001908, -0.00003216, 0.00003843, -0.00004676, 0.00004179, -0.00008398, 0.00007694, -0.00013347, 0.00012612, -0.00043752, 0.00034970, 0.00066845, -0.00052131, 0.00014053, -0.00015836, 0.00008079, -0.00009159, 0.00007278, -0.00007285, 0.00005376, -0.00004571, 0.00004531, -0.00005029, 0.00002536, -0.00001909, 0.00002907, -0.00004213, 0.00002075, -0.00001562, 0.00002265, -0.00002871, 0.00003171, -0.00002697, 0.00001887, -0.00002162, -0.00000013, -0.00000588, 0.00002628, -0.00001016, -0.00000711, -0.00000366, 0.00001525, -0.00001087, 0.00000309, -0.00000915, 0.00001069, -0.00000795, 0.00000517, 0.00000023, 0.00000699, -0.00001237, 0.00000108, -0.00000074, 0.00000263, -0.00000731, 0.00001513, -0.00001816, 0.00002034, -0.00000971, 0.00000449, 0.00000625, -0.00001286, 0.00000980, -0.00002107, 0.00000958, -0.00000174, -0.00000223, },
		{ -0.00001590, 0.00001658, -0.00002456, 0.00001858, -0.00001355, 0.00000213, 0.00001126, -0.00000829, 0.00000430, -0.00000153, 0.00000096, -0.00000816, 0.00001397, -0.00002510, 0.00002140, -0.00002119, 0.00002070, -0.00000148, 0.00000683, -0.00000108, 0.00002870, -0.00002233, 0.00001772, -0.00002601, 0.00001695, -0.00001114, 0.00000989, -0.00000474, -0.00000755, -0.00000219, -0.00000033, -0.00000021, 0.00001044, -0.00000787, -0.00000131, -0.00000242, 0.00000764, -0.00000139, -0.00000102, -0.00000541, 0.00000509, -0.00000085, -0.00000204, -0.00001606, 0.00002278, -0.00000364, -0.00000377, -0.00000747, 0.00001980, -0.00001840, 0.00001113, -0.00000127, -0.00000305, -0.00000059, -0.00000417, -0.00000298, 0.00001223, -0.00001035, -0.00000435, -0.00000410, 0.00001541, -0.00001657, 0.00000205, -0.00000295, 0.00000221, -0.00000790, -0.00001855, -0.00000891, -0.00001634, -0.00002210, -0.00009493, -0.00029097, 0.00095819, -0.00025650, -0.00011366, -0.00001197, -0.00003713, -0.00000467, -0.00000741, 0.00000198, -0.00001518, 0.00001102, -0.00000554, 0.00000450, -0.00002377, 0.00001242, 0.00000279, -0.00000534, -0.00001712, 0.00000910, 0.00000224, -0.00000412, 0.00000092, 0.00000754, -0.00001369, 0.00001078, -0.00002462, 0.00000553, 0.00000699, 0.00001647, -0.00002135, 0.00000128, 0.00000492, 0.00000514, -0.00000937, 0.00000094, -0.00000207, 0.00000546, -0.00000666, 0.00000713, 0.00000177, -0.00000076, -0.00001046, 0.00000437, -0.00000023, -0.00000367, 0.00000659, -0.00000205, -0.00000310, 0.00001618, -0.00001890, 0.00002444, -0.00002268, 0.00001569, -0.00002020, 0.00000266, -0.00000033, 0.00000460, },
		{ -0.00000707, -0.00000008, 0.00000118, -0.00001519, 0.00002157, -0.00002995, 0.00003208, -0.00001939, 0.00001449, -0.00001442, 0.00001728, -0.00002435, 0.00002538, -0.00002662, 0.00001376, -0.00001122, 0.00000730, 0.00000614, 0.00000908, -0.00000871, 0.00003142, -0.00001150, 0.00000775, -0.00001299, -0.00000127, 0.00000452, -0.00000449, 0.00000802, -0.00001120, -0.00000451, 0.00000195, -0.00000564, 0.00001259, -0.00000439, -0.00000081, -0.00000747, 0.00000956, -0.00000262, 0.00000477, -0.00001200, 0.00000988, -0.00000770, 0.00001097, -0.00002644, 0.00002004, 0.00000019, 0.00000422, -0.00002002, 0.00002365, -0.00001255, 0.00000428, -0.00000008, 0.00000406, -0.00001192, 0.00001146, -0.00002211, 0.00002579, -0.00001807, 0.00001101, -0.00002764, 0.00003356, -0.00002692, 0.00001696, -0.00002726, 0.00003180, -0.00003955, 0.00002990, -0.00007132, 0.00006318, -0.00011871, 0.00009729, -0.00045835, 0.00057724, 0.00039700, -0.00042620, 0.00011694, -0.00013785, 0.00006660, -0.00007571, 0.00006129, -0.00006402, 0.00004787, -0.00004026, 0.00004033, -0.00004813, 0.00002447, -0.00001547, 0.00002355, -0.00003913, 0.00001924, -0.00001205, 0.00001769, -0.00002387, 0.00002979, -0.00002901, 0.00002188, -0.00002635, 0.00000230, -0.00000352, 0.00002794, -0.00001807, -0.00000314, -0.00000135, 0.00001422, -0.00001246, 0.00000326, -0.00000732, 0.00000961, -0.00000831, 0.00000603, 0.00000028, 0.00000639, -0.00001333, 0.00000198, -0.00000038, 0.00000064, -0.00000312, 0.00001134, -0.00001604, 0.00002315, -0.00001751, 0.00001552, -0.00000713, -0.00000086, -0.00000046, -0.00001572, 0.00000796, 0.00000046, },
		{ 0.00000467, -0.00001702, 0.00001907, -0.00002889, 0.00002489, -0.00002247, 0.00001257, 0.00000300, -0.00000235, -0.00000125, 0.00000419, -0.00000650, 0.00000063, 0.00000443, -0.00001696, 0.00001278, -0.00001318, 0.00001671, 0.00000427, -0.00000032, 0.00001224, 0.00001690, -0.00001335, 0.00000939, -0.00002372, 0.00001912, -0.00001498, 0.00001370, -0.00000881, -0.00000672, -0.00000053, -0.00000247, 0.00000353, 0.00000713, -0.00000608, -0.00000326, 0.00000040, 0.00000541, 0.00000056, -0.00000577, 0.00000100, -0.00000030, 0.00000488, -0.00001210, -0.00000404, 0.00001737, -0.00000231, -0.00000936, 0.00000136, 0.00001246, -0.00001339, 0.00000921, -0.00000022, -0.00000634, 0.00000449, -0.00001150, 0.00000627, 0.00000326, -0.00000206, -0.00001402, 0.00000850, 0.00000228, -0.00000464, -0.00000966, 0.00001081, -0.00001453, 0.00000998, -0.00004372, 0.00002405, -0.00006014, 0.00003520, -0.00022332, -0.00005383, 0.00084805, -0.00035402, -0.00002066, -0.00006312, 0.00000309, -0.00003391, 0.00002120, -0.00002378, 0.00000846, -0.00000915, 0.00001291, -0.00001383, -0.00000819, 0.00000130, 0.00001317, -0.00001808, -0.00000532, 0.00000037, 0.00001085, -0.00001430, 0.00001393, -0.00000687, -0.00000140, -0.00000016, -0.00001805, 0.00000165, 0.00001649, 0.00000509, -0.00001779, 0.00000180, 0.00000849, -0.00000015, -0.00000666, -0.00000072, 0.00000131, 0.00000112, -0.00000286, 0.00000453, 0.00000447, -0.00000473, -0.00000772, 0.00000308, 0.00000048, -0.00000446, 0.00001026, -0.00000884, 0.00000786, 0.00000380, -0.00000790, 0.00001586, -0.00001909, 0.00001579, -0.00002493, 0.00000811, -0.00000263, },
		{ 0.00000389, -0.00001307, 0.00000739, -0.00000727, -0.00000647, 0.00001298, -0.00002335, 0.00002958, -0.00001849, 0.00001274, -0.00001189, 0.00001346, -0.00002076, 0.00002289, -0.00002626, 0.00001301, -0.00001168, 0.00001001, 0.00000730, 0.00000674, -0.00000362, 0.00003235, -0.00001614, 0.00001347, -0.00002404, 0.00001031, -0.00000456, 0.00000154, 0.00000431, -0.00001181, -0.00000257, 0.00000067, -0.00000393, 0.00001240, -0.00000519, -0.00000115, -0.00000627, 0.00000883, -0.00000109, 0.00000047, -0.00000770, 0.00000726, -0.00000431, 0.00000509, -0.00002288, 0.00002370, -0.00000327, 0.00000159, -0.00001672, 0.00002391, -0.00001460, 0.00000692, -0.00000045, 0.00000068, -0.00000688, 0.00000504, -0.00001554, 0.00002065, -0.00001301, 0.00000396, -0.00001898, 0.00002618, -0.00002055, 0.00000919, -0.00001641, 0.00001817, -0.00002416, 0.00000759, -0.00004440, 0.00003040, -0.00007786, 0.00002193, -0.00039107, 0.00071027, 0.00015718, -0.00031176, 0.00007130, -0.00010162, 0.00004007, -0.00004897, 0.00003917, -0.00004523, 0.00003275, -0.00002644, 0.00002753, -0.00003885, 0.00001771, -0.00000685, 0.00001297, -0.00003099, 0.00001314, -0.00000416, 0.00000806, -0.00001334, 0.00002104, -0.00002436, 0.00001923, -0.00002692, 0.00000194, 0.00000265, 0.00002413, -0.00002221, 0.00000014, 0.00000154, 0.00001139, -0.00001257, 0.00000284, -0.00000470, 0.00000694, -0.00000679, 0.00000524, 0.00000184, 0.00000416, -0.00001291, 0.00000202, 0.00000052, -0.00000172, 0.00000220, 0.00000471, -0.00000895, 0.00001929, -0.00001834, 0.00002033, -0.00001635, 0.00000976, -0.00001201, -0.00000756, 0.00000413, },
		{ -0.00000227, 0.00000172, -0.00001444, 0.00001773, -0.00002945, 0.00002752, -0.00002809, 0.00002079, -0.00000394, 0.00000223, -0.00000508, 0.00000768, -0.00001133, 0.00000684, -0.00000280, -0.00001120, 0.00000622, -0.00000638, 0.00001414, 0.00000803, -0.00000477, 0.00002406, 0.00000433, -0.00000204, -0.00000403, -0.00001469, 0.00001529, -0.00001459, 0.00001506, -0.00001232, -0.00000502, 0.00000011, -0.00000364, 0.00000648, 0.00000424, -0.00000475, -0.00000446, 0.00000227, 0.00000454, 0.00000007, -0.00000735, 0.00000464, -0.00000321, 0.00000747, -0.00001905, 0.00000734, 0.00001073, -0.00000114, -0.00001314, 0.00000885, 0.00000554, -0.00000830, 0.00000738, -0.00000067, -0.00000665, 0.00000568, -0.00001451, 0.00001118, -0.00000123, 0.00000119, -0.00001801, 0.00001515, -0.00000487, 0.00000092, -0.00001395, 0.00001549, -0.00002045, 0.00001439, -0.00005200, 0.00003672, -0.00007875, 0.00005646, -0.00030790, 0.00017956, 0.00067168, -0.00037255, 0.00002470, -0.00008499, 0.00002088, -0.00004403, 0.00003271, -0.00003486, 0.00001943, -0.00001837, 0.00002162, -0.00002371, 0.00000028, -0.00000317, 0.00001664, -0.00002417, 0.00000048, -0.00000241, 0.00001309, -0.00001772, 0.00002043, -0.00001641, 0.00000776, -0.00000928, -0.00001297, 0.00000078, 0.00002262, -0.00000574, -0.00001292, 0.00000266, 0.00001032, -0.00000437, -0.00000458, -0.00000098, 0.00000258, -0.00000130, -0.00000054, 0.00000328, 0.00000578, -0.00000766, -0.00000574, 0.00000275, -0.00000024, -0.00000265, 0.00001005, -0.00001117, 0.00001504, -0.00000696, 0.00000395, 0.00000368, -0.00000979, 0.00000933, -0.00002388, 0.00000984, },
		{ 0.00000022, 0.00000660, -0.00001809, 0.00001443, -0.00001596, 0.00000310, 0.00000276, -0.00001422, 0.00002442, -0.00001567, 0.00000959, -0.00000823, 0.00000842, -0.00001543, 0.00001822, -0.00002364, 0.00000948, -0.00000876, 0.00000984, 0.00001124, 0.00000118, 0.00000679, 0.00002616, -0.00001337, 0.00001094, -0.00002814, 0.00001808, -0.00001247, 0.00000802, -0.00000083, -0.00001099, -0.00000102, -0.00000089, -0.00000150, 0.00001119, -0.00000522, -0.00000197, -0.00000471, 0.00000804, -0.00000002, -0.00000368, -0.00000239, 0.00000315, 0.00000043, -0.00000316, -0.00001475, 0.00002284, -0.00000478, -0.00000206, -0.00001121, 0.00002139, -0.00001430, 0.00000819, -0.00000096, -0.00000186, -0.00000242, -0.00000133, -0.00000841, 0.00001480, -0.00000739, -0.00000374, -0.00000917, 0.00001724, -0.00001279, 0.00000053, -0.00000515, 0.00000417, -0.00000858, -0.00001530, -0.00001520, -0.00000678, -0.00002988, -0.00007601, -0.00025359, 0.00074229, -0.00002234, -0.00020795, 0.00002504, -0.00006535, 0.00001410, -0.00002272, 0.00001646, -0.00002527, 0.00001609, -0.00001099, 0.00001287, -0.00002768, 0.00000982, 0.00000220, 0.00000175, -0.00002194, 0.00000653, 0.00000391, -0.00000201, -0.00000134, 0.00000911, -0.00001574, 0.00001278, -0.00002475, 0.00000052, 0.00001055, 0.00001619, -0.00002264, 0.00000267, 0.00000435, 0.00000761, -0.00001180, 0.00000252, -0.00000243, 0.00000400, -0.00000459, 0.00000385, 0.00000393, 0.00000114, -0.00001179, 0.00000192, 0.00000107, -0.00000310, 0.00000661, -0.00000218, 0.00000053, 0.00001051, -0.00001267, 0.00001784, -0.00001876, 0.00001534, -0.00002112, 0.00000044, },
		{ 0.00000808, 0.00000049, -0.00000382, -0.00000846, 0.00001248, -0.00002587, 0.00002584, -0.00002947, 0.00002566, -0.00000861, 0.00000479, -0.00000674, 0.00000855, -0.00001302, 0.00000975, -0.00000684, -0.00000897, 0.00000331, -0.00000256, 0.00001486, 0.00000771, -0.00000374, 0.00003056, -0.00000456, 0.00000658, -0.00001774, -0.00000227, 0.00000707, -0.00001022, 0.00001342, -0.00001441, -0.00000306, -0.00000010, -0.00000375, 0.00000836, 0.00000205, -0.00000404, -0.00000487, 0.00000331, 0.00000465, -0.00000242, -0.00000576, 0.00000504, -0.00000301, 0.00000602, -0.00002117, 0.00001574, 0.00000451, -0.00000085, -0.00001446, 0.00001402, 0.00000003, -0.00000381, 0.00000563, -0.00000196, -0.00000483, 0.00000372, -0.00001373, 0.00001221, -0.00000167, -0.00000018, -0.00001662, 0.00001602, -0.00000625, 0.00000053, -0.00001154, 0.00001206, -0.00001672, 0.00000659, -0.00004424, 0.00002909, -0.00007085, 0.00003566, -0.00032761, 0.00036800, 0.00047289, -0.00033863, 0.00003377, -0.00008367, 0.00002089, -0.00003886, 0.00003003, -0.00003330, 0.00001912, -0.00001756, 0.00002127, -0.00002564, 0.00000218, -0.00000167, 0.00001393, -0.00002423, 0.00000089, 0.00000013, 0.00000948, -0.00001448, 0.00001976, -0.00001964, 0.00001217, -0.00001527, -0.00001004, 0.00000345, 0.00002402, -0.00001413, -0.00000779, 0.00000390, 0.00001039, -0.00000731, -0.00000302, -0.00000030, 0.00000212, -0.00000186, 0.00000024, 0.00000349, 0.00000552, -0.00000938, -0.00000449, 0.00000312, -0.00000187, 0.00000110, 0.00000627, -0.00000839, 0.00001643, -0.00001282, 0.00001259, -0.00000776, 0.00000151, -0.00000128, -0.00001805, },
		{ 0.00000581, -0.00000185, 0.00000691, -0.00002023, 0.00001889, -0.00002262, 0.00001124, -0.00000681, -0.00000459, 0.00001823, -0.00001230, 0.00000642, -0.00000499, 0.00000397, -0.00001043, 0.00001357, -0.00002120, 0.00000583, -0.00000510, 0.00000942, 0.00001512, -0.00000433, 0.00001903, 0.00001589, -0.00000554, 0.00000114, -0.00002429, 0.00001981, -0.00001676, 0.00001289, -0.00000609, -0.00000915, -0.00000009, -0.00000223, 0.00000108, 0.00000949, -0.00000493, -0.00000282, -0.00000328, 0.00000778, -0.00000040, -0.00000606, 0.00000193, -0.00000043, 0.00000434, -0.00001117, -0.00000439, 0.00001861, -0.00000487, -0.00000578, -0.00000477, 0.00001722, -0.00001235, 0.00000844, -0.00000196, -0.00000276, 0.00000000, -0.00000579, -0.00000284, 0.00001055, -0.00000380, -0.00000917, -0.00000145, 0.00001016, -0.00000721, -0.00000512, 0.00000204, -0.00000489, 0.00000064, -0.00003035, 0.00000547, -0.00003431, 0.00000511, -0.00016648, -0.00007890, 0.00068867, -0.00013305, -0.00013143, -0.00000833, -0.00003927, -0.00000274, -0.00000481, 0.00000044, -0.00001078, 0.00000392, 0.00000068, 0.00000118, -0.00001885, 0.00000454, 0.00000810, -0.00000657, -0.00001537, 0.00000270, 0.00000875, -0.00000879, 0.00000809, -0.00000200, -0.00000643, 0.00000494, -0.00002164, -0.00000007, 0.00001765, 0.00000625, -0.00002015, 0.00000460, 0.00000650, 0.00000363, -0.00001061, 0.00000271, -0.00000125, 0.00000184, -0.00000276, 0.00000282, 0.00000564, -0.00000187, -0.00001059, 0.00000219, 0.00000068, -0.00000258, 0.00000839, -0.00000668, 0.00000911, 0.00000013, -0.00000315, 0.00000950, -0.00001421, 0.00001421, -0.00002523, },
		{ -0.00000992, 0.00000443, 0.00000420, -0.00001034, -0.00000064, 0.00000462, -0.00001917, 0.00002055, -0.00002694, 0.00002722, -0.00001105, 0.00000560, -0.00000674, 0.00000752, -0.00001233, 0.00001003, -0.00000836, -0.00000972, 0.00000377, -0.00000163, 0.00001883, 0.00000329, 0.00000339, 0.00003007, -0.00000713, 0.00000868, -0.00002697, 0.00000925, -0.00000256, -0.00000353, 0.00000931, -0.00001483, -0.00000128, -0.00000087, -0.00000295, 0.00000921, 0.00000058, -0.00000388, -0.00000471, 0.00000397, 0.00000488, -0.00000573, -0.00000207, 0.00000292, -0.00000033, 0.00000104, -0.00001817, 0.00001993, -0.00000017, -0.00000171, -0.00001328, 0.00001656, -0.00000359, -0.00000028, 0.00000382, -0.00000316, -0.00000232, 0.00000021, -0.00001076, 0.00001099, 0.00000010, -0.00000416, -0.00001192, 0.00001322, -0.00000409, -0.00000326, -0.00000557, 0.00000438, -0.00000801, -0.00000778, -0.00002700, 0.00000856, -0.00004516, -0.00001882, -0.00027960, 0.00048596, 0.00028929, -0.00028125, 0.00002336, -0.00007033, 0.00001214, -0.00002628, 0.00001991, -0.00002476, 0.00001231, -0.00001079, 0.00001531, -0.00002260, 0.00000023, 0.00000303, 0.00000776, -0.00002082, -0.00000160, 0.00000543, 0.00000258, -0.00000692, 0.00001354, -0.00001727, 0.00001188, -0.00001811, -0.00000878, 0.00000885, 0.00002067, -0.00001906, -0.00000310, 0.00000536, 0.00000899, -0.00000909, -0.00000166, 0.00000058, 0.00000076, -0.00000122, -0.00000002, 0.00000473, 0.00000401, -0.00001013, -0.00000362, 0.00000357, -0.00000338, 0.00000527, 0.00000056, -0.00000162, 0.00001201, -0.00001239, 0.00001526, -0.00001466, 0.00001063, -0.00001240, },
		{ -0.00002371, 0.00000748, -0.00000129, 0.00000419, -0.00001863, 0.00001950, -0.00002575, 0.00001625, -0.00001393, 0.00000378, 0.00001244, -0.00000948, 0.00000420, -0.00000317, 0.00000132, -0.00000721, 0.00001055, -0.00002079, 0.00000425, -0.00000305, 0.00001122, 0.00001618, -0.00000634, 0.00002897, 0.00000584, 0.00000299, -0.00001227, -0.00001432, 0.00001554, -0.00001649, 0.00001486, -0.00001031, -0.00000684, 0.00000020, -0.00000306, 0.00000335, 0.00000775, -0.00000468, -0.00000340, -0.00000218, 0.00000813, -0.00000249, -0.00000586, 0.00000393, -0.00000199, 0.00000564, -0.00001651, 0.00000558, 0.00001271, -0.00000436, -0.00000863, 0.00000132, 0.00001253, -0.00000956, 0.00000791, -0.00000329, -0.00000202, 0.00000004, -0.00000760, 0.00000028, 0.00000886, -0.00000335, -0.00001091, 0.00000257, 0.00000665, -0.00000549, -0.00000609, 0.00000332, -0.00000678, 0.00000067, -0.00003351, 0.00001177, -0.00004385, 0.00001306, -0.00022273, 0.00009511, 0.00057910, -0.00018367, -0.00008518, -0.00002490, -0.00002624, -0.00000804, 0.00000226, -0.00000620, -0.00000448, -0.00000113, 0.00000606, -0.00000523, -0.00001432, 0.00000349, 0.00000936, -0.00001048, -0.00001270, 0.00000305, 0.00000882, -0.00001034, 0.00001236, -0.00000935, 0.00000080, -0.00000216, -0.00001895, 0.00000154, 0.00002186, -0.00000343, -0.00001588, 0.00000618, 0.00000759, 0.00000003, -0.00000924, 0.00000339, -0.00000133, 0.00000094, -0.00000186, 0.00000273, 0.00000635, -0.00000432, -0.00000962, 0.00000297, -0.00000068, -0.00000004, 0.00000684, -0.00000702, 0.00001384, -0.00000821, 0.00000642, -0.00000129, -0.00000498, 0.00000708, },
		{ -0.00002058, -0.00000235, 0.00000097, 0.00000701, -0.00001578, 0.00000679, -0.00000371, -0.00001127, 0.00001337, -0.00002184, 0.00002638, -0.00001204, 0.00000556, -0.00000621, 0.00000599, -0.00001081, 0.00000932, -0.00000916, -0.00001155, 0.00000579, -0.00000179, 0.00002415, -0.00000316, 0.00001476, 0.00002363, -0.00000359, 0.00000307, -0.00002882, 0.00001629, -0.00001046, 0.00000325, 0.00000395, -0.00001383, 0.00000007, -0.00000188, -0.00000153, 0.00000924, -0.00000035, -0.00000404, -0.00000431, 0.00000475, 0.00000431, -0.00000840, 0.00000206, -0.00000019, 0.00000329, -0.00000577, -0.00001117, 0.00001986, -0.00000294, -0.00000349, -0.00001012, 0.00001672, -0.00000534, 0.00000217, 0.00000180, -0.00000350, -0.00000051, -0.00000312, -0.00000744, 0.00000948, 0.00000186, -0.00000824, -0.00000660, 0.00000954, -0.00000135, -0.00000712, 0.00000006, -0.00000286, -0.00000008, -0.00002142, -0.00000926, -0.00001375, -0.00001656, -0.00008757, -0.00017850, 0.00052755, 0.00014463, -0.00022363, 0.00000881, -0.00005579, 0.00000356, -0.00001430, 0.00000960, -0.00001567, 0.00000465, -0.00000300, 0.00000806, -0.00001834, -0.00000213, 0.00000760, 0.00000142, -0.00001708, -0.00000387, 0.00001026, -0.00000417, 0.00000145, 0.00000489, -0.00001152, 0.00000825, -0.00001873, -0.00000792, 0.00001521, 0.00001375, -0.00002045, 0.00000082, 0.00000670, 0.00000665, -0.00000992, -0.00000022, 0.00000097, -0.00000056, -0.00000024, -0.00000056, 0.00000627, 0.00000183, -0.00001034, -0.00000266, 0.00000347, -0.00000373, 0.00000812, -0.00000469, 0.00000660, 0.00000372, -0.00000652, 0.00001148, -0.00001503, 0.00001443, },
		{ -0.00000334, -0.00001791, 0.00000597, 0.00000144, -0.00000092, -0.00001380, 0.00001639, -0.00002510, 0.00001762, -0.00001779, 0.00000997, 0.00000775, -0.00000763, 0.00000313, -0.00000284, 0.00000059, -0.00000600, 0.00000945, -0.00002290, 0.00000553, -0.00000355, 0.00001639, 0.00001296, -0.00000267, 0.00003344, -0.00000023, 0.00000768, -0.00002439, -0.00000201, 0.00000737, -0.00001237, 0.00001372, -0.00001287, -0.00000453, -0.00000006, -0.00000325, 0.00000506, 0.00000624, -0.00000463, -0.00000365, -0.00000129, 0.00000872, -0.00000577, -0.00000325, 0.00000333, -0.00000105, 0.00000361, -0.00001767, 0.00001299, 0.00000682, -0.00000404, -0.00000993, 0.00000611, 0.00000828, -0.00000661, 0.00000676, -0.00000449, -0.00000037, -0.00000171, -0.00000720, 0.00000131, 0.00000930, -0.00000561, -0.00000930, 0.00000311, 0.00000643, -0.00000719, -0.00000313, -0.00000030, -0.00000278, -0.00000704, -0.00002607, 0.00000449, -0.00003544, -0.00000936, -0.00022916, 0.00023599, 0.00044752, -0.00019303, -0.00006228, -0.00002841, -0.00002303, -0.00000475, 0.00000086, -0.00000517, -0.00000515, 0.00000011, 0.00000574, -0.00000682, -0.00001357, 0.00000598, 0.00000669, -0.00001064, -0.00001328, 0.00000684, 0.00000475, -0.00000694, 0.00001104, -0.00001173, 0.00000445, -0.00000732, -0.00001722, 0.00000562, 0.00002203, -0.00001106, -0.00001092, 0.00000759, 0.00000748, -0.00000286, -0.00000774, 0.00000421, -0.00000233, 0.00000118, -0.00000191, 0.00000362, 0.00000592, -0.00000605, -0.00000885, 0.00000398, -0.00000243, 0.00000371, 0.00000254, -0.00000296, 0.00001324, -0.00001179, 0.00001244, -0.00001047, 0.00000518, },
		{ 0.00001184, -0.00002371, 0.00000250, -0.00000064, 0.00000747, -0.00001849, 0.00001188, -0.00001047, -0.00000420, 0.00000623, -0.00001584, 0.00002438, -0.00001258, 0.00000566, -0.00000625, 0.00000529, -0.00001005, 0.00000933, -0.00001117, -0.00001223, 0.00000704, -0.00000052, 0.00002801, -0.00000837, 0.00002683, 0.00001442, 0.00000318, -0.00000841, -0.00002312, 0.00001717, -0.00001449, 0.00000814, -0.00000130, -0.00001196, 0.00000094, -0.00000284, 0.00000014, 0.00000877, -0.00000099, -0.00000426, -0.00000389, 0.00000591, 0.00000237, -0.00000922, 0.00000501, -0.00000262, 0.00000597, -0.00001215, -0.00000226, 0.00001655, -0.00000411, -0.00000559, -0.00000587, 0.00001521, -0.00000558, 0.00000356, -0.00000033, -0.00000267, -0.00000021, -0.00000510, -0.00000510, 0.00000910, 0.00000196, -0.00001048, -0.00000284, 0.00000723, -0.00000035, -0.00000856, 0.00000243, -0.00000613, 0.00000265, -0.00002855, 0.00000131, -0.00002771, -0.00000006, -0.00014703, -0.00005076, 0.00050451, 0.00004656, -0.00017877, -0.00000047, -0.00004669, 0.00000065, -0.00000808, 0.00000401, -0.00001057, 0.00000033, 0.00000203, 0.00000290, -0.00001576, -0.00000239, 0.00000958, -0.00000264, -0.00001535, -0.00000362, 0.00001212, -0.00000792, 0.00000738, -0.00000292, -0.00000513, 0.00000321, -0.00001836, -0.00000609, 0.00002046, 0.00000518, -0.00001900, 0.00000400, 0.00000752, 0.00000395, -0.00001007, 0.00000139, 0.00000050, -0.00000117, 0.00000036, -0.00000070, 0.00000743, -0.00000042, -0.00001037, -0.00000135, 0.00000250, -0.00000233, 0.00000836, -0.00000724, 0.00001326, -0.00000527, 0.00000207, 0.00000308, -0.00000931, },
		{ 0.00001245, -0.00001351, -0.00001035, 0.00000290, 0.00000487, -0.00000683, -0.00000734, 0.00001089, -0.00002168, 0.00001604, -0.00001869, 0.00001401, 0.00000406, -0.00000637, 0.00000260, -0.00000319, 0.00000085, -0.00000588, 0.00000927, -0.00002649, 0.00000880, -0.00000580, 0.00002418, 0.00000608, 0.00000665, 0.00003152, -0.00000065, 0.00000563, -0.00003101, 0.00000845, -0.00000155, -0.00000631, 0.00001028, -0.00001375, -0.00000250, -0.00000071, -0.00000283, 0.00000612, 0.00000507, -0.00000482, -0.00000367, -0.00000033, 0.00000884, -0.00000916, 0.00000073, 0.00000086, 0.00000177, -0.00000122, -0.00001445, 0.00001667, 0.00000213, -0.00000437, -0.00000943, 0.00000913, 0.00000506, -0.00000399, 0.00000507, -0.00000501, 0.00000119, -0.00000408, -0.00000569, 0.00000143, 0.00001050, -0.00000904, -0.00000593, 0.00000173, 0.00000783, -0.00001032, 0.00000134, -0.00000581, 0.00000343, -0.00001790, -0.00001346, -0.00001010, -0.00001684, -0.00005402, -0.00018510, 0.00032441, 0.00032274, -0.00018198, -0.00005169, -0.00002618, -0.00002341, 0.00000143, -0.00000394, -0.00000082, -0.00000908, 0.00000451, 0.00000237, -0.00000586, -0.00001446, 0.00000979, 0.00000230, -0.00000911, -0.00001500, 0.00001186, -0.00000117, -0.00000072, 0.00000566, -0.00000977, 0.00000450, -0.00001038, -0.00001606, 0.00001136, 0.00001829, -0.00001569, -0.00000606, 0.00000878, 0.00000632, -0.00000492, -0.00000607, 0.00000472, -0.00000361, 0.00000204, -0.00000252, 0.00000515, 0.00000457, -0.00000719, -0.00000804, 0.00000471, -0.00000368, 0.00000726, -0.00000281, 0.00000413, 0.00000767, -0.00000974, 0.00001284, -0.00001485, },
		{ -0.00000009, 0.00000406, -0.00002162, 0.00000388, 0.00000013, 0.00000522, -0.00001792, 0.00001370, -0.00001443, 0.00000066, 0.00000060, -0.00001028, 0.00002220, -0.00001332, 0.00000643, -0.00000737, 0.00000607, -0.00001085, 0.00001097, -0.00001554, -0.00001028, 0.00000583, 0.00000415, 0.00002800, -0.00000922, 0.00003585, 0.00000635, 0.00000890, -0.00002165, -0.00001235, 0.00001255, -0.00001414, 0.00001013, -0.00000541, -0.00000976, 0.00000134, -0.00000354, 0.00000174, 0.00000813, -0.00000157, -0.00000440, -0.00000351, 0.00000733, -0.00000091, -0.00000773, 0.00000580, -0.00000318, 0.00000631, -0.00001604, 0.00000621, 0.00001155, -0.00000439, -0.00000727, -0.00000150, 0.00001287, -0.00000486, 0.00000399, -0.00000228, -0.00000095, -0.00000146, -0.00000538, -0.00000416, 0.00001028, -0.00000014, -0.00001010, -0.00000156, 0.00000722, -0.00000197, -0.00000675, 0.00000068, -0.00000440, -0.00000116, -0.00002699, 0.00000133, -0.00002823, -0.00000506, -0.00017725, 0.00007401, 0.00043972, -0.00001066, -0.00014932, -0.00000212, -0.00004427, 0.00000423, -0.00000861, 0.00000441, -0.00001089, 0.00000083, 0.00000284, 0.00000114, -0.00001591, 0.00000026, 0.00000821, -0.00000364, -0.00001635, -0.00000015, 0.00001012, -0.00000743, 0.00000899, -0.00000754, -0.00000036, -0.00000157, -0.00001794, -0.00000246, 0.00002292, -0.00000306, -0.00001570, 0.00000661, 0.00000752, 0.00000139, -0.00000972, 0.00000305, -0.00000080, -0.00000083, 0.00000023, -0.00000005, 0.00000776, -0.00000235, -0.00001040, 0.00000030, 0.00000081, 0.00000066, 0.00000569, -0.00000582, 0.00001591, -0.00001171, 0.00000986, -0.00000656, },
		{ -0.00001312, 0.00001428, -0.00002034, -0.00000368, 0.00000019, 0.00000735, -0.00001169, -0.00000124, 0.00000491, -0.00001718, 0.00001293, -0.00001774, 0.00001662, 0.00000076, -0.00000496, 0.00000164, -0.00000298, 0.00000071, -0.00000533, 0.00000824, -0.00002958, 0.00001203, -0.00000762, 0.00003227, -0.00000200, 0.00001937, 0.00002474, 0.00000352, -0.00000327, -0.00003007, 0.00001394, -0.00000829, -0.00000055, 0.00000586, -0.00001338, -0.00000082, -0.00000161, -0.00000192, 0.00000662, 0.00000419, -0.00000516, -0.00000358, 0.00000094, 0.00000786, -0.00001145, 0.00000462, -0.00000214, 0.00000510, -0.00000733, -0.00000791, 0.00001655, -0.00000089, -0.00000535, -0.00000740, 0.00001039, 0.00000302, -0.00000203, 0.00000304, -0.00000451, 0.00000178, -0.00000591, -0.00000431, 0.00000190, 0.00001095, -0.00001186, -0.00000268, 0.00000038, 0.00000880, -0.00001259, 0.00000455, -0.00000986, 0.00000769, -0.00002655, -0.00000243, -0.00002347, 0.00000013, -0.00010463, -0.00010371, 0.00035684, 0.00022244, -0.00016684, -0.00004365, -0.00002491, -0.00002170, 0.00000513, -0.00000712, 0.00000232, -0.00001224, 0.00000852, -0.00000095, -0.00000502, -0.00001456, 0.00001248, -0.00000141, -0.00000821, -0.00001551, 0.00001556, -0.00000619, 0.00000541, -0.00000113, -0.00000537, 0.00000199, -0.00001193, -0.00001454, 0.00001721, 0.00001176, -0.00001728, -0.00000169, 0.00000955, 0.00000449, -0.00000622, -0.00000419, 0.00000453, -0.00000452, 0.00000290, -0.00000309, 0.00000675, 0.00000274, -0.00000800, -0.00000690, 0.00000472, -0.00000364, 0.00000915, -0.00000705, 0.00001176, -0.00000076, -0.00000327, 0.00000769, },
		{ -0.00001380, 0.00000898, -0.00000606, -0.00001587, 0.00000250, 0.00000270, 0.00000101, -0.00001470, 0.00001253, -0.00001555, 0.00000304, -0.00000304, -0.00000571, 0.00002015, -0.00001426, 0.00000766, -0.00000924, 0.00000798, -0.00001291, 0.00001391, -0.00002205, -0.00000569, 0.00000200, 0.00001269, 0.00002326, -0.00000412, 0.00003935, 0.00000233, 0.00000976, -0.00003195, -0.00000053, 0.00000487, -0.00001049, 0.00000934, -0.00000797, -0.00000757, 0.00000130, -0.00000383, 0.00000304, 0.00000753, -0.00000225, -0.00000439, -0.00000304, 0.00000857, -0.00000492, -0.00000427, 0.00000434, -0.00000159, 0.00000378, -0.00001621, 0.00001233, 0.00000644, -0.00000450, -0.00000795, 0.00000217, 0.00001047, -0.00000374, 0.00000360, -0.00000365, 0.00000099, -0.00000366, -0.00000441, -0.00000420, 0.00001246, -0.00000382, -0.00000761, -0.00000235, 0.00000903, -0.00000551, -0.00000268, -0.00000395, 0.00000081, -0.00000969, -0.00001866, -0.00000743, -0.00001697, -0.00003199, -0.00016803, 0.00017150, 0.00035876, -0.00004031, -0.00013079, 0.00000092, -0.00004555, 0.00001137, -0.00001335, 0.00000884, -0.00001510, 0.00000500, 0.00000033, 0.00000204, -0.00001798, 0.00000486, 0.00000444, -0.00000245, -0.00001914, 0.00000554, 0.00000520, -0.00000329, 0.00000625, -0.00000815, 0.00000160, -0.00000512, -0.00001773, 0.00000293, 0.00002188, -0.00000947, -0.00001155, 0.00000874, 0.00000660, -0.00000067, -0.00000898, 0.00000448, -0.00000257, 0.00000029, -0.00000057, 0.00000136, 0.00000718, -0.00000384, -0.00001036, 0.00000195, -0.00000097, 0.00000434, 0.00000094, -0.00000059, 0.00001356, -0.00001348, 0.00001382, },
		{ -0.00000090, -0.00000627, 0.00001021, -0.00002224, 0.00000030, -0.00000080, 0.00000776, -0.00001423, 0.00000295, 0.00000010, -0.00001318, 0.00000980, -0.00001614, 0.00001865, -0.00000279, -0.00000273, -0.00000049, -0.00000128, -0.00000100, -0.00000306, 0.00000480, -0.00003028, 0.00001310, -0.00000653, 0.00003780, -0.00000800, 0.00003201, 0.00001626, 0.00000906, -0.00001635, -0.00002233, 0.00001349, -0.00001118, 0.00000322, 0.00000175, -0.00001231, 0.00000053, -0.00000260, -0.00000075, 0.00000676, 0.00000344, -0.00000551, -0.00000346, 0.00000257, 0.00000548, -0.00001177, 0.00000711, -0.00000423, 0.00000736, -0.00001275, 0.00000007, 0.00001357, -0.00000241, -0.00000659, -0.00000443, 0.00001031, 0.00000203, -0.00000085, 0.00000094, -0.00000302, 0.00000100, -0.00000652, -0.00000381, 0.00000353, 0.00000967, -0.00001284, -0.00000096, 0.00000046, 0.00000792, -0.00001248, 0.00000474, -0.00001029, 0.00000726, -0.00002926, 0.00000189, -0.00002868, 0.00000457, -0.00014276, -0.00000669, 0.00034343, 0.00015182, -0.00015615, -0.00003289, -0.00002798, -0.00001511, 0.00000356, -0.00000580, 0.00000146, -0.00001195, 0.00000969, -0.00000207, -0.00000609, -0.00001231, 0.00001249, -0.00000284, -0.00000944, -0.00001330, 0.00001621, -0.00000822, 0.00000891, -0.00000667, -0.00000079, -0.00000157, -0.00001276, -0.00001176, 0.00002148, 0.00000415, -0.00001640, 0.00000214, 0.00000962, 0.00000248, -0.00000691, -0.00000219, 0.00000351, -0.00000463, 0.00000324, -0.00000315, 0.00000789, 0.00000086, -0.00000869, -0.00000532, 0.00000388, -0.00000199, 0.00000847, -0.00000831, 0.00001721, -0.00000905, 0.00000488, },
		{ 0.00001252, -0.00001609, 0.00001455, -0.00001521, -0.00000893, -0.00000007, 0.00000572, -0.00000370, -0.00001033, 0.00000965, -0.00001479, 0.00000362, -0.00000508, -0.00000192, 0.00001792, -0.00001484, 0.00000856, -0.00001083, 0.00000986, -0.00001501, 0.00001677, -0.00002918, 0.00000007, -0.00000296, 0.00002372, 0.00001494, 0.00000637, 0.00003707, 0.00000303, 0.00000395, -0.00003584, 0.00000843, -0.00000274, -0.00000556, 0.00000675, -0.00000916, -0.00000551, 0.00000085, -0.00000366, 0.00000392, 0.00000711, -0.00000308, -0.00000426, -0.00000231, 0.00000907, -0.00000863, 0.00000010, 0.00000139, 0.00000152, -0.00000102, -0.00001262, 0.00001510, 0.00000231, -0.00000494, -0.00000733, 0.00000468, 0.00000853, -0.00000268, 0.00000256, -0.00000407, 0.00000236, -0.00000588, -0.00000308, -0.00000432, 0.00001452, -0.00000782, -0.00000432, -0.00000390, 0.00001126, -0.00000934, 0.00000166, -0.00000899, 0.00000644, -0.00001906, -0.00000824, -0.00001934, -0.00000123, -0.00007238, -0.00012139, 0.00022850, 0.00028250, -0.00005670, -0.00011615, 0.00000368, -0.00004596, 0.00001763, -0.00001819, 0.00001367, -0.00002007, 0.00001019, -0.00000326, 0.00000365, -0.00002011, 0.00000948, 0.00000017, -0.00000087, -0.00002185, 0.00001142, -0.00000057, 0.00000246, 0.00000070, -0.00000550, 0.00000080, -0.00000730, -0.00001735, 0.00000916, 0.00001765, -0.00001333, -0.00000723, 0.00001033, 0.00000492, -0.00000209, -0.00000794, 0.00000537, -0.00000430, 0.00000173, -0.00000167, 0.00000319, 0.00000588, -0.00000498, -0.00001003, 0.00000317, -0.00000207, 0.00000739, -0.00000416, 0.00000685, 0.00000699, -0.00001021, },
		{ 0.00001155, -0.00000968, 0.00000293, 0.00000181, -0.00001943, 0.00000113, 0.00000027, 0.00000597, -0.00001420, 0.00000467, -0.00000274, -0.00001060, 0.00000755, -0.00001469, 0.00002054, -0.00000678, 0.00000040, -0.00000386, 0.00000207, -0.00000454, 0.00000128, -0.00000161, -0.00002771, 0.00001086, -0.00000094, 0.00003867, -0.00000915, 0.00004128, 0.00000941, 0.00001204, -0.00002924, -0.00001084, 0.00000835, -0.00001019, 0.00000437, -0.00000128, -0.00001097, 0.00000153, -0.00000348, 0.00000042, 0.00000677, 0.00000266, -0.00000577, -0.00000328, 0.00000434, 0.00000193, -0.00000989, 0.00000745, -0.00000448, 0.00000741, -0.00001576, 0.00000742, 0.00000913, -0.00000303, -0.00000755, -0.00000126, 0.00000947, 0.00000174, -0.00000044, -0.00000084, -0.00000096, -0.00000096, -0.00000585, -0.00000430, 0.00000635, 0.00000655, -0.00001169, -0.00000115, 0.00000232, 0.00000493, -0.00000980, 0.00000177, -0.00000694, 0.00000188, -0.00002533, -0.00000189, -0.00002331, -0.00000904, -0.00015413, 0.00008298, 0.00030229, 0.00010613, -0.00015099, -0.00001917, -0.00003479, -0.00000441, -0.00000282, -0.00000005, -0.00000363, -0.00000780, 0.00000754, -0.00000057, -0.00000933, -0.00000760, 0.00000970, -0.00000182, -0.00001289, -0.00000827, 0.00001350, -0.00000660, 0.00000858, -0.00000920, 0.00000220, -0.00000490, -0.00001342, -0.00000733, 0.00002297, -0.00000288, -0.00001392, 0.00000547, 0.00000876, 0.00000076, -0.00000717, -0.00000022, 0.00000180, -0.00000387, 0.00000289, -0.00000245, 0.00000826, -0.00000079, -0.00000932, -0.00000343, 0.00000247, 0.00000091, 0.00000526, -0.00000578, 0.00001852, -0.00001441, },
		{ -0.00000338, 0.00000653, -0.00001279, 0.00001471, -0.00002074, -0.00000316, -0.00000217, 0.00000782, -0.00000741, -0.00000643, 0.00000661, -0.00001353, 0.00000358, -0.00000642, 0.00000169, 0.00001492, -0.00001430, 0.00000820, -0.00001101, 0.00001037, -0.00001564, 0.00001775, -0.00003489, 0.00000478, -0.00000654, 0.00003459, 0.00000573, 0.00001981, 0.00003080, 0.00000673, -0.00000754, -0.00003226, 0.00001210, -0.00000776, -0.00000140, 0.00000366, -0.00000949, -0.00000355, 0.00000004, -0.00000307, 0.00000442, 0.00000683, -0.00000401, -0.00000403, -0.00000119, 0.00000835, -0.00001103, 0.00000408, -0.00000178, 0.00000489, -0.00000670, -0.00000630, 0.00001453, -0.00000035, -0.00000579, -0.00000556, 0.00000595, 0.00000724, -0.00000197, 0.00000117, -0.00000344, 0.00000259, -0.00000731, -0.00000221, -0.00000363, 0.00001538, -0.00001082, -0.00000163, -0.00000481, 0.00001242, -0.00001179, 0.00000427, -0.00001201, 0.00000947, -0.00002534, -0.00000085, -0.00002783, 0.00000943, -0.00011216, -0.00004992, 0.00024449, 0.00022277, -0.00007024, -0.00009982, 0.00000228, -0.00004200, 0.00001947, -0.00001964, 0.00001563, -0.00002279, 0.00001372, -0.00000560, 0.00000397, -0.00002046, 0.00001222, -0.00000273, -0.00000067, -0.00002265, 0.00001545, -0.00000491, 0.00000738, -0.00000534, -0.00000125, -0.00000188, -0.00000848, -0.00001609, 0.00001483, 0.00001140, -0.00001469, -0.00000305, 0.00001117, 0.00000289, -0.00000292, -0.00000666, 0.00000549, -0.00000549, 0.00000298, -0.00000259, 0.00000497, 0.00000421, -0.00000594, -0.00000923, 0.00000365, -0.00000192, 0.00000863, -0.00000767, 0.00001410, -0.00000164, },
		{ -0.00001522, 0.00001428, -0.00001548, 0.00001108, -0.00000805, -0.00001359, -0.00000037, 0.00000273, 0.00000277, -0.00001235, 0.00000439, -0.00000384, -0.00000937, 0.00000627, -0.00001351, 0.00002217, -0.00001080, 0.00000386, -0.00000777, 0.00000632, -0.00000917, 0.00000687, -0.00001020, -0.00002249, 0.00000578, 0.00000910, 0.00003444, -0.00000432, 0.00004536, 0.00000615, 0.00000953, -0.00003770, 0.00000034, 0.00000119, -0.00000672, 0.00000332, -0.00000309, -0.00000953, 0.00000211, -0.00000407, 0.00000140, 0.00000685, 0.00000170, -0.00000587, -0.00000295, 0.00000579, -0.00000213, -0.00000631, 0.00000574, -0.00000277, 0.00000494, -0.00001541, 0.00001251, 0.00000462, -0.00000340, -0.00000773, 0.00000148, 0.00000843, 0.00000173, -0.00000070, -0.00000194, 0.00000100, -0.00000346, -0.00000438, -0.00000529, 0.00000971, 0.00000230, -0.00000904, -0.00000267, 0.00000530, 0.00000067, -0.00000567, -0.00000298, -0.00000148, -0.00000636, -0.00001714, -0.00001120, -0.00001036, -0.00003870, -0.00013315, 0.00014777, 0.00025258, 0.00007563, -0.00014769, -0.00000560, -0.00004211, 0.00000720, -0.00001114, 0.00000777, -0.00001100, -0.00000133, 0.00000332, 0.00000247, -0.00001362, -0.00000166, 0.00000526, 0.00000058, -0.00001742, -0.00000161, 0.00000857, -0.00000225, 0.00000476, -0.00000833, 0.00000276, -0.00000730, -0.00001392, -0.00000164, 0.00002138, -0.00000821, -0.00001061, 0.00000828, 0.00000701, -0.00000037, -0.00000716, 0.00000149, -0.00000025, -0.00000245, 0.00000196, -0.00000108, 0.00000783, -0.00000213, -0.00000978, -0.00000158, 0.00000110, 0.00000410, 0.00000054, 0.00000001, 0.00001518, },
		{ -0.00000952, 0.00000436, -0.00000190, -0.00000532, 0.00000952, -0.00002153, -0.00000001, -0.00000271, 0.00000823, -0.00000929, -0.00000409, 0.00000460, -0.00001287, 0.00000392, -0.00000788, 0.00000566, 0.00001077, -0.00001222, 0.00000608, -0.00000909, 0.00000865, -0.00001381, 0.00001557, -0.00003754, 0.00000650, -0.00000630, 0.00004245, -0.00000118, 0.00003297, 0.00002343, 0.00001009, -0.00002136, -0.00002288, 0.00001028, -0.00000907, 0.00000069, 0.00000113, -0.00000941, -0.00000169, -0.00000101, -0.00000222, 0.00000469, 0.00000657, -0.00000496, -0.00000373, 0.00000023, 0.00000629, -0.00001147, 0.00000655, -0.00000396, 0.00000718, -0.00001155, 0.00000104, 0.00001149, -0.00000169, -0.00000681, -0.00000305, 0.00000622, 0.00000654, -0.00000178, -0.00000026, -0.00000195, 0.00000152, -0.00000752, -0.00000226, -0.00000170, 0.00001446, -0.00001203, -0.00000041, -0.00000423, 0.00001166, -0.00001195, 0.00000414, -0.00001176, 0.00000833, -0.00002628, 0.00000034, -0.00002837, 0.00000730, -0.00013689, 0.00002814, 0.00022913, 0.00018183, -0.00008538, -0.00007986, -0.00000427, -0.00003278, 0.00001572, -0.00001625, 0.00001317, -0.00002168, 0.00001413, -0.00000543, 0.00000196, -0.00001812, 0.00001214, -0.00000327, -0.00000276, -0.00002058, 0.00001645, -0.00000635, 0.00000956, -0.00000972, 0.00000269, -0.00000524, -0.00000917, -0.00001343, 0.00001861, 0.00000461, -0.00001409, 0.00000090, 0.00001098, 0.00000101, -0.00000335, -0.00000523, 0.00000482, -0.00000587, 0.00000369, -0.00000297, 0.00000629, 0.00000250, -0.00000684, -0.00000796, 0.00000339, -0.00000041, 0.00000746, -0.00000811, 0.00001877, },
		{ 0.00000823, -0.00001154, 0.00001220, -0.00001640, 0.00001533, -0.00001632, -0.00000706, -0.00000277, 0.00000545, -0.00000059, -0.00000999, 0.00000331, -0.00000413, -0.00000873, 0.00000539, -0.00001211, 0.00002296, -0.00001406, 0.00000667, -0.00001103, 0.00001013, -0.00001343, 0.00001204, -0.00001917, -0.00001646, -0.00000017, 0.00002180, 0.00002655, 0.00000565, 0.00004442, 0.00000627, 0.00000094, -0.00003915, 0.00000779, -0.00000505, -0.00000279, 0.00000118, -0.00000403, -0.00000794, 0.00000218, -0.00000426, 0.00000205, 0.00000712, 0.00000049, -0.00000577, -0.00000236, 0.00000646, -0.00000575, -0.00000203, 0.00000274, 0.00000028, 0.00000054, -0.00001180, 0.00001453, 0.00000104, -0.00000397, -0.00000689, 0.00000338, 0.00000757, 0.00000168, -0.00000144, -0.00000212, 0.00000222, -0.00000573, -0.00000284, -0.00000601, 0.00001267, -0.00000199, -0.00000599, -0.00000443, 0.00000819, -0.00000345, -0.00000176, -0.00000747, 0.00000353, -0.00001420, -0.00000874, -0.00002102, 0.00000329, -0.00007564, -0.00008419, 0.00017971, 0.00020882, 0.00005066, -0.00014155, 0.00000401, -0.00004622, 0.00001597, -0.00001783, 0.00001447, -0.00001784, 0.00000505, -0.00000090, 0.00000522, -0.00001722, 0.00000373, 0.00000088, 0.00000273, -0.00002134, 0.00000482, 0.00000333, 0.00000293, -0.00000098, -0.00000491, 0.00000102, -0.00000864, -0.00001389, 0.00000433, 0.00001725, -0.00001140, -0.00000694, 0.00001037, 0.00000464, -0.00000084, -0.00000698, 0.00000272, -0.00000220, -0.00000078, 0.00000079, 0.00000065, 0.00000677, -0.00000323, -0.00000990, -0.00000011, 0.00000040, 0.00000643, -0.00000402, 0.00000736, },
		{ 0.00001934, -0.00001429, 0.00001022, -0.00000974, 0.00000354, 0.00000083, -0.00001811, 0.00000031, -0.00000154, 0.00000700, -0.00000939, -0.00000346, 0.00000393, -0.00001320, 0.00000503, -0.00000975, 0.00001004, 0.00000560, -0.00000878, 0.00000240, -0.00000523, 0.00000477, -0.00000955, 0.00001010, -0.00003670, 0.00000450, -0.00000093, 0.00004545, -0.00000340, 0.00004312, 0.00001747, 0.00000982, -0.00003316, -0.00001117, 0.00000472, -0.00000717, 0.00000053, -0.00000039, -0.00000915, -0.00000003, -0.00000211, -0.00000133, 0.00000494, 0.00000616, -0.00000580, -0.00000332, 0.00000170, 0.00000322, -0.00000986, 0.00000697, -0.00000439, 0.00000746, -0.00001413, 0.00000759, 0.00000726, -0.00000221, -0.00000755, -0.00000036, 0.00000589, 0.00000621, -0.00000210, -0.00000134, -0.00000008, -0.00000054, -0.00000660, -0.00000316, 0.00000131, 0.00001189, -0.00001143, -0.00000075, -0.00000214, 0.00000906, -0.00001002, 0.00000154, -0.00000862, 0.00000348, -0.00002220, -0.00000468, -0.00002054, -0.00001019, -0.00013676, 0.00009492, 0.00019728, 0.00015490, -0.00010134, -0.00005782, -0.00001409, -0.00002010, 0.00000783, -0.00000897, 0.00000690, -0.00001707, 0.00001159, -0.00000290, -0.00000212, -0.00001342, 0.00000947, -0.00000167, -0.00000686, -0.00001591, 0.00001453, -0.00000470, 0.00000829, -0.00001119, 0.00000495, -0.00000829, -0.00000960, -0.00000937, 0.00001976, -0.00000141, -0.00001221, 0.00000459, 0.00000963, -0.00000026, -0.00000362, -0.00000377, 0.00000352, -0.00000542, 0.00000374, -0.00000266, 0.00000691, 0.00000097, -0.00000766, -0.00000639, 0.00000271, 0.00000198, 0.00000420, -0.00000506, },
		{ 0.00001398, -0.00000014, -0.00000501, 0.00000624, -0.00001237, 0.00001436, -0.00002084, -0.00000190, -0.00000465, 0.00000737, -0.00000305, -0.00000834, 0.00000262, -0.00000471, -0.00000771, 0.00000409, -0.00000988, 0.00002226, -0.00001577, 0.00000793, -0.00001255, 0.00001221, -0.00001590, 0.00001508, -0.00002649, -0.00001192, -0.00000433, 0.00003441, 0.00001768, 0.00001843, 0.00004022, 0.00000758, -0.00001173, -0.00003363, 0.00000986, -0.00000836, -0.00000013, -0.00000088, -0.00000458, -0.00000614, 0.00000173, -0.00000405, 0.00000240, 0.00000755, -0.00000095, -0.00000547, -0.00000149, 0.00000601, -0.00000807, 0.00000176, -0.00000039, 0.00000354, -0.00000451, -0.00000593, 0.00001360, -0.00000121, -0.00000484, -0.00000511, 0.00000434, 0.00000707, 0.00000136, -0.00000241, -0.00000136, 0.00000230, -0.00000712, -0.00000186, -0.00000585, 0.00001442, -0.00000531, -0.00000360, -0.00000537, 0.00000987, -0.00000620, 0.00000048, -0.00000996, 0.00000599, -0.00001883, -0.00000390, -0.00002631, 0.00000992, -0.00010765, -0.00001943, 0.00018089, 0.00017808, 0.00002512, -0.00012965, 0.00000739, -0.00004487, 0.00001949, -0.00002035, 0.00001756, -0.00002183, 0.00000927, -0.00000334, 0.00000614, -0.00001865, 0.00000705, -0.00000194, 0.00000326, -0.00002318, 0.00000938, -0.00000039, 0.00000688, -0.00000661, -0.00000047, -0.00000221, -0.00000918, -0.00001283, 0.00000935, 0.00001173, -0.00001262, -0.00000310, 0.00001141, 0.00000217, -0.00000080, -0.00000669, 0.00000335, -0.00000368, 0.00000074, -0.00000023, 0.00000233, 0.00000537, -0.00000420, -0.00000957, 0.00000075, 0.00000076, 0.00000696, -0.00000671, },
		{ 0.00000070, 0.00001566, -0.00001481, 0.00001230, -0.00001436, 0.00001068, -0.00000859, -0.00001221, -0.00000138, 0.00000071, 0.00000490, -0.00000851, -0.00000385, 0.00000411, -0.00001412, 0.00000660, -0.00001174, 0.00001433, 0.00000008, -0.00000476, -0.00000192, -0.00000043, -0.00000027, -0.00000397, 0.00000242, -0.00003338, -0.00000033, 0.00000919, 0.00004340, -0.00000010, 0.00004899, 0.00001390, 0.00000414, -0.00003939, -0.00000101, -0.00000178, -0.00000369, -0.00000116, -0.00000106, -0.00000869, 0.00000124, -0.00000301, -0.00000062, 0.00000539, 0.00000542, -0.00000641, -0.00000277, 0.00000283, -0.00000022, -0.00000676, 0.00000549, -0.00000300, 0.00000553, -0.00001374, 0.00001199, 0.00000306, -0.00000250, -0.00000762, 0.00000199, 0.00000535, 0.00000598, -0.00000286, -0.00000178, 0.00000157, -0.00000299, -0.00000502, -0.00000445, 0.00000477, 0.00000834, -0.00000966, -0.00000202, 0.00000073, 0.00000548, -0.00000705, -0.00000221, -0.00000420, -0.00000303, -0.00001561, -0.00001294, -0.00000813, -0.00003929, -0.00010972, 0.00013819, 0.00016337, 0.00013435, -0.00011451, -0.00003721, -0.00002375, -0.00000730, -0.00000120, -0.00000036, -0.00000107, -0.00001067, 0.00000754, 0.00000069, -0.00000697, -0.00000768, 0.00000544, 0.00000098, -0.00001178, -0.00000988, 0.00001085, -0.00000102, 0.00000421, -0.00000967, 0.00000495, -0.00001049, -0.00000966, -0.00000450, 0.00001832, -0.00000591, -0.00000956, 0.00000781, 0.00000727, -0.00000067, -0.00000393, -0.00000240, 0.00000189, -0.00000435, 0.00000326, -0.00000175, 0.00000681, -0.00000032, -0.00000831, -0.00000482, 0.00000214, 0.00000433, -0.00000003, },
		{ -0.00000643, 0.00001784, -0.00000743, 0.00000192, -0.00000132, -0.00000507, 0.00000867, -0.00002090, 0.00000077, -0.00000519, 0.00000801, -0.00000415, -0.00000800, 0.00000300, -0.00000623, -0.00000570, 0.00000184, -0.00000645, 0.00001979, -0.00001558, 0.00000724, -0.00001182, 0.00001192, -0.00001581, 0.00001496, -0.00003082, -0.00001058, -0.00000438, 0.00004426, 0.00001069, 0.00003130, 0.00003501, 0.00000706, -0.00002471, -0.00002358, 0.00000711, -0.00000826, 0.00000045, -0.00000211, -0.00000507, -0.00000417, 0.00000086, -0.00000356, 0.00000261, 0.00000797, -0.00000249, -0.00000496, -0.00000045, 0.00000443, -0.00000860, 0.00000411, -0.00000260, 0.00000586, -0.00000876, 0.00000067, 0.00001053, -0.00000227, -0.00000582, -0.00000269, 0.00000451, 0.00000688, 0.00000066, -0.00000328, 0.00000007, 0.00000119, -0.00000735, -0.00000177, -0.00000454, 0.00001460, -0.00000715, -0.00000242, -0.00000499, 0.00000985, -0.00000706, 0.00000052, -0.00000981, 0.00000510, -0.00001909, -0.00000435, -0.00002445, 0.00000435, -0.00012352, 0.00004564, 0.00016076, 0.00016010, -0.00000245, -0.00011206, 0.00000478, -0.00003819, 0.00001753, -0.00001813, 0.00001627, -0.00002205, 0.00001046, -0.00000324, 0.00000462, -0.00001738, 0.00000769, -0.00000259, 0.00000164, -0.00002238, 0.00001131, -0.00000160, 0.00000822, -0.00001039, 0.00000340, -0.00000595, -0.00000915, -0.00001056, 0.00001248, 0.00000608, -0.00001237, 0.00000085, 0.00001111, 0.00000019, -0.00000055, -0.00000630, 0.00000335, -0.00000446, 0.00000183, -0.00000081, 0.00000362, 0.00000390, -0.00000510, -0.00000881, 0.00000109, 0.00000212, 0.00000542, },
		{ -0.00000367, 0.00000741, 0.00000898, -0.00001144, 0.00001020, -0.00001445, 0.00001381, -0.00001603, -0.00000598, -0.00000379, 0.00000314, 0.00000289, -0.00000777, -0.00000421, 0.00000419, -0.00001476, 0.00000785, -0.00001313, 0.00001777, -0.00000485, -0.00000121, -0.00000570, 0.00000399, -0.00000500, 0.00000127, -0.00000560, -0.00002956, -0.00000585, 0.00002219, 0.00003773, 0.00000787, 0.00005097, 0.00001172, -0.00000629, -0.00003858, 0.00000483, -0.00000665, -0.00000052, -0.00000324, -0.00000133, -0.00000785, 0.00000194, -0.00000356, -0.00000020, 0.00000612, 0.00000426, -0.00000669, -0.00000204, 0.00000325, -0.00000319, -0.00000311, 0.00000286, -0.00000039, 0.00000197, -0.00001057, 0.00001361, -0.00000024, -0.00000294, -0.00000680, 0.00000366, 0.00000487, 0.00000563, -0.00000388, -0.00000141, 0.00000245, -0.00000514, -0.00000340, -0.00000551, 0.00000792, 0.00000470, -0.00000759, -0.00000334, 0.00000339, 0.00000208, -0.00000440, -0.00000549, -0.00000049, -0.00000860, -0.00000986, -0.00002011, 0.00000259, -0.00007138, -0.00006172, 0.00015393, 0.00013731, 0.00011351, -0.00012138, -0.00002127, -0.00003011, 0.00000248, -0.00000832, 0.00000679, -0.00000822, -0.00000463, 0.00000384, 0.00000370, -0.00001102, -0.00000250, 0.00000154, 0.00000328, -0.00001605, -0.00000409, 0.00000709, 0.00000298, -0.00000117, -0.00000607, 0.00000289, -0.00001169, -0.00000905, 0.00000019, 0.00001502, -0.00000874, -0.00000637, 0.00001018, 0.00000435, -0.00000027, -0.00000439, -0.00000127, 0.00000025, -0.00000296, 0.00000252, -0.00000052, 0.00000616, -0.00000141, -0.00000866, -0.00000351, 0.00000215, 0.00000568, },
		{ 0.00000231, -0.00000310, 0.00001783, -0.00001167, 0.00000689, -0.00000782, 0.00000282, 0.00000025, -0.00001724, 0.00000088, -0.00000434, 0.00000760, -0.00000423, -0.00000874, 0.00000438, -0.00000865, -0.00000271, -0.00000132, -0.00000201, 0.00001582, -0.00001380, 0.00000494, -0.00000918, 0.00000951, -0.00001341, 0.00001182, -0.00003207, -0.00001280, 0.00000074, 0.00004973, 0.00000756, 0.00004221, 0.00003025, 0.00000238, -0.00003402, -0.00001267, 0.00000170, -0.00000571, -0.00000081, -0.00000241, -0.00000552, -0.00000222, -0.00000017, -0.00000304, 0.00000291, 0.00000817, -0.00000397, -0.00000427, 0.00000052, 0.00000206, -0.00000737, 0.00000460, -0.00000324, 0.00000648, -0.00001104, 0.00000645, 0.00000647, -0.00000257, -0.00000656, -0.00000008, 0.00000412, 0.00000688, -0.00000040, -0.00000375, 0.00000171, -0.00000077, -0.00000654, -0.00000245, -0.00000227, 0.00001339, -0.00000761, -0.00000237, -0.00000345, 0.00000834, -0.00000636, -0.00000119, -0.00000762, 0.00000167, -0.00001589, -0.00000920, -0.00001636, -0.00001378, -0.00011693, 0.00009724, 0.00013176, 0.00014991, -0.00002987, -0.00009135, -0.00000144, -0.00002835, 0.00001189, -0.00001249, 0.00001158, -0.00001918, 0.00000913, -0.00000108, 0.00000115, -0.00001391, 0.00000603, -0.00000140, -0.00000170, -0.00001937, 0.00001092, -0.00000042, 0.00000665, -0.00001149, 0.00000561, -0.00000942, -0.00000862, -0.00000737, 0.00001337, 0.00000121, -0.00001111, 0.00000475, 0.00000940, -0.00000082, -0.00000045, -0.00000582, 0.00000283, -0.00000454, 0.00000239, -0.00000085, 0.00000435, 0.00000254, -0.00000591, -0.00000780, 0.00000120, 0.00000394, },
		{ 0.00000538, -0.00000533, 0.00001302, 0.00000137, -0.00000581, 0.00000505, -0.00001040, 0.00001214, -0.00001977, -0.00000116, -0.00000577, 0.00000503, 0.00000165, -0.00000795, -0.00000366, 0.00000333, -0.00001431, 0.00000801, -0.00001327, 0.00001965, -0.00000841, 0.00000104, -0.00000794, 0.00000690, -0.00000818, 0.00000465, -0.00001216, -0.00002738, -0.00000946, 0.00003550, 0.00003079, 0.00001861, 0.00005039, 0.00000864, -0.00001863, -0.00003183, 0.00000560, -0.00000849, 0.00000103, -0.00000474, -0.00000165, -0.00000654, 0.00000202, -0.00000374, -0.00000005, 0.00000709, 0.00000270, -0.00000662, -0.00000120, 0.00000275, -0.00000499, 0.00000006, 0.00000009, 0.00000248, -0.00000214, -0.00000553, 0.00001263, -0.00000230, -0.00000365, -0.00000511, 0.00000453, 0.00000460, 0.00000502, -0.00000494, -0.00000027, 0.00000227, -0.00000646, -0.00000227, -0.00000587, 0.00001017, 0.00000171, -0.00000599, -0.00000395, 0.00000503, -0.00000029, -0.00000302, -0.00000715, 0.00000112, -0.00001137, -0.00000750, -0.00002261, 0.00000566, -0.00009608, -0.00000425, 0.00014601, 0.00012272, 0.00008903, -0.00012036, -0.00001155, -0.00003162, 0.00000755, -0.00001165, 0.00001062, -0.00001277, -0.00000056, 0.00000188, 0.00000488, -0.00001308, 0.00000091, -0.00000109, 0.00000421, -0.00001860, 0.00000025, 0.00000461, 0.00000568, -0.00000612, -0.00000175, -0.00000056, -0.00001198, -0.00000756, 0.00000375, 0.00001091, -0.00001015, -0.00000270, 0.00001124, 0.00000155, 0.00000063, -0.00000499, -0.00000044, -0.00000109, -0.00000160, 0.00000183, 0.00000071, 0.00000518, -0.00000237, -0.00000866, -0.00000258, 0.00000296, },
		{ 0.00000540, -0.00000123, 0.00000231, 0.00001411, -0.00001212, 0.00000851, -0.00001120, 0.00000861, -0.00000833, -0.00001154, -0.00000079, -0.00000259, 0.00000676, -0.00000407, -0.00000976, 0.00000604, -0.00001129, 0.00000065, -0.00000482, 0.00000279, 0.00001109, -0.00001120, 0.00000192, -0.00000558, 0.00000601, -0.00000984, 0.00000683, -0.00003144, -0.00001751, 0.00001053, 0.00005074, 0.00000891, 0.00005043, 0.00002580, -0.00000681, -0.00003710, -0.00000436, -0.00000367, -0.00000243, -0.00000297, -0.00000216, -0.00000572, -0.00000059, -0.00000111, -0.00000270, 0.00000352, 0.00000793, -0.00000519, -0.00000342, 0.00000111, -0.00000051, -0.00000489, 0.00000343, -0.00000227, 0.00000528, -0.00001082, 0.00001029, 0.00000248, -0.00000262, -0.00000672, 0.00000229, 0.00000346, 0.00000691, -0.00000175, -0.00000358, 0.00000303, -0.00000303, -0.00000510, -0.00000354, 0.00000046, 0.00001134, -0.00000722, -0.00000291, -0.00000138, 0.00000610, -0.00000500, -0.00000353, -0.00000478, -0.00000248, -0.00001157, -0.00001548, -0.00000610, -0.00003978, -0.00008835, 0.00012713, 0.00010485, 0.00014124, -0.00005356, -0.00007101, -0.00000811, -0.00001831, 0.00000531, -0.00000582, 0.00000549, -0.00001486, 0.00000667, 0.00000188, -0.00000302, -0.00000945, 0.00000316, 0.00000065, -0.00000572, -0.00001525, 0.00000929, 0.00000209, 0.00000293, -0.00001002, 0.00000573, -0.00001212, -0.00000747, -0.00000398, 0.00001239, -0.00000249, -0.00000906, 0.00000820, 0.00000657, -0.00000068, -0.00000076, -0.00000524, 0.00000199, -0.00000406, 0.00000252, -0.00000044, 0.00000450, 0.00000135, -0.00000656, -0.00000677, 0.00000151, },
		{ 0.00000437, 0.00000340, -0.00000425, 0.00001585, -0.00000498, -0.00000010, -0.00000102, -0.00000398, 0.00000651, -0.00001944, 0.00000136, -0.00000670, 0.00000613, 0.00000124, -0.00000922, -0.00000192, 0.00000123, -0.00001240, 0.00000670, -0.00001185, 0.00001972, -0.00001031, 0.00000170, -0.00000833, 0.00000784, -0.00000929, 0.00000546, -0.00001631, -0.00002819, -0.00000915, 0.00004682, 0.00002486, 0.00003023, 0.00004849, 0.00000234, -0.00002903, -0.00002218, 0.00000251, -0.00000744, 0.00000062, -0.00000524, -0.00000218, -0.00000485, 0.00000159, -0.00000369, 0.00000003, 0.00000810, 0.00000089, -0.00000620, -0.00000036, 0.00000140, -0.00000526, 0.00000195, -0.00000190, 0.00000466, -0.00000558, 0.00000005, 0.00000978, -0.00000325, -0.00000449, -0.00000279, 0.00000462, 0.00000454, 0.00000409, -0.00000578, 0.00000135, 0.00000104, -0.00000672, -0.00000187, -0.00000535, 0.00001129, -0.00000033, -0.00000517, -0.00000360, 0.00000539, -0.00000136, -0.00000314, -0.00000696, 0.00000042, -0.00001101, -0.00000918, -0.00001926, -0.00000200, -0.00010502, 0.00004978, 0.00012342, 0.00011780, 0.00006119, -0.00011240, -0.00000742, -0.00002879, 0.00000813, -0.00001111, 0.00001080, -0.00001425, 0.00000108, 0.00000205, 0.00000391, -0.00001282, 0.00000214, -0.00000204, 0.00000349, -0.00001913, 0.00000270, 0.00000397, 0.00000614, -0.00000933, 0.00000198, -0.00000461, -0.00001137, -0.00000533, 0.00000566, 0.00000692, -0.00001054, 0.00000139, 0.00001067, -0.00000043, 0.00000157, -0.00000561, 0.00000002, -0.00000197, -0.00000048, 0.00000140, 0.00000166, 0.00000408, -0.00000323, -0.00000834, -0.00000192, },
		{ 0.00000232, 0.00000576, -0.00000383, 0.00000812, 0.00000799, -0.00000933, 0.00000673, -0.00001069, 0.00001058, -0.00001477, -0.00000570, -0.00000315, -0.00000061, 0.00000614, -0.00000452, -0.00001014, 0.00000708, -0.00001327, 0.00000349, -0.00000787, 0.00000714, 0.00000650, -0.00000872, -0.00000077, -0.00000221, 0.00000268, -0.00000657, 0.00000166, -0.00003078, -0.00002263, 0.00002332, 0.00004846, 0.00001411, 0.00005626, 0.00002008, -0.00001861, -0.00003375, -0.00000053, -0.00000693, -0.00000005, -0.00000496, -0.00000188, -0.00000540, 0.00000047, -0.00000177, -0.00000264, 0.00000453, 0.00000714, -0.00000604, -0.00000246, 0.00000108, -0.00000259, -0.00000201, 0.00000123, -0.00000019, 0.00000274, -0.00000833, 0.00001171, -0.00000068, -0.00000279, -0.00000609, 0.00000409, 0.00000272, 0.00000685, -0.00000327, -0.00000267, 0.00000358, -0.00000498, -0.00000355, -0.00000458, 0.00000310, 0.00000909, -0.00000660, -0.00000338, 0.00000046, 0.00000398, -0.00000392, -0.00000535, -0.00000269, -0.00000547, -0.00000864, -0.00001973, 0.00000098, -0.00006562, -0.00004450, 0.00013403, 0.00008664, 0.00012933, -0.00007063, -0.00005388, -0.00001273, -0.00001055, 0.00000022, -0.00000036, 0.00000004, -0.00001084, 0.00000463, 0.00000426, -0.00000656, -0.00000530, 0.00000023, 0.00000251, -0.00000931, -0.00001122, 0.00000770, 0.00000452, -0.00000160, -0.00000685, 0.00000393, -0.00001381, -0.00000557, -0.00000123, 0.00001033, -0.00000512, -0.00000621, 0.00001064, 0.00000328, 0.00000042, -0.00000156, -0.00000461, 0.00000104, -0.00000324, 0.00000240, 0.00000022, 0.00000418, 0.00000032, -0.00000697, -0.00000588, },
		{ -0.00000126, 0.00000581, 0.00000040, -0.00000064, 0.00001517, -0.00000853, 0.00000380, -0.00000581, 0.00000237, -0.00000115, -0.00001589, 0.00000160, -0.00000653, 0.00000671, 0.00000120, -0.00001110, 0.00000062, -0.00000179, -0.00000932, 0.00000415, -0.00000912, 0.00001824, -0.00001082, 0.00000111, -0.00000722, 0.00000718, -0.00000872, 0.00000405, -0.00001825, -0.00003206, -0.00000410, 0.00005486, 0.00002142, 0.00004153, 0.00004535, -0.00000809, -0.00003424, -0.00001324, -0.00000203, -0.00000478, -0.00000114, -0.00000496, -0.00000274, -0.00000309, 0.00000092, -0.00000363, 0.00000029, 0.00000889, -0.00000095, -0.00000549, 0.00000028, -0.00000048, -0.00000410, 0.00000225, -0.00000257, 0.00000553, -0.00000746, 0.00000489, 0.00000606, -0.00000346, -0.00000517, -0.00000019, 0.00000407, 0.00000466, 0.00000283, -0.00000613, 0.00000302, -0.00000092, -0.00000602, -0.00000216, -0.00000408, 0.00001143, -0.00000153, -0.00000497, -0.00000249, 0.00000473, -0.00000149, -0.00000426, -0.00000558, -0.00000167, -0.00000870, -0.00001347, -0.00001189, -0.00001920, -0.00009464, 0.00008996, 0.00009644, 0.00011773, 0.00003268, -0.00010013, -0.00000677, -0.00002356, 0.00000593, -0.00000802, 0.00000837, -0.00001338, 0.00000090, 0.00000381, 0.00000131, -0.00001072, 0.00000154, -0.00000163, 0.00000153, -0.00001809, 0.00000366, 0.00000487, 0.00000438, -0.00001025, 0.00000424, -0.00000858, -0.00000982, -0.00000293, 0.00000601, 0.00000354, -0.00001008, 0.00000557, 0.00000850, -0.00000113, 0.00000209, -0.00000612, 0.00000015, -0.00000234, 0.00000032, 0.00000131, 0.00000221, 0.00000301, -0.00000397, -0.00000786, },
		{ -0.00000516, 0.00000367, 0.00000464, -0.00000438, 0.00001252, 0.00000136, -0.00000484, 0.00000277, -0.00000698, 0.00000839, -0.00001776, -0.00000117, -0.00000531, 0.00000120, 0.00000604, -0.00000599, -0.00000931, 0.00000688, -0.00001388, 0.00000511, -0.00000980, 0.00001038, 0.00000271, -0.00000699, -0.00000243, 0.00000012, 0.00000042, -0.00000471, -0.00000233, -0.00003186, -0.00002593, 0.00003701, 0.00004457, 0.00002208, 0.00006026, 0.00001102, -0.00002951, -0.00002609, -0.00000105, -0.00000739, 0.00000059, -0.00000607, -0.00000192, -0.00000451, 0.00000091, -0.00000214, -0.00000280, 0.00000587, 0.00000583, -0.00000643, -0.00000149, 0.00000032, -0.00000363, 0.00000039, -0.00000111, 0.00000222, -0.00000027, -0.00000434, 0.00001089, -0.00000274, -0.00000320, -0.00000463, 0.00000514, 0.00000207, 0.00000660, -0.00000475, -0.00000110, 0.00000310, -0.00000615, -0.00000234, -0.00000523, 0.00000525, 0.00000711, -0.00000622, -0.00000331, 0.00000155, 0.00000250, -0.00000370, -0.00000600, -0.00000209, -0.00000633, -0.00000860, -0.00001972, 0.00000062, -0.00008291, 0.00000431, 0.00012269, 0.00007865, 0.00011239, -0.00008010, -0.00004109, -0.00001433, -0.00000610, -0.00000216, 0.00000263, -0.00000351, -0.00000824, 0.00000397, 0.00000513, -0.00000860, -0.00000236, -0.00000194, 0.00000354, -0.00001179, -0.00000807, 0.00000702, 0.00000572, -0.00000558, -0.00000310, 0.00000073, -0.00001432, -0.00000309, 0.00000030, 0.00000804, -0.00000695, -0.00000253, 0.00001150, 0.00000034, 0.00000201, -0.00000275, -0.00000398, 0.00000018, -0.00000231, 0.00000224, 0.00000089, 0.00000357, -0.00000057, -0.00000713, },
		{ -0.00000738, -0.00000029, 0.00000660, -0.00000250, 0.00000441, 0.00001138, -0.00000886, 0.00000497, -0.00000784, 0.00000649, -0.00000853, -0.00001066, 0.00000020, -0.00000558, 0.00000720, 0.00000081, -0.00001277, 0.00000318, -0.00000494, -0.00000579, 0.00000095, -0.00000569, 0.00001585, -0.00001057, 0.00000001, -0.00000544, 0.00000578, -0.00000749, 0.00000153, -0.00001918, -0.00003787, 0.00000524, 0.00005953, 0.00002092, 0.00005221, 0.00003970, -0.00002139, -0.00003302, -0.00000772, -0.00000565, -0.00000218, -0.00000322, -0.00000440, -0.00000301, -0.00000159, 0.00000029, -0.00000378, 0.00000096, 0.00000921, -0.00000260, -0.00000458, 0.00000049, -0.00000237, -0.00000198, 0.00000112, -0.00000187, 0.00000500, -0.00000738, 0.00000810, 0.00000239, -0.00000337, -0.00000538, 0.00000232, 0.00000305, 0.00000493, 0.00000127, -0.00000579, 0.00000426, -0.00000309, -0.00000470, -0.00000290, -0.00000239, 0.00001096, -0.00000228, -0.00000498, -0.00000115, 0.00000363, -0.00000134, -0.00000554, -0.00000406, -0.00000372, -0.00000640, -0.00001769, -0.00000440, -0.00004085, -0.00006713, 0.00011112, 0.00007327, 0.00011740, 0.00000675, -0.00008653, -0.00000720, -0.00001816, 0.00000309, -0.00000430, 0.00000496, -0.00001153, 0.00000005, 0.00000608, -0.00000191, -0.00000773, -0.00000009, -0.00000056, -0.00000089, -0.00001632, 0.00000401, 0.00000635, 0.00000115, -0.00000910, 0.00000470, -0.00001196, -0.00000728, -0.00000108, 0.00000539, 0.00000077, -0.00000869, 0.00000925, 0.00000519, -0.00000046, 0.00000188, -0.00000639, 0.00000001, -0.00000227, 0.00000083, 0.00000149, 0.00000233, 0.00000204, -0.00000454, },
		{ -0.00000710, -0.00000448, 0.00000526, 0.00000226, -0.00000250, 0.00001422, -0.00000394, -0.00000043, -0.00000157, -0.00000182, 0.00000306, -0.00001718, 0.00000134, -0.00000675, 0.00000276, 0.00000625, -0.00000832, -0.00000731, 0.00000537, -0.00001298, 0.00000524, -0.00001037, 0.00001230, -0.00000009, -0.00000616, -0.00000290, 0.00000122, -0.00000056, -0.00000456, -0.00000461, -0.00003556, -0.00002584, 0.00004991, 0.00004057, 0.00003198, 0.00006224, -0.00000252, -0.00003594, -0.00001756, -0.00000412, -0.00000573, -0.00000036, -0.00000624, -0.00000225, -0.00000320, 0.00000088, -0.00000238, -0.00000297, 0.00000730, 0.00000415, -0.00000640, -0.00000065, -0.00000105, -0.00000340, 0.00000167, -0.00000282, 0.00000421, -0.00000282, 0.00000005, 0.00000845, -0.00000375, -0.00000376, -0.00000248, 0.00000534, 0.00000161, 0.00000609, -0.00000595, 0.00000085, 0.00000168, -0.00000635, -0.00000169, -0.00000534, 0.00000679, 0.00000554, -0.00000619, -0.00000264, 0.00000181, 0.00000173, -0.00000431, -0.00000557, -0.00000279, -0.00000531, -0.00001120, -0.00001550, -0.00000844, -0.00008585, 0.00004769, 0.00010112, 0.00007846, 0.00009134, -0.00008294, -0.00003215, -0.00001344, -0.00000450, -0.00000205, 0.00000313, -0.00000500, -0.00000724, 0.00000486, 0.00000435, -0.00000893, -0.00000090, -0.00000314, 0.00000365, -0.00001311, -0.00000590, 0.00000744, 0.00000520, -0.00000807, 0.00000021, -0.00000331, -0.00001345, -0.00000053, 0.00000059, 0.00000601, -0.00000817, 0.00000186, 0.00001045, -0.00000147, 0.00000347, -0.00000410, -0.00000344, -0.00000042, -0.00000143, 0.00000219, 0.00000138, 0.00000283, -0.00000134, },
		{ -0.00000492, -0.00000697, 0.00000113, 0.00000622, -0.00000415, 0.00000933, 0.00000576, -0.00000661, 0.00000363, -0.00000682, 0.00000719, -0.00001378, -0.00000534, -0.00000198, -0.00000420, 0.00000790, -0.00000044, -0.00001352, 0.00000497, -0.00000740, -0.00000262, -0.00000215, -0.00000224, 0.00001320, -0.00001022, -0.00000089, -0.00000380, 0.00000453, -0.00000666, -0.00000083, -0.00002064, -0.00004389, 0.00001765, 0.00006149, 0.00002325, 0.00006228, 0.00002954, -0.00003438, -0.00002658, -0.00000649, -0.00000696, -0.00000085, -0.00000473, -0.00000407, -0.00000272, -0.00000060, -0.00000014, -0.00000422, 0.00000211, 0.00000894, -0.00000390, -0.00000355, 0.00000013, -0.00000372, 0.00000036, -0.00000087, -0.00000018, 0.00000341, -0.00000556, 0.00000932, -0.00000062, -0.00000331, -0.00000491, 0.00000441, 0.00000173, 0.00000530, -0.00000051, -0.00000467, 0.00000466, -0.00000491, -0.00000319, -0.00000379, -0.00000060, 0.00001027, -0.00000294, -0.00000477, -0.00000008, 0.00000262, -0.00000150, -0.00000627, -0.00000325, -0.00000459, -0.00000577, -0.00001941, -0.00000101, -0.00005995, -0.00002924, 0.00011377, 0.00005824, 0.00011340, -0.00001430, -0.00007375, -0.00000709, -0.00001417, 0.00000123, -0.00000148, 0.00000202, -0.00000993, -0.00000036, 0.00000785, -0.00000476, -0.00000476, -0.00000201, 0.00000056, -0.00000311, -0.00001456, 0.00000456, 0.00000739, -0.00000244, -0.00000659, 0.00000338, -0.00001430, -0.00000394, -0.00000038, 0.00000458, -0.00000164, -0.00000615, 0.00001167, 0.00000158, 0.00000126, 0.00000091, -0.00000639, -0.00000028, -0.00000188, 0.00000116, 0.00000179, 0.00000212, 0.00000119, },
		{ -0.00000195, -0.00000701, -0.00000359, 0.00000653, -0.00000063, 0.00000137, 0.00001282, -0.00000678, 0.00000248, -0.00000460, 0.00000276, -0.00000360, -0.00001388, 0.00000184, -0.00000738, 0.00000427, 0.00000625, -0.00001089, -0.00000465, 0.00000302, -0.00001092, 0.00000412, -0.00000978, 0.00001309, -0.00000209, -0.00000592, -0.00000257, 0.00000147, -0.00000075, -0.00000564, -0.00000567, -0.00004170, -0.00002179, 0.00006114, 0.00003730, 0.00004358, 0.00006071, -0.00001968, -0.00003580, -0.00001138, -0.00000733, -0.00000342, -0.00000205, -0.00000588, -0.00000257, -0.00000182, 0.00000066, -0.00000272, -0.00000286, 0.00000854, 0.00000234, -0.00000602, -0.00000008, -0.00000273, -0.00000202, 0.00000159, -0.00000345, 0.00000528, -0.00000424, 0.00000386, 0.00000523, -0.00000402, -0.00000426, 0.00000010, 0.00000470, 0.00000147, 0.00000524, -0.00000662, 0.00000273, -0.00000037, -0.00000564, -0.00000165, -0.00000496, 0.00000783, 0.00000424, -0.00000632, -0.00000161, 0.00000149, 0.00000133, -0.00000527, -0.00000466, -0.00000390, -0.00000367, -0.00001477, -0.00000939, -0.00002388, -0.00007313, 0.00007825, 0.00007746, 0.00008190, 0.00006862, -0.00008118, -0.00002566, -0.00001147, -0.00000441, -0.00000056, 0.00000206, -0.00000511, -0.00000726, 0.00000677, 0.00000237, -0.00000793, -0.00000068, -0.00000358, 0.00000317, -0.00001367, -0.00000428, 0.00000849, 0.00000319, -0.00000878, 0.00000238, -0.00000758, -0.00001108, 0.00000133, 0.00000015, 0.00000429, -0.00000861, 0.00000643, 0.00000766, -0.00000171, 0.00000426, -0.00000533, -0.00000306, -0.00000071, -0.00000069, 0.00000228, 0.00000159, 0.00000207, },
		{ 0.00000045, -0.00000512, -0.00000659, 0.00000289, 0.00000454, -0.00000378, 0.00001261, -0.00000001, -0.00000310, 0.00000090, -0.00000362, 0.00000452, -0.00001597, -0.00000115, -0.00000426, -0.00000256, 0.00000877, -0.00000262, -0.00001307, 0.00000560, -0.00000868, -0.00000040, -0.00000463, 0.00000078, 0.00001070, -0.00001008, -0.00000123, -0.00000273, 0.00000387, -0.00000681, -0.00000225, -0.00002382, -0.00004847, 0.00003179, 0.00006153, 0.00002834, 0.00007115, 0.00001342, -0.00004313, -0.00001802, -0.00000847, -0.00000600, -0.00000109, -0.00000532, -0.00000413, -0.00000189, -0.00000012, -0.00000041, -0.00000482, 0.00000362, 0.00000812, -0.00000479, -0.00000253, -0.00000084, -0.00000412, 0.00000221, -0.00000299, 0.00000191, 0.00000142, -0.00000261, 0.00000867, -0.00000267, -0.00000343, -0.00000364, 0.00000576, 0.00000035, 0.00000569, -0.00000233, -0.00000289, 0.00000404, -0.00000597, -0.00000189, -0.00000457, 0.00000109, 0.00000956, -0.00000372, -0.00000414, 0.00000045, 0.00000195, -0.00000216, -0.00000623, -0.00000335, -0.00000403, -0.00000731, -0.00001763, -0.00000427, -0.00007016, 0.00001025, 0.00010271, 0.00005169, 0.00010472, -0.00002976, -0.00006262, -0.00000601, -0.00001196, 0.00000090, -0.00000023, 0.00000023, -0.00000921, 0.00000023, 0.00000854, -0.00000670, -0.00000236, -0.00000378, 0.00000147, -0.00000490, -0.00001311, 0.00000569, 0.00000732, -0.00000545, -0.00000357, 0.00000057, -0.00001514, -0.00000035, -0.00000093, 0.00000410, -0.00000392, -0.00000236, 0.00001216, -0.00000136, 0.00000337, -0.00000063, -0.00000617, -0.00000058, -0.00000131, 0.00000143, 0.00000206, 0.00000169, },
		{ 0.00000134, -0.00000238, -0.00000696, -0.00000232, 0.00000694, -0.00000300, 0.00000609, 0.00000888, -0.00000699, 0.00000338, -0.00000541, 0.00000514, -0.00000967, -0.00000918, 0.00000078, -0.00000724, 0.00000589, 0.00000556, -0.00001300, -0.00000207, 0.00000056, -0.00000835, 0.00000223, -0.00000844, 0.00001314, -0.00000370, -0.00000579, -0.00000200, 0.00000149, -0.00000090, -0.00000703, -0.00000654, -0.00004930, -0.00001404, 0.00007048, 0.00003517, 0.00005693, 0.00005326, -0.00003738, -0.00002946, -0.00000928, -0.00000883, -0.00000187, -0.00000352, -0.00000556, -0.00000251, -0.00000071, 0.00000048, -0.00000333, -0.00000229, 0.00000936, 0.00000061, -0.00000539, 0.00000004, -0.00000427, 0.00000006, 0.00000030, -0.00000292, 0.00000534, -0.00000425, 0.00000638, 0.00000195, -0.00000392, -0.00000443, 0.00000278, 0.00000327, 0.00000174, 0.00000398, -0.00000652, 0.00000408, -0.00000255, -0.00000432, -0.00000210, -0.00000424, 0.00000857, 0.00000299, -0.00000630, -0.00000062, 0.00000100, 0.00000088, -0.00000601, -0.00000398, -0.00000444, -0.00000287, -0.00001717, -0.00000485, -0.00004070, -0.00004802, 0.00009317, 0.00005762, 0.00008510, 0.00004670, -0.00007682, -0.00002031, -0.00000972, -0.00000444, 0.00000101, 0.00000057, -0.00000478, -0.00000748, 0.00000894, -0.00000012, -0.00000618, -0.00000127, -0.00000359, 0.00000254, -0.00001399, -0.00000263, 0.00000944, 0.00000037, -0.00000795, 0.00000306, -0.00001147, -0.00000732, 0.00000181, -0.00000023, 0.00000250, -0.00000788, 0.00001034, 0.00000381, -0.00000045, 0.00000409, -0.00000624, -0.00000284, -0.00000069, -0.00000008, 0.00000246, 0.00000153, },
		{ 0.00000115, -0.00000015, -0.00000523, -0.00000609, 0.00000462, 0.00000197, -0.00000129, 0.00001332, -0.00000448, 0.00000028, -0.00000178, 0.00000019, -0.00000043, -0.00001514, 0.00000135, -0.00000618, -0.00000063, 0.00000943, -0.00000535, -0.00001164, 0.00000520, -0.00000875, 0.00000072, -0.00000633, 0.00000325, 0.00000841, -0.00001011, -0.00000111, -0.00000213, 0.00000367, -0.00000784, -0.00000270, -0.00002909, -0.00005064, 0.00004676, 0.00006015, 0.00003664, 0.00007688, -0.00000824, -0.00004472, -0.00001091, -0.00001142, -0.00000392, -0.00000232, -0.00000524, -0.00000437, -0.00000077, 0.00000004, -0.00000071, -0.00000533, 0.00000525, 0.00000691, -0.00000531, -0.00000164, -0.00000228, -0.00000343, 0.00000305, -0.00000461, 0.00000382, -0.00000032, 0.00000064, 0.00000667, -0.00000378, -0.00000372, -0.00000162, 0.00000614, -0.00000083, 0.00000594, -0.00000396, -0.00000075, 0.00000249, -0.00000607, -0.00000105, -0.00000514, 0.00000262, 0.00000883, -0.00000455, -0.00000316, 0.00000049, 0.00000153, -0.00000314, -0.00000569, -0.00000390, -0.00000272, -0.00001016, -0.00001336, -0.00001392, -0.00006809, 0.00004353, 0.00008457, 0.00005137, 0.00009223, -0.00004020, -0.00005293, -0.00000449, -0.00001095, 0.00000168, -0.00000029, -0.00000049, -0.00000930, 0.00000180, 0.00000813, -0.00000764, -0.00000067, -0.00000532, 0.00000226, -0.00000641, -0.00001180, 0.00000723, 0.00000606, -0.00000732, -0.00000077, -0.00000333, -0.00001408, 0.00000257, -0.00000221, 0.00000396, -0.00000594, 0.00000239, 0.00001049, -0.00000285, 0.00000515, -0.00000240, -0.00000585, -0.00000078, -0.00000068, 0.00000169, 0.00000219, },
		{ 0.00000126, 0.00000067, -0.00000262, -0.00000696, -0.00000062, 0.00000616, -0.00000391, 0.00001024, 0.00000362, -0.00000519, 0.00000265, -0.00000411, 0.00000467, -0.00001375, -0.00000438, -0.00000123, -0.00000639, 0.00000757, 0.00000398, -0.00001421, -0.00000009, -0.00000143, -0.00000589, 0.00000008, -0.00000672, 0.00001277, -0.00000515, -0.00000546, -0.00000155, 0.00000168, -0.00000156, -0.00000801, -0.00000826, -0.00005719, -0.00000313, 0.00007793, 0.00003473, 0.00007148, 0.00003768, -0.00005120, -0.00001964, -0.00001091, -0.00000808, -0.00000174, -0.00000419, -0.00000561, -0.00000193, -0.00000003, 0.00000042, -0.00000420, -0.00000124, 0.00000967, -0.00000091, -0.00000459, -0.00000039, -0.00000522, 0.00000221, -0.00000171, -0.00000150, 0.00000461, -0.00000303, 0.00000734, -0.00000084, -0.00000372, -0.00000402, 0.00000516, 0.00000125, 0.00000246, 0.00000232, -0.00000556, 0.00000450, -0.00000432, -0.00000277, -0.00000284, -0.00000330, 0.00000914, 0.00000162, -0.00000592, 0.00000006, 0.00000061, 0.00000010, -0.00000620, -0.00000387, -0.00000394, -0.00000371, -0.00001697, -0.00000482, -0.00005319, -0.00001682, 0.00009402, 0.00004423, 0.00008575, 0.00002718, -0.00007113, -0.00001546, -0.00000880, -0.00000385, 0.00000187, -0.00000058, -0.00000469, -0.00000725, 0.00001077, -0.00000258, -0.00000416, -0.00000236, -0.00000333, 0.00000192, -0.00001433, -0.00000058, 0.00000967, -0.00000248, -0.00000607, 0.00000212, -0.00001427, -0.00000272, 0.00000067, 0.00000009, 0.00000028, -0.00000556, 0.00001263, -0.00000006, 0.00000181, 0.00000300, -0.00000676, -0.00000276, -0.00000040, 0.00000043, 0.00000263, },
		{ 0.00000214, 0.00000056, -0.00000058, -0.00000535, -0.00000527, 0.00000581, -0.00000070, 0.00000269, 0.00001135, -0.00000687, 0.00000260, -0.00000331, 0.00000303, -0.00000608, -0.00001207, 0.00000207, -0.00000748, 0.00000160, 0.00000946, -0.00000803, -0.00000978, 0.00000422, -0.00000795, 0.00000090, -0.00000733, 0.00000527, 0.00000618, -0.00001005, -0.00000084, -0.00000163, 0.00000343, -0.00000916, -0.00000280, -0.00003608, -0.00005007, 0.00006214, 0.00005754, 0.00004889, 0.00007628, -0.00003247, -0.00003857, -0.00000780, -0.00001317, -0.00000221, -0.00000358, -0.00000504, -0.00000442, 0.00000030, 0.00000013, -0.00000125, -0.00000550, 0.00000674, 0.00000552, -0.00000551, -0.00000100, -0.00000389, -0.00000179, 0.00000268, -0.00000533, 0.00000517, -0.00000134, 0.00000348, 0.00000391, -0.00000414, -0.00000403, 0.00000100, 0.00000538, -0.00000148, 0.00000584, -0.00000506, 0.00000128, 0.00000037, -0.00000529, -0.00000077, -0.00000544, 0.00000408, 0.00000794, -0.00000522, -0.00000211, 0.00000031, 0.00000106, -0.00000400, -0.00000517, -0.00000415, -0.00000181, -0.00001261, -0.00000916, -0.00002688, -0.00005425, 0.00006586, 0.00006535, 0.00005416, 0.00007756, -0.00004669, -0.00004415, -0.00000330, -0.00001022, 0.00000274, -0.00000096, -0.00000068, -0.00000973, 0.00000392, 0.00000694, -0.00000778, 0.00000039, -0.00000672, 0.00000315, -0.00000797, -0.00001021, 0.00000865, 0.00000397, -0.00000794, 0.00000125, -0.00000780, -0.00001095, 0.00000389, -0.00000331, 0.00000369, -0.00000717, 0.00000733, 0.00000702, -0.00000255, 0.00000604, -0.00000402, -0.00000556, -0.00000079, -0.00000005, 0.00000195, },
		},
		{ { -0.00000107, 0.00000010, -0.00000355, 0.00000168, 0.00000644, -0.00000125, -0.00000645, 0.00000011, 0.00000073, 0.00000119, -0.00000486, 0.00000057, 0.00000283, 0.00000062, -0.00000488, 0.00000261, 0.00000169, 0.00000038, -0.00000948, -0.00000286, 0.00001362, -0.00000283, -0.00001114, 0.00000147, -0.00000044, -0.00000176, -0.00001916, -0.00000144, -0.00000780, -0.00001487, -0.00006385, 0.00001187, 0.00007644, 0.00003948, 0.00008332, 0.00001014, -0.00005945, -0.00001283, 0.00000390, -0.00001122, -0.00000557, -0.00000230, 0.00000863, -0.00000393, -0.00000457, -0.00000126, 0.00000712, -0.00000268, -0.00000296, -0.00000216, 0.00000529, -0.00000135, -0.00001009, 0.00000151, 0.00000702, -0.00000169, -0.00000338, -0.00000251, 0.00000493, -0.00000012, -0.00000504, -0.00000094, 0.00000451, -0.00000101, -0.00000114, -0.00000101, 0.00000451, -0.00000094, -0.00000504, -0.00000012, 0.00000493, -0.00000251, -0.00000338, -0.00000169, 0.00000702, 0.00000151, -0.00001009, -0.00000135, 0.00000529, -0.00000216, -0.00000296, -0.00000268, 0.00000712, -0.00000126, -0.00000457, -0.00000393, 0.00000863, -0.00000230, -0.00000557, -0.00001122, 0.00000390, -0.00001283, -0.00005945, 0.00001014, 0.00008332, 0.00003948, 0.00007644, 0.00001187, -0.00006385, -0.00001487, -0.00000780, -0.00000144, -0.00001916, -0.00000176, -0.00000044, 0.00000147, -0.00001114, -0.00000283, 0.00001362, -0.00000286, -0.00000948, 0.00000038, 0.00000169, 0.00000261, -0.00000488, 0.00000062, 0.00000283, 0.00000057, -0.00000486, 0.00000119, 0.00000073, 0.00000011, -0.00000645, -0.00000125, 0.00000644, 0.00000168, -0.00000355, 0.00000010, },
		{ -0.00000018, -0.00000050, -0.00000151, -0.00000232, 0.00000583, 0.00000339, -0.00000582, -0.00000308, 0.00000078, 0.00000175, -0.00000246, -0.00000329, 0.00000313, 0.00000212, -0.00000293, -0.00000201, 0.00000369, 0.00000088, -0.00000405, -0.00001045, 0.00000917, 0.00000781, -0.00001152, -0.00000419, 0.00000112, 0.00000131, -0.00001356, -0.00001139, -0.00000141, -0.00000991, -0.00003843, -0.00005275, 0.00007186, 0.00005170, 0.00006687, 0.00006423, -0.00004777, -0.00004030, 0.00000474, -0.00000616, -0.00000827, -0.00000605, 0.00000613, 0.00000264, -0.00000580, -0.00000372, 0.00000404, 0.00000371, -0.00000484, -0.00000256, 0.00000143, 0.00000469, -0.00000880, -0.00000518, 0.00000700, 0.00000248, -0.00000315, -0.00000385, 0.00000171, 0.00000404, -0.00000408, -0.00000396, 0.00000318, 0.00000194, -0.00000162, -0.00000146, 0.00000194, 0.00000321, -0.00000450, -0.00000312, 0.00000369, 0.00000164, -0.00000404, -0.00000284, 0.00000242, 0.00000722, -0.00000671, -0.00000733, 0.00000479, 0.00000090, -0.00000244, -0.00000392, 0.00000270, 0.00000461, -0.00000413, -0.00000542, 0.00000330, 0.00000504, -0.00000587, -0.00000741, -0.00000763, 0.00000669, -0.00004531, -0.00003440, 0.00006319, 0.00005883, 0.00004748, 0.00006615, -0.00004514, -0.00003930, -0.00000875, -0.00000176, -0.00001261, -0.00001170, 0.00000153, -0.00000050, -0.00000248, -0.00001253, 0.00001003, 0.00000687, -0.00001019, -0.00000316, 0.00000093, 0.00000334, -0.00000173, -0.00000321, 0.00000281, 0.00000215, -0.00000312, -0.00000203, 0.00000172, 0.00000053, -0.00000298, -0.00000601, 0.00000444, 0.00000503, -0.00000241, -0.00000177, },
		{ 0.00000019, -0.00000112, 0.00000010, -0.00000347, 0.00000234, 0.00000620, -0.00000190, -0.00000609, -0.00000002, 0.00000106, 0.00000093, -0.00000543, 0.00000140, 0.00000280, 0.00000033, -0.00000474, 0.00000251, 0.00000195, 0.00000005, -0.00001080, -0.00000076, 0.00001314, -0.00000366, -0.00001100, 0.00000059, 0.00000165, -0.00000415, -0.00001900, -0.00000021, -0.00000901, -0.00001398, -0.00006753, 0.00000816, 0.00008548, 0.00004233, 0.00009393, -0.00000317, -0.00006125, -0.00001029, 0.00000231, -0.00000958, -0.00000732, -0.00000101, 0.00000867, -0.00000481, -0.00000407, -0.00000177, 0.00000768, -0.00000270, -0.00000369, -0.00000186, 0.00000592, -0.00000285, -0.00000971, 0.00000240, 0.00000661, -0.00000165, -0.00000351, -0.00000240, 0.00000534, -0.00000066, -0.00000530, -0.00000042, 0.00000429, -0.00000103, -0.00000114, -0.00000097, 0.00000468, -0.00000148, -0.00000460, 0.00000022, 0.00000458, -0.00000265, -0.00000322, -0.00000168, 0.00000739, 0.00000044, -0.00001013, -0.00000001, 0.00000457, -0.00000236, -0.00000212, -0.00000287, 0.00000670, -0.00000082, -0.00000503, -0.00000291, 0.00000833, -0.00000347, -0.00000385, -0.00001283, 0.00000565, -0.00001614, -0.00005540, 0.00002080, 0.00007293, 0.00003726, 0.00006812, 0.00001470, -0.00005981, -0.00001585, -0.00000649, -0.00000302, -0.00001890, 0.00000054, -0.00000280, 0.00000274, -0.00001155, -0.00000167, 0.00001357, -0.00000468, -0.00000808, 0.00000064, 0.00000143, 0.00000264, -0.00000499, 0.00000101, 0.00000262, -0.00000007, -0.00000427, 0.00000136, 0.00000042, 0.00000021, -0.00000678, -0.00000047, 0.00000649, 0.00000105, -0.00000365, },
		{ -0.00000195, 0.00000010, -0.00000082, -0.00000119, -0.00000222, 0.00000654, 0.00000226, -0.00000545, -0.00000327, 0.00000110, 0.00000169, -0.00000293, -0.00000329, 0.00000409, 0.00000128, -0.00000243, -0.00000243, 0.00000411, 0.00000074, -0.00000497, -0.00001043, 0.00001129, 0.00000539, -0.00001025, -0.00000614, 0.00000325, 0.00000028, -0.00001460, -0.00001064, -0.00000077, -0.00001155, -0.00003686, -0.00006115, 0.00007773, 0.00005635, 0.00007619, 0.00006286, -0.00006089, -0.00003398, 0.00000162, -0.00000387, -0.00000968, -0.00000573, 0.00000682, 0.00000209, -0.00000614, -0.00000333, 0.00000350, 0.00000470, -0.00000592, -0.00000246, 0.00000192, 0.00000442, -0.00001013, -0.00000351, 0.00000654, 0.00000265, -0.00000344, -0.00000367, 0.00000182, 0.00000429, -0.00000507, -0.00000325, 0.00000302, 0.00000198, -0.00000176, -0.00000131, 0.00000198, 0.00000312, -0.00000483, -0.00000225, 0.00000330, 0.00000156, -0.00000420, -0.00000252, 0.00000246, 0.00000715, -0.00000803, -0.00000580, 0.00000475, 0.00000034, -0.00000209, -0.00000321, 0.00000176, 0.00000515, -0.00000452, -0.00000500, 0.00000404, 0.00000361, -0.00000523, -0.00000711, -0.00000824, 0.00000733, -0.00004858, -0.00002158, 0.00006041, 0.00005188, 0.00004368, 0.00006062, -0.00003828, -0.00003950, -0.00000803, -0.00000186, -0.00001423, -0.00000907, 0.00000090, -0.00000147, -0.00000117, -0.00001315, 0.00001193, 0.00000450, -0.00000970, -0.00000230, 0.00000090, 0.00000305, -0.00000160, -0.00000325, 0.00000329, 0.00000121, -0.00000283, -0.00000166, 0.00000163, 0.00000033, -0.00000299, -0.00000601, 0.00000538, 0.00000419, -0.00000249, },
		{ -0.00000372, 0.00000035, -0.00000127, 0.00000020, -0.00000340, 0.00000304, 0.00000577, -0.00000239, -0.00000575, -0.00000016, 0.00000137, 0.00000057, -0.00000593, 0.00000239, 0.00000253, 0.00000021, -0.00000463, 0.00000238, 0.00000220, -0.00000037, -0.00001202, 0.00000155, 0.00001215, -0.00000409, -0.00001118, 0.00000022, 0.00000324, -0.00000637, -0.00001860, 0.00000083, -0.00001025, -0.00001306, -0.00007100, 0.00000375, 0.00009496, 0.00004632, 0.00010412, -0.00001883, -0.00006057, -0.00000885, 0.00000114, -0.00000812, -0.00000894, 0.00000029, 0.00000851, -0.00000555, -0.00000354, -0.00000234, 0.00000835, -0.00000297, -0.00000426, -0.00000152, 0.00000646, -0.00000448, -0.00000899, 0.00000305, 0.00000621, -0.00000157, -0.00000360, -0.00000232, 0.00000580, -0.00000140, -0.00000535, 0.00000004, 0.00000405, -0.00000102, -0.00000114, -0.00000088, 0.00000477, -0.00000199, -0.00000405, 0.00000038, 0.00000429, -0.00000284, -0.00000301, -0.00000159, 0.00000768, -0.00000076, -0.00000987, 0.00000116, 0.00000379, -0.00000242, -0.00000127, -0.00000323, 0.00000641, -0.00000047, -0.00000545, -0.00000174, 0.00000773, -0.00000439, -0.00000233, -0.00001418, 0.00000723, -0.00001977, -0.00004949, 0.00002878, 0.00006328, 0.00003524, 0.00006074, 0.00001650, -0.00005529, -0.00001706, -0.00000496, -0.00000507, -0.00001802, 0.00000250, -0.00000518, 0.00000424, -0.00001215, -0.00000024, 0.00001304, -0.00000618, -0.00000665, 0.00000082, 0.00000122, 0.00000256, -0.00000501, 0.00000145, 0.00000224, -0.00000052, -0.00000368, 0.00000144, 0.00000015, 0.00000027, -0.00000705, 0.00000040, 0.00000636, 0.00000046, },
		{ -0.00000258, -0.00000202, 0.00000029, -0.00000110, -0.00000086, -0.00000208, 0.00000713, 0.00000110, -0.00000494, -0.00000354, 0.00000148, 0.00000152, -0.00000343, -0.00000310, 0.00000493, 0.00000036, -0.00000176, -0.00000299, 0.00000457, 0.00000051, -0.00000589, -0.00001009, 0.00001312, 0.00000291, -0.00000885, -0.00000814, 0.00000568, -0.00000141, -0.00001485, -0.00001037, 0.00000016, -0.00001361, -0.00003462, -0.00007024, 0.00008362, 0.00006162, 0.00008685, 0.00005847, -0.00007295, -0.00002688, -0.00000239, -0.00000093, -0.00001157, -0.00000495, 0.00000711, 0.00000169, -0.00000645, -0.00000296, 0.00000302, 0.00000559, -0.00000707, -0.00000216, 0.00000236, 0.00000395, -0.00001126, -0.00000178, 0.00000587, 0.00000293, -0.00000371, -0.00000350, 0.00000199, 0.00000438, -0.00000603, -0.00000242, 0.00000275, 0.00000204, -0.00000191, -0.00000115, 0.00000204, 0.00000290, -0.00000495, -0.00000154, 0.00000293, 0.00000146, -0.00000432, -0.00000220, 0.00000261, 0.00000680, -0.00000909, -0.00000428, 0.00000457, -0.00000022, -0.00000153, -0.00000278, 0.00000096, 0.00000560, -0.00000488, -0.00000452, 0.00000479, 0.00000192, -0.00000419, -0.00000733, -0.00000801, 0.00000660, -0.00004983, -0.00000997, 0.00005660, 0.00004572, 0.00004036, 0.00005522, -0.00003210, -0.00003913, -0.00000765, -0.00000182, -0.00001609, -0.00000580, -0.00000055, -0.00000170, -0.00000036, -0.00001330, 0.00001339, 0.00000219, -0.00000903, -0.00000148, 0.00000081, 0.00000280, -0.00000160, -0.00000305, 0.00000350, 0.00000040, -0.00000246, -0.00000134, 0.00000147, 0.00000019, -0.00000309, -0.00000579, 0.00000616, 0.00000332, },
		{ -0.00000010, -0.00000376, 0.00000054, -0.00000150, 0.00000040, -0.00000336, 0.00000376, 0.00000516, -0.00000270, -0.00000545, -0.00000026, 0.00000166, 0.00000011, -0.00000631, 0.00000350, 0.00000200, 0.00000028, -0.00000459, 0.00000227, 0.00000241, -0.00000085, -0.00001306, 0.00000400, 0.00001068, -0.00000415, -0.00001167, 0.00000041, 0.00000418, -0.00000819, -0.00001817, 0.00000179, -0.00001165, -0.00001195, -0.00007441, -0.00000116, 0.00010459, 0.00005203, 0.00011303, -0.00003632, -0.00005739, -0.00000876, 0.00000062, -0.00000704, -0.00001026, 0.00000146, 0.00000823, -0.00000617, -0.00000302, -0.00000294, 0.00000909, -0.00000349, -0.00000462, -0.00000115, 0.00000687, -0.00000617, -0.00000795, 0.00000344, 0.00000586, -0.00000145, -0.00000367, -0.00000224, 0.00000625, -0.00000232, -0.00000516, 0.00000040, 0.00000380, -0.00000100, -0.00000115, -0.00000075, 0.00000476, -0.00000241, -0.00000342, 0.00000040, 0.00000406, -0.00000306, -0.00000276, -0.00000142, 0.00000782, -0.00000202, -0.00000934, 0.00000215, 0.00000295, -0.00000231, -0.00000049, -0.00000369, 0.00000623, -0.00000022, -0.00000582, -0.00000044, 0.00000679, -0.00000497, -0.00000117, -0.00001506, 0.00000834, -0.00002325, -0.00004227, 0.00003421, 0.00005463, 0.00003312, 0.00005447, 0.00001721, -0.00005022, -0.00001859, -0.00000314, -0.00000765, -0.00001639, 0.00000387, -0.00000733, 0.00000578, -0.00001281, 0.00000137, 0.00001209, -0.00000736, -0.00000522, 0.00000092, 0.00000106, 0.00000235, -0.00000489, 0.00000184, 0.00000172, -0.00000078, -0.00000312, 0.00000143, -0.00000008, 0.00000024, -0.00000722, 0.00000132, 0.00000608, },
		{ 0.00000246, -0.00000269, -0.00000198, 0.00000037, -0.00000129, -0.00000055, -0.00000187, 0.00000756, -0.00000002, -0.00000433, -0.00000386, 0.00000190, 0.00000126, -0.00000395, -0.00000267, 0.00000559, -0.00000056, -0.00000098, -0.00000364, 0.00000506, 0.00000019, -0.00000681, -0.00000942, 0.00001456, 0.00000050, -0.00000744, -0.00001001, 0.00000817, -0.00000356, -0.00001442, -0.00001053, 0.00000127, -0.00001601, -0.00003185, -0.00007985, 0.00008931, 0.00006783, 0.00009871, 0.00005063, -0.00008316, -0.00001963, -0.00000694, 0.00000246, -0.00001379, -0.00000378, 0.00000699, 0.00000143, -0.00000671, -0.00000264, 0.00000266, 0.00000629, -0.00000819, -0.00000172, 0.00000276, 0.00000326, -0.00001210, -0.00000009, 0.00000504, 0.00000329, -0.00000395, -0.00000337, 0.00000224, 0.00000427, -0.00000688, -0.00000153, 0.00000239, 0.00000213, -0.00000204, -0.00000098, 0.00000210, 0.00000259, -0.00000486, -0.00000101, 0.00000261, 0.00000128, -0.00000438, -0.00000187, 0.00000284, 0.00000615, -0.00000983, -0.00000283, 0.00000430, -0.00000076, -0.00000078, -0.00000263, 0.00000037, 0.00000592, -0.00000517, -0.00000398, 0.00000549, 0.00000007, -0.00000286, -0.00000796, -0.00000705, 0.00000460, -0.00004893, -0.00000012, 0.00005243, 0.00003999, 0.00003764, 0.00004985, -0.00002647, -0.00003836, -0.00000746, -0.00000180, -0.00001799, -0.00000211, -0.00000271, -0.00000119, -0.00000012, -0.00001292, 0.00001435, 0.00000003, -0.00000823, -0.00000070, 0.00000068, 0.00000256, -0.00000169, -0.00000263, 0.00000345, -0.00000024, -0.00000206, -0.00000107, 0.00000126, 0.00000009, -0.00000327, -0.00000536, 0.00000676, },
		{ 0.00000567, -0.00000065, -0.00000373, 0.00000072, -0.00000177, 0.00000067, -0.00000334, 0.00000448, 0.00000439, -0.00000282, -0.00000521, -0.00000033, 0.00000189, -0.00000045, -0.00000654, 0.00000465, 0.00000125, 0.00000057, -0.00000465, 0.00000219, 0.00000255, -0.00000140, -0.00001388, 0.00000650, 0.00000880, -0.00000387, -0.00001242, 0.00000113, 0.00000438, -0.00000947, -0.00001785, 0.00000279, -0.00001331, -0.00001054, -0.00007786, -0.00000644, 0.00011410, 0.00006007, 0.00011973, -0.00005492, -0.00005187, -0.00001011, 0.00000089, -0.00000645, -0.00001115, 0.00000240, 0.00000788, -0.00000668, -0.00000251, -0.00000350, 0.00000985, -0.00000421, -0.00000475, -0.00000078, 0.00000714, -0.00000788, -0.00000663, 0.00000354, 0.00000558, -0.00000131, -0.00000374, -0.00000214, 0.00000666, -0.00000337, -0.00000475, 0.00000064, 0.00000356, -0.00000098, -0.00000116, -0.00000058, 0.00000465, -0.00000272, -0.00000280, 0.00000032, 0.00000387, -0.00000331, -0.00000248, -0.00000115, 0.00000779, -0.00000326, -0.00000859, 0.00000295, 0.00000207, -0.00000200, 0.00000013, -0.00000416, 0.00000612, -0.00000004, -0.00000612, 0.00000098, 0.00000551, -0.00000515, -0.00000048, -0.00001531, 0.00000872, -0.00002612, -0.00003433, 0.00003745, 0.00004710, 0.00003074, 0.00004935, 0.00001681, -0.00004459, -0.00002046, -0.00000100, -0.00001078, -0.00001391, 0.00000446, -0.00000899, 0.00000715, -0.00001340, 0.00000302, 0.00001079, -0.00000822, -0.00000381, 0.00000093, 0.00000097, 0.00000198, -0.00000457, 0.00000212, 0.00000116, -0.00000087, -0.00000260, 0.00000135, -0.00000025, 0.00000012, -0.00000724, 0.00000224, },
		{ 0.00000716, 0.00000161, -0.00000280, -0.00000182, 0.00000031, -0.00000136, -0.00000029, -0.00000158, 0.00000781, -0.00000105, -0.00000367, -0.00000421, 0.00000233, 0.00000089, -0.00000445, -0.00000202, 0.00000600, -0.00000137, -0.00000017, -0.00000433, 0.00000556, -0.00000020, -0.00000771, -0.00000839, 0.00001551, -0.00000173, -0.00000615, -0.00001156, 0.00001047, -0.00000594, -0.00001349, -0.00001098, 0.00000245, -0.00001856, -0.00002874, -0.00008974, 0.00009443, 0.00007545, 0.00011138, 0.00003908, -0.00009082, -0.00001282, -0.00001163, 0.00000602, -0.00001615, -0.00000233, 0.00000654, 0.00000129, -0.00000692, -0.00000239, 0.00000245, 0.00000671, -0.00000918, -0.00000119, 0.00000311, 0.00000235, -0.00001260, 0.00000149, 0.00000414, 0.00000371, -0.00000416, -0.00000327, 0.00000259, 0.00000391, -0.00000754, -0.00000062, 0.00000197, 0.00000223, -0.00000217, -0.00000081, 0.00000217, 0.00000221, -0.00000457, -0.00000069, 0.00000237, 0.00000102, -0.00000435, -0.00000154, 0.00000313, 0.00000524, -0.00001024, -0.00000149, 0.00000393, -0.00000123, 0.00000010, -0.00000277, 0.00000001, 0.00000607, -0.00000539, -0.00000336, 0.00000607, -0.00000181, -0.00000140, -0.00000882, -0.00000558, 0.00000155, -0.00004595, 0.00000762, 0.00004846, 0.00003436, 0.00003568, 0.00004435, -0.00002121, -0.00003735, -0.00000732, -0.00000197, -0.00001969, 0.00000174, -0.00000539, -0.00000001, -0.00000045, -0.00001200, 0.00001476, -0.00000190, -0.00000734, 0.00000002, 0.00000053, 0.00000230, -0.00000183, -0.00000203, 0.00000313, -0.00000067, -0.00000168, -0.00000085, 0.00000103, 0.00000001, -0.00000350, -0.00000473, },
		{ 0.00000310, 0.00000516, -0.00000119, -0.00000359, 0.00000085, -0.00000204, 0.00000100, -0.00000331, 0.00000517, 0.00000350, -0.00000277, -0.00000504, -0.00000035, 0.00000206, -0.00000110, -0.00000655, 0.00000576, 0.00000034, 0.00000106, -0.00000482, 0.00000216, 0.00000262, -0.00000202, -0.00001441, 0.00000892, 0.00000660, -0.00000331, -0.00001333, 0.00000231, 0.00000383, -0.00001010, -0.00001775, 0.00000389, -0.00001528, -0.00000879, -0.00008142, -0.00001200, 0.00012330, 0.00007095, 0.00012320, -0.00007372, -0.00004441, -0.00001284, 0.00000196, -0.00000642, -0.00001155, 0.00000303, 0.00000752, -0.00000709, -0.00000203, -0.00000399, 0.00001054, -0.00000509, -0.00000464, -0.00000043, 0.00000721, -0.00000951, -0.00000509, 0.00000338, 0.00000539, -0.00000115, -0.00000381, -0.00000196, 0.00000696, -0.00000449, -0.00000412, 0.00000076, 0.00000334, -0.00000095, -0.00000116, -0.00000037, 0.00000440, -0.00000286, -0.00000225, 0.00000019, 0.00000367, -0.00000356, -0.00000217, -0.00000079, 0.00000754, -0.00000440, -0.00000769, 0.00000357, 0.00000116, -0.00000146, 0.00000053, -0.00000457, 0.00000602, 0.00000008, -0.00000634, 0.00000247, 0.00000391, -0.00000491, -0.00000031, -0.00001482, 0.00000815, -0.00002795, -0.00002633, 0.00003893, 0.00004062, 0.00002806, 0.00004533, 0.00001539, -0.00003849, -0.00002263, 0.00000141, -0.00001441, -0.00001058, 0.00000414, -0.00000995, 0.00000816, -0.00001378, 0.00000460, 0.00000924, -0.00000877, -0.00000245, 0.00000088, 0.00000094, 0.00000145, -0.00000403, 0.00000223, 0.00000062, -0.00000084, -0.00000214, 0.00000120, -0.00000035, -0.00000011, -0.00000710, },
		{ -0.00000392, 0.00000735, 0.00000080, -0.00000290, -0.00000156, 0.00000011, -0.00000129, -0.00000011, -0.00000119, 0.00000785, -0.00000195, -0.00000300, -0.00000455, 0.00000276, 0.00000041, -0.00000492, -0.00000115, 0.00000610, -0.00000202, 0.00000060, -0.00000501, 0.00000601, -0.00000066, -0.00000858, -0.00000702, 0.00001592, -0.00000367, -0.00000509, -0.00001263, 0.00001232, -0.00000828, -0.00001225, -0.00001156, 0.00000349, -0.00002106, -0.00002553, -0.00009957, 0.00009854, 0.00008513, 0.00012426, 0.00002384, -0.00009539, -0.00000705, -0.00001603, 0.00000944, -0.00001844, -0.00000077, 0.00000583, 0.00000127, -0.00000707, -0.00000220, 0.00000239, 0.00000681, -0.00000994, -0.00000062, 0.00000342, 0.00000121, -0.00001271, 0.00000285, 0.00000321, 0.00000416, -0.00000432, -0.00000319, 0.00000302, 0.00000329, -0.00000796, 0.00000023, 0.00000152, 0.00000234, -0.00000227, -0.00000062, 0.00000220, 0.00000178, -0.00000411, -0.00000056, 0.00000224, 0.00000066, -0.00000423, -0.00000120, 0.00000342, 0.00000411, -0.00001032, -0.00000030, 0.00000350, -0.00000159, 0.00000104, -0.00000314, -0.00000010, 0.00000605, -0.00000551, -0.00000266, 0.00000646, -0.00000360, 0.00000005, -0.00000973, -0.00000382, -0.00000223, -0.00004111, 0.00001307, 0.00004512, 0.00002855, 0.00003461, 0.00003861, -0.00001618, -0.00003629, -0.00000704, -0.00000250, -0.00002094, 0.00000545, -0.00000834, 0.00000171, -0.00000130, -0.00001057, 0.00001459, -0.00000354, -0.00000641, 0.00000068, 0.00000039, 0.00000199, -0.00000197, -0.00000132, 0.00000258, -0.00000086, -0.00000135, -0.00000066, 0.00000079, -0.00000006, -0.00000376, },
		{ -0.00000675, 0.00000387, 0.00000458, -0.00000174, -0.00000334, 0.00000089, -0.00000226, 0.00000133, -0.00000325, 0.00000580, 0.00000252, -0.00000257, -0.00000493, -0.00000031, 0.00000214, -0.00000183, -0.00000631, 0.00000676, -0.00000067, 0.00000170, -0.00000510, 0.00000217, 0.00000261, -0.00000271, -0.00001460, 0.00001115, 0.00000422, -0.00000260, -0.00001425, 0.00000379, 0.00000263, -0.00001010, -0.00001788, 0.00000508, -0.00001753, -0.00000671, -0.00008506, -0.00001785, 0.00013216, 0.00008508, 0.00012248, -0.00009166, -0.00003557, -0.00001671, 0.00000376, -0.00000694, -0.00001143, 0.00000332, 0.00000717, -0.00000738, -0.00000163, -0.00000433, 0.00001106, -0.00000603, -0.00000432, -0.00000009, 0.00000707, -0.00001099, -0.00000340, 0.00000299, 0.00000529, -0.00000097, -0.00000392, -0.00000169, 0.00000709, -0.00000562, -0.00000332, 0.00000075, 0.00000315, -0.00000091, -0.00000117, -0.00000011, 0.00000403, -0.00000281, -0.00000181, 0.00000008, 0.00000343, -0.00000377, -0.00000184, -0.00000033, 0.00000705, -0.00000538, -0.00000670, 0.00000404, 0.00000026, -0.00000071, 0.00000065, -0.00000483, 0.00000590, 0.00000015, -0.00000648, 0.00000399, 0.00000204, -0.00000427, -0.00000068, -0.00001354, 0.00000652, -0.00002841, -0.00001886, 0.00003920, 0.00003502, 0.00002515, 0.00004229, 0.00001310, -0.00003204, -0.00002499, 0.00000399, -0.00001839, -0.00000647, 0.00000287, -0.00001006, 0.00000865, -0.00001382, 0.00000596, 0.00000755, -0.00000907, -0.00000115, 0.00000077, 0.00000097, 0.00000077, -0.00000325, 0.00000210, 0.00000018, -0.00000073, -0.00000173, 0.00000098, -0.00000039, -0.00000046, },
		{ -0.00000401, -0.00000297, 0.00000734, 0.00000001, -0.00000298, -0.00000123, -0.00000021, -0.00000107, -0.00000002, -0.00000069, 0.00000765, -0.00000267, -0.00000237, -0.00000484, 0.00000315, -0.00000016, -0.00000531, -0.00000008, 0.00000588, -0.00000242, 0.00000125, -0.00000561, 0.00000637, -0.00000116, -0.00000937, -0.00000533, 0.00001576, -0.00000525, -0.00000431, -0.00001310, 0.00001353, -0.00001033, -0.00001093, -0.00001210, 0.00000422, -0.00002327, -0.00002252, -0.00010899, 0.00010115, 0.00009762, 0.00013650, 0.00000523, -0.00009650, -0.00000279, -0.00001977, 0.00001244, -0.00002043, 0.00000077, 0.00000494, 0.00000130, -0.00000714, -0.00000208, 0.00000250, 0.00000655, -0.00001041, -0.00000006, 0.00000369, -0.00000013, -0.00001241, 0.00000394, 0.00000232, 0.00000462, -0.00000445, -0.00000314, 0.00000352, 0.00000239, -0.00000808, 0.00000097, 0.00000109, 0.00000242, -0.00000234, -0.00000042, 0.00000220, 0.00000137, -0.00000354, -0.00000061, 0.00000221, 0.00000019, -0.00000400, -0.00000085, 0.00000370, 0.00000280, -0.00001008, 0.00000071, 0.00000301, -0.00000181, 0.00000195, -0.00000371, 0.00000001, 0.00000583, -0.00000553, -0.00000187, 0.00000662, -0.00000519, 0.00000133, -0.00001049, -0.00000207, -0.00000637, -0.00003479, 0.00001626, 0.00004270, 0.00002235, 0.00003451, 0.00003253, -0.00001123, -0.00003533, -0.00000649, -0.00000355, -0.00002151, 0.00000873, -0.00001126, 0.00000378, -0.00000256, -0.00000869, 0.00001388, -0.00000485, -0.00000546, 0.00000128, 0.00000028, 0.00000159, -0.00000203, -0.00000058, 0.00000184, -0.00000082, -0.00000109, -0.00000051, 0.00000055, -0.00000016, },
		{ -0.00000092, -0.00000619, 0.00000450, 0.00000396, -0.00000229, -0.00000294, 0.00000079, -0.00000238, 0.00000162, -0.00000312, 0.00000631, 0.00000152, -0.00000225, -0.00000487, -0.00000024, 0.00000213, -0.00000264, -0.00000578, 0.00000754, -0.00000167, 0.00000245, -0.00000544, 0.00000223, 0.00000252, -0.00000349, -0.00001438, 0.00001303, 0.00000180, -0.00000185, -0.00001501, 0.00000535, 0.00000092, -0.00000953, -0.00001819, 0.00000628, -0.00001995, -0.00000443, -0.00008862, -0.00002414, 0.00014077, 0.00010268, 0.00011666, -0.00010765, -0.00002608, -0.00002139, 0.00000610, -0.00000789, -0.00001084, 0.00000328, 0.00000683, -0.00000754, -0.00000133, -0.00000448, 0.00001131, -0.00000693, -0.00000385, 0.00000025, 0.00000665, -0.00001223, -0.00000166, 0.00000241, 0.00000526, -0.00000076, -0.00000406, -0.00000128, 0.00000698, -0.00000665, -0.00000242, 0.00000064, 0.00000298, -0.00000087, -0.00000118, 0.00000019, 0.00000355, -0.00000256, -0.00000155, 0.00000003, 0.00000312, -0.00000392, -0.00000152, 0.00000023, 0.00000629, -0.00000615, -0.00000567, 0.00000436, -0.00000063, 0.00000025, 0.00000045, -0.00000488, 0.00000569, 0.00000022, -0.00000652, 0.00000548, -0.00000002, -0.00000329, -0.00000153, -0.00001154, 0.00000382, -0.00002730, -0.00001244, 0.00003880, 0.00003002, 0.00002220, 0.00004001, 0.00001017, -0.00002544, -0.00002738, 0.00000655, -0.00002254, -0.00000177, 0.00000068, -0.00000925, 0.00000849, -0.00001341, 0.00000700, 0.00000584, -0.00000913, 0.00000009, 0.00000061, 0.00000104, -0.00000003, -0.00000225, 0.00000172, -0.00000009, -0.00000059, -0.00000137, 0.00000072, -0.00000036, },
		{ -0.00000030, -0.00000421, -0.00000192, 0.00000714, -0.00000074, -0.00000301, -0.00000085, -0.00000063, -0.00000071, -0.00000004, -0.00000009, 0.00000723, -0.00000319, -0.00000183, -0.00000506, 0.00000346, -0.00000083, -0.00000557, 0.00000113, 0.00000534, -0.00000254, 0.00000173, -0.00000608, 0.00000661, -0.00000169, -0.00001007, -0.00000339, 0.00001502, -0.00000643, -0.00000385, -0.00001292, 0.00001392, -0.00001187, -0.00000973, -0.00001242, 0.00000443, -0.00002496, -0.00001997, -0.00011763, 0.00010175, 0.00011377, 0.00014703, -0.00001619, -0.00009407, -0.00000040, -0.00002252, 0.00001477, -0.00002192, 0.00000210, 0.00000398, 0.00000135, -0.00000712, -0.00000202, 0.00000272, 0.00000592, -0.00001053, 0.00000044, 0.00000390, -0.00000164, -0.00001169, 0.00000471, 0.00000154, 0.00000505, -0.00000454, -0.00000308, 0.00000404, 0.00000125, -0.00000789, 0.00000155, 0.00000070, 0.00000246, -0.00000238, -0.00000021, 0.00000214, 0.00000100, -0.00000291, -0.00000080, 0.00000227, -0.00000037, -0.00000367, -0.00000048, 0.00000390, 0.00000138, -0.00000955, 0.00000154, 0.00000247, -0.00000184, 0.00000276, -0.00000438, 0.00000032, 0.00000543, -0.00000543, -0.00000099, 0.00000649, -0.00000647, 0.00000230, -0.00001092, -0.00000060, -0.00001044, -0.00002745, 0.00001740, 0.00004130, 0.00001567, 0.00003540, 0.00002608, -0.00000630, -0.00003456, -0.00000559, -0.00000522, -0.00002122, 0.00001127, -0.00001385, 0.00000596, -0.00000411, -0.00000649, 0.00001267, -0.00000580, -0.00000454, 0.00000181, 0.00000020, 0.00000109, -0.00000197, 0.00000012, 0.00000100, -0.00000057, -0.00000092, -0.00000039, 0.00000032, },
		{ -0.00000028, -0.00000148, -0.00000541, 0.00000496, 0.00000332, -0.00000286, -0.00000240, 0.00000052, -0.00000234, 0.00000183, -0.00000290, 0.00000666, 0.00000056, -0.00000187, -0.00000483, -0.00000014, 0.00000203, -0.00000350, -0.00000493, 0.00000802, -0.00000257, 0.00000321, -0.00000580, 0.00000227, 0.00000238, -0.00000435, -0.00001370, 0.00001446, -0.00000053, -0.00000120, -0.00001542, 0.00000676, -0.00000106, -0.00000855, -0.00001857, 0.00000732, -0.00002235, -0.00000217, -0.00009187, -0.00003116, 0.00014946, 0.00012374, 0.00010503, -0.00012061, -0.00001674, -0.00002639, 0.00000869, -0.00000909, -0.00000989, 0.00000294, 0.00000647, -0.00000755, -0.00000118, -0.00000436, 0.00001119, -0.00000768, -0.00000328, 0.00000062, 0.00000592, -0.00001312, 0.00000001, 0.00000172, 0.00000528, -0.00000052, -0.00000426, -0.00000072, 0.00000657, -0.00000751, -0.00000147, 0.00000046, 0.00000281, -0.00000080, -0.00000119, 0.00000052, 0.00000295, -0.00000211, -0.00000149, 0.00000009, 0.00000268, -0.00000397, -0.00000120, 0.00000086, 0.00000527, -0.00000665, -0.00000468, 0.00000455, -0.00000145, 0.00000137, -0.00000009, -0.00000467, 0.00000534, 0.00000030, -0.00000645, 0.00000687, -0.00000216, -0.00000207, -0.00000276, -0.00000891, 0.00000016, -0.00002453, -0.00000750, 0.00003824, 0.00002532, 0.00001939, 0.00003825, 0.00000685, -0.00001892, -0.00002961, 0.00000889, -0.00002658, 0.00000329, -0.00000227, -0.00000751, 0.00000760, -0.00001247, 0.00000763, 0.00000419, -0.00000902, 0.00000124, 0.00000043, 0.00000111, -0.00000091, -0.00000106, 0.00000106, -0.00000015, -0.00000048, -0.00000103, 0.00000041, },
		{ 0.00000012, -0.00000050, -0.00000431, -0.00000082, 0.00000677, -0.00000146, -0.00000296, -0.00000048, -0.00000112, -0.00000023, -0.00000015, 0.00000060, 0.00000658, -0.00000349, -0.00000139, -0.00000517, 0.00000369, -0.00000159, -0.00000568, 0.00000242, 0.00000449, -0.00000235, 0.00000200, -0.00000639, 0.00000669, -0.00000223, -0.00001062, -0.00000127, 0.00001374, -0.00000721, -0.00000368, -0.00001210, 0.00001343, -0.00001274, -0.00000882, -0.00001238, 0.00000400, -0.00002596, -0.00001811, -0.00012517, 0.00009985, 0.00013443, 0.00015463, -0.00003952, -0.00008826, -0.00000005, -0.00002408, 0.00001622, -0.00002274, 0.00000311, 0.00000303, 0.00000139, -0.00000701, -0.00000197, 0.00000301, 0.00000495, -0.00001027, 0.00000084, 0.00000404, -0.00000328, -0.00001058, 0.00000513, 0.00000089, 0.00000544, -0.00000460, -0.00000300, 0.00000453, -0.00000008, -0.00000739, 0.00000195, 0.00000037, 0.00000246, -0.00000237, 0.00000001, 0.00000201, 0.00000072, -0.00000228, -0.00000108, 0.00000238, -0.00000100, -0.00000323, -0.00000008, 0.00000400, -0.00000008, -0.00000878, 0.00000217, 0.00000189, -0.00000166, 0.00000338, -0.00000508, 0.00000078, 0.00000487, -0.00000523, -0.00000005, 0.00000606, -0.00000737, 0.00000286, -0.00001085, 0.00000034, -0.00001400, -0.00001966, 0.00001684, 0.00004087, 0.00000855, 0.00003719, 0.00001931, -0.00000135, -0.00003404, -0.00000429, -0.00000753, -0.00001994, 0.00001286, -0.00001583, 0.00000801, -0.00000575, -0.00000407, 0.00001104, -0.00000642, -0.00000365, 0.00000226, 0.00000018, 0.00000046, -0.00000174, 0.00000071, 0.00000013, -0.00000015, -0.00000085, -0.00000029, },
		{ 0.00000006, -0.00000016, -0.00000213, -0.00000441, 0.00000522, 0.00000268, -0.00000343, -0.00000171, 0.00000007, -0.00000212, 0.00000190, -0.00000253, 0.00000680, -0.00000030, -0.00000147, -0.00000476, -0.00000005, 0.00000185, -0.00000440, -0.00000374, 0.00000812, -0.00000326, 0.00000388, -0.00000610, 0.00000226, 0.00000222, -0.00000531, -0.00001252, 0.00001531, -0.00000260, -0.00000080, -0.00001530, 0.00000774, -0.00000303, -0.00000739, -0.00001880, 0.00000800, -0.00002447, -0.00000022, -0.00009445, -0.00003933, 0.00015877, 0.00014794, 0.00008712, -0.00012955, -0.00000840, -0.00003117, 0.00001117, -0.00001030, -0.00000875, 0.00000242, 0.00000605, -0.00000735, -0.00000120, -0.00000394, 0.00001061, -0.00000814, -0.00000272, 0.00000104, 0.00000484, -0.00001357, 0.00000151, 0.00000100, 0.00000532, -0.00000022, -0.00000453, 0.00000001, 0.00000581, -0.00000809, -0.00000056, 0.00000025, 0.00000261, -0.00000070, -0.00000122, 0.00000086, 0.00000228, -0.00000146, -0.00000164, 0.00000028, 0.00000210, -0.00000389, -0.00000092, 0.00000156, 0.00000400, -0.00000687, -0.00000375, 0.00000463, -0.00000218, 0.00000260, -0.00000094, -0.00000417, 0.00000483, 0.00000043, -0.00000627, 0.00000810, -0.00000429, -0.00000074, -0.00000422, -0.00000586, -0.00000428, -0.00002019, -0.00000427, 0.00003795, 0.00002063, 0.00001695, 0.00003674, 0.00000343, -0.00001274, -0.00003145, 0.00001077, -0.00003023, 0.00000841, -0.00000579, -0.00000494, 0.00000599, -0.00001098, 0.00000778, 0.00000270, -0.00000876, 0.00000230, 0.00000025, 0.00000116, -0.00000181, 0.00000027, 0.00000015, 0.00000005, -0.00000043, -0.00000071, },
		{ -0.00000021, -0.00000005, -0.00000078, -0.00000430, 0.00000030, 0.00000625, -0.00000215, -0.00000281, -0.00000016, -0.00000161, 0.00000033, -0.00000032, 0.00000135, 0.00000574, -0.00000356, -0.00000107, -0.00000516, 0.00000378, -0.00000242, -0.00000557, 0.00000370, 0.00000341, -0.00000188, 0.00000203, -0.00000651, 0.00000659, -0.00000279, -0.00001097, 0.00000094, 0.00001199, -0.00000764, -0.00000372, -0.00001070, 0.00001205, -0.00001285, -0.00000830, -0.00001190, 0.00000284, -0.00002614, -0.00001712, -0.00013138, 0.00009509, 0.00016043, 0.00015798, -0.00006366, -0.00007950, -0.00000171, -0.00002436, 0.00001670, -0.00002279, 0.00000369, 0.00000217, 0.00000140, -0.00000681, -0.00000191, 0.00000330, 0.00000371, -0.00000964, 0.00000114, 0.00000406, -0.00000496, -0.00000912, 0.00000519, 0.00000042, 0.00000577, -0.00000463, -0.00000284, 0.00000494, -0.00000156, -0.00000658, 0.00000214, 0.00000012, 0.00000241, -0.00000232, 0.00000024, 0.00000181, 0.00000056, -0.00000171, -0.00000140, 0.00000251, -0.00000167, -0.00000272, 0.00000034, 0.00000394, -0.00000150, -0.00000782, 0.00000262, 0.00000130, -0.00000124, 0.00000373, -0.00000570, 0.00000131, 0.00000419, -0.00000490, 0.00000094, 0.00000535, -0.00000787, 0.00000293, -0.00001019, 0.00000055, -0.00001668, -0.00001198, 0.00001504, 0.00004121, 0.00000112, 0.00003971, 0.00001234, 0.00000354, -0.00003373, -0.00000265, -0.00001045, -0.00001765, 0.00001332, -0.00001695, 0.00000968, -0.00000733, -0.00000161, 0.00000911, -0.00000671, -0.00000281, 0.00000265, 0.00000020, -0.00000030, -0.00000131, 0.00000110, -0.00000068, 0.00000039, -0.00000087, },
		{ -0.00000039, -0.00000032, -0.00000002, -0.00000283, -0.00000321, 0.00000527, 0.00000206, -0.00000399, -0.00000090, -0.00000058, -0.00000168, 0.00000180, -0.00000198, 0.00000669, -0.00000100, -0.00000112, -0.00000462, -0.00000000, 0.00000159, -0.00000532, -0.00000222, 0.00000777, -0.00000363, 0.00000437, -0.00000625, 0.00000212, 0.00000208, -0.00000638, -0.00001081, 0.00001547, -0.00000428, -0.00000075, -0.00001448, 0.00000805, -0.00000472, -0.00000632, -0.00001867, 0.00000804, -0.00002599, 0.00000103, -0.00009592, -0.00004922, 0.00016942, 0.00017466, 0.00006278, -0.00013364, -0.00000187, -0.00003517, 0.00001314, -0.00001123, -0.00000763, 0.00000184, 0.00000549, -0.00000690, -0.00000145, -0.00000318, 0.00000948, -0.00000820, -0.00000224, 0.00000156, 0.00000338, -0.00001352, 0.00000273, 0.00000034, 0.00000532, 0.00000015, -0.00000487, 0.00000092, 0.00000467, -0.00000833, 0.00000024, 0.00000007, 0.00000237, -0.00000054, -0.00000127, 0.00000122, 0.00000154, -0.00000066, -0.00000201, 0.00000064, 0.00000136, -0.00000365, -0.00000066, 0.00000229, 0.00000252, -0.00000678, -0.00000295, 0.00000461, -0.00000278, 0.00000386, -0.00000206, -0.00000338, 0.00000412, 0.00000062, -0.00000596, 0.00000910, -0.00000629, 0.00000057, -0.00000574, -0.00000261, -0.00000919, -0.00001450, -0.00000284, 0.00003822, 0.00001573, 0.00001507, 0.00003524, 0.00000018, -0.00000714, -0.00003270, 0.00001198, -0.00003318, 0.00001324, -0.00000957, -0.00000171, 0.00000372, -0.00000896, 0.00000742, 0.00000142, -0.00000841, 0.00000328, 0.00000007, 0.00000115, -0.00000267, 0.00000167, -0.00000097, 0.00000050, -0.00000049, },
		{ -0.00000096, -0.00000016, -0.00000019, -0.00000113, -0.00000413, 0.00000137, 0.00000561, -0.00000282, -0.00000254, 0.00000006, -0.00000206, 0.00000092, -0.00000053, 0.00000212, 0.00000474, -0.00000343, -0.00000087, -0.00000503, 0.00000374, -0.00000330, -0.00000521, 0.00000489, 0.00000218, -0.00000117, 0.00000185, -0.00000643, 0.00000630, -0.00000336, -0.00001106, 0.00000312, 0.00000988, -0.00000779, -0.00000386, -0.00000888, 0.00000986, -0.00001222, -0.00000819, -0.00001100, 0.00000095, -0.00002550, -0.00001703, -0.00013617, 0.00008724, 0.00019244, 0.00015574, -0.00008729, -0.00006846, -0.00000516, -0.00002341, 0.00001621, -0.00002203, 0.00000378, 0.00000143, 0.00000135, -0.00000655, -0.00000178, 0.00000349, 0.00000230, -0.00000867, 0.00000135, 0.00000393, -0.00000659, -0.00000740, 0.00000492, 0.00000013, 0.00000603, -0.00000465, -0.00000259, 0.00000518, -0.00000308, -0.00000553, 0.00000215, -0.00000005, 0.00000231, -0.00000222, 0.00000047, 0.00000152, 0.00000054, -0.00000126, -0.00000168, 0.00000260, -0.00000234, -0.00000214, 0.00000081, 0.00000371, -0.00000283, -0.00000673, 0.00000290, 0.00000071, -0.00000060, 0.00000376, -0.00000616, 0.00000184, 0.00000342, -0.00000447, 0.00000194, 0.00000438, -0.00000796, 0.00000252, -0.00000889, -0.00000010, -0.00001817, -0.00000497, 0.00001252, 0.00004205, -0.00000637, 0.00004271, 0.00000535, 0.00000825, -0.00003354, -0.00000077, -0.00001384, -0.00001440, 0.00001259, -0.00001703, 0.00001078, -0.00000864, 0.00000076, 0.00000698, -0.00000672, -0.00000203, 0.00000297, 0.00000027, -0.00000116, -0.00000068, 0.00000125, -0.00000136, 0.00000098, },
		{ -0.00000068, -0.00000006, -0.00000073, 0.00000014, -0.00000355, -0.00000183, 0.00000510, 0.00000149, -0.00000454, 0.00000003, -0.00000141, -0.00000100, 0.00000150, -0.00000124, 0.00000629, -0.00000147, -0.00000087, -0.00000437, -0.00000005, 0.00000128, -0.00000623, -0.00000039, 0.00000696, -0.00000361, 0.00000458, -0.00000618, 0.00000180, 0.00000200, -0.00000756, -0.00000856, 0.00001487, -0.00000546, -0.00000116, -0.00001283, 0.00000743, -0.00000581, -0.00000561, -0.00001791, 0.00000716, -0.00002656, 0.00000118, -0.00009576, -0.00006146, 0.00018236, 0.00020297, 0.00003221, -0.00013230, 0.00000214, -0.00003783, 0.00001421, -0.00001157, -0.00000676, 0.00000135, 0.00000469, -0.00000614, -0.00000194, -0.00000206, 0.00000776, -0.00000776, -0.00000193, 0.00000219, 0.00000153, -0.00001290, 0.00000356, -0.00000017, 0.00000523, 0.00000060, -0.00000528, 0.00000198, 0.00000313, -0.00000815, 0.00000084, -0.00000004, 0.00000205, -0.00000030, -0.00000134, 0.00000158, 0.00000078, 0.00000027, -0.00000257, 0.00000115, 0.00000043, -0.00000324, -0.00000045, 0.00000303, 0.00000085, -0.00000639, -0.00000228, 0.00000451, -0.00000320, 0.00000508, -0.00000338, -0.00000231, 0.00000322, 0.00000089, -0.00000554, 0.00000983, -0.00000805, 0.00000173, -0.00000713, 0.00000057, -0.00001422, -0.00000782, -0.00000313, 0.00003922, 0.00001046, 0.00001390, 0.00003354, -0.00000270, -0.00000232, -0.00003317, 0.00001231, -0.00003516, 0.00001744, -0.00001329, 0.00000196, 0.00000091, -0.00000646, 0.00000656, 0.00000042, -0.00000800, 0.00000416, -0.00000008, 0.00000105, -0.00000344, 0.00000306, -0.00000224, 0.00000121, },
		{ 0.00000158, -0.00000111, -0.00000013, -0.00000029, -0.00000158, -0.00000378, 0.00000235, 0.00000488, -0.00000348, -0.00000214, 0.00000013, -0.00000241, 0.00000149, -0.00000073, 0.00000285, 0.00000366, -0.00000314, -0.00000078, -0.00000478, 0.00000356, -0.00000424, -0.00000457, 0.00000588, 0.00000089, -0.00000028, 0.00000148, -0.00000618, 0.00000583, -0.00000395, -0.00001084, 0.00000513, 0.00000755, -0.00000776, -0.00000395, -0.00000684, 0.00000705, -0.00001094, -0.00000842, -0.00000975, -0.00000154, -0.00002414, -0.00001775, -0.00013961, 0.00007629, 0.00023093, 0.00014665, -0.00010899, -0.00005603, -0.00001000, -0.00002145, 0.00001486, -0.00002053, 0.00000341, 0.00000081, 0.00000128, -0.00000625, -0.00000153, 0.00000347, 0.00000084, -0.00000744, 0.00000149, 0.00000360, -0.00000808, -0.00000552, 0.00000438, 0.00000001, 0.00000622, -0.00000467, -0.00000220, 0.00000518, -0.00000454, -0.00000429, 0.00000198, -0.00000014, 0.00000217, -0.00000207, 0.00000070, 0.00000115, 0.00000068, -0.00000096, -0.00000188, 0.00000259, -0.00000295, -0.00000152, 0.00000131, 0.00000328, -0.00000399, -0.00000558, 0.00000303, 0.00000014, 0.00000026, 0.00000342, -0.00000638, 0.00000230, 0.00000262, -0.00000393, 0.00000290, 0.00000319, -0.00000769, 0.00000164, -0.00000698, -0.00000164, -0.00001826, 0.00000091, 0.00000980, 0.00004301, -0.00001363, 0.00004590, -0.00000141, 0.00001260, -0.00003331, 0.00000116, -0.00001749, -0.00001035, 0.00001069, -0.00001598, 0.00001113, -0.00000955, 0.00000288, 0.00000478, -0.00000650, -0.00000130, 0.00000323, 0.00000035, -0.00000211, 0.00000016, 0.00000113, -0.00000182, },
		{ 0.00000213, -0.00000101, 0.00000029, -0.00000116, 0.00000029, -0.00000427, -0.00000031, 0.00000471, 0.00000096, -0.00000504, 0.00000102, -0.00000240, -0.00000011, 0.00000099, -0.00000029, 0.00000558, -0.00000167, -0.00000079, -0.00000395, -0.00000023, 0.00000093, -0.00000710, 0.00000171, 0.00000565, -0.00000313, 0.00000443, -0.00000580, 0.00000121, 0.00000202, -0.00000885, -0.00000578, 0.00001346, -0.00000605, -0.00000209, -0.00001026, 0.00000573, -0.00000605, -0.00000553, -0.00001627, 0.00000509, -0.00002583, -0.00000019, -0.00009348, -0.00007677, 0.00019866, 0.00023161, -0.00000403, -0.00012525, 0.00000308, -0.00003869, 0.00001401, -0.00001103, -0.00000637, 0.00000112, 0.00000357, -0.00000502, -0.00000271, -0.00000058, 0.00000542, -0.00000673, -0.00000186, 0.00000296, -0.00000071, -0.00001167, 0.00000392, -0.00000046, 0.00000502, 0.00000116, -0.00000575, 0.00000317, 0.00000121, -0.00000753, 0.00000118, -0.00000003, 0.00000162, 0.00000004, -0.00000146, 0.00000195, 0.00000002, 0.00000128, -0.00000330, 0.00000180, -0.00000066, -0.00000264, -0.00000029, 0.00000375, -0.00000094, -0.00000573, -0.00000178, 0.00000434, -0.00000342, 0.00000617, -0.00000482, -0.00000103, 0.00000213, 0.00000124, -0.00000501, 0.00001023, -0.00000948, 0.00000259, -0.00000821, 0.00000340, -0.00001897, -0.00000058, -0.00000490, 0.00004096, 0.00000479, 0.00001348, 0.00003152, -0.00000503, 0.00000159, -0.00003276, 0.00001163, -0.00003592, 0.00002067, -0.00001660, 0.00000580, -0.00000228, -0.00000359, 0.00000524, -0.00000030, -0.00000756, 0.00000495, -0.00000019, 0.00000085, -0.00000404, 0.00000436, -0.00000359, },
		{ -0.00000202, 0.00000210, -0.00000127, -0.00000013, -0.00000035, -0.00000210, -0.00000324, 0.00000319, 0.00000409, -0.00000411, -0.00000158, 0.00000003, -0.00000259, 0.00000196, -0.00000085, 0.00000350, 0.00000254, -0.00000273, -0.00000075, -0.00000445, 0.00000323, -0.00000520, -0.00000360, 0.00000658, -0.00000033, 0.00000068, 0.00000100, -0.00000580, 0.00000522, -0.00000460, -0.00001022, 0.00000682, 0.00000515, -0.00000770, -0.00000381, -0.00000483, 0.00000387, -0.00000921, -0.00000884, -0.00000835, -0.00000443, -0.00002229, -0.00001905, -0.00014195, 0.00006250, 0.00027611, 0.00012962, -0.00012729, -0.00004324, -0.00001564, -0.00001881, 0.00001287, -0.00001842, 0.00000265, 0.00000027, 0.00000123, -0.00000598, -0.00000109, 0.00000313, -0.00000054, -0.00000605, 0.00000160, 0.00000299, -0.00000930, -0.00000359, 0.00000365, 0.00000002, 0.00000637, -0.00000471, -0.00000165, 0.00000487, -0.00000585, -0.00000296, 0.00000169, -0.00000019, 0.00000200, -0.00000190, 0.00000093, 0.00000072, 0.00000100, -0.00000087, -0.00000195, 0.00000244, -0.00000345, -0.00000090, 0.00000185, 0.00000263, -0.00000493, -0.00000442, 0.00000302, -0.00000038, 0.00000131, 0.00000272, -0.00000631, 0.00000262, 0.00000184, -0.00000329, 0.00000379, 0.00000188, -0.00000712, 0.00000039, -0.00000456, -0.00000399, -0.00001688, 0.00000530, 0.00000737, 0.00004375, -0.00002032, 0.00004892, -0.00000765, 0.00001637, -0.00003283, 0.00000292, -0.00002115, -0.00000575, 0.00000777, -0.00001378, 0.00001063, -0.00000994, 0.00000463, 0.00000261, -0.00000610, -0.00000063, 0.00000345, 0.00000043, -0.00000310, 0.00000118, 0.00000071, },
		{ -0.00000493, 0.00000323, -0.00000149, 0.00000068, -0.00000161, 0.00000043, -0.00000495, 0.00000131, 0.00000410, 0.00000048, -0.00000548, 0.00000205, -0.00000351, 0.00000099, 0.00000027, 0.00000088, 0.00000455, -0.00000156, -0.00000091, -0.00000333, -0.00000060, 0.00000058, -0.00000790, 0.00000403, 0.00000386, -0.00000215, 0.00000384, -0.00000506, 0.00000031, 0.00000219, -0.00001025, -0.00000252, 0.00001124, -0.00000601, -0.00000356, -0.00000675, 0.00000282, -0.00000524, -0.00000629, -0.00001356, 0.00000157, -0.00002351, -0.00000347, -0.00008859, -0.00009583, 0.00021949, 0.00025909, -0.00004499, -0.00011254, 0.00000059, -0.00003738, 0.00001223, -0.00000936, -0.00000667, 0.00000129, 0.00000204, -0.00000348, -0.00000375, 0.00000122, 0.00000248, -0.00000508, -0.00000210, 0.00000387, -0.00000331, -0.00000983, 0.00000376, -0.00000044, 0.00000463, 0.00000183, -0.00000626, 0.00000445, -0.00000106, -0.00000643, 0.00000121, 0.00000015, 0.00000106, 0.00000049, -0.00000164, 0.00000231, -0.00000071, 0.00000231, -0.00000413, 0.00000255, -0.00000190, -0.00000186, -0.00000017, 0.00000440, -0.00000280, -0.00000481, -0.00000145, 0.00000412, -0.00000341, 0.00000705, -0.00000628, 0.00000041, 0.00000089, 0.00000169, -0.00000437, 0.00001029, -0.00001052, 0.00000308, -0.00000883, 0.00000565, -0.00002306, 0.00000671, -0.00000777, 0.00004330, -0.00000121, 0.00001378, 0.00002913, -0.00000674, 0.00000451, -0.00003141, 0.00000987, -0.00003532, 0.00002266, -0.00001918, 0.00000949, -0.00000563, -0.00000049, 0.00000354, -0.00000073, -0.00000710, 0.00000565, -0.00000027, 0.00000053, -0.00000445, 0.00000549, },
		{ 0.00000002, -0.00000192, 0.00000249, -0.00000142, -0.00000017, -0.00000038, -0.00000270, -0.00000250, 0.00000387, 0.00000327, -0.00000473, -0.00000087, -0.00000028, -0.00000256, 0.00000228, -0.00000085, 0.00000399, 0.00000147, -0.00000228, -0.00000075, -0.00000407, 0.00000277, -0.00000619, -0.00000229, 0.00000689, -0.00000135, 0.00000160, 0.00000051, -0.00000538, 0.00000452, -0.00000533, -0.00000914, 0.00000806, 0.00000284, -0.00000773, -0.00000324, -0.00000315, 0.00000063, -0.00000729, -0.00000920, -0.00000708, -0.00000738, -0.00002029, -0.00002055, -0.00014366, 0.00004640, 0.00032782, 0.00010382, -0.00014082, -0.00003123, -0.00002140, -0.00001594, 0.00001055, -0.00001593, 0.00000163, -0.00000027, 0.00000128, -0.00000579, -0.00000039, 0.00000236, -0.00000167, -0.00000460, 0.00000176, 0.00000204, -0.00001015, -0.00000177, 0.00000282, 0.00000011, 0.00000650, -0.00000479, -0.00000091, 0.00000418, -0.00000688, -0.00000164, 0.00000134, -0.00000022, 0.00000184, -0.00000171, 0.00000117, 0.00000021, 0.00000147, -0.00000099, -0.00000183, 0.00000211, -0.00000380, -0.00000029, 0.00000243, 0.00000176, -0.00000562, -0.00000332, 0.00000289, -0.00000083, 0.00000248, 0.00000166, -0.00000590, 0.00000275, 0.00000112, -0.00000256, 0.00000456, 0.00000051, -0.00000636, -0.00000111, -0.00000177, -0.00000701, -0.00001407, 0.00000798, 0.00000563, 0.00004390, -0.00002612, 0.00005144, -0.00001308, 0.00001932, -0.00003190, 0.00000425, -0.00002447, -0.00000092, 0.00000405, -0.00001052, 0.00000925, -0.00000972, 0.00000591, 0.00000059, -0.00000557, -0.00000000, 0.00000362, 0.00000049, -0.00000408, 0.00000231, },
		{ 0.00000637, -0.00000617, 0.00000445, -0.00000209, 0.00000110, -0.00000208, 0.00000053, -0.00000555, 0.00000299, 0.00000329, 0.00000007, -0.00000585, 0.00000306, -0.00000469, 0.00000226, -0.00000065, 0.00000225, 0.00000322, -0.00000114, -0.00000126, -0.00000248, -0.00000118, 0.00000024, -0.00000859, 0.00000648, 0.00000166, -0.00000068, 0.00000279, -0.00000389, -0.00000095, 0.00000254, -0.00001171, 0.00000116, 0.00000823, -0.00000531, -0.00000553, -0.00000237, -0.00000132, -0.00000326, -0.00000806, -0.00000962, -0.00000353, -0.00001936, -0.00000895, -0.00008068, -0.00011925, 0.00024601, 0.00028371, -0.00008936, -0.00009455, -0.00000543, -0.00003371, 0.00000870, -0.00000638, -0.00000782, 0.00000200, 0.00000004, -0.00000149, -0.00000508, 0.00000330, -0.00000100, -0.00000279, -0.00000269, 0.00000492, -0.00000621, -0.00000740, 0.00000305, -0.00000007, 0.00000403, 0.00000262, -0.00000679, 0.00000575, -0.00000360, -0.00000488, 0.00000091, 0.00000051, 0.00000035, 0.00000108, -0.00000188, 0.00000267, -0.00000139, 0.00000331, -0.00000502, 0.00000336, -0.00000324, -0.00000091, -0.00000010, 0.00000495, -0.00000466, -0.00000370, -0.00000130, 0.00000386, -0.00000316, 0.00000765, -0.00000765, 0.00000193, -0.00000047, 0.00000223, -0.00000365, 0.00001001, -0.00001112, 0.00000311, -0.00000888, 0.00000712, -0.00002613, 0.00001352, -0.00001128, 0.00004600, -0.00000736, 0.00001466, 0.00002643, -0.00000784, 0.00000647, -0.00002917, 0.00000705, -0.00003328, 0.00002321, -0.00002072, 0.00001275, -0.00000892, 0.00000268, 0.00000155, -0.00000090, -0.00000663, 0.00000625, -0.00000031, 0.00000009, -0.00000462, },
		{ 0.00000350, -0.00000091, -0.00000149, 0.00000270, -0.00000152, -0.00000026, -0.00000038, -0.00000336, -0.00000155, 0.00000434, 0.00000245, -0.00000532, -0.00000003, -0.00000081, -0.00000227, 0.00000239, -0.00000067, 0.00000425, 0.00000054, -0.00000184, -0.00000071, -0.00000369, 0.00000220, -0.00000719, -0.00000063, 0.00000674, -0.00000206, 0.00000234, 0.00000011, -0.00000500, 0.00000379, -0.00000619, -0.00000754, 0.00000870, 0.00000080, -0.00000801, -0.00000208, -0.00000207, -0.00000232, -0.00000551, -0.00000922, -0.00000626, -0.00000999, -0.00001861, -0.00002170, -0.00014540, 0.00002882, 0.00038554, 0.00006875, -0.00014833, -0.00002112, -0.00002649, -0.00001337, 0.00000829, -0.00001335, 0.00000056, -0.00000095, 0.00000151, -0.00000575, 0.00000063, 0.00000104, -0.00000239, -0.00000324, 0.00000202, 0.00000070, -0.00001049, -0.00000018, 0.00000200, 0.00000022, 0.00000664, -0.00000492, 0.00000005, 0.00000306, -0.00000754, -0.00000043, 0.00000098, -0.00000029, 0.00000172, -0.00000153, 0.00000143, -0.00000034, 0.00000209, -0.00000134, -0.00000152, 0.00000155, -0.00000395, 0.00000026, 0.00000303, 0.00000067, -0.00000601, -0.00000232, 0.00000269, -0.00000117, 0.00000373, 0.00000029, -0.00000515, 0.00000265, 0.00000050, -0.00000176, 0.00000515, -0.00000082, -0.00000551, -0.00000270, 0.00000118, -0.00001044, -0.00001004, 0.00000887, 0.00000484, 0.00004319, -0.00003073, 0.00005315, -0.00001743, 0.00002124, -0.00003028, 0.00000489, -0.00002715, 0.00000378, -0.00000016, -0.00000639, 0.00000703, -0.00000887, 0.00000666, -0.00000120, -0.00000498, 0.00000058, 0.00000375, 0.00000049, -0.00000500, },
		{ -0.00000454, 0.00000694, -0.00000722, 0.00000569, -0.00000280, 0.00000156, -0.00000257, 0.00000059, -0.00000604, 0.00000469, 0.00000231, -0.00000028, -0.00000612, 0.00000400, -0.00000588, 0.00000364, -0.00000173, 0.00000378, 0.00000161, -0.00000040, -0.00000186, -0.00000138, -0.00000201, -0.00000004, -0.00000915, 0.00000899, -0.00000089, 0.00000126, 0.00000128, -0.00000229, -0.00000260, 0.00000308, -0.00001321, 0.00000517, 0.00000451, -0.00000401, -0.00000792, 0.00000278, -0.00000662, -0.00000009, -0.00001089, -0.00000442, -0.00001029, -0.00001328, -0.00001680, -0.00006952, -0.00014749, 0.00027933, 0.00030369, -0.00013554, -0.00007201, -0.00001482, -0.00002766, 0.00000334, -0.00000201, -0.00000993, 0.00000331, -0.00000249, 0.00000095, -0.00000666, 0.00000557, -0.00000494, 0.00000010, -0.00000361, 0.00000607, -0.00000935, -0.00000446, 0.00000180, 0.00000067, 0.00000320, 0.00000353, -0.00000731, 0.00000700, -0.00000634, -0.00000292, 0.00000027, 0.00000108, -0.00000052, 0.00000181, -0.00000220, 0.00000302, -0.00000198, 0.00000423, -0.00000590, 0.00000418, -0.00000465, 0.00000019, -0.00000008, 0.00000538, -0.00000647, -0.00000244, -0.00000131, 0.00000359, -0.00000266, 0.00000790, -0.00000884, 0.00000342, -0.00000188, 0.00000285, -0.00000287, 0.00000941, -0.00001128, 0.00000267, -0.00000831, 0.00000766, -0.00002788, 0.00001935, -0.00001493, 0.00004871, -0.00001339, 0.00001591, 0.00002353, -0.00000840, 0.00000757, -0.00002615, 0.00000331, -0.00002985, 0.00002224, -0.00002103, 0.00001531, -0.00001192, 0.00000573, -0.00000060, -0.00000084, -0.00000616, 0.00000677, -0.00000033, -0.00000046, },
		{ -0.00000579, 0.00000467, -0.00000203, -0.00000076, 0.00000270, -0.00000155, -0.00000040, -0.00000036, -0.00000405, -0.00000042, 0.00000459, 0.00000166, -0.00000589, 0.00000094, -0.00000156, -0.00000171, 0.00000224, -0.00000025, 0.00000423, -0.00000019, -0.00000151, -0.00000056, -0.00000339, 0.00000158, -0.00000819, 0.00000138, 0.00000606, -0.00000233, 0.00000278, -0.00000008, -0.00000477, 0.00000313, -0.00000721, -0.00000536, 0.00000861, -0.00000081, -0.00000867, -0.00000014, -0.00000188, -0.00000465, -0.00000420, -0.00000854, -0.00000630, -0.00001180, -0.00001779, -0.00002189, -0.00014800, 0.00001086, 0.00044831, 0.00002435, -0.00014884, -0.00001396, -0.00003016, -0.00001165, 0.00000654, -0.00001101, -0.00000035, -0.00000192, 0.00000205, -0.00000595, 0.00000202, -0.00000091, -0.00000256, -0.00000211, 0.00000246, -0.00000111, -0.00001024, 0.00000102, 0.00000133, 0.00000026, 0.00000683, -0.00000514, 0.00000121, 0.00000148, -0.00000772, 0.00000056, 0.00000071, -0.00000044, 0.00000168, -0.00000139, 0.00000170, -0.00000093, 0.00000284, -0.00000190, -0.00000099, 0.00000075, -0.00000385, 0.00000074, 0.00000365, -0.00000060, -0.00000610, -0.00000146, 0.00000241, -0.00000137, 0.00000496, -0.00000133, -0.00000408, 0.00000229, 0.00000003, -0.00000090, 0.00000554, -0.00000203, -0.00000469, -0.00000422, 0.00000405, -0.00001399, -0.00000508, 0.00000809, 0.00000514, 0.00004145, -0.00003396, 0.00005381, -0.00002049, 0.00002194, -0.00002780, 0.00000462, -0.00002885, 0.00000797, -0.00000450, -0.00000164, 0.00000410, -0.00000744, 0.00000685, -0.00000270, -0.00000435, 0.00000114, 0.00000385, 0.00000041, },
		{ -0.00000108, -0.00000423, 0.00000717, -0.00000800, 0.00000690, -0.00000358, 0.00000204, -0.00000307, 0.00000062, -0.00000640, 0.00000635, 0.00000118, -0.00000057, -0.00000628, 0.00000484, -0.00000701, 0.00000507, -0.00000291, 0.00000542, -0.00000023, 0.00000062, -0.00000269, -0.00000003, -0.00000311, -0.00000026, -0.00000954, 0.00001144, -0.00000367, 0.00000358, -0.00000066, -0.00000026, -0.00000464, 0.00000381, -0.00001469, 0.00000938, 0.00000019, -0.00000216, -0.00001063, 0.00000851, -0.00001291, 0.00000420, -0.00001474, 0.00000199, -0.00001864, -0.00000530, -0.00002703, -0.00005505, -0.00018081, 0.00032034, 0.00031722, -0.00018174, -0.00004596, -0.00002717, -0.00001941, -0.00000375, 0.00000373, -0.00001303, 0.00000526, -0.00000554, 0.00000384, -0.00000846, 0.00000793, -0.00000921, 0.00000352, -0.00000486, 0.00000726, -0.00001261, -0.00000109, 0.00000004, 0.00000179, 0.00000212, 0.00000456, -0.00000780, 0.00000813, -0.00000916, -0.00000061, -0.00000069, 0.00000184, -0.00000155, 0.00000267, -0.00000260, 0.00000338, -0.00000249, 0.00000503, -0.00000671, 0.00000494, -0.00000605, 0.00000141, -0.00000009, 0.00000566, -0.00000815, -0.00000108, -0.00000147, 0.00000331, -0.00000193, 0.00000777, -0.00000976, 0.00000481, -0.00000329, 0.00000354, -0.00000207, 0.00000852, -0.00001103, 0.00000179, -0.00000711, 0.00000723, -0.00002813, 0.00002378, -0.00001819, 0.00005104, -0.00001896, 0.00001724, 0.00002063, -0.00000861, 0.00000798, -0.00002258, -0.00000115, -0.00002518, 0.00001975, -0.00001995, 0.00001695, -0.00001443, 0.00000851, -0.00000279, -0.00000061, -0.00000568, 0.00000719, -0.00000034, },
		{ 0.00000024, -0.00000640, 0.00000575, -0.00000324, 0.00000026, 0.00000246, -0.00000147, -0.00000062, -0.00000032, -0.00000477, 0.00000089, 0.00000459, 0.00000092, -0.00000641, 0.00000201, -0.00000252, -0.00000086, 0.00000180, 0.00000044, 0.00000388, -0.00000064, -0.00000136, -0.00000025, -0.00000322, 0.00000093, -0.00000918, 0.00000371, 0.00000481, -0.00000206, 0.00000279, 0.00000006, -0.00000479, 0.00000261, -0.00000844, -0.00000258, 0.00000770, -0.00000185, -0.00000982, 0.00000269, -0.00000281, -0.00000599, -0.00000373, -0.00000683, -0.00000758, -0.00001231, -0.00001840, -0.00002040, -0.00015241, -0.00000610, 0.00051478, -0.00002901, -0.00014169, -0.00001066, -0.00003167, -0.00001134, 0.00000572, -0.00000926, -0.00000083, -0.00000334, 0.00000302, -0.00000648, 0.00000384, -0.00000356, -0.00000204, -0.00000132, 0.00000315, -0.00000340, -0.00000931, 0.00000172, 0.00000092, 0.00000017, 0.00000713, -0.00000544, 0.00000257, -0.00000057, -0.00000737, 0.00000124, 0.00000059, -0.00000074, 0.00000178, -0.00000132, 0.00000200, -0.00000156, 0.00000368, -0.00000265, -0.00000026, -0.00000030, -0.00000350, 0.00000112, 0.00000427, -0.00000203, -0.00000588, -0.00000078, 0.00000210, -0.00000142, 0.00000611, -0.00000311, -0.00000272, 0.00000167, -0.00000028, -0.00000000, 0.00000570, -0.00000305, -0.00000402, -0.00000551, 0.00000659, -0.00001730, 0.00000041, 0.00000586, 0.00000653, 0.00003861, -0.00003569, 0.00005324, -0.00002212, 0.00002130, -0.00002434, 0.00000326, -0.00002931, 0.00001127, -0.00000858, 0.00000341, 0.00000063, -0.00000550, 0.00000651, -0.00000387, -0.00000374, 0.00000167, 0.00000392, },
		{ -0.00000036, -0.00000174, -0.00000370, 0.00000705, -0.00000845, 0.00000800, -0.00000438, 0.00000253, -0.00000358, 0.00000060, -0.00000660, 0.00000793, -0.00000008, -0.00000078, -0.00000632, 0.00000554, -0.00000803, 0.00000648, -0.00000412, 0.00000710, -0.00000222, 0.00000188, -0.00000372, 0.00000153, -0.00000447, -0.00000039, -0.00000974, 0.00001375, -0.00000657, 0.00000619, -0.00000296, 0.00000213, -0.00000704, 0.00000472, -0.00001608, 0.00001365, -0.00000460, 0.00000011, -0.00001348, 0.00001455, -0.00001996, 0.00000943, -0.00001947, 0.00000942, -0.00002837, 0.00000441, -0.00003948, -0.00003742, -0.00021916, 0.00036971, 0.00032262, -0.00022600, -0.00001767, -0.00004182, -0.00000932, -0.00001235, 0.00001069, -0.00001703, 0.00000781, -0.00000906, 0.00000711, -0.00001040, 0.00001026, -0.00001365, 0.00000737, -0.00000639, 0.00000843, -0.00001587, 0.00000257, -0.00000214, 0.00000325, 0.00000082, 0.00000569, -0.00000822, 0.00000906, -0.00001195, 0.00000194, -0.00000192, 0.00000277, -0.00000270, 0.00000366, -0.00000308, 0.00000373, -0.00000290, 0.00000567, -0.00000741, 0.00000558, -0.00000740, 0.00000271, -0.00000012, 0.00000576, -0.00000964, 0.00000030, -0.00000176, 0.00000307, -0.00000102, 0.00000725, -0.00001033, 0.00000601, -0.00000463, 0.00000427, -0.00000127, 0.00000740, -0.00001043, 0.00000052, -0.00000537, 0.00000586, -0.00002678, 0.00002648, -0.00002062, 0.00005257, -0.00002372, 0.00001833, 0.00001800, -0.00000868, 0.00000795, -0.00001871, -0.00000602, -0.00001952, 0.00001587, -0.00001747, 0.00001751, -0.00001627, 0.00001085, -0.00000489, -0.00000026, -0.00000518, 0.00000753, },
		{ 0.00000396, -0.00000002, -0.00000680, 0.00000665, -0.00000448, 0.00000152, 0.00000199, -0.00000128, -0.00000091, -0.00000025, -0.00000548, 0.00000235, 0.00000433, 0.00000027, -0.00000690, 0.00000314, -0.00000366, 0.00000028, 0.00000105, 0.00000143, 0.00000316, -0.00000074, -0.00000145, 0.00000031, -0.00000327, 0.00000033, -0.00001015, 0.00000633, 0.00000299, -0.00000120, 0.00000227, 0.00000064, -0.00000518, 0.00000232, -0.00000990, 0.00000083, 0.00000588, -0.00000219, -0.00001155, 0.00000651, -0.00000505, -0.00000609, -0.00000440, -0.00000377, -0.00001048, -0.00001106, -0.00002102, -0.00001650, -0.00015966, -0.00002053, 0.00058323, -0.00009049, -0.00012659, -0.00001190, -0.00003042, -0.00001293, 0.00000624, -0.00000845, -0.00000063, -0.00000539, 0.00000455, -0.00000741, 0.00000610, -0.00000694, -0.00000075, -0.00000099, 0.00000414, -0.00000621, -0.00000765, 0.00000180, 0.00000088, -0.00000016, 0.00000756, -0.00000584, 0.00000411, -0.00000308, -0.00000643, 0.00000153, 0.00000070, -0.00000122, 0.00000204, -0.00000136, 0.00000235, -0.00000220, 0.00000458, -0.00000355, 0.00000065, -0.00000160, -0.00000286, 0.00000137, 0.00000488, -0.00000357, -0.00000538, -0.00000030, 0.00000178, -0.00000130, 0.00000710, -0.00000495, -0.00000116, 0.00000082, -0.00000041, 0.00000092, 0.00000560, -0.00000380, -0.00000359, -0.00000642, 0.00000859, -0.00002004, 0.00000597, 0.00000257, 0.00000883, 0.00003473, -0.00003593, 0.00005141, -0.00002226, 0.00001931, -0.00001985, 0.00000073, -0.00002833, 0.00001337, -0.00001201, 0.00000842, -0.00000315, -0.00000318, 0.00000569, -0.00000470, -0.00000317, 0.00000219, },
		{ 0.00000780, -0.00000042, -0.00000240, -0.00000300, 0.00000659, -0.00000853, 0.00000891, -0.00000516, 0.00000302, -0.00000411, 0.00000054, -0.00000663, 0.00000938, -0.00000142, -0.00000092, -0.00000625, 0.00000605, -0.00000888, 0.00000778, -0.00000531, 0.00000876, -0.00000430, 0.00000331, -0.00000490, 0.00000327, -0.00000608, -0.00000042, -0.00000971, 0.00001580, -0.00000943, 0.00000896, -0.00000551, 0.00000481, -0.00000974, 0.00000576, -0.00001730, 0.00001783, -0.00000966, 0.00000267, -0.00001629, 0.00002064, -0.00002743, 0.00001531, -0.00002481, 0.00001758, -0.00003914, 0.00001549, -0.00005376, -0.00001706, -0.00026220, 0.00042772, 0.00031839, -0.00026636, 0.00001141, -0.00005792, 0.00000208, -0.00002208, 0.00001863, -0.00002179, 0.00001085, -0.00001295, 0.00001070, -0.00001242, 0.00001243, -0.00001807, 0.00001149, -0.00000810, 0.00000950, -0.00001899, 0.00000638, -0.00000463, 0.00000500, -0.00000067, 0.00000690, -0.00000854, 0.00000970, -0.00001457, 0.00000461, -0.00000335, 0.00000383, -0.00000394, 0.00000476, -0.00000363, 0.00000409, -0.00000321, 0.00000613, -0.00000793, 0.00000604, -0.00000862, 0.00000404, -0.00000018, 0.00000567, -0.00001090, 0.00000166, -0.00000215, 0.00000288, 0.00000005, 0.00000634, -0.00001051, 0.00000694, -0.00000583, 0.00000502, -0.00000051, 0.00000613, -0.00000955, -0.00000101, -0.00000320, 0.00000368, -0.00002388, 0.00002726, -0.00002183, 0.00005296, -0.00002731, 0.00001886, 0.00001588, -0.00000886, 0.00000775, -0.00001486, -0.00001095, -0.00001323, 0.00001086, -0.00001366, 0.00001694, -0.00001733, 0.00001263, -0.00000679, 0.00000013, -0.00000465, },
		{ 0.00000269, 0.00000395, -0.00000039, -0.00000695, 0.00000730, -0.00000564, 0.00000294, 0.00000129, -0.00000094, -0.00000130, -0.00000017, -0.00000616, 0.00000393, 0.00000381, -0.00000029, -0.00000734, 0.00000431, -0.00000493, 0.00000167, -0.00000001, 0.00000273, 0.00000205, -0.00000045, -0.00000185, 0.00000117, -0.00000359, -0.00000019, -0.00001110, 0.00000920, 0.00000062, 0.00000030, 0.00000115, 0.00000176, -0.00000603, 0.00000234, -0.00001164, 0.00000486, 0.00000312, -0.00000174, -0.00001389, 0.00001132, -0.00000870, -0.00000471, -0.00000648, 0.00000090, -0.00001531, -0.00000762, -0.00002617, -0.00000954, -0.00017074, -0.00003076, 0.00065162, -0.00015877, -0.00010368, -0.00001809, -0.00002597, -0.00001679, 0.00000846, -0.00000888, 0.00000048, -0.00000823, 0.00000676, -0.00000882, 0.00000881, -0.00001104, 0.00000140, -0.00000122, 0.00000547, -0.00000953, -0.00000523, 0.00000119, 0.00000131, -0.00000078, 0.00000819, -0.00000636, 0.00000579, -0.00000600, -0.00000490, 0.00000136, 0.00000108, -0.00000193, 0.00000251, -0.00000153, 0.00000276, -0.00000286, 0.00000550, -0.00000456, 0.00000170, -0.00000311, -0.00000194, 0.00000150, 0.00000546, -0.00000519, -0.00000461, -0.00000004, 0.00000146, -0.00000100, 0.00000785, -0.00000674, 0.00000054, -0.00000023, -0.00000035, 0.00000183, 0.00000526, -0.00000426, -0.00000345, -0.00000683, 0.00000983, -0.00002186, 0.00001112, -0.00000135, 0.00001179, 0.00003001, -0.00003478, 0.00004834, -0.00002096, 0.00001601, -0.00001438, -0.00000298, -0.00002580, 0.00001403, -0.00001446, 0.00001303, -0.00000698, -0.00000064, 0.00000447, -0.00000521, -0.00000264, },
		{ -0.00000408, 0.00000798, -0.00000054, -0.00000299, -0.00000221, 0.00000583, -0.00000823, 0.00000957, -0.00000587, 0.00000349, -0.00000463, 0.00000044, -0.00000648, 0.00001066, -0.00000282, -0.00000099, -0.00000606, 0.00000636, -0.00000950, 0.00000891, -0.00000638, 0.00001030, -0.00000636, 0.00000483, -0.00000616, 0.00000512, -0.00000789, -0.00000035, -0.00000944, 0.00001749, -0.00001212, 0.00001172, -0.00000817, 0.00000766, -0.00001265, 0.00000687, -0.00001826, 0.00002173, -0.00001481, 0.00000536, -0.00001886, 0.00002643, -0.00003494, 0.00002148, -0.00003043, 0.00002607, -0.00005051, 0.00002743, -0.00006931, 0.00000535, -0.00030920, 0.00049426, 0.00030341, -0.00030096, 0.00003972, -0.00007447, 0.00001411, -0.00003245, 0.00002717, -0.00002706, 0.00001422, -0.00001708, 0.00001449, -0.00001442, 0.00001430, -0.00002227, 0.00001572, -0.00000990, 0.00001036, -0.00002182, 0.00001015, -0.00000730, 0.00000694, -0.00000230, 0.00000815, -0.00000873, 0.00000999, -0.00001690, 0.00000729, -0.00000490, 0.00000496, -0.00000523, 0.00000592, -0.00000424, 0.00000445, -0.00000343, 0.00000640, -0.00000825, 0.00000627, -0.00000965, 0.00000535, -0.00000024, 0.00000538, -0.00001189, 0.00000293, -0.00000262, 0.00000277, 0.00000120, 0.00000509, -0.00001028, 0.00000754, -0.00000684, 0.00000577, 0.00000017, 0.00000477, -0.00000850, -0.00000268, -0.00000079, 0.00000090, -0.00001960, 0.00002607, -0.00002158, 0.00005190, -0.00002944, 0.00001852, 0.00001454, -0.00000939, 0.00000767, -0.00001135, -0.00001556, -0.00000670, 0.00000505, -0.00000873, 0.00001525, -0.00001755, 0.00001376, -0.00000839, 0.00000052, },
		{ -0.00000216, 0.00000319, 0.00000389, -0.00000083, -0.00000684, 0.00000766, -0.00000664, 0.00000444, 0.00000040, -0.00000047, -0.00000178, -0.00000006, -0.00000679, 0.00000560, 0.00000304, -0.00000072, -0.00000772, 0.00000547, -0.00000630, 0.00000327, -0.00000135, 0.00000432, 0.00000055, 0.00000024, -0.00000258, 0.00000236, -0.00000425, -0.00000056, -0.00001202, 0.00001224, -0.00000225, 0.00000244, -0.00000061, 0.00000348, -0.00000742, 0.00000275, -0.00001365, 0.00000946, -0.00000058, -0.00000047, -0.00001686, 0.00001708, -0.00001378, -0.00000174, -0.00001012, 0.00000734, -0.00002229, -0.00000169, -0.00003427, 0.00000108, -0.00018660, -0.00003511, 0.00071767, -0.00023211, -0.00007354, -0.00002933, -0.00001806, -0.00002318, 0.00001262, -0.00001079, 0.00000267, -0.00001197, 0.00000974, -0.00001075, 0.00001195, -0.00001582, 0.00000440, -0.00000207, 0.00000715, -0.00001331, -0.00000207, -0.00000014, 0.00000228, -0.00000177, 0.00000904, -0.00000698, 0.00000756, -0.00000927, -0.00000278, 0.00000070, 0.00000179, -0.00000291, 0.00000322, -0.00000187, 0.00000326, -0.00000352, 0.00000641, -0.00000564, 0.00000285, -0.00000479, -0.00000075, 0.00000148, 0.00000598, -0.00000684, -0.00000361, 0.00000001, 0.00000119, -0.00000052, 0.00000831, -0.00000838, 0.00000227, -0.00000142, -0.00000012, 0.00000273, 0.00000468, -0.00000442, -0.00000365, -0.00000668, 0.00001020, -0.00002251, 0.00001541, -0.00000539, 0.00001501, 0.00002474, -0.00003247, 0.00004421, -0.00001837, 0.00001156, -0.00000806, -0.00000776, -0.00002173, 0.00001311, -0.00001564, 0.00001691, -0.00001059, 0.00000192, 0.00000298, -0.00000544, },
		{ 0.00000086, -0.00000346, 0.00000809, -0.00000073, -0.00000346, -0.00000138, 0.00000483, -0.00000755, 0.00000995, -0.00000647, 0.00000390, -0.00000516, 0.00000032, -0.00000614, 0.00001173, -0.00000423, -0.00000100, -0.00000577, 0.00000646, -0.00000984, 0.00000980, -0.00000727, 0.00001165, -0.00000831, 0.00000634, -0.00000741, 0.00000700, -0.00000987, -0.00000019, -0.00000891, 0.00001875, -0.00001447, 0.00001431, -0.00001080, 0.00001055, -0.00001565, 0.00000798, -0.00001887, 0.00002519, -0.00001983, 0.00000799, -0.00002097, 0.00003162, -0.00004210, 0.00002755, -0.00003591, 0.00003440, -0.00006188, 0.00003960, -0.00008539, 0.00002887, -0.00035909, 0.00056874, 0.00027693, -0.00032814, 0.00006573, -0.00009041, 0.00002602, -0.00004286, 0.00003587, -0.00003253, 0.00001767, -0.00002124, 0.00001832, -0.00001628, 0.00001572, -0.00002605, 0.00001986, -0.00001168, 0.00001092, -0.00002421, 0.00001372, -0.00000999, 0.00000898, -0.00000400, 0.00000941, -0.00000877, 0.00000986, -0.00001882, 0.00000982, -0.00000645, 0.00000608, -0.00000651, 0.00000710, -0.00000487, 0.00000481, -0.00000356, 0.00000647, -0.00000835, 0.00000624, -0.00001044, 0.00000661, -0.00000031, 0.00000490, -0.00001257, 0.00000407, -0.00000313, 0.00000275, 0.00000237, 0.00000356, -0.00000965, 0.00000778, -0.00000762, 0.00000650, 0.00000074, 0.00000340, -0.00000739, -0.00000434, 0.00000166, -0.00000222, -0.00001422, 0.00002303, -0.00001975, 0.00004922, -0.00002989, 0.00001708, 0.00001419, -0.00001049, 0.00000797, -0.00000850, -0.00001945, -0.00000037, -0.00000116, -0.00000294, 0.00001255, -0.00001692, 0.00001419, -0.00000963, },
		{ -0.00000543, -0.00000172, 0.00000367, 0.00000377, -0.00000132, -0.00000649, 0.00000769, -0.00000740, 0.00000596, -0.00000066, 0.00000014, -0.00000236, 0.00000007, -0.00000735, 0.00000731, 0.00000203, -0.00000102, -0.00000805, 0.00000659, -0.00000770, 0.00000503, -0.00000293, 0.00000618, -0.00000131, 0.00000135, -0.00000367, 0.00000391, -0.00000529, -0.00000073, -0.00001291, 0.00001537, -0.00000554, 0.00000518, -0.00000303, 0.00000582, -0.00000939, 0.00000358, -0.00001593, 0.00001458, -0.00000518, 0.00000163, -0.00002040, 0.00002368, -0.00002020, 0.00000284, -0.00001539, 0.00001561, -0.00003151, 0.00000690, -0.00004557, 0.00001573, -0.00020800, -0.00003199, 0.00077899, -0.00030845, -0.00003712, -0.00004539, -0.00000671, -0.00003217, 0.00001882, -0.00001431, 0.00000608, -0.00001671, 0.00001354, -0.00001322, 0.00001546, -0.00002120, 0.00000824, -0.00000355, 0.00000916, -0.00001748, 0.00000177, -0.00000220, 0.00000381, -0.00000316, 0.00001015, -0.00000770, 0.00000935, -0.00001279, -0.00000014, -0.00000044, 0.00000283, -0.00000416, 0.00000418, -0.00000240, 0.00000386, -0.00000419, 0.00000729, -0.00000671, 0.00000403, -0.00000659, 0.00000068, 0.00000132, 0.00000644, -0.00000845, -0.00000243, -0.00000015, 0.00000097, 0.00000011, 0.00000845, -0.00000978, 0.00000394, -0.00000268, 0.00000028, 0.00000357, 0.00000389, -0.00000429, -0.00000416, -0.00000596, 0.00000962, -0.00002179, 0.00001841, -0.00000904, 0.00001809, 0.00001926, -0.00002930, 0.00003925, -0.00001473, 0.00000622, -0.00000112, -0.00001339, -0.00001626, 0.00001059, -0.00001538, 0.00001977, -0.00001374, 0.00000432, 0.00000134, },
		{ -0.00001046, 0.00000111, -0.00000278, 0.00000812, -0.00000101, -0.00000377, -0.00000061, 0.00000367, -0.00000652, 0.00001002, -0.00000691, 0.00000424, -0.00000567, 0.00000019, -0.00000561, 0.00001256, -0.00000561, -0.00000093, -0.00000539, 0.00000635, -0.00000990, 0.00001039, -0.00000790, 0.00001272, -0.00001004, 0.00000775, -0.00000857, 0.00000882, -0.00001193, 0.00000006, -0.00000812, 0.00001949, -0.00001637, 0.00001658, -0.00001321, 0.00001331, -0.00001863, 0.00000900, -0.00001903, 0.00002803, -0.00002452, 0.00001040, -0.00002245, 0.00003588, -0.00004847, 0.00003307, -0.00004080, 0.00004205, -0.00007264, 0.00005126, -0.00010111, 0.00005237, -0.00041044, 0.00065008, 0.00023872, -0.00034653, 0.00008801, -0.00010465, 0.00003699, -0.00005269, 0.00004423, -0.00003784, 0.00002095, -0.00002523, 0.00002205, -0.00001791, 0.00001658, -0.00002922, 0.00002372, -0.00001331, 0.00001109, -0.00002602, 0.00001689, -0.00001253, 0.00001098, -0.00000570, 0.00001064, -0.00000864, 0.00000928, -0.00002020, 0.00001209, -0.00000792, 0.00000711, -0.00000770, 0.00000826, -0.00000550, 0.00000516, -0.00000361, 0.00000635, -0.00000822, 0.00000592, -0.00001093, 0.00000776, -0.00000037, 0.00000424, -0.00001294, 0.00000503, -0.00000367, 0.00000285, 0.00000348, 0.00000183, -0.00000865, 0.00000764, -0.00000812, 0.00000719, 0.00000119, 0.00000209, -0.00000630, -0.00000583, 0.00000393, -0.00000536, -0.00000812, 0.00001839, -0.00001638, 0.00004487, -0.00002855, 0.00001438, 0.00001498, -0.00001234, 0.00000885, -0.00000656, -0.00002227, 0.00000530, -0.00000731, 0.00000334, 0.00000903, -0.00001551, 0.00001394, },
		{ -0.00000031, -0.00000526, -0.00000131, 0.00000415, 0.00000356, -0.00000182, -0.00000593, 0.00000738, -0.00000786, 0.00000739, -0.00000182, 0.00000087, -0.00000305, 0.00000022, -0.00000783, 0.00000904, 0.00000080, -0.00000116, -0.00000834, 0.00000763, -0.00000908, 0.00000688, -0.00000470, 0.00000825, -0.00000349, 0.00000283, -0.00000511, 0.00000583, -0.00000674, -0.00000066, -0.00001375, 0.00001852, -0.00000914, 0.00000843, -0.00000604, 0.00000877, -0.00001196, 0.00000486, -0.00001844, 0.00002010, -0.00001058, 0.00000451, -0.00002442, 0.00003093, -0.00002780, 0.00000892, -0.00002221, 0.00002563, -0.00004290, 0.00001815, -0.00006014, 0.00003459, -0.00023543, -0.00002000, 0.00083316, -0.00038545, 0.00000424, -0.00006570, 0.00000785, -0.00004363, 0.00002706, -0.00001947, 0.00001074, -0.00002244, 0.00001818, -0.00001624, 0.00001928, -0.00002704, 0.00001284, -0.00000564, 0.00001146, -0.00002196, 0.00000620, -0.00000494, 0.00000592, -0.00000496, 0.00001152, -0.00000852, 0.00001111, -0.00001645, 0.00000294, -0.00000205, 0.00000420, -0.00000569, 0.00000542, -0.00000312, 0.00000457, -0.00000488, 0.00000809, -0.00000775, 0.00000518, -0.00000844, 0.00000232, 0.00000103, 0.00000680, -0.00000998, -0.00000112, -0.00000050, 0.00000084, 0.00000086, 0.00000824, -0.00001084, 0.00000544, -0.00000395, 0.00000081, 0.00000433, 0.00000294, -0.00000391, -0.00000495, -0.00000472, 0.00000810, -0.00001962, 0.00001984, -0.00001183, 0.00002057, 0.00001399, -0.00002565, 0.00003378, -0.00001036, 0.00000030, 0.00000612, -0.00001957, -0.00000964, 0.00000658, -0.00001360, 0.00002143, -0.00001619, 0.00000638, },
		{ 0.00001304, -0.00001087, 0.00000125, -0.00000202, 0.00000806, -0.00000139, -0.00000387, 0.00000004, 0.00000243, -0.00000519, 0.00000977, -0.00000717, 0.00000448, -0.00000617, 0.00000004, -0.00000491, 0.00001313, -0.00000693, -0.00000081, -0.00000494, 0.00000604, -0.00000964, 0.00001064, -0.00000822, 0.00001344, -0.00001147, 0.00000895, -0.00000954, 0.00001047, -0.00001399, 0.00000038, -0.00000708, 0.00001967, -0.00001768, 0.00001836, -0.00001526, 0.00001577, -0.00002143, 0.00000985, -0.00001866, 0.00003011, -0.00002866, 0.00001243, -0.00002313, 0.00003895, -0.00005369, 0.00003762, -0.00004464, 0.00004848, -0.00008210, 0.00006162, -0.00011549, 0.00007454, -0.00046155, 0.00073675, 0.00018907, -0.00035519, 0.00010530, -0.00011619, 0.00004624, -0.00006130, 0.00005174, -0.00004258, 0.00002376, -0.00002879, 0.00002547, -0.00001918, 0.00001676, -0.00003159, 0.00002711, -0.00001467, 0.00001077, -0.00002713, 0.00001950, -0.00001477, 0.00001281, -0.00000730, 0.00001178, -0.00000833, 0.00000823, -0.00002098, 0.00001396, -0.00000918, 0.00000797, -0.00000873, 0.00000932, -0.00000609, 0.00000548, -0.00000361, 0.00000607, -0.00000786, 0.00000531, -0.00001109, 0.00000877, -0.00000043, 0.00000342, -0.00001298, 0.00000577, -0.00000419, 0.00000307, 0.00000446, -0.00000000, -0.00000735, 0.00000715, -0.00000832, 0.00000780, 0.00000150, 0.00000090, -0.00000535, -0.00000700, 0.00000578, -0.00000820, -0.00000173, 0.00001252, -0.00001167, 0.00003893, -0.00002542, 0.00001038, 0.00001697, -0.00001501, 0.00001046, -0.00000571, -0.00002375, 0.00000994, -0.00001295, 0.00000972, 0.00000493, -0.00001346, },
		{ 0.00000795, -0.00000186, -0.00000498, -0.00000091, 0.00000462, 0.00000324, -0.00000228, -0.00000521, 0.00000675, -0.00000798, 0.00000868, -0.00000303, 0.00000170, -0.00000385, 0.00000041, -0.00000820, 0.00001073, -0.00000063, -0.00000113, -0.00000858, 0.00000858, -0.00001038, 0.00000874, -0.00000658, 0.00001046, -0.00000591, 0.00000465, -0.00000686, 0.00000808, -0.00000860, -0.00000031, -0.00001453, 0.00002158, -0.00001292, 0.00001208, -0.00000957, 0.00001228, -0.00001511, 0.00000658, -0.00002113, 0.00002590, -0.00001665, 0.00000808, -0.00002879, 0.00003862, -0.00003632, 0.00001629, -0.00003041, 0.00003718, -0.00005624, 0.00003188, -0.00007782, 0.00005757, -0.00026909, 0.00000194, 0.00087791, -0.00046061, 0.00004889, -0.00008945, 0.00002515, -0.00005728, 0.00003714, -0.00002620, 0.00001658, -0.00002912, 0.00002362, -0.00001976, 0.00002330, -0.00003319, 0.00001807, -0.00000830, 0.00001398, -0.00002660, 0.00001107, -0.00000828, 0.00000856, -0.00000718, 0.00001315, -0.00000942, 0.00001275, -0.00002012, 0.00000636, -0.00000405, 0.00000586, -0.00000746, 0.00000690, -0.00000405, 0.00000540, -0.00000558, 0.00000882, -0.00000870, 0.00000625, -0.00001027, 0.00000411, 0.00000061, 0.00000707, -0.00001139, 0.00000027, -0.00000103, 0.00000082, 0.00000169, 0.00000769, -0.00001153, 0.00000669, -0.00000515, 0.00000145, 0.00000500, 0.00000189, -0.00000335, -0.00000593, -0.00000305, 0.00000575, -0.00001602, 0.00001949, -0.00001339, 0.00002205, 0.00000934, -0.00002191, 0.00002815, -0.00000560, -0.00000584, 0.00001331, -0.00002593, -0.00000221, 0.00000130, -0.00001037, 0.00002175, -0.00001779, },
		{ -0.00001093, 0.00001159, -0.00001088, 0.00000127, -0.00000120, 0.00000792, -0.00000185, -0.00000375, 0.00000049, 0.00000120, -0.00000364, 0.00000920, -0.00000722, 0.00000460, -0.00000663, -0.00000010, -0.00000405, 0.00001340, -0.00000814, -0.00000063, -0.00000444, 0.00000555, -0.00000909, 0.00001053, -0.00000819, 0.00001375, -0.00001252, 0.00000985, -0.00001022, 0.00001185, -0.00001597, 0.00000072, -0.00000579, 0.00001927, -0.00001832, 0.00001952, -0.00001677, 0.00001779, -0.00002391, 0.00001042, -0.00001769, 0.00003127, -0.00003208, 0.00001392, -0.00002288, 0.00004061, -0.00005740, 0.00004080, -0.00004702, 0.00005316, -0.00008963, 0.00006990, -0.00012755, 0.00009399, -0.00051046, 0.00082675, 0.00012882, -0.00035360, 0.00011665, -0.00012417, 0.00005306, -0.00006807, 0.00005789, -0.00004638, 0.00002579, -0.00003167, 0.00002842, -0.00002001, 0.00001618, -0.00003301, 0.00002987, -0.00001566, 0.00000992, -0.00002744, 0.00002140, -0.00001653, 0.00001435, -0.00000871, 0.00001280, -0.00000785, 0.00000672, -0.00002110, 0.00001534, -0.00001014, 0.00000858, -0.00000953, 0.00001023, -0.00000660, 0.00000576, -0.00000356, 0.00000565, -0.00000731, 0.00000441, -0.00001090, 0.00000959, -0.00000047, 0.00000246, -0.00001271, 0.00000629, -0.00000466, 0.00000342, 0.00000525, -0.00000183, -0.00000582, 0.00000632, -0.00000822, 0.00000832, 0.00000165, -0.00000013, -0.00000460, -0.00000775, 0.00000704, -0.00001041, 0.00000448, 0.00000591, -0.00000593, 0.00003161, -0.00002065, 0.00000513, 0.00002011, -0.00001850, 0.00001284, -0.00000605, -0.00002369, 0.00001320, -0.00001766, 0.00001578, 0.00000055, },
		{ -0.00001843, 0.00000892, -0.00000318, -0.00000467, -0.00000051, 0.00000507, 0.00000280, -0.00000265, -0.00000438, 0.00000584, -0.00000773, 0.00000974, -0.00000425, 0.00000261, -0.00000474, 0.00000064, -0.00000847, 0.00001235, -0.00000223, -0.00000093, -0.00000879, 0.00000941, -0.00001155, 0.00001054, -0.00000850, 0.00001274, -0.00000850, 0.00000674, -0.00000887, 0.00001061, -0.00001085, 0.00000032, -0.00001525, 0.00002447, -0.00001675, 0.00001599, -0.00001349, 0.00001624, -0.00001875, 0.00000869, -0.00002393, 0.00003181, -0.00002322, 0.00001222, -0.00003334, 0.00004646, -0.00004542, 0.00002466, -0.00003968, 0.00004993, -0.00007117, 0.00004770, -0.00009823, 0.00008429, -0.00030881, 0.00003455, 0.00091119, -0.00053145, 0.00009502, -0.00011553, 0.00004450, -0.00007264, 0.00004874, -0.00003427, 0.00002346, -0.00003661, 0.00002976, -0.00002372, 0.00002739, -0.00003945, 0.00002377, -0.00001142, 0.00001664, -0.00003126, 0.00001622, -0.00001209, 0.00001165, -0.00000976, 0.00001503, -0.00001037, 0.00001422, -0.00002367, 0.00000997, -0.00000637, 0.00000777, -0.00000945, 0.00000861, -0.00000517, 0.00000635, -0.00000631, 0.00000946, -0.00000952, 0.00000717, -0.00001201, 0.00000600, 0.00000008, 0.00000722, -0.00001262, 0.00000167, -0.00000171, 0.00000094, 0.00000254, 0.00000684, -0.00001180, 0.00000763, -0.00000621, 0.00000215, 0.00000555, 0.00000078, -0.00000268, -0.00000699, -0.00000110, 0.00000273, -0.00001114, 0.00001731, -0.00001345, 0.00002221, 0.00000565, -0.00001845, 0.00002272, -0.00000083, -0.00001181, 0.00002005, -0.00003203, 0.00000560, -0.00000490, -0.00000584, 0.00002073, },
		{ -0.00000380, -0.00000813, 0.00000972, -0.00001054, 0.00000117, -0.00000032, 0.00000768, -0.00000238, -0.00000339, 0.00000069, 0.00000009, -0.00000194, 0.00000835, -0.00000704, 0.00000458, -0.00000704, -0.00000023, -0.00000305, 0.00001337, -0.00000920, -0.00000043, -0.00000391, 0.00000492, -0.00000826, 0.00001006, -0.00000779, 0.00001363, -0.00001312, 0.00001039, -0.00001053, 0.00001285, -0.00001776, 0.00000107, -0.00000428, 0.00001828, -0.00001824, 0.00001997, -0.00001763, 0.00001920, -0.00002594, 0.00001064, -0.00001608, 0.00003144, -0.00003463, 0.00001477, -0.00002162, 0.00004071, -0.00005937, 0.00004231, -0.00004758, 0.00005565, -0.00009464, 0.00007535, -0.00013630, 0.00010932, -0.00055509, 0.00091774, 0.00005935, -0.00034175, 0.00012140, -0.00012792, 0.00005688, -0.00007251, 0.00006223, -0.00004887, 0.00002677, -0.00003365, 0.00003071, -0.00002030, 0.00001478, -0.00003338, 0.00003185, -0.00001619, 0.00000848, -0.00002689, 0.00002247, -0.00001769, 0.00001546, -0.00000984, 0.00001363, -0.00000720, 0.00000476, -0.00002052, 0.00001613, -0.00001073, 0.00000887, -0.00001003, 0.00001093, -0.00000698, 0.00000598, -0.00000346, 0.00000513, -0.00000659, 0.00000325, -0.00001036, 0.00001021, -0.00000051, 0.00000140, -0.00001215, 0.00000655, -0.00000505, 0.00000389, 0.00000578, -0.00000354, -0.00000417, 0.00000523, -0.00000783, 0.00000873, 0.00000167, -0.00000097, -0.00000409, -0.00000800, 0.00000755, -0.00001172, 0.00001004, -0.00000094, 0.00000041, 0.00002325, -0.00001449, -0.00000118, 0.00002427, -0.00002270, 0.00001592, -0.00000756, -0.00002202, 0.00001487, -0.00002110, 0.00002113, },
		{ 0.00001845, -0.00001809, 0.00000922, -0.00000421, -0.00000438, -0.00000008, 0.00000549, 0.00000224, -0.00000289, -0.00000354, 0.00000471, -0.00000714, 0.00001054, -0.00000540, 0.00000357, -0.00000572, 0.00000091, -0.00000862, 0.00001386, -0.00000395, -0.00000056, -0.00000898, 0.00001011, -0.00001255, 0.00001220, -0.00001036, 0.00001498, -0.00001116, 0.00000899, -0.00001104, 0.00001336, -0.00001344, 0.00000125, -0.00001590, 0.00002710, -0.00002046, 0.00001998, -0.00001765, 0.00002053, -0.00002281, 0.00001114, -0.00002672, 0.00003767, -0.00003009, 0.00001677, -0.00003790, 0.00005414, -0.00005471, 0.00003362, -0.00004963, 0.00006341, -0.00008715, 0.00006507, -0.00012075, 0.00011405, -0.00035399, 0.00007805, 0.00093135, -0.00059556, 0.00014067, -0.00014271, 0.00006504, -0.00008907, 0.00006141, -0.00004337, 0.00003114, -0.00004471, 0.00003645, -0.00002799, 0.00003141, -0.00004562, 0.00002974, -0.00001489, 0.00001931, -0.00003577, 0.00002146, -0.00001621, 0.00001510, -0.00001265, 0.00001712, -0.00001136, 0.00001543, -0.00002697, 0.00001364, -0.00000890, 0.00000984, -0.00001157, 0.00001049, -0.00000646, 0.00000741, -0.00000706, 0.00001000, -0.00001020, 0.00000791, -0.00001360, 0.00000794, -0.00000055, 0.00000726, -0.00001365, 0.00000305, -0.00000251, 0.00000121, 0.00000335, 0.00000573, -0.00001165, 0.00000820, -0.00000708, 0.00000290, 0.00000596, -0.00000032, -0.00000199, -0.00000800, 0.00000095, -0.00000070, -0.00000521, 0.00001341, -0.00001189, 0.00002081, 0.00000324, -0.00001560, 0.00001783, 0.00000361, -0.00001725, 0.00002598, -0.00003743, 0.00001331, -0.00001160, -0.00000029, },
		{ 0.00002540, -0.00000780, -0.00000527, 0.00000755, -0.00000990, 0.00000099, 0.00000061, 0.00000733, -0.00000294, -0.00000281, 0.00000061, -0.00000083, -0.00000019, 0.00000726, -0.00000664, 0.00000440, -0.00000739, -0.00000035, -0.00000192, 0.00001304, -0.00001008, -0.00000020, -0.00000337, 0.00000418, -0.00000719, 0.00000924, -0.00000701, 0.00001305, -0.00001323, 0.00001049, -0.00001040, 0.00001340, -0.00001927, 0.00000138, -0.00000259, 0.00001674, -0.00001741, 0.00001962, -0.00001772, 0.00001988, -0.00002739, 0.00001045, -0.00001380, 0.00003053, -0.00003617, 0.00001490, -0.00001935, 0.00003919, -0.00005942, 0.00004192, -0.00004606, 0.00005560, -0.00009667, 0.00007737, -0.00014091, 0.00011922, -0.00059331, 0.00100709, -0.00001748, -0.00032012, 0.00011930, -0.00012704, 0.00005729, -0.00007421, 0.00006441, -0.00004976, 0.00002646, -0.00003450, 0.00003218, -0.00001998, 0.00001255, -0.00003263, 0.00003293, -0.00001619, 0.00000647, -0.00002546, 0.00002265, -0.00001814, 0.00001606, -0.00001059, 0.00001425, -0.00000641, 0.00000242, -0.00001925, 0.00001631, -0.00001088, 0.00000878, -0.00001018, 0.00001136, -0.00000719, 0.00000611, -0.00000333, 0.00000452, -0.00000575, 0.00000187, -0.00000946, 0.00001060, -0.00000053, 0.00000027, -0.00001132, 0.00000655, -0.00000534, 0.00000445, 0.00000603, -0.00000504, -0.00000250, 0.00000393, -0.00000717, 0.00000902, 0.00000155, -0.00000162, -0.00000383, -0.00000770, 0.00000722, -0.00001192, 0.00001454, -0.00000748, 0.00000689, 0.00001427, -0.00000730, -0.00000829, 0.00002920, -0.00002742, 0.00001957, -0.00001010, -0.00001878, 0.00001485, -0.00002298, },
		{ 0.00000591, 0.00001508, -0.00001681, 0.00000883, -0.00000489, -0.00000414, 0.00000040, 0.00000588, 0.00000153, -0.00000294, -0.00000276, 0.00000344, -0.00000622, 0.00001104, -0.00000645, 0.00000453, -0.00000677, 0.00000124, -0.00000865, 0.00001522, -0.00000576, -0.00000001, -0.00000917, 0.00001068, -0.00001334, 0.00001367, -0.00001209, 0.00001711, -0.00001377, 0.00001130, -0.00001330, 0.00001621, -0.00001630, 0.00000245, -0.00001646, 0.00002939, -0.00002391, 0.00002389, -0.00002187, 0.00002496, -0.00002713, 0.00001383, -0.00002942, 0.00004327, -0.00003703, 0.00002154, -0.00004226, 0.00006134, -0.00006380, 0.00004273, -0.00005978, 0.00007707, -0.00010356, 0.00008327, -0.00014456, 0.00014589, -0.00040365, 0.00013214, 0.00093718, -0.00065079, 0.00018393, -0.00016963, 0.00008582, -0.00010585, 0.00007457, -0.00005309, 0.00003929, -0.00005315, 0.00004350, -0.00003246, 0.00003521, -0.00005145, 0.00003577, -0.00001855, 0.00002188, -0.00003996, 0.00002658, -0.00002045, 0.00001875, -0.00001574, 0.00001937, -0.00001235, 0.00001634, -0.00002988, 0.00001720, -0.00001151, 0.00001198, -0.00001376, 0.00001249, -0.00000787, 0.00000857, -0.00000785, 0.00001045, -0.00001072, 0.00000842, -0.00001497, 0.00000985, -0.00000125, 0.00000718, -0.00001444, 0.00000435, -0.00000340, 0.00000163, 0.00000405, 0.00000445, -0.00001112, 0.00000838, -0.00000770, 0.00000365, 0.00000622, -0.00000134, -0.00000137, -0.00000884, 0.00000290, -0.00000428, 0.00000141, 0.00000800, -0.00000871, 0.00001775, 0.00000229, -0.00001364, 0.00001374, 0.00000742, -0.00002182, 0.00003074, -0.00004173, 0.00002044, -0.00001832, },
		{ -0.00002318, 0.00002833, -0.00001119, -0.00000258, 0.00000527, -0.00000906, 0.00000073, 0.00000159, 0.00000687, -0.00000351, -0.00000203, 0.00000024, -0.00000148, 0.00000152, 0.00000597, -0.00000603, 0.00000404, -0.00000766, -0.00000044, -0.00000069, 0.00001239, -0.00001074, 0.00000001, -0.00000283, 0.00000337, -0.00000593, 0.00000810, -0.00000587, 0.00001201, -0.00001284, 0.00001013, -0.00000980, 0.00001341, -0.00002042, 0.00000162, -0.00000074, 0.00001471, -0.00001586, 0.00001847, -0.00001699, 0.00001974, -0.00002817, 0.00000979, -0.00001086, 0.00002855, -0.00003665, 0.00001427, -0.00001611, 0.00003607, -0.00005750, 0.00003953, -0.00004233, 0.00005283, -0.00009539, 0.00007550, -0.00014070, 0.00012254, -0.00062308, 0.00109204, -0.00009939, -0.00028969, 0.00011046, -0.00012139, 0.00005410, -0.00007294, 0.00006418, -0.00004884, 0.00002469, -0.00003408, 0.00003271, -0.00001903, 0.00000952, -0.00003075, 0.00003306, -0.00001563, 0.00000389, -0.00002317, 0.00002190, -0.00001781, 0.00001604, -0.00001091, 0.00001460, -0.00000550, -0.00000024, -0.00001735, 0.00001584, -0.00001055, 0.00000829, -0.00000994, 0.00001148, -0.00000719, 0.00000612, -0.00000317, 0.00000388, -0.00000483, 0.00000032, -0.00000825, 0.00001075, -0.00000055, -0.00000090, -0.00001027, 0.00000630, -0.00000549, 0.00000508, 0.00000597, -0.00000626, -0.00000091, 0.00000251, -0.00000628, 0.00000917, 0.00000133, -0.00000209, -0.00000380, -0.00000689, 0.00000605, -0.00001089, 0.00001764, -0.00001322, 0.00001299, 0.00000515, 0.00000050, -0.00001582, 0.00003459, -0.00003239, 0.00002353, -0.00001342, -0.00001417, 0.00001316, },
		{ -0.00002458, 0.00001232, 0.00001089, -0.00001472, 0.00000780, -0.00000519, -0.00000400, 0.00000092, 0.00000622, 0.00000069, -0.00000277, -0.00000212, 0.00000212, -0.00000503, 0.00001122, -0.00000734, 0.00000547, -0.00000788, 0.00000161, -0.00000857, 0.00001639, -0.00000760, 0.00000068, -0.00000936, 0.00001111, -0.00001391, 0.00001487, -0.00001361, 0.00001901, -0.00001623, 0.00001357, -0.00001551, 0.00001906, -0.00001934, 0.00000389, -0.00001692, 0.00003125, -0.00002695, 0.00002752, -0.00002594, 0.00002936, -0.00003155, 0.00001663, -0.00003188, 0.00004842, -0.00004380, 0.00002634, -0.00004623, 0.00006778, -0.00007225, 0.00005153, -0.00006961, 0.00009030, -0.00011966, 0.00010148, -0.00016867, 0.00017858, -0.00045641, 0.00019599, 0.00092805, -0.00069534, 0.00022296, -0.00019492, 0.00010581, -0.00012216, 0.00008761, -0.00006293, 0.00004753, -0.00006161, 0.00005066, -0.00003695, 0.00003862, -0.00005674, 0.00004161, -0.00002226, 0.00002421, -0.00004366, 0.00003138, -0.00002461, 0.00002245, -0.00001893, 0.00002171, -0.00001332, 0.00001690, -0.00003229, 0.00002051, -0.00001408, 0.00001409, -0.00001592, 0.00001452, -0.00000936, 0.00000980, -0.00000867, 0.00001081, -0.00001107, 0.00000868, -0.00001606, 0.00001169, -0.00000200, 0.00000698, -0.00001499, 0.00000552, -0.00000433, 0.00000223, 0.00000459, 0.00000308, -0.00001025, 0.00000817, -0.00000805, 0.00000436, 0.00000632, -0.00000226, -0.00000088, -0.00000938, 0.00000454, -0.00000769, 0.00000832, 0.00000145, -0.00000411, 0.00001307, 0.00000286, -0.00001272, 0.00001065, 0.00001037, -0.00002529, 0.00003407, -0.00004456, 0.00002651, },
		{ 0.00000998, -0.00002169, 0.00002974, -0.00001374, -0.00000024, 0.00000300, -0.00000809, 0.00000045, 0.00000258, 0.00000628, -0.00000404, -0.00000112, -0.00000042, -0.00000184, 0.00000312, 0.00000455, -0.00000521, 0.00000350, -0.00000782, -0.00000052, 0.00000061, 0.00001144, -0.00001115, 0.00000020, -0.00000230, 0.00000252, -0.00000452, 0.00000670, -0.00000441, 0.00001056, -0.00001195, 0.00000929, -0.00000869, 0.00001283, -0.00002112, 0.00000176, 0.00000120, 0.00001225, -0.00001365, 0.00001653, -0.00001542, 0.00001874, -0.00002821, 0.00000866, -0.00000732, 0.00002552, -0.00003603, 0.00001289, -0.00001199, 0.00003148, -0.00005366, 0.00003514, -0.00003640, 0.00004727, -0.00009067, 0.00006950, -0.00013525, 0.00011840, -0.00064258, 0.00116978, -0.00018381, -0.00025184, 0.00009539, -0.00011113, 0.00004731, -0.00006865, 0.00006143, -0.00004600, 0.00002138, -0.00003229, 0.00003222, -0.00001744, 0.00000576, -0.00002777, 0.00003220, -0.00001451, 0.00000082, -0.00002010, 0.00002024, -0.00001668, 0.00001538, -0.00001074, 0.00001467, -0.00000452, -0.00000311, -0.00001488, 0.00001476, -0.00000974, 0.00000737, -0.00000928, 0.00001126, -0.00000695, 0.00000598, -0.00000298, 0.00000322, -0.00000389, -0.00000133, -0.00000675, 0.00001067, -0.00000056, -0.00000206, -0.00000903, 0.00000582, -0.00000547, 0.00000573, 0.00000561, -0.00000712, 0.00000051, 0.00000105, -0.00000519, 0.00000918, 0.00000105, -0.00000243, -0.00000395, -0.00000563, 0.00000408, -0.00000861, 0.00001912, -0.00001773, 0.00001823, -0.00000362, 0.00000842, -0.00002338, 0.00004008, -0.00003725, 0.00002749, -0.00001720, -0.00000847, },
		{ 0.00003115, -0.00002989, 0.00001851, 0.00000619, -0.00001197, 0.00000620, -0.00000514, -0.00000394, 0.00000149, 0.00000651, -0.00000027, -0.00000235, -0.00000168, 0.00000086, -0.00000364, 0.00001110, -0.00000804, 0.00000634, -0.00000901, 0.00000204, -0.00000836, 0.00001735, -0.00000943, 0.00000151, -0.00000957, 0.00001143, -0.00001423, 0.00001577, -0.00001485, 0.00002062, -0.00001843, 0.00001567, -0.00001758, 0.00002177, -0.00002244, 0.00000552, -0.00001726, 0.00003265, -0.00002946, 0.00003071, -0.00002967, 0.00003351, -0.00003591, 0.00001943, -0.00003400, 0.00005294, -0.00005015, 0.00003095, -0.00004961, 0.00007315, -0.00007966, 0.00005955, -0.00007860, 0.00010246, -0.00013469, 0.00011878, -0.00019199, 0.00021071, -0.00051055, 0.00026822, 0.00090391, -0.00072783, 0.00025616, -0.00021732, 0.00012401, -0.00013719, 0.00009986, -0.00007239, 0.00005544, -0.00006972, 0.00005765, -0.00004129, 0.00004151, -0.00006127, 0.00004705, -0.00002583, 0.00002619, -0.00004673, 0.00003564, -0.00002850, 0.00002603, -0.00002209, 0.00002406, -0.00001424, 0.00001708, -0.00003411, 0.00002343, -0.00001647, 0.00001606, -0.00001795, 0.00001652, -0.00001087, 0.00001106, -0.00000950, 0.00001110, -0.00001126, 0.00000869, -0.00001683, 0.00001340, -0.00000279, 0.00000668, -0.00001528, 0.00000654, -0.00000526, 0.00000298, 0.00000491, 0.00000172, -0.00000912, 0.00000761, -0.00000811, 0.00000502, 0.00000626, -0.00000302, -0.00000060, -0.00000952, 0.00000570, -0.00001062, 0.00001507, -0.00000580, 0.00000163, 0.00000696, 0.00000491, -0.00001289, 0.00000866, 0.00001232, -0.00002748, 0.00003579, -0.00004566, },
		{ -0.00000208, 0.00000559, -0.00001862, 0.00002956, -0.00001529, 0.00000159, 0.00000089, -0.00000709, 0.00000018, 0.00000358, 0.00000556, -0.00000447, -0.00000013, -0.00000131, -0.00000186, 0.00000454, 0.00000307, -0.00000422, 0.00000277, -0.00000787, -0.00000057, 0.00000194, 0.00001021, -0.00001128, 0.00000032, -0.00000178, 0.00000168, -0.00000303, 0.00000510, -0.00000270, 0.00000874, -0.00001061, 0.00000802, -0.00000711, 0.00001166, -0.00002133, 0.00000177, 0.00000320, 0.00000948, -0.00001087, 0.00001388, -0.00001305, 0.00001686, -0.00002748, 0.00000705, -0.00000324, 0.00002151, -0.00003435, 0.00001083, -0.00000715, 0.00002561, -0.00004809, 0.00002891, -0.00002842, 0.00003907, -0.00008258, 0.00005937, -0.00012441, 0.00010628, -0.00065029, 0.00123764, -0.00026801, -0.00020832, 0.00007494, -0.00009670, 0.00003720, -0.00006145, 0.00005622, -0.00004128, 0.00001653, -0.00002910, 0.00003067, -0.00001524, 0.00000138, -0.00002378, 0.00003040, -0.00001287, -0.00000264, -0.00001634, 0.00001775, -0.00001477, 0.00001404, -0.00001004, 0.00001444, -0.00000349, -0.00000610, -0.00001195, 0.00001311, -0.00000848, 0.00000604, -0.00000820, 0.00001068, -0.00000644, 0.00000568, -0.00000274, 0.00000256, -0.00000298, -0.00000303, -0.00000501, 0.00001035, -0.00000057, -0.00000318, -0.00000765, 0.00000511, -0.00000527, 0.00000636, 0.00000497, -0.00000759, 0.00000167, -0.00000037, -0.00000396, 0.00000905, 0.00000072, -0.00000268, -0.00000420, -0.00000404, 0.00000145, -0.00000516, 0.00001886, -0.00002068, 0.00002221, -0.00001156, 0.00001601, -0.00003055, 0.00004526, -0.00004166, 0.00003111, -0.00002103, },
		{ -0.00004488, 0.00003404, -0.00003384, 0.00002401, 0.00000134, -0.00000881, 0.00000416, -0.00000478, -0.00000398, 0.00000212, 0.00000674, -0.00000133, -0.00000168, -0.00000152, -0.00000026, -0.00000214, 0.00001069, -0.00000854, 0.00000710, -0.00001015, 0.00000252, -0.00000805, 0.00001805, -0.00001119, 0.00000245, -0.00000980, 0.00001164, -0.00001432, 0.00001635, -0.00001575, 0.00002186, -0.00002028, 0.00001751, -0.00001938, 0.00002423, -0.00002549, 0.00000728, -0.00001748, 0.00003353, -0.00003135, 0.00003330, -0.00003289, 0.00003723, -0.00004000, 0.00002208, -0.00003565, 0.00005663, -0.00005585, 0.00003520, -0.00005226, 0.00007724, -0.00008570, 0.00006636, -0.00008627, 0.00011295, -0.00014792, 0.00013428, -0.00021336, 0.00024077, -0.00056409, 0.00034698, 0.00086535, -0.00074743, 0.00028225, -0.00023570, 0.00013950, -0.00015019, 0.00011068, -0.00008093, 0.00006259, -0.00007713, 0.00006418, -0.00004530, 0.00004372, -0.00006484, 0.00005186, -0.00002912, 0.00002771, -0.00004901, 0.00003920, -0.00003192, 0.00002931, -0.00002507, 0.00002634, -0.00001508, 0.00001687, -0.00003526, 0.00002583, -0.00001857, 0.00001778, -0.00001975, 0.00001838, -0.00001232, 0.00001231, -0.00001035, 0.00001134, -0.00001131, 0.00000845, -0.00001726, 0.00001492, -0.00000359, 0.00000629, -0.00001531, 0.00000736, -0.00000616, 0.00000387, 0.00000496, 0.00000046, -0.00000782, 0.00000673, -0.00000788, 0.00000560, 0.00000606, -0.00000360, -0.00000056, -0.00000917, 0.00000621, -0.00001281, 0.00002123, -0.00001325, 0.00000809, -0.00000029, 0.00000823, -0.00001407, 0.00000775, 0.00001327, -0.00002837, 0.00003585, },
		{ -0.00002450, 0.00000454, 0.00000039, -0.00001421, 0.00002782, -0.00001576, 0.00000279, -0.00000094, -0.00000612, -0.00000006, 0.00000455, 0.00000470, -0.00000476, 0.00000086, -0.00000238, -0.00000156, 0.00000571, 0.00000157, -0.00000310, 0.00000187, -0.00000779, -0.00000060, 0.00000329, 0.00000873, -0.00001113, 0.00000036, -0.00000128, 0.00000087, -0.00000153, 0.00000337, -0.00000080, 0.00000663, -0.00000887, 0.00000634, -0.00000507, 0.00000990, -0.00002101, 0.00000162, 0.00000518, 0.00000651, -0.00000765, 0.00001062, -0.00000997, 0.00001415, -0.00002600, 0.00000502, 0.00000125, 0.00001666, -0.00003167, 0.00000815, -0.00000179, 0.00001873, -0.00004103, 0.00002110, -0.00001867, 0.00002853, -0.00007141, 0.00004537, -0.00010836, 0.00008602, -0.00064511, 0.00129321, -0.00034918, -0.00016110, 0.00005022, -0.00007880, 0.00002421, -0.00005167, 0.00004873, -0.00003479, 0.00001024, -0.00002456, 0.00002807, -0.00001250, -0.00000348, -0.00001892, 0.00002772, -0.00001078, -0.00000638, -0.00001205, 0.00001452, -0.00001214, 0.00001207, -0.00000882, 0.00001390, -0.00000246, -0.00000908, -0.00000868, 0.00001098, -0.00000681, 0.00000435, -0.00000673, 0.00000975, -0.00000566, 0.00000519, -0.00000245, 0.00000193, -0.00000212, -0.00000471, -0.00000309, 0.00000981, -0.00000058, -0.00000422, -0.00000619, 0.00000421, -0.00000486, 0.00000691, 0.00000411, -0.00000767, 0.00000253, -0.00000167, -0.00000264, 0.00000877, 0.00000039, -0.00000290, -0.00000446, -0.00000226, -0.00000165, -0.00000073, 0.00001691, -0.00002189, 0.00002459, -0.00001828, 0.00002281, -0.00003692, 0.00004976, -0.00004524, 0.00003400, },
		{ 0.00003432, -0.00004219, 0.00003502, -0.00003612, 0.00002842, -0.00000331, -0.00000547, 0.00000183, -0.00000418, -0.00000407, 0.00000277, 0.00000688, -0.00000245, -0.00000077, -0.00000165, -0.00000115, -0.00000061, 0.00001002, -0.00000880, 0.00000773, -0.00001125, 0.00000304, -0.00000763, 0.00001849, -0.00001282, 0.00000346, -0.00001006, 0.00001177, -0.00001419, 0.00001660, -0.00001629, 0.00002268, -0.00002171, 0.00001899, -0.00002082, 0.00002629, -0.00002834, 0.00000910, -0.00001757, 0.00003388, -0.00003253, 0.00003517, -0.00003541, 0.00004031, -0.00004365, 0.00002445, -0.00003674, 0.00005935, -0.00006069, 0.00003889, -0.00005405, 0.00007986, -0.00009006, 0.00007160, -0.00009218, 0.00012124, -0.00015864, 0.00014711, -0.00023169, 0.00026719, -0.00061485, 0.00042998, 0.00081354, -0.00075384, 0.00030031, -0.00024914, 0.00015151, -0.00016045, 0.00011947, -0.00008808, 0.00006857, -0.00008347, 0.00006997, -0.00004880, 0.00004515, -0.00006731, 0.00005584, -0.00003198, 0.00002867, -0.00005042, 0.00004190, -0.00003469, 0.00003212, -0.00002774, 0.00002847, -0.00001582, 0.00001628, -0.00003572, 0.00002761, -0.00002026, 0.00001916, -0.00002124, 0.00002003, -0.00001365, 0.00001350, -0.00001119, 0.00001152, -0.00001124, 0.00000797, -0.00001733, 0.00001622, -0.00000438, 0.00000583, -0.00001510, 0.00000797, -0.00000699, 0.00000488, 0.00000471, -0.00000060, -0.00000645, 0.00000561, -0.00000739, 0.00000609, 0.00000572, -0.00000400, -0.00000077, -0.00000832, 0.00000598, -0.00001401, 0.00002637, -0.00002036, 0.00001478, -0.00000826, 0.00001254, -0.00001611, 0.00000780, 0.00001328, -0.00002800, },
		{ 0.00003585, -0.00002716, 0.00001090, -0.00000517, -0.00000883, 0.00002470, -0.00001517, 0.00000331, -0.00000243, -0.00000524, -0.00000022, 0.00000548, 0.00000371, -0.00000485, 0.00000176, -0.00000355, -0.00000097, 0.00000660, 0.00000013, -0.00000187, 0.00000079, -0.00000756, -0.00000061, 0.00000462, 0.00000702, -0.00001069, 0.00000029, -0.00000078, 0.00000011, -0.00000006, 0.00000159, 0.00000120, 0.00000432, -0.00000683, 0.00000434, -0.00000267, 0.00000760, -0.00002014, 0.00000131, 0.00000710, 0.00000346, -0.00000414, 0.00000690, -0.00000629, 0.00001071, -0.00002382, 0.00000263, 0.00000602, 0.00001113, -0.00002812, 0.00000500, 0.00000387, 0.00001116, -0.00003285, 0.00001207, -0.00000757, 0.00001608, -0.00005762, 0.00002798, -0.00008760, 0.00005792, -0.00062644, 0.00133445, -0.00042466, -0.00011231, 0.00002260, -0.00005831, 0.00000899, -0.00003976, 0.00003931, -0.00002680, 0.00000273, -0.00001882, 0.00002451, -0.00000931, -0.00000864, -0.00001338, 0.00002427, -0.00000832, -0.00001025, -0.00000741, 0.00001072, -0.00000890, 0.00000952, -0.00000710, 0.00001305, -0.00000148, -0.00001194, -0.00000522, 0.00000847, -0.00000481, 0.00000235, -0.00000491, 0.00000849, -0.00000462, 0.00000450, -0.00000210, 0.00000133, -0.00000137, -0.00000629, -0.00000105, 0.00000907, -0.00000059, -0.00000516, -0.00000470, 0.00000315, -0.00000425, 0.00000731, 0.00000308, -0.00000736, 0.00000303, -0.00000279, -0.00000128, 0.00000835, 0.00000010, -0.00000314, -0.00000461, -0.00000048, -0.00000498, 0.00000441, 0.00001342, -0.00002131, 0.00002519, -0.00002347, 0.00002846, -0.00004214, 0.00005325, -0.00004768, },
		{ -0.00002656, 0.00003138, -0.00003770, 0.00003406, -0.00003653, 0.00003141, -0.00000742, -0.00000220, -0.00000061, -0.00000343, -0.00000419, 0.00000344, 0.00000693, -0.00000360, 0.00000035, -0.00000210, -0.00000174, 0.00000086, 0.00000916, -0.00000884, 0.00000820, -0.00001229, 0.00000360, -0.00000712, 0.00001863, -0.00001427, 0.00000450, -0.00001033, 0.00001182, -0.00001386, 0.00001651, -0.00001645, 0.00002304, -0.00002264, 0.00002002, -0.00002179, 0.00002786, -0.00003088, 0.00001089, -0.00001751, 0.00003371, -0.00003296, 0.00003624, -0.00003712, 0.00004260, -0.00004669, 0.00002642, -0.00003718, 0.00006099, -0.00006450, 0.00004189, -0.00005489, 0.00008091, -0.00009256, 0.00007499, -0.00009599, 0.00012688, -0.00016629, 0.00015652, -0.00024593, 0.00028850, -0.00066062, 0.00051461, 0.00075020, -0.00074732, 0.00030982, -0.00025704, 0.00015945, -0.00016746, 0.00012577, -0.00009342, 0.00007304, -0.00008841, 0.00007475, -0.00005163, 0.00004571, -0.00006854, 0.00005884, -0.00003429, 0.00002904, -0.00005089, 0.00004362, -0.00003667, 0.00003432, -0.00002997, 0.00003034, -0.00001645, 0.00001534, -0.00003547, 0.00002872, -0.00002147, 0.00002012, -0.00002233, 0.00002138, -0.00001479, 0.00001456, -0.00001199, 0.00001165, -0.00001109, 0.00000730, -0.00001705, 0.00001726, -0.00000514, 0.00000531, -0.00001467, 0.00000836, -0.00000770, 0.00000596, 0.00000415, -0.00000140, -0.00000510, 0.00000433, -0.00000667, 0.00000646, 0.00000526, -0.00000423, -0.00000122, -0.00000696, 0.00000497, -0.00001410, 0.00003016, -0.00002664, 0.00002120, -0.00001648, 0.00001746, -0.00001872, 0.00000860, 0.00001254, },
		{ -0.00004870, 0.00003635, -0.00002865, 0.00001654, -0.00001058, -0.00000290, 0.00002045, -0.00001360, 0.00000316, -0.00000352, -0.00000450, -0.00000029, 0.00000635, 0.00000260, -0.00000472, 0.00000249, -0.00000473, -0.00000014, 0.00000720, -0.00000121, -0.00000059, -0.00000043, -0.00000719, -0.00000062, 0.00000589, 0.00000514, -0.00000996, 0.00000009, -0.00000028, -0.00000056, 0.00000131, -0.00000016, 0.00000321, 0.00000191, -0.00000457, 0.00000211, 0.00000001, 0.00000482, -0.00001875, 0.00000084, 0.00000890, 0.00000046, -0.00000050, 0.00000289, -0.00000218, 0.00000665, -0.00002103, -0.00000002, 0.00001088, 0.00000513, -0.00002384, 0.00000151, 0.00000957, 0.00000326, -0.00002395, 0.00000227, 0.00000437, 0.00000233, -0.00004187, 0.00000795, -0.00006294, 0.00002270, -0.00059424, 0.00135983, -0.00049198, -0.00006409, -0.00000646, -0.00003628, -0.00000768, -0.00002632, 0.00002841, -0.00001766, -0.00000573, -0.00001207, 0.00002012, -0.00000582, -0.00001389, -0.00000737, 0.00002020, -0.00000563, -0.00001409, -0.00000260, 0.00000654, -0.00000520, 0.00000648, -0.00000493, 0.00001191, -0.00000058, -0.00001457, -0.00000172, 0.00000572, -0.00000258, 0.00000013, -0.00000280, 0.00000693, -0.00000332, 0.00000362, -0.00000167, 0.00000077, -0.00000074, -0.00000773, 0.00000104, 0.00000815, -0.00000060, -0.00000598, -0.00000321, 0.00000197, -0.00000343, 0.00000753, 0.00000198, -0.00000673, 0.00000317, -0.00000369, 0.00000007, 0.00000780, -0.00000012, -0.00000347, -0.00000456, 0.00000113, -0.00000826, 0.00000991, 0.00000869, -0.00001905, 0.00002397, -0.00002695, 0.00003269, -0.00004593, 0.00005545, },
		{ 0.00001130, -0.00002431, 0.00002732, -0.00003167, 0.00003125, -0.00003500, 0.00003275, -0.00001072, 0.00000075, -0.00000298, -0.00000263, -0.00000429, 0.00000409, 0.00000689, -0.00000472, 0.00000161, -0.00000286, -0.00000199, 0.00000218, 0.00000816, -0.00000866, 0.00000847, -0.00001323, 0.00000417, -0.00000653, 0.00001848, -0.00001548, 0.00000551, -0.00001060, 0.00001181, -0.00001336, 0.00001612, -0.00001623, 0.00002293, -0.00002306, 0.00002057, -0.00002225, 0.00002882, -0.00003298, 0.00001258, -0.00001730, 0.00003303, -0.00003266, 0.00003644, -0.00003792, 0.00004396, -0.00004897, 0.00002788, -0.00003692, 0.00006146, -0.00006713, 0.00004407, -0.00005475, 0.00008037, -0.00009310, 0.00007637, -0.00009750, 0.00012958, -0.00017044, 0.00016193, -0.00025527, 0.00030337, -0.00069922, 0.00059808, 0.00067754, -0.00072869, 0.00031072, -0.00025908, 0.00016296, -0.00017085, 0.00012920, -0.00009662, 0.00007570, -0.00009168, 0.00007828, -0.00005364, 0.00004536, -0.00006849, 0.00006074, -0.00003595, 0.00002877, -0.00005040, 0.00004431, -0.00003777, 0.00003579, -0.00003164, 0.00003189, -0.00001695, 0.00001409, -0.00003454, 0.00002913, -0.00002214, 0.00002060, -0.00002295, 0.00002237, -0.00001567, 0.00001545, -0.00001273, 0.00001175, -0.00001086, 0.00000646, -0.00001642, 0.00001802, -0.00000586, 0.00000475, -0.00001403, 0.00000853, -0.00000827, 0.00000707, 0.00000330, -0.00000187, -0.00000388, 0.00000297, -0.00000576, 0.00000673, 0.00000471, -0.00000431, -0.00000186, -0.00000515, 0.00000319, -0.00001299, 0.00003233, -0.00003163, 0.00002684, -0.00002445, 0.00002254, -0.00002157, 0.00000987, },
		{ 0.00005621, -0.00004814, 0.00003533, -0.00002864, 0.00002101, -0.00001535, 0.00000310, 0.00001539, -0.00001123, 0.00000239, -0.00000422, -0.00000391, -0.00000025, 0.00000713, 0.00000138, -0.00000434, 0.00000298, -0.00000582, 0.00000087, 0.00000750, -0.00000241, 0.00000071, -0.00000177, -0.00000666, -0.00000061, 0.00000708, 0.00000312, -0.00000897, -0.00000025, 0.00000023, -0.00000116, 0.00000255, -0.00000182, 0.00000515, -0.00000050, -0.00000222, -0.00000024, 0.00000285, 0.00000167, -0.00001687, 0.00000020, 0.00001052, -0.00000238, 0.00000309, -0.00000120, 0.00000216, 0.00000215, -0.00001774, -0.00000281, 0.00001567, -0.00000112, -0.00001903, -0.00000215, 0.00001507, -0.00000461, -0.00001476, -0.00000781, 0.00001656, -0.00001205, -0.00002493, -0.00001382, -0.00003546, -0.00001847, -0.00054905, 0.00136840, -0.00054905, -0.00001847, -0.00003546, -0.00001382, -0.00002493, -0.00001205, 0.00001656, -0.00000781, -0.00001476, -0.00000461, 0.00001507, -0.00000215, -0.00001903, -0.00000112, 0.00001567, -0.00000281, -0.00001774, 0.00000215, 0.00000216, -0.00000120, 0.00000309, -0.00000238, 0.00001052, 0.00000020, -0.00001687, 0.00000167, 0.00000285, -0.00000024, -0.00000222, -0.00000050, 0.00000515, -0.00000182, 0.00000255, -0.00000116, 0.00000023, -0.00000025, -0.00000897, 0.00000312, 0.00000708, -0.00000061, -0.00000666, -0.00000177, 0.00000071, -0.00000241, 0.00000750, 0.00000087, -0.00000582, 0.00000298, -0.00000434, 0.00000138, 0.00000713, -0.00000025, -0.00000391, -0.00000422, 0.00000239, -0.00001123, 0.00001539, 0.00000310, -0.00001535, 0.00002101, -0.00002864, 0.00003533, -0.00004814, },
		{ 0.00001130, 0.00000987, -0.00002157, 0.00002254, -0.00002445, 0.00002684, -0.00003163, 0.00003233, -0.00001299, 0.00000319, -0.00000515, -0.00000186, -0.00000431, 0.00000471, 0.00000673, -0.00000576, 0.00000297, -0.00000388, -0.00000187, 0.00000330, 0.00000707, -0.00000827, 0.00000853, -0.00001403, 0.00000475, -0.00000586, 0.00001802, -0.00001642, 0.00000646, -0.00001086, 0.00001175, -0.00001273, 0.00001545, -0.00001567, 0.00002237, -0.00002295, 0.00002060, -0.00002214, 0.00002913, -0.00003454, 0.00001409, -0.00001695, 0.00003189, -0.00003164, 0.00003579, -0.00003777, 0.00004431, -0.00005040, 0.00002877, -0.00003595, 0.00006074, -0.00006849, 0.00004536, -0.00005364, 0.00007828, -0.00009168, 0.00007570, -0.00009662, 0.00012920, -0.00017085, 0.00016296, -0.00025908, 0.00031072, -0.00072869, 0.00067754, 0.00059808, -0.00069922, 0.00030337, -0.00025527, 0.00016193, -0.00017044, 0.00012958, -0.00009750, 0.00007637, -0.00009310, 0.00008037, -0.00005475, 0.00004407, -0.00006713, 0.00006146, -0.00003692, 0.00002788, -0.00004897, 0.00004396, -0.00003792, 0.00003644, -0.00003266, 0.00003303, -0.00001730, 0.00001258, -0.00003298, 0.00002882, -0.00002225, 0.00002057, -0.00002306, 0.00002293, -0.00001623, 0.00001612, -0.00001336, 0.00001181, -0.00001060, 0.00000551, -0.00001548, 0.00001848, -0.00000653, 0.00000417, -0.00001323, 0.00000847, -0.00000866, 0.00000816, 0.00000218, -0.00000199, -0.00000286, 0.00000161, -0.00000472, 0.00000689, 0.00000409, -0.00000429, -0.00000263, -0.00000298, 0.00000075, -0.00001072, 0.00003275, -0.00003500, 0.00003125, -0.00003167, 0.00002732, -0.00002431, },
		{ -0.00004870, 0.00005545, -0.00004593, 0.00003269, -0.00002695, 0.00002397, -0.00001905, 0.00000869, 0.00000991, -0.00000826, 0.00000113, -0.00000456, -0.00000347, -0.00000012, 0.00000780, 0.00000007, -0.00000369, 0.00000317, -0.00000673, 0.00000198, 0.00000753, -0.00000343, 0.00000197, -0.00000321, -0.00000598, -0.00000060, 0.00000815, 0.00000104, -0.00000773, -0.00000074, 0.00000077, -0.00000167, 0.00000362, -0.00000332, 0.00000693, -0.00000280, 0.00000013, -0.00000258, 0.00000572, -0.00000172, -0.00001457, -0.00000058, 0.00001191, -0.00000493, 0.00000648, -0.00000520, 0.00000654, -0.00000260, -0.00001409, -0.00000563, 0.00002020, -0.00000737, -0.00001389, -0.00000582, 0.00002012, -0.00001207, -0.00000573, -0.00001766, 0.00002841, -0.00002632, -0.00000768, -0.00003628, -0.00000646, -0.00006409, -0.00049198, 0.00135983, -0.00059424, 0.00002270, -0.00006294, 0.00000795, -0.00004187, 0.00000233, 0.00000437, 0.00000227, -0.00002395, 0.00000326, 0.00000957, 0.00000151, -0.00002384, 0.00000513, 0.00001088, -0.00000002, -0.00002103, 0.00000665, -0.00000218, 0.00000289, -0.00000050, 0.00000046, 0.00000890, 0.00000084, -0.00001875, 0.00000482, 0.00000001, 0.00000211, -0.00000457, 0.00000191, 0.00000321, -0.00000016, 0.00000131, -0.00000056, -0.00000028, 0.00000009, -0.00000996, 0.00000514, 0.00000589, -0.00000062, -0.00000719, -0.00000043, -0.00000059, -0.00000121, 0.00000720, -0.00000014, -0.00000473, 0.00000249, -0.00000472, 0.00000260, 0.00000635, -0.00000029, -0.00000450, -0.00000352, 0.00000316, -0.00001360, 0.00002045, -0.00000290, -0.00001058, 0.00001654, -0.00002865, 0.00003635, },
		{ -0.00002656, 0.00001254, 0.00000860, -0.00001872, 0.00001746, -0.00001648, 0.00002120, -0.00002664, 0.00003016, -0.00001410, 0.00000497, -0.00000696, -0.00000122, -0.00000423, 0.00000526, 0.00000646, -0.00000667, 0.00000433, -0.00000510, -0.00000140, 0.00000415, 0.00000596, -0.00000770, 0.00000836, -0.00001467, 0.00000531, -0.00000514, 0.00001726, -0.00001705, 0.00000730, -0.00001109, 0.00001165, -0.00001199, 0.00001456, -0.00001479, 0.00002138, -0.00002233, 0.00002012, -0.00002147, 0.00002872, -0.00003547, 0.00001534, -0.00001645, 0.00003034, -0.00002997, 0.00003432, -0.00003667, 0.00004362, -0.00005089, 0.00002904, -0.00003429, 0.00005884, -0.00006854, 0.00004571, -0.00005163, 0.00007475, -0.00008841, 0.00007304, -0.00009342, 0.00012577, -0.00016746, 0.00015945, -0.00025704, 0.00030982, -0.00074732, 0.00075020, 0.00051461, -0.00066062, 0.00028850, -0.00024593, 0.00015652, -0.00016629, 0.00012688, -0.00009599, 0.00007499, -0.00009256, 0.00008091, -0.00005489, 0.00004189, -0.00006450, 0.00006099, -0.00003718, 0.00002642, -0.00004669, 0.00004260, -0.00003712, 0.00003624, -0.00003296, 0.00003371, -0.00001751, 0.00001089, -0.00003088, 0.00002786, -0.00002179, 0.00002002, -0.00002264, 0.00002304, -0.00001645, 0.00001651, -0.00001386, 0.00001182, -0.00001033, 0.00000450, -0.00001427, 0.00001863, -0.00000712, 0.00000360, -0.00001229, 0.00000820, -0.00000884, 0.00000916, 0.00000086, -0.00000174, -0.00000210, 0.00000035, -0.00000360, 0.00000693, 0.00000344, -0.00000419, -0.00000343, -0.00000061, -0.00000220, -0.00000742, 0.00003141, -0.00003653, 0.00003406, -0.00003770, 0.00003138, },
		{ 0.00003585, -0.00004768, 0.00005325, -0.00004214, 0.00002846, -0.00002347, 0.00002519, -0.00002131, 0.00001342, 0.00000441, -0.00000498, -0.00000048, -0.00000461, -0.00000314, 0.00000010, 0.00000835, -0.00000128, -0.00000279, 0.00000303, -0.00000736, 0.00000308, 0.00000731, -0.00000425, 0.00000315, -0.00000470, -0.00000516, -0.00000059, 0.00000907, -0.00000105, -0.00000629, -0.00000137, 0.00000133, -0.00000210, 0.00000450, -0.00000462, 0.00000849, -0.00000491, 0.00000235, -0.00000481, 0.00000847, -0.00000522, -0.00001194, -0.00000148, 0.00001305, -0.00000710, 0.00000952, -0.00000890, 0.00001072, -0.00000741, -0.00001025, -0.00000832, 0.00002427, -0.00001338, -0.00000864, -0.00000931, 0.00002451, -0.00001882, 0.00000273, -0.00002680, 0.00003931, -0.00003976, 0.00000899, -0.00005831, 0.00002260, -0.00011231, -0.00042466, 0.00133445, -0.00062644, 0.00005792, -0.00008760, 0.00002798, -0.00005762, 0.00001608, -0.00000757, 0.00001207, -0.00003285, 0.00001116, 0.00000387, 0.00000500, -0.00002812, 0.00001113, 0.00000602, 0.00000263, -0.00002382, 0.00001071, -0.00000629, 0.00000690, -0.00000414, 0.00000346, 0.00000710, 0.00000131, -0.00002014, 0.00000760, -0.00000267, 0.00000434, -0.00000683, 0.00000432, 0.00000120, 0.00000159, -0.00000006, 0.00000011, -0.00000078, 0.00000029, -0.00001069, 0.00000702, 0.00000462, -0.00000061, -0.00000756, 0.00000079, -0.00000187, 0.00000013, 0.00000660, -0.00000097, -0.00000355, 0.00000176, -0.00000485, 0.00000371, 0.00000548, -0.00000022, -0.00000524, -0.00000243, 0.00000331, -0.00001517, 0.00002470, -0.00000883, -0.00000517, 0.00001090, -0.00002716, },
		{ 0.00003432, -0.00002800, 0.00001328, 0.00000780, -0.00001611, 0.00001254, -0.00000826, 0.00001478, -0.00002036, 0.00002637, -0.00001401, 0.00000598, -0.00000832, -0.00000077, -0.00000400, 0.00000572, 0.00000609, -0.00000739, 0.00000561, -0.00000645, -0.00000060, 0.00000471, 0.00000488, -0.00000699, 0.00000797, -0.00001510, 0.00000583, -0.00000438, 0.00001622, -0.00001733, 0.00000797, -0.00001124, 0.00001152, -0.00001119, 0.00001350, -0.00001365, 0.00002003, -0.00002124, 0.00001916, -0.00002026, 0.00002761, -0.00003572, 0.00001628, -0.00001582, 0.00002847, -0.00002774, 0.00003212, -0.00003469, 0.00004190, -0.00005042, 0.00002867, -0.00003198, 0.00005584, -0.00006731, 0.00004515, -0.00004880, 0.00006997, -0.00008347, 0.00006857, -0.00008808, 0.00011947, -0.00016045, 0.00015151, -0.00024914, 0.00030031, -0.00075384, 0.00081354, 0.00042998, -0.00061485, 0.00026719, -0.00023169, 0.00014711, -0.00015864, 0.00012124, -0.00009218, 0.00007160, -0.00009006, 0.00007986, -0.00005405, 0.00003889, -0.00006069, 0.00005935, -0.00003674, 0.00002445, -0.00004365, 0.00004031, -0.00003541, 0.00003517, -0.00003253, 0.00003388, -0.00001757, 0.00000910, -0.00002834, 0.00002629, -0.00002082, 0.00001899, -0.00002171, 0.00002268, -0.00001629, 0.00001660, -0.00001419, 0.00001177, -0.00001006, 0.00000346, -0.00001282, 0.00001849, -0.00000763, 0.00000304, -0.00001125, 0.00000773, -0.00000880, 0.00001002, -0.00000061, -0.00000115, -0.00000165, -0.00000077, -0.00000245, 0.00000688, 0.00000277, -0.00000407, -0.00000418, 0.00000183, -0.00000547, -0.00000331, 0.00002842, -0.00003612, 0.00003502, -0.00004219, },
		{ -0.00002450, 0.00003400, -0.00004524, 0.00004976, -0.00003692, 0.00002281, -0.00001828, 0.00002459, -0.00002189, 0.00001691, -0.00000073, -0.00000165, -0.00000226, -0.00000446, -0.00000290, 0.00000039, 0.00000877, -0.00000264, -0.00000167, 0.00000253, -0.00000767, 0.00000411, 0.00000691, -0.00000486, 0.00000421, -0.00000619, -0.00000422, -0.00000058, 0.00000981, -0.00000309, -0.00000471, -0.00000212, 0.00000193, -0.00000245, 0.00000519, -0.00000566, 0.00000975, -0.00000673, 0.00000435, -0.00000681, 0.00001098, -0.00000868, -0.00000908, -0.00000246, 0.00001390, -0.00000882, 0.00001207, -0.00001214, 0.00001452, -0.00001205, -0.00000638, -0.00001078, 0.00002772, -0.00001892, -0.00000348, -0.00001250, 0.00002807, -0.00002456, 0.00001024, -0.00003479, 0.00004873, -0.00005167, 0.00002421, -0.00007880, 0.00005022, -0.00016110, -0.00034918, 0.00129321, -0.00064511, 0.00008602, -0.00010836, 0.00004537, -0.00007141, 0.00002853, -0.00001867, 0.00002110, -0.00004103, 0.00001873, -0.00000179, 0.00000815, -0.00003167, 0.00001666, 0.00000125, 0.00000502, -0.00002600, 0.00001415, -0.00000997, 0.00001062, -0.00000765, 0.00000651, 0.00000518, 0.00000162, -0.00002101, 0.00000990, -0.00000507, 0.00000634, -0.00000887, 0.00000663, -0.00000080, 0.00000337, -0.00000153, 0.00000087, -0.00000128, 0.00000036, -0.00001113, 0.00000873, 0.00000329, -0.00000060, -0.00000779, 0.00000187, -0.00000310, 0.00000157, 0.00000571, -0.00000156, -0.00000238, 0.00000086, -0.00000476, 0.00000470, 0.00000455, -0.00000006, -0.00000612, -0.00000094, 0.00000279, -0.00001576, 0.00002782, -0.00001421, 0.00000039, 0.00000454, },
		{ -0.00004488, 0.00003585, -0.00002837, 0.00001327, 0.00000775, -0.00001407, 0.00000823, -0.00000029, 0.00000809, -0.00001325, 0.00002123, -0.00001281, 0.00000621, -0.00000917, -0.00000056, -0.00000360, 0.00000606, 0.00000560, -0.00000788, 0.00000673, -0.00000782, 0.00000046, 0.00000496, 0.00000387, -0.00000616, 0.00000736, -0.00001531, 0.00000629, -0.00000359, 0.00001492, -0.00001726, 0.00000845, -0.00001131, 0.00001134, -0.00001035, 0.00001231, -0.00001232, 0.00001838, -0.00001975, 0.00001778, -0.00001857, 0.00002583, -0.00003526, 0.00001687, -0.00001508, 0.00002634, -0.00002507, 0.00002931, -0.00003192, 0.00003920, -0.00004901, 0.00002771, -0.00002912, 0.00005186, -0.00006484, 0.00004372, -0.00004530, 0.00006418, -0.00007713, 0.00006259, -0.00008093, 0.00011068, -0.00015019, 0.00013950, -0.00023570, 0.00028225, -0.00074743, 0.00086535, 0.00034698, -0.00056409, 0.00024077, -0.00021336, 0.00013428, -0.00014792, 0.00011295, -0.00008627, 0.00006636, -0.00008570, 0.00007724, -0.00005226, 0.00003520, -0.00005585, 0.00005663, -0.00003565, 0.00002208, -0.00004000, 0.00003723, -0.00003289, 0.00003330, -0.00003135, 0.00003353, -0.00001748, 0.00000728, -0.00002549, 0.00002423, -0.00001938, 0.00001751, -0.00002028, 0.00002186, -0.00001575, 0.00001635, -0.00001432, 0.00001164, -0.00000980, 0.00000245, -0.00001119, 0.00001805, -0.00000805, 0.00000252, -0.00001015, 0.00000710, -0.00000854, 0.00001069, -0.00000214, -0.00000026, -0.00000152, -0.00000168, -0.00000133, 0.00000674, 0.00000212, -0.00000398, -0.00000478, 0.00000416, -0.00000881, 0.00000134, 0.00002401, -0.00003384, 0.00003404, },
		{ -0.00000208, -0.00002103, 0.00003111, -0.00004166, 0.00004526, -0.00003055, 0.00001601, -0.00001156, 0.00002221, -0.00002068, 0.00001886, -0.00000516, 0.00000145, -0.00000404, -0.00000420, -0.00000268, 0.00000072, 0.00000905, -0.00000396, -0.00000037, 0.00000167, -0.00000759, 0.00000497, 0.00000636, -0.00000527, 0.00000511, -0.00000765, -0.00000318, -0.00000057, 0.00001035, -0.00000501, -0.00000303, -0.00000298, 0.00000256, -0.00000274, 0.00000568, -0.00000644, 0.00001068, -0.00000820, 0.00000604, -0.00000848, 0.00001311, -0.00001195, -0.00000610, -0.00000349, 0.00001444, -0.00001004, 0.00001404, -0.00001477, 0.00001775, -0.00001634, -0.00000264, -0.00001287, 0.00003040, -0.00002378, 0.00000138, -0.00001524, 0.00003067, -0.00002910, 0.00001653, -0.00004128, 0.00005622, -0.00006145, 0.00003720, -0.00009670, 0.00007494, -0.00020832, -0.00026801, 0.00123764, -0.00065029, 0.00010628, -0.00012441, 0.00005937, -0.00008258, 0.00003907, -0.00002842, 0.00002891, -0.00004809, 0.00002561, -0.00000715, 0.00001083, -0.00003435, 0.00002151, -0.00000324, 0.00000705, -0.00002748, 0.00001686, -0.00001305, 0.00001388, -0.00001087, 0.00000948, 0.00000320, 0.00000177, -0.00002133, 0.00001166, -0.00000711, 0.00000802, -0.00001061, 0.00000874, -0.00000270, 0.00000510, -0.00000303, 0.00000168, -0.00000178, 0.00000032, -0.00001128, 0.00001021, 0.00000194, -0.00000057, -0.00000787, 0.00000277, -0.00000422, 0.00000307, 0.00000454, -0.00000186, -0.00000131, -0.00000013, -0.00000447, 0.00000556, 0.00000358, 0.00000018, -0.00000709, 0.00000089, 0.00000159, -0.00001529, 0.00002956, -0.00001862, 0.00000559, },
		{ 0.00003115, -0.00004566, 0.00003579, -0.00002748, 0.00001232, 0.00000866, -0.00001289, 0.00000491, 0.00000696, 0.00000163, -0.00000580, 0.00001507, -0.00001062, 0.00000570, -0.00000952, -0.00000060, -0.00000302, 0.00000626, 0.00000502, -0.00000811, 0.00000761, -0.00000912, 0.00000172, 0.00000491, 0.00000298, -0.00000526, 0.00000654, -0.00001528, 0.00000668, -0.00000279, 0.00001340, -0.00001683, 0.00000869, -0.00001126, 0.00001110, -0.00000950, 0.00001106, -0.00001087, 0.00001652, -0.00001795, 0.00001606, -0.00001647, 0.00002343, -0.00003411, 0.00001708, -0.00001424, 0.00002406, -0.00002209, 0.00002603, -0.00002850, 0.00003564, -0.00004673, 0.00002619, -0.00002583, 0.00004705, -0.00006127, 0.00004151, -0.00004129, 0.00005765, -0.00006972, 0.00005544, -0.00007239, 0.00009986, -0.00013719, 0.00012401, -0.00021732, 0.00025616, -0.00072783, 0.00090391, 0.00026822, -0.00051055, 0.00021071, -0.00019199, 0.00011878, -0.00013469, 0.00010246, -0.00007860, 0.00005955, -0.00007966, 0.00007315, -0.00004961, 0.00003095, -0.00005015, 0.00005294, -0.00003400, 0.00001943, -0.00003591, 0.00003351, -0.00002967, 0.00003071, -0.00002946, 0.00003265, -0.00001726, 0.00000552, -0.00002244, 0.00002177, -0.00001758, 0.00001567, -0.00001843, 0.00002062, -0.00001485, 0.00001577, -0.00001423, 0.00001143, -0.00000957, 0.00000151, -0.00000943, 0.00001735, -0.00000836, 0.00000204, -0.00000901, 0.00000634, -0.00000804, 0.00001110, -0.00000364, 0.00000086, -0.00000168, -0.00000235, -0.00000027, 0.00000651, 0.00000149, -0.00000394, -0.00000514, 0.00000620, -0.00001197, 0.00000619, 0.00001851, -0.00002989, },
		{ 0.00000998, -0.00000847, -0.00001720, 0.00002749, -0.00003725, 0.00004008, -0.00002338, 0.00000842, -0.00000362, 0.00001823, -0.00001773, 0.00001912, -0.00000861, 0.00000408, -0.00000563, -0.00000395, -0.00000243, 0.00000105, 0.00000918, -0.00000519, 0.00000105, 0.00000051, -0.00000712, 0.00000561, 0.00000573, -0.00000547, 0.00000582, -0.00000903, -0.00000206, -0.00000056, 0.00001067, -0.00000675, -0.00000133, -0.00000389, 0.00000322, -0.00000298, 0.00000598, -0.00000695, 0.00001126, -0.00000928, 0.00000737, -0.00000974, 0.00001476, -0.00001488, -0.00000311, -0.00000452, 0.00001467, -0.00001074, 0.00001538, -0.00001668, 0.00002024, -0.00002010, 0.00000082, -0.00001451, 0.00003220, -0.00002777, 0.00000576, -0.00001744, 0.00003222, -0.00003229, 0.00002138, -0.00004600, 0.00006143, -0.00006865, 0.00004731, -0.00011113, 0.00009539, -0.00025184, -0.00018381, 0.00116978, -0.00064258, 0.00011840, -0.00013525, 0.00006950, -0.00009067, 0.00004727, -0.00003640, 0.00003514, -0.00005366, 0.00003148, -0.00001199, 0.00001289, -0.00003603, 0.00002552, -0.00000732, 0.00000866, -0.00002821, 0.00001874, -0.00001542, 0.00001653, -0.00001365, 0.00001225, 0.00000120, 0.00000176, -0.00002112, 0.00001283, -0.00000869, 0.00000929, -0.00001195, 0.00001056, -0.00000441, 0.00000670, -0.00000452, 0.00000252, -0.00000230, 0.00000020, -0.00001115, 0.00001144, 0.00000061, -0.00000052, -0.00000782, 0.00000350, -0.00000521, 0.00000455, 0.00000312, -0.00000184, -0.00000042, -0.00000112, -0.00000404, 0.00000628, 0.00000258, 0.00000045, -0.00000809, 0.00000300, -0.00000024, -0.00001374, 0.00002974, -0.00002169, },
		{ -0.00002458, 0.00002651, -0.00004456, 0.00003407, -0.00002529, 0.00001037, 0.00001065, -0.00001272, 0.00000286, 0.00001307, -0.00000411, 0.00000145, 0.00000832, -0.00000769, 0.00000454, -0.00000938, -0.00000088, -0.00000226, 0.00000632, 0.00000436, -0.00000805, 0.00000817, -0.00001025, 0.00000308, 0.00000459, 0.00000223, -0.00000433, 0.00000552, -0.00001499, 0.00000698, -0.00000200, 0.00001169, -0.00001606, 0.00000868, -0.00001107, 0.00001081, -0.00000867, 0.00000980, -0.00000936, 0.00001452, -0.00001592, 0.00001409, -0.00001408, 0.00002051, -0.00003229, 0.00001690, -0.00001332, 0.00002171, -0.00001893, 0.00002245, -0.00002461, 0.00003138, -0.00004366, 0.00002421, -0.00002226, 0.00004161, -0.00005674, 0.00003862, -0.00003695, 0.00005066, -0.00006161, 0.00004753, -0.00006293, 0.00008761, -0.00012216, 0.00010581, -0.00019492, 0.00022296, -0.00069534, 0.00092805, 0.00019599, -0.00045641, 0.00017858, -0.00016867, 0.00010148, -0.00011966, 0.00009030, -0.00006961, 0.00005153, -0.00007225, 0.00006778, -0.00004623, 0.00002634, -0.00004380, 0.00004842, -0.00003188, 0.00001663, -0.00003155, 0.00002936, -0.00002594, 0.00002752, -0.00002695, 0.00003125, -0.00001692, 0.00000389, -0.00001934, 0.00001906, -0.00001551, 0.00001357, -0.00001623, 0.00001901, -0.00001361, 0.00001487, -0.00001391, 0.00001111, -0.00000936, 0.00000068, -0.00000760, 0.00001639, -0.00000857, 0.00000161, -0.00000788, 0.00000547, -0.00000734, 0.00001122, -0.00000503, 0.00000212, -0.00000212, -0.00000277, 0.00000069, 0.00000622, 0.00000092, -0.00000400, -0.00000519, 0.00000780, -0.00001472, 0.00001089, 0.00001232, },
		{ -0.00002318, 0.00001316, -0.00001417, -0.00001342, 0.00002353, -0.00003239, 0.00003459, -0.00001582, 0.00000050, 0.00000515, 0.00001299, -0.00001322, 0.00001764, -0.00001089, 0.00000605, -0.00000689, -0.00000380, -0.00000209, 0.00000133, 0.00000917, -0.00000628, 0.00000251, -0.00000091, -0.00000626, 0.00000597, 0.00000508, -0.00000549, 0.00000630, -0.00001027, -0.00000090, -0.00000055, 0.00001075, -0.00000825, 0.00000032, -0.00000483, 0.00000388, -0.00000317, 0.00000612, -0.00000719, 0.00001148, -0.00000994, 0.00000829, -0.00001055, 0.00001584, -0.00001735, -0.00000024, -0.00000550, 0.00001460, -0.00001091, 0.00001604, -0.00001781, 0.00002190, -0.00002317, 0.00000389, -0.00001563, 0.00003306, -0.00003075, 0.00000952, -0.00001903, 0.00003271, -0.00003408, 0.00002469, -0.00004884, 0.00006418, -0.00007294, 0.00005410, -0.00012139, 0.00011046, -0.00028969, -0.00009939, 0.00109204, -0.00062308, 0.00012254, -0.00014070, 0.00007550, -0.00009539, 0.00005283, -0.00004233, 0.00003953, -0.00005750, 0.00003607, -0.00001611, 0.00001427, -0.00003665, 0.00002855, -0.00001086, 0.00000979, -0.00002817, 0.00001974, -0.00001699, 0.00001847, -0.00001586, 0.00001471, -0.00000074, 0.00000162, -0.00002042, 0.00001341, -0.00000980, 0.00001013, -0.00001284, 0.00001201, -0.00000587, 0.00000810, -0.00000593, 0.00000337, -0.00000283, 0.00000001, -0.00001074, 0.00001239, -0.00000069, -0.00000044, -0.00000766, 0.00000404, -0.00000603, 0.00000597, 0.00000152, -0.00000148, 0.00000024, -0.00000203, -0.00000351, 0.00000687, 0.00000159, 0.00000073, -0.00000906, 0.00000527, -0.00000258, -0.00001119, 0.00002833, },
		{ 0.00000591, -0.00001832, 0.00002044, -0.00004173, 0.00003074, -0.00002182, 0.00000742, 0.00001374, -0.00001364, 0.00000229, 0.00001775, -0.00000871, 0.00000800, 0.00000141, -0.00000428, 0.00000290, -0.00000884, -0.00000137, -0.00000134, 0.00000622, 0.00000365, -0.00000770, 0.00000838, -0.00001112, 0.00000445, 0.00000405, 0.00000163, -0.00000340, 0.00000435, -0.00001444, 0.00000718, -0.00000125, 0.00000985, -0.00001497, 0.00000842, -0.00001072, 0.00001045, -0.00000785, 0.00000857, -0.00000787, 0.00001249, -0.00001376, 0.00001198, -0.00001151, 0.00001720, -0.00002988, 0.00001634, -0.00001235, 0.00001937, -0.00001574, 0.00001875, -0.00002045, 0.00002658, -0.00003996, 0.00002188, -0.00001855, 0.00003577, -0.00005145, 0.00003521, -0.00003246, 0.00004350, -0.00005315, 0.00003929, -0.00005309, 0.00007457, -0.00010585, 0.00008582, -0.00016963, 0.00018393, -0.00065079, 0.00093718, 0.00013214, -0.00040365, 0.00014589, -0.00014456, 0.00008327, -0.00010356, 0.00007707, -0.00005978, 0.00004273, -0.00006380, 0.00006134, -0.00004226, 0.00002154, -0.00003703, 0.00004327, -0.00002942, 0.00001383, -0.00002713, 0.00002496, -0.00002187, 0.00002389, -0.00002391, 0.00002939, -0.00001646, 0.00000245, -0.00001630, 0.00001621, -0.00001330, 0.00001130, -0.00001377, 0.00001711, -0.00001209, 0.00001367, -0.00001334, 0.00001068, -0.00000917, -0.00000001, -0.00000576, 0.00001522, -0.00000865, 0.00000124, -0.00000677, 0.00000453, -0.00000645, 0.00001104, -0.00000622, 0.00000344, -0.00000276, -0.00000294, 0.00000153, 0.00000588, 0.00000040, -0.00000414, -0.00000489, 0.00000883, -0.00001681, 0.00001508, },
		{ 0.00002540, -0.00002298, 0.00001485, -0.00001878, -0.00001010, 0.00001957, -0.00002742, 0.00002920, -0.00000829, -0.00000730, 0.00001427, 0.00000689, -0.00000748, 0.00001454, -0.00001192, 0.00000722, -0.00000770, -0.00000383, -0.00000162, 0.00000155, 0.00000902, -0.00000717, 0.00000393, -0.00000250, -0.00000504, 0.00000603, 0.00000445, -0.00000534, 0.00000655, -0.00001132, 0.00000027, -0.00000053, 0.00001060, -0.00000946, 0.00000187, -0.00000575, 0.00000452, -0.00000333, 0.00000611, -0.00000719, 0.00001136, -0.00001018, 0.00000878, -0.00001088, 0.00001631, -0.00001925, 0.00000242, -0.00000641, 0.00001425, -0.00001059, 0.00001606, -0.00001814, 0.00002265, -0.00002546, 0.00000647, -0.00001619, 0.00003293, -0.00003263, 0.00001255, -0.00001998, 0.00003218, -0.00003450, 0.00002646, -0.00004976, 0.00006441, -0.00007421, 0.00005729, -0.00012704, 0.00011930, -0.00032012, -0.00001748, 0.00100709, -0.00059331, 0.00011922, -0.00014091, 0.00007737, -0.00009667, 0.00005560, -0.00004606, 0.00004192, -0.00005942, 0.00003919, -0.00001935, 0.00001490, -0.00003617, 0.00003053, -0.00001380, 0.00001045, -0.00002739, 0.00001988, -0.00001772, 0.00001962, -0.00001741, 0.00001674, -0.00000259, 0.00000138, -0.00001927, 0.00001340, -0.00001040, 0.00001049, -0.00001323, 0.00001305, -0.00000701, 0.00000924, -0.00000719, 0.00000418, -0.00000337, -0.00000020, -0.00001008, 0.00001304, -0.00000192, -0.00000035, -0.00000739, 0.00000440, -0.00000664, 0.00000726, -0.00000019, -0.00000083, 0.00000061, -0.00000281, -0.00000294, 0.00000733, 0.00000061, 0.00000099, -0.00000990, 0.00000755, -0.00000527, -0.00000780, },
		{ 0.00001845, -0.00000029, -0.00001160, 0.00001331, -0.00003743, 0.00002598, -0.00001725, 0.00000361, 0.00001783, -0.00001560, 0.00000324, 0.00002081, -0.00001189, 0.00001341, -0.00000521, -0.00000070, 0.00000095, -0.00000800, -0.00000199, -0.00000032, 0.00000596, 0.00000290, -0.00000708, 0.00000820, -0.00001165, 0.00000573, 0.00000335, 0.00000121, -0.00000251, 0.00000305, -0.00001365, 0.00000726, -0.00000055, 0.00000794, -0.00001360, 0.00000791, -0.00001020, 0.00001000, -0.00000706, 0.00000741, -0.00000646, 0.00001049, -0.00001157, 0.00000984, -0.00000890, 0.00001364, -0.00002697, 0.00001543, -0.00001136, 0.00001712, -0.00001265, 0.00001510, -0.00001621, 0.00002146, -0.00003577, 0.00001931, -0.00001489, 0.00002974, -0.00004562, 0.00003141, -0.00002799, 0.00003645, -0.00004471, 0.00003114, -0.00004337, 0.00006141, -0.00008907, 0.00006504, -0.00014271, 0.00014067, -0.00059556, 0.00093135, 0.00007805, -0.00035399, 0.00011405, -0.00012075, 0.00006507, -0.00008715, 0.00006341, -0.00004963, 0.00003362, -0.00005471, 0.00005414, -0.00003790, 0.00001677, -0.00003009, 0.00003767, -0.00002672, 0.00001114, -0.00002281, 0.00002053, -0.00001765, 0.00001998, -0.00002046, 0.00002710, -0.00001590, 0.00000125, -0.00001344, 0.00001336, -0.00001104, 0.00000899, -0.00001116, 0.00001498, -0.00001036, 0.00001220, -0.00001255, 0.00001011, -0.00000898, -0.00000056, -0.00000395, 0.00001386, -0.00000862, 0.00000091, -0.00000572, 0.00000357, -0.00000540, 0.00001054, -0.00000714, 0.00000471, -0.00000354, -0.00000289, 0.00000224, 0.00000549, -0.00000008, -0.00000438, -0.00000421, 0.00000922, -0.00001809, },
		{ -0.00000380, 0.00002113, -0.00002110, 0.00001487, -0.00002202, -0.00000756, 0.00001592, -0.00002270, 0.00002427, -0.00000118, -0.00001449, 0.00002325, 0.00000041, -0.00000094, 0.00001004, -0.00001172, 0.00000755, -0.00000800, -0.00000409, -0.00000097, 0.00000167, 0.00000873, -0.00000783, 0.00000523, -0.00000417, -0.00000354, 0.00000578, 0.00000389, -0.00000505, 0.00000655, -0.00001215, 0.00000140, -0.00000051, 0.00001021, -0.00001036, 0.00000325, -0.00000659, 0.00000513, -0.00000346, 0.00000598, -0.00000698, 0.00001093, -0.00001003, 0.00000887, -0.00001073, 0.00001613, -0.00002052, 0.00000476, -0.00000720, 0.00001363, -0.00000984, 0.00001546, -0.00001769, 0.00002247, -0.00002689, 0.00000848, -0.00001619, 0.00003185, -0.00003338, 0.00001478, -0.00002030, 0.00003071, -0.00003365, 0.00002677, -0.00004887, 0.00006223, -0.00007251, 0.00005688, -0.00012792, 0.00012140, -0.00034175, 0.00005935, 0.00091774, -0.00055509, 0.00010932, -0.00013630, 0.00007535, -0.00009464, 0.00005565, -0.00004758, 0.00004231, -0.00005937, 0.00004071, -0.00002162, 0.00001477, -0.00003463, 0.00003144, -0.00001608, 0.00001064, -0.00002594, 0.00001920, -0.00001763, 0.00001997, -0.00001824, 0.00001828, -0.00000428, 0.00000107, -0.00001776, 0.00001285, -0.00001053, 0.00001039, -0.00001312, 0.00001363, -0.00000779, 0.00001006, -0.00000826, 0.00000492, -0.00000391, -0.00000043, -0.00000920, 0.00001337, -0.00000305, -0.00000023, -0.00000704, 0.00000458, -0.00000704, 0.00000835, -0.00000194, 0.00000009, 0.00000069, -0.00000339, -0.00000238, 0.00000768, -0.00000032, 0.00000117, -0.00001054, 0.00000972, -0.00000813, },
		{ -0.00001843, 0.00002073, -0.00000584, -0.00000490, 0.00000560, -0.00003203, 0.00002005, -0.00001181, -0.00000083, 0.00002272, -0.00001845, 0.00000565, 0.00002221, -0.00001345, 0.00001731, -0.00001114, 0.00000273, -0.00000110, -0.00000699, -0.00000268, 0.00000078, 0.00000555, 0.00000215, -0.00000621, 0.00000763, -0.00001180, 0.00000684, 0.00000254, 0.00000094, -0.00000171, 0.00000167, -0.00001262, 0.00000722, 0.00000008, 0.00000600, -0.00001201, 0.00000717, -0.00000952, 0.00000946, -0.00000631, 0.00000635, -0.00000517, 0.00000861, -0.00000945, 0.00000777, -0.00000637, 0.00000997, -0.00002367, 0.00001422, -0.00001037, 0.00001503, -0.00000976, 0.00001165, -0.00001209, 0.00001622, -0.00003126, 0.00001664, -0.00001142, 0.00002377, -0.00003945, 0.00002739, -0.00002372, 0.00002976, -0.00003661, 0.00002346, -0.00003427, 0.00004874, -0.00007264, 0.00004450, -0.00011553, 0.00009502, -0.00053145, 0.00091119, 0.00003455, -0.00030881, 0.00008429, -0.00009823, 0.00004770, -0.00007117, 0.00004993, -0.00003968, 0.00002466, -0.00004542, 0.00004646, -0.00003334, 0.00001222, -0.00002322, 0.00003181, -0.00002393, 0.00000869, -0.00001875, 0.00001624, -0.00001349, 0.00001599, -0.00001675, 0.00002447, -0.00001525, 0.00000032, -0.00001085, 0.00001061, -0.00000887, 0.00000674, -0.00000850, 0.00001274, -0.00000850, 0.00001054, -0.00001155, 0.00000941, -0.00000879, -0.00000093, -0.00000223, 0.00001235, -0.00000847, 0.00000064, -0.00000474, 0.00000261, -0.00000425, 0.00000974, -0.00000773, 0.00000584, -0.00000438, -0.00000265, 0.00000280, 0.00000507, -0.00000051, -0.00000467, -0.00000318, 0.00000892, },
		{ -0.00001093, 0.00000055, 0.00001578, -0.00001766, 0.00001320, -0.00002369, -0.00000605, 0.00001284, -0.00001850, 0.00002011, 0.00000513, -0.00002065, 0.00003161, -0.00000593, 0.00000591, 0.00000448, -0.00001041, 0.00000704, -0.00000775, -0.00000460, -0.00000013, 0.00000165, 0.00000832, -0.00000822, 0.00000632, -0.00000582, -0.00000183, 0.00000525, 0.00000342, -0.00000466, 0.00000629, -0.00001271, 0.00000246, -0.00000047, 0.00000959, -0.00001090, 0.00000441, -0.00000731, 0.00000565, -0.00000356, 0.00000576, -0.00000660, 0.00001023, -0.00000953, 0.00000858, -0.00001014, 0.00001534, -0.00002110, 0.00000672, -0.00000785, 0.00001280, -0.00000871, 0.00001435, -0.00001653, 0.00002140, -0.00002744, 0.00000992, -0.00001566, 0.00002987, -0.00003301, 0.00001618, -0.00002001, 0.00002842, -0.00003167, 0.00002579, -0.00004638, 0.00005789, -0.00006807, 0.00005306, -0.00012417, 0.00011665, -0.00035360, 0.00012882, 0.00082675, -0.00051046, 0.00009399, -0.00012755, 0.00006990, -0.00008963, 0.00005316, -0.00004702, 0.00004080, -0.00005740, 0.00004061, -0.00002288, 0.00001392, -0.00003208, 0.00003127, -0.00001769, 0.00001042, -0.00002391, 0.00001779, -0.00001677, 0.00001952, -0.00001832, 0.00001927, -0.00000579, 0.00000072, -0.00001597, 0.00001185, -0.00001022, 0.00000985, -0.00001252, 0.00001375, -0.00000819, 0.00001053, -0.00000909, 0.00000555, -0.00000444, -0.00000063, -0.00000814, 0.00001340, -0.00000405, -0.00000010, -0.00000663, 0.00000460, -0.00000722, 0.00000920, -0.00000364, 0.00000120, 0.00000049, -0.00000375, -0.00000185, 0.00000792, -0.00000120, 0.00000127, -0.00001088, 0.00001159, },
		{ 0.00000795, -0.00001779, 0.00002175, -0.00001037, 0.00000130, -0.00000221, -0.00002593, 0.00001331, -0.00000584, -0.00000560, 0.00002815, -0.00002191, 0.00000934, 0.00002205, -0.00001339, 0.00001949, -0.00001602, 0.00000575, -0.00000305, -0.00000593, -0.00000335, 0.00000189, 0.00000500, 0.00000145, -0.00000515, 0.00000669, -0.00001153, 0.00000769, 0.00000169, 0.00000082, -0.00000103, 0.00000027, -0.00001139, 0.00000707, 0.00000061, 0.00000411, -0.00001027, 0.00000625, -0.00000870, 0.00000882, -0.00000558, 0.00000540, -0.00000405, 0.00000690, -0.00000746, 0.00000586, -0.00000405, 0.00000636, -0.00002012, 0.00001275, -0.00000942, 0.00001315, -0.00000718, 0.00000856, -0.00000828, 0.00001107, -0.00002660, 0.00001398, -0.00000830, 0.00001807, -0.00003319, 0.00002330, -0.00001976, 0.00002362, -0.00002912, 0.00001658, -0.00002620, 0.00003714, -0.00005728, 0.00002515, -0.00008945, 0.00004889, -0.00046061, 0.00087791, 0.00000194, -0.00026909, 0.00005757, -0.00007782, 0.00003188, -0.00005624, 0.00003718, -0.00003041, 0.00001629, -0.00003632, 0.00003862, -0.00002879, 0.00000808, -0.00001665, 0.00002590, -0.00002113, 0.00000658, -0.00001511, 0.00001228, -0.00000957, 0.00001208, -0.00001292, 0.00002158, -0.00001453, -0.00000031, -0.00000860, 0.00000808, -0.00000686, 0.00000465, -0.00000591, 0.00001046, -0.00000658, 0.00000874, -0.00001038, 0.00000858, -0.00000858, -0.00000113, -0.00000063, 0.00001073, -0.00000820, 0.00000041, -0.00000385, 0.00000170, -0.00000303, 0.00000868, -0.00000798, 0.00000675, -0.00000521, -0.00000228, 0.00000324, 0.00000462, -0.00000091, -0.00000498, -0.00000186, },
		{ 0.00001304, -0.00001346, 0.00000493, 0.00000972, -0.00001295, 0.00000994, -0.00002375, -0.00000571, 0.00001046, -0.00001501, 0.00001697, 0.00001038, -0.00002542, 0.00003893, -0.00001167, 0.00001252, -0.00000173, -0.00000820, 0.00000578, -0.00000700, -0.00000535, 0.00000090, 0.00000150, 0.00000780, -0.00000832, 0.00000715, -0.00000735, -0.00000000, 0.00000446, 0.00000307, -0.00000419, 0.00000577, -0.00001298, 0.00000342, -0.00000043, 0.00000877, -0.00001109, 0.00000531, -0.00000786, 0.00000607, -0.00000361, 0.00000548, -0.00000609, 0.00000932, -0.00000873, 0.00000797, -0.00000918, 0.00001396, -0.00002098, 0.00000823, -0.00000833, 0.00001178, -0.00000730, 0.00001281, -0.00001477, 0.00001950, -0.00002713, 0.00001077, -0.00001467, 0.00002711, -0.00003159, 0.00001676, -0.00001918, 0.00002547, -0.00002879, 0.00002376, -0.00004258, 0.00005174, -0.00006130, 0.00004624, -0.00011619, 0.00010530, -0.00035519, 0.00018907, 0.00073675, -0.00046155, 0.00007454, -0.00011549, 0.00006162, -0.00008210, 0.00004848, -0.00004464, 0.00003762, -0.00005369, 0.00003895, -0.00002313, 0.00001243, -0.00002866, 0.00003011, -0.00001866, 0.00000985, -0.00002143, 0.00001577, -0.00001526, 0.00001836, -0.00001768, 0.00001967, -0.00000708, 0.00000038, -0.00001399, 0.00001047, -0.00000954, 0.00000895, -0.00001147, 0.00001344, -0.00000822, 0.00001064, -0.00000964, 0.00000604, -0.00000494, -0.00000081, -0.00000693, 0.00001313, -0.00000491, 0.00000004, -0.00000617, 0.00000448, -0.00000717, 0.00000977, -0.00000519, 0.00000243, 0.00000004, -0.00000387, -0.00000139, 0.00000806, -0.00000202, 0.00000125, -0.00001087, },
		{ -0.00000031, 0.00000638, -0.00001619, 0.00002143, -0.00001360, 0.00000658, -0.00000964, -0.00001957, 0.00000612, 0.00000030, -0.00001036, 0.00003378, -0.00002565, 0.00001399, 0.00002057, -0.00001183, 0.00001984, -0.00001962, 0.00000810, -0.00000472, -0.00000495, -0.00000391, 0.00000294, 0.00000433, 0.00000081, -0.00000395, 0.00000544, -0.00001084, 0.00000824, 0.00000086, 0.00000084, -0.00000050, -0.00000112, -0.00000998, 0.00000680, 0.00000103, 0.00000232, -0.00000844, 0.00000518, -0.00000775, 0.00000809, -0.00000488, 0.00000457, -0.00000312, 0.00000542, -0.00000569, 0.00000420, -0.00000205, 0.00000294, -0.00001645, 0.00001111, -0.00000852, 0.00001152, -0.00000496, 0.00000592, -0.00000494, 0.00000620, -0.00002196, 0.00001146, -0.00000564, 0.00001284, -0.00002704, 0.00001928, -0.00001624, 0.00001818, -0.00002244, 0.00001074, -0.00001947, 0.00002706, -0.00004363, 0.00000785, -0.00006570, 0.00000424, -0.00038545, 0.00083316, -0.00002000, -0.00023543, 0.00003459, -0.00006014, 0.00001815, -0.00004290, 0.00002563, -0.00002221, 0.00000892, -0.00002780, 0.00003093, -0.00002442, 0.00000451, -0.00001058, 0.00002010, -0.00001844, 0.00000486, -0.00001196, 0.00000877, -0.00000604, 0.00000843, -0.00000914, 0.00001852, -0.00001375, -0.00000066, -0.00000674, 0.00000583, -0.00000511, 0.00000283, -0.00000349, 0.00000825, -0.00000470, 0.00000688, -0.00000908, 0.00000763, -0.00000834, -0.00000116, 0.00000080, 0.00000904, -0.00000783, 0.00000022, -0.00000305, 0.00000087, -0.00000182, 0.00000739, -0.00000786, 0.00000738, -0.00000593, -0.00000182, 0.00000356, 0.00000415, -0.00000131, -0.00000526, },
		{ -0.00001046, 0.00001394, -0.00001551, 0.00000903, 0.00000334, -0.00000731, 0.00000530, -0.00002227, -0.00000656, 0.00000885, -0.00001234, 0.00001498, 0.00001438, -0.00002855, 0.00004487, -0.00001638, 0.00001839, -0.00000812, -0.00000536, 0.00000393, -0.00000583, -0.00000630, 0.00000209, 0.00000119, 0.00000719, -0.00000812, 0.00000764, -0.00000865, 0.00000183, 0.00000348, 0.00000285, -0.00000367, 0.00000503, -0.00001294, 0.00000424, -0.00000037, 0.00000776, -0.00001093, 0.00000592, -0.00000822, 0.00000635, -0.00000361, 0.00000516, -0.00000550, 0.00000826, -0.00000770, 0.00000711, -0.00000792, 0.00001209, -0.00002020, 0.00000928, -0.00000864, 0.00001064, -0.00000570, 0.00001098, -0.00001253, 0.00001689, -0.00002602, 0.00001109, -0.00001331, 0.00002372, -0.00002922, 0.00001658, -0.00001791, 0.00002205, -0.00002523, 0.00002095, -0.00003784, 0.00004423, -0.00005269, 0.00003699, -0.00010465, 0.00008801, -0.00034653, 0.00023872, 0.00065008, -0.00041044, 0.00005237, -0.00010111, 0.00005126, -0.00007264, 0.00004205, -0.00004080, 0.00003307, -0.00004847, 0.00003588, -0.00002245, 0.00001040, -0.00002452, 0.00002803, -0.00001903, 0.00000900, -0.00001863, 0.00001331, -0.00001321, 0.00001658, -0.00001637, 0.00001949, -0.00000812, 0.00000006, -0.00001193, 0.00000882, -0.00000857, 0.00000775, -0.00001004, 0.00001272, -0.00000790, 0.00001039, -0.00000990, 0.00000635, -0.00000539, -0.00000093, -0.00000561, 0.00001256, -0.00000561, 0.00000019, -0.00000567, 0.00000424, -0.00000691, 0.00001002, -0.00000652, 0.00000367, -0.00000061, -0.00000377, -0.00000101, 0.00000812, -0.00000278, 0.00000111, },
		{ -0.00000543, 0.00000134, 0.00000432, -0.00001374, 0.00001977, -0.00001538, 0.00001059, -0.00001626, -0.00001339, -0.00000112, 0.00000622, -0.00001473, 0.00003925, -0.00002930, 0.00001926, 0.00001809, -0.00000904, 0.00001841, -0.00002179, 0.00000962, -0.00000596, -0.00000416, -0.00000429, 0.00000389, 0.00000357, 0.00000028, -0.00000268, 0.00000394, -0.00000978, 0.00000845, 0.00000011, 0.00000097, -0.00000015, -0.00000243, -0.00000845, 0.00000644, 0.00000132, 0.00000068, -0.00000659, 0.00000403, -0.00000671, 0.00000729, -0.00000419, 0.00000386, -0.00000240, 0.00000418, -0.00000416, 0.00000283, -0.00000044, -0.00000014, -0.00001279, 0.00000935, -0.00000770, 0.00001015, -0.00000316, 0.00000381, -0.00000220, 0.00000177, -0.00001748, 0.00000916, -0.00000355, 0.00000824, -0.00002120, 0.00001546, -0.00001322, 0.00001354, -0.00001671, 0.00000608, -0.00001431, 0.00001882, -0.00003217, -0.00000671, -0.00004539, -0.00003712, -0.00030845, 0.00077899, -0.00003199, -0.00020800, 0.00001573, -0.00004557, 0.00000690, -0.00003151, 0.00001561, -0.00001539, 0.00000284, -0.00002020, 0.00002368, -0.00002040, 0.00000163, -0.00000518, 0.00001458, -0.00001593, 0.00000358, -0.00000939, 0.00000582, -0.00000303, 0.00000518, -0.00000554, 0.00001537, -0.00001291, -0.00000073, -0.00000529, 0.00000391, -0.00000367, 0.00000135, -0.00000131, 0.00000618, -0.00000293, 0.00000503, -0.00000770, 0.00000659, -0.00000805, -0.00000102, 0.00000203, 0.00000731, -0.00000735, 0.00000007, -0.00000236, 0.00000014, -0.00000066, 0.00000596, -0.00000740, 0.00000769, -0.00000649, -0.00000132, 0.00000377, 0.00000367, -0.00000172, },
		{ 0.00000086, -0.00000963, 0.00001419, -0.00001692, 0.00001255, -0.00000294, -0.00000116, -0.00000037, -0.00001945, -0.00000850, 0.00000797, -0.00001049, 0.00001419, 0.00001708, -0.00002989, 0.00004922, -0.00001975, 0.00002303, -0.00001422, -0.00000222, 0.00000166, -0.00000434, -0.00000739, 0.00000340, 0.00000074, 0.00000650, -0.00000762, 0.00000778, -0.00000965, 0.00000356, 0.00000237, 0.00000275, -0.00000313, 0.00000407, -0.00001257, 0.00000490, -0.00000031, 0.00000661, -0.00001044, 0.00000624, -0.00000835, 0.00000647, -0.00000356, 0.00000481, -0.00000487, 0.00000710, -0.00000651, 0.00000608, -0.00000645, 0.00000982, -0.00001882, 0.00000986, -0.00000877, 0.00000941, -0.00000400, 0.00000898, -0.00000999, 0.00001372, -0.00002421, 0.00001092, -0.00001168, 0.00001986, -0.00002605, 0.00001572, -0.00001628, 0.00001832, -0.00002124, 0.00001767, -0.00003253, 0.00003587, -0.00004286, 0.00002602, -0.00009041, 0.00006573, -0.00032814, 0.00027693, 0.00056874, -0.00035909, 0.00002887, -0.00008539, 0.00003960, -0.00006188, 0.00003440, -0.00003591, 0.00002755, -0.00004210, 0.00003162, -0.00002097, 0.00000799, -0.00001983, 0.00002519, -0.00001887, 0.00000798, -0.00001565, 0.00001055, -0.00001080, 0.00001431, -0.00001447, 0.00001875, -0.00000891, -0.00000019, -0.00000987, 0.00000700, -0.00000741, 0.00000634, -0.00000831, 0.00001165, -0.00000727, 0.00000980, -0.00000984, 0.00000646, -0.00000577, -0.00000100, -0.00000423, 0.00001173, -0.00000614, 0.00000032, -0.00000516, 0.00000390, -0.00000647, 0.00000995, -0.00000755, 0.00000483, -0.00000138, -0.00000346, -0.00000073, 0.00000809, -0.00000346, },
		{ -0.00000216, -0.00000544, 0.00000298, 0.00000192, -0.00001059, 0.00001691, -0.00001564, 0.00001311, -0.00002173, -0.00000776, -0.00000806, 0.00001156, -0.00001837, 0.00004421, -0.00003247, 0.00002474, 0.00001501, -0.00000539, 0.00001541, -0.00002251, 0.00001020, -0.00000668, -0.00000365, -0.00000442, 0.00000468, 0.00000273, -0.00000012, -0.00000142, 0.00000227, -0.00000838, 0.00000831, -0.00000052, 0.00000119, 0.00000001, -0.00000361, -0.00000684, 0.00000598, 0.00000148, -0.00000075, -0.00000479, 0.00000285, -0.00000564, 0.00000641, -0.00000352, 0.00000326, -0.00000187, 0.00000322, -0.00000291, 0.00000179, 0.00000070, -0.00000278, -0.00000927, 0.00000756, -0.00000698, 0.00000904, -0.00000177, 0.00000228, -0.00000014, -0.00000207, -0.00001331, 0.00000715, -0.00000207, 0.00000440, -0.00001582, 0.00001195, -0.00001075, 0.00000974, -0.00001197, 0.00000267, -0.00001079, 0.00001262, -0.00002318, -0.00001806, -0.00002933, -0.00007354, -0.00023211, 0.00071767, -0.00003511, -0.00018660, 0.00000108, -0.00003427, -0.00000169, -0.00002229, 0.00000734, -0.00001012, -0.00000174, -0.00001378, 0.00001708, -0.00001686, -0.00000047, -0.00000058, 0.00000946, -0.00001365, 0.00000275, -0.00000742, 0.00000348, -0.00000061, 0.00000244, -0.00000225, 0.00001224, -0.00001202, -0.00000056, -0.00000425, 0.00000236, -0.00000258, 0.00000024, 0.00000055, 0.00000432, -0.00000135, 0.00000327, -0.00000630, 0.00000547, -0.00000772, -0.00000072, 0.00000304, 0.00000560, -0.00000679, -0.00000006, -0.00000178, -0.00000047, 0.00000040, 0.00000444, -0.00000664, 0.00000766, -0.00000684, -0.00000083, 0.00000389, 0.00000319, },
		{ -0.00000408, 0.00000052, -0.00000839, 0.00001376, -0.00001755, 0.00001525, -0.00000873, 0.00000505, -0.00000670, -0.00001556, -0.00001135, 0.00000767, -0.00000939, 0.00001454, 0.00001852, -0.00002944, 0.00005190, -0.00002158, 0.00002607, -0.00001960, 0.00000090, -0.00000079, -0.00000268, -0.00000850, 0.00000477, 0.00000017, 0.00000577, -0.00000684, 0.00000754, -0.00001028, 0.00000509, 0.00000120, 0.00000277, -0.00000262, 0.00000293, -0.00001189, 0.00000538, -0.00000024, 0.00000535, -0.00000965, 0.00000627, -0.00000825, 0.00000640, -0.00000343, 0.00000445, -0.00000424, 0.00000592, -0.00000523, 0.00000496, -0.00000490, 0.00000729, -0.00001690, 0.00000999, -0.00000873, 0.00000815, -0.00000230, 0.00000694, -0.00000730, 0.00001015, -0.00002182, 0.00001036, -0.00000990, 0.00001572, -0.00002227, 0.00001430, -0.00001442, 0.00001449, -0.00001708, 0.00001422, -0.00002706, 0.00002717, -0.00003245, 0.00001411, -0.00007447, 0.00003972, -0.00030096, 0.00030341, 0.00049426, -0.00030920, 0.00000535, -0.00006931, 0.00002743, -0.00005051, 0.00002607, -0.00003043, 0.00002148, -0.00003494, 0.00002643, -0.00001886, 0.00000536, -0.00001481, 0.00002173, -0.00001826, 0.00000687, -0.00001265, 0.00000766, -0.00000817, 0.00001172, -0.00001212, 0.00001749, -0.00000944, -0.00000035, -0.00000789, 0.00000512, -0.00000616, 0.00000483, -0.00000636, 0.00001030, -0.00000638, 0.00000891, -0.00000950, 0.00000636, -0.00000606, -0.00000099, -0.00000282, 0.00001066, -0.00000648, 0.00000044, -0.00000463, 0.00000349, -0.00000587, 0.00000957, -0.00000823, 0.00000583, -0.00000221, -0.00000299, -0.00000054, 0.00000798, },
		{ 0.00000269, -0.00000264, -0.00000521, 0.00000447, -0.00000064, -0.00000698, 0.00001303, -0.00001446, 0.00001403, -0.00002580, -0.00000298, -0.00001438, 0.00001601, -0.00002096, 0.00004834, -0.00003478, 0.00003001, 0.00001179, -0.00000135, 0.00001112, -0.00002186, 0.00000983, -0.00000683, -0.00000345, -0.00000426, 0.00000526, 0.00000183, -0.00000035, -0.00000023, 0.00000054, -0.00000674, 0.00000785, -0.00000100, 0.00000146, -0.00000004, -0.00000461, -0.00000519, 0.00000546, 0.00000150, -0.00000194, -0.00000311, 0.00000170, -0.00000456, 0.00000550, -0.00000286, 0.00000276, -0.00000153, 0.00000251, -0.00000193, 0.00000108, 0.00000136, -0.00000490, -0.00000600, 0.00000579, -0.00000636, 0.00000819, -0.00000078, 0.00000131, 0.00000119, -0.00000523, -0.00000953, 0.00000547, -0.00000122, 0.00000140, -0.00001104, 0.00000881, -0.00000882, 0.00000676, -0.00000823, 0.00000048, -0.00000888, 0.00000846, -0.00001679, -0.00002597, -0.00001809, -0.00010368, -0.00015877, 0.00065162, -0.00003076, -0.00017074, -0.00000954, -0.00002617, -0.00000762, -0.00001531, 0.00000090, -0.00000648, -0.00000471, -0.00000870, 0.00001132, -0.00001389, -0.00000174, 0.00000312, 0.00000486, -0.00001164, 0.00000234, -0.00000603, 0.00000176, 0.00000115, 0.00000030, 0.00000062, 0.00000920, -0.00001110, -0.00000019, -0.00000359, 0.00000117, -0.00000185, -0.00000045, 0.00000205, 0.00000273, -0.00000001, 0.00000167, -0.00000493, 0.00000431, -0.00000734, -0.00000029, 0.00000381, 0.00000393, -0.00000616, -0.00000017, -0.00000130, -0.00000094, 0.00000129, 0.00000294, -0.00000564, 0.00000730, -0.00000695, -0.00000039, 0.00000395, },
		{ 0.00000780, -0.00000465, 0.00000013, -0.00000679, 0.00001263, -0.00001733, 0.00001694, -0.00001366, 0.00001086, -0.00001323, -0.00001095, -0.00001486, 0.00000775, -0.00000886, 0.00001588, 0.00001886, -0.00002731, 0.00005296, -0.00002183, 0.00002726, -0.00002388, 0.00000368, -0.00000320, -0.00000101, -0.00000955, 0.00000613, -0.00000051, 0.00000502, -0.00000583, 0.00000694, -0.00001051, 0.00000634, 0.00000005, 0.00000288, -0.00000215, 0.00000166, -0.00001090, 0.00000567, -0.00000018, 0.00000404, -0.00000862, 0.00000604, -0.00000793, 0.00000613, -0.00000321, 0.00000409, -0.00000363, 0.00000476, -0.00000394, 0.00000383, -0.00000335, 0.00000461, -0.00001457, 0.00000970, -0.00000854, 0.00000690, -0.00000067, 0.00000500, -0.00000463, 0.00000638, -0.00001899, 0.00000950, -0.00000810, 0.00001149, -0.00001807, 0.00001243, -0.00001242, 0.00001070, -0.00001295, 0.00001085, -0.00002179, 0.00001863, -0.00002208, 0.00000208, -0.00005792, 0.00001141, -0.00026636, 0.00031839, 0.00042772, -0.00026220, -0.00001706, -0.00005376, 0.00001549, -0.00003914, 0.00001758, -0.00002481, 0.00001531, -0.00002743, 0.00002064, -0.00001629, 0.00000267, -0.00000966, 0.00001783, -0.00001730, 0.00000576, -0.00000974, 0.00000481, -0.00000551, 0.00000896, -0.00000943, 0.00001580, -0.00000971, -0.00000042, -0.00000608, 0.00000327, -0.00000490, 0.00000331, -0.00000430, 0.00000876, -0.00000531, 0.00000778, -0.00000888, 0.00000605, -0.00000625, -0.00000092, -0.00000142, 0.00000938, -0.00000663, 0.00000054, -0.00000411, 0.00000302, -0.00000516, 0.00000891, -0.00000853, 0.00000659, -0.00000300, -0.00000240, -0.00000042, },
		{ 0.00000396, 0.00000219, -0.00000317, -0.00000470, 0.00000569, -0.00000318, -0.00000315, 0.00000842, -0.00001201, 0.00001337, -0.00002833, 0.00000073, -0.00001985, 0.00001931, -0.00002226, 0.00005141, -0.00003593, 0.00003473, 0.00000883, 0.00000257, 0.00000597, -0.00002004, 0.00000859, -0.00000642, -0.00000359, -0.00000380, 0.00000560, 0.00000092, -0.00000041, 0.00000082, -0.00000116, -0.00000495, 0.00000710, -0.00000130, 0.00000178, -0.00000030, -0.00000538, -0.00000357, 0.00000488, 0.00000137, -0.00000286, -0.00000160, 0.00000065, -0.00000355, 0.00000458, -0.00000220, 0.00000235, -0.00000136, 0.00000204, -0.00000122, 0.00000070, 0.00000153, -0.00000643, -0.00000308, 0.00000411, -0.00000584, 0.00000756, -0.00000016, 0.00000088, 0.00000180, -0.00000765, -0.00000621, 0.00000414, -0.00000099, -0.00000075, -0.00000694, 0.00000610, -0.00000741, 0.00000455, -0.00000539, -0.00000063, -0.00000845, 0.00000624, -0.00001293, -0.00003042, -0.00001190, -0.00012659, -0.00009049, 0.00058323, -0.00002053, -0.00015966, -0.00001650, -0.00002102, -0.00001106, -0.00001048, -0.00000377, -0.00000440, -0.00000609, -0.00000505, 0.00000651, -0.00001155, -0.00000219, 0.00000588, 0.00000083, -0.00000990, 0.00000232, -0.00000518, 0.00000064, 0.00000227, -0.00000120, 0.00000299, 0.00000633, -0.00001015, 0.00000033, -0.00000327, 0.00000031, -0.00000145, -0.00000074, 0.00000316, 0.00000143, 0.00000105, 0.00000028, -0.00000366, 0.00000314, -0.00000690, 0.00000027, 0.00000433, 0.00000235, -0.00000548, -0.00000025, -0.00000091, -0.00000128, 0.00000199, 0.00000152, -0.00000448, 0.00000665, -0.00000680, -0.00000002, },
		{ -0.00000036, 0.00000753, -0.00000518, -0.00000026, -0.00000489, 0.00001085, -0.00001627, 0.00001751, -0.00001747, 0.00001587, -0.00001952, -0.00000602, -0.00001871, 0.00000795, -0.00000868, 0.00001800, 0.00001833, -0.00002372, 0.00005257, -0.00002062, 0.00002648, -0.00002678, 0.00000586, -0.00000537, 0.00000052, -0.00001043, 0.00000740, -0.00000127, 0.00000427, -0.00000463, 0.00000601, -0.00001033, 0.00000725, -0.00000102, 0.00000307, -0.00000176, 0.00000030, -0.00000964, 0.00000576, -0.00000012, 0.00000271, -0.00000740, 0.00000558, -0.00000741, 0.00000567, -0.00000290, 0.00000373, -0.00000308, 0.00000366, -0.00000270, 0.00000277, -0.00000192, 0.00000194, -0.00001195, 0.00000906, -0.00000822, 0.00000569, 0.00000082, 0.00000325, -0.00000214, 0.00000257, -0.00001587, 0.00000843, -0.00000639, 0.00000737, -0.00001365, 0.00001026, -0.00001040, 0.00000711, -0.00000906, 0.00000781, -0.00001703, 0.00001069, -0.00001235, -0.00000932, -0.00004182, -0.00001767, -0.00022600, 0.00032262, 0.00036971, -0.00021916, -0.00003742, -0.00003948, 0.00000441, -0.00002837, 0.00000942, -0.00001947, 0.00000943, -0.00001996, 0.00001455, -0.00001348, 0.00000011, -0.00000460, 0.00001365, -0.00001608, 0.00000472, -0.00000704, 0.00000213, -0.00000296, 0.00000619, -0.00000657, 0.00001375, -0.00000974, -0.00000039, -0.00000447, 0.00000153, -0.00000372, 0.00000188, -0.00000222, 0.00000710, -0.00000412, 0.00000648, -0.00000803, 0.00000554, -0.00000632, -0.00000078, -0.00000008, 0.00000793, -0.00000660, 0.00000060, -0.00000358, 0.00000253, -0.00000438, 0.00000800, -0.00000845, 0.00000705, -0.00000370, -0.00000174, },
		{ 0.00000024, 0.00000392, 0.00000167, -0.00000374, -0.00000387, 0.00000651, -0.00000550, 0.00000063, 0.00000341, -0.00000858, 0.00001127, -0.00002931, 0.00000326, -0.00002434, 0.00002130, -0.00002212, 0.00005324, -0.00003569, 0.00003861, 0.00000653, 0.00000586, 0.00000041, -0.00001730, 0.00000659, -0.00000551, -0.00000402, -0.00000305, 0.00000570, -0.00000000, -0.00000028, 0.00000167, -0.00000272, -0.00000311, 0.00000611, -0.00000142, 0.00000210, -0.00000078, -0.00000588, -0.00000203, 0.00000427, 0.00000112, -0.00000350, -0.00000030, -0.00000026, -0.00000265, 0.00000368, -0.00000156, 0.00000200, -0.00000132, 0.00000178, -0.00000074, 0.00000059, 0.00000124, -0.00000737, -0.00000057, 0.00000257, -0.00000544, 0.00000713, 0.00000017, 0.00000092, 0.00000172, -0.00000931, -0.00000340, 0.00000315, -0.00000132, -0.00000204, -0.00000356, 0.00000384, -0.00000648, 0.00000302, -0.00000334, -0.00000083, -0.00000926, 0.00000572, -0.00001134, -0.00003167, -0.00001066, -0.00014169, -0.00002901, 0.00051478, -0.00000610, -0.00015241, -0.00002040, -0.00001840, -0.00001231, -0.00000758, -0.00000683, -0.00000373, -0.00000599, -0.00000281, 0.00000269, -0.00000982, -0.00000185, 0.00000770, -0.00000258, -0.00000844, 0.00000261, -0.00000479, 0.00000006, 0.00000279, -0.00000206, 0.00000481, 0.00000371, -0.00000918, 0.00000093, -0.00000322, -0.00000025, -0.00000136, -0.00000064, 0.00000388, 0.00000044, 0.00000180, -0.00000086, -0.00000252, 0.00000201, -0.00000641, 0.00000092, 0.00000459, 0.00000089, -0.00000477, -0.00000032, -0.00000062, -0.00000147, 0.00000246, 0.00000026, -0.00000324, 0.00000575, -0.00000640, },
		{ -0.00000108, -0.00000034, 0.00000719, -0.00000568, -0.00000061, -0.00000279, 0.00000851, -0.00001443, 0.00001695, -0.00001995, 0.00001975, -0.00002518, -0.00000115, -0.00002258, 0.00000798, -0.00000861, 0.00002063, 0.00001724, -0.00001896, 0.00005104, -0.00001819, 0.00002378, -0.00002813, 0.00000723, -0.00000711, 0.00000179, -0.00001103, 0.00000852, -0.00000207, 0.00000354, -0.00000329, 0.00000481, -0.00000976, 0.00000777, -0.00000193, 0.00000331, -0.00000147, -0.00000108, -0.00000815, 0.00000566, -0.00000009, 0.00000141, -0.00000605, 0.00000494, -0.00000671, 0.00000503, -0.00000249, 0.00000338, -0.00000260, 0.00000267, -0.00000155, 0.00000184, -0.00000069, -0.00000061, -0.00000916, 0.00000813, -0.00000780, 0.00000456, 0.00000212, 0.00000179, 0.00000004, -0.00000109, -0.00001261, 0.00000726, -0.00000486, 0.00000352, -0.00000921, 0.00000793, -0.00000846, 0.00000384, -0.00000554, 0.00000526, -0.00001303, 0.00000373, -0.00000375, -0.00001941, -0.00002717, -0.00004596, -0.00018174, 0.00031722, 0.00032034, -0.00018081, -0.00005505, -0.00002703, -0.00000530, -0.00001864, 0.00000199, -0.00001474, 0.00000420, -0.00001291, 0.00000851, -0.00001063, -0.00000216, 0.00000019, 0.00000938, -0.00001469, 0.00000381, -0.00000464, -0.00000026, -0.00000066, 0.00000358, -0.00000367, 0.00001144, -0.00000954, -0.00000026, -0.00000311, -0.00000003, -0.00000269, 0.00000062, -0.00000023, 0.00000542, -0.00000291, 0.00000507, -0.00000701, 0.00000484, -0.00000628, -0.00000057, 0.00000118, 0.00000635, -0.00000640, 0.00000062, -0.00000307, 0.00000204, -0.00000358, 0.00000690, -0.00000800, 0.00000717, -0.00000423, },
		{ -0.00000579, 0.00000041, 0.00000385, 0.00000114, -0.00000435, -0.00000270, 0.00000685, -0.00000744, 0.00000410, -0.00000164, -0.00000450, 0.00000797, -0.00002885, 0.00000462, -0.00002780, 0.00002194, -0.00002049, 0.00005381, -0.00003396, 0.00004145, 0.00000514, 0.00000809, -0.00000508, -0.00001399, 0.00000405, -0.00000422, -0.00000469, -0.00000203, 0.00000554, -0.00000090, 0.00000003, 0.00000229, -0.00000408, -0.00000133, 0.00000496, -0.00000137, 0.00000241, -0.00000146, -0.00000610, -0.00000060, 0.00000365, 0.00000074, -0.00000385, 0.00000075, -0.00000099, -0.00000190, 0.00000284, -0.00000093, 0.00000170, -0.00000139, 0.00000168, -0.00000044, 0.00000071, 0.00000056, -0.00000772, 0.00000148, 0.00000121, -0.00000514, 0.00000683, 0.00000026, 0.00000133, 0.00000102, -0.00001024, -0.00000111, 0.00000246, -0.00000211, -0.00000256, -0.00000091, 0.00000202, -0.00000595, 0.00000205, -0.00000192, -0.00000035, -0.00001101, 0.00000654, -0.00001165, -0.00003016, -0.00001396, -0.00014884, 0.00002435, 0.00044831, 0.00001086, -0.00014800, -0.00002189, -0.00001779, -0.00001180, -0.00000630, -0.00000854, -0.00000420, -0.00000465, -0.00000188, -0.00000014, -0.00000867, -0.00000081, 0.00000861, -0.00000536, -0.00000721, 0.00000313, -0.00000477, -0.00000008, 0.00000278, -0.00000233, 0.00000606, 0.00000138, -0.00000819, 0.00000158, -0.00000339, -0.00000056, -0.00000151, -0.00000019, 0.00000423, -0.00000025, 0.00000224, -0.00000171, -0.00000156, 0.00000094, -0.00000589, 0.00000166, 0.00000459, -0.00000042, -0.00000405, -0.00000036, -0.00000040, -0.00000155, 0.00000270, -0.00000076, -0.00000203, 0.00000467, },
		{ -0.00000454, -0.00000046, -0.00000033, 0.00000677, -0.00000616, -0.00000084, -0.00000060, 0.00000573, -0.00001192, 0.00001531, -0.00002103, 0.00002224, -0.00002985, 0.00000331, -0.00002615, 0.00000757, -0.00000840, 0.00002353, 0.00001591, -0.00001339, 0.00004871, -0.00001493, 0.00001935, -0.00002788, 0.00000766, -0.00000831, 0.00000267, -0.00001128, 0.00000941, -0.00000287, 0.00000285, -0.00000188, 0.00000342, -0.00000884, 0.00000790, -0.00000266, 0.00000359, -0.00000131, -0.00000244, -0.00000647, 0.00000538, -0.00000008, 0.00000019, -0.00000465, 0.00000418, -0.00000590, 0.00000423, -0.00000198, 0.00000302, -0.00000220, 0.00000181, -0.00000052, 0.00000108, 0.00000027, -0.00000292, -0.00000634, 0.00000700, -0.00000731, 0.00000353, 0.00000320, 0.00000067, 0.00000180, -0.00000446, -0.00000935, 0.00000607, -0.00000361, 0.00000010, -0.00000494, 0.00000557, -0.00000666, 0.00000095, -0.00000249, 0.00000331, -0.00000993, -0.00000201, 0.00000334, -0.00002766, -0.00001482, -0.00007201, -0.00013554, 0.00030369, 0.00027933, -0.00014749, -0.00006952, -0.00001680, -0.00001328, -0.00001029, -0.00000442, -0.00001089, -0.00000009, -0.00000662, 0.00000278, -0.00000792, -0.00000401, 0.00000451, 0.00000517, -0.00001321, 0.00000308, -0.00000260, -0.00000229, 0.00000128, 0.00000126, -0.00000089, 0.00000899, -0.00000915, -0.00000004, -0.00000201, -0.00000138, -0.00000186, -0.00000040, 0.00000161, 0.00000378, -0.00000173, 0.00000364, -0.00000588, 0.00000400, -0.00000612, -0.00000028, 0.00000231, 0.00000469, -0.00000604, 0.00000059, -0.00000257, 0.00000156, -0.00000280, 0.00000569, -0.00000722, 0.00000694, },
		{ 0.00000350, -0.00000500, 0.00000049, 0.00000375, 0.00000058, -0.00000498, -0.00000120, 0.00000666, -0.00000887, 0.00000703, -0.00000639, -0.00000016, 0.00000378, -0.00002715, 0.00000489, -0.00003028, 0.00002124, -0.00001743, 0.00005315, -0.00003073, 0.00004319, 0.00000484, 0.00000887, -0.00001004, -0.00001044, 0.00000118, -0.00000270, -0.00000551, -0.00000082, 0.00000515, -0.00000176, 0.00000050, 0.00000265, -0.00000515, 0.00000029, 0.00000373, -0.00000117, 0.00000269, -0.00000232, -0.00000601, 0.00000067, 0.00000303, 0.00000026, -0.00000395, 0.00000155, -0.00000152, -0.00000134, 0.00000209, -0.00000034, 0.00000143, -0.00000153, 0.00000172, -0.00000029, 0.00000098, -0.00000043, -0.00000754, 0.00000306, 0.00000005, -0.00000492, 0.00000664, 0.00000022, 0.00000200, -0.00000018, -0.00001049, 0.00000070, 0.00000202, -0.00000324, -0.00000239, 0.00000104, 0.00000063, -0.00000575, 0.00000151, -0.00000095, 0.00000056, -0.00001335, 0.00000829, -0.00001337, -0.00002649, -0.00002112, -0.00014833, 0.00006875, 0.00038554, 0.00002882, -0.00014540, -0.00002170, -0.00001861, -0.00000999, -0.00000626, -0.00000922, -0.00000551, -0.00000232, -0.00000207, -0.00000208, -0.00000801, 0.00000080, 0.00000870, -0.00000754, -0.00000619, 0.00000379, -0.00000500, 0.00000011, 0.00000234, -0.00000206, 0.00000674, -0.00000063, -0.00000719, 0.00000220, -0.00000369, -0.00000071, -0.00000184, 0.00000054, 0.00000425, -0.00000067, 0.00000239, -0.00000227, -0.00000081, -0.00000003, -0.00000532, 0.00000245, 0.00000434, -0.00000155, -0.00000336, -0.00000038, -0.00000026, -0.00000152, 0.00000270, -0.00000149, -0.00000091, },
		{ 0.00000637, -0.00000462, 0.00000009, -0.00000031, 0.00000625, -0.00000663, -0.00000090, 0.00000155, 0.00000268, -0.00000892, 0.00001275, -0.00002072, 0.00002321, -0.00003328, 0.00000705, -0.00002917, 0.00000647, -0.00000784, 0.00002643, 0.00001466, -0.00000736, 0.00004600, -0.00001128, 0.00001352, -0.00002613, 0.00000712, -0.00000888, 0.00000311, -0.00001112, 0.00001001, -0.00000365, 0.00000223, -0.00000047, 0.00000193, -0.00000765, 0.00000765, -0.00000316, 0.00000386, -0.00000130, -0.00000370, -0.00000466, 0.00000495, -0.00000010, -0.00000091, -0.00000324, 0.00000336, -0.00000502, 0.00000331, -0.00000139, 0.00000267, -0.00000188, 0.00000108, 0.00000035, 0.00000051, 0.00000091, -0.00000488, -0.00000360, 0.00000575, -0.00000679, 0.00000262, 0.00000403, -0.00000007, 0.00000305, -0.00000740, -0.00000621, 0.00000492, -0.00000269, -0.00000279, -0.00000100, 0.00000330, -0.00000508, -0.00000149, 0.00000004, 0.00000200, -0.00000782, -0.00000638, 0.00000870, -0.00003371, -0.00000543, -0.00009455, -0.00008936, 0.00028371, 0.00024601, -0.00011925, -0.00008068, -0.00000895, -0.00001936, -0.00000353, -0.00000962, -0.00000806, -0.00000326, -0.00000132, -0.00000237, -0.00000553, -0.00000531, 0.00000823, 0.00000116, -0.00001171, 0.00000254, -0.00000095, -0.00000389, 0.00000279, -0.00000068, 0.00000166, 0.00000648, -0.00000859, 0.00000024, -0.00000118, -0.00000248, -0.00000126, -0.00000114, 0.00000322, 0.00000225, -0.00000065, 0.00000226, -0.00000469, 0.00000306, -0.00000585, 0.00000007, 0.00000329, 0.00000299, -0.00000555, 0.00000053, -0.00000208, 0.00000110, -0.00000209, 0.00000445, -0.00000617, },
		{ 0.00000002, 0.00000231, -0.00000408, 0.00000049, 0.00000362, -0.00000000, -0.00000557, 0.00000059, 0.00000591, -0.00000972, 0.00000925, -0.00001052, 0.00000405, -0.00000092, -0.00002447, 0.00000425, -0.00003190, 0.00001932, -0.00001308, 0.00005144, -0.00002612, 0.00004390, 0.00000563, 0.00000798, -0.00001407, -0.00000701, -0.00000177, -0.00000111, -0.00000636, 0.00000051, 0.00000456, -0.00000256, 0.00000112, 0.00000275, -0.00000590, 0.00000166, 0.00000248, -0.00000083, 0.00000289, -0.00000332, -0.00000562, 0.00000176, 0.00000243, -0.00000029, -0.00000380, 0.00000211, -0.00000183, -0.00000099, 0.00000147, 0.00000021, 0.00000117, -0.00000171, 0.00000184, -0.00000022, 0.00000134, -0.00000164, -0.00000688, 0.00000418, -0.00000091, -0.00000479, 0.00000650, 0.00000011, 0.00000282, -0.00000177, -0.00001015, 0.00000204, 0.00000176, -0.00000460, -0.00000167, 0.00000236, -0.00000039, -0.00000579, 0.00000128, -0.00000027, 0.00000163, -0.00001593, 0.00001055, -0.00001594, -0.00002140, -0.00003123, -0.00014082, 0.00010382, 0.00032782, 0.00004640, -0.00014366, -0.00002055, -0.00002029, -0.00000738, -0.00000708, -0.00000920, -0.00000729, 0.00000063, -0.00000315, -0.00000324, -0.00000773, 0.00000284, 0.00000806, -0.00000914, -0.00000533, 0.00000452, -0.00000538, 0.00000051, 0.00000160, -0.00000135, 0.00000689, -0.00000229, -0.00000619, 0.00000277, -0.00000407, -0.00000075, -0.00000228, 0.00000147, 0.00000399, -0.00000085, 0.00000228, -0.00000256, -0.00000028, -0.00000087, -0.00000473, 0.00000327, 0.00000387, -0.00000250, -0.00000270, -0.00000038, -0.00000017, -0.00000142, 0.00000249, -0.00000192, },
		{ -0.00000493, 0.00000549, -0.00000445, 0.00000053, -0.00000027, 0.00000565, -0.00000710, -0.00000073, 0.00000354, -0.00000049, -0.00000563, 0.00000949, -0.00001918, 0.00002266, -0.00003532, 0.00000987, -0.00003141, 0.00000451, -0.00000674, 0.00002913, 0.00001378, -0.00000121, 0.00004330, -0.00000777, 0.00000671, -0.00002306, 0.00000565, -0.00000883, 0.00000308, -0.00001052, 0.00001029, -0.00000437, 0.00000169, 0.00000089, 0.00000041, -0.00000628, 0.00000705, -0.00000341, 0.00000412, -0.00000145, -0.00000481, -0.00000280, 0.00000440, -0.00000017, -0.00000186, -0.00000190, 0.00000255, -0.00000413, 0.00000231, -0.00000071, 0.00000231, -0.00000164, 0.00000049, 0.00000106, 0.00000015, 0.00000121, -0.00000643, -0.00000106, 0.00000445, -0.00000626, 0.00000183, 0.00000463, -0.00000044, 0.00000376, -0.00000983, -0.00000331, 0.00000387, -0.00000210, -0.00000508, 0.00000248, 0.00000122, -0.00000375, -0.00000348, 0.00000204, 0.00000129, -0.00000667, -0.00000936, 0.00001223, -0.00003738, 0.00000059, -0.00011254, -0.00004499, 0.00025909, 0.00021949, -0.00009583, -0.00008859, -0.00000347, -0.00002351, 0.00000157, -0.00001356, -0.00000629, -0.00000524, 0.00000282, -0.00000675, -0.00000356, -0.00000601, 0.00001124, -0.00000252, -0.00001025, 0.00000219, 0.00000031, -0.00000506, 0.00000384, -0.00000215, 0.00000386, 0.00000403, -0.00000790, 0.00000058, -0.00000060, -0.00000333, -0.00000091, -0.00000156, 0.00000455, 0.00000088, 0.00000027, 0.00000099, -0.00000351, 0.00000205, -0.00000548, 0.00000048, 0.00000410, 0.00000131, -0.00000495, 0.00000043, -0.00000161, 0.00000068, -0.00000149, 0.00000323, },
		{ -0.00000202, 0.00000071, 0.00000118, -0.00000310, 0.00000043, 0.00000345, -0.00000063, -0.00000610, 0.00000261, 0.00000463, -0.00000994, 0.00001063, -0.00001378, 0.00000777, -0.00000575, -0.00002115, 0.00000292, -0.00003283, 0.00001637, -0.00000765, 0.00004892, -0.00002032, 0.00004375, 0.00000737, 0.00000530, -0.00001688, -0.00000399, -0.00000456, 0.00000039, -0.00000712, 0.00000188, 0.00000379, -0.00000329, 0.00000184, 0.00000262, -0.00000631, 0.00000272, 0.00000131, -0.00000038, 0.00000302, -0.00000442, -0.00000493, 0.00000263, 0.00000185, -0.00000090, -0.00000345, 0.00000244, -0.00000195, -0.00000087, 0.00000100, 0.00000072, 0.00000093, -0.00000190, 0.00000200, -0.00000019, 0.00000169, -0.00000296, -0.00000585, 0.00000487, -0.00000165, -0.00000471, 0.00000637, 0.00000002, 0.00000365, -0.00000359, -0.00000930, 0.00000299, 0.00000160, -0.00000605, -0.00000054, 0.00000313, -0.00000109, -0.00000598, 0.00000123, 0.00000027, 0.00000265, -0.00001842, 0.00001287, -0.00001881, -0.00001564, -0.00004324, -0.00012729, 0.00012962, 0.00027611, 0.00006250, -0.00014195, -0.00001905, -0.00002229, -0.00000443, -0.00000835, -0.00000884, -0.00000921, 0.00000387, -0.00000483, -0.00000381, -0.00000770, 0.00000515, 0.00000682, -0.00001022, -0.00000460, 0.00000522, -0.00000580, 0.00000100, 0.00000068, -0.00000033, 0.00000658, -0.00000360, -0.00000520, 0.00000323, -0.00000445, -0.00000075, -0.00000273, 0.00000254, 0.00000350, -0.00000085, 0.00000196, -0.00000259, 0.00000003, -0.00000158, -0.00000411, 0.00000409, 0.00000319, -0.00000324, -0.00000210, -0.00000035, -0.00000013, -0.00000127, 0.00000210, },
		{ 0.00000213, -0.00000359, 0.00000436, -0.00000404, 0.00000085, -0.00000019, 0.00000495, -0.00000756, -0.00000030, 0.00000524, -0.00000359, -0.00000228, 0.00000580, -0.00001660, 0.00002067, -0.00003592, 0.00001163, -0.00003276, 0.00000159, -0.00000503, 0.00003152, 0.00001348, 0.00000479, 0.00004096, -0.00000490, -0.00000058, -0.00001897, 0.00000340, -0.00000821, 0.00000259, -0.00000948, 0.00001023, -0.00000501, 0.00000124, 0.00000213, -0.00000103, -0.00000482, 0.00000617, -0.00000342, 0.00000434, -0.00000178, -0.00000573, -0.00000094, 0.00000375, -0.00000029, -0.00000264, -0.00000066, 0.00000180, -0.00000330, 0.00000128, 0.00000002, 0.00000195, -0.00000146, 0.00000004, 0.00000162, -0.00000003, 0.00000118, -0.00000753, 0.00000121, 0.00000317, -0.00000575, 0.00000116, 0.00000502, -0.00000046, 0.00000392, -0.00001167, -0.00000071, 0.00000296, -0.00000186, -0.00000673, 0.00000542, -0.00000058, -0.00000271, -0.00000502, 0.00000357, 0.00000112, -0.00000637, -0.00001103, 0.00001401, -0.00003869, 0.00000308, -0.00012525, -0.00000403, 0.00023161, 0.00019866, -0.00007677, -0.00009348, -0.00000019, -0.00002583, 0.00000509, -0.00001627, -0.00000553, -0.00000605, 0.00000573, -0.00001026, -0.00000209, -0.00000605, 0.00001346, -0.00000578, -0.00000885, 0.00000202, 0.00000121, -0.00000580, 0.00000443, -0.00000313, 0.00000565, 0.00000171, -0.00000710, 0.00000093, -0.00000023, -0.00000395, -0.00000079, -0.00000167, 0.00000558, -0.00000029, 0.00000099, -0.00000011, -0.00000240, 0.00000102, -0.00000504, 0.00000096, 0.00000471, -0.00000031, -0.00000427, 0.00000029, -0.00000116, 0.00000029, -0.00000101, },
		{ 0.00000158, -0.00000182, 0.00000113, 0.00000016, -0.00000211, 0.00000035, 0.00000323, -0.00000130, -0.00000650, 0.00000478, 0.00000288, -0.00000955, 0.00001113, -0.00001598, 0.00001069, -0.00001035, -0.00001749, 0.00000116, -0.00003331, 0.00001260, -0.00000141, 0.00004590, -0.00001363, 0.00004301, 0.00000980, 0.00000091, -0.00001826, -0.00000164, -0.00000698, 0.00000164, -0.00000769, 0.00000319, 0.00000290, -0.00000393, 0.00000262, 0.00000230, -0.00000638, 0.00000342, 0.00000026, 0.00000014, 0.00000303, -0.00000558, -0.00000399, 0.00000328, 0.00000131, -0.00000152, -0.00000295, 0.00000259, -0.00000188, -0.00000096, 0.00000068, 0.00000115, 0.00000070, -0.00000207, 0.00000217, -0.00000014, 0.00000198, -0.00000429, -0.00000454, 0.00000518, -0.00000220, -0.00000467, 0.00000622, 0.00000001, 0.00000438, -0.00000552, -0.00000808, 0.00000360, 0.00000149, -0.00000744, 0.00000084, 0.00000347, -0.00000153, -0.00000625, 0.00000128, 0.00000081, 0.00000341, -0.00002053, 0.00001486, -0.00002145, -0.00001000, -0.00005603, -0.00010899, 0.00014665, 0.00023093, 0.00007629, -0.00013961, -0.00001775, -0.00002414, -0.00000154, -0.00000975, -0.00000842, -0.00001094, 0.00000705, -0.00000684, -0.00000395, -0.00000776, 0.00000755, 0.00000513, -0.00001084, -0.00000395, 0.00000583, -0.00000618, 0.00000148, -0.00000028, 0.00000089, 0.00000588, -0.00000457, -0.00000424, 0.00000356, -0.00000478, -0.00000078, -0.00000314, 0.00000366, 0.00000285, -0.00000073, 0.00000149, -0.00000241, 0.00000013, -0.00000214, -0.00000348, 0.00000488, 0.00000235, -0.00000378, -0.00000158, -0.00000029, -0.00000013, -0.00000111, },
		{ -0.00000068, 0.00000121, -0.00000224, 0.00000306, -0.00000344, 0.00000105, -0.00000008, 0.00000416, -0.00000800, 0.00000042, 0.00000656, -0.00000646, 0.00000091, 0.00000196, -0.00001329, 0.00001744, -0.00003516, 0.00001231, -0.00003317, -0.00000232, -0.00000270, 0.00003354, 0.00001390, 0.00001046, 0.00003922, -0.00000313, -0.00000782, -0.00001422, 0.00000057, -0.00000713, 0.00000173, -0.00000805, 0.00000983, -0.00000554, 0.00000089, 0.00000322, -0.00000231, -0.00000338, 0.00000508, -0.00000320, 0.00000451, -0.00000228, -0.00000639, 0.00000085, 0.00000303, -0.00000045, -0.00000324, 0.00000043, 0.00000115, -0.00000257, 0.00000027, 0.00000078, 0.00000158, -0.00000134, -0.00000030, 0.00000205, -0.00000004, 0.00000084, -0.00000815, 0.00000313, 0.00000198, -0.00000528, 0.00000060, 0.00000523, -0.00000017, 0.00000356, -0.00001290, 0.00000153, 0.00000219, -0.00000193, -0.00000776, 0.00000776, -0.00000206, -0.00000194, -0.00000614, 0.00000469, 0.00000135, -0.00000676, -0.00001157, 0.00001421, -0.00003783, 0.00000214, -0.00013230, 0.00003221, 0.00020297, 0.00018236, -0.00006146, -0.00009576, 0.00000118, -0.00002656, 0.00000716, -0.00001791, -0.00000561, -0.00000581, 0.00000743, -0.00001283, -0.00000116, -0.00000546, 0.00001487, -0.00000856, -0.00000756, 0.00000200, 0.00000180, -0.00000618, 0.00000458, -0.00000361, 0.00000696, -0.00000039, -0.00000623, 0.00000128, -0.00000005, -0.00000437, -0.00000087, -0.00000147, 0.00000629, -0.00000124, 0.00000150, -0.00000100, -0.00000141, 0.00000003, -0.00000454, 0.00000149, 0.00000510, -0.00000183, -0.00000355, 0.00000014, -0.00000073, -0.00000006, },
		{ -0.00000096, 0.00000098, -0.00000136, 0.00000125, -0.00000068, -0.00000116, 0.00000027, 0.00000297, -0.00000203, -0.00000672, 0.00000698, 0.00000076, -0.00000864, 0.00001078, -0.00001703, 0.00001259, -0.00001440, -0.00001384, -0.00000077, -0.00003354, 0.00000825, 0.00000535, 0.00004271, -0.00000637, 0.00004205, 0.00001252, -0.00000497, -0.00001817, -0.00000010, -0.00000889, 0.00000252, -0.00000796, 0.00000438, 0.00000194, -0.00000447, 0.00000342, 0.00000184, -0.00000616, 0.00000376, -0.00000060, 0.00000071, 0.00000290, -0.00000673, -0.00000283, 0.00000371, 0.00000081, -0.00000214, -0.00000234, 0.00000260, -0.00000168, -0.00000126, 0.00000054, 0.00000152, 0.00000047, -0.00000222, 0.00000231, -0.00000005, 0.00000215, -0.00000553, -0.00000308, 0.00000518, -0.00000259, -0.00000465, 0.00000603, 0.00000013, 0.00000492, -0.00000740, -0.00000659, 0.00000393, 0.00000135, -0.00000867, 0.00000230, 0.00000349, -0.00000178, -0.00000655, 0.00000135, 0.00000143, 0.00000378, -0.00002203, 0.00001621, -0.00002341, -0.00000516, -0.00006846, -0.00008729, 0.00015574, 0.00019244, 0.00008724, -0.00013617, -0.00001703, -0.00002550, 0.00000095, -0.00001100, -0.00000819, -0.00001222, 0.00000986, -0.00000888, -0.00000386, -0.00000779, 0.00000988, 0.00000312, -0.00001106, -0.00000336, 0.00000630, -0.00000643, 0.00000185, -0.00000117, 0.00000218, 0.00000489, -0.00000521, -0.00000330, 0.00000374, -0.00000503, -0.00000087, -0.00000343, 0.00000474, 0.00000212, -0.00000053, 0.00000092, -0.00000206, 0.00000006, -0.00000254, -0.00000282, 0.00000561, 0.00000137, -0.00000413, -0.00000113, -0.00000019, -0.00000016, },
		{ -0.00000039, -0.00000049, 0.00000050, -0.00000097, 0.00000167, -0.00000267, 0.00000115, 0.00000007, 0.00000328, -0.00000841, 0.00000142, 0.00000742, -0.00000896, 0.00000372, -0.00000171, -0.00000957, 0.00001324, -0.00003318, 0.00001198, -0.00003270, -0.00000714, 0.00000018, 0.00003524, 0.00001507, 0.00001573, 0.00003822, -0.00000284, -0.00001450, -0.00000919, -0.00000261, -0.00000574, 0.00000057, -0.00000629, 0.00000910, -0.00000596, 0.00000062, 0.00000412, -0.00000338, -0.00000206, 0.00000386, -0.00000278, 0.00000461, -0.00000295, -0.00000678, 0.00000252, 0.00000229, -0.00000066, -0.00000365, 0.00000136, 0.00000064, -0.00000201, -0.00000066, 0.00000154, 0.00000122, -0.00000127, -0.00000054, 0.00000237, 0.00000007, 0.00000024, -0.00000833, 0.00000467, 0.00000092, -0.00000487, 0.00000015, 0.00000532, 0.00000034, 0.00000273, -0.00001352, 0.00000338, 0.00000156, -0.00000224, -0.00000820, 0.00000948, -0.00000318, -0.00000145, -0.00000690, 0.00000549, 0.00000184, -0.00000763, -0.00001123, 0.00001314, -0.00003517, -0.00000187, -0.00013364, 0.00006278, 0.00017466, 0.00016942, -0.00004922, -0.00009592, 0.00000103, -0.00002599, 0.00000804, -0.00001867, -0.00000632, -0.00000472, 0.00000805, -0.00001448, -0.00000075, -0.00000428, 0.00001547, -0.00001081, -0.00000638, 0.00000208, 0.00000212, -0.00000625, 0.00000437, -0.00000363, 0.00000777, -0.00000222, -0.00000532, 0.00000159, -0.00000000, -0.00000462, -0.00000112, -0.00000100, 0.00000669, -0.00000198, 0.00000180, -0.00000168, -0.00000058, -0.00000090, -0.00000399, 0.00000206, 0.00000527, -0.00000321, -0.00000283, -0.00000002, -0.00000032, },
		{ -0.00000021, -0.00000087, 0.00000039, -0.00000068, 0.00000110, -0.00000131, -0.00000030, 0.00000020, 0.00000265, -0.00000281, -0.00000671, 0.00000911, -0.00000161, -0.00000733, 0.00000968, -0.00001695, 0.00001332, -0.00001765, -0.00001045, -0.00000265, -0.00003373, 0.00000354, 0.00001234, 0.00003971, 0.00000112, 0.00004121, 0.00001504, -0.00001198, -0.00001668, 0.00000055, -0.00001019, 0.00000293, -0.00000787, 0.00000535, 0.00000094, -0.00000490, 0.00000419, 0.00000131, -0.00000570, 0.00000373, -0.00000124, 0.00000130, 0.00000262, -0.00000782, -0.00000150, 0.00000394, 0.00000034, -0.00000272, -0.00000167, 0.00000251, -0.00000140, -0.00000171, 0.00000056, 0.00000181, 0.00000024, -0.00000232, 0.00000241, 0.00000012, 0.00000214, -0.00000658, -0.00000156, 0.00000494, -0.00000284, -0.00000463, 0.00000577, 0.00000042, 0.00000519, -0.00000912, -0.00000496, 0.00000406, 0.00000114, -0.00000964, 0.00000371, 0.00000330, -0.00000191, -0.00000681, 0.00000140, 0.00000217, 0.00000369, -0.00002279, 0.00001670, -0.00002436, -0.00000171, -0.00007950, -0.00006366, 0.00015798, 0.00016043, 0.00009509, -0.00013138, -0.00001712, -0.00002614, 0.00000284, -0.00001190, -0.00000830, -0.00001285, 0.00001205, -0.00001070, -0.00000372, -0.00000764, 0.00001199, 0.00000094, -0.00001097, -0.00000279, 0.00000659, -0.00000651, 0.00000203, -0.00000188, 0.00000341, 0.00000370, -0.00000557, -0.00000242, 0.00000378, -0.00000516, -0.00000107, -0.00000356, 0.00000574, 0.00000135, -0.00000032, 0.00000033, -0.00000161, -0.00000016, -0.00000281, -0.00000215, 0.00000625, 0.00000030, -0.00000430, -0.00000078, -0.00000005, },
		{ 0.00000006, -0.00000071, -0.00000043, 0.00000005, 0.00000015, 0.00000027, -0.00000181, 0.00000116, 0.00000025, 0.00000230, -0.00000876, 0.00000270, 0.00000778, -0.00001098, 0.00000599, -0.00000494, -0.00000579, 0.00000841, -0.00003023, 0.00001077, -0.00003145, -0.00001274, 0.00000343, 0.00003674, 0.00001695, 0.00002063, 0.00003795, -0.00000427, -0.00002019, -0.00000428, -0.00000586, -0.00000422, -0.00000074, -0.00000429, 0.00000810, -0.00000627, 0.00000043, 0.00000483, -0.00000417, -0.00000094, 0.00000260, -0.00000218, 0.00000463, -0.00000375, -0.00000687, 0.00000400, 0.00000156, -0.00000092, -0.00000389, 0.00000210, 0.00000028, -0.00000164, -0.00000146, 0.00000228, 0.00000086, -0.00000122, -0.00000070, 0.00000261, 0.00000025, -0.00000056, -0.00000809, 0.00000581, 0.00000001, -0.00000453, -0.00000022, 0.00000532, 0.00000100, 0.00000151, -0.00001357, 0.00000484, 0.00000104, -0.00000272, -0.00000814, 0.00001061, -0.00000394, -0.00000120, -0.00000735, 0.00000605, 0.00000242, -0.00000875, -0.00001030, 0.00001117, -0.00003117, -0.00000840, -0.00012955, 0.00008712, 0.00014794, 0.00015877, -0.00003933, -0.00009445, -0.00000022, -0.00002447, 0.00000800, -0.00001880, -0.00000739, -0.00000303, 0.00000774, -0.00001530, -0.00000080, -0.00000260, 0.00001531, -0.00001252, -0.00000531, 0.00000222, 0.00000226, -0.00000610, 0.00000388, -0.00000326, 0.00000812, -0.00000374, -0.00000440, 0.00000185, -0.00000005, -0.00000476, -0.00000147, -0.00000030, 0.00000680, -0.00000253, 0.00000190, -0.00000212, 0.00000007, -0.00000171, -0.00000343, 0.00000268, 0.00000522, -0.00000441, -0.00000213, -0.00000016, },
		{ 0.00000012, -0.00000029, -0.00000085, -0.00000015, 0.00000013, 0.00000071, -0.00000174, 0.00000046, 0.00000018, 0.00000226, -0.00000365, -0.00000642, 0.00001104, -0.00000407, -0.00000575, 0.00000801, -0.00001583, 0.00001286, -0.00001994, -0.00000753, -0.00000429, -0.00003404, -0.00000135, 0.00001931, 0.00003719, 0.00000855, 0.00004087, 0.00001684, -0.00001966, -0.00001400, 0.00000034, -0.00001085, 0.00000286, -0.00000737, 0.00000606, -0.00000005, -0.00000523, 0.00000487, 0.00000078, -0.00000508, 0.00000338, -0.00000166, 0.00000189, 0.00000217, -0.00000878, -0.00000008, 0.00000400, -0.00000008, -0.00000323, -0.00000100, 0.00000238, -0.00000108, -0.00000228, 0.00000072, 0.00000201, 0.00000001, -0.00000237, 0.00000246, 0.00000037, 0.00000195, -0.00000739, -0.00000008, 0.00000453, -0.00000300, -0.00000460, 0.00000544, 0.00000089, 0.00000513, -0.00001058, -0.00000328, 0.00000404, 0.00000084, -0.00001027, 0.00000495, 0.00000301, -0.00000197, -0.00000701, 0.00000139, 0.00000303, 0.00000311, -0.00002274, 0.00001622, -0.00002408, -0.00000005, -0.00008826, -0.00003952, 0.00015463, 0.00013443, 0.00009985, -0.00012517, -0.00001811, -0.00002596, 0.00000400, -0.00001238, -0.00000882, -0.00001274, 0.00001343, -0.00001210, -0.00000368, -0.00000721, 0.00001374, -0.00000127, -0.00001062, -0.00000223, 0.00000669, -0.00000639, 0.00000200, -0.00000235, 0.00000449, 0.00000242, -0.00000568, -0.00000159, 0.00000369, -0.00000517, -0.00000139, -0.00000349, 0.00000658, 0.00000060, -0.00000015, -0.00000023, -0.00000112, -0.00000048, -0.00000296, -0.00000146, 0.00000677, -0.00000082, -0.00000431, -0.00000050, },
		{ -0.00000028, 0.00000041, -0.00000103, -0.00000048, -0.00000015, 0.00000106, -0.00000106, -0.00000091, 0.00000111, 0.00000043, 0.00000124, -0.00000902, 0.00000419, 0.00000763, -0.00001247, 0.00000760, -0.00000751, -0.00000227, 0.00000329, -0.00002658, 0.00000889, -0.00002961, -0.00001892, 0.00000685, 0.00003825, 0.00001939, 0.00002532, 0.00003824, -0.00000750, -0.00002453, 0.00000016, -0.00000891, -0.00000276, -0.00000207, -0.00000216, 0.00000687, -0.00000645, 0.00000030, 0.00000534, -0.00000467, -0.00000009, 0.00000137, -0.00000145, 0.00000455, -0.00000468, -0.00000665, 0.00000527, 0.00000086, -0.00000120, -0.00000397, 0.00000268, 0.00000009, -0.00000149, -0.00000211, 0.00000295, 0.00000052, -0.00000119, -0.00000080, 0.00000281, 0.00000046, -0.00000147, -0.00000751, 0.00000657, -0.00000072, -0.00000426, -0.00000052, 0.00000528, 0.00000172, 0.00000001, -0.00001312, 0.00000592, 0.00000062, -0.00000328, -0.00000768, 0.00001119, -0.00000436, -0.00000118, -0.00000755, 0.00000647, 0.00000294, -0.00000989, -0.00000909, 0.00000869, -0.00002639, -0.00001674, -0.00012061, 0.00010503, 0.00012374, 0.00014946, -0.00003116, -0.00009187, -0.00000217, -0.00002235, 0.00000732, -0.00001857, -0.00000855, -0.00000106, 0.00000676, -0.00001542, -0.00000120, -0.00000053, 0.00001446, -0.00001370, -0.00000435, 0.00000238, 0.00000227, -0.00000580, 0.00000321, -0.00000257, 0.00000802, -0.00000493, -0.00000350, 0.00000203, -0.00000014, -0.00000483, -0.00000187, 0.00000056, 0.00000666, -0.00000290, 0.00000183, -0.00000234, 0.00000052, -0.00000240, -0.00000286, 0.00000332, 0.00000496, -0.00000541, -0.00000148, },
		{ -0.00000030, 0.00000032, -0.00000039, -0.00000092, -0.00000057, 0.00000100, 0.00000012, -0.00000197, 0.00000109, 0.00000020, 0.00000181, -0.00000454, -0.00000580, 0.00001267, -0.00000649, -0.00000411, 0.00000596, -0.00001385, 0.00001127, -0.00002122, -0.00000522, -0.00000559, -0.00003456, -0.00000630, 0.00002608, 0.00003540, 0.00001567, 0.00004130, 0.00001740, -0.00002745, -0.00001044, -0.00000060, -0.00001092, 0.00000230, -0.00000647, 0.00000649, -0.00000099, -0.00000543, 0.00000543, 0.00000032, -0.00000438, 0.00000276, -0.00000184, 0.00000247, 0.00000154, -0.00000955, 0.00000138, 0.00000390, -0.00000048, -0.00000367, -0.00000037, 0.00000227, -0.00000080, -0.00000291, 0.00000100, 0.00000214, -0.00000021, -0.00000238, 0.00000246, 0.00000070, 0.00000155, -0.00000789, 0.00000125, 0.00000404, -0.00000308, -0.00000454, 0.00000505, 0.00000154, 0.00000471, -0.00001169, -0.00000164, 0.00000390, 0.00000044, -0.00001053, 0.00000592, 0.00000272, -0.00000202, -0.00000712, 0.00000135, 0.00000398, 0.00000210, -0.00002192, 0.00001477, -0.00002252, -0.00000040, -0.00009407, -0.00001619, 0.00014703, 0.00011377, 0.00010175, -0.00011763, -0.00001997, -0.00002496, 0.00000443, -0.00001242, -0.00000973, -0.00001187, 0.00001392, -0.00001292, -0.00000385, -0.00000643, 0.00001502, -0.00000339, -0.00001007, -0.00000169, 0.00000661, -0.00000608, 0.00000173, -0.00000254, 0.00000534, 0.00000113, -0.00000557, -0.00000083, 0.00000346, -0.00000506, -0.00000183, -0.00000319, 0.00000723, -0.00000009, -0.00000004, -0.00000071, -0.00000063, -0.00000085, -0.00000301, -0.00000074, 0.00000714, -0.00000192, -0.00000421, },
		{ -0.00000092, -0.00000036, 0.00000072, -0.00000137, -0.00000059, -0.00000009, 0.00000172, -0.00000225, -0.00000003, 0.00000104, 0.00000061, 0.00000009, -0.00000913, 0.00000584, 0.00000700, -0.00001341, 0.00000849, -0.00000925, 0.00000068, -0.00000177, -0.00002254, 0.00000655, -0.00002738, -0.00002544, 0.00001017, 0.00004001, 0.00002220, 0.00003002, 0.00003880, -0.00001244, -0.00002730, 0.00000382, -0.00001154, -0.00000153, -0.00000329, -0.00000002, 0.00000548, -0.00000652, 0.00000022, 0.00000569, -0.00000488, 0.00000045, 0.00000025, -0.00000063, 0.00000436, -0.00000567, -0.00000615, 0.00000629, 0.00000023, -0.00000152, -0.00000392, 0.00000312, 0.00000003, -0.00000155, -0.00000256, 0.00000355, 0.00000019, -0.00000118, -0.00000087, 0.00000298, 0.00000064, -0.00000242, -0.00000665, 0.00000698, -0.00000128, -0.00000406, -0.00000076, 0.00000526, 0.00000241, -0.00000166, -0.00001223, 0.00000665, 0.00000025, -0.00000385, -0.00000693, 0.00001131, -0.00000448, -0.00000133, -0.00000754, 0.00000683, 0.00000328, -0.00001084, -0.00000789, 0.00000610, -0.00002139, -0.00002608, -0.00010765, 0.00011666, 0.00010268, 0.00014077, -0.00002414, -0.00008862, -0.00000443, -0.00001995, 0.00000628, -0.00001819, -0.00000953, 0.00000092, 0.00000535, -0.00001501, -0.00000185, 0.00000180, 0.00001303, -0.00001438, -0.00000349, 0.00000252, 0.00000223, -0.00000544, 0.00000245, -0.00000167, 0.00000754, -0.00000578, -0.00000264, 0.00000213, -0.00000024, -0.00000487, -0.00000225, 0.00000152, 0.00000631, -0.00000312, 0.00000162, -0.00000238, 0.00000079, -0.00000294, -0.00000229, 0.00000396, 0.00000450, -0.00000619, },
		{ -0.00000401, -0.00000016, 0.00000055, -0.00000051, -0.00000109, -0.00000082, 0.00000184, -0.00000058, -0.00000203, 0.00000159, 0.00000028, 0.00000128, -0.00000546, -0.00000485, 0.00001388, -0.00000869, -0.00000256, 0.00000378, -0.00001126, 0.00000873, -0.00002151, -0.00000355, -0.00000649, -0.00003533, -0.00001123, 0.00003253, 0.00003451, 0.00002235, 0.00004270, 0.00001626, -0.00003479, -0.00000637, -0.00000207, -0.00001049, 0.00000133, -0.00000519, 0.00000662, -0.00000187, -0.00000553, 0.00000583, 0.00000001, -0.00000371, 0.00000195, -0.00000181, 0.00000301, 0.00000071, -0.00001008, 0.00000280, 0.00000370, -0.00000085, -0.00000400, 0.00000019, 0.00000221, -0.00000061, -0.00000354, 0.00000137, 0.00000220, -0.00000042, -0.00000234, 0.00000242, 0.00000109, 0.00000097, -0.00000808, 0.00000239, 0.00000352, -0.00000314, -0.00000445, 0.00000462, 0.00000232, 0.00000394, -0.00001241, -0.00000013, 0.00000369, -0.00000006, -0.00001041, 0.00000655, 0.00000250, -0.00000208, -0.00000714, 0.00000130, 0.00000494, 0.00000077, -0.00002043, 0.00001244, -0.00001977, -0.00000279, -0.00009650, 0.00000523, 0.00013650, 0.00009762, 0.00010115, -0.00010899, -0.00002252, -0.00002327, 0.00000422, -0.00001210, -0.00001093, -0.00001033, 0.00001353, -0.00001310, -0.00000431, -0.00000525, 0.00001576, -0.00000533, -0.00000937, -0.00000116, 0.00000637, -0.00000561, 0.00000125, -0.00000242, 0.00000588, -0.00000008, -0.00000531, -0.00000016, 0.00000315, -0.00000484, -0.00000237, -0.00000267, 0.00000765, -0.00000069, -0.00000002, -0.00000107, -0.00000021, -0.00000123, -0.00000298, 0.00000001, 0.00000734, -0.00000297, },
		{ -0.00000675, -0.00000046, -0.00000039, 0.00000098, -0.00000173, -0.00000073, 0.00000018, 0.00000210, -0.00000325, 0.00000077, 0.00000097, 0.00000077, -0.00000115, -0.00000907, 0.00000755, 0.00000596, -0.00001382, 0.00000865, -0.00001006, 0.00000287, -0.00000647, -0.00001839, 0.00000399, -0.00002499, -0.00003204, 0.00001310, 0.00004229, 0.00002515, 0.00003502, 0.00003920, -0.00001886, -0.00002841, 0.00000652, -0.00001354, -0.00000068, -0.00000427, 0.00000204, 0.00000399, -0.00000648, 0.00000015, 0.00000590, -0.00000483, 0.00000065, -0.00000071, 0.00000026, 0.00000404, -0.00000670, -0.00000538, 0.00000705, -0.00000033, -0.00000184, -0.00000377, 0.00000343, 0.00000008, -0.00000181, -0.00000281, 0.00000403, -0.00000011, -0.00000117, -0.00000091, 0.00000315, 0.00000075, -0.00000332, -0.00000562, 0.00000709, -0.00000169, -0.00000392, -0.00000097, 0.00000529, 0.00000299, -0.00000340, -0.00001099, 0.00000707, -0.00000009, -0.00000432, -0.00000603, 0.00001106, -0.00000433, -0.00000163, -0.00000738, 0.00000717, 0.00000332, -0.00001143, -0.00000694, 0.00000376, -0.00001671, -0.00003557, -0.00009166, 0.00012248, 0.00008508, 0.00013216, -0.00001785, -0.00008506, -0.00000671, -0.00001753, 0.00000508, -0.00001788, -0.00001010, 0.00000263, 0.00000379, -0.00001425, -0.00000260, 0.00000422, 0.00001115, -0.00001460, -0.00000271, 0.00000261, 0.00000217, -0.00000510, 0.00000170, -0.00000067, 0.00000676, -0.00000631, -0.00000183, 0.00000214, -0.00000031, -0.00000493, -0.00000257, 0.00000252, 0.00000580, -0.00000325, 0.00000133, -0.00000226, 0.00000089, -0.00000334, -0.00000174, 0.00000458, 0.00000387, },
		{ -0.00000392, -0.00000376, -0.00000006, 0.00000079, -0.00000066, -0.00000135, -0.00000086, 0.00000258, -0.00000132, -0.00000197, 0.00000199, 0.00000039, 0.00000068, -0.00000641, -0.00000354, 0.00001459, -0.00001057, -0.00000130, 0.00000171, -0.00000834, 0.00000545, -0.00002094, -0.00000250, -0.00000704, -0.00003629, -0.00001618, 0.00003861, 0.00003461, 0.00002855, 0.00004512, 0.00001307, -0.00004111, -0.00000223, -0.00000382, -0.00000973, 0.00000005, -0.00000360, 0.00000646, -0.00000266, -0.00000551, 0.00000605, -0.00000010, -0.00000314, 0.00000104, -0.00000159, 0.00000350, -0.00000030, -0.00001032, 0.00000411, 0.00000342, -0.00000120, -0.00000423, 0.00000066, 0.00000224, -0.00000056, -0.00000411, 0.00000178, 0.00000220, -0.00000062, -0.00000227, 0.00000234, 0.00000152, 0.00000023, -0.00000796, 0.00000329, 0.00000302, -0.00000319, -0.00000432, 0.00000416, 0.00000321, 0.00000285, -0.00001271, 0.00000121, 0.00000342, -0.00000062, -0.00000994, 0.00000681, 0.00000239, -0.00000220, -0.00000707, 0.00000127, 0.00000583, -0.00000077, -0.00001844, 0.00000944, -0.00001603, -0.00000705, -0.00009539, 0.00002384, 0.00012426, 0.00008513, 0.00009854, -0.00009957, -0.00002553, -0.00002106, 0.00000349, -0.00001156, -0.00001225, -0.00000828, 0.00001232, -0.00001263, -0.00000509, -0.00000367, 0.00001592, -0.00000702, -0.00000858, -0.00000066, 0.00000601, -0.00000501, 0.00000060, -0.00000202, 0.00000610, -0.00000115, -0.00000492, 0.00000041, 0.00000276, -0.00000455, -0.00000300, -0.00000195, 0.00000785, -0.00000119, -0.00000011, -0.00000129, 0.00000011, -0.00000156, -0.00000290, 0.00000080, 0.00000735, },
		{ 0.00000310, -0.00000710, -0.00000011, -0.00000035, 0.00000120, -0.00000214, -0.00000084, 0.00000062, 0.00000223, -0.00000403, 0.00000145, 0.00000094, 0.00000088, -0.00000245, -0.00000877, 0.00000924, 0.00000460, -0.00001378, 0.00000816, -0.00000995, 0.00000414, -0.00001058, -0.00001441, 0.00000141, -0.00002263, -0.00003849, 0.00001539, 0.00004533, 0.00002806, 0.00004062, 0.00003893, -0.00002633, -0.00002795, 0.00000815, -0.00001482, -0.00000031, -0.00000491, 0.00000391, 0.00000247, -0.00000634, 0.00000008, 0.00000602, -0.00000457, 0.00000053, -0.00000146, 0.00000116, 0.00000357, -0.00000769, -0.00000440, 0.00000754, -0.00000079, -0.00000217, -0.00000356, 0.00000367, 0.00000019, -0.00000225, -0.00000286, 0.00000440, -0.00000037, -0.00000116, -0.00000095, 0.00000334, 0.00000076, -0.00000412, -0.00000449, 0.00000696, -0.00000196, -0.00000381, -0.00000115, 0.00000539, 0.00000338, -0.00000509, -0.00000951, 0.00000721, -0.00000043, -0.00000464, -0.00000509, 0.00001054, -0.00000399, -0.00000203, -0.00000709, 0.00000752, 0.00000303, -0.00001155, -0.00000642, 0.00000196, -0.00001284, -0.00004441, -0.00007372, 0.00012320, 0.00007095, 0.00012330, -0.00001200, -0.00008142, -0.00000879, -0.00001528, 0.00000389, -0.00001775, -0.00001010, 0.00000383, 0.00000231, -0.00001333, -0.00000331, 0.00000660, 0.00000892, -0.00001441, -0.00000202, 0.00000262, 0.00000216, -0.00000482, 0.00000106, 0.00000034, 0.00000576, -0.00000655, -0.00000110, 0.00000206, -0.00000035, -0.00000504, -0.00000277, 0.00000350, 0.00000517, -0.00000331, 0.00000100, -0.00000204, 0.00000085, -0.00000359, -0.00000119, 0.00000516, },
		{ 0.00000716, -0.00000473, -0.00000350, 0.00000001, 0.00000103, -0.00000085, -0.00000168, -0.00000067, 0.00000313, -0.00000203, -0.00000183, 0.00000230, 0.00000053, 0.00000002, -0.00000734, -0.00000190, 0.00001476, -0.00001200, -0.00000045, -0.00000001, -0.00000539, 0.00000174, -0.00001969, -0.00000197, -0.00000732, -0.00003735, -0.00002121, 0.00004435, 0.00003568, 0.00003436, 0.00004846, 0.00000762, -0.00004595, 0.00000155, -0.00000558, -0.00000882, -0.00000140, -0.00000181, 0.00000607, -0.00000336, -0.00000539, 0.00000607, 0.00000001, -0.00000277, 0.00000010, -0.00000123, 0.00000393, -0.00000149, -0.00001024, 0.00000524, 0.00000313, -0.00000154, -0.00000435, 0.00000102, 0.00000237, -0.00000069, -0.00000457, 0.00000221, 0.00000217, -0.00000081, -0.00000217, 0.00000223, 0.00000197, -0.00000062, -0.00000754, 0.00000391, 0.00000259, -0.00000327, -0.00000416, 0.00000371, 0.00000414, 0.00000149, -0.00001260, 0.00000235, 0.00000311, -0.00000119, -0.00000918, 0.00000671, 0.00000245, -0.00000239, -0.00000692, 0.00000129, 0.00000654, -0.00000233, -0.00001615, 0.00000602, -0.00001163, -0.00001282, -0.00009082, 0.00003908, 0.00011138, 0.00007545, 0.00009443, -0.00008974, -0.00002874, -0.00001856, 0.00000245, -0.00001098, -0.00001349, -0.00000594, 0.00001047, -0.00001156, -0.00000615, -0.00000173, 0.00001551, -0.00000839, -0.00000771, -0.00000020, 0.00000556, -0.00000433, -0.00000017, -0.00000137, 0.00000600, -0.00000202, -0.00000445, 0.00000089, 0.00000233, -0.00000421, -0.00000367, -0.00000105, 0.00000781, -0.00000158, -0.00000029, -0.00000136, 0.00000031, -0.00000182, -0.00000280, 0.00000161, },
		{ 0.00000567, 0.00000224, -0.00000724, 0.00000012, -0.00000025, 0.00000135, -0.00000260, -0.00000087, 0.00000116, 0.00000212, -0.00000457, 0.00000198, 0.00000097, 0.00000093, -0.00000381, -0.00000822, 0.00001079, 0.00000302, -0.00001340, 0.00000715, -0.00000899, 0.00000446, -0.00001391, -0.00001078, -0.00000100, -0.00002046, -0.00004459, 0.00001681, 0.00004935, 0.00003074, 0.00004710, 0.00003745, -0.00003433, -0.00002612, 0.00000872, -0.00001531, -0.00000048, -0.00000515, 0.00000551, 0.00000098, -0.00000612, -0.00000004, 0.00000612, -0.00000416, 0.00000013, -0.00000200, 0.00000207, 0.00000295, -0.00000859, -0.00000326, 0.00000779, -0.00000115, -0.00000248, -0.00000331, 0.00000387, 0.00000032, -0.00000280, -0.00000272, 0.00000465, -0.00000058, -0.00000116, -0.00000098, 0.00000356, 0.00000064, -0.00000475, -0.00000337, 0.00000666, -0.00000214, -0.00000374, -0.00000131, 0.00000558, 0.00000354, -0.00000663, -0.00000788, 0.00000714, -0.00000078, -0.00000475, -0.00000421, 0.00000985, -0.00000350, -0.00000251, -0.00000668, 0.00000788, 0.00000240, -0.00001115, -0.00000645, 0.00000089, -0.00001011, -0.00005187, -0.00005492, 0.00011973, 0.00006007, 0.00011410, -0.00000644, -0.00007786, -0.00001054, -0.00001331, 0.00000279, -0.00001785, -0.00000947, 0.00000438, 0.00000113, -0.00001242, -0.00000387, 0.00000880, 0.00000650, -0.00001388, -0.00000140, 0.00000255, 0.00000219, -0.00000465, 0.00000057, 0.00000125, 0.00000465, -0.00000654, -0.00000045, 0.00000189, -0.00000033, -0.00000521, -0.00000282, 0.00000439, 0.00000448, -0.00000334, 0.00000067, -0.00000177, 0.00000072, -0.00000373, -0.00000065, },
		{ 0.00000246, 0.00000676, -0.00000536, -0.00000327, 0.00000009, 0.00000126, -0.00000107, -0.00000206, -0.00000024, 0.00000345, -0.00000263, -0.00000169, 0.00000256, 0.00000068, -0.00000070, -0.00000823, 0.00000003, 0.00001435, -0.00001292, -0.00000012, -0.00000119, -0.00000271, -0.00000211, -0.00001799, -0.00000180, -0.00000746, -0.00003836, -0.00002647, 0.00004985, 0.00003764, 0.00003999, 0.00005243, -0.00000012, -0.00004893, 0.00000460, -0.00000705, -0.00000796, -0.00000286, 0.00000007, 0.00000549, -0.00000398, -0.00000517, 0.00000592, 0.00000037, -0.00000263, -0.00000078, -0.00000076, 0.00000430, -0.00000283, -0.00000983, 0.00000615, 0.00000284, -0.00000187, -0.00000438, 0.00000128, 0.00000261, -0.00000101, -0.00000486, 0.00000259, 0.00000210, -0.00000098, -0.00000204, 0.00000213, 0.00000239, -0.00000153, -0.00000688, 0.00000427, 0.00000224, -0.00000337, -0.00000395, 0.00000329, 0.00000504, -0.00000009, -0.00001210, 0.00000326, 0.00000276, -0.00000172, -0.00000819, 0.00000629, 0.00000266, -0.00000264, -0.00000671, 0.00000143, 0.00000699, -0.00000378, -0.00001379, 0.00000246, -0.00000694, -0.00001963, -0.00008316, 0.00005063, 0.00009871, 0.00006783, 0.00008931, -0.00007985, -0.00003185, -0.00001601, 0.00000127, -0.00001053, -0.00001442, -0.00000356, 0.00000817, -0.00001001, -0.00000744, 0.00000050, 0.00001456, -0.00000942, -0.00000681, 0.00000019, 0.00000506, -0.00000364, -0.00000098, -0.00000056, 0.00000559, -0.00000267, -0.00000395, 0.00000126, 0.00000190, -0.00000386, -0.00000433, -0.00000002, 0.00000756, -0.00000187, -0.00000055, -0.00000129, 0.00000037, -0.00000198, -0.00000269, },
		{ -0.00000010, 0.00000608, 0.00000132, -0.00000722, 0.00000024, -0.00000008, 0.00000143, -0.00000312, -0.00000078, 0.00000172, 0.00000184, -0.00000489, 0.00000235, 0.00000106, 0.00000092, -0.00000522, -0.00000736, 0.00001209, 0.00000137, -0.00001281, 0.00000578, -0.00000733, 0.00000387, -0.00001639, -0.00000765, -0.00000314, -0.00001859, -0.00005022, 0.00001721, 0.00005447, 0.00003312, 0.00005463, 0.00003421, -0.00004227, -0.00002325, 0.00000834, -0.00001506, -0.00000117, -0.00000497, 0.00000679, -0.00000044, -0.00000582, -0.00000022, 0.00000623, -0.00000369, -0.00000049, -0.00000231, 0.00000295, 0.00000215, -0.00000934, -0.00000202, 0.00000782, -0.00000142, -0.00000276, -0.00000306, 0.00000406, 0.00000040, -0.00000342, -0.00000241, 0.00000476, -0.00000075, -0.00000115, -0.00000100, 0.00000380, 0.00000040, -0.00000516, -0.00000232, 0.00000625, -0.00000224, -0.00000367, -0.00000145, 0.00000586, 0.00000344, -0.00000795, -0.00000617, 0.00000687, -0.00000115, -0.00000462, -0.00000349, 0.00000909, -0.00000294, -0.00000302, -0.00000617, 0.00000823, 0.00000146, -0.00001026, -0.00000704, 0.00000062, -0.00000876, -0.00005739, -0.00003632, 0.00011303, 0.00005203, 0.00010459, -0.00000116, -0.00007441, -0.00001195, -0.00001165, 0.00000179, -0.00001817, -0.00000819, 0.00000418, 0.00000041, -0.00001167, -0.00000415, 0.00001068, 0.00000400, -0.00001306, -0.00000085, 0.00000241, 0.00000227, -0.00000459, 0.00000028, 0.00000200, 0.00000350, -0.00000631, 0.00000011, 0.00000166, -0.00000026, -0.00000545, -0.00000270, 0.00000516, 0.00000376, -0.00000336, 0.00000040, -0.00000150, 0.00000054, -0.00000376, },
		{ -0.00000258, 0.00000332, 0.00000616, -0.00000579, -0.00000309, 0.00000019, 0.00000147, -0.00000134, -0.00000246, 0.00000040, 0.00000350, -0.00000305, -0.00000160, 0.00000280, 0.00000081, -0.00000148, -0.00000903, 0.00000219, 0.00001339, -0.00001330, -0.00000036, -0.00000170, -0.00000055, -0.00000580, -0.00001609, -0.00000182, -0.00000765, -0.00003913, -0.00003210, 0.00005522, 0.00004036, 0.00004572, 0.00005660, -0.00000997, -0.00004983, 0.00000660, -0.00000801, -0.00000733, -0.00000419, 0.00000192, 0.00000479, -0.00000452, -0.00000488, 0.00000560, 0.00000096, -0.00000278, -0.00000153, -0.00000022, 0.00000457, -0.00000428, -0.00000909, 0.00000680, 0.00000261, -0.00000220, -0.00000432, 0.00000146, 0.00000293, -0.00000154, -0.00000495, 0.00000290, 0.00000204, -0.00000115, -0.00000191, 0.00000204, 0.00000275, -0.00000242, -0.00000603, 0.00000438, 0.00000199, -0.00000350, -0.00000371, 0.00000293, 0.00000587, -0.00000178, -0.00001126, 0.00000395, 0.00000236, -0.00000216, -0.00000707, 0.00000559, 0.00000302, -0.00000296, -0.00000645, 0.00000169, 0.00000711, -0.00000495, -0.00001157, -0.00000093, -0.00000239, -0.00002688, -0.00007295, 0.00005847, 0.00008685, 0.00006162, 0.00008362, -0.00007024, -0.00003462, -0.00001361, 0.00000016, -0.00001037, -0.00001485, -0.00000141, 0.00000568, -0.00000814, -0.00000885, 0.00000291, 0.00001312, -0.00001009, -0.00000589, 0.00000051, 0.00000457, -0.00000299, -0.00000176, 0.00000036, 0.00000493, -0.00000310, -0.00000343, 0.00000152, 0.00000148, -0.00000354, -0.00000494, 0.00000110, 0.00000713, -0.00000208, -0.00000086, -0.00000110, 0.00000029, -0.00000202, },
		{ -0.00000372, 0.00000046, 0.00000636, 0.00000040, -0.00000705, 0.00000027, 0.00000015, 0.00000144, -0.00000368, -0.00000052, 0.00000224, 0.00000145, -0.00000501, 0.00000256, 0.00000122, 0.00000082, -0.00000665, -0.00000618, 0.00001304, -0.00000024, -0.00001215, 0.00000424, -0.00000518, 0.00000250, -0.00001802, -0.00000507, -0.00000496, -0.00001706, -0.00005529, 0.00001650, 0.00006074, 0.00003524, 0.00006328, 0.00002878, -0.00004949, -0.00001977, 0.00000723, -0.00001418, -0.00000233, -0.00000439, 0.00000773, -0.00000174, -0.00000545, -0.00000047, 0.00000641, -0.00000323, -0.00000127, -0.00000242, 0.00000379, 0.00000116, -0.00000987, -0.00000076, 0.00000768, -0.00000159, -0.00000301, -0.00000284, 0.00000429, 0.00000038, -0.00000405, -0.00000199, 0.00000477, -0.00000088, -0.00000114, -0.00000102, 0.00000405, 0.00000004, -0.00000535, -0.00000140, 0.00000580, -0.00000232, -0.00000360, -0.00000157, 0.00000621, 0.00000305, -0.00000899, -0.00000448, 0.00000646, -0.00000152, -0.00000426, -0.00000297, 0.00000835, -0.00000234, -0.00000354, -0.00000555, 0.00000851, 0.00000029, -0.00000894, -0.00000812, 0.00000114, -0.00000885, -0.00006057, -0.00001883, 0.00010412, 0.00004632, 0.00009496, 0.00000375, -0.00007100, -0.00001306, -0.00001025, 0.00000083, -0.00001860, -0.00000637, 0.00000324, 0.00000022, -0.00001118, -0.00000409, 0.00001215, 0.00000155, -0.00001202, -0.00000037, 0.00000220, 0.00000238, -0.00000463, 0.00000021, 0.00000253, 0.00000239, -0.00000593, 0.00000057, 0.00000137, -0.00000016, -0.00000575, -0.00000239, 0.00000577, 0.00000304, -0.00000340, 0.00000020, -0.00000127, 0.00000035, },
		{ -0.00000195, -0.00000249, 0.00000419, 0.00000538, -0.00000601, -0.00000299, 0.00000033, 0.00000163, -0.00000166, -0.00000283, 0.00000121, 0.00000329, -0.00000325, -0.00000160, 0.00000305, 0.00000090, -0.00000230, -0.00000970, 0.00000450, 0.00001193, -0.00001315, -0.00000117, -0.00000147, 0.00000090, -0.00000907, -0.00001423, -0.00000186, -0.00000803, -0.00003950, -0.00003828, 0.00006062, 0.00004368, 0.00005188, 0.00006041, -0.00002158, -0.00004858, 0.00000733, -0.00000824, -0.00000711, -0.00000523, 0.00000361, 0.00000404, -0.00000500, -0.00000452, 0.00000515, 0.00000176, -0.00000321, -0.00000209, 0.00000034, 0.00000475, -0.00000580, -0.00000803, 0.00000715, 0.00000246, -0.00000252, -0.00000420, 0.00000156, 0.00000330, -0.00000225, -0.00000483, 0.00000312, 0.00000198, -0.00000131, -0.00000176, 0.00000198, 0.00000302, -0.00000325, -0.00000507, 0.00000429, 0.00000182, -0.00000367, -0.00000344, 0.00000265, 0.00000654, -0.00000351, -0.00001013, 0.00000442, 0.00000192, -0.00000246, -0.00000592, 0.00000470, 0.00000350, -0.00000333, -0.00000614, 0.00000209, 0.00000682, -0.00000573, -0.00000968, -0.00000387, 0.00000162, -0.00003398, -0.00006089, 0.00006286, 0.00007619, 0.00005635, 0.00007773, -0.00006115, -0.00003686, -0.00001155, -0.00000077, -0.00001064, -0.00001460, 0.00000028, 0.00000325, -0.00000614, -0.00001025, 0.00000539, 0.00001129, -0.00001043, -0.00000497, 0.00000074, 0.00000411, -0.00000243, -0.00000243, 0.00000128, 0.00000409, -0.00000329, -0.00000293, 0.00000169, 0.00000110, -0.00000327, -0.00000545, 0.00000226, 0.00000654, -0.00000222, -0.00000119, -0.00000082, 0.00000010, },
		{ 0.00000019, -0.00000365, 0.00000105, 0.00000649, -0.00000047, -0.00000678, 0.00000021, 0.00000042, 0.00000136, -0.00000427, -0.00000007, 0.00000262, 0.00000101, -0.00000499, 0.00000264, 0.00000143, 0.00000064, -0.00000808, -0.00000468, 0.00001357, -0.00000167, -0.00001155, 0.00000274, -0.00000280, 0.00000054, -0.00001890, -0.00000302, -0.00000649, -0.00001585, -0.00005981, 0.00001470, 0.00006812, 0.00003726, 0.00007293, 0.00002080, -0.00005540, -0.00001614, 0.00000565, -0.00001283, -0.00000385, -0.00000347, 0.00000833, -0.00000291, -0.00000503, -0.00000082, 0.00000670, -0.00000287, -0.00000212, -0.00000236, 0.00000457, -0.00000001, -0.00001013, 0.00000044, 0.00000739, -0.00000168, -0.00000322, -0.00000265, 0.00000458, 0.00000022, -0.00000460, -0.00000148, 0.00000468, -0.00000097, -0.00000114, -0.00000103, 0.00000429, -0.00000042, -0.00000530, -0.00000066, 0.00000534, -0.00000240, -0.00000351, -0.00000165, 0.00000661, 0.00000240, -0.00000971, -0.00000285, 0.00000592, -0.00000186, -0.00000369, -0.00000270, 0.00000768, -0.00000177, -0.00000407, -0.00000481, 0.00000867, -0.00000101, -0.00000732, -0.00000958, 0.00000231, -0.00001029, -0.00006125, -0.00000317, 0.00009393, 0.00004233, 0.00008548, 0.00000816, -0.00006753, -0.00001398, -0.00000901, -0.00000021, -0.00001900, -0.00000415, 0.00000165, 0.00000059, -0.00001100, -0.00000366, 0.00001314, -0.00000076, -0.00001080, 0.00000005, 0.00000195, 0.00000251, -0.00000474, 0.00000033, 0.00000280, 0.00000140, -0.00000543, 0.00000093, 0.00000106, -0.00000002, -0.00000609, -0.00000190, 0.00000620, 0.00000234, -0.00000347, 0.00000010, -0.00000112, },
		{ -0.00000018, -0.00000177, -0.00000241, 0.00000503, 0.00000444, -0.00000601, -0.00000298, 0.00000053, 0.00000172, -0.00000203, -0.00000312, 0.00000215, 0.00000281, -0.00000321, -0.00000173, 0.00000334, 0.00000093, -0.00000316, -0.00001019, 0.00000687, 0.00001003, -0.00001253, -0.00000248, -0.00000050, 0.00000153, -0.00001170, -0.00001261, -0.00000176, -0.00000875, -0.00003930, -0.00004514, 0.00006615, 0.00004748, 0.00005883, 0.00006319, -0.00003440, -0.00004531, 0.00000669, -0.00000763, -0.00000741, -0.00000587, 0.00000504, 0.00000330, -0.00000542, -0.00000413, 0.00000461, 0.00000270, -0.00000392, -0.00000244, 0.00000090, 0.00000479, -0.00000733, -0.00000671, 0.00000722, 0.00000242, -0.00000284, -0.00000404, 0.00000164, 0.00000369, -0.00000312, -0.00000450, 0.00000321, 0.00000194, -0.00000146, -0.00000162, 0.00000194, 0.00000318, -0.00000396, -0.00000408, 0.00000404, 0.00000171, -0.00000385, -0.00000315, 0.00000248, 0.00000700, -0.00000518, -0.00000880, 0.00000469, 0.00000143, -0.00000256, -0.00000484, 0.00000371, 0.00000404, -0.00000372, -0.00000580, 0.00000264, 0.00000613, -0.00000605, -0.00000827, -0.00000616, 0.00000474, -0.00004030, -0.00004777, 0.00006423, 0.00006687, 0.00005170, 0.00007186, -0.00005275, -0.00003843, -0.00000991, -0.00000141, -0.00001139, -0.00001356, 0.00000131, 0.00000112, -0.00000419, -0.00001152, 0.00000781, 0.00000917, -0.00001045, -0.00000405, 0.00000088, 0.00000369, -0.00000201, -0.00000293, 0.00000212, 0.00000313, -0.00000329, -0.00000246, 0.00000175, 0.00000078, -0.00000308, -0.00000582, 0.00000339, 0.00000583, -0.00000232, -0.00000151, -0.00000050, },
		},
		{ { 0.00000330, 0.00000147, -0.00000995, -0.00000229, 0.00000391, 0.00000084, 0.00000707, -0.00000056, -0.00000136, -0.00000226, -0.00000236, -0.00000289, 0.00000762, -0.00000322, 0.00000063, -0.00000438, -0.00000160, 0.00000105, 0.00000389, -0.00000139, -0.00000191, -0.00000459, -0.00000028, -0.00000673, -0.00000878, -0.00000959, -0.00005913, 0.00000964, 0.00008321, 0.00003839, 0.00008903, 0.00000800, -0.00005288, -0.00001359, -0.00001041, -0.00000447, -0.00000338, -0.00000035, -0.00001576, 0.00000073, -0.00000440, -0.00000653, -0.00000185, 0.00000301, -0.00001354, 0.00000227, -0.00000974, 0.00000211, 0.00000114, -0.00000217, 0.00001215, -0.00000454, -0.00000562, -0.00000107, 0.00000349, 0.00000191, 0.00000230, 0.00000122, -0.00000426, 0.00000058, -0.00000488, 0.00000196, -0.00000118, 0.00000013, -0.00000025, 0.00000013, -0.00000118, 0.00000196, -0.00000488, 0.00000058, -0.00000426, 0.00000122, 0.00000230, 0.00000191, 0.00000349, -0.00000107, -0.00000562, -0.00000454, 0.00001215, -0.00000217, 0.00000114, 0.00000211, -0.00000974, 0.00000227, -0.00001354, 0.00000301, -0.00000185, -0.00000653, -0.00000440, 0.00000073, -0.00001576, -0.00000035, -0.00000338, -0.00000447, -0.00001041, -0.00001359, -0.00005288, 0.00000800, 0.00008903, 0.00003839, 0.00008321, 0.00000964, -0.00005913, -0.00000959, -0.00000878, -0.00000673, -0.00000028, -0.00000459, -0.00000191, -0.00000139, 0.00000389, 0.00000105, -0.00000160, -0.00000438, 0.00000063, -0.00000322, 0.00000762, -0.00000289, -0.00000236, -0.00000226, -0.00000136, -0.00000056, 0.00000707, 0.00000084, 0.00000391, -0.00000229, -0.00000995, 0.00000147, },
		{ 0.00000324, 0.00000334, -0.00000299, -0.00001039, 0.00000376, 0.00000070, 0.00000498, 0.00000372, -0.00000123, -0.00000209, -0.00000135, -0.00000464, 0.00000293, 0.00000378, -0.00000300, 0.00000007, -0.00000537, 0.00000086, 0.00000175, 0.00000301, -0.00000327, -0.00000215, -0.00000381, -0.00000039, -0.00001206, -0.00000316, -0.00003138, -0.00005371, 0.00006431, 0.00006345, 0.00005958, 0.00007760, -0.00004830, -0.00003082, -0.00001239, -0.00000570, -0.00000672, 0.00000285, -0.00001159, -0.00000783, 0.00000030, -0.00000583, -0.00000697, 0.00000426, -0.00000599, -0.00000792, -0.00000045, -0.00000868, 0.00000721, -0.00000535, 0.00000649, 0.00000686, -0.00000818, -0.00000289, 0.00000006, 0.00000479, 0.00000032, 0.00000407, -0.00000346, -0.00000084, -0.00000227, -0.00000211, 0.00000115, -0.00000125, 0.00000036, 0.00000007, -0.00000126, 0.00000138, -0.00000112, -0.00000362, -0.00000016, -0.00000413, 0.00000425, 0.00000034, 0.00000371, 0.00000130, -0.00000209, -0.00000794, 0.00000341, 0.00000984, -0.00000638, 0.00000768, -0.00000835, -0.00000184, -0.00000448, -0.00000954, 0.00000517, -0.00000608, -0.00000599, -0.00000065, -0.00000682, -0.00001233, 0.00000359, -0.00000781, -0.00000386, -0.00001241, -0.00002796, -0.00004489, 0.00006616, 0.00006014, 0.00005077, 0.00007025, -0.00003721, -0.00004005, -0.00000178, -0.00001236, -0.00000002, -0.00000414, -0.00000252, -0.00000280, 0.00000208, 0.00000240, 0.00000105, -0.00000495, -0.00000030, -0.00000294, 0.00000213, 0.00000496, -0.00000484, -0.00000130, -0.00000228, -0.00000170, 0.00000320, 0.00000575, 0.00000000, 0.00000452, -0.00000851, -0.00000524, },
		{ -0.00000013, 0.00000335, 0.00000256, -0.00000870, -0.00000508, 0.00000486, 0.00000034, 0.00000714, -0.00000031, -0.00000063, -0.00000265, -0.00000165, -0.00000425, 0.00000712, -0.00000179, 0.00000031, -0.00000397, -0.00000228, 0.00000093, 0.00000351, -0.00000031, -0.00000303, -0.00000311, -0.00000135, -0.00000548, -0.00001117, -0.00000513, -0.00005751, -0.00001222, 0.00009240, 0.00004353, 0.00009409, 0.00001561, -0.00006046, -0.00001233, -0.00001305, -0.00000374, -0.00000388, 0.00000023, -0.00001637, 0.00000087, -0.00000389, -0.00000631, -0.00000394, 0.00000500, -0.00001264, -0.00000065, -0.00000722, -0.00000003, 0.00000302, -0.00000427, 0.00001165, -0.00000149, -0.00000704, -0.00000170, 0.00000263, 0.00000354, 0.00000120, 0.00000227, -0.00000486, 0.00000117, -0.00000450, 0.00000108, -0.00000122, 0.00000029, -0.00000021, -0.00000003, -0.00000132, 0.00000292, -0.00000494, -0.00000033, -0.00000346, 0.00000004, 0.00000337, 0.00000061, 0.00000381, -0.00000026, -0.00000432, -0.00000663, 0.00001112, 0.00000104, -0.00000144, 0.00000481, -0.00001228, 0.00000437, -0.00001314, 0.00000025, 0.00000019, -0.00000648, -0.00000493, 0.00000053, -0.00001510, -0.00000104, -0.00000267, -0.00000552, -0.00000752, -0.00001512, -0.00004576, 0.00000185, 0.00008316, 0.00003557, 0.00007237, 0.00002699, -0.00005509, -0.00001644, -0.00000559, -0.00000847, 0.00000120, -0.00000617, -0.00000092, -0.00000223, 0.00000400, 0.00000127, -0.00000092, -0.00000476, 0.00000102, -0.00000450, 0.00000755, -0.00000117, -0.00000311, -0.00000184, -0.00000190, -0.00000106, 0.00000698, 0.00000153, 0.00000281, 0.00000011, -0.00001029, },
		{ -0.00000722, 0.00000282, 0.00000327, -0.00000075, -0.00001152, 0.00000225, 0.00000170, 0.00000417, 0.00000420, -0.00000101, -0.00000155, -0.00000163, -0.00000412, 0.00000083, 0.00000501, -0.00000270, 0.00000028, -0.00000561, 0.00000051, 0.00000132, 0.00000347, -0.00000307, -0.00000264, -0.00000252, -0.00000164, -0.00001066, -0.00000622, -0.00002152, -0.00006666, 0.00005143, 0.00007872, 0.00005892, 0.00008970, -0.00005111, -0.00003427, -0.00001236, -0.00000743, -0.00000593, 0.00000244, -0.00001114, -0.00000863, 0.00000099, -0.00000544, -0.00000766, 0.00000285, -0.00000251, -0.00001026, -0.00000057, -0.00000780, 0.00000596, -0.00000364, 0.00000289, 0.00000962, -0.00000749, -0.00000396, -0.00000110, 0.00000539, 0.00000087, 0.00000348, -0.00000253, -0.00000169, -0.00000097, -0.00000280, 0.00000085, -0.00000137, 0.00000069, -0.00000015, -0.00000134, 0.00000141, 0.00000017, -0.00000492, 0.00000024, -0.00000446, 0.00000397, 0.00000085, 0.00000232, 0.00000248, -0.00000146, -0.00000707, -0.00000016, 0.00001231, -0.00000623, 0.00000694, -0.00000654, -0.00000466, -0.00000032, -0.00001279, 0.00000547, -0.00000507, -0.00000583, -0.00000186, -0.00000557, -0.00001333, 0.00000459, -0.00000907, -0.00000209, -0.00001226, -0.00002575, -0.00004099, 0.00005555, 0.00006020, 0.00004089, 0.00007070, -0.00001975, -0.00004567, -0.00000293, -0.00001114, -0.00000079, -0.00000340, -0.00000376, -0.00000173, 0.00000079, 0.00000316, 0.00000112, -0.00000434, -0.00000085, -0.00000252, 0.00000014, 0.00000677, -0.00000462, -0.00000149, -0.00000218, -0.00000225, 0.00000254, 0.00000649, -0.00000038, 0.00000462, -0.00000619, },
		{ -0.00000972, -0.00000212, 0.00000347, 0.00000311, -0.00000669, -0.00000798, 0.00000545, 0.00000009, 0.00000717, -0.00000026, 0.00000013, -0.00000288, -0.00000107, -0.00000520, 0.00000616, -0.00000036, 0.00000015, -0.00000363, -0.00000289, 0.00000079, 0.00000308, 0.00000067, -0.00000384, -0.00000226, -0.00000144, -0.00000537, -0.00001198, -0.00000386, -0.00005038, -0.00003662, 0.00009759, 0.00005210, 0.00009790, 0.00002502, -0.00006870, -0.00001110, -0.00001576, -0.00000299, -0.00000451, 0.00000106, -0.00001725, 0.00000120, -0.00000363, -0.00000569, -0.00000608, 0.00000628, -0.00001082, -0.00000373, -0.00000541, -0.00000116, 0.00000391, -0.00000510, 0.00000974, 0.00000218, -0.00000829, -0.00000226, 0.00000136, 0.00000529, 0.00000027, 0.00000305, -0.00000519, 0.00000142, -0.00000389, 0.00000033, -0.00000139, 0.00000038, -0.00000009, -0.00000017, -0.00000164, 0.00000383, -0.00000459, -0.00000156, -0.00000254, -0.00000114, 0.00000423, -0.00000021, 0.00000354, 0.00000079, -0.00000331, -0.00000764, 0.00000875, 0.00000493, -0.00000419, 0.00000738, -0.00001405, 0.00000502, -0.00001121, -0.00000326, 0.00000218, -0.00000633, -0.00000521, 0.00000000, -0.00001412, -0.00000213, -0.00000146, -0.00000714, -0.00000419, -0.00001704, -0.00003899, -0.00000313, 0.00007693, 0.00003404, 0.00006163, 0.00003899, -0.00004631, -0.00002423, -0.00000270, -0.00000992, 0.00000242, -0.00000729, -0.00000048, -0.00000255, 0.00000367, 0.00000169, -0.00000033, -0.00000505, 0.00000135, -0.00000546, 0.00000684, 0.00000083, -0.00000378, -0.00000154, -0.00000215, -0.00000181, 0.00000678, 0.00000235, 0.00000174, 0.00000193, },
		{ -0.00000374, -0.00000864, 0.00000197, 0.00000320, 0.00000117, -0.00001170, -0.00000002, 0.00000293, 0.00000331, 0.00000477, -0.00000114, -0.00000067, -0.00000208, -0.00000336, -0.00000120, 0.00000577, -0.00000213, 0.00000043, -0.00000571, 0.00000004, 0.00000109, 0.00000349, -0.00000232, -0.00000373, -0.00000071, -0.00000318, -0.00000893, -0.00000972, -0.00001255, -0.00007381, 0.00003105, 0.00009551, 0.00005908, 0.00010198, -0.00005297, -0.00003843, -0.00001223, -0.00000917, -0.00000530, 0.00000221, -0.00001078, -0.00000942, 0.00000164, -0.00000504, -0.00000798, 0.00000101, 0.00000067, -0.00001140, -0.00000194, -0.00000626, 0.00000451, -0.00000184, -0.00000035, 0.00001122, -0.00000573, -0.00000532, -0.00000207, 0.00000539, 0.00000201, 0.00000253, -0.00000141, -0.00000264, 0.00000022, -0.00000324, 0.00000058, -0.00000167, 0.00000105, -0.00000031, -0.00000143, 0.00000119, 0.00000174, -0.00000604, 0.00000029, -0.00000444, 0.00000330, 0.00000170, 0.00000086, 0.00000338, -0.00000082, -0.00000592, -0.00000330, 0.00001343, -0.00000464, 0.00000487, -0.00000337, -0.00000847, 0.00000392, -0.00001521, 0.00000495, -0.00000390, -0.00000544, -0.00000320, -0.00000429, -0.00001437, 0.00000557, -0.00001018, -0.00000069, -0.00001161, -0.00002441, -0.00003653, 0.00004567, 0.00005976, 0.00003337, 0.00006753, -0.00000370, -0.00004703, -0.00000677, -0.00000851, -0.00000255, -0.00000177, -0.00000559, -0.00000034, -0.00000062, 0.00000389, 0.00000115, -0.00000358, -0.00000156, -0.00000176, -0.00000203, 0.00000821, -0.00000397, -0.00000189, -0.00000193, -0.00000270, 0.00000162, 0.00000719, -0.00000047, 0.00000424, },
		{ 0.00000310, -0.00000835, -0.00000430, 0.00000355, 0.00000322, -0.00000417, -0.00001065, 0.00000548, 0.00000015, 0.00000713, -0.00000027, 0.00000072, -0.00000281, -0.00000070, -0.00000572, 0.00000484, 0.00000094, 0.00000023, -0.00000338, -0.00000334, 0.00000051, 0.00000279, 0.00000128, -0.00000400, -0.00000239, -0.00000026, -0.00000663, -0.00001098, -0.00000575, -0.00003898, -0.00006073, 0.00009624, 0.00006493, 0.00010029, 0.00003641, -0.00007772, -0.00000970, -0.00001881, -0.00000195, -0.00000557, 0.00000236, -0.00001856, 0.00000188, -0.00000372, -0.00000465, -0.00000821, 0.00000693, -0.00000850, -0.00000637, -0.00000475, -0.00000111, 0.00000383, -0.00000479, 0.00000677, 0.00000598, -0.00000906, -0.00000288, -0.00000015, 0.00000691, -0.00000030, 0.00000347, -0.00000522, 0.00000138, -0.00000314, -0.00000026, -0.00000162, 0.00000034, 0.00000010, -0.00000025, -0.00000209, 0.00000457, -0.00000376, -0.00000303, -0.00000164, -0.00000218, 0.00000475, -0.00000047, 0.00000272, 0.00000202, -0.00000261, -0.00000765, 0.00000543, 0.00000888, -0.00000642, 0.00000910, -0.00001439, 0.00000384, -0.00000775, -0.00000730, 0.00000402, -0.00000619, -0.00000506, -0.00000100, -0.00001265, -0.00000377, 0.00000034, -0.00000936, -0.00000043, -0.00001928, -0.00003263, -0.00000709, 0.00007059, 0.00003314, 0.00005201, 0.00004584, -0.00003450, -0.00003121, -0.00000123, -0.00001035, 0.00000282, -0.00000753, -0.00000085, -0.00000221, 0.00000286, 0.00000232, 0.00000013, -0.00000515, 0.00000151, -0.00000597, 0.00000550, 0.00000300, -0.00000425, -0.00000142, -0.00000209, -0.00000268, 0.00000640, 0.00000330, 0.00000082, },
		{ 0.00000354, -0.00000147, -0.00000930, 0.00000061, 0.00000323, 0.00000257, -0.00001082, -0.00000295, 0.00000426, 0.00000244, 0.00000547, -0.00000164, 0.00000046, -0.00000259, -0.00000247, -0.00000308, 0.00000608, -0.00000139, 0.00000060, -0.00000576, -0.00000047, 0.00000096, 0.00000326, -0.00000132, -0.00000497, 0.00000098, -0.00000422, -0.00000783, -0.00001231, -0.00000636, -0.00007374, 0.00000387, 0.00011174, 0.00006149, 0.00011373, -0.00005335, -0.00004366, -0.00001172, -0.00001124, -0.00000450, 0.00000178, -0.00001014, -0.00001054, 0.00000256, -0.00000494, -0.00000772, -0.00000128, 0.00000347, -0.00001147, -0.00000406, -0.00000473, 0.00000347, -0.00000045, -0.00000274, 0.00001138, -0.00000292, -0.00000688, -0.00000283, 0.00000477, 0.00000364, 0.00000135, -0.00000022, -0.00000359, 0.00000126, -0.00000349, 0.00000040, -0.00000211, 0.00000139, -0.00000042, -0.00000147, 0.00000066, 0.00000347, -0.00000684, -0.00000008, -0.00000407, 0.00000233, 0.00000271, -0.00000045, 0.00000385, -0.00000002, -0.00000482, -0.00000558, 0.00001296, -0.00000162, 0.00000167, 0.00000071, -0.00001253, 0.00000745, -0.00001629, 0.00000343, -0.00000250, -0.00000500, -0.00000436, -0.00000327, -0.00001511, 0.00000615, -0.00001076, -0.00000004, -0.00001013, -0.00002417, -0.00003138, 0.00003634, 0.00005906, 0.00002752, 0.00006258, 0.00000922, -0.00004384, -0.00001274, -0.00000511, -0.00000478, 0.00000024, -0.00000755, 0.00000098, -0.00000185, 0.00000439, 0.00000123, -0.00000273, -0.00000239, -0.00000077, -0.00000420, 0.00000910, -0.00000286, -0.00000239, -0.00000165, -0.00000291, 0.00000042, 0.00000781, -0.00000027, },
		{ 0.00000012, 0.00000363, -0.00000641, -0.00000638, 0.00000344, 0.00000302, -0.00000147, -0.00001271, 0.00000473, 0.00000058, 0.00000695, -0.00000018, 0.00000096, -0.00000235, -0.00000057, -0.00000584, 0.00000326, 0.00000201, 0.00000053, -0.00000324, -0.00000361, 0.00000005, 0.00000276, 0.00000136, -0.00000338, -0.00000358, 0.00000213, -0.00000905, -0.00000852, -0.00000995, -0.00002542, -0.00008129, 0.00008616, 0.00008217, 0.00010159, 0.00004964, -0.00008739, -0.00000813, -0.00002227, -0.00000054, -0.00000709, 0.00000415, -0.00002028, 0.00000283, -0.00000409, -0.00000334, -0.00001016, 0.00000697, -0.00000597, -0.00000820, -0.00000536, -0.00000005, 0.00000308, -0.00000371, 0.00000327, 0.00000936, -0.00000902, -0.00000370, -0.00000171, 0.00000820, -0.00000036, 0.00000348, -0.00000494, 0.00000106, -0.00000229, -0.00000073, -0.00000180, 0.00000011, 0.00000036, -0.00000029, -0.00000258, 0.00000500, -0.00000241, -0.00000464, -0.00000089, -0.00000294, 0.00000483, -0.00000018, 0.00000147, 0.00000331, -0.00000214, -0.00000690, 0.00000175, 0.00001220, -0.00000752, 0.00000940, -0.00001289, 0.00000079, -0.00000308, -0.00001147, 0.00000553, -0.00000607, -0.00000447, -0.00000246, -0.00001076, -0.00000585, 0.00000259, -0.00001198, 0.00000350, -0.00002152, -0.00002694, -0.00000996, 0.00006416, 0.00003260, 0.00004376, 0.00004855, -0.00002175, -0.00003575, -0.00000199, -0.00000933, 0.00000212, -0.00000672, -0.00000209, -0.00000122, 0.00000164, 0.00000311, 0.00000044, -0.00000503, 0.00000141, -0.00000594, 0.00000356, 0.00000518, -0.00000441, -0.00000153, -0.00000178, -0.00000354, 0.00000571, 0.00000434, },
		{ 0.00000018, 0.00000271, 0.00000040, -0.00000912, -0.00000121, 0.00000340, 0.00000335, -0.00000894, -0.00000629, 0.00000547, 0.00000163, 0.00000629, -0.00000241, 0.00000166, -0.00000302, -0.00000156, -0.00000470, 0.00000598, -0.00000061, 0.00000088, -0.00000584, -0.00000090, 0.00000078, 0.00000302, -0.00000046, -0.00000583, 0.00000189, -0.00000404, -0.00000818, -0.00001295, -0.00000414, -0.00006624, -0.00002805, 0.00012448, 0.00006779, 0.00012418, -0.00005164, -0.00005040, -0.00001042, -0.00001407, -0.00000311, 0.00000073, -0.00000880, -0.00001235, 0.00000405, -0.00000541, -0.00000673, -0.00000400, 0.00000588, -0.00001078, -0.00000633, -0.00000384, 0.00000330, 0.00000017, -0.00000402, 0.00001006, 0.00000072, -0.00000843, -0.00000343, 0.00000357, 0.00000561, 0.00000013, 0.00000092, -0.00000446, 0.00000213, -0.00000360, 0.00000035, -0.00000264, 0.00000165, -0.00000047, -0.00000141, -0.00000015, 0.00000521, -0.00000717, -0.00000091, -0.00000342, 0.00000119, 0.00000367, -0.00000139, 0.00000376, 0.00000105, -0.00000400, -0.00000676, 0.00001093, 0.00000249, -0.00000212, 0.00000494, -0.00001594, 0.00000950, -0.00001562, 0.00000080, -0.00000079, -0.00000472, -0.00000507, -0.00000281, -0.00001524, 0.00000601, -0.00001048, -0.00000042, -0.00000761, -0.00002513, -0.00002555, 0.00002743, 0.00005832, 0.00002268, 0.00005724, 0.00001818, -0.00003673, -0.00001966, -0.00000188, -0.00000677, 0.00000204, -0.00000913, 0.00000183, -0.00000262, 0.00000450, 0.00000145, -0.00000187, -0.00000324, 0.00000031, -0.00000616, 0.00000933, -0.00000133, -0.00000292, -0.00000147, -0.00000279, -0.00000102, 0.00000827, },
		{ 0.00000544, -0.00000033, 0.00000363, -0.00000418, -0.00000810, 0.00000299, 0.00000270, 0.00000108, -0.00001384, 0.00000311, 0.00000141, 0.00000656, 0.00000018, 0.00000069, -0.00000148, -0.00000067, -0.00000560, 0.00000154, 0.00000282, 0.00000103, -0.00000315, -0.00000375, -0.00000058, 0.00000298, 0.00000094, -0.00000210, -0.00000554, 0.00000526, -0.00001197, -0.00000551, -0.00001502, -0.00001229, -0.00009525, 0.00006612, 0.00010292, 0.00010286, 0.00006414, -0.00009729, -0.00000668, -0.00002593, 0.00000101, -0.00000886, 0.00000617, -0.00002211, 0.00000374, -0.00000442, -0.00000207, -0.00001166, 0.00000637, -0.00000343, -0.00000909, -0.00000707, 0.00000155, 0.00000215, -0.00000236, -0.00000017, 0.00001180, -0.00000796, -0.00000478, -0.00000319, 0.00000896, 0.00000017, 0.00000307, -0.00000437, 0.00000050, -0.00000136, -0.00000113, -0.00000184, -0.00000031, 0.00000067, -0.00000028, -0.00000302, 0.00000501, -0.00000060, -0.00000623, -0.00000041, -0.00000335, 0.00000445, 0.00000056, -0.00000002, 0.00000447, -0.00000173, -0.00000578, -0.00000171, 0.00001427, -0.00000701, 0.00000796, -0.00000952, -0.00000375, 0.00000214, -0.00001518, 0.00000639, -0.00000585, -0.00000359, -0.00000414, -0.00000870, -0.00000809, 0.00000493, -0.00001458, 0.00000716, -0.00002329, -0.00002231, -0.00001153, 0.00005748, 0.00003247, 0.00003661, 0.00004859, -0.00001008, -0.00003670, -0.00000525, -0.00000689, 0.00000039, -0.00000499, -0.00000400, 0.00000018, 0.00000020, 0.00000393, 0.00000066, -0.00000467, 0.00000101, -0.00000536, 0.00000115, 0.00000719, -0.00000420, -0.00000182, -0.00000135, -0.00000418, 0.00000464, },
		{ 0.00000844, 0.00000088, 0.00000188, 0.00000173, -0.00000811, -0.00000335, 0.00000365, 0.00000351, -0.00000626, -0.00000971, 0.00000632, 0.00000100, 0.00000715, -0.00000329, 0.00000270, -0.00000321, -0.00000074, -0.00000602, 0.00000553, 0.00000011, 0.00000134, -0.00000603, -0.00000115, 0.00000040, 0.00000302, -0.00000010, -0.00000588, 0.00000155, -0.00000217, -0.00001041, -0.00001115, -0.00000617, -0.00005243, -0.00006144, 0.00013040, 0.00007947, 0.00013280, -0.00004738, -0.00005898, -0.00000796, -0.00001803, -0.00000076, -0.00000129, -0.00000647, -0.00001509, 0.00000629, -0.00000658, -0.00000497, -0.00000707, 0.00000791, -0.00000962, -0.00000821, -0.00000405, 0.00000418, -0.00000002, -0.00000419, 0.00000750, 0.00000477, -0.00000967, -0.00000396, 0.00000193, 0.00000773, -0.00000097, 0.00000188, -0.00000519, 0.00000280, -0.00000359, 0.00000041, -0.00000317, 0.00000176, -0.00000047, -0.00000123, -0.00000117, 0.00000679, -0.00000689, -0.00000214, -0.00000261, 0.00000005, 0.00000437, -0.00000181, 0.00000305, 0.00000241, -0.00000357, -0.00000683, 0.00000767, 0.00000708, -0.00000575, 0.00000847, -0.00001784, 0.00000949, -0.00001303, -0.00000283, 0.00000117, -0.00000474, -0.00000511, -0.00000311, -0.00001457, 0.00000499, -0.00000923, -0.00000189, -0.00000406, -0.00002716, -0.00001923, 0.00001898, 0.00005759, 0.00001846, 0.00005228, 0.00002328, -0.00002710, -0.00002600, 0.00000011, -0.00000781, 0.00000309, -0.00000989, 0.00000191, -0.00000275, 0.00000413, 0.00000184, -0.00000107, -0.00000401, 0.00000132, -0.00000768, 0.00000879, 0.00000054, -0.00000334, -0.00000148, -0.00000232, -0.00000258, },
		{ 0.00000315, 0.00000654, -0.00000050, 0.00000323, -0.00000193, -0.00000921, 0.00000208, 0.00000245, 0.00000315, -0.00001384, 0.00000061, 0.00000256, 0.00000594, 0.00000089, -0.00000011, -0.00000025, -0.00000092, -0.00000507, -0.00000025, 0.00000335, 0.00000163, -0.00000303, -0.00000381, -0.00000128, 0.00000333, 0.00000021, -0.00000048, -0.00000779, 0.00000839, -0.00001445, -0.00000310, -0.00001928, -0.00000208, -0.00010036, 0.00003629, 0.00012499, 0.00010584, 0.00007898, -0.00010666, -0.00000591, -0.00002929, 0.00000222, -0.00001039, 0.00000794, -0.00002355, 0.00000415, -0.00000430, -0.00000127, -0.00001240, 0.00000502, -0.00000093, -0.00000916, -0.00000941, 0.00000308, 0.00000160, -0.00000123, -0.00000302, 0.00001293, -0.00000580, -0.00000609, -0.00000449, 0.00000908, 0.00000131, 0.00000228, -0.00000356, -0.00000025, -0.00000037, -0.00000153, -0.00000167, -0.00000092, 0.00000098, -0.00000026, -0.00000331, 0.00000455, 0.00000158, -0.00000761, -0.00000034, -0.00000336, 0.00000366, 0.00000161, -0.00000150, 0.00000528, -0.00000118, -0.00000464, -0.00000439, 0.00001468, -0.00000475, 0.00000487, -0.00000468, -0.00000902, 0.00000705, -0.00001777, 0.00000626, -0.00000533, -0.00000267, -0.00000568, -0.00000687, -0.00001005, 0.00000689, -0.00001667, 0.00001003, -0.00002411, -0.00001911, -0.00001157, 0.00005030, 0.00003299, 0.00003006, 0.00004742, -0.00000100, -0.00003368, -0.00001062, -0.00000353, -0.00000192, -0.00000275, -0.00000617, 0.00000164, -0.00000119, 0.00000461, 0.00000085, -0.00000410, 0.00000032, -0.00000429, -0.00000152, 0.00000882, -0.00000356, -0.00000226, -0.00000094, -0.00000446, },
		{ -0.00000407, 0.00000820, 0.00000180, 0.00000116, 0.00000247, -0.00000643, -0.00000559, 0.00000388, 0.00000319, -0.00000309, -0.00001280, 0.00000654, 0.00000067, 0.00000791, -0.00000406, 0.00000335, -0.00000303, -0.00000007, -0.00000699, 0.00000477, 0.00000071, 0.00000200, -0.00000634, -0.00000117, -0.00000028, 0.00000341, -0.00000044, -0.00000489, -0.00000021, 0.00000141, -0.00001440, -0.00000719, -0.00001168, -0.00003464, -0.00009228, 0.00012628, 0.00009732, 0.00013969, -0.00004054, -0.00006947, -0.00000424, -0.00002328, 0.00000266, -0.00000437, -0.00000310, -0.00001873, 0.00000921, -0.00000834, -0.00000264, -0.00001025, 0.00000949, -0.00000824, -0.00000935, -0.00000550, 0.00000601, -0.00000083, -0.00000350, 0.00000413, 0.00000872, -0.00001030, -0.00000454, 0.00000001, 0.00000975, -0.00000175, 0.00000257, -0.00000570, 0.00000323, -0.00000346, 0.00000051, -0.00000357, 0.00000168, -0.00000042, -0.00000095, -0.00000228, 0.00000800, -0.00000595, -0.00000371, -0.00000179, -0.00000092, 0.00000469, -0.00000165, 0.00000180, 0.00000397, -0.00000347, -0.00000601, 0.00000372, 0.00001139, -0.00000845, 0.00001050, -0.00001761, 0.00000719, -0.00000872, -0.00000712, 0.00000320, -0.00000506, -0.00000447, -0.00000412, -0.00001316, 0.00000319, -0.00000716, -0.00000420, 0.00000021, -0.00002985, -0.00001282, 0.00001126, 0.00005669, 0.00001484, 0.00004784, 0.00002538, -0.00001676, -0.00003023, 0.00000004, -0.00000742, 0.00000303, -0.00000961, 0.00000112, -0.00000222, 0.00000333, 0.00000240, -0.00000036, -0.00000461, 0.00000210, -0.00000860, 0.00000747, 0.00000262, -0.00000356, -0.00000172, -0.00000159, },
		{ -0.00000429, 0.00000130, 0.00000752, -0.00000039, 0.00000258, 0.00000006, -0.00000956, 0.00000068, 0.00000236, 0.00000455, -0.00001262, -0.00000263, 0.00000389, 0.00000514, 0.00000197, -0.00000138, 0.00000115, -0.00000120, -0.00000432, -0.00000203, 0.00000361, 0.00000225, -0.00000279, -0.00000393, -0.00000190, 0.00000362, -0.00000051, 0.00000100, -0.00000966, 0.00001070, -0.00001553, -0.00000242, -0.00002125, 0.00000336, -0.00009569, -0.00000146, 0.00014500, 0.00011279, 0.00009299, -0.00011462, -0.00000653, -0.00003171, 0.00000244, -0.00001107, 0.00000884, -0.00002402, 0.00000355, -0.00000329, -0.00000133, -0.00001210, 0.00000286, 0.00000155, -0.00000865, -0.00001182, 0.00000391, 0.00000188, -0.00000070, -0.00000493, 0.00001258, -0.00000266, -0.00000751, -0.00000558, 0.00000851, 0.00000299, 0.00000122, -0.00000259, -0.00000116, 0.00000065, -0.00000196, -0.00000130, -0.00000165, 0.00000123, -0.00000023, -0.00000340, 0.00000362, 0.00000393, -0.00000856, -0.00000077, -0.00000299, 0.00000257, 0.00000274, -0.00000274, 0.00000556, -0.00000032, -0.00000380, -0.00000595, 0.00001332, -0.00000094, 0.00000059, 0.00000084, -0.00001402, 0.00001070, -0.00001864, 0.00000489, -0.00000432, -0.00000205, -0.00000669, -0.00000566, -0.00001131, 0.00000803, -0.00001781, 0.00001173, -0.00002363, -0.00001760, -0.00000999, 0.00004249, 0.00003431, 0.00002363, 0.00004615, 0.00000474, -0.00002717, -0.00001709, -0.00000012, -0.00000415, -0.00000058, -0.00000811, 0.00000279, -0.00000224, 0.00000500, 0.00000107, -0.00000337, -0.00000061, -0.00000288, -0.00000422, 0.00000988, -0.00000247, -0.00000273, -0.00000068, },
		{ -0.00000073, -0.00000529, 0.00000745, 0.00000291, 0.00000061, 0.00000267, -0.00000430, -0.00000768, 0.00000392, 0.00000256, 0.00000017, -0.00001514, 0.00000590, 0.00000076, 0.00000843, -0.00000447, 0.00000340, -0.00000241, 0.00000040, -0.00000758, 0.00000375, 0.00000116, 0.00000282, -0.00000673, -0.00000098, -0.00000125, 0.00000423, -0.00000149, -0.00000294, -0.00000319, 0.00000627, -0.00001947, -0.00000199, -0.00001907, -0.00001598, -0.00011645, 0.00010981, 0.00012093, 0.00014581, -0.00003162, -0.00008148, 0.00000048, -0.00002959, 0.00000692, -0.00000826, 0.00000101, -0.00002293, 0.00001243, -0.00001032, -0.00000013, -0.00001318, 0.00001048, -0.00000675, -0.00000964, -0.00000802, 0.00000836, -0.00000181, -0.00000238, 0.00000050, 0.00001202, -0.00001003, -0.00000528, -0.00000202, 0.00001143, -0.00000204, 0.00000288, -0.00000594, 0.00000339, -0.00000316, 0.00000054, -0.00000373, 0.00000135, -0.00000033, -0.00000061, -0.00000335, 0.00000869, -0.00000434, -0.00000544, -0.00000113, -0.00000160, 0.00000453, -0.00000093, 0.00000015, 0.00000556, -0.00000353, -0.00000465, -0.00000026, 0.00001469, -0.00000955, 0.00001050, -0.00001502, 0.00000283, -0.00000324, -0.00001149, 0.00000495, -0.00000550, -0.00000331, -0.00000560, -0.00001131, 0.00000094, -0.00000469, -0.00000688, 0.00000466, -0.00003261, -0.00000690, 0.00000468, 0.00005523, 0.00001209, 0.00004357, 0.00002576, -0.00000753, -0.00003123, -0.00000248, -0.00000553, 0.00000189, -0.00000835, -0.00000041, -0.00000119, 0.00000223, 0.00000302, 0.00000026, -0.00000498, 0.00000252, -0.00000882, 0.00000543, 0.00000474, -0.00000349, -0.00000216, },
		{ -0.00000066, -0.00000365, -0.00000076, 0.00000822, 0.00000002, 0.00000184, 0.00000162, -0.00000907, -0.00000116, 0.00000248, 0.00000518, -0.00001030, -0.00000632, 0.00000515, 0.00000427, 0.00000332, -0.00000293, 0.00000250, -0.00000136, -0.00000347, -0.00000370, 0.00000364, 0.00000279, -0.00000236, -0.00000419, -0.00000229, 0.00000364, -0.00000089, 0.00000188, -0.00001056, 0.00001148, -0.00001449, -0.00000421, -0.00002005, 0.00000323, -0.00008199, -0.00004358, 0.00015871, 0.00012599, 0.00010521, -0.00012040, -0.00000920, -0.00003254, 0.00000104, -0.00001032, 0.00000834, -0.00002305, 0.00000154, -0.00000107, -0.00000247, -0.00001065, -0.00000007, 0.00000400, -0.00000785, -0.00001378, 0.00000357, 0.00000323, -0.00000090, -0.00000577, 0.00001085, 0.00000116, -0.00000880, -0.00000650, 0.00000731, 0.00000508, 0.00000005, -0.00000155, -0.00000214, 0.00000167, -0.00000241, -0.00000076, -0.00000237, 0.00000137, -0.00000020, -0.00000327, 0.00000231, 0.00000623, -0.00000890, -0.00000171, -0.00000231, 0.00000134, 0.00000374, -0.00000351, 0.00000520, 0.00000093, -0.00000344, -0.00000628, 0.00001040, 0.00000389, -0.00000409, 0.00000606, -0.00001771, 0.00001228, -0.00001739, 0.00000218, -0.00000273, -0.00000195, -0.00000687, -0.00000535, -0.00001160, 0.00000810, -0.00001778, 0.00001208, -0.00002174, -0.00001773, -0.00000695, 0.00003412, 0.00003641, 0.00001715, 0.00004532, 0.00000726, -0.00001837, -0.00002326, 0.00000231, -0.00000559, 0.00000096, -0.00000938, 0.00000329, -0.00000275, 0.00000501, 0.00000138, -0.00000254, -0.00000166, -0.00000130, -0.00000666, 0.00001020, -0.00000096, -0.00000315, },
		{ -0.00000274, 0.00000009, -0.00000604, 0.00000613, 0.00000413, 0.00000028, 0.00000241, -0.00000200, -0.00000933, 0.00000361, 0.00000187, 0.00000313, -0.00001638, 0.00000427, 0.00000131, 0.00000859, -0.00000436, 0.00000277, -0.00000138, 0.00000069, -0.00000779, 0.00000250, 0.00000147, 0.00000371, -0.00000713, -0.00000066, -0.00000240, 0.00000532, -0.00000304, -0.00000040, -0.00000680, 0.00001156, -0.00002453, 0.00000302, -0.00002624, 0.00000027, -0.00013051, 0.00008021, 0.00014832, 0.00015314, -0.00002175, -0.00009420, 0.00000554, -0.00003635, 0.00001141, -0.00001237, 0.00000526, -0.00002706, 0.00001536, -0.00001198, 0.00000201, -0.00001540, 0.00001065, -0.00000513, -0.00000920, -0.00001117, 0.00001063, -0.00000245, -0.00000130, -0.00000283, 0.00001423, -0.00000871, -0.00000621, -0.00000400, 0.00001257, -0.00000172, 0.00000279, -0.00000587, 0.00000325, -0.00000266, 0.00000042, -0.00000357, 0.00000081, -0.00000023, -0.00000024, -0.00000423, 0.00000875, -0.00000214, -0.00000713, -0.00000079, -0.00000188, 0.00000389, 0.00000021, -0.00000166, 0.00000694, -0.00000353, -0.00000318, -0.00000363, 0.00001637, -0.00000864, 0.00000835, -0.00001033, -0.00000291, 0.00000250, -0.00001518, 0.00000596, -0.00000577, -0.00000201, -0.00000710, -0.00000949, -0.00000124, -0.00000239, -0.00000934, 0.00000866, -0.00003479, -0.00000208, -0.00000030, 0.00005278, 0.00001059, 0.00003893, 0.00002572, -0.00000081, -0.00002850, -0.00000721, -0.00000251, -0.00000001, -0.00000645, -0.00000235, 0.00000004, 0.00000107, 0.00000359, 0.00000080, -0.00000506, 0.00000252, -0.00000832, 0.00000283, 0.00000671, -0.00000305, },
		{ -0.00000342, -0.00000093, -0.00000264, -0.00000281, 0.00000849, 0.00000074, 0.00000110, 0.00000264, -0.00000784, -0.00000329, 0.00000275, 0.00000511, -0.00000713, -0.00001006, 0.00000604, 0.00000350, 0.00000476, -0.00000448, 0.00000352, -0.00000124, -0.00000261, -0.00000518, 0.00000342, 0.00000321, -0.00000170, -0.00000464, -0.00000236, 0.00000323, -0.00000071, 0.00000184, -0.00001014, 0.00001039, -0.00001106, -0.00000860, -0.00001566, -0.00000202, -0.00006158, -0.00008532, 0.00016168, 0.00014719, 0.00011527, -0.00012363, -0.00001428, -0.00003141, -0.00000232, -0.00000786, 0.00000619, -0.00002046, -0.00000196, 0.00000236, -0.00000467, -0.00000820, -0.00000354, 0.00000632, -0.00000695, -0.00001494, 0.00000190, 0.00000557, -0.00000170, -0.00000568, 0.00000806, 0.00000518, -0.00000966, -0.00000733, 0.00000559, 0.00000735, -0.00000106, -0.00000058, -0.00000309, 0.00000261, -0.00000283, -0.00000017, -0.00000295, 0.00000131, -0.00000018, -0.00000293, 0.00000078, 0.00000822, -0.00000849, -0.00000313, -0.00000148, 0.00000015, 0.00000443, -0.00000370, 0.00000418, 0.00000256, -0.00000356, -0.00000551, 0.00000640, 0.00000896, -0.00000827, 0.00000996, -0.00001920, 0.00001131, -0.00001399, -0.00000168, -0.00000067, -0.00000243, -0.00000612, -0.00000600, -0.00001091, 0.00000711, -0.00001667, 0.00001123, -0.00001869, -0.00001914, -0.00000288, 0.00002554, 0.00003895, 0.00001081, 0.00004489, 0.00000739, -0.00000896, -0.00002765, 0.00000287, -0.00000567, 0.00000149, -0.00000968, 0.00000298, -0.00000266, 0.00000461, 0.00000177, -0.00000166, -0.00000272, 0.00000022, -0.00000859, 0.00000968, 0.00000089, },
		{ -0.00000220, -0.00000337, 0.00000071, -0.00000619, 0.00000429, 0.00000533, 0.00000022, 0.00000185, 0.00000022, -0.00001035, 0.00000284, 0.00000130, 0.00000546, -0.00001631, 0.00000167, 0.00000227, 0.00000835, -0.00000365, 0.00000145, -0.00000003, 0.00000086, -0.00000765, 0.00000106, 0.00000168, 0.00000458, -0.00000742, -0.00000036, -0.00000353, 0.00000644, -0.00000469, 0.00000219, -0.00001027, 0.00001625, -0.00002831, 0.00000633, -0.00003115, 0.00001133, -0.00013245, 0.00003873, 0.00017587, 0.00016451, -0.00001245, -0.00010649, 0.00000998, -0.00004267, 0.00001527, -0.00001586, 0.00000885, -0.00003036, 0.00001730, -0.00001267, 0.00000322, -0.00001648, 0.00000980, -0.00000329, -0.00000831, -0.00001439, 0.00001217, -0.00000226, -0.00000068, -0.00000542, 0.00001504, -0.00000635, -0.00000728, -0.00000582, 0.00001303, -0.00000077, 0.00000229, -0.00000548, 0.00000278, -0.00000194, 0.00000008, -0.00000307, 0.00000010, -0.00000016, 0.00000010, -0.00000484, 0.00000816, 0.00000045, -0.00000853, -0.00000091, -0.00000172, 0.00000284, 0.00000159, -0.00000337, 0.00000788, -0.00000323, -0.00000198, -0.00000591, 0.00001613, -0.00000575, 0.00000434, -0.00000427, -0.00000898, 0.00000745, -0.00001740, 0.00000579, -0.00000554, -0.00000099, -0.00000811, -0.00000822, -0.00000280, -0.00000080, -0.00001099, 0.00001164, -0.00003582, 0.00000118, -0.00000337, 0.00004901, 0.00001065, 0.00003345, 0.00002625, 0.00000269, -0.00002237, -0.00001331, 0.00000084, -0.00000203, -0.00000441, -0.00000425, 0.00000112, 0.00000009, 0.00000397, 0.00000129, -0.00000485, 0.00000210, -0.00000719, -0.00000011, 0.00000829, },
		{ 0.00000296, -0.00000344, -0.00000146, -0.00000143, -0.00000460, 0.00000818, 0.00000173, 0.00000046, 0.00000313, -0.00000602, -0.00000548, 0.00000301, 0.00000450, -0.00000352, -0.00001336, 0.00000624, 0.00000304, 0.00000607, -0.00000572, 0.00000394, -0.00000073, -0.00000184, -0.00000639, 0.00000293, 0.00000353, -0.00000085, -0.00000527, -0.00000210, 0.00000235, 0.00000012, 0.00000081, -0.00000836, 0.00000751, -0.00000559, -0.00001499, -0.00000896, -0.00001074, -0.00003804, -0.00012133, 0.00015010, 0.00017689, 0.00012382, -0.00012469, -0.00002158, -0.00002842, -0.00000757, -0.00000379, 0.00000255, -0.00001646, -0.00000666, 0.00000668, -0.00000757, -0.00000515, -0.00000713, 0.00000828, -0.00000602, -0.00001519, -0.00000097, 0.00000852, -0.00000272, -0.00000502, 0.00000471, 0.00000889, -0.00000982, -0.00000817, 0.00000352, 0.00000955, -0.00000188, 0.00000018, -0.00000391, 0.00000336, -0.00000312, 0.00000031, -0.00000323, 0.00000102, -0.00000015, -0.00000244, -0.00000081, 0.00000968, -0.00000727, -0.00000490, -0.00000064, -0.00000084, 0.00000466, -0.00000326, 0.00000259, 0.00000441, -0.00000402, -0.00000397, 0.00000199, 0.00001341, -0.00001106, 0.00001175, -0.00001798, 0.00000778, -0.00000884, -0.00000617, 0.00000154, -0.00000335, -0.00000460, -0.00000736, -0.00000949, 0.00000540, -0.00001486, 0.00000966, -0.00001501, -0.00002122, 0.00000155, 0.00001733, 0.00004136, 0.00000511, 0.00004431, 0.00000639, -0.00000068, -0.00002904, 0.00000102, -0.00000420, 0.00000089, -0.00000899, 0.00000191, -0.00000205, 0.00000392, 0.00000221, -0.00000079, -0.00000365, 0.00000148, -0.00000980, 0.00000827, },
		{ 0.00000927, -0.00000092, -0.00000396, 0.00000100, -0.00000573, 0.00000207, 0.00000636, 0.00000045, 0.00000110, 0.00000211, -0.00001058, 0.00000157, 0.00000098, 0.00000696, -0.00001489, -0.00000175, 0.00000347, 0.00000777, -0.00000237, -0.00000038, 0.00000140, 0.00000103, -0.00000721, -0.00000052, 0.00000179, 0.00000534, -0.00000749, -0.00000024, -0.00000443, 0.00000729, -0.00000602, 0.00000419, -0.00001279, 0.00001934, -0.00002971, 0.00000674, -0.00003228, 0.00001554, -0.00012215, -0.00001120, 0.00019852, 0.00018319, -0.00000532, -0.00011718, 0.00001280, -0.00004761, 0.00001757, -0.00001790, 0.00001098, -0.00003211, 0.00001762, -0.00001189, 0.00000308, -0.00001613, 0.00000784, -0.00000116, -0.00000728, -0.00001710, 0.00001247, -0.00000096, -0.00000072, -0.00000703, 0.00001442, -0.00000313, -0.00000834, -0.00000745, 0.00001275, 0.00000077, 0.00000148, -0.00000484, 0.00000203, -0.00000100, -0.00000048, -0.00000230, -0.00000065, -0.00000018, 0.00000039, -0.00000510, 0.00000699, 0.00000319, -0.00000940, -0.00000159, -0.00000116, 0.00000151, 0.00000299, -0.00000473, 0.00000818, -0.00000246, -0.00000133, -0.00000686, 0.00001400, -0.00000126, -0.00000077, 0.00000212, -0.00001418, 0.00001057, -0.00001752, 0.00000413, -0.00000455, -0.00000061, -0.00000821, -0.00000792, -0.00000332, -0.00000035, -0.00001145, 0.00001327, -0.00003544, 0.00000271, -0.00000450, 0.00004388, 0.00001232, 0.00002699, 0.00002778, 0.00000305, -0.00001392, -0.00001949, 0.00000352, -0.00000345, -0.00000283, -0.00000565, 0.00000171, -0.00000047, 0.00000408, 0.00000176, -0.00000438, 0.00000131, -0.00000560, -0.00000307, },
		{ 0.00000606, 0.00000507, -0.00000317, -0.00000217, -0.00000021, -0.00000588, 0.00000718, 0.00000297, -0.00000003, 0.00000314, -0.00000386, -0.00000746, 0.00000309, 0.00000362, 0.00000008, -0.00001575, 0.00000549, 0.00000304, 0.00000700, -0.00000636, 0.00000357, 0.00000024, -0.00000122, -0.00000721, 0.00000213, 0.00000380, 0.00000008, -0.00000597, -0.00000162, 0.00000114, 0.00000145, -0.00000101, -0.00000559, 0.00000344, 0.00000103, -0.00002216, -0.00000156, -0.00002045, -0.00001555, -0.00014664, 0.00012170, 0.00021380, 0.00013287, -0.00012478, -0.00003027, -0.00002423, -0.00001409, 0.00000124, -0.00000195, -0.00001170, -0.00001190, 0.00001119, -0.00001050, -0.00000216, -0.00001028, 0.00000955, -0.00000494, -0.00001468, -0.00000458, 0.00001146, -0.00000345, -0.00000425, 0.00000136, 0.00001175, -0.00000903, -0.00000911, 0.00000131, 0.00001142, -0.00000223, 0.00000062, -0.00000446, 0.00000381, -0.00000315, 0.00000053, -0.00000309, 0.00000046, -0.00000013, -0.00000187, -0.00000225, 0.00001043, -0.00000526, -0.00000683, -0.00000001, -0.00000147, 0.00000436, -0.00000226, 0.00000065, 0.00000626, -0.00000457, -0.00000213, -0.00000209, 0.00001645, -0.00001179, 0.00001101, -0.00001408, 0.00000219, -0.00000274, -0.00001055, 0.00000341, -0.00000437, -0.00000270, -0.00000898, -0.00000786, 0.00000353, -0.00001299, 0.00000803, -0.00001142, -0.00002318, 0.00000558, 0.00001014, 0.00004298, 0.00000070, 0.00004288, 0.00000559, 0.00000502, -0.00002683, -0.00000320, -0.00000141, -0.00000058, -0.00000755, 0.00000032, -0.00000115, 0.00000309, 0.00000261, 0.00000003, -0.00000433, 0.00000230, -0.00001017, },
		{ -0.00000574, 0.00000945, 0.00000075, -0.00000441, 0.00000092, -0.00000472, -0.00000029, 0.00000701, 0.00000101, 0.00000028, 0.00000355, -0.00001002, -0.00000014, 0.00000095, 0.00000756, -0.00001228, -0.00000561, 0.00000463, 0.00000703, -0.00000072, -0.00000243, 0.00000263, 0.00000134, -0.00000656, -0.00000212, 0.00000178, 0.00000594, -0.00000728, -0.00000038, -0.00000493, 0.00000760, -0.00000666, 0.00000512, -0.00001377, 0.00002019, -0.00002812, 0.00000368, -0.00002911, 0.00001273, -0.00010156, -0.00006424, 0.00021042, 0.00021223, -0.00000146, -0.00012551, 0.00001328, -0.00005043, 0.00001764, -0.00001790, 0.00001113, -0.00003185, 0.00001600, -0.00000942, 0.00000144, -0.00001436, 0.00000490, 0.00000119, -0.00000633, -0.00001890, 0.00001127, 0.00000145, -0.00000137, -0.00000770, 0.00001256, 0.00000058, -0.00000914, -0.00000889, 0.00001178, 0.00000277, 0.00000048, -0.00000402, 0.00000107, 0.00000007, -0.00000118, -0.00000140, -0.00000128, -0.00000036, 0.00000062, -0.00000503, 0.00000538, 0.00000578, -0.00000953, -0.00000285, -0.00000030, 0.00000007, 0.00000417, -0.00000554, 0.00000776, -0.00000116, -0.00000131, -0.00000648, 0.00001036, 0.00000409, -0.00000601, 0.00000764, -0.00001737, 0.00001109, -0.00001524, 0.00000097, -0.00000276, -0.00000109, -0.00000715, -0.00000877, -0.00000265, -0.00000115, -0.00001068, 0.00001355, -0.00003373, 0.00000272, -0.00000398, 0.00003764, 0.00001529, 0.00001981, 0.00003012, 0.00000114, -0.00000472, -0.00002426, 0.00000456, -0.00000365, -0.00000215, -0.00000621, 0.00000160, -0.00000053, 0.00000386, 0.00000221, -0.00000370, 0.00000029, -0.00000379, },
		{ -0.00000967, 0.00000322, 0.00000701, -0.00000254, -0.00000295, 0.00000078, -0.00000646, 0.00000549, 0.00000433, -0.00000032, 0.00000280, -0.00000163, -0.00000895, 0.00000280, 0.00000274, 0.00000321, -0.00001685, 0.00000362, 0.00000361, 0.00000741, -0.00000620, 0.00000232, 0.00000163, -0.00000077, -0.00000759, 0.00000096, 0.00000411, 0.00000095, -0.00000659, -0.00000110, -0.00000017, 0.00000296, -0.00000316, -0.00000248, -0.00000092, 0.00000755, -0.00002852, 0.00000457, -0.00002841, 0.00000194, -0.00015759, 0.00007645, 0.00025445, 0.00014573, -0.00012586, -0.00003897, -0.00001996, -0.00002081, 0.00000621, -0.00000633, -0.00000716, -0.00001670, 0.00001500, -0.00001264, 0.00000000, -0.00001234, 0.00000973, -0.00000350, -0.00001377, -0.00000831, 0.00001369, -0.00000335, -0.00000382, -0.00000147, 0.00001337, -0.00000718, -0.00001018, -0.00000088, 0.00001272, -0.00000195, 0.00000061, -0.00000465, 0.00000384, -0.00000282, 0.00000036, -0.00000249, -0.00000032, -0.00000012, -0.00000132, -0.00000337, 0.00001035, -0.00000261, -0.00000867, 0.00000025, -0.00000166, 0.00000356, -0.00000087, -0.00000139, 0.00000782, -0.00000490, -0.00000046, -0.00000521, 0.00001754, -0.00001021, 0.00000780, -0.00000805, -0.00000447, 0.00000319, -0.00001390, 0.00000434, -0.00000504, -0.00000095, -0.00001023, -0.00000667, 0.00000216, -0.00001173, 0.00000705, -0.00000864, -0.00002432, 0.00000853, 0.00000455, 0.00004322, -0.00000185, 0.00003993, 0.00000604, 0.00000737, -0.00002112, -0.00000911, 0.00000202, -0.00000239, -0.00000581, -0.00000138, -0.00000026, 0.00000235, 0.00000289, 0.00000077, -0.00000467, 0.00000258, },
		{ -0.00000203, -0.00000780, 0.00000870, 0.00000275, -0.00000464, 0.00000049, -0.00000336, -0.00000249, 0.00000707, 0.00000192, -0.00000052, 0.00000448, -0.00000878, -0.00000210, 0.00000110, 0.00000739, -0.00000883, -0.00000944, 0.00000537, 0.00000636, 0.00000102, -0.00000432, 0.00000334, 0.00000196, -0.00000583, -0.00000362, 0.00000159, 0.00000642, -0.00000683, -0.00000081, -0.00000498, 0.00000729, -0.00000643, 0.00000477, -0.00001301, 0.00001867, -0.00002362, -0.00000249, -0.00002223, 0.00000432, -0.00007452, -0.00011381, 0.00020587, 0.00025364, -0.00000089, -0.00013147, 0.00001128, -0.00005094, 0.00001530, -0.00001573, 0.00000925, -0.00002962, 0.00001255, -0.00000546, -0.00000143, -0.00001151, 0.00000137, 0.00000354, -0.00000550, -0.00001967, 0.00000868, 0.00000463, -0.00000233, -0.00000768, 0.00000986, 0.00000428, -0.00000940, -0.00001024, 0.00001023, 0.00000498, -0.00000048, -0.00000320, 0.00000004, 0.00000114, -0.00000187, -0.00000056, -0.00000161, -0.00000078, 0.00000081, -0.00000469, 0.00000356, 0.00000791, -0.00000876, -0.00000462, 0.00000072, -0.00000128, 0.00000497, -0.00000570, 0.00000664, 0.00000057, -0.00000183, -0.00000506, 0.00000586, 0.00000936, -0.00001028, 0.00001120, -0.00001772, 0.00000868, -0.00001069, -0.00000336, -0.00000036, -0.00000237, -0.00000500, -0.00001065, -0.00000098, -0.00000293, -0.00000900, 0.00001289, -0.00003115, 0.00000176, -0.00000245, 0.00003089, 0.00001892, 0.00001258, 0.00003257, -0.00000173, 0.00000344, -0.00002632, 0.00000326, -0.00000227, -0.00000258, -0.00000580, 0.00000074, -0.00000009, 0.00000339, 0.00000263, -0.00000289, -0.00000080, },
		{ 0.00000231, -0.00000842, 0.00000002, 0.00000855, -0.00000153, -0.00000368, 0.00000139, -0.00000625, 0.00000321, 0.00000567, -0.00000038, 0.00000223, 0.00000039, -0.00000974, 0.00000201, 0.00000209, 0.00000551, -0.00001645, 0.00000065, 0.00000469, 0.00000722, -0.00000518, 0.00000027, 0.00000329, -0.00000045, -0.00000748, -0.00000058, 0.00000452, 0.00000163, -0.00000694, -0.00000078, -0.00000125, 0.00000422, -0.00000504, 0.00000017, -0.00000448, 0.00001261, -0.00003249, 0.00000758, -0.00003225, 0.00001147, -0.00015260, 0.00001706, 0.00029327, 0.00016675, -0.00013034, -0.00004602, -0.00001704, -0.00002643, 0.00000986, -0.00000942, -0.00000393, -0.00002005, 0.00001719, -0.00001320, 0.00000066, -0.00001281, 0.00000854, -0.00000149, -0.00001286, -0.00001149, 0.00001459, -0.00000204, -0.00000402, -0.00000342, 0.00001355, -0.00000432, -0.00001131, -0.00000289, 0.00001326, -0.00000095, 0.00000014, -0.00000442, 0.00000341, -0.00000206, -0.00000025, -0.00000145, -0.00000122, -0.00000015, -0.00000085, -0.00000406, 0.00000948, 0.00000045, -0.00001013, -0.00000001, -0.00000140, 0.00000237, 0.00000069, -0.00000322, 0.00000883, -0.00000476, 0.00000067, -0.00000696, 0.00001651, -0.00000649, 0.00000273, -0.00000093, -0.00001091, 0.00000774, -0.00001539, 0.00000383, -0.00000492, 0.00000013, -0.00001053, -0.00000648, 0.00000188, -0.00001166, 0.00000726, -0.00000717, -0.00002415, 0.00001000, 0.00000086, 0.00004175, -0.00000226, 0.00003518, 0.00000821, 0.00000638, -0.00001284, -0.00001556, 0.00000514, -0.00000383, -0.00000434, -0.00000274, 0.00000027, 0.00000190, 0.00000298, 0.00000141, -0.00000465, },
		{ -0.00000177, -0.00000060, -0.00000901, 0.00000699, 0.00000497, -0.00000462, -0.00000019, -0.00000190, -0.00000420, 0.00000638, 0.00000317, -0.00000125, 0.00000493, -0.00000706, -0.00000405, 0.00000124, 0.00000670, -0.00000503, -0.00001267, 0.00000530, 0.00000606, 0.00000251, -0.00000566, 0.00000322, 0.00000303, -0.00000516, -0.00000483, 0.00000108, 0.00000689, -0.00000627, -0.00000140, -0.00000467, 0.00000645, -0.00000544, 0.00000331, -0.00001079, 0.00001529, -0.00001706, -0.00001055, -0.00001333, -0.00000696, -0.00004608, -0.00015306, 0.00018038, 0.00030749, -0.00000205, -0.00013619, 0.00000753, -0.00004967, 0.00001104, -0.00001193, 0.00000587, -0.00002601, 0.00000793, -0.00000068, -0.00000486, -0.00000827, -0.00000211, 0.00000544, -0.00000462, -0.00001957, 0.00000514, 0.00000796, -0.00000306, -0.00000742, 0.00000690, 0.00000741, -0.00000883, -0.00001159, 0.00000833, 0.00000712, -0.00000117, -0.00000255, -0.00000086, 0.00000201, -0.00000235, -0.00000003, -0.00000145, -0.00000148, 0.00000097, -0.00000418, 0.00000178, 0.00000931, -0.00000704, -0.00000676, 0.00000171, -0.00000238, 0.00000525, -0.00000521, 0.00000497, 0.00000252, -0.00000265, -0.00000308, 0.00000130, 0.00001360, -0.00001267, 0.00001204, -0.00001492, 0.00000358, -0.00000452, -0.00000820, 0.00000218, -0.00000413, -0.00000211, -0.00001310, 0.00000121, -0.00000515, -0.00000702, 0.00001195, -0.00002843, 0.00000060, -0.00000073, 0.00002441, 0.00002236, 0.00000619, 0.00003417, -0.00000410, 0.00000904, -0.00002483, -0.00000056, 0.00000061, -0.00000401, -0.00000456, -0.00000069, 0.00000068, 0.00000277, 0.00000299, -0.00000204, },
		{ -0.00000425, 0.00000156, -0.00000661, -0.00000320, 0.00000945, -0.00000015, -0.00000425, 0.00000150, -0.00000528, 0.00000057, 0.00000677, -0.00000016, 0.00000156, 0.00000202, -0.00000973, 0.00000067, 0.00000182, 0.00000680, -0.00001455, -0.00000320, 0.00000609, 0.00000653, -0.00000341, -0.00000234, 0.00000499, -0.00000018, -0.00000690, -0.00000248, 0.00000505, 0.00000202, -0.00000691, -0.00000086, -0.00000180, 0.00000484, -0.00000609, 0.00000160, -0.00000633, 0.00001513, -0.00003293, 0.00000628, -0.00003056, 0.00001175, -0.00013263, -0.00005108, 0.00032306, 0.00020065, -0.00014047, -0.00004993, -0.00001672, -0.00002975, 0.00001108, -0.00001022, -0.00000291, -0.00002114, 0.00001710, -0.00001164, -0.00000061, -0.00001141, 0.00000590, 0.00000116, -0.00001228, -0.00001361, 0.00001377, 0.00000060, -0.00000494, -0.00000437, 0.00001230, -0.00000071, -0.00001236, -0.00000467, 0.00001297, 0.00000072, -0.00000075, -0.00000378, 0.00000251, -0.00000090, -0.00000125, -0.00000010, -0.00000210, -0.00000030, -0.00000051, -0.00000428, 0.00000792, 0.00000361, -0.00001096, -0.00000088, -0.00000074, 0.00000095, 0.00000218, -0.00000460, 0.00000911, -0.00000399, 0.00000104, -0.00000721, 0.00001356, -0.00000125, -0.00000323, 0.00000598, -0.00001576, 0.00000987, -0.00001440, 0.00000162, -0.00000376, 0.00000014, -0.00000947, -0.00000766, 0.00000302, -0.00001307, 0.00000890, -0.00000719, -0.00002258, 0.00000996, -0.00000101, 0.00003863, -0.00000065, 0.00002880, 0.00001189, 0.00000284, -0.00000347, -0.00002108, 0.00000690, -0.00000419, -0.00000364, -0.00000340, 0.00000020, 0.00000186, 0.00000284, 0.00000194, },
		{ -0.00000123, -0.00000244, 0.00000027, -0.00000922, 0.00000438, 0.00000724, -0.00000432, -0.00000098, -0.00000064, -0.00000515, 0.00000485, 0.00000471, -0.00000189, 0.00000503, -0.00000515, -0.00000569, 0.00000115, 0.00000584, -0.00000145, -0.00001477, 0.00000411, 0.00000635, 0.00000343, -0.00000608, 0.00000206, 0.00000462, -0.00000467, -0.00000555, 0.00000007, 0.00000754, -0.00000581, -0.00000194, -0.00000426, 0.00000541, -0.00000409, 0.00000125, -0.00000787, 0.00001110, -0.00000989, -0.00001862, -0.00000485, -0.00001761, -0.00002166, -0.00017606, 0.00013173, 0.00037132, -0.00000155, -0.00014196, 0.00000363, -0.00004790, 0.00000603, -0.00000764, 0.00000211, -0.00002209, 0.00000321, 0.00000386, -0.00000785, -0.00000559, -0.00000470, 0.00000632, -0.00000334, -0.00001906, 0.00000138, 0.00001066, -0.00000297, -0.00000741, 0.00000425, 0.00000946, -0.00000721, -0.00001304, 0.00000630, 0.00000884, -0.00000131, -0.00000228, -0.00000143, 0.00000246, -0.00000239, -0.00000002, -0.00000068, -0.00000249, 0.00000112, -0.00000365, 0.00000029, 0.00000979, -0.00000444, -0.00000904, 0.00000247, -0.00000310, 0.00000501, -0.00000421, 0.00000301, 0.00000438, -0.00000342, -0.00000106, -0.00000256, 0.00001605, -0.00001259, 0.00000993, -0.00000928, -0.00000339, 0.00000225, -0.00001263, 0.00000422, -0.00000587, 0.00000093, -0.00001544, 0.00000319, -0.00000706, -0.00000553, 0.00001155, -0.00002640, 0.00000011, 0.00000036, 0.00001898, 0.00002476, 0.00000150, 0.00003399, -0.00000475, 0.00001109, -0.00001967, -0.00000646, 0.00000448, -0.00000600, -0.00000286, -0.00000238, 0.00000152, 0.00000217, 0.00000325, },
		{ 0.00000234, -0.00000355, 0.00000049, -0.00000456, -0.00000606, 0.00000949, 0.00000160, -0.00000459, 0.00000112, -0.00000375, -0.00000210, 0.00000739, 0.00000040, 0.00000087, 0.00000315, -0.00000894, -0.00000111, 0.00000192, 0.00000706, -0.00001137, -0.00000749, 0.00000746, 0.00000557, -0.00000118, -0.00000511, 0.00000638, 0.00000019, -0.00000595, -0.00000459, 0.00000563, 0.00000213, -0.00000647, -0.00000145, -0.00000167, 0.00000456, -0.00000596, 0.00000137, -0.00000597, 0.00001461, -0.00002945, 0.00000045, -0.00002333, 0.00000349, -0.00010126, -0.00012035, 0.00033590, 0.00025159, -0.00015761, -0.00004995, -0.00001969, -0.00003007, 0.00000923, -0.00000820, -0.00000454, -0.00001966, 0.00001452, -0.00000789, -0.00000378, -0.00000830, 0.00000204, 0.00000433, -0.00001215, -0.00001443, 0.00001122, 0.00000431, -0.00000636, -0.00000447, 0.00000994, 0.00000325, -0.00001309, -0.00000621, 0.00001188, 0.00000293, -0.00000190, -0.00000285, 0.00000126, 0.00000054, -0.00000249, 0.00000134, -0.00000273, -0.00000062, -0.00000030, -0.00000406, 0.00000591, 0.00000650, -0.00001092, -0.00000234, 0.00000021, -0.00000052, 0.00000336, -0.00000535, 0.00000859, -0.00000261, 0.00000067, -0.00000614, 0.00000925, 0.00000459, -0.00000888, 0.00001134, -0.00001788, 0.00000890, -0.00001078, -0.00000224, -0.00000158, -0.00000102, -0.00000692, -0.00001026, 0.00000557, -0.00001587, 0.00001179, -0.00000844, -0.00001993, 0.00000883, -0.00000153, 0.00003435, 0.00000239, 0.00002150, 0.00001630, -0.00000186, 0.00000519, -0.00002424, 0.00000645, -0.00000299, -0.00000404, -0.00000313, -0.00000056, 0.00000225, 0.00000252, },
		{ 0.00000339, -0.00000057, -0.00000267, 0.00000046, -0.00000839, 0.00000107, 0.00000935, -0.00000373, -0.00000170, 0.00000017, -0.00000515, 0.00000250, 0.00000644, -0.00000242, 0.00000492, -0.00000335, -0.00000672, 0.00000057, 0.00000514, 0.00000140, -0.00001530, 0.00000160, 0.00000740, 0.00000353, -0.00000534, -0.00000021, 0.00000672, -0.00000449, -0.00000558, -0.00000161, 0.00000855, -0.00000571, -0.00000212, -0.00000414, 0.00000462, -0.00000297, -0.00000060, -0.00000528, 0.00000746, -0.00000388, -0.00002453, 0.00000060, -0.00002411, -0.00000599, -0.00017890, 0.00006085, 0.00043982, 0.00000580, -0.00015208, 0.00000186, -0.00004749, 0.00000200, -0.00000447, -0.00000053, -0.00001930, -0.00000026, 0.00000693, -0.00000927, -0.00000447, -0.00000559, 0.00000563, -0.00000123, -0.00001869, -0.00000181, 0.00001192, -0.00000154, -0.00000810, 0.00000243, 0.00001005, -0.00000444, -0.00001466, 0.00000439, 0.00000986, -0.00000068, -0.00000257, -0.00000148, 0.00000229, -0.00000182, -0.00000068, 0.00000073, -0.00000374, 0.00000130, -0.00000323, -0.00000069, 0.00000924, -0.00000112, -0.00001119, 0.00000284, -0.00000340, 0.00000431, -0.00000290, 0.00000108, 0.00000583, -0.00000378, 0.00000048, -0.00000511, 0.00001624, -0.00000991, 0.00000519, -0.00000166, -0.00001095, 0.00000831, -0.00001562, 0.00000508, -0.00000700, 0.00000344, -0.00001694, 0.00000423, -0.00000788, -0.00000531, 0.00001243, -0.00002577, 0.00000094, 0.00000017, 0.00001518, 0.00002550, -0.00000085, 0.00003144, -0.00000300, 0.00000943, -0.00001151, -0.00001346, 0.00000846, -0.00000788, -0.00000123, -0.00000390, 0.00000214, 0.00000175, },
		{ 0.00000208, 0.00000260, -0.00000266, -0.00000067, -0.00000260, -0.00000819, 0.00000853, 0.00000366, -0.00000467, 0.00000036, -0.00000193, -0.00000445, 0.00000730, 0.00000133, 0.00000021, 0.00000380, -0.00000753, -0.00000313, 0.00000225, 0.00000649, -0.00000740, -0.00001166, 0.00000837, 0.00000468, 0.00000111, -0.00000756, 0.00000710, 0.00000084, -0.00000481, -0.00000669, 0.00000608, 0.00000211, -0.00000573, -0.00000244, -0.00000089, 0.00000341, -0.00000465, -0.00000050, -0.00000347, 0.00001131, -0.00002263, -0.00000892, -0.00001200, -0.00001064, -0.00006414, -0.00018194, 0.00032436, 0.00032206, -0.00018147, -0.00004660, -0.00002567, -0.00002754, 0.00000446, -0.00000359, -0.00000854, -0.00001595, 0.00000989, -0.00000246, -0.00000830, -0.00000410, -0.00000241, 0.00000756, -0.00001229, -0.00001409, 0.00000735, 0.00000853, -0.00000779, -0.00000409, 0.00000695, 0.00000701, -0.00001320, -0.00000760, 0.00001016, 0.00000540, -0.00000307, -0.00000183, -0.00000013, 0.00000202, -0.00000371, 0.00000256, -0.00000291, -0.00000120, -0.00000020, -0.00000352, 0.00000373, 0.00000877, -0.00000988, -0.00000432, 0.00000127, -0.00000185, 0.00000409, -0.00000540, 0.00000736, -0.00000077, -0.00000025, -0.00000416, 0.00000437, 0.00000996, -0.00001303, 0.00001405, -0.00001659, 0.00000472, -0.00000486, -0.00000724, 0.00000129, -0.00000317, -0.00000315, -0.00001391, 0.00000910, -0.00001956, 0.00001536, -0.00001027, -0.00001689, 0.00000736, -0.00000151, 0.00002969, 0.00000597, 0.00001430, 0.00002027, -0.00000612, 0.00001144, -0.00002398, 0.00000336, -0.00000010, -0.00000554, -0.00000197, -0.00000194, 0.00000298, },
		{ 0.00000164, 0.00000341, -0.00000013, -0.00000239, -0.00000006, -0.00000666, -0.00000265, 0.00001108, -0.00000286, -0.00000219, 0.00000035, -0.00000414, -0.00000052, 0.00000817, -0.00000282, 0.00000476, -0.00000193, -0.00000692, -0.00000065, 0.00000491, 0.00000313, -0.00001404, -0.00000223, 0.00000918, 0.00000274, -0.00000341, -0.00000350, 0.00000921, -0.00000465, -0.00000478, -0.00000408, 0.00001009, -0.00000619, -0.00000166, -0.00000467, 0.00000460, -0.00000274, -0.00000139, -0.00000414, 0.00000577, -0.00000073, -0.00002634, 0.00000082, -0.00002375, -0.00000219, -0.00016056, -0.00002784, 0.00050500, 0.00002643, -0.00017029, 0.00000475, -0.00005051, 0.00000079, -0.00000414, -0.00000051, -0.00001905, -0.00000119, 0.00000737, -0.00000817, -0.00000574, -0.00000415, 0.00000294, 0.00000206, -0.00001900, -0.00000367, 0.00001116, 0.00000156, -0.00000972, 0.00000174, 0.00000901, -0.00000059, -0.00001642, 0.00000282, 0.00000991, 0.00000085, -0.00000352, -0.00000088, 0.00000139, -0.00000054, -0.00000207, 0.00000273, -0.00000515, 0.00000149, -0.00000305, -0.00000103, 0.00000772, 0.00000263, -0.00001294, 0.00000270, -0.00000328, 0.00000330, -0.00000155, -0.00000050, 0.00000659, -0.00000347, 0.00000120, -0.00000603, 0.00001415, -0.00000503, -0.00000132, 0.00000661, -0.00001763, 0.00001238, -0.00001633, 0.00000423, -0.00000706, 0.00000484, -0.00001697, 0.00000374, -0.00000705, -0.00000686, 0.00001505, -0.00002693, 0.00000342, -0.00000154, 0.00001319, 0.00002435, -0.00000071, 0.00002648, 0.00000108, 0.00000470, -0.00000171, -0.00002015, 0.00001150, -0.00000893, -0.00000019, -0.00000487, 0.00000229, },
		{ 0.00000385, 0.00000164, 0.00000273, -0.00000173, -0.00000166, -0.00000107, -0.00000931, 0.00000651, 0.00000594, -0.00000450, -0.00000060, -0.00000017, -0.00000607, 0.00000631, 0.00000266, -0.00000042, 0.00000406, -0.00000576, -0.00000506, 0.00000255, 0.00000547, -0.00000323, -0.00001504, 0.00000834, 0.00000424, 0.00000300, -0.00000919, 0.00000678, 0.00000197, -0.00000371, -0.00000845, 0.00000610, 0.00000223, -0.00000498, -0.00000357, 0.00000023, 0.00000173, -0.00000256, -0.00000349, 0.00000041, 0.00000628, -0.00001399, -0.00001978, 0.00000070, -0.00002654, -0.00002821, -0.00022707, 0.00028294, 0.00041190, -0.00020951, -0.00004196, -0.00003316, -0.00002337, -0.00000212, 0.00000253, -0.00001385, -0.00001107, 0.00000430, 0.00000358, -0.00001312, 0.00000018, -0.00000654, 0.00001018, -0.00001226, -0.00001308, 0.00000291, 0.00001239, -0.00000857, -0.00000377, 0.00000398, 0.00000997, -0.00001239, -0.00000899, 0.00000808, 0.00000775, -0.00000393, -0.00000099, -0.00000138, 0.00000324, -0.00000458, 0.00000326, -0.00000243, -0.00000210, -0.00000013, -0.00000285, 0.00000171, 0.00001011, -0.00000777, -0.00000667, 0.00000228, -0.00000290, 0.00000431, -0.00000484, 0.00000565, 0.00000122, -0.00000134, -0.00000187, -0.00000019, 0.00001385, -0.00001478, 0.00001348, -0.00001182, -0.00000213, 0.00000247, -0.00001254, 0.00000424, -0.00000579, 0.00000129, -0.00001795, 0.00001290, -0.00002334, 0.00001874, -0.00001181, -0.00001440, 0.00000647, -0.00000185, 0.00002555, 0.00000905, 0.00000835, 0.00002257, -0.00000845, 0.00001405, -0.00001988, -0.00000221, 0.00000414, -0.00000784, -0.00000019, -0.00000370, },
		{ 0.00000184, 0.00000185, 0.00000336, 0.00000004, -0.00000159, -0.00000122, -0.00000428, -0.00000639, 0.00001219, -0.00000171, -0.00000235, -0.00000019, -0.00000223, -0.00000392, 0.00000970, -0.00000304, 0.00000467, -0.00000108, -0.00000618, -0.00000257, 0.00000529, 0.00000353, -0.00001101, -0.00000712, 0.00001149, 0.00000116, -0.00000042, -0.00000751, 0.00001185, -0.00000512, -0.00000312, -0.00000733, 0.00001221, -0.00000737, -0.00000035, -0.00000617, 0.00000577, -0.00000393, -0.00000041, -0.00000533, 0.00000707, -0.00000160, -0.00002288, -0.00000536, -0.00001527, -0.00001106, -0.00012330, -0.00012660, 0.00055698, 0.00006709, -0.00019998, 0.00001439, -0.00005865, 0.00000395, -0.00000803, 0.00000338, -0.00002239, 0.00000131, 0.00000448, -0.00000396, -0.00000981, -0.00000013, -0.00000181, 0.00000663, -0.00002035, -0.00000376, 0.00000813, 0.00000629, -0.00001224, 0.00000224, 0.00000642, 0.00000409, -0.00001821, 0.00000171, 0.00000888, 0.00000328, -0.00000512, 0.00000038, -0.00000025, 0.00000143, -0.00000409, 0.00000512, -0.00000653, 0.00000166, -0.00000315, -0.00000070, 0.00000539, 0.00000646, -0.00001402, 0.00000200, -0.00000285, 0.00000218, -0.00000043, -0.00000148, 0.00000648, -0.00000238, 0.00000097, -0.00000534, 0.00001016, 0.00000122, -0.00000836, 0.00001401, -0.00002199, 0.00001345, -0.00001422, 0.00000146, -0.00000580, 0.00000479, -0.00001518, 0.00000144, -0.00000437, -0.00001031, 0.00001943, -0.00002981, 0.00000741, -0.00000457, 0.00001276, 0.00002156, 0.00000155, 0.00001966, 0.00000675, -0.00000175, 0.00000797, -0.00002504, 0.00001260, -0.00000852, -0.00000016, -0.00000501, },
		{ -0.00000548, 0.00000462, 0.00000132, 0.00000277, -0.00000092, -0.00000224, -0.00000027, -0.00000923, 0.00000350, 0.00000831, -0.00000412, -0.00000152, 0.00000115, -0.00000666, 0.00000432, 0.00000436, -0.00000107, 0.00000413, -0.00000400, -0.00000652, 0.00000249, 0.00000444, 0.00000046, -0.00001698, 0.00000695, 0.00000462, 0.00000402, -0.00000953, 0.00000515, 0.00000374, -0.00000294, -0.00000948, 0.00000532, 0.00000287, -0.00000464, -0.00000439, 0.00000116, 0.00000015, -0.00000047, -0.00000661, 0.00000438, 0.00000121, -0.00000573, -0.00002938, 0.00001130, -0.00003939, -0.00000041, -0.00024852, 0.00020925, 0.00051753, -0.00023670, -0.00003979, -0.00003942, -0.00001974, -0.00000852, 0.00000829, -0.00001871, -0.00000669, -0.00000067, 0.00000872, -0.00001683, 0.00000324, -0.00000921, 0.00001133, -0.00001141, -0.00001214, -0.00000112, 0.00001494, -0.00000801, -0.00000409, 0.00000166, 0.00001160, -0.00001040, -0.00001055, 0.00000598, 0.00000955, -0.00000412, -0.00000066, -0.00000215, 0.00000385, -0.00000478, 0.00000314, -0.00000114, -0.00000336, -0.00000001, -0.00000225, 0.00000018, 0.00001030, -0.00000469, -0.00000916, 0.00000306, -0.00000359, 0.00000404, -0.00000387, 0.00000378, 0.00000299, -0.00000217, 0.00000012, -0.00000364, 0.00001554, -0.00001363, 0.00000962, -0.00000418, -0.00001054, 0.00000995, -0.00001707, 0.00000649, -0.00000824, 0.00000563, -0.00002154, 0.00001608, -0.00002631, 0.00002102, -0.00001211, -0.00001335, 0.00000703, -0.00000340, 0.00002276, 0.00001070, 0.00000463, 0.00002224, -0.00000789, 0.00001255, -0.00001229, -0.00000951, 0.00000901, -0.00001038, 0.00000177, },
		{ -0.00000427, 0.00000078, 0.00000234, 0.00000331, -0.00000010, -0.00000035, -0.00000280, -0.00000160, -0.00000974, 0.00001244, -0.00000031, -0.00000212, -0.00000137, 0.00000033, -0.00000734, 0.00001075, -0.00000304, 0.00000472, -0.00000086, -0.00000456, -0.00000508, 0.00000627, 0.00000266, -0.00000655, -0.00001259, 0.00001394, -0.00000094, 0.00000325, -0.00001179, 0.00001430, -0.00000578, -0.00000068, -0.00001122, 0.00001477, -0.00000920, 0.00000180, -0.00000871, 0.00000827, -0.00000678, 0.00000265, -0.00000920, 0.00001174, -0.00000679, -0.00001406, -0.00001772, 0.00000071, -0.00003083, -0.00007255, -0.00022510, 0.00058514, 0.00013364, -0.00024322, 0.00003174, -0.00007268, 0.00001218, -0.00001673, 0.00001158, -0.00002963, 0.00000744, -0.00000182, 0.00000326, -0.00001649, 0.00000616, -0.00000828, 0.00001224, -0.00002273, -0.00000202, 0.00000304, 0.00001221, -0.00001534, 0.00000370, 0.00000264, 0.00000917, -0.00001981, 0.00000106, 0.00000681, 0.00000648, -0.00000720, 0.00000219, -0.00000249, 0.00000390, -0.00000651, 0.00000763, -0.00000766, 0.00000174, -0.00000356, 0.00000023, 0.00000252, 0.00000997, -0.00001420, 0.00000076, -0.00000225, 0.00000115, 0.00000025, -0.00000170, 0.00000549, -0.00000057, -0.00000009, -0.00000335, 0.00000500, 0.00000776, -0.00001459, 0.00001915, -0.00002295, 0.00001100, -0.00000932, -0.00000301, -0.00000335, 0.00000328, -0.00001163, -0.00000250, -0.00000006, -0.00001534, 0.00002516, -0.00003391, 0.00001231, -0.00000826, 0.00001321, 0.00001781, 0.00000511, 0.00001203, 0.00001274, -0.00000820, 0.00001568, -0.00002683, 0.00001111, -0.00000635, -0.00000126, },
		{ 0.00000338, -0.00000691, 0.00000506, 0.00000118, 0.00000279, -0.00000042, -0.00000219, -0.00000038, -0.00000792, -0.00000032, 0.00001059, -0.00000358, -0.00000216, 0.00000171, -0.00000600, 0.00000134, 0.00000636, -0.00000178, 0.00000423, -0.00000260, -0.00000716, 0.00000173, 0.00000388, 0.00000303, -0.00001698, 0.00000391, 0.00000605, 0.00000384, -0.00000827, 0.00000208, 0.00000624, -0.00000276, -0.00000940, 0.00000336, 0.00000441, -0.00000515, -0.00000434, 0.00000123, -0.00000050, 0.00000062, -0.00000858, 0.00000686, -0.00000192, -0.00000030, -0.00003480, 0.00001632, -0.00004469, 0.00001347, -0.00024186, 0.00010497, 0.00063170, -0.00025569, -0.00004512, -0.00004089, -0.00001957, -0.00001217, 0.00001132, -0.00002096, -0.00000480, -0.00000317, 0.00001128, -0.00001795, 0.00000377, -0.00000935, 0.00001020, -0.00000905, -0.00001212, -0.00000371, 0.00001527, -0.00000553, -0.00000553, 0.00000055, 0.00001147, -0.00000708, -0.00001243, 0.00000423, 0.00001038, -0.00000331, -0.00000110, -0.00000214, 0.00000356, -0.00000402, 0.00000199, 0.00000102, -0.00000496, 0.00000023, -0.00000195, -0.00000059, 0.00000925, -0.00000080, -0.00001154, 0.00000350, -0.00000393, 0.00000344, -0.00000275, 0.00000213, 0.00000415, -0.00000236, 0.00000129, -0.00000543, 0.00001469, -0.00000966, 0.00000306, 0.00000513, -0.00001900, 0.00001617, -0.00001977, 0.00000732, -0.00000984, 0.00000910, -0.00002383, 0.00001784, -0.00002765, 0.00002141, -0.00001046, -0.00001440, 0.00000961, -0.00000666, 0.00002178, 0.00001037, 0.00000367, 0.00001888, -0.00000419, 0.00000731, -0.00000230, -0.00001728, 0.00001349, -0.00001242, },
		{ -0.00000337, -0.00000280, -0.00000071, 0.00000296, 0.00000334, -0.00000050, 0.00000115, -0.00000450, 0.00000095, -0.00001229, 0.00001168, 0.00000130, -0.00000152, -0.00000301, 0.00000314, -0.00001031, 0.00001107, -0.00000274, 0.00000490, -0.00000121, -0.00000227, -0.00000791, 0.00000763, 0.00000083, -0.00000121, -0.00001794, 0.00001600, -0.00000310, 0.00000709, -0.00001575, 0.00001613, -0.00000642, 0.00000228, -0.00001540, 0.00001748, -0.00001141, 0.00000455, -0.00001209, 0.00001191, -0.00001108, 0.00000757, -0.00001546, 0.00001929, -0.00001556, -0.00000108, -0.00003452, 0.00002173, -0.00005740, -0.00001618, -0.00031178, 0.00057977, 0.00022969, -0.00029978, 0.00005600, -0.00009194, 0.00002498, -0.00002971, 0.00002349, -0.00004012, 0.00001649, -0.00001074, 0.00001268, -0.00002494, 0.00001385, -0.00001563, 0.00001826, -0.00002575, 0.00000115, -0.00000345, 0.00001854, -0.00001839, 0.00000565, -0.00000176, 0.00001408, -0.00002093, 0.00000079, 0.00000389, 0.00001010, -0.00000948, 0.00000428, -0.00000506, 0.00000655, -0.00000896, 0.00000986, -0.00000829, 0.00000166, -0.00000420, 0.00000159, -0.00000051, 0.00001278, -0.00001335, -0.00000092, -0.00000163, 0.00000038, 0.00000038, -0.00000119, 0.00000378, 0.00000170, -0.00000165, -0.00000064, -0.00000039, 0.00001340, -0.00001876, 0.00002098, -0.00002004, 0.00000521, -0.00000222, -0.00000852, -0.00000020, 0.00000069, -0.00000676, -0.00000753, 0.00000523, -0.00002121, 0.00003140, -0.00003832, 0.00001716, -0.00001165, 0.00001360, 0.00001403, 0.00000885, 0.00000492, 0.00001755, -0.00001291, 0.00001991, -0.00002477, 0.00000686, -0.00000254, },
		{ -0.00001330, 0.00000421, -0.00000770, 0.00000504, 0.00000124, 0.00000289, -0.00000037, -0.00000139, -0.00000145, -0.00000550, -0.00000466, 0.00001259, -0.00000296, -0.00000230, 0.00000131, -0.00000407, -0.00000246, 0.00000852, -0.00000258, 0.00000458, -0.00000190, -0.00000670, 0.00000002, 0.00000416, 0.00000402, -0.00001476, -0.00000079, 0.00000859, 0.00000230, -0.00000531, -0.00000237, 0.00000945, -0.00000340, -0.00000788, -0.00000006, 0.00000718, -0.00000693, -0.00000293, -0.00000020, 0.00000059, -0.00000030, -0.00000814, 0.00000631, -0.00000125, 0.00000014, -0.00003361, 0.00001303, -0.00003917, 0.00000992, -0.00020642, -0.00002381, 0.00074382, -0.00025753, -0.00006356, -0.00003374, -0.00002589, -0.00001039, 0.00000924, -0.00001851, -0.00000726, -0.00000155, 0.00000983, -0.00001524, 0.00000073, -0.00000615, 0.00000619, -0.00000463, -0.00001372, -0.00000404, 0.00001277, -0.00000084, -0.00000834, 0.00000100, 0.00000936, -0.00000242, -0.00001472, 0.00000316, 0.00000988, -0.00000124, -0.00000255, -0.00000112, 0.00000214, -0.00000216, -0.00000024, 0.00000398, -0.00000682, 0.00000066, -0.00000214, -0.00000043, 0.00000698, 0.00000361, -0.00001356, 0.00000353, -0.00000401, 0.00000273, -0.00000183, 0.00000106, 0.00000441, -0.00000164, 0.00000136, -0.00000535, 0.00001143, -0.00000347, -0.00000506, 0.00001456, -0.00002589, 0.00001981, -0.00001980, 0.00000622, -0.00001011, 0.00001110, -0.00002422, 0.00001762, -0.00002686, 0.00001947, -0.00000652, -0.00001779, 0.00001434, -0.00001168, 0.00002267, 0.00000797, 0.00000547, 0.00001274, 0.00000212, -0.00000049, 0.00000840, -0.00002404, 0.00001654, },
		{ 0.00000239, -0.00000601, -0.00000094, -0.00000236, 0.00000351, 0.00000352, -0.00000111, 0.00000269, -0.00000599, 0.00000297, -0.00001369, 0.00000984, 0.00000309, -0.00000064, -0.00000482, 0.00000576, -0.00001241, 0.00001044, -0.00000208, 0.00000513, -0.00000191, 0.00000032, -0.00001065, 0.00000900, -0.00000148, 0.00000424, -0.00002241, 0.00001709, -0.00000483, 0.00001047, -0.00001877, 0.00001692, -0.00000674, 0.00000538, -0.00001936, 0.00001981, -0.00001353, 0.00000743, -0.00001580, 0.00001612, -0.00001616, 0.00001353, -0.00002308, 0.00002840, -0.00002615, 0.00001371, -0.00005270, 0.00004381, -0.00008497, 0.00003665, -0.00037543, 0.00053364, 0.00035539, -0.00036651, 0.00008414, -0.00011410, 0.00004039, -0.00004520, 0.00003739, -0.00005222, 0.00002686, -0.00002072, 0.00002279, -0.00003371, 0.00002158, -0.00002263, 0.00002372, -0.00002867, 0.00000496, -0.00001028, 0.00002422, -0.00002061, 0.00000746, -0.00000607, 0.00001818, -0.00002122, 0.00000067, 0.00000050, 0.00001368, -0.00001152, 0.00000627, -0.00000754, 0.00000895, -0.00001101, 0.00001142, -0.00000819, 0.00000132, -0.00000496, 0.00000311, -0.00000330, 0.00001454, -0.00001144, -0.00000288, -0.00000116, -0.00000005, -0.00000004, -0.00000009, 0.00000166, 0.00000404, -0.00000325, 0.00000210, -0.00000508, 0.00001714, -0.00002003, 0.00001907, -0.00001347, -0.00000310, 0.00000600, -0.00001405, 0.00000290, -0.00000233, -0.00000132, -0.00001277, 0.00001059, -0.00002693, 0.00003711, -0.00004197, 0.00002090, -0.00001371, 0.00001293, 0.00001123, 0.00001160, -0.00000038, 0.00001987, -0.00001454, 0.00001980, -0.00001883, 0.00000032, },
		{ 0.00001729, -0.00001254, 0.00000400, -0.00000772, 0.00000453, 0.00000142, 0.00000318, -0.00000087, 0.00000018, -0.00000337, -0.00000225, -0.00000911, 0.00001408, -0.00000229, -0.00000180, -0.00000011, -0.00000099, -0.00000678, 0.00001065, -0.00000348, 0.00000534, -0.00000210, -0.00000500, -0.00000272, 0.00000549, 0.00000322, -0.00001035, -0.00000693, 0.00001208, -0.00000054, -0.00000080, -0.00000796, 0.00001323, -0.00000495, -0.00000475, -0.00000506, 0.00001134, -0.00001023, 0.00000022, -0.00000363, 0.00000405, -0.00000401, -0.00000437, 0.00000162, 0.00000446, -0.00000578, -0.00002447, 0.00000015, -0.00002157, -0.00001153, -0.00014565, -0.00016695, 0.00084098, -0.00023287, -0.00010022, -0.00001477, -0.00004121, -0.00000105, 0.00000022, -0.00000981, -0.00001539, 0.00000529, 0.00000347, -0.00000803, -0.00000638, 0.00000071, -0.00000087, 0.00000204, -0.00001736, -0.00000165, 0.00000727, 0.00000598, -0.00001250, 0.00000306, 0.00000535, 0.00000339, -0.00001741, 0.00000300, 0.00000785, 0.00000215, -0.00000504, 0.00000099, -0.00000046, 0.00000080, -0.00000346, 0.00000753, -0.00000880, 0.00000130, -0.00000291, 0.00000071, 0.00000370, 0.00000817, -0.00001499, 0.00000316, -0.00000399, 0.00000214, -0.00000137, 0.00000081, 0.00000361, 0.00000007, 0.00000028, -0.00000355, 0.00000632, 0.00000391, -0.00001332, 0.00002245, -0.00002979, 0.00001999, -0.00001681, 0.00000309, -0.00000890, 0.00001136, -0.00002248, 0.00001530, -0.00002389, 0.00001526, -0.00000045, -0.00002322, 0.00002082, -0.00001801, 0.00002495, 0.00000396, 0.00000939, 0.00000473, 0.00000979, -0.00000911, 0.00001794, -0.00002834, },
		{ -0.00000751, 0.00000756, -0.00000856, 0.00000083, -0.00000383, 0.00000383, 0.00000388, -0.00000176, 0.00000398, -0.00000692, 0.00000408, -0.00001373, 0.00000693, 0.00000497, 0.00000039, -0.00000646, 0.00000773, -0.00001326, 0.00000873, -0.00000102, 0.00000529, -0.00000269, 0.00000276, -0.00001284, 0.00000992, -0.00000365, 0.00000898, -0.00002522, 0.00001667, -0.00000562, 0.00001280, -0.00002025, 0.00001629, -0.00000643, 0.00000811, -0.00002245, 0.00002114, -0.00001491, 0.00000973, -0.00001906, 0.00001999, -0.00002093, 0.00001923, -0.00003045, 0.00003708, -0.00003605, 0.00002719, -0.00006843, 0.00006220, -0.00010711, 0.00007699, -0.00040691, 0.00044357, 0.00050656, -0.00043698, 0.00011100, -0.00013520, 0.00005517, -0.00006026, 0.00005058, -0.00006343, 0.00003623, -0.00002959, 0.00003155, -0.00004094, 0.00002766, -0.00002784, 0.00002747, -0.00003050, 0.00000831, -0.00001620, 0.00002811, -0.00002120, 0.00000843, -0.00000955, 0.00002083, -0.00002034, 0.00000043, -0.00000288, 0.00001665, -0.00001283, 0.00000772, -0.00000946, 0.00001061, -0.00001220, 0.00001194, -0.00000717, 0.00000064, -0.00000568, 0.00000447, -0.00000544, 0.00001499, -0.00000851, -0.00000493, -0.00000093, -0.00000013, -0.00000088, 0.00000132, -0.00000047, 0.00000600, -0.00000437, 0.00000421, -0.00000829, 0.00001831, -0.00001806, 0.00001365, -0.00000419, -0.00001254, 0.00001389, -0.00001846, 0.00000509, -0.00000500, 0.00000378, -0.00001725, 0.00001503, -0.00003148, 0.00004127, -0.00004387, 0.00002258, -0.00001359, 0.00001042, 0.00001020, 0.00001245, -0.00000289, 0.00001882, -0.00001242, 0.00001536, -0.00000977, },
		{ -0.00002911, 0.00001531, -0.00001005, 0.00000277, -0.00000703, 0.00000365, 0.00000160, 0.00000377, -0.00000194, 0.00000240, -0.00000590, 0.00000142, -0.00001323, 0.00001486, -0.00000163, -0.00000063, -0.00000245, 0.00000291, -0.00001122, 0.00001250, -0.00000446, 0.00000657, -0.00000325, -0.00000212, -0.00000640, 0.00000784, 0.00000069, -0.00000412, -0.00001399, 0.00001613, -0.00000440, 0.00000489, -0.00001421, 0.00001729, -0.00000738, -0.00000004, -0.00001152, 0.00001678, -0.00001503, 0.00000514, -0.00000920, 0.00001009, -0.00001078, 0.00000309, -0.00000756, 0.00001552, -0.00001822, -0.00000752, -0.00002177, 0.00000699, -0.00004818, -0.00006683, -0.00031125, 0.00090944, -0.00017347, -0.00015854, 0.00001771, -0.00006671, 0.00001685, -0.00001651, 0.00000566, -0.00002952, 0.00001751, -0.00000779, 0.00000357, -0.00001730, 0.00001088, -0.00001058, 0.00001071, -0.00002299, 0.00000343, -0.00000090, 0.00001441, -0.00001763, 0.00000653, -0.00000023, 0.00000996, -0.00002036, 0.00000383, 0.00000427, 0.00000675, -0.00000847, 0.00000411, -0.00000409, 0.00000465, -0.00000739, 0.00001136, -0.00001070, 0.00000213, -0.00000431, 0.00000274, -0.00000030, 0.00001249, -0.00001566, 0.00000248, -0.00000403, 0.00000189, -0.00000160, 0.00000151, 0.00000179, 0.00000264, -0.00000172, -0.00000050, 0.00000023, 0.00001125, -0.00002026, 0.00002742, -0.00002977, 0.00001643, -0.00001100, -0.00000169, -0.00000647, 0.00001004, -0.00001884, 0.00001123, -0.00001917, 0.00000933, 0.00000707, -0.00002994, 0.00002820, -0.00002475, 0.00002775, -0.00000079, 0.00001438, -0.00000381, 0.00001723, -0.00001665, 0.00002454, },
		{ 0.00000099, -0.00001525, 0.00001192, -0.00001032, 0.00000209, -0.00000486, 0.00000379, 0.00000437, -0.00000228, 0.00000477, -0.00000700, 0.00000405, -0.00001233, 0.00000313, 0.00000684, 0.00000143, -0.00000759, 0.00000865, -0.00001263, 0.00000591, 0.00000045, 0.00000522, -0.00000319, 0.00000458, -0.00001400, 0.00000993, -0.00000505, 0.00001224, -0.00002573, 0.00001434, -0.00000507, 0.00001356, -0.00001976, 0.00001395, -0.00000521, 0.00000994, -0.00002398, 0.00002080, -0.00001486, 0.00001067, -0.00002096, 0.00002241, -0.00002409, 0.00002311, -0.00003567, 0.00004303, -0.00004251, 0.00003603, -0.00007781, 0.00007226, -0.00011795, 0.00009752, -0.00040069, 0.00031134, 0.00067450, -0.00050179, 0.00012966, -0.00015016, 0.00006516, -0.00007126, 0.00005980, -0.00007083, 0.00004195, -0.00003496, 0.00003683, -0.00004473, 0.00003041, -0.00002985, 0.00002836, -0.00003025, 0.00001009, -0.00001997, 0.00002918, -0.00001948, 0.00000800, -0.00001154, 0.00002152, -0.00001802, -0.00000026, -0.00000571, 0.00001841, -0.00001293, 0.00000817, -0.00001035, 0.00001108, -0.00001215, 0.00001115, -0.00000514, -0.00000043, -0.00000617, 0.00000535, -0.00000659, 0.00001401, -0.00000474, -0.00000688, -0.00000100, 0.00000004, -0.00000190, 0.00000265, -0.00000216, 0.00000714, -0.00000459, 0.00000521, -0.00000955, 0.00001673, -0.00001314, 0.00000562, 0.00000634, -0.00002143, 0.00001995, -0.00002064, 0.00000563, -0.00000660, 0.00000772, -0.00002012, 0.00001772, -0.00003406, 0.00004313, -0.00004333, 0.00002160, -0.00001080, 0.00000569, 0.00001131, 0.00001093, -0.00000221, 0.00001424, -0.00000675, 0.00000744, },
		{ 0.00002698, -0.00002594, 0.00001071, -0.00000614, 0.00000082, -0.00000591, 0.00000263, 0.00000159, 0.00000474, -0.00000351, 0.00000504, -0.00000868, 0.00000504, -0.00001659, 0.00001477, -0.00000098, 0.00000115, -0.00000546, 0.00000719, -0.00001531, 0.00001381, -0.00000543, 0.00000824, -0.00000524, 0.00000168, -0.00001072, 0.00001095, -0.00000317, 0.00000329, -0.00002123, 0.00002017, -0.00000877, 0.00001116, -0.00002050, 0.00002123, -0.00001047, 0.00000595, -0.00001902, 0.00002309, -0.00002096, 0.00001154, -0.00001662, 0.00001840, -0.00002025, 0.00001379, -0.00002065, 0.00003110, -0.00003597, 0.00001546, -0.00005025, 0.00004306, -0.00009444, 0.00001988, -0.00044208, 0.00093653, -0.00007383, -0.00023912, 0.00006308, -0.00010173, 0.00004266, -0.00004022, 0.00002710, -0.00004876, 0.00003417, -0.00002297, 0.00001851, -0.00003100, 0.00002331, -0.00002199, 0.00002064, -0.00003007, 0.00001059, -0.00001089, 0.00002355, -0.00002308, 0.00001089, -0.00000682, 0.00001676, -0.00002332, 0.00000558, -0.00000064, 0.00001220, -0.00001251, 0.00000793, -0.00000843, 0.00000901, -0.00001160, 0.00001506, -0.00001228, 0.00000310, -0.00000626, 0.00000545, -0.00000461, 0.00001618, -0.00001545, 0.00000159, -0.00000431, 0.00000213, -0.00000257, 0.00000308, -0.00000083, 0.00000571, -0.00000420, 0.00000311, -0.00000581, 0.00001734, -0.00002466, 0.00002858, -0.00002564, 0.00000957, -0.00000321, -0.00000728, -0.00000344, 0.00000769, -0.00001394, 0.00000617, -0.00001353, 0.00000261, 0.00001504, -0.00003686, 0.00003533, -0.00003080, 0.00003001, -0.00000522, 0.00001915, -0.00001134, 0.00002277, -0.00002139, },
		{ -0.00000237, 0.00001162, -0.00002140, 0.00001450, -0.00001074, 0.00000251, -0.00000529, 0.00000338, 0.00000486, -0.00000245, 0.00000481, -0.00000605, 0.00000279, -0.00000961, -0.00000131, 0.00000856, 0.00000231, -0.00000791, 0.00000827, -0.00001044, 0.00000210, 0.00000228, 0.00000476, -0.00000310, 0.00000536, -0.00001372, 0.00000866, -0.00000518, 0.00001343, -0.00002359, 0.00000995, -0.00000298, 0.00001245, -0.00001704, 0.00000982, -0.00000286, 0.00001038, -0.00002336, 0.00001821, -0.00001276, 0.00000953, -0.00002060, 0.00002232, -0.00002434, 0.00002364, -0.00003692, 0.00004408, -0.00004301, 0.00003736, -0.00007766, 0.00007043, -0.00011326, 0.00009385, -0.00035585, 0.00014411, 0.00084637, -0.00054949, 0.00013241, -0.00015356, 0.00006603, -0.00007447, 0.00006179, -0.00007158, 0.00004156, -0.00003467, 0.00003674, -0.00004347, 0.00002849, -0.00002756, 0.00002553, -0.00002708, 0.00000931, -0.00002057, 0.00002672, -0.00001507, 0.00000577, -0.00001157, 0.00001990, -0.00001408, -0.00000167, -0.00000746, 0.00001845, -0.00001141, 0.00000728, -0.00000985, 0.00001004, -0.00001058, 0.00000892, -0.00000211, -0.00000190, -0.00000626, 0.00000547, -0.00000652, 0.00001160, -0.00000036, -0.00000854, -0.00000133, 0.00000028, -0.00000280, 0.00000354, -0.00000303, 0.00000715, -0.00000365, 0.00000484, -0.00000876, 0.00001269, -0.00000606, -0.00000368, 0.00001638, -0.00002811, 0.00002297, -0.00001989, 0.00000409, -0.00000671, 0.00000996, -0.00002083, 0.00001822, -0.00003427, 0.00004236, -0.00004016, 0.00001789, -0.00000543, -0.00000110, 0.00001441, 0.00000715, 0.00000142, 0.00000669, 0.00000149, },
		{ -0.00002225, 0.00002480, -0.00001915, 0.00000420, -0.00000151, -0.00000129, -0.00000473, 0.00000175, 0.00000125, 0.00000609, -0.00000541, 0.00000780, -0.00001128, 0.00000811, -0.00001882, 0.00001372, -0.00000037, 0.00000339, -0.00000875, 0.00001131, -0.00001858, 0.00001436, -0.00000629, 0.00001021, -0.00000779, 0.00000597, -0.00001520, 0.00001436, -0.00000777, 0.00001099, -0.00002777, 0.00002353, -0.00001305, 0.00001727, -0.00002613, 0.00002454, -0.00001385, 0.00001267, -0.00002684, 0.00002958, -0.00002737, 0.00001872, -0.00002514, 0.00002813, -0.00003142, 0.00002652, -0.00003612, 0.00004928, -0.00005657, 0.00004130, -0.00008124, 0.00008139, -0.00014278, 0.00010288, -0.00054532, 0.00091262, 0.00006742, -0.00033891, 0.00011797, -0.00014341, 0.00007391, -0.00006862, 0.00005226, -0.00007100, 0.00005324, -0.00004012, 0.00003494, -0.00004572, 0.00003639, -0.00003366, 0.00003063, -0.00003762, 0.00001880, -0.00002144, 0.00003222, -0.00002801, 0.00001547, -0.00001365, 0.00002311, -0.00002591, 0.00000796, -0.00000641, 0.00001796, -0.00001664, 0.00001198, -0.00001297, 0.00001334, -0.00001555, 0.00001817, -0.00001329, 0.00000408, -0.00000859, 0.00000849, -0.00000875, 0.00001888, -0.00001434, 0.00000066, -0.00000495, 0.00000291, -0.00000422, 0.00000529, -0.00000386, 0.00000883, -0.00000662, 0.00000655, -0.00001086, 0.00002122, -0.00002580, 0.00002574, -0.00001794, 0.00000055, 0.00000526, -0.00001254, -0.00000069, 0.00000509, -0.00000869, 0.00000113, -0.00000802, -0.00000379, 0.00002229, -0.00004279, 0.00004104, -0.00003504, 0.00003071, -0.00000827, 0.00002247, -0.00001652, 0.00002511, },
		{ 0.00001074, -0.00001214, 0.00002024, -0.00002469, 0.00001465, -0.00000955, 0.00000203, -0.00000515, 0.00000274, 0.00000517, -0.00000207, 0.00000396, -0.00000402, 0.00000040, -0.00000584, -0.00000601, 0.00000998, 0.00000291, -0.00000722, 0.00000650, -0.00000683, -0.00000246, 0.00000437, 0.00000381, -0.00000216, 0.00000480, -0.00001181, 0.00000589, -0.00000373, 0.00001224, -0.00001874, 0.00000365, 0.00000059, 0.00000939, -0.00001214, 0.00000402, 0.00000067, 0.00000906, -0.00002015, 0.00001303, -0.00000824, 0.00000579, -0.00001734, 0.00001887, -0.00002069, 0.00001964, -0.00003283, 0.00003872, -0.00003587, 0.00002940, -0.00006620, 0.00005497, -0.00009137, 0.00006531, -0.00027643, -0.00004608, 0.00100644, -0.00056794, 0.00011200, -0.00014063, 0.00005415, -0.00006686, 0.00005402, -0.00006358, 0.00003328, -0.00002724, 0.00003010, -0.00003621, 0.00002118, -0.00002047, 0.00001859, -0.00002059, 0.00000542, -0.00001746, 0.00002051, -0.00000799, 0.00000166, -0.00000944, 0.00001591, -0.00000857, -0.00000398, -0.00000772, 0.00001638, -0.00000806, 0.00000483, -0.00000775, 0.00000732, -0.00000746, 0.00000533, 0.00000177, -0.00000371, -0.00000582, 0.00000463, -0.00000511, 0.00000793, 0.00000430, -0.00000978, -0.00000180, 0.00000034, -0.00000327, 0.00000365, -0.00000280, 0.00000588, -0.00000152, 0.00000315, -0.00000619, 0.00000688, 0.00000198, -0.00001268, 0.00002425, -0.00003127, 0.00002222, -0.00001601, 0.00000053, -0.00000528, 0.00001037, -0.00001933, 0.00001657, -0.00003224, 0.00003921, -0.00003471, 0.00001194, 0.00000192, -0.00000929, 0.00001886, 0.00000173, 0.00000717, -0.00000263, },
		{ 0.00002356, -0.00001891, 0.00001844, -0.00000984, -0.00000304, 0.00000289, -0.00000293, -0.00000391, 0.00000125, 0.00000050, 0.00000772, -0.00000739, 0.00001029, -0.00001327, 0.00001022, -0.00001965, 0.00001170, 0.00000020, 0.00000586, -0.00001190, 0.00001474, -0.00002061, 0.00001395, -0.00000691, 0.00001222, -0.00001049, 0.00001019, -0.00001924, 0.00001747, -0.00001231, 0.00001801, -0.00003273, 0.00002555, -0.00001659, 0.00002247, -0.00003036, 0.00002674, -0.00001702, 0.00001939, -0.00003406, 0.00003532, -0.00003331, 0.00002567, -0.00003362, 0.00003793, -0.00004270, 0.00003940, -0.00005170, 0.00006729, -0.00007665, 0.00006587, -0.00010976, 0.00011581, -0.00018495, 0.00017085, -0.00060934, 0.00083277, 0.00024650, -0.00045086, 0.00017622, -0.00018688, 0.00010647, -0.00009796, 0.00007775, -0.00009309, 0.00007183, -0.00005657, 0.00005041, -0.00005923, 0.00004810, -0.00004385, 0.00003926, -0.00004439, 0.00002669, -0.00003109, 0.00003915, -0.00003152, 0.00001948, -0.00001989, 0.00002832, -0.00002770, 0.00001056, -0.00001243, 0.00002330, -0.00002025, 0.00001569, -0.00001709, 0.00001703, -0.00001868, 0.00002027, -0.00001351, 0.00000495, -0.00001104, 0.00001146, -0.00001225, 0.00002030, -0.00001237, -0.00000017, -0.00000599, 0.00000413, -0.00000633, 0.00000776, -0.00000682, 0.00001147, -0.00000843, 0.00000913, -0.00001419, 0.00002232, -0.00002355, 0.00001942, -0.00000790, -0.00000904, 0.00001284, -0.00001624, 0.00000086, 0.00000311, -0.00000406, -0.00000285, -0.00000368, -0.00000883, 0.00002777, -0.00004673, 0.00004439, -0.00003665, 0.00002911, -0.00000922, 0.00002349, -0.00001845, },
		{ -0.00001215, 0.00001918, -0.00001992, 0.00002531, -0.00002434, 0.00001219, -0.00000688, 0.00000085, -0.00000465, 0.00000207, 0.00000508, -0.00000099, 0.00000218, -0.00000103, -0.00000287, -0.00000144, -0.00001056, 0.00001096, 0.00000313, -0.00000547, 0.00000343, -0.00000210, -0.00000740, 0.00000655, 0.00000232, -0.00000024, 0.00000280, -0.00000824, 0.00000165, -0.00000065, 0.00000872, -0.00001151, -0.00000411, 0.00000537, 0.00000460, -0.00000534, -0.00000313, 0.00000529, 0.00000585, -0.00001425, 0.00000524, -0.00000125, -0.00000068, -0.00001092, 0.00001172, -0.00001270, 0.00001064, -0.00002288, 0.00002641, -0.00002064, 0.00001187, -0.00004351, 0.00002643, -0.00005351, 0.00001508, -0.00017104, -0.00024355, 0.00113788, -0.00054617, 0.00006304, -0.00010843, 0.00002741, -0.00004680, 0.00003524, -0.00004595, 0.00001652, -0.00001233, 0.00001673, -0.00002296, 0.00000859, -0.00000878, 0.00000781, -0.00001090, -0.00000163, -0.00001065, 0.00001086, 0.00000132, -0.00000406, -0.00000524, 0.00000975, -0.00000170, -0.00000723, -0.00000627, 0.00001211, -0.00000290, 0.00000088, -0.00000406, 0.00000301, -0.00000295, 0.00000061, 0.00000621, -0.00000574, -0.00000476, 0.00000273, -0.00000243, 0.00000328, 0.00000890, -0.00001051, -0.00000227, -0.00000001, -0.00000306, 0.00000277, -0.00000138, 0.00000341, 0.00000162, 0.00000045, -0.00000242, 0.00000029, 0.00000970, -0.00001991, 0.00002868, -0.00003023, 0.00001772, -0.00000948, -0.00000450, -0.00000275, 0.00000929, -0.00001602, 0.00001331, -0.00002859, 0.00003440, -0.00002782, 0.00000468, 0.00001020, -0.00001781, 0.00002365, -0.00000431, 0.00001377, },
		{ -0.00001690, 0.00001821, -0.00001195, 0.00000915, 0.00000039, -0.00000959, 0.00000606, -0.00000351, -0.00000377, 0.00000128, -0.00000066, 0.00000945, -0.00000913, 0.00001213, -0.00001429, 0.00001106, -0.00001898, 0.00000883, 0.00000070, 0.00000829, -0.00001443, 0.00001700, -0.00002111, 0.00001250, -0.00000718, 0.00001398, -0.00001285, 0.00001371, -0.00002224, 0.00001966, -0.00001602, 0.00002342, -0.00003533, 0.00002569, -0.00001881, 0.00002605, -0.00003258, 0.00002734, -0.00001942, 0.00002522, -0.00003966, 0.00003929, -0.00003773, 0.00003122, -0.00004069, 0.00004619, -0.00005219, 0.00005020, -0.00006476, 0.00008203, -0.00009253, 0.00008484, -0.00013077, 0.00014034, -0.00021339, 0.00021447, -0.00062671, 0.00069789, 0.00045422, -0.00056429, 0.00022947, -0.00022572, 0.00013504, -0.00012363, 0.00009942, -0.00011138, 0.00008666, -0.00006939, 0.00006230, -0.00006923, 0.00005642, -0.00005084, 0.00004508, -0.00004904, 0.00003284, -0.00003837, 0.00004319, -0.00003283, 0.00002220, -0.00002474, 0.00003172, -0.00002826, 0.00001288, -0.00001795, 0.00002745, -0.00002269, 0.00001843, -0.00002018, 0.00001948, -0.00002049, 0.00002100, -0.00001280, 0.00000556, -0.00001332, 0.00001392, -0.00001469, 0.00002029, -0.00000968, -0.00000077, -0.00000737, 0.00000560, -0.00000855, 0.00001005, -0.00000921, 0.00001318, -0.00000918, 0.00001041, -0.00001542, 0.00002063, -0.00001840, 0.00001077, 0.00000277, -0.00001742, 0.00001806, -0.00001735, 0.00000052, 0.00000243, -0.00000085, -0.00000497, -0.00000126, -0.00001178, 0.00003083, -0.00004810, 0.00004489, -0.00003530, 0.00002500, -0.00000782, 0.00002186, },
		{ 0.00001985, -0.00002024, 0.00002514, -0.00002419, 0.00002594, -0.00002028, 0.00000752, -0.00000324, -0.00000058, -0.00000413, 0.00000167, 0.00000441, 0.00000085, -0.00000043, 0.00000265, -0.00000663, 0.00000309, -0.00001454, 0.00001139, 0.00000294, -0.00000271, -0.00000065, 0.00000328, -0.00001229, 0.00000862, 0.00000032, 0.00000263, -0.00000053, -0.00000327, -0.00000377, 0.00000381, 0.00000325, -0.00000257, -0.00001262, 0.00001081, -0.00000148, 0.00000277, -0.00001106, 0.00001069, 0.00000084, -0.00000589, -0.00000475, 0.00000783, -0.00000960, -0.00000158, 0.00000111, -0.00000067, -0.00000303, -0.00000754, 0.00000784, 0.00000169, -0.00001378, -0.00001167, -0.00001233, -0.00000371, -0.00005025, -0.00005171, -0.00043083, 0.00122487, -0.00047613, -0.00001662, -0.00005667, -0.00001402, -0.00001460, 0.00000599, -0.00001941, -0.00000786, 0.00000914, -0.00000237, -0.00000473, -0.00000824, 0.00000648, -0.00000593, 0.00000128, -0.00001132, -0.00000077, -0.00000135, 0.00001200, -0.00001085, 0.00000061, 0.00000189, 0.00000610, -0.00001128, -0.00000310, 0.00000581, 0.00000375, -0.00000430, 0.00000095, -0.00000257, 0.00000256, -0.00000478, 0.00001084, -0.00000783, -0.00000310, -0.00000019, 0.00000130, -0.00000194, 0.00001310, -0.00001068, -0.00000256, -0.00000096, -0.00000200, 0.00000085, 0.00000112, -0.00000000, 0.00000537, -0.00000276, 0.00000181, -0.00000605, 0.00001590, -0.00002425, 0.00002902, -0.00002511, 0.00001022, -0.00000136, -0.00000997, 0.00000011, 0.00000740, -0.00001173, 0.00000937, -0.00002430, 0.00002900, -0.00002063, -0.00000269, 0.00001817, -0.00002546, 0.00002764, -0.00000981, },
		{ 0.00001784, -0.00001232, 0.00000991, -0.00000270, -0.00000125, 0.00000967, -0.00001408, 0.00000721, -0.00000265, -0.00000443, 0.00000183, -0.00000206, 0.00001101, -0.00001029, 0.00001299, -0.00001406, 0.00001050, -0.00001688, 0.00000531, 0.00000112, 0.00001042, -0.00001596, 0.00001777, -0.00001995, 0.00001003, -0.00000699, 0.00001520, -0.00001440, 0.00001599, -0.00002366, 0.00002037, -0.00001820, 0.00002648, -0.00003506, 0.00002363, -0.00001930, 0.00002750, -0.00003234, 0.00002602, -0.00002054, 0.00002932, -0.00004265, 0.00004055, -0.00003965, 0.00003420, -0.00004498, 0.00005130, -0.00005802, 0.00005675, -0.00007278, 0.00009058, -0.00010089, 0.00009446, -0.00014008, 0.00015024, -0.00022249, 0.00022783, -0.00059536, 0.00051509, 0.00067654, -0.00066586, 0.00026819, -0.00025297, 0.00015403, -0.00014084, 0.00011316, -0.00012227, 0.00009463, -0.00007589, 0.00006828, -0.00007370, 0.00005965, -0.00005322, 0.00004690, -0.00005044, 0.00003602, -0.00004209, 0.00004349, -0.00003140, 0.00002307, -0.00002754, 0.00003279, -0.00002722, 0.00001440, -0.00002221, 0.00002970, -0.00002336, 0.00001968, -0.00002169, 0.00002021, -0.00002062, 0.00002017, -0.00001109, 0.00000580, -0.00001512, 0.00001548, -0.00001575, 0.00001878, -0.00000643, -0.00000108, -0.00000894, 0.00000702, -0.00001049, 0.00001171, -0.00001060, 0.00001361, -0.00000862, 0.00001019, -0.00001457, 0.00001657, -0.00001133, 0.00000133, 0.00001226, -0.00002297, 0.00001983, -0.00001530, -0.00000202, 0.00000339, 0.00000050, -0.00000482, -0.00000106, -0.00001240, 0.00003131, -0.00004685, 0.00004262, -0.00003122, 0.00001870, -0.00000437, },
		{ -0.00001376, 0.00002422, -0.00002561, 0.00002742, -0.00002415, 0.00002209, -0.00001319, 0.00000154, 0.00000055, -0.00000167, -0.00000399, 0.00000179, 0.00000304, 0.00000339, -0.00000363, 0.00000664, -0.00001039, 0.00000724, -0.00001757, 0.00001117, 0.00000232, 0.00000086, -0.00000533, 0.00000877, -0.00001667, 0.00001038, -0.00000206, 0.00000628, -0.00000487, 0.00000264, -0.00000988, 0.00000913, -0.00000348, 0.00000717, -0.00002098, 0.00001626, -0.00000821, 0.00001146, -0.00001907, 0.00001643, -0.00000555, 0.00000430, -0.00001615, 0.00001821, -0.00002019, 0.00000988, -0.00001209, 0.00001438, -0.00002011, 0.00001169, -0.00001513, 0.00002878, -0.00004459, 0.00002556, -0.00005653, 0.00005182, -0.00012173, 0.00006777, -0.00059087, 0.00125488, -0.00035443, -0.00012499, 0.00001167, -0.00006744, 0.00002722, -0.00003129, 0.00001368, -0.00003761, 0.00003500, -0.00002514, 0.00001651, -0.00002746, 0.00002364, -0.00002113, 0.00001473, -0.00002257, 0.00001104, -0.00001483, 0.00002291, -0.00001792, 0.00000748, -0.00000696, 0.00001422, -0.00001579, 0.00000151, -0.00000206, 0.00001132, -0.00001020, 0.00000678, -0.00000887, 0.00000849, -0.00001030, 0.00001522, -0.00000978, -0.00000092, -0.00000393, 0.00000573, -0.00000725, 0.00001657, -0.00001032, -0.00000248, -0.00000264, -0.00000011, -0.00000199, 0.00000442, -0.00000392, 0.00000923, -0.00000589, 0.00000574, -0.00001123, 0.00001973, -0.00002518, 0.00002538, -0.00001681, 0.00000112, 0.00000686, -0.00001462, 0.00000231, 0.00000564, -0.00000749, 0.00000587, -0.00002049, 0.00002416, -0.00001434, -0.00000896, 0.00002465, -0.00003114, 0.00002986, },
		{ 0.00000035, 0.00001221, -0.00000571, 0.00000007, 0.00000705, -0.00001076, 0.00001629, -0.00001548, 0.00000593, -0.00000034, -0.00000576, 0.00000269, -0.00000343, 0.00001209, -0.00001059, 0.00001266, -0.00001251, 0.00000859, -0.00001359, 0.00000144, 0.00000141, 0.00001199, -0.00001619, 0.00001687, -0.00001717, 0.00000666, -0.00000632, 0.00001560, -0.00001475, 0.00001660, -0.00002312, 0.00001925, -0.00001838, 0.00002674, -0.00003174, 0.00001934, -0.00001792, 0.00002652, -0.00002946, 0.00002262, -0.00001996, 0.00003097, -0.00004228, 0.00003843, -0.00003830, 0.00003372, -0.00004539, 0.00005194, -0.00005867, 0.00005732, -0.00007379, 0.00009080, -0.00009940, 0.00009224, -0.00013515, 0.00014291, -0.00020960, 0.00020929, -0.00051892, 0.00029716, 0.00089584, -0.00074117, 0.00028307, -0.00026226, 0.00015850, -0.00014550, 0.00011556, -0.00012298, 0.00009340, -0.00007413, 0.00006676, -0.00007133, 0.00005674, -0.00005014, 0.00004401, -0.00004787, 0.00003540, -0.00004150, 0.00003966, -0.00002706, 0.00002179, -0.00002787, 0.00003124, -0.00002436, 0.00001468, -0.00002457, 0.00002947, -0.00002188, 0.00001907, -0.00002125, 0.00001894, -0.00001888, 0.00001774, -0.00000842, 0.00000556, -0.00001615, 0.00001579, -0.00001524, 0.00001588, -0.00000284, -0.00000108, -0.00001046, 0.00000805, -0.00001176, 0.00001235, -0.00001071, 0.00001263, -0.00000676, 0.00000860, -0.00001202, 0.00001098, -0.00000362, -0.00000729, 0.00001897, -0.00002461, 0.00001769, -0.00001015, -0.00000651, 0.00000585, 0.00000006, -0.00000254, -0.00000282, -0.00001102, 0.00002964, -0.00004352, 0.00003823, -0.00002518, 0.00001102, },
		{ 0.00002967, -0.00001552, 0.00002608, -0.00002749, 0.00002559, -0.00001988, 0.00001459, -0.00000437, -0.00000447, 0.00000352, -0.00000183, -0.00000457, 0.00000258, 0.00000098, 0.00000643, -0.00000705, 0.00001049, -0.00001365, 0.00001056, -0.00001940, 0.00001031, 0.00000132, 0.00000494, -0.00001008, 0.00001379, -0.00002011, 0.00001161, -0.00000464, 0.00001038, -0.00000972, 0.00000885, -0.00001600, 0.00001464, -0.00001062, 0.00001664, -0.00002825, 0.00002100, -0.00001487, 0.00001986, -0.00002636, 0.00002190, -0.00001269, 0.00001538, -0.00002782, 0.00002876, -0.00003130, 0.00002220, -0.00002643, 0.00003076, -0.00003864, 0.00003249, -0.00003970, 0.00005727, -0.00007652, 0.00006325, -0.00010026, 0.00010573, -0.00018947, 0.00017379, -0.00070927, 0.00122048, -0.00018335, -0.00025551, 0.00009036, -0.00012762, 0.00007414, -0.00007250, 0.00004952, -0.00006924, 0.00006206, -0.00004866, 0.00003813, -0.00004671, 0.00004056, -0.00003596, 0.00002788, -0.00003400, 0.00002331, -0.00002804, 0.00003276, -0.00002439, 0.00001458, -0.00001595, 0.00002196, -0.00002031, 0.00000709, -0.00001078, 0.00001901, -0.00001610, 0.00001277, -0.00001521, 0.00001417, -0.00001534, 0.00001891, -0.00001138, 0.00000163, -0.00000820, 0.00001039, -0.00001216, 0.00001903, -0.00000948, -0.00000193, -0.00000502, 0.00000249, -0.00000545, 0.00000808, -0.00000782, 0.00001264, -0.00000839, 0.00000875, -0.00001464, 0.00002080, -0.00002282, 0.00001860, -0.00000684, -0.00000782, 0.00001357, -0.00001723, 0.00000295, 0.00000488, -0.00000428, 0.00000382, -0.00001815, 0.00002088, -0.00000990, -0.00001322, 0.00002877, -0.00003413, },
		{ 0.00000311, 0.00000528, 0.00000606, 0.00000156, -0.00000960, 0.00001543, -0.00001756, 0.00001900, -0.00001335, 0.00000233, 0.00000308, -0.00000738, 0.00000351, -0.00000445, 0.00001239, -0.00000981, 0.00001107, -0.00000972, 0.00000556, -0.00000949, -0.00000245, 0.00000154, 0.00001283, -0.00001499, 0.00001436, -0.00001303, 0.00000266, -0.00000516, 0.00001500, -0.00001364, 0.00001532, -0.00002050, 0.00001618, -0.00001638, 0.00002412, -0.00002557, 0.00001311, -0.00001474, 0.00002313, -0.00002402, 0.00001725, -0.00001750, 0.00002970, -0.00003815, 0.00003263, -0.00003337, 0.00002929, -0.00004129, 0.00004737, -0.00005331, 0.00005102, -0.00006686, 0.00008173, -0.00008717, 0.00007748, -0.00011561, 0.00011838, -0.00017541, 0.00016166, -0.00040628, 0.00006121, 0.00109294, -0.00077672, 0.00026668, -0.00024909, 0.00014517, -0.00013508, 0.00010469, -0.00011204, 0.00008192, -0.00006335, 0.00005720, -0.00006180, 0.00004751, -0.00004155, 0.00003638, -0.00004119, 0.00003072, -0.00003648, 0.00003187, -0.00002004, 0.00001835, -0.00002559, 0.00002704, -0.00001965, 0.00001344, -0.00002458, 0.00002648, -0.00001810, 0.00001646, -0.00001874, 0.00001563, -0.00001537, 0.00001389, -0.00000496, 0.00000482, -0.00001619, 0.00001467, -0.00001315, 0.00001180, 0.00000085, -0.00000080, -0.00001168, 0.00000836, -0.00001204, 0.00001174, -0.00000944, 0.00001031, -0.00000382, 0.00000600, -0.00000841, 0.00000484, 0.00000343, -0.00001369, 0.00002185, -0.00002198, 0.00001197, -0.00000263, -0.00001218, 0.00000921, -0.00000167, 0.00000125, -0.00000580, -0.00000844, 0.00002669, -0.00003907, 0.00003276, -0.00001832, },
		{ -0.00003424, 0.00002701, -0.00001499, 0.00002525, -0.00002584, 0.00002000, -0.00001231, 0.00000496, 0.00000446, -0.00000912, 0.00000479, -0.00000059, -0.00000603, 0.00000405, -0.00000160, 0.00000965, -0.00001027, 0.00001371, -0.00001600, 0.00001269, -0.00001990, 0.00000887, -0.00000001, 0.00000916, -0.00001438, 0.00001781, -0.00002229, 0.00001216, -0.00000719, 0.00001450, -0.00001448, 0.00001464, -0.00002140, 0.00001958, -0.00001723, 0.00002479, -0.00003357, 0.00002437, -0.00002072, 0.00002712, -0.00003217, 0.00002647, -0.00001979, 0.00002618, -0.00003842, 0.00003815, -0.00004158, 0.00003386, -0.00004015, 0.00004641, -0.00005619, 0.00005203, -0.00006260, 0.00008333, -0.00010508, 0.00009620, -0.00013743, 0.00015076, -0.00024423, 0.00025484, -0.00077615, 0.00112066, 0.00002876, -0.00039735, 0.00017044, -0.00018751, 0.00012023, -0.00011242, 0.00008348, -0.00009864, 0.00008666, -0.00006970, 0.00005723, -0.00006342, 0.00005502, -0.00004850, 0.00003911, -0.00004407, 0.00003443, -0.00003940, 0.00004037, -0.00002940, 0.00002108, -0.00002417, 0.00002856, -0.00002428, 0.00001296, -0.00001944, 0.00002593, -0.00002124, 0.00001821, -0.00002084, 0.00001889, -0.00001933, 0.00002153, -0.00001244, 0.00000434, -0.00001260, 0.00001475, -0.00001620, 0.00002030, -0.00000826, -0.00000084, -0.00000797, 0.00000550, -0.00000913, 0.00001160, -0.00001118, 0.00001515, -0.00000984, 0.00001045, -0.00001604, 0.00001928, -0.00001788, 0.00001006, 0.00000295, -0.00001481, 0.00001736, -0.00001687, 0.00000139, 0.00000575, -0.00000282, 0.00000393, -0.00001791, 0.00001972, -0.00000782, -0.00001504, 0.00003020, },
		{ -0.00001190, -0.00000381, 0.00000929, 0.00000052, 0.00000811, -0.00001753, 0.00002088, -0.00002043, 0.00001733, -0.00000799, -0.00000292, 0.00000689, -0.00000870, 0.00000388, -0.00000479, 0.00001169, -0.00000788, 0.00000833, -0.00000597, 0.00000180, -0.00000503, -0.00000601, 0.00000147, 0.00001280, -0.00001237, 0.00001046, -0.00000791, -0.00000168, -0.00000361, 0.00001333, -0.00001103, 0.00001219, -0.00001596, 0.00001134, -0.00001237, 0.00001893, -0.00001711, 0.00000552, -0.00001013, 0.00001760, -0.00001644, 0.00001024, -0.00001319, 0.00002542, -0.00003031, 0.00002338, -0.00002503, 0.00002100, -0.00003271, 0.00003759, -0.00004200, 0.00003797, -0.00005225, 0.00006386, -0.00006501, 0.00005145, -0.00008335, 0.00007934, -0.00012384, 0.00009170, -0.00027033, -0.00017341, 0.00124945, -0.00076196, 0.00021488, -0.00021179, 0.00011322, -0.00010919, 0.00008053, -0.00008975, 0.00006066, -0.00004417, 0.00004032, -0.00004587, 0.00003274, -0.00002819, 0.00002467, -0.00003089, 0.00002234, -0.00002748, 0.00002081, -0.00001099, 0.00001308, -0.00002089, 0.00002045, -0.00001330, 0.00001061, -0.00002209, 0.00002078, -0.00001222, 0.00001203, -0.00001432, 0.00001052, -0.00001040, 0.00000897, -0.00000098, 0.00000361, -0.00001513, 0.00001209, -0.00000963, 0.00000689, 0.00000438, -0.00000033, -0.00001231, 0.00000770, -0.00001116, 0.00000982, -0.00000691, 0.00000694, -0.00000018, 0.00000290, -0.00000447, -0.00000084, 0.00000874, -0.00001697, 0.00002056, -0.00001558, 0.00000375, 0.00000597, -0.00001787, 0.00001254, -0.00000381, 0.00000557, -0.00000892, -0.00000575, 0.00002362, -0.00003469, 0.00002744, },
		{ 0.00002919, -0.00003186, 0.00002234, -0.00001258, 0.00002211, -0.00002129, 0.00001174, -0.00000300, -0.00000487, 0.00001153, -0.00001126, 0.00000379, 0.00000218, -0.00000829, 0.00000602, -0.00000440, 0.00001266, -0.00001285, 0.00001591, -0.00001711, 0.00001346, -0.00001909, 0.00000698, -0.00000158, 0.00001315, -0.00001774, 0.00002042, -0.00002300, 0.00001193, -0.00000948, 0.00001819, -0.00001853, 0.00001932, -0.00002541, 0.00002323, -0.00002245, 0.00003075, -0.00003627, 0.00002588, -0.00002516, 0.00003250, -0.00003584, 0.00002954, -0.00002597, 0.00003547, -0.00004662, 0.00004509, -0.00004964, 0.00004328, -0.00005140, 0.00005917, -0.00007032, 0.00006752, -0.00008061, 0.00010332, -0.00012614, 0.00011975, -0.00016285, 0.00018092, -0.00027885, 0.00030304, -0.00078743, 0.00096120, 0.00026806, -0.00053652, 0.00024139, -0.00023916, 0.00015902, -0.00014550, 0.00011079, -0.00012170, 0.00010526, -0.00008517, 0.00007115, -0.00007531, 0.00006506, -0.00005709, 0.00004696, -0.00005140, 0.00004297, -0.00004759, 0.00004476, -0.00003229, 0.00002626, -0.00003081, 0.00003332, -0.00002714, 0.00001840, -0.00002708, 0.00003117, -0.00002489, 0.00002239, -0.00002508, 0.00002207, -0.00002180, 0.00002281, -0.00001281, 0.00000694, -0.00001668, 0.00001831, -0.00001896, 0.00002027, -0.00000674, 0.00000074, -0.00001121, 0.00000852, -0.00001255, 0.00001448, -0.00001357, 0.00001644, -0.00001004, 0.00001076, -0.00001556, 0.00001574, -0.00001150, 0.00000143, 0.00001076, -0.00001839, 0.00001736, -0.00001317, -0.00000250, 0.00000844, -0.00000340, 0.00000641, -0.00001984, 0.00002071, -0.00000805, -0.00001458, },
		{ 0.00002339, -0.00000702, -0.00000876, 0.00001151, -0.00000347, 0.00001281, -0.00002253, 0.00002244, -0.00001892, 0.00001163, -0.00000035, -0.00000875, 0.00001018, -0.00000911, 0.00000340, -0.00000420, 0.00000991, -0.00000491, 0.00000472, -0.00000164, -0.00000222, -0.00000070, -0.00000894, 0.00000121, 0.00001188, -0.00000852, 0.00000558, -0.00000231, -0.00000596, -0.00000178, 0.00001067, -0.00000706, 0.00000748, -0.00000991, 0.00000517, -0.00000679, 0.00001179, -0.00000721, -0.00000266, -0.00000464, 0.00001048, -0.00000734, 0.00000214, -0.00000735, 0.00001841, -0.00001934, 0.00001142, -0.00001400, 0.00000950, -0.00002034, 0.00002341, -0.00002571, 0.00001939, -0.00003146, 0.00003908, -0.00003532, 0.00001718, -0.00004224, 0.00003071, -0.00006136, 0.00000893, -0.00012616, -0.00038742, 0.00135009, -0.00069105, 0.00012796, -0.00015223, 0.00006463, -0.00006982, 0.00004516, -0.00005819, 0.00003168, -0.00001857, 0.00001800, -0.00002530, 0.00001408, -0.00001156, 0.00001014, -0.00001800, 0.00001117, -0.00001554, 0.00000764, -0.00000082, 0.00000654, -0.00001423, 0.00001200, -0.00000573, 0.00000636, -0.00001725, 0.00001274, -0.00000474, 0.00000621, -0.00000841, 0.00000409, -0.00000449, 0.00000346, 0.00000320, 0.00000201, -0.00001297, 0.00000817, -0.00000502, 0.00000155, 0.00000753, 0.00000023, -0.00001213, 0.00000592, -0.00000908, 0.00000672, -0.00000340, 0.00000295, 0.00000363, -0.00000013, -0.00000089, -0.00000530, 0.00001165, -0.00001688, 0.00001559, -0.00000664, -0.00000536, 0.00001403, -0.00002222, 0.00001479, -0.00000535, 0.00000930, -0.00001104, -0.00000410, 0.00002156, -0.00003153, },
		{ -0.00001253, 0.00002652, -0.00002788, 0.00001659, -0.00000914, 0.00001754, -0.00001497, 0.00000237, 0.00000618, -0.00001297, 0.00001538, -0.00001015, 0.00000039, 0.00000626, -0.00001103, 0.00000811, -0.00000702, 0.00001502, -0.00001441, 0.00001681, -0.00001684, 0.00001288, -0.00001716, 0.00000485, -0.00000329, 0.00001652, -0.00001975, 0.00002137, -0.00002220, 0.00001093, -0.00001128, 0.00002101, -0.00002134, 0.00002232, -0.00002750, 0.00002505, -0.00002564, 0.00003390, -0.00003603, 0.00002531, -0.00002778, 0.00003546, -0.00003691, 0.00003065, -0.00003045, 0.00004215, -0.00005128, 0.00004853, -0.00005435, 0.00004910, -0.00005857, 0.00006721, -0.00007894, 0.00007662, -0.00009115, 0.00011439, -0.00013663, 0.00013061, -0.00017305, 0.00019250, -0.00028927, 0.00031511, -0.00074525, 0.00075418, 0.00051656, -0.00065749, 0.00029262, -0.00027502, 0.00018455, -0.00016684, 0.00012739, -0.00013505, 0.00011503, -0.00009272, 0.00007793, -0.00008072, 0.00006930, -0.00006059, 0.00005042, -0.00005498, 0.00004785, -0.00005168, 0.00004538, -0.00003266, 0.00002955, -0.00003519, 0.00003573, -0.00002841, 0.00002270, -0.00003284, 0.00003399, -0.00002646, 0.00002477, -0.00002739, 0.00002327, -0.00002249, 0.00002263, -0.00001242, 0.00000921, -0.00002001, 0.00002062, -0.00002018, 0.00001898, -0.00000505, 0.00000269, -0.00001441, 0.00001112, -0.00001525, 0.00001631, -0.00001468, 0.00001637, -0.00000902, 0.00000985, -0.00001368, 0.00001107, -0.00000500, -0.00000571, 0.00001515, -0.00001776, 0.00001341, -0.00000643, -0.00000827, 0.00001263, -0.00000574, 0.00001091, -0.00002348, 0.00002330, -0.00000996, },
		{ -0.00003037, 0.00002136, -0.00000438, -0.00001117, 0.00001149, -0.00000541, 0.00001506, -0.00002404, 0.00001994, -0.00001349, 0.00000307, 0.00000808, -0.00001383, 0.00001201, -0.00000808, 0.00000182, -0.00000261, 0.00000714, -0.00000117, 0.00000064, 0.00000275, -0.00000600, 0.00000306, -0.00001101, 0.00000077, 0.00001008, -0.00000374, 0.00000023, 0.00000322, -0.00000983, 0.00000015, 0.00000720, -0.00000207, 0.00000171, -0.00000298, -0.00000166, -0.00000033, 0.00000357, 0.00000310, -0.00001054, 0.00000108, 0.00000244, 0.00000244, -0.00000628, -0.00000053, 0.00000934, -0.00000624, -0.00000205, -0.00000145, -0.00000400, -0.00000548, 0.00000636, -0.00000624, -0.00000260, -0.00000700, 0.00001039, -0.00000168, -0.00002100, 0.00000249, -0.00002123, 0.00000414, -0.00007584, 0.00001103, -0.00056406, 0.00138482, -0.00056406, 0.00001103, -0.00007584, 0.00000414, -0.00002123, 0.00000249, -0.00002100, -0.00000168, 0.00001039, -0.00000700, -0.00000260, -0.00000624, 0.00000636, -0.00000548, -0.00000400, -0.00000145, -0.00000205, -0.00000624, 0.00000934, -0.00000053, -0.00000628, 0.00000244, 0.00000244, 0.00000108, -0.00001054, 0.00000310, 0.00000357, -0.00000033, -0.00000166, -0.00000298, 0.00000171, -0.00000207, 0.00000720, 0.00000015, -0.00000983, 0.00000322, 0.00000023, -0.00000374, 0.00001008, 0.00000077, -0.00001101, 0.00000306, -0.00000600, 0.00000275, 0.00000064, -0.00000117, 0.00000714, -0.00000261, 0.00000182, -0.00000808, 0.00001201, -0.00001383, 0.00000808, 0.00000307, -0.00001349, 0.00001994, -0.00002404, 0.00001506, -0.00000541, 0.00001149, -0.00001117, -0.00000438, 0.00002136, },
		{ -0.00001253, -0.00000996, 0.00002330, -0.00002348, 0.00001091, -0.00000574, 0.00001263, -0.00000827, -0.00000643, 0.00001341, -0.00001776, 0.00001515, -0.00000571, -0.00000500, 0.00001107, -0.00001368, 0.00000985, -0.00000902, 0.00001637, -0.00001468, 0.00001631, -0.00001525, 0.00001112, -0.00001441, 0.00000269, -0.00000505, 0.00001898, -0.00002018, 0.00002062, -0.00002001, 0.00000921, -0.00001242, 0.00002263, -0.00002249, 0.00002327, -0.00002739, 0.00002477, -0.00002646, 0.00003399, -0.00003284, 0.00002270, -0.00002841, 0.00003573, -0.00003519, 0.00002955, -0.00003266, 0.00004538, -0.00005168, 0.00004785, -0.00005498, 0.00005042, -0.00006059, 0.00006930, -0.00008072, 0.00007793, -0.00009272, 0.00011503, -0.00013505, 0.00012739, -0.00016684, 0.00018455, -0.00027502, 0.00029262, -0.00065749, 0.00051656, 0.00075418, -0.00074525, 0.00031511, -0.00028927, 0.00019250, -0.00017305, 0.00013061, -0.00013663, 0.00011439, -0.00009115, 0.00007662, -0.00007894, 0.00006721, -0.00005857, 0.00004910, -0.00005435, 0.00004853, -0.00005128, 0.00004215, -0.00003045, 0.00003065, -0.00003691, 0.00003546, -0.00002778, 0.00002531, -0.00003603, 0.00003390, -0.00002564, 0.00002505, -0.00002750, 0.00002232, -0.00002134, 0.00002101, -0.00001128, 0.00001093, -0.00002220, 0.00002137, -0.00001975, 0.00001652, -0.00000329, 0.00000485, -0.00001716, 0.00001288, -0.00001684, 0.00001681, -0.00001441, 0.00001502, -0.00000702, 0.00000811, -0.00001103, 0.00000626, 0.00000039, -0.00001015, 0.00001538, -0.00001297, 0.00000618, 0.00000237, -0.00001497, 0.00001754, -0.00000914, 0.00001659, -0.00002788, 0.00002652, },
		{ 0.00002339, -0.00003153, 0.00002156, -0.00000410, -0.00001104, 0.00000930, -0.00000535, 0.00001479, -0.00002222, 0.00001403, -0.00000536, -0.00000664, 0.00001559, -0.00001688, 0.00001165, -0.00000530, -0.00000089, -0.00000013, 0.00000363, 0.00000295, -0.00000340, 0.00000672, -0.00000908, 0.00000592, -0.00001213, 0.00000023, 0.00000753, 0.00000155, -0.00000502, 0.00000817, -0.00001297, 0.00000201, 0.00000320, 0.00000346, -0.00000449, 0.00000409, -0.00000841, 0.00000621, -0.00000474, 0.00001274, -0.00001725, 0.00000636, -0.00000573, 0.00001200, -0.00001423, 0.00000654, -0.00000082, 0.00000764, -0.00001554, 0.00001117, -0.00001800, 0.00001014, -0.00001156, 0.00001408, -0.00002530, 0.00001800, -0.00001857, 0.00003168, -0.00005819, 0.00004516, -0.00006982, 0.00006463, -0.00015223, 0.00012796, -0.00069105, 0.00135009, -0.00038742, -0.00012616, 0.00000893, -0.00006136, 0.00003071, -0.00004224, 0.00001718, -0.00003532, 0.00003908, -0.00003146, 0.00001939, -0.00002571, 0.00002341, -0.00002034, 0.00000950, -0.00001400, 0.00001142, -0.00001934, 0.00001841, -0.00000735, 0.00000214, -0.00000734, 0.00001048, -0.00000464, -0.00000266, -0.00000721, 0.00001179, -0.00000679, 0.00000517, -0.00000991, 0.00000748, -0.00000706, 0.00001067, -0.00000178, -0.00000596, -0.00000231, 0.00000558, -0.00000852, 0.00001188, 0.00000121, -0.00000894, -0.00000070, -0.00000222, -0.00000164, 0.00000472, -0.00000491, 0.00000991, -0.00000420, 0.00000340, -0.00000911, 0.00001018, -0.00000875, -0.00000035, 0.00001163, -0.00001892, 0.00002244, -0.00002253, 0.00001281, -0.00000347, 0.00001151, -0.00000876, -0.00000702, },
		{ 0.00002919, -0.00001458, -0.00000805, 0.00002071, -0.00001984, 0.00000641, -0.00000340, 0.00000844, -0.00000250, -0.00001317, 0.00001736, -0.00001839, 0.00001076, 0.00000143, -0.00001150, 0.00001574, -0.00001556, 0.00001076, -0.00001004, 0.00001644, -0.00001357, 0.00001448, -0.00001255, 0.00000852, -0.00001121, 0.00000074, -0.00000674, 0.00002027, -0.00001896, 0.00001831, -0.00001668, 0.00000694, -0.00001281, 0.00002281, -0.00002180, 0.00002207, -0.00002508, 0.00002239, -0.00002489, 0.00003117, -0.00002708, 0.00001840, -0.00002714, 0.00003332, -0.00003081, 0.00002626, -0.00003229, 0.00004476, -0.00004759, 0.00004297, -0.00005140, 0.00004696, -0.00005709, 0.00006506, -0.00007531, 0.00007115, -0.00008517, 0.00010526, -0.00012170, 0.00011079, -0.00014550, 0.00015902, -0.00023916, 0.00024139, -0.00053652, 0.00026806, 0.00096120, -0.00078743, 0.00030304, -0.00027885, 0.00018092, -0.00016285, 0.00011975, -0.00012614, 0.00010332, -0.00008061, 0.00006752, -0.00007032, 0.00005917, -0.00005140, 0.00004328, -0.00004964, 0.00004509, -0.00004662, 0.00003547, -0.00002597, 0.00002954, -0.00003584, 0.00003250, -0.00002516, 0.00002588, -0.00003627, 0.00003075, -0.00002245, 0.00002323, -0.00002541, 0.00001932, -0.00001853, 0.00001819, -0.00000948, 0.00001193, -0.00002300, 0.00002042, -0.00001774, 0.00001315, -0.00000158, 0.00000698, -0.00001909, 0.00001346, -0.00001711, 0.00001591, -0.00001285, 0.00001266, -0.00000440, 0.00000602, -0.00000829, 0.00000218, 0.00000379, -0.00001126, 0.00001153, -0.00000487, -0.00000300, 0.00001174, -0.00002129, 0.00002211, -0.00001258, 0.00002234, -0.00003186, },
		{ -0.00001190, 0.00002744, -0.00003469, 0.00002362, -0.00000575, -0.00000892, 0.00000557, -0.00000381, 0.00001254, -0.00001787, 0.00000597, 0.00000375, -0.00001558, 0.00002056, -0.00001697, 0.00000874, -0.00000084, -0.00000447, 0.00000290, -0.00000018, 0.00000694, -0.00000691, 0.00000982, -0.00001116, 0.00000770, -0.00001231, -0.00000033, 0.00000438, 0.00000689, -0.00000963, 0.00001209, -0.00001513, 0.00000361, -0.00000098, 0.00000897, -0.00001040, 0.00001052, -0.00001432, 0.00001203, -0.00001222, 0.00002078, -0.00002209, 0.00001061, -0.00001330, 0.00002045, -0.00002089, 0.00001308, -0.00001099, 0.00002081, -0.00002748, 0.00002234, -0.00003089, 0.00002467, -0.00002819, 0.00003274, -0.00004587, 0.00004032, -0.00004417, 0.00006066, -0.00008975, 0.00008053, -0.00010919, 0.00011322, -0.00021179, 0.00021488, -0.00076196, 0.00124945, -0.00017341, -0.00027033, 0.00009170, -0.00012384, 0.00007934, -0.00008335, 0.00005145, -0.00006501, 0.00006386, -0.00005225, 0.00003797, -0.00004200, 0.00003759, -0.00003271, 0.00002100, -0.00002503, 0.00002338, -0.00003031, 0.00002542, -0.00001319, 0.00001024, -0.00001644, 0.00001760, -0.00001013, 0.00000552, -0.00001711, 0.00001893, -0.00001237, 0.00001134, -0.00001596, 0.00001219, -0.00001103, 0.00001333, -0.00000361, -0.00000168, -0.00000791, 0.00001046, -0.00001237, 0.00001280, 0.00000147, -0.00000601, -0.00000503, 0.00000180, -0.00000597, 0.00000833, -0.00000788, 0.00001169, -0.00000479, 0.00000388, -0.00000870, 0.00000689, -0.00000292, -0.00000799, 0.00001733, -0.00002043, 0.00002088, -0.00001753, 0.00000811, 0.00000052, 0.00000929, -0.00000381, },
		{ -0.00003424, 0.00003020, -0.00001504, -0.00000782, 0.00001972, -0.00001791, 0.00000393, -0.00000282, 0.00000575, 0.00000139, -0.00001687, 0.00001736, -0.00001481, 0.00000295, 0.00001006, -0.00001788, 0.00001928, -0.00001604, 0.00001045, -0.00000984, 0.00001515, -0.00001118, 0.00001160, -0.00000913, 0.00000550, -0.00000797, -0.00000084, -0.00000826, 0.00002030, -0.00001620, 0.00001475, -0.00001260, 0.00000434, -0.00001244, 0.00002153, -0.00001933, 0.00001889, -0.00002084, 0.00001821, -0.00002124, 0.00002593, -0.00001944, 0.00001296, -0.00002428, 0.00002856, -0.00002417, 0.00002108, -0.00002940, 0.00004037, -0.00003940, 0.00003443, -0.00004407, 0.00003911, -0.00004850, 0.00005502, -0.00006342, 0.00005723, -0.00006970, 0.00008666, -0.00009864, 0.00008348, -0.00011242, 0.00012023, -0.00018751, 0.00017044, -0.00039735, 0.00002876, 0.00112066, -0.00077615, 0.00025484, -0.00024423, 0.00015076, -0.00013743, 0.00009620, -0.00010508, 0.00008333, -0.00006260, 0.00005203, -0.00005619, 0.00004641, -0.00004015, 0.00003386, -0.00004158, 0.00003815, -0.00003842, 0.00002618, -0.00001979, 0.00002647, -0.00003217, 0.00002712, -0.00002072, 0.00002437, -0.00003357, 0.00002479, -0.00001723, 0.00001958, -0.00002140, 0.00001464, -0.00001448, 0.00001450, -0.00000719, 0.00001216, -0.00002229, 0.00001781, -0.00001438, 0.00000916, -0.00000001, 0.00000887, -0.00001990, 0.00001269, -0.00001600, 0.00001371, -0.00001027, 0.00000965, -0.00000160, 0.00000405, -0.00000603, -0.00000059, 0.00000479, -0.00000912, 0.00000446, 0.00000496, -0.00001231, 0.00002000, -0.00002584, 0.00002525, -0.00001499, 0.00002701, },
		{ 0.00000311, -0.00001832, 0.00003276, -0.00003907, 0.00002669, -0.00000844, -0.00000580, 0.00000125, -0.00000167, 0.00000921, -0.00001218, -0.00000263, 0.00001197, -0.00002198, 0.00002185, -0.00001369, 0.00000343, 0.00000484, -0.00000841, 0.00000600, -0.00000382, 0.00001031, -0.00000944, 0.00001174, -0.00001204, 0.00000836, -0.00001168, -0.00000080, 0.00000085, 0.00001180, -0.00001315, 0.00001467, -0.00001619, 0.00000482, -0.00000496, 0.00001389, -0.00001537, 0.00001563, -0.00001874, 0.00001646, -0.00001810, 0.00002648, -0.00002458, 0.00001344, -0.00001965, 0.00002704, -0.00002559, 0.00001835, -0.00002004, 0.00003187, -0.00003648, 0.00003072, -0.00004119, 0.00003638, -0.00004155, 0.00004751, -0.00006180, 0.00005720, -0.00006335, 0.00008192, -0.00011204, 0.00010469, -0.00013508, 0.00014517, -0.00024909, 0.00026668, -0.00077672, 0.00109294, 0.00006121, -0.00040628, 0.00016166, -0.00017541, 0.00011838, -0.00011561, 0.00007748, -0.00008717, 0.00008173, -0.00006686, 0.00005102, -0.00005331, 0.00004737, -0.00004129, 0.00002929, -0.00003337, 0.00003263, -0.00003815, 0.00002970, -0.00001750, 0.00001725, -0.00002402, 0.00002313, -0.00001474, 0.00001311, -0.00002557, 0.00002412, -0.00001638, 0.00001618, -0.00002050, 0.00001532, -0.00001364, 0.00001500, -0.00000516, 0.00000266, -0.00001303, 0.00001436, -0.00001499, 0.00001283, 0.00000154, -0.00000245, -0.00000949, 0.00000556, -0.00000972, 0.00001107, -0.00000981, 0.00001239, -0.00000445, 0.00000351, -0.00000738, 0.00000308, 0.00000233, -0.00001335, 0.00001900, -0.00001756, 0.00001543, -0.00000960, 0.00000156, 0.00000606, 0.00000528, },
		{ 0.00002967, -0.00003413, 0.00002877, -0.00001322, -0.00000990, 0.00002088, -0.00001815, 0.00000382, -0.00000428, 0.00000488, 0.00000295, -0.00001723, 0.00001357, -0.00000782, -0.00000684, 0.00001860, -0.00002282, 0.00002080, -0.00001464, 0.00000875, -0.00000839, 0.00001264, -0.00000782, 0.00000808, -0.00000545, 0.00000249, -0.00000502, -0.00000193, -0.00000948, 0.00001903, -0.00001216, 0.00001039, -0.00000820, 0.00000163, -0.00001138, 0.00001891, -0.00001534, 0.00001417, -0.00001521, 0.00001277, -0.00001610, 0.00001901, -0.00001078, 0.00000709, -0.00002031, 0.00002196, -0.00001595, 0.00001458, -0.00002439, 0.00003276, -0.00002804, 0.00002331, -0.00003400, 0.00002788, -0.00003596, 0.00004056, -0.00004671, 0.00003813, -0.00004866, 0.00006206, -0.00006924, 0.00004952, -0.00007250, 0.00007414, -0.00012762, 0.00009036, -0.00025551, -0.00018335, 0.00122048, -0.00070927, 0.00017379, -0.00018947, 0.00010573, -0.00010026, 0.00006325, -0.00007652, 0.00005727, -0.00003970, 0.00003249, -0.00003864, 0.00003076, -0.00002643, 0.00002220, -0.00003130, 0.00002876, -0.00002782, 0.00001538, -0.00001269, 0.00002190, -0.00002636, 0.00001986, -0.00001487, 0.00002100, -0.00002825, 0.00001664, -0.00001062, 0.00001464, -0.00001600, 0.00000885, -0.00000972, 0.00001038, -0.00000464, 0.00001161, -0.00002011, 0.00001379, -0.00001008, 0.00000494, 0.00000132, 0.00001031, -0.00001940, 0.00001056, -0.00001365, 0.00001049, -0.00000705, 0.00000643, 0.00000098, 0.00000258, -0.00000457, -0.00000183, 0.00000352, -0.00000447, -0.00000437, 0.00001459, -0.00001988, 0.00002559, -0.00002749, 0.00002608, -0.00001552, },
		{ 0.00000035, 0.00001102, -0.00002518, 0.00003823, -0.00004352, 0.00002964, -0.00001102, -0.00000282, -0.00000254, 0.00000006, 0.00000585, -0.00000651, -0.00001015, 0.00001769, -0.00002461, 0.00001897, -0.00000729, -0.00000362, 0.00001098, -0.00001202, 0.00000860, -0.00000676, 0.00001263, -0.00001071, 0.00001235, -0.00001176, 0.00000805, -0.00001046, -0.00000108, -0.00000284, 0.00001588, -0.00001524, 0.00001579, -0.00001615, 0.00000556, -0.00000842, 0.00001774, -0.00001888, 0.00001894, -0.00002125, 0.00001907, -0.00002188, 0.00002947, -0.00002457, 0.00001468, -0.00002436, 0.00003124, -0.00002787, 0.00002179, -0.00002706, 0.00003966, -0.00004150, 0.00003540, -0.00004787, 0.00004401, -0.00005014, 0.00005674, -0.00007133, 0.00006676, -0.00007413, 0.00009340, -0.00012298, 0.00011556, -0.00014550, 0.00015850, -0.00026226, 0.00028307, -0.00074117, 0.00089584, 0.00029716, -0.00051892, 0.00020929, -0.00020960, 0.00014291, -0.00013515, 0.00009224, -0.00009940, 0.00009080, -0.00007379, 0.00005732, -0.00005867, 0.00005194, -0.00004539, 0.00003372, -0.00003830, 0.00003843, -0.00004228, 0.00003097, -0.00001996, 0.00002262, -0.00002946, 0.00002652, -0.00001792, 0.00001934, -0.00003174, 0.00002674, -0.00001838, 0.00001925, -0.00002312, 0.00001660, -0.00001475, 0.00001560, -0.00000632, 0.00000666, -0.00001717, 0.00001687, -0.00001619, 0.00001199, 0.00000141, 0.00000144, -0.00001359, 0.00000859, -0.00001251, 0.00001266, -0.00001059, 0.00001209, -0.00000343, 0.00000269, -0.00000576, -0.00000034, 0.00000593, -0.00001548, 0.00001629, -0.00001076, 0.00000705, 0.00000007, -0.00000571, 0.00001221, },
		{ -0.00001376, 0.00002986, -0.00003114, 0.00002465, -0.00000896, -0.00001434, 0.00002416, -0.00002049, 0.00000587, -0.00000749, 0.00000564, 0.00000231, -0.00001462, 0.00000686, 0.00000112, -0.00001681, 0.00002538, -0.00002518, 0.00001973, -0.00001123, 0.00000574, -0.00000589, 0.00000923, -0.00000392, 0.00000442, -0.00000199, -0.00000011, -0.00000264, -0.00000248, -0.00001032, 0.00001657, -0.00000725, 0.00000573, -0.00000393, -0.00000092, -0.00000978, 0.00001522, -0.00001030, 0.00000849, -0.00000887, 0.00000678, -0.00001020, 0.00001132, -0.00000206, 0.00000151, -0.00001579, 0.00001422, -0.00000696, 0.00000748, -0.00001792, 0.00002291, -0.00001483, 0.00001104, -0.00002257, 0.00001473, -0.00002113, 0.00002364, -0.00002746, 0.00001651, -0.00002514, 0.00003500, -0.00003761, 0.00001368, -0.00003129, 0.00002722, -0.00006744, 0.00001167, -0.00012499, -0.00035443, 0.00125488, -0.00059087, 0.00006777, -0.00012173, 0.00005182, -0.00005653, 0.00002556, -0.00004459, 0.00002878, -0.00001513, 0.00001169, -0.00002011, 0.00001438, -0.00001209, 0.00000988, -0.00002019, 0.00001821, -0.00001615, 0.00000430, -0.00000555, 0.00001643, -0.00001907, 0.00001146, -0.00000821, 0.00001626, -0.00002098, 0.00000717, -0.00000348, 0.00000913, -0.00000988, 0.00000264, -0.00000487, 0.00000628, -0.00000206, 0.00001038, -0.00001667, 0.00000877, -0.00000533, 0.00000086, 0.00000232, 0.00001117, -0.00001757, 0.00000724, -0.00001039, 0.00000664, -0.00000363, 0.00000339, 0.00000304, 0.00000179, -0.00000399, -0.00000167, 0.00000055, 0.00000154, -0.00001319, 0.00002209, -0.00002415, 0.00002742, -0.00002561, 0.00002422, },
		{ 0.00001784, -0.00000437, 0.00001870, -0.00003122, 0.00004262, -0.00004685, 0.00003131, -0.00001240, -0.00000106, -0.00000482, 0.00000050, 0.00000339, -0.00000202, -0.00001530, 0.00001983, -0.00002297, 0.00001226, 0.00000133, -0.00001133, 0.00001657, -0.00001457, 0.00001019, -0.00000862, 0.00001361, -0.00001060, 0.00001171, -0.00001049, 0.00000702, -0.00000894, -0.00000108, -0.00000643, 0.00001878, -0.00001575, 0.00001548, -0.00001512, 0.00000580, -0.00001109, 0.00002017, -0.00002062, 0.00002021, -0.00002169, 0.00001968, -0.00002336, 0.00002970, -0.00002221, 0.00001440, -0.00002722, 0.00003279, -0.00002754, 0.00002307, -0.00003140, 0.00004349, -0.00004209, 0.00003602, -0.00005044, 0.00004690, -0.00005322, 0.00005965, -0.00007370, 0.00006828, -0.00007589, 0.00009463, -0.00012227, 0.00011316, -0.00014084, 0.00015403, -0.00025297, 0.00026819, -0.00066586, 0.00067654, 0.00051509, -0.00059536, 0.00022783, -0.00022249, 0.00015024, -0.00014008, 0.00009446, -0.00010089, 0.00009058, -0.00007278, 0.00005675, -0.00005802, 0.00005130, -0.00004498, 0.00003420, -0.00003965, 0.00004055, -0.00004265, 0.00002932, -0.00002054, 0.00002602, -0.00003234, 0.00002750, -0.00001930, 0.00002363, -0.00003506, 0.00002648, -0.00001820, 0.00002037, -0.00002366, 0.00001599, -0.00001440, 0.00001520, -0.00000699, 0.00001003, -0.00001995, 0.00001777, -0.00001596, 0.00001042, 0.00000112, 0.00000531, -0.00001688, 0.00001050, -0.00001406, 0.00001299, -0.00001029, 0.00001101, -0.00000206, 0.00000183, -0.00000443, -0.00000265, 0.00000721, -0.00001408, 0.00000967, -0.00000125, -0.00000270, 0.00000991, -0.00001232, },
		{ 0.00001985, -0.00000981, 0.00002764, -0.00002546, 0.00001817, -0.00000269, -0.00002063, 0.00002900, -0.00002430, 0.00000937, -0.00001173, 0.00000740, 0.00000011, -0.00000997, -0.00000136, 0.00001022, -0.00002511, 0.00002902, -0.00002425, 0.00001590, -0.00000605, 0.00000181, -0.00000276, 0.00000537, -0.00000000, 0.00000112, 0.00000085, -0.00000200, -0.00000096, -0.00000256, -0.00001068, 0.00001310, -0.00000194, 0.00000130, -0.00000019, -0.00000310, -0.00000783, 0.00001084, -0.00000478, 0.00000256, -0.00000257, 0.00000095, -0.00000430, 0.00000375, 0.00000581, -0.00000310, -0.00001128, 0.00000610, 0.00000189, 0.00000061, -0.00001085, 0.00001200, -0.00000135, -0.00000077, -0.00001132, 0.00000128, -0.00000593, 0.00000648, -0.00000824, -0.00000473, -0.00000237, 0.00000914, -0.00000786, -0.00001941, 0.00000599, -0.00001460, -0.00001402, -0.00005667, -0.00001662, -0.00047613, 0.00122487, -0.00043083, -0.00005171, -0.00005025, -0.00000371, -0.00001233, -0.00001167, -0.00001378, 0.00000169, 0.00000784, -0.00000754, -0.00000303, -0.00000067, 0.00000111, -0.00000158, -0.00000960, 0.00000783, -0.00000475, -0.00000589, 0.00000084, 0.00001069, -0.00001106, 0.00000277, -0.00000148, 0.00001081, -0.00001262, -0.00000257, 0.00000325, 0.00000381, -0.00000377, -0.00000327, -0.00000053, 0.00000263, 0.00000032, 0.00000862, -0.00001229, 0.00000328, -0.00000065, -0.00000271, 0.00000294, 0.00001139, -0.00001454, 0.00000309, -0.00000663, 0.00000265, -0.00000043, 0.00000085, 0.00000441, 0.00000167, -0.00000413, -0.00000058, -0.00000324, 0.00000752, -0.00002028, 0.00002594, -0.00002419, 0.00002514, -0.00002024, },
		{ -0.00001690, 0.00002186, -0.00000782, 0.00002500, -0.00003530, 0.00004489, -0.00004810, 0.00003083, -0.00001178, -0.00000126, -0.00000497, -0.00000085, 0.00000243, 0.00000052, -0.00001735, 0.00001806, -0.00001742, 0.00000277, 0.00001077, -0.00001840, 0.00002063, -0.00001542, 0.00001041, -0.00000918, 0.00001318, -0.00000921, 0.00001005, -0.00000855, 0.00000560, -0.00000737, -0.00000077, -0.00000968, 0.00002029, -0.00001469, 0.00001392, -0.00001332, 0.00000556, -0.00001280, 0.00002100, -0.00002049, 0.00001948, -0.00002018, 0.00001843, -0.00002269, 0.00002745, -0.00001795, 0.00001288, -0.00002826, 0.00003172, -0.00002474, 0.00002220, -0.00003283, 0.00004319, -0.00003837, 0.00003284, -0.00004904, 0.00004508, -0.00005084, 0.00005642, -0.00006923, 0.00006230, -0.00006939, 0.00008666, -0.00011138, 0.00009942, -0.00012363, 0.00013504, -0.00022572, 0.00022947, -0.00056429, 0.00045422, 0.00069789, -0.00062671, 0.00021447, -0.00021339, 0.00014034, -0.00013077, 0.00008484, -0.00009253, 0.00008203, -0.00006476, 0.00005020, -0.00005219, 0.00004619, -0.00004069, 0.00003122, -0.00003773, 0.00003929, -0.00003966, 0.00002522, -0.00001942, 0.00002734, -0.00003258, 0.00002605, -0.00001881, 0.00002569, -0.00003533, 0.00002342, -0.00001602, 0.00001966, -0.00002224, 0.00001371, -0.00001285, 0.00001398, -0.00000718, 0.00001250, -0.00002111, 0.00001700, -0.00001443, 0.00000829, 0.00000070, 0.00000883, -0.00001898, 0.00001106, -0.00001429, 0.00001213, -0.00000913, 0.00000945, -0.00000066, 0.00000128, -0.00000377, -0.00000351, 0.00000606, -0.00000959, 0.00000039, 0.00000915, -0.00001195, 0.00001821, },
		{ -0.00001215, 0.00001377, -0.00000431, 0.00002365, -0.00001781, 0.00001020, 0.00000468, -0.00002782, 0.00003440, -0.00002859, 0.00001331, -0.00001602, 0.00000929, -0.00000275, -0.00000450, -0.00000948, 0.00001772, -0.00003023, 0.00002868, -0.00001991, 0.00000970, 0.00000029, -0.00000242, 0.00000045, 0.00000162, 0.00000341, -0.00000138, 0.00000277, -0.00000306, -0.00000001, -0.00000227, -0.00001051, 0.00000890, 0.00000328, -0.00000243, 0.00000273, -0.00000476, -0.00000574, 0.00000621, 0.00000061, -0.00000295, 0.00000301, -0.00000406, 0.00000088, -0.00000290, 0.00001211, -0.00000627, -0.00000723, -0.00000170, 0.00000975, -0.00000524, -0.00000406, 0.00000132, 0.00001086, -0.00001065, -0.00000163, -0.00001090, 0.00000781, -0.00000878, 0.00000859, -0.00002296, 0.00001673, -0.00001233, 0.00001652, -0.00004595, 0.00003524, -0.00004680, 0.00002741, -0.00010843, 0.00006304, -0.00054617, 0.00113788, -0.00024355, -0.00017104, 0.00001508, -0.00005351, 0.00002643, -0.00004351, 0.00001187, -0.00002064, 0.00002641, -0.00002288, 0.00001064, -0.00001270, 0.00001172, -0.00001092, -0.00000068, -0.00000125, 0.00000524, -0.00001425, 0.00000585, 0.00000529, -0.00000313, -0.00000534, 0.00000460, 0.00000537, -0.00000411, -0.00001151, 0.00000872, -0.00000065, 0.00000165, -0.00000824, 0.00000280, -0.00000024, 0.00000232, 0.00000655, -0.00000740, -0.00000210, 0.00000343, -0.00000547, 0.00000313, 0.00001096, -0.00001056, -0.00000144, -0.00000287, -0.00000103, 0.00000218, -0.00000099, 0.00000508, 0.00000207, -0.00000465, 0.00000085, -0.00000688, 0.00001219, -0.00002434, 0.00002531, -0.00001992, 0.00001918, },
		{ 0.00002356, -0.00001845, 0.00002349, -0.00000922, 0.00002911, -0.00003665, 0.00004439, -0.00004673, 0.00002777, -0.00000883, -0.00000368, -0.00000285, -0.00000406, 0.00000311, 0.00000086, -0.00001624, 0.00001284, -0.00000904, -0.00000790, 0.00001942, -0.00002355, 0.00002232, -0.00001419, 0.00000913, -0.00000843, 0.00001147, -0.00000682, 0.00000776, -0.00000633, 0.00000413, -0.00000599, -0.00000017, -0.00001237, 0.00002030, -0.00001225, 0.00001146, -0.00001104, 0.00000495, -0.00001351, 0.00002027, -0.00001868, 0.00001703, -0.00001709, 0.00001569, -0.00002025, 0.00002330, -0.00001243, 0.00001056, -0.00002770, 0.00002832, -0.00001989, 0.00001948, -0.00003152, 0.00003915, -0.00003109, 0.00002669, -0.00004439, 0.00003926, -0.00004385, 0.00004810, -0.00005923, 0.00005041, -0.00005657, 0.00007183, -0.00009309, 0.00007775, -0.00009796, 0.00010647, -0.00018688, 0.00017622, -0.00045086, 0.00024650, 0.00083277, -0.00060934, 0.00017085, -0.00018495, 0.00011581, -0.00010976, 0.00006587, -0.00007665, 0.00006729, -0.00005170, 0.00003940, -0.00004270, 0.00003793, -0.00003362, 0.00002567, -0.00003331, 0.00003532, -0.00003406, 0.00001939, -0.00001702, 0.00002674, -0.00003036, 0.00002247, -0.00001659, 0.00002555, -0.00003273, 0.00001801, -0.00001231, 0.00001747, -0.00001924, 0.00001019, -0.00001049, 0.00001222, -0.00000691, 0.00001395, -0.00002061, 0.00001474, -0.00001190, 0.00000586, 0.00000020, 0.00001170, -0.00001965, 0.00001022, -0.00001327, 0.00001029, -0.00000739, 0.00000772, 0.00000050, 0.00000125, -0.00000391, -0.00000293, 0.00000289, -0.00000304, -0.00000984, 0.00001844, -0.00001891, },
		{ 0.00001074, -0.00000263, 0.00000717, 0.00000173, 0.00001886, -0.00000929, 0.00000192, 0.00001194, -0.00003471, 0.00003921, -0.00003224, 0.00001657, -0.00001933, 0.00001037, -0.00000528, 0.00000053, -0.00001601, 0.00002222, -0.00003127, 0.00002425, -0.00001268, 0.00000198, 0.00000688, -0.00000619, 0.00000315, -0.00000152, 0.00000588, -0.00000280, 0.00000365, -0.00000327, 0.00000034, -0.00000180, -0.00000978, 0.00000430, 0.00000793, -0.00000511, 0.00000463, -0.00000582, -0.00000371, 0.00000177, 0.00000533, -0.00000746, 0.00000732, -0.00000775, 0.00000483, -0.00000806, 0.00001638, -0.00000772, -0.00000398, -0.00000857, 0.00001591, -0.00000944, 0.00000166, -0.00000799, 0.00002051, -0.00001746, 0.00000542, -0.00002059, 0.00001859, -0.00002047, 0.00002118, -0.00003621, 0.00003010, -0.00002724, 0.00003328, -0.00006358, 0.00005402, -0.00006686, 0.00005415, -0.00014063, 0.00011200, -0.00056794, 0.00100644, -0.00004608, -0.00027643, 0.00006531, -0.00009137, 0.00005497, -0.00006620, 0.00002940, -0.00003587, 0.00003872, -0.00003283, 0.00001964, -0.00002069, 0.00001887, -0.00001734, 0.00000579, -0.00000824, 0.00001303, -0.00002015, 0.00000906, 0.00000067, 0.00000402, -0.00001214, 0.00000939, 0.00000059, 0.00000365, -0.00001874, 0.00001224, -0.00000373, 0.00000589, -0.00001181, 0.00000480, -0.00000216, 0.00000381, 0.00000437, -0.00000246, -0.00000683, 0.00000650, -0.00000722, 0.00000291, 0.00000998, -0.00000601, -0.00000584, 0.00000040, -0.00000402, 0.00000396, -0.00000207, 0.00000517, 0.00000274, -0.00000515, 0.00000203, -0.00000955, 0.00001465, -0.00002469, 0.00002024, -0.00001214, },
		{ -0.00002225, 0.00002511, -0.00001652, 0.00002247, -0.00000827, 0.00003071, -0.00003504, 0.00004104, -0.00004279, 0.00002229, -0.00000379, -0.00000802, 0.00000113, -0.00000869, 0.00000509, -0.00000069, -0.00001254, 0.00000526, 0.00000055, -0.00001794, 0.00002574, -0.00002580, 0.00002122, -0.00001086, 0.00000655, -0.00000662, 0.00000883, -0.00000386, 0.00000529, -0.00000422, 0.00000291, -0.00000495, 0.00000066, -0.00001434, 0.00001888, -0.00000875, 0.00000849, -0.00000859, 0.00000408, -0.00001329, 0.00001817, -0.00001555, 0.00001334, -0.00001297, 0.00001198, -0.00001664, 0.00001796, -0.00000641, 0.00000796, -0.00002591, 0.00002311, -0.00001365, 0.00001547, -0.00002801, 0.00003222, -0.00002144, 0.00001880, -0.00003762, 0.00003063, -0.00003366, 0.00003639, -0.00004572, 0.00003494, -0.00004012, 0.00005324, -0.00007100, 0.00005226, -0.00006862, 0.00007391, -0.00014341, 0.00011797, -0.00033891, 0.00006742, 0.00091262, -0.00054532, 0.00010288, -0.00014278, 0.00008139, -0.00008124, 0.00004130, -0.00005657, 0.00004928, -0.00003612, 0.00002652, -0.00003142, 0.00002813, -0.00002514, 0.00001872, -0.00002737, 0.00002958, -0.00002684, 0.00001267, -0.00001385, 0.00002454, -0.00002613, 0.00001727, -0.00001305, 0.00002353, -0.00002777, 0.00001099, -0.00000777, 0.00001436, -0.00001520, 0.00000597, -0.00000779, 0.00001021, -0.00000629, 0.00001436, -0.00001858, 0.00001131, -0.00000875, 0.00000339, -0.00000037, 0.00001372, -0.00001882, 0.00000811, -0.00001128, 0.00000780, -0.00000541, 0.00000609, 0.00000125, 0.00000175, -0.00000473, -0.00000129, -0.00000151, 0.00000420, -0.00001915, 0.00002480, },
		{ -0.00000237, 0.00000149, 0.00000669, 0.00000142, 0.00000715, 0.00001441, -0.00000110, -0.00000543, 0.00001789, -0.00004016, 0.00004236, -0.00003427, 0.00001822, -0.00002083, 0.00000996, -0.00000671, 0.00000409, -0.00001989, 0.00002297, -0.00002811, 0.00001638, -0.00000368, -0.00000606, 0.00001269, -0.00000876, 0.00000484, -0.00000365, 0.00000715, -0.00000303, 0.00000354, -0.00000280, 0.00000028, -0.00000133, -0.00000854, -0.00000036, 0.00001160, -0.00000652, 0.00000547, -0.00000626, -0.00000190, -0.00000211, 0.00000892, -0.00001058, 0.00001004, -0.00000985, 0.00000728, -0.00001141, 0.00001845, -0.00000746, -0.00000167, -0.00001408, 0.00001990, -0.00001157, 0.00000577, -0.00001507, 0.00002672, -0.00002057, 0.00000931, -0.00002708, 0.00002553, -0.00002756, 0.00002849, -0.00004347, 0.00003674, -0.00003467, 0.00004156, -0.00007158, 0.00006179, -0.00007447, 0.00006603, -0.00015356, 0.00013241, -0.00054949, 0.00084637, 0.00014411, -0.00035585, 0.00009385, -0.00011326, 0.00007043, -0.00007766, 0.00003736, -0.00004301, 0.00004408, -0.00003692, 0.00002364, -0.00002434, 0.00002232, -0.00002060, 0.00000953, -0.00001276, 0.00001821, -0.00002336, 0.00001038, -0.00000286, 0.00000982, -0.00001704, 0.00001245, -0.00000298, 0.00000995, -0.00002359, 0.00001343, -0.00000518, 0.00000866, -0.00001372, 0.00000536, -0.00000310, 0.00000476, 0.00000228, 0.00000210, -0.00001044, 0.00000827, -0.00000791, 0.00000231, 0.00000856, -0.00000131, -0.00000961, 0.00000279, -0.00000605, 0.00000481, -0.00000245, 0.00000486, 0.00000338, -0.00000529, 0.00000251, -0.00001074, 0.00001450, -0.00002140, 0.00001162, },
		{ 0.00002698, -0.00002139, 0.00002277, -0.00001134, 0.00001915, -0.00000522, 0.00003001, -0.00003080, 0.00003533, -0.00003686, 0.00001504, 0.00000261, -0.00001353, 0.00000617, -0.00001394, 0.00000769, -0.00000344, -0.00000728, -0.00000321, 0.00000957, -0.00002564, 0.00002858, -0.00002466, 0.00001734, -0.00000581, 0.00000311, -0.00000420, 0.00000571, -0.00000083, 0.00000308, -0.00000257, 0.00000213, -0.00000431, 0.00000159, -0.00001545, 0.00001618, -0.00000461, 0.00000545, -0.00000626, 0.00000310, -0.00001228, 0.00001506, -0.00001160, 0.00000901, -0.00000843, 0.00000793, -0.00001251, 0.00001220, -0.00000064, 0.00000558, -0.00002332, 0.00001676, -0.00000682, 0.00001089, -0.00002308, 0.00002355, -0.00001089, 0.00001059, -0.00003007, 0.00002064, -0.00002199, 0.00002331, -0.00003100, 0.00001851, -0.00002297, 0.00003417, -0.00004876, 0.00002710, -0.00004022, 0.00004266, -0.00010173, 0.00006308, -0.00023912, -0.00007383, 0.00093653, -0.00044208, 0.00001988, -0.00009444, 0.00004306, -0.00005025, 0.00001546, -0.00003597, 0.00003110, -0.00002065, 0.00001379, -0.00002025, 0.00001840, -0.00001662, 0.00001154, -0.00002096, 0.00002309, -0.00001902, 0.00000595, -0.00001047, 0.00002123, -0.00002050, 0.00001116, -0.00000877, 0.00002017, -0.00002123, 0.00000329, -0.00000317, 0.00001095, -0.00001072, 0.00000168, -0.00000524, 0.00000824, -0.00000543, 0.00001381, -0.00001531, 0.00000719, -0.00000546, 0.00000115, -0.00000098, 0.00001477, -0.00001659, 0.00000504, -0.00000868, 0.00000504, -0.00000351, 0.00000474, 0.00000159, 0.00000263, -0.00000591, 0.00000082, -0.00000614, 0.00001071, -0.00002594, },
		{ 0.00000099, 0.00000744, -0.00000675, 0.00001424, -0.00000221, 0.00001093, 0.00001131, 0.00000569, -0.00001080, 0.00002160, -0.00004333, 0.00004313, -0.00003406, 0.00001772, -0.00002012, 0.00000772, -0.00000660, 0.00000563, -0.00002064, 0.00001995, -0.00002143, 0.00000634, 0.00000562, -0.00001314, 0.00001673, -0.00000955, 0.00000521, -0.00000459, 0.00000714, -0.00000216, 0.00000265, -0.00000190, 0.00000004, -0.00000100, -0.00000688, -0.00000474, 0.00001401, -0.00000659, 0.00000535, -0.00000617, -0.00000043, -0.00000514, 0.00001115, -0.00001215, 0.00001108, -0.00001035, 0.00000817, -0.00001293, 0.00001841, -0.00000571, -0.00000026, -0.00001802, 0.00002152, -0.00001154, 0.00000800, -0.00001948, 0.00002918, -0.00001997, 0.00001009, -0.00003025, 0.00002836, -0.00002985, 0.00003041, -0.00004473, 0.00003683, -0.00003496, 0.00004195, -0.00007083, 0.00005980, -0.00007126, 0.00006516, -0.00015016, 0.00012966, -0.00050179, 0.00067450, 0.00031134, -0.00040069, 0.00009752, -0.00011795, 0.00007226, -0.00007781, 0.00003603, -0.00004251, 0.00004303, -0.00003567, 0.00002311, -0.00002409, 0.00002241, -0.00002096, 0.00001067, -0.00001486, 0.00002080, -0.00002398, 0.00000994, -0.00000521, 0.00001395, -0.00001976, 0.00001356, -0.00000507, 0.00001434, -0.00002573, 0.00001224, -0.00000505, 0.00000993, -0.00001400, 0.00000458, -0.00000319, 0.00000522, 0.00000045, 0.00000591, -0.00001263, 0.00000865, -0.00000759, 0.00000143, 0.00000684, 0.00000313, -0.00001233, 0.00000405, -0.00000700, 0.00000477, -0.00000228, 0.00000437, 0.00000379, -0.00000486, 0.00000209, -0.00001032, 0.00001192, -0.00001525, },
		{ -0.00002911, 0.00002454, -0.00001665, 0.00001723, -0.00000381, 0.00001438, -0.00000079, 0.00002775, -0.00002475, 0.00002820, -0.00002994, 0.00000707, 0.00000933, -0.00001917, 0.00001123, -0.00001884, 0.00001004, -0.00000647, -0.00000169, -0.00001100, 0.00001643, -0.00002977, 0.00002742, -0.00002026, 0.00001125, 0.00000023, -0.00000050, -0.00000172, 0.00000264, 0.00000179, 0.00000151, -0.00000160, 0.00000189, -0.00000403, 0.00000248, -0.00001566, 0.00001249, -0.00000030, 0.00000274, -0.00000431, 0.00000213, -0.00001070, 0.00001136, -0.00000739, 0.00000465, -0.00000409, 0.00000411, -0.00000847, 0.00000675, 0.00000427, 0.00000383, -0.00002036, 0.00000996, -0.00000023, 0.00000653, -0.00001763, 0.00001441, -0.00000090, 0.00000343, -0.00002299, 0.00001071, -0.00001058, 0.00001088, -0.00001730, 0.00000357, -0.00000779, 0.00001751, -0.00002952, 0.00000566, -0.00001651, 0.00001685, -0.00006671, 0.00001771, -0.00015854, -0.00017347, 0.00090944, -0.00031125, -0.00006683, -0.00004818, 0.00000699, -0.00002177, -0.00000752, -0.00001822, 0.00001552, -0.00000756, 0.00000309, -0.00001078, 0.00001009, -0.00000920, 0.00000514, -0.00001503, 0.00001678, -0.00001152, -0.00000004, -0.00000738, 0.00001729, -0.00001421, 0.00000489, -0.00000440, 0.00001613, -0.00001399, -0.00000412, 0.00000069, 0.00000784, -0.00000640, -0.00000212, -0.00000325, 0.00000657, -0.00000446, 0.00001250, -0.00001122, 0.00000291, -0.00000245, -0.00000063, -0.00000163, 0.00001486, -0.00001323, 0.00000142, -0.00000590, 0.00000240, -0.00000194, 0.00000377, 0.00000160, 0.00000365, -0.00000703, 0.00000277, -0.00001005, 0.00001531, },
		{ -0.00000751, -0.00000977, 0.00001536, -0.00001242, 0.00001882, -0.00000289, 0.00001245, 0.00001020, 0.00001042, -0.00001359, 0.00002258, -0.00004387, 0.00004127, -0.00003148, 0.00001503, -0.00001725, 0.00000378, -0.00000500, 0.00000509, -0.00001846, 0.00001389, -0.00001254, -0.00000419, 0.00001365, -0.00001806, 0.00001831, -0.00000829, 0.00000421, -0.00000437, 0.00000600, -0.00000047, 0.00000132, -0.00000088, -0.00000013, -0.00000093, -0.00000493, -0.00000851, 0.00001499, -0.00000544, 0.00000447, -0.00000568, 0.00000064, -0.00000717, 0.00001194, -0.00001220, 0.00001061, -0.00000946, 0.00000772, -0.00001283, 0.00001665, -0.00000288, 0.00000043, -0.00002034, 0.00002083, -0.00000955, 0.00000843, -0.00002120, 0.00002811, -0.00001620, 0.00000831, -0.00003050, 0.00002747, -0.00002784, 0.00002766, -0.00004094, 0.00003155, -0.00002959, 0.00003623, -0.00006343, 0.00005058, -0.00006026, 0.00005517, -0.00013520, 0.00011100, -0.00043698, 0.00050656, 0.00044357, -0.00040691, 0.00007699, -0.00010711, 0.00006220, -0.00006843, 0.00002719, -0.00003605, 0.00003708, -0.00003045, 0.00001923, -0.00002093, 0.00001999, -0.00001906, 0.00000973, -0.00001491, 0.00002114, -0.00002245, 0.00000811, -0.00000643, 0.00001629, -0.00002025, 0.00001280, -0.00000562, 0.00001667, -0.00002522, 0.00000898, -0.00000365, 0.00000992, -0.00001284, 0.00000276, -0.00000269, 0.00000529, -0.00000102, 0.00000873, -0.00001326, 0.00000773, -0.00000646, 0.00000039, 0.00000497, 0.00000693, -0.00001373, 0.00000408, -0.00000692, 0.00000398, -0.00000176, 0.00000388, 0.00000383, -0.00000383, 0.00000083, -0.00000856, 0.00000756, },
		{ 0.00001729, -0.00002834, 0.00001794, -0.00000911, 0.00000979, 0.00000473, 0.00000939, 0.00000396, 0.00002495, -0.00001801, 0.00002082, -0.00002322, -0.00000045, 0.00001526, -0.00002389, 0.00001530, -0.00002248, 0.00001136, -0.00000890, 0.00000309, -0.00001681, 0.00001999, -0.00002979, 0.00002245, -0.00001332, 0.00000391, 0.00000632, -0.00000355, 0.00000028, 0.00000007, 0.00000361, 0.00000081, -0.00000137, 0.00000214, -0.00000399, 0.00000316, -0.00001499, 0.00000817, 0.00000370, 0.00000071, -0.00000291, 0.00000130, -0.00000880, 0.00000753, -0.00000346, 0.00000080, -0.00000046, 0.00000099, -0.00000504, 0.00000215, 0.00000785, 0.00000300, -0.00001741, 0.00000339, 0.00000535, 0.00000306, -0.00001250, 0.00000598, 0.00000727, -0.00000165, -0.00001736, 0.00000204, -0.00000087, 0.00000071, -0.00000638, -0.00000803, 0.00000347, 0.00000529, -0.00001539, -0.00000981, 0.00000022, -0.00000105, -0.00004121, -0.00001477, -0.00010022, -0.00023287, 0.00084098, -0.00016695, -0.00014565, -0.00001153, -0.00002157, 0.00000015, -0.00002447, -0.00000578, 0.00000446, 0.00000162, -0.00000437, -0.00000401, 0.00000405, -0.00000363, 0.00000022, -0.00001023, 0.00001134, -0.00000506, -0.00000475, -0.00000495, 0.00001323, -0.00000796, -0.00000080, -0.00000054, 0.00001208, -0.00000693, -0.00001035, 0.00000322, 0.00000549, -0.00000272, -0.00000500, -0.00000210, 0.00000534, -0.00000348, 0.00001065, -0.00000678, -0.00000099, -0.00000011, -0.00000180, -0.00000229, 0.00001408, -0.00000911, -0.00000225, -0.00000337, 0.00000018, -0.00000087, 0.00000318, 0.00000142, 0.00000453, -0.00000772, 0.00000400, -0.00001254, },
		{ 0.00000239, 0.00000032, -0.00001883, 0.00001980, -0.00001454, 0.00001987, -0.00000038, 0.00001160, 0.00001123, 0.00001293, -0.00001371, 0.00002090, -0.00004197, 0.00003711, -0.00002693, 0.00001059, -0.00001277, -0.00000132, -0.00000233, 0.00000290, -0.00001405, 0.00000600, -0.00000310, -0.00001347, 0.00001907, -0.00002003, 0.00001714, -0.00000508, 0.00000210, -0.00000325, 0.00000404, 0.00000166, -0.00000009, -0.00000004, -0.00000005, -0.00000116, -0.00000288, -0.00001144, 0.00001454, -0.00000330, 0.00000311, -0.00000496, 0.00000132, -0.00000819, 0.00001142, -0.00001101, 0.00000895, -0.00000754, 0.00000627, -0.00001152, 0.00001368, 0.00000050, 0.00000067, -0.00002122, 0.00001818, -0.00000607, 0.00000746, -0.00002061, 0.00002422, -0.00001028, 0.00000496, -0.00002867, 0.00002372, -0.00002263, 0.00002158, -0.00003371, 0.00002279, -0.00002072, 0.00002686, -0.00005222, 0.00003739, -0.00004520, 0.00004039, -0.00011410, 0.00008414, -0.00036651, 0.00035539, 0.00053364, -0.00037543, 0.00003665, -0.00008497, 0.00004381, -0.00005270, 0.00001371, -0.00002615, 0.00002840, -0.00002308, 0.00001353, -0.00001616, 0.00001612, -0.00001580, 0.00000743, -0.00001353, 0.00001981, -0.00001936, 0.00000538, -0.00000674, 0.00001692, -0.00001877, 0.00001047, -0.00000483, 0.00001709, -0.00002241, 0.00000424, -0.00000148, 0.00000900, -0.00001065, 0.00000032, -0.00000191, 0.00000513, -0.00000208, 0.00001044, -0.00001241, 0.00000576, -0.00000482, -0.00000064, 0.00000309, 0.00000984, -0.00001369, 0.00000297, -0.00000599, 0.00000269, -0.00000111, 0.00000352, 0.00000351, -0.00000236, -0.00000094, -0.00000601, },
		{ -0.00001330, 0.00001654, -0.00002404, 0.00000840, -0.00000049, 0.00000212, 0.00001274, 0.00000547, 0.00000797, 0.00002267, -0.00001168, 0.00001434, -0.00001779, -0.00000652, 0.00001947, -0.00002686, 0.00001762, -0.00002422, 0.00001110, -0.00001011, 0.00000622, -0.00001980, 0.00001981, -0.00002589, 0.00001456, -0.00000506, -0.00000347, 0.00001143, -0.00000535, 0.00000136, -0.00000164, 0.00000441, 0.00000106, -0.00000183, 0.00000273, -0.00000401, 0.00000353, -0.00001356, 0.00000361, 0.00000698, -0.00000043, -0.00000214, 0.00000066, -0.00000682, 0.00000398, -0.00000024, -0.00000216, 0.00000214, -0.00000112, -0.00000255, -0.00000124, 0.00000988, 0.00000316, -0.00001472, -0.00000242, 0.00000936, 0.00000100, -0.00000834, -0.00000084, 0.00001277, -0.00000404, -0.00001372, -0.00000463, 0.00000619, -0.00000615, 0.00000073, -0.00001524, 0.00000983, -0.00000155, -0.00000726, -0.00001851, 0.00000924, -0.00001039, -0.00002589, -0.00003374, -0.00006356, -0.00025753, 0.00074382, -0.00002381, -0.00020642, 0.00000992, -0.00003917, 0.00001303, -0.00003361, 0.00000014, -0.00000125, 0.00000631, -0.00000814, -0.00000030, 0.00000059, -0.00000020, -0.00000293, -0.00000693, 0.00000718, -0.00000006, -0.00000788, -0.00000340, 0.00000945, -0.00000237, -0.00000531, 0.00000230, 0.00000859, -0.00000079, -0.00001476, 0.00000402, 0.00000416, 0.00000002, -0.00000670, -0.00000190, 0.00000458, -0.00000258, 0.00000852, -0.00000246, -0.00000407, 0.00000131, -0.00000230, -0.00000296, 0.00001259, -0.00000466, -0.00000550, -0.00000145, -0.00000139, -0.00000037, 0.00000289, 0.00000124, 0.00000504, -0.00000770, 0.00000421, },
		{ -0.00000337, -0.00000254, 0.00000686, -0.00002477, 0.00001991, -0.00001291, 0.00001755, 0.00000492, 0.00000885, 0.00001403, 0.00001360, -0.00001165, 0.00001716, -0.00003832, 0.00003140, -0.00002121, 0.00000523, -0.00000753, -0.00000676, 0.00000069, -0.00000020, -0.00000852, -0.00000222, 0.00000521, -0.00002004, 0.00002098, -0.00001876, 0.00001340, -0.00000039, -0.00000064, -0.00000165, 0.00000170, 0.00000378, -0.00000119, 0.00000038, 0.00000038, -0.00000163, -0.00000092, -0.00001335, 0.00001278, -0.00000051, 0.00000159, -0.00000420, 0.00000166, -0.00000829, 0.00000986, -0.00000896, 0.00000655, -0.00000506, 0.00000428, -0.00000948, 0.00001010, 0.00000389, 0.00000079, -0.00002093, 0.00001408, -0.00000176, 0.00000565, -0.00001839, 0.00001854, -0.00000345, 0.00000115, -0.00002575, 0.00001826, -0.00001563, 0.00001385, -0.00002494, 0.00001268, -0.00001074, 0.00001649, -0.00004012, 0.00002349, -0.00002971, 0.00002498, -0.00009194, 0.00005600, -0.00029978, 0.00022969, 0.00057977, -0.00031178, -0.00001618, -0.00005740, 0.00002173, -0.00003452, -0.00000108, -0.00001556, 0.00001929, -0.00001546, 0.00000757, -0.00001108, 0.00001191, -0.00001209, 0.00000455, -0.00001141, 0.00001748, -0.00001540, 0.00000228, -0.00000642, 0.00001613, -0.00001575, 0.00000709, -0.00000310, 0.00001600, -0.00001794, -0.00000121, 0.00000083, 0.00000763, -0.00000791, -0.00000227, -0.00000121, 0.00000490, -0.00000274, 0.00001107, -0.00001031, 0.00000314, -0.00000301, -0.00000152, 0.00000130, 0.00001168, -0.00001229, 0.00000095, -0.00000450, 0.00000115, -0.00000050, 0.00000334, 0.00000296, -0.00000071, -0.00000280, },
		{ 0.00000338, -0.00001242, 0.00001349, -0.00001728, -0.00000230, 0.00000731, -0.00000419, 0.00001888, 0.00000367, 0.00001037, 0.00002178, -0.00000666, 0.00000961, -0.00001440, -0.00001046, 0.00002141, -0.00002765, 0.00001784, -0.00002383, 0.00000910, -0.00000984, 0.00000732, -0.00001977, 0.00001617, -0.00001900, 0.00000513, 0.00000306, -0.00000966, 0.00001469, -0.00000543, 0.00000129, -0.00000236, 0.00000415, 0.00000213, -0.00000275, 0.00000344, -0.00000393, 0.00000350, -0.00001154, -0.00000080, 0.00000925, -0.00000059, -0.00000195, 0.00000023, -0.00000496, 0.00000102, 0.00000199, -0.00000402, 0.00000356, -0.00000214, -0.00000110, -0.00000331, 0.00001038, 0.00000423, -0.00001243, -0.00000708, 0.00001147, 0.00000055, -0.00000553, -0.00000553, 0.00001527, -0.00000371, -0.00001212, -0.00000905, 0.00001020, -0.00000935, 0.00000377, -0.00001795, 0.00001128, -0.00000317, -0.00000480, -0.00002096, 0.00001132, -0.00001217, -0.00001957, -0.00004089, -0.00004512, -0.00025569, 0.00063170, 0.00010497, -0.00024186, 0.00001347, -0.00004469, 0.00001632, -0.00003480, -0.00000030, -0.00000192, 0.00000686, -0.00000858, 0.00000062, -0.00000050, 0.00000123, -0.00000434, -0.00000515, 0.00000441, 0.00000336, -0.00000940, -0.00000276, 0.00000624, 0.00000208, -0.00000827, 0.00000384, 0.00000605, 0.00000391, -0.00001698, 0.00000303, 0.00000388, 0.00000173, -0.00000716, -0.00000260, 0.00000423, -0.00000178, 0.00000636, 0.00000134, -0.00000600, 0.00000171, -0.00000216, -0.00000358, 0.00001059, -0.00000032, -0.00000792, -0.00000038, -0.00000219, -0.00000042, 0.00000279, 0.00000118, 0.00000506, -0.00000691, },
		{ -0.00000427, -0.00000126, -0.00000635, 0.00001111, -0.00002683, 0.00001568, -0.00000820, 0.00001274, 0.00001203, 0.00000511, 0.00001781, 0.00001321, -0.00000826, 0.00001231, -0.00003391, 0.00002516, -0.00001534, -0.00000006, -0.00000250, -0.00001163, 0.00000328, -0.00000335, -0.00000301, -0.00000932, 0.00001100, -0.00002295, 0.00001915, -0.00001459, 0.00000776, 0.00000500, -0.00000335, -0.00000009, -0.00000057, 0.00000549, -0.00000170, 0.00000025, 0.00000115, -0.00000225, 0.00000076, -0.00001420, 0.00000997, 0.00000252, 0.00000023, -0.00000356, 0.00000174, -0.00000766, 0.00000763, -0.00000651, 0.00000390, -0.00000249, 0.00000219, -0.00000720, 0.00000648, 0.00000681, 0.00000106, -0.00001981, 0.00000917, 0.00000264, 0.00000370, -0.00001534, 0.00001221, 0.00000304, -0.00000202, -0.00002273, 0.00001224, -0.00000828, 0.00000616, -0.00001649, 0.00000326, -0.00000182, 0.00000744, -0.00002963, 0.00001158, -0.00001673, 0.00001218, -0.00007268, 0.00003174, -0.00024322, 0.00013364, 0.00058514, -0.00022510, -0.00007255, -0.00003083, 0.00000071, -0.00001772, -0.00001406, -0.00000679, 0.00001174, -0.00000920, 0.00000265, -0.00000678, 0.00000827, -0.00000871, 0.00000180, -0.00000920, 0.00001477, -0.00001122, -0.00000068, -0.00000578, 0.00001430, -0.00001179, 0.00000325, -0.00000094, 0.00001394, -0.00001259, -0.00000655, 0.00000266, 0.00000627, -0.00000508, -0.00000456, -0.00000086, 0.00000472, -0.00000304, 0.00001075, -0.00000734, 0.00000033, -0.00000137, -0.00000212, -0.00000031, 0.00001244, -0.00000974, -0.00000160, -0.00000280, -0.00000035, -0.00000010, 0.00000331, 0.00000234, 0.00000078, },
		{ -0.00000548, 0.00000177, -0.00001038, 0.00000901, -0.00000951, -0.00001229, 0.00001255, -0.00000789, 0.00002224, 0.00000463, 0.00001070, 0.00002276, -0.00000340, 0.00000703, -0.00001335, -0.00001211, 0.00002102, -0.00002631, 0.00001608, -0.00002154, 0.00000563, -0.00000824, 0.00000649, -0.00001707, 0.00000995, -0.00001054, -0.00000418, 0.00000962, -0.00001363, 0.00001554, -0.00000364, 0.00000012, -0.00000217, 0.00000299, 0.00000378, -0.00000387, 0.00000404, -0.00000359, 0.00000306, -0.00000916, -0.00000469, 0.00001030, 0.00000018, -0.00000225, -0.00000001, -0.00000336, -0.00000114, 0.00000314, -0.00000478, 0.00000385, -0.00000215, -0.00000066, -0.00000412, 0.00000955, 0.00000598, -0.00001055, -0.00001040, 0.00001160, 0.00000166, -0.00000409, -0.00000801, 0.00001494, -0.00000112, -0.00001214, -0.00001141, 0.00001133, -0.00000921, 0.00000324, -0.00001683, 0.00000872, -0.00000067, -0.00000669, -0.00001871, 0.00000829, -0.00000852, -0.00001974, -0.00003942, -0.00003979, -0.00023670, 0.00051753, 0.00020925, -0.00024852, -0.00000041, -0.00003939, 0.00001130, -0.00002938, -0.00000573, 0.00000121, 0.00000438, -0.00000661, -0.00000047, 0.00000015, 0.00000116, -0.00000439, -0.00000464, 0.00000287, 0.00000532, -0.00000948, -0.00000294, 0.00000374, 0.00000515, -0.00000953, 0.00000402, 0.00000462, 0.00000695, -0.00001698, 0.00000046, 0.00000444, 0.00000249, -0.00000652, -0.00000400, 0.00000413, -0.00000107, 0.00000436, 0.00000432, -0.00000666, 0.00000115, -0.00000152, -0.00000412, 0.00000831, 0.00000350, -0.00000923, -0.00000027, -0.00000224, -0.00000092, 0.00000277, 0.00000132, 0.00000462, },
		{ 0.00000184, -0.00000501, -0.00000016, -0.00000852, 0.00001260, -0.00002504, 0.00000797, -0.00000175, 0.00000675, 0.00001966, 0.00000155, 0.00002156, 0.00001276, -0.00000457, 0.00000741, -0.00002981, 0.00001943, -0.00001031, -0.00000437, 0.00000144, -0.00001518, 0.00000479, -0.00000580, 0.00000146, -0.00001422, 0.00001345, -0.00002199, 0.00001401, -0.00000836, 0.00000122, 0.00001016, -0.00000534, 0.00000097, -0.00000238, 0.00000648, -0.00000148, -0.00000043, 0.00000218, -0.00000285, 0.00000200, -0.00001402, 0.00000646, 0.00000539, -0.00000070, -0.00000315, 0.00000166, -0.00000653, 0.00000512, -0.00000409, 0.00000143, -0.00000025, 0.00000038, -0.00000512, 0.00000328, 0.00000888, 0.00000171, -0.00001821, 0.00000409, 0.00000642, 0.00000224, -0.00001224, 0.00000629, 0.00000813, -0.00000376, -0.00002035, 0.00000663, -0.00000181, -0.00000013, -0.00000981, -0.00000396, 0.00000448, 0.00000131, -0.00002239, 0.00000338, -0.00000803, 0.00000395, -0.00005865, 0.00001439, -0.00019998, 0.00006709, 0.00055698, -0.00012660, -0.00012330, -0.00001106, -0.00001527, -0.00000536, -0.00002288, -0.00000160, 0.00000707, -0.00000533, -0.00000041, -0.00000393, 0.00000577, -0.00000617, -0.00000035, -0.00000737, 0.00001221, -0.00000733, -0.00000312, -0.00000512, 0.00001185, -0.00000751, -0.00000042, 0.00000116, 0.00001149, -0.00000712, -0.00001101, 0.00000353, 0.00000529, -0.00000257, -0.00000618, -0.00000108, 0.00000467, -0.00000304, 0.00000970, -0.00000392, -0.00000223, -0.00000019, -0.00000235, -0.00000171, 0.00001219, -0.00000639, -0.00000428, -0.00000122, -0.00000159, 0.00000004, 0.00000336, 0.00000185, },
		{ 0.00000385, -0.00000370, -0.00000019, -0.00000784, 0.00000414, -0.00000221, -0.00001988, 0.00001405, -0.00000845, 0.00002257, 0.00000835, 0.00000905, 0.00002555, -0.00000185, 0.00000647, -0.00001440, -0.00001181, 0.00001874, -0.00002334, 0.00001290, -0.00001795, 0.00000129, -0.00000579, 0.00000424, -0.00001254, 0.00000247, -0.00000213, -0.00001182, 0.00001348, -0.00001478, 0.00001385, -0.00000019, -0.00000187, -0.00000134, 0.00000122, 0.00000565, -0.00000484, 0.00000431, -0.00000290, 0.00000228, -0.00000667, -0.00000777, 0.00001011, 0.00000171, -0.00000285, -0.00000013, -0.00000210, -0.00000243, 0.00000326, -0.00000458, 0.00000324, -0.00000138, -0.00000099, -0.00000393, 0.00000775, 0.00000808, -0.00000899, -0.00001239, 0.00000997, 0.00000398, -0.00000377, -0.00000857, 0.00001239, 0.00000291, -0.00001308, -0.00001226, 0.00001018, -0.00000654, 0.00000018, -0.00001312, 0.00000358, 0.00000430, -0.00001107, -0.00001385, 0.00000253, -0.00000212, -0.00002337, -0.00003316, -0.00004196, -0.00020951, 0.00041190, 0.00028294, -0.00022707, -0.00002821, -0.00002654, 0.00000070, -0.00001978, -0.00001399, 0.00000628, 0.00000041, -0.00000349, -0.00000256, 0.00000173, 0.00000023, -0.00000357, -0.00000498, 0.00000223, 0.00000610, -0.00000845, -0.00000371, 0.00000197, 0.00000678, -0.00000919, 0.00000300, 0.00000424, 0.00000834, -0.00001504, -0.00000323, 0.00000547, 0.00000255, -0.00000506, -0.00000576, 0.00000406, -0.00000042, 0.00000266, 0.00000631, -0.00000607, -0.00000017, -0.00000060, -0.00000450, 0.00000594, 0.00000651, -0.00000931, -0.00000107, -0.00000166, -0.00000173, 0.00000273, 0.00000164, },
		{ 0.00000164, 0.00000229, -0.00000487, -0.00000019, -0.00000893, 0.00001150, -0.00002015, -0.00000171, 0.00000470, 0.00000108, 0.00002648, -0.00000071, 0.00002435, 0.00001319, -0.00000154, 0.00000342, -0.00002693, 0.00001505, -0.00000686, -0.00000705, 0.00000374, -0.00001697, 0.00000484, -0.00000706, 0.00000423, -0.00001633, 0.00001238, -0.00001763, 0.00000661, -0.00000132, -0.00000503, 0.00001415, -0.00000603, 0.00000120, -0.00000347, 0.00000659, -0.00000050, -0.00000155, 0.00000330, -0.00000328, 0.00000270, -0.00001294, 0.00000263, 0.00000772, -0.00000103, -0.00000305, 0.00000149, -0.00000515, 0.00000273, -0.00000207, -0.00000054, 0.00000139, -0.00000088, -0.00000352, 0.00000085, 0.00000991, 0.00000282, -0.00001642, -0.00000059, 0.00000901, 0.00000174, -0.00000972, 0.00000156, 0.00001116, -0.00000367, -0.00001900, 0.00000206, 0.00000294, -0.00000415, -0.00000574, -0.00000817, 0.00000737, -0.00000119, -0.00001905, -0.00000051, -0.00000414, 0.00000079, -0.00005051, 0.00000475, -0.00017029, 0.00002643, 0.00050500, -0.00002784, -0.00016056, -0.00000219, -0.00002375, 0.00000082, -0.00002634, -0.00000073, 0.00000577, -0.00000414, -0.00000139, -0.00000274, 0.00000460, -0.00000467, -0.00000166, -0.00000619, 0.00001009, -0.00000408, -0.00000478, -0.00000465, 0.00000921, -0.00000350, -0.00000341, 0.00000274, 0.00000918, -0.00000223, -0.00001404, 0.00000313, 0.00000491, -0.00000065, -0.00000692, -0.00000193, 0.00000476, -0.00000282, 0.00000817, -0.00000052, -0.00000414, 0.00000035, -0.00000219, -0.00000286, 0.00001108, -0.00000265, -0.00000666, -0.00000006, -0.00000239, -0.00000013, 0.00000341, },
		{ 0.00000208, 0.00000298, -0.00000194, -0.00000197, -0.00000554, -0.00000010, 0.00000336, -0.00002398, 0.00001144, -0.00000612, 0.00002027, 0.00001430, 0.00000597, 0.00002969, -0.00000151, 0.00000736, -0.00001689, -0.00001027, 0.00001536, -0.00001956, 0.00000910, -0.00001391, -0.00000315, -0.00000317, 0.00000129, -0.00000724, -0.00000486, 0.00000472, -0.00001659, 0.00001405, -0.00001303, 0.00000996, 0.00000437, -0.00000416, -0.00000025, -0.00000077, 0.00000736, -0.00000540, 0.00000409, -0.00000185, 0.00000127, -0.00000432, -0.00000988, 0.00000877, 0.00000373, -0.00000352, -0.00000020, -0.00000120, -0.00000291, 0.00000256, -0.00000371, 0.00000202, -0.00000013, -0.00000183, -0.00000307, 0.00000540, 0.00001016, -0.00000760, -0.00001320, 0.00000701, 0.00000695, -0.00000409, -0.00000779, 0.00000853, 0.00000735, -0.00001409, -0.00001229, 0.00000756, -0.00000241, -0.00000410, -0.00000830, -0.00000246, 0.00000989, -0.00001595, -0.00000854, -0.00000359, 0.00000446, -0.00002754, -0.00002567, -0.00004660, -0.00018147, 0.00032206, 0.00032436, -0.00018194, -0.00006414, -0.00001064, -0.00001200, -0.00000892, -0.00002263, 0.00001131, -0.00000347, -0.00000050, -0.00000465, 0.00000341, -0.00000089, -0.00000244, -0.00000573, 0.00000211, 0.00000608, -0.00000669, -0.00000481, 0.00000084, 0.00000710, -0.00000756, 0.00000111, 0.00000468, 0.00000837, -0.00001166, -0.00000740, 0.00000649, 0.00000225, -0.00000313, -0.00000753, 0.00000380, 0.00000021, 0.00000133, 0.00000730, -0.00000445, -0.00000193, 0.00000036, -0.00000467, 0.00000366, 0.00000853, -0.00000819, -0.00000260, -0.00000067, -0.00000266, 0.00000260, },
		{ 0.00000339, 0.00000175, 0.00000214, -0.00000390, -0.00000123, -0.00000788, 0.00000846, -0.00001346, -0.00001151, 0.00000943, -0.00000300, 0.00003144, -0.00000085, 0.00002550, 0.00001518, 0.00000017, 0.00000094, -0.00002577, 0.00001243, -0.00000531, -0.00000788, 0.00000423, -0.00001694, 0.00000344, -0.00000700, 0.00000508, -0.00001562, 0.00000831, -0.00001095, -0.00000166, 0.00000519, -0.00000991, 0.00001624, -0.00000511, 0.00000048, -0.00000378, 0.00000583, 0.00000108, -0.00000290, 0.00000431, -0.00000340, 0.00000284, -0.00001119, -0.00000112, 0.00000924, -0.00000069, -0.00000323, 0.00000130, -0.00000374, 0.00000073, -0.00000068, -0.00000182, 0.00000229, -0.00000148, -0.00000257, -0.00000068, 0.00000986, 0.00000439, -0.00001466, -0.00000444, 0.00001005, 0.00000243, -0.00000810, -0.00000154, 0.00001192, -0.00000181, -0.00001869, -0.00000123, 0.00000563, -0.00000559, -0.00000447, -0.00000927, 0.00000693, -0.00000026, -0.00001930, -0.00000053, -0.00000447, 0.00000200, -0.00004749, 0.00000186, -0.00015208, 0.00000580, 0.00043982, 0.00006085, -0.00017890, -0.00000599, -0.00002411, 0.00000060, -0.00002453, -0.00000388, 0.00000746, -0.00000528, -0.00000060, -0.00000297, 0.00000462, -0.00000414, -0.00000212, -0.00000571, 0.00000855, -0.00000161, -0.00000558, -0.00000449, 0.00000672, -0.00000021, -0.00000534, 0.00000353, 0.00000740, 0.00000160, -0.00001530, 0.00000140, 0.00000514, 0.00000057, -0.00000672, -0.00000335, 0.00000492, -0.00000242, 0.00000644, 0.00000250, -0.00000515, 0.00000017, -0.00000170, -0.00000373, 0.00000935, 0.00000107, -0.00000839, 0.00000046, -0.00000267, -0.00000057, },
		{ 0.00000234, 0.00000252, 0.00000225, -0.00000056, -0.00000313, -0.00000404, -0.00000299, 0.00000645, -0.00002424, 0.00000519, -0.00000186, 0.00001630, 0.00002150, 0.00000239, 0.00003435, -0.00000153, 0.00000883, -0.00001993, -0.00000844, 0.00001179, -0.00001587, 0.00000557, -0.00001026, -0.00000692, -0.00000102, -0.00000158, -0.00000224, -0.00001078, 0.00000890, -0.00001788, 0.00001134, -0.00000888, 0.00000459, 0.00000925, -0.00000614, 0.00000067, -0.00000261, 0.00000859, -0.00000535, 0.00000336, -0.00000052, 0.00000021, -0.00000234, -0.00001092, 0.00000650, 0.00000591, -0.00000406, -0.00000030, -0.00000062, -0.00000273, 0.00000134, -0.00000249, 0.00000054, 0.00000126, -0.00000285, -0.00000190, 0.00000293, 0.00001188, -0.00000621, -0.00001309, 0.00000325, 0.00000994, -0.00000447, -0.00000636, 0.00000431, 0.00001122, -0.00001443, -0.00001215, 0.00000433, 0.00000204, -0.00000830, -0.00000378, -0.00000789, 0.00001452, -0.00001966, -0.00000454, -0.00000820, 0.00000923, -0.00003007, -0.00001969, -0.00004995, -0.00015761, 0.00025159, 0.00033590, -0.00012035, -0.00010126, 0.00000349, -0.00002333, 0.00000045, -0.00002945, 0.00001461, -0.00000597, 0.00000137, -0.00000596, 0.00000456, -0.00000167, -0.00000145, -0.00000647, 0.00000213, 0.00000563, -0.00000459, -0.00000595, 0.00000019, 0.00000638, -0.00000511, -0.00000118, 0.00000557, 0.00000746, -0.00000749, -0.00001137, 0.00000706, 0.00000192, -0.00000111, -0.00000894, 0.00000315, 0.00000087, 0.00000040, 0.00000739, -0.00000210, -0.00000375, 0.00000112, -0.00000459, 0.00000160, 0.00000949, -0.00000606, -0.00000456, 0.00000049, -0.00000355, },
		{ -0.00000123, 0.00000325, 0.00000217, 0.00000152, -0.00000238, -0.00000286, -0.00000600, 0.00000448, -0.00000646, -0.00001967, 0.00001109, -0.00000475, 0.00003399, 0.00000150, 0.00002476, 0.00001898, 0.00000036, 0.00000011, -0.00002640, 0.00001155, -0.00000553, -0.00000706, 0.00000319, -0.00001544, 0.00000093, -0.00000587, 0.00000422, -0.00001263, 0.00000225, -0.00000339, -0.00000928, 0.00000993, -0.00001259, 0.00001605, -0.00000256, -0.00000106, -0.00000342, 0.00000438, 0.00000301, -0.00000421, 0.00000501, -0.00000310, 0.00000247, -0.00000904, -0.00000444, 0.00000979, 0.00000029, -0.00000365, 0.00000112, -0.00000249, -0.00000068, -0.00000002, -0.00000239, 0.00000246, -0.00000143, -0.00000228, -0.00000131, 0.00000884, 0.00000630, -0.00001304, -0.00000721, 0.00000946, 0.00000425, -0.00000741, -0.00000297, 0.00001066, 0.00000138, -0.00001906, -0.00000334, 0.00000632, -0.00000470, -0.00000559, -0.00000785, 0.00000386, 0.00000321, -0.00002209, 0.00000211, -0.00000764, 0.00000603, -0.00004790, 0.00000363, -0.00014196, -0.00000155, 0.00037132, 0.00013173, -0.00017606, -0.00002166, -0.00001761, -0.00000485, -0.00001862, -0.00000989, 0.00001110, -0.00000787, 0.00000125, -0.00000409, 0.00000541, -0.00000426, -0.00000194, -0.00000581, 0.00000754, 0.00000007, -0.00000555, -0.00000467, 0.00000462, 0.00000206, -0.00000608, 0.00000343, 0.00000635, 0.00000411, -0.00001477, -0.00000145, 0.00000584, 0.00000115, -0.00000569, -0.00000515, 0.00000503, -0.00000189, 0.00000471, 0.00000485, -0.00000515, -0.00000064, -0.00000098, -0.00000432, 0.00000724, 0.00000438, -0.00000922, 0.00000027, -0.00000244, },
		{ -0.00000425, 0.00000194, 0.00000284, 0.00000186, 0.00000020, -0.00000340, -0.00000364, -0.00000419, 0.00000690, -0.00002108, -0.00000347, 0.00000284, 0.00001189, 0.00002880, -0.00000065, 0.00003863, -0.00000101, 0.00000996, -0.00002258, -0.00000719, 0.00000890, -0.00001307, 0.00000302, -0.00000766, -0.00000947, 0.00000014, -0.00000376, 0.00000162, -0.00001440, 0.00000987, -0.00001576, 0.00000598, -0.00000323, -0.00000125, 0.00001356, -0.00000721, 0.00000104, -0.00000399, 0.00000911, -0.00000460, 0.00000218, 0.00000095, -0.00000074, -0.00000088, -0.00001096, 0.00000361, 0.00000792, -0.00000428, -0.00000051, -0.00000030, -0.00000210, -0.00000010, -0.00000125, -0.00000090, 0.00000251, -0.00000378, -0.00000075, 0.00000072, 0.00001297, -0.00000467, -0.00001236, -0.00000071, 0.00001230, -0.00000437, -0.00000494, 0.00000060, 0.00001377, -0.00001361, -0.00001228, 0.00000116, 0.00000590, -0.00001141, -0.00000061, -0.00001164, 0.00001710, -0.00002114, -0.00000291, -0.00001022, 0.00001108, -0.00002975, -0.00001672, -0.00004993, -0.00014047, 0.00020065, 0.00032306, -0.00005108, -0.00013263, 0.00001175, -0.00003056, 0.00000628, -0.00003293, 0.00001513, -0.00000633, 0.00000160, -0.00000609, 0.00000484, -0.00000180, -0.00000086, -0.00000691, 0.00000202, 0.00000505, -0.00000248, -0.00000690, -0.00000018, 0.00000499, -0.00000234, -0.00000341, 0.00000653, 0.00000609, -0.00000320, -0.00001455, 0.00000680, 0.00000182, 0.00000067, -0.00000973, 0.00000202, 0.00000156, -0.00000016, 0.00000677, 0.00000057, -0.00000528, 0.00000150, -0.00000425, -0.00000015, 0.00000945, -0.00000320, -0.00000661, 0.00000156, },
		{ -0.00000177, -0.00000204, 0.00000299, 0.00000277, 0.00000068, -0.00000069, -0.00000456, -0.00000401, 0.00000061, -0.00000056, -0.00002483, 0.00000904, -0.00000410, 0.00003417, 0.00000619, 0.00002236, 0.00002441, -0.00000073, 0.00000060, -0.00002843, 0.00001195, -0.00000702, -0.00000515, 0.00000121, -0.00001310, -0.00000211, -0.00000413, 0.00000218, -0.00000820, -0.00000452, 0.00000358, -0.00001492, 0.00001204, -0.00001267, 0.00001360, 0.00000130, -0.00000308, -0.00000265, 0.00000252, 0.00000497, -0.00000521, 0.00000525, -0.00000238, 0.00000171, -0.00000676, -0.00000704, 0.00000931, 0.00000178, -0.00000418, 0.00000097, -0.00000148, -0.00000145, -0.00000003, -0.00000235, 0.00000201, -0.00000086, -0.00000255, -0.00000117, 0.00000712, 0.00000833, -0.00001159, -0.00000883, 0.00000741, 0.00000690, -0.00000742, -0.00000306, 0.00000796, 0.00000514, -0.00001957, -0.00000462, 0.00000544, -0.00000211, -0.00000827, -0.00000486, -0.00000068, 0.00000793, -0.00002601, 0.00000587, -0.00001193, 0.00001104, -0.00004967, 0.00000753, -0.00013619, -0.00000205, 0.00030749, 0.00018038, -0.00015306, -0.00004608, -0.00000696, -0.00001333, -0.00001055, -0.00001706, 0.00001529, -0.00001079, 0.00000331, -0.00000544, 0.00000645, -0.00000467, -0.00000140, -0.00000627, 0.00000689, 0.00000108, -0.00000483, -0.00000516, 0.00000303, 0.00000322, -0.00000566, 0.00000251, 0.00000606, 0.00000530, -0.00001267, -0.00000503, 0.00000670, 0.00000124, -0.00000405, -0.00000706, 0.00000493, -0.00000125, 0.00000317, 0.00000638, -0.00000420, -0.00000190, -0.00000019, -0.00000462, 0.00000497, 0.00000699, -0.00000901, -0.00000060, },
		{ 0.00000231, -0.00000465, 0.00000141, 0.00000298, 0.00000190, 0.00000027, -0.00000274, -0.00000434, -0.00000383, 0.00000514, -0.00001556, -0.00001284, 0.00000638, 0.00000821, 0.00003518, -0.00000226, 0.00004175, 0.00000086, 0.00001000, -0.00002415, -0.00000717, 0.00000726, -0.00001166, 0.00000188, -0.00000648, -0.00001053, 0.00000013, -0.00000492, 0.00000383, -0.00001539, 0.00000774, -0.00001091, -0.00000093, 0.00000273, -0.00000649, 0.00001651, -0.00000696, 0.00000067, -0.00000476, 0.00000883, -0.00000322, 0.00000069, 0.00000237, -0.00000140, -0.00000001, -0.00001013, 0.00000045, 0.00000948, -0.00000406, -0.00000085, -0.00000015, -0.00000122, -0.00000145, -0.00000025, -0.00000206, 0.00000341, -0.00000442, 0.00000014, -0.00000095, 0.00001326, -0.00000289, -0.00001131, -0.00000432, 0.00001355, -0.00000342, -0.00000402, -0.00000204, 0.00001459, -0.00001149, -0.00001286, -0.00000149, 0.00000854, -0.00001281, 0.00000066, -0.00001320, 0.00001719, -0.00002005, -0.00000393, -0.00000942, 0.00000986, -0.00002643, -0.00001704, -0.00004602, -0.00013034, 0.00016675, 0.00029327, 0.00001706, -0.00015260, 0.00001147, -0.00003225, 0.00000758, -0.00003249, 0.00001261, -0.00000448, 0.00000017, -0.00000504, 0.00000422, -0.00000125, -0.00000078, -0.00000694, 0.00000163, 0.00000452, -0.00000058, -0.00000748, -0.00000045, 0.00000329, 0.00000027, -0.00000518, 0.00000722, 0.00000469, 0.00000065, -0.00001645, 0.00000551, 0.00000209, 0.00000201, -0.00000974, 0.00000039, 0.00000223, -0.00000038, 0.00000567, 0.00000321, -0.00000625, 0.00000139, -0.00000368, -0.00000153, 0.00000855, 0.00000002, -0.00000842, },
		{ -0.00000203, -0.00000080, -0.00000289, 0.00000263, 0.00000339, -0.00000009, 0.00000074, -0.00000580, -0.00000258, -0.00000227, 0.00000326, -0.00002632, 0.00000344, -0.00000173, 0.00003257, 0.00001258, 0.00001892, 0.00003089, -0.00000245, 0.00000176, -0.00003115, 0.00001289, -0.00000900, -0.00000293, -0.00000098, -0.00001065, -0.00000500, -0.00000237, -0.00000036, -0.00000336, -0.00001069, 0.00000868, -0.00001772, 0.00001120, -0.00001028, 0.00000936, 0.00000586, -0.00000506, -0.00000183, 0.00000057, 0.00000664, -0.00000570, 0.00000497, -0.00000128, 0.00000072, -0.00000462, -0.00000876, 0.00000791, 0.00000356, -0.00000469, 0.00000081, -0.00000078, -0.00000161, -0.00000056, -0.00000187, 0.00000114, 0.00000004, -0.00000320, -0.00000048, 0.00000498, 0.00001023, -0.00001024, -0.00000940, 0.00000428, 0.00000986, -0.00000768, -0.00000233, 0.00000463, 0.00000868, -0.00001967, -0.00000550, 0.00000354, 0.00000137, -0.00001151, -0.00000143, -0.00000546, 0.00001255, -0.00002962, 0.00000925, -0.00001573, 0.00001530, -0.00005094, 0.00001128, -0.00013147, -0.00000089, 0.00025364, 0.00020587, -0.00011381, -0.00007452, 0.00000432, -0.00002223, -0.00000249, -0.00002362, 0.00001867, -0.00001301, 0.00000477, -0.00000643, 0.00000729, -0.00000498, -0.00000081, -0.00000683, 0.00000642, 0.00000159, -0.00000362, -0.00000583, 0.00000196, 0.00000334, -0.00000432, 0.00000102, 0.00000636, 0.00000537, -0.00000944, -0.00000883, 0.00000739, 0.00000110, -0.00000210, -0.00000878, 0.00000448, -0.00000052, 0.00000192, 0.00000707, -0.00000249, -0.00000336, 0.00000049, -0.00000464, 0.00000275, 0.00000870, -0.00000780, },
		{ -0.00000967, 0.00000258, -0.00000467, 0.00000077, 0.00000289, 0.00000235, -0.00000026, -0.00000138, -0.00000581, -0.00000239, 0.00000202, -0.00000911, -0.00002112, 0.00000737, 0.00000604, 0.00003993, -0.00000185, 0.00004322, 0.00000455, 0.00000853, -0.00002432, -0.00000864, 0.00000705, -0.00001173, 0.00000216, -0.00000667, -0.00001023, -0.00000095, -0.00000504, 0.00000434, -0.00001390, 0.00000319, -0.00000447, -0.00000805, 0.00000780, -0.00001021, 0.00001754, -0.00000521, -0.00000046, -0.00000490, 0.00000782, -0.00000139, -0.00000087, 0.00000356, -0.00000166, 0.00000025, -0.00000867, -0.00000261, 0.00001035, -0.00000337, -0.00000132, -0.00000012, -0.00000032, -0.00000249, 0.00000036, -0.00000282, 0.00000384, -0.00000465, 0.00000061, -0.00000195, 0.00001272, -0.00000088, -0.00001018, -0.00000718, 0.00001337, -0.00000147, -0.00000382, -0.00000335, 0.00001369, -0.00000831, -0.00001377, -0.00000350, 0.00000973, -0.00001234, 0.00000000, -0.00001264, 0.00001500, -0.00001670, -0.00000716, -0.00000633, 0.00000621, -0.00002081, -0.00001996, -0.00003897, -0.00012586, 0.00014573, 0.00025445, 0.00007645, -0.00015759, 0.00000194, -0.00002841, 0.00000457, -0.00002852, 0.00000755, -0.00000092, -0.00000248, -0.00000316, 0.00000296, -0.00000017, -0.00000110, -0.00000659, 0.00000095, 0.00000411, 0.00000096, -0.00000759, -0.00000077, 0.00000163, 0.00000232, -0.00000620, 0.00000741, 0.00000361, 0.00000362, -0.00001685, 0.00000321, 0.00000274, 0.00000280, -0.00000895, -0.00000163, 0.00000280, -0.00000032, 0.00000433, 0.00000549, -0.00000646, 0.00000078, -0.00000295, -0.00000254, 0.00000701, 0.00000322, },
		{ -0.00000574, -0.00000379, 0.00000029, -0.00000370, 0.00000221, 0.00000386, -0.00000053, 0.00000160, -0.00000621, -0.00000215, -0.00000365, 0.00000456, -0.00002426, -0.00000472, 0.00000114, 0.00003012, 0.00001981, 0.00001529, 0.00003764, -0.00000398, 0.00000272, -0.00003373, 0.00001355, -0.00001068, -0.00000115, -0.00000265, -0.00000877, -0.00000715, -0.00000109, -0.00000276, 0.00000097, -0.00001524, 0.00001109, -0.00001737, 0.00000764, -0.00000601, 0.00000409, 0.00001036, -0.00000648, -0.00000131, -0.00000116, 0.00000776, -0.00000554, 0.00000417, 0.00000007, -0.00000030, -0.00000285, -0.00000953, 0.00000578, 0.00000538, -0.00000503, 0.00000062, -0.00000036, -0.00000128, -0.00000140, -0.00000118, 0.00000007, 0.00000107, -0.00000402, 0.00000048, 0.00000277, 0.00001178, -0.00000889, -0.00000914, 0.00000058, 0.00001256, -0.00000770, -0.00000137, 0.00000145, 0.00001127, -0.00001890, -0.00000633, 0.00000119, 0.00000490, -0.00001436, 0.00000144, -0.00000942, 0.00001600, -0.00003185, 0.00001113, -0.00001790, 0.00001764, -0.00005043, 0.00001328, -0.00012551, -0.00000146, 0.00021223, 0.00021042, -0.00006424, -0.00010156, 0.00001273, -0.00002911, 0.00000368, -0.00002812, 0.00002019, -0.00001377, 0.00000512, -0.00000666, 0.00000760, -0.00000493, -0.00000038, -0.00000728, 0.00000594, 0.00000178, -0.00000212, -0.00000656, 0.00000134, 0.00000263, -0.00000243, -0.00000072, 0.00000703, 0.00000463, -0.00000561, -0.00001228, 0.00000756, 0.00000095, -0.00000014, -0.00001002, 0.00000355, 0.00000028, 0.00000101, 0.00000701, -0.00000029, -0.00000472, 0.00000092, -0.00000441, 0.00000075, 0.00000945, },
		{ 0.00000606, -0.00001017, 0.00000230, -0.00000433, 0.00000003, 0.00000261, 0.00000309, -0.00000115, 0.00000032, -0.00000755, -0.00000058, -0.00000141, -0.00000320, -0.00002683, 0.00000502, 0.00000559, 0.00004288, 0.00000070, 0.00004298, 0.00001014, 0.00000558, -0.00002318, -0.00001142, 0.00000803, -0.00001299, 0.00000353, -0.00000786, -0.00000898, -0.00000270, -0.00000437, 0.00000341, -0.00001055, -0.00000274, 0.00000219, -0.00001408, 0.00001101, -0.00001179, 0.00001645, -0.00000209, -0.00000213, -0.00000457, 0.00000626, 0.00000065, -0.00000226, 0.00000436, -0.00000147, -0.00000001, -0.00000683, -0.00000526, 0.00001043, -0.00000225, -0.00000187, -0.00000013, 0.00000046, -0.00000309, 0.00000053, -0.00000315, 0.00000381, -0.00000446, 0.00000062, -0.00000223, 0.00001142, 0.00000131, -0.00000911, -0.00000903, 0.00001175, 0.00000136, -0.00000425, -0.00000345, 0.00001146, -0.00000458, -0.00001468, -0.00000494, 0.00000955, -0.00001028, -0.00000216, -0.00001050, 0.00001119, -0.00001190, -0.00001170, -0.00000195, 0.00000124, -0.00001409, -0.00002423, -0.00003027, -0.00012478, 0.00013287, 0.00021380, 0.00012170, -0.00014664, -0.00001555, -0.00002045, -0.00000156, -0.00002216, 0.00000103, 0.00000344, -0.00000559, -0.00000101, 0.00000145, 0.00000114, -0.00000162, -0.00000597, 0.00000008, 0.00000380, 0.00000213, -0.00000721, -0.00000122, 0.00000024, 0.00000357, -0.00000636, 0.00000700, 0.00000304, 0.00000549, -0.00001575, 0.00000008, 0.00000362, 0.00000309, -0.00000746, -0.00000386, 0.00000314, -0.00000003, 0.00000297, 0.00000718, -0.00000588, -0.00000021, -0.00000217, -0.00000317, 0.00000507, },
		{ 0.00000927, -0.00000307, -0.00000560, 0.00000131, -0.00000438, 0.00000176, 0.00000408, -0.00000047, 0.00000171, -0.00000565, -0.00000283, -0.00000345, 0.00000352, -0.00001949, -0.00001392, 0.00000305, 0.00002778, 0.00002699, 0.00001232, 0.00004388, -0.00000450, 0.00000271, -0.00003544, 0.00001327, -0.00001145, -0.00000035, -0.00000332, -0.00000792, -0.00000821, -0.00000061, -0.00000455, 0.00000413, -0.00001752, 0.00001057, -0.00001418, 0.00000212, -0.00000077, -0.00000126, 0.00001400, -0.00000686, -0.00000133, -0.00000246, 0.00000818, -0.00000473, 0.00000299, 0.00000151, -0.00000116, -0.00000159, -0.00000940, 0.00000319, 0.00000699, -0.00000510, 0.00000039, -0.00000018, -0.00000065, -0.00000230, -0.00000048, -0.00000100, 0.00000203, -0.00000484, 0.00000148, 0.00000077, 0.00001275, -0.00000745, -0.00000834, -0.00000313, 0.00001442, -0.00000703, -0.00000072, -0.00000096, 0.00001247, -0.00001710, -0.00000728, -0.00000116, 0.00000784, -0.00001613, 0.00000308, -0.00001189, 0.00001762, -0.00003211, 0.00001098, -0.00001790, 0.00001757, -0.00004761, 0.00001280, -0.00011718, -0.00000532, 0.00018319, 0.00019852, -0.00001120, -0.00012215, 0.00001554, -0.00003228, 0.00000674, -0.00002971, 0.00001934, -0.00001279, 0.00000419, -0.00000602, 0.00000729, -0.00000443, -0.00000024, -0.00000749, 0.00000534, 0.00000179, -0.00000052, -0.00000721, 0.00000103, 0.00000140, -0.00000038, -0.00000237, 0.00000777, 0.00000347, -0.00000175, -0.00001489, 0.00000696, 0.00000098, 0.00000157, -0.00001058, 0.00000211, 0.00000110, 0.00000045, 0.00000636, 0.00000207, -0.00000573, 0.00000100, -0.00000396, -0.00000092, },
		{ 0.00000296, 0.00000827, -0.00000980, 0.00000148, -0.00000365, -0.00000079, 0.00000221, 0.00000392, -0.00000205, 0.00000191, -0.00000899, 0.00000089, -0.00000420, 0.00000102, -0.00002904, -0.00000068, 0.00000639, 0.00004431, 0.00000511, 0.00004136, 0.00001733, 0.00000155, -0.00002122, -0.00001501, 0.00000966, -0.00001486, 0.00000540, -0.00000949, -0.00000736, -0.00000460, -0.00000335, 0.00000154, -0.00000617, -0.00000884, 0.00000778, -0.00001798, 0.00001175, -0.00001106, 0.00001341, 0.00000199, -0.00000397, -0.00000402, 0.00000441, 0.00000259, -0.00000326, 0.00000466, -0.00000084, -0.00000064, -0.00000490, -0.00000727, 0.00000968, -0.00000081, -0.00000244, -0.00000015, 0.00000102, -0.00000323, 0.00000031, -0.00000312, 0.00000336, -0.00000391, 0.00000018, -0.00000188, 0.00000955, 0.00000352, -0.00000817, -0.00000982, 0.00000889, 0.00000471, -0.00000502, -0.00000272, 0.00000852, -0.00000097, -0.00001519, -0.00000602, 0.00000828, -0.00000713, -0.00000515, -0.00000757, 0.00000668, -0.00000666, -0.00001646, 0.00000255, -0.00000379, -0.00000757, -0.00002842, -0.00002158, -0.00012469, 0.00012382, 0.00017689, 0.00015010, -0.00012133, -0.00003804, -0.00001074, -0.00000896, -0.00001499, -0.00000559, 0.00000751, -0.00000836, 0.00000081, 0.00000012, 0.00000235, -0.00000210, -0.00000527, -0.00000085, 0.00000353, 0.00000293, -0.00000639, -0.00000184, -0.00000073, 0.00000394, -0.00000572, 0.00000607, 0.00000304, 0.00000624, -0.00001336, -0.00000352, 0.00000450, 0.00000301, -0.00000548, -0.00000602, 0.00000313, 0.00000046, 0.00000173, 0.00000818, -0.00000460, -0.00000143, -0.00000146, -0.00000344, },
		{ -0.00000220, 0.00000829, -0.00000011, -0.00000719, 0.00000210, -0.00000485, 0.00000129, 0.00000397, 0.00000009, 0.00000112, -0.00000425, -0.00000441, -0.00000203, 0.00000084, -0.00001331, -0.00002237, 0.00000269, 0.00002625, 0.00003345, 0.00001065, 0.00004901, -0.00000337, 0.00000118, -0.00003582, 0.00001164, -0.00001099, -0.00000080, -0.00000280, -0.00000822, -0.00000811, -0.00000099, -0.00000554, 0.00000579, -0.00001740, 0.00000745, -0.00000898, -0.00000427, 0.00000434, -0.00000575, 0.00001613, -0.00000591, -0.00000198, -0.00000323, 0.00000788, -0.00000337, 0.00000159, 0.00000284, -0.00000172, -0.00000091, -0.00000853, 0.00000045, 0.00000816, -0.00000484, 0.00000010, -0.00000016, 0.00000010, -0.00000307, 0.00000008, -0.00000194, 0.00000278, -0.00000548, 0.00000229, -0.00000077, 0.00001303, -0.00000582, -0.00000728, -0.00000635, 0.00001504, -0.00000542, -0.00000068, -0.00000226, 0.00001217, -0.00001439, -0.00000831, -0.00000329, 0.00000980, -0.00001648, 0.00000322, -0.00001267, 0.00001730, -0.00003036, 0.00000885, -0.00001586, 0.00001527, -0.00004267, 0.00000998, -0.00010649, -0.00001245, 0.00016451, 0.00017587, 0.00003873, -0.00013245, 0.00001133, -0.00003115, 0.00000633, -0.00002831, 0.00001625, -0.00001027, 0.00000219, -0.00000469, 0.00000644, -0.00000353, -0.00000036, -0.00000742, 0.00000458, 0.00000168, 0.00000106, -0.00000765, 0.00000086, -0.00000003, 0.00000145, -0.00000365, 0.00000835, 0.00000227, 0.00000167, -0.00001631, 0.00000546, 0.00000130, 0.00000284, -0.00001035, 0.00000022, 0.00000185, 0.00000022, 0.00000533, 0.00000429, -0.00000619, 0.00000071, -0.00000337, },
		{ -0.00000342, 0.00000089, 0.00000968, -0.00000859, 0.00000022, -0.00000272, -0.00000166, 0.00000177, 0.00000461, -0.00000266, 0.00000298, -0.00000968, 0.00000149, -0.00000567, 0.00000287, -0.00002765, -0.00000896, 0.00000739, 0.00004489, 0.00001081, 0.00003895, 0.00002554, -0.00000288, -0.00001914, -0.00001869, 0.00001123, -0.00001667, 0.00000711, -0.00001091, -0.00000600, -0.00000612, -0.00000243, -0.00000067, -0.00000168, -0.00001399, 0.00001131, -0.00001920, 0.00000996, -0.00000827, 0.00000896, 0.00000640, -0.00000551, -0.00000356, 0.00000256, 0.00000418, -0.00000370, 0.00000443, 0.00000015, -0.00000148, -0.00000313, -0.00000849, 0.00000822, 0.00000078, -0.00000293, -0.00000018, 0.00000131, -0.00000295, -0.00000017, -0.00000283, 0.00000261, -0.00000309, -0.00000058, -0.00000106, 0.00000735, 0.00000559, -0.00000733, -0.00000966, 0.00000518, 0.00000806, -0.00000568, -0.00000170, 0.00000557, 0.00000190, -0.00001494, -0.00000695, 0.00000632, -0.00000354, -0.00000820, -0.00000467, 0.00000236, -0.00000196, -0.00002046, 0.00000619, -0.00000786, -0.00000232, -0.00003141, -0.00001428, -0.00012363, 0.00011527, 0.00014719, 0.00016168, -0.00008532, -0.00006158, -0.00000202, -0.00001566, -0.00000860, -0.00001106, 0.00001039, -0.00001014, 0.00000184, -0.00000071, 0.00000323, -0.00000236, -0.00000464, -0.00000170, 0.00000321, 0.00000342, -0.00000518, -0.00000261, -0.00000124, 0.00000352, -0.00000448, 0.00000476, 0.00000350, 0.00000604, -0.00001006, -0.00000713, 0.00000511, 0.00000275, -0.00000329, -0.00000784, 0.00000264, 0.00000110, 0.00000074, 0.00000849, -0.00000281, -0.00000264, -0.00000093, },
		{ -0.00000274, -0.00000305, 0.00000671, 0.00000283, -0.00000832, 0.00000252, -0.00000506, 0.00000080, 0.00000359, 0.00000107, 0.00000004, -0.00000235, -0.00000645, -0.00000001, -0.00000251, -0.00000721, -0.00002850, -0.00000081, 0.00002572, 0.00003893, 0.00001059, 0.00005278, -0.00000030, -0.00000208, -0.00003479, 0.00000866, -0.00000934, -0.00000239, -0.00000124, -0.00000949, -0.00000710, -0.00000201, -0.00000577, 0.00000596, -0.00001518, 0.00000250, -0.00000291, -0.00001033, 0.00000835, -0.00000864, 0.00001637, -0.00000363, -0.00000318, -0.00000353, 0.00000694, -0.00000166, 0.00000021, 0.00000389, -0.00000188, -0.00000079, -0.00000713, -0.00000214, 0.00000875, -0.00000423, -0.00000024, -0.00000023, 0.00000081, -0.00000357, 0.00000042, -0.00000266, 0.00000325, -0.00000587, 0.00000279, -0.00000172, 0.00001257, -0.00000400, -0.00000621, -0.00000871, 0.00001423, -0.00000283, -0.00000130, -0.00000245, 0.00001063, -0.00001117, -0.00000920, -0.00000513, 0.00001065, -0.00001540, 0.00000201, -0.00001198, 0.00001536, -0.00002706, 0.00000526, -0.00001237, 0.00001141, -0.00003635, 0.00000554, -0.00009420, -0.00002175, 0.00015314, 0.00014832, 0.00008021, -0.00013051, 0.00000027, -0.00002624, 0.00000302, -0.00002453, 0.00001156, -0.00000680, -0.00000040, -0.00000304, 0.00000532, -0.00000240, -0.00000066, -0.00000713, 0.00000371, 0.00000147, 0.00000250, -0.00000779, 0.00000069, -0.00000138, 0.00000277, -0.00000436, 0.00000859, 0.00000131, 0.00000427, -0.00001638, 0.00000313, 0.00000187, 0.00000361, -0.00000933, -0.00000200, 0.00000241, 0.00000028, 0.00000413, 0.00000613, -0.00000604, 0.00000009, },
		{ -0.00000066, -0.00000315, -0.00000096, 0.00001020, -0.00000666, -0.00000130, -0.00000166, -0.00000254, 0.00000138, 0.00000501, -0.00000275, 0.00000329, -0.00000938, 0.00000096, -0.00000559, 0.00000231, -0.00002326, -0.00001837, 0.00000726, 0.00004532, 0.00001715, 0.00003641, 0.00003412, -0.00000695, -0.00001773, -0.00002174, 0.00001208, -0.00001778, 0.00000810, -0.00001160, -0.00000535, -0.00000687, -0.00000195, -0.00000273, 0.00000218, -0.00001739, 0.00001228, -0.00001771, 0.00000606, -0.00000409, 0.00000389, 0.00001040, -0.00000628, -0.00000344, 0.00000093, 0.00000520, -0.00000351, 0.00000374, 0.00000134, -0.00000231, -0.00000171, -0.00000890, 0.00000623, 0.00000231, -0.00000327, -0.00000020, 0.00000137, -0.00000237, -0.00000076, -0.00000241, 0.00000167, -0.00000214, -0.00000155, 0.00000005, 0.00000508, 0.00000731, -0.00000650, -0.00000880, 0.00000116, 0.00001085, -0.00000577, -0.00000090, 0.00000323, 0.00000357, -0.00001378, -0.00000785, 0.00000400, -0.00000007, -0.00001065, -0.00000247, -0.00000107, 0.00000154, -0.00002305, 0.00000834, -0.00001032, 0.00000104, -0.00003254, -0.00000920, -0.00012040, 0.00010521, 0.00012599, 0.00015871, -0.00004358, -0.00008199, 0.00000323, -0.00002005, -0.00000421, -0.00001449, 0.00001148, -0.00001056, 0.00000188, -0.00000089, 0.00000364, -0.00000229, -0.00000419, -0.00000236, 0.00000279, 0.00000364, -0.00000370, -0.00000347, -0.00000136, 0.00000250, -0.00000293, 0.00000332, 0.00000427, 0.00000515, -0.00000632, -0.00001030, 0.00000518, 0.00000248, -0.00000116, -0.00000907, 0.00000162, 0.00000184, 0.00000002, 0.00000822, -0.00000076, -0.00000365, },
		{ -0.00000073, -0.00000216, -0.00000349, 0.00000474, 0.00000543, -0.00000882, 0.00000252, -0.00000498, 0.00000026, 0.00000302, 0.00000223, -0.00000119, -0.00000041, -0.00000835, 0.00000189, -0.00000553, -0.00000248, -0.00003123, -0.00000753, 0.00002576, 0.00004357, 0.00001209, 0.00005523, 0.00000468, -0.00000690, -0.00003261, 0.00000466, -0.00000688, -0.00000469, 0.00000094, -0.00001131, -0.00000560, -0.00000331, -0.00000550, 0.00000495, -0.00001149, -0.00000324, 0.00000283, -0.00001502, 0.00001050, -0.00000955, 0.00001469, -0.00000026, -0.00000465, -0.00000353, 0.00000556, 0.00000015, -0.00000093, 0.00000453, -0.00000160, -0.00000113, -0.00000544, -0.00000434, 0.00000869, -0.00000335, -0.00000061, -0.00000033, 0.00000135, -0.00000373, 0.00000054, -0.00000316, 0.00000339, -0.00000594, 0.00000288, -0.00000204, 0.00001143, -0.00000202, -0.00000528, -0.00001003, 0.00001202, 0.00000050, -0.00000238, -0.00000181, 0.00000836, -0.00000802, -0.00000964, -0.00000675, 0.00001048, -0.00001318, -0.00000013, -0.00001032, 0.00001243, -0.00002293, 0.00000101, -0.00000826, 0.00000692, -0.00002959, 0.00000048, -0.00008148, -0.00003162, 0.00014581, 0.00012093, 0.00010981, -0.00011645, -0.00001598, -0.00001907, -0.00000199, -0.00001947, 0.00000627, -0.00000319, -0.00000294, -0.00000149, 0.00000423, -0.00000125, -0.00000098, -0.00000673, 0.00000282, 0.00000116, 0.00000375, -0.00000758, 0.00000040, -0.00000241, 0.00000340, -0.00000447, 0.00000843, 0.00000076, 0.00000590, -0.00001514, 0.00000017, 0.00000256, 0.00000392, -0.00000768, -0.00000430, 0.00000267, 0.00000061, 0.00000291, 0.00000745, -0.00000529, },
		{ -0.00000429, -0.00000068, -0.00000273, -0.00000247, 0.00000988, -0.00000422, -0.00000288, -0.00000061, -0.00000337, 0.00000107, 0.00000500, -0.00000224, 0.00000279, -0.00000811, -0.00000058, -0.00000415, -0.00000012, -0.00001709, -0.00002717, 0.00000474, 0.00004615, 0.00002363, 0.00003431, 0.00004249, -0.00000999, -0.00001760, -0.00002363, 0.00001173, -0.00001781, 0.00000803, -0.00001131, -0.00000566, -0.00000669, -0.00000205, -0.00000432, 0.00000489, -0.00001864, 0.00001070, -0.00001402, 0.00000084, 0.00000059, -0.00000094, 0.00001332, -0.00000595, -0.00000380, -0.00000032, 0.00000556, -0.00000274, 0.00000274, 0.00000257, -0.00000299, -0.00000077, -0.00000856, 0.00000393, 0.00000362, -0.00000340, -0.00000023, 0.00000123, -0.00000165, -0.00000130, -0.00000196, 0.00000065, -0.00000116, -0.00000259, 0.00000122, 0.00000299, 0.00000851, -0.00000558, -0.00000751, -0.00000266, 0.00001258, -0.00000493, -0.00000070, 0.00000188, 0.00000391, -0.00001182, -0.00000865, 0.00000155, 0.00000286, -0.00001210, -0.00000133, -0.00000329, 0.00000355, -0.00002402, 0.00000884, -0.00001107, 0.00000244, -0.00003171, -0.00000653, -0.00011462, 0.00009299, 0.00011279, 0.00014500, -0.00000146, -0.00009569, 0.00000336, -0.00002125, -0.00000242, -0.00001553, 0.00001070, -0.00000966, 0.00000100, -0.00000051, 0.00000362, -0.00000190, -0.00000393, -0.00000279, 0.00000225, 0.00000361, -0.00000203, -0.00000432, -0.00000120, 0.00000115, -0.00000138, 0.00000197, 0.00000514, 0.00000389, -0.00000263, -0.00001262, 0.00000455, 0.00000236, 0.00000068, -0.00000956, 0.00000006, 0.00000258, -0.00000039, 0.00000752, 0.00000130, },
		{ -0.00000407, -0.00000159, -0.00000172, -0.00000356, 0.00000262, 0.00000747, -0.00000860, 0.00000210, -0.00000461, -0.00000036, 0.00000240, 0.00000333, -0.00000222, 0.00000112, -0.00000961, 0.00000303, -0.00000742, 0.00000004, -0.00003023, -0.00001676, 0.00002538, 0.00004784, 0.00001484, 0.00005669, 0.00001126, -0.00001282, -0.00002985, 0.00000021, -0.00000420, -0.00000716, 0.00000319, -0.00001316, -0.00000412, -0.00000447, -0.00000506, 0.00000320, -0.00000712, -0.00000872, 0.00000719, -0.00001761, 0.00001050, -0.00000845, 0.00001139, 0.00000372, -0.00000601, -0.00000347, 0.00000397, 0.00000180, -0.00000165, 0.00000469, -0.00000092, -0.00000179, -0.00000371, -0.00000595, 0.00000800, -0.00000228, -0.00000095, -0.00000042, 0.00000168, -0.00000357, 0.00000051, -0.00000346, 0.00000323, -0.00000570, 0.00000257, -0.00000175, 0.00000975, 0.00000001, -0.00000454, -0.00001030, 0.00000872, 0.00000413, -0.00000350, -0.00000083, 0.00000601, -0.00000550, -0.00000935, -0.00000824, 0.00000949, -0.00001025, -0.00000264, -0.00000834, 0.00000921, -0.00001873, -0.00000310, -0.00000437, 0.00000266, -0.00002328, -0.00000424, -0.00006947, -0.00004054, 0.00013969, 0.00009732, 0.00012628, -0.00009228, -0.00003464, -0.00001168, -0.00000719, -0.00001440, 0.00000141, -0.00000021, -0.00000489, -0.00000044, 0.00000341, -0.00000028, -0.00000117, -0.00000634, 0.00000200, 0.00000071, 0.00000477, -0.00000699, -0.00000007, -0.00000303, 0.00000335, -0.00000406, 0.00000791, 0.00000067, 0.00000654, -0.00001280, -0.00000309, 0.00000319, 0.00000388, -0.00000559, -0.00000643, 0.00000247, 0.00000116, 0.00000180, 0.00000820, },
		{ 0.00000315, -0.00000446, -0.00000094, -0.00000226, -0.00000356, 0.00000882, -0.00000152, -0.00000429, 0.00000032, -0.00000410, 0.00000085, 0.00000461, -0.00000119, 0.00000164, -0.00000617, -0.00000275, -0.00000192, -0.00000353, -0.00001062, -0.00003368, -0.00000100, 0.00004742, 0.00003006, 0.00003299, 0.00005030, -0.00001157, -0.00001911, -0.00002411, 0.00001003, -0.00001667, 0.00000689, -0.00001005, -0.00000687, -0.00000568, -0.00000267, -0.00000533, 0.00000626, -0.00001777, 0.00000705, -0.00000902, -0.00000468, 0.00000487, -0.00000475, 0.00001468, -0.00000439, -0.00000464, -0.00000118, 0.00000528, -0.00000150, 0.00000161, 0.00000366, -0.00000336, -0.00000034, -0.00000761, 0.00000158, 0.00000455, -0.00000331, -0.00000026, 0.00000098, -0.00000092, -0.00000167, -0.00000153, -0.00000037, -0.00000025, -0.00000356, 0.00000228, 0.00000131, 0.00000908, -0.00000449, -0.00000609, -0.00000580, 0.00001293, -0.00000302, -0.00000123, 0.00000160, 0.00000308, -0.00000941, -0.00000916, -0.00000093, 0.00000502, -0.00001240, -0.00000127, -0.00000430, 0.00000415, -0.00002355, 0.00000794, -0.00001039, 0.00000222, -0.00002929, -0.00000591, -0.00010666, 0.00007898, 0.00010584, 0.00012499, 0.00003629, -0.00010036, -0.00000208, -0.00001928, -0.00000310, -0.00001445, 0.00000839, -0.00000779, -0.00000048, 0.00000021, 0.00000333, -0.00000128, -0.00000381, -0.00000303, 0.00000163, 0.00000335, -0.00000025, -0.00000507, -0.00000092, -0.00000025, -0.00000011, 0.00000089, 0.00000594, 0.00000256, 0.00000061, -0.00001384, 0.00000315, 0.00000245, 0.00000208, -0.00000921, -0.00000193, 0.00000323, -0.00000050, 0.00000654, },
		{ 0.00000844, -0.00000258, -0.00000232, -0.00000148, -0.00000334, 0.00000054, 0.00000879, -0.00000768, 0.00000132, -0.00000401, -0.00000107, 0.00000184, 0.00000413, -0.00000275, 0.00000191, -0.00000989, 0.00000309, -0.00000781, 0.00000011, -0.00002600, -0.00002710, 0.00002328, 0.00005228, 0.00001846, 0.00005759, 0.00001898, -0.00001923, -0.00002716, -0.00000406, -0.00000189, -0.00000923, 0.00000499, -0.00001457, -0.00000311, -0.00000511, -0.00000474, 0.00000117, -0.00000283, -0.00001303, 0.00000949, -0.00001784, 0.00000847, -0.00000575, 0.00000708, 0.00000767, -0.00000683, -0.00000357, 0.00000241, 0.00000305, -0.00000181, 0.00000437, 0.00000005, -0.00000261, -0.00000214, -0.00000689, 0.00000679, -0.00000117, -0.00000123, -0.00000047, 0.00000176, -0.00000317, 0.00000041, -0.00000359, 0.00000280, -0.00000519, 0.00000188, -0.00000097, 0.00000773, 0.00000193, -0.00000396, -0.00000967, 0.00000477, 0.00000750, -0.00000419, -0.00000002, 0.00000418, -0.00000405, -0.00000821, -0.00000962, 0.00000791, -0.00000707, -0.00000497, -0.00000658, 0.00000629, -0.00001509, -0.00000647, -0.00000129, -0.00000076, -0.00001803, -0.00000796, -0.00005898, -0.00004738, 0.00013280, 0.00007947, 0.00013040, -0.00006144, -0.00005243, -0.00000617, -0.00001115, -0.00001041, -0.00000217, 0.00000155, -0.00000588, -0.00000010, 0.00000302, 0.00000040, -0.00000115, -0.00000603, 0.00000134, 0.00000011, 0.00000553, -0.00000602, -0.00000074, -0.00000321, 0.00000270, -0.00000329, 0.00000715, 0.00000100, 0.00000632, -0.00000971, -0.00000626, 0.00000351, 0.00000365, -0.00000335, -0.00000811, 0.00000173, 0.00000188, 0.00000088, },
		{ 0.00000544, 0.00000464, -0.00000418, -0.00000135, -0.00000182, -0.00000420, 0.00000719, 0.00000115, -0.00000536, 0.00000101, -0.00000467, 0.00000066, 0.00000393, 0.00000020, 0.00000018, -0.00000400, -0.00000499, 0.00000039, -0.00000689, -0.00000525, -0.00003670, -0.00001008, 0.00004859, 0.00003661, 0.00003247, 0.00005748, -0.00001153, -0.00002231, -0.00002329, 0.00000716, -0.00001458, 0.00000493, -0.00000809, -0.00000870, -0.00000414, -0.00000359, -0.00000585, 0.00000639, -0.00001518, 0.00000214, -0.00000375, -0.00000952, 0.00000796, -0.00000701, 0.00001427, -0.00000171, -0.00000578, -0.00000173, 0.00000447, -0.00000002, 0.00000056, 0.00000445, -0.00000335, -0.00000041, -0.00000623, -0.00000060, 0.00000501, -0.00000302, -0.00000028, 0.00000067, -0.00000031, -0.00000184, -0.00000113, -0.00000136, 0.00000050, -0.00000437, 0.00000307, 0.00000017, 0.00000896, -0.00000319, -0.00000478, -0.00000796, 0.00001180, -0.00000017, -0.00000236, 0.00000215, 0.00000155, -0.00000707, -0.00000909, -0.00000343, 0.00000637, -0.00001166, -0.00000207, -0.00000442, 0.00000374, -0.00002211, 0.00000617, -0.00000886, 0.00000101, -0.00002593, -0.00000668, -0.00009729, 0.00006414, 0.00010286, 0.00010292, 0.00006612, -0.00009525, -0.00001229, -0.00001502, -0.00000551, -0.00001197, 0.00000526, -0.00000554, -0.00000210, 0.00000094, 0.00000298, -0.00000058, -0.00000375, -0.00000315, 0.00000103, 0.00000282, 0.00000154, -0.00000560, -0.00000067, -0.00000148, 0.00000069, 0.00000018, 0.00000656, 0.00000141, 0.00000311, -0.00001384, 0.00000108, 0.00000270, 0.00000299, -0.00000810, -0.00000418, 0.00000363, -0.00000033, },
		{ 0.00000018, 0.00000827, -0.00000102, -0.00000279, -0.00000147, -0.00000292, -0.00000133, 0.00000933, -0.00000616, 0.00000031, -0.00000324, -0.00000187, 0.00000145, 0.00000450, -0.00000262, 0.00000183, -0.00000913, 0.00000204, -0.00000677, -0.00000188, -0.00001966, -0.00003673, 0.00001818, 0.00005724, 0.00002268, 0.00005832, 0.00002743, -0.00002555, -0.00002513, -0.00000761, -0.00000042, -0.00001048, 0.00000601, -0.00001524, -0.00000281, -0.00000507, -0.00000472, -0.00000079, 0.00000080, -0.00001562, 0.00000950, -0.00001594, 0.00000494, -0.00000212, 0.00000249, 0.00001093, -0.00000676, -0.00000400, 0.00000105, 0.00000376, -0.00000139, 0.00000367, 0.00000119, -0.00000342, -0.00000091, -0.00000717, 0.00000521, -0.00000015, -0.00000141, -0.00000047, 0.00000165, -0.00000264, 0.00000035, -0.00000360, 0.00000213, -0.00000446, 0.00000092, 0.00000013, 0.00000561, 0.00000357, -0.00000343, -0.00000843, 0.00000072, 0.00001006, -0.00000402, 0.00000017, 0.00000330, -0.00000384, -0.00000633, -0.00001078, 0.00000588, -0.00000400, -0.00000673, -0.00000541, 0.00000405, -0.00001235, -0.00000880, 0.00000073, -0.00000311, -0.00001407, -0.00001042, -0.00005040, -0.00005164, 0.00012418, 0.00006779, 0.00012448, -0.00002805, -0.00006624, -0.00000414, -0.00001295, -0.00000818, -0.00000404, 0.00000189, -0.00000583, -0.00000046, 0.00000302, 0.00000078, -0.00000090, -0.00000584, 0.00000088, -0.00000061, 0.00000598, -0.00000470, -0.00000156, -0.00000302, 0.00000166, -0.00000241, 0.00000629, 0.00000163, 0.00000547, -0.00000629, -0.00000894, 0.00000335, 0.00000340, -0.00000121, -0.00000912, 0.00000040, 0.00000271, },
		{ 0.00000012, 0.00000434, 0.00000571, -0.00000354, -0.00000178, -0.00000153, -0.00000441, 0.00000518, 0.00000356, -0.00000594, 0.00000141, -0.00000503, 0.00000044, 0.00000311, 0.00000164, -0.00000122, -0.00000209, -0.00000672, 0.00000212, -0.00000933, -0.00000199, -0.00003575, -0.00002175, 0.00004855, 0.00004376, 0.00003260, 0.00006416, -0.00000996, -0.00002694, -0.00002152, 0.00000350, -0.00001198, 0.00000259, -0.00000585, -0.00001076, -0.00000246, -0.00000447, -0.00000607, 0.00000553, -0.00001147, -0.00000308, 0.00000079, -0.00001289, 0.00000940, -0.00000752, 0.00001220, 0.00000175, -0.00000690, -0.00000214, 0.00000331, 0.00000147, -0.00000018, 0.00000483, -0.00000294, -0.00000089, -0.00000464, -0.00000241, 0.00000500, -0.00000258, -0.00000029, 0.00000036, 0.00000011, -0.00000180, -0.00000073, -0.00000229, 0.00000106, -0.00000494, 0.00000348, -0.00000036, 0.00000820, -0.00000171, -0.00000370, -0.00000902, 0.00000936, 0.00000327, -0.00000371, 0.00000308, -0.00000005, -0.00000536, -0.00000820, -0.00000597, 0.00000697, -0.00001016, -0.00000334, -0.00000409, 0.00000283, -0.00002028, 0.00000415, -0.00000709, -0.00000054, -0.00002227, -0.00000813, -0.00008739, 0.00004964, 0.00010159, 0.00008217, 0.00008616, -0.00008129, -0.00002542, -0.00000995, -0.00000852, -0.00000905, 0.00000213, -0.00000358, -0.00000338, 0.00000136, 0.00000276, 0.00000005, -0.00000361, -0.00000324, 0.00000053, 0.00000201, 0.00000326, -0.00000584, -0.00000057, -0.00000235, 0.00000096, -0.00000018, 0.00000695, 0.00000058, 0.00000473, -0.00001271, -0.00000147, 0.00000302, 0.00000344, -0.00000638, -0.00000641, 0.00000363, },
		{ 0.00000354, -0.00000027, 0.00000781, 0.00000042, -0.00000291, -0.00000165, -0.00000239, -0.00000286, 0.00000910, -0.00000420, -0.00000077, -0.00000239, -0.00000273, 0.00000123, 0.00000439, -0.00000185, 0.00000098, -0.00000755, 0.00000024, -0.00000478, -0.00000511, -0.00001274, -0.00004384, 0.00000922, 0.00006258, 0.00002752, 0.00005906, 0.00003634, -0.00003138, -0.00002417, -0.00001013, -0.00000004, -0.00001076, 0.00000615, -0.00001511, -0.00000327, -0.00000436, -0.00000500, -0.00000250, 0.00000343, -0.00001629, 0.00000745, -0.00001253, 0.00000071, 0.00000167, -0.00000162, 0.00001296, -0.00000558, -0.00000482, -0.00000002, 0.00000385, -0.00000045, 0.00000271, 0.00000233, -0.00000407, -0.00000008, -0.00000684, 0.00000347, 0.00000066, -0.00000147, -0.00000042, 0.00000139, -0.00000211, 0.00000040, -0.00000349, 0.00000126, -0.00000359, -0.00000022, 0.00000135, 0.00000364, 0.00000477, -0.00000283, -0.00000688, -0.00000292, 0.00001138, -0.00000274, -0.00000045, 0.00000347, -0.00000473, -0.00000406, -0.00001147, 0.00000347, -0.00000128, -0.00000772, -0.00000494, 0.00000256, -0.00001054, -0.00001014, 0.00000178, -0.00000450, -0.00001124, -0.00001172, -0.00004366, -0.00005335, 0.00011373, 0.00006149, 0.00011174, 0.00000387, -0.00007374, -0.00000636, -0.00001231, -0.00000783, -0.00000422, 0.00000098, -0.00000497, -0.00000132, 0.00000326, 0.00000096, -0.00000047, -0.00000576, 0.00000060, -0.00000139, 0.00000608, -0.00000308, -0.00000247, -0.00000259, 0.00000046, -0.00000164, 0.00000547, 0.00000244, 0.00000426, -0.00000295, -0.00001082, 0.00000257, 0.00000323, 0.00000061, -0.00000930, -0.00000147, },
		{ 0.00000310, 0.00000082, 0.00000330, 0.00000640, -0.00000268, -0.00000209, -0.00000142, -0.00000425, 0.00000300, 0.00000550, -0.00000597, 0.00000151, -0.00000515, 0.00000013, 0.00000232, 0.00000286, -0.00000221, -0.00000085, -0.00000753, 0.00000282, -0.00001035, -0.00000123, -0.00003121, -0.00003450, 0.00004584, 0.00005201, 0.00003314, 0.00007059, -0.00000709, -0.00003263, -0.00001928, -0.00000043, -0.00000936, 0.00000034, -0.00000377, -0.00001265, -0.00000100, -0.00000506, -0.00000619, 0.00000402, -0.00000730, -0.00000775, 0.00000384, -0.00001439, 0.00000910, -0.00000642, 0.00000888, 0.00000543, -0.00000765, -0.00000261, 0.00000202, 0.00000272, -0.00000047, 0.00000475, -0.00000218, -0.00000164, -0.00000303, -0.00000376, 0.00000457, -0.00000209, -0.00000025, 0.00000010, 0.00000034, -0.00000162, -0.00000026, -0.00000314, 0.00000138, -0.00000522, 0.00000347, -0.00000030, 0.00000691, -0.00000015, -0.00000288, -0.00000906, 0.00000598, 0.00000677, -0.00000479, 0.00000383, -0.00000111, -0.00000475, -0.00000637, -0.00000850, 0.00000693, -0.00000821, -0.00000465, -0.00000372, 0.00000188, -0.00001856, 0.00000236, -0.00000557, -0.00000195, -0.00001881, -0.00000970, -0.00007772, 0.00003641, 0.00010029, 0.00006493, 0.00009624, -0.00006073, -0.00003898, -0.00000575, -0.00001098, -0.00000663, -0.00000026, -0.00000239, -0.00000400, 0.00000128, 0.00000279, 0.00000051, -0.00000334, -0.00000338, 0.00000023, 0.00000094, 0.00000484, -0.00000572, -0.00000070, -0.00000281, 0.00000072, -0.00000027, 0.00000713, 0.00000015, 0.00000548, -0.00001065, -0.00000417, 0.00000322, 0.00000355, -0.00000430, -0.00000835, },
		{ -0.00000374, 0.00000424, -0.00000047, 0.00000719, 0.00000162, -0.00000270, -0.00000193, -0.00000189, -0.00000397, 0.00000821, -0.00000203, -0.00000176, -0.00000156, -0.00000358, 0.00000115, 0.00000389, -0.00000062, -0.00000034, -0.00000559, -0.00000177, -0.00000255, -0.00000851, -0.00000677, -0.00004703, -0.00000370, 0.00006753, 0.00003337, 0.00005976, 0.00004567, -0.00003653, -0.00002441, -0.00001161, -0.00000069, -0.00001018, 0.00000557, -0.00001437, -0.00000429, -0.00000320, -0.00000544, -0.00000390, 0.00000495, -0.00001521, 0.00000392, -0.00000847, -0.00000337, 0.00000487, -0.00000464, 0.00001343, -0.00000330, -0.00000592, -0.00000082, 0.00000338, 0.00000086, 0.00000170, 0.00000330, -0.00000444, 0.00000029, -0.00000604, 0.00000174, 0.00000119, -0.00000143, -0.00000031, 0.00000105, -0.00000167, 0.00000058, -0.00000324, 0.00000022, -0.00000264, -0.00000141, 0.00000253, 0.00000201, 0.00000539, -0.00000207, -0.00000532, -0.00000573, 0.00001122, -0.00000035, -0.00000184, 0.00000451, -0.00000626, -0.00000194, -0.00001140, 0.00000067, 0.00000101, -0.00000798, -0.00000504, 0.00000164, -0.00000942, -0.00001078, 0.00000221, -0.00000530, -0.00000917, -0.00001223, -0.00003843, -0.00005297, 0.00010198, 0.00005908, 0.00009551, 0.00003105, -0.00007381, -0.00001255, -0.00000972, -0.00000893, -0.00000318, -0.00000071, -0.00000373, -0.00000232, 0.00000349, 0.00000109, 0.00000004, -0.00000571, 0.00000043, -0.00000213, 0.00000577, -0.00000120, -0.00000336, -0.00000208, -0.00000067, -0.00000114, 0.00000477, 0.00000331, 0.00000293, -0.00000002, -0.00001170, 0.00000117, 0.00000320, 0.00000197, -0.00000864, },
		{ -0.00000972, 0.00000193, 0.00000174, 0.00000235, 0.00000678, -0.00000181, -0.00000215, -0.00000154, -0.00000378, 0.00000083, 0.00000684, -0.00000546, 0.00000135, -0.00000505, -0.00000033, 0.00000169, 0.00000367, -0.00000255, -0.00000048, -0.00000729, 0.00000242, -0.00000992, -0.00000270, -0.00002423, -0.00004631, 0.00003899, 0.00006163, 0.00003404, 0.00007693, -0.00000313, -0.00003899, -0.00001704, -0.00000419, -0.00000714, -0.00000146, -0.00000213, -0.00001412, 0.00000000, -0.00000521, -0.00000633, 0.00000218, -0.00000326, -0.00001121, 0.00000502, -0.00001405, 0.00000738, -0.00000419, 0.00000493, 0.00000875, -0.00000764, -0.00000331, 0.00000079, 0.00000354, -0.00000021, 0.00000423, -0.00000114, -0.00000254, -0.00000156, -0.00000459, 0.00000383, -0.00000164, -0.00000017, -0.00000009, 0.00000038, -0.00000139, 0.00000033, -0.00000389, 0.00000142, -0.00000519, 0.00000305, 0.00000027, 0.00000529, 0.00000136, -0.00000226, -0.00000829, 0.00000218, 0.00000974, -0.00000510, 0.00000391, -0.00000116, -0.00000541, -0.00000373, -0.00001082, 0.00000628, -0.00000608, -0.00000569, -0.00000363, 0.00000120, -0.00001725, 0.00000106, -0.00000451, -0.00000299, -0.00001576, -0.00001110, -0.00006870, 0.00002502, 0.00009790, 0.00005210, 0.00009759, -0.00003662, -0.00005038, -0.00000386, -0.00001198, -0.00000537, -0.00000144, -0.00000226, -0.00000384, 0.00000067, 0.00000308, 0.00000079, -0.00000289, -0.00000363, 0.00000015, -0.00000036, 0.00000616, -0.00000520, -0.00000107, -0.00000288, 0.00000013, -0.00000026, 0.00000717, 0.00000009, 0.00000545, -0.00000798, -0.00000669, 0.00000311, 0.00000347, -0.00000212, },
		{ -0.00000722, -0.00000619, 0.00000462, -0.00000038, 0.00000649, 0.00000254, -0.00000225, -0.00000218, -0.00000149, -0.00000462, 0.00000677, 0.00000014, -0.00000252, -0.00000085, -0.00000434, 0.00000112, 0.00000316, 0.00000079, -0.00000173, -0.00000376, -0.00000340, -0.00000079, -0.00001114, -0.00000293, -0.00004567, -0.00001975, 0.00007070, 0.00004089, 0.00006020, 0.00005555, -0.00004099, -0.00002575, -0.00001226, -0.00000209, -0.00000907, 0.00000459, -0.00001333, -0.00000557, -0.00000186, -0.00000583, -0.00000507, 0.00000547, -0.00001279, -0.00000032, -0.00000466, -0.00000654, 0.00000694, -0.00000623, 0.00001231, -0.00000016, -0.00000707, -0.00000146, 0.00000248, 0.00000232, 0.00000085, 0.00000397, -0.00000446, 0.00000024, -0.00000492, 0.00000017, 0.00000141, -0.00000134, -0.00000015, 0.00000069, -0.00000137, 0.00000085, -0.00000280, -0.00000097, -0.00000169, -0.00000253, 0.00000348, 0.00000087, 0.00000539, -0.00000110, -0.00000396, -0.00000749, 0.00000962, 0.00000289, -0.00000364, 0.00000596, -0.00000780, -0.00000057, -0.00001026, -0.00000251, 0.00000285, -0.00000766, -0.00000544, 0.00000099, -0.00000863, -0.00001114, 0.00000244, -0.00000593, -0.00000743, -0.00001236, -0.00003427, -0.00005111, 0.00008970, 0.00005892, 0.00007872, 0.00005143, -0.00006666, -0.00002152, -0.00000622, -0.00001066, -0.00000164, -0.00000252, -0.00000264, -0.00000307, 0.00000347, 0.00000132, 0.00000051, -0.00000561, 0.00000028, -0.00000270, 0.00000501, 0.00000083, -0.00000412, -0.00000163, -0.00000155, -0.00000101, 0.00000420, 0.00000417, 0.00000170, 0.00000225, -0.00001152, -0.00000075, 0.00000327, 0.00000282, },
		{ -0.00000013, -0.00001029, 0.00000011, 0.00000281, 0.00000153, 0.00000698, -0.00000106, -0.00000190, -0.00000184, -0.00000311, -0.00000117, 0.00000755, -0.00000450, 0.00000102, -0.00000476, -0.00000092, 0.00000127, 0.00000400, -0.00000223, -0.00000092, -0.00000617, 0.00000120, -0.00000847, -0.00000559, -0.00001644, -0.00005509, 0.00002699, 0.00007237, 0.00003557, 0.00008316, 0.00000185, -0.00004576, -0.00001512, -0.00000752, -0.00000552, -0.00000267, -0.00000104, -0.00001510, 0.00000053, -0.00000493, -0.00000648, 0.00000019, 0.00000025, -0.00001314, 0.00000437, -0.00001228, 0.00000481, -0.00000144, 0.00000104, 0.00001112, -0.00000663, -0.00000432, -0.00000026, 0.00000381, 0.00000061, 0.00000337, 0.00000004, -0.00000346, -0.00000033, -0.00000494, 0.00000292, -0.00000132, -0.00000003, -0.00000021, 0.00000029, -0.00000122, 0.00000108, -0.00000450, 0.00000117, -0.00000486, 0.00000227, 0.00000120, 0.00000354, 0.00000263, -0.00000170, -0.00000704, -0.00000149, 0.00001165, -0.00000427, 0.00000302, -0.00000003, -0.00000722, -0.00000065, -0.00001264, 0.00000500, -0.00000394, -0.00000631, -0.00000389, 0.00000087, -0.00001637, 0.00000023, -0.00000388, -0.00000374, -0.00001305, -0.00001233, -0.00006046, 0.00001561, 0.00009409, 0.00004353, 0.00009240, -0.00001222, -0.00005751, -0.00000513, -0.00001117, -0.00000548, -0.00000135, -0.00000311, -0.00000303, -0.00000031, 0.00000351, 0.00000093, -0.00000228, -0.00000397, 0.00000031, -0.00000179, 0.00000712, -0.00000425, -0.00000165, -0.00000265, -0.00000063, -0.00000031, 0.00000714, 0.00000034, 0.00000486, -0.00000508, -0.00000870, 0.00000256, 0.00000335, },
		{ 0.00000324, -0.00000524, -0.00000851, 0.00000452, 0.00000000, 0.00000575, 0.00000320, -0.00000170, -0.00000228, -0.00000130, -0.00000484, 0.00000496, 0.00000213, -0.00000294, -0.00000030, -0.00000495, 0.00000105, 0.00000240, 0.00000208, -0.00000280, -0.00000252, -0.00000414, -0.00000002, -0.00001236, -0.00000178, -0.00004005, -0.00003721, 0.00007025, 0.00005077, 0.00006014, 0.00006616, -0.00004489, -0.00002796, -0.00001241, -0.00000386, -0.00000781, 0.00000359, -0.00001233, -0.00000682, -0.00000065, -0.00000599, -0.00000608, 0.00000517, -0.00000954, -0.00000448, -0.00000184, -0.00000835, 0.00000768, -0.00000638, 0.00000984, 0.00000341, -0.00000794, -0.00000209, 0.00000130, 0.00000371, 0.00000034, 0.00000425, -0.00000413, -0.00000016, -0.00000362, -0.00000112, 0.00000138, -0.00000126, 0.00000007, 0.00000036, -0.00000125, 0.00000115, -0.00000211, -0.00000227, -0.00000084, -0.00000346, 0.00000407, 0.00000032, 0.00000479, 0.00000006, -0.00000289, -0.00000818, 0.00000686, 0.00000649, -0.00000535, 0.00000721, -0.00000868, -0.00000045, -0.00000792, -0.00000599, 0.00000426, -0.00000697, -0.00000583, 0.00000030, -0.00000783, -0.00001159, 0.00000285, -0.00000672, -0.00000570, -0.00001239, -0.00003082, -0.00004830, 0.00007760, 0.00005958, 0.00006345, 0.00006431, -0.00005371, -0.00003138, -0.00000316, -0.00001206, -0.00000039, -0.00000381, -0.00000215, -0.00000327, 0.00000301, 0.00000175, 0.00000086, -0.00000537, 0.00000007, -0.00000300, 0.00000378, 0.00000293, -0.00000464, -0.00000135, -0.00000209, -0.00000123, 0.00000372, 0.00000498, 0.00000070, 0.00000376, -0.00001039, -0.00000299, 0.00000334, },
		},
		{ { -0.00000522, -0.00000277, 0.00001198, 0.00000302, -0.00000373, -0.00000211, -0.00000484, -0.00000307, 0.00000443, -0.00000170, -0.00000115, 0.00000178, -0.00000840, 0.00000118, 0.00000621, 0.00000089, -0.00000135, -0.00000393, -0.00001748, -0.00000501, -0.00000948, -0.00001123, -0.00005648, 0.00000844, 0.00008260, 0.00003788, 0.00009212, 0.00001079, -0.00005989, -0.00001041, -0.00001577, -0.00000353, -0.00000204, -0.00000270, 0.00000018, -0.00000332, -0.00000231, -0.00000255, -0.00000304, -0.00000027, -0.00000047, 0.00000171, -0.00000202, -0.00000374, 0.00001362, -0.00000272, 0.00000212, -0.00000299, -0.00000226, 0.00000134, -0.00001534, 0.00000317, 0.00000174, -0.00000322, -0.00000545, 0.00000257, -0.00001249, 0.00000303, -0.00000066, 0.00000144, 0.00000439, -0.00000624, 0.00001097, -0.00000304, -0.00000080, -0.00000304, 0.00001097, -0.00000624, 0.00000439, 0.00000144, -0.00000066, 0.00000303, -0.00001249, 0.00000257, -0.00000545, -0.00000322, 0.00000174, 0.00000317, -0.00001534, 0.00000134, -0.00000226, -0.00000299, 0.00000212, -0.00000272, 0.00001362, -0.00000374, -0.00000202, 0.00000171, -0.00000047, -0.00000027, -0.00000304, -0.00000255, -0.00000231, -0.00000332, 0.00000018, -0.00000270, -0.00000204, -0.00000353, -0.00001577, -0.00001041, -0.00005989, 0.00001079, 0.00009212, 0.00003788, 0.00008260, 0.00000844, -0.00005648, -0.00001123, -0.00000948, -0.00000501, -0.00001748, -0.00000393, -0.00000135, 0.00000089, 0.00000621, 0.00000118, -0.00000840, 0.00000178, -0.00000115, -0.00000170, 0.00000443, -0.00000307, -0.00000484, -0.00000211, -0.00000373, 0.00000302, 0.00001198, -0.00000277, },
		{ -0.00000512, -0.00000543, 0.00000334, 0.00001196, -0.00000286, -0.00000196, -0.00000373, -0.00000440, -0.00000064, 0.00000431, -0.00000430, 0.00000277, -0.00000382, -0.00000592, 0.00000452, 0.00000529, -0.00000173, -0.00000005, -0.00001083, -0.00001417, -0.00000489, -0.00000943, -0.00002401, -0.00005809, 0.00005784, 0.00006587, 0.00005651, 0.00008689, -0.00004450, -0.00003947, -0.00000825, -0.00001466, 0.00000075, -0.00000425, -0.00000048, -0.00000160, -0.00000303, -0.00000240, -0.00000285, -0.00000237, 0.00000061, -0.00000098, 0.00000266, -0.00000665, 0.00000550, 0.00000925, -0.00000481, 0.00000403, -0.00000732, 0.00000420, -0.00000870, -0.00000894, 0.00000568, -0.00000046, -0.00000614, -0.00000101, -0.00000173, -0.00001134, 0.00000756, -0.00000538, 0.00000833, -0.00000461, 0.00000183, 0.00000733, -0.00000359, -0.00000205, 0.00000181, 0.00000767, -0.00000740, 0.00000946, -0.00000502, 0.00000567, -0.00000586, -0.00000697, 0.00000081, -0.00000580, -0.00000139, 0.00000446, -0.00000500, -0.00001268, 0.00000603, -0.00000823, 0.00000377, -0.00000370, 0.00000605, 0.00000856, -0.00000698, 0.00000189, 0.00000010, 0.00000029, -0.00000181, -0.00000292, -0.00000231, -0.00000259, -0.00000272, 0.00000106, -0.00000486, 0.00000066, -0.00001129, -0.00001014, -0.00002846, -0.00005118, 0.00006347, 0.00006670, 0.00004561, 0.00007409, -0.00003270, -0.00003948, -0.00000782, -0.00000682, -0.00000961, -0.00001505, 0.00000001, -0.00000205, 0.00000367, 0.00000590, -0.00000406, -0.00000577, 0.00000384, -0.00000449, 0.00000235, 0.00000220, -0.00000491, -0.00000396, -0.00000186, -0.00000365, 0.00000982, 0.00000628, },
		{ -0.00000050, -0.00000520, -0.00000442, 0.00001046, 0.00000594, -0.00000420, -0.00000174, -0.00000454, -0.00000358, 0.00000218, 0.00000147, -0.00000317, 0.00000310, -0.00000824, -0.00000070, 0.00000541, 0.00000357, -0.00000315, -0.00000049, -0.00001729, -0.00000740, -0.00000915, -0.00000716, -0.00004671, -0.00003089, 0.00009047, 0.00004519, 0.00009040, 0.00003900, -0.00006890, -0.00001445, -0.00001555, -0.00000653, -0.00000084, -0.00000315, -0.00000067, -0.00000212, -0.00000333, -0.00000196, -0.00000374, -0.00000017, -0.00000066, 0.00000084, -0.00000039, -0.00000573, 0.00001282, 0.00000038, -0.00000061, -0.00000030, -0.00000441, 0.00000323, -0.00001469, 0.00000022, 0.00000385, -0.00000258, -0.00000738, 0.00000373, -0.00000956, -0.00000317, 0.00000429, -0.00000294, 0.00000823, -0.00000912, 0.00000982, 0.00000064, -0.00000162, -0.00000379, 0.00000787, -0.00000017, -0.00000105, 0.00000648, -0.00000494, 0.00000657, -0.00001173, -0.00000083, -0.00000333, -0.00000336, -0.00000011, 0.00000474, -0.00001400, -0.00000236, 0.00000107, -0.00000623, 0.00000476, -0.00000505, 0.00001339, -0.00000131, -0.00000344, 0.00000200, 0.00000024, -0.00000058, -0.00000227, -0.00000317, -0.00000144, -0.00000391, -0.00000008, -0.00000120, -0.00000353, -0.00000149, -0.00001457, -0.00000910, -0.00004913, -0.00001086, 0.00008520, 0.00004035, 0.00006495, 0.00004056, -0.00005093, -0.00002225, -0.00000756, -0.00000526, -0.00001450, -0.00000908, 0.00000078, -0.00000137, 0.00000623, 0.00000286, -0.00000765, -0.00000032, 0.00000134, -0.00000434, 0.00000510, -0.00000129, -0.00000556, -0.00000230, -0.00000337, 0.00000070, 0.00001227, },
		{ 0.00000863, -0.00000429, -0.00000543, 0.00000037, 0.00001317, -0.00000121, -0.00000242, -0.00000335, -0.00000392, -0.00000261, 0.00000426, -0.00000226, 0.00000062, -0.00000130, -0.00000767, 0.00000323, 0.00000548, 0.00000059, -0.00000230, -0.00000509, -0.00001849, -0.00000315, -0.00001257, -0.00001020, -0.00006738, 0.00002265, 0.00008962, 0.00004814, 0.00010503, -0.00002879, -0.00005349, -0.00000660, -0.00001738, -0.00000034, -0.00000311, -0.00000166, -0.00000140, -0.00000257, -0.00000314, -0.00000229, -0.00000338, 0.00000128, -0.00000199, 0.00000291, -0.00000567, 0.00000217, 0.00001173, -0.00000464, 0.00000302, -0.00000541, 0.00000187, -0.00000502, -0.00001167, 0.00000540, 0.00000126, -0.00000592, -0.00000390, 0.00000297, -0.00001361, 0.00000531, -0.00000229, 0.00000445, 0.00000044, -0.00000470, 0.00001124, -0.00000347, -0.00000089, -0.00000246, 0.00000991, -0.00000588, 0.00000664, -0.00000089, 0.00000067, 0.00000067, -0.00001077, 0.00000085, -0.00000507, -0.00000143, 0.00000224, -0.00000084, -0.00001579, 0.00000627, -0.00000746, 0.00000195, -0.00000135, 0.00000233, 0.00001127, -0.00000681, 0.00000092, 0.00000087, 0.00000054, -0.00000173, -0.00000252, -0.00000278, -0.00000156, -0.00000409, 0.00000206, -0.00000435, -0.00000050, -0.00000792, -0.00001131, -0.00002092, -0.00005097, 0.00003911, 0.00007339, 0.00003398, 0.00007319, -0.00000096, -0.00004788, -0.00001183, -0.00000664, -0.00000713, -0.00001587, -0.00000303, -0.00000027, 0.00000089, 0.00000734, -0.00000254, -0.00000661, 0.00000338, -0.00000281, -0.00000096, 0.00000501, -0.00000483, -0.00000436, -0.00000186, -0.00000398, 0.00000736, },
		{ 0.00001152, 0.00000206, -0.00000506, -0.00000539, 0.00000795, 0.00000894, -0.00000431, -0.00000152, -0.00000435, -0.00000337, -0.00000050, 0.00000371, -0.00000358, 0.00000288, -0.00000671, -0.00000308, 0.00000456, 0.00000524, -0.00000284, -0.00000039, -0.00001304, -0.00001283, -0.00000571, -0.00001019, -0.00002744, -0.00006481, 0.00007870, 0.00006483, 0.00008077, 0.00007116, -0.00007339, -0.00002200, -0.00001431, -0.00000964, -0.00000073, -0.00000233, -0.00000228, -0.00000085, -0.00000408, -0.00000166, -0.00000412, -0.00000053, -0.00000024, -0.00000044, 0.00000112, -0.00000697, 0.00001088, 0.00000401, -0.00000300, 0.00000145, -0.00000526, 0.00000369, -0.00001288, -0.00000303, 0.00000531, -0.00000113, -0.00000879, 0.00000263, -0.00000439, -0.00000923, 0.00000700, -0.00000437, 0.00000875, -0.00000787, 0.00000478, 0.00000603, -0.00000323, -0.00000244, 0.00000245, 0.00000605, -0.00000490, 0.00000902, -0.00000567, 0.00000564, -0.00000716, -0.00000558, -0.00000154, -0.00000344, -0.00000086, 0.00000434, -0.00001054, -0.00000745, 0.00000482, -0.00000923, 0.00000668, -0.00000628, 0.00001210, 0.00000145, -0.00000461, 0.00000179, 0.00000118, -0.00000079, -0.00000161, -0.00000361, -0.00000099, -0.00000375, -0.00000116, 0.00000051, -0.00000428, -0.00000101, -0.00001203, -0.00000944, -0.00003877, -0.00002512, 0.00007155, 0.00004758, 0.00004676, 0.00005859, -0.00003185, -0.00003472, -0.00000714, -0.00000574, -0.00001095, -0.00001313, 0.00000096, -0.00000170, 0.00000466, 0.00000493, -0.00000687, -0.00000207, 0.00000299, -0.00000518, 0.00000352, 0.00000179, -0.00000662, -0.00000225, -0.00000327, -0.00000093, },
		{ 0.00000513, 0.00001007, -0.00000289, -0.00000528, -0.00000222, 0.00001313, 0.00000128, -0.00000311, -0.00000288, -0.00000360, -0.00000354, 0.00000260, 0.00000056, -0.00000136, 0.00000061, -0.00000835, 0.00000148, 0.00000492, 0.00000345, -0.00000439, -0.00000097, -0.00001939, -0.00000476, -0.00001255, -0.00000499, -0.00005734, -0.00002458, 0.00010588, 0.00004898, 0.00011339, -0.00000256, -0.00006962, -0.00000578, -0.00001965, -0.00000177, -0.00000274, -0.00000137, -0.00000270, -0.00000105, -0.00000453, -0.00000123, -0.00000478, 0.00000208, -0.00000267, 0.00000257, -0.00000421, -0.00000105, 0.00001306, -0.00000304, 0.00000101, -0.00000307, -0.00000028, -0.00000212, -0.00001316, 0.00000403, 0.00000305, -0.00000473, -0.00000722, 0.00000587, -0.00001223, 0.00000010, 0.00000199, 0.00000045, 0.00000485, -0.00000887, 0.00001132, -0.00000052, -0.00000134, -0.00000381, 0.00000758, -0.00000045, 0.00000127, 0.00000478, -0.00000435, 0.00000507, -0.00001115, -0.00000142, -0.00000390, -0.00000116, 0.00000020, 0.00000212, -0.00001667, 0.00000405, -0.00000464, -0.00000144, 0.00000208, -0.00000174, 0.00001364, -0.00000631, 0.00000004, 0.00000104, 0.00000145, -0.00000212, -0.00000164, -0.00000375, -0.00000019, -0.00000527, 0.00000229, -0.00000309, -0.00000168, -0.00000575, -0.00001076, -0.00001694, -0.00004605, 0.00001695, 0.00007428, 0.00003083, 0.00006201, 0.00002674, -0.00004405, -0.00002144, -0.00000501, -0.00000662, -0.00001380, -0.00000784, 0.00000202, -0.00000165, 0.00000803, -0.00000095, -0.00000680, 0.00000209, -0.00000033, -0.00000410, 0.00000640, -0.00000334, -0.00000542, -0.00000153, -0.00000434, },
		{ -0.00000222, 0.00001023, 0.00000444, -0.00000461, -0.00000582, 0.00000491, 0.00001140, -0.00000365, -0.00000168, -0.00000411, -0.00000287, -0.00000276, 0.00000430, -0.00000223, 0.00000145, -0.00000438, -0.00000538, 0.00000362, 0.00000554, -0.00000037, -0.00000288, -0.00000663, -0.00001872, -0.00000193, -0.00001546, -0.00000921, -0.00008057, 0.00004281, 0.00009291, 0.00006839, 0.00010207, -0.00006952, -0.00003414, -0.00001203, -0.00001287, -0.00000136, -0.00000124, -0.00000338, -0.00000058, -0.00000385, -0.00000216, -0.00000376, -0.00000167, 0.00000086, -0.00000192, 0.00000226, -0.00000731, 0.00000795, 0.00000769, -0.00000453, 0.00000200, -0.00000497, 0.00000326, -0.00001064, -0.00000595, 0.00000580, 0.00000082, -0.00000914, -0.00000036, 0.00000124, -0.00001296, 0.00000617, -0.00000274, 0.00000656, -0.00000390, -0.00000170, 0.00001052, -0.00000364, -0.00000103, -0.00000227, 0.00000908, -0.00000445, 0.00000720, -0.00000196, 0.00000085, -0.00000070, -0.00000969, -0.00000095, -0.00000345, -0.00000050, 0.00000246, -0.00000601, -0.00001249, 0.00000755, -0.00001084, 0.00000706, -0.00000585, 0.00000944, 0.00000477, -0.00000579, 0.00000148, 0.00000183, -0.00000048, -0.00000140, -0.00000350, -0.00000128, -0.00000273, -0.00000275, 0.00000187, -0.00000402, -0.00000159, -0.00000930, -0.00000969, -0.00003074, -0.00003210, 0.00005374, 0.00005523, 0.00003324, 0.00006276, -0.00000678, -0.00004138, -0.00001166, -0.00000453, -0.00000893, -0.00001419, -0.00000187, 0.00000022, 0.00000160, 0.00000744, -0.00000647, -0.00000292, 0.00000322, -0.00000396, 0.00000015, 0.00000529, -0.00000731, -0.00000242, -0.00000308, },
		{ -0.00000482, 0.00000324, 0.00001067, -0.00000114, -0.00000497, -0.00000424, 0.00001192, 0.00000417, -0.00000359, -0.00000260, -0.00000331, -0.00000370, 0.00000021, 0.00000284, -0.00000201, 0.00000105, -0.00000750, -0.00000094, 0.00000429, 0.00000535, -0.00000427, -0.00000053, -0.00001543, -0.00001064, -0.00000791, -0.00000922, -0.00003393, -0.00006912, 0.00010154, 0.00006462, 0.00011008, 0.00003342, -0.00008565, -0.00000675, -0.00002164, -0.00000280, -0.00000391, 0.00000064, -0.00000522, 0.00000105, -0.00000617, 0.00000001, -0.00000623, 0.00000261, -0.00000279, 0.00000174, -0.00000263, -0.00000360, 0.00001282, -0.00000002, -0.00000166, -0.00000079, -0.00000198, 0.00000010, -0.00001386, 0.00000229, 0.00000421, -0.00000250, -0.00001025, 0.00000639, -0.00000805, -0.00000577, 0.00000485, -0.00000151, 0.00000676, -0.00000931, 0.00000741, 0.00000454, -0.00000291, -0.00000268, 0.00000226, 0.00000599, -0.00000327, 0.00000839, -0.00000593, 0.00000490, -0.00000736, -0.00000558, -0.00000252, -0.00000118, -0.00000061, 0.00000292, -0.00001483, -0.00000056, -0.00000028, -0.00000577, 0.00000594, -0.00000560, 0.00001529, -0.00000532, -0.00000080, 0.00000070, 0.00000281, -0.00000274, -0.00000047, -0.00000497, 0.00000113, -0.00000597, 0.00000188, -0.00000183, -0.00000191, -0.00000540, -0.00000850, -0.00001550, -0.00003902, -0.00000070, 0.00006905, 0.00003356, 0.00004764, 0.00004436, -0.00002920, -0.00003197, -0.00000521, -0.00000602, -0.00001112, -0.00001182, 0.00000253, -0.00000230, 0.00000694, 0.00000148, -0.00000728, 0.00000109, 0.00000152, -0.00000557, 0.00000537, -0.00000012, -0.00000722, -0.00000079, },
		{ -0.00000229, -0.00000360, 0.00000886, 0.00000633, -0.00000382, -0.00000581, 0.00000177, 0.00001295, -0.00000222, -0.00000199, -0.00000400, -0.00000218, -0.00000435, 0.00000342, -0.00000003, 0.00000000, -0.00000253, -0.00000657, 0.00000205, 0.00000510, 0.00000280, -0.00000543, -0.00000158, -0.00002118, -0.00000193, -0.00001689, -0.00000118, -0.00007239, -0.00001138, 0.00011766, 0.00006233, 0.00012487, -0.00005312, -0.00005200, -0.00000815, -0.00001713, -0.00000131, -0.00000167, -0.00000236, -0.00000237, -0.00000198, -0.00000392, -0.00000226, -0.00000383, 0.00000260, -0.00000346, 0.00000300, -0.00000698, 0.00000465, 0.00001053, -0.00000450, 0.00000108, -0.00000366, 0.00000222, -0.00000814, -0.00000862, 0.00000579, 0.00000242, -0.00000787, -0.00000472, 0.00000594, -0.00001350, 0.00000246, 0.00000022, 0.00000376, 0.00000027, -0.00000667, 0.00001170, -0.00000138, -0.00000113, -0.00000413, 0.00000730, 0.00000054, 0.00000190, 0.00000419, -0.00000472, 0.00000426, -0.00001078, -0.00000253, -0.00000303, 0.00000019, 0.00000047, -0.00000213, -0.00001572, 0.00000784, -0.00001010, 0.00000524, -0.00000335, 0.00000520, 0.00000882, -0.00000720, 0.00000145, 0.00000173, 0.00000073, -0.00000197, -0.00000248, -0.00000261, -0.00000081, -0.00000472, 0.00000291, -0.00000344, -0.00000184, -0.00000803, -0.00000817, -0.00002623, -0.00003273, 0.00003435, 0.00006057, 0.00002565, 0.00005766, 0.00001589, -0.00003785, -0.00002096, -0.00000236, -0.00000836, -0.00001256, -0.00000665, 0.00000296, -0.00000161, 0.00000946, -0.00000590, -0.00000340, 0.00000266, -0.00000166, -0.00000350, 0.00000768, -0.00000654, -0.00000352, },
		{ -0.00000023, -0.00000501, 0.00000136, 0.00001081, 0.00000061, -0.00000443, -0.00000564, 0.00000981, 0.00000698, -0.00000354, -0.00000258, -0.00000313, -0.00000324, -0.00000226, 0.00000389, -0.00000114, 0.00000030, -0.00000572, -0.00000337, 0.00000351, 0.00000583, -0.00000169, -0.00000325, -0.00000837, -0.00001817, -0.00000171, -0.00001738, -0.00000959, -0.00009464, 0.00006840, 0.00009367, 0.00009870, 0.00007442, -0.00009697, -0.00001178, -0.00002242, -0.00000407, -0.00000580, 0.00000302, -0.00000735, 0.00000233, -0.00000699, 0.00000056, -0.00000692, 0.00000220, -0.00000200, 0.00000042, -0.00000113, -0.00000526, 0.00001105, 0.00000389, -0.00000422, 0.00000083, -0.00000296, 0.00000170, -0.00001399, 0.00000039, 0.00000474, 0.00000015, -0.00001200, 0.00000434, -0.00000239, -0.00001029, 0.00000498, -0.00000106, 0.00000633, -0.00000687, 0.00000147, 0.00000932, -0.00000375, -0.00000102, -0.00000293, 0.00000971, -0.00000363, 0.00000730, -0.00000247, 0.00000017, -0.00000094, -0.00000982, -0.00000183, -0.00000148, -0.00000011, 0.00000188, -0.00001108, -0.00000622, 0.00000414, -0.00000970, 0.00000907, -0.00000821, 0.00001540, -0.00000321, -0.00000207, 0.00000033, 0.00000397, -0.00000299, 0.00000043, -0.00000576, 0.00000172, -0.00000573, 0.00000079, -0.00000078, -0.00000123, -0.00000606, -0.00000600, -0.00001443, -0.00003284, -0.00001206, 0.00005820, 0.00003954, 0.00003438, 0.00005163, -0.00000950, -0.00003707, -0.00001027, -0.00000371, -0.00000974, -0.00001301, -0.00000002, -0.00000045, 0.00000387, 0.00000498, -0.00000849, 0.00000094, 0.00000208, -0.00000484, 0.00000209, 0.00000411, -0.00000900, },
		{ -0.00000580, -0.00000080, -0.00000513, 0.00000748, 0.00000779, -0.00000288, -0.00000535, -0.00000125, 0.00001355, -0.00000037, -0.00000205, -0.00000425, -0.00000127, -0.00000532, 0.00000171, 0.00000185, -0.00000029, -0.00000206, -0.00000608, -0.00000038, 0.00000462, 0.00000501, -0.00000565, -0.00000057, -0.00001793, -0.00000761, -0.00001184, -0.00000564, -0.00004547, -0.00006774, 0.00012294, 0.00007135, 0.00013446, -0.00002204, -0.00007514, -0.00000284, -0.00002306, 0.00000051, -0.00000476, 0.00000136, -0.00000625, 0.00000134, -0.00000678, 0.00000024, -0.00000678, 0.00000464, -0.00000480, 0.00000341, -0.00000639, 0.00000177, 0.00001170, -0.00000245, -0.00000133, -0.00000152, 0.00000069, -0.00000523, -0.00001141, 0.00000586, 0.00000304, -0.00000492, -0.00000952, 0.00000867, -0.00001113, -0.00000235, 0.00000235, 0.00000222, 0.00000278, -0.00000847, 0.00000893, 0.00000326, -0.00000265, -0.00000311, 0.00000184, 0.00000746, -0.00000340, 0.00000882, -0.00000704, 0.00000463, -0.00000750, -0.00000634, -0.00000207, 0.00000041, -0.00000036, -0.00000019, -0.00001614, 0.00000522, -0.00000700, 0.00000137, 0.00000096, -0.00000026, 0.00001319, -0.00000854, 0.00000163, 0.00000084, 0.00000279, -0.00000329, -0.00000055, -0.00000486, 0.00000174, -0.00000680, 0.00000381, -0.00000332, -0.00000067, -0.00000902, -0.00000458, -0.00002474, -0.00002927, 0.00001626, 0.00006181, 0.00002344, 0.00004820, 0.00003136, -0.00002530, -0.00003071, -0.00000235, -0.00000736, -0.00001032, -0.00001086, 0.00000420, -0.00000313, 0.00000967, -0.00000407, -0.00000461, 0.00000246, 0.00000025, -0.00000571, 0.00000760, -0.00000365, },
		{ -0.00000949, -0.00000074, -0.00000427, -0.00000091, 0.00001080, 0.00000220, -0.00000376, -0.00000635, 0.00000702, 0.00000946, -0.00000308, -0.00000242, -0.00000343, -0.00000206, -0.00000462, 0.00000384, 0.00000043, -0.00000041, -0.00000440, -0.00000459, 0.00000192, 0.00000556, 0.00000182, -0.00000641, -0.00000215, -0.00002266, 0.00000080, -0.00002164, 0.00000359, -0.00009169, 0.00000982, 0.00012489, 0.00008876, 0.00011227, -0.00009711, -0.00002419, -0.00001988, -0.00000761, -0.00000614, 0.00000320, -0.00000683, 0.00000110, -0.00000573, -0.00000065, -0.00000590, 0.00000012, 0.00000006, -0.00000149, 0.00000040, -0.00000631, 0.00000849, 0.00000751, -0.00000549, 0.00000096, -0.00000277, 0.00000242, -0.00001318, -0.00000215, 0.00000542, 0.00000204, -0.00001146, -0.00000013, 0.00000348, -0.00001246, 0.00000271, 0.00000046, 0.00000518, -0.00000368, -0.00000368, 0.00001131, -0.00000213, -0.00000072, -0.00000537, 0.00000838, 0.00000123, 0.00000180, 0.00000439, -0.00000613, 0.00000463, -0.00001147, -0.00000307, -0.00000145, 0.00000083, 0.00000038, -0.00000724, -0.00001095, 0.00000687, -0.00001181, 0.00001021, -0.00000847, 0.00001312, 0.00000062, -0.00000424, 0.00000051, 0.00000422, -0.00000216, 0.00000032, -0.00000531, 0.00000081, -0.00000402, -0.00000124, 0.00000046, -0.00000072, -0.00000589, -0.00000545, -0.00001136, -0.00002983, -0.00001647, 0.00004295, 0.00004665, 0.00002388, 0.00005171, 0.00000814, -0.00003298, -0.00001996, -0.00000044, -0.00000964, -0.00001152, -0.00000484, 0.00000275, -0.00000003, 0.00000866, -0.00000978, 0.00000106, 0.00000184, -0.00000272, -0.00000203, 0.00000772, },
		{ 0.00000090, -0.00000863, 0.00000079, -0.00000627, 0.00000568, 0.00000913, -0.00000204, -0.00000448, -0.00000394, 0.00001326, 0.00000162, -0.00000171, -0.00000473, -0.00000049, -0.00000548, -0.00000047, 0.00000286, 0.00000074, -0.00000273, -0.00000458, -0.00000293, 0.00000399, 0.00000568, -0.00000301, -0.00000348, -0.00001033, -0.00001662, -0.00000321, -0.00001690, -0.00001387, -0.00010620, 0.00009620, 0.00009747, 0.00013162, 0.00002114, -0.00009966, 0.00000143, -0.00002938, 0.00000322, -0.00000951, 0.00000628, -0.00001053, 0.00000460, -0.00000954, 0.00000271, -0.00000949, 0.00000603, -0.00000538, 0.00000335, -0.00000571, -0.00000033, 0.00001100, 0.00000128, -0.00000452, 0.00000073, -0.00000092, -0.00000207, -0.00001418, 0.00000587, 0.00000298, -0.00000124, -0.00001329, 0.00000860, -0.00000658, -0.00000662, 0.00000248, 0.00000235, 0.00000354, -0.00000748, 0.00000374, 0.00000819, -0.00000391, -0.00000105, -0.00000418, 0.00001221, -0.00000470, 0.00000848, -0.00000376, -0.00000023, -0.00000085, -0.00001086, -0.00000135, 0.00000008, 0.00000020, -0.00000017, -0.00001420, 0.00000069, -0.00000268, -0.00000346, 0.00000597, -0.00000570, 0.00001672, -0.00000890, 0.00000137, -0.00000029, 0.00000505, -0.00000472, 0.00000166, -0.00000723, 0.00000403, -0.00000828, 0.00000428, -0.00000369, 0.00000170, -0.00001128, -0.00000059, -0.00002389, -0.00002523, 0.00000262, 0.00005748, 0.00002598, 0.00003717, 0.00003965, -0.00000934, -0.00003495, -0.00000751, -0.00000431, -0.00000938, -0.00001242, 0.00000235, -0.00000194, 0.00000740, -0.00000038, -0.00000721, 0.00000330, 0.00000089, -0.00000561, 0.00000484, },
		{ 0.00000904, -0.00000765, -0.00000286, -0.00000239, -0.00000355, 0.00001055, 0.00000379, -0.00000325, -0.00000620, 0.00000366, 0.00001163, -0.00000258, -0.00000165, -0.00000458, -0.00000011, -0.00000683, 0.00000317, 0.00000161, 0.00000012, -0.00000460, -0.00000390, -0.00000081, 0.00000536, 0.00000434, -0.00000714, -0.00000028, -0.00002071, -0.00000370, -0.00001756, 0.00000091, -0.00006347, -0.00005781, 0.00013989, 0.00009157, 0.00013893, -0.00008057, -0.00004627, -0.00001247, -0.00001522, -0.00000288, -0.00000074, -0.00000231, -0.00000337, -0.00000192, -0.00000404, -0.00000278, -0.00000381, 0.00000331, -0.00000396, 0.00000216, -0.00000728, 0.00000618, 0.00000954, -0.00000444, -0.00000098, -0.00000114, 0.00000188, -0.00001088, -0.00000598, 0.00000698, 0.00000232, -0.00000836, -0.00000617, 0.00000831, -0.00001214, -0.00000061, 0.00000127, 0.00000490, -0.00000148, -0.00000620, 0.00000947, 0.00000211, -0.00000216, -0.00000447, 0.00000254, 0.00000909, -0.00000474, 0.00001043, -0.00000952, 0.00000567, -0.00000856, -0.00000683, -0.00000067, 0.00000121, -0.00000014, -0.00000484, -0.00001336, 0.00000694, -0.00001148, 0.00000880, -0.00000594, 0.00000827, 0.00000606, -0.00000717, 0.00000135, 0.00000325, 0.00000005, -0.00000116, -0.00000318, -0.00000189, -0.00000076, -0.00000429, 0.00000239, -0.00000154, -0.00000337, -0.00000809, -0.00000543, -0.00003032, -0.00001528, 0.00002585, 0.00005253, 0.00001691, 0.00004771, 0.00002015, -0.00002112, -0.00003001, 0.00000058, -0.00000892, -0.00000929, -0.00000954, 0.00000497, -0.00000279, 0.00001090, -0.00000972, 0.00000010, 0.00000206, -0.00000072, -0.00000509, },
		{ 0.00000058, 0.00000556, -0.00001059, 0.00000137, -0.00000622, 0.00000305, 0.00001056, -0.00000131, -0.00000346, -0.00000595, 0.00001197, 0.00000378, -0.00000135, -0.00000478, -0.00000059, -0.00000434, -0.00000319, 0.00000324, 0.00000219, -0.00000316, -0.00000366, -0.00000412, 0.00000237, 0.00000555, 0.00000093, -0.00000746, -0.00000271, -0.00002356, 0.00000261, -0.00002534, 0.00000702, -0.00011267, 0.00003703, 0.00013132, 0.00012476, 0.00006857, -0.00011809, -0.00000025, -0.00003263, 0.00000374, -0.00001275, 0.00000912, -0.00001225, 0.00000548, -0.00001032, 0.00000346, -0.00001039, 0.00000549, -0.00000437, 0.00000233, -0.00000461, -0.00000199, 0.00000919, 0.00000536, -0.00000698, 0.00000187, -0.00000172, 0.00000052, -0.00001599, 0.00000477, 0.00000345, 0.00000152, -0.00001442, 0.00000527, -0.00000071, -0.00000950, 0.00000091, 0.00000301, 0.00000382, -0.00000549, -0.00000123, 0.00001081, -0.00000297, -0.00000014, -0.00000760, 0.00001169, -0.00000007, 0.00000264, 0.00000407, -0.00000765, 0.00000570, -0.00001327, -0.00000236, -0.00000003, 0.00000124, -0.00000079, -0.00001157, -0.00000389, 0.00000108, -0.00000769, 0.00001009, -0.00000951, 0.00001794, -0.00000717, -0.00000009, -0.00000086, 0.00000656, -0.00000526, 0.00000302, -0.00000843, 0.00000476, -0.00000811, 0.00000356, -0.00000365, 0.00000364, -0.00001236, 0.00000098, -0.00002063, -0.00002402, -0.00000420, 0.00004699, 0.00003254, 0.00002564, 0.00004342, 0.00000396, -0.00003020, -0.00001769, 0.00000009, -0.00000983, -0.00001112, -0.00000221, 0.00000123, 0.00000345, 0.00000451, -0.00001057, 0.00000462, 0.00000064, -0.00000377, },
		{ -0.00000588, 0.00000741, -0.00000356, -0.00000616, 0.00000004, -0.00000590, 0.00000954, 0.00000577, -0.00000319, -0.00000518, -0.00000004, 0.00001344, -0.00000224, -0.00000021, -0.00000633, 0.00000208, -0.00000841, 0.00000198, 0.00000196, 0.00000206, -0.00000607, -0.00000196, -0.00000387, 0.00000513, 0.00000509, -0.00000447, -0.00000325, -0.00001292, -0.00001362, -0.00000723, -0.00001260, -0.00002484, -0.00011106, 0.00012173, 0.00011281, 0.00015175, -0.00004655, -0.00007638, -0.00000154, -0.00002629, 0.00000359, -0.00000824, 0.00000506, -0.00000969, 0.00000321, -0.00000864, 0.00000156, -0.00000863, 0.00000683, -0.00000636, 0.00000396, -0.00000838, 0.00000469, 0.00000941, -0.00000097, -0.00000461, 0.00000149, 0.00000030, -0.00000713, -0.00001093, 0.00000917, 0.00000139, -0.00000375, -0.00001201, 0.00001074, -0.00000950, -0.00000377, 0.00000048, 0.00000577, -0.00000050, -0.00000613, 0.00000491, 0.00000719, -0.00000387, -0.00000186, -0.00000475, 0.00001558, -0.00000766, 0.00001136, -0.00000675, 0.00000073, -0.00000151, -0.00001195, 0.00000027, 0.00000082, 0.00000070, -0.00000411, -0.00001353, 0.00000485, -0.00000932, 0.00000538, -0.00000124, 0.00000179, 0.00001202, -0.00000988, 0.00000224, 0.00000147, 0.00000321, -0.00000361, 0.00000017, -0.00000572, 0.00000322, -0.00000766, 0.00000475, -0.00000377, 0.00000129, -0.00001296, 0.00000168, -0.00003201, -0.00001205, 0.00001081, 0.00005424, 0.00001472, 0.00004071, 0.00002736, -0.00000706, -0.00003445, -0.00000387, -0.00000580, -0.00000835, -0.00001180, 0.00000428, -0.00000292, 0.00001051, -0.00000718, -0.00000294, 0.00000363, 0.00000004, },
		{ -0.00000170, -0.00000320, 0.00000840, -0.00001034, 0.00000011, -0.00000459, -0.00000038, 0.00001179, -0.00000031, -0.00000281, -0.00000682, 0.00000947, 0.00000640, -0.00000160, -0.00000358, -0.00000226, -0.00000158, -0.00000655, 0.00000353, 0.00000287, -0.00000195, -0.00000467, -0.00000294, -0.00000080, 0.00000566, 0.00000387, -0.00000900, 0.00000051, -0.00002368, 0.00000055, -0.00002409, 0.00000904, -0.00008694, -0.00003933, 0.00015351, 0.00012653, 0.00011046, -0.00012216, -0.00001300, -0.00002909, -0.00000140, -0.00001111, 0.00000674, -0.00000903, 0.00000229, -0.00000777, 0.00000122, -0.00000832, 0.00000217, -0.00000128, -0.00000002, -0.00000256, -0.00000399, 0.00000757, 0.00000807, -0.00000710, 0.00000067, -0.00000080, 0.00000153, -0.00001557, 0.00000133, 0.00000566, 0.00000204, -0.00001211, -0.00000089, 0.00000535, -0.00001069, -0.00000126, 0.00000265, 0.00000501, -0.00000415, -0.00000398, 0.00000972, 0.00000080, -0.00000125, -0.00000719, 0.00000570, 0.00000890, -0.00000558, 0.00001211, -0.00001274, 0.00000791, -0.00001098, -0.00000617, 0.00000083, 0.00000157, -0.00000050, -0.00000987, -0.00000697, 0.00000309, -0.00001029, 0.00001220, -0.00001051, 0.00001597, -0.00000288, -0.00000303, -0.00000045, 0.00000662, -0.00000411, 0.00000261, -0.00000742, 0.00000301, -0.00000561, 0.00000101, -0.00000212, 0.00000336, -0.00001005, -0.00000187, -0.00001330, -0.00002701, -0.00000421, 0.00003180, 0.00004115, 0.00001484, 0.00004467, 0.00001194, -0.00001801, -0.00002870, 0.00000260, -0.00000979, -0.00000875, -0.00000729, 0.00000420, -0.00000028, 0.00000886, -0.00001300, 0.00000484, 0.00000081, },
		{ -0.00000029, -0.00000461, 0.00000365, 0.00000144, -0.00000925, 0.00000177, -0.00000697, 0.00000720, 0.00000832, -0.00000349, -0.00000366, -0.00000354, 0.00001448, -0.00000169, 0.00000122, -0.00000761, 0.00000328, -0.00000837, -0.00000022, 0.00000200, 0.00000429, -0.00000724, -0.00000065, -0.00000551, 0.00000384, 0.00000487, 0.00000005, -0.00000820, -0.00000376, -0.00002311, 0.00000217, -0.00002581, 0.00000551, -0.00013064, 0.00006535, 0.00014590, 0.00015681, -0.00000081, -0.00010783, 0.00000774, -0.00003684, 0.00000971, -0.00001570, 0.00001157, -0.00001455, 0.00000699, -0.00001223, 0.00000505, -0.00001236, 0.00000891, -0.00000751, 0.00000504, -0.00000909, 0.00000366, 0.00000780, 0.00000363, -0.00000836, 0.00000372, -0.00000123, -0.00000302, -0.00001561, 0.00001043, 0.00000091, 0.00000024, -0.00001542, 0.00000960, -0.00000482, -0.00000623, -0.00000148, 0.00000663, 0.00000040, -0.00000500, 0.00000010, 0.00001047, -0.00000375, 0.00000000, -0.00000984, 0.00001679, -0.00000398, 0.00000562, 0.00000191, -0.00000799, 0.00000641, -0.00001554, -0.00000028, 0.00000057, 0.00000196, -0.00000385, -0.00001291, 0.00000219, -0.00000676, 0.00000120, 0.00000418, -0.00000459, 0.00001671, -0.00001096, 0.00000218, -0.00000023, 0.00000627, -0.00000589, 0.00000344, -0.00000912, 0.00000630, -0.00000993, 0.00000642, -0.00000611, 0.00000607, -0.00001707, 0.00000654, -0.00003121, -0.00001108, 0.00000170, 0.00004926, 0.00001841, 0.00003048, 0.00003273, 0.00000330, -0.00002961, -0.00001407, -0.00000069, -0.00000905, -0.00001105, 0.00000054, -0.00000059, 0.00000765, -0.00000233, -0.00000770, 0.00000614, },
		{ 0.00000256, -0.00000073, -0.00000496, 0.00000831, -0.00000748, -0.00000287, -0.00000189, -0.00000383, 0.00001209, 0.00000156, -0.00000303, -0.00000629, 0.00000574, 0.00000961, -0.00000279, -0.00000095, -0.00000536, 0.00000228, -0.00000997, 0.00000370, 0.00000236, 0.00000116, -0.00000751, 0.00000005, -0.00000480, 0.00000609, 0.00000468, -0.00000653, -0.00000209, -0.00001642, -0.00000912, -0.00001367, -0.00000447, -0.00004347, -0.00010699, 0.00014324, 0.00014566, 0.00014116, -0.00010741, -0.00003842, -0.00001788, -0.00001319, -0.00000369, -0.00000134, -0.00000111, -0.00000438, -0.00000236, -0.00000371, -0.00000356, -0.00000346, 0.00000332, -0.00000341, 0.00000048, -0.00000676, 0.00000706, 0.00000834, -0.00000419, -0.00000305, 0.00000193, 0.00000064, -0.00001247, -0.00000464, 0.00000957, 0.00000051, -0.00000722, -0.00000810, 0.00000997, -0.00000999, -0.00000312, 0.00000059, 0.00000722, -0.00000372, -0.00000426, 0.00000559, 0.00000599, -0.00000327, -0.00000421, -0.00000299, 0.00001777, -0.00001101, 0.00001521, -0.00001127, 0.00000340, -0.00000367, -0.00001209, 0.00000233, 0.00000088, 0.00000116, -0.00000945, -0.00000841, 0.00000348, -0.00001137, 0.00001213, -0.00000858, 0.00001108, 0.00000329, -0.00000671, 0.00000062, 0.00000526, -0.00000127, 0.00000035, -0.00000416, -0.00000101, -0.00000116, -0.00000310, 0.00000109, 0.00000030, -0.00000405, -0.00000862, -0.00000316, -0.00003240, -0.00000045, 0.00001580, 0.00004813, 0.00000735, 0.00004294, 0.00001634, -0.00000421, -0.00003432, -0.00000040, -0.00000720, -0.00000748, -0.00001038, 0.00000476, -0.00000193, 0.00001096, -0.00001282, 0.00000254, },
		{ 0.00000793, -0.00000041, -0.00000267, -0.00000031, 0.00000533, -0.00001053, 0.00000170, -0.00000616, 0.00000352, 0.00001104, -0.00000357, -0.00000244, -0.00000600, 0.00001407, -0.00000016, 0.00000151, -0.00000712, 0.00000224, -0.00000584, -0.00000394, 0.00000258, 0.00000529, -0.00000627, -0.00000181, -0.00000422, 0.00000051, 0.00000505, 0.00000371, -0.00001100, 0.00000134, -0.00002603, 0.00000368, -0.00002910, 0.00001524, -0.00011176, -0.00001647, 0.00017159, 0.00016670, 0.00004690, -0.00013115, 0.00000828, -0.00004133, 0.00001045, -0.00001839, 0.00001295, -0.00001454, 0.00000681, -0.00001258, 0.00000559, -0.00001304, 0.00000796, -0.00000629, 0.00000445, -0.00000846, 0.00000199, 0.00000624, 0.00000729, -0.00001011, 0.00000376, -0.00000121, -0.00000022, -0.00001800, 0.00000881, 0.00000272, 0.00000167, -0.00001486, 0.00000464, 0.00000120, -0.00000788, -0.00000344, 0.00000594, 0.00000249, -0.00000440, -0.00000280, 0.00001016, -0.00000069, -0.00000030, -0.00001073, 0.00001151, 0.00000559, -0.00000423, 0.00001226, -0.00001537, 0.00001046, -0.00001443, -0.00000389, 0.00000157, 0.00000222, -0.00000246, -0.00001308, 0.00000041, -0.00000492, -0.00000270, 0.00000897, -0.00000921, 0.00001862, -0.00000935, 0.00000051, -0.00000110, 0.00000819, -0.00000681, 0.00000520, -0.00001045, 0.00000690, -0.00000973, 0.00000605, -0.00000677, 0.00000831, -0.00001727, 0.00000612, -0.00002522, -0.00001498, 0.00000019, 0.00003734, 0.00002727, 0.00001777, 0.00003797, 0.00000765, -0.00001693, -0.00002604, 0.00000324, -0.00000964, -0.00000876, -0.00000440, 0.00000237, 0.00000395, 0.00000325, -0.00001247, },
		{ -0.00000244, 0.00000579, -0.00000103, -0.00000450, 0.00000569, -0.00000304, -0.00000633, 0.00000061, -0.00000613, 0.00001065, 0.00000462, -0.00000410, -0.00000466, 0.00000133, 0.00001289, -0.00000438, 0.00000220, -0.00000856, 0.00000564, -0.00001200, 0.00000286, 0.00000143, 0.00000487, -0.00001036, 0.00000268, -0.00000757, 0.00000558, 0.00000413, -0.00000142, -0.00000790, -0.00000604, -0.00002063, -0.00000140, -0.00002158, -0.00000378, -0.00014146, 0.00009046, 0.00017856, 0.00016414, -0.00007645, -0.00007219, -0.00000295, -0.00002833, 0.00000638, -0.00001182, 0.00000804, -0.00001141, 0.00000340, -0.00000925, 0.00000191, -0.00000945, 0.00000771, -0.00000660, 0.00000372, -0.00000979, 0.00000741, 0.00000654, 0.00000081, -0.00000796, 0.00000519, -0.00000118, -0.00000771, -0.00001164, 0.00001341, -0.00000119, -0.00000214, -0.00001368, 0.00001136, -0.00000710, -0.00000452, -0.00000248, 0.00000918, -0.00000312, -0.00000339, 0.00000071, 0.00001001, -0.00000401, -0.00000126, -0.00001031, 0.00002176, -0.00000949, 0.00001059, -0.00000249, -0.00000634, 0.00000575, -0.00001730, 0.00000271, 0.00000017, 0.00000315, -0.00000908, -0.00000955, 0.00000358, -0.00001192, 0.00001057, -0.00000460, 0.00000463, 0.00000968, -0.00000972, 0.00000140, 0.00000321, 0.00000246, -0.00000288, 0.00000026, -0.00000590, 0.00000368, -0.00000736, 0.00000484, -0.00000424, 0.00000354, -0.00001612, 0.00000600, -0.00003614, 0.00000227, 0.00000393, 0.00004931, 0.00000617, 0.00003636, 0.00002100, 0.00000495, -0.00003037, -0.00000983, -0.00000203, -0.00000803, -0.00001047, 0.00000224, -0.00000104, 0.00001027, -0.00000951, },
		{ -0.00001508, 0.00000715, 0.00000109, -0.00000159, -0.00000270, 0.00000665, -0.00000921, -0.00000040, -0.00000378, -0.00000068, 0.00001291, -0.00000251, -0.00000232, -0.00000676, 0.00001173, 0.00000292, -0.00000009, -0.00000424, -0.00000131, -0.00000102, -0.00000879, 0.00000379, 0.00000443, -0.00000256, -0.00000570, -0.00000024, -0.00000439, 0.00000602, 0.00000484, -0.00000925, -0.00000018, -0.00002029, -0.00000430, -0.00002064, 0.00000486, -0.00006711, -0.00009645, 0.00016595, 0.00019220, 0.00008929, -0.00013916, -0.00000459, -0.00003633, 0.00000280, -0.00001373, 0.00000726, -0.00000863, 0.00000213, -0.00000912, 0.00000250, -0.00001003, 0.00000364, -0.00000256, 0.00000192, -0.00000592, -0.00000117, 0.00000612, 0.00000836, -0.00000847, 0.00000069, 0.00000114, 0.00000019, -0.00001685, 0.00000339, 0.00000736, 0.00000018, -0.00001067, -0.00000272, 0.00000699, -0.00000837, -0.00000462, 0.00000311, 0.00000587, -0.00000469, -0.00000322, 0.00000648, 0.00000440, -0.00000218, -0.00000808, 0.00000201, 0.00001682, -0.00001270, 0.00001839, -0.00001618, 0.00000724, -0.00000742, -0.00001051, 0.00000383, 0.00000096, 0.00000061, -0.00001438, -0.00000040, -0.00000368, -0.00000624, 0.00001253, -0.00001130, 0.00001731, -0.00000515, -0.00000247, -0.00000109, 0.00000855, -0.00000576, 0.00000468, -0.00000894, 0.00000448, -0.00000672, 0.00000319, -0.00000475, 0.00000658, -0.00001228, -0.00000004, -0.00001439, -0.00002295, 0.00000434, 0.00002153, 0.00003783, 0.00000569, 0.00004162, 0.00000844, -0.00000244, -0.00003335, 0.00000202, -0.00000782, -0.00000720, -0.00000801, 0.00000368, 0.00000139, 0.00000753, },
		{ -0.00000420, -0.00000858, 0.00000893, -0.00000141, -0.00000301, 0.00000225, 0.00000104, -0.00000855, 0.00000156, -0.00000639, 0.00000731, 0.00000845, -0.00000530, -0.00000287, -0.00000268, 0.00001515, -0.00000511, 0.00000423, -0.00000995, 0.00000656, -0.00001107, -0.00000009, 0.00000140, 0.00000719, -0.00001085, 0.00000251, -0.00000702, 0.00000270, 0.00000399, 0.00000323, -0.00001224, 0.00000118, -0.00002653, 0.00000395, -0.00002994, 0.00001532, -0.00013251, 0.00000466, 0.00020657, 0.00019118, -0.00003747, -0.00010549, 0.00000820, -0.00004047, 0.00001326, -0.00001918, 0.00001338, -0.00001472, 0.00000632, -0.00001263, 0.00000545, -0.00001325, 0.00000974, -0.00000796, 0.00000578, -0.00001176, 0.00000731, 0.00000438, 0.00000563, -0.00001165, 0.00000687, -0.00000209, -0.00000340, -0.00001707, 0.00001451, -0.00000055, 0.00000051, -0.00001533, 0.00000848, -0.00000214, -0.00000566, -0.00000520, 0.00000920, -0.00000103, -0.00000299, -0.00000257, 0.00001068, -0.00000190, -0.00000034, -0.00001342, 0.00001854, -0.00000054, -0.00000010, 0.00000988, -0.00001622, 0.00001219, -0.00001807, -0.00000014, 0.00000098, 0.00000374, -0.00000695, -0.00001201, 0.00000469, -0.00001274, 0.00000819, 0.00000033, -0.00000173, 0.00001454, -0.00001071, 0.00000105, 0.00000130, 0.00000596, -0.00000571, 0.00000414, -0.00000965, 0.00000689, -0.00000993, 0.00000733, -0.00000791, 0.00000937, -0.00002029, 0.00001011, -0.00003442, 0.00000008, -0.00000045, 0.00004245, 0.00001244, 0.00002441, 0.00002809, 0.00000693, -0.00001777, -0.00002221, 0.00000286, -0.00000891, -0.00000865, -0.00000200, 0.00000107, 0.00000787, },
		{ 0.00000932, -0.00001431, 0.00000306, 0.00000451, -0.00000184, -0.00000294, 0.00000533, -0.00000591, -0.00000351, -0.00000105, -0.00000405, 0.00001287, 0.00000028, -0.00000353, -0.00000583, 0.00000771, 0.00000725, -0.00000318, 0.00000024, -0.00000608, 0.00000440, -0.00001309, 0.00000453, 0.00000258, 0.00000255, -0.00001046, 0.00000413, -0.00000864, 0.00000649, 0.00000406, -0.00000390, -0.00000631, -0.00000952, -0.00001658, -0.00000722, -0.00001379, -0.00002023, -0.00014523, 0.00011407, 0.00023211, 0.00012805, -0.00013151, -0.00002958, -0.00002360, -0.00001153, -0.00000363, -0.00000316, 0.00000046, -0.00000460, -0.00000373, -0.00000275, -0.00000475, -0.00000256, 0.00000233, -0.00000167, -0.00000196, -0.00000558, 0.00000752, 0.00000674, -0.00000382, -0.00000467, 0.00000500, -0.00000128, -0.00001271, -0.00000466, 0.00001318, -0.00000243, -0.00000521, -0.00000964, 0.00001037, -0.00000702, -0.00000511, -0.00000103, 0.00000914, -0.00000482, -0.00000227, 0.00000147, 0.00000903, -0.00000350, -0.00000444, -0.00000748, 0.00002424, -0.00001449, 0.00001616, -0.00000840, -0.00000277, 0.00000321, -0.00001753, 0.00000564, -0.00000072, 0.00000417, -0.00001537, -0.00000172, -0.00000170, -0.00001016, 0.00001503, -0.00001103, 0.00001347, 0.00000046, -0.00000563, -0.00000083, 0.00000764, -0.00000304, 0.00000223, -0.00000513, -0.00000008, -0.00000197, -0.00000124, -0.00000067, 0.00000163, -0.00000372, -0.00000913, -0.00000231, -0.00003106, 0.00000944, 0.00000715, 0.00004485, -0.00000126, 0.00004025, 0.00001055, 0.00000701, -0.00003111, -0.00000602, -0.00000303, -0.00000745, -0.00000887, 0.00000222, 0.00000097, },
		{ 0.00000562, 0.00000104, -0.00001348, 0.00000982, -0.00000029, -0.00000209, -0.00000012, 0.00000305, -0.00000842, 0.00000043, -0.00000470, 0.00000282, 0.00001186, -0.00000546, -0.00000206, -0.00000517, 0.00001535, -0.00000380, 0.00000375, -0.00000823, 0.00000407, -0.00000672, -0.00000525, 0.00000279, 0.00000700, -0.00000778, -0.00000137, -0.00000279, -0.00000256, 0.00000510, 0.00000511, -0.00001175, 0.00000134, -0.00002304, -0.00000135, -0.00002489, 0.00001062, -0.00008999, -0.00008627, 0.00020297, 0.00023443, 0.00000220, -0.00013017, 0.00000894, -0.00004425, 0.00001220, -0.00001931, 0.00001167, -0.00001218, 0.00000487, -0.00001235, 0.00000547, -0.00001338, 0.00000825, -0.00000666, 0.00000572, -0.00001150, 0.00000538, 0.00000367, 0.00000810, -0.00001208, 0.00000532, -0.00000062, -0.00000136, -0.00001876, 0.00001106, 0.00000373, -0.00000035, -0.00001265, 0.00000213, 0.00000367, -0.00000636, -0.00000664, 0.00000655, 0.00000273, -0.00000356, -0.00000334, 0.00000770, 0.00000271, -0.00000147, -0.00001216, 0.00000951, 0.00001217, -0.00001139, 0.00001939, -0.00002008, 0.00001119, -0.00001222, -0.00000691, 0.00000385, 0.00000203, -0.00000234, -0.00001609, 0.00000669, -0.00001339, 0.00000481, 0.00000562, -0.00000691, 0.00001693, -0.00000929, -0.00000050, -0.00000015, 0.00000844, -0.00000708, 0.00000617, -0.00001084, 0.00000720, -0.00000970, 0.00000726, -0.00000879, 0.00001092, -0.00001865, 0.00000742, -0.00002632, -0.00000757, 0.00000235, 0.00002903, 0.00002374, 0.00000982, 0.00003592, 0.00000443, -0.00000250, -0.00003110, 0.00000328, -0.00000772, -0.00000711, -0.00000554, 0.00000227, },
		{ 0.00000194, 0.00000903, -0.00001066, -0.00000324, 0.00000851, -0.00000247, -0.00000193, 0.00000280, -0.00000239, -0.00000595, 0.00000050, -0.00000542, 0.00001045, 0.00000451, -0.00000543, -0.00000409, 0.00000310, 0.00001160, -0.00000625, 0.00000429, -0.00000971, 0.00000791, -0.00001461, 0.00000310, 0.00000150, 0.00000663, -0.00001315, 0.00000581, -0.00000951, 0.00000451, 0.00000343, 0.00000165, -0.00001195, -0.00000062, -0.00002476, 0.00000099, -0.00002598, 0.00000759, -0.00014666, 0.00002174, 0.00026801, 0.00017477, -0.00011455, -0.00005973, -0.00000972, -0.00002657, 0.00000617, -0.00001273, 0.00000760, -0.00000922, 0.00000037, -0.00000735, 0.00000010, -0.00000799, 0.00000611, -0.00000460, 0.00000197, -0.00000988, 0.00000913, 0.00000404, 0.00000168, -0.00000993, 0.00000819, -0.00000274, -0.00000777, -0.00001243, 0.00001708, -0.00000316, -0.00000160, -0.00001318, 0.00000961, -0.00000336, -0.00000549, -0.00000483, 0.00001040, -0.00000337, -0.00000158, -0.00000242, 0.00001076, -0.00000232, -0.00000230, -0.00001328, 0.00002430, -0.00000767, 0.00000595, 0.00000510, -0.00001480, 0.00001222, -0.00002082, 0.00000435, -0.00000080, 0.00000600, -0.00001382, -0.00000545, 0.00000235, -0.00001501, 0.00001662, -0.00000895, 0.00000836, 0.00000588, -0.00000767, -0.00000106, 0.00000607, 0.00000051, -0.00000104, -0.00000054, -0.00000488, 0.00000254, -0.00000538, 0.00000360, -0.00000407, 0.00000482, -0.00001664, 0.00000625, -0.00003473, 0.00001068, -0.00000106, 0.00004425, 0.00000008, 0.00003162, 0.00001719, 0.00000818, -0.00001942, -0.00001814, 0.00000232, -0.00000839, -0.00000763, -0.00000117, },
		{ 0.00000115, 0.00000482, 0.00000458, -0.00001531, 0.00000718, 0.00000302, -0.00000251, -0.00000051, 0.00000254, -0.00000618, -0.00000208, -0.00000215, -0.00000134, 0.00001347, -0.00000364, -0.00000282, -0.00000572, 0.00001321, -0.00000022, 0.00000065, -0.00000376, -0.00000093, -0.00000039, -0.00001107, 0.00000455, 0.00000500, -0.00000222, -0.00000731, 0.00000294, -0.00000805, 0.00000631, 0.00000446, -0.00000676, -0.00000435, -0.00001281, -0.00001315, -0.00001207, -0.00000693, -0.00003869, -0.00014824, 0.00014778, 0.00029448, 0.00004402, -0.00014441, -0.00000180, -0.00003925, 0.00000311, -0.00001264, 0.00000404, -0.00000549, 0.00000062, -0.00000946, 0.00000265, -0.00001047, 0.00000407, -0.00000341, 0.00000384, -0.00000895, 0.00000135, 0.00000502, 0.00000751, -0.00000893, 0.00000069, 0.00000307, -0.00000180, -0.00001650, 0.00000358, 0.00001050, -0.00000328, -0.00000765, -0.00000499, 0.00000805, -0.00000573, -0.00000701, 0.00000211, 0.00000676, -0.00000417, -0.00000244, 0.00000277, 0.00000762, -0.00000279, -0.00000879, -0.00000122, 0.00002280, -0.00001702, 0.00002055, -0.00001447, 0.00000196, -0.00000112, -0.00001552, 0.00000729, -0.00000095, 0.00000347, -0.00002009, 0.00000768, -0.00001215, -0.00000052, 0.00001134, -0.00001068, 0.00001700, -0.00000617, -0.00000243, -0.00000142, 0.00000977, -0.00000672, 0.00000599, -0.00000929, 0.00000477, -0.00000701, 0.00000480, -0.00000666, 0.00000799, -0.00001183, -0.00000021, -0.00001464, -0.00001750, 0.00000836, 0.00001393, 0.00003471, -0.00000209, 0.00003989, 0.00000323, 0.00000799, -0.00003087, -0.00000318, -0.00000340, -0.00000724, -0.00000670, },
		{ -0.00000431, 0.00000243, 0.00000819, -0.00000591, -0.00000940, 0.00001083, -0.00000178, -0.00000128, 0.00000097, -0.00000050, -0.00000636, 0.00000005, -0.00000457, 0.00000623, 0.00000899, -0.00000669, -0.00000286, -0.00000068, 0.00001445, -0.00000750, 0.00000585, -0.00001011, 0.00000767, -0.00001206, -0.00000130, 0.00000233, 0.00000790, -0.00001181, 0.00000308, -0.00000589, -0.00000056, 0.00000441, 0.00000438, -0.00001278, 0.00000119, -0.00002332, -0.00000193, -0.00002402, 0.00000893, -0.00010700, -0.00008278, 0.00026973, 0.00024345, -0.00009494, -0.00008742, -0.00000177, -0.00003603, 0.00000991, -0.00001636, 0.00000866, -0.00000884, 0.00000100, -0.00000924, 0.00000238, -0.00001061, 0.00000710, -0.00000555, 0.00000445, -0.00001242, 0.00000919, 0.00000238, 0.00000551, -0.00001256, 0.00000856, -0.00000223, -0.00000443, -0.00001699, 0.00001635, 0.00000012, -0.00000169, -0.00001213, 0.00000476, 0.00000187, -0.00000586, -0.00000712, 0.00000868, -0.00000000, -0.00000187, -0.00000389, 0.00000873, 0.00000160, -0.00000235, -0.00001432, 0.00001724, 0.00000505, -0.00000723, 0.00001769, -0.00002202, 0.00001419, -0.00001712, -0.00000164, 0.00000200, 0.00000460, -0.00000850, -0.00001216, 0.00000829, -0.00001998, 0.00001660, -0.00000516, 0.00000282, 0.00001012, -0.00000807, -0.00000191, 0.00000415, 0.00000408, -0.00000395, 0.00000327, -0.00000815, 0.00000512, -0.00000765, 0.00000627, -0.00000798, 0.00000996, -0.00001902, 0.00000831, -0.00003173, 0.00000609, -0.00000168, 0.00003561, 0.00000900, 0.00001733, 0.00002713, 0.00000349, -0.00000388, -0.00002811, 0.00000401, -0.00000763, -0.00000632, },
		{ -0.00000570, -0.00000190, 0.00000523, 0.00000619, -0.00001387, 0.00000153, 0.00000755, -0.00000358, 0.00000048, 0.00000061, -0.00000336, -0.00000436, -0.00000034, -0.00000375, 0.00001247, 0.00000017, -0.00000471, -0.00000496, 0.00000957, 0.00000458, -0.00000359, 0.00000140, -0.00000591, 0.00000509, -0.00001486, 0.00000456, 0.00000329, 0.00000307, -0.00001195, 0.00000653, -0.00001045, 0.00000522, 0.00000363, -0.00000085, -0.00001061, -0.00000312, -0.00002238, -0.00000269, -0.00002057, -0.00000438, -0.00015860, 0.00004262, 0.00035241, 0.00010107, -0.00015389, -0.00001873, -0.00003074, -0.00000892, -0.00000433, -0.00000428, 0.00000042, -0.00000247, -0.00000696, -0.00000049, -0.00000695, -0.00000043, -0.00000028, 0.00000162, -0.00000532, -0.00000359, 0.00000734, 0.00000515, -0.00000398, -0.00000495, 0.00000703, -0.00000308, -0.00001224, -0.00000512, 0.00001648, -0.00000504, -0.00000377, -0.00000951, 0.00000870, -0.00000286, -0.00000716, -0.00000227, 0.00000896, -0.00000332, -0.00000146, -0.00000162, 0.00001018, -0.00000211, -0.00000609, -0.00000935, 0.00002661, -0.00001350, 0.00001218, -0.00000105, -0.00001143, 0.00001021, -0.00002163, 0.00000827, -0.00000271, 0.00000772, -0.00002109, 0.00000510, -0.00000717, -0.00000850, 0.00001739, -0.00001308, 0.00001545, -0.00000256, -0.00000371, -0.00000298, 0.00001014, -0.00000498, 0.00000422, -0.00000600, 0.00000102, -0.00000338, 0.00000136, -0.00000288, 0.00000258, -0.00000306, -0.00000845, -0.00000420, -0.00002493, 0.00001246, 0.00000262, 0.00003993, -0.00000645, 0.00003605, 0.00000779, 0.00000949, -0.00002065, -0.00001466, 0.00000220, -0.00000834, },
		{ -0.00000658, -0.00000531, 0.00000090, 0.00000820, -0.00000197, -0.00001312, 0.00000965, 0.00000127, -0.00000205, 0.00000105, -0.00000102, -0.00000462, -0.00000201, -0.00000240, 0.00000168, 0.00001212, -0.00000607, -0.00000306, -0.00000281, 0.00001499, -0.00000608, 0.00000417, -0.00000699, 0.00000395, -0.00000634, -0.00000749, 0.00000433, 0.00000667, -0.00000703, -0.00000290, 0.00000041, -0.00000681, 0.00000610, 0.00000406, -0.00000841, -0.00000386, -0.00001376, -0.00001314, -0.00001200, -0.00000670, -0.00005192, -0.00016013, 0.00021089, 0.00033708, -0.00007044, -0.00011113, -0.00000180, -0.00003823, 0.00000627, -0.00001338, 0.00000386, -0.00000449, -0.00000081, -0.00000894, 0.00000220, -0.00001048, 0.00000557, -0.00000474, 0.00000527, -0.00001261, 0.00000691, 0.00000285, 0.00000641, -0.00001163, 0.00000557, 0.00000075, -0.00000362, -0.00001729, 0.00001067, 0.00000691, -0.00000460, -0.00000804, -0.00000182, 0.00000649, -0.00000532, -0.00000807, 0.00000475, 0.00000398, -0.00000236, -0.00000344, 0.00000423, 0.00000643, -0.00000314, -0.00001230, 0.00000673, 0.00001782, -0.00001642, 0.00002266, -0.00001950, 0.00000686, -0.00000664, -0.00001107, 0.00000671, 0.00000063, -0.00000054, -0.00001999, 0.00001377, -0.00002266, 0.00001336, 0.00000077, -0.00000294, 0.00001314, -0.00000731, -0.00000269, 0.00000170, 0.00000733, -0.00000589, 0.00000552, -0.00000915, 0.00000531, -0.00000772, 0.00000675, -0.00000896, 0.00001046, -0.00001570, 0.00000449, -0.00002369, -0.00000228, 0.00000260, 0.00002262, 0.00002082, 0.00000270, 0.00003524, -0.00000073, 0.00000764, -0.00002972, -0.00000101, -0.00000366, },
		{ -0.00000807, -0.00000439, -0.00000492, 0.00000525, 0.00000700, -0.00001055, -0.00000497, 0.00001102, -0.00000361, 0.00000126, -0.00000079, -0.00000190, -0.00000487, -0.00000037, -0.00000385, 0.00000916, 0.00000489, -0.00000642, -0.00000420, 0.00000588, 0.00000889, -0.00000683, 0.00000469, -0.00000818, 0.00000715, -0.00001452, 0.00000127, 0.00000360, 0.00000566, -0.00001254, 0.00000536, -0.00000776, 0.00000041, 0.00000472, 0.00000222, -0.00001210, -0.00000054, -0.00002175, -0.00000495, -0.00001940, 0.00000079, -0.00011922, -0.00008391, 0.00037300, 0.00019144, -0.00016582, -0.00003497, -0.00002554, -0.00001742, -0.00000057, -0.00000766, 0.00000091, -0.00000113, -0.00000723, -0.00000171, -0.00000512, -0.00000301, 0.00000092, 0.00000049, -0.00000212, -0.00000768, 0.00000872, 0.00000327, 0.00000011, -0.00000886, 0.00000883, -0.00000296, -0.00000867, -0.00001161, 0.00001828, -0.00000280, -0.00000355, -0.00000948, 0.00000497, 0.00000205, -0.00000751, -0.00000514, 0.00000809, -0.00000047, -0.00000131, -0.00000385, 0.00000907, 0.00000134, -0.00000537, -0.00001286, 0.00002258, -0.00000225, -0.00000159, 0.00001382, -0.00002183, 0.00001552, -0.00002098, 0.00000421, -0.00000110, 0.00000802, -0.00001701, -0.00000234, 0.00000174, -0.00001826, 0.00002264, -0.00001363, 0.00001265, 0.00000081, -0.00000395, -0.00000475, 0.00000959, -0.00000233, 0.00000175, -0.00000235, -0.00000237, -0.00000049, -0.00000148, 0.00000070, -0.00000259, 0.00000394, -0.00001313, 0.00000120, -0.00002668, 0.00001161, -0.00000204, 0.00003715, -0.00000206, 0.00002443, 0.00001786, 0.00000378, -0.00000538, -0.00002528, 0.00000496, },
		{ 0.00000223, -0.00000809, -0.00000403, -0.00000261, 0.00000919, 0.00000045, -0.00001355, 0.00000491, 0.00000616, -0.00000388, 0.00000276, -0.00000316, -0.00000201, -0.00000422, -0.00000047, -0.00000161, 0.00001278, -0.00000326, -0.00000448, -0.00000361, 0.00001365, -0.00000263, 0.00000036, -0.00000219, -0.00000083, -0.00000030, -0.00001258, 0.00000511, 0.00000518, -0.00000166, -0.00000854, 0.00000534, -0.00001042, 0.00000567, 0.00000332, -0.00000255, -0.00001031, -0.00000371, -0.00002291, -0.00000221, -0.00002046, -0.00001217, -0.00017972, 0.00008733, 0.00043632, -0.00002465, -0.00013814, -0.00000468, -0.00003785, -0.00000017, -0.00000860, -0.00000172, -0.00000105, -0.00000116, -0.00000931, 0.00000192, -0.00000982, 0.00000368, -0.00000399, 0.00000561, -0.00001130, 0.00000315, 0.00000468, 0.00000532, -0.00000847, 0.00000089, 0.00000459, -0.00000409, -0.00001480, 0.00000250, 0.00001404, -0.00000705, -0.00000442, -0.00000648, 0.00000778, -0.00000256, -0.00000875, 0.00000065, 0.00000633, -0.00000152, -0.00000259, -0.00000029, 0.00000947, -0.00000238, -0.00001011, -0.00000259, 0.00002495, -0.00001666, 0.00001714, -0.00000736, -0.00000682, 0.00000623, -0.00001973, 0.00001019, -0.00000320, 0.00000686, -0.00002533, 0.00001528, -0.00002026, 0.00000545, 0.00000914, -0.00000893, 0.00001538, -0.00000636, -0.00000250, -0.00000160, 0.00001020, -0.00000680, 0.00000628, -0.00000836, 0.00000399, -0.00000658, 0.00000583, -0.00000768, 0.00000753, -0.00000917, -0.00000150, -0.00001495, -0.00001003, 0.00000699, 0.00001073, 0.00002956, -0.00000626, 0.00003614, 0.00000121, 0.00000994, -0.00002106, -0.00001178, },
		{ 0.00000130, -0.00000455, -0.00000463, -0.00000587, 0.00000325, 0.00000843, -0.00000731, -0.00000976, 0.00001122, -0.00000113, 0.00000044, -0.00000017, -0.00000290, -0.00000314, -0.00000218, -0.00000224, 0.00000486, 0.00000888, -0.00000657, -0.00000451, 0.00000318, 0.00001151, -0.00000767, 0.00000470, -0.00000673, 0.00000531, -0.00001023, -0.00000468, 0.00000565, 0.00000535, -0.00000901, -0.00000007, -0.00000125, -0.00000658, 0.00000722, 0.00000168, -0.00000765, -0.00000590, -0.00001152, -0.00001740, -0.00000581, -0.00001546, -0.00005657, -0.00018595, 0.00032069, 0.00032322, -0.00017802, -0.00004994, -0.00002558, -0.00002028, -0.00000328, -0.00000481, -0.00000427, 0.00000394, -0.00000949, -0.00000136, -0.00000504, -0.00000354, 0.00000023, 0.00000041, 0.00000020, -0.00000998, 0.00000801, 0.00000326, 0.00000177, -0.00000965, 0.00000743, -0.00000049, -0.00000726, -0.00001406, 0.00001472, 0.00000370, -0.00000671, -0.00000588, -0.00000121, 0.00000699, -0.00000721, -0.00000659, 0.00000483, 0.00000315, -0.00000135, -0.00000408, 0.00000513, 0.00000613, -0.00000566, -0.00001276, 0.00001359, 0.00001127, -0.00001351, 0.00002247, -0.00002290, 0.00001105, -0.00001237, -0.00000475, 0.00000382, 0.00000423, -0.00000831, -0.00001310, 0.00001210, -0.00002684, 0.00002468, -0.00001110, 0.00000829, 0.00000414, -0.00000376, -0.00000587, 0.00000765, 0.00000107, -0.00000086, 0.00000080, -0.00000450, 0.00000099, -0.00000312, 0.00000312, -0.00000587, 0.00000715, -0.00001279, 0.00000112, -0.00002330, 0.00000672, -0.00000135, 0.00002857, 0.00000766, 0.00000967, 0.00002836, -0.00000267, 0.00000688, -0.00002847, },
		{ -0.00002300, 0.00000617, -0.00000869, -0.00000210, -0.00000636, 0.00000982, 0.00000209, -0.00001163, -0.00000157, 0.00001077, -0.00000512, 0.00000453, -0.00000502, -0.00000051, -0.00000494, -0.00000008, -0.00000279, 0.00001097, 0.00000084, -0.00000590, -0.00000429, 0.00001175, 0.00000118, -0.00000329, 0.00000146, -0.00000357, 0.00000289, -0.00001376, 0.00000244, 0.00000579, 0.00000121, -0.00001026, 0.00000530, -0.00000836, 0.00000077, 0.00000508, 0.00000011, -0.00001165, -0.00000130, -0.00002190, -0.00000547, -0.00001765, -0.00000580, -0.00013732, -0.00007298, 0.00049953, 0.00006781, -0.00017914, -0.00000259, -0.00004220, -0.00000242, -0.00000884, -0.00000171, -0.00000387, 0.00000380, -0.00001316, 0.00000401, -0.00001110, 0.00000381, -0.00000522, 0.00000691, -0.00000998, -0.00000030, 0.00000588, 0.00000449, -0.00000573, -0.00000270, 0.00000675, -0.00000352, -0.00001232, -0.00000443, 0.00001759, -0.00000555, -0.00000447, -0.00000658, 0.00000438, 0.00000276, -0.00000992, -0.00000191, 0.00000553, 0.00000139, -0.00000239, -0.00000296, 0.00000895, 0.00000114, -0.00000955, -0.00000781, 0.00002409, -0.00000782, 0.00000396, 0.00000877, -0.00001980, 0.00001479, -0.00002261, 0.00000899, -0.00000384, 0.00001031, -0.00002489, 0.00001033, -0.00001157, -0.00000685, 0.00001887, -0.00001439, 0.00001687, -0.00000570, -0.00000109, -0.00000548, 0.00001235, -0.00000672, 0.00000604, -0.00000675, 0.00000249, -0.00000555, 0.00000475, -0.00000562, 0.00000359, -0.00000301, -0.00000546, -0.00000961, -0.00001360, 0.00000797, 0.00000366, 0.00003160, -0.00000664, 0.00002820, 0.00001027, 0.00000384, -0.00000623, },
		{ -0.00002123, -0.00000879, 0.00000158, -0.00000656, -0.00000418, -0.00000109, 0.00001089, -0.00000528, -0.00001135, 0.00000743, 0.00000383, -0.00000205, 0.00000258, -0.00000597, -0.00000013, -0.00000459, -0.00000032, 0.00000112, 0.00001087, -0.00000467, -0.00000579, 0.00000132, 0.00001257, -0.00000636, 0.00000220, -0.00000309, 0.00000175, -0.00000477, -0.00001030, 0.00000690, 0.00000445, -0.00000425, -0.00000602, 0.00000470, -0.00001146, 0.00000780, 0.00000050, -0.00000122, -0.00001343, 0.00000019, -0.00002947, 0.00000662, -0.00003176, -0.00000760, -0.00022221, 0.00018175, 0.00047911, -0.00017066, -0.00007424, -0.00002420, -0.00002288, -0.00000750, -0.00000090, -0.00000962, 0.00000730, -0.00000926, -0.00000276, -0.00000406, -0.00000445, 0.00000002, -0.00000058, 0.00000282, -0.00001123, 0.00000592, 0.00000448, 0.00000171, -0.00000838, 0.00000419, 0.00000303, -0.00000712, -0.00001349, 0.00000782, 0.00001152, -0.00001003, -0.00000228, -0.00000598, 0.00000887, -0.00000455, -0.00000794, 0.00000137, 0.00000514, -0.00000001, -0.00000375, 0.00000072, 0.00000955, -0.00000464, -0.00001187, 0.00000443, 0.00002071, -0.00001730, 0.00002031, -0.00001304, -0.00000173, 0.00000084, -0.00001502, 0.00000922, -0.00000113, 0.00000178, -0.00002315, 0.00001957, -0.00003031, 0.00002088, -0.00000432, 0.00000182, 0.00000812, -0.00000430, -0.00000520, 0.00000382, 0.00000524, -0.00000348, 0.00000336, -0.00000560, 0.00000164, -0.00000421, 0.00000475, -0.00000733, 0.00000711, -0.00000914, -0.00000147, -0.00001850, 0.00000156, 0.00000058, 0.00001906, 0.00001683, -0.00000168, 0.00003289, -0.00000311, 0.00000995, },
		{ 0.00000672, -0.00002778, 0.00000409, -0.00000611, -0.00000163, -0.00000813, 0.00000830, 0.00000438, -0.00000929, -0.00000706, 0.00001253, -0.00000396, 0.00000466, -0.00000471, -0.00000161, -0.00000333, -0.00000154, -0.00000210, 0.00000772, 0.00000465, -0.00000591, -0.00000591, 0.00001035, 0.00000399, -0.00000503, 0.00000206, -0.00000267, 0.00000207, -0.00001049, -0.00000365, 0.00000873, 0.00000078, -0.00000723, -0.00000003, -0.00000114, -0.00000771, 0.00000945, -0.00000198, -0.00000561, -0.00000876, -0.00000868, -0.00002227, 0.00000176, -0.00002782, -0.00005949, -0.00021667, 0.00047829, 0.00022492, -0.00023556, 0.00000565, -0.00005360, 0.00000224, -0.00001678, 0.00000593, -0.00001384, 0.00001376, -0.00001988, 0.00000825, -0.00001447, 0.00000619, -0.00000840, 0.00000910, -0.00000909, -0.00000238, 0.00000512, 0.00000552, -0.00000519, -0.00000350, 0.00000583, -0.00000058, -0.00001172, -0.00000769, 0.00001557, 0.00000093, -0.00000861, -0.00000246, -0.00000227, 0.00000887, -0.00001081, -0.00000288, 0.00000212, 0.00000522, -0.00000235, -0.00000362, 0.00000530, 0.00000643, -0.00001012, -0.00000925, 0.00001719, 0.00000538, -0.00000977, 0.00002066, -0.00002461, 0.00001386, -0.00001704, 0.00000196, -0.00000019, 0.00000849, -0.00001796, -0.00000055, 0.00000151, -0.00002060, 0.00002712, -0.00001746, 0.00001667, -0.00000473, 0.00000059, -0.00000869, 0.00001289, -0.00000537, 0.00000500, -0.00000494, 0.00000152, -0.00000516, 0.00000399, -0.00000367, 0.00000040, 0.00000047, -0.00000544, -0.00000870, -0.00001286, 0.00000564, 0.00000119, 0.00002774, -0.00000043, 0.00001501, 0.00002184, -0.00000429, },
		{ 0.00000358, -0.00000669, -0.00002070, 0.00000682, -0.00000828, -0.00000120, -0.00000632, 0.00001332, -0.00000411, -0.00001017, 0.00000099, 0.00000950, -0.00000479, 0.00000612, -0.00000930, 0.00000223, -0.00000595, 0.00000058, -0.00000104, 0.00001066, -0.00000156, -0.00000679, -0.00000095, 0.00001334, -0.00000451, -0.00000049, -0.00000031, -0.00000007, -0.00000183, -0.00001209, 0.00000436, 0.00000603, -0.00000209, -0.00000800, 0.00000533, -0.00000991, 0.00000286, 0.00000284, 0.00000117, -0.00001494, 0.00000304, -0.00002892, 0.00000333, -0.00002806, 0.00000131, -0.00017873, -0.00001985, 0.00061185, -0.00010876, -0.00012546, -0.00000983, -0.00003491, -0.00000431, -0.00000448, -0.00000706, 0.00000200, -0.00000132, -0.00000982, 0.00000125, -0.00000901, 0.00000332, -0.00000491, 0.00000752, -0.00001312, 0.00000444, 0.00000472, 0.00000241, -0.00000783, 0.00000203, 0.00000485, -0.00000577, -0.00001278, 0.00000154, 0.00001623, -0.00000936, -0.00000261, -0.00000590, 0.00000548, 0.00000146, -0.00001041, -0.00000018, 0.00000360, 0.00000371, -0.00000410, -0.00000200, 0.00000940, -0.00000065, -0.00001245, -0.00000130, 0.00002243, -0.00001116, 0.00000856, 0.00000333, -0.00001637, 0.00001200, -0.00002120, 0.00001112, -0.00000436, 0.00000903, -0.00002805, 0.00002021, -0.00002595, 0.00001029, 0.00000646, -0.00000636, 0.00001289, -0.00000615, -0.00000234, -0.00000174, 0.00000980, -0.00000598, 0.00000558, -0.00000642, 0.00000259, -0.00000584, 0.00000652, -0.00000808, 0.00000584, -0.00000549, -0.00000251, -0.00001634, -0.00000025, 0.00000019, 0.00001266, 0.00002100, -0.00000536, 0.00002826, 0.00000473, },
		{ -0.00000684, 0.00001066, -0.00002207, -0.00000499, -0.00000026, -0.00000332, -0.00000679, 0.00000382, 0.00000812, -0.00000808, -0.00000949, 0.00001001, 0.00000027, 0.00000245, -0.00000156, -0.00000543, 0.00000011, -0.00000404, -0.00000061, 0.00000443, 0.00000694, -0.00000403, -0.00000832, 0.00000921, 0.00000589, -0.00000493, 0.00000013, 0.00000059, -0.00000080, -0.00000541, -0.00001017, 0.00001132, -0.00000055, -0.00000237, -0.00000688, 0.00000662, -0.00001503, 0.00001263, -0.00000550, 0.00000360, -0.00002004, 0.00000811, -0.00004111, 0.00002268, -0.00005388, 0.00000924, -0.00028804, 0.00034333, 0.00043724, -0.00028725, 0.00000679, -0.00006347, 0.00000727, -0.00002667, 0.00001539, -0.00002470, 0.00002242, -0.00002411, 0.00001049, -0.00001653, 0.00000769, -0.00001057, 0.00000968, -0.00000707, -0.00000402, 0.00000317, 0.00000770, -0.00000609, -0.00000252, 0.00000304, 0.00000353, -0.00001229, -0.00000789, 0.00000954, 0.00000976, -0.00001371, 0.00000226, -0.00000814, 0.00001229, -0.00000931, -0.00000393, -0.00000144, 0.00000730, -0.00000073, -0.00000368, 0.00000077, 0.00001073, -0.00000948, -0.00000957, 0.00000892, 0.00001620, -0.00001667, 0.00002207, -0.00001780, 0.00000315, -0.00000494, -0.00000839, 0.00000576, 0.00000320, -0.00000718, -0.00001358, 0.00001436, -0.00003105, 0.00003015, -0.00001585, 0.00001338, -0.00000213, 0.00000090, -0.00000957, 0.00001075, -0.00000226, 0.00000293, -0.00000277, 0.00000070, -0.00000470, 0.00000279, -0.00000145, -0.00000199, 0.00000150, -0.00000278, -0.00000964, -0.00001114, 0.00000341, -0.00000017, 0.00002216, 0.00000693, 0.00000324, 0.00002884, },
		{ 0.00001697, -0.00000626, 0.00000729, -0.00002740, 0.00000677, -0.00000727, 0.00000082, -0.00001004, 0.00001378, -0.00000238, -0.00000804, -0.00000538, 0.00001316, -0.00000574, 0.00000846, -0.00001074, 0.00000213, -0.00000511, -0.00000013, -0.00000161, 0.00000913, 0.00000121, -0.00000596, -0.00000477, 0.00001492, -0.00000362, -0.00000130, -0.00000078, 0.00000208, -0.00000332, -0.00000864, -0.00000280, 0.00001108, -0.00000418, -0.00000411, -0.00000091, -0.00000117, -0.00000791, 0.00001002, -0.00000352, -0.00000629, -0.00000757, -0.00001146, -0.00001959, -0.00000007, -0.00003009, -0.00007939, -0.00022436, 0.00065758, 0.00004103, -0.00021414, 0.00002320, -0.00006160, 0.00001138, -0.00002040, 0.00000745, -0.00001458, 0.00001541, -0.00002303, 0.00001149, -0.00001803, 0.00001093, -0.00001301, 0.00001452, -0.00001632, 0.00000492, 0.00000226, 0.00000587, -0.00001016, 0.00000306, 0.00000309, -0.00000144, -0.00001427, -0.00000104, 0.00001495, -0.00000268, -0.00000818, -0.00000048, -0.00000244, 0.00000956, -0.00001353, 0.00000051, -0.00000116, 0.00000881, -0.00000478, -0.00000255, 0.00000560, 0.00000578, -0.00001445, -0.00000311, 0.00001727, 0.00000124, -0.00000625, 0.00001788, -0.00002471, 0.00001477, -0.00001931, 0.00000708, -0.00000317, 0.00001077, -0.00002555, 0.00001295, -0.00001429, -0.00000505, 0.00001871, -0.00001432, 0.00001731, -0.00000853, 0.00000166, -0.00000762, 0.00001358, -0.00000776, 0.00000735, -0.00000732, 0.00000433, -0.00000832, 0.00000866, -0.00000877, 0.00000495, -0.00000416, 0.00000012, -0.00001812, 0.00000169, -0.00000273, 0.00000975, 0.00001992, -0.00000207, 0.00001690, },
		{ 0.00002665, -0.00000012, 0.00000408, -0.00000781, -0.00001740, 0.00000596, -0.00000578, -0.00000316, -0.00000264, 0.00001264, -0.00000816, -0.00000857, 0.00000384, 0.00000640, -0.00000122, 0.00000357, -0.00001056, 0.00000373, -0.00000616, 0.00000053, 0.00000195, 0.00000763, -0.00000127, -0.00000989, 0.00000673, 0.00000841, -0.00000482, -0.00000175, 0.00000274, -0.00000172, -0.00000291, -0.00001267, 0.00000956, 0.00000127, -0.00000051, -0.00000951, 0.00000867, -0.00001518, 0.00000924, -0.00000449, 0.00000794, -0.00002453, 0.00001502, -0.00004568, 0.00002522, -0.00005637, 0.00003043, -0.00025751, 0.00010764, 0.00065673, -0.00029061, -0.00002533, -0.00005432, -0.00000129, -0.00002619, 0.00001510, -0.00002556, 0.00002016, -0.00001821, 0.00000467, -0.00001203, 0.00000346, -0.00000723, 0.00000492, -0.00000114, -0.00000766, 0.00000267, 0.00000819, -0.00000536, -0.00000305, 0.00000179, 0.00000562, -0.00001118, -0.00000821, 0.00000379, 0.00001592, -0.00001484, 0.00000279, -0.00000881, 0.00000981, -0.00000349, -0.00000700, -0.00000238, 0.00000514, 0.00000389, -0.00000467, -0.00000210, 0.00001133, -0.00000544, -0.00001148, 0.00000357, 0.00001977, -0.00001324, 0.00001239, -0.00000222, -0.00001200, 0.00000759, -0.00001679, 0.00000996, -0.00000179, 0.00000292, -0.00002356, 0.00002182, -0.00003394, 0.00002540, -0.00000843, 0.00000637, 0.00000295, -0.00000135, -0.00000711, 0.00000554, 0.00000267, -0.00000032, 0.00000021, -0.00000094, -0.00000283, -0.00000011, 0.00000204, -0.00000464, 0.00000203, -0.00000078, -0.00000829, -0.00001267, 0.00000501, -0.00000401, 0.00001896, 0.00001050, -0.00000180, },
		{ 0.00000538, 0.00002607, -0.00001119, 0.00001268, -0.00002379, -0.00000058, -0.00000264, 0.00000053, -0.00001050, 0.00001087, 0.00000119, -0.00000696, -0.00000901, 0.00001264, -0.00000350, 0.00000820, -0.00000888, -0.00000120, -0.00000205, -0.00000207, -0.00000130, 0.00000739, 0.00000249, -0.00000286, -0.00000985, 0.00001688, -0.00000351, -0.00000021, -0.00000419, 0.00000713, -0.00000753, -0.00000236, -0.00001168, 0.00001693, -0.00000817, 0.00000290, -0.00001043, 0.00001022, -0.00001973, 0.00001816, -0.00001203, 0.00000859, -0.00002595, 0.00001453, -0.00005054, 0.00003656, -0.00007543, 0.00002372, -0.00035874, 0.00056467, 0.00028555, -0.00032657, 0.00006301, -0.00009471, 0.00003334, -0.00004326, 0.00002842, -0.00003632, 0.00003448, -0.00003666, 0.00002208, -0.00002770, 0.00001931, -0.00002144, 0.00002083, -0.00001887, 0.00000627, -0.00000212, 0.00001137, -0.00001461, 0.00000637, -0.00000117, 0.00000486, -0.00001747, -0.00000011, 0.00000865, 0.00000786, -0.00001617, 0.00000683, -0.00001085, 0.00001602, -0.00001484, 0.00000146, -0.00000643, 0.00001235, -0.00000389, -0.00000229, 0.00000032, 0.00001191, -0.00001556, -0.00000350, 0.00000967, 0.00001307, -0.00001597, 0.00002297, -0.00002152, 0.00000707, -0.00000968, -0.00000170, 0.00000163, 0.00000758, -0.00001701, 0.00000053, 0.00000063, -0.00002040, 0.00002822, -0.00001902, 0.00001922, -0.00000951, 0.00000462, -0.00001158, 0.00001488, -0.00000773, 0.00000796, -0.00000773, 0.00000610, -0.00001055, 0.00001004, -0.00000868, 0.00000434, -0.00000492, 0.00000508, -0.00002124, 0.00000398, -0.00000473, 0.00000705, 0.00001729, 0.00000324, },
		{ 0.00000037, 0.00001641, 0.00001295, -0.00000773, 0.00000758, -0.00002626, 0.00000810, -0.00000649, 0.00000062, -0.00000879, 0.00001610, -0.00000829, -0.00000584, -0.00000353, 0.00001187, -0.00000439, 0.00000871, -0.00001478, 0.00000561, -0.00000669, 0.00000075, 0.00000027, 0.00000767, 0.00000050, -0.00000860, 0.00000124, 0.00001308, -0.00000662, -0.00000089, 0.00000060, 0.00000251, -0.00000604, -0.00000849, 0.00000128, 0.00000857, -0.00000472, -0.00000437, 0.00000153, -0.00000529, -0.00000319, 0.00000467, 0.00000173, -0.00001478, 0.00000334, -0.00002666, -0.00000062, -0.00002318, -0.00000625, -0.00014012, -0.00016790, 0.00080530, -0.00019362, -0.00011551, -0.00001131, -0.00003515, -0.00000502, -0.00000431, -0.00000828, 0.00000067, 0.00000209, -0.00001234, 0.00000194, -0.00000935, 0.00000420, -0.00000709, 0.00001015, -0.00001497, 0.00000595, 0.00000422, 0.00000002, -0.00000818, 0.00000506, 0.00000288, -0.00000572, -0.00001179, 0.00000234, 0.00001534, -0.00000864, -0.00000349, -0.00000250, 0.00000097, 0.00000631, -0.00001232, 0.00000022, -0.00000151, 0.00001095, -0.00000663, -0.00000229, 0.00000747, 0.00000215, -0.00001559, 0.00000250, 0.00001570, -0.00000175, -0.00000292, 0.00001426, -0.00002323, 0.00001358, -0.00001837, 0.00000904, -0.00000322, 0.00000854, -0.00002677, 0.00002103, -0.00002795, 0.00001332, 0.00000342, -0.00000312, 0.00000980, -0.00000576, -0.00000195, -0.00000167, 0.00000845, -0.00000426, 0.00000402, -0.00000389, 0.00000110, -0.00000518, 0.00000712, -0.00000821, 0.00000375, -0.00000203, -0.00000219, -0.00001911, 0.00001114, -0.00001071, 0.00001889, 0.00000926, },
		{ 0.00000515, -0.00000003, 0.00002598, -0.00000599, 0.00000648, -0.00001037, -0.00001274, 0.00000359, -0.00000147, -0.00000768, 0.00000475, 0.00000663, -0.00000781, -0.00000874, 0.00000764, 0.00000167, 0.00000532, -0.00000375, -0.00000709, 0.00000216, -0.00000430, -0.00000081, 0.00000593, 0.00000251, 0.00000109, -0.00001400, 0.00001692, -0.00000200, 0.00000042, -0.00000736, 0.00001095, -0.00000967, 0.00000144, -0.00001673, 0.00001835, -0.00000879, 0.00000696, -0.00001609, 0.00001631, -0.00002461, 0.00001982, -0.00001629, 0.00001935, -0.00003868, 0.00003197, -0.00006843, 0.00005551, -0.00009761, 0.00007632, -0.00037061, 0.00032501, 0.00058559, -0.00041640, 0.00007713, -0.00011133, 0.00004335, -0.00005723, 0.00004113, -0.00004928, 0.00004331, -0.00004028, 0.00002445, -0.00003048, 0.00002151, -0.00002375, 0.00002096, -0.00001651, 0.00000487, -0.00000483, 0.00001517, -0.00001721, 0.00000785, -0.00000385, 0.00000924, -0.00001892, 0.00000072, 0.00000202, 0.00001669, -0.00002086, 0.00001031, -0.00001424, 0.00001644, -0.00001129, -0.00000020, -0.00000857, 0.00001103, 0.00000063, -0.00000319, -0.00000347, 0.00001442, -0.00001292, -0.00000573, 0.00000457, 0.00001831, -0.00001541, 0.00001600, -0.00000778, -0.00000728, 0.00000270, -0.00001065, 0.00000652, 0.00000276, -0.00000660, -0.00001190, 0.00001315, -0.00003045, 0.00003104, -0.00001793, 0.00001680, -0.00000727, 0.00000449, -0.00001174, 0.00001245, -0.00000505, 0.00000657, -0.00000659, 0.00000634, -0.00001058, 0.00000873, -0.00000625, 0.00000248, -0.00000541, 0.00000858, -0.00002096, 0.00000169, -0.00000146, 0.00000058, 0.00001749, },
		{ 0.00000666, 0.00000523, 0.00000445, 0.00002463, -0.00001583, 0.00001533, -0.00002549, 0.00000321, -0.00000385, 0.00000265, -0.00001232, 0.00001654, -0.00000680, -0.00000361, -0.00000904, 0.00001402, -0.00000514, 0.00001192, -0.00001607, 0.00000440, -0.00000517, 0.00000013, -0.00000115, 0.00000816, -0.00000004, -0.00000370, -0.00000720, 0.00001961, -0.00001035, 0.00000311, -0.00000605, 0.00001140, -0.00001371, 0.00000066, -0.00001128, 0.00001928, -0.00001335, 0.00000669, -0.00001225, 0.00001111, -0.00002086, 0.00001919, -0.00001319, 0.00000711, -0.00002345, 0.00001004, -0.00004583, 0.00003267, -0.00007622, 0.00000797, -0.00039352, 0.00080436, 0.00003543, -0.00026532, 0.00006183, -0.00009131, 0.00003462, -0.00004086, 0.00002460, -0.00003233, 0.00003221, -0.00003628, 0.00002199, -0.00002792, 0.00002100, -0.00002355, 0.00002424, -0.00002439, 0.00001240, -0.00000407, 0.00001001, -0.00001776, 0.00001254, -0.00000425, 0.00000367, -0.00001879, 0.00000580, 0.00000801, 0.00000372, -0.00001463, 0.00000802, -0.00001067, 0.00001660, -0.00001749, 0.00000447, -0.00000995, 0.00001742, -0.00000764, -0.00000097, 0.00000096, 0.00001078, -0.00001976, 0.00000356, 0.00000789, 0.00001122, -0.00001536, 0.00002302, -0.00002388, 0.00000931, -0.00001191, 0.00000287, -0.00000066, 0.00000889, -0.00002299, 0.00001309, -0.00001568, -0.00000232, 0.00001587, -0.00001206, 0.00001613, -0.00001026, 0.00000360, -0.00000847, 0.00001311, -0.00000754, 0.00000772, -0.00000744, 0.00000621, -0.00001113, 0.00001240, -0.00001179, 0.00000647, -0.00000630, 0.00000737, -0.00002781, 0.00001815, -0.00001655, 0.00001857, },
		{ 0.00002184, 0.00000176, 0.00000245, 0.00001633, 0.00000780, -0.00000713, 0.00000629, -0.00002334, 0.00000729, -0.00000304, -0.00000319, -0.00000280, 0.00001257, -0.00000991, -0.00000541, -0.00000012, 0.00000786, 0.00000116, 0.00000326, -0.00001373, 0.00000591, -0.00000593, -0.00000032, 0.00000426, 0.00000279, 0.00000325, -0.00001414, 0.00001224, 0.00000352, -0.00000245, -0.00000639, 0.00000890, -0.00000495, -0.00000207, -0.00001330, 0.00001104, -0.00000154, 0.00000271, -0.00001177, 0.00001069, -0.00001650, 0.00000913, -0.00000904, 0.00001612, -0.00003326, 0.00002593, -0.00005624, 0.00003721, -0.00007199, 0.00005106, -0.00025650, -0.00000992, 0.00085618, -0.00041487, 0.00002466, -0.00008452, 0.00002046, -0.00004442, 0.00002960, -0.00003932, 0.00003012, -0.00002477, 0.00001117, -0.00001966, 0.00001123, -0.00001425, 0.00001020, -0.00000542, -0.00000295, -0.00000168, 0.00001266, -0.00001316, 0.00000275, -0.00000038, 0.00000740, -0.00001455, -0.00000298, 0.00000041, 0.00001814, -0.00001714, 0.00000553, -0.00000908, 0.00000865, -0.00000173, -0.00000597, -0.00000549, 0.00000352, 0.00000906, -0.00000602, -0.00000393, 0.00001154, -0.00000540, -0.00001131, 0.00000440, 0.00001510, -0.00000512, 0.00000093, 0.00000966, -0.00002034, 0.00001075, -0.00001454, 0.00000770, -0.00000006, 0.00000137, -0.00001969, 0.00001903, -0.00003213, 0.00002581, -0.00001085, 0.00000998, -0.00000149, 0.00000076, -0.00000784, 0.00000650, 0.00000004, 0.00000303, -0.00000324, 0.00000385, -0.00000698, 0.00000354, -0.00000051, -0.00000188, -0.00000325, 0.00000721, -0.00001382, -0.00000783, 0.00000863, -0.00001067, },
		{ 0.00001307, 0.00000777, 0.00000666, -0.00000192, 0.00002738, -0.00001331, 0.00001075, -0.00001401, -0.00000754, 0.00000111, 0.00000265, -0.00001225, 0.00001313, -0.00000273, -0.00000363, -0.00001056, 0.00001149, -0.00000270, 0.00001211, -0.00001342, -0.00000013, -0.00000196, -0.00000092, -0.00000255, 0.00000922, -0.00000239, 0.00000307, -0.00001589, 0.00002506, -0.00001320, 0.00000732, -0.00001351, 0.00002020, -0.00002044, 0.00000871, -0.00002165, 0.00002711, -0.00002008, 0.00001659, -0.00002440, 0.00002478, -0.00003427, 0.00003002, -0.00002719, 0.00002913, -0.00004992, 0.00004462, -0.00008545, 0.00007926, -0.00013324, 0.00011516, -0.00048826, 0.00061383, 0.00037902, -0.00043626, 0.00013327, -0.00014576, 0.00007330, -0.00007757, 0.00005722, -0.00006361, 0.00005804, -0.00005514, 0.00003799, -0.00004326, 0.00003482, -0.00003662, 0.00003428, -0.00003047, 0.00001756, -0.00001243, 0.00002026, -0.00002721, 0.00001959, -0.00001121, 0.00001269, -0.00002535, 0.00001037, -0.00000133, 0.00001649, -0.00002446, 0.00001641, -0.00001881, 0.00002199, -0.00001848, 0.00000660, -0.00001568, 0.00001914, -0.00000496, -0.00000055, -0.00000482, 0.00001659, -0.00002049, 0.00000295, 0.00000189, 0.00001868, -0.00001818, 0.00001949, -0.00001294, -0.00000311, -0.00000104, -0.00000503, 0.00000324, 0.00000632, -0.00001534, 0.00000226, -0.00000234, -0.00001611, 0.00002444, -0.00001712, 0.00001923, -0.00001226, 0.00000679, -0.00001222, 0.00001462, -0.00000860, 0.00000984, -0.00000992, 0.00001031, -0.00001530, 0.00001523, -0.00001307, 0.00000806, -0.00001066, 0.00001589, -0.00003307, 0.00002001, -0.00001599, },
		{ -0.00002296, 0.00002699, -0.00000378, 0.00000887, 0.00000290, 0.00002260, -0.00001901, 0.00001701, -0.00002573, 0.00000495, -0.00000230, 0.00000098, -0.00000940, 0.00001693, -0.00001154, -0.00000127, -0.00000761, 0.00001242, -0.00000233, 0.00001012, -0.00001893, 0.00000767, -0.00000646, 0.00000044, 0.00000152, 0.00000496, 0.00000149, -0.00000851, 0.00000182, 0.00001375, -0.00000981, 0.00000028, -0.00000060, 0.00000766, -0.00001337, -0.00000140, -0.00000451, 0.00001330, -0.00001014, 0.00000291, -0.00000654, 0.00000375, -0.00001288, 0.00001002, -0.00000292, -0.00000734, -0.00000543, -0.00001360, -0.00001661, -0.00000307, -0.00003674, -0.00006801, -0.00034040, 0.00099291, -0.00026055, -0.00011877, -0.00000276, -0.00004335, 0.00000177, -0.00001164, -0.00000245, -0.00000707, 0.00001017, -0.00001766, 0.00000502, -0.00001197, 0.00000729, -0.00001119, 0.00001399, -0.00001751, 0.00000855, 0.00000208, -0.00000059, -0.00001054, 0.00001056, -0.00000183, -0.00000321, -0.00001285, 0.00000601, 0.00001046, -0.00000434, -0.00000726, 0.00000344, -0.00000512, 0.00001138, -0.00001427, 0.00000166, -0.00000810, 0.00001881, -0.00000927, -0.00000161, 0.00000430, 0.00000523, -0.00001871, 0.00000773, 0.00000647, 0.00000905, -0.00001397, 0.00002171, -0.00002450, 0.00000956, -0.00001095, 0.00000399, 0.00000050, 0.00000501, -0.00002119, 0.00001763, -0.00002626, 0.00001473, -0.00000034, 0.00000096, 0.00000613, -0.00000503, -0.00000168, -0.00000099, 0.00000594, -0.00000166, 0.00000177, -0.00000112, -0.00000016, -0.00000469, 0.00000756, -0.00000819, 0.00000182, 0.00000052, -0.00000043, -0.00002245, 0.00002217, },
		{ -0.00000578, 0.00000000, 0.00001507, 0.00000150, 0.00000070, 0.00001891, 0.00000012, -0.00000357, 0.00000296, -0.00001875, 0.00000500, 0.00000192, -0.00000918, 0.00000662, 0.00000348, -0.00000620, -0.00000784, 0.00000490, 0.00000215, 0.00000951, -0.00000712, -0.00000713, 0.00000209, -0.00000221, -0.00000336, 0.00000955, -0.00000414, 0.00000796, -0.00002034, 0.00002497, -0.00001090, 0.00000710, -0.00001612, 0.00002229, -0.00001926, 0.00000850, -0.00002265, 0.00002506, -0.00001761, 0.00001699, -0.00002557, 0.00002566, -0.00003337, 0.00002707, -0.00002844, 0.00003627, -0.00005738, 0.00005453, -0.00009285, 0.00008472, -0.00013599, 0.00013008, -0.00041871, 0.00026097, 0.00075905, -0.00055163, 0.00014989, -0.00016080, 0.00008118, -0.00008895, 0.00006683, -0.00007276, 0.00006200, -0.00005445, 0.00003772, -0.00004427, 0.00003530, -0.00003682, 0.00003211, -0.00002638, 0.00001514, -0.00001440, 0.00002393, -0.00002950, 0.00001932, -0.00001149, 0.00001517, -0.00002566, 0.00001001, -0.00000575, 0.00002227, -0.00002596, 0.00001621, -0.00001782, 0.00001827, -0.00001226, 0.00000342, -0.00001502, 0.00001332, 0.00000290, -0.00000277, -0.00000698, 0.00001656, -0.00001537, -0.00000204, 0.00000143, 0.00001707, -0.00000996, 0.00000576, 0.00000418, -0.00001669, 0.00000758, -0.00000956, 0.00000480, 0.00000424, -0.00000798, -0.00000668, 0.00000726, -0.00002392, 0.00002639, -0.00001665, 0.00001773, -0.00001025, 0.00000600, -0.00001159, 0.00001216, -0.00000667, 0.00000927, -0.00000971, 0.00001114, -0.00001513, 0.00001321, -0.00000995, 0.00000615, -0.00001155, 0.00001850, -0.00002973, 0.00001225, },
		{ 0.00003209, -0.00002954, 0.00002688, -0.00000554, 0.00001238, -0.00000605, 0.00002953, -0.00002068, 0.00001546, -0.00001724, -0.00000350, 0.00000061, 0.00000379, -0.00001325, 0.00001803, -0.00001100, 0.00000124, -0.00001198, 0.00001322, -0.00000365, 0.00001497, -0.00002085, 0.00000648, -0.00000576, 0.00000159, -0.00000244, 0.00000919, -0.00000392, 0.00000148, -0.00001180, 0.00002574, -0.00001886, 0.00000988, -0.00001406, 0.00002355, -0.00002710, 0.00001299, -0.00002189, 0.00002942, -0.00002529, 0.00002099, -0.00002734, 0.00002700, -0.00003679, 0.00003200, -0.00002857, 0.00002776, -0.00004744, 0.00004020, -0.00008085, 0.00007591, -0.00013638, 0.00010957, -0.00055626, 0.00091623, 0.00006320, -0.00033530, 0.00011296, -0.00013063, 0.00006661, -0.00006936, 0.00004852, -0.00005539, 0.00005217, -0.00005132, 0.00003449, -0.00003996, 0.00003316, -0.00003582, 0.00003514, -0.00003361, 0.00002182, -0.00001292, 0.00001675, -0.00002796, 0.00002477, -0.00001438, 0.00001125, -0.00002550, 0.00001559, -0.00000232, 0.00001240, -0.00002232, 0.00001717, -0.00001854, 0.00002240, -0.00002072, 0.00000877, -0.00001892, 0.00002525, -0.00000982, 0.00000098, -0.00000390, 0.00001492, -0.00002413, 0.00001052, -0.00000179, 0.00001948, -0.00002069, 0.00002216, -0.00001695, -0.00000042, -0.00000217, -0.00000212, 0.00000270, 0.00000560, -0.00001826, 0.00001187, -0.00001683, 0.00000236, 0.00000946, -0.00000700, 0.00001276, -0.00001012, 0.00000381, -0.00000719, 0.00001038, -0.00000569, 0.00000681, -0.00000677, 0.00000754, -0.00001306, 0.00001494, -0.00001382, 0.00000746, -0.00000840, 0.00001436, -0.00003582, },
		{ -0.00000461, 0.00001198, -0.00001781, 0.00002596, -0.00000798, 0.00000945, 0.00000390, 0.00001771, -0.00001895, 0.00001637, -0.00002371, 0.00000429, 0.00000210, -0.00000460, -0.00000118, 0.00001030, -0.00001019, -0.00000243, -0.00000345, 0.00000744, 0.00000529, 0.00000150, -0.00001497, 0.00000595, -0.00000384, -0.00000253, 0.00000731, -0.00000244, 0.00000729, -0.00001689, 0.00001629, -0.00000069, -0.00000067, -0.00001000, 0.00001353, -0.00000642, -0.00000336, -0.00001119, 0.00001036, -0.00000285, 0.00000407, -0.00001146, 0.00000953, -0.00001437, 0.00000638, -0.00001119, 0.00002034, -0.00003596, 0.00002885, -0.00005681, 0.00003772, -0.00007280, 0.00004829, -0.00021832, -0.00015862, 0.00105478, -0.00052272, 0.00006204, -0.00010556, 0.00003520, -0.00005607, 0.00003730, -0.00004631, 0.00003363, -0.00002647, 0.00001469, -0.00002479, 0.00001662, -0.00001901, 0.00001354, -0.00000861, 0.00000140, -0.00000548, 0.00001605, -0.00001960, 0.00000715, -0.00000097, 0.00000727, -0.00001601, 0.00000068, -0.00000067, 0.00001676, -0.00001581, 0.00000510, -0.00000645, 0.00000543, 0.00000066, -0.00000517, -0.00000754, 0.00000062, 0.00001448, -0.00000715, -0.00000494, 0.00001039, -0.00000485, -0.00001105, 0.00000632, 0.00000803, 0.00000477, -0.00001075, 0.00001862, -0.00002344, 0.00000840, -0.00000757, 0.00000224, 0.00000447, -0.00000329, -0.00001129, 0.00001125, -0.00002504, 0.00002229, -0.00001171, 0.00001251, -0.00000485, 0.00000177, -0.00000739, 0.00000673, -0.00000246, 0.00000614, -0.00000638, 0.00000788, -0.00000986, 0.00000601, -0.00000228, 0.00000017, -0.00000742, 0.00001319, -0.00001672, },
		{ -0.00004083, 0.00003167, -0.00002484, 0.00001705, 0.00000077, 0.00000805, -0.00000515, 0.00002404, -0.00000959, 0.00000223, -0.00000145, -0.00001394, 0.00000326, 0.00000561, -0.00001381, 0.00001536, -0.00000748, 0.00000060, -0.00001170, 0.00000960, -0.00000231, 0.00001667, -0.00001842, 0.00000202, -0.00000375, 0.00000258, -0.00000633, 0.00001348, -0.00000984, 0.00001116, -0.00002292, 0.00003359, -0.00002378, 0.00001620, -0.00002415, 0.00003421, -0.00003421, 0.00002049, -0.00003149, 0.00003725, -0.00003280, 0.00003141, -0.00003934, 0.00003987, -0.00004892, 0.00004297, -0.00004495, 0.00005244, -0.00007592, 0.00007567, -0.00011986, 0.00012096, -0.00018897, 0.00019708, -0.00058470, 0.00061362, 0.00050286, -0.00055493, 0.00020722, -0.00019938, 0.00011595, -0.00011408, 0.00008704, -0.00009065, 0.00007990, -0.00007178, 0.00005323, -0.00005858, 0.00005011, -0.00005138, 0.00004719, -0.00004205, 0.00002984, -0.00002416, 0.00003039, -0.00004077, 0.00003372, -0.00002219, 0.00002119, -0.00003372, 0.00002186, -0.00001213, 0.00002445, -0.00003115, 0.00002401, -0.00002421, 0.00002530, -0.00002056, 0.00001105, -0.00002370, 0.00002405, -0.00000496, 0.00000112, -0.00000923, 0.00001950, -0.00002387, 0.00000879, -0.00000459, 0.00002090, -0.00001582, 0.00001102, -0.00000137, -0.00001339, 0.00000588, -0.00000593, 0.00000323, 0.00000606, -0.00001439, 0.00000592, -0.00000836, -0.00000705, 0.00001532, -0.00001147, 0.00001621, -0.00001219, 0.00000626, -0.00000997, 0.00001172, -0.00000755, 0.00001006, -0.00001073, 0.00001301, -0.00001799, 0.00001814, -0.00001554, 0.00001023, -0.00001479, 0.00002416, },
		{ 0.00000186, -0.00002461, 0.00003015, -0.00003309, 0.00003374, -0.00001533, 0.00001712, -0.00000891, 0.00002968, -0.00002572, 0.00001863, -0.00001855, -0.00000196, 0.00000344, 0.00000030, -0.00000829, 0.00001582, -0.00001372, 0.00000309, -0.00001057, 0.00001098, 0.00000100, 0.00001054, -0.00002164, 0.00000853, -0.00000581, 0.00000052, 0.00000163, 0.00000384, 0.00000009, -0.00000550, 0.00000008, 0.00001585, -0.00001454, 0.00000360, -0.00000431, 0.00001521, -0.00002322, 0.00000850, -0.00001239, 0.00001976, -0.00001796, 0.00001327, -0.00001808, 0.00001626, -0.00002547, 0.00001893, -0.00001332, 0.00000770, -0.00002311, 0.00000944, -0.00004319, 0.00002990, -0.00008493, 0.00002460, -0.00051315, 0.00114755, -0.00029163, -0.00014346, 0.00001977, -0.00006210, 0.00001804, -0.00002784, 0.00001122, -0.00002127, 0.00002246, -0.00002557, 0.00001073, -0.00001726, 0.00001286, -0.00001728, 0.00001907, -0.00002102, 0.00001276, -0.00000231, 0.00000129, -0.00001521, 0.00001827, -0.00000896, 0.00000168, -0.00001608, 0.00001255, 0.00000201, 0.00000279, -0.00001292, 0.00001063, -0.00001140, 0.00001555, -0.00001555, 0.00000332, -0.00001457, 0.00002546, -0.00001089, -0.00000065, 0.00000077, 0.00000758, -0.00002069, 0.00001285, -0.00000294, 0.00001831, -0.00002136, 0.00002296, -0.00001915, 0.00000051, -0.00000033, -0.00000267, 0.00000584, -0.00000064, -0.00001248, 0.00001135, -0.00002211, 0.00001545, -0.00000541, 0.00000609, 0.00000172, -0.00000361, -0.00000209, 0.00000074, 0.00000217, 0.00000177, -0.00000102, 0.00000167, -0.00000116, -0.00000408, 0.00000746, -0.00000778, 0.00000008, 0.00000225, },
		{ 0.00002476, -0.00003384, 0.00001877, -0.00000850, -0.00000176, 0.00001429, -0.00000354, 0.00000462, 0.00000891, 0.00000911, -0.00001519, 0.00001342, -0.00002010, 0.00000257, 0.00000729, -0.00001147, 0.00000955, -0.00000134, -0.00000327, -0.00000706, 0.00000258, 0.00000105, 0.00001502, -0.00001154, -0.00000536, -0.00000035, 0.00000217, -0.00000783, 0.00001451, -0.00001172, 0.00001483, -0.00002544, 0.00003148, -0.00001895, 0.00001320, -0.00002386, 0.00003202, -0.00002720, 0.00001380, -0.00002614, 0.00002969, -0.00002503, 0.00002549, -0.00003302, 0.00003244, -0.00003931, 0.00003237, -0.00003906, 0.00005028, -0.00007120, 0.00007013, -0.00010802, 0.00010289, -0.00015976, 0.00015964, -0.00042506, 0.00015700, 0.00093906, -0.00067304, 0.00021028, -0.00020114, 0.00011213, -0.00011438, 0.00008596, -0.00008942, 0.00007385, -0.00006315, 0.00004740, -0.00005496, 0.00004609, -0.00004695, 0.00004064, -0.00003454, 0.00002464, -0.00002326, 0.00003154, -0.00004016, 0.00002918, -0.00001772, 0.00001954, -0.00003071, 0.00001789, -0.00001155, 0.00002453, -0.00002737, 0.00001860, -0.00001794, 0.00001729, -0.00001178, 0.00000610, -0.00001979, 0.00001387, 0.00000558, -0.00000217, -0.00000956, 0.00001694, -0.00001661, 0.00000121, -0.00000040, 0.00001311, -0.00000206, -0.00000562, 0.00001409, -0.00002152, 0.00000738, -0.00000400, 0.00000010, 0.00000819, -0.00001187, 0.00000228, -0.00000307, -0.00001204, 0.00001679, -0.00001239, 0.00001620, -0.00001077, 0.00000515, -0.00000909, 0.00000993, -0.00000693, 0.00001057, -0.00001141, 0.00001407, -0.00001723, 0.00001523, -0.00001162, 0.00000793, -0.00001526, },
		{ -0.00000910, 0.00001881, -0.00003928, 0.00004014, -0.00003812, 0.00003179, -0.00001432, 0.00001690, -0.00001210, 0.00002921, -0.00001901, 0.00000822, -0.00000511, -0.00001079, 0.00000430, 0.00000518, -0.00001332, 0.00001854, -0.00001499, 0.00000624, -0.00001397, 0.00001118, -0.00000202, 0.00001792, -0.00002508, 0.00000867, -0.00000747, 0.00000490, -0.00000604, 0.00001266, -0.00001056, 0.00000919, -0.00001778, 0.00003238, -0.00002817, 0.00001799, -0.00002336, 0.00003643, -0.00004113, 0.00002607, -0.00003253, 0.00003952, -0.00003795, 0.00003640, -0.00004355, 0.00004373, -0.00005323, 0.00004641, -0.00004733, 0.00005238, -0.00007514, 0.00007342, -0.00011781, 0.00012092, -0.00019803, 0.00020607, -0.00068774, 0.00097398, 0.00012950, -0.00042029, 0.00017120, -0.00017482, 0.00010312, -0.00010184, 0.00007524, -0.00007978, 0.00007168, -0.00006541, 0.00004700, -0.00005232, 0.00004551, -0.00004790, 0.00004551, -0.00004264, 0.00003200, -0.00002312, 0.00002484, -0.00003901, 0.00003760, -0.00002534, 0.00001967, -0.00003293, 0.00002657, -0.00001391, 0.00002110, -0.00002902, 0.00002476, -0.00002409, 0.00002551, -0.00002209, 0.00001196, -0.00002609, 0.00003055, -0.00001047, 0.00000268, -0.00000803, 0.00001715, -0.00002643, 0.00001619, -0.00000975, 0.00002403, -0.00002088, 0.00001534, -0.00000584, -0.00001146, 0.00000696, -0.00000560, 0.00000529, 0.00000247, -0.00001308, 0.00001091, -0.00001888, 0.00000954, -0.00000066, 0.00000079, 0.00000714, -0.00000769, 0.00000176, -0.00000348, 0.00000538, -0.00000217, 0.00000447, -0.00000501, 0.00000766, -0.00001315, 0.00001528, -0.00001398, 0.00000707, },
		{ -0.00000983, 0.00001649, -0.00001693, -0.00000246, 0.00001371, -0.00002308, 0.00002880, -0.00001639, 0.00001649, -0.00000745, 0.00002559, -0.00002669, 0.00001912, -0.00001753, -0.00000266, 0.00000884, -0.00000683, 0.00000186, 0.00000616, -0.00000916, 0.00000008, -0.00000558, 0.00000505, 0.00001058, -0.00000120, -0.00001447, 0.00000417, -0.00000064, -0.00000503, 0.00000977, -0.00000645, 0.00000913, -0.00001660, 0.00001758, -0.00000303, -0.00000069, -0.00001117, 0.00001523, -0.00000527, -0.00000716, -0.00000618, 0.00000728, -0.00000217, 0.00000294, -0.00000809, 0.00000489, -0.00000871, 0.00000066, -0.00000973, 0.00001870, -0.00003067, 0.00002198, -0.00004589, 0.00002545, -0.00005777, 0.00002314, -0.00015151, -0.00031713, 0.00122489, -0.00059274, 0.00007425, -0.00010965, 0.00003716, -0.00005651, 0.00003451, -0.00004372, 0.00002883, -0.00002217, 0.00001405, -0.00002587, 0.00001799, -0.00001999, 0.00001373, -0.00000960, 0.00000408, -0.00000724, 0.00001676, -0.00002292, 0.00000870, 0.00000080, 0.00000472, -0.00001491, 0.00000190, 0.00000127, 0.00001150, -0.00001099, 0.00000192, -0.00000162, 0.00000102, 0.00000306, -0.00000441, -0.00000880, -0.00000253, 0.00001874, -0.00000724, -0.00000552, 0.00000843, -0.00000439, -0.00001005, 0.00000823, 0.00000056, 0.00001355, -0.00001918, 0.00002148, -0.00001967, 0.00000054, 0.00000312, -0.00000525, 0.00001089, -0.00001011, 0.00000028, 0.00000018, -0.00001440, 0.00001618, -0.00001183, 0.00001427, -0.00000721, 0.00000194, -0.00000621, 0.00000661, -0.00000484, 0.00000875, -0.00000878, 0.00001060, -0.00001107, 0.00000711, -0.00000319, 0.00000125, },
		{ 0.00000990, -0.00001580, 0.00002774, -0.00004229, 0.00003673, -0.00002936, 0.00001797, -0.00000324, 0.00000696, -0.00000419, 0.00001654, -0.00000172, -0.00000914, 0.00000961, -0.00001678, 0.00000227, 0.00000998, -0.00001541, 0.00001764, -0.00001290, 0.00000546, -0.00001246, 0.00000763, -0.00000294, 0.00002182, -0.00002354, 0.00000521, -0.00000748, 0.00000824, -0.00001218, 0.00001938, -0.00001860, 0.00001969, -0.00002901, 0.00004055, -0.00003325, 0.00002432, -0.00003356, 0.00004618, -0.00004587, 0.00003032, -0.00003881, 0.00004503, -0.00004379, 0.00004447, -0.00005216, 0.00005244, -0.00006086, 0.00005422, -0.00006168, 0.00007365, -0.00009804, 0.00010079, -0.00014591, 0.00015224, -0.00023048, 0.00025411, -0.00063410, 0.00056194, 0.00064504, -0.00066631, 0.00027144, -0.00024521, 0.00015271, -0.00014527, 0.00011119, -0.00011150, 0.00009516, -0.00008292, 0.00006474, -0.00007063, 0.00006212, -0.00006279, 0.00005692, -0.00005131, 0.00004082, -0.00003490, 0.00003923, -0.00005234, 0.00004570, -0.00003146, 0.00002813, -0.00004041, 0.00003192, -0.00002115, 0.00002937, -0.00003421, 0.00002791, -0.00002577, 0.00002515, -0.00002038, 0.00001373, -0.00002911, 0.00002631, -0.00000389, 0.00000263, -0.00001245, 0.00002017, -0.00002516, 0.00001321, -0.00000900, 0.00001966, -0.00000997, 0.00000015, 0.00000936, -0.00002005, 0.00000839, -0.00000284, 0.00000069, 0.00000775, -0.00001505, 0.00001203, -0.00001726, 0.00000595, 0.00000189, -0.00000291, 0.00001060, -0.00000936, 0.00000302, -0.00000506, 0.00000652, -0.00000473, 0.00000846, -0.00000975, 0.00001343, -0.00001769, 0.00001790, -0.00001528, },
		{ -0.00000657, -0.00000176, 0.00000385, 0.00000323, -0.00002359, 0.00003257, -0.00003804, 0.00003630, -0.00002288, 0.00002233, -0.00001609, 0.00003118, -0.00002543, 0.00001235, -0.00000667, -0.00001026, 0.00000886, -0.00000026, -0.00000621, 0.00001334, -0.00001506, 0.00000703, -0.00001221, 0.00000803, 0.00000465, 0.00001060, -0.00002318, 0.00000866, -0.00000555, 0.00000189, -0.00000027, 0.00000514, -0.00000427, 0.00000055, -0.00000380, 0.00001909, -0.00002050, 0.00000873, -0.00000989, 0.00002386, -0.00003382, 0.00001942, -0.00002076, 0.00002658, -0.00002675, 0.00002499, -0.00003106, 0.00003006, -0.00003907, 0.00003008, -0.00002802, 0.00002809, -0.00004651, 0.00003818, -0.00007500, 0.00006862, -0.00013987, 0.00011900, -0.00065939, 0.00124485, -0.00027486, -0.00019191, 0.00005552, -0.00009015, 0.00004234, -0.00005115, 0.00003068, -0.00003961, 0.00003676, -0.00003460, 0.00001827, -0.00002457, 0.00002019, -0.00002454, 0.00002468, -0.00002513, 0.00001834, -0.00000873, 0.00000539, -0.00002144, 0.00002692, -0.00001715, 0.00000787, -0.00002055, 0.00002073, -0.00000832, 0.00001085, -0.00001856, 0.00001716, -0.00001619, 0.00001762, -0.00001535, 0.00000412, -0.00001926, 0.00002930, -0.00001071, 0.00000005, -0.00000229, 0.00000837, -0.00002058, 0.00001637, -0.00001039, 0.00002366, -0.00002317, 0.00001741, -0.00000852, -0.00001096, 0.00001063, -0.00000852, 0.00001094, -0.00000649, -0.00000281, 0.00000438, -0.00001730, 0.00001649, -0.00001207, 0.00001220, -0.00000338, -0.00000125, -0.00000356, 0.00000372, -0.00000272, 0.00000574, -0.00000413, 0.00000439, -0.00000213, -0.00000288, 0.00000628, },
		{ -0.00001108, 0.00000745, -0.00001574, 0.00002642, -0.00003275, 0.00002106, -0.00000988, -0.00000353, 0.00001354, -0.00000836, 0.00000963, -0.00000125, 0.00001722, -0.00002372, 0.00001745, -0.00001539, -0.00000363, 0.00001397, -0.00001398, 0.00001304, -0.00000727, 0.00000052, -0.00000644, 0.00000114, -0.00000162, 0.00002110, -0.00001619, -0.00000231, -0.00000453, 0.00000784, -0.00001290, 0.00001920, -0.00001832, 0.00001961, -0.00002735, 0.00003469, -0.00002440, 0.00001686, -0.00002843, 0.00003785, -0.00003153, 0.00001590, -0.00002607, 0.00003107, -0.00002958, 0.00003072, -0.00003668, 0.00003509, -0.00004107, 0.00003414, -0.00004559, 0.00005795, -0.00007635, 0.00007483, -0.00010997, 0.00010568, -0.00016403, 0.00016091, -0.00039210, 0.00002897, 0.00110064, -0.00075868, 0.00024361, -0.00022222, 0.00012833, -0.00012698, 0.00009314, -0.00009495, 0.00007567, -0.00006412, 0.00005134, -0.00006015, 0.00005139, -0.00005180, 0.00004450, -0.00003911, 0.00003128, -0.00002918, 0.00003565, -0.00004667, 0.00003506, -0.00002077, 0.00002097, -0.00003266, 0.00002275, -0.00001392, 0.00002251, -0.00002460, 0.00001731, -0.00001482, 0.00001392, -0.00001009, 0.00000761, -0.00002261, 0.00001311, 0.00000778, -0.00000092, -0.00001128, 0.00001578, -0.00001687, 0.00000415, -0.00000143, 0.00000796, 0.00000567, -0.00001464, 0.00001844, -0.00001961, 0.00000152, 0.00000552, -0.00000671, 0.00001394, -0.00001746, 0.00001368, -0.00001589, 0.00000290, 0.00000428, -0.00000659, 0.00001299, -0.00000929, 0.00000247, -0.00000496, 0.00000643, -0.00000611, 0.00001041, -0.00001141, 0.00001464, -0.00001643, 0.00001452, },
		{ 0.00001287, -0.00001177, 0.00000471, -0.00000732, 0.00001905, -0.00003639, 0.00004040, -0.00004026, 0.00003178, -0.00001860, 0.00001766, -0.00001287, 0.00002342, -0.00001185, -0.00000315, 0.00000685, -0.00001554, 0.00000536, 0.00000768, -0.00001303, 0.00001837, -0.00001876, 0.00001110, -0.00001502, 0.00000862, -0.00000105, 0.00002098, -0.00002863, 0.00001100, -0.00001050, 0.00001004, -0.00001167, 0.00001800, -0.00001891, 0.00001794, -0.00002351, 0.00003786, -0.00003657, 0.00002558, -0.00003170, 0.00004723, -0.00005276, 0.00003708, -0.00004069, 0.00004722, -0.00004873, 0.00004991, -0.00005759, 0.00005798, -0.00006709, 0.00005945, -0.00006573, 0.00007595, -0.00010041, 0.00010244, -0.00014832, 0.00015746, -0.00024751, 0.00027899, -0.00076842, 0.00096787, 0.00022976, -0.00050933, 0.00022637, -0.00021558, 0.00013701, -0.00013181, 0.00009884, -0.00010021, 0.00008638, -0.00007537, 0.00005699, -0.00006253, 0.00005556, -0.00005736, 0.00005318, -0.00004964, 0.00004111, -0.00003271, 0.00003218, -0.00004840, 0.00004823, -0.00003455, 0.00002674, -0.00003895, 0.00003632, -0.00002407, 0.00002731, -0.00003257, 0.00002895, -0.00002595, 0.00002506, -0.00002093, 0.00001304, -0.00003010, 0.00003241, -0.00000942, 0.00000392, -0.00001087, 0.00001697, -0.00002606, 0.00001959, -0.00001482, 0.00002422, -0.00001640, 0.00000469, 0.00000579, -0.00001998, 0.00001241, -0.00000551, 0.00000568, 0.00000100, -0.00000897, 0.00001141, -0.00002229, 0.00001871, -0.00001340, 0.00001024, -0.00000005, -0.00000328, -0.00000234, 0.00000223, -0.00000126, 0.00000251, 0.00000092, -0.00000211, 0.00000617, -0.00001081, },
		{ 0.00000718, -0.00000360, 0.00000159, -0.00001056, 0.00001743, -0.00001540, -0.00000007, 0.00001220, -0.00002426, 0.00002801, -0.00002122, 0.00002089, -0.00001467, 0.00002817, -0.00002749, 0.00001389, -0.00000625, -0.00001143, 0.00001530, -0.00000858, 0.00000531, 0.00000105, -0.00000739, 0.00000219, -0.00000645, 0.00000131, 0.00001581, -0.00000380, -0.00001295, 0.00000146, 0.00000248, -0.00000650, 0.00001033, -0.00000792, 0.00000765, -0.00001258, 0.00001567, -0.00000305, -0.00000298, -0.00000935, 0.00001366, -0.00000170, -0.00001273, 0.00000086, 0.00000250, 0.00000008, -0.00000031, -0.00000223, -0.00000233, -0.00000059, -0.00000727, -0.00000473, 0.00001116, -0.00001793, 0.00000663, -0.00002566, 0.00000332, -0.00003112, -0.00001674, -0.00007184, -0.00045969, 0.00133989, -0.00060826, 0.00005398, -0.00009342, 0.00002426, -0.00004438, 0.00002064, -0.00003162, 0.00001635, -0.00001263, 0.00000945, -0.00002254, 0.00001486, -0.00001680, 0.00001048, -0.00000803, 0.00000444, -0.00000617, 0.00001401, -0.00002237, 0.00000705, 0.00000474, 0.00000016, -0.00001153, 0.00000086, 0.00000574, 0.00000344, -0.00000406, -0.00000256, 0.00000386, -0.00000322, 0.00000455, -0.00000321, -0.00000987, -0.00000475, 0.00002061, -0.00000574, -0.00000601, 0.00000624, -0.00000457, -0.00000773, 0.00000899, -0.00000532, 0.00001875, -0.00002202, 0.00001706, -0.00000983, -0.00001062, 0.00001487, -0.00001233, 0.00001721, -0.00001694, 0.00001246, -0.00001131, -0.00000296, 0.00000943, -0.00001223, 0.00001568, -0.00000890, 0.00000177, -0.00000495, 0.00000657, -0.00000713, 0.00001069, -0.00001024, 0.00001186, -0.00001069, },
		{ -0.00001422, 0.00001447, -0.00001231, 0.00000693, -0.00001313, 0.00002588, -0.00003703, 0.00003499, -0.00002941, 0.00001629, -0.00000513, 0.00000431, -0.00000056, 0.00000720, 0.00000681, -0.00001871, 0.00001536, -0.00001420, -0.00000228, 0.00001546, -0.00001673, 0.00001962, -0.00001843, 0.00001036, -0.00001276, 0.00000604, -0.00000472, 0.00002694, -0.00002798, 0.00000883, -0.00001246, 0.00001501, -0.00001866, 0.00002521, -0.00002661, 0.00002638, -0.00003206, 0.00004396, -0.00003962, 0.00002954, -0.00003926, 0.00005333, -0.00005279, 0.00003599, -0.00004172, 0.00004861, -0.00005085, 0.00005321, -0.00006020, 0.00005991, -0.00006771, 0.00006116, -0.00007374, 0.00008803, -0.00011096, 0.00011425, -0.00015750, 0.00016613, -0.00024937, 0.00027722, -0.00063240, 0.00046785, 0.00078550, -0.00075047, 0.00031139, -0.00027264, 0.00017499, -0.00016381, 0.00012353, -0.00012115, 0.00010006, -0.00008582, 0.00007005, -0.00007677, 0.00006817, -0.00006828, 0.00006119, -0.00005613, 0.00004821, -0.00004217, 0.00004418, -0.00005909, 0.00005275, -0.00003670, 0.00003159, -0.00004354, 0.00003844, -0.00002635, 0.00002975, -0.00003268, 0.00002753, -0.00002337, 0.00002172, -0.00001804, 0.00001430, -0.00003143, 0.00002600, -0.00000236, 0.00000422, -0.00001437, 0.00001865, -0.00002440, 0.00001606, -0.00001143, 0.00001600, -0.00000291, -0.00000966, 0.00001544, -0.00002043, 0.00000529, 0.00000438, -0.00000407, 0.00001123, -0.00001704, 0.00001977, -0.00002740, 0.00002048, -0.00001335, 0.00000660, 0.00000368, -0.00000465, -0.00000202, 0.00000142, 0.00000010, -0.00000083, 0.00000540, -0.00000706, 0.00001140, },
		{ -0.00000331, -0.00000066, 0.00000364, -0.00000427, -0.00000404, 0.00000610, 0.00000238, -0.00001810, 0.00002815, -0.00003621, 0.00003322, -0.00002518, 0.00002307, -0.00001736, 0.00002634, -0.00001865, 0.00000093, 0.00000610, -0.00001676, 0.00001179, 0.00000052, -0.00000406, 0.00001023, -0.00001591, 0.00001052, -0.00001269, 0.00000450, 0.00000746, 0.00001095, -0.00002389, 0.00000853, -0.00000625, 0.00000495, -0.00000432, 0.00000872, -0.00001102, 0.00000858, -0.00000863, 0.00002234, -0.00002653, 0.00001445, -0.00001575, 0.00003147, -0.00004272, 0.00002894, -0.00002750, 0.00003183, -0.00003445, 0.00003562, -0.00004270, 0.00004223, -0.00005104, 0.00004101, -0.00004391, 0.00004953, -0.00007013, 0.00006600, -0.00010427, 0.00010344, -0.00018813, 0.00019569, -0.00075619, 0.00126921, -0.00020811, -0.00025871, 0.00009824, -0.00012206, 0.00006976, -0.00007681, 0.00005141, -0.00005793, 0.00004965, -0.00004243, 0.00002596, -0.00003230, 0.00002765, -0.00003138, 0.00002949, -0.00002881, 0.00002426, -0.00001600, 0.00001040, -0.00002754, 0.00003441, -0.00002437, 0.00001357, -0.00002454, 0.00002847, -0.00001806, 0.00001747, -0.00002211, 0.00002124, -0.00001814, 0.00001683, -0.00001317, 0.00000335, -0.00002106, 0.00002955, -0.00000861, 0.00000035, -0.00000445, 0.00000720, -0.00001799, 0.00001744, -0.00001460, 0.00002413, -0.00001939, 0.00000677, 0.00000391, -0.00002098, 0.00001856, -0.00001101, 0.00001380, -0.00001042, 0.00000552, -0.00000107, -0.00001342, 0.00001830, -0.00001988, 0.00001859, -0.00000869, 0.00000188, -0.00000604, 0.00000764, -0.00000813, 0.00000974, -0.00000725, 0.00000693, },
		{ 0.00001299, -0.00001308, 0.00001163, -0.00000888, 0.00000512, -0.00001322, 0.00002390, -0.00002745, 0.00002032, -0.00001111, -0.00000369, 0.00001102, -0.00001107, 0.00001326, -0.00000854, 0.00002095, -0.00002588, 0.00001378, -0.00000547, -0.00001189, 0.00002031, -0.00001549, 0.00001606, -0.00001323, 0.00000433, -0.00000572, 0.00000053, -0.00000521, 0.00002645, -0.00001979, 0.00000097, -0.00000925, 0.00001324, -0.00001683, 0.00002186, -0.00002211, 0.00002067, -0.00002488, 0.00003360, -0.00002622, 0.00001697, -0.00002858, 0.00003856, -0.00003145, 0.00001447, -0.00002235, 0.00002899, -0.00003069, 0.00003192, -0.00003594, 0.00003320, -0.00003841, 0.00003179, -0.00004665, 0.00005756, -0.00007143, 0.00006776, -0.00009841, 0.00009318, -0.00014954, 0.00013733, -0.00033111, -0.00010061, 0.00121767, -0.00078982, 0.00023955, -0.00021786, 0.00012524, -0.00012330, 0.00008599, -0.00008788, 0.00006684, -0.00005708, 0.00004888, -0.00005869, 0.00004998, -0.00005024, 0.00004269, -0.00003902, 0.00003374, -0.00003067, 0.00003480, -0.00004750, 0.00003573, -0.00001992, 0.00001904, -0.00003106, 0.00002404, -0.00001260, 0.00001665, -0.00001852, 0.00001340, -0.00000970, 0.00000933, -0.00000795, 0.00000826, -0.00002380, 0.00001188, 0.00000857, 0.00000132, -0.00001232, 0.00001351, -0.00001653, 0.00000710, -0.00000246, 0.00000336, 0.00001076, -0.00001863, 0.00001550, -0.00001120, -0.00000832, 0.00001675, -0.00001347, 0.00001968, -0.00002273, 0.00002496, -0.00002805, 0.00001746, -0.00000825, -0.00000112, 0.00000928, -0.00000673, -0.00000102, -0.00000034, 0.00000261, -0.00000473, 0.00000906, -0.00000986, },
		{ 0.00000159, 0.00000313, -0.00000612, 0.00000791, -0.00000745, 0.00000051, -0.00000281, 0.00001463, -0.00002714, 0.00003311, -0.00003604, 0.00002716, -0.00001892, 0.00001519, -0.00000942, 0.00001411, -0.00000245, -0.00001407, 0.00001457, -0.00001538, 0.00000281, 0.00001152, -0.00001252, 0.00001763, -0.00002206, 0.00001542, -0.00001517, 0.00000619, -0.00000120, 0.00002391, -0.00003101, 0.00001313, -0.00001436, 0.00001615, -0.00001807, 0.00002364, -0.00002702, 0.00002551, -0.00002693, 0.00004034, -0.00004216, 0.00003061, -0.00003669, 0.00005323, -0.00005927, 0.00004355, -0.00004425, 0.00005046, -0.00005526, 0.00005876, -0.00006646, 0.00006674, -0.00007553, 0.00006807, -0.00007963, 0.00009317, -0.00011713, 0.00012052, -0.00016499, 0.00017693, -0.00027404, 0.00031550, -0.00078865, 0.00090037, 0.00034958, -0.00058600, 0.00026566, -0.00024317, 0.00016019, -0.00015217, 0.00011328, -0.00011165, 0.00009248, -0.00007855, 0.00006214, -0.00006819, 0.00006089, -0.00006193, 0.00005615, -0.00005278, 0.00004715, -0.00003949, 0.00003646, -0.00005356, 0.00005401, -0.00003964, 0.00003040, -0.00004156, 0.00004256, -0.00003051, 0.00002924, -0.00003173, 0.00002893, -0.00002384, 0.00002118, -0.00001741, 0.00001180, -0.00003054, 0.00003089, -0.00000718, 0.00000490, -0.00001236, 0.00001449, -0.00002315, 0.00002056, -0.00001690, 0.00002081, -0.00000929, -0.00000631, 0.00001386, -0.00002281, 0.00001231, -0.00000096, 0.00000342, 0.00000181, -0.00000653, 0.00001354, -0.00002637, 0.00002808, -0.00002650, 0.00001939, -0.00000739, 0.00000208, -0.00000752, 0.00000874, -0.00000830, 0.00000742, -0.00000321, },
		{ -0.00001073, 0.00001181, -0.00000921, 0.00000682, -0.00000406, 0.00000165, -0.00001008, 0.00001687, -0.00001349, 0.00000376, 0.00000652, -0.00002017, 0.00002240, -0.00002126, 0.00002109, -0.00001617, 0.00002421, -0.00002128, 0.00000285, 0.00000668, -0.00001888, 0.00001919, -0.00000840, 0.00000797, -0.00000389, -0.00000565, 0.00000416, -0.00000643, -0.00000266, 0.00001952, -0.00000520, -0.00001124, -0.00000122, 0.00000441, -0.00000596, 0.00000826, -0.00000639, 0.00000284, -0.00000534, 0.00001118, -0.00000131, -0.00000703, -0.00000510, 0.00000962, 0.00000303, -0.00001846, 0.00000823, -0.00000252, 0.00000275, -0.00000449, 0.00000460, -0.00001030, 0.00000804, -0.00001572, 0.00000242, -0.00000037, -0.00000075, -0.00001310, -0.00000131, -0.00002242, -0.00000080, -0.00005960, 0.00000207, -0.00056223, 0.00138056, -0.00056223, 0.00000207, -0.00005960, -0.00000080, -0.00002242, -0.00000131, -0.00001310, -0.00000075, -0.00000037, 0.00000242, -0.00001572, 0.00000804, -0.00001030, 0.00000460, -0.00000449, 0.00000275, -0.00000252, 0.00000823, -0.00001846, 0.00000303, 0.00000962, -0.00000510, -0.00000703, -0.00000131, 0.00001118, -0.00000534, 0.00000284, -0.00000639, 0.00000826, -0.00000596, 0.00000441, -0.00000122, -0.00001124, -0.00000520, 0.00001952, -0.00000266, -0.00000643, 0.00000416, -0.00000565, -0.00000389, 0.00000797, -0.00000840, 0.00001919, -0.00001888, 0.00000668, 0.00000285, -0.00002128, 0.00002421, -0.00001617, 0.00002109, -0.00002126, 0.00002240, -0.00002017, 0.00000652, 0.00000376, -0.00001349, 0.00001687, -0.00001008, 0.00000165, -0.00000406, 0.00000682, -0.00000921, 0.00001181, },
		{ 0.00000159, -0.00000321, 0.00000742, -0.00000830, 0.00000874, -0.00000752, 0.00000208, -0.00000739, 0.00001939, -0.00002650, 0.00002808, -0.00002637, 0.00001354, -0.00000653, 0.00000181, 0.00000342, -0.00000096, 0.00001231, -0.00002281, 0.00001386, -0.00000631, -0.00000929, 0.00002081, -0.00001690, 0.00002056, -0.00002315, 0.00001449, -0.00001236, 0.00000490, -0.00000718, 0.00003089, -0.00003054, 0.00001180, -0.00001741, 0.00002118, -0.00002384, 0.00002893, -0.00003173, 0.00002924, -0.00003051, 0.00004256, -0.00004156, 0.00003040, -0.00003964, 0.00005401, -0.00005356, 0.00003646, -0.00003949, 0.00004715, -0.00005278, 0.00005615, -0.00006193, 0.00006089, -0.00006819, 0.00006214, -0.00007855, 0.00009248, -0.00011165, 0.00011328, -0.00015217, 0.00016019, -0.00024317, 0.00026566, -0.00058600, 0.00034958, 0.00090037, -0.00078865, 0.00031550, -0.00027404, 0.00017693, -0.00016499, 0.00012052, -0.00011713, 0.00009317, -0.00007963, 0.00006807, -0.00007553, 0.00006674, -0.00006646, 0.00005876, -0.00005526, 0.00005046, -0.00004425, 0.00004355, -0.00005927, 0.00005323, -0.00003669, 0.00003061, -0.00004216, 0.00004034, -0.00002693, 0.00002551, -0.00002702, 0.00002364, -0.00001807, 0.00001615, -0.00001436, 0.00001313, -0.00003101, 0.00002391, -0.00000120, 0.00000619, -0.00001517, 0.00001542, -0.00002206, 0.00001763, -0.00001252, 0.00001152, 0.00000281, -0.00001538, 0.00001457, -0.00001407, -0.00000245, 0.00001411, -0.00000942, 0.00001519, -0.00001892, 0.00002716, -0.00003604, 0.00003311, -0.00002714, 0.00001463, -0.00000281, 0.00000051, -0.00000745, 0.00000791, -0.00000612, 0.00000313, },
		{ 0.00001299, -0.00000986, 0.00000906, -0.00000473, 0.00000261, -0.00000034, -0.00000102, -0.00000673, 0.00000928, -0.00000112, -0.00000825, 0.00001746, -0.00002805, 0.00002496, -0.00002273, 0.00001968, -0.00001347, 0.00001675, -0.00000832, -0.00001120, 0.00001550, -0.00001863, 0.00001076, 0.00000336, -0.00000246, 0.00000710, -0.00001653, 0.00001351, -0.00001232, 0.00000132, 0.00000857, 0.00001188, -0.00002380, 0.00000826, -0.00000795, 0.00000933, -0.00000970, 0.00001340, -0.00001852, 0.00001665, -0.00001260, 0.00002404, -0.00003106, 0.00001904, -0.00001992, 0.00003573, -0.00004750, 0.00003480, -0.00003067, 0.00003374, -0.00003902, 0.00004269, -0.00005024, 0.00004998, -0.00005869, 0.00004888, -0.00005708, 0.00006684, -0.00008788, 0.00008599, -0.00012330, 0.00012524, -0.00021786, 0.00023955, -0.00078982, 0.00121767, -0.00010061, -0.00033111, 0.00013733, -0.00014954, 0.00009318, -0.00009841, 0.00006776, -0.00007143, 0.00005756, -0.00004665, 0.00003179, -0.00003841, 0.00003320, -0.00003594, 0.00003192, -0.00003069, 0.00002899, -0.00002235, 0.00001447, -0.00003145, 0.00003856, -0.00002858, 0.00001697, -0.00002622, 0.00003360, -0.00002488, 0.00002067, -0.00002211, 0.00002186, -0.00001683, 0.00001324, -0.00000925, 0.00000097, -0.00001979, 0.00002645, -0.00000521, 0.00000053, -0.00000572, 0.00000433, -0.00001323, 0.00001606, -0.00001549, 0.00002031, -0.00001189, -0.00000547, 0.00001378, -0.00002588, 0.00002095, -0.00000854, 0.00001326, -0.00001107, 0.00001102, -0.00000369, -0.00001111, 0.00002032, -0.00002745, 0.00002390, -0.00001322, 0.00000512, -0.00000888, 0.00001163, -0.00001308, },
		{ -0.00000331, 0.00000693, -0.00000725, 0.00000974, -0.00000813, 0.00000764, -0.00000604, 0.00000188, -0.00000869, 0.00001859, -0.00001988, 0.00001830, -0.00001342, -0.00000107, 0.00000552, -0.00001042, 0.00001380, -0.00001101, 0.00001856, -0.00002098, 0.00000391, 0.00000677, -0.00001939, 0.00002413, -0.00001460, 0.00001744, -0.00001799, 0.00000720, -0.00000445, 0.00000035, -0.00000861, 0.00002955, -0.00002106, 0.00000335, -0.00001317, 0.00001683, -0.00001814, 0.00002124, -0.00002211, 0.00001747, -0.00001806, 0.00002847, -0.00002454, 0.00001357, -0.00002437, 0.00003441, -0.00002754, 0.00001040, -0.00001600, 0.00002426, -0.00002881, 0.00002949, -0.00003138, 0.00002765, -0.00003230, 0.00002596, -0.00004243, 0.00004965, -0.00005793, 0.00005141, -0.00007681, 0.00006976, -0.00012206, 0.00009824, -0.00025871, -0.00020811, 0.00126921, -0.00075619, 0.00019569, -0.00018813, 0.00010344, -0.00010427, 0.00006600, -0.00007013, 0.00004953, -0.00004391, 0.00004101, -0.00005104, 0.00004223, -0.00004270, 0.00003562, -0.00003445, 0.00003183, -0.00002750, 0.00002894, -0.00004272, 0.00003147, -0.00001575, 0.00001445, -0.00002653, 0.00002234, -0.00000863, 0.00000858, -0.00001102, 0.00000872, -0.00000432, 0.00000495, -0.00000625, 0.00000853, -0.00002389, 0.00001095, 0.00000746, 0.00000450, -0.00001269, 0.00001052, -0.00001591, 0.00001023, -0.00000406, 0.00000052, 0.00001179, -0.00001676, 0.00000610, 0.00000093, -0.00001865, 0.00002634, -0.00001736, 0.00002307, -0.00002518, 0.00003322, -0.00003621, 0.00002815, -0.00001810, 0.00000238, 0.00000610, -0.00000404, -0.00000427, 0.00000364, -0.00000066, },
		{ -0.00001422, 0.00001140, -0.00000706, 0.00000540, -0.00000083, 0.00000010, 0.00000142, -0.00000202, -0.00000465, 0.00000368, 0.00000660, -0.00001335, 0.00002048, -0.00002740, 0.00001977, -0.00001704, 0.00001123, -0.00000407, 0.00000438, 0.00000529, -0.00002043, 0.00001544, -0.00000966, -0.00000291, 0.00001600, -0.00001143, 0.00001606, -0.00002440, 0.00001865, -0.00001437, 0.00000422, -0.00000236, 0.00002600, -0.00003143, 0.00001430, -0.00001804, 0.00002172, -0.00002337, 0.00002753, -0.00003268, 0.00002975, -0.00002635, 0.00003844, -0.00004354, 0.00003159, -0.00003670, 0.00005275, -0.00005909, 0.00004418, -0.00004217, 0.00004821, -0.00005613, 0.00006119, -0.00006828, 0.00006817, -0.00007677, 0.00007005, -0.00008582, 0.00010006, -0.00012115, 0.00012353, -0.00016381, 0.00017499, -0.00027264, 0.00031139, -0.00075047, 0.00078550, 0.00046785, -0.00063240, 0.00027722, -0.00024937, 0.00016613, -0.00015750, 0.00011425, -0.00011096, 0.00008803, -0.00007374, 0.00006116, -0.00006771, 0.00005991, -0.00006020, 0.00005321, -0.00005085, 0.00004861, -0.00004172, 0.00003599, -0.00005279, 0.00005333, -0.00003926, 0.00002954, -0.00003962, 0.00004396, -0.00003206, 0.00002638, -0.00002661, 0.00002521, -0.00001866, 0.00001501, -0.00001246, 0.00000883, -0.00002798, 0.00002694, -0.00000472, 0.00000604, -0.00001276, 0.00001036, -0.00001843, 0.00001962, -0.00001673, 0.00001546, -0.00000228, -0.00001420, 0.00001536, -0.00001871, 0.00000681, 0.00000720, -0.00000056, 0.00000431, -0.00000513, 0.00001629, -0.00002941, 0.00003499, -0.00003703, 0.00002588, -0.00001313, 0.00000693, -0.00001231, 0.00001447, },
		{ 0.00000718, -0.00001069, 0.00001186, -0.00001024, 0.00001069, -0.00000713, 0.00000657, -0.00000495, 0.00000177, -0.00000890, 0.00001568, -0.00001223, 0.00000943, -0.00000296, -0.00001131, 0.00001246, -0.00001694, 0.00001721, -0.00001233, 0.00001487, -0.00001062, -0.00000983, 0.00001706, -0.00002202, 0.00001875, -0.00000532, 0.00000899, -0.00000773, -0.00000457, 0.00000624, -0.00000601, -0.00000574, 0.00002061, -0.00000475, -0.00000987, -0.00000321, 0.00000455, -0.00000322, 0.00000386, -0.00000256, -0.00000406, 0.00000344, 0.00000574, 0.00000086, -0.00001153, 0.00000016, 0.00000474, 0.00000705, -0.00002237, 0.00001401, -0.00000617, 0.00000444, -0.00000803, 0.00001048, -0.00001680, 0.00001486, -0.00002254, 0.00000945, -0.00001263, 0.00001635, -0.00003162, 0.00002064, -0.00004438, 0.00002426, -0.00009342, 0.00005398, -0.00060826, 0.00133989, -0.00045969, -0.00007184, -0.00001674, -0.00003112, 0.00000332, -0.00002566, 0.00000663, -0.00001793, 0.00001116, -0.00000473, -0.00000727, -0.00000059, -0.00000233, -0.00000223, -0.00000031, 0.00000008, 0.00000250, 0.00000086, -0.00001273, -0.00000170, 0.00001366, -0.00000935, -0.00000298, -0.00000305, 0.00001567, -0.00001258, 0.00000765, -0.00000792, 0.00001033, -0.00000650, 0.00000248, 0.00000146, -0.00001295, -0.00000380, 0.00001581, 0.00000131, -0.00000645, 0.00000219, -0.00000739, 0.00000105, 0.00000531, -0.00000858, 0.00001530, -0.00001143, -0.00000625, 0.00001389, -0.00002749, 0.00002817, -0.00001467, 0.00002089, -0.00002122, 0.00002801, -0.00002426, 0.00001220, -0.00000007, -0.00001540, 0.00001743, -0.00001056, 0.00000159, -0.00000360, },
		{ 0.00001287, -0.00001081, 0.00000617, -0.00000211, 0.00000092, 0.00000251, -0.00000126, 0.00000223, -0.00000234, -0.00000328, -0.00000005, 0.00001024, -0.00001340, 0.00001871, -0.00002229, 0.00001141, -0.00000897, 0.00000100, 0.00000568, -0.00000551, 0.00001241, -0.00001998, 0.00000579, 0.00000469, -0.00001640, 0.00002422, -0.00001482, 0.00001959, -0.00002606, 0.00001697, -0.00001087, 0.00000392, -0.00000942, 0.00003241, -0.00003010, 0.00001304, -0.00002093, 0.00002506, -0.00002595, 0.00002895, -0.00003257, 0.00002731, -0.00002407, 0.00003632, -0.00003895, 0.00002674, -0.00003455, 0.00004823, -0.00004840, 0.00003218, -0.00003271, 0.00004111, -0.00004964, 0.00005318, -0.00005736, 0.00005556, -0.00006253, 0.00005699, -0.00007537, 0.00008638, -0.00010021, 0.00009884, -0.00013181, 0.00013701, -0.00021558, 0.00022637, -0.00050933, 0.00022976, 0.00096787, -0.00076842, 0.00027899, -0.00024751, 0.00015746, -0.00014832, 0.00010244, -0.00010041, 0.00007595, -0.00006573, 0.00005945, -0.00006709, 0.00005798, -0.00005759, 0.00004991, -0.00004873, 0.00004722, -0.00004069, 0.00003708, -0.00005276, 0.00004723, -0.00003170, 0.00002558, -0.00003657, 0.00003786, -0.00002351, 0.00001794, -0.00001891, 0.00001800, -0.00001167, 0.00001004, -0.00001050, 0.00001100, -0.00002863, 0.00002098, -0.00000105, 0.00000862, -0.00001502, 0.00001110, -0.00001876, 0.00001837, -0.00001303, 0.00000768, 0.00000536, -0.00001554, 0.00000685, -0.00000315, -0.00001185, 0.00002342, -0.00001287, 0.00001766, -0.00001860, 0.00003178, -0.00004026, 0.00004040, -0.00003639, 0.00001905, -0.00000732, 0.00000471, -0.00001177, },
		{ -0.00001108, 0.00001452, -0.00001643, 0.00001464, -0.00001141, 0.00001041, -0.00000611, 0.00000643, -0.00000496, 0.00000247, -0.00000929, 0.00001299, -0.00000659, 0.00000428, 0.00000290, -0.00001589, 0.00001368, -0.00001746, 0.00001394, -0.00000671, 0.00000552, 0.00000152, -0.00001961, 0.00001844, -0.00001464, 0.00000567, 0.00000796, -0.00000143, 0.00000415, -0.00001687, 0.00001578, -0.00001128, -0.00000092, 0.00000778, 0.00001311, -0.00002261, 0.00000761, -0.00001009, 0.00001392, -0.00001482, 0.00001731, -0.00002460, 0.00002251, -0.00001392, 0.00002275, -0.00003266, 0.00002097, -0.00002077, 0.00003506, -0.00004667, 0.00003565, -0.00002918, 0.00003128, -0.00003911, 0.00004450, -0.00005180, 0.00005139, -0.00006015, 0.00005134, -0.00006412, 0.00007567, -0.00009495, 0.00009314, -0.00012698, 0.00012833, -0.00022222, 0.00024361, -0.00075868, 0.00110064, 0.00002897, -0.00039210, 0.00016091, -0.00016403, 0.00010568, -0.00010997, 0.00007483, -0.00007635, 0.00005795, -0.00004559, 0.00003414, -0.00004107, 0.00003509, -0.00003668, 0.00003072, -0.00002958, 0.00003107, -0.00002607, 0.00001590, -0.00003153, 0.00003785, -0.00002843, 0.00001686, -0.00002440, 0.00003469, -0.00002735, 0.00001961, -0.00001832, 0.00001920, -0.00001290, 0.00000784, -0.00000453, -0.00000231, -0.00001619, 0.00002110, -0.00000162, 0.00000114, -0.00000644, 0.00000052, -0.00000727, 0.00001304, -0.00001398, 0.00001397, -0.00000363, -0.00001539, 0.00001745, -0.00002372, 0.00001722, -0.00000125, 0.00000963, -0.00000836, 0.00001354, -0.00000353, -0.00000988, 0.00002106, -0.00003275, 0.00002642, -0.00001574, 0.00000745, },
		{ -0.00000657, 0.00000628, -0.00000288, -0.00000213, 0.00000439, -0.00000413, 0.00000574, -0.00000272, 0.00000372, -0.00000356, -0.00000125, -0.00000338, 0.00001220, -0.00001207, 0.00001649, -0.00001730, 0.00000438, -0.00000281, -0.00000649, 0.00001094, -0.00000852, 0.00001063, -0.00001096, -0.00000852, 0.00001741, -0.00002317, 0.00002366, -0.00001039, 0.00001637, -0.00002058, 0.00000837, -0.00000229, 0.00000005, -0.00001071, 0.00002930, -0.00001926, 0.00000412, -0.00001535, 0.00001762, -0.00001619, 0.00001716, -0.00001856, 0.00001085, -0.00000832, 0.00002073, -0.00002055, 0.00000787, -0.00001715, 0.00002692, -0.00002144, 0.00000539, -0.00000873, 0.00001834, -0.00002513, 0.00002468, -0.00002454, 0.00002019, -0.00002457, 0.00001827, -0.00003460, 0.00003676, -0.00003961, 0.00003068, -0.00005115, 0.00004234, -0.00009015, 0.00005552, -0.00019191, -0.00027486, 0.00124485, -0.00065939, 0.00011900, -0.00013987, 0.00006862, -0.00007500, 0.00003818, -0.00004651, 0.00002809, -0.00002802, 0.00003008, -0.00003907, 0.00003006, -0.00003106, 0.00002499, -0.00002675, 0.00002658, -0.00002076, 0.00001942, -0.00003382, 0.00002386, -0.00000989, 0.00000873, -0.00002050, 0.00001909, -0.00000380, 0.00000055, -0.00000427, 0.00000514, -0.00000027, 0.00000189, -0.00000555, 0.00000866, -0.00002318, 0.00001060, 0.00000465, 0.00000803, -0.00001221, 0.00000703, -0.00001506, 0.00001334, -0.00000621, -0.00000026, 0.00000886, -0.00001026, -0.00000667, 0.00001235, -0.00002543, 0.00003118, -0.00001609, 0.00002233, -0.00002288, 0.00003630, -0.00003804, 0.00003257, -0.00002359, 0.00000323, 0.00000385, -0.00000176, },
		{ 0.00000990, -0.00001528, 0.00001790, -0.00001769, 0.00001343, -0.00000975, 0.00000846, -0.00000473, 0.00000652, -0.00000506, 0.00000302, -0.00000936, 0.00001060, -0.00000291, 0.00000189, 0.00000595, -0.00001726, 0.00001203, -0.00001505, 0.00000775, 0.00000069, -0.00000284, 0.00000839, -0.00002005, 0.00000936, 0.00000015, -0.00000997, 0.00001966, -0.00000900, 0.00001321, -0.00002516, 0.00002017, -0.00001245, 0.00000263, -0.00000389, 0.00002631, -0.00002911, 0.00001373, -0.00002038, 0.00002515, -0.00002577, 0.00002791, -0.00003421, 0.00002937, -0.00002115, 0.00003192, -0.00004041, 0.00002813, -0.00003146, 0.00004570, -0.00005234, 0.00003923, -0.00003490, 0.00004082, -0.00005131, 0.00005692, -0.00006279, 0.00006212, -0.00007063, 0.00006474, -0.00008292, 0.00009516, -0.00011150, 0.00011119, -0.00014527, 0.00015271, -0.00024521, 0.00027144, -0.00066631, 0.00064504, 0.00056194, -0.00063410, 0.00025411, -0.00023048, 0.00015224, -0.00014591, 0.00010079, -0.00009804, 0.00007365, -0.00006168, 0.00005422, -0.00006086, 0.00005244, -0.00005216, 0.00004447, -0.00004379, 0.00004503, -0.00003881, 0.00003032, -0.00004587, 0.00004618, -0.00003356, 0.00002432, -0.00003325, 0.00004055, -0.00002901, 0.00001969, -0.00001860, 0.00001938, -0.00001218, 0.00000824, -0.00000748, 0.00000521, -0.00002354, 0.00002182, -0.00000294, 0.00000763, -0.00001246, 0.00000546, -0.00001290, 0.00001764, -0.00001541, 0.00000998, 0.00000227, -0.00001678, 0.00000961, -0.00000914, -0.00000172, 0.00001654, -0.00000419, 0.00000696, -0.00000324, 0.00001797, -0.00002936, 0.00003673, -0.00004229, 0.00002774, -0.00001580, },
		{ -0.00000983, 0.00000125, -0.00000319, 0.00000711, -0.00001107, 0.00001060, -0.00000878, 0.00000875, -0.00000484, 0.00000661, -0.00000621, 0.00000194, -0.00000721, 0.00001427, -0.00001183, 0.00001618, -0.00001440, 0.00000018, 0.00000028, -0.00001011, 0.00001089, -0.00000525, 0.00000312, 0.00000054, -0.00001967, 0.00002148, -0.00001918, 0.00001355, 0.00000056, 0.00000823, -0.00001005, -0.00000439, 0.00000843, -0.00000552, -0.00000724, 0.00001874, -0.00000253, -0.00000880, -0.00000441, 0.00000306, 0.00000102, -0.00000162, 0.00000192, -0.00001099, 0.00001150, 0.00000127, 0.00000190, -0.00001491, 0.00000472, 0.00000080, 0.00000870, -0.00002292, 0.00001676, -0.00000724, 0.00000408, -0.00000960, 0.00001373, -0.00001999, 0.00001799, -0.00002587, 0.00001405, -0.00002217, 0.00002883, -0.00004372, 0.00003451, -0.00005651, 0.00003716, -0.00010965, 0.00007425, -0.00059274, 0.00122489, -0.00031713, -0.00015151, 0.00002314, -0.00005777, 0.00002545, -0.00004589, 0.00002198, -0.00003067, 0.00001870, -0.00000973, 0.00000066, -0.00000871, 0.00000489, -0.00000809, 0.00000294, -0.00000217, 0.00000728, -0.00000618, -0.00000716, -0.00000527, 0.00001523, -0.00001117, -0.00000069, -0.00000303, 0.00001758, -0.00001660, 0.00000913, -0.00000645, 0.00000977, -0.00000503, -0.00000064, 0.00000417, -0.00001447, -0.00000120, 0.00001058, 0.00000505, -0.00000558, 0.00000008, -0.00000916, 0.00000616, 0.00000186, -0.00000683, 0.00000884, -0.00000266, -0.00001753, 0.00001912, -0.00002669, 0.00002559, -0.00000745, 0.00001649, -0.00001639, 0.00002880, -0.00002308, 0.00001371, -0.00000246, -0.00001693, 0.00001649, },
		{ -0.00000910, 0.00000707, -0.00001398, 0.00001528, -0.00001315, 0.00000766, -0.00000501, 0.00000447, -0.00000217, 0.00000538, -0.00000348, 0.00000176, -0.00000769, 0.00000714, 0.00000079, -0.00000066, 0.00000954, -0.00001888, 0.00001091, -0.00001308, 0.00000247, 0.00000529, -0.00000560, 0.00000696, -0.00001146, -0.00000584, 0.00001534, -0.00002088, 0.00002403, -0.00000975, 0.00001619, -0.00002643, 0.00001715, -0.00000803, 0.00000268, -0.00001047, 0.00003055, -0.00002609, 0.00001196, -0.00002209, 0.00002551, -0.00002409, 0.00002476, -0.00002902, 0.00002110, -0.00001391, 0.00002657, -0.00003293, 0.00001967, -0.00002534, 0.00003760, -0.00003901, 0.00002484, -0.00002312, 0.00003200, -0.00004264, 0.00004551, -0.00004790, 0.00004551, -0.00005232, 0.00004700, -0.00006541, 0.00007168, -0.00007978, 0.00007524, -0.00010184, 0.00010312, -0.00017482, 0.00017120, -0.00042029, 0.00012950, 0.00097398, -0.00068774, 0.00020607, -0.00019803, 0.00012092, -0.00011781, 0.00007342, -0.00007514, 0.00005238, -0.00004733, 0.00004641, -0.00005323, 0.00004373, -0.00004355, 0.00003640, -0.00003795, 0.00003952, -0.00003253, 0.00002607, -0.00004113, 0.00003643, -0.00002336, 0.00001799, -0.00002817, 0.00003238, -0.00001778, 0.00000919, -0.00001056, 0.00001266, -0.00000604, 0.00000490, -0.00000747, 0.00000867, -0.00002508, 0.00001792, -0.00000202, 0.00001118, -0.00001397, 0.00000624, -0.00001499, 0.00001854, -0.00001332, 0.00000518, 0.00000430, -0.00001079, -0.00000511, 0.00000822, -0.00001901, 0.00002921, -0.00001210, 0.00001690, -0.00001432, 0.00003179, -0.00003812, 0.00004014, -0.00003928, 0.00001881, },
		{ 0.00002476, -0.00001526, 0.00000793, -0.00001162, 0.00001523, -0.00001723, 0.00001407, -0.00001141, 0.00001057, -0.00000693, 0.00000993, -0.00000909, 0.00000515, -0.00001077, 0.00001620, -0.00001239, 0.00001679, -0.00001204, -0.00000307, 0.00000228, -0.00001187, 0.00000819, 0.00000010, -0.00000400, 0.00000738, -0.00002152, 0.00001409, -0.00000562, -0.00000206, 0.00001311, -0.00000040, 0.00000121, -0.00001661, 0.00001694, -0.00000956, -0.00000217, 0.00000558, 0.00001387, -0.00001979, 0.00000610, -0.00001178, 0.00001729, -0.00001794, 0.00001860, -0.00002737, 0.00002453, -0.00001155, 0.00001789, -0.00003071, 0.00001954, -0.00001772, 0.00002918, -0.00004016, 0.00003154, -0.00002326, 0.00002464, -0.00003454, 0.00004064, -0.00004695, 0.00004609, -0.00005496, 0.00004740, -0.00006315, 0.00007385, -0.00008942, 0.00008596, -0.00011438, 0.00011213, -0.00020114, 0.00021028, -0.00067304, 0.00093906, 0.00015700, -0.00042506, 0.00015964, -0.00015976, 0.00010289, -0.00010802, 0.00007013, -0.00007120, 0.00005028, -0.00003906, 0.00003237, -0.00003931, 0.00003244, -0.00003302, 0.00002549, -0.00002503, 0.00002969, -0.00002614, 0.00001380, -0.00002720, 0.00003202, -0.00002386, 0.00001320, -0.00001895, 0.00003148, -0.00002544, 0.00001483, -0.00001172, 0.00001451, -0.00000783, 0.00000217, -0.00000035, -0.00000536, -0.00001154, 0.00001502, 0.00000105, 0.00000258, -0.00000706, -0.00000327, -0.00000134, 0.00000955, -0.00001147, 0.00000729, 0.00000257, -0.00002010, 0.00001342, -0.00001519, 0.00000911, 0.00000891, 0.00000462, -0.00000354, 0.00001429, -0.00000176, -0.00000850, 0.00001877, -0.00003384, },
		{ 0.00000186, 0.00000225, 0.00000008, -0.00000778, 0.00000746, -0.00000408, -0.00000116, 0.00000167, -0.00000102, 0.00000177, 0.00000217, 0.00000074, -0.00000209, -0.00000361, 0.00000172, 0.00000609, -0.00000541, 0.00001545, -0.00002211, 0.00001135, -0.00001248, -0.00000064, 0.00000584, -0.00000267, -0.00000033, 0.00000051, -0.00001915, 0.00002296, -0.00002136, 0.00001831, -0.00000294, 0.00001285, -0.00002069, 0.00000758, 0.00000077, -0.00000065, -0.00001089, 0.00002546, -0.00001457, 0.00000332, -0.00001555, 0.00001555, -0.00001140, 0.00001063, -0.00001292, 0.00000279, 0.00000201, 0.00001255, -0.00001608, 0.00000168, -0.00000896, 0.00001827, -0.00001521, 0.00000129, -0.00000231, 0.00001276, -0.00002102, 0.00001907, -0.00001728, 0.00001286, -0.00001726, 0.00001073, -0.00002557, 0.00002246, -0.00002127, 0.00001122, -0.00002784, 0.00001804, -0.00006210, 0.00001977, -0.00014346, -0.00029163, 0.00114755, -0.00051315, 0.00002460, -0.00008493, 0.00002990, -0.00004319, 0.00000944, -0.00002311, 0.00000770, -0.00001332, 0.00001893, -0.00002547, 0.00001626, -0.00001808, 0.00001327, -0.00001796, 0.00001976, -0.00001239, 0.00000850, -0.00002322, 0.00001521, -0.00000431, 0.00000360, -0.00001454, 0.00001585, 0.00000008, -0.00000550, 0.00000009, 0.00000384, 0.00000163, 0.00000052, -0.00000581, 0.00000853, -0.00002164, 0.00001054, 0.00000100, 0.00001098, -0.00001057, 0.00000309, -0.00001372, 0.00001582, -0.00000829, 0.00000030, 0.00000344, -0.00000196, -0.00001855, 0.00001863, -0.00002572, 0.00002968, -0.00000891, 0.00001712, -0.00001533, 0.00003374, -0.00003309, 0.00003015, -0.00002461, },
		{ -0.00004083, 0.00002416, -0.00001479, 0.00001023, -0.00001554, 0.00001814, -0.00001799, 0.00001301, -0.00001073, 0.00001006, -0.00000755, 0.00001172, -0.00000997, 0.00000626, -0.00001219, 0.00001621, -0.00001147, 0.00001532, -0.00000705, -0.00000836, 0.00000592, -0.00001439, 0.00000606, 0.00000323, -0.00000593, 0.00000588, -0.00001339, -0.00000137, 0.00001102, -0.00001582, 0.00002090, -0.00000459, 0.00000879, -0.00002387, 0.00001950, -0.00000923, 0.00000112, -0.00000496, 0.00002405, -0.00002370, 0.00001105, -0.00002056, 0.00002530, -0.00002421, 0.00002401, -0.00003115, 0.00002445, -0.00001213, 0.00002186, -0.00003372, 0.00002119, -0.00002219, 0.00003372, -0.00004077, 0.00003039, -0.00002416, 0.00002984, -0.00004205, 0.00004719, -0.00005138, 0.00005011, -0.00005858, 0.00005323, -0.00007178, 0.00007990, -0.00009065, 0.00008704, -0.00011408, 0.00011595, -0.00019938, 0.00020722, -0.00055493, 0.00050286, 0.00061362, -0.00058470, 0.00019708, -0.00018897, 0.00012096, -0.00011986, 0.00007567, -0.00007592, 0.00005244, -0.00004495, 0.00004297, -0.00004892, 0.00003987, -0.00003934, 0.00003141, -0.00003280, 0.00003725, -0.00003149, 0.00002049, -0.00003421, 0.00003421, -0.00002415, 0.00001620, -0.00002378, 0.00003359, -0.00002292, 0.00001116, -0.00000984, 0.00001348, -0.00000633, 0.00000258, -0.00000375, 0.00000202, -0.00001842, 0.00001667, -0.00000231, 0.00000960, -0.00001170, 0.00000060, -0.00000748, 0.00001536, -0.00001381, 0.00000561, 0.00000326, -0.00001394, -0.00000145, 0.00000223, -0.00000959, 0.00002404, -0.00000515, 0.00000805, 0.00000077, 0.00001705, -0.00002484, 0.00003167, },
		{ -0.00000461, -0.00001672, 0.00001319, -0.00000742, 0.00000017, -0.00000228, 0.00000601, -0.00000986, 0.00000788, -0.00000638, 0.00000614, -0.00000246, 0.00000673, -0.00000739, 0.00000177, -0.00000485, 0.00001251, -0.00001171, 0.00002229, -0.00002504, 0.00001125, -0.00001129, -0.00000329, 0.00000447, 0.00000224, -0.00000757, 0.00000840, -0.00002344, 0.00001862, -0.00001075, 0.00000477, 0.00000803, 0.00000632, -0.00001105, -0.00000485, 0.00001039, -0.00000494, -0.00000715, 0.00001448, 0.00000062, -0.00000754, -0.00000517, 0.00000066, 0.00000543, -0.00000645, 0.00000510, -0.00001581, 0.00001676, -0.00000067, 0.00000068, -0.00001601, 0.00000727, -0.00000097, 0.00000715, -0.00001960, 0.00001605, -0.00000548, 0.00000140, -0.00000861, 0.00001354, -0.00001901, 0.00001662, -0.00002479, 0.00001469, -0.00002647, 0.00003363, -0.00004631, 0.00003730, -0.00005607, 0.00003520, -0.00010556, 0.00006204, -0.00052272, 0.00105478, -0.00015862, -0.00021832, 0.00004829, -0.00007280, 0.00003772, -0.00005681, 0.00002885, -0.00003596, 0.00002034, -0.00001119, 0.00000638, -0.00001437, 0.00000953, -0.00001146, 0.00000407, -0.00000285, 0.00001036, -0.00001119, -0.00000336, -0.00000642, 0.00001353, -0.00001000, -0.00000067, -0.00000069, 0.00001629, -0.00001689, 0.00000729, -0.00000244, 0.00000731, -0.00000253, -0.00000384, 0.00000595, -0.00001497, 0.00000150, 0.00000529, 0.00000744, -0.00000345, -0.00000243, -0.00001019, 0.00001030, -0.00000118, -0.00000460, 0.00000210, 0.00000429, -0.00002371, 0.00001637, -0.00001895, 0.00001771, 0.00000390, 0.00000945, -0.00000798, 0.00002596, -0.00001781, 0.00001198, },
		{ 0.00003209, -0.00003582, 0.00001436, -0.00000840, 0.00000746, -0.00001382, 0.00001494, -0.00001306, 0.00000754, -0.00000677, 0.00000681, -0.00000569, 0.00001038, -0.00000719, 0.00000381, -0.00001012, 0.00001276, -0.00000700, 0.00000946, 0.00000236, -0.00001683, 0.00001187, -0.00001826, 0.00000560, 0.00000270, -0.00000212, -0.00000217, -0.00000042, -0.00001695, 0.00002216, -0.00002069, 0.00001948, -0.00000179, 0.00001052, -0.00002413, 0.00001492, -0.00000390, 0.00000098, -0.00000982, 0.00002525, -0.00001892, 0.00000877, -0.00002072, 0.00002240, -0.00001854, 0.00001717, -0.00002232, 0.00001240, -0.00000232, 0.00001559, -0.00002550, 0.00001125, -0.00001438, 0.00002477, -0.00002796, 0.00001675, -0.00001292, 0.00002182, -0.00003361, 0.00003514, -0.00003582, 0.00003316, -0.00003996, 0.00003449, -0.00005132, 0.00005217, -0.00005539, 0.00004852, -0.00006936, 0.00006661, -0.00013063, 0.00011296, -0.00033530, 0.00006320, 0.00091623, -0.00055626, 0.00010957, -0.00013638, 0.00007591, -0.00008085, 0.00004020, -0.00004744, 0.00002776, -0.00002857, 0.00003200, -0.00003679, 0.00002700, -0.00002734, 0.00002099, -0.00002529, 0.00002942, -0.00002189, 0.00001299, -0.00002710, 0.00002355, -0.00001406, 0.00000988, -0.00001886, 0.00002574, -0.00001180, 0.00000148, -0.00000392, 0.00000919, -0.00000244, 0.00000159, -0.00000576, 0.00000648, -0.00002085, 0.00001497, -0.00000365, 0.00001322, -0.00001198, 0.00000124, -0.00001100, 0.00001803, -0.00001325, 0.00000379, 0.00000061, -0.00000350, -0.00001724, 0.00001546, -0.00002068, 0.00002953, -0.00000605, 0.00001238, -0.00000554, 0.00002688, -0.00002954, },
		{ -0.00000578, 0.00001225, -0.00002973, 0.00001850, -0.00001155, 0.00000615, -0.00000995, 0.00001321, -0.00001513, 0.00001114, -0.00000971, 0.00000927, -0.00000667, 0.00001216, -0.00001159, 0.00000600, -0.00001025, 0.00001773, -0.00001665, 0.00002639, -0.00002392, 0.00000726, -0.00000668, -0.00000798, 0.00000424, 0.00000480, -0.00000956, 0.00000758, -0.00001669, 0.00000418, 0.00000576, -0.00000996, 0.00001707, 0.00000143, -0.00000204, -0.00001537, 0.00001656, -0.00000698, -0.00000277, 0.00000290, 0.00001332, -0.00001502, 0.00000342, -0.00001226, 0.00001827, -0.00001782, 0.00001621, -0.00002596, 0.00002227, -0.00000575, 0.00001001, -0.00002566, 0.00001517, -0.00001149, 0.00001932, -0.00002950, 0.00002393, -0.00001440, 0.00001514, -0.00002638, 0.00003211, -0.00003682, 0.00003530, -0.00004427, 0.00003772, -0.00005445, 0.00006200, -0.00007276, 0.00006683, -0.00008895, 0.00008118, -0.00016080, 0.00014989, -0.00055163, 0.00075905, 0.00026097, -0.00041871, 0.00013008, -0.00013599, 0.00008472, -0.00009285, 0.00005453, -0.00005738, 0.00003627, -0.00002844, 0.00002707, -0.00003337, 0.00002566, -0.00002557, 0.00001699, -0.00001761, 0.00002506, -0.00002265, 0.00000850, -0.00001926, 0.00002229, -0.00001612, 0.00000710, -0.00001090, 0.00002497, -0.00002034, 0.00000796, -0.00000414, 0.00000955, -0.00000336, -0.00000221, 0.00000209, -0.00000713, -0.00000712, 0.00000951, 0.00000215, 0.00000490, -0.00000784, -0.00000620, 0.00000348, 0.00000662, -0.00000918, 0.00000192, 0.00000500, -0.00001875, 0.00000296, -0.00000357, 0.00000012, 0.00001891, 0.00000070, 0.00000150, 0.00001507, 0.00000000, },
		{ -0.00002296, 0.00002217, -0.00002245, -0.00000043, 0.00000052, 0.00000182, -0.00000819, 0.00000756, -0.00000469, -0.00000016, -0.00000112, 0.00000177, -0.00000166, 0.00000594, -0.00000099, -0.00000168, -0.00000503, 0.00000613, 0.00000096, -0.00000034, 0.00001473, -0.00002626, 0.00001763, -0.00002119, 0.00000501, 0.00000050, 0.00000399, -0.00001095, 0.00000956, -0.00002450, 0.00002171, -0.00001397, 0.00000905, 0.00000647, 0.00000773, -0.00001871, 0.00000523, 0.00000430, -0.00000161, -0.00000927, 0.00001881, -0.00000810, 0.00000166, -0.00001427, 0.00001138, -0.00000512, 0.00000344, -0.00000726, -0.00000434, 0.00001046, 0.00000601, -0.00001285, -0.00000321, -0.00000183, 0.00001056, -0.00001054, -0.00000059, 0.00000208, 0.00000855, -0.00001751, 0.00001399, -0.00001119, 0.00000729, -0.00001197, 0.00000502, -0.00001766, 0.00001017, -0.00000707, -0.00000245, -0.00001164, 0.00000177, -0.00004335, -0.00000276, -0.00011877, -0.00026055, 0.00099291, -0.00034040, -0.00006801, -0.00003674, -0.00000307, -0.00001661, -0.00001360, -0.00000543, -0.00000734, -0.00000292, 0.00001002, -0.00001288, 0.00000375, -0.00000654, 0.00000291, -0.00001014, 0.00001330, -0.00000451, -0.00000140, -0.00001337, 0.00000766, -0.00000060, 0.00000028, -0.00000981, 0.00001375, 0.00000182, -0.00000851, 0.00000149, 0.00000496, 0.00000152, 0.00000044, -0.00000646, 0.00000767, -0.00001893, 0.00001012, -0.00000233, 0.00001242, -0.00000761, -0.00000127, -0.00001154, 0.00001693, -0.00000940, 0.00000098, -0.00000230, 0.00000495, -0.00002573, 0.00001701, -0.00001901, 0.00002260, 0.00000290, 0.00000887, -0.00000378, 0.00002699, },
		{ 0.00001307, -0.00001599, 0.00002001, -0.00003307, 0.00001589, -0.00001066, 0.00000806, -0.00001307, 0.00001523, -0.00001530, 0.00001031, -0.00000992, 0.00000984, -0.00000860, 0.00001462, -0.00001222, 0.00000679, -0.00001226, 0.00001923, -0.00001712, 0.00002444, -0.00001611, -0.00000234, 0.00000226, -0.00001534, 0.00000632, 0.00000324, -0.00000503, -0.00000104, -0.00000311, -0.00001294, 0.00001949, -0.00001818, 0.00001868, 0.00000189, 0.00000295, -0.00002049, 0.00001659, -0.00000482, -0.00000055, -0.00000496, 0.00001914, -0.00001568, 0.00000660, -0.00001848, 0.00002199, -0.00001881, 0.00001641, -0.00002446, 0.00001649, -0.00000133, 0.00001037, -0.00002535, 0.00001269, -0.00001121, 0.00001959, -0.00002721, 0.00002026, -0.00001243, 0.00001756, -0.00003047, 0.00003428, -0.00003662, 0.00003482, -0.00004326, 0.00003799, -0.00005514, 0.00005804, -0.00006361, 0.00005722, -0.00007757, 0.00007330, -0.00014576, 0.00013327, -0.00043626, 0.00037902, 0.00061383, -0.00048826, 0.00011516, -0.00013324, 0.00007926, -0.00008545, 0.00004462, -0.00004992, 0.00002913, -0.00002719, 0.00003002, -0.00003427, 0.00002478, -0.00002440, 0.00001659, -0.00002008, 0.00002711, -0.00002165, 0.00000871, -0.00002044, 0.00002020, -0.00001351, 0.00000732, -0.00001320, 0.00002506, -0.00001589, 0.00000307, -0.00000239, 0.00000922, -0.00000255, -0.00000092, -0.00000196, -0.00000013, -0.00001342, 0.00001211, -0.00000270, 0.00001149, -0.00001056, -0.00000363, -0.00000273, 0.00001313, -0.00001225, 0.00000265, 0.00000111, -0.00000754, -0.00001401, 0.00001075, -0.00001331, 0.00002738, -0.00000192, 0.00000666, 0.00000777, },
		{ 0.00002184, -0.00001067, 0.00000863, -0.00000783, -0.00001382, 0.00000721, -0.00000325, -0.00000188, -0.00000051, 0.00000354, -0.00000698, 0.00000385, -0.00000324, 0.00000303, 0.00000004, 0.00000650, -0.00000784, 0.00000076, -0.00000149, 0.00000998, -0.00001085, 0.00002581, -0.00003213, 0.00001903, -0.00001969, 0.00000137, -0.00000006, 0.00000770, -0.00001454, 0.00001075, -0.00002034, 0.00000966, 0.00000093, -0.00000512, 0.00001510, 0.00000440, -0.00001131, -0.00000540, 0.00001154, -0.00000393, -0.00000602, 0.00000906, 0.00000352, -0.00000549, -0.00000597, -0.00000173, 0.00000865, -0.00000908, 0.00000553, -0.00001714, 0.00001814, 0.00000041, -0.00000298, -0.00001455, 0.00000740, -0.00000038, 0.00000275, -0.00001316, 0.00001266, -0.00000168, -0.00000295, -0.00000542, 0.00001020, -0.00001425, 0.00001123, -0.00001966, 0.00001117, -0.00002477, 0.00003012, -0.00003932, 0.00002960, -0.00004442, 0.00002046, -0.00008452, 0.00002466, -0.00041487, 0.00085618, -0.00000992, -0.00025650, 0.00005106, -0.00007199, 0.00003721, -0.00005624, 0.00002593, -0.00003326, 0.00001612, -0.00000904, 0.00000913, -0.00001650, 0.00001069, -0.00001177, 0.00000271, -0.00000154, 0.00001104, -0.00001330, -0.00000207, -0.00000495, 0.00000890, -0.00000639, -0.00000245, 0.00000352, 0.00001224, -0.00001414, 0.00000325, 0.00000279, 0.00000426, -0.00000032, -0.00000593, 0.00000591, -0.00001373, 0.00000326, 0.00000116, 0.00000786, -0.00000012, -0.00000541, -0.00000991, 0.00001257, -0.00000280, -0.00000319, -0.00000304, 0.00000729, -0.00002334, 0.00000629, -0.00000713, 0.00000780, 0.00001633, 0.00000245, 0.00000176, },
		{ 0.00000666, 0.00001857, -0.00001655, 0.00001815, -0.00002781, 0.00000737, -0.00000630, 0.00000647, -0.00001179, 0.00001240, -0.00001113, 0.00000621, -0.00000744, 0.00000772, -0.00000754, 0.00001311, -0.00000847, 0.00000360, -0.00001026, 0.00001613, -0.00001206, 0.00001587, -0.00000232, -0.00001568, 0.00001309, -0.00002299, 0.00000889, -0.00000066, 0.00000287, -0.00001191, 0.00000931, -0.00002388, 0.00002302, -0.00001536, 0.00001122, 0.00000789, 0.00000356, -0.00001976, 0.00001078, 0.00000096, -0.00000097, -0.00000764, 0.00001742, -0.00000995, 0.00000447, -0.00001749, 0.00001660, -0.00001067, 0.00000802, -0.00001463, 0.00000372, 0.00000801, 0.00000580, -0.00001879, 0.00000367, -0.00000425, 0.00001254, -0.00001776, 0.00001001, -0.00000407, 0.00001240, -0.00002439, 0.00002424, -0.00002355, 0.00002100, -0.00002792, 0.00002199, -0.00003628, 0.00003221, -0.00003233, 0.00002460, -0.00004086, 0.00003462, -0.00009131, 0.00006183, -0.00026532, 0.00003543, 0.00080436, -0.00039352, 0.00000797, -0.00007622, 0.00003267, -0.00004583, 0.00001004, -0.00002345, 0.00000711, -0.00001319, 0.00001919, -0.00002086, 0.00001111, -0.00001225, 0.00000669, -0.00001335, 0.00001928, -0.00001128, 0.00000066, -0.00001371, 0.00001140, -0.00000605, 0.00000311, -0.00001035, 0.00001961, -0.00000720, -0.00000370, -0.00000004, 0.00000816, -0.00000115, 0.00000013, -0.00000517, 0.00000440, -0.00001607, 0.00001192, -0.00000514, 0.00001402, -0.00000904, -0.00000361, -0.00000680, 0.00001654, -0.00001232, 0.00000265, -0.00000385, 0.00000321, -0.00002549, 0.00001533, -0.00001583, 0.00002463, 0.00000445, 0.00000523, },
		{ 0.00000515, 0.00001749, 0.00000058, -0.00000146, 0.00000169, -0.00002096, 0.00000858, -0.00000541, 0.00000248, -0.00000625, 0.00000873, -0.00001058, 0.00000634, -0.00000659, 0.00000657, -0.00000505, 0.00001245, -0.00001174, 0.00000449, -0.00000727, 0.00001680, -0.00001793, 0.00003104, -0.00003045, 0.00001315, -0.00001190, -0.00000660, 0.00000276, 0.00000652, -0.00001065, 0.00000270, -0.00000728, -0.00000778, 0.00001600, -0.00001541, 0.00001831, 0.00000457, -0.00000573, -0.00001292, 0.00001442, -0.00000347, -0.00000319, 0.00000063, 0.00001103, -0.00000857, -0.00000020, -0.00001129, 0.00001644, -0.00001424, 0.00001031, -0.00002086, 0.00001669, 0.00000202, 0.00000072, -0.00001892, 0.00000924, -0.00000385, 0.00000785, -0.00001721, 0.00001517, -0.00000483, 0.00000487, -0.00001651, 0.00002096, -0.00002375, 0.00002151, -0.00003048, 0.00002445, -0.00004028, 0.00004331, -0.00004928, 0.00004113, -0.00005723, 0.00004335, -0.00011133, 0.00007713, -0.00041640, 0.00058559, 0.00032501, -0.00037061, 0.00007632, -0.00009761, 0.00005551, -0.00006843, 0.00003197, -0.00003868, 0.00001935, -0.00001629, 0.00001982, -0.00002461, 0.00001631, -0.00001609, 0.00000696, -0.00000879, 0.00001835, -0.00001673, 0.00000144, -0.00000967, 0.00001095, -0.00000736, 0.00000042, -0.00000200, 0.00001692, -0.00001400, 0.00000109, 0.00000251, 0.00000593, -0.00000081, -0.00000430, 0.00000216, -0.00000709, -0.00000375, 0.00000532, 0.00000167, 0.00000764, -0.00000874, -0.00000781, 0.00000663, 0.00000475, -0.00000768, -0.00000147, 0.00000359, -0.00001274, -0.00001037, 0.00000648, -0.00000599, 0.00002598, -0.00000003, },
		{ 0.00000037, 0.00000926, 0.00001889, -0.00001071, 0.00001114, -0.00001911, -0.00000219, -0.00000203, 0.00000375, -0.00000821, 0.00000712, -0.00000518, 0.00000110, -0.00000389, 0.00000402, -0.00000426, 0.00000845, -0.00000167, -0.00000195, -0.00000576, 0.00000980, -0.00000312, 0.00000342, 0.00001332, -0.00002795, 0.00002103, -0.00002677, 0.00000854, -0.00000322, 0.00000904, -0.00001837, 0.00001358, -0.00002323, 0.00001426, -0.00000292, -0.00000175, 0.00001570, 0.00000250, -0.00001559, 0.00000215, 0.00000747, -0.00000229, -0.00000663, 0.00001095, -0.00000151, 0.00000022, -0.00001232, 0.00000631, 0.00000097, -0.00000250, -0.00000349, -0.00000864, 0.00001534, 0.00000234, -0.00001179, -0.00000572, 0.00000288, 0.00000506, -0.00000818, 0.00000002, 0.00000422, 0.00000595, -0.00001497, 0.00001015, -0.00000709, 0.00000420, -0.00000935, 0.00000194, -0.00001234, 0.00000209, 0.00000067, -0.00000828, -0.00000431, -0.00000502, -0.00003515, -0.00001131, -0.00011551, -0.00019362, 0.00080530, -0.00016790, -0.00014012, -0.00000625, -0.00002318, -0.00000062, -0.00002666, 0.00000334, -0.00001478, 0.00000173, 0.00000467, -0.00000319, -0.00000529, 0.00000153, -0.00000437, -0.00000472, 0.00000857, 0.00000128, -0.00000849, -0.00000604, 0.00000251, 0.00000060, -0.00000089, -0.00000662, 0.00001308, 0.00000124, -0.00000860, 0.00000050, 0.00000767, 0.00000027, 0.00000075, -0.00000669, 0.00000561, -0.00001478, 0.00000871, -0.00000439, 0.00001187, -0.00000353, -0.00000584, -0.00000829, 0.00001610, -0.00000879, 0.00000062, -0.00000649, 0.00000810, -0.00002626, 0.00000758, -0.00000773, 0.00001295, 0.00001641, },
		{ 0.00000538, 0.00000324, 0.00001729, 0.00000705, -0.00000473, 0.00000398, -0.00002124, 0.00000508, -0.00000492, 0.00000434, -0.00000868, 0.00001004, -0.00001055, 0.00000610, -0.00000773, 0.00000796, -0.00000773, 0.00001488, -0.00001158, 0.00000462, -0.00000951, 0.00001922, -0.00001902, 0.00002822, -0.00002040, 0.00000063, 0.00000053, -0.00001701, 0.00000758, 0.00000163, -0.00000170, -0.00000968, 0.00000707, -0.00002152, 0.00002297, -0.00001597, 0.00001307, 0.00000967, -0.00000350, -0.00001556, 0.00001191, 0.00000032, -0.00000229, -0.00000389, 0.00001235, -0.00000643, 0.00000146, -0.00001484, 0.00001602, -0.00001085, 0.00000683, -0.00001617, 0.00000786, 0.00000865, -0.00000011, -0.00001747, 0.00000486, -0.00000117, 0.00000637, -0.00001461, 0.00001137, -0.00000212, 0.00000627, -0.00001887, 0.00002083, -0.00002144, 0.00001931, -0.00002770, 0.00002208, -0.00003666, 0.00003448, -0.00003632, 0.00002842, -0.00004326, 0.00003334, -0.00009471, 0.00006301, -0.00032657, 0.00028555, 0.00056467, -0.00035874, 0.00002372, -0.00007543, 0.00003656, -0.00005054, 0.00001453, -0.00002595, 0.00000859, -0.00001203, 0.00001816, -0.00001973, 0.00001022, -0.00001043, 0.00000290, -0.00000817, 0.00001693, -0.00001168, -0.00000236, -0.00000753, 0.00000713, -0.00000419, -0.00000021, -0.00000351, 0.00001688, -0.00000985, -0.00000286, 0.00000249, 0.00000739, -0.00000130, -0.00000207, -0.00000205, -0.00000120, -0.00000888, 0.00000820, -0.00000350, 0.00001264, -0.00000901, -0.00000696, 0.00000119, 0.00001087, -0.00001050, 0.00000053, -0.00000264, -0.00000058, -0.00002379, 0.00001268, -0.00001119, 0.00002607, },
		{ 0.00002665, -0.00000180, 0.00001050, 0.00001896, -0.00000401, 0.00000501, -0.00001267, -0.00000829, -0.00000078, 0.00000203, -0.00000464, 0.00000204, -0.00000011, -0.00000283, -0.00000094, 0.00000021, -0.00000032, 0.00000267, 0.00000554, -0.00000711, -0.00000135, 0.00000295, 0.00000637, -0.00000843, 0.00002540, -0.00003394, 0.00002182, -0.00002356, 0.00000292, -0.00000179, 0.00000996, -0.00001679, 0.00000759, -0.00001200, -0.00000222, 0.00001239, -0.00001324, 0.00001977, 0.00000357, -0.00001148, -0.00000544, 0.00001133, -0.00000210, -0.00000467, 0.00000389, 0.00000514, -0.00000238, -0.00000700, -0.00000349, 0.00000981, -0.00000881, 0.00000279, -0.00001484, 0.00001592, 0.00000379, -0.00000821, -0.00001118, 0.00000562, 0.00000179, -0.00000305, -0.00000536, 0.00000819, 0.00000267, -0.00000766, -0.00000114, 0.00000492, -0.00000723, 0.00000346, -0.00001203, 0.00000467, -0.00001821, 0.00002016, -0.00002556, 0.00001510, -0.00002619, -0.00000129, -0.00005432, -0.00002533, -0.00029061, 0.00065673, 0.00010764, -0.00025751, 0.00003043, -0.00005637, 0.00002522, -0.00004568, 0.00001502, -0.00002453, 0.00000794, -0.00000449, 0.00000924, -0.00001518, 0.00000867, -0.00000951, -0.00000051, 0.00000127, 0.00000956, -0.00001267, -0.00000291, -0.00000172, 0.00000274, -0.00000175, -0.00000482, 0.00000841, 0.00000673, -0.00000989, -0.00000127, 0.00000763, 0.00000195, 0.00000053, -0.00000616, 0.00000373, -0.00001056, 0.00000357, -0.00000122, 0.00000640, 0.00000384, -0.00000857, -0.00000816, 0.00001264, -0.00000264, -0.00000316, -0.00000578, 0.00000596, -0.00001740, -0.00000781, 0.00000408, -0.00000012, },
		{ 0.00001697, 0.00001690, -0.00000207, 0.00001992, 0.00000975, -0.00000273, 0.00000169, -0.00001812, 0.00000012, -0.00000416, 0.00000495, -0.00000877, 0.00000866, -0.00000832, 0.00000433, -0.00000732, 0.00000735, -0.00000776, 0.00001358, -0.00000762, 0.00000166, -0.00000853, 0.00001731, -0.00001432, 0.00001871, -0.00000505, -0.00001429, 0.00001295, -0.00002555, 0.00001077, -0.00000317, 0.00000708, -0.00001931, 0.00001477, -0.00002471, 0.00001788, -0.00000625, 0.00000124, 0.00001727, -0.00000311, -0.00001445, 0.00000578, 0.00000560, -0.00000255, -0.00000478, 0.00000881, -0.00000116, 0.00000051, -0.00001353, 0.00000956, -0.00000244, -0.00000048, -0.00000818, -0.00000268, 0.00001495, -0.00000104, -0.00001427, -0.00000144, 0.00000309, 0.00000306, -0.00001016, 0.00000587, 0.00000226, 0.00000492, -0.00001632, 0.00001452, -0.00001301, 0.00001093, -0.00001803, 0.00001149, -0.00002303, 0.00001541, -0.00001458, 0.00000745, -0.00002040, 0.00001138, -0.00006160, 0.00002320, -0.00021414, 0.00004103, 0.00065758, -0.00022436, -0.00007939, -0.00003009, -0.00000007, -0.00001959, -0.00001146, -0.00000757, -0.00000629, -0.00000352, 0.00001002, -0.00000791, -0.00000117, -0.00000091, -0.00000411, -0.00000418, 0.00001108, -0.00000280, -0.00000864, -0.00000332, 0.00000208, -0.00000078, -0.00000130, -0.00000362, 0.00001492, -0.00000477, -0.00000596, 0.00000121, 0.00000913, -0.00000161, -0.00000013, -0.00000511, 0.00000213, -0.00001074, 0.00000846, -0.00000574, 0.00001316, -0.00000538, -0.00000804, -0.00000238, 0.00001378, -0.00001004, 0.00000082, -0.00000727, 0.00000677, -0.00002740, 0.00000729, -0.00000626, },
		{ -0.00000684, 0.00002884, 0.00000324, 0.00000693, 0.00002216, -0.00000017, 0.00000341, -0.00001114, -0.00000964, -0.00000278, 0.00000150, -0.00000199, -0.00000145, 0.00000279, -0.00000470, 0.00000070, -0.00000277, 0.00000293, -0.00000226, 0.00001075, -0.00000957, 0.00000090, -0.00000213, 0.00001338, -0.00001585, 0.00003015, -0.00003105, 0.00001436, -0.00001358, -0.00000718, 0.00000320, 0.00000576, -0.00000839, -0.00000494, 0.00000315, -0.00001780, 0.00002207, -0.00001667, 0.00001620, 0.00000892, -0.00000957, -0.00000948, 0.00001073, 0.00000077, -0.00000368, -0.00000073, 0.00000730, -0.00000144, -0.00000393, -0.00000931, 0.00001229, -0.00000814, 0.00000226, -0.00001371, 0.00000976, 0.00000954, -0.00000789, -0.00001229, 0.00000353, 0.00000304, -0.00000252, -0.00000609, 0.00000770, 0.00000317, -0.00000402, -0.00000707, 0.00000968, -0.00001057, 0.00000769, -0.00001653, 0.00001049, -0.00002411, 0.00002242, -0.00002470, 0.00001539, -0.00002667, 0.00000727, -0.00006347, 0.00000679, -0.00028725, 0.00043724, 0.00034333, -0.00028804, 0.00000924, -0.00005388, 0.00002268, -0.00004111, 0.00000811, -0.00002004, 0.00000360, -0.00000550, 0.00001263, -0.00001503, 0.00000662, -0.00000688, -0.00000237, -0.00000055, 0.00001132, -0.00001017, -0.00000541, -0.00000080, 0.00000059, 0.00000013, -0.00000493, 0.00000589, 0.00000921, -0.00000832, -0.00000403, 0.00000694, 0.00000443, -0.00000061, -0.00000404, 0.00000011, -0.00000543, -0.00000156, 0.00000245, 0.00000027, 0.00001001, -0.00000949, -0.00000808, 0.00000812, 0.00000382, -0.00000679, -0.00000332, -0.00000026, -0.00000499, -0.00002207, 0.00001066, },
		{ 0.00000358, 0.00000473, 0.00002826, -0.00000536, 0.00002100, 0.00001266, 0.00000019, -0.00000025, -0.00001634, -0.00000251, -0.00000549, 0.00000584, -0.00000808, 0.00000652, -0.00000584, 0.00000259, -0.00000642, 0.00000558, -0.00000598, 0.00000980, -0.00000174, -0.00000234, -0.00000615, 0.00001289, -0.00000636, 0.00000646, 0.00001029, -0.00002595, 0.00002021, -0.00002805, 0.00000903, -0.00000436, 0.00001112, -0.00002120, 0.00001200, -0.00001637, 0.00000333, 0.00000856, -0.00001116, 0.00002243, -0.00000130, -0.00001245, -0.00000065, 0.00000940, -0.00000200, -0.00000410, 0.00000371, 0.00000360, -0.00000018, -0.00001041, 0.00000146, 0.00000548, -0.00000590, -0.00000261, -0.00000936, 0.00001623, 0.00000154, -0.00001278, -0.00000577, 0.00000485, 0.00000203, -0.00000783, 0.00000241, 0.00000472, 0.00000444, -0.00001312, 0.00000752, -0.00000491, 0.00000332, -0.00000901, 0.00000125, -0.00000982, -0.00000132, 0.00000200, -0.00000706, -0.00000448, -0.00000431, -0.00003491, -0.00000983, -0.00012546, -0.00010876, 0.00061185, -0.00001985, -0.00017873, 0.00000131, -0.00002806, 0.00000333, -0.00002892, 0.00000304, -0.00001494, 0.00000117, 0.00000284, 0.00000286, -0.00000991, 0.00000533, -0.00000800, -0.00000209, 0.00000603, 0.00000436, -0.00001209, -0.00000183, -0.00000007, -0.00000031, -0.00000049, -0.00000451, 0.00001334, -0.00000095, -0.00000679, -0.00000156, 0.00001066, -0.00000104, 0.00000058, -0.00000595, 0.00000223, -0.00000930, 0.00000612, -0.00000479, 0.00000950, 0.00000099, -0.00001017, -0.00000411, 0.00001332, -0.00000632, -0.00000120, -0.00000828, 0.00000682, -0.00002070, -0.00000669, },
		{ 0.00000672, -0.00000429, 0.00002184, 0.00001501, -0.00000043, 0.00002774, 0.00000119, 0.00000564, -0.00001286, -0.00000870, -0.00000544, 0.00000047, 0.00000040, -0.00000367, 0.00000399, -0.00000516, 0.00000152, -0.00000494, 0.00000500, -0.00000537, 0.00001289, -0.00000869, 0.00000059, -0.00000473, 0.00001667, -0.00001746, 0.00002712, -0.00002060, 0.00000151, -0.00000055, -0.00001796, 0.00000849, -0.00000019, 0.00000196, -0.00001704, 0.00001386, -0.00002461, 0.00002066, -0.00000977, 0.00000538, 0.00001719, -0.00000925, -0.00001012, 0.00000643, 0.00000530, -0.00000362, -0.00000235, 0.00000522, 0.00000212, -0.00000288, -0.00001081, 0.00000887, -0.00000227, -0.00000246, -0.00000861, 0.00000093, 0.00001557, -0.00000769, -0.00001172, -0.00000058, 0.00000583, -0.00000350, -0.00000519, 0.00000552, 0.00000512, -0.00000238, -0.00000909, 0.00000910, -0.00000840, 0.00000619, -0.00001447, 0.00000825, -0.00001988, 0.00001376, -0.00001384, 0.00000593, -0.00001678, 0.00000224, -0.00005360, 0.00000565, -0.00023556, 0.00022492, 0.00047829, -0.00021667, -0.00005949, -0.00002782, 0.00000176, -0.00002227, -0.00000868, -0.00000876, -0.00000561, -0.00000198, 0.00000945, -0.00000771, -0.00000114, -0.00000003, -0.00000723, 0.00000078, 0.00000873, -0.00000365, -0.00001049, 0.00000207, -0.00000267, 0.00000206, -0.00000503, 0.00000399, 0.00001035, -0.00000591, -0.00000591, 0.00000465, 0.00000772, -0.00000210, -0.00000154, -0.00000333, -0.00000161, -0.00000471, 0.00000466, -0.00000396, 0.00001253, -0.00000706, -0.00000929, 0.00000438, 0.00000830, -0.00000813, -0.00000163, -0.00000611, 0.00000409, -0.00002778, },
		{ -0.00002123, 0.00000995, -0.00000311, 0.00003289, -0.00000168, 0.00001683, 0.00001906, 0.00000058, 0.00000156, -0.00001850, -0.00000147, -0.00000914, 0.00000711, -0.00000733, 0.00000475, -0.00000421, 0.00000164, -0.00000560, 0.00000336, -0.00000348, 0.00000524, 0.00000382, -0.00000520, -0.00000430, 0.00000812, 0.00000182, -0.00000432, 0.00002088, -0.00003031, 0.00001957, -0.00002315, 0.00000178, -0.00000113, 0.00000922, -0.00001502, 0.00000084, -0.00000173, -0.00001304, 0.00002031, -0.00001730, 0.00002071, 0.00000443, -0.00001187, -0.00000464, 0.00000955, 0.00000072, -0.00000375, -0.00000001, 0.00000514, 0.00000137, -0.00000794, -0.00000455, 0.00000887, -0.00000598, -0.00000228, -0.00001003, 0.00001152, 0.00000782, -0.00001349, -0.00000712, 0.00000303, 0.00000419, -0.00000838, 0.00000171, 0.00000448, 0.00000592, -0.00001123, 0.00000282, -0.00000058, 0.00000002, -0.00000445, -0.00000406, -0.00000276, -0.00000926, 0.00000730, -0.00000962, -0.00000090, -0.00000750, -0.00002288, -0.00002420, -0.00007424, -0.00017066, 0.00047911, 0.00018175, -0.00022221, -0.00000760, -0.00003176, 0.00000662, -0.00002947, 0.00000019, -0.00001343, -0.00000122, 0.00000050, 0.00000780, -0.00001146, 0.00000470, -0.00000602, -0.00000425, 0.00000445, 0.00000690, -0.00001030, -0.00000477, 0.00000175, -0.00000309, 0.00000220, -0.00000636, 0.00001257, 0.00000132, -0.00000579, -0.00000467, 0.00001087, 0.00000112, -0.00000032, -0.00000459, -0.00000013, -0.00000597, 0.00000258, -0.00000205, 0.00000383, 0.00000743, -0.00001135, -0.00000528, 0.00001089, -0.00000109, -0.00000418, -0.00000656, 0.00000158, -0.00000879, },
		{ -0.00002300, -0.00000623, 0.00000384, 0.00001027, 0.00002820, -0.00000664, 0.00003160, 0.00000366, 0.00000797, -0.00001360, -0.00000961, -0.00000546, -0.00000301, 0.00000359, -0.00000562, 0.00000475, -0.00000555, 0.00000249, -0.00000675, 0.00000604, -0.00000672, 0.00001235, -0.00000548, -0.00000109, -0.00000570, 0.00001687, -0.00001439, 0.00001887, -0.00000685, -0.00001157, 0.00001033, -0.00002489, 0.00001031, -0.00000384, 0.00000899, -0.00002261, 0.00001479, -0.00001980, 0.00000877, 0.00000396, -0.00000782, 0.00002409, -0.00000781, -0.00000955, 0.00000114, 0.00000895, -0.00000296, -0.00000239, 0.00000139, 0.00000553, -0.00000191, -0.00000992, 0.00000276, 0.00000438, -0.00000658, -0.00000447, -0.00000555, 0.00001759, -0.00000443, -0.00001232, -0.00000352, 0.00000675, -0.00000270, -0.00000573, 0.00000449, 0.00000588, -0.00000030, -0.00000998, 0.00000691, -0.00000522, 0.00000381, -0.00001110, 0.00000401, -0.00001316, 0.00000380, -0.00000387, -0.00000171, -0.00000884, -0.00000242, -0.00004220, -0.00000259, -0.00017914, 0.00006781, 0.00049953, -0.00007298, -0.00013732, -0.00000580, -0.00001765, -0.00000547, -0.00002190, -0.00000130, -0.00001165, 0.00000011, 0.00000508, 0.00000077, -0.00000836, 0.00000530, -0.00001026, 0.00000121, 0.00000579, 0.00000244, -0.00001376, 0.00000289, -0.00000357, 0.00000146, -0.00000329, 0.00000118, 0.00001175, -0.00000429, -0.00000590, 0.00000084, 0.00001097, -0.00000279, -0.00000008, -0.00000494, -0.00000051, -0.00000502, 0.00000453, -0.00000512, 0.00001077, -0.00000157, -0.00001163, 0.00000209, 0.00000982, -0.00000636, -0.00000210, -0.00000869, 0.00000617, },
		{ 0.00000130, -0.00002847, 0.00000688, -0.00000267, 0.00002836, 0.00000967, 0.00000766, 0.00002857, -0.00000135, 0.00000672, -0.00002330, 0.00000112, -0.00001279, 0.00000715, -0.00000587, 0.00000312, -0.00000312, 0.00000099, -0.00000450, 0.00000080, -0.00000086, 0.00000107, 0.00000765, -0.00000587, -0.00000376, 0.00000414, 0.00000829, -0.00001110, 0.00002468, -0.00002684, 0.00001210, -0.00001310, -0.00000831, 0.00000423, 0.00000382, -0.00000475, -0.00001237, 0.00001105, -0.00002290, 0.00002247, -0.00001351, 0.00001127, 0.00001359, -0.00001276, -0.00000566, 0.00000613, 0.00000513, -0.00000408, -0.00000135, 0.00000315, 0.00000483, -0.00000659, -0.00000721, 0.00000699, -0.00000121, -0.00000588, -0.00000671, 0.00000370, 0.00001472, -0.00001406, -0.00000726, -0.00000049, 0.00000743, -0.00000965, 0.00000177, 0.00000326, 0.00000801, -0.00000998, 0.00000020, 0.00000041, 0.00000023, -0.00000354, -0.00000504, -0.00000136, -0.00000949, 0.00000394, -0.00000427, -0.00000481, -0.00000328, -0.00002028, -0.00002558, -0.00004994, -0.00017802, 0.00032322, 0.00032069, -0.00018595, -0.00005657, -0.00001546, -0.00000581, -0.00001740, -0.00001152, -0.00000590, -0.00000765, 0.00000168, 0.00000722, -0.00000658, -0.00000125, -0.00000007, -0.00000901, 0.00000535, 0.00000565, -0.00000468, -0.00001023, 0.00000531, -0.00000673, 0.00000470, -0.00000767, 0.00001151, 0.00000318, -0.00000451, -0.00000657, 0.00000888, 0.00000486, -0.00000224, -0.00000218, -0.00000314, -0.00000290, -0.00000017, 0.00000044, -0.00000113, 0.00001122, -0.00000976, -0.00000731, 0.00000843, 0.00000325, -0.00000587, -0.00000463, -0.00000455, },
		{ 0.00000223, -0.00001178, -0.00002106, 0.00000994, 0.00000121, 0.00003614, -0.00000626, 0.00002956, 0.00001073, 0.00000699, -0.00001003, -0.00001495, -0.00000150, -0.00000917, 0.00000753, -0.00000768, 0.00000583, -0.00000658, 0.00000399, -0.00000836, 0.00000628, -0.00000680, 0.00001020, -0.00000160, -0.00000250, -0.00000636, 0.00001538, -0.00000893, 0.00000914, 0.00000545, -0.00002026, 0.00001528, -0.00002533, 0.00000686, -0.00000320, 0.00001019, -0.00001973, 0.00000623, -0.00000682, -0.00000736, 0.00001714, -0.00001666, 0.00002495, -0.00000259, -0.00001011, -0.00000238, 0.00000947, -0.00000029, -0.00000259, -0.00000152, 0.00000633, 0.00000065, -0.00000875, -0.00000256, 0.00000778, -0.00000648, -0.00000442, -0.00000705, 0.00001404, 0.00000250, -0.00001480, -0.00000409, 0.00000459, 0.00000089, -0.00000847, 0.00000532, 0.00000468, 0.00000315, -0.00001130, 0.00000561, -0.00000399, 0.00000368, -0.00000982, 0.00000192, -0.00000931, -0.00000116, -0.00000105, -0.00000172, -0.00000860, -0.00000017, -0.00003785, -0.00000468, -0.00013814, -0.00002465, 0.00043632, 0.00008733, -0.00017972, -0.00001217, -0.00002046, -0.00000221, -0.00002291, -0.00000371, -0.00001031, -0.00000255, 0.00000332, 0.00000567, -0.00001042, 0.00000534, -0.00000854, -0.00000166, 0.00000518, 0.00000511, -0.00001258, -0.00000030, -0.00000083, -0.00000219, 0.00000036, -0.00000263, 0.00001365, -0.00000361, -0.00000448, -0.00000326, 0.00001278, -0.00000161, -0.00000047, -0.00000422, -0.00000201, -0.00000316, 0.00000276, -0.00000388, 0.00000616, 0.00000491, -0.00001355, 0.00000045, 0.00000919, -0.00000261, -0.00000403, -0.00000809, },
		{ -0.00000807, 0.00000496, -0.00002528, -0.00000538, 0.00000378, 0.00001786, 0.00002443, -0.00000206, 0.00003715, -0.00000204, 0.00001161, -0.00002668, 0.00000120, -0.00001313, 0.00000394, -0.00000259, 0.00000070, -0.00000148, -0.00000049, -0.00000237, -0.00000235, 0.00000175, -0.00000233, 0.00000959, -0.00000475, -0.00000395, 0.00000081, 0.00001265, -0.00001363, 0.00002264, -0.00001826, 0.00000174, -0.00000234, -0.00001701, 0.00000802, -0.00000110, 0.00000421, -0.00002098, 0.00001552, -0.00002183, 0.00001382, -0.00000159, -0.00000225, 0.00002258, -0.00001286, -0.00000537, 0.00000134, 0.00000907, -0.00000385, -0.00000131, -0.00000047, 0.00000809, -0.00000514, -0.00000751, 0.00000205, 0.00000497, -0.00000948, -0.00000355, -0.00000280, 0.00001828, -0.00001161, -0.00000867, -0.00000296, 0.00000883, -0.00000886, 0.00000011, 0.00000327, 0.00000872, -0.00000768, -0.00000212, 0.00000049, 0.00000092, -0.00000301, -0.00000512, -0.00000171, -0.00000723, -0.00000113, 0.00000091, -0.00000766, -0.00000057, -0.00001742, -0.00002554, -0.00003497, -0.00016582, 0.00019144, 0.00037300, -0.00008391, -0.00011922, 0.00000079, -0.00001940, -0.00000495, -0.00002175, -0.00000054, -0.00001210, 0.00000222, 0.00000472, 0.00000041, -0.00000776, 0.00000536, -0.00001254, 0.00000566, 0.00000360, 0.00000127, -0.00001452, 0.00000715, -0.00000818, 0.00000469, -0.00000683, 0.00000889, 0.00000588, -0.00000420, -0.00000642, 0.00000489, 0.00000916, -0.00000385, -0.00000037, -0.00000487, -0.00000190, -0.00000079, 0.00000126, -0.00000361, 0.00001102, -0.00000497, -0.00001055, 0.00000700, 0.00000525, -0.00000492, -0.00000439, },
		{ -0.00000658, -0.00000366, -0.00000101, -0.00002972, 0.00000764, -0.00000073, 0.00003524, 0.00000270, 0.00002082, 0.00002262, 0.00000260, -0.00000228, -0.00002369, 0.00000449, -0.00001570, 0.00001046, -0.00000896, 0.00000675, -0.00000772, 0.00000531, -0.00000915, 0.00000552, -0.00000589, 0.00000733, 0.00000170, -0.00000269, -0.00000731, 0.00001314, -0.00000294, 0.00000077, 0.00001336, -0.00002266, 0.00001377, -0.00001999, -0.00000054, 0.00000063, 0.00000671, -0.00001107, -0.00000664, 0.00000686, -0.00001950, 0.00002266, -0.00001642, 0.00001782, 0.00000673, -0.00001230, -0.00000314, 0.00000643, 0.00000423, -0.00000344, -0.00000236, 0.00000398, 0.00000475, -0.00000807, -0.00000532, 0.00000649, -0.00000182, -0.00000804, -0.00000460, 0.00000691, 0.00001067, -0.00001729, -0.00000362, 0.00000075, 0.00000557, -0.00001163, 0.00000641, 0.00000285, 0.00000691, -0.00001261, 0.00000527, -0.00000474, 0.00000557, -0.00001048, 0.00000220, -0.00000894, -0.00000081, -0.00000449, 0.00000386, -0.00001338, 0.00000627, -0.00003823, -0.00000180, -0.00011113, -0.00007044, 0.00033708, 0.00021089, -0.00016013, -0.00005192, -0.00000670, -0.00001200, -0.00001314, -0.00001376, -0.00000386, -0.00000841, 0.00000406, 0.00000610, -0.00000681, 0.00000041, -0.00000290, -0.00000703, 0.00000667, 0.00000433, -0.00000749, -0.00000634, 0.00000395, -0.00000699, 0.00000417, -0.00000608, 0.00001499, -0.00000281, -0.00000306, -0.00000607, 0.00001212, 0.00000168, -0.00000240, -0.00000201, -0.00000462, -0.00000102, 0.00000105, -0.00000205, 0.00000127, 0.00000965, -0.00001312, -0.00000197, 0.00000820, 0.00000090, -0.00000531, },
		{ -0.00000570, -0.00000834, 0.00000220, -0.00001466, -0.00002065, 0.00000949, 0.00000779, 0.00003605, -0.00000645, 0.00003993, 0.00000262, 0.00001246, -0.00002493, -0.00000420, -0.00000845, -0.00000306, 0.00000258, -0.00000288, 0.00000136, -0.00000338, 0.00000102, -0.00000600, 0.00000422, -0.00000498, 0.00001014, -0.00000298, -0.00000371, -0.00000256, 0.00001545, -0.00001308, 0.00001739, -0.00000850, -0.00000717, 0.00000510, -0.00002109, 0.00000772, -0.00000271, 0.00000827, -0.00002163, 0.00001021, -0.00001143, -0.00000105, 0.00001218, -0.00001350, 0.00002661, -0.00000935, -0.00000609, -0.00000211, 0.00001018, -0.00000162, -0.00000146, -0.00000332, 0.00000896, -0.00000227, -0.00000716, -0.00000286, 0.00000870, -0.00000951, -0.00000377, -0.00000504, 0.00001648, -0.00000512, -0.00001224, -0.00000308, 0.00000703, -0.00000495, -0.00000398, 0.00000515, 0.00000734, -0.00000359, -0.00000532, 0.00000162, -0.00000028, -0.00000043, -0.00000695, -0.00000049, -0.00000696, -0.00000247, 0.00000042, -0.00000428, -0.00000433, -0.00000892, -0.00003074, -0.00001873, -0.00015389, 0.00010107, 0.00035241, 0.00004262, -0.00015860, -0.00000438, -0.00002057, -0.00000269, -0.00002238, -0.00000312, -0.00001061, -0.00000085, 0.00000363, 0.00000522, -0.00001045, 0.00000653, -0.00001195, 0.00000307, 0.00000329, 0.00000456, -0.00001486, 0.00000509, -0.00000591, 0.00000140, -0.00000359, 0.00000458, 0.00000957, -0.00000496, -0.00000471, 0.00000017, 0.00001247, -0.00000375, -0.00000034, -0.00000436, -0.00000336, 0.00000061, 0.00000048, -0.00000358, 0.00000755, 0.00000153, -0.00001387, 0.00000619, 0.00000523, -0.00000190, },
		{ -0.00000431, -0.00000632, -0.00000763, 0.00000401, -0.00002811, -0.00000388, 0.00000349, 0.00002713, 0.00001733, 0.00000900, 0.00003561, -0.00000168, 0.00000609, -0.00003173, 0.00000831, -0.00001902, 0.00000996, -0.00000798, 0.00000627, -0.00000765, 0.00000512, -0.00000815, 0.00000327, -0.00000395, 0.00000408, 0.00000415, -0.00000191, -0.00000807, 0.00001012, 0.00000282, -0.00000516, 0.00001660, -0.00001998, 0.00000829, -0.00001216, -0.00000850, 0.00000460, 0.00000200, -0.00000164, -0.00001712, 0.00001419, -0.00002202, 0.00001769, -0.00000723, 0.00000505, 0.00001724, -0.00001432, -0.00000235, 0.00000160, 0.00000873, -0.00000389, -0.00000187, -0.00000000, 0.00000868, -0.00000712, -0.00000586, 0.00000187, 0.00000476, -0.00001213, -0.00000169, 0.00000012, 0.00001635, -0.00001699, -0.00000443, -0.00000223, 0.00000856, -0.00001256, 0.00000551, 0.00000238, 0.00000919, -0.00001242, 0.00000445, -0.00000555, 0.00000710, -0.00001061, 0.00000238, -0.00000924, 0.00000100, -0.00000884, 0.00000866, -0.00001636, 0.00000991, -0.00003603, -0.00000177, -0.00008742, -0.00009494, 0.00024345, 0.00026973, -0.00008278, -0.00010700, 0.00000893, -0.00002402, -0.00000193, -0.00002332, 0.00000119, -0.00001278, 0.00000438, 0.00000441, -0.00000056, -0.00000589, 0.00000308, -0.00001181, 0.00000790, 0.00000233, -0.00000130, -0.00001206, 0.00000767, -0.00001011, 0.00000585, -0.00000750, 0.00001445, -0.00000068, -0.00000286, -0.00000669, 0.00000899, 0.00000623, -0.00000457, 0.00000005, -0.00000636, -0.00000050, 0.00000097, -0.00000128, -0.00000178, 0.00001083, -0.00000940, -0.00000591, 0.00000819, 0.00000243, },
		{ 0.00000115, -0.00000670, -0.00000724, -0.00000340, -0.00000318, -0.00003087, 0.00000799, 0.00000323, 0.00003989, -0.00000209, 0.00003471, 0.00001393, 0.00000836, -0.00001750, -0.00001464, -0.00000021, -0.00001183, 0.00000799, -0.00000666, 0.00000480, -0.00000701, 0.00000477, -0.00000929, 0.00000599, -0.00000672, 0.00000977, -0.00000142, -0.00000243, -0.00000617, 0.00001700, -0.00001068, 0.00001134, -0.00000052, -0.00001215, 0.00000768, -0.00002009, 0.00000347, -0.00000095, 0.00000729, -0.00001552, -0.00000112, 0.00000196, -0.00001447, 0.00002055, -0.00001702, 0.00002280, -0.00000122, -0.00000879, -0.00000279, 0.00000762, 0.00000277, -0.00000244, -0.00000417, 0.00000676, 0.00000211, -0.00000701, -0.00000573, 0.00000805, -0.00000499, -0.00000765, -0.00000328, 0.00001050, 0.00000358, -0.00001650, -0.00000180, 0.00000307, 0.00000069, -0.00000893, 0.00000751, 0.00000502, 0.00000135, -0.00000895, 0.00000384, -0.00000341, 0.00000407, -0.00001047, 0.00000265, -0.00000946, 0.00000062, -0.00000549, 0.00000404, -0.00001264, 0.00000311, -0.00003925, -0.00000180, -0.00014441, 0.00004402, 0.00029448, 0.00014778, -0.00014824, -0.00003869, -0.00000693, -0.00001207, -0.00001315, -0.00001281, -0.00000435, -0.00000676, 0.00000446, 0.00000631, -0.00000805, 0.00000294, -0.00000731, -0.00000222, 0.00000500, 0.00000455, -0.00001107, -0.00000039, -0.00000093, -0.00000376, 0.00000065, -0.00000022, 0.00001321, -0.00000572, -0.00000282, -0.00000364, 0.00001347, -0.00000134, -0.00000215, -0.00000208, -0.00000618, 0.00000254, -0.00000051, -0.00000251, 0.00000302, 0.00000718, -0.00001531, 0.00000458, 0.00000482, },
		{ 0.00000194, -0.00000117, -0.00000763, -0.00000839, 0.00000232, -0.00001814, -0.00001942, 0.00000818, 0.00001719, 0.00003162, 0.00000008, 0.00004425, -0.00000106, 0.00001068, -0.00003473, 0.00000625, -0.00001664, 0.00000482, -0.00000407, 0.00000360, -0.00000538, 0.00000254, -0.00000488, -0.00000054, -0.00000104, 0.00000051, 0.00000607, -0.00000106, -0.00000767, 0.00000588, 0.00000836, -0.00000895, 0.00001662, -0.00001501, 0.00000235, -0.00000545, -0.00001382, 0.00000600, -0.00000080, 0.00000435, -0.00002082, 0.00001222, -0.00001480, 0.00000510, 0.00000595, -0.00000767, 0.00002430, -0.00001328, -0.00000230, -0.00000232, 0.00001076, -0.00000242, -0.00000158, -0.00000337, 0.00001040, -0.00000483, -0.00000549, -0.00000336, 0.00000961, -0.00001318, -0.00000160, -0.00000316, 0.00001708, -0.00001243, -0.00000777, -0.00000274, 0.00000819, -0.00000993, 0.00000168, 0.00000404, 0.00000913, -0.00000988, 0.00000197, -0.00000460, 0.00000611, -0.00000799, 0.00000010, -0.00000735, 0.00000037, -0.00000922, 0.00000760, -0.00001273, 0.00000617, -0.00002657, -0.00000972, -0.00005973, -0.00011455, 0.00017477, 0.00026801, 0.00002174, -0.00014666, 0.00000759, -0.00002598, 0.00000099, -0.00002476, -0.00000062, -0.00001195, 0.00000165, 0.00000343, 0.00000451, -0.00000951, 0.00000581, -0.00001315, 0.00000663, 0.00000150, 0.00000310, -0.00001461, 0.00000791, -0.00000971, 0.00000429, -0.00000625, 0.00001160, 0.00000310, -0.00000409, -0.00000543, 0.00000451, 0.00001045, -0.00000542, 0.00000050, -0.00000595, -0.00000239, 0.00000280, -0.00000193, -0.00000247, 0.00000851, -0.00000324, -0.00001066, 0.00000903, },
		{ 0.00000562, 0.00000227, -0.00000554, -0.00000711, -0.00000772, 0.00000328, -0.00003110, -0.00000250, 0.00000443, 0.00003592, 0.00000982, 0.00002374, 0.00002903, 0.00000235, -0.00000757, -0.00002632, 0.00000742, -0.00001865, 0.00001092, -0.00000879, 0.00000726, -0.00000970, 0.00000720, -0.00001084, 0.00000617, -0.00000708, 0.00000844, -0.00000015, -0.00000050, -0.00000929, 0.00001693, -0.00000691, 0.00000562, 0.00000481, -0.00001339, 0.00000669, -0.00001609, -0.00000234, 0.00000203, 0.00000385, -0.00000691, -0.00001222, 0.00001119, -0.00002008, 0.00001939, -0.00001139, 0.00001217, 0.00000951, -0.00001216, -0.00000147, 0.00000271, 0.00000770, -0.00000334, -0.00000356, 0.00000273, 0.00000655, -0.00000664, -0.00000636, 0.00000367, 0.00000213, -0.00001265, -0.00000035, 0.00000373, 0.00001106, -0.00001876, -0.00000136, -0.00000062, 0.00000532, -0.00001208, 0.00000810, 0.00000367, 0.00000538, -0.00001150, 0.00000572, -0.00000666, 0.00000825, -0.00001338, 0.00000547, -0.00001235, 0.00000487, -0.00001218, 0.00001167, -0.00001931, 0.00001220, -0.00004425, 0.00000894, -0.00013017, 0.00000220, 0.00023443, 0.00020297, -0.00008627, -0.00008999, 0.00001062, -0.00002489, -0.00000135, -0.00002304, 0.00000134, -0.00001175, 0.00000511, 0.00000510, -0.00000256, -0.00000279, -0.00000137, -0.00000778, 0.00000700, 0.00000279, -0.00000525, -0.00000672, 0.00000407, -0.00000823, 0.00000375, -0.00000380, 0.00001535, -0.00000517, -0.00000206, -0.00000546, 0.00001186, 0.00000282, -0.00000470, 0.00000043, -0.00000842, 0.00000305, -0.00000012, -0.00000209, -0.00000029, 0.00000982, -0.00001348, 0.00000104, },
		{ 0.00000932, 0.00000097, 0.00000222, -0.00000887, -0.00000745, -0.00000303, -0.00000602, -0.00003111, 0.00000701, 0.00001055, 0.00004025, -0.00000126, 0.00004485, 0.00000715, 0.00000944, -0.00003106, -0.00000231, -0.00000913, -0.00000372, 0.00000163, -0.00000067, -0.00000124, -0.00000197, -0.00000008, -0.00000513, 0.00000223, -0.00000304, 0.00000764, -0.00000083, -0.00000563, 0.00000046, 0.00001347, -0.00001103, 0.00001503, -0.00001016, -0.00000170, -0.00000172, -0.00001537, 0.00000417, -0.00000072, 0.00000564, -0.00001753, 0.00000321, -0.00000277, -0.00000840, 0.00001616, -0.00001449, 0.00002424, -0.00000748, -0.00000444, -0.00000350, 0.00000903, 0.00000147, -0.00000227, -0.00000482, 0.00000914, -0.00000103, -0.00000511, -0.00000702, 0.00001037, -0.00000964, -0.00000521, -0.00000243, 0.00001318, -0.00000466, -0.00001271, -0.00000128, 0.00000500, -0.00000467, -0.00000382, 0.00000674, 0.00000752, -0.00000558, -0.00000196, -0.00000167, 0.00000233, -0.00000256, -0.00000475, -0.00000275, -0.00000373, -0.00000460, 0.00000046, -0.00000316, -0.00000363, -0.00001153, -0.00002360, -0.00002958, -0.00013151, 0.00012805, 0.00023211, 0.00011407, -0.00014523, -0.00002023, -0.00001379, -0.00000722, -0.00001658, -0.00000952, -0.00000631, -0.00000390, 0.00000406, 0.00000649, -0.00000864, 0.00000413, -0.00001046, 0.00000255, 0.00000258, 0.00000453, -0.00001309, 0.00000440, -0.00000608, 0.00000024, -0.00000318, 0.00000725, 0.00000771, -0.00000583, -0.00000353, 0.00000028, 0.00001287, -0.00000405, -0.00000105, -0.00000351, -0.00000591, 0.00000533, -0.00000294, -0.00000184, 0.00000451, 0.00000306, -0.00001431, },
		{ -0.00000420, 0.00000787, 0.00000107, -0.00000200, -0.00000865, -0.00000891, 0.00000286, -0.00002221, -0.00001777, 0.00000693, 0.00002809, 0.00002441, 0.00001244, 0.00004245, -0.00000045, 0.00000008, -0.00003442, 0.00001011, -0.00002029, 0.00000937, -0.00000791, 0.00000733, -0.00000993, 0.00000689, -0.00000965, 0.00000414, -0.00000571, 0.00000596, 0.00000130, 0.00000105, -0.00001071, 0.00001454, -0.00000173, 0.00000033, 0.00000819, -0.00001274, 0.00000469, -0.00001201, -0.00000695, 0.00000374, 0.00000098, -0.00000014, -0.00001807, 0.00001219, -0.00001622, 0.00000988, -0.00000010, -0.00000054, 0.00001854, -0.00001342, -0.00000034, -0.00000190, 0.00001068, -0.00000257, -0.00000299, -0.00000103, 0.00000920, -0.00000520, -0.00000566, -0.00000214, 0.00000848, -0.00001533, 0.00000051, -0.00000055, 0.00001451, -0.00001707, -0.00000340, -0.00000209, 0.00000687, -0.00001165, 0.00000563, 0.00000438, 0.00000731, -0.00001176, 0.00000578, -0.00000796, 0.00000974, -0.00001325, 0.00000545, -0.00001263, 0.00000632, -0.00001472, 0.00001338, -0.00001918, 0.00001326, -0.00004047, 0.00000820, -0.00010549, -0.00003747, 0.00019118, 0.00020657, 0.00000466, -0.00013251, 0.00001532, -0.00002994, 0.00000395, -0.00002653, 0.00000118, -0.00001224, 0.00000323, 0.00000399, 0.00000270, -0.00000702, 0.00000251, -0.00001085, 0.00000719, 0.00000140, -0.00000009, -0.00001107, 0.00000656, -0.00000995, 0.00000423, -0.00000511, 0.00001515, -0.00000268, -0.00000287, -0.00000530, 0.00000845, 0.00000731, -0.00000639, 0.00000156, -0.00000855, 0.00000104, 0.00000225, -0.00000301, -0.00000141, 0.00000893, -0.00000858, },
		{ -0.00001508, 0.00000753, 0.00000139, 0.00000368, -0.00000801, -0.00000720, -0.00000782, 0.00000202, -0.00003335, -0.00000244, 0.00000844, 0.00004162, 0.00000569, 0.00003783, 0.00002153, 0.00000434, -0.00002295, -0.00001439, -0.00000004, -0.00001228, 0.00000658, -0.00000475, 0.00000319, -0.00000672, 0.00000448, -0.00000894, 0.00000468, -0.00000576, 0.00000855, -0.00000109, -0.00000247, -0.00000515, 0.00001731, -0.00001130, 0.00001253, -0.00000624, -0.00000368, -0.00000040, -0.00001438, 0.00000061, 0.00000096, 0.00000383, -0.00001051, -0.00000742, 0.00000724, -0.00001618, 0.00001839, -0.00001270, 0.00001682, 0.00000201, -0.00000808, -0.00000218, 0.00000440, 0.00000648, -0.00000322, -0.00000469, 0.00000587, 0.00000311, -0.00000462, -0.00000837, 0.00000699, -0.00000272, -0.00001067, 0.00000018, 0.00000736, 0.00000339, -0.00001685, 0.00000019, 0.00000114, 0.00000069, -0.00000847, 0.00000836, 0.00000612, -0.00000117, -0.00000592, 0.00000192, -0.00000256, 0.00000364, -0.00001003, 0.00000250, -0.00000912, 0.00000213, -0.00000863, 0.00000726, -0.00001373, 0.00000280, -0.00003633, -0.00000459, -0.00013916, 0.00008929, 0.00019220, 0.00016595, -0.00009645, -0.00006711, 0.00000486, -0.00002064, -0.00000430, -0.00002029, -0.00000018, -0.00000925, 0.00000484, 0.00000602, -0.00000439, -0.00000024, -0.00000570, -0.00000256, 0.00000443, 0.00000379, -0.00000879, -0.00000102, -0.00000131, -0.00000424, -0.00000009, 0.00000292, 0.00001173, -0.00000676, -0.00000232, -0.00000251, 0.00001291, -0.00000068, -0.00000378, -0.00000040, -0.00000921, 0.00000665, -0.00000270, -0.00000159, 0.00000109, 0.00000715, },
		{ -0.00000244, -0.00000951, 0.00001027, -0.00000104, 0.00000224, -0.00001047, -0.00000803, -0.00000203, -0.00000983, -0.00003037, 0.00000495, 0.00002100, 0.00003636, 0.00000617, 0.00004931, 0.00000393, 0.00000227, -0.00003614, 0.00000600, -0.00001612, 0.00000354, -0.00000424, 0.00000484, -0.00000736, 0.00000368, -0.00000590, 0.00000026, -0.00000288, 0.00000246, 0.00000321, 0.00000140, -0.00000972, 0.00000968, 0.00000463, -0.00000460, 0.00001057, -0.00001192, 0.00000358, -0.00000955, -0.00000908, 0.00000315, 0.00000017, 0.00000271, -0.00001730, 0.00000575, -0.00000634, -0.00000249, 0.00001059, -0.00000949, 0.00002176, -0.00001031, -0.00000126, -0.00000401, 0.00001001, 0.00000071, -0.00000339, -0.00000312, 0.00000918, -0.00000248, -0.00000452, -0.00000710, 0.00001136, -0.00001368, -0.00000214, -0.00000119, 0.00001341, -0.00001164, -0.00000771, -0.00000118, 0.00000519, -0.00000796, 0.00000081, 0.00000654, 0.00000741, -0.00000979, 0.00000372, -0.00000660, 0.00000771, -0.00000945, 0.00000191, -0.00000925, 0.00000340, -0.00001141, 0.00000804, -0.00001182, 0.00000638, -0.00002833, -0.00000295, -0.00007219, -0.00007645, 0.00016414, 0.00017856, 0.00009046, -0.00014146, -0.00000378, -0.00002158, -0.00000140, -0.00002063, -0.00000604, -0.00000790, -0.00000142, 0.00000413, 0.00000558, -0.00000757, 0.00000268, -0.00001036, 0.00000487, 0.00000143, 0.00000286, -0.00001200, 0.00000564, -0.00000856, 0.00000220, -0.00000438, 0.00001289, 0.00000133, -0.00000466, -0.00000410, 0.00000462, 0.00001065, -0.00000613, 0.00000061, -0.00000633, -0.00000304, 0.00000569, -0.00000450, -0.00000103, 0.00000579, },
		{ 0.00000793, -0.00001247, 0.00000325, 0.00000395, 0.00000237, -0.00000440, -0.00000876, -0.00000964, 0.00000324, -0.00002604, -0.00001693, 0.00000765, 0.00003797, 0.00001777, 0.00002727, 0.00003734, 0.00000019, -0.00001498, -0.00002522, 0.00000612, -0.00001727, 0.00000831, -0.00000677, 0.00000605, -0.00000973, 0.00000690, -0.00001045, 0.00000520, -0.00000681, 0.00000819, -0.00000110, 0.00000051, -0.00000935, 0.00001862, -0.00000921, 0.00000897, -0.00000270, -0.00000492, 0.00000041, -0.00001308, -0.00000246, 0.00000222, 0.00000157, -0.00000389, -0.00001443, 0.00001046, -0.00001537, 0.00001226, -0.00000423, 0.00000559, 0.00001151, -0.00001073, -0.00000030, -0.00000069, 0.00001016, -0.00000280, -0.00000440, 0.00000249, 0.00000594, -0.00000344, -0.00000788, 0.00000120, 0.00000464, -0.00001486, 0.00000167, 0.00000272, 0.00000881, -0.00001800, -0.00000022, -0.00000121, 0.00000376, -0.00001011, 0.00000729, 0.00000624, 0.00000199, -0.00000846, 0.00000445, -0.00000629, 0.00000796, -0.00001304, 0.00000559, -0.00001258, 0.00000681, -0.00001454, 0.00001295, -0.00001839, 0.00001045, -0.00004133, 0.00000828, -0.00013115, 0.00004690, 0.00016670, 0.00017159, -0.00001647, -0.00011176, 0.00001524, -0.00002910, 0.00000368, -0.00002603, 0.00000134, -0.00001100, 0.00000371, 0.00000505, 0.00000051, -0.00000422, -0.00000181, -0.00000627, 0.00000529, 0.00000258, -0.00000394, -0.00000584, 0.00000224, -0.00000712, 0.00000151, -0.00000016, 0.00001407, -0.00000600, -0.00000244, -0.00000357, 0.00001104, 0.00000352, -0.00000616, 0.00000170, -0.00001053, 0.00000533, -0.00000031, -0.00000267, -0.00000041, },
		{ 0.00000256, 0.00000254, -0.00001282, 0.00001096, -0.00000193, 0.00000476, -0.00001038, -0.00000748, -0.00000720, -0.00000040, -0.00003432, -0.00000421, 0.00001634, 0.00004294, 0.00000735, 0.00004813, 0.00001580, -0.00000045, -0.00003240, -0.00000316, -0.00000862, -0.00000405, 0.00000030, 0.00000109, -0.00000310, -0.00000116, -0.00000101, -0.00000416, 0.00000035, -0.00000127, 0.00000526, 0.00000062, -0.00000671, 0.00000329, 0.00001108, -0.00000858, 0.00001213, -0.00001137, 0.00000348, -0.00000841, -0.00000945, 0.00000116, 0.00000088, 0.00000233, -0.00001209, -0.00000367, 0.00000340, -0.00001127, 0.00001521, -0.00001101, 0.00001777, -0.00000299, -0.00000421, -0.00000327, 0.00000599, 0.00000559, -0.00000426, -0.00000372, 0.00000722, 0.00000059, -0.00000312, -0.00000999, 0.00000997, -0.00000810, -0.00000722, 0.00000051, 0.00000957, -0.00000464, -0.00001247, 0.00000064, 0.00000193, -0.00000305, -0.00000419, 0.00000834, 0.00000706, -0.00000676, 0.00000048, -0.00000341, 0.00000332, -0.00000346, -0.00000356, -0.00000371, -0.00000236, -0.00000438, -0.00000111, -0.00000134, -0.00000369, -0.00001319, -0.00001788, -0.00003842, -0.00010741, 0.00014116, 0.00014566, 0.00014324, -0.00010699, -0.00004347, -0.00000447, -0.00001367, -0.00000912, -0.00001642, -0.00000209, -0.00000653, 0.00000468, 0.00000609, -0.00000480, 0.00000005, -0.00000751, 0.00000116, 0.00000236, 0.00000370, -0.00000997, 0.00000228, -0.00000536, -0.00000095, -0.00000279, 0.00000961, 0.00000574, -0.00000629, -0.00000303, 0.00000156, 0.00001209, -0.00000383, -0.00000189, -0.00000287, -0.00000748, 0.00000831, -0.00000496, -0.00000073, },
		{ -0.00000029, 0.00000614, -0.00000770, -0.00000233, 0.00000765, -0.00000059, 0.00000054, -0.00001105, -0.00000905, -0.00000069, -0.00001407, -0.00002961, 0.00000330, 0.00003273, 0.00003048, 0.00001841, 0.00004926, 0.00000170, -0.00001108, -0.00003121, 0.00000654, -0.00001707, 0.00000607, -0.00000611, 0.00000642, -0.00000993, 0.00000630, -0.00000912, 0.00000344, -0.00000589, 0.00000627, -0.00000023, 0.00000218, -0.00001096, 0.00001671, -0.00000459, 0.00000418, 0.00000120, -0.00000676, 0.00000219, -0.00001291, -0.00000385, 0.00000196, 0.00000057, -0.00000028, -0.00001554, 0.00000641, -0.00000799, 0.00000191, 0.00000562, -0.00000398, 0.00001679, -0.00000984, 0.00000000, -0.00000375, 0.00001047, 0.00000010, -0.00000500, 0.00000040, 0.00000663, -0.00000148, -0.00000623, -0.00000482, 0.00000960, -0.00001542, 0.00000024, 0.00000091, 0.00001043, -0.00001561, -0.00000302, -0.00000123, 0.00000372, -0.00000836, 0.00000363, 0.00000780, 0.00000366, -0.00000909, 0.00000504, -0.00000751, 0.00000891, -0.00001236, 0.00000505, -0.00001223, 0.00000699, -0.00001455, 0.00001157, -0.00001570, 0.00000971, -0.00003684, 0.00000774, -0.00010783, -0.00000081, 0.00015681, 0.00014590, 0.00006535, -0.00013064, 0.00000551, -0.00002581, 0.00000217, -0.00002311, -0.00000376, -0.00000820, 0.00000005, 0.00000487, 0.00000384, -0.00000551, -0.00000065, -0.00000724, 0.00000429, 0.00000200, -0.00000022, -0.00000837, 0.00000328, -0.00000761, 0.00000122, -0.00000169, 0.00001448, -0.00000354, -0.00000366, -0.00000349, 0.00000832, 0.00000720, -0.00000697, 0.00000177, -0.00000925, 0.00000144, 0.00000365, -0.00000461, },
		{ -0.00000170, 0.00000081, 0.00000484, -0.00001300, 0.00000886, -0.00000028, 0.00000420, -0.00000729, -0.00000875, -0.00000979, 0.00000260, -0.00002870, -0.00001801, 0.00001194, 0.00004467, 0.00001484, 0.00004115, 0.00003180, -0.00000421, -0.00002701, -0.00001330, -0.00000187, -0.00001005, 0.00000336, -0.00000212, 0.00000101, -0.00000561, 0.00000301, -0.00000742, 0.00000261, -0.00000411, 0.00000662, -0.00000045, -0.00000303, -0.00000288, 0.00001597, -0.00001051, 0.00001220, -0.00001029, 0.00000309, -0.00000697, -0.00000987, -0.00000050, 0.00000157, 0.00000083, -0.00000617, -0.00001098, 0.00000791, -0.00001274, 0.00001211, -0.00000558, 0.00000890, 0.00000570, -0.00000719, -0.00000125, 0.00000080, 0.00000972, -0.00000398, -0.00000415, 0.00000501, 0.00000265, -0.00000126, -0.00001069, 0.00000535, -0.00000089, -0.00001211, 0.00000204, 0.00000566, 0.00000133, -0.00001557, 0.00000153, -0.00000080, 0.00000067, -0.00000710, 0.00000807, 0.00000757, -0.00000399, -0.00000256, -0.00000002, -0.00000128, 0.00000217, -0.00000832, 0.00000122, -0.00000777, 0.00000229, -0.00000903, 0.00000674, -0.00001111, -0.00000140, -0.00002909, -0.00001300, -0.00012216, 0.00011046, 0.00012653, 0.00015351, -0.00003933, -0.00008694, 0.00000904, -0.00002409, 0.00000055, -0.00002368, 0.00000051, -0.00000900, 0.00000387, 0.00000566, -0.00000080, -0.00000294, -0.00000467, -0.00000195, 0.00000287, 0.00000353, -0.00000655, -0.00000158, -0.00000226, -0.00000358, -0.00000160, 0.00000640, 0.00000947, -0.00000682, -0.00000281, -0.00000031, 0.00001179, -0.00000038, -0.00000459, 0.00000011, -0.00001034, 0.00000840, -0.00000320, },
		{ -0.00000588, 0.00000004, 0.00000363, -0.00000294, -0.00000718, 0.00001051, -0.00000292, 0.00000428, -0.00001180, -0.00000835, -0.00000580, -0.00000387, -0.00003445, -0.00000706, 0.00002736, 0.00004071, 0.00001472, 0.00005424, 0.00001081, -0.00001205, -0.00003201, 0.00000168, -0.00001296, 0.00000129, -0.00000377, 0.00000475, -0.00000766, 0.00000322, -0.00000572, 0.00000017, -0.00000361, 0.00000321, 0.00000147, 0.00000224, -0.00000988, 0.00001202, 0.00000179, -0.00000124, 0.00000538, -0.00000932, 0.00000485, -0.00001353, -0.00000411, 0.00000070, 0.00000082, 0.00000027, -0.00001195, -0.00000151, 0.00000073, -0.00000675, 0.00001136, -0.00000766, 0.00001558, -0.00000475, -0.00000186, -0.00000387, 0.00000719, 0.00000491, -0.00000613, -0.00000050, 0.00000577, 0.00000048, -0.00000377, -0.00000950, 0.00001074, -0.00001201, -0.00000375, 0.00000139, 0.00000917, -0.00001093, -0.00000713, 0.00000030, 0.00000149, -0.00000461, -0.00000097, 0.00000941, 0.00000469, -0.00000838, 0.00000396, -0.00000636, 0.00000683, -0.00000863, 0.00000156, -0.00000864, 0.00000321, -0.00000969, 0.00000506, -0.00000824, 0.00000359, -0.00002629, -0.00000154, -0.00007638, -0.00004655, 0.00015175, 0.00011281, 0.00012173, -0.00011106, -0.00002484, -0.00001260, -0.00000723, -0.00001362, -0.00001292, -0.00000325, -0.00000447, 0.00000509, 0.00000513, -0.00000387, -0.00000196, -0.00000607, 0.00000206, 0.00000196, 0.00000198, -0.00000841, 0.00000208, -0.00000633, -0.00000021, -0.00000224, 0.00001344, -0.00000004, -0.00000518, -0.00000319, 0.00000577, 0.00000954, -0.00000590, 0.00000004, -0.00000616, -0.00000356, 0.00000741, },
		{ 0.00000058, -0.00000377, 0.00000064, 0.00000462, -0.00001057, 0.00000451, 0.00000345, 0.00000123, -0.00000221, -0.00001112, -0.00000983, 0.00000009, -0.00001769, -0.00003020, 0.00000396, 0.00004342, 0.00002564, 0.00003254, 0.00004699, -0.00000420, -0.00002402, -0.00002063, 0.00000098, -0.00001236, 0.00000364, -0.00000365, 0.00000356, -0.00000811, 0.00000476, -0.00000843, 0.00000302, -0.00000526, 0.00000656, -0.00000086, -0.00000009, -0.00000717, 0.00001794, -0.00000951, 0.00001009, -0.00000769, 0.00000108, -0.00000389, -0.00001157, -0.00000079, 0.00000124, -0.00000003, -0.00000236, -0.00001327, 0.00000570, -0.00000765, 0.00000407, 0.00000264, -0.00000007, 0.00001169, -0.00000760, -0.00000014, -0.00000297, 0.00001081, -0.00000123, -0.00000549, 0.00000382, 0.00000301, 0.00000091, -0.00000950, -0.00000071, 0.00000527, -0.00001442, 0.00000152, 0.00000345, 0.00000477, -0.00001599, 0.00000052, -0.00000172, 0.00000187, -0.00000698, 0.00000536, 0.00000919, -0.00000199, -0.00000461, 0.00000233, -0.00000437, 0.00000549, -0.00001039, 0.00000346, -0.00001032, 0.00000548, -0.00001225, 0.00000912, -0.00001275, 0.00000374, -0.00003263, -0.00000025, -0.00011809, 0.00006857, 0.00012476, 0.00013132, 0.00003703, -0.00011267, 0.00000702, -0.00002534, 0.00000261, -0.00002356, -0.00000271, -0.00000746, 0.00000093, 0.00000555, 0.00000237, -0.00000412, -0.00000366, -0.00000316, 0.00000219, 0.00000324, -0.00000319, -0.00000434, -0.00000059, -0.00000478, -0.00000135, 0.00000378, 0.00001197, -0.00000595, -0.00000346, -0.00000131, 0.00001056, 0.00000305, -0.00000622, 0.00000137, -0.00001059, 0.00000556, },
		{ 0.00000904, -0.00000509, -0.00000072, 0.00000206, 0.00000010, -0.00000972, 0.00001090, -0.00000279, 0.00000497, -0.00000954, -0.00000929, -0.00000892, 0.00000058, -0.00003001, -0.00002112, 0.00002015, 0.00004771, 0.00001691, 0.00005253, 0.00002585, -0.00001528, -0.00003032, -0.00000543, -0.00000809, -0.00000337, -0.00000154, 0.00000239, -0.00000429, -0.00000076, -0.00000189, -0.00000318, -0.00000116, 0.00000005, 0.00000325, 0.00000135, -0.00000717, 0.00000606, 0.00000827, -0.00000594, 0.00000880, -0.00001148, 0.00000694, -0.00001336, -0.00000484, -0.00000014, 0.00000121, -0.00000067, -0.00000683, -0.00000856, 0.00000567, -0.00000952, 0.00001043, -0.00000474, 0.00000909, 0.00000254, -0.00000447, -0.00000216, 0.00000211, 0.00000947, -0.00000620, -0.00000148, 0.00000490, 0.00000127, -0.00000061, -0.00001214, 0.00000831, -0.00000617, -0.00000836, 0.00000232, 0.00000698, -0.00000598, -0.00001088, 0.00000188, -0.00000114, -0.00000098, -0.00000444, 0.00000954, 0.00000618, -0.00000728, 0.00000216, -0.00000396, 0.00000331, -0.00000381, -0.00000278, -0.00000404, -0.00000192, -0.00000337, -0.00000231, -0.00000074, -0.00000288, -0.00001522, -0.00001247, -0.00004627, -0.00008057, 0.00013893, 0.00009157, 0.00013989, -0.00005781, -0.00006347, 0.00000091, -0.00001756, -0.00000370, -0.00002071, -0.00000028, -0.00000714, 0.00000434, 0.00000536, -0.00000081, -0.00000390, -0.00000460, 0.00000012, 0.00000161, 0.00000317, -0.00000683, -0.00000011, -0.00000458, -0.00000165, -0.00000258, 0.00001163, 0.00000366, -0.00000620, -0.00000325, 0.00000379, 0.00001055, -0.00000355, -0.00000239, -0.00000286, -0.00000765, },
		{ 0.00000090, 0.00000484, -0.00000561, 0.00000089, 0.00000330, -0.00000721, -0.00000038, 0.00000740, -0.00000194, 0.00000235, -0.00001242, -0.00000938, -0.00000431, -0.00000751, -0.00003495, -0.00000934, 0.00003965, 0.00003717, 0.00002598, 0.00005748, 0.00000262, -0.00002523, -0.00002389, -0.00000059, -0.00001128, 0.00000170, -0.00000369, 0.00000428, -0.00000828, 0.00000403, -0.00000723, 0.00000166, -0.00000472, 0.00000505, -0.00000029, 0.00000137, -0.00000890, 0.00001672, -0.00000570, 0.00000597, -0.00000346, -0.00000268, 0.00000069, -0.00001420, -0.00000017, 0.00000020, 0.00000008, -0.00000135, -0.00001086, -0.00000085, -0.00000023, -0.00000376, 0.00000848, -0.00000470, 0.00001221, -0.00000418, -0.00000105, -0.00000391, 0.00000819, 0.00000374, -0.00000748, 0.00000354, 0.00000235, 0.00000248, -0.00000662, -0.00000658, 0.00000860, -0.00001329, -0.00000124, 0.00000298, 0.00000587, -0.00001418, -0.00000207, -0.00000092, 0.00000073, -0.00000452, 0.00000128, 0.00001100, -0.00000033, -0.00000571, 0.00000335, -0.00000538, 0.00000603, -0.00000949, 0.00000271, -0.00000954, 0.00000460, -0.00001053, 0.00000628, -0.00000951, 0.00000322, -0.00002938, 0.00000143, -0.00009966, 0.00002114, 0.00013162, 0.00009747, 0.00009620, -0.00010620, -0.00001387, -0.00001690, -0.00000321, -0.00001662, -0.00001033, -0.00000348, -0.00000301, 0.00000568, 0.00000399, -0.00000293, -0.00000458, -0.00000273, 0.00000074, 0.00000286, -0.00000047, -0.00000548, -0.00000049, -0.00000473, -0.00000171, 0.00000162, 0.00001326, -0.00000394, -0.00000448, -0.00000204, 0.00000913, 0.00000568, -0.00000627, 0.00000079, -0.00000863, },
		{ -0.00000949, 0.00000772, -0.00000203, -0.00000272, 0.00000184, 0.00000106, -0.00000978, 0.00000866, -0.00000003, 0.00000275, -0.00000484, -0.00001152, -0.00000964, -0.00000044, -0.00001996, -0.00003298, 0.00000814, 0.00005171, 0.00002388, 0.00004665, 0.00004295, -0.00001647, -0.00002983, -0.00001136, -0.00000545, -0.00000589, -0.00000072, 0.00000046, -0.00000124, -0.00000402, 0.00000081, -0.00000531, 0.00000032, -0.00000216, 0.00000422, 0.00000051, -0.00000424, 0.00000062, 0.00001312, -0.00000847, 0.00001021, -0.00001181, 0.00000687, -0.00001095, -0.00000724, 0.00000038, 0.00000083, -0.00000145, -0.00000307, -0.00001147, 0.00000463, -0.00000613, 0.00000439, 0.00000180, 0.00000123, 0.00000838, -0.00000537, -0.00000072, -0.00000213, 0.00001131, -0.00000368, -0.00000368, 0.00000518, 0.00000046, 0.00000271, -0.00001246, 0.00000348, -0.00000013, -0.00001146, 0.00000204, 0.00000542, -0.00000215, -0.00001318, 0.00000242, -0.00000277, 0.00000096, -0.00000549, 0.00000751, 0.00000849, -0.00000631, 0.00000040, -0.00000149, 0.00000006, 0.00000012, -0.00000590, -0.00000065, -0.00000573, 0.00000110, -0.00000683, 0.00000320, -0.00000614, -0.00000761, -0.00001988, -0.00002419, -0.00009711, 0.00011227, 0.00008876, 0.00012489, 0.00000982, -0.00009169, 0.00000359, -0.00002164, 0.00000080, -0.00002266, -0.00000215, -0.00000641, 0.00000182, 0.00000556, 0.00000192, -0.00000459, -0.00000440, -0.00000041, 0.00000043, 0.00000384, -0.00000462, -0.00000206, -0.00000343, -0.00000242, -0.00000308, 0.00000946, 0.00000702, -0.00000635, -0.00000376, 0.00000220, 0.00001080, -0.00000091, -0.00000427, -0.00000074, },
		{ -0.00000580, -0.00000365, 0.00000760, -0.00000571, 0.00000025, 0.00000246, -0.00000461, -0.00000407, 0.00000967, -0.00000313, 0.00000420, -0.00001086, -0.00001032, -0.00000736, -0.00000235, -0.00003071, -0.00002530, 0.00003136, 0.00004820, 0.00002344, 0.00006181, 0.00001626, -0.00002927, -0.00002474, -0.00000458, -0.00000902, -0.00000067, -0.00000332, 0.00000381, -0.00000680, 0.00000174, -0.00000486, -0.00000055, -0.00000329, 0.00000279, 0.00000084, 0.00000163, -0.00000854, 0.00001319, -0.00000026, 0.00000096, 0.00000137, -0.00000700, 0.00000522, -0.00001614, -0.00000019, -0.00000036, 0.00000041, -0.00000207, -0.00000634, -0.00000750, 0.00000463, -0.00000704, 0.00000882, -0.00000340, 0.00000746, 0.00000184, -0.00000311, -0.00000265, 0.00000326, 0.00000893, -0.00000847, 0.00000278, 0.00000222, 0.00000235, -0.00000235, -0.00001113, 0.00000867, -0.00000952, -0.00000492, 0.00000304, 0.00000586, -0.00001141, -0.00000523, 0.00000069, -0.00000152, -0.00000133, -0.00000245, 0.00001170, 0.00000177, -0.00000639, 0.00000341, -0.00000480, 0.00000464, -0.00000678, 0.00000024, -0.00000678, 0.00000134, -0.00000625, 0.00000136, -0.00000476, 0.00000051, -0.00002306, -0.00000284, -0.00007514, -0.00002204, 0.00013446, 0.00007135, 0.00012294, -0.00006774, -0.00004547, -0.00000564, -0.00001184, -0.00000761, -0.00001793, -0.00000057, -0.00000565, 0.00000501, 0.00000462, -0.00000038, -0.00000608, -0.00000206, -0.00000029, 0.00000185, 0.00000171, -0.00000532, -0.00000127, -0.00000425, -0.00000205, -0.00000037, 0.00001355, -0.00000125, -0.00000535, -0.00000288, 0.00000779, 0.00000748, -0.00000513, -0.00000080, },
		{ -0.00000023, -0.00000900, 0.00000411, 0.00000209, -0.00000484, 0.00000208, 0.00000094, -0.00000849, 0.00000498, 0.00000387, -0.00000045, -0.00000002, -0.00001301, -0.00000974, -0.00000371, -0.00001027, -0.00003707, -0.00000950, 0.00005163, 0.00003438, 0.00003954, 0.00005820, -0.00001206, -0.00003284, -0.00001443, -0.00000600, -0.00000606, -0.00000123, -0.00000078, 0.00000079, -0.00000573, 0.00000172, -0.00000576, 0.00000043, -0.00000299, 0.00000397, 0.00000033, -0.00000207, -0.00000321, 0.00001540, -0.00000821, 0.00000907, -0.00000970, 0.00000414, -0.00000622, -0.00001108, 0.00000188, -0.00000011, -0.00000148, -0.00000183, -0.00000982, -0.00000094, 0.00000017, -0.00000247, 0.00000730, -0.00000363, 0.00000971, -0.00000293, -0.00000102, -0.00000375, 0.00000932, 0.00000147, -0.00000687, 0.00000633, -0.00000106, 0.00000498, -0.00001029, -0.00000239, 0.00000434, -0.00001200, 0.00000015, 0.00000474, 0.00000039, -0.00001399, 0.00000170, -0.00000296, 0.00000083, -0.00000422, 0.00000389, 0.00001105, -0.00000526, -0.00000113, 0.00000042, -0.00000200, 0.00000220, -0.00000692, 0.00000056, -0.00000699, 0.00000233, -0.00000735, 0.00000302, -0.00000580, -0.00000407, -0.00002242, -0.00001178, -0.00009697, 0.00007442, 0.00009870, 0.00009367, 0.00006840, -0.00009464, -0.00000959, -0.00001738, -0.00000171, -0.00001817, -0.00000837, -0.00000325, -0.00000169, 0.00000583, 0.00000351, -0.00000337, -0.00000572, 0.00000030, -0.00000114, 0.00000389, -0.00000226, -0.00000324, -0.00000313, -0.00000258, -0.00000354, 0.00000698, 0.00000981, -0.00000564, -0.00000443, 0.00000061, 0.00001081, 0.00000136, -0.00000501, },
		{ -0.00000229, -0.00000352, -0.00000654, 0.00000768, -0.00000350, -0.00000166, 0.00000266, -0.00000340, -0.00000590, 0.00000946, -0.00000161, 0.00000296, -0.00000665, -0.00001256, -0.00000836, -0.00000236, -0.00002096, -0.00003785, 0.00001589, 0.00005766, 0.00002565, 0.00006057, 0.00003435, -0.00003273, -0.00002623, -0.00000817, -0.00000803, -0.00000184, -0.00000344, 0.00000291, -0.00000472, -0.00000081, -0.00000261, -0.00000248, -0.00000197, 0.00000073, 0.00000173, 0.00000145, -0.00000720, 0.00000882, 0.00000520, -0.00000335, 0.00000524, -0.00001010, 0.00000784, -0.00001572, -0.00000213, 0.00000047, 0.00000019, -0.00000303, -0.00000253, -0.00001078, 0.00000426, -0.00000472, 0.00000419, 0.00000190, 0.00000054, 0.00000730, -0.00000413, -0.00000113, -0.00000138, 0.00001170, -0.00000667, 0.00000027, 0.00000376, 0.00000022, 0.00000246, -0.00001350, 0.00000594, -0.00000472, -0.00000787, 0.00000242, 0.00000579, -0.00000862, -0.00000814, 0.00000222, -0.00000366, 0.00000108, -0.00000450, 0.00001053, 0.00000465, -0.00000698, 0.00000300, -0.00000346, 0.00000260, -0.00000383, -0.00000226, -0.00000392, -0.00000198, -0.00000237, -0.00000236, -0.00000167, -0.00000131, -0.00001713, -0.00000815, -0.00005200, -0.00005312, 0.00012487, 0.00006233, 0.00011766, -0.00001138, -0.00007239, -0.00000118, -0.00001689, -0.00000193, -0.00002118, -0.00000158, -0.00000543, 0.00000280, 0.00000510, 0.00000205, -0.00000657, -0.00000253, 0.00000000, -0.00000003, 0.00000342, -0.00000435, -0.00000218, -0.00000400, -0.00000199, -0.00000222, 0.00001295, 0.00000177, -0.00000581, -0.00000382, 0.00000633, 0.00000886, -0.00000360, },
		{ -0.00000482, -0.00000079, -0.00000722, -0.00000012, 0.00000537, -0.00000557, 0.00000152, 0.00000109, -0.00000728, 0.00000148, 0.00000694, -0.00000230, 0.00000253, -0.00001182, -0.00001112, -0.00000602, -0.00000521, -0.00003197, -0.00002920, 0.00004436, 0.00004764, 0.00003356, 0.00006905, -0.00000070, -0.00003902, -0.00001550, -0.00000850, -0.00000540, -0.00000191, -0.00000183, 0.00000188, -0.00000597, 0.00000113, -0.00000497, -0.00000047, -0.00000274, 0.00000281, 0.00000070, -0.00000080, -0.00000532, 0.00001529, -0.00000560, 0.00000594, -0.00000577, -0.00000028, -0.00000056, -0.00001483, 0.00000292, -0.00000061, -0.00000118, -0.00000252, -0.00000558, -0.00000736, 0.00000490, -0.00000593, 0.00000839, -0.00000327, 0.00000599, 0.00000226, -0.00000268, -0.00000291, 0.00000454, 0.00000741, -0.00000931, 0.00000676, -0.00000151, 0.00000485, -0.00000577, -0.00000805, 0.00000639, -0.00001025, -0.00000250, 0.00000421, 0.00000229, -0.00001386, 0.00000010, -0.00000198, -0.00000079, -0.00000166, -0.00000002, 0.00001282, -0.00000360, -0.00000263, 0.00000174, -0.00000279, 0.00000261, -0.00000623, 0.00000001, -0.00000617, 0.00000105, -0.00000522, 0.00000064, -0.00000391, -0.00000280, -0.00002164, -0.00000675, -0.00008565, 0.00003342, 0.00011008, 0.00006462, 0.00010154, -0.00006912, -0.00003393, -0.00000922, -0.00000791, -0.00001064, -0.00001543, -0.00000053, -0.00000427, 0.00000535, 0.00000429, -0.00000094, -0.00000750, 0.00000105, -0.00000201, 0.00000284, 0.00000021, -0.00000370, -0.00000331, -0.00000260, -0.00000359, 0.00000417, 0.00001192, -0.00000424, -0.00000497, -0.00000114, 0.00001067, 0.00000324, },
		{ -0.00000222, -0.00000308, -0.00000242, -0.00000731, 0.00000529, 0.00000015, -0.00000396, 0.00000322, -0.00000292, -0.00000647, 0.00000744, 0.00000160, 0.00000022, -0.00000187, -0.00001419, -0.00000893, -0.00000453, -0.00001166, -0.00004138, -0.00000678, 0.00006276, 0.00003324, 0.00005523, 0.00005374, -0.00003210, -0.00003074, -0.00000969, -0.00000930, -0.00000159, -0.00000402, 0.00000187, -0.00000275, -0.00000273, -0.00000128, -0.00000350, -0.00000140, -0.00000048, 0.00000183, 0.00000148, -0.00000579, 0.00000477, 0.00000944, -0.00000585, 0.00000706, -0.00001084, 0.00000755, -0.00001249, -0.00000601, 0.00000246, -0.00000050, -0.00000345, -0.00000095, -0.00000969, -0.00000070, 0.00000085, -0.00000196, 0.00000720, -0.00000445, 0.00000908, -0.00000227, -0.00000103, -0.00000364, 0.00001052, -0.00000170, -0.00000390, 0.00000656, -0.00000274, 0.00000617, -0.00001296, 0.00000124, -0.00000036, -0.00000914, 0.00000082, 0.00000580, -0.00000595, -0.00001064, 0.00000326, -0.00000497, 0.00000200, -0.00000453, 0.00000769, 0.00000795, -0.00000731, 0.00000226, -0.00000192, 0.00000086, -0.00000167, -0.00000376, -0.00000216, -0.00000385, -0.00000058, -0.00000338, -0.00000124, -0.00000136, -0.00001287, -0.00001203, -0.00003414, -0.00006952, 0.00010207, 0.00006839, 0.00009291, 0.00004281, -0.00008057, -0.00000921, -0.00001546, -0.00000193, -0.00001872, -0.00000663, -0.00000288, -0.00000037, 0.00000554, 0.00000362, -0.00000538, -0.00000438, 0.00000145, -0.00000223, 0.00000430, -0.00000276, -0.00000287, -0.00000411, -0.00000168, -0.00000365, 0.00001140, 0.00000491, -0.00000582, -0.00000461, 0.00000444, 0.00001023, },
		{ 0.00000513, -0.00000434, -0.00000153, -0.00000542, -0.00000334, 0.00000640, -0.00000410, -0.00000033, 0.00000209, -0.00000680, -0.00000095, 0.00000803, -0.00000165, 0.00000202, -0.00000784, -0.00001380, -0.00000662, -0.00000501, -0.00002144, -0.00004405, 0.00002674, 0.00006201, 0.00003083, 0.00007428, 0.00001695, -0.00004605, -0.00001694, -0.00001076, -0.00000575, -0.00000168, -0.00000309, 0.00000229, -0.00000527, -0.00000019, -0.00000375, -0.00000164, -0.00000212, 0.00000145, 0.00000104, 0.00000004, -0.00000631, 0.00001364, -0.00000174, 0.00000208, -0.00000144, -0.00000464, 0.00000405, -0.00001667, 0.00000212, 0.00000020, -0.00000116, -0.00000390, -0.00000142, -0.00001115, 0.00000507, -0.00000435, 0.00000478, 0.00000127, -0.00000045, 0.00000758, -0.00000381, -0.00000134, -0.00000052, 0.00001132, -0.00000887, 0.00000485, 0.00000045, 0.00000199, 0.00000010, -0.00001223, 0.00000587, -0.00000722, -0.00000473, 0.00000305, 0.00000403, -0.00001316, -0.00000212, -0.00000028, -0.00000307, 0.00000101, -0.00000304, 0.00001306, -0.00000105, -0.00000421, 0.00000257, -0.00000267, 0.00000208, -0.00000478, -0.00000123, -0.00000453, -0.00000105, -0.00000270, -0.00000137, -0.00000274, -0.00000177, -0.00001965, -0.00000578, -0.00006962, -0.00000256, 0.00011339, 0.00004898, 0.00010588, -0.00002458, -0.00005734, -0.00000499, -0.00001255, -0.00000476, -0.00001939, -0.00000097, -0.00000439, 0.00000345, 0.00000492, 0.00000148, -0.00000835, 0.00000061, -0.00000136, 0.00000056, 0.00000260, -0.00000354, -0.00000360, -0.00000288, -0.00000311, 0.00000128, 0.00001313, -0.00000222, -0.00000528, -0.00000289, 0.00001007, },
		{ 0.00001152, -0.00000093, -0.00000327, -0.00000225, -0.00000662, 0.00000179, 0.00000352, -0.00000518, 0.00000299, -0.00000207, -0.00000687, 0.00000493, 0.00000466, -0.00000170, 0.00000096, -0.00001313, -0.00001095, -0.00000574, -0.00000714, -0.00003472, -0.00003185, 0.00005859, 0.00004676, 0.00004758, 0.00007155, -0.00002512, -0.00003877, -0.00000944, -0.00001203, -0.00000101, -0.00000428, 0.00000051, -0.00000116, -0.00000375, -0.00000099, -0.00000361, -0.00000161, -0.00000079, 0.00000118, 0.00000179, -0.00000461, 0.00000145, 0.00001210, -0.00000628, 0.00000668, -0.00000923, 0.00000482, -0.00000745, -0.00001054, 0.00000434, -0.00000086, -0.00000344, -0.00000154, -0.00000558, -0.00000716, 0.00000564, -0.00000567, 0.00000902, -0.00000490, 0.00000605, 0.00000245, -0.00000244, -0.00000323, 0.00000603, 0.00000478, -0.00000787, 0.00000875, -0.00000437, 0.00000700, -0.00000923, -0.00000439, 0.00000263, -0.00000879, -0.00000113, 0.00000531, -0.00000303, -0.00001288, 0.00000369, -0.00000526, 0.00000145, -0.00000300, 0.00000401, 0.00001088, -0.00000697, 0.00000112, -0.00000044, -0.00000024, -0.00000053, -0.00000412, -0.00000166, -0.00000408, -0.00000085, -0.00000228, -0.00000233, -0.00000073, -0.00000964, -0.00001431, -0.00002200, -0.00007339, 0.00007116, 0.00008077, 0.00006483, 0.00007870, -0.00006481, -0.00002744, -0.00001019, -0.00000571, -0.00001283, -0.00001304, -0.00000039, -0.00000284, 0.00000524, 0.00000456, -0.00000308, -0.00000671, 0.00000288, -0.00000358, 0.00000371, -0.00000050, -0.00000337, -0.00000435, -0.00000152, -0.00000431, 0.00000894, 0.00000795, -0.00000539, -0.00000506, 0.00000206, },
		{ 0.00000863, 0.00000736, -0.00000398, -0.00000186, -0.00000436, -0.00000483, 0.00000501, -0.00000096, -0.00000281, 0.00000338, -0.00000661, -0.00000254, 0.00000734, 0.00000089, -0.00000027, -0.00000303, -0.00001587, -0.00000713, -0.00000664, -0.00001183, -0.00004788, -0.00000096, 0.00007319, 0.00003398, 0.00007339, 0.00003911, -0.00005097, -0.00002092, -0.00001131, -0.00000792, -0.00000050, -0.00000435, 0.00000206, -0.00000409, -0.00000156, -0.00000278, -0.00000252, -0.00000173, 0.00000054, 0.00000087, 0.00000092, -0.00000681, 0.00001127, 0.00000233, -0.00000135, 0.00000195, -0.00000746, 0.00000627, -0.00001579, -0.00000084, 0.00000224, -0.00000143, -0.00000507, 0.00000085, -0.00001077, 0.00000067, 0.00000067, -0.00000089, 0.00000664, -0.00000588, 0.00000991, -0.00000246, -0.00000089, -0.00000347, 0.00001124, -0.00000470, 0.00000044, 0.00000445, -0.00000229, 0.00000531, -0.00001361, 0.00000297, -0.00000390, -0.00000592, 0.00000126, 0.00000540, -0.00001167, -0.00000502, 0.00000187, -0.00000541, 0.00000302, -0.00000464, 0.00001173, 0.00000217, -0.00000567, 0.00000291, -0.00000199, 0.00000128, -0.00000338, -0.00000229, -0.00000314, -0.00000257, -0.00000140, -0.00000166, -0.00000311, -0.00000034, -0.00001738, -0.00000660, -0.00005349, -0.00002879, 0.00010503, 0.00004814, 0.00008962, 0.00002265, -0.00006738, -0.00001020, -0.00001257, -0.00000315, -0.00001849, -0.00000509, -0.00000230, 0.00000059, 0.00000548, 0.00000323, -0.00000767, -0.00000130, 0.00000062, -0.00000226, 0.00000426, -0.00000261, -0.00000392, -0.00000335, -0.00000242, -0.00000121, 0.00001317, 0.00000037, -0.00000543, -0.00000429, },
		{ -0.00000050, 0.00001227, 0.00000070, -0.00000337, -0.00000230, -0.00000556, -0.00000129, 0.00000510, -0.00000434, 0.00000134, -0.00000032, -0.00000765, 0.00000286, 0.00000623, -0.00000137, 0.00000078, -0.00000908, -0.00001450, -0.00000526, -0.00000756, -0.00002225, -0.00005093, 0.00004056, 0.00006495, 0.00004035, 0.00008520, -0.00001086, -0.00004913, -0.00000910, -0.00001457, -0.00000149, -0.00000353, -0.00000120, -0.00000008, -0.00000391, -0.00000144, -0.00000317, -0.00000227, -0.00000058, 0.00000024, 0.00000200, -0.00000344, -0.00000131, 0.00001339, -0.00000505, 0.00000476, -0.00000623, 0.00000107, -0.00000236, -0.00001400, 0.00000474, -0.00000011, -0.00000336, -0.00000333, -0.00000083, -0.00001173, 0.00000657, -0.00000494, 0.00000648, -0.00000105, -0.00000017, 0.00000787, -0.00000379, -0.00000162, 0.00000064, 0.00000982, -0.00000912, 0.00000823, -0.00000294, 0.00000429, -0.00000317, -0.00000956, 0.00000373, -0.00000738, -0.00000258, 0.00000385, 0.00000022, -0.00001469, 0.00000323, -0.00000441, -0.00000030, -0.00000061, 0.00000038, 0.00001282, -0.00000573, -0.00000039, 0.00000084, -0.00000066, -0.00000017, -0.00000374, -0.00000196, -0.00000333, -0.00000212, -0.00000067, -0.00000315, -0.00000084, -0.00000653, -0.00001555, -0.00001445, -0.00006890, 0.00003900, 0.00009040, 0.00004519, 0.00009047, -0.00003089, -0.00004671, -0.00000716, -0.00000915, -0.00000740, -0.00001729, -0.00000049, -0.00000315, 0.00000357, 0.00000541, -0.00000070, -0.00000824, 0.00000310, -0.00000317, 0.00000147, 0.00000218, -0.00000358, -0.00000454, -0.00000174, -0.00000420, 0.00000594, 0.00001046, -0.00000442, -0.00000520, },
		{ -0.00000512, 0.00000628, 0.00000982, -0.00000365, -0.00000186, -0.00000396, -0.00000491, 0.00000220, 0.00000235, -0.00000449, 0.00000384, -0.00000577, -0.00000406, 0.00000590, 0.00000367, -0.00000205, 0.00000001, -0.00001505, -0.00000961, -0.00000682, -0.00000782, -0.00003948, -0.00003270, 0.00007409, 0.00004561, 0.00006670, 0.00006347, -0.00005118, -0.00002846, -0.00001014, -0.00001129, 0.00000066, -0.00000486, 0.00000106, -0.00000272, -0.00000259, -0.00000231, -0.00000292, -0.00000181, 0.00000029, 0.00000010, 0.00000189, -0.00000698, 0.00000856, 0.00000605, -0.00000370, 0.00000377, -0.00000823, 0.00000603, -0.00001268, -0.00000500, 0.00000446, -0.00000139, -0.00000580, 0.00000081, -0.00000697, -0.00000586, 0.00000567, -0.00000502, 0.00000946, -0.00000740, 0.00000767, 0.00000181, -0.00000205, -0.00000359, 0.00000733, 0.00000183, -0.00000461, 0.00000833, -0.00000538, 0.00000756, -0.00001134, -0.00000173, -0.00000101, -0.00000614, -0.00000046, 0.00000568, -0.00000894, -0.00000870, 0.00000420, -0.00000732, 0.00000403, -0.00000481, 0.00000925, 0.00000550, -0.00000665, 0.00000266, -0.00000098, 0.00000061, -0.00000237, -0.00000285, -0.00000240, -0.00000303, -0.00000160, -0.00000048, -0.00000425, 0.00000075, -0.00001466, -0.00000825, -0.00003947, -0.00004450, 0.00008689, 0.00005651, 0.00006587, 0.00005784, -0.00005809, -0.00002401, -0.00000943, -0.00000489, -0.00001417, -0.00001083, -0.00000005, -0.00000173, 0.00000529, 0.00000452, -0.00000592, -0.00000382, 0.00000277, -0.00000430, 0.00000431, -0.00000064, -0.00000440, -0.00000373, -0.00000196, -0.00000286, 0.00001196, 0.00000334, -0.00000543, },
		},
		{ { -0.00000043, -0.00000198, 0.00000331, -0.00000513, -0.00000162, -0.00000120, -0.00000288, 0.00000130, -0.00000441, 0.00000291, 0.00000293, -0.00000161, -0.00000925, -0.00000282, -0.00000548, -0.00000682, -0.00001002, -0.00001074, -0.00004964, 0.00000784, 0.00008053, 0.00003782, 0.00008537, 0.00001561, -0.00005906, -0.00001443, -0.00001892, -0.00000616, 0.00000657, -0.00000350, 0.00000211, -0.00000135, -0.00000111, -0.00000185, -0.00000914, 0.00000056, -0.00000064, 0.00000242, -0.00000779, 0.00000192, 0.00000363, 0.00000001, 0.00000027, -0.00000198, -0.00000159, 0.00000023, 0.00000068, 0.00000162, 0.00000110, -0.00000002, -0.00000126, 0.00000042, 0.00000511, -0.00000183, 0.00000903, -0.00000884, 0.00000802, -0.00000586, 0.00000643, -0.00000075, -0.00001370, 0.00000755, -0.00001365, -0.00000064, -0.00000402, -0.00000064, -0.00001365, 0.00000755, -0.00001370, -0.00000075, 0.00000643, -0.00000586, 0.00000802, -0.00000884, 0.00000903, -0.00000183, 0.00000511, 0.00000042, -0.00000126, -0.00000002, 0.00000110, 0.00000162, 0.00000068, 0.00000023, -0.00000159, -0.00000198, 0.00000027, 0.00000001, 0.00000363, 0.00000192, -0.00000779, 0.00000242, -0.00000064, 0.00000056, -0.00000914, -0.00000185, -0.00000111, -0.00000135, 0.00000211, -0.00000350, 0.00000657, -0.00000616, -0.00001892, -0.00001443, -0.00005906, 0.00001561, 0.00008537, 0.00003782, 0.00008053, 0.00000784, -0.00004964, -0.00001074, -0.00001002, -0.00000682, -0.00000548, -0.00000282, -0.00000925, -0.00000161, 0.00000293, 0.00000291, -0.00000441, 0.00000130, -0.00000288, -0.00000120, -0.00000162, -0.00000513, 0.00000331, -0.00000198, },
		{ -0.00000198, -0.00000088, -0.00000080, 0.00000211, -0.00000652, 0.00000125, -0.00000389, 0.00000000, -0.00000151, -0.00000264, 0.00000416, 0.00000190, -0.00000460, -0.00000806, -0.00000277, -0.00000673, -0.00000735, -0.00000980, -0.00001857, -0.00005387, 0.00004661, 0.00007067, 0.00004914, 0.00008699, -0.00003182, -0.00004583, -0.00001093, -0.00001961, 0.00000176, 0.00000330, -0.00000298, 0.00000269, -0.00000322, 0.00000027, -0.00000548, -0.00000617, 0.00000084, 0.00000026, 0.00000031, -0.00000794, 0.00000564, 0.00000107, 0.00000084, -0.00000099, -0.00000167, -0.00000160, 0.00000104, 0.00000041, 0.00000255, -0.00000030, 0.00000080, -0.00000313, 0.00000443, 0.00000128, 0.00000222, 0.00000419, -0.00000597, 0.00000847, -0.00000900, 0.00001230, -0.00001206, -0.00000292, -0.00000029, -0.00001075, -0.00000022, -0.00000261, -0.00000571, -0.00000902, 0.00000775, -0.00001973, 0.00001055, -0.00000289, 0.00000233, 0.00000094, -0.00000508, 0.00000918, -0.00000288, 0.00000667, -0.00000253, 0.00000068, -0.00000093, 0.00000252, 0.00000029, 0.00000126, -0.00000078, -0.00000144, -0.00000197, 0.00000112, -0.00000020, 0.00000540, -0.00000262, -0.00000519, 0.00000260, -0.00000045, -0.00000165, -0.00000989, 0.00000132, -0.00000290, 0.00000127, -0.00000011, -0.00000172, 0.00000728, -0.00001521, -0.00001314, -0.00002941, -0.00005022, 0.00005801, 0.00006418, 0.00004434, 0.00007273, -0.00002292, -0.00003783, -0.00000790, -0.00000918, -0.00000713, -0.00000291, -0.00000516, -0.00000818, 0.00000069, 0.00000378, 0.00000006, -0.00000334, 0.00000180, -0.00000437, 0.00000077, -0.00000498, -0.00000173, 0.00000246, },
		{ 0.00000053, -0.00000122, -0.00000183, 0.00000107, -0.00000091, -0.00000502, 0.00000168, -0.00000501, 0.00000195, -0.00000455, 0.00000094, 0.00000350, 0.00000091, -0.00000755, -0.00000542, -0.00000507, -0.00000569, -0.00001001, -0.00000754, -0.00003370, -0.00004143, 0.00007992, 0.00005286, 0.00007540, 0.00005781, -0.00006211, -0.00002470, -0.00001642, -0.00001329, 0.00000493, 0.00000062, -0.00000179, 0.00000216, -0.00000420, 0.00000069, -0.00000828, -0.00000225, -0.00000035, 0.00000202, -0.00000359, -0.00000487, 0.00000681, -0.00000074, 0.00000139, -0.00000207, -0.00000136, -0.00000124, 0.00000136, 0.00000086, 0.00000238, -0.00000071, 0.00000032, -0.00000300, 0.00000687, -0.00000198, 0.00000621, -0.00000193, 0.00000091, 0.00000260, -0.00000494, 0.00001025, -0.00001649, 0.00000456, -0.00000889, -0.00000436, -0.00000263, -0.00000042, -0.00001247, 0.00000142, -0.00000153, -0.00001542, 0.00001435, -0.00000828, 0.00000898, -0.00000776, 0.00000294, 0.00000487, -0.00000032, 0.00000488, -0.00000279, 0.00000108, -0.00000064, 0.00000281, -0.00000043, 0.00000147, -0.00000149, -0.00000149, -0.00000137, 0.00000129, 0.00000064, 0.00000527, -0.00000612, -0.00000175, 0.00000166, 0.00000011, -0.00000519, -0.00000797, 0.00000288, -0.00000409, 0.00000387, -0.00000358, 0.00000306, 0.00000224, -0.00001934, -0.00001041, -0.00004698, -0.00001850, 0.00007827, 0.00004206, 0.00005893, 0.00004907, -0.00003928, -0.00002595, -0.00000678, -0.00000910, -0.00000607, -0.00000146, -0.00000785, -0.00000561, 0.00000249, 0.00000345, -0.00000257, -0.00000070, 0.00000001, -0.00000357, 0.00000041, -0.00000662, 0.00000178, },
		{ 0.00000374, -0.00000119, -0.00000060, -0.00000238, 0.00000236, -0.00000419, -0.00000163, -0.00000046, -0.00000356, 0.00000121, -0.00000533, 0.00000407, 0.00000204, -0.00000060, -0.00000914, -0.00000344, -0.00000761, -0.00000404, -0.00001334, -0.00000570, -0.00005329, -0.00000803, 0.00009400, 0.00004425, 0.00009871, 0.00000478, -0.00006316, -0.00001197, -0.00002116, -0.00000486, 0.00000469, -0.00000109, -0.00000025, 0.00000042, -0.00000383, -0.00000055, -0.00000881, 0.00000063, -0.00000134, 0.00000306, -0.00000748, 0.00000035, 0.00000517, -0.00000099, 0.00000106, -0.00000261, -0.00000114, -0.00000064, 0.00000146, 0.00000161, 0.00000141, -0.00000034, -0.00000130, -0.00000044, 0.00000632, -0.00000257, 0.00000735, -0.00000519, 0.00000663, -0.00000484, 0.00000293, 0.00000177, -0.00001265, 0.00000431, -0.00001211, -0.00000003, -0.00000393, -0.00000147, -0.00001478, 0.00001034, -0.00001463, -0.00000258, 0.00000864, -0.00000544, 0.00000821, -0.00001145, 0.00000965, -0.00000025, 0.00000344, 0.00000136, -0.00000112, 0.00000011, 0.00000080, 0.00000180, -0.00000021, 0.00000105, -0.00000179, -0.00000168, -0.00000039, 0.00000102, 0.00000211, 0.00000325, -0.00000760, 0.00000130, 0.00000038, 0.00000022, -0.00000899, -0.00000360, 0.00000179, -0.00000310, 0.00000434, -0.00000566, 0.00000810, -0.00000738, -0.00001638, -0.00001749, -0.00005302, 0.00002296, 0.00007373, 0.00003244, 0.00006842, 0.00001894, -0.00004263, -0.00001696, -0.00000621, -0.00000981, -0.00000310, -0.00000268, -0.00000889, -0.00000278, 0.00000400, 0.00000135, -0.00000309, 0.00000115, -0.00000223, -0.00000164, -0.00000199, -0.00000559, },
		{ -0.00000251, 0.00000363, -0.00000188, -0.00000076, -0.00000182, 0.00000194, -0.00000577, 0.00000136, -0.00000336, -0.00000075, -0.00000179, -0.00000304, 0.00000504, 0.00000092, -0.00000302, -0.00000862, -0.00000379, -0.00000792, -0.00000474, -0.00001388, -0.00001016, -0.00006730, 0.00003982, 0.00008478, 0.00005750, 0.00009576, -0.00004793, -0.00004192, -0.00001331, -0.00001836, 0.00000020, 0.00000409, -0.00000279, 0.00000185, -0.00000264, -0.00000194, -0.00000362, -0.00000640, 0.00000098, -0.00000056, 0.00000159, -0.00000882, 0.00000519, 0.00000206, 0.00000000, -0.00000015, -0.00000252, -0.00000097, -0.00000004, 0.00000173, 0.00000198, 0.00000047, -0.00000017, -0.00000265, 0.00000315, 0.00000349, -0.00000066, 0.00000534, -0.00000399, 0.00000748, -0.00000870, 0.00000893, -0.00000752, -0.00000487, -0.00000191, -0.00000891, -0.00000013, -0.00000249, -0.00000723, -0.00000867, 0.00001004, -0.00002193, 0.00001040, -0.00000119, 0.00000294, 0.00000001, -0.00000722, 0.00001088, -0.00000246, 0.00000532, -0.00000120, 0.00000043, -0.00000076, 0.00000223, 0.00000034, 0.00000050, 0.00000028, -0.00000192, -0.00000166, 0.00000057, 0.00000080, 0.00000344, 0.00000006, -0.00000696, 0.00000322, -0.00000071, -0.00000082, -0.00001148, 0.00000184, -0.00000169, 0.00000073, 0.00000103, -0.00000326, 0.00000843, -0.00001579, -0.00001120, -0.00003250, -0.00003906, 0.00005641, 0.00005464, 0.00003737, 0.00006476, -0.00000755, -0.00003842, -0.00001036, -0.00000703, -0.00000935, -0.00000018, -0.00000565, -0.00000794, -0.00000009, 0.00000450, -0.00000147, -0.00000140, 0.00000090, -0.00000312, -0.00000046, -0.00000476, },
		{ -0.00000619, 0.00000106, 0.00000200, -0.00000161, -0.00000162, -0.00000035, -0.00000023, -0.00000472, 0.00000208, -0.00000482, 0.00000114, -0.00000483, 0.00000104, 0.00000379, 0.00000043, -0.00000593, -0.00000663, -0.00000636, -0.00000533, -0.00000928, -0.00000939, -0.00002637, -0.00006188, 0.00008477, 0.00006385, 0.00008820, 0.00005538, -0.00007426, -0.00001864, -0.00002092, -0.00001010, 0.00000147, 0.00000363, -0.00000390, 0.00000332, -0.00000562, 0.00000019, -0.00000699, -0.00000238, -0.00000062, 0.00000172, -0.00000244, -0.00000636, 0.00000740, -0.00000065, 0.00000125, -0.00000183, -0.00000181, -0.00000100, 0.00000073, 0.00000205, 0.00000182, -0.00000011, -0.00000067, -0.00000257, 0.00000591, 0.00000015, 0.00000219, 0.00000201, 0.00000011, 0.00000345, -0.00000699, 0.00000938, -0.00001235, 0.00000067, -0.00000838, -0.00000313, -0.00000258, -0.00000069, -0.00001408, 0.00000329, -0.00000092, -0.00001733, 0.00001495, -0.00000666, 0.00000919, -0.00000949, 0.00000178, 0.00000702, -0.00000109, 0.00000438, -0.00000158, 0.00000062, -0.00000049, 0.00000263, -0.00000067, 0.00000116, -0.00000058, -0.00000201, -0.00000130, 0.00000131, 0.00000076, 0.00000418, -0.00000353, -0.00000435, 0.00000341, -0.00000104, -0.00000369, -0.00001073, 0.00000547, -0.00000498, 0.00000522, -0.00000448, 0.00000335, 0.00000172, -0.00001763, -0.00001165, -0.00004534, -0.00000837, 0.00007033, 0.00003564, 0.00004950, 0.00004825, -0.00002488, -0.00003142, -0.00000528, -0.00000972, -0.00000605, 0.00000010, -0.00000798, -0.00000603, 0.00000268, 0.00000295, -0.00000284, 0.00000060, -0.00000066, -0.00000252, -0.00000099, },
		{ -0.00000275, -0.00000553, 0.00000363, -0.00000001, -0.00000107, -0.00000242, 0.00000101, -0.00000288, -0.00000193, 0.00000046, -0.00000408, 0.00000068, -0.00000568, 0.00000463, 0.00000150, 0.00000006, -0.00000839, -0.00000476, -0.00000910, -0.00000197, -0.00001558, -0.00000255, -0.00005303, -0.00002828, 0.00010751, 0.00005308, 0.00011286, -0.00000926, -0.00006520, -0.00001028, -0.00002299, -0.00000356, 0.00000258, 0.00000138, -0.00000252, 0.00000196, -0.00000613, 0.00000023, -0.00000810, 0.00000049, -0.00000180, 0.00000328, -0.00000681, -0.00000123, 0.00000646, -0.00000176, 0.00000181, -0.00000337, -0.00000065, -0.00000135, 0.00000194, 0.00000190, 0.00000161, -0.00000074, -0.00000134, -0.00000105, 0.00000690, -0.00000238, 0.00000465, -0.00000061, 0.00000414, -0.00000261, -0.00000141, 0.00000459, -0.00001134, 0.00000066, -0.00001022, 0.00000036, -0.00000368, -0.00000252, -0.00001546, 0.00001262, -0.00001552, -0.00000348, 0.00000930, -0.00000351, 0.00000727, -0.00001307, 0.00000932, 0.00000196, 0.00000155, 0.00000222, -0.00000076, -0.00000003, 0.00000063, 0.00000199, -0.00000101, 0.00000160, -0.00000157, -0.00000184, -0.00000075, 0.00000185, 0.00000084, 0.00000414, -0.00000682, -0.00000032, 0.00000168, -0.00000043, -0.00000835, -0.00000563, 0.00000458, -0.00000453, 0.00000621, -0.00000741, 0.00000916, -0.00000842, -0.00001369, -0.00002091, -0.00004534, 0.00002683, 0.00006434, 0.00002714, 0.00005846, 0.00002543, -0.00003319, -0.00002344, -0.00000261, -0.00001243, -0.00000088, -0.00000278, -0.00000824, -0.00000392, 0.00000506, -0.00000043, -0.00000160, 0.00000093, -0.00000175, -0.00000168, },
		{ -0.00000174, -0.00000454, -0.00000310, 0.00000439, -0.00000140, -0.00000099, -0.00000243, 0.00000122, -0.00000436, 0.00000070, -0.00000196, -0.00000218, -0.00000164, -0.00000365, 0.00000599, -0.00000034, -0.00000099, -0.00000948, -0.00000460, -0.00000942, -0.00000168, -0.00001837, -0.00000179, -0.00007820, 0.00002793, 0.00010088, 0.00006816, 0.00010347, -0.00006606, -0.00003560, -0.00001737, -0.00001569, -0.00000234, 0.00000540, -0.00000302, 0.00000146, -0.00000258, -0.00000343, -0.00000265, -0.00000578, 0.00000045, -0.00000072, 0.00000198, -0.00000869, 0.00000396, 0.00000349, -0.00000112, 0.00000103, -0.00000394, 0.00000026, -0.00000155, 0.00000330, 0.00000126, 0.00000149, -0.00000165, -0.00000152, 0.00000124, 0.00000603, -0.00000348, 0.00000618, -0.00000167, 0.00000585, -0.00000736, 0.00000453, -0.00000221, -0.00000711, -0.00000355, -0.00000692, -0.00000018, -0.00000224, -0.00000893, -0.00000775, 0.00001130, -0.00002304, 0.00000994, -0.00000009, 0.00000458, -0.00000190, -0.00000831, 0.00001130, -0.00000081, 0.00000298, 0.00000082, -0.00000031, -0.00000029, 0.00000160, 0.00000090, -0.00000076, 0.00000170, -0.00000259, -0.00000134, -0.00000006, 0.00000200, 0.00000127, 0.00000282, -0.00000855, 0.00000363, -0.00000093, -0.00000002, -0.00001276, 0.00000204, -0.00000055, 0.00000058, 0.00000158, -0.00000397, 0.00000840, -0.00001507, -0.00001059, -0.00003387, -0.00002899, 0.00005335, 0.00004689, 0.00003116, 0.00005721, 0.00000396, -0.00003559, -0.00001448, -0.00000403, -0.00001188, 0.00000269, -0.00000655, -0.00000709, -0.00000120, 0.00000544, -0.00000339, 0.00000099, -0.00000047, -0.00000157, },
		{ -0.00000099, -0.00000272, -0.00000542, 0.00000014, 0.00000340, -0.00000185, -0.00000154, -0.00000159, 0.00000015, -0.00000397, 0.00000181, -0.00000367, -0.00000070, -0.00000415, 0.00000034, 0.00000464, -0.00000074, -0.00000333, -0.00000883, -0.00000664, -0.00000612, -0.00000707, -0.00001317, -0.00001664, -0.00008266, 0.00008555, 0.00007821, 0.00010252, 0.00005021, -0.00008600, -0.00001169, -0.00002632, -0.00000590, -0.00000286, 0.00000709, -0.00000627, 0.00000465, -0.00000719, 0.00000000, -0.00000626, -0.00000187, -0.00000137, 0.00000185, -0.00000195, -0.00000688, 0.00000703, 0.00000016, 0.00000065, -0.00000121, -0.00000284, -0.00000004, -0.00000058, 0.00000375, 0.00000094, 0.00000079, -0.00000216, -0.00000135, 0.00000394, 0.00000324, -0.00000251, 0.00000639, -0.00000113, 0.00000448, -0.00000867, 0.00000784, -0.00000750, -0.00000367, -0.00000762, -0.00000198, -0.00000250, -0.00000101, -0.00001556, 0.00000532, -0.00000120, -0.00001763, 0.00001411, -0.00000436, 0.00000932, -0.00001127, 0.00000085, 0.00000851, -0.00000096, 0.00000296, 0.00000043, -0.00000055, 0.00000017, 0.00000200, -0.00000029, 0.00000014, 0.00000101, -0.00000309, -0.00000086, 0.00000099, 0.00000133, 0.00000257, -0.00000049, -0.00000717, 0.00000522, -0.00000234, -0.00000199, -0.00001331, 0.00000761, -0.00000557, 0.00000653, -0.00000554, 0.00000401, 0.00000036, -0.00001474, -0.00001438, -0.00004117, -0.00000150, 0.00006356, 0.00002958, 0.00004207, 0.00004533, -0.00001191, -0.00003434, -0.00000554, -0.00000917, -0.00000660, 0.00000179, -0.00000832, -0.00000596, 0.00000253, 0.00000260, -0.00000336, 0.00000226, -0.00000181, },
		{ -0.00000153, -0.00000134, -0.00000383, -0.00000502, 0.00000303, 0.00000146, -0.00000172, -0.00000226, -0.00000057, -0.00000135, -0.00000238, 0.00000139, -0.00000425, -0.00000044, -0.00000532, 0.00000449, 0.00000143, 0.00000027, -0.00000690, -0.00000694, -0.00000965, -0.00000100, -0.00001628, -0.00000173, -0.00004882, -0.00005193, 0.00011924, 0.00006599, 0.00012671, -0.00002606, -0.00006519, -0.00000942, -0.00002431, -0.00000236, 0.00000034, 0.00000375, -0.00000450, 0.00000306, -0.00000785, 0.00000052, -0.00000716, 0.00000030, -0.00000209, 0.00000321, -0.00000600, -0.00000256, 0.00000727, -0.00000209, 0.00000232, -0.00000406, -0.00000034, -0.00000169, 0.00000197, 0.00000257, 0.00000163, -0.00000112, -0.00000145, -0.00000133, 0.00000679, -0.00000128, 0.00000110, 0.00000465, 0.00000081, 0.00000039, -0.00000604, 0.00000727, -0.00000965, -0.00000334, -0.00000804, 0.00000056, -0.00000326, -0.00000379, -0.00001562, 0.00001433, -0.00001638, -0.00000332, 0.00000835, -0.00000023, 0.00000543, -0.00001385, 0.00000820, 0.00000454, -0.00000032, 0.00000284, -0.00000011, -0.00000052, 0.00000067, 0.00000205, -0.00000152, 0.00000172, -0.00000082, -0.00000253, -0.00000073, 0.00000235, -0.00000004, 0.00000450, -0.00000546, -0.00000234, 0.00000318, -0.00000132, -0.00000731, -0.00000772, 0.00000690, -0.00000535, 0.00000752, -0.00000859, 0.00000966, -0.00000914, -0.00001105, -0.00002438, -0.00003646, 0.00002750, 0.00005739, 0.00002141, 0.00005092, 0.00002796, -0.00002248, -0.00002922, 0.00000011, -0.00001420, 0.00000081, -0.00000288, -0.00000747, -0.00000487, 0.00000595, -0.00000221, -0.00000013, 0.00000080, },
		{ -0.00000196, -0.00000002, -0.00000286, -0.00000444, -0.00000337, 0.00000463, -0.00000047, -0.00000159, -0.00000261, 0.00000001, -0.00000236, -0.00000064, 0.00000016, -0.00000412, -0.00000122, -0.00000431, 0.00000686, -0.00000171, 0.00000131, -0.00001045, -0.00000537, -0.00001090, 0.00000127, -0.00002250, 0.00000547, -0.00008538, 0.00001056, 0.00011815, 0.00008248, 0.00010873, -0.00008512, -0.00002750, -0.00002271, -0.00001192, -0.00000562, 0.00000705, -0.00000362, 0.00000154, -0.00000315, -0.00000401, -0.00000267, -0.00000435, -0.00000066, -0.00000025, 0.00000153, -0.00000764, 0.00000213, 0.00000514, -0.00000231, 0.00000235, -0.00000569, 0.00000181, -0.00000322, 0.00000485, 0.00000063, 0.00000255, -0.00000340, 0.00000001, -0.00000098, 0.00000855, -0.00000593, 0.00000653, 0.00000107, 0.00000353, -0.00000511, -0.00000051, 0.00000336, -0.00000927, -0.00000532, -0.00000476, -0.00000038, -0.00000185, -0.00001080, -0.00000626, 0.00001146, -0.00002303, 0.00000926, 0.00000026, 0.00000723, -0.00000464, -0.00000846, 0.00001057, 0.00000183, -0.00000002, 0.00000326, -0.00000138, 0.00000034, 0.00000078, 0.00000178, -0.00000225, 0.00000320, -0.00000318, -0.00000120, -0.00000058, 0.00000314, -0.00000078, 0.00000528, -0.00000968, 0.00000362, -0.00000091, 0.00000051, -0.00001344, 0.00000176, 0.00000053, 0.00000085, 0.00000154, -0.00000391, 0.00000732, -0.00001322, -0.00001116, -0.00003356, -0.00002023, 0.00004917, 0.00004095, 0.00002509, 0.00005093, 0.00001144, -0.00002991, -0.00001954, -0.00000079, -0.00001417, 0.00000519, -0.00000748, -0.00000594, -0.00000237, 0.00000634, -0.00000541, 0.00000351, },
		{ 0.00000407, -0.00000319, 0.00000072, -0.00000449, -0.00000451, -0.00000079, 0.00000454, -0.00000178, -0.00000173, -0.00000252, 0.00000015, -0.00000274, 0.00000084, -0.00000158, -0.00000348, -0.00000263, -0.00000104, 0.00000595, -0.00000249, 0.00000008, -0.00001187, -0.00000601, -0.00000790, -0.00000370, -0.00001827, -0.00000571, -0.00010187, 0.00008051, 0.00009668, 0.00011845, 0.00004168, -0.00009653, -0.00000450, -0.00003213, -0.00000108, -0.00000771, 0.00001068, -0.00000864, 0.00000598, -0.00000886, 0.00000026, -0.00000624, -0.00000067, -0.00000263, 0.00000247, -0.00000220, -0.00000637, 0.00000567, 0.00000163, -0.00000037, -0.00000023, -0.00000437, 0.00000151, -0.00000238, 0.00000575, -0.00000006, 0.00000179, -0.00000396, 0.00000042, 0.00000126, 0.00000687, -0.00000742, 0.00001075, -0.00000241, 0.00000526, -0.00000963, 0.00000561, -0.00000220, -0.00000816, -0.00000680, -0.00000084, -0.00000242, -0.00000138, -0.00001686, 0.00000747, -0.00000235, -0.00001636, 0.00001196, -0.00000164, 0.00000955, -0.00001313, 0.00000019, 0.00000930, 0.00000003, 0.00000079, 0.00000304, -0.00000226, 0.00000121, 0.00000100, 0.00000062, -0.00000142, 0.00000306, -0.00000452, -0.00000023, 0.00000046, 0.00000216, 0.00000072, 0.00000262, -0.00000980, 0.00000673, -0.00000348, -0.00000043, -0.00001526, 0.00000887, -0.00000560, 0.00000761, -0.00000655, 0.00000484, -0.00000153, -0.00001104, -0.00001813, -0.00003494, 0.00000236, 0.00005806, 0.00002382, 0.00003610, 0.00004144, -0.00000131, -0.00003442, -0.00000748, -0.00000758, -0.00000753, 0.00000336, -0.00000867, -0.00000553, 0.00000216, 0.00000230, -0.00000401, },
		{ 0.00000117, 0.00000085, -0.00000158, -0.00000072, -0.00000508, -0.00000418, 0.00000206, 0.00000306, -0.00000240, -0.00000204, -0.00000217, 0.00000019, -0.00000283, 0.00000214, -0.00000392, -0.00000222, -0.00000421, 0.00000364, 0.00000183, -0.00000000, -0.00000471, -0.00000997, -0.00000914, -0.00000134, -0.00001529, -0.00000331, -0.00004111, -0.00007739, 0.00012697, 0.00008483, 0.00013900, -0.00004497, -0.00006325, -0.00000945, -0.00002507, -0.00000133, -0.00000189, 0.00000586, -0.00000602, 0.00000356, -0.00000888, 0.00000039, -0.00000616, 0.00000019, -0.00000235, 0.00000300, -0.00000522, -0.00000343, 0.00000738, -0.00000182, 0.00000245, -0.00000451, -0.00000038, -0.00000153, 0.00000148, 0.00000364, 0.00000145, -0.00000145, -0.00000168, -0.00000126, 0.00000600, 0.00000057, -0.00000298, 0.00001022, -0.00000292, 0.00000361, -0.00001030, 0.00000937, -0.00000744, -0.00000760, -0.00000567, 0.00000058, -0.00000268, -0.00000528, -0.00001524, 0.00001540, -0.00001718, -0.00000215, 0.00000591, 0.00000410, 0.00000302, -0.00001403, 0.00000656, 0.00000721, -0.00000193, 0.00000311, 0.00000084, -0.00000136, 0.00000098, 0.00000190, -0.00000164, 0.00000134, 0.00000044, -0.00000372, -0.00000035, 0.00000249, -0.00000044, 0.00000432, -0.00000363, -0.00000458, 0.00000470, -0.00000233, -0.00000603, -0.00000958, 0.00000842, -0.00000530, 0.00000807, -0.00000907, 0.00000949, -0.00000941, -0.00000870, -0.00002753, -0.00002696, 0.00002548, 0.00005276, 0.00001510, 0.00004560, 0.00002751, -0.00001175, -0.00003340, 0.00000141, -0.00001473, 0.00000170, -0.00000282, -0.00000671, -0.00000550, 0.00000653, -0.00000385, },
		{ -0.00000722, 0.00000581, -0.00000326, 0.00000125, -0.00000365, -0.00000454, -0.00000329, 0.00000433, 0.00000089, -0.00000252, -0.00000246, -0.00000153, 0.00000001, -0.00000246, 0.00000275, -0.00000628, -0.00000078, -0.00000481, 0.00000744, -0.00000298, 0.00000361, -0.00001133, -0.00000633, -0.00001198, 0.00000358, -0.00002554, 0.00001063, -0.00008798, -0.00001214, 0.00013527, 0.00010211, 0.00010989, -0.00010375, -0.00001853, -0.00002872, -0.00000750, -0.00000926, 0.00000876, -0.00000440, 0.00000200, -0.00000435, -0.00000364, -0.00000363, -0.00000230, -0.00000218, 0.00000070, 0.00000040, -0.00000590, -0.00000003, 0.00000675, -0.00000332, 0.00000357, -0.00000748, 0.00000336, -0.00000470, 0.00000607, 0.00000032, 0.00000341, -0.00000515, 0.00000165, -0.00000320, 0.00001070, -0.00000774, 0.00000625, 0.00000425, 0.00000058, -0.00000221, -0.00000564, 0.00000860, -0.00001098, -0.00000734, -0.00000242, -0.00000076, -0.00000133, -0.00001279, -0.00000423, 0.00001054, -0.00002192, 0.00000844, -0.00000017, 0.00001071, -0.00000792, -0.00000792, 0.00000897, 0.00000509, -0.00000331, 0.00000579, -0.00000253, 0.00000094, -0.00000006, 0.00000276, -0.00000369, 0.00000450, -0.00000348, -0.00000142, -0.00000082, 0.00000398, -0.00000239, 0.00000712, -0.00001010, 0.00000303, -0.00000053, 0.00000061, -0.00001334, 0.00000092, 0.00000150, 0.00000155, 0.00000096, -0.00000318, 0.00000539, -0.00001050, -0.00001268, -0.00003172, -0.00001292, 0.00004421, 0.00003673, 0.00001883, 0.00004631, 0.00001517, -0.00002225, -0.00002469, 0.00000202, -0.00001567, 0.00000686, -0.00000809, -0.00000477, -0.00000333, 0.00000693, },
		{ 0.00000192, -0.00000461, 0.00000579, -0.00000450, 0.00000227, -0.00000601, -0.00000368, -0.00000158, 0.00000526, -0.00000129, -0.00000228, -0.00000305, -0.00000023, -0.00000108, -0.00000073, 0.00000127, -0.00000694, -0.00000056, -0.00000282, 0.00000745, -0.00000452, 0.00000396, -0.00001544, -0.00000476, -0.00001028, 0.00000021, -0.00002377, 0.00000493, -0.00011750, 0.00006809, 0.00011970, 0.00013628, 0.00002902, -0.00010488, 0.00000210, -0.00003767, 0.00000377, -0.00001256, 0.00001397, -0.00001070, 0.00000709, -0.00001051, 0.00000098, -0.00000695, 0.00000114, -0.00000428, 0.00000350, -0.00000312, -0.00000492, 0.00000349, 0.00000360, -0.00000163, 0.00000093, -0.00000620, 0.00000338, -0.00000437, 0.00000774, -0.00000094, 0.00000267, -0.00000577, 0.00000241, -0.00000171, 0.00001054, -0.00001197, 0.00001459, -0.00000330, 0.00000536, -0.00000961, 0.00000276, 0.00000322, -0.00001245, -0.00000609, 0.00000033, -0.00000236, -0.00000180, -0.00001793, 0.00000962, -0.00000426, -0.00001365, 0.00000874, 0.00000122, 0.00001000, -0.00001505, -0.00000023, 0.00000943, 0.00000171, -0.00000185, 0.00000593, -0.00000427, 0.00000245, -0.00000022, 0.00000189, -0.00000327, 0.00000527, -0.00000601, 0.00000035, -0.00000006, 0.00000301, -0.00000104, 0.00000541, -0.00001185, 0.00000763, -0.00000418, 0.00000065, -0.00001621, 0.00000892, -0.00000485, 0.00000828, -0.00000735, 0.00000561, -0.00000363, -0.00000700, -0.00002232, -0.00002732, 0.00000367, 0.00005369, 0.00001850, 0.00003103, 0.00003756, 0.00000630, -0.00003173, -0.00001080, -0.00000527, -0.00000855, 0.00000453, -0.00000879, -0.00000495, 0.00000175, },
		{ 0.00000671, -0.00000521, 0.00000219, 0.00000110, -0.00000183, 0.00000004, -0.00000630, -0.00000318, 0.00000090, 0.00000461, -0.00000294, -0.00000196, -0.00000360, 0.00000157, -0.00000313, 0.00000258, -0.00000303, -0.00000464, -0.00000243, 0.00000217, 0.00000264, -0.00000069, -0.00000194, -0.00001370, -0.00000762, -0.00000293, -0.00001274, -0.00000693, -0.00003089, -0.00010263, 0.00012827, 0.00011144, 0.00014847, -0.00006533, -0.00005953, -0.00001039, -0.00002518, -0.00000060, -0.00000396, 0.00000752, -0.00000694, 0.00000336, -0.00000920, -0.00000006, -0.00000529, 0.00000029, -0.00000265, 0.00000277, -0.00000463, -0.00000369, 0.00000669, -0.00000089, 0.00000212, -0.00000462, -0.00000083, -0.00000085, 0.00000050, 0.00000502, 0.00000111, -0.00000173, -0.00000197, -0.00000090, 0.00000470, 0.00000288, -0.00000717, 0.00001560, -0.00000656, 0.00000644, -0.00001357, 0.00001048, -0.00000464, -0.00001200, -0.00000320, 0.00000048, -0.00000198, -0.00000697, -0.00001429, 0.00001574, -0.00001779, -0.00000012, 0.00000229, 0.00000902, 0.00000045, -0.00001388, 0.00000464, 0.00000969, -0.00000310, 0.00000299, 0.00000205, -0.00000251, 0.00000153, 0.00000150, -0.00000133, 0.00000049, 0.00000212, -0.00000527, 0.00000030, 0.00000229, -0.00000039, 0.00000370, -0.00000153, -0.00000680, 0.00000602, -0.00000327, -0.00000471, -0.00001091, 0.00000882, -0.00000420, 0.00000775, -0.00000878, 0.00000859, -0.00000911, -0.00000687, -0.00002995, -0.00001750, 0.00002150, 0.00005002, 0.00000843, 0.00004200, 0.00002520, -0.00000220, -0.00003532, 0.00000097, -0.00001383, 0.00000164, -0.00000250, -0.00000603, -0.00000574, },
		{ -0.00000386, 0.00000701, -0.00000857, 0.00000760, -0.00000411, 0.00000206, -0.00000403, -0.00000482, -0.00000289, 0.00000355, 0.00000255, -0.00000363, -0.00000212, -0.00000318, 0.00000249, -0.00000448, 0.00000547, -0.00000831, -0.00000061, -0.00000491, 0.00000754, -0.00000393, 0.00000565, -0.00001187, -0.00000769, -0.00001232, 0.00000473, -0.00002684, 0.00001294, -0.00008556, -0.00003944, 0.00015048, 0.00012895, 0.00010518, -0.00012046, -0.00000973, -0.00003466, -0.00000300, -0.00001282, 0.00001021, -0.00000515, 0.00000268, -0.00000606, -0.00000244, -0.00000530, 0.00000006, -0.00000382, 0.00000188, -0.00000115, -0.00000377, -0.00000221, 0.00000800, -0.00000388, 0.00000440, -0.00000899, 0.00000457, -0.00000569, 0.00000671, 0.00000053, 0.00000385, -0.00000664, 0.00000311, -0.00000510, 0.00001220, -0.00000871, 0.00000532, 0.00000777, -0.00000283, 0.00000099, -0.00001029, 0.00001295, -0.00001192, -0.00000971, 0.00000011, -0.00000130, -0.00000071, -0.00001484, -0.00000173, 0.00000858, -0.00001976, 0.00000748, -0.00000126, 0.00001468, -0.00001132, -0.00000705, 0.00000686, 0.00000854, -0.00000643, 0.00000807, -0.00000353, 0.00000136, -0.00000074, 0.00000361, -0.00000482, 0.00000535, -0.00000332, -0.00000210, -0.00000066, 0.00000436, -0.00000334, 0.00000812, -0.00000967, 0.00000180, 0.00000026, 0.00000019, -0.00001239, -0.00000045, 0.00000229, 0.00000271, -0.00000010, -0.00000193, 0.00000284, -0.00000718, -0.00001486, -0.00002856, -0.00000716, 0.00003881, 0.00003399, 0.00001228, 0.00004336, 0.00001583, -0.00001370, -0.00002902, 0.00000378, -0.00001590, 0.00000733, -0.00000806, -0.00000385, },
		{ -0.00000441, 0.00000143, 0.00000132, -0.00000498, 0.00000717, -0.00000544, 0.00000340, -0.00000706, -0.00000305, -0.00000210, 0.00000549, -0.00000036, -0.00000317, -0.00000324, -0.00000088, 0.00000082, -0.00000266, 0.00000456, -0.00001065, 0.00000167, -0.00000461, 0.00000879, -0.00000645, 0.00000783, -0.00001903, -0.00000335, -0.00001269, 0.00000387, -0.00002856, 0.00001370, -0.00012762, 0.00004707, 0.00014729, 0.00015654, 0.00001137, -0.00010995, 0.00000715, -0.00004216, 0.00000800, -0.00001683, 0.00001647, -0.00001206, 0.00000770, -0.00001196, 0.00000204, -0.00000825, 0.00000331, -0.00000608, 0.00000473, -0.00000451, -0.00000276, 0.00000076, 0.00000576, -0.00000289, 0.00000203, -0.00000803, 0.00000524, -0.00000617, 0.00000936, -0.00000141, 0.00000314, -0.00000727, 0.00000425, -0.00000451, 0.00001370, -0.00001563, 0.00001744, -0.00000342, 0.00000446, -0.00000846, -0.00000055, 0.00000838, -0.00001621, -0.00000565, 0.00000159, -0.00000233, -0.00000230, -0.00001867, 0.00001162, -0.00000674, -0.00000977, 0.00000474, 0.00000398, 0.00001071, -0.00001688, -0.00000054, 0.00000908, 0.00000377, -0.00000458, 0.00000873, -0.00000628, 0.00000366, -0.00000144, 0.00000325, -0.00000509, 0.00000729, -0.00000724, 0.00000067, -0.00000040, 0.00000366, -0.00000238, 0.00000749, -0.00001299, 0.00000765, -0.00000421, 0.00000101, -0.00001586, 0.00000755, -0.00000323, 0.00000845, -0.00000776, 0.00000609, -0.00000558, -0.00000310, -0.00002634, -0.00001905, 0.00000305, 0.00005018, 0.00001383, 0.00002631, 0.00003443, 0.00001071, -0.00002667, -0.00001496, -0.00000267, -0.00000933, 0.00000502, -0.00000845, },
		{ -0.00000541, -0.00000560, 0.00000648, -0.00000623, 0.00000289, 0.00000153, -0.00000218, 0.00000074, -0.00000728, -0.00000221, -0.00000028, 0.00000591, -0.00000317, -0.00000214, -0.00000470, 0.00000266, -0.00000320, 0.00000266, -0.00000160, -0.00000753, -0.00000021, 0.00000030, 0.00000365, -0.00000158, 0.00000112, -0.00001782, -0.00000537, -0.00000547, -0.00000907, -0.00001186, -0.00001956, -0.00012529, 0.00012070, 0.00014747, 0.00015403, -0.00008648, -0.00005420, -0.00001231, -0.00002452, -0.00000030, -0.00000569, 0.00000857, -0.00000712, 0.00000238, -0.00000880, -0.00000071, -0.00000468, 0.00000064, -0.00000301, 0.00000256, -0.00000429, -0.00000330, 0.00000523, 0.00000064, 0.00000135, -0.00000437, -0.00000169, 0.00000027, -0.00000082, 0.00000656, 0.00000073, -0.00000204, -0.00000224, -0.00000041, 0.00000311, 0.00000527, -0.00001098, 0.00002026, -0.00000955, 0.00000826, -0.00001529, 0.00001030, -0.00000128, -0.00001637, -0.00000078, 0.00000031, -0.00000119, -0.00000883, -0.00001275, 0.00001527, -0.00001806, 0.00000248, -0.00000210, 0.00001403, -0.00000188, -0.00001363, 0.00000267, 0.00001177, -0.00000373, 0.00000250, 0.00000339, -0.00000383, 0.00000224, 0.00000090, -0.00000065, -0.00000072, 0.00000401, -0.00000695, 0.00000107, 0.00000185, 0.00000001, 0.00000280, 0.00000058, -0.00000870, 0.00000689, -0.00000393, -0.00000357, -0.00001144, 0.00000790, -0.00000194, 0.00000654, -0.00000769, 0.00000694, -0.00000816, -0.00000578, -0.00003126, -0.00000876, 0.00001636, 0.00004857, 0.00000183, 0.00003941, 0.00002221, 0.00000522, -0.00003459, -0.00000128, -0.00001151, 0.00000066, -0.00000196, },
		{ -0.00000723, -0.00000333, -0.00000383, 0.00000644, -0.00000928, 0.00000869, -0.00000435, 0.00000230, -0.00000399, -0.00000519, -0.00000228, 0.00000244, 0.00000432, -0.00000471, -0.00000183, -0.00000466, 0.00000475, -0.00000634, 0.00000794, -0.00000986, -0.00000095, -0.00000444, 0.00000700, -0.00000435, 0.00000718, -0.00001186, -0.00000964, -0.00001163, 0.00000435, -0.00002597, 0.00001200, -0.00007817, -0.00007007, 0.00016171, 0.00016487, 0.00009291, -0.00013372, -0.00000220, -0.00003972, 0.00000094, -0.00001580, 0.00001103, -0.00000560, 0.00000336, -0.00000805, -0.00000064, -0.00000736, 0.00000234, -0.00000524, 0.00000298, -0.00000278, -0.00000159, -0.00000405, 0.00000859, -0.00000376, 0.00000462, -0.00000992, 0.00000515, -0.00000593, 0.00000658, 0.00000139, 0.00000374, -0.00000765, 0.00000415, -0.00000644, 0.00001285, -0.00000876, 0.00000379, 0.00001145, -0.00000642, 0.00000402, -0.00001388, 0.00001590, -0.00001183, -0.00001244, 0.00000275, -0.00000195, -0.00000004, -0.00001684, 0.00000112, 0.00000571, -0.00001663, 0.00000634, -0.00000280, 0.00001873, -0.00001436, -0.00000623, 0.00000463, 0.00001172, -0.00000898, 0.00000979, -0.00000419, 0.00000145, -0.00000112, 0.00000415, -0.00000541, 0.00000558, -0.00000262, -0.00000324, -0.00000009, 0.00000420, -0.00000350, 0.00000820, -0.00000840, 0.00000000, 0.00000139, -0.00000072, -0.00001061, -0.00000224, 0.00000278, 0.00000432, -0.00000153, -0.00000035, -0.00000005, -0.00000358, -0.00001738, -0.00002434, -0.00000296, 0.00003334, 0.00003236, 0.00000562, 0.00004174, 0.00001435, -0.00000540, -0.00003175, 0.00000403, -0.00001460, 0.00000638, },
		{ 0.00000461, -0.00000748, -0.00000405, 0.00000133, 0.00000042, -0.00000499, 0.00000800, -0.00000581, 0.00000388, -0.00000749, -0.00000271, -0.00000234, 0.00000524, 0.00000093, -0.00000429, -0.00000320, -0.00000162, 0.00000269, -0.00000462, 0.00000782, -0.00001411, 0.00000356, -0.00000596, 0.00000953, -0.00000783, 0.00001114, -0.00002207, -0.00000237, -0.00001441, 0.00000637, -0.00003148, 0.00001904, -0.00013063, 0.00001679, 0.00017900, 0.00017998, -0.00001221, -0.00011063, 0.00000960, -0.00004473, 0.00001083, -0.00001985, 0.00001764, -0.00001233, 0.00000749, -0.00001294, 0.00000322, -0.00000984, 0.00000549, -0.00000769, 0.00000584, -0.00000603, -0.00000030, -0.00000211, 0.00000774, -0.00000382, 0.00000278, -0.00000952, 0.00000668, -0.00000740, 0.00001028, -0.00000119, 0.00000295, -0.00000814, 0.00000553, -0.00000667, 0.00001584, -0.00001790, 0.00001892, -0.00000247, 0.00000237, -0.00000618, -0.00000406, 0.00001289, -0.00001911, -0.00000562, 0.00000295, -0.00000232, -0.00000292, -0.00001896, 0.00001330, -0.00000951, -0.00000504, 0.00000028, 0.00000643, 0.00001163, -0.00001844, -0.00000093, 0.00000848, 0.00000590, -0.00000700, 0.00001105, -0.00000795, 0.00000460, -0.00000244, 0.00000445, -0.00000655, 0.00000877, -0.00000794, 0.00000052, -0.00000038, 0.00000391, -0.00000305, 0.00000858, -0.00001299, 0.00000664, -0.00000346, 0.00000051, -0.00001408, 0.00000475, -0.00000078, 0.00000809, -0.00000769, 0.00000609, -0.00000706, 0.00000019, -0.00002960, -0.00001089, 0.00000117, 0.00004717, 0.00001001, 0.00002154, 0.00003247, 0.00001210, -0.00001992, -0.00001929, -0.00000026, -0.00000953, },
		{ -0.00000108, -0.00000128, -0.00000481, -0.00000513, 0.00000586, -0.00000689, 0.00000327, 0.00000205, -0.00000248, 0.00000119, -0.00000781, -0.00000144, -0.00000132, 0.00000683, -0.00000300, -0.00000267, -0.00000537, 0.00000338, -0.00000302, 0.00000242, 0.00000019, -0.00001061, 0.00000212, -0.00000162, 0.00000457, -0.00000238, 0.00000405, -0.00002185, -0.00000286, -0.00000843, -0.00000501, -0.00001707, -0.00000884, -0.00014293, 0.00010210, 0.00019410, 0.00015491, -0.00010793, -0.00004734, -0.00001532, -0.00002293, -0.00000065, -0.00000685, 0.00000879, -0.00000644, 0.00000057, -0.00000768, -0.00000151, -0.00000436, 0.00000122, -0.00000340, 0.00000235, -0.00000416, -0.00000236, 0.00000313, 0.00000262, 0.00000027, -0.00000385, -0.00000282, 0.00000164, -0.00000224, 0.00000801, 0.00000046, -0.00000249, -0.00000234, 0.00000001, 0.00000154, 0.00000732, -0.00001389, 0.00002368, -0.00001139, 0.00000859, -0.00001506, 0.00000869, 0.00000254, -0.00002049, 0.00000145, 0.00000013, -0.00000034, -0.00001080, -0.00001062, 0.00001393, -0.00001778, 0.00000529, -0.00000678, 0.00001861, -0.00000362, -0.00001347, 0.00000081, 0.00001332, -0.00000380, 0.00000177, 0.00000467, -0.00000513, 0.00000299, 0.00000019, 0.00000028, -0.00000208, 0.00000584, -0.00000849, 0.00000174, 0.00000133, 0.00000059, 0.00000185, 0.00000242, -0.00001001, 0.00000710, -0.00000414, -0.00000280, -0.00001098, 0.00000554, 0.00000143, 0.00000450, -0.00000585, 0.00000457, -0.00000651, -0.00000557, -0.00003114, -0.00000139, 0.00001093, 0.00004770, -0.00000415, 0.00003707, 0.00001955, 0.00000984, -0.00003115, -0.00000513, -0.00000799, },
		{ -0.00001167, 0.00000398, -0.00000556, -0.00000325, -0.00000321, 0.00000519, -0.00000926, 0.00000897, -0.00000391, 0.00000201, -0.00000365, -0.00000551, -0.00000165, 0.00000121, 0.00000596, -0.00000553, -0.00000181, -0.00000569, 0.00000649, -0.00000772, 0.00000980, -0.00001063, -0.00000201, -0.00000328, 0.00000573, -0.00000411, 0.00000801, -0.00001116, -0.00001227, -0.00000979, 0.00000225, -0.00002278, 0.00000777, -0.00006642, -0.00010226, 0.00016668, 0.00021164, 0.00007160, -0.00014215, 0.00000299, -0.00004315, 0.00000373, -0.00001773, 0.00001088, -0.00000549, 0.00000377, -0.00001004, 0.00000143, -0.00000941, 0.00000415, -0.00000610, 0.00000371, -0.00000419, 0.00000029, -0.00000528, 0.00000833, -0.00000280, 0.00000405, -0.00001007, 0.00000488, -0.00000526, 0.00000560, 0.00000291, 0.00000300, -0.00000803, 0.00000460, -0.00000707, 0.00001254, -0.00000792, 0.00000185, 0.00001501, -0.00000985, 0.00000642, -0.00001590, 0.00001706, -0.00001056, -0.00001547, 0.00000542, -0.00000263, 0.00000060, -0.00001867, 0.00000415, 0.00000210, -0.00001264, 0.00000496, -0.00000448, 0.00002236, -0.00001658, -0.00000582, 0.00000266, 0.00001424, -0.00001061, 0.00001071, -0.00000440, 0.00000116, -0.00000112, 0.00000426, -0.00000536, 0.00000512, -0.00000140, -0.00000475, 0.00000082, 0.00000351, -0.00000286, 0.00000738, -0.00000640, -0.00000219, 0.00000272, -0.00000199, -0.00000816, -0.00000427, 0.00000284, 0.00000635, -0.00000320, 0.00000138, -0.00000302, -0.00000001, -0.00001995, -0.00001934, -0.00000032, 0.00002817, 0.00003140, -0.00000080, 0.00004092, 0.00001178, 0.00000159, -0.00003227, 0.00000252, },
		{ 0.00000152, -0.00000891, 0.00000315, -0.00000579, -0.00000396, 0.00000150, -0.00000081, -0.00000456, 0.00000818, -0.00000548, 0.00000363, -0.00000728, -0.00000261, -0.00000237, 0.00000464, 0.00000239, -0.00000543, -0.00000317, -0.00000215, 0.00000418, -0.00000620, 0.00001054, -0.00001675, 0.00000457, -0.00000640, 0.00000925, -0.00000817, 0.00001329, -0.00002393, -0.00000242, -0.00001474, 0.00000684, -0.00003145, 0.00001963, -0.00012535, -0.00002278, 0.00021383, 0.00020757, -0.00004261, -0.00010583, 0.00000845, -0.00004451, 0.00001152, -0.00002099, 0.00001697, -0.00001111, 0.00000612, -0.00001312, 0.00000417, -0.00001131, 0.00000721, -0.00000867, 0.00000645, -0.00000727, 0.00000204, -0.00000466, 0.00000912, -0.00000409, 0.00000287, -0.00001030, 0.00000732, -0.00000769, 0.00001018, -0.00000009, 0.00000189, -0.00000810, 0.00000592, -0.00000781, 0.00001656, -0.00001843, 0.00001878, -0.00000034, -0.00000095, -0.00000293, -0.00000742, 0.00001636, -0.00002089, -0.00000608, 0.00000439, -0.00000229, -0.00000369, -0.00001871, 0.00001445, -0.00001227, 0.00000015, -0.00000432, 0.00000845, 0.00001264, -0.00001947, -0.00000161, 0.00000788, 0.00000771, -0.00000869, 0.00001251, -0.00000902, 0.00000505, -0.00000301, 0.00000525, -0.00000737, 0.00000943, -0.00000790, -0.00000022, 0.00000008, 0.00000362, -0.00000289, 0.00000853, -0.00001175, 0.00000459, -0.00000193, -0.00000086, -0.00001093, 0.00000065, 0.00000233, 0.00000727, -0.00000710, 0.00000549, -0.00000783, 0.00000253, -0.00003160, -0.00000353, -0.00000129, 0.00004433, 0.00000720, 0.00001651, 0.00003177, 0.00001096, -0.00001234, -0.00002311, },
		{ -0.00001013, -0.00000366, -0.00000331, -0.00000063, -0.00000412, -0.00000439, 0.00000490, -0.00000722, 0.00000337, 0.00000256, -0.00000259, 0.00000123, -0.00000777, -0.00000097, -0.00000213, 0.00000729, -0.00000241, -0.00000352, -0.00000564, 0.00000378, -0.00000267, 0.00000198, 0.00000208, -0.00001348, 0.00000414, -0.00000318, 0.00000503, -0.00000270, 0.00000636, -0.00002519, -0.00000069, -0.00001110, -0.00000146, -0.00002131, -0.00000057, -0.00015322, 0.00007077, 0.00025191, 0.00015079, -0.00012939, -0.00003893, -0.00001961, -0.00002016, -0.00000190, -0.00000718, 0.00000798, -0.00000478, -0.00000211, -0.00000585, -0.00000244, -0.00000428, 0.00000187, -0.00000366, 0.00000202, -0.00000410, -0.00000108, 0.00000069, 0.00000475, -0.00000089, -0.00000325, -0.00000401, 0.00000298, -0.00000347, 0.00000909, 0.00000051, -0.00000321, -0.00000212, 0.00000014, 0.00000033, 0.00000860, -0.00001545, 0.00002543, -0.00001168, 0.00000706, -0.00001270, 0.00000569, 0.00000658, -0.00002411, 0.00000333, 0.00000002, 0.00000051, -0.00001284, -0.00000795, 0.00001169, -0.00001678, 0.00000793, -0.00001124, 0.00002231, -0.00000451, -0.00001350, -0.00000085, 0.00001430, -0.00000338, 0.00000096, 0.00000567, -0.00000621, 0.00000360, -0.00000048, 0.00000129, -0.00000333, 0.00000731, -0.00000960, 0.00000209, 0.00000091, 0.00000118, 0.00000107, 0.00000372, -0.00001049, 0.00000647, -0.00000379, -0.00000253, -0.00000945, 0.00000182, 0.00000574, 0.00000181, -0.00000341, 0.00000160, -0.00000417, -0.00000633, -0.00002938, 0.00000410, 0.00000599, 0.00004670, -0.00000895, 0.00003431, 0.00001798, 0.00001139, -0.00002530, },
		{ -0.00003030, -0.00000073, -0.00000728, 0.00000033, -0.00000318, -0.00000353, -0.00000207, 0.00000334, -0.00000853, 0.00000846, -0.00000287, 0.00000130, -0.00000316, -0.00000557, -0.00000119, 0.00000008, 0.00000722, -0.00000586, -0.00000228, -0.00000603, 0.00000743, -0.00000834, 0.00001076, -0.00001041, -0.00000385, -0.00000142, 0.00000375, -0.00000318, 0.00000805, -0.00000975, -0.00001553, -0.00000683, -0.00000150, -0.00001741, 0.00000062, -0.00005137, -0.00013391, 0.00016318, 0.00027061, 0.00004020, -0.00014468, 0.00000501, -0.00004433, 0.00000490, -0.00001824, 0.00000949, -0.00000460, 0.00000367, -0.00001171, 0.00000341, -0.00001106, 0.00000511, -0.00000612, 0.00000380, -0.00000508, 0.00000162, -0.00000567, 0.00000708, -0.00000095, 0.00000263, -0.00000932, 0.00000369, -0.00000368, 0.00000383, 0.00000502, 0.00000166, -0.00000774, 0.00000439, -0.00000692, 0.00001132, -0.00000635, -0.00000024, 0.00001815, -0.00001273, 0.00000777, -0.00001596, 0.00001623, -0.00000812, -0.00001867, 0.00000796, -0.00000324, 0.00000108, -0.00002017, 0.00000716, -0.00000203, -0.00000791, 0.00000325, -0.00000600, 0.00002512, -0.00001757, -0.00000610, 0.00000124, 0.00001580, -0.00001110, 0.00001073, -0.00000413, 0.00000049, -0.00000074, 0.00000390, -0.00000464, 0.00000404, 0.00000019, -0.00000645, 0.00000191, 0.00000241, -0.00000156, 0.00000582, -0.00000389, -0.00000454, 0.00000402, -0.00000344, -0.00000525, -0.00000631, 0.00000238, 0.00000873, -0.00000495, 0.00000306, -0.00000580, 0.00000326, -0.00002227, -0.00001387, 0.00000083, 0.00002364, 0.00003063, -0.00000655, 0.00004023, 0.00000913, 0.00000646, },
		{ -0.00000485, -0.00002575, 0.00000232, -0.00000737, 0.00000068, -0.00000344, -0.00000411, 0.00000189, -0.00000231, -0.00000372, 0.00000769, -0.00000446, 0.00000270, -0.00000653, -0.00000261, -0.00000235, 0.00000389, 0.00000376, -0.00000629, -0.00000341, -0.00000218, 0.00000494, -0.00000696, 0.00001218, -0.00001801, 0.00000424, -0.00000553, 0.00000758, -0.00000706, 0.00001376, -0.00002404, -0.00000407, -0.00001305, 0.00000454, -0.00002762, 0.00001453, -0.00011127, -0.00007092, 0.00025027, 0.00024038, -0.00008063, -0.00009464, 0.00000280, -0.00004072, 0.00000941, -0.00001966, 0.00001402, -0.00000806, 0.00000329, -0.00001215, 0.00000449, -0.00001217, 0.00000796, -0.00000860, 0.00000615, -0.00000779, 0.00000375, -0.00000642, 0.00000952, -0.00000342, 0.00000203, -0.00001007, 0.00000684, -0.00000675, 0.00000887, 0.00000202, -0.00000013, -0.00000697, 0.00000519, -0.00000765, 0.00001559, -0.00001707, 0.00001697, 0.00000295, -0.00000533, 0.00000099, -0.00001027, 0.00001845, -0.00002137, -0.00000704, 0.00000587, -0.00000219, -0.00000469, -0.00001781, 0.00001489, -0.00001470, 0.00000541, -0.00000876, 0.00000997, 0.00001356, -0.00001974, -0.00000280, 0.00000752, 0.00000891, -0.00000935, 0.00001286, -0.00000927, 0.00000487, -0.00000302, 0.00000545, -0.00000735, 0.00000912, -0.00000701, -0.00000155, 0.00000098, 0.00000279, -0.00000186, 0.00000733, -0.00000934, 0.00000163, 0.00000024, -0.00000294, -0.00000661, -0.00000446, 0.00000583, 0.00000612, -0.00000599, 0.00000423, -0.00000771, 0.00000363, -0.00003198, 0.00000247, -0.00000375, 0.00004136, 0.00000543, 0.00001118, 0.00003217, 0.00000801, },
		{ 0.00000997, -0.00001763, -0.00001567, 0.00000098, -0.00000570, -0.00000016, -0.00000326, -0.00000350, 0.00000367, -0.00000722, 0.00000323, 0.00000301, -0.00000242, 0.00000078, -0.00000711, -0.00000084, -0.00000268, 0.00000731, -0.00000148, -0.00000458, -0.00000561, 0.00000399, -0.00000231, 0.00000159, 0.00000371, -0.00001568, 0.00000536, -0.00000395, 0.00000462, -0.00000212, 0.00000751, -0.00002722, 0.00000044, -0.00001266, 0.00000058, -0.00002330, 0.00000351, -0.00015422, 0.00002574, 0.00032068, 0.00014189, -0.00015083, -0.00002884, -0.00002544, -0.00001594, -0.00000434, -0.00000639, 0.00000592, -0.00000205, -0.00000569, -0.00000327, -0.00000358, -0.00000427, 0.00000235, -0.00000358, 0.00000137, -0.00000387, 0.00000022, -0.00000173, 0.00000668, -0.00000185, -0.00000280, -0.00000498, 0.00000397, -0.00000416, 0.00000951, 0.00000106, -0.00000434, -0.00000141, -0.00000023, -0.00000024, 0.00000876, -0.00001530, 0.00002520, -0.00001021, 0.00000357, -0.00000827, 0.00000154, 0.00001055, -0.00002695, 0.00000471, 0.00000006, 0.00000129, -0.00001485, -0.00000482, 0.00000855, -0.00001491, 0.00001002, -0.00001504, 0.00002476, -0.00000440, -0.00001373, -0.00000229, 0.00001472, -0.00000262, 0.00000030, 0.00000616, -0.00000685, 0.00000391, -0.00000095, 0.00000217, -0.00000423, 0.00000814, -0.00001004, 0.00000194, 0.00000073, 0.00000161, 0.00000065, 0.00000429, -0.00001000, 0.00000494, -0.00000282, -0.00000278, -0.00000689, -0.00000307, 0.00001068, -0.00000128, -0.00000056, -0.00000183, -0.00000125, -0.00000804, -0.00002590, 0.00000732, 0.00000220, 0.00004497, -0.00001213, 0.00003062, 0.00001796, },
		{ 0.00000725, 0.00000868, -0.00002586, -0.00000539, -0.00000184, -0.00000421, -0.00000037, -0.00000400, -0.00000060, 0.00000104, -0.00000721, 0.00000727, -0.00000138, 0.00000034, -0.00000274, -0.00000521, -0.00000109, -0.00000072, 0.00000788, -0.00000552, -0.00000338, -0.00000552, 0.00000741, -0.00000801, 0.00001063, -0.00000912, -0.00000644, 0.00000103, 0.00000117, -0.00000162, 0.00000732, -0.00000769, -0.00001926, -0.00000296, -0.00000662, -0.00001030, -0.00000869, -0.00003445, -0.00016272, 0.00014924, 0.00034255, -0.00000172, -0.00014067, 0.00000330, -0.00004285, 0.00000413, -0.00001712, 0.00000673, -0.00000280, 0.00000285, -0.00001278, 0.00000497, -0.00001196, 0.00000496, -0.00000514, 0.00000310, -0.00000528, 0.00000223, -0.00000514, 0.00000488, 0.00000169, 0.00000044, -0.00000769, 0.00000161, -0.00000131, 0.00000144, 0.00000751, -0.00000013, -0.00000682, 0.00000355, -0.00000605, 0.00000932, -0.00000429, -0.00000217, 0.00002051, -0.00001473, 0.00000774, -0.00001389, 0.00001341, -0.00000468, -0.00002179, 0.00001018, -0.00000362, 0.00000130, -0.00002118, 0.00000994, -0.00000643, -0.00000263, 0.00000114, -0.00000704, 0.00002663, -0.00001707, -0.00000723, 0.00000052, 0.00001623, -0.00001037, 0.00000986, -0.00000345, -0.00000045, -0.00000005, 0.00000313, -0.00000333, 0.00000247, 0.00000194, -0.00000810, 0.00000296, 0.00000111, 0.00000021, 0.00000378, -0.00000118, -0.00000674, 0.00000505, -0.00000484, -0.00000213, -0.00000813, 0.00000133, 0.00001134, -0.00000660, 0.00000448, -0.00000818, 0.00000598, -0.00002409, -0.00000822, 0.00000061, 0.00002001, 0.00002956, -0.00001115, 0.00003903, },
		{ 0.00003328, 0.00000410, 0.00000168, -0.00002673, 0.00000198, -0.00000494, -0.00000262, -0.00000058, -0.00000437, 0.00000238, -0.00000395, -0.00000256, 0.00000663, -0.00000288, 0.00000126, -0.00000545, -0.00000251, -0.00000250, 0.00000327, 0.00000472, -0.00000658, -0.00000422, -0.00000140, 0.00000463, -0.00000650, 0.00001229, -0.00001744, 0.00000219, -0.00000306, 0.00000426, -0.00000418, 0.00001212, -0.00002195, -0.00000777, -0.00000886, -0.00000103, -0.00001946, 0.00000330, -0.00008854, -0.00012619, 0.00028642, 0.00027945, -0.00012678, -0.00007642, -0.00000802, -0.00003279, 0.00000396, -0.00001543, 0.00000846, -0.00000293, -0.00000127, -0.00000972, 0.00000377, -0.00001195, 0.00000728, -0.00000707, 0.00000457, -0.00000718, 0.00000442, -0.00000698, 0.00000863, -0.00000158, 0.00000009, -0.00000861, 0.00000502, -0.00000446, 0.00000630, 0.00000512, -0.00000311, -0.00000472, 0.00000324, -0.00000606, 0.00001286, -0.00001381, 0.00001362, 0.00000716, -0.00001048, 0.00000517, -0.00001223, 0.00001895, -0.00002049, -0.00000844, 0.00000728, -0.00000194, -0.00000595, -0.00001617, 0.00001443, -0.00001648, 0.00001034, -0.00001278, 0.00001099, 0.00001419, -0.00001903, -0.00000462, 0.00000758, 0.00000927, -0.00000877, 0.00001195, -0.00000863, 0.00000401, -0.00000242, 0.00000498, -0.00000639, 0.00000777, -0.00000532, -0.00000340, 0.00000223, 0.00000148, -0.00000006, 0.00000510, -0.00000599, -0.00000196, 0.00000277, -0.00000546, -0.00000148, -0.00001012, 0.00000936, 0.00000482, -0.00000445, 0.00000232, -0.00000667, 0.00000340, -0.00003053, 0.00000670, -0.00000574, 0.00003806, 0.00000466, 0.00000569, },
		{ 0.00002573, 0.00001959, 0.00000605, -0.00000896, -0.00002101, 0.00000538, -0.00000792, -0.00000003, -0.00000218, -0.00000249, 0.00000221, -0.00000691, 0.00000287, 0.00000333, -0.00000196, -0.00000013, -0.00000590, -0.00000097, -0.00000305, 0.00000701, -0.00000039, -0.00000564, -0.00000549, 0.00000420, -0.00000218, 0.00000155, 0.00000465, -0.00001671, 0.00000531, -0.00000349, 0.00000298, -0.00000026, 0.00000699, -0.00002730, -0.00000014, -0.00001234, 0.00000016, -0.00002188, 0.00000191, -0.00014461, -0.00003304, 0.00039933, 0.00012897, -0.00017248, -0.00001685, -0.00003312, -0.00000996, -0.00000828, -0.00000420, 0.00000243, 0.00000185, -0.00001020, 0.00000010, -0.00000506, -0.00000407, 0.00000236, -0.00000288, 0.00000015, -0.00000318, 0.00000119, -0.00000370, 0.00000800, -0.00000229, -0.00000278, -0.00000542, 0.00000426, -0.00000402, 0.00000904, 0.00000225, -0.00000596, -0.00000010, -0.00000128, 0.00000006, 0.00000757, -0.00001327, 0.00002291, -0.00000695, -0.00000178, -0.00000207, -0.00000337, 0.00001406, -0.00002876, 0.00000546, 0.00000030, 0.00000195, -0.00001676, -0.00000131, 0.00000459, -0.00001206, 0.00001126, -0.00001779, 0.00002574, -0.00000330, -0.00001408, -0.00000357, 0.00001466, -0.00000167, -0.00000002, 0.00000593, -0.00000688, 0.00000378, -0.00000110, 0.00000275, -0.00000455, 0.00000810, -0.00000962, 0.00000117, 0.00000089, 0.00000178, 0.00000071, 0.00000402, -0.00000851, 0.00000257, -0.00000132, -0.00000349, -0.00000348, -0.00000875, 0.00001583, -0.00000441, 0.00000245, -0.00000549, 0.00000207, -0.00001057, -0.00002078, 0.00000813, 0.00000001, 0.00004203, -0.00001340, },
		{ -0.00001420, 0.00003679, 0.00000674, 0.00000811, -0.00001931, -0.00001094, 0.00000409, -0.00000914, 0.00000252, -0.00000440, 0.00000098, -0.00000149, -0.00000549, 0.00000560, 0.00000036, -0.00000064, -0.00000257, -0.00000426, -0.00000148, -0.00000105, 0.00000781, -0.00000444, -0.00000515, -0.00000414, 0.00000637, -0.00000668, 0.00000940, -0.00000683, -0.00000958, 0.00000383, -0.00000179, 0.00000042, 0.00000595, -0.00000519, -0.00002315, 0.00000144, -0.00001261, -0.00000215, -0.00001915, -0.00001736, -0.00018646, 0.00012341, 0.00042745, -0.00005383, -0.00013004, -0.00000231, -0.00003861, 0.00000136, -0.00001433, 0.00000262, -0.00000008, 0.00000121, -0.00001304, 0.00000582, -0.00001186, 0.00000357, -0.00000312, 0.00000159, -0.00000474, 0.00000206, -0.00000375, 0.00000188, 0.00000490, -0.00000235, -0.00000535, -0.00000116, 0.00000162, -0.00000129, 0.00001011, -0.00000218, -0.00000541, 0.00000219, -0.00000462, 0.00000676, -0.00000204, -0.00000363, 0.00002180, -0.00001556, 0.00000614, -0.00000973, 0.00000884, -0.00000056, -0.00002453, 0.00001186, -0.00000364, 0.00000113, -0.00002154, 0.00001227, -0.00001083, 0.00000299, -0.00000138, -0.00000738, 0.00002662, -0.00001496, -0.00000922, 0.00000054, 0.00001553, -0.00000849, 0.00000821, -0.00000250, -0.00000150, 0.00000081, 0.00000207, -0.00000160, 0.00000064, 0.00000357, -0.00000940, 0.00000373, -0.00000019, 0.00000220, 0.00000155, 0.00000142, -0.00000850, 0.00000558, -0.00000598, 0.00000093, -0.00000955, -0.00000031, 0.00001401, -0.00000795, 0.00000548, -0.00000995, 0.00000796, -0.00002519, -0.00000267, -0.00000082, 0.00001748, 0.00002780, },
		{ 0.00000469, 0.00000031, 0.00003459, 0.00000010, 0.00000653, -0.00002579, 0.00000052, -0.00000187, -0.00000645, 0.00000247, -0.00000455, 0.00000277, -0.00000550, -0.00000130, 0.00000521, -0.00000098, -0.00000041, -0.00000432, -0.00000205, -0.00000307, 0.00000304, 0.00000497, -0.00000601, -0.00000583, 0.00000042, 0.00000298, -0.00000454, 0.00001053, -0.00001475, -0.00000177, 0.00000116, -0.00000081, 0.00000061, 0.00000815, -0.00001739, -0.00001374, -0.00000196, -0.00001007, -0.00000689, -0.00001394, -0.00005804, -0.00018652, 0.00032009, 0.00032566, -0.00018117, -0.00005093, -0.00002432, -0.00002042, -0.00000508, -0.00000808, 0.00000016, 0.00000436, -0.00000767, -0.00000557, 0.00000164, -0.00001022, 0.00000481, -0.00000382, 0.00000147, -0.00000516, 0.00000370, -0.00000607, 0.00000627, 0.00000150, -0.00000303, -0.00000583, 0.00000181, -0.00000082, 0.00000255, 0.00000905, -0.00000690, -0.00000141, 0.00000012, -0.00000310, 0.00000848, -0.00000891, 0.00000903, 0.00001196, -0.00001596, 0.00000918, -0.00001300, 0.00001775, -0.00001832, -0.00001010, 0.00000849, -0.00000146, -0.00000754, -0.00001375, 0.00001297, -0.00001735, 0.00001459, -0.00001618, 0.00001153, 0.00001436, -0.00001723, -0.00000711, 0.00000813, 0.00000868, -0.00000690, 0.00000978, -0.00000712, 0.00000254, -0.00000126, 0.00000387, -0.00000453, 0.00000549, -0.00000298, -0.00000555, 0.00000365, -0.00000011, 0.00000232, 0.00000213, -0.00000203, -0.00000581, 0.00000532, -0.00000809, 0.00000402, -0.00001584, 0.00001254, 0.00000356, -0.00000257, -0.00000015, -0.00000474, 0.00000183, -0.00002722, 0.00000894, -0.00000694, 0.00003433, },
		{ 0.00003767, -0.00001268, 0.00001964, 0.00002266, 0.00000036, -0.00000020, -0.00002543, 0.00000904, -0.00000969, -0.00000031, -0.00000087, -0.00000141, 0.00000056, -0.00000628, 0.00000229, 0.00000354, -0.00000124, -0.00000139, -0.00000429, -0.00000120, -0.00000339, 0.00000658, 0.00000060, -0.00000642, -0.00000553, 0.00000467, -0.00000255, 0.00000216, 0.00000450, -0.00001613, 0.00000360, -0.00000146, -0.00000018, 0.00000319, 0.00000439, -0.00002494, -0.00000301, -0.00000946, -0.00000349, -0.00001613, -0.00000637, -0.00012380, -0.00010464, 0.00048587, 0.00011337, -0.00019476, -0.00000266, -0.00004296, -0.00000193, -0.00001399, -0.00000036, -0.00000265, 0.00000696, -0.00001565, 0.00000433, -0.00000708, -0.00000336, 0.00000155, -0.00000126, -0.00000189, -0.00000173, 0.00000142, -0.00000481, 0.00000834, -0.00000192, -0.00000342, -0.00000507, 0.00000360, -0.00000280, 0.00000751, 0.00000415, -0.00000809, 0.00000186, -0.00000307, 0.00000135, 0.00000492, -0.00000935, 0.00001867, -0.00000211, -0.00000862, 0.00000536, -0.00000851, 0.00001672, -0.00002930, 0.00000550, 0.00000079, 0.00000244, -0.00001845, 0.00000242, -0.00000008, -0.00000823, 0.00001142, -0.00001925, 0.00002516, -0.00000131, -0.00001441, -0.00000474, 0.00001425, -0.00000071, 0.00000018, 0.00000487, -0.00000620, 0.00000313, -0.00000085, 0.00000290, -0.00000416, 0.00000709, -0.00000828, -0.00000023, 0.00000140, 0.00000163, 0.00000130, 0.00000294, -0.00000612, -0.00000048, 0.00000055, -0.00000448, 0.00000047, -0.00001477, 0.00002069, -0.00000723, 0.00000532, -0.00000911, 0.00000558, -0.00001370, -0.00001426, 0.00000656, -0.00000033, },
		{ 0.00001613, 0.00002508, -0.00001544, 0.00003320, 0.00000787, 0.00000495, -0.00001130, -0.00001666, 0.00000983, -0.00001387, 0.00000511, -0.00000448, 0.00000242, -0.00000399, -0.00000357, 0.00000365, 0.00000210, -0.00000144, -0.00000279, -0.00000265, -0.00000244, -0.00000084, 0.00000695, -0.00000262, -0.00000750, -0.00000197, 0.00000442, -0.00000444, 0.00000717, -0.00000375, -0.00001297, 0.00000666, -0.00000482, 0.00000267, 0.00000415, -0.00000254, -0.00002682, 0.00000587, -0.00001881, 0.00000616, -0.00002957, -0.00000181, -0.00020318, 0.00008494, 0.00052434, -0.00011499, -0.00011333, -0.00001155, -0.00003183, -0.00000324, -0.00001006, -0.00000266, 0.00000343, -0.00000123, -0.00001237, 0.00000584, -0.00001069, 0.00000098, -0.00000020, -0.00000062, -0.00000352, 0.00000123, -0.00000168, -0.00000163, 0.00000838, -0.00000543, -0.00000254, -0.00000433, 0.00000476, -0.00000401, 0.00001251, -0.00000424, -0.00000370, 0.00000050, -0.00000282, 0.00000392, 0.00000008, -0.00000433, 0.00002181, -0.00001508, 0.00000300, -0.00000374, 0.00000295, 0.00000380, -0.00002653, 0.00001277, -0.00000315, 0.00000046, -0.00002112, 0.00001394, -0.00001491, 0.00000870, -0.00000431, -0.00000688, 0.00002501, -0.00001134, -0.00001191, 0.00000120, 0.00001380, -0.00000566, 0.00000602, -0.00000151, -0.00000248, 0.00000166, 0.00000090, 0.00000033, -0.00000116, 0.00000480, -0.00001011, 0.00000401, -0.00000124, 0.00000413, -0.00000057, 0.00000359, -0.00000958, 0.00000543, -0.00000667, 0.00000372, -0.00001041, -0.00000246, 0.00001651, -0.00000879, 0.00000590, -0.00001098, 0.00000904, -0.00002540, 0.00000247, -0.00000322, },
		{ -0.00000717, 0.00003016, 0.00000531, -0.00000461, 0.00003556, -0.00000327, 0.00000924, -0.00002292, -0.00000182, 0.00000148, -0.00001033, 0.00000535, -0.00000436, 0.00000280, -0.00000671, -0.00000016, 0.00000368, 0.00000093, -0.00000198, -0.00000346, -0.00000099, -0.00000427, 0.00000347, 0.00000425, -0.00000435, -0.00000842, 0.00000341, -0.00000013, -0.00000092, 0.00000673, -0.00000985, -0.00000762, 0.00000704, -0.00000754, 0.00000727, 0.00000182, -0.00001035, -0.00002196, 0.00000759, -0.00002240, 0.00000974, -0.00003649, -0.00002129, -0.00024931, 0.00034902, 0.00037954, -0.00024341, -0.00001845, -0.00004602, -0.00000366, -0.00001771, 0.00000237, -0.00001080, 0.00001371, -0.00001590, 0.00000040, -0.00000215, -0.00000669, 0.00000032, 0.00000127, -0.00000327, -0.00000156, 0.00000143, -0.00000355, 0.00000240, 0.00000576, -0.00000722, -0.00000181, -0.00000267, 0.00000398, -0.00000214, 0.00001355, -0.00001125, 0.00000273, -0.00000397, 0.00000101, 0.00000276, -0.00000274, 0.00000366, 0.00001687, -0.00002127, 0.00001255, -0.00001239, 0.00001492, -0.00001508, -0.00001180, 0.00000931, -0.00000064, -0.00000947, -0.00001051, 0.00001042, -0.00001711, 0.00001787, -0.00001878, 0.00001163, 0.00001396, -0.00001431, -0.00001019, 0.00000912, 0.00000718, -0.00000386, 0.00000651, -0.00000493, 0.00000062, 0.00000032, 0.00000223, -0.00000195, 0.00000250, -0.00000027, -0.00000774, 0.00000498, -0.00000176, 0.00000497, -0.00000122, 0.00000211, -0.00000951, 0.00000752, -0.00001042, 0.00000938, -0.00002110, 0.00001501, 0.00000253, -0.00000047, -0.00000303, -0.00000206, -0.00000091, -0.00002219, 0.00000915, },
		{ 0.00000287, 0.00000118, 0.00003184, -0.00001011, 0.00001259, 0.00002670, -0.00000622, 0.00000772, -0.00002831, 0.00001159, -0.00001083, -0.00000098, 0.00000059, -0.00000022, -0.00000132, -0.00000531, 0.00000146, 0.00000371, -0.00000043, -0.00000279, -0.00000255, -0.00000130, -0.00000390, 0.00000627, 0.00000120, -0.00000662, -0.00000604, 0.00000567, -0.00000366, 0.00000370, 0.00000293, -0.00001363, -0.00000003, 0.00000233, -0.00000502, 0.00000840, -0.00000054, -0.00001979, -0.00000857, -0.00000358, -0.00001087, -0.00000556, -0.00002175, -0.00009212, -0.00018710, 0.00057747, 0.00009684, -0.00021827, 0.00001399, -0.00005519, 0.00000834, -0.00002167, 0.00000532, -0.00000942, 0.00001328, -0.00002201, 0.00000947, -0.00000983, -0.00000183, -0.00000040, 0.00000150, -0.00000496, 0.00000074, 0.00000059, -0.00000469, 0.00000738, -0.00000051, -0.00000493, -0.00000371, 0.00000178, -0.00000040, 0.00000489, 0.00000674, -0.00001068, 0.00000441, -0.00000559, 0.00000359, 0.00000091, -0.00000377, 0.00001280, 0.00000388, -0.00001638, 0.00001334, -0.00001330, 0.00001819, -0.00002845, 0.00000481, 0.00000151, 0.00000273, -0.00001984, 0.00000623, -0.00000527, -0.00000349, 0.00001039, -0.00001930, 0.00002309, 0.00000136, -0.00001455, -0.00000590, 0.00001358, 0.00000012, 0.00000097, 0.00000295, -0.00000480, 0.00000198, -0.00000018, 0.00000258, -0.00000300, 0.00000510, -0.00000609, -0.00000217, 0.00000222, 0.00000120, 0.00000234, 0.00000116, -0.00000304, -0.00000393, 0.00000254, -0.00000549, 0.00000458, -0.00002057, 0.00002475, -0.00000936, 0.00000780, -0.00001244, 0.00000899, -0.00001712, -0.00000671, },
		{ 0.00000696, -0.00000630, 0.00001589, 0.00002130, -0.00001480, 0.00002817, 0.00001060, -0.00000025, -0.00000266, -0.00002175, 0.00001466, -0.00001784, 0.00000700, -0.00000400, 0.00000350, -0.00000625, -0.00000167, 0.00000168, 0.00000364, -0.00000189, -0.00000348, -0.00000042, -0.00000390, -0.00000010, 0.00000538, -0.00000021, -0.00001025, 0.00000077, 0.00000176, -0.00000149, 0.00000419, -0.00000023, -0.00001622, 0.00000914, -0.00000758, 0.00000484, 0.00000221, -0.00000006, -0.00002981, 0.00000977, -0.00002451, 0.00001373, -0.00003874, 0.00001059, -0.00021139, 0.00003396, 0.00063128, -0.00018320, -0.00009168, -0.00002373, -0.00002303, -0.00000924, -0.00000469, -0.00000874, 0.00000745, -0.00000435, -0.00001080, 0.00000498, -0.00000850, -0.00000260, 0.00000335, -0.00000330, -0.00000181, -0.00000006, 0.00000077, -0.00000528, 0.00001171, -0.00000845, 0.00000040, -0.00000754, 0.00000771, -0.00000634, 0.00001436, -0.00000604, -0.00000190, -0.00000131, -0.00000090, 0.00000108, 0.00000181, -0.00000409, 0.00002047, -0.00001332, -0.00000148, 0.00000356, -0.00000365, 0.00000786, -0.00002743, 0.00001273, -0.00000205, -0.00000080, -0.00001980, 0.00001476, -0.00001841, 0.00001423, -0.00000755, -0.00000551, 0.00002188, -0.00000645, -0.00001501, 0.00000227, 0.00001128, -0.00000217, 0.00000358, -0.00000071, -0.00000317, 0.00000233, -0.00000021, 0.00000223, -0.00000268, 0.00000539, -0.00001003, 0.00000362, -0.00000188, 0.00000576, -0.00000232, 0.00000511, -0.00000980, 0.00000453, -0.00000682, 0.00000610, -0.00001069, -0.00000494, 0.00001858, -0.00000892, 0.00000563, -0.00001117, 0.00000916, -0.00002459, },
		{ -0.00001571, 0.00000746, -0.00000637, 0.00002564, 0.00000623, -0.00000873, 0.00003572, -0.00000544, 0.00000964, -0.00001837, -0.00000463, 0.00000461, -0.00001377, 0.00000763, -0.00000351, 0.00000223, -0.00000732, 0.00000057, 0.00000236, 0.00000250, -0.00000309, -0.00000317, 0.00000088, -0.00000627, 0.00000474, 0.00000240, -0.00000151, -0.00001203, 0.00000757, -0.00000466, 0.00000431, 0.00000097, -0.00000289, -0.00001513, 0.00001432, -0.00001565, 0.00001553, -0.00000667, -0.00000102, -0.00003218, 0.00001942, -0.00003750, 0.00002966, -0.00006314, 0.00001964, -0.00031167, 0.00037109, 0.00044112, -0.00031248, 0.00002021, -0.00007255, 0.00001704, -0.00003356, 0.00001559, -0.00002408, 0.00002484, -0.00002577, 0.00000815, -0.00000767, -0.00000126, -0.00000618, 0.00000811, -0.00000955, 0.00000357, -0.00000239, 0.00000051, -0.00000280, 0.00001097, -0.00001226, 0.00000322, -0.00000817, 0.00000961, -0.00000738, 0.00001822, -0.00001580, 0.00000737, -0.00000871, 0.00000593, -0.00000386, 0.00000415, -0.00000197, 0.00002138, -0.00002592, 0.00001492, -0.00001035, 0.00001071, -0.00001110, -0.00001323, 0.00000958, 0.00000060, -0.00001175, -0.00000651, 0.00000680, -0.00001563, 0.00001994, -0.00002047, 0.00001136, 0.00001292, -0.00001042, -0.00001366, 0.00001037, 0.00000497, 0.00000010, 0.00000247, -0.00000233, -0.00000150, 0.00000208, 0.00000028, 0.00000110, -0.00000086, 0.00000248, -0.00000964, 0.00000595, -0.00000318, 0.00000755, -0.00000454, 0.00000599, -0.00001261, 0.00000900, -0.00001210, 0.00001415, -0.00002542, 0.00001646, 0.00000184, 0.00000172, -0.00000617, 0.00000116, -0.00000460, },
		{ -0.00002046, 0.00000138, -0.00000252, 0.00000434, 0.00002473, -0.00000602, 0.00000500, 0.00003107, -0.00001276, 0.00001400, -0.00002920, 0.00001278, -0.00001128, -0.00000194, 0.00000205, 0.00000115, -0.00000345, -0.00000392, 0.00000031, 0.00000396, 0.00000028, -0.00000403, -0.00000100, -0.00000100, -0.00000483, 0.00000635, 0.00000111, -0.00000594, -0.00000725, 0.00000740, -0.00000571, 0.00000638, -0.00000027, -0.00000904, -0.00000565, 0.00000789, -0.00001152, 0.00001536, -0.00000786, -0.00001177, -0.00001693, 0.00000546, -0.00002211, 0.00000985, -0.00004397, -0.00005077, -0.00027752, 0.00067064, 0.00008143, -0.00024360, 0.00003329, -0.00006994, 0.00002095, -0.00003139, 0.00001287, -0.00001784, 0.00002074, -0.00002920, 0.00001553, -0.00001348, 0.00000080, -0.00000374, 0.00000558, -0.00000920, 0.00000440, -0.00000154, -0.00000311, 0.00000495, 0.00000205, -0.00000736, -0.00000128, -0.00000124, 0.00000316, 0.00000125, 0.00000987, -0.00001358, 0.00000741, -0.00000870, 0.00000664, -0.00000420, 0.00000308, 0.00000580, 0.00001045, -0.00002437, 0.00002109, -0.00001716, 0.00001820, -0.00002616, 0.00000347, 0.00000241, 0.00000284, -0.00002085, 0.00000995, -0.00001076, 0.00000201, 0.00000816, -0.00001795, 0.00001971, 0.00000443, -0.00001433, -0.00000711, 0.00001273, 0.00000075, 0.00000234, 0.00000026, -0.00000280, 0.00000042, 0.00000083, 0.00000182, -0.00000115, 0.00000229, -0.00000322, -0.00000447, 0.00000319, 0.00000060, 0.00000368, -0.00000108, 0.00000039, -0.00000744, 0.00000434, -0.00000624, 0.00000842, -0.00002560, 0.00002753, -0.00001052, 0.00000965, -0.00001520, 0.00001201, },
		{ 0.00000828, -0.00002272, 0.00001054, -0.00000973, 0.00001660, 0.00001650, -0.00001238, 0.00002185, 0.00001459, -0.00000673, 0.00000565, -0.00002545, 0.00001798, -0.00002056, 0.00000793, -0.00000280, 0.00000404, -0.00000806, 0.00000001, -0.00000015, 0.00000480, -0.00000192, -0.00000460, 0.00000231, -0.00000573, 0.00000102, 0.00000328, 0.00000254, -0.00001310, 0.00000375, -0.00000128, 0.00000182, 0.00000082, 0.00000333, -0.00001891, 0.00001091, -0.00000973, 0.00000664, 0.00000042, 0.00000188, -0.00003170, 0.00001260, -0.00002899, 0.00001966, -0.00004558, 0.00001857, -0.00021026, -0.00002845, 0.00074531, -0.00025566, -0.00006678, -0.00003779, -0.00001303, -0.00001600, 0.00000125, -0.00001513, 0.00001162, -0.00000789, -0.00000847, 0.00000338, -0.00000551, -0.00000681, 0.00000714, -0.00000611, 0.00000010, -0.00000151, 0.00000325, -0.00000864, 0.00001448, -0.00001102, 0.00000312, -0.00001041, 0.00001010, -0.00000795, 0.00001539, -0.00000735, -0.00000021, -0.00000304, 0.00000092, -0.00000152, 0.00000293, -0.00000283, 0.00001782, -0.00001045, -0.00000689, 0.00001149, -0.00001022, 0.00001110, -0.00002694, 0.00001159, -0.00000027, -0.00000265, -0.00001754, 0.00001459, -0.00002105, 0.00001928, -0.00001093, -0.00000338, 0.00001746, -0.00000070, -0.00001814, 0.00000347, 0.00000830, 0.00000163, 0.00000118, -0.00000030, -0.00000342, 0.00000266, -0.00000107, 0.00000388, -0.00000370, 0.00000516, -0.00000904, 0.00000251, -0.00000198, 0.00000692, -0.00000350, 0.00000582, -0.00000912, 0.00000287, -0.00000643, 0.00000801, -0.00001040, -0.00000753, 0.00001994, -0.00000821, 0.00000461, -0.00001048, },
		{ 0.00000467, -0.00000889, -0.00000817, 0.00000413, -0.00000463, 0.00002087, 0.00000717, -0.00001177, 0.00003472, -0.00000612, 0.00000785, -0.00001263, -0.00000739, 0.00000699, -0.00001623, 0.00000893, -0.00000177, 0.00000084, -0.00000710, 0.00000067, 0.00000153, 0.00000341, -0.00000344, -0.00000368, 0.00000367, -0.00000913, 0.00000691, -0.00000062, 0.00000249, -0.00001654, 0.00001272, -0.00001039, 0.00001092, -0.00000649, 0.00000575, -0.00002386, 0.00002252, -0.00002469, 0.00002498, -0.00001689, 0.00001012, -0.00004386, 0.00003291, -0.00005451, 0.00005171, -0.00009228, 0.00006234, -0.00037061, 0.00038453, 0.00050981, -0.00038670, 0.00006361, -0.00010285, 0.00004085, -0.00005194, 0.00003098, -0.00003911, 0.00003723, -0.00003688, 0.00001745, -0.00001482, 0.00000596, -0.00001447, 0.00001638, -0.00001708, 0.00001000, -0.00000754, 0.00000587, -0.00000902, 0.00001675, -0.00001776, 0.00000891, -0.00001427, 0.00001561, -0.00001273, 0.00002263, -0.00002014, 0.00001213, -0.00001369, 0.00001120, -0.00001086, 0.00001119, -0.00000729, 0.00002503, -0.00002949, 0.00001606, -0.00000703, 0.00000551, -0.00000684, -0.00001406, 0.00000911, 0.00000231, -0.00001436, -0.00000181, 0.00000217, -0.00001287, 0.00002066, -0.00002117, 0.00001074, 0.00001131, -0.00000579, -0.00001719, 0.00001162, 0.00000235, 0.00000457, -0.00000195, 0.00000032, -0.00000349, 0.00000376, -0.00000169, 0.00000429, -0.00000423, 0.00000491, -0.00001095, 0.00000631, -0.00000412, 0.00000974, -0.00000744, 0.00000918, -0.00001476, 0.00000949, -0.00001284, 0.00001793, -0.00002841, 0.00001671, 0.00000156, 0.00000391, -0.00000935, },
		{ -0.00001719, 0.00001437, -0.00002337, 0.00000949, -0.00000904, 0.00000876, 0.00001668, -0.00000090, -0.00000258, 0.00003509, -0.00001840, 0.00001804, -0.00002790, 0.00001259, -0.00001114, -0.00000298, 0.00000331, 0.00000281, -0.00000591, -0.00000206, -0.00000124, 0.00000445, 0.00000065, -0.00000480, 0.00000004, -0.00000005, -0.00000637, 0.00000706, 0.00000008, -0.00000417, -0.00000935, 0.00001000, -0.00000878, 0.00001027, -0.00000516, -0.00000240, -0.00001313, 0.00001505, -0.00001946, 0.00002385, -0.00001737, -0.00000104, -0.00002794, 0.00001748, -0.00003690, 0.00002961, -0.00007210, -0.00000181, -0.00037219, 0.00076145, 0.00006931, -0.00027124, 0.00005521, -0.00008711, 0.00003577, -0.00004306, 0.00002221, -0.00002776, 0.00002913, -0.00003707, 0.00002246, -0.00001814, 0.00000472, -0.00000860, 0.00001099, -0.00001460, 0.00000929, -0.00000504, 0.00000002, 0.00000102, 0.00000572, -0.00001068, 0.00000218, -0.00000536, 0.00000772, -0.00000318, 0.00001331, -0.00001656, 0.00001065, -0.00001215, 0.00001021, -0.00001002, 0.00001065, -0.00000168, 0.00001693, -0.00003181, 0.00002784, -0.00001960, 0.00001663, -0.00002255, 0.00000163, 0.00000340, 0.00000278, -0.00002142, 0.00001339, -0.00001627, 0.00000804, 0.00000484, -0.00001535, 0.00001532, 0.00000760, -0.00001364, -0.00000836, 0.00001171, 0.00000118, 0.00000420, -0.00000300, -0.00000040, -0.00000137, 0.00000204, 0.00000076, 0.00000120, -0.00000110, 0.00000004, -0.00000686, 0.00000412, 0.00000001, 0.00000512, -0.00000347, 0.00000382, -0.00001062, 0.00000564, -0.00000643, 0.00001158, -0.00002936, 0.00002868, -0.00001048, 0.00001070, },
		{ 0.00000285, -0.00000896, 0.00000643, -0.00001977, 0.00001302, -0.00001313, 0.00001799, 0.00001088, -0.00000849, 0.00001461, 0.00001928, -0.00001356, 0.00001275, -0.00002712, 0.00001930, -0.00002169, 0.00000773, -0.00000083, 0.00000397, -0.00000932, 0.00000137, -0.00000168, 0.00000550, -0.00000152, -0.00000603, 0.00000531, -0.00000770, 0.00000232, 0.00000091, 0.00000531, -0.00001568, 0.00000661, -0.00000431, 0.00000513, -0.00000254, 0.00000647, -0.00002064, 0.00001162, -0.00001098, 0.00000780, -0.00000095, 0.00000298, -0.00003212, 0.00001389, -0.00003168, 0.00002325, -0.00004927, 0.00002129, -0.00019973, -0.00010023, 0.00086258, -0.00032894, -0.00004073, -0.00005237, -0.00000281, -0.00002276, 0.00000708, -0.00002126, 0.00001546, -0.00001153, -0.00000563, 0.00000127, -0.00000207, -0.00001118, 0.00001071, -0.00000868, 0.00000189, -0.00000282, 0.00000538, -0.00001130, 0.00001630, -0.00001280, 0.00000528, -0.00001259, 0.00001160, -0.00000857, 0.00001541, -0.00000802, 0.00000119, -0.00000451, 0.00000246, -0.00000368, 0.00000334, -0.00000061, 0.00001410, -0.00000681, -0.00001267, 0.00001925, -0.00001601, 0.00001300, -0.00002482, 0.00000934, 0.00000216, -0.00000507, -0.00001435, 0.00001338, -0.00002263, 0.00002356, -0.00001426, -0.00000068, 0.00001213, 0.00000540, -0.00002086, 0.00000447, 0.00000517, 0.00000538, -0.00000089, -0.00000045, -0.00000311, 0.00000257, -0.00000159, 0.00000512, -0.00000408, 0.00000406, -0.00000716, 0.00000069, -0.00000153, 0.00000754, -0.00000404, 0.00000570, -0.00000760, 0.00000060, -0.00000556, 0.00000945, -0.00000968, -0.00000994, 0.00002034, -0.00000654, },
		{ 0.00000604, -0.00001240, 0.00000817, -0.00001342, -0.00000002, -0.00000049, -0.00000209, 0.00001601, 0.00000793, -0.00001353, 0.00003239, -0.00000528, 0.00000429, -0.00000635, -0.00000952, 0.00000811, -0.00001726, 0.00000893, 0.00000104, -0.00000151, -0.00000589, -0.00000005, 0.00000143, 0.00000341, -0.00000282, -0.00000510, 0.00000736, -0.00001279, 0.00000991, -0.00000470, 0.00000745, -0.00002170, 0.00001855, -0.00001697, 0.00001849, -0.00001516, 0.00001551, -0.00003318, 0.00003107, -0.00003406, 0.00003501, -0.00002823, 0.00002244, -0.00005629, 0.00004718, -0.00007236, 0.00007450, -0.00012204, 0.00010425, -0.00042332, 0.00038809, 0.00058430, -0.00046372, 0.00010980, -0.00013544, 0.00006655, -0.00007186, 0.00004768, -0.00005509, 0.00005018, -0.00004869, 0.00002789, -0.00002329, 0.00001464, -0.00002409, 0.00002558, -0.00002539, 0.00001732, -0.00001365, 0.00001211, -0.00001580, 0.00002265, -0.00002329, 0.00001479, -0.00002048, 0.00002146, -0.00001769, 0.00002633, -0.00002386, 0.00001659, -0.00001848, 0.00001635, -0.00001766, 0.00001778, -0.00001181, 0.00002741, -0.00003166, 0.00001586, -0.00000273, -0.00000012, -0.00000280, -0.00001397, 0.00000777, 0.00000453, -0.00001721, 0.00000343, -0.00000329, -0.00000889, 0.00001995, -0.00002086, 0.00000978, 0.00000926, -0.00000078, -0.00002037, 0.00001253, -0.00000032, 0.00000910, -0.00000628, 0.00000264, -0.00000507, 0.00000508, -0.00000341, 0.00000725, -0.00000722, 0.00000668, -0.00001142, 0.00000589, -0.00000437, 0.00001125, -0.00000958, 0.00001137, -0.00001570, 0.00000882, -0.00001245, 0.00002044, -0.00002985, 0.00001570, 0.00000167, },
		{ -0.00000918, 0.00001088, -0.00001826, 0.00001583, -0.00002549, 0.00001706, -0.00001608, 0.00001394, 0.00000815, 0.00000472, -0.00000958, 0.00003809, -0.00002247, 0.00001953, -0.00002449, 0.00001119, -0.00001062, -0.00000382, 0.00000414, 0.00000487, -0.00000876, 0.00000034, -0.00000328, 0.00000531, 0.00000044, -0.00000481, 0.00000028, 0.00000175, -0.00000869, 0.00000855, -0.00000205, -0.00000118, -0.00001241, 0.00001348, -0.00001285, 0.00001528, -0.00001160, 0.00000602, -0.00002209, 0.00002340, -0.00002843, 0.00003348, -0.00002865, 0.00001193, -0.00004111, 0.00003192, -0.00005455, 0.00005271, -0.00010457, 0.00005202, -0.00046688, 0.00084575, 0.00006249, -0.00030145, 0.00007949, -0.00010636, 0.00005246, -0.00005637, 0.00003306, -0.00003889, 0.00003818, -0.00004538, 0.00003012, -0.00002378, 0.00000996, -0.00001491, 0.00001759, -0.00002101, 0.00001528, -0.00000983, 0.00000459, -0.00000426, 0.00001033, -0.00001470, 0.00000649, -0.00001034, 0.00001296, -0.00000806, 0.00001675, -0.00001933, 0.00001384, -0.00001563, 0.00001392, -0.00001604, 0.00001829, -0.00000895, 0.00002261, -0.00003795, 0.00003289, -0.00002030, 0.00001353, -0.00001786, -0.00000046, 0.00000434, 0.00000262, -0.00002149, 0.00001640, -0.00002153, 0.00001429, 0.00000065, -0.00001173, 0.00001023, 0.00001059, -0.00001240, -0.00000959, 0.00001049, 0.00000148, 0.00000631, -0.00000653, 0.00000211, -0.00000314, 0.00000324, -0.00000041, 0.00000380, -0.00000471, 0.00000330, -0.00000903, 0.00000479, -0.00000039, 0.00000639, -0.00000569, 0.00000684, -0.00001310, 0.00000617, -0.00000585, 0.00001371, -0.00003148, 0.00002794, },
		{ 0.00001958, -0.00000390, 0.00000045, -0.00000670, 0.00000373, -0.00001581, 0.00001426, -0.00001616, 0.00001970, 0.00000478, -0.00000356, 0.00000698, 0.00002399, -0.00001984, 0.00001785, -0.00002635, 0.00001839, -0.00002109, 0.00000640, 0.00000184, 0.00000331, -0.00001000, 0.00000234, -0.00000286, 0.00000571, -0.00000078, -0.00000756, 0.00000827, -0.00000953, 0.00000353, -0.00000142, 0.00000774, -0.00001763, 0.00000894, -0.00000695, 0.00000803, -0.00000548, 0.00000881, -0.00002107, 0.00001105, -0.00001114, 0.00000813, -0.00000170, 0.00000303, -0.00003084, 0.00001335, -0.00003216, 0.00002402, -0.00004933, 0.00001846, -0.00018053, -0.00017842, 0.00097858, -0.00039921, -0.00001589, -0.00006600, 0.00000655, -0.00002868, 0.00001213, -0.00002653, 0.00001852, -0.00001489, -0.00000258, -0.00000103, 0.00000141, -0.00001521, 0.00001357, -0.00001063, 0.00000325, -0.00000365, 0.00000682, -0.00001294, 0.00001686, -0.00001352, 0.00000663, -0.00001382, 0.00001198, -0.00000807, 0.00001433, -0.00000796, 0.00000223, -0.00000561, 0.00000359, -0.00000532, 0.00000305, 0.00000237, 0.00000964, -0.00000285, -0.00001818, 0.00002602, -0.00002033, 0.00001321, -0.00002102, 0.00000604, 0.00000513, -0.00000793, -0.00001033, 0.00001112, -0.00002300, 0.00002681, -0.00001728, 0.00000230, 0.00000634, 0.00001131, -0.00002277, 0.00000498, 0.00000222, 0.00000876, -0.00000243, -0.00000121, -0.00000225, 0.00000203, -0.00000170, 0.00000586, -0.00000376, 0.00000215, -0.00000453, -0.00000168, -0.00000060, 0.00000761, -0.00000394, 0.00000485, -0.00000542, -0.00000208, -0.00000439, 0.00001052, -0.00000869, -0.00001191, },
		{ 0.00001352, 0.00000205, 0.00000806, -0.00001511, 0.00001139, -0.00001781, 0.00000824, -0.00000596, 0.00000103, 0.00001119, 0.00000833, -0.00001396, 0.00002877, -0.00000313, -0.00000043, -0.00000025, -0.00001048, 0.00000757, -0.00001655, 0.00000746, 0.00000494, -0.00000483, -0.00000365, -0.00000166, 0.00000217, 0.00000236, -0.00000112, -0.00000742, 0.00001180, -0.00001706, 0.00001357, -0.00000960, 0.00001306, -0.00002714, 0.00002462, -0.00002388, 0.00002649, -0.00002441, 0.00002565, -0.00004237, 0.00003926, -0.00004310, 0.00004497, -0.00003996, 0.00003514, -0.00006863, 0.00006128, -0.00008983, 0.00009650, -0.00015046, 0.00014287, -0.00046738, 0.00038123, 0.00066257, -0.00054063, 0.00015637, -0.00016847, 0.00009266, -0.00009210, 0.00006462, -0.00007106, 0.00006289, -0.00006049, 0.00003888, -0.00003259, 0.00002425, -0.00003441, 0.00003505, -0.00003389, 0.00002500, -0.00002021, 0.00001871, -0.00002259, 0.00002813, -0.00002833, 0.00002038, -0.00002627, 0.00002663, -0.00002178, 0.00002894, -0.00002660, 0.00002038, -0.00002265, 0.00002089, -0.00002375, 0.00002340, -0.00001510, 0.00002827, -0.00003230, 0.00001443, 0.00000210, -0.00000558, 0.00000052, -0.00001269, 0.00000550, 0.00000722, -0.00002019, 0.00000901, -0.00000935, -0.00000384, 0.00001785, -0.00001953, 0.00000847, 0.00000699, 0.00000417, -0.00002279, 0.00001274, -0.00000264, 0.00001319, -0.00001005, 0.00000429, -0.00000596, 0.00000579, -0.00000462, 0.00000968, -0.00000950, 0.00000753, -0.00001090, 0.00000459, -0.00000383, 0.00001190, -0.00001074, 0.00001234, -0.00001533, 0.00000696, -0.00001092, 0.00002157, -0.00002966, },
		{ -0.00003173, 0.00002528, -0.00000669, 0.00001023, -0.00001834, 0.00001625, -0.00002653, 0.00002357, -0.00002297, 0.00001934, -0.00000035, 0.00001025, -0.00001546, 0.00003956, -0.00002454, 0.00001847, -0.00001927, 0.00000890, -0.00001001, -0.00000417, 0.00000431, 0.00000742, -0.00001201, 0.00000330, -0.00000583, 0.00000668, -0.00000054, -0.00000383, -0.00000049, 0.00000449, -0.00001181, 0.00001089, -0.00000530, 0.00000299, -0.00001635, 0.00001772, -0.00001771, 0.00002118, -0.00001923, 0.00001574, -0.00003198, 0.00003239, -0.00003789, 0.00004366, -0.00004105, 0.00002646, -0.00005571, 0.00004796, -0.00007400, 0.00007774, -0.00013933, 0.00010750, -0.00055713, 0.00091955, 0.00006266, -0.00033409, 0.00010551, -0.00012709, 0.00007045, -0.00007081, 0.00004497, -0.00005076, 0.00004745, -0.00005377, 0.00003823, -0.00003027, 0.00001640, -0.00002246, 0.00002506, -0.00002810, 0.00002207, -0.00001565, 0.00001034, -0.00001057, 0.00001556, -0.00001911, 0.00001132, -0.00001581, 0.00001847, -0.00001299, 0.00001984, -0.00002160, 0.00001667, -0.00001879, 0.00001737, -0.00002172, 0.00002534, -0.00001533, 0.00002686, -0.00004217, 0.00003576, -0.00001913, 0.00000911, -0.00001243, -0.00000250, 0.00000504, 0.00000245, -0.00002107, 0.00001883, -0.00002625, 0.00002043, -0.00000414, -0.00000738, 0.00000481, 0.00001318, -0.00001064, -0.00001069, 0.00000898, 0.00000179, 0.00000840, -0.00000994, 0.00000439, -0.00000461, 0.00000419, -0.00000145, 0.00000633, -0.00000816, 0.00000620, -0.00001067, 0.00000497, -0.00000041, 0.00000726, -0.00000743, 0.00000913, -0.00001460, 0.00000575, -0.00000433, 0.00001460, },
		{ -0.00000761, -0.00001317, 0.00001756, -0.00000037, -0.00000244, -0.00000388, 0.00000033, -0.00001095, 0.00001417, -0.00001849, 0.00002134, -0.00000140, 0.00000186, -0.00000043, 0.00002798, -0.00002474, 0.00002044, -0.00002299, 0.00001525, -0.00001885, 0.00000413, 0.00000500, 0.00000218, -0.00001015, 0.00000294, -0.00000369, 0.00000552, 0.00000013, -0.00000892, 0.00001086, -0.00001090, 0.00000437, -0.00000340, 0.00000948, -0.00001861, 0.00001040, -0.00000884, 0.00001020, -0.00000767, 0.00001003, -0.00002004, 0.00000913, -0.00001017, 0.00000758, -0.00000173, 0.00000193, -0.00002779, 0.00001085, -0.00003028, 0.00002183, -0.00004571, 0.00001037, -0.00015414, -0.00025935, 0.00108834, -0.00046246, 0.00000539, -0.00007723, 0.00001403, -0.00003300, 0.00001576, -0.00003038, 0.00002034, -0.00001759, 0.00000032, -0.00000318, 0.00000449, -0.00001841, 0.00001531, -0.00001165, 0.00000392, -0.00000379, 0.00000733, -0.00001332, 0.00001600, -0.00001305, 0.00000703, -0.00001396, 0.00001116, -0.00000645, 0.00001220, -0.00000721, 0.00000285, -0.00000628, 0.00000429, -0.00000642, 0.00000218, 0.00000581, 0.00000489, 0.00000092, -0.00002277, 0.00003105, -0.00002263, 0.00001153, -0.00001560, 0.00000191, 0.00000845, -0.00001104, -0.00000563, 0.00000792, -0.00002209, 0.00002882, -0.00001977, 0.00000523, 0.00000060, 0.00001649, -0.00002352, 0.00000479, -0.00000032, 0.00001152, -0.00000333, -0.00000257, -0.00000090, 0.00000111, -0.00000143, 0.00000611, -0.00000283, -0.00000039, -0.00000141, -0.00000435, 0.00000064, 0.00000723, -0.00000333, 0.00000344, -0.00000285, -0.00000487, -0.00000310, 0.00001133, },
		{ 0.00002135, -0.00002792, 0.00001038, 0.00000255, 0.00000995, -0.00001733, 0.00001408, -0.00002170, 0.00001614, -0.00001186, 0.00000452, 0.00000654, 0.00000831, -0.00001314, 0.00002408, -0.00000011, -0.00000559, 0.00000494, -0.00000986, 0.00000514, -0.00001396, 0.00000452, 0.00000983, -0.00000900, -0.00000046, -0.00000412, 0.00000379, 0.00000024, 0.00000159, -0.00001050, 0.00001672, -0.00002161, 0.00001757, -0.00001493, 0.00001888, -0.00003237, 0.00003039, -0.00003055, 0.00003426, -0.00003354, 0.00003540, -0.00005067, 0.00004644, -0.00005116, 0.00005414, -0.00005131, 0.00004739, -0.00007999, 0.00007418, -0.00010571, 0.00011622, -0.00017571, 0.00017598, -0.00050094, 0.00036419, 0.00074192, -0.00061403, 0.00020058, -0.00019988, 0.00011751, -0.00011130, 0.00008065, -0.00008598, 0.00007444, -0.00007151, 0.00004973, -0.00004208, 0.00003411, -0.00004466, 0.00004403, -0.00004189, 0.00003242, -0.00002665, 0.00002510, -0.00002884, 0.00003268, -0.00003243, 0.00002519, -0.00003115, 0.00003067, -0.00002464, 0.00003017, -0.00002810, 0.00002317, -0.00002585, 0.00002444, -0.00002868, 0.00002764, -0.00001694, 0.00002752, -0.00003145, 0.00001201, 0.00000688, -0.00001025, 0.00000268, -0.00001003, 0.00000231, 0.00001027, -0.00002311, 0.00001466, -0.00001570, 0.00000201, 0.00001446, -0.00001726, 0.00000678, 0.00000475, 0.00000862, -0.00002403, 0.00001194, -0.00000424, 0.00001642, -0.00001288, 0.00000500, -0.00000596, 0.00000575, -0.00000513, 0.00001133, -0.00001083, 0.00000732, -0.00000935, 0.00000245, -0.00000249, 0.00001161, -0.00001081, 0.00001204, -0.00001368, 0.00000406, -0.00000835, },
		{ -0.00000188, 0.00001417, -0.00003007, 0.00002083, -0.00000323, 0.00000889, -0.00001745, 0.00001554, -0.00002631, 0.00002861, -0.00002913, 0.00002441, -0.00000832, 0.00001519, -0.00001978, 0.00003916, -0.00002447, 0.00001515, -0.00001279, 0.00000617, -0.00000965, -0.00000375, 0.00000366, 0.00001050, -0.00001565, 0.00000676, -0.00000889, 0.00000859, -0.00000240, -0.00000172, -0.00000236, 0.00000817, -0.00001568, 0.00001400, -0.00000956, 0.00000816, -0.00002097, 0.00002245, -0.00002305, 0.00002756, -0.00002755, 0.00002610, -0.00004208, 0.00004131, -0.00004714, 0.00005368, -0.00005378, 0.00004163, -0.00007078, 0.00006451, -0.00009392, 0.00010299, -0.00017404, 0.00016117, -0.00063852, 0.00097929, 0.00007097, -0.00036862, 0.00013234, -0.00014839, 0.00008890, -0.00008568, 0.00005730, -0.00006279, 0.00005646, -0.00006181, 0.00004645, -0.00003731, 0.00002373, -0.00003082, 0.00003294, -0.00003541, 0.00002924, -0.00002207, 0.00001681, -0.00001747, 0.00002096, -0.00002352, 0.00001628, -0.00002132, 0.00002377, -0.00001754, 0.00002224, -0.00002308, 0.00001885, -0.00002131, 0.00002014, -0.00002655, 0.00003120, -0.00002023, 0.00002919, -0.00004404, 0.00003616, -0.00001620, 0.00000379, -0.00000674, -0.00000415, 0.00000534, 0.00000235, -0.00002015, 0.00002058, -0.00003017, 0.00002609, -0.00000920, -0.00000265, -0.00000058, 0.00001519, -0.00000847, -0.00001148, 0.00000705, 0.00000230, 0.00001016, -0.00001287, 0.00000611, -0.00000552, 0.00000468, -0.00000213, 0.00000848, -0.00001106, 0.00000838, -0.00001151, 0.00000452, 0.00000009, 0.00000755, -0.00000843, 0.00001042, -0.00001494, 0.00000430, },
		{ -0.00000193, 0.00001203, -0.00000662, -0.00001356, 0.00001428, 0.00000384, -0.00000556, -0.00000071, -0.00000354, -0.00000540, 0.00001277, -0.00001986, 0.00002256, -0.00000723, 0.00000723, -0.00000702, 0.00003062, -0.00002765, 0.00002027, -0.00001721, 0.00001016, -0.00001524, 0.00000121, 0.00000838, 0.00000078, -0.00000990, 0.00000326, -0.00000424, 0.00000504, 0.00000101, -0.00000983, 0.00001276, -0.00001154, 0.00000460, -0.00000474, 0.00001027, -0.00001835, 0.00001074, -0.00000974, 0.00001139, -0.00000886, 0.00000998, -0.00001751, 0.00000592, -0.00000815, 0.00000617, -0.00000107, -0.00000025, -0.00002308, 0.00000653, -0.00002615, 0.00001686, -0.00003877, -0.00000213, -0.00012269, -0.00033888, 0.00118681, -0.00051487, 0.00002093, -0.00008478, 0.00001874, -0.00003509, 0.00001745, -0.00003239, 0.00002058, -0.00001932, 0.00000277, -0.00000485, 0.00000679, -0.00002036, 0.00001562, -0.00001153, 0.00000375, -0.00000309, 0.00000680, -0.00001237, 0.00001370, -0.00001138, 0.00000644, -0.00001299, 0.00000921, -0.00000385, 0.00000919, -0.00000589, 0.00000313, -0.00000655, 0.00000458, -0.00000706, 0.00000092, 0.00000934, 0.00000033, 0.00000398, -0.00002585, 0.00003372, -0.00002258, 0.00000798, -0.00000885, -0.00000273, 0.00001184, -0.00001415, -0.00000052, 0.00000396, -0.00001995, 0.00002944, -0.00002149, 0.00000780, -0.00000462, 0.00002051, -0.00002287, 0.00000378, -0.00000228, 0.00001354, -0.00000358, -0.00000438, 0.00000077, -0.00000005, -0.00000090, 0.00000596, -0.00000144, -0.00000329, 0.00000188, -0.00000703, 0.00000197, 0.00000656, -0.00000237, 0.00000172, -0.00000021, -0.00000746, },
		{ 0.00000038, -0.00000500, 0.00001995, -0.00002486, 0.00000659, 0.00000298, 0.00001171, -0.00001890, 0.00001601, -0.00002479, 0.00002325, -0.00001773, 0.00000813, 0.00000215, 0.00000787, -0.00001128, 0.00001867, 0.00000323, -0.00001041, 0.00000864, -0.00000743, 0.00000081, -0.00000958, 0.00000032, 0.00001543, -0.00001378, 0.00000346, -0.00000728, 0.00000616, -0.00000281, 0.00000514, -0.00001403, 0.00002175, -0.00002608, 0.00002151, -0.00002026, 0.00002441, -0.00003690, 0.00003533, -0.00003639, 0.00004116, -0.00004180, 0.00004400, -0.00005740, 0.00005198, -0.00005765, 0.00006189, -0.00006153, 0.00005837, -0.00008953, 0.00008495, -0.00011887, 0.00013233, -0.00019620, 0.00020183, -0.00052288, 0.00033796, 0.00081910, -0.00068031, 0.00023963, -0.00022753, 0.00013941, -0.00012808, 0.00009460, -0.00009883, 0.00008396, -0.00008095, 0.00005969, -0.00005105, 0.00004346, -0.00005402, 0.00005174, -0.00004871, 0.00003895, -0.00003237, 0.00003071, -0.00003401, 0.00003585, -0.00003518, 0.00002884, -0.00003471, 0.00003320, -0.00002601, 0.00002985, -0.00002822, 0.00002477, -0.00002783, 0.00002671, -0.00003214, 0.00003027, -0.00001722, 0.00002526, -0.00002929, 0.00000900, 0.00001102, -0.00001355, 0.00000335, -0.00000595, -0.00000165, 0.00001351, -0.00002574, 0.00002010, -0.00002198, 0.00000834, 0.00000997, -0.00001414, 0.00000468, 0.00000282, 0.00001214, -0.00002377, 0.00000992, -0.00000485, 0.00001845, -0.00001447, 0.00000461, -0.00000503, 0.00000489, -0.00000485, 0.00001205, -0.00001109, 0.00000604, -0.00000688, -0.00000038, -0.00000047, 0.00001045, -0.00000984, 0.00001059, -0.00001097, },
		{ -0.00001406, 0.00000192, 0.00000138, 0.00001252, -0.00002665, 0.00001491, 0.00000086, 0.00000708, -0.00001569, 0.00001372, -0.00002473, 0.00003187, -0.00003407, 0.00002867, -0.00001533, 0.00001914, -0.00002230, 0.00003679, -0.00002240, 0.00001012, -0.00000574, 0.00000349, -0.00000984, -0.00000236, 0.00000210, 0.00001404, -0.00001955, 0.00001062, -0.00001236, 0.00001104, -0.00000516, 0.00000149, -0.00000524, 0.00001260, -0.00002008, 0.00001767, -0.00001456, 0.00001402, -0.00002591, 0.00002729, -0.00002842, 0.00003392, -0.00003592, 0.00003634, -0.00005159, 0.00004942, -0.00005548, 0.00006280, -0.00006597, 0.00005645, -0.00008522, 0.00008035, -0.00011285, 0.00012662, -0.00020624, 0.00020965, -0.00070707, 0.00102211, 0.00008787, -0.00040399, 0.00015870, -0.00016911, 0.00010680, -0.00010010, 0.00006929, -0.00007430, 0.00006460, -0.00006902, 0.00005431, -0.00004449, 0.00003149, -0.00003945, 0.00004063, -0.00004238, 0.00003623, -0.00002857, 0.00002348, -0.00002440, 0.00002606, -0.00002749, 0.00002092, -0.00002639, 0.00002836, -0.00002128, 0.00002364, -0.00002355, 0.00002013, -0.00002288, 0.00002189, -0.00003007, 0.00003537, -0.00002324, 0.00002931, -0.00004336, 0.00003409, -0.00001184, -0.00000192, -0.00000128, -0.00000508, 0.00000508, 0.00000241, -0.00001876, 0.00002157, -0.00003304, 0.00003092, -0.00001419, 0.00000212, -0.00000566, 0.00001654, -0.00000605, -0.00001176, 0.00000459, 0.00000317, 0.00001127, -0.00001495, 0.00000698, -0.00000566, 0.00000451, -0.00000224, 0.00000995, -0.00001308, 0.00000955, -0.00001140, 0.00000335, 0.00000115, 0.00000715, -0.00000855, 0.00001059, },
		{ 0.00000219, -0.00000957, -0.00000105, 0.00001270, -0.00000585, -0.00001299, 0.00000987, 0.00000847, -0.00000862, 0.00000253, -0.00000764, 0.00000059, 0.00001016, -0.00002007, 0.00002301, -0.00001231, 0.00001201, -0.00001227, 0.00003142, -0.00002822, 0.00001744, -0.00000947, 0.00000362, -0.00001072, -0.00000193, 0.00001160, -0.00000065, -0.00000940, 0.00000339, -0.00000460, 0.00000442, 0.00000165, -0.00001004, 0.00001370, -0.00001126, 0.00000405, -0.00000525, 0.00000993, -0.00001674, 0.00000985, -0.00000952, 0.00001148, -0.00000896, 0.00000863, -0.00001363, 0.00000167, -0.00000529, 0.00000410, 0.00000017, -0.00000332, -0.00001701, 0.00000070, -0.00002014, 0.00000960, -0.00002925, -0.00001776, -0.00008872, -0.00041273, 0.00126921, -0.00055307, 0.00002901, -0.00008773, 0.00002011, -0.00003455, 0.00001692, -0.00003229, 0.00001903, -0.00001987, 0.00000451, -0.00000577, 0.00000801, -0.00002080, 0.00001436, -0.00001022, 0.00000271, -0.00000156, 0.00000526, -0.00001017, 0.00001012, -0.00000867, 0.00000500, -0.00001106, 0.00000631, -0.00000053, 0.00000559, -0.00000419, 0.00000315, -0.00000650, 0.00000456, -0.00000737, -0.00000045, 0.00001254, -0.00000357, 0.00000590, -0.00002698, 0.00003368, -0.00002011, 0.00000282, -0.00000119, -0.00000744, 0.00001497, -0.00001693, 0.00000473, -0.00000050, -0.00001671, 0.00002861, -0.00002228, 0.00000972, -0.00000893, 0.00002303, -0.00002075, 0.00000197, -0.00000364, 0.00001478, -0.00000329, -0.00000643, 0.00000251, -0.00000124, -0.00000024, 0.00000555, 0.00000019, -0.00000622, 0.00000496, -0.00000938, 0.00000316, 0.00000579, -0.00000129, -0.00000001, },
		{ 0.00000822, -0.00000751, -0.00000370, -0.00000122, 0.00001767, -0.00002081, 0.00000251, 0.00000313, 0.00001332, -0.00001971, 0.00001701, -0.00002683, 0.00002917, -0.00002319, 0.00001163, -0.00000193, 0.00000710, -0.00000868, 0.00001300, 0.00000632, -0.00001421, 0.00001044, -0.00000320, -0.00000521, -0.00000372, -0.00000479, 0.00002134, -0.00001879, 0.00000781, -0.00001087, 0.00000906, -0.00000655, 0.00000921, -0.00001766, 0.00002643, -0.00003001, 0.00002499, -0.00002511, 0.00002916, -0.00004024, 0.00003895, -0.00004086, 0.00004660, -0.00004853, 0.00005076, -0.00006198, 0.00005542, -0.00006211, 0.00006768, -0.00006996, 0.00006737, -0.00009653, 0.00009281, -0.00012842, 0.00014380, -0.00021075, 0.00021926, -0.00053285, 0.00030429, 0.00089056, -0.00073585, 0.00027080, -0.00024944, 0.00015680, -0.00014118, 0.00010541, -0.00010868, 0.00009065, -0.00008809, 0.00006804, -0.00005877, 0.00005155, -0.00006171, 0.00005748, -0.00005375, 0.00004404, -0.00003686, 0.00003505, -0.00003770, 0.00003734, -0.00003632, 0.00003103, -0.00003669, 0.00003405, -0.00002580, 0.00002800, -0.00002695, 0.00002513, -0.00002847, 0.00002756, -0.00003399, 0.00003120, -0.00001607, 0.00002174, -0.00002617, 0.00000585, 0.00001394, -0.00001506, 0.00000237, -0.00000056, -0.00000616, 0.00001671, -0.00002782, 0.00002500, -0.00002780, 0.00001474, 0.00000467, -0.00001033, 0.00000219, 0.00000144, 0.00001439, -0.00002182, 0.00000658, -0.00000434, 0.00001908, -0.00001470, 0.00000312, -0.00000321, 0.00000328, -0.00000380, 0.00001183, -0.00001030, 0.00000383, -0.00000374, -0.00000363, 0.00000199, 0.00000858, -0.00000800, },
		{ -0.00000776, 0.00000965, -0.00001209, -0.00000120, 0.00000521, 0.00000987, -0.00002180, 0.00000800, 0.00000513, 0.00000509, -0.00001323, 0.00001090, -0.00002182, 0.00003318, -0.00003740, 0.00003172, -0.00002107, 0.00002181, -0.00002294, 0.00003259, -0.00001872, 0.00000411, 0.00000116, 0.00000133, -0.00001082, 0.00000008, -0.00000032, 0.00001790, -0.00002354, 0.00001468, -0.00001608, 0.00001390, -0.00000867, 0.00000566, -0.00000895, 0.00001749, -0.00002469, 0.00002160, -0.00001992, 0.00002011, -0.00003072, 0.00003178, -0.00003334, 0.00003971, -0.00004366, 0.00004564, -0.00005973, 0.00005601, -0.00006222, 0.00007029, -0.00007672, 0.00006988, -0.00009794, 0.00009426, -0.00012935, 0.00014684, -0.00023362, 0.00024990, -0.00075949, 0.00104608, 0.00011307, -0.00043871, 0.00018305, -0.00018794, 0.00012302, -0.00011312, 0.00008014, -0.00008455, 0.00007130, -0.00007485, 0.00006129, -0.00005129, 0.00003913, -0.00004767, 0.00004747, -0.00004842, 0.00004247, -0.00003455, 0.00002974, -0.00003080, 0.00003035, -0.00003059, 0.00002481, -0.00003057, 0.00003184, -0.00002389, 0.00002386, -0.00002288, 0.00002038, -0.00002333, 0.00002239, -0.00003199, 0.00003754, -0.00002414, 0.00002718, -0.00004021, 0.00002981, -0.00000659, -0.00000740, 0.00000344, -0.00000499, 0.00000415, 0.00000271, -0.00001692, 0.00002175, -0.00003470, 0.00003460, -0.00001876, 0.00000662, -0.00001016, 0.00001723, -0.00000359, -0.00001135, 0.00000150, 0.00000455, 0.00001151, -0.00001592, 0.00000680, -0.00000491, 0.00000361, -0.00000168, 0.00001057, -0.00001400, 0.00000957, -0.00001028, 0.00000151, 0.00000276, 0.00000604, },
		{ -0.00000030, -0.00000150, 0.00000408, -0.00001097, -0.00000060, 0.00001342, -0.00000535, -0.00001150, 0.00000459, 0.00001312, -0.00001129, 0.00000556, -0.00001169, 0.00000674, 0.00000650, -0.00001907, 0.00002248, -0.00001631, 0.00001580, -0.00001580, 0.00003012, -0.00002642, 0.00001236, -0.00000047, -0.00000369, -0.00000585, -0.00000483, 0.00001431, -0.00000187, -0.00000881, 0.00000346, -0.00000487, 0.00000381, 0.00000188, -0.00000937, 0.00001354, -0.00000996, 0.00000266, -0.00000486, 0.00000842, -0.00001381, 0.00000776, -0.00000821, 0.00001050, -0.00000803, 0.00000616, -0.00000872, -0.00000324, -0.00000196, 0.00000162, 0.00000173, -0.00000696, -0.00001002, -0.00000610, -0.00001285, 0.00000080, -0.00001819, -0.00003490, -0.00005501, -0.00047674, 0.00133135, -0.00057446, 0.00002850, -0.00008559, 0.00001790, -0.00003129, 0.00001409, -0.00003004, 0.00001569, -0.00001915, 0.00000539, -0.00000579, 0.00000800, -0.00001965, 0.00001156, -0.00000782, 0.00000091, 0.00000067, 0.00000288, -0.00000698, 0.00000554, -0.00000519, 0.00000295, -0.00000840, 0.00000279, 0.00000313, 0.00000175, -0.00000237, 0.00000306, -0.00000623, 0.00000433, -0.00000752, -0.00000169, 0.00001506, -0.00000644, 0.00000636, -0.00002592, 0.00003083, -0.00001546, -0.00000348, 0.00000682, -0.00001173, 0.00001748, -0.00001909, 0.00000979, -0.00000516, -0.00001257, 0.00002636, -0.00002204, 0.00001077, -0.00001203, 0.00002390, -0.00001723, -0.00000051, -0.00000444, 0.00001530, -0.00000262, -0.00000843, 0.00000406, -0.00000226, 0.00000036, 0.00000506, 0.00000180, -0.00000881, 0.00000746, -0.00001111, 0.00000400, 0.00000511, },
		{ 0.00000625, -0.00000557, 0.00000526, -0.00000371, -0.00000775, 0.00000261, 0.00001485, -0.00001616, -0.00000150, 0.00000284, 0.00001475, -0.00001966, 0.00001697, -0.00002766, 0.00003364, -0.00002788, 0.00001484, -0.00000566, 0.00000611, -0.00000566, 0.00000754, 0.00000858, -0.00001647, 0.00001015, 0.00000260, -0.00001246, 0.00000313, -0.00001027, 0.00002704, -0.00002358, 0.00001219, -0.00001455, 0.00001220, -0.00001063, 0.00001341, -0.00002096, 0.00003032, -0.00003300, 0.00002760, -0.00002905, 0.00003269, -0.00004199, 0.00004086, -0.00004354, 0.00005012, -0.00005323, 0.00005518, -0.00006403, 0.00005649, -0.00006426, 0.00007114, -0.00007610, 0.00007384, -0.00010050, 0.00009725, -0.00013377, 0.00014997, -0.00021868, 0.00022775, -0.00053128, 0.00026548, 0.00095262, -0.00077731, 0.00029176, -0.00026394, 0.00016839, -0.00014960, 0.00011222, -0.00011480, 0.00009392, -0.00009235, 0.00007415, -0.00006460, 0.00005769, -0.00006706, 0.00006072, -0.00005655, 0.00004726, -0.00003972, 0.00003778, -0.00003963, 0.00003697, -0.00003574, 0.00003166, -0.00003698, 0.00003317, -0.00002411, 0.00002478, -0.00002446, 0.00002430, -0.00002780, 0.00002699, -0.00003423, 0.00003055, -0.00001372, 0.00001735, -0.00002253, 0.00000303, 0.00001520, -0.00001453, -0.00000020, 0.00000588, -0.00001086, 0.00001957, -0.00002906, 0.00002903, -0.00003274, 0.00002079, -0.00000112, -0.00000604, -0.00000066, 0.00000076, 0.00001519, -0.00001816, 0.00000202, -0.00000270, 0.00001827, -0.00001359, 0.00000068, -0.00000071, 0.00000108, -0.00000212, 0.00001076, -0.00000860, 0.00000095, -0.00000025, -0.00000695, 0.00000462, },
		{ 0.00000476, 0.00000434, -0.00000616, 0.00000774, -0.00000925, -0.00000472, 0.00000922, 0.00000660, -0.00001597, 0.00000066, 0.00000912, 0.00000321, -0.00001028, 0.00000726, -0.00001772, 0.00003254, -0.00003892, 0.00003333, -0.00002531, 0.00002310, -0.00002179, 0.00002688, -0.00001400, -0.00000209, 0.00000719, 0.00000007, -0.00001269, 0.00000354, -0.00000344, 0.00002184, -0.00002733, 0.00001867, -0.00001978, 0.00001695, -0.00001272, 0.00001049, -0.00001313, 0.00002245, -0.00002911, 0.00002537, -0.00002519, 0.00002594, -0.00003493, 0.00003546, -0.00003731, 0.00004434, -0.00005009, 0.00005326, -0.00006577, 0.00006045, -0.00006679, 0.00007551, -0.00008525, 0.00008100, -0.00010794, 0.00010514, -0.00014210, 0.00016207, -0.00025421, 0.00027947, -0.00079342, 0.00105034, 0.00014548, -0.00047091, 0.00020372, -0.00020351, 0.00013642, -0.00012377, 0.00008900, -0.00009283, 0.00007596, -0.00007883, 0.00006689, -0.00005716, 0.00004600, -0.00005481, 0.00005283, -0.00005292, 0.00004738, -0.00003945, 0.00003502, -0.00003611, 0.00003341, -0.00003246, 0.00002760, -0.00003348, 0.00003386, -0.00002515, 0.00002281, -0.00002107, 0.00001956, -0.00002260, 0.00002155, -0.00003218, 0.00003761, -0.00002297, 0.00002300, -0.00003493, 0.00002383, -0.00000109, -0.00001205, 0.00000700, -0.00000368, 0.00000250, 0.00000327, -0.00001467, 0.00002111, -0.00003501, 0.00003687, -0.00002261, 0.00001055, -0.00001390, 0.00001730, -0.00000131, -0.00001012, -0.00000224, 0.00000650, 0.00001071, -0.00001561, 0.00000551, -0.00000326, 0.00000198, -0.00000041, 0.00001025, -0.00001374, 0.00000841, -0.00000823, -0.00000088, },
		{ 0.00000430, 0.00000465, 0.00000042, -0.00000255, 0.00000524, -0.00001152, -0.00000063, 0.00001415, -0.00000510, -0.00000923, -0.00000120, 0.00001739, -0.00001322, 0.00000811, -0.00001538, 0.00001271, 0.00000207, -0.00001689, 0.00002085, -0.00001901, 0.00001829, -0.00001741, 0.00002670, -0.00002247, 0.00000568, 0.00000895, -0.00001100, -0.00000121, -0.00000704, 0.00001618, -0.00000267, -0.00000828, 0.00000356, -0.00000515, 0.00000334, 0.00000157, -0.00000776, 0.00001223, -0.00000768, 0.00000050, -0.00000363, 0.00000586, -0.00000975, 0.00000466, -0.00000598, 0.00000862, -0.00000625, 0.00000286, -0.00000322, -0.00000832, 0.00000145, -0.00000093, 0.00000332, -0.00001078, -0.00000268, -0.00001325, -0.00000501, -0.00000862, -0.00000678, -0.00005181, -0.00002429, -0.00052727, 0.00137001, -0.00057741, 0.00001903, -0.00007842, 0.00001226, -0.00002550, 0.00000916, -0.00002583, 0.00001070, -0.00001724, 0.00000540, -0.00000490, 0.00000677, -0.00001698, 0.00000743, -0.00000458, -0.00000141, 0.00000334, -0.00000005, -0.00000314, 0.00000039, -0.00000133, 0.00000059, -0.00000535, -0.00000098, 0.00000674, -0.00000197, -0.00000067, 0.00000300, -0.00000586, 0.00000403, -0.00000765, -0.00000257, 0.00001659, -0.00000800, 0.00000525, -0.00002269, 0.00002539, -0.00000911, -0.00001029, 0.00001456, -0.00001511, 0.00001903, -0.00002030, 0.00001434, -0.00000966, -0.00000783, 0.00002281, -0.00002075, 0.00001085, -0.00001376, 0.00002312, -0.00001256, -0.00000341, -0.00000482, 0.00001525, -0.00000180, -0.00001011, 0.00000516, -0.00000291, 0.00000074, 0.00000467, 0.00000314, -0.00001078, 0.00000910, -0.00001199, },
		{ -0.00000997, 0.00000707, 0.00000378, -0.00000288, 0.00000212, -0.00000002, -0.00001132, 0.00000608, 0.00001185, -0.00001129, -0.00000509, 0.00000200, 0.00001595, -0.00001869, 0.00001585, -0.00002721, 0.00003645, -0.00003153, 0.00001757, -0.00000903, 0.00000506, -0.00000260, 0.00000273, 0.00000960, -0.00001686, 0.00000788, 0.00000948, -0.00002033, 0.00001034, -0.00001554, 0.00003199, -0.00002769, 0.00001619, -0.00001796, 0.00001522, -0.00001466, 0.00001732, -0.00002352, 0.00003303, -0.00003470, 0.00002901, -0.00003171, 0.00003466, -0.00004191, 0.00004083, -0.00004420, 0.00005144, -0.00005555, 0.00005699, -0.00006341, 0.00005513, -0.00006403, 0.00007211, -0.00007966, 0.00007746, -0.00010116, 0.00009801, -0.00013466, 0.00015061, -0.00021982, 0.00022743, -0.00051928, 0.00022421, 0.00100181, -0.00080189, 0.00030072, -0.00026984, 0.00017330, -0.00015264, 0.00011451, -0.00011675, 0.00009340, -0.00009334, 0.00007756, -0.00006804, 0.00006137, -0.00006962, 0.00006110, -0.00005685, 0.00004838, -0.00004074, 0.00003872, -0.00003971, 0.00003479, -0.00003351, 0.00003077, -0.00003564, 0.00003073, -0.00002117, 0.00002051, -0.00002101, 0.00002247, -0.00002596, 0.00002515, -0.00003302, 0.00002854, -0.00001055, 0.00001255, -0.00001880, 0.00000096, 0.00001449, -0.00001199, -0.00000410, 0.00001296, -0.00001534, 0.00002178, -0.00002920, 0.00003191, -0.00003643, 0.00002604, -0.00000704, -0.00000149, -0.00000379, 0.00000087, 0.00001447, -0.00001294, -0.00000354, -0.00000011, 0.00001616, -0.00001134, -0.00000246, 0.00000220, -0.00000144, -0.00000001, 0.00000904, -0.00000625, -0.00000228, 0.00000320, },
		{ -0.00000546, -0.00000357, 0.00000695, 0.00000224, -0.00000396, 0.00000515, -0.00000590, -0.00000825, 0.00001302, 0.00000312, -0.00000970, -0.00000651, 0.00001236, 0.00000173, -0.00000706, 0.00000304, -0.00001269, 0.00003008, -0.00003857, 0.00003339, -0.00002800, 0.00002304, -0.00001913, 0.00002015, -0.00000890, -0.00000768, 0.00001177, -0.00000008, -0.00001541, 0.00000780, -0.00000698, 0.00002551, -0.00003061, 0.00002227, -0.00002318, 0.00001992, -0.00001696, 0.00001559, -0.00001736, 0.00002702, -0.00003290, 0.00002856, -0.00002987, 0.00003099, -0.00003806, 0.00003789, -0.00003989, 0.00004735, -0.00005463, 0.00005855, -0.00006918, 0.00006231, -0.00006879, 0.00007802, -0.00009092, 0.00008901, -0.00011439, 0.00011208, -0.00015008, 0.00017113, -0.00026655, 0.00029667, -0.00080759, 0.00103514, 0.00018331, -0.00049848, 0.00021902, -0.00021450, 0.00014590, -0.00013118, 0.00009515, -0.00009852, 0.00007812, -0.00008053, 0.00007062, -0.00006156, 0.00005151, -0.00006023, 0.00005614, -0.00005541, 0.00005048, -0.00004278, 0.00003883, -0.00003992, 0.00003490, -0.00003287, 0.00002903, -0.00003489, 0.00003426, -0.00002499, 0.00002056, -0.00001825, 0.00001777, -0.00002074, 0.00001941, -0.00003067, 0.00003566, -0.00001996, 0.00001719, -0.00002807, 0.00001681, 0.00000395, -0.00001533, 0.00000909, -0.00000108, 0.00000020, 0.00000407, -0.00001204, 0.00001970, -0.00003395, 0.00003752, -0.00002549, 0.00001368, -0.00001677, 0.00001683, 0.00000060, -0.00000802, -0.00000656, 0.00000898, 0.00000886, -0.00001403, 0.00000316, -0.00000084, -0.00000026, 0.00000147, 0.00000903, -0.00001233, 0.00000620, },
		{ 0.00000967, -0.00001188, 0.00000401, 0.00000451, 0.00000078, -0.00000303, 0.00000559, -0.00001121, -0.00000107, 0.00001481, -0.00000498, -0.00000643, -0.00000707, 0.00002085, -0.00001412, 0.00000993, -0.00001848, 0.00001817, -0.00000283, -0.00001367, 0.00001809, -0.00002033, 0.00001936, -0.00001712, 0.00002137, -0.00001687, -0.00000178, 0.00001786, -0.00001755, 0.00000264, -0.00000817, 0.00001698, -0.00000291, -0.00000788, 0.00000375, -0.00000548, 0.00000307, 0.00000070, -0.00000523, 0.00000988, -0.00000459, -0.00000227, -0.00000172, 0.00000249, -0.00000487, 0.00000089, -0.00000314, 0.00000611, -0.00000392, -0.00000088, 0.00000234, -0.00001307, 0.00000449, -0.00000321, 0.00000464, -0.00001434, 0.00000441, -0.00002005, 0.00000259, -0.00001769, 0.00000374, -0.00006680, 0.00000103, -0.00056144, 0.00138313, -0.00056144, 0.00000103, -0.00006680, 0.00000374, -0.00001769, 0.00000259, -0.00002005, 0.00000441, -0.00001434, 0.00000464, -0.00000321, 0.00000449, -0.00001307, 0.00000234, -0.00000088, -0.00000392, 0.00000611, -0.00000314, 0.00000089, -0.00000487, 0.00000249, -0.00000172, -0.00000227, -0.00000459, 0.00000988, -0.00000523, 0.00000070, 0.00000307, -0.00000548, 0.00000375, -0.00000788, -0.00000291, 0.00001698, -0.00000817, 0.00000264, -0.00001755, 0.00001786, -0.00000178, -0.00001687, 0.00002137, -0.00001712, 0.00001936, -0.00002033, 0.00001809, -0.00001367, -0.00000283, 0.00001817, -0.00001848, 0.00000993, -0.00001412, 0.00002085, -0.00000707, -0.00000643, -0.00000498, 0.00001481, -0.00000107, -0.00001121, 0.00000559, -0.00000303, 0.00000078, 0.00000451, 0.00000401, -0.00001188, },
		{ -0.00000546, 0.00000620, -0.00001233, 0.00000903, 0.00000147, -0.00000026, -0.00000084, 0.00000316, -0.00001403, 0.00000886, 0.00000898, -0.00000656, -0.00000802, 0.00000060, 0.00001683, -0.00001677, 0.00001368, -0.00002549, 0.00003752, -0.00003395, 0.00001970, -0.00001204, 0.00000407, 0.00000020, -0.00000108, 0.00000909, -0.00001533, 0.00000395, 0.00001681, -0.00002807, 0.00001719, -0.00001996, 0.00003566, -0.00003067, 0.00001941, -0.00002074, 0.00001777, -0.00001825, 0.00002056, -0.00002499, 0.00003426, -0.00003489, 0.00002903, -0.00003287, 0.00003490, -0.00003992, 0.00003883, -0.00004278, 0.00005048, -0.00005541, 0.00005614, -0.00006023, 0.00005151, -0.00006156, 0.00007062, -0.00008053, 0.00007812, -0.00009852, 0.00009515, -0.00013118, 0.00014590, -0.00021450, 0.00021902, -0.00049848, 0.00018331, 0.00103514, -0.00080759, 0.00029667, -0.00026655, 0.00017113, -0.00015008, 0.00011208, -0.00011439, 0.00008901, -0.00009092, 0.00007802, -0.00006879, 0.00006231, -0.00006918, 0.00005855, -0.00005463, 0.00004735, -0.00003989, 0.00003789, -0.00003806, 0.00003099, -0.00002987, 0.00002856, -0.00003290, 0.00002702, -0.00001736, 0.00001559, -0.00001696, 0.00001992, -0.00002318, 0.00002227, -0.00003061, 0.00002551, -0.00000698, 0.00000780, -0.00001541, -0.00000008, 0.00001177, -0.00000768, -0.00000890, 0.00002015, -0.00001913, 0.00002304, -0.00002800, 0.00003339, -0.00003857, 0.00003008, -0.00001269, 0.00000304, -0.00000706, 0.00000173, 0.00001236, -0.00000651, -0.00000970, 0.00000312, 0.00001302, -0.00000825, -0.00000590, 0.00000515, -0.00000396, 0.00000224, 0.00000695, -0.00000357, },
		{ -0.00000997, 0.00000320, -0.00000228, -0.00000625, 0.00000904, -0.00000001, -0.00000144, 0.00000220, -0.00000246, -0.00001134, 0.00001616, -0.00000011, -0.00000354, -0.00001294, 0.00001447, 0.00000087, -0.00000379, -0.00000149, -0.00000704, 0.00002604, -0.00003643, 0.00003191, -0.00002920, 0.00002178, -0.00001534, 0.00001296, -0.00000410, -0.00001199, 0.00001449, 0.00000096, -0.00001880, 0.00001255, -0.00001055, 0.00002854, -0.00003302, 0.00002515, -0.00002596, 0.00002247, -0.00002101, 0.00002051, -0.00002117, 0.00003073, -0.00003564, 0.00003077, -0.00003351, 0.00003479, -0.00003971, 0.00003872, -0.00004074, 0.00004838, -0.00005685, 0.00006110, -0.00006962, 0.00006137, -0.00006804, 0.00007756, -0.00009334, 0.00009340, -0.00011675, 0.00011451, -0.00015264, 0.00017330, -0.00026984, 0.00030072, -0.00080189, 0.00100181, 0.00022421, -0.00051928, 0.00022743, -0.00021982, 0.00015061, -0.00013466, 0.00009801, -0.00010116, 0.00007746, -0.00007966, 0.00007211, -0.00006403, 0.00005513, -0.00006341, 0.00005699, -0.00005555, 0.00005144, -0.00004420, 0.00004083, -0.00004191, 0.00003466, -0.00003171, 0.00002901, -0.00003470, 0.00003303, -0.00002352, 0.00001732, -0.00001466, 0.00001522, -0.00001796, 0.00001619, -0.00002769, 0.00003199, -0.00001554, 0.00001034, -0.00002033, 0.00000948, 0.00000788, -0.00001686, 0.00000960, 0.00000273, -0.00000260, 0.00000506, -0.00000903, 0.00001757, -0.00003153, 0.00003645, -0.00002721, 0.00001585, -0.00001869, 0.00001595, 0.00000200, -0.00000509, -0.00001129, 0.00001185, 0.00000608, -0.00001132, -0.00000002, 0.00000212, -0.00000288, 0.00000378, 0.00000707, },
		{ 0.00000430, -0.00001199, 0.00000910, -0.00001078, 0.00000314, 0.00000467, 0.00000074, -0.00000291, 0.00000516, -0.00001011, -0.00000180, 0.00001525, -0.00000482, -0.00000341, -0.00001256, 0.00002312, -0.00001376, 0.00001085, -0.00002075, 0.00002281, -0.00000783, -0.00000966, 0.00001434, -0.00002030, 0.00001903, -0.00001511, 0.00001456, -0.00001029, -0.00000911, 0.00002539, -0.00002269, 0.00000525, -0.00000800, 0.00001659, -0.00000257, -0.00000765, 0.00000403, -0.00000586, 0.00000300, -0.00000067, -0.00000197, 0.00000674, -0.00000098, -0.00000535, 0.00000059, -0.00000133, 0.00000039, -0.00000314, -0.00000005, 0.00000334, -0.00000141, -0.00000458, 0.00000743, -0.00001698, 0.00000677, -0.00000490, 0.00000540, -0.00001724, 0.00001070, -0.00002583, 0.00000916, -0.00002550, 0.00001226, -0.00007842, 0.00001903, -0.00057741, 0.00137001, -0.00052727, -0.00002429, -0.00005181, -0.00000678, -0.00000862, -0.00000501, -0.00001325, -0.00000268, -0.00001078, 0.00000332, -0.00000093, 0.00000145, -0.00000832, -0.00000322, 0.00000286, -0.00000625, 0.00000862, -0.00000598, 0.00000466, -0.00000975, 0.00000586, -0.00000363, 0.00000050, -0.00000768, 0.00001223, -0.00000776, 0.00000157, 0.00000334, -0.00000515, 0.00000356, -0.00000828, -0.00000267, 0.00001618, -0.00000704, -0.00000121, -0.00001100, 0.00000895, 0.00000568, -0.00002247, 0.00002670, -0.00001741, 0.00001829, -0.00001901, 0.00002085, -0.00001689, 0.00000207, 0.00001271, -0.00001538, 0.00000811, -0.00001322, 0.00001739, -0.00000120, -0.00000923, -0.00000510, 0.00001415, -0.00000063, -0.00001152, 0.00000524, -0.00000255, 0.00000042, 0.00000465, },
		{ 0.00000476, -0.00000088, -0.00000823, 0.00000841, -0.00001374, 0.00001025, -0.00000041, 0.00000198, -0.00000326, 0.00000551, -0.00001561, 0.00001071, 0.00000650, -0.00000224, -0.00001012, -0.00000131, 0.00001730, -0.00001390, 0.00001055, -0.00002261, 0.00003687, -0.00003501, 0.00002111, -0.00001467, 0.00000327, 0.00000250, -0.00000368, 0.00000700, -0.00001205, -0.00000109, 0.00002383, -0.00003493, 0.00002300, -0.00002297, 0.00003761, -0.00003218, 0.00002155, -0.00002260, 0.00001956, -0.00002107, 0.00002281, -0.00002515, 0.00003386, -0.00003348, 0.00002760, -0.00003246, 0.00003341, -0.00003611, 0.00003502, -0.00003945, 0.00004738, -0.00005292, 0.00005283, -0.00005481, 0.00004600, -0.00005716, 0.00006689, -0.00007883, 0.00007596, -0.00009283, 0.00008900, -0.00012377, 0.00013642, -0.00020351, 0.00020372, -0.00047091, 0.00014548, 0.00105034, -0.00079342, 0.00027947, -0.00025421, 0.00016207, -0.00014210, 0.00010514, -0.00010794, 0.00008100, -0.00008525, 0.00007551, -0.00006679, 0.00006045, -0.00006577, 0.00005326, -0.00005009, 0.00004434, -0.00003731, 0.00003546, -0.00003493, 0.00002594, -0.00002519, 0.00002537, -0.00002911, 0.00002245, -0.00001313, 0.00001049, -0.00001272, 0.00001695, -0.00001978, 0.00001867, -0.00002733, 0.00002184, -0.00000344, 0.00000354, -0.00001269, 0.00000007, 0.00000719, -0.00000209, -0.00001400, 0.00002688, -0.00002179, 0.00002310, -0.00002531, 0.00003333, -0.00003892, 0.00003254, -0.00001772, 0.00000726, -0.00001028, 0.00000321, 0.00000912, 0.00000066, -0.00001597, 0.00000660, 0.00000922, -0.00000472, -0.00000925, 0.00000774, -0.00000616, 0.00000434, },
		{ 0.00000625, 0.00000462, -0.00000695, -0.00000025, 0.00000095, -0.00000860, 0.00001076, -0.00000212, 0.00000108, -0.00000071, 0.00000068, -0.00001359, 0.00001827, -0.00000270, 0.00000202, -0.00001816, 0.00001519, 0.00000076, -0.00000066, -0.00000604, -0.00000112, 0.00002079, -0.00003274, 0.00002903, -0.00002906, 0.00001957, -0.00001086, 0.00000588, -0.00000020, -0.00001453, 0.00001520, 0.00000303, -0.00002253, 0.00001735, -0.00001372, 0.00003055, -0.00003423, 0.00002699, -0.00002780, 0.00002430, -0.00002446, 0.00002478, -0.00002411, 0.00003317, -0.00003698, 0.00003166, -0.00003574, 0.00003697, -0.00003963, 0.00003778, -0.00003972, 0.00004726, -0.00005655, 0.00006072, -0.00006706, 0.00005769, -0.00006460, 0.00007415, -0.00009235, 0.00009392, -0.00011480, 0.00011222, -0.00014960, 0.00016839, -0.00026394, 0.00029176, -0.00077731, 0.00095262, 0.00026548, -0.00053128, 0.00022775, -0.00021868, 0.00014997, -0.00013377, 0.00009725, -0.00010050, 0.00007384, -0.00007610, 0.00007114, -0.00006426, 0.00005649, -0.00006403, 0.00005518, -0.00005323, 0.00005012, -0.00004354, 0.00004086, -0.00004199, 0.00003269, -0.00002905, 0.00002760, -0.00003300, 0.00003032, -0.00002096, 0.00001341, -0.00001063, 0.00001220, -0.00001455, 0.00001219, -0.00002358, 0.00002704, -0.00001027, 0.00000313, -0.00001246, 0.00000260, 0.00001015, -0.00001647, 0.00000858, 0.00000754, -0.00000566, 0.00000611, -0.00000566, 0.00001484, -0.00002788, 0.00003364, -0.00002766, 0.00001697, -0.00001966, 0.00001475, 0.00000284, -0.00000150, -0.00001616, 0.00001485, 0.00000261, -0.00000775, -0.00000371, 0.00000526, -0.00000557, },
		{ -0.00000030, 0.00000511, 0.00000400, -0.00001111, 0.00000746, -0.00000881, 0.00000180, 0.00000506, 0.00000036, -0.00000226, 0.00000406, -0.00000843, -0.00000262, 0.00001530, -0.00000444, -0.00000051, -0.00001723, 0.00002390, -0.00001203, 0.00001077, -0.00002204, 0.00002636, -0.00001257, -0.00000516, 0.00000979, -0.00001909, 0.00001748, -0.00001173, 0.00000682, -0.00000348, -0.00001546, 0.00003083, -0.00002592, 0.00000636, -0.00000644, 0.00001506, -0.00000169, -0.00000752, 0.00000433, -0.00000623, 0.00000306, -0.00000237, 0.00000175, 0.00000313, 0.00000279, -0.00000840, 0.00000295, -0.00000519, 0.00000554, -0.00000698, 0.00000288, 0.00000067, 0.00000091, -0.00000782, 0.00001156, -0.00001965, 0.00000800, -0.00000579, 0.00000539, -0.00001915, 0.00001569, -0.00003004, 0.00001409, -0.00003129, 0.00001790, -0.00008559, 0.00002850, -0.00057446, 0.00133135, -0.00047674, -0.00005501, -0.00003490, -0.00001819, 0.00000080, -0.00001285, -0.00000610, -0.00001002, -0.00000696, 0.00000173, 0.00000162, -0.00000196, -0.00000324, -0.00000872, 0.00000616, -0.00000803, 0.00001050, -0.00000821, 0.00000776, -0.00001381, 0.00000842, -0.00000486, 0.00000266, -0.00000996, 0.00001354, -0.00000937, 0.00000188, 0.00000381, -0.00000487, 0.00000346, -0.00000881, -0.00000187, 0.00001431, -0.00000483, -0.00000585, -0.00000369, -0.00000047, 0.00001236, -0.00002642, 0.00003012, -0.00001580, 0.00001580, -0.00001631, 0.00002248, -0.00001907, 0.00000650, 0.00000674, -0.00001169, 0.00000556, -0.00001129, 0.00001312, 0.00000459, -0.00001150, -0.00000535, 0.00001342, -0.00000060, -0.00001097, 0.00000408, -0.00000150, },
		{ -0.00000776, 0.00000604, 0.00000276, 0.00000151, -0.00001028, 0.00000957, -0.00001400, 0.00001057, -0.00000168, 0.00000361, -0.00000491, 0.00000680, -0.00001592, 0.00001151, 0.00000455, 0.00000150, -0.00001135, -0.00000359, 0.00001723, -0.00001016, 0.00000662, -0.00001876, 0.00003460, -0.00003470, 0.00002175, -0.00001692, 0.00000271, 0.00000415, -0.00000499, 0.00000344, -0.00000740, -0.00000659, 0.00002981, -0.00004021, 0.00002718, -0.00002414, 0.00003754, -0.00003199, 0.00002239, -0.00002333, 0.00002038, -0.00002288, 0.00002386, -0.00002389, 0.00003184, -0.00003057, 0.00002481, -0.00003059, 0.00003035, -0.00003080, 0.00002974, -0.00003455, 0.00004247, -0.00004842, 0.00004747, -0.00004767, 0.00003913, -0.00005129, 0.00006129, -0.00007485, 0.00007130, -0.00008455, 0.00008014, -0.00011312, 0.00012302, -0.00018794, 0.00018305, -0.00043871, 0.00011307, 0.00104608, -0.00075949, 0.00024990, -0.00023362, 0.00014684, -0.00012935, 0.00009426, -0.00009794, 0.00006988, -0.00007672, 0.00007029, -0.00006222, 0.00005601, -0.00005973, 0.00004564, -0.00004366, 0.00003971, -0.00003334, 0.00003178, -0.00003072, 0.00002011, -0.00001992, 0.00002160, -0.00002469, 0.00001749, -0.00000895, 0.00000566, -0.00000867, 0.00001390, -0.00001608, 0.00001468, -0.00002354, 0.00001790, -0.00000032, 0.00000008, -0.00001082, 0.00000133, 0.00000116, 0.00000411, -0.00001872, 0.00003259, -0.00002294, 0.00002181, -0.00002107, 0.00003172, -0.00003740, 0.00003318, -0.00002182, 0.00001090, -0.00001323, 0.00000509, 0.00000513, 0.00000800, -0.00002180, 0.00000987, 0.00000521, -0.00000120, -0.00001209, 0.00000965, },
		{ 0.00000822, -0.00000800, 0.00000858, 0.00000199, -0.00000363, -0.00000374, 0.00000383, -0.00001030, 0.00001183, -0.00000380, 0.00000328, -0.00000321, 0.00000312, -0.00001470, 0.00001908, -0.00000434, 0.00000658, -0.00002182, 0.00001439, 0.00000144, 0.00000219, -0.00001033, 0.00000467, 0.00001474, -0.00002780, 0.00002500, -0.00002782, 0.00001671, -0.00000616, -0.00000056, 0.00000237, -0.00001506, 0.00001394, 0.00000585, -0.00002617, 0.00002174, -0.00001607, 0.00003120, -0.00003399, 0.00002756, -0.00002847, 0.00002513, -0.00002695, 0.00002800, -0.00002580, 0.00003405, -0.00003669, 0.00003103, -0.00003632, 0.00003734, -0.00003770, 0.00003505, -0.00003686, 0.00004404, -0.00005375, 0.00005748, -0.00006171, 0.00005155, -0.00005877, 0.00006804, -0.00008809, 0.00009065, -0.00010868, 0.00010541, -0.00014118, 0.00015680, -0.00024944, 0.00027080, -0.00073585, 0.00089056, 0.00030429, -0.00053285, 0.00021926, -0.00021075, 0.00014380, -0.00012842, 0.00009281, -0.00009653, 0.00006737, -0.00006996, 0.00006768, -0.00006211, 0.00005542, -0.00006198, 0.00005076, -0.00004853, 0.00004660, -0.00004086, 0.00003895, -0.00004024, 0.00002916, -0.00002511, 0.00002499, -0.00003001, 0.00002643, -0.00001766, 0.00000921, -0.00000655, 0.00000906, -0.00001087, 0.00000781, -0.00001879, 0.00002134, -0.00000479, -0.00000372, -0.00000521, -0.00000320, 0.00001044, -0.00001421, 0.00000632, 0.00001300, -0.00000868, 0.00000710, -0.00000193, 0.00001163, -0.00002319, 0.00002917, -0.00002683, 0.00001701, -0.00001971, 0.00001332, 0.00000313, 0.00000251, -0.00002081, 0.00001767, -0.00000122, -0.00000370, -0.00000751, },
		{ 0.00000219, -0.00000001, -0.00000129, 0.00000579, 0.00000316, -0.00000938, 0.00000496, -0.00000622, 0.00000019, 0.00000555, -0.00000024, -0.00000124, 0.00000251, -0.00000643, -0.00000329, 0.00001478, -0.00000364, 0.00000197, -0.00002075, 0.00002303, -0.00000893, 0.00000972, -0.00002228, 0.00002861, -0.00001671, -0.00000050, 0.00000473, -0.00001693, 0.00001497, -0.00000744, -0.00000119, 0.00000282, -0.00002011, 0.00003368, -0.00002698, 0.00000590, -0.00000357, 0.00001254, -0.00000045, -0.00000737, 0.00000456, -0.00000650, 0.00000315, -0.00000419, 0.00000559, -0.00000053, 0.00000631, -0.00001106, 0.00000500, -0.00000867, 0.00001012, -0.00001017, 0.00000526, -0.00000156, 0.00000271, -0.00001022, 0.00001436, -0.00002080, 0.00000801, -0.00000577, 0.00000451, -0.00001987, 0.00001903, -0.00003229, 0.00001692, -0.00003455, 0.00002011, -0.00008773, 0.00002901, -0.00055307, 0.00126921, -0.00041273, -0.00008872, -0.00001776, -0.00002925, 0.00000960, -0.00002014, 0.00000070, -0.00001701, -0.00000332, 0.00000017, 0.00000410, -0.00000529, 0.00000167, -0.00001363, 0.00000863, -0.00000896, 0.00001148, -0.00000952, 0.00000985, -0.00001674, 0.00000993, -0.00000525, 0.00000405, -0.00001126, 0.00001370, -0.00001004, 0.00000165, 0.00000442, -0.00000460, 0.00000339, -0.00000940, -0.00000065, 0.00001160, -0.00000193, -0.00001072, 0.00000362, -0.00000947, 0.00001744, -0.00002822, 0.00003142, -0.00001227, 0.00001201, -0.00001231, 0.00002301, -0.00002007, 0.00001016, 0.00000059, -0.00000764, 0.00000253, -0.00000862, 0.00000847, 0.00000987, -0.00001299, -0.00000585, 0.00001270, -0.00000105, -0.00000957, },
		{ -0.00001406, 0.00001059, -0.00000855, 0.00000715, 0.00000115, 0.00000335, -0.00001140, 0.00000955, -0.00001308, 0.00000995, -0.00000224, 0.00000451, -0.00000566, 0.00000698, -0.00001495, 0.00001127, 0.00000317, 0.00000459, -0.00001176, -0.00000605, 0.00001654, -0.00000566, 0.00000212, -0.00001419, 0.00003092, -0.00003304, 0.00002157, -0.00001876, 0.00000241, 0.00000508, -0.00000508, -0.00000128, -0.00000192, -0.00001184, 0.00003409, -0.00004336, 0.00002931, -0.00002324, 0.00003537, -0.00003007, 0.00002189, -0.00002288, 0.00002013, -0.00002355, 0.00002364, -0.00002128, 0.00002836, -0.00002639, 0.00002092, -0.00002749, 0.00002606, -0.00002440, 0.00002348, -0.00002857, 0.00003623, -0.00004238, 0.00004063, -0.00003945, 0.00003149, -0.00004449, 0.00005431, -0.00006902, 0.00006460, -0.00007430, 0.00006929, -0.00010010, 0.00010680, -0.00016911, 0.00015870, -0.00040399, 0.00008787, 0.00102211, -0.00070707, 0.00020965, -0.00020624, 0.00012662, -0.00011285, 0.00008035, -0.00008522, 0.00005645, -0.00006597, 0.00006280, -0.00005548, 0.00004942, -0.00005159, 0.00003634, -0.00003592, 0.00003392, -0.00002842, 0.00002729, -0.00002591, 0.00001402, -0.00001456, 0.00001767, -0.00002008, 0.00001260, -0.00000524, 0.00000149, -0.00000516, 0.00001104, -0.00001236, 0.00001062, -0.00001955, 0.00001404, 0.00000210, -0.00000236, -0.00000984, 0.00000349, -0.00000574, 0.00001012, -0.00002240, 0.00003679, -0.00002230, 0.00001914, -0.00001533, 0.00002867, -0.00003407, 0.00003187, -0.00002473, 0.00001372, -0.00001569, 0.00000708, 0.00000086, 0.00001491, -0.00002665, 0.00001252, 0.00000138, 0.00000192, },
		{ 0.00000038, -0.00001097, 0.00001059, -0.00000984, 0.00001045, -0.00000047, -0.00000038, -0.00000688, 0.00000604, -0.00001109, 0.00001205, -0.00000485, 0.00000489, -0.00000503, 0.00000461, -0.00001447, 0.00001845, -0.00000485, 0.00000992, -0.00002377, 0.00001214, 0.00000282, 0.00000468, -0.00001414, 0.00000997, 0.00000834, -0.00002198, 0.00002010, -0.00002574, 0.00001351, -0.00000165, -0.00000595, 0.00000335, -0.00001355, 0.00001102, 0.00000900, -0.00002929, 0.00002526, -0.00001722, 0.00003027, -0.00003214, 0.00002671, -0.00002783, 0.00002477, -0.00002822, 0.00002985, -0.00002601, 0.00003320, -0.00003471, 0.00002884, -0.00003518, 0.00003585, -0.00003401, 0.00003071, -0.00003237, 0.00003895, -0.00004871, 0.00005174, -0.00005402, 0.00004346, -0.00005105, 0.00005969, -0.00008095, 0.00008396, -0.00009883, 0.00009460, -0.00012808, 0.00013941, -0.00022753, 0.00023963, -0.00068031, 0.00081910, 0.00033796, -0.00052288, 0.00020183, -0.00019620, 0.00013233, -0.00011887, 0.00008495, -0.00008953, 0.00005837, -0.00006153, 0.00006189, -0.00005765, 0.00005198, -0.00005740, 0.00004400, -0.00004180, 0.00004116, -0.00003639, 0.00003533, -0.00003690, 0.00002441, -0.00002026, 0.00002151, -0.00002608, 0.00002175, -0.00001403, 0.00000514, -0.00000281, 0.00000616, -0.00000728, 0.00000346, -0.00001378, 0.00001543, 0.00000032, -0.00000958, 0.00000081, -0.00000743, 0.00000864, -0.00001041, 0.00000323, 0.00001867, -0.00001128, 0.00000787, 0.00000215, 0.00000813, -0.00001773, 0.00002325, -0.00002479, 0.00001601, -0.00001890, 0.00001171, 0.00000298, 0.00000659, -0.00002486, 0.00001995, -0.00000500, },
		{ -0.00000193, -0.00000746, -0.00000021, 0.00000172, -0.00000237, 0.00000656, 0.00000197, -0.00000703, 0.00000188, -0.00000329, -0.00000144, 0.00000596, -0.00000090, -0.00000005, 0.00000077, -0.00000438, -0.00000358, 0.00001354, -0.00000228, 0.00000378, -0.00002287, 0.00002051, -0.00000462, 0.00000780, -0.00002149, 0.00002944, -0.00001995, 0.00000396, -0.00000052, -0.00001415, 0.00001184, -0.00000273, -0.00000885, 0.00000798, -0.00002258, 0.00003372, -0.00002585, 0.00000398, 0.00000033, 0.00000934, 0.00000092, -0.00000706, 0.00000458, -0.00000655, 0.00000313, -0.00000589, 0.00000919, -0.00000385, 0.00000921, -0.00001299, 0.00000644, -0.00001138, 0.00001370, -0.00001237, 0.00000680, -0.00000309, 0.00000375, -0.00001153, 0.00001562, -0.00002036, 0.00000679, -0.00000485, 0.00000277, -0.00001932, 0.00002058, -0.00003239, 0.00001745, -0.00003509, 0.00001874, -0.00008478, 0.00002093, -0.00051487, 0.00118681, -0.00033888, -0.00012269, -0.00000213, -0.00003877, 0.00001686, -0.00002615, 0.00000653, -0.00002308, -0.00000025, -0.00000107, 0.00000617, -0.00000815, 0.00000592, -0.00001751, 0.00000998, -0.00000886, 0.00001139, -0.00000974, 0.00001074, -0.00001835, 0.00001027, -0.00000474, 0.00000460, -0.00001154, 0.00001276, -0.00000983, 0.00000101, 0.00000504, -0.00000424, 0.00000326, -0.00000990, 0.00000078, 0.00000838, 0.00000121, -0.00001524, 0.00001016, -0.00001721, 0.00002027, -0.00002765, 0.00003062, -0.00000702, 0.00000723, -0.00000723, 0.00002256, -0.00001986, 0.00001277, -0.00000540, -0.00000354, -0.00000071, -0.00000556, 0.00000384, 0.00001428, -0.00001356, -0.00000662, 0.00001203, },
		{ -0.00000188, 0.00000430, -0.00001494, 0.00001042, -0.00000843, 0.00000755, 0.00000009, 0.00000452, -0.00001151, 0.00000838, -0.00001106, 0.00000848, -0.00000213, 0.00000468, -0.00000552, 0.00000611, -0.00001287, 0.00001016, 0.00000230, 0.00000705, -0.00001148, -0.00000847, 0.00001519, -0.00000058, -0.00000265, -0.00000920, 0.00002609, -0.00003017, 0.00002058, -0.00002015, 0.00000235, 0.00000534, -0.00000415, -0.00000674, 0.00000379, -0.00001620, 0.00003616, -0.00004404, 0.00002919, -0.00002023, 0.00003120, -0.00002655, 0.00002014, -0.00002131, 0.00001885, -0.00002308, 0.00002224, -0.00001754, 0.00002377, -0.00002132, 0.00001628, -0.00002352, 0.00002096, -0.00001747, 0.00001681, -0.00002207, 0.00002924, -0.00003541, 0.00003294, -0.00003082, 0.00002373, -0.00003731, 0.00004645, -0.00006181, 0.00005646, -0.00006279, 0.00005730, -0.00008568, 0.00008890, -0.00014839, 0.00013234, -0.00036862, 0.00007097, 0.00097929, -0.00063852, 0.00016117, -0.00017404, 0.00010299, -0.00009392, 0.00006451, -0.00007078, 0.00004163, -0.00005378, 0.00005368, -0.00004714, 0.00004131, -0.00004208, 0.00002610, -0.00002755, 0.00002756, -0.00002305, 0.00002245, -0.00002097, 0.00000816, -0.00000956, 0.00001400, -0.00001568, 0.00000817, -0.00000236, -0.00000172, -0.00000240, 0.00000859, -0.00000889, 0.00000676, -0.00001565, 0.00001050, 0.00000366, -0.00000375, -0.00000965, 0.00000617, -0.00001279, 0.00001515, -0.00002447, 0.00003916, -0.00001978, 0.00001519, -0.00000832, 0.00002441, -0.00002913, 0.00002861, -0.00002631, 0.00001554, -0.00001745, 0.00000889, -0.00000323, 0.00002083, -0.00003007, 0.00001417, },
		{ 0.00002135, -0.00000835, 0.00000406, -0.00001368, 0.00001204, -0.00001081, 0.00001161, -0.00000249, 0.00000245, -0.00000935, 0.00000732, -0.00001083, 0.00001133, -0.00000513, 0.00000575, -0.00000596, 0.00000500, -0.00001288, 0.00001642, -0.00000424, 0.00001194, -0.00002403, 0.00000862, 0.00000475, 0.00000678, -0.00001726, 0.00001446, 0.00000201, -0.00001570, 0.00001466, -0.00002311, 0.00001027, 0.00000231, -0.00001003, 0.00000268, -0.00001025, 0.00000688, 0.00001201, -0.00003145, 0.00002752, -0.00001694, 0.00002764, -0.00002868, 0.00002444, -0.00002585, 0.00002317, -0.00002810, 0.00003017, -0.00002464, 0.00003067, -0.00003115, 0.00002519, -0.00003243, 0.00003268, -0.00002884, 0.00002510, -0.00002665, 0.00003242, -0.00004189, 0.00004403, -0.00004466, 0.00003411, -0.00004208, 0.00004973, -0.00007151, 0.00007444, -0.00008598, 0.00008065, -0.00011130, 0.00011751, -0.00019988, 0.00020058, -0.00061403, 0.00074192, 0.00036419, -0.00050094, 0.00017598, -0.00017571, 0.00011622, -0.00010571, 0.00007418, -0.00007999, 0.00004739, -0.00005131, 0.00005414, -0.00005116, 0.00004644, -0.00005067, 0.00003540, -0.00003354, 0.00003426, -0.00003055, 0.00003039, -0.00003237, 0.00001888, -0.00001493, 0.00001757, -0.00002161, 0.00001672, -0.00001050, 0.00000159, 0.00000024, 0.00000379, -0.00000412, -0.00000046, -0.00000900, 0.00000983, 0.00000452, -0.00001396, 0.00000514, -0.00000986, 0.00000494, -0.00000559, -0.00000011, 0.00002408, -0.00001314, 0.00000831, 0.00000654, 0.00000452, -0.00001186, 0.00001614, -0.00002170, 0.00001408, -0.00001733, 0.00000995, 0.00000255, 0.00001038, -0.00002792, },
		{ -0.00000761, 0.00001133, -0.00000310, -0.00000487, -0.00000285, 0.00000344, -0.00000333, 0.00000723, 0.00000064, -0.00000435, -0.00000141, -0.00000039, -0.00000283, 0.00000611, -0.00000143, 0.00000111, -0.00000090, -0.00000257, -0.00000333, 0.00001152, -0.00000032, 0.00000479, -0.00002352, 0.00001649, 0.00000060, 0.00000523, -0.00001977, 0.00002882, -0.00002209, 0.00000792, -0.00000563, -0.00001104, 0.00000845, 0.00000191, -0.00001560, 0.00001153, -0.00002263, 0.00003105, -0.00002277, 0.00000092, 0.00000489, 0.00000581, 0.00000218, -0.00000642, 0.00000429, -0.00000628, 0.00000285, -0.00000721, 0.00001220, -0.00000645, 0.00001116, -0.00001396, 0.00000703, -0.00001305, 0.00001600, -0.00001332, 0.00000733, -0.00000379, 0.00000392, -0.00001165, 0.00001531, -0.00001841, 0.00000449, -0.00000318, 0.00000032, -0.00001759, 0.00002034, -0.00003038, 0.00001576, -0.00003300, 0.00001403, -0.00007723, 0.00000539, -0.00046246, 0.00108834, -0.00025935, -0.00015414, 0.00001037, -0.00004571, 0.00002183, -0.00003028, 0.00001085, -0.00002779, 0.00000193, -0.00000173, 0.00000758, -0.00001017, 0.00000913, -0.00002004, 0.00001003, -0.00000767, 0.00001020, -0.00000884, 0.00001040, -0.00001861, 0.00000948, -0.00000340, 0.00000437, -0.00001090, 0.00001086, -0.00000892, 0.00000013, 0.00000552, -0.00000369, 0.00000294, -0.00001015, 0.00000218, 0.00000500, 0.00000413, -0.00001885, 0.00001525, -0.00002299, 0.00002044, -0.00002474, 0.00002798, -0.00000043, 0.00000186, -0.00000140, 0.00002134, -0.00001849, 0.00001417, -0.00001095, 0.00000033, -0.00000388, -0.00000244, -0.00000037, 0.00001756, -0.00001317, },
		{ -0.00003173, 0.00001460, -0.00000433, 0.00000575, -0.00001460, 0.00000913, -0.00000743, 0.00000726, -0.00000041, 0.00000497, -0.00001067, 0.00000620, -0.00000816, 0.00000633, -0.00000145, 0.00000419, -0.00000461, 0.00000439, -0.00000994, 0.00000840, 0.00000179, 0.00000898, -0.00001069, -0.00001064, 0.00001318, 0.00000481, -0.00000738, -0.00000414, 0.00002043, -0.00002625, 0.00001883, -0.00002107, 0.00000245, 0.00000504, -0.00000250, -0.00001243, 0.00000911, -0.00001913, 0.00003576, -0.00004217, 0.00002686, -0.00001533, 0.00002534, -0.00002172, 0.00001737, -0.00001879, 0.00001667, -0.00002160, 0.00001984, -0.00001299, 0.00001847, -0.00001581, 0.00001132, -0.00001911, 0.00001556, -0.00001057, 0.00001034, -0.00001565, 0.00002207, -0.00002810, 0.00002506, -0.00002246, 0.00001640, -0.00003027, 0.00003823, -0.00005377, 0.00004745, -0.00005076, 0.00004497, -0.00007081, 0.00007045, -0.00012709, 0.00010551, -0.00033409, 0.00006266, 0.00091955, -0.00055713, 0.00010750, -0.00013933, 0.00007774, -0.00007400, 0.00004796, -0.00005571, 0.00002646, -0.00004105, 0.00004366, -0.00003789, 0.00003239, -0.00003198, 0.00001574, -0.00001923, 0.00002118, -0.00001771, 0.00001772, -0.00001635, 0.00000299, -0.00000530, 0.00001089, -0.00001181, 0.00000449, -0.00000049, -0.00000383, -0.00000054, 0.00000668, -0.00000583, 0.00000330, -0.00001201, 0.00000742, 0.00000431, -0.00000417, -0.00001001, 0.00000890, -0.00001927, 0.00001847, -0.00002454, 0.00003956, -0.00001546, 0.00001025, -0.00000035, 0.00001934, -0.00002297, 0.00002357, -0.00002653, 0.00001625, -0.00001834, 0.00001023, -0.00000669, 0.00002528, },
		{ 0.00001352, -0.00002966, 0.00002157, -0.00001092, 0.00000696, -0.00001533, 0.00001234, -0.00001074, 0.00001190, -0.00000383, 0.00000459, -0.00001090, 0.00000753, -0.00000950, 0.00000968, -0.00000462, 0.00000579, -0.00000596, 0.00000429, -0.00001005, 0.00001319, -0.00000264, 0.00001274, -0.00002279, 0.00000417, 0.00000699, 0.00000847, -0.00001953, 0.00001785, -0.00000384, -0.00000935, 0.00000901, -0.00002019, 0.00000722, 0.00000550, -0.00001269, 0.00000052, -0.00000558, 0.00000210, 0.00001443, -0.00003230, 0.00002827, -0.00001510, 0.00002340, -0.00002375, 0.00002089, -0.00002265, 0.00002038, -0.00002660, 0.00002894, -0.00002178, 0.00002663, -0.00002627, 0.00002038, -0.00002833, 0.00002813, -0.00002259, 0.00001871, -0.00002021, 0.00002500, -0.00003389, 0.00003505, -0.00003441, 0.00002425, -0.00003259, 0.00003888, -0.00006049, 0.00006289, -0.00007106, 0.00006462, -0.00009210, 0.00009266, -0.00016847, 0.00015637, -0.00054063, 0.00066257, 0.00038123, -0.00046738, 0.00014287, -0.00015046, 0.00009650, -0.00008983, 0.00006128, -0.00006863, 0.00003514, -0.00003996, 0.00004497, -0.00004310, 0.00003926, -0.00004237, 0.00002565, -0.00002441, 0.00002649, -0.00002388, 0.00002462, -0.00002714, 0.00001306, -0.00000960, 0.00001357, -0.00001706, 0.00001180, -0.00000742, -0.00000112, 0.00000236, 0.00000217, -0.00000166, -0.00000365, -0.00000483, 0.00000494, 0.00000746, -0.00001655, 0.00000757, -0.00001048, -0.00000025, -0.00000043, -0.00000313, 0.00002877, -0.00001396, 0.00000833, 0.00001119, 0.00000103, -0.00000596, 0.00000824, -0.00001781, 0.00001139, -0.00001511, 0.00000806, 0.00000205, },
		{ 0.00001958, -0.00001191, -0.00000869, 0.00001052, -0.00000439, -0.00000208, -0.00000542, 0.00000485, -0.00000394, 0.00000761, -0.00000060, -0.00000168, -0.00000453, 0.00000215, -0.00000376, 0.00000586, -0.00000170, 0.00000203, -0.00000225, -0.00000121, -0.00000243, 0.00000876, 0.00000222, 0.00000498, -0.00002277, 0.00001131, 0.00000634, 0.00000230, -0.00001728, 0.00002681, -0.00002300, 0.00001112, -0.00001033, -0.00000793, 0.00000513, 0.00000604, -0.00002102, 0.00001321, -0.00002033, 0.00002602, -0.00001818, -0.00000285, 0.00000964, 0.00000237, 0.00000305, -0.00000532, 0.00000359, -0.00000561, 0.00000223, -0.00000796, 0.00001433, -0.00000807, 0.00001198, -0.00001382, 0.00000663, -0.00001352, 0.00001686, -0.00001294, 0.00000682, -0.00000365, 0.00000325, -0.00001063, 0.00001357, -0.00001521, 0.00000141, -0.00000103, -0.00000258, -0.00001489, 0.00001852, -0.00002653, 0.00001213, -0.00002868, 0.00000655, -0.00006600, -0.00001589, -0.00039921, 0.00097858, -0.00017842, -0.00018053, 0.00001846, -0.00004933, 0.00002402, -0.00003216, 0.00001335, -0.00003084, 0.00000303, -0.00000170, 0.00000813, -0.00001114, 0.00001105, -0.00002107, 0.00000881, -0.00000548, 0.00000803, -0.00000695, 0.00000894, -0.00001763, 0.00000774, -0.00000142, 0.00000353, -0.00000953, 0.00000827, -0.00000756, -0.00000078, 0.00000571, -0.00000286, 0.00000234, -0.00001000, 0.00000331, 0.00000184, 0.00000640, -0.00002109, 0.00001839, -0.00002635, 0.00001785, -0.00001984, 0.00002399, 0.00000698, -0.00000356, 0.00000478, 0.00001970, -0.00001616, 0.00001426, -0.00001581, 0.00000373, -0.00000670, 0.00000045, -0.00000390, },
		{ -0.00000918, 0.00002794, -0.00003148, 0.00001371, -0.00000585, 0.00000617, -0.00001310, 0.00000684, -0.00000569, 0.00000639, -0.00000039, 0.00000479, -0.00000903, 0.00000330, -0.00000471, 0.00000380, -0.00000041, 0.00000324, -0.00000314, 0.00000211, -0.00000653, 0.00000631, 0.00000148, 0.00001049, -0.00000959, -0.00001240, 0.00001059, 0.00001023, -0.00001173, 0.00000065, 0.00001429, -0.00002153, 0.00001640, -0.00002149, 0.00000262, 0.00000434, -0.00000046, -0.00001786, 0.00001353, -0.00002030, 0.00003289, -0.00003795, 0.00002261, -0.00000895, 0.00001829, -0.00001604, 0.00001392, -0.00001563, 0.00001384, -0.00001933, 0.00001675, -0.00000806, 0.00001296, -0.00001034, 0.00000649, -0.00001470, 0.00001033, -0.00000426, 0.00000459, -0.00000983, 0.00001528, -0.00002101, 0.00001759, -0.00001491, 0.00000996, -0.00002378, 0.00003012, -0.00004538, 0.00003818, -0.00003889, 0.00003306, -0.00005637, 0.00005246, -0.00010636, 0.00007949, -0.00030145, 0.00006249, 0.00084575, -0.00046688, 0.00005202, -0.00010457, 0.00005271, -0.00005455, 0.00003192, -0.00004111, 0.00001193, -0.00002865, 0.00003348, -0.00002843, 0.00002340, -0.00002209, 0.00000602, -0.00001160, 0.00001528, -0.00001285, 0.00001348, -0.00001241, -0.00000118, -0.00000205, 0.00000855, -0.00000869, 0.00000175, 0.00000028, -0.00000481, 0.00000044, 0.00000531, -0.00000328, 0.00000034, -0.00000876, 0.00000487, 0.00000414, -0.00000382, -0.00001062, 0.00001119, -0.00002449, 0.00001953, -0.00002247, 0.00003809, -0.00000958, 0.00000472, 0.00000815, 0.00001394, -0.00001608, 0.00001706, -0.00002549, 0.00001583, -0.00001826, 0.00001088, },
		{ 0.00000604, 0.00000167, 0.00001570, -0.00002985, 0.00002044, -0.00001245, 0.00000882, -0.00001570, 0.00001137, -0.00000958, 0.00001125, -0.00000437, 0.00000589, -0.00001142, 0.00000668, -0.00000722, 0.00000725, -0.00000341, 0.00000508, -0.00000507, 0.00000264, -0.00000628, 0.00000910, -0.00000032, 0.00001253, -0.00002037, -0.00000078, 0.00000926, 0.00000978, -0.00002086, 0.00001995, -0.00000889, -0.00000329, 0.00000343, -0.00001721, 0.00000453, 0.00000777, -0.00001397, -0.00000280, -0.00000012, -0.00000273, 0.00001586, -0.00003166, 0.00002741, -0.00001181, 0.00001778, -0.00001766, 0.00001635, -0.00001848, 0.00001659, -0.00002386, 0.00002633, -0.00001769, 0.00002146, -0.00002048, 0.00001479, -0.00002329, 0.00002265, -0.00001580, 0.00001211, -0.00001365, 0.00001732, -0.00002539, 0.00002558, -0.00002409, 0.00001464, -0.00002329, 0.00002789, -0.00004869, 0.00005018, -0.00005509, 0.00004768, -0.00007186, 0.00006655, -0.00013544, 0.00010980, -0.00046372, 0.00058430, 0.00038809, -0.00042332, 0.00010425, -0.00012204, 0.00007450, -0.00007236, 0.00004718, -0.00005629, 0.00002244, -0.00002823, 0.00003501, -0.00003406, 0.00003107, -0.00003318, 0.00001551, -0.00001516, 0.00001849, -0.00001697, 0.00001855, -0.00002170, 0.00000745, -0.00000470, 0.00000991, -0.00001279, 0.00000736, -0.00000510, -0.00000282, 0.00000341, 0.00000143, -0.00000005, -0.00000589, -0.00000151, 0.00000104, 0.00000893, -0.00001726, 0.00000811, -0.00000952, -0.00000635, 0.00000429, -0.00000528, 0.00003239, -0.00001353, 0.00000793, 0.00001601, -0.00000209, -0.00000049, -0.00000002, -0.00001342, 0.00000817, -0.00001240, },
		{ 0.00000285, -0.00000654, 0.00002034, -0.00000994, -0.00000968, 0.00000945, -0.00000556, 0.00000060, -0.00000760, 0.00000570, -0.00000404, 0.00000754, -0.00000153, 0.00000069, -0.00000716, 0.00000406, -0.00000408, 0.00000512, -0.00000159, 0.00000257, -0.00000311, -0.00000045, -0.00000089, 0.00000538, 0.00000517, 0.00000447, -0.00002086, 0.00000540, 0.00001213, -0.00000068, -0.00001426, 0.00002356, -0.00002263, 0.00001338, -0.00001435, -0.00000507, 0.00000216, 0.00000934, -0.00002482, 0.00001300, -0.00001601, 0.00001925, -0.00001267, -0.00000681, 0.00001410, -0.00000061, 0.00000334, -0.00000368, 0.00000246, -0.00000451, 0.00000119, -0.00000802, 0.00001541, -0.00000857, 0.00001160, -0.00001259, 0.00000528, -0.00001280, 0.00001630, -0.00001130, 0.00000538, -0.00000282, 0.00000189, -0.00000868, 0.00001071, -0.00001118, -0.00000207, 0.00000127, -0.00000563, -0.00001153, 0.00001546, -0.00002126, 0.00000708, -0.00002276, -0.00000281, -0.00005237, -0.00004073, -0.00032894, 0.00086258, -0.00010023, -0.00019973, 0.00002129, -0.00004927, 0.00002325, -0.00003168, 0.00001389, -0.00003212, 0.00000298, -0.00000095, 0.00000780, -0.00001098, 0.00001162, -0.00002064, 0.00000647, -0.00000254, 0.00000513, -0.00000431, 0.00000661, -0.00001568, 0.00000531, 0.00000091, 0.00000232, -0.00000770, 0.00000531, -0.00000603, -0.00000152, 0.00000550, -0.00000168, 0.00000137, -0.00000932, 0.00000397, -0.00000083, 0.00000773, -0.00002169, 0.00001930, -0.00002712, 0.00001275, -0.00001356, 0.00001928, 0.00001461, -0.00000849, 0.00001088, 0.00001799, -0.00001313, 0.00001302, -0.00001977, 0.00000643, -0.00000896, },
		{ -0.00001719, 0.00001070, -0.00001048, 0.00002868, -0.00002936, 0.00001158, -0.00000643, 0.00000564, -0.00001062, 0.00000382, -0.00000347, 0.00000512, 0.00000001, 0.00000412, -0.00000686, 0.00000004, -0.00000110, 0.00000120, 0.00000076, 0.00000204, -0.00000137, -0.00000040, -0.00000300, 0.00000420, 0.00000118, 0.00001171, -0.00000836, -0.00001364, 0.00000760, 0.00001532, -0.00001535, 0.00000484, 0.00000804, -0.00001627, 0.00001339, -0.00002142, 0.00000278, 0.00000340, 0.00000163, -0.00002255, 0.00001663, -0.00001960, 0.00002784, -0.00003181, 0.00001693, -0.00000168, 0.00001065, -0.00001002, 0.00001021, -0.00001215, 0.00001065, -0.00001656, 0.00001331, -0.00000318, 0.00000772, -0.00000536, 0.00000218, -0.00001068, 0.00000572, 0.00000102, 0.00000002, -0.00000504, 0.00000929, -0.00001460, 0.00001099, -0.00000860, 0.00000472, -0.00001814, 0.00002246, -0.00003707, 0.00002913, -0.00002776, 0.00002221, -0.00004306, 0.00003577, -0.00008711, 0.00005521, -0.00027124, 0.00006931, 0.00076145, -0.00037219, -0.00000181, -0.00007210, 0.00002961, -0.00003690, 0.00001748, -0.00002794, -0.00000104, -0.00001737, 0.00002385, -0.00001946, 0.00001505, -0.00001313, -0.00000240, -0.00000516, 0.00001027, -0.00000878, 0.00001000, -0.00000935, -0.00000417, 0.00000008, 0.00000706, -0.00000637, -0.00000005, 0.00000004, -0.00000480, 0.00000065, 0.00000445, -0.00000124, -0.00000206, -0.00000591, 0.00000281, 0.00000331, -0.00000298, -0.00001114, 0.00001259, -0.00002790, 0.00001804, -0.00001840, 0.00003509, -0.00000258, -0.00000090, 0.00001668, 0.00000876, -0.00000904, 0.00000949, -0.00002337, 0.00001437, },
		{ 0.00000467, -0.00000935, 0.00000391, 0.00000156, 0.00001671, -0.00002841, 0.00001793, -0.00001284, 0.00000949, -0.00001476, 0.00000918, -0.00000744, 0.00000974, -0.00000412, 0.00000631, -0.00001095, 0.00000491, -0.00000423, 0.00000429, -0.00000169, 0.00000376, -0.00000349, 0.00000032, -0.00000195, 0.00000457, 0.00000235, 0.00001162, -0.00001719, -0.00000579, 0.00001131, 0.00001074, -0.00002117, 0.00002066, -0.00001287, 0.00000217, -0.00000181, -0.00001436, 0.00000231, 0.00000911, -0.00001406, -0.00000684, 0.00000551, -0.00000703, 0.00001606, -0.00002949, 0.00002503, -0.00000729, 0.00001119, -0.00001086, 0.00001120, -0.00001369, 0.00001213, -0.00002014, 0.00002263, -0.00001273, 0.00001561, -0.00001427, 0.00000891, -0.00001776, 0.00001675, -0.00000902, 0.00000587, -0.00000754, 0.00001000, -0.00001708, 0.00001638, -0.00001447, 0.00000596, -0.00001482, 0.00001745, -0.00003688, 0.00003723, -0.00003911, 0.00003098, -0.00005194, 0.00004085, -0.00010285, 0.00006361, -0.00038670, 0.00050981, 0.00038453, -0.00037061, 0.00006234, -0.00009228, 0.00005171, -0.00005451, 0.00003291, -0.00004386, 0.00001012, -0.00001689, 0.00002498, -0.00002469, 0.00002252, -0.00002386, 0.00000575, -0.00000649, 0.00001092, -0.00001039, 0.00001272, -0.00001654, 0.00000249, -0.00000062, 0.00000691, -0.00000913, 0.00000367, -0.00000368, -0.00000344, 0.00000341, 0.00000153, 0.00000067, -0.00000710, 0.00000084, -0.00000177, 0.00000893, -0.00001623, 0.00000699, -0.00000739, -0.00001263, 0.00000785, -0.00000612, 0.00003472, -0.00001177, 0.00000717, 0.00002087, -0.00000463, 0.00000413, -0.00000817, -0.00000889, },
		{ 0.00000828, -0.00001048, 0.00000461, -0.00000821, 0.00001994, -0.00000753, -0.00001040, 0.00000801, -0.00000643, 0.00000287, -0.00000912, 0.00000582, -0.00000350, 0.00000692, -0.00000198, 0.00000251, -0.00000904, 0.00000516, -0.00000370, 0.00000388, -0.00000107, 0.00000266, -0.00000342, -0.00000030, 0.00000118, 0.00000163, 0.00000830, 0.00000347, -0.00001814, -0.00000070, 0.00001746, -0.00000338, -0.00001093, 0.00001928, -0.00002105, 0.00001459, -0.00001754, -0.00000265, -0.00000027, 0.00001159, -0.00002694, 0.00001110, -0.00001022, 0.00001149, -0.00000689, -0.00001045, 0.00001782, -0.00000283, 0.00000293, -0.00000152, 0.00000092, -0.00000304, -0.00000021, -0.00000735, 0.00001539, -0.00000795, 0.00001010, -0.00001041, 0.00000312, -0.00001102, 0.00001448, -0.00000864, 0.00000325, -0.00000151, 0.00000010, -0.00000611, 0.00000714, -0.00000681, -0.00000551, 0.00000338, -0.00000847, -0.00000789, 0.00001162, -0.00001513, 0.00000125, -0.00001600, -0.00001303, -0.00003779, -0.00006678, -0.00025566, 0.00074531, -0.00002845, -0.00021026, 0.00001857, -0.00004558, 0.00001966, -0.00002899, 0.00001260, -0.00003170, 0.00000188, 0.00000042, 0.00000664, -0.00000973, 0.00001091, -0.00001891, 0.00000333, 0.00000082, 0.00000182, -0.00000128, 0.00000375, -0.00001310, 0.00000254, 0.00000328, 0.00000102, -0.00000573, 0.00000231, -0.00000460, -0.00000192, 0.00000480, -0.00000015, 0.00000001, -0.00000806, 0.00000404, -0.00000280, 0.00000793, -0.00002056, 0.00001798, -0.00002545, 0.00000565, -0.00000673, 0.00001459, 0.00002185, -0.00001238, 0.00001650, 0.00001660, -0.00000973, 0.00001054, -0.00002272, },
		{ -0.00002046, 0.00001201, -0.00001520, 0.00000965, -0.00001052, 0.00002753, -0.00002560, 0.00000842, -0.00000624, 0.00000434, -0.00000744, 0.00000039, -0.00000108, 0.00000368, 0.00000060, 0.00000319, -0.00000447, -0.00000322, 0.00000229, -0.00000115, 0.00000182, 0.00000083, 0.00000042, -0.00000280, 0.00000026, 0.00000234, 0.00000075, 0.00001273, -0.00000711, -0.00001433, 0.00000443, 0.00001971, -0.00001795, 0.00000816, 0.00000201, -0.00001076, 0.00000995, -0.00002085, 0.00000284, 0.00000241, 0.00000347, -0.00002616, 0.00001820, -0.00001716, 0.00002109, -0.00002437, 0.00001045, 0.00000580, 0.00000308, -0.00000420, 0.00000664, -0.00000870, 0.00000741, -0.00001358, 0.00000987, 0.00000125, 0.00000316, -0.00000124, -0.00000128, -0.00000736, 0.00000205, 0.00000495, -0.00000311, -0.00000154, 0.00000440, -0.00000920, 0.00000558, -0.00000374, 0.00000080, -0.00001348, 0.00001553, -0.00002920, 0.00002074, -0.00001784, 0.00001287, -0.00003139, 0.00002095, -0.00006994, 0.00003329, -0.00024360, 0.00008143, 0.00067064, -0.00027752, -0.00005077, -0.00004397, 0.00000985, -0.00002211, 0.00000546, -0.00001693, -0.00001177, -0.00000786, 0.00001536, -0.00001152, 0.00000789, -0.00000565, -0.00000904, -0.00000027, 0.00000638, -0.00000571, 0.00000740, -0.00000725, -0.00000594, 0.00000111, 0.00000635, -0.00000483, -0.00000100, -0.00000100, -0.00000403, 0.00000028, 0.00000396, 0.00000031, -0.00000392, -0.00000345, 0.00000115, 0.00000205, -0.00000194, -0.00001128, 0.00001278, -0.00002920, 0.00001400, -0.00001276, 0.00003107, 0.00000500, -0.00000602, 0.00002473, 0.00000434, -0.00000252, 0.00000138, },
		{ -0.00001571, -0.00000460, 0.00000116, -0.00000617, 0.00000172, 0.00000184, 0.00001646, -0.00002542, 0.00001415, -0.00001210, 0.00000900, -0.00001261, 0.00000599, -0.00000454, 0.00000755, -0.00000318, 0.00000595, -0.00000964, 0.00000248, -0.00000086, 0.00000110, 0.00000028, 0.00000208, -0.00000150, -0.00000233, 0.00000247, 0.00000010, 0.00000497, 0.00001037, -0.00001366, -0.00001042, 0.00001292, 0.00001136, -0.00002047, 0.00001994, -0.00001563, 0.00000680, -0.00000651, -0.00001175, 0.00000060, 0.00000958, -0.00001323, -0.00001110, 0.00001071, -0.00001035, 0.00001492, -0.00002592, 0.00002138, -0.00000197, 0.00000415, -0.00000386, 0.00000593, -0.00000871, 0.00000737, -0.00001580, 0.00001822, -0.00000738, 0.00000961, -0.00000817, 0.00000322, -0.00001226, 0.00001097, -0.00000280, 0.00000051, -0.00000239, 0.00000357, -0.00000955, 0.00000811, -0.00000618, -0.00000126, -0.00000767, 0.00000815, -0.00002577, 0.00002484, -0.00002408, 0.00001559, -0.00003356, 0.00001704, -0.00007255, 0.00002021, -0.00031248, 0.00044112, 0.00037109, -0.00031167, 0.00001964, -0.00006314, 0.00002966, -0.00003750, 0.00001942, -0.00003218, -0.00000102, -0.00000667, 0.00001553, -0.00001565, 0.00001432, -0.00001513, -0.00000289, 0.00000097, 0.00000431, -0.00000466, 0.00000757, -0.00001203, -0.00000151, 0.00000240, 0.00000474, -0.00000627, 0.00000088, -0.00000317, -0.00000309, 0.00000250, 0.00000236, 0.00000057, -0.00000732, 0.00000223, -0.00000351, 0.00000763, -0.00001377, 0.00000461, -0.00000463, -0.00001837, 0.00000964, -0.00000544, 0.00003572, -0.00000873, 0.00000623, 0.00002564, -0.00000637, 0.00000746, },
		{ 0.00000696, -0.00002459, 0.00000916, -0.00001117, 0.00000563, -0.00000892, 0.00001858, -0.00000494, -0.00001069, 0.00000610, -0.00000682, 0.00000453, -0.00000980, 0.00000511, -0.00000232, 0.00000576, -0.00000188, 0.00000362, -0.00001003, 0.00000539, -0.00000268, 0.00000223, -0.00000021, 0.00000233, -0.00000317, -0.00000071, 0.00000358, -0.00000217, 0.00001128, 0.00000227, -0.00001501, -0.00000645, 0.00002188, -0.00000551, -0.00000755, 0.00001423, -0.00001841, 0.00001476, -0.00001980, -0.00000080, -0.00000205, 0.00001273, -0.00002743, 0.00000786, -0.00000365, 0.00000356, -0.00000148, -0.00001332, 0.00002047, -0.00000409, 0.00000181, 0.00000108, -0.00000090, -0.00000131, -0.00000190, -0.00000604, 0.00001436, -0.00000634, 0.00000771, -0.00000754, 0.00000040, -0.00000845, 0.00001171, -0.00000528, 0.00000077, -0.00000006, -0.00000181, -0.00000330, 0.00000335, -0.00000260, -0.00000850, 0.00000498, -0.00001080, -0.00000435, 0.00000745, -0.00000874, -0.00000469, -0.00000924, -0.00002303, -0.00002373, -0.00009168, -0.00018320, 0.00063128, 0.00003396, -0.00021139, 0.00001059, -0.00003874, 0.00001373, -0.00002451, 0.00000977, -0.00002981, -0.00000006, 0.00000221, 0.00000484, -0.00000758, 0.00000914, -0.00001622, -0.00000023, 0.00000419, -0.00000149, 0.00000176, 0.00000077, -0.00001025, -0.00000021, 0.00000538, -0.00000010, -0.00000390, -0.00000042, -0.00000348, -0.00000189, 0.00000364, 0.00000168, -0.00000167, -0.00000625, 0.00000350, -0.00000400, 0.00000700, -0.00001784, 0.00001466, -0.00002175, -0.00000266, -0.00000025, 0.00001060, 0.00002817, -0.00001480, 0.00002130, 0.00001589, -0.00000630, },
		{ 0.00000287, -0.00000671, -0.00001712, 0.00000899, -0.00001244, 0.00000780, -0.00000936, 0.00002475, -0.00002057, 0.00000458, -0.00000549, 0.00000254, -0.00000393, -0.00000304, 0.00000116, 0.00000234, 0.00000120, 0.00000222, -0.00000217, -0.00000609, 0.00000510, -0.00000300, 0.00000258, -0.00000018, 0.00000198, -0.00000480, 0.00000295, 0.00000097, 0.00000012, 0.00001358, -0.00000590, -0.00001455, 0.00000136, 0.00002309, -0.00001930, 0.00001039, -0.00000349, -0.00000527, 0.00000623, -0.00001984, 0.00000273, 0.00000151, 0.00000481, -0.00002845, 0.00001819, -0.00001330, 0.00001334, -0.00001638, 0.00000388, 0.00001280, -0.00000377, 0.00000091, 0.00000359, -0.00000559, 0.00000441, -0.00001068, 0.00000674, 0.00000489, -0.00000040, 0.00000178, -0.00000371, -0.00000493, -0.00000051, 0.00000738, -0.00000469, 0.00000059, 0.00000074, -0.00000496, 0.00000150, -0.00000040, -0.00000183, -0.00000983, 0.00000947, -0.00002201, 0.00001328, -0.00000942, 0.00000532, -0.00002167, 0.00000834, -0.00005519, 0.00001399, -0.00021827, 0.00009684, 0.00057747, -0.00018710, -0.00009212, -0.00002175, -0.00000556, -0.00001087, -0.00000358, -0.00000857, -0.00001979, -0.00000054, 0.00000840, -0.00000502, 0.00000233, -0.00000003, -0.00001363, 0.00000293, 0.00000370, -0.00000366, 0.00000567, -0.00000604, -0.00000662, 0.00000120, 0.00000627, -0.00000390, -0.00000130, -0.00000255, -0.00000279, -0.00000043, 0.00000371, 0.00000146, -0.00000531, -0.00000132, -0.00000022, 0.00000059, -0.00000098, -0.00001083, 0.00001159, -0.00002831, 0.00000772, -0.00000622, 0.00002670, 0.00001259, -0.00001011, 0.00003184, 0.00000118, },
		{ -0.00000717, 0.00000915, -0.00002219, -0.00000091, -0.00000206, -0.00000303, -0.00000047, 0.00000253, 0.00001501, -0.00002110, 0.00000938, -0.00001042, 0.00000752, -0.00000951, 0.00000211, -0.00000122, 0.00000497, -0.00000176, 0.00000498, -0.00000774, -0.00000027, 0.00000250, -0.00000195, 0.00000223, 0.00000032, 0.00000062, -0.00000493, 0.00000651, -0.00000386, 0.00000718, 0.00000912, -0.00001019, -0.00001431, 0.00001396, 0.00001163, -0.00001878, 0.00001787, -0.00001711, 0.00001042, -0.00001051, -0.00000947, -0.00000064, 0.00000931, -0.00001180, -0.00001508, 0.00001492, -0.00001239, 0.00001255, -0.00002127, 0.00001687, 0.00000366, -0.00000274, 0.00000276, 0.00000101, -0.00000397, 0.00000273, -0.00001125, 0.00001355, -0.00000214, 0.00000398, -0.00000267, -0.00000181, -0.00000722, 0.00000576, 0.00000240, -0.00000355, 0.00000143, -0.00000156, -0.00000327, 0.00000127, 0.00000032, -0.00000669, -0.00000215, 0.00000040, -0.00001590, 0.00001371, -0.00001080, 0.00000237, -0.00001771, -0.00000366, -0.00004602, -0.00001845, -0.00024341, 0.00037954, 0.00034902, -0.00024931, -0.00002129, -0.00003649, 0.00000974, -0.00002240, 0.00000759, -0.00002196, -0.00001035, 0.00000182, 0.00000727, -0.00000754, 0.00000704, -0.00000762, -0.00000985, 0.00000673, -0.00000092, -0.00000013, 0.00000341, -0.00000842, -0.00000435, 0.00000425, 0.00000347, -0.00000427, -0.00000099, -0.00000346, -0.00000198, 0.00000093, 0.00000368, -0.00000016, -0.00000671, 0.00000280, -0.00000436, 0.00000535, -0.00001033, 0.00000148, -0.00000182, -0.00002292, 0.00000924, -0.00000327, 0.00003556, -0.00000461, 0.00000531, 0.00003016, },
		{ 0.00001613, -0.00000322, 0.00000247, -0.00002540, 0.00000904, -0.00001098, 0.00000590, -0.00000879, 0.00001651, -0.00000246, -0.00001041, 0.00000372, -0.00000667, 0.00000543, -0.00000958, 0.00000359, -0.00000057, 0.00000413, -0.00000124, 0.00000401, -0.00001011, 0.00000480, -0.00000116, 0.00000033, 0.00000090, 0.00000166, -0.00000248, -0.00000151, 0.00000602, -0.00000566, 0.00001380, 0.00000120, -0.00001191, -0.00001134, 0.00002501, -0.00000688, -0.00000431, 0.00000870, -0.00001491, 0.00001394, -0.00002112, 0.00000046, -0.00000315, 0.00001277, -0.00002653, 0.00000380, 0.00000295, -0.00000374, 0.00000300, -0.00001508, 0.00002181, -0.00000433, 0.00000008, 0.00000392, -0.00000282, 0.00000050, -0.00000370, -0.00000424, 0.00001251, -0.00000401, 0.00000476, -0.00000433, -0.00000254, -0.00000543, 0.00000838, -0.00000163, -0.00000168, 0.00000123, -0.00000352, -0.00000062, -0.00000020, 0.00000098, -0.00001069, 0.00000584, -0.00001237, -0.00000123, 0.00000343, -0.00000266, -0.00001006, -0.00000324, -0.00003183, -0.00001155, -0.00011333, -0.00011499, 0.00052434, 0.00008494, -0.00020318, -0.00000181, -0.00002957, 0.00000616, -0.00001881, 0.00000587, -0.00002682, -0.00000254, 0.00000415, 0.00000267, -0.00000482, 0.00000666, -0.00001297, -0.00000375, 0.00000717, -0.00000444, 0.00000442, -0.00000197, -0.00000750, -0.00000262, 0.00000695, -0.00000084, -0.00000244, -0.00000265, -0.00000279, -0.00000144, 0.00000210, 0.00000365, -0.00000357, -0.00000399, 0.00000242, -0.00000448, 0.00000511, -0.00001387, 0.00000983, -0.00001666, -0.00001130, 0.00000495, 0.00000787, 0.00003320, -0.00001544, 0.00002508, },
		{ 0.00003767, -0.00000033, 0.00000656, -0.00001426, -0.00001370, 0.00000558, -0.00000911, 0.00000532, -0.00000723, 0.00002069, -0.00001477, 0.00000047, -0.00000448, 0.00000055, -0.00000048, -0.00000612, 0.00000294, 0.00000130, 0.00000163, 0.00000140, -0.00000023, -0.00000828, 0.00000709, -0.00000416, 0.00000290, -0.00000085, 0.00000313, -0.00000620, 0.00000487, 0.00000018, -0.00000071, 0.00001425, -0.00000474, -0.00001441, -0.00000131, 0.00002516, -0.00001925, 0.00001142, -0.00000823, -0.00000008, 0.00000242, -0.00001845, 0.00000244, 0.00000079, 0.00000550, -0.00002930, 0.00001672, -0.00000851, 0.00000536, -0.00000862, -0.00000211, 0.00001867, -0.00000935, 0.00000492, 0.00000135, -0.00000307, 0.00000186, -0.00000809, 0.00000415, 0.00000751, -0.00000280, 0.00000360, -0.00000507, -0.00000342, -0.00000192, 0.00000834, -0.00000481, 0.00000142, -0.00000173, -0.00000189, -0.00000126, 0.00000155, -0.00000336, -0.00000708, 0.00000433, -0.00001565, 0.00000696, -0.00000265, -0.00000036, -0.00001399, -0.00000193, -0.00004296, -0.00000266, -0.00019476, 0.00011337, 0.00048587, -0.00010464, -0.00012380, -0.00000637, -0.00001613, -0.00000349, -0.00000946, -0.00000301, -0.00002494, 0.00000439, 0.00000319, -0.00000018, -0.00000146, 0.00000360, -0.00001613, 0.00000450, 0.00000216, -0.00000255, 0.00000467, -0.00000553, -0.00000642, 0.00000060, 0.00000658, -0.00000339, -0.00000120, -0.00000429, -0.00000139, -0.00000124, 0.00000354, 0.00000229, -0.00000628, 0.00000056, -0.00000141, -0.00000087, -0.00000031, -0.00000969, 0.00000904, -0.00002543, -0.00000020, 0.00000036, 0.00002266, 0.00001964, -0.00001268, },
		{ 0.00000469, 0.00003433, -0.00000694, 0.00000894, -0.00002722, 0.00000183, -0.00000474, -0.00000015, -0.00000257, 0.00000356, 0.00001254, -0.00001584, 0.00000402, -0.00000809, 0.00000532, -0.00000581, -0.00000203, 0.00000213, 0.00000232, -0.00000011, 0.00000365, -0.00000555, -0.00000298, 0.00000549, -0.00000453, 0.00000387, -0.00000126, 0.00000254, -0.00000712, 0.00000978, -0.00000690, 0.00000868, 0.00000813, -0.00000711, -0.00001723, 0.00001436, 0.00001153, -0.00001618, 0.00001459, -0.00001735, 0.00001297, -0.00001375, -0.00000754, -0.00000146, 0.00000849, -0.00001010, -0.00001832, 0.00001775, -0.00001300, 0.00000918, -0.00001596, 0.00001196, 0.00000903, -0.00000891, 0.00000848, -0.00000310, 0.00000012, -0.00000141, -0.00000690, 0.00000905, 0.00000255, -0.00000082, 0.00000181, -0.00000583, -0.00000303, 0.00000150, 0.00000627, -0.00000607, 0.00000370, -0.00000516, 0.00000147, -0.00000382, 0.00000481, -0.00001022, 0.00000164, -0.00000557, -0.00000767, 0.00000436, 0.00000016, -0.00000808, -0.00000508, -0.00002042, -0.00002432, -0.00005093, -0.00018117, 0.00032566, 0.00032009, -0.00018652, -0.00005804, -0.00001394, -0.00000689, -0.00001007, -0.00000196, -0.00001374, -0.00001739, 0.00000815, 0.00000061, -0.00000081, 0.00000116, -0.00000177, -0.00001475, 0.00001053, -0.00000454, 0.00000298, 0.00000042, -0.00000583, -0.00000601, 0.00000497, 0.00000304, -0.00000307, -0.00000205, -0.00000432, -0.00000041, -0.00000098, 0.00000521, -0.00000130, -0.00000550, 0.00000277, -0.00000455, 0.00000247, -0.00000645, -0.00000187, 0.00000052, -0.00002579, 0.00000653, 0.00000010, 0.00003459, 0.00000031, },
		{ -0.00001420, 0.00002780, 0.00001748, -0.00000082, -0.00000267, -0.00002519, 0.00000796, -0.00000995, 0.00000548, -0.00000795, 0.00001401, -0.00000031, -0.00000955, 0.00000093, -0.00000598, 0.00000558, -0.00000850, 0.00000142, 0.00000155, 0.00000220, -0.00000019, 0.00000373, -0.00000940, 0.00000357, 0.00000064, -0.00000160, 0.00000207, 0.00000081, -0.00000150, -0.00000250, 0.00000821, -0.00000849, 0.00001553, 0.00000054, -0.00000922, -0.00001496, 0.00002662, -0.00000738, -0.00000138, 0.00000299, -0.00001083, 0.00001227, -0.00002154, 0.00000113, -0.00000364, 0.00001186, -0.00002453, -0.00000056, 0.00000884, -0.00000973, 0.00000614, -0.00001556, 0.00002180, -0.00000363, -0.00000204, 0.00000676, -0.00000462, 0.00000219, -0.00000541, -0.00000218, 0.00001011, -0.00000129, 0.00000162, -0.00000116, -0.00000535, -0.00000235, 0.00000490, 0.00000188, -0.00000375, 0.00000206, -0.00000474, 0.00000159, -0.00000312, 0.00000357, -0.00001186, 0.00000582, -0.00001304, 0.00000121, -0.00000008, 0.00000262, -0.00001433, 0.00000136, -0.00003861, -0.00000231, -0.00013004, -0.00005383, 0.00042745, 0.00012341, -0.00018646, -0.00001736, -0.00001915, -0.00000215, -0.00001261, 0.00000144, -0.00002315, -0.00000519, 0.00000595, 0.00000042, -0.00000179, 0.00000383, -0.00000958, -0.00000683, 0.00000940, -0.00000668, 0.00000637, -0.00000414, -0.00000515, -0.00000444, 0.00000781, -0.00000105, -0.00000148, -0.00000426, -0.00000257, -0.00000064, 0.00000036, 0.00000560, -0.00000549, -0.00000149, 0.00000098, -0.00000440, 0.00000252, -0.00000914, 0.00000409, -0.00001094, -0.00001931, 0.00000811, 0.00000674, 0.00003679, },
		{ 0.00002573, -0.00001340, 0.00004203, 0.00000001, 0.00000813, -0.00002078, -0.00001057, 0.00000207, -0.00000549, 0.00000245, -0.00000441, 0.00001583, -0.00000875, -0.00000348, -0.00000349, -0.00000132, 0.00000257, -0.00000851, 0.00000402, 0.00000071, 0.00000178, 0.00000089, 0.00000117, -0.00000962, 0.00000810, -0.00000455, 0.00000275, -0.00000110, 0.00000378, -0.00000688, 0.00000593, -0.00000002, -0.00000167, 0.00001466, -0.00000357, -0.00001408, -0.00000330, 0.00002574, -0.00001779, 0.00001126, -0.00001206, 0.00000459, -0.00000131, -0.00001676, 0.00000195, 0.00000030, 0.00000546, -0.00002876, 0.00001406, -0.00000337, -0.00000207, -0.00000178, -0.00000695, 0.00002291, -0.00001327, 0.00000757, 0.00000006, -0.00000128, -0.00000010, -0.00000596, 0.00000225, 0.00000904, -0.00000402, 0.00000426, -0.00000542, -0.00000278, -0.00000229, 0.00000800, -0.00000370, 0.00000119, -0.00000318, 0.00000015, -0.00000288, 0.00000236, -0.00000407, -0.00000506, 0.00000010, -0.00001020, 0.00000185, 0.00000243, -0.00000420, -0.00000828, -0.00000996, -0.00003312, -0.00001685, -0.00017248, 0.00012897, 0.00039933, -0.00003304, -0.00014461, 0.00000191, -0.00002188, 0.00000016, -0.00001234, -0.00000014, -0.00002730, 0.00000699, -0.00000026, 0.00000298, -0.00000349, 0.00000531, -0.00001671, 0.00000465, 0.00000155, -0.00000218, 0.00000420, -0.00000549, -0.00000564, -0.00000039, 0.00000701, -0.00000305, -0.00000097, -0.00000590, -0.00000013, -0.00000196, 0.00000333, 0.00000287, -0.00000691, 0.00000221, -0.00000249, -0.00000218, -0.00000003, -0.00000792, 0.00000538, -0.00002101, -0.00000896, 0.00000605, 0.00001959, },
		{ 0.00003328, 0.00000569, 0.00000466, 0.00003806, -0.00000574, 0.00000670, -0.00003053, 0.00000340, -0.00000667, 0.00000232, -0.00000445, 0.00000482, 0.00000936, -0.00001012, -0.00000148, -0.00000546, 0.00000277, -0.00000196, -0.00000599, 0.00000510, -0.00000006, 0.00000148, 0.00000223, -0.00000340, -0.00000532, 0.00000777, -0.00000639, 0.00000498, -0.00000242, 0.00000401, -0.00000863, 0.00001195, -0.00000877, 0.00000927, 0.00000758, -0.00000462, -0.00001903, 0.00001419, 0.00001099, -0.00001278, 0.00001034, -0.00001648, 0.00001443, -0.00001617, -0.00000595, -0.00000194, 0.00000728, -0.00000844, -0.00002049, 0.00001895, -0.00001223, 0.00000517, -0.00001048, 0.00000716, 0.00001362, -0.00001381, 0.00001286, -0.00000606, 0.00000324, -0.00000472, -0.00000311, 0.00000512, 0.00000630, -0.00000446, 0.00000502, -0.00000861, 0.00000009, -0.00000158, 0.00000863, -0.00000698, 0.00000442, -0.00000718, 0.00000457, -0.00000707, 0.00000728, -0.00001195, 0.00000377, -0.00000972, -0.00000127, -0.00000293, 0.00000846, -0.00001543, 0.00000396, -0.00003279, -0.00000802, -0.00007642, -0.00012678, 0.00027945, 0.00028642, -0.00012619, -0.00008854, 0.00000330, -0.00001946, -0.00000103, -0.00000886, -0.00000777, -0.00002195, 0.00001212, -0.00000418, 0.00000426, -0.00000306, 0.00000219, -0.00001744, 0.00001229, -0.00000650, 0.00000463, -0.00000140, -0.00000422, -0.00000658, 0.00000472, 0.00000327, -0.00000250, -0.00000251, -0.00000545, 0.00000126, -0.00000288, 0.00000663, -0.00000256, -0.00000395, 0.00000238, -0.00000437, -0.00000058, -0.00000262, -0.00000494, 0.00000198, -0.00002673, 0.00000168, 0.00000410, },
		{ 0.00000725, 0.00003903, -0.00001115, 0.00002956, 0.00002001, 0.00000061, -0.00000822, -0.00002409, 0.00000598, -0.00000818, 0.00000448, -0.00000660, 0.00001134, 0.00000133, -0.00000813, -0.00000213, -0.00000484, 0.00000505, -0.00000674, -0.00000118, 0.00000378, 0.00000021, 0.00000111, 0.00000296, -0.00000810, 0.00000194, 0.00000247, -0.00000333, 0.00000313, -0.00000005, -0.00000045, -0.00000345, 0.00000986, -0.00001037, 0.00001623, 0.00000052, -0.00000723, -0.00001707, 0.00002663, -0.00000704, 0.00000114, -0.00000263, -0.00000643, 0.00000994, -0.00002118, 0.00000130, -0.00000362, 0.00001018, -0.00002179, -0.00000468, 0.00001341, -0.00001389, 0.00000774, -0.00001473, 0.00002051, -0.00000217, -0.00000429, 0.00000932, -0.00000605, 0.00000355, -0.00000682, -0.00000013, 0.00000751, 0.00000144, -0.00000131, 0.00000161, -0.00000769, 0.00000044, 0.00000169, 0.00000488, -0.00000514, 0.00000223, -0.00000528, 0.00000310, -0.00000514, 0.00000496, -0.00001196, 0.00000497, -0.00001278, 0.00000285, -0.00000280, 0.00000673, -0.00001712, 0.00000413, -0.00004285, 0.00000330, -0.00014067, -0.00000172, 0.00034255, 0.00014924, -0.00016272, -0.00003445, -0.00000869, -0.00001030, -0.00000662, -0.00000296, -0.00001926, -0.00000769, 0.00000732, -0.00000162, 0.00000117, 0.00000103, -0.00000644, -0.00000912, 0.00001063, -0.00000801, 0.00000741, -0.00000552, -0.00000338, -0.00000552, 0.00000788, -0.00000072, -0.00000109, -0.00000521, -0.00000274, 0.00000034, -0.00000138, 0.00000727, -0.00000721, 0.00000104, -0.00000060, -0.00000400, -0.00000037, -0.00000421, -0.00000184, -0.00000539, -0.00002586, 0.00000868, },
		{ 0.00000997, 0.00001796, 0.00003062, -0.00001213, 0.00004497, 0.00000220, 0.00000732, -0.00002590, -0.00000804, -0.00000125, -0.00000183, -0.00000056, -0.00000128, 0.00001068, -0.00000307, -0.00000689, -0.00000278, -0.00000282, 0.00000494, -0.00001000, 0.00000429, 0.00000065, 0.00000161, 0.00000073, 0.00000194, -0.00001004, 0.00000814, -0.00000423, 0.00000217, -0.00000095, 0.00000391, -0.00000685, 0.00000616, 0.00000030, -0.00000262, 0.00001472, -0.00000229, -0.00001373, -0.00000440, 0.00002476, -0.00001504, 0.00001002, -0.00001491, 0.00000855, -0.00000482, -0.00001485, 0.00000129, 0.00000006, 0.00000471, -0.00002695, 0.00001055, 0.00000154, -0.00000827, 0.00000357, -0.00001021, 0.00002520, -0.00001530, 0.00000876, -0.00000024, -0.00000023, -0.00000141, -0.00000434, 0.00000106, 0.00000951, -0.00000416, 0.00000397, -0.00000498, -0.00000280, -0.00000185, 0.00000668, -0.00000173, 0.00000022, -0.00000387, 0.00000137, -0.00000358, 0.00000235, -0.00000427, -0.00000358, -0.00000327, -0.00000569, -0.00000205, 0.00000592, -0.00000639, -0.00000434, -0.00001594, -0.00002544, -0.00002884, -0.00015083, 0.00014189, 0.00032068, 0.00002574, -0.00015422, 0.00000351, -0.00002330, 0.00000058, -0.00001266, 0.00000044, -0.00002722, 0.00000751, -0.00000212, 0.00000462, -0.00000395, 0.00000536, -0.00001568, 0.00000371, 0.00000159, -0.00000231, 0.00000399, -0.00000561, -0.00000458, -0.00000148, 0.00000731, -0.00000268, -0.00000084, -0.00000711, 0.00000078, -0.00000242, 0.00000301, 0.00000323, -0.00000722, 0.00000367, -0.00000350, -0.00000326, -0.00000016, -0.00000570, 0.00000098, -0.00001567, -0.00001763, },
		{ -0.00000485, 0.00000801, 0.00003217, 0.00001118, 0.00000543, 0.00004136, -0.00000375, 0.00000247, -0.00003198, 0.00000363, -0.00000771, 0.00000423, -0.00000599, 0.00000612, 0.00000583, -0.00000446, -0.00000661, -0.00000294, 0.00000024, 0.00000163, -0.00000934, 0.00000733, -0.00000186, 0.00000279, 0.00000098, -0.00000155, -0.00000701, 0.00000912, -0.00000735, 0.00000545, -0.00000302, 0.00000487, -0.00000927, 0.00001286, -0.00000935, 0.00000891, 0.00000752, -0.00000280, -0.00001974, 0.00001356, 0.00000997, -0.00000876, 0.00000541, -0.00001470, 0.00001489, -0.00001781, -0.00000469, -0.00000219, 0.00000587, -0.00000704, -0.00002137, 0.00001845, -0.00001027, 0.00000099, -0.00000533, 0.00000295, 0.00001697, -0.00001707, 0.00001559, -0.00000765, 0.00000519, -0.00000697, -0.00000013, 0.00000202, 0.00000887, -0.00000675, 0.00000684, -0.00001007, 0.00000203, -0.00000342, 0.00000952, -0.00000642, 0.00000375, -0.00000779, 0.00000615, -0.00000860, 0.00000796, -0.00001217, 0.00000449, -0.00001215, 0.00000329, -0.00000806, 0.00001402, -0.00001966, 0.00000941, -0.00004072, 0.00000280, -0.00009464, -0.00008063, 0.00024038, 0.00025027, -0.00007092, -0.00011127, 0.00001453, -0.00002762, 0.00000454, -0.00001305, -0.00000407, -0.00002404, 0.00001376, -0.00000706, 0.00000758, -0.00000553, 0.00000424, -0.00001801, 0.00001218, -0.00000696, 0.00000494, -0.00000218, -0.00000341, -0.00000629, 0.00000376, 0.00000389, -0.00000235, -0.00000261, -0.00000653, 0.00000270, -0.00000446, 0.00000769, -0.00000372, -0.00000231, 0.00000189, -0.00000411, -0.00000344, 0.00000068, -0.00000737, 0.00000232, -0.00002575, },
		{ -0.00003030, 0.00000646, 0.00000913, 0.00004023, -0.00000655, 0.00003063, 0.00002364, 0.00000083, -0.00001387, -0.00002227, 0.00000326, -0.00000580, 0.00000306, -0.00000495, 0.00000873, 0.00000238, -0.00000631, -0.00000525, -0.00000344, 0.00000402, -0.00000454, -0.00000389, 0.00000582, -0.00000156, 0.00000241, 0.00000191, -0.00000645, 0.00000019, 0.00000404, -0.00000464, 0.00000390, -0.00000074, 0.00000049, -0.00000413, 0.00001073, -0.00001110, 0.00001580, 0.00000124, -0.00000610, -0.00001757, 0.00002512, -0.00000600, 0.00000325, -0.00000791, -0.00000203, 0.00000716, -0.00002017, 0.00000108, -0.00000324, 0.00000796, -0.00001867, -0.00000812, 0.00001623, -0.00001596, 0.00000777, -0.00001273, 0.00001815, -0.00000024, -0.00000635, 0.00001132, -0.00000692, 0.00000439, -0.00000774, 0.00000166, 0.00000502, 0.00000383, -0.00000368, 0.00000369, -0.00000932, 0.00000263, -0.00000095, 0.00000708, -0.00000567, 0.00000162, -0.00000508, 0.00000380, -0.00000612, 0.00000511, -0.00001106, 0.00000341, -0.00001171, 0.00000367, -0.00000460, 0.00000949, -0.00001824, 0.00000490, -0.00004433, 0.00000501, -0.00014468, 0.00004020, 0.00027061, 0.00016318, -0.00013391, -0.00005137, 0.00000062, -0.00001741, -0.00000150, -0.00000683, -0.00001553, -0.00000975, 0.00000805, -0.00000318, 0.00000375, -0.00000142, -0.00000385, -0.00001041, 0.00001076, -0.00000834, 0.00000743, -0.00000603, -0.00000228, -0.00000586, 0.00000722, 0.00000008, -0.00000119, -0.00000557, -0.00000316, 0.00000130, -0.00000287, 0.00000846, -0.00000853, 0.00000334, -0.00000207, -0.00000353, -0.00000318, 0.00000033, -0.00000728, -0.00000073, },
		{ -0.00001013, -0.00002530, 0.00001139, 0.00001798, 0.00003431, -0.00000895, 0.00004670, 0.00000599, 0.00000410, -0.00002938, -0.00000633, -0.00000417, 0.00000160, -0.00000341, 0.00000181, 0.00000574, 0.00000182, -0.00000945, -0.00000253, -0.00000379, 0.00000647, -0.00001049, 0.00000372, 0.00000107, 0.00000118, 0.00000091, 0.00000209, -0.00000960, 0.00000731, -0.00000333, 0.00000129, -0.00000048, 0.00000360, -0.00000621, 0.00000567, 0.00000096, -0.00000338, 0.00001430, -0.00000085, -0.00001350, -0.00000451, 0.00002231, -0.00001124, 0.00000793, -0.00001678, 0.00001169, -0.00000795, -0.00001284, 0.00000051, 0.00000002, 0.00000333, -0.00002411, 0.00000658, 0.00000569, -0.00001270, 0.00000706, -0.00001168, 0.00002543, -0.00001545, 0.00000860, 0.00000033, 0.00000014, -0.00000212, -0.00000321, 0.00000051, 0.00000909, -0.00000347, 0.00000298, -0.00000401, -0.00000325, -0.00000089, 0.00000475, 0.00000069, -0.00000108, -0.00000410, 0.00000202, -0.00000366, 0.00000187, -0.00000428, -0.00000244, -0.00000585, -0.00000211, -0.00000478, 0.00000798, -0.00000718, -0.00000190, -0.00002016, -0.00001961, -0.00003893, -0.00012939, 0.00015079, 0.00025191, 0.00007077, -0.00015322, -0.00000057, -0.00002131, -0.00000146, -0.00001110, -0.00000069, -0.00002519, 0.00000636, -0.00000270, 0.00000503, -0.00000318, 0.00000414, -0.00001348, 0.00000208, 0.00000198, -0.00000267, 0.00000378, -0.00000564, -0.00000352, -0.00000241, 0.00000729, -0.00000213, -0.00000097, -0.00000777, 0.00000123, -0.00000259, 0.00000256, 0.00000337, -0.00000722, 0.00000490, -0.00000439, -0.00000412, -0.00000063, -0.00000331, -0.00000366, },
		{ 0.00000152, -0.00002311, -0.00001234, 0.00001096, 0.00003177, 0.00001651, 0.00000720, 0.00004433, -0.00000129, -0.00000353, -0.00003160, 0.00000253, -0.00000783, 0.00000549, -0.00000710, 0.00000727, 0.00000233, 0.00000065, -0.00001093, -0.00000086, -0.00000193, 0.00000459, -0.00001175, 0.00000853, -0.00000289, 0.00000362, 0.00000008, -0.00000022, -0.00000790, 0.00000943, -0.00000737, 0.00000525, -0.00000301, 0.00000505, -0.00000902, 0.00001251, -0.00000869, 0.00000771, 0.00000788, -0.00000161, -0.00001947, 0.00001264, 0.00000845, -0.00000432, 0.00000015, -0.00001227, 0.00001445, -0.00001871, -0.00000369, -0.00000229, 0.00000439, -0.00000608, -0.00002089, 0.00001636, -0.00000742, -0.00000293, -0.00000095, -0.00000034, 0.00001878, -0.00001843, 0.00001656, -0.00000781, 0.00000592, -0.00000810, 0.00000189, -0.00000009, 0.00001018, -0.00000769, 0.00000732, -0.00001030, 0.00000287, -0.00000409, 0.00000912, -0.00000466, 0.00000204, -0.00000727, 0.00000645, -0.00000867, 0.00000721, -0.00001131, 0.00000417, -0.00001312, 0.00000612, -0.00001111, 0.00001697, -0.00002099, 0.00001152, -0.00004451, 0.00000845, -0.00010583, -0.00004261, 0.00020757, 0.00021383, -0.00002278, -0.00012535, 0.00001963, -0.00003145, 0.00000684, -0.00001474, -0.00000242, -0.00002393, 0.00001329, -0.00000817, 0.00000925, -0.00000640, 0.00000457, -0.00001675, 0.00001054, -0.00000620, 0.00000418, -0.00000215, -0.00000317, -0.00000543, 0.00000239, 0.00000464, -0.00000237, -0.00000261, -0.00000728, 0.00000363, -0.00000548, 0.00000818, -0.00000456, -0.00000081, 0.00000150, -0.00000396, -0.00000579, 0.00000315, -0.00000891, },
		{ -0.00001167, 0.00000252, -0.00003227, 0.00000159, 0.00001178, 0.00004092, -0.00000080, 0.00003140, 0.00002817, -0.00000032, -0.00001934, -0.00001995, -0.00000001, -0.00000302, 0.00000138, -0.00000320, 0.00000635, 0.00000284, -0.00000427, -0.00000816, -0.00000199, 0.00000272, -0.00000219, -0.00000640, 0.00000738, -0.00000286, 0.00000351, 0.00000082, -0.00000475, -0.00000140, 0.00000512, -0.00000536, 0.00000426, -0.00000112, 0.00000116, -0.00000440, 0.00001071, -0.00001061, 0.00001424, 0.00000266, -0.00000582, -0.00001658, 0.00002236, -0.00000448, 0.00000496, -0.00001264, 0.00000210, 0.00000415, -0.00001867, 0.00000060, -0.00000263, 0.00000542, -0.00001547, -0.00001056, 0.00001706, -0.00001590, 0.00000642, -0.00000985, 0.00001501, 0.00000185, -0.00000792, 0.00001254, -0.00000707, 0.00000460, -0.00000803, 0.00000300, 0.00000291, 0.00000560, -0.00000526, 0.00000488, -0.00001007, 0.00000405, -0.00000280, 0.00000833, -0.00000528, 0.00000029, -0.00000419, 0.00000371, -0.00000610, 0.00000415, -0.00000941, 0.00000143, -0.00001004, 0.00000377, -0.00000549, 0.00001088, -0.00001773, 0.00000373, -0.00004315, 0.00000299, -0.00014215, 0.00007160, 0.00021164, 0.00016668, -0.00010226, -0.00006642, 0.00000777, -0.00002278, 0.00000225, -0.00000979, -0.00001227, -0.00001116, 0.00000801, -0.00000411, 0.00000573, -0.00000328, -0.00000201, -0.00001063, 0.00000980, -0.00000772, 0.00000649, -0.00000569, -0.00000181, -0.00000553, 0.00000596, 0.00000121, -0.00000165, -0.00000551, -0.00000365, 0.00000201, -0.00000391, 0.00000897, -0.00000926, 0.00000519, -0.00000321, -0.00000325, -0.00000556, 0.00000398, },
		{ -0.00000108, -0.00000799, -0.00000513, -0.00003115, 0.00000984, 0.00001955, 0.00003707, -0.00000415, 0.00004770, 0.00001093, -0.00000139, -0.00003114, -0.00000557, -0.00000651, 0.00000457, -0.00000585, 0.00000450, 0.00000143, 0.00000554, -0.00001098, -0.00000280, -0.00000414, 0.00000710, -0.00001001, 0.00000242, 0.00000185, 0.00000059, 0.00000133, 0.00000174, -0.00000849, 0.00000584, -0.00000208, 0.00000028, 0.00000019, 0.00000299, -0.00000513, 0.00000467, 0.00000177, -0.00000380, 0.00001332, 0.00000081, -0.00001347, -0.00000362, 0.00001861, -0.00000678, 0.00000529, -0.00001778, 0.00001393, -0.00001062, -0.00001080, -0.00000034, 0.00000013, 0.00000145, -0.00002049, 0.00000254, 0.00000869, -0.00001506, 0.00000859, -0.00001139, 0.00002368, -0.00001389, 0.00000732, 0.00000154, 0.00000001, -0.00000234, -0.00000249, 0.00000046, 0.00000801, -0.00000224, 0.00000164, -0.00000282, -0.00000385, 0.00000027, 0.00000262, 0.00000313, -0.00000236, -0.00000416, 0.00000235, -0.00000340, 0.00000122, -0.00000436, -0.00000151, -0.00000768, 0.00000057, -0.00000644, 0.00000879, -0.00000685, -0.00000065, -0.00002293, -0.00001532, -0.00004734, -0.00010793, 0.00015491, 0.00019410, 0.00010210, -0.00014293, -0.00000884, -0.00001707, -0.00000501, -0.00000843, -0.00000286, -0.00002185, 0.00000405, -0.00000238, 0.00000457, -0.00000162, 0.00000212, -0.00001061, 0.00000019, 0.00000242, -0.00000302, 0.00000338, -0.00000537, -0.00000267, -0.00000300, 0.00000683, -0.00000132, -0.00000144, -0.00000781, 0.00000119, -0.00000248, 0.00000205, 0.00000327, -0.00000689, 0.00000586, -0.00000513, -0.00000481, -0.00000128, },
		{ 0.00000461, -0.00000953, -0.00000026, -0.00001929, -0.00001992, 0.00001210, 0.00003247, 0.00002154, 0.00001001, 0.00004717, 0.00000117, -0.00001089, -0.00002960, 0.00000019, -0.00000706, 0.00000609, -0.00000769, 0.00000809, -0.00000078, 0.00000475, -0.00001408, 0.00000051, -0.00000346, 0.00000664, -0.00001299, 0.00000858, -0.00000305, 0.00000391, -0.00000038, 0.00000052, -0.00000794, 0.00000877, -0.00000655, 0.00000445, -0.00000244, 0.00000460, -0.00000795, 0.00001105, -0.00000700, 0.00000590, 0.00000848, -0.00000093, -0.00001844, 0.00001163, 0.00000643, 0.00000028, -0.00000504, -0.00000951, 0.00001330, -0.00001896, -0.00000292, -0.00000232, 0.00000295, -0.00000562, -0.00001911, 0.00001289, -0.00000406, -0.00000618, 0.00000237, -0.00000247, 0.00001892, -0.00001790, 0.00001584, -0.00000667, 0.00000553, -0.00000814, 0.00000295, -0.00000119, 0.00001028, -0.00000740, 0.00000668, -0.00000952, 0.00000278, -0.00000382, 0.00000774, -0.00000211, -0.00000030, -0.00000603, 0.00000584, -0.00000769, 0.00000549, -0.00000984, 0.00000322, -0.00001294, 0.00000749, -0.00001233, 0.00001764, -0.00001985, 0.00001083, -0.00004473, 0.00000960, -0.00011063, -0.00001221, 0.00017998, 0.00017900, 0.00001679, -0.00013063, 0.00001904, -0.00003148, 0.00000637, -0.00001441, -0.00000237, -0.00002207, 0.00001114, -0.00000783, 0.00000953, -0.00000596, 0.00000356, -0.00001411, 0.00000782, -0.00000462, 0.00000269, -0.00000162, -0.00000320, -0.00000429, 0.00000093, 0.00000524, -0.00000234, -0.00000271, -0.00000749, 0.00000388, -0.00000581, 0.00000800, -0.00000499, 0.00000042, 0.00000133, -0.00000405, -0.00000748, },
		{ -0.00000723, 0.00000638, -0.00001460, 0.00000403, -0.00003175, -0.00000540, 0.00001435, 0.00004174, 0.00000562, 0.00003236, 0.00003334, -0.00000296, -0.00002434, -0.00001738, -0.00000358, -0.00000005, -0.00000035, -0.00000153, 0.00000432, 0.00000278, -0.00000224, -0.00001061, -0.00000072, 0.00000139, 0.00000000, -0.00000840, 0.00000820, -0.00000350, 0.00000420, -0.00000009, -0.00000324, -0.00000262, 0.00000558, -0.00000541, 0.00000415, -0.00000112, 0.00000145, -0.00000419, 0.00000979, -0.00000898, 0.00001172, 0.00000463, -0.00000623, -0.00001436, 0.00001873, -0.00000280, 0.00000634, -0.00001663, 0.00000571, 0.00000112, -0.00001684, -0.00000004, -0.00000195, 0.00000275, -0.00001244, -0.00001183, 0.00001590, -0.00001388, 0.00000402, -0.00000642, 0.00001145, 0.00000379, -0.00000876, 0.00001285, -0.00000644, 0.00000415, -0.00000765, 0.00000374, 0.00000139, 0.00000658, -0.00000593, 0.00000515, -0.00000992, 0.00000462, -0.00000376, 0.00000859, -0.00000405, -0.00000159, -0.00000278, 0.00000298, -0.00000524, 0.00000234, -0.00000736, -0.00000064, -0.00000805, 0.00000336, -0.00000560, 0.00001103, -0.00001580, 0.00000094, -0.00003972, -0.00000220, -0.00013372, 0.00009291, 0.00016487, 0.00016171, -0.00007007, -0.00007817, 0.00001200, -0.00002597, 0.00000435, -0.00001163, -0.00000964, -0.00001186, 0.00000718, -0.00000435, 0.00000700, -0.00000444, -0.00000095, -0.00000986, 0.00000794, -0.00000634, 0.00000475, -0.00000466, -0.00000183, -0.00000471, 0.00000432, 0.00000244, -0.00000228, -0.00000519, -0.00000399, 0.00000230, -0.00000435, 0.00000869, -0.00000928, 0.00000644, -0.00000383, -0.00000333, },
		{ -0.00000541, -0.00000196, 0.00000066, -0.00001151, -0.00000128, -0.00003459, 0.00000522, 0.00002221, 0.00003941, 0.00000183, 0.00004857, 0.00001636, -0.00000876, -0.00003126, -0.00000578, -0.00000816, 0.00000694, -0.00000769, 0.00000654, -0.00000194, 0.00000790, -0.00001144, -0.00000357, -0.00000393, 0.00000689, -0.00000870, 0.00000058, 0.00000280, 0.00000001, 0.00000185, 0.00000107, -0.00000695, 0.00000401, -0.00000072, -0.00000065, 0.00000090, 0.00000224, -0.00000383, 0.00000339, 0.00000250, -0.00000373, 0.00001177, 0.00000267, -0.00001363, -0.00000188, 0.00001403, -0.00000210, 0.00000248, -0.00001806, 0.00001527, -0.00001275, -0.00000883, -0.00000119, 0.00000031, -0.00000078, -0.00001637, -0.00000128, 0.00001030, -0.00001529, 0.00000826, -0.00000955, 0.00002026, -0.00001098, 0.00000527, 0.00000311, -0.00000041, -0.00000224, -0.00000204, 0.00000073, 0.00000656, -0.00000082, 0.00000027, -0.00000169, -0.00000437, 0.00000135, 0.00000064, 0.00000523, -0.00000330, -0.00000429, 0.00000256, -0.00000301, 0.00000064, -0.00000468, -0.00000071, -0.00000880, 0.00000238, -0.00000712, 0.00000857, -0.00000569, -0.00000030, -0.00002452, -0.00001231, -0.00005420, -0.00008648, 0.00015403, 0.00014747, 0.00012070, -0.00012529, -0.00001956, -0.00001186, -0.00000907, -0.00000547, -0.00000537, -0.00001782, 0.00000112, -0.00000158, 0.00000365, 0.00000030, -0.00000021, -0.00000753, -0.00000160, 0.00000266, -0.00000320, 0.00000266, -0.00000470, -0.00000214, -0.00000317, 0.00000591, -0.00000028, -0.00000221, -0.00000728, 0.00000074, -0.00000218, 0.00000153, 0.00000289, -0.00000623, 0.00000648, -0.00000560, },
		{ -0.00000441, -0.00000845, 0.00000502, -0.00000933, -0.00000267, -0.00001496, -0.00002667, 0.00001071, 0.00003443, 0.00002631, 0.00001383, 0.00005018, 0.00000305, -0.00001905, -0.00002634, -0.00000310, -0.00000558, 0.00000609, -0.00000776, 0.00000845, -0.00000323, 0.00000755, -0.00001586, 0.00000101, -0.00000421, 0.00000765, -0.00001299, 0.00000749, -0.00000238, 0.00000366, -0.00000040, 0.00000067, -0.00000724, 0.00000729, -0.00000509, 0.00000325, -0.00000144, 0.00000366, -0.00000628, 0.00000873, -0.00000458, 0.00000377, 0.00000908, -0.00000054, -0.00001688, 0.00001071, 0.00000398, 0.00000474, -0.00000977, -0.00000674, 0.00001162, -0.00001867, -0.00000230, -0.00000233, 0.00000159, -0.00000565, -0.00001621, 0.00000838, -0.00000055, -0.00000846, 0.00000446, -0.00000342, 0.00001744, -0.00001563, 0.00001370, -0.00000451, 0.00000425, -0.00000727, 0.00000314, -0.00000141, 0.00000936, -0.00000617, 0.00000524, -0.00000803, 0.00000203, -0.00000289, 0.00000576, 0.00000076, -0.00000276, -0.00000451, 0.00000473, -0.00000608, 0.00000331, -0.00000825, 0.00000204, -0.00001196, 0.00000770, -0.00001206, 0.00001647, -0.00001683, 0.00000800, -0.00004216, 0.00000715, -0.00010995, 0.00001137, 0.00015654, 0.00014729, 0.00004707, -0.00012762, 0.00001370, -0.00002856, 0.00000387, -0.00001269, -0.00000335, -0.00001903, 0.00000783, -0.00000645, 0.00000879, -0.00000461, 0.00000167, -0.00001065, 0.00000456, -0.00000266, 0.00000082, -0.00000088, -0.00000324, -0.00000317, -0.00000036, 0.00000549, -0.00000210, -0.00000305, -0.00000706, 0.00000340, -0.00000544, 0.00000717, -0.00000498, 0.00000132, 0.00000143, },
		{ -0.00000386, -0.00000385, -0.00000806, 0.00000733, -0.00001590, 0.00000378, -0.00002902, -0.00001370, 0.00001583, 0.00004336, 0.00001228, 0.00003399, 0.00003881, -0.00000716, -0.00002856, -0.00001486, -0.00000718, 0.00000284, -0.00000193, -0.00000010, 0.00000271, 0.00000229, -0.00000045, -0.00001239, 0.00000019, 0.00000026, 0.00000180, -0.00000967, 0.00000812, -0.00000334, 0.00000436, -0.00000066, -0.00000210, -0.00000332, 0.00000535, -0.00000482, 0.00000361, -0.00000074, 0.00000136, -0.00000353, 0.00000807, -0.00000643, 0.00000854, 0.00000686, -0.00000705, -0.00001132, 0.00001468, -0.00000126, 0.00000748, -0.00001976, 0.00000858, -0.00000173, -0.00001484, -0.00000071, -0.00000130, 0.00000011, -0.00000971, -0.00001192, 0.00001295, -0.00001029, 0.00000099, -0.00000283, 0.00000777, 0.00000532, -0.00000871, 0.00001220, -0.00000510, 0.00000311, -0.00000664, 0.00000385, 0.00000053, 0.00000671, -0.00000569, 0.00000457, -0.00000899, 0.00000440, -0.00000388, 0.00000800, -0.00000221, -0.00000377, -0.00000115, 0.00000188, -0.00000382, 0.00000006, -0.00000530, -0.00000244, -0.00000606, 0.00000268, -0.00000515, 0.00001021, -0.00001282, -0.00000300, -0.00003466, -0.00000973, -0.00012046, 0.00010518, 0.00012895, 0.00015048, -0.00003944, -0.00008556, 0.00001294, -0.00002684, 0.00000473, -0.00001232, -0.00000769, -0.00001187, 0.00000565, -0.00000393, 0.00000754, -0.00000491, -0.00000061, -0.00000831, 0.00000547, -0.00000448, 0.00000249, -0.00000318, -0.00000212, -0.00000363, 0.00000255, 0.00000355, -0.00000289, -0.00000482, -0.00000403, 0.00000206, -0.00000411, 0.00000760, -0.00000857, 0.00000701, },
		{ 0.00000671, -0.00000574, -0.00000603, -0.00000250, 0.00000164, -0.00001383, 0.00000097, -0.00003532, -0.00000220, 0.00002520, 0.00004200, 0.00000843, 0.00005002, 0.00002150, -0.00001750, -0.00002995, -0.00000687, -0.00000911, 0.00000859, -0.00000878, 0.00000775, -0.00000420, 0.00000882, -0.00001091, -0.00000471, -0.00000327, 0.00000602, -0.00000680, -0.00000153, 0.00000370, -0.00000039, 0.00000229, 0.00000030, -0.00000527, 0.00000212, 0.00000049, -0.00000133, 0.00000150, 0.00000153, -0.00000251, 0.00000205, 0.00000299, -0.00000310, 0.00000969, 0.00000464, -0.00001388, 0.00000045, 0.00000902, 0.00000229, -0.00000012, -0.00001779, 0.00001574, -0.00001429, -0.00000697, -0.00000198, 0.00000048, -0.00000320, -0.00001200, -0.00000464, 0.00001048, -0.00001357, 0.00000644, -0.00000656, 0.00001560, -0.00000717, 0.00000288, 0.00000470, -0.00000090, -0.00000197, -0.00000173, 0.00000111, 0.00000502, 0.00000050, -0.00000085, -0.00000083, -0.00000462, 0.00000212, -0.00000089, 0.00000669, -0.00000369, -0.00000463, 0.00000277, -0.00000265, 0.00000029, -0.00000529, -0.00000006, -0.00000920, 0.00000336, -0.00000694, 0.00000752, -0.00000396, -0.00000060, -0.00002518, -0.00001039, -0.00005953, -0.00006533, 0.00014847, 0.00011144, 0.00012827, -0.00010263, -0.00003089, -0.00000693, -0.00001274, -0.00000293, -0.00000762, -0.00001370, -0.00000194, -0.00000069, 0.00000264, 0.00000217, -0.00000243, -0.00000464, -0.00000303, 0.00000258, -0.00000313, 0.00000157, -0.00000360, -0.00000196, -0.00000294, 0.00000461, 0.00000090, -0.00000318, -0.00000630, 0.00000004, -0.00000183, 0.00000110, 0.00000219, -0.00000521, },
		{ 0.00000192, 0.00000175, -0.00000495, -0.00000879, 0.00000453, -0.00000855, -0.00000527, -0.00001080, -0.00003173, 0.00000630, 0.00003756, 0.00003103, 0.00001850, 0.00005369, 0.00000367, -0.00002732, -0.00002232, -0.00000700, -0.00000363, 0.00000561, -0.00000735, 0.00000828, -0.00000485, 0.00000892, -0.00001621, 0.00000065, -0.00000418, 0.00000763, -0.00001185, 0.00000541, -0.00000104, 0.00000301, -0.00000006, 0.00000035, -0.00000601, 0.00000527, -0.00000327, 0.00000189, -0.00000022, 0.00000245, -0.00000427, 0.00000593, -0.00000185, 0.00000171, 0.00000943, -0.00000023, -0.00001505, 0.00001000, 0.00000122, 0.00000874, -0.00001365, -0.00000426, 0.00000962, -0.00001793, -0.00000180, -0.00000236, 0.00000033, -0.00000609, -0.00001245, 0.00000322, 0.00000276, -0.00000961, 0.00000536, -0.00000330, 0.00001459, -0.00001197, 0.00001054, -0.00000171, 0.00000241, -0.00000577, 0.00000267, -0.00000094, 0.00000774, -0.00000437, 0.00000338, -0.00000620, 0.00000093, -0.00000163, 0.00000360, 0.00000349, -0.00000492, -0.00000312, 0.00000350, -0.00000428, 0.00000114, -0.00000695, 0.00000098, -0.00001051, 0.00000709, -0.00001070, 0.00001397, -0.00001256, 0.00000377, -0.00003767, 0.00000210, -0.00010488, 0.00002902, 0.00013628, 0.00011970, 0.00006809, -0.00011750, 0.00000493, -0.00002377, 0.00000021, -0.00001028, -0.00000476, -0.00001544, 0.00000396, -0.00000452, 0.00000745, -0.00000282, -0.00000056, -0.00000694, 0.00000127, -0.00000073, -0.00000108, -0.00000023, -0.00000305, -0.00000228, -0.00000129, 0.00000526, -0.00000158, -0.00000368, -0.00000601, 0.00000227, -0.00000450, 0.00000579, -0.00000461, },
		{ -0.00000722, 0.00000693, -0.00000333, -0.00000477, -0.00000809, 0.00000686, -0.00001567, 0.00000202, -0.00002469, -0.00002225, 0.00001517, 0.00004631, 0.00001883, 0.00003673, 0.00004421, -0.00001292, -0.00003172, -0.00001268, -0.00001050, 0.00000539, -0.00000318, 0.00000096, 0.00000155, 0.00000150, 0.00000092, -0.00001334, 0.00000061, -0.00000053, 0.00000303, -0.00001010, 0.00000712, -0.00000239, 0.00000398, -0.00000082, -0.00000142, -0.00000348, 0.00000450, -0.00000369, 0.00000276, -0.00000006, 0.00000094, -0.00000253, 0.00000579, -0.00000331, 0.00000509, 0.00000897, -0.00000792, -0.00000792, 0.00001071, -0.00000017, 0.00000844, -0.00002192, 0.00001054, -0.00000423, -0.00001279, -0.00000133, -0.00000076, -0.00000242, -0.00000734, -0.00001098, 0.00000860, -0.00000564, -0.00000221, 0.00000058, 0.00000425, 0.00000625, -0.00000774, 0.00001070, -0.00000320, 0.00000165, -0.00000515, 0.00000341, 0.00000032, 0.00000607, -0.00000470, 0.00000336, -0.00000748, 0.00000357, -0.00000332, 0.00000675, -0.00000003, -0.00000590, 0.00000040, 0.00000070, -0.00000218, -0.00000230, -0.00000363, -0.00000364, -0.00000435, 0.00000200, -0.00000440, 0.00000876, -0.00000926, -0.00000750, -0.00002872, -0.00001853, -0.00010375, 0.00010989, 0.00010211, 0.00013527, -0.00001214, -0.00008798, 0.00001063, -0.00002554, 0.00000358, -0.00001198, -0.00000633, -0.00001133, 0.00000361, -0.00000298, 0.00000744, -0.00000481, -0.00000078, -0.00000628, 0.00000275, -0.00000246, 0.00000001, -0.00000153, -0.00000246, -0.00000252, 0.00000089, 0.00000433, -0.00000329, -0.00000454, -0.00000365, 0.00000125, -0.00000326, 0.00000581, },
		{ 0.00000117, -0.00000385, 0.00000653, -0.00000550, -0.00000671, -0.00000282, 0.00000170, -0.00001473, 0.00000141, -0.00003340, -0.00001175, 0.00002751, 0.00004560, 0.00001510, 0.00005276, 0.00002548, -0.00002696, -0.00002753, -0.00000870, -0.00000941, 0.00000949, -0.00000907, 0.00000807, -0.00000530, 0.00000842, -0.00000958, -0.00000603, -0.00000233, 0.00000470, -0.00000458, -0.00000363, 0.00000432, -0.00000044, 0.00000249, -0.00000035, -0.00000372, 0.00000044, 0.00000134, -0.00000164, 0.00000190, 0.00000098, -0.00000136, 0.00000084, 0.00000311, -0.00000193, 0.00000721, 0.00000656, -0.00001403, 0.00000302, 0.00000410, 0.00000591, -0.00000215, -0.00001718, 0.00001540, -0.00001524, -0.00000528, -0.00000268, 0.00000058, -0.00000567, -0.00000760, -0.00000744, 0.00000937, -0.00001030, 0.00000361, -0.00000292, 0.00001022, -0.00000298, 0.00000057, 0.00000600, -0.00000126, -0.00000168, -0.00000145, 0.00000145, 0.00000364, 0.00000148, -0.00000153, -0.00000038, -0.00000451, 0.00000245, -0.00000182, 0.00000738, -0.00000343, -0.00000522, 0.00000300, -0.00000235, 0.00000019, -0.00000616, 0.00000039, -0.00000888, 0.00000356, -0.00000602, 0.00000586, -0.00000189, -0.00000133, -0.00002507, -0.00000945, -0.00006325, -0.00004497, 0.00013900, 0.00008483, 0.00012697, -0.00007739, -0.00004111, -0.00000331, -0.00001529, -0.00000134, -0.00000914, -0.00000997, -0.00000471, -0.00000000, 0.00000183, 0.00000364, -0.00000421, -0.00000222, -0.00000392, 0.00000214, -0.00000283, 0.00000019, -0.00000217, -0.00000204, -0.00000240, 0.00000306, 0.00000206, -0.00000418, -0.00000508, -0.00000072, -0.00000158, 0.00000085, },
		{ 0.00000407, -0.00000401, 0.00000230, 0.00000216, -0.00000553, -0.00000867, 0.00000336, -0.00000753, -0.00000758, -0.00000748, -0.00003442, -0.00000131, 0.00004144, 0.00003610, 0.00002382, 0.00005806, 0.00000236, -0.00003494, -0.00001813, -0.00001104, -0.00000153, 0.00000484, -0.00000655, 0.00000761, -0.00000560, 0.00000887, -0.00001526, -0.00000043, -0.00000348, 0.00000673, -0.00000980, 0.00000262, 0.00000072, 0.00000216, 0.00000046, -0.00000023, -0.00000452, 0.00000306, -0.00000142, 0.00000062, 0.00000100, 0.00000121, -0.00000226, 0.00000304, 0.00000079, 0.00000003, 0.00000930, 0.00000019, -0.00001313, 0.00000955, -0.00000164, 0.00001196, -0.00001636, -0.00000235, 0.00000747, -0.00001686, -0.00000138, -0.00000242, -0.00000084, -0.00000680, -0.00000816, -0.00000220, 0.00000561, -0.00000963, 0.00000526, -0.00000241, 0.00001075, -0.00000742, 0.00000687, 0.00000126, 0.00000042, -0.00000396, 0.00000179, -0.00000006, 0.00000575, -0.00000238, 0.00000151, -0.00000437, -0.00000023, -0.00000037, 0.00000163, 0.00000567, -0.00000637, -0.00000220, 0.00000247, -0.00000263, -0.00000067, -0.00000624, 0.00000026, -0.00000886, 0.00000598, -0.00000864, 0.00001068, -0.00000771, -0.00000108, -0.00003213, -0.00000450, -0.00009653, 0.00004168, 0.00011845, 0.00009668, 0.00008051, -0.00010187, -0.00000571, -0.00001827, -0.00000370, -0.00000790, -0.00000601, -0.00001187, 0.00000008, -0.00000249, 0.00000595, -0.00000104, -0.00000263, -0.00000348, -0.00000158, 0.00000084, -0.00000274, 0.00000015, -0.00000252, -0.00000173, -0.00000178, 0.00000454, -0.00000079, -0.00000451, -0.00000449, 0.00000072, -0.00000319, },
		{ -0.00000196, 0.00000351, -0.00000541, 0.00000634, -0.00000237, -0.00000594, -0.00000748, 0.00000519, -0.00001417, -0.00000079, -0.00001954, -0.00002991, 0.00001144, 0.00005093, 0.00002509, 0.00004095, 0.00004917, -0.00002023, -0.00003356, -0.00001116, -0.00001322, 0.00000732, -0.00000391, 0.00000154, 0.00000085, 0.00000053, 0.00000176, -0.00001344, 0.00000051, -0.00000091, 0.00000362, -0.00000968, 0.00000528, -0.00000078, 0.00000314, -0.00000058, -0.00000120, -0.00000318, 0.00000320, -0.00000225, 0.00000178, 0.00000078, 0.00000034, -0.00000138, 0.00000326, -0.00000002, 0.00000183, 0.00001057, -0.00000846, -0.00000464, 0.00000723, 0.00000026, 0.00000926, -0.00002303, 0.00001146, -0.00000626, -0.00001080, -0.00000185, -0.00000038, -0.00000476, -0.00000532, -0.00000927, 0.00000336, -0.00000051, -0.00000511, 0.00000353, 0.00000107, 0.00000653, -0.00000593, 0.00000855, -0.00000098, 0.00000001, -0.00000340, 0.00000255, 0.00000063, 0.00000485, -0.00000322, 0.00000181, -0.00000569, 0.00000235, -0.00000231, 0.00000514, 0.00000213, -0.00000764, 0.00000153, -0.00000025, -0.00000066, -0.00000435, -0.00000267, -0.00000401, -0.00000315, 0.00000154, -0.00000362, 0.00000705, -0.00000562, -0.00001192, -0.00002271, -0.00002750, -0.00008512, 0.00010873, 0.00008248, 0.00011815, 0.00001056, -0.00008538, 0.00000547, -0.00002250, 0.00000127, -0.00001090, -0.00000537, -0.00001045, 0.00000131, -0.00000171, 0.00000686, -0.00000431, -0.00000122, -0.00000412, 0.00000016, -0.00000064, -0.00000236, 0.00000001, -0.00000261, -0.00000159, -0.00000047, 0.00000463, -0.00000337, -0.00000444, -0.00000286, -0.00000002, },
		{ -0.00000153, 0.00000080, -0.00000013, -0.00000221, 0.00000595, -0.00000487, -0.00000747, -0.00000288, 0.00000081, -0.00001420, 0.00000011, -0.00002922, -0.00002248, 0.00002796, 0.00005092, 0.00002141, 0.00005739, 0.00002750, -0.00003646, -0.00002438, -0.00001105, -0.00000914, 0.00000966, -0.00000859, 0.00000752, -0.00000535, 0.00000690, -0.00000772, -0.00000731, -0.00000132, 0.00000318, -0.00000234, -0.00000546, 0.00000450, -0.00000004, 0.00000235, -0.00000073, -0.00000253, -0.00000082, 0.00000172, -0.00000152, 0.00000205, 0.00000067, -0.00000052, -0.00000011, 0.00000284, -0.00000032, 0.00000454, 0.00000820, -0.00001385, 0.00000543, -0.00000023, 0.00000835, -0.00000332, -0.00001638, 0.00001433, -0.00001562, -0.00000379, -0.00000326, 0.00000056, -0.00000804, -0.00000334, -0.00000965, 0.00000727, -0.00000604, 0.00000039, 0.00000081, 0.00000465, 0.00000110, -0.00000128, 0.00000679, -0.00000133, -0.00000145, -0.00000112, 0.00000163, 0.00000257, 0.00000197, -0.00000169, -0.00000034, -0.00000406, 0.00000232, -0.00000209, 0.00000727, -0.00000256, -0.00000600, 0.00000321, -0.00000209, 0.00000030, -0.00000716, 0.00000052, -0.00000785, 0.00000306, -0.00000450, 0.00000375, 0.00000034, -0.00000236, -0.00002431, -0.00000942, -0.00006519, -0.00002606, 0.00012671, 0.00006599, 0.00011924, -0.00005193, -0.00004882, -0.00000173, -0.00001628, -0.00000100, -0.00000965, -0.00000694, -0.00000690, 0.00000027, 0.00000143, 0.00000449, -0.00000532, -0.00000044, -0.00000425, 0.00000139, -0.00000238, -0.00000135, -0.00000057, -0.00000226, -0.00000172, 0.00000146, 0.00000303, -0.00000502, -0.00000383, -0.00000134, },
		{ -0.00000099, -0.00000181, 0.00000226, -0.00000336, 0.00000260, 0.00000253, -0.00000596, -0.00000832, 0.00000179, -0.00000660, -0.00000917, -0.00000554, -0.00003434, -0.00001191, 0.00004533, 0.00004207, 0.00002958, 0.00006356, -0.00000150, -0.00004117, -0.00001438, -0.00001474, 0.00000036, 0.00000401, -0.00000554, 0.00000653, -0.00000557, 0.00000761, -0.00001331, -0.00000199, -0.00000234, 0.00000522, -0.00000717, -0.00000049, 0.00000257, 0.00000133, 0.00000099, -0.00000086, -0.00000309, 0.00000101, 0.00000014, -0.00000029, 0.00000200, 0.00000017, -0.00000055, 0.00000043, 0.00000296, -0.00000096, 0.00000851, 0.00000085, -0.00001127, 0.00000932, -0.00000436, 0.00001411, -0.00001763, -0.00000120, 0.00000532, -0.00001556, -0.00000101, -0.00000250, -0.00000198, -0.00000762, -0.00000367, -0.00000750, 0.00000784, -0.00000867, 0.00000448, -0.00000113, 0.00000639, -0.00000251, 0.00000324, 0.00000394, -0.00000135, -0.00000216, 0.00000079, 0.00000094, 0.00000375, -0.00000058, -0.00000004, -0.00000284, -0.00000121, 0.00000065, 0.00000016, 0.00000703, -0.00000688, -0.00000195, 0.00000185, -0.00000137, -0.00000187, -0.00000626, 0.00000000, -0.00000719, 0.00000465, -0.00000627, 0.00000709, -0.00000286, -0.00000590, -0.00002632, -0.00001169, -0.00008600, 0.00005021, 0.00010252, 0.00007821, 0.00008555, -0.00008266, -0.00001664, -0.00001317, -0.00000707, -0.00000612, -0.00000664, -0.00000883, -0.00000333, -0.00000074, 0.00000464, 0.00000034, -0.00000415, -0.00000070, -0.00000367, 0.00000181, -0.00000397, 0.00000015, -0.00000159, -0.00000154, -0.00000185, 0.00000340, 0.00000014, -0.00000542, -0.00000272, },
		{ -0.00000174, -0.00000157, -0.00000047, 0.00000099, -0.00000339, 0.00000544, -0.00000120, -0.00000709, -0.00000655, 0.00000269, -0.00001188, -0.00000403, -0.00001448, -0.00003559, 0.00000396, 0.00005721, 0.00003116, 0.00004689, 0.00005335, -0.00002899, -0.00003387, -0.00001059, -0.00001507, 0.00000840, -0.00000397, 0.00000158, 0.00000058, -0.00000055, 0.00000204, -0.00001276, -0.00000002, -0.00000093, 0.00000363, -0.00000855, 0.00000282, 0.00000127, 0.00000200, -0.00000006, -0.00000134, -0.00000259, 0.00000170, -0.00000076, 0.00000090, 0.00000160, -0.00000029, -0.00000031, 0.00000082, 0.00000298, -0.00000081, 0.00001130, -0.00000831, -0.00000190, 0.00000458, -0.00000009, 0.00000994, -0.00002304, 0.00001130, -0.00000775, -0.00000893, -0.00000224, -0.00000018, -0.00000692, -0.00000355, -0.00000711, -0.00000221, 0.00000453, -0.00000736, 0.00000585, -0.00000167, 0.00000618, -0.00000348, 0.00000603, 0.00000124, -0.00000152, -0.00000165, 0.00000149, 0.00000126, 0.00000330, -0.00000155, 0.00000026, -0.00000394, 0.00000103, -0.00000112, 0.00000349, 0.00000396, -0.00000869, 0.00000198, -0.00000072, 0.00000045, -0.00000578, -0.00000265, -0.00000343, -0.00000258, 0.00000146, -0.00000302, 0.00000540, -0.00000234, -0.00001569, -0.00001737, -0.00003560, -0.00006606, 0.00010347, 0.00006816, 0.00010088, 0.00002793, -0.00007820, -0.00000179, -0.00001837, -0.00000168, -0.00000942, -0.00000460, -0.00000948, -0.00000099, -0.00000034, 0.00000599, -0.00000365, -0.00000164, -0.00000218, -0.00000196, 0.00000070, -0.00000436, 0.00000122, -0.00000243, -0.00000099, -0.00000140, 0.00000439, -0.00000310, -0.00000454, },
		{ -0.00000275, -0.00000168, -0.00000175, 0.00000093, -0.00000160, -0.00000043, 0.00000506, -0.00000392, -0.00000824, -0.00000278, -0.00000088, -0.00001243, -0.00000261, -0.00002344, -0.00003319, 0.00002543, 0.00005846, 0.00002714, 0.00006434, 0.00002683, -0.00004534, -0.00002091, -0.00001369, -0.00000842, 0.00000916, -0.00000741, 0.00000621, -0.00000453, 0.00000458, -0.00000563, -0.00000835, -0.00000043, 0.00000168, -0.00000032, -0.00000682, 0.00000414, 0.00000084, 0.00000185, -0.00000075, -0.00000184, -0.00000157, 0.00000160, -0.00000101, 0.00000199, 0.00000063, -0.00000003, -0.00000076, 0.00000222, 0.00000155, 0.00000196, 0.00000932, -0.00001307, 0.00000727, -0.00000351, 0.00000930, -0.00000348, -0.00001552, 0.00001262, -0.00001546, -0.00000252, -0.00000368, 0.00000036, -0.00001022, 0.00000066, -0.00001134, 0.00000459, -0.00000141, -0.00000261, 0.00000414, -0.00000061, 0.00000465, -0.00000238, 0.00000690, -0.00000105, -0.00000134, -0.00000074, 0.00000161, 0.00000190, 0.00000194, -0.00000135, -0.00000065, -0.00000337, 0.00000181, -0.00000176, 0.00000646, -0.00000123, -0.00000681, 0.00000328, -0.00000180, 0.00000049, -0.00000810, 0.00000023, -0.00000613, 0.00000196, -0.00000252, 0.00000138, 0.00000258, -0.00000356, -0.00002299, -0.00001028, -0.00006520, -0.00000926, 0.00011286, 0.00005308, 0.00010751, -0.00002828, -0.00005303, -0.00000255, -0.00001558, -0.00000197, -0.00000910, -0.00000476, -0.00000839, 0.00000006, 0.00000150, 0.00000463, -0.00000568, 0.00000068, -0.00000408, 0.00000046, -0.00000193, -0.00000288, 0.00000101, -0.00000242, -0.00000107, -0.00000001, 0.00000363, -0.00000553, },
		{ -0.00000619, -0.00000099, -0.00000252, -0.00000066, 0.00000060, -0.00000284, 0.00000295, 0.00000268, -0.00000603, -0.00000798, 0.00000010, -0.00000605, -0.00000972, -0.00000528, -0.00003142, -0.00002488, 0.00004825, 0.00004950, 0.00003564, 0.00007033, -0.00000837, -0.00004534, -0.00001165, -0.00001763, 0.00000172, 0.00000335, -0.00000448, 0.00000522, -0.00000498, 0.00000547, -0.00001073, -0.00000369, -0.00000104, 0.00000341, -0.00000435, -0.00000353, 0.00000418, 0.00000076, 0.00000131, -0.00000130, -0.00000201, -0.00000058, 0.00000116, -0.00000067, 0.00000263, -0.00000049, 0.00000062, -0.00000158, 0.00000438, -0.00000109, 0.00000702, 0.00000178, -0.00000949, 0.00000919, -0.00000666, 0.00001495, -0.00001733, -0.00000092, 0.00000329, -0.00001408, -0.00000069, -0.00000258, -0.00000313, -0.00000838, 0.00000067, -0.00001235, 0.00000938, -0.00000699, 0.00000345, 0.00000011, 0.00000201, 0.00000219, 0.00000015, 0.00000591, -0.00000257, -0.00000067, -0.00000011, 0.00000182, 0.00000205, 0.00000073, -0.00000100, -0.00000181, -0.00000183, 0.00000125, -0.00000065, 0.00000740, -0.00000636, -0.00000244, 0.00000172, -0.00000062, -0.00000238, -0.00000699, 0.00000019, -0.00000562, 0.00000332, -0.00000390, 0.00000363, 0.00000147, -0.00001010, -0.00002092, -0.00001864, -0.00007426, 0.00005538, 0.00008820, 0.00006385, 0.00008477, -0.00006188, -0.00002637, -0.00000939, -0.00000928, -0.00000533, -0.00000636, -0.00000663, -0.00000593, 0.00000043, 0.00000379, 0.00000104, -0.00000483, 0.00000114, -0.00000482, 0.00000208, -0.00000472, -0.00000023, -0.00000035, -0.00000162, -0.00000161, 0.00000200, 0.00000106, },
		{ -0.00000251, -0.00000476, -0.00000046, -0.00000312, 0.00000090, -0.00000140, -0.00000147, 0.00000450, -0.00000009, -0.00000794, -0.00000565, -0.00000018, -0.00000935, -0.00000703, -0.00001036, -0.00003842, -0.00000755, 0.00006476, 0.00003737, 0.00005464, 0.00005641, -0.00003906, -0.00003250, -0.00001120, -0.00001579, 0.00000843, -0.00000326, 0.00000103, 0.00000073, -0.00000169, 0.00000184, -0.00001148, -0.00000082, -0.00000071, 0.00000322, -0.00000696, 0.00000006, 0.00000344, 0.00000080, 0.00000057, -0.00000166, -0.00000192, 0.00000028, 0.00000050, 0.00000034, 0.00000223, -0.00000076, 0.00000043, -0.00000120, 0.00000532, -0.00000246, 0.00001088, -0.00000722, 0.00000001, 0.00000294, -0.00000119, 0.00001040, -0.00002193, 0.00001004, -0.00000867, -0.00000723, -0.00000249, -0.00000013, -0.00000891, -0.00000191, -0.00000487, -0.00000752, 0.00000893, -0.00000870, 0.00000748, -0.00000399, 0.00000534, -0.00000066, 0.00000349, 0.00000315, -0.00000265, -0.00000017, 0.00000047, 0.00000198, 0.00000173, -0.00000004, -0.00000097, -0.00000252, -0.00000015, 0.00000000, 0.00000206, 0.00000519, -0.00000882, 0.00000159, -0.00000056, 0.00000098, -0.00000640, -0.00000362, -0.00000194, -0.00000264, 0.00000185, -0.00000279, 0.00000409, 0.00000020, -0.00001836, -0.00001331, -0.00004192, -0.00004793, 0.00009576, 0.00005750, 0.00008478, 0.00003982, -0.00006730, -0.00001016, -0.00001388, -0.00000474, -0.00000792, -0.00000379, -0.00000862, -0.00000302, 0.00000092, 0.00000504, -0.00000304, -0.00000179, -0.00000075, -0.00000336, 0.00000136, -0.00000577, 0.00000194, -0.00000182, -0.00000076, -0.00000188, 0.00000363, },
		{ 0.00000374, -0.00000559, -0.00000199, -0.00000164, -0.00000223, 0.00000115, -0.00000309, 0.00000135, 0.00000400, -0.00000278, -0.00000889, -0.00000268, -0.00000310, -0.00000981, -0.00000621, -0.00001696, -0.00004263, 0.00001894, 0.00006842, 0.00003244, 0.00007373, 0.00002296, -0.00005302, -0.00001749, -0.00001638, -0.00000738, 0.00000810, -0.00000566, 0.00000434, -0.00000310, 0.00000179, -0.00000360, -0.00000899, 0.00000022, 0.00000038, 0.00000130, -0.00000760, 0.00000325, 0.00000211, 0.00000102, -0.00000039, -0.00000168, -0.00000179, 0.00000105, -0.00000021, 0.00000180, 0.00000080, 0.00000011, -0.00000112, 0.00000136, 0.00000344, -0.00000025, 0.00000965, -0.00001145, 0.00000821, -0.00000544, 0.00000864, -0.00000258, -0.00001463, 0.00001034, -0.00001478, -0.00000147, -0.00000393, -0.00000003, -0.00001211, 0.00000431, -0.00001265, 0.00000177, 0.00000293, -0.00000484, 0.00000663, -0.00000519, 0.00000735, -0.00000257, 0.00000632, -0.00000044, -0.00000130, -0.00000034, 0.00000141, 0.00000161, 0.00000146, -0.00000064, -0.00000114, -0.00000261, 0.00000106, -0.00000099, 0.00000517, 0.00000035, -0.00000748, 0.00000306, -0.00000134, 0.00000063, -0.00000881, -0.00000055, -0.00000383, 0.00000042, -0.00000025, -0.00000109, 0.00000469, -0.00000486, -0.00002116, -0.00001197, -0.00006316, 0.00000478, 0.00009871, 0.00004425, 0.00009400, -0.00000803, -0.00005329, -0.00000570, -0.00001334, -0.00000404, -0.00000761, -0.00000344, -0.00000914, -0.00000060, 0.00000204, 0.00000407, -0.00000533, 0.00000121, -0.00000356, -0.00000046, -0.00000163, -0.00000419, 0.00000236, -0.00000238, -0.00000060, -0.00000119, },
		{ 0.00000053, 0.00000178, -0.00000662, 0.00000041, -0.00000357, 0.00000001, -0.00000070, -0.00000257, 0.00000345, 0.00000249, -0.00000561, -0.00000785, -0.00000146, -0.00000607, -0.00000910, -0.00000678, -0.00002595, -0.00003928, 0.00004907, 0.00005893, 0.00004206, 0.00007827, -0.00001850, -0.00004698, -0.00001041, -0.00001934, 0.00000224, 0.00000306, -0.00000358, 0.00000387, -0.00000409, 0.00000288, -0.00000797, -0.00000519, 0.00000011, 0.00000166, -0.00000175, -0.00000612, 0.00000527, 0.00000064, 0.00000129, -0.00000137, -0.00000149, -0.00000149, 0.00000147, -0.00000043, 0.00000281, -0.00000064, 0.00000108, -0.00000279, 0.00000488, -0.00000032, 0.00000487, 0.00000294, -0.00000776, 0.00000898, -0.00000828, 0.00001435, -0.00001542, -0.00000153, 0.00000142, -0.00001247, -0.00000042, -0.00000263, -0.00000436, -0.00000889, 0.00000456, -0.00001649, 0.00001025, -0.00000494, 0.00000260, 0.00000091, -0.00000193, 0.00000621, -0.00000198, 0.00000687, -0.00000300, 0.00000032, -0.00000071, 0.00000238, 0.00000086, 0.00000136, -0.00000124, -0.00000136, -0.00000207, 0.00000139, -0.00000074, 0.00000681, -0.00000487, -0.00000359, 0.00000202, -0.00000035, -0.00000225, -0.00000828, 0.00000069, -0.00000420, 0.00000216, -0.00000179, 0.00000062, 0.00000493, -0.00001329, -0.00001642, -0.00002470, -0.00006211, 0.00005781, 0.00007540, 0.00005286, 0.00007992, -0.00004143, -0.00003370, -0.00000754, -0.00001001, -0.00000569, -0.00000507, -0.00000542, -0.00000755, 0.00000091, 0.00000350, 0.00000094, -0.00000455, 0.00000195, -0.00000501, 0.00000168, -0.00000502, -0.00000091, 0.00000107, -0.00000183, -0.00000122, },
		{ -0.00000198, 0.00000246, -0.00000173, -0.00000498, 0.00000077, -0.00000437, 0.00000180, -0.00000334, 0.00000006, 0.00000378, 0.00000069, -0.00000818, -0.00000516, -0.00000291, -0.00000713, -0.00000918, -0.00000790, -0.00003783, -0.00002292, 0.00007273, 0.00004434, 0.00006418, 0.00005801, -0.00005022, -0.00002941, -0.00001314, -0.00001521, 0.00000728, -0.00000172, -0.00000011, 0.00000127, -0.00000290, 0.00000132, -0.00000989, -0.00000165, -0.00000045, 0.00000260, -0.00000519, -0.00000262, 0.00000540, -0.00000020, 0.00000112, -0.00000197, -0.00000144, -0.00000078, 0.00000126, 0.00000029, 0.00000252, -0.00000093, 0.00000068, -0.00000253, 0.00000667, -0.00000288, 0.00000918, -0.00000508, 0.00000094, 0.00000233, -0.00000289, 0.00001055, -0.00001973, 0.00000775, -0.00000902, -0.00000571, -0.00000261, -0.00000022, -0.00001075, -0.00000029, -0.00000292, -0.00001206, 0.00001230, -0.00000900, 0.00000847, -0.00000597, 0.00000419, 0.00000222, 0.00000128, 0.00000443, -0.00000313, 0.00000080, -0.00000030, 0.00000255, 0.00000041, 0.00000104, -0.00000160, -0.00000167, -0.00000099, 0.00000084, 0.00000107, 0.00000564, -0.00000794, 0.00000031, 0.00000026, 0.00000084, -0.00000617, -0.00000548, 0.00000027, -0.00000322, 0.00000269, -0.00000298, 0.00000330, 0.00000176, -0.00001961, -0.00001093, -0.00004583, -0.00003182, 0.00008699, 0.00004914, 0.00007067, 0.00004661, -0.00005387, -0.00001857, -0.00000980, -0.00000735, -0.00000673, -0.00000277, -0.00000806, -0.00000460, 0.00000190, 0.00000416, -0.00000264, -0.00000151, 0.00000000, -0.00000389, 0.00000125, -0.00000652, 0.00000211, -0.00000080, -0.00000088, },
		},
		{ { -0.00000082, -0.00000006, -0.00001202, 0.00000157, -0.00000053, 0.00000070, 0.00000229, -0.00000206, -0.00000485, -0.00000631, -0.00000382, -0.00000723, -0.00000396, -0.00001183, -0.00006238, 0.00001223, 0.00008200, 0.00003733, 0.00008261, 0.00001361, -0.00005756, -0.00001224, -0.00001010, -0.00000885, -0.00000322, 0.00000002, 0.00000389, -0.00000136, -0.00001255, -0.00000309, 0.00000253, -0.00000378, -0.00000219, 0.00000024, 0.00000596, -0.00000001, 0.00000147, -0.00000112, 0.00000879, -0.00000160, -0.00000438, -0.00000530, 0.00001334, -0.00000001, 0.00000022, -0.00000281, -0.00000047, -0.00000045, -0.00000666, 0.00000132, 0.00000067, -0.00000296, -0.00000455, 0.00000383, -0.00001619, 0.00000508, 0.00000380, 0.00000231, -0.00001532, -0.00000290, 0.00001341, -0.00000333, 0.00000158, 0.00000174, 0.00000265, 0.00000174, 0.00000158, -0.00000333, 0.00001341, -0.00000290, -0.00001532, 0.00000231, 0.00000380, 0.00000508, -0.00001619, 0.00000383, -0.00000455, -0.00000296, 0.00000067, 0.00000132, -0.00000666, -0.00000045, -0.00000047, -0.00000281, 0.00000022, -0.00000001, 0.00001334, -0.00000530, -0.00000438, -0.00000160, 0.00000879, -0.00000112, 0.00000147, -0.00000001, 0.00000596, 0.00000024, -0.00000219, -0.00000378, 0.00000253, -0.00000309, -0.00001255, -0.00000136, 0.00000389, 0.00000002, -0.00000322, -0.00000885, -0.00001010, -0.00001224, -0.00005756, 0.00001361, 0.00008261, 0.00003733, 0.00008200, 0.00001223, -0.00006238, -0.00001183, -0.00000396, -0.00000723, -0.00000382, -0.00000631, -0.00000485, -0.00000206, 0.00000229, 0.00000070, -0.00000053, 0.00000157, -0.00001202, -0.00000006, },
		{ 0.00000112, -0.00000046, -0.00000297, -0.00001089, 0.00000422, -0.00000289, 0.00000329, 0.00000009, -0.00000201, -0.00000631, -0.00000521, -0.00000474, -0.00000698, -0.00000292, -0.00002353, -0.00006373, 0.00004535, 0.00007373, 0.00004603, 0.00008549, -0.00002453, -0.00004868, -0.00000597, -0.00001263, -0.00000616, -0.00000306, 0.00000148, 0.00000372, -0.00000491, -0.00001240, 0.00000079, 0.00000009, -0.00000296, -0.00000269, 0.00000325, 0.00000370, 0.00000072, -0.00000006, 0.00000189, 0.00000747, -0.00000367, -0.00000491, -0.00000206, 0.00001433, -0.00000383, 0.00000262, -0.00000556, 0.00000238, -0.00000402, -0.00000399, 0.00000079, 0.00000149, -0.00000643, 0.00000023, -0.00000211, -0.00001348, 0.00000887, 0.00000104, 0.00000092, -0.00001857, 0.00000683, 0.00000836, -0.00000371, 0.00000287, 0.00000180, 0.00000240, 0.00000205, -0.00000078, 0.00000039, 0.00001396, -0.00001122, -0.00001011, 0.00000259, 0.00000710, -0.00000213, -0.00001363, 0.00000600, -0.00000759, 0.00000014, -0.00000018, 0.00000105, -0.00000814, 0.00000250, -0.00000310, -0.00000008, -0.00000241, 0.00000495, 0.00000981, -0.00000667, -0.00000437, 0.00000203, 0.00000732, -0.00000216, 0.00000191, 0.00000070, 0.00000667, -0.00000236, -0.00000162, -0.00000404, 0.00000447, -0.00000812, -0.00001004, 0.00000090, 0.00000376, -0.00000153, -0.00000338, -0.00001164, -0.00000614, -0.00002550, -0.00005135, 0.00004777, 0.00006693, 0.00004044, 0.00007921, -0.00001756, -0.00005319, -0.00000397, -0.00000591, -0.00000666, -0.00000345, -0.00000707, -0.00000344, -0.00000174, 0.00000376, -0.00000178, 0.00000251, -0.00000321, -0.00001017, },
		{ -0.00000637, 0.00000079, 0.00000021, -0.00000681, -0.00000706, 0.00000392, -0.00000319, 0.00000458, -0.00000180, -0.00000224, -0.00000719, -0.00000446, -0.00000546, -0.00000645, -0.00000332, -0.00003840, -0.00005382, 0.00007515, 0.00005925, 0.00006569, 0.00006840, -0.00005426, -0.00003149, -0.00000680, -0.00001202, -0.00000481, -0.00000225, 0.00000252, 0.00000296, -0.00000901, -0.00000952, 0.00000232, -0.00000144, -0.00000266, -0.00000203, 0.00000509, 0.00000148, 0.00000167, -0.00000144, 0.00000557, 0.00000394, -0.00000393, -0.00000608, 0.00000303, 0.00001213, -0.00000525, 0.00000351, -0.00000713, 0.00000435, -0.00000711, -0.00000106, 0.00000010, 0.00000139, -0.00000863, 0.00000454, -0.00000970, -0.00000610, 0.00000782, 0.00000084, -0.00000331, -0.00001716, 0.00001422, 0.00000146, -0.00000136, 0.00000283, 0.00000225, 0.00000189, 0.00000206, -0.00000284, 0.00000608, 0.00000958, -0.00001573, -0.00000531, 0.00000360, 0.00000828, -0.00000957, -0.00000759, 0.00000436, -0.00000791, 0.00000183, -0.00000044, -0.00000038, -0.00000786, 0.00000403, -0.00000462, 0.00000163, -0.00000398, 0.00000950, 0.00000500, -0.00000681, -0.00000405, 0.00000596, 0.00000382, -0.00000115, 0.00000100, 0.00000318, 0.00000488, -0.00000322, -0.00000212, -0.00000260, 0.00000416, -0.00001216, -0.00000623, 0.00000214, 0.00000310, -0.00000251, -0.00000465, -0.00001244, -0.00000470, -0.00004062, -0.00002983, 0.00006861, 0.00004839, 0.00005092, 0.00006495, -0.00003947, -0.00003985, 0.00000001, -0.00000802, -0.00000523, -0.00000402, -0.00000697, -0.00000251, -0.00000067, 0.00000375, -0.00000280, 0.00000438, -0.00000827, },
		{ -0.00001160, -0.00000232, -0.00000014, 0.00000027, -0.00001006, -0.00000200, 0.00000129, -0.00000117, 0.00000382, -0.00000259, -0.00000325, -0.00000708, -0.00000459, -0.00000535, -0.00000640, -0.00000559, -0.00005478, -0.00003069, 0.00009444, 0.00004715, 0.00008856, 0.00003168, -0.00006598, -0.00001556, -0.00001066, -0.00000947, -0.00000438, -0.00000111, 0.00000348, 0.00000081, -0.00001196, -0.00000540, 0.00000191, -0.00000188, -0.00000295, -0.00000012, 0.00000512, 0.00000037, 0.00000182, -0.00000146, 0.00000819, -0.00000027, -0.00000313, -0.00000721, 0.00000905, 0.00000707, -0.00000378, 0.00000201, -0.00000657, 0.00000442, -0.00000864, 0.00000116, -0.00000012, -0.00000031, -0.00000811, 0.00000577, -0.00001537, 0.00000280, 0.00000322, 0.00000304, -0.00001030, -0.00001003, 0.00001618, -0.00000389, 0.00000177, 0.00000220, 0.00000263, 0.00000161, 0.00000108, -0.00000287, 0.00001104, 0.00000210, -0.00001621, -0.00000176, 0.00000580, 0.00000589, -0.00001425, -0.00000096, 0.00000044, -0.00000587, 0.00000191, -0.00000000, -0.00000282, -0.00000576, 0.00000376, -0.00000458, 0.00000173, -0.00000358, 0.00001225, 0.00000030, -0.00000655, -0.00000245, 0.00000840, 0.00000013, 0.00000064, -0.00000019, 0.00000634, 0.00000117, -0.00000228, -0.00000377, 0.00000059, 0.00000083, -0.00001350, -0.00000268, 0.00000290, 0.00000180, -0.00000277, -0.00000725, -0.00001027, -0.00000926, -0.00005000, 0.00000076, 0.00007319, 0.00003508, 0.00006223, 0.00004225, -0.00005145, -0.00002569, 0.00000057, -0.00000945, -0.00000346, -0.00000537, -0.00000593, -0.00000217, 0.00000106, 0.00000222, -0.00000179, 0.00000372, },
		{ 0.00000038, -0.00001203, 0.00000053, -0.00000077, -0.00000102, -0.00001120, 0.00000222, -0.00000195, 0.00000215, 0.00000130, -0.00000218, -0.00000506, -0.00000605, -0.00000553, -0.00000435, -0.00000718, -0.00001046, -0.00006939, 0.00000460, 0.00009837, 0.00004652, 0.00010174, -0.00001503, -0.00005805, -0.00000779, -0.00001282, -0.00000739, -0.00000362, -0.00000036, 0.00000465, -0.00000326, -0.00001217, -0.00000211, 0.00000098, -0.00000207, -0.00000300, 0.00000209, 0.00000386, 0.00000038, 0.00000106, 0.00000029, 0.00000851, -0.00000353, -0.00000248, -0.00000710, 0.00001429, 0.00000059, -0.00000005, -0.00000177, -0.00000363, 0.00000216, -0.00000798, 0.00000205, 0.00000030, -0.00000348, -0.00000458, 0.00000242, -0.00001585, 0.00000896, -0.00000145, 0.00000511, -0.00001741, 0.00000099, 0.00001210, -0.00000544, 0.00000383, 0.00000186, 0.00000253, 0.00000169, -0.00000092, -0.00000000, 0.00001273, -0.00000556, -0.00001403, 0.00000111, 0.00000803, 0.00000046, -0.00001471, 0.00000375, -0.00000370, -0.00000268, 0.00000091, 0.00000071, -0.00000563, -0.00000237, 0.00000193, -0.00000307, 0.00000024, -0.00000101, 0.00001247, -0.00000333, -0.00000634, 0.00000080, 0.00000806, -0.00000185, 0.00000155, 0.00000002, 0.00000807, -0.00000251, -0.00000088, -0.00000531, 0.00000410, -0.00000456, -0.00001190, -0.00000018, 0.00000331, 0.00000029, -0.00000304, -0.00000992, -0.00000667, -0.00001993, -0.00004779, 0.00003113, 0.00006471, 0.00003084, 0.00006852, 0.00001595, -0.00005377, -0.00001325, -0.00000142, -0.00000956, -0.00000214, -0.00000685, -0.00000434, -0.00000205, 0.00000282, -0.00000004, 0.00000068, },
		{ 0.00000315, -0.00000448, -0.00000969, 0.00000159, -0.00000080, -0.00000368, -0.00000957, 0.00000398, -0.00000383, 0.00000499, -0.00000175, -0.00000131, -0.00000703, -0.00000479, -0.00000647, -0.00000333, -0.00000836, -0.00001933, -0.00007704, 0.00004653, 0.00008763, 0.00006178, 0.00009288, -0.00005596, -0.00003787, -0.00000880, -0.00001132, -0.00000697, -0.00000203, -0.00000006, 0.00000549, -0.00000845, -0.00000945, -0.00000083, 0.00000070, -0.00000275, -0.00000204, 0.00000348, 0.00000239, 0.00000082, 0.00000006, 0.00000324, 0.00000628, -0.00000479, -0.00000300, -0.00000454, 0.00001674, -0.00000501, 0.00000409, -0.00000652, 0.00000081, -0.00000190, -0.00000534, 0.00000166, 0.00000085, -0.00000693, 0.00000051, -0.00000464, -0.00001021, 0.00000942, -0.00000274, 0.00000378, -0.00002067, 0.00001167, 0.00000432, -0.00000324, 0.00000414, 0.00000211, 0.00000203, 0.00000167, -0.00000286, 0.00000492, 0.00001025, -0.00001114, -0.00001077, 0.00000417, 0.00000857, -0.00000594, -0.00001143, 0.00000532, -0.00000633, 0.00000022, -0.00000025, 0.00000094, -0.00000778, 0.00000122, -0.00000058, -0.00000087, -0.00000208, 0.00000310, 0.00001028, -0.00000566, -0.00000587, 0.00000489, 0.00000510, -0.00000145, 0.00000079, 0.00000254, 0.00000678, -0.00000419, -0.00000070, -0.00000508, 0.00000577, -0.00000982, -0.00000846, 0.00000134, 0.00000317, -0.00000079, -0.00000429, -0.00001075, -0.00000509, -0.00003276, -0.00003292, 0.00005325, 0.00004974, 0.00003519, 0.00006630, -0.00000885, -0.00004843, -0.00000411, -0.00000484, -0.00000806, -0.00000204, -0.00000761, -0.00000291, -0.00000158, 0.00000378, -0.00000180, },
		{ -0.00000206, 0.00000403, -0.00000896, -0.00000583, 0.00000122, -0.00000062, -0.00000687, -0.00000569, 0.00000283, -0.00000308, 0.00000568, -0.00000376, -0.00000116, -0.00000811, -0.00000436, -0.00000621, -0.00000377, -0.00000891, -0.00003373, -0.00007140, 0.00008549, 0.00006996, 0.00008836, 0.00005741, -0.00007667, -0.00001767, -0.00001356, -0.00000795, -0.00000722, -0.00000048, 0.00000073, 0.00000450, -0.00001258, -0.00000545, -0.00000104, 0.00000097, -0.00000375, -0.00000018, 0.00000365, 0.00000139, 0.00000107, -0.00000037, 0.00000620, 0.00000227, -0.00000397, -0.00000497, 0.00000091, 0.00001491, -0.00000740, 0.00000624, -0.00001005, 0.00000488, -0.00000624, -0.00000184, 0.00000074, 0.00000054, -0.00000875, 0.00000431, -0.00001210, -0.00000081, 0.00000450, 0.00000038, -0.00000245, -0.00001714, 0.00001748, -0.00000317, 0.00000076, 0.00000329, 0.00000255, 0.00000158, 0.00000083, -0.00000316, 0.00000968, 0.00000468, -0.00001389, -0.00000716, 0.00000742, 0.00000637, -0.00001097, -0.00000615, 0.00000396, -0.00000668, 0.00000181, -0.00000076, -0.00000001, -0.00000823, 0.00000373, -0.00000257, 0.00000083, -0.00000392, 0.00000742, 0.00000660, -0.00000709, -0.00000425, 0.00000805, 0.00000121, 0.00000031, -0.00000077, 0.00000648, 0.00000265, -0.00000327, -0.00000237, -0.00000240, 0.00000421, -0.00001292, -0.00000467, 0.00000223, 0.00000245, -0.00000137, -0.00000662, -0.00000898, -0.00000846, -0.00004172, -0.00000932, 0.00006314, 0.00003517, 0.00004455, 0.00005496, -0.00002804, -0.00003839, 0.00000115, -0.00000847, -0.00000532, -0.00000342, -0.00000707, -0.00000222, -0.00000036, 0.00000339, },
		{ 0.00000179, -0.00000068, 0.00000248, -0.00001144, -0.00000199, 0.00000024, -0.00000091, -0.00000927, -0.00000114, -0.00000026, 0.00000015, 0.00000356, -0.00000371, -0.00000262, -0.00000756, -0.00000542, -0.00000402, -0.00000674, -0.00000836, -0.00005370, -0.00004763, 0.00011082, 0.00005782, 0.00011245, 0.00000306, -0.00007176, -0.00000741, -0.00001573, -0.00000609, -0.00000614, -0.00000032, 0.00000295, 0.00000040, -0.00001337, -0.00000261, -0.00000110, 0.00000078, -0.00000409, 0.00000161, 0.00000314, 0.00000084, 0.00000103, 0.00000022, 0.00000796, -0.00000210, -0.00000199, -0.00000753, 0.00000837, 0.00000881, -0.00000543, 0.00000464, -0.00001032, 0.00000654, -0.00000902, 0.00000108, 0.00000015, -0.00000137, -0.00000751, 0.00000410, -0.00001573, 0.00000761, -0.00000223, 0.00000555, -0.00001179, -0.00000712, 0.00001616, -0.00000696, 0.00000413, 0.00000240, 0.00000264, 0.00000148, -0.00000101, -0.00000080, 0.00001192, -0.00000177, -0.00001436, -0.00000304, 0.00000985, 0.00000182, -0.00001316, -0.00000092, 0.00000087, -0.00000501, 0.00000181, -0.00000036, -0.00000227, -0.00000652, 0.00000427, -0.00000303, 0.00000097, -0.00000402, 0.00001048, 0.00000259, -0.00000809, -0.00000086, 0.00000856, -0.00000136, 0.00000147, -0.00000111, 0.00000944, -0.00000224, -0.00000104, -0.00000482, 0.00000167, -0.00000022, -0.00001310, -0.00000149, 0.00000259, 0.00000164, -0.00000223, -0.00000870, -0.00000613, -0.00001684, -0.00004208, 0.00001646, 0.00006115, 0.00002608, 0.00005397, 0.00003667, -0.00003947, -0.00002658, 0.00000262, -0.00001102, -0.00000234, -0.00000575, -0.00000532, -0.00000230, 0.00000149, },
		{ 0.00000326, -0.00000016, 0.00000148, -0.00000112, -0.00001126, 0.00000067, -0.00000067, -0.00000214, -0.00000975, 0.00000224, -0.00000334, 0.00000432, -0.00000051, -0.00000186, -0.00000552, -0.00000563, -0.00000750, -0.00000066, -0.00001163, -0.00000831, -0.00007569, -0.00000598, 0.00011608, 0.00006214, 0.00011656, -0.00005192, -0.00004800, -0.00000922, -0.00001244, -0.00000736, -0.00000300, -0.00000170, 0.00000617, -0.00000634, -0.00001017, -0.00000237, 0.00000019, -0.00000071, -0.00000296, 0.00000238, 0.00000272, 0.00000034, 0.00000102, 0.00000175, 0.00000767, -0.00000520, -0.00000047, -0.00000886, 0.00001546, 0.00000041, 0.00000003, -0.00000071, -0.00000668, 0.00000474, -0.00000906, 0.00000246, 0.00000015, -0.00000454, -0.00000344, -0.00000087, -0.00001288, 0.00001057, -0.00000599, 0.00000837, -0.00001966, 0.00000556, 0.00000895, -0.00000602, 0.00000547, 0.00000217, 0.00000222, 0.00000146, -0.00000308, 0.00000378, 0.00001050, -0.00000713, -0.00001340, 0.00000180, 0.00001017, -0.00000358, -0.00001232, 0.00000269, -0.00000239, -0.00000235, 0.00000077, 0.00000048, -0.00000518, -0.00000314, 0.00000293, -0.00000189, -0.00000066, -0.00000187, 0.00001142, -0.00000103, -0.00000852, 0.00000383, 0.00000596, -0.00000133, 0.00000068, 0.00000131, 0.00000906, -0.00000521, 0.00000032, -0.00000606, 0.00000490, -0.00000565, -0.00001097, 0.00000085, 0.00000230, 0.00000127, -0.00000426, -0.00000873, -0.00000536, -0.00002690, -0.00003273, 0.00003827, 0.00005067, 0.00002480, 0.00005867, 0.00001551, -0.00004320, -0.00001527, 0.00000083, -0.00001144, -0.00000046, -0.00000781, -0.00000322, -0.00000255, },
		{ -0.00000212, 0.00000408, -0.00000134, 0.00000302, -0.00000542, -0.00000887, 0.00000177, -0.00000126, -0.00000418, -0.00000795, 0.00000309, -0.00000437, 0.00000709, -0.00000448, 0.00000020, -0.00000849, -0.00000374, -0.00000880, 0.00000146, -0.00001580, -0.00001280, -0.00009169, 0.00004596, 0.00010367, 0.00008488, 0.00008898, -0.00008714, -0.00002056, -0.00001709, -0.00000614, -0.00001011, 0.00000064, -0.00000292, 0.00000810, -0.00001300, -0.00000489, -0.00000400, 0.00000250, -0.00000310, -0.00000067, 0.00000204, 0.00000262, -0.00000021, 0.00000126, 0.00000383, 0.00000512, -0.00000586, -0.00000091, -0.00000695, 0.00001913, -0.00000685, 0.00000606, -0.00000764, -0.00000051, 0.00000013, -0.00000653, 0.00000236, 0.00000013, -0.00000739, 0.00000119, -0.00000833, -0.00000444, 0.00000672, -0.00000399, 0.00000534, -0.00002133, 0.00001542, -0.00000014, -0.00000184, 0.00000494, 0.00000245, 0.00000165, 0.00000084, -0.00000387, 0.00000869, 0.00000606, -0.00001061, -0.00001126, 0.00000684, 0.00000775, -0.00000812, -0.00000926, 0.00000399, -0.00000447, -0.00000002, -0.00000029, 0.00000076, -0.00000740, 0.00000052, 0.00000084, -0.00000013, -0.00000315, 0.00000190, 0.00001031, -0.00000419, -0.00000752, 0.00000807, 0.00000173, 0.00000061, -0.00000152, 0.00000607, 0.00000483, -0.00000489, -0.00000057, -0.00000472, 0.00000539, -0.00000997, -0.00000770, 0.00000237, 0.00000159, 0.00000118, -0.00000724, -0.00000623, -0.00000910, -0.00003383, -0.00001616, 0.00005219, 0.00003655, 0.00003073, 0.00005561, -0.00000401, -0.00004095, -0.00000587, -0.00000326, -0.00000933, -0.00000075, -0.00000834, -0.00000188, },
		{ -0.00000189, -0.00000055, 0.00000348, -0.00000111, 0.00000282, -0.00000885, -0.00000543, 0.00000159, -0.00000169, -0.00000640, -0.00000450, 0.00000130, -0.00000229, 0.00000658, -0.00000612, 0.00000047, -0.00000965, -0.00000372, -0.00000724, -0.00000054, -0.00001601, -0.00002655, -0.00009132, 0.00009457, 0.00008538, 0.00011526, 0.00003208, -0.00009073, -0.00000418, -0.00002194, -0.00000206, -0.00001076, 0.00000200, -0.00000159, 0.00000598, -0.00001600, -0.00000090, -0.00000513, 0.00000407, -0.00000486, 0.00000146, 0.00000150, 0.00000233, -0.00000040, 0.00000171, 0.00000586, 0.00000085, -0.00000394, -0.00000380, -0.00000090, 0.00001714, -0.00000962, 0.00000910, -0.00001278, 0.00000535, -0.00000517, -0.00000283, 0.00000161, -0.00000092, -0.00000795, 0.00000317, -0.00001392, 0.00000522, -0.00000113, 0.00000260, -0.00000352, -0.00001490, 0.00001833, -0.00000669, 0.00000283, 0.00000371, 0.00000266, 0.00000134, -0.00000079, -0.00000217, 0.00001161, 0.00000042, -0.00001245, -0.00000765, 0.00001071, 0.00000321, -0.00001069, -0.00000526, 0.00000319, -0.00000476, 0.00000109, -0.00000056, -0.00000031, -0.00000771, 0.00000292, -0.00000052, 0.00000074, -0.00000491, 0.00000590, 0.00000792, -0.00000717, -0.00000418, 0.00000964, -0.00000155, 0.00000228, -0.00000287, 0.00001060, -0.00000128, -0.00000217, -0.00000320, -0.00000118, 0.00000280, -0.00001201, -0.00000419, 0.00000288, 0.00000120, 0.00000023, -0.00000942, -0.00000314, -0.00001689, -0.00003428, 0.00000348, 0.00005670, 0.00002392, 0.00004058, 0.00004461, -0.00001862, -0.00003495, 0.00000071, -0.00000816, -0.00000535, -0.00000321, -0.00000686, },
		{ -0.00000410, -0.00000287, 0.00000178, 0.00000186, 0.00000025, 0.00000068, -0.00001038, -0.00000208, 0.00000063, -0.00000226, -0.00000788, -0.00000079, -0.00000192, 0.00000223, 0.00000254, -0.00000450, -0.00000213, -0.00000805, -0.00000627, -0.00000226, -0.00000770, -0.00001120, -0.00005124, -0.00006685, 0.00012600, 0.00007687, 0.00013317, -0.00003632, -0.00006601, -0.00000473, -0.00001833, -0.00000369, -0.00000715, -0.00000025, 0.00000300, -0.00000104, -0.00001338, -0.00000072, -0.00000364, 0.00000327, -0.00000455, 0.00000208, 0.00000177, 0.00000123, 0.00000026, 0.00000220, 0.00000708, -0.00000371, -0.00000070, -0.00000787, 0.00000808, 0.00000962, -0.00000655, 0.00000688, -0.00001339, 0.00000812, -0.00000880, 0.00000041, 0.00000097, -0.00000336, -0.00000542, 0.00000047, -0.00001372, 0.00001065, -0.00000755, 0.00000914, -0.00001418, -0.00000273, 0.00001373, -0.00000825, 0.00000578, 0.00000285, 0.00000240, 0.00000125, -0.00000294, 0.00000200, 0.00001115, -0.00000462, -0.00001314, -0.00000241, 0.00001197, -0.00000200, -0.00001104, -0.00000155, 0.00000119, -0.00000358, 0.00000092, 0.00000000, -0.00000266, -0.00000584, 0.00000334, -0.00000032, -0.00000026, -0.00000464, 0.00000877, 0.00000499, -0.00000975, 0.00000136, 0.00000744, -0.00000190, 0.00000165, -0.00000105, 0.00001177, -0.00000600, 0.00000049, -0.00000554, 0.00000259, -0.00000155, -0.00001188, -0.00000082, 0.00000216, 0.00000186, -0.00000269, -0.00000863, -0.00000296, -0.00002516, -0.00002804, 0.00002251, 0.00005213, 0.00001718, 0.00004928, 0.00002845, -0.00002730, -0.00002691, 0.00000374, -0.00001189, -0.00000125, -0.00000658, },
		{ -0.00000889, -0.00000162, -0.00000364, 0.00000384, 0.00000022, 0.00000167, -0.00000259, -0.00000989, 0.00000040, -0.00000063, -0.00000315, -0.00000793, 0.00000161, -0.00000441, 0.00000695, -0.00000321, -0.00000077, -0.00000670, -0.00000466, -0.00000991, 0.00000390, -0.00001752, -0.00000486, -0.00008196, -0.00001867, 0.00013375, 0.00008821, 0.00011954, -0.00009059, -0.00002970, -0.00001668, -0.00000815, -0.00000963, -0.00000072, -0.00000437, 0.00000856, -0.00001052, -0.00000651, -0.00000411, 0.00000051, 0.00000010, -0.00000217, 0.00000108, 0.00000293, -0.00000057, 0.00000172, 0.00000281, 0.00000660, -0.00000668, 0.00000162, -0.00001045, 0.00001664, -0.00000048, 0.00000066, -0.00000006, -0.00000898, 0.00000666, -0.00000947, 0.00000229, 0.00000042, -0.00000619, -0.00000115, -0.00000599, -0.00000697, 0.00000875, -0.00000802, 0.00001048, -0.00002087, 0.00000971, 0.00000474, -0.00000524, 0.00000640, 0.00000265, 0.00000182, 0.00000082, -0.00000431, 0.00000711, 0.00000753, -0.00000818, -0.00001266, 0.00000375, 0.00001002, -0.00000632, -0.00000962, 0.00000097, -0.00000083, -0.00000191, 0.00000022, 0.00000060, -0.00000523, -0.00000276, 0.00000239, 0.00000096, -0.00000282, -0.00000221, 0.00001002, 0.00000161, -0.00001083, 0.00000734, 0.00000255, 0.00000055, -0.00000133, 0.00000428, 0.00000805, -0.00000695, 0.00000089, -0.00000562, 0.00000442, -0.00000575, -0.00001030, 0.00000207, 0.00000062, 0.00000323, -0.00000730, -0.00000436, -0.00000813, -0.00002946, -0.00001743, 0.00003834, 0.00004049, 0.00001869, 0.00005197, 0.00001166, -0.00003112, -0.00001783, 0.00000280, -0.00001271, 0.00000088, },
		{ -0.00000014, -0.00000865, -0.00000086, -0.00000304, 0.00000463, -0.00000050, 0.00000203, -0.00000574, -0.00000795, 0.00000172, -0.00000186, -0.00000430, -0.00000635, 0.00000166, -0.00000420, 0.00000902, -0.00000762, 0.00000228, -0.00001061, -0.00000215, -0.00001139, 0.00000671, -0.00002444, -0.00000469, -0.00010686, 0.00004253, 0.00012334, 0.00011581, 0.00006828, -0.00011041, -0.00000220, -0.00002784, 0.00000124, -0.00001459, 0.00000419, -0.00000642, 0.00001077, -0.00001760, 0.00000036, -0.00000803, 0.00000505, -0.00000359, 0.00000070, -0.00000025, 0.00000391, -0.00000205, 0.00000329, 0.00000377, 0.00000389, -0.00000661, 0.00000099, -0.00000879, 0.00002079, -0.00000859, 0.00000793, -0.00000853, -0.00000166, 0.00000198, -0.00000753, 0.00000296, -0.00000083, -0.00000746, 0.00000174, -0.00001234, 0.00000295, 0.00000095, -0.00000194, 0.00000448, -0.00001953, 0.00001691, -0.00000386, -0.00000015, 0.00000550, 0.00000271, 0.00000132, -0.00000047, -0.00000352, 0.00001094, 0.00000226, -0.00001040, -0.00001046, 0.00000920, 0.00000553, -0.00000883, -0.00000720, 0.00000198, -0.00000194, -0.00000076, -0.00000007, 0.00000020, -0.00000656, -0.00000009, 0.00000159, 0.00000177, -0.00000543, 0.00000134, 0.00001000, -0.00000255, -0.00000891, 0.00001094, -0.00000211, 0.00000343, -0.00000434, 0.00001070, 0.00000091, -0.00000425, -0.00000125, -0.00000313, 0.00000349, -0.00000859, -0.00000774, 0.00000374, -0.00000046, 0.00000360, -0.00001118, 0.00000086, -0.00001765, -0.00002749, -0.00000509, 0.00004892, 0.00002566, 0.00002755, 0.00004622, -0.00000178, -0.00003187, -0.00000859, -0.00000156, -0.00001003, },
		{ -0.00000506, -0.00000373, -0.00000590, -0.00000205, -0.00000079, 0.00000389, -0.00000008, 0.00000086, -0.00000782, -0.00000531, 0.00000184, -0.00000286, -0.00000544, -0.00000367, -0.00000050, -0.00000063, 0.00000667, -0.00000810, 0.00000194, -0.00001124, -0.00000304, -0.00000790, 0.00000216, -0.00002315, -0.00001859, -0.00011153, 0.00009957, 0.00011005, 0.00014172, -0.00000758, -0.00009189, 0.00000408, -0.00002824, 0.00000302, -0.00001363, 0.00000385, -0.00000336, 0.00000641, -0.00001809, 0.00000287, -0.00000886, 0.00000705, -0.00000537, 0.00000203, -0.00000027, 0.00000336, -0.00000197, 0.00000410, 0.00000505, -0.00000060, -0.00000350, -0.00000297, -0.00000193, 0.00001807, -0.00001093, 0.00001104, -0.00001439, 0.00000510, -0.00000350, -0.00000432, 0.00000300, -0.00000331, -0.00000576, 0.00000044, -0.00001396, 0.00001026, -0.00000748, 0.00000678, -0.00000638, -0.00001039, 0.00001631, -0.00000833, 0.00000425, 0.00000433, 0.00000253, 0.00000106, -0.00000247, -0.00000017, 0.00001178, -0.00000287, -0.00001168, -0.00000616, 0.00001218, 0.00000006, -0.00000942, -0.00000461, 0.00000181, -0.00000199, -0.00000054, 0.00000040, -0.00000155, -0.00000588, 0.00000101, 0.00000211, 0.00000069, -0.00000638, 0.00000464, 0.00000923, -0.00000738, -0.00000340, 0.00001003, -0.00000359, 0.00000385, -0.00000418, 0.00001428, -0.00000597, -0.00000039, -0.00000413, 0.00000037, 0.00000065, -0.00001002, -0.00000427, 0.00000342, 0.00000029, 0.00000101, -0.00001112, 0.00000246, -0.00002690, -0.00002062, 0.00000788, 0.00005211, 0.00001307, 0.00003927, 0.00003346, -0.00001030, -0.00003054, -0.00000062, -0.00000744, },
		{ -0.00001203, -0.00000040, -0.00000770, -0.00000228, -0.00000405, 0.00000222, 0.00000231, 0.00000080, -0.00000136, -0.00000856, -0.00000257, 0.00000093, -0.00000352, -0.00000614, -0.00000101, -0.00000338, 0.00000491, 0.00000054, -0.00000441, -0.00000242, -0.00000799, -0.00000749, 0.00000027, -0.00000992, -0.00001303, -0.00004880, -0.00008655, 0.00013737, 0.00010967, 0.00014247, -0.00008052, -0.00004974, -0.00000888, -0.00001675, -0.00000379, -0.00000631, -0.00000162, 0.00000408, -0.00000383, -0.00001154, -0.00000041, -0.00000502, 0.00000514, -0.00000420, 0.00000103, 0.00000159, 0.00000093, 0.00000004, 0.00000396, 0.00000593, -0.00000523, 0.00000081, -0.00000829, 0.00000816, 0.00000933, -0.00000668, 0.00000802, -0.00001497, 0.00000848, -0.00000753, -0.00000113, 0.00000252, -0.00000637, -0.00000174, -0.00000528, -0.00000869, 0.00001044, -0.00001076, 0.00001214, -0.00001632, 0.00000215, 0.00000944, -0.00000771, 0.00000653, 0.00000358, 0.00000201, 0.00000068, -0.00000416, 0.00000473, 0.00000933, -0.00000681, -0.00001199, -0.00000027, 0.00001174, -0.00000471, -0.00000850, -0.00000255, 0.00000124, -0.00000156, -0.00000074, 0.00000104, -0.00000388, -0.00000376, 0.00000081, 0.00000381, -0.00000242, -0.00000500, 0.00000691, 0.00000759, -0.00001157, 0.00000426, 0.00000487, -0.00000107, 0.00000072, 0.00000071, 0.00001219, -0.00000902, 0.00000164, -0.00000539, 0.00000268, -0.00000243, -0.00001048, -0.00000021, 0.00000120, 0.00000297, -0.00000487, -0.00000570, -0.00000312, -0.00003041, -0.00001256, 0.00002159, 0.00004637, 0.00000809, 0.00004750, 0.00001827, -0.00001509, -0.00002692, 0.00000413, },
		{ 0.00000419, -0.00001299, 0.00000136, -0.00000952, -0.00000002, -0.00000506, 0.00000458, 0.00000098, 0.00000111, -0.00000368, -0.00000818, -0.00000020, -0.00000069, -0.00000387, -0.00000600, 0.00000037, -0.00000477, 0.00000945, -0.00000637, 0.00000095, -0.00000858, -0.00000316, -0.00001247, 0.00000875, -0.00002428, -0.00000052, -0.00008792, -0.00003375, 0.00015152, 0.00012723, 0.00010312, -0.00012226, -0.00000862, -0.00002755, -0.00000122, -0.00001354, 0.00000270, -0.00000783, 0.00001116, -0.00001501, -0.00000189, -0.00000692, 0.00000173, 0.00000061, -0.00000121, -0.00000114, 0.00000414, -0.00000218, 0.00000306, 0.00000348, 0.00000524, -0.00000772, 0.00000343, -0.00001145, 0.00001722, -0.00000145, 0.00000142, 0.00000028, -0.00001037, 0.00000770, -0.00000913, 0.00000156, 0.00000091, -0.00000806, 0.00000174, -0.00001219, 0.00000127, 0.00000332, -0.00000640, 0.00001002, -0.00001978, 0.00001211, 0.00000059, -0.00000366, 0.00000681, 0.00000320, 0.00000143, -0.00000033, -0.00000422, 0.00000921, 0.00000464, -0.00000938, -0.00001082, 0.00000572, 0.00000823, -0.00000769, -0.00000680, -0.00000136, 0.00000104, -0.00000153, -0.00000046, 0.00000075, -0.00000530, -0.00000177, 0.00000078, 0.00000507, -0.00000610, -0.00000202, 0.00000842, 0.00000428, -0.00001287, 0.00001077, -0.00000162, 0.00000337, -0.00000405, 0.00000864, 0.00000482, -0.00000721, 0.00000058, -0.00000397, 0.00000259, -0.00000468, -0.00001003, 0.00000336, -0.00000138, 0.00000562, -0.00001144, 0.00000264, -0.00001519, -0.00002568, -0.00000618, 0.00003544, 0.00003261, 0.00001325, 0.00004710, 0.00000591, -0.00001864, -0.00002041, },
		{ -0.00001150, -0.00000012, -0.00000986, -0.00000063, -0.00000800, -0.00000041, -0.00000389, 0.00000531, 0.00000056, 0.00000025, -0.00000530, -0.00000708, 0.00000141, -0.00000249, -0.00000405, -0.00000491, -0.00000025, -0.00000298, 0.00000999, -0.00001026, 0.00000410, -0.00001272, -0.00000063, -0.00001337, 0.00001110, -0.00003257, 0.00000307, -0.00012037, 0.00003348, 0.00015066, 0.00015145, 0.00002841, -0.00011908, 0.00001106, -0.00003673, 0.00000784, -0.00001837, 0.00000710, -0.00000926, 0.00001225, -0.00002085, 0.00000491, -0.00001179, 0.00000756, -0.00000357, 0.00000114, -0.00000226, 0.00000528, -0.00000392, 0.00000546, 0.00000353, 0.00000227, -0.00000660, 0.00000218, -0.00000947, 0.00002088, -0.00000921, 0.00000869, -0.00000837, -0.00000316, 0.00000394, -0.00000861, 0.00000377, -0.00000243, -0.00000655, 0.00000127, -0.00001529, 0.00001022, -0.00000648, 0.00000297, 0.00000099, -0.00001487, 0.00001537, -0.00000589, 0.00000115, 0.00000610, 0.00000279, 0.00000101, -0.00000203, -0.00000193, 0.00001142, -0.00000068, -0.00001082, -0.00000770, 0.00000998, 0.00000304, -0.00000860, -0.00000506, -0.00000083, 0.00000138, -0.00000222, 0.00000070, -0.00000105, -0.00000478, -0.00000129, 0.00000240, 0.00000402, -0.00000829, 0.00000107, 0.00000974, -0.00000122, -0.00000944, 0.00001269, -0.00000545, 0.00000601, -0.00000658, 0.00001523, -0.00000419, -0.00000257, -0.00000235, -0.00000093, 0.00000068, -0.00000619, -0.00000812, 0.00000482, -0.00000205, 0.00000520, -0.00001427, 0.00000813, -0.00002820, -0.00001522, -0.00000107, 0.00004643, 0.00001548, 0.00002618, 0.00003719, -0.00000078, -0.00002245, },
		{ -0.00002581, -0.00000242, -0.00000640, -0.00000455, -0.00000492, -0.00000408, -0.00000294, -0.00000076, 0.00000447, 0.00000080, -0.00000152, -0.00000603, -0.00000544, 0.00000188, -0.00000387, -0.00000419, -0.00000324, -0.00000235, 0.00000191, 0.00000548, -0.00000893, 0.00000237, -0.00001208, -0.00000297, -0.00000743, 0.00000313, -0.00002861, -0.00001202, -0.00012936, 0.00009689, 0.00015026, 0.00015953, -0.00005697, -0.00007925, 0.00000415, -0.00002963, 0.00000502, -0.00001417, 0.00000385, -0.00000364, 0.00000498, -0.00001793, 0.00000481, -0.00001125, 0.00000924, -0.00000495, 0.00000106, -0.00000085, 0.00000365, -0.00000292, 0.00000614, 0.00000422, -0.00000234, -0.00000240, -0.00000272, -0.00000197, 0.00001730, -0.00001063, 0.00001120, -0.00001401, 0.00000345, -0.00000081, -0.00000662, 0.00000514, -0.00000682, -0.00000191, -0.00000436, -0.00001119, 0.00001247, -0.00001240, 0.00001122, -0.00001000, -0.00000437, 0.00001172, -0.00000784, 0.00000470, 0.00000525, 0.00000219, 0.00000056, -0.00000372, 0.00000220, 0.00001059, -0.00000539, -0.00001115, -0.00000293, 0.00001130, -0.00000217, -0.00000774, -0.00000398, -0.00000024, 0.00000161, -0.00000304, 0.00000211, -0.00000367, -0.00000270, -0.00000221, 0.00000575, 0.00000005, -0.00000774, 0.00000330, 0.00001074, -0.00000796, -0.00000164, 0.00000877, -0.00000425, 0.00000426, -0.00000373, 0.00001616, -0.00001018, 0.00000132, -0.00000455, 0.00000149, -0.00000142, -0.00000766, -0.00000437, 0.00000330, 0.00000042, -0.00000005, -0.00001007, 0.00000514, -0.00003463, -0.00000501, 0.00000597, 0.00004986, 0.00000235, 0.00004009, 0.00002179, -0.00000275, },
		{ -0.00000369, -0.00002637, 0.00000369, -0.00001118, -0.00000026, -0.00000859, -0.00000029, -0.00000562, 0.00000288, 0.00000303, 0.00000082, -0.00000329, -0.00000623, -0.00000339, 0.00000107, -0.00000447, -0.00000427, -0.00000183, -0.00000423, 0.00000765, -0.00000219, -0.00000341, -0.00000371, -0.00000720, -0.00000906, 0.00000336, -0.00001332, -0.00001357, -0.00004699, -0.00010584, 0.00014350, 0.00016077, 0.00013100, -0.00012023, -0.00002782, -0.00001813, -0.00001147, -0.00000634, -0.00000368, -0.00000428, 0.00000587, -0.00000753, -0.00000816, -0.00000163, -0.00000514, 0.00000639, -0.00000341, -0.00000111, 0.00000249, -0.00000012, 0.00000047, 0.00000531, 0.00000458, -0.00000648, 0.00000230, -0.00000846, 0.00000818, 0.00000834, -0.00000597, 0.00000796, -0.00001476, 0.00000733, -0.00000511, -0.00000350, 0.00000461, -0.00001002, 0.00000304, -0.00001258, -0.00000086, 0.00000627, -0.00001025, 0.00001280, -0.00001673, 0.00000613, 0.00000439, -0.00000582, 0.00000647, 0.00000450, 0.00000156, -0.00000036, -0.00000425, 0.00000664, 0.00000717, -0.00000891, -0.00001001, 0.00000231, 0.00000962, -0.00000611, -0.00000582, -0.00000388, 0.00000116, 0.00000077, -0.00000283, 0.00000237, -0.00000542, -0.00000083, -0.00000283, 0.00000901, -0.00000545, -0.00000494, 0.00000495, 0.00001016, -0.00001328, 0.00000742, 0.00000116, 0.00000095, -0.00000119, 0.00000425, 0.00001012, -0.00001049, 0.00000196, -0.00000416, 0.00000167, -0.00000249, -0.00000921, 0.00000042, -0.00000017, 0.00000474, -0.00000846, 0.00000009, -0.00000754, -0.00002996, -0.00000019, 0.00001768, 0.00004233, -0.00000022, 0.00004708, 0.00000759, },
		{ -0.00000004, -0.00000708, -0.00002229, 0.00000468, -0.00001199, 0.00000055, -0.00000916, 0.00000102, -0.00000637, 0.00000540, 0.00000198, -0.00000013, -0.00000423, -0.00000629, -0.00000117, -0.00000074, -0.00000429, -0.00000409, -0.00000155, -0.00000390, 0.00001089, -0.00000903, 0.00000247, -0.00001050, -0.00000180, -0.00001426, 0.00001253, -0.00003026, 0.00000300, -0.00009178, -0.00005334, 0.00017226, 0.00017822, 0.00006219, -0.00013817, 0.00000808, -0.00003720, 0.00000535, -0.00001706, 0.00000570, -0.00001071, 0.00001264, -0.00001838, 0.00000235, -0.00000975, 0.00000313, 0.00000142, -0.00000100, -0.00000317, 0.00000544, -0.00000376, 0.00000442, 0.00000417, 0.00000327, -0.00000781, 0.00000437, -0.00001121, 0.00001632, -0.00000152, 0.00000142, 0.00000100, -0.00001127, 0.00000813, -0.00000835, 0.00000068, 0.00000109, -0.00000937, 0.00000417, -0.00001798, 0.00001021, -0.00000468, -0.00000114, 0.00000637, -0.00001562, 0.00001170, -0.00000236, -0.00000199, 0.00000698, 0.00000367, 0.00000106, -0.00000186, -0.00000282, 0.00000970, 0.00000231, -0.00001105, -0.00000719, 0.00000647, 0.00000585, -0.00000809, -0.00000379, -0.00000441, 0.00000341, -0.00000153, -0.00000090, 0.00000048, -0.00000478, -0.00000101, -0.00000100, 0.00000953, -0.00000982, -0.00000165, 0.00000701, 0.00000645, -0.00001399, 0.00001324, -0.00000560, 0.00000621, -0.00000646, 0.00001337, -0.00000009, -0.00000607, -0.00000039, -0.00000154, -0.00000016, -0.00000278, -0.00000985, 0.00000420, -0.00000291, 0.00000732, -0.00001507, 0.00001026, -0.00002468, -0.00001612, -0.00000110, 0.00003284, 0.00002523, 0.00000947, 0.00004258, },
		{ 0.00002817, -0.00000012, -0.00001364, -0.00001405, 0.00000090, -0.00000889, -0.00000220, -0.00000625, -0.00000083, -0.00000432, 0.00000602, 0.00000160, -0.00000199, -0.00000420, -0.00000622, 0.00000068, -0.00000284, -0.00000366, -0.00000364, -0.00000257, -0.00000043, 0.00000921, -0.00001133, 0.00000455, -0.00001374, -0.00000020, -0.00001353, 0.00001289, -0.00003791, 0.00000777, -0.00012903, 0.00001471, 0.00019212, 0.00018441, -0.00002817, -0.00010983, 0.00001441, -0.00003973, 0.00001061, -0.00001920, 0.00000763, -0.00001003, 0.00001128, -0.00002141, 0.00000736, -0.00001404, 0.00000912, -0.00000248, -0.00000007, -0.00000295, 0.00000573, -0.00000493, 0.00000698, 0.00000366, -0.00000011, -0.00000542, 0.00000223, -0.00000857, 0.00001880, -0.00000783, 0.00000735, -0.00000626, -0.00000568, 0.00000650, -0.00001015, 0.00000515, -0.00000506, -0.00000407, -0.00000107, -0.00001574, 0.00001522, -0.00001333, 0.00000913, -0.00000427, -0.00000786, 0.00001080, -0.00000575, 0.00000163, 0.00000688, 0.00000269, 0.00000055, -0.00000342, 0.00000036, 0.00001039, -0.00000288, -0.00001154, -0.00000311, 0.00000853, 0.00000121, -0.00000792, -0.00000273, -0.00000445, 0.00000547, -0.00000442, 0.00000208, -0.00000306, -0.00000186, -0.00000352, 0.00000396, 0.00000575, -0.00001093, 0.00000060, 0.00000990, -0.00000062, -0.00000863, 0.00001259, -0.00000738, 0.00000741, -0.00000703, 0.00001817, -0.00000921, -0.00000057, -0.00000326, 0.00000112, -0.00000221, -0.00000350, -0.00000827, 0.00000494, -0.00000235, 0.00000479, -0.00001462, 0.00001274, -0.00003703, -0.00000069, -0.00000275, 0.00004558, 0.00000553, 0.00002651, },
		{ 0.00004160, 0.00001068, 0.00000368, -0.00002085, -0.00000461, -0.00000483, -0.00000428, -0.00000628, -0.00000181, -0.00000440, -0.00000035, 0.00000516, 0.00000130, -0.00000391, -0.00000374, -0.00000573, 0.00000155, -0.00000446, -0.00000294, -0.00000323, -0.00000391, 0.00000501, 0.00000295, -0.00000819, 0.00000120, -0.00001162, -0.00000393, -0.00000546, 0.00000169, -0.00003121, -0.00000847, -0.00014273, 0.00008357, 0.00021232, 0.00015814, -0.00010706, -0.00005636, -0.00000371, -0.00002542, 0.00000309, -0.00001171, 0.00000158, -0.00000224, 0.00000164, -0.00001531, 0.00000447, -0.00001179, 0.00001034, -0.00000370, -0.00000132, -0.00000006, 0.00000289, -0.00000290, 0.00000745, 0.00000357, -0.00000430, -0.00000076, -0.00000292, -0.00000127, 0.00001503, -0.00000867, 0.00000927, -0.00001123, 0.00000006, 0.00000305, -0.00000974, 0.00000793, -0.00001120, 0.00000334, -0.00001104, -0.00000528, 0.00001060, -0.00001384, 0.00001389, -0.00001273, 0.00000170, 0.00000557, -0.00000555, 0.00000421, 0.00000640, 0.00000177, -0.00000036, -0.00000408, 0.00000418, 0.00000871, -0.00000764, -0.00000999, 0.00000105, 0.00000838, -0.00000320, -0.00000590, -0.00000343, -0.00000271, 0.00000568, -0.00000604, 0.00000416, -0.00000618, 0.00000131, -0.00000642, 0.00001028, -0.00000140, -0.00000861, 0.00000177, 0.00001243, -0.00000883, 0.00000094, 0.00000587, -0.00000334, 0.00000330, -0.00000107, 0.00001528, -0.00001285, 0.00000231, -0.00000402, 0.00000168, -0.00000280, -0.00000546, -0.00000422, 0.00000238, 0.00000169, -0.00000306, -0.00000569, 0.00000329, -0.00003690, 0.00000792, 0.00000144, 0.00004853, -0.00000724, },
		{ -0.00000635, 0.00004622, -0.00000187, 0.00000580, -0.00002479, 0.00000222, -0.00000906, -0.00000128, -0.00000864, 0.00000128, -0.00000718, 0.00000369, 0.00000387, 0.00000028, -0.00000484, -0.00000361, -0.00000452, 0.00000111, -0.00000510, -0.00000236, -0.00000338, -0.00000388, 0.00000952, -0.00000477, -0.00000214, -0.00000548, -0.00000618, -0.00001021, 0.00000593, -0.00001672, -0.00001391, -0.00004484, -0.00012580, 0.00014565, 0.00023167, 0.00009092, -0.00014493, -0.00000791, -0.00002804, -0.00000520, -0.00000958, -0.00000061, -0.00000703, 0.00000719, -0.00001085, -0.00000439, -0.00000358, -0.00000450, 0.00000750, -0.00000302, -0.00000331, 0.00000367, -0.00000133, 0.00000110, 0.00000656, 0.00000273, -0.00000697, 0.00000319, -0.00000778, 0.00000729, 0.00000760, -0.00000524, 0.00000726, -0.00001312, 0.00000498, -0.00000191, -0.00000620, 0.00000660, -0.00001341, 0.00000780, -0.00002006, 0.00000845, -0.00000126, -0.00000555, 0.00001004, -0.00001423, 0.00000784, -0.00000014, -0.00000333, 0.00000593, 0.00000543, 0.00000112, -0.00000183, -0.00000307, 0.00000726, 0.00000522, -0.00001128, -0.00000643, 0.00000402, 0.00000659, -0.00000652, -0.00000292, -0.00000573, 0.00000127, 0.00000295, -0.00000481, 0.00000340, -0.00000648, 0.00000197, -0.00000660, 0.00001440, -0.00000881, -0.00000474, 0.00000321, 0.00001223, -0.00001425, 0.00001006, -0.00000286, 0.00000344, -0.00000331, 0.00000882, 0.00000574, -0.00001017, 0.00000141, -0.00000205, -0.00000026, -0.00000181, -0.00000798, 0.00000071, -0.00000135, 0.00000623, -0.00001215, 0.00000726, -0.00001516, -0.00002343, 0.00000627, 0.00001431, 0.00003801, },
		{ 0.00001750, 0.00000778, 0.00003748, -0.00000492, 0.00000254, -0.00002301, 0.00000418, -0.00000980, -0.00000156, -0.00000757, 0.00000112, -0.00000722, 0.00000622, 0.00000292, -0.00000176, -0.00000432, -0.00000419, -0.00000267, -0.00000046, -0.00000469, -0.00000210, -0.00000426, -0.00000127, 0.00001021, -0.00000986, 0.00000249, -0.00001121, -0.00000181, -0.00001377, 0.00001322, -0.00003295, 0.00000283, -0.00009030, -0.00008150, 0.00020199, 0.00023541, -0.00000355, -0.00013220, 0.00001377, -0.00004057, 0.00000782, -0.00001744, 0.00000612, -0.00001124, 0.00001140, -0.00001893, 0.00000455, -0.00001120, 0.00000372, 0.00000324, -0.00000244, -0.00000380, 0.00000567, -0.00000433, 0.00000497, 0.00000546, 0.00000024, -0.00000642, 0.00000388, -0.00000917, 0.00001319, 0.00000035, -0.00000042, 0.00000305, -0.00001244, 0.00000857, -0.00000768, 0.00000024, 0.00000029, -0.00000918, 0.00000488, -0.00002157, 0.00001753, -0.00001303, 0.00000635, 0.00000009, -0.00000856, 0.00000815, -0.00000338, -0.00000083, 0.00000714, 0.00000401, 0.00000059, -0.00000341, -0.00000052, 0.00000873, 0.00000054, -0.00001293, -0.00000174, 0.00000521, 0.00000380, -0.00000798, -0.00000050, -0.00000820, 0.00000635, -0.00000212, -0.00000081, -0.00000049, -0.00000318, -0.00000122, -0.00000199, 0.00001324, -0.00001308, -0.00000167, 0.00000628, 0.00000755, -0.00001353, 0.00001386, -0.00000830, 0.00000800, -0.00000732, 0.00001698, -0.00000552, -0.00000412, -0.00000150, 0.00000093, -0.00000305, -0.00000069, -0.00000924, 0.00000393, -0.00000306, 0.00000704, -0.00001619, 0.00001574, -0.00003328, -0.00000368, -0.00000159, 0.00003193, },
		{ 0.00004528, -0.00000337, 0.00002745, 0.00001969, 0.00000032, -0.00000578, -0.00001613, 0.00000177, -0.00000760, -0.00000433, -0.00000377, -0.00000202, -0.00000426, 0.00000675, 0.00000228, -0.00000422, -0.00000278, -0.00000525, -0.00000065, -0.00000258, -0.00000346, -0.00000245, -0.00000516, 0.00000343, 0.00000626, -0.00001001, 0.00000263, -0.00001268, -0.00000177, -0.00001087, 0.00001066, -0.00003845, 0.00000684, -0.00012975, -0.00001798, 0.00025538, 0.00020372, -0.00009360, -0.00008412, 0.00000623, -0.00003476, 0.00000765, -0.00001563, 0.00000467, -0.00000792, 0.00000720, -0.00001849, 0.00000676, -0.00001388, 0.00000915, -0.00000016, -0.00000310, -0.00000181, 0.00000463, -0.00000450, 0.00000733, 0.00000447, -0.00000332, -0.00000300, 0.00000108, -0.00000617, 0.00001451, -0.00000415, 0.00000344, -0.00000169, -0.00000969, 0.00000994, -0.00001236, 0.00000722, -0.00000877, 0.00000011, -0.00000565, -0.00001272, 0.00001612, -0.00001713, 0.00001425, -0.00000955, 0.00000009, 0.00000407, -0.00000362, 0.00000123, 0.00000781, 0.00000251, -0.00000026, -0.00000415, 0.00000264, 0.00000847, -0.00000482, -0.00001169, 0.00000241, 0.00000502, 0.00000040, -0.00000703, -0.00000034, -0.00000857, 0.00000997, -0.00000706, 0.00000379, -0.00000548, 0.00000190, -0.00000673, 0.00000649, 0.00000640, -0.00001274, -0.00000038, 0.00001064, -0.00000088, -0.00000637, 0.00001036, -0.00000738, 0.00000705, -0.00000493, 0.00001828, -0.00001286, 0.00000094, -0.00000339, 0.00000232, -0.00000441, -0.00000115, -0.00000778, 0.00000380, -0.00000099, 0.00000188, -0.00001117, 0.00001259, -0.00004066, 0.00001158, -0.00000680, },
		{ -0.00000196, 0.00004601, -0.00001409, 0.00004195, 0.00000161, 0.00000821, -0.00001541, -0.00000736, -0.00000245, -0.00000479, -0.00000709, 0.00000027, -0.00000602, 0.00000032, 0.00000589, 0.00000129, -0.00000596, -0.00000124, -0.00000612, 0.00000088, -0.00000454, -0.00000180, -0.00000374, -0.00000466, 0.00000788, -0.00000029, -0.00000621, -0.00000141, -0.00000997, -0.00000559, -0.00000251, -0.00000167, -0.00003116, -0.00000806, -0.00015141, 0.00005909, 0.00030032, 0.00012718, -0.00014616, -0.00003064, -0.00001578, -0.00001757, -0.00000160, -0.00000716, -0.00000216, -0.00000000, -0.00000270, -0.00001099, 0.00000230, -0.00001071, 0.00001062, -0.00000227, -0.00000428, 0.00000153, 0.00000144, -0.00000219, 0.00000821, 0.00000286, -0.00000602, 0.00000092, -0.00000302, -0.00000060, 0.00001209, -0.00000571, 0.00000574, -0.00000639, -0.00000477, 0.00000772, -0.00001321, 0.00001073, -0.00001560, 0.00000902, -0.00001855, 0.00000296, 0.00000459, -0.00001068, 0.00001313, -0.00001294, 0.00000617, -0.00000074, -0.00000226, 0.00000312, 0.00000757, 0.00000137, -0.00000171, -0.00000330, 0.00000522, 0.00000672, -0.00001007, -0.00000735, 0.00000444, 0.00000435, -0.00000328, -0.00000380, -0.00000327, -0.00000511, 0.00000959, -0.00000890, 0.00000589, -0.00000822, 0.00000517, -0.00001070, 0.00001498, -0.00000334, -0.00000906, 0.00000024, 0.00001392, -0.00000949, 0.00000370, 0.00000195, -0.00000131, 0.00000142, 0.00000341, 0.00001146, -0.00001338, 0.00000228, -0.00000251, 0.00000084, -0.00000333, -0.00000352, -0.00000397, 0.00000109, 0.00000340, -0.00000707, 0.00000112, -0.00000282, -0.00003248, 0.00001481, },
		{ 0.00000659, 0.00001340, 0.00003212, -0.00000941, 0.00004359, -0.00000875, 0.00001248, -0.00002184, -0.00000029, -0.00000577, -0.00000356, -0.00000752, 0.00000197, -0.00000834, 0.00000458, 0.00000459, -0.00000059, -0.00000603, -0.00000072, -0.00000616, 0.00000139, -0.00000568, -0.00000029, -0.00000589, -0.00000163, 0.00000931, -0.00000580, -0.00000195, -0.00000646, -0.00000618, -0.00000943, 0.00000597, -0.00001770, -0.00001668, -0.00003948, -0.00014995, 0.00014882, 0.00031889, 0.00001864, -0.00014570, 0.00000175, -0.00003278, -0.00000204, -0.00001053, 0.00000055, -0.00000790, 0.00000623, -0.00001195, -0.00000196, -0.00000483, -0.00000413, 0.00000922, -0.00000404, -0.00000426, 0.00000393, -0.00000169, 0.00000106, 0.00000833, -0.00000017, -0.00000601, 0.00000275, -0.00000554, 0.00000457, 0.00000826, -0.00000563, 0.00000698, -0.00001100, 0.00000221, 0.00000129, -0.00000843, 0.00000756, -0.00001533, 0.00001098, -0.00002574, 0.00001696, -0.00001020, 0.00000235, 0.00000393, -0.00000850, 0.00000651, -0.00000312, -0.00000098, 0.00000523, 0.00000625, 0.00000062, -0.00000344, -0.00000093, 0.00000659, 0.00000356, -0.00001379, -0.00000107, 0.00000393, 0.00000365, -0.00000640, 0.00000033, -0.00000828, 0.00000189, 0.00000450, -0.00000618, 0.00000361, -0.00000636, 0.00000362, -0.00000932, 0.00001878, -0.00001159, -0.00000492, 0.00000224, 0.00001316, -0.00001375, 0.00001119, -0.00000607, 0.00000533, -0.00000443, 0.00001304, 0.00000005, -0.00000844, 0.00000035, 0.00000035, -0.00000250, -0.00000085, -0.00000638, 0.00000008, -0.00000145, 0.00000619, -0.00001400, 0.00001325, -0.00002321, -0.00001288, },
		{ 0.00000740, -0.00000488, 0.00003201, 0.00000988, 0.00000756, 0.00003182, -0.00000834, 0.00000985, -0.00002255, 0.00000307, -0.00000699, -0.00000445, -0.00000510, 0.00000037, -0.00000753, 0.00000709, 0.00000342, -0.00000327, -0.00000427, -0.00000172, -0.00000499, 0.00000058, -0.00000550, 0.00000019, -0.00000782, 0.00000342, 0.00000659, -0.00000765, -0.00000040, -0.00000926, -0.00000458, -0.00000945, 0.00000876, -0.00002965, -0.00000428, -0.00007965, -0.00012332, 0.00024918, 0.00028822, -0.00008837, -0.00010312, 0.00000456, -0.00003398, 0.00000318, -0.00001237, 0.00000216, -0.00000796, 0.00000618, -0.00001535, 0.00000329, -0.00001003, 0.00000260, 0.00000653, -0.00000604, -0.00000218, 0.00000393, -0.00000306, 0.00000398, 0.00000783, -0.00000413, -0.00000324, 0.00000165, -0.00000515, 0.00000753, 0.00000473, -0.00000481, 0.00000720, -0.00001462, 0.00000962, -0.00000766, 0.00000071, -0.00000194, -0.00000686, 0.00000290, -0.00002139, 0.00002089, -0.00001897, 0.00001367, -0.00000712, 0.00000021, 0.00000203, -0.00000240, -0.00000038, 0.00000737, 0.00000428, -0.00000018, -0.00000446, 0.00000199, 0.00000685, -0.00000113, -0.00001422, 0.00000471, 0.00000196, 0.00000257, -0.00000760, 0.00000292, -0.00001249, 0.00000967, -0.00000321, -0.00000025, -0.00000206, -0.00000052, -0.00000270, -0.00000157, 0.00001534, -0.00001507, -0.00000253, 0.00000657, 0.00000725, -0.00001113, 0.00001206, -0.00000895, 0.00000792, -0.00000558, 0.00001800, -0.00001003, -0.00000223, -0.00000208, 0.00000266, -0.00000527, 0.00000093, -0.00000772, 0.00000212, -0.00000129, 0.00000400, -0.00001336, 0.00001659, -0.00003736, },
		{ -0.00003711, 0.00001782, -0.00000945, 0.00004342, -0.00000963, 0.00002735, 0.00001291, 0.00000003, 0.00000132, -0.00001799, 0.00000294, -0.00000675, -0.00000616, -0.00000130, -0.00000348, -0.00000386, 0.00000755, 0.00000231, -0.00000594, -0.00000138, -0.00000402, -0.00000270, -0.00000149, -0.00000371, -0.00000131, -0.00000764, 0.00000806, 0.00000137, -0.00000617, -0.00000198, -0.00000912, -0.00000562, -0.00000522, 0.00000403, -0.00003337, -0.00000096, -0.00012094, -0.00006701, 0.00034704, 0.00019692, -0.00015580, -0.00004836, -0.00001114, -0.00002260, -0.00000081, -0.00000779, -0.00000159, -0.00000322, 0.00000043, -0.00001233, 0.00000309, -0.00001114, 0.00000761, 0.00000300, -0.00000739, 0.00000088, 0.00000208, -0.00000264, 0.00000647, 0.00000585, -0.00000701, 0.00000026, -0.00000083, -0.00000290, 0.00000873, 0.00000132, -0.00000271, 0.00000517, -0.00001506, 0.00001413, -0.00001498, 0.00000960, -0.00001300, 0.00000540, -0.00001195, -0.00000634, 0.00001212, -0.00001607, 0.00001611, -0.00001281, 0.00000710, -0.00000338, -0.00000023, 0.00000024, 0.00000872, 0.00000237, -0.00000153, -0.00000386, 0.00000422, 0.00000632, -0.00000703, -0.00001030, 0.00000745, 0.00000037, 0.00000037, -0.00000576, 0.00000166, -0.00001256, 0.00001422, -0.00000952, 0.00000522, -0.00000753, 0.00000566, -0.00001024, 0.00000960, 0.00000596, -0.00001356, -0.00000184, 0.00001175, -0.00000170, -0.00000307, 0.00000633, -0.00000561, 0.00000510, -0.00000030, 0.00001496, -0.00001422, 0.00000141, -0.00000245, 0.00000233, -0.00000552, 0.00000058, -0.00000666, 0.00000163, 0.00000158, -0.00000295, -0.00000426, 0.00000716, },
		{ -0.00001103, -0.00002258, 0.00001465, -0.00000189, 0.00004067, -0.00001697, 0.00003964, -0.00000414, 0.00001012, -0.00000925, -0.00001068, 0.00000083, -0.00000639, -0.00000683, 0.00000154, -0.00000729, 0.00000113, 0.00000650, 0.00000084, -0.00000751, 0.00000138, -0.00000674, 0.00000014, -0.00000432, -0.00000065, -0.00000504, -0.00000383, 0.00000929, -0.00000291, -0.00000422, -0.00000435, -0.00000820, -0.00000665, -0.00000030, -0.00000497, -0.00003050, -0.00000862, -0.00015804, 0.00002686, 0.00041366, 0.00005951, -0.00016269, -0.00001152, -0.00002593, -0.00001025, -0.00000611, -0.00000284, -0.00000541, 0.00000129, -0.00000620, -0.00000664, -0.00000040, -0.00000893, 0.00001081, -0.00000172, -0.00000648, 0.00000273, 0.00000026, -0.00000155, 0.00000901, 0.00000150, -0.00000672, 0.00000181, -0.00000219, -0.00000103, 0.00000978, -0.00000302, 0.00000178, -0.00000055, -0.00001013, 0.00001235, -0.00001612, 0.00001254, -0.00001872, 0.00001354, -0.00002501, 0.00001158, -0.00000411, -0.00000330, 0.00000839, -0.00000971, 0.00000778, -0.00000585, 0.00000114, 0.00000184, 0.00000857, 0.00000098, -0.00000334, -0.00000152, 0.00000515, 0.00000487, -0.00001270, -0.00000278, 0.00000607, 0.00000034, -0.00000295, -0.00000118, -0.00000389, -0.00000685, 0.00001264, -0.00001091, 0.00000659, -0.00000890, 0.00000773, -0.00001384, 0.00001873, -0.00000501, -0.00000953, -0.00000082, 0.00001458, -0.00000919, 0.00000568, -0.00000194, 0.00000085, -0.00000028, 0.00000854, 0.00000542, -0.00001189, 0.00000137, -0.00000039, -0.00000064, -0.00000326, -0.00000159, -0.00000413, 0.00000024, 0.00000434, -0.00001026, 0.00000811, },
		{ 0.00001536, -0.00002809, -0.00000213, 0.00000295, 0.00001469, 0.00002477, -0.00000959, 0.00003890, -0.00001256, 0.00001609, -0.00001767, -0.00000339, -0.00000186, -0.00000664, -0.00000545, 0.00000176, -0.00000895, 0.00000560, 0.00000477, -0.00000130, -0.00000712, 0.00000261, -0.00000862, 0.00000242, -0.00000674, 0.00000232, -0.00000965, 0.00000311, 0.00000583, -0.00000371, -0.00000456, -0.00000491, -0.00000895, -0.00000477, 0.00000099, -0.00001317, -0.00002552, -0.00002669, -0.00018390, 0.00016162, 0.00041318, -0.00008307, -0.00011801, -0.00000511, -0.00002710, -0.00000608, -0.00000602, -0.00000274, -0.00000481, 0.00000118, -0.00000899, -0.00000270, -0.00000380, -0.00000518, 0.00001233, -0.00000729, -0.00000281, 0.00000209, -0.00000017, -0.00000050, 0.00001123, -0.00000458, -0.00000306, 0.00000045, -0.00000129, -0.00000059, 0.00001124, -0.00000821, 0.00000823, -0.00000949, 0.00000001, 0.00000362, -0.00000936, 0.00000663, -0.00001467, 0.00001114, -0.00002760, 0.00002202, -0.00001769, 0.00001117, -0.00000410, -0.00000035, 0.00000229, -0.00000412, 0.00000077, 0.00000457, 0.00000695, -0.00000011, -0.00000472, 0.00000156, 0.00000509, 0.00000184, -0.00001572, 0.00000548, 0.00000181, 0.00000146, -0.00000606, 0.00000387, -0.00001144, 0.00000308, 0.00000523, -0.00000669, 0.00000281, -0.00000473, 0.00000349, -0.00001010, 0.00002109, -0.00001289, -0.00000604, 0.00000245, 0.00001246, -0.00001121, 0.00001002, -0.00000746, 0.00000561, -0.00000332, 0.00001531, -0.00000550, -0.00000616, -0.00000054, 0.00000225, -0.00000428, -0.00000023, -0.00000396, -0.00000198, 0.00000023, 0.00000356, -0.00001226, },
		{ -0.00000636, 0.00001163, -0.00003439, 0.00001312, -0.00000728, 0.00003107, 0.00000390, 0.00000735, 0.00002652, -0.00001070, 0.00001516, -0.00002137, 0.00000204, -0.00000449, -0.00000711, -0.00000255, -0.00000072, -0.00000750, 0.00000819, 0.00000304, -0.00000405, -0.00000450, 0.00000134, -0.00000836, 0.00000278, -0.00000714, 0.00000300, -0.00001213, 0.00001000, -0.00000021, -0.00000174, -0.00000706, -0.00000368, -0.00001104, -0.00000035, -0.00000212, -0.00001854, -0.00002072, -0.00005688, -0.00018314, 0.00032273, 0.00032280, -0.00018072, -0.00005544, -0.00001914, -0.00001668, -0.00000953, -0.00000101, -0.00000686, -0.00000044, -0.00000330, -0.00000722, -0.00000202, -0.00000558, -0.00000065, 0.00001125, -0.00001161, 0.00000178, -0.00000005, 0.00000031, 0.00000122, 0.00001132, -0.00000962, 0.00000153, -0.00000207, 0.00000041, -0.00000020, 0.00001138, -0.00001173, 0.00001356, -0.00001807, 0.00001152, -0.00000842, 0.00000209, -0.00000549, -0.00000245, -0.00000190, -0.00001679, 0.00001875, -0.00002004, 0.00001814, -0.00001298, 0.00000873, -0.00000530, 0.00000002, -0.00000049, 0.00000752, 0.00000460, -0.00000144, -0.00000458, 0.00000395, 0.00000476, -0.00000329, -0.00001385, 0.00001078, -0.00000249, 0.00000202, -0.00000678, 0.00000599, -0.00001667, 0.00001281, -0.00000434, 0.00000034, -0.00000365, 0.00000262, -0.00000512, 0.00000021, 0.00001553, -0.00001542, -0.00000437, 0.00000780, 0.00000568, -0.00000699, 0.00000787, -0.00000736, 0.00000578, -0.00000089, 0.00001546, -0.00001242, -0.00000118, -0.00000162, 0.00000310, -0.00000624, 0.00000168, -0.00000521, -0.00000116, 0.00000222, -0.00000162, },
		{ -0.00000810, 0.00000430, -0.00000203, -0.00002679, 0.00001637, -0.00000799, 0.00003808, -0.00001174, 0.00002459, 0.00000903, -0.00000166, 0.00000801, -0.00001990, 0.00000478, -0.00000693, -0.00000690, 0.00000035, -0.00000460, -0.00000336, 0.00000843, 0.00000160, -0.00000686, -0.00000021, -0.00000258, -0.00000527, 0.00000032, -0.00000438, -0.00000043, -0.00000936, 0.00001240, -0.00000438, -0.00000071, -0.00000855, -0.00000373, -0.00001091, 0.00000228, -0.00000573, -0.00002388, -0.00001451, -0.00010394, -0.00013112, 0.00046990, 0.00015324, -0.00020095, -0.00001253, -0.00003180, -0.00000699, -0.00001218, 0.00000233, -0.00000950, 0.00000253, -0.00000745, -0.00000435, -0.00000264, -0.00000653, 0.00000511, 0.00000607, -0.00001175, 0.00000409, -0.00000114, -0.00000002, 0.00000495, 0.00000718, -0.00001034, 0.00000350, -0.00000263, 0.00000010, 0.00000279, 0.00000729, -0.00000993, 0.00001326, -0.00002098, 0.00001831, -0.00001724, 0.00001136, -0.00001655, 0.00001044, -0.00001844, 0.00000197, 0.00000397, -0.00000977, 0.00001344, -0.00001251, 0.00001144, -0.00000993, 0.00000340, -0.00000088, 0.00000935, 0.00000236, -0.00000321, -0.00000249, 0.00000476, 0.00000432, -0.00000972, -0.00000670, 0.00001065, -0.00000422, 0.00000056, -0.00000375, 0.00000272, -0.00001568, 0.00001739, -0.00001104, 0.00000562, -0.00000829, 0.00000829, -0.00001294, 0.00001235, 0.00000501, -0.00001369, -0.00000336, 0.00001265, -0.00000235, 0.00000035, 0.00000142, -0.00000285, 0.00000242, 0.00000596, 0.00000860, -0.00001307, 0.00000072, -0.00000058, 0.00000127, -0.00000558, 0.00000180, -0.00000533, -0.00000082, 0.00000425, },
		{ 0.00000326, -0.00001066, 0.00001246, -0.00001794, -0.00001052, 0.00000905, 0.00000184, 0.00003228, -0.00001570, 0.00003415, -0.00000607, 0.00000924, -0.00000251, -0.00001424, 0.00000459, -0.00000882, -0.00000547, 0.00000179, -0.00000798, 0.00000208, 0.00000666, 0.00000035, -0.00000878, 0.00000425, -0.00000796, -0.00000009, -0.00000439, 0.00000094, -0.00000756, -0.00000063, 0.00000782, -0.00000313, -0.00000408, -0.00000576, -0.00000819, -0.00000501, -0.00000148, -0.00000501, -0.00003296, -0.00000587, -0.00016824, -0.00000518, 0.00054547, -0.00004599, -0.00014842, -0.00000775, -0.00002745, -0.00000847, -0.00000657, -0.00000183, -0.00000563, -0.00000061, -0.00000663, -0.00000441, -0.00000181, -0.00000781, 0.00001191, -0.00000312, -0.00000655, 0.00000218, 0.00000053, -0.00000203, 0.00001065, -0.00000123, -0.00000558, 0.00000108, 0.00000034, -0.00000352, 0.00000939, -0.00000205, -0.00000113, 0.00000482, -0.00001473, 0.00001579, -0.00001739, 0.00001218, -0.00001912, 0.00001514, -0.00002816, 0.00001785, -0.00001280, 0.00000635, 0.00000059, -0.00000328, 0.00000612, -0.00000892, 0.00000395, 0.00000070, 0.00000931, 0.00000048, -0.00000481, 0.00000076, 0.00000426, 0.00000294, -0.00001500, 0.00000334, 0.00000544, -0.00000280, -0.00000265, 0.00000205, -0.00000546, -0.00000759, 0.00001435, -0.00001161, 0.00000588, -0.00000767, 0.00000812, -0.00001475, 0.00002039, -0.00000550, -0.00001061, -0.00000087, 0.00001381, -0.00000721, 0.00000587, -0.00000463, 0.00000202, -0.00000046, 0.00001264, -0.00000146, -0.00000911, 0.00000016, 0.00000163, -0.00000209, -0.00000313, 0.00000078, -0.00000530, 0.00000069, },
		{ 0.00000381, -0.00000188, -0.00000613, 0.00001177, -0.00002676, 0.00000466, -0.00000103, 0.00001624, 0.00001743, -0.00000820, 0.00003292, -0.00001378, 0.00001715, -0.00001293, -0.00000622, 0.00000163, -0.00000949, -0.00000320, 0.00000112, -0.00000925, 0.00000699, 0.00000391, -0.00000121, -0.00000859, 0.00000670, -0.00001231, 0.00000479, -0.00000889, 0.00000588, -0.00001482, 0.00001059, -0.00000163, 0.00000269, -0.00001142, 0.00000077, -0.00001605, 0.00000545, -0.00001114, -0.00000033, -0.00004391, -0.00000215, -0.00023381, 0.00019481, 0.00050025, -0.00020446, -0.00006349, -0.00003073, -0.00000830, -0.00001981, 0.00000597, -0.00001212, 0.00000353, -0.00000902, -0.00000084, -0.00000787, 0.00000068, -0.00000854, 0.00001721, -0.00001304, 0.00000156, -0.00000249, 0.00000385, -0.00000412, 0.00001557, -0.00001058, 0.00000199, -0.00000380, 0.00000489, -0.00000814, 0.00001677, -0.00001343, 0.00001168, -0.00000941, -0.00000085, 0.00000448, -0.00000851, 0.00000342, -0.00001090, 0.00000749, -0.00002430, 0.00002144, -0.00002079, 0.00001793, -0.00001170, 0.00000842, -0.00000374, -0.00000343, 0.00000191, 0.00000392, 0.00000763, -0.00000129, -0.00000515, 0.00000362, 0.00000345, -0.00000056, -0.00001601, 0.00001189, -0.00000173, 0.00000002, -0.00000543, 0.00000729, -0.00001476, 0.00000450, 0.00000534, -0.00000656, 0.00000134, -0.00000190, 0.00000162, -0.00000866, 0.00002079, -0.00001219, -0.00000836, 0.00000397, 0.00001008, -0.00000658, 0.00000625, -0.00000654, 0.00000377, 0.00000073, 0.00001430, -0.00000943, -0.00000428, -0.00000062, 0.00000295, -0.00000490, -0.00000044, -0.00000050, -0.00000560, },
		{ -0.00000527, 0.00000654, -0.00000852, 0.00000351, 0.00000163, -0.00002390, 0.00001112, -0.00000565, 0.00002689, 0.00000133, 0.00000540, 0.00002286, -0.00001287, 0.00001914, -0.00002012, 0.00000181, -0.00000325, -0.00000849, -0.00000092, -0.00000141, -0.00000754, 0.00000968, 0.00000149, -0.00000365, -0.00000534, 0.00000506, -0.00001277, 0.00000623, -0.00000979, 0.00000629, -0.00001672, 0.00001770, -0.00000952, 0.00000746, -0.00001735, 0.00000551, -0.00002104, 0.00001315, -0.00001911, 0.00000030, -0.00004671, -0.00002155, -0.00026237, 0.00042920, 0.00032492, -0.00026533, 0.00000123, -0.00005246, 0.00000862, -0.00002862, 0.00001539, -0.00001990, 0.00001034, -0.00001594, 0.00000454, -0.00001084, 0.00000185, -0.00000573, 0.00001673, -0.00001822, 0.00000731, -0.00000571, 0.00000535, -0.00000291, 0.00001540, -0.00001552, 0.00000711, -0.00000649, 0.00000646, -0.00000879, 0.00001920, -0.00002023, 0.00002141, -0.00002233, 0.00001392, -0.00000952, 0.00000378, -0.00000951, 0.00000310, -0.00000853, -0.00000851, 0.00001105, -0.00001504, 0.00001775, -0.00001534, 0.00001492, -0.00001209, 0.00000285, -0.00000070, 0.00000735, 0.00000511, -0.00000324, -0.00000352, 0.00000485, 0.00000302, -0.00000632, -0.00001094, 0.00001495, -0.00000689, 0.00000157, -0.00000511, 0.00000804, -0.00001991, 0.00001496, -0.00000480, 0.00000029, -0.00000441, 0.00000521, -0.00000751, 0.00000264, 0.00001418, -0.00001431, -0.00000685, 0.00000949, 0.00000349, -0.00000188, 0.00000202, -0.00000409, 0.00000220, 0.00000617, 0.00000934, -0.00001209, -0.00000140, -0.00000002, 0.00000214, -0.00000583, 0.00000156, -0.00000203, },
		{ -0.00000446, -0.00000256, 0.00000561, -0.00001146, 0.00001170, -0.00001188, -0.00001242, 0.00000817, -0.00000173, 0.00002865, -0.00000915, 0.00001839, 0.00000868, -0.00000510, 0.00001441, -0.00002181, 0.00000722, -0.00000825, -0.00000617, 0.00000069, -0.00000494, -0.00000292, 0.00000918, 0.00000039, -0.00000704, 0.00000081, -0.00000130, -0.00000783, 0.00000233, -0.00000507, -0.00000026, -0.00000940, 0.00001494, -0.00000919, 0.00000454, -0.00001528, 0.00000172, -0.00001563, 0.00000911, -0.00001560, -0.00001338, -0.00002991, -0.00008367, -0.00020402, 0.00062065, 0.00006684, -0.00021694, 0.00001240, -0.00004796, 0.00000642, -0.00002222, 0.00001134, -0.00001626, 0.00000684, -0.00001400, 0.00000317, -0.00000851, -0.00000148, 0.00000274, 0.00000781, -0.00001465, 0.00000632, -0.00000370, 0.00000224, 0.00000372, 0.00000756, -0.00001227, 0.00000565, -0.00000329, 0.00000155, -0.00000171, 0.00001204, -0.00001647, 0.00002096, -0.00002610, 0.00002135, -0.00001802, 0.00001129, -0.00001791, 0.00001339, -0.00002292, 0.00000970, -0.00000602, 0.00000041, 0.00000646, -0.00000819, 0.00001213, -0.00001433, 0.00000638, -0.00000169, 0.00000955, 0.00000239, -0.00000498, -0.00000039, 0.00000437, 0.00000245, -0.00001255, -0.00000096, 0.00001120, -0.00000772, 0.00000040, -0.00000075, 0.00000241, -0.00001732, 0.00001883, -0.00001101, 0.00000446, -0.00000705, 0.00000876, -0.00001366, 0.00001359, 0.00000445, -0.00001375, -0.00000435, 0.00001261, -0.00000195, 0.00000276, -0.00000311, -0.00000030, 0.00000039, 0.00001221, 0.00000043, -0.00000994, -0.00000074, 0.00000165, -0.00000030, -0.00000504, 0.00000294, },
		{ 0.00000392, -0.00000782, 0.00000291, -0.00000053, -0.00000691, 0.00001182, -0.00002006, -0.00000058, 0.00000150, 0.00000742, 0.00002220, -0.00001147, 0.00002618, -0.00000471, 0.00000612, 0.00000414, -0.00001723, 0.00000789, -0.00001122, -0.00000364, 0.00000134, -0.00000823, 0.00000338, 0.00000585, 0.00000046, -0.00001022, 0.00000777, -0.00001032, 0.00000093, -0.00000553, 0.00000358, -0.00001173, 0.00000557, 0.00000228, 0.00000073, -0.00000779, -0.00000347, -0.00001208, 0.00000167, -0.00000895, 0.00000189, -0.00004297, 0.00000568, -0.00018957, -0.00002528, 0.00068243, -0.00018281, -0.00010106, -0.00002471, -0.00001537, -0.00001640, 0.00000039, -0.00000689, -0.00000057, -0.00000764, -0.00000205, -0.00000626, -0.00000016, -0.00000873, 0.00001478, -0.00000721, -0.00000348, -0.00000119, 0.00000327, -0.00000448, 0.00001377, -0.00000577, -0.00000208, -0.00000178, 0.00000491, -0.00000853, 0.00001170, -0.00000381, -0.00000173, 0.00000834, -0.00001737, 0.00001707, -0.00001620, 0.00000889, -0.00001580, 0.00001249, -0.00002614, 0.00001914, -0.00001824, 0.00001526, -0.00000810, 0.00000490, 0.00000186, -0.00000989, 0.00000591, -0.00000021, 0.00000985, -0.00000035, -0.00000561, 0.00000281, 0.00000317, 0.00000037, -0.00001594, 0.00000989, 0.00000289, -0.00000485, -0.00000243, 0.00000562, -0.00000774, -0.00000739, 0.00001470, -0.00001100, 0.00000386, -0.00000457, 0.00000610, -0.00001291, 0.00001925, -0.00000414, -0.00001269, 0.00000039, 0.00001131, -0.00000317, 0.00000365, -0.00000529, 0.00000135, 0.00000192, 0.00001408, -0.00000748, -0.00000611, -0.00000062, 0.00000276, -0.00000275, -0.00000347, },
		{ -0.00000162, 0.00000380, -0.00001034, 0.00000865, -0.00000929, 0.00000372, 0.00000244, -0.00001789, 0.00000441, -0.00000176, 0.00001550, 0.00001218, -0.00000721, 0.00002717, -0.00001359, 0.00001667, -0.00000857, -0.00000778, 0.00000350, -0.00001088, -0.00000189, 0.00000087, -0.00000973, 0.00000907, 0.00000155, 0.00000030, -0.00001090, 0.00001177, -0.00001745, 0.00000886, -0.00001259, 0.00001064, -0.00002125, 0.00002044, -0.00001296, 0.00001367, -0.00002310, 0.00001134, -0.00002820, 0.00002185, -0.00003126, 0.00002213, -0.00007374, 0.00003674, -0.00030415, 0.00025864, 0.00056322, -0.00033005, 0.00000984, -0.00007222, 0.00002265, -0.00004294, 0.00002541, -0.00002760, 0.00001699, -0.00002408, 0.00001239, -0.00001766, 0.00000894, -0.00001441, 0.00002362, -0.00002078, 0.00000851, -0.00000966, 0.00001025, -0.00000967, 0.00002107, -0.00001770, 0.00000865, -0.00000949, 0.00001222, -0.00001718, 0.00002413, -0.00002082, 0.00001713, -0.00001089, -0.00000022, 0.00000385, -0.00000610, -0.00000163, -0.00000446, 0.00000038, -0.00001583, 0.00001437, -0.00001751, 0.00001995, -0.00001628, 0.00001551, -0.00000985, -0.00000195, 0.00000277, 0.00000307, 0.00000844, -0.00000308, -0.00000436, 0.00000461, 0.00000230, -0.00000412, -0.00001362, 0.00001661, -0.00000550, -0.00000115, -0.00000409, 0.00000990, -0.00001746, 0.00000548, 0.00000541, -0.00000636, -0.00000003, 0.00000123, -0.00000125, -0.00000550, 0.00001809, -0.00000955, -0.00001164, 0.00000639, 0.00000649, -0.00000041, 0.00000032, -0.00000352, 0.00000005, 0.00000757, 0.00000951, -0.00001073, -0.00000352, 0.00000048, 0.00000211, -0.00000404, },
		{ -0.00000434, 0.00000093, 0.00000114, -0.00000918, 0.00001023, -0.00001463, 0.00001368, -0.00001078, -0.00000789, 0.00000159, 0.00000132, 0.00001827, 0.00000327, 0.00000050, 0.00002183, -0.00001561, 0.00002237, -0.00001919, 0.00000276, -0.00000378, -0.00000785, -0.00000098, -0.00000104, -0.00000795, 0.00001154, -0.00000117, -0.00000197, -0.00000682, 0.00000918, -0.00001780, 0.00001056, -0.00001319, 0.00000973, -0.00002074, 0.00002508, -0.00001970, 0.00001840, -0.00002981, 0.00001696, -0.00003301, 0.00002904, -0.00003979, 0.00002428, -0.00007941, 0.00002297, -0.00035761, 0.00056989, 0.00028363, -0.00032649, 0.00005435, -0.00008720, 0.00003623, -0.00004992, 0.00003359, -0.00003431, 0.00002195, -0.00002938, 0.00001779, -0.00002146, 0.00001098, -0.00001166, 0.00002178, -0.00002438, 0.00001298, -0.00001179, 0.00001093, -0.00000744, 0.00001910, -0.00002065, 0.00001233, -0.00001045, 0.00001156, -0.00001651, 0.00002641, -0.00002858, 0.00002924, -0.00002628, 0.00001588, -0.00001001, 0.00000466, -0.00001254, 0.00000806, -0.00001501, 0.00000132, -0.00000049, -0.00000459, 0.00001189, -0.00001302, 0.00001721, -0.00001678, 0.00000505, -0.00000052, 0.00000667, 0.00000578, -0.00000537, -0.00000146, 0.00000459, 0.00000172, -0.00001008, -0.00000529, 0.00001615, -0.00001007, 0.00000058, -0.00000218, 0.00000846, -0.00002146, 0.00001538, -0.00000396, -0.00000103, -0.00000352, 0.00000614, -0.00000869, 0.00000458, 0.00001218, -0.00001241, -0.00000929, 0.00001080, 0.00000164, 0.00000299, -0.00000424, -0.00000025, -0.00000155, 0.00001414, 0.00000058, -0.00000922, -0.00000274, 0.00000238, 0.00000013, },
		{ -0.00000163, -0.00000449, 0.00000447, -0.00000463, -0.00000280, 0.00000453, -0.00001117, 0.00001513, -0.00001865, 0.00000176, -0.00000356, 0.00000780, 0.00001618, -0.00000273, 0.00000923, 0.00001152, -0.00000993, 0.00002013, -0.00002321, 0.00000960, -0.00001014, -0.00000432, -0.00000017, -0.00000450, -0.00000240, 0.00000934, -0.00000076, -0.00000692, 0.00000209, -0.00000081, -0.00000952, 0.00000365, -0.00000505, -0.00000145, -0.00000683, 0.00001409, -0.00001093, 0.00000719, -0.00001966, 0.00000473, -0.00001703, 0.00001186, -0.00002136, -0.00000692, -0.00004104, -0.00006828, -0.00027442, 0.00078869, -0.00006053, -0.00019664, 0.00001752, -0.00005229, 0.00001180, -0.00002626, 0.00001546, -0.00001878, 0.00000701, -0.00001654, 0.00000764, -0.00001227, 0.00000223, 0.00000173, 0.00000708, -0.00001469, 0.00000612, -0.00000427, 0.00000294, 0.00000375, 0.00000613, -0.00001186, 0.00000578, -0.00000198, 0.00000047, -0.00000343, 0.00001401, -0.00002060, 0.00002649, -0.00002893, 0.00002204, -0.00001626, 0.00000830, -0.00001569, 0.00001252, -0.00002310, 0.00001391, -0.00001456, 0.00001161, -0.00000306, -0.00000077, 0.00000929, -0.00001611, 0.00000818, -0.00000198, 0.00000933, 0.00000225, -0.00000633, 0.00000178, 0.00000354, 0.00000023, -0.00001462, 0.00000602, 0.00000917, -0.00000967, -0.00000029, 0.00000315, 0.00000073, -0.00001728, 0.00001828, -0.00000926, 0.00000167, -0.00000364, 0.00000658, -0.00001166, 0.00001243, 0.00000510, -0.00001434, -0.00000428, 0.00001107, 0.00000018, 0.00000318, -0.00000611, 0.00000093, 0.00000036, 0.00001658, -0.00000776, -0.00000588, -0.00000224, 0.00000346, },
		{ 0.00000243, -0.00000211, -0.00000462, 0.00000786, -0.00001155, 0.00000665, -0.00000676, 0.00000107, 0.00000520, -0.00001515, 0.00000346, -0.00000397, 0.00001194, 0.00001279, -0.00000642, 0.00001742, -0.00000146, 0.00000198, 0.00000966, -0.00001855, 0.00000949, -0.00001229, -0.00000243, 0.00000104, -0.00000856, 0.00000542, 0.00000350, 0.00000191, -0.00001245, 0.00001234, -0.00001431, 0.00000389, -0.00000850, 0.00000783, -0.00001778, 0.00001491, -0.00000787, 0.00000973, -0.00001674, 0.00000414, -0.00002153, 0.00001508, -0.00002482, 0.00001855, -0.00006427, 0.00003092, -0.00022948, -0.00002021, 0.00080648, -0.00033706, -0.00002546, -0.00006262, 0.00001183, -0.00003577, 0.00001637, -0.00001940, 0.00001081, -0.00002057, 0.00000844, -0.00001329, 0.00000566, -0.00001255, 0.00001974, -0.00001404, 0.00000290, -0.00000775, 0.00000883, -0.00000922, 0.00001848, -0.00001207, 0.00000371, -0.00000667, 0.00001122, -0.00001572, 0.00001662, -0.00000849, 0.00000048, 0.00000925, -0.00001736, 0.00001575, -0.00001235, 0.00000263, -0.00000866, 0.00000531, -0.00001818, 0.00001377, -0.00001772, 0.00002022, -0.00001482, 0.00001254, -0.00000345, -0.00000941, 0.00000721, -0.00000103, 0.00001037, -0.00000176, -0.00000524, 0.00000388, 0.00000256, -0.00000341, -0.00001442, 0.00001528, -0.00000050, -0.00000620, -0.00000190, 0.00000887, -0.00001007, -0.00000675, 0.00001410, -0.00000953, 0.00000119, -0.00000037, 0.00000222, -0.00000861, 0.00001532, -0.00000084, -0.00001575, 0.00000277, 0.00000731, 0.00000268, -0.00000094, -0.00000372, -0.00000131, 0.00000721, 0.00001183, -0.00001121, -0.00000389, -0.00000036, },
		{ 0.00000270, -0.00000015, -0.00000183, -0.00000349, 0.00000834, -0.00001527, 0.00001345, -0.00001685, 0.00001517, -0.00001074, -0.00000252, -0.00000367, 0.00000289, 0.00001052, 0.00001081, -0.00000841, 0.00002317, -0.00001331, 0.00001574, -0.00000543, -0.00000733, 0.00000290, -0.00000976, -0.00000253, 0.00000187, -0.00001087, 0.00001200, -0.00000247, 0.00000353, -0.00001425, 0.00001775, -0.00002386, 0.00001457, -0.00001785, 0.00001647, -0.00002834, 0.00003152, -0.00002694, 0.00002821, -0.00003882, 0.00002619, -0.00004461, 0.00004332, -0.00005808, 0.00005302, -0.00011398, 0.00008914, -0.00039536, 0.00035965, 0.00058625, -0.00044153, 0.00008916, -0.00012206, 0.00006095, -0.00007216, 0.00004981, -0.00004718, 0.00003375, -0.00004217, 0.00002908, -0.00003089, 0.00002021, -0.00002220, 0.00003068, -0.00002927, 0.00001683, -0.00001837, 0.00001813, -0.00001636, 0.00002687, -0.00002488, 0.00001584, -0.00001552, 0.00001934, -0.00002615, 0.00003178, -0.00002897, 0.00002348, -0.00001324, 0.00000134, 0.00000242, -0.00000304, -0.00000727, 0.00000318, -0.00000857, -0.00000376, 0.00000176, -0.00000758, 0.00001578, -0.00001595, 0.00001891, -0.00001424, -0.00000080, 0.00000388, 0.00000176, 0.00000946, -0.00000542, -0.00000231, 0.00000424, 0.00000191, -0.00000890, -0.00000809, 0.00001831, -0.00000822, -0.00000278, -0.00000151, 0.00001094, -0.00001869, 0.00000524, 0.00000612, -0.00000677, -0.00000041, 0.00000349, -0.00000389, -0.00000176, 0.00001386, -0.00000564, -0.00001523, 0.00000893, 0.00000267, 0.00000618, -0.00000668, 0.00000067, -0.00000448, 0.00001607, 0.00000145, -0.00000916, -0.00000408, },
		{ -0.00000460, 0.00000486, -0.00000221, -0.00000236, 0.00000039, 0.00000352, -0.00001185, 0.00001184, -0.00001778, 0.00002109, -0.00002186, 0.00000971, -0.00001280, 0.00001267, 0.00000573, 0.00000942, -0.00000741, 0.00002358, -0.00001911, 0.00002497, -0.00001848, 0.00000458, -0.00000588, -0.00000519, -0.00000290, 0.00000059, -0.00000876, 0.00001345, -0.00000449, 0.00000065, -0.00000857, 0.00001305, -0.00002259, 0.00001492, -0.00001661, 0.00001258, -0.00002310, 0.00003035, -0.00002843, 0.00002855, -0.00004187, 0.00002808, -0.00004410, 0.00004380, -0.00005993, 0.00004841, -0.00011294, 0.00006927, -0.00045971, 0.00073876, 0.00019464, -0.00035183, 0.00009167, -0.00011406, 0.00005908, -0.00006791, 0.00004914, -0.00004643, 0.00003121, -0.00004051, 0.00002959, -0.00003137, 0.00001987, -0.00001701, 0.00002502, -0.00002842, 0.00001704, -0.00001667, 0.00001562, -0.00001114, 0.00002125, -0.00002377, 0.00001593, -0.00001277, 0.00001435, -0.00002163, 0.00003109, -0.00003477, 0.00003514, -0.00002842, 0.00001619, -0.00000875, 0.00000345, -0.00001295, 0.00001043, -0.00001884, 0.00000969, -0.00001271, 0.00000888, 0.00000175, -0.00000626, 0.00001517, -0.00001846, 0.00000589, 0.00000036, 0.00000542, 0.00000646, -0.00000741, 0.00000108, 0.00000354, 0.00000053, -0.00001380, 0.00000244, 0.00001416, -0.00001136, -0.00000129, 0.00000209, 0.00000701, -0.00002091, 0.00001366, -0.00000147, -0.00000385, -0.00000057, 0.00000467, -0.00000767, 0.00000491, 0.00001057, -0.00001054, -0.00001095, 0.00001094, 0.00000112, 0.00000638, -0.00000955, 0.00000283, -0.00000394, 0.00002106, -0.00000911, -0.00000475, },
		{ -0.00000212, -0.00000299, 0.00000409, -0.00000203, -0.00000443, 0.00000669, -0.00000607, -0.00000130, 0.00000071, -0.00000641, 0.00001276, -0.00001926, 0.00001139, -0.00001447, 0.00001750, 0.00000314, 0.00000541, -0.00000129, 0.00001622, -0.00001510, 0.00002431, -0.00002312, 0.00001076, -0.00001140, -0.00000236, -0.00000148, -0.00000377, -0.00000139, 0.00000829, -0.00000099, -0.00000721, 0.00000424, -0.00000186, -0.00000935, 0.00000325, -0.00000345, -0.00000459, -0.00000100, 0.00000863, -0.00000779, 0.00000500, -0.00001920, 0.00000273, -0.00001236, 0.00000717, -0.00001876, -0.00000987, -0.00004103, -0.00006741, -0.00032756, 0.00095688, -0.00021869, -0.00014017, -0.00000145, -0.00004020, 0.00000506, -0.00002082, 0.00001177, -0.00001467, 0.00000102, -0.00001305, 0.00000696, -0.00001196, 0.00000305, 0.00000305, 0.00000323, -0.00001106, 0.00000255, -0.00000191, 0.00000124, 0.00000567, 0.00000245, -0.00000865, 0.00000345, 0.00000155, -0.00000342, -0.00000180, 0.00001234, -0.00002115, 0.00002854, -0.00002841, 0.00001962, -0.00001142, 0.00000194, -0.00000926, 0.00000690, -0.00001751, 0.00001221, -0.00001831, 0.00002031, -0.00001222, 0.00000766, 0.00000417, -0.00001566, 0.00000886, -0.00000186, 0.00000889, 0.00000165, -0.00000672, 0.00000327, 0.00000296, -0.00000297, -0.00001474, 0.00001267, 0.00000548, -0.00001033, -0.00000122, 0.00000738, -0.00000181, -0.00001593, 0.00001607, -0.00000616, -0.00000221, 0.00000139, 0.00000208, -0.00000696, 0.00000860, 0.00000735, -0.00001573, -0.00000301, 0.00000791, 0.00000421, 0.00000120, -0.00000693, 0.00000021, 0.00000316, 0.00001756, -0.00001420, },
		{ -0.00001191, -0.00000305, 0.00000116, 0.00000051, -0.00000005, -0.00000649, 0.00001219, -0.00001582, 0.00001095, -0.00001406, 0.00001184, -0.00000640, -0.00000314, -0.00000030, -0.00000428, 0.00001263, 0.00000655, -0.00000292, 0.00000988, 0.00000193, -0.00000175, 0.00001297, -0.00001723, 0.00000824, -0.00001078, -0.00000308, 0.00000188, -0.00000963, 0.00000859, -0.00000080, 0.00000524, -0.00001589, 0.00001817, -0.00002011, 0.00000911, -0.00001371, 0.00001395, -0.00002548, 0.00002677, -0.00002208, 0.00002369, -0.00003113, 0.00001751, -0.00003688, 0.00003541, -0.00004939, 0.00004572, -0.00009822, 0.00007218, -0.00029272, 0.00002155, 0.00089807, -0.00048964, 0.00006703, -0.00011612, 0.00005135, -0.00006500, 0.00004038, -0.00003861, 0.00002776, -0.00003858, 0.00002419, -0.00002525, 0.00001567, -0.00001933, 0.00002645, -0.00002284, 0.00001187, -0.00001690, 0.00001675, -0.00001583, 0.00002428, -0.00001939, 0.00001105, -0.00001284, 0.00001827, -0.00002389, 0.00002308, -0.00001524, 0.00000500, 0.00000767, -0.00001484, 0.00001219, -0.00000648, -0.00000563, 0.00000124, -0.00000531, -0.00000514, 0.00000183, -0.00001002, 0.00001892, -0.00001706, 0.00001725, -0.00000788, -0.00000862, 0.00000838, -0.00000208, 0.00001104, -0.00000385, -0.00000349, 0.00000354, 0.00000286, -0.00000865, -0.00000969, 0.00001799, -0.00000334, -0.00000761, -0.00000045, 0.00001096, -0.00001156, -0.00000648, 0.00001329, -0.00000797, -0.00000112, 0.00000372, -0.00000230, -0.00000292, 0.00000938, 0.00000389, -0.00001926, 0.00000562, 0.00000261, 0.00000945, -0.00000714, -0.00000049, -0.00000532, 0.00001477, 0.00000584, },
		{ -0.00000844, -0.00000531, -0.00000562, 0.00000547, -0.00000324, 0.00000122, -0.00000550, 0.00001230, -0.00001926, 0.00001667, -0.00002232, 0.00002523, -0.00002435, 0.00001609, -0.00001777, 0.00001284, 0.00000093, 0.00001389, -0.00001263, 0.00002178, -0.00001378, 0.00001506, -0.00000383, -0.00000478, -0.00000035, -0.00000577, -0.00000561, 0.00000459, -0.00001292, 0.00001565, -0.00000788, 0.00000831, -0.00001840, 0.00002412, -0.00003085, 0.00002124, -0.00002411, 0.00002278, -0.00003500, 0.00004208, -0.00004138, 0.00004405, -0.00005637, 0.00004316, -0.00006288, 0.00006685, -0.00008805, 0.00008830, -0.00016012, 0.00014978, -0.00050212, 0.00049771, 0.00055851, -0.00052174, 0.00015975, -0.00016998, 0.00009915, -0.00010183, 0.00007475, -0.00006723, 0.00005058, -0.00006018, 0.00004639, -0.00004520, 0.00003270, -0.00003058, 0.00003701, -0.00003681, 0.00002475, -0.00002698, 0.00002603, -0.00002289, 0.00003171, -0.00003075, 0.00002217, -0.00002057, 0.00002474, -0.00003317, 0.00003774, -0.00003589, 0.00002899, -0.00001517, 0.00000274, 0.00000132, -0.00000068, -0.00001173, 0.00000990, -0.00001682, 0.00000902, -0.00001363, 0.00000720, 0.00000583, -0.00001021, 0.00001754, -0.00001562, -0.00000088, 0.00000564, -0.00000011, 0.00001061, -0.00000798, 0.00000061, 0.00000272, 0.00000207, -0.00001429, 0.00000021, 0.00001644, -0.00000898, -0.00000536, 0.00000258, 0.00000995, -0.00001786, 0.00000321, 0.00000798, -0.00000826, 0.00000090, 0.00000384, -0.00000505, 0.00000122, 0.00000935, -0.00000145, -0.00001821, 0.00001060, -0.00000025, 0.00001179, -0.00001332, 0.00000471, -0.00000831, 0.00002437, },
		{ 0.00002505, -0.00001773, 0.00000005, -0.00000613, 0.00000658, -0.00000406, -0.00000057, 0.00000043, 0.00000455, -0.00001263, 0.00001053, -0.00001648, 0.00002320, -0.00002794, 0.00002424, -0.00002772, 0.00002547, -0.00000858, 0.00001808, -0.00001712, 0.00002721, -0.00002276, 0.00002648, -0.00001732, 0.00000633, -0.00000853, -0.00000134, -0.00000611, 0.00000313, -0.00000961, 0.00001477, -0.00000761, 0.00000342, -0.00000985, 0.00001575, -0.00002599, 0.00001809, -0.00001898, 0.00001394, -0.00002269, 0.00003172, -0.00003329, 0.00003507, -0.00005043, 0.00003573, -0.00005087, 0.00005326, -0.00007430, 0.00006631, -0.00013940, 0.00010693, -0.00055447, 0.00092197, 0.00006238, -0.00033554, 0.00010436, -0.00012529, 0.00007076, -0.00007730, 0.00005771, -0.00005270, 0.00003501, -0.00004630, 0.00003699, -0.00003792, 0.00002641, -0.00002031, 0.00002529, -0.00002899, 0.00001803, -0.00001892, 0.00001811, -0.00001292, 0.00002095, -0.00002392, 0.00001696, -0.00001266, 0.00001396, -0.00002299, 0.00003182, -0.00003713, 0.00003744, -0.00002744, 0.00001384, -0.00000486, -0.00000070, -0.00000957, 0.00000867, -0.00001786, 0.00001357, -0.00002187, 0.00002170, -0.00000996, 0.00000319, 0.00000961, -0.00001719, 0.00000524, 0.00000189, 0.00000379, 0.00000683, -0.00000879, 0.00000333, 0.00000236, -0.00000122, -0.00001626, 0.00001080, 0.00000969, -0.00001083, -0.00000385, 0.00000726, 0.00000403, -0.00001844, 0.00000996, 0.00000245, -0.00000785, 0.00000415, 0.00000077, -0.00000409, 0.00000298, 0.00001014, -0.00000933, -0.00001132, 0.00000941, 0.00000251, 0.00000742, -0.00001284, 0.00000413, -0.00000370, },
		{ 0.00000872, 0.00001447, -0.00001761, 0.00000038, -0.00000248, 0.00000313, -0.00000110, -0.00000507, 0.00000949, -0.00000848, 0.00000143, -0.00000513, 0.00000220, 0.00000453, -0.00001246, 0.00001346, -0.00002062, 0.00002395, -0.00000748, 0.00001256, -0.00001090, 0.00002087, -0.00001917, 0.00002588, -0.00002051, 0.00000943, -0.00001062, -0.00000164, -0.00000212, -0.00000353, 0.00000067, 0.00000539, 0.00000048, -0.00000865, 0.00000781, -0.00000505, -0.00000657, 0.00000026, 0.00000041, -0.00000997, 0.00000814, -0.00000174, 0.00000106, -0.00000333, -0.00001228, -0.00000590, 0.00000003, -0.00000702, -0.00000491, -0.00002619, -0.00002441, -0.00008969, -0.00034820, 0.00110409, -0.00039019, -0.00005533, -0.00004280, -0.00001143, -0.00001473, -0.00000480, -0.00000074, -0.00000320, -0.00001163, -0.00000290, 0.00000024, -0.00000657, 0.00000010, 0.00000702, -0.00000362, -0.00000388, -0.00000444, 0.00000351, -0.00000300, 0.00000947, -0.00000325, -0.00000288, -0.00000109, 0.00000682, -0.00000949, 0.00000273, 0.00000720, -0.00001795, 0.00002668, -0.00002425, 0.00001412, -0.00000380, -0.00000724, 0.00000084, -0.00000307, -0.00000613, 0.00000368, -0.00001509, 0.00002337, -0.00001798, 0.00001445, -0.00000127, -0.00001404, 0.00000892, -0.00000168, 0.00000852, 0.00000033, -0.00000574, 0.00000345, 0.00000321, -0.00000759, -0.00001189, 0.00001728, 0.00000158, -0.00001049, -0.00000169, 0.00001102, -0.00000429, -0.00001406, 0.00001293, -0.00000251, -0.00000615, 0.00000683, -0.00000366, -0.00000043, 0.00000263, 0.00001096, -0.00001768, -0.00000091, 0.00000363, 0.00000965, -0.00000287, -0.00000566, -0.00000230, },
		{ -0.00000936, 0.00002304, -0.00000289, -0.00000971, -0.00000358, 0.00000368, -0.00000265, 0.00000309, -0.00000867, 0.00001622, -0.00001961, 0.00001440, -0.00002038, 0.00002278, -0.00002022, 0.00001371, -0.00001122, 0.00000161, 0.00000807, 0.00000503, -0.00000263, 0.00000515, 0.00000399, -0.00000399, 0.00001346, -0.00001288, 0.00000364, -0.00000596, -0.00000639, 0.00000465, -0.00001195, 0.00001298, -0.00000703, 0.00001056, -0.00002060, 0.00002506, -0.00002738, 0.00001630, -0.00002096, 0.00002163, -0.00003407, 0.00003972, -0.00003869, 0.00004101, -0.00004956, 0.00003541, -0.00005674, 0.00006082, -0.00008054, 0.00008120, -0.00014272, 0.00012755, -0.00037887, 0.00010628, 0.00094039, -0.00061999, 0.00016096, -0.00017547, 0.00009664, -0.00009937, 0.00006885, -0.00006167, 0.00004771, -0.00005915, 0.00004299, -0.00004041, 0.00002867, -0.00002819, 0.00003386, -0.00003215, 0.00002187, -0.00002724, 0.00002579, -0.00002323, 0.00003007, -0.00002650, 0.00001866, -0.00001906, 0.00002460, -0.00003133, 0.00002927, -0.00002235, 0.00001045, 0.00000459, -0.00001072, 0.00000746, 0.00000006, -0.00001412, 0.00001183, -0.00001704, 0.00001050, -0.00001461, 0.00000396, 0.00001082, -0.00001347, 0.00001734, -0.00000976, -0.00000863, 0.00000995, -0.00000360, 0.00001193, -0.00000644, -0.00000053, 0.00000179, 0.00000400, -0.00001496, -0.00000177, 0.00001711, -0.00000449, -0.00000977, 0.00000242, 0.00001117, -0.00001140, -0.00000733, 0.00001297, -0.00000706, -0.00000212, 0.00000639, -0.00000599, 0.00000266, 0.00000275, 0.00000902, -0.00002232, 0.00000795, -0.00000165, 0.00001583, -0.00001364, 0.00000326, },
		{ 0.00000731, -0.00000992, 0.00003041, -0.00001814, -0.00000039, -0.00000736, 0.00000799, -0.00000633, 0.00000437, -0.00000700, 0.00001493, -0.00002137, 0.00001710, -0.00002375, 0.00003095, -0.00003446, 0.00003339, -0.00003402, 0.00002577, -0.00001180, 0.00002044, -0.00001940, 0.00002278, -0.00001504, 0.00001460, -0.00000336, -0.00000085, -0.00000550, 0.00000052, -0.00001082, 0.00000886, -0.00001564, 0.00001946, -0.00001390, 0.00001391, -0.00002265, 0.00002995, -0.00003726, 0.00002767, -0.00003029, 0.00002854, -0.00003991, 0.00005001, -0.00005356, 0.00005810, -0.00007249, 0.00005897, -0.00007941, 0.00008808, -0.00011582, 0.00012164, -0.00020458, 0.00020926, -0.00061300, 0.00066445, 0.00047731, -0.00055854, 0.00020860, -0.00020565, 0.00012914, -0.00012549, 0.00009498, -0.00008345, 0.00006372, -0.00007450, 0.00006090, -0.00005764, 0.00004405, -0.00003786, 0.00004115, -0.00004168, 0.00003041, -0.00003369, 0.00003232, -0.00002788, 0.00003437, -0.00003399, 0.00002633, -0.00002346, 0.00002713, -0.00003667, 0.00004019, -0.00003963, 0.00003182, -0.00001523, 0.00000277, 0.00000173, -0.00000033, -0.00001333, 0.00001360, -0.00002166, 0.00001907, -0.00002791, 0.00002337, -0.00000765, -0.00000023, 0.00001170, -0.00001365, -0.00000258, 0.00000814, -0.00000244, 0.00001161, -0.00001024, 0.00000372, 0.00000064, 0.00000212, -0.00001912, 0.00000999, 0.00001147, -0.00000763, -0.00000889, 0.00000795, 0.00000700, -0.00001491, -0.00000066, 0.00001101, -0.00001081, 0.00000399, 0.00000183, -0.00000393, 0.00000236, 0.00000572, 0.00000207, -0.00001980, 0.00001060, -0.00000132, 0.00001522, -0.00001825, },
		{ -0.00001370, 0.00000332, -0.00000036, 0.00002487, -0.00002357, 0.00000392, -0.00000658, 0.00000686, -0.00000477, 0.00000055, 0.00000129, 0.00000417, -0.00001155, 0.00000642, -0.00001053, 0.00001897, -0.00002669, 0.00003182, -0.00003865, 0.00003595, -0.00002149, 0.00002653, -0.00002648, 0.00003101, -0.00002537, 0.00002600, -0.00001466, 0.00000667, -0.00001025, 0.00000228, -0.00000942, 0.00000570, -0.00000981, 0.00001466, -0.00000948, 0.00000537, -0.00000981, 0.00001631, -0.00002684, 0.00001882, -0.00001920, 0.00001303, -0.00001879, 0.00002800, -0.00003241, 0.00003555, -0.00005276, 0.00003707, -0.00005028, 0.00005371, -0.00007816, 0.00007190, -0.00015089, 0.00012472, -0.00062498, 0.00109950, -0.00009987, -0.00027999, 0.00008936, -0.00011693, 0.00006744, -0.00007470, 0.00005639, -0.00005075, 0.00003136, -0.00004475, 0.00003788, -0.00003909, 0.00002888, -0.00002054, 0.00002202, -0.00002551, 0.00001518, -0.00001772, 0.00001762, -0.00001221, 0.00001786, -0.00002077, 0.00001510, -0.00001002, 0.00001040, -0.00002038, 0.00002811, -0.00003490, 0.00003523, -0.00002267, 0.00000848, 0.00000179, -0.00000782, -0.00000224, 0.00000233, -0.00001107, 0.00001102, -0.00002484, 0.00003009, -0.00001983, 0.00001263, 0.00000241, -0.00001394, 0.00000354, 0.00000371, 0.00000215, 0.00000651, -0.00000895, 0.00000456, 0.00000176, -0.00000418, -0.00001619, 0.00001791, 0.00000419, -0.00000926, -0.00000640, 0.00001242, 0.00000037, -0.00001479, 0.00000500, 0.00000705, -0.00001211, 0.00000964, -0.00000475, 0.00000154, -0.00000108, 0.00001105, -0.00000894, -0.00001043, 0.00000630, 0.00000577, 0.00000593, },
		{ -0.00000805, -0.00000311, -0.00000564, 0.00001596, 0.00000776, -0.00001759, 0.00000121, -0.00000065, 0.00000061, 0.00000111, -0.00000624, 0.00001240, -0.00001110, 0.00000427, -0.00001130, 0.00001265, -0.00000770, 0.00000069, 0.00000732, -0.00001979, 0.00002468, -0.00001323, 0.00001634, -0.00001748, 0.00002364, -0.00002083, 0.00002406, -0.00001472, 0.00000476, -0.00000670, -0.00000334, -0.00000097, -0.00000460, 0.00000419, 0.00000029, 0.00000415, -0.00001167, 0.00001299, -0.00001049, -0.00000096, -0.00000561, 0.00000674, -0.00001734, 0.00001980, -0.00001618, 0.00001509, -0.00001761, 0.00000118, -0.00002119, 0.00001994, -0.00003057, 0.00002043, -0.00005678, 0.00001082, -0.00013982, -0.00032428, 0.00120925, -0.00055338, 0.00004391, -0.00009911, 0.00002970, -0.00004485, 0.00001993, -0.00002065, 0.00001431, -0.00002961, 0.00001281, -0.00001188, 0.00000363, -0.00000651, 0.00001319, -0.00001250, 0.00000579, -0.00001394, 0.00001126, -0.00000914, 0.00001446, -0.00001008, 0.00000451, -0.00000694, 0.00001270, -0.00001637, 0.00000877, -0.00000015, -0.00001193, 0.00002153, -0.00001713, 0.00000644, 0.00000543, -0.00001771, 0.00001290, -0.00001555, 0.00000928, -0.00001059, -0.00000467, 0.00001918, -0.00001810, 0.00001730, -0.00000500, -0.00001258, 0.00000905, -0.00000185, 0.00000845, -0.00000173, -0.00000334, 0.00000211, 0.00000443, -0.00001355, -0.00000571, 0.00001855, -0.00000109, -0.00001107, -0.00000095, 0.00001311, -0.00000571, -0.00001261, 0.00000981, 0.00000073, -0.00000901, 0.00001124, -0.00000908, 0.00000645, -0.00000427, 0.00001509, -0.00001949, 0.00000120, -0.00000080, 0.00001539, },
		{ 0.00002043, -0.00001899, 0.00000616, -0.00001184, 0.00003000, -0.00001255, -0.00000553, -0.00000493, 0.00000650, -0.00000632, 0.00000665, -0.00001055, 0.00001915, -0.00002191, 0.00001563, -0.00002360, 0.00003083, -0.00003249, 0.00003136, -0.00002626, 0.00001234, -0.00000126, 0.00000821, -0.00000596, 0.00000379, 0.00000408, -0.00000436, 0.00001128, -0.00000595, -0.00000387, 0.00000197, -0.00001242, 0.00000949, -0.00001554, 0.00001827, -0.00001463, 0.00001738, -0.00002607, 0.00003224, -0.00003517, 0.00002449, -0.00002936, 0.00002999, -0.00004226, 0.00005173, -0.00005507, 0.00005889, -0.00006920, 0.00005502, -0.00007799, 0.00008754, -0.00011375, 0.00011984, -0.00019193, 0.00019030, -0.00048107, 0.00023240, 0.00092338, -0.00071044, 0.00024031, -0.00022902, 0.00013901, -0.00013219, 0.00009647, -0.00008421, 0.00006684, -0.00007867, 0.00006149, -0.00005597, 0.00004247, -0.00003755, 0.00004049, -0.00004019, 0.00003094, -0.00003689, 0.00003427, -0.00002996, 0.00003452, -0.00003193, 0.00002508, -0.00002398, 0.00002877, -0.00003625, 0.00003324, -0.00002779, 0.00001496, 0.00000150, -0.00000637, 0.00000300, 0.00000564, -0.00002079, 0.00002064, -0.00002687, 0.00002522, -0.00003189, 0.00002128, -0.00000257, -0.00000443, 0.00001241, -0.00000826, -0.00001012, 0.00001222, -0.00000563, 0.00001288, -0.00000913, 0.00000306, -0.00000090, 0.00000542, -0.00002133, 0.00000835, 0.00001270, -0.00000345, -0.00001296, 0.00000679, 0.00000925, -0.00000921, -0.00000967, 0.00001355, -0.00000717, -0.00000127, 0.00000674, -0.00000763, 0.00000668, -0.00000313, 0.00001340, -0.00002399, 0.00000880, -0.00000431, },
		{ -0.00000013, 0.00001583, -0.00002063, 0.00000772, -0.00000837, 0.00003252, -0.00002568, 0.00000423, -0.00000843, 0.00000941, -0.00000858, 0.00000696, -0.00000754, 0.00001583, -0.00002126, 0.00001433, -0.00002027, 0.00003044, -0.00003796, 0.00004494, -0.00004769, 0.00003798, -0.00002477, 0.00002817, -0.00002703, 0.00002494, -0.00001632, 0.00001370, -0.00000302, 0.00000308, -0.00001102, 0.00000771, -0.00001699, 0.00001385, -0.00001835, 0.00002252, -0.00001934, 0.00001922, -0.00002596, 0.00003404, -0.00004169, 0.00003236, -0.00003499, 0.00003256, -0.00004180, 0.00005343, -0.00006083, 0.00006719, -0.00008371, 0.00006999, -0.00009025, 0.00010225, -0.00013550, 0.00014577, -0.00023832, 0.00025563, -0.00071189, 0.00084359, 0.00034961, -0.00054773, 0.00022746, -0.00022142, 0.00014439, -0.00013765, 0.00010593, -0.00009206, 0.00006990, -0.00008196, 0.00006951, -0.00006537, 0.00005193, -0.00004244, 0.00004198, -0.00004269, 0.00003242, -0.00003709, 0.00003568, -0.00003025, 0.00003400, -0.00003377, 0.00002747, -0.00002357, 0.00002595, -0.00003587, 0.00003808, -0.00003889, 0.00003063, -0.00001239, 0.00000061, 0.00000435, -0.00000265, -0.00001111, 0.00001293, -0.00002112, 0.00002349, -0.00003721, 0.00003669, -0.00002115, 0.00001145, 0.00000299, -0.00000903, -0.00000559, 0.00001106, -0.00000487, 0.00001206, -0.00001155, 0.00000618, -0.00000119, 0.00000125, -0.00002195, 0.00001935, 0.00000473, -0.00000481, -0.00001274, 0.00001380, 0.00000281, -0.00001039, -0.00000587, 0.00001472, -0.00001382, 0.00000822, -0.00000218, -0.00000052, 0.00000125, 0.00000362, 0.00000435, -0.00001966, 0.00000864, },
		{ 0.00000230, 0.00001019, 0.00000246, -0.00001252, 0.00000090, 0.00000555, 0.00002034, -0.00002570, 0.00000605, -0.00000557, 0.00000542, -0.00000408, 0.00000089, 0.00000279, 0.00000278, -0.00000934, 0.00000068, -0.00000123, 0.00000925, -0.00001794, 0.00003053, -0.00004232, 0.00004081, -0.00002996, 0.00003201, -0.00003317, 0.00003307, -0.00002566, 0.00002266, -0.00000958, 0.00000436, -0.00000953, 0.00000411, -0.00001140, 0.00000720, -0.00000863, 0.00001241, -0.00000920, 0.00000563, -0.00000779, 0.00001408, -0.00002436, 0.00001625, -0.00001648, 0.00000944, -0.00001138, 0.00001903, -0.00002514, 0.00002887, -0.00004736, 0.00003054, -0.00004075, 0.00004318, -0.00006875, 0.00006127, -0.00014176, 0.00011364, -0.00065515, 0.00124851, -0.00027214, -0.00019541, 0.00005015, -0.00008991, 0.00004908, -0.00005966, 0.00004466, -0.00004023, 0.00002007, -0.00003555, 0.00003163, -0.00003405, 0.00002652, -0.00001741, 0.00001550, -0.00001835, 0.00000870, -0.00001313, 0.00001419, -0.00000918, 0.00001239, -0.00001478, 0.00001078, -0.00000551, 0.00000455, -0.00001469, 0.00002067, -0.00002847, 0.00002871, -0.00001445, 0.00000064, 0.00001043, -0.00001688, 0.00000796, -0.00000756, 0.00000081, 0.00000190, -0.00002017, 0.00003134, -0.00002481, 0.00001921, -0.00000413, -0.00001016, 0.00000163, 0.00000524, 0.00000090, 0.00000529, -0.00000760, 0.00000429, 0.00000215, -0.00000862, -0.00001280, 0.00002208, -0.00000064, -0.00000775, -0.00000797, 0.00001641, -0.00000282, -0.00001103, -0.00000018, 0.00001130, -0.00001544, 0.00001448, -0.00001046, 0.00000798, -0.00000642, 0.00001285, -0.00000900, -0.00000880, },
		{ 0.00000232, -0.00000421, 0.00002009, -0.00001295, -0.00000053, -0.00000836, 0.00002309, -0.00000115, -0.00001466, 0.00000060, 0.00000205, -0.00000293, 0.00000413, -0.00000749, 0.00001475, -0.00001298, 0.00000588, -0.00001574, 0.00002206, -0.00002062, 0.00001700, -0.00000511, -0.00001218, 0.00001905, -0.00001312, 0.00001552, -0.00001980, 0.00002343, -0.00001945, 0.00001878, -0.00000588, -0.00000326, 0.00000066, -0.00000801, 0.00000255, -0.00000738, 0.00000919, -0.00000682, 0.00000994, -0.00001617, 0.00001944, -0.00001769, 0.00000697, -0.00001392, 0.00001505, -0.00002580, 0.00003238, -0.00003268, 0.00003224, -0.00003594, 0.00001941, -0.00004118, 0.00004496, -0.00006066, 0.00005427, -0.00009860, 0.00006178, -0.00021639, -0.00025046, 0.00125580, -0.00068677, 0.00014051, -0.00015914, 0.00007536, -0.00007951, 0.00004892, -0.00004438, 0.00003472, -0.00004992, 0.00003141, -0.00002724, 0.00001706, -0.00001565, 0.00002041, -0.00002186, 0.00001620, -0.00002432, 0.00001991, -0.00001596, 0.00001944, -0.00001667, 0.00001212, -0.00001281, 0.00001775, -0.00002231, 0.00001439, -0.00000775, -0.00000485, 0.00001454, -0.00000849, -0.00000189, 0.00001450, -0.00002728, 0.00002440, -0.00002764, 0.00002556, -0.00002758, 0.00001091, 0.00000825, -0.00001200, 0.00001502, -0.00000562, -0.00001233, 0.00000981, -0.00000264, 0.00000870, -0.00000430, 0.00000013, -0.00000049, 0.00000629, -0.00002012, 0.00000327, 0.00001604, -0.00000165, -0.00001267, 0.00000143, 0.00001303, -0.00000533, -0.00001232, 0.00000747, 0.00000282, -0.00000987, 0.00001335, -0.00001265, 0.00001203, -0.00001059, 0.00001863, -0.00002026, },
		{ -0.00002369, 0.00000757, -0.00000463, 0.00002227, -0.00002206, 0.00000718, -0.00001153, 0.00003372, -0.00002105, -0.00000075, -0.00000625, 0.00000875, -0.00000961, 0.00000991, -0.00001153, 0.00002040, -0.00002209, 0.00001385, -0.00002231, 0.00003352, -0.00003961, 0.00004524, -0.00004092, 0.00002467, -0.00001306, 0.00001450, -0.00001187, 0.00000514, 0.00000255, -0.00000335, 0.00000740, 0.00000223, -0.00001292, 0.00001181, -0.00002025, 0.00001579, -0.00001986, 0.00002363, -0.00002250, 0.00002466, -0.00003132, 0.00003855, -0.00004208, 0.00003217, -0.00003748, 0.00003773, -0.00004854, 0.00006057, -0.00006826, 0.00007389, -0.00008635, 0.00007255, -0.00009651, 0.00011063, -0.00014303, 0.00015447, -0.00023724, 0.00024974, -0.00058625, 0.00038942, 0.00084644, -0.00074999, 0.00029222, -0.00026619, 0.00016999, -0.00015657, 0.00011759, -0.00010159, 0.00008110, -0.00009328, 0.00007604, -0.00006870, 0.00005446, -0.00004555, 0.00004488, -0.00004533, 0.00003725, -0.00004402, 0.00004055, -0.00003467, 0.00003646, -0.00003449, 0.00002907, -0.00002658, 0.00002976, -0.00003740, 0.00003347, -0.00002982, 0.00001682, -0.00000016, -0.00000310, 0.00000009, 0.00000892, -0.00002395, 0.00002554, -0.00003206, 0.00003546, -0.00004574, 0.00003768, -0.00001800, 0.00000795, 0.00000357, -0.00000366, -0.00001308, 0.00001502, -0.00000792, 0.00001353, -0.00001128, 0.00000648, -0.00000373, 0.00000627, -0.00002629, 0.00001888, 0.00000585, -0.00000060, -0.00001680, 0.00001212, 0.00000561, -0.00000525, -0.00001333, 0.00001490, -0.00000808, 0.00000123, 0.00000463, -0.00000668, 0.00000823, -0.00000719, 0.00001618, },
		{ 0.00000539, -0.00001802, 0.00000508, 0.00000297, 0.00001374, -0.00002040, 0.00000598, -0.00000366, 0.00002996, -0.00002964, 0.00000740, -0.00000818, 0.00000918, -0.00000944, 0.00000857, -0.00000700, 0.00001511, -0.00001927, 0.00000900, -0.00001241, 0.00002356, -0.00003352, 0.00004778, -0.00005475, 0.00004565, -0.00003464, 0.00003422, -0.00003320, 0.00002647, -0.00001639, 0.00001137, -0.00000159, 0.00000543, -0.00001509, 0.00001398, -0.00002251, 0.00001827, -0.00002006, 0.00002382, -0.00002298, 0.00002303, -0.00002731, 0.00003531, -0.00004289, 0.00003395, -0.00003690, 0.00003386, -0.00003988, 0.00005121, -0.00006141, 0.00006898, -0.00008726, 0.00007331, -0.00009227, 0.00010560, -0.00014228, 0.00015446, -0.00025310, 0.00027716, -0.00078097, 0.00101337, 0.00019113, -0.00049391, 0.00021479, -0.00021427, 0.00014175, -0.00013535, 0.00010501, -0.00009092, 0.00006735, -0.00008080, 0.00007030, -0.00006647, 0.00005468, -0.00004329, 0.00003904, -0.00003946, 0.00003023, -0.00003652, 0.00003551, -0.00002962, 0.00003048, -0.00003000, 0.00002549, -0.00002102, 0.00002155, -0.00003105, 0.00003150, -0.00003349, 0.00002508, -0.00000643, -0.00000378, 0.00000907, -0.00000736, -0.00000529, 0.00000790, -0.00001476, 0.00002091, -0.00003888, 0.00004351, -0.00003096, 0.00002164, -0.00000618, -0.00000325, -0.00000906, 0.00001376, -0.00000690, 0.00001156, -0.00001144, 0.00000730, -0.00000211, -0.00000114, -0.00002158, 0.00002631, -0.00000197, -0.00000169, -0.00001590, 0.00001893, -0.00000148, -0.00000537, -0.00001141, 0.00001817, -0.00001621, 0.00001238, -0.00000706, 0.00000433, -0.00000169, 0.00000305, },
		{ 0.00001460, -0.00000877, -0.00000734, -0.00000150, 0.00001454, -0.00000182, -0.00001029, -0.00000195, 0.00001258, 0.00001243, -0.00002416, 0.00000630, -0.00000326, 0.00000252, -0.00000216, 0.00000065, 0.00000444, 0.00000119, -0.00000719, -0.00000482, 0.00000895, -0.00000338, -0.00000383, 0.00002107, -0.00003773, 0.00003835, -0.00003208, 0.00003260, -0.00003543, 0.00003197, -0.00002276, 0.00001610, -0.00000174, -0.00000118, -0.00000548, 0.00000305, -0.00001093, 0.00000677, -0.00000560, 0.00000772, -0.00000633, 0.00000378, -0.00000362, 0.00000902, -0.00001847, 0.00001020, -0.00001070, 0.00000344, -0.00000132, 0.00000590, -0.00001237, 0.00001561, -0.00003457, 0.00001651, -0.00002289, 0.00002230, -0.00004641, 0.00003418, -0.00011075, 0.00006974, -0.00063363, 0.00134792, -0.00043155, -0.00009755, -0.00000430, -0.00004985, 0.00001946, -0.00003489, 0.00002462, -0.00002299, 0.00000297, -0.00002026, 0.00001933, -0.00002343, 0.00001968, -0.00001148, 0.00000679, -0.00000878, -0.00000025, -0.00000609, 0.00000862, -0.00000467, 0.00000559, -0.00000706, 0.00000504, -0.00000035, -0.00000203, -0.00000760, 0.00001119, -0.00001932, 0.00001910, -0.00000403, -0.00000833, 0.00001953, -0.00002593, 0.00001887, -0.00001862, 0.00001535, -0.00001178, -0.00000870, 0.00002471, -0.00002321, 0.00002090, -0.00000800, -0.00000730, 0.00000034, 0.00000597, 0.00000028, 0.00000316, -0.00000484, 0.00000252, 0.00000347, -0.00001418, -0.00000609, 0.00002234, -0.00000348, -0.00000726, -0.00000779, 0.00001826, -0.00000447, -0.00000820, -0.00000455, 0.00001422, -0.00001673, 0.00001723, -0.00001474, 0.00001361, -0.00001166, },
		{ -0.00001503, 0.00002056, -0.00001924, 0.00000167, -0.00000561, 0.00002255, -0.00001630, 0.00000089, -0.00000907, 0.00002809, -0.00001024, -0.00001002, -0.00000072, 0.00000482, -0.00000667, 0.00000730, -0.00000824, 0.00001590, -0.00001335, 0.00000521, -0.00001668, 0.00002763, -0.00003051, 0.00003211, -0.00001991, -0.00000027, 0.00000852, -0.00000793, 0.00001040, -0.00001792, 0.00002018, -0.00001532, 0.00001085, 0.00000495, -0.00001361, 0.00001069, -0.00001519, 0.00000825, -0.00001167, 0.00001514, -0.00001513, 0.00001712, -0.00002145, 0.00002623, -0.00002548, 0.00001597, -0.00002349, 0.00002424, -0.00003396, 0.00004375, -0.00004842, 0.00004940, -0.00005506, 0.00003913, -0.00006230, 0.00007081, -0.00009221, 0.00009070, -0.00014489, 0.00012059, -0.00031102, -0.00013032, 0.00123554, -0.00077348, 0.00021829, -0.00021074, 0.00011632, -0.00011144, 0.00007634, -0.00006713, 0.00005385, -0.00006864, 0.00004923, -0.00004270, 0.00003123, -0.00002552, 0.00002715, -0.00002994, 0.00002537, -0.00003362, 0.00002772, -0.00002200, 0.00002311, -0.00002162, 0.00001850, -0.00001742, 0.00002070, -0.00002573, 0.00001773, -0.00001358, 0.00000131, 0.00000748, -0.00000002, -0.00000918, 0.00002164, -0.00003380, 0.00003275, -0.00003619, 0.00003893, -0.00004311, 0.00002780, -0.00000685, -0.00000099, 0.00000795, -0.00000279, -0.00001371, 0.00001135, -0.00000403, 0.00000909, -0.00000687, 0.00000394, -0.00000364, 0.00000799, -0.00002597, 0.00001358, 0.00001040, -0.00000003, -0.00001530, 0.00000529, 0.00001078, -0.00000299, -0.00001345, 0.00000621, 0.00000356, -0.00000849, 0.00001257, -0.00001342, 0.00001506, },
		{ 0.00000726, -0.00000904, 0.00001706, -0.00002135, 0.00000429, -0.00000256, 0.00002106, -0.00002240, 0.00000597, -0.00000795, 0.00003300, -0.00002665, 0.00000330, -0.00000671, 0.00000967, -0.00001174, 0.00001226, -0.00001133, 0.00001984, -0.00002016, 0.00000919, -0.00001633, 0.00002981, -0.00003931, 0.00005162, -0.00005073, 0.00003461, -0.00002405, 0.00002121, -0.00001829, 0.00000764, 0.00000051, -0.00000204, 0.00000330, 0.00000983, -0.00002151, 0.00002164, -0.00002824, 0.00002227, -0.00002388, 0.00002792, -0.00002925, 0.00003105, -0.00003514, 0.00004265, -0.00004656, 0.00003767, -0.00004367, 0.00004343, -0.00005156, 0.00006440, -0.00007567, 0.00008283, -0.00009745, 0.00008423, -0.00010818, 0.00012519, -0.00016235, 0.00017762, -0.00026921, 0.00029332, -0.00067732, 0.00055920, 0.00071908, -0.00073660, 0.00030979, -0.00028011, 0.00018354, -0.00016733, 0.00012782, -0.00011013, 0.00008735, -0.00009999, 0.00008366, -0.00007581, 0.00006227, -0.00005054, 0.00004597, -0.00004654, 0.00003958, -0.00004738, 0.00004347, -0.00003645, 0.00003526, -0.00003354, 0.00002998, -0.00002643, 0.00002737, -0.00003448, 0.00002940, -0.00002761, 0.00001507, 0.00000046, -0.00000168, -0.00000065, 0.00000939, -0.00002291, 0.00002555, -0.00003105, 0.00003877, -0.00005259, 0.00004882, -0.00003144, 0.00002048, -0.00000692, 0.00000277, -0.00001679, 0.00001782, -0.00001000, 0.00001344, -0.00001229, 0.00000887, -0.00000585, 0.00000571, -0.00002837, 0.00002771, -0.00000168, 0.00000300, -0.00002033, 0.00001731, 0.00000124, -0.00000035, -0.00001757, 0.00001636, -0.00000904, 0.00000450, 0.00000077, -0.00000350, },
		{ 0.00000920, -0.00000539, 0.00000335, 0.00000574, -0.00001562, 0.00000082, 0.00000704, 0.00000980, -0.00001824, 0.00000295, 0.00000318, 0.00002316, -0.00002956, 0.00000860, -0.00000647, 0.00000722, -0.00000878, 0.00000919, -0.00000566, 0.00001336, -0.00001635, 0.00000265, -0.00000209, 0.00001204, -0.00002215, 0.00004139, -0.00005308, 0.00004616, -0.00003880, 0.00003615, -0.00003584, 0.00002569, -0.00001417, 0.00000684, 0.00000185, 0.00000497, -0.00001618, 0.00001764, -0.00002577, 0.00002084, -0.00001990, 0.00002256, -0.00002386, 0.00002444, -0.00002601, 0.00003316, -0.00004016, 0.00003161, -0.00003523, 0.00003197, -0.00003415, 0.00004338, -0.00005496, 0.00006265, -0.00008197, 0.00006766, -0.00008420, 0.00009655, -0.00013387, 0.00014437, -0.00024390, 0.00026550, -0.00080466, 0.00115063, 0.00002348, -0.00040938, 0.00017574, -0.00018651, 0.00012225, -0.00011900, 0.00009237, -0.00008019, 0.00005633, -0.00007115, 0.00006307, -0.00006046, 0.00005174, -0.00004023, 0.00003275, -0.00003260, 0.00002431, -0.00003227, 0.00003204, -0.00002636, 0.00002445, -0.00002336, 0.00002100, -0.00001663, 0.00001511, -0.00002351, 0.00002170, -0.00002443, 0.00001596, 0.00000183, -0.00000957, 0.00001486, -0.00001311, 0.00000266, 0.00000006, -0.00000392, 0.00001208, -0.00003246, 0.00004185, -0.00003437, 0.00002755, -0.00001332, 0.00000197, -0.00001193, 0.00001554, -0.00000809, 0.00000993, -0.00000973, 0.00000678, -0.00000187, -0.00000511, -0.00001752, 0.00002938, -0.00000691, 0.00000048, -0.00001730, 0.00002212, -0.00000459, -0.00000103, -0.00001617, 0.00002036, -0.00001688, 0.00001504, -0.00001128, },
		{ -0.00001633, 0.00001697, -0.00001541, 0.00001529, -0.00000744, -0.00000696, -0.00000400, 0.00001755, -0.00000559, -0.00000822, -0.00000379, 0.00001879, 0.00000291, -0.00001984, 0.00000515, -0.00000027, -0.00000120, 0.00000048, 0.00000029, 0.00000566, 0.00000020, -0.00000628, -0.00000816, 0.00001711, -0.00001534, 0.00001183, 0.00000652, -0.00002640, 0.00002905, -0.00002775, 0.00002785, -0.00003274, 0.00002723, -0.00001667, 0.00000677, 0.00000829, -0.00000954, 0.00000182, -0.00000114, -0.00000763, 0.00000411, -0.00000082, 0.00000091, -0.00000110, 0.00000006, 0.00000221, 0.00000185, -0.00000993, 0.00000137, -0.00000250, -0.00000401, 0.00000973, -0.00000920, 0.00000354, -0.00000194, -0.00001659, -0.00000263, 0.00000042, -0.00000553, -0.00001486, -0.00000537, -0.00006193, -0.00000401, -0.00055705, 0.00138284, -0.00055705, -0.00000401, -0.00006193, -0.00000537, -0.00001486, -0.00000553, 0.00000042, -0.00000263, -0.00001659, -0.00000194, 0.00000354, -0.00000920, 0.00000973, -0.00000401, -0.00000250, 0.00000137, -0.00000993, 0.00000185, 0.00000221, 0.00000006, -0.00000110, 0.00000091, -0.00000082, 0.00000411, -0.00000763, -0.00000114, 0.00000182, -0.00000954, 0.00000829, 0.00000677, -0.00001667, 0.00002723, -0.00003274, 0.00002785, -0.00002775, 0.00002905, -0.00002640, 0.00000652, 0.00001183, -0.00001534, 0.00001711, -0.00000816, -0.00000628, 0.00000020, 0.00000566, 0.00000029, 0.00000048, -0.00000120, -0.00000027, 0.00000515, -0.00001984, 0.00000291, 0.00001879, -0.00000379, -0.00000822, -0.00000559, 0.00001755, -0.00000400, -0.00000696, -0.00000744, 0.00001529, -0.00001541, 0.00001697, },
		{ 0.00000920, -0.00001128, 0.00001504, -0.00001688, 0.00002036, -0.00001617, -0.00000103, -0.00000459, 0.00002212, -0.00001730, 0.00000048, -0.00000691, 0.00002938, -0.00001752, -0.00000511, -0.00000187, 0.00000678, -0.00000973, 0.00000993, -0.00000809, 0.00001554, -0.00001193, 0.00000197, -0.00001332, 0.00002755, -0.00003437, 0.00004185, -0.00003246, 0.00001208, -0.00000392, 0.00000006, 0.00000266, -0.00001311, 0.00001486, -0.00000957, 0.00000183, 0.00001596, -0.00002443, 0.00002170, -0.00002351, 0.00001511, -0.00001663, 0.00002100, -0.00002336, 0.00002445, -0.00002636, 0.00003204, -0.00003227, 0.00002431, -0.00003260, 0.00003275, -0.00004023, 0.00005174, -0.00006046, 0.00006307, -0.00007115, 0.00005633, -0.00008019, 0.00009237, -0.00011900, 0.00012225, -0.00018651, 0.00017574, -0.00040938, 0.00002348, 0.00115063, -0.00080466, 0.00026550, -0.00024390, 0.00014437, -0.00013387, 0.00009655, -0.00008420, 0.00006766, -0.00008197, 0.00006265, -0.00005496, 0.00004338, -0.00003415, 0.00003197, -0.00003523, 0.00003161, -0.00004016, 0.00003316, -0.00002601, 0.00002444, -0.00002386, 0.00002256, -0.00001990, 0.00002084, -0.00002577, 0.00001764, -0.00001618, 0.00000497, 0.00000185, 0.00000684, -0.00001417, 0.00002569, -0.00003584, 0.00003615, -0.00003880, 0.00004616, -0.00005308, 0.00004139, -0.00002215, 0.00001204, -0.00000209, 0.00000265, -0.00001635, 0.00001336, -0.00000566, 0.00000919, -0.00000878, 0.00000722, -0.00000647, 0.00000860, -0.00002956, 0.00002316, 0.00000318, 0.00000295, -0.00001824, 0.00000980, 0.00000704, 0.00000082, -0.00001562, 0.00000574, 0.00000335, -0.00000539, },
		{ 0.00000726, -0.00000350, 0.00000077, 0.00000450, -0.00000904, 0.00001636, -0.00001757, -0.00000035, 0.00000124, 0.00001731, -0.00002033, 0.00000300, -0.00000168, 0.00002771, -0.00002837, 0.00000571, -0.00000585, 0.00000887, -0.00001229, 0.00001344, -0.00001000, 0.00001782, -0.00001679, 0.00000277, -0.00000692, 0.00002048, -0.00003144, 0.00004882, -0.00005259, 0.00003877, -0.00003105, 0.00002555, -0.00002291, 0.00000939, -0.00000065, -0.00000168, 0.00000046, 0.00001507, -0.00002761, 0.00002940, -0.00003448, 0.00002737, -0.00002643, 0.00002998, -0.00003354, 0.00003526, -0.00003645, 0.00004347, -0.00004738, 0.00003958, -0.00004654, 0.00004597, -0.00005054, 0.00006227, -0.00007581, 0.00008366, -0.00009999, 0.00008735, -0.00011013, 0.00012782, -0.00016733, 0.00018354, -0.00028011, 0.00030979, -0.00073660, 0.00071908, 0.00055920, -0.00067732, 0.00029332, -0.00026921, 0.00017762, -0.00016235, 0.00012519, -0.00010818, 0.00008423, -0.00009745, 0.00008283, -0.00007567, 0.00006440, -0.00005156, 0.00004343, -0.00004367, 0.00003767, -0.00004656, 0.00004265, -0.00003514, 0.00003105, -0.00002925, 0.00002792, -0.00002388, 0.00002227, -0.00002824, 0.00002164, -0.00002151, 0.00000983, 0.00000330, -0.00000204, 0.00000051, 0.00000764, -0.00001829, 0.00002121, -0.00002405, 0.00003461, -0.00005073, 0.00005162, -0.00003931, 0.00002981, -0.00001633, 0.00000919, -0.00002016, 0.00001984, -0.00001133, 0.00001226, -0.00001174, 0.00000967, -0.00000671, 0.00000330, -0.00002665, 0.00003300, -0.00000795, 0.00000597, -0.00002240, 0.00002106, -0.00000256, 0.00000429, -0.00002135, 0.00001706, -0.00000904, },
		{ -0.00001503, 0.00001506, -0.00001342, 0.00001257, -0.00000849, 0.00000356, 0.00000621, -0.00001345, -0.00000299, 0.00001078, 0.00000529, -0.00001530, -0.00000003, 0.00001040, 0.00001358, -0.00002597, 0.00000799, -0.00000364, 0.00000394, -0.00000687, 0.00000909, -0.00000403, 0.00001135, -0.00001371, -0.00000279, 0.00000795, -0.00000099, -0.00000685, 0.00002780, -0.00004311, 0.00003893, -0.00003619, 0.00003275, -0.00003380, 0.00002164, -0.00000918, -0.00000002, 0.00000748, 0.00000131, -0.00001358, 0.00001773, -0.00002573, 0.00002070, -0.00001742, 0.00001850, -0.00002162, 0.00002311, -0.00002200, 0.00002772, -0.00003362, 0.00002537, -0.00002994, 0.00002715, -0.00002552, 0.00003123, -0.00004270, 0.00004923, -0.00006864, 0.00005385, -0.00006713, 0.00007634, -0.00011144, 0.00011632, -0.00021074, 0.00021829, -0.00077348, 0.00123554, -0.00013032, -0.00031102, 0.00012059, -0.00014489, 0.00009070, -0.00009221, 0.00007081, -0.00006230, 0.00003913, -0.00005506, 0.00004940, -0.00004842, 0.00004375, -0.00003396, 0.00002424, -0.00002349, 0.00001597, -0.00002548, 0.00002623, -0.00002145, 0.00001712, -0.00001513, 0.00001514, -0.00001167, 0.00000825, -0.00001519, 0.00001069, -0.00001361, 0.00000495, 0.00001085, -0.00001532, 0.00002018, -0.00001792, 0.00001040, -0.00000793, 0.00000852, -0.00000027, -0.00001991, 0.00003211, -0.00003051, 0.00002763, -0.00001668, 0.00000521, -0.00001335, 0.00001590, -0.00000824, 0.00000730, -0.00000667, 0.00000482, -0.00000072, -0.00001002, -0.00001024, 0.00002809, -0.00000907, 0.00000089, -0.00001630, 0.00002255, -0.00000561, 0.00000167, -0.00001924, 0.00002056, },
		{ 0.00001460, -0.00001166, 0.00001361, -0.00001474, 0.00001723, -0.00001673, 0.00001422, -0.00000455, -0.00000820, -0.00000447, 0.00001826, -0.00000779, -0.00000726, -0.00000348, 0.00002234, -0.00000609, -0.00001418, 0.00000347, 0.00000252, -0.00000484, 0.00000316, 0.00000028, 0.00000597, 0.00000034, -0.00000730, -0.00000800, 0.00002090, -0.00002321, 0.00002471, -0.00000870, -0.00001178, 0.00001535, -0.00001862, 0.00001887, -0.00002593, 0.00001953, -0.00000833, -0.00000403, 0.00001910, -0.00001932, 0.00001119, -0.00000760, -0.00000203, -0.00000035, 0.00000504, -0.00000706, 0.00000559, -0.00000467, 0.00000862, -0.00000609, -0.00000025, -0.00000878, 0.00000679, -0.00001148, 0.00001968, -0.00002343, 0.00001933, -0.00002026, 0.00000297, -0.00002299, 0.00002462, -0.00003489, 0.00001946, -0.00004985, -0.00000430, -0.00009755, -0.00043155, 0.00134792, -0.00063363, 0.00006974, -0.00011075, 0.00003418, -0.00004641, 0.00002230, -0.00002289, 0.00001651, -0.00003457, 0.00001561, -0.00001237, 0.00000590, -0.00000132, 0.00000344, -0.00001070, 0.00001020, -0.00001847, 0.00000902, -0.00000362, 0.00000378, -0.00000633, 0.00000772, -0.00000560, 0.00000677, -0.00001093, 0.00000305, -0.00000548, -0.00000118, -0.00000174, 0.00001610, -0.00002276, 0.00003197, -0.00003543, 0.00003260, -0.00003208, 0.00003835, -0.00003773, 0.00002107, -0.00000383, -0.00000338, 0.00000895, -0.00000482, -0.00000719, 0.00000119, 0.00000444, 0.00000065, -0.00000216, 0.00000252, -0.00000326, 0.00000630, -0.00002416, 0.00001243, 0.00001258, -0.00000195, -0.00001029, -0.00000182, 0.00001454, -0.00000150, -0.00000734, -0.00000877, },
		{ 0.00000539, 0.00000305, -0.00000169, 0.00000433, -0.00000706, 0.00001238, -0.00001621, 0.00001817, -0.00001141, -0.00000537, -0.00000148, 0.00001893, -0.00001590, -0.00000169, -0.00000197, 0.00002631, -0.00002158, -0.00000114, -0.00000211, 0.00000730, -0.00001144, 0.00001156, -0.00000690, 0.00001376, -0.00000906, -0.00000325, -0.00000618, 0.00002164, -0.00003096, 0.00004351, -0.00003888, 0.00002091, -0.00001476, 0.00000790, -0.00000529, -0.00000736, 0.00000907, -0.00000378, -0.00000643, 0.00002508, -0.00003349, 0.00003150, -0.00003105, 0.00002155, -0.00002102, 0.00002549, -0.00003000, 0.00003048, -0.00002962, 0.00003551, -0.00003652, 0.00003023, -0.00003946, 0.00003904, -0.00004329, 0.00005468, -0.00006647, 0.00007030, -0.00008080, 0.00006735, -0.00009092, 0.00010501, -0.00013535, 0.00014175, -0.00021427, 0.00021479, -0.00049391, 0.00019113, 0.00101337, -0.00078097, 0.00027716, -0.00025310, 0.00015446, -0.00014228, 0.00010560, -0.00009227, 0.00007331, -0.00008726, 0.00006898, -0.00006141, 0.00005121, -0.00003988, 0.00003386, -0.00003690, 0.00003395, -0.00004289, 0.00003531, -0.00002731, 0.00002303, -0.00002298, 0.00002382, -0.00002006, 0.00001827, -0.00002251, 0.00001398, -0.00001509, 0.00000543, -0.00000159, 0.00001137, -0.00001639, 0.00002647, -0.00003320, 0.00003422, -0.00003464, 0.00004565, -0.00005475, 0.00004778, -0.00003352, 0.00002356, -0.00001241, 0.00000900, -0.00001927, 0.00001511, -0.00000700, 0.00000857, -0.00000944, 0.00000918, -0.00000818, 0.00000740, -0.00002964, 0.00002996, -0.00000366, 0.00000598, -0.00002040, 0.00001374, 0.00000297, 0.00000508, -0.00001802, },
		{ -0.00002369, 0.00001618, -0.00000719, 0.00000823, -0.00000668, 0.00000463, 0.00000123, -0.00000808, 0.00001490, -0.00001333, -0.00000525, 0.00000561, 0.00001212, -0.00001680, -0.00000060, 0.00000585, 0.00001888, -0.00002629, 0.00000627, -0.00000373, 0.00000648, -0.00001128, 0.00001353, -0.00000792, 0.00001502, -0.00001308, -0.00000366, 0.00000357, 0.00000795, -0.00001800, 0.00003768, -0.00004574, 0.00003546, -0.00003206, 0.00002554, -0.00002395, 0.00000892, 0.00000009, -0.00000310, -0.00000016, 0.00001682, -0.00002982, 0.00003347, -0.00003740, 0.00002976, -0.00002658, 0.00002907, -0.00003449, 0.00003646, -0.00003467, 0.00004055, -0.00004402, 0.00003725, -0.00004533, 0.00004488, -0.00004555, 0.00005446, -0.00006870, 0.00007604, -0.00009328, 0.00008110, -0.00010159, 0.00011759, -0.00015657, 0.00016999, -0.00026619, 0.00029222, -0.00074999, 0.00084644, 0.00038942, -0.00058625, 0.00024974, -0.00023724, 0.00015447, -0.00014303, 0.00011063, -0.00009651, 0.00007255, -0.00008635, 0.00007389, -0.00006826, 0.00006057, -0.00004854, 0.00003773, -0.00003748, 0.00003217, -0.00004208, 0.00003855, -0.00003132, 0.00002466, -0.00002250, 0.00002363, -0.00001986, 0.00001579, -0.00002025, 0.00001181, -0.00001292, 0.00000223, 0.00000740, -0.00000335, 0.00000255, 0.00000514, -0.00001187, 0.00001450, -0.00001306, 0.00002467, -0.00004092, 0.00004524, -0.00003961, 0.00003352, -0.00002231, 0.00001385, -0.00002209, 0.00002040, -0.00001153, 0.00000991, -0.00000961, 0.00000875, -0.00000625, -0.00000075, -0.00002105, 0.00003372, -0.00001153, 0.00000718, -0.00002206, 0.00002227, -0.00000463, 0.00000757, },
		{ 0.00000232, -0.00002026, 0.00001863, -0.00001059, 0.00001203, -0.00001265, 0.00001335, -0.00000987, 0.00000282, 0.00000747, -0.00001232, -0.00000533, 0.00001303, 0.00000143, -0.00001267, -0.00000165, 0.00001604, 0.00000327, -0.00002012, 0.00000629, -0.00000049, 0.00000013, -0.00000430, 0.00000870, -0.00000264, 0.00000981, -0.00001233, -0.00000562, 0.00001502, -0.00001200, 0.00000825, 0.00001091, -0.00002758, 0.00002556, -0.00002764, 0.00002440, -0.00002728, 0.00001450, -0.00000189, -0.00000849, 0.00001454, -0.00000485, -0.00000775, 0.00001439, -0.00002231, 0.00001775, -0.00001281, 0.00001212, -0.00001667, 0.00001944, -0.00001596, 0.00001991, -0.00002432, 0.00001620, -0.00002186, 0.00002041, -0.00001565, 0.00001706, -0.00002724, 0.00003141, -0.00004992, 0.00003472, -0.00004438, 0.00004892, -0.00007951, 0.00007536, -0.00015914, 0.00014051, -0.00068677, 0.00125580, -0.00025046, -0.00021639, 0.00006178, -0.00009860, 0.00005427, -0.00006066, 0.00004496, -0.00004118, 0.00001941, -0.00003594, 0.00003224, -0.00003268, 0.00003238, -0.00002580, 0.00001505, -0.00001392, 0.00000697, -0.00001769, 0.00001944, -0.00001617, 0.00000994, -0.00000682, 0.00000919, -0.00000738, 0.00000255, -0.00000801, 0.00000066, -0.00000326, -0.00000588, 0.00001878, -0.00001945, 0.00002343, -0.00001980, 0.00001552, -0.00001312, 0.00001905, -0.00001218, -0.00000511, 0.00001700, -0.00002062, 0.00002206, -0.00001574, 0.00000588, -0.00001298, 0.00001475, -0.00000749, 0.00000413, -0.00000293, 0.00000205, 0.00000060, -0.00001466, -0.00000115, 0.00002309, -0.00000836, -0.00000053, -0.00001295, 0.00002009, -0.00000421, },
		{ 0.00000230, -0.00000880, -0.00000900, 0.00001285, -0.00000642, 0.00000798, -0.00001046, 0.00001448, -0.00001544, 0.00001130, -0.00000018, -0.00001103, -0.00000282, 0.00001641, -0.00000797, -0.00000775, -0.00000064, 0.00002208, -0.00001280, -0.00000862, 0.00000215, 0.00000429, -0.00000760, 0.00000529, 0.00000090, 0.00000524, 0.00000163, -0.00001016, -0.00000413, 0.00001921, -0.00002481, 0.00003134, -0.00002017, 0.00000190, 0.00000081, -0.00000756, 0.00000796, -0.00001688, 0.00001043, 0.00000064, -0.00001445, 0.00002871, -0.00002847, 0.00002067, -0.00001469, 0.00000455, -0.00000551, 0.00001078, -0.00001478, 0.00001239, -0.00000918, 0.00001419, -0.00001313, 0.00000870, -0.00001835, 0.00001550, -0.00001741, 0.00002652, -0.00003405, 0.00003163, -0.00003555, 0.00002007, -0.00004023, 0.00004466, -0.00005966, 0.00004908, -0.00008991, 0.00005015, -0.00019541, -0.00027214, 0.00124851, -0.00065515, 0.00011364, -0.00014176, 0.00006127, -0.00006875, 0.00004318, -0.00004075, 0.00003054, -0.00004736, 0.00002887, -0.00002514, 0.00001903, -0.00001138, 0.00000944, -0.00001648, 0.00001625, -0.00002436, 0.00001408, -0.00000779, 0.00000563, -0.00000920, 0.00001241, -0.00000863, 0.00000720, -0.00001140, 0.00000411, -0.00000953, 0.00000436, -0.00000958, 0.00002266, -0.00002566, 0.00003307, -0.00003317, 0.00003201, -0.00002996, 0.00004081, -0.00004232, 0.00003053, -0.00001794, 0.00000925, -0.00000123, 0.00000068, -0.00000934, 0.00000278, 0.00000279, 0.00000089, -0.00000408, 0.00000542, -0.00000557, 0.00000605, -0.00002570, 0.00002034, 0.00000555, 0.00000090, -0.00001252, 0.00000246, 0.00001019, },
		{ -0.00000013, 0.00000864, -0.00001966, 0.00000435, 0.00000362, 0.00000125, -0.00000052, -0.00000218, 0.00000822, -0.00001382, 0.00001472, -0.00000587, -0.00001039, 0.00000281, 0.00001380, -0.00001274, -0.00000481, 0.00000473, 0.00001935, -0.00002195, 0.00000125, -0.00000119, 0.00000618, -0.00001155, 0.00001206, -0.00000487, 0.00001106, -0.00000559, -0.00000903, 0.00000299, 0.00001145, -0.00002115, 0.00003669, -0.00003721, 0.00002349, -0.00002112, 0.00001293, -0.00001111, -0.00000265, 0.00000435, 0.00000061, -0.00001239, 0.00003063, -0.00003889, 0.00003808, -0.00003587, 0.00002595, -0.00002357, 0.00002747, -0.00003377, 0.00003400, -0.00003025, 0.00003568, -0.00003709, 0.00003242, -0.00004269, 0.00004198, -0.00004244, 0.00005193, -0.00006537, 0.00006951, -0.00008196, 0.00006990, -0.00009206, 0.00010593, -0.00013765, 0.00014439, -0.00022142, 0.00022746, -0.00054773, 0.00034961, 0.00084359, -0.00071189, 0.00025563, -0.00023832, 0.00014577, -0.00013550, 0.00010225, -0.00009025, 0.00006999, -0.00008371, 0.00006719, -0.00006083, 0.00005343, -0.00004180, 0.00003256, -0.00003499, 0.00003236, -0.00004169, 0.00003404, -0.00002596, 0.00001922, -0.00001934, 0.00002252, -0.00001835, 0.00001385, -0.00001699, 0.00000771, -0.00001102, 0.00000308, -0.00000302, 0.00001370, -0.00001632, 0.00002494, -0.00002703, 0.00002817, -0.00002477, 0.00003798, -0.00004769, 0.00004494, -0.00003796, 0.00003044, -0.00002027, 0.00001433, -0.00002126, 0.00001583, -0.00000754, 0.00000696, -0.00000858, 0.00000941, -0.00000843, 0.00000423, -0.00002568, 0.00003252, -0.00000837, 0.00000772, -0.00002063, 0.00001583, },
		{ 0.00002043, -0.00000431, 0.00000880, -0.00002399, 0.00001340, -0.00000313, 0.00000668, -0.00000763, 0.00000674, -0.00000127, -0.00000717, 0.00001355, -0.00000967, -0.00000921, 0.00000925, 0.00000679, -0.00001296, -0.00000345, 0.00001270, 0.00000835, -0.00002133, 0.00000542, -0.00000090, 0.00000306, -0.00000913, 0.00001288, -0.00000563, 0.00001222, -0.00001012, -0.00000826, 0.00001241, -0.00000443, -0.00000257, 0.00002128, -0.00003189, 0.00002522, -0.00002687, 0.00002064, -0.00002079, 0.00000564, 0.00000300, -0.00000637, 0.00000150, 0.00001496, -0.00002779, 0.00003324, -0.00003625, 0.00002877, -0.00002398, 0.00002508, -0.00003193, 0.00003452, -0.00002996, 0.00003427, -0.00003689, 0.00003094, -0.00004019, 0.00004049, -0.00003755, 0.00004247, -0.00005597, 0.00006149, -0.00007867, 0.00006684, -0.00008421, 0.00009647, -0.00013219, 0.00013901, -0.00022902, 0.00024031, -0.00071044, 0.00092338, 0.00023240, -0.00048107, 0.00019030, -0.00019193, 0.00011984, -0.00011375, 0.00008754, -0.00007799, 0.00005502, -0.00006920, 0.00005889, -0.00005507, 0.00005173, -0.00004226, 0.00002999, -0.00002936, 0.00002449, -0.00003517, 0.00003224, -0.00002607, 0.00001738, -0.00001463, 0.00001827, -0.00001554, 0.00000949, -0.00001242, 0.00000197, -0.00000387, -0.00000595, 0.00001128, -0.00000436, 0.00000408, 0.00000379, -0.00000596, 0.00000821, -0.00000126, 0.00001234, -0.00002626, 0.00003136, -0.00003249, 0.00003083, -0.00002360, 0.00001563, -0.00002191, 0.00001915, -0.00001055, 0.00000665, -0.00000632, 0.00000650, -0.00000493, -0.00000553, -0.00001255, 0.00003000, -0.00001184, 0.00000616, -0.00001899, },
		{ -0.00000805, 0.00001539, -0.00000080, 0.00000120, -0.00001949, 0.00001509, -0.00000427, 0.00000645, -0.00000908, 0.00001124, -0.00000901, 0.00000073, 0.00000981, -0.00001261, -0.00000571, 0.00001311, -0.00000095, -0.00001107, -0.00000109, 0.00001855, -0.00000571, -0.00001355, 0.00000443, 0.00000211, -0.00000334, -0.00000173, 0.00000845, -0.00000185, 0.00000905, -0.00001258, -0.00000500, 0.00001730, -0.00001810, 0.00001918, -0.00000467, -0.00001059, 0.00000928, -0.00001555, 0.00001290, -0.00001771, 0.00000543, 0.00000644, -0.00001713, 0.00002153, -0.00001193, -0.00000015, 0.00000877, -0.00001637, 0.00001270, -0.00000694, 0.00000451, -0.00001008, 0.00001446, -0.00000914, 0.00001126, -0.00001394, 0.00000579, -0.00001250, 0.00001319, -0.00000651, 0.00000363, -0.00001188, 0.00001281, -0.00002961, 0.00001431, -0.00002065, 0.00001993, -0.00004485, 0.00002970, -0.00009911, 0.00004391, -0.00055338, 0.00120925, -0.00032428, -0.00013982, 0.00001082, -0.00005678, 0.00002043, -0.00003057, 0.00001994, -0.00002119, 0.00000118, -0.00001761, 0.00001509, -0.00001618, 0.00001980, -0.00001734, 0.00000674, -0.00000561, -0.00000096, -0.00001049, 0.00001299, -0.00001167, 0.00000415, 0.00000029, 0.00000419, -0.00000460, -0.00000097, -0.00000334, -0.00000670, 0.00000476, -0.00001472, 0.00002406, -0.00002083, 0.00002364, -0.00001748, 0.00001634, -0.00001323, 0.00002468, -0.00001979, 0.00000732, 0.00000069, -0.00000770, 0.00001265, -0.00001130, 0.00000427, -0.00001110, 0.00001240, -0.00000624, 0.00000111, 0.00000061, -0.00000065, 0.00000121, -0.00001759, 0.00000776, 0.00001596, -0.00000564, -0.00000311, },
		{ -0.00001370, 0.00000593, 0.00000577, 0.00000630, -0.00001043, -0.00000894, 0.00001105, -0.00000108, 0.00000154, -0.00000475, 0.00000964, -0.00001211, 0.00000705, 0.00000500, -0.00001479, 0.00000037, 0.00001242, -0.00000640, -0.00000926, 0.00000419, 0.00001791, -0.00001619, -0.00000418, 0.00000176, 0.00000456, -0.00000895, 0.00000651, 0.00000215, 0.00000371, 0.00000354, -0.00001394, 0.00000241, 0.00001263, -0.00001983, 0.00003009, -0.00002484, 0.00001102, -0.00001107, 0.00000233, -0.00000224, -0.00000782, 0.00000179, 0.00000848, -0.00002267, 0.00003523, -0.00003490, 0.00002811, -0.00002038, 0.00001040, -0.00001002, 0.00001510, -0.00002077, 0.00001786, -0.00001221, 0.00001762, -0.00001772, 0.00001518, -0.00002551, 0.00002202, -0.00002054, 0.00002888, -0.00003909, 0.00003788, -0.00004475, 0.00003136, -0.00005075, 0.00005639, -0.00007470, 0.00006744, -0.00011693, 0.00008936, -0.00027999, -0.00009987, 0.00109950, -0.00062498, 0.00012472, -0.00015089, 0.00007190, -0.00007816, 0.00005371, -0.00005028, 0.00003707, -0.00005276, 0.00003555, -0.00003241, 0.00002800, -0.00001879, 0.00001303, -0.00001920, 0.00001882, -0.00002684, 0.00001631, -0.00000981, 0.00000537, -0.00000948, 0.00001466, -0.00000981, 0.00000570, -0.00000942, 0.00000228, -0.00001025, 0.00000667, -0.00001466, 0.00002600, -0.00002537, 0.00003101, -0.00002648, 0.00002653, -0.00002149, 0.00003595, -0.00003865, 0.00003182, -0.00002669, 0.00001897, -0.00001053, 0.00000642, -0.00001155, 0.00000417, 0.00000129, 0.00000055, -0.00000477, 0.00000686, -0.00000658, 0.00000392, -0.00002357, 0.00002487, -0.00000036, 0.00000332, },
		{ 0.00000731, -0.00001825, 0.00001522, -0.00000132, 0.00001060, -0.00001980, 0.00000207, 0.00000572, 0.00000236, -0.00000393, 0.00000183, 0.00000399, -0.00001081, 0.00001101, -0.00000066, -0.00001491, 0.00000700, 0.00000795, -0.00000889, -0.00000763, 0.00001147, 0.00000999, -0.00001912, 0.00000212, 0.00000064, 0.00000372, -0.00001024, 0.00001161, -0.00000244, 0.00000814, -0.00000258, -0.00001365, 0.00001170, -0.00000023, -0.00000765, 0.00002337, -0.00002791, 0.00001907, -0.00002166, 0.00001360, -0.00001333, -0.00000033, 0.00000173, 0.00000277, -0.00001523, 0.00003182, -0.00003963, 0.00004019, -0.00003667, 0.00002713, -0.00002346, 0.00002633, -0.00003399, 0.00003437, -0.00002788, 0.00003232, -0.00003369, 0.00003041, -0.00004168, 0.00004115, -0.00003786, 0.00004405, -0.00005764, 0.00006090, -0.00007450, 0.00006372, -0.00008345, 0.00009498, -0.00012549, 0.00012914, -0.00020565, 0.00020860, -0.00055854, 0.00047731, 0.00066445, -0.00061300, 0.00020926, -0.00020458, 0.00012164, -0.00011582, 0.00008808, -0.00007941, 0.00005897, -0.00007249, 0.00005810, -0.00005356, 0.00005001, -0.00003991, 0.00002854, -0.00003029, 0.00002767, -0.00003726, 0.00002995, -0.00002265, 0.00001391, -0.00001390, 0.00001946, -0.00001564, 0.00000886, -0.00001082, 0.00000052, -0.00000550, -0.00000085, -0.00000336, 0.00001460, -0.00001504, 0.00002278, -0.00001940, 0.00002044, -0.00001180, 0.00002577, -0.00003402, 0.00003339, -0.00003446, 0.00003095, -0.00002375, 0.00001710, -0.00002137, 0.00001493, -0.00000700, 0.00000437, -0.00000633, 0.00000799, -0.00000736, -0.00000039, -0.00001814, 0.00003041, -0.00000992, },
		{ -0.00000936, 0.00000326, -0.00001364, 0.00001583, -0.00000165, 0.00000795, -0.00002232, 0.00000902, 0.00000275, 0.00000266, -0.00000599, 0.00000639, -0.00000212, -0.00000706, 0.00001297, -0.00000733, -0.00001140, 0.00001117, 0.00000242, -0.00000977, -0.00000449, 0.00001711, -0.00000177, -0.00001496, 0.00000400, 0.00000179, -0.00000053, -0.00000644, 0.00001193, -0.00000360, 0.00000995, -0.00000863, -0.00000976, 0.00001734, -0.00001347, 0.00001082, 0.00000396, -0.00001461, 0.00001050, -0.00001704, 0.00001183, -0.00001412, 0.00000006, 0.00000746, -0.00001072, 0.00000459, 0.00001045, -0.00002235, 0.00002927, -0.00003133, 0.00002460, -0.00001906, 0.00001866, -0.00002650, 0.00003007, -0.00002323, 0.00002579, -0.00002724, 0.00002187, -0.00003215, 0.00003386, -0.00002819, 0.00002867, -0.00004041, 0.00004299, -0.00005915, 0.00004771, -0.00006167, 0.00006885, -0.00009937, 0.00009664, -0.00017547, 0.00016096, -0.00061999, 0.00094039, 0.00010628, -0.00037887, 0.00012755, -0.00014272, 0.00008120, -0.00008054, 0.00006082, -0.00005674, 0.00003541, -0.00004956, 0.00004101, -0.00003869, 0.00003972, -0.00003407, 0.00002163, -0.00002096, 0.00001630, -0.00002738, 0.00002506, -0.00002060, 0.00001056, -0.00000703, 0.00001298, -0.00001195, 0.00000465, -0.00000639, -0.00000596, 0.00000364, -0.00001288, 0.00001346, -0.00000399, 0.00000399, 0.00000515, -0.00000263, 0.00000503, 0.00000807, 0.00000161, -0.00001122, 0.00001371, -0.00002022, 0.00002278, -0.00002038, 0.00001440, -0.00001961, 0.00001622, -0.00000867, 0.00000309, -0.00000265, 0.00000368, -0.00000358, -0.00000971, -0.00000289, 0.00002304, },
		{ 0.00000872, -0.00000230, -0.00000566, -0.00000287, 0.00000965, 0.00000363, -0.00000091, -0.00001768, 0.00001096, 0.00000263, -0.00000043, -0.00000366, 0.00000683, -0.00000615, -0.00000251, 0.00001293, -0.00001406, -0.00000429, 0.00001102, -0.00000169, -0.00001049, 0.00000158, 0.00001728, -0.00001189, -0.00000759, 0.00000321, 0.00000345, -0.00000574, 0.00000033, 0.00000852, -0.00000168, 0.00000892, -0.00001404, -0.00000127, 0.00001445, -0.00001798, 0.00002337, -0.00001509, 0.00000368, -0.00000613, -0.00000307, 0.00000084, -0.00000724, -0.00000380, 0.00001412, -0.00002425, 0.00002668, -0.00001795, 0.00000720, 0.00000273, -0.00000949, 0.00000682, -0.00000109, -0.00000288, -0.00000325, 0.00000947, -0.00000300, 0.00000351, -0.00000444, -0.00000388, -0.00000362, 0.00000702, 0.00000010, -0.00000657, 0.00000024, -0.00000290, -0.00001163, -0.00000320, -0.00000074, -0.00000480, -0.00001473, -0.00001143, -0.00004280, -0.00005533, -0.00039019, 0.00110409, -0.00034820, -0.00008969, -0.00002441, -0.00002619, -0.00000491, -0.00000702, 0.00000003, -0.00000590, -0.00001228, -0.00000333, 0.00000106, -0.00000174, 0.00000814, -0.00000997, 0.00000041, 0.00000026, -0.00000657, -0.00000505, 0.00000781, -0.00000865, 0.00000048, 0.00000539, 0.00000067, -0.00000353, -0.00000212, -0.00000164, -0.00001062, 0.00000943, -0.00002051, 0.00002588, -0.00001917, 0.00002087, -0.00001090, 0.00001256, -0.00000748, 0.00002395, -0.00002062, 0.00001346, -0.00001246, 0.00000453, 0.00000220, -0.00000513, 0.00000143, -0.00000848, 0.00000949, -0.00000507, -0.00000110, 0.00000313, -0.00000248, 0.00000038, -0.00001761, 0.00001447, },
		{ 0.00002505, -0.00000370, 0.00000413, -0.00001284, 0.00000742, 0.00000251, 0.00000941, -0.00001132, -0.00000933, 0.00001014, 0.00000298, -0.00000409, 0.00000077, 0.00000415, -0.00000785, 0.00000245, 0.00000996, -0.00001844, 0.00000403, 0.00000726, -0.00000385, -0.00001083, 0.00000969, 0.00001080, -0.00001626, -0.00000122, 0.00000236, 0.00000333, -0.00000879, 0.00000683, 0.00000379, 0.00000189, 0.00000524, -0.00001719, 0.00000961, 0.00000319, -0.00000996, 0.00002170, -0.00002187, 0.00001357, -0.00001786, 0.00000867, -0.00000957, -0.00000070, -0.00000486, 0.00001384, -0.00002744, 0.00003744, -0.00003713, 0.00003182, -0.00002299, 0.00001396, -0.00001266, 0.00001696, -0.00002392, 0.00002095, -0.00001292, 0.00001811, -0.00001892, 0.00001803, -0.00002899, 0.00002529, -0.00002031, 0.00002641, -0.00003792, 0.00003699, -0.00004630, 0.00003501, -0.00005270, 0.00005771, -0.00007730, 0.00007076, -0.00012529, 0.00010436, -0.00033554, 0.00006238, 0.00092197, -0.00055447, 0.00010693, -0.00013940, 0.00006631, -0.00007430, 0.00005326, -0.00005087, 0.00003573, -0.00005043, 0.00003507, -0.00003329, 0.00003172, -0.00002269, 0.00001394, -0.00001898, 0.00001809, -0.00002599, 0.00001575, -0.00000985, 0.00000342, -0.00000761, 0.00001477, -0.00000961, 0.00000313, -0.00000611, -0.00000134, -0.00000853, 0.00000633, -0.00001732, 0.00002648, -0.00002276, 0.00002721, -0.00001712, 0.00001808, -0.00000858, 0.00002547, -0.00002772, 0.00002424, -0.00002794, 0.00002320, -0.00001648, 0.00001053, -0.00001263, 0.00000455, 0.00000043, -0.00000057, -0.00000406, 0.00000658, -0.00000613, 0.00000005, -0.00001773, },
		{ -0.00000844, 0.00002437, -0.00000831, 0.00000471, -0.00001332, 0.00001179, -0.00000025, 0.00001060, -0.00001821, -0.00000145, 0.00000935, 0.00000122, -0.00000505, 0.00000384, 0.00000090, -0.00000826, 0.00000798, 0.00000321, -0.00001786, 0.00000995, 0.00000258, -0.00000536, -0.00000898, 0.00001644, 0.00000021, -0.00001429, 0.00000207, 0.00000272, 0.00000061, -0.00000798, 0.00001061, -0.00000011, 0.00000564, -0.00000088, -0.00001562, 0.00001754, -0.00001021, 0.00000583, 0.00000720, -0.00001363, 0.00000902, -0.00001682, 0.00000990, -0.00001173, -0.00000068, 0.00000132, 0.00000274, -0.00001517, 0.00002899, -0.00003589, 0.00003774, -0.00003317, 0.00002474, -0.00002057, 0.00002217, -0.00003075, 0.00003171, -0.00002289, 0.00002603, -0.00002698, 0.00002475, -0.00003681, 0.00003701, -0.00003058, 0.00003270, -0.00004520, 0.00004639, -0.00006018, 0.00005058, -0.00006723, 0.00007475, -0.00010183, 0.00009915, -0.00016998, 0.00015975, -0.00052174, 0.00055851, 0.00049771, -0.00050212, 0.00014978, -0.00016012, 0.00008830, -0.00008805, 0.00006685, -0.00006288, 0.00004316, -0.00005637, 0.00004405, -0.00004138, 0.00004208, -0.00003500, 0.00002278, -0.00002411, 0.00002124, -0.00003085, 0.00002412, -0.00001840, 0.00000831, -0.00000788, 0.00001565, -0.00001292, 0.00000459, -0.00000561, -0.00000577, -0.00000035, -0.00000478, -0.00000383, 0.00001506, -0.00001378, 0.00002178, -0.00001263, 0.00001389, 0.00000093, 0.00001284, -0.00001777, 0.00001609, -0.00002435, 0.00002523, -0.00002232, 0.00001667, -0.00001926, 0.00001230, -0.00000550, 0.00000122, -0.00000324, 0.00000547, -0.00000562, -0.00000531, },
		{ -0.00001191, 0.00000584, 0.00001477, -0.00000532, -0.00000049, -0.00000714, 0.00000945, 0.00000261, 0.00000562, -0.00001926, 0.00000389, 0.00000938, -0.00000292, -0.00000230, 0.00000372, -0.00000112, -0.00000797, 0.00001329, -0.00000648, -0.00001156, 0.00001096, -0.00000045, -0.00000761, -0.00000334, 0.00001799, -0.00000969, -0.00000865, 0.00000286, 0.00000354, -0.00000349, -0.00000385, 0.00001104, -0.00000208, 0.00000838, -0.00000862, -0.00000788, 0.00001725, -0.00001706, 0.00001892, -0.00001002, 0.00000183, -0.00000514, -0.00000531, 0.00000124, -0.00000563, -0.00000648, 0.00001219, -0.00001484, 0.00000767, 0.00000500, -0.00001524, 0.00002308, -0.00002389, 0.00001827, -0.00001284, 0.00001105, -0.00001939, 0.00002428, -0.00001583, 0.00001675, -0.00001690, 0.00001187, -0.00002284, 0.00002645, -0.00001933, 0.00001567, -0.00002525, 0.00002419, -0.00003858, 0.00002776, -0.00003861, 0.00004038, -0.00006500, 0.00005135, -0.00011612, 0.00006703, -0.00048964, 0.00089807, 0.00002155, -0.00029272, 0.00007218, -0.00009822, 0.00004572, -0.00004939, 0.00003541, -0.00003688, 0.00001751, -0.00003113, 0.00002369, -0.00002208, 0.00002677, -0.00002548, 0.00001395, -0.00001371, 0.00000911, -0.00002011, 0.00001817, -0.00001589, 0.00000524, -0.00000080, 0.00000859, -0.00000963, 0.00000188, -0.00000308, -0.00001078, 0.00000824, -0.00001723, 0.00001297, -0.00000175, 0.00000193, 0.00000988, -0.00000292, 0.00000655, 0.00001263, -0.00000428, -0.00000030, -0.00000314, -0.00000640, 0.00001184, -0.00001406, 0.00001095, -0.00001582, 0.00001219, -0.00000649, -0.00000005, 0.00000051, 0.00000116, -0.00000305, },
		{ -0.00000212, -0.00001420, 0.00001756, 0.00000316, 0.00000021, -0.00000693, 0.00000120, 0.00000421, 0.00000791, -0.00000301, -0.00001573, 0.00000735, 0.00000860, -0.00000696, 0.00000208, 0.00000139, -0.00000221, -0.00000616, 0.00001607, -0.00001593, -0.00000181, 0.00000738, -0.00000122, -0.00001033, 0.00000548, 0.00001267, -0.00001474, -0.00000297, 0.00000296, 0.00000327, -0.00000672, 0.00000165, 0.00000889, -0.00000186, 0.00000886, -0.00001566, 0.00000417, 0.00000766, -0.00001222, 0.00002031, -0.00001831, 0.00001221, -0.00001751, 0.00000690, -0.00000926, 0.00000194, -0.00001142, 0.00001962, -0.00002841, 0.00002854, -0.00002115, 0.00001234, -0.00000180, -0.00000342, 0.00000155, 0.00000345, -0.00000865, 0.00000245, 0.00000567, 0.00000124, -0.00000191, 0.00000255, -0.00001106, 0.00000323, 0.00000305, 0.00000305, -0.00001196, 0.00000696, -0.00001305, 0.00000102, -0.00001467, 0.00001177, -0.00002082, 0.00000506, -0.00004020, -0.00000145, -0.00014017, -0.00021869, 0.00095688, -0.00032756, -0.00006741, -0.00004103, -0.00000987, -0.00001876, 0.00000717, -0.00001236, 0.00000273, -0.00001920, 0.00000500, -0.00000779, 0.00000863, -0.00000100, -0.00000459, -0.00000345, 0.00000325, -0.00000935, -0.00000186, 0.00000424, -0.00000721, -0.00000099, 0.00000829, -0.00000139, -0.00000377, -0.00000148, -0.00000236, -0.00001140, 0.00001076, -0.00002312, 0.00002431, -0.00001510, 0.00001622, -0.00000129, 0.00000541, 0.00000314, 0.00001750, -0.00001447, 0.00001139, -0.00001926, 0.00001276, -0.00000641, 0.00000071, -0.00000130, -0.00000607, 0.00000669, -0.00000443, -0.00000203, 0.00000409, -0.00000299, },
		{ -0.00000460, -0.00000475, -0.00000911, 0.00002106, -0.00000394, 0.00000283, -0.00000955, 0.00000638, 0.00000112, 0.00001094, -0.00001095, -0.00001054, 0.00001057, 0.00000491, -0.00000767, 0.00000467, -0.00000057, -0.00000385, -0.00000147, 0.00001366, -0.00002091, 0.00000701, 0.00000209, -0.00000129, -0.00001136, 0.00001416, 0.00000244, -0.00001380, 0.00000053, 0.00000354, 0.00000108, -0.00000741, 0.00000646, 0.00000542, 0.00000036, 0.00000589, -0.00001846, 0.00001517, -0.00000626, 0.00000175, 0.00000888, -0.00001271, 0.00000969, -0.00001884, 0.00001043, -0.00001295, 0.00000345, -0.00000875, 0.00001619, -0.00002842, 0.00003514, -0.00003477, 0.00003109, -0.00002163, 0.00001435, -0.00001277, 0.00001593, -0.00002377, 0.00002125, -0.00001114, 0.00001562, -0.00001667, 0.00001704, -0.00002842, 0.00002502, -0.00001701, 0.00001987, -0.00003137, 0.00002959, -0.00004051, 0.00003121, -0.00004643, 0.00004914, -0.00006791, 0.00005908, -0.00011406, 0.00009167, -0.00035183, 0.00019464, 0.00073876, -0.00045971, 0.00006927, -0.00011294, 0.00004841, -0.00005993, 0.00004380, -0.00004410, 0.00002808, -0.00004187, 0.00002855, -0.00002843, 0.00003035, -0.00002310, 0.00001258, -0.00001661, 0.00001492, -0.00002259, 0.00001305, -0.00000857, 0.00000065, -0.00000449, 0.00001345, -0.00000876, 0.00000059, -0.00000290, -0.00000519, -0.00000588, 0.00000458, -0.00001848, 0.00002497, -0.00001911, 0.00002358, -0.00000741, 0.00000942, 0.00000573, 0.00001267, -0.00001280, 0.00000971, -0.00002186, 0.00002109, -0.00001778, 0.00001184, -0.00001185, 0.00000352, 0.00000039, -0.00000236, -0.00000221, 0.00000486, },
		{ 0.00000270, -0.00000408, -0.00000916, 0.00000145, 0.00001607, -0.00000448, 0.00000067, -0.00000668, 0.00000618, 0.00000267, 0.00000893, -0.00001523, -0.00000564, 0.00001386, -0.00000176, -0.00000389, 0.00000349, -0.00000041, -0.00000677, 0.00000612, 0.00000524, -0.00001869, 0.00001094, -0.00000151, -0.00000278, -0.00000822, 0.00001831, -0.00000809, -0.00000890, 0.00000191, 0.00000424, -0.00000231, -0.00000542, 0.00000946, 0.00000176, 0.00000388, -0.00000080, -0.00001424, 0.00001891, -0.00001595, 0.00001578, -0.00000758, 0.00000176, -0.00000376, -0.00000857, 0.00000318, -0.00000727, -0.00000304, 0.00000242, 0.00000134, -0.00001324, 0.00002348, -0.00002897, 0.00003178, -0.00002615, 0.00001934, -0.00001552, 0.00001584, -0.00002488, 0.00002687, -0.00001636, 0.00001813, -0.00001837, 0.00001683, -0.00002927, 0.00003068, -0.00002220, 0.00002021, -0.00003089, 0.00002908, -0.00004217, 0.00003375, -0.00004718, 0.00004981, -0.00007216, 0.00006095, -0.00012206, 0.00008916, -0.00044153, 0.00058625, 0.00035965, -0.00039536, 0.00008914, -0.00011398, 0.00005302, -0.00005808, 0.00004332, -0.00004461, 0.00002619, -0.00003882, 0.00002821, -0.00002694, 0.00003152, -0.00002834, 0.00001647, -0.00001785, 0.00001457, -0.00002386, 0.00001775, -0.00001425, 0.00000353, -0.00000247, 0.00001200, -0.00001087, 0.00000187, -0.00000253, -0.00000976, 0.00000290, -0.00000733, -0.00000543, 0.00001574, -0.00001331, 0.00002317, -0.00000841, 0.00001081, 0.00001052, 0.00000289, -0.00000367, -0.00000252, -0.00001074, 0.00001517, -0.00001685, 0.00001345, -0.00001527, 0.00000834, -0.00000349, -0.00000183, -0.00000015, },
		{ 0.00000243, -0.00000036, -0.00000389, -0.00001121, 0.00001183, 0.00000721, -0.00000131, -0.00000372, -0.00000094, 0.00000268, 0.00000731, 0.00000277, -0.00001575, -0.00000084, 0.00001532, -0.00000861, 0.00000222, -0.00000037, 0.00000119, -0.00000953, 0.00001410, -0.00000675, -0.00001007, 0.00000887, -0.00000190, -0.00000620, -0.00000050, 0.00001528, -0.00001442, -0.00000341, 0.00000256, 0.00000388, -0.00000524, -0.00000176, 0.00001037, -0.00000103, 0.00000721, -0.00000941, -0.00000345, 0.00001254, -0.00001482, 0.00002022, -0.00001772, 0.00001377, -0.00001818, 0.00000531, -0.00000866, 0.00000263, -0.00001235, 0.00001575, -0.00001736, 0.00000925, 0.00000048, -0.00000849, 0.00001662, -0.00001572, 0.00001122, -0.00000667, 0.00000371, -0.00001207, 0.00001848, -0.00000922, 0.00000883, -0.00000775, 0.00000290, -0.00001404, 0.00001974, -0.00001255, 0.00000566, -0.00001329, 0.00000844, -0.00002057, 0.00001081, -0.00001940, 0.00001637, -0.00003577, 0.00001183, -0.00006262, -0.00002546, -0.00033706, 0.00080648, -0.00002021, -0.00022948, 0.00003092, -0.00006427, 0.00001855, -0.00002482, 0.00001508, -0.00002153, 0.00000414, -0.00001674, 0.00000973, -0.00000787, 0.00001491, -0.00001778, 0.00000783, -0.00000850, 0.00000389, -0.00001431, 0.00001234, -0.00001245, 0.00000191, 0.00000350, 0.00000542, -0.00000856, 0.00000104, -0.00000243, -0.00001229, 0.00000949, -0.00001855, 0.00000966, 0.00000198, -0.00000146, 0.00001742, -0.00000642, 0.00001279, 0.00001194, -0.00000397, 0.00000346, -0.00001515, 0.00000520, 0.00000107, -0.00000676, 0.00000665, -0.00001155, 0.00000786, -0.00000462, -0.00000211, },
		{ -0.00000163, 0.00000346, -0.00000224, -0.00000588, -0.00000776, 0.00001658, 0.00000036, 0.00000093, -0.00000611, 0.00000318, 0.00000018, 0.00001107, -0.00000428, -0.00001434, 0.00000510, 0.00001243, -0.00001166, 0.00000658, -0.00000364, 0.00000167, -0.00000926, 0.00001828, -0.00001728, 0.00000073, 0.00000315, -0.00000029, -0.00000967, 0.00000917, 0.00000602, -0.00001462, 0.00000023, 0.00000354, 0.00000178, -0.00000633, 0.00000225, 0.00000933, -0.00000198, 0.00000818, -0.00001611, 0.00000929, -0.00000077, -0.00000306, 0.00001161, -0.00001456, 0.00001391, -0.00002310, 0.00001252, -0.00001569, 0.00000830, -0.00001626, 0.00002204, -0.00002893, 0.00002649, -0.00002060, 0.00001401, -0.00000343, 0.00000047, -0.00000198, 0.00000578, -0.00001186, 0.00000613, 0.00000375, 0.00000294, -0.00000427, 0.00000612, -0.00001469, 0.00000708, 0.00000173, 0.00000223, -0.00001227, 0.00000764, -0.00001654, 0.00000701, -0.00001878, 0.00001546, -0.00002626, 0.00001180, -0.00005229, 0.00001752, -0.00019664, -0.00006053, 0.00078869, -0.00027442, -0.00006828, -0.00004104, -0.00000692, -0.00002136, 0.00001186, -0.00001703, 0.00000473, -0.00001966, 0.00000719, -0.00001093, 0.00001409, -0.00000683, -0.00000145, -0.00000505, 0.00000365, -0.00000952, -0.00000081, 0.00000209, -0.00000692, -0.00000076, 0.00000934, -0.00000240, -0.00000450, -0.00000017, -0.00000432, -0.00001014, 0.00000960, -0.00002321, 0.00002013, -0.00000993, 0.00001152, 0.00000923, -0.00000273, 0.00001618, 0.00000780, -0.00000356, 0.00000176, -0.00001865, 0.00001513, -0.00001117, 0.00000453, -0.00000280, -0.00000463, 0.00000447, -0.00000449, },
		{ -0.00000434, 0.00000013, 0.00000238, -0.00000274, -0.00000922, 0.00000058, 0.00001414, -0.00000155, -0.00000025, -0.00000424, 0.00000299, 0.00000164, 0.00001080, -0.00000929, -0.00001241, 0.00001218, 0.00000458, -0.00000869, 0.00000614, -0.00000352, -0.00000103, -0.00000396, 0.00001538, -0.00002146, 0.00000846, -0.00000218, 0.00000058, -0.00001007, 0.00001615, -0.00000529, -0.00001008, 0.00000172, 0.00000459, -0.00000146, -0.00000537, 0.00000578, 0.00000667, -0.00000052, 0.00000505, -0.00001678, 0.00001721, -0.00001302, 0.00001189, -0.00000459, -0.00000049, 0.00000132, -0.00001501, 0.00000806, -0.00001254, 0.00000466, -0.00001001, 0.00001588, -0.00002628, 0.00002924, -0.00002858, 0.00002641, -0.00001651, 0.00001156, -0.00001045, 0.00001233, -0.00002065, 0.00001910, -0.00000744, 0.00001093, -0.00001179, 0.00001298, -0.00002438, 0.00002178, -0.00001166, 0.00001098, -0.00002146, 0.00001779, -0.00002938, 0.00002195, -0.00003431, 0.00003359, -0.00004992, 0.00003623, -0.00008720, 0.00005435, -0.00032649, 0.00028363, 0.00056989, -0.00035761, 0.00002297, -0.00007941, 0.00002428, -0.00003979, 0.00002904, -0.00003301, 0.00001696, -0.00002981, 0.00001840, -0.00001970, 0.00002508, -0.00002074, 0.00000973, -0.00001319, 0.00001056, -0.00001780, 0.00000918, -0.00000682, -0.00000197, -0.00000117, 0.00001154, -0.00000795, -0.00000104, -0.00000098, -0.00000785, -0.00000378, 0.00000276, -0.00001919, 0.00002237, -0.00001561, 0.00002183, 0.00000050, 0.00000327, 0.00001827, 0.00000132, 0.00000159, -0.00000789, -0.00001078, 0.00001368, -0.00001463, 0.00001023, -0.00000918, 0.00000114, 0.00000093, },
		{ -0.00000162, -0.00000404, 0.00000211, 0.00000048, -0.00000352, -0.00001073, 0.00000951, 0.00000757, 0.00000005, -0.00000352, 0.00000032, -0.00000041, 0.00000649, 0.00000639, -0.00001164, -0.00000955, 0.00001809, -0.00000550, -0.00000125, 0.00000123, -0.00000003, -0.00000636, 0.00000541, 0.00000548, -0.00001746, 0.00000990, -0.00000409, -0.00000115, -0.00000550, 0.00001661, -0.00001362, -0.00000412, 0.00000230, 0.00000461, -0.00000436, -0.00000308, 0.00000844, 0.00000307, 0.00000277, -0.00000195, -0.00000985, 0.00001551, -0.00001628, 0.00001995, -0.00001751, 0.00001437, -0.00001583, 0.00000038, -0.00000446, -0.00000163, -0.00000610, 0.00000385, -0.00000022, -0.00001089, 0.00001713, -0.00002082, 0.00002413, -0.00001718, 0.00001222, -0.00000949, 0.00000865, -0.00001770, 0.00002107, -0.00000967, 0.00001025, -0.00000966, 0.00000851, -0.00002078, 0.00002362, -0.00001441, 0.00000894, -0.00001766, 0.00001239, -0.00002408, 0.00001699, -0.00002760, 0.00002541, -0.00004294, 0.00002265, -0.00007222, 0.00000984, -0.00033005, 0.00056322, 0.00025864, -0.00030415, 0.00003674, -0.00007374, 0.00002213, -0.00003126, 0.00002185, -0.00002820, 0.00001134, -0.00002310, 0.00001367, -0.00001296, 0.00002044, -0.00002125, 0.00001064, -0.00001259, 0.00000886, -0.00001745, 0.00001177, -0.00001090, 0.00000030, 0.00000155, 0.00000907, -0.00000973, 0.00000087, -0.00000189, -0.00001088, 0.00000350, -0.00000778, -0.00000857, 0.00001667, -0.00001359, 0.00002717, -0.00000721, 0.00001218, 0.00001550, -0.00000176, 0.00000441, -0.00001789, 0.00000244, 0.00000372, -0.00000929, 0.00000865, -0.00001034, 0.00000380, },
		{ 0.00000392, -0.00000347, -0.00000275, 0.00000276, -0.00000062, -0.00000611, -0.00000748, 0.00001408, 0.00000192, 0.00000135, -0.00000529, 0.00000365, -0.00000317, 0.00001131, 0.00000039, -0.00001269, -0.00000414, 0.00001925, -0.00001291, 0.00000610, -0.00000457, 0.00000386, -0.00001100, 0.00001470, -0.00000739, -0.00000774, 0.00000562, -0.00000243, -0.00000485, 0.00000289, 0.00000989, -0.00001594, 0.00000037, 0.00000317, 0.00000281, -0.00000561, -0.00000035, 0.00000985, -0.00000021, 0.00000591, -0.00000989, 0.00000186, 0.00000490, -0.00000810, 0.00001526, -0.00001824, 0.00001914, -0.00002614, 0.00001249, -0.00001580, 0.00000889, -0.00001620, 0.00001707, -0.00001737, 0.00000834, -0.00000173, -0.00000381, 0.00001170, -0.00000853, 0.00000491, -0.00000178, -0.00000208, -0.00000577, 0.00001377, -0.00000448, 0.00000327, -0.00000119, -0.00000348, -0.00000721, 0.00001478, -0.00000873, -0.00000016, -0.00000626, -0.00000205, -0.00000764, -0.00000057, -0.00000689, 0.00000039, -0.00001640, -0.00001537, -0.00002471, -0.00010106, -0.00018281, 0.00068243, -0.00002528, -0.00018957, 0.00000568, -0.00004297, 0.00000189, -0.00000895, 0.00000167, -0.00001208, -0.00000347, -0.00000779, 0.00000073, 0.00000228, 0.00000557, -0.00001173, 0.00000358, -0.00000553, 0.00000093, -0.00001032, 0.00000777, -0.00001022, 0.00000046, 0.00000585, 0.00000338, -0.00000823, 0.00000134, -0.00000364, -0.00001122, 0.00000789, -0.00001723, 0.00000414, 0.00000612, -0.00000471, 0.00002618, -0.00001147, 0.00002220, 0.00000742, 0.00000150, -0.00000058, -0.00002006, 0.00001182, -0.00000691, -0.00000053, 0.00000291, -0.00000782, },
		{ -0.00000446, 0.00000294, -0.00000504, -0.00000030, 0.00000165, -0.00000074, -0.00000994, 0.00000043, 0.00001221, 0.00000039, -0.00000030, -0.00000311, 0.00000276, -0.00000195, 0.00001261, -0.00000435, -0.00001375, 0.00000445, 0.00001359, -0.00001366, 0.00000876, -0.00000705, 0.00000446, -0.00001101, 0.00001883, -0.00001732, 0.00000241, -0.00000075, 0.00000040, -0.00000772, 0.00001120, -0.00000096, -0.00001255, 0.00000245, 0.00000437, -0.00000039, -0.00000498, 0.00000239, 0.00000955, -0.00000169, 0.00000638, -0.00001433, 0.00001213, -0.00000819, 0.00000646, 0.00000041, -0.00000602, 0.00000970, -0.00002292, 0.00001339, -0.00001791, 0.00001129, -0.00001802, 0.00002135, -0.00002610, 0.00002096, -0.00001647, 0.00001204, -0.00000171, 0.00000155, -0.00000329, 0.00000565, -0.00001227, 0.00000756, 0.00000372, 0.00000224, -0.00000370, 0.00000632, -0.00001465, 0.00000781, 0.00000274, -0.00000148, -0.00000851, 0.00000317, -0.00001400, 0.00000684, -0.00001626, 0.00001134, -0.00002222, 0.00000642, -0.00004796, 0.00001240, -0.00021694, 0.00006684, 0.00062065, -0.00020402, -0.00008367, -0.00002991, -0.00001338, -0.00001560, 0.00000911, -0.00001563, 0.00000172, -0.00001528, 0.00000454, -0.00000919, 0.00001494, -0.00000940, -0.00000026, -0.00000507, 0.00000233, -0.00000783, -0.00000130, 0.00000081, -0.00000704, 0.00000039, 0.00000918, -0.00000292, -0.00000494, 0.00000069, -0.00000617, -0.00000825, 0.00000722, -0.00002181, 0.00001441, -0.00000510, 0.00000868, 0.00001839, -0.00000915, 0.00002865, -0.00000173, 0.00000817, -0.00001242, -0.00001188, 0.00001170, -0.00001146, 0.00000561, -0.00000256, },
		{ -0.00000527, -0.00000203, 0.00000156, -0.00000583, 0.00000214, -0.00000002, -0.00000140, -0.00001209, 0.00000934, 0.00000617, 0.00000220, -0.00000409, 0.00000202, -0.00000188, 0.00000349, 0.00000949, -0.00000685, -0.00001431, 0.00001418, 0.00000264, -0.00000751, 0.00000521, -0.00000441, 0.00000029, -0.00000480, 0.00001496, -0.00001991, 0.00000804, -0.00000511, 0.00000157, -0.00000689, 0.00001495, -0.00001094, -0.00000632, 0.00000302, 0.00000485, -0.00000352, -0.00000324, 0.00000511, 0.00000735, -0.00000070, 0.00000285, -0.00001209, 0.00001492, -0.00001534, 0.00001775, -0.00001504, 0.00001105, -0.00000851, -0.00000853, 0.00000310, -0.00000951, 0.00000378, -0.00000952, 0.00001392, -0.00002233, 0.00002141, -0.00002023, 0.00001920, -0.00000879, 0.00000646, -0.00000649, 0.00000711, -0.00001552, 0.00001540, -0.00000291, 0.00000535, -0.00000571, 0.00000731, -0.00001822, 0.00001673, -0.00000573, 0.00000185, -0.00001084, 0.00000454, -0.00001594, 0.00001034, -0.00001990, 0.00001539, -0.00002862, 0.00000862, -0.00005246, 0.00000123, -0.00026533, 0.00032492, 0.00042920, -0.00026237, -0.00002155, -0.00004671, 0.00000030, -0.00001911, 0.00001315, -0.00002104, 0.00000551, -0.00001735, 0.00000746, -0.00000952, 0.00001770, -0.00001672, 0.00000629, -0.00000979, 0.00000623, -0.00001277, 0.00000506, -0.00000534, -0.00000365, 0.00000149, 0.00000968, -0.00000754, -0.00000141, -0.00000092, -0.00000849, -0.00000325, 0.00000181, -0.00002012, 0.00001914, -0.00001287, 0.00002286, 0.00000540, 0.00000133, 0.00002689, -0.00000565, 0.00001112, -0.00002390, 0.00000163, 0.00000351, -0.00000852, 0.00000654, },
		{ 0.00000381, -0.00000560, -0.00000050, -0.00000044, -0.00000490, 0.00000295, -0.00000062, -0.00000428, -0.00000943, 0.00001430, 0.00000073, 0.00000377, -0.00000654, 0.00000625, -0.00000658, 0.00001008, 0.00000397, -0.00000836, -0.00001219, 0.00002079, -0.00000866, 0.00000162, -0.00000190, 0.00000134, -0.00000656, 0.00000534, 0.00000450, -0.00001476, 0.00000729, -0.00000543, 0.00000002, -0.00000173, 0.00001189, -0.00001601, -0.00000056, 0.00000345, 0.00000362, -0.00000515, -0.00000129, 0.00000763, 0.00000392, 0.00000191, -0.00000343, -0.00000374, 0.00000842, -0.00001170, 0.00001793, -0.00002079, 0.00002144, -0.00002430, 0.00000749, -0.00001090, 0.00000342, -0.00000851, 0.00000448, -0.00000085, -0.00000941, 0.00001168, -0.00001343, 0.00001677, -0.00000814, 0.00000489, -0.00000380, 0.00000199, -0.00001058, 0.00001557, -0.00000412, 0.00000385, -0.00000249, 0.00000156, -0.00001304, 0.00001721, -0.00000854, 0.00000068, -0.00000787, -0.00000084, -0.00000902, 0.00000353, -0.00001212, 0.00000597, -0.00001981, -0.00000830, -0.00003073, -0.00006349, -0.00020446, 0.00050025, 0.00019481, -0.00023381, -0.00000215, -0.00004391, -0.00000033, -0.00001114, 0.00000545, -0.00001605, 0.00000077, -0.00001142, 0.00000269, -0.00000163, 0.00001059, -0.00001482, 0.00000588, -0.00000889, 0.00000479, -0.00001231, 0.00000670, -0.00000859, -0.00000121, 0.00000391, 0.00000699, -0.00000925, 0.00000112, -0.00000320, -0.00000949, 0.00000163, -0.00000622, -0.00001293, 0.00001715, -0.00001378, 0.00003292, -0.00000820, 0.00001743, 0.00001624, -0.00000103, 0.00000466, -0.00002676, 0.00001177, -0.00000613, -0.00000188, },
		{ 0.00000326, 0.00000069, -0.00000530, 0.00000078, -0.00000313, -0.00000209, 0.00000163, 0.00000016, -0.00000911, -0.00000146, 0.00001264, -0.00000046, 0.00000202, -0.00000463, 0.00000587, -0.00000721, 0.00001381, -0.00000087, -0.00001061, -0.00000550, 0.00002039, -0.00001475, 0.00000812, -0.00000767, 0.00000588, -0.00001161, 0.00001435, -0.00000759, -0.00000546, 0.00000205, -0.00000265, -0.00000280, 0.00000544, 0.00000334, -0.00001500, 0.00000294, 0.00000426, 0.00000076, -0.00000481, 0.00000048, 0.00000931, 0.00000070, 0.00000395, -0.00000892, 0.00000612, -0.00000328, 0.00000059, 0.00000635, -0.00001280, 0.00001785, -0.00002816, 0.00001514, -0.00001912, 0.00001218, -0.00001739, 0.00001579, -0.00001473, 0.00000482, -0.00000113, -0.00000205, 0.00000939, -0.00000352, 0.00000034, 0.00000108, -0.00000558, -0.00000123, 0.00001065, -0.00000203, 0.00000053, 0.00000218, -0.00000655, -0.00000312, 0.00001191, -0.00000781, -0.00000181, -0.00000441, -0.00000663, -0.00000061, -0.00000563, -0.00000183, -0.00000657, -0.00000847, -0.00002745, -0.00000775, -0.00014842, -0.00004599, 0.00054547, -0.00000518, -0.00016824, -0.00000587, -0.00003296, -0.00000501, -0.00000148, -0.00000501, -0.00000819, -0.00000576, -0.00000408, -0.00000313, 0.00000782, -0.00000063, -0.00000756, 0.00000094, -0.00000439, -0.00000009, -0.00000796, 0.00000425, -0.00000878, 0.00000035, 0.00000666, 0.00000208, -0.00000798, 0.00000179, -0.00000547, -0.00000882, 0.00000459, -0.00001424, -0.00000251, 0.00000924, -0.00000607, 0.00003415, -0.00001570, 0.00003228, 0.00000184, 0.00000905, -0.00001052, -0.00001794, 0.00001246, -0.00001066, },
		{ -0.00000810, 0.00000425, -0.00000082, -0.00000533, 0.00000180, -0.00000558, 0.00000127, -0.00000058, 0.00000072, -0.00001307, 0.00000860, 0.00000596, 0.00000242, -0.00000285, 0.00000142, 0.00000035, -0.00000235, 0.00001265, -0.00000336, -0.00001369, 0.00000501, 0.00001235, -0.00001294, 0.00000829, -0.00000829, 0.00000562, -0.00001104, 0.00001739, -0.00001568, 0.00000272, -0.00000375, 0.00000056, -0.00000422, 0.00001065, -0.00000670, -0.00000972, 0.00000432, 0.00000476, -0.00000249, -0.00000321, 0.00000236, 0.00000935, -0.00000088, 0.00000340, -0.00000993, 0.00001144, -0.00001251, 0.00001344, -0.00000977, 0.00000397, 0.00000197, -0.00001844, 0.00001044, -0.00001655, 0.00001136, -0.00001724, 0.00001831, -0.00002098, 0.00001326, -0.00000993, 0.00000729, 0.00000279, 0.00000010, -0.00000263, 0.00000350, -0.00001034, 0.00000718, 0.00000495, -0.00000002, -0.00000114, 0.00000409, -0.00001175, 0.00000607, 0.00000511, -0.00000653, -0.00000264, -0.00000435, -0.00000745, 0.00000253, -0.00000950, 0.00000233, -0.00001218, -0.00000699, -0.00003180, -0.00001253, -0.00020095, 0.00015324, 0.00046990, -0.00013112, -0.00010394, -0.00001451, -0.00002388, -0.00000573, 0.00000228, -0.00001091, -0.00000373, -0.00000855, -0.00000071, -0.00000438, 0.00001240, -0.00000936, -0.00000043, -0.00000438, 0.00000032, -0.00000527, -0.00000258, -0.00000021, -0.00000686, 0.00000160, 0.00000843, -0.00000336, -0.00000460, 0.00000035, -0.00000690, -0.00000693, 0.00000478, -0.00001990, 0.00000801, -0.00000166, 0.00000903, 0.00002459, -0.00001174, 0.00003808, -0.00000799, 0.00001637, -0.00002679, -0.00000203, 0.00000430, },
		{ -0.00000636, -0.00000162, 0.00000222, -0.00000116, -0.00000521, 0.00000168, -0.00000624, 0.00000310, -0.00000162, -0.00000118, -0.00001242, 0.00001546, -0.00000089, 0.00000578, -0.00000736, 0.00000787, -0.00000699, 0.00000568, 0.00000780, -0.00000437, -0.00001542, 0.00001553, 0.00000021, -0.00000512, 0.00000262, -0.00000365, 0.00000034, -0.00000434, 0.00001281, -0.00001667, 0.00000599, -0.00000678, 0.00000202, -0.00000249, 0.00001078, -0.00001385, -0.00000329, 0.00000476, 0.00000395, -0.00000458, -0.00000144, 0.00000460, 0.00000752, -0.00000049, 0.00000002, -0.00000530, 0.00000873, -0.00001298, 0.00001814, -0.00002004, 0.00001875, -0.00001679, -0.00000190, -0.00000245, -0.00000549, 0.00000209, -0.00000842, 0.00001152, -0.00001807, 0.00001356, -0.00001173, 0.00001138, -0.00000020, 0.00000041, -0.00000207, 0.00000153, -0.00000962, 0.00001132, 0.00000122, 0.00000031, -0.00000005, 0.00000178, -0.00001161, 0.00001125, -0.00000065, -0.00000558, -0.00000202, -0.00000722, -0.00000330, -0.00000044, -0.00000686, -0.00000101, -0.00000953, -0.00001668, -0.00001914, -0.00005544, -0.00018072, 0.00032280, 0.00032273, -0.00018314, -0.00005688, -0.00002072, -0.00001854, -0.00000212, -0.00000035, -0.00001104, -0.00000368, -0.00000706, -0.00000174, -0.00000021, 0.00001000, -0.00001213, 0.00000300, -0.00000714, 0.00000278, -0.00000836, 0.00000134, -0.00000450, -0.00000405, 0.00000304, 0.00000819, -0.00000750, -0.00000072, -0.00000255, -0.00000711, -0.00000449, 0.00000204, -0.00002137, 0.00001516, -0.00001070, 0.00002652, 0.00000735, 0.00000390, 0.00003107, -0.00000728, 0.00001312, -0.00003439, 0.00001163, },
		{ 0.00001536, -0.00001226, 0.00000356, 0.00000023, -0.00000198, -0.00000396, -0.00000023, -0.00000428, 0.00000225, -0.00000054, -0.00000616, -0.00000550, 0.00001531, -0.00000332, 0.00000561, -0.00000746, 0.00001002, -0.00001121, 0.00001246, 0.00000245, -0.00000604, -0.00001289, 0.00002109, -0.00001010, 0.00000349, -0.00000473, 0.00000281, -0.00000669, 0.00000523, 0.00000308, -0.00001144, 0.00000387, -0.00000606, 0.00000146, 0.00000181, 0.00000548, -0.00001572, 0.00000184, 0.00000509, 0.00000156, -0.00000472, -0.00000011, 0.00000695, 0.00000457, 0.00000077, -0.00000412, 0.00000229, -0.00000035, -0.00000410, 0.00001117, -0.00001769, 0.00002202, -0.00002760, 0.00001114, -0.00001467, 0.00000663, -0.00000936, 0.00000362, 0.00000001, -0.00000949, 0.00000823, -0.00000821, 0.00001124, -0.00000059, -0.00000129, 0.00000045, -0.00000306, -0.00000458, 0.00001123, -0.00000050, -0.00000017, 0.00000209, -0.00000281, -0.00000729, 0.00001233, -0.00000518, -0.00000380, -0.00000270, -0.00000899, 0.00000118, -0.00000481, -0.00000274, -0.00000602, -0.00000608, -0.00002710, -0.00000511, -0.00011801, -0.00008307, 0.00041318, 0.00016162, -0.00018390, -0.00002669, -0.00002552, -0.00001317, 0.00000099, -0.00000477, -0.00000895, -0.00000491, -0.00000456, -0.00000371, 0.00000583, 0.00000311, -0.00000965, 0.00000232, -0.00000674, 0.00000242, -0.00000862, 0.00000261, -0.00000712, -0.00000130, 0.00000477, 0.00000560, -0.00000895, 0.00000176, -0.00000545, -0.00000664, -0.00000186, -0.00000339, -0.00001767, 0.00001609, -0.00001256, 0.00003890, -0.00000959, 0.00002477, 0.00001469, 0.00000295, -0.00000213, -0.00002809, },
		{ -0.00001103, 0.00000811, -0.00001026, 0.00000434, 0.00000024, -0.00000413, -0.00000159, -0.00000326, -0.00000064, -0.00000039, 0.00000137, -0.00001189, 0.00000542, 0.00000854, -0.00000028, 0.00000085, -0.00000194, 0.00000568, -0.00000919, 0.00001458, -0.00000082, -0.00000953, -0.00000501, 0.00001873, -0.00001384, 0.00000773, -0.00000890, 0.00000659, -0.00001091, 0.00001264, -0.00000685, -0.00000389, -0.00000118, -0.00000295, 0.00000034, 0.00000607, -0.00000278, -0.00001270, 0.00000487, 0.00000515, -0.00000152, -0.00000334, 0.00000098, 0.00000857, 0.00000184, 0.00000114, -0.00000585, 0.00000778, -0.00000971, 0.00000839, -0.00000330, -0.00000411, 0.00001158, -0.00002501, 0.00001354, -0.00001872, 0.00001254, -0.00001612, 0.00001235, -0.00001013, -0.00000055, 0.00000178, -0.00000302, 0.00000978, -0.00000103, -0.00000219, 0.00000181, -0.00000672, 0.00000150, 0.00000901, -0.00000155, 0.00000026, 0.00000273, -0.00000648, -0.00000172, 0.00001081, -0.00000893, -0.00000040, -0.00000664, -0.00000620, 0.00000129, -0.00000541, -0.00000284, -0.00000611, -0.00001025, -0.00002593, -0.00001152, -0.00016269, 0.00005951, 0.00041366, 0.00002686, -0.00015804, -0.00000862, -0.00003050, -0.00000497, -0.00000030, -0.00000665, -0.00000820, -0.00000435, -0.00000422, -0.00000291, 0.00000929, -0.00000383, -0.00000504, -0.00000065, -0.00000432, 0.00000014, -0.00000674, 0.00000138, -0.00000751, 0.00000084, 0.00000650, 0.00000113, -0.00000729, 0.00000154, -0.00000683, -0.00000639, 0.00000083, -0.00001068, -0.00000925, 0.00001012, -0.00000414, 0.00003964, -0.00001697, 0.00004067, -0.00000189, 0.00001465, -0.00002258, },
		{ -0.00003711, 0.00000716, -0.00000426, -0.00000295, 0.00000158, 0.00000163, -0.00000666, 0.00000058, -0.00000552, 0.00000233, -0.00000245, 0.00000141, -0.00001422, 0.00001496, -0.00000030, 0.00000510, -0.00000561, 0.00000633, -0.00000307, -0.00000170, 0.00001175, -0.00000184, -0.00001356, 0.00000596, 0.00000960, -0.00001024, 0.00000566, -0.00000753, 0.00000522, -0.00000952, 0.00001422, -0.00001256, 0.00000166, -0.00000576, 0.00000037, 0.00000037, 0.00000745, -0.00001030, -0.00000703, 0.00000632, 0.00000422, -0.00000386, -0.00000153, 0.00000237, 0.00000872, 0.00000024, -0.00000023, -0.00000338, 0.00000710, -0.00001281, 0.00001611, -0.00001607, 0.00001212, -0.00000634, -0.00001195, 0.00000540, -0.00001300, 0.00000960, -0.00001498, 0.00001413, -0.00001506, 0.00000517, -0.00000271, 0.00000132, 0.00000873, -0.00000290, -0.00000083, 0.00000026, -0.00000701, 0.00000585, 0.00000647, -0.00000264, 0.00000208, 0.00000088, -0.00000739, 0.00000300, 0.00000761, -0.00001114, 0.00000309, -0.00001233, 0.00000043, -0.00000322, -0.00000159, -0.00000779, -0.00000081, -0.00002260, -0.00001114, -0.00004836, -0.00015580, 0.00019692, 0.00034704, -0.00006701, -0.00012094, -0.00000096, -0.00003337, 0.00000403, -0.00000522, -0.00000562, -0.00000912, -0.00000198, -0.00000617, 0.00000137, 0.00000806, -0.00000764, -0.00000131, -0.00000371, -0.00000149, -0.00000270, -0.00000402, -0.00000138, -0.00000594, 0.00000231, 0.00000755, -0.00000386, -0.00000348, -0.00000130, -0.00000616, -0.00000675, 0.00000294, -0.00001799, 0.00000132, 0.00000003, 0.00001291, 0.00002735, -0.00000963, 0.00004342, -0.00000945, 0.00001782, },
		{ 0.00000740, -0.00003736, 0.00001659, -0.00001336, 0.00000400, -0.00000129, 0.00000212, -0.00000772, 0.00000093, -0.00000527, 0.00000266, -0.00000208, -0.00000223, -0.00001003, 0.00001800, -0.00000558, 0.00000792, -0.00000895, 0.00001206, -0.00001113, 0.00000725, 0.00000657, -0.00000253, -0.00001507, 0.00001534, -0.00000157, -0.00000270, -0.00000052, -0.00000206, -0.00000025, -0.00000321, 0.00000967, -0.00001249, 0.00000292, -0.00000760, 0.00000257, 0.00000196, 0.00000471, -0.00001422, -0.00000113, 0.00000685, 0.00000199, -0.00000446, -0.00000018, 0.00000428, 0.00000737, -0.00000038, -0.00000240, 0.00000203, 0.00000021, -0.00000712, 0.00001367, -0.00001897, 0.00002089, -0.00002139, 0.00000290, -0.00000686, -0.00000194, 0.00000071, -0.00000766, 0.00000962, -0.00001462, 0.00000720, -0.00000481, 0.00000473, 0.00000753, -0.00000515, 0.00000165, -0.00000324, -0.00000413, 0.00000783, 0.00000398, -0.00000306, 0.00000393, -0.00000218, -0.00000604, 0.00000653, 0.00000260, -0.00001003, 0.00000329, -0.00001535, 0.00000618, -0.00000796, 0.00000216, -0.00001237, 0.00000318, -0.00003398, 0.00000456, -0.00010312, -0.00008837, 0.00028822, 0.00024918, -0.00012332, -0.00007965, -0.00000428, -0.00002965, 0.00000876, -0.00000945, -0.00000458, -0.00000926, -0.00000040, -0.00000765, 0.00000659, 0.00000342, -0.00000782, 0.00000019, -0.00000550, 0.00000058, -0.00000499, -0.00000172, -0.00000427, -0.00000327, 0.00000342, 0.00000709, -0.00000753, 0.00000037, -0.00000510, -0.00000445, -0.00000699, 0.00000307, -0.00002255, 0.00000985, -0.00000834, 0.00003182, 0.00000756, 0.00000988, 0.00003201, -0.00000488, },
		{ 0.00000659, -0.00001288, -0.00002321, 0.00001325, -0.00001400, 0.00000619, -0.00000145, 0.00000008, -0.00000638, -0.00000085, -0.00000250, 0.00000035, 0.00000035, -0.00000844, 0.00000005, 0.00001304, -0.00000443, 0.00000533, -0.00000607, 0.00001119, -0.00001375, 0.00001316, 0.00000224, -0.00000492, -0.00001159, 0.00001878, -0.00000932, 0.00000362, -0.00000636, 0.00000361, -0.00000618, 0.00000450, 0.00000189, -0.00000828, 0.00000033, -0.00000640, 0.00000365, 0.00000393, -0.00000107, -0.00001379, 0.00000356, 0.00000659, -0.00000093, -0.00000344, 0.00000062, 0.00000625, 0.00000523, -0.00000098, -0.00000312, 0.00000651, -0.00000850, 0.00000393, 0.00000235, -0.00001020, 0.00001696, -0.00002574, 0.00001098, -0.00001533, 0.00000756, -0.00000843, 0.00000129, 0.00000221, -0.00001100, 0.00000698, -0.00000563, 0.00000826, 0.00000457, -0.00000554, 0.00000275, -0.00000601, -0.00000017, 0.00000833, 0.00000106, -0.00000169, 0.00000393, -0.00000426, -0.00000404, 0.00000922, -0.00000413, -0.00000483, -0.00000196, -0.00001195, 0.00000623, -0.00000790, 0.00000055, -0.00001053, -0.00000204, -0.00003278, 0.00000175, -0.00014570, 0.00001864, 0.00031889, 0.00014882, -0.00014995, -0.00003948, -0.00001668, -0.00001770, 0.00000597, -0.00000943, -0.00000618, -0.00000646, -0.00000195, -0.00000580, 0.00000931, -0.00000163, -0.00000589, -0.00000029, -0.00000568, 0.00000139, -0.00000616, -0.00000072, -0.00000603, -0.00000059, 0.00000459, 0.00000458, -0.00000834, 0.00000197, -0.00000752, -0.00000356, -0.00000577, -0.00000029, -0.00002184, 0.00001248, -0.00000875, 0.00004359, -0.00000941, 0.00003212, 0.00001340, },
		{ -0.00000196, 0.00001481, -0.00003248, -0.00000282, 0.00000112, -0.00000707, 0.00000340, 0.00000109, -0.00000397, -0.00000352, -0.00000333, 0.00000084, -0.00000251, 0.00000228, -0.00001338, 0.00001146, 0.00000341, 0.00000142, -0.00000131, 0.00000195, 0.00000370, -0.00000949, 0.00001392, 0.00000024, -0.00000906, -0.00000334, 0.00001498, -0.00001070, 0.00000517, -0.00000822, 0.00000589, -0.00000890, 0.00000959, -0.00000511, -0.00000327, -0.00000380, -0.00000328, 0.00000435, 0.00000444, -0.00000735, -0.00001007, 0.00000672, 0.00000522, -0.00000330, -0.00000171, 0.00000137, 0.00000757, 0.00000312, -0.00000226, -0.00000074, 0.00000617, -0.00001294, 0.00001313, -0.00001068, 0.00000459, 0.00000296, -0.00001855, 0.00000902, -0.00001560, 0.00001073, -0.00001321, 0.00000772, -0.00000477, -0.00000639, 0.00000574, -0.00000571, 0.00001209, -0.00000060, -0.00000302, 0.00000092, -0.00000602, 0.00000286, 0.00000821, -0.00000219, 0.00000144, 0.00000153, -0.00000428, -0.00000227, 0.00001062, -0.00001071, 0.00000230, -0.00001099, -0.00000270, -0.00000000, -0.00000216, -0.00000716, -0.00000160, -0.00001757, -0.00001578, -0.00003064, -0.00014616, 0.00012718, 0.00030032, 0.00005909, -0.00015141, -0.00000806, -0.00003116, -0.00000167, -0.00000251, -0.00000559, -0.00000997, -0.00000141, -0.00000621, -0.00000029, 0.00000788, -0.00000466, -0.00000374, -0.00000180, -0.00000454, 0.00000088, -0.00000612, -0.00000124, -0.00000596, 0.00000129, 0.00000589, 0.00000032, -0.00000602, 0.00000027, -0.00000709, -0.00000479, -0.00000245, -0.00000736, -0.00001541, 0.00000821, 0.00000161, 0.00004195, -0.00001409, 0.00004601, },
		{ 0.00004528, -0.00000680, 0.00001158, -0.00004066, 0.00001259, -0.00001117, 0.00000188, -0.00000099, 0.00000380, -0.00000778, -0.00000115, -0.00000441, 0.00000232, -0.00000339, 0.00000094, -0.00001286, 0.00001828, -0.00000493, 0.00000705, -0.00000738, 0.00001036, -0.00000637, -0.00000088, 0.00001064, -0.00000038, -0.00001274, 0.00000640, 0.00000649, -0.00000673, 0.00000190, -0.00000548, 0.00000379, -0.00000706, 0.00000997, -0.00000857, -0.00000034, -0.00000703, 0.00000040, 0.00000502, 0.00000241, -0.00001169, -0.00000482, 0.00000847, 0.00000264, -0.00000415, -0.00000026, 0.00000251, 0.00000781, 0.00000123, -0.00000362, 0.00000407, 0.00000009, -0.00000955, 0.00001425, -0.00001713, 0.00001612, -0.00001272, -0.00000565, 0.00000011, -0.00000877, 0.00000722, -0.00001236, 0.00000994, -0.00000969, -0.00000169, 0.00000344, -0.00000415, 0.00001451, -0.00000617, 0.00000108, -0.00000300, -0.00000332, 0.00000447, 0.00000733, -0.00000450, 0.00000463, -0.00000181, -0.00000310, -0.00000016, 0.00000915, -0.00001388, 0.00000676, -0.00001849, 0.00000720, -0.00000792, 0.00000467, -0.00001563, 0.00000765, -0.00003476, 0.00000623, -0.00008412, -0.00009360, 0.00020372, 0.00025538, -0.00001798, -0.00012975, 0.00000684, -0.00003845, 0.00001066, -0.00001087, -0.00000177, -0.00001268, 0.00000263, -0.00001001, 0.00000626, 0.00000343, -0.00000516, -0.00000245, -0.00000346, -0.00000258, -0.00000065, -0.00000525, -0.00000278, -0.00000422, 0.00000228, 0.00000675, -0.00000426, -0.00000202, -0.00000377, -0.00000433, -0.00000760, 0.00000177, -0.00001613, -0.00000578, 0.00000032, 0.00001969, 0.00002745, -0.00000337, },
		{ 0.00001750, 0.00003193, -0.00000159, -0.00000368, -0.00003328, 0.00001574, -0.00001619, 0.00000704, -0.00000306, 0.00000393, -0.00000924, -0.00000069, -0.00000305, 0.00000093, -0.00000150, -0.00000412, -0.00000552, 0.00001698, -0.00000732, 0.00000800, -0.00000830, 0.00001386, -0.00001353, 0.00000755, 0.00000628, -0.00000167, -0.00001308, 0.00001324, -0.00000199, -0.00000122, -0.00000318, -0.00000049, -0.00000081, -0.00000212, 0.00000635, -0.00000820, -0.00000050, -0.00000798, 0.00000380, 0.00000521, -0.00000174, -0.00001293, 0.00000054, 0.00000873, -0.00000052, -0.00000341, 0.00000059, 0.00000401, 0.00000714, -0.00000083, -0.00000338, 0.00000815, -0.00000856, 0.00000009, 0.00000635, -0.00001303, 0.00001753, -0.00002157, 0.00000488, -0.00000918, 0.00000029, 0.00000024, -0.00000768, 0.00000857, -0.00001244, 0.00000305, -0.00000042, 0.00000035, 0.00001319, -0.00000917, 0.00000388, -0.00000642, 0.00000024, 0.00000546, 0.00000497, -0.00000433, 0.00000567, -0.00000380, -0.00000244, 0.00000324, 0.00000372, -0.00001120, 0.00000455, -0.00001893, 0.00001140, -0.00001124, 0.00000612, -0.00001744, 0.00000782, -0.00004057, 0.00001377, -0.00013220, -0.00000355, 0.00023541, 0.00020199, -0.00008150, -0.00009030, 0.00000283, -0.00003295, 0.00001322, -0.00001377, -0.00000181, -0.00001121, 0.00000249, -0.00000986, 0.00001021, -0.00000127, -0.00000426, -0.00000210, -0.00000469, -0.00000046, -0.00000267, -0.00000419, -0.00000432, -0.00000176, 0.00000292, 0.00000622, -0.00000722, 0.00000112, -0.00000757, -0.00000156, -0.00000980, 0.00000418, -0.00002301, 0.00000254, -0.00000492, 0.00003748, 0.00000778, },
		{ -0.00000635, 0.00003801, 0.00001431, 0.00000627, -0.00002343, -0.00001516, 0.00000726, -0.00001215, 0.00000623, -0.00000135, 0.00000071, -0.00000798, -0.00000181, -0.00000026, -0.00000205, 0.00000141, -0.00001017, 0.00000574, 0.00000882, -0.00000331, 0.00000344, -0.00000286, 0.00001006, -0.00001425, 0.00001223, 0.00000321, -0.00000474, -0.00000881, 0.00001440, -0.00000660, 0.00000197, -0.00000648, 0.00000340, -0.00000481, 0.00000295, 0.00000127, -0.00000573, -0.00000292, -0.00000652, 0.00000659, 0.00000402, -0.00000643, -0.00001128, 0.00000522, 0.00000726, -0.00000307, -0.00000183, 0.00000112, 0.00000543, 0.00000593, -0.00000333, -0.00000014, 0.00000784, -0.00001423, 0.00001004, -0.00000555, -0.00000126, 0.00000845, -0.00002006, 0.00000780, -0.00001341, 0.00000660, -0.00000620, -0.00000191, 0.00000498, -0.00001312, 0.00000726, -0.00000524, 0.00000760, 0.00000729, -0.00000778, 0.00000319, -0.00000697, 0.00000273, 0.00000656, 0.00000110, -0.00000133, 0.00000367, -0.00000331, -0.00000302, 0.00000750, -0.00000450, -0.00000358, -0.00000439, -0.00001085, 0.00000719, -0.00000703, -0.00000061, -0.00000958, -0.00000520, -0.00002804, -0.00000791, -0.00014493, 0.00009092, 0.00023167, 0.00014565, -0.00012580, -0.00004484, -0.00001391, -0.00001672, 0.00000593, -0.00001021, -0.00000618, -0.00000548, -0.00000214, -0.00000477, 0.00000952, -0.00000388, -0.00000338, -0.00000236, -0.00000510, 0.00000111, -0.00000452, -0.00000361, -0.00000484, 0.00000028, 0.00000387, 0.00000369, -0.00000718, 0.00000128, -0.00000864, -0.00000128, -0.00000906, 0.00000222, -0.00002479, 0.00000580, -0.00000187, 0.00004622, },
		{ 0.00004160, -0.00000724, 0.00004853, 0.00000144, 0.00000792, -0.00003690, 0.00000329, -0.00000569, -0.00000306, 0.00000169, 0.00000238, -0.00000422, -0.00000546, -0.00000280, 0.00000168, -0.00000402, 0.00000231, -0.00001285, 0.00001528, -0.00000107, 0.00000330, -0.00000334, 0.00000587, 0.00000094, -0.00000883, 0.00001243, 0.00000177, -0.00000861, -0.00000140, 0.00001028, -0.00000642, 0.00000131, -0.00000618, 0.00000416, -0.00000604, 0.00000568, -0.00000271, -0.00000343, -0.00000590, -0.00000320, 0.00000838, 0.00000105, -0.00000999, -0.00000764, 0.00000871, 0.00000418, -0.00000408, -0.00000036, 0.00000177, 0.00000640, 0.00000421, -0.00000555, 0.00000557, 0.00000170, -0.00001273, 0.00001389, -0.00001384, 0.00001060, -0.00000528, -0.00001104, 0.00000334, -0.00001120, 0.00000793, -0.00000974, 0.00000305, 0.00000006, -0.00001123, 0.00000927, -0.00000867, 0.00001503, -0.00000127, -0.00000292, -0.00000076, -0.00000430, 0.00000357, 0.00000745, -0.00000290, 0.00000289, -0.00000006, -0.00000132, -0.00000370, 0.00001034, -0.00001179, 0.00000447, -0.00001531, 0.00000164, -0.00000224, 0.00000158, -0.00001171, 0.00000309, -0.00002542, -0.00000371, -0.00005636, -0.00010706, 0.00015814, 0.00021232, 0.00008357, -0.00014273, -0.00000847, -0.00003121, 0.00000169, -0.00000546, -0.00000393, -0.00001162, 0.00000120, -0.00000819, 0.00000295, 0.00000501, -0.00000391, -0.00000323, -0.00000294, -0.00000446, 0.00000155, -0.00000573, -0.00000374, -0.00000391, 0.00000130, 0.00000516, -0.00000035, -0.00000440, -0.00000181, -0.00000628, -0.00000428, -0.00000483, -0.00000461, -0.00002085, 0.00000368, 0.00001068, },
		{ 0.00002817, 0.00002651, 0.00000553, 0.00004558, -0.00000275, -0.00000069, -0.00003703, 0.00001274, -0.00001462, 0.00000479, -0.00000235, 0.00000494, -0.00000827, -0.00000350, -0.00000221, 0.00000112, -0.00000326, -0.00000057, -0.00000921, 0.00001817, -0.00000703, 0.00000741, -0.00000738, 0.00001259, -0.00000863, -0.00000062, 0.00000990, 0.00000060, -0.00001093, 0.00000575, 0.00000396, -0.00000352, -0.00000186, -0.00000306, 0.00000208, -0.00000442, 0.00000547, -0.00000445, -0.00000273, -0.00000792, 0.00000121, 0.00000853, -0.00000311, -0.00001154, -0.00000288, 0.00001039, 0.00000036, -0.00000342, 0.00000055, 0.00000269, 0.00000688, 0.00000163, -0.00000575, 0.00001080, -0.00000786, -0.00000427, 0.00000913, -0.00001333, 0.00001522, -0.00001574, -0.00000107, -0.00000407, -0.00000506, 0.00000515, -0.00001015, 0.00000650, -0.00000568, -0.00000626, 0.00000735, -0.00000783, 0.00001880, -0.00000857, 0.00000223, -0.00000542, -0.00000011, 0.00000366, 0.00000698, -0.00000493, 0.00000573, -0.00000295, -0.00000007, -0.00000248, 0.00000912, -0.00001404, 0.00000736, -0.00002141, 0.00001128, -0.00001003, 0.00000763, -0.00001920, 0.00001061, -0.00003973, 0.00001441, -0.00010983, -0.00002817, 0.00018441, 0.00019212, 0.00001471, -0.00012903, 0.00000777, -0.00003791, 0.00001289, -0.00001353, -0.00000020, -0.00001374, 0.00000455, -0.00001133, 0.00000921, -0.00000043, -0.00000257, -0.00000364, -0.00000366, -0.00000284, 0.00000068, -0.00000622, -0.00000420, -0.00000199, 0.00000160, 0.00000602, -0.00000432, -0.00000083, -0.00000625, -0.00000220, -0.00000889, 0.00000090, -0.00001405, -0.00001364, -0.00000012, },
		{ -0.00000004, 0.00004258, 0.00000947, 0.00002523, 0.00003284, -0.00000110, -0.00001612, -0.00002468, 0.00001026, -0.00001507, 0.00000732, -0.00000291, 0.00000420, -0.00000985, -0.00000278, -0.00000016, -0.00000154, -0.00000039, -0.00000607, -0.00000009, 0.00001337, -0.00000646, 0.00000621, -0.00000560, 0.00001324, -0.00001399, 0.00000645, 0.00000701, -0.00000165, -0.00000982, 0.00000953, -0.00000100, -0.00000101, -0.00000478, 0.00000048, -0.00000090, -0.00000153, 0.00000341, -0.00000441, -0.00000379, -0.00000809, 0.00000585, 0.00000647, -0.00000719, -0.00001105, 0.00000231, 0.00000970, -0.00000282, -0.00000186, 0.00000106, 0.00000367, 0.00000698, -0.00000199, -0.00000236, 0.00001170, -0.00001562, 0.00000637, -0.00000114, -0.00000468, 0.00001021, -0.00001798, 0.00000417, -0.00000937, 0.00000109, 0.00000068, -0.00000835, 0.00000813, -0.00001127, 0.00000100, 0.00000142, -0.00000152, 0.00001632, -0.00001121, 0.00000437, -0.00000781, 0.00000327, 0.00000417, 0.00000442, -0.00000376, 0.00000544, -0.00000317, -0.00000100, 0.00000142, 0.00000313, -0.00000975, 0.00000235, -0.00001838, 0.00001264, -0.00001071, 0.00000570, -0.00001706, 0.00000535, -0.00003720, 0.00000808, -0.00013817, 0.00006219, 0.00017822, 0.00017226, -0.00005334, -0.00009178, 0.00000300, -0.00003026, 0.00001253, -0.00001426, -0.00000180, -0.00001050, 0.00000247, -0.00000903, 0.00001089, -0.00000390, -0.00000155, -0.00000409, -0.00000429, -0.00000074, -0.00000117, -0.00000629, -0.00000423, -0.00000013, 0.00000198, 0.00000540, -0.00000637, 0.00000102, -0.00000916, 0.00000055, -0.00001199, 0.00000468, -0.00002229, -0.00000708, },
		{ -0.00000369, 0.00000759, 0.00004708, -0.00000022, 0.00004233, 0.00001768, -0.00000019, -0.00002996, -0.00000754, 0.00000009, -0.00000846, 0.00000474, -0.00000017, 0.00000042, -0.00000921, -0.00000249, 0.00000167, -0.00000416, 0.00000196, -0.00001049, 0.00001012, 0.00000425, -0.00000119, 0.00000095, 0.00000116, 0.00000742, -0.00001328, 0.00001016, 0.00000495, -0.00000494, -0.00000545, 0.00000901, -0.00000283, -0.00000083, -0.00000542, 0.00000237, -0.00000283, 0.00000077, 0.00000116, -0.00000388, -0.00000582, -0.00000611, 0.00000962, 0.00000231, -0.00001001, -0.00000891, 0.00000717, 0.00000664, -0.00000425, -0.00000036, 0.00000156, 0.00000450, 0.00000647, -0.00000582, 0.00000439, 0.00000613, -0.00001673, 0.00001280, -0.00001025, 0.00000627, -0.00000086, -0.00001258, 0.00000304, -0.00001002, 0.00000461, -0.00000350, -0.00000511, 0.00000733, -0.00001476, 0.00000796, -0.00000597, 0.00000834, 0.00000818, -0.00000846, 0.00000230, -0.00000648, 0.00000458, 0.00000531, 0.00000047, -0.00000012, 0.00000249, -0.00000111, -0.00000341, 0.00000639, -0.00000514, -0.00000163, -0.00000816, -0.00000753, 0.00000587, -0.00000428, -0.00000368, -0.00000634, -0.00001147, -0.00001813, -0.00002782, -0.00012023, 0.00013100, 0.00016077, 0.00014350, -0.00010584, -0.00004699, -0.00001357, -0.00001332, 0.00000336, -0.00000906, -0.00000720, -0.00000371, -0.00000341, -0.00000219, 0.00000765, -0.00000423, -0.00000183, -0.00000427, -0.00000447, 0.00000107, -0.00000339, -0.00000623, -0.00000329, 0.00000082, 0.00000303, 0.00000288, -0.00000562, -0.00000029, -0.00000859, -0.00000026, -0.00001118, 0.00000369, -0.00002637, },
		{ -0.00002581, -0.00000275, 0.00002179, 0.00004009, 0.00000235, 0.00004986, 0.00000597, -0.00000501, -0.00003463, 0.00000514, -0.00001007, -0.00000005, 0.00000042, 0.00000330, -0.00000437, -0.00000766, -0.00000142, 0.00000149, -0.00000455, 0.00000132, -0.00001018, 0.00001616, -0.00000373, 0.00000426, -0.00000425, 0.00000877, -0.00000164, -0.00000796, 0.00001074, 0.00000330, -0.00000774, 0.00000005, 0.00000575, -0.00000221, -0.00000270, -0.00000367, 0.00000211, -0.00000304, 0.00000161, -0.00000024, -0.00000398, -0.00000774, -0.00000217, 0.00001130, -0.00000293, -0.00001115, -0.00000539, 0.00001059, 0.00000220, -0.00000372, 0.00000056, 0.00000219, 0.00000525, 0.00000470, -0.00000784, 0.00001172, -0.00000437, -0.00001000, 0.00001122, -0.00001240, 0.00001247, -0.00001119, -0.00000436, -0.00000191, -0.00000682, 0.00000514, -0.00000662, -0.00000081, 0.00000345, -0.00001401, 0.00001120, -0.00001063, 0.00001730, -0.00000197, -0.00000272, -0.00000240, -0.00000234, 0.00000422, 0.00000614, -0.00000292, 0.00000365, -0.00000085, 0.00000106, -0.00000495, 0.00000924, -0.00001125, 0.00000481, -0.00001793, 0.00000498, -0.00000364, 0.00000385, -0.00001417, 0.00000502, -0.00002963, 0.00000415, -0.00007925, -0.00005697, 0.00015953, 0.00015026, 0.00009689, -0.00012936, -0.00001202, -0.00002861, 0.00000313, -0.00000743, -0.00000297, -0.00001208, 0.00000237, -0.00000893, 0.00000548, 0.00000191, -0.00000235, -0.00000324, -0.00000419, -0.00000387, 0.00000188, -0.00000544, -0.00000603, -0.00000152, 0.00000080, 0.00000447, -0.00000076, -0.00000294, -0.00000408, -0.00000492, -0.00000455, -0.00000640, -0.00000242, },
		{ -0.00001150, -0.00002245, -0.00000078, 0.00003719, 0.00002618, 0.00001548, 0.00004643, -0.00000107, -0.00001522, -0.00002820, 0.00000813, -0.00001427, 0.00000520, -0.00000205, 0.00000482, -0.00000812, -0.00000619, 0.00000068, -0.00000093, -0.00000235, -0.00000257, -0.00000419, 0.00001523, -0.00000658, 0.00000601, -0.00000545, 0.00001269, -0.00000944, -0.00000122, 0.00000974, 0.00000107, -0.00000829, 0.00000402, 0.00000240, -0.00000129, -0.00000478, -0.00000105, 0.00000070, -0.00000222, 0.00000138, -0.00000083, -0.00000506, -0.00000860, 0.00000304, 0.00000998, -0.00000770, -0.00001082, -0.00000068, 0.00001142, -0.00000193, -0.00000203, 0.00000101, 0.00000279, 0.00000610, 0.00000115, -0.00000589, 0.00001537, -0.00001487, 0.00000099, 0.00000297, -0.00000648, 0.00001022, -0.00001529, 0.00000127, -0.00000655, -0.00000243, 0.00000377, -0.00000861, 0.00000394, -0.00000316, -0.00000837, 0.00000869, -0.00000921, 0.00002088, -0.00000947, 0.00000218, -0.00000660, 0.00000227, 0.00000353, 0.00000546, -0.00000392, 0.00000528, -0.00000226, 0.00000114, -0.00000357, 0.00000756, -0.00001179, 0.00000491, -0.00002085, 0.00001225, -0.00000926, 0.00000710, -0.00001837, 0.00000784, -0.00003673, 0.00001106, -0.00011908, 0.00002841, 0.00015145, 0.00015066, 0.00003348, -0.00012037, 0.00000307, -0.00003257, 0.00001110, -0.00001337, -0.00000063, -0.00001272, 0.00000410, -0.00001026, 0.00000999, -0.00000298, -0.00000025, -0.00000491, -0.00000405, -0.00000249, 0.00000141, -0.00000708, -0.00000530, 0.00000025, 0.00000056, 0.00000531, -0.00000389, -0.00000041, -0.00000800, -0.00000063, -0.00000986, -0.00000012, },
		{ 0.00000419, -0.00002041, -0.00001864, 0.00000591, 0.00004710, 0.00001325, 0.00003261, 0.00003544, -0.00000618, -0.00002568, -0.00001519, 0.00000264, -0.00001144, 0.00000562, -0.00000138, 0.00000336, -0.00001003, -0.00000468, 0.00000259, -0.00000397, 0.00000058, -0.00000721, 0.00000482, 0.00000864, -0.00000405, 0.00000337, -0.00000162, 0.00001077, -0.00001287, 0.00000428, 0.00000842, -0.00000202, -0.00000610, 0.00000507, 0.00000078, -0.00000177, -0.00000530, 0.00000075, -0.00000046, -0.00000153, 0.00000104, -0.00000136, -0.00000680, -0.00000769, 0.00000823, 0.00000572, -0.00001082, -0.00000938, 0.00000464, 0.00000921, -0.00000422, -0.00000033, 0.00000143, 0.00000320, 0.00000681, -0.00000366, 0.00000059, 0.00001211, -0.00001978, 0.00001002, -0.00000640, 0.00000332, 0.00000127, -0.00001219, 0.00000174, -0.00000806, 0.00000091, 0.00000156, -0.00000913, 0.00000770, -0.00001037, 0.00000028, 0.00000142, -0.00000145, 0.00001722, -0.00001145, 0.00000343, -0.00000772, 0.00000524, 0.00000348, 0.00000306, -0.00000218, 0.00000414, -0.00000114, -0.00000121, 0.00000061, 0.00000173, -0.00000692, -0.00000189, -0.00001501, 0.00001116, -0.00000783, 0.00000270, -0.00001354, -0.00000122, -0.00002755, -0.00000862, -0.00012226, 0.00010312, 0.00012723, 0.00015152, -0.00003375, -0.00008792, -0.00000052, -0.00002428, 0.00000875, -0.00001247, -0.00000316, -0.00000858, 0.00000095, -0.00000637, 0.00000945, -0.00000477, 0.00000037, -0.00000600, -0.00000387, -0.00000069, -0.00000020, -0.00000818, -0.00000368, 0.00000111, 0.00000098, 0.00000458, -0.00000506, -0.00000002, -0.00000952, 0.00000136, -0.00001299, },
		{ -0.00001203, 0.00000413, -0.00002692, -0.00001509, 0.00001827, 0.00004750, 0.00000809, 0.00004637, 0.00002159, -0.00001256, -0.00003041, -0.00000312, -0.00000570, -0.00000487, 0.00000297, 0.00000120, -0.00000021, -0.00001048, -0.00000243, 0.00000268, -0.00000539, 0.00000164, -0.00000902, 0.00001219, 0.00000071, 0.00000072, -0.00000107, 0.00000487, 0.00000426, -0.00001157, 0.00000759, 0.00000691, -0.00000500, -0.00000242, 0.00000381, 0.00000081, -0.00000376, -0.00000388, 0.00000104, -0.00000074, -0.00000156, 0.00000124, -0.00000255, -0.00000850, -0.00000471, 0.00001174, -0.00000027, -0.00001199, -0.00000681, 0.00000933, 0.00000473, -0.00000416, 0.00000068, 0.00000201, 0.00000358, 0.00000653, -0.00000771, 0.00000944, 0.00000215, -0.00001632, 0.00001214, -0.00001076, 0.00001044, -0.00000869, -0.00000528, -0.00000174, -0.00000637, 0.00000252, -0.00000113, -0.00000753, 0.00000848, -0.00001497, 0.00000802, -0.00000668, 0.00000933, 0.00000816, -0.00000829, 0.00000081, -0.00000523, 0.00000593, 0.00000396, 0.00000004, 0.00000093, 0.00000159, 0.00000103, -0.00000420, 0.00000514, -0.00000502, -0.00000041, -0.00001154, -0.00000383, 0.00000408, -0.00000162, -0.00000631, -0.00000379, -0.00001675, -0.00000888, -0.00004974, -0.00008052, 0.00014247, 0.00010967, 0.00013737, -0.00008655, -0.00004880, -0.00001303, -0.00000992, 0.00000027, -0.00000749, -0.00000799, -0.00000242, -0.00000441, 0.00000054, 0.00000491, -0.00000338, -0.00000101, -0.00000614, -0.00000352, 0.00000093, -0.00000257, -0.00000856, -0.00000136, 0.00000080, 0.00000231, 0.00000222, -0.00000405, -0.00000228, -0.00000770, -0.00000040, },
		{ -0.00000506, -0.00000744, -0.00000062, -0.00003054, -0.00001030, 0.00003346, 0.00003927, 0.00001307, 0.00005211, 0.00000788, -0.00002062, -0.00002690, 0.00000246, -0.00001112, 0.00000101, 0.00000029, 0.00000342, -0.00000427, -0.00001002, 0.00000065, 0.00000037, -0.00000413, -0.00000039, -0.00000597, 0.00001428, -0.00000418, 0.00000385, -0.00000359, 0.00001003, -0.00000340, -0.00000738, 0.00000923, 0.00000464, -0.00000638, 0.00000069, 0.00000211, 0.00000101, -0.00000588, -0.00000155, 0.00000040, -0.00000054, -0.00000199, 0.00000181, -0.00000461, -0.00000942, 0.00000006, 0.00001218, -0.00000616, -0.00001168, -0.00000287, 0.00001178, -0.00000017, -0.00000247, 0.00000106, 0.00000253, 0.00000433, 0.00000425, -0.00000833, 0.00001631, -0.00001039, -0.00000638, 0.00000678, -0.00000748, 0.00001026, -0.00001396, 0.00000044, -0.00000576, -0.00000331, 0.00000300, -0.00000432, -0.00000350, 0.00000510, -0.00001439, 0.00001104, -0.00001093, 0.00001807, -0.00000193, -0.00000297, -0.00000350, -0.00000060, 0.00000505, 0.00000410, -0.00000197, 0.00000336, -0.00000027, 0.00000203, -0.00000537, 0.00000705, -0.00000886, 0.00000287, -0.00001809, 0.00000641, -0.00000336, 0.00000385, -0.00001363, 0.00000302, -0.00002824, 0.00000408, -0.00009189, -0.00000758, 0.00014172, 0.00011005, 0.00009957, -0.00011153, -0.00001859, -0.00002315, 0.00000216, -0.00000790, -0.00000304, -0.00001124, 0.00000194, -0.00000810, 0.00000667, -0.00000063, -0.00000050, -0.00000367, -0.00000544, -0.00000286, 0.00000184, -0.00000531, -0.00000782, 0.00000086, -0.00000008, 0.00000389, -0.00000079, -0.00000205, -0.00000590, -0.00000373, },
		{ -0.00000014, -0.00001003, -0.00000156, -0.00000859, -0.00003187, -0.00000178, 0.00004622, 0.00002755, 0.00002566, 0.00004892, -0.00000509, -0.00002749, -0.00001765, 0.00000086, -0.00001118, 0.00000360, -0.00000046, 0.00000374, -0.00000774, -0.00000859, 0.00000349, -0.00000313, -0.00000125, -0.00000425, 0.00000091, 0.00001070, -0.00000434, 0.00000343, -0.00000211, 0.00001094, -0.00000891, -0.00000255, 0.00001000, 0.00000134, -0.00000543, 0.00000177, 0.00000159, -0.00000009, -0.00000656, 0.00000020, -0.00000007, -0.00000076, -0.00000194, 0.00000198, -0.00000720, -0.00000883, 0.00000553, 0.00000920, -0.00001046, -0.00001040, 0.00000226, 0.00001094, -0.00000352, -0.00000047, 0.00000132, 0.00000271, 0.00000550, -0.00000015, -0.00000386, 0.00001691, -0.00001953, 0.00000448, -0.00000194, 0.00000095, 0.00000295, -0.00001234, 0.00000174, -0.00000746, -0.00000083, 0.00000296, -0.00000753, 0.00000198, -0.00000166, -0.00000853, 0.00000793, -0.00000859, 0.00002079, -0.00000879, 0.00000099, -0.00000661, 0.00000389, 0.00000377, 0.00000329, -0.00000205, 0.00000391, -0.00000025, 0.00000070, -0.00000359, 0.00000505, -0.00000803, 0.00000036, -0.00001760, 0.00001077, -0.00000642, 0.00000419, -0.00001459, 0.00000124, -0.00002784, -0.00000220, -0.00011041, 0.00006828, 0.00011581, 0.00012334, 0.00004253, -0.00010686, -0.00000469, -0.00002444, 0.00000671, -0.00001139, -0.00000215, -0.00001061, 0.00000228, -0.00000762, 0.00000902, -0.00000420, 0.00000166, -0.00000635, -0.00000430, -0.00000186, 0.00000172, -0.00000795, -0.00000574, 0.00000203, -0.00000050, 0.00000463, -0.00000304, -0.00000086, -0.00000865, },
		{ -0.00000889, 0.00000088, -0.00001271, 0.00000280, -0.00001783, -0.00003112, 0.00001166, 0.00005197, 0.00001869, 0.00004049, 0.00003834, -0.00001743, -0.00002946, -0.00000813, -0.00000436, -0.00000730, 0.00000323, 0.00000062, 0.00000207, -0.00001030, -0.00000575, 0.00000442, -0.00000562, 0.00000089, -0.00000695, 0.00000805, 0.00000428, -0.00000133, 0.00000055, 0.00000255, 0.00000734, -0.00001083, 0.00000161, 0.00001002, -0.00000221, -0.00000282, 0.00000096, 0.00000239, -0.00000276, -0.00000523, 0.00000060, 0.00000022, -0.00000191, -0.00000083, 0.00000097, -0.00000962, -0.00000632, 0.00001002, 0.00000375, -0.00001266, -0.00000818, 0.00000753, 0.00000711, -0.00000431, 0.00000082, 0.00000182, 0.00000265, 0.00000640, -0.00000524, 0.00000474, 0.00000971, -0.00002087, 0.00001048, -0.00000802, 0.00000875, -0.00000697, -0.00000599, -0.00000115, -0.00000619, 0.00000042, 0.00000229, -0.00000947, 0.00000666, -0.00000898, -0.00000006, 0.00000066, -0.00000048, 0.00001664, -0.00001045, 0.00000162, -0.00000668, 0.00000660, 0.00000281, 0.00000172, -0.00000057, 0.00000293, 0.00000108, -0.00000217, 0.00000010, 0.00000051, -0.00000411, -0.00000651, -0.00001052, 0.00000856, -0.00000437, -0.00000072, -0.00000963, -0.00000815, -0.00001668, -0.00002970, -0.00009059, 0.00011954, 0.00008821, 0.00013375, -0.00001867, -0.00008196, -0.00000486, -0.00001752, 0.00000390, -0.00000991, -0.00000466, -0.00000670, -0.00000077, -0.00000321, 0.00000695, -0.00000441, 0.00000161, -0.00000793, -0.00000315, -0.00000063, 0.00000040, -0.00000989, -0.00000259, 0.00000167, 0.00000022, 0.00000384, -0.00000364, -0.00000162, },
		{ -0.00000410, -0.00000658, -0.00000125, -0.00001189, 0.00000374, -0.00002691, -0.00002730, 0.00002845, 0.00004928, 0.00001718, 0.00005213, 0.00002251, -0.00002804, -0.00002516, -0.00000296, -0.00000863, -0.00000269, 0.00000186, 0.00000216, -0.00000082, -0.00001188, -0.00000155, 0.00000259, -0.00000554, 0.00000049, -0.00000600, 0.00001177, -0.00000105, 0.00000165, -0.00000190, 0.00000744, 0.00000136, -0.00000975, 0.00000499, 0.00000877, -0.00000464, -0.00000026, -0.00000032, 0.00000334, -0.00000584, -0.00000266, 0.00000000, 0.00000092, -0.00000358, 0.00000119, -0.00000155, -0.00001104, -0.00000200, 0.00001197, -0.00000241, -0.00001314, -0.00000462, 0.00001115, 0.00000200, -0.00000294, 0.00000125, 0.00000240, 0.00000285, 0.00000578, -0.00000825, 0.00001373, -0.00000273, -0.00001418, 0.00000914, -0.00000755, 0.00001065, -0.00001372, 0.00000047, -0.00000542, -0.00000336, 0.00000097, 0.00000041, -0.00000880, 0.00000812, -0.00001339, 0.00000688, -0.00000655, 0.00000962, 0.00000808, -0.00000787, -0.00000070, -0.00000371, 0.00000708, 0.00000220, 0.00000026, 0.00000123, 0.00000177, 0.00000208, -0.00000455, 0.00000327, -0.00000364, -0.00000072, -0.00001338, -0.00000104, 0.00000300, -0.00000025, -0.00000715, -0.00000369, -0.00001833, -0.00000473, -0.00006601, -0.00003632, 0.00013317, 0.00007687, 0.00012600, -0.00006685, -0.00005124, -0.00001120, -0.00000770, -0.00000226, -0.00000627, -0.00000805, -0.00000213, -0.00000450, 0.00000254, 0.00000223, -0.00000192, -0.00000079, -0.00000788, -0.00000226, 0.00000063, -0.00000208, -0.00001038, 0.00000068, 0.00000025, 0.00000186, 0.00000178, -0.00000287, },
		{ -0.00000189, -0.00000686, -0.00000321, -0.00000535, -0.00000816, 0.00000071, -0.00003495, -0.00001862, 0.00004461, 0.00004058, 0.00002392, 0.00005670, 0.00000348, -0.00003428, -0.00001689, -0.00000314, -0.00000942, 0.00000023, 0.00000120, 0.00000288, -0.00000419, -0.00001201, 0.00000280, -0.00000118, -0.00000320, -0.00000217, -0.00000128, 0.00001060, -0.00000287, 0.00000228, -0.00000155, 0.00000964, -0.00000418, -0.00000717, 0.00000792, 0.00000590, -0.00000491, 0.00000074, -0.00000052, 0.00000292, -0.00000771, -0.00000031, -0.00000056, 0.00000109, -0.00000476, 0.00000319, -0.00000526, -0.00001069, 0.00000321, 0.00001071, -0.00000765, -0.00001245, 0.00000042, 0.00001161, -0.00000217, -0.00000079, 0.00000134, 0.00000266, 0.00000371, 0.00000283, -0.00000669, 0.00001833, -0.00001490, -0.00000352, 0.00000260, -0.00000113, 0.00000522, -0.00001392, 0.00000317, -0.00000795, -0.00000092, 0.00000161, -0.00000283, -0.00000517, 0.00000535, -0.00001278, 0.00000910, -0.00000962, 0.00001714, -0.00000090, -0.00000380, -0.00000394, 0.00000085, 0.00000586, 0.00000171, -0.00000040, 0.00000233, 0.00000150, 0.00000146, -0.00000486, 0.00000407, -0.00000513, -0.00000090, -0.00001600, 0.00000598, -0.00000159, 0.00000200, -0.00001076, -0.00000206, -0.00002194, -0.00000418, -0.00009073, 0.00003208, 0.00011526, 0.00008538, 0.00009457, -0.00009132, -0.00002655, -0.00001601, -0.00000054, -0.00000724, -0.00000372, -0.00000965, 0.00000047, -0.00000612, 0.00000658, -0.00000229, 0.00000130, -0.00000450, -0.00000640, -0.00000169, 0.00000159, -0.00000543, -0.00000885, 0.00000282, -0.00000111, 0.00000348, -0.00000055, },
		{ -0.00000212, -0.00000188, -0.00000834, -0.00000075, -0.00000933, -0.00000326, -0.00000587, -0.00004095, -0.00000401, 0.00005561, 0.00003073, 0.00003655, 0.00005219, -0.00001616, -0.00003383, -0.00000910, -0.00000623, -0.00000724, 0.00000118, 0.00000159, 0.00000237, -0.00000770, -0.00000997, 0.00000539, -0.00000472, -0.00000057, -0.00000489, 0.00000483, 0.00000607, -0.00000152, 0.00000061, 0.00000173, 0.00000807, -0.00000752, -0.00000419, 0.00001031, 0.00000190, -0.00000315, -0.00000013, 0.00000084, 0.00000052, -0.00000740, 0.00000076, -0.00000029, -0.00000002, -0.00000447, 0.00000399, -0.00000926, -0.00000812, 0.00000775, 0.00000684, -0.00001126, -0.00001061, 0.00000606, 0.00000869, -0.00000387, 0.00000084, 0.00000165, 0.00000245, 0.00000494, -0.00000184, -0.00000014, 0.00001542, -0.00002133, 0.00000534, -0.00000399, 0.00000672, -0.00000444, -0.00000833, 0.00000119, -0.00000739, 0.00000013, 0.00000236, -0.00000653, 0.00000013, -0.00000051, -0.00000764, 0.00000606, -0.00000685, 0.00001913, -0.00000695, -0.00000091, -0.00000586, 0.00000512, 0.00000383, 0.00000126, -0.00000021, 0.00000262, 0.00000204, -0.00000067, -0.00000310, 0.00000250, -0.00000400, -0.00000489, -0.00001300, 0.00000810, -0.00000292, 0.00000064, -0.00001011, -0.00000614, -0.00001709, -0.00002056, -0.00008714, 0.00008898, 0.00008488, 0.00010367, 0.00004596, -0.00009169, -0.00001280, -0.00001580, 0.00000146, -0.00000880, -0.00000374, -0.00000849, 0.00000020, -0.00000448, 0.00000709, -0.00000437, 0.00000309, -0.00000795, -0.00000418, -0.00000126, 0.00000177, -0.00000887, -0.00000542, 0.00000302, -0.00000134, 0.00000408, },
		{ 0.00000326, -0.00000255, -0.00000322, -0.00000781, -0.00000046, -0.00001144, 0.00000083, -0.00001527, -0.00004320, 0.00001551, 0.00005867, 0.00002480, 0.00005067, 0.00003827, -0.00003273, -0.00002690, -0.00000536, -0.00000873, -0.00000426, 0.00000127, 0.00000230, 0.00000085, -0.00001097, -0.00000565, 0.00000490, -0.00000606, 0.00000032, -0.00000521, 0.00000906, 0.00000131, 0.00000068, -0.00000133, 0.00000596, 0.00000383, -0.00000852, -0.00000103, 0.00001142, -0.00000187, -0.00000066, -0.00000189, 0.00000293, -0.00000314, -0.00000518, 0.00000048, 0.00000077, -0.00000235, -0.00000239, 0.00000269, -0.00001232, -0.00000358, 0.00001017, 0.00000180, -0.00001340, -0.00000713, 0.00001050, 0.00000378, -0.00000308, 0.00000146, 0.00000222, 0.00000217, 0.00000547, -0.00000602, 0.00000895, 0.00000556, -0.00001966, 0.00000837, -0.00000599, 0.00001057, -0.00001288, -0.00000087, -0.00000344, -0.00000454, 0.00000015, 0.00000246, -0.00000906, 0.00000474, -0.00000668, -0.00000071, 0.00000003, 0.00000041, 0.00001546, -0.00000886, -0.00000047, -0.00000520, 0.00000767, 0.00000175, 0.00000102, 0.00000034, 0.00000272, 0.00000238, -0.00000296, -0.00000071, 0.00000019, -0.00000237, -0.00001017, -0.00000634, 0.00000617, -0.00000170, -0.00000300, -0.00000736, -0.00001244, -0.00000922, -0.00004800, -0.00005192, 0.00011656, 0.00006214, 0.00011608, -0.00000598, -0.00007569, -0.00000831, -0.00001163, -0.00000066, -0.00000750, -0.00000563, -0.00000552, -0.00000186, -0.00000051, 0.00000432, -0.00000334, 0.00000224, -0.00000975, -0.00000214, -0.00000067, 0.00000067, -0.00001126, -0.00000112, 0.00000148, -0.00000016, },
		{ 0.00000179, 0.00000149, -0.00000230, -0.00000532, -0.00000575, -0.00000234, -0.00001102, 0.00000262, -0.00002658, -0.00003947, 0.00003667, 0.00005397, 0.00002608, 0.00006115, 0.00001646, -0.00004208, -0.00001684, -0.00000613, -0.00000870, -0.00000223, 0.00000164, 0.00000259, -0.00000149, -0.00001310, -0.00000022, 0.00000167, -0.00000482, -0.00000104, -0.00000224, 0.00000944, -0.00000111, 0.00000147, -0.00000136, 0.00000856, -0.00000086, -0.00000809, 0.00000259, 0.00001048, -0.00000402, 0.00000097, -0.00000303, 0.00000427, -0.00000652, -0.00000227, -0.00000036, 0.00000181, -0.00000501, 0.00000087, -0.00000092, -0.00001316, 0.00000182, 0.00000985, -0.00000304, -0.00001436, -0.00000177, 0.00001192, -0.00000080, -0.00000101, 0.00000148, 0.00000264, 0.00000240, 0.00000413, -0.00000696, 0.00001616, -0.00000712, -0.00001179, 0.00000555, -0.00000223, 0.00000761, -0.00001573, 0.00000410, -0.00000751, -0.00000137, 0.00000015, 0.00000108, -0.00000902, 0.00000654, -0.00001032, 0.00000464, -0.00000543, 0.00000881, 0.00000837, -0.00000753, -0.00000199, -0.00000210, 0.00000796, 0.00000022, 0.00000103, 0.00000084, 0.00000314, 0.00000161, -0.00000409, 0.00000078, -0.00000110, -0.00000261, -0.00001337, 0.00000040, 0.00000295, -0.00000032, -0.00000614, -0.00000609, -0.00001573, -0.00000741, -0.00007176, 0.00000306, 0.00011245, 0.00005782, 0.00011082, -0.00004763, -0.00005370, -0.00000836, -0.00000674, -0.00000402, -0.00000542, -0.00000756, -0.00000262, -0.00000371, 0.00000356, 0.00000015, -0.00000026, -0.00000114, -0.00000927, -0.00000091, 0.00000024, -0.00000199, -0.00001144, 0.00000248, -0.00000068, },
		{ -0.00000206, 0.00000339, -0.00000036, -0.00000222, -0.00000707, -0.00000342, -0.00000532, -0.00000847, 0.00000115, -0.00003839, -0.00002804, 0.00005496, 0.00004455, 0.00003517, 0.00006314, -0.00000932, -0.00004172, -0.00000846, -0.00000898, -0.00000662, -0.00000137, 0.00000245, 0.00000223, -0.00000467, -0.00001292, 0.00000421, -0.00000240, -0.00000237, -0.00000327, 0.00000265, 0.00000648, -0.00000077, 0.00000031, 0.00000121, 0.00000805, -0.00000425, -0.00000709, 0.00000660, 0.00000742, -0.00000392, 0.00000083, -0.00000257, 0.00000373, -0.00000823, -0.00000001, -0.00000076, 0.00000181, -0.00000668, 0.00000396, -0.00000615, -0.00001097, 0.00000637, 0.00000742, -0.00000716, -0.00001389, 0.00000468, 0.00000968, -0.00000316, 0.00000083, 0.00000158, 0.00000255, 0.00000329, 0.00000076, -0.00000317, 0.00001748, -0.00001714, -0.00000245, 0.00000038, 0.00000450, -0.00000081, -0.00001210, 0.00000431, -0.00000875, 0.00000054, 0.00000074, -0.00000184, -0.00000624, 0.00000488, -0.00001005, 0.00000624, -0.00000740, 0.00001491, 0.00000091, -0.00000497, -0.00000397, 0.00000227, 0.00000620, -0.00000037, 0.00000107, 0.00000139, 0.00000365, -0.00000018, -0.00000375, 0.00000097, -0.00000104, -0.00000545, -0.00001258, 0.00000450, 0.00000073, -0.00000048, -0.00000722, -0.00000795, -0.00001356, -0.00001767, -0.00007667, 0.00005741, 0.00008836, 0.00006996, 0.00008549, -0.00007140, -0.00003373, -0.00000891, -0.00000377, -0.00000621, -0.00000436, -0.00000811, -0.00000116, -0.00000376, 0.00000568, -0.00000308, 0.00000283, -0.00000569, -0.00000687, -0.00000062, 0.00000122, -0.00000583, -0.00000896, 0.00000403, },
		{ 0.00000315, -0.00000180, 0.00000378, -0.00000158, -0.00000291, -0.00000761, -0.00000204, -0.00000806, -0.00000484, -0.00000411, -0.00004843, -0.00000885, 0.00006630, 0.00003519, 0.00004974, 0.00005325, -0.00003292, -0.00003276, -0.00000509, -0.00001075, -0.00000429, -0.00000079, 0.00000317, 0.00000134, -0.00000846, -0.00000982, 0.00000577, -0.00000508, -0.00000070, -0.00000419, 0.00000678, 0.00000254, 0.00000079, -0.00000145, 0.00000510, 0.00000489, -0.00000587, -0.00000566, 0.00001028, 0.00000310, -0.00000208, -0.00000087, -0.00000058, 0.00000122, -0.00000778, 0.00000094, -0.00000025, 0.00000022, -0.00000633, 0.00000532, -0.00001143, -0.00000594, 0.00000857, 0.00000417, -0.00001077, -0.00001114, 0.00001025, 0.00000492, -0.00000286, 0.00000167, 0.00000203, 0.00000211, 0.00000414, -0.00000324, 0.00000432, 0.00001167, -0.00002067, 0.00000378, -0.00000274, 0.00000942, -0.00001021, -0.00000464, 0.00000051, -0.00000693, 0.00000085, 0.00000166, -0.00000534, -0.00000190, 0.00000081, -0.00000652, 0.00000409, -0.00000501, 0.00001674, -0.00000454, -0.00000300, -0.00000479, 0.00000628, 0.00000324, 0.00000006, 0.00000082, 0.00000239, 0.00000348, -0.00000204, -0.00000275, 0.00000070, -0.00000083, -0.00000945, -0.00000845, 0.00000549, -0.00000006, -0.00000203, -0.00000697, -0.00001132, -0.00000880, -0.00003787, -0.00005596, 0.00009288, 0.00006178, 0.00008763, 0.00004653, -0.00007704, -0.00001933, -0.00000836, -0.00000333, -0.00000647, -0.00000479, -0.00000703, -0.00000131, -0.00000175, 0.00000499, -0.00000383, 0.00000398, -0.00000957, -0.00000368, -0.00000080, 0.00000159, -0.00000969, -0.00000448, },
		{ 0.00000038, 0.00000068, -0.00000004, 0.00000282, -0.00000205, -0.00000434, -0.00000685, -0.00000214, -0.00000956, -0.00000142, -0.00001325, -0.00005377, 0.00001595, 0.00006852, 0.00003084, 0.00006471, 0.00003113, -0.00004779, -0.00001993, -0.00000667, -0.00000992, -0.00000304, 0.00000029, 0.00000331, -0.00000018, -0.00001190, -0.00000456, 0.00000410, -0.00000531, -0.00000088, -0.00000251, 0.00000807, 0.00000002, 0.00000155, -0.00000185, 0.00000806, 0.00000080, -0.00000634, -0.00000333, 0.00001247, -0.00000101, 0.00000024, -0.00000307, 0.00000193, -0.00000237, -0.00000563, 0.00000071, 0.00000091, -0.00000268, -0.00000370, 0.00000375, -0.00001471, 0.00000046, 0.00000803, 0.00000111, -0.00001403, -0.00000556, 0.00001273, -0.00000000, -0.00000092, 0.00000169, 0.00000253, 0.00000186, 0.00000383, -0.00000544, 0.00001210, 0.00000099, -0.00001741, 0.00000511, -0.00000145, 0.00000896, -0.00001585, 0.00000242, -0.00000458, -0.00000348, 0.00000030, 0.00000205, -0.00000798, 0.00000216, -0.00000363, -0.00000177, -0.00000005, 0.00000059, 0.00001429, -0.00000710, -0.00000248, -0.00000353, 0.00000851, 0.00000029, 0.00000106, 0.00000038, 0.00000386, 0.00000209, -0.00000300, -0.00000207, 0.00000098, -0.00000211, -0.00001217, -0.00000326, 0.00000465, -0.00000036, -0.00000362, -0.00000739, -0.00001282, -0.00000779, -0.00005805, -0.00001503, 0.00010174, 0.00004652, 0.00009837, 0.00000460, -0.00006939, -0.00001046, -0.00000718, -0.00000435, -0.00000553, -0.00000605, -0.00000506, -0.00000218, 0.00000130, 0.00000215, -0.00000195, 0.00000222, -0.00001120, -0.00000102, -0.00000077, 0.00000053, -0.00001203, },
		{ -0.00001160, 0.00000372, -0.00000179, 0.00000222, 0.00000106, -0.00000217, -0.00000593, -0.00000537, -0.00000346, -0.00000945, 0.00000057, -0.00002569, -0.00005145, 0.00004225, 0.00006223, 0.00003508, 0.00007319, 0.00000076, -0.00005000, -0.00000926, -0.00001027, -0.00000725, -0.00000277, 0.00000180, 0.00000290, -0.00000268, -0.00001350, 0.00000083, 0.00000059, -0.00000377, -0.00000228, 0.00000117, 0.00000634, -0.00000019, 0.00000064, 0.00000013, 0.00000840, -0.00000245, -0.00000655, 0.00000030, 0.00001225, -0.00000358, 0.00000173, -0.00000458, 0.00000376, -0.00000576, -0.00000282, -0.00000000, 0.00000191, -0.00000587, 0.00000044, -0.00000096, -0.00001425, 0.00000589, 0.00000580, -0.00000176, -0.00001621, 0.00000210, 0.00001104, -0.00000287, 0.00000108, 0.00000161, 0.00000263, 0.00000220, 0.00000177, -0.00000389, 0.00001618, -0.00001003, -0.00001030, 0.00000304, 0.00000322, 0.00000280, -0.00001537, 0.00000577, -0.00000811, -0.00000031, -0.00000012, 0.00000116, -0.00000864, 0.00000442, -0.00000657, 0.00000201, -0.00000378, 0.00000707, 0.00000905, -0.00000721, -0.00000313, -0.00000027, 0.00000819, -0.00000146, 0.00000182, 0.00000037, 0.00000512, -0.00000012, -0.00000295, -0.00000188, 0.00000191, -0.00000540, -0.00001196, 0.00000081, 0.00000348, -0.00000111, -0.00000438, -0.00000947, -0.00001066, -0.00001556, -0.00006598, 0.00003168, 0.00008856, 0.00004715, 0.00009444, -0.00003069, -0.00005478, -0.00000559, -0.00000640, -0.00000535, -0.00000459, -0.00000708, -0.00000325, -0.00000259, 0.00000382, -0.00000117, 0.00000129, -0.00000200, -0.00001006, 0.00000027, -0.00000014, -0.00000232, },
		{ -0.00000637, -0.00000827, 0.00000438, -0.00000280, 0.00000375, -0.00000067, -0.00000251, -0.00000697, -0.00000402, -0.00000523, -0.00000802, 0.00000001, -0.00003985, -0.00003947, 0.00006495, 0.00005092, 0.00004839, 0.00006861, -0.00002983, -0.00004062, -0.00000470, -0.00001244, -0.00000465, -0.00000251, 0.00000310, 0.00000214, -0.00000623, -0.00001216, 0.00000416, -0.00000260, -0.00000212, -0.00000322, 0.00000488, 0.00000318, 0.00000100, -0.00000115, 0.00000382, 0.00000596, -0.00000405, -0.00000681, 0.00000500, 0.00000950, -0.00000398, 0.00000163, -0.00000462, 0.00000403, -0.00000786, -0.00000038, -0.00000044, 0.00000183, -0.00000791, 0.00000436, -0.00000759, -0.00000957, 0.00000828, 0.00000360, -0.00000531, -0.00001573, 0.00000958, 0.00000608, -0.00000284, 0.00000206, 0.00000189, 0.00000225, 0.00000283, -0.00000136, 0.00000146, 0.00001422, -0.00001716, -0.00000331, 0.00000084, 0.00000782, -0.00000610, -0.00000970, 0.00000454, -0.00000863, 0.00000139, 0.00000010, -0.00000106, -0.00000711, 0.00000435, -0.00000713, 0.00000351, -0.00000525, 0.00001213, 0.00000303, -0.00000608, -0.00000393, 0.00000394, 0.00000557, -0.00000144, 0.00000167, 0.00000148, 0.00000509, -0.00000203, -0.00000266, -0.00000144, 0.00000232, -0.00000952, -0.00000901, 0.00000296, 0.00000252, -0.00000225, -0.00000481, -0.00001202, -0.00000680, -0.00003149, -0.00005426, 0.00006840, 0.00006569, 0.00005925, 0.00007515, -0.00005382, -0.00003840, -0.00000332, -0.00000645, -0.00000546, -0.00000446, -0.00000719, -0.00000224, -0.00000180, 0.00000458, -0.00000319, 0.00000392, -0.00000706, -0.00000681, 0.00000021, 0.00000079, },
		{ 0.00000112, -0.00001017, -0.00000321, 0.00000251, -0.00000178, 0.00000376, -0.00000174, -0.00000344, -0.00000707, -0.00000345, -0.00000666, -0.00000591, -0.00000397, -0.00005319, -0.00001756, 0.00007921, 0.00004044, 0.00006693, 0.00004777, -0.00005135, -0.00002550, -0.00000614, -0.00001164, -0.00000338, -0.00000153, 0.00000376, 0.00000090, -0.00001004, -0.00000812, 0.00000447, -0.00000404, -0.00000162, -0.00000236, 0.00000667, 0.00000070, 0.00000191, -0.00000216, 0.00000732, 0.00000203, -0.00000437, -0.00000667, 0.00000981, 0.00000495, -0.00000241, -0.00000008, -0.00000310, 0.00000250, -0.00000814, 0.00000105, -0.00000018, 0.00000014, -0.00000759, 0.00000600, -0.00001363, -0.00000213, 0.00000710, 0.00000259, -0.00001011, -0.00001122, 0.00001396, 0.00000039, -0.00000078, 0.00000205, 0.00000240, 0.00000180, 0.00000287, -0.00000371, 0.00000836, 0.00000683, -0.00001857, 0.00000092, 0.00000104, 0.00000887, -0.00001348, -0.00000211, 0.00000023, -0.00000643, 0.00000149, 0.00000079, -0.00000399, -0.00000402, 0.00000238, -0.00000556, 0.00000262, -0.00000383, 0.00001433, -0.00000206, -0.00000491, -0.00000367, 0.00000747, 0.00000189, -0.00000006, 0.00000072, 0.00000370, 0.00000325, -0.00000269, -0.00000296, 0.00000009, 0.00000079, -0.00001240, -0.00000491, 0.00000372, 0.00000148, -0.00000306, -0.00000616, -0.00001263, -0.00000597, -0.00004868, -0.00002453, 0.00008549, 0.00004603, 0.00007373, 0.00004535, -0.00006373, -0.00002353, -0.00000292, -0.00000698, -0.00000474, -0.00000521, -0.00000631, -0.00000201, 0.00000009, 0.00000329, -0.00000289, 0.00000422, -0.00001089, -0.00000297, -0.00000046, },
		},
		{ { -0.00000158, -0.00000116, 0.00000121, -0.00000201, -0.00000683, -0.00000467, -0.00000574, -0.00000499, -0.00000877, -0.00001126, -0.00006194, 0.00001182, 0.00008525, 0.00003760, 0.00008342, 0.00000888, -0.00005855, -0.00001191, -0.00000365, -0.00000802, -0.00000159, -0.00000384, 0.00000133, -0.00000215, -0.00000349, -0.00000333, -0.00000408, 0.00000128, 0.00000219, -0.00000201, 0.00000721, -0.00000288, -0.00000189, -0.00000116, 0.00000093, 0.00000262, -0.00001467, 0.00000036, -0.00000694, -0.00000225, 0.00000225, 0.00000602, -0.00001511, 0.00000239, -0.00000615, 0.00000033, -0.00000468, -0.00000151, 0.00000719, 0.00000056, -0.00000759, 0.00000368, 0.00000136, 0.00000034, 0.00000365, -0.00000081, -0.00000225, -0.00000303, 0.00001576, -0.00000252, 0.00000587, -0.00000548, -0.00000242, 0.00000182, 0.00000444, 0.00000182, -0.00000242, -0.00000548, 0.00000587, -0.00000252, 0.00001576, -0.00000303, -0.00000225, -0.00000081, 0.00000365, 0.00000034, 0.00000136, 0.00000368, -0.00000759, 0.00000056, 0.00000719, -0.00000151, -0.00000468, 0.00000033, -0.00000615, 0.00000239, -0.00001511, 0.00000602, 0.00000225, -0.00000225, -0.00000694, 0.00000036, -0.00001467, 0.00000262, 0.00000093, -0.00000116, -0.00000189, -0.00000288, 0.00000721, -0.00000201, 0.00000219, 0.00000128, -0.00000408, -0.00000333, -0.00000349, -0.00000215, 0.00000133, -0.00000384, -0.00000159, -0.00000802, -0.00000365, -0.00001191, -0.00005855, 0.00000888, 0.00008342, 0.00003760, 0.00008525, 0.00001182, -0.00006194, -0.00001126, -0.00000877, -0.00000499, -0.00000574, -0.00000467, -0.00000683, -0.00000201, 0.00000121, -0.00000116, },
		{ -0.00000148, -0.00000159, -0.00000060, 0.00000115, -0.00000357, -0.00000651, -0.00000505, -0.00000512, -0.00000603, -0.00000825, -0.00001794, -0.00006598, 0.00003794, 0.00008011, 0.00004401, 0.00008688, -0.00002152, -0.00005233, -0.00000536, -0.00000605, -0.00000661, -0.00000171, -0.00000354, 0.00000178, -0.00000340, -0.00000260, -0.00000452, -0.00000285, 0.00000202, 0.00000140, -0.00000184, 0.00000810, -0.00000527, 0.00000061, -0.00000323, 0.00000373, -0.00000168, -0.00001299, 0.00000043, -0.00000720, -0.00000165, 0.00000487, 0.00000123, -0.00001426, 0.00000481, -0.00000877, 0.00000284, -0.00000746, 0.00000258, 0.00000549, -0.00000056, -0.00000744, 0.00000609, -0.00000138, 0.00000326, 0.00000064, 0.00000097, -0.00000515, 0.00000253, 0.00001309, -0.00000297, 0.00000596, -0.00000718, -0.00000081, 0.00000237, 0.00000423, 0.00000137, -0.00000413, -0.00000295, 0.00000445, -0.00000001, 0.00001572, -0.00000696, 0.00000012, -0.00000184, 0.00000598, -0.00000249, 0.00000443, 0.00000043, -0.00000694, 0.00000202, 0.00000774, -0.00000505, -0.00000168, -0.00000260, -0.00000258, -0.00000230, -0.00001257, 0.00000849, 0.00000012, -0.00000309, -0.00000590, -0.00000096, -0.00001469, 0.00000571, -0.00000134, 0.00000046, -0.00000419, -0.00000028, 0.00000587, -0.00000165, 0.00000265, 0.00000020, -0.00000471, -0.00000234, -0.00000445, -0.00000053, 0.00000027, -0.00000363, -0.00000198, -0.00000903, -0.00000100, -0.00002249, -0.00005540, 0.00003675, 0.00007181, 0.00003857, 0.00008362, -0.00001144, -0.00005398, -0.00000721, -0.00000867, -0.00000462, -0.00000596, -0.00000453, -0.00000688, -0.00000033, 0.00000073, },
		{ -0.00000011, -0.00000157, -0.00000157, 0.00000008, 0.00000054, -0.00000479, -0.00000617, -0.00000541, -0.00000434, -0.00000761, -0.00000724, -0.00002754, -0.00006403, 0.00006394, 0.00007008, 0.00005786, 0.00007821, -0.00004830, -0.00003963, -0.00000342, -0.00000703, -0.00000560, -0.00000181, -0.00000304, 0.00000176, -0.00000418, -0.00000209, -0.00000554, -0.00000125, 0.00000236, 0.00000033, -0.00000098, 0.00000822, -0.00000700, 0.00000279, -0.00000523, 0.00000650, -0.00000665, -0.00000990, -0.00000070, -0.00000669, -0.00000105, 0.00000719, -0.00000499, -0.00001012, 0.00000406, -0.00000919, 0.00000375, -0.00000888, 0.00000613, 0.00000329, -0.00000148, -0.00000640, 0.00000724, -0.00000308, 0.00000534, -0.00000204, 0.00000258, -0.00000757, 0.00000859, 0.00000840, -0.00000141, 0.00000447, -0.00000770, 0.00000049, 0.00000302, 0.00000365, 0.00000094, -0.00000560, -0.00000015, 0.00000236, 0.00000399, 0.00001300, -0.00000865, 0.00000128, -0.00000151, 0.00000682, -0.00000430, 0.00000699, -0.00000310, -0.00000562, 0.00000375, 0.00000683, -0.00000727, 0.00000057, -0.00000478, 0.00000041, -0.00000768, -0.00000754, 0.00000850, -0.00000114, -0.00000422, -0.00000427, -0.00000332, -0.00001307, 0.00000739, -0.00000279, 0.00000132, -0.00000591, 0.00000222, 0.00000435, -0.00000091, 0.00000270, -0.00000105, -0.00000471, -0.00000182, -0.00000516, 0.00000123, -0.00000137, -0.00000268, -0.00000337, -0.00000862, -0.00000007, -0.00003454, -0.00004270, 0.00005741, 0.00005689, 0.00004494, 0.00007501, -0.00002974, -0.00004415, -0.00000517, -0.00000811, -0.00000485, -0.00000568, -0.00000478, -0.00000640, 0.00000114, },
		{ 0.00000209, -0.00000105, -0.00000148, -0.00000144, 0.00000066, -0.00000048, -0.00000560, -0.00000593, -0.00000559, -0.00000369, -0.00000945, -0.00000618, -0.00003980, -0.00005435, 0.00008629, 0.00005866, 0.00007658, 0.00005586, -0.00006562, -0.00002511, -0.00000469, -0.00000667, -0.00000510, -0.00000179, -0.00000246, 0.00000131, -0.00000439, -0.00000222, -0.00000594, 0.00000031, 0.00000246, -0.00000104, 0.00000075, 0.00000721, -0.00000757, 0.00000403, -0.00000646, 0.00000848, -0.00001146, -0.00000607, -0.00000267, -0.00000560, -0.00000011, 0.00000824, -0.00001107, -0.00000390, 0.00000036, -0.00000695, 0.00000240, -0.00000808, 0.00000812, 0.00000139, -0.00000246, -0.00000444, 0.00000694, -0.00000333, 0.00000593, -0.00000354, 0.00000317, -0.00000842, 0.00001368, 0.00000294, 0.00000150, 0.00000163, -0.00000700, 0.00000137, 0.00000367, 0.00000285, 0.00000041, -0.00000650, 0.00000231, 0.00000044, 0.00000844, 0.00000832, -0.00000824, 0.00000118, 0.00000029, 0.00000576, -0.00000441, 0.00000829, -0.00000629, -0.00000377, 0.00000548, 0.00000461, -0.00000789, 0.00000160, -0.00000556, 0.00000161, -0.00001198, -0.00000147, 0.00000660, -0.00000157, -0.00000551, -0.00000235, -0.00000632, -0.00001008, 0.00000768, -0.00000335, 0.00000125, -0.00000678, 0.00000433, 0.00000290, -0.00000001, 0.00000234, -0.00000229, -0.00000414, -0.00000197, -0.00000524, 0.00000264, -0.00000320, -0.00000122, -0.00000576, -0.00000632, -0.00000264, -0.00004451, -0.00002285, 0.00006857, 0.00004299, 0.00005381, 0.00006052, -0.00004202, -0.00003423, -0.00000433, -0.00000744, -0.00000535, -0.00000504, -0.00000542, -0.00000526, },
		{ -0.00000351, 0.00000230, -0.00000180, -0.00000138, -0.00000109, 0.00000091, -0.00000160, -0.00000617, -0.00000574, -0.00000557, -0.00000337, -0.00001125, -0.00000570, -0.00005385, -0.00003597, 0.00010157, 0.00005043, 0.00009495, 0.00002199, -0.00006997, -0.00001347, -0.00000669, -0.00000601, -0.00000455, -0.00000199, -0.00000160, 0.00000036, -0.00000395, -0.00000320, -0.00000530, 0.00000132, 0.00000256, -0.00000272, 0.00000346, 0.00000477, -0.00000654, 0.00000372, -0.00000623, 0.00000880, -0.00001509, -0.00000238, -0.00000488, -0.00000429, 0.00000146, 0.00000711, -0.00001520, 0.00000243, -0.00000492, -0.00000272, -0.00000102, -0.00000495, 0.00000799, 0.00000037, -0.00000369, -0.00000168, 0.00000532, -0.00000216, 0.00000483, -0.00000346, 0.00000228, -0.00000698, 0.00001650, -0.00000182, 0.00000470, -0.00000191, -0.00000534, 0.00000191, 0.00000418, 0.00000201, -0.00000035, -0.00000658, 0.00000396, -0.00000052, 0.00001215, 0.00000285, -0.00000641, 0.00000030, 0.00000304, 0.00000305, -0.00000266, 0.00000786, -0.00000859, -0.00000157, 0.00000683, 0.00000157, -0.00000712, 0.00000147, -0.00000493, 0.00000053, -0.00001391, 0.00000408, 0.00000372, -0.00000150, -0.00000667, -0.00000056, -0.00000948, -0.00000613, 0.00000674, -0.00000306, 0.00000025, -0.00000663, 0.00000580, 0.00000178, 0.00000079, 0.00000169, -0.00000338, -0.00000314, -0.00000286, -0.00000432, 0.00000311, -0.00000449, -0.00000001, -0.00000835, -0.00000274, -0.00000925, -0.00004916, 0.00000025, 0.00007034, 0.00003311, 0.00006214, 0.00004219, -0.00004824, -0.00002541, -0.00000401, -0.00000703, -0.00000572, -0.00000437, -0.00000622, },
		{ -0.00000686, -0.00000137, 0.00000171, -0.00000215, -0.00000139, -0.00000045, 0.00000059, -0.00000251, -0.00000672, -0.00000539, -0.00000550, -0.00000344, -0.00001273, -0.00000668, -0.00006802, -0.00000918, 0.00010740, 0.00004981, 0.00010599, -0.00001745, -0.00006159, -0.00000765, -0.00000723, -0.00000619, -0.00000330, -0.00000283, -0.00000024, -0.00000115, -0.00000291, -0.00000496, -0.00000348, 0.00000150, 0.00000279, -0.00000458, 0.00000696, 0.00000097, -0.00000381, 0.00000160, -0.00000411, 0.00000686, -0.00001665, 0.00000026, -0.00000668, -0.00000303, 0.00000365, 0.00000343, -0.00001592, 0.00000667, -0.00000973, 0.00000198, -0.00000535, -0.00000021, 0.00000595, 0.00000037, -0.00000511, 0.00000152, 0.00000280, 0.00000004, 0.00000237, -0.00000193, 0.00000000, -0.00000314, 0.00001630, -0.00000467, 0.00000703, -0.00000531, -0.00000317, 0.00000226, 0.00000437, 0.00000133, -0.00000141, -0.00000576, 0.00000457, -0.00000003, 0.00001414, -0.00000217, -0.00000413, -0.00000055, 0.00000583, -0.00000052, 0.00000055, 0.00000563, -0.00000963, 0.00000076, 0.00000747, -0.00000169, -0.00000544, 0.00000066, -0.00000349, -0.00000249, -0.00001293, 0.00000788, 0.00000084, -0.00000139, -0.00000735, 0.00000069, -0.00001228, -0.00000174, 0.00000489, -0.00000212, -0.00000155, -0.00000542, 0.00000648, 0.00000116, 0.00000124, 0.00000093, -0.00000426, -0.00000191, -0.00000439, -0.00000227, 0.00000214, -0.00000449, -0.00000004, -0.00000985, 0.00000046, -0.00001876, -0.00004679, 0.00002250, 0.00006461, 0.00002871, 0.00006733, 0.00002251, -0.00004928, -0.00001826, -0.00000385, -0.00000707, -0.00000568, -0.00000396, },
		{ -0.00000402, -0.00000703, 0.00000079, 0.00000046, -0.00000201, -0.00000155, 0.00000044, -0.00000033, -0.00000304, -0.00000742, -0.00000475, -0.00000552, -0.00000396, -0.00001358, -0.00001016, -0.00007978, 0.00002398, 0.00010349, 0.00005940, 0.00010295, -0.00005396, -0.00004457, -0.00000771, -0.00000556, -0.00000759, -0.00000120, -0.00000434, 0.00000156, -0.00000303, -0.00000165, -0.00000703, -0.00000084, 0.00000094, 0.00000296, -0.00000611, 0.00001061, -0.00000365, 0.00000018, -0.00000203, -0.00000029, 0.00000258, -0.00001572, 0.00000123, -0.00000759, -0.00000192, 0.00000602, -0.00000235, -0.00001275, 0.00000729, -0.00001210, 0.00000524, -0.00000890, 0.00000471, 0.00000287, 0.00000101, -0.00000636, 0.00000459, 0.00000004, 0.00000254, -0.00000072, 0.00000041, -0.00000296, 0.00000243, 0.00001310, -0.00000495, 0.00000763, -0.00000773, -0.00000101, 0.00000260, 0.00000414, 0.00000089, -0.00000272, -0.00000417, 0.00000425, 0.00000191, 0.00001388, -0.00000577, -0.00000226, -0.00000055, 0.00000759, -0.00000382, 0.00000436, 0.00000202, -0.00000927, 0.00000297, 0.00000722, -0.00000457, -0.00000339, -0.00000030, -0.00000214, -0.00000644, -0.00000933, 0.00000930, -0.00000125, -0.00000172, -0.00000721, 0.00000095, -0.00001413, 0.00000249, 0.00000258, -0.00000090, -0.00000383, -0.00000331, 0.00000641, 0.00000104, 0.00000122, 0.00000018, -0.00000479, -0.00000083, -0.00000609, 0.00000058, -0.00000019, -0.00000294, -0.00000192, -0.00000909, 0.00000126, -0.00002868, -0.00003762, 0.00004065, 0.00005416, 0.00002981, 0.00006763, 0.00000398, -0.00004661, -0.00001269, -0.00000384, -0.00000742, -0.00000516, },
		{ -0.00000437, -0.00000455, -0.00000649, 0.00000255, -0.00000113, -0.00000148, -0.00000177, 0.00000132, -0.00000161, -0.00000329, -0.00000812, -0.00000395, -0.00000557, -0.00000511, -0.00001345, -0.00001734, -0.00008582, 0.00005957, 0.00009234, 0.00007835, 0.00008186, -0.00007898, -0.00002553, -0.00001109, -0.00000273, -0.00000942, 0.00000104, -0.00000592, 0.00000325, -0.00000470, -0.00000085, -0.00000853, 0.00000180, 0.00000022, 0.00000256, -0.00000649, 0.00001332, -0.00000795, 0.00000439, -0.00000623, 0.00000439, -0.00000334, -0.00001255, 0.00000047, -0.00000748, -0.00000079, 0.00000768, -0.00000888, -0.00000650, 0.00000398, -0.00001092, 0.00000565, -0.00001011, 0.00000820, 0.00000002, 0.00000153, -0.00000684, 0.00000684, -0.00000222, 0.00000452, -0.00000351, 0.00000265, -0.00000553, 0.00000846, 0.00000778, -0.00000281, 0.00000618, -0.00000867, 0.00000071, 0.00000303, 0.00000350, 0.00000067, -0.00000411, -0.00000209, 0.00000336, 0.00000486, 0.00001148, -0.00000753, -0.00000128, 0.00000072, 0.00000754, -0.00000574, 0.00000766, -0.00000224, -0.00000768, 0.00000484, 0.00000605, -0.00000670, -0.00000136, -0.00000105, -0.00000164, -0.00001009, -0.00000406, 0.00000841, -0.00000217, -0.00000276, -0.00000604, -0.00000010, -0.00001452, 0.00000589, 0.00000040, 0.00000010, -0.00000604, -0.00000070, 0.00000583, 0.00000120, 0.00000086, -0.00000059, -0.00000480, -0.00000038, -0.00000723, 0.00000334, -0.00000315, -0.00000038, -0.00000534, -0.00000590, -0.00000172, -0.00003613, -0.00002347, 0.00005285, 0.00004198, 0.00003524, 0.00006242, -0.00001139, -0.00004185, -0.00000829, -0.00000426, -0.00000773, },
		{ -0.00000760, -0.00000362, -0.00000538, -0.00000513, 0.00000347, -0.00000256, -0.00000090, -0.00000175, 0.00000177, -0.00000278, -0.00000368, -0.00000843, -0.00000341, -0.00000536, -0.00000719, -0.00001199, -0.00002920, -0.00008263, 0.00009234, 0.00007911, 0.00010166, 0.00004362, -0.00008687, -0.00001108, -0.00001396, -0.00000086, -0.00001013, 0.00000215, -0.00000647, 0.00000393, -0.00000539, -0.00000129, -0.00000851, 0.00000342, 0.00000010, 0.00000094, -0.00000481, 0.00001388, -0.00001058, 0.00000740, -0.00000961, 0.00000852, -0.00000960, -0.00000807, -0.00000157, -0.00000659, 0.00000069, 0.00000760, -0.00001423, 0.00000085, -0.00000206, -0.00000649, 0.00000285, -0.00000818, 0.00000912, -0.00000148, 0.00000118, -0.00000603, 0.00000765, -0.00000332, 0.00000528, -0.00000519, 0.00000393, -0.00000663, 0.00001343, 0.00000179, 0.00000090, 0.00000297, -0.00000801, 0.00000182, 0.00000352, 0.00000261, 0.00000053, -0.00000531, 0.00000009, 0.00000247, 0.00000803, 0.00000758, -0.00000762, -0.00000110, 0.00000313, 0.00000547, -0.00000557, 0.00000946, -0.00000623, -0.00000522, 0.00000623, 0.00000409, -0.00000783, 0.00000041, -0.00000150, -0.00000241, -0.00001227, 0.00000157, 0.00000581, -0.00000197, -0.00000447, -0.00000385, -0.00000256, -0.00001309, 0.00000790, -0.00000113, 0.00000034, -0.00000756, 0.00000186, 0.00000514, 0.00000126, 0.00000047, -0.00000153, -0.00000409, -0.00000102, -0.00000698, 0.00000489, -0.00000544, 0.00000188, -0.00000898, -0.00000140, -0.00000838, -0.00003901, -0.00000689, 0.00005848, 0.00003083, 0.00004298, 0.00005222, -0.00002251, -0.00003636, -0.00000463, -0.00000531, },
		{ -0.00000696, -0.00000684, -0.00000320, -0.00000627, -0.00000302, 0.00000325, -0.00000334, -0.00000062, -0.00000122, 0.00000138, -0.00000332, -0.00000468, -0.00000780, -0.00000365, -0.00000450, -0.00001044, -0.00000920, -0.00004599, -0.00006740, 0.00011687, 0.00007034, 0.00012084, -0.00000533, -0.00007705, -0.00000531, -0.00001327, -0.00000182, -0.00000833, 0.00000098, -0.00000506, 0.00000287, -0.00000459, -0.00000340, -0.00000637, 0.00000326, 0.00000114, -0.00000222, -0.00000060, 0.00001139, -0.00001039, 0.00000796, -0.00001080, 0.00001063, -0.00001463, -0.00000358, -0.00000403, -0.00000532, 0.00000273, 0.00000501, -0.00001652, 0.00000677, -0.00000854, -0.00000042, -0.00000216, -0.00000352, 0.00000725, -0.00000112, -0.00000041, -0.00000368, 0.00000674, -0.00000296, 0.00000448, -0.00000530, 0.00000377, -0.00000553, 0.00001601, -0.00000327, 0.00000486, -0.00000115, -0.00000610, 0.00000236, 0.00000394, 0.00000170, 0.00000024, -0.00000604, 0.00000197, 0.00000209, 0.00001054, 0.00000315, -0.00000667, -0.00000119, 0.00000598, 0.00000187, -0.00000324, 0.00000915, -0.00000913, -0.00000241, 0.00000708, 0.00000151, -0.00000790, 0.00000182, -0.00000186, -0.00000435, -0.00001229, 0.00000628, 0.00000248, -0.00000114, -0.00000647, -0.00000106, -0.00000616, -0.00000984, 0.00000823, -0.00000164, -0.00000053, -0.00000789, 0.00000386, 0.00000470, 0.00000087, 0.00000036, -0.00000274, -0.00000263, -0.00000290, -0.00000486, 0.00000431, -0.00000583, 0.00000232, -0.00001099, 0.00000222, -0.00001697, -0.00003674, 0.00000978, 0.00005785, 0.00002292, 0.00005050, 0.00003868, -0.00002933, -0.00003085, -0.00000163, },
		{ 0.00000046, -0.00000880, -0.00000554, -0.00000324, -0.00000692, -0.00000041, 0.00000185, -0.00000317, -0.00000093, -0.00000002, -0.00000004, -0.00000293, -0.00000656, -0.00000598, -0.00000500, -0.00000290, -0.00001470, -0.00000586, -0.00006645, -0.00003928, 0.00012944, 0.00007163, 0.00012633, -0.00005428, -0.00005445, -0.00000827, -0.00000826, -0.00000600, -0.00000383, -0.00000260, -0.00000155, 0.00000003, -0.00000244, -0.00000684, -0.00000233, 0.00000134, 0.00000320, -0.00000650, 0.00000579, 0.00000587, -0.00000703, 0.00000547, -0.00000905, 0.00000970, -0.00001718, -0.00000029, -0.00000610, -0.00000400, 0.00000516, -0.00000007, -0.00001469, 0.00000908, -0.00001291, 0.00000492, -0.00000741, 0.00000237, 0.00000346, 0.00000077, -0.00000299, -0.00000010, 0.00000430, -0.00000126, 0.00000225, -0.00000387, 0.00000222, -0.00000217, 0.00001548, -0.00000610, 0.00000774, -0.00000512, -0.00000355, 0.00000255, 0.00000411, 0.00000100, -0.00000038, -0.00000606, 0.00000332, 0.00000253, 0.00001169, -0.00000085, -0.00000543, -0.00000082, 0.00000823, -0.00000220, 0.00000063, 0.00000675, -0.00001046, 0.00000029, 0.00000737, -0.00000141, -0.00000691, 0.00000279, -0.00000242, -0.00000697, -0.00001001, 0.00000909, -0.00000054, -0.00000039, -0.00000809, 0.00000162, -0.00001022, -0.00000523, 0.00000701, -0.00000117, -0.00000250, -0.00000685, 0.00000508, 0.00000462, -0.00000006, 0.00000061, -0.00000411, -0.00000077, -0.00000562, -0.00000114, 0.00000146, -0.00000380, 0.00000005, -0.00000992, 0.00000270, -0.00002479, -0.00003021, 0.00002486, 0.00005190, 0.00001975, 0.00005534, 0.00002405, -0.00003271, -0.00002541, },
		{ -0.00001976, 0.00000116, -0.00001021, -0.00000407, -0.00000371, -0.00000706, 0.00000221, -0.00000043, -0.00000211, -0.00000173, 0.00000161, -0.00000225, -0.00000183, -0.00000904, -0.00000327, -0.00000716, -0.00000107, -0.00001904, -0.00000383, -0.00008739, -0.00000033, 0.00012961, 0.00008513, 0.00011082, -0.00009138, -0.00002789, -0.00001612, -0.00000093, -0.00001176, 0.00000197, -0.00000742, 0.00000305, -0.00000370, 0.00000008, -0.00001044, 0.00000243, -0.00000142, 0.00000535, -0.00001063, 0.00001294, -0.00000156, -0.00000126, 0.00000042, -0.00000460, 0.00000567, -0.00001675, 0.00000113, -0.00000728, -0.00000265, 0.00000728, -0.00000660, -0.00000907, 0.00000690, -0.00001335, 0.00000737, -0.00001069, 0.00000744, -0.00000067, 0.00000313, -0.00000570, 0.00000389, 0.00000101, 0.00000119, -0.00000085, -0.00000144, -0.00000011, 0.00000277, 0.00001201, -0.00000610, 0.00000857, -0.00000796, -0.00000102, 0.00000266, 0.00000389, 0.00000066, -0.00000144, -0.00000530, 0.00000405, 0.00000380, 0.00001120, -0.00000375, -0.00000449, 0.00000054, 0.00000891, -0.00000545, 0.00000491, 0.00000284, -0.00001015, 0.00000260, 0.00000707, -0.00000437, -0.00000494, 0.00000319, -0.00000343, -0.00000952, -0.00000589, 0.00000955, -0.00000243, -0.00000043, -0.00000859, 0.00000325, -0.00001372, -0.00000017, 0.00000483, -0.00000020, -0.00000512, -0.00000470, 0.00000561, 0.00000466, -0.00000125, 0.00000097, -0.00000522, 0.00000077, -0.00000817, 0.00000311, -0.00000270, -0.00000002, -0.00000454, -0.00000564, -0.00000119, -0.00002931, -0.00002116, 0.00003729, 0.00004211, 0.00002173, 0.00005567, 0.00001063, -0.00003395, },
		{ -0.00003415, -0.00001368, 0.00000015, -0.00001061, -0.00000284, -0.00000449, -0.00000642, 0.00000420, -0.00000290, -0.00000065, -0.00000259, 0.00000304, -0.00000450, -0.00000077, -0.00001123, -0.00000063, -0.00000917, -0.00000018, -0.00002189, -0.00000586, -0.00010379, 0.00004435, 0.00012098, 0.00010772, 0.00007257, -0.00010768, -0.00000676, -0.00002310, 0.00000504, -0.00001618, 0.00000655, -0.00001128, 0.00000689, -0.00000675, 0.00000146, -0.00001251, 0.00000618, -0.00000351, 0.00000611, -0.00001276, 0.00001871, -0.00000887, 0.00000516, -0.00000572, 0.00000129, -0.00000050, -0.00001377, 0.00000070, -0.00000753, -0.00000099, 0.00000801, -0.00001271, -0.00000141, 0.00000101, -0.00000961, 0.00000590, -0.00001050, 0.00000988, -0.00000344, 0.00000465, -0.00000742, 0.00000717, -0.00000215, 0.00000356, -0.00000395, 0.00000120, -0.00000224, 0.00000805, 0.00000657, -0.00000356, 0.00000708, -0.00000906, 0.00000095, 0.00000286, 0.00000329, 0.00000065, -0.00000283, -0.00000383, 0.00000433, 0.00000556, 0.00000924, -0.00000538, -0.00000398, 0.00000297, 0.00000752, -0.00000681, 0.00000833, -0.00000162, -0.00000854, 0.00000440, 0.00000609, -0.00000694, -0.00000222, 0.00000286, -0.00000496, -0.00001121, -0.00000079, 0.00000785, -0.00000282, -0.00000167, -0.00000744, 0.00000303, -0.00001558, 0.00000427, 0.00000256, 0.00000054, -0.00000761, -0.00000203, 0.00000588, 0.00000435, -0.00000216, 0.00000097, -0.00000551, 0.00000115, -0.00000928, 0.00000624, -0.00000644, 0.00000375, -0.00000968, 0.00000028, -0.00000880, -0.00002933, -0.00001117, 0.00004625, 0.00003052, 0.00002800, 0.00005088, -0.00000002, },
		{ -0.00000743, -0.00003382, -0.00000740, -0.00000242, -0.00000979, -0.00000214, -0.00000544, -0.00000479, 0.00000491, -0.00000468, 0.00000048, -0.00000282, 0.00000345, -0.00000581, -0.00000079, -0.00001195, 0.00000059, -0.00000970, -0.00000167, -0.00002150, -0.00001484, -0.00010992, 0.00008767, 0.00011040, 0.00013103, 0.00001705, -0.00010035, 0.00000245, -0.00002428, 0.00000618, -0.00001642, 0.00000744, -0.00001201, 0.00000813, -0.00000763, 0.00000042, -0.00001161, 0.00000733, -0.00000351, 0.00000418, -0.00001123, 0.00002096, -0.00001375, 0.00001008, -0.00001094, 0.00000673, -0.00000713, -0.00000943, -0.00000094, -0.00000714, 0.00000126, 0.00000634, -0.00001627, 0.00000561, -0.00000630, -0.00000310, 0.00000108, -0.00000668, 0.00000891, -0.00000378, 0.00000433, -0.00000721, 0.00000872, -0.00000421, 0.00000497, -0.00000621, 0.00000324, -0.00000322, 0.00001222, 0.00000068, 0.00000051, 0.00000369, -0.00000838, 0.00000215, 0.00000317, 0.00000246, 0.00000078, -0.00000431, -0.00000191, 0.00000442, 0.00000729, 0.00000639, -0.00000599, -0.00000361, 0.00000597, 0.00000426, -0.00000581, 0.00000992, -0.00000561, -0.00000619, 0.00000578, 0.00000433, -0.00000867, 0.00000086, 0.00000169, -0.00000687, -0.00001144, 0.00000416, 0.00000469, -0.00000193, -0.00000401, -0.00000461, 0.00000060, -0.00001509, 0.00000715, 0.00000102, 0.00000032, -0.00000914, 0.00000048, 0.00000629, 0.00000321, -0.00000224, 0.00000017, -0.00000460, -0.00000022, -0.00000800, 0.00000678, -0.00000790, 0.00000538, -0.00001291, 0.00000505, -0.00001770, -0.00002539, -0.00000104, 0.00005091, 0.00001967, 0.00003646, 0.00004181, },
		{ 0.00003050, -0.00001226, -0.00003271, -0.00000168, -0.00000587, -0.00000797, -0.00000203, -0.00000640, -0.00000218, 0.00000391, -0.00000507, 0.00000059, -0.00000188, 0.00000220, -0.00000542, -0.00000267, -0.00001030, -0.00000065, -0.00000777, -0.00000651, -0.00001695, -0.00003245, -0.00010100, 0.00012268, 0.00010547, 0.00014371, -0.00004389, -0.00007398, -0.00000140, -0.00001787, 0.00000121, -0.00001139, 0.00000359, -0.00000857, 0.00000587, -0.00000574, -0.00000337, -0.00000732, 0.00000523, -0.00000088, -0.00000082, -0.00000536, 0.00001838, -0.00001452, 0.00001171, -0.00001349, 0.00000996, -0.00001249, -0.00000513, -0.00000297, -0.00000643, 0.00000408, 0.00000194, -0.00001575, 0.00000940, -0.00001213, 0.00000359, -0.00000511, -0.00000052, 0.00000507, -0.00000165, 0.00000197, -0.00000477, 0.00000801, -0.00000453, 0.00000484, -0.00000704, 0.00000416, -0.00000243, 0.00001409, -0.00000405, 0.00000468, -0.00000062, -0.00000633, 0.00000263, 0.00000348, 0.00000164, 0.00000076, -0.00000553, 0.00000016, 0.00000461, 0.00000842, 0.00000335, -0.00000608, -0.00000281, 0.00000859, 0.00000001, -0.00000274, 0.00000930, -0.00000836, -0.00000367, 0.00000681, 0.00000174, -0.00000910, 0.00000374, -0.00000028, -0.00000886, -0.00000993, 0.00000788, 0.00000115, -0.00000048, -0.00000678, -0.00000073, -0.00000368, -0.00001215, 0.00000803, 0.00000060, -0.00000124, -0.00000921, 0.00000237, 0.00000702, 0.00000115, -0.00000131, -0.00000150, -0.00000256, -0.00000320, -0.00000422, 0.00000418, -0.00000608, 0.00000344, -0.00001223, 0.00000587, -0.00002460, -0.00001943, 0.00000924, 0.00005044, 0.00001220, 0.00004424, },
		{ 0.00004857, 0.00001942, -0.00001581, -0.00003015, 0.00000244, -0.00000917, -0.00000571, -0.00000239, -0.00000714, 0.00000104, 0.00000130, -0.00000385, -0.00000051, 0.00000028, -0.00000072, -0.00000327, -0.00000641, -0.00000624, -0.00000447, -0.00000341, -0.00001438, -0.00000904, -0.00005786, -0.00007516, 0.00014507, 0.00011133, 0.00013538, -0.00009547, -0.00003897, -0.00001396, -0.00000608, -0.00000816, -0.00000249, -0.00000388, -0.00000181, 0.00000076, -0.00000187, -0.00000887, -0.00000069, 0.00000068, 0.00000356, -0.00000784, 0.00000399, 0.00001121, -0.00001083, 0.00000940, -0.00001255, 0.00000997, -0.00001536, -0.00000195, -0.00000471, -0.00000547, 0.00000687, -0.00000448, -0.00001100, 0.00000849, -0.00001408, 0.00000776, -0.00001006, 0.00000572, -0.00000001, 0.00000193, -0.00000173, -0.00000060, 0.00000521, -0.00000308, 0.00000308, -0.00000630, 0.00000393, 0.00000012, 0.00001316, -0.00000647, 0.00000760, -0.00000469, -0.00000362, 0.00000266, 0.00000363, 0.00000105, 0.00000035, -0.00000617, 0.00000209, 0.00000502, 0.00000861, 0.00000072, -0.00000613, -0.00000109, 0.00000982, -0.00000394, 0.00000143, 0.00000676, -0.00000961, -0.00000130, 0.00000746, -0.00000151, -0.00000793, 0.00000582, -0.00000287, -0.00001055, -0.00000676, 0.00000961, -0.00000168, 0.00000056, -0.00000897, 0.00000306, -0.00000879, -0.00000740, 0.00000716, 0.00000116, -0.00000395, -0.00000781, 0.00000362, 0.00000780, -0.00000146, 0.00000032, -0.00000360, -0.00000014, -0.00000679, 0.00000099, -0.00000074, -0.00000148, -0.00000189, -0.00000725, 0.00000131, -0.00002683, -0.00001366, 0.00001993, 0.00004457, 0.00001008, },
		{ 0.00001385, 0.00004770, 0.00001064, -0.00001930, -0.00002560, 0.00000414, -0.00001136, -0.00000362, -0.00000309, -0.00000731, 0.00000415, -0.00000219, -0.00000149, -0.00000239, 0.00000301, -0.00000450, -0.00000024, -0.00001099, -0.00000094, -0.00000966, 0.00000190, -0.00002333, -0.00000072, -0.00008711, -0.00003454, 0.00015491, 0.00012776, 0.00010098, -0.00012487, -0.00000777, -0.00002754, 0.00000594, -0.00001776, 0.00000675, -0.00001197, 0.00000580, -0.00000511, 0.00000203, -0.00001390, 0.00000605, -0.00000432, 0.00000783, -0.00001458, 0.00001446, 0.00000126, -0.00000385, 0.00000387, -0.00000854, 0.00000689, -0.00001550, -0.00000032, -0.00000595, -0.00000398, 0.00000852, -0.00001112, -0.00000346, 0.00000321, -0.00001132, 0.00000774, -0.00001164, 0.00000973, -0.00000424, 0.00000532, -0.00000539, 0.00000409, 0.00000119, -0.00000039, 0.00000010, -0.00000430, 0.00000296, 0.00000378, 0.00000973, -0.00000617, 0.00000841, -0.00000754, -0.00000099, 0.00000255, 0.00000350, 0.00000081, -0.00000058, -0.00000603, 0.00000374, 0.00000561, 0.00000785, -0.00000123, -0.00000626, 0.00000167, 0.00000903, -0.00000637, 0.00000543, 0.00000312, -0.00000954, 0.00000082, 0.00000751, -0.00000498, -0.00000520, 0.00000656, -0.00000574, -0.00001151, -0.00000243, 0.00000909, -0.00000304, 0.00000038, -0.00000954, 0.00000543, -0.00001329, -0.00000206, 0.00000535, 0.00000200, -0.00000709, -0.00000542, 0.00000457, 0.00000804, -0.00000380, 0.00000186, -0.00000531, 0.00000157, -0.00000946, 0.00000573, -0.00000597, 0.00000391, -0.00000866, 0.00000031, -0.00000778, -0.00002360, -0.00000937, 0.00003069, 0.00003411, },
		{ 0.00002130, 0.00002227, 0.00004159, 0.00000501, -0.00002323, -0.00001913, 0.00000319, -0.00001191, -0.00000212, -0.00000405, -0.00000647, 0.00000616, -0.00000539, 0.00000095, -0.00000402, 0.00000510, -0.00000772, 0.00000211, -0.00001457, 0.00000352, -0.00001400, 0.00000560, -0.00003019, 0.00000346, -0.00011362, 0.00001502, 0.00015680, 0.00014819, 0.00004368, -0.00012581, 0.00000980, -0.00003462, 0.00001267, -0.00002298, 0.00001228, -0.00001708, 0.00001116, -0.00000920, 0.00000368, -0.00001606, 0.00001035, -0.00000736, 0.00000956, -0.00001825, 0.00002291, -0.00000858, 0.00000411, -0.00000307, -0.00000294, 0.00000184, -0.00001352, -0.00000002, -0.00000681, -0.00000158, 0.00000780, -0.00001569, 0.00000417, -0.00000432, -0.00000493, 0.00000357, -0.00000906, 0.00001010, -0.00000596, 0.00000697, -0.00000754, 0.00000783, -0.00000278, 0.00000258, -0.00000333, -0.00000165, 0.00000197, 0.00000749, 0.00000482, -0.00000359, 0.00000696, -0.00000863, 0.00000104, 0.00000250, 0.00000309, 0.00000086, -0.00000195, -0.00000510, 0.00000507, 0.00000613, 0.00000644, -0.00000261, -0.00000625, 0.00000504, 0.00000622, -0.00000656, 0.00000813, -0.00000066, -0.00000859, 0.00000282, 0.00000664, -0.00000798, -0.00000135, 0.00000569, -0.00000852, -0.00001139, 0.00000224, 0.00000669, -0.00000277, -0.00000137, -0.00000792, 0.00000537, -0.00001585, 0.00000255, 0.00000363, 0.00000223, -0.00000971, -0.00000276, 0.00000567, 0.00000710, -0.00000500, 0.00000248, -0.00000580, 0.00000155, -0.00000975, 0.00000800, -0.00000913, 0.00000742, -0.00001386, 0.00000711, -0.00001833, -0.00001641, -0.00000627, 0.00004016, },
		{ 0.00004612, 0.00000942, 0.00003248, 0.00003188, 0.00000205, -0.00002715, -0.00001162, 0.00000015, -0.00001088, -0.00000132, -0.00000524, -0.00000431, 0.00000620, -0.00000707, 0.00000229, -0.00000432, 0.00000530, -0.00000894, 0.00000219, -0.00001532, 0.00000497, -0.00001517, 0.00000508, -0.00003181, -0.00000111, -0.00012998, 0.00006582, 0.00015783, 0.00016134, -0.00002512, -0.00010092, 0.00001005, -0.00003120, 0.00001085, -0.00002094, 0.00001140, -0.00001666, 0.00001198, -0.00000969, 0.00000166, -0.00001382, 0.00001046, -0.00000678, 0.00000710, -0.00001671, 0.00002659, -0.00001542, 0.00001049, -0.00000925, 0.00000240, -0.00000359, -0.00001052, -0.00000050, -0.00000748, 0.00000182, 0.00000406, -0.00001629, 0.00000900, -0.00001098, 0.00000247, -0.00000295, -0.00000324, 0.00000689, -0.00000460, 0.00000605, -0.00000721, 0.00000943, -0.00000550, 0.00000479, -0.00000634, 0.00000099, 0.00000162, 0.00001009, -0.00000017, 0.00000020, 0.00000378, -0.00000798, 0.00000224, 0.00000259, 0.00000253, 0.00000099, -0.00000349, -0.00000351, 0.00000612, 0.00000629, 0.00000483, -0.00000375, -0.00000558, 0.00000812, 0.00000216, -0.00000457, 0.00000896, -0.00000378, -0.00000724, 0.00000478, 0.00000459, -0.00000972, 0.00000283, 0.00000330, -0.00001085, -0.00000996, 0.00000628, 0.00000327, -0.00000133, -0.00000430, -0.00000431, 0.00000260, -0.00001567, 0.00000545, 0.00000278, 0.00000117, -0.00001105, -0.00000042, 0.00000717, 0.00000471, -0.00000458, 0.00000172, -0.00000475, -0.00000051, -0.00000706, 0.00000667, -0.00000856, 0.00000694, -0.00001475, 0.00000960, -0.00002626, -0.00000836, -0.00000273, },
		{ 0.00000321, 0.00004641, 0.00000188, 0.00004096, 0.00002126, 0.00000024, -0.00002994, -0.00000447, -0.00000383, -0.00000882, -0.00000111, -0.00000652, -0.00000094, 0.00000396, -0.00000652, 0.00000179, -0.00000267, 0.00000297, -0.00000740, -0.00000078, -0.00001228, 0.00000223, -0.00001188, -0.00000106, -0.00002657, -0.00001719, -0.00013029, 0.00011108, 0.00016412, 0.00015534, -0.00008899, -0.00006082, -0.00000359, -0.00001841, 0.00000119, -0.00001200, 0.00000418, -0.00001049, 0.00000794, -0.00000648, -0.00000378, -0.00000744, 0.00000633, -0.00000249, 0.00000043, -0.00000948, 0.00002416, -0.00001740, 0.00001340, -0.00001297, 0.00000594, -0.00000803, -0.00000752, -0.00000129, -0.00000785, 0.00000563, -0.00000225, -0.00001233, 0.00000908, -0.00001396, 0.00000782, -0.00000901, 0.00000357, 0.00000155, -0.00000088, 0.00000286, -0.00000433, 0.00000837, -0.00000619, 0.00000549, -0.00000824, 0.00000315, 0.00000224, 0.00001080, -0.00000391, 0.00000387, -0.00000014, -0.00000606, 0.00000268, 0.00000276, 0.00000199, 0.00000093, -0.00000489, -0.00000147, 0.00000692, 0.00000592, 0.00000343, -0.00000495, -0.00000380, 0.00000987, -0.00000193, -0.00000119, 0.00000799, -0.00000592, -0.00000574, 0.00000661, 0.00000138, -0.00000959, 0.00000640, -0.00000021, -0.00001245, -0.00000718, 0.00000879, -0.00000003, 0.00000030, -0.00000743, 0.00000033, -0.00000223, -0.00001287, 0.00000637, 0.00000302, -0.00000134, -0.00001077, 0.00000139, 0.00000882, 0.00000120, -0.00000269, -0.00000024, -0.00000259, -0.00000393, -0.00000210, 0.00000212, -0.00000430, 0.00000209, -0.00001033, 0.00000570, -0.00002822, -0.00000272, },
		{ -0.00000129, 0.00001245, 0.00004008, 0.00000091, 0.00004476, 0.00001239, -0.00000208, -0.00003040, 0.00000089, -0.00000749, -0.00000644, -0.00000136, -0.00000753, 0.00000299, -0.00000004, -0.00000392, -0.00000042, 0.00000062, -0.00000144, -0.00000357, -0.00000623, -0.00000608, -0.00000415, -0.00000477, -0.00001202, -0.00001557, -0.00004416, -0.00011222, 0.00014749, 0.00017730, 0.00012273, -0.00013195, -0.00002011, -0.00002246, -0.00000184, -0.00001196, 0.00000026, -0.00000644, -0.00000090, 0.00000086, -0.00000132, -0.00001065, 0.00000105, -0.00000025, 0.00000379, -0.00000859, 0.00000190, 0.00001623, -0.00001433, 0.00001231, -0.00001363, 0.00000707, -0.00001081, -0.00000502, -0.00000230, -0.00000746, 0.00000862, -0.00000944, -0.00000498, 0.00000434, -0.00001193, 0.00000891, -0.00001200, 0.00000878, -0.00000370, 0.00000355, -0.00000137, 0.00000020, 0.00000505, -0.00000478, 0.00000438, -0.00000867, 0.00000469, 0.00000373, 0.00000941, -0.00000559, 0.00000631, -0.00000374, -0.00000355, 0.00000262, 0.00000291, 0.00000160, 0.00000050, -0.00000583, 0.00000080, 0.00000738, 0.00000507, 0.00000237, -0.00000626, -0.00000079, 0.00000959, -0.00000478, 0.00000240, 0.00000586, -0.00000717, -0.00000402, 0.00000794, -0.00000257, -0.00000741, 0.00000850, -0.00000423, -0.00001311, -0.00000333, 0.00000928, -0.00000221, 0.00000107, -0.00000950, 0.00000451, -0.00000777, -0.00000833, 0.00000578, 0.00000394, -0.00000484, -0.00000910, 0.00000289, 0.00001000, -0.00000251, -0.00000015, -0.00000259, -0.00000038, -0.00000722, 0.00000328, -0.00000371, 0.00000183, -0.00000538, -0.00000207, -0.00000398, -0.00002318, },
		{ -0.00001297, -0.00000352, 0.00002405, 0.00002821, 0.00000673, 0.00004255, 0.00000695, -0.00000595, -0.00002794, 0.00000360, -0.00000992, -0.00000428, -0.00000205, -0.00000766, 0.00000632, -0.00000452, -0.00000036, -0.00000329, 0.00000428, -0.00000647, 0.00000093, -0.00001227, 0.00000112, -0.00001188, 0.00000355, -0.00002473, -0.00000273, -0.00007764, -0.00007765, 0.00017616, 0.00019257, 0.00006448, -0.00014420, 0.00000807, -0.00003679, 0.00001131, -0.00002262, 0.00001067, -0.00001595, 0.00000831, -0.00000613, 0.00000301, -0.00001599, 0.00000855, -0.00000639, 0.00000920, -0.00001681, 0.00001432, 0.00000517, -0.00000763, 0.00000799, -0.00001170, 0.00000610, -0.00001201, -0.00000303, -0.00000372, -0.00000565, 0.00000926, -0.00001503, 0.00000310, -0.00000327, -0.00000564, 0.00000531, -0.00001061, 0.00001045, -0.00000678, 0.00000682, -0.00000497, 0.00000488, 0.00000055, -0.00000185, 0.00000174, -0.00000767, 0.00000575, 0.00000553, 0.00000641, -0.00000510, 0.00000692, -0.00000623, -0.00000111, 0.00000233, 0.00000298, 0.00000138, -0.00000034, -0.00000614, 0.00000315, 0.00000731, 0.00000405, 0.00000146, -0.00000732, 0.00000299, 0.00000720, -0.00000559, 0.00000512, 0.00000340, -0.00000791, -0.00000188, 0.00000821, -0.00000641, -0.00000353, 0.00000861, -0.00000815, -0.00001268, 0.00000101, 0.00000779, -0.00000277, 0.00000029, -0.00000952, 0.00000682, -0.00001242, -0.00000343, 0.00000462, 0.00000469, -0.00000844, -0.00000663, 0.00000447, 0.00000992, -0.00000531, 0.00000197, -0.00000432, 0.00000069, -0.00000870, 0.00000683, -0.00000820, 0.00000691, -0.00001223, 0.00000651, -0.00001620, },
		{ -0.00002622, -0.00000159, -0.00000674, 0.00003521, 0.00001386, 0.00001725, 0.00003503, 0.00000506, -0.00001135, -0.00002289, 0.00000365, -0.00001081, -0.00000258, -0.00000324, -0.00000634, 0.00000781, -0.00000787, 0.00000258, -0.00000533, 0.00000662, -0.00001015, 0.00000394, -0.00001640, 0.00000641, -0.00001775, 0.00000947, -0.00003489, 0.00000639, -0.00011072, -0.00003154, 0.00020131, 0.00019974, -0.00000906, -0.00012566, 0.00001684, -0.00003989, 0.00001559, -0.00002599, 0.00001485, -0.00002031, 0.00001358, -0.00001013, 0.00000406, -0.00001727, 0.00001230, -0.00000937, 0.00001103, -0.00002103, 0.00002420, -0.00000584, 0.00000038, 0.00000206, -0.00000835, 0.00000390, -0.00001218, -0.00000131, -0.00000570, -0.00000211, 0.00000648, -0.00001673, 0.00000873, -0.00001051, 0.00000231, -0.00000138, -0.00000535, 0.00000817, -0.00000663, 0.00000764, -0.00000653, 0.00000817, -0.00000372, 0.00000158, -0.00000181, -0.00000553, 0.00000660, 0.00000691, 0.00000273, -0.00000301, 0.00000575, -0.00000722, 0.00000081, 0.00000202, 0.00000298, 0.00000122, -0.00000150, -0.00000575, 0.00000544, 0.00000655, 0.00000329, 0.00000032, -0.00000752, 0.00000657, 0.00000344, -0.00000434, 0.00000637, 0.00000124, -0.00000845, 0.00000082, 0.00000693, -0.00000914, 0.00000118, 0.00000673, -0.00001145, -0.00001102, 0.00000501, 0.00000494, -0.00000188, -0.00000203, -0.00000717, 0.00000638, -0.00001490, 0.00000058, 0.00000383, 0.00000444, -0.00001126, -0.00000397, 0.00000638, 0.00000814, -0.00000636, 0.00000282, -0.00000475, -0.00000005, -0.00000741, 0.00000698, -0.00000921, 0.00000828, -0.00001514, 0.00001127, },
		{ 0.00000935, -0.00002985, 0.00000656, -0.00000747, 0.00004236, 0.00000120, 0.00002888, 0.00002458, 0.00000549, -0.00001728, -0.00001637, 0.00000176, -0.00001041, -0.00000134, -0.00000489, -0.00000335, 0.00000668, -0.00000882, 0.00000360, -0.00000530, 0.00000631, -0.00001096, 0.00000378, -0.00001666, 0.00000744, -0.00001908, 0.00000984, -0.00003875, 0.00000657, -0.00013613, 0.00002054, 0.00022705, 0.00018705, -0.00008091, -0.00008595, 0.00000755, -0.00003084, 0.00000982, -0.00002070, 0.00001125, -0.00001778, 0.00001318, -0.00000981, 0.00000095, -0.00001359, 0.00001110, -0.00000794, 0.00000791, -0.00001929, 0.00002881, -0.00001397, 0.00000738, -0.00000383, -0.00000474, 0.00000133, -0.00001179, 0.00000020, -0.00000795, 0.00000269, 0.00000042, -0.00001350, 0.00000954, -0.00001419, 0.00000856, -0.00000828, 0.00000168, 0.00000309, -0.00000356, 0.00000581, -0.00000546, 0.00000903, -0.00000653, 0.00000443, -0.00000547, -0.00000263, 0.00000739, 0.00000729, -0.00000059, -0.00000024, 0.00000333, -0.00000681, 0.00000201, 0.00000183, 0.00000293, 0.00000098, -0.00000276, -0.00000468, 0.00000749, 0.00000507, 0.00000308, -0.00000136, -0.00000630, 0.00000883, -0.00000046, -0.00000176, 0.00000625, -0.00000045, -0.00000879, 0.00000390, 0.00000398, -0.00000990, 0.00000559, 0.00000325, -0.00001379, -0.00000813, 0.00000784, 0.00000175, -0.00000033, -0.00000518, -0.00000302, 0.00000326, -0.00001471, 0.00000298, 0.00000395, 0.00000272, -0.00001271, -0.00000144, 0.00000847, 0.00000483, -0.00000554, 0.00000223, -0.00000392, -0.00000229, -0.00000362, 0.00000364, -0.00000618, 0.00000493, -0.00001239, },
		{ -0.00000488, 0.00000068, -0.00002534, 0.00000863, -0.00000317, 0.00004272, -0.00000604, 0.00003781, 0.00001420, 0.00000636, -0.00002231, -0.00000981, -0.00000101, -0.00000919, -0.00000057, -0.00000664, 0.00000083, 0.00000306, -0.00000707, 0.00000229, -0.00000291, 0.00000307, -0.00000858, 0.00000014, -0.00001265, 0.00000351, -0.00001498, 0.00000354, -0.00003478, -0.00000500, -0.00014859, 0.00007498, 0.00025370, 0.00014662, -0.00013319, -0.00004056, -0.00001157, -0.00001439, -0.00000257, -0.00000937, 0.00000181, -0.00000968, 0.00000797, -0.00000604, -0.00000512, -0.00000619, 0.00000585, -0.00000283, 0.00000055, -0.00001168, 0.00002720, -0.00001764, 0.00001187, -0.00000860, -0.00000153, -0.00000125, -0.00001089, 0.00000114, -0.00000964, 0.00000732, -0.00000732, -0.00000623, 0.00000507, -0.00001251, 0.00001044, -0.00001239, 0.00000767, -0.00000258, 0.00000094, 0.00000224, -0.00000219, 0.00000731, -0.00000721, 0.00000588, -0.00000851, 0.00000072, 0.00000808, 0.00000643, -0.00000277, 0.00000224, 0.00000046, -0.00000539, 0.00000251, 0.00000181, 0.00000284, 0.00000056, -0.00000393, -0.00000296, 0.00000898, 0.00000312, 0.00000342, -0.00000357, -0.00000349, 0.00000894, -0.00000316, 0.00000098, 0.00000539, -0.00000194, -0.00000859, 0.00000675, -0.00000019, -0.00000832, 0.00000866, -0.00000114, -0.00001497, -0.00000423, 0.00000887, -0.00000074, 0.00000084, -0.00000794, 0.00000164, -0.00000158, -0.00001223, 0.00000380, 0.00000491, -0.00000036, -0.00001265, 0.00000098, 0.00001011, 0.00000082, -0.00000343, 0.00000069, -0.00000260, -0.00000484, 0.00000109, -0.00000160, -0.00000057, -0.00000187, },
		{ -0.00000902, 0.00000410, -0.00001178, -0.00001419, 0.00000463, 0.00000640, 0.00003560, -0.00000591, 0.00004134, 0.00000637, 0.00000605, -0.00002530, -0.00000439, -0.00000376, -0.00000756, -0.00000042, -0.00000776, 0.00000504, -0.00000190, -0.00000346, -0.00000062, 0.00000093, -0.00000203, -0.00000410, -0.00000569, -0.00000581, -0.00000393, -0.00000690, -0.00000798, -0.00002445, -0.00002771, -0.00014595, 0.00013143, 0.00027554, 0.00007914, -0.00015370, -0.00000484, -0.00002963, 0.00000174, -0.00001535, 0.00000279, -0.00000904, 0.00000039, 0.00000079, -0.00000134, -0.00001142, 0.00000212, -0.00000087, 0.00000349, -0.00000853, -0.00000029, 0.00002032, -0.00001679, 0.00001345, -0.00001192, 0.00000123, -0.00000394, -0.00000917, 0.00000083, -0.00000959, 0.00000986, -0.00001408, 0.00000242, -0.00000290, -0.00000597, 0.00000713, -0.00001195, 0.00001034, -0.00000655, 0.00000495, -0.00000151, 0.00000203, 0.00000366, -0.00000580, 0.00000557, -0.00001042, 0.00000426, 0.00000838, 0.00000461, -0.00000353, 0.00000377, -0.00000212, -0.00000347, 0.00000245, 0.00000201, 0.00000264, -0.00000004, -0.00000488, -0.00000062, 0.00000954, 0.00000124, 0.00000393, -0.00000585, 0.00000044, 0.00000680, -0.00000384, 0.00000285, 0.00000456, -0.00000368, -0.00000733, 0.00000852, -0.00000463, -0.00000471, 0.00000973, -0.00000571, -0.00001485, 0.00000018, 0.00000799, -0.00000184, 0.00000075, -0.00000915, 0.00000531, -0.00000668, -0.00000850, 0.00000364, 0.00000612, -0.00000422, -0.00001136, 0.00000348, 0.00001057, -0.00000275, -0.00000105, -0.00000090, -0.00000183, -0.00000612, 0.00000457, -0.00000616, 0.00000480, },
		{ 0.00000708, -0.00001299, 0.00001021, -0.00002312, -0.00000062, -0.00000268, 0.00001894, 0.00002292, 0.00000104, 0.00003875, 0.00000227, 0.00000376, -0.00002581, -0.00000066, -0.00000599, -0.00000571, -0.00000111, -0.00000744, 0.00000782, -0.00000645, 0.00000040, -0.00000363, 0.00000453, -0.00000708, 0.00000039, -0.00001134, 0.00000113, -0.00001194, 0.00000195, -0.00002121, -0.00001172, -0.00005801, -0.00012867, 0.00019170, 0.00028154, -0.00000454, -0.00014048, 0.00001175, -0.00003803, 0.00001070, -0.00002262, 0.00001050, -0.00001657, 0.00000836, -0.00000507, 0.00000148, -0.00001505, 0.00000822, -0.00000606, 0.00000806, -0.00001606, 0.00001167, 0.00001046, -0.00001254, 0.00001248, -0.00001396, 0.00000376, -0.00000704, -0.00000630, -0.00000123, -0.00000698, 0.00000867, -0.00001708, 0.00000896, -0.00001093, 0.00000283, 0.00000001, -0.00000710, 0.00000887, -0.00000741, 0.00000691, -0.00000388, 0.00000569, -0.00000069, -0.00000290, 0.00000356, -0.00001089, 0.00000777, 0.00000792, 0.00000241, -0.00000309, 0.00000415, -0.00000395, -0.00000151, 0.00000204, 0.00000243, 0.00000224, -0.00000072, -0.00000553, 0.00000226, 0.00000883, 0.00000006, 0.00000398, -0.00000737, 0.00000442, 0.00000316, -0.00000253, 0.00000337, 0.00000424, -0.00000597, -0.00000468, 0.00000840, -0.00000807, 0.00000003, 0.00000868, -0.00000989, -0.00001330, 0.00000434, 0.00000566, -0.00000150, -0.00000086, -0.00000817, 0.00000684, -0.00001062, -0.00000472, 0.00000335, 0.00000680, -0.00000810, -0.00000920, 0.00000616, 0.00000938, -0.00000494, 0.00000066, -0.00000183, -0.00000230, -0.00000511, 0.00000511, -0.00000776, },
		{ -0.00000555, 0.00000488, -0.00001161, 0.00001010, -0.00002864, 0.00001030, -0.00000902, 0.00003047, 0.00000848, 0.00001216, 0.00003126, 0.00000165, -0.00000033, -0.00002412, 0.00000146, -0.00000760, -0.00000371, -0.00000273, -0.00000519, 0.00000802, -0.00000898, 0.00000289, -0.00000517, 0.00000621, -0.00001020, 0.00000287, -0.00001450, 0.00000541, -0.00001740, 0.00000803, -0.00003215, -0.00000133, -0.00009109, -0.00009774, 0.00025667, 0.00025938, -0.00008592, -0.00010255, 0.00000866, -0.00003401, 0.00000972, -0.00002162, 0.00001099, -0.00001795, 0.00001157, -0.00000751, 0.00000082, -0.00001437, 0.00001016, -0.00000769, 0.00000873, -0.00001937, 0.00002113, 0.00000022, -0.00000643, 0.00000958, -0.00001481, 0.00000603, -0.00001045, -0.00000245, -0.00000483, -0.00000196, 0.00000339, -0.00001475, 0.00001053, -0.00001540, 0.00001021, -0.00000789, 0.00000015, 0.00000414, -0.00000514, 0.00000631, -0.00000400, 0.00000756, -0.00000448, 0.00000055, 0.00000028, -0.00000977, 0.00001094, 0.00000649, 0.00000054, -0.00000203, 0.00000359, -0.00000487, 0.00000018, 0.00000149, 0.00000301, 0.00000156, -0.00000139, -0.00000578, 0.00000545, 0.00000676, 0.00000006, 0.00000299, -0.00000730, 0.00000711, -0.00000058, -0.00000007, 0.00000282, 0.00000432, -0.00000859, -0.00000081, 0.00000605, -0.00000945, 0.00000467, 0.00000588, -0.00001326, -0.00001033, 0.00000740, 0.00000279, -0.00000032, -0.00000351, -0.00000519, 0.00000582, -0.00001255, -0.00000178, 0.00000355, 0.00000639, -0.00001136, -0.00000638, 0.00000871, 0.00000665, -0.00000544, 0.00000134, -0.00000201, -0.00000381, -0.00000196, 0.00000242, },
		{ -0.00000207, -0.00000072, -0.00000089, -0.00000529, 0.00000309, -0.00002587, 0.00001480, -0.00001061, 0.00003707, -0.00000342, 0.00002374, 0.00002130, 0.00000328, -0.00000544, -0.00002093, 0.00000230, -0.00000862, -0.00000180, -0.00000489, -0.00000130, 0.00000544, -0.00000876, 0.00000316, -0.00000450, 0.00000519, -0.00001047, 0.00000238, -0.00001406, 0.00000555, -0.00001846, 0.00000908, -0.00003807, 0.00000306, -0.00012286, -0.00005247, 0.00032247, 0.00020137, -0.00014505, -0.00005620, -0.00000665, -0.00002118, 0.00000113, -0.00001384, 0.00000511, -0.00001344, 0.00000992, -0.00000656, -0.00000287, -0.00000979, 0.00000800, -0.00000563, 0.00000521, -0.00001745, 0.00002620, -0.00000843, 0.00000027, 0.00000513, -0.00001420, 0.00000746, -0.00001344, 0.00000148, -0.00000882, 0.00000405, -0.00000463, -0.00000759, 0.00000620, -0.00001405, 0.00001300, -0.00001322, 0.00000685, -0.00000169, -0.00000102, 0.00000375, -0.00000197, 0.00000713, -0.00000674, 0.00000362, -0.00000369, -0.00000703, 0.00001329, 0.00000423, -0.00000057, -0.00000095, 0.00000249, -0.00000496, 0.00000140, 0.00000100, 0.00000363, 0.00000063, -0.00000199, -0.00000546, 0.00000849, 0.00000362, 0.00000132, 0.00000081, -0.00000526, 0.00000748, -0.00000295, 0.00000220, 0.00000209, 0.00000415, -0.00001082, 0.00000350, 0.00000187, -0.00000824, 0.00000807, 0.00000197, -0.00001556, -0.00000611, 0.00000868, 0.00000042, 0.00000073, -0.00000620, -0.00000117, 0.00000277, -0.00001239, 0.00000003, 0.00000443, 0.00000470, -0.00001364, -0.00000294, 0.00001050, 0.00000318, -0.00000464, 0.00000129, -0.00000201, -0.00000530, 0.00000190, },
		{ 0.00000439, -0.00000591, 0.00000405, -0.00000734, 0.00000300, -0.00000831, -0.00001575, 0.00001201, -0.00000588, 0.00003653, -0.00000963, 0.00003246, 0.00001142, 0.00000571, -0.00001071, -0.00001688, 0.00000217, -0.00000896, -0.00000042, -0.00000674, 0.00000310, 0.00000101, -0.00000629, 0.00000153, -0.00000206, 0.00000197, -0.00000837, -0.00000051, -0.00001064, 0.00000201, -0.00001536, 0.00000501, -0.00003848, -0.00000009, -0.00015083, 0.00001039, 0.00037826, 0.00010988, -0.00016796, -0.00001841, -0.00002291, -0.00000706, -0.00000929, -0.00000398, -0.00000333, -0.00000597, 0.00000558, -0.00000416, -0.00000744, -0.00000355, 0.00000362, -0.00000155, -0.00000097, -0.00001117, 0.00002657, -0.00001451, 0.00000679, -0.00000077, -0.00001151, 0.00000705, -0.00001478, 0.00000401, -0.00001143, 0.00000870, -0.00001254, 0.00000178, -0.00000239, -0.00000702, 0.00000992, -0.00001372, 0.00001039, -0.00000618, 0.00000304, 0.00000064, 0.00000124, 0.00000473, -0.00000713, 0.00000561, -0.00000771, -0.00000279, 0.00001429, 0.00000164, -0.00000086, -0.00000020, 0.00000127, -0.00000443, 0.00000206, 0.00000080, 0.00000404, -0.00000036, -0.00000260, -0.00000432, 0.00001070, 0.00000016, 0.00000334, -0.00000212, -0.00000161, 0.00000531, -0.00000298, 0.00000312, 0.00000214, 0.00000288, -0.00001165, 0.00000702, -0.00000303, -0.00000472, 0.00000952, -0.00000237, -0.00001654, -0.00000117, 0.00000800, -0.00000072, 0.00000077, -0.00000779, 0.00000258, -0.00000124, -0.00001074, 0.00000095, 0.00000570, 0.00000191, -0.00001482, 0.00000113, 0.00001080, 0.00000004, -0.00000342, 0.00000116, -0.00000270, -0.00000534, },
		{ -0.00000303, 0.00000386, -0.00000677, 0.00000588, -0.00001095, 0.00000898, -0.00001939, -0.00000223, 0.00000424, 0.00000390, 0.00002922, -0.00000909, 0.00003639, 0.00000349, 0.00000782, -0.00001559, -0.00001232, 0.00000113, -0.00000839, -0.00000015, -0.00000720, 0.00000641, -0.00000356, -0.00000301, -0.00000071, 0.00000077, -0.00000196, -0.00000545, -0.00000407, -0.00000621, -0.00000319, -0.00001007, -0.00000240, -0.00003486, -0.00001048, -0.00017327, 0.00009496, 0.00040728, -0.00000065, -0.00015232, -0.00000014, -0.00003068, 0.00000118, -0.00001552, 0.00000274, -0.00000975, 0.00000064, 0.00000152, -0.00000279, -0.00001025, 0.00000157, -0.00000035, 0.00000206, -0.00000725, -0.00000262, 0.00002315, -0.00001780, 0.00001261, -0.00000791, -0.00000636, 0.00000394, -0.00001337, 0.00000381, -0.00001102, 0.00000968, -0.00001706, 0.00000950, -0.00001163, 0.00000303, 0.00000218, -0.00000920, 0.00000956, -0.00000768, 0.00000539, -0.00000151, 0.00000428, 0.00000126, -0.00000582, 0.00000612, -0.00001111, 0.00000263, 0.00001346, -0.00000056, -0.00000067, 0.00000016, 0.00000015, -0.00000347, 0.00000216, 0.00000109, 0.00000395, -0.00000116, -0.00000336, -0.00000213, 0.00001137, -0.00000262, 0.00000522, -0.00000481, 0.00000252, 0.00000138, -0.00000075, 0.00000225, 0.00000352, -0.00000001, -0.00001032, 0.00000850, -0.00000712, 0.00000008, 0.00000893, -0.00000664, -0.00001594, 0.00000366, 0.00000578, -0.00000049, -0.00000056, -0.00000754, 0.00000489, -0.00000496, -0.00000848, 0.00000152, 0.00000689, -0.00000161, -0.00001480, 0.00000561, 0.00000925, -0.00000191, -0.00000254, 0.00000146, -0.00000455, },
		{ -0.00000712, 0.00000126, 0.00000015, -0.00000384, 0.00000331, -0.00000940, 0.00000915, -0.00002532, 0.00000953, -0.00000428, 0.00001528, 0.00001777, -0.00000295, 0.00003527, -0.00000163, 0.00000895, -0.00001978, -0.00000739, -0.00000084, -0.00000677, -0.00000131, -0.00000554, 0.00000731, -0.00000657, -0.00000050, -0.00000214, 0.00000248, -0.00000497, -0.00000334, -0.00000648, -0.00000295, -0.00000768, -0.00000505, -0.00001079, -0.00002926, -0.00002763, -0.00018686, 0.00020218, 0.00039193, -0.00010658, -0.00010865, -0.00000246, -0.00002664, 0.00000038, -0.00001454, 0.00000337, -0.00001125, 0.00000377, -0.00000028, -0.00000394, -0.00000974, 0.00000370, -0.00000201, 0.00000325, -0.00001136, 0.00000593, 0.00001722, -0.00001835, 0.00001699, -0.00001549, 0.00000082, -0.00000186, -0.00000897, 0.00000051, -0.00000703, 0.00000586, -0.00001592, 0.00001211, -0.00001737, 0.00001204, -0.00000707, -0.00000157, 0.00000504, -0.00000592, 0.00000533, -0.00000176, 0.00000597, -0.00000219, -0.00000336, 0.00000503, -0.00001324, 0.00000861, 0.00001071, -0.00000171, -0.00000055, 0.00000037, -0.00000083, -0.00000225, 0.00000175, 0.00000192, 0.00000318, -0.00000153, -0.00000431, 0.00000114, 0.00001001, -0.00000381, 0.00000598, -0.00000612, 0.00000556, -0.00000271, 0.00000268, 0.00000009, 0.00000592, -0.00000435, -0.00000670, 0.00000722, -0.00000901, 0.00000478, 0.00000672, -0.00001054, -0.00001355, 0.00000740, 0.00000293, 0.00000058, -0.00000292, -0.00000549, 0.00000523, -0.00000758, -0.00000631, 0.00000219, 0.00000756, -0.00000550, -0.00001335, 0.00000979, 0.00000620, -0.00000246, -0.00000220, 0.00000205, },
		{ 0.00000211, -0.00000905, 0.00000582, -0.00000513, 0.00000159, -0.00000275, -0.00000298, 0.00000262, -0.00002335, 0.00001540, -0.00000938, 0.00002431, 0.00000586, 0.00000629, 0.00003008, -0.00000383, 0.00000877, -0.00002302, -0.00000238, -0.00000350, -0.00000437, -0.00000358, -0.00000193, 0.00000544, -0.00000720, 0.00000035, -0.00000204, 0.00000236, -0.00000627, -0.00000279, -0.00000693, -0.00000191, -0.00001015, -0.00000175, -0.00001881, -0.00002276, -0.00005247, -0.00018429, 0.00032500, 0.00032125, -0.00018237, -0.00005631, -0.00001708, -0.00001455, -0.00000711, -0.00000783, -0.00000128, -0.00000804, 0.00000334, 0.00000007, -0.00000711, -0.00000632, 0.00000286, -0.00000120, 0.00000165, -0.00001233, 0.00001305, 0.00000974, -0.00001595, 0.00001873, -0.00002181, 0.00000845, -0.00000905, -0.00000267, -0.00000486, -0.00000058, -0.00000187, -0.00000908, 0.00000812, -0.00001671, 0.00001626, -0.00001401, 0.00000604, -0.00000104, -0.00000206, 0.00000333, -0.00000009, 0.00000577, -0.00000474, -0.00000042, 0.00000243, -0.00001348, 0.00001420, 0.00000644, -0.00000153, -0.00000093, 0.00000073, -0.00000172, -0.00000091, 0.00000103, 0.00000316, 0.00000176, -0.00000144, -0.00000530, 0.00000515, 0.00000665, -0.00000295, 0.00000505, -0.00000529, 0.00000614, -0.00000512, 0.00000561, -0.00000205, 0.00000829, -0.00000912, -0.00000160, 0.00000339, -0.00000797, 0.00000811, 0.00000358, -0.00001384, -0.00000942, 0.00000918, 0.00000058, 0.00000154, -0.00000537, -0.00000236, 0.00000385, -0.00000885, -0.00000455, 0.00000313, 0.00000752, -0.00000955, -0.00001013, 0.00001263, 0.00000267, -0.00000218, -0.00000194, },
		{ -0.00000096, 0.00000059, -0.00000863, 0.00000823, -0.00000917, 0.00000658, -0.00000926, 0.00000530, -0.00000824, -0.00001419, 0.00001409, -0.00000881, 0.00002820, -0.00000331, 0.00001590, 0.00002232, -0.00000336, 0.00000690, -0.00002476, 0.00000208, -0.00000619, -0.00000198, -0.00000586, 0.00000239, 0.00000172, -0.00000590, -0.00000011, -0.00000086, 0.00000090, -0.00000636, -0.00000314, -0.00000623, -0.00000243, -0.00001108, 0.00000012, -0.00002676, -0.00001487, -0.00008759, -0.00015374, 0.00044550, 0.00019790, -0.00020989, -0.00001579, -0.00003115, -0.00000269, -0.00001502, -0.00000032, -0.00000728, -0.00000310, 0.00000153, 0.00000057, -0.00001007, -0.00000220, 0.00000085, 0.00000056, -0.00000145, -0.00001071, 0.00001838, 0.00000116, -0.00001021, 0.00001637, -0.00002463, 0.00001405, -0.00001529, 0.00000331, -0.00001004, 0.00000577, -0.00001064, 0.00000100, -0.00000115, -0.00000938, 0.00001382, -0.00001584, 0.00001063, -0.00000607, 0.00000199, 0.00000071, 0.00000262, 0.00000388, -0.00000595, 0.00000238, -0.00000139, -0.00001135, 0.00001827, 0.00000156, -0.00000026, -0.00000192, 0.00000138, -0.00000254, 0.00000034, 0.00000031, 0.00000445, -0.00000001, -0.00000113, -0.00000587, 0.00000910, 0.00000198, -0.00000033, 0.00000257, -0.00000236, 0.00000376, -0.00000461, 0.00000648, -0.00000270, 0.00000919, -0.00001281, 0.00000346, -0.00000180, -0.00000425, 0.00000934, 0.00000012, -0.00001624, -0.00000399, 0.00000869, -0.00000043, 0.00000146, -0.00000684, 0.00000067, 0.00000157, -0.00000913, -0.00000308, 0.00000420, 0.00000676, -0.00001353, -0.00000498, 0.00001312, -0.00000003, -0.00000204, },
		{ -0.00000282, 0.00000121, -0.00000305, -0.00000481, 0.00000660, -0.00000935, 0.00000793, -0.00001240, 0.00001099, -0.00001870, -0.00000162, 0.00000753, -0.00000312, 0.00002642, -0.00000806, 0.00002390, 0.00001332, -0.00000050, 0.00000289, -0.00002420, 0.00000501, -0.00000791, -0.00000072, -0.00000669, 0.00000559, -0.00000201, -0.00000414, -0.00000063, 0.00000006, -0.00000049, -0.00000666, -0.00000280, -0.00000621, -0.00000263, -0.00001221, 0.00000204, -0.00003590, -0.00000468, -0.00013527, -0.00008171, 0.00053632, 0.00004128, -0.00018576, -0.00000039, -0.00003386, 0.00000099, -0.00001691, 0.00000249, -0.00000992, -0.00000026, 0.00000129, -0.00000125, -0.00001022, -0.00000008, 0.00000012, 0.00000105, -0.00000396, -0.00000796, 0.00002224, -0.00000820, -0.00000115, 0.00000903, -0.00002210, 0.00001526, -0.00001813, 0.00000652, -0.00001251, 0.00000899, -0.00001668, 0.00001013, -0.00001190, 0.00000203, 0.00000563, -0.00001193, 0.00001058, -0.00000813, 0.00000444, -0.00000100, 0.00000510, 0.00000107, -0.00000586, 0.00000455, -0.00000594, -0.00000671, 0.00001979, -0.00000271, 0.00000136, -0.00000316, 0.00000220, -0.00000321, 0.00000128, -0.00000001, 0.00000527, -0.00000161, -0.00000101, -0.00000545, 0.00001193, -0.00000275, 0.00000308, -0.00000053, 0.00000162, -0.00000089, -0.00000108, 0.00000458, -0.00000092, 0.00000750, -0.00001401, 0.00000675, -0.00000655, 0.00000101, 0.00000848, -0.00000333, -0.00001728, 0.00000185, 0.00000627, 0.00000010, -0.00000007, -0.00000661, 0.00000261, -0.00000069, -0.00000895, -0.00000163, 0.00000517, 0.00000527, -0.00001696, 0.00000173, 0.00001080, -0.00000095, },
		{ -0.00000005, -0.00000450, 0.00000418, -0.00000813, 0.00000193, 0.00000074, -0.00000473, 0.00000399, -0.00000964, 0.00001037, -0.00002380, 0.00000918, -0.00000025, 0.00000492, 0.00002054, -0.00000845, 0.00002929, 0.00000417, 0.00000433, -0.00000350, -0.00002070, 0.00000554, -0.00000786, -0.00000139, -0.00000499, 0.00000607, -0.00000397, -0.00000345, 0.00000010, -0.00000059, -0.00000049, -0.00000839, -0.00000046, -0.00000844, -0.00000086, -0.00001503, 0.00000501, -0.00004656, 0.00000661, -0.00019344, 0.00004076, 0.00056728, -0.00011568, -0.00012354, -0.00001099, -0.00002185, -0.00000648, -0.00000996, -0.00000210, -0.00000659, -0.00000177, 0.00000419, -0.00000652, -0.00000631, -0.00000147, 0.00000213, -0.00000116, -0.00000430, -0.00000533, 0.00002474, -0.00001741, 0.00001015, -0.00000278, -0.00001383, 0.00001095, -0.00001612, 0.00000546, -0.00001068, 0.00000706, -0.00001693, 0.00001414, -0.00001940, 0.00001307, -0.00000502, -0.00000403, 0.00000632, -0.00000672, 0.00000447, -0.00000083, 0.00000622, -0.00000169, -0.00000477, 0.00000573, -0.00001049, 0.00000006, 0.00001817, -0.00000523, 0.00000238, -0.00000405, 0.00000283, -0.00000352, 0.00000170, 0.00000037, 0.00000517, -0.00000252, -0.00000150, -0.00000352, 0.00001261, -0.00000616, 0.00000596, -0.00000291, 0.00000492, -0.00000602, 0.00000423, 0.00000047, 0.00000316, 0.00000301, -0.00001204, 0.00000701, -0.00000908, 0.00000615, 0.00000616, -0.00000673, -0.00001645, 0.00000686, 0.00000295, 0.00000161, -0.00000269, -0.00000469, 0.00000300, -0.00000234, -0.00000870, -0.00000003, 0.00000595, 0.00000285, -0.00001900, 0.00000889, 0.00000624, },
		{ 0.00000097, 0.00000170, -0.00000615, 0.00000660, -0.00001275, 0.00000936, -0.00000734, 0.00000317, -0.00000422, -0.00000128, 0.00000249, -0.00002077, 0.00001408, -0.00000521, 0.00001187, 0.00001316, -0.00000559, 0.00003172, -0.00000409, 0.00001022, -0.00001180, -0.00001432, 0.00000350, -0.00000595, -0.00000387, -0.00000083, 0.00000348, -0.00000346, -0.00000443, 0.00000241, -0.00000304, 0.00000105, -0.00001144, 0.00000365, -0.00001279, 0.00000273, -0.00001909, 0.00000782, -0.00005613, 0.00001326, -0.00025158, 0.00021061, 0.00051585, -0.00023567, -0.00004927, -0.00003630, -0.00000082, -0.00002122, 0.00000308, -0.00001215, 0.00000144, -0.00000672, 0.00000916, -0.00001384, 0.00000045, -0.00000556, 0.00000627, -0.00000563, -0.00000248, -0.00000290, 0.00002512, -0.00002455, 0.00002111, -0.00001658, -0.00000155, 0.00000206, -0.00000974, 0.00000046, -0.00000494, 0.00000014, -0.00001059, 0.00001080, -0.00001999, 0.00001926, -0.00001387, 0.00000456, -0.00000006, -0.00000290, 0.00000246, 0.00000126, 0.00000546, -0.00000366, -0.00000306, 0.00000561, -0.00001409, 0.00000799, 0.00001358, -0.00000538, 0.00000203, -0.00000394, 0.00000283, -0.00000328, 0.00000150, 0.00000155, 0.00000394, -0.00000245, -0.00000275, 0.00000001, 0.00001059, -0.00000717, 0.00000719, -0.00000337, 0.00000581, -0.00000939, 0.00000918, -0.00000417, 0.00000823, -0.00000330, -0.00000739, 0.00000406, -0.00000830, 0.00000963, 0.00000328, -0.00001012, -0.00001341, 0.00000987, 0.00000005, 0.00000303, -0.00000541, -0.00000186, 0.00000216, -0.00000335, -0.00000837, 0.00000157, 0.00000668, -0.00000082, -0.00001856, 0.00001473, },
		{ 0.00001741, -0.00000302, 0.00000270, -0.00000634, 0.00000669, -0.00001438, 0.00001424, -0.00001404, 0.00001068, -0.00001305, 0.00000925, -0.00001002, -0.00001048, 0.00001194, -0.00000523, 0.00001537, 0.00000651, -0.00000073, 0.00003099, -0.00001014, 0.00001552, -0.00002053, -0.00000621, -0.00000024, -0.00000315, -0.00000682, 0.00000430, -0.00000092, -0.00000138, -0.00000616, 0.00000512, -0.00000595, 0.00000260, -0.00001393, 0.00000725, -0.00001683, 0.00000557, -0.00002143, 0.00000648, -0.00005870, 0.00000456, -0.00028912, 0.00040728, 0.00037764, -0.00028952, 0.00000827, -0.00005851, 0.00001754, -0.00003428, 0.00001501, -0.00002182, 0.00000953, -0.00001138, 0.00001273, -0.00001964, 0.00000664, -0.00000942, 0.00000994, -0.00001004, -0.00000019, 0.00000046, 0.00002187, -0.00002696, 0.00002810, -0.00002840, 0.00001110, -0.00000846, -0.00000145, -0.00000619, 0.00000225, -0.00000915, 0.00000018, 0.00000098, -0.00001279, 0.00001805, -0.00001748, 0.00001041, -0.00000577, 0.00000136, -0.00000024, 0.00000437, 0.00000310, -0.00000453, -0.00000107, 0.00000397, -0.00001572, 0.00001559, 0.00000698, -0.00000331, 0.00000009, -0.00000251, 0.00000196, -0.00000245, 0.00000082, 0.00000329, 0.00000177, -0.00000152, -0.00000445, 0.00000464, 0.00000611, -0.00000551, 0.00000635, -0.00000149, 0.00000333, -0.00000919, 0.00001157, -0.00000721, 0.00001230, -0.00000961, -0.00000164, -0.00000106, -0.00000426, 0.00001054, 0.00000056, -0.00001336, -0.00000829, 0.00001017, -0.00000130, 0.00000323, -0.00000703, 0.00000070, 0.00000092, -0.00000419, -0.00000761, 0.00000281, 0.00000755, -0.00000594, -0.00001472, },
		{ -0.00000736, 0.00001580, -0.00000411, 0.00000161, -0.00000396, 0.00000318, -0.00001112, 0.00001370, -0.00001574, 0.00001365, -0.00001782, 0.00001670, -0.00002178, 0.00000280, 0.00000502, -0.00000104, 0.00001515, 0.00000155, 0.00000544, 0.00002673, -0.00001243, 0.00001803, -0.00002736, 0.00000151, -0.00000393, -0.00000118, -0.00000819, 0.00000789, -0.00000467, 0.00000012, -0.00000659, 0.00000600, -0.00000691, 0.00000163, -0.00001289, 0.00000685, -0.00001675, 0.00000366, -0.00001774, -0.00000421, -0.00004727, -0.00003201, -0.00027864, 0.00059284, 0.00017253, -0.00026710, 0.00002878, -0.00006191, 0.00002172, -0.00003630, 0.00001794, -0.00002434, 0.00001206, -0.00001120, 0.00001091, -0.00001989, 0.00000830, -0.00000948, 0.00000987, -0.00001135, 0.00000018, 0.00000627, 0.00001357, -0.00002239, 0.00002775, -0.00003402, 0.00001983, -0.00001672, 0.00000530, -0.00001123, 0.00000733, -0.00001661, 0.00001093, -0.00001145, -0.00000022, 0.00000993, -0.00001468, 0.00001147, -0.00000857, 0.00000411, -0.00000201, 0.00000714, -0.00000003, -0.00000443, 0.00000097, 0.00000081, -0.00001454, 0.00002121, -0.00000003, 0.00000005, -0.00000299, 0.00000005, 0.00000029, -0.00000120, 0.00000004, 0.00000504, -0.00000074, -0.00000028, -0.00000588, 0.00000924, 0.00000026, -0.00000184, 0.00000400, 0.00000201, -0.00000214, -0.00000492, 0.00001010, -0.00000697, 0.00001349, -0.00001386, 0.00000313, -0.00000638, 0.00000181, 0.00000895, -0.00000172, -0.00001599, -0.00000186, 0.00000788, -0.00000066, 0.00000163, -0.00000671, 0.00000195, 0.00000018, -0.00000542, -0.00000605, 0.00000343, 0.00000853, -0.00001221, },
		{ -0.00001851, 0.00000246, 0.00001015, -0.00000192, -0.00000185, 0.00000091, -0.00000367, -0.00000290, 0.00000685, -0.00001067, 0.00000953, -0.00001504, 0.00001646, -0.00002709, 0.00001330, -0.00000224, 0.00000443, 0.00001281, -0.00000205, 0.00001270, 0.00001867, -0.00000972, 0.00001574, -0.00002990, 0.00000657, -0.00000576, -0.00000162, -0.00000617, 0.00000771, -0.00000536, -0.00000110, -0.00000377, 0.00000306, -0.00000390, -0.00000387, -0.00000610, -0.00000023, -0.00000960, -0.00000600, -0.00000512, -0.00002726, -0.00001822, -0.00010439, -0.00019437, 0.00071952, -0.00005726, -0.00018224, 0.00000847, -0.00004064, 0.00000675, -0.00002284, 0.00000777, -0.00001592, 0.00000577, -0.00000369, 0.00000173, -0.00001256, 0.00000322, -0.00000367, 0.00000409, -0.00000761, -0.00000286, 0.00001506, 0.00000018, -0.00001036, 0.00001852, -0.00003081, 0.00002139, -0.00001953, 0.00000762, -0.00001194, 0.00000728, -0.00001833, 0.00001673, -0.00002113, 0.00001290, -0.00000182, -0.00000695, 0.00000781, -0.00000772, 0.00000434, -0.00000174, 0.00000832, -0.00000289, -0.00000372, 0.00000285, -0.00000359, -0.00001018, 0.00002346, -0.00000568, 0.00000328, -0.00000618, 0.00000304, -0.00000175, 0.00000010, -0.00000035, 0.00000610, -0.00000279, 0.00000054, -0.00000612, 0.00001240, -0.00000531, 0.00000248, 0.00000143, 0.00000545, -0.00000881, 0.00000232, 0.00000498, -0.00000303, 0.00001086, -0.00001465, 0.00000508, -0.00000972, 0.00000799, 0.00000571, -0.00000378, -0.00001723, 0.00000454, 0.00000399, 0.00000149, -0.00000152, -0.00000441, 0.00000151, 0.00000035, -0.00000725, -0.00000364, 0.00000351, 0.00000912, },
		{ 0.00000835, -0.00002301, 0.00001245, 0.00000224, 0.00000240, -0.00000656, 0.00000668, -0.00001175, 0.00000798, -0.00000438, -0.00000011, -0.00000103, -0.00000454, 0.00000702, -0.00002266, 0.00001647, -0.00000553, 0.00000781, 0.00001061, -0.00000505, 0.00002072, 0.00000733, -0.00000195, 0.00000785, -0.00002686, 0.00000760, -0.00000483, -0.00000489, -0.00000041, 0.00000306, -0.00000207, -0.00000574, 0.00000269, -0.00000395, 0.00000313, -0.00001363, 0.00000600, -0.00001364, 0.00000426, -0.00002267, 0.00001483, -0.00005985, 0.00002376, -0.00020820, -0.00002386, 0.00074357, -0.00025776, -0.00006879, -0.00003857, -0.00000180, -0.00002245, 0.00000257, -0.00001293, 0.00000163, -0.00000792, 0.00000956, -0.00001299, 0.00000076, -0.00000738, 0.00000698, -0.00000649, 0.00000060, -0.00000863, 0.00002529, -0.00001595, 0.00000685, 0.00000192, -0.00001908, 0.00001514, -0.00001591, 0.00000460, -0.00000765, 0.00000129, -0.00001257, 0.00001440, -0.00002349, 0.00002137, -0.00001248, 0.00000229, 0.00000143, -0.00000409, 0.00000231, 0.00000071, 0.00000732, -0.00000466, -0.00000280, 0.00000427, -0.00000853, -0.00000303, 0.00002167, -0.00000855, 0.00000496, -0.00000826, 0.00000552, -0.00000354, 0.00000103, 0.00000004, 0.00000592, -0.00000369, 0.00000031, -0.00000449, 0.00001293, -0.00000899, 0.00000593, 0.00000009, 0.00000680, -0.00001408, 0.00001010, -0.00000205, 0.00000353, 0.00000489, -0.00001186, 0.00000346, -0.00000952, 0.00001230, 0.00000210, -0.00000608, -0.00001628, 0.00000938, 0.00000002, 0.00000395, -0.00000512, -0.00000101, -0.00000013, 0.00000113, -0.00000931, -0.00000084, 0.00000357, },
		{ 0.00000439, 0.00000513, -0.00002370, 0.00001973, -0.00000510, 0.00000654, -0.00001031, 0.00001074, -0.00001769, 0.00001746, -0.00001569, 0.00001172, -0.00001399, 0.00000999, -0.00000886, -0.00000943, 0.00001127, -0.00000294, 0.00000721, 0.00000991, -0.00000761, 0.00002828, -0.00000536, 0.00000911, -0.00000441, -0.00001882, 0.00000493, -0.00000185, -0.00000961, 0.00000734, -0.00000443, 0.00000375, -0.00001217, 0.00001075, -0.00001266, 0.00001145, -0.00002425, 0.00001937, -0.00002897, 0.00001998, -0.00004060, 0.00003458, -0.00009161, 0.00006309, -0.00032167, 0.00022149, 0.00064121, -0.00037977, 0.00003109, -0.00008667, 0.00003747, -0.00005266, 0.00002923, -0.00003533, 0.00002106, -0.00002290, 0.00002293, -0.00002768, 0.00001483, -0.00001889, 0.00001822, -0.00001795, 0.00000988, -0.00001407, 0.00003337, -0.00003032, 0.00002428, -0.00001747, -0.00000238, 0.00000353, -0.00000768, -0.00000206, -0.00000029, -0.00000848, -0.00000095, 0.00000416, -0.00001685, 0.00002179, -0.00001795, 0.00000925, -0.00000476, 0.00000033, -0.00000067, 0.00000447, 0.00000435, -0.00000504, -0.00000189, 0.00000488, -0.00001295, 0.00000577, 0.00001618, -0.00000809, 0.00000418, -0.00000826, 0.00000661, -0.00000452, 0.00000134, 0.00000131, 0.00000433, -0.00000319, -0.00000114, -0.00000090, 0.00001034, -0.00000973, 0.00000747, 0.00000089, 0.00000460, -0.00001552, 0.00001554, -0.00000832, 0.00001044, -0.00000263, -0.00000674, -0.00000108, -0.00000544, 0.00001349, -0.00000079, -0.00000891, -0.00001274, 0.00001142, -0.00000248, 0.00000525, -0.00000768, 0.00000204, -0.00000182, 0.00000165, -0.00001077, 0.00000154, },
		{ 0.00000271, 0.00000647, -0.00000106, -0.00001922, 0.00002185, -0.00000908, 0.00000802, -0.00001082, 0.00001057, -0.00001822, 0.00002122, -0.00002204, 0.00001931, -0.00002337, 0.00002217, -0.00002485, 0.00000750, 0.00000075, 0.00000405, 0.00000333, 0.00001030, -0.00000867, 0.00003314, -0.00001618, 0.00001993, -0.00001784, -0.00000836, 0.00000061, 0.00000097, -0.00001294, 0.00001362, -0.00001116, 0.00000866, -0.00001680, 0.00001640, -0.00001867, 0.00001620, -0.00003006, 0.00002737, -0.00003879, 0.00002933, -0.00005031, 0.00004237, -0.00010676, 0.00007576, -0.00040651, 0.00050207, 0.00042071, -0.00039562, 0.00008236, -0.00010983, 0.00005792, -0.00006825, 0.00004396, -0.00004813, 0.00003274, -0.00003102, 0.00002908, -0.00003527, 0.00002291, -0.00002507, 0.00002424, -0.00002481, 0.00001541, -0.00001495, 0.00003485, -0.00003757, 0.00003584, -0.00003337, 0.00001362, -0.00000877, 0.00000129, -0.00000876, 0.00000625, -0.00001760, 0.00001190, -0.00001017, -0.00000333, 0.00001411, -0.00001648, 0.00001143, -0.00000828, 0.00000349, -0.00000291, 0.00000812, 0.00000033, -0.00000423, -0.00000102, 0.00000428, -0.00001567, 0.00001448, 0.00000832, -0.00000478, 0.00000093, -0.00000585, 0.00000585, -0.00000438, 0.00000103, 0.00000318, 0.00000169, -0.00000158, -0.00000334, 0.00000396, 0.00000511, -0.00000751, 0.00000700, 0.00000365, -0.00000130, -0.00001188, 0.00001645, -0.00001131, 0.00001517, -0.00000929, -0.00000148, -0.00000663, 0.00000143, 0.00001142, -0.00000249, -0.00001210, -0.00000693, 0.00001025, -0.00000258, 0.00000439, -0.00000798, 0.00000337, -0.00000233, 0.00000088, -0.00001071, },
		{ -0.00000869, 0.00000243, 0.00000958, -0.00000965, -0.00000971, 0.00001789, -0.00000812, 0.00000544, -0.00000701, 0.00000511, -0.00001181, 0.00001679, -0.00001992, 0.00001835, -0.00002395, 0.00002568, -0.00003368, 0.00002092, -0.00000934, 0.00001129, -0.00000096, 0.00000983, -0.00000639, 0.00003263, -0.00002147, 0.00002629, -0.00002824, 0.00000091, -0.00000249, 0.00000099, -0.00001183, 0.00001473, -0.00001315, 0.00000876, -0.00001572, 0.00001547, -0.00001755, 0.00001266, -0.00002569, 0.00002364, -0.00003584, 0.00002427, -0.00004276, 0.00002717, -0.00009047, 0.00003745, -0.00041693, 0.00075561, 0.00012507, -0.00030959, 0.00007024, -0.00009336, 0.00004753, -0.00005882, 0.00003754, -0.00004303, 0.00002941, -0.00002617, 0.00002271, -0.00003060, 0.00001989, -0.00002116, 0.00002053, -0.00002269, 0.00001328, -0.00000811, 0.00002663, -0.00003385, 0.00003664, -0.00004015, 0.00002332, -0.00001686, 0.00000679, -0.00001173, 0.00000799, -0.00002118, 0.00002020, -0.00002257, 0.00001194, 0.00000158, -0.00000926, 0.00000858, -0.00000812, 0.00000417, -0.00000312, 0.00001023, -0.00000356, -0.00000277, -0.00000012, 0.00000223, -0.00001565, 0.00002120, 0.00000005, -0.00000001, -0.00000389, -0.00000153, 0.00000339, -0.00000323, 0.00000043, 0.00000503, -0.00000121, 0.00000037, -0.00000534, 0.00000872, -0.00000135, -0.00000330, 0.00000540, 0.00000698, -0.00000941, -0.00000378, 0.00001229, -0.00000967, 0.00001590, -0.00001302, 0.00000168, -0.00001076, 0.00000898, 0.00000703, -0.00000325, -0.00001490, 0.00000005, 0.00000650, -0.00000039, 0.00000126, -0.00000567, 0.00000236, -0.00000099, -0.00000172, },
		{ -0.00000579, -0.00000510, 0.00000128, 0.00001259, -0.00001875, 0.00000286, 0.00000900, -0.00000270, -0.00000073, 0.00000026, -0.00000434, 0.00000023, 0.00000490, -0.00000929, 0.00000800, -0.00001396, 0.00001737, -0.00003046, 0.00002482, -0.00001331, 0.00001419, -0.00000225, 0.00000639, 0.00000068, 0.00002506, -0.00001871, 0.00002497, -0.00003213, 0.00000595, -0.00000226, -0.00000327, -0.00000480, 0.00000875, -0.00000814, 0.00000193, -0.00000699, 0.00000604, -0.00000739, -0.00000095, -0.00000921, 0.00000569, -0.00001716, 0.00000160, -0.00001466, -0.00001481, -0.00003731, -0.00006469, -0.00031505, 0.00091277, -0.00017733, -0.00015764, 0.00000563, -0.00004069, 0.00000789, -0.00002488, 0.00000980, -0.00001934, 0.00001026, -0.00000787, 0.00000369, -0.00001346, 0.00000530, -0.00000660, 0.00000644, -0.00001074, 0.00000279, 0.00000654, 0.00000894, -0.00001880, 0.00002524, -0.00003525, 0.00002357, -0.00001775, 0.00000638, -0.00000899, 0.00000292, -0.00001631, 0.00001957, -0.00002735, 0.00002289, -0.00001063, 0.00000020, 0.00000258, -0.00000502, 0.00000248, -0.00000099, 0.00001000, -0.00000628, -0.00000126, 0.00000086, -0.00000115, -0.00001241, 0.00002438, -0.00000655, 0.00000433, -0.00000872, 0.00000352, -0.00000005, -0.00000150, -0.00000000, 0.00000612, -0.00000344, 0.00000169, -0.00000606, 0.00001181, -0.00000719, 0.00000137, 0.00000409, 0.00000881, -0.00001703, 0.00000644, 0.00000439, -0.00000382, 0.00001228, -0.00001294, 0.00000139, -0.00001141, 0.00001485, 0.00000194, -0.00000384, -0.00001627, 0.00000653, 0.00000165, 0.00000300, -0.00000316, -0.00000153, -0.00000053, 0.00000188, },
		{ 0.00000502, -0.00001015, -0.00000108, 0.00000046, 0.00001368, -0.00002557, 0.00001512, -0.00000177, 0.00000476, -0.00000818, 0.00000824, -0.00001432, 0.00001382, -0.00001033, 0.00000602, -0.00000821, 0.00000362, -0.00000088, -0.00001521, 0.00001726, -0.00000843, 0.00001020, 0.00000132, -0.00000070, 0.00001218, 0.00001101, -0.00000798, 0.00001535, -0.00002837, 0.00000575, 0.00000140, -0.00001092, 0.00000680, -0.00000308, 0.00000274, -0.00001048, 0.00000762, -0.00000981, 0.00000928, -0.00002143, 0.00001562, -0.00002245, 0.00001297, -0.00003364, 0.00002751, -0.00007517, 0.00004182, -0.00022101, -0.00008812, 0.00091876, -0.00041215, 0.00000454, -0.00007967, 0.00002763, -0.00004554, 0.00002128, -0.00002926, 0.00001474, -0.00001786, 0.00001759, -0.00002186, 0.00001048, -0.00001580, 0.00001396, -0.00001373, 0.00000725, -0.00001256, 0.00002509, -0.00001392, 0.00000356, 0.00000470, -0.00002038, 0.00001504, -0.00001178, 0.00000063, -0.00000159, -0.00000762, -0.00000374, 0.00000923, -0.00002176, 0.00002514, -0.00001772, 0.00000777, -0.00000356, -0.00000089, -0.00000035, 0.00000277, 0.00000741, -0.00000725, -0.00000019, 0.00000185, -0.00000538, -0.00000620, 0.00002329, -0.00000993, 0.00000654, -0.00001188, 0.00000782, -0.00000345, 0.00000021, 0.00000016, 0.00000591, -0.00000429, 0.00000171, -0.00000476, 0.00001206, -0.00001080, 0.00000503, 0.00000428, 0.00000727, -0.00002117, 0.00001545, -0.00000445, 0.00000423, 0.00000562, -0.00000971, -0.00000222, -0.00000777, 0.00001726, -0.00000228, -0.00000507, -0.00001523, 0.00001082, -0.00000243, 0.00000589, -0.00000719, 0.00000279, -0.00000390, },
		{ -0.00000593, 0.00000676, -0.00001317, 0.00000193, 0.00000130, 0.00001117, -0.00002736, 0.00002329, -0.00001054, 0.00001083, -0.00001363, 0.00001334, -0.00002044, 0.00002358, -0.00002280, 0.00001958, -0.00002355, 0.00002180, -0.00002232, 0.00000668, 0.00000154, 0.00000356, 0.00000040, 0.00000871, -0.00000969, 0.00002529, -0.00000612, 0.00000747, 0.00000002, -0.00001882, 0.00000198, 0.00000614, -0.00001856, 0.00001868, -0.00001615, 0.00001489, -0.00002353, 0.00002259, -0.00002597, 0.00002563, -0.00004082, 0.00003953, -0.00005045, 0.00004306, -0.00006821, 0.00006756, -0.00013305, 0.00011841, -0.00039464, 0.00023934, 0.00075378, -0.00052257, 0.00012087, -0.00014609, 0.00008271, -0.00008955, 0.00006029, -0.00006312, 0.00004511, -0.00004280, 0.00003930, -0.00004362, 0.00003147, -0.00003441, 0.00003209, -0.00003201, 0.00002403, -0.00002607, 0.00004071, -0.00003464, 0.00002586, -0.00001820, -0.00000111, 0.00000197, -0.00000247, -0.00000704, 0.00000659, -0.00001914, 0.00001195, -0.00000729, -0.00000737, 0.00001794, -0.00001729, 0.00001055, -0.00000723, 0.00000217, -0.00000269, 0.00000683, 0.00000319, -0.00000648, 0.00000025, 0.00000263, -0.00000961, 0.00000199, 0.00001823, -0.00000954, 0.00000560, -0.00001210, 0.00001008, -0.00000584, 0.00000139, 0.00000109, 0.00000431, -0.00000356, 0.00000031, -0.00000144, 0.00000913, -0.00001131, 0.00000694, 0.00000631, 0.00000161, -0.00001979, 0.00002021, -0.00001107, 0.00001162, -0.00000174, -0.00000527, -0.00000756, -0.00000057, 0.00001564, -0.00000465, -0.00000732, -0.00001143, 0.00001183, -0.00000424, 0.00000675, -0.00000910, 0.00000544, },
		{ 0.00000516, -0.00000529, 0.00000576, -0.00001355, 0.00000285, 0.00000449, 0.00000434, -0.00002257, 0.00002468, -0.00001409, 0.00001251, -0.00001435, 0.00001278, -0.00001931, 0.00002502, -0.00002702, 0.00002507, -0.00003081, 0.00003237, -0.00003794, 0.00002641, -0.00001506, 0.00001702, -0.00001077, 0.00001623, -0.00001671, 0.00003533, -0.00002093, 0.00002213, -0.00001594, -0.00000768, -0.00000183, 0.00000823, -0.00002152, 0.00002517, -0.00002427, 0.00002191, -0.00003049, 0.00003062, -0.00003451, 0.00003299, -0.00004921, 0.00005097, -0.00006517, 0.00005834, -0.00008507, 0.00008462, -0.00016127, 0.00015319, -0.00052691, 0.00060460, 0.00044436, -0.00048835, 0.00015487, -0.00016267, 0.00009987, -0.00010336, 0.00007414, -0.00007588, 0.00005781, -0.00005246, 0.00004647, -0.00005140, 0.00003979, -0.00004141, 0.00003912, -0.00004009, 0.00003184, -0.00003063, 0.00004631, -0.00004560, 0.00004006, -0.00003550, 0.00001540, -0.00000966, 0.00000517, -0.00001203, 0.00001063, -0.00002575, 0.00002407, -0.00002329, 0.00001041, 0.00000446, -0.00001028, 0.00000805, -0.00000729, 0.00000296, -0.00000328, 0.00000982, -0.00000155, -0.00000438, 0.00000010, 0.00000290, -0.00001283, 0.00001058, 0.00001050, -0.00000594, 0.00000158, -0.00000905, 0.00000961, -0.00000659, 0.00000181, 0.00000257, 0.00000171, -0.00000166, -0.00000195, 0.00000307, 0.00000372, -0.00000896, 0.00000732, 0.00000929, -0.00000724, -0.00001264, 0.00001914, -0.00001317, 0.00001585, -0.00000736, -0.00000196, -0.00001216, 0.00000817, 0.00001074, -0.00000514, -0.00001027, -0.00000542, 0.00000951, -0.00000321, 0.00000484, -0.00000793, },
		{ -0.00000394, 0.00000191, -0.00000174, 0.00000168, -0.00001096, 0.00000161, 0.00000954, -0.00000591, -0.00001172, 0.00001878, -0.00001122, 0.00000861, -0.00000953, 0.00000603, -0.00001029, 0.00001672, -0.00002058, 0.00001921, -0.00002569, 0.00002958, -0.00004033, 0.00003552, -0.00002448, 0.00002506, -0.00001759, 0.00001922, -0.00001753, 0.00003763, -0.00002803, 0.00003025, -0.00002701, 0.00000040, -0.00000212, 0.00000447, -0.00001621, 0.00002185, -0.00002238, 0.00001861, -0.00002602, 0.00002600, -0.00002936, 0.00002491, -0.00003931, 0.00004130, -0.00005650, 0.00004728, -0.00007089, 0.00006225, -0.00013752, 0.00011034, -0.00055357, 0.00091971, 0.00006008, -0.00033320, 0.00010234, -0.00011990, 0.00006988, -0.00007822, 0.00005456, -0.00005968, 0.00004559, -0.00004054, 0.00003362, -0.00004002, 0.00003031, -0.00003184, 0.00003021, -0.00003308, 0.00002602, -0.00002208, 0.00003778, -0.00004207, 0.00004067, -0.00004119, 0.00002334, -0.00001493, 0.00000713, -0.00001092, 0.00000706, -0.00002301, 0.00002666, -0.00003173, 0.00002447, -0.00000955, -0.00000043, 0.00000221, -0.00000450, 0.00000159, -0.00000177, 0.00001085, -0.00000569, -0.00000166, -0.00000042, 0.00000239, -0.00001412, 0.00001780, 0.00000206, -0.00000068, -0.00000437, -0.00000337, 0.00000655, -0.00000562, 0.00000158, 0.00000410, -0.00000109, 0.00000055, -0.00000408, 0.00000736, -0.00000262, -0.00000482, 0.00000717, 0.00001140, -0.00001681, -0.00000158, 0.00001272, -0.00001017, 0.00001569, -0.00000973, -0.00000155, -0.00001369, 0.00001585, 0.00000425, -0.00000449, -0.00001298, 0.00000146, 0.00000491, 0.00000007, 0.00000062, },
		{ -0.00000441, 0.00000140, -0.00000303, 0.00000361, -0.00000451, -0.00000623, -0.00000077, 0.00001472, -0.00001702, 0.00000246, 0.00000763, -0.00000326, 0.00000041, -0.00000080, -0.00000462, 0.00000392, 0.00000127, -0.00000558, 0.00000358, -0.00000904, 0.00001303, -0.00002713, 0.00002974, -0.00002163, 0.00002282, -0.00001591, 0.00001458, -0.00000978, 0.00002977, -0.00002443, 0.00002832, -0.00002964, 0.00000262, 0.00000263, -0.00000565, -0.00000233, 0.00000806, -0.00000945, 0.00000394, -0.00000924, 0.00000793, -0.00000988, 0.00000103, -0.00001080, 0.00000970, -0.00002306, 0.00000816, -0.00002378, -0.00000222, -0.00005666, -0.00002541, -0.00042745, 0.00109602, -0.00030487, -0.00011707, -0.00000813, -0.00003364, 0.00000358, -0.00002201, 0.00000745, -0.00001889, 0.00001179, -0.00001004, 0.00000380, -0.00001226, 0.00000538, -0.00000774, 0.00000711, -0.00001226, 0.00000755, -0.00000158, 0.00001626, -0.00002456, 0.00002721, -0.00003378, 0.00002073, -0.00001215, 0.00000245, -0.00000346, -0.00000408, -0.00001022, 0.00001739, -0.00002828, 0.00002912, -0.00001833, 0.00000762, -0.00000372, -0.00000089, -0.00000072, 0.00000124, 0.00000969, -0.00000841, 0.00000094, -0.00000099, 0.00000097, -0.00001289, 0.00002213, -0.00000504, 0.00000421, -0.00001038, 0.00000341, 0.00000181, -0.00000338, 0.00000104, 0.00000510, -0.00000323, 0.00000214, -0.00000503, 0.00000996, -0.00000815, -0.00000032, 0.00000767, 0.00001062, -0.00002390, 0.00001011, 0.00000333, -0.00000338, 0.00001166, -0.00000885, -0.00000440, -0.00001083, 0.00002021, -0.00000184, -0.00000384, -0.00001424, 0.00000742, -0.00000020, 0.00000406, },
		{ 0.00000692, -0.00000825, 0.00000595, -0.00000753, 0.00000870, -0.00001085, -0.00000112, -0.00000250, 0.00001762, -0.00002555, 0.00001582, -0.00000478, 0.00000635, -0.00000885, 0.00000825, -0.00001484, 0.00001801, -0.00001549, 0.00001197, -0.00001569, 0.00001295, -0.00001158, -0.00000257, 0.00001108, -0.00000695, 0.00001005, -0.00000547, 0.00000273, 0.00000532, 0.00001316, -0.00001117, 0.00001681, -0.00002388, -0.00000057, 0.00001068, -0.00001902, 0.00001610, -0.00001197, 0.00001031, -0.00001757, 0.00001481, -0.00001803, 0.00001767, -0.00003129, 0.00002784, -0.00003457, 0.00002515, -0.00004774, 0.00004284, -0.00009247, 0.00006148, -0.00023570, -0.00014004, 0.00107279, -0.00055986, 0.00008436, -0.00012714, 0.00006158, -0.00007214, 0.00004334, -0.00004880, 0.00003099, -0.00003037, 0.00002724, -0.00003196, 0.00002156, -0.00002552, 0.00002207, -0.00002197, 0.00001494, -0.00001679, 0.00002427, -0.00001172, 0.00000098, 0.00000461, -0.00001691, 0.00001020, -0.00000368, -0.00000619, 0.00000693, -0.00001869, 0.00000858, -0.00000102, -0.00001348, 0.00002253, -0.00001862, 0.00001042, -0.00000687, 0.00000141, -0.00000239, 0.00000461, 0.00000674, -0.00000927, 0.00000275, -0.00000130, -0.00000131, -0.00000905, 0.00002271, -0.00000918, 0.00000688, -0.00001448, 0.00000939, -0.00000322, -0.00000064, 0.00000064, 0.00000511, -0.00000410, 0.00000250, -0.00000417, 0.00000987, -0.00001144, 0.00000340, 0.00000946, 0.00000573, -0.00002579, 0.00001881, -0.00000569, 0.00000457, 0.00000564, -0.00000616, -0.00000936, -0.00000385, 0.00002014, -0.00000603, -0.00000413, -0.00001314, 0.00001089, -0.00000394, },
		{ -0.00000504, 0.00000724, -0.00000927, 0.00000774, -0.00000944, 0.00001129, -0.00001515, 0.00000245, -0.00000187, 0.00001613, -0.00002835, 0.00002419, -0.00001404, 0.00001352, -0.00001532, 0.00001367, -0.00001997, 0.00002620, -0.00002676, 0.00002454, -0.00003026, 0.00003108, -0.00003443, 0.00002397, -0.00001265, 0.00001355, -0.00000845, 0.00000963, -0.00001213, 0.00002293, -0.00000698, 0.00000676, 0.00000009, -0.00001344, -0.00000558, 0.00001751, -0.00002965, 0.00003181, -0.00003025, 0.00002858, -0.00003710, 0.00003652, -0.00004141, 0.00004177, -0.00005894, 0.00006150, -0.00007448, 0.00006912, -0.00009863, 0.00010283, -0.00017580, 0.00017375, -0.00046610, 0.00026702, 0.00083991, -0.00064646, 0.00020590, -0.00020430, 0.00012700, -0.00012548, 0.00009083, -0.00009078, 0.00006954, -0.00006312, 0.00005545, -0.00005891, 0.00004754, -0.00004940, 0.00004533, -0.00004537, 0.00003763, -0.00003694, 0.00004573, -0.00003651, 0.00002526, -0.00001864, 0.00000219, -0.00000236, 0.00000532, -0.00001369, 0.00001460, -0.00003007, 0.00002598, -0.00002149, 0.00000724, 0.00000761, -0.00001105, 0.00000729, -0.00000612, 0.00000130, -0.00000236, 0.00000726, 0.00000270, -0.00000819, 0.00000331, -0.00000107, -0.00000420, -0.00000306, 0.00001949, -0.00000964, 0.00000618, -0.00001515, 0.00001289, -0.00000712, 0.00000175, 0.00000068, 0.00000401, -0.00000355, 0.00000154, -0.00000161, 0.00000699, -0.00001193, 0.00000596, 0.00001214, -0.00000297, -0.00002133, 0.00002200, -0.00001137, 0.00001090, 0.00000003, -0.00000385, -0.00001415, 0.00000550, 0.00001591, -0.00000769, -0.00000567, -0.00000947, 0.00001105, },
		{ 0.00000819, -0.00000331, 0.00000469, -0.00000690, 0.00000587, -0.00000755, 0.00000993, -0.00001589, 0.00000325, 0.00000190, 0.00000951, -0.00002393, 0.00002497, -0.00001695, 0.00001518, -0.00001624, 0.00001289, -0.00001718, 0.00002473, -0.00002771, 0.00002633, -0.00003324, 0.00003702, -0.00004563, 0.00004167, -0.00003112, 0.00003075, -0.00002482, 0.00002260, -0.00002354, 0.00003627, -0.00002349, 0.00002226, -0.00001527, -0.00000392, -0.00000748, 0.00001789, -0.00003125, 0.00003718, -0.00003819, 0.00003625, -0.00004501, 0.00004550, -0.00005101, 0.00005024, -0.00006814, 0.00007419, -0.00009166, 0.00008790, -0.00012024, 0.00012664, -0.00021401, 0.00022611, -0.00063769, 0.00070657, 0.00044586, -0.00055723, 0.00021489, -0.00020801, 0.00013590, -0.00013324, 0.00009994, -0.00009997, 0.00008014, -0.00007180, 0.00006157, -0.00006493, 0.00005415, -0.00005530, 0.00005154, -0.00005285, 0.00004604, -0.00004375, 0.00005393, -0.00004915, 0.00003936, -0.00003398, 0.00001589, -0.00001057, 0.00000949, -0.00001515, 0.00001435, -0.00003188, 0.00003421, -0.00003541, 0.00002565, -0.00000920, 0.00000030, 0.00000044, -0.00000249, -0.00000080, -0.00000057, 0.00000858, -0.00000170, -0.00000545, 0.00000242, -0.00000020, -0.00000728, 0.00000413, 0.00001334, -0.00000679, 0.00000210, -0.00001190, 0.00001290, -0.00000887, 0.00000318, 0.00000124, 0.00000207, -0.00000193, -0.00000024, 0.00000186, 0.00000211, -0.00000994, 0.00000776, 0.00001431, -0.00001351, -0.00001150, 0.00001909, -0.00001217, 0.00001375, -0.00000337, -0.00000378, -0.00001632, 0.00001463, 0.00000901, -0.00000721, -0.00000801, -0.00000394, },
		{ 0.00000208, 0.00000353, 0.00000034, 0.00000011, -0.00000193, 0.00000089, -0.00000211, 0.00000460, -0.00001287, 0.00000133, 0.00000814, -0.00000114, -0.00001312, 0.00001814, -0.00001280, 0.00001056, -0.00001120, 0.00000603, -0.00000688, 0.00001368, -0.00001769, 0.00001600, -0.00002240, 0.00002712, -0.00003951, 0.00004291, -0.00003593, 0.00003652, -0.00003168, 0.00002703, -0.00002577, 0.00003954, -0.00003024, 0.00002913, -0.00002349, -0.00000006, -0.00000275, 0.00000879, -0.00002039, 0.00002785, -0.00003062, 0.00002774, -0.00003543, 0.00003546, -0.00004014, 0.00003597, -0.00005092, 0.00005634, -0.00007464, 0.00006810, -0.00009670, 0.00009429, -0.00017959, 0.00017404, -0.00067093, 0.00106746, -0.00001348, -0.00033754, 0.00012117, -0.00013734, 0.00008480, -0.00009090, 0.00006564, -0.00007103, 0.00005741, -0.00005152, 0.00004137, -0.00004610, 0.00003746, -0.00003933, 0.00003674, -0.00004023, 0.00003582, -0.00003307, 0.00004472, -0.00004513, 0.00003838, -0.00003633, 0.00001945, -0.00001064, 0.00000604, -0.00000852, 0.00000426, -0.00002128, 0.00002862, -0.00003631, 0.00003423, -0.00002080, 0.00000984, -0.00000626, 0.00000149, -0.00000327, 0.00000207, 0.00000857, -0.00000580, -0.00000161, 0.00000030, 0.00000115, -0.00000995, 0.00001124, 0.00000576, -0.00000195, -0.00000427, -0.00000540, 0.00000948, -0.00000811, 0.00000339, 0.00000217, -0.00000017, 0.00000009, -0.00000202, 0.00000504, -0.00000341, -0.00000643, 0.00000960, 0.00001411, -0.00002287, 0.00000086, 0.00001162, -0.00000847, 0.00001285, -0.00000407, -0.00000666, -0.00001426, 0.00002103, 0.00000151, -0.00000564, -0.00001017, },
		{ -0.00001108, 0.00000699, -0.00000117, 0.00000426, -0.00000474, 0.00000374, -0.00000534, 0.00000511, -0.00000322, -0.00000730, -0.00000197, 0.00001481, -0.00001294, 0.00000094, 0.00000634, -0.00000362, 0.00000151, -0.00000235, -0.00000402, 0.00000719, -0.00000288, -0.00000075, -0.00000248, -0.00000152, 0.00000442, -0.00001741, 0.00002674, -0.00002438, 0.00002741, -0.00002553, 0.00002007, -0.00001661, 0.00003065, -0.00002488, 0.00002481, -0.00002220, -0.00000326, 0.00000848, -0.00000845, 0.00000113, 0.00000542, -0.00000872, 0.00000420, -0.00000966, 0.00000785, -0.00001052, 0.00000124, -0.00000990, 0.00001035, -0.00002538, 0.00001139, -0.00002958, 0.00000663, -0.00007060, 0.00000371, -0.00051457, 0.00124746, -0.00042247, -0.00006952, -0.00002890, -0.00002138, -0.00000515, -0.00001482, 0.00000103, -0.00001448, 0.00000992, -0.00000972, 0.00000183, -0.00000890, 0.00000332, -0.00000677, 0.00000567, -0.00001149, 0.00001029, -0.00000785, 0.00002071, -0.00002637, 0.00002352, -0.00002614, 0.00001303, -0.00000314, -0.00000366, 0.00000397, -0.00001280, -0.00000078, 0.00001021, -0.00002284, 0.00002942, -0.00002251, 0.00001314, -0.00000925, 0.00000324, -0.00000426, 0.00000434, 0.00000756, -0.00000900, 0.00000246, -0.00000243, 0.00000255, -0.00001153, 0.00001692, -0.00000142, 0.00000299, -0.00001099, 0.00000265, 0.00000364, -0.00000525, 0.00000256, 0.00000312, -0.00000210, 0.00000179, -0.00000303, 0.00000680, -0.00000813, -0.00000241, 0.00001205, 0.00001013, -0.00002798, 0.00001212, 0.00000255, -0.00000226, 0.00000947, -0.00000306, -0.00001166, -0.00000783, 0.00002310, -0.00000471, -0.00000416, },
		{ -0.00000358, -0.00001004, 0.00000956, -0.00000431, 0.00000676, -0.00000793, 0.00000785, -0.00001028, 0.00001152, -0.00001104, -0.00000134, -0.00000456, 0.00001923, -0.00002228, 0.00001392, -0.00000612, 0.00000672, -0.00000832, 0.00000651, -0.00001278, 0.00001942, -0.00001837, 0.00001595, -0.00002131, 0.00002093, -0.00002229, 0.00001249, -0.00000048, -0.00000087, 0.00000643, -0.00000860, 0.00000395, 0.00000124, 0.00001246, -0.00001006, 0.00001161, -0.00001358, -0.00001071, 0.00002189, -0.00002785, 0.00002612, -0.00002237, 0.00001958, -0.00002607, 0.00002336, -0.00002765, 0.00002720, -0.00004185, 0.00004109, -0.00004850, 0.00003949, -0.00006364, 0.00005949, -0.00011069, 0.00008236, -0.00025488, -0.00016917, 0.00118660, -0.00068108, 0.00015647, -0.00017161, 0.00009331, -0.00009703, 0.00006448, -0.00006790, 0.00004736, -0.00004316, 0.00003664, -0.00004142, 0.00003205, -0.00003472, 0.00002971, -0.00002979, 0.00002247, -0.00002071, 0.00002279, -0.00000960, -0.00000056, 0.00000145, -0.00000903, 0.00000141, 0.00000733, -0.00001463, 0.00001647, -0.00002994, 0.00002222, -0.00001439, 0.00000028, 0.00001326, -0.00001415, 0.00000892, -0.00000710, 0.00000153, -0.00000272, 0.00000536, 0.00000583, -0.00001069, 0.00000569, -0.00000481, 0.00000332, -0.00001133, 0.00002000, -0.00000655, 0.00000610, -0.00001589, 0.00001006, -0.00000290, -0.00000130, 0.00000117, 0.00000371, -0.00000322, 0.00000268, -0.00000284, 0.00000649, -0.00001103, 0.00000145, 0.00001498, 0.00000217, -0.00002699, 0.00001914, -0.00000497, 0.00000389, 0.00000564, -0.00000221, -0.00001673, 0.00000156, 0.00002062, -0.00000850, },
		{ -0.00000966, -0.00000412, -0.00000704, 0.00000929, -0.00000495, 0.00000675, -0.00000814, 0.00000865, -0.00001178, 0.00001463, -0.00001636, 0.00000284, -0.00000452, 0.00001918, -0.00002605, 0.00002180, -0.00001488, 0.00001405, -0.00001508, 0.00001167, -0.00001614, 0.00002469, -0.00002653, 0.00002517, -0.00003216, 0.00003525, -0.00004187, 0.00003837, -0.00002756, 0.00002484, -0.00001815, 0.00001203, -0.00001472, 0.00002104, -0.00000817, 0.00000763, -0.00000429, -0.00000330, -0.00001665, 0.00003059, -0.00004087, 0.00004446, -0.00004431, 0.00004242, -0.00005051, 0.00005011, -0.00005646, 0.00005712, -0.00007542, 0.00008157, -0.00009700, 0.00009384, -0.00012706, 0.00013496, -0.00021364, 0.00022199, -0.00052961, 0.00030446, 0.00088930, -0.00073585, 0.00027236, -0.00025073, 0.00016198, -0.00015356, 0.00011496, -0.00011301, 0.00008966, -0.00007998, 0.00006816, -0.00007053, 0.00006002, -0.00006091, 0.00005520, -0.00005538, 0.00004819, -0.00004473, 0.00004720, -0.00003523, 0.00002215, -0.00001858, 0.00000692, -0.00000847, 0.00001447, -0.00002059, 0.00002212, -0.00003903, 0.00003829, -0.00003527, 0.00002405, -0.00000740, -0.00000002, -0.00000037, -0.00000114, -0.00000281, 0.00000093, 0.00000509, 0.00000341, -0.00001031, 0.00000701, -0.00000577, 0.00000272, -0.00000886, 0.00001982, -0.00000855, 0.00000606, -0.00001722, 0.00001474, -0.00000824, 0.00000249, -0.00000013, 0.00000367, -0.00000333, 0.00000263, -0.00000158, 0.00000415, -0.00001176, 0.00000507, 0.00001734, -0.00000838, -0.00002000, 0.00002041, -0.00000879, 0.00000791, 0.00000310, -0.00000332, -0.00001945, 0.00001150, 0.00001464, },
		{ 0.00000700, -0.00000880, -0.00000534, -0.00000273, 0.00000662, -0.00000316, 0.00000421, -0.00000519, 0.00000561, -0.00000888, 0.00001304, -0.00001754, 0.00000387, -0.00000099, 0.00001386, -0.00002291, 0.00002236, -0.00001714, 0.00001561, -0.00001632, 0.00001114, -0.00001228, 0.00002062, -0.00002409, 0.00002279, -0.00002989, 0.00003476, -0.00004561, 0.00004962, -0.00004315, 0.00004181, -0.00003629, 0.00002749, -0.00002785, 0.00003484, -0.00002329, 0.00002045, -0.00001573, 0.00000229, -0.00001528, 0.00002841, -0.00004019, 0.00004731, -0.00005032, 0.00004891, -0.00005753, 0.00005811, -0.00006515, 0.00006470, -0.00008316, 0.00009262, -0.00011321, 0.00011242, -0.00014940, 0.00016083, -0.00025578, 0.00028279, -0.00072482, 0.00079699, 0.00042653, -0.00059554, 0.00025391, -0.00023834, 0.00015961, -0.00015236, 0.00011644, -0.00011579, 0.00009544, -0.00008534, 0.00007134, -0.00007311, 0.00006324, -0.00006402, 0.00005886, -0.00006047, 0.00005530, -0.00005193, 0.00005603, -0.00004721, 0.00003332, -0.00002879, 0.00001486, -0.00001096, 0.00001353, -0.00001728, 0.00001646, -0.00003446, 0.00004002, -0.00004342, 0.00003882, -0.00002385, 0.00001307, -0.00000989, 0.00000521, -0.00000734, 0.00000503, 0.00000426, 0.00000017, -0.00000756, 0.00000573, -0.00000456, 0.00000030, -0.00000409, 0.00001636, -0.00000726, 0.00000252, -0.00001423, 0.00001536, -0.00001091, 0.00000497, -0.00000081, 0.00000290, -0.00000248, 0.00000186, 0.00000021, 0.00000044, -0.00001057, 0.00000873, 0.00001758, -0.00001885, -0.00000907, 0.00001651, -0.00000847, 0.00000906, 0.00000250, -0.00000717, -0.00001808, 0.00001943, },
		{ 0.00002352, -0.00000038, -0.00000692, -0.00000652, 0.00000176, 0.00000267, 0.00000011, 0.00000012, -0.00000016, -0.00000028, -0.00000226, 0.00000698, -0.00001445, 0.00000181, 0.00000539, 0.00000437, -0.00001381, 0.00001594, -0.00001268, 0.00001106, -0.00001200, 0.00000554, -0.00000253, 0.00000872, -0.00001230, 0.00000975, -0.00001498, 0.00001892, -0.00003110, 0.00004139, -0.00004057, 0.00004258, -0.00004058, 0.00003109, -0.00002951, 0.00003699, -0.00002733, 0.00002309, -0.00001793, -0.00000054, -0.00000407, 0.00001328, -0.00002339, 0.00003139, -0.00003617, 0.00003427, -0.00004192, 0.00004174, -0.00004769, 0.00004358, -0.00005796, 0.00006556, -0.00008640, 0.00008225, -0.00011479, 0.00011662, -0.00020808, 0.00021645, -0.00075192, 0.00118050, -0.00008386, -0.00032592, 0.00012588, -0.00014374, 0.00009023, -0.00009489, 0.00006888, -0.00007505, 0.00006272, -0.00005715, 0.00004443, -0.00004757, 0.00004002, -0.00004223, 0.00003866, -0.00004255, 0.00004083, -0.00003919, 0.00004610, -0.00004227, 0.00002976, -0.00002618, 0.00001228, -0.00000443, 0.00000384, -0.00000481, -0.00000000, -0.00001606, 0.00002547, -0.00003466, 0.00003848, -0.00002924, 0.00001898, -0.00001472, 0.00000825, -0.00000926, 0.00000759, 0.00000382, -0.00000387, -0.00000272, 0.00000194, -0.00000113, -0.00000381, 0.00000244, 0.00001044, -0.00000351, -0.00000366, -0.00000746, 0.00001179, -0.00001027, 0.00000549, -0.00000056, 0.00000149, -0.00000097, 0.00000077, 0.00000181, -0.00000368, -0.00000799, 0.00001270, 0.00001434, -0.00002626, 0.00000256, 0.00000966, -0.00000523, 0.00000810, 0.00000309, -0.00001308, -0.00001222, },
		{ -0.00000302, 0.00002317, -0.00000609, -0.00000494, -0.00000693, 0.00000530, -0.00000115, 0.00000343, -0.00000389, 0.00000501, -0.00000688, 0.00000598, -0.00000178, -0.00000848, -0.00000196, 0.00001264, -0.00000663, -0.00000173, 0.00000528, -0.00000379, 0.00000245, -0.00000432, -0.00000222, 0.00000910, -0.00000621, 0.00000362, -0.00000828, 0.00000648, -0.00000614, -0.00000350, 0.00001678, -0.00002071, 0.00002644, -0.00002950, 0.00002135, -0.00001869, 0.00002675, -0.00001963, 0.00001506, -0.00001084, -0.00001075, 0.00001443, -0.00001097, 0.00000506, 0.00000116, -0.00000599, 0.00000254, -0.00000795, 0.00000548, -0.00000900, -0.00000054, -0.00000631, 0.00000711, -0.00002313, 0.00000991, -0.00003009, 0.00000887, -0.00007477, 0.00001493, -0.00056214, 0.00134747, -0.00051271, -0.00002569, -0.00005078, -0.00000756, -0.00001574, -0.00000529, -0.00000784, -0.00000729, 0.00000524, -0.00000727, -0.00000170, -0.00000404, -0.00000034, -0.00000406, 0.00000237, -0.00000846, 0.00001054, -0.00001143, 0.00002164, -0.00002402, 0.00001464, -0.00001361, 0.00000199, 0.00000784, -0.00001055, 0.00001186, -0.00002144, 0.00001019, -0.00000045, -0.00001171, 0.00002308, -0.00002166, 0.00001502, -0.00001232, 0.00000591, -0.00000691, 0.00000731, 0.00000426, -0.00000819, 0.00000315, -0.00000333, 0.00000363, -0.00000872, 0.00000954, 0.00000350, 0.00000109, -0.00001065, 0.00000137, 0.00000513, -0.00000671, 0.00000408, 0.00000056, -0.00000027, 0.00000083, -0.00000029, 0.00000260, -0.00000732, -0.00000447, 0.00001671, 0.00000721, -0.00002839, 0.00001163, 0.00000271, -0.00000117, 0.00000664, 0.00000318, -0.00001910, },
		{ -0.00002280, 0.00000725, 0.00001902, -0.00000948, -0.00000347, -0.00000621, 0.00000709, -0.00000365, 0.00000550, -0.00000625, 0.00000826, -0.00001168, 0.00001304, -0.00001052, -0.00000190, -0.00000532, 0.00001823, -0.00001586, 0.00000944, -0.00000563, 0.00000581, -0.00000675, 0.00000333, -0.00000830, 0.00001770, -0.00001809, 0.00001668, -0.00002331, 0.00002527, -0.00003004, 0.00002638, -0.00001431, 0.00000831, -0.00000025, -0.00000799, 0.00000262, 0.00000018, 0.00000865, -0.00000471, 0.00000090, 0.00000078, -0.00002278, 0.00003305, -0.00003547, 0.00003489, -0.00003232, 0.00002874, -0.00003429, 0.00003166, -0.00003703, 0.00003612, -0.00005109, 0.00005300, -0.00006179, 0.00005354, -0.00007885, 0.00007504, -0.00012749, 0.00010249, -0.00027865, -0.00016908, 0.00124574, -0.00075907, 0.00020768, -0.00020379, 0.00011596, -0.00011478, 0.00008009, -0.00008247, 0.00006036, -0.00005346, 0.00004359, -0.00004816, 0.00003987, -0.00004143, 0.00003512, -0.00003555, 0.00002843, -0.00002351, 0.00002051, -0.00000762, -0.00000110, -0.00000418, 0.00000179, -0.00000943, 0.00001929, -0.00002271, 0.00002496, -0.00003883, 0.00003423, -0.00002771, 0.00001666, -0.00000102, -0.00000483, 0.00000317, -0.00000376, -0.00000111, -0.00000073, 0.00000439, 0.00000519, -0.00001162, 0.00000829, -0.00000823, 0.00000807, -0.00001290, 0.00001561, -0.00000268, 0.00000463, -0.00001620, 0.00000981, -0.00000259, -0.00000152, 0.00000140, 0.00000213, -0.00000195, 0.00000251, -0.00000109, 0.00000229, -0.00000987, -0.00000021, 0.00001986, -0.00000280, -0.00002468, 0.00001613, -0.00000203, 0.00000182, 0.00000622, 0.00000103, },
		{ -0.00000416, -0.00002229, 0.00001610, 0.00001255, -0.00001063, -0.00000264, -0.00000441, 0.00000690, -0.00000425, 0.00000558, -0.00000600, 0.00000821, -0.00001272, 0.00001644, -0.00001660, 0.00000297, -0.00000631, 0.00002000, -0.00002059, 0.00001631, -0.00001305, 0.00001246, -0.00001319, 0.00000798, -0.00000976, 0.00001959, -0.00002224, 0.00002125, -0.00002861, 0.00003306, -0.00004248, 0.00004647, -0.00003930, 0.00003457, -0.00002685, 0.00001474, -0.00001686, 0.00001917, -0.00000951, 0.00000984, -0.00001195, 0.00000966, -0.00002901, 0.00004297, -0.00004972, 0.00005395, -0.00005551, 0.00005358, -0.00006095, 0.00006057, -0.00006824, 0.00006865, -0.00008688, 0.00009576, -0.00011358, 0.00011241, -0.00014819, 0.00015806, -0.00023995, 0.00025545, -0.00057733, 0.00034809, 0.00089693, -0.00078011, 0.00030972, -0.00027690, 0.00018093, -0.00016823, 0.00012779, -0.00012535, 0.00010141, -0.00008996, 0.00007466, -0.00007601, 0.00006648, -0.00006660, 0.00005952, -0.00005997, 0.00005374, -0.00004801, 0.00004450, -0.00003071, 0.00001670, -0.00001792, 0.00001215, -0.00001493, 0.00002335, -0.00002608, 0.00002744, -0.00004383, 0.00004607, -0.00004509, 0.00003912, -0.00002382, 0.00001347, -0.00001090, 0.00000675, -0.00000956, 0.00000681, 0.00000042, 0.00000543, -0.00001271, 0.00001081, -0.00001069, 0.00001023, -0.00001468, 0.00001905, -0.00000653, 0.00000546, -0.00001828, 0.00001545, -0.00000904, 0.00000354, -0.00000145, 0.00000354, -0.00000310, 0.00000376, -0.00000161, 0.00000094, -0.00001107, 0.00000479, 0.00002079, -0.00001348, -0.00001642, 0.00001583, -0.00000370, 0.00000295, 0.00000739, },
		{ 0.00000936, -0.00001172, -0.00001702, 0.00002160, 0.00000544, -0.00001006, -0.00000231, -0.00000195, 0.00000509, -0.00000304, 0.00000374, -0.00000323, 0.00000470, -0.00000933, 0.00001490, -0.00001836, 0.00000466, -0.00000390, 0.00001708, -0.00001964, 0.00001716, -0.00001481, 0.00001395, -0.00001493, 0.00000831, -0.00000586, 0.00001401, -0.00001737, 0.00001555, -0.00002170, 0.00002616, -0.00003770, 0.00004856, -0.00004812, 0.00004737, -0.00004299, 0.00002934, -0.00002874, 0.00003047, -0.00002025, 0.00001692, -0.00001682, 0.00000966, -0.00002361, 0.00003787, -0.00004629, 0.00005333, -0.00005815, 0.00005727, -0.00006533, 0.00006572, -0.00007410, 0.00007345, -0.00009105, 0.00010242, -0.00012536, 0.00012680, -0.00016664, 0.00018036, -0.00027842, 0.00031286, -0.00077535, 0.00086393, 0.00039144, -0.00060135, 0.00026743, -0.00024919, 0.00016717, -0.00015746, 0.00012070, -0.00012049, 0.00010089, -0.00009060, 0.00007382, -0.00007425, 0.00006541, -0.00006593, 0.00005955, -0.00006137, 0.00005799, -0.00005388, 0.00005210, -0.00004001, 0.00002276, -0.00002085, 0.00001259, -0.00001057, 0.00001677, -0.00001786, 0.00001652, -0.00003271, 0.00004013, -0.00004498, 0.00004651, -0.00003594, 0.00002499, -0.00002048, 0.00001391, -0.00001518, 0.00001237, -0.00000236, 0.00000372, -0.00001039, 0.00000938, -0.00000924, 0.00000867, -0.00001282, 0.00001885, -0.00000717, 0.00000284, -0.00001586, 0.00001664, -0.00001233, 0.00000681, -0.00000337, 0.00000415, -0.00000331, 0.00000431, -0.00000191, -0.00000115, -0.00001085, 0.00001039, 0.00001826, -0.00002205, -0.00000629, 0.00001222, -0.00000297, 0.00000276, },
		{ 0.00000268, 0.00001035, -0.00001959, -0.00000804, 0.00002301, -0.00000095, -0.00000838, -0.00000223, 0.00000060, 0.00000239, -0.00000072, 0.00000082, 0.00000100, -0.00000116, -0.00000238, 0.00000874, -0.00001563, 0.00000302, 0.00000157, 0.00001016, -0.00001374, 0.00001244, -0.00001102, 0.00001020, -0.00001208, 0.00000512, 0.00000158, 0.00000341, -0.00000616, 0.00000243, -0.00000557, 0.00000722, -0.00001711, 0.00003170, -0.00003741, 0.00004169, -0.00004291, 0.00003031, -0.00002825, 0.00002986, -0.00001982, 0.00001331, -0.00001109, -0.00000097, -0.00000576, 0.00001731, -0.00002450, 0.00003163, -0.00003790, 0.00003685, -0.00004406, 0.00004335, -0.00005050, 0.00004622, -0.00005885, 0.00006697, -0.00008918, 0.00008654, -0.00012126, 0.00012440, -0.00021684, 0.00022875, -0.00078340, 0.00124316, -0.00013896, -0.00030412, 0.00011840, -0.00013964, 0.00008631, -0.00009025, 0.00006428, -0.00007150, 0.00006093, -0.00005677, 0.00004242, -0.00004424, 0.00003778, -0.00004025, 0.00003569, -0.00003963, 0.00004032, -0.00003978, 0.00004181, -0.00003419, 0.00001631, -0.00001277, 0.00000347, 0.00000253, 0.00000145, -0.00000068, -0.00000461, -0.00000843, 0.00001801, -0.00002675, 0.00003575, -0.00003262, 0.00002460, -0.00002095, 0.00001385, -0.00001473, 0.00001333, -0.00000230, -0.00000049, -0.00000465, 0.00000400, -0.00000370, 0.00000312, -0.00000714, 0.00001499, -0.00000483, -0.00000276, -0.00000926, 0.00001309, -0.00001160, 0.00000732, -0.00000358, 0.00000356, -0.00000237, 0.00000397, -0.00000209, -0.00000360, -0.00000916, 0.00001608, 0.00001188, -0.00002624, 0.00000272, 0.00000764, -0.00000145, },
		{ -0.00000073, 0.00000415, 0.00000847, -0.00002516, 0.00000245, 0.00002076, -0.00000589, -0.00000611, -0.00000218, 0.00000269, -0.00000033, 0.00000177, -0.00000198, 0.00000505, -0.00000734, 0.00000598, -0.00000019, -0.00000969, -0.00000078, 0.00000858, 0.00000124, -0.00000518, 0.00000445, -0.00000373, 0.00000307, -0.00000643, 0.00000072, 0.00000912, -0.00000769, 0.00000609, -0.00001195, 0.00001258, -0.00001570, 0.00001120, 0.00000270, -0.00001184, 0.00002012, -0.00002772, 0.00001826, -0.00001623, 0.00001857, -0.00000987, 0.00000121, 0.00000234, -0.00001829, 0.00001924, -0.00001228, 0.00000853, -0.00000386, -0.00000192, -0.00000056, -0.00000451, 0.00000118, -0.00000573, -0.00000376, -0.00000076, 0.00000066, -0.00001672, 0.00000382, -0.00002499, 0.00000364, -0.00006760, 0.00000462, -0.00056173, 0.00138244, -0.00056173, 0.00000462, -0.00006760, 0.00000364, -0.00002499, 0.00000382, -0.00001672, 0.00000066, -0.00000076, -0.00000376, -0.00000573, 0.00000118, -0.00000451, -0.00000056, -0.00000192, -0.00000386, 0.00000853, -0.00001228, 0.00001924, -0.00001829, 0.00000234, 0.00000121, -0.00000987, 0.00001857, -0.00001623, 0.00001826, -0.00002772, 0.00002012, -0.00001184, 0.00000270, 0.00001120, -0.00001570, 0.00001258, -0.00001195, 0.00000609, -0.00000769, 0.00000912, 0.00000072, -0.00000643, 0.00000307, -0.00000373, 0.00000445, -0.00000518, 0.00000124, 0.00000858, -0.00000078, -0.00000969, -0.00000019, 0.00000598, -0.00000734, 0.00000505, -0.00000198, 0.00000177, -0.00000033, 0.00000269, -0.00000218, -0.00000611, -0.00000589, 0.00002076, 0.00000245, -0.00002516, 0.00000847, 0.00000415, },
		{ 0.00000268, -0.00000145, 0.00000764, 0.00000272, -0.00002624, 0.00001188, 0.00001608, -0.00000916, -0.00000360, -0.00000209, 0.00000397, -0.00000237, 0.00000356, -0.00000358, 0.00000732, -0.00001160, 0.00001309, -0.00000926, -0.00000276, -0.00000483, 0.00001499, -0.00000714, 0.00000312, -0.00000370, 0.00000400, -0.00000465, -0.00000049, -0.00000230, 0.00001333, -0.00001473, 0.00001385, -0.00002095, 0.00002460, -0.00003262, 0.00003575, -0.00002675, 0.00001801, -0.00000843, -0.00000461, -0.00000068, 0.00000145, 0.00000253, 0.00000347, -0.00001277, 0.00001631, -0.00003419, 0.00004181, -0.00003978, 0.00004032, -0.00003963, 0.00003569, -0.00004025, 0.00003778, -0.00004424, 0.00004242, -0.00005677, 0.00006093, -0.00007150, 0.00006428, -0.00009025, 0.00008631, -0.00013964, 0.00011840, -0.00030412, -0.00013896, 0.00124316, -0.00078340, 0.00022875, -0.00021684, 0.00012440, -0.00012126, 0.00008654, -0.00008918, 0.00006697, -0.00005885, 0.00004622, -0.00005050, 0.00004335, -0.00004406, 0.00003685, -0.00003790, 0.00003163, -0.00002450, 0.00001731, -0.00000576, -0.00000097, -0.00001109, 0.00001331, -0.00001982, 0.00002986, -0.00002825, 0.00003031, -0.00004291, 0.00004169, -0.00003741, 0.00003170, -0.00001711, 0.00000722, -0.00000557, 0.00000243, -0.00000616, 0.00000341, 0.00000158, 0.00000512, -0.00001208, 0.00001020, -0.00001102, 0.00001244, -0.00001374, 0.00001016, 0.00000157, 0.00000302, -0.00001563, 0.00000874, -0.00000238, -0.00000116, 0.00000100, 0.00000082, -0.00000072, 0.00000239, 0.00000060, -0.00000223, -0.00000838, -0.00000095, 0.00002301, -0.00000804, -0.00001959, 0.00001035, },
		{ 0.00000936, 0.00000276, -0.00000297, 0.00001222, -0.00000629, -0.00002205, 0.00001826, 0.00001039, -0.00001085, -0.00000115, -0.00000191, 0.00000431, -0.00000331, 0.00000415, -0.00000337, 0.00000681, -0.00001233, 0.00001664, -0.00001586, 0.00000284, -0.00000717, 0.00001885, -0.00001282, 0.00000867, -0.00000924, 0.00000938, -0.00001039, 0.00000372, -0.00000236, 0.00001237, -0.00001518, 0.00001391, -0.00002048, 0.00002499, -0.00003594, 0.00004651, -0.00004498, 0.00004013, -0.00003271, 0.00001652, -0.00001786, 0.00001677, -0.00001057, 0.00001259, -0.00002085, 0.00002276, -0.00004001, 0.00005210, -0.00005388, 0.00005799, -0.00006137, 0.00005955, -0.00006593, 0.00006541, -0.00007425, 0.00007382, -0.00009060, 0.00010089, -0.00012049, 0.00012070, -0.00015746, 0.00016717, -0.00024919, 0.00026743, -0.00060135, 0.00039144, 0.00086393, -0.00077535, 0.00031286, -0.00027842, 0.00018036, -0.00016664, 0.00012680, -0.00012536, 0.00010242, -0.00009105, 0.00007345, -0.00007410, 0.00006572, -0.00006533, 0.00005727, -0.00005815, 0.00005333, -0.00004629, 0.00003787, -0.00002361, 0.00000966, -0.00001682, 0.00001692, -0.00002025, 0.00003047, -0.00002874, 0.00002934, -0.00004299, 0.00004737, -0.00004812, 0.00004856, -0.00003770, 0.00002616, -0.00002170, 0.00001555, -0.00001737, 0.00001401, -0.00000586, 0.00000831, -0.00001493, 0.00001395, -0.00001481, 0.00001716, -0.00001964, 0.00001708, -0.00000390, 0.00000466, -0.00001836, 0.00001490, -0.00000933, 0.00000470, -0.00000323, 0.00000374, -0.00000304, 0.00000509, -0.00000195, -0.00000231, -0.00001006, 0.00000544, 0.00002160, -0.00001702, -0.00001172, },
		{ -0.00000416, 0.00000739, 0.00000295, -0.00000370, 0.00001583, -0.00001642, -0.00001348, 0.00002079, 0.00000479, -0.00001107, 0.00000094, -0.00000161, 0.00000376, -0.00000310, 0.00000354, -0.00000145, 0.00000354, -0.00000904, 0.00001545, -0.00001828, 0.00000546, -0.00000653, 0.00001905, -0.00001468, 0.00001023, -0.00001069, 0.00001081, -0.00001271, 0.00000543, 0.00000042, 0.00000681, -0.00000956, 0.00000675, -0.00001090, 0.00001347, -0.00002382, 0.00003912, -0.00004509, 0.00004607, -0.00004383, 0.00002744, -0.00002608, 0.00002335, -0.00001493, 0.00001215, -0.00001792, 0.00001670, -0.00003071, 0.00004450, -0.00004801, 0.00005374, -0.00005997, 0.00005952, -0.00006660, 0.00006648, -0.00007601, 0.00007466, -0.00008996, 0.00010141, -0.00012535, 0.00012779, -0.00016823, 0.00018093, -0.00027690, 0.00030972, -0.00078011, 0.00089693, 0.00034809, -0.00057733, 0.00025545, -0.00023995, 0.00015806, -0.00014819, 0.00011241, -0.00011358, 0.00009576, -0.00008688, 0.00006865, -0.00006824, 0.00006057, -0.00006095, 0.00005358, -0.00005551, 0.00005395, -0.00004972, 0.00004297, -0.00002901, 0.00000966, -0.00001195, 0.00000984, -0.00000951, 0.00001917, -0.00001686, 0.00001474, -0.00002685, 0.00003457, -0.00003930, 0.00004647, -0.00004248, 0.00003306, -0.00002861, 0.00002125, -0.00002224, 0.00001959, -0.00000976, 0.00000798, -0.00001319, 0.00001246, -0.00001305, 0.00001631, -0.00002059, 0.00002000, -0.00000631, 0.00000297, -0.00001660, 0.00001644, -0.00001272, 0.00000821, -0.00000600, 0.00000558, -0.00000425, 0.00000690, -0.00000441, -0.00000264, -0.00001063, 0.00001255, 0.00001610, -0.00002229, },
		{ -0.00002280, 0.00000103, 0.00000622, 0.00000182, -0.00000203, 0.00001613, -0.00002468, -0.00000280, 0.00001986, -0.00000021, -0.00000987, 0.00000229, -0.00000109, 0.00000251, -0.00000195, 0.00000213, 0.00000140, -0.00000152, -0.00000259, 0.00000981, -0.00001620, 0.00000463, -0.00000268, 0.00001561, -0.00001290, 0.00000807, -0.00000823, 0.00000829, -0.00001162, 0.00000519, 0.00000439, -0.00000073, -0.00000111, -0.00000376, 0.00000317, -0.00000483, -0.00000102, 0.00001666, -0.00002771, 0.00003423, -0.00003883, 0.00002496, -0.00002271, 0.00001929, -0.00000943, 0.00000179, -0.00000418, -0.00000110, -0.00000762, 0.00002051, -0.00002351, 0.00002843, -0.00003555, 0.00003512, -0.00004143, 0.00003987, -0.00004816, 0.00004359, -0.00005346, 0.00006036, -0.00008247, 0.00008009, -0.00011478, 0.00011596, -0.00020379, 0.00020768, -0.00075907, 0.00124574, -0.00016908, -0.00027865, 0.00010249, -0.00012749, 0.00007504, -0.00007885, 0.00005354, -0.00006179, 0.00005300, -0.00005109, 0.00003612, -0.00003703, 0.00003166, -0.00003429, 0.00002874, -0.00003232, 0.00003489, -0.00003547, 0.00003305, -0.00002278, 0.00000078, 0.00000090, -0.00000471, 0.00000865, 0.00000018, 0.00000262, -0.00000799, -0.00000025, 0.00000831, -0.00001431, 0.00002638, -0.00003004, 0.00002527, -0.00002331, 0.00001668, -0.00001809, 0.00001770, -0.00000830, 0.00000333, -0.00000675, 0.00000581, -0.00000563, 0.00000944, -0.00001586, 0.00001823, -0.00000532, -0.00000190, -0.00001052, 0.00001304, -0.00001168, 0.00000826, -0.00000625, 0.00000550, -0.00000365, 0.00000709, -0.00000621, -0.00000347, -0.00000948, 0.00001902, 0.00000725, },
		{ -0.00000302, -0.00001910, 0.00000318, 0.00000664, -0.00000117, 0.00000271, 0.00001163, -0.00002839, 0.00000721, 0.00001671, -0.00000447, -0.00000732, 0.00000260, -0.00000029, 0.00000083, -0.00000027, 0.00000056, 0.00000408, -0.00000671, 0.00000513, 0.00000137, -0.00001065, 0.00000109, 0.00000350, 0.00000954, -0.00000872, 0.00000363, -0.00000333, 0.00000315, -0.00000819, 0.00000426, 0.00000731, -0.00000691, 0.00000591, -0.00001232, 0.00001502, -0.00002166, 0.00002308, -0.00001171, -0.00000045, 0.00001019, -0.00002144, 0.00001186, -0.00001055, 0.00000784, 0.00000199, -0.00001361, 0.00001464, -0.00002402, 0.00002164, -0.00001143, 0.00001054, -0.00000846, 0.00000237, -0.00000406, -0.00000034, -0.00000404, -0.00000170, -0.00000727, 0.00000524, -0.00000729, -0.00000784, -0.00000529, -0.00001574, -0.00000756, -0.00005078, -0.00002569, -0.00051271, 0.00134747, -0.00056214, 0.00001493, -0.00007477, 0.00000887, -0.00003009, 0.00000991, -0.00002313, 0.00000711, -0.00000631, -0.00000054, -0.00000900, 0.00000548, -0.00000795, 0.00000254, -0.00000599, 0.00000116, 0.00000506, -0.00001097, 0.00001443, -0.00001075, -0.00001084, 0.00001506, -0.00001963, 0.00002675, -0.00001869, 0.00002135, -0.00002950, 0.00002644, -0.00002071, 0.00001678, -0.00000350, -0.00000614, 0.00000648, -0.00000828, 0.00000362, -0.00000621, 0.00000910, -0.00000222, -0.00000432, 0.00000245, -0.00000379, 0.00000528, -0.00000173, -0.00000663, 0.00001264, -0.00000196, -0.00000848, -0.00000178, 0.00000598, -0.00000688, 0.00000501, -0.00000389, 0.00000343, -0.00000115, 0.00000530, -0.00000693, -0.00000494, -0.00000609, 0.00002317, },
		{ 0.00002352, -0.00001222, -0.00001308, 0.00000309, 0.00000810, -0.00000523, 0.00000966, 0.00000256, -0.00002626, 0.00001434, 0.00001270, -0.00000799, -0.00000368, 0.00000181, 0.00000077, -0.00000097, 0.00000149, -0.00000056, 0.00000549, -0.00001027, 0.00001179, -0.00000746, -0.00000366, -0.00000351, 0.00001044, 0.00000244, -0.00000381, -0.00000113, 0.00000194, -0.00000272, -0.00000387, 0.00000382, 0.00000759, -0.00000926, 0.00000825, -0.00001472, 0.00001898, -0.00002924, 0.00003848, -0.00003466, 0.00002547, -0.00001606, -0.00000000, -0.00000481, 0.00000384, -0.00000443, 0.00001228, -0.00002618, 0.00002976, -0.00004227, 0.00004610, -0.00003919, 0.00004083, -0.00004255, 0.00003866, -0.00004223, 0.00004002, -0.00004757, 0.00004443, -0.00005715, 0.00006272, -0.00007505, 0.00006888, -0.00009489, 0.00009023, -0.00014374, 0.00012588, -0.00032592, -0.00008386, 0.00118050, -0.00075192, 0.00021645, -0.00020808, 0.00011662, -0.00011479, 0.00008225, -0.00008640, 0.00006556, -0.00005796, 0.00004358, -0.00004769, 0.00004174, -0.00004192, 0.00003427, -0.00003617, 0.00003139, -0.00002339, 0.00001328, -0.00000407, -0.00000054, -0.00001793, 0.00002309, -0.00002733, 0.00003699, -0.00002951, 0.00003109, -0.00004058, 0.00004258, -0.00004057, 0.00004139, -0.00003110, 0.00001892, -0.00001498, 0.00000975, -0.00001230, 0.00000872, -0.00000253, 0.00000554, -0.00001200, 0.00001106, -0.00001268, 0.00001594, -0.00001381, 0.00000437, 0.00000539, 0.00000181, -0.00001445, 0.00000698, -0.00000226, -0.00000028, -0.00000016, 0.00000012, 0.00000011, 0.00000267, 0.00000176, -0.00000652, -0.00000692, -0.00000038, },
		{ 0.00000700, 0.00001943, -0.00001808, -0.00000717, 0.00000250, 0.00000906, -0.00000847, 0.00001651, -0.00000907, -0.00001885, 0.00001758, 0.00000873, -0.00001057, 0.00000044, 0.00000021, 0.00000186, -0.00000248, 0.00000290, -0.00000081, 0.00000497, -0.00001091, 0.00001536, -0.00001423, 0.00000252, -0.00000726, 0.00001636, -0.00000409, 0.00000030, -0.00000456, 0.00000573, -0.00000756, 0.00000017, 0.00000426, 0.00000503, -0.00000734, 0.00000521, -0.00000989, 0.00001307, -0.00002385, 0.00003882, -0.00004342, 0.00004002, -0.00003446, 0.00001646, -0.00001728, 0.00001353, -0.00001096, 0.00001486, -0.00002879, 0.00003332, -0.00004721, 0.00005603, -0.00005193, 0.00005530, -0.00006047, 0.00005886, -0.00006402, 0.00006324, -0.00007311, 0.00007134, -0.00008534, 0.00009544, -0.00011579, 0.00011644, -0.00015236, 0.00015961, -0.00023834, 0.00025391, -0.00059554, 0.00042653, 0.00079699, -0.00072482, 0.00028279, -0.00025578, 0.00016083, -0.00014940, 0.00011242, -0.00011321, 0.00009262, -0.00008316, 0.00006470, -0.00006515, 0.00005811, -0.00005753, 0.00004891, -0.00005032, 0.00004731, -0.00004019, 0.00002841, -0.00001528, 0.00000229, -0.00001573, 0.00002045, -0.00002329, 0.00003484, -0.00002785, 0.00002749, -0.00003629, 0.00004181, -0.00004315, 0.00004962, -0.00004561, 0.00003476, -0.00002989, 0.00002279, -0.00002409, 0.00002062, -0.00001228, 0.00001114, -0.00001632, 0.00001561, -0.00001714, 0.00002236, -0.00002291, 0.00001386, -0.00000099, 0.00000387, -0.00001754, 0.00001304, -0.00000888, 0.00000561, -0.00000519, 0.00000421, -0.00000316, 0.00000662, -0.00000273, -0.00000534, -0.00000880, },
		{ -0.00000966, 0.00001464, 0.00001150, -0.00001945, -0.00000332, 0.00000310, 0.00000791, -0.00000879, 0.00002041, -0.00002000, -0.00000838, 0.00001734, 0.00000507, -0.00001176, 0.00000415, -0.00000158, 0.00000263, -0.00000333, 0.00000367, -0.00000013, 0.00000249, -0.00000824, 0.00001474, -0.00001722, 0.00000606, -0.00000855, 0.00001982, -0.00000886, 0.00000272, -0.00000577, 0.00000701, -0.00001031, 0.00000341, 0.00000509, 0.00000093, -0.00000281, -0.00000114, -0.00000037, -0.00000002, -0.00000740, 0.00002405, -0.00003527, 0.00003829, -0.00003903, 0.00002212, -0.00002059, 0.00001447, -0.00000847, 0.00000692, -0.00001858, 0.00002215, -0.00003523, 0.00004720, -0.00004473, 0.00004819, -0.00005538, 0.00005520, -0.00006091, 0.00006002, -0.00007053, 0.00006816, -0.00007998, 0.00008966, -0.00011301, 0.00011496, -0.00015356, 0.00016198, -0.00025073, 0.00027236, -0.00073585, 0.00088930, 0.00030446, -0.00052961, 0.00022199, -0.00021364, 0.00013496, -0.00012706, 0.00009384, -0.00009700, 0.00008157, -0.00007542, 0.00005712, -0.00005646, 0.00005011, -0.00005051, 0.00004242, -0.00004431, 0.00004446, -0.00004087, 0.00003059, -0.00001665, -0.00000330, -0.00000429, 0.00000763, -0.00000817, 0.00002104, -0.00001472, 0.00001203, -0.00001815, 0.00002484, -0.00002756, 0.00003837, -0.00004187, 0.00003525, -0.00003216, 0.00002517, -0.00002653, 0.00002469, -0.00001614, 0.00001167, -0.00001508, 0.00001405, -0.00001488, 0.00002180, -0.00002605, 0.00001918, -0.00000452, 0.00000284, -0.00001636, 0.00001463, -0.00001178, 0.00000865, -0.00000814, 0.00000675, -0.00000495, 0.00000929, -0.00000704, -0.00000412, },
		{ -0.00000358, -0.00000850, 0.00002062, 0.00000156, -0.00001673, -0.00000221, 0.00000564, 0.00000389, -0.00000497, 0.00001914, -0.00002699, 0.00000217, 0.00001498, 0.00000145, -0.00001103, 0.00000649, -0.00000284, 0.00000268, -0.00000322, 0.00000371, 0.00000117, -0.00000130, -0.00000290, 0.00001006, -0.00001589, 0.00000610, -0.00000655, 0.00002000, -0.00001133, 0.00000332, -0.00000481, 0.00000569, -0.00001069, 0.00000583, 0.00000536, -0.00000272, 0.00000153, -0.00000710, 0.00000892, -0.00001415, 0.00001326, 0.00000028, -0.00001439, 0.00002222, -0.00002994, 0.00001647, -0.00001463, 0.00000733, 0.00000141, -0.00000903, 0.00000145, -0.00000056, -0.00000960, 0.00002279, -0.00002071, 0.00002247, -0.00002979, 0.00002971, -0.00003472, 0.00003205, -0.00004142, 0.00003664, -0.00004316, 0.00004736, -0.00006790, 0.00006448, -0.00009703, 0.00009331, -0.00017161, 0.00015647, -0.00068108, 0.00118660, -0.00016917, -0.00025488, 0.00008236, -0.00011069, 0.00005949, -0.00006364, 0.00003949, -0.00004850, 0.00004109, -0.00004185, 0.00002720, -0.00002765, 0.00002336, -0.00002607, 0.00001958, -0.00002237, 0.00002612, -0.00002785, 0.00002189, -0.00001071, -0.00001358, 0.00001161, -0.00001006, 0.00001246, 0.00000124, 0.00000395, -0.00000860, 0.00000643, -0.00000087, -0.00000048, 0.00001249, -0.00002229, 0.00002093, -0.00002131, 0.00001595, -0.00001837, 0.00001942, -0.00001278, 0.00000651, -0.00000832, 0.00000672, -0.00000612, 0.00001392, -0.00002228, 0.00001923, -0.00000456, -0.00000134, -0.00001104, 0.00001152, -0.00001028, 0.00000785, -0.00000793, 0.00000676, -0.00000431, 0.00000956, -0.00001004, },
		{ -0.00001108, -0.00000416, -0.00000471, 0.00002310, -0.00000783, -0.00001166, -0.00000306, 0.00000947, -0.00000226, 0.00000255, 0.00001212, -0.00002798, 0.00001013, 0.00001205, -0.00000241, -0.00000813, 0.00000680, -0.00000303, 0.00000179, -0.00000210, 0.00000312, 0.00000256, -0.00000525, 0.00000364, 0.00000265, -0.00001099, 0.00000299, -0.00000142, 0.00001692, -0.00001153, 0.00000255, -0.00000243, 0.00000246, -0.00000900, 0.00000756, 0.00000434, -0.00000426, 0.00000324, -0.00000925, 0.00001314, -0.00002251, 0.00002942, -0.00002284, 0.00001021, -0.00000078, -0.00001280, 0.00000397, -0.00000366, -0.00000314, 0.00001303, -0.00002614, 0.00002352, -0.00002637, 0.00002071, -0.00000785, 0.00001029, -0.00001149, 0.00000567, -0.00000677, 0.00000332, -0.00000890, 0.00000183, -0.00000972, 0.00000992, -0.00001448, 0.00000103, -0.00001482, -0.00000515, -0.00002138, -0.00002890, -0.00006952, -0.00042247, 0.00124746, -0.00051457, 0.00000371, -0.00007060, 0.00000663, -0.00002958, 0.00001139, -0.00002538, 0.00001035, -0.00000990, 0.00000124, -0.00001052, 0.00000785, -0.00000966, 0.00000420, -0.00000872, 0.00000542, 0.00000113, -0.00000845, 0.00000848, -0.00000326, -0.00002220, 0.00002481, -0.00002488, 0.00003065, -0.00001661, 0.00002007, -0.00002553, 0.00002741, -0.00002438, 0.00002674, -0.00001741, 0.00000442, -0.00000152, -0.00000248, -0.00000075, -0.00000288, 0.00000719, -0.00000402, -0.00000235, 0.00000151, -0.00000362, 0.00000634, 0.00000094, -0.00001294, 0.00001481, -0.00000197, -0.00000730, -0.00000322, 0.00000511, -0.00000534, 0.00000374, -0.00000474, 0.00000426, -0.00000117, 0.00000699, },
		{ 0.00000208, -0.00001017, -0.00000564, 0.00000151, 0.00002103, -0.00001426, -0.00000666, -0.00000407, 0.00001285, -0.00000847, 0.00001162, 0.00000086, -0.00002287, 0.00001411, 0.00000960, -0.00000643, -0.00000341, 0.00000504, -0.00000202, 0.00000009, -0.00000017, 0.00000217, 0.00000339, -0.00000811, 0.00000948, -0.00000540, -0.00000427, -0.00000195, 0.00000576, 0.00001124, -0.00000995, 0.00000115, 0.00000030, -0.00000161, -0.00000580, 0.00000857, 0.00000207, -0.00000327, 0.00000149, -0.00000626, 0.00000984, -0.00002080, 0.00003423, -0.00003631, 0.00002862, -0.00002128, 0.00000426, -0.00000852, 0.00000604, -0.00001064, 0.00001945, -0.00003633, 0.00003838, -0.00004513, 0.00004472, -0.00003307, 0.00003582, -0.00004023, 0.00003674, -0.00003933, 0.00003746, -0.00004610, 0.00004137, -0.00005152, 0.00005741, -0.00007103, 0.00006564, -0.00009090, 0.00008480, -0.00013734, 0.00012117, -0.00033754, -0.00001348, 0.00106746, -0.00067093, 0.00017404, -0.00017959, 0.00009429, -0.00009670, 0.00006810, -0.00007464, 0.00005634, -0.00005092, 0.00003597, -0.00004014, 0.00003546, -0.00003543, 0.00002774, -0.00003062, 0.00002785, -0.00002039, 0.00000879, -0.00000275, -0.00000006, -0.00002349, 0.00002913, -0.00003024, 0.00003954, -0.00002577, 0.00002703, -0.00003168, 0.00003652, -0.00003593, 0.00004291, -0.00003951, 0.00002712, -0.00002240, 0.00001600, -0.00001769, 0.00001368, -0.00000688, 0.00000603, -0.00001120, 0.00001056, -0.00001280, 0.00001814, -0.00001312, -0.00000114, 0.00000814, 0.00000133, -0.00001287, 0.00000460, -0.00000211, 0.00000089, -0.00000193, 0.00000011, 0.00000034, 0.00000353, },
		{ 0.00000819, -0.00000394, -0.00000801, -0.00000721, 0.00000901, 0.00001463, -0.00001632, -0.00000378, -0.00000337, 0.00001375, -0.00001217, 0.00001909, -0.00001150, -0.00001351, 0.00001431, 0.00000776, -0.00000994, 0.00000211, 0.00000186, -0.00000024, -0.00000193, 0.00000207, 0.00000124, 0.00000318, -0.00000887, 0.00001290, -0.00001190, 0.00000210, -0.00000679, 0.00001334, 0.00000413, -0.00000728, -0.00000020, 0.00000242, -0.00000545, -0.00000170, 0.00000858, -0.00000057, -0.00000080, -0.00000249, 0.00000044, 0.00000030, -0.00000920, 0.00002565, -0.00003541, 0.00003421, -0.00003188, 0.00001435, -0.00001515, 0.00000949, -0.00001057, 0.00001589, -0.00003398, 0.00003936, -0.00004915, 0.00005393, -0.00004375, 0.00004604, -0.00005285, 0.00005154, -0.00005530, 0.00005415, -0.00006493, 0.00006157, -0.00007180, 0.00008014, -0.00009997, 0.00009994, -0.00013324, 0.00013590, -0.00020801, 0.00021489, -0.00055723, 0.00044586, 0.00070657, -0.00063769, 0.00022611, -0.00021401, 0.00012664, -0.00012024, 0.00008790, -0.00009166, 0.00007419, -0.00006814, 0.00005024, -0.00005101, 0.00004550, -0.00004501, 0.00003625, -0.00003819, 0.00003718, -0.00003125, 0.00001789, -0.00000748, -0.00000392, -0.00001527, 0.00002226, -0.00002349, 0.00003627, -0.00002354, 0.00002260, -0.00002482, 0.00003075, -0.00003112, 0.00004167, -0.00004563, 0.00003702, -0.00003324, 0.00002633, -0.00002771, 0.00002473, -0.00001718, 0.00001289, -0.00001624, 0.00001518, -0.00001695, 0.00002497, -0.00002393, 0.00000951, 0.00000190, 0.00000325, -0.00001589, 0.00000993, -0.00000755, 0.00000587, -0.00000690, 0.00000469, -0.00000331, },
		{ -0.00000504, 0.00001105, -0.00000947, -0.00000567, -0.00000769, 0.00001591, 0.00000550, -0.00001415, -0.00000385, 0.00000003, 0.00001090, -0.00001137, 0.00002200, -0.00002133, -0.00000297, 0.00001214, 0.00000596, -0.00001193, 0.00000699, -0.00000161, 0.00000154, -0.00000355, 0.00000401, 0.00000068, 0.00000175, -0.00000712, 0.00001289, -0.00001515, 0.00000618, -0.00000964, 0.00001949, -0.00000306, -0.00000420, -0.00000107, 0.00000331, -0.00000819, 0.00000270, 0.00000726, -0.00000236, 0.00000130, -0.00000612, 0.00000729, -0.00001105, 0.00000761, 0.00000724, -0.00002149, 0.00002598, -0.00003007, 0.00001460, -0.00001369, 0.00000532, -0.00000236, 0.00000219, -0.00001864, 0.00002526, -0.00003651, 0.00004573, -0.00003694, 0.00003763, -0.00004537, 0.00004533, -0.00004940, 0.00004754, -0.00005891, 0.00005545, -0.00006312, 0.00006954, -0.00009078, 0.00009083, -0.00012548, 0.00012700, -0.00020430, 0.00020590, -0.00064646, 0.00083991, 0.00026702, -0.00046610, 0.00017375, -0.00017580, 0.00010283, -0.00009863, 0.00006912, -0.00007448, 0.00006150, -0.00005894, 0.00004177, -0.00004141, 0.00003652, -0.00003710, 0.00002858, -0.00003025, 0.00003181, -0.00002965, 0.00001751, -0.00000558, -0.00001344, 0.00000009, 0.00000676, -0.00000698, 0.00002293, -0.00001213, 0.00000963, -0.00000845, 0.00001355, -0.00001265, 0.00002397, -0.00003443, 0.00003108, -0.00003026, 0.00002454, -0.00002676, 0.00002620, -0.00001997, 0.00001367, -0.00001532, 0.00001352, -0.00001404, 0.00002419, -0.00002835, 0.00001613, -0.00000187, 0.00000245, -0.00001515, 0.00001129, -0.00000944, 0.00000774, -0.00000927, 0.00000724, },
		{ 0.00000692, -0.00000394, 0.00001089, -0.00001314, -0.00000413, -0.00000603, 0.00002014, -0.00000385, -0.00000936, -0.00000616, 0.00000564, 0.00000457, -0.00000569, 0.00001881, -0.00002579, 0.00000573, 0.00000946, 0.00000340, -0.00001144, 0.00000987, -0.00000417, 0.00000250, -0.00000410, 0.00000511, 0.00000064, -0.00000064, -0.00000322, 0.00000939, -0.00001448, 0.00000688, -0.00000918, 0.00002271, -0.00000905, -0.00000131, -0.00000130, 0.00000275, -0.00000927, 0.00000674, 0.00000461, -0.00000239, 0.00000141, -0.00000687, 0.00001042, -0.00001862, 0.00002253, -0.00001348, -0.00000102, 0.00000858, -0.00001869, 0.00000693, -0.00000619, -0.00000368, 0.00001020, -0.00001691, 0.00000461, 0.00000098, -0.00001172, 0.00002427, -0.00001679, 0.00001494, -0.00002197, 0.00002207, -0.00002552, 0.00002156, -0.00003196, 0.00002724, -0.00003037, 0.00003099, -0.00004880, 0.00004334, -0.00007214, 0.00006158, -0.00012714, 0.00008436, -0.00055986, 0.00107279, -0.00014004, -0.00023570, 0.00006148, -0.00009247, 0.00004284, -0.00004774, 0.00002515, -0.00003457, 0.00002784, -0.00003129, 0.00001767, -0.00001803, 0.00001481, -0.00001757, 0.00001031, -0.00001197, 0.00001610, -0.00001902, 0.00001068, -0.00000057, -0.00002388, 0.00001681, -0.00001117, 0.00001316, 0.00000532, 0.00000273, -0.00000547, 0.00001005, -0.00000695, 0.00001108, -0.00000257, -0.00001158, 0.00001295, -0.00001569, 0.00001197, -0.00001549, 0.00001801, -0.00001484, 0.00000825, -0.00000885, 0.00000635, -0.00000478, 0.00001582, -0.00002555, 0.00001762, -0.00000250, -0.00000112, -0.00001085, 0.00000870, -0.00000753, 0.00000595, -0.00000825, },
		{ -0.00000441, 0.00000406, -0.00000020, 0.00000742, -0.00001424, -0.00000384, -0.00000184, 0.00002021, -0.00001083, -0.00000440, -0.00000885, 0.00001166, -0.00000338, 0.00000333, 0.00001011, -0.00002390, 0.00001062, 0.00000767, -0.00000032, -0.00000815, 0.00000996, -0.00000503, 0.00000214, -0.00000323, 0.00000510, 0.00000104, -0.00000338, 0.00000181, 0.00000341, -0.00001038, 0.00000421, -0.00000504, 0.00002213, -0.00001289, 0.00000097, -0.00000099, 0.00000094, -0.00000841, 0.00000969, 0.00000124, -0.00000072, -0.00000089, -0.00000372, 0.00000762, -0.00001833, 0.00002912, -0.00002828, 0.00001739, -0.00001022, -0.00000408, -0.00000346, 0.00000245, -0.00001215, 0.00002073, -0.00003378, 0.00002721, -0.00002456, 0.00001626, -0.00000158, 0.00000755, -0.00001226, 0.00000711, -0.00000774, 0.00000538, -0.00001226, 0.00000380, -0.00001004, 0.00001179, -0.00001889, 0.00000745, -0.00002201, 0.00000358, -0.00003364, -0.00000813, -0.00011707, -0.00030487, 0.00109602, -0.00042745, -0.00002541, -0.00005666, -0.00000222, -0.00002378, 0.00000816, -0.00002306, 0.00000970, -0.00001080, 0.00000103, -0.00000988, 0.00000793, -0.00000924, 0.00000394, -0.00000945, 0.00000806, -0.00000233, -0.00000565, 0.00000263, 0.00000262, -0.00002964, 0.00002832, -0.00002443, 0.00002977, -0.00000978, 0.00001458, -0.00001591, 0.00002282, -0.00002163, 0.00002974, -0.00002713, 0.00001303, -0.00000904, 0.00000358, -0.00000558, 0.00000127, 0.00000392, -0.00000462, -0.00000080, 0.00000041, -0.00000326, 0.00000763, 0.00000246, -0.00001702, 0.00001472, -0.00000077, -0.00000623, -0.00000451, 0.00000361, -0.00000303, 0.00000140, },
		{ -0.00000394, 0.00000062, 0.00000007, 0.00000491, 0.00000146, -0.00001298, -0.00000449, 0.00000425, 0.00001585, -0.00001369, -0.00000155, -0.00000973, 0.00001569, -0.00001017, 0.00001272, -0.00000158, -0.00001681, 0.00001140, 0.00000717, -0.00000482, -0.00000262, 0.00000736, -0.00000408, 0.00000055, -0.00000109, 0.00000410, 0.00000158, -0.00000562, 0.00000655, -0.00000337, -0.00000437, -0.00000068, 0.00000206, 0.00001780, -0.00001412, 0.00000239, -0.00000042, -0.00000166, -0.00000569, 0.00001085, -0.00000177, 0.00000159, -0.00000450, 0.00000221, -0.00000043, -0.00000955, 0.00002447, -0.00003173, 0.00002666, -0.00002301, 0.00000706, -0.00001092, 0.00000713, -0.00001493, 0.00002334, -0.00004119, 0.00004067, -0.00004207, 0.00003778, -0.00002208, 0.00002602, -0.00003308, 0.00003021, -0.00003184, 0.00003031, -0.00004002, 0.00003362, -0.00004054, 0.00004559, -0.00005968, 0.00005456, -0.00007822, 0.00006988, -0.00011990, 0.00010234, -0.00033320, 0.00006008, 0.00091971, -0.00055357, 0.00011034, -0.00013752, 0.00006225, -0.00007089, 0.00004728, -0.00005650, 0.00004130, -0.00003931, 0.00002491, -0.00002936, 0.00002600, -0.00002602, 0.00001861, -0.00002238, 0.00002185, -0.00001621, 0.00000447, -0.00000212, 0.00000040, -0.00002701, 0.00003025, -0.00002803, 0.00003763, -0.00001753, 0.00001922, -0.00001759, 0.00002506, -0.00002448, 0.00003552, -0.00004033, 0.00002958, -0.00002569, 0.00001921, -0.00002058, 0.00001672, -0.00001029, 0.00000603, -0.00000953, 0.00000861, -0.00001122, 0.00001878, -0.00001172, -0.00000591, 0.00000954, 0.00000161, -0.00001096, 0.00000168, -0.00000174, 0.00000191, },
		{ 0.00000516, -0.00000793, 0.00000484, -0.00000321, 0.00000951, -0.00000542, -0.00001027, -0.00000514, 0.00001074, 0.00000817, -0.00001216, -0.00000196, -0.00000736, 0.00001585, -0.00001317, 0.00001914, -0.00001264, -0.00000724, 0.00000929, 0.00000732, -0.00000896, 0.00000372, 0.00000307, -0.00000195, -0.00000166, 0.00000171, 0.00000257, 0.00000181, -0.00000659, 0.00000961, -0.00000905, 0.00000158, -0.00000594, 0.00001050, 0.00001058, -0.00001283, 0.00000290, 0.00000010, -0.00000438, -0.00000155, 0.00000982, -0.00000328, 0.00000296, -0.00000729, 0.00000805, -0.00001028, 0.00000446, 0.00001041, -0.00002329, 0.00002407, -0.00002575, 0.00001063, -0.00001203, 0.00000517, -0.00000966, 0.00001540, -0.00003550, 0.00004006, -0.00004560, 0.00004631, -0.00003063, 0.00003184, -0.00004009, 0.00003912, -0.00004141, 0.00003979, -0.00005140, 0.00004647, -0.00005246, 0.00005781, -0.00007588, 0.00007414, -0.00010336, 0.00009987, -0.00016267, 0.00015487, -0.00048835, 0.00044436, 0.00060460, -0.00052691, 0.00015319, -0.00016127, 0.00008462, -0.00008507, 0.00005834, -0.00006517, 0.00005097, -0.00004921, 0.00003299, -0.00003451, 0.00003062, -0.00003049, 0.00002191, -0.00002427, 0.00002517, -0.00002152, 0.00000823, -0.00000183, -0.00000768, -0.00001594, 0.00002213, -0.00002093, 0.00003533, -0.00001671, 0.00001623, -0.00001077, 0.00001702, -0.00001506, 0.00002641, -0.00003794, 0.00003237, -0.00003081, 0.00002507, -0.00002702, 0.00002502, -0.00001931, 0.00001278, -0.00001435, 0.00001251, -0.00001409, 0.00002468, -0.00002257, 0.00000434, 0.00000449, 0.00000285, -0.00001355, 0.00000576, -0.00000529, },
		{ -0.00000593, 0.00000544, -0.00000910, 0.00000675, -0.00000424, 0.00001183, -0.00001143, -0.00000732, -0.00000465, 0.00001564, -0.00000057, -0.00000756, -0.00000527, -0.00000174, 0.00001162, -0.00001107, 0.00002021, -0.00001979, 0.00000161, 0.00000631, 0.00000694, -0.00001131, 0.00000913, -0.00000144, 0.00000031, -0.00000356, 0.00000431, 0.00000109, 0.00000139, -0.00000584, 0.00001008, -0.00001210, 0.00000560, -0.00000954, 0.00001823, 0.00000199, -0.00000961, 0.00000263, 0.00000025, -0.00000648, 0.00000319, 0.00000683, -0.00000269, 0.00000217, -0.00000723, 0.00001055, -0.00001729, 0.00001794, -0.00000737, -0.00000729, 0.00001195, -0.00001914, 0.00000659, -0.00000704, -0.00000247, 0.00000197, -0.00000111, -0.00001820, 0.00002586, -0.00003464, 0.00004071, -0.00002607, 0.00002403, -0.00003201, 0.00003209, -0.00003441, 0.00003147, -0.00004362, 0.00003930, -0.00004280, 0.00004511, -0.00006312, 0.00006029, -0.00008955, 0.00008271, -0.00014609, 0.00012087, -0.00052257, 0.00075378, 0.00023934, -0.00039464, 0.00011841, -0.00013305, 0.00006756, -0.00006821, 0.00004306, -0.00005045, 0.00003953, -0.00004082, 0.00002563, -0.00002597, 0.00002259, -0.00002353, 0.00001489, -0.00001615, 0.00001868, -0.00001856, 0.00000614, 0.00000198, -0.00001882, 0.00000002, 0.00000747, -0.00000612, 0.00002529, -0.00000969, 0.00000871, 0.00000040, 0.00000356, 0.00000154, 0.00000668, -0.00002232, 0.00002180, -0.00002355, 0.00001958, -0.00002280, 0.00002358, -0.00002044, 0.00001334, -0.00001363, 0.00001083, -0.00001054, 0.00002329, -0.00002736, 0.00001117, 0.00000130, 0.00000193, -0.00001317, 0.00000676, },
		{ 0.00000502, -0.00000390, 0.00000279, -0.00000719, 0.00000589, -0.00000243, 0.00001082, -0.00001523, -0.00000507, -0.00000228, 0.00001726, -0.00000777, -0.00000222, -0.00000971, 0.00000562, 0.00000423, -0.00000445, 0.00001545, -0.00002117, 0.00000727, 0.00000428, 0.00000503, -0.00001080, 0.00001206, -0.00000476, 0.00000171, -0.00000429, 0.00000591, 0.00000016, 0.00000021, -0.00000345, 0.00000782, -0.00001188, 0.00000654, -0.00000993, 0.00002329, -0.00000620, -0.00000538, 0.00000185, -0.00000019, -0.00000725, 0.00000741, 0.00000277, -0.00000035, -0.00000089, -0.00000356, 0.00000777, -0.00001772, 0.00002514, -0.00002176, 0.00000923, -0.00000374, -0.00000762, -0.00000159, 0.00000063, -0.00001178, 0.00001504, -0.00002038, 0.00000470, 0.00000356, -0.00001392, 0.00002509, -0.00001256, 0.00000725, -0.00001373, 0.00001396, -0.00001580, 0.00001048, -0.00002186, 0.00001759, -0.00001786, 0.00001474, -0.00002926, 0.00002128, -0.00004554, 0.00002763, -0.00007967, 0.00000454, -0.00041215, 0.00091876, -0.00008812, -0.00022101, 0.00004182, -0.00007517, 0.00002751, -0.00003364, 0.00001297, -0.00002245, 0.00001562, -0.00002143, 0.00000928, -0.00000981, 0.00000762, -0.00001048, 0.00000274, -0.00000308, 0.00000680, -0.00001092, 0.00000140, 0.00000575, -0.00002837, 0.00001535, -0.00000798, 0.00001101, 0.00001218, -0.00000070, 0.00000132, 0.00001020, -0.00000843, 0.00001726, -0.00001521, -0.00000088, 0.00000362, -0.00000821, 0.00000602, -0.00001033, 0.00001382, -0.00001432, 0.00000824, -0.00000818, 0.00000476, -0.00000177, 0.00001512, -0.00002557, 0.00001368, 0.00000046, -0.00000108, -0.00001015, },
		{ -0.00000579, 0.00000188, -0.00000053, -0.00000153, -0.00000316, 0.00000300, 0.00000165, 0.00000653, -0.00001627, -0.00000384, 0.00000194, 0.00001485, -0.00001141, 0.00000139, -0.00001294, 0.00001228, -0.00000382, 0.00000439, 0.00000644, -0.00001703, 0.00000881, 0.00000409, 0.00000137, -0.00000719, 0.00001181, -0.00000606, 0.00000169, -0.00000344, 0.00000612, -0.00000000, -0.00000150, -0.00000005, 0.00000352, -0.00000872, 0.00000433, -0.00000655, 0.00002438, -0.00001241, -0.00000115, 0.00000086, -0.00000126, -0.00000628, 0.00001000, -0.00000099, 0.00000248, -0.00000502, 0.00000258, 0.00000020, -0.00001063, 0.00002289, -0.00002735, 0.00001957, -0.00001631, 0.00000292, -0.00000899, 0.00000638, -0.00001775, 0.00002357, -0.00003525, 0.00002524, -0.00001880, 0.00000894, 0.00000654, 0.00000279, -0.00001074, 0.00000644, -0.00000660, 0.00000530, -0.00001346, 0.00000369, -0.00000787, 0.00001026, -0.00001934, 0.00000980, -0.00002488, 0.00000789, -0.00004069, 0.00000563, -0.00015764, -0.00017733, 0.00091277, -0.00031505, -0.00006469, -0.00003731, -0.00001481, -0.00001466, 0.00000160, -0.00001716, 0.00000569, -0.00000921, -0.00000095, -0.00000739, 0.00000604, -0.00000699, 0.00000193, -0.00000814, 0.00000875, -0.00000480, -0.00000327, -0.00000226, 0.00000595, -0.00003213, 0.00002497, -0.00001871, 0.00002506, 0.00000068, 0.00000639, -0.00000225, 0.00001419, -0.00001331, 0.00002482, -0.00003046, 0.00001737, -0.00001396, 0.00000800, -0.00000929, 0.00000490, 0.00000023, -0.00000434, 0.00000026, -0.00000073, -0.00000270, 0.00000900, 0.00000286, -0.00001875, 0.00001259, 0.00000128, -0.00000510, },
		{ -0.00000869, -0.00000172, -0.00000099, 0.00000236, -0.00000567, 0.00000126, -0.00000039, 0.00000650, 0.00000005, -0.00001490, -0.00000325, 0.00000703, 0.00000898, -0.00001076, 0.00000168, -0.00001302, 0.00001590, -0.00000967, 0.00001229, -0.00000378, -0.00000941, 0.00000698, 0.00000540, -0.00000330, -0.00000135, 0.00000872, -0.00000534, 0.00000037, -0.00000121, 0.00000503, 0.00000043, -0.00000323, 0.00000339, -0.00000153, -0.00000389, -0.00000001, 0.00000005, 0.00002120, -0.00001565, 0.00000223, -0.00000012, -0.00000277, -0.00000356, 0.00001023, -0.00000312, 0.00000417, -0.00000812, 0.00000858, -0.00000926, 0.00000158, 0.00001194, -0.00002257, 0.00002020, -0.00002118, 0.00000799, -0.00001173, 0.00000679, -0.00001686, 0.00002332, -0.00004015, 0.00003664, -0.00003385, 0.00002663, -0.00000811, 0.00001328, -0.00002269, 0.00002053, -0.00002116, 0.00001989, -0.00003060, 0.00002271, -0.00002617, 0.00002941, -0.00004303, 0.00003754, -0.00005882, 0.00004753, -0.00009336, 0.00007024, -0.00030959, 0.00012507, 0.00075561, -0.00041693, 0.00003745, -0.00009047, 0.00002717, -0.00004276, 0.00002427, -0.00003584, 0.00002364, -0.00002569, 0.00001266, -0.00001755, 0.00001547, -0.00001572, 0.00000876, -0.00001315, 0.00001473, -0.00001183, 0.00000099, -0.00000249, 0.00000091, -0.00002824, 0.00002629, -0.00002147, 0.00003263, -0.00000639, 0.00000983, -0.00000096, 0.00001129, -0.00000934, 0.00002092, -0.00003368, 0.00002568, -0.00002395, 0.00001835, -0.00001992, 0.00001679, -0.00001181, 0.00000511, -0.00000701, 0.00000544, -0.00000812, 0.00001789, -0.00000971, -0.00000965, 0.00000958, 0.00000243, },
		{ 0.00000271, -0.00001071, 0.00000088, -0.00000233, 0.00000337, -0.00000798, 0.00000439, -0.00000258, 0.00001025, -0.00000693, -0.00001210, -0.00000249, 0.00001142, 0.00000143, -0.00000663, -0.00000148, -0.00000929, 0.00001517, -0.00001131, 0.00001645, -0.00001188, -0.00000130, 0.00000365, 0.00000700, -0.00000751, 0.00000511, 0.00000396, -0.00000334, -0.00000158, 0.00000169, 0.00000318, 0.00000103, -0.00000438, 0.00000585, -0.00000585, 0.00000093, -0.00000478, 0.00000832, 0.00001448, -0.00001567, 0.00000428, -0.00000102, -0.00000423, 0.00000033, 0.00000812, -0.00000291, 0.00000349, -0.00000828, 0.00001143, -0.00001648, 0.00001411, -0.00000333, -0.00001017, 0.00001190, -0.00001760, 0.00000625, -0.00000876, 0.00000129, -0.00000877, 0.00001362, -0.00003337, 0.00003584, -0.00003757, 0.00003485, -0.00001495, 0.00001541, -0.00002481, 0.00002424, -0.00002507, 0.00002291, -0.00003527, 0.00002908, -0.00003102, 0.00003274, -0.00004813, 0.00004396, -0.00006825, 0.00005792, -0.00010983, 0.00008236, -0.00039562, 0.00042071, 0.00050207, -0.00040651, 0.00007576, -0.00010676, 0.00004237, -0.00005031, 0.00002933, -0.00003879, 0.00002737, -0.00003006, 0.00001620, -0.00001867, 0.00001640, -0.00001680, 0.00000866, -0.00001116, 0.00001362, -0.00001294, 0.00000097, 0.00000061, -0.00000836, -0.00001784, 0.00001993, -0.00001618, 0.00003314, -0.00000867, 0.00001030, 0.00000333, 0.00000405, 0.00000075, 0.00000750, -0.00002485, 0.00002217, -0.00002337, 0.00001931, -0.00002204, 0.00002122, -0.00001822, 0.00001057, -0.00001082, 0.00000802, -0.00000908, 0.00002185, -0.00001922, -0.00000106, 0.00000647, },
		{ 0.00000439, 0.00000154, -0.00001077, 0.00000165, -0.00000182, 0.00000204, -0.00000768, 0.00000525, -0.00000248, 0.00001142, -0.00001274, -0.00000891, -0.00000079, 0.00001349, -0.00000544, -0.00000108, -0.00000674, -0.00000263, 0.00001044, -0.00000832, 0.00001554, -0.00001552, 0.00000460, 0.00000089, 0.00000747, -0.00000973, 0.00001034, -0.00000090, -0.00000114, -0.00000319, 0.00000433, 0.00000131, 0.00000134, -0.00000452, 0.00000661, -0.00000826, 0.00000418, -0.00000809, 0.00001618, 0.00000577, -0.00001295, 0.00000488, -0.00000189, -0.00000504, 0.00000435, 0.00000447, -0.00000067, 0.00000033, -0.00000476, 0.00000925, -0.00001795, 0.00002179, -0.00001685, 0.00000416, -0.00000095, -0.00000848, -0.00000029, -0.00000206, -0.00000768, 0.00000353, -0.00000238, -0.00001747, 0.00002428, -0.00003032, 0.00003337, -0.00001407, 0.00000988, -0.00001795, 0.00001822, -0.00001889, 0.00001483, -0.00002768, 0.00002293, -0.00002290, 0.00002106, -0.00003533, 0.00002923, -0.00005266, 0.00003747, -0.00008667, 0.00003109, -0.00037977, 0.00064121, 0.00022149, -0.00032167, 0.00006309, -0.00009161, 0.00003458, -0.00004060, 0.00001998, -0.00002897, 0.00001937, -0.00002425, 0.00001145, -0.00001266, 0.00001075, -0.00001217, 0.00000375, -0.00000443, 0.00000734, -0.00000961, -0.00000185, 0.00000493, -0.00001882, -0.00000441, 0.00000911, -0.00000536, 0.00002828, -0.00000761, 0.00000991, 0.00000721, -0.00000294, 0.00001127, -0.00000943, -0.00000886, 0.00000999, -0.00001399, 0.00001172, -0.00001569, 0.00001746, -0.00001769, 0.00001074, -0.00001031, 0.00000654, -0.00000510, 0.00001973, -0.00002370, 0.00000513, },
		{ 0.00000835, 0.00000357, -0.00000084, -0.00000931, 0.00000113, -0.00000013, -0.00000101, -0.00000512, 0.00000395, 0.00000002, 0.00000938, -0.00001628, -0.00000608, 0.00000210, 0.00001230, -0.00000952, 0.00000346, -0.00001186, 0.00000489, 0.00000353, -0.00000205, 0.00001010, -0.00001408, 0.00000680, 0.00000009, 0.00000593, -0.00000899, 0.00001293, -0.00000449, 0.00000031, -0.00000369, 0.00000592, 0.00000004, 0.00000103, -0.00000354, 0.00000552, -0.00000826, 0.00000496, -0.00000855, 0.00002167, -0.00000303, -0.00000853, 0.00000427, -0.00000280, -0.00000466, 0.00000732, 0.00000071, 0.00000231, -0.00000409, 0.00000143, 0.00000229, -0.00001248, 0.00002137, -0.00002349, 0.00001440, -0.00001257, 0.00000129, -0.00000765, 0.00000460, -0.00001591, 0.00001514, -0.00001908, 0.00000192, 0.00000685, -0.00001595, 0.00002529, -0.00000863, 0.00000060, -0.00000649, 0.00000698, -0.00000738, 0.00000076, -0.00001299, 0.00000956, -0.00000792, 0.00000163, -0.00001293, 0.00000257, -0.00002245, -0.00000180, -0.00003857, -0.00006879, -0.00025776, 0.00074357, -0.00002386, -0.00020820, 0.00002376, -0.00005985, 0.00001483, -0.00002267, 0.00000426, -0.00001364, 0.00000600, -0.00001363, 0.00000313, -0.00000395, 0.00000269, -0.00000574, -0.00000207, 0.00000306, -0.00000041, -0.00000489, -0.00000483, 0.00000760, -0.00002686, 0.00000785, -0.00000195, 0.00000733, 0.00002072, -0.00000505, 0.00001061, 0.00000781, -0.00000553, 0.00001647, -0.00002266, 0.00000702, -0.00000454, -0.00000103, -0.00000011, -0.00000438, 0.00000798, -0.00001175, 0.00000668, -0.00000656, 0.00000240, 0.00000224, 0.00001245, -0.00002301, },
		{ -0.00001851, 0.00000912, 0.00000351, -0.00000364, -0.00000725, 0.00000035, 0.00000151, -0.00000441, -0.00000152, 0.00000149, 0.00000399, 0.00000454, -0.00001723, -0.00000378, 0.00000571, 0.00000799, -0.00000972, 0.00000508, -0.00001465, 0.00001086, -0.00000303, 0.00000498, 0.00000232, -0.00000881, 0.00000545, 0.00000143, 0.00000248, -0.00000531, 0.00001240, -0.00000612, 0.00000054, -0.00000279, 0.00000610, -0.00000035, 0.00000010, -0.00000175, 0.00000304, -0.00000618, 0.00000328, -0.00000568, 0.00002346, -0.00001018, -0.00000359, 0.00000285, -0.00000372, -0.00000289, 0.00000832, -0.00000174, 0.00000434, -0.00000772, 0.00000781, -0.00000695, -0.00000182, 0.00001290, -0.00002113, 0.00001673, -0.00001833, 0.00000728, -0.00001194, 0.00000762, -0.00001953, 0.00002139, -0.00003081, 0.00001852, -0.00001036, 0.00000018, 0.00001506, -0.00000286, -0.00000761, 0.00000409, -0.00000367, 0.00000322, -0.00001256, 0.00000173, -0.00000369, 0.00000577, -0.00001592, 0.00000777, -0.00002284, 0.00000675, -0.00004064, 0.00000847, -0.00018224, -0.00005726, 0.00071952, -0.00019437, -0.00010439, -0.00001822, -0.00002726, -0.00000512, -0.00000600, -0.00000960, -0.00000023, -0.00000610, -0.00000387, -0.00000390, 0.00000306, -0.00000377, -0.00000110, -0.00000536, 0.00000771, -0.00000617, -0.00000162, -0.00000576, 0.00000657, -0.00002990, 0.00001574, -0.00000972, 0.00001867, 0.00001270, -0.00000205, 0.00001281, 0.00000443, -0.00000224, 0.00001330, -0.00002709, 0.00001646, -0.00001504, 0.00000953, -0.00001067, 0.00000685, -0.00000290, -0.00000367, 0.00000091, -0.00000185, -0.00000192, 0.00001015, 0.00000246, },
		{ -0.00000736, -0.00001221, 0.00000853, 0.00000343, -0.00000605, -0.00000542, 0.00000018, 0.00000195, -0.00000671, 0.00000163, -0.00000066, 0.00000788, -0.00000186, -0.00001599, -0.00000172, 0.00000895, 0.00000181, -0.00000638, 0.00000313, -0.00001386, 0.00001349, -0.00000697, 0.00001010, -0.00000492, -0.00000214, 0.00000201, 0.00000400, -0.00000184, 0.00000026, 0.00000924, -0.00000588, -0.00000028, -0.00000074, 0.00000504, 0.00000004, -0.00000120, 0.00000029, 0.00000005, -0.00000299, 0.00000005, -0.00000003, 0.00002121, -0.00001454, 0.00000081, 0.00000097, -0.00000443, -0.00000003, 0.00000714, -0.00000201, 0.00000411, -0.00000857, 0.00001147, -0.00001468, 0.00000993, -0.00000022, -0.00001145, 0.00001093, -0.00001661, 0.00000733, -0.00001123, 0.00000530, -0.00001672, 0.00001983, -0.00003402, 0.00002775, -0.00002239, 0.00001357, 0.00000627, 0.00000018, -0.00001135, 0.00000987, -0.00000948, 0.00000830, -0.00001989, 0.00001091, -0.00001120, 0.00001206, -0.00002434, 0.00001794, -0.00003630, 0.00002172, -0.00006191, 0.00002878, -0.00026710, 0.00017253, 0.00059284, -0.00027864, -0.00003201, -0.00004727, -0.00000421, -0.00001774, 0.00000366, -0.00001675, 0.00000685, -0.00001289, 0.00000163, -0.00000691, 0.00000600, -0.00000659, 0.00000012, -0.00000467, 0.00000789, -0.00000819, -0.00000118, -0.00000393, 0.00000151, -0.00002736, 0.00001803, -0.00001243, 0.00002673, 0.00000544, 0.00000155, 0.00001515, -0.00000104, 0.00000502, 0.00000280, -0.00002178, 0.00001670, -0.00001782, 0.00001365, -0.00001574, 0.00001370, -0.00001112, 0.00000318, -0.00000396, 0.00000161, -0.00000411, 0.00001580, },
		{ 0.00001741, -0.00001472, -0.00000594, 0.00000755, 0.00000281, -0.00000761, -0.00000419, 0.00000092, 0.00000070, -0.00000703, 0.00000323, -0.00000130, 0.00001017, -0.00000829, -0.00001336, 0.00000056, 0.00001054, -0.00000426, -0.00000106, -0.00000164, -0.00000961, 0.00001230, -0.00000721, 0.00001157, -0.00000919, 0.00000333, -0.00000149, 0.00000635, -0.00000551, 0.00000611, 0.00000464, -0.00000445, -0.00000152, 0.00000177, 0.00000329, 0.00000082, -0.00000245, 0.00000196, -0.00000251, 0.00000009, -0.00000331, 0.00000698, 0.00001559, -0.00001572, 0.00000397, -0.00000107, -0.00000453, 0.00000310, 0.00000437, -0.00000024, 0.00000136, -0.00000577, 0.00001041, -0.00001748, 0.00001805, -0.00001279, 0.00000098, 0.00000018, -0.00000915, 0.00000225, -0.00000619, -0.00000145, -0.00000846, 0.00001110, -0.00002840, 0.00002810, -0.00002696, 0.00002187, 0.00000046, -0.00000019, -0.00001004, 0.00000994, -0.00000942, 0.00000664, -0.00001964, 0.00001273, -0.00001138, 0.00000953, -0.00002182, 0.00001501, -0.00003428, 0.00001754, -0.00005851, 0.00000827, -0.00028952, 0.00037764, 0.00040728, -0.00028912, 0.00000456, -0.00005870, 0.00000648, -0.00002143, 0.00000557, -0.00001683, 0.00000725, -0.00001393, 0.00000260, -0.00000595, 0.00000512, -0.00000616, -0.00000138, -0.00000092, 0.00000430, -0.00000682, -0.00000315, -0.00000024, -0.00000621, -0.00002053, 0.00001552, -0.00001014, 0.00003099, -0.00000073, 0.00000651, 0.00001537, -0.00000523, 0.00001194, -0.00001048, -0.00001002, 0.00000925, -0.00001305, 0.00001068, -0.00001404, 0.00001424, -0.00001438, 0.00000669, -0.00000634, 0.00000270, -0.00000302, },
		{ 0.00000097, 0.00001473, -0.00001856, -0.00000082, 0.00000668, 0.00000157, -0.00000837, -0.00000335, 0.00000216, -0.00000186, -0.00000541, 0.00000303, 0.00000005, 0.00000987, -0.00001341, -0.00001012, 0.00000328, 0.00000963, -0.00000830, 0.00000406, -0.00000739, -0.00000330, 0.00000823, -0.00000417, 0.00000918, -0.00000939, 0.00000581, -0.00000337, 0.00000719, -0.00000717, 0.00001059, 0.00000001, -0.00000275, -0.00000245, 0.00000394, 0.00000155, 0.00000150, -0.00000328, 0.00000283, -0.00000394, 0.00000203, -0.00000538, 0.00001358, 0.00000799, -0.00001409, 0.00000561, -0.00000306, -0.00000366, 0.00000546, 0.00000126, 0.00000246, -0.00000290, -0.00000006, 0.00000456, -0.00001387, 0.00001926, -0.00001999, 0.00001080, -0.00001059, 0.00000014, -0.00000494, 0.00000046, -0.00000974, 0.00000206, -0.00000155, -0.00001658, 0.00002111, -0.00002455, 0.00002512, -0.00000290, -0.00000248, -0.00000563, 0.00000627, -0.00000556, 0.00000045, -0.00001384, 0.00000916, -0.00000672, 0.00000144, -0.00001215, 0.00000308, -0.00002122, -0.00000082, -0.00003630, -0.00004927, -0.00023567, 0.00051585, 0.00021061, -0.00025158, 0.00001326, -0.00005613, 0.00000782, -0.00001909, 0.00000273, -0.00001279, 0.00000365, -0.00001144, 0.00000105, -0.00000304, 0.00000241, -0.00000443, -0.00000346, 0.00000348, -0.00000083, -0.00000387, -0.00000595, 0.00000350, -0.00001432, -0.00001180, 0.00001022, -0.00000409, 0.00003172, -0.00000559, 0.00001316, 0.00001187, -0.00000521, 0.00001408, -0.00002077, 0.00000249, -0.00000128, -0.00000422, 0.00000317, -0.00000734, 0.00000936, -0.00001275, 0.00000660, -0.00000615, 0.00000170, },
		{ -0.00000005, 0.00000624, 0.00000889, -0.00001900, 0.00000285, 0.00000595, -0.00000003, -0.00000870, -0.00000234, 0.00000300, -0.00000469, -0.00000269, 0.00000161, 0.00000295, 0.00000686, -0.00001645, -0.00000673, 0.00000616, 0.00000615, -0.00000908, 0.00000701, -0.00001204, 0.00000301, 0.00000316, 0.00000047, 0.00000423, -0.00000602, 0.00000492, -0.00000291, 0.00000596, -0.00000616, 0.00001261, -0.00000352, -0.00000150, -0.00000252, 0.00000517, 0.00000037, 0.00000170, -0.00000352, 0.00000283, -0.00000405, 0.00000238, -0.00000523, 0.00001817, 0.00000006, -0.00001049, 0.00000573, -0.00000477, -0.00000169, 0.00000622, -0.00000083, 0.00000447, -0.00000672, 0.00000632, -0.00000403, -0.00000502, 0.00001307, -0.00001940, 0.00001414, -0.00001693, 0.00000706, -0.00001068, 0.00000546, -0.00001612, 0.00001095, -0.00001383, -0.00000278, 0.00001015, -0.00001741, 0.00002474, -0.00000533, -0.00000430, -0.00000116, 0.00000213, -0.00000147, -0.00000631, -0.00000652, 0.00000419, -0.00000177, -0.00000659, -0.00000210, -0.00000996, -0.00000648, -0.00002185, -0.00001099, -0.00012354, -0.00011568, 0.00056728, 0.00004076, -0.00019344, 0.00000661, -0.00004656, 0.00000501, -0.00001503, -0.00000086, -0.00000844, -0.00000046, -0.00000839, -0.00000049, -0.00000059, 0.00000010, -0.00000345, -0.00000397, 0.00000607, -0.00000499, -0.00000139, -0.00000786, 0.00000554, -0.00002070, -0.00000350, 0.00000433, 0.00000417, 0.00002929, -0.00000845, 0.00002054, 0.00000492, -0.00000025, 0.00000918, -0.00002380, 0.00001037, -0.00000964, 0.00000399, -0.00000473, 0.00000074, 0.00000193, -0.00000813, 0.00000418, -0.00000450, },
		{ -0.00000282, -0.00000095, 0.00001080, 0.00000173, -0.00001696, 0.00000527, 0.00000517, -0.00000163, -0.00000895, -0.00000069, 0.00000261, -0.00000661, -0.00000007, 0.00000010, 0.00000627, 0.00000185, -0.00001728, -0.00000333, 0.00000848, 0.00000101, -0.00000655, 0.00000675, -0.00001401, 0.00000750, -0.00000092, 0.00000458, -0.00000108, -0.00000089, 0.00000162, -0.00000053, 0.00000308, -0.00000275, 0.00001193, -0.00000545, -0.00000101, -0.00000161, 0.00000527, -0.00000001, 0.00000128, -0.00000321, 0.00000220, -0.00000316, 0.00000136, -0.00000271, 0.00001979, -0.00000671, -0.00000594, 0.00000455, -0.00000586, 0.00000107, 0.00000510, -0.00000100, 0.00000444, -0.00000813, 0.00001058, -0.00001193, 0.00000563, 0.00000203, -0.00001190, 0.00001013, -0.00001668, 0.00000899, -0.00001251, 0.00000652, -0.00001813, 0.00001526, -0.00002210, 0.00000903, -0.00000115, -0.00000820, 0.00002224, -0.00000796, -0.00000396, 0.00000105, 0.00000012, -0.00000008, -0.00001022, -0.00000125, 0.00000129, -0.00000026, -0.00000992, 0.00000249, -0.00001691, 0.00000099, -0.00003386, -0.00000039, -0.00018576, 0.00004128, 0.00053632, -0.00008171, -0.00013527, -0.00000468, -0.00003590, 0.00000204, -0.00001221, -0.00000263, -0.00000621, -0.00000280, -0.00000666, -0.00000049, 0.00000006, -0.00000063, -0.00000414, -0.00000201, 0.00000559, -0.00000669, -0.00000072, -0.00000791, 0.00000501, -0.00002420, 0.00000289, -0.00000050, 0.00001332, 0.00002390, -0.00000806, 0.00002642, -0.00000312, 0.00000753, -0.00000162, -0.00001870, 0.00001099, -0.00001240, 0.00000793, -0.00000935, 0.00000660, -0.00000481, -0.00000305, 0.00000121, },
		{ -0.00000096, -0.00000204, -0.00000003, 0.00001312, -0.00000498, -0.00001353, 0.00000676, 0.00000420, -0.00000308, -0.00000913, 0.00000157, 0.00000067, -0.00000684, 0.00000146, -0.00000043, 0.00000869, -0.00000399, -0.00001624, 0.00000012, 0.00000934, -0.00000425, -0.00000180, 0.00000346, -0.00001281, 0.00000919, -0.00000270, 0.00000648, -0.00000461, 0.00000376, -0.00000236, 0.00000257, -0.00000033, 0.00000198, 0.00000910, -0.00000587, -0.00000113, -0.00000001, 0.00000445, 0.00000031, 0.00000034, -0.00000254, 0.00000138, -0.00000192, -0.00000026, 0.00000156, 0.00001827, -0.00001135, -0.00000139, 0.00000238, -0.00000595, 0.00000388, 0.00000262, 0.00000071, 0.00000199, -0.00000607, 0.00001063, -0.00001584, 0.00001382, -0.00000938, -0.00000115, 0.00000100, -0.00001064, 0.00000577, -0.00001004, 0.00000331, -0.00001529, 0.00001405, -0.00002463, 0.00001637, -0.00001021, 0.00000116, 0.00001838, -0.00001071, -0.00000145, 0.00000056, 0.00000085, -0.00000220, -0.00001007, 0.00000057, 0.00000153, -0.00000310, -0.00000728, -0.00000032, -0.00001502, -0.00000269, -0.00003115, -0.00001579, -0.00020989, 0.00019790, 0.00044550, -0.00015374, -0.00008759, -0.00001487, -0.00002676, 0.00000012, -0.00001108, -0.00000243, -0.00000623, -0.00000314, -0.00000636, 0.00000090, -0.00000086, -0.00000011, -0.00000590, 0.00000172, 0.00000239, -0.00000586, -0.00000198, -0.00000619, 0.00000208, -0.00002476, 0.00000690, -0.00000336, 0.00002232, 0.00001590, -0.00000331, 0.00002820, -0.00000881, 0.00001409, -0.00001419, -0.00000824, 0.00000530, -0.00000926, 0.00000658, -0.00000917, 0.00000823, -0.00000863, 0.00000059, },
		{ 0.00000211, -0.00000194, -0.00000218, 0.00000267, 0.00001263, -0.00001013, -0.00000955, 0.00000752, 0.00000313, -0.00000455, -0.00000885, 0.00000385, -0.00000236, -0.00000537, 0.00000154, 0.00000058, 0.00000918, -0.00000942, -0.00001384, 0.00000358, 0.00000811, -0.00000797, 0.00000339, -0.00000160, -0.00000912, 0.00000829, -0.00000205, 0.00000561, -0.00000512, 0.00000614, -0.00000529, 0.00000505, -0.00000295, 0.00000665, 0.00000515, -0.00000530, -0.00000144, 0.00000176, 0.00000316, 0.00000103, -0.00000091, -0.00000172, 0.00000073, -0.00000093, -0.00000153, 0.00000644, 0.00001420, -0.00001348, 0.00000243, -0.00000042, -0.00000474, 0.00000577, -0.00000009, 0.00000333, -0.00000206, -0.00000104, 0.00000604, -0.00001401, 0.00001626, -0.00001671, 0.00000812, -0.00000908, -0.00000187, -0.00000058, -0.00000486, -0.00000267, -0.00000905, 0.00000845, -0.00002181, 0.00001873, -0.00001595, 0.00000974, 0.00001305, -0.00001233, 0.00000165, -0.00000120, 0.00000286, -0.00000632, -0.00000711, 0.00000007, 0.00000334, -0.00000804, -0.00000128, -0.00000783, -0.00000711, -0.00001455, -0.00001708, -0.00005631, -0.00018237, 0.00032125, 0.00032500, -0.00018429, -0.00005247, -0.00002276, -0.00001881, -0.00000175, -0.00001015, -0.00000191, -0.00000693, -0.00000279, -0.00000627, 0.00000236, -0.00000204, 0.00000035, -0.00000720, 0.00000544, -0.00000193, -0.00000358, -0.00000437, -0.00000350, -0.00000238, -0.00002302, 0.00000877, -0.00000383, 0.00003008, 0.00000629, 0.00000586, 0.00002431, -0.00000938, 0.00001540, -0.00002335, 0.00000262, -0.00000298, -0.00000275, 0.00000159, -0.00000513, 0.00000582, -0.00000905, },
		{ -0.00000712, 0.00000205, -0.00000220, -0.00000246, 0.00000620, 0.00000979, -0.00001335, -0.00000550, 0.00000756, 0.00000219, -0.00000631, -0.00000758, 0.00000523, -0.00000549, -0.00000292, 0.00000058, 0.00000293, 0.00000740, -0.00001355, -0.00001054, 0.00000672, 0.00000478, -0.00000901, 0.00000722, -0.00000670, -0.00000435, 0.00000592, 0.00000009, 0.00000268, -0.00000271, 0.00000556, -0.00000612, 0.00000598, -0.00000381, 0.00001001, 0.00000114, -0.00000431, -0.00000153, 0.00000318, 0.00000192, 0.00000175, -0.00000225, -0.00000083, 0.00000037, -0.00000055, -0.00000171, 0.00001071, 0.00000861, -0.00001324, 0.00000503, -0.00000336, -0.00000219, 0.00000597, -0.00000176, 0.00000533, -0.00000592, 0.00000504, -0.00000157, -0.00000707, 0.00001204, -0.00001737, 0.00001211, -0.00001592, 0.00000586, -0.00000703, 0.00000051, -0.00000897, -0.00000186, 0.00000082, -0.00001549, 0.00001699, -0.00001835, 0.00001722, 0.00000593, -0.00001136, 0.00000325, -0.00000201, 0.00000370, -0.00000974, -0.00000394, -0.00000028, 0.00000377, -0.00001125, 0.00000337, -0.00001454, 0.00000038, -0.00002664, -0.00000246, -0.00010865, -0.00010658, 0.00039193, 0.00020218, -0.00018686, -0.00002763, -0.00002926, -0.00001079, -0.00000505, -0.00000768, -0.00000295, -0.00000648, -0.00000334, -0.00000497, 0.00000248, -0.00000214, -0.00000050, -0.00000657, 0.00000731, -0.00000554, -0.00000131, -0.00000677, -0.00000084, -0.00000739, -0.00001978, 0.00000895, -0.00000163, 0.00003527, -0.00000295, 0.00001777, 0.00001528, -0.00000428, 0.00000953, -0.00002532, 0.00000915, -0.00000940, 0.00000331, -0.00000384, 0.00000015, 0.00000126, },
		{ -0.00000303, -0.00000455, 0.00000146, -0.00000254, -0.00000191, 0.00000925, 0.00000561, -0.00001480, -0.00000161, 0.00000689, 0.00000152, -0.00000848, -0.00000496, 0.00000489, -0.00000754, -0.00000056, -0.00000049, 0.00000578, 0.00000366, -0.00001594, -0.00000664, 0.00000893, 0.00000008, -0.00000712, 0.00000850, -0.00001032, -0.00000001, 0.00000352, 0.00000225, -0.00000075, 0.00000138, 0.00000252, -0.00000481, 0.00000522, -0.00000262, 0.00001137, -0.00000213, -0.00000336, -0.00000116, 0.00000395, 0.00000109, 0.00000216, -0.00000347, 0.00000015, 0.00000016, -0.00000067, -0.00000056, 0.00001346, 0.00000263, -0.00001111, 0.00000612, -0.00000582, 0.00000126, 0.00000428, -0.00000151, 0.00000539, -0.00000768, 0.00000956, -0.00000920, 0.00000218, 0.00000303, -0.00001163, 0.00000950, -0.00001706, 0.00000968, -0.00001102, 0.00000381, -0.00001337, 0.00000394, -0.00000636, -0.00000791, 0.00001261, -0.00001780, 0.00002315, -0.00000262, -0.00000725, 0.00000206, -0.00000035, 0.00000157, -0.00001025, -0.00000279, 0.00000152, 0.00000064, -0.00000975, 0.00000274, -0.00001552, 0.00000118, -0.00003068, -0.00000014, -0.00015232, -0.00000065, 0.00040728, 0.00009496, -0.00017327, -0.00001048, -0.00003486, -0.00000240, -0.00001007, -0.00000319, -0.00000621, -0.00000407, -0.00000545, -0.00000196, 0.00000077, -0.00000071, -0.00000301, -0.00000356, 0.00000641, -0.00000720, -0.00000015, -0.00000839, 0.00000113, -0.00001232, -0.00001559, 0.00000782, 0.00000349, 0.00003639, -0.00000909, 0.00002922, 0.00000390, 0.00000424, -0.00000223, -0.00001939, 0.00000898, -0.00001095, 0.00000588, -0.00000677, 0.00000386, },
		{ 0.00000439, -0.00000534, -0.00000270, 0.00000116, -0.00000342, 0.00000004, 0.00001080, 0.00000113, -0.00001482, 0.00000191, 0.00000570, 0.00000095, -0.00001074, -0.00000124, 0.00000258, -0.00000779, 0.00000077, -0.00000072, 0.00000800, -0.00000117, -0.00001654, -0.00000237, 0.00000952, -0.00000472, -0.00000303, 0.00000702, -0.00001165, 0.00000288, 0.00000214, 0.00000312, -0.00000298, 0.00000531, -0.00000161, -0.00000212, 0.00000334, 0.00000016, 0.00001070, -0.00000432, -0.00000260, -0.00000036, 0.00000404, 0.00000080, 0.00000206, -0.00000443, 0.00000127, -0.00000020, -0.00000086, 0.00000164, 0.00001429, -0.00000279, -0.00000771, 0.00000561, -0.00000713, 0.00000473, 0.00000124, 0.00000064, 0.00000304, -0.00000618, 0.00001039, -0.00001372, 0.00000992, -0.00000702, -0.00000239, 0.00000178, -0.00001254, 0.00000870, -0.00001143, 0.00000401, -0.00001478, 0.00000705, -0.00001151, -0.00000077, 0.00000679, -0.00001451, 0.00002657, -0.00001117, -0.00000097, -0.00000155, 0.00000362, -0.00000355, -0.00000744, -0.00000416, 0.00000558, -0.00000597, -0.00000333, -0.00000398, -0.00000929, -0.00000706, -0.00002291, -0.00001841, -0.00016796, 0.00010988, 0.00037826, 0.00001039, -0.00015083, -0.00000009, -0.00003848, 0.00000501, -0.00001536, 0.00000201, -0.00001064, -0.00000051, -0.00000837, 0.00000197, -0.00000206, 0.00000153, -0.00000629, 0.00000101, 0.00000310, -0.00000674, -0.00000042, -0.00000896, 0.00000217, -0.00001688, -0.00001071, 0.00000571, 0.00001142, 0.00003246, -0.00000963, 0.00003653, -0.00000588, 0.00001201, -0.00001575, -0.00000831, 0.00000300, -0.00000734, 0.00000405, -0.00000591, },
		{ -0.00000207, 0.00000190, -0.00000530, -0.00000201, 0.00000129, -0.00000464, 0.00000318, 0.00001050, -0.00000294, -0.00001364, 0.00000470, 0.00000443, 0.00000003, -0.00001239, 0.00000277, -0.00000117, -0.00000620, 0.00000073, 0.00000042, 0.00000868, -0.00000611, -0.00001556, 0.00000197, 0.00000807, -0.00000824, 0.00000187, 0.00000350, -0.00001082, 0.00000415, 0.00000209, 0.00000220, -0.00000295, 0.00000748, -0.00000526, 0.00000081, 0.00000132, 0.00000362, 0.00000849, -0.00000546, -0.00000199, 0.00000063, 0.00000363, 0.00000100, 0.00000140, -0.00000496, 0.00000249, -0.00000095, -0.00000057, 0.00000423, 0.00001329, -0.00000703, -0.00000369, 0.00000362, -0.00000674, 0.00000713, -0.00000197, 0.00000375, -0.00000102, -0.00000169, 0.00000685, -0.00001322, 0.00001300, -0.00001405, 0.00000620, -0.00000759, -0.00000463, 0.00000405, -0.00000882, 0.00000148, -0.00001344, 0.00000746, -0.00001420, 0.00000513, 0.00000027, -0.00000843, 0.00002620, -0.00001745, 0.00000521, -0.00000563, 0.00000800, -0.00000979, -0.00000287, -0.00000656, 0.00000992, -0.00001344, 0.00000511, -0.00001384, 0.00000113, -0.00002118, -0.00000665, -0.00005620, -0.00014505, 0.00020137, 0.00032247, -0.00005247, -0.00012286, 0.00000306, -0.00003807, 0.00000908, -0.00001846, 0.00000555, -0.00001406, 0.00000238, -0.00001047, 0.00000519, -0.00000450, 0.00000316, -0.00000876, 0.00000544, -0.00000130, -0.00000489, -0.00000180, -0.00000862, 0.00000230, -0.00002093, -0.00000544, 0.00000328, 0.00002130, 0.00002374, -0.00000342, 0.00003707, -0.00001061, 0.00001480, -0.00002587, 0.00000309, -0.00000529, -0.00000089, -0.00000072, },
		{ -0.00000555, 0.00000242, -0.00000196, -0.00000381, -0.00000201, 0.00000134, -0.00000544, 0.00000665, 0.00000871, -0.00000638, -0.00001136, 0.00000639, 0.00000355, -0.00000178, -0.00001255, 0.00000582, -0.00000519, -0.00000351, -0.00000032, 0.00000279, 0.00000740, -0.00001033, -0.00001326, 0.00000588, 0.00000467, -0.00000945, 0.00000605, -0.00000081, -0.00000859, 0.00000432, 0.00000282, -0.00000007, -0.00000058, 0.00000711, -0.00000730, 0.00000299, 0.00000006, 0.00000676, 0.00000545, -0.00000578, -0.00000139, 0.00000156, 0.00000301, 0.00000149, 0.00000018, -0.00000487, 0.00000359, -0.00000203, 0.00000054, 0.00000649, 0.00001094, -0.00000977, 0.00000028, 0.00000055, -0.00000448, 0.00000756, -0.00000400, 0.00000631, -0.00000514, 0.00000414, 0.00000015, -0.00000789, 0.00001021, -0.00001540, 0.00001053, -0.00001475, 0.00000339, -0.00000196, -0.00000483, -0.00000245, -0.00001045, 0.00000603, -0.00001481, 0.00000958, -0.00000643, 0.00000022, 0.00002113, -0.00001937, 0.00000873, -0.00000769, 0.00001016, -0.00001437, 0.00000082, -0.00000751, 0.00001157, -0.00001795, 0.00001099, -0.00002162, 0.00000972, -0.00003401, 0.00000866, -0.00010255, -0.00008592, 0.00025938, 0.00025667, -0.00009774, -0.00009109, -0.00000133, -0.00003215, 0.00000803, -0.00001740, 0.00000541, -0.00001450, 0.00000287, -0.00001020, 0.00000621, -0.00000517, 0.00000289, -0.00000898, 0.00000802, -0.00000519, -0.00000273, -0.00000371, -0.00000760, 0.00000146, -0.00002412, -0.00000033, 0.00000165, 0.00003126, 0.00001216, 0.00000848, 0.00003047, -0.00000902, 0.00001030, -0.00002864, 0.00001010, -0.00001161, 0.00000488, },
		{ 0.00000708, -0.00000776, 0.00000511, -0.00000511, -0.00000230, -0.00000183, 0.00000066, -0.00000494, 0.00000938, 0.00000616, -0.00000920, -0.00000810, 0.00000680, 0.00000335, -0.00000472, -0.00001062, 0.00000684, -0.00000817, -0.00000086, -0.00000150, 0.00000566, 0.00000434, -0.00001330, -0.00000989, 0.00000868, 0.00000003, -0.00000807, 0.00000840, -0.00000468, -0.00000597, 0.00000424, 0.00000337, -0.00000253, 0.00000316, 0.00000442, -0.00000737, 0.00000398, 0.00000006, 0.00000883, 0.00000226, -0.00000553, -0.00000072, 0.00000224, 0.00000243, 0.00000204, -0.00000151, -0.00000395, 0.00000415, -0.00000309, 0.00000241, 0.00000792, 0.00000777, -0.00001089, 0.00000356, -0.00000290, -0.00000069, 0.00000569, -0.00000388, 0.00000691, -0.00000741, 0.00000887, -0.00000710, 0.00000001, 0.00000283, -0.00001093, 0.00000896, -0.00001708, 0.00000867, -0.00000698, -0.00000123, -0.00000630, -0.00000704, 0.00000376, -0.00001396, 0.00001248, -0.00001254, 0.00001046, 0.00001167, -0.00001606, 0.00000806, -0.00000606, 0.00000822, -0.00001505, 0.00000148, -0.00000507, 0.00000836, -0.00001657, 0.00001050, -0.00002262, 0.00001070, -0.00003803, 0.00001175, -0.00014048, -0.00000454, 0.00028154, 0.00019170, -0.00012867, -0.00005801, -0.00001172, -0.00002121, 0.00000195, -0.00001194, 0.00000113, -0.00001134, 0.00000039, -0.00000708, 0.00000453, -0.00000363, 0.00000040, -0.00000645, 0.00000782, -0.00000744, -0.00000111, -0.00000571, -0.00000599, -0.00000066, -0.00002581, 0.00000376, 0.00000227, 0.00003875, 0.00000104, 0.00002292, 0.00001894, -0.00000268, -0.00000062, -0.00002312, 0.00001021, -0.00001299, },
		{ -0.00000902, 0.00000480, -0.00000616, 0.00000457, -0.00000612, -0.00000183, -0.00000090, -0.00000105, -0.00000275, 0.00001057, 0.00000348, -0.00001136, -0.00000422, 0.00000612, 0.00000364, -0.00000850, -0.00000668, 0.00000531, -0.00000915, 0.00000075, -0.00000184, 0.00000799, 0.00000018, -0.00001485, -0.00000571, 0.00000973, -0.00000471, -0.00000463, 0.00000852, -0.00000733, -0.00000368, 0.00000456, 0.00000285, -0.00000384, 0.00000680, 0.00000044, -0.00000585, 0.00000393, 0.00000124, 0.00000954, -0.00000062, -0.00000488, -0.00000004, 0.00000264, 0.00000201, 0.00000245, -0.00000347, -0.00000212, 0.00000377, -0.00000353, 0.00000461, 0.00000838, 0.00000426, -0.00001042, 0.00000557, -0.00000580, 0.00000366, 0.00000203, -0.00000151, 0.00000495, -0.00000655, 0.00001034, -0.00001195, 0.00000713, -0.00000597, -0.00000290, 0.00000242, -0.00001408, 0.00000986, -0.00000959, 0.00000083, -0.00000917, -0.00000394, 0.00000123, -0.00001192, 0.00001345, -0.00001679, 0.00002032, -0.00000029, -0.00000853, 0.00000349, -0.00000087, 0.00000212, -0.00001142, -0.00000134, 0.00000079, 0.00000039, -0.00000904, 0.00000279, -0.00001535, 0.00000174, -0.00002963, -0.00000484, -0.00015370, 0.00007914, 0.00027554, 0.00013143, -0.00014595, -0.00002771, -0.00002445, -0.00000798, -0.00000690, -0.00000393, -0.00000581, -0.00000569, -0.00000410, -0.00000203, 0.00000093, -0.00000062, -0.00000346, -0.00000190, 0.00000504, -0.00000776, -0.00000042, -0.00000756, -0.00000376, -0.00000439, -0.00002530, 0.00000605, 0.00000637, 0.00004134, -0.00000591, 0.00003560, 0.00000640, 0.00000463, -0.00001419, -0.00001178, 0.00000410, },
		{ -0.00000488, -0.00000187, -0.00000057, -0.00000160, 0.00000109, -0.00000484, -0.00000260, 0.00000069, -0.00000343, 0.00000082, 0.00001011, 0.00000098, -0.00001265, -0.00000036, 0.00000491, 0.00000380, -0.00001223, -0.00000158, 0.00000164, -0.00000794, 0.00000084, -0.00000074, 0.00000887, -0.00000423, -0.00001497, -0.00000114, 0.00000866, -0.00000832, -0.00000019, 0.00000675, -0.00000859, -0.00000194, 0.00000539, 0.00000098, -0.00000316, 0.00000894, -0.00000349, -0.00000357, 0.00000342, 0.00000312, 0.00000898, -0.00000296, -0.00000393, 0.00000056, 0.00000284, 0.00000181, 0.00000251, -0.00000539, 0.00000046, 0.00000224, -0.00000277, 0.00000643, 0.00000808, 0.00000072, -0.00000851, 0.00000588, -0.00000721, 0.00000731, -0.00000219, 0.00000224, 0.00000094, -0.00000258, 0.00000767, -0.00001239, 0.00001044, -0.00001251, 0.00000507, -0.00000623, -0.00000732, 0.00000732, -0.00000964, 0.00000114, -0.00001089, -0.00000125, -0.00000153, -0.00000860, 0.00001187, -0.00001764, 0.00002720, -0.00001168, 0.00000055, -0.00000283, 0.00000585, -0.00000619, -0.00000512, -0.00000604, 0.00000797, -0.00000968, 0.00000181, -0.00000937, -0.00000257, -0.00001439, -0.00001157, -0.00004056, -0.00013319, 0.00014662, 0.00025370, 0.00007498, -0.00014859, -0.00000500, -0.00003478, 0.00000354, -0.00001498, 0.00000351, -0.00001265, 0.00000014, -0.00000858, 0.00000307, -0.00000291, 0.00000229, -0.00000707, 0.00000306, 0.00000083, -0.00000664, -0.00000057, -0.00000919, -0.00000101, -0.00000981, -0.00002231, 0.00000636, 0.00001420, 0.00003781, -0.00000604, 0.00004272, -0.00000317, 0.00000863, -0.00002534, 0.00000068, },
		{ 0.00000935, -0.00001239, 0.00000493, -0.00000618, 0.00000364, -0.00000362, -0.00000229, -0.00000392, 0.00000223, -0.00000554, 0.00000483, 0.00000847, -0.00000144, -0.00001271, 0.00000272, 0.00000395, 0.00000298, -0.00001471, 0.00000326, -0.00000302, -0.00000518, -0.00000033, 0.00000175, 0.00000784, -0.00000813, -0.00001379, 0.00000325, 0.00000559, -0.00000990, 0.00000398, 0.00000390, -0.00000879, -0.00000045, 0.00000625, -0.00000176, -0.00000046, 0.00000883, -0.00000630, -0.00000136, 0.00000308, 0.00000507, 0.00000749, -0.00000468, -0.00000276, 0.00000098, 0.00000293, 0.00000183, 0.00000201, -0.00000681, 0.00000333, -0.00000024, -0.00000059, 0.00000729, 0.00000739, -0.00000263, -0.00000547, 0.00000443, -0.00000653, 0.00000903, -0.00000546, 0.00000581, -0.00000356, 0.00000309, 0.00000168, -0.00000828, 0.00000856, -0.00001419, 0.00000954, -0.00001350, 0.00000042, 0.00000269, -0.00000795, 0.00000020, -0.00001179, 0.00000133, -0.00000474, -0.00000383, 0.00000738, -0.00001397, 0.00002881, -0.00001929, 0.00000791, -0.00000794, 0.00001110, -0.00001359, 0.00000095, -0.00000981, 0.00001318, -0.00001778, 0.00001125, -0.00002070, 0.00000982, -0.00003084, 0.00000755, -0.00008595, -0.00008091, 0.00018705, 0.00022705, 0.00002054, -0.00013613, 0.00000657, -0.00003875, 0.00000984, -0.00001908, 0.00000744, -0.00001666, 0.00000378, -0.00001096, 0.00000631, -0.00000530, 0.00000360, -0.00000882, 0.00000668, -0.00000335, -0.00000489, -0.00000134, -0.00001041, 0.00000176, -0.00001637, -0.00001728, 0.00000549, 0.00002458, 0.00002888, 0.00000120, 0.00004236, -0.00000747, 0.00000656, -0.00002985, },
		{ -0.00002622, 0.00001127, -0.00001514, 0.00000828, -0.00000921, 0.00000698, -0.00000741, -0.00000005, -0.00000475, 0.00000282, -0.00000636, 0.00000814, 0.00000638, -0.00000397, -0.00001126, 0.00000444, 0.00000383, 0.00000058, -0.00001490, 0.00000638, -0.00000717, -0.00000203, -0.00000188, 0.00000494, 0.00000501, -0.00001102, -0.00001145, 0.00000673, 0.00000118, -0.00000914, 0.00000693, 0.00000082, -0.00000845, 0.00000124, 0.00000637, -0.00000434, 0.00000344, 0.00000657, -0.00000752, 0.00000032, 0.00000329, 0.00000655, 0.00000544, -0.00000575, -0.00000150, 0.00000122, 0.00000298, 0.00000202, 0.00000081, -0.00000722, 0.00000575, -0.00000301, 0.00000273, 0.00000691, 0.00000660, -0.00000553, -0.00000181, 0.00000158, -0.00000372, 0.00000817, -0.00000653, 0.00000764, -0.00000663, 0.00000817, -0.00000535, -0.00000138, 0.00000231, -0.00001051, 0.00000873, -0.00001673, 0.00000648, -0.00000211, -0.00000570, -0.00000131, -0.00001218, 0.00000390, -0.00000835, 0.00000206, 0.00000038, -0.00000584, 0.00002420, -0.00002103, 0.00001103, -0.00000937, 0.00001230, -0.00001727, 0.00000406, -0.00001013, 0.00001358, -0.00002031, 0.00001485, -0.00002599, 0.00001559, -0.00003989, 0.00001684, -0.00012566, -0.00000906, 0.00019974, 0.00020131, -0.00003154, -0.00011072, 0.00000639, -0.00003489, 0.00000947, -0.00001775, 0.00000641, -0.00001640, 0.00000394, -0.00001015, 0.00000662, -0.00000533, 0.00000258, -0.00000787, 0.00000781, -0.00000634, -0.00000324, -0.00000258, -0.00001081, 0.00000365, -0.00002289, -0.00001135, 0.00000506, 0.00003503, 0.00001725, 0.00001386, 0.00003521, -0.00000674, -0.00000159, },
		{ -0.00001297, -0.00001620, 0.00000651, -0.00001223, 0.00000691, -0.00000820, 0.00000683, -0.00000870, 0.00000069, -0.00000432, 0.00000197, -0.00000531, 0.00000992, 0.00000447, -0.00000663, -0.00000844, 0.00000469, 0.00000462, -0.00000343, -0.00001242, 0.00000682, -0.00000952, 0.00000029, -0.00000277, 0.00000779, 0.00000101, -0.00001268, -0.00000815, 0.00000861, -0.00000353, -0.00000641, 0.00000821, -0.00000188, -0.00000791, 0.00000340, 0.00000512, -0.00000559, 0.00000720, 0.00000299, -0.00000732, 0.00000146, 0.00000405, 0.00000731, 0.00000315, -0.00000614, -0.00000034, 0.00000138, 0.00000298, 0.00000233, -0.00000111, -0.00000623, 0.00000692, -0.00000510, 0.00000641, 0.00000553, 0.00000575, -0.00000767, 0.00000174, -0.00000185, 0.00000055, 0.00000488, -0.00000497, 0.00000682, -0.00000678, 0.00001045, -0.00001061, 0.00000531, -0.00000564, -0.00000327, 0.00000310, -0.00001503, 0.00000926, -0.00000565, -0.00000372, -0.00000303, -0.00001201, 0.00000610, -0.00001170, 0.00000799, -0.00000763, 0.00000517, 0.00001432, -0.00001681, 0.00000920, -0.00000639, 0.00000855, -0.00001599, 0.00000301, -0.00000613, 0.00000831, -0.00001595, 0.00001067, -0.00002262, 0.00001131, -0.00003679, 0.00000807, -0.00014420, 0.00006448, 0.00019257, 0.00017616, -0.00007765, -0.00007764, -0.00000273, -0.00002473, 0.00000355, -0.00001188, 0.00000112, -0.00001227, 0.00000093, -0.00000647, 0.00000428, -0.00000329, -0.00000036, -0.00000452, 0.00000632, -0.00000766, -0.00000205, -0.00000428, -0.00000992, 0.00000360, -0.00002794, -0.00000595, 0.00000695, 0.00004255, 0.00000673, 0.00002821, 0.00002405, -0.00000352, },
		{ -0.00000129, -0.00002318, -0.00000398, -0.00000207, -0.00000538, 0.00000183, -0.00000371, 0.00000328, -0.00000722, -0.00000038, -0.00000259, -0.00000015, -0.00000251, 0.00001000, 0.00000289, -0.00000910, -0.00000484, 0.00000394, 0.00000578, -0.00000833, -0.00000777, 0.00000451, -0.00000950, 0.00000107, -0.00000221, 0.00000928, -0.00000333, -0.00001311, -0.00000423, 0.00000850, -0.00000741, -0.00000257, 0.00000794, -0.00000402, -0.00000717, 0.00000586, 0.00000240, -0.00000478, 0.00000959, -0.00000079, -0.00000626, 0.00000237, 0.00000507, 0.00000738, 0.00000080, -0.00000583, 0.00000050, 0.00000160, 0.00000291, 0.00000262, -0.00000355, -0.00000374, 0.00000631, -0.00000559, 0.00000941, 0.00000373, 0.00000469, -0.00000867, 0.00000438, -0.00000478, 0.00000505, 0.00000020, -0.00000137, 0.00000355, -0.00000370, 0.00000878, -0.00001200, 0.00000891, -0.00001193, 0.00000434, -0.00000498, -0.00000944, 0.00000862, -0.00000746, -0.00000230, -0.00000502, -0.00001081, 0.00000707, -0.00001363, 0.00001231, -0.00001433, 0.00001623, 0.00000190, -0.00000859, 0.00000379, -0.00000025, 0.00000105, -0.00001065, -0.00000132, 0.00000086, -0.00000090, -0.00000644, 0.00000026, -0.00001196, -0.00000184, -0.00002246, -0.00002011, -0.00013195, 0.00012273, 0.00017730, 0.00014749, -0.00011222, -0.00004416, -0.00001557, -0.00001202, -0.00000477, -0.00000415, -0.00000608, -0.00000623, -0.00000357, -0.00000144, 0.00000062, -0.00000042, -0.00000392, -0.00000004, 0.00000299, -0.00000753, -0.00000136, -0.00000644, -0.00000749, 0.00000089, -0.00003040, -0.00000208, 0.00001239, 0.00004476, 0.00000091, 0.00004008, 0.00001245, },
		{ 0.00000321, -0.00000272, -0.00002822, 0.00000570, -0.00001033, 0.00000209, -0.00000430, 0.00000212, -0.00000210, -0.00000393, -0.00000259, -0.00000024, -0.00000269, 0.00000120, 0.00000882, 0.00000139, -0.00001077, -0.00000134, 0.00000302, 0.00000637, -0.00001287, -0.00000223, 0.00000033, -0.00000743, 0.00000030, -0.00000003, 0.00000879, -0.00000718, -0.00001245, -0.00000021, 0.00000640, -0.00000959, 0.00000138, 0.00000661, -0.00000574, -0.00000592, 0.00000799, -0.00000119, -0.00000193, 0.00000987, -0.00000380, -0.00000495, 0.00000343, 0.00000592, 0.00000692, -0.00000147, -0.00000489, 0.00000093, 0.00000199, 0.00000276, 0.00000268, -0.00000606, -0.00000014, 0.00000387, -0.00000391, 0.00001080, 0.00000224, 0.00000315, -0.00000824, 0.00000549, -0.00000619, 0.00000837, -0.00000433, 0.00000286, -0.00000088, 0.00000155, 0.00000357, -0.00000901, 0.00000782, -0.00001396, 0.00000908, -0.00001233, -0.00000225, 0.00000563, -0.00000785, -0.00000129, -0.00000752, -0.00000803, 0.00000594, -0.00001297, 0.00001340, -0.00001740, 0.00002416, -0.00000948, 0.00000043, -0.00000249, 0.00000633, -0.00000744, -0.00000378, -0.00000648, 0.00000794, -0.00001049, 0.00000418, -0.00001200, 0.00000119, -0.00001841, -0.00000359, -0.00006082, -0.00008899, 0.00015534, 0.00016412, 0.00011108, -0.00013029, -0.00001719, -0.00002657, -0.00000106, -0.00001188, 0.00000223, -0.00001228, -0.00000078, -0.00000740, 0.00000297, -0.00000267, 0.00000179, -0.00000652, 0.00000396, -0.00000094, -0.00000652, -0.00000111, -0.00000882, -0.00000383, -0.00000447, -0.00002994, 0.00000024, 0.00002126, 0.00004096, 0.00000188, 0.00004641, },
		{ 0.00004612, -0.00000273, -0.00000836, -0.00002626, 0.00000960, -0.00001475, 0.00000694, -0.00000856, 0.00000667, -0.00000706, -0.00000051, -0.00000475, 0.00000172, -0.00000458, 0.00000471, 0.00000717, -0.00000042, -0.00001105, 0.00000117, 0.00000278, 0.00000545, -0.00001567, 0.00000260, -0.00000431, -0.00000430, -0.00000133, 0.00000327, 0.00000628, -0.00000996, -0.00001085, 0.00000330, 0.00000283, -0.00000972, 0.00000459, 0.00000478, -0.00000724, -0.00000378, 0.00000896, -0.00000457, 0.00000216, 0.00000812, -0.00000558, -0.00000375, 0.00000483, 0.00000629, 0.00000612, -0.00000351, -0.00000349, 0.00000099, 0.00000253, 0.00000259, 0.00000224, -0.00000798, 0.00000378, 0.00000020, -0.00000017, 0.00001009, 0.00000162, 0.00000099, -0.00000634, 0.00000479, -0.00000550, 0.00000943, -0.00000721, 0.00000605, -0.00000460, 0.00000689, -0.00000324, -0.00000295, 0.00000247, -0.00001098, 0.00000900, -0.00001629, 0.00000406, 0.00000182, -0.00000748, -0.00000050, -0.00001052, -0.00000359, 0.00000240, -0.00000925, 0.00001049, -0.00001542, 0.00002659, -0.00001671, 0.00000710, -0.00000678, 0.00001046, -0.00001382, 0.00000166, -0.00000969, 0.00001198, -0.00001666, 0.00001140, -0.00002094, 0.00001085, -0.00003120, 0.00001005, -0.00010092, -0.00002512, 0.00016134, 0.00015783, 0.00006582, -0.00012998, -0.00000111, -0.00003181, 0.00000508, -0.00001517, 0.00000497, -0.00001532, 0.00000219, -0.00000894, 0.00000530, -0.00000432, 0.00000229, -0.00000707, 0.00000620, -0.00000431, -0.00000524, -0.00000132, -0.00001088, 0.00000015, -0.00001162, -0.00002715, 0.00000205, 0.00003188, 0.00003248, 0.00000942, },
		{ 0.00002130, 0.00004016, -0.00000627, -0.00001641, -0.00001833, 0.00000711, -0.00001386, 0.00000742, -0.00000913, 0.00000800, -0.00000975, 0.00000155, -0.00000580, 0.00000248, -0.00000500, 0.00000710, 0.00000567, -0.00000276, -0.00000971, 0.00000223, 0.00000363, 0.00000255, -0.00001585, 0.00000537, -0.00000792, -0.00000137, -0.00000277, 0.00000669, 0.00000224, -0.00001139, -0.00000852, 0.00000569, -0.00000135, -0.00000798, 0.00000664, 0.00000282, -0.00000859, -0.00000066, 0.00000813, -0.00000656, 0.00000622, 0.00000504, -0.00000625, -0.00000261, 0.00000644, 0.00000613, 0.00000507, -0.00000510, -0.00000195, 0.00000086, 0.00000309, 0.00000250, 0.00000104, -0.00000863, 0.00000696, -0.00000359, 0.00000482, 0.00000749, 0.00000197, -0.00000165, -0.00000333, 0.00000258, -0.00000278, 0.00000783, -0.00000754, 0.00000697, -0.00000596, 0.00001010, -0.00000906, 0.00000357, -0.00000493, -0.00000432, 0.00000417, -0.00001569, 0.00000780, -0.00000158, -0.00000681, -0.00000002, -0.00001352, 0.00000184, -0.00000294, -0.00000307, 0.00000411, -0.00000858, 0.00002291, -0.00001825, 0.00000956, -0.00000736, 0.00001035, -0.00001606, 0.00000368, -0.00000920, 0.00001116, -0.00001708, 0.00001228, -0.00002298, 0.00001267, -0.00003462, 0.00000980, -0.00012581, 0.00004368, 0.00014819, 0.00015680, 0.00001502, -0.00011362, 0.00000346, -0.00003019, 0.00000560, -0.00001400, 0.00000352, -0.00001457, 0.00000211, -0.00000772, 0.00000510, -0.00000402, 0.00000095, -0.00000539, 0.00000616, -0.00000647, -0.00000405, -0.00000212, -0.00001191, 0.00000319, -0.00001913, -0.00002323, 0.00000501, 0.00004159, 0.00002227, },
		{ 0.00001385, 0.00003411, 0.00003069, -0.00000937, -0.00002360, -0.00000778, 0.00000031, -0.00000866, 0.00000391, -0.00000597, 0.00000573, -0.00000946, 0.00000157, -0.00000531, 0.00000186, -0.00000380, 0.00000804, 0.00000457, -0.00000542, -0.00000709, 0.00000200, 0.00000535, -0.00000206, -0.00001329, 0.00000543, -0.00000954, 0.00000038, -0.00000304, 0.00000909, -0.00000243, -0.00001151, -0.00000574, 0.00000656, -0.00000520, -0.00000498, 0.00000751, 0.00000082, -0.00000954, 0.00000312, 0.00000543, -0.00000637, 0.00000903, 0.00000167, -0.00000626, -0.00000123, 0.00000785, 0.00000561, 0.00000374, -0.00000603, -0.00000058, 0.00000081, 0.00000350, 0.00000255, -0.00000099, -0.00000754, 0.00000841, -0.00000617, 0.00000973, 0.00000378, 0.00000296, -0.00000430, 0.00000010, -0.00000039, 0.00000119, 0.00000409, -0.00000539, 0.00000532, -0.00000424, 0.00000973, -0.00001164, 0.00000774, -0.00001132, 0.00000321, -0.00000346, -0.00001112, 0.00000852, -0.00000398, -0.00000595, -0.00000032, -0.00001550, 0.00000689, -0.00000854, 0.00000387, -0.00000385, 0.00000126, 0.00001446, -0.00001458, 0.00000783, -0.00000432, 0.00000605, -0.00001390, 0.00000203, -0.00000511, 0.00000580, -0.00001197, 0.00000675, -0.00001776, 0.00000594, -0.00002754, -0.00000777, -0.00012487, 0.00010098, 0.00012776, 0.00015491, -0.00003454, -0.00008711, -0.00000072, -0.00002333, 0.00000190, -0.00000966, -0.00000094, -0.00001099, -0.00000024, -0.00000450, 0.00000301, -0.00000239, -0.00000149, -0.00000219, 0.00000415, -0.00000731, -0.00000309, -0.00000362, -0.00001136, 0.00000414, -0.00002560, -0.00001930, 0.00001064, 0.00004770, },
		{ 0.00004857, 0.00001008, 0.00004457, 0.00001993, -0.00001366, -0.00002683, 0.00000131, -0.00000725, -0.00000189, -0.00000148, -0.00000074, 0.00000099, -0.00000679, -0.00000014, -0.00000360, 0.00000032, -0.00000146, 0.00000780, 0.00000362, -0.00000781, -0.00000395, 0.00000116, 0.00000716, -0.00000740, -0.00000879, 0.00000306, -0.00000897, 0.00000056, -0.00000168, 0.00000961, -0.00000676, -0.00001055, -0.00000287, 0.00000582, -0.00000793, -0.00000151, 0.00000746, -0.00000130, -0.00000961, 0.00000676, 0.00000143, -0.00000394, 0.00000982, -0.00000109, -0.00000613, 0.00000072, 0.00000861, 0.00000502, 0.00000209, -0.00000617, 0.00000035, 0.00000105, 0.00000363, 0.00000266, -0.00000362, -0.00000469, 0.00000760, -0.00000647, 0.00001316, 0.00000012, 0.00000393, -0.00000630, 0.00000308, -0.00000308, 0.00000521, -0.00000060, -0.00000173, 0.00000193, -0.00000001, 0.00000572, -0.00001006, 0.00000776, -0.00001408, 0.00000849, -0.00001100, -0.00000448, 0.00000687, -0.00000547, -0.00000471, -0.00000195, -0.00001536, 0.00000997, -0.00001255, 0.00000940, -0.00001083, 0.00001121, 0.00000399, -0.00000784, 0.00000356, 0.00000068, -0.00000069, -0.00000887, -0.00000187, 0.00000076, -0.00000181, -0.00000388, -0.00000249, -0.00000816, -0.00000608, -0.00001396, -0.00003897, -0.00009547, 0.00013538, 0.00011133, 0.00014507, -0.00007516, -0.00005786, -0.00000904, -0.00001438, -0.00000341, -0.00000447, -0.00000624, -0.00000641, -0.00000327, -0.00000072, 0.00000028, -0.00000051, -0.00000385, 0.00000130, 0.00000104, -0.00000714, -0.00000239, -0.00000571, -0.00000917, 0.00000244, -0.00003015, -0.00001581, 0.00001942, },
		{ 0.00003050, 0.00004424, 0.00001220, 0.00005044, 0.00000924, -0.00001943, -0.00002460, 0.00000587, -0.00001223, 0.00000344, -0.00000608, 0.00000418, -0.00000422, -0.00000320, -0.00000256, -0.00000150, -0.00000131, 0.00000115, 0.00000702, 0.00000237, -0.00000921, -0.00000124, 0.00000060, 0.00000803, -0.00001215, -0.00000368, -0.00000073, -0.00000678, -0.00000048, 0.00000115, 0.00000788, -0.00000993, -0.00000886, -0.00000028, 0.00000374, -0.00000910, 0.00000174, 0.00000681, -0.00000367, -0.00000836, 0.00000930, -0.00000274, 0.00000001, 0.00000859, -0.00000281, -0.00000608, 0.00000335, 0.00000842, 0.00000461, 0.00000016, -0.00000553, 0.00000076, 0.00000164, 0.00000348, 0.00000263, -0.00000633, -0.00000062, 0.00000468, -0.00000405, 0.00001409, -0.00000243, 0.00000416, -0.00000704, 0.00000484, -0.00000453, 0.00000801, -0.00000477, 0.00000197, -0.00000165, 0.00000507, -0.00000052, -0.00000511, 0.00000359, -0.00001213, 0.00000940, -0.00001575, 0.00000194, 0.00000408, -0.00000643, -0.00000297, -0.00000513, -0.00001249, 0.00000996, -0.00001349, 0.00001171, -0.00001452, 0.00001838, -0.00000536, -0.00000082, -0.00000088, 0.00000523, -0.00000732, -0.00000337, -0.00000574, 0.00000587, -0.00000857, 0.00000359, -0.00001139, 0.00000121, -0.00001787, -0.00000140, -0.00007398, -0.00004389, 0.00014371, 0.00010547, 0.00012268, -0.00010100, -0.00003245, -0.00001695, -0.00000651, -0.00000777, -0.00000065, -0.00001030, -0.00000267, -0.00000542, 0.00000220, -0.00000188, 0.00000059, -0.00000507, 0.00000391, -0.00000218, -0.00000640, -0.00000203, -0.00000797, -0.00000587, -0.00000168, -0.00003271, -0.00001226, },
		{ -0.00000743, 0.00004181, 0.00003646, 0.00001967, 0.00005091, -0.00000104, -0.00002539, -0.00001770, 0.00000505, -0.00001291, 0.00000538, -0.00000790, 0.00000678, -0.00000800, -0.00000022, -0.00000460, 0.00000017, -0.00000224, 0.00000321, 0.00000629, 0.00000048, -0.00000914, 0.00000032, 0.00000102, 0.00000715, -0.00001509, 0.00000060, -0.00000461, -0.00000401, -0.00000193, 0.00000469, 0.00000416, -0.00001144, -0.00000687, 0.00000169, 0.00000086, -0.00000867, 0.00000433, 0.00000578, -0.00000619, -0.00000561, 0.00000992, -0.00000581, 0.00000426, 0.00000597, -0.00000361, -0.00000599, 0.00000639, 0.00000729, 0.00000442, -0.00000191, -0.00000431, 0.00000078, 0.00000246, 0.00000317, 0.00000215, -0.00000838, 0.00000369, 0.00000051, 0.00000068, 0.00001222, -0.00000322, 0.00000324, -0.00000621, 0.00000497, -0.00000421, 0.00000872, -0.00000721, 0.00000433, -0.00000378, 0.00000891, -0.00000668, 0.00000108, -0.00000310, -0.00000630, 0.00000561, -0.00001627, 0.00000634, 0.00000126, -0.00000714, -0.00000094, -0.00000943, -0.00000713, 0.00000673, -0.00001094, 0.00001008, -0.00001375, 0.00002096, -0.00001123, 0.00000418, -0.00000351, 0.00000733, -0.00001161, 0.00000042, -0.00000763, 0.00000813, -0.00001201, 0.00000744, -0.00001642, 0.00000618, -0.00002428, 0.00000245, -0.00010035, 0.00001705, 0.00013103, 0.00011040, 0.00008767, -0.00010992, -0.00001484, -0.00002150, -0.00000167, -0.00000970, 0.00000059, -0.00001195, -0.00000079, -0.00000581, 0.00000345, -0.00000282, 0.00000048, -0.00000468, 0.00000491, -0.00000479, -0.00000544, -0.00000214, -0.00000979, -0.00000242, -0.00000740, -0.00003382, },
		{ -0.00003415, -0.00000002, 0.00005088, 0.00002800, 0.00003052, 0.00004625, -0.00001117, -0.00002933, -0.00000880, 0.00000028, -0.00000968, 0.00000375, -0.00000644, 0.00000624, -0.00000928, 0.00000115, -0.00000551, 0.00000097, -0.00000216, 0.00000435, 0.00000588, -0.00000203, -0.00000761, 0.00000054, 0.00000256, 0.00000427, -0.00001558, 0.00000303, -0.00000744, -0.00000167, -0.00000282, 0.00000785, -0.00000079, -0.00001121, -0.00000496, 0.00000286, -0.00000222, -0.00000694, 0.00000609, 0.00000440, -0.00000854, -0.00000162, 0.00000833, -0.00000681, 0.00000752, 0.00000297, -0.00000398, -0.00000538, 0.00000924, 0.00000556, 0.00000433, -0.00000383, -0.00000283, 0.00000065, 0.00000329, 0.00000286, 0.00000095, -0.00000906, 0.00000708, -0.00000356, 0.00000657, 0.00000805, -0.00000224, 0.00000120, -0.00000395, 0.00000356, -0.00000215, 0.00000717, -0.00000742, 0.00000465, -0.00000344, 0.00000988, -0.00001050, 0.00000590, -0.00000961, 0.00000101, -0.00000141, -0.00001271, 0.00000801, -0.00000099, -0.00000753, 0.00000070, -0.00001377, -0.00000050, 0.00000129, -0.00000572, 0.00000516, -0.00000887, 0.00001871, -0.00001276, 0.00000611, -0.00000351, 0.00000618, -0.00001251, 0.00000146, -0.00000675, 0.00000689, -0.00001128, 0.00000655, -0.00001618, 0.00000504, -0.00002310, -0.00000676, -0.00010768, 0.00007257, 0.00010772, 0.00012098, 0.00004435, -0.00010379, -0.00000586, -0.00002189, -0.00000018, -0.00000917, -0.00000063, -0.00001123, -0.00000077, -0.00000450, 0.00000304, -0.00000259, -0.00000065, -0.00000290, 0.00000420, -0.00000642, -0.00000449, -0.00000284, -0.00001061, 0.00000015, -0.00001368, },
		{ -0.00001976, -0.00003395, 0.00001063, 0.00005567, 0.00002173, 0.00004211, 0.00003729, -0.00002116, -0.00002931, -0.00000119, -0.00000564, -0.00000454, -0.00000002, -0.00000270, 0.00000311, -0.00000817, 0.00000077, -0.00000522, 0.00000097, -0.00000125, 0.00000466, 0.00000561, -0.00000470, -0.00000512, -0.00000020, 0.00000483, -0.00000017, -0.00001372, 0.00000325, -0.00000859, -0.00000043, -0.00000243, 0.00000955, -0.00000589, -0.00000952, -0.00000343, 0.00000319, -0.00000494, -0.00000437, 0.00000707, 0.00000260, -0.00001015, 0.00000284, 0.00000491, -0.00000545, 0.00000891, 0.00000054, -0.00000449, -0.00000375, 0.00001120, 0.00000380, 0.00000405, -0.00000530, -0.00000144, 0.00000066, 0.00000389, 0.00000266, -0.00000102, -0.00000796, 0.00000857, -0.00000610, 0.00001201, 0.00000277, -0.00000011, -0.00000144, -0.00000085, 0.00000119, 0.00000101, 0.00000389, -0.00000570, 0.00000313, -0.00000067, 0.00000744, -0.00001069, 0.00000737, -0.00001335, 0.00000690, -0.00000907, -0.00000660, 0.00000728, -0.00000265, -0.00000728, 0.00000113, -0.00001675, 0.00000567, -0.00000460, 0.00000042, -0.00000126, -0.00000156, 0.00001294, -0.00001063, 0.00000535, -0.00000142, 0.00000243, -0.00001044, 0.00000008, -0.00000370, 0.00000305, -0.00000742, 0.00000197, -0.00001176, -0.00000093, -0.00001612, -0.00002789, -0.00009138, 0.00011082, 0.00008513, 0.00012961, -0.00000033, -0.00008739, -0.00000383, -0.00001904, -0.00000107, -0.00000716, -0.00000327, -0.00000904, -0.00000183, -0.00000225, 0.00000161, -0.00000173, -0.00000211, -0.00000043, 0.00000221, -0.00000706, -0.00000371, -0.00000407, -0.00001021, 0.00000116, },
		{ 0.00000046, -0.00002541, -0.00003271, 0.00002405, 0.00005534, 0.00001975, 0.00005190, 0.00002486, -0.00003021, -0.00002479, 0.00000270, -0.00000992, 0.00000005, -0.00000380, 0.00000146, -0.00000114, -0.00000562, -0.00000077, -0.00000411, 0.00000061, -0.00000006, 0.00000462, 0.00000508, -0.00000685, -0.00000250, -0.00000117, 0.00000701, -0.00000523, -0.00001022, 0.00000162, -0.00000809, -0.00000039, -0.00000054, 0.00000909, -0.00001001, -0.00000697, -0.00000242, 0.00000279, -0.00000691, -0.00000141, 0.00000737, 0.00000029, -0.00001046, 0.00000675, 0.00000063, -0.00000220, 0.00000823, -0.00000082, -0.00000543, -0.00000085, 0.00001169, 0.00000253, 0.00000332, -0.00000606, -0.00000038, 0.00000100, 0.00000411, 0.00000255, -0.00000355, -0.00000512, 0.00000774, -0.00000610, 0.00001548, -0.00000217, 0.00000222, -0.00000387, 0.00000225, -0.00000126, 0.00000430, -0.00000010, -0.00000299, 0.00000077, 0.00000346, 0.00000237, -0.00000741, 0.00000492, -0.00001291, 0.00000908, -0.00001469, -0.00000007, 0.00000516, -0.00000400, -0.00000610, -0.00000029, -0.00001718, 0.00000970, -0.00000905, 0.00000547, -0.00000703, 0.00000587, 0.00000579, -0.00000650, 0.00000320, 0.00000134, -0.00000233, -0.00000684, -0.00000244, 0.00000003, -0.00000155, -0.00000260, -0.00000383, -0.00000600, -0.00000826, -0.00000827, -0.00005445, -0.00005428, 0.00012633, 0.00007163, 0.00012944, -0.00003928, -0.00006645, -0.00000586, -0.00001470, -0.00000290, -0.00000500, -0.00000598, -0.00000656, -0.00000293, -0.00000004, -0.00000002, -0.00000093, -0.00000317, 0.00000185, -0.00000041, -0.00000692, -0.00000324, -0.00000554, -0.00000880, },
		{ -0.00000696, -0.00000163, -0.00003085, -0.00002933, 0.00003868, 0.00005050, 0.00002292, 0.00005785, 0.00000978, -0.00003674, -0.00001697, 0.00000222, -0.00001099, 0.00000232, -0.00000583, 0.00000431, -0.00000486, -0.00000290, -0.00000263, -0.00000274, 0.00000036, 0.00000087, 0.00000470, 0.00000386, -0.00000789, -0.00000053, -0.00000164, 0.00000823, -0.00000984, -0.00000616, -0.00000106, -0.00000647, -0.00000114, 0.00000248, 0.00000628, -0.00001229, -0.00000435, -0.00000186, 0.00000182, -0.00000790, 0.00000151, 0.00000708, -0.00000241, -0.00000913, 0.00000915, -0.00000324, 0.00000187, 0.00000598, -0.00000119, -0.00000667, 0.00000315, 0.00001054, 0.00000209, 0.00000197, -0.00000604, 0.00000024, 0.00000170, 0.00000394, 0.00000236, -0.00000610, -0.00000115, 0.00000486, -0.00000327, 0.00001601, -0.00000553, 0.00000377, -0.00000530, 0.00000448, -0.00000296, 0.00000674, -0.00000368, -0.00000041, -0.00000112, 0.00000725, -0.00000352, -0.00000216, -0.00000042, -0.00000854, 0.00000677, -0.00001652, 0.00000501, 0.00000273, -0.00000532, -0.00000403, -0.00000358, -0.00001463, 0.00001063, -0.00001080, 0.00000796, -0.00001039, 0.00001139, -0.00000060, -0.00000222, 0.00000114, 0.00000326, -0.00000637, -0.00000340, -0.00000459, 0.00000287, -0.00000506, 0.00000098, -0.00000833, -0.00000182, -0.00001327, -0.00000531, -0.00007705, -0.00000533, 0.00012084, 0.00007034, 0.00011687, -0.00006740, -0.00004599, -0.00000920, -0.00001044, -0.00000450, -0.00000365, -0.00000780, -0.00000468, -0.00000332, 0.00000138, -0.00000122, -0.00000062, -0.00000334, 0.00000325, -0.00000302, -0.00000627, -0.00000320, -0.00000684, },
		{ -0.00000760, -0.00000531, -0.00000463, -0.00003636, -0.00002251, 0.00005222, 0.00004298, 0.00003083, 0.00005848, -0.00000689, -0.00003901, -0.00000838, -0.00000140, -0.00000898, 0.00000188, -0.00000544, 0.00000489, -0.00000698, -0.00000102, -0.00000409, -0.00000153, 0.00000047, 0.00000126, 0.00000514, 0.00000186, -0.00000756, 0.00000034, -0.00000113, 0.00000790, -0.00001309, -0.00000256, -0.00000385, -0.00000447, -0.00000197, 0.00000581, 0.00000157, -0.00001227, -0.00000241, -0.00000150, 0.00000041, -0.00000783, 0.00000409, 0.00000623, -0.00000522, -0.00000623, 0.00000946, -0.00000557, 0.00000547, 0.00000313, -0.00000110, -0.00000762, 0.00000758, 0.00000803, 0.00000247, 0.00000009, -0.00000531, 0.00000053, 0.00000261, 0.00000352, 0.00000182, -0.00000801, 0.00000297, 0.00000090, 0.00000179, 0.00001343, -0.00000663, 0.00000393, -0.00000519, 0.00000528, -0.00000332, 0.00000765, -0.00000603, 0.00000118, -0.00000148, 0.00000912, -0.00000818, 0.00000285, -0.00000649, -0.00000206, 0.00000085, -0.00001423, 0.00000760, 0.00000069, -0.00000659, -0.00000157, -0.00000807, -0.00000960, 0.00000852, -0.00000961, 0.00000740, -0.00001058, 0.00001388, -0.00000481, 0.00000094, 0.00000010, 0.00000342, -0.00000851, -0.00000129, -0.00000539, 0.00000393, -0.00000647, 0.00000215, -0.00001013, -0.00000086, -0.00001396, -0.00001108, -0.00008687, 0.00004362, 0.00010166, 0.00007911, 0.00009234, -0.00008263, -0.00002920, -0.00001199, -0.00000719, -0.00000536, -0.00000341, -0.00000843, -0.00000368, -0.00000278, 0.00000177, -0.00000175, -0.00000090, -0.00000256, 0.00000347, -0.00000513, -0.00000538, -0.00000362, },
		{ -0.00000437, -0.00000773, -0.00000426, -0.00000829, -0.00004185, -0.00001139, 0.00006242, 0.00003524, 0.00004198, 0.00005285, -0.00002347, -0.00003613, -0.00000172, -0.00000590, -0.00000534, -0.00000038, -0.00000315, 0.00000334, -0.00000723, -0.00000038, -0.00000480, -0.00000059, 0.00000086, 0.00000120, 0.00000583, -0.00000070, -0.00000604, 0.00000010, 0.00000040, 0.00000589, -0.00001452, -0.00000010, -0.00000604, -0.00000276, -0.00000217, 0.00000841, -0.00000406, -0.00001009, -0.00000164, -0.00000105, -0.00000136, -0.00000670, 0.00000605, 0.00000484, -0.00000768, -0.00000224, 0.00000766, -0.00000574, 0.00000754, 0.00000072, -0.00000128, -0.00000753, 0.00001148, 0.00000486, 0.00000336, -0.00000209, -0.00000411, 0.00000067, 0.00000350, 0.00000303, 0.00000071, -0.00000867, 0.00000618, -0.00000281, 0.00000778, 0.00000846, -0.00000553, 0.00000265, -0.00000351, 0.00000452, -0.00000222, 0.00000684, -0.00000684, 0.00000153, 0.00000002, 0.00000820, -0.00001011, 0.00000565, -0.00001092, 0.00000398, -0.00000650, -0.00000888, 0.00000768, -0.00000079, -0.00000748, 0.00000047, -0.00001255, -0.00000334, 0.00000439, -0.00000623, 0.00000439, -0.00000795, 0.00001332, -0.00000649, 0.00000256, 0.00000022, 0.00000180, -0.00000853, -0.00000085, -0.00000470, 0.00000325, -0.00000592, 0.00000104, -0.00000942, -0.00000273, -0.00001109, -0.00002553, -0.00007898, 0.00008186, 0.00007835, 0.00009234, 0.00005957, -0.00008582, -0.00001734, -0.00001345, -0.00000511, -0.00000557, -0.00000395, -0.00000812, -0.00000329, -0.00000161, 0.00000132, -0.00000177, -0.00000148, -0.00000113, 0.00000255, -0.00000649, -0.00000455, },
		{ -0.00000402, -0.00000516, -0.00000742, -0.00000384, -0.00001269, -0.00004661, 0.00000398, 0.00006763, 0.00002981, 0.00005416, 0.00004065, -0.00003762, -0.00002868, 0.00000126, -0.00000909, -0.00000192, -0.00000294, -0.00000019, 0.00000058, -0.00000609, -0.00000083, -0.00000479, 0.00000018, 0.00000122, 0.00000104, 0.00000641, -0.00000331, -0.00000383, -0.00000090, 0.00000258, 0.00000249, -0.00001413, 0.00000095, -0.00000721, -0.00000172, -0.00000125, 0.00000930, -0.00000933, -0.00000644, -0.00000214, -0.00000030, -0.00000339, -0.00000457, 0.00000722, 0.00000297, -0.00000927, 0.00000202, 0.00000436, -0.00000382, 0.00000759, -0.00000055, -0.00000226, -0.00000577, 0.00001388, 0.00000191, 0.00000425, -0.00000417, -0.00000272, 0.00000089, 0.00000414, 0.00000260, -0.00000101, -0.00000773, 0.00000763, -0.00000495, 0.00001310, 0.00000243, -0.00000296, 0.00000041, -0.00000072, 0.00000254, 0.00000004, 0.00000459, -0.00000636, 0.00000101, 0.00000287, 0.00000471, -0.00000890, 0.00000524, -0.00001210, 0.00000729, -0.00001275, -0.00000235, 0.00000602, -0.00000192, -0.00000759, 0.00000123, -0.00001572, 0.00000258, -0.00000029, -0.00000203, 0.00000018, -0.00000365, 0.00001061, -0.00000611, 0.00000296, 0.00000094, -0.00000084, -0.00000703, -0.00000165, -0.00000303, 0.00000156, -0.00000434, -0.00000120, -0.00000759, -0.00000556, -0.00000771, -0.00004457, -0.00005396, 0.00010295, 0.00005940, 0.00010349, 0.00002398, -0.00007978, -0.00001016, -0.00001358, -0.00000396, -0.00000552, -0.00000475, -0.00000742, -0.00000304, -0.00000033, 0.00000044, -0.00000155, -0.00000201, 0.00000046, 0.00000079, -0.00000703, },
		{ -0.00000686, -0.00000396, -0.00000568, -0.00000707, -0.00000385, -0.00001826, -0.00004928, 0.00002251, 0.00006733, 0.00002871, 0.00006461, 0.00002250, -0.00004679, -0.00001876, 0.00000046, -0.00000985, -0.00000004, -0.00000449, 0.00000214, -0.00000227, -0.00000439, -0.00000191, -0.00000426, 0.00000093, 0.00000124, 0.00000116, 0.00000648, -0.00000542, -0.00000155, -0.00000212, 0.00000489, -0.00000174, -0.00001228, 0.00000069, -0.00000735, -0.00000139, 0.00000084, 0.00000788, -0.00001293, -0.00000249, -0.00000349, 0.00000066, -0.00000544, -0.00000169, 0.00000747, 0.00000076, -0.00000963, 0.00000563, 0.00000055, -0.00000052, 0.00000583, -0.00000055, -0.00000413, -0.00000217, 0.00001414, -0.00000003, 0.00000457, -0.00000576, -0.00000141, 0.00000133, 0.00000437, 0.00000226, -0.00000317, -0.00000531, 0.00000703, -0.00000467, 0.00001630, -0.00000314, 0.00000000, -0.00000193, 0.00000237, 0.00000004, 0.00000280, 0.00000152, -0.00000511, 0.00000037, 0.00000595, -0.00000021, -0.00000535, 0.00000198, -0.00000973, 0.00000667, -0.00001592, 0.00000343, 0.00000365, -0.00000303, -0.00000668, 0.00000026, -0.00001665, 0.00000686, -0.00000411, 0.00000160, -0.00000381, 0.00000097, 0.00000696, -0.00000458, 0.00000279, 0.00000150, -0.00000348, -0.00000496, -0.00000291, -0.00000115, -0.00000024, -0.00000283, -0.00000330, -0.00000619, -0.00000723, -0.00000765, -0.00006159, -0.00001745, 0.00010599, 0.00004981, 0.00010740, -0.00000918, -0.00006802, -0.00000668, -0.00001273, -0.00000344, -0.00000550, -0.00000539, -0.00000672, -0.00000251, 0.00000059, -0.00000045, -0.00000139, -0.00000215, 0.00000171, -0.00000137, },
		{ -0.00000351, -0.00000622, -0.00000437, -0.00000572, -0.00000703, -0.00000401, -0.00002541, -0.00004824, 0.00004219, 0.00006214, 0.00003311, 0.00007034, 0.00000025, -0.00004916, -0.00000925, -0.00000274, -0.00000835, -0.00000001, -0.00000449, 0.00000311, -0.00000432, -0.00000286, -0.00000314, -0.00000338, 0.00000169, 0.00000079, 0.00000178, 0.00000580, -0.00000663, 0.00000025, -0.00000306, 0.00000674, -0.00000613, -0.00000948, -0.00000056, -0.00000667, -0.00000150, 0.00000372, 0.00000408, -0.00001391, 0.00000053, -0.00000493, 0.00000147, -0.00000712, 0.00000157, 0.00000683, -0.00000157, -0.00000859, 0.00000786, -0.00000266, 0.00000305, 0.00000304, 0.00000030, -0.00000641, 0.00000285, 0.00001215, -0.00000052, 0.00000396, -0.00000658, -0.00000035, 0.00000201, 0.00000418, 0.00000191, -0.00000534, -0.00000191, 0.00000470, -0.00000182, 0.00001650, -0.00000698, 0.00000228, -0.00000346, 0.00000483, -0.00000216, 0.00000532, -0.00000168, -0.00000369, 0.00000037, 0.00000799, -0.00000495, -0.00000102, -0.00000272, -0.00000492, 0.00000243, -0.00001520, 0.00000711, 0.00000146, -0.00000429, -0.00000488, -0.00000238, -0.00001509, 0.00000880, -0.00000623, 0.00000372, -0.00000654, 0.00000477, 0.00000346, -0.00000272, 0.00000256, 0.00000132, -0.00000530, -0.00000320, -0.00000395, 0.00000036, -0.00000160, -0.00000199, -0.00000455, -0.00000601, -0.00000669, -0.00001347, -0.00006997, 0.00002199, 0.00009495, 0.00005043, 0.00010157, -0.00003597, -0.00005385, -0.00000570, -0.00001125, -0.00000337, -0.00000557, -0.00000574, -0.00000617, -0.00000160, 0.00000091, -0.00000109, -0.00000138, -0.00000180, 0.00000230, },
		{ 0.00000209, -0.00000526, -0.00000542, -0.00000504, -0.00000535, -0.00000744, -0.00000433, -0.00003423, -0.00004202, 0.00006052, 0.00005381, 0.00004299, 0.00006857, -0.00002285, -0.00004451, -0.00000264, -0.00000632, -0.00000576, -0.00000122, -0.00000320, 0.00000264, -0.00000524, -0.00000197, -0.00000414, -0.00000229, 0.00000234, -0.00000001, 0.00000290, 0.00000433, -0.00000678, 0.00000125, -0.00000335, 0.00000768, -0.00001008, -0.00000632, -0.00000235, -0.00000551, -0.00000157, 0.00000660, -0.00000147, -0.00001198, 0.00000161, -0.00000556, 0.00000160, -0.00000789, 0.00000461, 0.00000548, -0.00000377, -0.00000629, 0.00000829, -0.00000441, 0.00000576, 0.00000029, 0.00000118, -0.00000824, 0.00000832, 0.00000844, 0.00000044, 0.00000231, -0.00000650, 0.00000041, 0.00000285, 0.00000367, 0.00000137, -0.00000700, 0.00000163, 0.00000150, 0.00000294, 0.00001368, -0.00000842, 0.00000317, -0.00000354, 0.00000593, -0.00000333, 0.00000694, -0.00000444, -0.00000246, 0.00000139, 0.00000812, -0.00000808, 0.00000240, -0.00000695, 0.00000036, -0.00000390, -0.00001107, 0.00000824, -0.00000011, -0.00000560, -0.00000267, -0.00000607, -0.00001146, 0.00000848, -0.00000646, 0.00000403, -0.00000757, 0.00000721, 0.00000075, -0.00000104, 0.00000246, 0.00000031, -0.00000594, -0.00000222, -0.00000439, 0.00000131, -0.00000246, -0.00000179, -0.00000510, -0.00000667, -0.00000469, -0.00002511, -0.00006562, 0.00005586, 0.00007658, 0.00005866, 0.00008629, -0.00005435, -0.00003980, -0.00000618, -0.00000945, -0.00000369, -0.00000559, -0.00000593, -0.00000560, -0.00000048, 0.00000066, -0.00000144, -0.00000148, -0.00000105, },
		{ -0.00000011, 0.00000114, -0.00000640, -0.00000478, -0.00000568, -0.00000485, -0.00000811, -0.00000517, -0.00004415, -0.00002974, 0.00007501, 0.00004494, 0.00005689, 0.00005741, -0.00004270, -0.00003454, -0.00000007, -0.00000862, -0.00000337, -0.00000268, -0.00000137, 0.00000123, -0.00000516, -0.00000182, -0.00000471, -0.00000105, 0.00000270, -0.00000091, 0.00000435, 0.00000222, -0.00000591, 0.00000132, -0.00000279, 0.00000739, -0.00001307, -0.00000332, -0.00000427, -0.00000422, -0.00000114, 0.00000850, -0.00000754, -0.00000768, 0.00000041, -0.00000478, 0.00000057, -0.00000727, 0.00000683, 0.00000375, -0.00000562, -0.00000310, 0.00000699, -0.00000430, 0.00000682, -0.00000151, 0.00000128, -0.00000865, 0.00001300, 0.00000399, 0.00000236, -0.00000015, -0.00000560, 0.00000094, 0.00000365, 0.00000302, 0.00000049, -0.00000770, 0.00000447, -0.00000141, 0.00000840, 0.00000859, -0.00000757, 0.00000258, -0.00000204, 0.00000534, -0.00000308, 0.00000724, -0.00000640, -0.00000148, 0.00000329, 0.00000613, -0.00000888, 0.00000375, -0.00000919, 0.00000406, -0.00001012, -0.00000499, 0.00000719, -0.00000105, -0.00000669, -0.00000070, -0.00000990, -0.00000665, 0.00000650, -0.00000523, 0.00000279, -0.00000700, 0.00000822, -0.00000098, 0.00000033, 0.00000236, -0.00000125, -0.00000554, -0.00000209, -0.00000418, 0.00000176, -0.00000304, -0.00000181, -0.00000560, -0.00000703, -0.00000342, -0.00003963, -0.00004830, 0.00007821, 0.00005786, 0.00007008, 0.00006394, -0.00006403, -0.00002754, -0.00000724, -0.00000761, -0.00000434, -0.00000541, -0.00000617, -0.00000479, 0.00000054, 0.00000008, -0.00000157, -0.00000157, },
		{ -0.00000148, 0.00000073, -0.00000033, -0.00000688, -0.00000453, -0.00000596, -0.00000462, -0.00000867, -0.00000721, -0.00005398, -0.00001144, 0.00008362, 0.00003857, 0.00007181, 0.00003675, -0.00005540, -0.00002249, -0.00000100, -0.00000903, -0.00000198, -0.00000363, 0.00000027, -0.00000053, -0.00000445, -0.00000234, -0.00000471, 0.00000020, 0.00000265, -0.00000165, 0.00000587, -0.00000028, -0.00000419, 0.00000046, -0.00000134, 0.00000571, -0.00001469, -0.00000096, -0.00000590, -0.00000309, 0.00000012, 0.00000849, -0.00001257, -0.00000230, -0.00000258, -0.00000260, -0.00000168, -0.00000505, 0.00000774, 0.00000202, -0.00000694, 0.00000043, 0.00000443, -0.00000249, 0.00000598, -0.00000184, 0.00000012, -0.00000696, 0.00001572, -0.00000001, 0.00000445, -0.00000295, -0.00000413, 0.00000137, 0.00000423, 0.00000237, -0.00000081, -0.00000718, 0.00000596, -0.00000297, 0.00001309, 0.00000253, -0.00000515, 0.00000097, 0.00000064, 0.00000326, -0.00000138, 0.00000609, -0.00000744, -0.00000056, 0.00000549, 0.00000258, -0.00000746, 0.00000284, -0.00000877, 0.00000481, -0.00001426, 0.00000123, 0.00000487, -0.00000165, -0.00000720, 0.00000043, -0.00001299, -0.00000168, 0.00000373, -0.00000323, 0.00000061, -0.00000527, 0.00000810, -0.00000184, 0.00000140, 0.00000202, -0.00000285, -0.00000452, -0.00000260, -0.00000340, 0.00000178, -0.00000354, -0.00000171, -0.00000661, -0.00000605, -0.00000536, -0.00005233, -0.00002152, 0.00008688, 0.00004401, 0.00008011, 0.00003794, -0.00006598, -0.00001794, -0.00000825, -0.00000603, -0.00000512, -0.00000505, -0.00000651, -0.00000357, 0.00000115, -0.00000060, -0.00000159, },
		},
		{ { -0.00000423, -0.00000473, -0.00000410, -0.00000837, -0.00000830, -0.00001209, -0.00006021, 0.00000876, 0.00008159, 0.00003737, 0.00008677, 0.00001154, -0.00006068, -0.00001164, -0.00000642, -0.00000357, -0.00000440, -0.00000732, 0.00000637, -0.00000407, -0.00000556, -0.00000025, -0.00000353, 0.00000240, -0.00001211, 0.00000283, -0.00000414, -0.00000205, -0.00000163, 0.00000373, -0.00002337, 0.00000445, 0.00000197, -0.00000020, -0.00000095, -0.00000390, 0.00001260, -0.00000490, 0.00000527, -0.00000242, 0.00000414, -0.00000475, 0.00000427, -0.00000477, 0.00000446, 0.00000088, 0.00000374, 0.00000182, -0.00000064, -0.00000297, 0.00000367, -0.00000212, 0.00000026, 0.00000105, -0.00000393, 0.00000255, 0.00000045, 0.00000186, -0.00000600, 0.00000264, -0.00000521, 0.00000516, 0.00000393, -0.00000421, -0.00000693, -0.00000421, 0.00000393, 0.00000516, -0.00000521, 0.00000264, -0.00000600, 0.00000186, 0.00000045, 0.00000255, -0.00000393, 0.00000105, 0.00000026, -0.00000212, 0.00000367, -0.00000297, -0.00000064, 0.00000182, 0.00000374, 0.00000088, 0.00000446, -0.00000477, 0.00000427, -0.00000475, 0.00000414, -0.00000242, 0.00000527, -0.00000490, 0.00001260, -0.00000390, -0.00000095, -0.00000020, 0.00000197, 0.00000445, -0.00002337, 0.00000373, -0.00000163, -0.00000205, -0.00000414, 0.00000283, -0.00001211, 0.00000240, -0.00000353, -0.00000025, -0.00000556, -0.00000407, 0.00000637, -0.00000732, -0.00000440, -0.00000357, -0.00000642, -0.00001164, -0.00006068, 0.00001154, 0.00008677, 0.00003737, 0.00008159, 0.00000876, -0.00006021, -0.00001209, -0.00000830, -0.00000837, -0.00000410, -0.00000473, },
		{ -0.00000472, -0.00000432, -0.00000453, -0.00000454, -0.00000888, -0.00000763, -0.00001631, -0.00006410, 0.00002463, 0.00008126, 0.00004126, 0.00009276, -0.00001274, -0.00005653, -0.00000731, -0.00000722, -0.00000274, -0.00000522, -0.00000683, 0.00000744, -0.00000642, -0.00000330, -0.00000173, -0.00000230, 0.00000110, -0.00001169, 0.00000334, -0.00000493, -0.00000187, -0.00000028, 0.00000106, -0.00002418, 0.00000957, -0.00000249, 0.00000315, -0.00000508, 0.00000117, 0.00000991, -0.00000434, 0.00000529, -0.00000256, 0.00000364, -0.00000448, 0.00000425, -0.00000478, 0.00000544, -0.00000019, 0.00000467, 0.00000063, -0.00000031, -0.00000267, 0.00000422, -0.00000395, 0.00000230, -0.00000111, -0.00000188, 0.00000140, 0.00000165, -0.00000036, -0.00000435, 0.00000172, -0.00000469, 0.00000646, 0.00000225, -0.00000473, -0.00000683, -0.00000368, 0.00000577, 0.00000300, -0.00000459, 0.00000247, -0.00000671, 0.00000354, -0.00000041, 0.00000320, -0.00000562, 0.00000313, -0.00000199, 0.00000032, 0.00000230, -0.00000280, -0.00000113, 0.00000326, 0.00000242, 0.00000250, 0.00000268, -0.00000396, 0.00000359, -0.00000434, 0.00000406, -0.00000161, 0.00000414, -0.00000380, 0.00001333, -0.00000757, 0.00000255, -0.00000310, 0.00000635, -0.00000180, -0.00002052, 0.00000490, -0.00000239, -0.00000256, -0.00000286, 0.00000162, -0.00001174, 0.00000294, -0.00000395, 0.00000040, -0.00000719, -0.00000200, 0.00000529, -0.00000774, -0.00000346, -0.00000484, -0.00000487, -0.00001859, -0.00005907, 0.00003344, 0.00007678, 0.00003712, 0.00007881, -0.00000498, -0.00005538, -0.00000861, -0.00000923, -0.00000748, -0.00000401, },
		{ -0.00000425, -0.00000452, -0.00000455, -0.00000418, -0.00000528, -0.00000901, -0.00000732, -0.00002130, -0.00006655, 0.00004217, 0.00007785, 0.00004948, 0.00009255, -0.00003663, -0.00004775, -0.00000584, -0.00000685, -0.00000266, -0.00000572, -0.00000631, 0.00000842, -0.00000891, -0.00000052, -0.00000400, -0.00000023, -0.00000109, -0.00001037, 0.00000299, -0.00000506, -0.00000208, 0.00000159, -0.00000313, -0.00002249, 0.00001263, -0.00000595, 0.00000598, -0.00000891, 0.00000679, 0.00000583, -0.00000233, 0.00000419, -0.00000198, 0.00000259, -0.00000358, 0.00000358, -0.00000401, 0.00000555, -0.00000056, 0.00000502, -0.00000014, -0.00000024, -0.00000185, 0.00000381, -0.00000480, 0.00000364, -0.00000277, 0.00000001, 0.00000022, 0.00000276, -0.00000263, -0.00000220, 0.00000004, -0.00000317, 0.00000684, 0.00000086, -0.00000528, -0.00000658, -0.00000302, 0.00000750, 0.00000030, -0.00000299, 0.00000116, -0.00000623, 0.00000436, -0.00000063, 0.00000301, -0.00000645, 0.00000460, -0.00000386, 0.00000286, 0.00000046, -0.00000233, -0.00000158, 0.00000466, 0.00000101, 0.00000434, 0.00000036, -0.00000249, 0.00000232, -0.00000333, 0.00000355, -0.00000039, 0.00000221, -0.00000125, 0.00001194, -0.00000931, 0.00000473, -0.00000477, 0.00000964, -0.00000808, -0.00001639, 0.00000485, -0.00000270, -0.00000317, -0.00000135, -0.00000003, -0.00001085, 0.00000296, -0.00000380, 0.00000037, -0.00000823, -0.00000018, 0.00000414, -0.00000793, -0.00000275, -0.00000602, -0.00000338, -0.00002716, -0.00005159, 0.00005090, 0.00006503, 0.00003945, 0.00007330, -0.00001639, -0.00004994, -0.00000591, -0.00001026, -0.00000630, },
		{ -0.00000503, -0.00000470, -0.00000419, -0.00000483, -0.00000383, -0.00000616, -0.00000890, -0.00000733, -0.00002718, -0.00006692, 0.00006053, 0.00007192, 0.00006199, 0.00008437, -0.00005708, -0.00003633, -0.00000670, -0.00000541, -0.00000329, -0.00000593, -0.00000565, 0.00000907, -0.00001120, 0.00000242, -0.00000675, 0.00000248, -0.00000402, -0.00000822, 0.00000184, -0.00000455, -0.00000259, 0.00000371, -0.00000850, -0.00001830, 0.00001310, -0.00000759, 0.00000741, -0.00001148, 0.00001192, 0.00000120, 0.00000062, 0.00000220, -0.00000082, 0.00000114, -0.00000223, 0.00000247, -0.00000266, 0.00000492, -0.00000026, 0.00000477, -0.00000046, -0.00000039, -0.00000064, 0.00000251, -0.00000453, 0.00000398, -0.00000354, 0.00000126, -0.00000056, 0.00000335, -0.00000441, -0.00000013, -0.00000191, -0.00000099, 0.00000640, -0.00000022, -0.00000580, -0.00000627, -0.00000208, 0.00000876, -0.00000250, -0.00000080, -0.00000104, -0.00000461, 0.00000425, -0.00000013, 0.00000186, -0.00000617, 0.00000506, -0.00000484, 0.00000492, -0.00000145, -0.00000183, -0.00000174, 0.00000573, -0.00000013, 0.00000600, -0.00000211, -0.00000067, 0.00000070, -0.00000196, 0.00000285, 0.00000086, 0.00000003, 0.00000223, 0.00000874, -0.00000906, 0.00000534, -0.00000480, 0.00001110, -0.00001340, -0.00001179, 0.00000403, -0.00000277, -0.00000366, 0.00000007, -0.00000181, -0.00000972, 0.00000275, -0.00000337, -0.00000013, -0.00000882, 0.00000146, 0.00000285, -0.00000772, -0.00000256, -0.00000660, -0.00000290, -0.00003586, -0.00003912, 0.00006283, 0.00005342, 0.00004320, 0.00006565, -0.00002550, -0.00004413, -0.00000403, -0.00001117, },
		{ -0.00001176, -0.00000384, -0.00000520, -0.00000387, -0.00000503, -0.00000366, -0.00000693, -0.00000877, -0.00000753, -0.00003417, -0.00006436, 0.00007849, 0.00006468, 0.00007778, 0.00006744, -0.00007134, -0.00002467, -0.00000875, -0.00000347, -0.00000426, -0.00000609, -0.00000456, 0.00000898, -0.00001280, 0.00000497, -0.00000946, 0.00000540, -0.00000734, -0.00000555, 0.00000014, -0.00000363, -0.00000312, 0.00000553, -0.00001430, -0.00001215, 0.00001099, -0.00000703, 0.00000686, -0.00001203, 0.00001559, -0.00000304, 0.00000382, -0.00000024, 0.00000065, -0.00000046, -0.00000070, 0.00000121, -0.00000104, 0.00000382, 0.00000051, 0.00000406, -0.00000042, -0.00000062, 0.00000067, 0.00000061, -0.00000332, 0.00000331, -0.00000328, 0.00000160, -0.00000063, 0.00000311, -0.00000530, 0.00000132, -0.00000356, 0.00000137, 0.00000540, -0.00000109, -0.00000620, -0.00000601, -0.00000070, 0.00000922, -0.00000485, 0.00000144, -0.00000366, -0.00000217, 0.00000335, 0.00000093, -0.00000009, -0.00000481, 0.00000436, -0.00000461, 0.00000608, -0.00000303, -0.00000152, -0.00000143, 0.00000621, -0.00000067, 0.00000707, -0.00000431, 0.00000112, -0.00000094, -0.00000050, 0.00000224, 0.00000173, -0.00000178, 0.00000583, 0.00000440, -0.00000720, 0.00000457, -0.00000323, 0.00001044, -0.00001709, -0.00000746, 0.00000286, -0.00000280, -0.00000385, 0.00000120, -0.00000349, -0.00000856, 0.00000255, -0.00000291, -0.00000088, -0.00000908, 0.00000298, 0.00000140, -0.00000709, -0.00000303, -0.00000624, -0.00000422, -0.00004300, -0.00002326, 0.00006903, 0.00004326, 0.00004733, 0.00005649, -0.00003241, -0.00003816, -0.00000299, },
		{ -0.00000275, -0.00001188, -0.00000293, -0.00000559, -0.00000370, -0.00000499, -0.00000386, -0.00000734, -0.00000894, -0.00000771, -0.00004250, -0.00005797, 0.00009456, 0.00005792, 0.00009479, 0.00004234, -0.00007752, -0.00001506, -0.00001062, -0.00000184, -0.00000504, -0.00000660, -0.00000271, 0.00000775, -0.00001317, 0.00000649, -0.00001149, 0.00000791, -0.00001051, -0.00000283, -0.00000169, -0.00000265, -0.00000328, 0.00000641, -0.00001954, -0.00000503, 0.00000688, -0.00000444, 0.00000428, -0.00001022, 0.00001714, -0.00000608, 0.00000656, -0.00000264, 0.00000212, -0.00000194, 0.00000073, 0.00000009, 0.00000050, 0.00000258, 0.00000149, 0.00000309, -0.00000018, -0.00000073, 0.00000175, -0.00000143, -0.00000155, 0.00000193, -0.00000212, 0.00000104, 0.00000005, 0.00000196, -0.00000508, 0.00000179, -0.00000444, 0.00000344, 0.00000417, -0.00000189, -0.00000639, -0.00000588, 0.00000120, 0.00000861, -0.00000629, 0.00000313, -0.00000609, 0.00000060, 0.00000203, 0.00000219, -0.00000247, -0.00000265, 0.00000266, -0.00000316, 0.00000611, -0.00000402, -0.00000155, -0.00000055, 0.00000596, -0.00000041, 0.00000724, -0.00000584, 0.00000254, -0.00000230, 0.00000079, 0.00000195, 0.00000187, -0.00000263, 0.00000873, -0.00000021, -0.00000440, 0.00000295, -0.00000056, 0.00000786, -0.00001886, -0.00000387, 0.00000164, -0.00000288, -0.00000369, 0.00000194, -0.00000497, -0.00000744, 0.00000245, -0.00000253, -0.00000177, -0.00000905, 0.00000433, -0.00000009, -0.00000621, -0.00000404, -0.00000500, -0.00000769, -0.00004716, -0.00000596, 0.00007009, 0.00003532, 0.00005102, 0.00004645, -0.00003729, -0.00003223, },
		{ -0.00002654, -0.00000323, -0.00001146, -0.00000246, -0.00000572, -0.00000379, -0.00000462, -0.00000459, -0.00000717, -0.00000967, -0.00000770, -0.00005231, -0.00004695, 0.00010720, 0.00005373, 0.00011004, 0.00001114, -0.00007508, -0.00000916, -0.00001109, -0.00000130, -0.00000508, -0.00000777, 0.00000015, 0.00000506, -0.00001189, 0.00000645, -0.00001220, 0.00000941, -0.00001298, -0.00000056, -0.00000320, -0.00000199, -0.00000271, 0.00000576, -0.00002319, 0.00000179, 0.00000184, -0.00000052, 0.00000011, -0.00000623, 0.00001635, -0.00000741, 0.00000828, -0.00000456, 0.00000329, -0.00000308, 0.00000187, -0.00000068, 0.00000169, 0.00000151, 0.00000239, 0.00000211, 0.00000007, -0.00000054, 0.00000224, -0.00000310, 0.00000023, 0.00000031, -0.00000049, -0.00000016, 0.00000129, 0.00000009, -0.00000383, 0.00000117, -0.00000428, 0.00000486, 0.00000296, -0.00000270, -0.00000633, -0.00000586, 0.00000354, 0.00000690, -0.00000653, 0.00000380, -0.00000777, 0.00000315, 0.00000072, 0.00000323, -0.00000476, -0.00000022, 0.00000041, -0.00000073, 0.00000502, -0.00000434, -0.00000191, 0.00000082, 0.00000505, 0.00000067, 0.00000639, -0.00000648, 0.00000333, -0.00000315, 0.00000176, 0.00000206, 0.00000117, -0.00000216, 0.00001023, -0.00000419, -0.00000145, 0.00000121, 0.00000246, 0.00000395, -0.00001885, -0.00000120, 0.00000051, -0.00000299, -0.00000326, 0.00000233, -0.00000630, -0.00000632, 0.00000238, -0.00000224, -0.00000280, -0.00000865, 0.00000538, -0.00000140, -0.00000539, -0.00000517, -0.00000333, -0.00001307, -0.00004748, 0.00001093, 0.00006704, 0.00002988, 0.00005370, 0.00003606, -0.00004034, },
		{ -0.00004177, -0.00002125, -0.00000426, -0.00001050, -0.00000247, -0.00000551, -0.00000421, -0.00000387, -0.00000588, -0.00000632, -0.00001111, -0.00000748, -0.00006348, -0.00003082, 0.00011515, 0.00005416, 0.00012005, -0.00002276, -0.00006500, -0.00000761, -0.00000951, -0.00000227, -0.00000412, -0.00000970, 0.00000402, 0.00000090, -0.00000880, 0.00000456, -0.00001120, 0.00000942, -0.00001430, 0.00000086, -0.00000406, -0.00000190, -0.00000124, 0.00000323, -0.00002440, 0.00000708, -0.00000289, 0.00000367, -0.00000476, -0.00000072, 0.00001347, -0.00000691, 0.00000870, -0.00000572, 0.00000399, -0.00000377, 0.00000262, -0.00000099, 0.00000235, 0.00000083, 0.00000298, 0.00000131, 0.00000023, 0.00000003, 0.00000194, -0.00000399, 0.00000152, -0.00000099, 0.00000108, -0.00000152, 0.00000268, -0.00000210, -0.00000190, -0.00000034, -0.00000309, 0.00000546, 0.00000195, -0.00000357, -0.00000607, -0.00000583, 0.00000609, 0.00000424, -0.00000553, 0.00000321, -0.00000828, 0.00000503, -0.00000020, 0.00000365, -0.00000644, 0.00000191, -0.00000180, 0.00000216, 0.00000308, -0.00000411, -0.00000247, 0.00000248, 0.00000367, 0.00000237, 0.00000460, -0.00000619, 0.00000341, -0.00000341, 0.00000236, 0.00000249, -0.00000025, -0.00000032, 0.00000994, -0.00000690, 0.00000097, 0.00000002, 0.00000500, -0.00000049, -0.00001748, 0.00000059, -0.00000051, -0.00000301, -0.00000272, 0.00000249, -0.00000756, -0.00000505, 0.00000220, -0.00000193, -0.00000407, -0.00000777, 0.00000593, -0.00000226, -0.00000504, -0.00000587, -0.00000200, -0.00001955, -0.00004380, 0.00002590, 0.00006108, 0.00002687, 0.00005499, 0.00002582, },
		{ 0.00001614, -0.00004186, -0.00001651, -0.00000566, -0.00000908, -0.00000296, -0.00000495, -0.00000491, -0.00000283, -0.00000762, -0.00000489, -0.00001319, -0.00000731, -0.00007546, -0.00000968, 0.00011777, 0.00006059, 0.00012144, -0.00005511, -0.00004963, -0.00000990, -0.00000596, -0.00000468, -0.00000228, -0.00001214, 0.00000852, -0.00000438, -0.00000418, 0.00000090, -0.00000842, 0.00000779, -0.00001428, 0.00000126, -0.00000412, -0.00000239, 0.00000100, -0.00000113, -0.00002276, 0.00000989, -0.00000612, 0.00000697, -0.00000923, 0.00000533, 0.00000916, -0.00000486, 0.00000782, -0.00000602, 0.00000415, -0.00000401, 0.00000302, -0.00000089, 0.00000249, 0.00000061, 0.00000318, 0.00000077, 0.00000028, 0.00000091, 0.00000084, -0.00000390, 0.00000195, -0.00000152, 0.00000205, -0.00000254, 0.00000374, -0.00000406, 0.00000017, -0.00000231, -0.00000112, 0.00000528, 0.00000115, -0.00000441, -0.00000572, -0.00000559, 0.00000849, 0.00000103, -0.00000354, 0.00000144, -0.00000746, 0.00000596, -0.00000051, 0.00000321, -0.00000714, 0.00000323, -0.00000339, 0.00000493, 0.00000071, -0.00000354, -0.00000300, 0.00000413, 0.00000219, 0.00000434, 0.00000216, -0.00000512, 0.00000286, -0.00000315, 0.00000271, 0.00000299, -0.00000204, 0.00000256, 0.00000786, -0.00000800, 0.00000246, -0.00000019, 0.00000638, -0.00000464, -0.00001533, 0.00000169, -0.00000147, -0.00000285, -0.00000222, 0.00000253, -0.00000884, -0.00000352, 0.00000178, -0.00000150, -0.00000562, -0.00000631, 0.00000587, -0.00000251, -0.00000546, -0.00000566, -0.00000184, -0.00002595, -0.00003667, 0.00003803, 0.00005335, 0.00002604, 0.00005468, },
		{ 0.00005269, 0.00000738, -0.00004087, -0.00001240, -0.00000722, -0.00000737, -0.00000381, -0.00000415, -0.00000574, -0.00000173, -0.00000952, -0.00000322, -0.00001556, -0.00000775, -0.00008720, 0.00001564, 0.00011535, 0.00007335, 0.00011169, -0.00008152, -0.00003225, -0.00001452, -0.00000127, -0.00000790, -0.00000008, -0.00001450, 0.00001295, -0.00001003, 0.00000133, -0.00000399, -0.00000421, 0.00000472, -0.00001304, 0.00000075, -0.00000352, -0.00000324, 0.00000356, -0.00000678, -0.00001848, 0.00000983, -0.00000708, 0.00000846, -0.00001224, 0.00001082, 0.00000433, -0.00000182, 0.00000588, -0.00000555, 0.00000387, -0.00000389, 0.00000319, -0.00000055, 0.00000224, 0.00000076, 0.00000303, 0.00000048, 0.00000032, 0.00000190, -0.00000084, -0.00000288, 0.00000143, -0.00000103, 0.00000205, -0.00000278, 0.00000402, -0.00000528, 0.00000180, -0.00000417, 0.00000121, 0.00000452, 0.00000049, -0.00000512, -0.00000546, -0.00000490, 0.00001031, -0.00000218, -0.00000105, -0.00000119, -0.00000545, 0.00000591, -0.00000019, 0.00000191, -0.00000673, 0.00000344, -0.00000387, 0.00000699, -0.00000162, -0.00000292, -0.00000324, 0.00000544, 0.00000097, 0.00000615, -0.00000053, -0.00000357, 0.00000190, -0.00000252, 0.00000298, 0.00000323, -0.00000365, 0.00000590, 0.00000441, -0.00000758, 0.00000299, 0.00000065, 0.00000629, -0.00000790, -0.00001291, 0.00000230, -0.00000240, -0.00000246, -0.00000184, 0.00000250, -0.00001014, -0.00000171, 0.00000107, -0.00000094, -0.00000742, -0.00000433, 0.00000523, -0.00000216, -0.00000671, -0.00000428, -0.00000349, -0.00003101, -0.00002712, 0.00004693, 0.00004480, 0.00002703, },
		{ 0.00002944, 0.00004904, -0.00000016, -0.00003909, -0.00000899, -0.00000871, -0.00000560, -0.00000481, -0.00000332, -0.00000648, -0.00000090, -0.00001113, -0.00000185, -0.00001763, -0.00000969, -0.00009717, 0.00004353, 0.00010924, 0.00009126, 0.00008982, -0.00009836, -0.00001641, -0.00001940, 0.00000324, -0.00001093, 0.00000163, -0.00001592, 0.00001638, -0.00001506, 0.00000674, -0.00000926, 0.00000073, 0.00000072, -0.00001097, -0.00000031, -0.00000263, -0.00000398, 0.00000571, -0.00001275, -0.00001234, 0.00000718, -0.00000559, 0.00000766, -0.00001307, 0.00001482, -0.00000009, 0.00000153, 0.00000331, -0.00000452, 0.00000326, -0.00000355, 0.00000331, -0.00000021, 0.00000183, 0.00000113, 0.00000265, 0.00000032, 0.00000051, 0.00000267, -0.00000272, -0.00000127, 0.00000015, 0.00000042, 0.00000097, -0.00000205, 0.00000329, -0.00000542, 0.00000252, -0.00000539, 0.00000340, 0.00000349, -0.00000017, -0.00000554, -0.00000543, -0.00000356, 0.00001117, -0.00000484, 0.00000134, -0.00000412, -0.00000263, 0.00000507, 0.00000057, -0.00000001, -0.00000535, 0.00000254, -0.00000304, 0.00000794, -0.00000352, -0.00000251, -0.00000300, 0.00000616, 0.00000034, 0.00000737, -0.00000301, -0.00000190, 0.00000080, -0.00000176, 0.00000337, 0.00000290, -0.00000453, 0.00000892, 0.00000026, -0.00000605, 0.00000285, 0.00000226, 0.00000479, -0.00000998, -0.00001060, 0.00000256, -0.00000325, -0.00000191, -0.00000153, 0.00000231, -0.00001131, 0.00000023, 0.00000017, -0.00000038, -0.00000927, -0.00000200, 0.00000421, -0.00000147, -0.00000852, -0.00000192, -0.00000715, -0.00003370, -0.00001637, 0.00005266, 0.00003617, },
		{ 0.00002809, 0.00003276, 0.00004391, -0.00000635, -0.00003678, -0.00000632, -0.00000992, -0.00000401, -0.00000570, -0.00000270, -0.00000684, -0.00000072, -0.00001194, -0.00000141, -0.00001870, -0.00001415, -0.00010351, 0.00007181, 0.00010172, 0.00011160, 0.00005685, -0.00010345, -0.00000509, -0.00002246, 0.00000615, -0.00001267, 0.00000197, -0.00001559, 0.00001785, -0.00001843, 0.00001100, -0.00001390, 0.00000552, -0.00000349, -0.00000866, -0.00000140, -0.00000192, -0.00000410, 0.00000666, -0.00001786, -0.00000560, 0.00000286, -0.00000210, 0.00000466, -0.00001149, 0.00001674, -0.00000339, 0.00000451, 0.00000056, -0.00000317, 0.00000248, -0.00000311, 0.00000351, -0.00000004, 0.00000149, 0.00000148, 0.00000221, 0.00000017, 0.00000103, 0.00000289, -0.00000430, 0.00000045, -0.00000143, 0.00000245, -0.00000096, -0.00000044, 0.00000158, -0.00000438, 0.00000208, -0.00000558, 0.00000503, 0.00000248, -0.00000097, -0.00000560, -0.00000570, -0.00000150, 0.00001081, -0.00000647, 0.00000304, -0.00000671, 0.00000048, 0.00000380, 0.00000142, -0.00000214, -0.00000341, 0.00000082, -0.00000099, 0.00000760, -0.00000473, -0.00000244, -0.00000220, 0.00000618, 0.00000050, 0.00000771, -0.00000492, -0.00000045, -0.00000017, -0.00000103, 0.00000399, 0.00000181, -0.00000422, 0.00001089, -0.00000378, -0.00000399, 0.00000253, 0.00000405, 0.00000229, -0.00001090, -0.00000854, 0.00000251, -0.00000387, -0.00000137, -0.00000115, 0.00000177, -0.00001210, 0.00000204, -0.00000069, -0.00000009, -0.00001087, 0.00000032, 0.00000319, -0.00000094, -0.00001031, 0.00000083, -0.00001245, -0.00003354, -0.00000552, 0.00005547, },
		{ 0.00005569, 0.00002110, 0.00003643, 0.00003764, -0.00001116, -0.00003410, -0.00000441, -0.00001067, -0.00000285, -0.00000624, -0.00000253, -0.00000655, -0.00000150, -0.00001152, -0.00000247, -0.00001812, -0.00002208, -0.00010435, 0.00009806, 0.00009561, 0.00013036, 0.00001584, -0.00009663, -0.00000013, -0.00002219, 0.00000636, -0.00001228, 0.00000031, -0.00001291, 0.00001669, -0.00001930, 0.00001317, -0.00001698, 0.00000931, -0.00000720, -0.00000667, -0.00000206, -0.00000180, -0.00000318, 0.00000574, -0.00002097, 0.00000031, -0.00000183, 0.00000239, 0.00000015, -0.00000780, 0.00001648, -0.00000514, 0.00000666, -0.00000199, -0.00000171, 0.00000161, -0.00000263, 0.00000384, -0.00000018, 0.00000139, 0.00000166, 0.00000187, -0.00000007, 0.00000196, 0.00000229, -0.00000514, 0.00000173, -0.00000275, 0.00000448, -0.00000328, 0.00000168, -0.00000080, -0.00000240, 0.00000052, -0.00000459, 0.00000581, 0.00000170, -0.00000196, -0.00000528, -0.00000623, 0.00000117, 0.00000921, -0.00000682, 0.00000362, -0.00000842, 0.00000333, 0.00000249, 0.00000199, -0.00000398, -0.00000145, -0.00000119, 0.00000191, 0.00000608, -0.00000519, -0.00000271, -0.00000092, 0.00000556, 0.00000143, 0.00000705, -0.00000602, 0.00000056, -0.00000083, -0.00000044, 0.00000482, -0.00000004, -0.00000254, 0.00001128, -0.00000702, -0.00000199, 0.00000256, 0.00000537, -0.00000056, -0.00001097, -0.00000672, 0.00000209, -0.00000410, -0.00000106, -0.00000050, 0.00000067, -0.00001219, 0.00000337, -0.00000118, -0.00000037, -0.00001181, 0.00000223, 0.00000263, -0.00000113, -0.00001134, 0.00000302, -0.00001845, -0.00003066, 0.00000466, },
		{ 0.00001384, 0.00005353, 0.00001571, 0.00003979, 0.00003072, -0.00001475, -0.00003118, -0.00000325, -0.00001082, -0.00000230, -0.00000622, -0.00000298, -0.00000548, -0.00000340, -0.00000964, -0.00000538, -0.00001552, -0.00003404, -0.00009819, 0.00012008, 0.00009371, 0.00014293, -0.00002837, -0.00007982, -0.00000178, -0.00001805, 0.00000350, -0.00000947, -0.00000350, -0.00000780, 0.00001271, -0.00001731, 0.00001275, -0.00001787, 0.00001145, -0.00000987, -0.00000540, -0.00000199, -0.00000247, -0.00000111, 0.00000266, -0.00002128, 0.00000412, -0.00000550, 0.00000662, -0.00000481, -0.00000278, 0.00001438, -0.00000533, 0.00000773, -0.00000408, -0.00000027, 0.00000070, -0.00000205, 0.00000422, -0.00000059, 0.00000159, 0.00000159, 0.00000168, -0.00000039, 0.00000327, 0.00000082, -0.00000493, 0.00000211, -0.00000323, 0.00000589, -0.00000536, 0.00000378, -0.00000332, 0.00000006, -0.00000181, -0.00000258, 0.00000572, 0.00000121, -0.00000308, -0.00000470, -0.00000683, 0.00000418, 0.00000658, -0.00000592, 0.00000291, -0.00000887, 0.00000551, 0.00000145, 0.00000200, -0.00000512, 0.00000000, -0.00000287, 0.00000507, 0.00000369, -0.00000502, -0.00000319, 0.00000064, 0.00000452, 0.00000292, 0.00000549, -0.00000626, 0.00000103, -0.00000112, 0.00000003, 0.00000570, -0.00000237, 0.00000037, 0.00000991, -0.00000899, -0.00000045, 0.00000330, 0.00000566, -0.00000310, -0.00001060, -0.00000503, 0.00000126, -0.00000380, -0.00000114, 0.00000054, -0.00000115, -0.00001134, 0.00000394, -0.00000106, -0.00000152, -0.00001177, 0.00000337, 0.00000285, -0.00000253, -0.00001089, 0.00000363, -0.00002390, -0.00002576, },
		{ -0.00001988, 0.00002203, 0.00004915, 0.00001240, 0.00004217, 0.00002373, -0.00001737, -0.00002805, -0.00000281, -0.00001035, -0.00000243, -0.00000560, -0.00000406, -0.00000368, -0.00000633, -0.00000639, -0.00001006, -0.00001105, -0.00004995, -0.00008427, 0.00013647, 0.00009798, 0.00014493, -0.00007003, -0.00005669, -0.00000872, -0.00001067, -0.00000198, -0.00000467, -0.00000897, -0.00000076, 0.00000632, -0.00001268, 0.00000974, -0.00001642, 0.00001172, -0.00001131, -0.00000493, -0.00000121, -0.00000378, 0.00000183, -0.00000234, -0.00001859, 0.00000506, -0.00000703, 0.00000939, -0.00000905, 0.00000253, 0.00001117, -0.00000426, 0.00000776, -0.00000563, 0.00000108, -0.00000029, -0.00000127, 0.00000449, -0.00000114, 0.00000199, 0.00000131, 0.00000159, -0.00000067, 0.00000474, -0.00000138, -0.00000366, 0.00000141, -0.00000253, 0.00000613, -0.00000661, 0.00000526, -0.00000538, 0.00000236, -0.00000434, 0.00000007, 0.00000491, 0.00000093, -0.00000420, -0.00000406, -0.00000722, 0.00000712, 0.00000337, -0.00000407, 0.00000105, -0.00000798, 0.00000676, 0.00000085, 0.00000135, -0.00000534, 0.00000054, -0.00000368, 0.00000788, 0.00000091, -0.00000445, -0.00000368, 0.00000220, 0.00000339, 0.00000464, 0.00000330, -0.00000579, 0.00000105, -0.00000110, 0.00000049, 0.00000634, -0.00000473, 0.00000407, 0.00000696, -0.00000959, 0.00000049, 0.00000479, 0.00000464, -0.00000485, -0.00001016, -0.00000337, 0.00000003, -0.00000295, -0.00000166, 0.00000199, -0.00000369, -0.00000949, 0.00000362, -0.00000021, -0.00000361, -0.00001061, 0.00000363, 0.00000398, -0.00000532, -0.00000856, 0.00000188, -0.00002751, },
		{ -0.00002834, -0.00001408, 0.00002943, 0.00004270, 0.00001150, 0.00004297, 0.00001727, -0.00001935, -0.00002475, -0.00000301, -0.00000932, -0.00000315, -0.00000447, -0.00000561, -0.00000143, -0.00000989, -0.00000227, -0.00001599, -0.00000545, -0.00006891, -0.00006285, 0.00014698, 0.00010896, 0.00013322, -0.00010343, -0.00003190, -0.00001840, -0.00000163, -0.00000885, 0.00000108, -0.00001507, 0.00000712, -0.00000151, -0.00000618, 0.00000470, -0.00001298, 0.00001032, -0.00001166, -0.00000502, -0.00000004, -0.00000530, 0.00000494, -0.00000843, -0.00001342, 0.00000306, -0.00000587, 0.00000986, -0.00001156, 0.00000710, 0.00000768, -0.00000248, 0.00000696, -0.00000664, 0.00000235, -0.00000141, -0.00000017, 0.00000443, -0.00000161, 0.00000244, 0.00000094, 0.00000149, -0.00000070, 0.00000604, -0.00000394, -0.00000158, -0.00000028, -0.00000058, 0.00000495, -0.00000662, 0.00000567, -0.00000644, 0.00000390, -0.00000644, 0.00000283, 0.00000369, 0.00000069, -0.00000510, -0.00000359, -0.00000712, 0.00000951, 0.00000013, -0.00000180, -0.00000158, -0.00000592, 0.00000711, 0.00000063, 0.00000013, -0.00000465, 0.00000003, -0.00000323, 0.00000978, -0.00000180, -0.00000377, -0.00000394, 0.00000347, 0.00000251, 0.00000618, 0.00000086, -0.00000487, 0.00000077, -0.00000091, 0.00000110, 0.00000641, -0.00000656, 0.00000786, 0.00000292, -0.00000901, 0.00000102, 0.00000678, 0.00000241, -0.00000558, -0.00000987, -0.00000175, -0.00000144, -0.00000170, -0.00000248, 0.00000364, -0.00000670, -0.00000681, 0.00000254, 0.00000122, -0.00000649, -0.00000843, 0.00000312, 0.00000575, -0.00000925, -0.00000446, -0.00000252, },
		{ -0.00000920, -0.00002604, -0.00000914, 0.00003617, 0.00003450, 0.00001312, 0.00004178, 0.00001181, -0.00002098, -0.00002131, -0.00000371, -0.00000792, -0.00000425, -0.00000309, -0.00000727, 0.00000077, -0.00001342, 0.00000190, -0.00002217, -0.00000003, -0.00008920, -0.00003522, 0.00015270, 0.00012528, 0.00010676, -0.00012408, -0.00001017, -0.00002775, 0.00000694, -0.00001538, 0.00000632, -0.00002041, 0.00001446, -0.00000942, 0.00000097, -0.00000137, -0.00000832, 0.00000778, -0.00001139, -0.00000520, 0.00000097, -0.00000634, 0.00000728, -0.00001438, -0.00000693, -0.00000118, -0.00000221, 0.00000779, -0.00001178, 0.00001014, 0.00000468, -0.00000054, 0.00000564, -0.00000720, 0.00000353, -0.00000268, 0.00000134, 0.00000387, -0.00000176, 0.00000271, 0.00000068, 0.00000122, -0.00000025, 0.00000675, -0.00000631, 0.00000081, -0.00000255, 0.00000232, 0.00000240, -0.00000527, 0.00000480, -0.00000619, 0.00000422, -0.00000757, 0.00000516, 0.00000243, 0.00000028, -0.00000559, -0.00000349, -0.00000629, 0.00001095, -0.00000257, 0.00000030, -0.00000438, -0.00000311, 0.00000672, 0.00000063, -0.00000135, -0.00000335, -0.00000138, -0.00000145, 0.00001042, -0.00000403, -0.00000320, -0.00000382, 0.00000427, 0.00000215, 0.00000717, -0.00000142, -0.00000379, 0.00000041, -0.00000071, 0.00000200, 0.00000567, -0.00000733, 0.00001100, -0.00000150, -0.00000770, 0.00000152, 0.00000873, -0.00000060, -0.00000541, -0.00000975, -0.00000030, -0.00000290, -0.00000033, -0.00000331, 0.00000513, -0.00000976, -0.00000372, 0.00000104, 0.00000285, -0.00000975, -0.00000564, 0.00000229, 0.00000760, -0.00001363, 0.00000077, },
		{ 0.00000595, -0.00001707, -0.00002099, -0.00000533, 0.00004213, 0.00002508, 0.00001699, 0.00003851, 0.00000767, -0.00002247, -0.00001780, -0.00000467, -0.00000641, -0.00000540, -0.00000183, -0.00000858, 0.00000234, -0.00001611, 0.00000509, -0.00002738, 0.00000354, -0.00010850, -0.00000354, 0.00015587, 0.00014369, 0.00006712, -0.00012959, 0.00000475, -0.00003392, 0.00001292, -0.00001986, 0.00000960, -0.00002363, 0.00001982, -0.00001599, 0.00000744, -0.00000731, -0.00000338, 0.00000482, -0.00001106, -0.00000493, 0.00000124, -0.00000626, 0.00000789, -0.00001880, -0.00000066, -0.00000635, 0.00000310, 0.00000362, -0.00000973, 0.00001133, 0.00000268, 0.00000108, 0.00000408, -0.00000737, 0.00000457, -0.00000402, 0.00000322, 0.00000273, -0.00000143, 0.00000263, 0.00000067, 0.00000064, 0.00000085, 0.00000652, -0.00000796, 0.00000290, -0.00000476, 0.00000560, -0.00000107, -0.00000278, 0.00000276, -0.00000465, 0.00000317, -0.00000739, 0.00000666, 0.00000142, -0.00000044, -0.00000557, -0.00000384, -0.00000464, 0.00001119, -0.00000430, 0.00000168, -0.00000676, -0.00000003, 0.00000589, 0.00000056, -0.00000268, -0.00000189, -0.00000326, 0.00000143, 0.00000969, -0.00000556, -0.00000288, -0.00000329, 0.00000453, 0.00000242, 0.00000738, -0.00000324, -0.00000282, 0.00000017, -0.00000059, 0.00000321, 0.00000402, -0.00000671, 0.00001284, -0.00000558, -0.00000614, 0.00000245, 0.00000997, -0.00000373, -0.00000470, -0.00000963, 0.00000076, -0.00000398, 0.00000075, -0.00000375, 0.00000600, -0.00001230, -0.00000077, -0.00000035, 0.00000414, -0.00001279, -0.00000278, 0.00000172, 0.00000872, -0.00001745, },
		{ -0.00001960, 0.00000972, -0.00002459, -0.00001424, -0.00000237, 0.00004690, 0.00001530, 0.00002250, 0.00003338, 0.00000487, -0.00002387, -0.00001438, -0.00000567, -0.00000511, -0.00000625, -0.00000108, -0.00000912, 0.00000271, -0.00001724, 0.00000632, -0.00003038, 0.00000355, -0.00012430, 0.00002964, 0.00015935, 0.00015945, 0.00001841, -0.00012019, 0.00001085, -0.00003500, 0.00001482, -0.00002105, 0.00000988, -0.00002372, 0.00002213, -0.00002004, 0.00001210, -0.00001205, 0.00000095, 0.00000212, -0.00001113, -0.00000379, 0.00000038, -0.00000463, 0.00000609, -0.00002050, 0.00000379, -0.00001085, 0.00000870, -0.00000168, -0.00000597, 0.00001081, 0.00000183, 0.00000212, 0.00000246, -0.00000715, 0.00000534, -0.00000526, 0.00000532, 0.00000106, -0.00000059, 0.00000214, 0.00000096, -0.00000024, 0.00000264, 0.00000513, -0.00000845, 0.00000410, -0.00000624, 0.00000848, -0.00000480, 0.00000033, -0.00000002, -0.00000214, 0.00000092, -0.00000587, 0.00000711, 0.00000084, -0.00000152, -0.00000504, -0.00000460, -0.00000226, 0.00001021, -0.00000486, 0.00000198, -0.00000823, 0.00000282, 0.00000493, 0.00000018, -0.00000347, -0.00000081, -0.00000503, 0.00000496, 0.00000775, -0.00000631, -0.00000281, -0.00000242, 0.00000434, 0.00000328, 0.00000674, -0.00000442, -0.00000211, 0.00000012, -0.00000057, 0.00000460, 0.00000160, -0.00000463, 0.00001299, -0.00000878, -0.00000473, 0.00000411, 0.00000995, -0.00000628, -0.00000393, -0.00000927, 0.00000120, -0.00000438, 0.00000120, -0.00000343, 0.00000581, -0.00001378, 0.00000145, -0.00000109, 0.00000450, -0.00001499, -0.00000046, 0.00000201, 0.00000831, },
		{ 0.00000584, -0.00001923, 0.00001079, -0.00003011, -0.00000725, 0.00000046, 0.00004978, 0.00000627, 0.00002871, 0.00002698, 0.00000320, -0.00002514, -0.00001120, -0.00000646, -0.00000428, -0.00000650, -0.00000114, -0.00000856, 0.00000153, -0.00001631, 0.00000491, -0.00003029, -0.00000136, -0.00013437, 0.00006201, 0.00016583, 0.00016723, -0.00003327, -0.00009873, 0.00000816, -0.00003051, 0.00001218, -0.00001854, 0.00000682, -0.00002036, 0.00002094, -0.00002098, 0.00001423, -0.00001491, 0.00000407, 0.00000006, -0.00001181, -0.00000169, -0.00000166, -0.00000146, 0.00000174, -0.00001882, 0.00000521, -0.00001318, 0.00001310, -0.00000681, -0.00000148, 0.00000913, 0.00000193, 0.00000257, 0.00000087, -0.00000651, 0.00000567, -0.00000614, 0.00000734, -0.00000090, 0.00000065, 0.00000130, 0.00000149, -0.00000132, 0.00000496, 0.00000261, -0.00000757, 0.00000402, -0.00000641, 0.00001023, -0.00000799, 0.00000335, -0.00000291, 0.00000075, -0.00000207, -0.00000326, 0.00000659, 0.00000068, -0.00000285, -0.00000414, -0.00000557, 0.00000057, 0.00000825, -0.00000433, 0.00000111, -0.00000848, 0.00000509, 0.00000403, -0.00000063, -0.00000348, -0.00000053, -0.00000608, 0.00000850, 0.00000496, -0.00000640, -0.00000289, -0.00000140, 0.00000392, 0.00000450, 0.00000537, -0.00000495, -0.00000171, 0.00000027, -0.00000056, 0.00000595, -0.00000122, -0.00000136, 0.00001140, -0.00001080, -0.00000365, 0.00000658, 0.00000836, -0.00000772, -0.00000352, -0.00000847, 0.00000090, -0.00000391, 0.00000081, -0.00000217, 0.00000431, -0.00001384, 0.00000253, -0.00000078, 0.00000352, -0.00001587, 0.00000090, 0.00000354, },
		{ 0.00000631, 0.00000122, -0.00001598, 0.00000837, -0.00003223, -0.00000151, 0.00000405, 0.00004998, -0.00000086, 0.00003450, 0.00002010, 0.00000226, -0.00002610, -0.00000844, -0.00000690, -0.00000405, -0.00000602, -0.00000209, -0.00000688, -0.00000123, -0.00001328, 0.00000067, -0.00002683, -0.00001180, -0.00013723, 0.00009205, 0.00017697, 0.00016229, -0.00008089, -0.00007004, -0.00000130, -0.00002142, 0.00000570, -0.00001286, 0.00000094, -0.00001401, 0.00001653, -0.00001887, 0.00001377, -0.00001573, 0.00000582, -0.00000134, -0.00001292, 0.00000109, -0.00000448, 0.00000270, -0.00000463, -0.00001389, 0.00000309, -0.00001235, 0.00001506, -0.00001050, 0.00000263, 0.00000711, 0.00000253, 0.00000260, -0.00000074, -0.00000535, 0.00000536, -0.00000636, 0.00000891, -0.00000283, 0.00000204, 0.00000030, 0.00000207, -0.00000231, 0.00000747, -0.00000073, -0.00000542, 0.00000257, -0.00000499, 0.00001031, -0.00000995, 0.00000561, -0.00000524, 0.00000332, -0.00000512, -0.00000008, 0.00000534, 0.00000077, -0.00000422, -0.00000313, -0.00000645, 0.00000344, 0.00000574, -0.00000303, -0.00000072, -0.00000749, 0.00000659, 0.00000327, -0.00000181, -0.00000268, -0.00000128, -0.00000595, 0.00001144, 0.00000179, -0.00000603, -0.00000299, -0.00000044, 0.00000355, 0.00000577, 0.00000355, -0.00000497, -0.00000153, 0.00000052, -0.00000041, 0.00000691, -0.00000393, 0.00000255, 0.00000836, -0.00001168, -0.00000278, 0.00000959, 0.00000528, -0.00000782, -0.00000372, -0.00000719, -0.00000010, -0.00000262, -0.00000039, -0.00000002, 0.00000155, -0.00001241, 0.00000234, 0.00000070, 0.00000106, -0.00001522, 0.00000118, },
		{ 0.00000061, 0.00000988, -0.00000505, -0.00001019, 0.00000245, -0.00003023, 0.00000186, 0.00000912, 0.00004690, -0.00000506, 0.00003878, 0.00001359, 0.00000157, -0.00002658, -0.00000619, -0.00000695, -0.00000440, -0.00000490, -0.00000378, -0.00000434, -0.00000517, -0.00000859, -0.00000591, -0.00002051, -0.00002746, -0.00013242, 0.00011951, 0.00019259, 0.00014162, -0.00011787, -0.00003996, -0.00001415, -0.00000991, -0.00000296, -0.00000540, -0.00000652, -0.00000584, 0.00000987, -0.00001444, 0.00001123, -0.00001486, 0.00000649, -0.00000241, -0.00001394, 0.00000388, -0.00000731, 0.00000685, -0.00001184, -0.00000665, -0.00000218, -0.00000814, 0.00001394, -0.00001187, 0.00000542, 0.00000554, 0.00000308, 0.00000246, -0.00000242, -0.00000364, 0.00000428, -0.00000568, 0.00000970, -0.00000436, 0.00000326, -0.00000060, 0.00000245, -0.00000290, 0.00000968, -0.00000440, -0.00000243, 0.00000001, -0.00000205, 0.00000850, -0.00001025, 0.00000660, -0.00000645, 0.00000494, -0.00000755, 0.00000308, 0.00000377, 0.00000087, -0.00000539, -0.00000225, -0.00000693, 0.00000590, 0.00000319, -0.00000148, -0.00000310, -0.00000550, 0.00000735, 0.00000256, -0.00000312, -0.00000128, -0.00000301, -0.00000439, 0.00001327, -0.00000131, -0.00000540, -0.00000297, 0.00000029, 0.00000348, 0.00000674, 0.00000159, -0.00000468, -0.00000142, 0.00000073, 0.00000005, 0.00000718, -0.00000595, 0.00000639, 0.00000443, -0.00001169, -0.00000179, 0.00001256, 0.00000116, -0.00000668, -0.00000452, -0.00000556, -0.00000153, -0.00000081, -0.00000209, 0.00000267, -0.00000212, -0.00000976, 0.00000108, 0.00000312, -0.00000265, -0.00001322, },
		{ -0.00001036, -0.00000022, 0.00001340, -0.00001193, -0.00000291, -0.00000616, -0.00002428, 0.00000245, 0.00001593, 0.00004037, -0.00000570, 0.00004069, 0.00000819, 0.00000069, -0.00002644, -0.00000448, -0.00000672, -0.00000511, -0.00000341, -0.00000581, -0.00000148, -0.00000958, -0.00000316, -0.00001380, -0.00001254, -0.00004708, -0.00012058, 0.00014549, 0.00021038, 0.00010495, -0.00013934, -0.00001406, -0.00002653, 0.00000129, -0.00001160, 0.00000201, -0.00001387, 0.00000258, 0.00000237, -0.00000886, 0.00000751, -0.00001299, 0.00000663, -0.00000371, -0.00001419, 0.00000586, -0.00000921, 0.00000974, -0.00001829, 0.00000127, -0.00000938, -0.00000121, 0.00000983, -0.00001060, 0.00000633, 0.00000500, 0.00000313, 0.00000241, -0.00000421, -0.00000141, 0.00000241, -0.00000398, 0.00000944, -0.00000516, 0.00000401, -0.00000114, 0.00000239, -0.00000274, 0.00001109, -0.00000773, 0.00000078, -0.00000307, 0.00000192, 0.00000503, -0.00000882, 0.00000611, -0.00000625, 0.00000517, -0.00000881, 0.00000565, 0.00000229, 0.00000073, -0.00000611, -0.00000175, -0.00000678, 0.00000759, 0.00000110, -0.00000022, -0.00000546, -0.00000288, 0.00000752, 0.00000173, -0.00000421, 0.00000028, -0.00000537, -0.00000149, 0.00001372, -0.00000399, -0.00000469, -0.00000278, 0.00000069, 0.00000386, 0.00000714, -0.00000018, -0.00000432, -0.00000123, 0.00000079, 0.00000089, 0.00000655, -0.00000684, 0.00000948, 0.00000025, -0.00001124, -0.00000026, 0.00001479, -0.00000327, -0.00000475, -0.00000567, -0.00000390, -0.00000297, 0.00000106, -0.00000381, 0.00000532, -0.00000608, -0.00000647, -0.00000072, 0.00000591, -0.00000707, },
		{ -0.00001144, -0.00000732, -0.00000059, 0.00001584, -0.00001808, 0.00000433, -0.00001590, -0.00001543, 0.00000060, 0.00002413, 0.00003079, -0.00000270, 0.00003978, 0.00000443, -0.00000070, -0.00002562, -0.00000322, -0.00000639, -0.00000587, -0.00000198, -0.00000767, 0.00000100, -0.00001354, 0.00000186, -0.00002163, -0.00000464, -0.00006876, -0.00010307, 0.00017209, 0.00022593, 0.00005510, -0.00014324, 0.00000352, -0.00003510, 0.00000963, -0.00001813, 0.00000759, -0.00001948, 0.00000968, -0.00000451, -0.00000337, 0.00000366, -0.00001095, 0.00000687, -0.00000581, -0.00001302, 0.00000629, -0.00000931, 0.00001023, -0.00002234, 0.00000795, -0.00001669, 0.00000704, 0.00000359, -0.00000706, 0.00000537, 0.00000571, 0.00000242, 0.00000256, -0.00000604, 0.00000116, -0.00000007, -0.00000136, 0.00000807, -0.00000506, 0.00000411, -0.00000114, 0.00000172, -0.00000160, 0.00001126, -0.00001010, 0.00000349, -0.00000591, 0.00000617, 0.00000049, -0.00000598, 0.00000429, -0.00000469, 0.00000392, -0.00000859, 0.00000722, 0.00000120, 0.00000017, -0.00000625, -0.00000174, -0.00000589, 0.00000829, -0.00000019, 0.00000031, -0.00000727, -0.00000009, 0.00000732, 0.00000059, -0.00000469, 0.00000148, -0.00000781, 0.00000239, 0.00001276, -0.00000603, -0.00000400, -0.00000244, 0.00000082, 0.00000470, 0.00000686, -0.00000155, -0.00000405, -0.00000085, 0.00000067, 0.00000209, 0.00000502, -0.00000634, 0.00001127, -0.00000358, -0.00001067, 0.00000216, 0.00001559, -0.00000717, -0.00000263, -0.00000678, -0.00000259, -0.00000393, 0.00000245, -0.00000497, 0.00000728, -0.00000958, -0.00000327, -0.00000236, 0.00000836, },
		{ 0.00000972, -0.00001500, -0.00000476, 0.00000017, 0.00001626, -0.00002215, 0.00000983, -0.00002480, -0.00000540, -0.00000261, 0.00003275, 0.00001915, 0.00000341, 0.00003610, 0.00000248, -0.00000273, -0.00002415, -0.00000231, -0.00000620, -0.00000633, -0.00000102, -0.00000884, 0.00000247, -0.00001621, 0.00000533, -0.00002799, 0.00000141, -0.00009038, -0.00008147, 0.00020168, 0.00023355, -0.00000222, -0.00013070, 0.00001086, -0.00003782, 0.00001343, -0.00002111, 0.00001007, -0.00002219, 0.00001426, -0.00000961, 0.00000100, 0.00000050, -0.00000936, 0.00000761, -0.00000898, -0.00001010, 0.00000477, -0.00000719, 0.00000760, -0.00002274, 0.00001160, -0.00002214, 0.00001482, -0.00000336, -0.00000220, 0.00000304, 0.00000746, 0.00000101, 0.00000288, -0.00000771, 0.00000378, -0.00000284, 0.00000192, 0.00000573, -0.00000407, 0.00000350, -0.00000059, 0.00000045, 0.00000055, 0.00000996, -0.00001106, 0.00000508, -0.00000775, 0.00000980, -0.00000429, -0.00000235, 0.00000163, -0.00000216, 0.00000143, -0.00000694, 0.00000764, 0.00000065, -0.00000084, -0.00000581, -0.00000218, -0.00000436, 0.00000803, -0.00000058, -0.00000013, -0.00000814, 0.00000248, 0.00000694, -0.00000095, -0.00000430, 0.00000181, -0.00000969, 0.00000670, 0.00001060, -0.00000737, -0.00000334, -0.00000208, 0.00000083, 0.00000585, 0.00000592, -0.00000245, -0.00000393, -0.00000030, 0.00000042, 0.00000346, 0.00000281, -0.00000452, 0.00001151, -0.00000668, -0.00001017, 0.00000557, 0.00001453, -0.00000983, -0.00000092, -0.00000747, -0.00000196, -0.00000404, 0.00000292, -0.00000508, 0.00000796, -0.00001193, -0.00000088, -0.00000317, },
		{ -0.00000267, 0.00000949, -0.00001719, -0.00000314, 0.00000243, 0.00001407, -0.00002320, 0.00001219, -0.00003096, 0.00000383, -0.00000566, 0.00004042, 0.00000690, 0.00001162, 0.00003014, 0.00000221, -0.00000537, -0.00002214, -0.00000164, -0.00000632, -0.00000625, -0.00000083, -0.00000899, 0.00000250, -0.00001702, 0.00000647, -0.00003184, 0.00000416, -0.00011015, -0.00005685, 0.00023592, 0.00022741, -0.00005931, -0.00010580, 0.00000859, -0.00003441, 0.00001234, -0.00002012, 0.00000910, -0.00002159, 0.00001585, -0.00001241, 0.00000376, -0.00000156, -0.00000846, 0.00000887, -0.00001308, -0.00000560, 0.00000143, -0.00000299, 0.00000190, -0.00001899, 0.00001103, -0.00002410, 0.00002034, -0.00000942, 0.00000272, 0.00000024, 0.00000970, -0.00000083, 0.00000318, -0.00000896, 0.00000604, -0.00000546, 0.00000542, 0.00000276, -0.00000239, 0.00000233, 0.00000039, -0.00000124, 0.00000351, 0.00000726, -0.00001041, 0.00000518, -0.00000801, 0.00001200, -0.00000840, 0.00000128, -0.00000121, 0.00000070, -0.00000175, -0.00000419, 0.00000702, 0.00000061, -0.00000221, -0.00000491, -0.00000290, -0.00000243, 0.00000706, -0.00000021, -0.00000152, -0.00000789, 0.00000456, 0.00000644, -0.00000283, -0.00000297, 0.00000097, -0.00001042, 0.00001078, 0.00000762, -0.00000808, -0.00000265, -0.00000186, 0.00000096, 0.00000704, 0.00000452, -0.00000295, -0.00000390, 0.00000034, 0.00000021, 0.00000472, 0.00000033, -0.00000173, 0.00001029, -0.00000894, -0.00000966, 0.00000974, 0.00001155, -0.00001084, -0.00000003, -0.00000751, -0.00000214, -0.00000313, 0.00000227, -0.00000392, 0.00000708, -0.00001270, 0.00000023, },
		{ -0.00000004, -0.00000077, 0.00000752, -0.00001781, -0.00000250, 0.00000605, 0.00000931, -0.00002091, 0.00001066, -0.00003298, 0.00001051, -0.00000706, 0.00004568, -0.00000431, 0.00002058, 0.00002269, 0.00000327, -0.00000849, -0.00001967, -0.00000117, -0.00000675, -0.00000558, -0.00000144, -0.00000811, 0.00000110, -0.00001591, 0.00000507, -0.00003281, 0.00000293, -0.00012706, -0.00002919, 0.00027472, 0.00020303, -0.00010798, -0.00007468, -0.00000052, -0.00002630, 0.00000731, -0.00001590, 0.00000526, -0.00001818, 0.00001476, -0.00001305, 0.00000502, -0.00000269, -0.00000800, 0.00001023, -0.00001749, -0.00000020, -0.00000301, 0.00000244, -0.00000603, -0.00001154, 0.00000604, -0.00002166, 0.00002232, -0.00001321, 0.00000643, -0.00000203, 0.00001172, -0.00000265, 0.00000320, -0.00000949, 0.00000757, -0.00000745, 0.00000860, -0.00000038, -0.00000036, 0.00000087, 0.00000152, -0.00000299, 0.00000689, 0.00000348, -0.00000833, 0.00000375, -0.00000646, 0.00001220, -0.00001107, 0.00000415, -0.00000352, 0.00000317, -0.00000491, -0.00000089, 0.00000569, 0.00000088, -0.00000370, -0.00000379, -0.00000364, -0.00000045, 0.00000575, 0.00000055, -0.00000356, -0.00000658, 0.00000606, 0.00000577, -0.00000480, -0.00000089, -0.00000110, -0.00000963, 0.00001404, 0.00000424, -0.00000828, -0.00000190, -0.00000190, 0.00000144, 0.00000795, 0.00000295, -0.00000321, -0.00000384, 0.00000093, 0.00000020, 0.00000552, -0.00000188, 0.00000144, 0.00000797, -0.00001052, -0.00000883, 0.00001410, 0.00000703, -0.00001020, -0.00000009, -0.00000690, -0.00000302, -0.00000135, 0.00000061, -0.00000160, 0.00000467, -0.00001184, },
		{ -0.00000969, -0.00000141, 0.00000223, 0.00000405, -0.00001704, -0.00000253, 0.00001044, 0.00000264, -0.00001579, 0.00000539, -0.00003035, 0.00001351, -0.00000568, 0.00004733, -0.00001298, 0.00002890, 0.00001471, 0.00000515, -0.00001188, -0.00001680, -0.00000098, -0.00000734, -0.00000452, -0.00000261, -0.00000652, -0.00000133, -0.00001335, 0.00000157, -0.00003128, -0.00000214, -0.00014105, 0.00000299, 0.00031552, 0.00015864, -0.00014118, -0.00004414, -0.00001237, -0.00001612, 0.00000033, -0.00001001, -0.00000008, -0.00001315, 0.00001194, -0.00001230, 0.00000542, -0.00000347, -0.00000733, 0.00001088, -0.00002123, 0.00000501, -0.00000740, 0.00000773, -0.00001460, -0.00000184, -0.00000254, -0.00001495, 0.00002024, -0.00001387, 0.00000799, -0.00000289, 0.00001281, -0.00000401, 0.00000269, -0.00000910, 0.00000805, -0.00000840, 0.00001099, -0.00000321, 0.00000161, -0.00000055, 0.00000247, -0.00000439, 0.00001013, -0.00000080, -0.00000527, 0.00000116, -0.00000327, 0.00001028, -0.00001186, 0.00000571, -0.00000474, 0.00000463, -0.00000737, 0.00000230, 0.00000407, 0.00000116, -0.00000504, -0.00000272, -0.00000414, 0.00000120, 0.00000456, 0.00000122, -0.00000580, -0.00000449, 0.00000703, 0.00000480, -0.00000648, 0.00000151, -0.00000413, -0.00000723, 0.00001604, 0.00000087, -0.00000809, -0.00000107, -0.00000227, 0.00000244, 0.00000830, 0.00000149, -0.00000342, -0.00000359, 0.00000131, 0.00000056, 0.00000560, -0.00000332, 0.00000434, 0.00000508, -0.00001171, -0.00000722, 0.00001786, 0.00000171, -0.00000828, -0.00000091, -0.00000590, -0.00000422, 0.00000089, -0.00000161, 0.00000141, 0.00000120, },
		{ -0.00000263, -0.00000693, -0.00000326, 0.00000570, -0.00000030, -0.00001542, -0.00000256, 0.00001460, -0.00000470, -0.00000907, -0.00000252, -0.00002359, 0.00001265, -0.00000112, 0.00004479, -0.00001802, 0.00003539, 0.00000712, 0.00000735, -0.00001534, -0.00001357, -0.00000125, -0.00000778, -0.00000344, -0.00000388, -0.00000481, -0.00000403, -0.00001024, -0.00000306, -0.00002828, -0.00001014, -0.00015284, 0.00004241, 0.00035310, 0.00009613, -0.00015455, -0.00002015, -0.00002268, -0.00000689, -0.00000621, -0.00000443, -0.00000522, -0.00000799, 0.00000869, -0.00001120, 0.00000586, -0.00000473, -0.00000568, 0.00000995, -0.00002319, 0.00000882, -0.00001045, 0.00001132, -0.00002183, 0.00000803, -0.00001294, -0.00000510, 0.00001455, -0.00001132, 0.00000702, -0.00000185, 0.00001251, -0.00000456, 0.00000153, -0.00000772, 0.00000738, -0.00000808, 0.00001220, -0.00000530, 0.00000314, -0.00000159, 0.00000291, -0.00000502, 0.00001263, -0.00000489, -0.00000191, -0.00000195, 0.00000097, 0.00000652, -0.00001071, 0.00000578, -0.00000462, 0.00000470, -0.00000858, 0.00000483, 0.00000259, 0.00000120, -0.00000601, -0.00000189, -0.00000422, 0.00000228, 0.00000384, 0.00000134, -0.00000772, -0.00000199, 0.00000762, 0.00000334, -0.00000744, 0.00000366, -0.00000760, -0.00000344, 0.00001659, -0.00000221, -0.00000759, -0.00000026, -0.00000289, 0.00000396, 0.00000795, 0.00000036, -0.00000375, -0.00000304, 0.00000140, 0.00000132, 0.00000482, -0.00000363, 0.00000639, 0.00000217, -0.00001283, -0.00000442, 0.00002018, -0.00000345, -0.00000574, -0.00000206, -0.00000494, -0.00000521, 0.00000298, -0.00000376, 0.00000442, },
		{ 0.00000667, -0.00000599, -0.00000434, -0.00000484, 0.00000884, -0.00000478, -0.00001358, -0.00000187, 0.00001741, -0.00001126, -0.00000242, -0.00001133, -0.00001411, 0.00000867, 0.00000618, 0.00003818, -0.00001898, 0.00003924, 0.00000064, 0.00000942, -0.00001869, -0.00000998, -0.00000217, -0.00000770, -0.00000281, -0.00000471, -0.00000365, -0.00000617, -0.00000763, -0.00000763, -0.00002514, -0.00001983, -0.00016340, 0.00009236, 0.00038006, 0.00002128, -0.00014744, -0.00000650, -0.00002818, -0.00000110, -0.00001026, -0.00000087, -0.00000866, -0.00000408, 0.00000619, -0.00001071, 0.00000711, -0.00000713, -0.00000241, 0.00000674, -0.00002251, 0.00001022, -0.00001108, 0.00001185, -0.00002583, 0.00001573, -0.00002283, 0.00000599, 0.00000652, -0.00000626, 0.00000380, 0.00000110, 0.00001066, -0.00000420, -0.00000025, -0.00000547, 0.00000563, -0.00000649, 0.00001209, -0.00000639, 0.00000399, -0.00000204, 0.00000263, -0.00000457, 0.00001393, -0.00000813, 0.00000102, -0.00000477, 0.00000542, 0.00000162, -0.00000801, 0.00000451, -0.00000324, 0.00000333, -0.00000831, 0.00000634, 0.00000156, 0.00000081, -0.00000644, -0.00000141, -0.00000384, 0.00000269, 0.00000379, 0.00000062, -0.00000887, 0.00000051, 0.00000796, 0.00000131, -0.00000734, 0.00000496, -0.00001081, 0.00000124, 0.00001575, -0.00000482, -0.00000679, 0.00000039, -0.00000357, 0.00000586, 0.00000689, -0.00000033, -0.00000426, -0.00000219, 0.00000121, 0.00000241, 0.00000329, -0.00000272, 0.00000723, -0.00000036, -0.00001405, -0.00000024, 0.00002042, -0.00000757, -0.00000331, -0.00000307, -0.00000446, -0.00000545, 0.00000430, -0.00000516, },
		{ -0.00000530, 0.00000756, -0.00000820, -0.00000260, -0.00000545, 0.00001091, -0.00000867, -0.00001209, 0.00000010, 0.00001797, -0.00001571, 0.00000247, -0.00001899, -0.00000391, 0.00000304, 0.00001508, 0.00002837, -0.00001603, 0.00004012, -0.00000422, 0.00001099, -0.00002174, -0.00000610, -0.00000385, -0.00000686, -0.00000295, -0.00000470, -0.00000354, -0.00000710, -0.00000635, -0.00001113, -0.00002297, -0.00003027, -0.00017311, 0.00015558, 0.00038810, -0.00005692, -0.00012313, -0.00000407, -0.00002743, 0.00000001, -0.00001075, -0.00000025, -0.00000955, -0.00000220, 0.00000509, -0.00001132, 0.00000952, -0.00001091, 0.00000263, 0.00000108, -0.00001888, 0.00000876, -0.00000880, 0.00000860, -0.00002528, 0.00001929, -0.00002991, 0.00001609, -0.00000206, 0.00000002, -0.00000084, 0.00000548, 0.00000748, -0.00000305, -0.00000243, -0.00000267, 0.00000310, -0.00000383, 0.00001073, -0.00000645, 0.00000406, -0.00000184, 0.00000161, -0.00000297, 0.00001374, -0.00001007, 0.00000291, -0.00000655, 0.00000913, -0.00000349, -0.00000442, 0.00000239, -0.00000102, 0.00000085, -0.00000666, 0.00000671, 0.00000109, -0.00000004, -0.00000635, -0.00000124, -0.00000310, 0.00000255, 0.00000435, -0.00000102, -0.00000899, 0.00000273, 0.00000811, -0.00000124, -0.00000602, 0.00000493, -0.00001308, 0.00000618, 0.00001377, -0.00000692, -0.00000566, 0.00000067, -0.00000403, 0.00000786, 0.00000531, -0.00000063, -0.00000488, -0.00000111, 0.00000088, 0.00000357, 0.00000131, -0.00000080, 0.00000682, -0.00000236, -0.00001529, 0.00000518, 0.00001833, -0.00000999, -0.00000157, -0.00000354, -0.00000476, -0.00000461, 0.00000440, },
		{ 0.00000318, -0.00000401, 0.00000684, -0.00000890, -0.00000213, -0.00000470, 0.00001147, -0.00001151, -0.00001121, 0.00000353, 0.00001587, -0.00001723, 0.00000437, -0.00002374, 0.00000496, -0.00000245, 0.00002400, 0.00001674, -0.00000992, 0.00003810, -0.00000723, 0.00001177, -0.00002426, -0.00000214, -0.00000618, -0.00000524, -0.00000392, -0.00000377, -0.00000456, -0.00000665, -0.00000671, -0.00001317, -0.00002221, -0.00004138, -0.00018091, 0.00023283, 0.00036970, -0.00012772, -0.00008821, -0.00001073, -0.00002111, -0.00000318, -0.00000789, -0.00000243, -0.00000802, -0.00000228, 0.00000531, -0.00001284, 0.00001278, -0.00001567, 0.00000893, -0.00000650, -0.00001269, 0.00000475, -0.00000392, 0.00000177, -0.00001984, 0.00001759, -0.00003240, 0.00002317, -0.00000930, 0.00000597, -0.00000572, 0.00001045, 0.00000346, -0.00000141, -0.00000467, 0.00000029, 0.00000023, -0.00000052, 0.00000839, -0.00000561, 0.00000345, -0.00000111, -0.00000001, -0.00000035, 0.00001206, -0.00001050, 0.00000344, -0.00000675, 0.00001126, -0.00000786, -0.00000077, 0.00000009, 0.00000139, -0.00000217, -0.00000404, 0.00000612, 0.00000112, -0.00000125, -0.00000584, -0.00000124, -0.00000227, 0.00000215, 0.00000525, -0.00000339, -0.00000803, 0.00000451, 0.00000802, -0.00000407, -0.00000359, 0.00000338, -0.00001385, 0.00001071, 0.00001100, -0.00000853, -0.00000419, 0.00000041, -0.00000395, 0.00000956, 0.00000350, -0.00000073, -0.00000546, 0.00000000, 0.00000061, 0.00000448, -0.00000063, 0.00000165, 0.00000540, -0.00000397, -0.00001618, 0.00001127, 0.00001414, -0.00001052, -0.00000084, -0.00000332, -0.00000587, -0.00000262, },
		{ 0.00000018, 0.00000092, -0.00000154, 0.00000468, -0.00000815, -0.00000289, -0.00000260, 0.00001047, -0.00001324, -0.00001079, 0.00000801, 0.00001139, -0.00001579, 0.00000284, -0.00002451, 0.00001090, -0.00000617, 0.00003131, 0.00000490, -0.00000173, 0.00003358, -0.00000831, 0.00001149, -0.00002590, 0.00000150, -0.00000879, -0.00000316, -0.00000541, -0.00000229, -0.00000630, -0.00000528, -0.00000826, -0.00001421, -0.00002234, -0.00005432, -0.00018365, 0.00032167, 0.00032013, -0.00018065, -0.00005115, -0.00002209, -0.00001180, -0.00000880, -0.00000313, -0.00000617, -0.00000511, -0.00000342, 0.00000602, -0.00001441, 0.00001596, -0.00002038, 0.00001540, -0.00001485, -0.00000504, -0.00000078, 0.00000240, -0.00000742, -0.00001035, 0.00001070, -0.00002953, 0.00002591, -0.00001368, 0.00001018, -0.00000960, 0.00001507, -0.00000076, 0.00000032, -0.00000662, 0.00000297, -0.00000250, 0.00000295, 0.00000548, -0.00000415, 0.00000239, -0.00000008, -0.00000190, 0.00000294, 0.00000917, -0.00000954, 0.00000259, -0.00000520, 0.00001133, -0.00001070, 0.00000216, -0.00000170, 0.00000330, -0.00000502, -0.00000101, 0.00000492, 0.00000142, -0.00000259, -0.00000511, -0.00000120, -0.00000162, 0.00000188, 0.00000604, -0.00000610, -0.00000616, 0.00000585, 0.00000753, -0.00000677, -0.00000044, 0.00000042, -0.00001284, 0.00001428, 0.00000784, -0.00000970, -0.00000240, -0.00000051, -0.00000309, 0.00001062, 0.00000181, -0.00000086, -0.00000578, 0.00000094, 0.00000062, 0.00000480, -0.00000200, 0.00000400, 0.00000346, -0.00000549, -0.00001615, 0.00001717, 0.00000853, -0.00000940, -0.00000102, -0.00000260, -0.00000751, },
		{ -0.00000909, 0.00000318, -0.00000178, 0.00000151, 0.00000165, -0.00000643, -0.00000445, 0.00000040, 0.00000826, -0.00001419, -0.00001028, 0.00001268, 0.00000549, -0.00001218, -0.00000158, -0.00002125, 0.00001315, -0.00000706, 0.00003576, -0.00000569, 0.00000740, 0.00002711, -0.00000749, 0.00000989, -0.00002626, 0.00000425, -0.00001105, -0.00000125, -0.00000675, -0.00000104, -0.00000785, -0.00000404, -0.00000989, -0.00001549, -0.00002188, -0.00007146, -0.00017593, 0.00041568, 0.00023911, -0.00020772, -0.00002042, -0.00003271, -0.00000312, -0.00001408, 0.00000128, -0.00000958, -0.00000249, -0.00000417, 0.00000590, -0.00001469, 0.00001768, -0.00002364, 0.00002049, -0.00002232, 0.00000245, -0.00000626, 0.00000843, -0.00001700, 0.00000129, -0.00000009, -0.00002176, 0.00002395, -0.00001443, 0.00001168, -0.00001151, 0.00001846, -0.00000454, 0.00000178, -0.00000799, 0.00000503, -0.00000468, 0.00000613, 0.00000244, -0.00000242, 0.00000117, 0.00000093, -0.00000367, 0.00000640, 0.00000555, -0.00000762, 0.00000075, -0.00000215, 0.00000923, -0.00001159, 0.00000386, -0.00000248, 0.00000415, -0.00000703, 0.00000179, 0.00000354, 0.00000171, -0.00000383, -0.00000436, -0.00000096, -0.00000143, 0.00000211, 0.00000626, -0.00000863, -0.00000374, 0.00000685, 0.00000643, -0.00000884, 0.00000284, -0.00000351, -0.00001009, 0.00001654, 0.00000460, -0.00001044, -0.00000042, -0.00000204, -0.00000135, 0.00001077, 0.00000054, -0.00000125, -0.00000565, 0.00000151, 0.00000107, 0.00000429, -0.00000234, 0.00000563, 0.00000157, -0.00000732, -0.00001463, 0.00002185, 0.00000253, -0.00000727, -0.00000173, -0.00000182, },
		{ -0.00000152, -0.00000996, 0.00000560, -0.00000411, 0.00000431, -0.00000145, -0.00000449, -0.00000609, 0.00000359, 0.00000549, -0.00001491, -0.00000894, 0.00001636, -0.00000043, -0.00000783, -0.00000750, -0.00001493, 0.00001193, -0.00000491, 0.00003674, -0.00001399, 0.00001642, 0.00001927, -0.00000487, 0.00000676, -0.00002490, 0.00000547, -0.00001221, -0.00000028, -0.00000709, -0.00000098, -0.00000806, -0.00000428, -0.00001005, -0.00001874, -0.00001887, -0.00009585, -0.00015073, 0.00050454, 0.00013180, -0.00020521, -0.00000255, -0.00003768, 0.00000138, -0.00001620, 0.00000303, -0.00001068, -0.00000189, -0.00000300, 0.00000359, -0.00001237, 0.00001657, -0.00002400, 0.00002269, -0.00002727, 0.00000812, -0.00001003, 0.00001225, -0.00002459, 0.00001253, -0.00001251, -0.00001062, 0.00001802, -0.00001162, 0.00001020, -0.00001097, 0.00002005, -0.00000743, 0.00000271, -0.00000860, 0.00000627, -0.00000603, 0.00000864, -0.00000036, -0.00000073, 0.00000005, 0.00000165, -0.00000494, 0.00000951, 0.00000179, -0.00000530, -0.00000149, 0.00000177, 0.00000536, -0.00001053, 0.00000414, -0.00000205, 0.00000365, -0.00000777, 0.00000384, 0.00000239, 0.00000173, -0.00000475, -0.00000373, -0.00000044, -0.00000180, 0.00000308, 0.00000552, -0.00001046, -0.00000115, 0.00000763, 0.00000457, -0.00000981, 0.00000554, -0.00000773, -0.00000595, 0.00001738, 0.00000150, -0.00001069, 0.00000153, -0.00000396, 0.00000119, 0.00000991, -0.00000011, -0.00000204, -0.00000496, 0.00000161, 0.00000199, 0.00000290, -0.00000143, 0.00000607, 0.00000016, -0.00000973, -0.00001116, 0.00002439, -0.00000276, -0.00000496, -0.00000237, },
		{ -0.00000239, -0.00000223, -0.00000951, 0.00000675, -0.00000532, 0.00000608, -0.00000377, -0.00000315, -0.00000703, 0.00000617, 0.00000287, -0.00001597, -0.00000605, 0.00001794, -0.00000489, -0.00000445, -0.00001303, -0.00000727, 0.00000839, -0.00000034, 0.00003437, -0.00001944, 0.00002456, 0.00001064, -0.00000064, 0.00000198, -0.00002152, 0.00000464, -0.00001162, -0.00000099, -0.00000567, -0.00000296, -0.00000592, -0.00000722, -0.00000738, -0.00002549, -0.00001161, -0.00013011, -0.00010066, 0.00057508, 0.00000873, -0.00017475, -0.00000065, -0.00003395, -0.00000060, -0.00001324, 0.00000053, -0.00000811, -0.00000449, 0.00000108, -0.00000180, -0.00000660, 0.00001173, -0.00002050, 0.00002089, -0.00002842, 0.00001062, -0.00001076, 0.00001233, -0.00002808, 0.00002071, -0.00002385, 0.00000158, 0.00000971, -0.00000621, 0.00000617, -0.00000803, 0.00001964, -0.00000919, 0.00000302, -0.00000848, 0.00000668, -0.00000648, 0.00001030, -0.00000266, 0.00000070, -0.00000079, 0.00000188, -0.00000543, 0.00001184, -0.00000156, -0.00000320, -0.00000338, 0.00000572, 0.00000046, -0.00000793, 0.00000319, -0.00000057, 0.00000186, -0.00000707, 0.00000486, 0.00000172, 0.00000135, -0.00000527, -0.00000326, 0.00000029, -0.00000265, 0.00000478, 0.00000365, -0.00001123, 0.00000127, 0.00000828, 0.00000193, -0.00000937, 0.00000704, -0.00001148, -0.00000099, 0.00001692, -0.00000137, -0.00001039, 0.00000315, -0.00000595, 0.00000425, 0.00000816, -0.00000013, -0.00000322, -0.00000374, 0.00000131, 0.00000324, 0.00000083, 0.00000068, 0.00000515, -0.00000052, -0.00001273, -0.00000568, 0.00002419, -0.00000643, -0.00000326, },
		{ -0.00000268, -0.00000149, -0.00000415, -0.00000745, 0.00000622, -0.00000493, 0.00000627, -0.00000474, -0.00000297, -0.00000669, 0.00000756, 0.00000092, -0.00001771, -0.00000129, 0.00001670, -0.00000677, -0.00000349, -0.00001632, -0.00000031, 0.00000416, 0.00000543, 0.00002939, -0.00002203, 0.00003127, 0.00000177, 0.00000486, -0.00000430, -0.00001612, 0.00000157, -0.00000894, -0.00000370, -0.00000215, -0.00000736, -0.00000099, -0.00001341, -0.00000124, -0.00003637, 0.00000034, -0.00017513, -0.00001985, 0.00061328, -0.00011542, -0.00012327, -0.00001366, -0.00002130, -0.00000940, -0.00000490, -0.00000649, -0.00000166, -0.00001045, 0.00000814, -0.00001022, 0.00000260, 0.00000311, -0.00001298, 0.00001482, -0.00002533, 0.00000943, -0.00000794, 0.00000795, -0.00002616, 0.00002377, -0.00003155, 0.00001234, 0.00000108, 0.00000031, 0.00000061, -0.00000324, 0.00001742, -0.00000983, 0.00000277, -0.00000777, 0.00000640, -0.00000610, 0.00001106, -0.00000435, 0.00000175, -0.00000128, 0.00000159, -0.00000501, 0.00001311, -0.00000411, -0.00000178, -0.00000429, 0.00000877, -0.00000453, -0.00000448, 0.00000152, 0.00000152, -0.00000084, -0.00000514, 0.00000483, 0.00000160, 0.00000056, -0.00000541, -0.00000284, 0.00000101, -0.00000372, 0.00000698, 0.00000074, -0.00001078, 0.00000329, 0.00000876, -0.00000131, -0.00000746, 0.00000691, -0.00001404, 0.00000409, 0.00001543, -0.00000398, -0.00000942, 0.00000412, -0.00000757, 0.00000741, 0.00000582, 0.00000031, -0.00000465, -0.00000214, 0.00000076, 0.00000452, -0.00000151, 0.00000361, 0.00000306, -0.00000059, -0.00001599, 0.00000143, 0.00002120, -0.00000798, },
		{ -0.00000744, -0.00000327, 0.00000028, -0.00000713, -0.00000392, 0.00000405, -0.00000293, 0.00000478, -0.00000417, -0.00000414, -0.00000490, 0.00000753, -0.00000024, -0.00001998, 0.00000510, 0.00001263, -0.00000567, -0.00000571, -0.00001605, 0.00000422, 0.00000095, 0.00001093, 0.00002287, -0.00002206, 0.00003621, -0.00000673, 0.00001106, -0.00001160, -0.00000912, -0.00000344, -0.00000442, -0.00000816, 0.00000313, -0.00001373, 0.00000621, -0.00002230, 0.00000772, -0.00005045, 0.00001536, -0.00022869, 0.00009393, 0.00060688, -0.00022413, -0.00006178, -0.00003661, -0.00000240, -0.00002304, 0.00000730, -0.00001672, 0.00000750, -0.00001866, 0.00001707, -0.00002055, 0.00001414, -0.00000829, -0.00000234, 0.00000523, -0.00001854, 0.00000496, -0.00000207, -0.00000041, -0.00001876, 0.00002076, -0.00003391, 0.00001956, -0.00000590, 0.00000630, -0.00000521, 0.00000251, 0.00001382, -0.00000951, 0.00000213, -0.00000668, 0.00000567, -0.00000509, 0.00001104, -0.00000544, 0.00000244, -0.00000147, 0.00000084, -0.00000374, 0.00001326, -0.00000569, -0.00000129, -0.00000381, 0.00001018, -0.00000864, -0.00000103, -0.00000022, 0.00000356, -0.00000382, -0.00000245, 0.00000398, 0.00000189, -0.00000049, -0.00000531, -0.00000231, 0.00000141, -0.00000458, 0.00000922, -0.00000285, -0.00000920, 0.00000488, 0.00000890, -0.00000476, -0.00000435, 0.00000504, -0.00001491, 0.00000866, 0.00001327, -0.00000635, -0.00000774, 0.00000419, -0.00000840, 0.00001016, 0.00000329, 0.00000093, -0.00000605, -0.00000044, 0.00000024, 0.00000545, -0.00000355, 0.00000670, 0.00000026, -0.00000041, -0.00001887, 0.00000933, 0.00001592, },
		{ 0.00000937, -0.00000540, -0.00000468, 0.00000247, -0.00001060, 0.00000050, 0.00000076, 0.00000023, 0.00000202, -0.00000241, -0.00000634, -0.00000197, 0.00000636, -0.00000092, -0.00002216, 0.00001224, 0.00000654, -0.00000216, -0.00001083, -0.00001192, 0.00000535, -0.00000004, 0.00001494, 0.00001588, -0.00001997, 0.00003906, -0.00001413, 0.00001711, -0.00001904, -0.00000144, -0.00000949, 0.00000107, -0.00001342, 0.00000906, -0.00002076, 0.00001414, -0.00003213, 0.00001736, -0.00006499, 0.00002928, -0.00028462, 0.00023747, 0.00054822, -0.00030194, -0.00000323, -0.00006188, 0.00001774, -0.00003770, 0.00002021, -0.00002748, 0.00001697, -0.00002693, 0.00002578, -0.00003069, 0.00002595, -0.00002054, 0.00000960, -0.00000617, -0.00000956, -0.00000137, 0.00000533, -0.00001115, -0.00000715, 0.00001214, -0.00003042, 0.00002207, -0.00000978, 0.00001037, -0.00001007, 0.00000828, 0.00000943, -0.00000848, 0.00000123, -0.00000539, 0.00000469, -0.00000368, 0.00001038, -0.00000601, 0.00000281, -0.00000147, -0.00000016, -0.00000182, 0.00001245, -0.00000638, -0.00000166, -0.00000188, 0.00000954, -0.00001113, 0.00000169, -0.00000139, 0.00000489, -0.00000636, 0.00000038, 0.00000273, 0.00000234, -0.00000160, -0.00000514, -0.00000149, 0.00000119, -0.00000478, 0.00001095, -0.00000658, -0.00000679, 0.00000609, 0.00000846, -0.00000789, -0.00000058, 0.00000170, -0.00001391, 0.00001222, 0.00001075, -0.00000847, -0.00000542, 0.00000322, -0.00000812, 0.00001206, 0.00000104, 0.00000138, -0.00000710, 0.00000105, 0.00000002, 0.00000565, -0.00000470, 0.00000924, -0.00000258, -0.00000054, -0.00002055, 0.00001684, },
		{ 0.00002271, 0.00000284, -0.00000279, -0.00000619, 0.00000433, -0.00001369, 0.00000488, -0.00000277, 0.00000370, -0.00000120, -0.00000022, -0.00000882, 0.00000137, 0.00000472, -0.00000183, -0.00002324, 0.00001870, -0.00000007, 0.00000235, -0.00001756, -0.00000478, 0.00000319, 0.00000155, 0.00001680, 0.00000922, -0.00001609, 0.00003948, -0.00001958, 0.00002194, -0.00002542, 0.00000558, -0.00001519, 0.00000610, -0.00001793, 0.00001385, -0.00002635, 0.00002035, -0.00004003, 0.00002429, -0.00007562, 0.00003561, -0.00033264, 0.00040105, 0.00043657, -0.00033749, 0.00004015, -0.00008104, 0.00003316, -0.00004863, 0.00002990, -0.00003537, 0.00002372, -0.00003247, 0.00003161, -0.00003801, 0.00003542, -0.00003112, 0.00002042, -0.00001710, -0.00000052, -0.00000756, 0.00001207, -0.00002176, 0.00000621, -0.00000025, -0.00002198, 0.00001984, -0.00001003, 0.00001171, -0.00001308, 0.00001328, 0.00000475, -0.00000694, 0.00000015, -0.00000400, 0.00000361, -0.00000202, 0.00000923, -0.00000613, 0.00000295, -0.00000139, -0.00000122, 0.00000045, 0.00001094, -0.00000643, -0.00000255, 0.00000118, 0.00000686, -0.00001164, 0.00000317, -0.00000154, 0.00000503, -0.00000785, 0.00000273, 0.00000153, 0.00000266, -0.00000253, -0.00000507, -0.00000025, 0.00000014, -0.00000394, 0.00001163, -0.00000984, -0.00000395, 0.00000707, 0.00000721, -0.00001012, 0.00000313, -0.00000254, -0.00001118, 0.00001452, 0.00000813, -0.00001026, -0.00000262, 0.00000128, -0.00000659, 0.00001276, -0.00000056, 0.00000134, -0.00000753, 0.00000208, 0.00000032, 0.00000488, -0.00000449, 0.00001052, -0.00000477, -0.00000153, -0.00002024, },
		{ -0.00001740, 0.00002588, -0.00000239, -0.00000067, -0.00000695, 0.00000504, -0.00001546, 0.00000821, -0.00000550, 0.00000644, -0.00000385, 0.00000143, -0.00001062, 0.00000414, 0.00000350, -0.00000384, -0.00002207, 0.00002290, -0.00000533, 0.00000587, -0.00002382, 0.00000352, -0.00000109, 0.00000514, 0.00001658, 0.00000317, -0.00001058, 0.00003707, -0.00002219, 0.00002438, -0.00002941, 0.00001044, -0.00001893, 0.00000900, -0.00001988, 0.00001549, -0.00002815, 0.00002204, -0.00004273, 0.00002453, -0.00007724, 0.00002657, -0.00035935, 0.00056859, 0.00027941, -0.00032600, 0.00005952, -0.00008703, 0.00003855, -0.00005148, 0.00003270, -0.00003719, 0.00002490, -0.00003268, 0.00003206, -0.00004000, 0.00004002, -0.00003752, 0.00002770, -0.00002516, 0.00000634, -0.00001148, 0.00001584, -0.00002944, 0.00001826, -0.00001362, -0.00001063, 0.00001403, -0.00000704, 0.00001028, -0.00001389, 0.00001701, 0.00000016, -0.00000500, -0.00000113, -0.00000247, 0.00000243, -0.00000019, 0.00000766, -0.00000582, 0.00000286, -0.00000130, -0.00000216, 0.00000280, 0.00000905, -0.00000623, -0.00000347, 0.00000475, 0.00000258, -0.00001020, 0.00000327, -0.00000054, 0.00000377, -0.00000790, 0.00000413, 0.00000075, 0.00000262, -0.00000312, -0.00000516, 0.00000136, -0.00000175, -0.00000191, 0.00001088, -0.00001209, -0.00000110, 0.00000795, 0.00000500, -0.00001096, 0.00000602, -0.00000693, -0.00000716, 0.00001551, 0.00000553, -0.00001158, 0.00000034, -0.00000140, -0.00000390, 0.00001215, -0.00000127, 0.00000064, -0.00000715, 0.00000246, 0.00000120, 0.00000309, -0.00000272, 0.00001011, -0.00000584, -0.00000373, },
		{ -0.00000714, -0.00001196, 0.00002575, -0.00000538, 0.00000009, -0.00000631, 0.00000395, -0.00001515, 0.00000961, -0.00000653, 0.00000755, -0.00000502, 0.00000162, -0.00001084, 0.00000547, 0.00000348, -0.00000761, -0.00001779, 0.00002349, -0.00000753, 0.00000643, -0.00002729, 0.00001062, -0.00000557, 0.00000938, 0.00001506, -0.00000238, -0.00000346, 0.00003152, -0.00002122, 0.00002343, -0.00002984, 0.00001178, -0.00001928, 0.00000828, -0.00001772, 0.00001224, -0.00002416, 0.00001680, -0.00003735, 0.00001462, -0.00006536, -0.00000518, -0.00035019, 0.00071923, 0.00009228, -0.00027052, 0.00005132, -0.00007600, 0.00003076, -0.00004360, 0.00002631, -0.00003095, 0.00001874, -0.00002590, 0.00002553, -0.00003500, 0.00003797, -0.00003793, 0.00002962, -0.00002855, 0.00000931, -0.00001154, 0.00001485, -0.00003184, 0.00002602, -0.00002484, 0.00000092, 0.00000657, -0.00000205, 0.00000672, -0.00001268, 0.00001931, -0.00000413, -0.00000270, -0.00000270, -0.00000071, 0.00000107, 0.00000182, 0.00000567, -0.00000507, 0.00000250, -0.00000121, -0.00000286, 0.00000501, 0.00000707, -0.00000609, -0.00000385, 0.00000800, -0.00000251, -0.00000726, 0.00000221, 0.00000140, 0.00000125, -0.00000648, 0.00000438, 0.00000060, 0.00000210, -0.00000333, -0.00000534, 0.00000316, -0.00000426, 0.00000120, 0.00000861, -0.00001301, 0.00000143, 0.00000874, 0.00000191, -0.00001017, 0.00000746, -0.00001065, -0.00000248, 0.00001537, 0.00000298, -0.00001227, 0.00000307, -0.00000438, -0.00000036, 0.00001036, -0.00000109, -0.00000077, -0.00000594, 0.00000221, 0.00000255, 0.00000050, 0.00000047, 0.00000790, -0.00000559, },
		{ -0.00000423, -0.00001135, -0.00000441, 0.00002238, -0.00000573, -0.00000094, -0.00000395, 0.00000083, -0.00001244, 0.00000864, -0.00000539, 0.00000652, -0.00000418, -0.00000015, -0.00000898, 0.00000485, 0.00000505, -0.00001335, -0.00001013, 0.00001983, -0.00000562, 0.00000265, -0.00002608, 0.00001424, -0.00000815, 0.00001259, 0.00001334, -0.00000782, 0.00000523, 0.00002277, -0.00001635, 0.00001856, -0.00002608, 0.00000886, -0.00001540, 0.00000315, -0.00001066, 0.00000329, -0.00001344, 0.00000351, -0.00002251, -0.00000716, -0.00003767, -0.00006433, -0.00029236, 0.00083015, -0.00010299, -0.00018182, 0.00001820, -0.00004841, 0.00000992, -0.00002494, 0.00001065, -0.00001656, 0.00000516, -0.00001212, 0.00001196, -0.00002276, 0.00002882, -0.00003175, 0.00002546, -0.00002649, 0.00000764, -0.00000713, 0.00000842, -0.00002764, 0.00002738, -0.00003126, 0.00001001, -0.00000030, 0.00000330, 0.00000215, -0.00001001, 0.00002033, -0.00000804, -0.00000007, -0.00000467, 0.00000143, -0.00000058, 0.00000407, 0.00000326, -0.00000383, 0.00000179, -0.00000104, -0.00000334, 0.00000698, 0.00000514, -0.00000624, -0.00000328, 0.00001016, -0.00000749, -0.00000354, 0.00000050, 0.00000379, -0.00000205, -0.00000389, 0.00000355, 0.00000108, 0.00000117, -0.00000327, -0.00000543, 0.00000480, -0.00000694, 0.00000500, 0.00000503, -0.00001250, 0.00000349, 0.00000934, -0.00000176, -0.00000783, 0.00000713, -0.00001305, 0.00000219, 0.00001440, 0.00000047, -0.00001215, 0.00000516, -0.00000718, 0.00000354, 0.00000770, -0.00000018, -0.00000271, -0.00000406, 0.00000147, 0.00000409, -0.00000244, 0.00000461, 0.00000420, },
		{ -0.00000037, -0.00000229, -0.00001557, 0.00000426, 0.00001649, -0.00000371, -0.00000362, -0.00000013, -0.00000403, -0.00000759, 0.00000547, -0.00000218, 0.00000342, -0.00000137, -0.00000382, -0.00000509, 0.00000236, 0.00000798, -0.00002054, 0.00000030, 0.00001227, 0.00000034, -0.00000575, -0.00001933, 0.00001290, -0.00000718, 0.00001328, 0.00001250, -0.00001345, 0.00001521, 0.00001124, -0.00000785, 0.00000998, -0.00001832, 0.00000181, -0.00000747, -0.00000613, 0.00000087, -0.00001077, 0.00000328, -0.00001706, 0.00000091, -0.00003971, 0.00000449, -0.00015081, -0.00017794, 0.00088045, -0.00028149, -0.00007658, -0.00003155, -0.00000917, -0.00002038, 0.00000166, -0.00001191, 0.00000386, -0.00001390, 0.00000687, -0.00000699, -0.00000472, 0.00001369, -0.00001983, 0.00001595, -0.00001956, 0.00000192, 0.00000099, -0.00000272, -0.00001712, 0.00002169, -0.00003136, 0.00001465, -0.00000467, 0.00000740, -0.00000223, -0.00000663, 0.00002029, -0.00001150, 0.00000282, -0.00000702, 0.00000397, -0.00000257, 0.00000656, 0.00000043, -0.00000209, 0.00000068, -0.00000069, -0.00000364, 0.00000870, 0.00000329, -0.00000668, -0.00000156, 0.00001061, -0.00001144, 0.00000013, -0.00000121, 0.00000596, -0.00000544, -0.00000068, 0.00000198, 0.00000197, 0.00000002, -0.00000312, -0.00000518, 0.00000584, -0.00000923, 0.00000887, 0.00000060, -0.00001073, 0.00000509, 0.00000950, -0.00000553, -0.00000432, 0.00000504, -0.00001372, 0.00000627, 0.00001291, -0.00000203, -0.00001113, 0.00000630, -0.00000929, 0.00000725, 0.00000466, 0.00000110, -0.00000486, -0.00000181, 0.00000052, 0.00000542, -0.00000513, 0.00000896, },
		{ 0.00001267, -0.00000498, -0.00000050, -0.00001882, 0.00001268, 0.00000933, -0.00000020, -0.00000724, 0.00000440, -0.00000975, -0.00000144, 0.00000085, 0.00000238, -0.00000105, 0.00000272, -0.00000870, 0.00000009, -0.00000128, 0.00001141, -0.00002797, 0.00001200, 0.00000224, 0.00000911, -0.00001771, -0.00000757, 0.00000638, -0.00000201, 0.00001068, 0.00001314, -0.00001922, 0.00002576, -0.00000204, 0.00000322, -0.00000124, -0.00000770, -0.00000814, 0.00000319, -0.00001802, 0.00001503, -0.00002775, 0.00002341, -0.00004189, 0.00002934, -0.00007863, 0.00005529, -0.00025814, -0.00000678, 0.00085512, -0.00041933, 0.00002557, -0.00008575, 0.00003346, -0.00005368, 0.00003088, -0.00003682, 0.00002630, -0.00003475, 0.00002763, -0.00002807, 0.00001616, -0.00000482, -0.00000446, 0.00000310, -0.00000957, -0.00000611, 0.00001088, -0.00001639, -0.00000212, 0.00000996, -0.00002515, 0.00001409, -0.00000543, 0.00000915, -0.00000548, -0.00000319, 0.00001942, -0.00001436, 0.00000571, -0.00000954, 0.00000676, -0.00000477, 0.00000918, -0.00000267, 0.00000007, -0.00000082, -0.00000013, -0.00000384, 0.00001021, 0.00000144, -0.00000724, 0.00000119, 0.00000906, -0.00001367, 0.00000301, -0.00000229, 0.00000727, -0.00000817, 0.00000246, 0.00000016, 0.00000294, -0.00000109, -0.00000312, -0.00000433, 0.00000591, -0.00001054, 0.00001211, -0.00000400, -0.00000808, 0.00000633, 0.00000894, -0.00000875, -0.00000032, 0.00000158, -0.00001258, 0.00000935, 0.00001117, -0.00000449, -0.00000925, 0.00000627, -0.00001030, 0.00001023, 0.00000175, 0.00000233, -0.00000685, 0.00000045, -0.00000032, 0.00000613, -0.00000690, },
		{ -0.00000725, 0.00001495, -0.00000881, 0.00000042, -0.00002015, 0.00001943, 0.00000243, 0.00000353, -0.00001073, 0.00000853, -0.00001514, 0.00000477, -0.00000402, 0.00000712, -0.00000570, 0.00000689, -0.00001356, 0.00000534, -0.00000485, 0.00001400, -0.00003392, 0.00002285, -0.00000798, 0.00001836, -0.00003087, 0.00000718, -0.00000404, 0.00000671, 0.00000505, 0.00001509, -0.00002454, 0.00003562, -0.00001542, 0.00001507, -0.00001323, 0.00000374, -0.00001883, 0.00001425, -0.00002990, 0.00002877, -0.00004405, 0.00004271, -0.00006600, 0.00005683, -0.00011643, 0.00010452, -0.00037296, 0.00021164, 0.00074853, -0.00049790, 0.00010585, -0.00013100, 0.00006967, -0.00008210, 0.00005613, -0.00005836, 0.00004566, -0.00005273, 0.00004573, -0.00004707, 0.00003594, -0.00002314, 0.00001116, -0.00001016, 0.00000079, -0.00001386, 0.00001976, -0.00002946, 0.00001419, -0.00000522, -0.00001423, 0.00000896, -0.00000260, 0.00000823, -0.00000712, -0.00000004, 0.00001775, -0.00001632, 0.00000815, -0.00001180, 0.00000943, -0.00000688, 0.00001163, -0.00000575, 0.00000239, -0.00000256, 0.00000061, -0.00000395, 0.00001156, -0.00000056, -0.00000759, 0.00000458, 0.00000564, -0.00001384, 0.00000461, -0.00000232, 0.00000723, -0.00000962, 0.00000485, -0.00000138, 0.00000366, -0.00000189, -0.00000343, -0.00000277, 0.00000474, -0.00001042, 0.00001408, -0.00000808, -0.00000502, 0.00000740, 0.00000741, -0.00001082, 0.00000337, -0.00000258, -0.00000989, 0.00001127, 0.00000936, -0.00000680, -0.00000666, 0.00000509, -0.00001000, 0.00001206, -0.00000056, 0.00000312, -0.00000826, 0.00000234, -0.00000076, 0.00000593, },
		{ 0.00000468, -0.00000589, 0.00001526, -0.00001122, -0.00000003, -0.00001888, 0.00002332, -0.00000277, 0.00000618, -0.00001293, 0.00001108, -0.00001889, 0.00000965, -0.00000777, 0.00001069, -0.00000916, 0.00000978, -0.00001700, 0.00000923, -0.00000694, 0.00001427, -0.00003654, 0.00003050, -0.00001574, 0.00002524, -0.00004211, 0.00002183, -0.00001583, 0.00001710, -0.00000229, 0.00001741, -0.00002828, 0.00004316, -0.00002682, 0.00002545, -0.00002363, 0.00001345, -0.00002753, 0.00002277, -0.00003851, 0.00003840, -0.00005538, 0.00005609, -0.00008333, 0.00007600, -0.00014367, 0.00013886, -0.00047600, 0.00045799, 0.00056663, -0.00050739, 0.00014993, -0.00015571, 0.00009046, -0.00009821, 0.00007104, -0.00007098, 0.00005694, -0.00006321, 0.00005681, -0.00005977, 0.00005056, -0.00003753, 0.00002363, -0.00002071, 0.00000867, -0.00001864, 0.00002470, -0.00003849, 0.00002803, -0.00002030, -0.00000141, 0.00000111, 0.00000277, 0.00000514, -0.00000733, 0.00000282, 0.00001515, -0.00001694, 0.00000956, -0.00001321, 0.00001141, -0.00000840, 0.00001346, -0.00000838, 0.00000450, -0.00000427, 0.00000136, -0.00000388, 0.00001271, -0.00000278, -0.00000735, 0.00000799, 0.00000081, -0.00001202, 0.00000477, -0.00000116, 0.00000567, -0.00000942, 0.00000602, -0.00000223, 0.00000384, -0.00000221, -0.00000411, -0.00000051, 0.00000232, -0.00000867, 0.00001435, -0.00001106, -0.00000199, 0.00000839, 0.00000482, -0.00001130, 0.00000599, -0.00000662, -0.00000616, 0.00001205, 0.00000755, -0.00000880, -0.00000366, 0.00000293, -0.00000838, 0.00001253, -0.00000193, 0.00000315, -0.00000881, 0.00000358, -0.00000063, },
		{ 0.00000010, 0.00000250, -0.00000288, 0.00001342, -0.00001190, -0.00000200, -0.00001484, 0.00002371, -0.00000522, 0.00000670, -0.00001289, 0.00001110, -0.00001995, 0.00001207, -0.00000925, 0.00001191, -0.00001025, 0.00001016, -0.00001777, 0.00001050, -0.00000635, 0.00001103, -0.00003437, 0.00003301, -0.00001869, 0.00002702, -0.00004819, 0.00003286, -0.00002576, 0.00002647, -0.00000927, 0.00001862, -0.00002907, 0.00004661, -0.00003414, 0.00003209, -0.00003007, 0.00001890, -0.00003156, 0.00002589, -0.00004075, 0.00004042, -0.00005767, 0.00005872, -0.00008808, 0.00007970, -0.00015096, 0.00014423, -0.00054462, 0.00070475, 0.00032711, -0.00044890, 0.00015077, -0.00015271, 0.00008988, -0.00009686, 0.00007110, -0.00007067, 0.00005652, -0.00006282, 0.00005760, -0.00006289, 0.00005681, -0.00004494, 0.00003017, -0.00002601, 0.00001178, -0.00001836, 0.00002348, -0.00004066, 0.00003593, -0.00003161, 0.00000998, -0.00000685, 0.00000885, 0.00000105, -0.00000672, 0.00000568, 0.00001135, -0.00001578, 0.00000933, -0.00001313, 0.00001208, -0.00000878, 0.00001414, -0.00001005, 0.00000593, -0.00000553, 0.00000183, -0.00000345, 0.00001350, -0.00000520, -0.00000624, 0.00001074, -0.00000467, -0.00000864, 0.00000367, 0.00000099, 0.00000276, -0.00000758, 0.00000579, -0.00000219, 0.00000337, -0.00000204, -0.00000507, 0.00000218, -0.00000108, -0.00000541, 0.00001280, -0.00001257, 0.00000070, 0.00000928, 0.00000134, -0.00001003, 0.00000699, -0.00000980, -0.00000202, 0.00001193, 0.00000573, -0.00001029, -0.00000062, 0.00000014, -0.00000566, 0.00001165, -0.00000226, 0.00000228, -0.00000833, 0.00000402, },
		{ 0.00000366, 0.00000127, -0.00000028, 0.00000139, 0.00000963, -0.00001096, -0.00000519, -0.00000845, 0.00002064, -0.00000456, 0.00000463, -0.00001022, 0.00000819, -0.00001780, 0.00001140, -0.00000781, 0.00001014, -0.00000829, 0.00000737, -0.00001518, 0.00000846, -0.00000248, 0.00000382, -0.00002681, 0.00002936, -0.00001544, 0.00002190, -0.00004665, 0.00003722, -0.00003071, 0.00003201, -0.00001362, 0.00001712, -0.00002562, 0.00004458, -0.00003571, 0.00003321, -0.00003075, 0.00001824, -0.00002899, 0.00002164, -0.00003452, 0.00003253, -0.00004826, 0.00004735, -0.00007623, 0.00006283, -0.00013129, 0.00010889, -0.00055648, 0.00091972, 0.00005752, -0.00033449, 0.00011005, -0.00012100, 0.00006655, -0.00007659, 0.00005489, -0.00005610, 0.00004317, -0.00005036, 0.00004683, -0.00005501, 0.00005308, -0.00004379, 0.00002933, -0.00002479, 0.00000906, -0.00001219, 0.00001522, -0.00003449, 0.00003557, -0.00003620, 0.00001694, -0.00001232, 0.00001361, -0.00000258, -0.00000616, 0.00000894, 0.00000615, -0.00001255, 0.00000703, -0.00001106, 0.00001093, -0.00000752, 0.00001322, -0.00001031, 0.00000624, -0.00000597, 0.00000172, -0.00000240, 0.00001371, -0.00000765, -0.00000413, 0.00001219, -0.00000993, -0.00000438, 0.00000179, 0.00000366, -0.00000101, -0.00000442, 0.00000426, -0.00000128, 0.00000233, -0.00000150, -0.00000606, 0.00000491, -0.00000498, -0.00000108, 0.00000963, -0.00001254, 0.00000290, 0.00000990, -0.00000265, -0.00000721, 0.00000616, -0.00001153, 0.00000192, 0.00001117, 0.00000387, -0.00001112, 0.00000208, -0.00000283, -0.00000222, 0.00000966, -0.00000162, 0.00000058, -0.00000686, },
		{ -0.00000460, 0.00000268, 0.00000263, -0.00000315, 0.00000626, 0.00000448, -0.00000885, -0.00000889, -0.00000066, 0.00001488, -0.00000118, 0.00000020, -0.00000517, 0.00000262, -0.00001266, 0.00000776, -0.00000357, 0.00000550, -0.00000342, 0.00000152, -0.00000934, 0.00000325, 0.00000436, -0.00000685, -0.00001443, 0.00001987, -0.00000602, 0.00000959, -0.00003654, 0.00003314, -0.00002852, 0.00003158, -0.00001352, 0.00001176, -0.00001729, 0.00003646, -0.00003081, 0.00002808, -0.00002503, 0.00001087, -0.00001928, 0.00000957, -0.00001949, 0.00001442, -0.00002681, 0.00002149, -0.00004698, 0.00002413, -0.00008239, 0.00002674, -0.00049404, 0.00107092, -0.00020867, -0.00018530, 0.00003773, -0.00006603, 0.00002425, -0.00004024, 0.00002464, -0.00002914, 0.00001852, -0.00002726, 0.00002569, -0.00003696, 0.00003988, -0.00003439, 0.00002142, -0.00001740, 0.00000103, -0.00000093, 0.00000081, -0.00002042, 0.00002646, -0.00003271, 0.00001754, -0.00001333, 0.00001538, -0.00000451, -0.00000640, 0.00001285, -0.00000042, -0.00000727, 0.00000259, -0.00000685, 0.00000775, -0.00000441, 0.00001048, -0.00000890, 0.00000513, -0.00000529, 0.00000077, -0.00000050, 0.00001304, -0.00000980, -0.00000111, 0.00001195, -0.00001415, -0.00000002, -0.00000027, 0.00000621, -0.00000495, -0.00000057, 0.00000185, 0.00000024, 0.00000095, -0.00000084, -0.00000675, 0.00000715, -0.00000870, 0.00000367, 0.00000537, -0.00001115, 0.00000465, 0.00000997, -0.00000654, -0.00000339, 0.00000368, -0.00001154, 0.00000516, 0.00001006, 0.00000195, -0.00001118, 0.00000413, -0.00000552, 0.00000144, 0.00000694, -0.00000027, -0.00000173, },
		{ -0.00000427, -0.00000189, 0.00000134, 0.00000384, -0.00000552, 0.00001091, -0.00000123, -0.00000625, -0.00001219, 0.00000718, 0.00000776, 0.00000387, -0.00000568, 0.00000133, -0.00000463, -0.00000546, 0.00000203, 0.00000262, -0.00000113, 0.00000346, -0.00000641, -0.00000126, -0.00000409, 0.00001297, -0.00001943, 0.00000102, 0.00000627, 0.00000794, -0.00000842, -0.00001881, 0.00002074, -0.00001864, 0.00002435, -0.00000823, 0.00000232, -0.00000440, 0.00002280, -0.00001996, 0.00001731, -0.00001372, -0.00000220, -0.00000365, -0.00000883, 0.00000254, -0.00001172, 0.00000411, -0.00001599, -0.00000358, -0.00003270, -0.00000837, -0.00010003, -0.00034877, 0.00113205, -0.00043723, -0.00002782, -0.00005001, 0.00000137, -0.00002875, 0.00000552, -0.00001404, 0.00000533, -0.00001307, 0.00000252, -0.00000226, -0.00001184, 0.00001979, -0.00001893, 0.00000847, -0.00000582, -0.00001026, 0.00001308, -0.00001714, -0.00000084, 0.00001020, -0.00002169, 0.00001149, -0.00000915, 0.00001333, -0.00000407, -0.00000778, 0.00001733, -0.00000792, -0.00000043, -0.00000360, -0.00000080, 0.00000276, 0.00000039, 0.00000603, -0.00000586, 0.00000260, -0.00000342, -0.00000111, 0.00000236, 0.00001125, -0.00001126, 0.00000248, 0.00000989, -0.00001671, 0.00000371, -0.00000191, 0.00000802, -0.00000827, 0.00000324, -0.00000088, 0.00000198, -0.00000045, -0.00000036, -0.00000682, 0.00000842, -0.00001158, 0.00000805, 0.00000070, -0.00000881, 0.00000608, 0.00000918, -0.00000966, 0.00000067, 0.00000007, -0.00000991, 0.00000741, 0.00000881, 0.00000002, -0.00001048, 0.00000532, -0.00000756, 0.00000482, 0.00000396, 0.00000138, },
		{ 0.00000293, -0.00000663, 0.00000089, -0.00000006, 0.00000459, -0.00000688, 0.00001454, -0.00000665, -0.00000392, -0.00001415, 0.00001366, 0.00000092, 0.00000911, -0.00001165, 0.00000790, -0.00001211, 0.00000231, -0.00000436, 0.00000931, -0.00000821, 0.00001076, -0.00001478, 0.00000737, -0.00001182, 0.00002144, -0.00003163, 0.00001684, -0.00000858, 0.00002344, -0.00002900, 0.00000366, 0.00000212, -0.00000246, 0.00001115, 0.00000161, -0.00001023, 0.00001154, 0.00000537, -0.00000503, 0.00000287, 0.00000093, -0.00001842, 0.00001498, -0.00003021, 0.00002767, -0.00004135, 0.00003921, -0.00005891, 0.00004674, -0.00009890, 0.00007981, -0.00025948, -0.00012428, 0.00108744, -0.00059886, 0.00011090, -0.00013413, 0.00006736, -0.00008145, 0.00005154, -0.00005333, 0.00004042, -0.00004536, 0.00003328, -0.00003177, 0.00001565, -0.00000309, -0.00000103, -0.00000623, 0.00000681, -0.00002166, 0.00002641, -0.00003478, 0.00002032, -0.00000983, -0.00000558, 0.00000026, -0.00000052, 0.00000779, -0.00000140, -0.00001006, 0.00002181, -0.00001550, 0.00000701, -0.00001062, 0.00000628, -0.00000335, 0.00000627, 0.00000041, -0.00000163, -0.00000103, -0.00000059, -0.00000374, 0.00000609, 0.00000826, -0.00001165, 0.00000615, 0.00000620, -0.00001728, 0.00000629, -0.00000268, 0.00000858, -0.00001033, 0.00000628, -0.00000333, 0.00000350, -0.00000154, -0.00000030, -0.00000604, 0.00000837, -0.00001306, 0.00001135, -0.00000365, -0.00000597, 0.00000733, 0.00000733, -0.00001138, 0.00000411, -0.00000392, -0.00000702, 0.00000861, 0.00000755, -0.00000186, -0.00000914, 0.00000559, -0.00000867, 0.00000751, 0.00000121, },
		{ -0.00000092, 0.00000397, -0.00000840, 0.00000332, -0.00000124, 0.00000465, -0.00000689, 0.00001652, -0.00001104, -0.00000243, -0.00001408, 0.00001759, -0.00000411, 0.00001299, -0.00001623, 0.00001301, -0.00001819, 0.00000898, -0.00000972, 0.00001478, -0.00001399, 0.00001661, -0.00002164, 0.00001455, -0.00001787, 0.00002753, -0.00004083, 0.00002995, -0.00002125, 0.00003672, -0.00004798, 0.00002660, -0.00001891, 0.00001694, -0.00000555, 0.00001398, -0.00002378, 0.00002803, -0.00001302, 0.00001109, -0.00001218, 0.00001556, -0.00003405, 0.00003248, -0.00004992, 0.00005057, -0.00006828, 0.00007121, -0.00009872, 0.00009373, -0.00016174, 0.00016535, -0.00043037, 0.00016237, 0.00093566, -0.00067420, 0.00020815, -0.00019671, 0.00011811, -0.00012247, 0.00008816, -0.00008484, 0.00006869, -0.00007161, 0.00005874, -0.00005700, 0.00004016, -0.00002401, 0.00001515, -0.00001888, 0.00001695, -0.00002975, 0.00003537, -0.00004785, 0.00003841, -0.00002912, 0.00001179, -0.00001329, 0.00001057, 0.00000011, 0.00000252, -0.00001240, 0.00002534, -0.00002196, 0.00001374, -0.00001720, 0.00001318, -0.00000950, 0.00001226, -0.00000551, 0.00000302, -0.00000512, 0.00000266, -0.00000668, 0.00001031, 0.00000421, -0.00001074, 0.00000933, 0.00000134, -0.00001588, 0.00000746, -0.00000237, 0.00000767, -0.00001072, 0.00000798, -0.00000498, 0.00000444, -0.00000208, -0.00000079, -0.00000435, 0.00000688, -0.00001282, 0.00001307, -0.00000707, -0.00000308, 0.00000848, 0.00000437, -0.00001134, 0.00000620, -0.00000745, -0.00000343, 0.00000887, 0.00000633, -0.00000354, -0.00000737, 0.00000502, -0.00000876, 0.00000922, },
		{ 0.00000980, -0.00000219, 0.00000423, -0.00000926, 0.00000508, -0.00000198, 0.00000398, -0.00000542, 0.00001652, -0.00001392, -0.00000208, -0.00001172, 0.00001829, -0.00000624, 0.00001429, -0.00001817, 0.00001537, -0.00002147, 0.00001306, -0.00001256, 0.00001750, -0.00001682, 0.00001932, -0.00002519, 0.00001839, -0.00002036, 0.00002928, -0.00004475, 0.00003758, -0.00002855, 0.00004410, -0.00006101, 0.00004524, -0.00003770, 0.00003538, -0.00002220, 0.00002592, -0.00003552, 0.00004198, -0.00002903, 0.00002498, -0.00002436, 0.00002644, -0.00004501, 0.00004436, -0.00006293, 0.00006555, -0.00008594, 0.00009237, -0.00012615, 0.00012610, -0.00020680, 0.00022823, -0.00058452, 0.00048180, 0.00069075, -0.00065716, 0.00024959, -0.00022462, 0.00014281, -0.00014262, 0.00010739, -0.00010153, 0.00008383, -0.00008600, 0.00007346, -0.00007280, 0.00005688, -0.00003868, 0.00002584, -0.00002615, 0.00002159, -0.00003169, 0.00003696, -0.00005275, 0.00004914, -0.00004311, 0.00002614, -0.00002554, 0.00002130, -0.00000762, 0.00000618, -0.00001362, 0.00002678, -0.00002600, 0.00001832, -0.00002189, 0.00001852, -0.00001439, 0.00001717, -0.00001064, 0.00000710, -0.00000879, 0.00000560, -0.00000927, 0.00001444, -0.00000050, -0.00000850, 0.00001150, -0.00000405, -0.00001282, 0.00000728, -0.00000103, 0.00000537, -0.00000935, 0.00000809, -0.00000553, 0.00000460, -0.00000197, -0.00000183, -0.00000191, 0.00000410, -0.00001090, 0.00001300, -0.00000916, -0.00000044, 0.00000946, 0.00000056, -0.00000950, 0.00000651, -0.00000983, 0.00000024, 0.00000841, 0.00000515, -0.00000492, -0.00000539, 0.00000381, -0.00000789, },
		{ -0.00000623, 0.00000932, -0.00000249, 0.00000359, -0.00000903, 0.00000596, -0.00000218, 0.00000266, -0.00000264, 0.00001450, -0.00001514, -0.00000281, -0.00000733, 0.00001574, -0.00000502, 0.00001239, -0.00001681, 0.00001426, -0.00002111, 0.00001364, -0.00001193, 0.00001650, -0.00001572, 0.00001779, -0.00002427, 0.00001769, -0.00001808, 0.00002551, -0.00004199, 0.00003794, -0.00002829, 0.00004292, -0.00006467, 0.00005533, -0.00004972, 0.00004852, -0.00003494, 0.00003419, -0.00004254, 0.00005042, -0.00003954, 0.00003351, -0.00003055, 0.00003032, -0.00004784, 0.00004684, -0.00006510, 0.00006792, -0.00008889, 0.00009619, -0.00013331, 0.00013398, -0.00022099, 0.00024891, -0.00069109, 0.00079581, 0.00038089, -0.00055604, 0.00023176, -0.00021213, 0.00013608, -0.00013699, 0.00010473, -0.00009933, 0.00008213, -0.00008507, 0.00007411, -0.00007587, 0.00006263, -0.00004424, 0.00002865, -0.00002602, 0.00001903, -0.00002603, 0.00002971, -0.00004746, 0.00004961, -0.00004825, 0.00003371, -0.00003305, 0.00002881, -0.00001316, 0.00000794, -0.00001252, 0.00002511, -0.00002651, 0.00001956, -0.00002346, 0.00002107, -0.00001686, 0.00001990, -0.00001393, 0.00000965, -0.00001116, 0.00000741, -0.00001075, 0.00001774, -0.00000526, -0.00000515, 0.00001230, -0.00000927, -0.00000861, 0.00000601, 0.00000102, 0.00000204, -0.00000647, 0.00000665, -0.00000493, 0.00000396, -0.00000126, -0.00000321, 0.00000095, 0.00000044, -0.00000762, 0.00001128, -0.00000981, 0.00000180, 0.00001006, -0.00000362, -0.00000619, 0.00000497, -0.00001061, 0.00000345, 0.00000751, 0.00000400, -0.00000592, -0.00000344, 0.00000219, },
		{ 0.00000041, -0.00000405, 0.00000794, -0.00000191, 0.00000206, -0.00000771, 0.00000586, -0.00000186, 0.00000094, 0.00000106, 0.00001078, -0.00001488, -0.00000419, -0.00000167, 0.00001063, -0.00000084, 0.00000746, -0.00001224, 0.00000972, -0.00001710, 0.00001058, -0.00000771, 0.00001168, -0.00001053, 0.00001187, -0.00001866, 0.00001221, -0.00001089, 0.00001617, -0.00003252, 0.00003077, -0.00001994, 0.00003226, -0.00005726, 0.00005425, -0.00005175, 0.00005294, -0.00004056, 0.00003615, -0.00004263, 0.00005120, -0.00004236, 0.00003459, -0.00002883, 0.00002533, -0.00004060, 0.00003790, -0.00005424, 0.00005525, -0.00007427, 0.00007916, -0.00011569, 0.00011142, -0.00019577, 0.00021263, -0.00072186, 0.00106285, 0.00004443, -0.00039195, 0.00016229, -0.00016202, 0.00009917, -0.00010605, 0.00008027, -0.00007811, 0.00006337, -0.00006850, 0.00006021, -0.00006553, 0.00005662, -0.00003996, 0.00002311, -0.00001831, 0.00000937, -0.00001322, 0.00001421, -0.00003212, 0.00003911, -0.00004289, 0.00003225, -0.00003339, 0.00003090, -0.00001477, 0.00000657, -0.00000832, 0.00001977, -0.00002294, 0.00001679, -0.00002115, 0.00002001, -0.00001608, 0.00001965, -0.00001460, 0.00000993, -0.00001152, 0.00000745, -0.00001047, 0.00001947, -0.00000932, -0.00000115, 0.00001161, -0.00001371, -0.00000388, 0.00000411, 0.00000331, -0.00000174, -0.00000262, 0.00000400, -0.00000335, 0.00000270, -0.00000018, -0.00000465, 0.00000376, -0.00000351, -0.00000354, 0.00000832, -0.00000917, 0.00000368, 0.00001000, -0.00000750, -0.00000207, 0.00000191, -0.00000969, 0.00000581, 0.00000637, 0.00000294, -0.00000654, -0.00000166, },
		{ -0.00000016, -0.00000130, -0.00000165, 0.00000595, -0.00000065, -0.00000014, -0.00000543, 0.00000486, -0.00000110, -0.00000089, 0.00000512, 0.00000589, -0.00001355, -0.00000560, 0.00000416, 0.00000420, 0.00000526, 0.00000039, -0.00000533, 0.00000260, -0.00001021, 0.00000466, -0.00000068, 0.00000387, -0.00000214, 0.00000243, -0.00000924, 0.00000287, 0.00000018, 0.00000253, -0.00001773, 0.00001744, -0.00000484, 0.00001331, -0.00003940, 0.00004169, -0.00004262, 0.00004700, -0.00003734, 0.00003044, -0.00003490, 0.00004362, -0.00003683, 0.00002777, -0.00001907, 0.00001157, -0.00002358, 0.00001797, -0.00003092, 0.00002824, -0.00004288, 0.00004197, -0.00007375, 0.00005839, -0.00012988, 0.00011334, -0.00065681, 0.00124461, -0.00027593, -0.00019509, 0.00005811, -0.00008473, 0.00003964, -0.00005568, 0.00003872, -0.00004185, 0.00003099, -0.00003931, 0.00003437, -0.00004393, 0.00004054, -0.00002736, 0.00001082, -0.00000481, -0.00000537, 0.00000438, -0.00000690, -0.00000923, 0.00001938, -0.00002779, 0.00002160, -0.00002585, 0.00002667, -0.00001169, 0.00000162, -0.00000094, 0.00001095, -0.00001550, 0.00001013, -0.00001499, 0.00001526, -0.00001189, 0.00001620, -0.00001240, 0.00000764, -0.00000956, 0.00000538, -0.00000804, 0.00001911, -0.00001200, 0.00000290, 0.00000950, -0.00001689, 0.00000074, 0.00000205, 0.00000533, -0.00000528, 0.00000151, 0.00000069, -0.00000119, 0.00000112, 0.00000098, -0.00000582, 0.00000603, -0.00000711, 0.00000063, 0.00000473, -0.00000757, 0.00000528, 0.00000898, -0.00001038, 0.00000198, -0.00000199, -0.00000731, 0.00000718, 0.00000515, 0.00000201, -0.00000684, },
		{ -0.00000696, 0.00000108, -0.00000281, 0.00000070, 0.00000368, 0.00000097, -0.00000271, -0.00000249, 0.00000313, -0.00000006, -0.00000246, 0.00000895, 0.00000049, -0.00001164, -0.00000637, 0.00000898, -0.00000203, 0.00001176, -0.00000737, 0.00000252, -0.00000565, -0.00000192, -0.00000268, 0.00000765, -0.00000533, 0.00000776, -0.00000874, 0.00000217, -0.00000846, 0.00001311, -0.00001308, -0.00000028, 0.00000080, 0.00001398, -0.00001072, -0.00001406, 0.00001981, -0.00002364, 0.00003133, -0.00002553, 0.00001749, -0.00002017, 0.00002873, -0.00002413, 0.00001446, -0.00000310, -0.00000866, 0.00000056, -0.00000988, 0.00000132, -0.00000905, 0.00000064, -0.00001011, -0.00001341, -0.00001839, -0.00003076, -0.00004359, -0.00048885, 0.00131245, -0.00053991, 0.00000075, -0.00005821, 0.00000389, -0.00003025, 0.00000417, -0.00001155, 0.00000226, -0.00000863, -0.00000322, 0.00000167, -0.00001551, 0.00001825, -0.00000988, -0.00000492, 0.00001116, -0.00002175, 0.00002298, -0.00002937, 0.00001680, -0.00000565, -0.00000598, 0.00000378, -0.00001159, 0.00001682, -0.00000442, -0.00000634, 0.00000882, -0.00000033, -0.00000524, 0.00000054, -0.00000579, 0.00000749, -0.00000484, 0.00001001, -0.00000769, 0.00000307, -0.00000549, 0.00000135, -0.00000350, 0.00001646, -0.00001279, 0.00000637, 0.00000628, -0.00001852, 0.00000476, 0.00000025, 0.00000663, -0.00000798, 0.00000520, -0.00000267, 0.00000110, -0.00000042, 0.00000193, -0.00000641, 0.00000739, -0.00000982, 0.00000424, 0.00000115, -0.00000541, 0.00000670, 0.00000684, -0.00001169, 0.00000511, -0.00000589, -0.00000396, 0.00000760, 0.00000393, 0.00000130, },
		{ 0.00000088, -0.00000702, 0.00000208, -0.00000402, 0.00000276, 0.00000145, 0.00000260, -0.00000526, 0.00000072, 0.00000095, 0.00000107, -0.00000350, 0.00001200, -0.00000479, -0.00000962, -0.00000600, 0.00001177, -0.00000660, 0.00001705, -0.00001418, 0.00000961, -0.00001328, 0.00000597, -0.00000961, 0.00001537, -0.00001389, 0.00001702, -0.00001937, 0.00001318, -0.00001930, 0.00002525, -0.00002772, 0.00001651, -0.00001546, 0.00003239, -0.00003524, 0.00001402, -0.00000705, 0.00000157, 0.00000878, -0.00000744, -0.00000048, -0.00000091, 0.00000921, -0.00000706, -0.00000226, 0.00001552, -0.00003127, 0.00002714, -0.00004037, 0.00003648, -0.00004974, 0.00004833, -0.00006787, 0.00005457, -0.00010605, 0.00008570, -0.00024068, -0.00022649, 0.00125265, -0.00071616, 0.00016407, -0.00016305, 0.00008611, -0.00009621, 0.00006154, -0.00006034, 0.00004530, -0.00004754, 0.00003259, -0.00003134, 0.00001390, -0.00000519, 0.00000802, -0.00001993, 0.00002554, -0.00003568, 0.00003822, -0.00004825, 0.00004062, -0.00003068, 0.00001781, -0.00001740, 0.00000651, 0.00000347, 0.00000536, -0.00001577, 0.00001937, -0.00001235, 0.00000606, -0.00001029, 0.00000487, -0.00000191, 0.00000383, 0.00000216, -0.00000142, -0.00000294, -0.00000004, -0.00000399, 0.00000262, 0.00001176, -0.00001153, 0.00000877, 0.00000235, -0.00001853, 0.00000781, -0.00000098, 0.00000694, -0.00000938, 0.00000783, -0.00000547, 0.00000306, -0.00000159, 0.00000242, -0.00000627, 0.00000759, -0.00001126, 0.00000677, -0.00000185, -0.00000307, 0.00000796, 0.00000363, -0.00001113, 0.00000665, -0.00000898, -0.00000027, 0.00000724, 0.00000275, },
		{ 0.00000165, 0.00000074, -0.00000709, 0.00000290, -0.00000488, 0.00000435, -0.00000042, 0.00000390, -0.00000737, 0.00000377, -0.00000137, 0.00000213, -0.00000380, 0.00001384, -0.00000941, -0.00000779, -0.00000429, 0.00001197, -0.00000844, 0.00001978, -0.00001859, 0.00001441, -0.00001866, 0.00001174, -0.00001435, 0.00002062, -0.00001980, 0.00002346, -0.00002714, 0.00002138, -0.00002718, 0.00003400, -0.00003848, 0.00002930, -0.00002758, 0.00004610, -0.00005522, 0.00003926, -0.00003329, 0.00002787, -0.00001614, 0.00001312, -0.00001994, 0.00001922, -0.00001111, 0.00001048, -0.00001831, 0.00003224, -0.00005111, 0.00005039, -0.00006702, 0.00006723, -0.00008544, 0.00009043, -0.00011984, 0.00011648, -0.00018761, 0.00019735, -0.00044991, 0.00010593, 0.00106924, -0.00078713, 0.00027109, -0.00023636, 0.00014595, -0.00014497, 0.00010510, -0.00009781, 0.00007860, -0.00007794, 0.00006099, -0.00005817, 0.00003844, -0.00002464, 0.00002186, -0.00003017, 0.00003451, -0.00004345, 0.00004621, -0.00005908, 0.00005705, -0.00005019, 0.00003825, -0.00003721, 0.00002457, -0.00001030, 0.00001512, -0.00002449, 0.00002863, -0.00002298, 0.00001626, -0.00002026, 0.00001502, -0.00001110, 0.00001247, -0.00000584, 0.00000505, -0.00000914, 0.00000566, -0.00000962, 0.00000939, 0.00000571, -0.00000844, 0.00000978, -0.00000184, -0.00001704, 0.00000970, -0.00000152, 0.00000618, -0.00000927, 0.00000901, -0.00000727, 0.00000436, -0.00000215, 0.00000233, -0.00000537, 0.00000664, -0.00001132, 0.00000799, -0.00000390, -0.00000082, 0.00000896, -0.00000033, -0.00000877, 0.00000625, -0.00001062, 0.00000315, 0.00000635, },
		{ 0.00000514, 0.00000074, 0.00000077, -0.00000717, 0.00000355, -0.00000539, 0.00000534, -0.00000171, 0.00000459, -0.00000870, 0.00000622, -0.00000348, 0.00000295, -0.00000330, 0.00001427, -0.00001305, -0.00000625, -0.00000141, 0.00000954, -0.00000708, 0.00001920, -0.00001968, 0.00001589, -0.00002063, 0.00001416, -0.00001563, 0.00002206, -0.00002159, 0.00002548, -0.00003029, 0.00002488, -0.00003012, 0.00003734, -0.00004317, 0.00003558, -0.00003267, 0.00005164, -0.00006634, 0.00005647, -0.00005331, 0.00004967, -0.00003828, 0.00003160, -0.00003675, 0.00003613, -0.00002809, 0.00002438, -0.00002951, 0.00004256, -0.00006318, 0.00006472, -0.00008356, 0.00008649, -0.00010801, 0.00011739, -0.00015459, 0.00015836, -0.00024525, 0.00027952, -0.00063666, 0.00047095, 0.00078393, -0.00075169, 0.00030949, -0.00026543, 0.00017251, -0.00016707, 0.00012651, -0.00011659, 0.00009558, -0.00009386, 0.00007650, -0.00007371, 0.00005348, -0.00003610, 0.00002829, -0.00003274, 0.00003546, -0.00004267, 0.00004452, -0.00005890, 0.00006230, -0.00005974, 0.00005058, -0.00005108, 0.00003860, -0.00002118, 0.00002216, -0.00003020, 0.00003452, -0.00003014, 0.00002325, -0.00002731, 0.00002264, -0.00001820, 0.00001932, -0.00001238, 0.00001021, -0.00001416, 0.00001032, -0.00001429, 0.00001563, -0.00000068, -0.00000410, 0.00000939, -0.00000590, -0.00001426, 0.00001042, -0.00000137, 0.00000447, -0.00000772, 0.00000861, -0.00000784, 0.00000481, -0.00000200, 0.00000166, -0.00000385, 0.00000474, -0.00001016, 0.00000790, -0.00000486, 0.00000121, 0.00000945, -0.00000448, -0.00000509, 0.00000400, -0.00001049, 0.00000583, },
		{ 0.00000753, 0.00000374, 0.00000014, 0.00000080, -0.00000716, 0.00000398, -0.00000552, 0.00000569, -0.00000228, 0.00000446, -0.00000896, 0.00000772, -0.00000506, 0.00000343, -0.00000203, 0.00001323, -0.00001553, -0.00000488, 0.00000211, 0.00000506, -0.00000280, 0.00001528, -0.00001723, 0.00001376, -0.00001882, 0.00001277, -0.00001296, 0.00001914, -0.00001867, 0.00002238, -0.00002801, 0.00002279, -0.00002721, 0.00003434, -0.00004083, 0.00003421, -0.00002934, 0.00004721, -0.00006599, 0.00006201, -0.00006268, 0.00006223, -0.00005298, 0.00004384, -0.00004721, 0.00004628, -0.00003825, 0.00003130, -0.00003264, 0.00004313, -0.00006379, 0.00006602, -0.00008541, 0.00008906, -0.00011143, 0.00012201, -0.00016332, 0.00016911, -0.00026410, 0.00030985, -0.00076528, 0.00082319, 0.00043303, -0.00062474, 0.00027990, -0.00024695, 0.00016198, -0.00015872, 0.00012216, -0.00011331, 0.00009316, -0.00009244, 0.00007639, -0.00007538, 0.00005664, -0.00003757, 0.00002583, -0.00002660, 0.00002766, -0.00003289, 0.00003284, -0.00004707, 0.00005491, -0.00005694, 0.00005173, -0.00005568, 0.00004544, -0.00002646, 0.00002428, -0.00003110, 0.00003546, -0.00003232, 0.00002550, -0.00002985, 0.00002616, -0.00002166, 0.00002292, -0.00001608, 0.00001280, -0.00001676, 0.00001277, -0.00001686, 0.00002013, -0.00000625, 0.00000062, 0.00000783, -0.00000955, -0.00001051, 0.00001006, -0.00000069, 0.00000213, -0.00000501, 0.00000678, -0.00000718, 0.00000441, -0.00000120, 0.00000053, -0.00000197, 0.00000224, -0.00000814, 0.00000679, -0.00000480, 0.00000301, 0.00000918, -0.00000814, -0.00000082, 0.00000041, -0.00000867, },
		{ -0.00000553, 0.00000822, 0.00000224, 0.00000000, 0.00000058, -0.00000688, 0.00000410, -0.00000526, 0.00000540, -0.00000208, 0.00000346, -0.00000805, 0.00000802, -0.00000587, 0.00000351, -0.00000015, 0.00001086, -0.00001686, -0.00000346, 0.00000548, -0.00000046, 0.00000350, 0.00000872, -0.00001184, 0.00000851, -0.00001368, 0.00000799, -0.00000677, 0.00001234, -0.00001148, 0.00001455, -0.00002061, 0.00001539, -0.00001875, 0.00002542, -0.00003196, 0.00002575, -0.00001814, 0.00003318, -0.00005392, 0.00005466, -0.00005923, 0.00006271, -0.00005716, 0.00004702, -0.00004893, 0.00004752, -0.00003961, 0.00002953, -0.00002627, 0.00003265, -0.00005159, 0.00005277, -0.00007086, 0.00007299, -0.00009337, 0.00010133, -0.00014204, 0.00014320, -0.00023584, 0.00027314, -0.00080544, 0.00111628, 0.00006176, -0.00043408, 0.00019499, -0.00018714, 0.00011817, -0.00012244, 0.00009378, -0.00008923, 0.00007233, -0.00007447, 0.00006126, -0.00006355, 0.00004819, -0.00002946, 0.00001524, -0.00001283, 0.00001248, -0.00001582, 0.00001319, -0.00002551, 0.00003614, -0.00004210, 0.00004106, -0.00004966, 0.00004353, -0.00002474, 0.00002033, -0.00002634, 0.00003080, -0.00002893, 0.00002237, -0.00002720, 0.00002480, -0.00002070, 0.00002248, -0.00001619, 0.00001207, -0.00001622, 0.00001228, -0.00001652, 0.00002199, -0.00000998, 0.00000482, 0.00000550, -0.00001258, -0.00000611, 0.00000880, 0.00000033, -0.00000046, -0.00000164, 0.00000388, -0.00000552, 0.00000331, 0.00000010, -0.00000086, -0.00000002, -0.00000043, -0.00000572, 0.00000506, -0.00000395, 0.00000461, 0.00000789, -0.00001062, 0.00000312, -0.00000377, },
		{ -0.00000763, -0.00000169, 0.00000804, 0.00000064, 0.00000047, -0.00000011, -0.00000615, 0.00000381, -0.00000461, 0.00000453, -0.00000121, 0.00000168, -0.00000600, 0.00000704, -0.00000577, 0.00000316, 0.00000214, 0.00000742, -0.00001713, -0.00000175, 0.00000792, -0.00000575, 0.00001047, 0.00000080, -0.00000469, 0.00000133, -0.00000640, 0.00000100, 0.00000169, 0.00000300, -0.00000142, 0.00000344, -0.00000955, 0.00000413, -0.00000625, 0.00001226, -0.00001849, 0.00001231, -0.00000143, 0.00001207, -0.00003241, 0.00003593, -0.00004353, 0.00005083, -0.00005001, 0.00004035, -0.00004143, 0.00003966, -0.00003218, 0.00001942, -0.00001125, 0.00001235, -0.00002801, 0.00002659, -0.00004167, 0.00004021, -0.00005594, 0.00005753, -0.00009288, 0.00008243, -0.00016120, 0.00016524, -0.00073798, 0.00131024, -0.00028281, -0.00021519, 0.00007621, -0.00009997, 0.00005136, -0.00006631, 0.00004790, -0.00004984, 0.00003784, -0.00004412, 0.00003477, -0.00004141, 0.00003091, -0.00001440, -0.00000073, 0.00000559, -0.00000692, 0.00000505, -0.00001049, 0.00000166, 0.00000974, -0.00001819, 0.00002051, -0.00003406, 0.00003331, -0.00001625, 0.00001060, -0.00001630, 0.00002108, -0.00002052, 0.00001434, -0.00001971, 0.00001882, -0.00001545, 0.00001807, -0.00001272, 0.00000802, -0.00001248, 0.00000872, -0.00001307, 0.00002077, -0.00001116, 0.00000771, 0.00000288, -0.00001490, -0.00000142, 0.00000689, 0.00000144, -0.00000286, 0.00000184, 0.00000042, -0.00000320, 0.00000179, 0.00000166, -0.00000230, 0.00000171, -0.00000288, -0.00000336, 0.00000318, -0.00000260, 0.00000604, 0.00000551, -0.00001143, 0.00000588, },
		{ 0.00000688, -0.00001037, 0.00000219, 0.00000723, -0.00000100, 0.00000156, -0.00000143, -0.00000483, 0.00000302, -0.00000360, 0.00000321, 0.00000018, -0.00000066, -0.00000304, 0.00000488, -0.00000472, 0.00000244, 0.00000457, 0.00000321, -0.00001644, 0.00000038, 0.00000877, -0.00000962, 0.00001666, -0.00000695, 0.00000267, -0.00000620, 0.00000140, -0.00000650, 0.00001060, -0.00000696, 0.00000943, -0.00000879, 0.00000294, -0.00000871, 0.00000793, -0.00000259, -0.00000325, -0.00000288, 0.00001710, -0.00001197, -0.00000578, 0.00000980, -0.00001881, 0.00002898, -0.00003321, 0.00002525, -0.00002631, 0.00002452, -0.00001800, 0.00000344, 0.00000936, -0.00001413, 0.00000279, -0.00000795, -0.00000294, -0.00000354, -0.00000564, -0.00000206, -0.00002408, -0.00000381, -0.00005076, -0.00000481, -0.00055925, 0.00137808, -0.00055925, -0.00000481, -0.00005076, -0.00000381, -0.00002408, -0.00000206, -0.00000564, -0.00000354, -0.00000294, -0.00000795, 0.00000279, -0.00001413, 0.00000936, 0.00000344, -0.00001800, 0.00002452, -0.00002631, 0.00002525, -0.00003321, 0.00002898, -0.00001881, 0.00000980, -0.00000578, -0.00001197, 0.00001710, -0.00000288, -0.00000325, -0.00000259, 0.00000793, -0.00000871, 0.00000294, -0.00000879, 0.00000943, -0.00000696, 0.00001060, -0.00000650, 0.00000140, -0.00000620, 0.00000267, -0.00000695, 0.00001666, -0.00000962, 0.00000877, 0.00000038, -0.00001644, 0.00000321, 0.00000457, 0.00000244, -0.00000472, 0.00000488, -0.00000304, -0.00000066, 0.00000018, 0.00000321, -0.00000360, 0.00000302, -0.00000483, -0.00000143, 0.00000156, -0.00000100, 0.00000723, 0.00000219, -0.00001037, },
		{ -0.00000763, 0.00000588, -0.00001143, 0.00000551, 0.00000604, -0.00000260, 0.00000318, -0.00000336, -0.00000288, 0.00000171, -0.00000230, 0.00000166, 0.00000179, -0.00000320, 0.00000042, 0.00000184, -0.00000286, 0.00000144, 0.00000689, -0.00000142, -0.00001490, 0.00000288, 0.00000771, -0.00001116, 0.00002077, -0.00001307, 0.00000872, -0.00001248, 0.00000802, -0.00001272, 0.00001807, -0.00001545, 0.00001882, -0.00001971, 0.00001434, -0.00002052, 0.00002108, -0.00001630, 0.00001060, -0.00001625, 0.00003331, -0.00003406, 0.00002051, -0.00001819, 0.00000974, 0.00000166, -0.00001049, 0.00000505, -0.00000692, 0.00000559, -0.00000073, -0.00001440, 0.00003091, -0.00004141, 0.00003477, -0.00004412, 0.00003784, -0.00004984, 0.00004790, -0.00006631, 0.00005136, -0.00009997, 0.00007621, -0.00021519, -0.00028281, 0.00131024, -0.00073798, 0.00016524, -0.00016120, 0.00008243, -0.00009288, 0.00005753, -0.00005594, 0.00004021, -0.00004167, 0.00002659, -0.00002801, 0.00001235, -0.00001125, 0.00001942, -0.00003218, 0.00003966, -0.00004143, 0.00004035, -0.00005001, 0.00005083, -0.00004353, 0.00003593, -0.00003241, 0.00001207, -0.00000143, 0.00001231, -0.00001849, 0.00001226, -0.00000625, 0.00000413, -0.00000955, 0.00000344, -0.00000142, 0.00000300, 0.00000169, 0.00000100, -0.00000640, 0.00000133, -0.00000469, 0.00000080, 0.00001047, -0.00000575, 0.00000792, -0.00000175, -0.00001713, 0.00000742, 0.00000214, 0.00000316, -0.00000577, 0.00000704, -0.00000600, 0.00000168, -0.00000121, 0.00000453, -0.00000461, 0.00000381, -0.00000615, -0.00000011, 0.00000047, 0.00000064, 0.00000804, -0.00000169, },
		{ -0.00000553, -0.00000377, 0.00000312, -0.00001062, 0.00000789, 0.00000461, -0.00000395, 0.00000506, -0.00000572, -0.00000043, -0.00000002, -0.00000086, 0.00000010, 0.00000331, -0.00000552, 0.00000388, -0.00000164, -0.00000046, 0.00000033, 0.00000880, -0.00000611, -0.00001258, 0.00000550, 0.00000482, -0.00000998, 0.00002199, -0.00001652, 0.00001228, -0.00001622, 0.00001207, -0.00001619, 0.00002248, -0.00002070, 0.00002480, -0.00002720, 0.00002237, -0.00002893, 0.00003080, -0.00002634, 0.00002033, -0.00002474, 0.00004353, -0.00004966, 0.00004106, -0.00004210, 0.00003614, -0.00002551, 0.00001319, -0.00001582, 0.00001248, -0.00001283, 0.00001524, -0.00002946, 0.00004819, -0.00006355, 0.00006126, -0.00007447, 0.00007233, -0.00008923, 0.00009378, -0.00012244, 0.00011817, -0.00018714, 0.00019499, -0.00043408, 0.00006176, 0.00111628, -0.00080544, 0.00027314, -0.00023584, 0.00014320, -0.00014204, 0.00010133, -0.00009337, 0.00007299, -0.00007086, 0.00005277, -0.00005159, 0.00003265, -0.00002627, 0.00002953, -0.00003961, 0.00004752, -0.00004893, 0.00004702, -0.00005716, 0.00006271, -0.00005923, 0.00005466, -0.00005392, 0.00003318, -0.00001814, 0.00002575, -0.00003196, 0.00002542, -0.00001875, 0.00001539, -0.00002061, 0.00001455, -0.00001148, 0.00001234, -0.00000677, 0.00000799, -0.00001368, 0.00000851, -0.00001184, 0.00000872, 0.00000350, -0.00000046, 0.00000548, -0.00000346, -0.00001686, 0.00001086, -0.00000015, 0.00000351, -0.00000587, 0.00000802, -0.00000805, 0.00000346, -0.00000208, 0.00000540, -0.00000526, 0.00000410, -0.00000688, 0.00000058, 0.00000000, 0.00000224, 0.00000822, },
		{ 0.00000753, -0.00000867, 0.00000041, -0.00000082, -0.00000814, 0.00000918, 0.00000301, -0.00000480, 0.00000679, -0.00000814, 0.00000224, -0.00000197, 0.00000053, -0.00000120, 0.00000441, -0.00000718, 0.00000678, -0.00000501, 0.00000213, -0.00000069, 0.00001006, -0.00001051, -0.00000955, 0.00000783, 0.00000062, -0.00000625, 0.00002013, -0.00001686, 0.00001277, -0.00001676, 0.00001280, -0.00001608, 0.00002292, -0.00002166, 0.00002616, -0.00002985, 0.00002550, -0.00003232, 0.00003546, -0.00003110, 0.00002428, -0.00002646, 0.00004544, -0.00005568, 0.00005173, -0.00005694, 0.00005491, -0.00004707, 0.00003284, -0.00003289, 0.00002766, -0.00002660, 0.00002583, -0.00003757, 0.00005664, -0.00007538, 0.00007639, -0.00009244, 0.00009316, -0.00011331, 0.00012216, -0.00015872, 0.00016198, -0.00024695, 0.00027990, -0.00062474, 0.00043303, 0.00082319, -0.00076528, 0.00030985, -0.00026410, 0.00016911, -0.00016332, 0.00012201, -0.00011143, 0.00008906, -0.00008541, 0.00006602, -0.00006379, 0.00004313, -0.00003264, 0.00003130, -0.00003825, 0.00004628, -0.00004721, 0.00004384, -0.00005298, 0.00006223, -0.00006268, 0.00006201, -0.00006599, 0.00004721, -0.00002934, 0.00003421, -0.00004083, 0.00003434, -0.00002721, 0.00002279, -0.00002801, 0.00002238, -0.00001867, 0.00001914, -0.00001296, 0.00001277, -0.00001882, 0.00001376, -0.00001723, 0.00001528, -0.00000280, 0.00000506, 0.00000211, -0.00000488, -0.00001553, 0.00001323, -0.00000203, 0.00000343, -0.00000506, 0.00000772, -0.00000896, 0.00000446, -0.00000228, 0.00000569, -0.00000552, 0.00000398, -0.00000716, 0.00000080, 0.00000014, 0.00000374, },
		{ 0.00000514, 0.00000583, -0.00001049, 0.00000400, -0.00000509, -0.00000448, 0.00000945, 0.00000121, -0.00000486, 0.00000790, -0.00001016, 0.00000474, -0.00000385, 0.00000166, -0.00000200, 0.00000481, -0.00000784, 0.00000861, -0.00000772, 0.00000447, -0.00000137, 0.00001042, -0.00001426, -0.00000590, 0.00000939, -0.00000410, -0.00000068, 0.00001563, -0.00001429, 0.00001032, -0.00001416, 0.00001021, -0.00001238, 0.00001932, -0.00001820, 0.00002264, -0.00002731, 0.00002325, -0.00003014, 0.00003452, -0.00003020, 0.00002216, -0.00002118, 0.00003860, -0.00005108, 0.00005058, -0.00005974, 0.00006230, -0.00005890, 0.00004452, -0.00004267, 0.00003546, -0.00003274, 0.00002829, -0.00003610, 0.00005348, -0.00007371, 0.00007650, -0.00009386, 0.00009558, -0.00011659, 0.00012651, -0.00016707, 0.00017251, -0.00026543, 0.00030949, -0.00075169, 0.00078393, 0.00047095, -0.00063666, 0.00027952, -0.00024525, 0.00015836, -0.00015459, 0.00011739, -0.00010801, 0.00008649, -0.00008356, 0.00006472, -0.00006318, 0.00004256, -0.00002951, 0.00002438, -0.00002809, 0.00003613, -0.00003675, 0.00003160, -0.00003828, 0.00004967, -0.00005331, 0.00005647, -0.00006634, 0.00005164, -0.00003267, 0.00003558, -0.00004317, 0.00003734, -0.00003012, 0.00002488, -0.00003029, 0.00002548, -0.00002159, 0.00002206, -0.00001563, 0.00001416, -0.00002063, 0.00001589, -0.00001968, 0.00001920, -0.00000708, 0.00000954, -0.00000141, -0.00000625, -0.00001305, 0.00001427, -0.00000330, 0.00000295, -0.00000348, 0.00000622, -0.00000870, 0.00000459, -0.00000171, 0.00000534, -0.00000539, 0.00000355, -0.00000717, 0.00000077, 0.00000074, },
		{ 0.00000165, 0.00000635, 0.00000315, -0.00001062, 0.00000625, -0.00000877, -0.00000033, 0.00000896, -0.00000082, -0.00000390, 0.00000799, -0.00001132, 0.00000664, -0.00000537, 0.00000233, -0.00000215, 0.00000436, -0.00000727, 0.00000901, -0.00000927, 0.00000618, -0.00000152, 0.00000970, -0.00001704, -0.00000184, 0.00000978, -0.00000844, 0.00000571, 0.00000939, -0.00000962, 0.00000566, -0.00000914, 0.00000505, -0.00000584, 0.00001247, -0.00001110, 0.00001502, -0.00002026, 0.00001626, -0.00002298, 0.00002863, -0.00002449, 0.00001512, -0.00001030, 0.00002457, -0.00003721, 0.00003825, -0.00005019, 0.00005705, -0.00005908, 0.00004621, -0.00004345, 0.00003451, -0.00003017, 0.00002186, -0.00002464, 0.00003844, -0.00005817, 0.00006099, -0.00007794, 0.00007860, -0.00009781, 0.00010510, -0.00014497, 0.00014595, -0.00023636, 0.00027109, -0.00078713, 0.00106924, 0.00010593, -0.00044991, 0.00019735, -0.00018761, 0.00011648, -0.00011984, 0.00009043, -0.00008544, 0.00006723, -0.00006702, 0.00005039, -0.00005111, 0.00003224, -0.00001831, 0.00001048, -0.00001111, 0.00001922, -0.00001994, 0.00001312, -0.00001614, 0.00002787, -0.00003329, 0.00003926, -0.00005522, 0.00004610, -0.00002758, 0.00002930, -0.00003848, 0.00003400, -0.00002718, 0.00002138, -0.00002714, 0.00002346, -0.00001980, 0.00002062, -0.00001435, 0.00001174, -0.00001866, 0.00001441, -0.00001859, 0.00001978, -0.00000844, 0.00001197, -0.00000429, -0.00000779, -0.00000941, 0.00001384, -0.00000380, 0.00000213, -0.00000137, 0.00000377, -0.00000737, 0.00000390, -0.00000042, 0.00000435, -0.00000488, 0.00000290, -0.00000709, 0.00000074, },
		{ 0.00000088, 0.00000275, 0.00000724, -0.00000027, -0.00000898, 0.00000665, -0.00001113, 0.00000363, 0.00000796, -0.00000307, -0.00000185, 0.00000677, -0.00001126, 0.00000759, -0.00000627, 0.00000242, -0.00000159, 0.00000306, -0.00000547, 0.00000783, -0.00000938, 0.00000694, -0.00000098, 0.00000781, -0.00001853, 0.00000235, 0.00000877, -0.00001153, 0.00001176, 0.00000262, -0.00000399, -0.00000004, -0.00000294, -0.00000142, 0.00000216, 0.00000383, -0.00000191, 0.00000487, -0.00001029, 0.00000606, -0.00001235, 0.00001937, -0.00001577, 0.00000536, 0.00000347, 0.00000651, -0.00001740, 0.00001781, -0.00003068, 0.00004062, -0.00004825, 0.00003822, -0.00003568, 0.00002554, -0.00001993, 0.00000802, -0.00000519, 0.00001390, -0.00003134, 0.00003259, -0.00004754, 0.00004530, -0.00006034, 0.00006154, -0.00009621, 0.00008611, -0.00016305, 0.00016407, -0.00071616, 0.00125265, -0.00022649, -0.00024068, 0.00008570, -0.00010605, 0.00005457, -0.00006787, 0.00004833, -0.00004974, 0.00003648, -0.00004037, 0.00002714, -0.00003127, 0.00001552, -0.00000226, -0.00000706, 0.00000921, -0.00000091, -0.00000048, -0.00000744, 0.00000878, 0.00000157, -0.00000705, 0.00001402, -0.00003524, 0.00003239, -0.00001546, 0.00001651, -0.00002772, 0.00002525, -0.00001930, 0.00001318, -0.00001937, 0.00001702, -0.00001389, 0.00001537, -0.00000961, 0.00000597, -0.00001328, 0.00000961, -0.00001418, 0.00001705, -0.00000660, 0.00001177, -0.00000600, -0.00000962, -0.00000479, 0.00001200, -0.00000350, 0.00000107, 0.00000095, 0.00000072, -0.00000526, 0.00000260, 0.00000145, 0.00000276, -0.00000402, 0.00000208, -0.00000702, },
		{ -0.00000696, 0.00000130, 0.00000393, 0.00000760, -0.00000396, -0.00000589, 0.00000511, -0.00001169, 0.00000684, 0.00000670, -0.00000541, 0.00000115, 0.00000424, -0.00000982, 0.00000739, -0.00000641, 0.00000193, -0.00000042, 0.00000110, -0.00000267, 0.00000520, -0.00000798, 0.00000663, 0.00000025, 0.00000476, -0.00001852, 0.00000628, 0.00000637, -0.00001279, 0.00001646, -0.00000350, 0.00000135, -0.00000549, 0.00000307, -0.00000769, 0.00001001, -0.00000484, 0.00000749, -0.00000579, 0.00000054, -0.00000524, -0.00000033, 0.00000882, -0.00000634, -0.00000442, 0.00001682, -0.00001159, 0.00000378, -0.00000598, -0.00000565, 0.00001680, -0.00002937, 0.00002298, -0.00002175, 0.00001116, -0.00000492, -0.00000988, 0.00001825, -0.00001551, 0.00000167, -0.00000322, -0.00000863, 0.00000226, -0.00001155, 0.00000417, -0.00003025, 0.00000389, -0.00005821, 0.00000075, -0.00053991, 0.00131245, -0.00048885, -0.00004359, -0.00003076, -0.00001839, -0.00001341, -0.00001011, 0.00000064, -0.00000905, 0.00000132, -0.00000988, 0.00000056, -0.00000866, -0.00000310, 0.00001446, -0.00002413, 0.00002873, -0.00002017, 0.00001749, -0.00002553, 0.00003133, -0.00002364, 0.00001981, -0.00001406, -0.00001072, 0.00001398, 0.00000080, -0.00000028, -0.00001308, 0.00001311, -0.00000846, 0.00000217, -0.00000874, 0.00000776, -0.00000533, 0.00000765, -0.00000268, -0.00000192, -0.00000565, 0.00000252, -0.00000737, 0.00001176, -0.00000203, 0.00000898, -0.00000637, -0.00001164, 0.00000049, 0.00000895, -0.00000246, -0.00000006, 0.00000313, -0.00000249, -0.00000271, 0.00000097, 0.00000368, 0.00000070, -0.00000281, 0.00000108, },
		{ -0.00000016, -0.00000684, 0.00000201, 0.00000515, 0.00000718, -0.00000731, -0.00000199, 0.00000198, -0.00001038, 0.00000898, 0.00000528, -0.00000757, 0.00000473, 0.00000063, -0.00000711, 0.00000603, -0.00000582, 0.00000098, 0.00000112, -0.00000119, 0.00000069, 0.00000151, -0.00000528, 0.00000533, 0.00000205, 0.00000074, -0.00001689, 0.00000950, 0.00000290, -0.00001200, 0.00001911, -0.00000804, 0.00000538, -0.00000956, 0.00000764, -0.00001240, 0.00001620, -0.00001189, 0.00001526, -0.00001499, 0.00001013, -0.00001550, 0.00001095, -0.00000094, 0.00000162, -0.00001169, 0.00002667, -0.00002585, 0.00002160, -0.00002779, 0.00001938, -0.00000923, -0.00000690, 0.00000438, -0.00000537, -0.00000481, 0.00001082, -0.00002736, 0.00004054, -0.00004393, 0.00003437, -0.00003931, 0.00003099, -0.00004185, 0.00003872, -0.00005568, 0.00003964, -0.00008473, 0.00005811, -0.00019509, -0.00027593, 0.00124461, -0.00065681, 0.00011334, -0.00012988, 0.00005839, -0.00007375, 0.00004197, -0.00004288, 0.00002824, -0.00003092, 0.00001797, -0.00002358, 0.00001157, -0.00001907, 0.00002777, -0.00003683, 0.00004362, -0.00003490, 0.00003044, -0.00003734, 0.00004700, -0.00004262, 0.00004169, -0.00003940, 0.00001331, -0.00000484, 0.00001744, -0.00001773, 0.00000253, 0.00000018, 0.00000287, -0.00000924, 0.00000243, -0.00000214, 0.00000387, -0.00000068, 0.00000466, -0.00001021, 0.00000260, -0.00000533, 0.00000039, 0.00000526, 0.00000420, 0.00000416, -0.00000560, -0.00001355, 0.00000589, 0.00000512, -0.00000089, -0.00000110, 0.00000486, -0.00000543, -0.00000014, -0.00000065, 0.00000595, -0.00000165, -0.00000130, },
		{ 0.00000041, -0.00000166, -0.00000654, 0.00000294, 0.00000637, 0.00000581, -0.00000969, 0.00000191, -0.00000207, -0.00000750, 0.00001000, 0.00000368, -0.00000917, 0.00000832, -0.00000354, -0.00000351, 0.00000376, -0.00000465, -0.00000018, 0.00000270, -0.00000335, 0.00000400, -0.00000262, -0.00000174, 0.00000331, 0.00000411, -0.00000388, -0.00001371, 0.00001161, -0.00000115, -0.00000932, 0.00001947, -0.00001047, 0.00000745, -0.00001152, 0.00000993, -0.00001460, 0.00001965, -0.00001608, 0.00002001, -0.00002115, 0.00001679, -0.00002294, 0.00001977, -0.00000832, 0.00000657, -0.00001477, 0.00003090, -0.00003339, 0.00003225, -0.00004289, 0.00003911, -0.00003212, 0.00001421, -0.00001322, 0.00000937, -0.00001831, 0.00002311, -0.00003996, 0.00005662, -0.00006553, 0.00006021, -0.00006850, 0.00006337, -0.00007811, 0.00008027, -0.00010605, 0.00009917, -0.00016202, 0.00016229, -0.00039195, 0.00004443, 0.00106285, -0.00072186, 0.00021263, -0.00019577, 0.00011142, -0.00011569, 0.00007916, -0.00007427, 0.00005525, -0.00005424, 0.00003790, -0.00004060, 0.00002533, -0.00002883, 0.00003459, -0.00004236, 0.00005120, -0.00004263, 0.00003615, -0.00004056, 0.00005294, -0.00005175, 0.00005425, -0.00005726, 0.00003226, -0.00001994, 0.00003077, -0.00003252, 0.00001617, -0.00001089, 0.00001221, -0.00001866, 0.00001187, -0.00001053, 0.00001168, -0.00000771, 0.00001058, -0.00001710, 0.00000972, -0.00001224, 0.00000746, -0.00000084, 0.00001063, -0.00000167, -0.00000419, -0.00001488, 0.00001078, 0.00000106, 0.00000094, -0.00000186, 0.00000586, -0.00000771, 0.00000206, -0.00000191, 0.00000794, -0.00000405, },
		{ -0.00000623, 0.00000219, -0.00000344, -0.00000592, 0.00000400, 0.00000751, 0.00000345, -0.00001061, 0.00000497, -0.00000619, -0.00000362, 0.00001006, 0.00000180, -0.00000981, 0.00001128, -0.00000762, 0.00000044, 0.00000095, -0.00000321, -0.00000126, 0.00000396, -0.00000493, 0.00000665, -0.00000647, 0.00000204, 0.00000102, 0.00000601, -0.00000861, -0.00000927, 0.00001230, -0.00000515, -0.00000526, 0.00001774, -0.00001075, 0.00000741, -0.00001116, 0.00000965, -0.00001393, 0.00001990, -0.00001686, 0.00002107, -0.00002346, 0.00001956, -0.00002651, 0.00002511, -0.00001252, 0.00000794, -0.00001316, 0.00002881, -0.00003305, 0.00003371, -0.00004825, 0.00004961, -0.00004746, 0.00002971, -0.00002603, 0.00001903, -0.00002602, 0.00002865, -0.00004424, 0.00006263, -0.00007587, 0.00007411, -0.00008507, 0.00008213, -0.00009933, 0.00010473, -0.00013699, 0.00013608, -0.00021213, 0.00023176, -0.00055604, 0.00038089, 0.00079581, -0.00069109, 0.00024891, -0.00022099, 0.00013398, -0.00013331, 0.00009619, -0.00008889, 0.00006792, -0.00006510, 0.00004684, -0.00004784, 0.00003032, -0.00003055, 0.00003351, -0.00003954, 0.00005042, -0.00004254, 0.00003419, -0.00003494, 0.00004852, -0.00004972, 0.00005533, -0.00006467, 0.00004292, -0.00002829, 0.00003794, -0.00004199, 0.00002551, -0.00001808, 0.00001769, -0.00002427, 0.00001779, -0.00001572, 0.00001650, -0.00001193, 0.00001364, -0.00002111, 0.00001426, -0.00001681, 0.00001239, -0.00000502, 0.00001574, -0.00000733, -0.00000281, -0.00001514, 0.00001450, -0.00000264, 0.00000266, -0.00000218, 0.00000596, -0.00000903, 0.00000359, -0.00000249, 0.00000932, },
		{ 0.00000980, -0.00000789, 0.00000381, -0.00000539, -0.00000492, 0.00000515, 0.00000841, 0.00000024, -0.00000983, 0.00000651, -0.00000950, 0.00000056, 0.00000946, -0.00000044, -0.00000916, 0.00001300, -0.00001090, 0.00000410, -0.00000191, -0.00000183, -0.00000197, 0.00000460, -0.00000553, 0.00000809, -0.00000935, 0.00000537, -0.00000103, 0.00000728, -0.00001282, -0.00000405, 0.00001150, -0.00000850, -0.00000050, 0.00001444, -0.00000927, 0.00000560, -0.00000879, 0.00000710, -0.00001064, 0.00001717, -0.00001439, 0.00001852, -0.00002189, 0.00001832, -0.00002600, 0.00002678, -0.00001362, 0.00000618, -0.00000762, 0.00002130, -0.00002554, 0.00002614, -0.00004311, 0.00004914, -0.00005275, 0.00003696, -0.00003169, 0.00002159, -0.00002615, 0.00002584, -0.00003868, 0.00005688, -0.00007280, 0.00007346, -0.00008600, 0.00008383, -0.00010153, 0.00010739, -0.00014262, 0.00014281, -0.00022462, 0.00024959, -0.00065716, 0.00069075, 0.00048180, -0.00058452, 0.00022823, -0.00020680, 0.00012610, -0.00012615, 0.00009237, -0.00008594, 0.00006555, -0.00006293, 0.00004436, -0.00004501, 0.00002644, -0.00002436, 0.00002498, -0.00002903, 0.00004198, -0.00003552, 0.00002592, -0.00002220, 0.00003538, -0.00003770, 0.00004524, -0.00006101, 0.00004410, -0.00002855, 0.00003758, -0.00004475, 0.00002928, -0.00002036, 0.00001839, -0.00002519, 0.00001932, -0.00001682, 0.00001750, -0.00001256, 0.00001306, -0.00002147, 0.00001537, -0.00001817, 0.00001429, -0.00000624, 0.00001829, -0.00001172, -0.00000208, -0.00001392, 0.00001652, -0.00000542, 0.00000398, -0.00000198, 0.00000508, -0.00000926, 0.00000423, -0.00000219, },
		{ -0.00000092, 0.00000922, -0.00000876, 0.00000502, -0.00000737, -0.00000354, 0.00000633, 0.00000887, -0.00000343, -0.00000745, 0.00000620, -0.00001134, 0.00000437, 0.00000848, -0.00000308, -0.00000707, 0.00001307, -0.00001282, 0.00000688, -0.00000435, -0.00000079, -0.00000208, 0.00000444, -0.00000498, 0.00000798, -0.00001072, 0.00000767, -0.00000237, 0.00000746, -0.00001588, 0.00000134, 0.00000933, -0.00001074, 0.00000421, 0.00001031, -0.00000668, 0.00000266, -0.00000512, 0.00000302, -0.00000551, 0.00001226, -0.00000950, 0.00001318, -0.00001720, 0.00001374, -0.00002196, 0.00002534, -0.00001240, 0.00000252, 0.00000011, 0.00001057, -0.00001329, 0.00001179, -0.00002912, 0.00003841, -0.00004785, 0.00003537, -0.00002975, 0.00001695, -0.00001888, 0.00001515, -0.00002401, 0.00004016, -0.00005700, 0.00005874, -0.00007161, 0.00006869, -0.00008484, 0.00008816, -0.00012247, 0.00011811, -0.00019671, 0.00020815, -0.00067420, 0.00093566, 0.00016237, -0.00043037, 0.00016535, -0.00016174, 0.00009373, -0.00009872, 0.00007121, -0.00006828, 0.00005057, -0.00004992, 0.00003248, -0.00003405, 0.00001556, -0.00001218, 0.00001109, -0.00001302, 0.00002803, -0.00002378, 0.00001398, -0.00000555, 0.00001694, -0.00001891, 0.00002660, -0.00004798, 0.00003672, -0.00002125, 0.00002995, -0.00004083, 0.00002753, -0.00001787, 0.00001455, -0.00002164, 0.00001661, -0.00001399, 0.00001478, -0.00000972, 0.00000898, -0.00001819, 0.00001301, -0.00001623, 0.00001299, -0.00000411, 0.00001759, -0.00001408, -0.00000243, -0.00001104, 0.00001652, -0.00000689, 0.00000465, -0.00000124, 0.00000332, -0.00000840, 0.00000397, },
		{ 0.00000293, 0.00000121, 0.00000751, -0.00000867, 0.00000559, -0.00000914, -0.00000186, 0.00000755, 0.00000861, -0.00000702, -0.00000392, 0.00000411, -0.00001138, 0.00000733, 0.00000733, -0.00000597, -0.00000365, 0.00001135, -0.00001306, 0.00000837, -0.00000604, -0.00000030, -0.00000154, 0.00000350, -0.00000333, 0.00000628, -0.00001033, 0.00000858, -0.00000268, 0.00000629, -0.00001728, 0.00000620, 0.00000615, -0.00001165, 0.00000826, 0.00000609, -0.00000374, -0.00000059, -0.00000103, -0.00000163, 0.00000041, 0.00000627, -0.00000335, 0.00000628, -0.00001062, 0.00000701, -0.00001550, 0.00002181, -0.00001006, -0.00000140, 0.00000779, -0.00000052, 0.00000026, -0.00000558, -0.00000983, 0.00002032, -0.00003478, 0.00002641, -0.00002166, 0.00000681, -0.00000623, -0.00000103, -0.00000309, 0.00001565, -0.00003177, 0.00003328, -0.00004536, 0.00004042, -0.00005333, 0.00005154, -0.00008145, 0.00006736, -0.00013413, 0.00011090, -0.00059886, 0.00108744, -0.00012428, -0.00025948, 0.00007981, -0.00009890, 0.00004674, -0.00005891, 0.00003921, -0.00004135, 0.00002767, -0.00003021, 0.00001498, -0.00001842, 0.00000093, 0.00000287, -0.00000503, 0.00000537, 0.00001154, -0.00001023, 0.00000161, 0.00001115, -0.00000246, 0.00000212, 0.00000366, -0.00002900, 0.00002344, -0.00000858, 0.00001684, -0.00003163, 0.00002144, -0.00001182, 0.00000737, -0.00001478, 0.00001076, -0.00000821, 0.00000931, -0.00000436, 0.00000231, -0.00001211, 0.00000790, -0.00001165, 0.00000911, 0.00000092, 0.00001366, -0.00001415, -0.00000392, -0.00000665, 0.00001454, -0.00000688, 0.00000459, -0.00000006, 0.00000089, -0.00000663, },
		{ -0.00000427, 0.00000138, 0.00000396, 0.00000482, -0.00000756, 0.00000532, -0.00001048, 0.00000002, 0.00000881, 0.00000741, -0.00000991, 0.00000007, 0.00000067, -0.00000966, 0.00000918, 0.00000608, -0.00000881, 0.00000070, 0.00000805, -0.00001158, 0.00000842, -0.00000682, -0.00000036, -0.00000045, 0.00000198, -0.00000088, 0.00000324, -0.00000827, 0.00000802, -0.00000191, 0.00000371, -0.00001671, 0.00000989, 0.00000248, -0.00001126, 0.00001125, 0.00000236, -0.00000111, -0.00000342, 0.00000260, -0.00000586, 0.00000603, 0.00000039, 0.00000276, -0.00000080, -0.00000360, -0.00000043, -0.00000792, 0.00001733, -0.00000778, -0.00000407, 0.00001333, -0.00000915, 0.00001149, -0.00002169, 0.00001020, -0.00000084, -0.00001714, 0.00001308, -0.00001026, -0.00000582, 0.00000847, -0.00001893, 0.00001979, -0.00001184, -0.00000226, 0.00000252, -0.00001307, 0.00000533, -0.00001404, 0.00000552, -0.00002875, 0.00000137, -0.00005001, -0.00002782, -0.00043723, 0.00113205, -0.00034877, -0.00010003, -0.00000837, -0.00003270, -0.00000358, -0.00001599, 0.00000411, -0.00001172, 0.00000254, -0.00000883, -0.00000365, -0.00000220, -0.00001372, 0.00001731, -0.00001996, 0.00002280, -0.00000440, 0.00000232, -0.00000823, 0.00002435, -0.00001864, 0.00002074, -0.00001881, -0.00000842, 0.00000794, 0.00000627, 0.00000102, -0.00001943, 0.00001297, -0.00000409, -0.00000126, -0.00000641, 0.00000346, -0.00000113, 0.00000262, 0.00000203, -0.00000546, -0.00000463, 0.00000133, -0.00000568, 0.00000387, 0.00000776, 0.00000718, -0.00001219, -0.00000625, -0.00000123, 0.00001091, -0.00000552, 0.00000384, 0.00000134, -0.00000189, },
		{ -0.00000460, -0.00000173, -0.00000027, 0.00000694, 0.00000144, -0.00000552, 0.00000413, -0.00001118, 0.00000195, 0.00001006, 0.00000516, -0.00001154, 0.00000368, -0.00000339, -0.00000654, 0.00000997, 0.00000465, -0.00001115, 0.00000537, 0.00000367, -0.00000870, 0.00000715, -0.00000675, -0.00000084, 0.00000095, 0.00000024, 0.00000185, -0.00000057, -0.00000495, 0.00000621, -0.00000027, -0.00000002, -0.00001415, 0.00001195, -0.00000111, -0.00000980, 0.00001304, -0.00000050, 0.00000077, -0.00000529, 0.00000513, -0.00000890, 0.00001048, -0.00000441, 0.00000775, -0.00000685, 0.00000259, -0.00000727, -0.00000042, 0.00001285, -0.00000640, -0.00000451, 0.00001538, -0.00001333, 0.00001754, -0.00003271, 0.00002646, -0.00002042, 0.00000081, -0.00000093, 0.00000103, -0.00001740, 0.00002142, -0.00003439, 0.00003988, -0.00003696, 0.00002569, -0.00002726, 0.00001852, -0.00002914, 0.00002464, -0.00004024, 0.00002425, -0.00006603, 0.00003773, -0.00018530, -0.00020867, 0.00107092, -0.00049404, 0.00002674, -0.00008239, 0.00002413, -0.00004698, 0.00002149, -0.00002681, 0.00001442, -0.00001949, 0.00000957, -0.00001928, 0.00001087, -0.00002503, 0.00002808, -0.00003081, 0.00003646, -0.00001729, 0.00001176, -0.00001352, 0.00003158, -0.00002852, 0.00003314, -0.00003654, 0.00000959, -0.00000602, 0.00001987, -0.00001443, -0.00000685, 0.00000436, 0.00000325, -0.00000934, 0.00000152, -0.00000342, 0.00000550, -0.00000357, 0.00000776, -0.00001266, 0.00000262, -0.00000517, 0.00000020, -0.00000118, 0.00001488, -0.00000066, -0.00000889, -0.00000885, 0.00000448, 0.00000626, -0.00000315, 0.00000263, 0.00000268, },
		{ 0.00000366, -0.00000686, 0.00000058, -0.00000162, 0.00000966, -0.00000222, -0.00000283, 0.00000208, -0.00001112, 0.00000387, 0.00001117, 0.00000192, -0.00001153, 0.00000616, -0.00000721, -0.00000265, 0.00000990, 0.00000290, -0.00001254, 0.00000963, -0.00000108, -0.00000498, 0.00000491, -0.00000606, -0.00000150, 0.00000233, -0.00000128, 0.00000426, -0.00000442, -0.00000101, 0.00000366, 0.00000179, -0.00000438, -0.00000993, 0.00001219, -0.00000413, -0.00000765, 0.00001371, -0.00000240, 0.00000172, -0.00000597, 0.00000624, -0.00001031, 0.00001322, -0.00000752, 0.00001093, -0.00001106, 0.00000703, -0.00001255, 0.00000615, 0.00000894, -0.00000616, -0.00000258, 0.00001361, -0.00001232, 0.00001694, -0.00003620, 0.00003557, -0.00003449, 0.00001522, -0.00001219, 0.00000906, -0.00002479, 0.00002933, -0.00004379, 0.00005308, -0.00005501, 0.00004683, -0.00005036, 0.00004317, -0.00005610, 0.00005489, -0.00007659, 0.00006655, -0.00012100, 0.00011005, -0.00033449, 0.00005752, 0.00091972, -0.00055648, 0.00010889, -0.00013129, 0.00006283, -0.00007623, 0.00004735, -0.00004826, 0.00003253, -0.00003452, 0.00002164, -0.00002899, 0.00001824, -0.00003075, 0.00003321, -0.00003571, 0.00004458, -0.00002562, 0.00001712, -0.00001362, 0.00003201, -0.00003071, 0.00003722, -0.00004665, 0.00002190, -0.00001544, 0.00002936, -0.00002681, 0.00000382, -0.00000248, 0.00000846, -0.00001518, 0.00000737, -0.00000829, 0.00001014, -0.00000781, 0.00001140, -0.00001780, 0.00000819, -0.00001022, 0.00000463, -0.00000456, 0.00002064, -0.00000845, -0.00000519, -0.00001096, 0.00000963, 0.00000139, -0.00000028, 0.00000127, },
		{ 0.00000010, 0.00000402, -0.00000833, 0.00000228, -0.00000226, 0.00001165, -0.00000566, 0.00000014, -0.00000062, -0.00001029, 0.00000573, 0.00001193, -0.00000202, -0.00000980, 0.00000699, -0.00001003, 0.00000134, 0.00000928, 0.00000070, -0.00001257, 0.00001280, -0.00000541, -0.00000108, 0.00000218, -0.00000507, -0.00000204, 0.00000337, -0.00000219, 0.00000579, -0.00000758, 0.00000276, 0.00000099, 0.00000367, -0.00000864, -0.00000467, 0.00001074, -0.00000624, -0.00000520, 0.00001350, -0.00000345, 0.00000183, -0.00000553, 0.00000593, -0.00001005, 0.00001414, -0.00000878, 0.00001208, -0.00001313, 0.00000933, -0.00001578, 0.00001135, 0.00000568, -0.00000672, 0.00000105, 0.00000885, -0.00000685, 0.00000998, -0.00003161, 0.00003593, -0.00004066, 0.00002348, -0.00001836, 0.00001178, -0.00002601, 0.00003017, -0.00004494, 0.00005681, -0.00006289, 0.00005760, -0.00006282, 0.00005652, -0.00007067, 0.00007110, -0.00009686, 0.00008988, -0.00015271, 0.00015077, -0.00044890, 0.00032711, 0.00070475, -0.00054462, 0.00014423, -0.00015096, 0.00007970, -0.00008808, 0.00005872, -0.00005767, 0.00004042, -0.00004075, 0.00002589, -0.00003156, 0.00001890, -0.00003007, 0.00003209, -0.00003414, 0.00004661, -0.00002907, 0.00001862, -0.00000927, 0.00002647, -0.00002576, 0.00003286, -0.00004819, 0.00002702, -0.00001869, 0.00003301, -0.00003437, 0.00001103, -0.00000635, 0.00001050, -0.00001777, 0.00001016, -0.00001025, 0.00001191, -0.00000925, 0.00001207, -0.00001995, 0.00001110, -0.00001289, 0.00000670, -0.00000522, 0.00002371, -0.00001484, -0.00000200, -0.00001190, 0.00001342, -0.00000288, 0.00000250, },
		{ 0.00000468, -0.00000063, 0.00000358, -0.00000881, 0.00000315, -0.00000193, 0.00001253, -0.00000838, 0.00000293, -0.00000366, -0.00000880, 0.00000755, 0.00001205, -0.00000616, -0.00000662, 0.00000599, -0.00001130, 0.00000482, 0.00000839, -0.00000199, -0.00001106, 0.00001435, -0.00000867, 0.00000232, -0.00000051, -0.00000411, -0.00000221, 0.00000384, -0.00000223, 0.00000602, -0.00000942, 0.00000567, -0.00000116, 0.00000477, -0.00001202, 0.00000081, 0.00000799, -0.00000735, -0.00000278, 0.00001271, -0.00000388, 0.00000136, -0.00000427, 0.00000450, -0.00000838, 0.00001346, -0.00000840, 0.00001141, -0.00001321, 0.00000956, -0.00001694, 0.00001515, 0.00000282, -0.00000733, 0.00000514, 0.00000277, 0.00000111, -0.00000141, -0.00002030, 0.00002803, -0.00003849, 0.00002470, -0.00001864, 0.00000867, -0.00002071, 0.00002363, -0.00003753, 0.00005056, -0.00005977, 0.00005681, -0.00006321, 0.00005694, -0.00007098, 0.00007104, -0.00009821, 0.00009046, -0.00015571, 0.00014993, -0.00050739, 0.00056663, 0.00045799, -0.00047600, 0.00013886, -0.00014367, 0.00007600, -0.00008333, 0.00005609, -0.00005538, 0.00003840, -0.00003851, 0.00002277, -0.00002753, 0.00001345, -0.00002363, 0.00002545, -0.00002682, 0.00004316, -0.00002828, 0.00001741, -0.00000229, 0.00001710, -0.00001583, 0.00002183, -0.00004211, 0.00002524, -0.00001574, 0.00003050, -0.00003654, 0.00001427, -0.00000694, 0.00000923, -0.00001700, 0.00000978, -0.00000916, 0.00001069, -0.00000777, 0.00000965, -0.00001889, 0.00001108, -0.00001293, 0.00000618, -0.00000277, 0.00002332, -0.00001888, -0.00000003, -0.00001122, 0.00001526, -0.00000589, },
		{ -0.00000725, 0.00000593, -0.00000076, 0.00000234, -0.00000826, 0.00000312, -0.00000056, 0.00001206, -0.00001000, 0.00000509, -0.00000666, -0.00000680, 0.00000936, 0.00001127, -0.00000989, -0.00000258, 0.00000337, -0.00001082, 0.00000741, 0.00000740, -0.00000502, -0.00000808, 0.00001408, -0.00001042, 0.00000474, -0.00000277, -0.00000343, -0.00000189, 0.00000366, -0.00000138, 0.00000485, -0.00000962, 0.00000723, -0.00000232, 0.00000461, -0.00001384, 0.00000564, 0.00000458, -0.00000759, -0.00000056, 0.00001156, -0.00000395, 0.00000061, -0.00000256, 0.00000239, -0.00000575, 0.00001163, -0.00000688, 0.00000943, -0.00001180, 0.00000815, -0.00001632, 0.00001775, -0.00000004, -0.00000712, 0.00000823, -0.00000260, 0.00000896, -0.00001423, -0.00000522, 0.00001419, -0.00002946, 0.00001976, -0.00001386, 0.00000079, -0.00001016, 0.00001116, -0.00002314, 0.00003594, -0.00004707, 0.00004573, -0.00005273, 0.00004566, -0.00005836, 0.00005613, -0.00008210, 0.00006967, -0.00013100, 0.00010585, -0.00049790, 0.00074853, 0.00021164, -0.00037296, 0.00010452, -0.00011643, 0.00005683, -0.00006600, 0.00004271, -0.00004405, 0.00002877, -0.00002990, 0.00001425, -0.00001883, 0.00000374, -0.00001323, 0.00001507, -0.00001542, 0.00003562, -0.00002454, 0.00001509, 0.00000505, 0.00000671, -0.00000404, 0.00000718, -0.00003087, 0.00001836, -0.00000798, 0.00002285, -0.00003392, 0.00001400, -0.00000485, 0.00000534, -0.00001356, 0.00000689, -0.00000570, 0.00000712, -0.00000402, 0.00000477, -0.00001514, 0.00000853, -0.00001073, 0.00000353, 0.00000243, 0.00001943, -0.00002015, 0.00000042, -0.00000881, 0.00001495, },
		{ 0.00001267, -0.00000690, 0.00000613, -0.00000032, 0.00000045, -0.00000685, 0.00000233, 0.00000175, 0.00001023, -0.00001030, 0.00000627, -0.00000925, -0.00000449, 0.00001117, 0.00000935, -0.00001258, 0.00000158, -0.00000032, -0.00000875, 0.00000894, 0.00000633, -0.00000808, -0.00000400, 0.00001211, -0.00001054, 0.00000591, -0.00000433, -0.00000312, -0.00000109, 0.00000294, 0.00000016, 0.00000246, -0.00000817, 0.00000727, -0.00000229, 0.00000301, -0.00001367, 0.00000906, 0.00000119, -0.00000724, 0.00000144, 0.00001021, -0.00000384, -0.00000013, -0.00000082, 0.00000007, -0.00000267, 0.00000918, -0.00000477, 0.00000676, -0.00000954, 0.00000571, -0.00001436, 0.00001942, -0.00000319, -0.00000548, 0.00000915, -0.00000543, 0.00001409, -0.00002515, 0.00000996, -0.00000212, -0.00001639, 0.00001088, -0.00000611, -0.00000957, 0.00000310, -0.00000446, -0.00000482, 0.00001616, -0.00002807, 0.00002763, -0.00003475, 0.00002630, -0.00003682, 0.00003088, -0.00005368, 0.00003346, -0.00008575, 0.00002557, -0.00041933, 0.00085512, -0.00000678, -0.00025814, 0.00005529, -0.00007863, 0.00002934, -0.00004189, 0.00002341, -0.00002775, 0.00001503, -0.00001802, 0.00000319, -0.00000814, -0.00000770, -0.00000124, 0.00000322, -0.00000204, 0.00002576, -0.00001922, 0.00001314, 0.00001068, -0.00000201, 0.00000638, -0.00000757, -0.00001771, 0.00000911, 0.00000224, 0.00001200, -0.00002797, 0.00001141, -0.00000128, 0.00000009, -0.00000870, 0.00000272, -0.00000105, 0.00000238, 0.00000085, -0.00000144, -0.00000975, 0.00000440, -0.00000724, -0.00000020, 0.00000933, 0.00001268, -0.00001882, -0.00000050, -0.00000498, },
		{ -0.00000037, 0.00000896, -0.00000513, 0.00000542, 0.00000052, -0.00000181, -0.00000486, 0.00000110, 0.00000466, 0.00000725, -0.00000929, 0.00000630, -0.00001113, -0.00000203, 0.00001291, 0.00000627, -0.00001372, 0.00000504, -0.00000432, -0.00000553, 0.00000950, 0.00000509, -0.00001073, 0.00000060, 0.00000887, -0.00000923, 0.00000584, -0.00000518, -0.00000312, 0.00000002, 0.00000197, 0.00000198, -0.00000068, -0.00000544, 0.00000596, -0.00000121, 0.00000013, -0.00001144, 0.00001061, -0.00000156, -0.00000668, 0.00000329, 0.00000870, -0.00000364, -0.00000069, 0.00000068, -0.00000209, 0.00000043, 0.00000656, -0.00000257, 0.00000397, -0.00000702, 0.00000282, -0.00001150, 0.00002029, -0.00000663, -0.00000223, 0.00000740, -0.00000467, 0.00001465, -0.00003136, 0.00002169, -0.00001712, -0.00000272, 0.00000099, 0.00000192, -0.00001956, 0.00001595, -0.00001983, 0.00001369, -0.00000472, -0.00000699, 0.00000687, -0.00001390, 0.00000386, -0.00001191, 0.00000166, -0.00002038, -0.00000917, -0.00003155, -0.00007658, -0.00028149, 0.00088045, -0.00017794, -0.00015081, 0.00000449, -0.00003971, 0.00000091, -0.00001706, 0.00000328, -0.00001077, 0.00000087, -0.00000613, -0.00000747, 0.00000181, -0.00001832, 0.00000998, -0.00000785, 0.00001124, 0.00001521, -0.00001345, 0.00001250, 0.00001328, -0.00000718, 0.00001290, -0.00001933, -0.00000575, 0.00000034, 0.00001227, 0.00000030, -0.00002054, 0.00000798, 0.00000236, -0.00000509, -0.00000382, -0.00000137, 0.00000342, -0.00000218, 0.00000547, -0.00000759, -0.00000403, -0.00000013, -0.00000362, -0.00000371, 0.00001649, 0.00000426, -0.00001557, -0.00000229, },
		{ -0.00000423, 0.00000420, 0.00000461, -0.00000244, 0.00000409, 0.00000147, -0.00000406, -0.00000271, -0.00000018, 0.00000770, 0.00000354, -0.00000718, 0.00000516, -0.00001215, 0.00000047, 0.00001440, 0.00000219, -0.00001305, 0.00000713, -0.00000783, -0.00000176, 0.00000934, 0.00000349, -0.00001250, 0.00000503, 0.00000500, -0.00000694, 0.00000480, -0.00000543, -0.00000327, 0.00000117, 0.00000108, 0.00000355, -0.00000389, -0.00000205, 0.00000379, 0.00000050, -0.00000354, -0.00000749, 0.00001016, -0.00000328, -0.00000624, 0.00000514, 0.00000698, -0.00000334, -0.00000104, 0.00000179, -0.00000383, 0.00000326, 0.00000407, -0.00000058, 0.00000143, -0.00000467, -0.00000007, -0.00000804, 0.00002033, -0.00001001, 0.00000215, 0.00000330, -0.00000030, 0.00001001, -0.00003126, 0.00002738, -0.00002764, 0.00000842, -0.00000713, 0.00000764, -0.00002649, 0.00002546, -0.00003175, 0.00002882, -0.00002276, 0.00001196, -0.00001212, 0.00000516, -0.00001656, 0.00001065, -0.00002494, 0.00000992, -0.00004841, 0.00001820, -0.00018182, -0.00010299, 0.00083015, -0.00029236, -0.00006433, -0.00003767, -0.00000716, -0.00002251, 0.00000351, -0.00001344, 0.00000329, -0.00001066, 0.00000315, -0.00001540, 0.00000886, -0.00002608, 0.00001856, -0.00001635, 0.00002277, 0.00000523, -0.00000782, 0.00001334, 0.00001259, -0.00000815, 0.00001424, -0.00002608, 0.00000265, -0.00000562, 0.00001983, -0.00001013, -0.00001335, 0.00000505, 0.00000485, -0.00000898, -0.00000015, -0.00000418, 0.00000652, -0.00000539, 0.00000864, -0.00001244, 0.00000083, -0.00000395, -0.00000094, -0.00000573, 0.00002238, -0.00000441, -0.00001135, },
		{ -0.00000714, -0.00000559, 0.00000790, 0.00000047, 0.00000050, 0.00000255, 0.00000221, -0.00000594, -0.00000077, -0.00000109, 0.00001036, -0.00000036, -0.00000438, 0.00000307, -0.00001227, 0.00000298, 0.00001537, -0.00000248, -0.00001065, 0.00000746, -0.00001017, 0.00000191, 0.00000874, 0.00000143, -0.00001301, 0.00000861, 0.00000120, -0.00000426, 0.00000316, -0.00000534, -0.00000333, 0.00000210, 0.00000060, 0.00000438, -0.00000648, 0.00000125, 0.00000140, 0.00000221, -0.00000726, -0.00000251, 0.00000800, -0.00000385, -0.00000609, 0.00000707, 0.00000501, -0.00000286, -0.00000121, 0.00000250, -0.00000507, 0.00000567, 0.00000182, 0.00000107, -0.00000071, -0.00000270, -0.00000270, -0.00000413, 0.00001931, -0.00001268, 0.00000672, -0.00000205, 0.00000657, 0.00000092, -0.00002484, 0.00002602, -0.00003184, 0.00001485, -0.00001154, 0.00000931, -0.00002855, 0.00002962, -0.00003793, 0.00003797, -0.00003500, 0.00002553, -0.00002590, 0.00001874, -0.00003095, 0.00002631, -0.00004360, 0.00003076, -0.00007600, 0.00005132, -0.00027052, 0.00009228, 0.00071923, -0.00035019, -0.00000518, -0.00006536, 0.00001462, -0.00003735, 0.00001680, -0.00002416, 0.00001224, -0.00001772, 0.00000828, -0.00001928, 0.00001178, -0.00002984, 0.00002343, -0.00002122, 0.00003152, -0.00000346, -0.00000238, 0.00001506, 0.00000938, -0.00000557, 0.00001062, -0.00002729, 0.00000643, -0.00000753, 0.00002349, -0.00001779, -0.00000761, 0.00000348, 0.00000547, -0.00001084, 0.00000162, -0.00000502, 0.00000755, -0.00000653, 0.00000961, -0.00001515, 0.00000395, -0.00000631, 0.00000009, -0.00000538, 0.00002575, -0.00001196, },
		{ -0.00001740, -0.00000373, -0.00000584, 0.00001011, -0.00000272, 0.00000309, 0.00000120, 0.00000246, -0.00000715, 0.00000064, -0.00000127, 0.00001215, -0.00000390, -0.00000140, 0.00000034, -0.00001158, 0.00000553, 0.00001551, -0.00000716, -0.00000693, 0.00000602, -0.00001096, 0.00000500, 0.00000795, -0.00000110, -0.00001209, 0.00001088, -0.00000191, -0.00000175, 0.00000136, -0.00000516, -0.00000312, 0.00000262, 0.00000075, 0.00000413, -0.00000790, 0.00000377, -0.00000054, 0.00000327, -0.00001020, 0.00000258, 0.00000475, -0.00000347, -0.00000623, 0.00000905, 0.00000280, -0.00000216, -0.00000130, 0.00000286, -0.00000582, 0.00000766, -0.00000019, 0.00000243, -0.00000247, -0.00000113, -0.00000500, 0.00000016, 0.00001701, -0.00001389, 0.00001028, -0.00000704, 0.00001403, -0.00001063, -0.00001362, 0.00001826, -0.00002944, 0.00001584, -0.00001148, 0.00000634, -0.00002516, 0.00002770, -0.00003752, 0.00004002, -0.00004000, 0.00003206, -0.00003268, 0.00002490, -0.00003719, 0.00003270, -0.00005148, 0.00003855, -0.00008703, 0.00005952, -0.00032600, 0.00027941, 0.00056859, -0.00035935, 0.00002657, -0.00007724, 0.00002453, -0.00004273, 0.00002204, -0.00002815, 0.00001549, -0.00001988, 0.00000900, -0.00001893, 0.00001044, -0.00002941, 0.00002438, -0.00002219, 0.00003707, -0.00001058, 0.00000317, 0.00001658, 0.00000514, -0.00000109, 0.00000352, -0.00002382, 0.00000587, -0.00000533, 0.00002290, -0.00002207, -0.00000384, 0.00000350, 0.00000414, -0.00001062, 0.00000143, -0.00000385, 0.00000644, -0.00000550, 0.00000821, -0.00001546, 0.00000504, -0.00000695, -0.00000067, -0.00000239, 0.00002588, },
		{ 0.00002271, -0.00002024, -0.00000153, -0.00000477, 0.00001052, -0.00000449, 0.00000488, 0.00000032, 0.00000208, -0.00000753, 0.00000134, -0.00000056, 0.00001276, -0.00000659, 0.00000128, -0.00000262, -0.00001026, 0.00000813, 0.00001452, -0.00001118, -0.00000254, 0.00000313, -0.00001012, 0.00000721, 0.00000707, -0.00000395, -0.00000984, 0.00001163, -0.00000394, 0.00000014, -0.00000025, -0.00000507, -0.00000253, 0.00000266, 0.00000153, 0.00000273, -0.00000785, 0.00000503, -0.00000154, 0.00000317, -0.00001164, 0.00000686, 0.00000118, -0.00000255, -0.00000643, 0.00001094, 0.00000045, -0.00000122, -0.00000139, 0.00000295, -0.00000613, 0.00000923, -0.00000202, 0.00000361, -0.00000400, 0.00000015, -0.00000694, 0.00000475, 0.00001328, -0.00001308, 0.00001171, -0.00001003, 0.00001984, -0.00002198, -0.00000025, 0.00000621, -0.00002176, 0.00001207, -0.00000756, -0.00000052, -0.00001710, 0.00002042, -0.00003112, 0.00003542, -0.00003801, 0.00003161, -0.00003247, 0.00002372, -0.00003537, 0.00002990, -0.00004863, 0.00003316, -0.00008104, 0.00004015, -0.00033749, 0.00043657, 0.00040105, -0.00033264, 0.00003561, -0.00007562, 0.00002429, -0.00004003, 0.00002035, -0.00002635, 0.00001385, -0.00001793, 0.00000610, -0.00001519, 0.00000558, -0.00002542, 0.00002194, -0.00001958, 0.00003948, -0.00001609, 0.00000922, 0.00001680, 0.00000155, 0.00000319, -0.00000478, -0.00001756, 0.00000235, -0.00000007, 0.00001870, -0.00002324, -0.00000183, 0.00000472, 0.00000137, -0.00000882, -0.00000022, -0.00000120, 0.00000370, -0.00000277, 0.00000488, -0.00001369, 0.00000433, -0.00000619, -0.00000279, 0.00000284, },
		{ 0.00000937, 0.00001684, -0.00002055, -0.00000054, -0.00000258, 0.00000924, -0.00000470, 0.00000565, 0.00000002, 0.00000105, -0.00000710, 0.00000138, 0.00000104, 0.00001206, -0.00000812, 0.00000322, -0.00000542, -0.00000847, 0.00001075, 0.00001222, -0.00001391, 0.00000170, -0.00000058, -0.00000789, 0.00000846, 0.00000609, -0.00000679, -0.00000658, 0.00001095, -0.00000478, 0.00000119, -0.00000149, -0.00000514, -0.00000160, 0.00000234, 0.00000273, 0.00000038, -0.00000636, 0.00000489, -0.00000139, 0.00000169, -0.00001113, 0.00000954, -0.00000188, -0.00000166, -0.00000638, 0.00001245, -0.00000182, -0.00000016, -0.00000147, 0.00000281, -0.00000601, 0.00001038, -0.00000368, 0.00000469, -0.00000539, 0.00000123, -0.00000848, 0.00000943, 0.00000828, -0.00001007, 0.00001037, -0.00000978, 0.00002207, -0.00003042, 0.00001214, -0.00000715, -0.00001115, 0.00000533, -0.00000137, -0.00000956, -0.00000617, 0.00000960, -0.00002054, 0.00002595, -0.00003069, 0.00002578, -0.00002693, 0.00001697, -0.00002748, 0.00002021, -0.00003770, 0.00001774, -0.00006188, -0.00000323, -0.00030194, 0.00054822, 0.00023747, -0.00028462, 0.00002928, -0.00006499, 0.00001736, -0.00003213, 0.00001414, -0.00002076, 0.00000906, -0.00001342, 0.00000107, -0.00000949, -0.00000144, -0.00001904, 0.00001711, -0.00001413, 0.00003906, -0.00001997, 0.00001588, 0.00001494, -0.00000004, 0.00000535, -0.00001192, -0.00001083, -0.00000216, 0.00000654, 0.00001224, -0.00002216, -0.00000092, 0.00000636, -0.00000197, -0.00000634, -0.00000241, 0.00000202, 0.00000023, 0.00000076, 0.00000050, -0.00001060, 0.00000247, -0.00000468, -0.00000540, },
		{ -0.00000744, 0.00001592, 0.00000933, -0.00001887, -0.00000041, 0.00000026, 0.00000670, -0.00000355, 0.00000545, 0.00000024, -0.00000044, -0.00000605, 0.00000093, 0.00000329, 0.00001016, -0.00000840, 0.00000419, -0.00000774, -0.00000635, 0.00001327, 0.00000866, -0.00001491, 0.00000504, -0.00000435, -0.00000476, 0.00000890, 0.00000488, -0.00000920, -0.00000285, 0.00000922, -0.00000458, 0.00000141, -0.00000231, -0.00000531, -0.00000049, 0.00000189, 0.00000398, -0.00000245, -0.00000382, 0.00000356, -0.00000022, -0.00000103, -0.00000864, 0.00001018, -0.00000381, -0.00000129, -0.00000569, 0.00001326, -0.00000374, 0.00000084, -0.00000147, 0.00000244, -0.00000544, 0.00001104, -0.00000509, 0.00000567, -0.00000668, 0.00000213, -0.00000951, 0.00001382, 0.00000251, -0.00000521, 0.00000630, -0.00000590, 0.00001956, -0.00003391, 0.00002076, -0.00001876, -0.00000041, -0.00000207, 0.00000496, -0.00001854, 0.00000523, -0.00000234, -0.00000829, 0.00001414, -0.00002055, 0.00001707, -0.00001866, 0.00000750, -0.00001672, 0.00000730, -0.00002304, -0.00000240, -0.00003661, -0.00006178, -0.00022413, 0.00060688, 0.00009393, -0.00022869, 0.00001536, -0.00005045, 0.00000772, -0.00002230, 0.00000621, -0.00001373, 0.00000313, -0.00000816, -0.00000442, -0.00000344, -0.00000912, -0.00001160, 0.00001106, -0.00000673, 0.00003621, -0.00002206, 0.00002287, 0.00001093, 0.00000095, 0.00000422, -0.00001605, -0.00000571, -0.00000567, 0.00001263, 0.00000510, -0.00001998, -0.00000024, 0.00000753, -0.00000490, -0.00000414, -0.00000417, 0.00000478, -0.00000293, 0.00000405, -0.00000392, -0.00000713, 0.00000028, -0.00000327, },
		{ -0.00000268, -0.00000798, 0.00002120, 0.00000143, -0.00001599, -0.00000059, 0.00000306, 0.00000361, -0.00000151, 0.00000452, 0.00000076, -0.00000214, -0.00000465, 0.00000031, 0.00000582, 0.00000741, -0.00000757, 0.00000412, -0.00000942, -0.00000398, 0.00001543, 0.00000409, -0.00001404, 0.00000691, -0.00000746, -0.00000131, 0.00000876, 0.00000329, -0.00001078, 0.00000074, 0.00000698, -0.00000372, 0.00000101, -0.00000284, -0.00000541, 0.00000056, 0.00000160, 0.00000483, -0.00000514, -0.00000084, 0.00000152, 0.00000152, -0.00000448, -0.00000453, 0.00000877, -0.00000429, -0.00000178, -0.00000411, 0.00001311, -0.00000501, 0.00000159, -0.00000128, 0.00000175, -0.00000435, 0.00001106, -0.00000610, 0.00000640, -0.00000777, 0.00000277, -0.00000983, 0.00001742, -0.00000324, 0.00000061, 0.00000031, 0.00000108, 0.00001234, -0.00003155, 0.00002377, -0.00002616, 0.00000795, -0.00000794, 0.00000943, -0.00002533, 0.00001482, -0.00001298, 0.00000311, 0.00000260, -0.00001022, 0.00000814, -0.00001045, -0.00000166, -0.00000649, -0.00000490, -0.00000940, -0.00002130, -0.00001366, -0.00012327, -0.00011542, 0.00061328, -0.00001985, -0.00017513, 0.00000034, -0.00003637, -0.00000124, -0.00001341, -0.00000099, -0.00000736, -0.00000215, -0.00000370, -0.00000894, 0.00000157, -0.00001612, -0.00000430, 0.00000486, 0.00000177, 0.00003127, -0.00002203, 0.00002939, 0.00000543, 0.00000416, -0.00000031, -0.00001632, -0.00000349, -0.00000677, 0.00001670, -0.00000129, -0.00001771, 0.00000092, 0.00000756, -0.00000669, -0.00000297, -0.00000474, 0.00000627, -0.00000493, 0.00000622, -0.00000745, -0.00000415, -0.00000149, },
		{ -0.00000239, -0.00000326, -0.00000643, 0.00002419, -0.00000568, -0.00001273, -0.00000052, 0.00000515, 0.00000068, 0.00000083, 0.00000324, 0.00000131, -0.00000374, -0.00000322, -0.00000013, 0.00000816, 0.00000425, -0.00000595, 0.00000315, -0.00001039, -0.00000137, 0.00001692, -0.00000099, -0.00001148, 0.00000704, -0.00000937, 0.00000193, 0.00000828, 0.00000127, -0.00001123, 0.00000365, 0.00000478, -0.00000265, 0.00000029, -0.00000326, -0.00000527, 0.00000135, 0.00000172, 0.00000486, -0.00000707, 0.00000186, -0.00000057, 0.00000319, -0.00000793, 0.00000046, 0.00000572, -0.00000338, -0.00000320, -0.00000156, 0.00001184, -0.00000543, 0.00000188, -0.00000079, 0.00000070, -0.00000266, 0.00001030, -0.00000648, 0.00000668, -0.00000848, 0.00000302, -0.00000919, 0.00001964, -0.00000803, 0.00000617, -0.00000621, 0.00000971, 0.00000158, -0.00002385, 0.00002071, -0.00002808, 0.00001233, -0.00001076, 0.00001062, -0.00002842, 0.00002089, -0.00002050, 0.00001173, -0.00000660, -0.00000180, 0.00000108, -0.00000449, -0.00000811, 0.00000053, -0.00001324, -0.00000060, -0.00003395, -0.00000065, -0.00017475, 0.00000873, 0.00057508, -0.00010066, -0.00013011, -0.00001161, -0.00002549, -0.00000738, -0.00000722, -0.00000592, -0.00000296, -0.00000567, -0.00000099, -0.00001162, 0.00000464, -0.00002152, 0.00000198, -0.00000064, 0.00001064, 0.00002456, -0.00001944, 0.00003437, -0.00000034, 0.00000839, -0.00000727, -0.00001303, -0.00000445, -0.00000489, 0.00001794, -0.00000605, -0.00001597, 0.00000287, 0.00000617, -0.00000703, -0.00000315, -0.00000377, 0.00000608, -0.00000532, 0.00000675, -0.00000951, -0.00000223, },
		{ -0.00000152, -0.00000237, -0.00000496, -0.00000276, 0.00002439, -0.00001116, -0.00000973, 0.00000016, 0.00000607, -0.00000143, 0.00000290, 0.00000199, 0.00000161, -0.00000496, -0.00000204, -0.00000011, 0.00000991, 0.00000119, -0.00000396, 0.00000153, -0.00001069, 0.00000150, 0.00001738, -0.00000595, -0.00000773, 0.00000554, -0.00000981, 0.00000457, 0.00000763, -0.00000115, -0.00001046, 0.00000552, 0.00000308, -0.00000180, -0.00000044, -0.00000373, -0.00000475, 0.00000173, 0.00000239, 0.00000384, -0.00000777, 0.00000365, -0.00000205, 0.00000414, -0.00001053, 0.00000536, 0.00000177, -0.00000149, -0.00000530, 0.00000179, 0.00000951, -0.00000494, 0.00000165, 0.00000005, -0.00000073, -0.00000036, 0.00000864, -0.00000603, 0.00000627, -0.00000860, 0.00000271, -0.00000743, 0.00002005, -0.00001097, 0.00001020, -0.00001162, 0.00001802, -0.00001062, -0.00001251, 0.00001253, -0.00002459, 0.00001225, -0.00001003, 0.00000812, -0.00002727, 0.00002269, -0.00002400, 0.00001657, -0.00001237, 0.00000359, -0.00000300, -0.00000189, -0.00001068, 0.00000303, -0.00001620, 0.00000138, -0.00003768, -0.00000255, -0.00020521, 0.00013180, 0.00050454, -0.00015073, -0.00009585, -0.00001887, -0.00001874, -0.00001005, -0.00000428, -0.00000806, -0.00000098, -0.00000709, -0.00000028, -0.00001221, 0.00000547, -0.00002490, 0.00000676, -0.00000487, 0.00001927, 0.00001642, -0.00001399, 0.00003674, -0.00000491, 0.00001193, -0.00001493, -0.00000750, -0.00000783, -0.00000043, 0.00001636, -0.00000894, -0.00001491, 0.00000549, 0.00000359, -0.00000609, -0.00000449, -0.00000145, 0.00000431, -0.00000411, 0.00000560, -0.00000996, },
		{ -0.00000909, -0.00000182, -0.00000173, -0.00000727, 0.00000253, 0.00002185, -0.00001463, -0.00000732, 0.00000157, 0.00000563, -0.00000234, 0.00000429, 0.00000107, 0.00000151, -0.00000565, -0.00000125, 0.00000054, 0.00001077, -0.00000135, -0.00000204, -0.00000042, -0.00001044, 0.00000460, 0.00001654, -0.00001009, -0.00000351, 0.00000284, -0.00000884, 0.00000643, 0.00000685, -0.00000374, -0.00000863, 0.00000626, 0.00000211, -0.00000143, -0.00000096, -0.00000436, -0.00000383, 0.00000171, 0.00000354, 0.00000179, -0.00000703, 0.00000415, -0.00000248, 0.00000386, -0.00001159, 0.00000923, -0.00000215, 0.00000075, -0.00000762, 0.00000555, 0.00000640, -0.00000367, 0.00000093, 0.00000117, -0.00000242, 0.00000244, 0.00000613, -0.00000468, 0.00000503, -0.00000799, 0.00000178, -0.00000454, 0.00001846, -0.00001151, 0.00001168, -0.00001443, 0.00002395, -0.00002176, -0.00000009, 0.00000129, -0.00001700, 0.00000843, -0.00000626, 0.00000245, -0.00002232, 0.00002049, -0.00002364, 0.00001768, -0.00001469, 0.00000590, -0.00000417, -0.00000249, -0.00000958, 0.00000128, -0.00001408, -0.00000312, -0.00003271, -0.00002042, -0.00020772, 0.00023911, 0.00041568, -0.00017593, -0.00007146, -0.00002188, -0.00001549, -0.00000989, -0.00000404, -0.00000785, -0.00000104, -0.00000675, -0.00000125, -0.00001105, 0.00000425, -0.00002626, 0.00000989, -0.00000749, 0.00002711, 0.00000740, -0.00000569, 0.00003576, -0.00000706, 0.00001315, -0.00002125, -0.00000158, -0.00001218, 0.00000549, 0.00001268, -0.00001028, -0.00001419, 0.00000826, 0.00000040, -0.00000445, -0.00000643, 0.00000165, 0.00000151, -0.00000178, 0.00000318, },
		{ 0.00000018, -0.00000751, -0.00000260, -0.00000102, -0.00000940, 0.00000853, 0.00001717, -0.00001615, -0.00000549, 0.00000346, 0.00000400, -0.00000200, 0.00000480, 0.00000062, 0.00000094, -0.00000578, -0.00000086, 0.00000181, 0.00001062, -0.00000309, -0.00000051, -0.00000240, -0.00000970, 0.00000784, 0.00001428, -0.00001284, 0.00000042, -0.00000044, -0.00000677, 0.00000753, 0.00000585, -0.00000616, -0.00000610, 0.00000604, 0.00000188, -0.00000162, -0.00000120, -0.00000511, -0.00000259, 0.00000142, 0.00000492, -0.00000101, -0.00000502, 0.00000330, -0.00000170, 0.00000216, -0.00001070, 0.00001133, -0.00000520, 0.00000259, -0.00000954, 0.00000917, 0.00000294, -0.00000190, -0.00000008, 0.00000239, -0.00000415, 0.00000548, 0.00000295, -0.00000250, 0.00000297, -0.00000662, 0.00000032, -0.00000076, 0.00001507, -0.00000960, 0.00001018, -0.00001368, 0.00002591, -0.00002953, 0.00001070, -0.00001035, -0.00000742, 0.00000240, -0.00000078, -0.00000504, -0.00001485, 0.00001540, -0.00002038, 0.00001596, -0.00001441, 0.00000602, -0.00000342, -0.00000511, -0.00000617, -0.00000313, -0.00000880, -0.00001180, -0.00002209, -0.00005115, -0.00018065, 0.00032013, 0.00032167, -0.00018365, -0.00005432, -0.00002234, -0.00001421, -0.00000826, -0.00000528, -0.00000630, -0.00000229, -0.00000541, -0.00000316, -0.00000879, 0.00000150, -0.00002590, 0.00001149, -0.00000831, 0.00003358, -0.00000173, 0.00000490, 0.00003131, -0.00000617, 0.00001090, -0.00002451, 0.00000284, -0.00001579, 0.00001139, 0.00000801, -0.00001079, -0.00001324, 0.00001047, -0.00000260, -0.00000289, -0.00000815, 0.00000468, -0.00000154, 0.00000092, },
		{ 0.00000318, -0.00000262, -0.00000587, -0.00000332, -0.00000084, -0.00001052, 0.00001414, 0.00001127, -0.00001618, -0.00000397, 0.00000540, 0.00000165, -0.00000063, 0.00000448, 0.00000061, 0.00000000, -0.00000546, -0.00000073, 0.00000350, 0.00000956, -0.00000395, 0.00000041, -0.00000419, -0.00000853, 0.00001100, 0.00001071, -0.00001385, 0.00000338, -0.00000359, -0.00000407, 0.00000802, 0.00000451, -0.00000803, -0.00000339, 0.00000525, 0.00000215, -0.00000227, -0.00000124, -0.00000584, -0.00000125, 0.00000112, 0.00000612, -0.00000404, -0.00000217, 0.00000139, 0.00000009, -0.00000077, -0.00000786, 0.00001126, -0.00000675, 0.00000344, -0.00001050, 0.00001206, -0.00000035, -0.00000001, -0.00000111, 0.00000345, -0.00000561, 0.00000839, -0.00000052, 0.00000023, 0.00000029, -0.00000467, -0.00000141, 0.00000346, 0.00001045, -0.00000572, 0.00000597, -0.00000930, 0.00002317, -0.00003240, 0.00001759, -0.00001984, 0.00000177, -0.00000392, 0.00000475, -0.00001269, -0.00000650, 0.00000893, -0.00001567, 0.00001278, -0.00001284, 0.00000531, -0.00000228, -0.00000802, -0.00000243, -0.00000789, -0.00000318, -0.00002111, -0.00001073, -0.00008821, -0.00012772, 0.00036970, 0.00023283, -0.00018091, -0.00004138, -0.00002221, -0.00001317, -0.00000671, -0.00000665, -0.00000456, -0.00000377, -0.00000392, -0.00000524, -0.00000618, -0.00000214, -0.00002426, 0.00001177, -0.00000723, 0.00003810, -0.00000992, 0.00001674, 0.00002400, -0.00000245, 0.00000496, -0.00002374, 0.00000437, -0.00001723, 0.00001587, 0.00000353, -0.00001121, -0.00001151, 0.00001147, -0.00000470, -0.00000213, -0.00000890, 0.00000684, -0.00000401, },
		{ -0.00000530, 0.00000440, -0.00000461, -0.00000476, -0.00000354, -0.00000157, -0.00000999, 0.00001833, 0.00000518, -0.00001529, -0.00000236, 0.00000682, -0.00000080, 0.00000131, 0.00000357, 0.00000088, -0.00000111, -0.00000488, -0.00000063, 0.00000531, 0.00000786, -0.00000403, 0.00000067, -0.00000566, -0.00000692, 0.00001377, 0.00000618, -0.00001308, 0.00000493, -0.00000602, -0.00000124, 0.00000811, 0.00000273, -0.00000899, -0.00000102, 0.00000435, 0.00000255, -0.00000310, -0.00000124, -0.00000635, -0.00000004, 0.00000109, 0.00000671, -0.00000666, 0.00000085, -0.00000102, 0.00000239, -0.00000442, -0.00000349, 0.00000913, -0.00000655, 0.00000291, -0.00001007, 0.00001374, -0.00000297, 0.00000161, -0.00000184, 0.00000406, -0.00000645, 0.00001073, -0.00000383, 0.00000310, -0.00000267, -0.00000243, -0.00000305, 0.00000748, 0.00000548, -0.00000084, 0.00000002, -0.00000206, 0.00001609, -0.00002991, 0.00001929, -0.00002528, 0.00000860, -0.00000880, 0.00000876, -0.00001888, 0.00000108, 0.00000263, -0.00001091, 0.00000952, -0.00001132, 0.00000509, -0.00000220, -0.00000955, -0.00000025, -0.00001075, 0.00000001, -0.00002743, -0.00000407, -0.00012313, -0.00005692, 0.00038810, 0.00015558, -0.00017311, -0.00003027, -0.00002297, -0.00001113, -0.00000635, -0.00000710, -0.00000354, -0.00000470, -0.00000295, -0.00000686, -0.00000385, -0.00000610, -0.00002174, 0.00001099, -0.00000422, 0.00004012, -0.00001603, 0.00002837, 0.00001508, 0.00000304, -0.00000391, -0.00001899, 0.00000247, -0.00001571, 0.00001797, 0.00000010, -0.00001209, -0.00000867, 0.00001091, -0.00000545, -0.00000260, -0.00000820, 0.00000756, },
		{ 0.00000667, -0.00000516, 0.00000430, -0.00000545, -0.00000446, -0.00000307, -0.00000331, -0.00000757, 0.00002042, -0.00000024, -0.00001405, -0.00000036, 0.00000723, -0.00000272, 0.00000329, 0.00000241, 0.00000121, -0.00000219, -0.00000426, -0.00000033, 0.00000689, 0.00000586, -0.00000357, 0.00000039, -0.00000679, -0.00000482, 0.00001575, 0.00000124, -0.00001081, 0.00000496, -0.00000734, 0.00000131, 0.00000796, 0.00000051, -0.00000887, 0.00000062, 0.00000379, 0.00000269, -0.00000384, -0.00000141, -0.00000644, 0.00000081, 0.00000156, 0.00000634, -0.00000831, 0.00000333, -0.00000324, 0.00000451, -0.00000801, 0.00000162, 0.00000542, -0.00000477, 0.00000102, -0.00000813, 0.00001393, -0.00000457, 0.00000263, -0.00000204, 0.00000399, -0.00000639, 0.00001209, -0.00000649, 0.00000563, -0.00000547, -0.00000025, -0.00000420, 0.00001066, 0.00000110, 0.00000380, -0.00000626, 0.00000652, 0.00000599, -0.00002283, 0.00001573, -0.00002583, 0.00001185, -0.00001108, 0.00001022, -0.00002251, 0.00000674, -0.00000241, -0.00000713, 0.00000711, -0.00001071, 0.00000619, -0.00000408, -0.00000866, -0.00000087, -0.00001026, -0.00000110, -0.00002818, -0.00000650, -0.00014744, 0.00002128, 0.00038006, 0.00009236, -0.00016340, -0.00001983, -0.00002514, -0.00000763, -0.00000763, -0.00000617, -0.00000365, -0.00000471, -0.00000281, -0.00000770, -0.00000217, -0.00000998, -0.00001869, 0.00000942, 0.00000064, 0.00003924, -0.00001898, 0.00003818, 0.00000618, 0.00000867, -0.00001411, -0.00001133, -0.00000242, -0.00001126, 0.00001741, -0.00000187, -0.00001358, -0.00000478, 0.00000884, -0.00000484, -0.00000434, -0.00000599, },
		{ -0.00000263, 0.00000442, -0.00000376, 0.00000298, -0.00000521, -0.00000494, -0.00000206, -0.00000574, -0.00000345, 0.00002018, -0.00000442, -0.00001283, 0.00000217, 0.00000639, -0.00000363, 0.00000482, 0.00000132, 0.00000140, -0.00000304, -0.00000375, 0.00000036, 0.00000795, 0.00000396, -0.00000289, -0.00000026, -0.00000759, -0.00000221, 0.00001659, -0.00000344, -0.00000760, 0.00000366, -0.00000744, 0.00000334, 0.00000762, -0.00000199, -0.00000772, 0.00000134, 0.00000384, 0.00000228, -0.00000422, -0.00000189, -0.00000601, 0.00000120, 0.00000259, 0.00000483, -0.00000858, 0.00000470, -0.00000462, 0.00000578, -0.00001071, 0.00000652, 0.00000097, -0.00000195, -0.00000191, -0.00000489, 0.00001263, -0.00000502, 0.00000291, -0.00000159, 0.00000314, -0.00000530, 0.00001220, -0.00000808, 0.00000738, -0.00000772, 0.00000153, -0.00000456, 0.00001251, -0.00000185, 0.00000702, -0.00001132, 0.00001455, -0.00000510, -0.00001294, 0.00000803, -0.00002183, 0.00001132, -0.00001045, 0.00000882, -0.00002319, 0.00000995, -0.00000568, -0.00000473, 0.00000586, -0.00001120, 0.00000869, -0.00000799, -0.00000522, -0.00000443, -0.00000621, -0.00000689, -0.00002268, -0.00002015, -0.00015455, 0.00009613, 0.00035310, 0.00004241, -0.00015284, -0.00001014, -0.00002828, -0.00000306, -0.00001024, -0.00000403, -0.00000481, -0.00000388, -0.00000344, -0.00000778, -0.00000125, -0.00001357, -0.00001534, 0.00000735, 0.00000712, 0.00003539, -0.00001802, 0.00004479, -0.00000112, 0.00001265, -0.00002359, -0.00000252, -0.00000907, -0.00000470, 0.00001460, -0.00000256, -0.00001542, -0.00000030, 0.00000570, -0.00000326, -0.00000693, },
		{ -0.00000969, 0.00000120, 0.00000141, -0.00000161, 0.00000089, -0.00000422, -0.00000590, -0.00000091, -0.00000828, 0.00000171, 0.00001786, -0.00000722, -0.00001171, 0.00000508, 0.00000434, -0.00000332, 0.00000560, 0.00000056, 0.00000131, -0.00000359, -0.00000342, 0.00000149, 0.00000830, 0.00000244, -0.00000227, -0.00000107, -0.00000809, 0.00000087, 0.00001604, -0.00000723, -0.00000413, 0.00000151, -0.00000648, 0.00000480, 0.00000703, -0.00000449, -0.00000580, 0.00000122, 0.00000456, 0.00000120, -0.00000414, -0.00000272, -0.00000504, 0.00000116, 0.00000407, 0.00000230, -0.00000737, 0.00000463, -0.00000474, 0.00000571, -0.00001186, 0.00001028, -0.00000327, 0.00000116, -0.00000527, -0.00000080, 0.00001013, -0.00000439, 0.00000247, -0.00000055, 0.00000161, -0.00000321, 0.00001099, -0.00000840, 0.00000805, -0.00000910, 0.00000269, -0.00000401, 0.00001281, -0.00000289, 0.00000799, -0.00001387, 0.00002024, -0.00001495, -0.00000254, -0.00000184, -0.00001460, 0.00000773, -0.00000740, 0.00000501, -0.00002123, 0.00001088, -0.00000733, -0.00000347, 0.00000542, -0.00001230, 0.00001194, -0.00001315, -0.00000008, -0.00001001, 0.00000033, -0.00001612, -0.00001237, -0.00004414, -0.00014118, 0.00015864, 0.00031552, 0.00000299, -0.00014105, -0.00000214, -0.00003128, 0.00000157, -0.00001335, -0.00000133, -0.00000652, -0.00000261, -0.00000452, -0.00000734, -0.00000098, -0.00001680, -0.00001188, 0.00000515, 0.00001471, 0.00002890, -0.00001298, 0.00004733, -0.00000568, 0.00001351, -0.00003035, 0.00000539, -0.00001579, 0.00000264, 0.00001044, -0.00000253, -0.00001704, 0.00000405, 0.00000223, -0.00000141, },
		{ -0.00000004, -0.00001184, 0.00000467, -0.00000160, 0.00000061, -0.00000135, -0.00000302, -0.00000690, -0.00000009, -0.00001020, 0.00000703, 0.00001410, -0.00000883, -0.00001052, 0.00000797, 0.00000144, -0.00000188, 0.00000552, 0.00000020, 0.00000093, -0.00000384, -0.00000321, 0.00000295, 0.00000795, 0.00000144, -0.00000190, -0.00000190, -0.00000828, 0.00000424, 0.00001404, -0.00000963, -0.00000110, -0.00000089, -0.00000480, 0.00000577, 0.00000606, -0.00000658, -0.00000356, 0.00000055, 0.00000575, -0.00000045, -0.00000364, -0.00000379, -0.00000370, 0.00000088, 0.00000569, -0.00000089, -0.00000491, 0.00000317, -0.00000352, 0.00000415, -0.00001107, 0.00001220, -0.00000646, 0.00000375, -0.00000833, 0.00000348, 0.00000689, -0.00000299, 0.00000152, 0.00000087, -0.00000036, -0.00000038, 0.00000860, -0.00000745, 0.00000757, -0.00000949, 0.00000320, -0.00000265, 0.00001172, -0.00000203, 0.00000643, -0.00001321, 0.00002232, -0.00002166, 0.00000604, -0.00001154, -0.00000603, 0.00000244, -0.00000301, -0.00000020, -0.00001749, 0.00001023, -0.00000800, -0.00000269, 0.00000502, -0.00001305, 0.00001476, -0.00001818, 0.00000526, -0.00001590, 0.00000731, -0.00002630, -0.00000052, -0.00007468, -0.00010798, 0.00020303, 0.00027472, -0.00002919, -0.00012706, 0.00000293, -0.00003281, 0.00000507, -0.00001591, 0.00000110, -0.00000811, -0.00000144, -0.00000558, -0.00000675, -0.00000117, -0.00001967, -0.00000849, 0.00000327, 0.00002269, 0.00002058, -0.00000431, 0.00004568, -0.00000706, 0.00001051, -0.00003298, 0.00001066, -0.00002091, 0.00000931, 0.00000605, -0.00000250, -0.00001781, 0.00000752, -0.00000077, },
		{ -0.00000267, 0.00000023, -0.00001270, 0.00000708, -0.00000392, 0.00000227, -0.00000313, -0.00000214, -0.00000751, -0.00000003, -0.00001084, 0.00001155, 0.00000974, -0.00000966, -0.00000894, 0.00001029, -0.00000173, 0.00000033, 0.00000472, 0.00000021, 0.00000034, -0.00000390, -0.00000295, 0.00000452, 0.00000704, 0.00000096, -0.00000186, -0.00000265, -0.00000808, 0.00000762, 0.00001078, -0.00001042, 0.00000097, -0.00000297, -0.00000283, 0.00000644, 0.00000456, -0.00000789, -0.00000152, -0.00000021, 0.00000706, -0.00000243, -0.00000290, -0.00000491, -0.00000221, 0.00000061, 0.00000702, -0.00000419, -0.00000175, 0.00000070, -0.00000121, 0.00000128, -0.00000840, 0.00001200, -0.00000801, 0.00000518, -0.00001041, 0.00000726, 0.00000351, -0.00000124, 0.00000039, 0.00000233, -0.00000239, 0.00000276, 0.00000542, -0.00000546, 0.00000604, -0.00000896, 0.00000318, -0.00000083, 0.00000970, 0.00000024, 0.00000272, -0.00000942, 0.00002034, -0.00002410, 0.00001103, -0.00001899, 0.00000190, -0.00000299, 0.00000143, -0.00000560, -0.00001308, 0.00000887, -0.00000846, -0.00000156, 0.00000376, -0.00001241, 0.00001585, -0.00002159, 0.00000910, -0.00002012, 0.00001234, -0.00003441, 0.00000859, -0.00010580, -0.00005931, 0.00022741, 0.00023592, -0.00005685, -0.00011015, 0.00000416, -0.00003184, 0.00000647, -0.00001702, 0.00000250, -0.00000899, -0.00000083, -0.00000625, -0.00000632, -0.00000164, -0.00002214, -0.00000537, 0.00000221, 0.00003014, 0.00001162, 0.00000690, 0.00004042, -0.00000566, 0.00000383, -0.00003096, 0.00001219, -0.00002320, 0.00001407, 0.00000243, -0.00000314, -0.00001719, 0.00000949, },
		{ 0.00000972, -0.00000317, -0.00000088, -0.00001193, 0.00000796, -0.00000508, 0.00000292, -0.00000404, -0.00000196, -0.00000747, -0.00000092, -0.00000983, 0.00001453, 0.00000557, -0.00001017, -0.00000668, 0.00001151, -0.00000452, 0.00000281, 0.00000346, 0.00000042, -0.00000030, -0.00000393, -0.00000245, 0.00000592, 0.00000585, 0.00000083, -0.00000208, -0.00000334, -0.00000737, 0.00001060, 0.00000670, -0.00000969, 0.00000181, -0.00000430, -0.00000095, 0.00000694, 0.00000248, -0.00000814, -0.00000013, -0.00000058, 0.00000803, -0.00000436, -0.00000218, -0.00000581, -0.00000084, 0.00000065, 0.00000764, -0.00000694, 0.00000143, -0.00000216, 0.00000163, -0.00000235, -0.00000429, 0.00000980, -0.00000775, 0.00000508, -0.00001106, 0.00000996, 0.00000055, 0.00000045, -0.00000059, 0.00000350, -0.00000407, 0.00000573, 0.00000192, -0.00000284, 0.00000378, -0.00000771, 0.00000288, 0.00000101, 0.00000746, 0.00000304, -0.00000220, -0.00000336, 0.00001482, -0.00002214, 0.00001160, -0.00002274, 0.00000760, -0.00000719, 0.00000477, -0.00001010, -0.00000898, 0.00000761, -0.00000936, 0.00000050, 0.00000100, -0.00000961, 0.00001426, -0.00002219, 0.00001007, -0.00002111, 0.00001343, -0.00003782, 0.00001086, -0.00013070, -0.00000222, 0.00023355, 0.00020168, -0.00008147, -0.00009038, 0.00000141, -0.00002799, 0.00000533, -0.00001621, 0.00000247, -0.00000884, -0.00000102, -0.00000633, -0.00000620, -0.00000231, -0.00002415, -0.00000273, 0.00000248, 0.00003610, 0.00000341, 0.00001915, 0.00003275, -0.00000261, -0.00000540, -0.00002480, 0.00000983, -0.00002215, 0.00001626, 0.00000017, -0.00000476, -0.00001500, },
		{ -0.00001144, 0.00000836, -0.00000236, -0.00000327, -0.00000958, 0.00000728, -0.00000497, 0.00000245, -0.00000393, -0.00000259, -0.00000678, -0.00000263, -0.00000717, 0.00001559, 0.00000216, -0.00001067, -0.00000358, 0.00001127, -0.00000634, 0.00000502, 0.00000209, 0.00000067, -0.00000085, -0.00000405, -0.00000155, 0.00000686, 0.00000470, 0.00000082, -0.00000244, -0.00000400, -0.00000603, 0.00001276, 0.00000239, -0.00000781, 0.00000148, -0.00000469, 0.00000059, 0.00000732, -0.00000009, -0.00000727, 0.00000031, -0.00000019, 0.00000829, -0.00000589, -0.00000174, -0.00000625, 0.00000017, 0.00000120, 0.00000722, -0.00000859, 0.00000392, -0.00000469, 0.00000429, -0.00000598, 0.00000049, 0.00000617, -0.00000591, 0.00000349, -0.00001010, 0.00001126, -0.00000160, 0.00000172, -0.00000114, 0.00000411, -0.00000506, 0.00000807, -0.00000136, -0.00000007, 0.00000116, -0.00000604, 0.00000256, 0.00000242, 0.00000571, 0.00000537, -0.00000706, 0.00000359, 0.00000704, -0.00001669, 0.00000795, -0.00002234, 0.00001023, -0.00000931, 0.00000629, -0.00001302, -0.00000581, 0.00000687, -0.00001095, 0.00000366, -0.00000337, -0.00000451, 0.00000968, -0.00001948, 0.00000759, -0.00001813, 0.00000963, -0.00003510, 0.00000352, -0.00014324, 0.00005510, 0.00022593, 0.00017209, -0.00010307, -0.00006876, -0.00000464, -0.00002163, 0.00000186, -0.00001354, 0.00000100, -0.00000767, -0.00000198, -0.00000587, -0.00000639, -0.00000322, -0.00002562, -0.00000070, 0.00000443, 0.00003978, -0.00000270, 0.00003079, 0.00002413, 0.00000060, -0.00001543, -0.00001590, 0.00000433, -0.00001808, 0.00001584, -0.00000059, -0.00000732, },
		{ -0.00001036, -0.00000707, 0.00000591, -0.00000072, -0.00000647, -0.00000608, 0.00000532, -0.00000381, 0.00000106, -0.00000297, -0.00000390, -0.00000567, -0.00000475, -0.00000327, 0.00001479, -0.00000026, -0.00001124, 0.00000025, 0.00000948, -0.00000684, 0.00000655, 0.00000089, 0.00000079, -0.00000123, -0.00000432, -0.00000018, 0.00000714, 0.00000386, 0.00000069, -0.00000278, -0.00000469, -0.00000399, 0.00001372, -0.00000149, -0.00000537, 0.00000028, -0.00000421, 0.00000173, 0.00000752, -0.00000288, -0.00000546, -0.00000022, 0.00000110, 0.00000759, -0.00000678, -0.00000175, -0.00000611, 0.00000073, 0.00000229, 0.00000565, -0.00000881, 0.00000517, -0.00000625, 0.00000611, -0.00000882, 0.00000503, 0.00000192, -0.00000307, 0.00000078, -0.00000773, 0.00001109, -0.00000274, 0.00000239, -0.00000114, 0.00000401, -0.00000516, 0.00000944, -0.00000398, 0.00000241, -0.00000141, -0.00000421, 0.00000241, 0.00000313, 0.00000500, 0.00000633, -0.00001060, 0.00000983, -0.00000121, -0.00000938, 0.00000127, -0.00001829, 0.00000974, -0.00000921, 0.00000586, -0.00001419, -0.00000371, 0.00000663, -0.00001299, 0.00000751, -0.00000886, 0.00000237, 0.00000258, -0.00001387, 0.00000201, -0.00001160, 0.00000129, -0.00002653, -0.00001406, -0.00013934, 0.00010495, 0.00021038, 0.00014549, -0.00012058, -0.00004708, -0.00001254, -0.00001380, -0.00000316, -0.00000958, -0.00000148, -0.00000581, -0.00000341, -0.00000511, -0.00000672, -0.00000448, -0.00002644, 0.00000069, 0.00000819, 0.00004069, -0.00000570, 0.00004037, 0.00001593, 0.00000245, -0.00002428, -0.00000616, -0.00000291, -0.00001193, 0.00001340, -0.00000022, },
		{ 0.00000061, -0.00001322, -0.00000265, 0.00000312, 0.00000108, -0.00000976, -0.00000212, 0.00000267, -0.00000209, -0.00000081, -0.00000153, -0.00000556, -0.00000452, -0.00000668, 0.00000116, 0.00001256, -0.00000179, -0.00001169, 0.00000443, 0.00000639, -0.00000595, 0.00000718, 0.00000005, 0.00000073, -0.00000142, -0.00000468, 0.00000159, 0.00000674, 0.00000348, 0.00000029, -0.00000297, -0.00000540, -0.00000131, 0.00001327, -0.00000439, -0.00000301, -0.00000128, -0.00000312, 0.00000256, 0.00000735, -0.00000550, -0.00000310, -0.00000148, 0.00000319, 0.00000590, -0.00000693, -0.00000225, -0.00000539, 0.00000087, 0.00000377, 0.00000308, -0.00000755, 0.00000494, -0.00000645, 0.00000660, -0.00001025, 0.00000850, -0.00000205, 0.00000001, -0.00000243, -0.00000440, 0.00000968, -0.00000290, 0.00000245, -0.00000060, 0.00000326, -0.00000436, 0.00000970, -0.00000568, 0.00000428, -0.00000364, -0.00000242, 0.00000246, 0.00000308, 0.00000554, 0.00000542, -0.00001187, 0.00001394, -0.00000814, -0.00000218, -0.00000665, -0.00001184, 0.00000685, -0.00000731, 0.00000388, -0.00001394, -0.00000241, 0.00000649, -0.00001486, 0.00001123, -0.00001444, 0.00000987, -0.00000584, -0.00000652, -0.00000540, -0.00000296, -0.00000991, -0.00001415, -0.00003996, -0.00011787, 0.00014162, 0.00019259, 0.00011951, -0.00013242, -0.00002746, -0.00002051, -0.00000591, -0.00000859, -0.00000517, -0.00000434, -0.00000378, -0.00000490, -0.00000440, -0.00000695, -0.00000619, -0.00002658, 0.00000157, 0.00001359, 0.00003878, -0.00000506, 0.00004690, 0.00000912, 0.00000186, -0.00003023, 0.00000245, -0.00001019, -0.00000505, 0.00000988, },
		{ 0.00000631, 0.00000118, -0.00001522, 0.00000106, 0.00000070, 0.00000234, -0.00001241, 0.00000155, -0.00000002, -0.00000039, -0.00000262, -0.00000010, -0.00000719, -0.00000372, -0.00000782, 0.00000528, 0.00000959, -0.00000278, -0.00001168, 0.00000836, 0.00000255, -0.00000393, 0.00000691, -0.00000041, 0.00000052, -0.00000153, -0.00000497, 0.00000355, 0.00000577, 0.00000355, -0.00000044, -0.00000299, -0.00000603, 0.00000179, 0.00001144, -0.00000595, -0.00000128, -0.00000268, -0.00000181, 0.00000327, 0.00000659, -0.00000749, -0.00000072, -0.00000303, 0.00000574, 0.00000344, -0.00000645, -0.00000313, -0.00000422, 0.00000077, 0.00000534, -0.00000008, -0.00000512, 0.00000332, -0.00000524, 0.00000561, -0.00000995, 0.00001031, -0.00000499, 0.00000257, -0.00000542, -0.00000073, 0.00000747, -0.00000231, 0.00000207, 0.00000030, 0.00000204, -0.00000283, 0.00000891, -0.00000636, 0.00000536, -0.00000535, -0.00000074, 0.00000260, 0.00000253, 0.00000711, 0.00000263, -0.00001050, 0.00001506, -0.00001235, 0.00000309, -0.00001389, -0.00000463, 0.00000270, -0.00000448, 0.00000109, -0.00001292, -0.00000134, 0.00000582, -0.00001573, 0.00001377, -0.00001887, 0.00001653, -0.00001401, 0.00000094, -0.00001286, 0.00000570, -0.00002142, -0.00000130, -0.00007004, -0.00008089, 0.00016229, 0.00017697, 0.00009205, -0.00013723, -0.00001180, -0.00002683, 0.00000067, -0.00001328, -0.00000123, -0.00000688, -0.00000209, -0.00000602, -0.00000405, -0.00000690, -0.00000844, -0.00002610, 0.00000226, 0.00002010, 0.00003450, -0.00000086, 0.00004998, 0.00000405, -0.00000151, -0.00003223, 0.00000837, -0.00001598, 0.00000122, },
		{ 0.00000584, 0.00000354, 0.00000090, -0.00001587, 0.00000352, -0.00000078, 0.00000253, -0.00001384, 0.00000431, -0.00000217, 0.00000081, -0.00000391, 0.00000090, -0.00000847, -0.00000352, -0.00000772, 0.00000836, 0.00000658, -0.00000365, -0.00001080, 0.00001140, -0.00000136, -0.00000122, 0.00000595, -0.00000056, 0.00000027, -0.00000171, -0.00000495, 0.00000537, 0.00000450, 0.00000392, -0.00000140, -0.00000289, -0.00000640, 0.00000496, 0.00000850, -0.00000608, -0.00000053, -0.00000348, -0.00000063, 0.00000403, 0.00000509, -0.00000848, 0.00000111, -0.00000433, 0.00000825, 0.00000057, -0.00000557, -0.00000414, -0.00000285, 0.00000068, 0.00000659, -0.00000326, -0.00000207, 0.00000075, -0.00000291, 0.00000335, -0.00000799, 0.00001023, -0.00000641, 0.00000402, -0.00000757, 0.00000261, 0.00000496, -0.00000132, 0.00000149, 0.00000130, 0.00000065, -0.00000090, 0.00000734, -0.00000614, 0.00000567, -0.00000651, 0.00000087, 0.00000257, 0.00000193, 0.00000913, -0.00000148, -0.00000681, 0.00001310, -0.00001318, 0.00000521, -0.00001882, 0.00000174, -0.00000146, -0.00000166, -0.00000169, -0.00001181, 0.00000006, 0.00000407, -0.00001491, 0.00001423, -0.00002098, 0.00002094, -0.00002036, 0.00000682, -0.00001854, 0.00001218, -0.00003051, 0.00000816, -0.00009873, -0.00003327, 0.00016723, 0.00016583, 0.00006201, -0.00013437, -0.00000136, -0.00003029, 0.00000491, -0.00001631, 0.00000153, -0.00000856, -0.00000114, -0.00000650, -0.00000428, -0.00000646, -0.00001120, -0.00002514, 0.00000320, 0.00002698, 0.00002871, 0.00000627, 0.00004978, 0.00000046, -0.00000725, -0.00003011, 0.00001079, -0.00001923, },
		{ -0.00001960, 0.00000831, 0.00000201, -0.00000046, -0.00001499, 0.00000450, -0.00000109, 0.00000145, -0.00001378, 0.00000581, -0.00000343, 0.00000120, -0.00000438, 0.00000120, -0.00000927, -0.00000393, -0.00000628, 0.00000995, 0.00000411, -0.00000473, -0.00000878, 0.00001299, -0.00000463, 0.00000160, 0.00000460, -0.00000057, 0.00000012, -0.00000211, -0.00000442, 0.00000674, 0.00000328, 0.00000434, -0.00000242, -0.00000281, -0.00000631, 0.00000775, 0.00000496, -0.00000503, -0.00000081, -0.00000347, 0.00000018, 0.00000493, 0.00000282, -0.00000823, 0.00000198, -0.00000486, 0.00001021, -0.00000226, -0.00000460, -0.00000504, -0.00000152, 0.00000084, 0.00000711, -0.00000587, 0.00000092, -0.00000214, -0.00000002, 0.00000033, -0.00000480, 0.00000848, -0.00000624, 0.00000410, -0.00000845, 0.00000513, 0.00000264, -0.00000024, 0.00000096, 0.00000214, -0.00000059, 0.00000106, 0.00000532, -0.00000526, 0.00000534, -0.00000715, 0.00000246, 0.00000212, 0.00000183, 0.00001081, -0.00000597, -0.00000168, 0.00000870, -0.00001085, 0.00000379, -0.00002050, 0.00000609, -0.00000463, 0.00000038, -0.00000379, -0.00001113, 0.00000212, 0.00000095, -0.00001205, 0.00001210, -0.00002004, 0.00002213, -0.00002372, 0.00000988, -0.00002105, 0.00001482, -0.00003500, 0.00001085, -0.00012019, 0.00001841, 0.00015945, 0.00015935, 0.00002964, -0.00012430, 0.00000355, -0.00003038, 0.00000632, -0.00001724, 0.00000271, -0.00000912, -0.00000108, -0.00000625, -0.00000511, -0.00000567, -0.00001438, -0.00002387, 0.00000487, 0.00003338, 0.00002250, 0.00001530, 0.00004690, -0.00000237, -0.00001424, -0.00002459, 0.00000972, },
		{ 0.00000595, -0.00001745, 0.00000872, 0.00000172, -0.00000278, -0.00001279, 0.00000414, -0.00000035, -0.00000077, -0.00001230, 0.00000600, -0.00000375, 0.00000075, -0.00000398, 0.00000076, -0.00000963, -0.00000470, -0.00000373, 0.00000997, 0.00000245, -0.00000614, -0.00000558, 0.00001284, -0.00000671, 0.00000402, 0.00000321, -0.00000059, 0.00000017, -0.00000282, -0.00000324, 0.00000738, 0.00000242, 0.00000453, -0.00000329, -0.00000288, -0.00000556, 0.00000969, 0.00000143, -0.00000326, -0.00000189, -0.00000268, 0.00000056, 0.00000589, -0.00000003, -0.00000676, 0.00000168, -0.00000430, 0.00001119, -0.00000464, -0.00000384, -0.00000557, -0.00000044, 0.00000142, 0.00000666, -0.00000739, 0.00000317, -0.00000465, 0.00000276, -0.00000278, -0.00000107, 0.00000560, -0.00000476, 0.00000290, -0.00000796, 0.00000652, 0.00000085, 0.00000064, 0.00000067, 0.00000263, -0.00000143, 0.00000273, 0.00000322, -0.00000402, 0.00000457, -0.00000737, 0.00000408, 0.00000108, 0.00000268, 0.00001133, -0.00000973, 0.00000362, 0.00000310, -0.00000635, -0.00000066, -0.00001880, 0.00000789, -0.00000626, 0.00000124, -0.00000493, -0.00001106, 0.00000482, -0.00000338, -0.00000731, 0.00000744, -0.00001599, 0.00001982, -0.00002363, 0.00000960, -0.00001986, 0.00001292, -0.00003392, 0.00000475, -0.00012959, 0.00006712, 0.00014369, 0.00015587, -0.00000354, -0.00010850, 0.00000354, -0.00002738, 0.00000509, -0.00001611, 0.00000234, -0.00000858, -0.00000183, -0.00000540, -0.00000641, -0.00000467, -0.00001780, -0.00002247, 0.00000767, 0.00003851, 0.00001699, 0.00002508, 0.00004213, -0.00000533, -0.00002099, -0.00001707, },
		{ -0.00000920, 0.00000077, -0.00001363, 0.00000760, 0.00000229, -0.00000564, -0.00000975, 0.00000285, 0.00000104, -0.00000372, -0.00000976, 0.00000513, -0.00000331, -0.00000033, -0.00000290, -0.00000030, -0.00000975, -0.00000541, -0.00000060, 0.00000873, 0.00000152, -0.00000770, -0.00000150, 0.00001100, -0.00000733, 0.00000567, 0.00000200, -0.00000071, 0.00000041, -0.00000379, -0.00000142, 0.00000717, 0.00000215, 0.00000427, -0.00000382, -0.00000320, -0.00000403, 0.00001042, -0.00000145, -0.00000138, -0.00000335, -0.00000135, 0.00000063, 0.00000672, -0.00000311, -0.00000438, 0.00000030, -0.00000257, 0.00001095, -0.00000629, -0.00000349, -0.00000559, 0.00000028, 0.00000243, 0.00000516, -0.00000757, 0.00000422, -0.00000619, 0.00000480, -0.00000527, 0.00000240, 0.00000232, -0.00000255, 0.00000081, -0.00000631, 0.00000675, -0.00000025, 0.00000122, 0.00000068, 0.00000271, -0.00000176, 0.00000387, 0.00000134, -0.00000268, 0.00000353, -0.00000720, 0.00000564, -0.00000054, 0.00000468, 0.00001014, -0.00001178, 0.00000779, -0.00000221, -0.00000118, -0.00000693, -0.00001438, 0.00000728, -0.00000634, 0.00000097, -0.00000520, -0.00001139, 0.00000778, -0.00000832, -0.00000137, 0.00000097, -0.00000942, 0.00001446, -0.00002041, 0.00000632, -0.00001538, 0.00000694, -0.00002775, -0.00001017, -0.00012408, 0.00010676, 0.00012528, 0.00015270, -0.00003522, -0.00008920, -0.00000003, -0.00002217, 0.00000190, -0.00001342, 0.00000077, -0.00000727, -0.00000309, -0.00000425, -0.00000792, -0.00000371, -0.00002131, -0.00002098, 0.00001181, 0.00004178, 0.00001312, 0.00003450, 0.00003617, -0.00000914, -0.00002604, },
		{ -0.00002834, -0.00000252, -0.00000446, -0.00000925, 0.00000575, 0.00000312, -0.00000843, -0.00000649, 0.00000122, 0.00000254, -0.00000681, -0.00000670, 0.00000364, -0.00000248, -0.00000170, -0.00000144, -0.00000175, -0.00000987, -0.00000558, 0.00000241, 0.00000678, 0.00000102, -0.00000901, 0.00000292, 0.00000786, -0.00000656, 0.00000641, 0.00000110, -0.00000091, 0.00000077, -0.00000487, 0.00000086, 0.00000618, 0.00000251, 0.00000347, -0.00000394, -0.00000377, -0.00000180, 0.00000978, -0.00000323, 0.00000003, -0.00000465, 0.00000013, 0.00000063, 0.00000711, -0.00000592, -0.00000158, -0.00000180, 0.00000013, 0.00000951, -0.00000712, -0.00000359, -0.00000510, 0.00000069, 0.00000369, 0.00000283, -0.00000644, 0.00000390, -0.00000644, 0.00000567, -0.00000662, 0.00000495, -0.00000058, -0.00000028, -0.00000158, -0.00000394, 0.00000604, -0.00000070, 0.00000149, 0.00000094, 0.00000244, -0.00000161, 0.00000443, -0.00000017, -0.00000141, 0.00000235, -0.00000664, 0.00000696, -0.00000248, 0.00000768, 0.00000710, -0.00001156, 0.00000986, -0.00000587, 0.00000306, -0.00001342, -0.00000843, 0.00000494, -0.00000530, -0.00000004, -0.00000502, -0.00001166, 0.00001032, -0.00001298, 0.00000470, -0.00000618, -0.00000151, 0.00000712, -0.00001507, 0.00000108, -0.00000885, -0.00000163, -0.00001840, -0.00003190, -0.00010343, 0.00013322, 0.00010896, 0.00014698, -0.00006285, -0.00006891, -0.00000545, -0.00001599, -0.00000227, -0.00000989, -0.00000143, -0.00000561, -0.00000447, -0.00000315, -0.00000932, -0.00000301, -0.00002475, -0.00001935, 0.00001727, 0.00004297, 0.00001150, 0.00004270, 0.00002943, -0.00001408, },
		{ -0.00001988, -0.00002751, 0.00000188, -0.00000856, -0.00000532, 0.00000398, 0.00000363, -0.00001061, -0.00000361, -0.00000021, 0.00000362, -0.00000949, -0.00000369, 0.00000199, -0.00000166, -0.00000295, 0.00000003, -0.00000337, -0.00001016, -0.00000485, 0.00000464, 0.00000479, 0.00000049, -0.00000959, 0.00000696, 0.00000407, -0.00000473, 0.00000634, 0.00000049, -0.00000110, 0.00000105, -0.00000579, 0.00000330, 0.00000464, 0.00000339, 0.00000220, -0.00000368, -0.00000445, 0.00000091, 0.00000788, -0.00000368, 0.00000054, -0.00000534, 0.00000135, 0.00000085, 0.00000676, -0.00000798, 0.00000105, -0.00000407, 0.00000337, 0.00000712, -0.00000722, -0.00000406, -0.00000420, 0.00000093, 0.00000491, 0.00000007, -0.00000434, 0.00000236, -0.00000538, 0.00000526, -0.00000661, 0.00000613, -0.00000253, 0.00000141, -0.00000366, -0.00000138, 0.00000474, -0.00000067, 0.00000159, 0.00000131, 0.00000199, -0.00000114, 0.00000449, -0.00000127, -0.00000029, 0.00000108, -0.00000563, 0.00000776, -0.00000426, 0.00001117, 0.00000253, -0.00000905, 0.00000939, -0.00000703, 0.00000506, -0.00001859, -0.00000234, 0.00000183, -0.00000378, -0.00000121, -0.00000493, -0.00001131, 0.00001172, -0.00001642, 0.00000974, -0.00001268, 0.00000632, -0.00000076, -0.00000897, -0.00000467, -0.00000198, -0.00001067, -0.00000872, -0.00005669, -0.00007003, 0.00014493, 0.00009798, 0.00013647, -0.00008427, -0.00004995, -0.00001105, -0.00001006, -0.00000639, -0.00000633, -0.00000368, -0.00000406, -0.00000560, -0.00000243, -0.00001035, -0.00000281, -0.00002805, -0.00001737, 0.00002373, 0.00004217, 0.00001240, 0.00004915, 0.00002203, },
		{ 0.00001384, -0.00002576, -0.00002390, 0.00000363, -0.00001089, -0.00000253, 0.00000285, 0.00000337, -0.00001177, -0.00000152, -0.00000106, 0.00000394, -0.00001134, -0.00000115, 0.00000054, -0.00000114, -0.00000380, 0.00000126, -0.00000503, -0.00001060, -0.00000310, 0.00000566, 0.00000330, -0.00000045, -0.00000899, 0.00000991, 0.00000037, -0.00000237, 0.00000570, 0.00000003, -0.00000112, 0.00000103, -0.00000626, 0.00000549, 0.00000292, 0.00000452, 0.00000064, -0.00000319, -0.00000502, 0.00000369, 0.00000507, -0.00000287, 0.00000000, -0.00000512, 0.00000200, 0.00000145, 0.00000551, -0.00000887, 0.00000291, -0.00000592, 0.00000658, 0.00000418, -0.00000683, -0.00000470, -0.00000308, 0.00000121, 0.00000572, -0.00000258, -0.00000181, 0.00000006, -0.00000332, 0.00000378, -0.00000536, 0.00000589, -0.00000323, 0.00000211, -0.00000493, 0.00000082, 0.00000327, -0.00000039, 0.00000168, 0.00000159, 0.00000159, -0.00000059, 0.00000422, -0.00000205, 0.00000070, -0.00000027, -0.00000408, 0.00000773, -0.00000533, 0.00001438, -0.00000278, -0.00000481, 0.00000662, -0.00000550, 0.00000412, -0.00002128, 0.00000266, -0.00000111, -0.00000247, -0.00000199, -0.00000540, -0.00000987, 0.00001145, -0.00001787, 0.00001275, -0.00001731, 0.00001271, -0.00000780, -0.00000350, -0.00000947, 0.00000350, -0.00001805, -0.00000178, -0.00007982, -0.00002837, 0.00014293, 0.00009371, 0.00012008, -0.00009819, -0.00003404, -0.00001552, -0.00000538, -0.00000964, -0.00000340, -0.00000548, -0.00000298, -0.00000622, -0.00000230, -0.00001082, -0.00000325, -0.00003118, -0.00001475, 0.00003072, 0.00003979, 0.00001571, 0.00005353, },
		{ 0.00005569, 0.00000466, -0.00003066, -0.00001845, 0.00000302, -0.00001134, -0.00000113, 0.00000263, 0.00000223, -0.00001181, -0.00000037, -0.00000118, 0.00000337, -0.00001219, 0.00000067, -0.00000050, -0.00000106, -0.00000410, 0.00000209, -0.00000672, -0.00001097, -0.00000056, 0.00000537, 0.00000256, -0.00000199, -0.00000702, 0.00001128, -0.00000254, -0.00000004, 0.00000482, -0.00000044, -0.00000083, 0.00000056, -0.00000602, 0.00000705, 0.00000143, 0.00000556, -0.00000092, -0.00000271, -0.00000519, 0.00000608, 0.00000191, -0.00000119, -0.00000145, -0.00000398, 0.00000199, 0.00000249, 0.00000333, -0.00000842, 0.00000362, -0.00000682, 0.00000921, 0.00000117, -0.00000623, -0.00000528, -0.00000196, 0.00000170, 0.00000581, -0.00000459, 0.00000052, -0.00000240, -0.00000080, 0.00000168, -0.00000328, 0.00000448, -0.00000275, 0.00000173, -0.00000514, 0.00000229, 0.00000196, -0.00000007, 0.00000187, 0.00000166, 0.00000139, -0.00000018, 0.00000384, -0.00000263, 0.00000161, -0.00000171, -0.00000199, 0.00000666, -0.00000514, 0.00001648, -0.00000780, 0.00000015, 0.00000239, -0.00000183, 0.00000031, -0.00002097, 0.00000574, -0.00000318, -0.00000180, -0.00000206, -0.00000667, -0.00000720, 0.00000931, -0.00001698, 0.00001317, -0.00001930, 0.00001669, -0.00001291, 0.00000031, -0.00001228, 0.00000636, -0.00002219, -0.00000013, -0.00009663, 0.00001584, 0.00013036, 0.00009561, 0.00009806, -0.00010435, -0.00002208, -0.00001812, -0.00000247, -0.00001152, -0.00000150, -0.00000655, -0.00000253, -0.00000624, -0.00000285, -0.00001067, -0.00000441, -0.00003410, -0.00001116, 0.00003764, 0.00003643, 0.00002110, },
		{ 0.00002809, 0.00005547, -0.00000552, -0.00003354, -0.00001245, 0.00000083, -0.00001031, -0.00000094, 0.00000319, 0.00000032, -0.00001087, -0.00000009, -0.00000069, 0.00000204, -0.00001210, 0.00000177, -0.00000115, -0.00000137, -0.00000387, 0.00000251, -0.00000854, -0.00001090, 0.00000229, 0.00000405, 0.00000253, -0.00000399, -0.00000378, 0.00001089, -0.00000422, 0.00000181, 0.00000399, -0.00000103, -0.00000017, -0.00000045, -0.00000492, 0.00000771, 0.00000050, 0.00000618, -0.00000220, -0.00000244, -0.00000473, 0.00000760, -0.00000099, 0.00000082, -0.00000341, -0.00000214, 0.00000142, 0.00000380, 0.00000048, -0.00000671, 0.00000304, -0.00000647, 0.00001081, -0.00000150, -0.00000570, -0.00000560, -0.00000097, 0.00000248, 0.00000503, -0.00000558, 0.00000208, -0.00000438, 0.00000158, -0.00000044, -0.00000096, 0.00000245, -0.00000143, 0.00000045, -0.00000430, 0.00000289, 0.00000103, 0.00000017, 0.00000221, 0.00000148, 0.00000149, -0.00000004, 0.00000351, -0.00000311, 0.00000248, -0.00000317, 0.00000056, 0.00000451, -0.00000339, 0.00001674, -0.00001149, 0.00000466, -0.00000210, 0.00000286, -0.00000560, -0.00001786, 0.00000666, -0.00000410, -0.00000192, -0.00000140, -0.00000866, -0.00000349, 0.00000552, -0.00001390, 0.00001100, -0.00001843, 0.00001785, -0.00001559, 0.00000197, -0.00001267, 0.00000615, -0.00002246, -0.00000509, -0.00010345, 0.00005685, 0.00011160, 0.00010172, 0.00007181, -0.00010351, -0.00001415, -0.00001870, -0.00000141, -0.00001194, -0.00000072, -0.00000684, -0.00000270, -0.00000570, -0.00000401, -0.00000992, -0.00000632, -0.00003678, -0.00000635, 0.00004391, 0.00003276, },
		{ 0.00002944, 0.00003617, 0.00005266, -0.00001637, -0.00003370, -0.00000715, -0.00000192, -0.00000852, -0.00000147, 0.00000421, -0.00000200, -0.00000927, -0.00000038, 0.00000017, 0.00000023, -0.00001131, 0.00000231, -0.00000153, -0.00000191, -0.00000325, 0.00000256, -0.00001060, -0.00000998, 0.00000479, 0.00000226, 0.00000285, -0.00000605, 0.00000026, 0.00000892, -0.00000453, 0.00000290, 0.00000337, -0.00000176, 0.00000080, -0.00000190, -0.00000301, 0.00000737, 0.00000034, 0.00000616, -0.00000300, -0.00000251, -0.00000352, 0.00000794, -0.00000304, 0.00000254, -0.00000535, -0.00000001, 0.00000057, 0.00000507, -0.00000263, -0.00000412, 0.00000134, -0.00000484, 0.00001117, -0.00000356, -0.00000543, -0.00000554, -0.00000017, 0.00000349, 0.00000340, -0.00000539, 0.00000252, -0.00000542, 0.00000329, -0.00000205, 0.00000097, 0.00000042, 0.00000015, -0.00000127, -0.00000272, 0.00000267, 0.00000051, 0.00000032, 0.00000265, 0.00000113, 0.00000183, -0.00000021, 0.00000331, -0.00000355, 0.00000326, -0.00000452, 0.00000331, 0.00000153, -0.00000009, 0.00001482, -0.00001307, 0.00000766, -0.00000559, 0.00000718, -0.00001234, -0.00001275, 0.00000571, -0.00000398, -0.00000263, -0.00000031, -0.00001097, 0.00000072, 0.00000073, -0.00000926, 0.00000674, -0.00001506, 0.00001638, -0.00001592, 0.00000163, -0.00001093, 0.00000324, -0.00001940, -0.00001641, -0.00009836, 0.00008982, 0.00009126, 0.00010924, 0.00004353, -0.00009717, -0.00000969, -0.00001763, -0.00000185, -0.00001113, -0.00000090, -0.00000648, -0.00000332, -0.00000481, -0.00000560, -0.00000871, -0.00000899, -0.00003909, -0.00000016, 0.00004904, },
		{ 0.00005269, 0.00002703, 0.00004480, 0.00004693, -0.00002712, -0.00003101, -0.00000349, -0.00000428, -0.00000671, -0.00000216, 0.00000523, -0.00000433, -0.00000742, -0.00000094, 0.00000107, -0.00000171, -0.00001014, 0.00000250, -0.00000184, -0.00000246, -0.00000240, 0.00000230, -0.00001291, -0.00000790, 0.00000629, 0.00000065, 0.00000299, -0.00000758, 0.00000441, 0.00000590, -0.00000365, 0.00000323, 0.00000298, -0.00000252, 0.00000190, -0.00000357, -0.00000053, 0.00000615, 0.00000097, 0.00000544, -0.00000324, -0.00000292, -0.00000162, 0.00000699, -0.00000387, 0.00000344, -0.00000673, 0.00000191, -0.00000019, 0.00000591, -0.00000545, -0.00000119, -0.00000105, -0.00000218, 0.00001031, -0.00000490, -0.00000546, -0.00000512, 0.00000049, 0.00000452, 0.00000121, -0.00000417, 0.00000180, -0.00000528, 0.00000402, -0.00000278, 0.00000205, -0.00000103, 0.00000143, -0.00000288, -0.00000084, 0.00000190, 0.00000032, 0.00000048, 0.00000303, 0.00000076, 0.00000224, -0.00000055, 0.00000319, -0.00000389, 0.00000387, -0.00000555, 0.00000588, -0.00000182, 0.00000433, 0.00001082, -0.00001224, 0.00000846, -0.00000708, 0.00000983, -0.00001848, -0.00000678, 0.00000356, -0.00000324, -0.00000352, 0.00000075, -0.00001304, 0.00000472, -0.00000421, -0.00000399, 0.00000133, -0.00001003, 0.00001295, -0.00001450, -0.00000008, -0.00000790, -0.00000127, -0.00001452, -0.00003225, -0.00008152, 0.00011169, 0.00007335, 0.00011535, 0.00001564, -0.00008720, -0.00000775, -0.00001556, -0.00000322, -0.00000952, -0.00000173, -0.00000574, -0.00000415, -0.00000381, -0.00000737, -0.00000722, -0.00001240, -0.00004087, 0.00000738, },
		{ 0.00001614, 0.00005468, 0.00002604, 0.00005335, 0.00003803, -0.00003667, -0.00002595, -0.00000184, -0.00000566, -0.00000546, -0.00000251, 0.00000587, -0.00000631, -0.00000562, -0.00000150, 0.00000178, -0.00000352, -0.00000884, 0.00000253, -0.00000222, -0.00000285, -0.00000147, 0.00000169, -0.00001533, -0.00000464, 0.00000638, -0.00000019, 0.00000246, -0.00000800, 0.00000786, 0.00000256, -0.00000204, 0.00000299, 0.00000271, -0.00000315, 0.00000286, -0.00000512, 0.00000216, 0.00000434, 0.00000219, 0.00000413, -0.00000300, -0.00000354, 0.00000071, 0.00000493, -0.00000339, 0.00000323, -0.00000714, 0.00000321, -0.00000051, 0.00000596, -0.00000746, 0.00000144, -0.00000354, 0.00000103, 0.00000849, -0.00000559, -0.00000572, -0.00000441, 0.00000115, 0.00000528, -0.00000112, -0.00000231, 0.00000017, -0.00000406, 0.00000374, -0.00000254, 0.00000205, -0.00000152, 0.00000195, -0.00000390, 0.00000084, 0.00000091, 0.00000028, 0.00000077, 0.00000318, 0.00000061, 0.00000249, -0.00000089, 0.00000302, -0.00000401, 0.00000415, -0.00000602, 0.00000782, -0.00000486, 0.00000916, 0.00000533, -0.00000923, 0.00000697, -0.00000612, 0.00000989, -0.00002276, -0.00000113, 0.00000100, -0.00000239, -0.00000412, 0.00000126, -0.00001428, 0.00000779, -0.00000842, 0.00000090, -0.00000418, -0.00000438, 0.00000852, -0.00001214, -0.00000228, -0.00000468, -0.00000596, -0.00000990, -0.00004963, -0.00005511, 0.00012144, 0.00006059, 0.00011777, -0.00000968, -0.00007546, -0.00000731, -0.00001319, -0.00000489, -0.00000762, -0.00000283, -0.00000491, -0.00000495, -0.00000296, -0.00000908, -0.00000566, -0.00001651, -0.00004186, },
		{ -0.00004177, 0.00002582, 0.00005499, 0.00002687, 0.00006108, 0.00002590, -0.00004380, -0.00001955, -0.00000200, -0.00000587, -0.00000504, -0.00000226, 0.00000593, -0.00000777, -0.00000407, -0.00000193, 0.00000220, -0.00000505, -0.00000756, 0.00000249, -0.00000272, -0.00000301, -0.00000051, 0.00000059, -0.00001748, -0.00000049, 0.00000500, 0.00000002, 0.00000097, -0.00000690, 0.00000994, -0.00000032, -0.00000025, 0.00000249, 0.00000236, -0.00000341, 0.00000341, -0.00000619, 0.00000460, 0.00000237, 0.00000367, 0.00000248, -0.00000247, -0.00000411, 0.00000308, 0.00000216, -0.00000180, 0.00000191, -0.00000644, 0.00000365, -0.00000020, 0.00000503, -0.00000828, 0.00000321, -0.00000553, 0.00000424, 0.00000609, -0.00000583, -0.00000607, -0.00000357, 0.00000195, 0.00000546, -0.00000309, -0.00000034, -0.00000190, -0.00000210, 0.00000268, -0.00000152, 0.00000108, -0.00000099, 0.00000152, -0.00000399, 0.00000194, 0.00000003, 0.00000023, 0.00000131, 0.00000298, 0.00000083, 0.00000235, -0.00000099, 0.00000262, -0.00000377, 0.00000399, -0.00000572, 0.00000870, -0.00000691, 0.00001347, -0.00000072, -0.00000476, 0.00000367, -0.00000289, 0.00000708, -0.00002440, 0.00000323, -0.00000124, -0.00000190, -0.00000406, 0.00000086, -0.00001430, 0.00000942, -0.00001120, 0.00000456, -0.00000880, 0.00000090, 0.00000402, -0.00000970, -0.00000412, -0.00000227, -0.00000951, -0.00000761, -0.00006500, -0.00002276, 0.00012005, 0.00005416, 0.00011515, -0.00003082, -0.00006348, -0.00000748, -0.00001111, -0.00000632, -0.00000588, -0.00000387, -0.00000421, -0.00000551, -0.00000247, -0.00001050, -0.00000426, -0.00002125, },
		{ -0.00002654, -0.00004034, 0.00003606, 0.00005370, 0.00002988, 0.00006704, 0.00001093, -0.00004748, -0.00001307, -0.00000333, -0.00000517, -0.00000539, -0.00000140, 0.00000538, -0.00000865, -0.00000280, -0.00000224, 0.00000238, -0.00000632, -0.00000630, 0.00000233, -0.00000326, -0.00000299, 0.00000051, -0.00000120, -0.00001885, 0.00000395, 0.00000246, 0.00000121, -0.00000145, -0.00000419, 0.00001023, -0.00000216, 0.00000117, 0.00000206, 0.00000176, -0.00000315, 0.00000333, -0.00000648, 0.00000639, 0.00000067, 0.00000505, 0.00000082, -0.00000191, -0.00000434, 0.00000502, -0.00000073, 0.00000041, -0.00000022, -0.00000476, 0.00000323, 0.00000072, 0.00000315, -0.00000777, 0.00000380, -0.00000653, 0.00000690, 0.00000354, -0.00000586, -0.00000633, -0.00000270, 0.00000296, 0.00000486, -0.00000428, 0.00000117, -0.00000383, 0.00000009, 0.00000129, -0.00000016, -0.00000049, 0.00000031, 0.00000023, -0.00000310, 0.00000224, -0.00000054, 0.00000007, 0.00000211, 0.00000239, 0.00000151, 0.00000169, -0.00000068, 0.00000187, -0.00000308, 0.00000329, -0.00000456, 0.00000828, -0.00000741, 0.00001635, -0.00000623, 0.00000011, -0.00000052, 0.00000184, 0.00000179, -0.00002319, 0.00000576, -0.00000271, -0.00000199, -0.00000320, -0.00000056, -0.00001298, 0.00000941, -0.00001220, 0.00000645, -0.00001189, 0.00000506, 0.00000015, -0.00000777, -0.00000508, -0.00000130, -0.00001109, -0.00000916, -0.00007508, 0.00001114, 0.00011004, 0.00005373, 0.00010720, -0.00004695, -0.00005231, -0.00000770, -0.00000967, -0.00000717, -0.00000459, -0.00000462, -0.00000379, -0.00000572, -0.00000246, -0.00001146, -0.00000323, },
		{ -0.00000275, -0.00003223, -0.00003729, 0.00004645, 0.00005102, 0.00003532, 0.00007009, -0.00000596, -0.00004716, -0.00000769, -0.00000500, -0.00000404, -0.00000621, -0.00000009, 0.00000433, -0.00000905, -0.00000177, -0.00000253, 0.00000245, -0.00000744, -0.00000497, 0.00000194, -0.00000369, -0.00000288, 0.00000164, -0.00000387, -0.00001886, 0.00000786, -0.00000056, 0.00000295, -0.00000440, -0.00000021, 0.00000873, -0.00000263, 0.00000187, 0.00000195, 0.00000079, -0.00000230, 0.00000254, -0.00000584, 0.00000724, -0.00000041, 0.00000596, -0.00000055, -0.00000155, -0.00000402, 0.00000611, -0.00000316, 0.00000266, -0.00000265, -0.00000247, 0.00000219, 0.00000203, 0.00000060, -0.00000609, 0.00000313, -0.00000629, 0.00000861, 0.00000120, -0.00000588, -0.00000639, -0.00000189, 0.00000417, 0.00000344, -0.00000444, 0.00000179, -0.00000508, 0.00000196, 0.00000005, 0.00000104, -0.00000212, 0.00000193, -0.00000155, -0.00000143, 0.00000175, -0.00000073, -0.00000018, 0.00000309, 0.00000149, 0.00000258, 0.00000050, 0.00000009, 0.00000073, -0.00000194, 0.00000212, -0.00000264, 0.00000656, -0.00000608, 0.00001714, -0.00001022, 0.00000428, -0.00000444, 0.00000688, -0.00000503, -0.00001954, 0.00000641, -0.00000328, -0.00000265, -0.00000169, -0.00000283, -0.00001051, 0.00000791, -0.00001149, 0.00000649, -0.00001317, 0.00000775, -0.00000271, -0.00000660, -0.00000504, -0.00000184, -0.00001062, -0.00001506, -0.00007752, 0.00004234, 0.00009479, 0.00005792, 0.00009456, -0.00005797, -0.00004250, -0.00000771, -0.00000894, -0.00000734, -0.00000386, -0.00000499, -0.00000370, -0.00000559, -0.00000293, -0.00001188, },
		{ -0.00001176, -0.00000299, -0.00003816, -0.00003241, 0.00005649, 0.00004733, 0.00004326, 0.00006903, -0.00002326, -0.00004300, -0.00000422, -0.00000624, -0.00000303, -0.00000709, 0.00000140, 0.00000298, -0.00000908, -0.00000088, -0.00000291, 0.00000255, -0.00000856, -0.00000349, 0.00000120, -0.00000385, -0.00000280, 0.00000286, -0.00000746, -0.00001709, 0.00001044, -0.00000323, 0.00000457, -0.00000720, 0.00000440, 0.00000583, -0.00000178, 0.00000173, 0.00000224, -0.00000050, -0.00000094, 0.00000112, -0.00000431, 0.00000707, -0.00000067, 0.00000621, -0.00000143, -0.00000152, -0.00000303, 0.00000608, -0.00000461, 0.00000436, -0.00000481, -0.00000009, 0.00000093, 0.00000335, -0.00000217, -0.00000366, 0.00000144, -0.00000485, 0.00000922, -0.00000070, -0.00000601, -0.00000620, -0.00000109, 0.00000540, 0.00000137, -0.00000356, 0.00000132, -0.00000530, 0.00000311, -0.00000063, 0.00000160, -0.00000328, 0.00000331, -0.00000332, 0.00000061, 0.00000067, -0.00000062, -0.00000042, 0.00000406, 0.00000051, 0.00000382, -0.00000104, 0.00000121, -0.00000070, -0.00000046, 0.00000065, -0.00000024, 0.00000382, -0.00000304, 0.00001559, -0.00001203, 0.00000686, -0.00000703, 0.00001099, -0.00001215, -0.00001430, 0.00000553, -0.00000312, -0.00000363, 0.00000014, -0.00000555, -0.00000734, 0.00000540, -0.00000946, 0.00000497, -0.00001280, 0.00000898, -0.00000456, -0.00000609, -0.00000426, -0.00000347, -0.00000875, -0.00002467, -0.00007134, 0.00006744, 0.00007778, 0.00006468, 0.00007849, -0.00006436, -0.00003417, -0.00000753, -0.00000877, -0.00000693, -0.00000366, -0.00000503, -0.00000387, -0.00000520, -0.00000384, },
		{ -0.00000503, -0.00001117, -0.00000403, -0.00004413, -0.00002550, 0.00006565, 0.00004320, 0.00005342, 0.00006283, -0.00003912, -0.00003586, -0.00000290, -0.00000660, -0.00000256, -0.00000772, 0.00000285, 0.00000146, -0.00000882, -0.00000013, -0.00000337, 0.00000275, -0.00000972, -0.00000181, 0.00000007, -0.00000366, -0.00000277, 0.00000403, -0.00001179, -0.00001340, 0.00001110, -0.00000480, 0.00000534, -0.00000906, 0.00000874, 0.00000223, 0.00000003, 0.00000086, 0.00000285, -0.00000196, 0.00000070, -0.00000067, -0.00000211, 0.00000600, -0.00000013, 0.00000573, -0.00000174, -0.00000183, -0.00000145, 0.00000492, -0.00000484, 0.00000506, -0.00000617, 0.00000186, -0.00000013, 0.00000425, -0.00000461, -0.00000104, -0.00000080, -0.00000250, 0.00000876, -0.00000208, -0.00000627, -0.00000580, -0.00000022, 0.00000640, -0.00000099, -0.00000191, -0.00000013, -0.00000441, 0.00000335, -0.00000056, 0.00000126, -0.00000354, 0.00000398, -0.00000453, 0.00000251, -0.00000064, -0.00000039, -0.00000046, 0.00000477, -0.00000026, 0.00000492, -0.00000266, 0.00000247, -0.00000223, 0.00000114, -0.00000082, 0.00000220, 0.00000062, 0.00000120, 0.00001192, -0.00001148, 0.00000741, -0.00000759, 0.00001310, -0.00001830, -0.00000850, 0.00000371, -0.00000259, -0.00000455, 0.00000184, -0.00000822, -0.00000402, 0.00000248, -0.00000675, 0.00000242, -0.00001120, 0.00000907, -0.00000565, -0.00000593, -0.00000329, -0.00000541, -0.00000670, -0.00003633, -0.00005708, 0.00008437, 0.00006199, 0.00007192, 0.00006053, -0.00006692, -0.00002718, -0.00000733, -0.00000890, -0.00000616, -0.00000383, -0.00000483, -0.00000419, -0.00000470, },
		{ -0.00000425, -0.00000630, -0.00001026, -0.00000591, -0.00004994, -0.00001639, 0.00007330, 0.00003945, 0.00006503, 0.00005090, -0.00005159, -0.00002716, -0.00000338, -0.00000602, -0.00000275, -0.00000793, 0.00000414, -0.00000018, -0.00000823, 0.00000037, -0.00000380, 0.00000296, -0.00001085, -0.00000003, -0.00000135, -0.00000317, -0.00000270, 0.00000485, -0.00001639, -0.00000808, 0.00000964, -0.00000477, 0.00000473, -0.00000931, 0.00001194, -0.00000125, 0.00000221, -0.00000039, 0.00000355, -0.00000333, 0.00000232, -0.00000249, 0.00000036, 0.00000434, 0.00000101, 0.00000466, -0.00000158, -0.00000233, 0.00000046, 0.00000286, -0.00000386, 0.00000460, -0.00000645, 0.00000301, -0.00000063, 0.00000436, -0.00000623, 0.00000116, -0.00000299, 0.00000030, 0.00000750, -0.00000302, -0.00000658, -0.00000528, 0.00000086, 0.00000684, -0.00000317, 0.00000004, -0.00000220, -0.00000263, 0.00000276, 0.00000022, 0.00000001, -0.00000277, 0.00000364, -0.00000480, 0.00000381, -0.00000185, -0.00000024, -0.00000014, 0.00000502, -0.00000056, 0.00000555, -0.00000401, 0.00000358, -0.00000358, 0.00000259, -0.00000198, 0.00000419, -0.00000233, 0.00000583, 0.00000679, -0.00000891, 0.00000598, -0.00000595, 0.00001263, -0.00002249, -0.00000313, 0.00000159, -0.00000208, -0.00000506, 0.00000299, -0.00001037, -0.00000109, -0.00000023, -0.00000400, -0.00000052, -0.00000891, 0.00000842, -0.00000631, -0.00000572, -0.00000266, -0.00000685, -0.00000584, -0.00004775, -0.00003663, 0.00009255, 0.00004948, 0.00007785, 0.00004217, -0.00006655, -0.00002130, -0.00000732, -0.00000901, -0.00000528, -0.00000418, -0.00000455, -0.00000452, },
		{ -0.00000472, -0.00000401, -0.00000748, -0.00000923, -0.00000861, -0.00005538, -0.00000498, 0.00007881, 0.00003712, 0.00007678, 0.00003344, -0.00005907, -0.00001859, -0.00000487, -0.00000484, -0.00000346, -0.00000774, 0.00000529, -0.00000200, -0.00000719, 0.00000040, -0.00000395, 0.00000294, -0.00001174, 0.00000162, -0.00000286, -0.00000256, -0.00000239, 0.00000490, -0.00002052, -0.00000180, 0.00000635, -0.00000310, 0.00000255, -0.00000757, 0.00001333, -0.00000380, 0.00000414, -0.00000161, 0.00000406, -0.00000434, 0.00000359, -0.00000396, 0.00000268, 0.00000250, 0.00000242, 0.00000326, -0.00000113, -0.00000280, 0.00000230, 0.00000032, -0.00000199, 0.00000313, -0.00000562, 0.00000320, -0.00000041, 0.00000354, -0.00000671, 0.00000247, -0.00000459, 0.00000300, 0.00000577, -0.00000368, -0.00000683, -0.00000473, 0.00000225, 0.00000646, -0.00000469, 0.00000172, -0.00000435, -0.00000036, 0.00000165, 0.00000140, -0.00000188, -0.00000111, 0.00000230, -0.00000395, 0.00000422, -0.00000267, -0.00000031, 0.00000063, 0.00000467, -0.00000019, 0.00000544, -0.00000478, 0.00000425, -0.00000448, 0.00000364, -0.00000256, 0.00000529, -0.00000434, 0.00000991, 0.00000117, -0.00000508, 0.00000315, -0.00000249, 0.00000957, -0.00002418, 0.00000106, -0.00000028, -0.00000187, -0.00000493, 0.00000334, -0.00001169, 0.00000110, -0.00000230, -0.00000173, -0.00000330, -0.00000642, 0.00000744, -0.00000683, -0.00000522, -0.00000274, -0.00000722, -0.00000731, -0.00005653, -0.00001274, 0.00009276, 0.00004126, 0.00008126, 0.00002463, -0.00006410, -0.00001631, -0.00000763, -0.00000888, -0.00000454, -0.00000453, -0.00000432, },
		},
		{ { -0.00002172, -0.00001759, -0.00005912, 0.00001057, 0.00008009, 0.00003096, 0.00009720, 0.00000890, -0.00005960, -0.00001109, -0.00002009, -0.00000227, -0.00001782, -0.00000245, -0.00000013, -0.00000565, -0.00000210, 0.00000181, -0.00002095, 0.00000091, 0.00000781, -0.00000453, 0.00000037, -0.00000437, 0.00001323, -0.00000122, -0.00000109, 0.00000045, -0.00000674, -0.00000090, 0.00001278, -0.00000049, -0.00000047, -0.00000148, -0.00000588, -0.00000282, 0.00000330, 0.00000122, 0.00000153, 0.00000218, -0.00000586, 0.00000145, -0.00000654, 0.00000356, 0.00000008, 0.00000094, 0.00000055, -0.00000189, -0.00000124, -0.00000010, 0.00000316, 0.00000008, 0.00000055, -0.00000244, 0.00000309, -0.00000305, 0.00000042, -0.00000015, -0.00000005, -0.00000024, -0.00000186, -0.00000117, -0.00000349, 0.00000359, 0.00000484, 0.00000359, -0.00000349, -0.00000117, -0.00000186, -0.00000024, -0.00000005, -0.00000015, 0.00000042, -0.00000305, 0.00000309, -0.00000244, 0.00000055, 0.00000008, 0.00000316, -0.00000010, -0.00000124, -0.00000189, 0.00000055, 0.00000094, 0.00000008, 0.00000356, -0.00000654, 0.00000145, -0.00000586, 0.00000218, 0.00000153, 0.00000122, 0.00000330, -0.00000282, -0.00000588, -0.00000148, -0.00000047, -0.00000049, 0.00001278, -0.00000090, -0.00000674, 0.00000045, -0.00000109, -0.00000122, 0.00001323, -0.00000437, 0.00000037, -0.00000453, 0.00000781, 0.00000091, -0.00002095, 0.00000181, -0.00000210, -0.00000565, -0.00000013, -0.00000245, -0.00001782, -0.00000227, -0.00002009, -0.00001109, -0.00005960, 0.00000890, 0.00009720, 0.00003096, 0.00008009, 0.00001057, -0.00005912, -0.00001759, },
		{ -0.00001692, -0.00002174, -0.00001835, -0.00006220, 0.00001781, 0.00008171, 0.00003396, 0.00010342, -0.00000731, -0.00005707, -0.00000932, -0.00002098, -0.00000132, -0.00001847, -0.00000132, -0.00000120, -0.00000529, -0.00000134, 0.00000041, -0.00002162, 0.00000395, 0.00000550, -0.00000321, -0.00000104, -0.00000240, 0.00001289, -0.00000236, -0.00000044, -0.00000009, -0.00000680, 0.00000022, 0.00001292, -0.00000194, 0.00000048, -0.00000246, -0.00000519, -0.00000303, 0.00000393, 0.00000053, 0.00000222, 0.00000110, -0.00000531, 0.00000124, -0.00000664, 0.00000451, -0.00000071, 0.00000147, -0.00000008, -0.00000137, -0.00000191, 0.00000080, 0.00000265, 0.00000009, 0.00000038, -0.00000199, 0.00000299, -0.00000349, 0.00000106, -0.00000066, 0.00000040, -0.00000072, -0.00000163, -0.00000172, -0.00000286, 0.00000399, 0.00000488, 0.00000299, -0.00000392, -0.00000067, -0.00000207, 0.00000020, -0.00000050, 0.00000039, -0.00000031, -0.00000241, 0.00000296, -0.00000275, 0.00000063, 0.00000027, 0.00000342, -0.00000081, -0.00000062, -0.00000242, 0.00000122, 0.00000038, 0.00000092, 0.00000240, -0.00000610, 0.00000131, -0.00000609, 0.00000302, 0.00000099, 0.00000179, 0.00000280, -0.00000278, -0.00000644, -0.00000052, -0.00000146, 0.00000117, 0.00001226, -0.00000179, -0.00000673, 0.00000107, -0.00000195, 0.00000036, 0.00001288, -0.00000580, 0.00000144, -0.00000552, 0.00000996, -0.00000244, -0.00001963, 0.00000281, -0.00000282, -0.00000599, 0.00000102, -0.00000374, -0.00001699, -0.00000350, -0.00001861, -0.00001422, -0.00005951, 0.00002354, 0.00008968, 0.00002929, 0.00007771, 0.00000416, -0.00005613, },
		{ -0.00005324, -0.00001634, -0.00002180, -0.00001922, -0.00006534, 0.00002590, 0.00008249, 0.00003852, 0.00010768, -0.00002444, -0.00005201, -0.00000911, -0.00002101, -0.00000088, -0.00001877, -0.00000051, -0.00000204, -0.00000502, -0.00000047, -0.00000143, -0.00002152, 0.00000641, 0.00000332, -0.00000178, -0.00000262, -0.00000004, 0.00001191, -0.00000303, -0.00000002, -0.00000051, -0.00000693, 0.00000157, 0.00001263, -0.00000309, 0.00000129, -0.00000335, -0.00000447, -0.00000332, 0.00000462, -0.00000025, 0.00000303, -0.00000017, -0.00000444, 0.00000067, -0.00000636, 0.00000517, -0.00000135, 0.00000190, -0.00000062, -0.00000095, -0.00000254, 0.00000178, 0.00000194, 0.00000028, 0.00000011, -0.00000142, 0.00000267, -0.00000370, 0.00000154, -0.00000108, 0.00000081, -0.00000120, -0.00000140, -0.00000226, -0.00000208, 0.00000423, 0.00000500, 0.00000220, -0.00000412, -0.00000027, -0.00000220, 0.00000056, -0.00000089, 0.00000090, -0.00000109, -0.00000160, 0.00000261, -0.00000292, 0.00000062, 0.00000063, 0.00000341, -0.00000128, -0.00000015, -0.00000289, 0.00000185, -0.00000011, 0.00000171, 0.00000113, -0.00000542, 0.00000089, -0.00000599, 0.00000362, 0.00000062, 0.00000221, 0.00000243, -0.00000294, -0.00000680, 0.00000034, -0.00000238, 0.00000295, 0.00001141, -0.00000246, -0.00000677, 0.00000175, -0.00000292, 0.00000233, 0.00001185, -0.00000661, 0.00000204, -0.00000599, 0.00001169, -0.00000585, -0.00001786, 0.00000348, -0.00000356, -0.00000619, 0.00000212, -0.00000505, -0.00001617, -0.00000474, -0.00001691, -0.00001829, -0.00005691, 0.00003618, 0.00008144, 0.00002868, 0.00007465, -0.00000145, },
		{ -0.00000630, -0.00005050, -0.00001582, -0.00002191, -0.00002022, -0.00006848, 0.00003484, 0.00008239, 0.00004484, 0.00010932, -0.00004168, -0.00004478, -0.00001054, -0.00002002, -0.00000110, -0.00001859, -0.00000011, -0.00000258, -0.00000490, 0.00000054, -0.00000374, -0.00002055, 0.00000808, 0.00000152, -0.00000049, -0.00000413, 0.00000253, 0.00001042, -0.00000325, 0.00000016, -0.00000080, -0.00000710, 0.00000311, 0.00001192, -0.00000390, 0.00000191, -0.00000407, -0.00000383, -0.00000362, 0.00000532, -0.00000106, 0.00000387, -0.00000158, -0.00000330, -0.00000024, -0.00000568, 0.00000547, -0.00000178, 0.00000218, -0.00000100, -0.00000070, -0.00000304, 0.00000277, 0.00000109, 0.00000061, -0.00000024, -0.00000076, 0.00000214, -0.00000368, 0.00000186, -0.00000138, 0.00000113, -0.00000164, -0.00000124, -0.00000272, -0.00000122, 0.00000433, 0.00000516, 0.00000124, -0.00000406, -0.00000001, -0.00000223, 0.00000079, -0.00000119, 0.00000134, -0.00000185, -0.00000069, 0.00000207, -0.00000296, 0.00000055, 0.00000114, 0.00000314, -0.00000148, 0.00000011, -0.00000321, 0.00000237, -0.00000049, 0.00000237, -0.00000013, -0.00000459, 0.00000025, -0.00000562, 0.00000399, 0.00000040, 0.00000249, 0.00000220, -0.00000332, -0.00000689, 0.00000101, -0.00000317, 0.00000473, 0.00001032, -0.00000294, -0.00000682, 0.00000241, -0.00000391, 0.00000456, 0.00001020, -0.00000680, 0.00000213, -0.00000583, 0.00001279, -0.00000907, -0.00001580, 0.00000393, -0.00000439, -0.00000614, 0.00000304, -0.00000623, -0.00001555, -0.00000573, -0.00001536, -0.00002282, -0.00005210, 0.00004657, 0.00007294, 0.00002888, 0.00007100, },
		{ 0.00006685, -0.00001041, -0.00004793, -0.00001535, -0.00002206, -0.00002136, -0.00007157, 0.00004460, 0.00008145, 0.00005303, 0.00010767, -0.00005811, -0.00003593, -0.00001346, -0.00001800, -0.00000203, -0.00001788, -0.00000016, -0.00000277, -0.00000496, 0.00000165, -0.00000644, -0.00001873, 0.00000884, 0.00000034, 0.00000042, -0.00000536, 0.00000508, 0.00000857, -0.00000308, 0.00000010, -0.00000099, -0.00000728, 0.00000480, 0.00001080, -0.00000433, 0.00000226, -0.00000455, -0.00000336, -0.00000382, 0.00000594, -0.00000182, 0.00000467, -0.00000302, -0.00000198, -0.00000144, -0.00000460, 0.00000539, -0.00000196, 0.00000226, -0.00000118, -0.00000067, -0.00000331, 0.00000365, 0.00000019, 0.00000104, -0.00000062, -0.00000005, 0.00000145, -0.00000344, 0.00000198, -0.00000154, 0.00000132, -0.00000200, -0.00000115, -0.00000306, -0.00000036, 0.00000436, 0.00000531, 0.00000016, -0.00000375, 0.00000010, -0.00000214, 0.00000089, -0.00000135, 0.00000166, -0.00000253, 0.00000026, 0.00000139, -0.00000290, 0.00000045, 0.00000174, 0.00000265, -0.00000141, 0.00000013, -0.00000335, 0.00000275, -0.00000071, 0.00000281, -0.00000128, -0.00000373, -0.00000052, -0.00000503, 0.00000416, 0.00000032, 0.00000265, 0.00000207, -0.00000392, -0.00000668, 0.00000143, -0.00000372, 0.00000641, 0.00000906, -0.00000328, -0.00000684, 0.00000299, -0.00000482, 0.00000689, 0.00000805, -0.00000641, 0.00000172, -0.00000501, 0.00001313, -0.00001187, -0.00001362, 0.00000422, -0.00000536, -0.00000579, 0.00000368, -0.00000718, -0.00001525, -0.00000625, -0.00001432, -0.00002728, -0.00004554, 0.00005467, 0.00006455, 0.00002970, },
		{ 0.00003094, 0.00006227, -0.00001380, -0.00004555, -0.00001491, -0.00002225, -0.00002271, -0.00007451, 0.00005510, 0.00007978, 0.00006307, 0.00010219, -0.00007275, -0.00002618, -0.00001754, -0.00001508, -0.00000364, -0.00001669, -0.00000061, -0.00000264, -0.00000517, 0.00000280, -0.00000942, -0.00001613, 0.00000864, -0.00000007, 0.00000078, -0.00000609, 0.00000739, 0.00000654, -0.00000263, -0.00000014, -0.00000109, -0.00000742, 0.00000656, 0.00000932, -0.00000437, 0.00000234, -0.00000472, -0.00000315, -0.00000382, 0.00000640, -0.00000243, 0.00000533, -0.00000439, -0.00000059, -0.00000281, -0.00000320, 0.00000494, -0.00000187, 0.00000213, -0.00000115, -0.00000091, -0.00000328, 0.00000433, -0.00000067, 0.00000151, -0.00000100, 0.00000067, 0.00000065, -0.00000301, 0.00000193, -0.00000154, 0.00000138, -0.00000227, -0.00000118, -0.00000323, 0.00000044, 0.00000438, 0.00000539, -0.00000098, -0.00000324, 0.00000004, -0.00000193, 0.00000083, -0.00000135, 0.00000180, -0.00000307, 0.00000118, 0.00000060, -0.00000276, 0.00000037, 0.00000237, 0.00000201, -0.00000109, -0.00000009, -0.00000329, 0.00000297, -0.00000072, 0.00000300, -0.00000223, -0.00000292, -0.00000132, -0.00000431, 0.00000418, 0.00000034, 0.00000274, 0.00000199, -0.00000468, -0.00000617, 0.00000156, -0.00000397, 0.00000791, 0.00000773, -0.00000353, -0.00000680, 0.00000342, -0.00000551, 0.00000917, 0.00000557, -0.00000557, 0.00000090, -0.00000359, 0.00001266, -0.00001413, -0.00001145, 0.00000439, -0.00000646, -0.00000510, 0.00000401, -0.00000786, -0.00001533, -0.00000617, -0.00001407, -0.00003116, -0.00003777, 0.00006058, 0.00005654, },
		{ 0.00004906, 0.00003246, 0.00005735, -0.00001653, -0.00004338, -0.00001451, -0.00002244, -0.00002432, -0.00007715, 0.00006619, 0.00007758, 0.00007474, 0.00009253, -0.00008467, -0.00001634, -0.00002232, -0.00001151, -0.00000574, -0.00001513, -0.00000137, -0.00000228, -0.00000543, 0.00000386, -0.00001247, -0.00001293, 0.00000755, 0.00000033, 0.00000045, -0.00000618, 0.00000926, 0.00000451, -0.00000201, -0.00000051, -0.00000115, -0.00000746, 0.00000830, 0.00000755, -0.00000406, 0.00000214, -0.00000459, -0.00000327, -0.00000355, 0.00000660, -0.00000282, 0.00000576, -0.00000557, 0.00000074, -0.00000426, -0.00000157, 0.00000418, -0.00000152, 0.00000181, -0.00000092, -0.00000141, -0.00000293, 0.00000474, -0.00000142, 0.00000197, -0.00000133, 0.00000132, -0.00000020, -0.00000244, 0.00000172, -0.00000140, 0.00000130, -0.00000243, -0.00000132, -0.00000320, 0.00000113, 0.00000443, 0.00000534, -0.00000210, -0.00000255, -0.00000015, -0.00000161, 0.00000063, -0.00000119, 0.00000176, -0.00000342, 0.00000201, -0.00000024, -0.00000258, 0.00000034, 0.00000296, 0.00000128, -0.00000061, -0.00000052, -0.00000302, 0.00000301, -0.00000055, 0.00000290, -0.00000293, -0.00000226, -0.00000208, -0.00000352, 0.00000411, 0.00000042, 0.00000280, 0.00000191, -0.00000556, -0.00000538, 0.00000139, -0.00000388, 0.00000912, 0.00000642, -0.00000374, -0.00000665, 0.00000363, -0.00000588, 0.00001121, 0.00000296, -0.00000444, -0.00000017, -0.00000170, 0.00001140, -0.00001575, -0.00000937, 0.00000445, -0.00000767, -0.00000412, 0.00000403, -0.00000829, -0.00001576, -0.00000548, -0.00001476, -0.00003404, -0.00002933, 0.00006449, },
		{ 0.00006667, 0.00004221, 0.00003412, 0.00005215, -0.00001863, -0.00004141, -0.00001417, -0.00002259, -0.00002629, -0.00007932, 0.00007766, 0.00007519, 0.00008764, 0.00007854, -0.00009305, -0.00000728, -0.00002722, -0.00000766, -0.00000811, -0.00001339, -0.00000226, -0.00000183, -0.00000563, 0.00000464, -0.00001536, -0.00000941, 0.00000575, 0.00000150, -0.00000060, -0.00000552, 0.00001051, 0.00000266, -0.00000133, -0.00000094, -0.00000121, -0.00000734, 0.00000992, 0.00000559, -0.00000345, 0.00000169, -0.00000415, -0.00000372, -0.00000295, 0.00000648, -0.00000291, 0.00000586, -0.00000644, 0.00000187, -0.00000562, 0.00000017, 0.00000320, -0.00000099, 0.00000133, -0.00000055, -0.00000213, -0.00000225, 0.00000483, -0.00000199, 0.00000235, -0.00000158, 0.00000187, -0.00000103, -0.00000178, 0.00000140, -0.00000115, 0.00000108, -0.00000249, -0.00000157, -0.00000296, 0.00000165, 0.00000457, 0.00000510, -0.00000311, -0.00000177, -0.00000044, -0.00000119, 0.00000031, -0.00000087, 0.00000152, -0.00000354, 0.00000269, -0.00000108, -0.00000237, 0.00000041, 0.00000345, 0.00000056, -0.00000002, -0.00000112, -0.00000258, 0.00000292, -0.00000022, 0.00000253, -0.00000334, -0.00000179, -0.00000271, -0.00000271, 0.00000398, 0.00000054, 0.00000285, 0.00000178, -0.00000646, -0.00000437, 0.00000094, -0.00000343, 0.00001000, 0.00000520, -0.00000398, -0.00000635, 0.00000360, -0.00000585, 0.00001284, 0.00000042, -0.00000321, -0.00000129, 0.00000046, 0.00000948, -0.00001675, -0.00000742, 0.00000437, -0.00000889, -0.00000292, 0.00000383, -0.00000858, -0.00001643, -0.00000427, -0.00001643, -0.00003561, -0.00002075, },
		{ -0.00001246, 0.00006740, 0.00003608, 0.00003580, 0.00004680, -0.00002019, -0.00003960, -0.00001393, -0.00002262, -0.00002877, -0.00008076, 0.00008924, 0.00007303, 0.00010114, 0.00006041, -0.00009730, 0.00000023, -0.00003166, -0.00000395, -0.00001045, -0.00001171, -0.00000308, -0.00000148, -0.00000559, 0.00000495, -0.00001779, -0.00000588, 0.00000351, 0.00000329, -0.00000230, -0.00000409, 0.00001106, 0.00000113, -0.00000072, -0.00000136, -0.00000131, -0.00000701, 0.00001133, 0.00000355, -0.00000262, 0.00000107, -0.00000347, -0.00000451, -0.00000200, 0.00000601, -0.00000265, 0.00000556, -0.00000690, 0.00000268, -0.00000676, 0.00000187, 0.00000208, -0.00000032, 0.00000076, -0.00000011, -0.00000299, -0.00000127, 0.00000458, -0.00000234, 0.00000262, -0.00000171, 0.00000228, -0.00000179, -0.00000109, 0.00000101, -0.00000081, 0.00000076, -0.00000247, -0.00000189, -0.00000252, 0.00000200, 0.00000482, 0.00000463, -0.00000394, -0.00000096, -0.00000080, -0.00000073, -0.00000011, -0.00000042, 0.00000108, -0.00000342, 0.00000318, -0.00000188, -0.00000217, 0.00000060, 0.00000376, -0.00000006, 0.00000057, -0.00000180, -0.00000203, 0.00000275, 0.00000021, 0.00000194, -0.00000348, -0.00000154, -0.00000319, -0.00000195, 0.00000382, 0.00000066, 0.00000293, 0.00000152, -0.00000733, -0.00000320, 0.00000024, -0.00000263, 0.00001050, 0.00000414, -0.00000426, -0.00000590, 0.00000330, -0.00000536, 0.00001394, -0.00000187, -0.00000205, -0.00000226, 0.00000266, 0.00000709, -0.00001718, -0.00000558, 0.00000409, -0.00001000, -0.00000163, 0.00000354, -0.00000887, -0.00001713, -0.00000275, -0.00001892, -0.00003575, },
		{ -0.00003449, -0.00000475, 0.00006693, 0.00003069, 0.00003736, 0.00004140, -0.00002130, -0.00003791, -0.00001387, -0.00002244, -0.00003191, -0.00008123, 0.00010062, 0.00007161, 0.00011440, 0.00003861, -0.00009711, 0.00000554, -0.00003509, -0.00000077, -0.00001245, -0.00001032, -0.00000362, -0.00000141, -0.00000516, 0.00000462, -0.00001950, -0.00000270, 0.00000115, 0.00000545, -0.00000450, -0.00000197, 0.00001086, 0.00000002, -0.00000025, -0.00000171, -0.00000148, -0.00000644, 0.00001244, 0.00000154, -0.00000164, 0.00000035, -0.00000263, -0.00000556, -0.00000072, 0.00000518, -0.00000202, 0.00000484, -0.00000688, 0.00000306, -0.00000755, 0.00000340, 0.00000097, 0.00000038, 0.00000018, 0.00000033, -0.00000392, -0.00000005, 0.00000402, -0.00000244, 0.00000275, -0.00000170, 0.00000250, -0.00000241, -0.00000043, 0.00000060, -0.00000043, 0.00000037, -0.00000241, -0.00000224, -0.00000193, 0.00000218, 0.00000518, 0.00000389, -0.00000452, -0.00000020, -0.00000117, -0.00000026, -0.00000057, 0.00000012, 0.00000049, -0.00000306, 0.00000345, -0.00000261, -0.00000197, 0.00000091, 0.00000387, -0.00000051, 0.00000107, -0.00000249, -0.00000142, 0.00000255, 0.00000068, 0.00000118, -0.00000338, -0.00000151, -0.00000348, -0.00000126, 0.00000367, 0.00000078, 0.00000304, 0.00000112, -0.00000806, -0.00000196, -0.00000066, -0.00000151, 0.00001060, 0.00000328, -0.00000463, -0.00000530, 0.00000273, -0.00000438, 0.00001442, -0.00000376, -0.00000113, -0.00000288, 0.00000463, 0.00000444, -0.00001715, -0.00000382, 0.00000353, -0.00001087, -0.00000041, 0.00000329, -0.00000936, -0.00001762, -0.00000122, -0.00002195, },
		{ -0.00002516, -0.00003202, 0.00000220, 0.00006546, 0.00002611, 0.00003866, 0.00003609, -0.00002208, -0.00003624, -0.00001406, -0.00002193, -0.00003588, -0.00008043, 0.00011147, 0.00007149, 0.00012642, 0.00001398, -0.00009249, 0.00000822, -0.00003709, 0.00000152, -0.00001385, -0.00000943, -0.00000371, -0.00000176, -0.00000421, 0.00000349, -0.00002024, -0.00000022, -0.00000096, 0.00000768, -0.00000697, 0.00000069, 0.00000996, -0.00000062, 0.00000001, -0.00000197, -0.00000174, -0.00000561, 0.00001318, -0.00000034, -0.00000063, -0.00000037, -0.00000174, -0.00000680, 0.00000083, 0.00000404, -0.00000106, 0.00000369, -0.00000637, 0.00000295, -0.00000788, 0.00000462, -0.00000003, 0.00000102, -0.00000033, 0.00000067, -0.00000480, 0.00000129, 0.00000319, -0.00000231, 0.00000274, -0.00000155, 0.00000253, -0.00000287, 0.00000014, 0.00000023, -0.00000007, -0.00000006, -0.00000233, -0.00000258, -0.00000122, 0.00000222, 0.00000564, 0.00000290, -0.00000479, 0.00000045, -0.00000151, 0.00000017, -0.00000103, 0.00000071, -0.00000022, -0.00000249, 0.00000348, -0.00000324, -0.00000179, 0.00000134, 0.00000375, -0.00000074, 0.00000140, -0.00000310, -0.00000083, 0.00000239, 0.00000109, 0.00000034, -0.00000310, -0.00000166, -0.00000359, -0.00000066, 0.00000353, 0.00000089, 0.00000318, 0.00000055, -0.00000861, -0.00000072, -0.00000167, -0.00000012, 0.00001032, 0.00000264, -0.00000508, -0.00000456, 0.00000196, -0.00000296, 0.00001423, -0.00000516, -0.00000056, -0.00000301, 0.00000616, 0.00000176, -0.00001679, -0.00000208, 0.00000261, -0.00001137, 0.00000061, 0.00000325, -0.00001021, -0.00001764, -0.00000004, },
		{ 0.00000045, -0.00002809, -0.00002866, 0.00000835, 0.00006314, 0.00002239, 0.00003956, 0.00003102, -0.00002267, -0.00003451, -0.00001460, -0.00002100, -0.00004081, -0.00007812, 0.00012149, 0.00007319, 0.00013609, -0.00001239, -0.00008379, 0.00000810, -0.00003736, 0.00000269, -0.00001447, -0.00000918, -0.00000322, -0.00000265, -0.00000266, 0.00000148, -0.00001984, 0.00000127, -0.00000246, 0.00000964, -0.00000944, 0.00000369, 0.00000845, -0.00000079, 0.00000004, -0.00000211, -0.00000208, -0.00000451, 0.00001350, -0.00000199, 0.00000032, -0.00000100, -0.00000093, -0.00000810, 0.00000256, 0.00000264, 0.00000019, 0.00000217, -0.00000537, 0.00000232, -0.00000769, 0.00000543, -0.00000081, 0.00000152, -0.00000071, 0.00000081, -0.00000551, 0.00000266, 0.00000217, -0.00000197, 0.00000259, -0.00000127, 0.00000236, -0.00000314, 0.00000060, -0.00000007, 0.00000024, -0.00000047, -0.00000228, -0.00000285, -0.00000048, 0.00000216, 0.00000615, 0.00000167, -0.00000475, 0.00000094, -0.00000176, 0.00000052, -0.00000143, 0.00000128, -0.00000100, -0.00000173, 0.00000328, -0.00000377, -0.00000161, 0.00000186, 0.00000342, -0.00000072, 0.00000150, -0.00000357, -0.00000032, 0.00000233, 0.00000137, -0.00000047, -0.00000271, -0.00000196, -0.00000353, -0.00000014, 0.00000340, 0.00000102, 0.00000333, -0.00000019, -0.00000892, 0.00000043, -0.00000271, 0.00000146, 0.00000970, 0.00000223, -0.00000562, -0.00000373, 0.00000103, -0.00000117, 0.00001340, -0.00000601, -0.00000040, -0.00000255, 0.00000705, -0.00000073, -0.00001622, -0.00000033, 0.00000130, -0.00001141, 0.00000130, 0.00000354, -0.00001160, -0.00001693, },
		{ -0.00001535, -0.00000005, -0.00003030, -0.00002483, 0.00001379, 0.00006006, 0.00001958, 0.00003990, 0.00002636, -0.00002322, -0.00003264, -0.00001555, -0.00001957, -0.00004682, -0.00007407, 0.00013046, 0.00007716, 0.00014228, -0.00003916, -0.00007172, 0.00000530, -0.00003586, 0.00000262, -0.00001421, -0.00000964, -0.00000212, -0.00000409, -0.00000053, -0.00000140, -0.00001822, 0.00000155, -0.00000304, 0.00001101, -0.00001160, 0.00000676, 0.00000650, -0.00000056, -0.00000017, -0.00000213, -0.00000247, -0.00000317, 0.00001337, -0.00000334, 0.00000115, -0.00000143, -0.00000032, -0.00000931, 0.00000435, 0.00000111, 0.00000161, 0.00000037, -0.00000395, 0.00000121, -0.00000695, 0.00000578, -0.00000129, 0.00000182, -0.00000089, 0.00000070, -0.00000595, 0.00000392, 0.00000105, -0.00000147, 0.00000234, -0.00000090, 0.00000203, -0.00000321, 0.00000092, -0.00000027, 0.00000046, -0.00000084, -0.00000229, -0.00000300, 0.00000023, 0.00000208, 0.00000666, 0.00000026, -0.00000438, 0.00000123, -0.00000190, 0.00000076, -0.00000174, 0.00000179, -0.00000178, -0.00000083, 0.00000286, -0.00000418, -0.00000140, 0.00000243, 0.00000291, -0.00000044, 0.00000135, -0.00000383, 0.00000007, 0.00000242, 0.00000145, -0.00000118, -0.00000232, -0.00000233, -0.00000335, 0.00000031, 0.00000326, 0.00000118, 0.00000344, -0.00000107, -0.00000896, 0.00000143, -0.00000371, 0.00000314, 0.00000878, 0.00000202, -0.00000621, -0.00000285, 0.00000004, 0.00000089, 0.00001201, -0.00000635, -0.00000063, -0.00000148, 0.00000720, -0.00000286, -0.00001556, 0.00000146, -0.00000043, -0.00001093, 0.00000158, 0.00000421, -0.00001361, },
		{ -0.00001624, -0.00001280, -0.00000175, -0.00003139, -0.00002093, 0.00001863, 0.00005626, 0.00001775, 0.00003956, 0.00002225, -0.00002385, -0.00003054, -0.00001696, -0.00001762, -0.00005396, -0.00006814, 0.00013825, 0.00008369, 0.00014392, -0.00006486, -0.00005723, 0.00000021, -0.00003272, 0.00000138, -0.00001313, -0.00001073, -0.00000045, -0.00000601, 0.00000210, -0.00000504, -0.00001543, 0.00000054, -0.00000249, 0.00001152, -0.00001317, 0.00000964, 0.00000430, -0.00000001, -0.00000057, -0.00000206, -0.00000288, -0.00000166, 0.00001281, -0.00000433, 0.00000178, -0.00000160, -0.00000003, -0.00001029, 0.00000604, -0.00000043, 0.00000309, -0.00000158, -0.00000223, -0.00000031, -0.00000571, 0.00000564, -0.00000144, 0.00000188, -0.00000085, 0.00000029, -0.00000604, 0.00000498, -0.00000006, -0.00000089, 0.00000203, -0.00000047, 0.00000156, -0.00000311, 0.00000109, -0.00000034, 0.00000056, -0.00000113, -0.00000237, -0.00000298, 0.00000085, 0.00000202, 0.00000708, -0.00000125, -0.00000372, 0.00000131, -0.00000190, 0.00000086, -0.00000192, 0.00000219, -0.00000250, 0.00000013, 0.00000226, -0.00000449, -0.00000115, 0.00000298, 0.00000228, 0.00000004, 0.00000094, -0.00000387, 0.00000031, 0.00000269, 0.00000127, -0.00000171, -0.00000198, -0.00000271, -0.00000309, 0.00000070, 0.00000310, 0.00000140, 0.00000348, -0.00000205, -0.00000872, 0.00000221, -0.00000458, 0.00000485, 0.00000765, 0.00000198, -0.00000684, -0.00000199, -0.00000091, 0.00000307, 0.00001017, -0.00000625, -0.00000116, 0.00000013, 0.00000657, -0.00000451, -0.00001489, 0.00000327, -0.00000253, -0.00000994, 0.00000147, 0.00000527, },
		{ 0.00000661, -0.00001939, -0.00000936, -0.00000472, -0.00003108, -0.00001738, 0.00002307, 0.00005179, 0.00001692, 0.00003844, 0.00001879, -0.00002465, -0.00002819, -0.00001883, -0.00001518, -0.00006219, -0.00006032, 0.00014492, 0.00009283, 0.00014018, -0.00008800, -0.00004149, -0.00000657, -0.00002827, -0.00000083, -0.00001140, -0.00001231, 0.00000161, -0.00000826, 0.00000502, -0.00000922, -0.00001164, -0.00000173, -0.00000070, 0.00001097, -0.00001389, 0.00001205, 0.00000206, 0.00000075, -0.00000110, -0.00000194, -0.00000324, -0.00000003, 0.00001185, -0.00000496, 0.00000219, -0.00000147, -0.00000015, -0.00001091, 0.00000750, -0.00000184, 0.00000446, -0.00000353, -0.00000036, -0.00000211, -0.00000403, 0.00000505, -0.00000126, 0.00000169, -0.00000060, -0.00000042, -0.00000572, 0.00000572, -0.00000108, -0.00000029, 0.00000171, -0.00000003, 0.00000101, -0.00000286, 0.00000112, -0.00000029, 0.00000052, -0.00000134, -0.00000255, -0.00000277, 0.00000132, 0.00000207, 0.00000734, -0.00000277, -0.00000281, 0.00000118, -0.00000176, 0.00000082, -0.00000196, 0.00000244, -0.00000311, 0.00000111, 0.00000149, -0.00000470, -0.00000083, 0.00000346, 0.00000162, 0.00000067, 0.00000030, -0.00000370, 0.00000040, 0.00000312, 0.00000083, -0.00000200, -0.00000177, -0.00000304, -0.00000277, 0.00000105, 0.00000292, 0.00000169, 0.00000339, -0.00000308, -0.00000821, 0.00000273, -0.00000525, 0.00000647, 0.00000637, 0.00000208, -0.00000746, -0.00000119, -0.00000172, 0.00000522, 0.00000805, -0.00000583, -0.00000188, 0.00000215, 0.00000521, -0.00000564, -0.00001422, 0.00000504, -0.00000490, -0.00000851, 0.00000101, },
		{ 0.00000035, 0.00000807, -0.00002284, -0.00000521, -0.00000887, -0.00002924, -0.00001448, 0.00002726, 0.00004669, 0.00001708, 0.00003648, 0.00001606, -0.00002569, -0.00002560, -0.00002109, -0.00001238, -0.00007137, -0.00005070, 0.00015070, 0.00010432, 0.00013047, -0.00010722, -0.00002576, -0.00001424, -0.00002300, -0.00000364, -0.00000930, -0.00001411, 0.00000383, -0.00001059, 0.00000797, -0.00001361, -0.00000717, -0.00000507, 0.00000231, 0.00000926, -0.00001361, 0.00001378, -0.00000001, 0.00000157, -0.00000171, -0.00000181, -0.00000352, 0.00000163, 0.00001055, -0.00000523, 0.00000238, -0.00000102, -0.00000074, -0.00001106, 0.00000860, -0.00000296, 0.00000559, -0.00000531, 0.00000150, -0.00000404, -0.00000206, 0.00000407, -0.00000077, 0.00000130, -0.00000017, -0.00000141, -0.00000499, 0.00000610, -0.00000192, 0.00000026, 0.00000142, 0.00000036, 0.00000043, -0.00000250, 0.00000105, -0.00000013, 0.00000036, -0.00000145, -0.00000282, -0.00000236, 0.00000159, 0.00000226, 0.00000737, -0.00000421, -0.00000174, 0.00000088, -0.00000150, 0.00000064, -0.00000183, 0.00000252, -0.00000357, 0.00000205, 0.00000062, -0.00000482, -0.00000042, 0.00000380, 0.00000100, 0.00000136, -0.00000051, -0.00000334, 0.00000037, 0.00000370, 0.00000013, -0.00000203, -0.00000172, -0.00000327, -0.00000245, 0.00000138, 0.00000272, 0.00000207, 0.00000314, -0.00000408, -0.00000747, 0.00000299, -0.00000567, 0.00000793, 0.00000504, 0.00000225, -0.00000804, -0.00000052, -0.00000229, 0.00000717, 0.00000582, -0.00000526, -0.00000259, 0.00000438, 0.00000324, -0.00000626, -0.00001355, 0.00000667, -0.00000740, -0.00000679, },
		{ -0.00000495, -0.00000032, 0.00000941, -0.00002631, -0.00000066, -0.00001396, -0.00002589, -0.00001242, 0.00003130, 0.00004105, 0.00001817, 0.00003370, 0.00001407, -0.00002695, -0.00002283, -0.00002363, -0.00000939, -0.00008126, -0.00003948, 0.00015599, 0.00011761, 0.00011463, -0.00012136, -0.00001126, -0.00002195, -0.00001751, -0.00000665, -0.00000716, -0.00001581, 0.00000591, -0.00001271, 0.00001063, -0.00001784, -0.00000239, -0.00000916, 0.00000635, 0.00000644, -0.00001225, 0.00001467, -0.00000174, 0.00000234, -0.00000231, -0.00000171, -0.00000363, 0.00000322, 0.00000901, -0.00000519, 0.00000238, -0.00000029, -0.00000182, -0.00001066, 0.00000922, -0.00000369, 0.00000632, -0.00000676, 0.00000316, -0.00000591, 0.00000006, 0.00000282, -0.00000005, 0.00000074, 0.00000037, -0.00000259, -0.00000387, 0.00000608, -0.00000252, 0.00000070, 0.00000123, 0.00000065, -0.00000012, -0.00000209, 0.00000090, 0.00000010, 0.00000009, -0.00000148, -0.00000315, -0.00000175, 0.00000166, 0.00000264, 0.00000710, -0.00000546, -0.00000057, 0.00000043, -0.00000114, 0.00000035, -0.00000157, 0.00000242, -0.00000383, 0.00000288, -0.00000033, -0.00000486, 0.00000010, 0.00000396, 0.00000050, 0.00000202, -0.00000141, -0.00000284, 0.00000027, 0.00000437, -0.00000078, -0.00000181, -0.00000187, -0.00000338, -0.00000213, 0.00000168, 0.00000250, 0.00000251, 0.00000271, -0.00000501, -0.00000654, 0.00000297, -0.00000580, 0.00000915, 0.00000373, 0.00000245, -0.00000855, 0.00000000, -0.00000254, 0.00000880, 0.00000364, -0.00000468, -0.00000312, 0.00000659, 0.00000085, -0.00000645, -0.00001283, 0.00000803, -0.00000984, },
		{ -0.00001202, -0.00000320, -0.00000078, 0.00001036, -0.00002944, 0.00000387, -0.00001958, -0.00002126, -0.00001126, 0.00003522, 0.00003499, 0.00002006, 0.00003019, 0.00001277, -0.00002840, -0.00001999, -0.00002625, -0.00000649, -0.00009154, -0.00002699, 0.00016139, 0.00013178, 0.00009293, -0.00012965, 0.00000091, -0.00002886, -0.00001241, -0.00000938, -0.00000538, -0.00001708, 0.00000752, -0.00001432, 0.00001265, -0.00002149, 0.00000221, -0.00001359, 0.00001113, 0.00000265, -0.00000983, 0.00001463, -0.00000299, 0.00000296, -0.00000286, -0.00000169, -0.00000353, 0.00000464, 0.00000731, -0.00000488, 0.00000224, 0.00000065, -0.00000335, -0.00000969, 0.00000930, -0.00000394, 0.00000657, -0.00000774, 0.00000446, -0.00000755, 0.00000215, 0.00000144, 0.00000081, 0.00000011, 0.00000092, -0.00000387, -0.00000244, 0.00000567, -0.00000284, 0.00000100, 0.00000115, 0.00000082, -0.00000059, -0.00000168, 0.00000073, 0.00000038, -0.00000027, -0.00000146, -0.00000352, -0.00000100, 0.00000152, 0.00000321, 0.00000648, -0.00000643, 0.00000059, -0.00000010, -0.00000072, -0.00000003, -0.00000118, 0.00000215, -0.00000389, 0.00000357, -0.00000133, -0.00000481, 0.00000071, 0.00000391, 0.00000021, 0.00000255, -0.00000230, -0.00000228, 0.00000015, 0.00000504, -0.00000185, -0.00000138, -0.00000220, -0.00000334, -0.00000184, 0.00000194, 0.00000230, 0.00000301, 0.00000208, -0.00000579, -0.00000548, 0.00000270, -0.00000562, 0.00001008, 0.00000251, 0.00000264, -0.00000897, 0.00000035, -0.00000243, 0.00000998, 0.00000167, -0.00000424, -0.00000328, 0.00000851, -0.00000174, -0.00000633, -0.00001197, 0.00000898, },
		{ 0.00000939, -0.00001376, -0.00000176, -0.00000082, 0.00001064, -0.00003185, 0.00000793, -0.00002524, -0.00001570, -0.00001089, 0.00003897, 0.00002872, 0.00002255, 0.00002607, 0.00001204, -0.00002992, -0.00001725, -0.00002875, -0.00000397, -0.00010181, -0.00001362, 0.00016761, 0.00014563, 0.00006613, -0.00013171, 0.00000992, -0.00003423, -0.00000826, -0.00001141, -0.00000429, -0.00001760, 0.00000837, -0.00001511, 0.00001372, -0.00002416, 0.00000616, -0.00001787, 0.00001627, -0.00000184, -0.00000650, 0.00001367, -0.00000367, 0.00000333, -0.00000330, -0.00000176, -0.00000319, 0.00000580, 0.00000557, -0.00000440, 0.00000205, 0.00000170, -0.00000524, -0.00000818, 0.00000884, -0.00000367, 0.00000625, -0.00000814, 0.00000526, -0.00000879, 0.00000404, 0.00000006, 0.00000170, -0.00000051, 0.00000138, -0.00000513, -0.00000080, 0.00000491, -0.00000289, 0.00000113, 0.00000120, 0.00000082, -0.00000093, -0.00000132, 0.00000057, 0.00000065, -0.00000066, -0.00000140, -0.00000387, -0.00000014, 0.00000122, 0.00000396, 0.00000552, -0.00000706, 0.00000166, -0.00000065, -0.00000027, -0.00000045, -0.00000070, 0.00000173, -0.00000373, 0.00000408, -0.00000233, -0.00000468, 0.00000139, 0.00000364, 0.00000017, 0.00000287, -0.00000311, -0.00000174, 0.00000009, 0.00000565, -0.00000299, -0.00000078, -0.00000269, -0.00000318, -0.00000157, 0.00000216, 0.00000214, 0.00000351, 0.00000126, -0.00000639, -0.00000435, 0.00000221, -0.00000513, 0.00001066, 0.00000145, 0.00000277, -0.00000929, 0.00000052, -0.00000195, 0.00001065, 0.00000002, -0.00000404, -0.00000293, 0.00000993, -0.00000429, -0.00000603, -0.00001090, },
		{ -0.00000955, 0.00000917, -0.00001489, -0.00000080, -0.00000026, 0.00001000, -0.00003320, 0.00001104, -0.00003037, -0.00000969, -0.00001110, 0.00004242, 0.00002247, 0.00002539, 0.00002156, 0.00001172, -0.00003137, -0.00001478, -0.00003088, -0.00000217, -0.00011166, 0.00000023, 0.00017540, 0.00015769, 0.00003544, -0.00012767, 0.00001527, -0.00003755, -0.00000548, -0.00001239, -0.00000417, -0.00001713, 0.00000822, -0.00001488, 0.00001358, -0.00002550, 0.00000899, -0.00002149, 0.00002129, -0.00000667, -0.00000248, 0.00001188, -0.00000379, 0.00000343, -0.00000361, -0.00000192, -0.00000259, 0.00000664, 0.00000389, -0.00000384, 0.00000190, 0.00000272, -0.00000737, -0.00000621, 0.00000786, -0.00000289, 0.00000537, -0.00000793, 0.00000548, -0.00000949, 0.00000559, -0.00000118, 0.00000251, -0.00000103, 0.00000166, -0.00000622, 0.00000092, 0.00000390, -0.00000269, 0.00000111, 0.00000138, 0.00000065, -0.00000111, -0.00000105, 0.00000048, 0.00000088, -0.00000107, -0.00000136, -0.00000415, 0.00000075, 0.00000079, 0.00000483, 0.00000422, -0.00000729, 0.00000258, -0.00000118, 0.00000015, -0.00000088, -0.00000018, 0.00000120, -0.00000337, 0.00000439, -0.00000332, -0.00000444, 0.00000211, 0.00000316, 0.00000040, 0.00000291, -0.00000373, -0.00000127, 0.00000016, 0.00000610, -0.00000410, -0.00000009, -0.00000328, -0.00000290, -0.00000131, 0.00000230, 0.00000207, 0.00000396, 0.00000029, -0.00000678, -0.00000321, 0.00000154, -0.00000435, 0.00001089, 0.00000059, 0.00000281, -0.00000950, 0.00000055, -0.00000114, 0.00001080, -0.00000126, -0.00000413, -0.00000197, 0.00001065, -0.00000658, -0.00000569, },
		{ -0.00000540, -0.00000791, 0.00000828, -0.00001531, -0.00000043, 0.00000103, 0.00000830, -0.00003322, 0.00001279, -0.00003444, -0.00000378, -0.00001156, 0.00004535, 0.00001652, 0.00002830, 0.00001690, 0.00001162, -0.00003260, -0.00001278, -0.00003242, -0.00000140, -0.00012072, 0.00001420, 0.00018548, 0.00016638, 0.00000249, -0.00011813, 0.00001684, -0.00003853, -0.00000434, -0.00001213, -0.00000518, -0.00001552, 0.00000694, -0.00001348, 0.00001210, -0.00002527, 0.00001034, -0.00002398, 0.00002575, -0.00001145, 0.00000192, 0.00000940, -0.00000336, 0.00000325, -0.00000378, -0.00000218, -0.00000173, 0.00000709, 0.00000236, -0.00000326, 0.00000188, 0.00000358, -0.00000957, -0.00000391, 0.00000647, -0.00000168, 0.00000396, -0.00000710, 0.00000506, -0.00000959, 0.00000666, -0.00000215, 0.00000316, -0.00000138, 0.00000165, -0.00000702, 0.00000257, 0.00000273, -0.00000230, 0.00000095, 0.00000167, 0.00000033, -0.00000113, -0.00000090, 0.00000049, 0.00000101, -0.00000143, -0.00000135, -0.00000431, 0.00000161, 0.00000032, 0.00000577, 0.00000264, -0.00000712, 0.00000328, -0.00000163, 0.00000052, -0.00000126, 0.00000034, 0.00000061, -0.00000284, 0.00000450, -0.00000427, -0.00000408, 0.00000282, 0.00000252, 0.00000091, 0.00000263, -0.00000412, -0.00000094, 0.00000041, 0.00000631, -0.00000508, 0.00000060, -0.00000393, -0.00000254, -0.00000105, 0.00000236, 0.00000213, 0.00000429, -0.00000078, -0.00000695, -0.00000210, 0.00000075, -0.00000332, 0.00001076, -0.00000003, 0.00000275, -0.00000963, 0.00000049, -0.00000004, 0.00001044, -0.00000215, -0.00000452, -0.00000040, 0.00001056, -0.00000844, },
		{ -0.00000976, -0.00000523, -0.00000598, 0.00000673, -0.00001501, -0.00000068, 0.00000307, 0.00000547, -0.00003176, 0.00001289, -0.00003695, 0.00000149, -0.00001192, 0.00004754, 0.00001117, 0.00003099, 0.00001233, 0.00001157, -0.00003346, -0.00001138, -0.00003319, -0.00000189, -0.00012867, 0.00002808, 0.00019842, 0.00017011, -0.00003083, -0.00010416, 0.00001491, -0.00003719, -0.00000487, -0.00001060, -0.00000732, -0.00001279, 0.00000454, -0.00001093, 0.00000926, -0.00002339, 0.00000995, -0.00002497, 0.00002919, -0.00001578, 0.00000639, 0.00000647, -0.00000249, 0.00000281, -0.00000383, -0.00000248, -0.00000064, 0.00000714, 0.00000108, -0.00000276, 0.00000208, 0.00000412, -0.00001166, -0.00000146, 0.00000480, -0.00000015, 0.00000213, -0.00000573, 0.00000405, -0.00000905, 0.00000720, -0.00000279, 0.00000358, -0.00000151, 0.00000131, -0.00000742, 0.00000401, 0.00000152, -0.00000179, 0.00000071, 0.00000204, -0.00000012, -0.00000099, -0.00000088, 0.00000062, 0.00000101, -0.00000173, -0.00000142, -0.00000431, 0.00000235, -0.00000011, 0.00000669, 0.00000084, -0.00000655, 0.00000373, -0.00000198, 0.00000080, -0.00000157, 0.00000082, -0.00000001, -0.00000218, 0.00000440, -0.00000517, -0.00000359, 0.00000346, 0.00000178, 0.00000164, 0.00000204, -0.00000424, -0.00000079, 0.00000088, 0.00000623, -0.00000587, 0.00000122, -0.00000458, -0.00000212, -0.00000078, 0.00000230, 0.00000235, 0.00000444, -0.00000188, -0.00000689, -0.00000109, -0.00000010, -0.00000209, 0.00001030, -0.00000041, 0.00000257, -0.00000969, 0.00000040, 0.00000124, 0.00000965, -0.00000270, -0.00000514, 0.00000173, 0.00000962, },
		{ 0.00000787, -0.00001050, -0.00000519, -0.00000385, 0.00000462, -0.00001406, -0.00000150, 0.00000576, 0.00000158, -0.00002882, 0.00001118, -0.00003755, 0.00000565, -0.00001182, 0.00004876, 0.00000668, 0.00003318, 0.00000808, 0.00001138, -0.00003386, -0.00001068, -0.00003310, -0.00000379, -0.00013534, 0.00004182, 0.00021452, 0.00016744, -0.00006244, -0.00008716, 0.00001010, -0.00003385, -0.00000687, -0.00000795, -0.00001042, -0.00000908, 0.00000115, -0.00000738, 0.00000521, -0.00001992, 0.00000776, -0.00002421, 0.00003127, -0.00001926, 0.00001056, 0.00000332, -0.00000130, 0.00000217, -0.00000381, -0.00000279, 0.00000063, 0.00000678, 0.00000008, -0.00000237, 0.00000256, 0.00000421, -0.00001344, 0.00000097, 0.00000302, 0.00000154, 0.00000003, -0.00000393, 0.00000253, -0.00000791, 0.00000719, -0.00000305, 0.00000374, -0.00000140, 0.00000061, -0.00000734, 0.00000513, 0.00000039, -0.00000126, 0.00000044, 0.00000243, -0.00000066, -0.00000071, -0.00000098, 0.00000087, 0.00000088, -0.00000194, -0.00000157, -0.00000412, 0.00000292, -0.00000043, 0.00000748, -0.00000108, -0.00000563, 0.00000392, -0.00000219, 0.00000098, -0.00000178, 0.00000121, -0.00000059, -0.00000142, 0.00000411, -0.00000600, -0.00000295, 0.00000396, 0.00000102, 0.00000253, 0.00000116, -0.00000408, -0.00000083, 0.00000158, 0.00000583, -0.00000640, 0.00000170, -0.00000516, -0.00000168, -0.00000050, 0.00000215, 0.00000276, 0.00000436, -0.00000296, -0.00000663, -0.00000020, -0.00000094, -0.00000073, 0.00000954, -0.00000055, 0.00000230, -0.00000972, 0.00000036, 0.00000258, 0.00000855, -0.00000301, -0.00000585, 0.00000429, },
		{ 0.00000705, 0.00000545, -0.00001069, -0.00000523, -0.00000166, 0.00000212, -0.00001259, -0.00000272, 0.00000894, -0.00000315, -0.00002456, 0.00000771, -0.00003607, 0.00000832, -0.00001097, 0.00004884, 0.00000325, 0.00003465, 0.00000432, 0.00001094, -0.00003374, -0.00001070, -0.00003214, -0.00000712, -0.00014071, 0.00005566, 0.00023369, 0.00015724, -0.00009033, -0.00006877, 0.00000334, -0.00002906, -0.00000996, -0.00000453, -0.00001418, -0.00000470, -0.00000294, -0.00000312, 0.00000024, -0.00001511, 0.00000385, -0.00002161, 0.00003176, -0.00002160, 0.00001412, 0.00000020, 0.00000007, 0.00000139, -0.00000374, -0.00000304, 0.00000199, 0.00000604, -0.00000061, -0.00000215, 0.00000336, 0.00000375, -0.00001475, 0.00000317, 0.00000132, 0.00000320, -0.00000217, -0.00000189, 0.00000063, -0.00000627, 0.00000666, -0.00000294, 0.00000365, -0.00000109, -0.00000043, -0.00000676, 0.00000583, -0.00000055, -0.00000077, 0.00000020, 0.00000280, -0.00000124, -0.00000033, -0.00000119, 0.00000124, 0.00000059, -0.00000203, -0.00000182, -0.00000371, 0.00000326, -0.00000058, 0.00000806, -0.00000299, -0.00000442, 0.00000385, -0.00000227, 0.00000104, -0.00000188, 0.00000150, -0.00000110, -0.00000062, 0.00000363, -0.00000675, -0.00000214, 0.00000427, 0.00000034, 0.00000349, 0.00000006, -0.00000368, -0.00000104, 0.00000248, 0.00000510, -0.00000664, 0.00000200, -0.00000565, -0.00000124, -0.00000021, 0.00000190, 0.00000335, 0.00000399, -0.00000393, -0.00000620, 0.00000052, -0.00000173, 0.00000069, 0.00000855, -0.00000048, 0.00000192, -0.00000975, 0.00000045, 0.00000386, 0.00000727, -0.00000318, -0.00000650, },
		{ -0.00000690, 0.00000979, 0.00000256, -0.00001041, -0.00000526, 0.00000042, -0.00000053, -0.00001083, -0.00000412, 0.00001229, -0.00000839, -0.00001931, 0.00000272, -0.00003255, 0.00000933, -0.00000918, 0.00004766, 0.00000102, 0.00003526, 0.00000119, 0.00001021, -0.00003314, -0.00001135, -0.00003046, -0.00001174, -0.00014494, 0.00007009, 0.00025543, 0.00013884, -0.00011264, -0.00005071, -0.00000426, -0.00002357, -0.00001356, -0.00000079, -0.00001816, -0.00000004, -0.00000734, 0.00000147, -0.00000521, -0.00000935, -0.00000149, -0.00001726, 0.00003056, -0.00002261, 0.00001683, -0.00000265, 0.00000147, 0.00000057, -0.00000370, -0.00000317, 0.00000338, 0.00000498, -0.00000098, -0.00000208, 0.00000449, 0.00000268, -0.00001544, 0.00000496, -0.00000012, 0.00000465, -0.00000425, 0.00000019, -0.00000145, -0.00000426, 0.00000571, -0.00000250, 0.00000334, -0.00000064, -0.00000174, -0.00000569, 0.00000607, -0.00000123, -0.00000041, 0.00000005, 0.00000308, -0.00000180, 0.00000010, -0.00000146, 0.00000170, 0.00000016, -0.00000202, -0.00000214, -0.00000309, 0.00000335, -0.00000048, 0.00000834, -0.00000478, -0.00000303, 0.00000357, -0.00000223, 0.00000100, -0.00000187, 0.00000166, -0.00000150, 0.00000017, 0.00000300, -0.00000741, -0.00000117, 0.00000433, -0.00000016, 0.00000440, -0.00000118, -0.00000307, -0.00000137, 0.00000353, 0.00000408, -0.00000658, 0.00000210, -0.00000602, -0.00000080, 0.00000006, 0.00000159, 0.00000410, 0.00000333, -0.00000474, -0.00000564, 0.00000107, -0.00000240, 0.00000210, 0.00000738, -0.00000022, 0.00000147, -0.00000981, 0.00000072, 0.00000495, 0.00000595, -0.00000336, },
		{ -0.00000367, -0.00000688, 0.00001224, -0.00000057, -0.00000982, -0.00000515, 0.00000218, -0.00000309, -0.00000906, -0.00000540, 0.00001548, -0.00001371, -0.00001352, -0.00000338, -0.00002720, 0.00000866, -0.00000639, 0.00004525, 0.00000001, 0.00003495, -0.00000130, 0.00000920, -0.00003217, -0.00001246, -0.00002828, -0.00001739, -0.00014837, 0.00008590, 0.00027871, 0.00011215, -0.00012791, -0.00003461, -0.00001152, -0.00001822, -0.00001704, 0.00000270, -0.00002186, 0.00000439, -0.00001159, 0.00000591, -0.00001065, -0.00000318, -0.00000782, -0.00001143, 0.00002777, -0.00002220, 0.00001850, -0.00000505, 0.00000277, -0.00000024, -0.00000370, -0.00000313, 0.00000469, 0.00000368, -0.00000106, -0.00000214, 0.00000588, 0.00000099, -0.00001542, 0.00000621, -0.00000114, 0.00000570, -0.00000602, 0.00000212, -0.00000353, -0.00000207, 0.00000446, -0.00000182, 0.00000291, -0.00000014, -0.00000321, -0.00000420, 0.00000583, -0.00000159, -0.00000023, 0.00000004, 0.00000323, -0.00000228, 0.00000054, -0.00000177, 0.00000223, -0.00000038, -0.00000192, -0.00000252, -0.00000230, 0.00000318, -0.00000013, 0.00000826, -0.00000635, -0.00000153, 0.00000312, -0.00000208, 0.00000088, -0.00000175, 0.00000170, -0.00000177, 0.00000092, 0.00000223, -0.00000795, -0.00000007, 0.00000412, -0.00000042, 0.00000517, -0.00000245, -0.00000235, -0.00000176, 0.00000468, 0.00000282, -0.00000624, 0.00000200, -0.00000626, -0.00000035, 0.00000029, 0.00000129, 0.00000495, 0.00000237, -0.00000533, -0.00000498, 0.00000144, -0.00000291, 0.00000340, 0.00000611, 0.00000019, 0.00000095, -0.00000990, 0.00000122, 0.00000574, 0.00000473, },
		{ 0.00000370, -0.00000421, -0.00000627, 0.00001414, -0.00000367, -0.00000908, -0.00000479, 0.00000343, -0.00000528, -0.00000755, -0.00000626, 0.00001812, -0.00001865, -0.00000774, -0.00001002, -0.00002048, 0.00000650, -0.00000269, 0.00004172, 0.00000013, 0.00003378, -0.00000319, 0.00000802, -0.00003100, -0.00001378, -0.00002592, -0.00002368, -0.00015145, 0.00010413, 0.00030200, 0.00007775, -0.00013520, -0.00002185, -0.00001735, -0.00001383, -0.00001972, 0.00000540, -0.00002475, 0.00000811, -0.00001521, 0.00000970, -0.00001551, 0.00000280, -0.00001458, -0.00000453, 0.00002359, -0.00002044, 0.00001907, -0.00000686, 0.00000387, -0.00000098, -0.00000380, -0.00000285, 0.00000582, 0.00000223, -0.00000090, -0.00000227, 0.00000747, -0.00000127, -0.00001466, 0.00000681, -0.00000161, 0.00000623, -0.00000730, 0.00000370, -0.00000538, 0.00000011, 0.00000306, -0.00000100, 0.00000243, 0.00000031, -0.00000471, -0.00000240, 0.00000516, -0.00000162, -0.00000026, 0.00000020, 0.00000320, -0.00000264, 0.00000094, -0.00000206, 0.00000277, -0.00000101, -0.00000175, -0.00000291, -0.00000137, 0.00000279, 0.00000047, 0.00000776, -0.00000758, -0.00000003, 0.00000254, -0.00000186, 0.00000069, -0.00000155, 0.00000162, -0.00000188, 0.00000159, 0.00000135, -0.00000835, 0.00000114, 0.00000362, -0.00000037, 0.00000568, -0.00000364, -0.00000159, -0.00000214, 0.00000582, 0.00000140, -0.00000569, 0.00000173, -0.00000639, 0.00000012, 0.00000042, 0.00000105, 0.00000583, 0.00000116, -0.00000568, -0.00000428, 0.00000162, -0.00000321, 0.00000455, 0.00000482, 0.00000070, 0.00000038, -0.00001004, 0.00000199, 0.00000617, },
		{ 0.00000620, 0.00000293, -0.00000502, -0.00000499, 0.00001529, -0.00000651, -0.00000833, -0.00000405, 0.00000400, -0.00000687, -0.00000655, -0.00000642, 0.00001985, -0.00002272, -0.00000256, -0.00001656, -0.00001294, 0.00000318, 0.00000170, 0.00003730, 0.00000122, 0.00003187, -0.00000460, 0.00000681, -0.00002987, -0.00001502, -0.00002378, -0.00003019, -0.00015473, 0.00012597, 0.00032334, 0.00003691, -0.00013422, -0.00001345, -0.00002086, -0.00001113, -0.00002104, 0.00000681, -0.00002637, 0.00001067, -0.00001776, 0.00001239, -0.00001927, 0.00000798, -0.00002114, 0.00000290, 0.00001839, -0.00001750, 0.00001855, -0.00000801, 0.00000469, -0.00000160, -0.00000399, -0.00000232, 0.00000670, 0.00000073, -0.00000058, -0.00000240, 0.00000912, -0.00000399, -0.00001318, 0.00000675, -0.00000149, 0.00000612, -0.00000798, 0.00000475, -0.00000682, 0.00000209, 0.00000167, -0.00000017, 0.00000200, 0.00000057, -0.00000608, -0.00000043, 0.00000414, -0.00000134, -0.00000050, 0.00000053, 0.00000299, -0.00000284, 0.00000124, -0.00000229, 0.00000327, -0.00000169, -0.00000153, -0.00000327, -0.00000037, 0.00000222, 0.00000130, 0.00000686, -0.00000842, 0.00000137, 0.00000191, -0.00000161, 0.00000049, -0.00000130, 0.00000145, -0.00000185, 0.00000214, 0.00000037, -0.00000857, 0.00000240, 0.00000285, 0.00000003, 0.00000585, -0.00000465, -0.00000089, -0.00000242, 0.00000687, -0.00000008, -0.00000498, 0.00000131, -0.00000642, 0.00000063, 0.00000042, 0.00000095, 0.00000666, -0.00000025, -0.00000578, -0.00000356, 0.00000163, -0.00000329, 0.00000547, 0.00000358, 0.00000126, -0.00000024, -0.00001020, 0.00000300, },
		{ 0.00000422, 0.00000582, 0.00000243, -0.00000611, -0.00000301, 0.00001553, -0.00000889, -0.00000772, -0.00000287, 0.00000377, -0.00000768, -0.00000627, -0.00000565, 0.00002038, -0.00002552, 0.00000147, -0.00002232, -0.00000526, -0.00000082, 0.00000647, 0.00003230, 0.00000300, 0.00002941, -0.00000569, 0.00000575, -0.00002901, -0.00001588, -0.00002222, -0.00003650, -0.00015870, 0.00015268, 0.00034042, -0.00000844, -0.00012532, -0.00000988, -0.00002145, -0.00001062, -0.00002056, 0.00000655, -0.00002638, 0.00001173, -0.00001891, 0.00001361, -0.00002147, 0.00001179, -0.00002686, 0.00001029, 0.00001261, -0.00001368, 0.00001707, -0.00000851, 0.00000521, -0.00000210, -0.00000429, -0.00000152, 0.00000725, -0.00000072, -0.00000018, -0.00000242, 0.00001069, -0.00000701, -0.00001109, 0.00000606, -0.00000077, 0.00000537, -0.00000798, 0.00000519, -0.00000772, 0.00000371, 0.00000043, 0.00000057, 0.00000172, 0.00000055, -0.00000716, 0.00000154, 0.00000288, -0.00000082, -0.00000095, 0.00000104, 0.00000259, -0.00000285, 0.00000141, -0.00000242, 0.00000371, -0.00000237, -0.00000131, -0.00000355, 0.00000062, 0.00000155, 0.00000229, 0.00000557, -0.00000882, 0.00000261, 0.00000127, -0.00000135, 0.00000028, -0.00000104, 0.00000121, -0.00000168, 0.00000256, -0.00000068, -0.00000860, 0.00000366, 0.00000186, 0.00000077, 0.00000563, -0.00000540, -0.00000031, -0.00000254, 0.00000774, -0.00000153, -0.00000419, 0.00000081, -0.00000638, 0.00000120, 0.00000024, 0.00000106, 0.00000732, -0.00000176, -0.00000564, -0.00000286, 0.00000148, -0.00000313, 0.00000612, 0.00000245, 0.00000185, -0.00000089, -0.00001036, },
		{ -0.00001047, 0.00000558, 0.00000510, 0.00000218, -0.00000740, -0.00000039, 0.00001482, -0.00001067, -0.00000731, -0.00000125, 0.00000273, -0.00000764, -0.00000682, -0.00000384, 0.00001952, -0.00002674, 0.00000391, -0.00002668, 0.00000189, -0.00000499, 0.00001121, 0.00002708, 0.00000519, 0.00002662, -0.00000663, 0.00000500, -0.00002862, -0.00001611, -0.00002156, -0.00004225, -0.00016374, 0.00018541, 0.00035075, -0.00005583, -0.00010959, -0.00001113, -0.00001892, -0.00001250, -0.00001809, 0.00000447, -0.00002461, 0.00001110, -0.00001845, 0.00001316, -0.00002182, 0.00001378, -0.00003117, 0.00001704, 0.00000675, -0.00000932, 0.00001481, -0.00000842, 0.00000543, -0.00000248, -0.00000466, -0.00000045, 0.00000741, -0.00000201, 0.00000023, -0.00000226, 0.00001201, -0.00001014, -0.00000854, 0.00000484, 0.00000047, 0.00000400, -0.00000730, 0.00000495, -0.00000798, 0.00000485, -0.00000053, 0.00000113, 0.00000165, 0.00000015, -0.00000781, 0.00000334, 0.00000152, -0.00000012, -0.00000154, 0.00000169, 0.00000200, -0.00000267, 0.00000143, -0.00000241, 0.00000402, -0.00000301, -0.00000112, -0.00000371, 0.00000152, 0.00000085, 0.00000336, 0.00000396, -0.00000875, 0.00000361, 0.00000067, -0.00000113, 0.00000012, -0.00000079, 0.00000094, -0.00000140, 0.00000284, -0.00000178, -0.00000839, 0.00000482, 0.00000073, 0.00000182, 0.00000501, -0.00000582, 0.00000007, -0.00000244, 0.00000835, -0.00000286, -0.00000340, 0.00000027, -0.00000630, 0.00000185, -0.00000014, 0.00000145, 0.00000772, -0.00000330, -0.00000528, -0.00000220, 0.00000121, -0.00000275, 0.00000648, 0.00000148, 0.00000242, -0.00000159, },
		{ -0.00000235, -0.00001046, 0.00000697, 0.00000413, 0.00000209, -0.00000876, 0.00000275, 0.00001320, -0.00001180, -0.00000712, 0.00000075, 0.00000094, -0.00000676, -0.00000817, -0.00000099, 0.00001724, -0.00002625, 0.00000446, -0.00002917, 0.00000791, -0.00000879, 0.00001555, 0.00002200, 0.00000748, 0.00002374, -0.00000757, 0.00000468, -0.00002884, -0.00001557, -0.00002198, -0.00004730, -0.00016997, 0.00022503, 0.00035186, -0.00010236, -0.00008870, -0.00001658, -0.00001351, -0.00001669, -0.00001371, 0.00000063, -0.00002110, 0.00000879, -0.00001640, 0.00001102, -0.00002020, 0.00001368, -0.00003364, 0.00002263, 0.00000128, -0.00000480, 0.00001203, -0.00000783, 0.00000538, -0.00000275, -0.00000508, 0.00000087, 0.00000715, -0.00000306, 0.00000056, -0.00000181, 0.00001290, -0.00001313, -0.00000575, 0.00000326, 0.00000210, 0.00000214, -0.00000603, 0.00000408, -0.00000757, 0.00000544, -0.00000115, 0.00000145, 0.00000182, -0.00000067, -0.00000792, 0.00000481, 0.00000020, 0.00000065, -0.00000223, 0.00000244, 0.00000127, -0.00000231, 0.00000130, -0.00000225, 0.00000417, -0.00000359, -0.00000097, -0.00000370, 0.00000227, 0.00000023, 0.00000439, 0.00000212, -0.00000826, 0.00000436, 0.00000015, -0.00000095, 0.00000002, -0.00000059, 0.00000069, -0.00000102, 0.00000297, -0.00000292, -0.00000793, 0.00000583, -0.00000045, 0.00000311, 0.00000400, -0.00000589, 0.00000024, -0.00000209, 0.00000866, -0.00000400, -0.00000266, -0.00000027, -0.00000618, 0.00000256, -0.00000074, 0.00000216, 0.00000777, -0.00000474, -0.00000476, -0.00000161, 0.00000085, -0.00000217, 0.00000655, 0.00000071, 0.00000295, },
		{ 0.00000344, -0.00000319, -0.00001027, 0.00000828, 0.00000302, 0.00000204, -0.00001002, 0.00000616, 0.00001080, -0.00001233, -0.00000708, 0.00000296, -0.00000144, -0.00000519, -0.00001020, 0.00000275, 0.00001368, -0.00002411, 0.00000305, -0.00002949, 0.00001231, -0.00001179, 0.00001913, 0.00001737, 0.00000963, 0.00002090, -0.00000860, 0.00000483, -0.00002968, -0.00001426, -0.00002348, -0.00005172, -0.00017713, 0.00027199, 0.00034153, -0.00014496, -0.00006480, -0.00002516, -0.00000583, -0.00002275, -0.00000775, -0.00000468, -0.00001610, 0.00000504, -0.00001296, 0.00000736, -0.00001672, 0.00001143, -0.00003401, 0.00002667, -0.00000335, -0.00000052, 0.00000897, -0.00000688, 0.00000511, -0.00000296, -0.00000549, 0.00000237, 0.00000645, -0.00000379, 0.00000074, -0.00000100, 0.00001321, -0.00001576, -0.00000295, 0.00000152, 0.00000394, -0.00000005, -0.00000429, 0.00000267, -0.00000656, 0.00000548, -0.00000140, 0.00000153, 0.00000221, -0.00000191, -0.00000744, 0.00000583, -0.00000094, 0.00000140, -0.00000293, 0.00000324, 0.00000042, -0.00000179, 0.00000101, -0.00000193, 0.00000415, -0.00000405, -0.00000088, -0.00000351, 0.00000281, -0.00000024, 0.00000529, 0.00000016, -0.00000739, 0.00000483, -0.00000028, -0.00000083, -0.00000002, -0.00000047, 0.00000048, -0.00000059, 0.00000295, -0.00000407, -0.00000720, 0.00000658, -0.00000158, 0.00000453, 0.00000267, -0.00000564, 0.00000018, -0.00000149, 0.00000865, -0.00000491, -0.00000201, -0.00000078, -0.00000604, 0.00000331, -0.00000152, 0.00000319, 0.00000740, -0.00000600, -0.00000411, -0.00000111, 0.00000045, -0.00000144, 0.00000634, 0.00000013, },
		{ -0.00000024, 0.00000388, -0.00000414, -0.00000982, 0.00000939, 0.00000192, 0.00000189, -0.00001097, 0.00000959, 0.00000784, -0.00001238, -0.00000705, 0.00000517, -0.00000413, -0.00000316, -0.00001266, 0.00000709, 0.00000913, -0.00002058, -0.00000019, -0.00002757, 0.00001483, -0.00001366, 0.00002171, 0.00001342, 0.00001146, 0.00001822, -0.00000971, 0.00000534, -0.00003100, -0.00001236, -0.00002584, -0.00005588, -0.00018455, 0.00032611, 0.00031801, -0.00018061, -0.00004035, -0.00003539, 0.00000315, -0.00002995, -0.00000080, -0.00001093, -0.00001010, 0.00000024, -0.00000852, 0.00000257, -0.00001171, 0.00000720, -0.00003221, 0.00002892, -0.00000683, 0.00000322, 0.00000590, -0.00000569, 0.00000469, -0.00000313, -0.00000587, 0.00000402, 0.00000531, -0.00000416, 0.00000073, 0.00000022, 0.00001282, -0.00001781, -0.00000038, -0.00000016, 0.00000576, -0.00000234, -0.00000229, 0.00000088, -0.00000506, 0.00000503, -0.00000130, 0.00000139, 0.00000278, -0.00000350, -0.00000635, 0.00000632, -0.00000180, 0.00000202, -0.00000358, 0.00000402, -0.00000048, -0.00000114, 0.00000059, -0.00000145, 0.00000395, -0.00000441, -0.00000086, -0.00000312, 0.00000309, -0.00000048, 0.00000594, -0.00000178, -0.00000622, 0.00000504, -0.00000061, -0.00000077, 0.00000003, -0.00000043, 0.00000035, -0.00000014, 0.00000279, -0.00000519, -0.00000621, 0.00000703, -0.00000253, 0.00000596, 0.00000109, -0.00000509, -0.00000006, -0.00000068, 0.00000833, -0.00000558, -0.00000148, -0.00000125, -0.00000584, 0.00000407, -0.00000244, 0.00000452, 0.00000658, -0.00000701, -0.00000337, -0.00000071, 0.00000005, -0.00000061, 0.00000589, },
		{ 0.00000527, -0.00000044, 0.00000427, -0.00000520, -0.00000906, 0.00001018, 0.00000096, 0.00000150, -0.00001138, 0.00001272, 0.00000461, -0.00001212, -0.00000685, 0.00000710, -0.00000682, -0.00000101, -0.00001519, 0.00001162, 0.00000403, -0.00001609, -0.00000486, -0.00002361, 0.00001539, -0.00001427, 0.00002319, 0.00001024, 0.00001291, 0.00001566, -0.00001077, 0.00000601, -0.00003252, -0.00001024, -0.00002858, -0.00006048, -0.00019102, 0.00038649, 0.00028024, -0.00020660, -0.00001789, -0.00004556, 0.00001223, -0.00003735, 0.00000635, -0.00001743, -0.00000372, -0.00000500, -0.00000362, -0.00000279, -0.00000570, 0.00000138, -0.00002843, 0.00002933, -0.00000898, 0.00000617, 0.00000301, -0.00000440, 0.00000415, -0.00000328, -0.00000615, 0.00000575, 0.00000377, -0.00000413, 0.00000050, 0.00000186, 0.00001163, -0.00001910, 0.00000173, -0.00000153, 0.00000732, -0.00000451, -0.00000026, -0.00000106, -0.00000323, 0.00000420, -0.00000094, 0.00000111, 0.00000343, -0.00000534, -0.00000472, 0.00000624, -0.00000230, 0.00000244, -0.00000412, 0.00000475, -0.00000138, -0.00000042, 0.00000006, -0.00000084, 0.00000356, -0.00000463, -0.00000090, -0.00000256, 0.00000309, -0.00000044, 0.00000626, -0.00000358, -0.00000486, 0.00000503, -0.00000085, -0.00000076, 0.00000014, -0.00000049, 0.00000031, 0.00000029, 0.00000248, -0.00000627, -0.00000495, 0.00000711, -0.00000321, 0.00000728, -0.00000060, -0.00000433, -0.00000043, 0.00000028, 0.00000774, -0.00000603, -0.00000106, -0.00000170, -0.00000556, 0.00000475, -0.00000341, 0.00000607, 0.00000533, -0.00000771, -0.00000260, -0.00000044, -0.00000028, 0.00000024, },
		{ 0.00000102, 0.00000454, -0.00000051, 0.00000464, -0.00000640, -0.00000794, 0.00001054, 0.00000027, 0.00000073, -0.00001106, 0.00001523, 0.00000142, -0.00001178, -0.00000628, 0.00000846, -0.00000914, 0.00000089, -0.00001739, 0.00001587, -0.00000105, -0.00001126, -0.00001040, -0.00001805, 0.00001419, -0.00001368, 0.00002361, 0.00000781, 0.00001404, 0.00001309, -0.00001156, 0.00000650, -0.00003382, -0.00000845, -0.00003102, -0.00006653, -0.00019484, 0.00045143, 0.00022807, -0.00022079, 0.00000022, -0.00005391, 0.00002007, -0.00004392, 0.00001282, -0.00002338, 0.00000231, -0.00001003, 0.00000106, -0.00000803, 0.00000063, -0.00000543, -0.00002306, 0.00002804, -0.00000976, 0.00000820, 0.00000048, -0.00000310, 0.00000354, -0.00000342, -0.00000632, 0.00000750, 0.00000188, -0.00000369, 0.00000006, 0.00000389, 0.00000963, -0.00001948, 0.00000318, -0.00000239, 0.00000841, -0.00000631, 0.00000155, -0.00000293, -0.00000129, 0.00000316, -0.00000043, 0.00000080, 0.00000401, -0.00000726, -0.00000265, 0.00000562, -0.00000241, 0.00000260, -0.00000448, 0.00000535, -0.00000225, 0.00000035, -0.00000054, -0.00000012, 0.00000301, -0.00000474, -0.00000097, -0.00000185, 0.00000284, -0.00000011, 0.00000617, -0.00000513, -0.00000339, 0.00000484, -0.00000103, -0.00000078, 0.00000029, -0.00000062, 0.00000039, 0.00000068, 0.00000204, -0.00000726, -0.00000348, 0.00000680, -0.00000353, 0.00000835, -0.00000228, -0.00000345, -0.00000086, 0.00000131, 0.00000697, -0.00000631, -0.00000071, -0.00000216, -0.00000514, 0.00000527, -0.00000432, 0.00000774, 0.00000370, -0.00000807, -0.00000182, -0.00000032, -0.00000048, },
		{ -0.00000049, 0.00000164, 0.00000381, -0.00000050, 0.00000500, -0.00000774, -0.00000643, 0.00001041, -0.00000006, -0.00000050, -0.00000985, 0.00001684, -0.00000142, -0.00001159, -0.00000512, 0.00000896, -0.00001072, 0.00000214, -0.00001883, 0.00001935, -0.00000552, -0.00000675, -0.00001608, -0.00001150, 0.00001163, -0.00001211, 0.00002318, 0.00000594, 0.00001503, 0.00001029, -0.00001176, 0.00000638, -0.00003435, -0.00000766, -0.00003231, -0.00007524, -0.00019385, 0.00051839, 0.00016235, -0.00022181, 0.00001197, -0.00005882, 0.00002543, -0.00004859, 0.00001772, -0.00002797, 0.00000724, -0.00001414, 0.00000485, -0.00001245, 0.00000652, -0.00001252, -0.00001664, 0.00002540, -0.00000930, 0.00000928, -0.00000164, -0.00000184, 0.00000287, -0.00000355, -0.00000636, 0.00000919, -0.00000032, -0.00000287, -0.00000058, 0.00000625, 0.00000684, -0.00001890, 0.00000385, -0.00000258, 0.00000885, -0.00000753, 0.00000292, -0.00000449, 0.00000055, 0.00000207, 0.00000010, 0.00000057, 0.00000439, -0.00000909, -0.00000031, 0.00000454, -0.00000214, 0.00000249, -0.00000466, 0.00000581, -0.00000302, 0.00000111, -0.00000117, 0.00000067, 0.00000231, -0.00000473, -0.00000105, -0.00000104, 0.00000237, 0.00000050, 0.00000564, -0.00000631, -0.00000193, 0.00000453, -0.00000116, -0.00000079, 0.00000047, -0.00000083, 0.00000058, 0.00000100, 0.00000149, -0.00000811, -0.00000183, 0.00000610, -0.00000342, 0.00000905, -0.00000383, -0.00000255, -0.00000126, 0.00000232, 0.00000610, -0.00000646, -0.00000040, -0.00000267, -0.00000453, 0.00000554, -0.00000505, 0.00000940, 0.00000177, -0.00000810, -0.00000104, -0.00000036, },
		{ -0.00000058, -0.00000026, 0.00000203, 0.00000316, -0.00000047, 0.00000537, -0.00000921, -0.00000454, 0.00000974, 0.00000003, -0.00000222, -0.00000766, 0.00001733, -0.00000364, -0.00001176, -0.00000321, 0.00000836, -0.00001126, 0.00000239, -0.00001912, 0.00002157, -0.00000882, -0.00000327, -0.00002112, -0.00000474, 0.00000828, -0.00000997, 0.00002224, 0.00000434, 0.00001615, 0.00000695, -0.00001098, 0.00000515, -0.00003351, -0.00000863, -0.00003153, -0.00008800, -0.00018556, 0.00058408, 0.00008497, -0.00020922, 0.00001595, -0.00005900, 0.00002720, -0.00005046, 0.00002024, -0.00003047, 0.00001038, -0.00001669, 0.00000711, -0.00001533, 0.00001119, -0.00001908, -0.00000987, 0.00002186, -0.00000788, 0.00000951, -0.00000331, -0.00000065, 0.00000212, -0.00000364, -0.00000627, 0.00001077, -0.00000273, -0.00000169, -0.00000137, 0.00000882, 0.00000336, -0.00001737, 0.00000366, -0.00000201, 0.00000850, -0.00000801, 0.00000366, -0.00000552, 0.00000206, 0.00000113, 0.00000050, 0.00000056, 0.00000440, -0.00001063, 0.00000211, 0.00000313, -0.00000153, 0.00000211, -0.00000463, 0.00000611, -0.00000369, 0.00000183, -0.00000179, 0.00000150, 0.00000149, -0.00000463, -0.00000110, -0.00000021, 0.00000174, 0.00000135, 0.00000468, -0.00000706, -0.00000058, 0.00000415, -0.00000128, -0.00000078, 0.00000065, -0.00000107, 0.00000089, 0.00000121, 0.00000082, -0.00000880, -0.00000008, 0.00000506, -0.00000287, 0.00000931, -0.00000512, -0.00000174, -0.00000154, 0.00000320, 0.00000523, -0.00000658, -0.00000006, -0.00000327, -0.00000367, 0.00000547, -0.00000545, 0.00001087, -0.00000032, -0.00000786, -0.00000027, },
		{ 0.00000049, -0.00000100, 0.00000024, 0.00000211, 0.00000267, -0.00000047, 0.00000576, -0.00001080, -0.00000230, 0.00000856, 0.00000054, -0.00000441, -0.00000450, 0.00001658, -0.00000502, -0.00001244, -0.00000045, 0.00000653, -0.00001054, 0.00000140, -0.00001796, 0.00002216, -0.00001043, -0.00000145, -0.00002473, 0.00000142, 0.00000481, -0.00000777, 0.00002122, 0.00000268, 0.00001771, 0.00000275, -0.00000884, 0.00000235, -0.00003075, -0.00001206, -0.00002777, -0.00010615, -0.00016730, 0.00064462, -0.00000111, -0.00018364, 0.00001149, -0.00005361, 0.00002465, -0.00004886, 0.00001982, -0.00003035, 0.00001121, -0.00001717, 0.00000732, -0.00001610, 0.00001394, -0.00002434, -0.00000346, 0.00001799, -0.00000584, 0.00000908, -0.00000461, 0.00000050, 0.00000127, -0.00000365, -0.00000606, 0.00001219, -0.00000529, -0.00000022, -0.00000224, 0.00001147, -0.00000064, -0.00001497, 0.00000263, -0.00000066, 0.00000732, -0.00000766, 0.00000363, -0.00000586, 0.00000306, 0.00000049, 0.00000065, 0.00000086, 0.00000393, -0.00001168, 0.00000440, 0.00000153, -0.00000068, 0.00000151, -0.00000441, 0.00000623, -0.00000422, 0.00000248, -0.00000238, 0.00000231, 0.00000059, -0.00000445, -0.00000109, 0.00000057, 0.00000103, 0.00000236, 0.00000333, -0.00000734, 0.00000059, 0.00000376, -0.00000141, -0.00000073, 0.00000079, -0.00000133, 0.00000130, 0.00000132, 0.00000007, -0.00000927, 0.00000168, 0.00000372, -0.00000187, 0.00000907, -0.00000606, -0.00000110, -0.00000162, 0.00000387, 0.00000444, -0.00000674, 0.00000036, -0.00000401, -0.00000252, 0.00000497, -0.00000541, 0.00001200, -0.00000243, -0.00000738, },
		{ -0.00000676, 0.00000125, -0.00000162, 0.00000105, 0.00000183, 0.00000241, -0.00000055, 0.00000617, -0.00001245, 0.00000022, 0.00000691, 0.00000141, -0.00000696, -0.00000047, 0.00001457, -0.00000547, -0.00001369, 0.00000315, 0.00000341, -0.00000844, -0.00000099, -0.00001520, 0.00002091, -0.00001004, -0.00000177, -0.00002625, 0.00000619, 0.00000191, -0.00000606, 0.00002057, 0.00000056, 0.00002001, -0.00000261, -0.00000501, -0.00000241, -0.00002559, -0.00001853, -0.00002030, -0.00013084, -0.00013647, 0.00069568, -0.00009206, -0.00014668, -0.00000127, -0.00004242, 0.00001746, -0.00004348, 0.00001615, -0.00002731, 0.00000945, -0.00001530, 0.00000521, -0.00001438, 0.00001423, -0.00002764, 0.00000189, 0.00001438, -0.00000362, 0.00000823, -0.00000563, 0.00000164, 0.00000028, -0.00000353, -0.00000576, 0.00001339, -0.00000789, 0.00000148, -0.00000310, 0.00001402, -0.00000496, -0.00001187, 0.00000088, 0.00000141, 0.00000534, -0.00000646, 0.00000278, -0.00000541, 0.00000344, 0.00000027, 0.00000047, 0.00000154, 0.00000287, -0.00001207, 0.00000634, -0.00000007, 0.00000031, 0.00000074, -0.00000402, 0.00000621, -0.00000461, 0.00000305, -0.00000291, 0.00000309, -0.00000036, -0.00000422, -0.00000097, 0.00000124, 0.00000034, 0.00000343, 0.00000165, -0.00000716, 0.00000154, 0.00000339, -0.00000158, -0.00000060, 0.00000087, -0.00000157, 0.00000179, 0.00000130, -0.00000075, -0.00000949, 0.00000338, 0.00000219, -0.00000049, 0.00000832, -0.00000660, -0.00000069, -0.00000147, 0.00000426, 0.00000381, -0.00000698, 0.00000090, -0.00000491, -0.00000105, 0.00000399, -0.00000482, 0.00001264, -0.00000440, },
		{ -0.00000610, -0.00000607, 0.00000202, -0.00000241, 0.00000215, 0.00000117, 0.00000243, -0.00000074, 0.00000660, -0.00001411, 0.00000292, 0.00000489, 0.00000254, -0.00000970, 0.00000424, 0.00001139, -0.00000497, -0.00001548, 0.00000750, -0.00000088, -0.00000501, -0.00000475, -0.00001084, 0.00001778, -0.00000752, -0.00000450, -0.00002520, 0.00000894, 0.00000020, -0.00000531, 0.00002071, -0.00000232, 0.00002331, -0.00000931, 0.00000070, -0.00000937, -0.00001778, -0.00002838, -0.00000872, -0.00016284, -0.00009078, 0.00073288, -0.00018335, -0.00010092, -0.00002134, -0.00002583, 0.00000581, -0.00003442, 0.00000929, -0.00002137, 0.00000508, -0.00001106, 0.00000071, -0.00001004, 0.00001177, -0.00002849, 0.00000560, 0.00001157, -0.00000164, 0.00000723, -0.00000650, 0.00000283, -0.00000089, -0.00000324, -0.00000539, 0.00001434, -0.00001045, 0.00000332, -0.00000386, 0.00001630, -0.00000934, -0.00000830, -0.00000144, 0.00000406, 0.00000268, -0.00000451, 0.00000115, -0.00000417, 0.00000312, 0.00000055, -0.00000011, 0.00000264, 0.00000119, -0.00001169, 0.00000777, -0.00000151, 0.00000132, -0.00000013, -0.00000352, 0.00000605, -0.00000487, 0.00000352, -0.00000335, 0.00000380, -0.00000132, -0.00000394, -0.00000074, 0.00000173, -0.00000023, 0.00000444, -0.00000023, -0.00000654, 0.00000222, 0.00000307, -0.00000178, -0.00000039, 0.00000087, -0.00000176, 0.00000233, 0.00000114, -0.00000162, -0.00000943, 0.00000491, 0.00000057, 0.00000118, 0.00000710, -0.00000672, -0.00000053, -0.00000104, 0.00000435, 0.00000339, -0.00000736, 0.00000158, -0.00000596, 0.00000072, 0.00000253, -0.00000362, 0.00001267, },
		{ 0.00001201, -0.00000741, -0.00000539, 0.00000278, -0.00000334, 0.00000351, 0.00000013, 0.00000276, -0.00000106, 0.00000703, -0.00001571, 0.00000569, 0.00000264, 0.00000378, -0.00001241, 0.00000935, 0.00000726, -0.00000365, -0.00001767, 0.00001237, -0.00000612, -0.00000046, -0.00000969, -0.00000508, 0.00001295, -0.00000299, -0.00000965, -0.00002137, 0.00000925, 0.00000015, -0.00000592, 0.00002194, -0.00000618, 0.00002771, -0.00001741, 0.00000829, -0.00001851, -0.00000734, -0.00004158, 0.00000697, -0.00020234, -0.00002855, 0.00075200, -0.00027003, -0.00004972, -0.00004700, -0.00000485, -0.00000959, -0.00002218, -0.00000033, -0.00001291, -0.00000159, -0.00000472, -0.00000592, -0.00000326, 0.00000654, -0.00002665, 0.00000726, 0.00000999, -0.00000029, 0.00000634, -0.00000734, 0.00000411, -0.00000225, -0.00000274, -0.00000497, 0.00001499, -0.00001285, 0.00000522, -0.00000444, 0.00001812, -0.00001351, -0.00000453, -0.00000407, 0.00000706, -0.00000047, -0.00000197, -0.00000114, -0.00000219, 0.00000214, 0.00000132, -0.00000106, 0.00000410, -0.00000109, -0.00001050, 0.00000855, -0.00000264, 0.00000223, -0.00000102, -0.00000295, 0.00000579, -0.00000501, 0.00000389, -0.00000370, 0.00000440, -0.00000228, -0.00000363, -0.00000038, 0.00000199, -0.00000060, 0.00000527, -0.00000220, -0.00000556, 0.00000264, 0.00000281, -0.00000202, -0.00000011, 0.00000077, -0.00000188, 0.00000290, 0.00000083, -0.00000251, -0.00000908, 0.00000618, -0.00000101, 0.00000304, 0.00000547, -0.00000643, -0.00000063, -0.00000037, 0.00000414, 0.00000318, -0.00000786, 0.00000239, -0.00000714, 0.00000274, 0.00000063, -0.00000182, },
		{ 0.00000054, 0.00001063, -0.00000823, -0.00000476, 0.00000353, -0.00000436, 0.00000506, -0.00000124, 0.00000336, -0.00000150, 0.00000741, -0.00001715, 0.00000837, 0.00000032, 0.00000495, -0.00001481, 0.00001448, 0.00000249, -0.00000172, -0.00002002, 0.00001743, -0.00001194, 0.00000487, -0.00001545, 0.00000170, 0.00000682, 0.00000316, -0.00001693, -0.00001486, 0.00000696, 0.00000198, -0.00000808, 0.00002439, -0.00001105, 0.00003316, -0.00002675, 0.00001755, -0.00002955, 0.00000537, -0.00005770, 0.00002617, -0.00024878, 0.00005107, 0.00074941, -0.00034710, 0.00000306, -0.00007588, 0.00001894, -0.00002759, -0.00000769, -0.00001193, -0.00000258, -0.00000997, 0.00000317, -0.00001414, 0.00000552, -0.00000115, -0.00002217, 0.00000671, 0.00000993, 0.00000015, 0.00000578, -0.00000826, 0.00000550, -0.00000380, -0.00000203, -0.00000453, 0.00001530, -0.00001499, 0.00000707, -0.00000475, 0.00001932, -0.00001719, -0.00000086, -0.00000675, 0.00001014, -0.00000383, 0.00000094, -0.00000388, 0.00000035, 0.00000058, 0.00000251, -0.00000232, 0.00000584, -0.00000386, -0.00000851, 0.00000862, -0.00000336, 0.00000295, -0.00000186, -0.00000234, 0.00000546, -0.00000504, 0.00000416, -0.00000394, 0.00000489, -0.00000321, -0.00000329, 0.00000010, 0.00000199, -0.00000070, 0.00000581, -0.00000411, -0.00000430, 0.00000283, 0.00000262, -0.00000228, 0.00000024, 0.00000057, -0.00000189, 0.00000347, 0.00000039, -0.00000338, -0.00000843, 0.00000713, -0.00000243, 0.00000493, 0.00000356, -0.00000581, -0.00000094, 0.00000050, 0.00000367, 0.00000317, -0.00000848, 0.00000329, -0.00000838, 0.00000492, -0.00000162, },
		{ -0.00000409, 0.00000334, 0.00000858, -0.00000854, -0.00000422, 0.00000421, -0.00000537, 0.00000670, -0.00000285, 0.00000420, -0.00000201, 0.00000769, -0.00001833, 0.00001082, -0.00000190, 0.00000586, -0.00001663, 0.00001925, -0.00000251, 0.00000051, -0.00002222, 0.00002227, -0.00001786, 0.00001051, -0.00002155, 0.00000897, -0.00000003, 0.00001033, -0.00002576, -0.00000609, 0.00000227, 0.00000566, -0.00001174, 0.00002798, -0.00001680, 0.00003938, -0.00003694, 0.00002797, -0.00004191, 0.00001963, -0.00007584, 0.00004772, -0.00030076, 0.00014785, 0.00072238, -0.00040985, 0.00005330, -0.00010521, 0.00004356, -0.00004667, 0.00000783, -0.00002447, 0.00000867, -0.00001922, 0.00001181, -0.00002318, 0.00001554, -0.00001071, -0.00001540, 0.00000403, 0.00001146, -0.00000045, 0.00000567, -0.00000929, 0.00000698, -0.00000550, -0.00000114, -0.00000404, 0.00001522, -0.00001677, 0.00000879, -0.00000472, 0.00001976, -0.00002014, 0.00000240, -0.00000916, 0.00001302, -0.00000712, 0.00000391, -0.00000681, 0.00000325, -0.00000138, 0.00000401, -0.00000376, 0.00000771, -0.00000697, -0.00000583, 0.00000796, -0.00000360, 0.00000340, -0.00000259, -0.00000175, 0.00000508, -0.00000498, 0.00000435, -0.00000409, 0.00000524, -0.00000408, -0.00000290, 0.00000067, 0.00000172, -0.00000048, 0.00000596, -0.00000582, -0.00000288, 0.00000281, 0.00000246, -0.00000255, 0.00000064, 0.00000029, -0.00000179, 0.00000400, -0.00000019, -0.00000420, -0.00000751, 0.00000769, -0.00000357, 0.00000670, 0.00000150, -0.00000493, -0.00000137, 0.00000149, 0.00000301, 0.00000329, -0.00000915, 0.00000420, -0.00000958, 0.00000713, },
		{ 0.00000920, -0.00000659, 0.00000640, 0.00000598, -0.00000835, -0.00000377, 0.00000479, -0.00000628, 0.00000830, -0.00000458, 0.00000520, -0.00000255, 0.00000780, -0.00001916, 0.00001289, -0.00000386, 0.00000634, -0.00001759, 0.00002324, -0.00000732, 0.00000267, -0.00002393, 0.00002645, -0.00002337, 0.00001590, -0.00002741, 0.00001607, -0.00000687, 0.00001773, -0.00003532, 0.00000424, -0.00000436, 0.00001087, -0.00001663, 0.00003243, -0.00002304, 0.00004589, -0.00004737, 0.00003882, -0.00005471, 0.00003435, -0.00009467, 0.00006984, -0.00035594, 0.00026029, 0.00066936, -0.00045424, 0.00009700, -0.00013203, 0.00006681, -0.00006508, 0.00002293, -0.00003667, 0.00001971, -0.00002831, 0.00002023, -0.00003206, 0.00002585, -0.00002130, -0.00000695, -0.00000043, 0.00001444, -0.00000211, 0.00000604, -0.00001042, 0.00000847, -0.00000726, -0.00000013, -0.00000349, 0.00001470, -0.00001808, 0.00001028, -0.00000431, 0.00001934, -0.00002215, 0.00000499, -0.00001103, 0.00001538, -0.00001002, 0.00000664, -0.00000962, 0.00000622, -0.00000353, 0.00000564, -0.00000521, 0.00000950, -0.00001019, -0.00000263, 0.00000665, -0.00000335, 0.00000354, -0.00000319, -0.00000119, 0.00000467, -0.00000484, 0.00000444, -0.00000412, 0.00000545, -0.00000488, -0.00000246, 0.00000132, 0.00000122, 0.00000007, 0.00000565, -0.00000721, -0.00000142, 0.00000266, 0.00000233, -0.00000281, 0.00000106, -0.00000008, -0.00000157, 0.00000447, -0.00000089, -0.00000492, -0.00000635, 0.00000784, -0.00000432, 0.00000821, -0.00000055, -0.00000393, -0.00000185, 0.00000249, 0.00000226, 0.00000345, -0.00000978, 0.00000504, -0.00001063, },
		{ -0.00001138, 0.00001098, -0.00000891, 0.00000950, 0.00000298, -0.00000773, -0.00000337, 0.00000518, -0.00000696, 0.00000971, -0.00000628, 0.00000623, -0.00000302, 0.00000767, -0.00001954, 0.00001444, -0.00000539, 0.00000623, -0.00001747, 0.00002608, -0.00001151, 0.00000441, -0.00002476, 0.00002950, -0.00002791, 0.00002046, -0.00003240, 0.00002231, -0.00001290, 0.00002442, -0.00004463, 0.00001517, -0.00001215, 0.00001703, -0.00002222, 0.00003721, -0.00002921, 0.00005204, -0.00005724, 0.00004914, -0.00006680, 0.00004815, -0.00011246, 0.00009017, -0.00041111, 0.00038559, 0.00059025, -0.00047724, 0.00013059, -0.00015351, 0.00008646, -0.00008101, 0.00003609, -0.00004721, 0.00002936, -0.00003614, 0.00002738, -0.00003972, 0.00003537, -0.00003189, 0.00000233, -0.00000610, 0.00001854, -0.00000463, 0.00000680, -0.00001155, 0.00000985, -0.00000894, 0.00000092, -0.00000283, 0.00001369, -0.00001882, 0.00001146, -0.00000349, 0.00001801, -0.00002306, 0.00000667, -0.00001211, 0.00001697, -0.00001223, 0.00000881, -0.00001199, 0.00000896, -0.00000561, 0.00000718, -0.00000649, 0.00001100, -0.00001326, 0.00000086, 0.00000482, -0.00000268, 0.00000340, -0.00000364, -0.00000066, 0.00000423, -0.00000461, 0.00000443, -0.00000406, 0.00000551, -0.00000560, -0.00000197, 0.00000197, 0.00000052, 0.00000092, 0.00000485, -0.00000817, -0.00000002, 0.00000242, 0.00000219, -0.00000301, 0.00000148, -0.00000050, -0.00000122, 0.00000484, -0.00000171, -0.00000550, -0.00000501, 0.00000758, -0.00000463, 0.00000934, -0.00000246, -0.00000291, -0.00000226, 0.00000339, 0.00000155, 0.00000355, -0.00001026, 0.00000568, },
		{ 0.00000603, -0.00001171, 0.00001226, -0.00001081, 0.00001240, -0.00000021, -0.00000680, -0.00000297, 0.00000531, -0.00000729, 0.00001078, -0.00000780, 0.00000716, -0.00000333, 0.00000722, -0.00001940, 0.00001537, -0.00000637, 0.00000544, -0.00001610, 0.00002746, -0.00001467, 0.00000536, -0.00002441, 0.00003101, -0.00003099, 0.00002365, -0.00003592, 0.00002702, -0.00001732, 0.00002948, -0.00005259, 0.00002560, -0.00002015, 0.00002335, -0.00002780, 0.00004165, -0.00003461, 0.00005701, -0.00006560, 0.00005784, -0.00007689, 0.00005944, -0.00012719, 0.00010595, -0.00046227, 0.00051964, 0.00048652, -0.00047708, 0.00015122, -0.00016718, 0.00010048, -0.00009276, 0.00004588, -0.00005482, 0.00003645, -0.00004165, 0.00003225, -0.00004511, 0.00004297, -0.00004135, 0.00001144, -0.00001221, 0.00002325, -0.00000773, 0.00000777, -0.00001254, 0.00001095, -0.00001038, 0.00000185, -0.00000200, 0.00001215, -0.00001889, 0.00001225, -0.00000227, 0.00001579, -0.00002279, 0.00000729, -0.00001220, 0.00001756, -0.00001351, 0.00001015, -0.00001365, 0.00001119, -0.00000737, 0.00000843, -0.00000740, 0.00001200, -0.00001591, 0.00000436, 0.00000266, -0.00000166, 0.00000301, -0.00000398, -0.00000014, 0.00000375, -0.00000429, 0.00000434, -0.00000389, 0.00000541, -0.00000621, -0.00000140, 0.00000258, -0.00000030, 0.00000203, 0.00000357, -0.00000863, 0.00000122, 0.00000217, 0.00000201, -0.00000314, 0.00000186, -0.00000096, -0.00000074, 0.00000509, -0.00000262, -0.00000592, -0.00000355, 0.00000693, -0.00000446, 0.00000997, -0.00000407, -0.00000201, -0.00000250, 0.00000407, 0.00000100, 0.00000347, -0.00001049, },
		{ -0.00001037, 0.00000597, -0.00001150, 0.00001291, -0.00001209, 0.00001483, -0.00000333, -0.00000570, -0.00000248, 0.00000508, -0.00000715, 0.00001134, -0.00000895, 0.00000787, -0.00000339, 0.00000639, -0.00001866, 0.00001559, -0.00000673, 0.00000391, -0.00001342, 0.00002716, -0.00001650, 0.00000525, -0.00002262, 0.00003067, -0.00003219, 0.00002499, -0.00003748, 0.00002960, -0.00001944, 0.00003203, -0.00005817, 0.00003436, -0.00002727, 0.00002890, -0.00003253, 0.00004496, -0.00003843, 0.00005993, -0.00007147, 0.00006375, -0.00008363, 0.00006656, -0.00013675, 0.00011420, -0.00050485, 0.00065714, 0.00036124, -0.00045346, 0.00015706, -0.00017117, 0.00010725, -0.00009895, 0.00005111, -0.00005845, 0.00004001, -0.00004395, 0.00003395, -0.00004731, 0.00004762, -0.00004859, 0.00001935, -0.00001792, 0.00002794, -0.00001098, 0.00000870, -0.00001318, 0.00001159, -0.00001139, 0.00000254, -0.00000092, 0.00001002, -0.00001823, 0.00001259, -0.00000068, 0.00001274, -0.00002134, 0.00000679, -0.00001121, 0.00001702, -0.00001369, 0.00001045, -0.00001433, 0.00001264, -0.00000857, 0.00000918, -0.00000777, 0.00001228, -0.00001788, 0.00000760, 0.00000039, -0.00000045, 0.00000244, -0.00000422, 0.00000038, 0.00000322, -0.00000387, 0.00000413, -0.00000361, 0.00000515, -0.00000671, -0.00000075, 0.00000309, -0.00000116, 0.00000332, 0.00000185, -0.00000855, 0.00000221, 0.00000194, 0.00000176, -0.00000317, 0.00000218, -0.00000142, -0.00000014, 0.00000520, -0.00000360, -0.00000613, -0.00000205, 0.00000596, -0.00000380, 0.00001005, -0.00000529, -0.00000133, -0.00000247, 0.00000442, 0.00000071, 0.00000311, },
		{ 0.00000238, -0.00000982, 0.00000544, -0.00001067, 0.00001278, -0.00001257, 0.00001655, -0.00000616, -0.00000459, -0.00000182, 0.00000442, -0.00000643, 0.00001126, -0.00000959, 0.00000820, -0.00000310, 0.00000513, -0.00001729, 0.00001506, -0.00000645, 0.00000167, -0.00000944, 0.00002507, -0.00001678, 0.00000390, -0.00001925, 0.00002828, -0.00003124, 0.00002418, -0.00003671, 0.00002966, -0.00001873, 0.00003138, -0.00006044, 0.00004038, -0.00003246, 0.00003272, -0.00003555, 0.00004634, -0.00003988, 0.00005997, -0.00007390, 0.00006580, -0.00008576, 0.00006794, -0.00013911, 0.00011193, -0.00053398, 0.00079187, 0.00021897, -0.00040759, 0.00014740, -0.00016442, 0.00010575, -0.00009864, 0.00005094, -0.00005734, 0.00003936, -0.00004238, 0.00003187, -0.00004563, 0.00004852, -0.00005265, 0.00002507, -0.00002236, 0.00003194, -0.00001391, 0.00000926, -0.00001325, 0.00001156, -0.00001177, 0.00000282, 0.00000047, 0.00000729, -0.00001678, 0.00001245, 0.00000123, 0.00000899, -0.00001881, 0.00000517, -0.00000912, 0.00001531, -0.00001268, 0.00000959, -0.00001388, 0.00001312, -0.00000902, 0.00000930, -0.00000747, 0.00001170, -0.00001893, 0.00001031, -0.00000175, 0.00000080, 0.00000180, -0.00000445, 0.00000095, 0.00000259, -0.00000332, 0.00000380, -0.00000322, 0.00000472, -0.00000707, -0.00000002, 0.00000344, -0.00000197, 0.00000467, -0.00000022, -0.00000794, 0.00000291, 0.00000178, 0.00000145, -0.00000309, 0.00000242, -0.00000187, 0.00000057, 0.00000512, -0.00000460, -0.00000611, -0.00000059, 0.00000475, -0.00000271, 0.00000955, -0.00000603, -0.00000095, -0.00000211, 0.00000438, 0.00000079, },
		{ 0.00000127, 0.00000124, -0.00000879, 0.00000439, -0.00000918, 0.00001182, -0.00001210, 0.00001737, -0.00000845, -0.00000364, -0.00000092, 0.00000329, -0.00000507, 0.00001043, -0.00000957, 0.00000804, -0.00000239, 0.00000339, -0.00001531, 0.00001380, -0.00000554, -0.00000116, -0.00000432, 0.00002125, -0.00001545, 0.00000125, -0.00001430, 0.00002383, -0.00002804, 0.00002107, -0.00003346, 0.00002699, -0.00001491, 0.00002709, -0.00005872, 0.00004276, -0.00003476, 0.00003396, -0.00003607, 0.00004505, -0.00003828, 0.00005642, -0.00007211, 0.00006310, -0.00008224, 0.00006230, -0.00013255, 0.00009645, -0.00054483, 0.00091698, 0.00006559, -0.00034217, 0.00012276, -0.00014679, 0.00009562, -0.00009143, 0.00004503, -0.00005117, 0.00003420, -0.00003667, 0.00002571, -0.00003970, 0.00004512, -0.00005281, 0.00002779, -0.00002475, 0.00003459, -0.00001605, 0.00000917, -0.00001254, 0.00001068, -0.00001137, 0.00000257, 0.00000224, 0.00000395, -0.00001452, 0.00001181, 0.00000335, 0.00000471, -0.00001534, 0.00000255, -0.00000602, 0.00001248, -0.00001049, 0.00000754, -0.00001223, 0.00001253, -0.00000861, 0.00000866, -0.00000642, 0.00001018, -0.00001890, 0.00001223, -0.00000354, 0.00000192, 0.00000117, -0.00000471, 0.00000161, 0.00000185, -0.00000263, 0.00000334, -0.00000272, 0.00000412, -0.00000727, 0.00000077, 0.00000359, -0.00000264, 0.00000596, -0.00000253, -0.00000684, 0.00000328, 0.00000171, 0.00000106, -0.00000291, 0.00000256, -0.00000229, 0.00000137, 0.00000484, -0.00000559, -0.00000585, 0.00000074, 0.00000340, -0.00000127, 0.00000851, -0.00000626, -0.00000090, -0.00000139, 0.00000390, },
		{ 0.00000298, 0.00000217, -0.00000033, -0.00000729, 0.00000285, -0.00000704, 0.00001002, -0.00001064, 0.00001714, -0.00001003, -0.00000298, 0.00000028, 0.00000165, -0.00000304, 0.00000878, -0.00000882, 0.00000732, -0.00000120, 0.00000119, -0.00001273, 0.00001185, -0.00000408, -0.00000442, 0.00000171, 0.00001587, -0.00001258, -0.00000261, -0.00000793, 0.00001746, -0.00002269, 0.00001574, -0.00002781, 0.00002165, -0.00000801, 0.00001904, -0.00005268, 0.00004090, -0.00003347, 0.00003192, -0.00003347, 0.00004057, -0.00003314, 0.00004879, -0.00006560, 0.00005508, -0.00007242, 0.00004879, -0.00011587, 0.00006569, -0.00053301, 0.00102545, -0.00009206, -0.00026124, 0.00008487, -0.00011905, 0.00007728, -0.00007758, 0.00003355, -0.00004009, 0.00002462, -0.00002693, 0.00001558, -0.00002955, 0.00003727, -0.00004870, 0.00002693, -0.00002447, 0.00003534, -0.00001697, 0.00000815, -0.00001088, 0.00000883, -0.00001005, 0.00000167, 0.00000440, 0.00000004, -0.00001147, 0.00001068, 0.00000559, 0.00000013, -0.00001117, -0.00000087, -0.00000207, 0.00000869, -0.00000727, 0.00000440, -0.00000940, 0.00001083, -0.00000729, 0.00000727, -0.00000464, 0.00000771, -0.00001771, 0.00001322, -0.00000480, 0.00000278, 0.00000068, -0.00000506, 0.00000240, 0.00000097, -0.00000177, 0.00000273, -0.00000211, 0.00000335, -0.00000730, 0.00000161, 0.00000352, -0.00000308, 0.00000706, -0.00000494, -0.00000533, 0.00000334, 0.00000174, 0.00000060, -0.00000262, 0.00000261, -0.00000265, 0.00000225, 0.00000434, -0.00000651, -0.00000537, 0.00000188, 0.00000202, 0.00000041, 0.00000698, -0.00000600, -0.00000119, -0.00000032, },
		{ 0.00000104, 0.00000168, 0.00000345, -0.00000229, -0.00000535, 0.00000086, -0.00000435, 0.00000744, -0.00000822, 0.00001580, -0.00001076, -0.00000272, 0.00000179, -0.00000047, -0.00000037, 0.00000633, -0.00000728, 0.00000597, 0.00000048, -0.00000146, -0.00000965, 0.00000931, -0.00000221, -0.00000788, 0.00000832, 0.00000924, -0.00000838, -0.00000748, -0.00000043, 0.00000950, -0.00001549, 0.00000847, -0.00002006, 0.00001400, 0.00000165, 0.00000749, -0.00004236, 0.00003460, -0.00002820, 0.00002623, -0.00002740, 0.00003262, -0.00002425, 0.00003693, -0.00005421, 0.00004160, -0.00005612, 0.00002715, -0.00008861, 0.00001841, -0.00049492, 0.00111050, -0.00024657, -0.00016987, 0.00003651, -0.00008291, 0.00005189, -0.00005794, 0.00001722, -0.00002471, 0.00001120, -0.00001367, 0.00000199, -0.00001561, 0.00002523, -0.00004032, 0.00002226, -0.00002116, 0.00003382, -0.00001638, 0.00000602, -0.00000818, 0.00000595, -0.00000776, 0.00000007, 0.00000695, -0.00000436, -0.00000769, 0.00000910, 0.00000782, -0.00000451, -0.00000659, -0.00000483, 0.00000247, 0.00000418, -0.00000322, 0.00000035, -0.00000555, 0.00000811, -0.00000512, 0.00000516, -0.00000221, 0.00000439, -0.00001537, 0.00001316, -0.00000540, 0.00000324, 0.00000038, -0.00000555, 0.00000334, -0.00000006, -0.00000075, 0.00000196, -0.00000138, 0.00000242, -0.00000713, 0.00000247, 0.00000321, -0.00000323, 0.00000787, -0.00000729, -0.00000352, 0.00000312, 0.00000187, 0.00000009, -0.00000225, 0.00000256, -0.00000294, 0.00000319, 0.00000360, -0.00000731, -0.00000468, 0.00000277, 0.00000073, 0.00000220, 0.00000509, -0.00000531, -0.00000179, },
		{ -0.00000261, 0.00000262, 0.00000008, 0.00000503, -0.00000454, -0.00000307, -0.00000146, -0.00000124, 0.00000425, -0.00000493, 0.00001339, -0.00001059, -0.00000290, 0.00000361, -0.00000299, 0.00000287, 0.00000313, -0.00000499, 0.00000403, 0.00000261, -0.00000448, -0.00000616, 0.00000631, -0.00000008, -0.00001127, 0.00001510, 0.00000176, -0.00000317, -0.00001301, 0.00000776, 0.00000042, -0.00000691, -0.00000025, -0.00001074, 0.00000460, 0.00001343, -0.00000697, -0.00002822, 0.00002405, -0.00001898, 0.00001683, -0.00001784, 0.00002124, -0.00001176, 0.00002107, -0.00003823, 0.00002298, -0.00003378, -0.00000218, -0.00005118, -0.00004547, -0.00042819, 0.00116613, -0.00039046, -0.00007386, -0.00001867, -0.00004080, 0.00002125, -0.00003393, -0.00000276, -0.00000611, -0.00000511, 0.00000220, -0.00001419, 0.00000128, 0.00000969, -0.00002807, 0.00001390, -0.00001474, 0.00002985, -0.00001412, 0.00000271, -0.00000445, 0.00000207, -0.00000451, -0.00000220, 0.00000983, -0.00000912, -0.00000329, 0.00000712, 0.00000991, -0.00000897, -0.00000190, -0.00000902, 0.00000728, -0.00000072, 0.00000134, -0.00000432, -0.00000092, 0.00000455, -0.00000224, 0.00000248, 0.00000072, 0.00000038, -0.00001199, 0.00001207, -0.00000526, 0.00000325, 0.00000033, -0.00000619, 0.00000443, -0.00000125, 0.00000043, 0.00000104, -0.00000056, 0.00000134, -0.00000674, 0.00000331, 0.00000267, -0.00000306, 0.00000827, -0.00000944, -0.00000152, 0.00000266, 0.00000206, -0.00000044, -0.00000182, 0.00000243, -0.00000315, 0.00000416, 0.00000262, -0.00000793, -0.00000381, 0.00000337, -0.00000038, 0.00000395, 0.00000296, -0.00000428, },
		{ -0.00000302, -0.00000355, 0.00000428, -0.00000169, 0.00000678, -0.00000696, -0.00000059, -0.00000395, 0.00000211, 0.00000064, -0.00000099, 0.00001003, -0.00000954, -0.00000352, 0.00000565, -0.00000578, 0.00000651, -0.00000067, -0.00000205, 0.00000157, 0.00000512, -0.00000777, -0.00000242, 0.00000301, 0.00000213, -0.00001431, 0.00002161, -0.00000607, 0.00000260, -0.00001877, 0.00001611, -0.00000916, 0.00000243, -0.00000978, -0.00000055, -0.00000573, 0.00002643, -0.00002338, -0.00001110, 0.00000988, -0.00000621, 0.00000408, -0.00000514, 0.00000686, 0.00000382, 0.00000182, -0.00001839, 0.00000009, -0.00000640, -0.00003798, -0.00000496, -0.00012477, -0.00033191, 0.00118751, -0.00051663, 0.00002071, -0.00007649, 0.00000428, -0.00001234, -0.00000741, -0.00002479, 0.00001431, -0.00002301, 0.00001943, -0.00003173, 0.00001995, -0.00000832, -0.00001276, 0.00000235, -0.00000546, 0.00002354, -0.00001024, -0.00000168, 0.00000017, -0.00000266, -0.00000044, -0.00000507, 0.00001291, -0.00001407, 0.00000158, 0.00000484, 0.00001174, -0.00001298, 0.00000258, -0.00001310, 0.00001202, -0.00000566, 0.00000605, -0.00000927, 0.00000419, 0.00000042, 0.00000116, -0.00000058, 0.00000391, -0.00000406, -0.00000777, 0.00001004, -0.00000441, 0.00000278, 0.00000054, -0.00000697, 0.00000565, -0.00000256, 0.00000175, -0.00000000, 0.00000035, 0.00000014, -0.00000613, 0.00000407, 0.00000195, -0.00000254, 0.00000819, -0.00001124, 0.00000053, 0.00000203, 0.00000229, -0.00000097, -0.00000136, 0.00000224, -0.00000328, 0.00000515, 0.00000140, -0.00000832, -0.00000283, 0.00000366, -0.00000120, 0.00000553, 0.00000075, },
		{ -0.00000138, -0.00000168, -0.00000449, 0.00000590, -0.00000347, 0.00000855, -0.00000940, 0.00000192, -0.00000642, 0.00000545, -0.00000312, 0.00000334, 0.00000593, -0.00000774, -0.00000449, 0.00000780, -0.00000868, 0.00001035, -0.00000489, 0.00000139, -0.00000130, 0.00000788, -0.00001118, 0.00000141, -0.00000039, 0.00000423, -0.00001672, 0.00002740, -0.00001373, 0.00000846, -0.00002427, 0.00002406, -0.00001858, 0.00001181, -0.00001937, 0.00000968, -0.00001605, 0.00003956, -0.00004055, 0.00000780, -0.00000690, 0.00000926, -0.00001129, 0.00000999, -0.00000977, 0.00002162, -0.00001979, 0.00000415, -0.00002572, 0.00002439, -0.00007832, 0.00004767, -0.00021694, -0.00020694, 0.00117142, -0.00061887, 0.00010790, -0.00013255, 0.00004901, -0.00004627, 0.00001948, -0.00004702, 0.00003488, -0.00004098, 0.00003661, -0.00004923, 0.00003898, -0.00002746, 0.00000447, -0.00001156, 0.00000614, 0.00001521, -0.00000496, -0.00000695, 0.00000546, -0.00000798, 0.00000425, -0.00000836, 0.00001602, -0.00001899, 0.00000674, 0.00000237, 0.00001317, -0.00001633, 0.00000654, -0.00001675, 0.00001634, -0.00001027, 0.00001053, -0.00001410, 0.00000940, -0.00000398, 0.00000479, -0.00000378, 0.00000710, -0.00000861, -0.00000301, 0.00000725, -0.00000293, 0.00000187, 0.00000098, -0.00000785, 0.00000696, -0.00000394, 0.00000317, -0.00000115, 0.00000130, -0.00000117, -0.00000529, 0.00000472, 0.00000108, -0.00000171, 0.00000759, -0.00001257, 0.00000250, 0.00000130, 0.00000253, -0.00000145, -0.00000090, 0.00000201, -0.00000332, 0.00000611, -0.00000005, -0.00000843, -0.00000180, 0.00000366, -0.00000169, 0.00000681, },
		{ 0.00000767, -0.00000329, -0.00000039, -0.00000531, 0.00000733, -0.00000511, 0.00001019, -0.00001169, 0.00000430, -0.00000867, 0.00000854, -0.00000672, 0.00000774, 0.00000139, -0.00000536, -0.00000570, 0.00000990, -0.00001148, 0.00001416, -0.00000927, 0.00000511, -0.00000438, 0.00001073, -0.00001454, 0.00000513, -0.00000371, 0.00000604, -0.00001824, 0.00003206, -0.00002070, 0.00001391, -0.00002904, 0.00003100, -0.00002715, 0.00002051, -0.00002824, 0.00001909, -0.00002537, 0.00005162, -0.00005711, 0.00002706, -0.00002495, 0.00002628, -0.00002823, 0.00002652, -0.00002757, 0.00004043, -0.00004244, 0.00002787, -0.00005268, 0.00005648, -0.00012060, 0.00010346, -0.00031808, -0.00005596, 0.00111655, -0.00069229, 0.00018238, -0.00018261, 0.00009008, -0.00007785, 0.00004451, -0.00006752, 0.00005386, -0.00005743, 0.00005220, -0.00006515, 0.00005682, -0.00004619, 0.00002225, -0.00002672, 0.00001924, 0.00000544, 0.00000131, -0.00001273, 0.00001104, -0.00001357, 0.00000926, -0.00001185, 0.00001896, -0.00002363, 0.00001193, -0.00000016, 0.00001412, -0.00001882, 0.00000970, -0.00001965, 0.00001990, -0.00001418, 0.00001440, -0.00001842, 0.00001433, -0.00000831, 0.00000837, -0.00000683, 0.00000999, -0.00001293, 0.00000196, 0.00000395, -0.00000097, 0.00000063, 0.00000157, -0.00000875, 0.00000827, -0.00000533, 0.00000462, -0.00000234, 0.00000225, -0.00000252, -0.00000424, 0.00000520, 0.00000015, -0.00000061, 0.00000647, -0.00001336, 0.00000427, 0.00000056, 0.00000273, -0.00000187, -0.00000048, 0.00000175, -0.00000326, 0.00000701, -0.00000170, -0.00000824, -0.00000079, 0.00000340, -0.00000179, },
		{ -0.00000153, 0.00000806, -0.00000484, 0.00000071, -0.00000588, 0.00000844, -0.00000645, 0.00001151, -0.00001365, 0.00000635, -0.00001050, 0.00001113, -0.00000988, 0.00001185, -0.00000328, -0.00000265, -0.00000698, 0.00001177, -0.00001394, 0.00001764, -0.00001353, 0.00000885, -0.00000746, 0.00001347, -0.00001768, 0.00000855, -0.00000677, 0.00000739, -0.00001869, 0.00003524, -0.00002650, 0.00001849, -0.00003261, 0.00003640, -0.00003423, 0.00002782, -0.00003564, 0.00002684, -0.00003271, 0.00006145, -0.00007167, 0.00004512, -0.00004275, 0.00004346, -0.00004544, 0.00004319, -0.00004523, 0.00005885, -0.00006453, 0.00005097, -0.00007873, 0.00008735, -0.00016174, 0.00015840, -0.00042307, 0.00011652, 0.00102366, -0.00073367, 0.00023973, -0.00022291, 0.00012443, -0.00010456, 0.00006556, -0.00008443, 0.00006959, -0.00007083, 0.00006475, -0.00007804, 0.00007194, -0.00006296, 0.00003908, -0.00004185, 0.00003285, -0.00000504, 0.00000801, -0.00001857, 0.00001649, -0.00001899, 0.00001422, -0.00001526, 0.00002146, -0.00002769, 0.00001692, -0.00000260, 0.00001448, -0.00002030, 0.00001185, -0.00002157, 0.00002245, -0.00001710, 0.00001732, -0.00002187, 0.00001860, -0.00001220, 0.00001159, -0.00000949, 0.00001230, -0.00001666, 0.00000676, 0.00000043, 0.00000126, -0.00000082, 0.00000222, -0.00000959, 0.00000949, -0.00000664, 0.00000602, -0.00000352, 0.00000315, -0.00000385, -0.00000301, 0.00000548, -0.00000079, 0.00000069, 0.00000486, -0.00001356, 0.00000576, -0.00000015, 0.00000287, -0.00000219, -0.00000010, 0.00000148, -0.00000311, 0.00000781, -0.00000349, -0.00000771, 0.00000014, 0.00000292, },
		{ 0.00000230, -0.00000092, 0.00000792, -0.00000594, 0.00000152, -0.00000611, 0.00000910, -0.00000736, 0.00001239, -0.00001513, 0.00000793, -0.00001175, 0.00001301, -0.00001229, 0.00001535, -0.00000772, 0.00000013, -0.00000812, 0.00001320, -0.00001582, 0.00002053, -0.00001735, 0.00001234, -0.00001030, 0.00001589, -0.00002038, 0.00001148, -0.00000937, 0.00000818, -0.00001794, 0.00003667, -0.00003074, 0.00002180, -0.00003462, 0.00003983, -0.00003928, 0.00003315, -0.00004093, 0.00003224, -0.00003724, 0.00006800, -0.00008291, 0.00006047, -0.00005873, 0.00005928, -0.00006147, 0.00005860, -0.00006132, 0.00007535, -0.00008438, 0.00007156, -0.00010166, 0.00011436, -0.00019840, 0.00020803, -0.00052588, 0.00030433, 0.00089564, -0.00074164, 0.00027685, -0.00025049, 0.00014955, -0.00012427, 0.00008081, -0.00009618, 0.00008064, -0.00007986, 0.00007301, -0.00008661, 0.00008296, -0.00007627, 0.00005347, -0.00005561, 0.00004585, -0.00001537, 0.00001451, -0.00002393, 0.00002132, -0.00002380, 0.00001873, -0.00001827, 0.00002327, -0.00003091, 0.00002144, -0.00000483, 0.00001423, -0.00002071, 0.00001286, -0.00002233, 0.00002377, -0.00001880, 0.00001904, -0.00002413, 0.00002188, -0.00001536, 0.00001420, -0.00001151, 0.00001377, -0.00001949, 0.00001104, -0.00000298, 0.00000352, -0.00000230, 0.00000279, -0.00001025, 0.00001052, -0.00000776, 0.00000729, -0.00000461, 0.00000393, -0.00000511, -0.00000163, 0.00000554, -0.00000164, 0.00000208, 0.00000284, -0.00001314, 0.00000689, -0.00000076, 0.00000294, -0.00000242, 0.00000021, 0.00000121, -0.00000284, 0.00000846, -0.00000538, -0.00000685, 0.00000091, },
		{ 0.00000148, 0.00000162, -0.00000002, 0.00000728, -0.00000654, 0.00000197, -0.00000593, 0.00000925, -0.00000773, 0.00001270, -0.00001602, 0.00000891, -0.00001227, 0.00001402, -0.00001374, 0.00001790, -0.00001158, 0.00000269, -0.00000892, 0.00001402, -0.00001692, 0.00002255, -0.00002045, 0.00001528, -0.00001264, 0.00001778, -0.00002247, 0.00001375, -0.00001139, 0.00000834, -0.00001596, 0.00003621, -0.00003315, 0.00002356, -0.00003481, 0.00004098, -0.00004191, 0.00003605, -0.00004365, 0.00003477, -0.00003835, 0.00007044, -0.00008972, 0.00007173, -0.00007138, 0.00007226, -0.00007490, 0.00007136, -0.00007444, 0.00008843, -0.00010034, 0.00008777, -0.00011934, 0.00013489, -0.00022721, 0.00024770, -0.00061983, 0.00049988, 0.00073738, -0.00071682, 0.00029209, -0.00026338, 0.00016364, -0.00013541, 0.00008894, -0.00010157, 0.00008593, -0.00008354, 0.00007605, -0.00008990, 0.00008880, -0.00008490, 0.00006410, -0.00006672, 0.00005714, -0.00002467, 0.00002013, -0.00002827, 0.00002505, -0.00002754, 0.00002239, -0.00002057, 0.00002414, -0.00003301, 0.00002522, -0.00000669, 0.00001332, -0.00002001, 0.00001265, -0.00002185, 0.00002377, -0.00001913, 0.00001936, -0.00002500, 0.00002391, -0.00001754, 0.00001596, -0.00001270, 0.00001424, -0.00002116, 0.00001446, -0.00000596, 0.00000557, -0.00000365, 0.00000317, -0.00001061, 0.00001126, -0.00000861, 0.00000832, -0.00000555, 0.00000453, -0.00000621, -0.00000017, 0.00000537, -0.00000235, 0.00000346, 0.00000051, -0.00001215, 0.00000762, -0.00000124, 0.00000292, -0.00000254, 0.00000045, 0.00000095, -0.00000246, 0.00000890, -0.00000728, -0.00000568, },
		{ -0.00000425, 0.00000180, 0.00000096, 0.00000108, 0.00000617, -0.00000660, 0.00000201, -0.00000531, 0.00000883, -0.00000750, 0.00001239, -0.00001623, 0.00000921, -0.00001201, 0.00001404, -0.00001406, 0.00001927, -0.00001456, 0.00000477, -0.00000921, 0.00001406, -0.00001704, 0.00002349, -0.00002256, 0.00001742, -0.00001426, 0.00001895, -0.00002379, 0.00001520, -0.00001271, 0.00000784, -0.00001278, 0.00003384, -0.00003359, 0.00002362, -0.00003309, 0.00003973, -0.00004193, 0.00003627, -0.00004353, 0.00003416, -0.00003569, 0.00006827, -0.00009134, 0.00007782, -0.00007943, 0.00008109, -0.00008444, 0.00008020, -0.00008333, 0.00009675, -0.00011095, 0.00009801, -0.00012988, 0.00014661, -0.00024513, 0.00027299, -0.00069809, 0.00069465, 0.00055553, -0.00066175, 0.00028538, -0.00026080, 0.00016581, -0.00013713, 0.00008919, -0.00009996, 0.00008490, -0.00008135, 0.00007336, -0.00008736, 0.00008876, -0.00008795, 0.00006991, -0.00007409, 0.00006572, -0.00003212, 0.00002424, -0.00003109, 0.00002723, -0.00002979, 0.00002483, -0.00002189, 0.00002386, -0.00003376, 0.00002804, -0.00000808, 0.00001179, -0.00001827, 0.00001126, -0.00002013, 0.00002243, -0.00001808, 0.00001824, -0.00002436, 0.00002455, -0.00001856, 0.00001674, -0.00001297, 0.00001358, -0.00002148, 0.00001676, -0.00000825, 0.00000718, -0.00000468, 0.00000322, -0.00001058, 0.00001160, -0.00000909, 0.00000905, -0.00000625, 0.00000489, -0.00000709, 0.00000131, 0.00000498, -0.00000285, 0.00000471, -0.00000201, -0.00001064, 0.00000794, -0.00000157, 0.00000281, -0.00000257, 0.00000062, 0.00000070, -0.00000196, 0.00000908, -0.00000911, },
		{ -0.00001077, -0.00000261, 0.00000185, 0.00000039, 0.00000228, 0.00000466, -0.00000617, 0.00000165, -0.00000426, 0.00000785, -0.00000667, 0.00001142, -0.00001575, 0.00000883, -0.00001095, 0.00001306, -0.00001318, 0.00001929, -0.00001643, 0.00000615, -0.00000883, 0.00001323, -0.00001608, 0.00002317, -0.00002348, 0.00001854, -0.00001497, 0.00001923, -0.00002420, 0.00001574, -0.00001326, 0.00000672, -0.00000855, 0.00002967, -0.00003205, 0.00002196, -0.00002949, 0.00003613, -0.00003934, 0.00003381, -0.00004057, 0.00003042, -0.00002927, 0.00006136, -0.00008740, 0.00007806, -0.00008196, 0.00008476, -0.00008906, 0.00008413, -0.00008699, 0.00009931, -0.00011512, 0.00010102, -0.00013185, 0.00014770, -0.00024969, 0.00028008, -0.00075413, 0.00087960, 0.00035813, -0.00058071, 0.00025817, -0.00024316, 0.00015610, -0.00012935, 0.00008148, -0.00009129, 0.00007747, -0.00007325, 0.00006494, -0.00007892, 0.00008264, -0.00008499, 0.00007022, -0.00007691, 0.00007080, -0.00003704, 0.00002633, -0.00003198, 0.00002754, -0.00003024, 0.00002574, -0.00002199, 0.00002228, -0.00003300, 0.00002971, -0.00000890, 0.00000969, -0.00001560, 0.00000878, -0.00001727, 0.00001986, -0.00001571, 0.00001572, -0.00002221, 0.00002373, -0.00001834, 0.00001648, -0.00001230, 0.00001180, -0.00002039, 0.00001776, -0.00000964, 0.00000817, -0.00000527, 0.00000285, -0.00001007, 0.00001147, -0.00000912, 0.00000939, -0.00000667, 0.00000495, -0.00000766, 0.00000272, 0.00000439, -0.00000311, 0.00000575, -0.00000456, -0.00000872, 0.00000788, -0.00000174, 0.00000262, -0.00000253, 0.00000075, 0.00000045, -0.00000132, 0.00000896, },
		{ 0.00000849, -0.00001217, -0.00000086, 0.00000165, -0.00000002, 0.00000348, 0.00000287, -0.00000530, 0.00000092, -0.00000282, 0.00000638, -0.00000528, 0.00000985, -0.00001460, 0.00000780, -0.00000917, 0.00001115, -0.00001113, 0.00001792, -0.00001703, 0.00000668, -0.00000772, 0.00001148, -0.00001399, 0.00002154, -0.00002308, 0.00001851, -0.00001463, 0.00001852, -0.00002364, 0.00001531, -0.00001302, 0.00000506, -0.00000350, 0.00002393, -0.00002870, 0.00001873, -0.00002422, 0.00003042, -0.00003435, 0.00002887, -0.00003501, 0.00002385, -0.00001941, 0.00005000, -0.00007798, 0.00007226, -0.00007853, 0.00008268, -0.00008813, 0.00008255, -0.00008484, 0.00009551, -0.00011221, 0.00009611, -0.00012444, 0.00013709, -0.00023925, 0.00026607, -0.00078219, 0.00104579, 0.00015407, -0.00047938, 0.00021329, -0.00021201, 0.00013551, -0.00011278, 0.00006642, -0.00007608, 0.00006414, -0.00005972, 0.00005124, -0.00006503, 0.00007073, -0.00007607, 0.00006481, -0.00007475, 0.00007187, -0.00003896, 0.00002603, -0.00003071, 0.00002575, -0.00002869, 0.00002493, -0.00002075, 0.00001935, -0.00003065, 0.00003009, -0.00000909, 0.00000711, -0.00001215, 0.00000539, -0.00001345, 0.00001625, -0.00001221, 0.00001197, -0.00001869, 0.00002152, -0.00001692, 0.00001520, -0.00001074, 0.00000900, -0.00001793, 0.00001740, -0.00000999, 0.00000840, -0.00000531, 0.00000200, -0.00000903, 0.00001082, -0.00000866, 0.00000929, -0.00000676, 0.00000469, -0.00000787, 0.00000400, 0.00000366, -0.00000311, 0.00000647, -0.00000703, -0.00000649, 0.00000746, -0.00000177, 0.00000238, -0.00000244, 0.00000085, 0.00000019, -0.00000055, },
		{ 0.00000035, 0.00000765, -0.00001321, 0.00000091, 0.00000121, -0.00000022, 0.00000457, 0.00000091, -0.00000408, -0.00000009, -0.00000109, 0.00000449, -0.00000341, 0.00000774, -0.00001289, 0.00000624, -0.00000679, 0.00000846, -0.00000805, 0.00001521, -0.00001633, 0.00000630, -0.00000585, 0.00000886, -0.00001083, 0.00001860, -0.00002134, 0.00001725, -0.00001318, 0.00001677, -0.00002208, 0.00001391, -0.00001201, 0.00000299, 0.00000211, 0.00001696, -0.00002378, 0.00001417, -0.00001763, 0.00002300, -0.00002736, 0.00002186, -0.00002730, 0.00001501, -0.00000678, 0.00003485, -0.00006363, 0.00006071, -0.00006915, 0.00007472, -0.00008146, 0.00007526, -0.00007675, 0.00008525, -0.00010213, 0.00008320, -0.00010755, 0.00011456, -0.00021326, 0.00022936, -0.00077776, 0.00118489, -0.00004738, -0.00036448, 0.00015470, -0.00016988, 0.00010584, -0.00008883, 0.00004518, -0.00005540, 0.00004590, -0.00004171, 0.00003322, -0.00004659, 0.00005382, -0.00006177, 0.00005393, -0.00006759, 0.00006876, -0.00003768, 0.00002321, -0.00002718, 0.00002184, -0.00002508, 0.00002234, -0.00001813, 0.00001508, -0.00002672, 0.00002914, -0.00000865, 0.00000418, -0.00000815, 0.00000131, -0.00000892, 0.00001184, -0.00000784, 0.00000725, -0.00001400, 0.00001810, -0.00001441, 0.00001302, -0.00000845, 0.00000536, -0.00001425, 0.00001573, -0.00000928, 0.00000783, -0.00000475, 0.00000065, -0.00000748, 0.00000966, -0.00000770, 0.00000874, -0.00000649, 0.00000408, -0.00000768, 0.00000505, 0.00000284, -0.00000286, 0.00000681, -0.00000926, -0.00000409, 0.00000676, -0.00000167, 0.00000211, -0.00000233, 0.00000094, -0.00000007, },
		{ -0.00000035, 0.00000137, 0.00000643, -0.00001382, 0.00000260, 0.00000058, -0.00000019, 0.00000546, -0.00000108, -0.00000263, -0.00000131, 0.00000082, 0.00000231, -0.00000119, 0.00000524, -0.00001074, 0.00000428, -0.00000401, 0.00000520, -0.00000415, 0.00001132, -0.00001440, 0.00000503, -0.00000329, 0.00000547, -0.00000673, 0.00001448, -0.00001832, 0.00001482, -0.00001065, 0.00001401, -0.00001957, 0.00001160, -0.00001030, 0.00000068, 0.00000794, 0.00000915, -0.00001768, 0.00000865, -0.00001016, 0.00001439, -0.00001892, 0.00001334, -0.00001811, 0.00000465, 0.00000772, 0.00001691, -0.00004533, 0.00004419, -0.00005439, 0.00006126, -0.00006937, 0.00006260, -0.00006310, 0.00006898, -0.00008541, 0.00006291, -0.00008190, 0.00008084, -0.00017234, 0.00016981, -0.00073795, 0.00128980, -0.00023722, -0.00024321, 0.00008715, -0.00012007, 0.00006958, -0.00005948, 0.00001950, -0.00003079, 0.00002415, -0.00002054, 0.00001220, -0.00002491, 0.00003314, -0.00004310, 0.00003831, -0.00005585, 0.00006164, -0.00003329, 0.00001796, -0.00002154, 0.00001594, -0.00001954, 0.00001804, -0.00001423, 0.00000964, -0.00002132, 0.00002687, -0.00000760, 0.00000103, -0.00000383, -0.00000318, -0.00000396, 0.00000696, -0.00000294, 0.00000188, -0.00000847, 0.00001370, -0.00001103, 0.00001014, -0.00000566, 0.00000115, -0.00000961, 0.00001290, -0.00000757, 0.00000647, -0.00000363, -0.00000116, -0.00000544, 0.00000803, -0.00000627, 0.00000775, -0.00000589, 0.00000315, -0.00000707, 0.00000582, 0.00000201, -0.00000238, 0.00000674, -0.00001114, -0.00000165, 0.00000583, -0.00000148, 0.00000184, -0.00000223, 0.00000105, },
		{ 0.00000119, -0.00000065, 0.00000247, 0.00000484, -0.00001394, 0.00000411, -0.00000016, 0.00000008, 0.00000607, -0.00000297, -0.00000106, -0.00000260, 0.00000279, -0.00000000, 0.00000123, 0.00000250, -0.00000831, 0.00000212, -0.00000105, 0.00000167, 0.00000026, 0.00000653, -0.00001139, 0.00000298, -0.00000018, 0.00000151, -0.00000191, 0.00000939, -0.00001420, 0.00001133, -0.00000715, 0.00001035, -0.00001623, 0.00000848, -0.00000799, -0.00000171, 0.00001365, 0.00000096, -0.00001083, 0.00000261, -0.00000234, 0.00000521, -0.00000969, 0.00000402, -0.00000818, -0.00000630, 0.00002299, -0.00000254, -0.00002439, 0.00002394, -0.00003526, 0.00004317, -0.00005263, 0.00004534, -0.00004475, 0.00004766, -0.00006312, 0.00003646, -0.00004896, 0.00003764, -0.00011835, 0.00008893, -0.00066182, 0.00135503, -0.00040723, -0.00012285, 0.00001577, -0.00006631, 0.00002964, -0.00002712, -0.00000859, -0.00000409, 0.00000057, 0.00000217, -0.00001025, -0.00000156, 0.00001023, -0.00002146, 0.00001906, -0.00004031, 0.00005106, -0.00002614, 0.00001059, -0.00001410, 0.00000838, -0.00001234, 0.00001225, -0.00000925, 0.00000327, -0.00001469, 0.00002340, -0.00000601, -0.00000216, 0.00000054, -0.00000778, 0.00000110, 0.00000194, 0.00000213, -0.00000374, -0.00000247, 0.00000866, -0.00000705, 0.00000679, -0.00000263, -0.00000328, -0.00000433, 0.00000915, -0.00000501, 0.00000442, -0.00000202, -0.00000335, -0.00000300, 0.00000599, -0.00000443, 0.00000637, -0.00000499, 0.00000193, -0.00000605, 0.00000626, 0.00000123, -0.00000172, 0.00000625, -0.00001258, 0.00000073, 0.00000477, -0.00000123, 0.00000158, -0.00000216, },
		{ -0.00000213, 0.00000136, -0.00000094, 0.00000362, 0.00000292, -0.00001353, 0.00000535, -0.00000096, 0.00000057, 0.00000634, -0.00000466, 0.00000050, -0.00000387, 0.00000468, -0.00000230, 0.00000367, -0.00000030, -0.00000579, -0.00000004, 0.00000184, -0.00000184, 0.00000482, 0.00000118, -0.00000758, 0.00000035, 0.00000326, -0.00000276, 0.00000334, 0.00000362, -0.00000923, 0.00000700, -0.00000286, 0.00000597, -0.00001222, 0.00000473, -0.00000523, -0.00000400, 0.00001890, -0.00000713, -0.00000370, -0.00000351, 0.00000531, -0.00000391, -0.00000037, -0.00000536, 0.00000165, -0.00001687, 0.00003782, -0.00002210, -0.00000238, 0.00000147, -0.00001317, 0.00002171, -0.00003242, 0.00002470, -0.00002299, 0.00002271, -0.00003686, 0.00000570, -0.00001088, -0.00001250, -0.00005432, -0.00001017, -0.00055050, 0.00137717, -0.00055050, -0.00001017, -0.00005432, -0.00001250, -0.00001088, 0.00000570, -0.00003686, 0.00002271, -0.00002299, 0.00002470, -0.00003242, 0.00002171, -0.00001317, 0.00000147, -0.00000238, -0.00002210, 0.00003782, -0.00001687, 0.00000165, -0.00000536, -0.00000037, -0.00000391, 0.00000531, -0.00000351, -0.00000370, -0.00000713, 0.00001890, -0.00000400, -0.00000523, 0.00000473, -0.00001222, 0.00000597, -0.00000286, 0.00000700, -0.00000923, 0.00000362, 0.00000334, -0.00000276, 0.00000326, 0.00000035, -0.00000758, 0.00000118, 0.00000482, -0.00000184, 0.00000184, -0.00000004, -0.00000579, -0.00000030, 0.00000367, -0.00000230, 0.00000468, -0.00000387, 0.00000050, -0.00000466, 0.00000634, 0.00000057, -0.00000096, 0.00000535, -0.00001353, 0.00000292, 0.00000362, -0.00000094, 0.00000136, },
		{ 0.00000119, -0.00000216, 0.00000158, -0.00000123, 0.00000477, 0.00000073, -0.00001258, 0.00000625, -0.00000172, 0.00000123, 0.00000626, -0.00000605, 0.00000193, -0.00000499, 0.00000637, -0.00000443, 0.00000599, -0.00000300, -0.00000335, -0.00000202, 0.00000442, -0.00000501, 0.00000915, -0.00000433, -0.00000328, -0.00000263, 0.00000679, -0.00000705, 0.00000866, -0.00000247, -0.00000374, 0.00000213, 0.00000194, 0.00000110, -0.00000778, 0.00000054, -0.00000216, -0.00000601, 0.00002340, -0.00001469, 0.00000327, -0.00000925, 0.00001225, -0.00001234, 0.00000838, -0.00001410, 0.00001059, -0.00002614, 0.00005106, -0.00004031, 0.00001906, -0.00002146, 0.00001023, -0.00000156, -0.00001025, 0.00000217, 0.00000057, -0.00000409, -0.00000859, -0.00002712, 0.00002964, -0.00006631, 0.00001577, -0.00012285, -0.00040723, 0.00135503, -0.00066182, 0.00008893, -0.00011835, 0.00003764, -0.00004896, 0.00003646, -0.00006312, 0.00004766, -0.00004475, 0.00004534, -0.00005263, 0.00004317, -0.00003526, 0.00002394, -0.00002439, -0.00000254, 0.00002299, -0.00000630, -0.00000818, 0.00000402, -0.00000969, 0.00000521, -0.00000234, 0.00000261, -0.00001083, 0.00000096, 0.00001365, -0.00000171, -0.00000799, 0.00000848, -0.00001623, 0.00001035, -0.00000715, 0.00001133, -0.00001420, 0.00000939, -0.00000191, 0.00000151, -0.00000018, 0.00000298, -0.00001139, 0.00000653, 0.00000026, 0.00000167, -0.00000105, 0.00000212, -0.00000831, 0.00000250, 0.00000123, -0.00000000, 0.00000279, -0.00000260, -0.00000106, -0.00000297, 0.00000607, 0.00000008, -0.00000016, 0.00000411, -0.00001394, 0.00000484, 0.00000247, -0.00000065, },
		{ -0.00000035, 0.00000105, -0.00000223, 0.00000184, -0.00000148, 0.00000583, -0.00000165, -0.00001114, 0.00000674, -0.00000238, 0.00000201, 0.00000582, -0.00000707, 0.00000315, -0.00000589, 0.00000775, -0.00000627, 0.00000803, -0.00000544, -0.00000116, -0.00000363, 0.00000647, -0.00000757, 0.00001290, -0.00000961, 0.00000115, -0.00000566, 0.00001014, -0.00001103, 0.00001370, -0.00000847, 0.00000188, -0.00000294, 0.00000696, -0.00000396, -0.00000318, -0.00000383, 0.00000103, -0.00000760, 0.00002687, -0.00002132, 0.00000964, -0.00001423, 0.00001804, -0.00001954, 0.00001594, -0.00002154, 0.00001796, -0.00003329, 0.00006164, -0.00005585, 0.00003831, -0.00004310, 0.00003314, -0.00002491, 0.00001220, -0.00002054, 0.00002415, -0.00003079, 0.00001950, -0.00005948, 0.00006958, -0.00012007, 0.00008715, -0.00024321, -0.00023722, 0.00128980, -0.00073795, 0.00016981, -0.00017234, 0.00008084, -0.00008190, 0.00006291, -0.00008541, 0.00006898, -0.00006310, 0.00006260, -0.00006937, 0.00006126, -0.00005439, 0.00004419, -0.00004533, 0.00001691, 0.00000772, 0.00000465, -0.00001811, 0.00001334, -0.00001892, 0.00001439, -0.00001016, 0.00000865, -0.00001768, 0.00000915, 0.00000794, 0.00000068, -0.00001030, 0.00001160, -0.00001957, 0.00001401, -0.00001065, 0.00001482, -0.00001832, 0.00001448, -0.00000673, 0.00000547, -0.00000329, 0.00000503, -0.00001440, 0.00001132, -0.00000415, 0.00000520, -0.00000401, 0.00000428, -0.00001074, 0.00000524, -0.00000119, 0.00000231, 0.00000082, -0.00000131, -0.00000263, -0.00000108, 0.00000546, -0.00000019, 0.00000058, 0.00000260, -0.00001382, 0.00000643, 0.00000137, },
		{ 0.00000035, -0.00000007, 0.00000094, -0.00000233, 0.00000211, -0.00000167, 0.00000676, -0.00000409, -0.00000926, 0.00000681, -0.00000286, 0.00000284, 0.00000505, -0.00000768, 0.00000408, -0.00000649, 0.00000874, -0.00000770, 0.00000966, -0.00000748, 0.00000065, -0.00000475, 0.00000783, -0.00000928, 0.00001573, -0.00001425, 0.00000536, -0.00000845, 0.00001302, -0.00001441, 0.00001810, -0.00001400, 0.00000725, -0.00000784, 0.00001184, -0.00000892, 0.00000131, -0.00000815, 0.00000418, -0.00000865, 0.00002914, -0.00002672, 0.00001508, -0.00001813, 0.00002234, -0.00002508, 0.00002184, -0.00002718, 0.00002321, -0.00003768, 0.00006876, -0.00006759, 0.00005393, -0.00006177, 0.00005382, -0.00004659, 0.00003322, -0.00004171, 0.00004590, -0.00005540, 0.00004518, -0.00008883, 0.00010584, -0.00016988, 0.00015470, -0.00036448, -0.00004738, 0.00118489, -0.00077776, 0.00022936, -0.00021326, 0.00011456, -0.00010755, 0.00008320, -0.00010213, 0.00008525, -0.00007675, 0.00007526, -0.00008146, 0.00007472, -0.00006915, 0.00006071, -0.00006363, 0.00003485, -0.00000678, 0.00001501, -0.00002730, 0.00002186, -0.00002736, 0.00002300, -0.00001763, 0.00001417, -0.00002378, 0.00001696, 0.00000211, 0.00000299, -0.00001201, 0.00001391, -0.00002208, 0.00001677, -0.00001318, 0.00001725, -0.00002134, 0.00001860, -0.00001083, 0.00000886, -0.00000585, 0.00000630, -0.00001633, 0.00001521, -0.00000805, 0.00000846, -0.00000679, 0.00000624, -0.00001289, 0.00000774, -0.00000341, 0.00000449, -0.00000109, -0.00000009, -0.00000408, 0.00000091, 0.00000457, -0.00000022, 0.00000121, 0.00000091, -0.00001321, 0.00000765, },
		{ 0.00000849, -0.00000055, 0.00000019, 0.00000085, -0.00000244, 0.00000238, -0.00000177, 0.00000746, -0.00000649, -0.00000703, 0.00000647, -0.00000311, 0.00000366, 0.00000400, -0.00000787, 0.00000469, -0.00000676, 0.00000929, -0.00000866, 0.00001082, -0.00000903, 0.00000200, -0.00000531, 0.00000840, -0.00000999, 0.00001740, -0.00001793, 0.00000900, -0.00001074, 0.00001520, -0.00001692, 0.00002152, -0.00001869, 0.00001197, -0.00001221, 0.00001625, -0.00001345, 0.00000539, -0.00001215, 0.00000711, -0.00000909, 0.00003009, -0.00003065, 0.00001935, -0.00002075, 0.00002493, -0.00002869, 0.00002575, -0.00003071, 0.00002603, -0.00003896, 0.00007187, -0.00007475, 0.00006481, -0.00007607, 0.00007073, -0.00006503, 0.00005124, -0.00005972, 0.00006414, -0.00007608, 0.00006642, -0.00011278, 0.00013551, -0.00021201, 0.00021329, -0.00047938, 0.00015407, 0.00104579, -0.00078219, 0.00026607, -0.00023925, 0.00013709, -0.00012444, 0.00009611, -0.00011221, 0.00009551, -0.00008484, 0.00008255, -0.00008813, 0.00008268, -0.00007853, 0.00007226, -0.00007798, 0.00005000, -0.00001941, 0.00002385, -0.00003501, 0.00002887, -0.00003435, 0.00003042, -0.00002422, 0.00001873, -0.00002870, 0.00002393, -0.00000350, 0.00000506, -0.00001302, 0.00001531, -0.00002364, 0.00001852, -0.00001463, 0.00001851, -0.00002308, 0.00002154, -0.00001399, 0.00001148, -0.00000772, 0.00000668, -0.00001703, 0.00001792, -0.00001113, 0.00001115, -0.00000917, 0.00000780, -0.00001460, 0.00000985, -0.00000528, 0.00000638, -0.00000282, 0.00000092, -0.00000530, 0.00000287, 0.00000348, -0.00000002, 0.00000165, -0.00000086, -0.00001217, },
		{ -0.00001077, 0.00000896, -0.00000132, 0.00000045, 0.00000075, -0.00000253, 0.00000262, -0.00000174, 0.00000788, -0.00000872, -0.00000456, 0.00000575, -0.00000311, 0.00000439, 0.00000272, -0.00000766, 0.00000495, -0.00000667, 0.00000939, -0.00000912, 0.00001147, -0.00001007, 0.00000285, -0.00000527, 0.00000817, -0.00000964, 0.00001776, -0.00002039, 0.00001180, -0.00001230, 0.00001648, -0.00001834, 0.00002373, -0.00002221, 0.00001572, -0.00001571, 0.00001986, -0.00001727, 0.00000878, -0.00001560, 0.00000969, -0.00000890, 0.00002971, -0.00003300, 0.00002228, -0.00002199, 0.00002574, -0.00003024, 0.00002754, -0.00003198, 0.00002633, -0.00003704, 0.00007080, -0.00007691, 0.00007022, -0.00008499, 0.00008264, -0.00007892, 0.00006494, -0.00007325, 0.00007747, -0.00009129, 0.00008148, -0.00012935, 0.00015610, -0.00024316, 0.00025817, -0.00058071, 0.00035813, 0.00087960, -0.00075413, 0.00028008, -0.00024969, 0.00014770, -0.00013185, 0.00010102, -0.00011512, 0.00009931, -0.00008699, 0.00008413, -0.00008906, 0.00008476, -0.00008196, 0.00007806, -0.00008740, 0.00006136, -0.00002927, 0.00003042, -0.00004057, 0.00003381, -0.00003934, 0.00003613, -0.00002949, 0.00002196, -0.00003205, 0.00002967, -0.00000855, 0.00000672, -0.00001326, 0.00001574, -0.00002420, 0.00001923, -0.00001497, 0.00001854, -0.00002348, 0.00002317, -0.00001608, 0.00001323, -0.00000883, 0.00000615, -0.00001643, 0.00001929, -0.00001318, 0.00001306, -0.00001095, 0.00000883, -0.00001575, 0.00001142, -0.00000667, 0.00000785, -0.00000426, 0.00000165, -0.00000617, 0.00000466, 0.00000228, 0.00000039, 0.00000185, -0.00000261, },
		{ -0.00000425, -0.00000911, 0.00000908, -0.00000196, 0.00000070, 0.00000062, -0.00000257, 0.00000281, -0.00000157, 0.00000794, -0.00001064, -0.00000201, 0.00000471, -0.00000285, 0.00000498, 0.00000131, -0.00000709, 0.00000489, -0.00000625, 0.00000905, -0.00000909, 0.00001160, -0.00001058, 0.00000322, -0.00000468, 0.00000718, -0.00000825, 0.00001676, -0.00002148, 0.00001358, -0.00001297, 0.00001674, -0.00001856, 0.00002455, -0.00002436, 0.00001824, -0.00001808, 0.00002243, -0.00002013, 0.00001126, -0.00001827, 0.00001179, -0.00000808, 0.00002804, -0.00003376, 0.00002386, -0.00002189, 0.00002483, -0.00002979, 0.00002723, -0.00003109, 0.00002424, -0.00003212, 0.00006572, -0.00007409, 0.00006991, -0.00008795, 0.00008876, -0.00008736, 0.00007336, -0.00008135, 0.00008490, -0.00009996, 0.00008919, -0.00013713, 0.00016581, -0.00026080, 0.00028538, -0.00066175, 0.00055553, 0.00069465, -0.00069809, 0.00027299, -0.00024513, 0.00014661, -0.00012988, 0.00009801, -0.00011095, 0.00009675, -0.00008333, 0.00008020, -0.00008444, 0.00008109, -0.00007943, 0.00007782, -0.00009134, 0.00006827, -0.00003569, 0.00003416, -0.00004353, 0.00003627, -0.00004193, 0.00003973, -0.00003309, 0.00002362, -0.00003359, 0.00003384, -0.00001278, 0.00000784, -0.00001271, 0.00001520, -0.00002379, 0.00001895, -0.00001426, 0.00001742, -0.00002256, 0.00002349, -0.00001704, 0.00001406, -0.00000921, 0.00000477, -0.00001456, 0.00001927, -0.00001406, 0.00001404, -0.00001201, 0.00000921, -0.00001623, 0.00001239, -0.00000750, 0.00000883, -0.00000531, 0.00000201, -0.00000660, 0.00000617, 0.00000108, 0.00000096, 0.00000180, },
		{ 0.00000148, -0.00000568, -0.00000728, 0.00000890, -0.00000246, 0.00000095, 0.00000045, -0.00000254, 0.00000292, -0.00000124, 0.00000762, -0.00001215, 0.00000051, 0.00000346, -0.00000235, 0.00000537, -0.00000017, -0.00000621, 0.00000453, -0.00000555, 0.00000832, -0.00000861, 0.00001126, -0.00001061, 0.00000317, -0.00000365, 0.00000557, -0.00000596, 0.00001446, -0.00002116, 0.00001424, -0.00001270, 0.00001596, -0.00001754, 0.00002391, -0.00002500, 0.00001936, -0.00001913, 0.00002377, -0.00002185, 0.00001265, -0.00002001, 0.00001332, -0.00000669, 0.00002522, -0.00003301, 0.00002414, -0.00002057, 0.00002239, -0.00002754, 0.00002505, -0.00002827, 0.00002013, -0.00002467, 0.00005714, -0.00006672, 0.00006410, -0.00008490, 0.00008880, -0.00008990, 0.00007605, -0.00008354, 0.00008593, -0.00010157, 0.00008894, -0.00013541, 0.00016364, -0.00026338, 0.00029209, -0.00071682, 0.00073738, 0.00049988, -0.00061983, 0.00024770, -0.00022721, 0.00013489, -0.00011934, 0.00008777, -0.00010034, 0.00008843, -0.00007444, 0.00007136, -0.00007490, 0.00007226, -0.00007138, 0.00007173, -0.00008972, 0.00007044, -0.00003835, 0.00003477, -0.00004365, 0.00003605, -0.00004191, 0.00004098, -0.00003481, 0.00002356, -0.00003315, 0.00003621, -0.00001596, 0.00000834, -0.00001139, 0.00001375, -0.00002247, 0.00001778, -0.00001264, 0.00001528, -0.00002045, 0.00002255, -0.00001692, 0.00001402, -0.00000892, 0.00000269, -0.00001158, 0.00001790, -0.00001374, 0.00001402, -0.00001227, 0.00000891, -0.00001602, 0.00001270, -0.00000773, 0.00000925, -0.00000593, 0.00000197, -0.00000654, 0.00000728, -0.00000002, 0.00000162, },
		{ 0.00000230, 0.00000091, -0.00000685, -0.00000538, 0.00000846, -0.00000284, 0.00000121, 0.00000021, -0.00000242, 0.00000294, -0.00000076, 0.00000689, -0.00001314, 0.00000284, 0.00000208, -0.00000164, 0.00000554, -0.00000163, -0.00000511, 0.00000393, -0.00000461, 0.00000729, -0.00000776, 0.00001052, -0.00001025, 0.00000279, -0.00000230, 0.00000352, -0.00000298, 0.00001104, -0.00001949, 0.00001377, -0.00001151, 0.00001420, -0.00001536, 0.00002188, -0.00002413, 0.00001904, -0.00001880, 0.00002377, -0.00002233, 0.00001286, -0.00002071, 0.00001423, -0.00000483, 0.00002144, -0.00003091, 0.00002327, -0.00001827, 0.00001873, -0.00002380, 0.00002132, -0.00002393, 0.00001451, -0.00001537, 0.00004585, -0.00005561, 0.00005347, -0.00007627, 0.00008296, -0.00008661, 0.00007301, -0.00007986, 0.00008064, -0.00009618, 0.00008081, -0.00012427, 0.00014955, -0.00025049, 0.00027685, -0.00074164, 0.00089564, 0.00030433, -0.00052588, 0.00020803, -0.00019840, 0.00011436, -0.00010166, 0.00007156, -0.00008438, 0.00007535, -0.00006132, 0.00005860, -0.00006147, 0.00005928, -0.00005873, 0.00006047, -0.00008291, 0.00006800, -0.00003724, 0.00003224, -0.00004093, 0.00003315, -0.00003928, 0.00003983, -0.00003462, 0.00002180, -0.00003074, 0.00003667, -0.00001794, 0.00000818, -0.00000937, 0.00001148, -0.00002038, 0.00001589, -0.00001030, 0.00001234, -0.00001735, 0.00002053, -0.00001582, 0.00001320, -0.00000812, 0.00000013, -0.00000772, 0.00001535, -0.00001229, 0.00001301, -0.00001175, 0.00000793, -0.00001513, 0.00001239, -0.00000736, 0.00000910, -0.00000611, 0.00000152, -0.00000594, 0.00000792, -0.00000092, },
		{ -0.00000153, 0.00000292, 0.00000014, -0.00000771, -0.00000349, 0.00000781, -0.00000311, 0.00000148, -0.00000010, -0.00000219, 0.00000287, -0.00000015, 0.00000576, -0.00001356, 0.00000486, 0.00000069, -0.00000079, 0.00000548, -0.00000301, -0.00000385, 0.00000315, -0.00000352, 0.00000602, -0.00000664, 0.00000949, -0.00000959, 0.00000222, -0.00000082, 0.00000126, 0.00000043, 0.00000676, -0.00001666, 0.00001230, -0.00000949, 0.00001159, -0.00001220, 0.00001860, -0.00002187, 0.00001732, -0.00001710, 0.00002245, -0.00002157, 0.00001185, -0.00002030, 0.00001448, -0.00000260, 0.00001692, -0.00002769, 0.00002146, -0.00001526, 0.00001422, -0.00001899, 0.00001649, -0.00001857, 0.00000801, -0.00000504, 0.00003285, -0.00004185, 0.00003908, -0.00006296, 0.00007194, -0.00007804, 0.00006475, -0.00007083, 0.00006959, -0.00008443, 0.00006556, -0.00010456, 0.00012443, -0.00022291, 0.00023973, -0.00073367, 0.00102366, 0.00011652, -0.00042307, 0.00015840, -0.00016174, 0.00008735, -0.00007873, 0.00005097, -0.00006453, 0.00005885, -0.00004523, 0.00004319, -0.00004544, 0.00004346, -0.00004275, 0.00004512, -0.00007167, 0.00006145, -0.00003271, 0.00002684, -0.00003564, 0.00002782, -0.00003423, 0.00003640, -0.00003261, 0.00001849, -0.00002650, 0.00003524, -0.00001869, 0.00000739, -0.00000677, 0.00000855, -0.00001768, 0.00001347, -0.00000746, 0.00000885, -0.00001353, 0.00001764, -0.00001394, 0.00001177, -0.00000698, -0.00000265, -0.00000328, 0.00001185, -0.00000988, 0.00001113, -0.00001050, 0.00000635, -0.00001365, 0.00001151, -0.00000645, 0.00000844, -0.00000588, 0.00000071, -0.00000484, 0.00000806, },
		{ 0.00000767, -0.00000179, 0.00000340, -0.00000079, -0.00000824, -0.00000170, 0.00000701, -0.00000326, 0.00000175, -0.00000048, -0.00000187, 0.00000273, 0.00000056, 0.00000427, -0.00001336, 0.00000647, -0.00000061, 0.00000015, 0.00000520, -0.00000424, -0.00000252, 0.00000225, -0.00000234, 0.00000462, -0.00000533, 0.00000827, -0.00000875, 0.00000157, 0.00000063, -0.00000097, 0.00000395, 0.00000196, -0.00001293, 0.00000999, -0.00000683, 0.00000837, -0.00000831, 0.00001433, -0.00001842, 0.00001440, -0.00001418, 0.00001990, -0.00001965, 0.00000970, -0.00001882, 0.00001412, -0.00000016, 0.00001193, -0.00002363, 0.00001896, -0.00001185, 0.00000926, -0.00001357, 0.00001104, -0.00001273, 0.00000131, 0.00000544, 0.00001924, -0.00002672, 0.00002225, -0.00004619, 0.00005682, -0.00006515, 0.00005220, -0.00005743, 0.00005386, -0.00006752, 0.00004451, -0.00007785, 0.00009008, -0.00018261, 0.00018238, -0.00069229, 0.00111655, -0.00005596, -0.00031808, 0.00010346, -0.00012060, 0.00005648, -0.00005268, 0.00002787, -0.00004244, 0.00004043, -0.00002757, 0.00002652, -0.00002823, 0.00002628, -0.00002495, 0.00002706, -0.00005711, 0.00005162, -0.00002537, 0.00001909, -0.00002824, 0.00002051, -0.00002715, 0.00003100, -0.00002904, 0.00001391, -0.00002070, 0.00003206, -0.00001824, 0.00000604, -0.00000371, 0.00000513, -0.00001454, 0.00001073, -0.00000438, 0.00000511, -0.00000927, 0.00001416, -0.00001148, 0.00000990, -0.00000570, -0.00000536, 0.00000139, 0.00000774, -0.00000672, 0.00000854, -0.00000867, 0.00000430, -0.00001169, 0.00001019, -0.00000511, 0.00000733, -0.00000531, -0.00000039, -0.00000329, },
		{ -0.00000138, 0.00000681, -0.00000169, 0.00000366, -0.00000180, -0.00000843, -0.00000005, 0.00000611, -0.00000332, 0.00000201, -0.00000090, -0.00000145, 0.00000253, 0.00000130, 0.00000250, -0.00001257, 0.00000759, -0.00000171, 0.00000108, 0.00000472, -0.00000529, -0.00000117, 0.00000130, -0.00000115, 0.00000317, -0.00000394, 0.00000696, -0.00000785, 0.00000098, 0.00000187, -0.00000293, 0.00000725, -0.00000301, -0.00000861, 0.00000710, -0.00000378, 0.00000479, -0.00000398, 0.00000940, -0.00001410, 0.00001053, -0.00001027, 0.00001634, -0.00001675, 0.00000654, -0.00001633, 0.00001317, 0.00000237, 0.00000674, -0.00001899, 0.00001602, -0.00000836, 0.00000425, -0.00000798, 0.00000546, -0.00000695, -0.00000496, 0.00001521, 0.00000614, -0.00001156, 0.00000447, -0.00002746, 0.00003898, -0.00004923, 0.00003661, -0.00004098, 0.00003488, -0.00004702, 0.00001948, -0.00004627, 0.00004901, -0.00013255, 0.00010790, -0.00061887, 0.00117142, -0.00020694, -0.00021694, 0.00004767, -0.00007832, 0.00002439, -0.00002572, 0.00000415, -0.00001979, 0.00002162, -0.00000977, 0.00000999, -0.00001129, 0.00000926, -0.00000690, 0.00000780, -0.00004055, 0.00003956, -0.00001605, 0.00000968, -0.00001937, 0.00001181, -0.00001858, 0.00002406, -0.00002427, 0.00000846, -0.00001373, 0.00002740, -0.00001672, 0.00000423, -0.00000039, 0.00000141, -0.00001118, 0.00000788, -0.00000130, 0.00000139, -0.00000489, 0.00001035, -0.00000868, 0.00000780, -0.00000449, -0.00000774, 0.00000593, 0.00000334, -0.00000312, 0.00000545, -0.00000642, 0.00000192, -0.00000940, 0.00000855, -0.00000347, 0.00000590, -0.00000449, -0.00000168, },
		{ -0.00000302, 0.00000075, 0.00000553, -0.00000120, 0.00000366, -0.00000283, -0.00000832, 0.00000140, 0.00000515, -0.00000328, 0.00000224, -0.00000136, -0.00000097, 0.00000229, 0.00000203, 0.00000053, -0.00001124, 0.00000819, -0.00000254, 0.00000195, 0.00000407, -0.00000613, 0.00000014, 0.00000035, -0.00000000, 0.00000175, -0.00000256, 0.00000565, -0.00000697, 0.00000054, 0.00000278, -0.00000441, 0.00001004, -0.00000777, -0.00000406, 0.00000391, -0.00000058, 0.00000116, 0.00000042, 0.00000419, -0.00000927, 0.00000605, -0.00000566, 0.00001202, -0.00001310, 0.00000258, -0.00001298, 0.00001174, 0.00000484, 0.00000158, -0.00001407, 0.00001291, -0.00000507, -0.00000044, -0.00000266, 0.00000017, -0.00000168, -0.00001024, 0.00002354, -0.00000546, 0.00000235, -0.00001276, -0.00000832, 0.00001995, -0.00003173, 0.00001943, -0.00002301, 0.00001431, -0.00002479, -0.00000741, -0.00001234, 0.00000428, -0.00007649, 0.00002071, -0.00051663, 0.00118751, -0.00033191, -0.00012477, -0.00000496, -0.00003798, -0.00000640, 0.00000009, -0.00001839, 0.00000182, 0.00000382, 0.00000686, -0.00000514, 0.00000408, -0.00000621, 0.00000988, -0.00001110, -0.00002338, 0.00002643, -0.00000573, -0.00000055, -0.00000978, 0.00000243, -0.00000916, 0.00001611, -0.00001877, 0.00000260, -0.00000607, 0.00002161, -0.00001431, 0.00000213, 0.00000301, -0.00000242, -0.00000777, 0.00000512, 0.00000157, -0.00000205, -0.00000067, 0.00000651, -0.00000578, 0.00000565, -0.00000352, -0.00000954, 0.00001003, -0.00000099, 0.00000064, 0.00000211, -0.00000395, -0.00000059, -0.00000696, 0.00000678, -0.00000169, 0.00000428, -0.00000355, },
		{ -0.00000261, -0.00000428, 0.00000296, 0.00000395, -0.00000038, 0.00000337, -0.00000381, -0.00000793, 0.00000262, 0.00000416, -0.00000315, 0.00000243, -0.00000182, -0.00000044, 0.00000206, 0.00000266, -0.00000152, -0.00000944, 0.00000827, -0.00000306, 0.00000267, 0.00000331, -0.00000674, 0.00000134, -0.00000056, 0.00000104, 0.00000043, -0.00000125, 0.00000443, -0.00000619, 0.00000033, 0.00000325, -0.00000526, 0.00001207, -0.00001199, 0.00000038, 0.00000072, 0.00000248, -0.00000224, 0.00000455, -0.00000092, -0.00000432, 0.00000134, -0.00000072, 0.00000728, -0.00000902, -0.00000190, -0.00000897, 0.00000991, 0.00000712, -0.00000329, -0.00000912, 0.00000983, -0.00000220, -0.00000451, 0.00000207, -0.00000445, 0.00000271, -0.00001412, 0.00002985, -0.00001474, 0.00001390, -0.00002807, 0.00000969, 0.00000128, -0.00001419, 0.00000220, -0.00000511, -0.00000611, -0.00000276, -0.00003393, 0.00002125, -0.00004080, -0.00001867, -0.00007386, -0.00039046, 0.00116613, -0.00042819, -0.00004547, -0.00005118, -0.00000218, -0.00003378, 0.00002298, -0.00003823, 0.00002107, -0.00001176, 0.00002124, -0.00001784, 0.00001683, -0.00001898, 0.00002405, -0.00002822, -0.00000697, 0.00001343, 0.00000460, -0.00001074, -0.00000025, -0.00000691, 0.00000042, 0.00000776, -0.00001301, -0.00000317, 0.00000176, 0.00001510, -0.00001127, -0.00000008, 0.00000631, -0.00000616, -0.00000448, 0.00000261, 0.00000403, -0.00000499, 0.00000313, 0.00000287, -0.00000299, 0.00000361, -0.00000290, -0.00001059, 0.00001339, -0.00000493, 0.00000425, -0.00000124, -0.00000146, -0.00000307, -0.00000454, 0.00000503, 0.00000008, 0.00000262, },
		{ 0.00000104, -0.00000179, -0.00000531, 0.00000509, 0.00000220, 0.00000073, 0.00000277, -0.00000468, -0.00000731, 0.00000360, 0.00000319, -0.00000294, 0.00000256, -0.00000225, 0.00000009, 0.00000187, 0.00000312, -0.00000352, -0.00000729, 0.00000787, -0.00000323, 0.00000321, 0.00000247, -0.00000713, 0.00000242, -0.00000138, 0.00000196, -0.00000075, -0.00000006, 0.00000334, -0.00000555, 0.00000038, 0.00000324, -0.00000540, 0.00001316, -0.00001537, 0.00000439, -0.00000221, 0.00000516, -0.00000512, 0.00000811, -0.00000555, 0.00000035, -0.00000322, 0.00000418, 0.00000247, -0.00000483, -0.00000659, -0.00000451, 0.00000782, 0.00000910, -0.00000769, -0.00000436, 0.00000695, 0.00000007, -0.00000776, 0.00000595, -0.00000818, 0.00000602, -0.00001638, 0.00003382, -0.00002116, 0.00002226, -0.00004032, 0.00002523, -0.00001561, 0.00000199, -0.00001367, 0.00001120, -0.00002471, 0.00001722, -0.00005794, 0.00005189, -0.00008291, 0.00003651, -0.00016987, -0.00024657, 0.00111050, -0.00049492, 0.00001841, -0.00008861, 0.00002715, -0.00005612, 0.00004160, -0.00005421, 0.00003693, -0.00002425, 0.00003262, -0.00002740, 0.00002623, -0.00002820, 0.00003460, -0.00004236, 0.00000749, 0.00000165, 0.00001400, -0.00002006, 0.00000847, -0.00001549, 0.00000950, -0.00000043, -0.00000748, -0.00000838, 0.00000924, 0.00000832, -0.00000788, -0.00000221, 0.00000931, -0.00000965, -0.00000146, 0.00000048, 0.00000597, -0.00000728, 0.00000633, -0.00000037, -0.00000047, 0.00000179, -0.00000272, -0.00001076, 0.00001580, -0.00000822, 0.00000744, -0.00000435, 0.00000086, -0.00000535, -0.00000229, 0.00000345, 0.00000168, },
		{ 0.00000298, -0.00000032, -0.00000119, -0.00000600, 0.00000698, 0.00000041, 0.00000202, 0.00000188, -0.00000537, -0.00000651, 0.00000434, 0.00000225, -0.00000265, 0.00000261, -0.00000262, 0.00000060, 0.00000174, 0.00000334, -0.00000533, -0.00000494, 0.00000706, -0.00000308, 0.00000352, 0.00000161, -0.00000730, 0.00000335, -0.00000211, 0.00000273, -0.00000177, 0.00000097, 0.00000240, -0.00000506, 0.00000068, 0.00000278, -0.00000480, 0.00001322, -0.00001771, 0.00000771, -0.00000464, 0.00000727, -0.00000729, 0.00001083, -0.00000940, 0.00000440, -0.00000727, 0.00000869, -0.00000207, -0.00000087, -0.00001117, 0.00000013, 0.00000559, 0.00001068, -0.00001147, 0.00000004, 0.00000440, 0.00000167, -0.00001005, 0.00000883, -0.00001088, 0.00000815, -0.00001697, 0.00003534, -0.00002447, 0.00002693, -0.00004870, 0.00003727, -0.00002955, 0.00001558, -0.00002693, 0.00002462, -0.00004009, 0.00003355, -0.00007758, 0.00007728, -0.00011905, 0.00008487, -0.00026124, -0.00009206, 0.00102545, -0.00053301, 0.00006569, -0.00011587, 0.00004879, -0.00007242, 0.00005508, -0.00006560, 0.00004879, -0.00003314, 0.00004057, -0.00003347, 0.00003192, -0.00003347, 0.00004090, -0.00005268, 0.00001904, -0.00000801, 0.00002165, -0.00002781, 0.00001574, -0.00002269, 0.00001746, -0.00000793, -0.00000261, -0.00001258, 0.00001587, 0.00000171, -0.00000442, -0.00000408, 0.00001185, -0.00001273, 0.00000119, -0.00000120, 0.00000732, -0.00000882, 0.00000878, -0.00000304, 0.00000165, 0.00000028, -0.00000298, -0.00001003, 0.00001714, -0.00001064, 0.00001002, -0.00000704, 0.00000285, -0.00000729, -0.00000033, 0.00000217, },
		{ 0.00000127, 0.00000390, -0.00000139, -0.00000090, -0.00000626, 0.00000851, -0.00000127, 0.00000340, 0.00000074, -0.00000585, -0.00000559, 0.00000484, 0.00000137, -0.00000229, 0.00000256, -0.00000291, 0.00000106, 0.00000171, 0.00000328, -0.00000684, -0.00000253, 0.00000596, -0.00000264, 0.00000359, 0.00000077, -0.00000727, 0.00000412, -0.00000272, 0.00000334, -0.00000263, 0.00000185, 0.00000161, -0.00000471, 0.00000117, 0.00000192, -0.00000354, 0.00001223, -0.00001890, 0.00001018, -0.00000642, 0.00000866, -0.00000861, 0.00001253, -0.00001223, 0.00000754, -0.00001049, 0.00001248, -0.00000602, 0.00000255, -0.00001534, 0.00000471, 0.00000335, 0.00001181, -0.00001452, 0.00000395, 0.00000224, 0.00000257, -0.00001137, 0.00001068, -0.00001254, 0.00000917, -0.00001605, 0.00003459, -0.00002475, 0.00002779, -0.00005281, 0.00004512, -0.00003970, 0.00002571, -0.00003667, 0.00003420, -0.00005117, 0.00004503, -0.00009143, 0.00009562, -0.00014679, 0.00012276, -0.00034217, 0.00006559, 0.00091698, -0.00054483, 0.00009645, -0.00013255, 0.00006230, -0.00008224, 0.00006310, -0.00007211, 0.00005642, -0.00003828, 0.00004505, -0.00003607, 0.00003396, -0.00003476, 0.00004276, -0.00005872, 0.00002709, -0.00001491, 0.00002699, -0.00003346, 0.00002107, -0.00002804, 0.00002383, -0.00001430, 0.00000125, -0.00001545, 0.00002125, -0.00000432, -0.00000116, -0.00000554, 0.00001380, -0.00001531, 0.00000339, -0.00000239, 0.00000804, -0.00000957, 0.00001043, -0.00000507, 0.00000329, -0.00000092, -0.00000364, -0.00000845, 0.00001737, -0.00001210, 0.00001182, -0.00000918, 0.00000439, -0.00000879, 0.00000124, },
		{ 0.00000238, 0.00000079, 0.00000438, -0.00000211, -0.00000095, -0.00000603, 0.00000955, -0.00000271, 0.00000475, -0.00000059, -0.00000611, -0.00000460, 0.00000512, 0.00000057, -0.00000187, 0.00000242, -0.00000309, 0.00000145, 0.00000178, 0.00000291, -0.00000794, -0.00000022, 0.00000467, -0.00000197, 0.00000344, -0.00000002, -0.00000707, 0.00000472, -0.00000322, 0.00000380, -0.00000332, 0.00000259, 0.00000095, -0.00000445, 0.00000180, 0.00000080, -0.00000175, 0.00001031, -0.00001893, 0.00001170, -0.00000747, 0.00000930, -0.00000902, 0.00001312, -0.00001388, 0.00000959, -0.00001268, 0.00001531, -0.00000912, 0.00000517, -0.00001881, 0.00000899, 0.00000123, 0.00001245, -0.00001678, 0.00000729, 0.00000047, 0.00000282, -0.00001177, 0.00001156, -0.00001325, 0.00000926, -0.00001391, 0.00003194, -0.00002236, 0.00002507, -0.00005265, 0.00004852, -0.00004563, 0.00003187, -0.00004238, 0.00003936, -0.00005734, 0.00005094, -0.00009864, 0.00010575, -0.00016442, 0.00014740, -0.00040759, 0.00021897, 0.00079187, -0.00053398, 0.00011193, -0.00013911, 0.00006794, -0.00008576, 0.00006580, -0.00007390, 0.00005997, -0.00003988, 0.00004634, -0.00003555, 0.00003272, -0.00003246, 0.00004038, -0.00006044, 0.00003138, -0.00001873, 0.00002966, -0.00003671, 0.00002418, -0.00003124, 0.00002828, -0.00001925, 0.00000390, -0.00001678, 0.00002507, -0.00000944, 0.00000167, -0.00000645, 0.00001506, -0.00001729, 0.00000513, -0.00000310, 0.00000820, -0.00000959, 0.00001126, -0.00000643, 0.00000442, -0.00000182, -0.00000459, -0.00000616, 0.00001655, -0.00001257, 0.00001278, -0.00001067, 0.00000544, -0.00000982, },
		{ -0.00001037, 0.00000311, 0.00000071, 0.00000442, -0.00000247, -0.00000133, -0.00000529, 0.00001005, -0.00000380, 0.00000596, -0.00000205, -0.00000613, -0.00000360, 0.00000520, -0.00000014, -0.00000142, 0.00000218, -0.00000317, 0.00000176, 0.00000194, 0.00000221, -0.00000855, 0.00000185, 0.00000332, -0.00000116, 0.00000309, -0.00000075, -0.00000671, 0.00000515, -0.00000361, 0.00000413, -0.00000387, 0.00000322, 0.00000038, -0.00000422, 0.00000244, -0.00000045, 0.00000039, 0.00000760, -0.00001788, 0.00001228, -0.00000777, 0.00000918, -0.00000857, 0.00001264, -0.00001433, 0.00001045, -0.00001369, 0.00001702, -0.00001121, 0.00000679, -0.00002134, 0.00001274, -0.00000068, 0.00001259, -0.00001823, 0.00001002, -0.00000092, 0.00000254, -0.00001139, 0.00001159, -0.00001318, 0.00000870, -0.00001098, 0.00002794, -0.00001792, 0.00001935, -0.00004859, 0.00004762, -0.00004731, 0.00003395, -0.00004395, 0.00004001, -0.00005845, 0.00005111, -0.00009895, 0.00010725, -0.00017117, 0.00015706, -0.00045346, 0.00036124, 0.00065714, -0.00050485, 0.00011420, -0.00013675, 0.00006656, -0.00008363, 0.00006375, -0.00007147, 0.00005993, -0.00003843, 0.00004496, -0.00003253, 0.00002890, -0.00002727, 0.00003436, -0.00005817, 0.00003203, -0.00001944, 0.00002960, -0.00003748, 0.00002499, -0.00003219, 0.00003067, -0.00002262, 0.00000525, -0.00001650, 0.00002716, -0.00001342, 0.00000391, -0.00000673, 0.00001559, -0.00001866, 0.00000639, -0.00000339, 0.00000787, -0.00000895, 0.00001134, -0.00000715, 0.00000508, -0.00000248, -0.00000570, -0.00000333, 0.00001483, -0.00001209, 0.00001291, -0.00001150, 0.00000597, },
		{ 0.00000603, -0.00001049, 0.00000347, 0.00000100, 0.00000407, -0.00000250, -0.00000201, -0.00000407, 0.00000997, -0.00000446, 0.00000693, -0.00000355, -0.00000592, -0.00000262, 0.00000509, -0.00000074, -0.00000096, 0.00000186, -0.00000314, 0.00000201, 0.00000217, 0.00000122, -0.00000863, 0.00000357, 0.00000203, -0.00000030, 0.00000258, -0.00000140, -0.00000621, 0.00000541, -0.00000389, 0.00000434, -0.00000429, 0.00000375, -0.00000014, -0.00000398, 0.00000301, -0.00000166, 0.00000266, 0.00000436, -0.00001591, 0.00001200, -0.00000740, 0.00000843, -0.00000737, 0.00001119, -0.00001365, 0.00001015, -0.00001351, 0.00001756, -0.00001220, 0.00000729, -0.00002279, 0.00001579, -0.00000227, 0.00001225, -0.00001889, 0.00001215, -0.00000200, 0.00000185, -0.00001038, 0.00001095, -0.00001254, 0.00000777, -0.00000773, 0.00002325, -0.00001221, 0.00001144, -0.00004135, 0.00004297, -0.00004511, 0.00003225, -0.00004165, 0.00003645, -0.00005482, 0.00004588, -0.00009276, 0.00010048, -0.00016718, 0.00015122, -0.00047708, 0.00048652, 0.00051964, -0.00046227, 0.00010595, -0.00012719, 0.00005944, -0.00007689, 0.00005784, -0.00006560, 0.00005701, -0.00003461, 0.00004165, -0.00002780, 0.00002335, -0.00002015, 0.00002560, -0.00005259, 0.00002948, -0.00001732, 0.00002702, -0.00003592, 0.00002365, -0.00003099, 0.00003101, -0.00002441, 0.00000536, -0.00001467, 0.00002746, -0.00001610, 0.00000544, -0.00000637, 0.00001537, -0.00001940, 0.00000722, -0.00000333, 0.00000716, -0.00000780, 0.00001078, -0.00000729, 0.00000531, -0.00000297, -0.00000680, -0.00000021, 0.00001240, -0.00001081, 0.00001226, -0.00001171, },
		{ -0.00001138, 0.00000568, -0.00001026, 0.00000355, 0.00000155, 0.00000339, -0.00000226, -0.00000291, -0.00000246, 0.00000934, -0.00000463, 0.00000758, -0.00000501, -0.00000550, -0.00000171, 0.00000484, -0.00000122, -0.00000050, 0.00000148, -0.00000301, 0.00000219, 0.00000242, -0.00000002, -0.00000817, 0.00000485, 0.00000092, 0.00000052, 0.00000197, -0.00000197, -0.00000560, 0.00000551, -0.00000406, 0.00000443, -0.00000461, 0.00000423, -0.00000066, -0.00000364, 0.00000340, -0.00000268, 0.00000482, 0.00000086, -0.00001326, 0.00001100, -0.00000649, 0.00000718, -0.00000561, 0.00000896, -0.00001199, 0.00000881, -0.00001223, 0.00001697, -0.00001211, 0.00000667, -0.00002306, 0.00001801, -0.00000349, 0.00001146, -0.00001882, 0.00001369, -0.00000283, 0.00000092, -0.00000894, 0.00000985, -0.00001155, 0.00000680, -0.00000463, 0.00001854, -0.00000610, 0.00000233, -0.00003189, 0.00003537, -0.00003972, 0.00002738, -0.00003614, 0.00002936, -0.00004721, 0.00003609, -0.00008101, 0.00008646, -0.00015351, 0.00013059, -0.00047724, 0.00059025, 0.00038559, -0.00041111, 0.00009017, -0.00011246, 0.00004815, -0.00006680, 0.00004914, -0.00005724, 0.00005204, -0.00002921, 0.00003721, -0.00002222, 0.00001703, -0.00001215, 0.00001517, -0.00004463, 0.00002442, -0.00001290, 0.00002231, -0.00003240, 0.00002046, -0.00002791, 0.00002950, -0.00002476, 0.00000441, -0.00001151, 0.00002608, -0.00001747, 0.00000623, -0.00000539, 0.00001444, -0.00001954, 0.00000767, -0.00000302, 0.00000623, -0.00000628, 0.00000971, -0.00000696, 0.00000518, -0.00000337, -0.00000773, 0.00000298, 0.00000950, -0.00000891, 0.00001098, },
		{ 0.00000920, -0.00001063, 0.00000504, -0.00000978, 0.00000345, 0.00000226, 0.00000249, -0.00000185, -0.00000393, -0.00000055, 0.00000821, -0.00000432, 0.00000784, -0.00000635, -0.00000492, -0.00000089, 0.00000447, -0.00000157, -0.00000008, 0.00000106, -0.00000281, 0.00000233, 0.00000266, -0.00000142, -0.00000721, 0.00000565, 0.00000007, 0.00000122, 0.00000132, -0.00000246, -0.00000488, 0.00000545, -0.00000412, 0.00000444, -0.00000484, 0.00000467, -0.00000119, -0.00000319, 0.00000354, -0.00000335, 0.00000665, -0.00000263, -0.00001019, 0.00000950, -0.00000521, 0.00000564, -0.00000353, 0.00000622, -0.00000962, 0.00000664, -0.00001002, 0.00001538, -0.00001103, 0.00000499, -0.00002215, 0.00001934, -0.00000431, 0.00001028, -0.00001808, 0.00001470, -0.00000349, -0.00000013, -0.00000726, 0.00000847, -0.00001042, 0.00000604, -0.00000211, 0.00001444, -0.00000043, -0.00000695, -0.00002130, 0.00002585, -0.00003206, 0.00002023, -0.00002831, 0.00001971, -0.00003667, 0.00002293, -0.00006508, 0.00006681, -0.00013203, 0.00009700, -0.00045424, 0.00066936, 0.00026029, -0.00035594, 0.00006984, -0.00009467, 0.00003435, -0.00005471, 0.00003882, -0.00004737, 0.00004589, -0.00002304, 0.00003243, -0.00001663, 0.00001087, -0.00000436, 0.00000424, -0.00003532, 0.00001773, -0.00000687, 0.00001607, -0.00002741, 0.00001590, -0.00002337, 0.00002645, -0.00002393, 0.00000267, -0.00000732, 0.00002324, -0.00001759, 0.00000634, -0.00000386, 0.00001289, -0.00001916, 0.00000780, -0.00000255, 0.00000520, -0.00000458, 0.00000830, -0.00000628, 0.00000479, -0.00000377, -0.00000835, 0.00000598, 0.00000640, -0.00000659, },
		{ -0.00000409, 0.00000713, -0.00000958, 0.00000420, -0.00000915, 0.00000329, 0.00000301, 0.00000149, -0.00000137, -0.00000493, 0.00000150, 0.00000670, -0.00000357, 0.00000769, -0.00000751, -0.00000420, -0.00000019, 0.00000400, -0.00000179, 0.00000029, 0.00000064, -0.00000255, 0.00000246, 0.00000281, -0.00000288, -0.00000582, 0.00000596, -0.00000048, 0.00000172, 0.00000067, -0.00000290, -0.00000408, 0.00000524, -0.00000409, 0.00000435, -0.00000498, 0.00000508, -0.00000175, -0.00000259, 0.00000340, -0.00000360, 0.00000796, -0.00000583, -0.00000697, 0.00000771, -0.00000376, 0.00000401, -0.00000138, 0.00000325, -0.00000681, 0.00000391, -0.00000712, 0.00001302, -0.00000916, 0.00000240, -0.00002014, 0.00001976, -0.00000472, 0.00000879, -0.00001677, 0.00001522, -0.00000404, -0.00000114, -0.00000550, 0.00000698, -0.00000929, 0.00000567, -0.00000045, 0.00001146, 0.00000403, -0.00001540, -0.00001071, 0.00001554, -0.00002318, 0.00001181, -0.00001922, 0.00000867, -0.00002447, 0.00000783, -0.00004667, 0.00004356, -0.00010521, 0.00005330, -0.00040985, 0.00072238, 0.00014785, -0.00030076, 0.00004772, -0.00007584, 0.00001963, -0.00004191, 0.00002797, -0.00003694, 0.00003938, -0.00001680, 0.00002798, -0.00001174, 0.00000566, 0.00000227, -0.00000609, -0.00002576, 0.00001033, -0.00000003, 0.00000897, -0.00002155, 0.00001051, -0.00001786, 0.00002227, -0.00002222, 0.00000051, -0.00000251, 0.00001925, -0.00001663, 0.00000586, -0.00000190, 0.00001082, -0.00001833, 0.00000769, -0.00000201, 0.00000420, -0.00000285, 0.00000670, -0.00000537, 0.00000421, -0.00000422, -0.00000854, 0.00000858, 0.00000334, },
		{ 0.00000054, -0.00000162, 0.00000492, -0.00000838, 0.00000329, -0.00000848, 0.00000317, 0.00000367, 0.00000050, -0.00000094, -0.00000581, 0.00000356, 0.00000493, -0.00000243, 0.00000713, -0.00000843, -0.00000338, 0.00000039, 0.00000347, -0.00000189, 0.00000057, 0.00000024, -0.00000228, 0.00000262, 0.00000283, -0.00000430, -0.00000411, 0.00000581, -0.00000070, 0.00000199, 0.00000010, -0.00000329, -0.00000321, 0.00000489, -0.00000394, 0.00000416, -0.00000504, 0.00000546, -0.00000234, -0.00000186, 0.00000295, -0.00000336, 0.00000862, -0.00000851, -0.00000386, 0.00000584, -0.00000232, 0.00000251, 0.00000058, 0.00000035, -0.00000388, 0.00000094, -0.00000383, 0.00001014, -0.00000675, -0.00000086, -0.00001719, 0.00001932, -0.00000475, 0.00000707, -0.00001499, 0.00001530, -0.00000453, -0.00000203, -0.00000380, 0.00000550, -0.00000826, 0.00000578, 0.00000015, 0.00000993, 0.00000671, -0.00002217, -0.00000115, 0.00000552, -0.00001414, 0.00000317, -0.00000997, -0.00000258, -0.00001193, -0.00000769, -0.00002759, 0.00001894, -0.00007588, 0.00000306, -0.00034710, 0.00074941, 0.00005107, -0.00024878, 0.00002617, -0.00005770, 0.00000537, -0.00002955, 0.00001755, -0.00002675, 0.00003316, -0.00001105, 0.00002439, -0.00000808, 0.00000198, 0.00000696, -0.00001486, -0.00001693, 0.00000316, 0.00000682, 0.00000170, -0.00001545, 0.00000487, -0.00001194, 0.00001743, -0.00002002, -0.00000172, 0.00000249, 0.00001448, -0.00001481, 0.00000495, 0.00000032, 0.00000837, -0.00001715, 0.00000741, -0.00000150, 0.00000336, -0.00000124, 0.00000506, -0.00000436, 0.00000353, -0.00000476, -0.00000823, 0.00001063, },
		{ 0.00001201, -0.00000182, 0.00000063, 0.00000274, -0.00000714, 0.00000239, -0.00000786, 0.00000318, 0.00000414, -0.00000037, -0.00000063, -0.00000643, 0.00000547, 0.00000304, -0.00000101, 0.00000618, -0.00000908, -0.00000251, 0.00000083, 0.00000290, -0.00000188, 0.00000077, -0.00000011, -0.00000202, 0.00000281, 0.00000264, -0.00000556, -0.00000220, 0.00000527, -0.00000060, 0.00000199, -0.00000038, -0.00000363, -0.00000228, 0.00000440, -0.00000370, 0.00000389, -0.00000501, 0.00000579, -0.00000295, -0.00000102, 0.00000223, -0.00000264, 0.00000855, -0.00001050, -0.00000109, 0.00000410, -0.00000106, 0.00000132, 0.00000214, -0.00000219, -0.00000114, -0.00000197, -0.00000047, 0.00000706, -0.00000407, -0.00000453, -0.00001351, 0.00001812, -0.00000444, 0.00000522, -0.00001285, 0.00001499, -0.00000497, -0.00000274, -0.00000225, 0.00000411, -0.00000734, 0.00000634, -0.00000029, 0.00000999, 0.00000726, -0.00002665, 0.00000654, -0.00000326, -0.00000592, -0.00000472, -0.00000159, -0.00001291, -0.00000033, -0.00002218, -0.00000959, -0.00000485, -0.00004700, -0.00004972, -0.00027003, 0.00075200, -0.00002855, -0.00020234, 0.00000697, -0.00004158, -0.00000734, -0.00001851, 0.00000829, -0.00001741, 0.00002771, -0.00000618, 0.00002194, -0.00000592, 0.00000015, 0.00000925, -0.00002137, -0.00000965, -0.00000299, 0.00001295, -0.00000508, -0.00000969, -0.00000046, -0.00000612, 0.00001237, -0.00001767, -0.00000365, 0.00000726, 0.00000935, -0.00001241, 0.00000378, 0.00000264, 0.00000569, -0.00001571, 0.00000703, -0.00000106, 0.00000276, 0.00000013, 0.00000351, -0.00000334, 0.00000278, -0.00000539, -0.00000741, },
		{ -0.00000610, 0.00001267, -0.00000362, 0.00000253, 0.00000072, -0.00000596, 0.00000158, -0.00000736, 0.00000339, 0.00000435, -0.00000104, -0.00000053, -0.00000672, 0.00000710, 0.00000118, 0.00000057, 0.00000491, -0.00000943, -0.00000162, 0.00000114, 0.00000233, -0.00000176, 0.00000087, -0.00000039, -0.00000178, 0.00000307, 0.00000222, -0.00000654, -0.00000023, 0.00000444, -0.00000023, 0.00000173, -0.00000074, -0.00000394, -0.00000132, 0.00000380, -0.00000335, 0.00000352, -0.00000487, 0.00000605, -0.00000352, -0.00000013, 0.00000132, -0.00000151, 0.00000777, -0.00001169, 0.00000119, 0.00000264, -0.00000011, 0.00000055, 0.00000312, -0.00000417, 0.00000115, -0.00000451, 0.00000268, 0.00000406, -0.00000144, -0.00000830, -0.00000934, 0.00001630, -0.00000386, 0.00000332, -0.00001045, 0.00001434, -0.00000539, -0.00000324, -0.00000089, 0.00000283, -0.00000650, 0.00000723, -0.00000164, 0.00001157, 0.00000560, -0.00002849, 0.00001177, -0.00001004, 0.00000071, -0.00001106, 0.00000508, -0.00002137, 0.00000929, -0.00003442, 0.00000581, -0.00002583, -0.00002134, -0.00010092, -0.00018335, 0.00073288, -0.00009078, -0.00016284, -0.00000872, -0.00002838, -0.00001778, -0.00000937, 0.00000070, -0.00000931, 0.00002331, -0.00000232, 0.00002071, -0.00000531, 0.00000020, 0.00000894, -0.00002520, -0.00000450, -0.00000752, 0.00001778, -0.00001084, -0.00000475, -0.00000501, -0.00000088, 0.00000750, -0.00001548, -0.00000497, 0.00001139, 0.00000424, -0.00000970, 0.00000254, 0.00000489, 0.00000292, -0.00001411, 0.00000660, -0.00000074, 0.00000243, 0.00000117, 0.00000215, -0.00000241, 0.00000202, -0.00000607, },
		{ -0.00000676, -0.00000440, 0.00001264, -0.00000482, 0.00000399, -0.00000105, -0.00000491, 0.00000090, -0.00000698, 0.00000381, 0.00000426, -0.00000147, -0.00000069, -0.00000660, 0.00000832, -0.00000049, 0.00000219, 0.00000338, -0.00000949, -0.00000075, 0.00000130, 0.00000179, -0.00000157, 0.00000087, -0.00000060, -0.00000158, 0.00000339, 0.00000154, -0.00000716, 0.00000165, 0.00000343, 0.00000034, 0.00000124, -0.00000097, -0.00000422, -0.00000036, 0.00000309, -0.00000291, 0.00000305, -0.00000461, 0.00000621, -0.00000402, 0.00000074, 0.00000031, -0.00000007, 0.00000634, -0.00001207, 0.00000287, 0.00000154, 0.00000047, 0.00000027, 0.00000344, -0.00000541, 0.00000278, -0.00000646, 0.00000534, 0.00000141, 0.00000088, -0.00001187, -0.00000496, 0.00001402, -0.00000310, 0.00000148, -0.00000789, 0.00001339, -0.00000576, -0.00000353, 0.00000028, 0.00000164, -0.00000563, 0.00000823, -0.00000362, 0.00001438, 0.00000189, -0.00002764, 0.00001423, -0.00001438, 0.00000521, -0.00001530, 0.00000945, -0.00002731, 0.00001615, -0.00004348, 0.00001746, -0.00004242, -0.00000127, -0.00014668, -0.00009206, 0.00069568, -0.00013647, -0.00013084, -0.00002030, -0.00001853, -0.00002559, -0.00000241, -0.00000501, -0.00000261, 0.00002001, 0.00000056, 0.00002057, -0.00000606, 0.00000191, 0.00000619, -0.00002625, -0.00000177, -0.00001004, 0.00002091, -0.00001520, -0.00000099, -0.00000844, 0.00000341, 0.00000315, -0.00001369, -0.00000547, 0.00001457, -0.00000047, -0.00000696, 0.00000141, 0.00000691, 0.00000022, -0.00001245, 0.00000617, -0.00000055, 0.00000241, 0.00000183, 0.00000105, -0.00000162, 0.00000125, },
		{ 0.00000049, -0.00000738, -0.00000243, 0.00001200, -0.00000541, 0.00000497, -0.00000252, -0.00000401, 0.00000036, -0.00000674, 0.00000444, 0.00000387, -0.00000162, -0.00000110, -0.00000606, 0.00000907, -0.00000187, 0.00000372, 0.00000168, -0.00000927, 0.00000007, 0.00000132, 0.00000130, -0.00000133, 0.00000079, -0.00000073, -0.00000141, 0.00000376, 0.00000059, -0.00000734, 0.00000333, 0.00000236, 0.00000103, 0.00000057, -0.00000109, -0.00000445, 0.00000059, 0.00000231, -0.00000238, 0.00000248, -0.00000422, 0.00000623, -0.00000441, 0.00000151, -0.00000068, 0.00000153, 0.00000440, -0.00001168, 0.00000393, 0.00000086, 0.00000065, 0.00000049, 0.00000306, -0.00000586, 0.00000363, -0.00000766, 0.00000732, -0.00000066, 0.00000263, -0.00001497, -0.00000064, 0.00001147, -0.00000224, -0.00000022, -0.00000529, 0.00001219, -0.00000606, -0.00000365, 0.00000127, 0.00000050, -0.00000461, 0.00000908, -0.00000584, 0.00001799, -0.00000346, -0.00002434, 0.00001394, -0.00001610, 0.00000732, -0.00001717, 0.00001121, -0.00003035, 0.00001982, -0.00004886, 0.00002465, -0.00005361, 0.00001149, -0.00018364, -0.00000111, 0.00064462, -0.00016730, -0.00010615, -0.00002777, -0.00001206, -0.00003075, 0.00000235, -0.00000884, 0.00000275, 0.00001771, 0.00000268, 0.00002122, -0.00000777, 0.00000481, 0.00000142, -0.00002473, -0.00000145, -0.00001043, 0.00002216, -0.00001796, 0.00000140, -0.00001054, 0.00000653, -0.00000045, -0.00001244, -0.00000502, 0.00001658, -0.00000450, -0.00000441, 0.00000054, 0.00000856, -0.00000230, -0.00001080, 0.00000576, -0.00000047, 0.00000267, 0.00000211, 0.00000024, -0.00000100, },
		{ -0.00000058, -0.00000027, -0.00000786, -0.00000032, 0.00001087, -0.00000545, 0.00000547, -0.00000367, -0.00000327, -0.00000006, -0.00000658, 0.00000523, 0.00000320, -0.00000154, -0.00000174, -0.00000512, 0.00000931, -0.00000287, 0.00000506, -0.00000008, -0.00000880, 0.00000082, 0.00000121, 0.00000089, -0.00000107, 0.00000065, -0.00000078, -0.00000128, 0.00000415, -0.00000058, -0.00000706, 0.00000468, 0.00000135, 0.00000174, -0.00000021, -0.00000110, -0.00000463, 0.00000149, 0.00000150, -0.00000179, 0.00000183, -0.00000369, 0.00000611, -0.00000463, 0.00000211, -0.00000153, 0.00000313, 0.00000211, -0.00001063, 0.00000440, 0.00000056, 0.00000050, 0.00000113, 0.00000206, -0.00000552, 0.00000366, -0.00000801, 0.00000850, -0.00000201, 0.00000366, -0.00001737, 0.00000336, 0.00000882, -0.00000137, -0.00000169, -0.00000273, 0.00001077, -0.00000627, -0.00000364, 0.00000212, -0.00000065, -0.00000331, 0.00000951, -0.00000788, 0.00002186, -0.00000987, -0.00001908, 0.00001119, -0.00001533, 0.00000711, -0.00001669, 0.00001038, -0.00003047, 0.00002024, -0.00005046, 0.00002720, -0.00005900, 0.00001595, -0.00020922, 0.00008497, 0.00058408, -0.00018556, -0.00008800, -0.00003153, -0.00000863, -0.00003351, 0.00000515, -0.00001098, 0.00000695, 0.00001615, 0.00000434, 0.00002224, -0.00000997, 0.00000828, -0.00000474, -0.00002112, -0.00000327, -0.00000882, 0.00002157, -0.00001912, 0.00000239, -0.00001126, 0.00000836, -0.00000321, -0.00001176, -0.00000364, 0.00001733, -0.00000766, -0.00000222, 0.00000003, 0.00000974, -0.00000454, -0.00000921, 0.00000537, -0.00000047, 0.00000316, 0.00000203, -0.00000026, },
		{ -0.00000049, -0.00000036, -0.00000104, -0.00000810, 0.00000177, 0.00000940, -0.00000505, 0.00000554, -0.00000453, -0.00000267, -0.00000040, -0.00000646, 0.00000610, 0.00000232, -0.00000126, -0.00000255, -0.00000383, 0.00000905, -0.00000342, 0.00000610, -0.00000183, -0.00000811, 0.00000149, 0.00000100, 0.00000058, -0.00000083, 0.00000047, -0.00000079, -0.00000116, 0.00000453, -0.00000193, -0.00000631, 0.00000564, 0.00000050, 0.00000237, -0.00000104, -0.00000105, -0.00000473, 0.00000231, 0.00000067, -0.00000117, 0.00000111, -0.00000302, 0.00000581, -0.00000466, 0.00000249, -0.00000214, 0.00000454, -0.00000031, -0.00000909, 0.00000439, 0.00000057, 0.00000010, 0.00000207, 0.00000055, -0.00000449, 0.00000292, -0.00000753, 0.00000885, -0.00000258, 0.00000385, -0.00001890, 0.00000684, 0.00000625, -0.00000058, -0.00000287, -0.00000032, 0.00000919, -0.00000636, -0.00000355, 0.00000287, -0.00000184, -0.00000164, 0.00000928, -0.00000930, 0.00002540, -0.00001664, -0.00001252, 0.00000652, -0.00001245, 0.00000485, -0.00001414, 0.00000724, -0.00002797, 0.00001772, -0.00004859, 0.00002543, -0.00005882, 0.00001197, -0.00022181, 0.00016235, 0.00051839, -0.00019385, -0.00007524, -0.00003231, -0.00000766, -0.00003435, 0.00000638, -0.00001176, 0.00001029, 0.00001503, 0.00000594, 0.00002318, -0.00001211, 0.00001163, -0.00001150, -0.00001608, -0.00000675, -0.00000552, 0.00001935, -0.00001883, 0.00000214, -0.00001072, 0.00000896, -0.00000512, -0.00001159, -0.00000142, 0.00001684, -0.00000985, -0.00000050, -0.00000006, 0.00001041, -0.00000643, -0.00000774, 0.00000500, -0.00000050, 0.00000381, 0.00000164, },
		{ 0.00000102, -0.00000048, -0.00000032, -0.00000182, -0.00000807, 0.00000370, 0.00000774, -0.00000432, 0.00000527, -0.00000514, -0.00000216, -0.00000071, -0.00000631, 0.00000697, 0.00000131, -0.00000086, -0.00000345, -0.00000228, 0.00000835, -0.00000353, 0.00000680, -0.00000348, -0.00000726, 0.00000204, 0.00000068, 0.00000039, -0.00000062, 0.00000029, -0.00000078, -0.00000103, 0.00000484, -0.00000339, -0.00000513, 0.00000617, -0.00000011, 0.00000284, -0.00000185, -0.00000097, -0.00000474, 0.00000301, -0.00000012, -0.00000054, 0.00000035, -0.00000225, 0.00000535, -0.00000448, 0.00000260, -0.00000241, 0.00000562, -0.00000265, -0.00000726, 0.00000401, 0.00000080, -0.00000043, 0.00000316, -0.00000129, -0.00000293, 0.00000155, -0.00000631, 0.00000841, -0.00000239, 0.00000318, -0.00001948, 0.00000963, 0.00000389, 0.00000006, -0.00000369, 0.00000188, 0.00000750, -0.00000632, -0.00000342, 0.00000354, -0.00000310, 0.00000048, 0.00000820, -0.00000976, 0.00002804, -0.00002306, -0.00000543, 0.00000063, -0.00000803, 0.00000106, -0.00001003, 0.00000231, -0.00002338, 0.00001282, -0.00004392, 0.00002007, -0.00005391, 0.00000022, -0.00022079, 0.00022807, 0.00045143, -0.00019484, -0.00006653, -0.00003102, -0.00000845, -0.00003382, 0.00000650, -0.00001156, 0.00001309, 0.00001404, 0.00000781, 0.00002361, -0.00001368, 0.00001419, -0.00001805, -0.00001040, -0.00001126, -0.00000105, 0.00001587, -0.00001739, 0.00000089, -0.00000914, 0.00000846, -0.00000628, -0.00001178, 0.00000142, 0.00001523, -0.00001106, 0.00000073, 0.00000027, 0.00001054, -0.00000794, -0.00000640, 0.00000464, -0.00000051, 0.00000454, },
		{ 0.00000527, 0.00000024, -0.00000028, -0.00000044, -0.00000260, -0.00000771, 0.00000533, 0.00000607, -0.00000341, 0.00000475, -0.00000556, -0.00000170, -0.00000106, -0.00000603, 0.00000774, 0.00000028, -0.00000043, -0.00000433, -0.00000060, 0.00000728, -0.00000321, 0.00000711, -0.00000495, -0.00000627, 0.00000248, 0.00000029, 0.00000031, -0.00000049, 0.00000014, -0.00000076, -0.00000085, 0.00000503, -0.00000486, -0.00000358, 0.00000626, -0.00000044, 0.00000309, -0.00000256, -0.00000090, -0.00000463, 0.00000356, -0.00000084, 0.00000006, -0.00000042, -0.00000138, 0.00000475, -0.00000412, 0.00000244, -0.00000230, 0.00000624, -0.00000472, -0.00000534, 0.00000343, 0.00000111, -0.00000094, 0.00000420, -0.00000323, -0.00000106, -0.00000026, -0.00000451, 0.00000732, -0.00000153, 0.00000173, -0.00001910, 0.00001163, 0.00000186, 0.00000050, -0.00000413, 0.00000377, 0.00000575, -0.00000615, -0.00000328, 0.00000415, -0.00000440, 0.00000301, 0.00000617, -0.00000898, 0.00002933, -0.00002843, 0.00000138, -0.00000570, -0.00000279, -0.00000362, -0.00000500, -0.00000372, -0.00001743, 0.00000635, -0.00003735, 0.00001223, -0.00004556, -0.00001789, -0.00020660, 0.00028024, 0.00038649, -0.00019102, -0.00006048, -0.00002858, -0.00001024, -0.00003252, 0.00000601, -0.00001077, 0.00001566, 0.00001291, 0.00001024, 0.00002319, -0.00001427, 0.00001539, -0.00002361, -0.00000486, -0.00001609, 0.00000403, 0.00001162, -0.00001519, -0.00000101, -0.00000682, 0.00000710, -0.00000685, -0.00001212, 0.00000461, 0.00001272, -0.00001138, 0.00000150, 0.00000096, 0.00001018, -0.00000906, -0.00000520, 0.00000427, -0.00000044, },
		{ -0.00000024, 0.00000589, -0.00000061, 0.00000005, -0.00000071, -0.00000337, -0.00000701, 0.00000658, 0.00000452, -0.00000244, 0.00000407, -0.00000584, -0.00000125, -0.00000148, -0.00000558, 0.00000833, -0.00000068, -0.00000006, -0.00000509, 0.00000109, 0.00000596, -0.00000253, 0.00000703, -0.00000621, -0.00000519, 0.00000279, -0.00000014, 0.00000035, -0.00000043, 0.00000003, -0.00000077, -0.00000061, 0.00000504, -0.00000622, -0.00000178, 0.00000594, -0.00000048, 0.00000309, -0.00000312, -0.00000086, -0.00000441, 0.00000395, -0.00000145, 0.00000059, -0.00000114, -0.00000048, 0.00000402, -0.00000358, 0.00000202, -0.00000180, 0.00000632, -0.00000635, -0.00000350, 0.00000278, 0.00000139, -0.00000130, 0.00000503, -0.00000506, 0.00000088, -0.00000229, -0.00000234, 0.00000576, -0.00000016, -0.00000038, -0.00001781, 0.00001282, 0.00000022, 0.00000073, -0.00000416, 0.00000531, 0.00000402, -0.00000587, -0.00000313, 0.00000469, -0.00000569, 0.00000590, 0.00000322, -0.00000683, 0.00002892, -0.00003221, 0.00000720, -0.00001171, 0.00000257, -0.00000852, 0.00000024, -0.00001010, -0.00001093, -0.00000080, -0.00002995, 0.00000315, -0.00003539, -0.00004035, -0.00018061, 0.00031801, 0.00032611, -0.00018455, -0.00005588, -0.00002584, -0.00001236, -0.00003100, 0.00000534, -0.00000971, 0.00001822, 0.00001146, 0.00001342, 0.00002171, -0.00001366, 0.00001483, -0.00002757, -0.00000019, -0.00002058, 0.00000913, 0.00000709, -0.00001266, -0.00000316, -0.00000413, 0.00000517, -0.00000705, -0.00001238, 0.00000784, 0.00000959, -0.00001097, 0.00000189, 0.00000192, 0.00000939, -0.00000982, -0.00000414, 0.00000388, },
		{ 0.00000344, 0.00000013, 0.00000634, -0.00000144, 0.00000045, -0.00000111, -0.00000411, -0.00000600, 0.00000740, 0.00000319, -0.00000152, 0.00000331, -0.00000604, -0.00000078, -0.00000201, -0.00000491, 0.00000865, -0.00000149, 0.00000018, -0.00000564, 0.00000267, 0.00000453, -0.00000158, 0.00000658, -0.00000720, -0.00000407, 0.00000295, -0.00000059, 0.00000048, -0.00000047, -0.00000002, -0.00000083, -0.00000028, 0.00000483, -0.00000739, 0.00000016, 0.00000529, -0.00000024, 0.00000281, -0.00000351, -0.00000088, -0.00000405, 0.00000415, -0.00000193, 0.00000101, -0.00000179, 0.00000042, 0.00000324, -0.00000293, 0.00000140, -0.00000094, 0.00000583, -0.00000744, -0.00000191, 0.00000221, 0.00000153, -0.00000140, 0.00000548, -0.00000656, 0.00000267, -0.00000429, -0.00000005, 0.00000394, 0.00000152, -0.00000295, -0.00001576, 0.00001321, -0.00000100, 0.00000074, -0.00000379, 0.00000645, 0.00000237, -0.00000549, -0.00000296, 0.00000511, -0.00000688, 0.00000897, -0.00000052, -0.00000335, 0.00002667, -0.00003401, 0.00001143, -0.00001672, 0.00000736, -0.00001296, 0.00000504, -0.00001610, -0.00000468, -0.00000775, -0.00002275, -0.00000583, -0.00002516, -0.00006480, -0.00014496, 0.00034153, 0.00027199, -0.00017713, -0.00005172, -0.00002348, -0.00001426, -0.00002968, 0.00000483, -0.00000860, 0.00002090, 0.00000963, 0.00001737, 0.00001913, -0.00001179, 0.00001231, -0.00002949, 0.00000305, -0.00002411, 0.00001368, 0.00000275, -0.00001020, -0.00000519, -0.00000144, 0.00000296, -0.00000708, -0.00001233, 0.00001080, 0.00000616, -0.00001002, 0.00000204, 0.00000302, 0.00000828, -0.00001027, -0.00000319, },
		{ -0.00000235, 0.00000295, 0.00000071, 0.00000655, -0.00000217, 0.00000085, -0.00000161, -0.00000476, -0.00000474, 0.00000777, 0.00000216, -0.00000074, 0.00000256, -0.00000618, -0.00000027, -0.00000266, -0.00000400, 0.00000866, -0.00000209, 0.00000024, -0.00000589, 0.00000400, 0.00000311, -0.00000045, 0.00000583, -0.00000793, -0.00000292, 0.00000297, -0.00000102, 0.00000069, -0.00000059, 0.00000002, -0.00000095, 0.00000015, 0.00000436, -0.00000826, 0.00000212, 0.00000439, 0.00000023, 0.00000227, -0.00000370, -0.00000097, -0.00000359, 0.00000417, -0.00000225, 0.00000130, -0.00000231, 0.00000127, 0.00000244, -0.00000223, 0.00000065, 0.00000020, 0.00000481, -0.00000792, -0.00000067, 0.00000182, 0.00000145, -0.00000115, 0.00000544, -0.00000757, 0.00000408, -0.00000603, 0.00000214, 0.00000210, 0.00000326, -0.00000575, -0.00001313, 0.00001290, -0.00000181, 0.00000056, -0.00000306, 0.00000715, 0.00000087, -0.00000508, -0.00000275, 0.00000538, -0.00000783, 0.00001203, -0.00000480, 0.00000128, 0.00002263, -0.00003364, 0.00001368, -0.00002020, 0.00001102, -0.00001640, 0.00000879, -0.00002110, 0.00000063, -0.00001371, -0.00001669, -0.00001351, -0.00001658, -0.00008870, -0.00010236, 0.00035186, 0.00022503, -0.00016997, -0.00004730, -0.00002198, -0.00001557, -0.00002884, 0.00000468, -0.00000757, 0.00002374, 0.00000748, 0.00002200, 0.00001555, -0.00000879, 0.00000791, -0.00002917, 0.00000446, -0.00002625, 0.00001724, -0.00000099, -0.00000817, -0.00000676, 0.00000094, 0.00000075, -0.00000712, -0.00001180, 0.00001320, 0.00000275, -0.00000876, 0.00000209, 0.00000413, 0.00000697, -0.00001046, },
		{ -0.00001047, -0.00000159, 0.00000242, 0.00000148, 0.00000648, -0.00000275, 0.00000121, -0.00000220, -0.00000528, -0.00000330, 0.00000772, 0.00000145, -0.00000014, 0.00000185, -0.00000630, 0.00000027, -0.00000340, -0.00000286, 0.00000835, -0.00000244, 0.00000007, -0.00000582, 0.00000501, 0.00000182, 0.00000073, 0.00000482, -0.00000839, -0.00000178, 0.00000284, -0.00000140, 0.00000094, -0.00000079, 0.00000012, -0.00000113, 0.00000067, 0.00000361, -0.00000875, 0.00000396, 0.00000336, 0.00000085, 0.00000152, -0.00000371, -0.00000112, -0.00000301, 0.00000402, -0.00000241, 0.00000143, -0.00000267, 0.00000200, 0.00000169, -0.00000154, -0.00000012, 0.00000152, 0.00000334, -0.00000781, 0.00000015, 0.00000165, 0.00000113, -0.00000053, 0.00000485, -0.00000798, 0.00000495, -0.00000730, 0.00000400, 0.00000047, 0.00000484, -0.00000854, -0.00001014, 0.00001201, -0.00000226, 0.00000023, -0.00000201, 0.00000741, -0.00000045, -0.00000466, -0.00000248, 0.00000543, -0.00000842, 0.00001481, -0.00000932, 0.00000675, 0.00001704, -0.00003117, 0.00001378, -0.00002182, 0.00001316, -0.00001845, 0.00001110, -0.00002461, 0.00000447, -0.00001809, -0.00001250, -0.00001892, -0.00001113, -0.00010959, -0.00005583, 0.00035075, 0.00018541, -0.00016374, -0.00004225, -0.00002156, -0.00001611, -0.00002862, 0.00000500, -0.00000663, 0.00002662, 0.00000519, 0.00002708, 0.00001121, -0.00000499, 0.00000189, -0.00002668, 0.00000391, -0.00002674, 0.00001952, -0.00000384, -0.00000682, -0.00000764, 0.00000273, -0.00000125, -0.00000731, -0.00001067, 0.00001482, -0.00000039, -0.00000740, 0.00000218, 0.00000510, 0.00000558, },
		{ 0.00000422, -0.00001036, -0.00000089, 0.00000185, 0.00000245, 0.00000612, -0.00000313, 0.00000148, -0.00000286, -0.00000564, -0.00000176, 0.00000732, 0.00000106, 0.00000024, 0.00000120, -0.00000638, 0.00000081, -0.00000419, -0.00000153, 0.00000774, -0.00000254, -0.00000031, -0.00000540, 0.00000563, 0.00000077, 0.00000186, 0.00000366, -0.00000860, -0.00000068, 0.00000256, -0.00000168, 0.00000121, -0.00000104, 0.00000028, -0.00000135, 0.00000127, 0.00000261, -0.00000882, 0.00000557, 0.00000229, 0.00000155, 0.00000062, -0.00000355, -0.00000131, -0.00000237, 0.00000371, -0.00000242, 0.00000141, -0.00000285, 0.00000259, 0.00000104, -0.00000095, -0.00000082, 0.00000288, 0.00000154, -0.00000716, 0.00000055, 0.00000172, 0.00000057, 0.00000043, 0.00000371, -0.00000772, 0.00000519, -0.00000798, 0.00000537, -0.00000077, 0.00000606, -0.00001109, -0.00000701, 0.00001069, -0.00000242, -0.00000018, -0.00000072, 0.00000725, -0.00000152, -0.00000429, -0.00000210, 0.00000521, -0.00000851, 0.00001707, -0.00001368, 0.00001261, 0.00001029, -0.00002686, 0.00001179, -0.00002147, 0.00001361, -0.00001891, 0.00001173, -0.00002638, 0.00000655, -0.00002056, -0.00001062, -0.00002145, -0.00000988, -0.00012532, -0.00000844, 0.00034042, 0.00015268, -0.00015870, -0.00003650, -0.00002222, -0.00001588, -0.00002901, 0.00000575, -0.00000569, 0.00002941, 0.00000300, 0.00003230, 0.00000647, -0.00000082, -0.00000526, -0.00002232, 0.00000147, -0.00002552, 0.00002038, -0.00000565, -0.00000627, -0.00000768, 0.00000377, -0.00000287, -0.00000772, -0.00000889, 0.00001553, -0.00000301, -0.00000611, 0.00000243, 0.00000582, },
		{ 0.00000620, 0.00000300, -0.00001020, -0.00000024, 0.00000126, 0.00000358, 0.00000547, -0.00000329, 0.00000163, -0.00000356, -0.00000578, -0.00000025, 0.00000666, 0.00000095, 0.00000042, 0.00000063, -0.00000642, 0.00000131, -0.00000498, -0.00000008, 0.00000687, -0.00000242, -0.00000089, -0.00000465, 0.00000585, 0.00000003, 0.00000285, 0.00000240, -0.00000857, 0.00000037, 0.00000214, -0.00000185, 0.00000145, -0.00000130, 0.00000049, -0.00000161, 0.00000191, 0.00000137, -0.00000842, 0.00000686, 0.00000130, 0.00000222, -0.00000037, -0.00000327, -0.00000153, -0.00000169, 0.00000327, -0.00000229, 0.00000124, -0.00000284, 0.00000299, 0.00000053, -0.00000050, -0.00000134, 0.00000414, -0.00000043, -0.00000608, 0.00000057, 0.00000200, -0.00000017, 0.00000167, 0.00000209, -0.00000682, 0.00000475, -0.00000798, 0.00000612, -0.00000149, 0.00000675, -0.00001318, -0.00000399, 0.00000912, -0.00000240, -0.00000058, 0.00000073, 0.00000670, -0.00000232, -0.00000399, -0.00000160, 0.00000469, -0.00000801, 0.00001855, -0.00001750, 0.00001839, 0.00000290, -0.00002114, 0.00000798, -0.00001927, 0.00001239, -0.00001776, 0.00001067, -0.00002637, 0.00000681, -0.00002104, -0.00001113, -0.00002086, -0.00001345, -0.00013422, 0.00003691, 0.00032334, 0.00012597, -0.00015473, -0.00003019, -0.00002378, -0.00001502, -0.00002987, 0.00000681, -0.00000460, 0.00003187, 0.00000122, 0.00003730, 0.00000170, 0.00000318, -0.00001294, -0.00001656, -0.00000256, -0.00002272, 0.00001985, -0.00000642, -0.00000655, -0.00000687, 0.00000400, -0.00000405, -0.00000833, -0.00000651, 0.00001529, -0.00000499, -0.00000502, 0.00000293, },
		{ 0.00000370, 0.00000617, 0.00000199, -0.00001004, 0.00000038, 0.00000070, 0.00000482, 0.00000455, -0.00000321, 0.00000162, -0.00000428, -0.00000568, 0.00000116, 0.00000583, 0.00000105, 0.00000042, 0.00000012, -0.00000639, 0.00000173, -0.00000569, 0.00000140, 0.00000582, -0.00000214, -0.00000159, -0.00000364, 0.00000568, -0.00000037, 0.00000362, 0.00000114, -0.00000835, 0.00000135, 0.00000159, -0.00000188, 0.00000162, -0.00000155, 0.00000069, -0.00000186, 0.00000254, -0.00000003, -0.00000758, 0.00000776, 0.00000047, 0.00000279, -0.00000137, -0.00000291, -0.00000175, -0.00000101, 0.00000277, -0.00000206, 0.00000094, -0.00000264, 0.00000320, 0.00000020, -0.00000026, -0.00000162, 0.00000516, -0.00000240, -0.00000471, 0.00000031, 0.00000243, -0.00000100, 0.00000306, 0.00000011, -0.00000538, 0.00000370, -0.00000730, 0.00000623, -0.00000161, 0.00000681, -0.00001466, -0.00000127, 0.00000747, -0.00000227, -0.00000090, 0.00000223, 0.00000582, -0.00000285, -0.00000380, -0.00000098, 0.00000387, -0.00000686, 0.00001907, -0.00002044, 0.00002359, -0.00000453, -0.00001458, 0.00000280, -0.00001551, 0.00000970, -0.00001521, 0.00000811, -0.00002475, 0.00000540, -0.00001972, -0.00001383, -0.00001735, -0.00002185, -0.00013520, 0.00007775, 0.00030200, 0.00010413, -0.00015145, -0.00002368, -0.00002592, -0.00001378, -0.00003100, 0.00000802, -0.00000319, 0.00003378, 0.00000013, 0.00004172, -0.00000269, 0.00000650, -0.00002048, -0.00001002, -0.00000774, -0.00001865, 0.00001812, -0.00000626, -0.00000755, -0.00000528, 0.00000343, -0.00000479, -0.00000908, -0.00000367, 0.00001414, -0.00000627, -0.00000421, },
		{ -0.00000367, 0.00000473, 0.00000574, 0.00000122, -0.00000990, 0.00000095, 0.00000019, 0.00000611, 0.00000340, -0.00000291, 0.00000144, -0.00000498, -0.00000533, 0.00000237, 0.00000495, 0.00000129, 0.00000029, -0.00000035, -0.00000626, 0.00000200, -0.00000624, 0.00000282, 0.00000468, -0.00000176, -0.00000235, -0.00000245, 0.00000517, -0.00000042, 0.00000412, -0.00000007, -0.00000795, 0.00000223, 0.00000092, -0.00000177, 0.00000170, -0.00000175, 0.00000088, -0.00000208, 0.00000312, -0.00000153, -0.00000635, 0.00000826, -0.00000013, 0.00000318, -0.00000230, -0.00000252, -0.00000192, -0.00000038, 0.00000223, -0.00000177, 0.00000054, -0.00000228, 0.00000323, 0.00000004, -0.00000023, -0.00000159, 0.00000583, -0.00000420, -0.00000321, -0.00000014, 0.00000291, -0.00000182, 0.00000446, -0.00000207, -0.00000353, 0.00000212, -0.00000602, 0.00000570, -0.00000114, 0.00000621, -0.00001542, 0.00000099, 0.00000588, -0.00000214, -0.00000106, 0.00000368, 0.00000469, -0.00000313, -0.00000370, -0.00000024, 0.00000277, -0.00000505, 0.00001850, -0.00002220, 0.00002777, -0.00001143, -0.00000782, -0.00000318, -0.00001065, 0.00000591, -0.00001159, 0.00000439, -0.00002186, 0.00000270, -0.00001704, -0.00001822, -0.00001152, -0.00003461, -0.00012791, 0.00011215, 0.00027871, 0.00008590, -0.00014837, -0.00001739, -0.00002828, -0.00001246, -0.00003217, 0.00000920, -0.00000130, 0.00003495, 0.00000001, 0.00004525, -0.00000639, 0.00000866, -0.00002720, -0.00000338, -0.00001352, -0.00001371, 0.00001548, -0.00000540, -0.00000906, -0.00000309, 0.00000218, -0.00000515, -0.00000982, -0.00000057, 0.00001224, -0.00000688, },
		{ -0.00000690, -0.00000336, 0.00000595, 0.00000495, 0.00000072, -0.00000981, 0.00000147, -0.00000022, 0.00000738, 0.00000210, -0.00000240, 0.00000107, -0.00000564, -0.00000474, 0.00000333, 0.00000410, 0.00000159, 0.00000006, -0.00000080, -0.00000602, 0.00000210, -0.00000658, 0.00000408, 0.00000353, -0.00000137, -0.00000307, -0.00000118, 0.00000440, -0.00000016, 0.00000433, -0.00000117, -0.00000741, 0.00000300, 0.00000017, -0.00000150, 0.00000166, -0.00000187, 0.00000100, -0.00000223, 0.00000357, -0.00000303, -0.00000478, 0.00000834, -0.00000048, 0.00000335, -0.00000309, -0.00000214, -0.00000202, 0.00000016, 0.00000170, -0.00000146, 0.00000010, -0.00000180, 0.00000308, 0.00000005, -0.00000041, -0.00000123, 0.00000607, -0.00000569, -0.00000174, -0.00000064, 0.00000334, -0.00000250, 0.00000571, -0.00000426, -0.00000145, 0.00000019, -0.00000425, 0.00000465, -0.00000012, 0.00000496, -0.00001544, 0.00000268, 0.00000449, -0.00000208, -0.00000098, 0.00000498, 0.00000338, -0.00000317, -0.00000370, 0.00000057, 0.00000147, -0.00000265, 0.00001683, -0.00002261, 0.00003056, -0.00001726, -0.00000149, -0.00000935, -0.00000521, 0.00000147, -0.00000734, -0.00000004, -0.00001816, -0.00000079, -0.00001356, -0.00002357, -0.00000426, -0.00005071, -0.00011264, 0.00013884, 0.00025543, 0.00007009, -0.00014494, -0.00001174, -0.00003046, -0.00001135, -0.00003314, 0.00001021, 0.00000119, 0.00003526, 0.00000102, 0.00004766, -0.00000918, 0.00000933, -0.00003255, 0.00000272, -0.00001931, -0.00000839, 0.00001229, -0.00000412, -0.00001083, -0.00000053, 0.00000042, -0.00000526, -0.00001041, 0.00000256, 0.00000979, },
		{ 0.00000705, -0.00000650, -0.00000318, 0.00000727, 0.00000386, 0.00000045, -0.00000975, 0.00000192, -0.00000048, 0.00000855, 0.00000069, -0.00000173, 0.00000052, -0.00000620, -0.00000393, 0.00000399, 0.00000335, 0.00000190, -0.00000021, -0.00000124, -0.00000565, 0.00000200, -0.00000664, 0.00000510, 0.00000248, -0.00000104, -0.00000368, 0.00000006, 0.00000349, 0.00000034, 0.00000427, -0.00000214, -0.00000675, 0.00000363, -0.00000062, -0.00000110, 0.00000150, -0.00000188, 0.00000104, -0.00000227, 0.00000385, -0.00000442, -0.00000299, 0.00000806, -0.00000058, 0.00000326, -0.00000371, -0.00000182, -0.00000203, 0.00000059, 0.00000124, -0.00000119, -0.00000033, -0.00000124, 0.00000280, 0.00000020, -0.00000077, -0.00000055, 0.00000583, -0.00000676, -0.00000043, -0.00000109, 0.00000365, -0.00000294, 0.00000666, -0.00000627, 0.00000063, -0.00000189, -0.00000217, 0.00000320, 0.00000132, 0.00000317, -0.00001475, 0.00000375, 0.00000336, -0.00000215, -0.00000061, 0.00000604, 0.00000199, -0.00000304, -0.00000374, 0.00000139, 0.00000007, 0.00000020, 0.00001412, -0.00002160, 0.00003176, -0.00002161, 0.00000385, -0.00001511, 0.00000024, -0.00000312, -0.00000294, -0.00000470, -0.00001418, -0.00000453, -0.00000996, -0.00002906, 0.00000334, -0.00006877, -0.00009033, 0.00015724, 0.00023369, 0.00005566, -0.00014071, -0.00000712, -0.00003214, -0.00001070, -0.00003374, 0.00001094, 0.00000432, 0.00003465, 0.00000325, 0.00004884, -0.00001097, 0.00000832, -0.00003607, 0.00000771, -0.00002456, -0.00000315, 0.00000894, -0.00000272, -0.00001259, 0.00000212, -0.00000166, -0.00000523, -0.00001069, 0.00000545, },
		{ 0.00000787, 0.00000429, -0.00000585, -0.00000301, 0.00000855, 0.00000258, 0.00000036, -0.00000972, 0.00000230, -0.00000055, 0.00000954, -0.00000073, -0.00000094, -0.00000020, -0.00000663, -0.00000296, 0.00000436, 0.00000276, 0.00000215, -0.00000050, -0.00000168, -0.00000516, 0.00000170, -0.00000640, 0.00000583, 0.00000158, -0.00000083, -0.00000408, 0.00000116, 0.00000253, 0.00000102, 0.00000396, -0.00000295, -0.00000600, 0.00000411, -0.00000142, -0.00000059, 0.00000121, -0.00000178, 0.00000098, -0.00000219, 0.00000392, -0.00000563, -0.00000108, 0.00000748, -0.00000043, 0.00000292, -0.00000412, -0.00000157, -0.00000194, 0.00000088, 0.00000087, -0.00000098, -0.00000071, -0.00000066, 0.00000243, 0.00000044, -0.00000126, 0.00000039, 0.00000513, -0.00000734, 0.00000061, -0.00000140, 0.00000374, -0.00000305, 0.00000719, -0.00000791, 0.00000253, -0.00000393, 0.00000003, 0.00000154, 0.00000302, 0.00000097, -0.00001344, 0.00000421, 0.00000256, -0.00000237, 0.00000008, 0.00000678, 0.00000063, -0.00000279, -0.00000381, 0.00000217, -0.00000130, 0.00000332, 0.00001056, -0.00001926, 0.00003127, -0.00002421, 0.00000776, -0.00001992, 0.00000521, -0.00000738, 0.00000115, -0.00000908, -0.00001042, -0.00000795, -0.00000687, -0.00003385, 0.00001010, -0.00008716, -0.00006244, 0.00016744, 0.00021452, 0.00004182, -0.00013534, -0.00000379, -0.00003310, -0.00001068, -0.00003386, 0.00001138, 0.00000808, 0.00003318, 0.00000668, 0.00004876, -0.00001182, 0.00000565, -0.00003755, 0.00001118, -0.00002882, 0.00000158, 0.00000576, -0.00000150, -0.00001406, 0.00000462, -0.00000385, -0.00000519, -0.00001050, },
		{ -0.00000976, 0.00000962, 0.00000173, -0.00000514, -0.00000270, 0.00000965, 0.00000124, 0.00000040, -0.00000969, 0.00000257, -0.00000041, 0.00001030, -0.00000209, -0.00000010, -0.00000109, -0.00000689, -0.00000188, 0.00000444, 0.00000235, 0.00000230, -0.00000078, -0.00000212, -0.00000458, 0.00000122, -0.00000587, 0.00000623, 0.00000088, -0.00000079, -0.00000424, 0.00000204, 0.00000164, 0.00000178, 0.00000346, -0.00000359, -0.00000517, 0.00000440, -0.00000218, -0.00000001, 0.00000082, -0.00000157, 0.00000080, -0.00000198, 0.00000373, -0.00000655, 0.00000084, 0.00000669, -0.00000011, 0.00000235, -0.00000431, -0.00000142, -0.00000173, 0.00000101, 0.00000062, -0.00000088, -0.00000099, -0.00000012, 0.00000204, 0.00000071, -0.00000179, 0.00000152, 0.00000401, -0.00000742, 0.00000131, -0.00000151, 0.00000358, -0.00000279, 0.00000720, -0.00000905, 0.00000405, -0.00000573, 0.00000213, -0.00000015, 0.00000480, -0.00000146, -0.00001166, 0.00000412, 0.00000208, -0.00000276, 0.00000108, 0.00000714, -0.00000064, -0.00000248, -0.00000383, 0.00000281, -0.00000249, 0.00000647, 0.00000639, -0.00001578, 0.00002919, -0.00002497, 0.00000995, -0.00002339, 0.00000926, -0.00001093, 0.00000454, -0.00001279, -0.00000732, -0.00001060, -0.00000487, -0.00003719, 0.00001491, -0.00010416, -0.00003083, 0.00017011, 0.00019842, 0.00002808, -0.00012867, -0.00000189, -0.00003319, -0.00001138, -0.00003346, 0.00001157, 0.00001233, 0.00003099, 0.00001117, 0.00004754, -0.00001192, 0.00000149, -0.00003695, 0.00001289, -0.00003176, 0.00000547, 0.00000307, -0.00000068, -0.00001501, 0.00000673, -0.00000598, -0.00000523, },
		{ -0.00000540, -0.00000844, 0.00001056, -0.00000040, -0.00000452, -0.00000215, 0.00001044, -0.00000004, 0.00000049, -0.00000963, 0.00000275, -0.00000003, 0.00001076, -0.00000332, 0.00000075, -0.00000210, -0.00000695, -0.00000078, 0.00000429, 0.00000213, 0.00000236, -0.00000105, -0.00000254, -0.00000393, 0.00000060, -0.00000508, 0.00000631, 0.00000041, -0.00000094, -0.00000412, 0.00000263, 0.00000091, 0.00000252, 0.00000282, -0.00000408, -0.00000427, 0.00000450, -0.00000284, 0.00000061, 0.00000034, -0.00000126, 0.00000052, -0.00000163, 0.00000328, -0.00000712, 0.00000264, 0.00000577, 0.00000032, 0.00000161, -0.00000431, -0.00000135, -0.00000143, 0.00000101, 0.00000049, -0.00000090, -0.00000113, 0.00000033, 0.00000167, 0.00000095, -0.00000230, 0.00000273, 0.00000257, -0.00000702, 0.00000165, -0.00000138, 0.00000316, -0.00000215, 0.00000666, -0.00000959, 0.00000506, -0.00000710, 0.00000396, -0.00000168, 0.00000647, -0.00000391, -0.00000957, 0.00000358, 0.00000188, -0.00000326, 0.00000236, 0.00000709, -0.00000173, -0.00000218, -0.00000378, 0.00000325, -0.00000336, 0.00000940, 0.00000192, -0.00001145, 0.00002575, -0.00002398, 0.00001034, -0.00002527, 0.00001210, -0.00001348, 0.00000694, -0.00001552, -0.00000518, -0.00001213, -0.00000434, -0.00003853, 0.00001684, -0.00011813, 0.00000249, 0.00016638, 0.00018548, 0.00001420, -0.00012072, -0.00000140, -0.00003242, -0.00001278, -0.00003260, 0.00001162, 0.00001690, 0.00002830, 0.00001652, 0.00004535, -0.00001156, -0.00000378, -0.00003444, 0.00001279, -0.00003322, 0.00000830, 0.00000103, -0.00000043, -0.00001531, 0.00000828, -0.00000791, },
		{ -0.00000955, -0.00000569, -0.00000658, 0.00001065, -0.00000197, -0.00000413, -0.00000126, 0.00001080, -0.00000114, 0.00000055, -0.00000950, 0.00000281, 0.00000059, 0.00001089, -0.00000435, 0.00000154, -0.00000321, -0.00000678, 0.00000029, 0.00000396, 0.00000207, 0.00000230, -0.00000131, -0.00000290, -0.00000328, -0.00000009, -0.00000410, 0.00000610, 0.00000016, -0.00000127, -0.00000373, 0.00000291, 0.00000040, 0.00000316, 0.00000211, -0.00000444, -0.00000332, 0.00000439, -0.00000337, 0.00000120, -0.00000018, -0.00000088, 0.00000015, -0.00000118, 0.00000258, -0.00000729, 0.00000422, 0.00000483, 0.00000079, 0.00000075, -0.00000415, -0.00000136, -0.00000107, 0.00000088, 0.00000048, -0.00000105, -0.00000111, 0.00000065, 0.00000138, 0.00000111, -0.00000269, 0.00000390, 0.00000092, -0.00000622, 0.00000166, -0.00000103, 0.00000251, -0.00000118, 0.00000559, -0.00000949, 0.00000548, -0.00000793, 0.00000537, -0.00000289, 0.00000786, -0.00000621, -0.00000737, 0.00000272, 0.00000190, -0.00000384, 0.00000389, 0.00000664, -0.00000259, -0.00000192, -0.00000361, 0.00000343, -0.00000379, 0.00001188, -0.00000248, -0.00000667, 0.00002129, -0.00002149, 0.00000899, -0.00002550, 0.00001358, -0.00001488, 0.00000822, -0.00001713, -0.00000417, -0.00001239, -0.00000548, -0.00003755, 0.00001527, -0.00012767, 0.00003544, 0.00015769, 0.00017540, 0.00000023, -0.00011166, -0.00000217, -0.00003088, -0.00001478, -0.00003137, 0.00001172, 0.00002156, 0.00002539, 0.00002247, 0.00004242, -0.00001110, -0.00000969, -0.00003037, 0.00001104, -0.00003320, 0.00001000, -0.00000026, -0.00000080, -0.00001489, 0.00000917, },
		{ 0.00000939, -0.00001090, -0.00000603, -0.00000429, 0.00000993, -0.00000293, -0.00000404, 0.00000002, 0.00001065, -0.00000195, 0.00000052, -0.00000929, 0.00000277, 0.00000145, 0.00001066, -0.00000513, 0.00000221, -0.00000435, -0.00000639, 0.00000126, 0.00000351, 0.00000214, 0.00000216, -0.00000157, -0.00000318, -0.00000269, -0.00000078, -0.00000299, 0.00000565, 0.00000009, -0.00000174, -0.00000311, 0.00000287, 0.00000017, 0.00000364, 0.00000139, -0.00000468, -0.00000233, 0.00000408, -0.00000373, 0.00000173, -0.00000070, -0.00000045, -0.00000027, -0.00000065, 0.00000166, -0.00000706, 0.00000552, 0.00000396, 0.00000122, -0.00000014, -0.00000387, -0.00000140, -0.00000066, 0.00000065, 0.00000057, -0.00000132, -0.00000093, 0.00000082, 0.00000120, 0.00000113, -0.00000289, 0.00000491, -0.00000080, -0.00000513, 0.00000138, -0.00000051, 0.00000170, 0.00000006, 0.00000404, -0.00000879, 0.00000526, -0.00000814, 0.00000625, -0.00000367, 0.00000884, -0.00000818, -0.00000524, 0.00000170, 0.00000205, -0.00000440, 0.00000557, 0.00000580, -0.00000319, -0.00000176, -0.00000330, 0.00000333, -0.00000367, 0.00001367, -0.00000650, -0.00000184, 0.00001627, -0.00001787, 0.00000616, -0.00002416, 0.00001372, -0.00001511, 0.00000837, -0.00001760, -0.00000429, -0.00001141, -0.00000826, -0.00003423, 0.00000992, -0.00013171, 0.00006613, 0.00014563, 0.00016761, -0.00001362, -0.00010181, -0.00000397, -0.00002875, -0.00001725, -0.00002992, 0.00001204, 0.00002607, 0.00002255, 0.00002872, 0.00003897, -0.00001089, -0.00001570, -0.00002524, 0.00000793, -0.00003185, 0.00001064, -0.00000082, -0.00000176, -0.00001376, },
		{ -0.00001202, 0.00000898, -0.00001197, -0.00000633, -0.00000174, 0.00000851, -0.00000328, -0.00000424, 0.00000167, 0.00000998, -0.00000243, 0.00000035, -0.00000897, 0.00000264, 0.00000251, 0.00001008, -0.00000562, 0.00000270, -0.00000548, -0.00000579, 0.00000208, 0.00000301, 0.00000230, 0.00000194, -0.00000184, -0.00000334, -0.00000220, -0.00000138, -0.00000185, 0.00000504, 0.00000015, -0.00000228, -0.00000230, 0.00000255, 0.00000021, 0.00000391, 0.00000071, -0.00000481, -0.00000133, 0.00000357, -0.00000389, 0.00000215, -0.00000118, -0.00000003, -0.00000072, -0.00000010, 0.00000059, -0.00000643, 0.00000648, 0.00000321, 0.00000152, -0.00000100, -0.00000352, -0.00000146, -0.00000027, 0.00000038, 0.00000073, -0.00000168, -0.00000059, 0.00000082, 0.00000115, 0.00000100, -0.00000284, 0.00000567, -0.00000244, -0.00000387, 0.00000092, 0.00000011, 0.00000081, 0.00000144, 0.00000215, -0.00000755, 0.00000446, -0.00000774, 0.00000657, -0.00000394, 0.00000930, -0.00000969, -0.00000335, 0.00000065, 0.00000224, -0.00000488, 0.00000731, 0.00000464, -0.00000353, -0.00000169, -0.00000286, 0.00000296, -0.00000299, 0.00001463, -0.00000983, 0.00000265, 0.00001113, -0.00001359, 0.00000221, -0.00002149, 0.00001265, -0.00001432, 0.00000752, -0.00001708, -0.00000538, -0.00000938, -0.00001241, -0.00002886, 0.00000091, -0.00012965, 0.00009293, 0.00013178, 0.00016139, -0.00002699, -0.00009154, -0.00000649, -0.00002625, -0.00001999, -0.00002840, 0.00001277, 0.00003019, 0.00002006, 0.00003499, 0.00003522, -0.00001126, -0.00002126, -0.00001958, 0.00000387, -0.00002944, 0.00001036, -0.00000078, -0.00000320, },
		{ -0.00000495, -0.00000984, 0.00000803, -0.00001283, -0.00000645, 0.00000085, 0.00000659, -0.00000312, -0.00000468, 0.00000364, 0.00000880, -0.00000254, 0.00000000, -0.00000855, 0.00000245, 0.00000373, 0.00000915, -0.00000580, 0.00000297, -0.00000654, -0.00000501, 0.00000271, 0.00000251, 0.00000250, 0.00000168, -0.00000213, -0.00000338, -0.00000187, -0.00000181, -0.00000078, 0.00000437, 0.00000027, -0.00000284, -0.00000141, 0.00000202, 0.00000050, 0.00000396, 0.00000010, -0.00000486, -0.00000033, 0.00000288, -0.00000383, 0.00000242, -0.00000157, 0.00000035, -0.00000114, 0.00000043, -0.00000057, -0.00000546, 0.00000710, 0.00000264, 0.00000166, -0.00000175, -0.00000315, -0.00000148, 0.00000009, 0.00000010, 0.00000090, -0.00000209, -0.00000012, 0.00000065, 0.00000123, 0.00000070, -0.00000252, 0.00000608, -0.00000387, -0.00000259, 0.00000037, 0.00000074, -0.00000005, 0.00000282, 0.00000006, -0.00000591, 0.00000316, -0.00000676, 0.00000632, -0.00000369, 0.00000922, -0.00001066, -0.00000182, -0.00000029, 0.00000238, -0.00000519, 0.00000901, 0.00000322, -0.00000363, -0.00000171, -0.00000231, 0.00000234, -0.00000174, 0.00001467, -0.00001225, 0.00000644, 0.00000635, -0.00000916, -0.00000239, -0.00001784, 0.00001063, -0.00001271, 0.00000591, -0.00001581, -0.00000716, -0.00000665, -0.00001751, -0.00002195, -0.00001126, -0.00012136, 0.00011463, 0.00011761, 0.00015599, -0.00003948, -0.00008126, -0.00000939, -0.00002363, -0.00002283, -0.00002695, 0.00001407, 0.00003370, 0.00001817, 0.00004105, 0.00003130, -0.00001242, -0.00002589, -0.00001396, -0.00000066, -0.00002631, 0.00000941, -0.00000032, },
		{ 0.00000035, -0.00000679, -0.00000740, 0.00000667, -0.00001355, -0.00000626, 0.00000324, 0.00000438, -0.00000259, -0.00000526, 0.00000582, 0.00000717, -0.00000229, -0.00000052, -0.00000804, 0.00000225, 0.00000504, 0.00000793, -0.00000567, 0.00000299, -0.00000747, -0.00000408, 0.00000314, 0.00000207, 0.00000272, 0.00000138, -0.00000245, -0.00000327, -0.00000172, -0.00000203, 0.00000013, 0.00000370, 0.00000037, -0.00000334, -0.00000051, 0.00000136, 0.00000100, 0.00000380, -0.00000042, -0.00000482, 0.00000062, 0.00000205, -0.00000357, 0.00000252, -0.00000183, 0.00000064, -0.00000150, 0.00000088, -0.00000174, -0.00000421, 0.00000737, 0.00000226, 0.00000159, -0.00000236, -0.00000282, -0.00000145, 0.00000036, -0.00000013, 0.00000105, -0.00000250, 0.00000043, 0.00000036, 0.00000142, 0.00000026, -0.00000192, 0.00000610, -0.00000499, -0.00000141, -0.00000017, 0.00000130, -0.00000077, 0.00000407, -0.00000206, -0.00000404, 0.00000150, -0.00000531, 0.00000559, -0.00000296, 0.00000860, -0.00001106, -0.00000074, -0.00000102, 0.00000238, -0.00000523, 0.00001055, 0.00000163, -0.00000352, -0.00000181, -0.00000171, 0.00000157, -0.00000001, 0.00001378, -0.00001361, 0.00000926, 0.00000231, -0.00000507, -0.00000717, -0.00001361, 0.00000797, -0.00001059, 0.00000383, -0.00001411, -0.00000930, -0.00000364, -0.00002300, -0.00001424, -0.00002576, -0.00010722, 0.00013047, 0.00010432, 0.00015070, -0.00005070, -0.00007137, -0.00001238, -0.00002109, -0.00002560, -0.00002569, 0.00001606, 0.00003648, 0.00001708, 0.00004669, 0.00002726, -0.00001448, -0.00002924, -0.00000887, -0.00000521, -0.00002284, 0.00000807, },
		{ 0.00000661, 0.00000101, -0.00000851, -0.00000490, 0.00000504, -0.00001422, -0.00000564, 0.00000521, 0.00000215, -0.00000188, -0.00000583, 0.00000805, 0.00000522, -0.00000172, -0.00000119, -0.00000746, 0.00000208, 0.00000637, 0.00000647, -0.00000525, 0.00000273, -0.00000821, -0.00000308, 0.00000339, 0.00000169, 0.00000292, 0.00000105, -0.00000277, -0.00000304, -0.00000177, -0.00000200, 0.00000083, 0.00000312, 0.00000040, -0.00000370, 0.00000030, 0.00000067, 0.00000162, 0.00000346, -0.00000083, -0.00000470, 0.00000149, 0.00000111, -0.00000311, 0.00000244, -0.00000196, 0.00000082, -0.00000176, 0.00000118, -0.00000281, -0.00000277, 0.00000734, 0.00000207, 0.00000132, -0.00000277, -0.00000255, -0.00000134, 0.00000052, -0.00000029, 0.00000112, -0.00000286, 0.00000101, -0.00000003, 0.00000171, -0.00000029, -0.00000108, 0.00000572, -0.00000572, -0.00000042, -0.00000060, 0.00000169, -0.00000126, 0.00000505, -0.00000403, -0.00000211, -0.00000036, -0.00000353, 0.00000446, -0.00000184, 0.00000750, -0.00001091, -0.00000015, -0.00000147, 0.00000219, -0.00000496, 0.00001185, -0.00000003, -0.00000324, -0.00000194, -0.00000110, 0.00000075, 0.00000206, 0.00001205, -0.00001389, 0.00001097, -0.00000070, -0.00000173, -0.00001164, -0.00000922, 0.00000502, -0.00000826, 0.00000161, -0.00001231, -0.00001140, -0.00000083, -0.00002827, -0.00000657, -0.00004149, -0.00008800, 0.00014018, 0.00009283, 0.00014492, -0.00006032, -0.00006219, -0.00001518, -0.00001883, -0.00002819, -0.00002465, 0.00001879, 0.00003844, 0.00001692, 0.00005179, 0.00002307, -0.00001738, -0.00003108, -0.00000472, -0.00000936, -0.00001939, },
		{ -0.00001624, 0.00000527, 0.00000147, -0.00000994, -0.00000253, 0.00000327, -0.00001489, -0.00000451, 0.00000657, 0.00000013, -0.00000116, -0.00000625, 0.00001017, 0.00000307, -0.00000091, -0.00000199, -0.00000684, 0.00000198, 0.00000765, 0.00000485, -0.00000458, 0.00000221, -0.00000872, -0.00000205, 0.00000348, 0.00000140, 0.00000310, 0.00000070, -0.00000309, -0.00000271, -0.00000198, -0.00000171, 0.00000127, 0.00000269, 0.00000031, -0.00000387, 0.00000094, 0.00000004, 0.00000228, 0.00000298, -0.00000115, -0.00000449, 0.00000226, 0.00000013, -0.00000250, 0.00000219, -0.00000192, 0.00000086, -0.00000190, 0.00000131, -0.00000372, -0.00000125, 0.00000708, 0.00000202, 0.00000085, -0.00000298, -0.00000237, -0.00000113, 0.00000056, -0.00000034, 0.00000109, -0.00000311, 0.00000156, -0.00000047, 0.00000203, -0.00000089, -0.00000006, 0.00000498, -0.00000604, 0.00000029, -0.00000085, 0.00000188, -0.00000144, 0.00000564, -0.00000571, -0.00000031, -0.00000223, -0.00000158, 0.00000309, -0.00000043, 0.00000604, -0.00001029, -0.00000003, -0.00000160, 0.00000178, -0.00000433, 0.00001281, -0.00000166, -0.00000288, -0.00000206, -0.00000057, -0.00000001, 0.00000430, 0.00000964, -0.00001317, 0.00001152, -0.00000249, 0.00000054, -0.00001543, -0.00000504, 0.00000210, -0.00000601, -0.00000045, -0.00001073, -0.00001313, 0.00000138, -0.00003272, 0.00000021, -0.00005723, -0.00006486, 0.00014392, 0.00008369, 0.00013825, -0.00006814, -0.00005396, -0.00001762, -0.00001696, -0.00003054, -0.00002385, 0.00002225, 0.00003956, 0.00001775, 0.00005626, 0.00001863, -0.00002093, -0.00003139, -0.00000175, -0.00001280, },
		{ -0.00001535, -0.00001361, 0.00000421, 0.00000158, -0.00001093, -0.00000043, 0.00000146, -0.00001556, -0.00000286, 0.00000720, -0.00000148, -0.00000063, -0.00000635, 0.00001201, 0.00000089, 0.00000004, -0.00000285, -0.00000621, 0.00000202, 0.00000878, 0.00000314, -0.00000371, 0.00000143, -0.00000896, -0.00000107, 0.00000344, 0.00000118, 0.00000326, 0.00000031, -0.00000335, -0.00000233, -0.00000232, -0.00000118, 0.00000145, 0.00000242, 0.00000007, -0.00000383, 0.00000135, -0.00000044, 0.00000291, 0.00000243, -0.00000140, -0.00000418, 0.00000286, -0.00000083, -0.00000178, 0.00000179, -0.00000174, 0.00000076, -0.00000190, 0.00000123, -0.00000438, 0.00000026, 0.00000666, 0.00000208, 0.00000023, -0.00000300, -0.00000229, -0.00000084, 0.00000046, -0.00000027, 0.00000092, -0.00000321, 0.00000203, -0.00000090, 0.00000234, -0.00000147, 0.00000105, 0.00000392, -0.00000595, 0.00000070, -0.00000089, 0.00000182, -0.00000129, 0.00000578, -0.00000695, 0.00000121, -0.00000395, 0.00000037, 0.00000161, 0.00000111, 0.00000435, -0.00000931, -0.00000032, -0.00000143, 0.00000115, -0.00000334, 0.00001337, -0.00000317, -0.00000247, -0.00000213, -0.00000017, -0.00000056, 0.00000650, 0.00000676, -0.00001160, 0.00001101, -0.00000304, 0.00000155, -0.00001822, -0.00000140, -0.00000053, -0.00000409, -0.00000212, -0.00000964, -0.00001421, 0.00000262, -0.00003586, 0.00000530, -0.00007172, -0.00003916, 0.00014228, 0.00007716, 0.00013046, -0.00007407, -0.00004682, -0.00001957, -0.00001555, -0.00003264, -0.00002322, 0.00002636, 0.00003990, 0.00001958, 0.00006006, 0.00001379, -0.00002483, -0.00003030, -0.00000005, },
		{ 0.00000045, -0.00001693, -0.00001160, 0.00000354, 0.00000130, -0.00001141, 0.00000130, -0.00000033, -0.00001622, -0.00000073, 0.00000705, -0.00000255, -0.00000040, -0.00000601, 0.00001340, -0.00000117, 0.00000103, -0.00000373, -0.00000562, 0.00000223, 0.00000970, 0.00000146, -0.00000271, 0.00000043, -0.00000892, -0.00000019, 0.00000333, 0.00000102, 0.00000340, -0.00000014, -0.00000353, -0.00000196, -0.00000271, -0.00000047, 0.00000137, 0.00000233, -0.00000032, -0.00000357, 0.00000150, -0.00000072, 0.00000342, 0.00000186, -0.00000161, -0.00000377, 0.00000328, -0.00000173, -0.00000100, 0.00000128, -0.00000143, 0.00000052, -0.00000176, 0.00000094, -0.00000475, 0.00000167, 0.00000615, 0.00000216, -0.00000048, -0.00000285, -0.00000228, -0.00000047, 0.00000024, -0.00000007, 0.00000060, -0.00000314, 0.00000236, -0.00000127, 0.00000259, -0.00000197, 0.00000217, 0.00000266, -0.00000551, 0.00000081, -0.00000071, 0.00000152, -0.00000081, 0.00000543, -0.00000769, 0.00000232, -0.00000537, 0.00000217, 0.00000019, 0.00000264, 0.00000256, -0.00000810, -0.00000093, -0.00000100, 0.00000032, -0.00000199, 0.00001350, -0.00000451, -0.00000208, -0.00000211, 0.00000004, -0.00000079, 0.00000845, 0.00000369, -0.00000944, 0.00000964, -0.00000246, 0.00000127, -0.00001984, 0.00000148, -0.00000266, -0.00000265, -0.00000322, -0.00000918, -0.00001447, 0.00000269, -0.00003736, 0.00000810, -0.00008379, -0.00001239, 0.00013609, 0.00007319, 0.00012149, -0.00007812, -0.00004081, -0.00002100, -0.00001460, -0.00003451, -0.00002267, 0.00003102, 0.00003956, 0.00002239, 0.00006314, 0.00000835, -0.00002866, -0.00002809, },
		{ -0.00002516, -0.00000004, -0.00001764, -0.00001021, 0.00000325, 0.00000061, -0.00001137, 0.00000261, -0.00000208, -0.00001679, 0.00000176, 0.00000616, -0.00000301, -0.00000056, -0.00000516, 0.00001423, -0.00000296, 0.00000196, -0.00000456, -0.00000508, 0.00000264, 0.00001032, -0.00000012, -0.00000167, -0.00000072, -0.00000861, 0.00000055, 0.00000318, 0.00000089, 0.00000353, -0.00000066, -0.00000359, -0.00000166, -0.00000310, 0.00000034, 0.00000109, 0.00000239, -0.00000083, -0.00000310, 0.00000140, -0.00000074, 0.00000375, 0.00000134, -0.00000179, -0.00000324, 0.00000348, -0.00000249, -0.00000022, 0.00000071, -0.00000103, 0.00000017, -0.00000151, 0.00000045, -0.00000479, 0.00000290, 0.00000564, 0.00000222, -0.00000122, -0.00000258, -0.00000233, -0.00000006, -0.00000007, 0.00000023, 0.00000014, -0.00000287, 0.00000253, -0.00000155, 0.00000274, -0.00000231, 0.00000319, 0.00000129, -0.00000480, 0.00000067, -0.00000033, 0.00000102, -0.00000003, 0.00000462, -0.00000788, 0.00000295, -0.00000637, 0.00000369, -0.00000106, 0.00000404, 0.00000083, -0.00000680, -0.00000174, -0.00000037, -0.00000063, -0.00000034, 0.00001318, -0.00000561, -0.00000174, -0.00000197, 0.00000001, -0.00000062, 0.00000996, 0.00000069, -0.00000697, 0.00000768, -0.00000096, -0.00000022, -0.00002024, 0.00000349, -0.00000421, -0.00000176, -0.00000371, -0.00000943, -0.00001385, 0.00000152, -0.00003709, 0.00000822, -0.00009249, 0.00001398, 0.00012642, 0.00007149, 0.00011147, -0.00008043, -0.00003588, -0.00002193, -0.00001406, -0.00003624, -0.00002208, 0.00003609, 0.00003866, 0.00002611, 0.00006546, 0.00000220, -0.00003202, },
		{ -0.00003449, -0.00002195, -0.00000122, -0.00001762, -0.00000936, 0.00000329, -0.00000041, -0.00001087, 0.00000353, -0.00000382, -0.00001715, 0.00000444, 0.00000463, -0.00000288, -0.00000113, -0.00000376, 0.00001442, -0.00000438, 0.00000273, -0.00000530, -0.00000463, 0.00000328, 0.00001060, -0.00000151, -0.00000066, -0.00000196, -0.00000806, 0.00000112, 0.00000304, 0.00000078, 0.00000367, -0.00000126, -0.00000348, -0.00000151, -0.00000338, 0.00000118, 0.00000068, 0.00000255, -0.00000142, -0.00000249, 0.00000107, -0.00000051, 0.00000387, 0.00000091, -0.00000197, -0.00000261, 0.00000345, -0.00000306, 0.00000049, 0.00000012, -0.00000057, -0.00000026, -0.00000117, -0.00000020, -0.00000452, 0.00000389, 0.00000518, 0.00000218, -0.00000193, -0.00000224, -0.00000241, 0.00000037, -0.00000043, 0.00000060, -0.00000043, -0.00000241, 0.00000250, -0.00000170, 0.00000275, -0.00000244, 0.00000402, -0.00000005, -0.00000392, 0.00000033, 0.00000018, 0.00000038, 0.00000097, 0.00000340, -0.00000755, 0.00000306, -0.00000688, 0.00000484, -0.00000202, 0.00000518, -0.00000072, -0.00000556, -0.00000263, 0.00000035, -0.00000164, 0.00000154, 0.00001244, -0.00000644, -0.00000148, -0.00000171, -0.00000025, 0.00000002, 0.00001086, -0.00000197, -0.00000450, 0.00000545, 0.00000115, -0.00000270, -0.00001950, 0.00000462, -0.00000516, -0.00000141, -0.00000362, -0.00001032, -0.00001245, -0.00000077, -0.00003509, 0.00000554, -0.00009711, 0.00003861, 0.00011440, 0.00007161, 0.00010062, -0.00008123, -0.00003191, -0.00002244, -0.00001387, -0.00003791, -0.00002130, 0.00004140, 0.00003736, 0.00003069, 0.00006693, -0.00000475, },
		{ -0.00001246, -0.00003575, -0.00001892, -0.00000275, -0.00001713, -0.00000887, 0.00000354, -0.00000163, -0.00001000, 0.00000409, -0.00000558, -0.00001718, 0.00000709, 0.00000266, -0.00000226, -0.00000205, -0.00000187, 0.00001394, -0.00000536, 0.00000330, -0.00000590, -0.00000426, 0.00000414, 0.00001050, -0.00000263, 0.00000024, -0.00000320, -0.00000733, 0.00000152, 0.00000293, 0.00000066, 0.00000382, -0.00000195, -0.00000319, -0.00000154, -0.00000348, 0.00000194, 0.00000021, 0.00000275, -0.00000203, -0.00000180, 0.00000057, -0.00000006, 0.00000376, 0.00000060, -0.00000217, -0.00000188, 0.00000318, -0.00000342, 0.00000108, -0.00000042, -0.00000011, -0.00000073, -0.00000080, -0.00000096, -0.00000394, 0.00000463, 0.00000482, 0.00000200, -0.00000252, -0.00000189, -0.00000247, 0.00000076, -0.00000081, 0.00000101, -0.00000109, -0.00000179, 0.00000228, -0.00000171, 0.00000262, -0.00000234, 0.00000458, -0.00000127, -0.00000299, -0.00000011, 0.00000076, -0.00000032, 0.00000208, 0.00000187, -0.00000676, 0.00000268, -0.00000690, 0.00000556, -0.00000265, 0.00000601, -0.00000200, -0.00000451, -0.00000347, 0.00000107, -0.00000262, 0.00000355, 0.00001133, -0.00000701, -0.00000131, -0.00000136, -0.00000072, 0.00000113, 0.00001106, -0.00000409, -0.00000230, 0.00000329, 0.00000351, -0.00000588, -0.00001779, 0.00000495, -0.00000559, -0.00000148, -0.00000308, -0.00001171, -0.00001045, -0.00000395, -0.00003166, 0.00000023, -0.00009730, 0.00006041, 0.00010114, 0.00007303, 0.00008924, -0.00008076, -0.00002877, -0.00002262, -0.00001393, -0.00003960, -0.00002019, 0.00004680, 0.00003580, 0.00003608, 0.00006740, },
		{ 0.00006667, -0.00002075, -0.00003561, -0.00001643, -0.00000427, -0.00001643, -0.00000858, 0.00000383, -0.00000292, -0.00000889, 0.00000437, -0.00000742, -0.00001675, 0.00000948, 0.00000046, -0.00000129, -0.00000321, 0.00000042, 0.00001284, -0.00000585, 0.00000360, -0.00000635, -0.00000398, 0.00000520, 0.00001000, -0.00000343, 0.00000094, -0.00000437, -0.00000646, 0.00000178, 0.00000285, 0.00000054, 0.00000398, -0.00000271, -0.00000271, -0.00000179, -0.00000334, 0.00000253, -0.00000022, 0.00000292, -0.00000258, -0.00000112, -0.00000002, 0.00000056, 0.00000345, 0.00000041, -0.00000237, -0.00000108, 0.00000269, -0.00000354, 0.00000152, -0.00000087, 0.00000031, -0.00000119, -0.00000044, -0.00000177, -0.00000311, 0.00000510, 0.00000457, 0.00000165, -0.00000296, -0.00000157, -0.00000249, 0.00000108, -0.00000115, 0.00000140, -0.00000178, -0.00000103, 0.00000187, -0.00000158, 0.00000235, -0.00000199, 0.00000483, -0.00000225, -0.00000213, -0.00000055, 0.00000133, -0.00000099, 0.00000320, 0.00000017, -0.00000562, 0.00000187, -0.00000644, 0.00000586, -0.00000291, 0.00000648, -0.00000295, -0.00000372, -0.00000415, 0.00000169, -0.00000345, 0.00000559, 0.00000992, -0.00000734, -0.00000121, -0.00000094, -0.00000133, 0.00000266, 0.00001051, -0.00000552, -0.00000060, 0.00000150, 0.00000575, -0.00000941, -0.00001536, 0.00000464, -0.00000563, -0.00000183, -0.00000226, -0.00001339, -0.00000811, -0.00000766, -0.00002722, -0.00000728, -0.00009305, 0.00007854, 0.00008764, 0.00007519, 0.00007766, -0.00007932, -0.00002629, -0.00002259, -0.00001417, -0.00004141, -0.00001863, 0.00005215, 0.00003412, 0.00004221, },
		{ 0.00004906, 0.00006449, -0.00002933, -0.00003404, -0.00001476, -0.00000548, -0.00001576, -0.00000829, 0.00000403, -0.00000412, -0.00000767, 0.00000445, -0.00000937, -0.00001575, 0.00001140, -0.00000170, -0.00000017, -0.00000444, 0.00000296, 0.00001121, -0.00000588, 0.00000363, -0.00000665, -0.00000374, 0.00000642, 0.00000912, -0.00000388, 0.00000139, -0.00000538, -0.00000556, 0.00000191, 0.00000280, 0.00000042, 0.00000411, -0.00000352, -0.00000208, -0.00000226, -0.00000293, 0.00000290, -0.00000055, 0.00000301, -0.00000302, -0.00000052, -0.00000061, 0.00000128, 0.00000296, 0.00000034, -0.00000258, -0.00000024, 0.00000201, -0.00000342, 0.00000176, -0.00000119, 0.00000063, -0.00000161, -0.00000015, -0.00000255, -0.00000210, 0.00000534, 0.00000443, 0.00000113, -0.00000320, -0.00000132, -0.00000243, 0.00000130, -0.00000140, 0.00000172, -0.00000244, -0.00000020, 0.00000132, -0.00000133, 0.00000197, -0.00000142, 0.00000474, -0.00000293, -0.00000141, -0.00000092, 0.00000181, -0.00000152, 0.00000418, -0.00000157, -0.00000426, 0.00000074, -0.00000557, 0.00000576, -0.00000282, 0.00000660, -0.00000355, -0.00000327, -0.00000459, 0.00000214, -0.00000406, 0.00000755, 0.00000830, -0.00000746, -0.00000115, -0.00000051, -0.00000201, 0.00000451, 0.00000926, -0.00000618, 0.00000045, 0.00000033, 0.00000755, -0.00001293, -0.00001247, 0.00000386, -0.00000543, -0.00000228, -0.00000137, -0.00001513, -0.00000574, -0.00001151, -0.00002232, -0.00001634, -0.00008467, 0.00009253, 0.00007474, 0.00007758, 0.00006619, -0.00007715, -0.00002432, -0.00002244, -0.00001451, -0.00004338, -0.00001653, 0.00005735, 0.00003246, },
		{ 0.00003094, 0.00005654, 0.00006058, -0.00003777, -0.00003116, -0.00001407, -0.00000617, -0.00001533, -0.00000786, 0.00000401, -0.00000510, -0.00000646, 0.00000439, -0.00001145, -0.00001413, 0.00001266, -0.00000359, 0.00000090, -0.00000557, 0.00000557, 0.00000917, -0.00000551, 0.00000342, -0.00000680, -0.00000353, 0.00000773, 0.00000791, -0.00000397, 0.00000156, -0.00000617, -0.00000468, 0.00000199, 0.00000274, 0.00000034, 0.00000418, -0.00000431, -0.00000132, -0.00000292, -0.00000223, 0.00000300, -0.00000072, 0.00000297, -0.00000329, -0.00000009, -0.00000109, 0.00000201, 0.00000237, 0.00000037, -0.00000276, 0.00000060, 0.00000118, -0.00000307, 0.00000180, -0.00000135, 0.00000083, -0.00000193, 0.00000004, -0.00000324, -0.00000098, 0.00000539, 0.00000438, 0.00000044, -0.00000323, -0.00000118, -0.00000227, 0.00000138, -0.00000154, 0.00000193, -0.00000301, 0.00000065, 0.00000067, -0.00000100, 0.00000151, -0.00000067, 0.00000433, -0.00000328, -0.00000091, -0.00000115, 0.00000213, -0.00000187, 0.00000494, -0.00000320, -0.00000281, -0.00000059, -0.00000439, 0.00000533, -0.00000243, 0.00000640, -0.00000382, -0.00000315, -0.00000472, 0.00000234, -0.00000437, 0.00000932, 0.00000656, -0.00000742, -0.00000109, -0.00000014, -0.00000263, 0.00000654, 0.00000739, -0.00000609, 0.00000078, -0.00000007, 0.00000864, -0.00001613, -0.00000942, 0.00000280, -0.00000517, -0.00000264, -0.00000061, -0.00001669, -0.00000364, -0.00001508, -0.00001754, -0.00002618, -0.00007275, 0.00010219, 0.00006307, 0.00007978, 0.00005510, -0.00007451, -0.00002271, -0.00002225, -0.00001491, -0.00004555, -0.00001380, 0.00006227, },
		{ 0.00006685, 0.00002970, 0.00006455, 0.00005467, -0.00004554, -0.00002728, -0.00001432, -0.00000625, -0.00001525, -0.00000718, 0.00000368, -0.00000579, -0.00000536, 0.00000422, -0.00001362, -0.00001187, 0.00001313, -0.00000501, 0.00000172, -0.00000641, 0.00000805, 0.00000689, -0.00000482, 0.00000299, -0.00000684, -0.00000328, 0.00000906, 0.00000641, -0.00000372, 0.00000143, -0.00000668, -0.00000392, 0.00000207, 0.00000265, 0.00000032, 0.00000416, -0.00000503, -0.00000052, -0.00000373, -0.00000128, 0.00000281, -0.00000071, 0.00000275, -0.00000335, 0.00000013, -0.00000141, 0.00000265, 0.00000174, 0.00000045, -0.00000290, 0.00000139, 0.00000026, -0.00000253, 0.00000166, -0.00000135, 0.00000089, -0.00000214, 0.00000010, -0.00000375, 0.00000016, 0.00000531, 0.00000436, -0.00000036, -0.00000306, -0.00000115, -0.00000200, 0.00000132, -0.00000154, 0.00000198, -0.00000344, 0.00000145, -0.00000005, -0.00000062, 0.00000104, 0.00000019, 0.00000365, -0.00000331, -0.00000067, -0.00000118, 0.00000226, -0.00000196, 0.00000539, -0.00000460, -0.00000144, -0.00000198, -0.00000302, 0.00000467, -0.00000182, 0.00000594, -0.00000382, -0.00000336, -0.00000455, 0.00000226, -0.00000433, 0.00001080, 0.00000480, -0.00000728, -0.00000099, 0.00000010, -0.00000308, 0.00000857, 0.00000508, -0.00000536, 0.00000042, 0.00000034, 0.00000884, -0.00001873, -0.00000644, 0.00000165, -0.00000496, -0.00000277, -0.00000016, -0.00001788, -0.00000203, -0.00001800, -0.00001346, -0.00003593, -0.00005811, 0.00010767, 0.00005303, 0.00008145, 0.00004460, -0.00007157, -0.00002136, -0.00002206, -0.00001535, -0.00004793, -0.00001041, },
		{ -0.00000630, 0.00007100, 0.00002888, 0.00007294, 0.00004657, -0.00005210, -0.00002282, -0.00001536, -0.00000573, -0.00001555, -0.00000623, 0.00000304, -0.00000614, -0.00000439, 0.00000393, -0.00001580, -0.00000907, 0.00001279, -0.00000583, 0.00000213, -0.00000680, 0.00001020, 0.00000456, -0.00000391, 0.00000241, -0.00000682, -0.00000294, 0.00001032, 0.00000473, -0.00000317, 0.00000101, -0.00000689, -0.00000332, 0.00000220, 0.00000249, 0.00000040, 0.00000399, -0.00000562, 0.00000025, -0.00000459, -0.00000013, 0.00000237, -0.00000049, 0.00000237, -0.00000321, 0.00000011, -0.00000148, 0.00000314, 0.00000114, 0.00000055, -0.00000296, 0.00000207, -0.00000069, -0.00000185, 0.00000134, -0.00000119, 0.00000079, -0.00000223, -0.00000001, -0.00000406, 0.00000124, 0.00000516, 0.00000433, -0.00000122, -0.00000272, -0.00000124, -0.00000164, 0.00000113, -0.00000138, 0.00000186, -0.00000368, 0.00000214, -0.00000076, -0.00000024, 0.00000061, 0.00000109, 0.00000277, -0.00000304, -0.00000070, -0.00000100, 0.00000218, -0.00000178, 0.00000547, -0.00000568, -0.00000024, -0.00000330, -0.00000158, 0.00000387, -0.00000106, 0.00000532, -0.00000362, -0.00000383, -0.00000407, 0.00000191, -0.00000390, 0.00001192, 0.00000311, -0.00000710, -0.00000080, 0.00000016, -0.00000325, 0.00001042, 0.00000253, -0.00000413, -0.00000049, 0.00000152, 0.00000808, -0.00002055, -0.00000374, 0.00000054, -0.00000490, -0.00000258, -0.00000011, -0.00001859, -0.00000110, -0.00002002, -0.00001054, -0.00004478, -0.00004168, 0.00010932, 0.00004484, 0.00008239, 0.00003484, -0.00006848, -0.00002022, -0.00002191, -0.00001582, -0.00005050, },
		{ -0.00005324, -0.00000145, 0.00007465, 0.00002868, 0.00008144, 0.00003618, -0.00005691, -0.00001829, -0.00001691, -0.00000474, -0.00001617, -0.00000505, 0.00000212, -0.00000619, -0.00000356, 0.00000348, -0.00001786, -0.00000585, 0.00001169, -0.00000599, 0.00000204, -0.00000661, 0.00001185, 0.00000233, -0.00000292, 0.00000175, -0.00000677, -0.00000246, 0.00001141, 0.00000295, -0.00000238, 0.00000034, -0.00000680, -0.00000294, 0.00000243, 0.00000221, 0.00000062, 0.00000362, -0.00000599, 0.00000089, -0.00000542, 0.00000113, 0.00000171, -0.00000011, 0.00000185, -0.00000289, -0.00000015, -0.00000128, 0.00000341, 0.00000063, 0.00000062, -0.00000292, 0.00000261, -0.00000160, -0.00000109, 0.00000090, -0.00000089, 0.00000056, -0.00000220, -0.00000027, -0.00000412, 0.00000220, 0.00000500, 0.00000423, -0.00000208, -0.00000226, -0.00000140, -0.00000120, 0.00000081, -0.00000108, 0.00000154, -0.00000370, 0.00000267, -0.00000142, 0.00000011, 0.00000028, 0.00000194, 0.00000178, -0.00000254, -0.00000095, -0.00000062, 0.00000190, -0.00000135, 0.00000517, -0.00000636, 0.00000067, -0.00000444, -0.00000017, 0.00000303, -0.00000025, 0.00000462, -0.00000332, -0.00000447, -0.00000335, 0.00000129, -0.00000309, 0.00001263, 0.00000157, -0.00000693, -0.00000051, -0.00000002, -0.00000303, 0.00001191, -0.00000004, -0.00000262, -0.00000178, 0.00000332, 0.00000641, -0.00002152, -0.00000143, -0.00000047, -0.00000502, -0.00000204, -0.00000051, -0.00001877, -0.00000088, -0.00002101, -0.00000911, -0.00005201, -0.00002444, 0.00010768, 0.00003852, 0.00008249, 0.00002590, -0.00006534, -0.00001922, -0.00002180, -0.00001634, },
		{ -0.00001692, -0.00005613, 0.00000416, 0.00007771, 0.00002929, 0.00008968, 0.00002354, -0.00005951, -0.00001422, -0.00001861, -0.00000350, -0.00001699, -0.00000374, 0.00000102, -0.00000599, -0.00000282, 0.00000281, -0.00001963, -0.00000244, 0.00000996, -0.00000552, 0.00000144, -0.00000580, 0.00001288, 0.00000036, -0.00000195, 0.00000107, -0.00000673, -0.00000179, 0.00001226, 0.00000117, -0.00000146, -0.00000052, -0.00000644, -0.00000278, 0.00000280, 0.00000179, 0.00000099, 0.00000302, -0.00000609, 0.00000131, -0.00000610, 0.00000240, 0.00000092, 0.00000038, 0.00000122, -0.00000242, -0.00000062, -0.00000081, 0.00000342, 0.00000027, 0.00000063, -0.00000275, 0.00000296, -0.00000241, -0.00000031, 0.00000039, -0.00000050, 0.00000020, -0.00000207, -0.00000067, -0.00000392, 0.00000299, 0.00000488, 0.00000399, -0.00000286, -0.00000172, -0.00000163, -0.00000072, 0.00000040, -0.00000066, 0.00000106, -0.00000349, 0.00000299, -0.00000199, 0.00000038, 0.00000009, 0.00000265, 0.00000080, -0.00000191, -0.00000137, -0.00000008, 0.00000147, -0.00000071, 0.00000451, -0.00000664, 0.00000124, -0.00000531, 0.00000110, 0.00000222, 0.00000053, 0.00000393, -0.00000303, -0.00000519, -0.00000246, 0.00000048, -0.00000194, 0.00001292, 0.00000022, -0.00000680, -0.00000009, -0.00000044, -0.00000236, 0.00001289, -0.00000240, -0.00000104, -0.00000321, 0.00000550, 0.00000395, -0.00002162, 0.00000041, -0.00000134, -0.00000529, -0.00000120, -0.00000132, -0.00001847, -0.00000132, -0.00002098, -0.00000932, -0.00005707, -0.00000731, 0.00010342, 0.00003396, 0.00008171, 0.00001781, -0.00006220, -0.00001835, -0.00002174, },
		},
		{ { 0.00014965, 0.00005446, 0.00002221, -0.00000702, -0.00006808, -0.00001646, -0.00001343, -0.00000998, -0.00001096, -0.00000012, -0.00000078, -0.00000847, 0.00001079, -0.00000498, -0.00000161, 0.00000133, -0.00000650, 0.00000149, 0.00000299, 0.00000025, -0.00000965, 0.00000240, 0.00000080, 0.00000365, -0.00000768, -0.00000051, 0.00000012, -0.00000017, 0.00000309, -0.00000020, -0.00000400, 0.00000346, -0.00000650, 0.00000293, 0.00000226, 0.00000443, -0.00000708, 0.00000229, -0.00000433, 0.00000541, -0.00000576, 0.00000315, -0.00000134, 0.00000317, -0.00000300, 0.00000250, -0.00000528, 0.00000627, -0.00000556, 0.00000368, -0.00000344, 0.00000435, -0.00000714, 0.00000519, -0.00000443, 0.00000659, -0.00000584, 0.00000430, -0.00000452, 0.00000568, -0.00000457, 0.00000578, -0.00000544, 0.00000419, -0.00000756, 0.00000419, -0.00000544, 0.00000578, -0.00000457, 0.00000568, -0.00000452, 0.00000430, -0.00000584, 0.00000659, -0.00000443, 0.00000519, -0.00000714, 0.00000435, -0.00000344, 0.00000368, -0.00000556, 0.00000627, -0.00000528, 0.00000250, -0.00000300, 0.00000317, -0.00000134, 0.00000315, -0.00000576, 0.00000541, -0.00000433, 0.00000229, -0.00000708, 0.00000443, 0.00000226, 0.00000293, -0.00000650, 0.00000346, -0.00000400, -0.00000020, 0.00000309, -0.00000017, 0.00000012, -0.00000051, -0.00000768, 0.00000365, 0.00000080, 0.00000240, -0.00000965, 0.00000025, 0.00000299, 0.00000149, -0.00000650, 0.00000133, -0.00000161, -0.00000498, 0.00001079, -0.00000847, -0.00000078, -0.00000012, -0.00001096, -0.00000998, -0.00001343, -0.00001646, -0.00006808, -0.00000702, 0.00002221, 0.00005446, },
		{ 0.00005183, 0.00014982, 0.00005705, 0.00002965, -0.00001384, -0.00006858, -0.00001578, -0.00001437, -0.00000767, -0.00001318, 0.00000103, -0.00000098, -0.00000816, 0.00001100, -0.00000625, -0.00000075, 0.00000169, -0.00000698, 0.00000156, 0.00000342, -0.00000032, -0.00000977, 0.00000269, 0.00000046, 0.00000422, -0.00000816, -0.00000021, 0.00000007, 0.00000016, 0.00000290, -0.00000054, -0.00000409, 0.00000380, -0.00000669, 0.00000306, 0.00000244, 0.00000417, -0.00000713, 0.00000241, -0.00000450, 0.00000578, -0.00000600, 0.00000338, -0.00000165, 0.00000340, -0.00000322, 0.00000247, -0.00000526, 0.00000640, -0.00000554, 0.00000366, -0.00000339, 0.00000429, -0.00000718, 0.00000534, -0.00000462, 0.00000672, -0.00000602, 0.00000439, -0.00000457, 0.00000575, -0.00000451, 0.00000571, -0.00000540, 0.00000413, -0.00000755, 0.00000422, -0.00000547, 0.00000584, -0.00000465, 0.00000562, -0.00000449, 0.00000422, -0.00000566, 0.00000645, -0.00000423, 0.00000502, -0.00000706, 0.00000440, -0.00000349, 0.00000370, -0.00000559, 0.00000614, -0.00000532, 0.00000257, -0.00000282, 0.00000297, -0.00000104, 0.00000292, -0.00000550, 0.00000501, -0.00000414, 0.00000214, -0.00000701, 0.00000468, 0.00000205, 0.00000284, -0.00000632, 0.00000312, -0.00000392, 0.00000016, 0.00000323, -0.00000047, 0.00000013, -0.00000078, -0.00000720, 0.00000308, 0.00000114, 0.00000210, -0.00000948, 0.00000078, 0.00000255, 0.00000142, -0.00000598, 0.00000089, -0.00000244, -0.00000362, 0.00001042, -0.00000861, -0.00000074, -0.00000113, -0.00000881, -0.00001232, -0.00001235, -0.00001742, -0.00006699, -0.00000087, 0.00001490, },
		{ 0.00000776, 0.00004913, 0.00015033, 0.00005964, 0.00003715, -0.00002133, -0.00006843, -0.00001545, -0.00001511, -0.00000544, -0.00001542, 0.00000228, -0.00000133, -0.00000771, 0.00001105, -0.00000742, 0.00000013, 0.00000198, -0.00000742, 0.00000163, 0.00000383, -0.00000092, -0.00000983, 0.00000297, 0.00000012, 0.00000478, -0.00000864, 0.00000010, 0.00000001, 0.00000049, 0.00000267, -0.00000085, -0.00000416, 0.00000414, -0.00000690, 0.00000323, 0.00000257, 0.00000390, -0.00000716, 0.00000252, -0.00000464, 0.00000612, -0.00000622, 0.00000360, -0.00000197, 0.00000364, -0.00000347, 0.00000249, -0.00000526, 0.00000654, -0.00000553, 0.00000365, -0.00000334, 0.00000419, -0.00000719, 0.00000547, -0.00000479, 0.00000685, -0.00000621, 0.00000451, -0.00000463, 0.00000584, -0.00000446, 0.00000565, -0.00000536, 0.00000407, -0.00000752, 0.00000425, -0.00000549, 0.00000591, -0.00000473, 0.00000558, -0.00000449, 0.00000417, -0.00000550, 0.00000631, -0.00000403, 0.00000483, -0.00000696, 0.00000442, -0.00000353, 0.00000373, -0.00000562, 0.00000602, -0.00000537, 0.00000268, -0.00000268, 0.00000280, -0.00000077, 0.00000269, -0.00000523, 0.00000458, -0.00000392, 0.00000198, -0.00000690, 0.00000492, 0.00000180, 0.00000280, -0.00000615, 0.00000278, -0.00000384, 0.00000056, 0.00000333, -0.00000076, 0.00000012, -0.00000102, -0.00000673, 0.00000252, 0.00000147, 0.00000180, -0.00000928, 0.00000128, 0.00000209, 0.00000138, -0.00000543, 0.00000039, -0.00000324, -0.00000218, 0.00000988, -0.00000859, -0.00000088, -0.00000196, -0.00000678, -0.00001462, -0.00001119, -0.00001858, -0.00006537, 0.00000461, },
		{ 0.00000944, 0.00000085, 0.00004631, 0.00015118, 0.00006229, 0.00004462, -0.00002944, -0.00006762, -0.00001551, -0.00001558, -0.00000337, -0.00001761, 0.00000360, -0.00000179, -0.00000712, 0.00001094, -0.00000848, 0.00000102, 0.00000219, -0.00000781, 0.00000169, 0.00000423, -0.00000157, -0.00000981, 0.00000322, -0.00000021, 0.00000533, -0.00000910, 0.00000042, -0.00000007, 0.00000083, 0.00000242, -0.00000113, -0.00000422, 0.00000446, -0.00000710, 0.00000343, 0.00000267, 0.00000364, -0.00000716, 0.00000261, -0.00000474, 0.00000641, -0.00000641, 0.00000381, -0.00000230, 0.00000389, -0.00000375, 0.00000253, -0.00000527, 0.00000669, -0.00000553, 0.00000366, -0.00000328, 0.00000408, -0.00000717, 0.00000557, -0.00000494, 0.00000696, -0.00000639, 0.00000464, -0.00000471, 0.00000595, -0.00000443, 0.00000560, -0.00000531, 0.00000399, -0.00000748, 0.00000425, -0.00000550, 0.00000597, -0.00000483, 0.00000556, -0.00000450, 0.00000414, -0.00000535, 0.00000618, -0.00000382, 0.00000463, -0.00000684, 0.00000441, -0.00000356, 0.00000376, -0.00000566, 0.00000590, -0.00000543, 0.00000283, -0.00000258, 0.00000267, -0.00000053, 0.00000248, -0.00000497, 0.00000413, -0.00000368, 0.00000180, -0.00000678, 0.00000516, 0.00000152, 0.00000279, -0.00000602, 0.00000245, -0.00000376, 0.00000099, 0.00000338, -0.00000100, 0.00000007, -0.00000123, -0.00000628, 0.00000198, 0.00000177, 0.00000153, -0.00000905, 0.00000176, 0.00000160, 0.00000136, -0.00000488, -0.00000018, -0.00000400, -0.00000066, 0.00000918, -0.00000838, -0.00000122, -0.00000257, -0.00000493, -0.00001682, -0.00001001, -0.00001988, -0.00006329, },
		{ -0.00006082, 0.00001364, -0.00000583, 0.00004336, 0.00015234, 0.00006505, 0.00005199, -0.00003813, -0.00006611, -0.00001600, -0.00001573, -0.00000151, -0.00001969, 0.00000493, -0.00000234, -0.00000642, 0.00001069, -0.00000943, 0.00000190, 0.00000233, -0.00000813, 0.00000174, 0.00000463, -0.00000227, -0.00000972, 0.00000343, -0.00000051, 0.00000585, -0.00000954, 0.00000074, -0.00000016, 0.00000117, 0.00000214, -0.00000140, -0.00000427, 0.00000475, -0.00000730, 0.00000366, 0.00000273, 0.00000337, -0.00000713, 0.00000269, -0.00000482, 0.00000667, -0.00000657, 0.00000399, -0.00000261, 0.00000414, -0.00000404, 0.00000261, -0.00000530, 0.00000685, -0.00000554, 0.00000367, -0.00000323, 0.00000395, -0.00000712, 0.00000564, -0.00000507, 0.00000705, -0.00000658, 0.00000478, -0.00000481, 0.00000608, -0.00000442, 0.00000555, -0.00000527, 0.00000390, -0.00000742, 0.00000425, -0.00000549, 0.00000602, -0.00000493, 0.00000556, -0.00000453, 0.00000415, -0.00000522, 0.00000606, -0.00000362, 0.00000442, -0.00000669, 0.00000438, -0.00000358, 0.00000378, -0.00000570, 0.00000578, -0.00000551, 0.00000302, -0.00000253, 0.00000258, -0.00000033, 0.00000229, -0.00000470, 0.00000367, -0.00000342, 0.00000161, -0.00000662, 0.00000537, 0.00000122, 0.00000284, -0.00000591, 0.00000213, -0.00000369, 0.00000146, 0.00000336, -0.00000122, -0.00000002, -0.00000139, -0.00000586, 0.00000146, 0.00000204, 0.00000127, -0.00000880, 0.00000223, 0.00000108, 0.00000138, -0.00000432, -0.00000080, -0.00000470, 0.00000091, 0.00000832, -0.00000800, -0.00000176, -0.00000295, -0.00000331, -0.00001885, -0.00000889, -0.00002124, },
		{ -0.00002257, -0.00005806, 0.00001726, -0.00001225, 0.00004027, 0.00015381, 0.00006799, 0.00005915, -0.00004733, -0.00006390, -0.00001696, -0.00001554, 0.00000009, -0.00002162, 0.00000624, -0.00000295, -0.00000563, 0.00001031, -0.00001026, 0.00000278, 0.00000240, -0.00000839, 0.00000177, 0.00000501, -0.00000302, -0.00000955, 0.00000359, -0.00000079, 0.00000633, -0.00000995, 0.00000106, -0.00000025, 0.00000150, 0.00000184, -0.00000165, -0.00000430, 0.00000503, -0.00000748, 0.00000391, 0.00000274, 0.00000311, -0.00000709, 0.00000274, -0.00000487, 0.00000688, -0.00000668, 0.00000414, -0.00000292, 0.00000439, -0.00000434, 0.00000272, -0.00000534, 0.00000702, -0.00000557, 0.00000370, -0.00000319, 0.00000380, -0.00000705, 0.00000567, -0.00000517, 0.00000712, -0.00000674, 0.00000493, -0.00000491, 0.00000622, -0.00000442, 0.00000551, -0.00000523, 0.00000381, -0.00000736, 0.00000422, -0.00000547, 0.00000605, -0.00000503, 0.00000556, -0.00000457, 0.00000418, -0.00000512, 0.00000595, -0.00000343, 0.00000420, -0.00000653, 0.00000432, -0.00000359, 0.00000380, -0.00000574, 0.00000567, -0.00000559, 0.00000325, -0.00000253, 0.00000253, -0.00000017, 0.00000212, -0.00000445, 0.00000320, -0.00000315, 0.00000141, -0.00000644, 0.00000558, 0.00000088, 0.00000292, -0.00000583, 0.00000184, -0.00000362, 0.00000197, 0.00000328, -0.00000138, -0.00000015, -0.00000151, -0.00000547, 0.00000098, 0.00000227, 0.00000106, -0.00000855, 0.00000269, 0.00000053, 0.00000144, -0.00000378, -0.00000146, -0.00000534, 0.00000253, 0.00000732, -0.00000743, -0.00000251, -0.00000308, -0.00000195, -0.00002067, -0.00000789, },
		{ -0.00000707, -0.00002381, -0.00005509, 0.00002035, -0.00001841, 0.00003702, 0.00015554, 0.00007118, 0.00006598, -0.00005696, -0.00006102, -0.00001841, -0.00001495, 0.00000139, -0.00002334, 0.00000749, -0.00000359, -0.00000476, 0.00000980, -0.00001097, 0.00000363, 0.00000241, -0.00000857, 0.00000177, 0.00000539, -0.00000382, -0.00000928, 0.00000370, -0.00000103, 0.00000677, -0.00001031, 0.00000137, -0.00000035, 0.00000183, 0.00000153, -0.00000189, -0.00000432, 0.00000527, -0.00000765, 0.00000419, 0.00000271, 0.00000286, -0.00000702, 0.00000278, -0.00000488, 0.00000704, -0.00000676, 0.00000426, -0.00000320, 0.00000462, -0.00000465, 0.00000285, -0.00000539, 0.00000719, -0.00000561, 0.00000375, -0.00000317, 0.00000365, -0.00000694, 0.00000568, -0.00000525, 0.00000716, -0.00000690, 0.00000508, -0.00000502, 0.00000638, -0.00000445, 0.00000549, -0.00000520, 0.00000372, -0.00000728, 0.00000418, -0.00000543, 0.00000608, -0.00000513, 0.00000557, -0.00000464, 0.00000424, -0.00000504, 0.00000586, -0.00000325, 0.00000399, -0.00000635, 0.00000424, -0.00000359, 0.00000381, -0.00000578, 0.00000555, -0.00000567, 0.00000352, -0.00000258, 0.00000254, -0.00000006, 0.00000200, -0.00000423, 0.00000274, -0.00000288, 0.00000121, -0.00000624, 0.00000576, 0.00000052, 0.00000305, -0.00000580, 0.00000157, -0.00000357, 0.00000252, 0.00000313, -0.00000150, -0.00000033, -0.00000158, -0.00000513, 0.00000054, 0.00000245, 0.00000089, -0.00000831, 0.00000315, -0.00000007, 0.00000156, -0.00000327, -0.00000217, -0.00000593, 0.00000417, 0.00000618, -0.00000670, -0.00000346, -0.00000294, -0.00000089, -0.00002222, },
		{ -0.00002345, -0.00000648, -0.00002488, -0.00005200, 0.00002296, -0.00002431, 0.00003363, 0.00015753, 0.00007471, 0.00007236, -0.00006693, -0.00005748, -0.00002035, -0.00001397, 0.00000233, -0.00002480, 0.00000862, -0.00000424, -0.00000385, 0.00000917, -0.00001155, 0.00000445, 0.00000235, -0.00000868, 0.00000176, 0.00000576, -0.00000466, -0.00000892, 0.00000374, -0.00000122, 0.00000716, -0.00001064, 0.00000167, -0.00000045, 0.00000213, 0.00000122, -0.00000211, -0.00000431, 0.00000548, -0.00000779, 0.00000448, 0.00000264, 0.00000261, -0.00000694, 0.00000280, -0.00000488, 0.00000716, -0.00000680, 0.00000435, -0.00000347, 0.00000484, -0.00000496, 0.00000300, -0.00000546, 0.00000737, -0.00000566, 0.00000381, -0.00000316, 0.00000348, -0.00000681, 0.00000566, -0.00000529, 0.00000718, -0.00000703, 0.00000522, -0.00000514, 0.00000655, -0.00000450, 0.00000549, -0.00000517, 0.00000363, -0.00000719, 0.00000412, -0.00000537, 0.00000609, -0.00000523, 0.00000559, -0.00000471, 0.00000432, -0.00000498, 0.00000579, -0.00000309, 0.00000378, -0.00000617, 0.00000415, -0.00000357, 0.00000382, -0.00000582, 0.00000542, -0.00000575, 0.00000382, -0.00000267, 0.00000259, -0.00000001, 0.00000191, -0.00000403, 0.00000228, -0.00000260, 0.00000100, -0.00000602, 0.00000592, 0.00000015, 0.00000322, -0.00000579, 0.00000133, -0.00000353, 0.00000310, 0.00000291, -0.00000157, -0.00000056, -0.00000159, -0.00000484, 0.00000015, 0.00000259, 0.00000077, -0.00000809, 0.00000361, -0.00000071, 0.00000173, -0.00000280, -0.00000290, -0.00000644, 0.00000583, 0.00000491, -0.00000580, -0.00000461, -0.00000253, -0.00000015, },
		{ 0.00000025, -0.00002434, -0.00000616, -0.00002571, -0.00004888, 0.00002515, -0.00002995, 0.00003008, 0.00015975, 0.00007863, 0.00007815, -0.00007712, -0.00005335, -0.00002277, -0.00001256, 0.00000291, -0.00002597, 0.00000961, -0.00000485, -0.00000290, 0.00000845, -0.00001201, 0.00000523, 0.00000224, -0.00000873, 0.00000172, 0.00000612, -0.00000555, -0.00000846, 0.00000372, -0.00000137, 0.00000750, -0.00001091, 0.00000194, -0.00000054, 0.00000242, 0.00000090, -0.00000233, -0.00000428, 0.00000565, -0.00000789, 0.00000478, 0.00000254, 0.00000238, -0.00000684, 0.00000282, -0.00000484, 0.00000724, -0.00000679, 0.00000439, -0.00000370, 0.00000503, -0.00000525, 0.00000317, -0.00000553, 0.00000755, -0.00000572, 0.00000390, -0.00000317, 0.00000332, -0.00000666, 0.00000561, -0.00000531, 0.00000716, -0.00000714, 0.00000537, -0.00000526, 0.00000673, -0.00000457, 0.00000550, -0.00000516, 0.00000354, -0.00000709, 0.00000405, -0.00000530, 0.00000608, -0.00000532, 0.00000561, -0.00000480, 0.00000443, -0.00000496, 0.00000575, -0.00000295, 0.00000359, -0.00000598, 0.00000403, -0.00000354, 0.00000382, -0.00000584, 0.00000529, -0.00000583, 0.00000414, -0.00000282, 0.00000270, -0.00000001, 0.00000187, -0.00000387, 0.00000185, -0.00000232, 0.00000079, -0.00000578, 0.00000607, -0.00000023, 0.00000342, -0.00000583, 0.00000111, -0.00000351, 0.00000373, 0.00000262, -0.00000158, -0.00000085, -0.00000155, -0.00000459, -0.00000019, 0.00000268, 0.00000071, -0.00000789, 0.00000409, -0.00000140, 0.00000197, -0.00000238, -0.00000366, -0.00000688, 0.00000749, 0.00000354, -0.00000476, -0.00000595, -0.00000185, },
		{ -0.00000091, 0.00000030, -0.00002486, -0.00000616, -0.00002626, -0.00004582, 0.00002700, -0.00003536, 0.00002639, 0.00016216, 0.00008304, 0.00008323, -0.00008740, -0.00004868, -0.00002564, -0.00001075, 0.00000309, -0.00002681, 0.00001041, -0.00000541, -0.00000196, 0.00000765, -0.00001235, 0.00000595, 0.00000208, -0.00000870, 0.00000165, 0.00000646, -0.00000647, -0.00000791, 0.00000363, -0.00000146, 0.00000777, -0.00001112, 0.00000219, -0.00000063, 0.00000269, 0.00000057, -0.00000253, -0.00000423, 0.00000578, -0.00000796, 0.00000509, 0.00000239, 0.00000217, -0.00000673, 0.00000283, -0.00000479, 0.00000727, -0.00000674, 0.00000440, -0.00000390, 0.00000519, -0.00000554, 0.00000335, -0.00000560, 0.00000773, -0.00000580, 0.00000401, -0.00000320, 0.00000315, -0.00000650, 0.00000554, -0.00000529, 0.00000711, -0.00000722, 0.00000550, -0.00000537, 0.00000693, -0.00000465, 0.00000552, -0.00000516, 0.00000346, -0.00000699, 0.00000397, -0.00000521, 0.00000606, -0.00000540, 0.00000563, -0.00000490, 0.00000457, -0.00000496, 0.00000573, -0.00000283, 0.00000341, -0.00000579, 0.00000391, -0.00000350, 0.00000380, -0.00000586, 0.00000515, -0.00000590, 0.00000449, -0.00000300, 0.00000285, -0.00000006, 0.00000187, -0.00000375, 0.00000143, -0.00000205, 0.00000059, -0.00000552, 0.00000618, -0.00000063, 0.00000366, -0.00000590, 0.00000093, -0.00000349, 0.00000438, 0.00000226, -0.00000154, -0.00000118, -0.00000145, -0.00000439, -0.00000047, 0.00000271, 0.00000070, -0.00000772, 0.00000458, -0.00000213, 0.00000228, -0.00000202, -0.00000442, -0.00000723, 0.00000913, 0.00000208, -0.00000359, -0.00000745, },
		{ -0.00000909, 0.00000026, 0.00000001, -0.00002500, -0.00000650, -0.00002646, -0.00004288, 0.00002856, -0.00004055, 0.00002255, 0.00016474, 0.00008799, 0.00008744, -0.00009764, -0.00004355, -0.00002892, -0.00000855, 0.00000288, -0.00002730, 0.00001100, -0.00000589, -0.00000102, 0.00000678, -0.00001256, 0.00000660, 0.00000190, -0.00000861, 0.00000157, 0.00000677, -0.00000741, -0.00000727, 0.00000346, -0.00000150, 0.00000799, -0.00001128, 0.00000242, -0.00000072, 0.00000294, 0.00000025, -0.00000273, -0.00000416, 0.00000586, -0.00000799, 0.00000540, 0.00000220, 0.00000197, -0.00000661, 0.00000283, -0.00000473, 0.00000728, -0.00000665, 0.00000436, -0.00000407, 0.00000532, -0.00000580, 0.00000353, -0.00000568, 0.00000791, -0.00000589, 0.00000414, -0.00000326, 0.00000300, -0.00000632, 0.00000544, -0.00000525, 0.00000702, -0.00000727, 0.00000562, -0.00000548, 0.00000712, -0.00000475, 0.00000557, -0.00000518, 0.00000339, -0.00000690, 0.00000388, -0.00000511, 0.00000602, -0.00000547, 0.00000565, -0.00000499, 0.00000472, -0.00000499, 0.00000574, -0.00000275, 0.00000326, -0.00000561, 0.00000377, -0.00000344, 0.00000378, -0.00000587, 0.00000499, -0.00000596, 0.00000485, -0.00000322, 0.00000306, -0.00000018, 0.00000193, -0.00000367, 0.00000105, -0.00000180, 0.00000039, -0.00000524, 0.00000627, -0.00000102, 0.00000393, -0.00000601, 0.00000077, -0.00000349, 0.00000507, 0.00000183, -0.00000144, -0.00000157, -0.00000130, -0.00000425, -0.00000071, 0.00000268, 0.00000076, -0.00000759, 0.00000508, -0.00000291, 0.00000267, -0.00000174, -0.00000519, -0.00000751, 0.00001072, 0.00000055, -0.00000230, },
		{ -0.00000094, -0.00001084, 0.00000164, -0.00000062, -0.00002474, -0.00000720, -0.00002627, -0.00004015, 0.00002989, -0.00004553, 0.00001857, 0.00016749, 0.00009357, 0.00009066, -0.00010768, -0.00003807, -0.00003256, -0.00000598, 0.00000228, -0.00002741, 0.00001136, -0.00000628, -0.00000013, 0.00000586, -0.00001264, 0.00000718, 0.00000169, -0.00000847, 0.00000148, 0.00000703, -0.00000837, -0.00000655, 0.00000323, -0.00000149, 0.00000815, -0.00001137, 0.00000262, -0.00000080, 0.00000316, -0.00000007, -0.00000291, -0.00000407, 0.00000591, -0.00000798, 0.00000570, 0.00000198, 0.00000180, -0.00000650, 0.00000284, -0.00000466, 0.00000725, -0.00000652, 0.00000429, -0.00000421, 0.00000542, -0.00000604, 0.00000371, -0.00000575, 0.00000809, -0.00000599, 0.00000429, -0.00000335, 0.00000286, -0.00000613, 0.00000532, -0.00000519, 0.00000691, -0.00000729, 0.00000572, -0.00000558, 0.00000732, -0.00000487, 0.00000563, -0.00000522, 0.00000333, -0.00000680, 0.00000379, -0.00000499, 0.00000595, -0.00000552, 0.00000567, -0.00000509, 0.00000490, -0.00000504, 0.00000577, -0.00000270, 0.00000312, -0.00000544, 0.00000363, -0.00000338, 0.00000375, -0.00000587, 0.00000482, -0.00000599, 0.00000522, -0.00000348, 0.00000331, -0.00000035, 0.00000205, -0.00000365, 0.00000070, -0.00000156, 0.00000021, -0.00000495, 0.00000634, -0.00000141, 0.00000421, -0.00000614, 0.00000065, -0.00000349, 0.00000577, 0.00000134, -0.00000128, -0.00000200, -0.00000109, -0.00000417, -0.00000088, 0.00000261, 0.00000088, -0.00000749, 0.00000560, -0.00000373, 0.00000312, -0.00000155, -0.00000595, -0.00000769, 0.00001226, -0.00000102, },
		{ -0.00000261, 0.00000048, -0.00001267, 0.00000320, -0.00000156, -0.00002411, -0.00000825, -0.00002569, -0.00003768, 0.00003106, -0.00005033, 0.00001446, 0.00017037, 0.00009981, 0.00009274, -0.00011736, -0.00003233, -0.00003649, -0.00000309, 0.00000130, -0.00002716, 0.00001147, -0.00000654, 0.00000072, 0.00000490, -0.00001261, 0.00000766, 0.00000148, -0.00000828, 0.00000138, 0.00000726, -0.00000932, -0.00000576, 0.00000294, -0.00000143, 0.00000825, -0.00001141, 0.00000279, -0.00000087, 0.00000337, -0.00000039, -0.00000308, -0.00000396, 0.00000591, -0.00000791, 0.00000600, 0.00000171, 0.00000164, -0.00000638, 0.00000285, -0.00000458, 0.00000719, -0.00000636, 0.00000418, -0.00000431, 0.00000548, -0.00000624, 0.00000389, -0.00000582, 0.00000827, -0.00000609, 0.00000447, -0.00000348, 0.00000274, -0.00000595, 0.00000519, -0.00000510, 0.00000676, -0.00000728, 0.00000580, -0.00000567, 0.00000751, -0.00000500, 0.00000571, -0.00000527, 0.00000329, -0.00000672, 0.00000369, -0.00000487, 0.00000587, -0.00000556, 0.00000567, -0.00000519, 0.00000508, -0.00000512, 0.00000584, -0.00000268, 0.00000302, -0.00000528, 0.00000349, -0.00000331, 0.00000371, -0.00000586, 0.00000463, -0.00000600, 0.00000559, -0.00000376, 0.00000360, -0.00000058, 0.00000221, -0.00000368, 0.00000039, -0.00000135, 0.00000004, -0.00000466, 0.00000637, -0.00000178, 0.00000451, -0.00000631, 0.00000055, -0.00000349, 0.00000650, 0.00000078, -0.00000107, -0.00000248, -0.00000083, -0.00000413, -0.00000101, 0.00000248, 0.00000105, -0.00000743, 0.00000614, -0.00000458, 0.00000365, -0.00000144, -0.00000669, -0.00000778, 0.00001372, },
		{ 0.00001508, -0.00000420, 0.00000193, -0.00001454, 0.00000490, -0.00000278, -0.00002312, -0.00000966, -0.00002468, -0.00003553, 0.00003212, -0.00005495, 0.00001021, 0.00017340, 0.00010679, 0.00009356, -0.00012654, -0.00002645, -0.00004064, 0.00000007, -0.00000003, -0.00002653, 0.00001132, -0.00000668, 0.00000150, 0.00000391, -0.00001245, 0.00000804, 0.00000128, -0.00000805, 0.00000129, 0.00000742, -0.00001026, -0.00000490, 0.00000259, -0.00000132, 0.00000829, -0.00001138, 0.00000294, -0.00000095, 0.00000356, -0.00000071, -0.00000323, -0.00000385, 0.00000588, -0.00000780, 0.00000629, 0.00000142, 0.00000152, -0.00000626, 0.00000288, -0.00000451, 0.00000712, -0.00000618, 0.00000404, -0.00000438, 0.00000550, -0.00000642, 0.00000406, -0.00000588, 0.00000844, -0.00000620, 0.00000467, -0.00000364, 0.00000264, -0.00000576, 0.00000505, -0.00000499, 0.00000659, -0.00000724, 0.00000586, -0.00000574, 0.00000770, -0.00000514, 0.00000581, -0.00000535, 0.00000327, -0.00000664, 0.00000360, -0.00000473, 0.00000577, -0.00000558, 0.00000566, -0.00000527, 0.00000528, -0.00000522, 0.00000592, -0.00000269, 0.00000295, -0.00000514, 0.00000335, -0.00000323, 0.00000365, -0.00000583, 0.00000442, -0.00000598, 0.00000595, -0.00000406, 0.00000392, -0.00000085, 0.00000243, -0.00000377, 0.00000013, -0.00000116, -0.00000012, -0.00000435, 0.00000637, -0.00000215, 0.00000482, -0.00000649, 0.00000048, -0.00000349, 0.00000723, 0.00000017, -0.00000080, -0.00000300, -0.00000053, -0.00000414, -0.00000108, 0.00000232, 0.00000127, -0.00000740, 0.00000668, -0.00000547, 0.00000424, -0.00000142, -0.00000740, -0.00000778, },
		{ -0.00000768, 0.00001633, -0.00000575, 0.00000337, -0.00001641, 0.00000669, -0.00000424, -0.00002179, -0.00001141, -0.00002326, -0.00003372, 0.00003310, -0.00005941, 0.00000581, 0.00017657, 0.00011451, 0.00009298, -0.00013504, -0.00002057, -0.00004491, 0.00000343, -0.00000166, -0.00002555, 0.00001090, -0.00000667, 0.00000221, 0.00000291, -0.00001218, 0.00000831, 0.00000111, -0.00000781, 0.00000120, 0.00000750, -0.00001116, -0.00000400, 0.00000218, -0.00000118, 0.00000828, -0.00001130, 0.00000306, -0.00000103, 0.00000374, -0.00000104, -0.00000336, -0.00000373, 0.00000581, -0.00000765, 0.00000656, 0.00000108, 0.00000143, -0.00000616, 0.00000291, -0.00000444, 0.00000704, -0.00000598, 0.00000388, -0.00000443, 0.00000549, -0.00000656, 0.00000422, -0.00000594, 0.00000860, -0.00000632, 0.00000489, -0.00000384, 0.00000257, -0.00000559, 0.00000490, -0.00000486, 0.00000638, -0.00000716, 0.00000590, -0.00000579, 0.00000788, -0.00000529, 0.00000592, -0.00000544, 0.00000326, -0.00000658, 0.00000350, -0.00000459, 0.00000565, -0.00000558, 0.00000563, -0.00000535, 0.00000547, -0.00000533, 0.00000603, -0.00000273, 0.00000290, -0.00000502, 0.00000321, -0.00000316, 0.00000359, -0.00000579, 0.00000419, -0.00000592, 0.00000630, -0.00000437, 0.00000428, -0.00000118, 0.00000269, -0.00000391, -0.00000009, -0.00000100, -0.00000025, -0.00000404, 0.00000635, -0.00000248, 0.00000512, -0.00000670, 0.00000043, -0.00000348, 0.00000797, -0.00000049, -0.00000049, -0.00000356, -0.00000018, -0.00000419, -0.00000111, 0.00000211, 0.00000154, -0.00000740, 0.00000723, -0.00000638, 0.00000490, -0.00000150, -0.00000809, },
		{ -0.00000874, -0.00000747, 0.00001744, -0.00000724, 0.00000478, -0.00001823, 0.00000853, -0.00000590, -0.00002016, -0.00001345, -0.00002143, -0.00003229, 0.00003404, -0.00006371, 0.00000124, 0.00017990, 0.00012302, 0.00009091, -0.00014272, -0.00001481, -0.00004921, 0.00000692, -0.00000356, -0.00002423, 0.00001022, -0.00000653, 0.00000282, 0.00000190, -0.00001179, 0.00000845, 0.00000097, -0.00000755, 0.00000114, 0.00000750, -0.00001200, -0.00000306, 0.00000174, -0.00000100, 0.00000822, -0.00001117, 0.00000316, -0.00000112, 0.00000391, -0.00000137, -0.00000346, -0.00000361, 0.00000572, -0.00000744, 0.00000681, 0.00000072, 0.00000137, -0.00000606, 0.00000297, -0.00000440, 0.00000696, -0.00000577, 0.00000369, -0.00000445, 0.00000544, -0.00000667, 0.00000437, -0.00000598, 0.00000874, -0.00000644, 0.00000513, -0.00000407, 0.00000254, -0.00000544, 0.00000476, -0.00000473, 0.00000616, -0.00000705, 0.00000591, -0.00000582, 0.00000805, -0.00000544, 0.00000604, -0.00000556, 0.00000328, -0.00000653, 0.00000342, -0.00000444, 0.00000552, -0.00000556, 0.00000559, -0.00000540, 0.00000567, -0.00000545, 0.00000616, -0.00000281, 0.00000289, -0.00000493, 0.00000309, -0.00000308, 0.00000352, -0.00000574, 0.00000394, -0.00000583, 0.00000662, -0.00000469, 0.00000465, -0.00000154, 0.00000300, -0.00000410, -0.00000026, -0.00000087, -0.00000036, -0.00000373, 0.00000629, -0.00000279, 0.00000542, -0.00000691, 0.00000039, -0.00000346, 0.00000869, -0.00000119, -0.00000014, -0.00000414, 0.00000019, -0.00000428, -0.00000111, 0.00000188, 0.00000185, -0.00000743, 0.00000779, -0.00000730, 0.00000560, -0.00000168, },
		{ -0.00000194, -0.00000934, -0.00000717, 0.00001841, -0.00000864, 0.00000611, -0.00001997, 0.00001035, -0.00000770, -0.00001829, -0.00001575, -0.00001922, -0.00003123, 0.00003497, -0.00006784, -0.00000351, 0.00018342, 0.00013231, 0.00008724, -0.00014941, -0.00000930, -0.00005343, 0.00001045, -0.00000565, -0.00002261, 0.00000930, -0.00000625, 0.00000335, 0.00000091, -0.00001129, 0.00000845, 0.00000090, -0.00000729, 0.00000111, 0.00000739, -0.00001278, -0.00000211, 0.00000127, -0.00000080, 0.00000812, -0.00001098, 0.00000324, -0.00000123, 0.00000408, -0.00000171, -0.00000353, -0.00000350, 0.00000560, -0.00000720, 0.00000705, 0.00000032, 0.00000135, -0.00000598, 0.00000305, -0.00000437, 0.00000688, -0.00000555, 0.00000350, -0.00000445, 0.00000537, -0.00000675, 0.00000450, -0.00000601, 0.00000887, -0.00000655, 0.00000539, -0.00000435, 0.00000254, -0.00000531, 0.00000462, -0.00000459, 0.00000591, -0.00000691, 0.00000589, -0.00000582, 0.00000820, -0.00000559, 0.00000618, -0.00000569, 0.00000331, -0.00000650, 0.00000335, -0.00000429, 0.00000537, -0.00000553, 0.00000553, -0.00000544, 0.00000586, -0.00000558, 0.00000631, -0.00000291, 0.00000292, -0.00000487, 0.00000298, -0.00000300, 0.00000344, -0.00000568, 0.00000367, -0.00000569, 0.00000691, -0.00000499, 0.00000504, -0.00000193, 0.00000335, -0.00000435, -0.00000038, -0.00000077, -0.00000044, -0.00000342, 0.00000619, -0.00000307, 0.00000570, -0.00000713, 0.00000036, -0.00000342, 0.00000939, -0.00000192, 0.00000024, -0.00000474, 0.00000059, -0.00000441, -0.00000106, 0.00000163, 0.00000218, -0.00000748, 0.00000833, -0.00000822, 0.00000635, },
		{ 0.00000712, -0.00000230, -0.00000990, -0.00000676, 0.00001921, -0.00000994, 0.00000735, -0.00002157, 0.00001211, -0.00000959, -0.00001622, -0.00001825, -0.00001669, -0.00003056, 0.00003589, -0.00007179, -0.00000848, 0.00018717, 0.00014237, 0.00008189, -0.00015497, -0.00000419, -0.00005746, 0.00001394, -0.00000788, -0.00002074, 0.00000815, -0.00000584, 0.00000379, -0.00000008, -0.00001068, 0.00000832, 0.00000089, -0.00000705, 0.00000113, 0.00000717, -0.00001345, -0.00000117, 0.00000079, -0.00000058, 0.00000800, -0.00001076, 0.00000331, -0.00000135, 0.00000425, -0.00000207, -0.00000356, -0.00000341, 0.00000547, -0.00000691, 0.00000727, -0.00000011, 0.00000136, -0.00000593, 0.00000316, -0.00000438, 0.00000682, -0.00000535, 0.00000329, -0.00000444, 0.00000526, -0.00000679, 0.00000462, -0.00000602, 0.00000899, -0.00000666, 0.00000566, -0.00000466, 0.00000259, -0.00000520, 0.00000450, -0.00000446, 0.00000566, -0.00000674, 0.00000585, -0.00000580, 0.00000834, -0.00000574, 0.00000632, -0.00000585, 0.00000337, -0.00000649, 0.00000328, -0.00000415, 0.00000521, -0.00000547, 0.00000545, -0.00000546, 0.00000604, -0.00000571, 0.00000647, -0.00000304, 0.00000297, -0.00000483, 0.00000288, -0.00000293, 0.00000336, -0.00000560, 0.00000337, -0.00000550, 0.00000716, -0.00000529, 0.00000543, -0.00000234, 0.00000372, -0.00000464, -0.00000045, -0.00000071, -0.00000050, -0.00000310, 0.00000607, -0.00000330, 0.00000595, -0.00000734, 0.00000033, -0.00000335, 0.00001006, -0.00000267, 0.00000065, -0.00000536, 0.00000101, -0.00000455, -0.00000100, 0.00000138, 0.00000253, -0.00000753, 0.00000885, -0.00000913, },
		{ -0.00001000, 0.00000790, -0.00000274, -0.00001041, -0.00000624, 0.00001983, -0.00001110, 0.00000845, -0.00002300, 0.00001374, -0.00001149, -0.00001403, -0.00002089, -0.00001387, -0.00003023, 0.00003681, -0.00007555, -0.00001370, 0.00019122, 0.00015315, 0.00007482, -0.00015927, 0.00000041, -0.00006120, 0.00001729, -0.00001016, -0.00001865, 0.00000679, -0.00000531, 0.00000414, -0.00000105, -0.00000996, 0.00000804, 0.00000098, -0.00000684, 0.00000120, 0.00000681, -0.00001401, -0.00000026, 0.00000031, -0.00000037, 0.00000785, -0.00001051, 0.00000337, -0.00000149, 0.00000443, -0.00000245, -0.00000355, -0.00000335, 0.00000534, -0.00000659, 0.00000747, -0.00000058, 0.00000142, -0.00000589, 0.00000330, -0.00000442, 0.00000677, -0.00000515, 0.00000309, -0.00000443, 0.00000514, -0.00000680, 0.00000472, -0.00000602, 0.00000909, -0.00000677, 0.00000595, -0.00000502, 0.00000267, -0.00000513, 0.00000440, -0.00000433, 0.00000540, -0.00000655, 0.00000579, -0.00000576, 0.00000845, -0.00000588, 0.00000646, -0.00000601, 0.00000346, -0.00000650, 0.00000324, -0.00000401, 0.00000504, -0.00000540, 0.00000535, -0.00000545, 0.00000621, -0.00000584, 0.00000664, -0.00000320, 0.00000306, -0.00000483, 0.00000280, -0.00000287, 0.00000327, -0.00000552, 0.00000305, -0.00000527, 0.00000736, -0.00000555, 0.00000581, -0.00000277, 0.00000412, -0.00000498, -0.00000047, -0.00000068, -0.00000054, -0.00000279, 0.00000591, -0.00000350, 0.00000617, -0.00000753, 0.00000031, -0.00000324, 0.00001069, -0.00000342, 0.00000107, -0.00000598, 0.00000143, -0.00000471, -0.00000092, 0.00000112, 0.00000287, -0.00000758, 0.00000934, },
		{ 0.00000978, -0.00001083, 0.00000868, -0.00000325, -0.00001087, -0.00000563, 0.00002026, -0.00001211, 0.00000940, -0.00002422, 0.00001520, -0.00001336, -0.00001176, -0.00002361, -0.00001084, -0.00003023, 0.00003771, -0.00007908, -0.00001924, 0.00019561, 0.00016461, 0.00006598, -0.00016219, 0.00000435, -0.00006454, 0.00002041, -0.00001243, -0.00001641, 0.00000527, -0.00000468, 0.00000441, -0.00000199, -0.00000914, 0.00000759, 0.00000116, -0.00000667, 0.00000134, 0.00000631, -0.00001444, 0.00000061, -0.00000016, -0.00000017, 0.00000770, -0.00001023, 0.00000343, -0.00000167, 0.00000463, -0.00000286, -0.00000348, -0.00000332, 0.00000520, -0.00000624, 0.00000765, -0.00000107, 0.00000151, -0.00000589, 0.00000347, -0.00000450, 0.00000676, -0.00000498, 0.00000290, -0.00000441, 0.00000501, -0.00000678, 0.00000481, -0.00000600, 0.00000917, -0.00000687, 0.00000624, -0.00000541, 0.00000281, -0.00000509, 0.00000432, -0.00000422, 0.00000513, -0.00000635, 0.00000570, -0.00000569, 0.00000854, -0.00000601, 0.00000661, -0.00000620, 0.00000356, -0.00000653, 0.00000321, -0.00000388, 0.00000486, -0.00000531, 0.00000523, -0.00000541, 0.00000636, -0.00000596, 0.00000680, -0.00000337, 0.00000317, -0.00000485, 0.00000274, -0.00000281, 0.00000318, -0.00000542, 0.00000271, -0.00000498, 0.00000751, -0.00000578, 0.00000618, -0.00000320, 0.00000454, -0.00000534, -0.00000044, -0.00000068, -0.00000054, -0.00000249, 0.00000572, -0.00000364, 0.00000635, -0.00000771, 0.00000028, -0.00000310, 0.00001126, -0.00000417, 0.00000150, -0.00000659, 0.00000184, -0.00000487, -0.00000083, 0.00000088, 0.00000321, -0.00000762, },
		{ -0.00000763, 0.00001017, -0.00001160, 0.00000943, -0.00000382, -0.00001127, -0.00000491, 0.00002049, -0.00001294, 0.00001015, -0.00002518, 0.00001643, -0.00001512, -0.00000951, -0.00002632, -0.00000767, -0.00003050, 0.00003857, -0.00008234, -0.00002515, 0.00020046, 0.00017665, 0.00005536, -0.00016363, 0.00000753, -0.00006736, 0.00002320, -0.00001459, -0.00001407, 0.00000362, -0.00000397, 0.00000460, -0.00000291, -0.00000822, 0.00000699, 0.00000146, -0.00000656, 0.00000156, 0.00000565, -0.00001470, 0.00000141, -0.00000059, 0.00000001, 0.00000755, -0.00000994, 0.00000349, -0.00000188, 0.00000486, -0.00000330, -0.00000336, -0.00000333, 0.00000508, -0.00000587, 0.00000782, -0.00000160, 0.00000165, -0.00000591, 0.00000368, -0.00000462, 0.00000678, -0.00000484, 0.00000274, -0.00000441, 0.00000486, -0.00000674, 0.00000489, -0.00000597, 0.00000923, -0.00000695, 0.00000654, -0.00000583, 0.00000298, -0.00000509, 0.00000427, -0.00000413, 0.00000487, -0.00000613, 0.00000560, -0.00000559, 0.00000860, -0.00000613, 0.00000675, -0.00000639, 0.00000369, -0.00000659, 0.00000321, -0.00000377, 0.00000468, -0.00000521, 0.00000508, -0.00000535, 0.00000648, -0.00000606, 0.00000697, -0.00000355, 0.00000331, -0.00000490, 0.00000270, -0.00000277, 0.00000309, -0.00000532, 0.00000235, -0.00000465, 0.00000761, -0.00000597, 0.00000651, -0.00000362, 0.00000496, -0.00000574, -0.00000038, -0.00000071, -0.00000052, -0.00000219, 0.00000550, -0.00000373, 0.00000648, -0.00000786, 0.00000023, -0.00000290, 0.00001177, -0.00000489, 0.00000192, -0.00000718, 0.00000224, -0.00000502, -0.00000074, 0.00000066, 0.00000352, },
		{ 0.00000379, -0.00000761, 0.00001049, -0.00001228, 0.00001014, -0.00000444, -0.00001162, -0.00000409, 0.00002052, -0.00001358, 0.00001070, -0.00002585, 0.00001736, -0.00001671, -0.00000733, -0.00002894, -0.00000444, -0.00003099, 0.00003936, -0.00008529, -0.00003149, 0.00020584, 0.00018919, 0.00004298, -0.00016351, 0.00000984, -0.00006957, 0.00002556, -0.00001657, -0.00001171, 0.00000187, -0.00000321, 0.00000473, -0.00000381, -0.00000721, 0.00000622, 0.00000189, -0.00000652, 0.00000186, 0.00000483, -0.00001479, 0.00000211, -0.00000096, 0.00000014, 0.00000742, -0.00000964, 0.00000357, -0.00000214, 0.00000511, -0.00000378, -0.00000317, -0.00000339, 0.00000499, -0.00000548, 0.00000797, -0.00000215, 0.00000184, -0.00000596, 0.00000393, -0.00000479, 0.00000685, -0.00000474, 0.00000260, -0.00000443, 0.00000472, -0.00000668, 0.00000495, -0.00000593, 0.00000927, -0.00000703, 0.00000684, -0.00000628, 0.00000321, -0.00000514, 0.00000426, -0.00000407, 0.00000463, -0.00000590, 0.00000548, -0.00000547, 0.00000864, -0.00000623, 0.00000689, -0.00000659, 0.00000383, -0.00000667, 0.00000323, -0.00000367, 0.00000451, -0.00000509, 0.00000492, -0.00000525, 0.00000657, -0.00000615, 0.00000712, -0.00000375, 0.00000347, -0.00000498, 0.00000269, -0.00000274, 0.00000301, -0.00000521, 0.00000198, -0.00000426, 0.00000764, -0.00000611, 0.00000682, -0.00000402, 0.00000537, -0.00000616, -0.00000028, -0.00000077, -0.00000048, -0.00000190, 0.00000525, -0.00000377, 0.00000656, -0.00000797, 0.00000016, -0.00000264, 0.00001219, -0.00000559, 0.00000232, -0.00000774, 0.00000262, -0.00000517, -0.00000066, 0.00000048, },
		{ 0.00000034, 0.00000402, -0.00000753, 0.00001073, -0.00001287, 0.00001079, -0.00000509, -0.00001192, -0.00000317, 0.00002032, -0.00001402, 0.00001103, -0.00002620, 0.00001797, -0.00001807, -0.00000530, -0.00003140, -0.00000125, -0.00003163, 0.00004003, -0.00008787, -0.00003835, 0.00021188, 0.00020208, 0.00002888, -0.00016176, 0.00001118, -0.00007108, 0.00002742, -0.00001828, -0.00000938, 0.00000009, -0.00000242, 0.00000482, -0.00000468, -0.00000610, 0.00000529, 0.00000245, -0.00000656, 0.00000225, 0.00000382, -0.00001469, 0.00000271, -0.00000127, 0.00000022, 0.00000732, -0.00000934, 0.00000366, -0.00000244, 0.00000540, -0.00000430, -0.00000291, -0.00000352, 0.00000493, -0.00000509, 0.00000810, -0.00000273, 0.00000208, -0.00000605, 0.00000422, -0.00000502, 0.00000696, -0.00000468, 0.00000249, -0.00000448, 0.00000459, -0.00000661, 0.00000501, -0.00000587, 0.00000930, -0.00000708, 0.00000713, -0.00000677, 0.00000348, -0.00000523, 0.00000429, -0.00000403, 0.00000440, -0.00000566, 0.00000534, -0.00000532, 0.00000864, -0.00000632, 0.00000702, -0.00000679, 0.00000399, -0.00000677, 0.00000327, -0.00000359, 0.00000433, -0.00000497, 0.00000474, -0.00000512, 0.00000663, -0.00000621, 0.00000726, -0.00000394, 0.00000364, -0.00000509, 0.00000269, -0.00000272, 0.00000292, -0.00000510, 0.00000159, -0.00000382, 0.00000760, -0.00000619, 0.00000707, -0.00000440, 0.00000577, -0.00000658, -0.00000014, -0.00000086, -0.00000041, -0.00000162, 0.00000498, -0.00000375, 0.00000657, -0.00000804, 0.00000007, -0.00000232, 0.00001253, -0.00000624, 0.00000270, -0.00000827, 0.00000296, -0.00000528, -0.00000060, },
		{ -0.00000057, 0.00000027, 0.00000417, -0.00000739, 0.00001088, -0.00001334, 0.00001135, -0.00000577, -0.00001218, -0.00000216, 0.00001991, -0.00001424, 0.00001111, -0.00002620, 0.00001820, -0.00001913, -0.00000350, -0.00003360, 0.00000181, -0.00003234, 0.00004053, -0.00009002, -0.00004580, 0.00021870, 0.00021519, 0.00001314, -0.00015836, 0.00001148, -0.00007181, 0.00002867, -0.00001964, -0.00000717, -0.00000170, -0.00000164, 0.00000488, -0.00000554, -0.00000491, 0.00000419, 0.00000316, -0.00000669, 0.00000274, 0.00000264, -0.00001438, 0.00000318, -0.00000150, 0.00000023, 0.00000725, -0.00000907, 0.00000377, -0.00000280, 0.00000573, -0.00000488, -0.00000257, -0.00000372, 0.00000491, -0.00000470, 0.00000822, -0.00000335, 0.00000236, -0.00000617, 0.00000456, -0.00000530, 0.00000712, -0.00000468, 0.00000243, -0.00000456, 0.00000448, -0.00000654, 0.00000507, -0.00000581, 0.00000930, -0.00000713, 0.00000742, -0.00000727, 0.00000379, -0.00000537, 0.00000436, -0.00000404, 0.00000419, -0.00000544, 0.00000520, -0.00000516, 0.00000863, -0.00000638, 0.00000713, -0.00000700, 0.00000417, -0.00000690, 0.00000334, -0.00000353, 0.00000417, -0.00000484, 0.00000455, -0.00000496, 0.00000666, -0.00000624, 0.00000738, -0.00000413, 0.00000383, -0.00000521, 0.00000272, -0.00000272, 0.00000284, -0.00000499, 0.00000119, -0.00000334, 0.00000750, -0.00000621, 0.00000728, -0.00000474, 0.00000614, -0.00000701, 0.00000001, -0.00000096, -0.00000032, -0.00000135, 0.00000467, -0.00000368, 0.00000652, -0.00000807, -0.00000005, -0.00000194, 0.00001277, -0.00000683, 0.00000305, -0.00000875, 0.00000326, -0.00000537, },
		{ -0.00000541, -0.00000058, 0.00000025, 0.00000425, -0.00000717, 0.00001091, -0.00001367, 0.00001181, -0.00000644, -0.00001238, -0.00000106, 0.00001929, -0.00001423, 0.00001093, -0.00002584, 0.00001801, -0.00001984, -0.00000198, -0.00003547, 0.00000465, -0.00003304, 0.00004079, -0.00009168, -0.00005393, 0.00022643, 0.00022835, -0.00000414, -0.00015330, 0.00001069, -0.00007169, 0.00002926, -0.00002057, -0.00000513, -0.00000342, -0.00000090, 0.00000494, -0.00000638, -0.00000364, 0.00000292, 0.00000401, -0.00000691, 0.00000332, 0.00000126, -0.00001385, 0.00000349, -0.00000162, 0.00000016, 0.00000724, -0.00000881, 0.00000391, -0.00000322, 0.00000612, -0.00000550, -0.00000215, -0.00000399, 0.00000494, -0.00000432, 0.00000834, -0.00000398, 0.00000269, -0.00000633, 0.00000493, -0.00000563, 0.00000734, -0.00000473, 0.00000243, -0.00000468, 0.00000439, -0.00000647, 0.00000512, -0.00000574, 0.00000929, -0.00000715, 0.00000770, -0.00000779, 0.00000414, -0.00000555, 0.00000447, -0.00000408, 0.00000402, -0.00000522, 0.00000506, -0.00000497, 0.00000858, -0.00000642, 0.00000723, -0.00000720, 0.00000435, -0.00000704, 0.00000343, -0.00000350, 0.00000401, -0.00000471, 0.00000434, -0.00000477, 0.00000665, -0.00000624, 0.00000747, -0.00000431, 0.00000401, -0.00000535, 0.00000277, -0.00000273, 0.00000277, -0.00000489, 0.00000079, -0.00000281, 0.00000732, -0.00000616, 0.00000742, -0.00000503, 0.00000647, -0.00000743, 0.00000019, -0.00000109, -0.00000021, -0.00000109, 0.00000435, -0.00000355, 0.00000641, -0.00000804, -0.00000020, -0.00000148, 0.00001290, -0.00000736, 0.00000334, -0.00000917, 0.00000350, },
		{ 0.00000367, -0.00000541, -0.00000062, 0.00000031, 0.00000425, -0.00000687, 0.00001083, -0.00001387, 0.00001215, -0.00000711, -0.00001255, 0.00000012, 0.00001845, -0.00001400, 0.00001050, -0.00002511, 0.00001739, -0.00002016, -0.00000081, -0.00003693, 0.00000718, -0.00003364, 0.00004076, -0.00009278, -0.00006284, 0.00023522, 0.00024136, -0.00002285, -0.00014659, 0.00000876, -0.00007066, 0.00002911, -0.00002100, -0.00000333, -0.00000505, -0.00000024, 0.00000500, -0.00000721, -0.00000230, 0.00000150, 0.00000502, -0.00000724, 0.00000401, -0.00000030, -0.00001309, 0.00000365, -0.00000163, 0.00000000, 0.00000729, -0.00000859, 0.00000409, -0.00000370, 0.00000655, -0.00000618, -0.00000165, -0.00000435, 0.00000504, -0.00000396, 0.00000845, -0.00000465, 0.00000306, -0.00000652, 0.00000534, -0.00000601, 0.00000761, -0.00000485, 0.00000247, -0.00000486, 0.00000434, -0.00000640, 0.00000518, -0.00000567, 0.00000926, -0.00000715, 0.00000796, -0.00000833, 0.00000454, -0.00000578, 0.00000463, -0.00000417, 0.00000388, -0.00000501, 0.00000491, -0.00000477, 0.00000851, -0.00000644, 0.00000731, -0.00000739, 0.00000454, -0.00000720, 0.00000355, -0.00000349, 0.00000387, -0.00000458, 0.00000412, -0.00000455, 0.00000660, -0.00000620, 0.00000753, -0.00000448, 0.00000420, -0.00000551, 0.00000284, -0.00000276, 0.00000271, -0.00000479, 0.00000039, -0.00000224, 0.00000708, -0.00000604, 0.00000749, -0.00000527, 0.00000675, -0.00000783, 0.00000038, -0.00000122, -0.00000009, -0.00000084, 0.00000400, -0.00000336, 0.00000622, -0.00000795, -0.00000039, -0.00000095, 0.00001291, -0.00000780, 0.00000358, -0.00000952, },
		{ -0.00000980, 0.00000378, -0.00000535, -0.00000071, 0.00000045, 0.00000415, -0.00000648, 0.00001062, -0.00001390, 0.00001234, -0.00000775, -0.00001267, 0.00000138, 0.00001740, -0.00001354, 0.00000982, -0.00002399, 0.00001630, -0.00002004, -0.00000004, -0.00003791, 0.00000930, -0.00003407, 0.00004037, -0.00009325, -0.00007260, 0.00024521, 0.00025403, -0.00004282, -0.00013827, 0.00000568, -0.00006869, 0.00002817, -0.00002087, -0.00000184, -0.00000651, 0.00000032, 0.00000510, -0.00000803, -0.00000088, -0.00000008, 0.00000617, -0.00000767, 0.00000480, -0.00000205, -0.00001210, 0.00000362, -0.00000151, -0.00000026, 0.00000742, -0.00000842, 0.00000431, -0.00000424, 0.00000704, -0.00000692, -0.00000106, -0.00000479, 0.00000521, -0.00000363, 0.00000856, -0.00000534, 0.00000348, -0.00000675, 0.00000580, -0.00000645, 0.00000795, -0.00000504, 0.00000258, -0.00000509, 0.00000433, -0.00000635, 0.00000525, -0.00000559, 0.00000921, -0.00000713, 0.00000821, -0.00000887, 0.00000496, -0.00000606, 0.00000483, -0.00000430, 0.00000378, -0.00000483, 0.00000477, -0.00000457, 0.00000841, -0.00000643, 0.00000737, -0.00000758, 0.00000474, -0.00000738, 0.00000369, -0.00000351, 0.00000375, -0.00000445, 0.00000390, -0.00000430, 0.00000652, -0.00000612, 0.00000755, -0.00000462, 0.00000438, -0.00000568, 0.00000292, -0.00000281, 0.00000266, -0.00000470, -0.00000000, -0.00000163, 0.00000677, -0.00000585, 0.00000750, -0.00000545, 0.00000698, -0.00000820, 0.00000057, -0.00000137, 0.00000005, -0.00000061, 0.00000363, -0.00000312, 0.00000595, -0.00000781, -0.00000062, -0.00000034, 0.00001280, -0.00000816, 0.00000376, },
		{ 0.00000386, -0.00001000, 0.00000381, -0.00000522, -0.00000085, 0.00000068, 0.00000394, -0.00000598, 0.00001028, -0.00001376, 0.00001237, -0.00000834, -0.00001276, 0.00000270, 0.00001615, -0.00001287, 0.00000888, -0.00002249, 0.00001473, -0.00001946, 0.00000028, -0.00003835, 0.00001092, -0.00003423, 0.00003955, -0.00009301, -0.00008333, 0.00025656, 0.00026613, -0.00006388, -0.00012843, 0.00000145, -0.00006575, 0.00002641, -0.00002012, -0.00000072, -0.00000778, 0.00000073, 0.00000526, -0.00000886, 0.00000059, -0.00000180, 0.00000748, -0.00000820, 0.00000568, -0.00000397, -0.00001086, 0.00000341, -0.00000125, -0.00000064, 0.00000763, -0.00000829, 0.00000457, -0.00000485, 0.00000759, -0.00000772, -0.00000038, -0.00000533, 0.00000546, -0.00000333, 0.00000867, -0.00000606, 0.00000395, -0.00000701, 0.00000628, -0.00000694, 0.00000834, -0.00000529, 0.00000276, -0.00000538, 0.00000437, -0.00000632, 0.00000534, -0.00000552, 0.00000916, -0.00000710, 0.00000843, -0.00000941, 0.00000542, -0.00000637, 0.00000508, -0.00000449, 0.00000373, -0.00000467, 0.00000465, -0.00000435, 0.00000829, -0.00000640, 0.00000741, -0.00000775, 0.00000494, -0.00000757, 0.00000385, -0.00000356, 0.00000365, -0.00000433, 0.00000367, -0.00000403, 0.00000640, -0.00000600, 0.00000754, -0.00000474, 0.00000455, -0.00000585, 0.00000302, -0.00000287, 0.00000262, -0.00000462, -0.00000039, -0.00000100, 0.00000639, -0.00000558, 0.00000742, -0.00000555, 0.00000714, -0.00000853, 0.00000075, -0.00000151, 0.00000020, -0.00000039, 0.00000325, -0.00000283, 0.00000562, -0.00000760, -0.00000088, 0.00000034, 0.00001257, -0.00000842, },
		{ -0.00000858, 0.00000389, -0.00001011, 0.00000376, -0.00000504, -0.00000105, 0.00000098, 0.00000363, -0.00000539, 0.00000980, -0.00001344, 0.00001223, -0.00000887, -0.00001281, 0.00000407, 0.00001471, -0.00001198, 0.00000770, -0.00002062, 0.00001269, -0.00001840, 0.00000012, -0.00003818, 0.00001197, -0.00003404, 0.00003823, -0.00009201, -0.00009509, 0.00026941, 0.00027744, -0.00008582, -0.00011714, -0.00000389, -0.00006185, 0.00002380, -0.00001872, -0.00000001, -0.00000881, 0.00000097, 0.00000550, -0.00000970, 0.00000212, -0.00000366, 0.00000893, -0.00000884, 0.00000665, -0.00000607, -0.00000939, 0.00000300, -0.00000085, -0.00000114, 0.00000794, -0.00000823, 0.00000487, -0.00000552, 0.00000820, -0.00000858, 0.00000039, -0.00000596, 0.00000580, -0.00000308, 0.00000879, -0.00000679, 0.00000446, -0.00000731, 0.00000680, -0.00000747, 0.00000878, -0.00000561, 0.00000300, -0.00000573, 0.00000447, -0.00000633, 0.00000544, -0.00000546, 0.00000909, -0.00000704, 0.00000862, -0.00000995, 0.00000589, -0.00000672, 0.00000538, -0.00000472, 0.00000373, -0.00000454, 0.00000453, -0.00000414, 0.00000816, -0.00000635, 0.00000742, -0.00000790, 0.00000513, -0.00000776, 0.00000403, -0.00000363, 0.00000357, -0.00000423, 0.00000345, -0.00000374, 0.00000624, -0.00000585, 0.00000748, -0.00000482, 0.00000470, -0.00000602, 0.00000313, -0.00000295, 0.00000258, -0.00000455, -0.00000076, -0.00000035, 0.00000596, -0.00000524, 0.00000727, -0.00000558, 0.00000723, -0.00000882, 0.00000093, -0.00000165, 0.00000036, -0.00000018, 0.00000286, -0.00000249, 0.00000522, -0.00000733, -0.00000118, 0.00000109, 0.00001221, },
		{ 0.00001172, -0.00000863, 0.00000383, -0.00001013, 0.00000362, -0.00000479, -0.00000129, 0.00000137, 0.00000322, -0.00000468, 0.00000919, -0.00001294, 0.00001190, -0.00000932, -0.00001283, 0.00000550, 0.00001310, -0.00001089, 0.00000630, -0.00001840, 0.00001018, -0.00001684, -0.00000055, -0.00003738, 0.00001238, -0.00003342, 0.00003635, -0.00009018, -0.00010797, 0.00028392, 0.00028771, -0.00010842, -0.00010455, -0.00001031, -0.00005700, 0.00002033, -0.00001663, 0.00000024, -0.00000955, 0.00000100, 0.00000584, -0.00001055, 0.00000370, -0.00000564, 0.00001052, -0.00000957, 0.00000770, -0.00000832, -0.00000768, 0.00000240, -0.00000029, -0.00000177, 0.00000833, -0.00000822, 0.00000522, -0.00000626, 0.00000886, -0.00000950, 0.00000124, -0.00000669, 0.00000623, -0.00000287, 0.00000891, -0.00000755, 0.00000501, -0.00000763, 0.00000735, -0.00000804, 0.00000928, -0.00000600, 0.00000331, -0.00000615, 0.00000462, -0.00000636, 0.00000557, -0.00000541, 0.00000902, -0.00000696, 0.00000878, -0.00001047, 0.00000639, -0.00000711, 0.00000571, -0.00000500, 0.00000377, -0.00000445, 0.00000444, -0.00000392, 0.00000800, -0.00000627, 0.00000740, -0.00000802, 0.00000531, -0.00000796, 0.00000423, -0.00000373, 0.00000351, -0.00000413, 0.00000323, -0.00000343, 0.00000605, -0.00000565, 0.00000738, -0.00000487, 0.00000482, -0.00000618, 0.00000325, -0.00000304, 0.00000256, -0.00000450, -0.00000111, 0.00000032, 0.00000547, -0.00000483, 0.00000705, -0.00000553, 0.00000723, -0.00000905, 0.00000108, -0.00000178, 0.00000051, 0.00000001, 0.00000247, -0.00000211, 0.00000474, -0.00000701, -0.00000152, 0.00000192, },
		{ 0.00000281, 0.00001110, -0.00000858, 0.00000369, -0.00001006, 0.00000339, -0.00000446, -0.00000159, 0.00000185, 0.00000270, -0.00000388, 0.00000843, -0.00001224, 0.00001137, -0.00000969, -0.00001283, 0.00000695, 0.00001134, -0.00000963, 0.00000469, -0.00001585, 0.00000721, -0.00001479, -0.00000174, -0.00003590, 0.00001209, -0.00003231, 0.00003386, -0.00008747, -0.00012204, 0.00030023, 0.00029671, -0.00013143, -0.00009078, -0.00001772, -0.00005124, 0.00001601, -0.00001384, -0.00000000, -0.00000997, 0.00000080, 0.00000629, -0.00001143, 0.00000532, -0.00000772, 0.00001224, -0.00001039, 0.00000881, -0.00001072, -0.00000575, 0.00000159, 0.00000042, -0.00000252, 0.00000883, -0.00000829, 0.00000562, -0.00000707, 0.00000958, -0.00001047, 0.00000218, -0.00000750, 0.00000675, -0.00000273, 0.00000905, -0.00000833, 0.00000560, -0.00000798, 0.00000792, -0.00000865, 0.00000982, -0.00000645, 0.00000369, -0.00000664, 0.00000484, -0.00000643, 0.00000572, -0.00000538, 0.00000895, -0.00000687, 0.00000891, -0.00001098, 0.00000689, -0.00000752, 0.00000608, -0.00000533, 0.00000387, -0.00000440, 0.00000438, -0.00000371, 0.00000783, -0.00000616, 0.00000735, -0.00000813, 0.00000548, -0.00000816, 0.00000445, -0.00000386, 0.00000348, -0.00000406, 0.00000302, -0.00000311, 0.00000583, -0.00000541, 0.00000723, -0.00000487, 0.00000492, -0.00000633, 0.00000337, -0.00000314, 0.00000256, -0.00000447, -0.00000143, 0.00000100, 0.00000493, -0.00000436, 0.00000674, -0.00000541, 0.00000716, -0.00000922, 0.00000121, -0.00000190, 0.00000067, 0.00000019, 0.00000207, -0.00000170, 0.00000421, -0.00000662, -0.00000188, },
		{ -0.00000228, 0.00000376, 0.00001036, -0.00000841, 0.00000346, -0.00000990, 0.00000308, -0.00000408, -0.00000193, 0.00000240, 0.00000207, -0.00000296, 0.00000754, -0.00001137, 0.00001064, -0.00000995, -0.00001280, 0.00000843, 0.00000943, -0.00000820, 0.00000291, -0.00001301, 0.00000382, -0.00001226, -0.00000345, -0.00003374, 0.00001106, -0.00003064, 0.00003070, -0.00008383, -0.00013735, 0.00031846, 0.00030418, -0.00015461, -0.00007602, -0.00002603, -0.00004463, 0.00001088, -0.00001034, -0.00000076, -0.00001006, 0.00000036, 0.00000689, -0.00001233, 0.00000697, -0.00000990, 0.00001407, -0.00001130, 0.00000998, -0.00001323, -0.00000360, 0.00000059, 0.00000128, -0.00000340, 0.00000944, -0.00000842, 0.00000606, -0.00000793, 0.00001035, -0.00001149, 0.00000319, -0.00000841, 0.00000737, -0.00000265, 0.00000919, -0.00000912, 0.00000622, -0.00000836, 0.00000851, -0.00000928, 0.00001040, -0.00000695, 0.00000412, -0.00000719, 0.00000513, -0.00000655, 0.00000591, -0.00000537, 0.00000888, -0.00000675, 0.00000900, -0.00001146, 0.00000739, -0.00000795, 0.00000648, -0.00000570, 0.00000402, -0.00000439, 0.00000434, -0.00000352, 0.00000766, -0.00000604, 0.00000727, -0.00000820, 0.00000563, -0.00000836, 0.00000467, -0.00000401, 0.00000347, -0.00000400, 0.00000282, -0.00000279, 0.00000559, -0.00000514, 0.00000703, -0.00000484, 0.00000498, -0.00000647, 0.00000350, -0.00000324, 0.00000256, -0.00000445, -0.00000171, 0.00000166, 0.00000436, -0.00000382, 0.00000636, -0.00000520, 0.00000699, -0.00000932, 0.00000131, -0.00000199, 0.00000083, 0.00000036, 0.00000168, -0.00000125, 0.00000362, -0.00000618, },
		{ -0.00000569, -0.00000269, 0.00000475, 0.00000951, -0.00000813, 0.00000315, -0.00000964, 0.00000268, -0.00000362, -0.00000232, 0.00000302, 0.00000135, -0.00000196, 0.00000652, -0.00001030, 0.00000970, -0.00001011, -0.00001274, 0.00000991, 0.00000741, -0.00000663, 0.00000098, -0.00000991, 0.00000004, -0.00000928, -0.00000567, -0.00003089, 0.00000925, -0.00002838, 0.00002684, -0.00007924, -0.00015395, 0.00033873, 0.00030989, -0.00017767, -0.00006045, -0.00003514, -0.00003723, 0.00000497, -0.00000615, -0.00000204, -0.00000978, -0.00000036, 0.00000763, -0.00001327, 0.00000863, -0.00001215, 0.00001600, -0.00001227, 0.00001118, -0.00001585, -0.00000126, -0.00000060, 0.00000230, -0.00000441, 0.00001014, -0.00000862, 0.00000654, -0.00000884, 0.00001116, -0.00001255, 0.00000427, -0.00000941, 0.00000808, -0.00000263, 0.00000934, -0.00000992, 0.00000688, -0.00000875, 0.00000910, -0.00000993, 0.00001102, -0.00000751, 0.00000462, -0.00000781, 0.00000548, -0.00000672, 0.00000614, -0.00000538, 0.00000882, -0.00000663, 0.00000905, -0.00001191, 0.00000790, -0.00000839, 0.00000691, -0.00000611, 0.00000422, -0.00000442, 0.00000434, -0.00000334, 0.00000747, -0.00000590, 0.00000716, -0.00000825, 0.00000577, -0.00000855, 0.00000490, -0.00000418, 0.00000349, -0.00000397, 0.00000265, -0.00000246, 0.00000533, -0.00000483, 0.00000679, -0.00000476, 0.00000500, -0.00000658, 0.00000362, -0.00000336, 0.00000257, -0.00000447, -0.00000196, 0.00000232, 0.00000376, -0.00000324, 0.00000591, -0.00000491, 0.00000674, -0.00000935, 0.00000138, -0.00000207, 0.00000098, 0.00000050, 0.00000129, -0.00000078, 0.00000297, },
		{ 0.00000228, -0.00000516, -0.00000311, 0.00000579, 0.00000854, -0.00000775, 0.00000275, -0.00000930, 0.00000220, -0.00000311, -0.00000274, 0.00000370, 0.00000053, -0.00000086, 0.00000538, -0.00000907, 0.00000856, -0.00001015, -0.00001267, 0.00001138, 0.00000529, -0.00000495, -0.00000106, -0.00000659, -0.00000408, -0.00000588, -0.00000837, -0.00002737, 0.00000667, -0.00002549, 0.00002226, -0.00007368, -0.00017186, 0.00036115, 0.00031361, -0.00020035, -0.00004427, -0.00004490, -0.00002915, -0.00000165, -0.00000129, -0.00000385, -0.00000913, -0.00000136, 0.00000853, -0.00001425, 0.00001030, -0.00001445, 0.00001801, -0.00001330, 0.00001239, -0.00001854, 0.00000125, -0.00000198, 0.00000346, -0.00000554, 0.00001095, -0.00000889, 0.00000707, -0.00000980, 0.00001202, -0.00001365, 0.00000541, -0.00001048, 0.00000889, -0.00000268, 0.00000951, -0.00001073, 0.00000756, -0.00000916, 0.00000970, -0.00001060, 0.00001166, -0.00000812, 0.00000517, -0.00000849, 0.00000590, -0.00000693, 0.00000640, -0.00000542, 0.00000876, -0.00000649, 0.00000907, -0.00001232, 0.00000838, -0.00000884, 0.00000735, -0.00000656, 0.00000448, -0.00000450, 0.00000438, -0.00000318, 0.00000729, -0.00000574, 0.00000703, -0.00000827, 0.00000588, -0.00000872, 0.00000514, -0.00000438, 0.00000354, -0.00000397, 0.00000249, -0.00000214, 0.00000504, -0.00000449, 0.00000651, -0.00000464, 0.00000499, -0.00000667, 0.00000373, -0.00000348, 0.00000260, -0.00000450, -0.00000216, 0.00000295, 0.00000313, -0.00000261, 0.00000540, -0.00000456, 0.00000640, -0.00000930, 0.00000140, -0.00000211, 0.00000112, 0.00000064, 0.00000092, -0.00000030, },
		{ 0.00000019, 0.00000156, -0.00000459, -0.00000354, 0.00000685, 0.00000748, -0.00000726, 0.00000228, -0.00000888, 0.00000164, -0.00000254, -0.00000319, 0.00000443, -0.00000036, 0.00000031, 0.00000414, -0.00000766, 0.00000720, -0.00001006, -0.00001256, 0.00001282, 0.00000311, -0.00000318, -0.00000317, -0.00000311, -0.00000850, -0.00000210, -0.00001151, -0.00002321, 0.00000330, -0.00002195, 0.00001695, -0.00006715, -0.00019108, 0.00038579, 0.00031511, -0.00022236, -0.00002770, -0.00005517, -0.00002049, -0.00000892, 0.00000419, -0.00000619, -0.00000810, -0.00000264, 0.00000959, -0.00001526, 0.00001196, -0.00001678, 0.00002009, -0.00001436, 0.00001360, -0.00002127, 0.00000392, -0.00000351, 0.00000476, -0.00000678, 0.00001185, -0.00000923, 0.00000762, -0.00001079, 0.00001290, -0.00001477, 0.00000660, -0.00001162, 0.00000979, -0.00000281, 0.00000970, -0.00001155, 0.00000826, -0.00000958, 0.00001030, -0.00001126, 0.00001231, -0.00000876, 0.00000576, -0.00000922, 0.00000638, -0.00000719, 0.00000670, -0.00000549, 0.00000871, -0.00000634, 0.00000904, -0.00001269, 0.00000885, -0.00000929, 0.00000781, -0.00000703, 0.00000477, -0.00000463, 0.00000445, -0.00000305, 0.00000711, -0.00000558, 0.00000686, -0.00000825, 0.00000596, -0.00000888, 0.00000537, -0.00000459, 0.00000361, -0.00000399, 0.00000237, -0.00000183, 0.00000475, -0.00000412, 0.00000619, -0.00000447, 0.00000492, -0.00000673, 0.00000383, -0.00000360, 0.00000263, -0.00000456, -0.00000232, 0.00000356, 0.00000250, -0.00000194, 0.00000483, -0.00000413, 0.00000597, -0.00000918, 0.00000138, -0.00000213, 0.00000125, 0.00000075, 0.00000057, },
		{ 0.00000025, 0.00000067, 0.00000081, -0.00000400, -0.00000396, 0.00000792, 0.00000634, -0.00000669, 0.00000174, -0.00000838, 0.00000101, -0.00000193, -0.00000366, 0.00000520, -0.00000131, 0.00000155, 0.00000280, -0.00000611, 0.00000565, -0.00000985, -0.00001243, 0.00001422, 0.00000088, -0.00000137, -0.00000532, 0.00000048, -0.00001314, 0.00000198, -0.00001504, -0.00001845, -0.00000083, -0.00001777, 0.00001091, -0.00005967, -0.00021161, 0.00041271, 0.00031420, -0.00024343, -0.00001096, -0.00006580, -0.00001136, -0.00001672, 0.00001024, -0.00000901, -0.00000671, -0.00000420, 0.00001083, -0.00001631, 0.00001361, -0.00001912, 0.00002221, -0.00001545, 0.00001478, -0.00002402, 0.00000670, -0.00000520, 0.00000618, -0.00000813, 0.00001284, -0.00000963, 0.00000821, -0.00001181, 0.00001380, -0.00001591, 0.00000782, -0.00001281, 0.00001076, -0.00000300, 0.00000989, -0.00001236, 0.00000897, -0.00001000, 0.00001088, -0.00001192, 0.00001297, -0.00000943, 0.00000639, -0.00001000, 0.00000693, -0.00000750, 0.00000704, -0.00000559, 0.00000868, -0.00000618, 0.00000898, -0.00001302, 0.00000930, -0.00000974, 0.00000828, -0.00000753, 0.00000512, -0.00000480, 0.00000456, -0.00000294, 0.00000693, -0.00000540, 0.00000668, -0.00000820, 0.00000601, -0.00000901, 0.00000560, -0.00000481, 0.00000371, -0.00000404, 0.00000227, -0.00000153, 0.00000446, -0.00000373, 0.00000583, -0.00000426, 0.00000482, -0.00000675, 0.00000392, -0.00000372, 0.00000267, -0.00000464, -0.00000242, 0.00000412, 0.00000187, -0.00000125, 0.00000422, -0.00000364, 0.00000546, -0.00000898, 0.00000132, -0.00000212, 0.00000136, 0.00000085, },
		{ 0.00000093, -0.00000005, 0.00000114, 0.00000004, -0.00000338, -0.00000436, 0.00000898, 0.00000512, -0.00000604, 0.00000113, -0.00000781, 0.00000033, -0.00000128, -0.00000414, 0.00000599, -0.00000232, 0.00000283, 0.00000139, -0.00000443, 0.00000392, -0.00000952, -0.00001228, 0.00001557, -0.00000136, 0.00000047, -0.00000746, 0.00000412, -0.00001795, 0.00000632, -0.00001889, -0.00001315, -0.00000567, -0.00001297, 0.00000417, -0.00005126, -0.00023340, 0.00044193, 0.00031068, -0.00026328, 0.00000571, -0.00007660, -0.00000191, -0.00002497, 0.00001679, -0.00001231, -0.00000496, -0.00000604, 0.00001223, -0.00001739, 0.00001522, -0.00002145, 0.00002434, -0.00001654, 0.00001591, -0.00002674, 0.00000958, -0.00000701, 0.00000770, -0.00000957, 0.00001391, -0.00001009, 0.00000881, -0.00001285, 0.00001471, -0.00001704, 0.00000907, -0.00001405, 0.00001182, -0.00000326, 0.00001010, -0.00001316, 0.00000970, -0.00001042, 0.00001145, -0.00001256, 0.00001362, -0.00001011, 0.00000705, -0.00001082, 0.00000753, -0.00000786, 0.00000742, -0.00000572, 0.00000867, -0.00000602, 0.00000887, -0.00001329, 0.00000971, -0.00001016, 0.00000874, -0.00000805, 0.00000550, -0.00000501, 0.00000471, -0.00000287, 0.00000677, -0.00000522, 0.00000647, -0.00000812, 0.00000604, -0.00000912, 0.00000581, -0.00000505, 0.00000383, -0.00000411, 0.00000221, -0.00000126, 0.00000416, -0.00000333, 0.00000543, -0.00000400, 0.00000467, -0.00000674, 0.00000399, -0.00000384, 0.00000271, -0.00000474, -0.00000246, 0.00000463, 0.00000125, -0.00000055, 0.00000357, -0.00000309, 0.00000487, -0.00000870, 0.00000121, -0.00000207, 0.00000146, },
		{ 0.00000155, 0.00000099, -0.00000031, 0.00000159, -0.00000073, -0.00000277, -0.00000473, 0.00001003, 0.00000386, -0.00000531, 0.00000047, -0.00000719, -0.00000041, -0.00000060, -0.00000461, 0.00000679, -0.00000335, 0.00000414, -0.00000006, -0.00000264, 0.00000201, -0.00000907, -0.00001209, 0.00001683, -0.00000359, 0.00000231, -0.00000955, 0.00000774, -0.00002285, 0.00001084, -0.00002300, -0.00000740, -0.00001117, -0.00000758, -0.00000323, -0.00004200, -0.00025637, 0.00047346, 0.00030440, -0.00028166, 0.00002207, -0.00008741, 0.00000772, -0.00003355, 0.00002375, -0.00001602, -0.00000289, -0.00000814, 0.00001380, -0.00001850, 0.00001679, -0.00002373, 0.00002647, -0.00001761, 0.00001696, -0.00002942, 0.00001252, -0.00000892, 0.00000932, -0.00001108, 0.00001504, -0.00001060, 0.00000942, -0.00001389, 0.00001562, -0.00001817, 0.00001033, -0.00001532, 0.00001294, -0.00000358, 0.00001031, -0.00001396, 0.00001043, -0.00001084, 0.00001199, -0.00001317, 0.00001425, -0.00001080, 0.00000772, -0.00001166, 0.00000818, -0.00000827, 0.00000784, -0.00000589, 0.00000868, -0.00000586, 0.00000872, -0.00001351, 0.00001008, -0.00001056, 0.00000919, -0.00000857, 0.00000591, -0.00000526, 0.00000491, -0.00000283, 0.00000662, -0.00000504, 0.00000624, -0.00000801, 0.00000603, -0.00000920, 0.00000601, -0.00000529, 0.00000397, -0.00000422, 0.00000218, -0.00000101, 0.00000388, -0.00000292, 0.00000502, -0.00000371, 0.00000447, -0.00000669, 0.00000404, -0.00000395, 0.00000276, -0.00000487, -0.00000244, 0.00000508, 0.00000066, 0.00000015, 0.00000290, -0.00000250, 0.00000421, -0.00000835, 0.00000106, -0.00000199, },
		{ -0.00000188, 0.00000162, 0.00000103, -0.00000053, 0.00000201, -0.00000149, -0.00000215, -0.00000506, 0.00001105, 0.00000255, -0.00000453, -0.00000023, -0.00000651, -0.00000119, 0.00000010, -0.00000506, 0.00000757, -0.00000439, 0.00000545, -0.00000155, -0.00000076, -0.00000006, -0.00000850, -0.00001186, 0.00001801, -0.00000577, 0.00000410, -0.00001155, 0.00001130, -0.00002777, 0.00001546, -0.00002728, -0.00000127, -0.00001726, -0.00000165, -0.00001124, -0.00003195, -0.00028046, 0.00050726, 0.00029523, -0.00029831, 0.00003791, -0.00009804, 0.00001737, -0.00004232, 0.00003104, -0.00002009, -0.00000051, -0.00001048, 0.00001552, -0.00001964, 0.00001830, -0.00002596, 0.00002857, -0.00001865, 0.00001791, -0.00003201, 0.00001548, -0.00001091, 0.00001101, -0.00001265, 0.00001623, -0.00001114, 0.00001004, -0.00001492, 0.00001651, -0.00001927, 0.00001158, -0.00001660, 0.00001412, -0.00000396, 0.00001053, -0.00001473, 0.00001115, -0.00001125, 0.00001250, -0.00001374, 0.00001486, -0.00001148, 0.00000840, -0.00001252, 0.00000887, -0.00000871, 0.00000830, -0.00000610, 0.00000871, -0.00000571, 0.00000854, -0.00001368, 0.00001040, -0.00001093, 0.00000962, -0.00000909, 0.00000635, -0.00000555, 0.00000514, -0.00000282, 0.00000649, -0.00000486, 0.00000599, -0.00000786, 0.00000598, -0.00000926, 0.00000619, -0.00000553, 0.00000413, -0.00000435, 0.00000220, -0.00000080, 0.00000361, -0.00000252, 0.00000458, -0.00000339, 0.00000424, -0.00000661, 0.00000407, -0.00000405, 0.00000281, -0.00000503, -0.00000236, 0.00000547, 0.00000011, 0.00000083, 0.00000222, -0.00000188, 0.00000350, -0.00000793, 0.00000087, },
		{ 0.00000064, -0.00000174, 0.00000168, 0.00000104, -0.00000070, 0.00000238, -0.00000223, -0.00000155, -0.00000534, 0.00001201, 0.00000123, -0.00000371, -0.00000097, -0.00000581, -0.00000199, 0.00000080, -0.00000547, 0.00000832, -0.00000543, 0.00000675, -0.00000303, 0.00000117, -0.00000226, -0.00000782, -0.00001160, 0.00001907, -0.00000789, 0.00000581, -0.00001342, 0.00001472, -0.00003262, 0.00002009, -0.00003164, 0.00000514, -0.00002385, 0.00000474, -0.00001979, -0.00002122, -0.00030552, 0.00054327, 0.00028306, -0.00031302, 0.00005297, -0.00010830, 0.00002691, -0.00005116, 0.00003855, -0.00002445, 0.00000212, -0.00001304, 0.00001738, -0.00002079, 0.00001973, -0.00002809, 0.00003062, -0.00001963, 0.00001874, -0.00003447, 0.00001843, -0.00001295, 0.00001274, -0.00001426, 0.00001746, -0.00001172, 0.00001064, -0.00001594, 0.00001738, -0.00002033, 0.00001281, -0.00001788, 0.00001533, -0.00000440, 0.00001075, -0.00001549, 0.00001187, -0.00001164, 0.00001297, -0.00001427, 0.00001542, -0.00001214, 0.00000906, -0.00001339, 0.00000959, -0.00000918, 0.00000878, -0.00000633, 0.00000876, -0.00000556, 0.00000832, -0.00001378, 0.00001068, -0.00001126, 0.00001002, -0.00000960, 0.00000680, -0.00000587, 0.00000541, -0.00000285, 0.00000639, -0.00000468, 0.00000573, -0.00000769, 0.00000591, -0.00000927, 0.00000635, -0.00000577, 0.00000430, -0.00000451, 0.00000225, -0.00000061, 0.00000336, -0.00000211, 0.00000413, -0.00000304, 0.00000396, -0.00000649, 0.00000407, -0.00000415, 0.00000286, -0.00000520, -0.00000222, 0.00000578, -0.00000040, 0.00000149, 0.00000153, -0.00000124, 0.00000273, -0.00000746, },
		{ -0.00000693, 0.00000037, -0.00000157, 0.00000172, 0.00000104, -0.00000083, 0.00000271, -0.00000294, -0.00000098, -0.00000556, 0.00001290, -0.00000009, -0.00000286, -0.00000171, -0.00000508, -0.00000281, 0.00000149, -0.00000585, 0.00000902, -0.00000643, 0.00000801, -0.00000449, 0.00000314, -0.00000456, -0.00000705, -0.00001128, 0.00002000, -0.00000991, 0.00000742, -0.00001511, 0.00001795, -0.00003733, 0.00002464, -0.00003598, 0.00001171, -0.00003084, 0.00001152, -0.00002877, -0.00000991, -0.00033143, 0.00058139, 0.00026783, -0.00032556, 0.00006707, -0.00011803, 0.00003616, -0.00005992, 0.00004617, -0.00002905, 0.00000496, -0.00001577, 0.00001935, -0.00002195, 0.00002107, -0.00003012, 0.00003259, -0.00002054, 0.00001942, -0.00003678, 0.00002133, -0.00001500, 0.00001449, -0.00001589, 0.00001871, -0.00001231, 0.00001122, -0.00001691, 0.00001819, -0.00002134, 0.00001399, -0.00001914, 0.00001658, -0.00000489, 0.00001097, -0.00001621, 0.00001257, -0.00001200, 0.00001340, -0.00001473, 0.00001593, -0.00001277, 0.00000971, -0.00001424, 0.00001034, -0.00000969, 0.00000930, -0.00000660, 0.00000884, -0.00000541, 0.00000806, -0.00001382, 0.00001089, -0.00001155, 0.00001038, -0.00001010, 0.00000727, -0.00000622, 0.00000572, -0.00000292, 0.00000630, -0.00000452, 0.00000546, -0.00000749, 0.00000580, -0.00000926, 0.00000649, -0.00000600, 0.00000448, -0.00000469, 0.00000234, -0.00000047, 0.00000313, -0.00000173, 0.00000367, -0.00000267, 0.00000365, -0.00000633, 0.00000405, -0.00000423, 0.00000291, -0.00000539, -0.00000201, 0.00000601, -0.00000085, 0.00000212, 0.00000087, -0.00000059, 0.00000192, },
		{ 0.00000109, -0.00000635, 0.00000008, -0.00000138, 0.00000174, 0.00000101, -0.00000090, 0.00000298, -0.00000361, -0.00000044, -0.00000570, 0.00001370, -0.00000139, -0.00000200, -0.00000246, -0.00000435, -0.00000363, 0.00000215, -0.00000615, 0.00000966, -0.00000738, 0.00000921, -0.00000590, 0.00000509, -0.00000695, -0.00000619, -0.00001091, 0.00002079, -0.00001180, 0.00000890, -0.00001658, 0.00002092, -0.00004182, 0.00002903, -0.00004020, 0.00001833, -0.00003811, 0.00001857, -0.00003809, 0.00000184, -0.00035801, 0.00062149, 0.00024951, -0.00033577, 0.00007997, -0.00012703, 0.00004496, -0.00006847, 0.00005379, -0.00003378, 0.00000795, -0.00001865, 0.00002143, -0.00002309, 0.00002230, -0.00003201, 0.00003446, -0.00002135, 0.00001993, -0.00003890, 0.00002415, -0.00001705, 0.00001625, -0.00001751, 0.00001997, -0.00001290, 0.00001177, -0.00001784, 0.00001896, -0.00002228, 0.00001512, -0.00002036, 0.00001784, -0.00000541, 0.00001118, -0.00001689, 0.00001324, -0.00001234, 0.00001377, -0.00001514, 0.00001638, -0.00001336, 0.00001032, -0.00001508, 0.00001109, -0.00001021, 0.00000984, -0.00000689, 0.00000895, -0.00000529, 0.00000778, -0.00001380, 0.00001105, -0.00001178, 0.00001070, -0.00001056, 0.00000774, -0.00000658, 0.00000605, -0.00000302, 0.00000625, -0.00000438, 0.00000519, -0.00000727, 0.00000566, -0.00000921, 0.00000659, -0.00000622, 0.00000467, -0.00000489, 0.00000247, -0.00000038, 0.00000294, -0.00000136, 0.00000322, -0.00000229, 0.00000331, -0.00000613, 0.00000400, -0.00000429, 0.00000295, -0.00000560, -0.00000173, 0.00000616, -0.00000123, 0.00000269, 0.00000023, 0.00000005, },
		{ 0.00000066, 0.00000024, -0.00000575, -0.00000024, -0.00000117, 0.00000176, 0.00000096, -0.00000091, 0.00000318, -0.00000423, 0.00000006, -0.00000575, 0.00001441, -0.00000265, -0.00000115, -0.00000320, -0.00000362, -0.00000443, 0.00000279, -0.00000639, 0.00001020, -0.00000825, 0.00001032, -0.00000723, 0.00000702, -0.00000939, -0.00000527, -0.00001048, 0.00002142, -0.00001355, 0.00001021, -0.00001780, 0.00002358, -0.00004601, 0.00003317, -0.00004421, 0.00002487, -0.00004554, 0.00002579, -0.00004762, 0.00001387, -0.00038506, 0.00066339, 0.00022810, -0.00034348, 0.00009152, -0.00013515, 0.00005318, -0.00007666, 0.00006128, -0.00003858, 0.00001104, -0.00002162, 0.00002357, -0.00002422, 0.00002341, -0.00003374, 0.00003621, -0.00002205, 0.00002025, -0.00004080, 0.00002684, -0.00001904, 0.00001797, -0.00001911, 0.00002122, -0.00001349, 0.00001228, -0.00001870, 0.00001964, -0.00002313, 0.00001617, -0.00002152, 0.00001909, -0.00000596, 0.00001137, -0.00001753, 0.00001389, -0.00001265, 0.00001409, -0.00001547, 0.00001675, -0.00001389, 0.00001088, -0.00001588, 0.00001184, -0.00001074, 0.00001040, -0.00000721, 0.00000908, -0.00000517, 0.00000746, -0.00001372, 0.00001115, -0.00001195, 0.00001097, -0.00001099, 0.00000820, -0.00000696, 0.00000642, -0.00000316, 0.00000622, -0.00000425, 0.00000492, -0.00000703, 0.00000549, -0.00000912, 0.00000666, -0.00000642, 0.00000486, -0.00000511, 0.00000264, -0.00000032, 0.00000279, -0.00000102, 0.00000277, -0.00000189, 0.00000295, -0.00000591, 0.00000393, -0.00000434, 0.00000299, -0.00000582, -0.00000139, 0.00000622, -0.00000154, 0.00000320, -0.00000037, },
		{ -0.00000091, 0.00000125, -0.00000060, -0.00000512, -0.00000056, -0.00000094, 0.00000177, 0.00000087, -0.00000087, 0.00000331, -0.00000479, 0.00000050, -0.00000570, 0.00001499, -0.00000385, -0.00000033, -0.00000391, -0.00000292, -0.00000520, 0.00000337, -0.00000653, 0.00001064, -0.00000903, 0.00001132, -0.00000844, 0.00000887, -0.00001185, -0.00000429, -0.00000997, 0.00002187, -0.00001513, 0.00001134, -0.00001874, 0.00002589, -0.00004983, 0.00003696, -0.00004789, 0.00003122, -0.00005299, 0.00003306, -0.00005722, 0.00002601, -0.00041236, 0.00070689, 0.00020366, -0.00034859, 0.00010152, -0.00014223, 0.00006066, -0.00008435, 0.00006852, -0.00004334, 0.00001416, -0.00002464, 0.00002575, -0.00002531, 0.00002438, -0.00003530, 0.00003781, -0.00002262, 0.00002036, -0.00004244, 0.00002937, -0.00002096, 0.00001964, -0.00002066, 0.00002243, -0.00001406, 0.00001273, -0.00001949, 0.00002025, -0.00002389, 0.00001713, -0.00002260, 0.00002032, -0.00000653, 0.00001155, -0.00001812, 0.00001450, -0.00001292, 0.00001434, -0.00001572, 0.00001705, -0.00001435, 0.00001138, -0.00001663, 0.00001258, -0.00001128, 0.00001096, -0.00000755, 0.00000924, -0.00000508, 0.00000713, -0.00001359, 0.00001118, -0.00001207, 0.00001118, -0.00001137, 0.00000865, -0.00000735, 0.00000680, -0.00000333, 0.00000622, -0.00000414, 0.00000465, -0.00000678, 0.00000529, -0.00000899, 0.00000669, -0.00000661, 0.00000506, -0.00000535, 0.00000285, -0.00000032, 0.00000269, -0.00000072, 0.00000235, -0.00000150, 0.00000256, -0.00000565, 0.00000384, -0.00000438, 0.00000302, -0.00000605, -0.00000099, 0.00000619, -0.00000176, 0.00000364, },
		{ 0.00000399, -0.00000139, 0.00000178, -0.00000142, -0.00000448, -0.00000089, -0.00000071, 0.00000178, 0.00000077, -0.00000076, 0.00000336, -0.00000528, 0.00000087, -0.00000555, 0.00001543, -0.00000496, 0.00000044, -0.00000457, -0.00000226, -0.00000593, 0.00000389, -0.00000657, 0.00001095, -0.00000968, 0.00001219, -0.00000952, 0.00001061, -0.00001430, -0.00000329, -0.00000939, 0.00002214, -0.00001651, 0.00001226, -0.00001937, 0.00002778, -0.00005321, 0.00004034, -0.00005114, 0.00003724, -0.00006033, 0.00004023, -0.00006676, 0.00003809, -0.00043969, 0.00075176, 0.00017627, -0.00035100, 0.00010985, -0.00014813, 0.00006726, -0.00009142, 0.00007539, -0.00004798, 0.00001724, -0.00002766, 0.00002794, -0.00002635, 0.00002520, -0.00003665, 0.00003925, -0.00002305, 0.00002025, -0.00004381, 0.00003170, -0.00002278, 0.00002122, -0.00002213, 0.00002359, -0.00001460, 0.00001311, -0.00002019, 0.00002076, -0.00002454, 0.00001798, -0.00002358, 0.00002151, -0.00000711, 0.00001170, -0.00001865, 0.00001506, -0.00001314, 0.00001453, -0.00001589, 0.00001726, -0.00001474, 0.00001181, -0.00001731, 0.00001328, -0.00001180, 0.00001153, -0.00000790, 0.00000942, -0.00000500, 0.00000677, -0.00001339, 0.00001115, -0.00001212, 0.00001133, -0.00001171, 0.00000907, -0.00000773, 0.00000720, -0.00000353, 0.00000625, -0.00000406, 0.00000438, -0.00000651, 0.00000507, -0.00000884, 0.00000669, -0.00000677, 0.00000524, -0.00000560, 0.00000310, -0.00000036, 0.00000263, -0.00000047, 0.00000196, -0.00000112, 0.00000217, -0.00000537, 0.00000373, -0.00000440, 0.00000304, -0.00000629, -0.00000054, 0.00000607, -0.00000189, },
		{ -0.00000193, 0.00000425, -0.00000179, 0.00000225, -0.00000221, -0.00000385, -0.00000122, -0.00000048, 0.00000179, 0.00000063, -0.00000058, 0.00000333, -0.00000569, 0.00000117, -0.00000528, 0.00001573, -0.00000598, 0.00000116, -0.00000517, -0.00000164, -0.00000660, 0.00000434, -0.00000650, 0.00001113, -0.00001020, 0.00001290, -0.00001044, 0.00001222, -0.00001670, -0.00000227, -0.00000872, 0.00002220, -0.00001769, 0.00001294, -0.00001966, 0.00002922, -0.00005609, 0.00004321, -0.00005388, 0.00004280, -0.00006741, 0.00004717, -0.00007608, 0.00004992, -0.00046677, 0.00079773, 0.00014605, -0.00035067, 0.00011639, -0.00015274, 0.00007288, -0.00009772, 0.00008178, -0.00005240, 0.00002022, -0.00003063, 0.00003010, -0.00002731, 0.00002586, -0.00003779, 0.00004050, -0.00002332, 0.00001990, -0.00004487, 0.00003381, -0.00002445, 0.00002270, -0.00002351, 0.00002469, -0.00001508, 0.00001342, -0.00002079, 0.00002115, -0.00002506, 0.00001870, -0.00002445, 0.00002265, -0.00000769, 0.00001182, -0.00001911, 0.00001558, -0.00001333, 0.00001465, -0.00001597, 0.00001737, -0.00001504, 0.00001216, -0.00001792, 0.00001395, -0.00001231, 0.00001208, -0.00000827, 0.00000962, -0.00000495, 0.00000641, -0.00001314, 0.00001106, -0.00001211, 0.00001141, -0.00001198, 0.00000946, -0.00000810, 0.00000761, -0.00000375, 0.00000631, -0.00000401, 0.00000414, -0.00000623, 0.00000482, -0.00000864, 0.00000665, -0.00000691, 0.00000542, -0.00000585, 0.00000338, -0.00000045, 0.00000261, -0.00000026, 0.00000160, -0.00000076, 0.00000177, -0.00000507, 0.00000359, -0.00000441, 0.00000304, -0.00000653, -0.00000002, 0.00000585, },
		{ 0.00000554, -0.00000187, 0.00000442, -0.00000210, 0.00000264, -0.00000296, -0.00000323, -0.00000154, -0.00000025, 0.00000180, 0.00000046, -0.00000034, 0.00000321, -0.00000601, 0.00000139, -0.00000489, 0.00001587, -0.00000687, 0.00000179, -0.00000570, -0.00000110, -0.00000719, 0.00000471, -0.00000630, 0.00001116, -0.00001056, 0.00001343, -0.00001116, 0.00001366, -0.00001902, -0.00000127, -0.00000796, 0.00002206, -0.00001863, 0.00001339, -0.00001960, 0.00003019, -0.00005841, 0.00004551, -0.00005601, 0.00004780, -0.00007410, 0.00005374, -0.00008503, 0.00006129, -0.00049334, 0.00084449, 0.00011318, -0.00034758, 0.00012104, -0.00015596, 0.00007739, -0.00010316, 0.00008758, -0.00005651, 0.00002302, -0.00003349, 0.00003220, -0.00002819, 0.00002634, -0.00003870, 0.00004155, -0.00002343, 0.00001932, -0.00004561, 0.00003566, -0.00002596, 0.00002404, -0.00002476, 0.00002569, -0.00001551, 0.00001364, -0.00002128, 0.00002143, -0.00002545, 0.00001930, -0.00002519, 0.00002371, -0.00000825, 0.00001190, -0.00001951, 0.00001604, -0.00001346, 0.00001470, -0.00001596, 0.00001737, -0.00001525, 0.00001242, -0.00001844, 0.00001456, -0.00001278, 0.00001262, -0.00000863, 0.00000984, -0.00000493, 0.00000603, -0.00001285, 0.00001090, -0.00001203, 0.00001141, -0.00001219, 0.00000981, -0.00000846, 0.00000803, -0.00000400, 0.00000640, -0.00000399, 0.00000390, -0.00000596, 0.00000456, -0.00000842, 0.00000658, -0.00000701, 0.00000559, -0.00000611, 0.00000369, -0.00000059, 0.00000266, -0.00000010, 0.00000128, -0.00000043, 0.00000137, -0.00000476, 0.00000344, -0.00000440, 0.00000304, -0.00000677, 0.00000054, },
		{ 0.00000114, 0.00000514, -0.00000170, 0.00000448, -0.00000231, 0.00000295, -0.00000364, -0.00000264, -0.00000182, -0.00000004, 0.00000182, 0.00000026, -0.00000004, 0.00000299, -0.00000624, 0.00000153, -0.00000437, 0.00001584, -0.00000764, 0.00000233, -0.00000614, -0.00000064, -0.00000770, 0.00000498, -0.00000598, 0.00001101, -0.00001075, 0.00001377, -0.00001168, 0.00001490, -0.00002123, -0.00000029, -0.00000709, 0.00002170, -0.00001934, 0.00001357, -0.00001919, 0.00003065, -0.00006013, 0.00004718, -0.00005745, 0.00005211, -0.00008025, 0.00005979, -0.00009344, 0.00007199, -0.00051910, 0.00089174, 0.00007786, -0.00034177, 0.00012375, -0.00015770, 0.00008071, -0.00010762, 0.00009267, -0.00006023, 0.00002559, -0.00003619, 0.00003419, -0.00002896, 0.00002664, -0.00003936, 0.00004238, -0.00002337, 0.00001848, -0.00004601, 0.00003722, -0.00002727, 0.00002522, -0.00002588, 0.00002658, -0.00001586, 0.00001377, -0.00002165, 0.00002158, -0.00002569, 0.00001974, -0.00002579, 0.00002468, -0.00000879, 0.00001193, -0.00001983, 0.00001644, -0.00001354, 0.00001468, -0.00001586, 0.00001728, -0.00001536, 0.00001257, -0.00001886, 0.00001511, -0.00001321, 0.00001314, -0.00000900, 0.00001008, -0.00000493, 0.00000564, -0.00001250, 0.00001068, -0.00001188, 0.00001135, -0.00001232, 0.00001012, -0.00000879, 0.00000843, -0.00000427, 0.00000652, -0.00000399, 0.00000369, -0.00000569, 0.00000429, -0.00000817, 0.00000647, -0.00000709, 0.00000573, -0.00000637, 0.00000402, -0.00000077, 0.00000275, -0.00000001, 0.00000101, -0.00000013, 0.00000099, -0.00000445, 0.00000328, -0.00000439, 0.00000302, -0.00000701, },
		{ -0.00000724, 0.00000179, 0.00000466, -0.00000144, 0.00000442, -0.00000242, 0.00000316, -0.00000426, -0.00000209, -0.00000208, 0.00000014, 0.00000185, 0.00000003, 0.00000033, 0.00000269, -0.00000637, 0.00000158, -0.00000373, 0.00001563, -0.00000826, 0.00000277, -0.00000647, -0.00000026, -0.00000812, 0.00000514, -0.00000551, 0.00001070, -0.00001076, 0.00001390, -0.00001196, 0.00001591, -0.00002329, 0.00000063, -0.00000611, 0.00002112, -0.00001979, 0.00001350, -0.00001840, 0.00003058, -0.00006122, 0.00004816, -0.00005813, 0.00005563, -0.00008573, 0.00006519, -0.00010116, 0.00008183, -0.00054377, 0.00093911, 0.00004033, -0.00033329, 0.00012449, -0.00015792, 0.00008277, -0.00011102, 0.00009697, -0.00006348, 0.00002784, -0.00003866, 0.00003605, -0.00002960, 0.00002674, -0.00003977, 0.00004299, -0.00002313, 0.00001739, -0.00004606, 0.00003848, -0.00002837, 0.00002623, -0.00002684, 0.00002736, -0.00001614, 0.00001379, -0.00002189, 0.00002160, -0.00002579, 0.00002002, -0.00002622, 0.00002555, -0.00000929, 0.00001192, -0.00002006, 0.00001677, -0.00001357, 0.00001459, -0.00001566, 0.00001708, -0.00001537, 0.00001262, -0.00001917, 0.00001558, -0.00001359, 0.00001362, -0.00000935, 0.00001033, -0.00000495, 0.00000526, -0.00001212, 0.00001041, -0.00001167, 0.00001121, -0.00001238, 0.00001037, -0.00000909, 0.00000883, -0.00000455, 0.00000666, -0.00000403, 0.00000350, -0.00000542, 0.00000401, -0.00000790, 0.00000632, -0.00000714, 0.00000586, -0.00000663, 0.00000438, -0.00000100, 0.00000290, 0.00000003, 0.00000080, 0.00000013, 0.00000063, -0.00000413, 0.00000311, -0.00000436, 0.00000299, },
		{ 0.00000294, -0.00000746, 0.00000247, 0.00000409, -0.00000108, 0.00000426, -0.00000241, 0.00000326, -0.00000478, -0.00000159, -0.00000229, 0.00000030, 0.00000189, -0.00000024, 0.00000076, 0.00000229, -0.00000640, 0.00000153, -0.00000295, 0.00001523, -0.00000871, 0.00000307, -0.00000669, 0.00000001, -0.00000843, 0.00000519, -0.00000490, 0.00001020, -0.00001057, 0.00001380, -0.00001199, 0.00001666, -0.00002518, 0.00000148, -0.00000502, 0.00002031, -0.00001999, 0.00001315, -0.00001725, 0.00002999, -0.00006164, 0.00004842, -0.00005799, 0.00005827, -0.00009043, 0.00006980, -0.00010804, 0.00009060, -0.00056705, 0.00098626, 0.00000085, -0.00032224, 0.00012326, -0.00015658, 0.00008352, -0.00011329, 0.00010039, -0.00006618, 0.00002973, -0.00004088, 0.00003774, -0.00003009, 0.00002664, -0.00003991, 0.00004337, -0.00002272, 0.00001605, -0.00004575, 0.00003942, -0.00002923, 0.00002704, -0.00002761, 0.00002799, -0.00001631, 0.00001371, -0.00002200, 0.00002147, -0.00002572, 0.00002014, -0.00002648, 0.00002629, -0.00000974, 0.00001184, -0.00002021, 0.00001704, -0.00001355, 0.00001442, -0.00001537, 0.00001677, -0.00001528, 0.00001256, -0.00001936, 0.00001596, -0.00001391, 0.00001405, -0.00000968, 0.00001058, -0.00000500, 0.00000487, -0.00001170, 0.00001008, -0.00001139, 0.00001099, -0.00001236, 0.00001055, -0.00000935, 0.00000922, -0.00000484, 0.00000682, -0.00000410, 0.00000334, -0.00000517, 0.00000373, -0.00000760, 0.00000614, -0.00000715, 0.00000597, -0.00000687, 0.00000475, -0.00000127, 0.00000311, -0.00000000, 0.00000064, 0.00000034, 0.00000030, -0.00000382, 0.00000293, -0.00000432, },
		{ -0.00000428, 0.00000289, -0.00000766, 0.00000317, 0.00000346, -0.00000062, 0.00000399, -0.00000229, 0.00000324, -0.00000521, -0.00000116, -0.00000245, 0.00000042, 0.00000196, -0.00000054, 0.00000126, 0.00000180, -0.00000633, 0.00000139, -0.00000205, 0.00001465, -0.00000900, 0.00000325, -0.00000679, 0.00000017, -0.00000862, 0.00000513, -0.00000414, 0.00000952, -0.00001017, 0.00001348, -0.00001176, 0.00001713, -0.00002686, 0.00000224, -0.00000381, 0.00001928, -0.00001993, 0.00001254, -0.00001574, 0.00002887, -0.00006139, 0.00004792, -0.00005699, 0.00005994, -0.00009422, 0.00007350, -0.00011392, 0.00009809, -0.00058864, 0.00103280, -0.00004026, -0.00030875, 0.00012009, -0.00015368, 0.00008294, -0.00011437, 0.00010286, -0.00006825, 0.00003120, -0.00004277, 0.00003921, -0.00003043, 0.00002634, -0.00003978, 0.00004351, -0.00002213, 0.00001448, -0.00004507, 0.00004001, -0.00002983, 0.00002764, -0.00002820, 0.00002847, -0.00001639, 0.00001351, -0.00002197, 0.00002120, -0.00002550, 0.00002009, -0.00002656, 0.00002691, -0.00001014, 0.00001171, -0.00002027, 0.00001723, -0.00001347, 0.00001419, -0.00001499, 0.00001636, -0.00001507, 0.00001238, -0.00001943, 0.00001625, -0.00001415, 0.00001443, -0.00000998, 0.00001084, -0.00000508, 0.00000450, -0.00001125, 0.00000971, -0.00001105, 0.00001070, -0.00001226, 0.00001068, -0.00000957, 0.00000957, -0.00000512, 0.00000700, -0.00000420, 0.00000321, -0.00000493, 0.00000345, -0.00000729, 0.00000593, -0.00000713, 0.00000605, -0.00000710, 0.00000514, -0.00000157, 0.00000336, -0.00000009, 0.00000056, 0.00000049, 0.00000000, -0.00000352, 0.00000276, },
		{ 0.00000260, -0.00000423, 0.00000282, -0.00000784, 0.00000389, 0.00000275, -0.00000007, 0.00000361, -0.00000206, 0.00000311, -0.00000553, -0.00000080, -0.00000254, 0.00000050, 0.00000205, -0.00000087, 0.00000182, 0.00000121, -0.00000616, 0.00000117, -0.00000102, 0.00001388, -0.00000911, 0.00000328, -0.00000676, 0.00000020, -0.00000870, 0.00000495, -0.00000324, 0.00000865, -0.00000957, 0.00001291, -0.00001126, 0.00001731, -0.00002831, 0.00000289, -0.00000248, 0.00001802, -0.00001960, 0.00001167, -0.00001388, 0.00002723, -0.00006046, 0.00004666, -0.00005509, 0.00006058, -0.00009701, 0.00007616, -0.00011867, 0.00010413, -0.00060824, 0.00107833, -0.00008269, -0.00029300, 0.00011504, -0.00014925, 0.00008101, -0.00011424, 0.00010432, -0.00006966, 0.00003219, -0.00004430, 0.00004044, -0.00003060, 0.00002582, -0.00003938, 0.00004340, -0.00002137, 0.00001267, -0.00004404, 0.00004026, -0.00003017, 0.00002801, -0.00002858, 0.00002879, -0.00001636, 0.00001320, -0.00002181, 0.00002079, -0.00002512, 0.00001986, -0.00002644, 0.00002738, -0.00001047, 0.00001151, -0.00002024, 0.00001734, -0.00001334, 0.00001390, -0.00001452, 0.00001584, -0.00001477, 0.00001209, -0.00001937, 0.00001643, -0.00001432, 0.00001475, -0.00001025, 0.00001109, -0.00000519, 0.00000414, -0.00001077, 0.00000929, -0.00001066, 0.00001034, -0.00001209, 0.00001074, -0.00000973, 0.00000990, -0.00000541, 0.00000720, -0.00000433, 0.00000311, -0.00000471, 0.00000318, -0.00000697, 0.00000569, -0.00000708, 0.00000610, -0.00000731, 0.00000553, -0.00000191, 0.00000366, -0.00000025, 0.00000054, 0.00000058, -0.00000025, -0.00000325, },
		{ -0.00000300, 0.00000244, -0.00000419, 0.00000274, -0.00000801, 0.00000462, 0.00000199, 0.00000056, 0.00000311, -0.00000171, 0.00000285, -0.00000574, -0.00000052, -0.00000257, 0.00000053, 0.00000218, -0.00000124, 0.00000243, 0.00000054, -0.00000589, 0.00000085, 0.00000013, 0.00001293, -0.00000904, 0.00000316, -0.00000660, 0.00000012, -0.00000865, 0.00000464, -0.00000219, 0.00000760, -0.00000876, 0.00001210, -0.00001048, 0.00001717, -0.00002951, 0.00000341, -0.00000102, 0.00001654, -0.00001901, 0.00001055, -0.00001170, 0.00002508, -0.00005886, 0.00004462, -0.00005228, 0.00006014, -0.00009871, 0.00007770, -0.00012218, 0.00010854, -0.00062558, 0.00112249, -0.00012607, -0.00027518, 0.00010821, -0.00014333, 0.00007776, -0.00011287, 0.00010474, -0.00007034, 0.00003266, -0.00004543, 0.00004140, -0.00003057, 0.00002509, -0.00003870, 0.00004305, -0.00002044, 0.00001065, -0.00004266, 0.00004016, -0.00003023, 0.00002815, -0.00002874, 0.00002894, -0.00001622, 0.00001277, -0.00002151, 0.00002023, -0.00002458, 0.00001945, -0.00002613, 0.00002769, -0.00001073, 0.00001125, -0.00002010, 0.00001738, -0.00001315, 0.00001354, -0.00001398, 0.00001523, -0.00001436, 0.00001168, -0.00001918, 0.00001650, -0.00001439, 0.00001500, -0.00001047, 0.00001133, -0.00000532, 0.00000380, -0.00001028, 0.00000884, -0.00001022, 0.00000992, -0.00001183, 0.00001072, -0.00000984, 0.00001019, -0.00000568, 0.00000740, -0.00000449, 0.00000305, -0.00000452, 0.00000292, -0.00000664, 0.00000543, -0.00000700, 0.00000613, -0.00000750, 0.00000592, -0.00000227, 0.00000401, -0.00000048, 0.00000059, 0.00000060, -0.00000045, },
		{ -0.00000059, -0.00000279, 0.00000231, -0.00000415, 0.00000265, -0.00000814, 0.00000535, 0.00000118, 0.00000128, 0.00000252, -0.00000124, 0.00000247, -0.00000583, -0.00000033, -0.00000252, 0.00000049, 0.00000233, -0.00000163, 0.00000310, -0.00000021, -0.00000552, 0.00000044, 0.00000139, 0.00001179, -0.00000878, 0.00000289, -0.00000629, -0.00000009, -0.00000847, 0.00000421, -0.00000101, 0.00000637, -0.00000774, 0.00001106, -0.00000942, 0.00001671, -0.00003043, 0.00000380, 0.00000056, 0.00001484, -0.00001818, 0.00000919, -0.00000921, 0.00002247, -0.00005661, 0.00004183, -0.00004856, 0.00005859, -0.00009927, 0.00007802, -0.00012432, 0.00011116, -0.00064038, 0.00116486, -0.00017004, -0.00025552, 0.00009972, -0.00013600, 0.00007322, -0.00011027, 0.00010410, -0.00007027, 0.00003258, -0.00004613, 0.00004206, -0.00003035, 0.00002415, -0.00003776, 0.00004247, -0.00001937, 0.00000843, -0.00004095, 0.00003971, -0.00003001, 0.00002804, -0.00002869, 0.00002890, -0.00001595, 0.00001222, -0.00002107, 0.00001953, -0.00002388, 0.00001887, -0.00002562, 0.00002784, -0.00001090, 0.00001091, -0.00001988, 0.00001733, -0.00001290, 0.00001313, -0.00001336, 0.00001452, -0.00001386, 0.00001116, -0.00001885, 0.00001646, -0.00001438, 0.00001517, -0.00001065, 0.00001155, -0.00000547, 0.00000347, -0.00000978, 0.00000836, -0.00000974, 0.00000943, -0.00001149, 0.00001063, -0.00000989, 0.00001043, -0.00000594, 0.00000762, -0.00000467, 0.00000301, -0.00000435, 0.00000268, -0.00000632, 0.00000514, -0.00000688, 0.00000613, -0.00000766, 0.00000631, -0.00000266, 0.00000440, -0.00000077, 0.00000071, 0.00000055, },
		{ 0.00000043, -0.00000067, -0.00000261, 0.00000219, -0.00000412, 0.00000255, -0.00000826, 0.00000607, 0.00000033, 0.00000206, 0.00000183, -0.00000067, 0.00000196, -0.00000580, -0.00000025, -0.00000238, 0.00000040, 0.00000253, -0.00000207, 0.00000381, -0.00000104, -0.00000505, -0.00000004, 0.00000276, 0.00001048, -0.00000834, 0.00000246, -0.00000583, -0.00000044, -0.00000816, 0.00000367, 0.00000030, 0.00000497, -0.00000652, 0.00000978, -0.00000809, 0.00001592, -0.00003105, 0.00000402, 0.00000225, 0.00001295, -0.00001710, 0.00000762, -0.00000645, 0.00001941, -0.00005373, 0.00003831, -0.00004395, 0.00005592, -0.00009862, 0.00007706, -0.00012502, 0.00011185, -0.00065239, 0.00120506, -0.00021421, -0.00023426, 0.00008971, -0.00012735, 0.00006746, -0.00010647, 0.00010238, -0.00006942, 0.00003194, -0.00004637, 0.00004240, -0.00002992, 0.00002301, -0.00003656, 0.00004165, -0.00001816, 0.00000603, -0.00003891, 0.00003891, -0.00002951, 0.00002769, -0.00002840, 0.00002869, -0.00001558, 0.00001157, -0.00002049, 0.00001869, -0.00002303, 0.00001812, -0.00002491, 0.00002783, -0.00001100, 0.00001050, -0.00001955, 0.00001721, -0.00001261, 0.00001266, -0.00001267, 0.00001374, -0.00001326, 0.00001054, -0.00001840, 0.00001629, -0.00001427, 0.00001526, -0.00001077, 0.00001176, -0.00000564, 0.00000317, -0.00000927, 0.00000785, -0.00000922, 0.00000889, -0.00001108, 0.00001048, -0.00000987, 0.00001063, -0.00000618, 0.00000783, -0.00000488, 0.00000301, -0.00000422, 0.00000246, -0.00000600, 0.00000484, -0.00000674, 0.00000609, -0.00000780, 0.00000669, -0.00000306, 0.00000483, -0.00000111, 0.00000090, },
		{ 0.00000116, 0.00000023, -0.00000068, -0.00000248, 0.00000211, -0.00000411, 0.00000245, -0.00000834, 0.00000678, -0.00000055, 0.00000290, 0.00000105, 0.00000001, 0.00000134, -0.00000564, -0.00000026, -0.00000216, 0.00000024, 0.00000276, -0.00000253, 0.00000456, -0.00000194, -0.00000450, -0.00000060, 0.00000422, 0.00000901, -0.00000772, 0.00000188, -0.00000524, -0.00000091, -0.00000772, 0.00000301, 0.00000174, 0.00000341, -0.00000512, 0.00000828, -0.00000648, 0.00001480, -0.00003137, 0.00000408, 0.00000407, 0.00001085, -0.00001579, 0.00000584, -0.00000346, 0.00001596, -0.00005028, 0.00003409, -0.00003848, 0.00005214, -0.00009675, 0.00007478, -0.00012421, 0.00011050, -0.00066139, 0.00124273, -0.00025821, -0.00021168, 0.00007834, -0.00011750, 0.00006057, -0.00010153, 0.00009960, -0.00006780, 0.00003070, -0.00004613, 0.00004240, -0.00002928, 0.00002166, -0.00003510, 0.00004061, -0.00001683, 0.00000349, -0.00003658, 0.00003778, -0.00002873, 0.00002710, -0.00002789, 0.00002829, -0.00001508, 0.00001081, -0.00001979, 0.00001772, -0.00002204, 0.00001721, -0.00002401, 0.00002764, -0.00001100, 0.00001002, -0.00001913, 0.00001700, -0.00001227, 0.00001215, -0.00001193, 0.00001287, -0.00001258, 0.00000981, -0.00001781, 0.00001601, -0.00001405, 0.00001526, -0.00001083, 0.00001194, -0.00000583, 0.00000289, -0.00000877, 0.00000733, -0.00000868, 0.00000830, -0.00001060, 0.00001025, -0.00000980, 0.00001078, -0.00000639, 0.00000804, -0.00000510, 0.00000304, -0.00000411, 0.00000227, -0.00000569, 0.00000454, -0.00000658, 0.00000603, -0.00000790, 0.00000705, -0.00000347, 0.00000528, -0.00000151, },
		{ -0.00000196, 0.00000149, -0.00000004, -0.00000063, -0.00000241, 0.00000206, -0.00000411, 0.00000234, -0.00000839, 0.00000747, -0.00000145, 0.00000379, 0.00000020, 0.00000079, 0.00000059, -0.00000536, -0.00000037, -0.00000185, 0.00000001, 0.00000303, -0.00000302, 0.00000535, -0.00000290, -0.00000386, -0.00000122, 0.00000576, 0.00000738, -0.00000693, 0.00000115, -0.00000451, -0.00000151, -0.00000716, 0.00000224, 0.00000328, 0.00000170, -0.00000354, 0.00000658, -0.00000462, 0.00001336, -0.00003139, 0.00000398, 0.00000599, 0.00000859, -0.00001427, 0.00000388, -0.00000027, 0.00001217, -0.00004630, 0.00002922, -0.00003220, 0.00004728, -0.00009366, 0.00007115, -0.00012185, 0.00010703, -0.00066717, 0.00127750, -0.00030164, -0.00018805, 0.00006581, -0.00010659, 0.00005263, -0.00009549, 0.00009578, -0.00006539, 0.00002887, -0.00004540, 0.00004204, -0.00002843, 0.00002013, -0.00003341, 0.00003937, -0.00001540, 0.00000082, -0.00003398, 0.00003633, -0.00002768, 0.00002627, -0.00002716, 0.00002770, -0.00001447, 0.00000995, -0.00001898, 0.00001663, -0.00002092, 0.00001615, -0.00002292, 0.00002728, -0.00001091, 0.00000947, -0.00001861, 0.00001672, -0.00001188, 0.00001160, -0.00001114, 0.00001195, -0.00001184, 0.00000900, -0.00001711, 0.00001562, -0.00001374, 0.00001517, -0.00001082, 0.00001208, -0.00000604, 0.00000265, -0.00000827, 0.00000681, -0.00000811, 0.00000766, -0.00001006, 0.00000995, -0.00000965, 0.00001087, -0.00000657, 0.00000824, -0.00000534, 0.00000311, -0.00000405, 0.00000210, -0.00000539, 0.00000422, -0.00000639, 0.00000594, -0.00000797, 0.00000739, -0.00000387, 0.00000576, },
		{ 0.00000626, -0.00000246, 0.00000188, -0.00000038, -0.00000050, -0.00000238, 0.00000204, -0.00000413, 0.00000223, -0.00000842, 0.00000814, -0.00000236, 0.00000472, -0.00000072, 0.00000165, -0.00000026, -0.00000495, -0.00000060, -0.00000145, -0.00000029, 0.00000335, -0.00000353, 0.00000616, -0.00000392, -0.00000316, -0.00000190, 0.00000736, 0.00000563, -0.00000597, 0.00000027, -0.00000364, -0.00000222, -0.00000648, 0.00000137, 0.00000491, -0.00000013, -0.00000180, 0.00000468, -0.00000253, 0.00001159, -0.00003108, 0.00000369, 0.00000801, 0.00000616, -0.00001255, 0.00000177, 0.00000308, 0.00000808, -0.00004185, 0.00002378, -0.00002518, 0.00004137, -0.00008935, 0.00006617, -0.00011792, 0.00010138, -0.00066957, 0.00130905, -0.00034411, -0.00016368, 0.00005231, -0.00009476, 0.00004378, -0.00008846, 0.00009098, -0.00006224, 0.00002647, -0.00004417, 0.00004133, -0.00002736, 0.00001842, -0.00003149, 0.00003792, -0.00001388, -0.00000193, -0.00003115, 0.00003459, -0.00002637, 0.00002521, -0.00002621, 0.00002694, -0.00001375, 0.00000900, -0.00001806, 0.00001544, -0.00001968, 0.00001494, -0.00002164, 0.00002676, -0.00001073, 0.00000885, -0.00001801, 0.00001636, -0.00001145, 0.00001102, -0.00001031, 0.00001097, -0.00001102, 0.00000810, -0.00001628, 0.00001510, -0.00001332, 0.00001498, -0.00001074, 0.00001219, -0.00000626, 0.00000243, -0.00000779, 0.00000628, -0.00000753, 0.00000700, -0.00000946, 0.00000959, -0.00000945, 0.00001090, -0.00000671, 0.00000842, -0.00000560, 0.00000320, -0.00000401, 0.00000197, -0.00000512, 0.00000391, -0.00000619, 0.00000582, -0.00000801, 0.00000771, -0.00000428, },
		{ -0.00000467, 0.00000676, -0.00000299, 0.00000233, -0.00000080, -0.00000031, -0.00000242, 0.00000207, -0.00000418, 0.00000213, -0.00000841, 0.00000876, -0.00000326, 0.00000567, -0.00000170, 0.00000258, -0.00000121, -0.00000443, -0.00000092, -0.00000096, -0.00000067, 0.00000370, -0.00000407, 0.00000700, -0.00000498, -0.00000238, -0.00000262, 0.00000901, 0.00000375, -0.00000487, -0.00000075, -0.00000265, -0.00000305, -0.00000569, 0.00000042, 0.00000662, -0.00000208, 0.00000008, 0.00000261, -0.00000021, 0.00000953, -0.00003046, 0.00000324, 0.00001013, 0.00000359, -0.00001066, -0.00000046, 0.00000654, 0.00000377, -0.00003700, 0.00001783, -0.00001749, 0.00003450, -0.00008387, 0.00005987, -0.00011243, 0.00009352, -0.00066845, 0.00133706, -0.00038524, -0.00013887, 0.00003806, -0.00008220, 0.00003414, -0.00008052, 0.00008526, -0.00005836, 0.00002349, -0.00004246, 0.00004027, -0.00002609, 0.00001655, -0.00002936, 0.00003629, -0.00001230, -0.00000474, -0.00002811, 0.00003257, -0.00002482, 0.00002394, -0.00002506, 0.00002600, -0.00001293, 0.00000796, -0.00001704, 0.00001414, -0.00001833, 0.00001362, -0.00002020, 0.00002606, -0.00001046, 0.00000817, -0.00001731, 0.00001593, -0.00001098, 0.00001042, -0.00000945, 0.00000995, -0.00001016, 0.00000713, -0.00001536, 0.00001448, -0.00001281, 0.00001470, -0.00001059, 0.00001226, -0.00000648, 0.00000224, -0.00000732, 0.00000576, -0.00000695, 0.00000632, -0.00000882, 0.00000916, -0.00000918, 0.00001088, -0.00000682, 0.00000859, -0.00000586, 0.00000333, -0.00000401, 0.00000188, -0.00000488, 0.00000360, -0.00000598, 0.00000568, -0.00000801, 0.00000800, },
		{ 0.00000826, -0.00000505, 0.00000726, -0.00000354, 0.00000282, -0.00000128, -0.00000004, -0.00000251, 0.00000213, -0.00000425, 0.00000204, -0.00000837, 0.00000935, -0.00000415, 0.00000663, -0.00000272, 0.00000358, -0.00000226, -0.00000379, -0.00000134, -0.00000038, -0.00000111, 0.00000411, -0.00000463, 0.00000785, -0.00000607, -0.00000155, -0.00000337, 0.00001069, 0.00000177, -0.00000362, -0.00000190, -0.00000154, -0.00000398, -0.00000479, -0.00000062, 0.00000839, -0.00000411, 0.00000207, 0.00000039, 0.00000229, 0.00000718, -0.00002954, 0.00000260, 0.00001232, 0.00000090, -0.00000861, -0.00000278, 0.00001007, -0.00000071, -0.00003182, 0.00001147, -0.00000923, 0.00002675, -0.00007729, 0.00005228, -0.00010541, 0.00008346, -0.00066370, 0.00136125, -0.00042466, -0.00011392, 0.00002328, -0.00006908, 0.00002388, -0.00007181, 0.00007869, -0.00005381, 0.00001999, -0.00004027, 0.00003884, -0.00002461, 0.00001454, -0.00002705, 0.00003451, -0.00001068, -0.00000757, -0.00002490, 0.00003031, -0.00002306, 0.00002247, -0.00002371, 0.00002490, -0.00001202, 0.00000686, -0.00001594, 0.00001277, -0.00001690, 0.00001218, -0.00001860, 0.00002521, -0.00001011, 0.00000742, -0.00001654, 0.00001543, -0.00001048, 0.00000979, -0.00000858, 0.00000890, -0.00000927, 0.00000611, -0.00001434, 0.00001376, -0.00001219, 0.00001433, -0.00001035, 0.00001227, -0.00000671, 0.00000209, -0.00000688, 0.00000526, -0.00000638, 0.00000562, -0.00000813, 0.00000869, -0.00000885, 0.00001079, -0.00000688, 0.00000873, -0.00000612, 0.00000348, -0.00000404, 0.00000182, -0.00000466, 0.00000330, -0.00000575, 0.00000551, -0.00000798, },
		{ -0.00000791, 0.00000848, -0.00000541, 0.00000776, -0.00000412, 0.00000336, -0.00000182, 0.00000030, -0.00000267, 0.00000225, -0.00000435, 0.00000195, -0.00000830, 0.00000990, -0.00000501, 0.00000759, -0.00000376, 0.00000463, -0.00000337, -0.00000305, -0.00000186, 0.00000029, -0.00000163, 0.00000455, -0.00000521, 0.00000871, -0.00000719, -0.00000068, -0.00000415, 0.00001239, -0.00000029, -0.00000225, -0.00000317, -0.00000032, -0.00000501, -0.00000381, -0.00000172, 0.00001020, -0.00000620, 0.00000416, -0.00000194, 0.00000496, 0.00000458, -0.00002832, 0.00000181, 0.00001457, -0.00000189, -0.00000644, -0.00000517, 0.00001360, -0.00000529, -0.00002639, 0.00000476, -0.00000050, 0.00001820, -0.00006967, 0.00004348, -0.00009693, 0.00007124, -0.00065528, 0.00138139, -0.00046204, -0.00008914, 0.00000822, -0.00005558, 0.00001314, -0.00006244, 0.00007136, -0.00004864, 0.00001598, -0.00003762, 0.00003707, -0.00002294, 0.00001240, -0.00002457, 0.00003258, -0.00000905, -0.00001039, -0.00002157, 0.00002784, -0.00002109, 0.00002082, -0.00002219, 0.00002365, -0.00001102, 0.00000570, -0.00001477, 0.00001134, -0.00001539, 0.00001065, -0.00001685, 0.00002421, -0.00000967, 0.00000663, -0.00001569, 0.00001486, -0.00000994, 0.00000916, -0.00000770, 0.00000784, -0.00000835, 0.00000505, -0.00001324, 0.00001295, -0.00001149, 0.00001386, -0.00001004, 0.00001224, -0.00000693, 0.00000196, -0.00000646, 0.00000478, -0.00000581, 0.00000492, -0.00000741, 0.00000817, -0.00000846, 0.00001065, -0.00000690, 0.00000885, -0.00000639, 0.00000365, -0.00000411, 0.00000181, -0.00000448, 0.00000302, -0.00000553, 0.00000533, },
		{ 0.00000513, -0.00000781, 0.00000866, -0.00000573, 0.00000824, -0.00000471, 0.00000394, -0.00000242, 0.00000071, -0.00000288, 0.00000241, -0.00000449, 0.00000188, -0.00000821, 0.00001039, -0.00000584, 0.00000854, -0.00000482, 0.00000571, -0.00000456, -0.00000222, -0.00000247, 0.00000104, -0.00000222, 0.00000503, -0.00000580, 0.00000957, -0.00000832, 0.00000023, -0.00000493, 0.00001408, -0.00000240, -0.00000078, -0.00000453, 0.00000098, -0.00000611, -0.00000274, -0.00000287, 0.00001203, -0.00000834, 0.00000631, -0.00000436, 0.00000775, 0.00000174, -0.00002681, 0.00000085, 0.00001687, -0.00000475, -0.00000416, -0.00000759, 0.00001711, -0.00000990, -0.00002079, -0.00000217, 0.00000858, 0.00000899, -0.00006113, 0.00003357, -0.00008707, 0.00005692, -0.00064316, 0.00139726, -0.00049705, -0.00006482, -0.00000692, -0.00004189, 0.00000209, -0.00005255, 0.00006339, -0.00004292, 0.00001154, -0.00003456, 0.00003498, -0.00002110, 0.00001015, -0.00002195, 0.00003052, -0.00000743, -0.00001316, -0.00001815, 0.00002519, -0.00001896, 0.00001901, -0.00002050, 0.00002226, -0.00000995, 0.00000450, -0.00001355, 0.00000986, -0.00001384, 0.00000905, -0.00001499, 0.00002306, -0.00000915, 0.00000578, -0.00001477, 0.00001423, -0.00000939, 0.00000853, -0.00000682, 0.00000677, -0.00000742, 0.00000396, -0.00001207, 0.00001206, -0.00001071, 0.00001331, -0.00000965, 0.00001215, -0.00000716, 0.00000187, -0.00000607, 0.00000432, -0.00000527, 0.00000422, -0.00000668, 0.00000761, -0.00000802, 0.00001045, -0.00000687, 0.00000893, -0.00000664, 0.00000384, -0.00000421, 0.00000183, -0.00000433, 0.00000276, -0.00000530, },
		{ -0.00000508, 0.00000492, -0.00000768, 0.00000881, -0.00000602, 0.00000870, -0.00000531, 0.00000454, -0.00000306, 0.00000118, -0.00000315, 0.00000261, -0.00000466, 0.00000182, -0.00000808, 0.00001083, -0.00000663, 0.00000947, -0.00000588, 0.00000681, -0.00000578, -0.00000131, -0.00000316, 0.00000186, -0.00000286, 0.00000554, -0.00000640, 0.00001042, -0.00000946, 0.00000116, -0.00000571, 0.00001575, -0.00000455, 0.00000078, -0.00000598, 0.00000237, -0.00000728, -0.00000161, -0.00000405, 0.00001386, -0.00001048, 0.00000850, -0.00000685, 0.00001065, -0.00000130, -0.00002504, -0.00000026, 0.00001920, -0.00000766, -0.00000181, -0.00001001, 0.00002054, -0.00001448, -0.00001512, -0.00000924, 0.00001789, -0.00000076, -0.00005177, 0.00002265, -0.00007595, 0.00004062, -0.00062736, 0.00140872, -0.00052939, -0.00004124, -0.00002189, -0.00002822, -0.00000909, -0.00004231, 0.00005487, -0.00003674, 0.00000670, -0.00003111, 0.00003258, -0.00001911, 0.00000782, -0.00001922, 0.00002837, -0.00000584, -0.00001584, -0.00001469, 0.00002240, -0.00001669, 0.00001706, -0.00001868, 0.00002075, -0.00000883, 0.00000327, -0.00001230, 0.00000834, -0.00001225, 0.00000741, -0.00001302, 0.00002179, -0.00000856, 0.00000490, -0.00001379, 0.00001355, -0.00000881, 0.00000789, -0.00000596, 0.00000572, -0.00000650, 0.00000286, -0.00001085, 0.00001109, -0.00000985, 0.00001268, -0.00000919, 0.00001200, -0.00000737, 0.00000181, -0.00000572, 0.00000390, -0.00000477, 0.00000355, -0.00000593, 0.00000702, -0.00000754, 0.00001019, -0.00000679, 0.00000898, -0.00000689, 0.00000404, -0.00000434, 0.00000190, -0.00000423, 0.00000252, },
		{ 0.00000231, -0.00000487, 0.00000470, -0.00000752, 0.00000891, -0.00000628, 0.00000913, -0.00000590, 0.00000516, -0.00000374, 0.00000170, -0.00000348, 0.00000287, -0.00000487, 0.00000178, -0.00000794, 0.00001121, -0.00000736, 0.00001035, -0.00000693, 0.00000791, -0.00000704, -0.00000033, -0.00000392, 0.00000275, -0.00000357, 0.00000609, -0.00000701, 0.00001125, -0.00001058, 0.00000211, -0.00000648, 0.00001737, -0.00000671, 0.00000240, -0.00000749, 0.00000381, -0.00000850, -0.00000043, -0.00000526, 0.00001567, -0.00001262, 0.00001070, -0.00000935, 0.00001360, -0.00000450, -0.00002304, -0.00000150, 0.00002153, -0.00001059, 0.00000059, -0.00001240, 0.00002384, -0.00001895, -0.00000945, -0.00001633, 0.00002731, -0.00001092, -0.00004173, 0.00001085, -0.00006370, 0.00002247, -0.00060796, 0.00141565, -0.00055880, -0.00001867, -0.00003649, -0.00001475, -0.00002023, -0.00003185, 0.00004593, -0.00003019, 0.00000154, -0.00002733, 0.00002990, -0.00001697, 0.00000543, -0.00001639, 0.00002614, -0.00000431, -0.00001840, -0.00001124, 0.00001951, -0.00001431, 0.00001501, -0.00001674, 0.00001913, -0.00000766, 0.00000203, -0.00001103, 0.00000682, -0.00001065, 0.00000573, -0.00001097, 0.00002041, -0.00000792, 0.00000398, -0.00001276, 0.00001282, -0.00000822, 0.00000728, -0.00000513, 0.00000469, -0.00000560, 0.00000177, -0.00000959, 0.00001007, -0.00000892, 0.00001197, -0.00000865, 0.00001180, -0.00000758, 0.00000177, -0.00000540, 0.00000351, -0.00000429, 0.00000289, -0.00000519, 0.00000640, -0.00000702, 0.00000988, -0.00000666, 0.00000900, -0.00000712, 0.00000426, -0.00000449, 0.00000200, -0.00000416, },
		{ -0.00000414, 0.00000214, -0.00000467, 0.00000448, -0.00000733, 0.00000898, -0.00000649, 0.00000953, -0.00000647, 0.00000578, -0.00000445, 0.00000228, -0.00000386, 0.00000317, -0.00000512, 0.00000176, -0.00000777, 0.00001153, -0.00000804, 0.00001119, -0.00000795, 0.00000901, -0.00000832, 0.00000070, -0.00000474, 0.00000370, -0.00000433, 0.00000667, -0.00000762, 0.00001205, -0.00001168, 0.00000305, -0.00000722, 0.00001893, -0.00000886, 0.00000406, -0.00000906, 0.00000530, -0.00000976, 0.00000079, -0.00000646, 0.00001743, -0.00001471, 0.00001288, -0.00001185, 0.00001657, -0.00000782, -0.00002081, -0.00000285, 0.00002385, -0.00001351, 0.00000302, -0.00001473, 0.00002698, -0.00002325, -0.00000387, -0.00002335, 0.00003670, -0.00002133, -0.00003117, -0.00000167, -0.00005049, 0.00000264, -0.00058505, 0.00141796, -0.00058505, 0.00000264, -0.00005049, -0.00000167, -0.00003117, -0.00002133, 0.00003670, -0.00002335, -0.00000387, -0.00002325, 0.00002698, -0.00001473, 0.00000302, -0.00001351, 0.00002385, -0.00000285, -0.00002081, -0.00000782, 0.00001657, -0.00001185, 0.00001288, -0.00001471, 0.00001743, -0.00000646, 0.00000079, -0.00000976, 0.00000530, -0.00000906, 0.00000406, -0.00000886, 0.00001893, -0.00000722, 0.00000305, -0.00001168, 0.00001205, -0.00000762, 0.00000667, -0.00000433, 0.00000370, -0.00000474, 0.00000070, -0.00000832, 0.00000901, -0.00000795, 0.00001119, -0.00000804, 0.00001153, -0.00000777, 0.00000176, -0.00000512, 0.00000317, -0.00000386, 0.00000228, -0.00000445, 0.00000578, -0.00000647, 0.00000953, -0.00000649, 0.00000898, -0.00000733, 0.00000448, -0.00000467, 0.00000214, },
		{ 0.00000231, -0.00000416, 0.00000200, -0.00000449, 0.00000426, -0.00000712, 0.00000900, -0.00000666, 0.00000988, -0.00000702, 0.00000640, -0.00000519, 0.00000289, -0.00000429, 0.00000351, -0.00000540, 0.00000177, -0.00000758, 0.00001180, -0.00000865, 0.00001197, -0.00000892, 0.00001007, -0.00000959, 0.00000177, -0.00000560, 0.00000469, -0.00000513, 0.00000728, -0.00000822, 0.00001282, -0.00001276, 0.00000398, -0.00000792, 0.00002041, -0.00001097, 0.00000573, -0.00001065, 0.00000682, -0.00001103, 0.00000203, -0.00000766, 0.00001913, -0.00001674, 0.00001501, -0.00001431, 0.00001951, -0.00001124, -0.00001840, -0.00000431, 0.00002614, -0.00001639, 0.00000543, -0.00001697, 0.00002990, -0.00002733, 0.00000154, -0.00003019, 0.00004593, -0.00003185, -0.00002023, -0.00001475, -0.00003649, -0.00001867, -0.00055880, 0.00141565, -0.00060796, 0.00002247, -0.00006370, 0.00001085, -0.00004173, -0.00001092, 0.00002731, -0.00001633, -0.00000945, -0.00001895, 0.00002384, -0.00001240, 0.00000059, -0.00001059, 0.00002153, -0.00000150, -0.00002304, -0.00000450, 0.00001360, -0.00000935, 0.00001070, -0.00001262, 0.00001567, -0.00000526, -0.00000043, -0.00000850, 0.00000381, -0.00000749, 0.00000240, -0.00000671, 0.00001737, -0.00000648, 0.00000211, -0.00001058, 0.00001125, -0.00000701, 0.00000609, -0.00000357, 0.00000275, -0.00000392, -0.00000033, -0.00000704, 0.00000791, -0.00000693, 0.00001035, -0.00000736, 0.00001121, -0.00000794, 0.00000178, -0.00000487, 0.00000287, -0.00000348, 0.00000170, -0.00000374, 0.00000516, -0.00000590, 0.00000913, -0.00000628, 0.00000891, -0.00000752, 0.00000470, -0.00000487, },
		{ -0.00000508, 0.00000252, -0.00000423, 0.00000190, -0.00000434, 0.00000404, -0.00000689, 0.00000898, -0.00000679, 0.00001019, -0.00000754, 0.00000702, -0.00000593, 0.00000355, -0.00000477, 0.00000390, -0.00000572, 0.00000181, -0.00000737, 0.00001200, -0.00000919, 0.00001268, -0.00000985, 0.00001109, -0.00001085, 0.00000286, -0.00000650, 0.00000572, -0.00000596, 0.00000789, -0.00000881, 0.00001355, -0.00001379, 0.00000490, -0.00000856, 0.00002179, -0.00001302, 0.00000741, -0.00001225, 0.00000834, -0.00001230, 0.00000327, -0.00000883, 0.00002075, -0.00001868, 0.00001706, -0.00001669, 0.00002240, -0.00001469, -0.00001584, -0.00000584, 0.00002837, -0.00001922, 0.00000782, -0.00001911, 0.00003258, -0.00003111, 0.00000670, -0.00003674, 0.00005487, -0.00004231, -0.00000909, -0.00002822, -0.00002189, -0.00004124, -0.00052939, 0.00140872, -0.00062736, 0.00004062, -0.00007595, 0.00002265, -0.00005177, -0.00000076, 0.00001789, -0.00000924, -0.00001512, -0.00001448, 0.00002054, -0.00001001, -0.00000181, -0.00000766, 0.00001920, -0.00000026, -0.00002504, -0.00000130, 0.00001065, -0.00000685, 0.00000850, -0.00001048, 0.00001386, -0.00000405, -0.00000161, -0.00000728, 0.00000237, -0.00000598, 0.00000078, -0.00000455, 0.00001575, -0.00000571, 0.00000116, -0.00000946, 0.00001042, -0.00000640, 0.00000554, -0.00000286, 0.00000186, -0.00000316, -0.00000131, -0.00000578, 0.00000681, -0.00000588, 0.00000947, -0.00000663, 0.00001083, -0.00000808, 0.00000182, -0.00000466, 0.00000261, -0.00000315, 0.00000118, -0.00000306, 0.00000454, -0.00000531, 0.00000870, -0.00000602, 0.00000881, -0.00000768, 0.00000492, },
		{ 0.00000513, -0.00000530, 0.00000276, -0.00000433, 0.00000183, -0.00000421, 0.00000384, -0.00000664, 0.00000893, -0.00000687, 0.00001045, -0.00000802, 0.00000761, -0.00000668, 0.00000422, -0.00000527, 0.00000432, -0.00000607, 0.00000187, -0.00000716, 0.00001215, -0.00000965, 0.00001331, -0.00001071, 0.00001206, -0.00001207, 0.00000396, -0.00000742, 0.00000677, -0.00000682, 0.00000853, -0.00000939, 0.00001423, -0.00001477, 0.00000578, -0.00000915, 0.00002306, -0.00001499, 0.00000905, -0.00001384, 0.00000986, -0.00001355, 0.00000450, -0.00000995, 0.00002226, -0.00002050, 0.00001901, -0.00001896, 0.00002519, -0.00001815, -0.00001316, -0.00000743, 0.00003052, -0.00002195, 0.00001015, -0.00002110, 0.00003498, -0.00003456, 0.00001154, -0.00004292, 0.00006339, -0.00005255, 0.00000209, -0.00004189, -0.00000692, -0.00006482, -0.00049705, 0.00139726, -0.00064316, 0.00005692, -0.00008707, 0.00003357, -0.00006113, 0.00000899, 0.00000858, -0.00000217, -0.00002079, -0.00000990, 0.00001711, -0.00000759, -0.00000416, -0.00000475, 0.00001687, 0.00000085, -0.00002681, 0.00000174, 0.00000775, -0.00000436, 0.00000631, -0.00000834, 0.00001203, -0.00000287, -0.00000274, -0.00000611, 0.00000098, -0.00000453, -0.00000078, -0.00000240, 0.00001408, -0.00000493, 0.00000023, -0.00000832, 0.00000957, -0.00000580, 0.00000503, -0.00000222, 0.00000104, -0.00000247, -0.00000222, -0.00000456, 0.00000571, -0.00000482, 0.00000854, -0.00000584, 0.00001039, -0.00000821, 0.00000188, -0.00000449, 0.00000241, -0.00000288, 0.00000071, -0.00000242, 0.00000394, -0.00000471, 0.00000824, -0.00000573, 0.00000866, -0.00000781, },
		{ -0.00000791, 0.00000533, -0.00000553, 0.00000302, -0.00000448, 0.00000181, -0.00000411, 0.00000365, -0.00000639, 0.00000885, -0.00000690, 0.00001065, -0.00000846, 0.00000817, -0.00000741, 0.00000492, -0.00000581, 0.00000478, -0.00000646, 0.00000196, -0.00000693, 0.00001224, -0.00001004, 0.00001386, -0.00001149, 0.00001295, -0.00001324, 0.00000505, -0.00000835, 0.00000784, -0.00000770, 0.00000916, -0.00000994, 0.00001486, -0.00001569, 0.00000663, -0.00000967, 0.00002421, -0.00001685, 0.00001065, -0.00001539, 0.00001134, -0.00001477, 0.00000570, -0.00001102, 0.00002365, -0.00002219, 0.00002082, -0.00002109, 0.00002784, -0.00002157, -0.00001039, -0.00000905, 0.00003258, -0.00002457, 0.00001240, -0.00002294, 0.00003707, -0.00003762, 0.00001598, -0.00004864, 0.00007136, -0.00006244, 0.00001314, -0.00005558, 0.00000822, -0.00008914, -0.00046204, 0.00138139, -0.00065528, 0.00007124, -0.00009693, 0.00004348, -0.00006967, 0.00001820, -0.00000050, 0.00000476, -0.00002639, -0.00000529, 0.00001360, -0.00000517, -0.00000644, -0.00000189, 0.00001457, 0.00000181, -0.00002832, 0.00000458, 0.00000496, -0.00000194, 0.00000416, -0.00000620, 0.00001020, -0.00000172, -0.00000381, -0.00000501, -0.00000032, -0.00000317, -0.00000225, -0.00000029, 0.00001239, -0.00000415, -0.00000068, -0.00000719, 0.00000871, -0.00000521, 0.00000455, -0.00000163, 0.00000029, -0.00000186, -0.00000305, -0.00000337, 0.00000463, -0.00000376, 0.00000759, -0.00000501, 0.00000990, -0.00000830, 0.00000195, -0.00000435, 0.00000225, -0.00000267, 0.00000030, -0.00000182, 0.00000336, -0.00000412, 0.00000776, -0.00000541, 0.00000848, },
		{ 0.00000826, -0.00000798, 0.00000551, -0.00000575, 0.00000330, -0.00000466, 0.00000182, -0.00000404, 0.00000348, -0.00000612, 0.00000873, -0.00000688, 0.00001079, -0.00000885, 0.00000869, -0.00000813, 0.00000562, -0.00000638, 0.00000526, -0.00000688, 0.00000209, -0.00000671, 0.00001227, -0.00001035, 0.00001433, -0.00001219, 0.00001376, -0.00001434, 0.00000611, -0.00000927, 0.00000890, -0.00000858, 0.00000979, -0.00001048, 0.00001543, -0.00001654, 0.00000742, -0.00001011, 0.00002521, -0.00001860, 0.00001218, -0.00001690, 0.00001277, -0.00001594, 0.00000686, -0.00001202, 0.00002490, -0.00002371, 0.00002247, -0.00002306, 0.00003031, -0.00002490, -0.00000757, -0.00001068, 0.00003451, -0.00002705, 0.00001454, -0.00002461, 0.00003884, -0.00004027, 0.00001999, -0.00005381, 0.00007869, -0.00007181, 0.00002388, -0.00006908, 0.00002328, -0.00011392, -0.00042466, 0.00136125, -0.00066370, 0.00008346, -0.00010541, 0.00005228, -0.00007729, 0.00002675, -0.00000923, 0.00001147, -0.00003182, -0.00000071, 0.00001007, -0.00000278, -0.00000861, 0.00000090, 0.00001232, 0.00000260, -0.00002954, 0.00000718, 0.00000229, 0.00000039, 0.00000207, -0.00000411, 0.00000839, -0.00000062, -0.00000479, -0.00000398, -0.00000154, -0.00000190, -0.00000362, 0.00000177, 0.00001069, -0.00000337, -0.00000155, -0.00000607, 0.00000785, -0.00000463, 0.00000411, -0.00000111, -0.00000038, -0.00000134, -0.00000379, -0.00000226, 0.00000358, -0.00000272, 0.00000663, -0.00000415, 0.00000935, -0.00000837, 0.00000204, -0.00000425, 0.00000213, -0.00000251, -0.00000004, -0.00000128, 0.00000282, -0.00000354, 0.00000726, -0.00000505, },
		{ -0.00000467, 0.00000800, -0.00000801, 0.00000568, -0.00000598, 0.00000360, -0.00000488, 0.00000188, -0.00000401, 0.00000333, -0.00000586, 0.00000859, -0.00000682, 0.00001088, -0.00000918, 0.00000916, -0.00000882, 0.00000632, -0.00000695, 0.00000576, -0.00000732, 0.00000224, -0.00000648, 0.00001226, -0.00001059, 0.00001470, -0.00001281, 0.00001448, -0.00001536, 0.00000713, -0.00001016, 0.00000995, -0.00000945, 0.00001042, -0.00001098, 0.00001593, -0.00001731, 0.00000817, -0.00001046, 0.00002606, -0.00002020, 0.00001362, -0.00001833, 0.00001414, -0.00001704, 0.00000796, -0.00001293, 0.00002600, -0.00002506, 0.00002394, -0.00002482, 0.00003257, -0.00002811, -0.00000474, -0.00001230, 0.00003629, -0.00002936, 0.00001655, -0.00002609, 0.00004027, -0.00004246, 0.00002349, -0.00005836, 0.00008526, -0.00008052, 0.00003414, -0.00008220, 0.00003806, -0.00013887, -0.00038524, 0.00133706, -0.00066845, 0.00009352, -0.00011243, 0.00005987, -0.00008387, 0.00003450, -0.00001749, 0.00001783, -0.00003700, 0.00000377, 0.00000654, -0.00000046, -0.00001066, 0.00000359, 0.00001013, 0.00000324, -0.00003046, 0.00000953, -0.00000021, 0.00000261, 0.00000008, -0.00000208, 0.00000662, 0.00000042, -0.00000569, -0.00000305, -0.00000265, -0.00000075, -0.00000487, 0.00000375, 0.00000901, -0.00000262, -0.00000238, -0.00000498, 0.00000700, -0.00000407, 0.00000370, -0.00000067, -0.00000096, -0.00000092, -0.00000443, -0.00000121, 0.00000258, -0.00000170, 0.00000567, -0.00000326, 0.00000876, -0.00000841, 0.00000213, -0.00000418, 0.00000207, -0.00000242, -0.00000031, -0.00000080, 0.00000233, -0.00000299, 0.00000676, },
		{ 0.00000626, -0.00000428, 0.00000771, -0.00000801, 0.00000582, -0.00000619, 0.00000391, -0.00000512, 0.00000197, -0.00000401, 0.00000320, -0.00000560, 0.00000842, -0.00000671, 0.00001090, -0.00000945, 0.00000959, -0.00000946, 0.00000700, -0.00000753, 0.00000628, -0.00000779, 0.00000243, -0.00000626, 0.00001219, -0.00001074, 0.00001498, -0.00001332, 0.00001510, -0.00001628, 0.00000810, -0.00001102, 0.00001097, -0.00001031, 0.00001102, -0.00001145, 0.00001636, -0.00001801, 0.00000885, -0.00001073, 0.00002676, -0.00002164, 0.00001494, -0.00001968, 0.00001544, -0.00001806, 0.00000900, -0.00001375, 0.00002694, -0.00002621, 0.00002521, -0.00002637, 0.00003459, -0.00003115, -0.00000193, -0.00001388, 0.00003792, -0.00003149, 0.00001842, -0.00002736, 0.00004133, -0.00004417, 0.00002647, -0.00006224, 0.00009098, -0.00008846, 0.00004378, -0.00009476, 0.00005231, -0.00016368, -0.00034411, 0.00130905, -0.00066957, 0.00010138, -0.00011792, 0.00006617, -0.00008935, 0.00004137, -0.00002518, 0.00002378, -0.00004185, 0.00000808, 0.00000308, 0.00000177, -0.00001255, 0.00000616, 0.00000801, 0.00000369, -0.00003108, 0.00001159, -0.00000253, 0.00000468, -0.00000180, -0.00000013, 0.00000491, 0.00000137, -0.00000648, -0.00000222, -0.00000364, 0.00000027, -0.00000597, 0.00000563, 0.00000736, -0.00000190, -0.00000316, -0.00000392, 0.00000616, -0.00000353, 0.00000335, -0.00000029, -0.00000145, -0.00000060, -0.00000495, -0.00000026, 0.00000165, -0.00000072, 0.00000472, -0.00000236, 0.00000814, -0.00000842, 0.00000223, -0.00000413, 0.00000204, -0.00000238, -0.00000050, -0.00000038, 0.00000188, -0.00000246, },
		{ -0.00000196, 0.00000576, -0.00000387, 0.00000739, -0.00000797, 0.00000594, -0.00000639, 0.00000422, -0.00000539, 0.00000210, -0.00000405, 0.00000311, -0.00000534, 0.00000824, -0.00000657, 0.00001087, -0.00000965, 0.00000995, -0.00001006, 0.00000766, -0.00000811, 0.00000681, -0.00000827, 0.00000265, -0.00000604, 0.00001208, -0.00001082, 0.00001517, -0.00001374, 0.00001562, -0.00001711, 0.00000900, -0.00001184, 0.00001195, -0.00001114, 0.00001160, -0.00001188, 0.00001672, -0.00001861, 0.00000947, -0.00001091, 0.00002728, -0.00002292, 0.00001615, -0.00002092, 0.00001663, -0.00001898, 0.00000995, -0.00001447, 0.00002770, -0.00002716, 0.00002627, -0.00002768, 0.00003633, -0.00003398, 0.00000082, -0.00001540, 0.00003937, -0.00003341, 0.00002013, -0.00002843, 0.00004204, -0.00004540, 0.00002887, -0.00006539, 0.00009578, -0.00009549, 0.00005263, -0.00010659, 0.00006581, -0.00018805, -0.00030164, 0.00127750, -0.00066717, 0.00010703, -0.00012185, 0.00007115, -0.00009366, 0.00004728, -0.00003220, 0.00002922, -0.00004630, 0.00001217, -0.00000027, 0.00000388, -0.00001427, 0.00000859, 0.00000599, 0.00000398, -0.00003139, 0.00001336, -0.00000462, 0.00000658, -0.00000354, 0.00000170, 0.00000328, 0.00000224, -0.00000716, -0.00000151, -0.00000451, 0.00000115, -0.00000693, 0.00000738, 0.00000576, -0.00000122, -0.00000386, -0.00000290, 0.00000535, -0.00000302, 0.00000303, 0.00000001, -0.00000185, -0.00000037, -0.00000536, 0.00000059, 0.00000079, 0.00000020, 0.00000379, -0.00000145, 0.00000747, -0.00000839, 0.00000234, -0.00000411, 0.00000206, -0.00000241, -0.00000063, -0.00000004, 0.00000149, },
		{ 0.00000116, -0.00000151, 0.00000528, -0.00000347, 0.00000705, -0.00000790, 0.00000603, -0.00000658, 0.00000454, -0.00000569, 0.00000227, -0.00000411, 0.00000304, -0.00000510, 0.00000804, -0.00000639, 0.00001078, -0.00000980, 0.00001025, -0.00001060, 0.00000830, -0.00000868, 0.00000733, -0.00000877, 0.00000289, -0.00000583, 0.00001194, -0.00001083, 0.00001526, -0.00001405, 0.00001601, -0.00001781, 0.00000981, -0.00001258, 0.00001287, -0.00001193, 0.00001215, -0.00001227, 0.00001700, -0.00001913, 0.00001002, -0.00001100, 0.00002764, -0.00002401, 0.00001721, -0.00002204, 0.00001772, -0.00001979, 0.00001081, -0.00001508, 0.00002829, -0.00002789, 0.00002710, -0.00002873, 0.00003778, -0.00003658, 0.00000349, -0.00001683, 0.00004061, -0.00003510, 0.00002166, -0.00002928, 0.00004240, -0.00004613, 0.00003070, -0.00006780, 0.00009960, -0.00010153, 0.00006057, -0.00011750, 0.00007834, -0.00021168, -0.00025821, 0.00124273, -0.00066139, 0.00011050, -0.00012421, 0.00007478, -0.00009675, 0.00005214, -0.00003848, 0.00003409, -0.00005028, 0.00001596, -0.00000346, 0.00000584, -0.00001579, 0.00001085, 0.00000407, 0.00000408, -0.00003137, 0.00001480, -0.00000648, 0.00000828, -0.00000512, 0.00000341, 0.00000174, 0.00000301, -0.00000772, -0.00000091, -0.00000524, 0.00000188, -0.00000772, 0.00000901, 0.00000422, -0.00000060, -0.00000450, -0.00000194, 0.00000456, -0.00000253, 0.00000276, 0.00000024, -0.00000216, -0.00000026, -0.00000564, 0.00000134, 0.00000001, 0.00000105, 0.00000290, -0.00000055, 0.00000678, -0.00000834, 0.00000245, -0.00000411, 0.00000211, -0.00000248, -0.00000068, 0.00000023, },
		{ 0.00000043, 0.00000090, -0.00000111, 0.00000483, -0.00000306, 0.00000669, -0.00000780, 0.00000609, -0.00000674, 0.00000484, -0.00000600, 0.00000246, -0.00000422, 0.00000301, -0.00000488, 0.00000783, -0.00000618, 0.00001063, -0.00000987, 0.00001048, -0.00001108, 0.00000889, -0.00000922, 0.00000785, -0.00000927, 0.00000317, -0.00000564, 0.00001176, -0.00001077, 0.00001526, -0.00001427, 0.00001629, -0.00001840, 0.00001054, -0.00001326, 0.00001374, -0.00001267, 0.00001266, -0.00001261, 0.00001721, -0.00001955, 0.00001050, -0.00001100, 0.00002783, -0.00002491, 0.00001812, -0.00002303, 0.00001869, -0.00002049, 0.00001157, -0.00001558, 0.00002869, -0.00002840, 0.00002769, -0.00002951, 0.00003891, -0.00003891, 0.00000603, -0.00001816, 0.00004165, -0.00003656, 0.00002301, -0.00002992, 0.00004240, -0.00004637, 0.00003194, -0.00006942, 0.00010238, -0.00010647, 0.00006746, -0.00012735, 0.00008971, -0.00023426, -0.00021421, 0.00120506, -0.00065239, 0.00011185, -0.00012502, 0.00007706, -0.00009862, 0.00005592, -0.00004395, 0.00003831, -0.00005373, 0.00001941, -0.00000645, 0.00000762, -0.00001710, 0.00001295, 0.00000225, 0.00000402, -0.00003105, 0.00001592, -0.00000809, 0.00000978, -0.00000652, 0.00000497, 0.00000030, 0.00000367, -0.00000816, -0.00000044, -0.00000583, 0.00000246, -0.00000834, 0.00001048, 0.00000276, -0.00000004, -0.00000505, -0.00000104, 0.00000381, -0.00000207, 0.00000253, 0.00000040, -0.00000238, -0.00000025, -0.00000580, 0.00000196, -0.00000067, 0.00000183, 0.00000206, 0.00000033, 0.00000607, -0.00000826, 0.00000255, -0.00000412, 0.00000219, -0.00000261, -0.00000067, },
		{ -0.00000059, 0.00000055, 0.00000071, -0.00000077, 0.00000440, -0.00000266, 0.00000631, -0.00000766, 0.00000613, -0.00000688, 0.00000514, -0.00000632, 0.00000268, -0.00000435, 0.00000301, -0.00000467, 0.00000762, -0.00000594, 0.00001043, -0.00000989, 0.00001063, -0.00001149, 0.00000943, -0.00000974, 0.00000836, -0.00000978, 0.00000347, -0.00000547, 0.00001155, -0.00001065, 0.00001517, -0.00001438, 0.00001646, -0.00001885, 0.00001116, -0.00001386, 0.00001452, -0.00001336, 0.00001313, -0.00001290, 0.00001733, -0.00001988, 0.00001091, -0.00001090, 0.00002784, -0.00002562, 0.00001887, -0.00002388, 0.00001953, -0.00002107, 0.00001222, -0.00001595, 0.00002890, -0.00002869, 0.00002804, -0.00003001, 0.00003971, -0.00004095, 0.00000843, -0.00001937, 0.00004247, -0.00003776, 0.00002415, -0.00003035, 0.00004206, -0.00004613, 0.00003258, -0.00007027, 0.00010410, -0.00011027, 0.00007322, -0.00013600, 0.00009972, -0.00025552, -0.00017004, 0.00116486, -0.00064038, 0.00011116, -0.00012432, 0.00007802, -0.00009927, 0.00005859, -0.00004856, 0.00004183, -0.00005661, 0.00002247, -0.00000921, 0.00000919, -0.00001818, 0.00001484, 0.00000056, 0.00000380, -0.00003043, 0.00001671, -0.00000942, 0.00001106, -0.00000774, 0.00000637, -0.00000101, 0.00000421, -0.00000847, -0.00000009, -0.00000629, 0.00000289, -0.00000878, 0.00001179, 0.00000139, 0.00000044, -0.00000552, -0.00000021, 0.00000310, -0.00000163, 0.00000233, 0.00000049, -0.00000252, -0.00000033, -0.00000583, 0.00000247, -0.00000124, 0.00000252, 0.00000128, 0.00000118, 0.00000535, -0.00000814, 0.00000265, -0.00000415, 0.00000231, -0.00000279, },
		{ -0.00000300, -0.00000045, 0.00000060, 0.00000059, -0.00000048, 0.00000401, -0.00000227, 0.00000592, -0.00000750, 0.00000613, -0.00000700, 0.00000543, -0.00000664, 0.00000292, -0.00000452, 0.00000305, -0.00000449, 0.00000740, -0.00000568, 0.00001019, -0.00000984, 0.00001072, -0.00001183, 0.00000992, -0.00001022, 0.00000884, -0.00001028, 0.00000380, -0.00000532, 0.00001133, -0.00001047, 0.00001500, -0.00001439, 0.00001650, -0.00001918, 0.00001168, -0.00001436, 0.00001523, -0.00001398, 0.00001354, -0.00001315, 0.00001738, -0.00002010, 0.00001125, -0.00001073, 0.00002769, -0.00002613, 0.00001945, -0.00002458, 0.00002023, -0.00002151, 0.00001277, -0.00001622, 0.00002894, -0.00002874, 0.00002815, -0.00003023, 0.00004016, -0.00004266, 0.00001065, -0.00002044, 0.00004305, -0.00003870, 0.00002509, -0.00003057, 0.00004140, -0.00004543, 0.00003266, -0.00007034, 0.00010474, -0.00011287, 0.00007776, -0.00014333, 0.00010821, -0.00027518, -0.00012607, 0.00112249, -0.00062558, 0.00010854, -0.00012218, 0.00007770, -0.00009871, 0.00006014, -0.00005228, 0.00004462, -0.00005886, 0.00002508, -0.00001170, 0.00001055, -0.00001901, 0.00001654, -0.00000102, 0.00000341, -0.00002951, 0.00001717, -0.00001048, 0.00001210, -0.00000876, 0.00000760, -0.00000219, 0.00000464, -0.00000865, 0.00000012, -0.00000660, 0.00000316, -0.00000904, 0.00001293, 0.00000013, 0.00000085, -0.00000589, 0.00000054, 0.00000243, -0.00000124, 0.00000218, 0.00000053, -0.00000257, -0.00000052, -0.00000574, 0.00000285, -0.00000171, 0.00000311, 0.00000056, 0.00000199, 0.00000462, -0.00000801, 0.00000274, -0.00000419, 0.00000244, },
		{ 0.00000260, -0.00000325, -0.00000025, 0.00000058, 0.00000054, -0.00000025, 0.00000366, -0.00000191, 0.00000553, -0.00000731, 0.00000610, -0.00000708, 0.00000569, -0.00000697, 0.00000318, -0.00000471, 0.00000311, -0.00000433, 0.00000720, -0.00000541, 0.00000990, -0.00000973, 0.00001074, -0.00001209, 0.00001034, -0.00001066, 0.00000929, -0.00001077, 0.00000414, -0.00000519, 0.00001109, -0.00001025, 0.00001475, -0.00001432, 0.00001643, -0.00001937, 0.00001209, -0.00001477, 0.00001584, -0.00001452, 0.00001390, -0.00001334, 0.00001734, -0.00002024, 0.00001151, -0.00001047, 0.00002738, -0.00002644, 0.00001986, -0.00002512, 0.00002079, -0.00002181, 0.00001320, -0.00001636, 0.00002879, -0.00002858, 0.00002801, -0.00003017, 0.00004026, -0.00004404, 0.00001267, -0.00002137, 0.00004340, -0.00003938, 0.00002582, -0.00003060, 0.00004044, -0.00004430, 0.00003219, -0.00006966, 0.00010432, -0.00011424, 0.00008101, -0.00014925, 0.00011504, -0.00029300, -0.00008269, 0.00107833, -0.00060824, 0.00010413, -0.00011867, 0.00007616, -0.00009701, 0.00006058, -0.00005509, 0.00004666, -0.00006046, 0.00002723, -0.00001388, 0.00001167, -0.00001960, 0.00001802, -0.00000248, 0.00000289, -0.00002831, 0.00001731, -0.00001126, 0.00001291, -0.00000957, 0.00000865, -0.00000324, 0.00000495, -0.00000870, 0.00000020, -0.00000676, 0.00000328, -0.00000911, 0.00001388, -0.00000102, 0.00000117, -0.00000616, 0.00000121, 0.00000182, -0.00000087, 0.00000205, 0.00000050, -0.00000254, -0.00000080, -0.00000553, 0.00000311, -0.00000206, 0.00000361, -0.00000007, 0.00000275, 0.00000389, -0.00000784, 0.00000282, -0.00000423, },
		{ -0.00000428, 0.00000276, -0.00000352, 0.00000000, 0.00000049, 0.00000056, -0.00000009, 0.00000336, -0.00000157, 0.00000514, -0.00000710, 0.00000605, -0.00000713, 0.00000593, -0.00000729, 0.00000345, -0.00000493, 0.00000321, -0.00000420, 0.00000700, -0.00000512, 0.00000957, -0.00000957, 0.00001068, -0.00001226, 0.00001070, -0.00001105, 0.00000971, -0.00001125, 0.00000450, -0.00000508, 0.00001084, -0.00000998, 0.00001443, -0.00001415, 0.00001625, -0.00001943, 0.00001238, -0.00001507, 0.00001636, -0.00001499, 0.00001419, -0.00001347, 0.00001723, -0.00002027, 0.00001171, -0.00001014, 0.00002691, -0.00002656, 0.00002009, -0.00002550, 0.00002120, -0.00002197, 0.00001351, -0.00001639, 0.00002847, -0.00002820, 0.00002764, -0.00002983, 0.00004001, -0.00004507, 0.00001448, -0.00002213, 0.00004351, -0.00003978, 0.00002634, -0.00003043, 0.00003921, -0.00004277, 0.00003120, -0.00006825, 0.00010286, -0.00011437, 0.00008294, -0.00015368, 0.00012009, -0.00030875, -0.00004026, 0.00103280, -0.00058864, 0.00009809, -0.00011392, 0.00007350, -0.00009422, 0.00005994, -0.00005699, 0.00004792, -0.00006139, 0.00002887, -0.00001574, 0.00001254, -0.00001993, 0.00001928, -0.00000381, 0.00000224, -0.00002686, 0.00001713, -0.00001176, 0.00001348, -0.00001017, 0.00000952, -0.00000414, 0.00000513, -0.00000862, 0.00000017, -0.00000679, 0.00000325, -0.00000900, 0.00001465, -0.00000205, 0.00000139, -0.00000633, 0.00000180, 0.00000126, -0.00000054, 0.00000196, 0.00000042, -0.00000245, -0.00000116, -0.00000521, 0.00000324, -0.00000229, 0.00000399, -0.00000062, 0.00000346, 0.00000317, -0.00000766, 0.00000289, },
		{ 0.00000294, -0.00000432, 0.00000293, -0.00000382, 0.00000030, 0.00000034, 0.00000064, -0.00000000, 0.00000311, -0.00000127, 0.00000475, -0.00000687, 0.00000597, -0.00000715, 0.00000614, -0.00000760, 0.00000373, -0.00000517, 0.00000334, -0.00000410, 0.00000682, -0.00000484, 0.00000922, -0.00000935, 0.00001055, -0.00001236, 0.00001099, -0.00001139, 0.00001008, -0.00001170, 0.00000487, -0.00000500, 0.00001058, -0.00000968, 0.00001405, -0.00001391, 0.00001596, -0.00001936, 0.00001256, -0.00001528, 0.00001677, -0.00001537, 0.00001442, -0.00001355, 0.00001704, -0.00002021, 0.00001184, -0.00000974, 0.00002629, -0.00002648, 0.00002014, -0.00002572, 0.00002147, -0.00002200, 0.00001371, -0.00001631, 0.00002799, -0.00002761, 0.00002704, -0.00002923, 0.00003942, -0.00004575, 0.00001605, -0.00002272, 0.00004337, -0.00003991, 0.00002664, -0.00003009, 0.00003774, -0.00004088, 0.00002973, -0.00006618, 0.00010039, -0.00011329, 0.00008352, -0.00015658, 0.00012326, -0.00032224, 0.00000085, 0.00098626, -0.00056705, 0.00009060, -0.00010804, 0.00006980, -0.00009043, 0.00005827, -0.00005799, 0.00004842, -0.00006164, 0.00002999, -0.00001725, 0.00001315, -0.00001999, 0.00002031, -0.00000502, 0.00000148, -0.00002518, 0.00001666, -0.00001199, 0.00001380, -0.00001057, 0.00001020, -0.00000490, 0.00000519, -0.00000843, 0.00000001, -0.00000669, 0.00000307, -0.00000871, 0.00001523, -0.00000295, 0.00000153, -0.00000640, 0.00000229, 0.00000076, -0.00000024, 0.00000189, 0.00000030, -0.00000229, -0.00000159, -0.00000478, 0.00000326, -0.00000241, 0.00000426, -0.00000108, 0.00000409, 0.00000247, -0.00000746, },
		{ -0.00000724, 0.00000299, -0.00000436, 0.00000311, -0.00000413, 0.00000063, 0.00000013, 0.00000080, 0.00000003, 0.00000290, -0.00000100, 0.00000438, -0.00000663, 0.00000586, -0.00000714, 0.00000632, -0.00000790, 0.00000401, -0.00000542, 0.00000350, -0.00000403, 0.00000666, -0.00000455, 0.00000883, -0.00000909, 0.00001037, -0.00001238, 0.00001121, -0.00001167, 0.00001041, -0.00001212, 0.00000526, -0.00000495, 0.00001033, -0.00000935, 0.00001362, -0.00001359, 0.00001558, -0.00001917, 0.00001262, -0.00001537, 0.00001708, -0.00001566, 0.00001459, -0.00001357, 0.00001677, -0.00002006, 0.00001192, -0.00000929, 0.00002555, -0.00002622, 0.00002002, -0.00002579, 0.00002160, -0.00002189, 0.00001379, -0.00001614, 0.00002736, -0.00002684, 0.00002623, -0.00002837, 0.00003848, -0.00004606, 0.00001739, -0.00002313, 0.00004299, -0.00003977, 0.00002674, -0.00002960, 0.00003605, -0.00003866, 0.00002784, -0.00006348, 0.00009697, -0.00011102, 0.00008277, -0.00015792, 0.00012449, -0.00033329, 0.00004033, 0.00093911, -0.00054377, 0.00008183, -0.00010116, 0.00006519, -0.00008573, 0.00005563, -0.00005813, 0.00004816, -0.00006122, 0.00003058, -0.00001840, 0.00001350, -0.00001979, 0.00002112, -0.00000611, 0.00000063, -0.00002329, 0.00001591, -0.00001196, 0.00001390, -0.00001076, 0.00001070, -0.00000551, 0.00000514, -0.00000812, -0.00000026, -0.00000647, 0.00000277, -0.00000826, 0.00001563, -0.00000373, 0.00000158, -0.00000637, 0.00000269, 0.00000033, 0.00000003, 0.00000185, 0.00000014, -0.00000208, -0.00000209, -0.00000426, 0.00000316, -0.00000242, 0.00000442, -0.00000144, 0.00000466, 0.00000179, },
		{ 0.00000114, -0.00000701, 0.00000302, -0.00000439, 0.00000328, -0.00000445, 0.00000099, -0.00000013, 0.00000101, -0.00000001, 0.00000275, -0.00000077, 0.00000402, -0.00000637, 0.00000573, -0.00000709, 0.00000647, -0.00000817, 0.00000429, -0.00000569, 0.00000369, -0.00000399, 0.00000652, -0.00000427, 0.00000843, -0.00000879, 0.00001012, -0.00001232, 0.00001135, -0.00001188, 0.00001068, -0.00001250, 0.00000564, -0.00000493, 0.00001008, -0.00000900, 0.00001314, -0.00001321, 0.00001511, -0.00001886, 0.00001257, -0.00001536, 0.00001728, -0.00001586, 0.00001468, -0.00001354, 0.00001644, -0.00001983, 0.00001193, -0.00000879, 0.00002468, -0.00002579, 0.00001974, -0.00002569, 0.00002158, -0.00002165, 0.00001377, -0.00001586, 0.00002658, -0.00002588, 0.00002522, -0.00002727, 0.00003722, -0.00004601, 0.00001848, -0.00002337, 0.00004238, -0.00003936, 0.00002664, -0.00002896, 0.00003419, -0.00003619, 0.00002559, -0.00006023, 0.00009267, -0.00010762, 0.00008071, -0.00015770, 0.00012375, -0.00034177, 0.00007786, 0.00089174, -0.00051910, 0.00007199, -0.00009344, 0.00005979, -0.00008025, 0.00005211, -0.00005745, 0.00004718, -0.00006013, 0.00003065, -0.00001919, 0.00001357, -0.00001934, 0.00002170, -0.00000709, -0.00000029, -0.00002123, 0.00001490, -0.00001168, 0.00001377, -0.00001075, 0.00001101, -0.00000598, 0.00000498, -0.00000770, -0.00000064, -0.00000614, 0.00000233, -0.00000764, 0.00001584, -0.00000437, 0.00000153, -0.00000624, 0.00000299, -0.00000004, 0.00000026, 0.00000182, -0.00000004, -0.00000182, -0.00000264, -0.00000364, 0.00000295, -0.00000231, 0.00000448, -0.00000170, 0.00000514, },
		{ 0.00000554, 0.00000054, -0.00000677, 0.00000304, -0.00000440, 0.00000344, -0.00000476, 0.00000137, -0.00000043, 0.00000128, -0.00000010, 0.00000266, -0.00000059, 0.00000369, -0.00000611, 0.00000559, -0.00000701, 0.00000658, -0.00000842, 0.00000456, -0.00000596, 0.00000390, -0.00000399, 0.00000640, -0.00000400, 0.00000803, -0.00000846, 0.00000981, -0.00001219, 0.00001141, -0.00001203, 0.00001090, -0.00001285, 0.00000603, -0.00000493, 0.00000984, -0.00000863, 0.00001262, -0.00001278, 0.00001456, -0.00001844, 0.00001242, -0.00001525, 0.00001737, -0.00001596, 0.00001470, -0.00001346, 0.00001604, -0.00001951, 0.00001190, -0.00000825, 0.00002371, -0.00002519, 0.00001930, -0.00002545, 0.00002143, -0.00002128, 0.00001364, -0.00001551, 0.00002569, -0.00002476, 0.00002404, -0.00002596, 0.00003566, -0.00004561, 0.00001932, -0.00002343, 0.00004155, -0.00003870, 0.00002634, -0.00002819, 0.00003220, -0.00003349, 0.00002302, -0.00005651, 0.00008758, -0.00010316, 0.00007739, -0.00015596, 0.00012104, -0.00034758, 0.00011318, 0.00084449, -0.00049334, 0.00006129, -0.00008503, 0.00005374, -0.00007410, 0.00004780, -0.00005601, 0.00004551, -0.00005841, 0.00003019, -0.00001960, 0.00001339, -0.00001863, 0.00002206, -0.00000796, -0.00000127, -0.00001902, 0.00001366, -0.00001116, 0.00001343, -0.00001056, 0.00001116, -0.00000630, 0.00000471, -0.00000719, -0.00000110, -0.00000570, 0.00000179, -0.00000687, 0.00001587, -0.00000489, 0.00000139, -0.00000601, 0.00000321, -0.00000034, 0.00000046, 0.00000180, -0.00000025, -0.00000154, -0.00000323, -0.00000296, 0.00000264, -0.00000210, 0.00000442, -0.00000187, },
		{ -0.00000193, 0.00000585, -0.00000002, -0.00000653, 0.00000304, -0.00000441, 0.00000359, -0.00000507, 0.00000177, -0.00000076, 0.00000160, -0.00000026, 0.00000261, -0.00000045, 0.00000338, -0.00000585, 0.00000542, -0.00000691, 0.00000665, -0.00000864, 0.00000482, -0.00000623, 0.00000414, -0.00000401, 0.00000631, -0.00000375, 0.00000761, -0.00000810, 0.00000946, -0.00001198, 0.00001141, -0.00001211, 0.00001106, -0.00001314, 0.00000641, -0.00000495, 0.00000962, -0.00000827, 0.00001208, -0.00001231, 0.00001395, -0.00001792, 0.00001216, -0.00001504, 0.00001737, -0.00001597, 0.00001465, -0.00001333, 0.00001558, -0.00001911, 0.00001182, -0.00000769, 0.00002265, -0.00002445, 0.00001870, -0.00002506, 0.00002115, -0.00002079, 0.00001342, -0.00001508, 0.00002469, -0.00002351, 0.00002270, -0.00002445, 0.00003381, -0.00004487, 0.00001990, -0.00002332, 0.00004050, -0.00003779, 0.00002586, -0.00002731, 0.00003010, -0.00003063, 0.00002022, -0.00005240, 0.00008178, -0.00009772, 0.00007288, -0.00015274, 0.00011639, -0.00035067, 0.00014605, 0.00079773, -0.00046677, 0.00004992, -0.00007608, 0.00004717, -0.00006741, 0.00004280, -0.00005388, 0.00004321, -0.00005609, 0.00002922, -0.00001966, 0.00001294, -0.00001769, 0.00002220, -0.00000872, -0.00000227, -0.00001670, 0.00001222, -0.00001044, 0.00001290, -0.00001020, 0.00001113, -0.00000650, 0.00000434, -0.00000660, -0.00000164, -0.00000517, 0.00000116, -0.00000598, 0.00001573, -0.00000528, 0.00000117, -0.00000569, 0.00000333, -0.00000058, 0.00000063, 0.00000179, -0.00000048, -0.00000122, -0.00000385, -0.00000221, 0.00000225, -0.00000179, 0.00000425, },
		{ 0.00000399, -0.00000189, 0.00000607, -0.00000054, -0.00000629, 0.00000304, -0.00000440, 0.00000373, -0.00000537, 0.00000217, -0.00000112, 0.00000196, -0.00000047, 0.00000263, -0.00000036, 0.00000310, -0.00000560, 0.00000524, -0.00000677, 0.00000669, -0.00000884, 0.00000507, -0.00000651, 0.00000438, -0.00000406, 0.00000625, -0.00000353, 0.00000720, -0.00000773, 0.00000907, -0.00001171, 0.00001133, -0.00001212, 0.00001115, -0.00001339, 0.00000677, -0.00000500, 0.00000942, -0.00000790, 0.00001153, -0.00001180, 0.00001328, -0.00001731, 0.00001181, -0.00001474, 0.00001726, -0.00001589, 0.00001453, -0.00001314, 0.00001506, -0.00001865, 0.00001170, -0.00000711, 0.00002151, -0.00002358, 0.00001798, -0.00002454, 0.00002076, -0.00002019, 0.00001311, -0.00001460, 0.00002359, -0.00002213, 0.00002122, -0.00002278, 0.00003170, -0.00004381, 0.00002025, -0.00002305, 0.00003925, -0.00003665, 0.00002520, -0.00002635, 0.00002794, -0.00002766, 0.00001724, -0.00004798, 0.00007539, -0.00009142, 0.00006726, -0.00014813, 0.00010985, -0.00035100, 0.00017627, 0.00075176, -0.00043969, 0.00003809, -0.00006676, 0.00004023, -0.00006033, 0.00003724, -0.00005114, 0.00004034, -0.00005321, 0.00002778, -0.00001937, 0.00001226, -0.00001651, 0.00002214, -0.00000939, -0.00000329, -0.00001430, 0.00001061, -0.00000952, 0.00001219, -0.00000968, 0.00001095, -0.00000657, 0.00000389, -0.00000593, -0.00000226, -0.00000457, 0.00000044, -0.00000496, 0.00001543, -0.00000555, 0.00000087, -0.00000528, 0.00000336, -0.00000076, 0.00000077, 0.00000178, -0.00000071, -0.00000089, -0.00000448, -0.00000142, 0.00000178, -0.00000139, },
		{ -0.00000091, 0.00000364, -0.00000176, 0.00000619, -0.00000099, -0.00000605, 0.00000302, -0.00000438, 0.00000384, -0.00000565, 0.00000256, -0.00000150, 0.00000235, -0.00000072, 0.00000269, -0.00000032, 0.00000285, -0.00000535, 0.00000506, -0.00000661, 0.00000669, -0.00000899, 0.00000529, -0.00000678, 0.00000465, -0.00000414, 0.00000622, -0.00000333, 0.00000680, -0.00000735, 0.00000865, -0.00001137, 0.00001118, -0.00001207, 0.00001118, -0.00001359, 0.00000713, -0.00000508, 0.00000924, -0.00000755, 0.00001096, -0.00001128, 0.00001258, -0.00001663, 0.00001138, -0.00001435, 0.00001705, -0.00001572, 0.00001434, -0.00001292, 0.00001450, -0.00001812, 0.00001155, -0.00000653, 0.00002032, -0.00002260, 0.00001713, -0.00002389, 0.00002025, -0.00001949, 0.00001273, -0.00001406, 0.00002243, -0.00002066, 0.00001964, -0.00002096, 0.00002937, -0.00004244, 0.00002036, -0.00002262, 0.00003781, -0.00003530, 0.00002438, -0.00002531, 0.00002575, -0.00002464, 0.00001416, -0.00004334, 0.00006852, -0.00008435, 0.00006066, -0.00014223, 0.00010152, -0.00034859, 0.00020366, 0.00070689, -0.00041236, 0.00002601, -0.00005722, 0.00003306, -0.00005299, 0.00003122, -0.00004789, 0.00003696, -0.00004983, 0.00002589, -0.00001874, 0.00001134, -0.00001513, 0.00002187, -0.00000997, -0.00000429, -0.00001185, 0.00000887, -0.00000844, 0.00001132, -0.00000903, 0.00001064, -0.00000653, 0.00000337, -0.00000520, -0.00000292, -0.00000391, -0.00000033, -0.00000385, 0.00001499, -0.00000570, 0.00000050, -0.00000479, 0.00000331, -0.00000087, 0.00000087, 0.00000177, -0.00000094, -0.00000056, -0.00000512, -0.00000060, 0.00000125, },
		{ 0.00000066, -0.00000037, 0.00000320, -0.00000154, 0.00000622, -0.00000139, -0.00000582, 0.00000299, -0.00000434, 0.00000393, -0.00000591, 0.00000295, -0.00000189, 0.00000277, -0.00000102, 0.00000279, -0.00000032, 0.00000264, -0.00000511, 0.00000486, -0.00000642, 0.00000666, -0.00000912, 0.00000549, -0.00000703, 0.00000492, -0.00000425, 0.00000622, -0.00000316, 0.00000642, -0.00000696, 0.00000820, -0.00001099, 0.00001097, -0.00001195, 0.00001115, -0.00001372, 0.00000746, -0.00000517, 0.00000908, -0.00000721, 0.00001040, -0.00001074, 0.00001184, -0.00001588, 0.00001088, -0.00001389, 0.00001675, -0.00001547, 0.00001409, -0.00001265, 0.00001389, -0.00001753, 0.00001137, -0.00000596, 0.00001909, -0.00002152, 0.00001617, -0.00002313, 0.00001964, -0.00001870, 0.00001228, -0.00001349, 0.00002122, -0.00001911, 0.00001797, -0.00001904, 0.00002684, -0.00004080, 0.00002025, -0.00002205, 0.00003621, -0.00003374, 0.00002341, -0.00002422, 0.00002357, -0.00002162, 0.00001104, -0.00003858, 0.00006128, -0.00007666, 0.00005318, -0.00013515, 0.00009152, -0.00034348, 0.00022810, 0.00066339, -0.00038506, 0.00001387, -0.00004762, 0.00002579, -0.00004554, 0.00002487, -0.00004421, 0.00003317, -0.00004601, 0.00002358, -0.00001780, 0.00001021, -0.00001355, 0.00002142, -0.00001048, -0.00000527, -0.00000939, 0.00000702, -0.00000723, 0.00001032, -0.00000825, 0.00001020, -0.00000639, 0.00000279, -0.00000443, -0.00000362, -0.00000320, -0.00000115, -0.00000265, 0.00001441, -0.00000575, 0.00000006, -0.00000423, 0.00000318, -0.00000091, 0.00000096, 0.00000176, -0.00000117, -0.00000024, -0.00000575, 0.00000024, },
		{ 0.00000109, 0.00000005, 0.00000023, 0.00000269, -0.00000123, 0.00000616, -0.00000173, -0.00000560, 0.00000295, -0.00000429, 0.00000400, -0.00000613, 0.00000331, -0.00000229, 0.00000322, -0.00000136, 0.00000294, -0.00000038, 0.00000247, -0.00000489, 0.00000467, -0.00000622, 0.00000659, -0.00000921, 0.00000566, -0.00000727, 0.00000519, -0.00000438, 0.00000625, -0.00000302, 0.00000605, -0.00000658, 0.00000774, -0.00001056, 0.00001070, -0.00001178, 0.00001105, -0.00001380, 0.00000778, -0.00000529, 0.00000895, -0.00000689, 0.00000984, -0.00001021, 0.00001109, -0.00001508, 0.00001032, -0.00001336, 0.00001638, -0.00001514, 0.00001377, -0.00001234, 0.00001324, -0.00001689, 0.00001118, -0.00000541, 0.00001784, -0.00002036, 0.00001512, -0.00002228, 0.00001896, -0.00001784, 0.00001177, -0.00001290, 0.00001997, -0.00001751, 0.00001625, -0.00001705, 0.00002415, -0.00003890, 0.00001993, -0.00002135, 0.00003446, -0.00003201, 0.00002230, -0.00002309, 0.00002143, -0.00001865, 0.00000795, -0.00003378, 0.00005379, -0.00006847, 0.00004496, -0.00012703, 0.00007997, -0.00033577, 0.00024951, 0.00062149, -0.00035801, 0.00000184, -0.00003809, 0.00001857, -0.00003811, 0.00001833, -0.00004020, 0.00002903, -0.00004182, 0.00002092, -0.00001658, 0.00000890, -0.00001180, 0.00002079, -0.00001091, -0.00000619, -0.00000695, 0.00000509, -0.00000590, 0.00000921, -0.00000738, 0.00000966, -0.00000615, 0.00000215, -0.00000363, -0.00000435, -0.00000246, -0.00000200, -0.00000139, 0.00001370, -0.00000570, -0.00000044, -0.00000361, 0.00000298, -0.00000090, 0.00000101, 0.00000174, -0.00000138, 0.00000008, -0.00000635, },
		{ -0.00000693, 0.00000192, -0.00000059, 0.00000087, 0.00000212, -0.00000085, 0.00000601, -0.00000201, -0.00000539, 0.00000291, -0.00000423, 0.00000405, -0.00000633, 0.00000365, -0.00000267, 0.00000367, -0.00000173, 0.00000313, -0.00000047, 0.00000234, -0.00000469, 0.00000448, -0.00000600, 0.00000649, -0.00000926, 0.00000580, -0.00000749, 0.00000546, -0.00000452, 0.00000630, -0.00000292, 0.00000572, -0.00000622, 0.00000727, -0.00001010, 0.00001038, -0.00001155, 0.00001089, -0.00001382, 0.00000806, -0.00000541, 0.00000884, -0.00000660, 0.00000930, -0.00000969, 0.00001034, -0.00001424, 0.00000971, -0.00001277, 0.00001593, -0.00001473, 0.00001340, -0.00001200, 0.00001257, -0.00001621, 0.00001097, -0.00000489, 0.00001658, -0.00001914, 0.00001399, -0.00002134, 0.00001819, -0.00001691, 0.00001122, -0.00001231, 0.00001871, -0.00001589, 0.00001449, -0.00001500, 0.00002133, -0.00003678, 0.00001942, -0.00002054, 0.00003259, -0.00003012, 0.00002107, -0.00002195, 0.00001935, -0.00001577, 0.00000496, -0.00002905, 0.00004617, -0.00005992, 0.00003616, -0.00011803, 0.00006707, -0.00032556, 0.00026783, 0.00058139, -0.00033143, -0.00000991, -0.00002877, 0.00001152, -0.00003084, 0.00001171, -0.00003598, 0.00002464, -0.00003733, 0.00001795, -0.00001511, 0.00000742, -0.00000991, 0.00002000, -0.00001128, -0.00000705, -0.00000456, 0.00000314, -0.00000449, 0.00000801, -0.00000643, 0.00000902, -0.00000585, 0.00000149, -0.00000281, -0.00000508, -0.00000171, -0.00000286, -0.00000009, 0.00001290, -0.00000556, -0.00000098, -0.00000294, 0.00000271, -0.00000083, 0.00000104, 0.00000172, -0.00000157, 0.00000037, },
		{ 0.00000064, -0.00000746, 0.00000273, -0.00000124, 0.00000153, 0.00000149, -0.00000040, 0.00000578, -0.00000222, -0.00000520, 0.00000286, -0.00000415, 0.00000407, -0.00000649, 0.00000396, -0.00000304, 0.00000413, -0.00000211, 0.00000336, -0.00000061, 0.00000225, -0.00000451, 0.00000430, -0.00000577, 0.00000635, -0.00000927, 0.00000591, -0.00000769, 0.00000573, -0.00000468, 0.00000639, -0.00000285, 0.00000541, -0.00000587, 0.00000680, -0.00000960, 0.00001002, -0.00001126, 0.00001068, -0.00001378, 0.00000832, -0.00000556, 0.00000876, -0.00000633, 0.00000878, -0.00000918, 0.00000959, -0.00001339, 0.00000906, -0.00001214, 0.00001542, -0.00001427, 0.00001297, -0.00001164, 0.00001187, -0.00001549, 0.00001075, -0.00000440, 0.00001533, -0.00001788, 0.00001281, -0.00002033, 0.00001738, -0.00001594, 0.00001064, -0.00001172, 0.00001746, -0.00001426, 0.00001274, -0.00001295, 0.00001843, -0.00003447, 0.00001874, -0.00001963, 0.00003062, -0.00002809, 0.00001973, -0.00002079, 0.00001738, -0.00001304, 0.00000212, -0.00002445, 0.00003855, -0.00005116, 0.00002691, -0.00010830, 0.00005297, -0.00031302, 0.00028306, 0.00054327, -0.00030552, -0.00002122, -0.00001979, 0.00000474, -0.00002385, 0.00000514, -0.00003164, 0.00002009, -0.00003262, 0.00001472, -0.00001342, 0.00000581, -0.00000789, 0.00001907, -0.00001160, -0.00000782, -0.00000226, 0.00000117, -0.00000303, 0.00000675, -0.00000543, 0.00000832, -0.00000547, 0.00000080, -0.00000199, -0.00000581, -0.00000097, -0.00000371, 0.00000123, 0.00001201, -0.00000534, -0.00000155, -0.00000223, 0.00000238, -0.00000070, 0.00000104, 0.00000168, -0.00000174, },
		{ -0.00000188, 0.00000087, -0.00000793, 0.00000350, -0.00000188, 0.00000222, 0.00000083, 0.00000011, 0.00000547, -0.00000236, -0.00000503, 0.00000281, -0.00000405, 0.00000407, -0.00000661, 0.00000424, -0.00000339, 0.00000458, -0.00000252, 0.00000361, -0.00000080, 0.00000220, -0.00000435, 0.00000413, -0.00000553, 0.00000619, -0.00000926, 0.00000598, -0.00000786, 0.00000599, -0.00000486, 0.00000649, -0.00000282, 0.00000514, -0.00000555, 0.00000635, -0.00000909, 0.00000962, -0.00001093, 0.00001040, -0.00001368, 0.00000854, -0.00000571, 0.00000871, -0.00000610, 0.00000830, -0.00000871, 0.00000887, -0.00001252, 0.00000840, -0.00001148, 0.00001486, -0.00001374, 0.00001250, -0.00001125, 0.00001115, -0.00001473, 0.00001053, -0.00000396, 0.00001412, -0.00001660, 0.00001158, -0.00001927, 0.00001651, -0.00001492, 0.00001004, -0.00001114, 0.00001623, -0.00001265, 0.00001101, -0.00001091, 0.00001548, -0.00003201, 0.00001791, -0.00001865, 0.00002857, -0.00002596, 0.00001830, -0.00001964, 0.00001552, -0.00001048, -0.00000051, -0.00002009, 0.00003104, -0.00004232, 0.00001737, -0.00009804, 0.00003791, -0.00029831, 0.00029523, 0.00050726, -0.00028046, -0.00003195, -0.00001124, -0.00000165, -0.00001726, -0.00000127, -0.00002728, 0.00001546, -0.00002777, 0.00001130, -0.00001155, 0.00000410, -0.00000577, 0.00001801, -0.00001186, -0.00000850, -0.00000006, -0.00000076, -0.00000155, 0.00000545, -0.00000439, 0.00000757, -0.00000506, 0.00000010, -0.00000119, -0.00000651, -0.00000023, -0.00000453, 0.00000255, 0.00001105, -0.00000506, -0.00000215, -0.00000149, 0.00000201, -0.00000053, 0.00000103, 0.00000162, },
		{ 0.00000155, -0.00000199, 0.00000106, -0.00000835, 0.00000421, -0.00000250, 0.00000290, 0.00000015, 0.00000066, 0.00000508, -0.00000244, -0.00000487, 0.00000276, -0.00000395, 0.00000404, -0.00000669, 0.00000447, -0.00000371, 0.00000502, -0.00000292, 0.00000388, -0.00000101, 0.00000218, -0.00000422, 0.00000397, -0.00000529, 0.00000601, -0.00000920, 0.00000603, -0.00000801, 0.00000624, -0.00000504, 0.00000662, -0.00000283, 0.00000491, -0.00000526, 0.00000591, -0.00000857, 0.00000919, -0.00001056, 0.00001008, -0.00001351, 0.00000872, -0.00000586, 0.00000868, -0.00000589, 0.00000784, -0.00000827, 0.00000818, -0.00001166, 0.00000772, -0.00001080, 0.00001425, -0.00001317, 0.00001199, -0.00001084, 0.00001043, -0.00001396, 0.00001031, -0.00000358, 0.00001294, -0.00001532, 0.00001033, -0.00001817, 0.00001562, -0.00001389, 0.00000942, -0.00001060, 0.00001504, -0.00001108, 0.00000932, -0.00000892, 0.00001252, -0.00002942, 0.00001696, -0.00001761, 0.00002647, -0.00002373, 0.00001679, -0.00001850, 0.00001380, -0.00000814, -0.00000289, -0.00001602, 0.00002375, -0.00003355, 0.00000772, -0.00008741, 0.00002207, -0.00028166, 0.00030440, 0.00047346, -0.00025637, -0.00004200, -0.00000323, -0.00000758, -0.00001117, -0.00000740, -0.00002300, 0.00001084, -0.00002285, 0.00000774, -0.00000955, 0.00000231, -0.00000359, 0.00001683, -0.00001209, -0.00000907, 0.00000201, -0.00000264, -0.00000006, 0.00000414, -0.00000335, 0.00000679, -0.00000461, -0.00000060, -0.00000041, -0.00000719, 0.00000047, -0.00000531, 0.00000386, 0.00001003, -0.00000473, -0.00000277, -0.00000073, 0.00000159, -0.00000031, 0.00000099, },
		{ 0.00000093, 0.00000146, -0.00000207, 0.00000121, -0.00000870, 0.00000487, -0.00000309, 0.00000357, -0.00000055, 0.00000125, 0.00000463, -0.00000246, -0.00000474, 0.00000271, -0.00000384, 0.00000399, -0.00000674, 0.00000467, -0.00000400, 0.00000543, -0.00000333, 0.00000416, -0.00000126, 0.00000221, -0.00000411, 0.00000383, -0.00000505, 0.00000581, -0.00000912, 0.00000604, -0.00000812, 0.00000647, -0.00000522, 0.00000677, -0.00000287, 0.00000471, -0.00000501, 0.00000550, -0.00000805, 0.00000874, -0.00001016, 0.00000971, -0.00001329, 0.00000887, -0.00000602, 0.00000867, -0.00000572, 0.00000742, -0.00000786, 0.00000753, -0.00001082, 0.00000705, -0.00001011, 0.00001362, -0.00001256, 0.00001145, -0.00001042, 0.00000970, -0.00001316, 0.00001010, -0.00000326, 0.00001182, -0.00001405, 0.00000907, -0.00001704, 0.00001471, -0.00001285, 0.00000881, -0.00001009, 0.00001391, -0.00000957, 0.00000770, -0.00000701, 0.00000958, -0.00002674, 0.00001591, -0.00001654, 0.00002434, -0.00002145, 0.00001522, -0.00001739, 0.00001223, -0.00000604, -0.00000496, -0.00001231, 0.00001679, -0.00002497, -0.00000191, -0.00007660, 0.00000571, -0.00026328, 0.00031068, 0.00044193, -0.00023340, -0.00005126, 0.00000417, -0.00001297, -0.00000567, -0.00001315, -0.00001889, 0.00000632, -0.00001795, 0.00000412, -0.00000746, 0.00000047, -0.00000136, 0.00001557, -0.00001228, -0.00000952, 0.00000392, -0.00000443, 0.00000139, 0.00000283, -0.00000232, 0.00000599, -0.00000414, -0.00000128, 0.00000033, -0.00000781, 0.00000113, -0.00000604, 0.00000512, 0.00000898, -0.00000436, -0.00000338, 0.00000004, 0.00000114, -0.00000005, },
		{ 0.00000025, 0.00000085, 0.00000136, -0.00000212, 0.00000132, -0.00000898, 0.00000546, -0.00000364, 0.00000422, -0.00000125, 0.00000187, 0.00000412, -0.00000242, -0.00000464, 0.00000267, -0.00000372, 0.00000392, -0.00000675, 0.00000482, -0.00000426, 0.00000583, -0.00000373, 0.00000446, -0.00000153, 0.00000227, -0.00000404, 0.00000371, -0.00000481, 0.00000560, -0.00000901, 0.00000601, -0.00000820, 0.00000668, -0.00000540, 0.00000693, -0.00000294, 0.00000456, -0.00000480, 0.00000512, -0.00000753, 0.00000828, -0.00000974, 0.00000930, -0.00001302, 0.00000898, -0.00000618, 0.00000868, -0.00000559, 0.00000704, -0.00000750, 0.00000693, -0.00001000, 0.00000639, -0.00000943, 0.00001297, -0.00001192, 0.00001088, -0.00001000, 0.00000897, -0.00001236, 0.00000989, -0.00000300, 0.00001076, -0.00001281, 0.00000782, -0.00001591, 0.00001380, -0.00001181, 0.00000821, -0.00000963, 0.00001284, -0.00000813, 0.00000618, -0.00000520, 0.00000670, -0.00002402, 0.00001478, -0.00001545, 0.00002221, -0.00001912, 0.00001361, -0.00001631, 0.00001083, -0.00000420, -0.00000671, -0.00000901, 0.00001024, -0.00001672, -0.00001136, -0.00006580, -0.00001096, -0.00024343, 0.00031420, 0.00041271, -0.00021161, -0.00005967, 0.00001091, -0.00001777, -0.00000083, -0.00001845, -0.00001504, 0.00000198, -0.00001314, 0.00000048, -0.00000532, -0.00000137, 0.00000088, 0.00001422, -0.00001243, -0.00000985, 0.00000565, -0.00000611, 0.00000280, 0.00000155, -0.00000131, 0.00000520, -0.00000366, -0.00000193, 0.00000101, -0.00000838, 0.00000174, -0.00000669, 0.00000634, 0.00000792, -0.00000396, -0.00000400, 0.00000081, 0.00000067, },
		{ 0.00000019, 0.00000057, 0.00000075, 0.00000125, -0.00000213, 0.00000138, -0.00000918, 0.00000597, -0.00000413, 0.00000483, -0.00000194, 0.00000250, 0.00000356, -0.00000232, -0.00000456, 0.00000263, -0.00000360, 0.00000383, -0.00000673, 0.00000492, -0.00000447, 0.00000619, -0.00000412, 0.00000475, -0.00000183, 0.00000237, -0.00000399, 0.00000361, -0.00000459, 0.00000537, -0.00000888, 0.00000596, -0.00000825, 0.00000686, -0.00000558, 0.00000711, -0.00000305, 0.00000445, -0.00000463, 0.00000477, -0.00000703, 0.00000781, -0.00000929, 0.00000885, -0.00001269, 0.00000904, -0.00000634, 0.00000871, -0.00000549, 0.00000670, -0.00000719, 0.00000638, -0.00000922, 0.00000576, -0.00000876, 0.00001231, -0.00001126, 0.00001030, -0.00000958, 0.00000826, -0.00001155, 0.00000970, -0.00000281, 0.00000979, -0.00001162, 0.00000660, -0.00001477, 0.00001290, -0.00001079, 0.00000762, -0.00000923, 0.00001185, -0.00000678, 0.00000476, -0.00000351, 0.00000392, -0.00002127, 0.00001360, -0.00001436, 0.00002009, -0.00001678, 0.00001196, -0.00001526, 0.00000959, -0.00000264, -0.00000810, -0.00000619, 0.00000419, -0.00000892, -0.00002049, -0.00005517, -0.00002770, -0.00022236, 0.00031511, 0.00038579, -0.00019108, -0.00006715, 0.00001695, -0.00002195, 0.00000330, -0.00002321, -0.00001151, -0.00000210, -0.00000850, -0.00000311, -0.00000317, -0.00000318, 0.00000311, 0.00001282, -0.00001256, -0.00001006, 0.00000720, -0.00000766, 0.00000414, 0.00000031, -0.00000036, 0.00000443, -0.00000319, -0.00000254, 0.00000164, -0.00000888, 0.00000228, -0.00000726, 0.00000748, 0.00000685, -0.00000354, -0.00000459, 0.00000156, },
		{ 0.00000228, -0.00000030, 0.00000092, 0.00000064, 0.00000112, -0.00000211, 0.00000140, -0.00000930, 0.00000640, -0.00000456, 0.00000540, -0.00000261, 0.00000313, 0.00000295, -0.00000216, -0.00000450, 0.00000260, -0.00000348, 0.00000373, -0.00000667, 0.00000499, -0.00000464, 0.00000651, -0.00000449, 0.00000504, -0.00000214, 0.00000249, -0.00000397, 0.00000354, -0.00000438, 0.00000514, -0.00000872, 0.00000588, -0.00000827, 0.00000703, -0.00000574, 0.00000729, -0.00000318, 0.00000438, -0.00000450, 0.00000448, -0.00000656, 0.00000735, -0.00000884, 0.00000838, -0.00001232, 0.00000907, -0.00000649, 0.00000876, -0.00000542, 0.00000640, -0.00000693, 0.00000590, -0.00000849, 0.00000517, -0.00000812, 0.00001166, -0.00001060, 0.00000970, -0.00000916, 0.00000756, -0.00001073, 0.00000951, -0.00000268, 0.00000889, -0.00001048, 0.00000541, -0.00001365, 0.00001202, -0.00000980, 0.00000707, -0.00000889, 0.00001095, -0.00000554, 0.00000346, -0.00000198, 0.00000125, -0.00001854, 0.00001239, -0.00001330, 0.00001801, -0.00001445, 0.00001030, -0.00001425, 0.00000853, -0.00000136, -0.00000913, -0.00000385, -0.00000129, -0.00000165, -0.00002915, -0.00004490, -0.00004427, -0.00020035, 0.00031361, 0.00036115, -0.00017186, -0.00007368, 0.00002226, -0.00002549, 0.00000667, -0.00002737, -0.00000837, -0.00000588, -0.00000408, -0.00000659, -0.00000106, -0.00000495, 0.00000529, 0.00001138, -0.00001267, -0.00001015, 0.00000856, -0.00000907, 0.00000538, -0.00000086, 0.00000053, 0.00000370, -0.00000274, -0.00000311, 0.00000220, -0.00000930, 0.00000275, -0.00000775, 0.00000854, 0.00000579, -0.00000311, -0.00000516, },
		{ -0.00000569, 0.00000297, -0.00000078, 0.00000129, 0.00000050, 0.00000098, -0.00000207, 0.00000138, -0.00000935, 0.00000674, -0.00000491, 0.00000591, -0.00000324, 0.00000376, 0.00000232, -0.00000196, -0.00000447, 0.00000257, -0.00000336, 0.00000362, -0.00000658, 0.00000500, -0.00000476, 0.00000679, -0.00000483, 0.00000533, -0.00000246, 0.00000265, -0.00000397, 0.00000349, -0.00000418, 0.00000490, -0.00000855, 0.00000577, -0.00000825, 0.00000716, -0.00000590, 0.00000747, -0.00000334, 0.00000434, -0.00000442, 0.00000422, -0.00000611, 0.00000691, -0.00000839, 0.00000790, -0.00001191, 0.00000905, -0.00000663, 0.00000882, -0.00000538, 0.00000614, -0.00000672, 0.00000548, -0.00000781, 0.00000462, -0.00000751, 0.00001102, -0.00000993, 0.00000910, -0.00000875, 0.00000688, -0.00000992, 0.00000934, -0.00000263, 0.00000808, -0.00000941, 0.00000427, -0.00001255, 0.00001116, -0.00000884, 0.00000654, -0.00000862, 0.00001014, -0.00000441, 0.00000230, -0.00000060, -0.00000126, -0.00001585, 0.00001118, -0.00001227, 0.00001600, -0.00001215, 0.00000863, -0.00001327, 0.00000763, -0.00000036, -0.00000978, -0.00000204, -0.00000615, 0.00000497, -0.00003723, -0.00003514, -0.00006045, -0.00017767, 0.00030989, 0.00033873, -0.00015395, -0.00007924, 0.00002684, -0.00002838, 0.00000925, -0.00003089, -0.00000567, -0.00000928, 0.00000004, -0.00000991, 0.00000098, -0.00000663, 0.00000741, 0.00000991, -0.00001274, -0.00001011, 0.00000970, -0.00001030, 0.00000652, -0.00000196, 0.00000135, 0.00000302, -0.00000232, -0.00000362, 0.00000268, -0.00000964, 0.00000315, -0.00000813, 0.00000951, 0.00000475, -0.00000269, },
		{ -0.00000228, -0.00000618, 0.00000362, -0.00000125, 0.00000168, 0.00000036, 0.00000083, -0.00000199, 0.00000131, -0.00000932, 0.00000699, -0.00000520, 0.00000636, -0.00000382, 0.00000436, 0.00000166, -0.00000171, -0.00000445, 0.00000256, -0.00000324, 0.00000350, -0.00000647, 0.00000498, -0.00000484, 0.00000703, -0.00000514, 0.00000559, -0.00000279, 0.00000282, -0.00000400, 0.00000347, -0.00000401, 0.00000467, -0.00000836, 0.00000563, -0.00000820, 0.00000727, -0.00000604, 0.00000766, -0.00000352, 0.00000434, -0.00000439, 0.00000402, -0.00000570, 0.00000648, -0.00000795, 0.00000739, -0.00001146, 0.00000900, -0.00000675, 0.00000888, -0.00000537, 0.00000591, -0.00000655, 0.00000513, -0.00000719, 0.00000412, -0.00000695, 0.00001040, -0.00000928, 0.00000851, -0.00000836, 0.00000622, -0.00000912, 0.00000919, -0.00000265, 0.00000737, -0.00000841, 0.00000319, -0.00001149, 0.00001035, -0.00000793, 0.00000606, -0.00000842, 0.00000944, -0.00000340, 0.00000128, 0.00000059, -0.00000360, -0.00001323, 0.00000998, -0.00001130, 0.00001407, -0.00000990, 0.00000697, -0.00001233, 0.00000689, 0.00000036, -0.00001006, -0.00000076, -0.00001034, 0.00001088, -0.00004463, -0.00002603, -0.00007602, -0.00015461, 0.00030418, 0.00031846, -0.00013735, -0.00008383, 0.00003070, -0.00003064, 0.00001106, -0.00003374, -0.00000345, -0.00001226, 0.00000382, -0.00001301, 0.00000291, -0.00000820, 0.00000943, 0.00000843, -0.00001280, -0.00000995, 0.00001064, -0.00001137, 0.00000754, -0.00000296, 0.00000207, 0.00000240, -0.00000193, -0.00000408, 0.00000308, -0.00000990, 0.00000346, -0.00000841, 0.00001036, 0.00000376, },
		{ 0.00000281, -0.00000188, -0.00000662, 0.00000421, -0.00000170, 0.00000207, 0.00000019, 0.00000067, -0.00000190, 0.00000121, -0.00000922, 0.00000716, -0.00000541, 0.00000674, -0.00000436, 0.00000493, 0.00000100, -0.00000143, -0.00000447, 0.00000256, -0.00000314, 0.00000337, -0.00000633, 0.00000492, -0.00000487, 0.00000723, -0.00000541, 0.00000583, -0.00000311, 0.00000302, -0.00000406, 0.00000348, -0.00000386, 0.00000445, -0.00000816, 0.00000548, -0.00000813, 0.00000735, -0.00000616, 0.00000783, -0.00000371, 0.00000438, -0.00000440, 0.00000387, -0.00000533, 0.00000608, -0.00000752, 0.00000689, -0.00001098, 0.00000891, -0.00000687, 0.00000895, -0.00000538, 0.00000572, -0.00000643, 0.00000484, -0.00000664, 0.00000369, -0.00000645, 0.00000982, -0.00000865, 0.00000792, -0.00000798, 0.00000560, -0.00000833, 0.00000905, -0.00000273, 0.00000675, -0.00000750, 0.00000218, -0.00001047, 0.00000958, -0.00000707, 0.00000562, -0.00000829, 0.00000883, -0.00000252, 0.00000042, 0.00000159, -0.00000575, -0.00001072, 0.00000881, -0.00001039, 0.00001224, -0.00000772, 0.00000532, -0.00001143, 0.00000629, 0.00000080, -0.00000997, -0.00000000, -0.00001384, 0.00001601, -0.00005124, -0.00001772, -0.00009078, -0.00013143, 0.00029671, 0.00030023, -0.00012204, -0.00008747, 0.00003386, -0.00003231, 0.00001209, -0.00003590, -0.00000174, -0.00001479, 0.00000721, -0.00001585, 0.00000469, -0.00000963, 0.00001134, 0.00000695, -0.00001283, -0.00000969, 0.00001137, -0.00001224, 0.00000843, -0.00000388, 0.00000270, 0.00000185, -0.00000159, -0.00000446, 0.00000339, -0.00001006, 0.00000369, -0.00000858, 0.00001110, },
		{ 0.00001172, 0.00000192, -0.00000152, -0.00000701, 0.00000474, -0.00000211, 0.00000247, 0.00000001, 0.00000051, -0.00000178, 0.00000108, -0.00000905, 0.00000723, -0.00000553, 0.00000705, -0.00000483, 0.00000547, 0.00000032, -0.00000111, -0.00000450, 0.00000256, -0.00000304, 0.00000325, -0.00000618, 0.00000482, -0.00000487, 0.00000738, -0.00000565, 0.00000605, -0.00000343, 0.00000323, -0.00000413, 0.00000351, -0.00000373, 0.00000423, -0.00000796, 0.00000531, -0.00000802, 0.00000740, -0.00000627, 0.00000800, -0.00000392, 0.00000444, -0.00000445, 0.00000377, -0.00000500, 0.00000571, -0.00000711, 0.00000639, -0.00001047, 0.00000878, -0.00000696, 0.00000902, -0.00000541, 0.00000557, -0.00000636, 0.00000462, -0.00000615, 0.00000331, -0.00000600, 0.00000928, -0.00000804, 0.00000735, -0.00000763, 0.00000501, -0.00000755, 0.00000891, -0.00000287, 0.00000623, -0.00000669, 0.00000124, -0.00000950, 0.00000886, -0.00000626, 0.00000522, -0.00000822, 0.00000833, -0.00000177, -0.00000029, 0.00000240, -0.00000768, -0.00000832, 0.00000770, -0.00000957, 0.00001052, -0.00000564, 0.00000370, -0.00001055, 0.00000584, 0.00000100, -0.00000955, 0.00000024, -0.00001663, 0.00002033, -0.00005700, -0.00001031, -0.00010455, -0.00010842, 0.00028771, 0.00028392, -0.00010797, -0.00009018, 0.00003635, -0.00003342, 0.00001238, -0.00003738, -0.00000055, -0.00001684, 0.00001018, -0.00001840, 0.00000630, -0.00001089, 0.00001310, 0.00000550, -0.00001283, -0.00000932, 0.00001190, -0.00001294, 0.00000919, -0.00000468, 0.00000322, 0.00000137, -0.00000129, -0.00000479, 0.00000362, -0.00001013, 0.00000383, -0.00000863, },
		{ -0.00000858, 0.00001221, 0.00000109, -0.00000118, -0.00000733, 0.00000522, -0.00000249, 0.00000286, -0.00000018, 0.00000036, -0.00000165, 0.00000093, -0.00000882, 0.00000723, -0.00000558, 0.00000727, -0.00000524, 0.00000596, -0.00000035, -0.00000076, -0.00000455, 0.00000258, -0.00000295, 0.00000313, -0.00000602, 0.00000470, -0.00000482, 0.00000748, -0.00000585, 0.00000624, -0.00000374, 0.00000345, -0.00000423, 0.00000357, -0.00000363, 0.00000403, -0.00000776, 0.00000513, -0.00000790, 0.00000742, -0.00000635, 0.00000816, -0.00000414, 0.00000453, -0.00000454, 0.00000373, -0.00000472, 0.00000538, -0.00000672, 0.00000589, -0.00000995, 0.00000862, -0.00000704, 0.00000909, -0.00000546, 0.00000544, -0.00000633, 0.00000447, -0.00000573, 0.00000300, -0.00000561, 0.00000878, -0.00000747, 0.00000680, -0.00000731, 0.00000446, -0.00000679, 0.00000879, -0.00000308, 0.00000580, -0.00000596, 0.00000039, -0.00000858, 0.00000820, -0.00000552, 0.00000487, -0.00000823, 0.00000794, -0.00000114, -0.00000085, 0.00000300, -0.00000939, -0.00000607, 0.00000665, -0.00000884, 0.00000893, -0.00000366, 0.00000212, -0.00000970, 0.00000550, 0.00000097, -0.00000881, -0.00000001, -0.00001872, 0.00002380, -0.00006185, -0.00000389, -0.00011714, -0.00008582, 0.00027744, 0.00026941, -0.00009509, -0.00009201, 0.00003823, -0.00003404, 0.00001197, -0.00003818, 0.00000012, -0.00001840, 0.00001269, -0.00002062, 0.00000770, -0.00001198, 0.00001471, 0.00000407, -0.00001281, -0.00000887, 0.00001223, -0.00001344, 0.00000980, -0.00000539, 0.00000363, 0.00000098, -0.00000105, -0.00000504, 0.00000376, -0.00001011, 0.00000389, },
		{ 0.00000386, -0.00000842, 0.00001257, 0.00000034, -0.00000088, -0.00000760, 0.00000562, -0.00000283, 0.00000325, -0.00000039, 0.00000020, -0.00000151, 0.00000075, -0.00000853, 0.00000714, -0.00000555, 0.00000742, -0.00000558, 0.00000639, -0.00000100, -0.00000039, -0.00000462, 0.00000262, -0.00000287, 0.00000302, -0.00000585, 0.00000455, -0.00000474, 0.00000754, -0.00000600, 0.00000640, -0.00000403, 0.00000367, -0.00000433, 0.00000365, -0.00000356, 0.00000385, -0.00000757, 0.00000494, -0.00000775, 0.00000741, -0.00000640, 0.00000829, -0.00000435, 0.00000465, -0.00000467, 0.00000373, -0.00000449, 0.00000508, -0.00000637, 0.00000542, -0.00000941, 0.00000843, -0.00000710, 0.00000916, -0.00000552, 0.00000534, -0.00000632, 0.00000437, -0.00000538, 0.00000276, -0.00000529, 0.00000834, -0.00000694, 0.00000628, -0.00000701, 0.00000395, -0.00000606, 0.00000867, -0.00000333, 0.00000546, -0.00000533, -0.00000038, -0.00000772, 0.00000759, -0.00000485, 0.00000457, -0.00000829, 0.00000763, -0.00000064, -0.00000125, 0.00000341, -0.00001086, -0.00000397, 0.00000568, -0.00000820, 0.00000748, -0.00000180, 0.00000059, -0.00000886, 0.00000526, 0.00000073, -0.00000778, -0.00000072, -0.00002012, 0.00002641, -0.00006575, 0.00000145, -0.00012843, -0.00006388, 0.00026613, 0.00025656, -0.00008333, -0.00009301, 0.00003955, -0.00003423, 0.00001092, -0.00003835, 0.00000028, -0.00001946, 0.00001473, -0.00002249, 0.00000888, -0.00001287, 0.00001615, 0.00000270, -0.00001276, -0.00000834, 0.00001237, -0.00001376, 0.00001028, -0.00000598, 0.00000394, 0.00000068, -0.00000085, -0.00000522, 0.00000381, -0.00001000, },
		{ -0.00000980, 0.00000376, -0.00000816, 0.00001280, -0.00000034, -0.00000062, -0.00000781, 0.00000595, -0.00000312, 0.00000363, -0.00000061, 0.00000005, -0.00000137, 0.00000057, -0.00000820, 0.00000698, -0.00000545, 0.00000750, -0.00000585, 0.00000677, -0.00000163, -0.00000000, -0.00000470, 0.00000266, -0.00000281, 0.00000292, -0.00000568, 0.00000438, -0.00000462, 0.00000755, -0.00000612, 0.00000652, -0.00000430, 0.00000390, -0.00000445, 0.00000375, -0.00000351, 0.00000369, -0.00000738, 0.00000474, -0.00000758, 0.00000737, -0.00000643, 0.00000841, -0.00000457, 0.00000477, -0.00000483, 0.00000378, -0.00000430, 0.00000483, -0.00000606, 0.00000496, -0.00000887, 0.00000821, -0.00000713, 0.00000921, -0.00000559, 0.00000525, -0.00000635, 0.00000433, -0.00000509, 0.00000258, -0.00000504, 0.00000795, -0.00000645, 0.00000580, -0.00000675, 0.00000348, -0.00000534, 0.00000856, -0.00000363, 0.00000521, -0.00000479, -0.00000106, -0.00000692, 0.00000704, -0.00000424, 0.00000431, -0.00000842, 0.00000742, -0.00000026, -0.00000151, 0.00000362, -0.00001210, -0.00000205, 0.00000480, -0.00000767, 0.00000617, -0.00000008, -0.00000088, -0.00000803, 0.00000510, 0.00000032, -0.00000651, -0.00000184, -0.00002087, 0.00002817, -0.00006869, 0.00000568, -0.00013827, -0.00004282, 0.00025403, 0.00024521, -0.00007260, -0.00009325, 0.00004037, -0.00003407, 0.00000930, -0.00003791, -0.00000004, -0.00002004, 0.00001630, -0.00002399, 0.00000982, -0.00001354, 0.00001740, 0.00000138, -0.00001267, -0.00000775, 0.00001234, -0.00001390, 0.00001062, -0.00000648, 0.00000415, 0.00000045, -0.00000071, -0.00000535, 0.00000378, },
		{ 0.00000367, -0.00000952, 0.00000358, -0.00000780, 0.00001291, -0.00000095, -0.00000039, -0.00000795, 0.00000622, -0.00000336, 0.00000400, -0.00000084, -0.00000009, -0.00000122, 0.00000038, -0.00000783, 0.00000675, -0.00000527, 0.00000749, -0.00000604, 0.00000708, -0.00000224, 0.00000039, -0.00000479, 0.00000271, -0.00000276, 0.00000284, -0.00000551, 0.00000420, -0.00000448, 0.00000753, -0.00000620, 0.00000660, -0.00000455, 0.00000412, -0.00000458, 0.00000387, -0.00000349, 0.00000355, -0.00000720, 0.00000454, -0.00000739, 0.00000731, -0.00000644, 0.00000851, -0.00000477, 0.00000491, -0.00000501, 0.00000388, -0.00000417, 0.00000463, -0.00000578, 0.00000454, -0.00000833, 0.00000796, -0.00000715, 0.00000926, -0.00000567, 0.00000518, -0.00000640, 0.00000434, -0.00000486, 0.00000247, -0.00000485, 0.00000761, -0.00000601, 0.00000534, -0.00000652, 0.00000306, -0.00000465, 0.00000845, -0.00000396, 0.00000504, -0.00000435, -0.00000165, -0.00000618, 0.00000655, -0.00000370, 0.00000409, -0.00000859, 0.00000729, 0.00000000, -0.00000163, 0.00000365, -0.00001309, -0.00000030, 0.00000401, -0.00000724, 0.00000502, 0.00000150, -0.00000230, -0.00000721, 0.00000500, -0.00000024, -0.00000505, -0.00000333, -0.00002100, 0.00002911, -0.00007066, 0.00000876, -0.00014659, -0.00002285, 0.00024136, 0.00023522, -0.00006284, -0.00009278, 0.00004076, -0.00003364, 0.00000718, -0.00003693, -0.00000081, -0.00002016, 0.00001739, -0.00002511, 0.00001050, -0.00001400, 0.00001845, 0.00000012, -0.00001255, -0.00000711, 0.00001215, -0.00001387, 0.00001083, -0.00000687, 0.00000425, 0.00000031, -0.00000062, -0.00000541, },
		{ -0.00000541, 0.00000350, -0.00000917, 0.00000334, -0.00000736, 0.00001290, -0.00000148, -0.00000020, -0.00000804, 0.00000641, -0.00000355, 0.00000435, -0.00000109, -0.00000021, -0.00000109, 0.00000019, -0.00000743, 0.00000647, -0.00000503, 0.00000742, -0.00000616, 0.00000732, -0.00000281, 0.00000079, -0.00000489, 0.00000277, -0.00000273, 0.00000277, -0.00000535, 0.00000401, -0.00000431, 0.00000747, -0.00000624, 0.00000665, -0.00000477, 0.00000434, -0.00000471, 0.00000401, -0.00000350, 0.00000343, -0.00000704, 0.00000435, -0.00000720, 0.00000723, -0.00000642, 0.00000858, -0.00000497, 0.00000506, -0.00000522, 0.00000402, -0.00000408, 0.00000447, -0.00000555, 0.00000414, -0.00000779, 0.00000770, -0.00000715, 0.00000929, -0.00000574, 0.00000512, -0.00000647, 0.00000439, -0.00000468, 0.00000243, -0.00000473, 0.00000734, -0.00000563, 0.00000493, -0.00000633, 0.00000269, -0.00000398, 0.00000834, -0.00000432, 0.00000494, -0.00000399, -0.00000215, -0.00000550, 0.00000612, -0.00000322, 0.00000391, -0.00000881, 0.00000724, 0.00000016, -0.00000162, 0.00000349, -0.00001385, 0.00000126, 0.00000332, -0.00000691, 0.00000401, 0.00000292, -0.00000364, -0.00000638, 0.00000494, -0.00000090, -0.00000342, -0.00000513, -0.00002057, 0.00002926, -0.00007169, 0.00001069, -0.00015330, -0.00000414, 0.00022835, 0.00022643, -0.00005393, -0.00009168, 0.00004079, -0.00003304, 0.00000465, -0.00003547, -0.00000198, -0.00001984, 0.00001801, -0.00002584, 0.00001093, -0.00001423, 0.00001929, -0.00000106, -0.00001238, -0.00000644, 0.00001181, -0.00001367, 0.00001091, -0.00000717, 0.00000425, 0.00000025, -0.00000058, },
		{ -0.00000057, -0.00000537, 0.00000326, -0.00000875, 0.00000305, -0.00000683, 0.00001277, -0.00000194, -0.00000005, -0.00000807, 0.00000652, -0.00000368, 0.00000467, -0.00000135, -0.00000032, -0.00000096, 0.00000001, -0.00000701, 0.00000614, -0.00000474, 0.00000728, -0.00000621, 0.00000750, -0.00000334, 0.00000119, -0.00000499, 0.00000284, -0.00000272, 0.00000272, -0.00000521, 0.00000383, -0.00000413, 0.00000738, -0.00000624, 0.00000666, -0.00000496, 0.00000455, -0.00000484, 0.00000417, -0.00000353, 0.00000334, -0.00000690, 0.00000417, -0.00000700, 0.00000713, -0.00000638, 0.00000863, -0.00000516, 0.00000520, -0.00000544, 0.00000419, -0.00000404, 0.00000436, -0.00000537, 0.00000379, -0.00000727, 0.00000742, -0.00000713, 0.00000930, -0.00000581, 0.00000507, -0.00000654, 0.00000448, -0.00000456, 0.00000243, -0.00000468, 0.00000712, -0.00000530, 0.00000456, -0.00000617, 0.00000236, -0.00000335, 0.00000822, -0.00000470, 0.00000491, -0.00000372, -0.00000257, -0.00000488, 0.00000573, -0.00000280, 0.00000377, -0.00000907, 0.00000725, 0.00000023, -0.00000150, 0.00000318, -0.00001438, 0.00000264, 0.00000274, -0.00000669, 0.00000316, 0.00000419, -0.00000491, -0.00000554, 0.00000488, -0.00000164, -0.00000170, -0.00000717, -0.00001964, 0.00002867, -0.00007181, 0.00001148, -0.00015836, 0.00001314, 0.00021519, 0.00021870, -0.00004580, -0.00009002, 0.00004053, -0.00003234, 0.00000181, -0.00003360, -0.00000350, -0.00001913, 0.00001820, -0.00002620, 0.00001111, -0.00001424, 0.00001991, -0.00000216, -0.00001218, -0.00000577, 0.00001135, -0.00001334, 0.00001088, -0.00000739, 0.00000417, 0.00000027, },
		{ 0.00000034, -0.00000060, -0.00000528, 0.00000296, -0.00000827, 0.00000270, -0.00000624, 0.00001253, -0.00000232, 0.00000007, -0.00000804, 0.00000657, -0.00000375, 0.00000498, -0.00000162, -0.00000041, -0.00000086, -0.00000014, -0.00000658, 0.00000577, -0.00000440, 0.00000707, -0.00000619, 0.00000760, -0.00000382, 0.00000159, -0.00000510, 0.00000292, -0.00000272, 0.00000269, -0.00000509, 0.00000364, -0.00000394, 0.00000726, -0.00000621, 0.00000663, -0.00000512, 0.00000474, -0.00000497, 0.00000433, -0.00000359, 0.00000327, -0.00000677, 0.00000399, -0.00000679, 0.00000702, -0.00000632, 0.00000864, -0.00000532, 0.00000534, -0.00000566, 0.00000440, -0.00000403, 0.00000429, -0.00000523, 0.00000348, -0.00000677, 0.00000713, -0.00000708, 0.00000930, -0.00000587, 0.00000501, -0.00000661, 0.00000459, -0.00000448, 0.00000249, -0.00000468, 0.00000696, -0.00000502, 0.00000422, -0.00000605, 0.00000208, -0.00000273, 0.00000810, -0.00000509, 0.00000493, -0.00000352, -0.00000291, -0.00000430, 0.00000540, -0.00000244, 0.00000366, -0.00000934, 0.00000732, 0.00000022, -0.00000127, 0.00000271, -0.00001469, 0.00000382, 0.00000225, -0.00000656, 0.00000245, 0.00000529, -0.00000610, -0.00000468, 0.00000482, -0.00000242, 0.00000009, -0.00000938, -0.00001828, 0.00002742, -0.00007108, 0.00001118, -0.00016176, 0.00002888, 0.00020208, 0.00021188, -0.00003835, -0.00008787, 0.00004003, -0.00003163, -0.00000125, -0.00003140, -0.00000530, -0.00001807, 0.00001797, -0.00002620, 0.00001103, -0.00001402, 0.00002032, -0.00000317, -0.00001192, -0.00000509, 0.00001079, -0.00001287, 0.00001073, -0.00000753, 0.00000402, },
		{ 0.00000379, 0.00000048, -0.00000066, -0.00000517, 0.00000262, -0.00000774, 0.00000232, -0.00000559, 0.00001219, -0.00000264, 0.00000016, -0.00000797, 0.00000656, -0.00000377, 0.00000525, -0.00000190, -0.00000048, -0.00000077, -0.00000028, -0.00000616, 0.00000537, -0.00000402, 0.00000682, -0.00000611, 0.00000764, -0.00000426, 0.00000198, -0.00000521, 0.00000301, -0.00000274, 0.00000269, -0.00000498, 0.00000347, -0.00000375, 0.00000712, -0.00000615, 0.00000657, -0.00000525, 0.00000492, -0.00000509, 0.00000451, -0.00000367, 0.00000323, -0.00000667, 0.00000383, -0.00000659, 0.00000689, -0.00000623, 0.00000864, -0.00000547, 0.00000548, -0.00000590, 0.00000463, -0.00000407, 0.00000426, -0.00000514, 0.00000321, -0.00000628, 0.00000684, -0.00000703, 0.00000927, -0.00000593, 0.00000495, -0.00000668, 0.00000472, -0.00000443, 0.00000260, -0.00000474, 0.00000685, -0.00000479, 0.00000393, -0.00000596, 0.00000184, -0.00000215, 0.00000797, -0.00000548, 0.00000499, -0.00000339, -0.00000317, -0.00000378, 0.00000511, -0.00000214, 0.00000357, -0.00000964, 0.00000742, 0.00000014, -0.00000096, 0.00000211, -0.00001479, 0.00000483, 0.00000186, -0.00000652, 0.00000189, 0.00000622, -0.00000721, -0.00000381, 0.00000473, -0.00000321, 0.00000187, -0.00001171, -0.00001657, 0.00002556, -0.00006957, 0.00000984, -0.00016351, 0.00004298, 0.00018919, 0.00020584, -0.00003149, -0.00008529, 0.00003936, -0.00003099, -0.00000444, -0.00002894, -0.00000733, -0.00001671, 0.00001736, -0.00002585, 0.00001070, -0.00001358, 0.00002052, -0.00000409, -0.00001162, -0.00000444, 0.00001014, -0.00001228, 0.00001049, -0.00000761, },
		{ -0.00000763, 0.00000352, 0.00000066, -0.00000074, -0.00000502, 0.00000224, -0.00000718, 0.00000192, -0.00000489, 0.00001177, -0.00000290, 0.00000023, -0.00000786, 0.00000648, -0.00000373, 0.00000550, -0.00000219, -0.00000052, -0.00000071, -0.00000038, -0.00000574, 0.00000496, -0.00000362, 0.00000651, -0.00000597, 0.00000761, -0.00000465, 0.00000235, -0.00000532, 0.00000309, -0.00000277, 0.00000270, -0.00000490, 0.00000331, -0.00000355, 0.00000697, -0.00000606, 0.00000648, -0.00000535, 0.00000508, -0.00000521, 0.00000468, -0.00000377, 0.00000321, -0.00000659, 0.00000369, -0.00000639, 0.00000675, -0.00000613, 0.00000860, -0.00000559, 0.00000560, -0.00000613, 0.00000487, -0.00000413, 0.00000427, -0.00000509, 0.00000298, -0.00000583, 0.00000654, -0.00000695, 0.00000923, -0.00000597, 0.00000489, -0.00000674, 0.00000486, -0.00000441, 0.00000274, -0.00000484, 0.00000678, -0.00000462, 0.00000368, -0.00000591, 0.00000165, -0.00000160, 0.00000782, -0.00000587, 0.00000508, -0.00000333, -0.00000336, -0.00000330, 0.00000486, -0.00000188, 0.00000349, -0.00000994, 0.00000755, 0.00000001, -0.00000059, 0.00000141, -0.00001470, 0.00000565, 0.00000156, -0.00000656, 0.00000146, 0.00000699, -0.00000822, -0.00000291, 0.00000460, -0.00000397, 0.00000362, -0.00001407, -0.00001459, 0.00002320, -0.00006736, 0.00000753, -0.00016363, 0.00005536, 0.00017665, 0.00020046, -0.00002515, -0.00008234, 0.00003857, -0.00003050, -0.00000767, -0.00002632, -0.00000951, -0.00001512, 0.00001643, -0.00002518, 0.00001015, -0.00001294, 0.00002049, -0.00000491, -0.00001127, -0.00000382, 0.00000943, -0.00001160, 0.00001017, },
		{ 0.00000978, -0.00000762, 0.00000321, 0.00000088, -0.00000083, -0.00000487, 0.00000184, -0.00000659, 0.00000150, -0.00000417, 0.00001126, -0.00000310, 0.00000028, -0.00000771, 0.00000635, -0.00000364, 0.00000572, -0.00000249, -0.00000054, -0.00000068, -0.00000044, -0.00000534, 0.00000454, -0.00000320, 0.00000618, -0.00000578, 0.00000751, -0.00000498, 0.00000271, -0.00000542, 0.00000318, -0.00000281, 0.00000274, -0.00000485, 0.00000317, -0.00000337, 0.00000680, -0.00000596, 0.00000636, -0.00000541, 0.00000523, -0.00000531, 0.00000486, -0.00000388, 0.00000321, -0.00000653, 0.00000356, -0.00000620, 0.00000661, -0.00000601, 0.00000854, -0.00000569, 0.00000570, -0.00000635, 0.00000513, -0.00000422, 0.00000432, -0.00000509, 0.00000281, -0.00000541, 0.00000624, -0.00000687, 0.00000917, -0.00000600, 0.00000481, -0.00000678, 0.00000501, -0.00000441, 0.00000290, -0.00000498, 0.00000676, -0.00000450, 0.00000347, -0.00000589, 0.00000151, -0.00000107, 0.00000765, -0.00000624, 0.00000520, -0.00000332, -0.00000348, -0.00000286, 0.00000463, -0.00000167, 0.00000343, -0.00001023, 0.00000770, -0.00000017, -0.00000016, 0.00000061, -0.00001444, 0.00000631, 0.00000134, -0.00000667, 0.00000116, 0.00000759, -0.00000914, -0.00000199, 0.00000441, -0.00000468, 0.00000527, -0.00001641, -0.00001243, 0.00002041, -0.00006454, 0.00000435, -0.00016219, 0.00006598, 0.00016461, 0.00019561, -0.00001924, -0.00007908, 0.00003771, -0.00003023, -0.00001084, -0.00002361, -0.00001176, -0.00001336, 0.00001520, -0.00002422, 0.00000940, -0.00001211, 0.00002026, -0.00000563, -0.00001087, -0.00000325, 0.00000868, -0.00001083, },
		{ -0.00001000, 0.00000934, -0.00000758, 0.00000287, 0.00000112, -0.00000092, -0.00000471, 0.00000143, -0.00000598, 0.00000107, -0.00000342, 0.00001069, -0.00000324, 0.00000031, -0.00000753, 0.00000617, -0.00000350, 0.00000591, -0.00000279, -0.00000054, -0.00000068, -0.00000047, -0.00000498, 0.00000412, -0.00000277, 0.00000581, -0.00000555, 0.00000736, -0.00000527, 0.00000305, -0.00000552, 0.00000327, -0.00000287, 0.00000280, -0.00000483, 0.00000306, -0.00000320, 0.00000664, -0.00000584, 0.00000621, -0.00000545, 0.00000535, -0.00000540, 0.00000504, -0.00000401, 0.00000324, -0.00000650, 0.00000346, -0.00000601, 0.00000646, -0.00000588, 0.00000845, -0.00000576, 0.00000579, -0.00000655, 0.00000540, -0.00000433, 0.00000440, -0.00000513, 0.00000267, -0.00000502, 0.00000595, -0.00000677, 0.00000909, -0.00000602, 0.00000472, -0.00000680, 0.00000514, -0.00000443, 0.00000309, -0.00000515, 0.00000677, -0.00000442, 0.00000330, -0.00000589, 0.00000142, -0.00000058, 0.00000747, -0.00000659, 0.00000534, -0.00000335, -0.00000355, -0.00000245, 0.00000443, -0.00000149, 0.00000337, -0.00001051, 0.00000785, -0.00000037, 0.00000031, -0.00000026, -0.00001401, 0.00000681, 0.00000120, -0.00000684, 0.00000098, 0.00000804, -0.00000996, -0.00000105, 0.00000414, -0.00000531, 0.00000679, -0.00001865, -0.00001016, 0.00001729, -0.00006120, 0.00000041, -0.00015927, 0.00007482, 0.00015315, 0.00019122, -0.00001370, -0.00007555, 0.00003681, -0.00003023, -0.00001387, -0.00002089, -0.00001403, -0.00001149, 0.00001374, -0.00002300, 0.00000845, -0.00001110, 0.00001983, -0.00000624, -0.00001041, -0.00000274, 0.00000790, },
		{ 0.00000712, -0.00000913, 0.00000885, -0.00000753, 0.00000253, 0.00000138, -0.00000100, -0.00000455, 0.00000101, -0.00000536, 0.00000065, -0.00000267, 0.00001006, -0.00000335, 0.00000033, -0.00000734, 0.00000595, -0.00000330, 0.00000607, -0.00000310, -0.00000050, -0.00000071, -0.00000045, -0.00000464, 0.00000372, -0.00000234, 0.00000543, -0.00000529, 0.00000716, -0.00000550, 0.00000337, -0.00000560, 0.00000336, -0.00000293, 0.00000288, -0.00000483, 0.00000297, -0.00000304, 0.00000647, -0.00000571, 0.00000604, -0.00000546, 0.00000545, -0.00000547, 0.00000521, -0.00000415, 0.00000328, -0.00000649, 0.00000337, -0.00000585, 0.00000632, -0.00000574, 0.00000834, -0.00000580, 0.00000585, -0.00000674, 0.00000566, -0.00000446, 0.00000450, -0.00000520, 0.00000259, -0.00000466, 0.00000566, -0.00000666, 0.00000899, -0.00000602, 0.00000462, -0.00000679, 0.00000526, -0.00000444, 0.00000329, -0.00000535, 0.00000682, -0.00000438, 0.00000316, -0.00000593, 0.00000136, -0.00000011, 0.00000727, -0.00000691, 0.00000547, -0.00000341, -0.00000356, -0.00000207, 0.00000425, -0.00000135, 0.00000331, -0.00001076, 0.00000800, -0.00000058, 0.00000079, -0.00000117, -0.00001345, 0.00000717, 0.00000113, -0.00000705, 0.00000089, 0.00000832, -0.00001068, -0.00000008, 0.00000379, -0.00000584, 0.00000815, -0.00002074, -0.00000788, 0.00001394, -0.00005746, -0.00000419, -0.00015497, 0.00008189, 0.00014237, 0.00018717, -0.00000848, -0.00007179, 0.00003589, -0.00003056, -0.00001669, -0.00001825, -0.00001622, -0.00000959, 0.00001211, -0.00002157, 0.00000735, -0.00000994, 0.00001921, -0.00000676, -0.00000990, -0.00000230, },
		{ -0.00000194, 0.00000635, -0.00000822, 0.00000833, -0.00000748, 0.00000218, 0.00000163, -0.00000106, -0.00000441, 0.00000059, -0.00000474, 0.00000024, -0.00000192, 0.00000939, -0.00000342, 0.00000036, -0.00000713, 0.00000570, -0.00000307, 0.00000619, -0.00000342, -0.00000044, -0.00000077, -0.00000038, -0.00000435, 0.00000335, -0.00000193, 0.00000504, -0.00000499, 0.00000691, -0.00000569, 0.00000367, -0.00000568, 0.00000344, -0.00000300, 0.00000298, -0.00000487, 0.00000292, -0.00000291, 0.00000631, -0.00000558, 0.00000586, -0.00000544, 0.00000553, -0.00000553, 0.00000537, -0.00000429, 0.00000335, -0.00000650, 0.00000331, -0.00000569, 0.00000618, -0.00000559, 0.00000820, -0.00000582, 0.00000589, -0.00000691, 0.00000591, -0.00000459, 0.00000462, -0.00000531, 0.00000254, -0.00000435, 0.00000539, -0.00000655, 0.00000887, -0.00000601, 0.00000450, -0.00000675, 0.00000537, -0.00000445, 0.00000350, -0.00000555, 0.00000688, -0.00000437, 0.00000305, -0.00000598, 0.00000135, 0.00000032, 0.00000705, -0.00000720, 0.00000560, -0.00000350, -0.00000353, -0.00000171, 0.00000408, -0.00000123, 0.00000324, -0.00001098, 0.00000812, -0.00000080, 0.00000127, -0.00000211, -0.00001278, 0.00000739, 0.00000111, -0.00000729, 0.00000090, 0.00000845, -0.00001129, 0.00000091, 0.00000335, -0.00000625, 0.00000930, -0.00002261, -0.00000565, 0.00001045, -0.00005343, -0.00000930, -0.00014941, 0.00008724, 0.00013231, 0.00018342, -0.00000351, -0.00006784, 0.00003497, -0.00003123, -0.00001922, -0.00001575, -0.00001829, -0.00000770, 0.00001035, -0.00001997, 0.00000611, -0.00000864, 0.00001841, -0.00000717, -0.00000934, },
		{ -0.00000874, -0.00000168, 0.00000560, -0.00000730, 0.00000779, -0.00000743, 0.00000185, 0.00000188, -0.00000111, -0.00000428, 0.00000019, -0.00000414, -0.00000014, -0.00000119, 0.00000869, -0.00000346, 0.00000039, -0.00000691, 0.00000542, -0.00000279, 0.00000629, -0.00000373, -0.00000036, -0.00000087, -0.00000026, -0.00000410, 0.00000300, -0.00000154, 0.00000465, -0.00000469, 0.00000662, -0.00000583, 0.00000394, -0.00000574, 0.00000352, -0.00000308, 0.00000309, -0.00000493, 0.00000289, -0.00000281, 0.00000616, -0.00000545, 0.00000567, -0.00000540, 0.00000559, -0.00000556, 0.00000552, -0.00000444, 0.00000342, -0.00000653, 0.00000328, -0.00000556, 0.00000604, -0.00000544, 0.00000805, -0.00000582, 0.00000591, -0.00000705, 0.00000616, -0.00000473, 0.00000476, -0.00000544, 0.00000254, -0.00000407, 0.00000513, -0.00000644, 0.00000874, -0.00000598, 0.00000437, -0.00000667, 0.00000544, -0.00000445, 0.00000369, -0.00000577, 0.00000696, -0.00000440, 0.00000297, -0.00000606, 0.00000137, 0.00000072, 0.00000681, -0.00000744, 0.00000572, -0.00000361, -0.00000346, -0.00000137, 0.00000391, -0.00000112, 0.00000316, -0.00001117, 0.00000822, -0.00000100, 0.00000174, -0.00000306, -0.00001200, 0.00000750, 0.00000114, -0.00000755, 0.00000097, 0.00000845, -0.00001179, 0.00000190, 0.00000282, -0.00000653, 0.00001022, -0.00002423, -0.00000356, 0.00000692, -0.00004921, -0.00001481, -0.00014272, 0.00009091, 0.00012302, 0.00017990, 0.00000124, -0.00006371, 0.00003404, -0.00003229, -0.00002143, -0.00001345, -0.00002016, -0.00000590, 0.00000853, -0.00001823, 0.00000478, -0.00000724, 0.00001744, -0.00000747, },
		{ -0.00000768, -0.00000809, -0.00000150, 0.00000490, -0.00000638, 0.00000723, -0.00000740, 0.00000154, 0.00000211, -0.00000111, -0.00000419, -0.00000018, -0.00000356, -0.00000049, -0.00000049, 0.00000797, -0.00000348, 0.00000043, -0.00000670, 0.00000512, -0.00000248, 0.00000635, -0.00000404, -0.00000025, -0.00000100, -0.00000009, -0.00000391, 0.00000269, -0.00000118, 0.00000428, -0.00000437, 0.00000630, -0.00000592, 0.00000419, -0.00000579, 0.00000359, -0.00000316, 0.00000321, -0.00000502, 0.00000290, -0.00000273, 0.00000603, -0.00000533, 0.00000547, -0.00000535, 0.00000563, -0.00000558, 0.00000565, -0.00000459, 0.00000350, -0.00000658, 0.00000326, -0.00000544, 0.00000592, -0.00000529, 0.00000788, -0.00000579, 0.00000590, -0.00000716, 0.00000638, -0.00000486, 0.00000490, -0.00000559, 0.00000257, -0.00000384, 0.00000489, -0.00000632, 0.00000860, -0.00000594, 0.00000422, -0.00000656, 0.00000549, -0.00000443, 0.00000388, -0.00000598, 0.00000704, -0.00000444, 0.00000291, -0.00000616, 0.00000143, 0.00000108, 0.00000656, -0.00000765, 0.00000581, -0.00000373, -0.00000336, -0.00000104, 0.00000374, -0.00000103, 0.00000306, -0.00001130, 0.00000828, -0.00000118, 0.00000218, -0.00000400, -0.00001116, 0.00000750, 0.00000120, -0.00000781, 0.00000111, 0.00000831, -0.00001218, 0.00000291, 0.00000221, -0.00000667, 0.00001090, -0.00002555, -0.00000166, 0.00000343, -0.00004491, -0.00002057, -0.00013504, 0.00009298, 0.00011451, 0.00017657, 0.00000581, -0.00005941, 0.00003310, -0.00003372, -0.00002326, -0.00001141, -0.00002179, -0.00000424, 0.00000669, -0.00001641, 0.00000337, -0.00000575, 0.00001633, },
		{ 0.00001508, -0.00000778, -0.00000740, -0.00000142, 0.00000424, -0.00000547, 0.00000668, -0.00000740, 0.00000127, 0.00000232, -0.00000108, -0.00000414, -0.00000053, -0.00000300, -0.00000080, 0.00000017, 0.00000723, -0.00000349, 0.00000048, -0.00000649, 0.00000482, -0.00000215, 0.00000637, -0.00000435, -0.00000012, -0.00000116, 0.00000013, -0.00000377, 0.00000243, -0.00000085, 0.00000392, -0.00000406, 0.00000595, -0.00000598, 0.00000442, -0.00000583, 0.00000365, -0.00000323, 0.00000335, -0.00000514, 0.00000295, -0.00000269, 0.00000592, -0.00000522, 0.00000528, -0.00000527, 0.00000566, -0.00000558, 0.00000577, -0.00000473, 0.00000360, -0.00000664, 0.00000327, -0.00000535, 0.00000581, -0.00000514, 0.00000770, -0.00000574, 0.00000586, -0.00000724, 0.00000659, -0.00000499, 0.00000505, -0.00000576, 0.00000264, -0.00000364, 0.00000467, -0.00000620, 0.00000844, -0.00000588, 0.00000406, -0.00000642, 0.00000550, -0.00000438, 0.00000404, -0.00000618, 0.00000712, -0.00000451, 0.00000288, -0.00000626, 0.00000152, 0.00000142, 0.00000629, -0.00000780, 0.00000588, -0.00000385, -0.00000323, -0.00000071, 0.00000356, -0.00000095, 0.00000294, -0.00001138, 0.00000829, -0.00000132, 0.00000259, -0.00000490, -0.00001026, 0.00000742, 0.00000129, -0.00000805, 0.00000128, 0.00000804, -0.00001245, 0.00000391, 0.00000150, -0.00000668, 0.00001132, -0.00002653, -0.00000003, 0.00000007, -0.00004064, -0.00002645, -0.00012654, 0.00009356, 0.00010679, 0.00017340, 0.00001021, -0.00005495, 0.00003212, -0.00003553, -0.00002468, -0.00000966, -0.00002312, -0.00000278, 0.00000490, -0.00001454, 0.00000193, -0.00000420, },
		{ -0.00000261, 0.00001372, -0.00000778, -0.00000669, -0.00000144, 0.00000365, -0.00000458, 0.00000614, -0.00000743, 0.00000105, 0.00000248, -0.00000101, -0.00000413, -0.00000083, -0.00000248, -0.00000107, 0.00000078, 0.00000650, -0.00000349, 0.00000055, -0.00000631, 0.00000451, -0.00000178, 0.00000637, -0.00000466, 0.00000004, -0.00000135, 0.00000039, -0.00000368, 0.00000221, -0.00000058, 0.00000360, -0.00000376, 0.00000559, -0.00000600, 0.00000463, -0.00000586, 0.00000371, -0.00000331, 0.00000349, -0.00000528, 0.00000302, -0.00000268, 0.00000584, -0.00000512, 0.00000508, -0.00000519, 0.00000567, -0.00000556, 0.00000587, -0.00000487, 0.00000369, -0.00000672, 0.00000329, -0.00000527, 0.00000571, -0.00000500, 0.00000751, -0.00000567, 0.00000580, -0.00000728, 0.00000676, -0.00000510, 0.00000519, -0.00000595, 0.00000274, -0.00000348, 0.00000447, -0.00000609, 0.00000827, -0.00000582, 0.00000389, -0.00000624, 0.00000548, -0.00000431, 0.00000418, -0.00000636, 0.00000719, -0.00000458, 0.00000285, -0.00000638, 0.00000164, 0.00000171, 0.00000600, -0.00000791, 0.00000591, -0.00000396, -0.00000308, -0.00000039, 0.00000337, -0.00000087, 0.00000279, -0.00001141, 0.00000825, -0.00000143, 0.00000294, -0.00000576, -0.00000932, 0.00000726, 0.00000138, -0.00000828, 0.00000148, 0.00000766, -0.00001261, 0.00000490, 0.00000072, -0.00000654, 0.00001147, -0.00002716, 0.00000130, -0.00000309, -0.00003649, -0.00003233, -0.00011736, 0.00009274, 0.00009981, 0.00017037, 0.00001446, -0.00005033, 0.00003106, -0.00003768, -0.00002569, -0.00000825, -0.00002411, -0.00000156, 0.00000320, -0.00001267, 0.00000048, },
		{ -0.00000094, -0.00000102, 0.00001226, -0.00000769, -0.00000595, -0.00000155, 0.00000312, -0.00000373, 0.00000560, -0.00000749, 0.00000088, 0.00000261, -0.00000088, -0.00000417, -0.00000109, -0.00000200, -0.00000128, 0.00000134, 0.00000577, -0.00000349, 0.00000065, -0.00000614, 0.00000421, -0.00000141, 0.00000634, -0.00000495, 0.00000021, -0.00000156, 0.00000070, -0.00000365, 0.00000205, -0.00000035, 0.00000331, -0.00000348, 0.00000522, -0.00000599, 0.00000482, -0.00000587, 0.00000375, -0.00000338, 0.00000363, -0.00000544, 0.00000312, -0.00000270, 0.00000577, -0.00000504, 0.00000490, -0.00000509, 0.00000567, -0.00000552, 0.00000595, -0.00000499, 0.00000379, -0.00000680, 0.00000333, -0.00000522, 0.00000563, -0.00000487, 0.00000732, -0.00000558, 0.00000572, -0.00000729, 0.00000691, -0.00000519, 0.00000532, -0.00000613, 0.00000286, -0.00000335, 0.00000429, -0.00000599, 0.00000809, -0.00000575, 0.00000371, -0.00000604, 0.00000542, -0.00000421, 0.00000429, -0.00000652, 0.00000725, -0.00000466, 0.00000284, -0.00000650, 0.00000180, 0.00000198, 0.00000570, -0.00000798, 0.00000591, -0.00000407, -0.00000291, -0.00000007, 0.00000316, -0.00000080, 0.00000262, -0.00001137, 0.00000815, -0.00000149, 0.00000323, -0.00000655, -0.00000837, 0.00000703, 0.00000148, -0.00000847, 0.00000169, 0.00000718, -0.00001264, 0.00000586, -0.00000013, -0.00000628, 0.00001136, -0.00002741, 0.00000228, -0.00000598, -0.00003256, -0.00003807, -0.00010768, 0.00009066, 0.00009357, 0.00016749, 0.00001857, -0.00004553, 0.00002989, -0.00004015, -0.00002627, -0.00000720, -0.00002474, -0.00000062, 0.00000164, -0.00001084, },
		{ -0.00000909, -0.00000230, 0.00000055, 0.00001072, -0.00000751, -0.00000519, -0.00000174, 0.00000267, -0.00000291, 0.00000508, -0.00000759, 0.00000076, 0.00000268, -0.00000071, -0.00000425, -0.00000130, -0.00000157, -0.00000144, 0.00000183, 0.00000507, -0.00000349, 0.00000077, -0.00000601, 0.00000393, -0.00000102, 0.00000627, -0.00000524, 0.00000039, -0.00000180, 0.00000105, -0.00000367, 0.00000193, -0.00000018, 0.00000306, -0.00000322, 0.00000485, -0.00000596, 0.00000499, -0.00000587, 0.00000378, -0.00000344, 0.00000377, -0.00000561, 0.00000326, -0.00000275, 0.00000574, -0.00000499, 0.00000472, -0.00000499, 0.00000565, -0.00000547, 0.00000602, -0.00000511, 0.00000388, -0.00000690, 0.00000339, -0.00000518, 0.00000557, -0.00000475, 0.00000712, -0.00000548, 0.00000562, -0.00000727, 0.00000702, -0.00000525, 0.00000544, -0.00000632, 0.00000300, -0.00000326, 0.00000414, -0.00000589, 0.00000791, -0.00000568, 0.00000353, -0.00000580, 0.00000532, -0.00000407, 0.00000436, -0.00000665, 0.00000728, -0.00000473, 0.00000283, -0.00000661, 0.00000197, 0.00000220, 0.00000540, -0.00000799, 0.00000586, -0.00000416, -0.00000273, 0.00000025, 0.00000294, -0.00000072, 0.00000242, -0.00001128, 0.00000799, -0.00000150, 0.00000346, -0.00000727, -0.00000741, 0.00000677, 0.00000157, -0.00000861, 0.00000190, 0.00000660, -0.00001256, 0.00000678, -0.00000102, -0.00000589, 0.00001100, -0.00002730, 0.00000288, -0.00000855, -0.00002892, -0.00004355, -0.00009764, 0.00008744, 0.00008799, 0.00016474, 0.00002255, -0.00004055, 0.00002856, -0.00004288, -0.00002646, -0.00000650, -0.00002500, 0.00000001, 0.00000026, },
		{ -0.00000091, -0.00000745, -0.00000359, 0.00000208, 0.00000913, -0.00000723, -0.00000442, -0.00000202, 0.00000228, -0.00000213, 0.00000458, -0.00000772, 0.00000070, 0.00000271, -0.00000047, -0.00000439, -0.00000145, -0.00000118, -0.00000154, 0.00000226, 0.00000438, -0.00000349, 0.00000093, -0.00000590, 0.00000366, -0.00000063, 0.00000618, -0.00000552, 0.00000059, -0.00000205, 0.00000143, -0.00000375, 0.00000187, -0.00000006, 0.00000285, -0.00000300, 0.00000449, -0.00000590, 0.00000515, -0.00000586, 0.00000380, -0.00000350, 0.00000391, -0.00000579, 0.00000341, -0.00000283, 0.00000573, -0.00000496, 0.00000457, -0.00000490, 0.00000563, -0.00000540, 0.00000606, -0.00000521, 0.00000397, -0.00000699, 0.00000346, -0.00000516, 0.00000552, -0.00000465, 0.00000693, -0.00000537, 0.00000550, -0.00000722, 0.00000711, -0.00000529, 0.00000554, -0.00000650, 0.00000315, -0.00000320, 0.00000401, -0.00000580, 0.00000773, -0.00000560, 0.00000335, -0.00000554, 0.00000519, -0.00000390, 0.00000440, -0.00000674, 0.00000727, -0.00000479, 0.00000283, -0.00000673, 0.00000217, 0.00000239, 0.00000509, -0.00000796, 0.00000578, -0.00000423, -0.00000253, 0.00000057, 0.00000269, -0.00000063, 0.00000219, -0.00001112, 0.00000777, -0.00000146, 0.00000363, -0.00000791, -0.00000647, 0.00000646, 0.00000165, -0.00000870, 0.00000208, 0.00000595, -0.00001235, 0.00000765, -0.00000196, -0.00000541, 0.00001041, -0.00002681, 0.00000309, -0.00001075, -0.00002564, -0.00004868, -0.00008740, 0.00008323, 0.00008304, 0.00016216, 0.00002639, -0.00003536, 0.00002700, -0.00004582, -0.00002626, -0.00000616, -0.00002486, 0.00000030, },
		{ 0.00000025, -0.00000185, -0.00000595, -0.00000476, 0.00000354, 0.00000749, -0.00000688, -0.00000366, -0.00000238, 0.00000197, -0.00000140, 0.00000409, -0.00000789, 0.00000071, 0.00000268, -0.00000019, -0.00000459, -0.00000155, -0.00000085, -0.00000158, 0.00000262, 0.00000373, -0.00000351, 0.00000111, -0.00000583, 0.00000342, -0.00000023, 0.00000607, -0.00000578, 0.00000079, -0.00000232, 0.00000185, -0.00000387, 0.00000187, -0.00000001, 0.00000270, -0.00000282, 0.00000414, -0.00000583, 0.00000529, -0.00000584, 0.00000382, -0.00000354, 0.00000403, -0.00000598, 0.00000359, -0.00000295, 0.00000575, -0.00000496, 0.00000443, -0.00000480, 0.00000561, -0.00000532, 0.00000608, -0.00000530, 0.00000405, -0.00000709, 0.00000354, -0.00000516, 0.00000550, -0.00000457, 0.00000673, -0.00000526, 0.00000537, -0.00000714, 0.00000716, -0.00000531, 0.00000561, -0.00000666, 0.00000332, -0.00000317, 0.00000390, -0.00000572, 0.00000755, -0.00000553, 0.00000317, -0.00000525, 0.00000503, -0.00000370, 0.00000439, -0.00000679, 0.00000724, -0.00000484, 0.00000282, -0.00000684, 0.00000238, 0.00000254, 0.00000478, -0.00000789, 0.00000565, -0.00000428, -0.00000233, 0.00000090, 0.00000242, -0.00000054, 0.00000194, -0.00001091, 0.00000750, -0.00000137, 0.00000372, -0.00000846, -0.00000555, 0.00000612, 0.00000172, -0.00000873, 0.00000224, 0.00000523, -0.00001201, 0.00000845, -0.00000290, -0.00000485, 0.00000961, -0.00002597, 0.00000291, -0.00001256, -0.00002277, -0.00005335, -0.00007712, 0.00007815, 0.00007863, 0.00015975, 0.00003008, -0.00002995, 0.00002515, -0.00004888, -0.00002571, -0.00000616, -0.00002434, },
		{ -0.00002345, -0.00000015, -0.00000253, -0.00000461, -0.00000580, 0.00000491, 0.00000583, -0.00000644, -0.00000290, -0.00000280, 0.00000173, -0.00000071, 0.00000361, -0.00000809, 0.00000077, 0.00000259, 0.00000015, -0.00000484, -0.00000159, -0.00000056, -0.00000157, 0.00000291, 0.00000310, -0.00000353, 0.00000133, -0.00000579, 0.00000322, 0.00000015, 0.00000592, -0.00000602, 0.00000100, -0.00000260, 0.00000228, -0.00000403, 0.00000191, -0.00000001, 0.00000259, -0.00000267, 0.00000382, -0.00000575, 0.00000542, -0.00000582, 0.00000382, -0.00000357, 0.00000415, -0.00000617, 0.00000378, -0.00000309, 0.00000579, -0.00000498, 0.00000432, -0.00000471, 0.00000559, -0.00000523, 0.00000609, -0.00000537, 0.00000412, -0.00000719, 0.00000363, -0.00000517, 0.00000549, -0.00000450, 0.00000655, -0.00000514, 0.00000522, -0.00000703, 0.00000718, -0.00000529, 0.00000566, -0.00000681, 0.00000348, -0.00000316, 0.00000381, -0.00000566, 0.00000737, -0.00000546, 0.00000300, -0.00000496, 0.00000484, -0.00000347, 0.00000435, -0.00000680, 0.00000716, -0.00000488, 0.00000280, -0.00000694, 0.00000261, 0.00000264, 0.00000448, -0.00000779, 0.00000548, -0.00000431, -0.00000211, 0.00000122, 0.00000213, -0.00000045, 0.00000167, -0.00001064, 0.00000716, -0.00000122, 0.00000374, -0.00000892, -0.00000466, 0.00000576, 0.00000176, -0.00000868, 0.00000235, 0.00000445, -0.00001155, 0.00000917, -0.00000385, -0.00000424, 0.00000862, -0.00002480, 0.00000233, -0.00001397, -0.00002035, -0.00005748, -0.00006693, 0.00007236, 0.00007471, 0.00015753, 0.00003363, -0.00002431, 0.00002296, -0.00005200, -0.00002488, -0.00000648, },
		{ -0.00000707, -0.00002222, -0.00000089, -0.00000294, -0.00000346, -0.00000670, 0.00000618, 0.00000417, -0.00000593, -0.00000217, -0.00000327, 0.00000156, -0.00000007, 0.00000315, -0.00000831, 0.00000089, 0.00000245, 0.00000054, -0.00000513, -0.00000158, -0.00000033, -0.00000150, 0.00000313, 0.00000252, -0.00000357, 0.00000157, -0.00000580, 0.00000305, 0.00000052, 0.00000576, -0.00000624, 0.00000121, -0.00000288, 0.00000274, -0.00000423, 0.00000200, -0.00000006, 0.00000254, -0.00000258, 0.00000352, -0.00000567, 0.00000555, -0.00000578, 0.00000381, -0.00000359, 0.00000424, -0.00000635, 0.00000399, -0.00000325, 0.00000586, -0.00000504, 0.00000424, -0.00000464, 0.00000557, -0.00000513, 0.00000608, -0.00000543, 0.00000418, -0.00000728, 0.00000372, -0.00000520, 0.00000549, -0.00000445, 0.00000638, -0.00000502, 0.00000508, -0.00000690, 0.00000716, -0.00000525, 0.00000568, -0.00000694, 0.00000365, -0.00000317, 0.00000375, -0.00000561, 0.00000719, -0.00000539, 0.00000285, -0.00000465, 0.00000462, -0.00000320, 0.00000426, -0.00000676, 0.00000704, -0.00000488, 0.00000278, -0.00000702, 0.00000286, 0.00000271, 0.00000419, -0.00000765, 0.00000527, -0.00000432, -0.00000189, 0.00000153, 0.00000183, -0.00000035, 0.00000137, -0.00001031, 0.00000677, -0.00000103, 0.00000370, -0.00000928, -0.00000382, 0.00000539, 0.00000177, -0.00000857, 0.00000241, 0.00000363, -0.00001097, 0.00000980, -0.00000476, -0.00000359, 0.00000749, -0.00002334, 0.00000139, -0.00001495, -0.00001841, -0.00006102, -0.00005696, 0.00006598, 0.00007118, 0.00015554, 0.00003702, -0.00001841, 0.00002035, -0.00005509, -0.00002381, },
		{ -0.00002257, -0.00000789, -0.00002067, -0.00000195, -0.00000308, -0.00000251, -0.00000743, 0.00000732, 0.00000253, -0.00000534, -0.00000146, -0.00000378, 0.00000144, 0.00000053, 0.00000269, -0.00000855, 0.00000106, 0.00000227, 0.00000098, -0.00000547, -0.00000151, -0.00000015, -0.00000138, 0.00000328, 0.00000197, -0.00000362, 0.00000184, -0.00000583, 0.00000292, 0.00000088, 0.00000558, -0.00000644, 0.00000141, -0.00000315, 0.00000320, -0.00000445, 0.00000212, -0.00000017, 0.00000253, -0.00000253, 0.00000325, -0.00000559, 0.00000567, -0.00000574, 0.00000380, -0.00000359, 0.00000432, -0.00000653, 0.00000420, -0.00000343, 0.00000595, -0.00000512, 0.00000418, -0.00000457, 0.00000556, -0.00000503, 0.00000605, -0.00000547, 0.00000422, -0.00000736, 0.00000381, -0.00000523, 0.00000551, -0.00000442, 0.00000622, -0.00000491, 0.00000493, -0.00000674, 0.00000712, -0.00000517, 0.00000567, -0.00000705, 0.00000380, -0.00000319, 0.00000370, -0.00000557, 0.00000702, -0.00000534, 0.00000272, -0.00000434, 0.00000439, -0.00000292, 0.00000414, -0.00000668, 0.00000688, -0.00000487, 0.00000274, -0.00000709, 0.00000311, 0.00000274, 0.00000391, -0.00000748, 0.00000503, -0.00000430, -0.00000165, 0.00000184, 0.00000150, -0.00000025, 0.00000106, -0.00000995, 0.00000633, -0.00000079, 0.00000359, -0.00000955, -0.00000302, 0.00000501, 0.00000177, -0.00000839, 0.00000240, 0.00000278, -0.00001026, 0.00001031, -0.00000563, -0.00000295, 0.00000624, -0.00002162, 0.00000009, -0.00001554, -0.00001696, -0.00006390, -0.00004733, 0.00005915, 0.00006799, 0.00015381, 0.00004027, -0.00001225, 0.00001726, -0.00005806, },
		{ -0.00006082, -0.00002124, -0.00000889, -0.00001885, -0.00000331, -0.00000295, -0.00000176, -0.00000800, 0.00000832, 0.00000091, -0.00000470, -0.00000080, -0.00000432, 0.00000138, 0.00000108, 0.00000223, -0.00000880, 0.00000127, 0.00000204, 0.00000146, -0.00000586, -0.00000139, -0.00000002, -0.00000122, 0.00000336, 0.00000146, -0.00000369, 0.00000213, -0.00000591, 0.00000284, 0.00000122, 0.00000537, -0.00000662, 0.00000161, -0.00000342, 0.00000367, -0.00000470, 0.00000229, -0.00000033, 0.00000258, -0.00000253, 0.00000302, -0.00000551, 0.00000578, -0.00000570, 0.00000378, -0.00000358, 0.00000438, -0.00000669, 0.00000442, -0.00000362, 0.00000606, -0.00000522, 0.00000415, -0.00000453, 0.00000556, -0.00000493, 0.00000602, -0.00000549, 0.00000425, -0.00000742, 0.00000390, -0.00000527, 0.00000555, -0.00000442, 0.00000608, -0.00000481, 0.00000478, -0.00000658, 0.00000705, -0.00000507, 0.00000564, -0.00000712, 0.00000395, -0.00000323, 0.00000367, -0.00000554, 0.00000685, -0.00000530, 0.00000261, -0.00000404, 0.00000414, -0.00000261, 0.00000399, -0.00000657, 0.00000667, -0.00000482, 0.00000269, -0.00000713, 0.00000337, 0.00000273, 0.00000366, -0.00000730, 0.00000475, -0.00000427, -0.00000140, 0.00000214, 0.00000117, -0.00000016, 0.00000074, -0.00000954, 0.00000585, -0.00000051, 0.00000343, -0.00000972, -0.00000227, 0.00000463, 0.00000174, -0.00000813, 0.00000233, 0.00000190, -0.00000943, 0.00001069, -0.00000642, -0.00000234, 0.00000493, -0.00001969, -0.00000151, -0.00001573, -0.00001600, -0.00006611, -0.00003813, 0.00005199, 0.00006505, 0.00015234, 0.00004336, -0.00000583, 0.00001364, },
		{ 0.00000944, -0.00006329, -0.00001988, -0.00001001, -0.00001682, -0.00000493, -0.00000257, -0.00000122, -0.00000838, 0.00000918, -0.00000066, -0.00000400, -0.00000018, -0.00000488, 0.00000136, 0.00000160, 0.00000176, -0.00000905, 0.00000153, 0.00000177, 0.00000198, -0.00000628, -0.00000123, 0.00000007, -0.00000100, 0.00000338, 0.00000099, -0.00000376, 0.00000245, -0.00000602, 0.00000279, 0.00000152, 0.00000516, -0.00000678, 0.00000180, -0.00000368, 0.00000413, -0.00000497, 0.00000248, -0.00000053, 0.00000267, -0.00000258, 0.00000283, -0.00000543, 0.00000590, -0.00000566, 0.00000376, -0.00000356, 0.00000441, -0.00000684, 0.00000463, -0.00000382, 0.00000618, -0.00000535, 0.00000414, -0.00000450, 0.00000556, -0.00000483, 0.00000597, -0.00000550, 0.00000425, -0.00000748, 0.00000399, -0.00000531, 0.00000560, -0.00000443, 0.00000595, -0.00000471, 0.00000464, -0.00000639, 0.00000696, -0.00000494, 0.00000557, -0.00000717, 0.00000408, -0.00000328, 0.00000366, -0.00000553, 0.00000669, -0.00000527, 0.00000253, -0.00000375, 0.00000389, -0.00000230, 0.00000381, -0.00000641, 0.00000641, -0.00000474, 0.00000261, -0.00000716, 0.00000364, 0.00000267, 0.00000343, -0.00000710, 0.00000446, -0.00000422, -0.00000113, 0.00000242, 0.00000083, -0.00000007, 0.00000042, -0.00000910, 0.00000533, -0.00000021, 0.00000322, -0.00000981, -0.00000157, 0.00000423, 0.00000169, -0.00000781, 0.00000219, 0.00000102, -0.00000848, 0.00001094, -0.00000712, -0.00000179, 0.00000360, -0.00001761, -0.00000337, -0.00001558, -0.00001551, -0.00006762, -0.00002944, 0.00004462, 0.00006229, 0.00015118, 0.00004631, 0.00000085, },
		{ 0.00000776, 0.00000461, -0.00006537, -0.00001858, -0.00001119, -0.00001462, -0.00000678, -0.00000196, -0.00000088, -0.00000859, 0.00000988, -0.00000218, -0.00000324, 0.00000039, -0.00000543, 0.00000138, 0.00000209, 0.00000128, -0.00000928, 0.00000180, 0.00000147, 0.00000252, -0.00000673, -0.00000102, 0.00000012, -0.00000076, 0.00000333, 0.00000056, -0.00000384, 0.00000278, -0.00000615, 0.00000280, 0.00000180, 0.00000492, -0.00000690, 0.00000198, -0.00000392, 0.00000458, -0.00000523, 0.00000269, -0.00000077, 0.00000280, -0.00000268, 0.00000268, -0.00000537, 0.00000602, -0.00000562, 0.00000373, -0.00000353, 0.00000442, -0.00000696, 0.00000483, -0.00000403, 0.00000631, -0.00000550, 0.00000417, -0.00000449, 0.00000558, -0.00000473, 0.00000591, -0.00000549, 0.00000425, -0.00000752, 0.00000407, -0.00000536, 0.00000565, -0.00000446, 0.00000584, -0.00000463, 0.00000451, -0.00000621, 0.00000685, -0.00000479, 0.00000547, -0.00000719, 0.00000419, -0.00000334, 0.00000365, -0.00000553, 0.00000654, -0.00000526, 0.00000249, -0.00000347, 0.00000364, -0.00000197, 0.00000360, -0.00000622, 0.00000612, -0.00000464, 0.00000252, -0.00000716, 0.00000390, 0.00000257, 0.00000323, -0.00000690, 0.00000414, -0.00000416, -0.00000085, 0.00000267, 0.00000049, 0.00000001, 0.00000010, -0.00000864, 0.00000478, 0.00000012, 0.00000297, -0.00000983, -0.00000092, 0.00000383, 0.00000163, -0.00000742, 0.00000198, 0.00000013, -0.00000742, 0.00001105, -0.00000771, -0.00000133, 0.00000228, -0.00001542, -0.00000544, -0.00001511, -0.00001545, -0.00006843, -0.00002133, 0.00003715, 0.00005964, 0.00015033, 0.00004913, },
		{ 0.00005183, 0.00001490, -0.00000087, -0.00006699, -0.00001742, -0.00001235, -0.00001232, -0.00000881, -0.00000113, -0.00000074, -0.00000861, 0.00001042, -0.00000362, -0.00000244, 0.00000089, -0.00000598, 0.00000142, 0.00000255, 0.00000078, -0.00000948, 0.00000210, 0.00000114, 0.00000308, -0.00000720, -0.00000078, 0.00000013, -0.00000047, 0.00000323, 0.00000016, -0.00000392, 0.00000312, -0.00000632, 0.00000284, 0.00000205, 0.00000468, -0.00000701, 0.00000214, -0.00000414, 0.00000501, -0.00000550, 0.00000292, -0.00000104, 0.00000297, -0.00000282, 0.00000257, -0.00000532, 0.00000614, -0.00000559, 0.00000370, -0.00000349, 0.00000440, -0.00000706, 0.00000502, -0.00000423, 0.00000645, -0.00000566, 0.00000422, -0.00000449, 0.00000562, -0.00000465, 0.00000584, -0.00000547, 0.00000422, -0.00000755, 0.00000413, -0.00000540, 0.00000571, -0.00000451, 0.00000575, -0.00000457, 0.00000439, -0.00000602, 0.00000672, -0.00000462, 0.00000534, -0.00000718, 0.00000429, -0.00000339, 0.00000366, -0.00000554, 0.00000640, -0.00000526, 0.00000247, -0.00000322, 0.00000340, -0.00000165, 0.00000338, -0.00000600, 0.00000578, -0.00000450, 0.00000241, -0.00000713, 0.00000417, 0.00000244, 0.00000306, -0.00000669, 0.00000380, -0.00000409, -0.00000054, 0.00000290, 0.00000016, 0.00000007, -0.00000021, -0.00000816, 0.00000422, 0.00000046, 0.00000269, -0.00000977, -0.00000032, 0.00000342, 0.00000156, -0.00000698, 0.00000169, -0.00000075, -0.00000625, 0.00001100, -0.00000816, -0.00000098, 0.00000103, -0.00001318, -0.00000767, -0.00001437, -0.00001578, -0.00006858, -0.00001384, 0.00002965, 0.00005705, 0.00014982, },
		},
	}
	;
}
